#include "icu-collate.hpp"

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2011-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  appendable.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010dec07
*   created by: Markus W. Scherer
*/


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2011-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  appendable.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010dec07
*   created by: Markus W. Scherer
*/


/**
 * \file
 * \brief C++ API: Appendable class: Sink for Unicode code points and 16-bit code units (char16_ts).
 */



#if U_SHOW_CPLUSPLUS_API



U_NAMESPACE_BEGIN

class UnicodeString;

/**
 * Base class for objects to which Unicode characters and strings can be appended.
 * Combines elements of Java Appendable and ICU4C ByteSink.
 *
 * This class can be used in APIs where it does not matter whether the actual destination is
 * a UnicodeString, a char16_t[] array, a UnicodeSet, or any other object
 * that receives and processes characters and/or strings.
 *
 * Implementation classes must implement at least appendCodeUnit(char16_t).
 * The base class provides default implementations for the other methods.
 *
 * The methods do not take UErrorCode parameters.
 * If an error occurs (e.g., out-of-memory),
 * in addition to returning FALSE from failing operations,
 * the implementation must prevent unexpected behavior (e.g., crashes)
 * from further calls and should make the error condition available separately
 * (e.g., store a UErrorCode, make/keep a UnicodeString bogus).
 * @stable ICU 4.8
 */
class U_COMMON_API Appendable : public UObject {
public:
    /**
     * Destructor.
     * @stable ICU 4.8
     */
    ~Appendable();

    /**
     * Appends a 16-bit code unit.
     * @param c code unit
     * @return TRUE if the operation succeeded
     * @stable ICU 4.8
     */
    virtual UBool appendCodeUnit(char16_t c) = 0;

    /**
     * Appends a code point.
     * The default implementation calls appendCodeUnit(char16_t) once or twice.
     * @param c code point 0..0x10ffff
     * @return TRUE if the operation succeeded
     * @stable ICU 4.8
     */
    virtual UBool appendCodePoint(UChar32 c);

    /**
     * Appends a string.
     * The default implementation calls appendCodeUnit(char16_t) for each code unit.
     * @param s string, must not be NULL if length!=0
     * @param length string length, or -1 if NUL-terminated
     * @return TRUE if the operation succeeded
     * @stable ICU 4.8
     */
    virtual UBool appendString(const char16_t *s, int32_t length);

    /**
     * Tells the object that the caller is going to append roughly
     * appendCapacity char16_ts. A subclass might use this to pre-allocate
     * a larger buffer if necessary.
     * The default implementation does nothing. (It always returns TRUE.)
     * @param appendCapacity estimated number of char16_ts that will be appended
     * @return TRUE if the operation succeeded
     * @stable ICU 4.8
     */
    virtual UBool reserveAppendCapacity(int32_t appendCapacity);

    /**
     * Returns a writable buffer for appending and writes the buffer's capacity to
     * *resultCapacity. Guarantees *resultCapacity>=minCapacity.
     * May return a pointer to the caller-owned scratch buffer which must have
     * scratchCapacity>=minCapacity.
     * The returned buffer is only valid until the next operation
     * on this Appendable.
     *
     * After writing at most *resultCapacity char16_ts, call appendString() with the
     * pointer returned from this function and the number of char16_ts written.
     * Many appendString() implementations will avoid copying char16_ts if this function
     * returned an internal buffer.
     *
     * Partial usage example:
     * \code
     *  int32_t capacity;
     *  char16_t* buffer = app.getAppendBuffer(..., &capacity);
     *  ... Write n char16_ts into buffer, with n <= capacity.
     *  app.appendString(buffer, n);
     * \endcode
     * In many implementations, that call to append will avoid copying char16_ts.
     *
     * If the Appendable allocates or reallocates an internal buffer, it should use
     * the desiredCapacityHint if appropriate.
     * If a caller cannot provide a reasonable guess at the desired capacity,
     * it should pass desiredCapacityHint=0.
     *
     * If a non-scratch buffer is returned, the caller may only pass
     * a prefix to it to appendString().
     * That is, it is not correct to pass an interior pointer to appendString().
     *
     * The default implementation always returns the scratch buffer.
     *
     * @param minCapacity required minimum capacity of the returned buffer;
     *                    must be non-negative
     * @param desiredCapacityHint desired capacity of the returned buffer;
     *                            must be non-negative
     * @param scratch default caller-owned buffer
     * @param scratchCapacity capacity of the scratch buffer
     * @param resultCapacity pointer to an integer which will be set to the
     *                       capacity of the returned buffer
     * @return a buffer with *resultCapacity>=minCapacity
     * @stable ICU 4.8
     */
    virtual char16_t *getAppendBuffer(int32_t minCapacity,
                                   int32_t desiredCapacityHint,
                                   char16_t *scratch, int32_t scratchCapacity,
                                   int32_t *resultCapacity);
};

/**
 * An Appendable implementation which writes to a UnicodeString.
 *
 * This class is not intended for public subclassing.
 * @stable ICU 4.8
 */
class U_COMMON_API UnicodeStringAppendable : public Appendable {
public:
    /**
     * Aliases the UnicodeString (keeps its reference) for writing.
     * @param s The UnicodeString to which this Appendable will write.
     * @stable ICU 4.8
     */
    explicit UnicodeStringAppendable(UnicodeString &s) : str(s) {}

    /**
     * Destructor.
     * @stable ICU 4.8
     */
    ~UnicodeStringAppendable();

    /**
     * Appends a 16-bit code unit to the string.
     * @param c code unit
     * @return TRUE if the operation succeeded
     * @stable ICU 4.8
     */
    virtual UBool appendCodeUnit(char16_t c);

    /**
     * Appends a code point to the string.
     * @param c code point 0..0x10ffff
     * @return TRUE if the operation succeeded
     * @stable ICU 4.8
     */
    virtual UBool appendCodePoint(UChar32 c);

    /**
     * Appends a string to the UnicodeString.
     * @param s string, must not be NULL if length!=0
     * @param length string length, or -1 if NUL-terminated
     * @return TRUE if the operation succeeded
     * @stable ICU 4.8
     */
    virtual UBool appendString(const char16_t *s, int32_t length);

    /**
     * Tells the UnicodeString that the caller is going to append roughly
     * appendCapacity char16_ts.
     * @param appendCapacity estimated number of char16_ts that will be appended
     * @return TRUE if the operation succeeded
     * @stable ICU 4.8
     */
    virtual UBool reserveAppendCapacity(int32_t appendCapacity);

    /**
     * Returns a writable buffer for appending and writes the buffer's capacity to
     * *resultCapacity. Guarantees *resultCapacity>=minCapacity.
     * May return a pointer to the caller-owned scratch buffer which must have
     * scratchCapacity>=minCapacity.
     * The returned buffer is only valid until the next write operation
     * on the UnicodeString.
     *
     * For details see Appendable::getAppendBuffer().
     *
     * @param minCapacity required minimum capacity of the returned buffer;
     *                    must be non-negative
     * @param desiredCapacityHint desired capacity of the returned buffer;
     *                            must be non-negative
     * @param scratch default caller-owned buffer
     * @param scratchCapacity capacity of the scratch buffer
     * @param resultCapacity pointer to an integer which will be set to the
     *                       capacity of the returned buffer
     * @return a buffer with *resultCapacity>=minCapacity
     * @stable ICU 4.8
     */
    virtual char16_t *getAppendBuffer(int32_t minCapacity,
                                   int32_t desiredCapacityHint,
                                   char16_t *scratch, int32_t scratchCapacity,
                                   int32_t *resultCapacity);

private:
    UnicodeString &str;
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */




U_NAMESPACE_BEGIN

Appendable::~Appendable() {}

UBool
Appendable::appendCodePoint(UChar32 c) {
    if(c<=0xffff) {
        return appendCodeUnit((UChar)c);
    } else {
        return appendCodeUnit(U16_LEAD(c)) && appendCodeUnit(U16_TRAIL(c));
    }
}

UBool
Appendable::appendString(const UChar *s, int32_t length) {
    if(length<0) {
        UChar c;
        while((c=*s++)!=0) {
            if(!appendCodeUnit(c)) {
                return FALSE;
            }
        }
    } else if(length>0) {
        const UChar *limit=s+length;
        do {
            if(!appendCodeUnit(*s++)) {
                return FALSE;
            }
        } while(s<limit);
    }
    return TRUE;
}

UBool
Appendable::reserveAppendCapacity(int32_t /*appendCapacity*/) {
    return TRUE;
}

UChar *
Appendable::getAppendBuffer(int32_t minCapacity,
                            int32_t /*desiredCapacityHint*/,
                            UChar *scratch, int32_t scratchCapacity,
                            int32_t *resultCapacity) {
    if(minCapacity<1 || scratchCapacity<minCapacity) {
        *resultCapacity=0;
        return NULL;
    }
    *resultCapacity=scratchCapacity;
    return scratch;
}

// UnicodeStringAppendable is implemented in unistr.cpp.

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2007-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  bmpset.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2007jan29
*   created by: Markus W. Scherer
*/





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File CMEMORY.H
*
*  Contains stdlib.h/string.h memory functions
*
* @author       Bertrand A. Damiba
*
* Modification History:
*
*   Date        Name        Description
*   6/20/98     Bertrand    Created.
*  05/03/99     stephen     Changed from functions to macros.
*
******************************************************************************
*/




#include <stddef.h>
#include <string.h>


#if U_DEBUG && defined(UPRV_MALLOC_COUNT)
#include <stdio.h>
#endif


#define uprv_memcpy(dst, src, size) U_STANDARD_CPP_NAMESPACE memcpy(dst, src, size)
#define uprv_memmove(dst, src, size) U_STANDARD_CPP_NAMESPACE memmove(dst, src, size)

/**
 * \def UPRV_LENGTHOF
 * Convenience macro to determine the length of a fixed array at compile-time.
 * @param array A fixed length array
 * @return The length of the array, in elements
 * @internal
 */
#define UPRV_LENGTHOF(array) (int32_t)(sizeof(array)/sizeof((array)[0]))
#define uprv_memset(buffer, mark, size) U_STANDARD_CPP_NAMESPACE memset(buffer, mark, size)
#define uprv_memcmp(buffer1, buffer2, size) U_STANDARD_CPP_NAMESPACE memcmp(buffer1, buffer2,size)
#define uprv_memchr(ptr, value, num) U_STANDARD_CPP_NAMESPACE memchr(ptr, value, num)

U_CAPI void * U_EXPORT2
uprv_malloc(size_t s) U_MALLOC_ATTR U_ALLOC_SIZE_ATTR(1);

U_CAPI void * U_EXPORT2
uprv_realloc(void *mem, size_t size) U_ALLOC_SIZE_ATTR(2);

U_CAPI void U_EXPORT2
uprv_free(void *mem);

U_CAPI void * U_EXPORT2
uprv_calloc(size_t num, size_t size) U_MALLOC_ATTR U_ALLOC_SIZE_ATTR2(1,2);

/**
 * Get the least significant bits of a pointer (a memory address).
 * For example, with a mask of 3, the macro gets the 2 least significant bits,
 * which will be 0 if the pointer is 32-bit (4-byte) aligned.
 *
 * uintptr_t is the most appropriate integer type to cast to.
 */
#define U_POINTER_MASK_LSB(ptr, mask) ((uintptr_t)(ptr) & (mask))

/**
 * Create & return an instance of "type" in statically allocated storage.
 * e.g.
 *    static std::mutex *myMutex = STATIC_NEW(std::mutex);
 * To destroy an object created in this way, invoke the destructor explicitly, e.g.
 *    myMutex->~mutex();
 * DO NOT use delete.
 * DO NOT use with class UMutex, which has specific support for static instances.
 *
 * STATIC_NEW is intended for use when
 *   - We want a static (or global) object.
 *   - We don't want it to ever be destructed, or to explicitly control destruction,
 *     to avoid use-after-destruction problems.
 *   - We want to avoid an ordinary heap allocated object,
 *     to avoid the possibility of memory allocation failures, and
 *     to avoid memory leak reports, from valgrind, for example.
 * This is defined as a macro rather than a template function because each invocation
 * must define distinct static storage for the object being returned.
 */
#define STATIC_NEW(type) [] () { \
    alignas(type) static char storage[sizeof(type)]; \
    return new(storage) type();} ()

/**
  *  Heap clean up function, called from u_cleanup()
  *    Clears any user heap functions from u_setMemoryFunctions()
  *    Does NOT deallocate any remaining allocated memory.
  */
U_CFUNC UBool 
cmemory_cleanup(void);

/**
 * A function called by <TT>uhash_remove</TT>,
 * <TT>uhash_close</TT>, or <TT>uhash_put</TT> to delete
 * an existing key or value.
 * @param obj A key or value stored in a hashtable
 * @see uprv_deleteUObject
 */
typedef void U_CALLCONV UObjectDeleter(void* obj);

/**
 * Deleter for UObject instances.
 * Works for all subclasses of UObject because it has a virtual destructor.
 */
U_CAPI void U_EXPORT2
uprv_deleteUObject(void *obj);

#ifdef __cplusplus

#include <utility>


U_NAMESPACE_BEGIN

/**
 * "Smart pointer" class, deletes memory via uprv_free().
 * For most methods see the LocalPointerBase base class.
 * Adds operator[] for array item access.
 *
 * @see LocalPointerBase
 */
template<typename T>
class LocalMemory : public LocalPointerBase<T> {
public:
    using LocalPointerBase<T>::operator*;
    using LocalPointerBase<T>::operator->;
    /**
     * Constructor takes ownership.
     * @param p simple pointer to an array of T items that is adopted
     */
    explicit LocalMemory(T *p=NULL) : LocalPointerBase<T>(p) {}
    /**
     * Move constructor, leaves src with isNull().
     * @param src source smart pointer
     */
    LocalMemory(LocalMemory<T> &&src) U_NOEXCEPT : LocalPointerBase<T>(src.ptr) {
        src.ptr=NULL;
    }
    /**
     * Destructor deletes the memory it owns.
     */
    ~LocalMemory() {
        uprv_free(LocalPointerBase<T>::ptr);
    }
    /**
     * Move assignment operator, leaves src with isNull().
     * The behavior is undefined if *this and src are the same object.
     * @param src source smart pointer
     * @return *this
     */
    LocalMemory<T> &operator=(LocalMemory<T> &&src) U_NOEXCEPT {
        uprv_free(LocalPointerBase<T>::ptr);
        LocalPointerBase<T>::ptr=src.ptr;
        src.ptr=NULL;
        return *this;
    }
    /**
     * Swap pointers.
     * @param other other smart pointer
     */
    void swap(LocalMemory<T> &other) U_NOEXCEPT {
        T *temp=LocalPointerBase<T>::ptr;
        LocalPointerBase<T>::ptr=other.ptr;
        other.ptr=temp;
    }
    /**
     * Non-member LocalMemory swap function.
     * @param p1 will get p2's pointer
     * @param p2 will get p1's pointer
     */
    friend inline void swap(LocalMemory<T> &p1, LocalMemory<T> &p2) U_NOEXCEPT {
        p1.swap(p2);
    }
    /**
     * Deletes the array it owns,
     * and adopts (takes ownership of) the one passed in.
     * @param p simple pointer to an array of T items that is adopted
     */
    void adoptInstead(T *p) {
        uprv_free(LocalPointerBase<T>::ptr);
        LocalPointerBase<T>::ptr=p;
    }
    /**
     * Deletes the array it owns, allocates a new one and reset its bytes to 0.
     * Returns the new array pointer.
     * If the allocation fails, then the current array is unchanged and
     * this method returns NULL.
     * @param newCapacity must be >0
     * @return the allocated array pointer, or NULL if the allocation failed
     */
    inline T *allocateInsteadAndReset(int32_t newCapacity=1);
    /**
     * Deletes the array it owns and allocates a new one, copying length T items.
     * Returns the new array pointer.
     * If the allocation fails, then the current array is unchanged and
     * this method returns NULL.
     * @param newCapacity must be >0
     * @param length number of T items to be copied from the old array to the new one;
     *               must be no more than the capacity of the old array,
     *               which the caller must track because the LocalMemory does not track it
     * @return the allocated array pointer, or NULL if the allocation failed
     */
    inline T *allocateInsteadAndCopy(int32_t newCapacity=1, int32_t length=0);
    /**
     * Array item access (writable).
     * No index bounds check.
     * @param i array index
     * @return reference to the array item
     */
    T &operator[](ptrdiff_t i) const { return LocalPointerBase<T>::ptr[i]; }
};

template<typename T>
inline T *LocalMemory<T>::allocateInsteadAndReset(int32_t newCapacity) {
    if(newCapacity>0) {
        T *p=(T *)uprv_malloc(newCapacity*sizeof(T));
        if(p!=NULL) {
            uprv_memset(p, 0, newCapacity*sizeof(T));
            uprv_free(LocalPointerBase<T>::ptr);
            LocalPointerBase<T>::ptr=p;
        }
        return p;
    } else {
        return NULL;
    }
}


template<typename T>
inline T *LocalMemory<T>::allocateInsteadAndCopy(int32_t newCapacity, int32_t length) {
    if(newCapacity>0) {
        T *p=(T *)uprv_malloc(newCapacity*sizeof(T));
        if(p!=NULL) {
            if(length>0) {
                if(length>newCapacity) {
                    length=newCapacity;
                }
                uprv_memcpy(p, LocalPointerBase<T>::ptr, (size_t)length*sizeof(T));
            }
            uprv_free(LocalPointerBase<T>::ptr);
            LocalPointerBase<T>::ptr=p;
        }
        return p;
    } else {
        return NULL;
    }
}

/**
 * Simple array/buffer management class using uprv_malloc() and uprv_free().
 * Provides an internal array with fixed capacity. Can alias another array
 * or allocate one.
 *
 * The array address is properly aligned for type T. It might not be properly
 * aligned for types larger than T (or larger than the largest subtype of T).
 *
 * Unlike LocalMemory and LocalArray, this class never adopts
 * (takes ownership of) another array.
 *
 * WARNING: MaybeStackArray only works with primitive (plain-old data) types.
 * It does NOT know how to call a destructor! If you work with classes with
 * destructors, consider LocalArray in localpointer.h or MemoryPool.
 */
template<typename T, int32_t stackCapacity>
class MaybeStackArray {
public:
    // No heap allocation. Use only on the stack.
    static void* U_EXPORT2 operator new(size_t) U_NOEXCEPT = delete;
    static void* U_EXPORT2 operator new[](size_t) U_NOEXCEPT = delete;
#if U_HAVE_PLACEMENT_NEW
    static void* U_EXPORT2 operator new(size_t, void*) U_NOEXCEPT = delete;
#endif

    /**
     * Default constructor initializes with internal T[stackCapacity] buffer.
     */
    MaybeStackArray() : ptr(stackArray), capacity(stackCapacity), needToRelease(FALSE) {}
    /**
     * Automatically allocates the heap array if the argument is larger than the stack capacity.
     * Intended for use when an approximate capacity is known at compile time but the true
     * capacity is not known until runtime.
     */
    MaybeStackArray(int32_t newCapacity) : MaybeStackArray() {
        if (capacity < newCapacity) { resize(newCapacity); }
    }
    /**
     * Destructor deletes the array (if owned).
     */
    ~MaybeStackArray() { releaseArray(); }
    /**
     * Move constructor: transfers ownership or copies the stack array.
     */
    MaybeStackArray(MaybeStackArray<T, stackCapacity> &&src) U_NOEXCEPT;
    /**
     * Move assignment: transfers ownership or copies the stack array.
     */
    MaybeStackArray<T, stackCapacity> &operator=(MaybeStackArray<T, stackCapacity> &&src) U_NOEXCEPT;
    /**
     * Returns the array capacity (number of T items).
     * @return array capacity
     */
    int32_t getCapacity() const { return capacity; }
    /**
     * Access without ownership change.
     * @return the array pointer
     */
    T *getAlias() const { return ptr; }
    /**
     * Returns the array limit. Simple convenience method.
     * @return getAlias()+getCapacity()
     */
    T *getArrayLimit() const { return getAlias()+capacity; }
    // No "operator T *() const" because that can make
    // expressions like mbs[index] ambiguous for some compilers.
    /**
     * Array item access (const).
     * No index bounds check.
     * @param i array index
     * @return reference to the array item
     */
    const T &operator[](ptrdiff_t i) const { return ptr[i]; }
    /**
     * Array item access (writable).
     * No index bounds check.
     * @param i array index
     * @return reference to the array item
     */
    T &operator[](ptrdiff_t i) { return ptr[i]; }
    /**
     * Deletes the array (if owned) and aliases another one, no transfer of ownership.
     * If the arguments are illegal, then the current array is unchanged.
     * @param otherArray must not be NULL
     * @param otherCapacity must be >0
     */
    void aliasInstead(T *otherArray, int32_t otherCapacity) {
        if(otherArray!=NULL && otherCapacity>0) {
            releaseArray();
            ptr=otherArray;
            capacity=otherCapacity;
            needToRelease=FALSE;
        }
    }
    /**
     * Deletes the array (if owned) and allocates a new one, copying length T items.
     * Returns the new array pointer.
     * If the allocation fails, then the current array is unchanged and
     * this method returns NULL.
     * @param newCapacity can be less than or greater than the current capacity;
     *                    must be >0
     * @param length number of T items to be copied from the old array to the new one
     * @return the allocated array pointer, or NULL if the allocation failed
     */
    inline T *resize(int32_t newCapacity, int32_t length=0);
    /**
     * Gives up ownership of the array if owned, or else clones it,
     * copying length T items; resets itself to the internal stack array.
     * Returns NULL if the allocation failed.
     * @param length number of T items to copy when cloning,
     *        and capacity of the clone when cloning
     * @param resultCapacity will be set to the returned array's capacity (output-only)
     * @return the array pointer;
     *         caller becomes responsible for deleting the array
     */
    inline T *orphanOrClone(int32_t length, int32_t &resultCapacity);
private:
    T *ptr;
    int32_t capacity;
    UBool needToRelease;
    T stackArray[stackCapacity];
    void releaseArray() {
        if(needToRelease) {
            uprv_free(ptr);
        }
    }
    void resetToStackArray() {
        ptr=stackArray;
        capacity=stackCapacity;
        needToRelease=FALSE;
    }
    /* No comparison operators with other MaybeStackArray's. */
    bool operator==(const MaybeStackArray & /*other*/) {return FALSE;}
    bool operator!=(const MaybeStackArray & /*other*/) {return TRUE;}
    /* No ownership transfer: No copy constructor, no assignment operator. */
    MaybeStackArray(const MaybeStackArray & /*other*/) {}
    void operator=(const MaybeStackArray & /*other*/) {}
};

template<typename T, int32_t stackCapacity>
icu::MaybeStackArray<T, stackCapacity>::MaybeStackArray(
        MaybeStackArray <T, stackCapacity>&& src) U_NOEXCEPT
        : ptr(src.ptr), capacity(src.capacity), needToRelease(src.needToRelease) {
    if (src.ptr == src.stackArray) {
        ptr = stackArray;
        uprv_memcpy(stackArray, src.stackArray, sizeof(T) * src.capacity);
    } else {
        src.resetToStackArray();  // take ownership away from src
    }
}

template<typename T, int32_t stackCapacity>
inline MaybeStackArray <T, stackCapacity>&
MaybeStackArray<T, stackCapacity>::operator=(MaybeStackArray <T, stackCapacity>&& src) U_NOEXCEPT {
    releaseArray();  // in case this instance had its own memory allocated
    capacity = src.capacity;
    needToRelease = src.needToRelease;
    if (src.ptr == src.stackArray) {
        ptr = stackArray;
        uprv_memcpy(stackArray, src.stackArray, sizeof(T) * src.capacity);
    } else {
        ptr = src.ptr;
        src.resetToStackArray();  // take ownership away from src
    }
    return *this;
}

template<typename T, int32_t stackCapacity>
inline T *MaybeStackArray<T, stackCapacity>::resize(int32_t newCapacity, int32_t length) {
    if(newCapacity>0) {
#if U_DEBUG && defined(UPRV_MALLOC_COUNT)
      ::fprintf(::stderr,"MaybeStacArray (resize) alloc %d * %lu\n", newCapacity,sizeof(T));
#endif
        T *p=(T *)uprv_malloc(newCapacity*sizeof(T));
        if(p!=NULL) {
            if(length>0) {
                if(length>capacity) {
                    length=capacity;
                }
                if(length>newCapacity) {
                    length=newCapacity;
                }
                uprv_memcpy(p, ptr, (size_t)length*sizeof(T));
            }
            releaseArray();
            ptr=p;
            capacity=newCapacity;
            needToRelease=TRUE;
        }
        return p;
    } else {
        return NULL;
    }
}

template<typename T, int32_t stackCapacity>
inline T *MaybeStackArray<T, stackCapacity>::orphanOrClone(int32_t length, int32_t &resultCapacity) {
    T *p;
    if(needToRelease) {
        p=ptr;
    } else if(length<=0) {
        return NULL;
    } else {
        if(length>capacity) {
            length=capacity;
        }
        p=(T *)uprv_malloc(length*sizeof(T));
#if U_DEBUG && defined(UPRV_MALLOC_COUNT)
      ::fprintf(::stderr,"MaybeStacArray (orphan) alloc %d * %lu\n", length,sizeof(T));
#endif
        if(p==NULL) {
            return NULL;
        }
        uprv_memcpy(p, ptr, (size_t)length*sizeof(T));
    }
    resultCapacity=length;
    resetToStackArray();
    return p;
}

/**
 * Variant of MaybeStackArray that allocates a header struct and an array
 * in one contiguous memory block, using uprv_malloc() and uprv_free().
 * Provides internal memory with fixed array capacity. Can alias another memory
 * block or allocate one.
 * The stackCapacity is the number of T items in the internal memory,
 * not counting the H header.
 * Unlike LocalMemory and LocalArray, this class never adopts
 * (takes ownership of) another memory block.
 */
template<typename H, typename T, int32_t stackCapacity>
class MaybeStackHeaderAndArray {
public:
    // No heap allocation. Use only on the stack.
    static void* U_EXPORT2 operator new(size_t) U_NOEXCEPT = delete;
    static void* U_EXPORT2 operator new[](size_t) U_NOEXCEPT = delete;
#if U_HAVE_PLACEMENT_NEW
    static void* U_EXPORT2 operator new(size_t, void*) U_NOEXCEPT = delete;
#endif

    /**
     * Default constructor initializes with internal H+T[stackCapacity] buffer.
     */
    MaybeStackHeaderAndArray() : ptr(&stackHeader), capacity(stackCapacity), needToRelease(FALSE) {}
    /**
     * Destructor deletes the memory (if owned).
     */
    ~MaybeStackHeaderAndArray() { releaseMemory(); }
    /**
     * Returns the array capacity (number of T items).
     * @return array capacity
     */
    int32_t getCapacity() const { return capacity; }
    /**
     * Access without ownership change.
     * @return the header pointer
     */
    H *getAlias() const { return ptr; }
    /**
     * Returns the array start.
     * @return array start, same address as getAlias()+1
     */
    T *getArrayStart() const { return reinterpret_cast<T *>(getAlias()+1); }
    /**
     * Returns the array limit.
     * @return array limit
     */
    T *getArrayLimit() const { return getArrayStart()+capacity; }
    /**
     * Access without ownership change. Same as getAlias().
     * A class instance can be used directly in expressions that take a T *.
     * @return the header pointer
     */
    operator H *() const { return ptr; }
    /**
     * Array item access (writable).
     * No index bounds check.
     * @param i array index
     * @return reference to the array item
     */
    T &operator[](ptrdiff_t i) { return getArrayStart()[i]; }
    /**
     * Deletes the memory block (if owned) and aliases another one, no transfer of ownership.
     * If the arguments are illegal, then the current memory is unchanged.
     * @param otherArray must not be NULL
     * @param otherCapacity must be >0
     */
    void aliasInstead(H *otherMemory, int32_t otherCapacity) {
        if(otherMemory!=NULL && otherCapacity>0) {
            releaseMemory();
            ptr=otherMemory;
            capacity=otherCapacity;
            needToRelease=FALSE;
        }
    }
    /**
     * Deletes the memory block (if owned) and allocates a new one,
     * copying the header and length T array items.
     * Returns the new header pointer.
     * If the allocation fails, then the current memory is unchanged and
     * this method returns NULL.
     * @param newCapacity can be less than or greater than the current capacity;
     *                    must be >0
     * @param length number of T items to be copied from the old array to the new one
     * @return the allocated pointer, or NULL if the allocation failed
     */
    inline H *resize(int32_t newCapacity, int32_t length=0);
    /**
     * Gives up ownership of the memory if owned, or else clones it,
     * copying the header and length T array items; resets itself to the internal memory.
     * Returns NULL if the allocation failed.
     * @param length number of T items to copy when cloning,
     *        and array capacity of the clone when cloning
     * @param resultCapacity will be set to the returned array's capacity (output-only)
     * @return the header pointer;
     *         caller becomes responsible for deleting the array
     */
    inline H *orphanOrClone(int32_t length, int32_t &resultCapacity);
private:
    H *ptr;
    int32_t capacity;
    UBool needToRelease;
    // stackHeader must precede stackArray immediately.
    H stackHeader;
    T stackArray[stackCapacity];
    void releaseMemory() {
        if(needToRelease) {
            uprv_free(ptr);
        }
    }
    /* No comparison operators with other MaybeStackHeaderAndArray's. */
    bool operator==(const MaybeStackHeaderAndArray & /*other*/) {return FALSE;}
    bool operator!=(const MaybeStackHeaderAndArray & /*other*/) {return TRUE;}
    /* No ownership transfer: No copy constructor, no assignment operator. */
    MaybeStackHeaderAndArray(const MaybeStackHeaderAndArray & /*other*/) {}
    void operator=(const MaybeStackHeaderAndArray & /*other*/) {}
};

template<typename H, typename T, int32_t stackCapacity>
inline H *MaybeStackHeaderAndArray<H, T, stackCapacity>::resize(int32_t newCapacity,
                                                                int32_t length) {
    if(newCapacity>=0) {
#if U_DEBUG && defined(UPRV_MALLOC_COUNT)
      ::fprintf(::stderr,"MaybeStackHeaderAndArray alloc %d + %d * %ul\n", sizeof(H),newCapacity,sizeof(T));
#endif
        H *p=(H *)uprv_malloc(sizeof(H)+newCapacity*sizeof(T));
        if(p!=NULL) {
            if(length<0) {
                length=0;
            } else if(length>0) {
                if(length>capacity) {
                    length=capacity;
                }
                if(length>newCapacity) {
                    length=newCapacity;
                }
            }
            uprv_memcpy(p, ptr, sizeof(H)+(size_t)length*sizeof(T));
            releaseMemory();
            ptr=p;
            capacity=newCapacity;
            needToRelease=TRUE;
        }
        return p;
    } else {
        return NULL;
    }
}

template<typename H, typename T, int32_t stackCapacity>
inline H *MaybeStackHeaderAndArray<H, T, stackCapacity>::orphanOrClone(int32_t length,
                                                                       int32_t &resultCapacity) {
    H *p;
    if(needToRelease) {
        p=ptr;
    } else {
        if(length<0) {
            length=0;
        } else if(length>capacity) {
            length=capacity;
        }
#if U_DEBUG && defined(UPRV_MALLOC_COUNT)
      ::fprintf(::stderr,"MaybeStackHeaderAndArray (orphan) alloc %ul + %d * %lu\n", sizeof(H),length,sizeof(T));
#endif
        p=(H *)uprv_malloc(sizeof(H)+length*sizeof(T));
        if(p==NULL) {
            return NULL;
        }
        uprv_memcpy(p, ptr, sizeof(H)+(size_t)length*sizeof(T));
    }
    resultCapacity=length;
    ptr=&stackHeader;
    capacity=stackCapacity;
    needToRelease=FALSE;
    return p;
}

/**
 * A simple memory management class that creates new heap allocated objects (of
 * any class that has a public constructor), keeps track of them and eventually
 * deletes them all in its own destructor.
 *
 * A typical use-case would be code like this:
 *
 *     MemoryPool<MyType> pool;
 *
 *     MyType* o1 = pool.create();
 *     if (o1 != nullptr) {
 *         foo(o1);
 *     }
 *
 *     MyType* o2 = pool.create(1, 2, 3);
 *     if (o2 != nullptr) {
 *         bar(o2);
 *     }
 *
 *     // MemoryPool will take care of deleting the MyType objects.
 *
 * It doesn't do anything more than that, and is intentionally kept minimalist.
 */
template<typename T, int32_t stackCapacity = 8>
class MemoryPool : public UMemory {
public:
    MemoryPool() : count(0), pool() {}

    ~MemoryPool() {
        for (int32_t i = 0; i < count; ++i) {
            delete pool[i];
        }
    }

    MemoryPool(const MemoryPool&) = delete;
    MemoryPool& operator=(const MemoryPool&) = delete;

    MemoryPool(MemoryPool&& other) U_NOEXCEPT : count(other.count),
                                                pool(std::move(other.pool)) {
        other.count = 0;
    }

    MemoryPool& operator=(MemoryPool&& other) U_NOEXCEPT {
        count = other.count;
        pool = std::move(other.pool);
        other.count = 0;
        return *this;
    }

    /**
     * Creates a new object of typename T, by forwarding any and all arguments
     * to the typename T constructor.
     *
     * @param args Arguments to be forwarded to the typename T constructor.
     * @return A pointer to the newly created object, or nullptr on error.
     */
    template<typename... Args>
    T* create(Args&&... args) {
        int32_t capacity = pool.getCapacity();
        if (count == capacity &&
            pool.resize(capacity == stackCapacity ? 4 * capacity : 2 * capacity,
                        capacity) == nullptr) {
            return nullptr;
        }
        return pool[count++] = new T(std::forward<Args>(args)...);
    }

private:
    int32_t count;
    MaybeStackArray<T*, stackCapacity> pool;
};

U_NAMESPACE_END

#endif  /* __cplusplus */

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2007, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  bmpset.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2007jan29
*   created by: Markus W. Scherer
*/





U_NAMESPACE_BEGIN

/*
 * Helper class for frozen UnicodeSets, implements contains() and span()
 * optimized for BMP code points. Structured to be UTF-8-friendly.
 *
 * Latin-1: Look up bytes.
 * 2-byte characters: Bits organized vertically.
 * 3-byte characters: Use zero/one/mixed data per 64-block in U+0000..U+FFFF,
 *                    with mixed for illegal ranges.
 * Supplementary characters: Binary search over
 * the supplementary part of the parent set's inversion list.
 */
class BMPSet : public UMemory {
public:
    BMPSet(const int32_t *parentList, int32_t parentListLength);
    BMPSet(const BMPSet &otherBMPSet, const int32_t *newParentList, int32_t newParentListLength);
    virtual ~BMPSet();

    virtual UBool contains(UChar32 c) const;

    /*
     * Span the initial substring for which each character c has spanCondition==contains(c).
     * It must be s<limit and spanCondition==0 or 1.
     * @return The string pointer which limits the span.
     */
    const UChar *span(const UChar *s, const UChar *limit, USetSpanCondition spanCondition) const;

    /*
     * Span the trailing substring for which each character c has spanCondition==contains(c).
     * It must be s<limit and spanCondition==0 or 1.
     * @return The string pointer which starts the span.
     */
    const UChar *spanBack(const UChar *s, const UChar *limit, USetSpanCondition spanCondition) const;

    /*
     * Span the initial substring for which each character c has spanCondition==contains(c).
     * It must be length>0 and spanCondition==0 or 1.
     * @return The string pointer which limits the span.
     */
    const uint8_t *spanUTF8(const uint8_t *s, int32_t length, USetSpanCondition spanCondition) const;

    /*
     * Span the trailing substring for which each character c has spanCondition==contains(c).
     * It must be length>0 and spanCondition==0 or 1.
     * @return The start of the span.
     */
    int32_t spanBackUTF8(const uint8_t *s, int32_t length, USetSpanCondition spanCondition) const;

private:
    void initBits();
    void overrideIllegal();

    /**
     * Same as UnicodeSet::findCodePoint(UChar32 c) const except that the
     * binary search is restricted for finding code points in a certain range.
     *
     * For restricting the search for finding in the range start..end,
     * pass in
     *   lo=findCodePoint(start) and
     *   hi=findCodePoint(end)
     * with 0<=lo<=hi<len.
     * findCodePoint(c) defaults to lo=0 and hi=len-1.
     *
     * @param c a character in a subrange of MIN_VALUE..MAX_VALUE
     * @param lo The lowest index to be returned.
     * @param hi The highest index to be returned.
     * @return the smallest integer i in the range lo..hi,
     *         inclusive, such that c < list[i]
     */
    int32_t findCodePoint(UChar32 c, int32_t lo, int32_t hi) const;

    inline UBool containsSlow(UChar32 c, int32_t lo, int32_t hi) const;

    /*
     * One byte 0 or 1 per Latin-1 character.
     */
    UBool latin1Contains[0x100];

    /* TRUE if contains(U+FFFD). */
    UBool containsFFFD;

    /*
     * One bit per code point from U+0000..U+07FF.
     * The bits are organized vertically; consecutive code points
     * correspond to the same bit positions in consecutive table words.
     * With code point parts
     *   lead=c{10..6}
     *   trail=c{5..0}
     * it is set.contains(c)==(table7FF[trail] bit lead)
     *
     * Bits for 0..7F (non-shortest forms) are set to the result of contains(FFFD)
     * for faster validity checking at runtime.
     */
    uint32_t table7FF[64];

    /*
     * One bit per 64 BMP code points.
     * The bits are organized vertically; consecutive 64-code point blocks
     * correspond to the same bit position in consecutive table words.
     * With code point parts
     *   lead=c{15..12}
     *   t1=c{11..6}
     * test bits (lead+16) and lead in bmpBlockBits[t1].
     * If the upper bit is 0, then the lower bit indicates if contains(c)
     * for all code points in the 64-block.
     * If the upper bit is 1, then the block is mixed and set.contains(c)
     * must be called.
     *
     * Bits for 0..7FF (non-shortest forms) and D800..DFFF are set to
     * the result of contains(FFFD) for faster validity checking at runtime.
     */
    uint32_t bmpBlockBits[64];

    /*
     * Inversion list indexes for restricted binary searches in
     * findCodePoint(), from
     * findCodePoint(U+0800, U+1000, U+2000, .., U+F000, U+10000).
     * U+0800 is the first 3-byte-UTF-8 code point. Code points below U+0800 are
     * always looked up in the bit tables.
     * The last pair of indexes is for finding supplementary code points.
     */
    int32_t list4kStarts[18];

    /*
     * The inversion list of the parent set, for the slower contains() implementation
     * for mixed BMP blocks and for supplementary code points.
     * The list is terminated with list[listLength-1]=0x110000.
     */
    const int32_t *list;
    int32_t listLength;
};

inline UBool BMPSet::containsSlow(UChar32 c, int32_t lo, int32_t hi) const {
    return (UBool)(findCodePoint(c, lo, hi) & 1);
}

U_NAMESPACE_END


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2002-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File uassert.h
*
*  Contains the U_ASSERT and UPRV_UNREACHABLE macros
*
******************************************************************************
*/

/* utypes.h is included to get the proper define for uint8_t */

/* for abort */
#include <stdlib.h>

/**
 * \def U_ASSERT
 * By default, U_ASSERT just wraps the C library assert macro.
 * By changing the definition here, the assert behavior for ICU can be changed
 * without affecting other non - ICU uses of the C library assert().
*/
#if U_DEBUG
#   include <assert.h>
#   define U_ASSERT(exp) assert(exp)
#else
#   define U_ASSERT(exp)
#endif

/**
 * \def UPRV_UNREACHABLE
 * This macro is used to unconditionally abort if unreachable code is ever executed.
 * @internal
*/
#if defined(UPRV_UNREACHABLE)
    // Use the predefined value.
#else
#   define UPRV_UNREACHABLE abort()
#endif



U_NAMESPACE_BEGIN

BMPSet::BMPSet(const int32_t *parentList, int32_t parentListLength) :
        list(parentList), listLength(parentListLength) {
    uprv_memset(latin1Contains, 0, sizeof(latin1Contains));
    uprv_memset(table7FF, 0, sizeof(table7FF));
    uprv_memset(bmpBlockBits, 0, sizeof(bmpBlockBits));

    /*
     * Set the list indexes for binary searches for
     * U+0800, U+1000, U+2000, .., U+F000, U+10000.
     * U+0800 is the first 3-byte-UTF-8 code point. Lower code points are
     * looked up in the bit tables.
     * The last pair of indexes is for finding supplementary code points.
     */
    list4kStarts[0]=findCodePoint(0x800, 0, listLength-1);
    int32_t i;
    for(i=1; i<=0x10; ++i) {
        list4kStarts[i]=findCodePoint(i<<12, list4kStarts[i-1], listLength-1);
    }
    list4kStarts[0x11]=listLength-1;
    containsFFFD=containsSlow(0xfffd, list4kStarts[0xf], list4kStarts[0x10]);

    initBits();
    overrideIllegal();
}

BMPSet::BMPSet(const BMPSet &otherBMPSet, const int32_t *newParentList, int32_t newParentListLength) :
        containsFFFD(otherBMPSet.containsFFFD),
        list(newParentList), listLength(newParentListLength) {
    uprv_memcpy(latin1Contains, otherBMPSet.latin1Contains, sizeof(latin1Contains));
    uprv_memcpy(table7FF, otherBMPSet.table7FF, sizeof(table7FF));
    uprv_memcpy(bmpBlockBits, otherBMPSet.bmpBlockBits, sizeof(bmpBlockBits));
    uprv_memcpy(list4kStarts, otherBMPSet.list4kStarts, sizeof(list4kStarts));
}

BMPSet::~BMPSet() {
}

/*
 * Set bits in a bit rectangle in "vertical" bit organization.
 * start<limit<=0x800
 */
static void set32x64Bits(uint32_t table[64], int32_t start, int32_t limit) {
    U_ASSERT(start<limit);
    U_ASSERT(limit<=0x800);

    int32_t lead=start>>6;  // Named for UTF-8 2-byte lead byte with upper 5 bits.
    int32_t trail=start&0x3f;  // Named for UTF-8 2-byte trail byte with lower 6 bits.

    // Set one bit indicating an all-one block.
    uint32_t bits=(uint32_t)1<<lead;
    if((start+1)==limit) {  // Single-character shortcut.
        table[trail]|=bits;
        return;
    }

    int32_t limitLead=limit>>6;
    int32_t limitTrail=limit&0x3f;

    if(lead==limitLead) {
        // Partial vertical bit column.
        while(trail<limitTrail) {
            table[trail++]|=bits;
        }
    } else {
        // Partial vertical bit column,
        // followed by a bit rectangle,
        // followed by another partial vertical bit column.
        if(trail>0) {
            do {
                table[trail++]|=bits;
            } while(trail<64);
            ++lead;
        }
        if(lead<limitLead) {
            bits=~(((unsigned)1<<lead)-1);
            if(limitLead<0x20) {
                bits&=((unsigned)1<<limitLead)-1;
            }
            for(trail=0; trail<64; ++trail) {
                table[trail]|=bits;
            }
        }
        // limit<=0x800. If limit==0x800 then limitLead=32 and limitTrail=0.
        // In that case, bits=1<<limitLead is undefined but the bits value
        // is not used because trail<limitTrail is already false.
        bits=(uint32_t)1<<((limitLead == 0x20) ? (limitLead - 1) : limitLead);
        for(trail=0; trail<limitTrail; ++trail) {
            table[trail]|=bits;
        }
    }
}

void BMPSet::initBits() {
    UChar32 start, limit;
    int32_t listIndex=0;

    // Set latin1Contains[].
    do {
        start=list[listIndex++];
        if(listIndex<listLength) {
            limit=list[listIndex++];
        } else {
            limit=0x110000;
        }
        if(start>=0x100) {
            break;
        }
        do {
            latin1Contains[start++]=1;
        } while(start<limit && start<0x100);
    } while(limit<=0x100);

    // Find the first range overlapping with (or after) 80..FF again,
    // to include them in table7FF as well.
    for(listIndex=0;;) {
        start=list[listIndex++];
        if(listIndex<listLength) {
            limit=list[listIndex++];
        } else {
            limit=0x110000;
        }
        if(limit>0x80) {
            if(start<0x80) {
                start=0x80;
            }
            break;
        }
    }

    // Set table7FF[].
    while(start<0x800) {
        set32x64Bits(table7FF, start, limit<=0x800 ? limit : 0x800);
        if(limit>0x800) {
            start=0x800;
            break;
        }

        start=list[listIndex++];
        if(listIndex<listLength) {
            limit=list[listIndex++];
        } else {
            limit=0x110000;
        }
    }

    // Set bmpBlockBits[].
    int32_t minStart=0x800;
    while(start<0x10000) {
        if(limit>0x10000) {
            limit=0x10000;
        }

        if(start<minStart) {
            start=minStart;
        }
        if(start<limit) {  // Else: Another range entirely in a known mixed-value block.
            if(start&0x3f) {
                // Mixed-value block of 64 code points.
                start>>=6;
                bmpBlockBits[start&0x3f]|=0x10001<<(start>>6);
                start=(start+1)<<6;  // Round up to the next block boundary.
                minStart=start;      // Ignore further ranges in this block.
            }
            if(start<limit) {
                if(start<(limit&~0x3f)) {
                    // Multiple all-ones blocks of 64 code points each.
                    set32x64Bits(bmpBlockBits, start>>6, limit>>6);
                }

                if(limit&0x3f) {
                    // Mixed-value block of 64 code points.
                    limit>>=6;
                    bmpBlockBits[limit&0x3f]|=0x10001<<(limit>>6);
                    limit=(limit+1)<<6;  // Round up to the next block boundary.
                    minStart=limit;      // Ignore further ranges in this block.
                }
            }
        }

        if(limit==0x10000) {
            break;
        }

        start=list[listIndex++];
        if(listIndex<listLength) {
            limit=list[listIndex++];
        } else {
            limit=0x110000;
        }
    }
}

/*
 * Override some bits and bytes to the result of contains(FFFD)
 * for faster validity checking at runtime.
 * No need to set 0 values where they were reset to 0 in the constructor
 * and not modified by initBits().
 * (table7FF[] 0..7F, bmpBlockBits[] 0..7FF)
 * Need to set 0 values for surrogates D800..DFFF.
 */
void BMPSet::overrideIllegal() {
    uint32_t bits, mask;
    int32_t i;

    if(containsFFFD) {
        bits=3;                 // Lead bytes 0xC0 and 0xC1.
        for(i=0; i<64; ++i) {
            table7FF[i]|=bits;
        }

        bits=1;                 // Lead byte 0xE0.
        for(i=0; i<32; ++i) {   // First half of 4k block.
            bmpBlockBits[i]|=bits;
        }

        mask= static_cast<uint32_t>(~(0x10001<<0xd));   // Lead byte 0xED.
        bits=1<<0xd;
        for(i=32; i<64; ++i) {  // Second half of 4k block.
            bmpBlockBits[i]=(bmpBlockBits[i]&mask)|bits;
        }
    } else {
        mask= static_cast<uint32_t>(~(0x10001<<0xd));   // Lead byte 0xED.
        for(i=32; i<64; ++i) {  // Second half of 4k block.
            bmpBlockBits[i]&=mask;
        }
    }
}

int32_t BMPSet::findCodePoint(UChar32 c, int32_t lo, int32_t hi) const {
    /* Examples:
                                       findCodePoint(c)
       set              list[]         c=0 1 3 4 7 8
       ===              ==============   ===========
       []               [110000]         0 0 0 0 0 0
       [\u0000-\u0003]  [0, 4, 110000]   1 1 1 2 2 2
       [\u0004-\u0007]  [4, 8, 110000]   0 0 0 1 1 2
       [:Any:]          [0, 110000]      1 1 1 1 1 1
     */

    // Return the smallest i such that c < list[i].  Assume
    // list[len - 1] == HIGH and that c is legal (0..HIGH-1).
    if (c < list[lo])
        return lo;
    // High runner test.  c is often after the last range, so an
    // initial check for this condition pays off.
    if (lo >= hi || c >= list[hi-1])
        return hi;
    // invariant: c >= list[lo]
    // invariant: c < list[hi]
    for (;;) {
        int32_t i = (lo + hi) >> 1;
        if (i == lo) {
            break; // Found!
        } else if (c < list[i]) {
            hi = i;
        } else {
            lo = i;
        }
    }
    return hi;
}

UBool
BMPSet::contains(UChar32 c) const {
    if((uint32_t)c<=0xff) {
        return (UBool)latin1Contains[c];
    } else if((uint32_t)c<=0x7ff) {
        return (UBool)((table7FF[c&0x3f]&((uint32_t)1<<(c>>6)))!=0);
    } else if((uint32_t)c<0xd800 || (c>=0xe000 && c<=0xffff)) {
        int lead=c>>12;
        uint32_t twoBits=(bmpBlockBits[(c>>6)&0x3f]>>lead)&0x10001;
        if(twoBits<=1) {
            // All 64 code points with the same bits 15..6
            // are either in the set or not.
            return (UBool)twoBits;
        } else {
            // Look up the code point in its 4k block of code points.
            return containsSlow(c, list4kStarts[lead], list4kStarts[lead+1]);
        }
    } else if((uint32_t)c<=0x10ffff) {
        // surrogate or supplementary code point
        return containsSlow(c, list4kStarts[0xd], list4kStarts[0x11]);
    } else {
        // Out-of-range code points get FALSE, consistent with long-standing
        // behavior of UnicodeSet::contains(c).
        return FALSE;
    }
}

/*
 * Check for sufficient length for trail unit for each surrogate pair.
 * Handle single surrogates as surrogate code points as usual in ICU.
 */
const UChar *
BMPSet::span(const UChar *s, const UChar *limit, USetSpanCondition spanCondition) const {
    UChar c, c2;

    if(spanCondition) {
        // span
        do {
            c=*s;
            if(c<=0xff) {
                if(!latin1Contains[c]) {
                    break;
                }
            } else if(c<=0x7ff) {
                if((table7FF[c&0x3f]&((uint32_t)1<<(c>>6)))==0) {
                    break;
                }
            } else if(c<0xd800 || c>=0xe000) {
                int lead=c>>12;
                uint32_t twoBits=(bmpBlockBits[(c>>6)&0x3f]>>lead)&0x10001;
                if(twoBits<=1) {
                    // All 64 code points with the same bits 15..6
                    // are either in the set or not.
                    if(twoBits==0) {
                        break;
                    }
                } else {
                    // Look up the code point in its 4k block of code points.
                    if(!containsSlow(c, list4kStarts[lead], list4kStarts[lead+1])) {
                        break;
                    }
                }
            } else if(c>=0xdc00 || (s+1)==limit || (c2=s[1])<0xdc00 || c2>=0xe000) {
                // surrogate code point
                if(!containsSlow(c, list4kStarts[0xd], list4kStarts[0xe])) {
                    break;
                }
            } else {
                // surrogate pair
                if(!containsSlow(U16_GET_SUPPLEMENTARY(c, c2), list4kStarts[0x10], list4kStarts[0x11])) {
                    break;
                }
                ++s;
            }
        } while(++s<limit);
    } else {
        // span not
        do {
            c=*s;
            if(c<=0xff) {
                if(latin1Contains[c]) {
                    break;
                }
            } else if(c<=0x7ff) {
                if((table7FF[c&0x3f]&((uint32_t)1<<(c>>6)))!=0) {
                    break;
                }
            } else if(c<0xd800 || c>=0xe000) {
                int lead=c>>12;
                uint32_t twoBits=(bmpBlockBits[(c>>6)&0x3f]>>lead)&0x10001;
                if(twoBits<=1) {
                    // All 64 code points with the same bits 15..6
                    // are either in the set or not.
                    if(twoBits!=0) {
                        break;
                    }
                } else {
                    // Look up the code point in its 4k block of code points.
                    if(containsSlow(c, list4kStarts[lead], list4kStarts[lead+1])) {
                        break;
                    }
                }
            } else if(c>=0xdc00 || (s+1)==limit || (c2=s[1])<0xdc00 || c2>=0xe000) {
                // surrogate code point
                if(containsSlow(c, list4kStarts[0xd], list4kStarts[0xe])) {
                    break;
                }
            } else {
                // surrogate pair
                if(containsSlow(U16_GET_SUPPLEMENTARY(c, c2), list4kStarts[0x10], list4kStarts[0x11])) {
                    break;
                }
                ++s;
            }
        } while(++s<limit);
    }
    return s;
}

/* Symmetrical with span(). */
const UChar *
BMPSet::spanBack(const UChar *s, const UChar *limit, USetSpanCondition spanCondition) const {
    UChar c, c2;

    if(spanCondition) {
        // span
        for(;;) {
            c=*(--limit);
            if(c<=0xff) {
                if(!latin1Contains[c]) {
                    break;
                }
            } else if(c<=0x7ff) {
                if((table7FF[c&0x3f]&((uint32_t)1<<(c>>6)))==0) {
                    break;
                }
            } else if(c<0xd800 || c>=0xe000) {
                int lead=c>>12;
                uint32_t twoBits=(bmpBlockBits[(c>>6)&0x3f]>>lead)&0x10001;
                if(twoBits<=1) {
                    // All 64 code points with the same bits 15..6
                    // are either in the set or not.
                    if(twoBits==0) {
                        break;
                    }
                } else {
                    // Look up the code point in its 4k block of code points.
                    if(!containsSlow(c, list4kStarts[lead], list4kStarts[lead+1])) {
                        break;
                    }
                }
            } else if(c<0xdc00 || s==limit || (c2=*(limit-1))<0xd800 || c2>=0xdc00) {
                // surrogate code point
                if(!containsSlow(c, list4kStarts[0xd], list4kStarts[0xe])) {
                    break;
                }
            } else {
                // surrogate pair
                if(!containsSlow(U16_GET_SUPPLEMENTARY(c2, c), list4kStarts[0x10], list4kStarts[0x11])) {
                    break;
                }
                --limit;
            }
            if(s==limit) {
                return s;
            }
        }
    } else {
        // span not
        for(;;) {
            c=*(--limit);
            if(c<=0xff) {
                if(latin1Contains[c]) {
                    break;
                }
            } else if(c<=0x7ff) {
                if((table7FF[c&0x3f]&((uint32_t)1<<(c>>6)))!=0) {
                    break;
                }
            } else if(c<0xd800 || c>=0xe000) {
                int lead=c>>12;
                uint32_t twoBits=(bmpBlockBits[(c>>6)&0x3f]>>lead)&0x10001;
                if(twoBits<=1) {
                    // All 64 code points with the same bits 15..6
                    // are either in the set or not.
                    if(twoBits!=0) {
                        break;
                    }
                } else {
                    // Look up the code point in its 4k block of code points.
                    if(containsSlow(c, list4kStarts[lead], list4kStarts[lead+1])) {
                        break;
                    }
                }
            } else if(c<0xdc00 || s==limit || (c2=*(limit-1))<0xd800 || c2>=0xdc00) {
                // surrogate code point
                if(containsSlow(c, list4kStarts[0xd], list4kStarts[0xe])) {
                    break;
                }
            } else {
                // surrogate pair
                if(containsSlow(U16_GET_SUPPLEMENTARY(c2, c), list4kStarts[0x10], list4kStarts[0x11])) {
                    break;
                }
                --limit;
            }
            if(s==limit) {
                return s;
            }
        }
    }
    return limit+1;
}

/*
 * Precheck for sufficient trail bytes at end of string only once per span.
 * Check validity.
 */
const uint8_t *
BMPSet::spanUTF8(const uint8_t *s, int32_t length, USetSpanCondition spanCondition) const {
    const uint8_t *limit=s+length;
    uint8_t b=*s;
    if(U8_IS_SINGLE(b)) {
        // Initial all-ASCII span.
        if(spanCondition) {
            do {
                if(!latin1Contains[b] || ++s==limit) {
                    return s;
                }
                b=*s;
            } while(U8_IS_SINGLE(b));
        } else {
            do {
                if(latin1Contains[b] || ++s==limit) {
                    return s;
                }
                b=*s;
            } while(U8_IS_SINGLE(b));
        }
        length=(int32_t)(limit-s);
    }

    if(spanCondition!=USET_SPAN_NOT_CONTAINED) {
        spanCondition=USET_SPAN_CONTAINED;  // Pin to 0/1 values.
    }

    const uint8_t *limit0=limit;

    /*
     * Make sure that the last 1/2/3/4-byte sequence before limit is complete
     * or runs into a lead byte.
     * In the span loop compare s with limit only once
     * per multi-byte character.
     *
     * Give a trailing illegal sequence the same value as the result of contains(FFFD),
     * including it if that is part of the span, otherwise set limit0 to before
     * the truncated sequence.
     */
    b=*(limit-1);
    if((int8_t)b<0) {
        // b>=0x80: lead or trail byte
        if(b<0xc0) {
            // single trail byte, check for preceding 3- or 4-byte lead byte
            if(length>=2 && (b=*(limit-2))>=0xe0) {
                limit-=2;
                if(containsFFFD!=spanCondition) {
                    limit0=limit;
                }
            } else if(b<0xc0 && b>=0x80 && length>=3 && (b=*(limit-3))>=0xf0) {
                // 4-byte lead byte with only two trail bytes
                limit-=3;
                if(containsFFFD!=spanCondition) {
                    limit0=limit;
                }
            }
        } else {
            // lead byte with no trail bytes
            --limit;
            if(containsFFFD!=spanCondition) {
                limit0=limit;
            }
        }
    }

    uint8_t t1, t2, t3;

    while(s<limit) {
        b=*s;
        if(U8_IS_SINGLE(b)) {
            // ASCII
            if(spanCondition) {
                do {
                    if(!latin1Contains[b]) {
                        return s;
                    } else if(++s==limit) {
                        return limit0;
                    }
                    b=*s;
                } while(U8_IS_SINGLE(b));
            } else {
                do {
                    if(latin1Contains[b]) {
                        return s;
                    } else if(++s==limit) {
                        return limit0;
                    }
                    b=*s;
                } while(U8_IS_SINGLE(b));
            }
        }
        ++s;  // Advance past the lead byte.
        if(b>=0xe0) {
            if(b<0xf0) {
                if( /* handle U+0000..U+FFFF inline */
                    (t1=(uint8_t)(s[0]-0x80)) <= 0x3f &&
                    (t2=(uint8_t)(s[1]-0x80)) <= 0x3f
                ) {
                    b&=0xf;
                    uint32_t twoBits=(bmpBlockBits[t1]>>b)&0x10001;
                    if(twoBits<=1) {
                        // All 64 code points with this lead byte and middle trail byte
                        // are either in the set or not.
                        if(twoBits!=(uint32_t)spanCondition) {
                            return s-1;
                        }
                    } else {
                        // Look up the code point in its 4k block of code points.
                        UChar32 c=(b<<12)|(t1<<6)|t2;
                        if(containsSlow(c, list4kStarts[b], list4kStarts[b+1]) != spanCondition) {
                            return s-1;
                        }
                    }
                    s+=2;
                    continue;
                }
            } else if( /* handle U+10000..U+10FFFF inline */
                (t1=(uint8_t)(s[0]-0x80)) <= 0x3f &&
                (t2=(uint8_t)(s[1]-0x80)) <= 0x3f &&
                (t3=(uint8_t)(s[2]-0x80)) <= 0x3f
            ) {
                // Give an illegal sequence the same value as the result of contains(FFFD).
                UChar32 c=((UChar32)(b-0xf0)<<18)|((UChar32)t1<<12)|(t2<<6)|t3;
                if( (   (0x10000<=c && c<=0x10ffff) ?
                            containsSlow(c, list4kStarts[0x10], list4kStarts[0x11]) :
                            containsFFFD
                    ) != spanCondition
                ) {
                    return s-1;
                }
                s+=3;
                continue;
            }
        } else {
            if( /* handle U+0000..U+07FF inline */
                b>=0xc0 &&
                (t1=(uint8_t)(*s-0x80)) <= 0x3f
            ) {
                if((USetSpanCondition)((table7FF[t1]&((uint32_t)1<<(b&0x1f)))!=0) != spanCondition) {
                    return s-1;
                }
                ++s;
                continue;
            }
        }

        // Give an illegal sequence the same value as the result of contains(FFFD).
        // Handle each byte of an illegal sequence separately to simplify the code;
        // no need to optimize error handling.
        if(containsFFFD!=spanCondition) {
            return s-1;
        }
    }

    return limit0;
}

/*
 * While going backwards through UTF-8 optimize only for ASCII.
 * Unlike UTF-16, UTF-8 is not forward-backward symmetrical, that is, it is not
 * possible to tell from the last byte in a multi-byte sequence how many
 * preceding bytes there should be. Therefore, going backwards through UTF-8
 * is much harder than going forward.
 */
int32_t
BMPSet::spanBackUTF8(const uint8_t *s, int32_t length, USetSpanCondition spanCondition) const {
    if(spanCondition!=USET_SPAN_NOT_CONTAINED) {
        spanCondition=USET_SPAN_CONTAINED;  // Pin to 0/1 values.
    }

    uint8_t b;

    do {
        b=s[--length];
        if(U8_IS_SINGLE(b)) {
            // ASCII sub-span
            if(spanCondition) {
                do {
                    if(!latin1Contains[b]) {
                        return length+1;
                    } else if(length==0) {
                        return 0;
                    }
                    b=s[--length];
                } while(U8_IS_SINGLE(b));
            } else {
                do {
                    if(latin1Contains[b]) {
                        return length+1;
                    } else if(length==0) {
                        return 0;
                    }
                    b=s[--length];
                } while(U8_IS_SINGLE(b));
            }
        }

        int32_t prev=length;
        UChar32 c;
        // trail byte: collect a multi-byte character
        // (or  lead byte in last-trail position)
        c=utf8_prevCharSafeBody(s, 0, &length, b, -3);
        // c is a valid code point, not ASCII, not a surrogate
        if(c<=0x7ff) {
            if((USetSpanCondition)((table7FF[c&0x3f]&((uint32_t)1<<(c>>6)))!=0) != spanCondition) {
                return prev+1;
            }
        } else if(c<=0xffff) {
            int lead=c>>12;
            uint32_t twoBits=(bmpBlockBits[(c>>6)&0x3f]>>lead)&0x10001;
            if(twoBits<=1) {
                // All 64 code points with the same bits 15..6
                // are either in the set or not.
                if(twoBits!=(uint32_t)spanCondition) {
                    return prev+1;
                }
            } else {
                // Look up the code point in its 4k block of code points.
                if(containsSlow(c, list4kStarts[lead], list4kStarts[lead+1]) != spanCondition) {
                    return prev+1;
                }
            }
        } else {
            if(containsSlow(c, list4kStarts[0x10], list4kStarts[0x11]) != spanCondition) {
                return prev+1;
            }
        }
    } while(length>0);
    return 0;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ************************************************************************************
 * Copyright (C) 2006-2016, International Business Machines Corporation
 * and others. All Rights Reserved.
 ************************************************************************************
 */



// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/uchar.h"
// #include "unicode/uniset.h"
// #include "unicode/chariter.h"
// #include "unicode/ures.h"
// #include "unicode/udata.h"
// #include "unicode/putil.h"
// #include "unicode/ustring.h"
// #include "unicode/uscript.h"
// #include "unicode/ucharstrie.h"
// #include "unicode/bytestrie.h"

// #include "brkeng.h"
// #include "cmemory.h"
// #include "dictbe.h"
// #include "charstr.h"
// #include "dictionarydata.h"
// #include "mutex.h"
// #include "uvector.h"
// #include "umutex.h"
// #include "uresimp.h"
// #include "ubrkimpl.h"

// U_NAMESPACE_BEGIN

// /*
//  ******************************************************************
//  */

// LanguageBreakEngine::LanguageBreakEngine() {
// }

// LanguageBreakEngine::~LanguageBreakEngine() {
// }

// /*
//  ******************************************************************
//  */

// LanguageBreakFactory::LanguageBreakFactory() {
// }

// LanguageBreakFactory::~LanguageBreakFactory() {
// }

// /*
//  ******************************************************************
//  */

// UnhandledEngine::UnhandledEngine(UErrorCode &status) : fHandled(nullptr) {
//     (void)status;
// }

// UnhandledEngine::~UnhandledEngine() {
//     delete fHandled;
//     fHandled = nullptr;
// }

// UBool
// UnhandledEngine::handles(UChar32 c) const {
//     return fHandled && fHandled->contains(c);
// }

// int32_t
// UnhandledEngine::findBreaks( UText *text,
//                              int32_t /* startPos */,
//                              int32_t endPos,
//                              UVector32 &/*foundBreaks*/ ) const {
//     UChar32 c = utext_current32(text);
//     while((int32_t)utext_getNativeIndex(text) < endPos && fHandled->contains(c)) {
//         utext_next32(text);            // TODO:  recast loop to work with post-increment operations.
//         c = utext_current32(text);
//     }
//     return 0;
// }

// void
// UnhandledEngine::handleCharacter(UChar32 c) {
//     if (fHandled == nullptr) {
//         fHandled = new UnicodeSet();
//         if (fHandled == nullptr) {
//             return;
//         }
//     }
//     if (!fHandled->contains(c)) {
//         UErrorCode status = U_ZERO_ERROR;
//         // Apply the entire script of the character.
//         int32_t script = u_getIntPropertyValue(c, UCHAR_SCRIPT);
//         fHandled->applyIntPropertyValue(UCHAR_SCRIPT, script, status);
//     }
// }

// /*
//  ******************************************************************
//  */

// ICULanguageBreakFactory::ICULanguageBreakFactory(UErrorCode &/*status*/) {
//     fEngines = 0;
// }

// ICULanguageBreakFactory::~ICULanguageBreakFactory() {
//     if (fEngines != 0) {
//         delete fEngines;
//     }
// }

// U_NAMESPACE_END
// U_CDECL_BEGIN
// static void U_CALLCONV _deleteEngine(void *obj) {
//     delete (const icu::LanguageBreakEngine *) obj;
// }
// U_CDECL_END
// U_NAMESPACE_BEGIN

// const LanguageBreakEngine *
// ICULanguageBreakFactory::getEngineFor(UChar32 c) {
//     const LanguageBreakEngine *lbe = NULL;
//     UErrorCode  status = U_ZERO_ERROR;

//     static UMutex gBreakEngineMutex;
//     Mutex m(&gBreakEngineMutex);

//     if (fEngines == NULL) {
//         UStack  *engines = new UStack(_deleteEngine, NULL, status);
//         if (U_FAILURE(status) || engines == NULL) {
//             // Note: no way to return error code to caller.
//             delete engines;
//             return NULL;
//         }
//         fEngines = engines;
//     } else {
//         int32_t i = fEngines->size();
//         while (--i >= 0) {
//             lbe = (const LanguageBreakEngine *)(fEngines->elementAt(i));
//             if (lbe != NULL && lbe->handles(c)) {
//                 return lbe;
//             }
//         }
//     }

//     // We didn't find an engine. Create one.
//     lbe = loadEngineFor(c);
//     if (lbe != NULL) {
//         fEngines->push((void *)lbe, status);
//     }
//     return lbe;
// }

// const LanguageBreakEngine *
// ICULanguageBreakFactory::loadEngineFor(UChar32 c) {
//     UErrorCode status = U_ZERO_ERROR;
//     UScriptCode code = uscript_getScript(c, &status);
//     if (U_SUCCESS(status)) {
//         DictionaryMatcher *m = loadDictionaryMatcherFor(code);
//         if (m != NULL) {
//             const LanguageBreakEngine *engine = NULL;
//             switch(code) {
//             case USCRIPT_THAI:
//                 engine = new ThaiBreakEngine(m, status);
//                 break;
//             case USCRIPT_LAO:
//                 engine = new LaoBreakEngine(m, status);
//                 break;
//             case USCRIPT_MYANMAR:
//                 engine = new BurmeseBreakEngine(m, status);
//                 break;
//             case USCRIPT_KHMER:
//                 engine = new KhmerBreakEngine(m, status);
//                 break;

// #if !UCONFIG_NO_NORMALIZATION
//                 // CJK not available w/o normalization
//             case USCRIPT_HANGUL:
//                 engine = new CjkBreakEngine(m, kKorean, status);
//                 break;

//             // use same BreakEngine and dictionary for both Chinese and Japanese
//             case USCRIPT_HIRAGANA:
//             case USCRIPT_KATAKANA:
//             case USCRIPT_HAN:
//                 engine = new CjkBreakEngine(m, kChineseJapanese, status);
//                 break;
// #if 0
//             // TODO: Have to get some characters with script=common handled
//             // by CjkBreakEngine (e.g. U+309B). Simply subjecting
//             // them to CjkBreakEngine does not work. The engine has to
//             // special-case them.
//             case USCRIPT_COMMON:
//             {
//                 UBlockCode block = ublock_getCode(code);
//                 if (block == UBLOCK_HIRAGANA || block == UBLOCK_KATAKANA)
//                    engine = new CjkBreakEngine(dict, kChineseJapanese, status);
//                 break;
//             }
// #endif
// #endif

//             default:
//                 break;
//             }
//             if (engine == NULL) {
//                 delete m;
//             }
//             else if (U_FAILURE(status)) {
//                 delete engine;
//                 engine = NULL;
//             }
//             return engine;
//         }
//     }
//     return NULL;
// }

// DictionaryMatcher *
// ICULanguageBreakFactory::loadDictionaryMatcherFor(UScriptCode script) {
//     UErrorCode status = U_ZERO_ERROR;
//     // open root from brkitr tree.
//     UResourceBundle *b = ures_open(U_ICUDATA_BRKITR, "", &status);
//     b = ures_getByKeyWithFallback(b, "dictionaries", b, &status);
//     int32_t dictnlength = 0;
//     const UChar *dictfname =
//         ures_getStringByKeyWithFallback(b, uscript_getShortName(script), &dictnlength, &status);
//     if (U_FAILURE(status)) {
//         ures_close(b);
//         return NULL;
//     }
//     CharString dictnbuf;
//     CharString ext;
//     const UChar *extStart = u_memrchr(dictfname, 0x002e, dictnlength);  // last dot
//     if (extStart != NULL) {
//         int32_t len = (int32_t)(extStart - dictfname);
//         ext.appendInvariantChars(UnicodeString(FALSE, extStart + 1, dictnlength - len - 1), status);
//         dictnlength = len;
//     }
//     dictnbuf.appendInvariantChars(UnicodeString(FALSE, dictfname, dictnlength), status);
//     ures_close(b);

//     UDataMemory *file = udata_open(U_ICUDATA_BRKITR, ext.data(), dictnbuf.data(), &status);
//     if (U_SUCCESS(status)) {
//         // build trie
//         const uint8_t *data = (const uint8_t *)udata_getMemory(file);
//         const int32_t *indexes = (const int32_t *)data;
//         const int32_t offset = indexes[DictionaryData::IX_STRING_TRIE_OFFSET];
//         const int32_t trieType = indexes[DictionaryData::IX_TRIE_TYPE] & DictionaryData::TRIE_TYPE_MASK;
//         DictionaryMatcher *m = NULL;
//         if (trieType == DictionaryData::TRIE_TYPE_BYTES) {
//             const int32_t transform = indexes[DictionaryData::IX_TRANSFORM];
//             const char *characters = (const char *)(data + offset);
//             m = new BytesDictionaryMatcher(characters, transform, file);
//         }
//         else if (trieType == DictionaryData::TRIE_TYPE_UCHARS) {
//             const UChar *characters = (const UChar *)(data + offset);
//             m = new UCharsDictionaryMatcher(characters, file);
//         }
//         if (m == NULL) {
//             // no matcher exists to take ownership - either we are an invalid
//             // type or memory allocation failed
//             udata_close(file);
//         }
//         return m;
//     } else if (dictfname != NULL) {
//         // we don't have a dictionary matcher.
//         // returning NULL here will cause us to fail to find a dictionary break engine, as expected
//         status = U_ZERO_ERROR;
//         return NULL;
//     }
//     return NULL;
// }

// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2015, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File brkiter.cpp
*
* Modification History:
*
*   Date        Name        Description
*   02/18/97    aliu        Converted from OpenClass.  Added DONE.
*   01/13/2000  helena      Added UErrorCode parameter to createXXXInstance methods.
*****************************************************************************************
*/

// *****************************************************************************
// This file was generated from the java source file BreakIterator.java
// *****************************************************************************



// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/rbbi.h"
// #include "unicode/brkiter.h"
// #include "unicode/udata.h"
// #include "unicode/ures.h"
// #include "unicode/ustring.h"
// #include "unicode/filteredbrk.h"
// #include "ucln_cmn.h"
// #include "cstring.h"
// #include "umutex.h"
// #include "servloc.h"
// #include "locbased.h"
// #include "uresimp.h"
// #include "uassert.h"
// #include "ubrkimpl.h"
// #include "charstr.h"

// // *****************************************************************************
// // class BreakIterator
// // This class implements methods for finding the location of boundaries in text.
// // Instances of BreakIterator maintain a current position and scan over text
// // returning the index of characters where boundaries occur.
// // *****************************************************************************

// U_NAMESPACE_BEGIN

// // -------------------------------------

// BreakIterator*
// BreakIterator::buildInstance(const Locale& loc, const char *type, UErrorCode &status)
// {
//     char fnbuff[256];
//     char ext[4]={'\0'};
//     CharString actualLocale;
//     int32_t size;
//     const UChar* brkfname = NULL;
//     UResourceBundle brkRulesStack;
//     UResourceBundle brkNameStack;
//     UResourceBundle *brkRules = &brkRulesStack;
//     UResourceBundle *brkName  = &brkNameStack;
//     RuleBasedBreakIterator *result = NULL;

//     if (U_FAILURE(status))
//         return NULL;

//     ures_initStackObject(brkRules);
//     ures_initStackObject(brkName);

//     // Get the locale
//     UResourceBundle *b = ures_openNoDefault(U_ICUDATA_BRKITR, loc.getName(), &status);

//     // Get the "boundaries" array.
//     if (U_SUCCESS(status)) {
//         brkRules = ures_getByKeyWithFallback(b, "boundaries", brkRules, &status);
//         // Get the string object naming the rules file
//         brkName = ures_getByKeyWithFallback(brkRules, type, brkName, &status);
//         // Get the actual string
//         brkfname = ures_getString(brkName, &size, &status);
//         U_ASSERT((size_t)size<sizeof(fnbuff));
//         if ((size_t)size>=sizeof(fnbuff)) {
//             size=0;
//             if (U_SUCCESS(status)) {
//                 status = U_BUFFER_OVERFLOW_ERROR;
//             }
//         }

//         // Use the string if we found it
//         if (U_SUCCESS(status) && brkfname) {
//             actualLocale.append(ures_getLocaleInternal(brkName, &status), -1, status);

//             UChar* extStart=u_strchr(brkfname, 0x002e);
//             int len = 0;
//             if(extStart!=NULL){
//                 len = (int)(extStart-brkfname);
//                 u_UCharsToChars(extStart+1, ext, sizeof(ext)); // nul terminates the buff
//                 u_UCharsToChars(brkfname, fnbuff, len);
//             }
//             fnbuff[len]=0; // nul terminate
//         }
//     }

//     ures_close(brkRules);
//     ures_close(brkName);

//     UDataMemory* file = udata_open(U_ICUDATA_BRKITR, ext, fnbuff, &status);
//     if (U_FAILURE(status)) {
//         ures_close(b);
//         return NULL;
//     }

//     // Create a RuleBasedBreakIterator
//     result = new RuleBasedBreakIterator(file, status);

//     // If there is a result, set the valid locale and actual locale, and the kind
//     if (U_SUCCESS(status) && result != NULL) {
//         U_LOCALE_BASED(locBased, *(BreakIterator*)result);
//         locBased.setLocaleIDs(ures_getLocaleByType(b, ULOC_VALID_LOCALE, &status),
//                               actualLocale.data());
//     }

//     ures_close(b);

//     if (U_FAILURE(status) && result != NULL) {  // Sometimes redundant check, but simple
//         delete result;
//         return NULL;
//     }

//     if (result == NULL) {
//         udata_close(file);
//         if (U_SUCCESS(status)) {
//             status = U_MEMORY_ALLOCATION_ERROR;
//         }
//     }

//     return result;
// }

// // Creates a break iterator for word breaks.
// BreakIterator* U_EXPORT2
// BreakIterator::createWordInstance(const Locale& key, UErrorCode& status)
// {
//     return createInstance(key, UBRK_WORD, status);
// }

// // -------------------------------------

// // Creates a break iterator  for line breaks.
// BreakIterator* U_EXPORT2
// BreakIterator::createLineInstance(const Locale& key, UErrorCode& status)
// {
//     return createInstance(key, UBRK_LINE, status);
// }

// // -------------------------------------

// // Creates a break iterator  for character breaks.
// BreakIterator* U_EXPORT2
// BreakIterator::createCharacterInstance(const Locale& key, UErrorCode& status)
// {
//     return createInstance(key, UBRK_CHARACTER, status);
// }

// // -------------------------------------

// // Creates a break iterator  for sentence breaks.
// BreakIterator* U_EXPORT2
// BreakIterator::createSentenceInstance(const Locale& key, UErrorCode& status)
// {
//     return createInstance(key, UBRK_SENTENCE, status);
// }

// // -------------------------------------

// // Creates a break iterator for title casing breaks.
// BreakIterator* U_EXPORT2
// BreakIterator::createTitleInstance(const Locale& key, UErrorCode& status)
// {
//     return createInstance(key, UBRK_TITLE, status);
// }

// // -------------------------------------

// // Gets all the available locales that has localized text boundary data.
// const Locale* U_EXPORT2
// BreakIterator::getAvailableLocales(int32_t& count)
// {
//     return Locale::getAvailableLocales(count);
// }

// // ------------------------------------------
// //
// // Constructors, destructor and assignment operator
// //
// //-------------------------------------------

// BreakIterator::BreakIterator()
// {
//     *validLocale = *actualLocale = 0;
// }

// BreakIterator::BreakIterator(const BreakIterator &other) : UObject(other) {
//     uprv_strncpy(actualLocale, other.actualLocale, sizeof(actualLocale));
//     uprv_strncpy(validLocale, other.validLocale, sizeof(validLocale));
// }

// BreakIterator &BreakIterator::operator =(const BreakIterator &other) {
//     if (this != &other) {
//         uprv_strncpy(actualLocale, other.actualLocale, sizeof(actualLocale));
//         uprv_strncpy(validLocale, other.validLocale, sizeof(validLocale));
//     }
//     return *this;
// }

// BreakIterator::~BreakIterator()
// {
// }

// // ------------------------------------------
// //
// // Registration
// //
// //-------------------------------------------
// #if !UCONFIG_NO_SERVICE

// // -------------------------------------

// class ICUBreakIteratorFactory : public ICUResourceBundleFactory {
// public:
//     virtual ~ICUBreakIteratorFactory();
// protected:
//     virtual UObject* handleCreate(const Locale& loc, int32_t kind, const ICUService* /*service*/, UErrorCode& status) const {
//         return BreakIterator::makeInstance(loc, kind, status);
//     }
// };

// ICUBreakIteratorFactory::~ICUBreakIteratorFactory() {}

// // -------------------------------------

// class ICUBreakIteratorService : public ICULocaleService {
// public:
//     ICUBreakIteratorService()
//         : ICULocaleService(UNICODE_STRING("Break Iterator", 14))
//     {
//         UErrorCode status = U_ZERO_ERROR;
//         registerFactory(new ICUBreakIteratorFactory(), status);
//     }

//     virtual ~ICUBreakIteratorService();

//     virtual UObject* cloneInstance(UObject* instance) const {
//         return ((BreakIterator*)instance)->clone();
//     }

//     virtual UObject* handleDefault(const ICUServiceKey& key, UnicodeString* /*actualID*/, UErrorCode& status) const {
//         LocaleKey& lkey = (LocaleKey&)key;
//         int32_t kind = lkey.kind();
//         Locale loc;
//         lkey.currentLocale(loc);
//         return BreakIterator::makeInstance(loc, kind, status);
//     }

//     virtual UBool isDefault() const {
//         return countFactories() == 1;
//     }
// };

// ICUBreakIteratorService::~ICUBreakIteratorService() {}

// // -------------------------------------

// // defined in ucln_cmn.h
// U_NAMESPACE_END

// static icu::UInitOnce gInitOnceBrkiter = U_INITONCE_INITIALIZER;
// static icu::ICULocaleService* gService = NULL;



// /**
//  * Release all static memory held by breakiterator.
//  */
// U_CDECL_BEGIN
// static UBool U_CALLCONV breakiterator_cleanup(void) {
// #if !UCONFIG_NO_SERVICE
//     if (gService) {
//         delete gService;
//         gService = NULL;
//     }
//     gInitOnceBrkiter.reset();
// #endif
//     return TRUE;
// }
// U_CDECL_END
// U_NAMESPACE_BEGIN

// static void U_CALLCONV
// initService(void) {
//     gService = new ICUBreakIteratorService();
//     ucln_common_registerCleanup(UCLN_COMMON_BREAKITERATOR, breakiterator_cleanup);
// }

// static ICULocaleService*
// getService(void)
// {
//     umtx_initOnce(gInitOnceBrkiter, &initService);
//     return gService;
// }


// // -------------------------------------

// static inline UBool
// hasService(void)
// {
//     return !gInitOnceBrkiter.isReset() && getService() != NULL;
// }

// // -------------------------------------

// URegistryKey U_EXPORT2
// BreakIterator::registerInstance(BreakIterator* toAdopt, const Locale& locale, UBreakIteratorType kind, UErrorCode& status)
// {
//     ICULocaleService *service = getService();
//     if (service == NULL) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//         return NULL;
//     }
//     return service->registerInstance(toAdopt, locale, kind, status);
// }

// // -------------------------------------

// UBool U_EXPORT2
// BreakIterator::unregister(URegistryKey key, UErrorCode& status)
// {
//     if (U_SUCCESS(status)) {
//         if (hasService()) {
//             return gService->unregister(key, status);
//         }
//         status = U_MEMORY_ALLOCATION_ERROR;
//     }
//     return FALSE;
// }

// // -------------------------------------

// StringEnumeration* U_EXPORT2
// BreakIterator::getAvailableLocales(void)
// {
//     ICULocaleService *service = getService();
//     if (service == NULL) {
//         return NULL;
//     }
//     return service->getAvailableLocales();
// }
// #endif /* UCONFIG_NO_SERVICE */

// // -------------------------------------

// BreakIterator*
// BreakIterator::createInstance(const Locale& loc, int32_t kind, UErrorCode& status)
// {
//     if (U_FAILURE(status)) {
//         return NULL;
//     }

// #if !UCONFIG_NO_SERVICE
//     if (hasService()) {
//         Locale actualLoc("");
//         BreakIterator *result = (BreakIterator*)gService->get(loc, kind, &actualLoc, status);
//         // TODO: The way the service code works in ICU 2.8 is that if
//         // there is a real registered break iterator, the actualLoc
//         // will be populated, but if the handleDefault path is taken
//         // (because nothing is registered that can handle the
//         // requested locale) then the actualLoc comes back empty.  In
//         // that case, the returned object already has its actual/valid
//         // locale data populated (by makeInstance, which is what
//         // handleDefault calls), so we don't touch it.  YES, A COMMENT
//         // THIS LONG is a sign of bad code -- so the action item is to
//         // revisit this in ICU 3.0 and clean it up/fix it/remove it.
//         if (U_SUCCESS(status) && (result != NULL) && *actualLoc.getName() != 0) {
//             U_LOCALE_BASED(locBased, *result);
//             locBased.setLocaleIDs(actualLoc.getName(), actualLoc.getName());
//         }
//         return result;
//     }
//     else
// #endif
//     {
//         return makeInstance(loc, kind, status);
//     }
// }

// // -------------------------------------
// enum { kKeyValueLenMax = 32 };

// BreakIterator*
// BreakIterator::makeInstance(const Locale& loc, int32_t kind, UErrorCode& status)
// {

//     if (U_FAILURE(status)) {
//         return NULL;
//     }
//     char lbType[kKeyValueLenMax];

//     BreakIterator *result = NULL;
//     switch (kind) {
//     case UBRK_CHARACTER:
//         result = BreakIterator::buildInstance(loc, "grapheme", status);
//         break;
//     case UBRK_WORD:
//         result = BreakIterator::buildInstance(loc, "word", status);
//         break;
//     case UBRK_LINE:
//         uprv_strcpy(lbType, "line");
//         {
//             char lbKeyValue[kKeyValueLenMax] = {0};
//             UErrorCode kvStatus = U_ZERO_ERROR;
//             int32_t kLen = loc.getKeywordValue("lb", lbKeyValue, kKeyValueLenMax, kvStatus);
//             if (U_SUCCESS(kvStatus) && kLen > 0 && (uprv_strcmp(lbKeyValue,"strict")==0 || uprv_strcmp(lbKeyValue,"normal")==0 || uprv_strcmp(lbKeyValue,"loose")==0)) {
//                 uprv_strcat(lbType, "_");
//                 uprv_strcat(lbType, lbKeyValue);
//             }
//         }
//         result = BreakIterator::buildInstance(loc, lbType, status);
//         break;
//     case UBRK_SENTENCE:
//         result = BreakIterator::buildInstance(loc, "sentence", status);
// #if !UCONFIG_NO_FILTERED_BREAK_ITERATION
//         {
//             char ssKeyValue[kKeyValueLenMax] = {0};
//             UErrorCode kvStatus = U_ZERO_ERROR;
//             int32_t kLen = loc.getKeywordValue("ss", ssKeyValue, kKeyValueLenMax, kvStatus);
//             if (U_SUCCESS(kvStatus) && kLen > 0 && uprv_strcmp(ssKeyValue,"standard")==0) {
//                 FilteredBreakIteratorBuilder* fbiBuilder = FilteredBreakIteratorBuilder::createInstance(loc, kvStatus);
//                 if (U_SUCCESS(kvStatus)) {
//                     result = fbiBuilder->build(result, status);
//                     delete fbiBuilder;
//                 }
//             }
//         }
// #endif
//         break;
//     case UBRK_TITLE:
//         result = BreakIterator::buildInstance(loc, "title", status);
//         break;
//     default:
//         status = U_ILLEGAL_ARGUMENT_ERROR;
//     }

//     if (U_FAILURE(status)) {
//         return NULL;
//     }

//     return result;
// }

// Locale
// BreakIterator::getLocale(ULocDataLocaleType type, UErrorCode& status) const {
//     U_LOCALE_BASED(locBased, *this);
//     return locBased.getLocale(type, status);
// }

// const char *
// BreakIterator::getLocaleID(ULocDataLocaleType type, UErrorCode& status) const {
//     U_LOCALE_BASED(locBased, *this);
//     return locBased.getLocaleID(type, status);
// }


// // This implementation of getRuleStatus is a do-nothing stub, here to
// // provide a default implementation for any derived BreakIterator classes that
// // do not implement it themselves.
// int32_t BreakIterator::getRuleStatus() const {
//     return 0;
// }

// // This implementation of getRuleStatusVec is a do-nothing stub, here to
// // provide a default implementation for any derived BreakIterator classes that
// // do not implement it themselves.
// int32_t BreakIterator::getRuleStatusVec(int32_t *fillInVec, int32_t capacity, UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return 0;
//     }
//     if (capacity < 1) {
//         status = U_BUFFER_OVERFLOW_ERROR;
//         return 1;
//     }
//     *fillInVec = 0;
//     return 1;
// }

// BreakIterator::BreakIterator (const Locale& valid, const Locale& actual) {
//   U_LOCALE_BASED(locBased, (*this));
//   locBased.setLocaleIDs(valid, actual);
// }

// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */

//eof
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// bytesinkutil.cpp
// created: 2017sep14 Markus W. Scherer



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// edits.h
// created: 2016dec30 Markus W. Scherer




#if U_SHOW_CPLUSPLUS_API



/**
 * \file
 * \brief C++ API: C++ class Edits for low-level string transformations on styled text.
 */

U_NAMESPACE_BEGIN

class UnicodeString;

/**
 * Records lengths of string edits but not replacement text. Supports replacements, insertions, deletions
 * in linear progression. Does not support moving/reordering of text.
 *
 * There are two types of edits: <em>change edits</em> and <em>no-change edits</em>. Add edits to
 * instances of this class using {@link #addReplace(int32_t, int32_t)} (for change edits) and
 * {@link #addUnchanged(int32_t)} (for no-change edits). Change edits are retained with full granularity,
 * whereas adjacent no-change edits are always merged together. In no-change edits, there is a one-to-one
 * mapping between code points in the source and destination strings.
 *
 * After all edits have been added, instances of this class should be considered immutable, and an
 * {@link Edits::Iterator} can be used for queries.
 *
 * There are four flavors of Edits::Iterator:
 *
 * <ul>
 * <li>{@link #getFineIterator()} retains full granularity of change edits.
 * <li>{@link #getFineChangesIterator()} retains full granularity of change edits, and when calling
 * next() on the iterator, skips over no-change edits (unchanged regions).
 * <li>{@link #getCoarseIterator()} treats adjacent change edits as a single edit. (Adjacent no-change
 * edits are automatically merged during the construction phase.)
 * <li>{@link #getCoarseChangesIterator()} treats adjacent change edits as a single edit, and when
 * calling next() on the iterator, skips over no-change edits (unchanged regions).
 * </ul>
 *
 * For example, consider the string "abcßDeF", which case-folds to "abcssdef". This string has the
 * following fine edits:
 * <ul>
 * <li>abc ⇨ abc (no-change)
 * <li>ß ⇨ ss (change)
 * <li>D ⇨ d (change)
 * <li>e ⇨ e (no-change)
 * <li>F ⇨ f (change)
 * </ul>
 * and the following coarse edits (note how adjacent change edits get merged together):
 * <ul>
 * <li>abc ⇨ abc (no-change)
 * <li>ßD ⇨ ssd (change)
 * <li>e ⇨ e (no-change)
 * <li>F ⇨ f (change)
 * </ul>
 *
 * The "fine changes" and "coarse changes" iterators will step through only the change edits when their
 * `Edits::Iterator::next()` methods are called. They are identical to the non-change iterators when
 * their `Edits::Iterator::findSourceIndex()` or `Edits::Iterator::findDestinationIndex()`
 * methods are used to walk through the string.
 *
 * For examples of how to use this class, see the test `TestCaseMapEditsIteratorDocs` in
 * UCharacterCaseTest.java.
 *
 * An Edits object tracks a separate UErrorCode, but ICU string transformation functions
 * (e.g., case mapping functions) merge any such errors into their API's UErrorCode.
 *
 * @stable ICU 59
 */
class U_COMMON_API Edits U_FINAL : public UMemory {
public:
    /**
     * Constructs an empty object.
     * @stable ICU 59
     */
    Edits() :
            array(stackArray), capacity(STACK_CAPACITY), length(0), delta(0), numChanges(0),
            errorCode_(U_ZERO_ERROR) {}
    /**
     * Copy constructor.
     * @param other source edits
     * @stable ICU 60
     */
    Edits(const Edits &other) :
            array(stackArray), capacity(STACK_CAPACITY), length(other.length),
            delta(other.delta), numChanges(other.numChanges),
            errorCode_(other.errorCode_) {
        copyArray(other);
    }
    /**
     * Move constructor, might leave src empty.
     * This object will have the same contents that the source object had.
     * @param src source edits
     * @stable ICU 60
     */
    Edits(Edits &&src) U_NOEXCEPT :
            array(stackArray), capacity(STACK_CAPACITY), length(src.length),
            delta(src.delta), numChanges(src.numChanges),
            errorCode_(src.errorCode_) {
        moveArray(src);
    }

    /**
     * Destructor.
     * @stable ICU 59
     */
    ~Edits();

    /**
     * Assignment operator.
     * @param other source edits
     * @return *this
     * @stable ICU 60
     */
    Edits &operator=(const Edits &other);

    /**
     * Move assignment operator, might leave src empty.
     * This object will have the same contents that the source object had.
     * The behavior is undefined if *this and src are the same object.
     * @param src source edits
     * @return *this
     * @stable ICU 60
     */
    Edits &operator=(Edits &&src) U_NOEXCEPT;

    /**
     * Resets the data but may not release memory.
     * @stable ICU 59
     */
    void reset() U_NOEXCEPT;

    /**
     * Adds a no-change edit: a record for an unchanged segment of text.
     * Normally called from inside ICU string transformation functions, not user code.
     * @stable ICU 59
     */
    void addUnchanged(int32_t unchangedLength);
    /**
     * Adds a change edit: a record for a text replacement/insertion/deletion.
     * Normally called from inside ICU string transformation functions, not user code.
     * @stable ICU 59
     */
    void addReplace(int32_t oldLength, int32_t newLength);
    /**
     * Sets the UErrorCode if an error occurred while recording edits.
     * Preserves older error codes in the outErrorCode.
     * Normally called from inside ICU string transformation functions, not user code.
     * @param outErrorCode Set to an error code if it does not contain one already
     *                  and an error occurred while recording edits.
     *                  Otherwise unchanged.
     * @return TRUE if U_FAILURE(outErrorCode)
     * @stable ICU 59
     */
    UBool copyErrorTo(UErrorCode &outErrorCode) const;

    /**
     * How much longer is the new text compared with the old text?
     * @return new length minus old length
     * @stable ICU 59
     */
    int32_t lengthDelta() const { return delta; }
    /**
     * @return TRUE if there are any change edits
     * @stable ICU 59
     */
    UBool hasChanges() const { return numChanges != 0; }

    /**
     * @return the number of change edits
     * @stable ICU 60
     */
    int32_t numberOfChanges() const { return numChanges; }

    /**
     * Access to the list of edits.
     *
     * At any moment in time, an instance of this class points to a single edit: a "window" into a span
     * of the source string and the corresponding span of the destination string. The source string span
     * starts at {@link #sourceIndex()} and runs for {@link #oldLength()} chars; the destination string
     * span starts at {@link #destinationIndex()} and runs for {@link #newLength()} chars.
     *
     * The iterator can be moved between edits using the `next()`, `findSourceIndex(int32_t, UErrorCode &)`,
     * and `findDestinationIndex(int32_t, UErrorCode &)` methods.
     * Calling any of these methods mutates the iterator to make it point to the corresponding edit.
     *
     * For more information, see the documentation for {@link Edits}.
     *
     * @see getCoarseIterator
     * @see getFineIterator
     * @stable ICU 59
     */
    struct U_COMMON_API Iterator U_FINAL : public UMemory {
        /**
         * Default constructor, empty iterator.
         * @stable ICU 60
         */
        Iterator() :
                array(nullptr), index(0), length(0),
                remaining(0), onlyChanges_(FALSE), coarse(FALSE),
                dir(0), changed(FALSE), oldLength_(0), newLength_(0),
                srcIndex(0), replIndex(0), destIndex(0) {}
        /**
         * Copy constructor.
         * @stable ICU 59
         */
        Iterator(const Iterator &other) = default;
        /**
         * Assignment operator.
         * @stable ICU 59
         */
        Iterator &operator=(const Iterator &other) = default;

        /**
         * Advances the iterator to the next edit.
         * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
         *                  or else the function returns immediately. Check for U_FAILURE()
         *                  on output or use with function chaining. (See User Guide for details.)
         * @return TRUE if there is another edit
         * @stable ICU 59
         */
        UBool next(UErrorCode &errorCode) { return next(onlyChanges_, errorCode); }

        /**
         * Moves the iterator to the edit that contains the source index.
         * The source index may be found in a no-change edit
         * even if normal iteration would skip no-change edits.
         * Normal iteration can continue from a found edit.
         *
         * The iterator state before this search logically does not matter.
         * (It may affect the performance of the search.)
         *
         * The iterator state after this search is undefined
         * if the source index is out of bounds for the source string.
         *
         * @param i source index
         * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
         *                  or else the function returns immediately. Check for U_FAILURE()
         *                  on output or use with function chaining. (See User Guide for details.)
         * @return TRUE if the edit for the source index was found
         * @stable ICU 59
         */
        UBool findSourceIndex(int32_t i, UErrorCode &errorCode) {
            return findIndex(i, TRUE, errorCode) == 0;
        }

        /**
         * Moves the iterator to the edit that contains the destination index.
         * The destination index may be found in a no-change edit
         * even if normal iteration would skip no-change edits.
         * Normal iteration can continue from a found edit.
         *
         * The iterator state before this search logically does not matter.
         * (It may affect the performance of the search.)
         *
         * The iterator state after this search is undefined
         * if the source index is out of bounds for the source string.
         *
         * @param i destination index
         * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
         *                  or else the function returns immediately. Check for U_FAILURE()
         *                  on output or use with function chaining. (See User Guide for details.)
         * @return TRUE if the edit for the destination index was found
         * @stable ICU 60
         */
        UBool findDestinationIndex(int32_t i, UErrorCode &errorCode) {
            return findIndex(i, FALSE, errorCode) == 0;
        }

        /**
         * Computes the destination index corresponding to the given source index.
         * If the source index is inside a change edit (not at its start),
         * then the destination index at the end of that edit is returned,
         * since there is no information about index mapping inside a change edit.
         *
         * (This means that indexes to the start and middle of an edit,
         * for example around a grapheme cluster, are mapped to indexes
         * encompassing the entire edit.
         * The alternative, mapping an interior index to the start,
         * would map such an interval to an empty one.)
         *
         * This operation will usually but not always modify this object.
         * The iterator state after this search is undefined.
         *
         * @param i source index
         * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
         *                  or else the function returns immediately. Check for U_FAILURE()
         *                  on output or use with function chaining. (See User Guide for details.)
         * @return destination index; undefined if i is not 0..string length
         * @stable ICU 60
         */
        int32_t destinationIndexFromSourceIndex(int32_t i, UErrorCode &errorCode);

        /**
         * Computes the source index corresponding to the given destination index.
         * If the destination index is inside a change edit (not at its start),
         * then the source index at the end of that edit is returned,
         * since there is no information about index mapping inside a change edit.
         *
         * (This means that indexes to the start and middle of an edit,
         * for example around a grapheme cluster, are mapped to indexes
         * encompassing the entire edit.
         * The alternative, mapping an interior index to the start,
         * would map such an interval to an empty one.)
         *
         * This operation will usually but not always modify this object.
         * The iterator state after this search is undefined.
         *
         * @param i destination index
         * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
         *                  or else the function returns immediately. Check for U_FAILURE()
         *                  on output or use with function chaining. (See User Guide for details.)
         * @return source index; undefined if i is not 0..string length
         * @stable ICU 60
         */
        int32_t sourceIndexFromDestinationIndex(int32_t i, UErrorCode &errorCode);

        /**
         * Returns whether the edit currently represented by the iterator is a change edit.
         *
         * @return TRUE if this edit replaces oldLength() units with newLength() different ones.
         *         FALSE if oldLength units remain unchanged.
         * @stable ICU 59
         */
        UBool hasChange() const { return changed; }

        /**
         * The length of the current span in the source string, which starts at {@link #sourceIndex}.
         *
         * @return the number of units in the original string which are replaced or remain unchanged.
         * @stable ICU 59
         */
        int32_t oldLength() const { return oldLength_; }

        /**
         * The length of the current span in the destination string, which starts at
         * {@link #destinationIndex}, or in the replacement string, which starts at
         * {@link #replacementIndex}.
         *
         * @return the number of units in the modified string, if hasChange() is TRUE.
         *         Same as oldLength if hasChange() is FALSE.
         * @stable ICU 59
         */
        int32_t newLength() const { return newLength_; }

        /**
         * The start index of the current span in the source string; the span has length
         * {@link #oldLength}.
         *
         * @return the current index into the source string
         * @stable ICU 59
         */
        int32_t sourceIndex() const { return srcIndex; }

        /**
         * The start index of the current span in the replacement string; the span has length
         * {@link #newLength}. Well-defined only if the current edit is a change edit.
         *
         * The *replacement string* is the concatenation of all substrings of the destination
         * string corresponding to change edits.
         *
         * This method is intended to be used together with operations that write only replacement
         * characters (e.g. operations specifying the \ref U_OMIT_UNCHANGED_TEXT option).
         * The source string can then be modified in-place.
         *
         * @return the current index into the replacement-characters-only string,
         *         not counting unchanged spans
         * @stable ICU 59
         */
        int32_t replacementIndex() const {
            // TODO: Throw an exception if we aren't in a change edit?
            return replIndex;
        }

        /**
         * The start index of the current span in the destination string; the span has length
         * {@link #newLength}.
         *
         * @return the current index into the full destination string
         * @stable ICU 59
         */
        int32_t destinationIndex() const { return destIndex; }

#ifndef U_HIDE_INTERNAL_API
        /**
         * A string representation of the current edit represented by the iterator for debugging. You
         * should not depend on the contents of the return string.
         * @internal
         */
        UnicodeString& toString(UnicodeString& appendTo) const;
#endif  // U_HIDE_INTERNAL_API

    private:
        friend class Edits;

        Iterator(const uint16_t *a, int32_t len, UBool oc, UBool crs);

        int32_t readLength(int32_t head);
        void updateNextIndexes();
        void updatePreviousIndexes();
        UBool noNext();
        UBool next(UBool onlyChanges, UErrorCode &errorCode);
        UBool previous(UErrorCode &errorCode);
        /** @return -1: error or i<0; 0: found; 1: i>=string length */
        int32_t findIndex(int32_t i, UBool findSource, UErrorCode &errorCode);

        const uint16_t *array;
        int32_t index, length;
        // 0 if we are not within compressed equal-length changes.
        // Otherwise the number of remaining changes, including the current one.
        int32_t remaining;
        UBool onlyChanges_, coarse;

        int8_t dir;  // iteration direction: back(<0), initial(0), forward(>0)
        UBool changed;
        int32_t oldLength_, newLength_;
        int32_t srcIndex, replIndex, destIndex;
    };

    /**
     * Returns an Iterator for coarse-grained change edits
     * (adjacent change edits are treated as one).
     * Can be used to perform simple string updates.
     * Skips no-change edits.
     * @return an Iterator that merges adjacent changes.
     * @stable ICU 59
     */
    Iterator getCoarseChangesIterator() const {
        return Iterator(array, length, TRUE, TRUE);
    }

    /**
     * Returns an Iterator for coarse-grained change and no-change edits
     * (adjacent change edits are treated as one).
     * Can be used to perform simple string updates.
     * Adjacent change edits are treated as one edit.
     * @return an Iterator that merges adjacent changes.
     * @stable ICU 59
     */
    Iterator getCoarseIterator() const {
        return Iterator(array, length, FALSE, TRUE);
    }

    /**
     * Returns an Iterator for fine-grained change edits
     * (full granularity of change edits is retained).
     * Can be used for modifying styled text.
     * Skips no-change edits.
     * @return an Iterator that separates adjacent changes.
     * @stable ICU 59
     */
    Iterator getFineChangesIterator() const {
        return Iterator(array, length, TRUE, FALSE);
    }

    /**
     * Returns an Iterator for fine-grained change and no-change edits
     * (full granularity of change edits is retained).
     * Can be used for modifying styled text.
     * @return an Iterator that separates adjacent changes.
     * @stable ICU 59
     */
    Iterator getFineIterator() const {
        return Iterator(array, length, FALSE, FALSE);
    }

    /**
     * Merges the two input Edits and appends the result to this object.
     *
     * Consider two string transformations (for example, normalization and case mapping)
     * where each records Edits in addition to writing an output string.<br>
     * Edits ab reflect how substrings of input string a
     * map to substrings of intermediate string b.<br>
     * Edits bc reflect how substrings of intermediate string b
     * map to substrings of output string c.<br>
     * This function merges ab and bc such that the additional edits
     * recorded in this object reflect how substrings of input string a
     * map to substrings of output string c.
     *
     * If unrelated Edits are passed in where the output string of the first
     * has a different length than the input string of the second,
     * then a U_ILLEGAL_ARGUMENT_ERROR is reported.
     *
     * @param ab reflects how substrings of input string a
     *     map to substrings of intermediate string b.
     * @param bc reflects how substrings of intermediate string b
     *     map to substrings of output string c.
     * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
     *                  or else the function returns immediately. Check for U_FAILURE()
     *                  on output or use with function chaining. (See User Guide for details.)
     * @return *this, with the merged edits appended
     * @stable ICU 60
     */
    Edits &mergeAndAppend(const Edits &ab, const Edits &bc, UErrorCode &errorCode);

private:
    void releaseArray() U_NOEXCEPT;
    Edits &copyArray(const Edits &other);
    Edits &moveArray(Edits &src) U_NOEXCEPT;

    void setLastUnit(int32_t last) { array[length - 1] = (uint16_t)last; }
    int32_t lastUnit() const { return length > 0 ? array[length - 1] : 0xffff; }

    void append(int32_t r);
    UBool growArray();

    static const int32_t STACK_CAPACITY = 100;
    uint16_t *array;
    int32_t capacity;
    int32_t length;
    int32_t delta;
    int32_t numChanges;
    UErrorCode errorCode_;
    uint16_t stackArray[STACK_CAPACITY];
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */





// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// bytesinkutil.h
// created: 2017sep14 Markus W. Scherer







U_NAMESPACE_BEGIN

class ByteSink;
class CharString;
class Edits;

class U_COMMON_API ByteSinkUtil {
public:
    ByteSinkUtil() = delete;  // all static

    /** (length) bytes were mapped to valid (s16, s16Length). */
    static UBool appendChange(int32_t length,
                              const char16_t *s16, int32_t s16Length,
                              ByteSink &sink, Edits *edits, UErrorCode &errorCode);

    /** The bytes at [s, limit[ were mapped to valid (s16, s16Length). */
    static UBool appendChange(const uint8_t *s, const uint8_t *limit,
                              const char16_t *s16, int32_t s16Length,
                              ByteSink &sink, Edits *edits, UErrorCode &errorCode);

    /** (length) bytes were mapped/changed to valid code point c. */
    static void appendCodePoint(int32_t length, UChar32 c, ByteSink &sink, Edits *edits = nullptr);

    /** The few bytes at [src, nextSrc[ were mapped/changed to valid code point c. */
    static inline void appendCodePoint(const uint8_t *src, const uint8_t *nextSrc, UChar32 c,
                                       ByteSink &sink, Edits *edits = nullptr) {
        appendCodePoint((int32_t)(nextSrc - src), c, sink, edits);
    }

    /** Append the two-byte character (U+0080..U+07FF). */
    static void appendTwoBytes(UChar32 c, ByteSink &sink);

    static UBool appendUnchanged(const uint8_t *s, int32_t length,
                                 ByteSink &sink, uint32_t options, Edits *edits,
                                 UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return FALSE; }
        if (length > 0) { appendNonEmptyUnchanged(s, length, sink, options, edits); }
        return TRUE;
    }

    static UBool appendUnchanged(const uint8_t *s, const uint8_t *limit,
                                 ByteSink &sink, uint32_t options, Edits *edits,
                                 UErrorCode &errorCode);

private:
    static void appendNonEmptyUnchanged(const uint8_t *s, int32_t length,
                                        ByteSink &sink, uint32_t options, Edits *edits);
};

class U_COMMON_API CharStringByteSink : public ByteSink {
public:
    CharStringByteSink(CharString* dest);
    ~CharStringByteSink() override;

    CharStringByteSink() = delete;
    CharStringByteSink(const CharStringByteSink&) = delete;
    CharStringByteSink& operator=(const CharStringByteSink&) = delete;

    void Append(const char* bytes, int32_t n) override;

    char* GetAppendBuffer(int32_t min_capacity,
                          int32_t desired_capacity_hint,
                          char* scratch,
                          int32_t scratch_capacity,
                          int32_t* result_capacity) override;

private:
    CharString& dest_;
};

U_NAMESPACE_END

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (c) 2001-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*   Date        Name        Description
*   11/19/2001  aliu        Creation.
*   05/19/2010  markus      Rewritten from scratch
**********************************************************************
*/







U_NAMESPACE_BEGIN

// Windows needs us to DLL-export the MaybeStackArray template specialization,
// but MacOS X cannot handle it. Same as in digitlst.h.
#if !U_PLATFORM_IS_DARWIN_BASED
template class U_COMMON_API MaybeStackArray<char, 40>;
#endif

/**
 * ICU-internal char * string class.
 * This class does not assume or enforce any particular character encoding.
 * Raw bytes can be stored. The string object owns its characters.
 * A terminating NUL is stored, but the class does not prevent embedded NUL characters.
 *
 * This class wants to be convenient but is also deliberately minimalist.
 * Please do not add methods if they only add minor convenience.
 * For example:
 *   cs.data()[5]='a';  // no need for setCharAt(5, 'a')
 */
class U_COMMON_API CharString : public UMemory {
public:
    CharString() : len(0) { buffer[0]=0; }
    CharString(StringPiece s, UErrorCode &errorCode) : len(0) {
        buffer[0]=0;
        append(s, errorCode);
    }
    CharString(const CharString &s, UErrorCode &errorCode) : len(0) {
        buffer[0]=0;
        append(s, errorCode);
    }
    CharString(const char *s, int32_t sLength, UErrorCode &errorCode) : len(0) {
        buffer[0]=0;
        append(s, sLength, errorCode);
    }
    ~CharString() {}

    /**
     * Move constructor; might leave src in an undefined state.
     * This string will have the same contents and state that the source string had.
     */
    CharString(CharString &&src) U_NOEXCEPT;
    /**
     * Move assignment operator; might leave src in an undefined state.
     * This string will have the same contents and state that the source string had.
     * The behavior is undefined if *this and src are the same object.
     */
    CharString &operator=(CharString &&src) U_NOEXCEPT;

    /**
     * Replaces this string's contents with the other string's contents.
     * CharString does not support the standard copy constructor nor
     * the assignment operator, to make copies explicit and to
     * use a UErrorCode where memory allocations might be needed.
     */
    CharString &copyFrom(const CharString &other, UErrorCode &errorCode);

    UBool isEmpty() const { return len==0; }
    int32_t length() const { return len; }
    char operator[](int32_t index) const { return buffer[index]; }
    StringPiece toStringPiece() const { return StringPiece(buffer.getAlias(), len); }

    const char *data() const { return buffer.getAlias(); }
    char *data() { return buffer.getAlias(); }
    /**
     * Allocates length()+1 chars and copies the NUL-terminated data().
     * The caller must uprv_free() the result.
     */
    char *cloneData(UErrorCode &errorCode) const;

    bool operator==(StringPiece other) const {
        return len == other.length() && (len == 0 || uprv_memcmp(data(), other.data(), len) == 0);
    }
    bool operator!=(StringPiece other) const {
        return !operator==(other);
    }

    /** @return last index of c, or -1 if c is not in this string */
    int32_t lastIndexOf(char c) const;

    bool contains(StringPiece s) const;

    CharString &clear() { len=0; buffer[0]=0; return *this; }
    CharString &truncate(int32_t newLength);

    CharString &append(char c, UErrorCode &errorCode);
    CharString &append(StringPiece s, UErrorCode &errorCode) {
        return append(s.data(), s.length(), errorCode);
    }
    CharString &append(const CharString &s, UErrorCode &errorCode) {
        return append(s.data(), s.length(), errorCode);
    }
    CharString &append(const char *s, int32_t sLength, UErrorCode &status);
    /**
     * Returns a writable buffer for appending and writes the buffer's capacity to
     * resultCapacity. Guarantees resultCapacity>=minCapacity if U_SUCCESS().
     * There will additionally be space for a terminating NUL right at resultCapacity.
     * (This function is similar to ByteSink.GetAppendBuffer().)
     *
     * The returned buffer is only valid until the next write operation
     * on this string.
     *
     * After writing at most resultCapacity bytes, call append() with the
     * pointer returned from this function and the number of bytes written.
     *
     * @param minCapacity required minimum capacity of the returned buffer;
     *                    must be non-negative
     * @param desiredCapacityHint desired capacity of the returned buffer;
     *                            must be non-negative
     * @param resultCapacity will be set to the capacity of the returned buffer
     * @param errorCode in/out error code
     * @return a buffer with resultCapacity>=min_capacity
     */
    char *getAppendBuffer(int32_t minCapacity,
                          int32_t desiredCapacityHint,
                          int32_t &resultCapacity,
                          UErrorCode &errorCode);

    CharString &appendInvariantChars(const UnicodeString &s, UErrorCode &errorCode);
    CharString &appendInvariantChars(const UChar* uchars, int32_t ucharsLen, UErrorCode& errorCode);

    /**
     * Appends a filename/path part, e.g., a directory name.
     * First appends a U_FILE_SEP_CHAR if necessary.
     * Does nothing if s is empty.
     */
    CharString &appendPathPart(StringPiece s, UErrorCode &errorCode);

    /**
     * Appends a U_FILE_SEP_CHAR if this string is not empty
     * and does not already end with a U_FILE_SEP_CHAR or U_FILE_ALT_SEP_CHAR.
     */
    CharString &ensureEndsWithFileSeparator(UErrorCode &errorCode);

private:
    MaybeStackArray<char, 40> buffer;
    int32_t len;

    UBool ensureCapacity(int32_t capacity, int32_t desiredCapacityHint, UErrorCode &errorCode);

    CharString(const CharString &other); // forbid copying of this class
    CharString &operator=(const CharString &other); // forbid copying of this class
};

U_NAMESPACE_END





U_NAMESPACE_BEGIN

UBool
ByteSinkUtil::appendChange(int32_t length, const char16_t *s16, int32_t s16Length,
                           ByteSink &sink, Edits *edits, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return FALSE; }
    char scratch[200];
    int32_t s8Length = 0;
    for (int32_t i = 0; i < s16Length;) {
        int32_t capacity;
        int32_t desiredCapacity = s16Length - i;
        if (desiredCapacity < (INT32_MAX / 3)) {
            desiredCapacity *= 3;  // max 3 UTF-8 bytes per UTF-16 code unit
        } else if (desiredCapacity < (INT32_MAX / 2)) {
            desiredCapacity *= 2;
        } else {
            desiredCapacity = INT32_MAX;
        }
        char *buffer = sink.GetAppendBuffer(U8_MAX_LENGTH, desiredCapacity,
                                            scratch, UPRV_LENGTHOF(scratch), &capacity);
        capacity -= U8_MAX_LENGTH - 1;
        int32_t j = 0;
        for (; i < s16Length && j < capacity;) {
            UChar32 c;
            U16_NEXT_UNSAFE(s16, i, c);
            U8_APPEND_UNSAFE(buffer, j, c);
        }
        if (j > (INT32_MAX - s8Length)) {
            errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
            return FALSE;
        }
        sink.Append(buffer, j);
        s8Length += j;
    }
    if (edits != nullptr) {
        edits->addReplace(length, s8Length);
    }
    return TRUE;
}

UBool
ByteSinkUtil::appendChange(const uint8_t *s, const uint8_t *limit,
                           const char16_t *s16, int32_t s16Length,
                           ByteSink &sink, Edits *edits, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return FALSE; }
    if ((limit - s) > INT32_MAX) {
        errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
        return FALSE;
    }
    return appendChange((int32_t)(limit - s), s16, s16Length, sink, edits, errorCode);
}

void
ByteSinkUtil::appendCodePoint(int32_t length, UChar32 c, ByteSink &sink, Edits *edits) {
    char s8[U8_MAX_LENGTH];
    int32_t s8Length = 0;
    U8_APPEND_UNSAFE(s8, s8Length, c);
    if (edits != nullptr) {
        edits->addReplace(length, s8Length);
    }
    sink.Append(s8, s8Length);
}

// See unicode/utf8.h U8_APPEND_UNSAFE().
inline uint8_t bytesinkutil_getTwoByteLead(UChar32 c) { return (uint8_t)((c >> 6) | 0xc0); }
inline uint8_t bytesinkutil_getTwoByteTrail(UChar32 c) { return (uint8_t)((c & 0x3f) | 0x80); }

void
ByteSinkUtil::appendTwoBytes(UChar32 c, ByteSink &sink) {
    U_ASSERT(0x80 <= c && c <= 0x7ff);  // 2-byte UTF-8
    char s8[2] = { (char)bytesinkutil_getTwoByteLead(c), (char)bytesinkutil_getTwoByteTrail(c) };
    sink.Append(s8, 2);
}

void
ByteSinkUtil::appendNonEmptyUnchanged(const uint8_t *s, int32_t length,
                                      ByteSink &sink, uint32_t options, Edits *edits) {
    U_ASSERT(length > 0);
    if (edits != nullptr) {
        edits->addUnchanged(length);
    }
    if ((options & U_OMIT_UNCHANGED_TEXT) == 0) {
        sink.Append(reinterpret_cast<const char *>(s), length);
    }
}

UBool
ByteSinkUtil::appendUnchanged(const uint8_t *s, const uint8_t *limit,
                              ByteSink &sink, uint32_t options, Edits *edits,
                              UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return FALSE; }
    if ((limit - s) > INT32_MAX) {
        errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
        return FALSE;
    }
    int32_t length = (int32_t)(limit - s);
    if (length > 0) {
        appendNonEmptyUnchanged(s, length, sink, options, edits);
    }
    return TRUE;
}

CharStringByteSink::CharStringByteSink(CharString* dest) : dest_(*dest) {
}

CharStringByteSink::~CharStringByteSink() = default;

void
CharStringByteSink::Append(const char* bytes, int32_t n) {
    UErrorCode status = U_ZERO_ERROR;
    dest_.append(bytes, n, status);
    // Any errors are silently ignored.
}

char*
CharStringByteSink::GetAppendBuffer(int32_t min_capacity,
                                    int32_t desired_capacity_hint,
                                    char* scratch,
                                    int32_t scratch_capacity,
                                    int32_t* result_capacity) {
    if (min_capacity < 1 || scratch_capacity < min_capacity) {
        *result_capacity = 0;
        return nullptr;
    }

    UErrorCode status = U_ZERO_ERROR;
    char* result = dest_.getAppendBuffer(
            min_capacity,
            desired_capacity_hint,
            *result_capacity,
            status);
    if (U_SUCCESS(status)) {
        return result;
    }

    *result_capacity = scratch_capacity;
    return scratch;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
// Copyright (C) 2009-2011, International Business Machines
// Corporation and others. All Rights Reserved.
//
// Copyright 2007 Google Inc. All Rights Reserved.
// Author: sanjay@google.com (Sanjay Ghemawat)





U_NAMESPACE_BEGIN

ByteSink::~ByteSink() {}

char* ByteSink::GetAppendBuffer(int32_t min_capacity,
                                int32_t /*desired_capacity_hint*/,
                                char* scratch, int32_t scratch_capacity,
                                int32_t* result_capacity) {
  if (min_capacity < 1 || scratch_capacity < min_capacity) {
    *result_capacity = 0;
    return NULL;
  }
  *result_capacity = scratch_capacity;
  return scratch;
}

void ByteSink::Flush() {}

CheckedArrayByteSink::CheckedArrayByteSink(char* outbuf, int32_t capacity)
    : outbuf_(outbuf), capacity_(capacity < 0 ? 0 : capacity),
      size_(0), appended_(0), overflowed_(FALSE) {
}

CheckedArrayByteSink::~CheckedArrayByteSink() {}

CheckedArrayByteSink& CheckedArrayByteSink::Reset() {
  size_ = appended_ = 0;
  overflowed_ = FALSE;
  return *this;
}

void CheckedArrayByteSink::Append(const char* bytes, int32_t n) {
  if (n <= 0) {
    return;
  }
  if (n > (INT32_MAX - appended_)) {
    // TODO: Report as integer overflow, not merely buffer overflow.
    appended_ = INT32_MAX;
    overflowed_ = TRUE;
    return;
  }
  appended_ += n;
  int32_t available = capacity_ - size_;
  if (n > available) {
    n = available;
    overflowed_ = TRUE;
  }
  if (n > 0 && bytes != (outbuf_ + size_)) {
    uprv_memcpy(outbuf_ + size_, bytes, n);
  }
  size_ += n;
}

char* CheckedArrayByteSink::GetAppendBuffer(int32_t min_capacity,
                                            int32_t /*desired_capacity_hint*/,
                                            char* scratch,
                                            int32_t scratch_capacity,
                                            int32_t* result_capacity) {
  if (min_capacity < 1 || scratch_capacity < min_capacity) {
    *result_capacity = 0;
    return NULL;
  }
  int32_t available = capacity_ - size_;
  if (available >= min_capacity) {
    *result_capacity = available;
    return outbuf_ + size_;
  } else {
    *result_capacity = scratch_capacity;
    return scratch;
  }
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  bytestrie.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010sep25
*   created by: Markus W. Scherer
*/



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  bytestrie.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010sep25
*   created by: Markus W. Scherer
*/


/**
 * \file
 * \brief C++ API: Trie for mapping byte sequences to integer values.
 */



#if U_SHOW_CPLUSPLUS_API



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  udicttrie.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010dec17
*   created by: Markus W. Scherer
*/


/**
 * \file
 * \brief C API: Helper definitions for dictionary trie APIs.
 */




/**
 * Return values for BytesTrie::next(), UCharsTrie::next() and similar methods.
 * @see USTRINGTRIE_MATCHES
 * @see USTRINGTRIE_HAS_VALUE
 * @see USTRINGTRIE_HAS_NEXT
 * @stable ICU 4.8
 */
enum UStringTrieResult {
    /**
     * The input unit(s) did not continue a matching string.
     * Once current()/next() return USTRINGTRIE_NO_MATCH,
     * all further calls to current()/next() will also return USTRINGTRIE_NO_MATCH,
     * until the trie is reset to its original state or to a saved state.
     * @stable ICU 4.8
     */
    USTRINGTRIE_NO_MATCH,
    /**
     * The input unit(s) continued a matching string
     * but there is no value for the string so far.
     * (It is a prefix of a longer string.)
     * @stable ICU 4.8
     */
    USTRINGTRIE_NO_VALUE,
    /**
     * The input unit(s) continued a matching string
     * and there is a value for the string so far.
     * This value will be returned by getValue().
     * No further input byte/unit can continue a matching string.
     * @stable ICU 4.8
     */
    USTRINGTRIE_FINAL_VALUE,
    /**
     * The input unit(s) continued a matching string
     * and there is a value for the string so far.
     * This value will be returned by getValue().
     * Another input byte/unit can continue a matching string.
     * @stable ICU 4.8
     */
    USTRINGTRIE_INTERMEDIATE_VALUE
};

/**
 * Same as (result!=USTRINGTRIE_NO_MATCH).
 * @param result A result from BytesTrie::first(), UCharsTrie::next() etc.
 * @return true if the input bytes/units so far are part of a matching string/byte sequence.
 * @stable ICU 4.8
 */
#define USTRINGTRIE_MATCHES(result) ((result)!=USTRINGTRIE_NO_MATCH)

/**
 * Equivalent to (result==USTRINGTRIE_INTERMEDIATE_VALUE || result==USTRINGTRIE_FINAL_VALUE) but
 * this macro evaluates result exactly once.
 * @param result A result from BytesTrie::first(), UCharsTrie::next() etc.
 * @return true if there is a value for the input bytes/units so far.
 * @see BytesTrie::getValue
 * @see UCharsTrie::getValue
 * @stable ICU 4.8
 */
#define USTRINGTRIE_HAS_VALUE(result) ((result)>=USTRINGTRIE_FINAL_VALUE)

/**
 * Equivalent to (result==USTRINGTRIE_NO_VALUE || result==USTRINGTRIE_INTERMEDIATE_VALUE) but
 * this macro evaluates result exactly once.
 * @param result A result from BytesTrie::first(), UCharsTrie::next() etc.
 * @return true if another input byte/unit can continue a matching string.
 * @stable ICU 4.8
 */
#define USTRINGTRIE_HAS_NEXT(result) ((result)&1)



U_NAMESPACE_BEGIN

class ByteSink;
class BytesTrieBuilder;
class CharString;
class UVector32;

/**
 * Light-weight, non-const reader class for a BytesTrie.
 * Traverses a byte-serialized data structure with minimal state,
 * for mapping byte sequences to non-negative integer values.
 *
 * This class owns the serialized trie data only if it was constructed by
 * the builder's build() method.
 * The public constructor and the copy constructor only alias the data (only copy the pointer).
 * There is no assignment operator.
 *
 * This class is not intended for public subclassing.
 * @stable ICU 4.8
 */
class U_COMMON_API BytesTrie : public UMemory {
public:
    /**
     * Constructs a BytesTrie reader instance.
     *
     * The trieBytes must contain a copy of a byte sequence from the BytesTrieBuilder,
     * starting with the first byte of that sequence.
     * The BytesTrie object will not read more bytes than
     * the BytesTrieBuilder generated in the corresponding build() call.
     *
     * The array is not copied/cloned and must not be modified while
     * the BytesTrie object is in use.
     *
     * @param trieBytes The byte array that contains the serialized trie.
     * @stable ICU 4.8
     */
    BytesTrie(const void *trieBytes)
            : ownedArray_(NULL), bytes_(static_cast<const uint8_t *>(trieBytes)),
              pos_(bytes_), remainingMatchLength_(-1) {}

    /**
     * Destructor.
     * @stable ICU 4.8
     */
    ~BytesTrie();

    /**
     * Copy constructor, copies the other trie reader object and its state,
     * but not the byte array which will be shared. (Shallow copy.)
     * @param other Another BytesTrie object.
     * @stable ICU 4.8
     */
    BytesTrie(const BytesTrie &other)
            : ownedArray_(NULL), bytes_(other.bytes_),
              pos_(other.pos_), remainingMatchLength_(other.remainingMatchLength_) {}

    /**
     * Resets this trie to its initial state.
     * @return *this
     * @stable ICU 4.8
     */
    BytesTrie &reset() {
        pos_=bytes_;
        remainingMatchLength_=-1;
        return *this;
    }

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns the state of this trie as a 64-bit integer.
     * The state value is never 0.
     *
     * @return opaque state value
     * @see resetToState64
     * @draft ICU 65
     */
    uint64_t getState64() const {
        return (static_cast<uint64_t>(remainingMatchLength_ + 2) << kState64RemainingShift) |
            (uint64_t)(pos_ - bytes_);
    }

    /**
     * Resets this trie to the saved state.
     * Unlike resetToState(State), the 64-bit state value
     * must be from getState64() from the same trie object or
     * from one initialized the exact same way.
     * Because of no validation, this method is faster.
     *
     * @param state The opaque trie state value from getState64().
     * @return *this
     * @see getState64
     * @see resetToState
     * @see reset
     * @draft ICU 65
     */
    BytesTrie &resetToState64(uint64_t state) {
        remainingMatchLength_ = static_cast<int32_t>(state >> kState64RemainingShift) - 2;
        pos_ = bytes_ + (state & kState64PosMask);
        return *this;
    }
#endif  /* U_HIDE_DRAFT_API */

    /**
     * BytesTrie state object, for saving a trie's current state
     * and resetting the trie back to this state later.
     * @stable ICU 4.8
     */
    class State : public UMemory {
    public:
        /**
         * Constructs an empty State.
         * @stable ICU 4.8
         */
        State() { bytes=NULL; }
    private:
        friend class BytesTrie;

        const uint8_t *bytes;
        const uint8_t *pos;
        int32_t remainingMatchLength;
    };

    /**
     * Saves the state of this trie.
     * @param state The State object to hold the trie's state.
     * @return *this
     * @see resetToState
     * @stable ICU 4.8
     */
    const BytesTrie &saveState(State &state) const {
        state.bytes=bytes_;
        state.pos=pos_;
        state.remainingMatchLength=remainingMatchLength_;
        return *this;
    }

    /**
     * Resets this trie to the saved state.
     * If the state object contains no state, or the state of a different trie,
     * then this trie remains unchanged.
     * @param state The State object which holds a saved trie state.
     * @return *this
     * @see saveState
     * @see reset
     * @stable ICU 4.8
     */
    BytesTrie &resetToState(const State &state) {
        if(bytes_==state.bytes && bytes_!=NULL) {
            pos_=state.pos;
            remainingMatchLength_=state.remainingMatchLength;
        }
        return *this;
    }

    /**
     * Determines whether the byte sequence so far matches, whether it has a value,
     * and whether another input byte can continue a matching byte sequence.
     * @return The match/value Result.
     * @stable ICU 4.8
     */
    UStringTrieResult current() const;

    /**
     * Traverses the trie from the initial state for this input byte.
     * Equivalent to reset().next(inByte).
     * @param inByte Input byte value. Values -0x100..-1 are treated like 0..0xff.
     *               Values below -0x100 and above 0xff will never match.
     * @return The match/value Result.
     * @stable ICU 4.8
     */
    inline UStringTrieResult first(int32_t inByte) {
        remainingMatchLength_=-1;
        if(inByte<0) {
            inByte+=0x100;
        }
        return nextImpl(bytes_, inByte);
    }

    /**
     * Traverses the trie from the current state for this input byte.
     * @param inByte Input byte value. Values -0x100..-1 are treated like 0..0xff.
     *               Values below -0x100 and above 0xff will never match.
     * @return The match/value Result.
     * @stable ICU 4.8
     */
    UStringTrieResult next(int32_t inByte);

    /**
     * Traverses the trie from the current state for this byte sequence.
     * Equivalent to
     * \code
     * Result result=current();
     * for(each c in s)
     *   if(!USTRINGTRIE_HAS_NEXT(result)) return USTRINGTRIE_NO_MATCH;
     *   result=next(c);
     * return result;
     * \endcode
     * @param s A string or byte sequence. Can be NULL if length is 0.
     * @param length The length of the byte sequence. Can be -1 if NUL-terminated.
     * @return The match/value Result.
     * @stable ICU 4.8
     */
    UStringTrieResult next(const char *s, int32_t length);

    /**
     * Returns a matching byte sequence's value if called immediately after
     * current()/first()/next() returned USTRINGTRIE_INTERMEDIATE_VALUE or USTRINGTRIE_FINAL_VALUE.
     * getValue() can be called multiple times.
     *
     * Do not call getValue() after USTRINGTRIE_NO_MATCH or USTRINGTRIE_NO_VALUE!
     * @return The value for the byte sequence so far.
     * @stable ICU 4.8
     */
    inline int32_t getValue() const {
        const uint8_t *pos=pos_;
        int32_t leadByte=*pos++;
        // U_ASSERT(leadByte>=kMinValueLead);
        return readValue(pos, leadByte>>1);
    }

    /**
     * Determines whether all byte sequences reachable from the current state
     * map to the same value.
     * @param uniqueValue Receives the unique value, if this function returns TRUE.
     *                    (output-only)
     * @return TRUE if all byte sequences reachable from the current state
     *         map to the same value.
     * @stable ICU 4.8
     */
    inline UBool hasUniqueValue(int32_t &uniqueValue) const {
        const uint8_t *pos=pos_;
        // Skip the rest of a pending linear-match node.
        return pos!=NULL && findUniqueValue(pos+remainingMatchLength_+1, FALSE, uniqueValue);
    }

    /**
     * Finds each byte which continues the byte sequence from the current state.
     * That is, each byte b for which it would be next(b)!=USTRINGTRIE_NO_MATCH now.
     * @param out Each next byte is appended to this object.
     *            (Only uses the out.Append(s, length) method.)
     * @return the number of bytes which continue the byte sequence from here
     * @stable ICU 4.8
     */
    int32_t getNextBytes(ByteSink &out) const;

    /**
     * Iterator for all of the (byte sequence, value) pairs in a BytesTrie.
     * @stable ICU 4.8
     */
    class U_COMMON_API Iterator : public UMemory {
    public:
        /**
         * Iterates from the root of a byte-serialized BytesTrie.
         * @param trieBytes The trie bytes.
         * @param maxStringLength If 0, the iterator returns full strings/byte sequences.
         *                        Otherwise, the iterator returns strings with this maximum length.
         * @param errorCode Standard ICU error code. Its input value must
         *                  pass the U_SUCCESS() test, or else the function returns
         *                  immediately. Check for U_FAILURE() on output or use with
         *                  function chaining. (See User Guide for details.)
         * @stable ICU 4.8
         */
        Iterator(const void *trieBytes, int32_t maxStringLength, UErrorCode &errorCode);

        /**
         * Iterates from the current state of the specified BytesTrie.
         * @param trie The trie whose state will be copied for iteration.
         * @param maxStringLength If 0, the iterator returns full strings/byte sequences.
         *                        Otherwise, the iterator returns strings with this maximum length.
         * @param errorCode Standard ICU error code. Its input value must
         *                  pass the U_SUCCESS() test, or else the function returns
         *                  immediately. Check for U_FAILURE() on output or use with
         *                  function chaining. (See User Guide for details.)
         * @stable ICU 4.8
         */
        Iterator(const BytesTrie &trie, int32_t maxStringLength, UErrorCode &errorCode);

        /**
         * Destructor.
         * @stable ICU 4.8
         */
        ~Iterator();

        /**
         * Resets this iterator to its initial state.
         * @return *this
         * @stable ICU 4.8
         */
        Iterator &reset();

        /**
         * @return TRUE if there are more elements.
         * @stable ICU 4.8
         */
        UBool hasNext() const;

        /**
         * Finds the next (byte sequence, value) pair if there is one.
         *
         * If the byte sequence is truncated to the maximum length and does not
         * have a real value, then the value is set to -1.
         * In this case, this "not a real value" is indistinguishable from
         * a real value of -1.
         * @param errorCode Standard ICU error code. Its input value must
         *                  pass the U_SUCCESS() test, or else the function returns
         *                  immediately. Check for U_FAILURE() on output or use with
         *                  function chaining. (See User Guide for details.)
         * @return TRUE if there is another element.
         * @stable ICU 4.8
         */
        UBool next(UErrorCode &errorCode);

        /**
         * @return The NUL-terminated byte sequence for the last successful next().
         * @stable ICU 4.8
         */
        StringPiece getString() const;
        /**
         * @return The value for the last successful next().
         * @stable ICU 4.8
         */
        int32_t getValue() const { return value_; }

    private:
        UBool truncateAndStop();

        const uint8_t *branchNext(const uint8_t *pos, int32_t length, UErrorCode &errorCode);

        const uint8_t *bytes_;
        const uint8_t *pos_;
        const uint8_t *initialPos_;
        int32_t remainingMatchLength_;
        int32_t initialRemainingMatchLength_;

        CharString *str_;
        int32_t maxLength_;
        int32_t value_;

        // The stack stores pairs of integers for backtracking to another
        // outbound edge of a branch node.
        // The first integer is an offset from bytes_.
        // The second integer has the str_->length() from before the node in bits 15..0,
        // and the remaining branch length in bits 24..16. (Bits 31..25 are unused.)
        // (We could store the remaining branch length minus 1 in bits 23..16 and not use bits 31..24,
        // but the code looks more confusing that way.)
        UVector32 *stack_;
    };

private:
    friend class BytesTrieBuilder;

    /**
     * Constructs a BytesTrie reader instance.
     * Unlike the public constructor which just aliases an array,
     * this constructor adopts the builder's array.
     * This constructor is only called by the builder.
     */
    BytesTrie(void *adoptBytes, const void *trieBytes)
            : ownedArray_(static_cast<uint8_t *>(adoptBytes)),
              bytes_(static_cast<const uint8_t *>(trieBytes)),
              pos_(bytes_), remainingMatchLength_(-1) {}

    // No assignment operator.
    BytesTrie &operator=(const BytesTrie &other);

    inline void stop() {
        pos_=NULL;
    }

    // Reads a compact 32-bit integer.
    // pos is already after the leadByte, and the lead byte is already shifted right by 1.
    static int32_t readValue(const uint8_t *pos, int32_t leadByte);
    static inline const uint8_t *skipValue(const uint8_t *pos, int32_t leadByte) {
        // U_ASSERT(leadByte>=kMinValueLead);
        if(leadByte>=(kMinTwoByteValueLead<<1)) {
            if(leadByte<(kMinThreeByteValueLead<<1)) {
                ++pos;
            } else if(leadByte<(kFourByteValueLead<<1)) {
                pos+=2;
            } else {
                pos+=3+((leadByte>>1)&1);
            }
        }
        return pos;
    }
    static inline const uint8_t *skipValue(const uint8_t *pos) {
        int32_t leadByte=*pos++;
        return skipValue(pos, leadByte);
    }

    // Reads a jump delta and jumps.
    static const uint8_t *jumpByDelta(const uint8_t *pos);

    static inline const uint8_t *skipDelta(const uint8_t *pos) {
        int32_t delta=*pos++;
        if(delta>=kMinTwoByteDeltaLead) {
            if(delta<kMinThreeByteDeltaLead) {
                ++pos;
            } else if(delta<kFourByteDeltaLead) {
                pos+=2;
            } else {
                pos+=3+(delta&1);
            }
        }
        return pos;
    }

    static inline UStringTrieResult valueResult(int32_t node) {
        return (UStringTrieResult)(USTRINGTRIE_INTERMEDIATE_VALUE-(node&kValueIsFinal));
    }

    // Handles a branch node for both next(byte) and next(string).
    UStringTrieResult branchNext(const uint8_t *pos, int32_t length, int32_t inByte);

    // Requires remainingLength_<0.
    UStringTrieResult nextImpl(const uint8_t *pos, int32_t inByte);

    // Helper functions for hasUniqueValue().
    // Recursively finds a unique value (or whether there is not a unique one)
    // from a branch.
    static const uint8_t *findUniqueValueFromBranch(const uint8_t *pos, int32_t length,
                                                    UBool haveUniqueValue, int32_t &uniqueValue);
    // Recursively finds a unique value (or whether there is not a unique one)
    // starting from a position on a node lead byte.
    static UBool findUniqueValue(const uint8_t *pos, UBool haveUniqueValue, int32_t &uniqueValue);

    // Helper functions for getNextBytes().
    // getNextBytes() when pos is on a branch node.
    static void getNextBranchBytes(const uint8_t *pos, int32_t length, ByteSink &out);
    static void append(ByteSink &out, int c);

    // BytesTrie data structure
    //
    // The trie consists of a series of byte-serialized nodes for incremental
    // string/byte sequence matching. The root node is at the beginning of the trie data.
    //
    // Types of nodes are distinguished by their node lead byte ranges.
    // After each node, except a final-value node, another node follows to
    // encode match values or continue matching further bytes.
    //
    // Node types:
    //  - Value node: Stores a 32-bit integer in a compact, variable-length format.
    //    The value is for the string/byte sequence so far.
    //    One node bit indicates whether the value is final or whether
    //    matching continues with the next node.
    //  - Linear-match node: Matches a number of bytes.
    //  - Branch node: Branches to other nodes according to the current input byte.
    //    The node byte is the length of the branch (number of bytes to select from)
    //    minus 1. It is followed by a sub-node:
    //    - If the length is at most kMaxBranchLinearSubNodeLength, then
    //      there are length-1 (key, value) pairs and then one more comparison byte.
    //      If one of the key bytes matches, then the value is either a final value for
    //      the string/byte sequence so far, or a "jump" delta to the next node.
    //      If the last byte matches, then matching continues with the next node.
    //      (Values have the same encoding as value nodes.)
    //    - If the length is greater than kMaxBranchLinearSubNodeLength, then
    //      there is one byte and one "jump" delta.
    //      If the input byte is less than the sub-node byte, then "jump" by delta to
    //      the next sub-node which will have a length of length/2.
    //      (The delta has its own compact encoding.)
    //      Otherwise, skip the "jump" delta to the next sub-node
    //      which will have a length of length-length/2.

    // Node lead byte values.

    // 00..0f: Branch node. If node!=0 then the length is node+1, otherwise
    // the length is one more than the next byte.

    // For a branch sub-node with at most this many entries, we drop down
    // to a linear search.
    static const int32_t kMaxBranchLinearSubNodeLength=5;

    // 10..1f: Linear-match node, match 1..16 bytes and continue reading the next node.
    static const int32_t kMinLinearMatch=0x10;
    static const int32_t kMaxLinearMatchLength=0x10;

    // 20..ff: Variable-length value node.
    // If odd, the value is final. (Otherwise, intermediate value or jump delta.)
    // Then shift-right by 1 bit.
    // The remaining lead byte value indicates the number of following bytes (0..4)
    // and contains the value's top bits.
    static const int32_t kMinValueLead=kMinLinearMatch+kMaxLinearMatchLength;  // 0x20
    // It is a final value if bit 0 is set.
    static const int32_t kValueIsFinal=1;

    // Compact value: After testing bit 0, shift right by 1 and then use the following thresholds.
    static const int32_t kMinOneByteValueLead=kMinValueLead/2;  // 0x10
    static const int32_t kMaxOneByteValue=0x40;  // At least 6 bits in the first byte.

    static const int32_t kMinTwoByteValueLead=kMinOneByteValueLead+kMaxOneByteValue+1;  // 0x51
    static const int32_t kMaxTwoByteValue=0x1aff;

    static const int32_t kMinThreeByteValueLead=kMinTwoByteValueLead+(kMaxTwoByteValue>>8)+1;  // 0x6c
    static const int32_t kFourByteValueLead=0x7e;

    // A little more than Unicode code points. (0x11ffff)
    static const int32_t kMaxThreeByteValue=((kFourByteValueLead-kMinThreeByteValueLead)<<16)-1;

    static const int32_t kFiveByteValueLead=0x7f;

    // Compact delta integers.
    static const int32_t kMaxOneByteDelta=0xbf;
    static const int32_t kMinTwoByteDeltaLead=kMaxOneByteDelta+1;  // 0xc0
    static const int32_t kMinThreeByteDeltaLead=0xf0;
    static const int32_t kFourByteDeltaLead=0xfe;
    static const int32_t kFiveByteDeltaLead=0xff;

    static const int32_t kMaxTwoByteDelta=((kMinThreeByteDeltaLead-kMinTwoByteDeltaLead)<<8)-1;  // 0x2fff
    static const int32_t kMaxThreeByteDelta=((kFourByteDeltaLead-kMinThreeByteDeltaLead)<<16)-1;  // 0xdffff

    // For getState64():
    // The remainingMatchLength_ is -1..14=(kMaxLinearMatchLength=0x10)-2
    // so we need at least 5 bits for that.
    // We add 2 to store it as a positive value 1..16=kMaxLinearMatchLength.
    static constexpr int32_t kState64RemainingShift = 59;
    static constexpr uint64_t kState64PosMask = (UINT64_C(1) << kState64RemainingShift) - 1;

    uint8_t *ownedArray_;

    // Fixed value referencing the BytesTrie bytes.
    const uint8_t *bytes_;

    // Iterator variables.

    // Pointer to next trie byte to read. NULL if no more matches.
    const uint8_t *pos_;
    // Remaining length of a linear-match node, minus 1. Negative if not in such a node.
    int32_t remainingMatchLength_;
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */






U_NAMESPACE_BEGIN

BytesTrie::~BytesTrie() {
    uprv_free(ownedArray_);
}

// lead byte already shifted right by 1.
int32_t
BytesTrie::readValue(const uint8_t *pos, int32_t leadByte) {
    int32_t value;
    if(leadByte<kMinTwoByteValueLead) {
        value=leadByte-kMinOneByteValueLead;
    } else if(leadByte<kMinThreeByteValueLead) {
        value=((leadByte-kMinTwoByteValueLead)<<8)|*pos;
    } else if(leadByte<kFourByteValueLead) {
        value=((leadByte-kMinThreeByteValueLead)<<16)|(pos[0]<<8)|pos[1];
    } else if(leadByte==kFourByteValueLead) {
        value=(pos[0]<<16)|(pos[1]<<8)|pos[2];
    } else {
        value=(pos[0]<<24)|(pos[1]<<16)|(pos[2]<<8)|pos[3];
    }
    return value;
}

const uint8_t *
BytesTrie::jumpByDelta(const uint8_t *pos) {
    int32_t delta=*pos++;
    if(delta<kMinTwoByteDeltaLead) {
        // nothing to do
    } else if(delta<kMinThreeByteDeltaLead) {
        delta=((delta-kMinTwoByteDeltaLead)<<8)|*pos++;
    } else if(delta<kFourByteDeltaLead) {
        delta=((delta-kMinThreeByteDeltaLead)<<16)|(pos[0]<<8)|pos[1];
        pos+=2;
    } else if(delta==kFourByteDeltaLead) {
        delta=(pos[0]<<16)|(pos[1]<<8)|pos[2];
        pos+=3;
    } else {
        delta=(pos[0]<<24)|(pos[1]<<16)|(pos[2]<<8)|pos[3];
        pos+=4;
    }
    return pos+delta;
}

UStringTrieResult
BytesTrie::current() const {
    const uint8_t *pos=pos_;
    if(pos==NULL) {
        return USTRINGTRIE_NO_MATCH;
    } else {
        int32_t node;
        return (remainingMatchLength_<0 && (node=*pos)>=kMinValueLead) ?
                valueResult(node) : USTRINGTRIE_NO_VALUE;
    }
}

UStringTrieResult
BytesTrie::branchNext(const uint8_t *pos, int32_t length, int32_t inByte) {
    // Branch according to the current byte.
    if(length==0) {
        length=*pos++;
    }
    ++length;
    // The length of the branch is the number of bytes to select from.
    // The data structure encodes a binary search.
    while(length>kMaxBranchLinearSubNodeLength) {
        if(inByte<*pos++) {
            length>>=1;
            pos=jumpByDelta(pos);
        } else {
            length=length-(length>>1);
            pos=skipDelta(pos);
        }
    }
    // Drop down to linear search for the last few bytes.
    // length>=2 because the loop body above sees length>kMaxBranchLinearSubNodeLength>=3
    // and divides length by 2.
    do {
        if(inByte==*pos++) {
            UStringTrieResult result;
            int32_t node=*pos;
            U_ASSERT(node>=kMinValueLead);
            if(node&kValueIsFinal) {
                // Leave the final value for getValue() to read.
                result=USTRINGTRIE_FINAL_VALUE;
            } else {
                // Use the non-final value as the jump delta.
                ++pos;
                // int32_t delta=readValue(pos, node>>1);
                node>>=1;
                int32_t delta;
                if(node<kMinTwoByteValueLead) {
                    delta=node-kMinOneByteValueLead;
                } else if(node<kMinThreeByteValueLead) {
                    delta=((node-kMinTwoByteValueLead)<<8)|*pos++;
                } else if(node<kFourByteValueLead) {
                    delta=((node-kMinThreeByteValueLead)<<16)|(pos[0]<<8)|pos[1];
                    pos+=2;
                } else if(node==kFourByteValueLead) {
                    delta=(pos[0]<<16)|(pos[1]<<8)|pos[2];
                    pos+=3;
                } else {
                    delta=(pos[0]<<24)|(pos[1]<<16)|(pos[2]<<8)|pos[3];
                    pos+=4;
                }
                // end readValue()
                pos+=delta;
                node=*pos;
                result= node>=kMinValueLead ? valueResult(node) : USTRINGTRIE_NO_VALUE;
            }
            pos_=pos;
            return result;
        }
        --length;
        pos=skipValue(pos);
    } while(length>1);
    if(inByte==*pos++) {
        pos_=pos;
        int32_t node=*pos;
        return node>=kMinValueLead ? valueResult(node) : USTRINGTRIE_NO_VALUE;
    } else {
        stop();
        return USTRINGTRIE_NO_MATCH;
    }
}

UStringTrieResult
BytesTrie::nextImpl(const uint8_t *pos, int32_t inByte) {
    for(;;) {
        int32_t node=*pos++;
        if(node<kMinLinearMatch) {
            return branchNext(pos, node, inByte);
        } else if(node<kMinValueLead) {
            // Match the first of length+1 bytes.
            int32_t length=node-kMinLinearMatch;  // Actual match length minus 1.
            if(inByte==*pos++) {
                remainingMatchLength_=--length;
                pos_=pos;
                return (length<0 && (node=*pos)>=kMinValueLead) ?
                        valueResult(node) : USTRINGTRIE_NO_VALUE;
            } else {
                // No match.
                break;
            }
        } else if(node&kValueIsFinal) {
            // No further matching bytes.
            break;
        } else {
            // Skip intermediate value.
            pos=skipValue(pos, node);
            // The next node must not also be a value node.
            U_ASSERT(*pos<kMinValueLead);
        }
    }
    stop();
    return USTRINGTRIE_NO_MATCH;
}

UStringTrieResult
BytesTrie::next(int32_t inByte) {
    const uint8_t *pos=pos_;
    if(pos==NULL) {
        return USTRINGTRIE_NO_MATCH;
    }
    if(inByte<0) {
        inByte+=0x100;
    }
    int32_t length=remainingMatchLength_;  // Actual remaining match length minus 1.
    if(length>=0) {
        // Remaining part of a linear-match node.
        if(inByte==*pos++) {
            remainingMatchLength_=--length;
            pos_=pos;
            int32_t node;
            return (length<0 && (node=*pos)>=kMinValueLead) ?
                    valueResult(node) : USTRINGTRIE_NO_VALUE;
        } else {
            stop();
            return USTRINGTRIE_NO_MATCH;
        }
    }
    return nextImpl(pos, inByte);
}

UStringTrieResult
BytesTrie::next(const char *s, int32_t sLength) {
    if(sLength<0 ? *s==0 : sLength==0) {
        // Empty input.
        return current();
    }
    const uint8_t *pos=pos_;
    if(pos==NULL) {
        return USTRINGTRIE_NO_MATCH;
    }
    int32_t length=remainingMatchLength_;  // Actual remaining match length minus 1.
    for(;;) {
        // Fetch the next input byte, if there is one.
        // Continue a linear-match node without rechecking sLength<0.
        int32_t inByte;
        if(sLength<0) {
            for(;;) {
                if((inByte=*s++)==0) {
                    remainingMatchLength_=length;
                    pos_=pos;
                    int32_t node;
                    return (length<0 && (node=*pos)>=kMinValueLead) ?
                            valueResult(node) : USTRINGTRIE_NO_VALUE;
                }
                if(length<0) {
                    remainingMatchLength_=length;
                    break;
                }
                if(inByte!=*pos) {
                    stop();
                    return USTRINGTRIE_NO_MATCH;
                }
                ++pos;
                --length;
            }
        } else {
            for(;;) {
                if(sLength==0) {
                    remainingMatchLength_=length;
                    pos_=pos;
                    int32_t node;
                    return (length<0 && (node=*pos)>=kMinValueLead) ?
                            valueResult(node) : USTRINGTRIE_NO_VALUE;
                }
                inByte=*s++;
                --sLength;
                if(length<0) {
                    remainingMatchLength_=length;
                    break;
                }
                if(inByte!=*pos) {
                    stop();
                    return USTRINGTRIE_NO_MATCH;
                }
                ++pos;
                --length;
            }
        }
        for(;;) {
            int32_t node=*pos++;
            if(node<kMinLinearMatch) {
                UStringTrieResult result=branchNext(pos, node, inByte);
                if(result==USTRINGTRIE_NO_MATCH) {
                    return USTRINGTRIE_NO_MATCH;
                }
                // Fetch the next input byte, if there is one.
                if(sLength<0) {
                    if((inByte=*s++)==0) {
                        return result;
                    }
                } else {
                    if(sLength==0) {
                        return result;
                    }
                    inByte=*s++;
                    --sLength;
                }
                if(result==USTRINGTRIE_FINAL_VALUE) {
                    // No further matching bytes.
                    stop();
                    return USTRINGTRIE_NO_MATCH;
                }
                pos=pos_;  // branchNext() advanced pos and wrote it to pos_ .
            } else if(node<kMinValueLead) {
                // Match length+1 bytes.
                length=node-kMinLinearMatch;  // Actual match length minus 1.
                if(inByte!=*pos) {
                    stop();
                    return USTRINGTRIE_NO_MATCH;
                }
                ++pos;
                --length;
                break;
            } else if(node&kValueIsFinal) {
                // No further matching bytes.
                stop();
                return USTRINGTRIE_NO_MATCH;
            } else {
                // Skip intermediate value.
                pos=skipValue(pos, node);
                // The next node must not also be a value node.
                U_ASSERT(*pos<kMinValueLead);
            }
        }
    }
}

const uint8_t *
BytesTrie::findUniqueValueFromBranch(const uint8_t *pos, int32_t length,
                                     UBool haveUniqueValue, int32_t &uniqueValue) {
    while(length>kMaxBranchLinearSubNodeLength) {
        ++pos;  // ignore the comparison byte
        if(NULL==findUniqueValueFromBranch(jumpByDelta(pos), length>>1, haveUniqueValue, uniqueValue)) {
            return NULL;
        }
        length=length-(length>>1);
        pos=skipDelta(pos);
    }
    do {
        ++pos;  // ignore a comparison byte
        // handle its value
        int32_t node=*pos++;
        UBool isFinal=(UBool)(node&kValueIsFinal);
        int32_t value=readValue(pos, node>>1);
        pos=skipValue(pos, node);
        if(isFinal) {
            if(haveUniqueValue) {
                if(value!=uniqueValue) {
                    return NULL;
                }
            } else {
                uniqueValue=value;
                haveUniqueValue=TRUE;
            }
        } else {
            if(!findUniqueValue(pos+value, haveUniqueValue, uniqueValue)) {
                return NULL;
            }
            haveUniqueValue=TRUE;
        }
    } while(--length>1);
    return pos+1;  // ignore the last comparison byte
}

UBool
BytesTrie::findUniqueValue(const uint8_t *pos, UBool haveUniqueValue, int32_t &uniqueValue) {
    for(;;) {
        int32_t node=*pos++;
        if(node<kMinLinearMatch) {
            if(node==0) {
                node=*pos++;
            }
            pos=findUniqueValueFromBranch(pos, node+1, haveUniqueValue, uniqueValue);
            if(pos==NULL) {
                return FALSE;
            }
            haveUniqueValue=TRUE;
        } else if(node<kMinValueLead) {
            // linear-match node
            pos+=node-kMinLinearMatch+1;  // Ignore the match bytes.
        } else {
            UBool isFinal=(UBool)(node&kValueIsFinal);
            int32_t value=readValue(pos, node>>1);
            if(haveUniqueValue) {
                if(value!=uniqueValue) {
                    return FALSE;
                }
            } else {
                uniqueValue=value;
                haveUniqueValue=TRUE;
            }
            if(isFinal) {
                return TRUE;
            }
            pos=skipValue(pos, node);
        }
    }
}

int32_t
BytesTrie::getNextBytes(ByteSink &out) const {
    const uint8_t *pos=pos_;
    if(pos==NULL) {
        return 0;
    }
    if(remainingMatchLength_>=0) {
        append(out, *pos);  // Next byte of a pending linear-match node.
        return 1;
    }
    int32_t node=*pos++;
    if(node>=kMinValueLead) {
        if(node&kValueIsFinal) {
            return 0;
        } else {
            pos=skipValue(pos, node);
            node=*pos++;
            U_ASSERT(node<kMinValueLead);
        }
    }
    if(node<kMinLinearMatch) {
        if(node==0) {
            node=*pos++;
        }
        getNextBranchBytes(pos, ++node, out);
        return node;
    } else {
        // First byte of the linear-match node.
        append(out, *pos);
        return 1;
    }
}

void
BytesTrie::getNextBranchBytes(const uint8_t *pos, int32_t length, ByteSink &out) {
    while(length>kMaxBranchLinearSubNodeLength) {
        ++pos;  // ignore the comparison byte
        getNextBranchBytes(jumpByDelta(pos), length>>1, out);
        length=length-(length>>1);
        pos=skipDelta(pos);
    }
    do {
        append(out, *pos++);
        pos=skipValue(pos);
    } while(--length>1);
    append(out, *pos);
}

void
BytesTrie::append(ByteSink &out, int c) {
    char ch=(char)c;
    out.Append(&ch, 1);
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  bytestriebuilder.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010sep25
*   created by: Markus W. Scherer
*/



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  bytestriebuilder.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010sep25
*   created by: Markus W. Scherer
*/

/**
 * \file
 * \brief C++ API: Builder for icu::BytesTrie
 */




#if U_SHOW_CPLUSPLUS_API



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2012,2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  stringtriebuilder.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010dec24
*   created by: Markus W. Scherer
*/




#if U_SHOW_CPLUSPLUS_API



/**
 * \file
 * \brief C++ API: Builder API for trie builders
 */

// Forward declaration.
/// \cond
struct UHashtable;
typedef struct UHashtable UHashtable;
/// \endcond

/**
 * Build options for BytesTrieBuilder and CharsTrieBuilder.
 * @stable ICU 4.8
 */
enum UStringTrieBuildOption {
    /**
     * Builds a trie quickly.
     * @stable ICU 4.8
     */
    USTRINGTRIE_BUILD_FAST,
    /**
     * Builds a trie more slowly, attempting to generate
     * a shorter but equivalent serialization.
     * This build option also uses more memory.
     *
     * This option can be effective when many integer values are the same
     * and string/byte sequence suffixes can be shared.
     * Runtime speed is not expected to improve.
     * @stable ICU 4.8
     */
    USTRINGTRIE_BUILD_SMALL
};

U_NAMESPACE_BEGIN

/**
 * Base class for string trie builder classes.
 *
 * This class is not intended for public subclassing.
 * @stable ICU 4.8
 */
class U_COMMON_API StringTrieBuilder : public UObject {
public:
#ifndef U_HIDE_INTERNAL_API
    /** @internal */
    static int32_t hashNode(const void *node);
    /** @internal */
    static UBool equalNodes(const void *left, const void *right);
#endif  /* U_HIDE_INTERNAL_API */

protected:
    // Do not enclose the protected default constructor with #ifndef U_HIDE_INTERNAL_API
    // or else the compiler will create a public default constructor.
    /** @internal */
    StringTrieBuilder();
    /** @internal */
    virtual ~StringTrieBuilder();

#ifndef U_HIDE_INTERNAL_API
    /** @internal */
    void createCompactBuilder(int32_t sizeGuess, UErrorCode &errorCode);
    /** @internal */
    void deleteCompactBuilder();

    /** @internal */
    void build(UStringTrieBuildOption buildOption, int32_t elementsLength, UErrorCode &errorCode);

    /** @internal */
    int32_t writeNode(int32_t start, int32_t limit, int32_t unitIndex);
    /** @internal */
    int32_t writeBranchSubNode(int32_t start, int32_t limit, int32_t unitIndex, int32_t length);
#endif  /* U_HIDE_INTERNAL_API */

    class Node;

#ifndef U_HIDE_INTERNAL_API
    /** @internal */
    Node *makeNode(int32_t start, int32_t limit, int32_t unitIndex, UErrorCode &errorCode);
    /** @internal */
    Node *makeBranchSubNode(int32_t start, int32_t limit, int32_t unitIndex,
                            int32_t length, UErrorCode &errorCode);
#endif  /* U_HIDE_INTERNAL_API */

    /** @internal */
    virtual int32_t getElementStringLength(int32_t i) const = 0;
    /** @internal */
    virtual char16_t getElementUnit(int32_t i, int32_t unitIndex) const = 0;
    /** @internal */
    virtual int32_t getElementValue(int32_t i) const = 0;

    // Finds the first unit index after this one where
    // the first and last element have different units again.
    /** @internal */
    virtual int32_t getLimitOfLinearMatch(int32_t first, int32_t last, int32_t unitIndex) const = 0;

    // Number of different units at unitIndex.
    /** @internal */
    virtual int32_t countElementUnits(int32_t start, int32_t limit, int32_t unitIndex) const = 0;
    /** @internal */
    virtual int32_t skipElementsBySomeUnits(int32_t i, int32_t unitIndex, int32_t count) const = 0;
    /** @internal */
    virtual int32_t indexOfElementWithNextUnit(int32_t i, int32_t unitIndex, char16_t unit) const = 0;

    /** @internal */
    virtual UBool matchNodesCanHaveValues() const = 0;

    /** @internal */
    virtual int32_t getMaxBranchLinearSubNodeLength() const = 0;
    /** @internal */
    virtual int32_t getMinLinearMatch() const = 0;
    /** @internal */
    virtual int32_t getMaxLinearMatchLength() const = 0;

#ifndef U_HIDE_INTERNAL_API
    // max(BytesTrie::kMaxBranchLinearSubNodeLength, UCharsTrie::kMaxBranchLinearSubNodeLength).
    /** @internal */
    static const int32_t kMaxBranchLinearSubNodeLength=5;

    // Maximum number of nested split-branch levels for a branch on all 2^16 possible char16_t units.
    // log2(2^16/kMaxBranchLinearSubNodeLength) rounded up.
    /** @internal */
    static const int32_t kMaxSplitBranchLevels=14;

    /**
     * Makes sure that there is only one unique node registered that is
     * equivalent to newNode.
     * @param newNode Input node. The builder takes ownership.
     * @param errorCode ICU in/out UErrorCode.
                        Set to U_MEMORY_ALLOCATION_ERROR if it was success but newNode==NULL.
     * @return newNode if it is the first of its kind, or
     *         an equivalent node if newNode is a duplicate.
     * @internal
     */
    Node *registerNode(Node *newNode, UErrorCode &errorCode);
    /**
     * Makes sure that there is only one unique FinalValueNode registered
     * with this value.
     * Avoids creating a node if the value is a duplicate.
     * @param value A final value.
     * @param errorCode ICU in/out UErrorCode.
                        Set to U_MEMORY_ALLOCATION_ERROR if it was success but newNode==NULL.
     * @return A FinalValueNode with the given value.
     * @internal
     */
    Node *registerFinalValue(int32_t value, UErrorCode &errorCode);
#endif  /* U_HIDE_INTERNAL_API */

    /*
     * C++ note:
     * registerNode() and registerFinalValue() take ownership of their input nodes,
     * and only return owned nodes.
     * If they see a failure UErrorCode, they will delete the input node.
     * If they get a NULL pointer, they will record a U_MEMORY_ALLOCATION_ERROR.
     * If there is a failure, they return NULL.
     *
     * NULL Node pointers can be safely passed into other Nodes because
     * they call the static Node::hashCode() which checks for a NULL pointer first.
     *
     * Therefore, as long as builder functions register a new node,
     * they need to check for failures only before explicitly dereferencing
     * a Node pointer, or before setting a new UErrorCode.
     */

    // Hash set of nodes, maps from nodes to integer 1.
    /** @internal */
    UHashtable *nodes;

    // Do not conditionalize the following with #ifndef U_HIDE_INTERNAL_API,
    // it is needed for layout of other objects.
    /**
     * @internal
     * \cond
     */
    class Node : public UObject {
    public:
        Node(int32_t initialHash) : hash(initialHash), offset(0) {}
        inline int32_t hashCode() const { return hash; }
        // Handles node==NULL.
        static inline int32_t hashCode(const Node *node) { return node==NULL ? 0 : node->hashCode(); }
        // Base class operator==() compares the actual class types.
        virtual UBool operator==(const Node &other) const;
        inline UBool operator!=(const Node &other) const { return !operator==(other); }
        /**
         * Traverses the Node graph and numbers branch edges, with rightmost edges first.
         * This is to avoid writing a duplicate node twice.
         *
         * Branch nodes in this trie data structure are not symmetric.
         * Most branch edges "jump" to other nodes but the rightmost branch edges
         * just continue without a jump.
         * Therefore, write() must write the rightmost branch edge last
         * (trie units are written backwards), and must write it at that point even if
         * it is a duplicate of a node previously written elsewhere.
         *
         * This function visits and marks right branch edges first.
         * Edges are numbered with increasingly negative values because we share the
         * offset field which gets positive values when nodes are written.
         * A branch edge also remembers the first number for any of its edges.
         *
         * When a further-left branch edge has a number in the range of the rightmost
         * edge's numbers, then it will be written as part of the required right edge
         * and we can avoid writing it first.
         *
         * After root.markRightEdgesFirst(-1) the offsets of all nodes are negative
         * edge numbers.
         *
         * @param edgeNumber The first edge number for this node and its sub-nodes.
         * @return An edge number that is at least the maximum-negative
         *         of the input edge number and the numbers of this node and all of its sub-nodes.
         */
        virtual int32_t markRightEdgesFirst(int32_t edgeNumber);
        // write() must set the offset to a positive value.
        virtual void write(StringTrieBuilder &builder) = 0;
        // See markRightEdgesFirst.
        inline void writeUnlessInsideRightEdge(int32_t firstRight, int32_t lastRight,
                                               StringTrieBuilder &builder) {
            // Note: Edge numbers are negative, lastRight<=firstRight.
            // If offset>0 then this node and its sub-nodes have been written already
            // and we need not write them again.
            // If this node is part of the unwritten right branch edge,
            // then we wait until that is written.
            if(offset<0 && (offset<lastRight || firstRight<offset)) {
                write(builder);
            }
        }
        inline int32_t getOffset() const { return offset; }
    protected:
        int32_t hash;
        int32_t offset;
    };

#ifndef U_HIDE_INTERNAL_API
    // This class should not be overridden because
    // registerFinalValue() compares a stack-allocated FinalValueNode
    // (stack-allocated so that we don't unnecessarily create lots of duplicate nodes)
    // with the input node, and the
    // !Node::operator==(other) used inside FinalValueNode::operator==(other)
    // will be false if the typeid's are different.
    /** @internal */
    class FinalValueNode : public Node {
    public:
        FinalValueNode(int32_t v) : Node(0x111111u*37u+v), value(v) {}
        virtual UBool operator==(const Node &other) const;
        virtual void write(StringTrieBuilder &builder);
    protected:
        int32_t value;
    };
#endif  /* U_HIDE_INTERNAL_API */

    // Do not conditionalize the following with #ifndef U_HIDE_INTERNAL_API,
    // it is needed for layout of other objects.
    /**
     * @internal 
     */
    class ValueNode : public Node {
    public:
        ValueNode(int32_t initialHash) : Node(initialHash), hasValue(FALSE), value(0) {}
        virtual UBool operator==(const Node &other) const;
        void setValue(int32_t v) {
            hasValue=TRUE;
            value=v;
            hash=hash*37u+v;
        }
    protected:
        UBool hasValue;
        int32_t value;
    };

#ifndef U_HIDE_INTERNAL_API
    /** 
     * @internal 
     */
    class IntermediateValueNode : public ValueNode {
    public:
        IntermediateValueNode(int32_t v, Node *nextNode)
                : ValueNode(0x222222u*37u+hashCode(nextNode)), next(nextNode) { setValue(v); }
        virtual UBool operator==(const Node &other) const;
        virtual int32_t markRightEdgesFirst(int32_t edgeNumber);
        virtual void write(StringTrieBuilder &builder);
    protected:
        Node *next;
    };
#endif  /* U_HIDE_INTERNAL_API */

    // Do not conditionalize the following with #ifndef U_HIDE_INTERNAL_API,
    // it is needed for layout of other objects.
    /**
     * @internal 
     */
    class LinearMatchNode : public ValueNode {
    public:
        LinearMatchNode(int32_t len, Node *nextNode)
                : ValueNode((0x333333u*37u+len)*37u+hashCode(nextNode)),
                  length(len), next(nextNode) {}
        virtual UBool operator==(const Node &other) const;
        virtual int32_t markRightEdgesFirst(int32_t edgeNumber);
    protected:
        int32_t length;
        Node *next;
    };

#ifndef U_HIDE_INTERNAL_API
    /**
     * @internal 
     */
    class BranchNode : public Node {
    public:
        BranchNode(int32_t initialHash) : Node(initialHash) {}
    protected:
        int32_t firstEdgeNumber;
    };

    /**
     * @internal 
     */
    class ListBranchNode : public BranchNode {
    public:
        ListBranchNode() : BranchNode(0x444444), length(0) {}
        virtual UBool operator==(const Node &other) const;
        virtual int32_t markRightEdgesFirst(int32_t edgeNumber);
        virtual void write(StringTrieBuilder &builder);
        // Adds a unit with a final value.
        void add(int32_t c, int32_t value) {
            units[length]=(char16_t)c;
            equal[length]=NULL;
            values[length]=value;
            ++length;
            hash=(hash*37u+c)*37u+value;
        }
        // Adds a unit which leads to another match node.
        void add(int32_t c, Node *node) {
            units[length]=(char16_t)c;
            equal[length]=node;
            values[length]=0;
            ++length;
            hash=(hash*37u+c)*37u+hashCode(node);
        }
    protected:
        Node *equal[kMaxBranchLinearSubNodeLength];  // NULL means "has final value".
        int32_t length;
        int32_t values[kMaxBranchLinearSubNodeLength];
        char16_t units[kMaxBranchLinearSubNodeLength];
    };

    /**
     * @internal 
     */
    class SplitBranchNode : public BranchNode {
    public:
        SplitBranchNode(char16_t middleUnit, Node *lessThanNode, Node *greaterOrEqualNode)
                : BranchNode(((0x555555u*37u+middleUnit)*37u+
                              hashCode(lessThanNode))*37u+hashCode(greaterOrEqualNode)),
                  unit(middleUnit), lessThan(lessThanNode), greaterOrEqual(greaterOrEqualNode) {}
        virtual UBool operator==(const Node &other) const;
        virtual int32_t markRightEdgesFirst(int32_t edgeNumber);
        virtual void write(StringTrieBuilder &builder);
    protected:
        char16_t unit;
        Node *lessThan;
        Node *greaterOrEqual;
    };

    // Branch head node, for writing the actual node lead unit.
    /** @internal */
    class BranchHeadNode : public ValueNode {
    public:
        BranchHeadNode(int32_t len, Node *subNode)
                : ValueNode((0x666666u*37u+len)*37u+hashCode(subNode)),
                  length(len), next(subNode) {}
        virtual UBool operator==(const Node &other) const;
        virtual int32_t markRightEdgesFirst(int32_t edgeNumber);
        virtual void write(StringTrieBuilder &builder);
    protected:
        int32_t length;
        Node *next;  // A branch sub-node.
    };

#endif  /* U_HIDE_INTERNAL_API */
    /// \endcond

    /** @internal */
    virtual Node *createLinearMatchNode(int32_t i, int32_t unitIndex, int32_t length,
                                        Node *nextNode) const = 0;

    /** @internal */
    virtual int32_t write(int32_t unit) = 0;
    /** @internal */
    virtual int32_t writeElementUnits(int32_t i, int32_t unitIndex, int32_t length) = 0;
    /** @internal */
    virtual int32_t writeValueAndFinal(int32_t i, UBool isFinal) = 0;
    /** @internal */
    virtual int32_t writeValueAndType(UBool hasValue, int32_t value, int32_t node) = 0;
    /** @internal */
    virtual int32_t writeDeltaTo(int32_t jumpTarget) = 0;
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */



U_NAMESPACE_BEGIN

class BytesTrieElement;
class CharString;
/**
 * Builder class for BytesTrie.
 *
 * This class is not intended for public subclassing.
 * @stable ICU 4.8
 */
class U_COMMON_API BytesTrieBuilder : public StringTrieBuilder {
public:
    /**
     * Constructs an empty builder.
     * @param errorCode Standard ICU error code.
     * @stable ICU 4.8
     */
    BytesTrieBuilder(UErrorCode &errorCode);

    /**
     * Destructor.
     * @stable ICU 4.8
     */
    virtual ~BytesTrieBuilder();

    /**
     * Adds a (byte sequence, value) pair.
     * The byte sequence must be unique.
     * The bytes will be copied; the builder does not keep
     * a reference to the input StringPiece or its data().
     * @param s The input byte sequence.
     * @param value The value associated with this byte sequence.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return *this
     * @stable ICU 4.8
     */
    BytesTrieBuilder &add(StringPiece s, int32_t value, UErrorCode &errorCode);

    /**
     * Builds a BytesTrie for the add()ed data.
     * Once built, no further data can be add()ed until clear() is called.
     *
     * A BytesTrie cannot be empty. At least one (byte sequence, value) pair
     * must have been add()ed.
     *
     * This method passes ownership of the builder's internal result array to the new trie object.
     * Another call to any build() variant will re-serialize the trie.
     * After clear() has been called, a new array will be used as well.
     * @param buildOption Build option, see UStringTrieBuildOption.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return A new BytesTrie for the add()ed data.
     * @stable ICU 4.8
     */
    BytesTrie *build(UStringTrieBuildOption buildOption, UErrorCode &errorCode);

    /**
     * Builds a BytesTrie for the add()ed data and byte-serializes it.
     * Once built, no further data can be add()ed until clear() is called.
     *
     * A BytesTrie cannot be empty. At least one (byte sequence, value) pair
     * must have been add()ed.
     *
     * Multiple calls to buildStringPiece() return StringPieces referring to the
     * builder's same byte array, without rebuilding.
     * If buildStringPiece() is called after build(), the trie will be
     * re-serialized into a new array.
     * If build() is called after buildStringPiece(), the trie object will become
     * the owner of the previously returned array.
     * After clear() has been called, a new array will be used as well.
     * @param buildOption Build option, see UStringTrieBuildOption.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return A StringPiece which refers to the byte-serialized BytesTrie for the add()ed data.
     * @stable ICU 4.8
     */
    StringPiece buildStringPiece(UStringTrieBuildOption buildOption, UErrorCode &errorCode);

    /**
     * Removes all (byte sequence, value) pairs.
     * New data can then be add()ed and a new trie can be built.
     * @return *this
     * @stable ICU 4.8
     */
    BytesTrieBuilder &clear();

private:
    BytesTrieBuilder(const BytesTrieBuilder &other);  // no copy constructor
    BytesTrieBuilder &operator=(const BytesTrieBuilder &other);  // no assignment operator

    void buildBytes(UStringTrieBuildOption buildOption, UErrorCode &errorCode);

    virtual int32_t getElementStringLength(int32_t i) const;
    virtual char16_t getElementUnit(int32_t i, int32_t byteIndex) const;
    virtual int32_t getElementValue(int32_t i) const;

    virtual int32_t getLimitOfLinearMatch(int32_t first, int32_t last, int32_t byteIndex) const;

    virtual int32_t countElementUnits(int32_t start, int32_t limit, int32_t byteIndex) const;
    virtual int32_t skipElementsBySomeUnits(int32_t i, int32_t byteIndex, int32_t count) const;
    virtual int32_t indexOfElementWithNextUnit(int32_t i, int32_t byteIndex, char16_t byte) const;

    virtual UBool matchNodesCanHaveValues() const { return FALSE; }

    virtual int32_t getMaxBranchLinearSubNodeLength() const { return BytesTrie::kMaxBranchLinearSubNodeLength; }
    virtual int32_t getMinLinearMatch() const { return BytesTrie::kMinLinearMatch; }
    virtual int32_t getMaxLinearMatchLength() const { return BytesTrie::kMaxLinearMatchLength; }

    /**
     * @internal (private)
     */
    class BTLinearMatchNode : public LinearMatchNode {
    public:
        BTLinearMatchNode(const char *units, int32_t len, Node *nextNode);
        virtual UBool operator==(const Node &other) const;
        virtual void write(StringTrieBuilder &builder);
    private:
        const char *s;
    };
    
    virtual Node *createLinearMatchNode(int32_t i, int32_t byteIndex, int32_t length,
                                        Node *nextNode) const;

    UBool ensureCapacity(int32_t length);
    virtual int32_t write(int32_t byte);
    int32_t write(const char *b, int32_t length);
    virtual int32_t writeElementUnits(int32_t i, int32_t byteIndex, int32_t length);
    virtual int32_t writeValueAndFinal(int32_t i, UBool isFinal);
    virtual int32_t writeValueAndType(UBool hasValue, int32_t value, int32_t node);
    virtual int32_t writeDeltaTo(int32_t jumpTarget);

    CharString *strings;  // Pointer not object so we need not #include internal charstr.h.
    BytesTrieElement *elements;
    int32_t elementsCapacity;
    int32_t elementsLength;

    // Byte serialization of the trie.
    // Grows from the back: bytesLength measures from the end of the buffer!
    char *bytes;
    int32_t bytesCapacity;
    int32_t bytesLength;
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 1997-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*   Date        Name        Description
*   03/22/00    aliu        Adapted from original C++ ICU Hashtable.
*   07/06/01    aliu        Modified to support int32_t keys on
*                           platforms with sizeof(void*) < 32.
******************************************************************************
*/




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 1997-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  uelement.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011jul04
*   created by: Markus W. Scherer
*
*   Common definitions for UHashTable and UVector.
*   UHashTok moved here from uhash.h and renamed UElement.
*   This allows users of UVector to avoid the confusing #include of uhash.h.
*   uhash.h aliases UElement to UHashTok,
*   so that we need not change all of its code and its users.
*/




U_CDECL_BEGIN

/**
 * A UVector element, or a key or value within a UHashtable.
 * It may be either a 32-bit integral value or an opaque void* pointer.
 * The void* pointer may be smaller than 32 bits (e.g. 24 bits)
 * or may be larger (e.g. 64 bits).
 *
 * Because a UElement is the size of a native pointer or a 32-bit
 * integer, we pass it around by value.
 */
union UElement {
    void*   pointer;
    int32_t integer;
};
typedef union UElement UElement;

/**
 * An element-equality (boolean) comparison function.
 * @param e1 An element (object or integer)
 * @param e2 An element (object or integer)
 * @return TRUE if the two elements are equal.
 */
typedef UBool U_CALLCONV UElementsAreEqual(const UElement e1, const UElement e2);

/**
 * An element sorting (three-way) comparison function.
 * @param e1 An element (object or integer)
 * @param e2 An element (object or integer)
 * @return 0 if the two elements are equal, -1 if e1 is < e2, or +1 if e1 is > e2.
 */
typedef int8_t U_CALLCONV UElementComparator(UElement e1, UElement e2);

/**
 * An element assignment function.  It may copy an integer, copy
 * a pointer, or clone a pointer, as appropriate.
 * @param dst The element to be assigned to
 * @param src The element to assign from
 */
typedef void U_CALLCONV UElementAssigner(UElement *dst, UElement *src);

U_CDECL_END

/**
 * Comparator function for UnicodeString* keys. Implements UElementsAreEqual.
 * @param key1 The string for comparison
 * @param key2 The string for comparison
 * @return true if key1 and key2 are equal, return false otherwise.
 */
U_CAPI UBool U_EXPORT2 
uhash_compareUnicodeString(const UElement key1, const UElement key2);

/**
 * Comparator function for UnicodeString* keys (case insensitive).
 * Make sure to use together with uhash_hashCaselessUnicodeString.
 * Implements UElementsAreEqual.
 * @param key1 The string for comparison
 * @param key2 The string for comparison
 * @return true if key1 and key2 are equal, return false otherwise.
 */
U_CAPI UBool U_EXPORT2 
uhash_compareCaselessUnicodeString(const UElement key1, const UElement key2);




/**
 * UHashtable stores key-value pairs and does moderately fast lookup
 * based on keys.  It provides a good tradeoff between access time and
 * storage space.  As elements are added to it, it grows to accomodate
 * them.  By default, the table never shrinks, even if all elements
 * are removed from it.
 *
 * Keys and values are stored as void* pointers.  These void* pointers
 * may be actual pointers to strings, objects, or any other structure
 * in memory, or they may simply be integral values cast to void*.
 * UHashtable doesn't care and manipulates them via user-supplied
 * functions.  These functions hash keys, compare keys, delete keys,
 * and delete values.  Some function pointers are optional (may be
 * NULL); others must be supplied.  Several prebuilt functions exist
 * to handle common key types.
 *
 * UHashtable ownership of keys and values is flexible, and controlled
 * by whether or not the key deleter and value deleter functions are
 * set.  If a void* key is actually a pointer to a deletable object,
 * then UHashtable can be made to delete that object by setting the
 * key deleter function pointer to a non-NULL value.  If this is done,
 * then keys passed to uhash_put() are owned by the hashtable and will
 * be deleted by it at some point, either as keys are replaced, or
 * when uhash_close() is finally called.  The same is true of values
 * and the value deleter function pointer.  Keys passed to methods
 * other than uhash_put() are never owned by the hashtable.
 *
 * NULL values are not allowed.  uhash_get() returns NULL to indicate
 * a key that is not in the table, and having a NULL value in the
 * table would generate an ambiguous result.  If a key and a NULL
 * value is passed to uhash_put(), this has the effect of doing a
 * uhash_remove() on that key.  This keeps uhash_get(), uhash_count(),
 * and uhash_nextElement() consistent with one another.
 *
 * To see everything in a hashtable, use uhash_nextElement() to
 * iterate through its contents.  Each call to this function returns a
 * UHashElement pointer.  A hash element contains a key, value, and
 * hashcode.  During iteration an element may be deleted by calling
 * uhash_removeElement(); iteration may safely continue thereafter.
 * The uhash_remove() function may also be safely called in
 * mid-iteration.  If uhash_put() is called during iteration,
 * the iteration is still guaranteed to terminate reasonably, but
 * there is no guarantee that every element will be returned or that
 * some won't be returned more than once.
 *
 * Under no circumstances should the UHashElement returned by
 * uhash_nextElement be modified directly.
 *
 * By default, the hashtable grows when necessary, but never shrinks,
 * even if all items are removed.  For most applications this is
 * optimal.  However, in a highly dynamic usage where memory is at a
 * premium, the table can be set to both grow and shrink by calling
 * uhash_setResizePolicy() with the policy U_GROW_AND_SHRINK.  In a
 * situation where memory is critical and the client wants a table
 * that does not grow at all, the constant U_FIXED can be used.
 */

/********************************************************************
 * Data Structures
 ********************************************************************/

U_CDECL_BEGIN

/**
 * A key or value within a UHashtable.
 * The hashing and comparison functions take a pointer to a
 * UHashTok, but the deleter receives the void* pointer within it.
 */
typedef UElement UHashTok;

/**
 * This is a single hash element.
 */
struct UHashElement {
    /* Reorder these elements to pack nicely if necessary */
    int32_t  hashcode;
    UHashTok value;
    UHashTok key;
};
typedef struct UHashElement UHashElement;

/**
 * A hashing function.
 * @param key A key stored in a hashtable
 * @return A NON-NEGATIVE hash code for parm.
 */
typedef int32_t U_CALLCONV UHashFunction(const UHashTok key);

/**
 * A key equality (boolean) comparison function.
 */
typedef UElementsAreEqual UKeyComparator;

/**
 * A value equality (boolean) comparison function.
 */
typedef UElementsAreEqual UValueComparator;

/* see cmemory.h for UObjectDeleter and uprv_deleteUObject() */

/**
 * This specifies whether or not, and how, the hastable resizes itself.
 * See uhash_setResizePolicy().
 */
enum UHashResizePolicy {
    U_GROW,            /* Grow on demand, do not shrink */
    U_GROW_AND_SHRINK, /* Grow and shrink on demand */
    U_FIXED            /* Never change size */
};

/**
 * The UHashtable struct.  Clients should treat this as an opaque data
 * type and manipulate it only through the uhash_... API.
 */
struct UHashtable {

    /* Main key-value pair storage array */

    UHashElement *elements;

    /* Function pointers */

    UHashFunction *keyHasher;      /* Computes hash from key.
                                   * Never null. */
    UKeyComparator *keyComparator; /* Compares keys for equality.
                                   * Never null. */
    UValueComparator *valueComparator; /* Compares the values for equality */

    UObjectDeleter *keyDeleter;    /* Deletes keys when required.
                                   * If NULL won't do anything */
    UObjectDeleter *valueDeleter;  /* Deletes values when required.
                                   * If NULL won't do anything */

    /* Size parameters */

    int32_t     count;      /* The number of key-value pairs in this table.
                             * 0 <= count <= length.  In practice we
                             * never let count == length (see code). */
    int32_t     length;     /* The physical size of the arrays hashes, keys
                             * and values.  Must be prime. */

    /* Rehashing thresholds */

    int32_t     highWaterMark;  /* If count > highWaterMark, rehash */
    int32_t     lowWaterMark;   /* If count < lowWaterMark, rehash */
    float       highWaterRatio; /* 0..1; high water as a fraction of length */
    float       lowWaterRatio;  /* 0..1; low water as a fraction of length */

    int8_t      primeIndex;     /* Index into our prime table for length.
                                 * length == PRIMES[primeIndex] */
    UBool       allocated; /* Was this UHashtable allocated? */
};
typedef struct UHashtable UHashtable;

U_CDECL_END

/********************************************************************
 * API
 ********************************************************************/

/**
 * Initialize a new UHashtable.
 * @param keyHash A pointer to the key hashing function.  Must not be
 * NULL.
 * @param keyComp A pointer to the function that compares keys.  Must
 * not be NULL.
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return A pointer to a UHashtable, or 0 if an error occurred.
 * @see uhash_openSize
 */
U_CAPI UHashtable* U_EXPORT2
uhash_open(UHashFunction *keyHash,
           UKeyComparator *keyComp,
           UValueComparator *valueComp,
           UErrorCode *status);

/**
 * Initialize a new UHashtable with a given initial size.
 * @param keyHash A pointer to the key hashing function.  Must not be
 * NULL.
 * @param keyComp A pointer to the function that compares keys.  Must
 * not be NULL.
 * @param size The initial capacity of this hash table.
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return A pointer to a UHashtable, or 0 if an error occurred.
 * @see uhash_open
 */
U_CAPI UHashtable* U_EXPORT2
uhash_openSize(UHashFunction *keyHash,
               UKeyComparator *keyComp,
               UValueComparator *valueComp,
               int32_t size,
               UErrorCode *status);

/**
 * Initialize an existing UHashtable.
 * @param keyHash A pointer to the key hashing function.  Must not be
 * NULL.
 * @param keyComp A pointer to the function that compares keys.  Must
 * not be NULL.
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return A pointer to a UHashtable, or 0 if an error occurred.
 * @see uhash_openSize
 */
U_CAPI UHashtable* U_EXPORT2
uhash_init(UHashtable *hash,
           UHashFunction *keyHash,
           UKeyComparator *keyComp,
           UValueComparator *valueComp,
           UErrorCode *status);

/**
 * Initialize an existing UHashtable.
 * @param keyHash A pointer to the key hashing function.  Must not be
 * NULL.
 * @param keyComp A pointer to the function that compares keys.  Must
 * not be NULL.
 * @param size The initial capacity of this hash table.
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return A pointer to a UHashtable, or 0 if an error occurred.
 * @see uhash_openSize
 */
U_CAPI UHashtable* U_EXPORT2
uhash_initSize(UHashtable *hash,
               UHashFunction *keyHash,
               UKeyComparator *keyComp,
               UValueComparator *valueComp,
               int32_t size,
               UErrorCode *status);

/**
 * Close a UHashtable, releasing the memory used.
 * @param hash The UHashtable to close. If hash is NULL no operation is performed.
 */
U_CAPI void U_EXPORT2
uhash_close(UHashtable *hash);



/**
 * Set the function used to hash keys.
 * @param hash The UHashtable to set
 * @param fn the function to be used hash keys; must not be NULL
 * @return the previous key hasher; non-NULL
 */
U_CAPI UHashFunction *U_EXPORT2
uhash_setKeyHasher(UHashtable *hash, UHashFunction *fn);

/**
 * Set the function used to compare keys.  The default comparison is a
 * void* pointer comparison.
 * @param hash The UHashtable to set
 * @param fn the function to be used compare keys; must not be NULL
 * @return the previous key comparator; non-NULL
 */
U_CAPI UKeyComparator *U_EXPORT2
uhash_setKeyComparator(UHashtable *hash, UKeyComparator *fn);

/**
 * Set the function used to compare values.  The default comparison is a
 * void* pointer comparison.
 * @param hash The UHashtable to set
 * @param fn the function to be used compare keys; must not be NULL
 * @return the previous key comparator; non-NULL
 */
U_CAPI UValueComparator *U_EXPORT2
uhash_setValueComparator(UHashtable *hash, UValueComparator *fn);

/**
 * Set the function used to delete keys.  If this function pointer is
 * NULL, this hashtable does not delete keys.  If it is non-NULL, this
 * hashtable does delete keys.  This function should be set once
 * before any elements are added to the hashtable and should not be
 * changed thereafter.
 * @param hash The UHashtable to set
 * @param fn the function to be used delete keys, or NULL
 * @return the previous key deleter; may be NULL
 */
U_CAPI UObjectDeleter *U_EXPORT2
uhash_setKeyDeleter(UHashtable *hash, UObjectDeleter *fn);

/**
 * Set the function used to delete values.  If this function pointer
 * is NULL, this hashtable does not delete values.  If it is non-NULL,
 * this hashtable does delete values.  This function should be set
 * once before any elements are added to the hashtable and should not
 * be changed thereafter.
 * @param hash The UHashtable to set
 * @param fn the function to be used delete values, or NULL
 * @return the previous value deleter; may be NULL
 */
U_CAPI UObjectDeleter *U_EXPORT2
uhash_setValueDeleter(UHashtable *hash, UObjectDeleter *fn);

/**
 * Specify whether or not, and how, the hastable resizes itself.
 * By default, tables grow but do not shrink (policy U_GROW).
 * See enum UHashResizePolicy.
 * @param hash The UHashtable to set
 * @param policy The way the hashtable resizes itself, {U_GROW, U_GROW_AND_SHRINK, U_FIXED}
 */
U_CAPI void U_EXPORT2
uhash_setResizePolicy(UHashtable *hash, enum UHashResizePolicy policy);

/**
 * Get the number of key-value pairs stored in a UHashtable.
 * @param hash The UHashtable to query.
 * @return The number of key-value pairs stored in hash.
 */
U_CAPI int32_t U_EXPORT2
uhash_count(const UHashtable *hash);

/**
 * Put a (key=pointer, value=pointer) item in a UHashtable.  If the
 * keyDeleter is non-NULL, then the hashtable owns 'key' after this
 * call.  If the valueDeleter is non-NULL, then the hashtable owns
 * 'value' after this call.  Storing a NULL value is the same as
 * calling uhash_remove().
 * @param hash The target UHashtable.
 * @param key The key to store.
 * @param value The value to store, may be NULL (see above).
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return The previous value, or NULL if none.
 * @see uhash_get
 */
U_CAPI void* U_EXPORT2
uhash_put(UHashtable *hash,
          void *key,
          void *value,
          UErrorCode *status);

/**
 * Put a (key=integer, value=pointer) item in a UHashtable.
 * keyDeleter must be NULL.  If the valueDeleter is non-NULL, then the
 * hashtable owns 'value' after this call.  Storing a NULL value is
 * the same as calling uhash_remove().
 * @param hash The target UHashtable.
 * @param key The integer key to store.
 * @param value The value to store, may be NULL (see above).
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return The previous value, or NULL if none.
 * @see uhash_get
 */
U_CAPI void* U_EXPORT2
uhash_iput(UHashtable *hash,
           int32_t key,
           void* value,
           UErrorCode *status);

/**
 * Put a (key=pointer, value=integer) item in a UHashtable.  If the
 * keyDeleter is non-NULL, then the hashtable owns 'key' after this
 * call.  valueDeleter must be NULL.  Storing a 0 value is the same as
 * calling uhash_remove().
 * @param hash The target UHashtable.
 * @param key The key to store.
 * @param value The integer value to store.
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return The previous value, or 0 if none.
 * @see uhash_get
 */
U_CAPI int32_t U_EXPORT2
uhash_puti(UHashtable *hash,
           void* key,
           int32_t value,
           UErrorCode *status);

/**
 * Put a (key=integer, value=integer) item in a UHashtable.  If the
 * keyDeleter is non-NULL, then the hashtable owns 'key' after this
 * call.  valueDeleter must be NULL.  Storing a 0 value is the same as
 * calling uhash_remove().
 * @param hash The target UHashtable.
 * @param key The key to store.
 * @param value The integer value to store.
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return The previous value, or 0 if none.
 * @see uhash_get
 */
U_CAPI int32_t U_EXPORT2
uhash_iputi(UHashtable *hash,
           int32_t key,
           int32_t value,
           UErrorCode *status);

/**
 * Retrieve a pointer value from a UHashtable using a pointer key,
 * as previously stored by uhash_put().
 * @param hash The target UHashtable.
 * @param key A pointer key stored in a hashtable
 * @return The requested item, or NULL if not found.
 */
U_CAPI void* U_EXPORT2
uhash_get(const UHashtable *hash,
          const void *key);

/**
 * Retrieve a pointer value from a UHashtable using a integer key,
 * as previously stored by uhash_iput().
 * @param hash The target UHashtable.
 * @param key An integer key stored in a hashtable
 * @return The requested item, or NULL if not found.
 */
U_CAPI void* U_EXPORT2
uhash_iget(const UHashtable *hash,
           int32_t key);

/**
 * Retrieve an integer value from a UHashtable using a pointer key,
 * as previously stored by uhash_puti().
 * @param hash The target UHashtable.
 * @param key A pointer key stored in a hashtable
 * @return The requested item, or 0 if not found.
 */
U_CAPI int32_t U_EXPORT2
uhash_geti(const UHashtable *hash,
           const void* key);
/**
 * Retrieve an integer value from a UHashtable using an integer key,
 * as previously stored by uhash_iputi().
 * @param hash The target UHashtable.
 * @param key An integer key stored in a hashtable
 * @return The requested item, or 0 if not found.
 */
U_CAPI int32_t U_EXPORT2
uhash_igeti(const UHashtable *hash,
           int32_t key);

/**
 * Remove an item from a UHashtable stored by uhash_put().
 * @param hash The target UHashtable.
 * @param key A key stored in a hashtable
 * @return The item removed, or NULL if not found.
 */
U_CAPI void* U_EXPORT2
uhash_remove(UHashtable *hash,
             const void *key);

/**
 * Remove an item from a UHashtable stored by uhash_iput().
 * @param hash The target UHashtable.
 * @param key An integer key stored in a hashtable
 * @return The item removed, or NULL if not found.
 */
U_CAPI void* U_EXPORT2
uhash_iremove(UHashtable *hash,
              int32_t key);

/**
 * Remove an item from a UHashtable stored by uhash_puti().
 * @param hash The target UHashtable.
 * @param key An key stored in a hashtable
 * @return The item removed, or 0 if not found.
 */
U_CAPI int32_t U_EXPORT2
uhash_removei(UHashtable *hash,
              const void* key);

/**
 * Remove an item from a UHashtable stored by uhash_iputi().
 * @param hash The target UHashtable.
 * @param key An integer key stored in a hashtable
 * @return The item removed, or 0 if not found.
 */
U_CAPI int32_t U_EXPORT2
uhash_iremovei(UHashtable *hash,
               int32_t key);

/**
 * Remove all items from a UHashtable.
 * @param hash The target UHashtable.
 */
U_CAPI void U_EXPORT2
uhash_removeAll(UHashtable *hash);

/**
 * Locate an element of a UHashtable.  The caller must not modify the
 * returned object.  The primary use of this function is to obtain the
 * stored key when it may not be identical to the search key.  For
 * example, if the compare function is a case-insensitive string
 * compare, then the hash key may be desired in order to obtain the
 * canonical case corresponding to a search key.
 * @param hash The target UHashtable.
 * @param key A key stored in a hashtable
 * @return a hash element, or NULL if the key is not found.
 */
U_CAPI const UHashElement* U_EXPORT2
uhash_find(const UHashtable *hash, const void* key);

/**
 * \def UHASH_FIRST
 * Constant for use with uhash_nextElement
 * @see uhash_nextElement
 */
#define UHASH_FIRST (-1)

/**
 * Iterate through the elements of a UHashtable.  The caller must not
 * modify the returned object.  However, uhash_removeElement() may be
 * called during iteration to remove an element from the table.
 * Iteration may safely be resumed afterwards.  If uhash_put() is
 * called during iteration the iteration will then be out of sync and
 * should be restarted.
 * @param hash The target UHashtable.
 * @param pos This should be set to UHASH_FIRST initially, and left untouched
 * thereafter.
 * @return a hash element, or NULL if no further key-value pairs
 * exist in the table.
 */
U_CAPI const UHashElement* U_EXPORT2
uhash_nextElement(const UHashtable *hash,
                  int32_t *pos);

/**
 * Remove an element, returned by uhash_nextElement(), from the table.
 * Iteration may be safely continued afterwards.
 * @param hash The hashtable
 * @param e The element, returned by uhash_nextElement(), to remove.
 * Must not be NULL.  Must not be an empty or deleted element (as long
 * as this was returned by uhash_nextElement() it will not be empty or
 * deleted).  Note: Although this parameter is const, it will be
 * modified.
 * @return the value that was removed.
 */
U_CAPI void* U_EXPORT2
uhash_removeElement(UHashtable *hash, const UHashElement* e);

/********************************************************************
 * UHashTok convenience
 ********************************************************************/

/**
 * Return a UHashTok for an integer.
 * @param i The given integer
 * @return a UHashTok for an integer.
 */
/*U_CAPI UHashTok U_EXPORT2
uhash_toki(int32_t i);*/

/**
 * Return a UHashTok for a pointer.
 * @param p The given pointer
 * @return a UHashTok for a pointer.
 */
/*U_CAPI UHashTok U_EXPORT2
uhash_tokp(void* p);*/

/********************************************************************
 * UChar* and char* Support Functions
 ********************************************************************/

/**
 * Generate a hash code for a null-terminated UChar* string.  If the
 * string is not null-terminated do not use this function.  Use
 * together with uhash_compareUChars.
 * @param key The string (const UChar*) to hash.
 * @return A hash code for the key.
 */
U_CAPI int32_t U_EXPORT2
uhash_hashUChars(const UHashTok key);

/**
 * Generate a hash code for a null-terminated char* string.  If the
 * string is not null-terminated do not use this function.  Use
 * together with uhash_compareChars.
 * @param key The string (const char*) to hash.
 * @return A hash code for the key.
 */
U_CAPI int32_t U_EXPORT2
uhash_hashChars(const UHashTok key);

/**
 * Generate a case-insensitive hash code for a null-terminated char*
 * string.  If the string is not null-terminated do not use this
 * function.  Use together with uhash_compareIChars.
 * @param key The string (const char*) to hash.
 * @return A hash code for the key.
 */
U_CAPI int32_t U_EXPORT2
uhash_hashIChars(const UHashTok key);

/**
 * Comparator for null-terminated UChar* strings.  Use together with
 * uhash_hashUChars.
 * @param key1 The string for comparison
 * @param key2 The string for comparison
 * @return true if key1 and key2 are equal, return false otherwise.
 */
U_CAPI UBool U_EXPORT2
uhash_compareUChars(const UHashTok key1, const UHashTok key2);

/**
 * Comparator for null-terminated char* strings.  Use together with
 * uhash_hashChars.
 * @param key1 The string for comparison
 * @param key2 The string for comparison
 * @return true if key1 and key2 are equal, return false otherwise.
 */
U_CAPI UBool U_EXPORT2
uhash_compareChars(const UHashTok key1, const UHashTok key2);

/**
 * Case-insensitive comparator for null-terminated char* strings.  Use
 * together with uhash_hashIChars.
 * @param key1 The string for comparison
 * @param key2 The string for comparison
 * @return true if key1 and key2 are equal, return false otherwise.
 */
U_CAPI UBool U_EXPORT2
uhash_compareIChars(const UHashTok key1, const UHashTok key2);

/********************************************************************
 * UnicodeString Support Functions
 ********************************************************************/

/**
 * Hash function for UnicodeString* keys.
 * @param key The string (const char*) to hash.
 * @return A hash code for the key.
 */
U_CAPI int32_t U_EXPORT2
uhash_hashUnicodeString(const UElement key);

/**
 * Hash function for UnicodeString* keys (case insensitive).
 * Make sure to use together with uhash_compareCaselessUnicodeString.
 * @param key The string (const char*) to hash.
 * @return A hash code for the key.
 */
U_CAPI int32_t U_EXPORT2
uhash_hashCaselessUnicodeString(const UElement key);

/********************************************************************
 * int32_t Support Functions
 ********************************************************************/

/**
 * Hash function for 32-bit integer keys.
 * @param key The string (const char*) to hash.
 * @return A hash code for the key.
 */
U_CAPI int32_t U_EXPORT2
uhash_hashLong(const UHashTok key);

/**
 * Comparator function for 32-bit integer keys.
 * @param key1 The integer for comparison
 * @param Key2 The integer for comparison
 * @return true if key1 and key2 are equal, return false otherwise
 */
U_CAPI UBool U_EXPORT2
uhash_compareLong(const UHashTok key1, const UHashTok key2);

/********************************************************************
 * Other Support Functions
 ********************************************************************/

/**
 * Deleter for Hashtable objects.
 * @param obj The object to be deleted
 */
U_CAPI void U_EXPORT2
uhash_deleteHashtable(void *obj);

/* Use uprv_free() itself as a deleter for any key or value allocated using uprv_malloc. */

/**
 * Checks if the given hash tables are equal or not.
 * @param hash1
 * @param hash2
 * @return true if the hashtables are equal and false if not.
 */
U_CAPI UBool U_EXPORT2
uhash_equals(const UHashtable* hash1, const UHashtable* hash2);


#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUHashtablePointer
 * "Smart pointer" class, closes a UHashtable via uhash_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUHashtablePointer, UHashtable, uhash_close);

U_NAMESPACE_END

#endif


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2003-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uarrsort.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2003aug04
*   created by: Markus W. Scherer
*
*   Internal function for sorting arrays.
*/




U_CDECL_BEGIN
/**
 * Function type for comparing two items as part of sorting an array or similar.
 * Callback function for uprv_sortArray().
 *
 * @param context Application-specific pointer, passed through by uprv_sortArray().
 * @param left    Pointer to the "left" item.
 * @param right   Pointer to the "right" item.
 * @return 32-bit signed integer comparison result:
 *                <0 if left<right
 *               ==0 if left==right
 *                >0 if left>right
 *
 * @internal
 */
typedef int32_t U_CALLCONV
UComparator(const void *context, const void *left, const void *right);
U_CDECL_END

/**
 * Array sorting function.
 * Uses a UComparator for comparing array items to each other, and simple
 * memory copying to move items.
 *
 * @param array      The array to be sorted.
 * @param length     The number of items in the array.
 * @param itemSize   The size in bytes of each array item.
 * @param cmp        UComparator function used to compare two items each.
 * @param context    Application-specific pointer, passed through to the UComparator.
 * @param sortStable If true, a stable sorting algorithm must be used.
 * @param pErrorCode ICU in/out UErrorCode parameter.
 *
 * @internal
 */
U_CAPI void U_EXPORT2
uprv_sortArray(void *array, int32_t length, int32_t itemSize,
               UComparator *cmp, const void *context,
               UBool sortStable, UErrorCode *pErrorCode);

/**
 * Convenience UComparator implementation for uint16_t arrays.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
uprv_uint16Comparator(const void *context, const void *left, const void *right);

/**
 * Convenience UComparator implementation for int32_t arrays.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
uprv_int32Comparator(const void *context, const void *left, const void *right);

/**
 * Convenience UComparator implementation for uint32_t arrays.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
uprv_uint32Comparator(const void *context, const void *left, const void *right);

/**
 * Much like Java Collections.binarySearch(list, key, comparator).
 *
 * Except: Java documents "If the list contains multiple elements equal to
 * the specified object, there is no guarantee which one will be found."
 *
 * This version here will return the largest index of any equal item,
 * for use in stable sorting.
 *
 * @return the index>=0 where the item was found:
 *         the largest such index, if multiple, for stable sorting;
 *         or the index<0 for inserting the item at ~index in sorted order
 */
U_CAPI int32_t U_EXPORT2
uprv_stableBinarySearch(char *array, int32_t length, void *item, int32_t itemSize,
                        UComparator *cmp, const void *context);



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*  
**********************************************************************
*   Copyright (C) 1999-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*   file name:  ustr_imp.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001jan30
*   created by: Markus W. Scherer
*/





/**
 * Internal option for unorm_cmpEquivFold() for strncmp style.
 * If set, checks for both string length and terminating NUL.
 */
#define _STRNCMP_STYLE 0x1000

/**
 * Compare two strings in code point order or code unit order.
 * Works in strcmp style (both lengths -1),
 * strncmp style (lengths equal and >=0, flag TRUE),
 * and memcmp/UnicodeString style (at least one length >=0).
 */
U_CFUNC int32_t U_EXPORT2
uprv_strCompare(const UChar *s1, int32_t length1,
                const UChar *s2, int32_t length2,
                UBool strncmpStyle, UBool codePointOrder);

U_CAPI int32_t U_EXPORT2 
ustr_hashUCharsN(const UChar *str, int32_t length);

U_CAPI int32_t U_EXPORT2 
ustr_hashCharsN(const char *str, int32_t length);

U_CAPI int32_t U_EXPORT2
ustr_hashICharsN(const char *str, int32_t length);

/**
 * NUL-terminate a UChar * string if possible.
 * If length  < destCapacity then NUL-terminate.
 * If length == destCapacity then do not terminate but set U_STRING_NOT_TERMINATED_WARNING.
 * If length  > destCapacity then do not terminate but set U_BUFFER_OVERFLOW_ERROR.
 *
 * @param dest Destination buffer, can be NULL if destCapacity==0.
 * @param destCapacity Number of UChars available at dest.
 * @param length Number of UChars that were (to be) written to dest.
 * @param pErrorCode ICU error code.
 * @return length
 */
U_CAPI int32_t U_EXPORT2
u_terminateUChars(UChar *dest, int32_t destCapacity, int32_t length, UErrorCode *pErrorCode);

/**
 * NUL-terminate a char * string if possible.
 * Same as u_terminateUChars() but for a different string type.
 */
U_CAPI int32_t U_EXPORT2
u_terminateChars(char *dest, int32_t destCapacity, int32_t length, UErrorCode *pErrorCode);

/**
 * NUL-terminate a UChar32 * string if possible.
 * Same as u_terminateUChars() but for a different string type.
 */
U_CAPI int32_t U_EXPORT2
u_terminateUChar32s(UChar32 *dest, int32_t destCapacity, int32_t length, UErrorCode *pErrorCode);

/**
 * NUL-terminate a wchar_t * string if possible.
 * Same as u_terminateUChars() but for a different string type.
 */
U_CAPI int32_t U_EXPORT2
u_terminateWChars(wchar_t *dest, int32_t destCapacity, int32_t length, UErrorCode *pErrorCode);

/**
 * Counts the bytes of any whole valid sequence for a UTF-8 lead byte.
 * Returns 1 for ASCII 0..0x7f.
 * Returns 0 for 0x80..0xc1 as well as for 0xf5..0xff.
 * leadByte might be evaluated multiple times.
 *
 * @param leadByte The first byte of a UTF-8 sequence. Must be 0..0xff.
 * @return 0..4
 */
#define U8_COUNT_BYTES(leadByte) \
    (U8_IS_SINGLE(leadByte) ? 1 : U8_COUNT_BYTES_NON_ASCII(leadByte))

/**
 * Counts the bytes of any whole valid sequence for a UTF-8 lead byte.
 * Returns 0 for 0x00..0xc1 as well as for 0xf5..0xff.
 * leadByte might be evaluated multiple times.
 *
 * @param leadByte The first byte of a UTF-8 sequence. Must be 0..0xff.
 * @return 0 or 2..4
 */
#define U8_COUNT_BYTES_NON_ASCII(leadByte) \
    (U8_IS_LEAD(leadByte) ? ((uint8_t)(leadByte)>=0xe0)+((uint8_t)(leadByte)>=0xf0)+2 : 0)

#ifdef __cplusplus

U_NAMESPACE_BEGIN

class UTF8 {
public:
    UTF8() = delete;  // all static

    /**
     * Is t a valid UTF-8 trail byte?
     *
     * @param prev Must be the preceding lead byte if i==1 and length>=3;
     *             otherwise ignored.
     * @param t The i-th byte following the lead byte.
     * @param i The index (1..3) of byte t in the byte sequence. 0<i<length
     * @param length The length (2..4) of the byte sequence according to the lead byte.
     * @return TRUE if t is a valid trail byte in this context.
     */
    static inline UBool isValidTrail(int32_t prev, uint8_t t, int32_t i, int32_t length) {
        // The first trail byte after a 3- or 4-byte lead byte
        // needs to be validated together with its lead byte.
        if (length <= 2 || i > 1) {
            return U8_IS_TRAIL(t);
        } else if (length == 3) {
            return U8_IS_VALID_LEAD3_AND_T1(prev, t);
        } else {  // length == 4
            return U8_IS_VALID_LEAD4_AND_T1(prev, t);
        }
    }
};

U_NAMESPACE_END

#endif  // __cplusplus



U_NAMESPACE_BEGIN

/*
 * Note: This builder implementation stores (bytes, value) pairs with full copies
 * of the byte sequences, until the BytesTrie is built.
 * It might(!) take less memory if we collected the data in a temporary, dynamic trie.
 */

class BytesTrieElement : public UMemory {
public:
    // Use compiler's default constructor, initializes nothing.

    void setTo(StringPiece s, int32_t val, CharString &strings, UErrorCode &errorCode);

    StringPiece getString(const CharString &strings) const {
        int32_t offset=stringOffset;
        int32_t length;
        if(offset>=0) {
            length=(uint8_t)strings[offset++];
        } else {
            offset=~offset;
            length=((int32_t)(uint8_t)strings[offset]<<8)|(uint8_t)strings[offset+1];
            offset+=2;
        }
        return StringPiece(strings.data()+offset, length);
    }
    int32_t getStringLength(const CharString &strings) const {
        int32_t offset=stringOffset;
        if(offset>=0) {
            return (uint8_t)strings[offset];
        } else {
            offset=~offset;
            return ((int32_t)(uint8_t)strings[offset]<<8)|(uint8_t)strings[offset+1];
        }
    }

    char charAt(int32_t index, const CharString &strings) const { return data(strings)[index]; }

    int32_t getValue() const { return value; }

    int32_t compareStringTo(const BytesTrieElement &o, const CharString &strings) const;

private:
    const char *data(const CharString &strings) const {
        int32_t offset=stringOffset;
        if(offset>=0) {
            ++offset;
        } else {
            offset=~offset+2;
        }
        return strings.data()+offset;
    }

    // If the stringOffset is non-negative, then the first strings byte contains
    // the string length.
    // If the stringOffset is negative, then the first two strings bytes contain
    // the string length (big-endian), and the offset needs to be bit-inverted.
    // (Compared with a stringLength field here, this saves 3 bytes per string for most strings.)
    int32_t stringOffset;
    int32_t value;
};

void
BytesTrieElement::setTo(StringPiece s, int32_t val,
                        CharString &strings, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    int32_t length=s.length();
    if(length>0xffff) {
        // Too long: We store the length in 1 or 2 bytes.
        errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return;
    }
    int32_t offset=strings.length();
    if(length>0xff) {
        offset=~offset;
        strings.append((char)(length>>8), errorCode);
    }
    strings.append((char)length, errorCode);
    stringOffset=offset;
    value=val;
    strings.append(s, errorCode);
}

int32_t
BytesTrieElement::compareStringTo(const BytesTrieElement &other, const CharString &strings) const {
    // TODO: add StringPiece::compare(), see ticket #8187
    StringPiece thisString=getString(strings);
    StringPiece otherString=other.getString(strings);
    int32_t lengthDiff=thisString.length()-otherString.length();
    int32_t commonLength;
    if(lengthDiff<=0) {
        commonLength=thisString.length();
    } else {
        commonLength=otherString.length();
    }
    int32_t diff=uprv_memcmp(thisString.data(), otherString.data(), commonLength);
    return diff!=0 ? diff : lengthDiff;
}

BytesTrieBuilder::BytesTrieBuilder(UErrorCode &errorCode)
        : strings(NULL), elements(NULL), elementsCapacity(0), elementsLength(0),
          bytes(NULL), bytesCapacity(0), bytesLength(0) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    strings=new CharString();
    if(strings==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
    }
}

BytesTrieBuilder::~BytesTrieBuilder() {
    delete strings;
    delete[] elements;
    uprv_free(bytes);
}

BytesTrieBuilder &
BytesTrieBuilder::add(StringPiece s, int32_t value, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return *this;
    }
    if(bytesLength>0) {
        // Cannot add elements after building.
        errorCode=U_NO_WRITE_PERMISSION;
        return *this;
    }
    if(elementsLength==elementsCapacity) {
        int32_t newCapacity;
        if(elementsCapacity==0) {
            newCapacity=1024;
        } else {
            newCapacity=4*elementsCapacity;
        }
        BytesTrieElement *newElements=new BytesTrieElement[newCapacity];
        if(newElements==NULL) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            return *this; // error instead of dereferencing null
        }
        if(elementsLength>0) {
            uprv_memcpy(newElements, elements, (size_t)elementsLength*sizeof(BytesTrieElement));
        }
        delete[] elements;
        elements=newElements;
        elementsCapacity=newCapacity;
    }
    elements[elementsLength++].setTo(s, value, *strings, errorCode);
    return *this;
}

U_CDECL_BEGIN

static int32_t U_CALLCONV
bytestriebuilder_compareElementStrings(const void *context, const void *left, const void *right) {
    const CharString *strings=static_cast<const CharString *>(context);
    const BytesTrieElement *leftElement=static_cast<const BytesTrieElement *>(left);
    const BytesTrieElement *rightElement=static_cast<const BytesTrieElement *>(right);
    return leftElement->compareStringTo(*rightElement, *strings);
}

U_CDECL_END

BytesTrie *
BytesTrieBuilder::build(UStringTrieBuildOption buildOption, UErrorCode &errorCode) {
    buildBytes(buildOption, errorCode);
    BytesTrie *newTrie=NULL;
    if(U_SUCCESS(errorCode)) {
        newTrie=new BytesTrie(bytes, bytes+(bytesCapacity-bytesLength));
        if(newTrie==NULL) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
        } else {
            bytes=NULL;  // The new trie now owns the array.
            bytesCapacity=0;
        }
    }
    return newTrie;
}

StringPiece
BytesTrieBuilder::buildStringPiece(UStringTrieBuildOption buildOption, UErrorCode &errorCode) {
    buildBytes(buildOption, errorCode);
    StringPiece result;
    if(U_SUCCESS(errorCode)) {
        result.set(bytes+(bytesCapacity-bytesLength), bytesLength);
    }
    return result;
}

void
BytesTrieBuilder::buildBytes(UStringTrieBuildOption buildOption, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    if(bytes!=NULL && bytesLength>0) {
        // Already built.
        return;
    }
    if(bytesLength==0) {
        if(elementsLength==0) {
            errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return;
        }
        uprv_sortArray(elements, elementsLength, (int32_t)sizeof(BytesTrieElement),
                      bytestriebuilder_compareElementStrings, strings,
                      FALSE,  // need not be a stable sort
                      &errorCode);
        if(U_FAILURE(errorCode)) {
            return;
        }
        // Duplicate strings are not allowed.
        StringPiece prev=elements[0].getString(*strings);
        for(int32_t i=1; i<elementsLength; ++i) {
            StringPiece current=elements[i].getString(*strings);
            if(prev==current) {
                errorCode=U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
            prev=current;
        }
    }
    // Create and byte-serialize the trie for the elements.
    bytesLength=0;
    int32_t capacity=strings->length();
    if(capacity<1024) {
        capacity=1024;
    }
    if(bytesCapacity<capacity) {
        uprv_free(bytes);
        bytes=static_cast<char *>(uprv_malloc(capacity));
        if(bytes==NULL) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            bytesCapacity=0;
            return;
        }
        bytesCapacity=capacity;
    }
    StringTrieBuilder::build(buildOption, elementsLength, errorCode);
    if(bytes==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
    }
}

BytesTrieBuilder &
BytesTrieBuilder::clear() {
    strings->clear();
    elementsLength=0;
    bytesLength=0;
    return *this;
}

int32_t
BytesTrieBuilder::getElementStringLength(int32_t i) const {
    return elements[i].getStringLength(*strings);
}

UChar
BytesTrieBuilder::getElementUnit(int32_t i, int32_t byteIndex) const {
    return (uint8_t)elements[i].charAt(byteIndex, *strings);
}

int32_t
BytesTrieBuilder::getElementValue(int32_t i) const {
    return elements[i].getValue();
}

int32_t
BytesTrieBuilder::getLimitOfLinearMatch(int32_t first, int32_t last, int32_t byteIndex) const {
    const BytesTrieElement &firstElement=elements[first];
    const BytesTrieElement &lastElement=elements[last];
    int32_t minStringLength=firstElement.getStringLength(*strings);
    while(++byteIndex<minStringLength &&
            firstElement.charAt(byteIndex, *strings)==
            lastElement.charAt(byteIndex, *strings)) {}
    return byteIndex;
}

int32_t
BytesTrieBuilder::countElementUnits(int32_t start, int32_t limit, int32_t byteIndex) const {
    int32_t length=0;  // Number of different bytes at byteIndex.
    int32_t i=start;
    do {
        char byte=elements[i++].charAt(byteIndex, *strings);
        while(i<limit && byte==elements[i].charAt(byteIndex, *strings)) {
            ++i;
        }
        ++length;
    } while(i<limit);
    return length;
}

int32_t
BytesTrieBuilder::skipElementsBySomeUnits(int32_t i, int32_t byteIndex, int32_t count) const {
    do {
        char byte=elements[i++].charAt(byteIndex, *strings);
        while(byte==elements[i].charAt(byteIndex, *strings)) {
            ++i;
        }
    } while(--count>0);
    return i;
}

int32_t
BytesTrieBuilder::indexOfElementWithNextUnit(int32_t i, int32_t byteIndex, UChar byte) const {
    char b=(char)byte;
    while(b==elements[i].charAt(byteIndex, *strings)) {
        ++i;
    }
    return i;
}

BytesTrieBuilder::BTLinearMatchNode::BTLinearMatchNode(const char *bytes, int32_t len, Node *nextNode)
        : LinearMatchNode(len, nextNode), s(bytes) {
    hash=static_cast<int32_t>(
        static_cast<uint32_t>(hash)*37u + static_cast<uint32_t>(ustr_hashCharsN(bytes, len)));
}

UBool
BytesTrieBuilder::BTLinearMatchNode::operator==(const Node &other) const {
    if(this==&other) {
        return TRUE;
    }
    if(!LinearMatchNode::operator==(other)) {
        return FALSE;
    }
    const BTLinearMatchNode &o=(const BTLinearMatchNode &)other;
    return 0==uprv_memcmp(s, o.s, length);
}

void
BytesTrieBuilder::BTLinearMatchNode::write(StringTrieBuilder &builder) {
    BytesTrieBuilder &b=(BytesTrieBuilder &)builder;
    next->write(builder);
    b.write(s, length);
    offset=b.write(b.getMinLinearMatch()+length-1);
}

StringTrieBuilder::Node *
BytesTrieBuilder::createLinearMatchNode(int32_t i, int32_t byteIndex, int32_t length,
                                        Node *nextNode) const {
    return new BTLinearMatchNode(
            elements[i].getString(*strings).data()+byteIndex,
            length,
            nextNode);
}

UBool
BytesTrieBuilder::ensureCapacity(int32_t length) {
    if(bytes==NULL) {
        return FALSE;  // previous memory allocation had failed
    }
    if(length>bytesCapacity) {
        int32_t newCapacity=bytesCapacity;
        do {
            newCapacity*=2;
        } while(newCapacity<=length);
        char *newBytes=static_cast<char *>(uprv_malloc(newCapacity));
        if(newBytes==NULL) {
            // unable to allocate memory
            uprv_free(bytes);
            bytes=NULL;
            bytesCapacity=0;
            return FALSE;
        }
        uprv_memcpy(newBytes+(newCapacity-bytesLength),
                    bytes+(bytesCapacity-bytesLength), bytesLength);
        uprv_free(bytes);
        bytes=newBytes;
        bytesCapacity=newCapacity;
    }
    return TRUE;
}

int32_t
BytesTrieBuilder::write(int32_t byte) {
    int32_t newLength=bytesLength+1;
    if(ensureCapacity(newLength)) {
        bytesLength=newLength;
        bytes[bytesCapacity-bytesLength]=(char)byte;
    }
    return bytesLength;
}

int32_t
BytesTrieBuilder::write(const char *b, int32_t length) {
    int32_t newLength=bytesLength+length;
    if(ensureCapacity(newLength)) {
        bytesLength=newLength;
        uprv_memcpy(bytes+(bytesCapacity-bytesLength), b, length);
    }
    return bytesLength;
}

int32_t
BytesTrieBuilder::writeElementUnits(int32_t i, int32_t byteIndex, int32_t length) {
    return write(elements[i].getString(*strings).data()+byteIndex, length);
}

int32_t
BytesTrieBuilder::writeValueAndFinal(int32_t i, UBool isFinal) {
    if(0<=i && i<=BytesTrie::kMaxOneByteValue) {
        return write(((BytesTrie::kMinOneByteValueLead+i)<<1)|isFinal);
    }
    char intBytes[5];
    int32_t length=1;
    if(i<0 || i>0xffffff) {
        intBytes[0]=(char)BytesTrie::kFiveByteValueLead;
        intBytes[1]=(char)((uint32_t)i>>24);
        intBytes[2]=(char)((uint32_t)i>>16);
        intBytes[3]=(char)((uint32_t)i>>8);
        intBytes[4]=(char)i;
        length=5;
    // } else if(i<=BytesTrie::kMaxOneByteValue) {
    //     intBytes[0]=(char)(BytesTrie::kMinOneByteValueLead+i);
    } else {
        if(i<=BytesTrie::kMaxTwoByteValue) {
            intBytes[0]=(char)(BytesTrie::kMinTwoByteValueLead+(i>>8));
        } else {
            if(i<=BytesTrie::kMaxThreeByteValue) {
                intBytes[0]=(char)(BytesTrie::kMinThreeByteValueLead+(i>>16));
            } else {
                intBytes[0]=(char)BytesTrie::kFourByteValueLead;
                intBytes[1]=(char)(i>>16);
                length=2;
            }
            intBytes[length++]=(char)(i>>8);
        }
        intBytes[length++]=(char)i;
    }
    intBytes[0]=(char)((intBytes[0]<<1)|isFinal);
    return write(intBytes, length);
}

int32_t
BytesTrieBuilder::writeValueAndType(UBool hasValue, int32_t value, int32_t node) {
    int32_t offset=write(node);
    if(hasValue) {
        offset=writeValueAndFinal(value, FALSE);
    }
    return offset;
}

int32_t
BytesTrieBuilder::writeDeltaTo(int32_t jumpTarget) {
    int32_t i=bytesLength-jumpTarget;
    U_ASSERT(i>=0);
    if(i<=BytesTrie::kMaxOneByteDelta) {
        return write(i);
    }
    char intBytes[5];
    int32_t length;
    if(i<=BytesTrie::kMaxTwoByteDelta) {
        intBytes[0]=(char)(BytesTrie::kMinTwoByteDeltaLead+(i>>8));
        length=1;
    } else {
        if(i<=BytesTrie::kMaxThreeByteDelta) {
            intBytes[0]=(char)(BytesTrie::kMinThreeByteDeltaLead+(i>>16));
            length=2;
        } else {
            if(i<=0xffffff) {
                intBytes[0]=(char)BytesTrie::kFourByteDeltaLead;
                length=3;
            } else {
                intBytes[0]=(char)BytesTrie::kFiveByteDeltaLead;
                intBytes[1]=(char)(i>>24);
                length=4;
            }
            intBytes[1]=(char)(i>>16);
        }
        intBytes[1]=(char)(i>>8);
    }
    intBytes[length++]=(char)i;
    return write(intBytes, length);
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  bytestrieiterator.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010nov03
*   created by: Markus W. Scherer
*/





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1999-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/

//
//  UVector32 is a class implementing a vector of 32 bit integers.
//            It is similar to UVector, but holds int32_t values rather than pointers.
//            Most of the code is unchanged from UVector.
//







U_NAMESPACE_BEGIN



/**
 * <p>Ultralightweight C++ implementation of a <tt>void*</tt> vector
 * that is (mostly) compatible with java.util.Vector.
 *
 * <p>This is a very simple implementation, written to satisfy an
 * immediate porting need.  As such, it is not completely fleshed out,
 * and it aims for simplicity and conformity.  Nonetheless, it serves
 * its purpose (porting code from java that uses java.util.Vector)
 * well, and it could be easily made into a more robust vector class.
 *
 * <p><b>Design notes</b>
 *
 * <p>There is index bounds checking, but little is done about it.  If
 * indices are out of bounds, either nothing happens, or zero is
 * returned.  We <em>do</em> avoid indexing off into the weeds.
 *
 * <p>There is detection of out of memory, but the handling is very
 * coarse-grained -- similar to UnicodeString's protocol, but even
 * coarser.  The class contains <em>one static flag</em> that is set
 * when any call to <tt>new</tt> returns zero.  This allows the caller
 * to use several vectors and make just one check at the end to see if
 * a memory failure occurred.  This is more efficient than making a
 * check after each call on each vector when doing many operations on
 * multiple vectors.  The single static flag works best when memory
 * failures are infrequent, and when recovery options are limited or
 * nonexistent.
 *
 * <p><b>To do</b>
 *
 * <p>Improve the handling of index out of bounds errors.
 *
 * @author Alan Liu
 */
class U_COMMON_API UVector32 : public UObject {
private:
    int32_t   count;

    int32_t   capacity;
    
    int32_t   maxCapacity;   // Limit beyond which capacity is not permitted to grow.

    int32_t*  elements;

public:
    UVector32(UErrorCode &status);

    UVector32(int32_t initialCapacity, UErrorCode &status);

    virtual ~UVector32();

    /**
     * Assign this object to another (make this a copy of 'other').
     * Use the 'assign' function to assign each element.
     */
    void assign(const UVector32& other, UErrorCode &ec);

    /**
     * Compare this vector with another.  They will be considered
     * equal if they are of the same size and all elements are equal,
     * as compared using this object's comparer.
     */
    UBool operator==(const UVector32& other);

    /**
     * Equivalent to !operator==()
     */
    inline UBool operator!=(const UVector32& other);

    //------------------------------------------------------------
    // java.util.Vector API
    //------------------------------------------------------------

    inline void addElement(int32_t elem, UErrorCode &status);

    void setElementAt(int32_t elem, int32_t index);

    void insertElementAt(int32_t elem, int32_t index, UErrorCode &status);
    
    inline int32_t elementAti(int32_t index) const;

    UBool equals(const UVector32 &other) const;

    inline int32_t lastElementi(void) const;

    int32_t indexOf(int32_t elem, int32_t startIndex = 0) const;

    inline UBool contains(int32_t elem) const;

    UBool containsAll(const UVector32& other) const;

    UBool removeAll(const UVector32& other);

    UBool retainAll(const UVector32& other);

    void removeElementAt(int32_t index);

    void removeAllElements();

    inline int32_t size(void) const;

    inline UBool isEmpty(void) const;

    // Inline.  Use this one for speedy size check.
    inline UBool ensureCapacity(int32_t minimumCapacity, UErrorCode &status);

    // Out-of-line, handles actual growth.  Called by ensureCapacity() when necessary.
    UBool expandCapacity(int32_t minimumCapacity, UErrorCode &status);

    /**
     * Change the size of this vector as follows: If newSize is
     * smaller, then truncate the array, possibly deleting held
     * elements for i >= newSize.  If newSize is larger, grow the
     * array, filling in new slows with zero.
     */
    void setSize(int32_t newSize);

    //------------------------------------------------------------
    // New API
    //------------------------------------------------------------

    /**
     * Returns true if this vector contains none of the elements
     * of the given vector.
     * @param other vector to be checked for containment
     * @return true if the test condition is met
     */
    UBool containsNone(const UVector32& other) const;


    /**
     * Insert the given integer into this vector at its sorted position.
     * The current elements are assumed to be sorted already.
     */
    void sortedInsert(int32_t elem, UErrorCode& ec);

    /**
     * Returns a pointer to the internal array holding the vector.
     */
    inline int32_t *getBuffer() const;

    /**
     * Set the maximum allowed buffer capacity for this vector/stack.
     * Default with no limit set is unlimited, go until malloc() fails.
     * A Limit of zero means unlimited capacity.
     * Units are vector elements (32 bits each), not bytes.
     */
    void setMaxCapacity(int32_t limit);

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     */
    virtual UClassID getDynamicClassID() const;

private:
    void _init(int32_t initialCapacity, UErrorCode &status);

    // Disallow
    UVector32(const UVector32&);

    // Disallow
    UVector32& operator=(const UVector32&);


    //  API Functions for Stack operations.
    //  In the original UVector, these were in a separate derived class, UStack.
    //  Here in UVector32, they are all together.
public:
    inline UBool empty(void) const;   // TODO:  redundant, same as empty().  Remove it?

    inline int32_t peeki(void) const;
    
    inline int32_t popi(void);
    
    inline int32_t push(int32_t i, UErrorCode &status);

    inline int32_t *reserveBlock(int32_t size, UErrorCode &status);
    inline int32_t *popFrame(int32_t size);
};


// UVector32 inlines

inline UBool UVector32::ensureCapacity(int32_t minimumCapacity, UErrorCode &status) {
    if ((minimumCapacity >= 0) && (capacity >= minimumCapacity)) {
        return TRUE;
    } else {
        return expandCapacity(minimumCapacity, status);
    }
}

inline int32_t UVector32::elementAti(int32_t index) const {
    return (index >= 0 && count > 0 && count - index > 0) ? elements[index] : 0;
}


inline void UVector32::addElement(int32_t elem, UErrorCode &status) {
    if (ensureCapacity(count + 1, status)) {
        elements[count] = elem;
        count++;
    }
}

inline int32_t *UVector32::reserveBlock(int32_t size, UErrorCode &status) {
    if (ensureCapacity(count+size, status) == FALSE) {
        return NULL;
    }
    int32_t  *rp = elements+count;
    count += size;
    return rp;
}

inline int32_t *UVector32::popFrame(int32_t size) {
    U_ASSERT(count >= size);
    count -= size;
    if (count < 0) {
        count = 0;
    }
    return elements+count-size;
}



inline int32_t UVector32::size(void) const {
    return count;
}

inline UBool UVector32::isEmpty(void) const {
    return count == 0;
}

inline UBool UVector32::contains(int32_t obj) const {
    return indexOf(obj) >= 0;
}

inline int32_t UVector32::lastElementi(void) const {
    return elementAti(count-1);
}

inline UBool UVector32::operator!=(const UVector32& other) {
    return !operator==(other);
}

inline int32_t *UVector32::getBuffer() const {
    return elements;
}


// UStack inlines

inline UBool UVector32::empty(void) const {
    return isEmpty();
}

inline int32_t UVector32::peeki(void) const {
    return lastElementi();
}

inline int32_t UVector32::push(int32_t i, UErrorCode &status) {
    addElement(i, status);
    return i;
}

inline int32_t UVector32::popi(void) {
    int32_t result = 0;
    if (count > 0) {
        count--;
        result = elements[count];
    }
    return result;
}

U_NAMESPACE_END



U_NAMESPACE_BEGIN

BytesTrie::Iterator::Iterator(const void *trieBytes, int32_t maxStringLength,
                              UErrorCode &errorCode)
        : bytes_(static_cast<const uint8_t *>(trieBytes)),
          pos_(bytes_), initialPos_(bytes_),
          remainingMatchLength_(-1), initialRemainingMatchLength_(-1),
          str_(NULL), maxLength_(maxStringLength), value_(0), stack_(NULL) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    // str_ and stack_ are pointers so that it's easy to turn bytestrie.h into
    // a public API header for which we would want it to depend only on
    // other public headers.
    // Unlike BytesTrie itself, its Iterator performs memory allocations anyway
    // via the CharString and UVector32 implementations, so this additional
    // cost is minimal.
    str_=new CharString();
    stack_=new UVector32(errorCode);
    if(U_SUCCESS(errorCode) && (str_==NULL || stack_==NULL)) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
    }
}

BytesTrie::Iterator::Iterator(const BytesTrie &trie, int32_t maxStringLength,
                              UErrorCode &errorCode)
        : bytes_(trie.bytes_), pos_(trie.pos_), initialPos_(trie.pos_),
          remainingMatchLength_(trie.remainingMatchLength_),
          initialRemainingMatchLength_(trie.remainingMatchLength_),
          str_(NULL), maxLength_(maxStringLength), value_(0), stack_(NULL) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    str_=new CharString();
    stack_=new UVector32(errorCode);
    if(U_FAILURE(errorCode)) {
        return;
    }
    if(str_==NULL || stack_==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    int32_t length=remainingMatchLength_;  // Actual remaining match length minus 1.
    if(length>=0) {
        // Pending linear-match node, append remaining bytes to str_.
        ++length;
        if(maxLength_>0 && length>maxLength_) {
            length=maxLength_;  // This will leave remainingMatchLength>=0 as a signal.
        }
        str_->append(reinterpret_cast<const char *>(pos_), length, errorCode);
        pos_+=length;
        remainingMatchLength_-=length;
    }
}

BytesTrie::Iterator::~Iterator() {
    delete str_;
    delete stack_;
}

BytesTrie::Iterator &
BytesTrie::Iterator::reset() {
    pos_=initialPos_;
    remainingMatchLength_=initialRemainingMatchLength_;
    int32_t length=remainingMatchLength_+1;  // Remaining match length.
    if(maxLength_>0 && length>maxLength_) {
        length=maxLength_;
    }
    str_->truncate(length);
    pos_+=length;
    remainingMatchLength_-=length;
    stack_->setSize(0);
    return *this;
}

UBool
BytesTrie::Iterator::hasNext() const { return pos_!=NULL || !stack_->isEmpty(); }

UBool
BytesTrie::Iterator::next(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return FALSE;
    }
    const uint8_t *pos=pos_;
    if(pos==NULL) {
        if(stack_->isEmpty()) {
            return FALSE;
        }
        // Pop the state off the stack and continue with the next outbound edge of
        // the branch node.
        int32_t stackSize=stack_->size();
        int32_t length=stack_->elementAti(stackSize-1);
        pos=bytes_+stack_->elementAti(stackSize-2);
        stack_->setSize(stackSize-2);
        str_->truncate(length&0xffff);
        length=(int32_t)((uint32_t)length>>16);
        if(length>1) {
            pos=branchNext(pos, length, errorCode);
            if(pos==NULL) {
                return TRUE;  // Reached a final value.
            }
        } else {
            str_->append((char)*pos++, errorCode);
        }
    }
    if(remainingMatchLength_>=0) {
        // We only get here if we started in a pending linear-match node
        // with more than maxLength remaining bytes.
        return truncateAndStop();
    }
    for(;;) {
        int32_t node=*pos++;
        if(node>=kMinValueLead) {
            // Deliver value for the byte sequence so far.
            UBool isFinal=(UBool)(node&kValueIsFinal);
            value_=readValue(pos, node>>1);
            if(isFinal || (maxLength_>0 && str_->length()==maxLength_)) {
                pos_=NULL;
            } else {
                pos_=skipValue(pos, node);
            }
            return TRUE;
        }
        if(maxLength_>0 && str_->length()==maxLength_) {
            return truncateAndStop();
        }
        if(node<kMinLinearMatch) {
            if(node==0) {
                node=*pos++;
            }
            pos=branchNext(pos, node+1, errorCode);
            if(pos==NULL) {
                return TRUE;  // Reached a final value.
            }
        } else {
            // Linear-match node, append length bytes to str_.
            int32_t length=node-kMinLinearMatch+1;
            if(maxLength_>0 && str_->length()+length>maxLength_) {
                str_->append(reinterpret_cast<const char *>(pos),
                            maxLength_-str_->length(), errorCode);
                return truncateAndStop();
            }
            str_->append(reinterpret_cast<const char *>(pos), length, errorCode);
            pos+=length;
        }
    }
}

StringPiece
BytesTrie::Iterator::getString() const {
    return str_ == NULL ? StringPiece() : str_->toStringPiece();
}

UBool
BytesTrie::Iterator::truncateAndStop() {
    pos_=NULL;
    value_=-1;  // no real value for str
    return TRUE;
}

// Branch node, needs to take the first outbound edge and push state for the rest.
const uint8_t *
BytesTrie::Iterator::branchNext(const uint8_t *pos, int32_t length, UErrorCode &errorCode) {
    while(length>kMaxBranchLinearSubNodeLength) {
        ++pos;  // ignore the comparison byte
        // Push state for the greater-or-equal edge.
        stack_->addElement((int32_t)(skipDelta(pos)-bytes_), errorCode);
        stack_->addElement(((length-(length>>1))<<16)|str_->length(), errorCode);
        // Follow the less-than edge.
        length>>=1;
        pos=jumpByDelta(pos);
    }
    // List of key-value pairs where values are either final values or jump deltas.
    // Read the first (key, value) pair.
    uint8_t trieByte=*pos++;
    int32_t node=*pos++;
    UBool isFinal=(UBool)(node&kValueIsFinal);
    int32_t value=readValue(pos, node>>1);
    pos=skipValue(pos, node);
    stack_->addElement((int32_t)(pos-bytes_), errorCode);
    stack_->addElement(((length-1)<<16)|str_->length(), errorCode);
    str_->append((char)trieByte, errorCode);
    if(isFinal) {
        pos_=NULL;
        value_=value;
        return NULL;
    } else {
        return pos+value;
    }
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *****************************************************************************
 * Copyright (C) 1996-2015, International Business Machines Corporation and
 * others. All Rights Reserved.
 *****************************************************************************
 */



#if !UCONFIG_NO_NORMALIZATION

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 1996-2014, International Business Machines Corporation and
 * others. All Rights Reserved.
 *******************************************************************************
 */




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_NORMALIZATION




/**
 * \file
 * \brief C++ API: Canonical Iterator
 */
 
/** Should permutation skip characters with combining class zero
 *  Should be either TRUE or FALSE. This is a compile time option
 *  @stable ICU 2.4
 */
#ifndef CANITER_SKIP_ZEROES
#define CANITER_SKIP_ZEROES TRUE
#endif

U_NAMESPACE_BEGIN

class Hashtable;
class Normalizer2;
class Normalizer2Impl;

/**
 * This class allows one to iterate through all the strings that are canonically equivalent to a given
 * string. For example, here are some sample results:
Results for: {LATIN CAPITAL LETTER A WITH RING ABOVE}{LATIN SMALL LETTER D}{COMBINING DOT ABOVE}{COMBINING CEDILLA}
1: \\u0041\\u030A\\u0064\\u0307\\u0327
 = {LATIN CAPITAL LETTER A}{COMBINING RING ABOVE}{LATIN SMALL LETTER D}{COMBINING DOT ABOVE}{COMBINING CEDILLA}
2: \\u0041\\u030A\\u0064\\u0327\\u0307
 = {LATIN CAPITAL LETTER A}{COMBINING RING ABOVE}{LATIN SMALL LETTER D}{COMBINING CEDILLA}{COMBINING DOT ABOVE}
3: \\u0041\\u030A\\u1E0B\\u0327
 = {LATIN CAPITAL LETTER A}{COMBINING RING ABOVE}{LATIN SMALL LETTER D WITH DOT ABOVE}{COMBINING CEDILLA}
4: \\u0041\\u030A\\u1E11\\u0307
 = {LATIN CAPITAL LETTER A}{COMBINING RING ABOVE}{LATIN SMALL LETTER D WITH CEDILLA}{COMBINING DOT ABOVE}
5: \\u00C5\\u0064\\u0307\\u0327
 = {LATIN CAPITAL LETTER A WITH RING ABOVE}{LATIN SMALL LETTER D}{COMBINING DOT ABOVE}{COMBINING CEDILLA}
6: \\u00C5\\u0064\\u0327\\u0307
 = {LATIN CAPITAL LETTER A WITH RING ABOVE}{LATIN SMALL LETTER D}{COMBINING CEDILLA}{COMBINING DOT ABOVE}
7: \\u00C5\\u1E0B\\u0327
 = {LATIN CAPITAL LETTER A WITH RING ABOVE}{LATIN SMALL LETTER D WITH DOT ABOVE}{COMBINING CEDILLA}
8: \\u00C5\\u1E11\\u0307
 = {LATIN CAPITAL LETTER A WITH RING ABOVE}{LATIN SMALL LETTER D WITH CEDILLA}{COMBINING DOT ABOVE}
9: \\u212B\\u0064\\u0307\\u0327
 = {ANGSTROM SIGN}{LATIN SMALL LETTER D}{COMBINING DOT ABOVE}{COMBINING CEDILLA}
10: \\u212B\\u0064\\u0327\\u0307
 = {ANGSTROM SIGN}{LATIN SMALL LETTER D}{COMBINING CEDILLA}{COMBINING DOT ABOVE}
11: \\u212B\\u1E0B\\u0327
 = {ANGSTROM SIGN}{LATIN SMALL LETTER D WITH DOT ABOVE}{COMBINING CEDILLA}
12: \\u212B\\u1E11\\u0307
 = {ANGSTROM SIGN}{LATIN SMALL LETTER D WITH CEDILLA}{COMBINING DOT ABOVE}
 *<br>Note: the code is intended for use with small strings, and is not suitable for larger ones,
 * since it has not been optimized for that situation.
 * Note, CanonicalIterator is not intended to be subclassed.
 * @author M. Davis
 * @author C++ port by V. Weinstein
 * @stable ICU 2.4
 */
class U_COMMON_API CanonicalIterator U_FINAL : public UObject {
public:
    /**
     * Construct a CanonicalIterator object
     * @param source    string to get results for
     * @param status    Fill-in parameter which receives the status of this operation.
     * @stable ICU 2.4
     */
    CanonicalIterator(const UnicodeString &source, UErrorCode &status);

    /** Destructor
     *  Cleans pieces
     * @stable ICU 2.4
     */
    virtual ~CanonicalIterator();

    /**
     * Gets the NFD form of the current source we are iterating over.
     * @return gets the source: NOTE: it is the NFD form of source
     * @stable ICU 2.4
     */
    UnicodeString getSource();

    /**
     * Resets the iterator so that one can start again from the beginning.
     * @stable ICU 2.4
     */
    void reset();

    /**
     * Get the next canonically equivalent string.
     * <br><b>Warning: The strings are not guaranteed to be in any particular order.</b>
     * @return the next string that is canonically equivalent. A bogus string is returned when
     * the iteration is done.
     * @stable ICU 2.4
     */
    UnicodeString next();

    /**
     * Set a new source for this iterator. Allows object reuse.
     * @param newSource     the source string to iterate against. This allows the same iterator to be used
     *                     while changing the source string, saving object creation.
     * @param status        Fill-in parameter which receives the status of this operation.
     * @stable ICU 2.4
     */
    void setSource(const UnicodeString &newSource, UErrorCode &status);

#ifndef U_HIDE_INTERNAL_API
    /**
     * Dumb recursive implementation of permutation.
     * TODO: optimize
     * @param source     the string to find permutations for
     * @param skipZeros  determine if skip zeros
     * @param result     the results in a set.
     * @param status       Fill-in parameter which receives the status of this operation.
     * @internal
     */
    static void U_EXPORT2 permute(UnicodeString &source, UBool skipZeros, Hashtable *result, UErrorCode &status);
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 2.2
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 2.2
     */
    virtual UClassID getDynamicClassID() const;

private:
    // ===================== PRIVATES ==============================
    // private default constructor
    CanonicalIterator();


    /**
     * Copy constructor. Private for now.
     * @internal (private)
     */
    CanonicalIterator(const CanonicalIterator& other);

    /**
     * Assignment operator. Private for now.
     * @internal (private)
     */
    CanonicalIterator& operator=(const CanonicalIterator& other);

    // fields
    UnicodeString source;
    UBool done;

    // 2 dimensional array holds the pieces of the string with
    // their different canonically equivalent representations
    UnicodeString **pieces;
    int32_t pieces_length;
    int32_t *pieces_lengths;

    // current is used in iterating to combine pieces
    int32_t *current;
    int32_t current_length;

    // transient fields
    UnicodeString buffer;

    const Normalizer2 &nfd;
    const Normalizer2Impl &nfcImpl;

    // we have a segment, in NFD. Find all the strings that are canonically equivalent to it.
    UnicodeString *getEquivalents(const UnicodeString &segment, int32_t &result_len, UErrorCode &status); //private String[] getEquivalents(String segment)

    //Set getEquivalents2(String segment);
    Hashtable *getEquivalents2(Hashtable *fillinResult, const char16_t *segment, int32_t segLen, UErrorCode &status);
    //Hashtable *getEquivalents2(const UnicodeString &segment, int32_t segLen, UErrorCode &status);

    /**
     * See if the decomposition of cp2 is at segment starting at segmentPos
     * (with canonical rearrangment!)
     * If so, take the remainder, and return the equivalents
     */
    //Set extract(int comp, String segment, int segmentPos, StringBuffer buffer);
    Hashtable *extract(Hashtable *fillinResult, UChar32 comp, const char16_t *segment, int32_t segLen, int32_t segmentPos, UErrorCode &status);
    //Hashtable *extract(UChar32 comp, const UnicodeString &segment, int32_t segLen, int32_t segmentPos, UErrorCode &status);

    void cleanPieces();

};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_NORMALIZATION */

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2009-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  normalizer2.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2009nov22
*   created by: Markus W. Scherer
*/


/**
 * \file
 * \brief C++ API: New API for Unicode Normalization.
 */



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_NORMALIZATION






U_NAMESPACE_BEGIN

class ByteSink;

/**
 * Unicode normalization functionality for standard Unicode normalization or
 * for using custom mapping tables.
 * All instances of this class are unmodifiable/immutable.
 * Instances returned by getInstance() are singletons that must not be deleted by the caller.
 * The Normalizer2 class is not intended for public subclassing.
 *
 * The primary functions are to produce a normalized string and to detect whether
 * a string is already normalized.
 * The most commonly used normalization forms are those defined in
 * http://www.unicode.org/unicode/reports/tr15/
 * However, this API supports additional normalization forms for specialized purposes.
 * For example, NFKC_Casefold is provided via getInstance("nfkc_cf", COMPOSE)
 * and can be used in implementations of UTS #46.
 *
 * Not only are the standard compose and decompose modes supplied,
 * but additional modes are provided as documented in the Mode enum.
 *
 * Some of the functions in this class identify normalization boundaries.
 * At a normalization boundary, the portions of the string
 * before it and starting from it do not interact and can be handled independently.
 *
 * The spanQuickCheckYes() stops at a normalization boundary.
 * When the goal is a normalized string, then the text before the boundary
 * can be copied, and the remainder can be processed with normalizeSecondAndAppend().
 *
 * The hasBoundaryBefore(), hasBoundaryAfter() and isInert() functions test whether
 * a character is guaranteed to be at a normalization boundary,
 * regardless of context.
 * This is used for moving from one normalization boundary to the next
 * or preceding boundary, and for performing iterative normalization.
 *
 * Iterative normalization is useful when only a small portion of a
 * longer string needs to be processed.
 * For example, in ICU, iterative normalization is used by the NormalizationTransliterator
 * (to avoid replacing already-normalized text) and ucol_nextSortKeyPart()
 * (to process only the substring for which sort key bytes are computed).
 *
 * The set of normalization boundaries returned by these functions may not be
 * complete: There may be more boundaries that could be returned.
 * Different functions may return different boundaries.
 * @stable ICU 4.4
 */
class U_COMMON_API Normalizer2 : public UObject {
public:
    /**
     * Destructor.
     * @stable ICU 4.4
     */
    ~Normalizer2();

    /**
     * Returns a Normalizer2 instance for Unicode NFC normalization.
     * Same as getInstance(NULL, "nfc", UNORM2_COMPOSE, errorCode).
     * Returns an unmodifiable singleton instance. Do not delete it.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return the requested Normalizer2, if successful
     * @stable ICU 49
     */
    static const Normalizer2 *
    getNFCInstance(UErrorCode &errorCode);

    /**
     * Returns a Normalizer2 instance for Unicode NFD normalization.
     * Same as getInstance(NULL, "nfc", UNORM2_DECOMPOSE, errorCode).
     * Returns an unmodifiable singleton instance. Do not delete it.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return the requested Normalizer2, if successful
     * @stable ICU 49
     */
    static const Normalizer2 *
    getNFDInstance(UErrorCode &errorCode);

    /**
     * Returns a Normalizer2 instance for Unicode NFKC normalization.
     * Same as getInstance(NULL, "nfkc", UNORM2_COMPOSE, errorCode).
     * Returns an unmodifiable singleton instance. Do not delete it.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return the requested Normalizer2, if successful
     * @stable ICU 49
     */
    static const Normalizer2 *
    getNFKCInstance(UErrorCode &errorCode);

    /**
     * Returns a Normalizer2 instance for Unicode NFKD normalization.
     * Same as getInstance(NULL, "nfkc", UNORM2_DECOMPOSE, errorCode).
     * Returns an unmodifiable singleton instance. Do not delete it.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return the requested Normalizer2, if successful
     * @stable ICU 49
     */
    static const Normalizer2 *
    getNFKDInstance(UErrorCode &errorCode);

    /**
     * Returns a Normalizer2 instance for Unicode NFKC_Casefold normalization.
     * Same as getInstance(NULL, "nfkc_cf", UNORM2_COMPOSE, errorCode).
     * Returns an unmodifiable singleton instance. Do not delete it.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return the requested Normalizer2, if successful
     * @stable ICU 49
     */
    static const Normalizer2 *
    getNFKCCasefoldInstance(UErrorCode &errorCode);

    /**
     * Returns a Normalizer2 instance which uses the specified data file
     * (packageName/name similar to ucnv_openPackage() and ures_open()/ResourceBundle)
     * and which composes or decomposes text according to the specified mode.
     * Returns an unmodifiable singleton instance. Do not delete it.
     *
     * Use packageName=NULL for data files that are part of ICU's own data.
     * Use name="nfc" and UNORM2_COMPOSE/UNORM2_DECOMPOSE for Unicode standard NFC/NFD.
     * Use name="nfkc" and UNORM2_COMPOSE/UNORM2_DECOMPOSE for Unicode standard NFKC/NFKD.
     * Use name="nfkc_cf" and UNORM2_COMPOSE for Unicode standard NFKC_CF=NFKC_Casefold.
     *
     * @param packageName NULL for ICU built-in data, otherwise application data package name
     * @param name "nfc" or "nfkc" or "nfkc_cf" or name of custom data file
     * @param mode normalization mode (compose or decompose etc.)
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return the requested Normalizer2, if successful
     * @stable ICU 4.4
     */
    static const Normalizer2 *
    getInstance(const char *packageName,
                const char *name,
                UNormalization2Mode mode,
                UErrorCode &errorCode);

    /**
     * Returns the normalized form of the source string.
     * @param src source string
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return normalized src
     * @stable ICU 4.4
     */
    UnicodeString
    normalize(const UnicodeString &src, UErrorCode &errorCode) const {
        UnicodeString result;
        normalize(src, result, errorCode);
        return result;
    }
    /**
     * Writes the normalized form of the source string to the destination string
     * (replacing its contents) and returns the destination string.
     * The source and destination strings must be different objects.
     * @param src source string
     * @param dest destination string; its contents is replaced with normalized src
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return dest
     * @stable ICU 4.4
     */
    virtual UnicodeString &
    normalize(const UnicodeString &src,
              UnicodeString &dest,
              UErrorCode &errorCode) const = 0;

    /**
     * Normalizes a UTF-8 string and optionally records how source substrings
     * relate to changed and unchanged result substrings.
     *
     * Currently implemented completely only for "compose" modes,
     * such as for NFC, NFKC, and NFKC_Casefold
     * (UNORM2_COMPOSE and UNORM2_COMPOSE_CONTIGUOUS).
     * Otherwise currently converts to & from UTF-16 and does not support edits.
     *
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT and U_EDITS_NO_RESET.
     * @param src       Source UTF-8 string.
     * @param sink      A ByteSink to which the normalized UTF-8 result string is written.
     *                  sink.Flush() is called at the end.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be nullptr.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @stable ICU 60
     */
    virtual void
    normalizeUTF8(uint32_t options, StringPiece src, ByteSink &sink,
                  Edits *edits, UErrorCode &errorCode) const;

    /**
     * Appends the normalized form of the second string to the first string
     * (merging them at the boundary) and returns the first string.
     * The result is normalized if the first string was normalized.
     * The first and second strings must be different objects.
     * @param first string, should be normalized
     * @param second string, will be normalized
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return first
     * @stable ICU 4.4
     */
    virtual UnicodeString &
    normalizeSecondAndAppend(UnicodeString &first,
                             const UnicodeString &second,
                             UErrorCode &errorCode) const = 0;
    /**
     * Appends the second string to the first string
     * (merging them at the boundary) and returns the first string.
     * The result is normalized if both the strings were normalized.
     * The first and second strings must be different objects.
     * @param first string, should be normalized
     * @param second string, should be normalized
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return first
     * @stable ICU 4.4
     */
    virtual UnicodeString &
    append(UnicodeString &first,
           const UnicodeString &second,
           UErrorCode &errorCode) const = 0;

    /**
     * Gets the decomposition mapping of c.
     * Roughly equivalent to normalizing the String form of c
     * on a UNORM2_DECOMPOSE Normalizer2 instance, but much faster, and except that this function
     * returns FALSE and does not write a string
     * if c does not have a decomposition mapping in this instance's data.
     * This function is independent of the mode of the Normalizer2.
     * @param c code point
     * @param decomposition String object which will be set to c's
     *                      decomposition mapping, if there is one.
     * @return TRUE if c has a decomposition, otherwise FALSE
     * @stable ICU 4.6
     */
    virtual UBool
    getDecomposition(UChar32 c, UnicodeString &decomposition) const = 0;

    /**
     * Gets the raw decomposition mapping of c.
     *
     * This is similar to the getDecomposition() method but returns the
     * raw decomposition mapping as specified in UnicodeData.txt or
     * (for custom data) in the mapping files processed by the gennorm2 tool.
     * By contrast, getDecomposition() returns the processed,
     * recursively-decomposed version of this mapping.
     *
     * When used on a standard NFKC Normalizer2 instance,
     * getRawDecomposition() returns the Unicode Decomposition_Mapping (dm) property.
     *
     * When used on a standard NFC Normalizer2 instance,
     * it returns the Decomposition_Mapping only if the Decomposition_Type (dt) is Canonical (Can);
     * in this case, the result contains either one or two code points (=1..4 char16_ts).
     *
     * This function is independent of the mode of the Normalizer2.
     * The default implementation returns FALSE.
     * @param c code point
     * @param decomposition String object which will be set to c's
     *                      raw decomposition mapping, if there is one.
     * @return TRUE if c has a decomposition, otherwise FALSE
     * @stable ICU 49
     */
    virtual UBool
    getRawDecomposition(UChar32 c, UnicodeString &decomposition) const;

    /**
     * Performs pairwise composition of a & b and returns the composite if there is one.
     *
     * Returns a composite code point c only if c has a two-way mapping to a+b.
     * In standard Unicode normalization, this means that
     * c has a canonical decomposition to a+b
     * and c does not have the Full_Composition_Exclusion property.
     *
     * This function is independent of the mode of the Normalizer2.
     * The default implementation returns a negative value.
     * @param a A (normalization starter) code point.
     * @param b Another code point.
     * @return The non-negative composite code point if there is one; otherwise a negative value.
     * @stable ICU 49
     */
    virtual UChar32
    composePair(UChar32 a, UChar32 b) const;

    /**
     * Gets the combining class of c.
     * The default implementation returns 0
     * but all standard implementations return the Unicode Canonical_Combining_Class value.
     * @param c code point
     * @return c's combining class
     * @stable ICU 49
     */
    virtual uint8_t
    getCombiningClass(UChar32 c) const;

    /**
     * Tests if the string is normalized.
     * Internally, in cases where the quickCheck() method would return "maybe"
     * (which is only possible for the two COMPOSE modes) this method
     * resolves to "yes" or "no" to provide a definitive result,
     * at the cost of doing more work in those cases.
     * @param s input string
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return TRUE if s is normalized
     * @stable ICU 4.4
     */
    virtual UBool
    isNormalized(const UnicodeString &s, UErrorCode &errorCode) const = 0;
    /**
     * Tests if the UTF-8 string is normalized.
     * Internally, in cases where the quickCheck() method would return "maybe"
     * (which is only possible for the two COMPOSE modes) this method
     * resolves to "yes" or "no" to provide a definitive result,
     * at the cost of doing more work in those cases.
     *
     * This works for all normalization modes,
     * but it is currently optimized for UTF-8 only for "compose" modes,
     * such as for NFC, NFKC, and NFKC_Casefold
     * (UNORM2_COMPOSE and UNORM2_COMPOSE_CONTIGUOUS).
     * For other modes it currently converts to UTF-16 and calls isNormalized().
     *
     * @param s UTF-8 input string
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return TRUE if s is normalized
     * @stable ICU 60
     */
    virtual UBool
    isNormalizedUTF8(StringPiece s, UErrorCode &errorCode) const;


    /**
     * Tests if the string is normalized.
     * For the two COMPOSE modes, the result could be "maybe" in cases that
     * would take a little more work to resolve definitively.
     * Use spanQuickCheckYes() and normalizeSecondAndAppend() for a faster
     * combination of quick check + normalization, to avoid
     * re-checking the "yes" prefix.
     * @param s input string
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return UNormalizationCheckResult
     * @stable ICU 4.4
     */
    virtual UNormalizationCheckResult
    quickCheck(const UnicodeString &s, UErrorCode &errorCode) const = 0;

    /**
     * Returns the end of the normalized substring of the input string.
     * In other words, with <code>end=spanQuickCheckYes(s, ec);</code>
     * the substring <code>UnicodeString(s, 0, end)</code>
     * will pass the quick check with a "yes" result.
     *
     * The returned end index is usually one or more characters before the
     * "no" or "maybe" character: The end index is at a normalization boundary.
     * (See the class documentation for more about normalization boundaries.)
     *
     * When the goal is a normalized string and most input strings are expected
     * to be normalized already, then call this method,
     * and if it returns a prefix shorter than the input string,
     * copy that prefix and use normalizeSecondAndAppend() for the remainder.
     * @param s input string
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return "yes" span end index
     * @stable ICU 4.4
     */
    virtual int32_t
    spanQuickCheckYes(const UnicodeString &s, UErrorCode &errorCode) const = 0;

    /**
     * Tests if the character always has a normalization boundary before it,
     * regardless of context.
     * If true, then the character does not normalization-interact with
     * preceding characters.
     * In other words, a string containing this character can be normalized
     * by processing portions before this character and starting from this
     * character independently.
     * This is used for iterative normalization. See the class documentation for details.
     * @param c character to test
     * @return TRUE if c has a normalization boundary before it
     * @stable ICU 4.4
     */
    virtual UBool hasBoundaryBefore(UChar32 c) const = 0;

    /**
     * Tests if the character always has a normalization boundary after it,
     * regardless of context.
     * If true, then the character does not normalization-interact with
     * following characters.
     * In other words, a string containing this character can be normalized
     * by processing portions up to this character and after this
     * character independently.
     * This is used for iterative normalization. See the class documentation for details.
     * Note that this operation may be significantly slower than hasBoundaryBefore().
     * @param c character to test
     * @return TRUE if c has a normalization boundary after it
     * @stable ICU 4.4
     */
    virtual UBool hasBoundaryAfter(UChar32 c) const = 0;

    /**
     * Tests if the character is normalization-inert.
     * If true, then the character does not change, nor normalization-interact with
     * preceding or following characters.
     * In other words, a string containing this character can be normalized
     * by processing portions before this character and after this
     * character independently.
     * This is used for iterative normalization. See the class documentation for details.
     * Note that this operation may be significantly slower than hasBoundaryBefore().
     * @param c character to test
     * @return TRUE if c is normalization-inert
     * @stable ICU 4.4
     */
    virtual UBool isInert(UChar32 c) const = 0;
};

/**
 * Normalization filtered by a UnicodeSet.
 * Normalizes portions of the text contained in the filter set and leaves
 * portions not contained in the filter set unchanged.
 * Filtering is done via UnicodeSet::span(..., USET_SPAN_SIMPLE).
 * Not-in-the-filter text is treated as "is normalized" and "quick check yes".
 * This class implements all of (and only) the Normalizer2 API.
 * An instance of this class is unmodifiable/immutable but is constructed and
 * must be destructed by the owner.
 * @stable ICU 4.4
 */
class U_COMMON_API FilteredNormalizer2 : public Normalizer2 {
public:
    /**
     * Constructs a filtered normalizer wrapping any Normalizer2 instance
     * and a filter set.
     * Both are aliased and must not be modified or deleted while this object
     * is used.
     * The filter set should be frozen; otherwise the performance will suffer greatly.
     * @param n2 wrapped Normalizer2 instance
     * @param filterSet UnicodeSet which determines the characters to be normalized
     * @stable ICU 4.4
     */
    FilteredNormalizer2(const Normalizer2 &n2, const UnicodeSet &filterSet) :
            norm2(n2), set(filterSet) {}

    /**
     * Destructor.
     * @stable ICU 4.4
     */
    ~FilteredNormalizer2();

    /**
     * Writes the normalized form of the source string to the destination string
     * (replacing its contents) and returns the destination string.
     * The source and destination strings must be different objects.
     * @param src source string
     * @param dest destination string; its contents is replaced with normalized src
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return dest
     * @stable ICU 4.4
     */
    virtual UnicodeString &
    normalize(const UnicodeString &src,
              UnicodeString &dest,
              UErrorCode &errorCode) const U_OVERRIDE;

    /**
     * Normalizes a UTF-8 string and optionally records how source substrings
     * relate to changed and unchanged result substrings.
     *
     * Currently implemented completely only for "compose" modes,
     * such as for NFC, NFKC, and NFKC_Casefold
     * (UNORM2_COMPOSE and UNORM2_COMPOSE_CONTIGUOUS).
     * Otherwise currently converts to & from UTF-16 and does not support edits.
     *
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT and U_EDITS_NO_RESET.
     * @param src       Source UTF-8 string.
     * @param sink      A ByteSink to which the normalized UTF-8 result string is written.
     *                  sink.Flush() is called at the end.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be nullptr.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @stable ICU 60
     */
    virtual void
    normalizeUTF8(uint32_t options, StringPiece src, ByteSink &sink,
                  Edits *edits, UErrorCode &errorCode) const U_OVERRIDE;

    /**
     * Appends the normalized form of the second string to the first string
     * (merging them at the boundary) and returns the first string.
     * The result is normalized if the first string was normalized.
     * The first and second strings must be different objects.
     * @param first string, should be normalized
     * @param second string, will be normalized
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return first
     * @stable ICU 4.4
     */
    virtual UnicodeString &
    normalizeSecondAndAppend(UnicodeString &first,
                             const UnicodeString &second,
                             UErrorCode &errorCode) const U_OVERRIDE;
    /**
     * Appends the second string to the first string
     * (merging them at the boundary) and returns the first string.
     * The result is normalized if both the strings were normalized.
     * The first and second strings must be different objects.
     * @param first string, should be normalized
     * @param second string, should be normalized
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return first
     * @stable ICU 4.4
     */
    virtual UnicodeString &
    append(UnicodeString &first,
           const UnicodeString &second,
           UErrorCode &errorCode) const U_OVERRIDE;

    /**
     * Gets the decomposition mapping of c.
     * For details see the base class documentation.
     *
     * This function is independent of the mode of the Normalizer2.
     * @param c code point
     * @param decomposition String object which will be set to c's
     *                      decomposition mapping, if there is one.
     * @return TRUE if c has a decomposition, otherwise FALSE
     * @stable ICU 4.6
     */
    virtual UBool
    getDecomposition(UChar32 c, UnicodeString &decomposition) const U_OVERRIDE;

    /**
     * Gets the raw decomposition mapping of c.
     * For details see the base class documentation.
     *
     * This function is independent of the mode of the Normalizer2.
     * @param c code point
     * @param decomposition String object which will be set to c's
     *                      raw decomposition mapping, if there is one.
     * @return TRUE if c has a decomposition, otherwise FALSE
     * @stable ICU 49
     */
    virtual UBool
    getRawDecomposition(UChar32 c, UnicodeString &decomposition) const U_OVERRIDE;

    /**
     * Performs pairwise composition of a & b and returns the composite if there is one.
     * For details see the base class documentation.
     *
     * This function is independent of the mode of the Normalizer2.
     * @param a A (normalization starter) code point.
     * @param b Another code point.
     * @return The non-negative composite code point if there is one; otherwise a negative value.
     * @stable ICU 49
     */
    virtual UChar32
    composePair(UChar32 a, UChar32 b) const U_OVERRIDE;

    /**
     * Gets the combining class of c.
     * The default implementation returns 0
     * but all standard implementations return the Unicode Canonical_Combining_Class value.
     * @param c code point
     * @return c's combining class
     * @stable ICU 49
     */
    virtual uint8_t
    getCombiningClass(UChar32 c) const U_OVERRIDE;

    /**
     * Tests if the string is normalized.
     * For details see the Normalizer2 base class documentation.
     * @param s input string
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return TRUE if s is normalized
     * @stable ICU 4.4
     */
    virtual UBool
    isNormalized(const UnicodeString &s, UErrorCode &errorCode) const U_OVERRIDE;
    /**
     * Tests if the UTF-8 string is normalized.
     * Internally, in cases where the quickCheck() method would return "maybe"
     * (which is only possible for the two COMPOSE modes) this method
     * resolves to "yes" or "no" to provide a definitive result,
     * at the cost of doing more work in those cases.
     *
     * This works for all normalization modes,
     * but it is currently optimized for UTF-8 only for "compose" modes,
     * such as for NFC, NFKC, and NFKC_Casefold
     * (UNORM2_COMPOSE and UNORM2_COMPOSE_CONTIGUOUS).
     * For other modes it currently converts to UTF-16 and calls isNormalized().
     *
     * @param s UTF-8 input string
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return TRUE if s is normalized
     * @stable ICU 60
     */
    virtual UBool
    isNormalizedUTF8(StringPiece s, UErrorCode &errorCode) const U_OVERRIDE;
    /**
     * Tests if the string is normalized.
     * For details see the Normalizer2 base class documentation.
     * @param s input string
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return UNormalizationCheckResult
     * @stable ICU 4.4
     */
    virtual UNormalizationCheckResult
    quickCheck(const UnicodeString &s, UErrorCode &errorCode) const U_OVERRIDE;
    /**
     * Returns the end of the normalized substring of the input string.
     * For details see the Normalizer2 base class documentation.
     * @param s input string
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return "yes" span end index
     * @stable ICU 4.4
     */
    virtual int32_t
    spanQuickCheckYes(const UnicodeString &s, UErrorCode &errorCode) const U_OVERRIDE;

    /**
     * Tests if the character always has a normalization boundary before it,
     * regardless of context.
     * For details see the Normalizer2 base class documentation.
     * @param c character to test
     * @return TRUE if c has a normalization boundary before it
     * @stable ICU 4.4
     */
    virtual UBool hasBoundaryBefore(UChar32 c) const U_OVERRIDE;

    /**
     * Tests if the character always has a normalization boundary after it,
     * regardless of context.
     * For details see the Normalizer2 base class documentation.
     * @param c character to test
     * @return TRUE if c has a normalization boundary after it
     * @stable ICU 4.4
     */
    virtual UBool hasBoundaryAfter(UChar32 c) const U_OVERRIDE;

    /**
     * Tests if the character is normalization-inert.
     * For details see the Normalizer2 base class documentation.
     * @param c character to test
     * @return TRUE if c is normalization-inert
     * @stable ICU 4.4
     */
    virtual UBool isInert(UChar32 c) const U_OVERRIDE;
private:
    UnicodeString &
    normalize(const UnicodeString &src,
              UnicodeString &dest,
              USetSpanCondition spanCondition,
              UErrorCode &errorCode) const;

    void
    normalizeUTF8(uint32_t options, const char *src, int32_t length,
                  ByteSink &sink, Edits *edits,
                  USetSpanCondition spanCondition,
                  UErrorCode &errorCode) const;

    UnicodeString &
    normalizeSecondAndAppend(UnicodeString &first,
                             const UnicodeString &second,
                             UBool doNormalize,
                             UErrorCode &errorCode) const;

    const Normalizer2 &norm2;
    const UnicodeSet &set;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_NORMALIZATION

#endif /* U_SHOW_CPLUSPLUS_API */




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2002-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
*/



#if U_SHOW_CPLUSPLUS_API




/**
 * \file 
 * \brief C++ API: UnicodeSetIterator iterates over the contents of a UnicodeSet.
 */

U_NAMESPACE_BEGIN

class UnicodeSet;
class UnicodeString;

/**
 *
 * UnicodeSetIterator iterates over the contents of a UnicodeSet.  It
 * iterates over either code points or code point ranges.  After all
 * code points or ranges have been returned, it returns the
 * multicharacter strings of the UnicodeSet, if any.
 *
 * This class is not intended to be subclassed.  Consider any fields
 *  or methods declared as "protected" to be private.  The use of
 *  protected in this class is an artifact of history.
 *
 * <p>To iterate over code points and strings, use a loop like this:
 * <pre>
 * UnicodeSetIterator it(set);
 * while (it.next()) {
 *     processItem(it.getString());
 * }
 * </pre>
 * <p>Each item in the set is accessed as a string.  Set elements
 *    consisting of single code points are returned as strings containing
 *    just the one code point.
 *
 * <p>To iterate over code point ranges, instead of individual code points,
 *    use a loop like this:
 * <pre>
 * UnicodeSetIterator it(set);
 * while (it.nextRange()) {
 *   if (it.isString()) {
 *     processString(it.getString());
 *   } else {
 *     processCodepointRange(it.getCodepoint(), it.getCodepointEnd());
 *   }
 * }
 * </pre>
 * @author M. Davis
 * @stable ICU 2.4
 */
class U_COMMON_API UnicodeSetIterator : public UObject {

 protected:

    /**
     * Value of <tt>codepoint</tt> if the iterator points to a string.
     * If <tt>codepoint == IS_STRING</tt>, then examine
     * <tt>string</tt> for the current iteration result.
     * @stable ICU 2.4
     */
    enum { IS_STRING = -1 };

    /**
     * Current code point, or the special value <tt>IS_STRING</tt>, if
     * the iterator points to a string.
     * @stable ICU 2.4
     */
    UChar32 codepoint;

    /**
     * When iterating over ranges using <tt>nextRange()</tt>,
     * <tt>codepointEnd</tt> contains the inclusive end of the
     * iteration range, if <tt>codepoint != IS_STRING</tt>.  If
     * iterating over code points using <tt>next()</tt>, or if
     * <tt>codepoint == IS_STRING</tt>, then the value of
     * <tt>codepointEnd</tt> is undefined.
     * @stable ICU 2.4
     */
    UChar32 codepointEnd;

    /**
     * If <tt>codepoint == IS_STRING</tt>, then <tt>string</tt> points
     * to the current string.  If <tt>codepoint != IS_STRING</tt>, the
     * value of <tt>string</tt> is undefined.
     * @stable ICU 2.4
     */
    const UnicodeString* string;

 public:

    /**
     * Create an iterator over the given set.  The iterator is valid
     * only so long as <tt>set</tt> is valid.
     * @param set set to iterate over
     * @stable ICU 2.4
     */
    UnicodeSetIterator(const UnicodeSet& set);

    /**
     * Create an iterator over nothing.  <tt>next()</tt> and
     * <tt>nextRange()</tt> return false. This is a convenience
     * constructor allowing the target to be set later.
     * @stable ICU 2.4
     */
    UnicodeSetIterator();

    /**
     * Destructor.
     * @stable ICU 2.4
     */
    virtual ~UnicodeSetIterator();

    /**
     * Returns true if the current element is a string.  If so, the
     * caller can retrieve it with <tt>getString()</tt>.  If this
     * method returns false, the current element is a code point or
     * code point range, depending on whether <tt>next()</tt> or
     * <tt>nextRange()</tt> was called.
     * Elements of types string and codepoint can both be retrieved
     * with the function <tt>getString()</tt>.
     * Elements of type codepoint can also be retrieved with
     * <tt>getCodepoint()</tt>.
     * For ranges, <tt>getCodepoint()</tt> returns the starting codepoint
     * of the range, and <tt>getCodepointEnd()</tt> returns the end
     * of the range.
     * @stable ICU 2.4
     */
    inline UBool isString() const;

    /**
     * Returns the current code point, if <tt>isString()</tt> returned
     * false.  Otherwise returns an undefined result.
     * @stable ICU 2.4
     */
    inline UChar32 getCodepoint() const;

    /**
     * Returns the end of the current code point range, if
     * <tt>isString()</tt> returned false and <tt>nextRange()</tt> was
     * called.  Otherwise returns an undefined result.
     * @stable ICU 2.4
     */
    inline UChar32 getCodepointEnd() const;

    /**
     * Returns the current string, if <tt>isString()</tt> returned
     * true.  If the current iteration item is a code point, a UnicodeString
     * containing that single code point is returned.
     *
     * Ownership of the returned string remains with the iterator.
     * The string is guaranteed to remain valid only until the iterator is
     *   advanced to the next item, or until the iterator is deleted.
     * 
     * @stable ICU 2.4
     */
    const UnicodeString& getString();

    /**
     * Advances the iteration position to the next element in the set, 
     * which can be either a single code point or a string.  
     * If there are no more elements in the set, return false.
     *
     * <p>
     * If <tt>isString() == TRUE</tt>, the value is a
     * string, otherwise the value is a
     * single code point.  Elements of either type can be retrieved
     * with the function <tt>getString()</tt>, while elements of
     * consisting of a single code point can be retrieved with
     * <tt>getCodepoint()</tt>
     *
     * <p>The order of iteration is all code points in sorted order,
     * followed by all strings sorted order.    Do not mix
     * calls to <tt>next()</tt> and <tt>nextRange()</tt> without
     * calling <tt>reset()</tt> between them.  The results of doing so
     * are undefined.
     *
     * @return true if there was another element in the set.
     * @stable ICU 2.4
     */
    UBool next();

    /**
     * Returns the next element in the set, either a code point range
     * or a string.  If there are no more elements in the set, return
     * false.  If <tt>isString() == TRUE</tt>, the value is a
     * string and can be accessed with <tt>getString()</tt>.  Otherwise the value is a
     * range of one or more code points from <tt>getCodepoint()</tt> to
     * <tt>getCodepointeEnd()</tt> inclusive.
     *
     * <p>The order of iteration is all code points ranges in sorted
     * order, followed by all strings sorted order.  Ranges are
     * disjoint and non-contiguous.  The value returned from <tt>getString()</tt>
     * is undefined unless <tt>isString() == TRUE</tt>.  Do not mix calls to
     * <tt>next()</tt> and <tt>nextRange()</tt> without calling
     * <tt>reset()</tt> between them.  The results of doing so are
     * undefined.
     *
     * @return true if there was another element in the set.
     * @stable ICU 2.4
     */
    UBool nextRange();

    /**
     * Sets this iterator to visit the elements of the given set and
     * resets it to the start of that set.  The iterator is valid only
     * so long as <tt>set</tt> is valid.
     * @param set the set to iterate over.
     * @stable ICU 2.4
     */
    void reset(const UnicodeSet& set);

    /**
     * Resets this iterator to the start of the set.
     * @stable ICU 2.4
     */
    void reset();

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 2.4
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 2.4
     */
    virtual UClassID getDynamicClassID() const;

    // ======================= PRIVATES ===========================

 protected:

    // endElement and nextElements are really UChar32's, but we keep
    // them as signed int32_t's so we can do comparisons with
    // endElement set to -1.  Leave them as int32_t's.
    /** The set
     * @stable ICU 2.4
     */
    const UnicodeSet* set;
    /** End range
     * @stable ICU 2.4
     */
    int32_t endRange;
    /** Range
     * @stable ICU 2.4
     */
    int32_t range;
    /** End element
     * @stable ICU 2.4
     */
    int32_t endElement;
    /** Next element
     * @stable ICU 2.4
     */
    int32_t nextElement;
    //UBool abbreviated;
    /** Next string
     * @stable ICU 2.4
     */
    int32_t nextString;
    /** String count
     * @stable ICU 2.4
     */
    int32_t stringCount;

    /**
     *  Points to the string to use when the caller asks for a
     *  string and the current iteration item is a code point, not a string.
     *  @internal
     */
    UnicodeString *cpString;

    /** Copy constructor. Disallowed.
     * @stable ICU 2.4
     */
    UnicodeSetIterator(const UnicodeSetIterator&); // disallow

    /** Assignment operator. Disallowed.
     * @stable ICU 2.4
     */
    UnicodeSetIterator& operator=(const UnicodeSetIterator&); // disallow

    /** Load range
     * @stable ICU 2.4
     */
    virtual void loadRange(int32_t range);

};

inline UBool UnicodeSetIterator::isString() const {
    return codepoint == (UChar32)IS_STRING;
}

inline UChar32 UnicodeSetIterator::getCodepoint() const {
    return codepoint;
}

inline UChar32 UnicodeSetIterator::getCodepointEnd() const {
    return codepointEnd;
}


U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1998-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*
* File ustring.h
*
* Modification History:
*
*   Date        Name        Description
*   12/07/98    bertrand    Creation.
******************************************************************************
*/






/**
 * \def UBRK_TYPEDEF_UBREAK_ITERATOR
 * @internal 
 */

#ifndef UBRK_TYPEDEF_UBREAK_ITERATOR
#   define UBRK_TYPEDEF_UBREAK_ITERATOR
/** Simple declaration for u_strToTitle() to avoid including unicode/ubrk.h. @stable ICU 2.1*/
    typedef struct UBreakIterator UBreakIterator;
#endif

/**
 * \file
 * \brief C API: Unicode string handling functions
 *
 * These C API functions provide general Unicode string handling.
 *
 * Some functions are equivalent in name, signature, and behavior to the ANSI C <string.h>
 * functions. (For example, they do not check for bad arguments like NULL string pointers.)
 * In some cases, only the thread-safe variant of such a function is implemented here
 * (see u_strtok_r()).
 *
 * Other functions provide more Unicode-specific functionality like locale-specific
 * upper/lower-casing and string comparison in code point order.
 *
 * ICU uses 16-bit Unicode (UTF-16) in the form of arrays of UChar code units.
 * UTF-16 encodes each Unicode code point with either one or two UChar code units.
 * (This is the default form of Unicode, and a forward-compatible extension of the original,
 * fixed-width form that was known as UCS-2. UTF-16 superseded UCS-2 with Unicode 2.0
 * in 1996.)
 *
 * Some APIs accept a 32-bit UChar32 value for a single code point.
 *
 * ICU also handles 16-bit Unicode text with unpaired surrogates.
 * Such text is not well-formed UTF-16.
 * Code-point-related functions treat unpaired surrogates as surrogate code points,
 * i.e., as separate units.
 *
 * Although UTF-16 is a variable-width encoding form (like some legacy multi-byte encodings),
 * it is much more efficient even for random access because the code unit values
 * for single-unit characters vs. lead units vs. trail units are completely disjoint.
 * This means that it is easy to determine character (code point) boundaries from
 * random offsets in the string.
 *
 * Unicode (UTF-16) string processing is optimized for the single-unit case.
 * Although it is important to support supplementary characters
 * (which use pairs of lead/trail code units called "surrogates"),
 * their occurrence is rare. Almost all characters in modern use require only
 * a single UChar code unit (i.e., their code point values are <=0xffff).
 *
 * For more details see the User Guide Strings chapter (http://icu-project.org/userguide/strings.html).
 * For a discussion of the handling of unpaired surrogates see also
 * Jitterbug 2145 and its icu mailing list proposal on 2002-sep-18.
 */

/**
 * \defgroup ustring_ustrlen String Length
 * \ingroup ustring_strlen
 */
/*@{*/
/**
 * Determine the length of an array of UChar.
 *
 * @param s The array of UChars, NULL (U+0000) terminated.
 * @return The number of UChars in <code>chars</code>, minus the terminator.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strlen(const UChar *s);
/*@}*/

/**
 * Count Unicode code points in the length UChar code units of the string.
 * A code point may occupy either one or two UChar code units.
 * Counting code points involves reading all code units.
 *
 * This functions is basically the inverse of the U16_FWD_N() macro (see utf.h).
 *
 * @param s The input string.
 * @param length The number of UChar code units to be checked, or -1 to count all
 *               code points before the first NUL (U+0000).
 * @return The number of code points in the specified code units.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_countChar32(const UChar *s, int32_t length);

/**
 * Check if the string contains more Unicode code points than a certain number.
 * This is more efficient than counting all code points in the entire string
 * and comparing that number with a threshold.
 * This function may not need to scan the string at all if the length is known
 * (not -1 for NUL-termination) and falls within a certain range, and
 * never needs to count more than 'number+1' code points.
 * Logically equivalent to (u_countChar32(s, length)>number).
 * A Unicode code point may occupy either one or two UChar code units.
 *
 * @param s The input string.
 * @param length The length of the string, or -1 if it is NUL-terminated.
 * @param number The number of code points in the string is compared against
 *               the 'number' parameter.
 * @return Boolean value for whether the string contains more Unicode code points
 *         than 'number'. Same as (u_countChar32(s, length)>number).
 * @stable ICU 2.4
 */
U_STABLE UBool U_EXPORT2
u_strHasMoreChar32Than(const UChar *s, int32_t length, int32_t number);

/**
 * Concatenate two ustrings.  Appends a copy of <code>src</code>,
 * including the null terminator, to <code>dst</code>. The initial copied
 * character from <code>src</code> overwrites the null terminator in <code>dst</code>.
 *
 * @param dst The destination string.
 * @param src The source string.
 * @return A pointer to <code>dst</code>.
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2
u_strcat(UChar     *dst, 
    const UChar     *src);

/**
 * Concatenate two ustrings.  
 * Appends at most <code>n</code> characters from <code>src</code> to <code>dst</code>.
 * Adds a terminating NUL.
 * If src is too long, then only <code>n-1</code> characters will be copied
 * before the terminating NUL.
 * If <code>n&lt;=0</code> then dst is not modified.
 *
 * @param dst The destination string.
 * @param src The source string (can be NULL/invalid if n<=0).
 * @param n The maximum number of characters to append; no-op if <=0.
 * @return A pointer to <code>dst</code>.
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2
u_strncat(UChar     *dst, 
     const UChar     *src, 
     int32_t     n);

/**
 * Find the first occurrence of a substring in a string.
 * The substring is found at code point boundaries.
 * That means that if the substring begins with
 * a trail surrogate or ends with a lead surrogate,
 * then it is found only if these surrogates stand alone in the text.
 * Otherwise, the substring edge units would be matched against
 * halves of surrogate pairs.
 *
 * @param s The string to search (NUL-terminated).
 * @param substring The substring to find (NUL-terminated).
 * @return A pointer to the first occurrence of <code>substring</code> in <code>s</code>,
 *         or <code>s</code> itself if the <code>substring</code> is empty,
 *         or <code>NULL</code> if <code>substring</code> is not in <code>s</code>.
 * @stable ICU 2.0
 *
 * @see u_strrstr
 * @see u_strFindFirst
 * @see u_strFindLast
 */
U_STABLE UChar * U_EXPORT2
u_strstr(const UChar *s, const UChar *substring);

/**
 * Find the first occurrence of a substring in a string.
 * The substring is found at code point boundaries.
 * That means that if the substring begins with
 * a trail surrogate or ends with a lead surrogate,
 * then it is found only if these surrogates stand alone in the text.
 * Otherwise, the substring edge units would be matched against
 * halves of surrogate pairs.
 *
 * @param s The string to search.
 * @param length The length of s (number of UChars), or -1 if it is NUL-terminated.
 * @param substring The substring to find (NUL-terminated).
 * @param subLength The length of substring (number of UChars), or -1 if it is NUL-terminated.
 * @return A pointer to the first occurrence of <code>substring</code> in <code>s</code>,
 *         or <code>s</code> itself if the <code>substring</code> is empty,
 *         or <code>NULL</code> if <code>substring</code> is not in <code>s</code>.
 * @stable ICU 2.4
 *
 * @see u_strstr
 * @see u_strFindLast
 */
U_STABLE UChar * U_EXPORT2
u_strFindFirst(const UChar *s, int32_t length, const UChar *substring, int32_t subLength);

/**
 * Find the first occurrence of a BMP code point in a string.
 * A surrogate code point is found only if its match in the text is not
 * part of a surrogate pair.
 * A NUL character is found at the string terminator.
 *
 * @param s The string to search (NUL-terminated).
 * @param c The BMP code point to find.
 * @return A pointer to the first occurrence of <code>c</code> in <code>s</code>
 *         or <code>NULL</code> if <code>c</code> is not in <code>s</code>.
 * @stable ICU 2.0
 *
 * @see u_strchr32
 * @see u_memchr
 * @see u_strstr
 * @see u_strFindFirst
 */
U_STABLE UChar * U_EXPORT2
u_strchr(const UChar *s, UChar c);

/**
 * Find the first occurrence of a code point in a string.
 * A surrogate code point is found only if its match in the text is not
 * part of a surrogate pair.
 * A NUL character is found at the string terminator.
 *
 * @param s The string to search (NUL-terminated).
 * @param c The code point to find.
 * @return A pointer to the first occurrence of <code>c</code> in <code>s</code>
 *         or <code>NULL</code> if <code>c</code> is not in <code>s</code>.
 * @stable ICU 2.0
 *
 * @see u_strchr
 * @see u_memchr32
 * @see u_strstr
 * @see u_strFindFirst
 */
U_STABLE UChar * U_EXPORT2
u_strchr32(const UChar *s, UChar32 c);

/**
 * Find the last occurrence of a substring in a string.
 * The substring is found at code point boundaries.
 * That means that if the substring begins with
 * a trail surrogate or ends with a lead surrogate,
 * then it is found only if these surrogates stand alone in the text.
 * Otherwise, the substring edge units would be matched against
 * halves of surrogate pairs.
 *
 * @param s The string to search (NUL-terminated).
 * @param substring The substring to find (NUL-terminated).
 * @return A pointer to the last occurrence of <code>substring</code> in <code>s</code>,
 *         or <code>s</code> itself if the <code>substring</code> is empty,
 *         or <code>NULL</code> if <code>substring</code> is not in <code>s</code>.
 * @stable ICU 2.4
 *
 * @see u_strstr
 * @see u_strFindFirst
 * @see u_strFindLast
 */
U_STABLE UChar * U_EXPORT2
u_strrstr(const UChar *s, const UChar *substring);

/**
 * Find the last occurrence of a substring in a string.
 * The substring is found at code point boundaries.
 * That means that if the substring begins with
 * a trail surrogate or ends with a lead surrogate,
 * then it is found only if these surrogates stand alone in the text.
 * Otherwise, the substring edge units would be matched against
 * halves of surrogate pairs.
 *
 * @param s The string to search.
 * @param length The length of s (number of UChars), or -1 if it is NUL-terminated.
 * @param substring The substring to find (NUL-terminated).
 * @param subLength The length of substring (number of UChars), or -1 if it is NUL-terminated.
 * @return A pointer to the last occurrence of <code>substring</code> in <code>s</code>,
 *         or <code>s</code> itself if the <code>substring</code> is empty,
 *         or <code>NULL</code> if <code>substring</code> is not in <code>s</code>.
 * @stable ICU 2.4
 *
 * @see u_strstr
 * @see u_strFindLast
 */
U_STABLE UChar * U_EXPORT2
u_strFindLast(const UChar *s, int32_t length, const UChar *substring, int32_t subLength);

/**
 * Find the last occurrence of a BMP code point in a string.
 * A surrogate code point is found only if its match in the text is not
 * part of a surrogate pair.
 * A NUL character is found at the string terminator.
 *
 * @param s The string to search (NUL-terminated).
 * @param c The BMP code point to find.
 * @return A pointer to the last occurrence of <code>c</code> in <code>s</code>
 *         or <code>NULL</code> if <code>c</code> is not in <code>s</code>.
 * @stable ICU 2.4
 *
 * @see u_strrchr32
 * @see u_memrchr
 * @see u_strrstr
 * @see u_strFindLast
 */
U_STABLE UChar * U_EXPORT2
u_strrchr(const UChar *s, UChar c);

/**
 * Find the last occurrence of a code point in a string.
 * A surrogate code point is found only if its match in the text is not
 * part of a surrogate pair.
 * A NUL character is found at the string terminator.
 *
 * @param s The string to search (NUL-terminated).
 * @param c The code point to find.
 * @return A pointer to the last occurrence of <code>c</code> in <code>s</code>
 *         or <code>NULL</code> if <code>c</code> is not in <code>s</code>.
 * @stable ICU 2.4
 *
 * @see u_strrchr
 * @see u_memchr32
 * @see u_strrstr
 * @see u_strFindLast
 */
U_STABLE UChar * U_EXPORT2
u_strrchr32(const UChar *s, UChar32 c);

/**
 * Locates the first occurrence in the string <code>string</code> of any of the characters
 * in the string <code>matchSet</code>.
 * Works just like C's strpbrk but with Unicode.
 *
 * @param string The string in which to search, NUL-terminated.
 * @param matchSet A NUL-terminated string defining a set of code points
 *                 for which to search in the text string.
 * @return A pointer to the  character in <code>string</code> that matches one of the
 *         characters in <code>matchSet</code>, or NULL if no such character is found.
 * @stable ICU 2.0
 */
U_STABLE UChar * U_EXPORT2
u_strpbrk(const UChar *string, const UChar *matchSet);

/**
 * Returns the number of consecutive characters in <code>string</code>,
 * beginning with the first, that do not occur somewhere in <code>matchSet</code>.
 * Works just like C's strcspn but with Unicode.
 *
 * @param string The string in which to search, NUL-terminated.
 * @param matchSet A NUL-terminated string defining a set of code points
 *                 for which to search in the text string.
 * @return The number of initial characters in <code>string</code> that do not
 *         occur in <code>matchSet</code>.
 * @see u_strspn
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strcspn(const UChar *string, const UChar *matchSet);

/**
 * Returns the number of consecutive characters in <code>string</code>,
 * beginning with the first, that occur somewhere in <code>matchSet</code>.
 * Works just like C's strspn but with Unicode.
 *
 * @param string The string in which to search, NUL-terminated.
 * @param matchSet A NUL-terminated string defining a set of code points
 *                 for which to search in the text string.
 * @return The number of initial characters in <code>string</code> that do
 *         occur in <code>matchSet</code>.
 * @see u_strcspn
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strspn(const UChar *string, const UChar *matchSet);

/**
 * The string tokenizer API allows an application to break a string into
 * tokens. Unlike strtok(), the saveState (the current pointer within the
 * original string) is maintained in saveState. In the first call, the
 * argument src is a pointer to the string. In subsequent calls to
 * return successive tokens of that string, src must be specified as
 * NULL. The value saveState is set by this function to maintain the
 * function's position within the string, and on each subsequent call
 * you must give this argument the same variable. This function does
 * handle surrogate pairs. This function is similar to the strtok_r()
 * the POSIX Threads Extension (1003.1c-1995) version.
 *
 * @param src String containing token(s). This string will be modified.
 *            After the first call to u_strtok_r(), this argument must
 *            be NULL to get to the next token.
 * @param delim Set of delimiter characters (Unicode code points).
 * @param saveState The current pointer within the original string,
 *              which is set by this function. The saveState
 *              parameter should the address of a local variable of type
 *              UChar *. (i.e. defined "UChar *myLocalSaveState" and use
 *              &myLocalSaveState for this parameter).
 * @return A pointer to the next token found in src, or NULL
 *         when there are no more tokens.
 * @stable ICU 2.0
 */
U_STABLE UChar * U_EXPORT2
u_strtok_r(UChar    *src, 
     const UChar    *delim,
           UChar   **saveState);

/**
 * Compare two Unicode strings for bitwise equality (code unit order).
 *
 * @param s1 A string to compare.
 * @param s2 A string to compare.
 * @return 0 if <code>s1</code> and <code>s2</code> are bitwise equal; a negative
 * value if <code>s1</code> is bitwise less than <code>s2,</code>; a positive
 * value if <code>s1</code> is bitwise greater than <code>s2</code>.
 * @stable ICU 2.0
 */
U_STABLE int32_t  U_EXPORT2
u_strcmp(const UChar     *s1, 
         const UChar     *s2);

/**
 * Compare two Unicode strings in code point order.
 * See u_strCompare for details.
 *
 * @param s1 A string to compare.
 * @param s2 A string to compare.
 * @return a negative/zero/positive integer corresponding to whether
 * the first string is less than/equal to/greater than the second one
 * in code point order
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strcmpCodePointOrder(const UChar *s1, const UChar *s2);

/**
 * Compare two Unicode strings (binary order).
 *
 * The comparison can be done in code unit order or in code point order.
 * They differ only in UTF-16 when
 * comparing supplementary code points (U+10000..U+10ffff)
 * to BMP code points near the end of the BMP (i.e., U+e000..U+ffff).
 * In code unit order, high BMP code points sort after supplementary code points
 * because they are stored as pairs of surrogates which are at U+d800..U+dfff.
 *
 * This functions works with strings of different explicitly specified lengths
 * unlike the ANSI C-like u_strcmp() and u_memcmp() etc.
 * NUL-terminated strings are possible with length arguments of -1.
 *
 * @param s1 First source string.
 * @param length1 Length of first source string, or -1 if NUL-terminated.
 *
 * @param s2 Second source string.
 * @param length2 Length of second source string, or -1 if NUL-terminated.
 *
 * @param codePointOrder Choose between code unit order (FALSE)
 *                       and code point order (TRUE).
 *
 * @return <0 or 0 or >0 as usual for string comparisons
 *
 * @stable ICU 2.2
 */
U_STABLE int32_t U_EXPORT2
u_strCompare(const UChar *s1, int32_t length1,
             const UChar *s2, int32_t length2,
             UBool codePointOrder);

/**
 * Compare two Unicode strings (binary order)
 * as presented by UCharIterator objects.
 * Works otherwise just like u_strCompare().
 *
 * Both iterators are reset to their start positions.
 * When the function returns, it is undefined where the iterators
 * have stopped.
 *
 * @param iter1 First source string iterator.
 * @param iter2 Second source string iterator.
 * @param codePointOrder Choose between code unit order (FALSE)
 *                       and code point order (TRUE).
 *
 * @return <0 or 0 or >0 as usual for string comparisons
 *
 * @see u_strCompare
 *
 * @stable ICU 2.6
 */
U_STABLE int32_t U_EXPORT2
u_strCompareIter(UCharIterator *iter1, UCharIterator *iter2, UBool codePointOrder);

/**
 * Compare two strings case-insensitively using full case folding.
 * This is equivalent to
 *   u_strCompare(u_strFoldCase(s1, options),
 *                u_strFoldCase(s2, options),
 *                (options&U_COMPARE_CODE_POINT_ORDER)!=0).
 *
 * The comparison can be done in UTF-16 code unit order or in code point order.
 * They differ only when comparing supplementary code points (U+10000..U+10ffff)
 * to BMP code points near the end of the BMP (i.e., U+e000..U+ffff).
 * In code unit order, high BMP code points sort after supplementary code points
 * because they are stored as pairs of surrogates which are at U+d800..U+dfff.
 *
 * This functions works with strings of different explicitly specified lengths
 * unlike the ANSI C-like u_strcmp() and u_memcmp() etc.
 * NUL-terminated strings are possible with length arguments of -1.
 *
 * @param s1 First source string.
 * @param length1 Length of first source string, or -1 if NUL-terminated.
 *
 * @param s2 Second source string.
 * @param length2 Length of second source string, or -1 if NUL-terminated.
 *
 * @param options A bit set of options:
 *   - U_FOLD_CASE_DEFAULT or 0 is used for default options:
 *     Comparison in code unit order with default case folding.
 *
 *   - U_COMPARE_CODE_POINT_ORDER
 *     Set to choose code point order instead of code unit order
 *     (see u_strCompare for details).
 *
 *   - U_FOLD_CASE_EXCLUDE_SPECIAL_I
 *
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 *
 * @return <0 or 0 or >0 as usual for string comparisons
 *
 * @stable ICU 2.2
 */
U_STABLE int32_t U_EXPORT2
u_strCaseCompare(const UChar *s1, int32_t length1,
                 const UChar *s2, int32_t length2,
                 uint32_t options,
                 UErrorCode *pErrorCode);

/**
 * Compare two ustrings for bitwise equality. 
 * Compares at most <code>n</code> characters.
 *
 * @param ucs1 A string to compare (can be NULL/invalid if n<=0).
 * @param ucs2 A string to compare (can be NULL/invalid if n<=0).
 * @param n The maximum number of characters to compare; always returns 0 if n<=0.
 * @return 0 if <code>s1</code> and <code>s2</code> are bitwise equal; a negative
 * value if <code>s1</code> is bitwise less than <code>s2</code>; a positive
 * value if <code>s1</code> is bitwise greater than <code>s2</code>.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strncmp(const UChar     *ucs1, 
     const UChar     *ucs2, 
     int32_t     n);

/**
 * Compare two Unicode strings in code point order.
 * This is different in UTF-16 from u_strncmp() if supplementary characters are present.
 * For details, see u_strCompare().
 *
 * @param s1 A string to compare.
 * @param s2 A string to compare.
 * @param n The maximum number of characters to compare.
 * @return a negative/zero/positive integer corresponding to whether
 * the first string is less than/equal to/greater than the second one
 * in code point order
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strncmpCodePointOrder(const UChar *s1, const UChar *s2, int32_t n);

/**
 * Compare two strings case-insensitively using full case folding.
 * This is equivalent to u_strcmp(u_strFoldCase(s1, options), u_strFoldCase(s2, options)).
 *
 * @param s1 A string to compare.
 * @param s2 A string to compare.
 * @param options A bit set of options:
 *   - U_FOLD_CASE_DEFAULT or 0 is used for default options:
 *     Comparison in code unit order with default case folding.
 *
 *   - U_COMPARE_CODE_POINT_ORDER
 *     Set to choose code point order instead of code unit order
 *     (see u_strCompare for details).
 *
 *   - U_FOLD_CASE_EXCLUDE_SPECIAL_I
 *
 * @return A negative, zero, or positive integer indicating the comparison result.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strcasecmp(const UChar *s1, const UChar *s2, uint32_t options);

/**
 * Compare two strings case-insensitively using full case folding.
 * This is equivalent to u_strcmp(u_strFoldCase(s1, at most n, options),
 * u_strFoldCase(s2, at most n, options)).
 *
 * @param s1 A string to compare.
 * @param s2 A string to compare.
 * @param n The maximum number of characters each string to case-fold and then compare.
 * @param options A bit set of options:
 *   - U_FOLD_CASE_DEFAULT or 0 is used for default options:
 *     Comparison in code unit order with default case folding.
 *
 *   - U_COMPARE_CODE_POINT_ORDER
 *     Set to choose code point order instead of code unit order
 *     (see u_strCompare for details).
 *
 *   - U_FOLD_CASE_EXCLUDE_SPECIAL_I
 *
 * @return A negative, zero, or positive integer indicating the comparison result.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strncasecmp(const UChar *s1, const UChar *s2, int32_t n, uint32_t options);

/**
 * Compare two strings case-insensitively using full case folding.
 * This is equivalent to u_strcmp(u_strFoldCase(s1, n, options),
 * u_strFoldCase(s2, n, options)).
 *
 * @param s1 A string to compare.
 * @param s2 A string to compare.
 * @param length The number of characters in each string to case-fold and then compare.
 * @param options A bit set of options:
 *   - U_FOLD_CASE_DEFAULT or 0 is used for default options:
 *     Comparison in code unit order with default case folding.
 *
 *   - U_COMPARE_CODE_POINT_ORDER
 *     Set to choose code point order instead of code unit order
 *     (see u_strCompare for details).
 *
 *   - U_FOLD_CASE_EXCLUDE_SPECIAL_I
 *
 * @return A negative, zero, or positive integer indicating the comparison result.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_memcasecmp(const UChar *s1, const UChar *s2, int32_t length, uint32_t options);

/**
 * Copy a ustring. Adds a null terminator.
 *
 * @param dst The destination string.
 * @param src The source string.
 * @return A pointer to <code>dst</code>.
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2
u_strcpy(UChar     *dst, 
    const UChar     *src);

/**
 * Copy a ustring.
 * Copies at most <code>n</code> characters.  The result will be null terminated
 * if the length of <code>src</code> is less than <code>n</code>.
 *
 * @param dst The destination string.
 * @param src The source string (can be NULL/invalid if n<=0).
 * @param n The maximum number of characters to copy; no-op if <=0.
 * @return A pointer to <code>dst</code>.
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2
u_strncpy(UChar     *dst, 
     const UChar     *src, 
     int32_t     n);

#if !UCONFIG_NO_CONVERSION

/**
 * Copy a byte string encoded in the default codepage to a ustring.
 * Adds a null terminator.
 * Performs a host byte to UChar conversion
 *
 * @param dst The destination string.
 * @param src The source string.
 * @return A pointer to <code>dst</code>.
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2 u_uastrcpy(UChar *dst,
               const char *src );

/**
 * Copy a byte string encoded in the default codepage to a ustring.
 * Copies at most <code>n</code> characters.  The result will be null terminated
 * if the length of <code>src</code> is less than <code>n</code>.
 * Performs a host byte to UChar conversion
 *
 * @param dst The destination string.
 * @param src The source string.
 * @param n The maximum number of characters to copy.
 * @return A pointer to <code>dst</code>.
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2 u_uastrncpy(UChar *dst,
            const char *src,
            int32_t n);

/**
 * Copy ustring to a byte string encoded in the default codepage.
 * Adds a null terminator.
 * Performs a UChar to host byte conversion
 *
 * @param dst The destination string.
 * @param src The source string.
 * @return A pointer to <code>dst</code>.
 * @stable ICU 2.0
 */
U_STABLE char* U_EXPORT2 u_austrcpy(char *dst,
            const UChar *src );

/**
 * Copy ustring to a byte string encoded in the default codepage.
 * Copies at most <code>n</code> characters.  The result will be null terminated
 * if the length of <code>src</code> is less than <code>n</code>.
 * Performs a UChar to host byte conversion
 *
 * @param dst The destination string.
 * @param src The source string.
 * @param n The maximum number of characters to copy.
 * @return A pointer to <code>dst</code>.
 * @stable ICU 2.0
 */
U_STABLE char* U_EXPORT2 u_austrncpy(char *dst,
            const UChar *src,
            int32_t n );

#endif

/**
 * Synonym for memcpy(), but with UChars only.
 * @param dest The destination string
 * @param src The source string (can be NULL/invalid if count<=0)
 * @param count The number of characters to copy; no-op if <=0
 * @return A pointer to <code>dest</code>
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2
u_memcpy(UChar *dest, const UChar *src, int32_t count);

/**
 * Synonym for memmove(), but with UChars only.
 * @param dest The destination string
 * @param src The source string (can be NULL/invalid if count<=0)
 * @param count The number of characters to move; no-op if <=0
 * @return A pointer to <code>dest</code>
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2
u_memmove(UChar *dest, const UChar *src, int32_t count);

/**
 * Initialize <code>count</code> characters of <code>dest</code> to <code>c</code>.
 *
 * @param dest The destination string.
 * @param c The character to initialize the string.
 * @param count The maximum number of characters to set.
 * @return A pointer to <code>dest</code>.
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2
u_memset(UChar *dest, UChar c, int32_t count);

/**
 * Compare the first <code>count</code> UChars of each buffer.
 *
 * @param buf1 The first string to compare.
 * @param buf2 The second string to compare.
 * @param count The maximum number of UChars to compare.
 * @return When buf1 < buf2, a negative number is returned.
 *      When buf1 == buf2, 0 is returned.
 *      When buf1 > buf2, a positive number is returned.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_memcmp(const UChar *buf1, const UChar *buf2, int32_t count);

/**
 * Compare two Unicode strings in code point order.
 * This is different in UTF-16 from u_memcmp() if supplementary characters are present.
 * For details, see u_strCompare().
 *
 * @param s1 A string to compare.
 * @param s2 A string to compare.
 * @param count The maximum number of characters to compare.
 * @return a negative/zero/positive integer corresponding to whether
 * the first string is less than/equal to/greater than the second one
 * in code point order
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_memcmpCodePointOrder(const UChar *s1, const UChar *s2, int32_t count);

/**
 * Find the first occurrence of a BMP code point in a string.
 * A surrogate code point is found only if its match in the text is not
 * part of a surrogate pair.
 * A NUL character is found at the string terminator.
 *
 * @param s The string to search (contains <code>count</code> UChars).
 * @param c The BMP code point to find.
 * @param count The length of the string.
 * @return A pointer to the first occurrence of <code>c</code> in <code>s</code>
 *         or <code>NULL</code> if <code>c</code> is not in <code>s</code>.
 * @stable ICU 2.0
 *
 * @see u_strchr
 * @see u_memchr32
 * @see u_strFindFirst
 */
U_STABLE UChar* U_EXPORT2
u_memchr(const UChar *s, UChar c, int32_t count);

/**
 * Find the first occurrence of a code point in a string.
 * A surrogate code point is found only if its match in the text is not
 * part of a surrogate pair.
 * A NUL character is found at the string terminator.
 *
 * @param s The string to search (contains <code>count</code> UChars).
 * @param c The code point to find.
 * @param count The length of the string.
 * @return A pointer to the first occurrence of <code>c</code> in <code>s</code>
 *         or <code>NULL</code> if <code>c</code> is not in <code>s</code>.
 * @stable ICU 2.0
 *
 * @see u_strchr32
 * @see u_memchr
 * @see u_strFindFirst
 */
U_STABLE UChar* U_EXPORT2
u_memchr32(const UChar *s, UChar32 c, int32_t count);

/**
 * Find the last occurrence of a BMP code point in a string.
 * A surrogate code point is found only if its match in the text is not
 * part of a surrogate pair.
 * A NUL character is found at the string terminator.
 *
 * @param s The string to search (contains <code>count</code> UChars).
 * @param c The BMP code point to find.
 * @param count The length of the string.
 * @return A pointer to the last occurrence of <code>c</code> in <code>s</code>
 *         or <code>NULL</code> if <code>c</code> is not in <code>s</code>.
 * @stable ICU 2.4
 *
 * @see u_strrchr
 * @see u_memrchr32
 * @see u_strFindLast
 */
U_STABLE UChar* U_EXPORT2
u_memrchr(const UChar *s, UChar c, int32_t count);

/**
 * Find the last occurrence of a code point in a string.
 * A surrogate code point is found only if its match in the text is not
 * part of a surrogate pair.
 * A NUL character is found at the string terminator.
 *
 * @param s The string to search (contains <code>count</code> UChars).
 * @param c The code point to find.
 * @param count The length of the string.
 * @return A pointer to the last occurrence of <code>c</code> in <code>s</code>
 *         or <code>NULL</code> if <code>c</code> is not in <code>s</code>.
 * @stable ICU 2.4
 *
 * @see u_strrchr32
 * @see u_memrchr
 * @see u_strFindLast
 */
U_STABLE UChar* U_EXPORT2
u_memrchr32(const UChar *s, UChar32 c, int32_t count);

/**
 * Unicode String literals in C.
 * We need one macro to declare a variable for the string
 * and to statically preinitialize it if possible,
 * and a second macro to dynamically initialize such a string variable if necessary.
 *
 * The macros are defined for maximum performance.
 * They work only for strings that contain "invariant characters", i.e.,
 * only latin letters, digits, and some punctuation.
 * See utypes.h for details.
 *
 * A pair of macros for a single string must be used with the same
 * parameters.
 * The string parameter must be a C string literal.
 * The length of the string, not including the terminating
 * `NUL`, must be specified as a constant.
 * The U_STRING_DECL macro should be invoked exactly once for one
 * such string variable before it is used.
 *
 * Usage:
 *
 *     U_STRING_DECL(ustringVar1, "Quick-Fox 2", 11);
 *     U_STRING_DECL(ustringVar2, "jumps 5%", 8);
 *     static UBool didInit=FALSE;
 *
 *     int32_t function() {
 *         if(!didInit) {
 *             U_STRING_INIT(ustringVar1, "Quick-Fox 2", 11);
 *             U_STRING_INIT(ustringVar2, "jumps 5%", 8);
 *             didInit=TRUE;
 *         }
 *         return u_strcmp(ustringVar1, ustringVar2);
 *     }
 * 
 * Note that the macros will NOT consistently work if their argument is another #`define`.
 * The following will not work on all platforms, don't use it.
 * 
 *     #define GLUCK "Mr. Gluck"
 *     U_STRING_DECL(var, GLUCK, 9)
 *     U_STRING_INIT(var, GLUCK, 9)
 *
 * Instead, use the string literal "Mr. Gluck"  as the argument to both macro
 * calls.
 *
 *
 * @stable ICU 2.0
 */
#if defined(U_DECLARE_UTF16)
#   define U_STRING_DECL(var, cs, length) static const UChar *var=(const UChar *)U_DECLARE_UTF16(cs)
    /**@stable ICU 2.0 */
#   define U_STRING_INIT(var, cs, length)
#elif U_SIZEOF_WCHAR_T==U_SIZEOF_UCHAR && (U_CHARSET_FAMILY==U_ASCII_FAMILY || (U_SIZEOF_UCHAR == 2 && defined(U_WCHAR_IS_UTF16)))
#   define U_STRING_DECL(var, cs, length) static const UChar var[(length)+1]=L ## cs
    /**@stable ICU 2.0 */
#   define U_STRING_INIT(var, cs, length)
#elif U_SIZEOF_UCHAR==1 && U_CHARSET_FAMILY==U_ASCII_FAMILY
#   define U_STRING_DECL(var, cs, length) static const UChar var[(length)+1]=cs
    /**@stable ICU 2.0 */
#   define U_STRING_INIT(var, cs, length)
#else
#   define U_STRING_DECL(var, cs, length) static UChar var[(length)+1]
    /**@stable ICU 2.0 */
#   define U_STRING_INIT(var, cs, length) u_charsToUChars(cs, var, length+1)
#endif

/**
 * Unescape a string of characters and write the resulting
 * Unicode characters to the destination buffer.  The following escape
 * sequences are recognized:
 *
 * \\uhhhh       4 hex digits; h in [0-9A-Fa-f]
 * \\Uhhhhhhhh   8 hex digits
 * \\xhh         1-2 hex digits
 * \\x{h...}     1-8 hex digits
 * \\ooo         1-3 octal digits; o in [0-7]
 * \\cX          control-X; X is masked with 0x1F
 *
 * as well as the standard ANSI C escapes:
 *
 * \\a => U+0007, \\b => U+0008, \\t => U+0009, \\n => U+000A,
 * \\v => U+000B, \\f => U+000C, \\r => U+000D, \\e => U+001B,
 * \\&quot; => U+0022, \\' => U+0027, \\? => U+003F, \\\\ => U+005C
 *
 * Anything else following a backslash is generically escaped.  For
 * example, "[a\\-z]" returns "[a-z]".
 *
 * If an escape sequence is ill-formed, this method returns an empty
 * string.  An example of an ill-formed sequence is "\\u" followed by
 * fewer than 4 hex digits.
 *
 * The above characters are recognized in the compiler's codepage,
 * that is, they are coded as 'u', '\\', etc.  Characters that are
 * not parts of escape sequences are converted using u_charsToUChars().
 *
 * This function is similar to UnicodeString::unescape() but not
 * identical to it.  The latter takes a source UnicodeString, so it
 * does escape recognition but no conversion.
 *
 * @param src a zero-terminated string of invariant characters
 * @param dest pointer to buffer to receive converted and unescaped
 * text and, if there is room, a zero terminator.  May be NULL for
 * preflighting, in which case no UChars will be written, but the
 * return value will still be valid.  On error, an empty string is
 * stored here (if possible).
 * @param destCapacity the number of UChars that may be written at
 * dest.  Ignored if dest == NULL.
 * @return the length of unescaped string.
 * @see u_unescapeAt
 * @see UnicodeString#unescape()
 * @see UnicodeString#unescapeAt()
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_unescape(const char *src,
           UChar *dest, int32_t destCapacity);

U_CDECL_BEGIN
/**
 * Callback function for u_unescapeAt() that returns a character of
 * the source text given an offset and a context pointer.  The context
 * pointer will be whatever is passed into u_unescapeAt().
 *
 * @param offset pointer to the offset that will be passed to u_unescapeAt().
 * @param context an opaque pointer passed directly into u_unescapeAt()
 * @return the character represented by the escape sequence at
 * offset
 * @see u_unescapeAt
 * @stable ICU 2.0
 */
typedef UChar (U_CALLCONV *UNESCAPE_CHAR_AT)(int32_t offset, void *context);
U_CDECL_END

/**
 * Unescape a single sequence. The character at offset-1 is assumed
 * (without checking) to be a backslash.  This method takes a callback
 * pointer to a function that returns the UChar at a given offset.  By
 * varying this callback, ICU functions are able to unescape char*
 * strings, UnicodeString objects, and UFILE pointers.
 *
 * If offset is out of range, or if the escape sequence is ill-formed,
 * (UChar32)0xFFFFFFFF is returned.  See documentation of u_unescape()
 * for a list of recognized sequences.
 *
 * @param charAt callback function that returns a UChar of the source
 * text given an offset and a context pointer.
 * @param offset pointer to the offset that will be passed to charAt.
 * The offset value will be updated upon return to point after the
 * last parsed character of the escape sequence.  On error the offset
 * is unchanged.
 * @param length the number of characters in the source text.  The
 * last character of the source text is considered to be at offset
 * length-1.
 * @param context an opaque pointer passed directly into charAt.
 * @return the character represented by the escape sequence at
 * offset, or (UChar32)0xFFFFFFFF on error.
 * @see u_unescape()
 * @see UnicodeString#unescape()
 * @see UnicodeString#unescapeAt()
 * @stable ICU 2.0
 */
U_STABLE UChar32 U_EXPORT2
u_unescapeAt(UNESCAPE_CHAR_AT charAt,
             int32_t *offset,
             int32_t length,
             void *context);

/**
 * Uppercase the characters in a string.
 * Casing is locale-dependent and context-sensitive.
 * The result may be longer or shorter than the original.
 * The source string and the destination buffer are allowed to overlap.
 *
 * @param dest      A buffer for the result string. The result will be zero-terminated if
 *                  the buffer is large enough.
 * @param destCapacity The size of the buffer (number of UChars). If it is 0, then
 *                  dest may be NULL and the function will only return the length of the result
 *                  without writing any of the result string.
 * @param src       The original string
 * @param srcLength The length of the original string. If -1, then src must be zero-terminated.
 * @param locale    The locale to consider, or "" for the root locale or NULL for the default locale.
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return The length of the result string. It may be greater than destCapacity. In that case,
 *         only some of the result was written to the destination buffer.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strToUpper(UChar *dest, int32_t destCapacity,
             const UChar *src, int32_t srcLength,
             const char *locale,
             UErrorCode *pErrorCode);

/**
 * Lowercase the characters in a string.
 * Casing is locale-dependent and context-sensitive.
 * The result may be longer or shorter than the original.
 * The source string and the destination buffer are allowed to overlap.
 *
 * @param dest      A buffer for the result string. The result will be zero-terminated if
 *                  the buffer is large enough.
 * @param destCapacity The size of the buffer (number of UChars). If it is 0, then
 *                  dest may be NULL and the function will only return the length of the result
 *                  without writing any of the result string.
 * @param src       The original string
 * @param srcLength The length of the original string. If -1, then src must be zero-terminated.
 * @param locale    The locale to consider, or "" for the root locale or NULL for the default locale.
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return The length of the result string. It may be greater than destCapacity. In that case,
 *         only some of the result was written to the destination buffer.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strToLower(UChar *dest, int32_t destCapacity,
             const UChar *src, int32_t srcLength,
             const char *locale,
             UErrorCode *pErrorCode);

#if !UCONFIG_NO_BREAK_ITERATION

/**
 * Titlecase a string.
 * Casing is locale-dependent and context-sensitive.
 * Titlecasing uses a break iterator to find the first characters of words
 * that are to be titlecased. It titlecases those characters and lowercases
 * all others.
 *
 * The titlecase break iterator can be provided to customize for arbitrary
 * styles, using rules and dictionaries beyond the standard iterators.
 * It may be more efficient to always provide an iterator to avoid
 * opening and closing one for each string.
 * The standard titlecase iterator for the root locale implements the
 * algorithm of Unicode TR 21.
 *
 * This function uses only the setText(), first() and next() methods of the
 * provided break iterator.
 *
 * The result may be longer or shorter than the original.
 * The source string and the destination buffer are allowed to overlap.
 *
 * @param dest      A buffer for the result string. The result will be zero-terminated if
 *                  the buffer is large enough.
 * @param destCapacity The size of the buffer (number of UChars). If it is 0, then
 *                  dest may be NULL and the function will only return the length of the result
 *                  without writing any of the result string.
 * @param src       The original string
 * @param srcLength The length of the original string. If -1, then src must be zero-terminated.
 * @param titleIter A break iterator to find the first characters of words
 *                  that are to be titlecased.
 *                  If none is provided (NULL), then a standard titlecase
 *                  break iterator is opened.
 * @param locale    The locale to consider, or "" for the root locale or NULL for the default locale.
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return The length of the result string. It may be greater than destCapacity. In that case,
 *         only some of the result was written to the destination buffer.
 * @stable ICU 2.1
 */
U_STABLE int32_t U_EXPORT2
u_strToTitle(UChar *dest, int32_t destCapacity,
             const UChar *src, int32_t srcLength,
             UBreakIterator *titleIter,
             const char *locale,
             UErrorCode *pErrorCode);

#endif

/**
 * Case-folds the characters in a string.
 *
 * Case-folding is locale-independent and not context-sensitive,
 * but there is an option for whether to include or exclude mappings for dotted I
 * and dotless i that are marked with 'T' in CaseFolding.txt.
 *
 * The result may be longer or shorter than the original.
 * The source string and the destination buffer are allowed to overlap.
 *
 * @param dest      A buffer for the result string. The result will be zero-terminated if
 *                  the buffer is large enough.
 * @param destCapacity The size of the buffer (number of UChars). If it is 0, then
 *                  dest may be NULL and the function will only return the length of the result
 *                  without writing any of the result string.
 * @param src       The original string
 * @param srcLength The length of the original string. If -1, then src must be zero-terminated.
 * @param options   Either U_FOLD_CASE_DEFAULT or U_FOLD_CASE_EXCLUDE_SPECIAL_I
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return The length of the result string. It may be greater than destCapacity. In that case,
 *         only some of the result was written to the destination buffer.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_strFoldCase(UChar *dest, int32_t destCapacity,
              const UChar *src, int32_t srcLength,
              uint32_t options,
              UErrorCode *pErrorCode);

#if defined(U_WCHAR_IS_UTF16) || defined(U_WCHAR_IS_UTF32) || !UCONFIG_NO_CONVERSION
/**
 * Convert a UTF-16 string to a wchar_t string.
 * If it is known at compile time that wchar_t strings are in UTF-16 or UTF-32, then
 * this function simply calls the fast, dedicated function for that.
 * Otherwise, two conversions UTF-16 -> default charset -> wchar_t* are performed.
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of wchar_t's). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param pErrorCode    Must be a valid pointer to an error code value,
 *                      which must not indicate a failure before the function call.
 * @return The pointer to destination buffer.
 * @stable ICU 2.0
 */
U_STABLE wchar_t* U_EXPORT2
u_strToWCS(wchar_t *dest, 
           int32_t destCapacity,
           int32_t *pDestLength,
           const UChar *src, 
           int32_t srcLength,
           UErrorCode *pErrorCode);
/**
 * Convert a wchar_t string to UTF-16.
 * If it is known at compile time that wchar_t strings are in UTF-16 or UTF-32, then
 * this function simply calls the fast, dedicated function for that.
 * Otherwise, two conversions wchar_t* -> default charset -> UTF-16 are performed.
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of UChars). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param pErrorCode    Must be a valid pointer to an error code value,
 *                      which must not indicate a failure before the function call.
 * @return The pointer to destination buffer.
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2
u_strFromWCS(UChar   *dest,
             int32_t destCapacity, 
             int32_t *pDestLength,
             const wchar_t *src,
             int32_t srcLength,
             UErrorCode *pErrorCode);
#endif /* defined(U_WCHAR_IS_UTF16) || defined(U_WCHAR_IS_UTF32) || !UCONFIG_NO_CONVERSION */

/**
 * Convert a UTF-16 string to UTF-8.
 * If the input string is not well-formed, then the U_INVALID_CHAR_FOUND error code is set.
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of chars). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param pErrorCode    Must be a valid pointer to an error code value,
 *                      which must not indicate a failure before the function call.
 * @return The pointer to destination buffer.
 * @stable ICU 2.0
 * @see u_strToUTF8WithSub
 * @see u_strFromUTF8
 */
U_STABLE char* U_EXPORT2 
u_strToUTF8(char *dest,           
            int32_t destCapacity,
            int32_t *pDestLength,
            const UChar *src, 
            int32_t srcLength,
            UErrorCode *pErrorCode);

/**
 * Convert a UTF-8 string to UTF-16.
 * If the input string is not well-formed, then the U_INVALID_CHAR_FOUND error code is set.
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of UChars). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param pErrorCode    Must be a valid pointer to an error code value,
 *                      which must not indicate a failure before the function call.
 * @return The pointer to destination buffer.
 * @stable ICU 2.0
 * @see u_strFromUTF8WithSub
 * @see u_strFromUTF8Lenient
 */
U_STABLE UChar* U_EXPORT2
u_strFromUTF8(UChar *dest,             
              int32_t destCapacity,
              int32_t *pDestLength,
              const char *src, 
              int32_t srcLength,
              UErrorCode *pErrorCode);

/**
 * Convert a UTF-16 string to UTF-8.
 *
 * Same as u_strToUTF8() except for the additional subchar which is output for
 * illegal input sequences, instead of stopping with the U_INVALID_CHAR_FOUND error code.
 * With subchar==U_SENTINEL, this function behaves exactly like u_strToUTF8().
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of chars). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param subchar       The substitution character to use in place of an illegal input sequence,
 *                      or U_SENTINEL if the function is to return with U_INVALID_CHAR_FOUND instead.
 *                      A substitution character can be any valid Unicode code point (up to U+10FFFF)
 *                      except for surrogate code points (U+D800..U+DFFF).
 *                      The recommended value is U+FFFD "REPLACEMENT CHARACTER".
 * @param pNumSubstitutions Output parameter receiving the number of substitutions if subchar>=0.
 *                      Set to 0 if no substitutions occur or subchar<0.
 *                      pNumSubstitutions can be NULL.
 * @param pErrorCode    Pointer to a standard ICU error code. Its input value must
 *                      pass the U_SUCCESS() test, or else the function returns
 *                      immediately. Check for U_FAILURE() on output or use with
 *                      function chaining. (See User Guide for details.)
 * @return The pointer to destination buffer.
 * @see u_strToUTF8
 * @see u_strFromUTF8WithSub
 * @stable ICU 3.6
 */
U_STABLE char* U_EXPORT2
u_strToUTF8WithSub(char *dest,
            int32_t destCapacity,
            int32_t *pDestLength,
            const UChar *src,
            int32_t srcLength,
            UChar32 subchar, int32_t *pNumSubstitutions,
            UErrorCode *pErrorCode);

/**
 * Convert a UTF-8 string to UTF-16.
 *
 * Same as u_strFromUTF8() except for the additional subchar which is output for
 * illegal input sequences, instead of stopping with the U_INVALID_CHAR_FOUND error code.
 * With subchar==U_SENTINEL, this function behaves exactly like u_strFromUTF8().
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of UChars). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param subchar       The substitution character to use in place of an illegal input sequence,
 *                      or U_SENTINEL if the function is to return with U_INVALID_CHAR_FOUND instead.
 *                      A substitution character can be any valid Unicode code point (up to U+10FFFF)
 *                      except for surrogate code points (U+D800..U+DFFF).
 *                      The recommended value is U+FFFD "REPLACEMENT CHARACTER".
 * @param pNumSubstitutions Output parameter receiving the number of substitutions if subchar>=0.
 *                      Set to 0 if no substitutions occur or subchar<0.
 *                      pNumSubstitutions can be NULL.
 * @param pErrorCode    Pointer to a standard ICU error code. Its input value must
 *                      pass the U_SUCCESS() test, or else the function returns
 *                      immediately. Check for U_FAILURE() on output or use with
 *                      function chaining. (See User Guide for details.)
 * @return The pointer to destination buffer.
 * @see u_strFromUTF8
 * @see u_strFromUTF8Lenient
 * @see u_strToUTF8WithSub
 * @stable ICU 3.6
 */
U_STABLE UChar* U_EXPORT2
u_strFromUTF8WithSub(UChar *dest,
              int32_t destCapacity,
              int32_t *pDestLength,
              const char *src,
              int32_t srcLength,
              UChar32 subchar, int32_t *pNumSubstitutions,
              UErrorCode *pErrorCode);

/**
 * Convert a UTF-8 string to UTF-16.
 *
 * Same as u_strFromUTF8() except that this function is designed to be very fast,
 * which it achieves by being lenient about malformed UTF-8 sequences.
 * This function is intended for use in environments where UTF-8 text is
 * expected to be well-formed.
 *
 * Its semantics are:
 * - Well-formed UTF-8 text is correctly converted to well-formed UTF-16 text.
 * - The function will not read beyond the input string, nor write beyond
 *   the destCapacity.
 * - Malformed UTF-8 results in "garbage" 16-bit Unicode strings which may not
 *   be well-formed UTF-16.
 *   The function will resynchronize to valid code point boundaries
 *   within a small number of code points after an illegal sequence.
 * - Non-shortest forms are not detected and will result in "spoofing" output.
 *
 * For further performance improvement, if srcLength is given (>=0),
 * then it must be destCapacity>=srcLength.
 *
 * There is no inverse u_strToUTF8Lenient() function because there is practically
 * no performance gain from not checking that a UTF-16 string is well-formed.
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of UChars). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 *                      Unlike for other ICU functions, if srcLength>=0 then it
 *                      must be destCapacity>=srcLength.
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 *                      Unlike for other ICU functions, if srcLength>=0 but
 *                      destCapacity<srcLength, then *pDestLength will be set to srcLength
 *                      (and U_BUFFER_OVERFLOW_ERROR will be set)
 *                      regardless of the actual result length.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param pErrorCode    Pointer to a standard ICU error code. Its input value must
 *                      pass the U_SUCCESS() test, or else the function returns
 *                      immediately. Check for U_FAILURE() on output or use with
 *                      function chaining. (See User Guide for details.)
 * @return The pointer to destination buffer.
 * @see u_strFromUTF8
 * @see u_strFromUTF8WithSub
 * @see u_strToUTF8WithSub
 * @stable ICU 3.6
 */
U_STABLE UChar * U_EXPORT2
u_strFromUTF8Lenient(UChar *dest,
                     int32_t destCapacity,
                     int32_t *pDestLength,
                     const char *src,
                     int32_t srcLength,
                     UErrorCode *pErrorCode);

/**
 * Convert a UTF-16 string to UTF-32.
 * If the input string is not well-formed, then the U_INVALID_CHAR_FOUND error code is set.
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of UChar32s). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param pErrorCode    Must be a valid pointer to an error code value,
 *                      which must not indicate a failure before the function call.
 * @return The pointer to destination buffer.
 * @see u_strToUTF32WithSub
 * @see u_strFromUTF32
 * @stable ICU 2.0
 */
U_STABLE UChar32* U_EXPORT2 
u_strToUTF32(UChar32 *dest, 
             int32_t  destCapacity,
             int32_t  *pDestLength,
             const UChar *src, 
             int32_t  srcLength,
             UErrorCode *pErrorCode);

/**
 * Convert a UTF-32 string to UTF-16.
 * If the input string is not well-formed, then the U_INVALID_CHAR_FOUND error code is set.
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of UChars). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param pErrorCode    Must be a valid pointer to an error code value,
 *                      which must not indicate a failure before the function call.
 * @return The pointer to destination buffer.
 * @see u_strFromUTF32WithSub
 * @see u_strToUTF32
 * @stable ICU 2.0
 */
U_STABLE UChar* U_EXPORT2 
u_strFromUTF32(UChar   *dest,
               int32_t destCapacity, 
               int32_t *pDestLength,
               const UChar32 *src,
               int32_t srcLength,
               UErrorCode *pErrorCode);

/**
 * Convert a UTF-16 string to UTF-32.
 *
 * Same as u_strToUTF32() except for the additional subchar which is output for
 * illegal input sequences, instead of stopping with the U_INVALID_CHAR_FOUND error code.
 * With subchar==U_SENTINEL, this function behaves exactly like u_strToUTF32().
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of UChar32s). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If
 *                      pDestLength!=NULL then *pDestLength is always set to the
 *                      number of output units corresponding to the transformation of
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param subchar       The substitution character to use in place of an illegal input sequence,
 *                      or U_SENTINEL if the function is to return with U_INVALID_CHAR_FOUND instead.
 *                      A substitution character can be any valid Unicode code point (up to U+10FFFF)
 *                      except for surrogate code points (U+D800..U+DFFF).
 *                      The recommended value is U+FFFD "REPLACEMENT CHARACTER".
 * @param pNumSubstitutions Output parameter receiving the number of substitutions if subchar>=0.
 *                      Set to 0 if no substitutions occur or subchar<0.
 *                      pNumSubstitutions can be NULL.
 * @param pErrorCode    Pointer to a standard ICU error code. Its input value must
 *                      pass the U_SUCCESS() test, or else the function returns
 *                      immediately. Check for U_FAILURE() on output or use with
 *                      function chaining. (See User Guide for details.)
 * @return The pointer to destination buffer.
 * @see u_strToUTF32
 * @see u_strFromUTF32WithSub
 * @stable ICU 4.2
 */
U_STABLE UChar32* U_EXPORT2
u_strToUTF32WithSub(UChar32 *dest,
             int32_t destCapacity,
             int32_t *pDestLength,
             const UChar *src,
             int32_t srcLength,
             UChar32 subchar, int32_t *pNumSubstitutions,
             UErrorCode *pErrorCode);

/**
 * Convert a UTF-32 string to UTF-16.
 *
 * Same as u_strFromUTF32() except for the additional subchar which is output for
 * illegal input sequences, instead of stopping with the U_INVALID_CHAR_FOUND error code.
 * With subchar==U_SENTINEL, this function behaves exactly like u_strFromUTF32().
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of UChars). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If
 *                      pDestLength!=NULL then *pDestLength is always set to the
 *                      number of output units corresponding to the transformation of
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param subchar       The substitution character to use in place of an illegal input sequence,
 *                      or U_SENTINEL if the function is to return with U_INVALID_CHAR_FOUND instead.
 *                      A substitution character can be any valid Unicode code point (up to U+10FFFF)
 *                      except for surrogate code points (U+D800..U+DFFF).
 *                      The recommended value is U+FFFD "REPLACEMENT CHARACTER".
 * @param pNumSubstitutions Output parameter receiving the number of substitutions if subchar>=0.
 *                      Set to 0 if no substitutions occur or subchar<0.
 *                      pNumSubstitutions can be NULL.
 * @param pErrorCode    Pointer to a standard ICU error code. Its input value must
 *                      pass the U_SUCCESS() test, or else the function returns
 *                      immediately. Check for U_FAILURE() on output or use with
 *                      function chaining. (See User Guide for details.)
 * @return The pointer to destination buffer.
 * @see u_strFromUTF32
 * @see u_strToUTF32WithSub
 * @stable ICU 4.2
 */
U_STABLE UChar* U_EXPORT2
u_strFromUTF32WithSub(UChar *dest,
               int32_t destCapacity,
               int32_t *pDestLength,
               const UChar32 *src,
               int32_t srcLength,
               UChar32 subchar, int32_t *pNumSubstitutions,
               UErrorCode *pErrorCode);

/**
 * Convert a 16-bit Unicode string to Java Modified UTF-8.
 * See http://java.sun.com/javase/6/docs/api/java/io/DataInput.html#modified-utf-8
 *
 * This function behaves according to the documentation for Java DataOutput.writeUTF()
 * except that it does not encode the output length in the destination buffer
 * and does not have an output length restriction.
 * See http://java.sun.com/javase/6/docs/api/java/io/DataOutput.html#writeUTF(java.lang.String)
 *
 * The input string need not be well-formed UTF-16.
 * (Therefore there is no subchar parameter.)
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of chars). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param pErrorCode    Pointer to a standard ICU error code. Its input value must
 *                      pass the U_SUCCESS() test, or else the function returns
 *                      immediately. Check for U_FAILURE() on output or use with
 *                      function chaining. (See User Guide for details.)
 * @return The pointer to destination buffer.
 * @stable ICU 4.4
 * @see u_strToUTF8WithSub
 * @see u_strFromJavaModifiedUTF8WithSub
 */
U_STABLE char* U_EXPORT2 
u_strToJavaModifiedUTF8(
        char *dest,
        int32_t destCapacity,
        int32_t *pDestLength,
        const UChar *src, 
        int32_t srcLength,
        UErrorCode *pErrorCode);

/**
 * Convert a Java Modified UTF-8 string to a 16-bit Unicode string.
 * If the input string is not well-formed and no substitution char is specified, 
 * then the U_INVALID_CHAR_FOUND error code is set.
 *
 * This function behaves according to the documentation for Java DataInput.readUTF()
 * except that it takes a length parameter rather than
 * interpreting the first two input bytes as the length.
 * See http://java.sun.com/javase/6/docs/api/java/io/DataInput.html#readUTF()
 *
 * The output string may not be well-formed UTF-16.
 *
 * @param dest          A buffer for the result string. The result will be zero-terminated if
 *                      the buffer is large enough.
 * @param destCapacity  The size of the buffer (number of UChars). If it is 0, then
 *                      dest may be NULL and the function will only return the length of the 
 *                      result without writing any of the result string (pre-flighting).
 * @param pDestLength   A pointer to receive the number of units written to the destination. If 
 *                      pDestLength!=NULL then *pDestLength is always set to the 
 *                      number of output units corresponding to the transformation of 
 *                      all the input units, even in case of a buffer overflow.
 * @param src           The original source string
 * @param srcLength     The length of the original string. If -1, then src must be zero-terminated.
 * @param subchar       The substitution character to use in place of an illegal input sequence,
 *                      or U_SENTINEL if the function is to return with U_INVALID_CHAR_FOUND instead.
 *                      A substitution character can be any valid Unicode code point (up to U+10FFFF)
 *                      except for surrogate code points (U+D800..U+DFFF).
 *                      The recommended value is U+FFFD "REPLACEMENT CHARACTER".
 * @param pNumSubstitutions Output parameter receiving the number of substitutions if subchar>=0.
 *                      Set to 0 if no substitutions occur or subchar<0.
 *                      pNumSubstitutions can be NULL.
 * @param pErrorCode    Pointer to a standard ICU error code. Its input value must
 *                      pass the U_SUCCESS() test, or else the function returns
 *                      immediately. Check for U_FAILURE() on output or use with
 *                      function chaining. (See User Guide for details.)
 * @return The pointer to destination buffer.
 * @see u_strFromUTF8WithSub
 * @see u_strFromUTF8Lenient
 * @see u_strToJavaModifiedUTF8
 * @stable ICU 4.4
 */
U_STABLE UChar* U_EXPORT2
u_strFromJavaModifiedUTF8WithSub(
        UChar *dest,
        int32_t destCapacity,
        int32_t *pDestLength,
        const char *src,
        int32_t srcLength,
        UChar32 subchar, int32_t *pNumSubstitutions,
        UErrorCode *pErrorCode);




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 1997-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*   Date        Name        Description
*   03/28/00    aliu        Creation.
******************************************************************************
*/







U_NAMESPACE_BEGIN

/**
 * Hashtable is a thin C++ wrapper around UHashtable, a general-purpose void*
 * hashtable implemented in C.  Hashtable is designed to be idiomatic and
 * easy-to-use in C++.
 *
 * Hashtable is an INTERNAL CLASS.
 */
class U_COMMON_API Hashtable : public UMemory {
    UHashtable* hash;
    UHashtable hashObj;

    inline void init(UHashFunction *keyHash, UKeyComparator *keyComp, UValueComparator *valueComp, UErrorCode& status);

    inline void initSize(UHashFunction *keyHash, UKeyComparator *keyComp, UValueComparator *valueComp, int32_t size, UErrorCode& status);

public:
    /**
     * Construct a hashtable
     * @param ignoreKeyCase If true, keys are case insensitive.
     * @param status Error code
    */
    inline Hashtable(UBool ignoreKeyCase, UErrorCode& status);

    /**
     * Construct a hashtable
     * @param ignoreKeyCase If true, keys are case insensitive.
     * @param size initial size allocation
     * @param status Error code
    */
    inline Hashtable(UBool ignoreKeyCase, int32_t size, UErrorCode& status);

    /**
     * Construct a hashtable
     * @param keyComp Comparator for comparing the keys
     * @param valueComp Comparator for comparing the values
     * @param status Error code
    */
    inline Hashtable(UKeyComparator *keyComp, UValueComparator *valueComp, UErrorCode& status);

    /**
     * Construct a hashtable
     * @param status Error code
    */
    inline Hashtable(UErrorCode& status);

    /**
     * Construct a hashtable, _disregarding any error_.  Use this constructor
     * with caution.
     */
    inline Hashtable();

    /**
     * Non-virtual destructor; make this virtual if Hashtable is subclassed
     * in the future.
     */
    inline ~Hashtable();

    inline UObjectDeleter *setValueDeleter(UObjectDeleter *fn);

    inline int32_t count() const;

    inline void* put(const UnicodeString& key, void* value, UErrorCode& status);

    inline int32_t puti(const UnicodeString& key, int32_t value, UErrorCode& status);

    inline void* get(const UnicodeString& key) const;

    inline int32_t geti(const UnicodeString& key) const;

    inline void* remove(const UnicodeString& key);

    inline int32_t removei(const UnicodeString& key);

    inline void removeAll(void);

    inline const UHashElement* find(const UnicodeString& key) const;

    /**
     * @param pos - must be UHASH_FIRST on first call, and untouched afterwards.
     * @see uhash_nextElement
     */
    inline const UHashElement* nextElement(int32_t& pos) const;

    inline UKeyComparator* setKeyComparator(UKeyComparator*keyComp);

    inline UValueComparator* setValueComparator(UValueComparator* valueComp);

    inline UBool equals(const Hashtable& that) const;
private:
    Hashtable(const Hashtable &other); // forbid copying of this class
    Hashtable &operator=(const Hashtable &other); // forbid copying of this class
};

/*********************************************************************
 * Implementation
 ********************************************************************/

inline void Hashtable::init(UHashFunction *keyHash, UKeyComparator *keyComp,
                            UValueComparator *valueComp, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    uhash_init(&hashObj, keyHash, keyComp, valueComp, &status);
    if (U_SUCCESS(status)) {
        hash = &hashObj;
        uhash_setKeyDeleter(hash, uprv_deleteUObject);
    }
}

inline void Hashtable::initSize(UHashFunction *keyHash, UKeyComparator *keyComp,
                                UValueComparator *valueComp, int32_t size, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    uhash_initSize(&hashObj, keyHash, keyComp, valueComp, size, &status);
    if (U_SUCCESS(status)) {
        hash = &hashObj;
        uhash_setKeyDeleter(hash, uprv_deleteUObject);
    }
}

inline Hashtable::Hashtable(UKeyComparator *keyComp, UValueComparator *valueComp,
                 UErrorCode& status) : hash(0) {
    init( uhash_hashUnicodeString, keyComp, valueComp, status);
}

inline Hashtable::Hashtable(UBool ignoreKeyCase, UErrorCode& status)
 : hash(0)
{
    init(ignoreKeyCase ? uhash_hashCaselessUnicodeString
                        : uhash_hashUnicodeString,
            ignoreKeyCase ? uhash_compareCaselessUnicodeString
                        : uhash_compareUnicodeString,
            NULL,
            status);
}

inline Hashtable::Hashtable(UBool ignoreKeyCase, int32_t size, UErrorCode& status)
 : hash(0)
{
    initSize(ignoreKeyCase ? uhash_hashCaselessUnicodeString
                        : uhash_hashUnicodeString,
            ignoreKeyCase ? uhash_compareCaselessUnicodeString
                        : uhash_compareUnicodeString,
            NULL, size,
            status);
}

inline Hashtable::Hashtable(UErrorCode& status)
 : hash(0)
{
    init(uhash_hashUnicodeString, uhash_compareUnicodeString, NULL, status);
}

inline Hashtable::Hashtable()
 : hash(0)
{
    UErrorCode status = U_ZERO_ERROR;
    init(uhash_hashUnicodeString, uhash_compareUnicodeString, NULL, status);
}

inline Hashtable::~Hashtable() {
    if (hash != NULL) {
        uhash_close(hash);
    }
}

inline UObjectDeleter *Hashtable::setValueDeleter(UObjectDeleter *fn) {
    return uhash_setValueDeleter(hash, fn);
}

inline int32_t Hashtable::count() const {
    return uhash_count(hash);
}

inline void* Hashtable::put(const UnicodeString& key, void* value, UErrorCode& status) {
    return uhash_put(hash, new UnicodeString(key), value, &status);
}

inline int32_t Hashtable::puti(const UnicodeString& key, int32_t value, UErrorCode& status) {
    return uhash_puti(hash, new UnicodeString(key), value, &status);
}

inline void* Hashtable::get(const UnicodeString& key) const {
    return uhash_get(hash, &key);
}

inline int32_t Hashtable::geti(const UnicodeString& key) const {
    return uhash_geti(hash, &key);
}

inline void* Hashtable::remove(const UnicodeString& key) {
    return uhash_remove(hash, &key);
}

inline int32_t Hashtable::removei(const UnicodeString& key) {
    return uhash_removei(hash, &key);
}

inline const UHashElement* Hashtable::find(const UnicodeString& key) const {
    return uhash_find(hash, &key);
}

inline const UHashElement* Hashtable::nextElement(int32_t& pos) const {
    return uhash_nextElement(hash, &pos);
}

inline void Hashtable::removeAll(void) {
    uhash_removeAll(hash);
}

inline UKeyComparator* Hashtable::setKeyComparator(UKeyComparator*keyComp){
    return uhash_setKeyComparator(hash, keyComp);
}

inline UValueComparator* Hashtable::setValueComparator(UValueComparator* valueComp){
    return uhash_setValueComparator(hash, valueComp);
}

inline UBool Hashtable::equals(const Hashtable& that)const{
   return uhash_equals(hash, that.hash);
}
U_NAMESPACE_END


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2009-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  normalizer2impl.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2009nov22
*   created by: Markus W. Scherer
*/




#if !UCONFIG_NO_NORMALIZATION


// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// ucptrie.h (modified from utrie2.h)
// created: 2017dec29 Markus W. Scherer








U_CDECL_BEGIN

/**
 * \file
 *
 * This file defines an immutable Unicode code point trie.
 *
 * @see UCPTrie
 * @see UMutableCPTrie
 */

#ifndef U_IN_DOXYGEN
/** @internal */
typedef union UCPTrieData {
    /** @internal */
    const void *ptr0;
    /** @internal */
    const uint16_t *ptr16;
    /** @internal */
    const uint32_t *ptr32;
    /** @internal */
    const uint8_t *ptr8;
} UCPTrieData;
#endif

/**
 * Immutable Unicode code point trie structure.
 * Fast, reasonably compact, map from Unicode code points (U+0000..U+10FFFF) to integer values.
 * For details see http://site.icu-project.org/design/struct/utrie
 *
 * Do not access UCPTrie fields directly; use public functions and macros.
 * Functions are easy to use: They support all trie types and value widths.
 *
 * When performance is really important, macros provide faster access.
 * Most macros are specific to either "fast" or "small" tries, see UCPTrieType.
 * There are "fast" macros for special optimized use cases.
 *
 * The macros will return bogus values, or may crash, if used on the wrong type or value width.
 *
 * @see UMutableCPTrie
 * @stable ICU 63
 */
struct UCPTrie {
#ifndef U_IN_DOXYGEN
    /** @internal */
    const uint16_t *index;
    /** @internal */
    UCPTrieData data;

    /** @internal */
    int32_t indexLength;
    /** @internal */
    int32_t dataLength;
    /** Start of the last range which ends at U+10FFFF. @internal */
    UChar32 highStart;
    /** highStart>>12 @internal */
    uint16_t shifted12HighStart;

    /** @internal */
    int8_t type;  // UCPTrieType
    /** @internal */
    int8_t valueWidth;  // UCPTrieValueWidth

    /** padding/reserved @internal */
    uint32_t reserved32;
    /** padding/reserved @internal */
    uint16_t reserved16;

    /**
     * Internal index-3 null block offset.
     * Set to an impossibly high value (e.g., 0xffff) if there is no dedicated index-3 null block.
     * @internal
     */
    uint16_t index3NullOffset;
    /**
     * Internal data null block offset, not shifted.
     * Set to an impossibly high value (e.g., 0xfffff) if there is no dedicated data null block.
     * @internal
     */
    int32_t dataNullOffset;
    /** @internal */
    uint32_t nullValue;

#ifdef UCPTRIE_DEBUG
    /** @internal */
    const char *name;
#endif
#endif
};
#ifndef U_IN_DOXYGEN
typedef struct UCPTrie UCPTrie;
#endif

/**
 * Selectors for the type of a UCPTrie.
 * Different trade-offs for size vs. speed.
 *
 * @see umutablecptrie_buildImmutable
 * @see ucptrie_openFromBinary
 * @see ucptrie_getType
 * @stable ICU 63
 */
enum UCPTrieType {
    /**
     * For ucptrie_openFromBinary() to accept any type.
     * ucptrie_getType() will return the actual type.
     * @stable ICU 63
     */
    UCPTRIE_TYPE_ANY = -1,
    /**
     * Fast/simple/larger BMP data structure. Use functions and "fast" macros.
     * @stable ICU 63
     */
    UCPTRIE_TYPE_FAST,
    /**
     * Small/slower BMP data structure. Use functions and "small" macros.
     * @stable ICU 63
     */
    UCPTRIE_TYPE_SMALL
};
#ifndef U_IN_DOXYGEN
typedef enum UCPTrieType UCPTrieType;
#endif

/**
 * Selectors for the number of bits in a UCPTrie data value.
 *
 * @see umutablecptrie_buildImmutable
 * @see ucptrie_openFromBinary
 * @see ucptrie_getValueWidth
 * @stable ICU 63
 */
enum UCPTrieValueWidth {
    /**
     * For ucptrie_openFromBinary() to accept any data value width.
     * ucptrie_getValueWidth() will return the actual data value width.
     * @stable ICU 63
     */
    UCPTRIE_VALUE_BITS_ANY = -1,
    /**
     * The trie stores 16 bits per data value.
     * It returns them as unsigned values 0..0xffff=65535.
     * @stable ICU 63
     */
    UCPTRIE_VALUE_BITS_16,
    /**
     * The trie stores 32 bits per data value.
     * @stable ICU 63
     */
    UCPTRIE_VALUE_BITS_32,
    /**
     * The trie stores 8 bits per data value.
     * It returns them as unsigned values 0..0xff=255.
     * @stable ICU 63
     */
    UCPTRIE_VALUE_BITS_8
};
#ifndef U_IN_DOXYGEN
typedef enum UCPTrieValueWidth UCPTrieValueWidth;
#endif

/**
 * Opens a trie from its binary form, stored in 32-bit-aligned memory.
 * Inverse of ucptrie_toBinary().
 *
 * The memory must remain valid and unchanged as long as the trie is used.
 * You must ucptrie_close() the trie once you are done using it.
 *
 * @param type selects the trie type; results in an
 *             U_INVALID_FORMAT_ERROR if it does not match the binary data;
 *             use UCPTRIE_TYPE_ANY to accept any type
 * @param valueWidth selects the number of bits in a data value; results in an
 *                  U_INVALID_FORMAT_ERROR if it does not match the binary data;
 *                  use UCPTRIE_VALUE_BITS_ANY to accept any data value width
 * @param data a pointer to 32-bit-aligned memory containing the binary data of a UCPTrie
 * @param length the number of bytes available at data;
 *               can be more than necessary
 * @param pActualLength receives the actual number of bytes at data taken up by the trie data;
 *                      can be NULL
 * @param pErrorCode an in/out ICU UErrorCode
 * @return the trie
 *
 * @see umutablecptrie_open
 * @see umutablecptrie_buildImmutable
 * @see ucptrie_toBinary
 * @stable ICU 63
 */
U_CAPI UCPTrie * U_EXPORT2
ucptrie_openFromBinary(UCPTrieType type, UCPTrieValueWidth valueWidth,
                       const void *data, int32_t length, int32_t *pActualLength,
                       UErrorCode *pErrorCode);

/**
 * Closes a trie and releases associated memory.
 *
 * @param trie the trie
 * @stable ICU 63
 */
U_CAPI void U_EXPORT2
ucptrie_close(UCPTrie *trie);

/**
 * Returns the trie type.
 *
 * @param trie the trie
 * @return the trie type
 * @see ucptrie_openFromBinary
 * @see UCPTRIE_TYPE_ANY
 * @stable ICU 63
 */
U_CAPI UCPTrieType U_EXPORT2
ucptrie_getType(const UCPTrie *trie);

/**
 * Returns the number of bits in a trie data value.
 *
 * @param trie the trie
 * @return the number of bits in a trie data value
 * @see ucptrie_openFromBinary
 * @see UCPTRIE_VALUE_BITS_ANY
 * @stable ICU 63
 */
U_CAPI UCPTrieValueWidth U_EXPORT2
ucptrie_getValueWidth(const UCPTrie *trie);

/**
 * Returns the value for a code point as stored in the trie, with range checking.
 * Returns the trie error value if c is not in the range 0..U+10FFFF.
 *
 * Easier to use than UCPTRIE_FAST_GET() and similar macros but slower.
 * Easier to use because, unlike the macros, this function works on all UCPTrie
 * objects, for all types and value widths.
 *
 * @param trie the trie
 * @param c the code point
 * @return the trie value,
 *         or the trie error value if the code point is not in the range 0..U+10FFFF
 * @stable ICU 63
 */
U_CAPI uint32_t U_EXPORT2
ucptrie_get(const UCPTrie *trie, UChar32 c);

/**
 * Returns the last code point such that all those from start to there have the same value.
 * Can be used to efficiently iterate over all same-value ranges in a trie.
 * (This is normally faster than iterating over code points and get()ting each value,
 * but much slower than a data structure that stores ranges directly.)
 *
 * If the UCPMapValueFilter function pointer is not NULL, then
 * the value to be delivered is passed through that function, and the return value is the end
 * of the range where all values are modified to the same actual value.
 * The value is unchanged if that function pointer is NULL.
 *
 * Example:
 * \code
 * UChar32 start = 0, end;
 * uint32_t value;
 * while ((end = ucptrie_getRange(trie, start, UCPMAP_RANGE_NORMAL, 0,
 *                                NULL, NULL, &value)) >= 0) {
 *     // Work with the range start..end and its value.
 *     start = end + 1;
 * }
 * \endcode
 *
 * @param trie the trie
 * @param start range start
 * @param option defines whether surrogates are treated normally,
 *               or as having the surrogateValue; usually UCPMAP_RANGE_NORMAL
 * @param surrogateValue value for surrogates; ignored if option==UCPMAP_RANGE_NORMAL
 * @param filter a pointer to a function that may modify the trie data value,
 *     or NULL if the values from the trie are to be used unmodified
 * @param context an opaque pointer that is passed on to the filter function
 * @param pValue if not NULL, receives the value that every code point start..end has;
 *     may have been modified by filter(context, trie value)
 *     if that function pointer is not NULL
 * @return the range end code point, or -1 if start is not a valid code point
 * @stable ICU 63
 */
U_CAPI UChar32 U_EXPORT2
ucptrie_getRange(const UCPTrie *trie, UChar32 start,
                 UCPMapRangeOption option, uint32_t surrogateValue,
                 UCPMapValueFilter *filter, const void *context, uint32_t *pValue);

/**
 * Writes a memory-mappable form of the trie into 32-bit aligned memory.
 * Inverse of ucptrie_openFromBinary().
 *
 * @param trie the trie
 * @param data a pointer to 32-bit-aligned memory to be filled with the trie data;
 *             can be NULL if capacity==0
 * @param capacity the number of bytes available at data, or 0 for pure preflighting
 * @param pErrorCode an in/out ICU UErrorCode;
 *                   U_BUFFER_OVERFLOW_ERROR if the capacity is too small
 * @return the number of bytes written or (if buffer overflow) needed for the trie
 *
 * @see ucptrie_openFromBinary()
 * @stable ICU 63
 */
U_CAPI int32_t U_EXPORT2
ucptrie_toBinary(const UCPTrie *trie, void *data, int32_t capacity, UErrorCode *pErrorCode);

/**
 * Macro parameter value for a trie with 16-bit data values.
 * Use the name of this macro as a "dataAccess" parameter in other macros.
 * Do not use this macro in any other way.
 *
 * @see UCPTRIE_VALUE_BITS_16
 * @stable ICU 63
 */
#define UCPTRIE_16(trie, i) ((trie)->data.ptr16[i])

/**
 * Macro parameter value for a trie with 32-bit data values.
 * Use the name of this macro as a "dataAccess" parameter in other macros.
 * Do not use this macro in any other way.
 *
 * @see UCPTRIE_VALUE_BITS_32
 * @stable ICU 63
 */
#define UCPTRIE_32(trie, i) ((trie)->data.ptr32[i])

/**
 * Macro parameter value for a trie with 8-bit data values.
 * Use the name of this macro as a "dataAccess" parameter in other macros.
 * Do not use this macro in any other way.
 *
 * @see UCPTRIE_VALUE_BITS_8
 * @stable ICU 63
 */
#define UCPTRIE_8(trie, i) ((trie)->data.ptr8[i])

/**
 * Returns a trie value for a code point, with range checking.
 * Returns the trie error value if c is not in the range 0..U+10FFFF.
 *
 * @param trie (const UCPTrie *, in) the trie; must have type UCPTRIE_TYPE_FAST
 * @param dataAccess UCPTRIE_16, UCPTRIE_32, or UCPTRIE_8 according to the trie’s value width
 * @param c (UChar32, in) the input code point
 * @return The code point's trie value.
 * @stable ICU 63
 */
#define UCPTRIE_FAST_GET(trie, dataAccess, c) dataAccess(trie, _UCPTRIE_CP_INDEX(trie, 0xffff, c))

/**
 * Returns a 16-bit trie value for a code point, with range checking.
 * Returns the trie error value if c is not in the range U+0000..U+10FFFF.
 *
 * @param trie (const UCPTrie *, in) the trie; must have type UCPTRIE_TYPE_SMALL
 * @param dataAccess UCPTRIE_16, UCPTRIE_32, or UCPTRIE_8 according to the trie’s value width
 * @param c (UChar32, in) the input code point
 * @return The code point's trie value.
 * @stable ICU 63
 */
#define UCPTRIE_SMALL_GET(trie, dataAccess, c) \
    dataAccess(trie, _UCPTRIE_CP_INDEX(trie, UCPTRIE_SMALL_MAX, c))

/**
 * UTF-16: Reads the next code point (UChar32 c, out), post-increments src,
 * and gets a value from the trie.
 * Sets the trie error value if c is an unpaired surrogate.
 *
 * @param trie (const UCPTrie *, in) the trie; must have type UCPTRIE_TYPE_FAST
 * @param dataAccess UCPTRIE_16, UCPTRIE_32, or UCPTRIE_8 according to the trie’s value width
 * @param src (const UChar *, in/out) the source text pointer
 * @param limit (const UChar *, in) the limit pointer for the text, or NULL if NUL-terminated
 * @param c (UChar32, out) variable for the code point
 * @param result (out) variable for the trie lookup result
 * @stable ICU 63
 */
#define UCPTRIE_FAST_U16_NEXT(trie, dataAccess, src, limit, c, result) UPRV_BLOCK_MACRO_BEGIN { \
    (c) = *(src)++; \
    int32_t __index; \
    if (!U16_IS_SURROGATE(c)) { \
        __index = _UCPTRIE_FAST_INDEX(trie, c); \
    } else { \
        uint16_t __c2; \
        if (U16_IS_SURROGATE_LEAD(c) && (src) != (limit) && U16_IS_TRAIL(__c2 = *(src))) { \
            ++(src); \
            (c) = U16_GET_SUPPLEMENTARY((c), __c2); \
            __index = _UCPTRIE_SMALL_INDEX(trie, c); \
        } else { \
            __index = (trie)->dataLength - UCPTRIE_ERROR_VALUE_NEG_DATA_OFFSET; \
        } \
    } \
    (result) = dataAccess(trie, __index); \
} UPRV_BLOCK_MACRO_END

/**
 * UTF-16: Reads the previous code point (UChar32 c, out), pre-decrements src,
 * and gets a value from the trie.
 * Sets the trie error value if c is an unpaired surrogate.
 *
 * @param trie (const UCPTrie *, in) the trie; must have type UCPTRIE_TYPE_FAST
 * @param dataAccess UCPTRIE_16, UCPTRIE_32, or UCPTRIE_8 according to the trie’s value width
 * @param start (const UChar *, in) the start pointer for the text
 * @param src (const UChar *, in/out) the source text pointer
 * @param c (UChar32, out) variable for the code point
 * @param result (out) variable for the trie lookup result
 * @stable ICU 63
 */
#define UCPTRIE_FAST_U16_PREV(trie, dataAccess, start, src, c, result) UPRV_BLOCK_MACRO_BEGIN { \
    (c) = *--(src); \
    int32_t __index; \
    if (!U16_IS_SURROGATE(c)) { \
        __index = _UCPTRIE_FAST_INDEX(trie, c); \
    } else { \
        uint16_t __c2; \
        if (U16_IS_SURROGATE_TRAIL(c) && (src) != (start) && U16_IS_LEAD(__c2 = *((src) - 1))) { \
            --(src); \
            (c) = U16_GET_SUPPLEMENTARY(__c2, (c)); \
            __index = _UCPTRIE_SMALL_INDEX(trie, c); \
        } else { \
            __index = (trie)->dataLength - UCPTRIE_ERROR_VALUE_NEG_DATA_OFFSET; \
        } \
    } \
    (result) = dataAccess(trie, __index); \
} UPRV_BLOCK_MACRO_END

/**
 * UTF-8: Post-increments src and gets a value from the trie.
 * Sets the trie error value for an ill-formed byte sequence.
 *
 * Unlike UCPTRIE_FAST_U16_NEXT() this UTF-8 macro does not provide the code point
 * because it would be more work to do so and is often not needed.
 * If the trie value differs from the error value, then the byte sequence is well-formed,
 * and the code point can be assembled without revalidation.
 *
 * @param trie (const UCPTrie *, in) the trie; must have type UCPTRIE_TYPE_FAST
 * @param dataAccess UCPTRIE_16, UCPTRIE_32, or UCPTRIE_8 according to the trie’s value width
 * @param src (const char *, in/out) the source text pointer
 * @param limit (const char *, in) the limit pointer for the text (must not be NULL)
 * @param result (out) variable for the trie lookup result
 * @stable ICU 63
 */
#define UCPTRIE_FAST_U8_NEXT(trie, dataAccess, src, limit, result) UPRV_BLOCK_MACRO_BEGIN { \
    int32_t __lead = (uint8_t)*(src)++; \
    if (!U8_IS_SINGLE(__lead)) { \
        uint8_t __t1, __t2, __t3; \
        if ((src) != (limit) && \
            (__lead >= 0xe0 ? \
                __lead < 0xf0 ?  /* U+0800..U+FFFF except surrogates */ \
                    U8_LEAD3_T1_BITS[__lead &= 0xf] & (1 << ((__t1 = *(src)) >> 5)) && \
                    ++(src) != (limit) && (__t2 = *(src) - 0x80) <= 0x3f && \
                    (__lead = ((int32_t)(trie)->index[(__lead << 6) + (__t1 & 0x3f)]) + __t2, 1) \
                :  /* U+10000..U+10FFFF */ \
                    (__lead -= 0xf0) <= 4 && \
                    U8_LEAD4_T1_BITS[(__t1 = *(src)) >> 4] & (1 << __lead) && \
                    (__lead = (__lead << 6) | (__t1 & 0x3f), ++(src) != (limit)) && \
                    (__t2 = *(src) - 0x80) <= 0x3f && \
                    ++(src) != (limit) && (__t3 = *(src) - 0x80) <= 0x3f && \
                    (__lead = __lead >= (trie)->shifted12HighStart ? \
                        (trie)->dataLength - UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET : \
                        ucptrie_internalSmallU8Index((trie), __lead, __t2, __t3), 1) \
            :  /* U+0080..U+07FF */ \
                __lead >= 0xc2 && (__t1 = *(src) - 0x80) <= 0x3f && \
                (__lead = (int32_t)(trie)->index[__lead & 0x1f] + __t1, 1))) { \
            ++(src); \
        } else { \
            __lead = (trie)->dataLength - UCPTRIE_ERROR_VALUE_NEG_DATA_OFFSET;  /* ill-formed*/ \
        } \
    } \
    (result) = dataAccess(trie, __lead); \
} UPRV_BLOCK_MACRO_END

/**
 * UTF-8: Pre-decrements src and gets a value from the trie.
 * Sets the trie error value for an ill-formed byte sequence.
 *
 * Unlike UCPTRIE_FAST_U16_PREV() this UTF-8 macro does not provide the code point
 * because it would be more work to do so and is often not needed.
 * If the trie value differs from the error value, then the byte sequence is well-formed,
 * and the code point can be assembled without revalidation.
 *
 * @param trie (const UCPTrie *, in) the trie; must have type UCPTRIE_TYPE_FAST
 * @param dataAccess UCPTRIE_16, UCPTRIE_32, or UCPTRIE_8 according to the trie’s value width
 * @param start (const char *, in) the start pointer for the text
 * @param src (const char *, in/out) the source text pointer
 * @param result (out) variable for the trie lookup result
 * @stable ICU 63
 */
#define UCPTRIE_FAST_U8_PREV(trie, dataAccess, start, src, result) UPRV_BLOCK_MACRO_BEGIN { \
    int32_t __index = (uint8_t)*--(src); \
    if (!U8_IS_SINGLE(__index)) { \
        __index = ucptrie_internalU8PrevIndex((trie), __index, (const uint8_t *)(start), \
                                              (const uint8_t *)(src)); \
        (src) -= __index & 7; \
        __index >>= 3; \
    } \
    (result) = dataAccess(trie, __index); \
} UPRV_BLOCK_MACRO_END

/**
 * Returns a trie value for an ASCII code point, without range checking.
 *
 * @param trie (const UCPTrie *, in) the trie (of either fast or small type)
 * @param dataAccess UCPTRIE_16, UCPTRIE_32, or UCPTRIE_8 according to the trie’s value width
 * @param c (UChar32, in) the input code point; must be U+0000..U+007F
 * @return The ASCII code point's trie value.
 * @stable ICU 63
 */
#define UCPTRIE_ASCII_GET(trie, dataAccess, c) dataAccess(trie, c)

/**
 * Returns a trie value for a BMP code point (U+0000..U+FFFF), without range checking.
 * Can be used to look up a value for a UTF-16 code unit if other parts of
 * the string processing check for surrogates.
 *
 * @param trie (const UCPTrie *, in) the trie; must have type UCPTRIE_TYPE_FAST
 * @param dataAccess UCPTRIE_16, UCPTRIE_32, or UCPTRIE_8 according to the trie’s value width
 * @param c (UChar32, in) the input code point, must be U+0000..U+FFFF
 * @return The BMP code point's trie value.
 * @stable ICU 63
 */
#define UCPTRIE_FAST_BMP_GET(trie, dataAccess, c) dataAccess(trie, _UCPTRIE_FAST_INDEX(trie, c))

/**
 * Returns a trie value for a supplementary code point (U+10000..U+10FFFF),
 * without range checking.
 *
 * @param trie (const UCPTrie *, in) the trie; must have type UCPTRIE_TYPE_FAST
 * @param dataAccess UCPTRIE_16, UCPTRIE_32, or UCPTRIE_8 according to the trie’s value width
 * @param c (UChar32, in) the input code point, must be U+10000..U+10FFFF
 * @return The supplementary code point's trie value.
 * @stable ICU 63
 */
#define UCPTRIE_FAST_SUPP_GET(trie, dataAccess, c) dataAccess(trie, _UCPTRIE_SMALL_INDEX(trie, c))

/* Internal definitions ----------------------------------------------------- */

#ifndef U_IN_DOXYGEN

/**
 * Internal implementation constants.
 * These are needed for the API macros, but users should not use these directly.
 * @internal
 */
enum {
    /** @internal */
    UCPTRIE_FAST_SHIFT = 6,

    /** Number of entries in a data block for code points below the fast limit. 64=0x40 @internal */
    UCPTRIE_FAST_DATA_BLOCK_LENGTH = 1 << UCPTRIE_FAST_SHIFT,

    /** Mask for getting the lower bits for the in-fast-data-block offset. @internal */
    UCPTRIE_FAST_DATA_MASK = UCPTRIE_FAST_DATA_BLOCK_LENGTH - 1,

    /** @internal */
    UCPTRIE_SMALL_MAX = 0xfff,

    /**
     * Offset from dataLength (to be subtracted) for fetching the
     * value returned for out-of-range code points and ill-formed UTF-8/16.
     * @internal
     */
    UCPTRIE_ERROR_VALUE_NEG_DATA_OFFSET = 1,
    /**
     * Offset from dataLength (to be subtracted) for fetching the
     * value returned for code points highStart..U+10FFFF.
     * @internal
     */
    UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET = 2
};

/* Internal functions and macros -------------------------------------------- */
// Do not conditionalize with #ifndef U_HIDE_INTERNAL_API, needed for public API

/** @internal */
U_INTERNAL int32_t U_EXPORT2
ucptrie_internalSmallIndex(const UCPTrie *trie, UChar32 c);

/** @internal */
U_INTERNAL int32_t U_EXPORT2
ucptrie_internalSmallU8Index(const UCPTrie *trie, int32_t lt1, uint8_t t2, uint8_t t3);

/**
 * Internal function for part of the UCPTRIE_FAST_U8_PREVxx() macro implementations.
 * Do not call directly.
 * @internal
 */
U_INTERNAL int32_t U_EXPORT2
ucptrie_internalU8PrevIndex(const UCPTrie *trie, UChar32 c,
                            const uint8_t *start, const uint8_t *src);

/** Internal trie getter for a code point below the fast limit. Returns the data index. @internal */
#define _UCPTRIE_FAST_INDEX(trie, c) \
    ((int32_t)(trie)->index[(c) >> UCPTRIE_FAST_SHIFT] + ((c) & UCPTRIE_FAST_DATA_MASK))

/** Internal trie getter for a code point at or above the fast limit. Returns the data index. @internal */
#define _UCPTRIE_SMALL_INDEX(trie, c) \
    ((c) >= (trie)->highStart ? \
        (trie)->dataLength - UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET : \
        ucptrie_internalSmallIndex(trie, c))

/**
 * Internal trie getter for a code point, with checking that c is in U+0000..10FFFF.
 * Returns the data index.
 * @internal
 */
#define _UCPTRIE_CP_INDEX(trie, fastMax, c) \
    ((uint32_t)(c) <= (uint32_t)(fastMax) ? \
        _UCPTRIE_FAST_INDEX(trie, c) : \
        (uint32_t)(c) <= 0x10ffff ? \
            _UCPTRIE_SMALL_INDEX(trie, c) : \
            (trie)->dataLength - UCPTRIE_ERROR_VALUE_NEG_DATA_OFFSET)

U_CDECL_END

#endif  // U_IN_DOXYGEN

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUCPTriePointer
 * "Smart pointer" class, closes a UCPTrie via ucptrie_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 63
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUCPTriePointer, UCPTrie, ucptrie_close);

U_NAMESPACE_END

#endif  // U_SHOW_CPLUSPLUS_API






// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1997-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*/
//----------------------------------------------------------------------------
// File:     mutex.h
//
// Lightweight C++ wrapper for umtx_ C mutex functions
//
// Author:   Alan Liu  1/31/97
// History:
// 06/04/97   helena         Updated setImplementation as per feedback from 5/21 drop.
// 04/07/1999  srl               refocused as a thin wrapper
//
//----------------------------------------------------------------------------



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1997-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*
* File UMUTEX.H
*
* Modification History:
*
*   Date        Name        Description
*   04/02/97  aliu        Creation.
*   04/07/99  srl         rewrite - C interface, multiple mutices
*   05/13/99  stephen     Changed to umutex (from cmutex)
******************************************************************************
*/


#include <atomic>
#include <condition_variable>
#include <mutex>
#include <type_traits>


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2001-2014, International Business Machines
*                Corporation and others. All Rights Reserved.
******************************************************************************
*   file name:  uclean.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001July05
*   created by: George Rhoten
*/



/**
 * \file
 * \brief C API: Initialize and clean up ICU
 */
 
/**
 *  Initialize ICU.
 *
 *  Use of this function is optional.  It is OK to simply use ICU
 *  services and functions without first having initialized
 *  ICU by calling u_init().
 *
 *  u_init() will attempt to load some part of ICU's data, and is
 *  useful as a test for configuration or installation problems that
 *  leave the ICU data inaccessible.  A successful invocation of u_init()
 *  does not, however, guarantee that all ICU data is accessible.
 *
 *  Multiple calls to u_init() cause no harm, aside from the small amount
 *  of time required.
 *
 *  In old versions of ICU, u_init() was required in multi-threaded applications
 *  to ensure the thread safety of ICU.  u_init() is no longer needed for this purpose.
 *
 * @param status An ICU UErrorCode parameter. It must not be <code>NULL</code>.
 *    An Error will be returned if some required part of ICU data can not
 *    be loaded or initialized.
 *    The function returns immediately if the input error code indicates a
 *    failure, as usual.
 *
 * @stable ICU 2.6
 */  
U_STABLE void U_EXPORT2 
u_init(UErrorCode *status);

#ifndef U_HIDE_SYSTEM_API
/**
 * Clean up the system resources, such as allocated memory or open files,
 * used in all ICU libraries. This will free/delete all memory owned by the
 * ICU libraries, and return them to their original load state. All open ICU
 * items (collators, resource bundles, converters, etc.) must be closed before
 * calling this function, otherwise ICU may not free its allocated memory
 * (e.g. close your converters and resource bundles before calling this
 * function). Generally, this function should be called once just before
 * an application exits. For applications that dynamically load and unload
 * the ICU libraries (relatively uncommon), u_cleanup() should be called
 * just before the library unload.
 * <p>
 * u_cleanup() also clears any ICU heap functions, mutex functions or
 * trace functions that may have been set for the process.  
 * This has the effect of restoring ICU to its initial condition, before
 * any of these override functions were installed.  Refer to
 * u_setMemoryFunctions(), u_setMutexFunctions and 
 * utrace_setFunctions().  If ICU is to be reinitialized after
 * calling u_cleanup(), these runtime override functions will need to
 * be set up again if they are still required.
 * <p>
 * u_cleanup() is not thread safe.  All other threads should stop using ICU
 * before calling this function.
 * <p>
 * Any open ICU items will be left in an undefined state by u_cleanup(),
 * and any subsequent attempt to use such an item will give unpredictable
 * results.
 * <p>
 * After calling u_cleanup(), an application may continue to use ICU by
 * calling u_init().  An application must invoke u_init() first from one single
 * thread before allowing other threads call u_init().  All threads existing
 * at the time of the first thread's call to u_init() must also call
 * u_init() themselves before continuing with other ICU operations.  
 * <p>
 * The use of u_cleanup() just before an application terminates is optional,
 * but it should be called only once for performance reasons. The primary
 * benefit is to eliminate reports of memory or resource leaks originating
 * in ICU code from the results generated by heap analysis tools.
 * <p>
 * <strong>Use this function with great care!</strong>
 * </p>
 *
 * @stable ICU 2.0
 * @system
 */
U_STABLE void U_EXPORT2 
u_cleanup(void);

U_CDECL_BEGIN
/**
  *  Pointer type for a user supplied memory allocation function.
  *  @param context user supplied value, obtained from u_setMemoryFunctions().
  *  @param size    The number of bytes to be allocated
  *  @return        Pointer to the newly allocated memory, or NULL if the allocation failed.
  *  @stable ICU 2.8
  *  @system
  */
typedef void *U_CALLCONV UMemAllocFn(const void *context, size_t size);
/**
  *  Pointer type for a user supplied memory re-allocation function.
  *  @param context user supplied value, obtained from u_setMemoryFunctions().
  *  @param size    The number of bytes to be allocated
  *  @return        Pointer to the newly allocated memory, or NULL if the allocation failed.
  *  @stable ICU 2.8
  *  @system
  */
typedef void *U_CALLCONV UMemReallocFn(const void *context, void *mem, size_t size);
/**
  *  Pointer type for a user supplied memory free  function.  Behavior should be
  *  similar the standard C library free().
  *  @param context user supplied value, obtained from u_setMemoryFunctions().
  *  @param mem     Pointer to the memory block to be resized
  *  @param size    The new size for the block
  *  @return        Pointer to the resized memory block, or NULL if the resizing failed.
  *  @stable ICU 2.8
  *  @system
  */
typedef void  U_CALLCONV UMemFreeFn (const void *context, void *mem);

/**
 *  Set the functions that ICU will use for memory allocation.
 *  Use of this function is optional; by default (without this function), ICU will
 *  use the standard C library malloc() and free() functions.
 *  This function can only be used when ICU is in an initial, unused state, before
 *  u_init() has been called.
 *  @param context This pointer value will be saved, and then (later) passed as
 *                 a parameter to the memory functions each time they
 *                 are called.
 *  @param a       Pointer to a user-supplied malloc function.
 *  @param r       Pointer to a user-supplied realloc function.
 *  @param f       Pointer to a user-supplied free function.
 *  @param status  Receives error values.
 *  @stable ICU 2.8
 *  @system
 */  
U_STABLE void U_EXPORT2 
u_setMemoryFunctions(const void *context, UMemAllocFn * U_CALLCONV_FPTR a, UMemReallocFn * U_CALLCONV_FPTR r, UMemFreeFn * U_CALLCONV_FPTR f, 
                    UErrorCode *status);

U_CDECL_END

#ifndef U_HIDE_DEPRECATED_API
/*********************************************************************************
 *
 * Deprecated Functions
 *
 *    The following functions for user supplied mutexes are no longer supported.
 *    Any attempt to use them will return a U_UNSUPPORTED_ERROR.
 *
 **********************************************************************************/

/**
  * An opaque pointer type that represents an ICU mutex.
  * For user-implemented mutexes, the value will typically point to a
  *  struct or object that implements the mutex.
  * @deprecated ICU 52. This type is no longer supported.
  * @system
  */
typedef void *UMTX;

U_CDECL_BEGIN
/**
  *  Function Pointer type for a user supplied mutex initialization function.
  *  The user-supplied function will be called by ICU whenever ICU needs to create a
  *  new mutex.  The function implementation should create a mutex, and store a pointer
  *  to something that uniquely identifies the mutex into the UMTX that is supplied
  *  as a parameter.
  *  @param context user supplied value, obtained from u_setMutexFunctions().
  *  @param mutex   Receives a pointer that identifies the new mutex.
  *                 The mutex init function must set the UMTX to a non-null value.   
  *                 Subsequent calls by ICU to lock, unlock, or destroy a mutex will 
  *                 identify the mutex by the UMTX value.
  *  @param status  Error status.  Report errors back to ICU by setting this variable
  *                 with an error code.
  *  @deprecated ICU 52. This function is no longer supported.
  *  @system
  */
typedef void U_CALLCONV UMtxInitFn (const void *context, UMTX  *mutex, UErrorCode* status);


/**
  *  Function Pointer type for a user supplied mutex functions.
  *  One of the  user-supplied functions with this signature will be called by ICU
  *  whenever ICU needs to lock, unlock, or destroy a mutex.
  *  @param context user supplied value, obtained from u_setMutexFunctions().
  *  @param mutex   specify the mutex on which to operate.
  *  @deprecated ICU 52. This function is no longer supported.
  *  @system
  */
typedef void U_CALLCONV UMtxFn   (const void *context, UMTX  *mutex);
U_CDECL_END

/**
  *  Set the functions that ICU will use for mutex operations
  *  Use of this function is optional; by default (without this function), ICU will
  *  directly access system functions for mutex operations
  *  This function can only be used when ICU is in an initial, unused state, before
  *  u_init() has been called.
  *  @param context This pointer value will be saved, and then (later) passed as
  *                 a parameter to the user-supplied mutex functions each time they
  *                 are called. 
  *  @param init    Pointer to a mutex initialization function.  Must be non-null.
  *  @param destroy Pointer to the mutex destroy function.  Must be non-null.
  *  @param lock    pointer to the mutex lock function.  Must be non-null.
  *  @param unlock  Pointer to the mutex unlock function.  Must be non-null.
  *  @param status  Receives error values.
  *  @deprecated ICU 52. This function is no longer supported.
  *  @system
  */  
U_DEPRECATED void U_EXPORT2 
u_setMutexFunctions(const void *context, UMtxInitFn *init, UMtxFn *destroy, UMtxFn *lock, UMtxFn *unlock,
                    UErrorCode *status);


/**
  *  Pointer type for a user supplied atomic increment or decrement function.
  *  @param context user supplied value, obtained from u_setAtomicIncDecFunctions().
  *  @param p   Pointer to a 32 bit int to be incremented or decremented
  *  @return    The value of the variable after the inc or dec operation.
  *  @deprecated ICU 52. This function is no longer supported.
  *  @system
  */
typedef int32_t U_CALLCONV UMtxAtomicFn(const void *context, int32_t *p);

/**
 *  Set the functions that ICU will use for atomic increment and decrement of int32_t values.
 *  Use of this function is optional; by default (without this function), ICU will
 *  use its own internal implementation of atomic increment/decrement.
 *  This function can only be used when ICU is in an initial, unused state, before
 *  u_init() has been called.
 *  @param context This pointer value will be saved, and then (later) passed as
 *                 a parameter to the increment and decrement functions each time they
 *                 are called.  This function can only be called 
 *  @param inc     Pointer to a function to do an atomic increment operation.  Must be non-null.
 *  @param dec     Pointer to a function to do an atomic decrement operation.  Must be non-null.
 *  @param status  Receives error values.
 *  @deprecated ICU 52. This function is no longer supported.
 *  @system
 */  
U_DEPRECATED void U_EXPORT2 
u_setAtomicIncDecFunctions(const void *context, UMtxAtomicFn *inc, UMtxAtomicFn *dec,
                    UErrorCode *status);

#endif  /* U_HIDE_DEPRECATED_API */
#endif  /* U_HIDE_SYSTEM_API */




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
*  FILE NAME : putilimp.h
*
*   Date        Name        Description
*   10/17/04    grhoten     Move internal functions from putil.h to this file.
******************************************************************************
*/





/**
 * \def U_SIGNED_RIGHT_SHIFT_IS_ARITHMETIC
 * Nearly all CPUs and compilers implement a right-shift of a signed integer
 * as an Arithmetic Shift Right which copies the sign bit (the Most Significant Bit (MSB))
 * into the vacated bits (sign extension).
 * For example, (int32_t)0xfff5fff3>>4 becomes 0xffff5fff and -1>>1=-1.
 *
 * This can be useful for storing a signed value in the upper bits
 * and another bit field in the lower bits.
 * The signed value can be retrieved by simple right-shifting.
 *
 * This is consistent with the Java language.
 *
 * However, the C standard allows compilers to implement a right-shift of a signed integer
 * as a Logical Shift Right which copies a 0 into the vacated bits.
 * For example, (int32_t)0xfff5fff3>>4 becomes 0x0fff5fff and -1>>1=0x7fffffff.
 *
 * Code that depends on the natural behavior should be guarded with this macro,
 * with an alternate path for unusual platforms.
 * @internal
 */
#ifdef U_SIGNED_RIGHT_SHIFT_IS_ARITHMETIC
    /* Use the predefined value. */
#else
    /*
     * Nearly all CPUs & compilers implement a right-shift of a signed integer
     * as an Arithmetic Shift Right (with sign extension).
     */
#   define U_SIGNED_RIGHT_SHIFT_IS_ARITHMETIC 1
#endif

/** Define this to 1 if your platform supports IEEE 754 floating point,
   to 0 if it does not. */
#ifndef IEEE_754
#   define IEEE_754 1
#endif

/**
 * uintptr_t is an optional part of the standard definitions in stdint.h.
 * The opengroup.org documentation for stdint.h says
 * "On XSI-conformant systems, the intptr_t and uintptr_t types are required;
 * otherwise, they are optional."
 * We assume that when uintptr_t is defined, UINTPTR_MAX is defined as well.
 *
 * Do not use ptrdiff_t since it is signed. size_t is unsigned.
 */
/* TODO: This check fails on some z environments. Filed a ticket #9357 for this. */
#if !defined(__intptr_t_defined) && !defined(UINTPTR_MAX) && (U_PLATFORM != U_PF_OS390)
typedef size_t uintptr_t;
#endif

/*===========================================================================*/
/** @{ Information about POSIX support                                       */
/*===========================================================================*/

#ifdef U_HAVE_NL_LANGINFO_CODESET
    /* Use the predefined value. */
#elif U_PLATFORM_USES_ONLY_WIN32_API || U_PLATFORM == U_PF_ANDROID || U_PLATFORM == U_PF_QNX
#   define U_HAVE_NL_LANGINFO_CODESET 0
#else
#   define U_HAVE_NL_LANGINFO_CODESET 1
#endif

#ifdef U_NL_LANGINFO_CODESET
    /* Use the predefined value. */
#elif !U_HAVE_NL_LANGINFO_CODESET
#   define U_NL_LANGINFO_CODESET -1
#elif U_PLATFORM == U_PF_OS400
   /* not defined */
#else
#   define U_NL_LANGINFO_CODESET CODESET
#endif

#if defined(U_TZSET) || defined(U_HAVE_TZSET)
    /* Use the predefined value. */
#elif U_PLATFORM_USES_ONLY_WIN32_API
    // UWP doesn't support tzset or environment variables for tz
#if U_PLATFORM_HAS_WINUWP_API == 0
#   define U_TZSET _tzset
#endif
#elif U_PLATFORM == U_PF_OS400
   /* not defined */
#else
#   define U_TZSET tzset
#endif

#if defined(U_TIMEZONE) || defined(U_HAVE_TIMEZONE)
    /* Use the predefined value. */
#elif U_PLATFORM == U_PF_ANDROID
#   define U_TIMEZONE timezone
#elif defined(__UCLIBC__)
    // uClibc does not have __timezone or _timezone.
#elif defined(_NEWLIB_VERSION)
#   define U_TIMEZONE _timezone
#elif defined(__GLIBC__)
    // glibc
#   define U_TIMEZONE __timezone
#elif U_PLATFORM_IS_LINUX_BASED
    // not defined
#elif U_PLATFORM_USES_ONLY_WIN32_API
#   define U_TIMEZONE _timezone
#elif U_PLATFORM == U_PF_BSD && !defined(__NetBSD__)
   /* not defined */
#elif U_PLATFORM == U_PF_OS400
   /* not defined */
#elif U_PLATFORM == U_PF_IPHONE
   /* not defined */
#else
#   define U_TIMEZONE timezone
#endif

#if defined(U_TZNAME) || defined(U_HAVE_TZNAME)
    /* Use the predefined value. */
#elif U_PLATFORM_USES_ONLY_WIN32_API
    /* not usable on all windows platforms */
#if U_PLATFORM_HAS_WINUWP_API == 0
#   define U_TZNAME _tzname
#endif
#elif U_PLATFORM == U_PF_OS400
   /* not defined */
#else
#   define U_TZNAME tzname
#endif

#ifdef U_HAVE_MMAP
    /* Use the predefined value. */
#elif U_PLATFORM_USES_ONLY_WIN32_API
#   define U_HAVE_MMAP 0
#else
#   define U_HAVE_MMAP 1
#endif

#ifdef U_HAVE_POPEN
    /* Use the predefined value. */
#elif U_PLATFORM_USES_ONLY_WIN32_API
#   define U_HAVE_POPEN 0
#elif U_PLATFORM == U_PF_OS400
#   define U_HAVE_POPEN 0
#else
#   define U_HAVE_POPEN 1
#endif

/**
 * \def U_HAVE_DIRENT_H
 * Defines whether dirent.h is available.
 * @internal
 */
#ifdef U_HAVE_DIRENT_H
    /* Use the predefined value. */
#elif U_PLATFORM_USES_ONLY_WIN32_API
#   define U_HAVE_DIRENT_H 0
#else
#   define U_HAVE_DIRENT_H 1
#endif

/** @} */

/*===========================================================================*/
/** @{ Programs used by ICU code                                             */
/*===========================================================================*/

/**
 * \def U_MAKE_IS_NMAKE
 * Defines whether the "make" program is Windows nmake.
 */
#ifdef U_MAKE_IS_NMAKE
    /* Use the predefined value. */
#elif U_PLATFORM == U_PF_WINDOWS
#   define U_MAKE_IS_NMAKE 1
#else
#   define U_MAKE_IS_NMAKE 0
#endif

/** @} */

/*==========================================================================*/
/* Platform utilities                                                       */
/*==========================================================================*/

/**
 * Platform utilities isolates the platform dependencies of the
 * library.  For each platform which this code is ported to, these
 * functions may have to be re-implemented.
 */

/**
 * Floating point utility to determine if a double is Not a Number (NaN).
 * @internal
 */
U_INTERNAL UBool   U_EXPORT2 uprv_isNaN(double d);
/**
 * Floating point utility to determine if a double has an infinite value.
 * @internal
 */
U_INTERNAL UBool   U_EXPORT2 uprv_isInfinite(double d);
/**
 * Floating point utility to determine if a double has a positive infinite value.
 * @internal
 */
U_INTERNAL UBool   U_EXPORT2 uprv_isPositiveInfinity(double d);
/**
 * Floating point utility to determine if a double has a negative infinite value.
 * @internal
 */
U_INTERNAL UBool   U_EXPORT2 uprv_isNegativeInfinity(double d);
/**
 * Floating point utility that returns a Not a Number (NaN) value.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_getNaN(void);
/**
 * Floating point utility that returns an infinite value.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_getInfinity(void);

/**
 * Floating point utility to truncate a double.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_trunc(double d);
/**
 * Floating point utility to calculate the floor of a double.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_floor(double d);
/**
 * Floating point utility to calculate the ceiling of a double.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_ceil(double d);
/**
 * Floating point utility to calculate the absolute value of a double.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_fabs(double d);
/**
 * Floating point utility to calculate the fractional and integer parts of a double.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_modf(double d, double* pinteger);
/**
 * Floating point utility to calculate the remainder of a double divided by another double.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_fmod(double d, double y);
/**
 * Floating point utility to calculate d to the power of exponent (d^exponent).
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_pow(double d, double exponent);
/**
 * Floating point utility to calculate 10 to the power of exponent (10^exponent).
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_pow10(int32_t exponent);
/**
 * Floating point utility to calculate the maximum value of two doubles.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_fmax(double d, double y);
/**
 * Floating point utility to calculate the minimum value of two doubles.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_fmin(double d, double y);
/**
 * Private utility to calculate the maximum value of two integers.
 * @internal
 */
U_INTERNAL int32_t U_EXPORT2 uprv_max(int32_t d, int32_t y);
/**
 * Private utility to calculate the minimum value of two integers.
 * @internal
 */
U_INTERNAL int32_t U_EXPORT2 uprv_min(int32_t d, int32_t y);

#if U_IS_BIG_ENDIAN
#   define uprv_isNegative(number) (*((signed char *)&(number))<0)
#else
#   define uprv_isNegative(number) (*((signed char *)&(number)+sizeof(number)-1)<0)
#endif

/**
 * Return the largest positive number that can be represented by an integer
 * type of arbitrary bit length.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_maxMantissa(void);

/**
 * Floating point utility to calculate the logarithm of a double.
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_log(double d);

/**
 * Does common notion of rounding e.g. uprv_floor(x + 0.5);
 * @param x the double number
 * @return the rounded double
 * @internal
 */
U_INTERNAL double  U_EXPORT2 uprv_round(double x);

/**
 * Adds the signed integers a and b, storing the result in res.
 * Checks for signed integer overflow.
 * Similar to the GCC/Clang extension __builtin_add_overflow
 *
 * @param a The first operand.
 * @param b The second operand.
 * @param res a + b
 * @return true if overflow occurred; false if no overflow occurred.
 * @internal
 */
U_INTERNAL UBool U_EXPORT2 uprv_add32_overflow(int32_t a, int32_t b, int32_t* res);

/**
 * Multiplies the signed integers a and b, storing the result in res.
 * Checks for signed integer overflow.
 * Similar to the GCC/Clang extension __builtin_mul_overflow
 *
 * @param a The first multiplicand.
 * @param b The second multiplicand.
 * @param res a * b
 * @return true if overflow occurred; false if no overflow occurred.
 * @internal
 */
U_INTERNAL UBool U_EXPORT2 uprv_mul32_overflow(int32_t a, int32_t b, int32_t* res);

#if 0
/**
 * Returns the number of digits after the decimal point in a double number x.
 *
 * @param x the double number
 * @return the number of digits after the decimal point in a double number x.
 * @internal
 */
/*U_INTERNAL int32_t  U_EXPORT2 uprv_digitsAfterDecimal(double x);*/
#endif

#if !U_CHARSET_IS_UTF8
/**
 * Please use ucnv_getDefaultName() instead.
 * Return the default codepage for this platform and locale.
 * This function can call setlocale() on Unix platforms. Please read the
 * platform documentation on setlocale() before calling this function.
 * @return the default codepage for this platform
 * @internal
 */
U_INTERNAL const char*  U_EXPORT2 uprv_getDefaultCodepage(void);
#endif

/**
 * Please use uloc_getDefault() instead.
 * Return the default locale ID string by querying the system, or
 *     zero if one cannot be found.
 * This function can call setlocale() on Unix platforms. Please read the
 * platform documentation on setlocale() before calling this function.
 * @return the default locale ID string
 * @internal
 */
U_INTERNAL const char*  U_EXPORT2 uprv_getDefaultLocaleID(void);

/**
 * Time zone utilities
 *
 * Wrappers for C runtime library functions relating to timezones.
 * The t_tzset() function (similar to tzset) uses the current setting
 * of the environment variable TZ to assign values to three global
 * variables: daylight, timezone, and tzname. These variables have the
 * following meanings, and are declared in &lt;time.h&gt;.
 *
 *   daylight   Nonzero if daylight-saving-time zone (DST) is specified
 *              in TZ; otherwise, 0. Default value is 1.
 *   timezone   Difference in seconds between coordinated universal
 *              time and local time. E.g., -28,800 for PST (GMT-8hrs)
 *   tzname(0)  Three-letter time-zone name derived from TZ environment
 *              variable. E.g., "PST".
 *   tzname(1)  Three-letter DST zone name derived from TZ environment
 *              variable.  E.g., "PDT". If DST zone is omitted from TZ,
 *              tzname(1) is an empty string.
 *
 * Notes: For example, to set the TZ environment variable to correspond
 * to the current time zone in Germany, you can use one of the
 * following statements:
 *
 *   set TZ=GST1GDT
 *   set TZ=GST+1GDT
 *
 * If the TZ value is not set, t_tzset() attempts to use the time zone
 * information specified by the operating system. Under Windows NT
 * and Windows 95, this information is specified in the Control Panel's
 * Date/Time application.
 * @internal
 */
U_INTERNAL void     U_EXPORT2 uprv_tzset(void);

/**
 * Difference in seconds between coordinated universal
 * time and local time. E.g., -28,800 for PST (GMT-8hrs)
 * @return the difference in seconds between coordinated universal time and local time.
 * @internal
 */
U_INTERNAL int32_t  U_EXPORT2 uprv_timezone(void);

/**
 *   tzname(0)  Three-letter time-zone name derived from TZ environment
 *              variable. E.g., "PST".
 *   tzname(1)  Three-letter DST zone name derived from TZ environment
 *              variable.  E.g., "PDT". If DST zone is omitted from TZ,
 *              tzname(1) is an empty string.
 * @internal
 */
U_INTERNAL const char* U_EXPORT2 uprv_tzname(int n);

/**
 * Reset the global tzname cache.
 * @internal
 */
U_INTERNAL void uprv_tzname_clear_cache();

/**
 * Get UTC (GMT) time measured in milliseconds since 0:00 on 1/1/1970.
 * This function is affected by 'faketime' and should be the bottleneck for all user-visible ICU time functions.
 * @return the UTC time measured in milliseconds
 * @internal
 */
U_INTERNAL UDate U_EXPORT2 uprv_getUTCtime(void);

/**
 * Get UTC (GMT) time measured in milliseconds since 0:00 on 1/1/1970.
 * This function is not affected by 'faketime', so it should only be used by low level test functions- not by anything that
 * exposes time to the end user.
 * @return the UTC time measured in milliseconds
 * @internal
 */
U_INTERNAL UDate U_EXPORT2 uprv_getRawUTCtime(void);

/**
 * Determine whether a pathname is absolute or not, as defined by the platform.
 * @param path Pathname to test
 * @return TRUE if the path is absolute
 * @internal (ICU 3.0)
 */
U_INTERNAL UBool U_EXPORT2 uprv_pathIsAbsolute(const char *path);

/**
 * Use U_MAX_PTR instead of this function.
 * @param void pointer to test
 * @return the largest possible pointer greater than the base
 * @internal (ICU 3.8)
 */
U_INTERNAL void * U_EXPORT2 uprv_maximumPtr(void *base);

/**
 * Maximum value of a (void*) - use to indicate the limit of an 'infinite' buffer.
 * In fact, buffer sizes must not exceed 2GB so that the difference between
 * the buffer limit and the buffer start can be expressed in an int32_t.
 *
 * The definition of U_MAX_PTR must fulfill the following conditions:
 * - return the largest possible pointer greater than base
 * - return a valid pointer according to the machine architecture (AS/400, 64-bit, etc.)
 * - avoid wrapping around at high addresses
 * - make sure that the returned pointer is not farther from base than 0x7fffffff bytes
 *
 * @param base The beginning of a buffer to find the maximum offset from
 * @internal
 */
#ifndef U_MAX_PTR
#  if U_PLATFORM == U_PF_OS390 && !defined(_LP64)
    /* We have 31-bit pointers. */
#    define U_MAX_PTR(base) ((void *)0x7fffffff)
#  elif U_PLATFORM == U_PF_OS400
#    define U_MAX_PTR(base) uprv_maximumPtr((void *)base)
#  elif 0
    /*
     * For platforms where pointers are scalar values (which is normal, but unlike i5/OS)
     * but that do not define uintptr_t.
     *
     * However, this does not work on modern compilers:
     * The C++ standard does not define pointer overflow, and allows compilers to
     * assume that p+u>p for any pointer p and any integer u>0.
     * Thus, modern compilers optimize away the ">" comparison.
     * (See ICU tickets #7187 and #8096.)
     */
#    define U_MAX_PTR(base) \
    ((void *)(((char *)(base)+0x7fffffffu) > (char *)(base) \
        ? ((char *)(base)+0x7fffffffu) \
        : (char *)-1))
#  else
    /* Default version. C++ standard compliant for scalar pointers. */
#    define U_MAX_PTR(base) \
    ((void *)(((uintptr_t)(base)+0x7fffffffu) > (uintptr_t)(base) \
        ? ((uintptr_t)(base)+0x7fffffffu) \
        : (uintptr_t)-1))
#  endif
#endif


#ifdef __cplusplus
/**
 * Pin a buffer capacity such that doing pointer arithmetic
 * on the destination pointer and capacity cannot overflow.
 *
 * The pinned capacity must fulfill the following conditions (for positive capacities):
 *   - dest + capacity is a valid pointer according to the machine arcitecture (AS/400, 64-bit, etc.)
 *   - (dest + capacity) >= dest
 *   - The size (in bytes) of T[capacity] does not exceed 0x7fffffff
 *
 * @param dest the destination buffer pointer.
 * @param capacity the requested buffer capacity, in units of type T.
 * @return the pinned capacity.
 * @internal
 */
template <typename T>
inline int32_t pinCapacity(T *dest, int32_t capacity) {
    if (capacity <= 0) { return capacity; }

    uintptr_t destInt = (uintptr_t)dest;
    uintptr_t maxInt;

#  if U_PLATFORM == U_PF_OS390 && !defined(_LP64)
    // We have 31-bit pointers.
    maxInt = 0x7fffffff;
#  elif U_PLATFORM == U_PF_OS400
    maxInt = (uintptr_t)uprv_maximumPtr((void *)dest);
#  else
    maxInt = destInt + 0x7fffffffu;
    if (maxInt < destInt) {
        // Less than 2GB to the end of the address space.
        // Pin to that to prevent address overflow.
        maxInt = (uintptr_t)-1;
    }
#  endif

    uintptr_t maxBytes = maxInt - destInt;  // max. 2GB
    int32_t maxCapacity = (int32_t)(maxBytes / sizeof(T));
    return capacity <= maxCapacity ? capacity : maxCapacity;
}
#endif   // __cplusplus

/*  Dynamic Library Functions */

typedef void (UVoidFunction)(void);

/**
 * Define malloc and related functions
 * @internal
 */
#if U_PLATFORM == U_PF_OS400
# define uprv_default_malloc(x) _C_TS_malloc(x)
# define uprv_default_realloc(x,y) _C_TS_realloc(x,y)
# define uprv_default_free(x) _C_TS_free(x)
/* also _C_TS_calloc(x) */
#else
/* C defaults */
# define uprv_default_malloc(x) malloc(x)
# define uprv_default_realloc(x,y) realloc(x,y)
# define uprv_default_free(x) free(x)
#endif




#if defined(U_USER_ATOMICS_H) || defined(U_USER_MUTEX_H)
// Support for including an alternate implementation of atomic & mutex operations has been withdrawn.
// See issue ICU-20185.
#error U_USER_ATOMICS and U_USER_MUTEX_H are not supported
#endif

// Export an explicit template instantiation of std::atomic<int32_t>. 
// When building DLLs for Windows this is required as it is used as a data member of the exported SharedObject class.
// See digitlst.h, pluralaffix.h, datefmt.h, and others for similar examples.
//
// Similar story for std::atomic<std::mutex *>, and the exported UMutex class.
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN && !defined(U_IN_DOXYGEN)
#if defined(__clang__) || defined(_MSC_VER)
  #if defined(__clang__)
    // Suppress the warning that the explicit instantiation after explicit specialization has no effect.
    #pragma clang diagnostic push
    #pragma clang diagnostic ignored "-Winstantiation-after-specialization"
  #endif
template struct U_COMMON_API std::atomic<int32_t>;
template struct U_COMMON_API std::atomic<std::mutex *>;
  #if defined(__clang__)
    #pragma clang diagnostic pop
  #endif
#elif defined(__GNUC__)
// For GCC this class is already exported/visible, so no need for U_COMMON_API.
template struct std::atomic<int32_t>;
template struct std::atomic<std::mutex *>;
#endif
#endif


U_NAMESPACE_BEGIN

/****************************************************************************
 *
 *   Low Level Atomic Operations, ICU wrappers for.
 *
 ****************************************************************************/

typedef std::atomic<int32_t> u_atomic_int32_t;
#define ATOMIC_INT32_T_INITIALIZER(val) ATOMIC_VAR_INIT(val)

inline int32_t umtx_loadAcquire(u_atomic_int32_t &var) {
    return var.load(std::memory_order_acquire);
}

inline void umtx_storeRelease(u_atomic_int32_t &var, int32_t val) {
    var.store(val, std::memory_order_release);
}

inline int32_t umtx_atomic_inc(u_atomic_int32_t *var) {
    return var->fetch_add(1) + 1;
}

inline int32_t umtx_atomic_dec(u_atomic_int32_t *var) {
    return var->fetch_sub(1) - 1;
}


/*************************************************************************************************
 *
 *  UInitOnce Definitions.
 *
 *************************************************************************************************/

struct UInitOnce {
    u_atomic_int32_t   fState;
    UErrorCode       fErrCode;
    void reset() {fState = 0;}
    UBool isReset() {return umtx_loadAcquire(fState) == 0;}
// Note: isReset() is used by service registration code.
//                 Thread safety of this usage needs review.
};

#define U_INITONCE_INITIALIZER {ATOMIC_INT32_T_INITIALIZER(0), U_ZERO_ERROR}


U_COMMON_API UBool U_EXPORT2 umtx_initImplPreInit(UInitOnce &);
U_COMMON_API void  U_EXPORT2 umtx_initImplPostInit(UInitOnce &);

template<class T> void umtx_initOnce(UInitOnce &uio, T *obj, void (U_CALLCONV T::*fp)()) {
    if (umtx_loadAcquire(uio.fState) == 2) {
        return;
    }
    if (umtx_initImplPreInit(uio)) {
        (obj->*fp)();
        umtx_initImplPostInit(uio);
    }
}


// umtx_initOnce variant for plain functions, or static class functions.
//               No context parameter.
inline void umtx_initOnce(UInitOnce &uio, void (U_CALLCONV *fp)()) {
    if (umtx_loadAcquire(uio.fState) == 2) {
        return;
    }
    if (umtx_initImplPreInit(uio)) {
        (*fp)();
        umtx_initImplPostInit(uio);
    }
}

// umtx_initOnce variant for plain functions, or static class functions.
//               With ErrorCode, No context parameter.
inline void umtx_initOnce(UInitOnce &uio, void (U_CALLCONV *fp)(UErrorCode &), UErrorCode &errCode) {
    if (U_FAILURE(errCode)) {
        return;
    }
    if (umtx_loadAcquire(uio.fState) != 2 && umtx_initImplPreInit(uio)) {
        // We run the initialization.
        (*fp)(errCode);
        uio.fErrCode = errCode;
        umtx_initImplPostInit(uio);
    } else {
        // Someone else already ran the initialization.
        if (U_FAILURE(uio.fErrCode)) {
            errCode = uio.fErrCode;
        }
    }
}

// umtx_initOnce variant for plain functions, or static class functions,
//               with a context parameter.
template<class T> void umtx_initOnce(UInitOnce &uio, void (U_CALLCONV *fp)(T), T context) {
    if (umtx_loadAcquire(uio.fState) == 2) {
        return;
    }
    if (umtx_initImplPreInit(uio)) {
        (*fp)(context);
        umtx_initImplPostInit(uio);
    }
}

// umtx_initOnce variant for plain functions, or static class functions,
//               with a context parameter and an error code.
template<class T> void umtx_initOnce(UInitOnce &uio, void (U_CALLCONV *fp)(T, UErrorCode &), T context, UErrorCode &errCode) {
    if (U_FAILURE(errCode)) {
        return;
    }
    if (umtx_loadAcquire(uio.fState) != 2 && umtx_initImplPreInit(uio)) {
        // We run the initialization.
        (*fp)(context, errCode);
        uio.fErrCode = errCode;
        umtx_initImplPostInit(uio);
    } else {
        // Someone else already ran the initialization.
        if (U_FAILURE(uio.fErrCode)) {
            errCode = uio.fErrCode;
        }
    }
}

// UMutex should be constexpr-constructible, so that no initialization code
// is run during startup.
// This works on all C++ libraries except MS VS before VS2019.
#if (defined(_CPPLIB_VER) && !defined(_MSVC_STL_VERSION)) || \
    (defined(_MSVC_STL_VERSION) && _MSVC_STL_VERSION < 142)
    // (VS std lib older than VS2017) || (VS std lib version < VS2019)
#   define UMUTEX_CONSTEXPR
#else
#   define UMUTEX_CONSTEXPR constexpr
#endif

/**
 * UMutex - ICU Mutex class.
 *
 * This is the preferred Mutex class for use within ICU implementation code.
 * It is a thin wrapper over C++ std::mutex, with these additions:
 *    - Static instances are safe, not triggering static construction or destruction,
 *      and the associated order of construction or destruction issues.
 *    - Plumbed into u_cleanup() for destructing the underlying std::mutex,
 *      which frees any OS level resources they may be holding.
 *
 * Limitations:
 *    - Static or global instances only. Cannot be heap allocated. Cannot appear as a
 *      member of another class.
 *    - No condition variables or other advanced features. If needed, you will need to use
 *      std::mutex and std::condition_variable directly. For an example, see unifiedcache.cpp
 *
 * Typical Usage:
 *    static UMutex myMutex;
 *
 *    {
 *       Mutex lock(myMutex);
 *       ...    // Do stuff that is protected by myMutex;
 *    }         // myMutex is released when lock goes out of scope.
 */

class U_COMMON_API UMutex {
public:
    UMUTEX_CONSTEXPR UMutex() {}
    ~UMutex() = default;

    UMutex(const UMutex &other) = delete;
    UMutex &operator =(const UMutex &other) = delete;
    void *operator new(size_t) = delete;

    // requirements for C++ BasicLockable, allows UMutex to work with std::lock_guard
    void lock() {
        std::mutex *m = fMutex.load(std::memory_order_acquire);
        if (m == nullptr) { m = getMutex(); }
        m->lock();
    }
    void unlock() { fMutex.load(std::memory_order_relaxed)->unlock(); }

    static void cleanup();

private:
    alignas(std::mutex) char fStorage[sizeof(std::mutex)] {};
    std::atomic<std::mutex *> fMutex { nullptr };

    /** All initialized UMutexes are kept in a linked list, so that they can be found,
     * and the underlying std::mutex destructed, by u_cleanup().
     */
    UMutex *fListLink { nullptr };
    static UMutex *gListHead;

    /** Out-of-line function to lazily initialize a UMutex on first use.
     * Initial fast check is inline, in lock().  The returned value may never
     * be nullptr.
     */
    std::mutex *getMutex();
};


/* Lock a mutex.
 * @param mutex The given mutex to be locked.  Pass NULL to specify
 *              the global ICU mutex.  Recursive locks are an error
 *              and may cause a deadlock on some platforms.
 */
U_INTERNAL void U_EXPORT2 umtx_lock(UMutex* mutex);

/* Unlock a mutex.
 * @param mutex The given mutex to be unlocked.  Pass NULL to specify
 *              the global ICU mutex.
 */
U_INTERNAL void U_EXPORT2 umtx_unlock (UMutex* mutex);


U_NAMESPACE_END



U_NAMESPACE_BEGIN

/**
  * Mutex is a helper class for convenient locking and unlocking of a UMutex.
  *
  * Creating a local scope Mutex will lock a UMutex, holding the lock until the Mutex
  * goes out of scope.
  *
  *  If no UMutex is specified, the ICU global mutex is implied.
  *
  *  For example:
  *
  *  static UMutex myMutex;
  *
  *  void Function(int arg1, int arg2)
  *  {
  *     static Object* foo;      // Shared read-write object
  *     Mutex mutex(&myMutex);   // or no args for the global lock
  *     foo->Method();
  *     // When 'mutex' goes out of scope and gets destroyed here, the lock is released
  *  }
  *
  *  Note:  Do NOT use the form 'Mutex mutex();' as that merely forward-declares a function
  *         returning a Mutex. This is a common mistake which silently slips through the
  *         compiler!!
  */

class U_COMMON_API Mutex : public UMemory {
public:
    Mutex(UMutex *mutex = nullptr) : fMutex(mutex) {
        umtx_lock(fMutex);
    }
    ~Mutex() {
        umtx_unlock(fMutex);
    }

    Mutex(const Mutex &other) = delete; // forbid assigning of this class
    Mutex &operator=(const Mutex &other) = delete; // forbid copying of this class
    void *operator new(size_t s) = delete;  // forbid heap allocation. Locals only.

private:
    UMutex   *fMutex;
};


U_NAMESPACE_END


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2003-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  udataswp.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2003jun05
*   created by: Markus W. Scherer
*
*   Definitions for ICU data transformations for different platforms,
*   changing between big- and little-endian data and/or between
*   charset families (ASCII<->EBCDIC).
*/


#include <stdarg.h>


/* forward declaration */

U_CDECL_BEGIN

struct UDataSwapper;
typedef struct UDataSwapper UDataSwapper;

/**
 * Function type for data transformation.
 * Transforms data, or just returns the length of the data if
 * the input length is -1.
 * Swap functions assume that their data pointers are aligned properly.
 *
 * Quick implementation outline:
 * (best to copy and adapt and existing swapper implementation)
 * check that the data looks like the expected format
 * if(length<0) {
 *   preflight:
 *   never dereference outData
 *   read inData and determine the data size
 *   assume that inData is long enough for this
 * } else {
 *   outData can be NULL if length==0
 *   inData==outData (in-place swapping) possible but not required!
 *   verify that length>=(actual size)
 *   if there is a chance that not every byte up to size is reached
 *     due to padding etc.:
 *   if(inData!=outData) {
 *     memcpy(outData, inData, actual size);
 *   }
 *   swap contents
 * }
 * return actual size
 *
 * Further implementation notes:
 * - read integers from inData before swapping them
 *   because in-place swapping can make them unreadable
 * - compareInvChars compares a local Unicode string with already-swapped
 *   output charset strings
 *
 * @param ds Pointer to UDataSwapper containing global data about the
 *           transformation and function pointers for handling primitive
 *           types.
 * @param inData Pointer to the input data to be transformed or examined.
 * @param length Length of the data, counting bytes. May be -1 for preflighting.
 *               If length>=0, then transform the data.
 *               If length==-1, then only determine the length of the data.
 *               The length cannot be determined from the data itself for all
 *               types of data (e.g., not for simple arrays of integers).
 * @param outData Pointer to the output data buffer.
 *                If length>=0 (transformation), then the output buffer must
 *                have a capacity of at least length.
 *                If length==-1, then outData will not be used and can be NULL.
 * @param pErrorCode ICU UErrorCode parameter, must not be NULL and must
 *                   fulfill U_SUCCESS on input.
 * @return The actual length of the data.
 *
 * @see UDataSwapper
 * @internal ICU 2.8
 */
typedef int32_t U_CALLCONV
UDataSwapFn(const UDataSwapper *ds,
            const void *inData, int32_t length, void *outData,
            UErrorCode *pErrorCode);

/**
 * Convert one uint16_t from input to platform endianness.
 * @internal ICU 2.8
 */
typedef uint16_t U_CALLCONV
UDataReadUInt16(uint16_t x);

/**
 * Convert one uint32_t from input to platform endianness.
 * @internal ICU 2.8
 */
typedef uint32_t U_CALLCONV
UDataReadUInt32(uint32_t x);

/**
 * Convert one uint16_t from platform to input endianness.
 * @internal ICU 2.8
 */
typedef void U_CALLCONV
UDataWriteUInt16(uint16_t *p, uint16_t x);

/**
 * Convert one uint32_t from platform to input endianness.
 * @internal ICU 2.8
 */
typedef void U_CALLCONV
UDataWriteUInt32(uint32_t *p, uint32_t x);

/**
 * Compare invariant-character strings, one in the output data and the
 * other one caller-provided in Unicode.
 * An output data string is compared because strings are usually swapped
 * before the rest of the data, to allow for sorting of string tables
 * according to the output charset.
 * You can use -1 for the length parameters of NUL-terminated strings as usual.
 * Returns Unicode code point order for invariant characters.
 * @internal ICU 2.8
 */
typedef int32_t U_CALLCONV
UDataCompareInvChars(const UDataSwapper *ds,
                     const char *outString, int32_t outLength,
                     const UChar *localString, int32_t localLength);

/**
 * Function for message output when an error occurs during data swapping.
 * A format string and variable number of arguments are passed
 * like for vprintf().
 *
 * @param context A function-specific context pointer.
 * @param fmt The format string.
 * @param args The arguments for format string inserts.
 *
 * @internal ICU 2.8
 */
typedef void U_CALLCONV
UDataPrintError(void *context, const char *fmt, va_list args);

struct UDataSwapper {
    /** Input endianness. @internal ICU 2.8 */
    UBool inIsBigEndian;
    /** Input charset family. @see U_CHARSET_FAMILY @internal ICU 2.8 */
    uint8_t inCharset;
    /** Output endianness. @internal ICU 2.8 */
    UBool outIsBigEndian;
    /** Output charset family. @see U_CHARSET_FAMILY @internal ICU 2.8 */
    uint8_t outCharset;

    /* basic functions for reading data values */

    /** Convert one uint16_t from input to platform endianness. @internal ICU 2.8 */
    UDataReadUInt16 *readUInt16;
    /** Convert one uint32_t from input to platform endianness. @internal ICU 2.8 */
    UDataReadUInt32 *readUInt32;
    /** Compare an invariant-character output string with a local one. @internal ICU 2.8 */
    UDataCompareInvChars *compareInvChars;

    /* basic functions for writing data values */

    /** Convert one uint16_t from platform to input endianness. @internal ICU 2.8 */
    UDataWriteUInt16 *writeUInt16;
    /** Convert one uint32_t from platform to input endianness. @internal ICU 2.8 */
    UDataWriteUInt32 *writeUInt32;

    /* basic functions for data transformations */

    /** Transform an array of 16-bit integers. @internal ICU 2.8 */
    UDataSwapFn *swapArray16;
    /** Transform an array of 32-bit integers. @internal ICU 2.8 */
    UDataSwapFn *swapArray32;
    /** Transform an array of 64-bit integers. @internal ICU 53 */
    UDataSwapFn *swapArray64;
    /** Transform an invariant-character string. @internal ICU 2.8 */
    UDataSwapFn *swapInvChars;

    /**
     * Function for message output when an error occurs during data swapping.
     * Can be NULL.
     * @internal ICU 2.8
     */
    UDataPrintError *printError;
    /** Context pointer for printError. @internal ICU 2.8 */
    void *printErrorContext;
};

U_CDECL_END

U_CAPI UDataSwapper * U_EXPORT2
udata_openSwapper(UBool inIsBigEndian, uint8_t inCharset,
                  UBool outIsBigEndian, uint8_t outCharset,
                  UErrorCode *pErrorCode);

/**
 * Open a UDataSwapper for the given input data and the specified output
 * characteristics.
 * Values of -1 for any of the characteristics mean the local platform's
 * characteristics.
 *
 * @see udata_swap
 * @internal ICU 2.8
 */
U_CAPI UDataSwapper * U_EXPORT2
udata_openSwapperForInputData(const void *data, int32_t length,
                              UBool outIsBigEndian, uint8_t outCharset,
                              UErrorCode *pErrorCode);

U_CAPI void U_EXPORT2
udata_closeSwapper(UDataSwapper *ds);

/**
 * Read the beginning of an ICU data piece, recognize magic bytes,
 * swap the structure.
 * Set a U_UNSUPPORTED_ERROR if it does not look like an ICU data piece.
 *
 * @return The size of the data header, in bytes.
 *
 * @internal ICU 2.8
 */
U_CAPI int32_t U_EXPORT2
udata_swapDataHeader(const UDataSwapper *ds,
                     const void *inData, int32_t length, void *outData,
                     UErrorCode *pErrorCode);

/**
 * Convert one int16_t from input to platform endianness.
 * @internal ICU 2.8
 */
U_CAPI int16_t U_EXPORT2
udata_readInt16(const UDataSwapper *ds, int16_t x);

/**
 * Convert one int32_t from input to platform endianness.
 * @internal ICU 2.8
 */
U_CAPI int32_t U_EXPORT2
udata_readInt32(const UDataSwapper *ds, int32_t x);

/**
 * Swap a block of invariant, NUL-terminated strings, but not padding
 * bytes after the last string.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
udata_swapInvStringBlock(const UDataSwapper *ds,
                         const void *inData, int32_t length, void *outData,
                         UErrorCode *pErrorCode);

U_CAPI void U_EXPORT2
udata_printError(const UDataSwapper *ds,
                 const char *fmt,
                 ...);

/* internal exports from putil.c -------------------------------------------- */

/* declared here to keep them out of the public putil.h */

/**
 * Swap invariant char * strings ASCII->EBCDIC.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
uprv_ebcdicFromAscii(const UDataSwapper *ds,
                     const void *inData, int32_t length, void *outData,
                     UErrorCode *pErrorCode);

/**
 * Copy invariant ASCII char * strings and verify they are invariant.
 * @internal
 */
U_CFUNC int32_t
uprv_copyAscii(const UDataSwapper *ds,
               const void *inData, int32_t length, void *outData,
               UErrorCode *pErrorCode);

/**
 * Swap invariant char * strings EBCDIC->ASCII.
 * @internal
 */
U_CFUNC int32_t
uprv_asciiFromEbcdic(const UDataSwapper *ds,
                     const void *inData, int32_t length, void *outData,
                     UErrorCode *pErrorCode);

/**
 * Copy invariant EBCDIC char * strings and verify they are invariant.
 * @internal
 */
U_CFUNC int32_t
uprv_copyEbcdic(const UDataSwapper *ds,
                const void *inData, int32_t length, void *outData,
                UErrorCode *pErrorCode);

/**
 * Compare ASCII invariant char * with Unicode invariant UChar *
 * @internal
 */
U_CFUNC int32_t
uprv_compareInvAscii(const UDataSwapper *ds,
                     const char *outString, int32_t outLength,
                     const UChar *localString, int32_t localLength);

/**
 * Compare EBCDIC invariant char * with Unicode invariant UChar *
 * @internal
 */
U_CFUNC int32_t
uprv_compareInvEbcdic(const UDataSwapper *ds,
                      const char *outString, int32_t outLength,
                      const UChar *localString, int32_t localLength);

/**
 * \def uprv_compareInvWithUChar
 * Compare an invariant-character strings with a UChar string
 * @internal
 */
#if U_CHARSET_FAMILY==U_ASCII_FAMILY
#   define uprv_compareInvWithUChar uprv_compareInvAscii
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
#   define uprv_compareInvWithUChar uprv_compareInvEbcdic
#else
#   error Unknown charset family!
#endif

// utrie_swap.cpp -----------------------------------------------------------***

/**
 * Swaps a serialized UTrie.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
utrie_swap(const UDataSwapper *ds,
           const void *inData, int32_t length, void *outData,
           UErrorCode *pErrorCode);

/**
 * Swaps a serialized UTrie2.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
utrie2_swap(const UDataSwapper *ds,
            const void *inData, int32_t length, void *outData,
            UErrorCode *pErrorCode);

/**
 * Swaps a serialized UCPTrie.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
ucptrie_swap(const UDataSwapper *ds,
             const void *inData, int32_t length, void *outData,
             UErrorCode *pErrorCode);

/**
 * Swaps a serialized UTrie, UTrie2, or UCPTrie.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
utrie_swapAnyVersion(const UDataSwapper *ds,
                     const void *inData, int32_t length, void *outData,
                     UErrorCode *pErrorCode);

/* material... -------------------------------------------------------------- */

#if 0

/* udata.h */

/**
 * Public API function in udata.c
 *
 * Same as udata_openChoice() but automatically swaps the data.
 * isAcceptable, if not NULL, may accept data with endianness and charset family
 * different from the current platform's properties.
 * If the data is acceptable and the platform properties do not match, then
 * the swap function is called to swap an allocated version of the data.
 * Preflighting may or may not be performed depending on whether the size of
 * the loaded data item is known.
 *
 * @param isAcceptable Same as for udata_openChoice(). May be NULL.
 *
 * @internal ICU 2.8
 */
U_CAPI UDataMemory * U_EXPORT2
udata_openSwap(const char *path, const char *type, const char *name,
               UDataMemoryIsAcceptable *isAcceptable, void *isAcceptableContext,
               UDataSwapFn *swap,
               UDataPrintError *printError, void *printErrorContext,
               UErrorCode *pErrorCode);

#endif


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2004-2007, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uset_imp.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004sep07
*   created by: Markus W. Scherer
*
*   Internal USet definitions.
*/





U_CDECL_BEGIN

typedef void U_CALLCONV
USetAdd(USet *set, UChar32 c);

typedef void U_CALLCONV
USetAddRange(USet *set, UChar32 start, UChar32 end);

typedef void U_CALLCONV
USetAddString(USet *set, const UChar *str, int32_t length);

typedef void U_CALLCONV
USetRemove(USet *set, UChar32 c);

typedef void U_CALLCONV
USetRemoveRange(USet *set, UChar32 start, UChar32 end);

/**
 * Interface for adding items to a USet, to keep low-level code from
 * statically depending on the USet implementation.
 * Calls will look like sa->add(sa->set, c);
 */
struct USetAdder {
    USet *set;
    USetAdd *add;
    USetAddRange *addRange;
    USetAddString *addString;
    USetRemove *remove;
    USetRemoveRange *removeRange;
};
typedef struct USetAdder USetAdder;

U_CDECL_END



// When the nfc.nrm data is *not* hardcoded into the common library
// (with this constant set to 0),
// then it needs to be built into the data package:
// Add nfc.nrm to icu4c/source/data/Makefile.in DAT_FILES_SHORT
#define NORM2_HARDCODE_NFC_DATA 1

U_NAMESPACE_BEGIN

struct CanonIterData;

class ByteSink;
class Edits;
class InitCanonIterData;
class LcccContext;

class U_COMMON_API Hangul {
public:
    /* Korean Hangul and Jamo constants */
    enum {
        JAMO_L_BASE=0x1100,     /* "lead" jamo */
        JAMO_L_END=0x1112,
        JAMO_V_BASE=0x1161,     /* "vowel" jamo */
        JAMO_V_END=0x1175,
        JAMO_T_BASE=0x11a7,     /* "trail" jamo */
        JAMO_T_END=0x11c2,

        HANGUL_BASE=0xac00,
        HANGUL_END=0xd7a3,

        JAMO_L_COUNT=19,
        JAMO_V_COUNT=21,
        JAMO_T_COUNT=28,

        JAMO_VT_COUNT=JAMO_V_COUNT*JAMO_T_COUNT,

        HANGUL_COUNT=JAMO_L_COUNT*JAMO_V_COUNT*JAMO_T_COUNT,
        HANGUL_LIMIT=HANGUL_BASE+HANGUL_COUNT
    };

    static inline UBool isHangul(UChar32 c) {
        return HANGUL_BASE<=c && c<HANGUL_LIMIT;
    }
    static inline UBool
    isHangulLV(UChar32 c) {
        c-=HANGUL_BASE;
        return 0<=c && c<HANGUL_COUNT && c%JAMO_T_COUNT==0;
    }
    static inline UBool isJamoL(UChar32 c) {
        return (uint32_t)(c-JAMO_L_BASE)<JAMO_L_COUNT;
    }
    static inline UBool isJamoV(UChar32 c) {
        return (uint32_t)(c-JAMO_V_BASE)<JAMO_V_COUNT;
    }
    static inline UBool isJamoT(UChar32 c) {
        int32_t t=c-JAMO_T_BASE;
        return 0<t && t<JAMO_T_COUNT;  // not JAMO_T_BASE itself
    }
    static UBool isJamo(UChar32 c) {
        return JAMO_L_BASE<=c && c<=JAMO_T_END &&
            (c<=JAMO_L_END || (JAMO_V_BASE<=c && c<=JAMO_V_END) || JAMO_T_BASE<c);
    }

    /**
     * Decomposes c, which must be a Hangul syllable, into buffer
     * and returns the length of the decomposition (2 or 3).
     */
    static inline int32_t decompose(UChar32 c, UChar buffer[3]) {
        c-=HANGUL_BASE;
        UChar32 c2=c%JAMO_T_COUNT;
        c/=JAMO_T_COUNT;
        buffer[0]=(UChar)(JAMO_L_BASE+c/JAMO_V_COUNT);
        buffer[1]=(UChar)(JAMO_V_BASE+c%JAMO_V_COUNT);
        if(c2==0) {
            return 2;
        } else {
            buffer[2]=(UChar)(JAMO_T_BASE+c2);
            return 3;
        }
    }

    /**
     * Decomposes c, which must be a Hangul syllable, into buffer.
     * This is the raw, not recursive, decomposition. Its length is always 2.
     */
    static inline void getRawDecomposition(UChar32 c, UChar buffer[2]) {
        UChar32 orig=c;
        c-=HANGUL_BASE;
        UChar32 c2=c%JAMO_T_COUNT;
        if(c2==0) {
            c/=JAMO_T_COUNT;
            buffer[0]=(UChar)(JAMO_L_BASE+c/JAMO_V_COUNT);
            buffer[1]=(UChar)(JAMO_V_BASE+c%JAMO_V_COUNT);
        } else {
            buffer[0]=(UChar)(orig-c2);  // LV syllable
            buffer[1]=(UChar)(JAMO_T_BASE+c2);
        }
    }
private:
    Hangul();  // no instantiation
};

class Normalizer2Impl;

class U_COMMON_API ReorderingBuffer : public UMemory {
public:
    /** Constructs only; init() should be called. */
    ReorderingBuffer(const Normalizer2Impl &ni, UnicodeString &dest) :
        impl(ni), str(dest),
        start(NULL), reorderStart(NULL), limit(NULL),
        remainingCapacity(0), lastCC(0) {}
    /** Constructs, removes the string contents, and initializes for a small initial capacity. */
    ReorderingBuffer(const Normalizer2Impl &ni, UnicodeString &dest, UErrorCode &errorCode);
    ~ReorderingBuffer() {
        if(start!=NULL) {
            str.releaseBuffer((int32_t)(limit-start));
        }
    }
    UBool init(int32_t destCapacity, UErrorCode &errorCode);

    UBool isEmpty() const { return start==limit; }
    int32_t length() const { return (int32_t)(limit-start); }
    UChar *getStart() { return start; }
    UChar *getLimit() { return limit; }
    uint8_t getLastCC() const { return lastCC; }

    UBool equals(const UChar *start, const UChar *limit) const;
    UBool equals(const uint8_t *otherStart, const uint8_t *otherLimit) const;

    UBool append(UChar32 c, uint8_t cc, UErrorCode &errorCode) {
        return (c<=0xffff) ?
            appendBMP((UChar)c, cc, errorCode) :
            appendSupplementary(c, cc, errorCode);
    }
    UBool append(const UChar *s, int32_t length, UBool isNFD,
                 uint8_t leadCC, uint8_t trailCC,
                 UErrorCode &errorCode);
    UBool appendBMP(UChar c, uint8_t cc, UErrorCode &errorCode) {
        if(remainingCapacity==0 && !resize(1, errorCode)) {
            return FALSE;
        }
        if(lastCC<=cc || cc==0) {
            *limit++=c;
            lastCC=cc;
            if(cc<=1) {
                reorderStart=limit;
            }
        } else {
            insert(c, cc);
        }
        --remainingCapacity;
        return TRUE;
    }
    UBool appendZeroCC(UChar32 c, UErrorCode &errorCode);
    UBool appendZeroCC(const UChar *s, const UChar *sLimit, UErrorCode &errorCode);
    void remove();
    void removeSuffix(int32_t suffixLength);
    void setReorderingLimit(UChar *newLimit) {
        remainingCapacity+=(int32_t)(limit-newLimit);
        reorderStart=limit=newLimit;
        lastCC=0;
    }
    void copyReorderableSuffixTo(UnicodeString &s) const {
        s.setTo(ConstChar16Ptr(reorderStart), (int32_t)(limit-reorderStart));
    }
private:
    /*
     * TODO: Revisit whether it makes sense to track reorderStart.
     * It is set to after the last known character with cc<=1,
     * which stops previousCC() before it reads that character and looks up its cc.
     * previousCC() is normally only called from insert().
     * In other words, reorderStart speeds up the insertion of a combining mark
     * into a multi-combining mark sequence where it does not belong at the end.
     * This might not be worth the trouble.
     * On the other hand, it's not a huge amount of trouble.
     *
     * We probably need it for UNORM_SIMPLE_APPEND.
     */

    UBool appendSupplementary(UChar32 c, uint8_t cc, UErrorCode &errorCode);
    void insert(UChar32 c, uint8_t cc);
    static void writeCodePoint(UChar *p, UChar32 c) {
        if(c<=0xffff) {
            *p=(UChar)c;
        } else {
            p[0]=U16_LEAD(c);
            p[1]=U16_TRAIL(c);
        }
    }
    UBool resize(int32_t appendLength, UErrorCode &errorCode);

    const Normalizer2Impl &impl;
    UnicodeString &str;
    UChar *start, *reorderStart, *limit;
    int32_t remainingCapacity;
    uint8_t lastCC;

    // private backward iterator
    void setIterator() { codePointStart=limit; }
    void skipPrevious();  // Requires start<codePointStart.
    uint8_t previousCC();  // Returns 0 if there is no previous character.

    UChar *codePointStart, *codePointLimit;
};

/**
 * Low-level implementation of the Unicode Normalization Algorithm.
 * For the data structure and details see the documentation at the end of
 * this normalizer2impl.h and in the design doc at
 * http://site.icu-project.org/design/normalization/custom
 */
class U_COMMON_API Normalizer2Impl : public UObject {
public:
    Normalizer2Impl() : normTrie(NULL), fCanonIterData(NULL) { }
    virtual ~Normalizer2Impl();

    void init(const int32_t *inIndexes, const UCPTrie *inTrie,
              const uint16_t *inExtraData, const uint8_t *inSmallFCD);

    void addLcccChars(UnicodeSet &set) const;
    void addPropertyStarts(const USetAdder *sa, UErrorCode &errorCode) const;
    void addCanonIterPropertyStarts(const USetAdder *sa, UErrorCode &errorCode) const;

    // low-level properties ------------------------------------------------ ***

    UBool ensureCanonIterData(UErrorCode &errorCode) const;

    // The trie stores values for lead surrogate code *units*.
    // Surrogate code *points* are inert.
    uint16_t getNorm16(UChar32 c) const {
        return U_IS_LEAD(c) ?
            static_cast<uint16_t>(INERT) :
            UCPTRIE_FAST_GET(normTrie, UCPTRIE_16, c);
    }
    uint16_t getRawNorm16(UChar32 c) const { return UCPTRIE_FAST_GET(normTrie, UCPTRIE_16, c); }

    UNormalizationCheckResult getCompQuickCheck(uint16_t norm16) const {
        if(norm16<minNoNo || MIN_YES_YES_WITH_CC<=norm16) {
            return UNORM_YES;
        } else if(minMaybeYes<=norm16) {
            return UNORM_MAYBE;
        } else {
            return UNORM_NO;
        }
    }
    UBool isAlgorithmicNoNo(uint16_t norm16) const { return limitNoNo<=norm16 && norm16<minMaybeYes; }
    UBool isCompNo(uint16_t norm16) const { return minNoNo<=norm16 && norm16<minMaybeYes; }
    UBool isDecompYes(uint16_t norm16) const { return norm16<minYesNo || minMaybeYes<=norm16; }

    uint8_t getCC(uint16_t norm16) const {
        if(norm16>=MIN_NORMAL_MAYBE_YES) {
            return getCCFromNormalYesOrMaybe(norm16);
        }
        if(norm16<minNoNo || limitNoNo<=norm16) {
            return 0;
        }
        return getCCFromNoNo(norm16);
    }
    static uint8_t getCCFromNormalYesOrMaybe(uint16_t norm16) {
        return (uint8_t)(norm16 >> OFFSET_SHIFT);
    }
    static uint8_t getCCFromYesOrMaybe(uint16_t norm16) {
        return norm16>=MIN_NORMAL_MAYBE_YES ? getCCFromNormalYesOrMaybe(norm16) : 0;
    }
    uint8_t getCCFromYesOrMaybeCP(UChar32 c) const {
        if (c < minCompNoMaybeCP) { return 0; }
        return getCCFromYesOrMaybe(getNorm16(c));
    }

    /**
     * Returns the FCD data for code point c.
     * @param c A Unicode code point.
     * @return The lccc(c) in bits 15..8 and tccc(c) in bits 7..0.
     */
    uint16_t getFCD16(UChar32 c) const {
        if(c<minDecompNoCP) {
            return 0;
        } else if(c<=0xffff) {
            if(!singleLeadMightHaveNonZeroFCD16(c)) { return 0; }
        }
        return getFCD16FromNormData(c);
    }
    /**
     * Returns the FCD data for the next code point (post-increment).
     * Might skip only a lead surrogate rather than the whole surrogate pair if none of
     * the supplementary code points associated with the lead surrogate have non-zero FCD data.
     * @param s A valid pointer into a string. Requires s!=limit.
     * @param limit The end of the string, or NULL.
     * @return The lccc(c) in bits 15..8 and tccc(c) in bits 7..0.
     */
    uint16_t nextFCD16(const UChar *&s, const UChar *limit) const {
        UChar32 c=*s++;
        if(c<minDecompNoCP || !singleLeadMightHaveNonZeroFCD16(c)) {
            return 0;
        }
        UChar c2;
        if(U16_IS_LEAD(c) && s!=limit && U16_IS_TRAIL(c2=*s)) {
            c=U16_GET_SUPPLEMENTARY(c, c2);
            ++s;
        }
        return getFCD16FromNormData(c);
    }
    /**
     * Returns the FCD data for the previous code point (pre-decrement).
     * @param start The start of the string.
     * @param s A valid pointer into a string. Requires start<s.
     * @return The lccc(c) in bits 15..8 and tccc(c) in bits 7..0.
     */
    uint16_t previousFCD16(const UChar *start, const UChar *&s) const {
        UChar32 c=*--s;
        if(c<minDecompNoCP) {
            return 0;
        }
        if(!U16_IS_TRAIL(c)) {
            if(!singleLeadMightHaveNonZeroFCD16(c)) {
                return 0;
            }
        } else {
            UChar c2;
            if(start<s && U16_IS_LEAD(c2=*(s-1))) {
                c=U16_GET_SUPPLEMENTARY(c2, c);
                --s;
            }
        }
        return getFCD16FromNormData(c);
    }

    /** Returns TRUE if the single-or-lead code unit c might have non-zero FCD data. */
    UBool singleLeadMightHaveNonZeroFCD16(UChar32 lead) const {
        // 0<=lead<=0xffff
        uint8_t bits=smallFCD[lead>>8];
        if(bits==0) { return false; }
        return (UBool)((bits>>((lead>>5)&7))&1);
    }
    /** Returns the FCD value from the regular normalization data. */
    uint16_t getFCD16FromNormData(UChar32 c) const;

    /**
     * Gets the decomposition for one code point.
     * @param c code point
     * @param buffer out-only buffer for algorithmic decompositions
     * @param length out-only, takes the length of the decomposition, if any
     * @return pointer to the decomposition, or NULL if none
     */
    const UChar *getDecomposition(UChar32 c, UChar buffer[4], int32_t &length) const;

    /**
     * Gets the raw decomposition for one code point.
     * @param c code point
     * @param buffer out-only buffer for algorithmic decompositions
     * @param length out-only, takes the length of the decomposition, if any
     * @return pointer to the decomposition, or NULL if none
     */
    const UChar *getRawDecomposition(UChar32 c, UChar buffer[30], int32_t &length) const;

    UChar32 composePair(UChar32 a, UChar32 b) const;

    UBool isCanonSegmentStarter(UChar32 c) const;
    UBool getCanonStartSet(UChar32 c, UnicodeSet &set) const;

    enum {
        // Fixed norm16 values.
        MIN_YES_YES_WITH_CC=0xfe02,
        JAMO_VT=0xfe00,
        MIN_NORMAL_MAYBE_YES=0xfc00,
        JAMO_L=2,  // offset=1 hasCompBoundaryAfter=FALSE
        INERT=1,  // offset=0 hasCompBoundaryAfter=TRUE

        // norm16 bit 0 is comp-boundary-after.
        HAS_COMP_BOUNDARY_AFTER=1,
        OFFSET_SHIFT=1,

        // For algorithmic one-way mappings, norm16 bits 2..1 indicate the
        // tccc (0, 1, >1) for quick FCC boundary-after tests.
        DELTA_TCCC_0=0,
        DELTA_TCCC_1=2,
        DELTA_TCCC_GT_1=4,
        DELTA_TCCC_MASK=6,
        DELTA_SHIFT=3,

        MAX_DELTA=0x40
    };

    enum {
        // Byte offsets from the start of the data, after the generic header.
        IX_NORM_TRIE_OFFSET,
        IX_EXTRA_DATA_OFFSET,
        IX_SMALL_FCD_OFFSET,
        IX_RESERVED3_OFFSET,
        IX_RESERVED4_OFFSET,
        IX_RESERVED5_OFFSET,
        IX_RESERVED6_OFFSET,
        IX_TOTAL_SIZE,

        // Code point thresholds for quick check codes.
        IX_MIN_DECOMP_NO_CP,
        IX_MIN_COMP_NO_MAYBE_CP,

        // Norm16 value thresholds for quick check combinations and types of extra data.

        /** Mappings & compositions in [minYesNo..minYesNoMappingsOnly[. */
        IX_MIN_YES_NO,
        /** Mappings are comp-normalized. */
        IX_MIN_NO_NO,
        IX_LIMIT_NO_NO,
        IX_MIN_MAYBE_YES,

        /** Mappings only in [minYesNoMappingsOnly..minNoNo[. */
        IX_MIN_YES_NO_MAPPINGS_ONLY,
        /** Mappings are not comp-normalized but have a comp boundary before. */
        IX_MIN_NO_NO_COMP_BOUNDARY_BEFORE,
        /** Mappings do not have a comp boundary before. */
        IX_MIN_NO_NO_COMP_NO_MAYBE_CC,
        /** Mappings to the empty string. */
        IX_MIN_NO_NO_EMPTY,

        IX_MIN_LCCC_CP,
        IX_RESERVED19,
        IX_COUNT
    };

    enum {
        MAPPING_HAS_CCC_LCCC_WORD=0x80,
        MAPPING_HAS_RAW_MAPPING=0x40,
        // unused bit 0x20,
        MAPPING_LENGTH_MASK=0x1f
    };

    enum {
        COMP_1_LAST_TUPLE=0x8000,
        COMP_1_TRIPLE=1,
        COMP_1_TRAIL_LIMIT=0x3400,
        COMP_1_TRAIL_MASK=0x7ffe,
        COMP_1_TRAIL_SHIFT=9,  // 10-1 for the "triple" bit
        COMP_2_TRAIL_SHIFT=6,
        COMP_2_TRAIL_MASK=0xffc0
    };

    // higher-level functionality ------------------------------------------ ***

    // NFD without an NFD Normalizer2 instance.
    UnicodeString &decompose(const UnicodeString &src, UnicodeString &dest,
                             UErrorCode &errorCode) const;
    /**
     * Decomposes [src, limit[ and writes the result to dest.
     * limit can be NULL if src is NUL-terminated.
     * destLengthEstimate is the initial dest buffer capacity and can be -1.
     */
    void decompose(const UChar *src, const UChar *limit,
                   UnicodeString &dest, int32_t destLengthEstimate,
                   UErrorCode &errorCode) const;

    const UChar *decompose(const UChar *src, const UChar *limit,
                           ReorderingBuffer *buffer, UErrorCode &errorCode) const;
    void decomposeAndAppend(const UChar *src, const UChar *limit,
                            UBool doDecompose,
                            UnicodeString &safeMiddle,
                            ReorderingBuffer &buffer,
                            UErrorCode &errorCode) const;
    UBool compose(const UChar *src, const UChar *limit,
                  UBool onlyContiguous,
                  UBool doCompose,
                  ReorderingBuffer &buffer,
                  UErrorCode &errorCode) const;
    const UChar *composeQuickCheck(const UChar *src, const UChar *limit,
                                   UBool onlyContiguous,
                                   UNormalizationCheckResult *pQCResult) const;
    void composeAndAppend(const UChar *src, const UChar *limit,
                          UBool doCompose,
                          UBool onlyContiguous,
                          UnicodeString &safeMiddle,
                          ReorderingBuffer &buffer,
                          UErrorCode &errorCode) const;

    /** sink==nullptr: isNormalized() */
    UBool composeUTF8(uint32_t options, UBool onlyContiguous,
                      const uint8_t *src, const uint8_t *limit,
                      ByteSink *sink, icu::Edits *edits, UErrorCode &errorCode) const;

    const UChar *makeFCD(const UChar *src, const UChar *limit,
                         ReorderingBuffer *buffer, UErrorCode &errorCode) const;
    void makeFCDAndAppend(const UChar *src, const UChar *limit,
                          UBool doMakeFCD,
                          UnicodeString &safeMiddle,
                          ReorderingBuffer &buffer,
                          UErrorCode &errorCode) const;

    UBool hasDecompBoundaryBefore(UChar32 c) const;
    UBool norm16HasDecompBoundaryBefore(uint16_t norm16) const;
    UBool hasDecompBoundaryAfter(UChar32 c) const;
    UBool norm16HasDecompBoundaryAfter(uint16_t norm16) const;
    UBool isDecompInert(UChar32 c) const { return isDecompYesAndZeroCC(getNorm16(c)); }

    UBool hasCompBoundaryBefore(UChar32 c) const {
        return c<minCompNoMaybeCP || norm16HasCompBoundaryBefore(getNorm16(c));
    }
    UBool hasCompBoundaryAfter(UChar32 c, UBool onlyContiguous) const {
        return norm16HasCompBoundaryAfter(getNorm16(c), onlyContiguous);
    }
    UBool isCompInert(UChar32 c, UBool onlyContiguous) const {
        uint16_t norm16=getNorm16(c);
        return isCompYesAndZeroCC(norm16) &&
            (norm16 & HAS_COMP_BOUNDARY_AFTER) != 0 &&
            (!onlyContiguous || isInert(norm16) || *getMapping(norm16) <= 0x1ff);
    }

    UBool hasFCDBoundaryBefore(UChar32 c) const { return hasDecompBoundaryBefore(c); }
    UBool hasFCDBoundaryAfter(UChar32 c) const { return hasDecompBoundaryAfter(c); }
    UBool isFCDInert(UChar32 c) const { return getFCD16(c)<=1; }
private:
    friend class InitCanonIterData;
    friend class LcccContext;

    UBool isMaybe(uint16_t norm16) const { return minMaybeYes<=norm16 && norm16<=JAMO_VT; }
    UBool isMaybeOrNonZeroCC(uint16_t norm16) const { return norm16>=minMaybeYes; }
    static UBool isInert(uint16_t norm16) { return norm16==INERT; }
    static UBool isJamoL(uint16_t norm16) { return norm16==JAMO_L; }
    static UBool isJamoVT(uint16_t norm16) { return norm16==JAMO_VT; }
    uint16_t hangulLVT() const { return minYesNoMappingsOnly|HAS_COMP_BOUNDARY_AFTER; }
    UBool isHangulLV(uint16_t norm16) const { return norm16==minYesNo; }
    UBool isHangulLVT(uint16_t norm16) const {
        return norm16==hangulLVT();
    }
    UBool isCompYesAndZeroCC(uint16_t norm16) const { return norm16<minNoNo; }
    // UBool isCompYes(uint16_t norm16) const {
    //     return norm16>=MIN_YES_YES_WITH_CC || norm16<minNoNo;
    // }
    // UBool isCompYesOrMaybe(uint16_t norm16) const {
    //     return norm16<minNoNo || minMaybeYes<=norm16;
    // }
    // UBool hasZeroCCFromDecompYes(uint16_t norm16) const {
    //     return norm16<=MIN_NORMAL_MAYBE_YES || norm16==JAMO_VT;
    // }
    UBool isDecompYesAndZeroCC(uint16_t norm16) const {
        return norm16<minYesNo ||
               norm16==JAMO_VT ||
               (minMaybeYes<=norm16 && norm16<=MIN_NORMAL_MAYBE_YES);
    }
    /**
     * A little faster and simpler than isDecompYesAndZeroCC() but does not include
     * the MaybeYes which combine-forward and have ccc=0.
     * (Standard Unicode 10 normalization does not have such characters.)
     */
    UBool isMostDecompYesAndZeroCC(uint16_t norm16) const {
        return norm16<minYesNo || norm16==MIN_NORMAL_MAYBE_YES || norm16==JAMO_VT;
    }
    UBool isDecompNoAlgorithmic(uint16_t norm16) const { return norm16>=limitNoNo; }

    // For use with isCompYes().
    // Perhaps the compiler can combine the two tests for MIN_YES_YES_WITH_CC.
    // static uint8_t getCCFromYes(uint16_t norm16) {
    //     return norm16>=MIN_YES_YES_WITH_CC ? getCCFromNormalYesOrMaybe(norm16) : 0;
    // }
    uint8_t getCCFromNoNo(uint16_t norm16) const {
        const uint16_t *mapping=getMapping(norm16);
        if(*mapping&MAPPING_HAS_CCC_LCCC_WORD) {
            return (uint8_t)*(mapping-1);
        } else {
            return 0;
        }
    }
    // requires that the [cpStart..cpLimit[ character passes isCompYesAndZeroCC()
    uint8_t getTrailCCFromCompYesAndZeroCC(uint16_t norm16) const {
        if(norm16<=minYesNo) {
            return 0;  // yesYes and Hangul LV have ccc=tccc=0
        } else {
            // For Hangul LVT we harmlessly fetch a firstUnit with tccc=0 here.
            return (uint8_t)(*getMapping(norm16)>>8);  // tccc from yesNo
        }
    }
    uint8_t getPreviousTrailCC(const UChar *start, const UChar *p) const;
    uint8_t getPreviousTrailCC(const uint8_t *start, const uint8_t *p) const;

    // Requires algorithmic-NoNo.
    UChar32 mapAlgorithmic(UChar32 c, uint16_t norm16) const {
        return c+(norm16>>DELTA_SHIFT)-centerNoNoDelta;
    }
    UChar32 getAlgorithmicDelta(uint16_t norm16) const {
        return (norm16>>DELTA_SHIFT)-centerNoNoDelta;
    }

    // Requires minYesNo<norm16<limitNoNo.
    const uint16_t *getMapping(uint16_t norm16) const { return extraData+(norm16>>OFFSET_SHIFT); }
    const uint16_t *getCompositionsListForDecompYes(uint16_t norm16) const {
        if(norm16<JAMO_L || MIN_NORMAL_MAYBE_YES<=norm16) {
            return NULL;
        } else if(norm16<minMaybeYes) {
            return getMapping(norm16);  // for yesYes; if Jamo L: harmless empty list
        } else {
            return maybeYesCompositions+norm16-minMaybeYes;
        }
    }
    const uint16_t *getCompositionsListForComposite(uint16_t norm16) const {
        // A composite has both mapping & compositions list.
        const uint16_t *list=getMapping(norm16);
        return list+  // mapping pointer
            1+  // +1 to skip the first unit with the mapping length
            (*list&MAPPING_LENGTH_MASK);  // + mapping length
    }
    const uint16_t *getCompositionsListForMaybe(uint16_t norm16) const {
        // minMaybeYes<=norm16<MIN_NORMAL_MAYBE_YES
        return maybeYesCompositions+((norm16-minMaybeYes)>>OFFSET_SHIFT);
    }
    /**
     * @param c code point must have compositions
     * @return compositions list pointer
     */
    const uint16_t *getCompositionsList(uint16_t norm16) const {
        return isDecompYes(norm16) ?
                getCompositionsListForDecompYes(norm16) :
                getCompositionsListForComposite(norm16);
    }

    const UChar *copyLowPrefixFromNulTerminated(const UChar *src,
                                                UChar32 minNeedDataCP,
                                                ReorderingBuffer *buffer,
                                                UErrorCode &errorCode) const;
    const UChar *decomposeShort(const UChar *src, const UChar *limit,
                                UBool stopAtCompBoundary, UBool onlyContiguous,
                                ReorderingBuffer &buffer, UErrorCode &errorCode) const;
    UBool decompose(UChar32 c, uint16_t norm16,
                    ReorderingBuffer &buffer, UErrorCode &errorCode) const;

    const uint8_t *decomposeShort(const uint8_t *src, const uint8_t *limit,
                                  UBool stopAtCompBoundary, UBool onlyContiguous,
                                  ReorderingBuffer &buffer, UErrorCode &errorCode) const;

    static int32_t combine(const uint16_t *list, UChar32 trail);
    void addComposites(const uint16_t *list, UnicodeSet &set) const;
    void recompose(ReorderingBuffer &buffer, int32_t recomposeStartIndex,
                   UBool onlyContiguous) const;

    UBool hasCompBoundaryBefore(UChar32 c, uint16_t norm16) const {
        return c<minCompNoMaybeCP || norm16HasCompBoundaryBefore(norm16);
    }
    UBool norm16HasCompBoundaryBefore(uint16_t norm16) const  {
        return norm16 < minNoNoCompNoMaybeCC || isAlgorithmicNoNo(norm16);
    }
    UBool hasCompBoundaryBefore(const UChar *src, const UChar *limit) const;
    UBool hasCompBoundaryBefore(const uint8_t *src, const uint8_t *limit) const;
    UBool hasCompBoundaryAfter(const UChar *start, const UChar *p,
                               UBool onlyContiguous) const;
    UBool hasCompBoundaryAfter(const uint8_t *start, const uint8_t *p,
                               UBool onlyContiguous) const;
    UBool norm16HasCompBoundaryAfter(uint16_t norm16, UBool onlyContiguous) const {
        return (norm16 & HAS_COMP_BOUNDARY_AFTER) != 0 &&
            (!onlyContiguous || isTrailCC01ForCompBoundaryAfter(norm16));
    }
    /** For FCC: Given norm16 HAS_COMP_BOUNDARY_AFTER, does it have tccc<=1? */
    UBool isTrailCC01ForCompBoundaryAfter(uint16_t norm16) const {
        return isInert(norm16) || (isDecompNoAlgorithmic(norm16) ?
            (norm16 & DELTA_TCCC_MASK) <= DELTA_TCCC_1 : *getMapping(norm16) <= 0x1ff);
    }

    const UChar *findPreviousCompBoundary(const UChar *start, const UChar *p, UBool onlyContiguous) const;
    const UChar *findNextCompBoundary(const UChar *p, const UChar *limit, UBool onlyContiguous) const;

    const UChar *findPreviousFCDBoundary(const UChar *start, const UChar *p) const;
    const UChar *findNextFCDBoundary(const UChar *p, const UChar *limit) const;

    void makeCanonIterDataFromNorm16(UChar32 start, UChar32 end, const uint16_t norm16,
                                     CanonIterData &newData, UErrorCode &errorCode) const;

    int32_t getCanonValue(UChar32 c) const;
    const UnicodeSet &getCanonStartSet(int32_t n) const;

    // UVersionInfo dataVersion;

    // BMP code point thresholds for quick check loops looking at single UTF-16 code units.
    UChar minDecompNoCP;
    UChar minCompNoMaybeCP;
    UChar minLcccCP;

    // Norm16 value thresholds for quick check combinations and types of extra data.
    uint16_t minYesNo;
    uint16_t minYesNoMappingsOnly;
    uint16_t minNoNo;
    uint16_t minNoNoCompBoundaryBefore;
    uint16_t minNoNoCompNoMaybeCC;
    uint16_t minNoNoEmpty;
    uint16_t limitNoNo;
    uint16_t centerNoNoDelta;
    uint16_t minMaybeYes;

    const UCPTrie *normTrie;
    const uint16_t *maybeYesCompositions;
    const uint16_t *extraData;  // mappings and/or compositions for yesYes, yesNo & noNo characters
    const uint8_t *smallFCD;  // [0x100] one bit per 32 BMP code points, set if any FCD!=0

    UInitOnce       fCanonIterDataInitOnce = U_INITONCE_INITIALIZER;
    CanonIterData  *fCanonIterData;
};

// bits in canonIterData
#define CANON_NOT_SEGMENT_STARTER 0x80000000
#define CANON_HAS_COMPOSITIONS 0x40000000
#define CANON_HAS_SET 0x200000
#define CANON_VALUE_MASK 0x1fffff

/**
 * ICU-internal shortcut for quick access to standard Unicode normalization.
 */
class U_COMMON_API Normalizer2Factory {
public:
    static const Normalizer2 *getFCDInstance(UErrorCode &errorCode);
    static const Normalizer2 *getFCCInstance(UErrorCode &errorCode);
    static const Normalizer2 *getNoopInstance(UErrorCode &errorCode);

    static const Normalizer2 *getInstance(UNormalizationMode mode, UErrorCode &errorCode);

    static const Normalizer2Impl *getNFCImpl(UErrorCode &errorCode);
    static const Normalizer2Impl *getNFKCImpl(UErrorCode &errorCode);
    static const Normalizer2Impl *getNFKC_CFImpl(UErrorCode &errorCode);

    // Get the Impl instance of the Normalizer2.
    // Must be used only when it is known that norm2 is a Normalizer2WithImpl instance.
    static const Normalizer2Impl *getImpl(const Normalizer2 *norm2);
private:
    Normalizer2Factory();  // No instantiation.
};

U_NAMESPACE_END

U_CAPI int32_t U_EXPORT2
unorm2_swap(const UDataSwapper *ds,
            const void *inData, int32_t length, void *outData,
            UErrorCode *pErrorCode);

/**
 * Get the NF*_QC property for a code point, for u_getIntPropertyValue().
 * @internal
 */
U_CFUNC UNormalizationCheckResult
unorm_getQuickCheck(UChar32 c, UNormalizationMode mode);

/**
 * Gets the 16-bit FCD value (lead & trail CCs) for a code point, for u_getIntPropertyValue().
 * @internal
 */
U_CFUNC uint16_t
unorm_getFCD16(UChar32 c);

/**
 * Format of Normalizer2 .nrm data files.
 * Format version 4.0.
 *
 * Normalizer2 .nrm data files provide data for the Unicode Normalization algorithms.
 * ICU ships with data files for standard Unicode Normalization Forms
 * NFC and NFD (nfc.nrm), NFKC and NFKD (nfkc.nrm) and NFKC_Casefold (nfkc_cf.nrm).
 * Custom (application-specific) data can be built into additional .nrm files
 * with the gennorm2 build tool.
 * ICU ships with one such file, uts46.nrm, for the implementation of UTS #46.
 *
 * Normalizer2.getInstance() causes a .nrm file to be loaded, unless it has been
 * cached already. Internally, Normalizer2Impl.load() reads the .nrm file.
 *
 * A .nrm file begins with a standard ICU data file header
 * (DataHeader, see ucmndata.h and unicode/udata.h).
 * The UDataInfo.dataVersion field usually contains the Unicode version
 * for which the data was generated.
 *
 * After the header, the file contains the following parts.
 * Constants are defined as enum values of the Normalizer2Impl class.
 *
 * Many details of the data structures are described in the design doc
 * which is at http://site.icu-project.org/design/normalization/custom
 *
 * int32_t indexes[indexesLength]; -- indexesLength=indexes[IX_NORM_TRIE_OFFSET]/4;
 *
 *      The first eight indexes are byte offsets in ascending order.
 *      Each byte offset marks the start of the next part in the data file,
 *      and the end of the previous one.
 *      When two consecutive byte offsets are the same, then the corresponding part is empty.
 *      Byte offsets are offsets from after the header,
 *      that is, from the beginning of the indexes[].
 *      Each part starts at an offset with proper alignment for its data.
 *      If necessary, the previous part may include padding bytes to achieve this alignment.
 *
 *      minDecompNoCP=indexes[IX_MIN_DECOMP_NO_CP] is the lowest code point
 *      with a decomposition mapping, that is, with NF*D_QC=No.
 *      minCompNoMaybeCP=indexes[IX_MIN_COMP_NO_MAYBE_CP] is the lowest code point
 *      with NF*C_QC=No (has a one-way mapping) or Maybe (combines backward).
 *      minLcccCP=indexes[IX_MIN_LCCC_CP] (index 18, new in formatVersion 3)
 *      is the lowest code point with lccc!=0.
 *
 *      The next eight indexes are thresholds of 16-bit trie values for ranges of
 *      values indicating multiple normalization properties.
 *      They are listed here in threshold order, not in the order they are stored in the indexes.
 *          minYesNo=indexes[IX_MIN_YES_NO];
 *          minYesNoMappingsOnly=indexes[IX_MIN_YES_NO_MAPPINGS_ONLY];
 *          minNoNo=indexes[IX_MIN_NO_NO];
 *          minNoNoCompBoundaryBefore=indexes[IX_MIN_NO_NO_COMP_BOUNDARY_BEFORE];
 *          minNoNoCompNoMaybeCC=indexes[IX_MIN_NO_NO_COMP_NO_MAYBE_CC];
 *          minNoNoEmpty=indexes[IX_MIN_NO_NO_EMPTY];
 *          limitNoNo=indexes[IX_LIMIT_NO_NO];
 *          minMaybeYes=indexes[IX_MIN_MAYBE_YES];
 *      See the normTrie description below and the design doc for details.
 *
 * UCPTrie normTrie; -- see ucptrie_impl.h and ucptrie.h, same as Java CodePointTrie
 *
 *      The trie holds the main normalization data. Each code point is mapped to a 16-bit value.
 *      Rather than using independent bits in the value (which would require more than 16 bits),
 *      information is extracted primarily via range checks.
 *      Except, format version 3 uses bit 0 for hasCompBoundaryAfter().
 *      For example, a 16-bit value norm16 in the range minYesNo<=norm16<minNoNo
 *      means that the character has NF*C_QC=Yes and NF*D_QC=No properties,
 *      which means it has a two-way (round-trip) decomposition mapping.
 *      Values in the range 2<=norm16<limitNoNo are also directly indexes into the extraData
 *      pointing to mappings, compositions lists, or both.
 *      Value norm16==INERT (0 in versions 1 & 2, 1 in version 3)
 *      means that the character is normalization-inert, that is,
 *      it does not have a mapping, does not participate in composition, has a zero
 *      canonical combining class, and forms a boundary where text before it and after it
 *      can be normalized independently.
 *      For details about how multiple properties are encoded in 16-bit values
 *      see the design doc.
 *      Note that the encoding cannot express all combinations of the properties involved;
 *      it only supports those combinations that are allowed by
 *      the Unicode Normalization algorithms. Details are in the design doc as well.
 *      The gennorm2 tool only builds .nrm files for data that conforms to the limitations.
 *
 *      The trie has a value for each lead surrogate code unit representing the "worst case"
 *      properties of the 1024 supplementary characters whose UTF-16 form starts with
 *      the lead surrogate. If all of the 1024 supplementary characters are normalization-inert,
 *      then their lead surrogate code unit has the trie value INERT.
 *      When the lead surrogate unit's value exceeds the quick check minimum during processing,
 *      the properties for the full supplementary code point need to be looked up.
 *
 * uint16_t maybeYesCompositions[MIN_NORMAL_MAYBE_YES-minMaybeYes];
 * uint16_t extraData[];
 *
 *      There is only one byte offset for the end of these two arrays.
 *      The split between them is given by the constant and variable mentioned above.
 *      In version 3, the difference must be shifted right by OFFSET_SHIFT.
 *
 *      The maybeYesCompositions array contains compositions lists for characters that
 *      combine both forward (as starters in composition pairs)
 *      and backward (as trailing characters in composition pairs).
 *      Such characters do not occur in Unicode 5.2 but are allowed by
 *      the Unicode Normalization algorithms.
 *      If there are no such characters, then minMaybeYes==MIN_NORMAL_MAYBE_YES
 *      and the maybeYesCompositions array is empty.
 *      If there are such characters, then minMaybeYes is subtracted from their norm16 values
 *      to get the index into this array.
 *
 *      The extraData array contains compositions lists for "YesYes" characters,
 *      followed by mappings and optional compositions lists for "YesNo" characters,
 *      followed by only mappings for "NoNo" characters.
 *      (Referring to pairs of NFC/NFD quick check values.)
 *      The norm16 values of those characters are directly indexes into the extraData array.
 *      In version 3, the norm16 values must be shifted right by OFFSET_SHIFT
 *      for accessing extraData.
 *
 *      The data structures for compositions lists and mappings are described in the design doc.
 *
 * uint8_t smallFCD[0x100]; -- new in format version 2
 *
 *      This is a bit set to help speed up FCD value lookups in the absence of a full
 *      UTrie2 or other large data structure with the full FCD value mapping.
 *
 *      Each smallFCD bit is set if any of the corresponding 32 BMP code points
 *      has a non-zero FCD value (lccc!=0 or tccc!=0).
 *      Bit 0 of smallFCD[0] is for U+0000..U+001F. Bit 7 of smallFCD[0xff] is for U+FFE0..U+FFFF.
 *      A bit for 32 lead surrogates is set if any of the 32k corresponding
 *      _supplementary_ code points has a non-zero FCD value.
 *
 *      This bit set is most useful for the large blocks of CJK characters with FCD=0.
 *
 * Changes from format version 1 to format version 2 ---------------------------
 *
 * - Addition of data for raw (not recursively decomposed) mappings.
 *   + The MAPPING_NO_COMP_BOUNDARY_AFTER bit in the extraData is now also set when
 *     the mapping is to an empty string or when the character combines-forward.
 *     This subsumes the one actual use of the MAPPING_PLUS_COMPOSITION_LIST bit which
 *     is then repurposed for the MAPPING_HAS_RAW_MAPPING bit.
 *   + For details see the design doc.
 * - Addition of indexes[IX_MIN_YES_NO_MAPPINGS_ONLY] and separation of the yesNo extraData into
 *   distinct ranges (combines-forward vs. not)
 *   so that a range check can be used to find out if there is a compositions list.
 *   This is fully equivalent with formatVersion 1's MAPPING_PLUS_COMPOSITION_LIST flag.
 *   It is needed for the new (in ICU 49) composePair(), not for other normalization.
 * - Addition of the smallFCD[] bit set.
 *
 * Changes from format version 2 to format version 3 (ICU 60) ------------------
 *
 * - norm16 bit 0 indicates hasCompBoundaryAfter(),
 *   except that for contiguous composition (FCC) the tccc must be checked as well.
 *   Data indexes and ccc values are shifted left by one (OFFSET_SHIFT).
 *   Thresholds like minNoNo are tested before shifting.
 *
 * - Algorithmic mapping deltas are shifted left by two more bits (total DELTA_SHIFT),
 *   to make room for two bits (three values) indicating whether the tccc is 0, 1, or greater.
 *   See DELTA_TCCC_MASK etc.
 *   This helps with fetching tccc/FCD values and FCC hasCompBoundaryAfter().
 *   minMaybeYes is 8-aligned so that the DELTA_TCCC_MASK bits can be tested directly.
 *
 * - Algorithmic mappings are only used for mapping to "comp yes and ccc=0" characters,
 *   and ASCII characters are mapped algorithmically only to other ASCII characters.
 *   This helps with hasCompBoundaryBefore() and compose() fast paths.
 *   It is never necessary any more to loop for algorithmic mappings.
 *
 * - Addition of indexes[IX_MIN_NO_NO_COMP_BOUNDARY_BEFORE],
 *   indexes[IX_MIN_NO_NO_COMP_NO_MAYBE_CC], and indexes[IX_MIN_NO_NO_EMPTY],
 *   and separation of the noNo extraData into distinct ranges.
 *   With this, the noNo norm16 value indicates whether the mapping is
 *   compose-normalized, not normalized but hasCompBoundaryBefore(),
 *   not even that, or maps to an empty string.
 *   hasCompBoundaryBefore() can be determined solely from the norm16 value.
 *
 * - The norm16 value for Hangul LVT is now different from that for Hangul LV,
 *   so that hasCompBoundaryAfter() need not check for the syllable type.
 *   For Hangul LV, minYesNo continues to be used (no comp-boundary-after).
 *   For Hangul LVT, minYesNoMappingsOnly|HAS_COMP_BOUNDARY_AFTER is used.
 *   The extraData units at these indexes are set to firstUnit=2 and firstUnit=3, respectively,
 *   to simplify some code.
 *
 * - The extraData firstUnit bit 5 is no longer necessary
 *   (norm16 bit 0 used instead of firstUnit MAPPING_NO_COMP_BOUNDARY_AFTER),
 *   is reserved again, and always set to 0.
 *
 * - Addition of indexes[IX_MIN_LCCC_CP], the first code point where lccc!=0.
 *   This used to be hardcoded to U+0300, but in data like NFKC_Casefold it is lower:
 *   U+00AD Soft Hyphen maps to an empty string,
 *   which is artificially assigned "worst case" values lccc=1 and tccc=255.
 *
 * - A mapping to an empty string has explicit lccc=1 and tccc=255 values.
 *
 * Changes from format version 3 to format version 4 (ICU 63) ------------------
 *
 * Switched from UTrie2 to UCPTrie/CodePointTrie.
 *
 * The new trie no longer stores different values for surrogate code *units* vs.
 * surrogate code *points*.
 * Lead surrogates still have values for optimized UTF-16 string processing.
 * When looking up code point properties, the code now checks for lead surrogates and
 * treats them as inert.
 *
 * gennorm2 now has to reject mappings for surrogate code points.
 * UTS #46 maps unpaired surrogates to U+FFFD in code rather than via its
 * custom normalization data file.
 */

#endif  /* !UCONFIG_NO_NORMALIZATION */


/**
 * This class allows one to iterate through all the strings that are canonically equivalent to a given
 * string. For example, here are some sample results:
Results for: {LATIN CAPITAL LETTER A WITH RING ABOVE}{LATIN SMALL LETTER D}{COMBINING DOT ABOVE}{COMBINING CEDILLA}
1: \u0041\u030A\u0064\u0307\u0327
 = {LATIN CAPITAL LETTER A}{COMBINING RING ABOVE}{LATIN SMALL LETTER D}{COMBINING DOT ABOVE}{COMBINING CEDILLA}
2: \u0041\u030A\u0064\u0327\u0307
 = {LATIN CAPITAL LETTER A}{COMBINING RING ABOVE}{LATIN SMALL LETTER D}{COMBINING CEDILLA}{COMBINING DOT ABOVE}
3: \u0041\u030A\u1E0B\u0327
 = {LATIN CAPITAL LETTER A}{COMBINING RING ABOVE}{LATIN SMALL LETTER D WITH DOT ABOVE}{COMBINING CEDILLA}
4: \u0041\u030A\u1E11\u0307
 = {LATIN CAPITAL LETTER A}{COMBINING RING ABOVE}{LATIN SMALL LETTER D WITH CEDILLA}{COMBINING DOT ABOVE}
5: \u00C5\u0064\u0307\u0327
 = {LATIN CAPITAL LETTER A WITH RING ABOVE}{LATIN SMALL LETTER D}{COMBINING DOT ABOVE}{COMBINING CEDILLA}
6: \u00C5\u0064\u0327\u0307
 = {LATIN CAPITAL LETTER A WITH RING ABOVE}{LATIN SMALL LETTER D}{COMBINING CEDILLA}{COMBINING DOT ABOVE}
7: \u00C5\u1E0B\u0327
 = {LATIN CAPITAL LETTER A WITH RING ABOVE}{LATIN SMALL LETTER D WITH DOT ABOVE}{COMBINING CEDILLA}
8: \u00C5\u1E11\u0307
 = {LATIN CAPITAL LETTER A WITH RING ABOVE}{LATIN SMALL LETTER D WITH CEDILLA}{COMBINING DOT ABOVE}
9: \u212B\u0064\u0307\u0327
 = {ANGSTROM SIGN}{LATIN SMALL LETTER D}{COMBINING DOT ABOVE}{COMBINING CEDILLA}
10: \u212B\u0064\u0327\u0307
 = {ANGSTROM SIGN}{LATIN SMALL LETTER D}{COMBINING CEDILLA}{COMBINING DOT ABOVE}
11: \u212B\u1E0B\u0327
 = {ANGSTROM SIGN}{LATIN SMALL LETTER D WITH DOT ABOVE}{COMBINING CEDILLA}
12: \u212B\u1E11\u0307
 = {ANGSTROM SIGN}{LATIN SMALL LETTER D WITH CEDILLA}{COMBINING DOT ABOVE}
 *<br>Note: the code is intended for use with small strings, and is not suitable for larger ones,
 * since it has not been optimized for that situation.
 *@author M. Davis
 *@draft
 */

// public

U_NAMESPACE_BEGIN

// TODO: add boilerplate methods.

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(CanonicalIterator)

/**
 *@param source string to get results for
 */
CanonicalIterator::CanonicalIterator(const UnicodeString &sourceStr, UErrorCode &status) :
    pieces(NULL),
    pieces_length(0),
    pieces_lengths(NULL),
    current(NULL),
    current_length(0),
    nfd(*Normalizer2::getNFDInstance(status)),
    nfcImpl(*Normalizer2Factory::getNFCImpl(status))
{
    if(U_SUCCESS(status) && nfcImpl.ensureCanonIterData(status)) {
      setSource(sourceStr, status);
    }
}

CanonicalIterator::~CanonicalIterator() {
  cleanPieces();
}

void CanonicalIterator::cleanPieces() {
    int32_t i = 0;
    if(pieces != NULL) {
        for(i = 0; i < pieces_length; i++) {
            if(pieces[i] != NULL) {
                delete[] pieces[i];
            }
        }
        uprv_free(pieces);
        pieces = NULL;
        pieces_length = 0;
    }
    if(pieces_lengths != NULL) {
        uprv_free(pieces_lengths);
        pieces_lengths = NULL;
    }
    if(current != NULL) {
        uprv_free(current);
        current = NULL;
        current_length = 0;
    }
}

/**
 *@return gets the source: NOTE: it is the NFD form of source
 */
UnicodeString CanonicalIterator::getSource() {
  return source;
}

/**
 * Resets the iterator so that one can start again from the beginning.
 */
void CanonicalIterator::reset() {
    done = FALSE;
    for (int i = 0; i < current_length; ++i) {
        current[i] = 0;
    }
}

/**
 *@return the next string that is canonically equivalent. The value null is returned when
 * the iteration is done.
 */
UnicodeString CanonicalIterator::next() {
    int32_t i = 0;

    if (done) {
      buffer.setToBogus();
      return buffer;
    }

    // delete old contents
    buffer.remove();

    // construct return value

    for (i = 0; i < pieces_length; ++i) {
        buffer.append(pieces[i][current[i]]);
    }
    //String result = buffer.toString(); // not needed

    // find next value for next time

    for (i = current_length - 1; ; --i) {
        if (i < 0) {
            done = TRUE;
            break;
        }
        current[i]++;
        if (current[i] < pieces_lengths[i]) break; // got sequence
        current[i] = 0;
    }
    return buffer;
}

/**
 *@param set the source string to iterate against. This allows the same iterator to be used
 * while changing the source string, saving object creation.
 */
void CanonicalIterator::setSource(const UnicodeString &newSource, UErrorCode &status) {
    int32_t list_length = 0;
    UChar32 cp = 0;
    int32_t start = 0;
    int32_t i = 0;
    UnicodeString *list = NULL;

    nfd.normalize(newSource, source, status);
    if(U_FAILURE(status)) {
      return;
    }
    done = FALSE;

    cleanPieces();

    // catch degenerate case
    if (newSource.length() == 0) {
        pieces = (UnicodeString **)uprv_malloc(sizeof(UnicodeString *));
        pieces_lengths = (int32_t*)uprv_malloc(1 * sizeof(int32_t));
        pieces_length = 1;
        current = (int32_t*)uprv_malloc(1 * sizeof(int32_t));
        current_length = 1;
        if (pieces == NULL || pieces_lengths == NULL || current == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            goto CleanPartialInitialization;
        }
        current[0] = 0;
        pieces[0] = new UnicodeString[1];
        pieces_lengths[0] = 1;
        if (pieces[0] == 0) {
            status = U_MEMORY_ALLOCATION_ERROR;
            goto CleanPartialInitialization;
        }
        return;
    }


    list = new UnicodeString[source.length()];
    if (list == 0) {
        status = U_MEMORY_ALLOCATION_ERROR;
        goto CleanPartialInitialization;
    }

    // i should initialy be the number of code units at the 
    // start of the string
    i = U16_LENGTH(source.char32At(0));
    //int32_t i = 1;
    // find the segments
    // This code iterates through the source string and 
    // extracts segments that end up on a codepoint that
    // doesn't start any decompositions. (Analysis is done
    // on the NFD form - see above).
    for (; i < source.length(); i += U16_LENGTH(cp)) {
        cp = source.char32At(i);
        if (nfcImpl.isCanonSegmentStarter(cp)) {
            source.extract(start, i-start, list[list_length++]); // add up to i
            start = i;
        }
    }
    source.extract(start, i-start, list[list_length++]); // add last one


    // allocate the arrays, and find the strings that are CE to each segment
    pieces = (UnicodeString **)uprv_malloc(list_length * sizeof(UnicodeString *));
    pieces_length = list_length;
    pieces_lengths = (int32_t*)uprv_malloc(list_length * sizeof(int32_t));
    current = (int32_t*)uprv_malloc(list_length * sizeof(int32_t));
    current_length = list_length;
    if (pieces == NULL || pieces_lengths == NULL || current == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        goto CleanPartialInitialization;
    }

    for (i = 0; i < current_length; i++) {
        current[i] = 0;
    }
    // for each segment, get all the combinations that can produce 
    // it after NFD normalization
    for (i = 0; i < pieces_length; ++i) {
        //if (PROGRESS) printf("SEGMENT\n");
        pieces[i] = getEquivalents(list[i], pieces_lengths[i], status);
    }

    delete[] list;
    return;
// Common section to cleanup all local variables and reset object variables.
CleanPartialInitialization:
    if (list != NULL) {
        delete[] list;
    }
    cleanPieces();
}

/**
 * Dumb recursive implementation of permutation.
 * TODO: optimize
 * @param source the string to find permutations for
 * @return the results in a set.
 */
void U_EXPORT2 CanonicalIterator::permute(UnicodeString &source, UBool skipZeros, Hashtable *result, UErrorCode &status) {
    if(U_FAILURE(status)) {
        return;
    }
    //if (PROGRESS) printf("Permute: %s\n", UToS(Tr(source)));
    int32_t i = 0;

    // optimization:
    // if zero or one character, just return a set with it
    // we check for length < 2 to keep from counting code points all the time
    if (source.length() <= 2 && source.countChar32() <= 1) {
        UnicodeString *toPut = new UnicodeString(source);
        /* test for NULL */
        if (toPut == 0) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        result->put(source, toPut, status);
        return;
    }

    // otherwise iterate through the string, and recursively permute all the other characters
    UChar32 cp;
    Hashtable subpermute(status);
    if(U_FAILURE(status)) {
        return;
    }
    subpermute.setValueDeleter(uprv_deleteUObject);

    for (i = 0; i < source.length(); i += U16_LENGTH(cp)) {
        cp = source.char32At(i);
        const UHashElement *ne = NULL;
        int32_t el = UHASH_FIRST;
        UnicodeString subPermuteString = source;

        // optimization:
        // if the character is canonical combining class zero,
        // don't permute it
        if (skipZeros && i != 0 && u_getCombiningClass(cp) == 0) {
            //System.out.println("Skipping " + Utility.hex(UTF16.valueOf(source, i)));
            continue;
        }

        subpermute.removeAll();

        // see what the permutations of the characters before and after this one are
        //Hashtable *subpermute = permute(source.substring(0,i) + source.substring(i + UTF16.getCharCount(cp)));
        permute(subPermuteString.remove(i, U16_LENGTH(cp)), skipZeros, &subpermute, status);
        /* Test for buffer overflows */
        if(U_FAILURE(status)) {
            return;
        }
        // The upper remove is destructive. The question is do we have to make a copy, or we don't care about the contents 
        // of source at this point.

        // prefix this character to all of them
        ne = subpermute.nextElement(el);
        while (ne != NULL) {
            UnicodeString *permRes = (UnicodeString *)(ne->value.pointer);
            UnicodeString *chStr = new UnicodeString(cp);
            //test for  NULL
            if (chStr == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            chStr->append(*permRes); //*((UnicodeString *)(ne->value.pointer));
            //if (PROGRESS) printf("  Piece: %s\n", UToS(*chStr));
            result->put(*chStr, chStr, status);
            ne = subpermute.nextElement(el);
        }
    }
    //return result;
}

// privates

// we have a segment, in NFD. Find all the strings that are canonically equivalent to it.
UnicodeString* CanonicalIterator::getEquivalents(const UnicodeString &segment, int32_t &result_len, UErrorCode &status) {
    Hashtable result(status);
    Hashtable permutations(status);
    Hashtable basic(status);
    if (U_FAILURE(status)) {
        return 0;
    }
    result.setValueDeleter(uprv_deleteUObject);
    permutations.setValueDeleter(uprv_deleteUObject);
    basic.setValueDeleter(uprv_deleteUObject);

    UChar USeg[256];
    int32_t segLen = segment.extract(USeg, 256, status);
    getEquivalents2(&basic, USeg, segLen, status);

    // now get all the permutations
    // add only the ones that are canonically equivalent
    // TODO: optimize by not permuting any class zero.

    const UHashElement *ne = NULL;
    int32_t el = UHASH_FIRST;
    //Iterator it = basic.iterator();
    ne = basic.nextElement(el);
    //while (it.hasNext())
    while (ne != NULL) {
        //String item = (String) it.next();
        UnicodeString item = *((UnicodeString *)(ne->value.pointer));

        permutations.removeAll();
        permute(item, CANITER_SKIP_ZEROES, &permutations, status);
        const UHashElement *ne2 = NULL;
        int32_t el2 = UHASH_FIRST;
        //Iterator it2 = permutations.iterator();
        ne2 = permutations.nextElement(el2);
        //while (it2.hasNext())
        while (ne2 != NULL) {
            //String possible = (String) it2.next();
            //UnicodeString *possible = new UnicodeString(*((UnicodeString *)(ne2->value.pointer)));
            UnicodeString possible(*((UnicodeString *)(ne2->value.pointer)));
            UnicodeString attempt;
            nfd.normalize(possible, attempt, status);

            // TODO: check if operator == is semanticaly the same as attempt.equals(segment)
            if (attempt==segment) {
                //if (PROGRESS) printf("Adding Permutation: %s\n", UToS(Tr(*possible)));
                // TODO: use the hashtable just to catch duplicates - store strings directly (somehow).
                result.put(possible, new UnicodeString(possible), status); //add(possible);
            } else {
                //if (PROGRESS) printf("-Skipping Permutation: %s\n", UToS(Tr(*possible)));
            }

            ne2 = permutations.nextElement(el2);
        }
        ne = basic.nextElement(el);
    }

    /* Test for buffer overflows */
    if(U_FAILURE(status)) {
        return 0;
    }
    // convert into a String[] to clean up storage
    //String[] finalResult = new String[result.size()];
    UnicodeString *finalResult = NULL;
    int32_t resultCount;
    if((resultCount = result.count()) != 0) {
        finalResult = new UnicodeString[resultCount];
        if (finalResult == 0) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
    }
    else {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    //result.toArray(finalResult);
    result_len = 0;
    el = UHASH_FIRST;
    ne = result.nextElement(el);
    while(ne != NULL) {
        finalResult[result_len++] = *((UnicodeString *)(ne->value.pointer));
        ne = result.nextElement(el);
    }


    return finalResult;
}

Hashtable *CanonicalIterator::getEquivalents2(Hashtable *fillinResult, const UChar *segment, int32_t segLen, UErrorCode &status) {

    if (U_FAILURE(status)) {
        return NULL;
    }

    //if (PROGRESS) printf("Adding: %s\n", UToS(Tr(segment)));

    UnicodeString toPut(segment, segLen);

    fillinResult->put(toPut, new UnicodeString(toPut), status);

    UnicodeSet starts;

    // cycle through all the characters
    UChar32 cp;
    for (int32_t i = 0; i < segLen; i += U16_LENGTH(cp)) {
        // see if any character is at the start of some decomposition
        U16_GET(segment, 0, i, segLen, cp);
        if (!nfcImpl.getCanonStartSet(cp, starts)) {
            continue;
        }
        // if so, see which decompositions match
        UnicodeSetIterator iter(starts);
        while (iter.next()) {
            UChar32 cp2 = iter.getCodepoint();
            Hashtable remainder(status);
            remainder.setValueDeleter(uprv_deleteUObject);
            if (extract(&remainder, cp2, segment, segLen, i, status) == NULL) {
                continue;
            }

            // there were some matches, so add all the possibilities to the set.
            UnicodeString prefix(segment, i);
            prefix += cp2;

            int32_t el = UHASH_FIRST;
            const UHashElement *ne = remainder.nextElement(el);
            while (ne != NULL) {
                UnicodeString item = *((UnicodeString *)(ne->value.pointer));
                UnicodeString *toAdd = new UnicodeString(prefix);
                /* test for NULL */
                if (toAdd == 0) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                    return NULL;
                }
                *toAdd += item;
                fillinResult->put(*toAdd, toAdd, status);

                //if (PROGRESS) printf("Adding: %s\n", UToS(Tr(*toAdd)));

                ne = remainder.nextElement(el);
            }
        }
    }

    /* Test for buffer overflows */
    if(U_FAILURE(status)) {
        return NULL;
    }
    return fillinResult;
}

/**
 * See if the decomposition of cp2 is at segment starting at segmentPos 
 * (with canonical rearrangment!)
 * If so, take the remainder, and return the equivalents 
 */
Hashtable *CanonicalIterator::extract(Hashtable *fillinResult, UChar32 comp, const UChar *segment, int32_t segLen, int32_t segmentPos, UErrorCode &status) {
//Hashtable *CanonicalIterator::extract(UChar32 comp, const UnicodeString &segment, int32_t segLen, int32_t segmentPos, UErrorCode &status) {
    //if (PROGRESS) printf(" extract: %s, ", UToS(Tr(UnicodeString(comp))));
    //if (PROGRESS) printf("%s, %i\n", UToS(Tr(segment)), segmentPos);

    if (U_FAILURE(status)) {
        return NULL;
    }

    UnicodeString temp(comp);
    int32_t inputLen=temp.length();
    UnicodeString decompString;
    nfd.normalize(temp, decompString, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    if (decompString.isBogus()) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    const UChar *decomp=decompString.getBuffer();
    int32_t decompLen=decompString.length();

    // See if it matches the start of segment (at segmentPos)
    UBool ok = FALSE;
    UChar32 cp;
    int32_t decompPos = 0;
    UChar32 decompCp;
    U16_NEXT(decomp, decompPos, decompLen, decompCp);

    int32_t i = segmentPos;
    while(i < segLen) {
        U16_NEXT(segment, i, segLen, cp);

        if (cp == decompCp) { // if equal, eat another cp from decomp

            //if (PROGRESS) printf("  matches: %s\n", UToS(Tr(UnicodeString(cp))));

            if (decompPos == decompLen) { // done, have all decomp characters!
                temp.append(segment+i, segLen-i);
                ok = TRUE;
                break;
            }
            U16_NEXT(decomp, decompPos, decompLen, decompCp);
        } else {
            //if (PROGRESS) printf("  buffer: %s\n", UToS(Tr(UnicodeString(cp))));

            // brute force approach
            temp.append(cp);

            /* TODO: optimize
            // since we know that the classes are monotonically increasing, after zero
            // e.g. 0 5 7 9 0 3
            // we can do an optimization
            // there are only a few cases that work: zero, less, same, greater
            // if both classes are the same, we fail
            // if the decomp class < the segment class, we fail

            segClass = getClass(cp);
            if (decompClass <= segClass) return null;
            */
        }
    }
    if (!ok)
        return NULL; // we failed, characters left over

    //if (PROGRESS) printf("Matches\n");

    if (inputLen == temp.length()) {
        fillinResult->put(UnicodeString(), new UnicodeString(), status);
        return fillinResult; // succeed, but no remainder
    }

    // brute force approach
    // check to make sure result is canonically equivalent
    UnicodeString trial;
    nfd.normalize(temp, trial, status);
    if(U_FAILURE(status) || trial.compare(segment+segmentPos, segLen - segmentPos) != 0) {
        return NULL;
    }

    return getEquivalents2(fillinResult, temp.getBuffer()+inputLen, temp.length()-inputLen, status);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_NORMALIZATION */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// characterproperties.cpp
// created: 2018sep03 Markus W. Scherer






// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// umutablecptrie.h (split out of ucptrie.h)
// created: 2018jan24 Markus W. Scherer









U_CDECL_BEGIN

/**
 * \file
 *
 * This file defines a mutable Unicode code point trie.
 *
 * @see UCPTrie
 * @see UMutableCPTrie
 */

/**
 * Mutable Unicode code point trie.
 * Fast map from Unicode code points (U+0000..U+10FFFF) to 32-bit integer values.
 * For details see http://site.icu-project.org/design/struct/utrie
 *
 * Setting values (especially ranges) and lookup is fast.
 * The mutable trie is only somewhat space-efficient.
 * It builds a compacted, immutable UCPTrie.
 *
 * This trie can be modified while iterating over its contents.
 * For example, it is possible to merge its values with those from another
 * set of ranges (e.g., another mutable or immutable trie):
 * Iterate over those source ranges; for each of them iterate over this trie;
 * add the source value into the value of each trie range.
 *
 * @see UCPTrie
 * @see umutablecptrie_buildImmutable
 * @stable ICU 63
 */
typedef struct UMutableCPTrie UMutableCPTrie;

/**
 * Creates a mutable trie that initially maps each Unicode code point to the same value.
 * It uses 32-bit data values until umutablecptrie_buildImmutable() is called.
 * umutablecptrie_buildImmutable() takes a valueWidth parameter which
 * determines the number of bits in the data value in the resulting UCPTrie.
 * You must umutablecptrie_close() the trie once you are done using it.
 *
 * @param initialValue the initial value that is set for all code points
 * @param errorValue the value for out-of-range code points and ill-formed UTF-8/16
 * @param pErrorCode an in/out ICU UErrorCode
 * @return the trie
 * @stable ICU 63
 */
U_CAPI UMutableCPTrie * U_EXPORT2
umutablecptrie_open(uint32_t initialValue, uint32_t errorValue, UErrorCode *pErrorCode);

/**
 * Clones a mutable trie.
 * You must umutablecptrie_close() the clone once you are done using it.
 *
 * @param other the trie to clone
 * @param pErrorCode an in/out ICU UErrorCode
 * @return the trie clone
 * @stable ICU 63
 */
U_CAPI UMutableCPTrie * U_EXPORT2
umutablecptrie_clone(const UMutableCPTrie *other, UErrorCode *pErrorCode);

/**
 * Closes a mutable trie and releases associated memory.
 *
 * @param trie the trie
 * @stable ICU 63
 */
U_CAPI void U_EXPORT2
umutablecptrie_close(UMutableCPTrie *trie);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUMutableCPTriePointer
 * "Smart pointer" class, closes a UMutableCPTrie via umutablecptrie_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 63
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUMutableCPTriePointer, UMutableCPTrie, umutablecptrie_close);

U_NAMESPACE_END

#endif

/**
 * Creates a mutable trie with the same contents as the UCPMap.
 * You must umutablecptrie_close() the mutable trie once you are done using it.
 *
 * @param map the source map
 * @param pErrorCode an in/out ICU UErrorCode
 * @return the mutable trie
 * @stable ICU 63
 */
U_CAPI UMutableCPTrie * U_EXPORT2
umutablecptrie_fromUCPMap(const UCPMap *map, UErrorCode *pErrorCode);

/**
 * Creates a mutable trie with the same contents as the immutable one.
 * You must umutablecptrie_close() the mutable trie once you are done using it.
 *
 * @param trie the immutable trie
 * @param pErrorCode an in/out ICU UErrorCode
 * @return the mutable trie
 * @stable ICU 63
 */
U_CAPI UMutableCPTrie * U_EXPORT2
umutablecptrie_fromUCPTrie(const UCPTrie *trie, UErrorCode *pErrorCode);

/**
 * Returns the value for a code point as stored in the trie.
 *
 * @param trie the trie
 * @param c the code point
 * @return the value
 * @stable ICU 63
 */
U_CAPI uint32_t U_EXPORT2
umutablecptrie_get(const UMutableCPTrie *trie, UChar32 c);

/**
 * Returns the last code point such that all those from start to there have the same value.
 * Can be used to efficiently iterate over all same-value ranges in a trie.
 * (This is normally faster than iterating over code points and get()ting each value,
 * but much slower than a data structure that stores ranges directly.)
 *
 * The trie can be modified between calls to this function.
 *
 * If the UCPMapValueFilter function pointer is not NULL, then
 * the value to be delivered is passed through that function, and the return value is the end
 * of the range where all values are modified to the same actual value.
 * The value is unchanged if that function pointer is NULL.
 *
 * See the same-signature ucptrie_getRange() for a code sample.
 *
 * @param trie the trie
 * @param start range start
 * @param option defines whether surrogates are treated normally,
 *               or as having the surrogateValue; usually UCPMAP_RANGE_NORMAL
 * @param surrogateValue value for surrogates; ignored if option==UCPMAP_RANGE_NORMAL
 * @param filter a pointer to a function that may modify the trie data value,
 *     or NULL if the values from the trie are to be used unmodified
 * @param context an opaque pointer that is passed on to the filter function
 * @param pValue if not NULL, receives the value that every code point start..end has;
 *     may have been modified by filter(context, trie value)
 *     if that function pointer is not NULL
 * @return the range end code point, or -1 if start is not a valid code point
 * @stable ICU 63
 */
U_CAPI UChar32 U_EXPORT2
umutablecptrie_getRange(const UMutableCPTrie *trie, UChar32 start,
                        UCPMapRangeOption option, uint32_t surrogateValue,
                        UCPMapValueFilter *filter, const void *context, uint32_t *pValue);

/**
 * Sets a value for a code point.
 *
 * @param trie the trie
 * @param c the code point
 * @param value the value
 * @param pErrorCode an in/out ICU UErrorCode
 * @stable ICU 63
 */
U_CAPI void U_EXPORT2
umutablecptrie_set(UMutableCPTrie *trie, UChar32 c, uint32_t value, UErrorCode *pErrorCode);

/**
 * Sets a value for each code point [start..end].
 * Faster and more space-efficient than setting the value for each code point separately.
 *
 * @param trie the trie
 * @param start the first code point to get the value
 * @param end the last code point to get the value (inclusive)
 * @param value the value
 * @param pErrorCode an in/out ICU UErrorCode
 * @stable ICU 63
 */
U_CAPI void U_EXPORT2
umutablecptrie_setRange(UMutableCPTrie *trie,
                        UChar32 start, UChar32 end,
                        uint32_t value, UErrorCode *pErrorCode);

/**
 * Compacts the data and builds an immutable UCPTrie according to the parameters.
 * After this, the mutable trie will be empty.
 *
 * The mutable trie stores 32-bit values until buildImmutable() is called.
 * If values shorter than 32 bits are to be stored in the immutable trie,
 * then the upper bits are discarded.
 * For example, when the mutable trie contains values 0x81, -0x7f, and 0xa581,
 * and the value width is 8 bits, then each of these is stored as 0x81
 * and the immutable trie will return that as an unsigned value.
 * (Some implementations may want to make productive temporary use of the upper bits
 * until buildImmutable() discards them.)
 *
 * Not every possible set of mappings can be built into a UCPTrie,
 * because of limitations resulting from speed and space optimizations.
 * Every Unicode assigned character can be mapped to a unique value.
 * Typical data yields data structures far smaller than the limitations.
 *
 * It is possible to construct extremely unusual mappings that exceed the data structure limits.
 * In such a case this function will fail with a U_INDEX_OUTOFBOUNDS_ERROR.
 *
 * @param trie the trie trie
 * @param type selects the trie type
 * @param valueWidth selects the number of bits in a trie data value; if smaller than 32 bits,
 *                   then the values stored in the trie will be truncated first
 * @param pErrorCode an in/out ICU UErrorCode
 *
 * @see umutablecptrie_fromUCPTrie
 * @stable ICU 63
 */
U_CAPI UCPTrie * U_EXPORT2
umutablecptrie_buildImmutable(UMutableCPTrie *trie, UCPTrieType type, UCPTrieValueWidth valueWidth,
                              UErrorCode *pErrorCode);

U_CDECL_END









// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2004-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ubidi_props.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004dec30
*   created by: Markus W. Scherer
*
*   Low-level Unicode bidi/shaping properties access.
*/








U_CDECL_BEGIN

/* library API -------------------------------------------------------------- */

U_CFUNC void
ubidi_addPropertyStarts(const USetAdder *sa, UErrorCode *pErrorCode);

/* property access functions */

U_CFUNC int32_t
ubidi_getMaxValue(UProperty which);

U_CAPI UCharDirection
ubidi_getClass(UChar32 c);

U_CFUNC UBool
ubidi_isMirrored(UChar32 c);

U_CFUNC UChar32
ubidi_getMirror(UChar32 c);

U_CFUNC UBool
ubidi_isBidiControl(UChar32 c);

U_CFUNC UBool
ubidi_isJoinControl(UChar32 c);

U_CFUNC UJoiningType
ubidi_getJoiningType(UChar32 c);

U_CFUNC UJoiningGroup
ubidi_getJoiningGroup(UChar32 c);

U_CFUNC UBidiPairedBracketType
ubidi_getPairedBracketType(UChar32 c);

U_CFUNC UChar32
ubidi_getPairedBracket(UChar32 c);

/* file definitions --------------------------------------------------------- */

#define UBIDI_DATA_NAME "ubidi"
#define UBIDI_DATA_TYPE "icu"

/* format "BiDi" */
#define UBIDI_FMT_0 0x42
#define UBIDI_FMT_1 0x69
#define UBIDI_FMT_2 0x44
#define UBIDI_FMT_3 0x69

/* indexes into indexes[] */
enum {
    UBIDI_IX_INDEX_TOP,
    UBIDI_IX_LENGTH,
    UBIDI_IX_TRIE_SIZE,
    UBIDI_IX_MIRROR_LENGTH,

    UBIDI_IX_JG_START,
    UBIDI_IX_JG_LIMIT,
    UBIDI_IX_JG_START2,  /* new in format version 2.2, ICU 54 */
    UBIDI_IX_JG_LIMIT2,

    UBIDI_MAX_VALUES_INDEX=15,
    UBIDI_IX_TOP=16
};

/* definitions for 16-bit bidi/shaping properties word ---------------------- */

enum {
 /* UBIDI_CLASS_SHIFT=0, */     /* bidi class: 5 bits (4..0) */
    UBIDI_JT_SHIFT=5,           /* joining type: 3 bits (7..5) */

    UBIDI_BPT_SHIFT=8,          /* Bidi_Paired_Bracket_Type(bpt): 2 bits (9..8) */

    UBIDI_JOIN_CONTROL_SHIFT=10,
    UBIDI_BIDI_CONTROL_SHIFT=11,

    UBIDI_IS_MIRRORED_SHIFT=12,         /* 'is mirrored' */
    UBIDI_MIRROR_DELTA_SHIFT=13,        /* bidi mirroring delta: 3 bits (15..13) */

    UBIDI_MAX_JG_SHIFT=16               /* max JG value in indexes[UBIDI_MAX_VALUES_INDEX] bits 23..16 */
};

#define UBIDI_CLASS_MASK        0x0000001f
#define UBIDI_JT_MASK           0x000000e0
#define UBIDI_BPT_MASK          0x00000300

#define UBIDI_MAX_JG_MASK       0x00ff0000

#define UBIDI_GET_CLASS(props) ((props)&UBIDI_CLASS_MASK)
#define UBIDI_GET_FLAG(props, shift) (((props)>>(shift))&1)

#if U_SIGNED_RIGHT_SHIFT_IS_ARITHMETIC
#   define UBIDI_GET_MIRROR_DELTA(props) ((int16_t)(props)>>UBIDI_MIRROR_DELTA_SHIFT)
#else
#   define UBIDI_GET_MIRROR_DELTA(props) (int16_t)(((props)&0x8000) ? (((props)>>UBIDI_MIRROR_DELTA_SHIFT)|0xe000) : ((props)>>UBIDI_MIRROR_DELTA_SHIFT))
#endif

enum {
    UBIDI_ESC_MIRROR_DELTA=-4,
    UBIDI_MIN_MIRROR_DELTA=-3,
    UBIDI_MAX_MIRROR_DELTA=3
};

/* definitions for 32-bit mirror table entry -------------------------------- */

enum {
    /* the source Unicode code point takes 21 bits (20..0) */
    UBIDI_MIRROR_INDEX_SHIFT=21,
    UBIDI_MAX_MIRROR_INDEX=0x7ff
};

#define UBIDI_GET_MIRROR_CODE_POINT(m) (UChar32)((m)&0x1fffff)

#define UBIDI_GET_MIRROR_INDEX(m) ((m)>>UBIDI_MIRROR_INDEX_SHIFT)

U_CDECL_END


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2004-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ucase.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004aug30
*   created by: Markus W. Scherer
*
*   Low-level Unicode character/string case mapping code.
*/







// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2001-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  utrie2.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2008aug16 (starting from a copy of utrie.h)
*   created by: Markus W. Scherer
*/






U_CDECL_BEGIN

struct UTrie;  /* forward declaration */
#ifndef __UTRIE_H__
typedef struct UTrie UTrie;
#endif

/**
 * \file
 *
 * This is a common implementation of a Unicode trie.
 * It is a kind of compressed, serializable table of 16- or 32-bit values associated with
 * Unicode code points (0..0x10ffff). (A map from code points to integers.)
 *
 * This is the second common version of a Unicode trie (hence the name UTrie2).
 * Compared with UTrie version 1:
 * - Still splitting BMP code points 11:5 bits for index and data table lookups.
 * - Still separate data for lead surrogate code _units_ vs. code _points_,
 *   but the lead surrogate code unit values are not required any more
 *   for data lookup for supplementary code points.
 * - The "folding" mechanism is removed. In UTrie version 1, this somewhat
 *   hard-to-explain mechanism was meant to be used for optimized UTF-16
 *   processing, with application-specific encoding of indexing bits
 *   in the lead surrogate data for the associated supplementary code points.
 * - For the last single-value code point range (ending with U+10ffff),
 *   the starting code point ("highStart") and the value are stored.
 * - For supplementary code points U+10000..highStart-1 a three-table lookup
 *   (two index tables and one data table) is used. The first index
 *   is truncated, omitting both the BMP portion and the high range.
 * - There is a special small index for 2-byte UTF-8, and the initial data
 *   entries are designed for fast 1/2-byte UTF-8 lookup.
 *   Starting with ICU 60, C0 and C1 are not recognized as UTF-8 lead bytes any more at all,
 *   and the associated 2-byte indexes are unused.
 */

/**
 * Trie structure.
 * Use only with public API macros and functions.
 */
struct UTrie2;
typedef struct UTrie2 UTrie2;

/* Public UTrie2 API functions: read-only access ---------------------------- */

/**
 * Selectors for the width of a UTrie2 data value.
 */
enum UTrie2ValueBits {
    /** 16 bits per UTrie2 data value. */
    UTRIE2_16_VALUE_BITS,
    /** 32 bits per UTrie2 data value. */
    UTRIE2_32_VALUE_BITS,
    /** Number of selectors for the width of UTrie2 data values. */
    UTRIE2_COUNT_VALUE_BITS
};
typedef enum UTrie2ValueBits UTrie2ValueBits;

/**
 * Open a frozen trie from its serialized from, stored in 32-bit-aligned memory.
 * Inverse of utrie2_serialize().
 * The memory must remain valid and unchanged as long as the trie is used.
 * You must utrie2_close() the trie once you are done using it.
 *
 * @param valueBits selects the data entry size; results in an
 *                  U_INVALID_FORMAT_ERROR if it does not match the serialized form
 * @param data a pointer to 32-bit-aligned memory containing the serialized form of a UTrie2
 * @param length the number of bytes available at data;
 *               can be more than necessary
 * @param pActualLength receives the actual number of bytes at data taken up by the trie data;
 *                      can be NULL
 * @param pErrorCode an in/out ICU UErrorCode
 * @return the unserialized trie
 *
 * @see utrie2_open
 * @see utrie2_serialize
 */
U_CAPI UTrie2 * U_EXPORT2
utrie2_openFromSerialized(UTrie2ValueBits valueBits,
                          const void *data, int32_t length, int32_t *pActualLength,
                          UErrorCode *pErrorCode);

/**
 * Open a frozen, empty "dummy" trie.
 * A dummy trie is an empty trie, used when a real data trie cannot
 * be loaded. Equivalent to calling utrie2_open() and utrie2_freeze(),
 * but without internally creating and compacting/serializing the
 * builder data structure.
 *
 * The trie always returns the initialValue,
 * or the errorValue for out-of-range code points and illegal UTF-8.
 *
 * You must utrie2_close() the trie once you are done using it.
 *
 * @param valueBits selects the data entry size
 * @param initialValue the initial value that is set for all code points
 * @param errorValue the value for out-of-range code points and illegal UTF-8
 * @param pErrorCode an in/out ICU UErrorCode
 * @return the dummy trie
 *
 * @see utrie2_openFromSerialized
 * @see utrie2_open
 */
U_CAPI UTrie2 * U_EXPORT2
utrie2_openDummy(UTrie2ValueBits valueBits,
                 uint32_t initialValue, uint32_t errorValue,
                 UErrorCode *pErrorCode);

/**
 * Get a value from a code point as stored in the trie.
 * Easier to use than UTRIE2_GET16() and UTRIE2_GET32() but slower.
 * Easier to use because, unlike the macros, this function works on all UTrie2
 * objects, frozen or not, holding 16-bit or 32-bit data values.
 *
 * @param trie the trie
 * @param c the code point
 * @return the value
 */
U_CAPI uint32_t U_EXPORT2
utrie2_get32(const UTrie2 *trie, UChar32 c);

/* enumeration callback types */

/**
 * Callback from utrie2_enum(), extracts a uint32_t value from a
 * trie value. This value will be passed on to the UTrie2EnumRange function.
 *
 * @param context an opaque pointer, as passed into utrie2_enum()
 * @param value a value from the trie
 * @return the value that is to be passed on to the UTrie2EnumRange function
 */
typedef uint32_t U_CALLCONV
UTrie2EnumValue(const void *context, uint32_t value);

/**
 * Callback from utrie2_enum(), is called for each contiguous range
 * of code points with the same value as retrieved from the trie and
 * transformed by the UTrie2EnumValue function.
 *
 * The callback function can stop the enumeration by returning FALSE.
 *
 * @param context an opaque pointer, as passed into utrie2_enum()
 * @param start the first code point in a contiguous range with value
 * @param end the last code point in a contiguous range with value (inclusive)
 * @param value the value that is set for all code points in [start..end]
 * @return FALSE to stop the enumeration
 */
typedef UBool U_CALLCONV
UTrie2EnumRange(const void *context, UChar32 start, UChar32 end, uint32_t value);

/**
 * Enumerate efficiently all values in a trie.
 * Do not modify the trie during the enumeration.
 *
 * For each entry in the trie, the value to be delivered is passed through
 * the UTrie2EnumValue function.
 * The value is unchanged if that function pointer is NULL.
 *
 * For each contiguous range of code points with a given (transformed) value,
 * the UTrie2EnumRange function is called.
 *
 * @param trie a pointer to the trie
 * @param enumValue a pointer to a function that may transform the trie entry value,
 *                  or NULL if the values from the trie are to be used directly
 * @param enumRange a pointer to a function that is called for each contiguous range
 *                  of code points with the same (transformed) value
 * @param context an opaque pointer that is passed on to the callback functions
 */
U_CAPI void U_EXPORT2
utrie2_enum(const UTrie2 *trie,
            UTrie2EnumValue *enumValue, UTrie2EnumRange *enumRange, const void *context);

/* Building a trie ---------------------------------------------------------- */

/**
 * Open an empty, writable trie. At build time, 32-bit data values are used.
 * utrie2_freeze() takes a valueBits parameter
 * which determines the data value width in the serialized and frozen forms.
 * You must utrie2_close() the trie once you are done using it.
 *
 * @param initialValue the initial value that is set for all code points
 * @param errorValue the value for out-of-range code points and illegal UTF-8
 * @param pErrorCode an in/out ICU UErrorCode
 * @return a pointer to the allocated and initialized new trie
 */
U_CAPI UTrie2 * U_EXPORT2
utrie2_open(uint32_t initialValue, uint32_t errorValue, UErrorCode *pErrorCode);

/**
 * Clone a trie.
 * You must utrie2_close() the clone once you are done using it.
 *
 * @param other the trie to clone
 * @param pErrorCode an in/out ICU UErrorCode
 * @return a pointer to the new trie clone
 */
U_CAPI UTrie2 * U_EXPORT2
utrie2_clone(const UTrie2 *other, UErrorCode *pErrorCode);

/**
 * Clone a trie. The clone will be mutable/writable even if the other trie
 * is frozen. (See utrie2_freeze().)
 * You must utrie2_close() the clone once you are done using it.
 *
 * @param other the trie to clone
 * @param pErrorCode an in/out ICU UErrorCode
 * @return a pointer to the new trie clone
 */
U_CAPI UTrie2 * U_EXPORT2
utrie2_cloneAsThawed(const UTrie2 *other, UErrorCode *pErrorCode);

/**
 * Close a trie and release associated memory.
 *
 * @param trie the trie
 */
U_CAPI void U_EXPORT2
utrie2_close(UTrie2 *trie);

/**
 * Set a value for a code point.
 *
 * @param trie the unfrozen trie
 * @param c the code point
 * @param value the value
 * @param pErrorCode an in/out ICU UErrorCode; among other possible error codes:
 * - U_NO_WRITE_PERMISSION if the trie is frozen
 */
U_CAPI void U_EXPORT2
utrie2_set32(UTrie2 *trie, UChar32 c, uint32_t value, UErrorCode *pErrorCode);

/**
 * Set a value in a range of code points [start..end].
 * All code points c with start<=c<=end will get the value if
 * overwrite is TRUE or if the old value is the initial value.
 *
 * @param trie the unfrozen trie
 * @param start the first code point to get the value
 * @param end the last code point to get the value (inclusive)
 * @param value the value
 * @param overwrite flag for whether old non-initial values are to be overwritten
 * @param pErrorCode an in/out ICU UErrorCode; among other possible error codes:
 * - U_NO_WRITE_PERMISSION if the trie is frozen
 */
U_CAPI void U_EXPORT2
utrie2_setRange32(UTrie2 *trie,
                  UChar32 start, UChar32 end,
                  uint32_t value, UBool overwrite,
                  UErrorCode *pErrorCode);

/**
 * Freeze a trie. Make it immutable (read-only) and compact it,
 * ready for serialization and for use with fast macros.
 * Functions to set values will fail after serializing.
 *
 * A trie can be frozen only once. If this function is called again with different
 * valueBits then it will set a U_ILLEGAL_ARGUMENT_ERROR.
 *
 * @param trie the trie
 * @param valueBits selects the data entry size; if smaller than 32 bits, then
 *                  the values stored in the trie will be truncated
 * @param pErrorCode an in/out ICU UErrorCode; among other possible error codes:
 * - U_INDEX_OUTOFBOUNDS_ERROR if the compacted index or data arrays are too long
 *                             for serialization
 *                             (the trie will be immutable and usable,
 *                             but not frozen and not usable with the fast macros)
 *
 * @see utrie2_cloneAsThawed
 */
U_CAPI void U_EXPORT2
utrie2_freeze(UTrie2 *trie, UTrie2ValueBits valueBits, UErrorCode *pErrorCode);

/**
 * Test if the trie is frozen. (See utrie2_freeze().)
 *
 * @param trie the trie
 * @return TRUE if the trie is frozen, that is, immutable, ready for serialization
 *         and for use with fast macros
 */
U_CAPI UBool U_EXPORT2
utrie2_isFrozen(const UTrie2 *trie);

/**
 * Serialize a frozen trie into 32-bit aligned memory.
 * If the trie is not frozen, then the function returns with a U_ILLEGAL_ARGUMENT_ERROR.
 * A trie can be serialized multiple times.
 *
 * @param trie the frozen trie
 * @param data a pointer to 32-bit-aligned memory to be filled with the trie data,
 *             can be NULL if capacity==0
 * @param capacity the number of bytes available at data,
 *                 or 0 for preflighting
 * @param pErrorCode an in/out ICU UErrorCode; among other possible error codes:
 * - U_BUFFER_OVERFLOW_ERROR if the data storage block is too small for serialization
 * - U_ILLEGAL_ARGUMENT_ERROR if the trie is not frozen or the data and capacity
 *                            parameters are bad
 * @return the number of bytes written or needed for the trie
 *
 * @see utrie2_openFromSerialized()
 */
U_CAPI int32_t U_EXPORT2
utrie2_serialize(const UTrie2 *trie,
                 void *data, int32_t capacity,
                 UErrorCode *pErrorCode);

/* Public UTrie2 API: miscellaneous functions ------------------------------- */

/**
 * Build a UTrie2 (version 2) from a UTrie (version 1).
 * Enumerates all values in the UTrie and builds a UTrie2 with the same values.
 * The resulting UTrie2 will be frozen.
 *
 * @param trie1 the runtime UTrie structure to be enumerated
 * @param errorValue the value for out-of-range code points and illegal UTF-8
 * @param pErrorCode an in/out ICU UErrorCode
 * @return The frozen UTrie2 with the same values as the UTrie.
 */
U_CAPI UTrie2 * U_EXPORT2
utrie2_fromUTrie(const UTrie *trie1, uint32_t errorValue, UErrorCode *pErrorCode);

/* Public UTrie2 API macros ------------------------------------------------- */

/*
 * These macros provide fast data lookup from a frozen trie.
 * They will crash when used on an unfrozen trie.
 */

/**
 * Return a 16-bit trie value from a code point, with range checking.
 * Returns trie->errorValue if c is not in the range 0..U+10ffff.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param c (UChar32, in) the input code point
 * @return (uint16_t) The code point's trie value.
 */
#define UTRIE2_GET16(trie, c) _UTRIE2_GET((trie), index, (trie)->indexLength, (c))

/**
 * Return a 32-bit trie value from a code point, with range checking.
 * Returns trie->errorValue if c is not in the range 0..U+10ffff.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param c (UChar32, in) the input code point
 * @return (uint32_t) The code point's trie value.
 */
#define UTRIE2_GET32(trie, c) _UTRIE2_GET((trie), data32, 0, (c))

/**
 * UTF-16: Get the next code point (UChar32 c, out), post-increment src,
 * and get a 16-bit value from the trie.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param src (const UChar *, in/out) the source text pointer
 * @param limit (const UChar *, in) the limit pointer for the text, or NULL if NUL-terminated
 * @param c (UChar32, out) variable for the code point
 * @param result (uint16_t, out) uint16_t variable for the trie lookup result
 */
#define UTRIE2_U16_NEXT16(trie, src, limit, c, result) _UTRIE2_U16_NEXT(trie, index, src, limit, c, result)

/**
 * UTF-16: Get the next code point (UChar32 c, out), post-increment src,
 * and get a 32-bit value from the trie.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param src (const UChar *, in/out) the source text pointer
 * @param limit (const UChar *, in) the limit pointer for the text, or NULL if NUL-terminated
 * @param c (UChar32, out) variable for the code point
 * @param result (uint32_t, out) uint32_t variable for the trie lookup result
 */
#define UTRIE2_U16_NEXT32(trie, src, limit, c, result) _UTRIE2_U16_NEXT(trie, data32, src, limit, c, result)

/**
 * UTF-16: Get the previous code point (UChar32 c, out), pre-decrement src,
 * and get a 16-bit value from the trie.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param start (const UChar *, in) the start pointer for the text
 * @param src (const UChar *, in/out) the source text pointer
 * @param c (UChar32, out) variable for the code point
 * @param result (uint16_t, out) uint16_t variable for the trie lookup result
 */
#define UTRIE2_U16_PREV16(trie, start, src, c, result) _UTRIE2_U16_PREV(trie, index, start, src, c, result)

/**
 * UTF-16: Get the previous code point (UChar32 c, out), pre-decrement src,
 * and get a 32-bit value from the trie.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param start (const UChar *, in) the start pointer for the text
 * @param src (const UChar *, in/out) the source text pointer
 * @param c (UChar32, out) variable for the code point
 * @param result (uint32_t, out) uint32_t variable for the trie lookup result
 */
#define UTRIE2_U16_PREV32(trie, start, src, c, result) _UTRIE2_U16_PREV(trie, data32, start, src, c, result)

/**
 * UTF-8: Post-increment src and get a 16-bit value from the trie.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param src (const char *, in/out) the source text pointer
 * @param limit (const char *, in) the limit pointer for the text (must not be NULL)
 * @param result (uint16_t, out) uint16_t variable for the trie lookup result
 */
#define UTRIE2_U8_NEXT16(trie, src, limit, result)\
    _UTRIE2_U8_NEXT(trie, data16, index, src, limit, result)

/**
 * UTF-8: Post-increment src and get a 32-bit value from the trie.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param src (const char *, in/out) the source text pointer
 * @param limit (const char *, in) the limit pointer for the text (must not be NULL)
 * @param result (uint16_t, out) uint32_t variable for the trie lookup result
 */
#define UTRIE2_U8_NEXT32(trie, src, limit, result) \
    _UTRIE2_U8_NEXT(trie, data32, data32, src, limit, result)

/**
 * UTF-8: Pre-decrement src and get a 16-bit value from the trie.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param start (const char *, in) the start pointer for the text
 * @param src (const char *, in/out) the source text pointer
 * @param result (uint16_t, out) uint16_t variable for the trie lookup result
 */
#define UTRIE2_U8_PREV16(trie, start, src, result) \
    _UTRIE2_U8_PREV(trie, data16, index, start, src, result)

/**
 * UTF-8: Pre-decrement src and get a 32-bit value from the trie.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param start (const char *, in) the start pointer for the text
 * @param src (const char *, in/out) the source text pointer
 * @param result (uint16_t, out) uint32_t variable for the trie lookup result
 */
#define UTRIE2_U8_PREV32(trie, start, src, result) \
    _UTRIE2_U8_PREV(trie, data32, data32, start, src, result)

/* Public UTrie2 API: optimized UTF-16 access ------------------------------- */

/*
 * The following functions and macros are used for highly optimized UTF-16
 * text processing. The UTRIE2_U16_NEXTxy() macros do not depend on these.
 *
 * A UTrie2 stores separate values for lead surrogate code _units_ vs. code _points_.
 * UTF-16 text processing can be optimized by detecting surrogate pairs and
 * assembling supplementary code points only when there is non-trivial data
 * available.
 *
 * At build-time, use utrie2_enumForLeadSurrogate() to see if there
 * is non-trivial (non-initialValue) data for any of the supplementary
 * code points associated with a lead surrogate.
 * If so, then set a special (application-specific) value for the
 * lead surrogate code _unit_, with utrie2_set32ForLeadSurrogateCodeUnit().
 *
 * At runtime, use UTRIE2_GET16_FROM_U16_SINGLE_LEAD() or
 * UTRIE2_GET32_FROM_U16_SINGLE_LEAD() per code unit. If there is non-trivial
 * data and the code unit is a lead surrogate, then check if a trail surrogate
 * follows. If so, assemble the supplementary code point with
 * U16_GET_SUPPLEMENTARY() and look up its value with UTRIE2_GET16_FROM_SUPP()
 * or UTRIE2_GET32_FROM_SUPP(); otherwise reset the lead
 * surrogate's value or do a code point lookup for it.
 *
 * If there is only trivial data for lead and trail surrogates, then processing
 * can often skip them. For example, in normalization or case mapping
 * all characters that do not have any mappings are simply copied as is.
 */

/**
 * Get a value from a lead surrogate code unit as stored in the trie.
 *
 * @param trie the trie
 * @param c the code unit (U+D800..U+DBFF)
 * @return the value
 */
U_CAPI uint32_t U_EXPORT2
utrie2_get32FromLeadSurrogateCodeUnit(const UTrie2 *trie, UChar32 c);

/**
 * Enumerate the trie values for the 1024=0x400 code points
 * corresponding to a given lead surrogate.
 * For example, for the lead surrogate U+D87E it will enumerate the values
 * for [U+2F800..U+2FC00[.
 * Used by data builder code that sets special lead surrogate code unit values
 * for optimized UTF-16 string processing.
 *
 * Do not modify the trie during the enumeration.
 *
 * Except for the limited code point range, this functions just like utrie2_enum():
 * For each entry in the trie, the value to be delivered is passed through
 * the UTrie2EnumValue function.
 * The value is unchanged if that function pointer is NULL.
 *
 * For each contiguous range of code points with a given (transformed) value,
 * the UTrie2EnumRange function is called.
 *
 * @param trie a pointer to the trie
 * @param enumValue a pointer to a function that may transform the trie entry value,
 *                  or NULL if the values from the trie are to be used directly
 * @param enumRange a pointer to a function that is called for each contiguous range
 *                  of code points with the same (transformed) value
 * @param context an opaque pointer that is passed on to the callback functions
 */
U_CAPI void U_EXPORT2
utrie2_enumForLeadSurrogate(const UTrie2 *trie, UChar32 lead,
                            UTrie2EnumValue *enumValue, UTrie2EnumRange *enumRange,
                            const void *context);

/**
 * Set a value for a lead surrogate code unit.
 *
 * @param trie the unfrozen trie
 * @param lead the lead surrogate code unit (U+D800..U+DBFF)
 * @param value the value
 * @param pErrorCode an in/out ICU UErrorCode; among other possible error codes:
 * - U_NO_WRITE_PERMISSION if the trie is frozen
 */
U_CAPI void U_EXPORT2
utrie2_set32ForLeadSurrogateCodeUnit(UTrie2 *trie,
                                     UChar32 lead, uint32_t value,
                                     UErrorCode *pErrorCode);

/**
 * Return a 16-bit trie value from a UTF-16 single/lead code unit (<=U+ffff).
 * Same as UTRIE2_GET16() if c is a BMP code point except for lead surrogates,
 * but smaller and faster.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param c (UChar32, in) the input code unit, must be 0<=c<=U+ffff
 * @return (uint16_t) The code unit's trie value.
 */
#define UTRIE2_GET16_FROM_U16_SINGLE_LEAD(trie, c) _UTRIE2_GET_FROM_U16_SINGLE_LEAD((trie), index, c)

/**
 * Return a 32-bit trie value from a UTF-16 single/lead code unit (<=U+ffff).
 * Same as UTRIE2_GET32() if c is a BMP code point except for lead surrogates,
 * but smaller and faster.
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param c (UChar32, in) the input code unit, must be 0<=c<=U+ffff
 * @return (uint32_t) The code unit's trie value.
 */
#define UTRIE2_GET32_FROM_U16_SINGLE_LEAD(trie, c) _UTRIE2_GET_FROM_U16_SINGLE_LEAD((trie), data32, c)

/**
 * Return a 16-bit trie value from a supplementary code point (U+10000..U+10ffff).
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param c (UChar32, in) the input code point, must be U+10000<=c<=U+10ffff
 * @return (uint16_t) The code point's trie value.
 */
#define UTRIE2_GET16_FROM_SUPP(trie, c) _UTRIE2_GET_FROM_SUPP((trie), index, c)

/**
 * Return a 32-bit trie value from a supplementary code point (U+10000..U+10ffff).
 *
 * @param trie (const UTrie2 *, in) a frozen trie
 * @param c (UChar32, in) the input code point, must be U+10000<=c<=U+10ffff
 * @return (uint32_t) The code point's trie value.
 */
#define UTRIE2_GET32_FROM_SUPP(trie, c) _UTRIE2_GET_FROM_SUPP((trie), data32, c)

U_CDECL_END

/* C++ convenience wrappers ------------------------------------------------- */

#ifdef __cplusplus




U_NAMESPACE_BEGIN

// Use the Forward/Backward subclasses below.
class UTrie2StringIterator : public UMemory {
public:
    UTrie2StringIterator(const UTrie2 *t, const UChar *p) :
        trie(t), codePointStart(p), codePointLimit(p), codePoint(U_SENTINEL) {}

    const UTrie2 *trie;
    const UChar *codePointStart, *codePointLimit;
    UChar32 codePoint;
};

class BackwardUTrie2StringIterator : public UTrie2StringIterator {
public:
    BackwardUTrie2StringIterator(const UTrie2 *t, const UChar *s, const UChar *p) :
        UTrie2StringIterator(t, p), start(s) {}

    uint16_t previous16();

    const UChar *start;
};

class ForwardUTrie2StringIterator : public UTrie2StringIterator {
public:
    // Iteration limit l can be NULL.
    // In that case, the caller must detect c==0 and stop.
    ForwardUTrie2StringIterator(const UTrie2 *t, const UChar *p, const UChar *l) :
        UTrie2StringIterator(t, p), limit(l) {}

    uint16_t next16();

    const UChar *limit;
};

U_NAMESPACE_END

#endif

/* Internal definitions ----------------------------------------------------- */

U_CDECL_BEGIN

/** Build-time trie structure. */
struct UNewTrie2;
typedef struct UNewTrie2 UNewTrie2;

/*
 * Trie structure definition.
 *
 * Either the data table is 16 bits wide and accessed via the index
 * pointer, with each index item increased by indexLength;
 * in this case, data32==NULL, and data16 is used for direct ASCII access.
 *
 * Or the data table is 32 bits wide and accessed via the data32 pointer.
 */
struct UTrie2 {
    /* protected: used by macros and functions for reading values */
    const uint16_t *index;
    const uint16_t *data16;     /* for fast UTF-8 ASCII access, if 16b data */
    const uint32_t *data32;     /* NULL if 16b data is used via index */

    int32_t indexLength, dataLength;
    uint16_t index2NullOffset;  /* 0xffff if there is no dedicated index-2 null block */
    uint16_t dataNullOffset;
    uint32_t initialValue;
    /** Value returned for out-of-range code points and illegal UTF-8. */
    uint32_t errorValue;

    /* Start of the last range which ends at U+10ffff, and its value. */
    UChar32 highStart;
    int32_t highValueIndex;

    /* private: used by builder and unserialization functions */
    void *memory;           /* serialized bytes; NULL if not frozen yet */
    int32_t length;         /* number of serialized bytes at memory; 0 if not frozen yet */
    UBool isMemoryOwned;    /* TRUE if the trie owns the memory */
    UBool padding1;
    int16_t padding2;
    UNewTrie2 *newTrie;     /* builder object; NULL when frozen */

#ifdef UTRIE2_DEBUG
    const char *name;
#endif
};

/**
 * Trie constants, defining shift widths, index array lengths, etc.
 *
 * These are needed for the runtime macros but users can treat these as
 * implementation details and skip to the actual public API further below.
 */
enum {
    /** Shift size for getting the index-1 table offset. */
    UTRIE2_SHIFT_1=6+5,

    /** Shift size for getting the index-2 table offset. */
    UTRIE2_SHIFT_2=5,

    /**
     * Difference between the two shift sizes,
     * for getting an index-1 offset from an index-2 offset. 6=11-5
     */
    UTRIE2_SHIFT_1_2=UTRIE2_SHIFT_1-UTRIE2_SHIFT_2,

    /**
     * Number of index-1 entries for the BMP. 32=0x20
     * This part of the index-1 table is omitted from the serialized form.
     */
    UTRIE2_OMITTED_BMP_INDEX_1_LENGTH=0x10000>>UTRIE2_SHIFT_1,

    /** Number of code points per index-1 table entry. 2048=0x800 */
    UTRIE2_CP_PER_INDEX_1_ENTRY=1<<UTRIE2_SHIFT_1,

    /** Number of entries in an index-2 block. 64=0x40 */
    UTRIE2_INDEX_2_BLOCK_LENGTH=1<<UTRIE2_SHIFT_1_2,

    /** Mask for getting the lower bits for the in-index-2-block offset. */
    UTRIE2_INDEX_2_MASK=UTRIE2_INDEX_2_BLOCK_LENGTH-1,

    /** Number of entries in a data block. 32=0x20 */
    UTRIE2_DATA_BLOCK_LENGTH=1<<UTRIE2_SHIFT_2,

    /** Mask for getting the lower bits for the in-data-block offset. */
    UTRIE2_DATA_MASK=UTRIE2_DATA_BLOCK_LENGTH-1,

    /**
     * Shift size for shifting left the index array values.
     * Increases possible data size with 16-bit index values at the cost
     * of compactability.
     * This requires data blocks to be aligned by UTRIE2_DATA_GRANULARITY.
     */
    UTRIE2_INDEX_SHIFT=2,

    /** The alignment size of a data block. Also the granularity for compaction. */
    UTRIE2_DATA_GRANULARITY=1<<UTRIE2_INDEX_SHIFT,

    /* Fixed layout of the first part of the index array. ------------------- */

    /**
     * The BMP part of the index-2 table is fixed and linear and starts at offset 0.
     * Length=2048=0x800=0x10000>>UTRIE2_SHIFT_2.
     */
    UTRIE2_INDEX_2_OFFSET=0,

    /**
     * The part of the index-2 table for U+D800..U+DBFF stores values for
     * lead surrogate code _units_ not code _points_.
     * Values for lead surrogate code _points_ are indexed with this portion of the table.
     * Length=32=0x20=0x400>>UTRIE2_SHIFT_2. (There are 1024=0x400 lead surrogates.)
     */
    UTRIE2_LSCP_INDEX_2_OFFSET=0x10000>>UTRIE2_SHIFT_2,
    UTRIE2_LSCP_INDEX_2_LENGTH=0x400>>UTRIE2_SHIFT_2,

    /** Count the lengths of both BMP pieces. 2080=0x820 */
    UTRIE2_INDEX_2_BMP_LENGTH=UTRIE2_LSCP_INDEX_2_OFFSET+UTRIE2_LSCP_INDEX_2_LENGTH,

    /**
     * The 2-byte UTF-8 version of the index-2 table follows at offset 2080=0x820.
     * Length 32=0x20 for lead bytes C0..DF, regardless of UTRIE2_SHIFT_2.
     */
    UTRIE2_UTF8_2B_INDEX_2_OFFSET=UTRIE2_INDEX_2_BMP_LENGTH,
    UTRIE2_UTF8_2B_INDEX_2_LENGTH=0x800>>6,  /* U+0800 is the first code point after 2-byte UTF-8 */

    /**
     * The index-1 table, only used for supplementary code points, at offset 2112=0x840.
     * Variable length, for code points up to highStart, where the last single-value range starts.
     * Maximum length 512=0x200=0x100000>>UTRIE2_SHIFT_1.
     * (For 0x100000 supplementary code points U+10000..U+10ffff.)
     *
     * The part of the index-2 table for supplementary code points starts
     * after this index-1 table.
     *
     * Both the index-1 table and the following part of the index-2 table
     * are omitted completely if there is only BMP data.
     */
    UTRIE2_INDEX_1_OFFSET=UTRIE2_UTF8_2B_INDEX_2_OFFSET+UTRIE2_UTF8_2B_INDEX_2_LENGTH,
    UTRIE2_MAX_INDEX_1_LENGTH=0x100000>>UTRIE2_SHIFT_1,

    /*
     * Fixed layout of the first part of the data array. -----------------------
     * Starts with 4 blocks (128=0x80 entries) for ASCII.
     */

    /**
     * The illegal-UTF-8 data block follows the ASCII block, at offset 128=0x80.
     * Used with linear access for single bytes 0..0xbf for simple error handling.
     * Length 64=0x40, not UTRIE2_DATA_BLOCK_LENGTH.
     */
    UTRIE2_BAD_UTF8_DATA_OFFSET=0x80,

    /** The start of non-linear-ASCII data blocks, at offset 192=0xc0. */
    UTRIE2_DATA_START_OFFSET=0xc0
};

/* Internal functions and macros -------------------------------------------- */

/**
 * Internal function for part of the UTRIE2_U8_NEXTxx() macro implementations.
 * Do not call directly.
 * @internal
 */
U_INTERNAL int32_t U_EXPORT2
utrie2_internalU8NextIndex(const UTrie2 *trie, UChar32 c,
                           const uint8_t *src, const uint8_t *limit);

/**
 * Internal function for part of the UTRIE2_U8_PREVxx() macro implementations.
 * Do not call directly.
 * @internal
 */
U_INTERNAL int32_t U_EXPORT2
utrie2_internalU8PrevIndex(const UTrie2 *trie, UChar32 c,
                           const uint8_t *start, const uint8_t *src);


/** Internal low-level trie getter. Returns a data index. */
#define _UTRIE2_INDEX_RAW(offset, trieIndex, c) \
    (((int32_t)((trieIndex)[(offset)+((c)>>UTRIE2_SHIFT_2)]) \
    <<UTRIE2_INDEX_SHIFT)+ \
    ((c)&UTRIE2_DATA_MASK))

/** Internal trie getter from a UTF-16 single/lead code unit. Returns the data index. */
#define _UTRIE2_INDEX_FROM_U16_SINGLE_LEAD(trieIndex, c) _UTRIE2_INDEX_RAW(0, trieIndex, c)

/** Internal trie getter from a lead surrogate code point (D800..DBFF). Returns the data index. */
#define _UTRIE2_INDEX_FROM_LSCP(trieIndex, c) \
    _UTRIE2_INDEX_RAW(UTRIE2_LSCP_INDEX_2_OFFSET-(0xd800>>UTRIE2_SHIFT_2), trieIndex, c)

/** Internal trie getter from a BMP code point. Returns the data index. */
#define _UTRIE2_INDEX_FROM_BMP(trieIndex, c) \
    _UTRIE2_INDEX_RAW(U_IS_LEAD(c) ? UTRIE2_LSCP_INDEX_2_OFFSET-(0xd800>>UTRIE2_SHIFT_2) : 0, \
                      trieIndex, c)

/** Internal trie getter from a supplementary code point below highStart. Returns the data index. */
#define _UTRIE2_INDEX_FROM_SUPP(trieIndex, c) \
    (((int32_t)((trieIndex)[ \
        (trieIndex)[(UTRIE2_INDEX_1_OFFSET-UTRIE2_OMITTED_BMP_INDEX_1_LENGTH)+ \
                      ((c)>>UTRIE2_SHIFT_1)]+ \
        (((c)>>UTRIE2_SHIFT_2)&UTRIE2_INDEX_2_MASK)]) \
    <<UTRIE2_INDEX_SHIFT)+ \
    ((c)&UTRIE2_DATA_MASK))

/**
 * Internal trie getter from a code point, with checking that c is in 0..10FFFF.
 * Returns the data index.
 */
#define _UTRIE2_INDEX_FROM_CP(trie, asciiOffset, c) \
    ((uint32_t)(c)<0xd800 ? \
        _UTRIE2_INDEX_RAW(0, (trie)->index, c) : \
        (uint32_t)(c)<=0xffff ? \
            _UTRIE2_INDEX_RAW( \
                (c)<=0xdbff ? UTRIE2_LSCP_INDEX_2_OFFSET-(0xd800>>UTRIE2_SHIFT_2) : 0, \
                (trie)->index, c) : \
            (uint32_t)(c)>0x10ffff ? \
                (asciiOffset)+UTRIE2_BAD_UTF8_DATA_OFFSET : \
                (c)>=(trie)->highStart ? \
                    (trie)->highValueIndex : \
                    _UTRIE2_INDEX_FROM_SUPP((trie)->index, c))

/** Internal trie getter from a UTF-16 single/lead code unit. Returns the data. */
#define _UTRIE2_GET_FROM_U16_SINGLE_LEAD(trie, data, c) \
    (trie)->data[_UTRIE2_INDEX_FROM_U16_SINGLE_LEAD((trie)->index, c)]

/** Internal trie getter from a supplementary code point. Returns the data. */
#define _UTRIE2_GET_FROM_SUPP(trie, data, c) \
    (trie)->data[(c)>=(trie)->highStart ? (trie)->highValueIndex : \
                 _UTRIE2_INDEX_FROM_SUPP((trie)->index, c)]

/**
 * Internal trie getter from a code point, with checking that c is in 0..10FFFF.
 * Returns the data.
 */
#define _UTRIE2_GET(trie, data, asciiOffset, c) \
    (trie)->data[_UTRIE2_INDEX_FROM_CP(trie, asciiOffset, c)]

/** Internal next-post-increment: get the next code point (c) and its data. */
#define _UTRIE2_U16_NEXT(trie, data, src, limit, c, result) UPRV_BLOCK_MACRO_BEGIN { \
    { \
        uint16_t __c2; \
        (c)=*(src)++; \
        if(!U16_IS_LEAD(c)) { \
            (result)=_UTRIE2_GET_FROM_U16_SINGLE_LEAD(trie, data, c); \
        } else if((src)==(limit) || !U16_IS_TRAIL(__c2=*(src))) { \
            (result)=(trie)->data[_UTRIE2_INDEX_FROM_LSCP((trie)->index, c)]; \
        } else { \
            ++(src); \
            (c)=U16_GET_SUPPLEMENTARY((c), __c2); \
            (result)=_UTRIE2_GET_FROM_SUPP((trie), data, (c)); \
        } \
    } \
} UPRV_BLOCK_MACRO_END

/** Internal pre-decrement-previous: get the previous code point (c) and its data */
#define _UTRIE2_U16_PREV(trie, data, start, src, c, result) UPRV_BLOCK_MACRO_BEGIN { \
    { \
        uint16_t __c2; \
        (c)=*--(src); \
        if(!U16_IS_TRAIL(c) || (src)==(start) || !U16_IS_LEAD(__c2=*((src)-1))) { \
            (result)=(trie)->data[_UTRIE2_INDEX_FROM_BMP((trie)->index, c)]; \
        } else { \
            --(src); \
            (c)=U16_GET_SUPPLEMENTARY(__c2, (c)); \
            (result)=_UTRIE2_GET_FROM_SUPP((trie), data, (c)); \
        } \
    } \
} UPRV_BLOCK_MACRO_END

/** Internal UTF-8 next-post-increment: get the next code point's data. */
#define _UTRIE2_U8_NEXT(trie, ascii, data, src, limit, result) UPRV_BLOCK_MACRO_BEGIN { \
    uint8_t __lead=(uint8_t)*(src)++; \
    if(U8_IS_SINGLE(__lead)) { \
        (result)=(trie)->ascii[__lead]; \
    } else { \
        uint8_t __t1, __t2; \
        if( /* handle U+0800..U+FFFF inline */ \
            0xe0<=__lead && __lead<0xf0 && ((src)+1)<(limit) && \
            U8_IS_VALID_LEAD3_AND_T1(__lead, __t1=(uint8_t)*(src)) && \
            (__t2=(uint8_t)(*((src)+1)-0x80))<= 0x3f \
        ) { \
            (src)+=2; \
            (result)=(trie)->data[ \
                ((int32_t)((trie)->index[((__lead-0xe0)<<(12-UTRIE2_SHIFT_2))+ \
                                         ((__t1&0x3f)<<(6-UTRIE2_SHIFT_2))+(__t2>>UTRIE2_SHIFT_2)]) \
                <<UTRIE2_INDEX_SHIFT)+ \
                (__t2&UTRIE2_DATA_MASK)]; \
        } else if( /* handle U+0080..U+07FF inline */ \
            __lead<0xe0 && __lead>=0xc2 && (src)<(limit) && \
            (__t1=(uint8_t)(*(src)-0x80))<=0x3f \
        ) { \
            ++(src); \
            (result)=(trie)->data[ \
                (trie)->index[(UTRIE2_UTF8_2B_INDEX_2_OFFSET-0xc0)+__lead]+ \
                __t1]; \
        } else { \
            int32_t __index=utrie2_internalU8NextIndex((trie), __lead, (const uint8_t *)(src), \
                                                                       (const uint8_t *)(limit)); \
            (src)+=__index&7; \
            (result)=(trie)->data[__index>>3]; \
        } \
    } \
} UPRV_BLOCK_MACRO_END

/** Internal UTF-8 pre-decrement-previous: get the previous code point's data. */
#define _UTRIE2_U8_PREV(trie, ascii, data, start, src, result) UPRV_BLOCK_MACRO_BEGIN { \
    uint8_t __b=(uint8_t)*--(src); \
    if(U8_IS_SINGLE(__b)) { \
        (result)=(trie)->ascii[__b]; \
    } else { \
        int32_t __index=utrie2_internalU8PrevIndex((trie), __b, (const uint8_t *)(start), \
                                                                (const uint8_t *)(src)); \
        (src)-=__index&7; \
        (result)=(trie)->data[__index>>3]; \
    } \
} UPRV_BLOCK_MACRO_END

U_CDECL_END



#ifdef __cplusplus
U_NAMESPACE_BEGIN

class UnicodeString;

U_NAMESPACE_END
#endif

/* library API -------------------------------------------------------------- */

U_CFUNC void U_EXPORT2
ucase_addPropertyStarts(const USetAdder *sa, UErrorCode *pErrorCode);

/**
 * Requires non-NULL locale ID but otherwise does the equivalent of
 * checking for language codes as if uloc_getLanguage() were called:
 * Accepts both 2- and 3-letter codes and accepts case variants.
 */
U_CFUNC int32_t
ucase_getCaseLocale(const char *locale);

/* Casing locale types for ucase_getCaseLocale */
enum {
    UCASE_LOC_UNKNOWN,
    UCASE_LOC_ROOT,
    UCASE_LOC_TURKISH,
    UCASE_LOC_LITHUANIAN,
    UCASE_LOC_GREEK,
    UCASE_LOC_DUTCH
};

/**
 * Bit mask for getting just the options from a string compare options word
 * that are relevant for case-insensitive string comparison.
 * See stringoptions.h. Also include _STRNCMP_STYLE and U_COMPARE_CODE_POINT_ORDER.
 * @internal
 */
#define _STRCASECMP_OPTIONS_MASK 0xffff

/**
 * Bit mask for getting just the options from a string compare options word
 * that are relevant for case folding (of a single string or code point).
 *
 * Currently only bit 0 for U_FOLD_CASE_EXCLUDE_SPECIAL_I.
 * It is conceivable that at some point we might use one more bit for using uppercase sharp s.
 * It is conceivable that at some point we might want the option to use only simple case foldings
 * when operating on strings.
 *
 * See stringoptions.h.
 * @internal
 */
#define _FOLD_CASE_OPTIONS_MASK 7

/* single-code point functions */

U_CAPI UChar32 U_EXPORT2
ucase_tolower(UChar32 c);

U_CAPI UChar32 U_EXPORT2
ucase_toupper(UChar32 c);

U_CAPI UChar32 U_EXPORT2
ucase_totitle(UChar32 c);

U_CAPI UChar32 U_EXPORT2
ucase_fold(UChar32 c, uint32_t options);

/**
 * Adds all simple case mappings and the full case folding for c to sa,
 * and also adds special case closure mappings.
 * c itself is not added.
 * For example, the mappings
 * - for s include long s
 * - for sharp s include ss
 * - for k include the Kelvin sign
 */
U_CFUNC void U_EXPORT2
ucase_addCaseClosure(UChar32 c, const USetAdder *sa);

/**
 * Maps the string to single code points and adds the associated case closure
 * mappings.
 * The string is mapped to code points if it is their full case folding string.
 * In other words, this performs a reverse full case folding and then
 * adds the case closure items of the resulting code points.
 * If the string is found and its closure applied, then
 * the string itself is added as well as part of its code points' closure.
 * It must be length>=0.
 *
 * @return TRUE if the string was found
 */
U_CFUNC UBool U_EXPORT2
ucase_addStringCaseClosure(const UChar *s, int32_t length, const USetAdder *sa);

#ifdef __cplusplus
U_NAMESPACE_BEGIN

/**
 * Iterator over characters with more than one code point in the full default Case_Folding.
 */
class U_COMMON_API FullCaseFoldingIterator {
public:
    /** Constructor. */
    FullCaseFoldingIterator();
    /**
     * Returns the next (cp, full) pair where "full" is cp's full default Case_Folding.
     * Returns a negative cp value at the end of the iteration.
     */
    UChar32 next(UnicodeString &full);
private:
    FullCaseFoldingIterator(const FullCaseFoldingIterator &);  // no copy
    FullCaseFoldingIterator &operator=(const FullCaseFoldingIterator &);  // no assignment

    const UChar *unfold;
    int32_t unfoldRows;
    int32_t unfoldRowWidth;
    int32_t unfoldStringWidth;
    int32_t currentRow;
    int32_t rowCpIndex;
};

/**
 * Fast case mapping data for ASCII/Latin.
 * Linear arrays of delta bytes: 0=no mapping; EXC=exception.
 * Deltas must not cross the ASCII boundary, or else they cannot be easily used
 * in simple UTF-8 code.
 */
namespace LatinCase {

/** Case mapping/folding data for code points up to U+017F. */
constexpr UChar LIMIT = 0x180;
/** U+017F case-folds and uppercases crossing the ASCII boundary. */
constexpr UChar LONG_S = 0x17f;
/** Exception: Complex mapping, or too-large delta. */
constexpr int8_t EXC = -0x80;

/** Deltas for lowercasing for most locales, and default case folding. */
extern const int8_t TO_LOWER_NORMAL[LIMIT];
/** Deltas for lowercasing for tr/az/lt, and Turkic case folding. */
extern const int8_t TO_LOWER_TR_LT[LIMIT];

/** Deltas for uppercasing for most locales. */
extern const int8_t TO_UPPER_NORMAL[LIMIT];
/** Deltas for uppercasing for tr/az. */
extern const int8_t TO_UPPER_TR[LIMIT];

}  // namespace LatinCase

U_NAMESPACE_END
#endif

/** @return UCASE_NONE, UCASE_LOWER, UCASE_UPPER, UCASE_TITLE */
U_CAPI int32_t U_EXPORT2
ucase_getType(UChar32 c);

/** @return like ucase_getType() but also sets UCASE_IGNORABLE if c is case-ignorable */
U_CAPI int32_t U_EXPORT2
ucase_getTypeOrIgnorable(UChar32 c);

U_CAPI UBool U_EXPORT2
ucase_isSoftDotted(UChar32 c);

U_CAPI UBool U_EXPORT2
ucase_isCaseSensitive(UChar32 c);

/* string case mapping functions */

U_CDECL_BEGIN

/**
 * Iterator function for string case mappings, which need to look at the
 * context (surrounding text) of a given character for conditional mappings.
 *
 * The iterator only needs to go backward or forward away from the
 * character in question. It does not use any indexes on this interface.
 * It does not support random access or an arbitrary change of
 * iteration direction.
 *
 * The code point being case-mapped itself is never returned by
 * this iterator.
 *
 * @param context A pointer to the iterator's working data.
 * @param dir If <0 then start iterating backward from the character;
 *            if >0 then start iterating forward from the character;
 *            if 0 then continue iterating in the current direction.
 * @return Next code point, or <0 when the iteration is done.
 */
typedef UChar32 U_CALLCONV
UCaseContextIterator(void *context, int8_t dir);

/**
 * Sample struct which may be used by some implementations of
 * UCaseContextIterator.
 */
struct UCaseContext {
    void *p;
    int32_t start, index, limit;
    int32_t cpStart, cpLimit;
    int8_t dir;
    int8_t b1, b2, b3;
};
typedef struct UCaseContext UCaseContext;

U_CDECL_END

#define UCASECONTEXT_INITIALIZER { NULL,  0, 0, 0,  0, 0,  0,  0, 0, 0 }

enum {
    /**
     * For string case mappings, a single character (a code point) is mapped
     * either to itself (in which case in-place mapping functions do nothing),
     * or to another single code point, or to a string.
     * Aside from the string contents, these are indicated with a single int32_t
     * value as follows:
     *
     * Mapping to self: Negative values (~self instead of -self to support U+0000)
     *
     * Mapping to another code point: Positive values >UCASE_MAX_STRING_LENGTH
     *
     * Mapping to a string: The string length (0..UCASE_MAX_STRING_LENGTH) is
     * returned. Note that the string result may indeed have zero length.
     */
    UCASE_MAX_STRING_LENGTH=0x1f
};

/**
 * Get the full lowercase mapping for c.
 *
 * @param csp Case mapping properties.
 * @param c Character to be mapped.
 * @param iter Character iterator, used for context-sensitive mappings.
 *             See UCaseContextIterator for details.
 *             If iter==NULL then a context-independent result is returned.
 * @param context Pointer to be passed into iter.
 * @param pString If the mapping result is a string, then the pointer is
 *                written to *pString.
 * @param caseLocale Case locale value from ucase_getCaseLocale().
 * @return Output code point or string length, see UCASE_MAX_STRING_LENGTH.
 *
 * @see UCaseContextIterator
 * @see UCASE_MAX_STRING_LENGTH
 * @internal
 */
U_CAPI int32_t U_EXPORT2
ucase_toFullLower(UChar32 c,
                  UCaseContextIterator *iter, void *context,
                  const UChar **pString,
                  int32_t caseLocale);

U_CAPI int32_t U_EXPORT2
ucase_toFullUpper(UChar32 c,
                  UCaseContextIterator *iter, void *context,
                  const UChar **pString,
                  int32_t caseLocale);

U_CAPI int32_t U_EXPORT2
ucase_toFullTitle(UChar32 c,
                  UCaseContextIterator *iter, void *context,
                  const UChar **pString,
                  int32_t caseLocale);

U_CAPI int32_t U_EXPORT2
ucase_toFullFolding(UChar32 c,
                    const UChar **pString,
                    uint32_t options);

U_CFUNC int32_t U_EXPORT2
ucase_hasBinaryProperty(UChar32 c, UProperty which);


U_CDECL_BEGIN

/**
 * @internal
 */
typedef int32_t U_CALLCONV
UCaseMapFull(UChar32 c,
             UCaseContextIterator *iter, void *context,
             const UChar **pString,
             int32_t caseLocale);

U_CDECL_END

/* file definitions --------------------------------------------------------- */

#define UCASE_DATA_NAME "ucase"
#define UCASE_DATA_TYPE "icu"

/* format "cAsE" */
#define UCASE_FMT_0 0x63
#define UCASE_FMT_1 0x41
#define UCASE_FMT_2 0x53
#define UCASE_FMT_3 0x45

/* indexes into indexes[] */
enum {
    UCASE_IX_INDEX_TOP,
    UCASE_IX_LENGTH,
    UCASE_IX_TRIE_SIZE,
    UCASE_IX_EXC_LENGTH,
    UCASE_IX_UNFOLD_LENGTH,

    UCASE_IX_MAX_FULL_LENGTH=15,
    UCASE_IX_TOP=16
};

/* definitions for 16-bit case properties word ------------------------------ */

U_CFUNC const UTrie2 * U_EXPORT2
ucase_getTrie();

/* 2-bit constants for types of cased characters */
#define UCASE_TYPE_MASK     3
enum {
    UCASE_NONE,
    UCASE_LOWER,
    UCASE_UPPER,
    UCASE_TITLE
};

#define UCASE_GET_TYPE(props) ((props)&UCASE_TYPE_MASK)
#define UCASE_GET_TYPE_AND_IGNORABLE(props) ((props)&7)

#define UCASE_IS_UPPER_OR_TITLE(props) ((props)&2)

#define UCASE_IGNORABLE         4
#define UCASE_EXCEPTION         8
#define UCASE_SENSITIVE         0x10

#define UCASE_HAS_EXCEPTION(props) ((props)&UCASE_EXCEPTION)

#define UCASE_DOT_MASK      0x60
enum {
    UCASE_NO_DOT=0,         /* normal characters with cc=0 */
    UCASE_SOFT_DOTTED=0x20, /* soft-dotted characters with cc=0 */
    UCASE_ABOVE=0x40,       /* "above" accents with cc=230 */
    UCASE_OTHER_ACCENT=0x60 /* other accent character (0<cc!=230) */
};

/* no exception: bits 15..7 are a 9-bit signed case mapping delta */
#define UCASE_DELTA_SHIFT   7
#define UCASE_DELTA_MASK    0xff80
#define UCASE_MAX_DELTA     0xff
#define UCASE_MIN_DELTA     (-UCASE_MAX_DELTA-1)

#if U_SIGNED_RIGHT_SHIFT_IS_ARITHMETIC
#   define UCASE_GET_DELTA(props) ((int16_t)(props)>>UCASE_DELTA_SHIFT)
#else
#   define UCASE_GET_DELTA(props) (int16_t)(((props)&0x8000) ? (((props)>>UCASE_DELTA_SHIFT)|0xfe00) : ((uint16_t)(props)>>UCASE_DELTA_SHIFT))
#endif

/* exception: bits 15..4 are an unsigned 12-bit index into the exceptions array */
#define UCASE_EXC_SHIFT     4
#define UCASE_EXC_MASK      0xfff0
#define UCASE_MAX_EXCEPTIONS ((UCASE_EXC_MASK>>UCASE_EXC_SHIFT)+1)

/* definitions for 16-bit main exceptions word ------------------------------ */

/* first 8 bits indicate values in optional slots */
enum {
    UCASE_EXC_LOWER,
    UCASE_EXC_FOLD,
    UCASE_EXC_UPPER,
    UCASE_EXC_TITLE,
    UCASE_EXC_DELTA,
    UCASE_EXC_5,            /* reserved */
    UCASE_EXC_CLOSURE,
    UCASE_EXC_FULL_MAPPINGS,
    UCASE_EXC_ALL_SLOTS     /* one past the last slot */
};

/* each slot is 2 uint16_t instead of 1 */
#define UCASE_EXC_DOUBLE_SLOTS      0x100

enum {
    UCASE_EXC_NO_SIMPLE_CASE_FOLDING=0x200,
    UCASE_EXC_DELTA_IS_NEGATIVE=0x400,
    UCASE_EXC_SENSITIVE=0x800
};

/* UCASE_EXC_DOT_MASK=UCASE_DOT_MASK<<UCASE_EXC_DOT_SHIFT */
#define UCASE_EXC_DOT_SHIFT     7

/* normally stored in the main word, but pushed out for larger exception indexes */
#define UCASE_EXC_DOT_MASK      0x3000
enum {
    UCASE_EXC_NO_DOT=0,
    UCASE_EXC_SOFT_DOTTED=0x1000,
    UCASE_EXC_ABOVE=0x2000,         /* "above" accents with cc=230 */
    UCASE_EXC_OTHER_ACCENT=0x3000   /* other character (0<cc!=230) */
};

/* complex/conditional mappings */
#define UCASE_EXC_CONDITIONAL_SPECIAL   0x4000
#define UCASE_EXC_CONDITIONAL_FOLD      0x8000

/* definitions for lengths word for full case mappings */
#define UCASE_FULL_LOWER    0xf
#define UCASE_FULL_FOLDING  0xf0
#define UCASE_FULL_UPPER    0xf00
#define UCASE_FULL_TITLE    0xf000

/* maximum lengths */
#define UCASE_FULL_MAPPINGS_MAX_LENGTH (4*0xf)
#define UCASE_CLOSURE_MAX_LENGTH 0xf

/* constants for reverse case folding ("unfold") data */
enum {
    UCASE_UNFOLD_ROWS,
    UCASE_UNFOLD_ROW_WIDTH,
    UCASE_UNFOLD_STRING_WIDTH
};


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2001-2016, International Business Machines
*                Corporation and others. All Rights Reserved.
******************************************************************************
*   file name:  ucln_cmn.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001July05
*   created by: George Rhoten
*/



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
* Copyright (C) 2001-2013, International Business Machines
*                Corporation and others. All Rights Reserved.
*
******************************************************************************
*   file name:  ucln.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001July05
*   created by: George Rhoten
*/




/** These are the functions used to register a library's memory cleanup
 * functions.  Each library should define a single library register function
 * to call this API.  In the i18n library, it is ucln_i18n_registerCleanup().
 *
 * None of the cleanup functions should use a mutex to clean up an API's
 * allocated memory because a cleanup function is not meant to be thread safe,
 * and plenty of data cannot be reference counted in order to make sure that
 * no one else needs the allocated data.
 *
 * In order to make a cleanup function get called when u_cleanup is called,
 * You should add your function to the library specific cleanup function.
 * If the cleanup function is not in the common library, the code that
 * allocates the memory should call the library specific cleanup function.
 * For instance, in the i18n library, any memory allocated statically must
 * call ucln_i18n_registerCleanup() from the ucln_in.h header.  These library
 * cleanup functions are needed in order to prevent a circular dependency
 * between the common library and any other library.
 *
 * The order of the cleanup is very important.  In general, an API that
 * depends on a second API should be cleaned up before the second API.
 * For instance, the default converter in ustring depends upon the converter
 * API.  So the default converter should be closed before the converter API
 * has its cache flushed.  This will prevent any memory leaks due to
 * reference counting.
 *
 * Please see common/ucln_cmn.{h,c} and i18n/ucln_in.{h,c} for examples.
 */

/**
 * Data Type for cleanup function selector. These roughly correspond to libraries.
 */
typedef enum ECleanupLibraryType {
    UCLN_START = -1,
    UCLN_UPLUG,     /* ICU plugins */
    UCLN_CUSTOM,    /* Custom is for anyone else. */
    UCLN_CTESTFW,
    UCLN_TOOLUTIL,
    UCLN_LAYOUTEX,
    UCLN_LAYOUT,
    UCLN_IO,
    UCLN_I18N,
    UCLN_COMMON /* This must be the last one to cleanup. */
} ECleanupLibraryType;

/**
 * Data type for cleanup function pointer
 */
U_CDECL_BEGIN
typedef UBool U_CALLCONV cleanupFunc(void);
typedef void U_CALLCONV initFunc(UErrorCode *);
U_CDECL_END

/**
 * Register a cleanup function
 * @param type which library to register for.
 * @param func the function pointer
 */
U_CAPI void U_EXPORT2 ucln_registerCleanup(ECleanupLibraryType type,
                                           cleanupFunc *func);

/**
 * Request cleanup for one specific library.
 * Not thread safe.
 * @param type which library to cleanup
 */
U_CAPI void U_EXPORT2 ucln_cleanupOne(ECleanupLibraryType type);



/* These are the cleanup functions for various APIs. */
/* @return true if cleanup complete successfully.*/
U_CFUNC UBool utrace_cleanup(void);

U_CFUNC UBool ucln_lib_cleanup(void);

/*
Please keep the order of enums declared in same order
as the cleanup functions are suppose to be called. */
typedef enum ECleanupCommonType {
    UCLN_COMMON_START = -1,
    UCLN_COMMON_NUMPARSE_UNISETS,
    UCLN_COMMON_USPREP,
    UCLN_COMMON_BREAKITERATOR,
    UCLN_COMMON_RBBI,
    UCLN_COMMON_SERVICE,
    UCLN_COMMON_LOCALE_KEY_TYPE,
    UCLN_COMMON_LOCALE,
    UCLN_COMMON_LOCALE_AVAILABLE,
    UCLN_COMMON_LIKELY_SUBTAGS,
    UCLN_COMMON_LOCALE_DISTANCE,
    UCLN_COMMON_ULOC,
    UCLN_COMMON_CURRENCY,
    UCLN_COMMON_LOADED_NORMALIZER2,
    UCLN_COMMON_NORMALIZER2,
    UCLN_COMMON_CHARACTERPROPERTIES,
    UCLN_COMMON_USET,
    UCLN_COMMON_UNAMES,
    UCLN_COMMON_UPROPS,
    UCLN_COMMON_UCNV,
    UCLN_COMMON_UCNV_IO,
    UCLN_COMMON_UDATA,
    UCLN_COMMON_PUTIL,
    UCLN_COMMON_UINIT,

    /*
       Unified caches caches collation stuff. Collation data structures
       contain resource bundles which means that unified cache cleanup
       must happen before resource bundle clean up.
    */
    UCLN_COMMON_UNIFIED_CACHE,
    UCLN_COMMON_URES,
    UCLN_COMMON_MUTEX,    // Mutexes should be the last to be cleaned up.
    UCLN_COMMON_COUNT /* This must be last */
} ECleanupCommonType;

/* Main library cleanup registration function. */
/* See common/ucln.h for details on adding a cleanup function. */
/* Note: the global mutex must not be held when calling this function. */
U_CFUNC void U_EXPORT2 ucln_common_registerCleanup(ECleanupCommonType type,
                                                   cleanupFunc *func);



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2002-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uprops.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2002feb24
*   created by: Markus W. Scherer
*
*   Constants for mostly non-core Unicode character properties
*   stored in uprops.icu.
*/







/* indexes[] entries */
enum {
    UPROPS_PROPS32_INDEX,
    UPROPS_EXCEPTIONS_INDEX,
    UPROPS_EXCEPTIONS_TOP_INDEX,

    UPROPS_ADDITIONAL_TRIE_INDEX,
    UPROPS_ADDITIONAL_VECTORS_INDEX,
    UPROPS_ADDITIONAL_VECTORS_COLUMNS_INDEX,

    UPROPS_SCRIPT_EXTENSIONS_INDEX,

    UPROPS_RESERVED_INDEX_7,
    UPROPS_RESERVED_INDEX_8,

    /* size of the data file (number of 32-bit units after the header) */
    UPROPS_DATA_TOP_INDEX,

    /* maximum values for code values in vector word 0 */
    UPROPS_MAX_VALUES_INDEX=10,
    /* maximum values for code values in vector word 2 */
    UPROPS_MAX_VALUES_2_INDEX,

    UPROPS_INDEX_COUNT=16
};

/* definitions for the main properties words */
enum {
    /* general category shift==0                                0 (5 bits) */
    /* reserved                                                 5 (1 bit) */
    UPROPS_NUMERIC_TYPE_VALUE_SHIFT=6                       /*  6 (10 bits) */
};

#define GET_CATEGORY(props) ((props)&0x1f)
#define CAT_MASK(props) U_MASK(GET_CATEGORY(props))

#define GET_NUMERIC_TYPE_VALUE(props) ((props)>>UPROPS_NUMERIC_TYPE_VALUE_SHIFT)

/* constants for the storage form of numeric types and values */
enum {
    /** No numeric value. */
    UPROPS_NTV_NONE=0,
    /** Decimal digits: nv=0..9 */
    UPROPS_NTV_DECIMAL_START=1,
    /** Other digits: nv=0..9 */
    UPROPS_NTV_DIGIT_START=11,
    /** Small integers: nv=0..154 */
    UPROPS_NTV_NUMERIC_START=21,
    /** Fractions: ((ntv>>4)-12) / ((ntv&0xf)+1) = -1..17 / 1..16 */
    UPROPS_NTV_FRACTION_START=0xb0,
    /**
     * Large integers:
     * ((ntv>>5)-14) * 10^((ntv&0x1f)+2) = (1..9)*(10^2..10^33)
     * (only one significant decimal digit)
     */
    UPROPS_NTV_LARGE_START=0x1e0,
    /**
     * Sexagesimal numbers:
     * ((ntv>>2)-0xbf) * 60^((ntv&3)+1) = (1..9)*(60^1..60^4)
     */
    UPROPS_NTV_BASE60_START=0x300,
    /**
     * Fraction-20 values:
     * frac20 = ntv-0x324 = 0..0x17 -> 1|3|5|7 / 20|40|80|160|320|640
     * numerator: num = 2*(frac20&3)+1
     * denominator: den = 20<<(frac20>>2)
     */
    UPROPS_NTV_FRACTION20_START=UPROPS_NTV_BASE60_START+36,  // 0x300+9*4=0x324
    /**
     * Fraction-32 values:
     * frac32 = ntv-0x34c = 0..15 -> 1|3|5|7 / 32|64|128|256
     * numerator: num = 2*(frac32&3)+1
     * denominator: den = 32<<(frac32>>2)
     */
    UPROPS_NTV_FRACTION32_START=UPROPS_NTV_FRACTION20_START+24,  // 0x324+6*4=0x34c
    /** No numeric value (yet). */
    UPROPS_NTV_RESERVED_START=UPROPS_NTV_FRACTION32_START+16,  // 0x34c+4*4=0x35c

    UPROPS_NTV_MAX_SMALL_INT=UPROPS_NTV_FRACTION_START-UPROPS_NTV_NUMERIC_START-1
};

#define UPROPS_NTV_GET_TYPE(ntv) \
    ((ntv==UPROPS_NTV_NONE) ? U_NT_NONE : \
    (ntv<UPROPS_NTV_DIGIT_START) ?  U_NT_DECIMAL : \
    (ntv<UPROPS_NTV_NUMERIC_START) ? U_NT_DIGIT : \
    U_NT_NUMERIC)

/* number of properties vector words */
#define UPROPS_VECTOR_WORDS     3

/*
 * Properties in vector word 0
 * Bits
 * 31..24   DerivedAge version major/minor one nibble each
 * 23..22   3..1: Bits 21..20 & 7..0 = Script_Extensions index
 *             3: Script value from Script_Extensions
 *             2: Script=Inherited
 *             1: Script=Common
 *             0: Script=bits 21..20 & 7..0
 * 21..20   Bits 9..8 of the UScriptCode, or index to Script_Extensions
 * 19..17   East Asian Width
 * 16.. 8   UBlockCode
 *  7.. 0   UScriptCode, or index to Script_Extensions
 */

/* derived age: one nibble each for major and minor version numbers */
#define UPROPS_AGE_MASK         0xff000000
#define UPROPS_AGE_SHIFT        24

/* Script_Extensions: mask includes Script */
#define UPROPS_SCRIPT_X_MASK    0x00f000ff
#define UPROPS_SCRIPT_X_SHIFT   22

// The UScriptCode or Script_Extensions index is split across two bit fields.
// (Starting with Unicode 13/ICU 66/2019 due to more varied Script_Extensions.)
// Shift the high bits right by 12 to assemble the full value.
#define UPROPS_SCRIPT_HIGH_MASK    0x00300000
#define UPROPS_SCRIPT_HIGH_SHIFT   12
#define UPROPS_MAX_SCRIPT          0x3ff

#define UPROPS_EA_MASK          0x000e0000
#define UPROPS_EA_SHIFT         17

#define UPROPS_BLOCK_MASK       0x0001ff00
#define UPROPS_BLOCK_SHIFT      8

#define UPROPS_SCRIPT_LOW_MASK  0x000000ff

/* UPROPS_SCRIPT_X_WITH_COMMON must be the lowest value that involves Script_Extensions. */
#define UPROPS_SCRIPT_X_WITH_COMMON     0x400000
#define UPROPS_SCRIPT_X_WITH_INHERITED  0x800000
#define UPROPS_SCRIPT_X_WITH_OTHER      0xc00000

#ifdef __cplusplus

namespace {

inline uint32_t uprops_mergeScriptCodeOrIndex(uint32_t scriptX) {
    return
        ((scriptX & UPROPS_SCRIPT_HIGH_MASK) >> UPROPS_SCRIPT_HIGH_SHIFT) |
        (scriptX & UPROPS_SCRIPT_LOW_MASK);
}

}  // namespace

#endif  // __cplusplus

/*
 * Properties in vector word 1
 * Each bit encodes one binary property.
 * The following constants represent the bit number, use 1<<UPROPS_XYZ.
 * UPROPS_BINARY_1_TOP<=32!
 *
 * Keep this list of property enums in sync with
 * propListNames[] in icu/source/tools/genprops/props2.c!
 *
 * ICU 2.6/uprops format version 3.2 stores full properties instead of "Other_".
 */
enum {
    UPROPS_WHITE_SPACE,
    UPROPS_DASH,
    UPROPS_HYPHEN,
    UPROPS_QUOTATION_MARK,
    UPROPS_TERMINAL_PUNCTUATION,
    UPROPS_MATH,
    UPROPS_HEX_DIGIT,
    UPROPS_ASCII_HEX_DIGIT,
    UPROPS_ALPHABETIC,
    UPROPS_IDEOGRAPHIC,
    UPROPS_DIACRITIC,
    UPROPS_EXTENDER,
    UPROPS_NONCHARACTER_CODE_POINT,
    UPROPS_GRAPHEME_EXTEND,
    UPROPS_GRAPHEME_LINK,
    UPROPS_IDS_BINARY_OPERATOR,
    UPROPS_IDS_TRINARY_OPERATOR,
    UPROPS_RADICAL,
    UPROPS_UNIFIED_IDEOGRAPH,
    UPROPS_DEFAULT_IGNORABLE_CODE_POINT,
    UPROPS_DEPRECATED,
    UPROPS_LOGICAL_ORDER_EXCEPTION,
    UPROPS_XID_START,
    UPROPS_XID_CONTINUE,
    UPROPS_ID_START,                            /* ICU 2.6, uprops format version 3.2 */
    UPROPS_ID_CONTINUE,
    UPROPS_GRAPHEME_BASE,
    UPROPS_S_TERM,                              /* new in ICU 3.0 and Unicode 4.0.1 */
    UPROPS_VARIATION_SELECTOR,
    UPROPS_PATTERN_SYNTAX,                      /* new in ICU 3.4 and Unicode 4.1 */
    UPROPS_PATTERN_WHITE_SPACE,
    UPROPS_PREPENDED_CONCATENATION_MARK,        // new in ICU 60 and Unicode 10
    UPROPS_BINARY_1_TOP                         /* ==32 - full! */
};

/*
 * Properties in vector word 2
 * Bits
 * 31..26   http://www.unicode.org/reports/tr51/#Emoji_Properties
 * 25..20   Line Break
 * 19..15   Sentence Break
 * 14..10   Word Break
 *  9.. 5   Grapheme Cluster Break
 *  4.. 0   Decomposition Type
 */
enum {
    UPROPS_2_EXTENDED_PICTOGRAPHIC=26,
    UPROPS_2_EMOJI_COMPONENT,
    UPROPS_2_EMOJI,
    UPROPS_2_EMOJI_PRESENTATION,
    UPROPS_2_EMOJI_MODIFIER,
    UPROPS_2_EMOJI_MODIFIER_BASE
};

#define UPROPS_LB_MASK          0x03f00000
#define UPROPS_LB_SHIFT         20

#define UPROPS_SB_MASK          0x000f8000
#define UPROPS_SB_SHIFT         15

#define UPROPS_WB_MASK          0x00007c00
#define UPROPS_WB_SHIFT         10

#define UPROPS_GCB_MASK         0x000003e0
#define UPROPS_GCB_SHIFT        5

#define UPROPS_DT_MASK          0x0000001f

/**
 * Gets the main properties value for a code point.
 * Implemented in uchar.c for uprops.cpp.
 */
U_CFUNC uint32_t
u_getMainProperties(UChar32 c);

/**
 * Get a properties vector word for a code point.
 * Implemented in uchar.c for uprops.cpp.
 * @return 0 if no data or illegal argument
 */
U_CFUNC uint32_t
u_getUnicodeProperties(UChar32 c, int32_t column);

/**
 * Get the the maximum values for some enum/int properties.
 * Use the same column numbers as for u_getUnicodeProperties().
 * The returned value will contain maximum values stored in the same bit fields
 * as where the enum values are stored in the u_getUnicodeProperties()
 * return values for the same columns.
 *
 * Valid columns are those for properties words that contain enumerated values.
 * (ICU 2.6: columns 0 and 2)
 * For other column numbers, this function will return 0.
 *
 * @internal
 */
U_CFUNC int32_t
uprv_getMaxValues(int32_t column);

/**
 * Checks if c is alphabetic, or a decimal digit; implements UCHAR_POSIX_ALNUM.
 * @internal
 */
U_CFUNC UBool
u_isalnumPOSIX(UChar32 c);

/**
 * Checks if c is in
 * [^\p{space}\p{gc=Control}\p{gc=Surrogate}\p{gc=Unassigned}]
 * with space=\p{Whitespace} and Control=Cc.
 * Implements UCHAR_POSIX_GRAPH.
 * @internal
 */
U_CFUNC UBool
u_isgraphPOSIX(UChar32 c);

/**
 * Checks if c is in \p{graph}\p{blank} - \p{cntrl}.
 * Implements UCHAR_POSIX_PRINT.
 * @internal
 */
U_CFUNC UBool
u_isprintPOSIX(UChar32 c);

/** Turn a bit index into a bit flag. @internal */
#define UPROPS_FLAG(n) ((uint32_t)1<<(n))

/** Flags for general categories in the order of UCharCategory. @internal */
#define _Cn     UPROPS_FLAG(U_GENERAL_OTHER_TYPES)
#define _Lu     UPROPS_FLAG(U_UPPERCASE_LETTER)
#define _Ll     UPROPS_FLAG(U_LOWERCASE_LETTER)
#define _Lt     UPROPS_FLAG(U_TITLECASE_LETTER)
#define _Lm     UPROPS_FLAG(U_MODIFIER_LETTER)
/* #define _Lo     UPROPS_FLAG(U_OTHER_LETTER) -- conflicts with MS Visual Studio 9.0 xiosbase */
#define _Mn     UPROPS_FLAG(U_NON_SPACING_MARK)
#define _Me     UPROPS_FLAG(U_ENCLOSING_MARK)
#define _Mc     UPROPS_FLAG(U_COMBINING_SPACING_MARK)
#define _Nd     UPROPS_FLAG(U_DECIMAL_DIGIT_NUMBER)
#define _Nl     UPROPS_FLAG(U_LETTER_NUMBER)
#define _No     UPROPS_FLAG(U_OTHER_NUMBER)
#define _Zs     UPROPS_FLAG(U_SPACE_SEPARATOR)
#define _Zl     UPROPS_FLAG(U_LINE_SEPARATOR)
#define _Zp     UPROPS_FLAG(U_PARAGRAPH_SEPARATOR)
#define _Cc     UPROPS_FLAG(U_CONTROL_CHAR)
#define _Cf     UPROPS_FLAG(U_FORMAT_CHAR)
#define _Co     UPROPS_FLAG(U_PRIVATE_USE_CHAR)
#define _Cs     UPROPS_FLAG(U_SURROGATE)
#define _Pd     UPROPS_FLAG(U_DASH_PUNCTUATION)
#define _Ps     UPROPS_FLAG(U_START_PUNCTUATION)
/* #define _Pe     UPROPS_FLAG(U_END_PUNCTUATION) -- conflicts with MS Visual Studio 9.0 xlocnum */
/* #define _Pc     UPROPS_FLAG(U_CONNECTOR_PUNCTUATION) -- conflicts with MS Visual Studio 9.0 streambuf */
#define _Po     UPROPS_FLAG(U_OTHER_PUNCTUATION)
#define _Sm     UPROPS_FLAG(U_MATH_SYMBOL)
#define _Sc     UPROPS_FLAG(U_CURRENCY_SYMBOL)
#define _Sk     UPROPS_FLAG(U_MODIFIER_SYMBOL)
#define _So     UPROPS_FLAG(U_OTHER_SYMBOL)
#define _Pi     UPROPS_FLAG(U_INITIAL_PUNCTUATION)
/* #define _Pf     UPROPS_FLAG(U_FINAL_PUNCTUATION) -- conflicts with MS Visual Studio 9.0 streambuf */

/** Some code points. @internal */
enum {
    uprops_TAB     =0x0009,
    uprops_LF      =0x000a,
    uprops_FF      =0x000c,
    uprops_CR      =0x000d,
    uprops_U_A     =0x0041,
    uprops_U_F     =0x0046,
    uprops_U_Z     =0x005a,
    uprops_U_a     =0x0061,
    uprops_U_f     =0x0066,
    uprops_U_z     =0x007a,
    uprops_DEL     =0x007f,
    uprops_NL      =0x0085,
    uprops_NBSP    =0x00a0,
    uprops_CGJ     =0x034f,
    uprops_FIGURESP=0x2007,
    uprops_HAIRSP  =0x200a,
    uprops_ZWNJ    =0x200c,
    uprops_ZWJ     =0x200d,
    uprops_RLM     =0x200f,
    uprops_NNBSP   =0x202f,
    uprops_WJ      =0x2060,
    uprops_INHSWAP =0x206a,
    uprops_NOMDIG  =0x206f,
    uprops_U_FW_A  =0xff21,
    uprops_U_FW_F  =0xff26,
    uprops_U_FW_Z  =0xff3a,
    uprops_U_FW_a  =0xff41,
    uprops_U_FW_f  =0xff46,
    uprops_U_FW_z  =0xff5a,
    uprops_ZWNBSP  =0xfeff
};

/**
 * Get the maximum length of a (regular/1.0/extended) character name.
 * @return 0 if no character names available.
 */
U_CAPI int32_t U_EXPORT2
uprv_getMaxCharNameLength(void);

/**
 * Fills set with characters that are used in Unicode character names.
 * Includes all characters that are used in regular/Unicode 1.0/extended names.
 * Just empties the set if no character names are available.
 * @param sa USetAdder to receive characters.
 */
U_CAPI void U_EXPORT2
uprv_getCharNameCharacters(const USetAdder *sa);

/**
 * Constants for which data and implementation files provide which properties.
 * Used by UnicodeSet for service-specific property enumeration.
 * @internal
 */
enum UPropertySource {
    /** No source, not a supported property. */
    UPROPS_SRC_NONE,
    /** From uchar.c/uprops.icu main trie */
    UPROPS_SRC_CHAR,
    /** From uchar.c/uprops.icu properties vectors trie */
    UPROPS_SRC_PROPSVEC,
    /** From unames.c/unames.icu */
    UPROPS_SRC_NAMES,
    /** From ucase.c/ucase.icu */
    UPROPS_SRC_CASE,
    /** From ubidi_props.c/ubidi.icu */
    UPROPS_SRC_BIDI,
    /** From uchar.c/uprops.icu main trie as well as properties vectors trie */
    UPROPS_SRC_CHAR_AND_PROPSVEC,
    /** From ucase.c/ucase.icu as well as unorm.cpp/unorm.icu */
    UPROPS_SRC_CASE_AND_NORM,
    /** From normalizer2impl.cpp/nfc.nrm */
    UPROPS_SRC_NFC,
    /** From normalizer2impl.cpp/nfkc.nrm */
    UPROPS_SRC_NFKC,
    /** From normalizer2impl.cpp/nfkc_cf.nrm */
    UPROPS_SRC_NFKC_CF,
    /** From normalizer2impl.cpp/nfc.nrm canonical iterator data */
    UPROPS_SRC_NFC_CANON_ITER,
    // Text layout properties.
    UPROPS_SRC_INPC,
    UPROPS_SRC_INSC,
    UPROPS_SRC_VO,
    /** One more than the highest UPropertySource (UPROPS_SRC_) constant. */
    UPROPS_SRC_COUNT
};
typedef enum UPropertySource UPropertySource;

/**
 * @see UPropertySource
 * @internal
 */
U_CFUNC UPropertySource U_EXPORT2
uprops_getSource(UProperty which);

/**
 * Enumerate uprops.icu's main data trie and add the
 * start of each range of same properties to the set.
 * @internal
 */
U_CFUNC void U_EXPORT2
uchar_addPropertyStarts(const USetAdder *sa, UErrorCode *pErrorCode);

/**
 * Enumerate uprops.icu's properties vectors trie and add the
 * start of each range of same properties to the set.
 * @internal
 */
U_CFUNC void U_EXPORT2
upropsvec_addPropertyStarts(const USetAdder *sa, UErrorCode *pErrorCode);

U_CFUNC void U_EXPORT2
uprops_addPropertyStarts(UPropertySource src, const USetAdder *sa, UErrorCode *pErrorCode);

/**
 * Return a set of characters for property enumeration.
 * For each two consecutive characters (start, limit) in the set,
 * all of the properties for start..limit-1 are all the same.
 *
 * @param sa USetAdder to receive result. Existing contents are lost.
 * @internal
 */
/*U_CFUNC void U_EXPORT2
uprv_getInclusions(const USetAdder *sa, UErrorCode *pErrorCode);
*/

/**
 * Swap the ICU Unicode character names file. See uchar.c.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
uchar_swapNames(const UDataSwapper *ds,
                const void *inData, int32_t length, void *outData,
                UErrorCode *pErrorCode);

#ifdef __cplusplus

U_NAMESPACE_BEGIN

class UnicodeSet;

class CharacterProperties {
public:
    CharacterProperties() = delete;
    static const UnicodeSet *getInclusionsForProperty(UProperty prop, UErrorCode &errorCode);
};

// implemented in uniset_props.cpp
U_CFUNC UnicodeSet *
uniset_getUnicode32Instance(UErrorCode &errorCode);

U_NAMESPACE_END

#endif



using icu::LocalPointer;
#if !UCONFIG_NO_NORMALIZATION
using icu::Normalizer2Factory;
using icu::Normalizer2Impl;
#endif
using icu::UInitOnce;
using icu::UnicodeSet;

namespace {

UBool U_CALLCONV characterproperties_cleanup();

constexpr int32_t NUM_INCLUSIONS = UPROPS_SRC_COUNT + UCHAR_INT_LIMIT - UCHAR_INT_START;

struct Inclusion {
    UnicodeSet  *fSet = nullptr;
    UInitOnce    fInitOnce = U_INITONCE_INITIALIZER;
};
Inclusion gInclusions[NUM_INCLUSIONS]; // cached getInclusions()

UnicodeSet *sets[UCHAR_BINARY_LIMIT] = {};

UCPMap *maps[UCHAR_INT_LIMIT - UCHAR_INT_START] = {};

icu::UMutex cpMutex;

//----------------------------------------------------------------
// Inclusions list
//----------------------------------------------------------------

// USetAdder implementation
// Does not use uset.h to reduce code dependencies
void U_CALLCONV
_set_add(USet *set, UChar32 c) {
    ((UnicodeSet *)set)->add(c);
}

void U_CALLCONV
_set_addRange(USet *set, UChar32 start, UChar32 end) {
    ((UnicodeSet *)set)->add(start, end);
}

void U_CALLCONV
_set_addString(USet *set, const UChar *str, int32_t length) {
    ((UnicodeSet *)set)->add(icu::UnicodeString((UBool)(length<0), str, length));
}

UBool U_CALLCONV characterproperties_cleanup() {
    for (Inclusion &in: gInclusions) {
        delete in.fSet;
        in.fSet = nullptr;
        in.fInitOnce.reset();
    }
    for (int32_t i = 0; i < UPRV_LENGTHOF(sets); ++i) {
        delete sets[i];
        sets[i] = nullptr;
    }
    for (int32_t i = 0; i < UPRV_LENGTHOF(maps); ++i) {
        ucptrie_close(reinterpret_cast<UCPTrie *>(maps[i]));
        maps[i] = nullptr;
    }
    return TRUE;
}

void U_CALLCONV initInclusion(UPropertySource src, UErrorCode &errorCode) {
    // This function is invoked only via umtx_initOnce().
    U_ASSERT(0 <= src && src < UPROPS_SRC_COUNT);
    if (src == UPROPS_SRC_NONE) {
        errorCode = U_INTERNAL_PROGRAM_ERROR;
        return;
    }
    U_ASSERT(gInclusions[src].fSet == nullptr);

    LocalPointer<UnicodeSet> incl(new UnicodeSet());
    if (incl.isNull()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    USetAdder sa = {
        (USet *)incl.getAlias(),
        _set_add,
        _set_addRange,
        _set_addString,
        nullptr, // don't need remove()
        nullptr // don't need removeRange()
    };

    switch(src) {
    case UPROPS_SRC_CHAR:
        uchar_addPropertyStarts(&sa, &errorCode);
        break;
    case UPROPS_SRC_PROPSVEC:
        upropsvec_addPropertyStarts(&sa, &errorCode);
        break;
    case UPROPS_SRC_CHAR_AND_PROPSVEC:
        uchar_addPropertyStarts(&sa, &errorCode);
        upropsvec_addPropertyStarts(&sa, &errorCode);
        break;
#if !UCONFIG_NO_NORMALIZATION
    case UPROPS_SRC_CASE_AND_NORM: {
        const Normalizer2Impl *impl=Normalizer2Factory::getNFCImpl(errorCode);
        if(U_SUCCESS(errorCode)) {
            impl->addPropertyStarts(&sa, errorCode);
        }
        ucase_addPropertyStarts(&sa, &errorCode);
        break;
    }
    case UPROPS_SRC_NFC: {
        const Normalizer2Impl *impl=Normalizer2Factory::getNFCImpl(errorCode);
        if(U_SUCCESS(errorCode)) {
            impl->addPropertyStarts(&sa, errorCode);
        }
        break;
    }
    case UPROPS_SRC_NFKC: {
        const Normalizer2Impl *impl=Normalizer2Factory::getNFKCImpl(errorCode);
        if(U_SUCCESS(errorCode)) {
            impl->addPropertyStarts(&sa, errorCode);
        }
        break;
    }
    case UPROPS_SRC_NFKC_CF: {
        const Normalizer2Impl *impl=Normalizer2Factory::getNFKC_CFImpl(errorCode);
        if(U_SUCCESS(errorCode)) {
            impl->addPropertyStarts(&sa, errorCode);
        }
        break;
    }
    case UPROPS_SRC_NFC_CANON_ITER: {
        const Normalizer2Impl *impl=Normalizer2Factory::getNFCImpl(errorCode);
        if(U_SUCCESS(errorCode)) {
            impl->addCanonIterPropertyStarts(&sa, errorCode);
        }
        break;
    }
#endif
    case UPROPS_SRC_CASE:
        ucase_addPropertyStarts(&sa, &errorCode);
        break;
    case UPROPS_SRC_BIDI:
        ubidi_addPropertyStarts(&sa, &errorCode);
        break;
    case UPROPS_SRC_INPC:
    case UPROPS_SRC_INSC:
    case UPROPS_SRC_VO:
        uprops_addPropertyStarts((UPropertySource)src, &sa, &errorCode);
        break;
    default:
        errorCode = U_INTERNAL_PROGRAM_ERROR;
        break;
    }

    if (U_FAILURE(errorCode)) {
        return;
    }
    if (incl->isBogus()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    // Compact for caching.
    incl->compact();
    gInclusions[src].fSet = incl.orphan();
    ucln_common_registerCleanup(UCLN_COMMON_CHARACTERPROPERTIES, characterproperties_cleanup);
}

const UnicodeSet *getInclusionsForSource(UPropertySource src, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return nullptr; }
    if (src < 0 || UPROPS_SRC_COUNT <= src) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    Inclusion &i = gInclusions[src];
    umtx_initOnce(i.fInitOnce, &initInclusion, src, errorCode);
    return i.fSet;
}

void U_CALLCONV initIntPropInclusion(UProperty prop, UErrorCode &errorCode) {
    // This function is invoked only via umtx_initOnce().
    U_ASSERT(UCHAR_INT_START <= prop && prop < UCHAR_INT_LIMIT);
    int32_t inclIndex = UPROPS_SRC_COUNT + prop - UCHAR_INT_START;
    U_ASSERT(gInclusions[inclIndex].fSet == nullptr);
    UPropertySource src = uprops_getSource(prop);
    const UnicodeSet *incl = getInclusionsForSource(src, errorCode);
    if (U_FAILURE(errorCode)) {
        return;
    }

    LocalPointer<UnicodeSet> intPropIncl(new UnicodeSet(0, 0));
    if (intPropIncl.isNull()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    int32_t numRanges = incl->getRangeCount();
    int32_t prevValue = 0;
    for (int32_t i = 0; i < numRanges; ++i) {
        UChar32 rangeEnd = incl->getRangeEnd(i);
        for (UChar32 c = incl->getRangeStart(i); c <= rangeEnd; ++c) {
            // TODO: Get a UCharacterProperty.IntProperty to avoid the property dispatch.
            int32_t value = u_getIntPropertyValue(c, prop);
            if (value != prevValue) {
                intPropIncl->add(c);
                prevValue = value;
            }
        }
    }

    if (intPropIncl->isBogus()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    // Compact for caching.
    intPropIncl->compact();
    gInclusions[inclIndex].fSet = intPropIncl.orphan();
    ucln_common_registerCleanup(UCLN_COMMON_CHARACTERPROPERTIES, characterproperties_cleanup);
}

}  // namespace

U_NAMESPACE_BEGIN

const UnicodeSet *CharacterProperties::getInclusionsForProperty(
        UProperty prop, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return nullptr; }
    if (UCHAR_INT_START <= prop && prop < UCHAR_INT_LIMIT) {
        int32_t inclIndex = UPROPS_SRC_COUNT + prop - UCHAR_INT_START;
        Inclusion &i = gInclusions[inclIndex];
        umtx_initOnce(i.fInitOnce, &initIntPropInclusion, prop, errorCode);
        return i.fSet;
    } else {
        UPropertySource src = uprops_getSource(prop);
        return getInclusionsForSource(src, errorCode);
    }
}

U_NAMESPACE_END

namespace {

UnicodeSet *makeSet(UProperty property, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return nullptr; }
    LocalPointer<UnicodeSet> set(new UnicodeSet());
    if (set.isNull()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    const UnicodeSet *inclusions =
        icu::CharacterProperties::getInclusionsForProperty(property, errorCode);
    if (U_FAILURE(errorCode)) { return nullptr; }
    int32_t numRanges = inclusions->getRangeCount();
    UChar32 startHasProperty = -1;

    for (int32_t i = 0; i < numRanges; ++i) {
        UChar32 rangeEnd = inclusions->getRangeEnd(i);
        for (UChar32 c = inclusions->getRangeStart(i); c <= rangeEnd; ++c) {
            // TODO: Get a UCharacterProperty.BinaryProperty to avoid the property dispatch.
            if (u_hasBinaryProperty(c, property)) {
                if (startHasProperty < 0) {
                    // Transition from false to true.
                    startHasProperty = c;
                }
            } else if (startHasProperty >= 0) {
                // Transition from true to false.
                set->add(startHasProperty, c - 1);
                startHasProperty = -1;
            }
        }
    }
    if (startHasProperty >= 0) {
        set->add(startHasProperty, 0x10FFFF);
    }
    set->freeze();
    return set.orphan();
}

UCPMap *makeMap(UProperty property, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return nullptr; }
    uint32_t nullValue = property == UCHAR_SCRIPT ? USCRIPT_UNKNOWN : 0;
    icu::LocalUMutableCPTriePointer mutableTrie(
        umutablecptrie_open(nullValue, nullValue, &errorCode));
    const UnicodeSet *inclusions =
        icu::CharacterProperties::getInclusionsForProperty(property, errorCode);
    if (U_FAILURE(errorCode)) { return nullptr; }
    int32_t numRanges = inclusions->getRangeCount();
    UChar32 start = 0;
    uint32_t value = nullValue;

    for (int32_t i = 0; i < numRanges; ++i) {
        UChar32 rangeEnd = inclusions->getRangeEnd(i);
        for (UChar32 c = inclusions->getRangeStart(i); c <= rangeEnd; ++c) {
            // TODO: Get a UCharacterProperty.IntProperty to avoid the property dispatch.
            uint32_t nextValue = u_getIntPropertyValue(c, property);
            if (value != nextValue) {
                if (value != nullValue) {
                    umutablecptrie_setRange(mutableTrie.getAlias(), start, c - 1, value, &errorCode);
                }
                start = c;
                value = nextValue;
            }
        }
    }
    if (value != 0) {
        umutablecptrie_setRange(mutableTrie.getAlias(), start, 0x10FFFF, value, &errorCode);
    }

    UCPTrieType type;
    if (property == UCHAR_BIDI_CLASS || property == UCHAR_GENERAL_CATEGORY) {
        type = UCPTRIE_TYPE_FAST;
    } else {
        type = UCPTRIE_TYPE_SMALL;
    }
    UCPTrieValueWidth valueWidth;
    // TODO: UCharacterProperty.IntProperty
    int32_t max = u_getIntPropertyMaxValue(property);
    if (max <= 0xff) {
        valueWidth = UCPTRIE_VALUE_BITS_8;
    } else if (max <= 0xffff) {
        valueWidth = UCPTRIE_VALUE_BITS_16;
    } else {
        valueWidth = UCPTRIE_VALUE_BITS_32;
    }
    return reinterpret_cast<UCPMap *>(
        umutablecptrie_buildImmutable(mutableTrie.getAlias(), type, valueWidth, &errorCode));
}

}  // namespace

U_NAMESPACE_USE

U_CAPI const USet * U_EXPORT2
u_getBinaryPropertySet(UProperty property, UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) { return nullptr; }
    if (property < 0 || UCHAR_BINARY_LIMIT <= property) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    Mutex m(&cpMutex);
    UnicodeSet *set = sets[property];
    if (set == nullptr) {
        sets[property] = set = makeSet(property, *pErrorCode);
    }
    if (U_FAILURE(*pErrorCode)) { return nullptr; }
    return set->toUSet();
}

U_CAPI const UCPMap * U_EXPORT2
u_getIntPropertyMap(UProperty property, UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) { return nullptr; }
    if (property < UCHAR_INT_START || UCHAR_INT_LIMIT <= property) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    Mutex m(&cpMutex);
    UCPMap *map = maps[property - UCHAR_INT_START];
    if (map == nullptr) {
        maps[property - UCHAR_INT_START] = map = makeMap(property, *pErrorCode);
    }
    return map;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1999-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************
*
*   Copyright (C) 1997-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
********************************************************************
*/




#if U_SHOW_CPLUSPLUS_API



/**
 * \file
 * \brief C++ API: Character Iterator
 */
 
U_NAMESPACE_BEGIN
/**
 * Abstract class that defines an API for forward-only iteration
 * on text objects.
 * This is a minimal interface for iteration without random access
 * or backwards iteration. It is especially useful for wrapping
 * streams with converters into an object for collation or
 * normalization.
 *
 * <p>Characters can be accessed in two ways: as code units or as
 * code points.
 * Unicode code points are 21-bit integers and are the scalar values
 * of Unicode characters. ICU uses the type UChar32 for them.
 * Unicode code units are the storage units of a given
 * Unicode/UCS Transformation Format (a character encoding scheme).
 * With UTF-16, all code points can be represented with either one
 * or two code units ("surrogates").
 * String storage is typically based on code units, while properties
 * of characters are typically determined using code point values.
 * Some processes may be designed to work with sequences of code units,
 * or it may be known that all characters that are important to an
 * algorithm can be represented with single code units.
 * Other processes will need to use the code point access functions.</p>
 *
 * <p>ForwardCharacterIterator provides nextPostInc() to access
 * a code unit and advance an internal position into the text object,
 * similar to a <code>return text[position++]</code>.<br>
 * It provides next32PostInc() to access a code point and advance an internal
 * position.</p>
 *
 * <p>next32PostInc() assumes that the current position is that of
 * the beginning of a code point, i.e., of its first code unit.
 * After next32PostInc(), this will be true again.
 * In general, access to code units and code points in the same
 * iteration loop should not be mixed. In UTF-16, if the current position
 * is on a second code unit (Low Surrogate), then only that code unit
 * is returned even by next32PostInc().</p>
 *
 * <p>For iteration with either function, there are two ways to
 * check for the end of the iteration. When there are no more
 * characters in the text object:
 * <ul>
 * <li>The hasNext() function returns FALSE.</li>
 * <li>nextPostInc() and next32PostInc() return DONE
 *     when one attempts to read beyond the end of the text object.</li>
 * </ul>
 *
 * Example:
 * \code 
 * void function1(ForwardCharacterIterator &it) {
 *     UChar32 c;
 *     while(it.hasNext()) {
 *         c=it.next32PostInc();
 *         // use c
 *     }
 * }
 *
 * void function1(ForwardCharacterIterator &it) {
 *     char16_t c;
 *     while((c=it.nextPostInc())!=ForwardCharacterIterator::DONE) {
 *         // use c
 *      }
 *  }
 * \endcode
 * </p>
 *
 * @stable ICU 2.0
 */
class U_COMMON_API ForwardCharacterIterator : public UObject {
public:
    /**
     * Value returned by most of ForwardCharacterIterator's functions
     * when the iterator has reached the limits of its iteration.
     * @stable ICU 2.0
     */
    enum { DONE = 0xffff };
    
    /**
     * Destructor.  
     * @stable ICU 2.0
     */
    virtual ~ForwardCharacterIterator();
    
    /**
     * Returns true when both iterators refer to the same
     * character in the same character-storage object.  
     * @param that The ForwardCharacterIterator to be compared for equality
     * @return true when both iterators refer to the same
     * character in the same character-storage object
     * @stable ICU 2.0
     */
    virtual UBool operator==(const ForwardCharacterIterator& that) const = 0;
    
    /**
     * Returns true when the iterators refer to different
     * text-storage objects, or to different characters in the
     * same text-storage object.  
     * @param that The ForwardCharacterIterator to be compared for inequality
     * @return true when the iterators refer to different
     * text-storage objects, or to different characters in the
     * same text-storage object
     * @stable ICU 2.0
     */
    inline UBool operator!=(const ForwardCharacterIterator& that) const;
    
    /**
     * Generates a hash code for this iterator.  
     * @return the hash code.
     * @stable ICU 2.0
     */
    virtual int32_t hashCode(void) const = 0;
    
    /**
     * Returns a UClassID for this ForwardCharacterIterator ("poor man's
     * RTTI").<P> Despite the fact that this function is public,
     * DO NOT CONSIDER IT PART OF CHARACTERITERATOR'S API! 
     * @return a UClassID for this ForwardCharacterIterator 
     * @stable ICU 2.0
     */
    virtual UClassID getDynamicClassID(void) const = 0;
    
    /**
     * Gets the current code unit for returning and advances to the next code unit
     * in the iteration range
     * (toward endIndex()).  If there are
     * no more code units to return, returns DONE.
     * @return the current code unit.
     * @stable ICU 2.0
     */
    virtual char16_t         nextPostInc(void) = 0;
    
    /**
     * Gets the current code point for returning and advances to the next code point
     * in the iteration range
     * (toward endIndex()).  If there are
     * no more code points to return, returns DONE.
     * @return the current code point.
     * @stable ICU 2.0
     */
    virtual UChar32       next32PostInc(void) = 0;
    
    /**
     * Returns FALSE if there are no more code units or code points
     * at or after the current position in the iteration range.
     * This is used with nextPostInc() or next32PostInc() in forward
     * iteration.
     * @returns FALSE if there are no more code units or code points
     * at or after the current position in the iteration range.
     * @stable ICU 2.0
     */
    virtual UBool        hasNext() = 0;
    
protected:
    /** Default constructor to be overridden in the implementing class. @stable ICU 2.0*/
    ForwardCharacterIterator();
    
    /** Copy constructor to be overridden in the implementing class. @stable ICU 2.0*/
    ForwardCharacterIterator(const ForwardCharacterIterator &other);
    
    /**
     * Assignment operator to be overridden in the implementing class.
     * @stable ICU 2.0
     */
    ForwardCharacterIterator &operator=(const ForwardCharacterIterator&) { return *this; }
};

/**
 * Abstract class that defines an API for iteration
 * on text objects.
 * This is an interface for forward and backward iteration
 * and random access into a text object.
 *
 * <p>The API provides backward compatibility to the Java and older ICU
 * CharacterIterator classes but extends them significantly:
 * <ol>
 * <li>CharacterIterator is now a subclass of ForwardCharacterIterator.</li>
 * <li>While the old API functions provided forward iteration with
 *     "pre-increment" semantics, the new one also provides functions
 *     with "post-increment" semantics. They are more efficient and should
 *     be the preferred iterator functions for new implementations.
 *     The backward iteration always had "pre-decrement" semantics, which
 *     are efficient.</li>
 * <li>Just like ForwardCharacterIterator, it provides access to
 *     both code units and code points. Code point access versions are available
 *     for the old and the new iteration semantics.</li>
 * <li>There are new functions for setting and moving the current position
 *     without returning a character, for efficiency.</li>
 * </ol>
 *
 * See ForwardCharacterIterator for examples for using the new forward iteration
 * functions. For backward iteration, there is also a hasPrevious() function
 * that can be used analogously to hasNext().
 * The old functions work as before and are shown below.</p>
 *
 * <p>Examples for some of the new functions:</p>
 *
 * Forward iteration with hasNext():
 * \code
 * void forward1(CharacterIterator &it) {
 *     UChar32 c;
 *     for(it.setToStart(); it.hasNext();) {
 *         c=it.next32PostInc();
 *         // use c
 *     }
 *  }
 * \endcode
 * Forward iteration more similar to loops with the old forward iteration,
 * showing a way to convert simple for() loops:
 * \code
 * void forward2(CharacterIterator &it) {
 *     char16_t c;
 *     for(c=it.firstPostInc(); c!=CharacterIterator::DONE; c=it.nextPostInc()) {
 *          // use c
 *      }
 * }
 * \endcode
 * Backward iteration with setToEnd() and hasPrevious():
 * \code
 *  void backward1(CharacterIterator &it) {
 *      UChar32 c;
 *      for(it.setToEnd(); it.hasPrevious();) {
 *         c=it.previous32();
 *          // use c
 *      }
 *  }
 * \endcode
 * Backward iteration with a more traditional for() loop:
 * \code
 * void backward2(CharacterIterator &it) {
 *     char16_t c;
 *     for(c=it.last(); c!=CharacterIterator::DONE; c=it.previous()) {
 *         // use c
 *      }
 *  }
 * \endcode
 *
 * Example for random access:
 * \code
 *  void random(CharacterIterator &it) {
 *      // set to the third code point from the beginning
 *      it.move32(3, CharacterIterator::kStart);
 *      // get a code point from here without moving the position
 *      UChar32 c=it.current32();
 *      // get the position
 *      int32_t pos=it.getIndex();
 *      // get the previous code unit
 *      char16_t u=it.previous();
 *      // move back one more code unit
 *      it.move(-1, CharacterIterator::kCurrent);
 *      // set the position back to where it was
 *      // and read the same code point c and move beyond it
 *      it.setIndex(pos);
 *      if(c!=it.next32PostInc()) {
 *          exit(1); // CharacterIterator inconsistent
 *      }
 *  }
 * \endcode
 *
 * <p>Examples, especially for the old API:</p>
 *
 * Function processing characters, in this example simple output
 * <pre>
 * \code
 *  void processChar( char16_t c )
 *  {
 *      cout << " " << c;
 *  }
 * \endcode
 * </pre>
 * Traverse the text from start to finish
 * <pre> 
 * \code
 *  void traverseForward(CharacterIterator& iter)
 *  {
 *      for(char16_t c = iter.first(); c != CharacterIterator.DONE; c = iter.next()) {
 *          processChar(c);
 *      }
 *  }
 * \endcode
 * </pre>
 * Traverse the text backwards, from end to start
 * <pre>
 * \code
 *  void traverseBackward(CharacterIterator& iter)
 *  {
 *      for(char16_t c = iter.last(); c != CharacterIterator.DONE; c = iter.previous()) {
 *          processChar(c);
 *      }
 *  }
 * \endcode
 * </pre>
 * Traverse both forward and backward from a given position in the text. 
 * Calls to notBoundary() in this example represents some additional stopping criteria.
 * <pre>
 * \code
 * void traverseOut(CharacterIterator& iter, int32_t pos)
 * {
 *      char16_t c;
 *      for (c = iter.setIndex(pos);
 *      c != CharacterIterator.DONE && (Unicode::isLetter(c) || Unicode::isDigit(c));
 *          c = iter.next()) {}
 *      int32_t end = iter.getIndex();
 *      for (c = iter.setIndex(pos);
 *          c != CharacterIterator.DONE && (Unicode::isLetter(c) || Unicode::isDigit(c));
 *          c = iter.previous()) {}
 *      int32_t start = iter.getIndex() + 1;
 *  
 *      cout << "start: " << start << " end: " << end << endl;
 *      for (c = iter.setIndex(start); iter.getIndex() < end; c = iter.next() ) {
 *          processChar(c);
 *     }
 *  }
 * \endcode
 * </pre>
 * Creating a StringCharacterIterator and calling the test functions
 * <pre>
 * \code
 *  void CharacterIterator_Example( void )
 *   {
 *       cout << endl << "===== CharacterIterator_Example: =====" << endl;
 *       UnicodeString text("Ein kleiner Satz.");
 *       StringCharacterIterator iterator(text);
 *       cout << "----- traverseForward: -----------" << endl;
 *       traverseForward( iterator );
 *       cout << endl << endl << "----- traverseBackward: ----------" << endl;
 *       traverseBackward( iterator );
 *       cout << endl << endl << "----- traverseOut: ---------------" << endl;
 *       traverseOut( iterator, 7 );
 *       cout << endl << endl << "-----" << endl;
 *   }
 * \endcode
 * </pre>
 *
 * @stable ICU 2.0
 */
class U_COMMON_API CharacterIterator : public ForwardCharacterIterator {
public:
    /**
     * Origin enumeration for the move() and move32() functions.
     * @stable ICU 2.0
     */
    enum EOrigin { kStart, kCurrent, kEnd };

    /**
     * Destructor.
     * @stable ICU 2.0
     */
    virtual ~CharacterIterator();

    /**
     * Returns a pointer to a new CharacterIterator of the same
     * concrete class as this one, and referring to the same
     * character in the same text-storage object as this one.  The
     * caller is responsible for deleting the new clone.  
     * @return a pointer to a new CharacterIterator
     * @stable ICU 2.0
     */
    virtual CharacterIterator* clone() const = 0;

    /**
     * Sets the iterator to refer to the first code unit in its
     * iteration range, and returns that code unit.
     * This can be used to begin an iteration with next().
     * @return the first code unit in its iteration range.
     * @stable ICU 2.0
     */
    virtual char16_t         first(void) = 0;

    /**
     * Sets the iterator to refer to the first code unit in its
     * iteration range, returns that code unit, and moves the position
     * to the second code unit. This is an alternative to setToStart()
     * for forward iteration with nextPostInc().
     * @return the first code unit in its iteration range.
     * @stable ICU 2.0
     */
    virtual char16_t         firstPostInc(void);

    /**
     * Sets the iterator to refer to the first code point in its
     * iteration range, and returns that code unit,
     * This can be used to begin an iteration with next32().
     * Note that an iteration with next32PostInc(), beginning with,
     * e.g., setToStart() or firstPostInc(), is more efficient.
     * @return the first code point in its iteration range.
     * @stable ICU 2.0
     */
    virtual UChar32       first32(void) = 0;

    /**
     * Sets the iterator to refer to the first code point in its
     * iteration range, returns that code point, and moves the position
     * to the second code point. This is an alternative to setToStart()
     * for forward iteration with next32PostInc().
     * @return the first code point in its iteration range.
     * @stable ICU 2.0
     */
    virtual UChar32       first32PostInc(void);

    /**
     * Sets the iterator to refer to the first code unit or code point in its
     * iteration range. This can be used to begin a forward
     * iteration with nextPostInc() or next32PostInc().
     * @return the start position of the iteration range
     * @stable ICU 2.0
     */
    inline int32_t    setToStart();

    /**
     * Sets the iterator to refer to the last code unit in its
     * iteration range, and returns that code unit.
     * This can be used to begin an iteration with previous().
     * @return the last code unit.
     * @stable ICU 2.0
     */
    virtual char16_t         last(void) = 0;
        
    /**
     * Sets the iterator to refer to the last code point in its
     * iteration range, and returns that code unit.
     * This can be used to begin an iteration with previous32().
     * @return the last code point.
     * @stable ICU 2.0
     */
    virtual UChar32       last32(void) = 0;

    /**
     * Sets the iterator to the end of its iteration range, just behind
     * the last code unit or code point. This can be used to begin a backward
     * iteration with previous() or previous32().
     * @return the end position of the iteration range
     * @stable ICU 2.0
     */
    inline int32_t    setToEnd();

    /**
     * Sets the iterator to refer to the "position"-th code unit
     * in the text-storage object the iterator refers to, and
     * returns that code unit.  
     * @param position the "position"-th code unit in the text-storage object
     * @return the "position"-th code unit.
     * @stable ICU 2.0
     */
    virtual char16_t         setIndex(int32_t position) = 0;

    /**
     * Sets the iterator to refer to the beginning of the code point
     * that contains the "position"-th code unit
     * in the text-storage object the iterator refers to, and
     * returns that code point.
     * The current position is adjusted to the beginning of the code point
     * (its first code unit).
     * @param position the "position"-th code unit in the text-storage object
     * @return the "position"-th code point.
     * @stable ICU 2.0
     */
    virtual UChar32       setIndex32(int32_t position) = 0;

    /**
     * Returns the code unit the iterator currently refers to. 
     * @return the current code unit. 
     * @stable ICU 2.0
     */
    virtual char16_t         current(void) const = 0;
        
    /**
     * Returns the code point the iterator currently refers to.  
     * @return the current code point.
     * @stable ICU 2.0
     */
    virtual UChar32       current32(void) const = 0;
        
    /**
     * Advances to the next code unit in the iteration range
     * (toward endIndex()), and returns that code unit.  If there are
     * no more code units to return, returns DONE.
     * @return the next code unit.
     * @stable ICU 2.0
     */
    virtual char16_t         next(void) = 0;
        
    /**
     * Advances to the next code point in the iteration range
     * (toward endIndex()), and returns that code point.  If there are
     * no more code points to return, returns DONE.
     * Note that iteration with "pre-increment" semantics is less
     * efficient than iteration with "post-increment" semantics
     * that is provided by next32PostInc().
     * @return the next code point.
     * @stable ICU 2.0
     */
    virtual UChar32       next32(void) = 0;
        
    /**
     * Advances to the previous code unit in the iteration range
     * (toward startIndex()), and returns that code unit.  If there are
     * no more code units to return, returns DONE.  
     * @return the previous code unit.
     * @stable ICU 2.0
     */
    virtual char16_t         previous(void) = 0;

    /**
     * Advances to the previous code point in the iteration range
     * (toward startIndex()), and returns that code point.  If there are
     * no more code points to return, returns DONE. 
     * @return the previous code point. 
     * @stable ICU 2.0
     */
    virtual UChar32       previous32(void) = 0;

    /**
     * Returns FALSE if there are no more code units or code points
     * before the current position in the iteration range.
     * This is used with previous() or previous32() in backward
     * iteration.
     * @return FALSE if there are no more code units or code points
     * before the current position in the iteration range, return TRUE otherwise.
     * @stable ICU 2.0
     */
    virtual UBool        hasPrevious() = 0;

    /**
     * Returns the numeric index in the underlying text-storage
     * object of the character returned by first().  Since it's
     * possible to create an iterator that iterates across only
     * part of a text-storage object, this number isn't
     * necessarily 0.  
     * @returns the numeric index in the underlying text-storage
     * object of the character returned by first().
     * @stable ICU 2.0
     */
    inline int32_t       startIndex(void) const;
        
    /**
     * Returns the numeric index in the underlying text-storage
     * object of the position immediately BEYOND the character
     * returned by last().  
     * @return the numeric index in the underlying text-storage
     * object of the position immediately BEYOND the character
     * returned by last().
     * @stable ICU 2.0
     */
    inline int32_t       endIndex(void) const;
        
    /**
     * Returns the numeric index in the underlying text-storage
     * object of the character the iterator currently refers to
     * (i.e., the character returned by current()).  
     * @return the numeric index in the text-storage object of 
     * the character the iterator currently refers to
     * @stable ICU 2.0
     */
    inline int32_t       getIndex(void) const;

    /**
     * Returns the length of the entire text in the underlying
     * text-storage object.
     * @return the length of the entire text in the text-storage object
     * @stable ICU 2.0
     */
    inline int32_t           getLength() const;

    /**
     * Moves the current position relative to the start or end of the
     * iteration range, or relative to the current position itself.
     * The movement is expressed in numbers of code units forward
     * or backward by specifying a positive or negative delta.
     * @param delta the position relative to origin. A positive delta means forward;
     * a negative delta means backward.
     * @param origin Origin enumeration {kStart, kCurrent, kEnd}
     * @return the new position
     * @stable ICU 2.0
     */
    virtual int32_t      move(int32_t delta, EOrigin origin) = 0;

    /**
     * Moves the current position relative to the start or end of the
     * iteration range, or relative to the current position itself.
     * The movement is expressed in numbers of code points forward
     * or backward by specifying a positive or negative delta.
     * @param delta the position relative to origin. A positive delta means forward;
     * a negative delta means backward.
     * @param origin Origin enumeration {kStart, kCurrent, kEnd}
     * @return the new position
     * @stable ICU 2.0
     */
#ifdef move32
     // One of the system headers right now is sometimes defining a conflicting macro we don't use
#undef move32
#endif
    virtual int32_t      move32(int32_t delta, EOrigin origin) = 0;

    /**
     * Copies the text under iteration into the UnicodeString
     * referred to by "result".  
     * @param result Receives a copy of the text under iteration.  
     * @stable ICU 2.0
     */
    virtual void            getText(UnicodeString&  result) = 0;

protected:
    /**
     * Empty constructor.
     * @stable ICU 2.0
     */
    CharacterIterator();

    /**
     * Constructor, just setting the length field in this base class.
     * @stable ICU 2.0
     */
    CharacterIterator(int32_t length);

    /**
     * Constructor, just setting the length and position fields in this base class.
     * @stable ICU 2.0
     */
    CharacterIterator(int32_t length, int32_t position);

    /**
     * Constructor, just setting the length, start, end, and position fields in this base class.
     * @stable ICU 2.0
     */
    CharacterIterator(int32_t length, int32_t textBegin, int32_t textEnd, int32_t position);
  
    /**
     * Copy constructor.
     *
     * @param that The CharacterIterator to be copied
     * @stable ICU 2.0
     */
    CharacterIterator(const CharacterIterator &that);

    /**
     * Assignment operator.  Sets this CharacterIterator to have the same behavior,
     * as the one passed in.
     * @param that The CharacterIterator passed in.
     * @return the newly set CharacterIterator.
     * @stable ICU 2.0
     */
    CharacterIterator &operator=(const CharacterIterator &that);

    /**
     * Base class text length field.
     * Necessary this for correct getText() and hashCode().
     * @stable ICU 2.0
     */
    int32_t textLength;

    /**
     * Base class field for the current position.
     * @stable ICU 2.0
     */
    int32_t  pos;

    /**
     * Base class field for the start of the iteration range.
     * @stable ICU 2.0
     */
    int32_t  begin;

    /**
     * Base class field for the end of the iteration range.
     * @stable ICU 2.0
     */
    int32_t  end;
};

inline UBool
ForwardCharacterIterator::operator!=(const ForwardCharacterIterator& that) const {
    return !operator==(that);
}

inline int32_t
CharacterIterator::setToStart() {
    return move(0, kStart);
}

inline int32_t
CharacterIterator::setToEnd() {
    return move(0, kEnd);
}

inline int32_t
CharacterIterator::startIndex(void) const {
    return begin;
}

inline int32_t
CharacterIterator::endIndex(void) const {
    return end;
}

inline int32_t
CharacterIterator::getIndex(void) const {
    return pos;
}

inline int32_t
CharacterIterator::getLength(void) const {
    return textLength;
}

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */



U_NAMESPACE_BEGIN

ForwardCharacterIterator::~ForwardCharacterIterator() {}
ForwardCharacterIterator::ForwardCharacterIterator()
: UObject()
{}
ForwardCharacterIterator::ForwardCharacterIterator(const ForwardCharacterIterator &other)
: UObject(other)
{}


CharacterIterator::CharacterIterator()
: textLength(0), pos(0), begin(0), end(0) {
}

CharacterIterator::CharacterIterator(int32_t length)
: textLength(length), pos(0), begin(0), end(length) {
    if(textLength < 0) {
        textLength = end = 0;
    }
}

CharacterIterator::CharacterIterator(int32_t length, int32_t position)
: textLength(length), pos(position), begin(0), end(length) {
    if(textLength < 0) {
        textLength = end = 0;
    }
    if(pos < 0) {
        pos = 0;
    } else if(pos > end) {
        pos = end;
    }
}

CharacterIterator::CharacterIterator(int32_t length, int32_t textBegin, int32_t textEnd, int32_t position)
: textLength(length), pos(position), begin(textBegin), end(textEnd) {
    if(textLength < 0) {
        textLength = 0;
    }
    if(begin < 0) {
        begin = 0;
    } else if(begin > textLength) {
        begin = textLength;
    }
    if(end < begin) {
        end = begin;
    } else if(end > textLength) {
        end = textLength;
    }
    if(pos < begin) {
        pos = begin;
    } else if(pos > end) {
        pos = end;
    }
}

CharacterIterator::~CharacterIterator() {}

CharacterIterator::CharacterIterator(const CharacterIterator &that) :
ForwardCharacterIterator(that),
textLength(that.textLength), pos(that.pos), begin(that.begin), end(that.end)
{
}

CharacterIterator &
CharacterIterator::operator=(const CharacterIterator &that) {
    ForwardCharacterIterator::operator=(that);
    textLength = that.textLength;
    pos = that.pos;
    begin = that.begin;
    end = that.end;
    return *this;
}

// implementing first[32]PostInc() directly in a subclass should be faster
// but these implementations make subclassing a little easier
UChar
CharacterIterator::firstPostInc(void) {
    setToStart();
    return nextPostInc();
}

UChar32
CharacterIterator::first32PostInc(void) {
    setToStart();
    return next32PostInc();
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  charstr.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010may19
*   created by: Markus W. Scherer
*/





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1997-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File CSTRING.H
*
* Contains CString interface
*
* @author       Helena Shih
*
* Modification History:
*
*   Date        Name        Description
*   6/17/98     hshih       Created.
*  05/03/99     stephen     Changed from functions to macros.
*  06/14/99     stephen     Added icu_strncat, icu_strncmp, icu_tolower
*
******************************************************************************
*/




#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#define uprv_strcpy(dst, src) U_STANDARD_CPP_NAMESPACE  strcpy(dst, src)
#define uprv_strlen(str) U_STANDARD_CPP_NAMESPACE strlen(str)
#define uprv_strcmp(s1, s2) U_STANDARD_CPP_NAMESPACE strcmp(s1, s2)
#define uprv_strcat(dst, src) U_STANDARD_CPP_NAMESPACE strcat(dst, src)
#define uprv_strchr(s, c) U_STANDARD_CPP_NAMESPACE strchr(s, c)
#define uprv_strstr(s, c) U_STANDARD_CPP_NAMESPACE strstr(s, c)
#define uprv_strrchr(s, c) U_STANDARD_CPP_NAMESPACE strrchr(s, c)
#define uprv_strncpy(dst, src, size) U_STANDARD_CPP_NAMESPACE strncpy(dst, src, size)
#define uprv_strncmp(s1, s2, n) U_STANDARD_CPP_NAMESPACE strncmp(s1, s2, n)
#define uprv_strncat(dst, src, n) U_STANDARD_CPP_NAMESPACE strncat(dst, src, n)

/**
 * Is c an ASCII-repertoire letter a-z or A-Z?
 * Note: The implementation is specific to whether ICU is compiled for
 * an ASCII-based or EBCDIC-based machine. There just does not seem to be a better name for this.
 */
U_CAPI UBool U_EXPORT2
uprv_isASCIILetter(char c);

U_CAPI char U_EXPORT2
uprv_toupper(char c);


U_CAPI char U_EXPORT2
uprv_asciitolower(char c);

U_CAPI char U_EXPORT2
uprv_ebcdictolower(char c);

#if U_CHARSET_FAMILY==U_ASCII_FAMILY
#   define uprv_tolower uprv_asciitolower
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
#   define uprv_tolower uprv_ebcdictolower
#else
#   error U_CHARSET_FAMILY is not valid
#endif

#define uprv_strtod(source, end) U_STANDARD_CPP_NAMESPACE strtod(source, end)
#define uprv_strtoul(str, end, base) U_STANDARD_CPP_NAMESPACE strtoul(str, end, base)
#define uprv_strtol(str, end, base) U_STANDARD_CPP_NAMESPACE strtol(str, end, base)

/* Conversion from a digit to the character with radix base from 2-19 */
/* May need to use U_UPPER_ORDINAL*/
#define T_CString_itosOffset(a) ((a)<=9?('0'+(a)):('A'+(a)-10))

U_CAPI char* U_EXPORT2
uprv_strdup(const char *src);

/**
 * uprv_malloc n+1 bytes, and copy n bytes from src into the new string.
 * Terminate with a null at offset n.   If n is -1, works like uprv_strdup
 * @param src
 * @param n length of the input string, not including null.
 * @return new string (owned by caller, use uprv_free to free).
 * @internal
 */
U_CAPI char* U_EXPORT2
uprv_strndup(const char *src, int32_t n);

U_CAPI char* U_EXPORT2
T_CString_toLowerCase(char* str);

U_CAPI char* U_EXPORT2
T_CString_toUpperCase(char* str);

U_CAPI int32_t U_EXPORT2
T_CString_integerToString(char *buffer, int32_t n, int32_t radix);

U_CAPI int32_t U_EXPORT2
T_CString_int64ToString(char *buffer, int64_t n, uint32_t radix);

U_CAPI int32_t U_EXPORT2
T_CString_stringToInteger(const char *integerString, int32_t radix);

/**
 * Case-insensitive, language-independent string comparison
 * limited to the ASCII character repertoire.
 */
U_CAPI int U_EXPORT2
uprv_stricmp(const char *str1, const char *str2);

/**
 * Case-insensitive, language-independent string comparison
 * limited to the ASCII character repertoire.
 */
U_CAPI int U_EXPORT2
uprv_strnicmp(const char *str1, const char *str2, uint32_t n);


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1999-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uinvchar.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:2
*
*   created on: 2004sep14
*   created by: Markus W. Scherer
*
*   Definitions for handling invariant characters, moved here from putil.c
*   for better modularization.
*/



#ifdef __cplusplus

#endif

/**
 * Check if a char string only contains invariant characters.
 * See utypes.h for details.
 *
 * @param s Input string pointer.
 * @param length Length of the string, can be -1 if NUL-terminated.
 * @return TRUE if s contains only invariant characters.
 *
 * @internal (ICU 2.8)
 */
U_INTERNAL UBool U_EXPORT2
uprv_isInvariantString(const char *s, int32_t length);

/**
 * Check if a Unicode string only contains invariant characters.
 * See utypes.h for details.
 *
 * @param s Input string pointer.
 * @param length Length of the string, can be -1 if NUL-terminated.
 * @return TRUE if s contains only invariant characters.
 *
 * @internal (ICU 2.8)
 */
U_INTERNAL UBool U_EXPORT2
uprv_isInvariantUString(const UChar *s, int32_t length);

/**
 * \def U_UPPER_ORDINAL
 * Get the ordinal number of an uppercase invariant character
 * @internal
 */
#if U_CHARSET_FAMILY==U_ASCII_FAMILY
#   define U_UPPER_ORDINAL(x) ((x)-'A')
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
#   define U_UPPER_ORDINAL(x) (((x) < 'J') ? ((x)-'A') : \
                              (((x) < 'S') ? ((x)-'J'+9) : \
                               ((x)-'S'+18)))
#else
#   error Unknown charset family!
#endif

#ifdef __cplusplus

U_NAMESPACE_BEGIN

/**
 * Like U_UPPER_ORDINAL(x) but with validation.
 * Returns 0..25 for A..Z else a value outside 0..25.
 */
inline int32_t uprv_upperOrdinal(int32_t c) {
#if U_CHARSET_FAMILY==U_ASCII_FAMILY
    return c - 'A';
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
    // EBCDIC: A-Z (26 letters) is split into three ranges A-I (9 letters), J-R (9), S-Z (8).
    // https://en.wikipedia.org/wiki/EBCDIC_037#Codepage_layout
    if (c <= 'I') { return c - 'A'; }  // A-I --> 0-8
    if (c < 'J') { return -1; }
    if (c <= 'R') { return c - 'J' + 9; }  // J-R --> 9..17
    if (c < 'S') { return -1; }
    return c - 'S' + 18;  // S-Z --> 18..25
#else
#   error Unknown charset family!
#endif
}

// Like U_UPPER_ORDINAL(x) but for lowercase and with validation.
// Returns 0..25 for a..z else a value outside 0..25.
inline int32_t uprv_lowerOrdinal(int32_t c) {
#if U_CHARSET_FAMILY==U_ASCII_FAMILY
    return c - 'a';
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
    // EBCDIC: a-z (26 letters) is split into three ranges a-i (9 letters), j-r (9), s-z (8).
    // https://en.wikipedia.org/wiki/EBCDIC_037#Codepage_layout
    if (c <= 'i') { return c - 'a'; }  // a-i --> 0-8
    if (c < 'j') { return -1; }
    if (c <= 'r') { return c - 'j' + 9; }  // j-r --> 9..17
    if (c < 's') { return -1; }
    return c - 's' + 18;  // s-z --> 18..25
#else
#   error Unknown charset family!
#endif
}

U_NAMESPACE_END

#endif

/**
 * Returns true if c == '@' is possible.
 * The @ sign is variant, and the @ sign used on one
 * EBCDIC machine won't be compiled the same way on other EBCDIC based machines.
 * @internal
 */
U_CFUNC UBool
uprv_isEbcdicAtSign(char c);

/**
 * \def uprv_isAtSign
 * Returns true if c == '@' is possible.
 * For ASCII, checks for exactly '@'. For EBCDIC, calls uprv_isEbcdicAtSign().
 * @internal
 */
#if U_CHARSET_FAMILY==U_ASCII_FAMILY
#   define uprv_isAtSign(c) ((c)=='@')
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
#   define uprv_isAtSign(c) uprv_isEbcdicAtSign(c)
#else
#   error Unknown charset family!
#endif

/**
 * Compare two EBCDIC invariant-character strings in ASCII order.
 * @internal
 */
U_INTERNAL int32_t U_EXPORT2
uprv_compareInvEbcdicAsAscii(const char *s1, const char *s2);

/**
 * \def uprv_compareInvCharsAsAscii
 * Compare two invariant-character strings in ASCII order.
 * @internal
 */
#if U_CHARSET_FAMILY==U_ASCII_FAMILY
#   define uprv_compareInvCharsAsAscii(s1, s2) uprv_strcmp(s1, s2)
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
#   define uprv_compareInvCharsAsAscii(s1, s2) uprv_compareInvEbcdicAsAscii(s1, s2)
#else
#   error Unknown charset family!
#endif

/**
 * Converts an EBCDIC invariant character to ASCII.
 * @internal
 */
U_INTERNAL char U_EXPORT2
uprv_ebcdicToAscii(char c);

/**
 * \def uprv_invCharToAscii
 * Converts an invariant character to ASCII.
 * @internal
 */
#if U_CHARSET_FAMILY==U_ASCII_FAMILY
#   define uprv_invCharToAscii(c) (c)
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
#   define uprv_invCharToAscii(c) uprv_ebcdicToAscii(c)
#else
#   error Unknown charset family!
#endif

/**
 * Converts an EBCDIC invariant character to lowercase ASCII.
 * @internal
 */
U_INTERNAL char U_EXPORT2
uprv_ebcdicToLowercaseAscii(char c);

/**
 * \def uprv_invCharToLowercaseAscii
 * Converts an invariant character to lowercase ASCII.
 * @internal
 */
#if U_CHARSET_FAMILY==U_ASCII_FAMILY
#   define uprv_invCharToLowercaseAscii uprv_asciitolower
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
#   define uprv_invCharToLowercaseAscii uprv_ebcdicToLowercaseAscii
#else
#   error Unknown charset family!
#endif

/**
 * Copy EBCDIC to ASCII
 * @internal
 * @see uprv_strncpy
 */
U_INTERNAL uint8_t* U_EXPORT2
uprv_aestrncpy(uint8_t *dst, const uint8_t *src, int32_t n);


/**
 * Copy ASCII to EBCDIC
 * @internal
 * @see uprv_strncpy
 */
U_INTERNAL uint8_t* U_EXPORT2
uprv_eastrncpy(uint8_t *dst, const uint8_t *src, int32_t n);





U_NAMESPACE_BEGIN

CharString::CharString(CharString&& src) U_NOEXCEPT
        : buffer(std::move(src.buffer)), len(src.len) {
    src.len = 0;  // not strictly necessary because we make no guarantees on the source string
}

CharString& CharString::operator=(CharString&& src) U_NOEXCEPT {
    buffer = std::move(src.buffer);
    len = src.len;
    src.len = 0;  // not strictly necessary because we make no guarantees on the source string
    return *this;
}

char *CharString::cloneData(UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) { return nullptr; }
    char *p = static_cast<char *>(uprv_malloc(len + 1));
    if (p == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    uprv_memcpy(p, buffer.getAlias(), len + 1);
    return p;
}

CharString &CharString::copyFrom(const CharString &s, UErrorCode &errorCode) {
    if(U_SUCCESS(errorCode) && this!=&s && ensureCapacity(s.len+1, 0, errorCode)) {
        len=s.len;
        uprv_memcpy(buffer.getAlias(), s.buffer.getAlias(), len+1);
    }
    return *this;
}

int32_t CharString::lastIndexOf(char c) const {
    for(int32_t i=len; i>0;) {
        if(buffer[--i]==c) {
            return i;
        }
    }
    return -1;
}

bool CharString::contains(StringPiece s) const {
    if (s.empty()) { return false; }
    const char *p = buffer.getAlias();
    int32_t lastStart = len - s.length();
    for (int32_t i = 0; i <= lastStart; ++i) {
        if (uprv_memcmp(p + i, s.data(), s.length()) == 0) {
            return true;
        }
    }
    return false;
}

CharString &CharString::truncate(int32_t newLength) {
    if(newLength<0) {
        newLength=0;
    }
    if(newLength<len) {
        buffer[len=newLength]=0;
    }
    return *this;
}

CharString &CharString::append(char c, UErrorCode &errorCode) {
    if(ensureCapacity(len+2, 0, errorCode)) {
        buffer[len++]=c;
        buffer[len]=0;
    }
    return *this;
}

CharString &CharString::append(const char *s, int32_t sLength, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return *this;
    }
    if(sLength<-1 || (s==NULL && sLength!=0)) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return *this;
    }
    if(sLength<0) {
        sLength= static_cast<int32_t>(uprv_strlen(s));
    }
    if(sLength>0) {
        if(s==(buffer.getAlias()+len)) {
            // The caller wrote into the getAppendBuffer().
            if(sLength>=(buffer.getCapacity()-len)) {
                // The caller wrote too much.
                errorCode=U_INTERNAL_PROGRAM_ERROR;
            } else {
                buffer[len+=sLength]=0;
            }
        } else if(buffer.getAlias()<=s && s<(buffer.getAlias()+len) &&
                  sLength>=(buffer.getCapacity()-len)
        ) {
            // (Part of) this string is appended to itself which requires reallocation,
            // so we have to make a copy of the substring and append that.
            return append(CharString(s, sLength, errorCode), errorCode);
        } else if(ensureCapacity(len+sLength+1, 0, errorCode)) {
            uprv_memcpy(buffer.getAlias()+len, s, sLength);
            buffer[len+=sLength]=0;
        }
    }
    return *this;
}

char *CharString::getAppendBuffer(int32_t minCapacity,
                                  int32_t desiredCapacityHint,
                                  int32_t &resultCapacity,
                                  UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        resultCapacity=0;
        return NULL;
    }
    int32_t appendCapacity=buffer.getCapacity()-len-1;  // -1 for NUL
    if(appendCapacity>=minCapacity) {
        resultCapacity=appendCapacity;
        return buffer.getAlias()+len;
    }
    if(ensureCapacity(len+minCapacity+1, len+desiredCapacityHint+1, errorCode)) {
        resultCapacity=buffer.getCapacity()-len-1;
        return buffer.getAlias()+len;
    }
    resultCapacity=0;
    return NULL;
}

CharString &CharString::appendInvariantChars(const UnicodeString &s, UErrorCode &errorCode) {
    return appendInvariantChars(s.getBuffer(), s.length(), errorCode);
}

CharString &CharString::appendInvariantChars(const UChar* uchars, int32_t ucharsLen, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return *this;
    }
    if (!uprv_isInvariantUString(uchars, ucharsLen)) {
        errorCode = U_INVARIANT_CONVERSION_ERROR;
        return *this;
    }
    if(ensureCapacity(len+ucharsLen+1, 0, errorCode)) {
        u_UCharsToChars(uchars, buffer.getAlias()+len, ucharsLen);
        len += ucharsLen;
        buffer[len] = 0;
    }
    return *this;
}

UBool CharString::ensureCapacity(int32_t capacity,
                                 int32_t desiredCapacityHint,
                                 UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return FALSE;
    }
    if(capacity>buffer.getCapacity()) {
        if(desiredCapacityHint==0) {
            desiredCapacityHint=capacity+buffer.getCapacity();
        }
        if( (desiredCapacityHint<=capacity || buffer.resize(desiredCapacityHint, len+1)==NULL) &&
            buffer.resize(capacity, len+1)==NULL
        ) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            return FALSE;
        }
    }
    return TRUE;
}

CharString &CharString::appendPathPart(StringPiece s, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return *this;
    }
    if(s.length()==0) {
        return *this;
    }
    char c;
    if(len>0 && (c=buffer[len-1])!=U_FILE_SEP_CHAR && c!=U_FILE_ALT_SEP_CHAR) {
        append(U_FILE_SEP_CHAR, errorCode);
    }
    append(s, errorCode);
    return *this;
}

CharString &CharString::ensureEndsWithFileSeparator(UErrorCode &errorCode) {
    char c;
    if(U_SUCCESS(errorCode) && len>0 &&
            (c=buffer[len-1])!=U_FILE_SEP_CHAR && c!=U_FILE_ALT_SEP_CHAR) {
        append(U_FILE_SEP_CHAR, errorCode);
    }
    return *this;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2002-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File cmemory.c      ICU Heap allocation.
*                     All ICU heap allocation, both for C and C++ new of ICU
*                     class types, comes through these functions.
*
*                     If you have a need to replace ICU allocation, this is the
*                     place to do it.
*
*                     Note that uprv_malloc(0) returns a non-NULL pointer, and
*                     that a subsequent free of that pointer value is a NOP.
*
******************************************************************************
*/




#include <stdlib.h>

/* uprv_malloc(0) returns a pointer to this read-only data. */
static const int32_t zeroMem[] = {0, 0, 0, 0, 0, 0};

/* Function Pointers for user-supplied heap functions  */
static const void     *pContext;
static UMemAllocFn    *pAlloc;
static UMemReallocFn  *pRealloc;
static UMemFreeFn     *pFree;

#if U_DEBUG && defined(UPRV_MALLOC_COUNT)
#include <stdio.h>
static int n=0;
static long b=0; 
#endif

U_CAPI void * U_EXPORT2
uprv_malloc(size_t s) {
#if U_DEBUG && defined(UPRV_MALLOC_COUNT)
#if 1
  putchar('>');
  fflush(stdout);
#else
  fprintf(stderr,"MALLOC\t#%d\t%ul bytes\t%ul total\n", ++n,s,(b+=s)); fflush(stderr);
#endif
#endif
    if (s > 0) {
        if (pAlloc) {
            return (*pAlloc)(pContext, s);
        } else {
            return uprv_default_malloc(s);
        }
    } else {
        return (void *)zeroMem;
    }
}

U_CAPI void * U_EXPORT2
uprv_realloc(void * buffer, size_t size) {
#if U_DEBUG && defined(UPRV_MALLOC_COUNT)
  putchar('~');
  fflush(stdout);
#endif
    if (buffer == zeroMem) {
        return uprv_malloc(size);
    } else if (size == 0) {
        if (pFree) {
            (*pFree)(pContext, buffer);
        } else {
            uprv_default_free(buffer);
        }
        return (void *)zeroMem;
    } else {
        if (pRealloc) {
            return (*pRealloc)(pContext, buffer, size);
        } else {
            return uprv_default_realloc(buffer, size);
        }
    }
}

U_CAPI void U_EXPORT2
uprv_free(void *buffer) {
#if U_DEBUG && defined(UPRV_MALLOC_COUNT)
  putchar('<');
  fflush(stdout);
#endif
    if (buffer != zeroMem) {
        if (pFree) {
            (*pFree)(pContext, buffer);
        } else {
            uprv_default_free(buffer);
        }
    }
}

U_CAPI void * U_EXPORT2
uprv_calloc(size_t num, size_t size) {
    void *mem = NULL;
    size *= num;
    mem = uprv_malloc(size);
    if (mem) {
        uprv_memset(mem, 0, size);
    }
    return mem;
}

U_CAPI void U_EXPORT2
u_setMemoryFunctions(const void *context, UMemAllocFn *a, UMemReallocFn *r, UMemFreeFn *f,  UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return;
    }
    if (a==NULL || r==NULL || f==NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    pContext  = context;
    pAlloc    = a;
    pRealloc  = r;
    pFree     = f;
}


U_CFUNC UBool cmemory_cleanup(void) {
    pContext   = NULL;
    pAlloc     = NULL;
    pRealloc   = NULL;
    pFree      = NULL;
    return TRUE;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2015-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  charstr.cpp
*/




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File: cstr.h
*/








/**
 * ICU-internal class CStr, a small helper class to facilitate passing UnicodeStrings
 * to functions needing (const char *) strings, such as printf().
 *
 * It is intended primarily for use in debugging or in tests. Uses platform 
 * default code page conversion, which will do the best job possible,
 * but may be lossy, depending on the platform.
 *
 * If no other conversion is available, use invariant conversion and substitue
 * '?' for non-invariant characters.
 *
 * Example Usage:
 *   UnicodeString s = whatever;
 *   printf("%s", CStr(s)());
 *
 *   The explicit call to the CStr() constructor creates a temporary object.
 *   Operator () on the temporary object returns a (const char *) pointer.
 *   The lifetime of the (const char *) data is that of the temporary object,
 *   which works well when passing it as a parameter to another function, such as printf.
 */

U_NAMESPACE_BEGIN

class U_COMMON_API CStr : public UMemory {
  public:
    CStr(const UnicodeString &in);
    ~CStr();
    const char * operator ()() const;

  private:
    CharString s;
    CStr(const CStr &other);               //  Forbid copying of this class.
    CStr &operator =(const CStr &other);   //  Forbid assignment.
};

U_NAMESPACE_END






U_NAMESPACE_BEGIN

CStr::CStr(const UnicodeString &in) {
    UErrorCode status = U_ZERO_ERROR;
#if !UCONFIG_NO_CONVERSION || U_CHARSET_IS_UTF8
    int32_t length = in.extract(0, in.length(), static_cast<char *>(NULL), static_cast<uint32_t>(0));
    int32_t resultCapacity = 0;
    char *buf = s.getAppendBuffer(length, length, resultCapacity, status);
    if (U_SUCCESS(status)) {
        in.extract(0, in.length(), buf, resultCapacity);
        s.append(buf, length, status);
    }
#else
    // No conversion available. Convert any invariant characters; substitute '?' for the rest.
    // Note: can't just call u_UCharsToChars() or CharString.appendInvariantChars() on the 
    //       whole string because they require that the entire input be invariant.
    char buf[2];
    for (int i=0; i<in.length(); i = in.moveIndex32(i, 1)) {
        if (uprv_isInvariantUString(in.getBuffer()+i, 1)) {
            u_UCharsToChars(in.getBuffer()+i, buf, 1);
        } else {
            buf[0] = '?';
        }
        s.append(buf, 1, status);
    }
#endif
}

CStr::~CStr() {
}

const char * CStr::operator ()() const {
    return s.data();
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1997-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File CSTRING.C
*
* @author       Helena Shih
*
* Modification History:
*
*   Date        Name        Description
*   6/18/98     hshih       Created
*   09/08/98    stephen     Added include for ctype, for Mac Port
*   11/15/99    helena      Integrated S/390 IEEE changes. 
******************************************************************************
*/



#include <stdlib.h>
#include <stdio.h>





/*
 * We hardcode case conversion for invariant characters to match our expectation
 * and the compiler execution charset.
 * This prevents problems on systems
 * - with non-default casing behavior, like Turkish system locales where
 *   tolower('I') maps to dotless i and toupper('i') maps to dotted I
 * - where there are no lowercase Latin characters at all, or using different
 *   codes (some old EBCDIC codepages)
 *
 * This works because the compiler usually runs on a platform where the execution
 * charset includes all of the invariant characters at their expected
 * code positions, so that the char * string literals in ICU code match
 * the char literals here.
 *
 * Note that the set of lowercase Latin letters is discontiguous in EBCDIC
 * and the set of uppercase Latin letters is discontiguous as well.
 */

U_CAPI UBool U_EXPORT2
uprv_isASCIILetter(char c) {
#if U_CHARSET_FAMILY==U_EBCDIC_FAMILY
    return
        ('a'<=c && c<='i') || ('j'<=c && c<='r') || ('s'<=c && c<='z') ||
        ('A'<=c && c<='I') || ('J'<=c && c<='R') || ('S'<=c && c<='Z');
#else
    return ('a'<=c && c<='z') || ('A'<=c && c<='Z');
#endif
}

U_CAPI char U_EXPORT2
uprv_toupper(char c) {
#if U_CHARSET_FAMILY==U_EBCDIC_FAMILY
    if(('a'<=c && c<='i') || ('j'<=c && c<='r') || ('s'<=c && c<='z')) {
        c=(char)(c+('A'-'a'));
    }
#else
    if('a'<=c && c<='z') {
        c=(char)(c+('A'-'a'));
    }
#endif
    return c;
}


#if 0
/*
 * Commented out because cstring.h defines uprv_tolower() to be
 * the same as either uprv_asciitolower() or uprv_ebcdictolower()
 * to reduce the amount of code to cover with tests.
 *
 * Note that this uprv_tolower() definition is likely to work for most
 * charset families, not just ASCII and EBCDIC, because its #else branch
 * is written generically.
 */
U_CAPI char U_EXPORT2
uprv_tolower(char c) {
#if U_CHARSET_FAMILY==U_EBCDIC_FAMILY
    if(('A'<=c && c<='I') || ('J'<=c && c<='R') || ('S'<=c && c<='Z')) {
        c=(char)(c+('a'-'A'));
    }
#else
    if('A'<=c && c<='Z') {
        c=(char)(c+('a'-'A'));
    }
#endif
    return c;
}
#endif

U_CAPI char U_EXPORT2
uprv_asciitolower(char c) {
    if(0x41<=c && c<=0x5a) {
        c=(char)(c+0x20);
    }
    return c;
}

U_CAPI char U_EXPORT2
uprv_ebcdictolower(char c) {
    if( (0xc1<=(uint8_t)c && (uint8_t)c<=0xc9) ||
        (0xd1<=(uint8_t)c && (uint8_t)c<=0xd9) ||
        (0xe2<=(uint8_t)c && (uint8_t)c<=0xe9)
    ) {
        c=(char)(c-0x40);
    }
    return c;
}


U_CAPI char* U_EXPORT2
T_CString_toLowerCase(char* str)
{
    char* origPtr = str;

    if (str) {
        do
            *str = (char)uprv_tolower(*str);
        while (*(str++));
    }

    return origPtr;
}

U_CAPI char* U_EXPORT2
T_CString_toUpperCase(char* str)
{
    char* origPtr = str;

    if (str) {
        do
            *str = (char)uprv_toupper(*str);
        while (*(str++));
    }

    return origPtr;
}

/*
 * Takes a int32_t and fills in  a char* string with that number "radix"-based.
 * Does not handle negative values (makes an empty string for them).
 * Writes at most 12 chars ("-2147483647" plus NUL).
 * Returns the length of the string (not including the NUL).
 */
U_CAPI int32_t U_EXPORT2
T_CString_integerToString(char* buffer, int32_t v, int32_t radix)
{
    char      tbuf[30];
    int32_t   tbx    = sizeof(tbuf);
    uint8_t   digit;
    int32_t   length = 0;
    uint32_t  uval;
    
    U_ASSERT(radix>=2 && radix<=16);
    uval = (uint32_t) v;
    if(v<0 && radix == 10) {
        /* Only in base 10 do we conside numbers to be signed. */
        uval = (uint32_t)(-v); 
        buffer[length++] = '-';
    }
    
    tbx = sizeof(tbuf)-1;
    tbuf[tbx] = 0;   /* We are generating the digits backwards.  Null term the end. */
    do {
        digit = (uint8_t)(uval % radix);
        tbuf[--tbx] = (char)(T_CString_itosOffset(digit));
        uval  = uval / radix;
    } while (uval != 0);
    
    /* copy converted number into user buffer  */
    uprv_strcpy(buffer+length, tbuf+tbx);
    length += sizeof(tbuf) - tbx -1;
    return length;
}



/*
 * Takes a int64_t and fills in  a char* string with that number "radix"-based.
 * Writes at most 21: chars ("-9223372036854775807" plus NUL).
 * Returns the length of the string, not including the terminating NULL.
 */
U_CAPI int32_t U_EXPORT2
T_CString_int64ToString(char* buffer, int64_t v, uint32_t radix)
{
    char      tbuf[30];
    int32_t   tbx    = sizeof(tbuf);
    uint8_t   digit;
    int32_t   length = 0;
    uint64_t  uval;
    
    U_ASSERT(radix>=2 && radix<=16);
    uval = (uint64_t) v;
    if(v<0 && radix == 10) {
        /* Only in base 10 do we conside numbers to be signed. */
        uval = (uint64_t)(-v); 
        buffer[length++] = '-';
    }
    
    tbx = sizeof(tbuf)-1;
    tbuf[tbx] = 0;   /* We are generating the digits backwards.  Null term the end. */
    do {
        digit = (uint8_t)(uval % radix);
        tbuf[--tbx] = (char)(T_CString_itosOffset(digit));
        uval  = uval / radix;
    } while (uval != 0);
    
    /* copy converted number into user buffer  */
    uprv_strcpy(buffer+length, tbuf+tbx);
    length += sizeof(tbuf) - tbx -1;
    return length;
}


U_CAPI int32_t U_EXPORT2
T_CString_stringToInteger(const char *integerString, int32_t radix)
{
    char *end;
    return uprv_strtoul(integerString, &end, radix);

}

U_CAPI int U_EXPORT2
uprv_stricmp(const char *str1, const char *str2) {
    if(str1==NULL) {
        if(str2==NULL) {
            return 0;
        } else {
            return -1;
        }
    } else if(str2==NULL) {
        return 1;
    } else {
        /* compare non-NULL strings lexically with lowercase */
        int rc;
        unsigned char c1, c2;

        for(;;) {
            c1=(unsigned char)*str1;
            c2=(unsigned char)*str2;
            if(c1==0) {
                if(c2==0) {
                    return 0;
                } else {
                    return -1;
                }
            } else if(c2==0) {
                return 1;
            } else {
                /* compare non-zero characters with lowercase */
                rc=(int)(unsigned char)uprv_tolower(c1)-(int)(unsigned char)uprv_tolower(c2);
                if(rc!=0) {
                    return rc;
                }
            }
            ++str1;
            ++str2;
        }
    }
}

U_CAPI int U_EXPORT2
uprv_strnicmp(const char *str1, const char *str2, uint32_t n) {
    if(str1==NULL) {
        if(str2==NULL) {
            return 0;
        } else {
            return -1;
        }
    } else if(str2==NULL) {
        return 1;
    } else {
        /* compare non-NULL strings lexically with lowercase */
        int rc;
        unsigned char c1, c2;

        for(; n--;) {
            c1=(unsigned char)*str1;
            c2=(unsigned char)*str2;
            if(c1==0) {
                if(c2==0) {
                    return 0;
                } else {
                    return -1;
                }
            } else if(c2==0) {
                return 1;
            } else {
                /* compare non-zero characters with lowercase */
                rc=(int)(unsigned char)uprv_tolower(c1)-(int)(unsigned char)uprv_tolower(c2);
                if(rc!=0) {
                    return rc;
                }
            }
            ++str1;
            ++str2;
        }
    }

    return 0;
}

U_CAPI char* U_EXPORT2
uprv_strdup(const char *src) {
    size_t len = uprv_strlen(src) + 1;
    char *dup = (char *) uprv_malloc(len);

    if (dup) {
        uprv_memcpy(dup, src, len);
    }

    return dup;
}

U_CAPI char* U_EXPORT2
uprv_strndup(const char *src, int32_t n) {
    char *dup;

    if(n < 0) {
        dup = uprv_strdup(src);
    } else {
        dup = (char*)uprv_malloc(n+1);
        if (dup) { 
            uprv_memcpy(dup, src, n);
            dup[n] = 0;
        }
    }

    return dup;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*  
******************************************************************************
*
*   Copyright (C) 2001, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  cwchar.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001may25
*   created by: Markus W. Scherer
*/



#if !U_HAVE_WCSCPY

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*  
******************************************************************************
*
*   Copyright (C) 2001, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  cwchar.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001may25
*   created by: Markus W. Scherer
*
*   This file contains ICU-internal definitions of wchar_t operations.
*   These definitions were moved here from cstring.h so that fewer
*   ICU implementation files include wchar.h.
*/


#include <string.h>
#include <stdlib.h>


/* Do this after utypes.h so that we have U_HAVE_WCHAR_H . */
#if U_HAVE_WCHAR_H
#   include <wchar.h>
#endif

/*===========================================================================*/
/* Wide-character functions                                                  */
/*===========================================================================*/

/* The following are not available on all systems, defined in wchar.h or string.h. */
#if U_HAVE_WCSCPY
#   define uprv_wcscpy wcscpy
#   define uprv_wcscat wcscat
#   define uprv_wcslen wcslen
#else
U_CAPI wchar_t* U_EXPORT2 
uprv_wcscpy(wchar_t *dst, const wchar_t *src);
U_CAPI wchar_t* U_EXPORT2 
uprv_wcscat(wchar_t *dst, const wchar_t *src);
U_CAPI size_t U_EXPORT2 
uprv_wcslen(const wchar_t *src);
#endif

/* The following are part of the ANSI C standard, defined in stdlib.h . */
#define uprv_wcstombs(mbstr, wcstr, count) U_STANDARD_CPP_NAMESPACE wcstombs(mbstr, wcstr, count)
#define uprv_mbstowcs(wcstr, mbstr, count) U_STANDARD_CPP_NAMESPACE mbstowcs(wcstr, mbstr, count)




U_CAPI wchar_t *uprv_wcscat(wchar_t *dst, const wchar_t *src) {
    wchar_t *start=dst;
    while(*dst!=0) {
        ++dst;
    }
    while((*dst=*src)!=0) {
        ++dst;
        ++src;
    }
    return start;
}

U_CAPI wchar_t *uprv_wcscpy(wchar_t *dst, const wchar_t *src) {
    wchar_t *start=dst;
    while((*dst=*src)!=0) {
        ++dst;
        ++src;
    }
    return start;
}

U_CAPI size_t uprv_wcslen(const wchar_t *src) {
    const wchar_t *start=src;
    while(*src!=0) {
        ++src;
    }
    return src-start;
}

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
 *******************************************************************************
 * Copyright (C) 2006-2016, International Business Machines Corporation
 * and others. All Rights Reserved.
 *******************************************************************************
 */

#include <utility>



// #if !UCONFIG_NO_BREAK_ITERATION

// #include "brkeng.h"
// #include "dictbe.h"
// #include "unicode/uniset.h"
// #include "unicode/chariter.h"
// #include "unicode/ubrk.h"
// #include "uvectr32.h"
// #include "uvector.h"
// #include "uassert.h"
// #include "unicode/normlzr.h"
// #include "cmemory.h"
// #include "dictionarydata.h"

// U_NAMESPACE_BEGIN

// /*
//  ******************************************************************
//  */

// DictionaryBreakEngine::DictionaryBreakEngine() {
// }

// DictionaryBreakEngine::~DictionaryBreakEngine() {
// }

// UBool
// DictionaryBreakEngine::handles(UChar32 c) const {
//     return fSet.contains(c);
// }

// int32_t
// DictionaryBreakEngine::findBreaks( UText *text,
//                                  int32_t startPos,
//                                  int32_t endPos,
//                                  UVector32 &foundBreaks ) const {
//     (void)startPos;            // TODO: remove this param?
//     int32_t result = 0;

//     // Find the span of characters included in the set.
//     //   The span to break begins at the current position in the text, and
//     //   extends towards the start or end of the text, depending on 'reverse'.

//     int32_t start = (int32_t)utext_getNativeIndex(text);
//     int32_t current;
//     int32_t rangeStart;
//     int32_t rangeEnd;
//     UChar32 c = utext_current32(text);
//     while((current = (int32_t)utext_getNativeIndex(text)) < endPos && fSet.contains(c)) {
//         utext_next32(text);         // TODO:  recast loop for postincrement
//         c = utext_current32(text);
//     }
//     rangeStart = start;
//     rangeEnd = current;
//     result = divideUpDictionaryRange(text, rangeStart, rangeEnd, foundBreaks);
//     utext_setNativeIndex(text, current);

//     return result;
// }

// void
// DictionaryBreakEngine::setCharacters( const UnicodeSet &set ) {
//     fSet = set;
//     // Compact for caching
//     fSet.compact();
// }

// /*
//  ******************************************************************
//  * PossibleWord
//  */

// // Helper class for improving readability of the Thai/Lao/Khmer word break
// // algorithm. The implementation is completely inline.

// // List size, limited by the maximum number of words in the dictionary
// // that form a nested sequence.
// static const int32_t POSSIBLE_WORD_LIST_MAX = 20;

// class PossibleWord {
// private:
//     // list of word candidate lengths, in increasing length order
//     // TODO: bytes would be sufficient for word lengths.
//     int32_t   count;      // Count of candidates
//     int32_t   prefix;     // The longest match with a dictionary word
//     int32_t   offset;     // Offset in the text of these candidates
//     int32_t   mark;       // The preferred candidate's offset
//     int32_t   current;    // The candidate we're currently looking at
//     int32_t   cuLengths[POSSIBLE_WORD_LIST_MAX];   // Word Lengths, in code units.
//     int32_t   cpLengths[POSSIBLE_WORD_LIST_MAX];   // Word Lengths, in code points.

// public:
//     PossibleWord() : count(0), prefix(0), offset(-1), mark(0), current(0) {}
//     ~PossibleWord() {}

//     // Fill the list of candidates if needed, select the longest, and return the number found
//     int32_t   candidates( UText *text, DictionaryMatcher *dict, int32_t rangeEnd );

//     // Select the currently marked candidate, point after it in the text, and invalidate self
//     int32_t   acceptMarked( UText *text );

//     // Back up from the current candidate to the next shorter one; return TRUE if that exists
//     // and point the text after it
//     UBool     backUp( UText *text );

//     // Return the longest prefix this candidate location shares with a dictionary word
//     // Return value is in code points.
//     int32_t   longestPrefix() { return prefix; }

//     // Mark the current candidate as the one we like
//     void      markCurrent() { mark = current; }

//     // Get length in code points of the marked word.
//     int32_t   markedCPLength() { return cpLengths[mark]; }
// };


// int32_t PossibleWord::candidates( UText *text, DictionaryMatcher *dict, int32_t rangeEnd ) {
//     // TODO: If getIndex is too slow, use offset < 0 and add discardAll()
//     int32_t start = (int32_t)utext_getNativeIndex(text);
//     if (start != offset) {
//         offset = start;
//         count = dict->matches(text, rangeEnd-start, UPRV_LENGTHOF(cuLengths), cuLengths, cpLengths, NULL, &prefix);
//         // Dictionary leaves text after longest prefix, not longest word. Back up.
//         if (count <= 0) {
//             utext_setNativeIndex(text, start);
//         }
//     }
//     if (count > 0) {
//         utext_setNativeIndex(text, start+cuLengths[count-1]);
//     }
//     current = count-1;
//     mark = current;
//     return count;
// }

// int32_t
// PossibleWord::acceptMarked( UText *text ) {
//     utext_setNativeIndex(text, offset + cuLengths[mark]);
//     return cuLengths[mark];
// }


// UBool
// PossibleWord::backUp( UText *text ) {
//     if (current > 0) {
//         utext_setNativeIndex(text, offset + cuLengths[--current]);
//         return TRUE;
//     }
//     return FALSE;
// }

// /*
//  ******************************************************************
//  * ThaiBreakEngine
//  */

// // How many words in a row are "good enough"?
// static const int32_t THAI_LOOKAHEAD = 3;

// // Will not combine a non-word with a preceding dictionary word longer than this
// static const int32_t THAI_ROOT_COMBINE_THRESHOLD = 3;

// // Will not combine a non-word that shares at least this much prefix with a
// // dictionary word, with a preceding word
// static const int32_t THAI_PREFIX_COMBINE_THRESHOLD = 3;

// // Ellision character
// static const int32_t THAI_PAIYANNOI = 0x0E2F;

// // Repeat character
// static const int32_t THAI_MAIYAMOK = 0x0E46;

// // Minimum word size
// static const int32_t THAI_MIN_WORD = 2;

// // Minimum number of characters for two words
// static const int32_t THAI_MIN_WORD_SPAN = THAI_MIN_WORD * 2;

// ThaiBreakEngine::ThaiBreakEngine(DictionaryMatcher *adoptDictionary, UErrorCode &status)
//     : DictionaryBreakEngine(),
//       fDictionary(adoptDictionary)
// {
//     fThaiWordSet.applyPattern(UNICODE_STRING_SIMPLE("[[:Thai:]&[:LineBreak=SA:]]"), status);
//     if (U_SUCCESS(status)) {
//         setCharacters(fThaiWordSet);
//     }
//     fMarkSet.applyPattern(UNICODE_STRING_SIMPLE("[[:Thai:]&[:LineBreak=SA:]&[:M:]]"), status);
//     fMarkSet.add(0x0020);
//     fEndWordSet = fThaiWordSet;
//     fEndWordSet.remove(0x0E31);             // MAI HAN-AKAT
//     fEndWordSet.remove(0x0E40, 0x0E44);     // SARA E through SARA AI MAIMALAI
//     fBeginWordSet.add(0x0E01, 0x0E2E);      // KO KAI through HO NOKHUK
//     fBeginWordSet.add(0x0E40, 0x0E44);      // SARA E through SARA AI MAIMALAI
//     fSuffixSet.add(THAI_PAIYANNOI);
//     fSuffixSet.add(THAI_MAIYAMOK);

//     // Compact for caching.
//     fMarkSet.compact();
//     fEndWordSet.compact();
//     fBeginWordSet.compact();
//     fSuffixSet.compact();
// }

// ThaiBreakEngine::~ThaiBreakEngine() {
//     delete fDictionary;
// }

// int32_t
// ThaiBreakEngine::divideUpDictionaryRange( UText *text,
//                                                 int32_t rangeStart,
//                                                 int32_t rangeEnd,
//                                                 UVector32 &foundBreaks ) const {
//     utext_setNativeIndex(text, rangeStart);
//     utext_moveIndex32(text, THAI_MIN_WORD_SPAN);
//     if (utext_getNativeIndex(text) >= rangeEnd) {
//         return 0;       // Not enough characters for two words
//     }
//     utext_setNativeIndex(text, rangeStart);


//     uint32_t wordsFound = 0;
//     int32_t cpWordLength = 0;    // Word Length in Code Points.
//     int32_t cuWordLength = 0;    // Word length in code units (UText native indexing)
//     int32_t current;
//     UErrorCode status = U_ZERO_ERROR;
//     PossibleWord words[THAI_LOOKAHEAD];

//     utext_setNativeIndex(text, rangeStart);

//     while (U_SUCCESS(status) && (current = (int32_t)utext_getNativeIndex(text)) < rangeEnd) {
//         cpWordLength = 0;
//         cuWordLength = 0;

//         // Look for candidate words at the current position
//         int32_t candidates = words[wordsFound%THAI_LOOKAHEAD].candidates(text, fDictionary, rangeEnd);

//         // If we found exactly one, use that
//         if (candidates == 1) {
//             cuWordLength = words[wordsFound % THAI_LOOKAHEAD].acceptMarked(text);
//             cpWordLength = words[wordsFound % THAI_LOOKAHEAD].markedCPLength();
//             wordsFound += 1;
//         }
//         // If there was more than one, see which one can take us forward the most words
//         else if (candidates > 1) {
//             // If we're already at the end of the range, we're done
//             if ((int32_t)utext_getNativeIndex(text) >= rangeEnd) {
//                 goto foundBest;
//             }
//             do {
//                 int32_t wordsMatched = 1;
//                 if (words[(wordsFound + 1) % THAI_LOOKAHEAD].candidates(text, fDictionary, rangeEnd) > 0) {
//                     if (wordsMatched < 2) {
//                         // Followed by another dictionary word; mark first word as a good candidate
//                         words[wordsFound%THAI_LOOKAHEAD].markCurrent();
//                         wordsMatched = 2;
//                     }

//                     // If we're already at the end of the range, we're done
//                     if ((int32_t)utext_getNativeIndex(text) >= rangeEnd) {
//                         goto foundBest;
//                     }

//                     // See if any of the possible second words is followed by a third word
//                     do {
//                         // If we find a third word, stop right away
//                         if (words[(wordsFound + 2) % THAI_LOOKAHEAD].candidates(text, fDictionary, rangeEnd)) {
//                             words[wordsFound % THAI_LOOKAHEAD].markCurrent();
//                             goto foundBest;
//                         }
//                     }
//                     while (words[(wordsFound + 1) % THAI_LOOKAHEAD].backUp(text));
//                 }
//             }
//             while (words[wordsFound % THAI_LOOKAHEAD].backUp(text));
// foundBest:
//             // Set UText position to after the accepted word.
//             cuWordLength = words[wordsFound % THAI_LOOKAHEAD].acceptMarked(text);
//             cpWordLength = words[wordsFound % THAI_LOOKAHEAD].markedCPLength();
//             wordsFound += 1;
//         }

//         // We come here after having either found a word or not. We look ahead to the
//         // next word. If it's not a dictionary word, we will combine it with the word we
//         // just found (if there is one), but only if the preceding word does not exceed
//         // the threshold.
//         // The text iterator should now be positioned at the end of the word we found.

//         UChar32 uc = 0;
//         if ((int32_t)utext_getNativeIndex(text) < rangeEnd &&  cpWordLength < THAI_ROOT_COMBINE_THRESHOLD) {
//             // if it is a dictionary word, do nothing. If it isn't, then if there is
//             // no preceding word, or the non-word shares less than the minimum threshold
//             // of characters with a dictionary word, then scan to resynchronize
//             if (words[wordsFound % THAI_LOOKAHEAD].candidates(text, fDictionary, rangeEnd) <= 0
//                   && (cuWordLength == 0
//                       || words[wordsFound%THAI_LOOKAHEAD].longestPrefix() < THAI_PREFIX_COMBINE_THRESHOLD)) {
//                 // Look for a plausible word boundary
//                 int32_t remaining = rangeEnd - (current+cuWordLength);
//                 UChar32 pc;
//                 int32_t chars = 0;
//                 for (;;) {
//                     int32_t pcIndex = (int32_t)utext_getNativeIndex(text);
//                     pc = utext_next32(text);
//                     int32_t pcSize = (int32_t)utext_getNativeIndex(text) - pcIndex;
//                     chars += pcSize;
//                     remaining -= pcSize;
//                     if (remaining <= 0) {
//                         break;
//                     }
//                     uc = utext_current32(text);
//                     if (fEndWordSet.contains(pc) && fBeginWordSet.contains(uc)) {
//                         // Maybe. See if it's in the dictionary.
//                         // NOTE: In the original Apple code, checked that the next
//                         // two characters after uc were not 0x0E4C THANTHAKHAT before
//                         // checking the dictionary. That is just a performance filter,
//                         // but it's not clear it's faster than checking the trie.
//                         int32_t num_candidates = words[(wordsFound + 1) % THAI_LOOKAHEAD].candidates(text, fDictionary, rangeEnd);
//                         utext_setNativeIndex(text, current + cuWordLength + chars);
//                         if (num_candidates > 0) {
//                             break;
//                         }
//                     }
//                 }

//                 // Bump the word count if there wasn't already one
//                 if (cuWordLength <= 0) {
//                     wordsFound += 1;
//                 }

//                 // Update the length with the passed-over characters
//                 cuWordLength += chars;
//             }
//             else {
//                 // Back up to where we were for next iteration
//                 utext_setNativeIndex(text, current+cuWordLength);
//             }
//         }

//         // Never stop before a combining mark.
//         int32_t currPos;
//         while ((currPos = (int32_t)utext_getNativeIndex(text)) < rangeEnd && fMarkSet.contains(utext_current32(text))) {
//             utext_next32(text);
//             cuWordLength += (int32_t)utext_getNativeIndex(text) - currPos;
//         }

//         // Look ahead for possible suffixes if a dictionary word does not follow.
//         // We do this in code rather than using a rule so that the heuristic
//         // resynch continues to function. For example, one of the suffix characters
//         // could be a typo in the middle of a word.
//         if ((int32_t)utext_getNativeIndex(text) < rangeEnd && cuWordLength > 0) {
//             if (words[wordsFound%THAI_LOOKAHEAD].candidates(text, fDictionary, rangeEnd) <= 0
//                 && fSuffixSet.contains(uc = utext_current32(text))) {
//                 if (uc == THAI_PAIYANNOI) {
//                     if (!fSuffixSet.contains(utext_previous32(text))) {
//                         // Skip over previous end and PAIYANNOI
//                         utext_next32(text);
//                         int32_t paiyannoiIndex = (int32_t)utext_getNativeIndex(text);
//                         utext_next32(text);
//                         cuWordLength += (int32_t)utext_getNativeIndex(text) - paiyannoiIndex;    // Add PAIYANNOI to word
//                         uc = utext_current32(text);     // Fetch next character
//                     }
//                     else {
//                         // Restore prior position
//                         utext_next32(text);
//                     }
//                 }
//                 if (uc == THAI_MAIYAMOK) {
//                     if (utext_previous32(text) != THAI_MAIYAMOK) {
//                         // Skip over previous end and MAIYAMOK
//                         utext_next32(text);
//                         int32_t maiyamokIndex = (int32_t)utext_getNativeIndex(text);
//                         utext_next32(text);
//                         cuWordLength += (int32_t)utext_getNativeIndex(text) - maiyamokIndex;    // Add MAIYAMOK to word
//                     }
//                     else {
//                         // Restore prior position
//                         utext_next32(text);
//                     }
//                 }
//             }
//             else {
//                 utext_setNativeIndex(text, current+cuWordLength);
//             }
//         }

//         // Did we find a word on this iteration? If so, push it on the break stack
//         if (cuWordLength > 0) {
//             foundBreaks.push((current+cuWordLength), status);
//         }
//     }

//     // Don't return a break for the end of the dictionary range if there is one there.
//     if (foundBreaks.peeki() >= rangeEnd) {
//         (void) foundBreaks.popi();
//         wordsFound -= 1;
//     }

//     return wordsFound;
// }

// /*
//  ******************************************************************
//  * LaoBreakEngine
//  */

// // How many words in a row are "good enough"?
// static const int32_t LAO_LOOKAHEAD = 3;

// // Will not combine a non-word with a preceding dictionary word longer than this
// static const int32_t LAO_ROOT_COMBINE_THRESHOLD = 3;

// // Will not combine a non-word that shares at least this much prefix with a
// // dictionary word, with a preceding word
// static const int32_t LAO_PREFIX_COMBINE_THRESHOLD = 3;

// // Minimum word size
// static const int32_t LAO_MIN_WORD = 2;

// // Minimum number of characters for two words
// static const int32_t LAO_MIN_WORD_SPAN = LAO_MIN_WORD * 2;

// LaoBreakEngine::LaoBreakEngine(DictionaryMatcher *adoptDictionary, UErrorCode &status)
//     : DictionaryBreakEngine(),
//       fDictionary(adoptDictionary)
// {
//     fLaoWordSet.applyPattern(UNICODE_STRING_SIMPLE("[[:Laoo:]&[:LineBreak=SA:]]"), status);
//     if (U_SUCCESS(status)) {
//         setCharacters(fLaoWordSet);
//     }
//     fMarkSet.applyPattern(UNICODE_STRING_SIMPLE("[[:Laoo:]&[:LineBreak=SA:]&[:M:]]"), status);
//     fMarkSet.add(0x0020);
//     fEndWordSet = fLaoWordSet;
//     fEndWordSet.remove(0x0EC0, 0x0EC4);     // prefix vowels
//     fBeginWordSet.add(0x0E81, 0x0EAE);      // basic consonants (including holes for corresponding Thai characters)
//     fBeginWordSet.add(0x0EDC, 0x0EDD);      // digraph consonants (no Thai equivalent)
//     fBeginWordSet.add(0x0EC0, 0x0EC4);      // prefix vowels

//     // Compact for caching.
//     fMarkSet.compact();
//     fEndWordSet.compact();
//     fBeginWordSet.compact();
// }

// LaoBreakEngine::~LaoBreakEngine() {
//     delete fDictionary;
// }

// int32_t
// LaoBreakEngine::divideUpDictionaryRange( UText *text,
//                                                 int32_t rangeStart,
//                                                 int32_t rangeEnd,
//                                                 UVector32 &foundBreaks ) const {
//     if ((rangeEnd - rangeStart) < LAO_MIN_WORD_SPAN) {
//         return 0;       // Not enough characters for two words
//     }

//     uint32_t wordsFound = 0;
//     int32_t cpWordLength = 0;
//     int32_t cuWordLength = 0;
//     int32_t current;
//     UErrorCode status = U_ZERO_ERROR;
//     PossibleWord words[LAO_LOOKAHEAD];

//     utext_setNativeIndex(text, rangeStart);

//     while (U_SUCCESS(status) && (current = (int32_t)utext_getNativeIndex(text)) < rangeEnd) {
//         cuWordLength = 0;
//         cpWordLength = 0;

//         // Look for candidate words at the current position
//         int32_t candidates = words[wordsFound%LAO_LOOKAHEAD].candidates(text, fDictionary, rangeEnd);

//         // If we found exactly one, use that
//         if (candidates == 1) {
//             cuWordLength = words[wordsFound % LAO_LOOKAHEAD].acceptMarked(text);
//             cpWordLength = words[wordsFound % LAO_LOOKAHEAD].markedCPLength();
//             wordsFound += 1;
//         }
//         // If there was more than one, see which one can take us forward the most words
//         else if (candidates > 1) {
//             // If we're already at the end of the range, we're done
//             if (utext_getNativeIndex(text) >= rangeEnd) {
//                 goto foundBest;
//             }
//             do {
//                 int32_t wordsMatched = 1;
//                 if (words[(wordsFound + 1) % LAO_LOOKAHEAD].candidates(text, fDictionary, rangeEnd) > 0) {
//                     if (wordsMatched < 2) {
//                         // Followed by another dictionary word; mark first word as a good candidate
//                         words[wordsFound%LAO_LOOKAHEAD].markCurrent();
//                         wordsMatched = 2;
//                     }

//                     // If we're already at the end of the range, we're done
//                     if ((int32_t)utext_getNativeIndex(text) >= rangeEnd) {
//                         goto foundBest;
//                     }

//                     // See if any of the possible second words is followed by a third word
//                     do {
//                         // If we find a third word, stop right away
//                         if (words[(wordsFound + 2) % LAO_LOOKAHEAD].candidates(text, fDictionary, rangeEnd)) {
//                             words[wordsFound % LAO_LOOKAHEAD].markCurrent();
//                             goto foundBest;
//                         }
//                     }
//                     while (words[(wordsFound + 1) % LAO_LOOKAHEAD].backUp(text));
//                 }
//             }
//             while (words[wordsFound % LAO_LOOKAHEAD].backUp(text));
// foundBest:
//             cuWordLength = words[wordsFound % LAO_LOOKAHEAD].acceptMarked(text);
//             cpWordLength = words[wordsFound % LAO_LOOKAHEAD].markedCPLength();
//             wordsFound += 1;
//         }

//         // We come here after having either found a word or not. We look ahead to the
//         // next word. If it's not a dictionary word, we will combine it withe the word we
//         // just found (if there is one), but only if the preceding word does not exceed
//         // the threshold.
//         // The text iterator should now be positioned at the end of the word we found.
//         if ((int32_t)utext_getNativeIndex(text) < rangeEnd && cpWordLength < LAO_ROOT_COMBINE_THRESHOLD) {
//             // if it is a dictionary word, do nothing. If it isn't, then if there is
//             // no preceding word, or the non-word shares less than the minimum threshold
//             // of characters with a dictionary word, then scan to resynchronize
//             if (words[wordsFound % LAO_LOOKAHEAD].candidates(text, fDictionary, rangeEnd) <= 0
//                   && (cuWordLength == 0
//                       || words[wordsFound%LAO_LOOKAHEAD].longestPrefix() < LAO_PREFIX_COMBINE_THRESHOLD)) {
//                 // Look for a plausible word boundary
//                 int32_t remaining = rangeEnd - (current + cuWordLength);
//                 UChar32 pc;
//                 UChar32 uc;
//                 int32_t chars = 0;
//                 for (;;) {
//                     int32_t pcIndex = (int32_t)utext_getNativeIndex(text);
//                     pc = utext_next32(text);
//                     int32_t pcSize = (int32_t)utext_getNativeIndex(text) - pcIndex;
//                     chars += pcSize;
//                     remaining -= pcSize;
//                     if (remaining <= 0) {
//                         break;
//                     }
//                     uc = utext_current32(text);
//                     if (fEndWordSet.contains(pc) && fBeginWordSet.contains(uc)) {
//                         // Maybe. See if it's in the dictionary.
//                         // TODO: this looks iffy; compare with old code.
//                         int32_t num_candidates = words[(wordsFound + 1) % LAO_LOOKAHEAD].candidates(text, fDictionary, rangeEnd);
//                         utext_setNativeIndex(text, current + cuWordLength + chars);
//                         if (num_candidates > 0) {
//                             break;
//                         }
//                     }
//                 }

//                 // Bump the word count if there wasn't already one
//                 if (cuWordLength <= 0) {
//                     wordsFound += 1;
//                 }

//                 // Update the length with the passed-over characters
//                 cuWordLength += chars;
//             }
//             else {
//                 // Back up to where we were for next iteration
//                 utext_setNativeIndex(text, current + cuWordLength);
//             }
//         }

//         // Never stop before a combining mark.
//         int32_t currPos;
//         while ((currPos = (int32_t)utext_getNativeIndex(text)) < rangeEnd && fMarkSet.contains(utext_current32(text))) {
//             utext_next32(text);
//             cuWordLength += (int32_t)utext_getNativeIndex(text) - currPos;
//         }

//         // Look ahead for possible suffixes if a dictionary word does not follow.
//         // We do this in code rather than using a rule so that the heuristic
//         // resynch continues to function. For example, one of the suffix characters
//         // could be a typo in the middle of a word.
//         // NOT CURRENTLY APPLICABLE TO LAO

//         // Did we find a word on this iteration? If so, push it on the break stack
//         if (cuWordLength > 0) {
//             foundBreaks.push((current+cuWordLength), status);
//         }
//     }

//     // Don't return a break for the end of the dictionary range if there is one there.
//     if (foundBreaks.peeki() >= rangeEnd) {
//         (void) foundBreaks.popi();
//         wordsFound -= 1;
//     }

//     return wordsFound;
// }

// /*
//  ******************************************************************
//  * BurmeseBreakEngine
//  */

// // How many words in a row are "good enough"?
// static const int32_t BURMESE_LOOKAHEAD = 3;

// // Will not combine a non-word with a preceding dictionary word longer than this
// static const int32_t BURMESE_ROOT_COMBINE_THRESHOLD = 3;

// // Will not combine a non-word that shares at least this much prefix with a
// // dictionary word, with a preceding word
// static const int32_t BURMESE_PREFIX_COMBINE_THRESHOLD = 3;

// // Minimum word size
// static const int32_t BURMESE_MIN_WORD = 2;

// // Minimum number of characters for two words
// static const int32_t BURMESE_MIN_WORD_SPAN = BURMESE_MIN_WORD * 2;

// BurmeseBreakEngine::BurmeseBreakEngine(DictionaryMatcher *adoptDictionary, UErrorCode &status)
//     : DictionaryBreakEngine(),
//       fDictionary(adoptDictionary)
// {
//     fBurmeseWordSet.applyPattern(UNICODE_STRING_SIMPLE("[[:Mymr:]&[:LineBreak=SA:]]"), status);
//     if (U_SUCCESS(status)) {
//         setCharacters(fBurmeseWordSet);
//     }
//     fMarkSet.applyPattern(UNICODE_STRING_SIMPLE("[[:Mymr:]&[:LineBreak=SA:]&[:M:]]"), status);
//     fMarkSet.add(0x0020);
//     fEndWordSet = fBurmeseWordSet;
//     fBeginWordSet.add(0x1000, 0x102A);      // basic consonants and independent vowels

//     // Compact for caching.
//     fMarkSet.compact();
//     fEndWordSet.compact();
//     fBeginWordSet.compact();
// }

// BurmeseBreakEngine::~BurmeseBreakEngine() {
//     delete fDictionary;
// }

// int32_t
// BurmeseBreakEngine::divideUpDictionaryRange( UText *text,
//                                                 int32_t rangeStart,
//                                                 int32_t rangeEnd,
//                                                 UVector32 &foundBreaks ) const {
//     if ((rangeEnd - rangeStart) < BURMESE_MIN_WORD_SPAN) {
//         return 0;       // Not enough characters for two words
//     }

//     uint32_t wordsFound = 0;
//     int32_t cpWordLength = 0;
//     int32_t cuWordLength = 0;
//     int32_t current;
//     UErrorCode status = U_ZERO_ERROR;
//     PossibleWord words[BURMESE_LOOKAHEAD];

//     utext_setNativeIndex(text, rangeStart);

//     while (U_SUCCESS(status) && (current = (int32_t)utext_getNativeIndex(text)) < rangeEnd) {
//         cuWordLength = 0;
//         cpWordLength = 0;

//         // Look for candidate words at the current position
//         int32_t candidates = words[wordsFound%BURMESE_LOOKAHEAD].candidates(text, fDictionary, rangeEnd);

//         // If we found exactly one, use that
//         if (candidates == 1) {
//             cuWordLength = words[wordsFound % BURMESE_LOOKAHEAD].acceptMarked(text);
//             cpWordLength = words[wordsFound % BURMESE_LOOKAHEAD].markedCPLength();
//             wordsFound += 1;
//         }
//         // If there was more than one, see which one can take us forward the most words
//         else if (candidates > 1) {
//             // If we're already at the end of the range, we're done
//             if (utext_getNativeIndex(text) >= rangeEnd) {
//                 goto foundBest;
//             }
//             do {
//                 int32_t wordsMatched = 1;
//                 if (words[(wordsFound + 1) % BURMESE_LOOKAHEAD].candidates(text, fDictionary, rangeEnd) > 0) {
//                     if (wordsMatched < 2) {
//                         // Followed by another dictionary word; mark first word as a good candidate
//                         words[wordsFound%BURMESE_LOOKAHEAD].markCurrent();
//                         wordsMatched = 2;
//                     }

//                     // If we're already at the end of the range, we're done
//                     if ((int32_t)utext_getNativeIndex(text) >= rangeEnd) {
//                         goto foundBest;
//                     }

//                     // See if any of the possible second words is followed by a third word
//                     do {
//                         // If we find a third word, stop right away
//                         if (words[(wordsFound + 2) % BURMESE_LOOKAHEAD].candidates(text, fDictionary, rangeEnd)) {
//                             words[wordsFound % BURMESE_LOOKAHEAD].markCurrent();
//                             goto foundBest;
//                         }
//                     }
//                     while (words[(wordsFound + 1) % BURMESE_LOOKAHEAD].backUp(text));
//                 }
//             }
//             while (words[wordsFound % BURMESE_LOOKAHEAD].backUp(text));
// foundBest:
//             cuWordLength = words[wordsFound % BURMESE_LOOKAHEAD].acceptMarked(text);
//             cpWordLength = words[wordsFound % BURMESE_LOOKAHEAD].markedCPLength();
//             wordsFound += 1;
//         }

//         // We come here after having either found a word or not. We look ahead to the
//         // next word. If it's not a dictionary word, we will combine it withe the word we
//         // just found (if there is one), but only if the preceding word does not exceed
//         // the threshold.
//         // The text iterator should now be positioned at the end of the word we found.
//         if ((int32_t)utext_getNativeIndex(text) < rangeEnd && cpWordLength < BURMESE_ROOT_COMBINE_THRESHOLD) {
//             // if it is a dictionary word, do nothing. If it isn't, then if there is
//             // no preceding word, or the non-word shares less than the minimum threshold
//             // of characters with a dictionary word, then scan to resynchronize
//             if (words[wordsFound % BURMESE_LOOKAHEAD].candidates(text, fDictionary, rangeEnd) <= 0
//                   && (cuWordLength == 0
//                       || words[wordsFound%BURMESE_LOOKAHEAD].longestPrefix() < BURMESE_PREFIX_COMBINE_THRESHOLD)) {
//                 // Look for a plausible word boundary
//                 int32_t remaining = rangeEnd - (current + cuWordLength);
//                 UChar32 pc;
//                 UChar32 uc;
//                 int32_t chars = 0;
//                 for (;;) {
//                     int32_t pcIndex = (int32_t)utext_getNativeIndex(text);
//                     pc = utext_next32(text);
//                     int32_t pcSize = (int32_t)utext_getNativeIndex(text) - pcIndex;
//                     chars += pcSize;
//                     remaining -= pcSize;
//                     if (remaining <= 0) {
//                         break;
//                     }
//                     uc = utext_current32(text);
//                     if (fEndWordSet.contains(pc) && fBeginWordSet.contains(uc)) {
//                         // Maybe. See if it's in the dictionary.
//                         // TODO: this looks iffy; compare with old code.
//                         int32_t num_candidates = words[(wordsFound + 1) % BURMESE_LOOKAHEAD].candidates(text, fDictionary, rangeEnd);
//                         utext_setNativeIndex(text, current + cuWordLength + chars);
//                         if (num_candidates > 0) {
//                             break;
//                         }
//                     }
//                 }

//                 // Bump the word count if there wasn't already one
//                 if (cuWordLength <= 0) {
//                     wordsFound += 1;
//                 }

//                 // Update the length with the passed-over characters
//                 cuWordLength += chars;
//             }
//             else {
//                 // Back up to where we were for next iteration
//                 utext_setNativeIndex(text, current + cuWordLength);
//             }
//         }

//         // Never stop before a combining mark.
//         int32_t currPos;
//         while ((currPos = (int32_t)utext_getNativeIndex(text)) < rangeEnd && fMarkSet.contains(utext_current32(text))) {
//             utext_next32(text);
//             cuWordLength += (int32_t)utext_getNativeIndex(text) - currPos;
//         }

//         // Look ahead for possible suffixes if a dictionary word does not follow.
//         // We do this in code rather than using a rule so that the heuristic
//         // resynch continues to function. For example, one of the suffix characters
//         // could be a typo in the middle of a word.
//         // NOT CURRENTLY APPLICABLE TO BURMESE

//         // Did we find a word on this iteration? If so, push it on the break stack
//         if (cuWordLength > 0) {
//             foundBreaks.push((current+cuWordLength), status);
//         }
//     }

//     // Don't return a break for the end of the dictionary range if there is one there.
//     if (foundBreaks.peeki() >= rangeEnd) {
//         (void) foundBreaks.popi();
//         wordsFound -= 1;
//     }

//     return wordsFound;
// }

// /*
//  ******************************************************************
//  * KhmerBreakEngine
//  */

// // How many words in a row are "good enough"?
// static const int32_t KHMER_LOOKAHEAD = 3;

// // Will not combine a non-word with a preceding dictionary word longer than this
// static const int32_t KHMER_ROOT_COMBINE_THRESHOLD = 3;

// // Will not combine a non-word that shares at least this much prefix with a
// // dictionary word, with a preceding word
// static const int32_t KHMER_PREFIX_COMBINE_THRESHOLD = 3;

// // Minimum word size
// static const int32_t KHMER_MIN_WORD = 2;

// // Minimum number of characters for two words
// static const int32_t KHMER_MIN_WORD_SPAN = KHMER_MIN_WORD * 2;

// KhmerBreakEngine::KhmerBreakEngine(DictionaryMatcher *adoptDictionary, UErrorCode &status)
//     : DictionaryBreakEngine(),
//       fDictionary(adoptDictionary)
// {
//     fKhmerWordSet.applyPattern(UNICODE_STRING_SIMPLE("[[:Khmr:]&[:LineBreak=SA:]]"), status);
//     if (U_SUCCESS(status)) {
//         setCharacters(fKhmerWordSet);
//     }
//     fMarkSet.applyPattern(UNICODE_STRING_SIMPLE("[[:Khmr:]&[:LineBreak=SA:]&[:M:]]"), status);
//     fMarkSet.add(0x0020);
//     fEndWordSet = fKhmerWordSet;
//     fBeginWordSet.add(0x1780, 0x17B3);
//     //fBeginWordSet.add(0x17A3, 0x17A4);      // deprecated vowels
//     //fEndWordSet.remove(0x17A5, 0x17A9);     // Khmer independent vowels that can't end a word
//     //fEndWordSet.remove(0x17B2);             // Khmer independent vowel that can't end a word
//     fEndWordSet.remove(0x17D2);             // KHMER SIGN COENG that combines some following characters
//     //fEndWordSet.remove(0x17B6, 0x17C5);     // Remove dependent vowels
// //    fEndWordSet.remove(0x0E31);             // MAI HAN-AKAT
// //    fEndWordSet.remove(0x0E40, 0x0E44);     // SARA E through SARA AI MAIMALAI
// //    fBeginWordSet.add(0x0E01, 0x0E2E);      // KO KAI through HO NOKHUK
// //    fBeginWordSet.add(0x0E40, 0x0E44);      // SARA E through SARA AI MAIMALAI
// //    fSuffixSet.add(THAI_PAIYANNOI);
// //    fSuffixSet.add(THAI_MAIYAMOK);

//     // Compact for caching.
//     fMarkSet.compact();
//     fEndWordSet.compact();
//     fBeginWordSet.compact();
// //    fSuffixSet.compact();
// }

// KhmerBreakEngine::~KhmerBreakEngine() {
//     delete fDictionary;
// }

// int32_t
// KhmerBreakEngine::divideUpDictionaryRange( UText *text,
//                                                 int32_t rangeStart,
//                                                 int32_t rangeEnd,
//                                                 UVector32 &foundBreaks ) const {
//     if ((rangeEnd - rangeStart) < KHMER_MIN_WORD_SPAN) {
//         return 0;       // Not enough characters for two words
//     }

//     uint32_t wordsFound = 0;
//     int32_t cpWordLength = 0;
//     int32_t cuWordLength = 0;
//     int32_t current;
//     UErrorCode status = U_ZERO_ERROR;
//     PossibleWord words[KHMER_LOOKAHEAD];

//     utext_setNativeIndex(text, rangeStart);

//     while (U_SUCCESS(status) && (current = (int32_t)utext_getNativeIndex(text)) < rangeEnd) {
//         cuWordLength = 0;
//         cpWordLength = 0;

//         // Look for candidate words at the current position
//         int32_t candidates = words[wordsFound%KHMER_LOOKAHEAD].candidates(text, fDictionary, rangeEnd);

//         // If we found exactly one, use that
//         if (candidates == 1) {
//             cuWordLength = words[wordsFound % KHMER_LOOKAHEAD].acceptMarked(text);
//             cpWordLength = words[wordsFound % KHMER_LOOKAHEAD].markedCPLength();
//             wordsFound += 1;
//         }

//         // If there was more than one, see which one can take us forward the most words
//         else if (candidates > 1) {
//             // If we're already at the end of the range, we're done
//             if ((int32_t)utext_getNativeIndex(text) >= rangeEnd) {
//                 goto foundBest;
//             }
//             do {
//                 int32_t wordsMatched = 1;
//                 if (words[(wordsFound + 1) % KHMER_LOOKAHEAD].candidates(text, fDictionary, rangeEnd) > 0) {
//                     if (wordsMatched < 2) {
//                         // Followed by another dictionary word; mark first word as a good candidate
//                         words[wordsFound % KHMER_LOOKAHEAD].markCurrent();
//                         wordsMatched = 2;
//                     }

//                     // If we're already at the end of the range, we're done
//                     if ((int32_t)utext_getNativeIndex(text) >= rangeEnd) {
//                         goto foundBest;
//                     }

//                     // See if any of the possible second words is followed by a third word
//                     do {
//                         // If we find a third word, stop right away
//                         if (words[(wordsFound + 2) % KHMER_LOOKAHEAD].candidates(text, fDictionary, rangeEnd)) {
//                             words[wordsFound % KHMER_LOOKAHEAD].markCurrent();
//                             goto foundBest;
//                         }
//                     }
//                     while (words[(wordsFound + 1) % KHMER_LOOKAHEAD].backUp(text));
//                 }
//             }
//             while (words[wordsFound % KHMER_LOOKAHEAD].backUp(text));
// foundBest:
//             cuWordLength = words[wordsFound % KHMER_LOOKAHEAD].acceptMarked(text);
//             cpWordLength = words[wordsFound % KHMER_LOOKAHEAD].markedCPLength();
//             wordsFound += 1;
//         }

//         // We come here after having either found a word or not. We look ahead to the
//         // next word. If it's not a dictionary word, we will combine it with the word we
//         // just found (if there is one), but only if the preceding word does not exceed
//         // the threshold.
//         // The text iterator should now be positioned at the end of the word we found.
//         if ((int32_t)utext_getNativeIndex(text) < rangeEnd && cpWordLength < KHMER_ROOT_COMBINE_THRESHOLD) {
//             // if it is a dictionary word, do nothing. If it isn't, then if there is
//             // no preceding word, or the non-word shares less than the minimum threshold
//             // of characters with a dictionary word, then scan to resynchronize
//             if (words[wordsFound % KHMER_LOOKAHEAD].candidates(text, fDictionary, rangeEnd) <= 0
//                   && (cuWordLength == 0
//                       || words[wordsFound % KHMER_LOOKAHEAD].longestPrefix() < KHMER_PREFIX_COMBINE_THRESHOLD)) {
//                 // Look for a plausible word boundary
//                 int32_t remaining = rangeEnd - (current+cuWordLength);
//                 UChar32 pc;
//                 UChar32 uc;
//                 int32_t chars = 0;
//                 for (;;) {
//                     int32_t pcIndex = (int32_t)utext_getNativeIndex(text);
//                     pc = utext_next32(text);
//                     int32_t pcSize = (int32_t)utext_getNativeIndex(text) - pcIndex;
//                     chars += pcSize;
//                     remaining -= pcSize;
//                     if (remaining <= 0) {
//                         break;
//                     }
//                     uc = utext_current32(text);
//                     if (fEndWordSet.contains(pc) && fBeginWordSet.contains(uc)) {
//                         // Maybe. See if it's in the dictionary.
//                         int32_t num_candidates = words[(wordsFound + 1) % KHMER_LOOKAHEAD].candidates(text, fDictionary, rangeEnd);
//                         utext_setNativeIndex(text, current+cuWordLength+chars);
//                         if (num_candidates > 0) {
//                             break;
//                         }
//                     }
//                 }

//                 // Bump the word count if there wasn't already one
//                 if (cuWordLength <= 0) {
//                     wordsFound += 1;
//                 }

//                 // Update the length with the passed-over characters
//                 cuWordLength += chars;
//             }
//             else {
//                 // Back up to where we were for next iteration
//                 utext_setNativeIndex(text, current+cuWordLength);
//             }
//         }

//         // Never stop before a combining mark.
//         int32_t currPos;
//         while ((currPos = (int32_t)utext_getNativeIndex(text)) < rangeEnd && fMarkSet.contains(utext_current32(text))) {
//             utext_next32(text);
//             cuWordLength += (int32_t)utext_getNativeIndex(text) - currPos;
//         }

//         // Look ahead for possible suffixes if a dictionary word does not follow.
//         // We do this in code rather than using a rule so that the heuristic
//         // resynch continues to function. For example, one of the suffix characters
//         // could be a typo in the middle of a word.
// //        if ((int32_t)utext_getNativeIndex(text) < rangeEnd && wordLength > 0) {
// //            if (words[wordsFound%KHMER_LOOKAHEAD].candidates(text, fDictionary, rangeEnd) <= 0
// //                && fSuffixSet.contains(uc = utext_current32(text))) {
// //                if (uc == KHMER_PAIYANNOI) {
// //                    if (!fSuffixSet.contains(utext_previous32(text))) {
// //                        // Skip over previous end and PAIYANNOI
// //                        utext_next32(text);
// //                        utext_next32(text);
// //                        wordLength += 1;            // Add PAIYANNOI to word
// //                        uc = utext_current32(text);     // Fetch next character
// //                    }
// //                    else {
// //                        // Restore prior position
// //                        utext_next32(text);
// //                    }
// //                }
// //                if (uc == KHMER_MAIYAMOK) {
// //                    if (utext_previous32(text) != KHMER_MAIYAMOK) {
// //                        // Skip over previous end and MAIYAMOK
// //                        utext_next32(text);
// //                        utext_next32(text);
// //                        wordLength += 1;            // Add MAIYAMOK to word
// //                    }
// //                    else {
// //                        // Restore prior position
// //                        utext_next32(text);
// //                    }
// //                }
// //            }
// //            else {
// //                utext_setNativeIndex(text, current+wordLength);
// //            }
// //        }

//         // Did we find a word on this iteration? If so, push it on the break stack
//         if (cuWordLength > 0) {
//             foundBreaks.push((current+cuWordLength), status);
//         }
//     }

//     // Don't return a break for the end of the dictionary range if there is one there.
//     if (foundBreaks.peeki() >= rangeEnd) {
//         (void) foundBreaks.popi();
//         wordsFound -= 1;
//     }

//     return wordsFound;
// }

// #if !UCONFIG_NO_NORMALIZATION
// /*
//  ******************************************************************
//  * CjkBreakEngine
//  */
// static const uint32_t kuint32max = 0xFFFFFFFF;
// CjkBreakEngine::CjkBreakEngine(DictionaryMatcher *adoptDictionary, LanguageType type, UErrorCode &status)
// : DictionaryBreakEngine(), fDictionary(adoptDictionary) {
//     // Korean dictionary only includes Hangul syllables
//     fHangulWordSet.applyPattern(UNICODE_STRING_SIMPLE("[\\uac00-\\ud7a3]"), status);
//     fHanWordSet.applyPattern(UNICODE_STRING_SIMPLE("[:Han:]"), status);
//     fKatakanaWordSet.applyPattern(UNICODE_STRING_SIMPLE("[[:Katakana:]\\uff9e\\uff9f]"), status);
//     fHiraganaWordSet.applyPattern(UNICODE_STRING_SIMPLE("[:Hiragana:]"), status);
//     nfkcNorm2 = Normalizer2::getNFKCInstance(status);

//     if (U_SUCCESS(status)) {
//         // handle Korean and Japanese/Chinese using different dictionaries
//         if (type == kKorean) {
//             setCharacters(fHangulWordSet);
//         } else { //Chinese and Japanese
//             UnicodeSet cjSet;
//             cjSet.addAll(fHanWordSet);
//             cjSet.addAll(fKatakanaWordSet);
//             cjSet.addAll(fHiraganaWordSet);
//             cjSet.add(0xFF70); // HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
//             cjSet.add(0x30FC); // KATAKANA-HIRAGANA PROLONGED SOUND MARK
//             setCharacters(cjSet);
//         }
//     }
// }

// CjkBreakEngine::~CjkBreakEngine(){
//     delete fDictionary;
// }

// // The katakanaCost values below are based on the length frequencies of all
// // katakana phrases in the dictionary
// static const int32_t kMaxKatakanaLength = 8;
// static const int32_t kMaxKatakanaGroupLength = 20;
// static const uint32_t maxSnlp = 255;

// static inline uint32_t getKatakanaCost(int32_t wordLength){
//     //TODO: fill array with actual values from dictionary!
//     static const uint32_t katakanaCost[kMaxKatakanaLength + 1]
//                                        = {8192, 984, 408, 240, 204, 252, 300, 372, 480};
//     return (wordLength > kMaxKatakanaLength) ? 8192 : katakanaCost[wordLength];
// }

// static inline bool isKatakana(UChar32 value) {
//     return (value >= 0x30A1 && value <= 0x30FE && value != 0x30FB) ||
//             (value >= 0xFF66 && value <= 0xFF9f);
// }


// // Function for accessing internal utext flags.
// //   Replicates an internal UText function.

// static inline int32_t utext_i32_flag(int32_t bitIndex) {
//     return (int32_t)1 << bitIndex;
// }


// /*
//  * @param text A UText representing the text
//  * @param rangeStart The start of the range of dictionary characters
//  * @param rangeEnd The end of the range of dictionary characters
//  * @param foundBreaks vector<int32> to receive the break positions
//  * @return The number of breaks found
//  */
// int32_t
// CjkBreakEngine::divideUpDictionaryRange( UText *inText,
//         int32_t rangeStart,
//         int32_t rangeEnd,
//         UVector32 &foundBreaks ) const {
//     if (rangeStart >= rangeEnd) {
//         return 0;
//     }

//     // UnicodeString version of input UText, NFKC normalized if necessary.
//     UnicodeString inString;

//     // inputMap[inStringIndex] = corresponding native index from UText inText.
//     // If NULL then mapping is 1:1
//     LocalPointer<UVector32>     inputMap;

//     UErrorCode     status      = U_ZERO_ERROR;


//     // if UText has the input string as one contiguous UTF-16 chunk
//     if ((inText->providerProperties & utext_i32_flag(UTEXT_PROVIDER_STABLE_CHUNKS)) &&
//          inText->chunkNativeStart <= rangeStart &&
//          inText->chunkNativeLimit >= rangeEnd   &&
//          inText->nativeIndexingLimit >= rangeEnd - inText->chunkNativeStart) {

//         // Input UText is in one contiguous UTF-16 chunk.
//         // Use Read-only aliasing UnicodeString.
//         inString.setTo(FALSE,
//                        inText->chunkContents + rangeStart - inText->chunkNativeStart,
//                        rangeEnd - rangeStart);
//     } else {
//         // Copy the text from the original inText (UText) to inString (UnicodeString).
//         // Create a map from UnicodeString indices -> UText offsets.
//         utext_setNativeIndex(inText, rangeStart);
//         int32_t limit = rangeEnd;
//         U_ASSERT(limit <= utext_nativeLength(inText));
//         if (limit > utext_nativeLength(inText)) {
//             limit = (int32_t)utext_nativeLength(inText);
//         }
//         inputMap.adoptInsteadAndCheckErrorCode(new UVector32(status), status);
//         if (U_FAILURE(status)) {
//             return 0;
//         }
//         while (utext_getNativeIndex(inText) < limit) {
//             int32_t nativePosition = (int32_t)utext_getNativeIndex(inText);
//             UChar32 c = utext_next32(inText);
//             U_ASSERT(c != U_SENTINEL);
//             inString.append(c);
//             while (inputMap->size() < inString.length()) {
//                 inputMap->addElement(nativePosition, status);
//             }
//         }
//         inputMap->addElement(limit, status);
//     }


//     if (!nfkcNorm2->isNormalized(inString, status)) {
//         UnicodeString normalizedInput;
//         //  normalizedMap[normalizedInput position] ==  original UText position.
//         LocalPointer<UVector32> normalizedMap(new UVector32(status), status);
//         if (U_FAILURE(status)) {
//             return 0;
//         }

//         UnicodeString fragment;
//         UnicodeString normalizedFragment;
//         for (int32_t srcI = 0; srcI < inString.length();) {  // Once per normalization chunk
//             fragment.remove();
//             int32_t fragmentStartI = srcI;
//             UChar32 c = inString.char32At(srcI);
//             for (;;) {
//                 fragment.append(c);
//                 srcI = inString.moveIndex32(srcI, 1);
//                 if (srcI == inString.length()) {
//                     break;
//                 }
//                 c = inString.char32At(srcI);
//                 if (nfkcNorm2->hasBoundaryBefore(c)) {
//                     break;
//                 }
//             }
//             nfkcNorm2->normalize(fragment, normalizedFragment, status);
//             normalizedInput.append(normalizedFragment);

//             // Map every position in the normalized chunk to the start of the chunk
//             //   in the original input.
//             int32_t fragmentOriginalStart = inputMap.isValid() ?
//                     inputMap->elementAti(fragmentStartI) : fragmentStartI+rangeStart;
//             while (normalizedMap->size() < normalizedInput.length()) {
//                 normalizedMap->addElement(fragmentOriginalStart, status);
//                 if (U_FAILURE(status)) {
//                     break;
//                 }
//             }
//         }
//         U_ASSERT(normalizedMap->size() == normalizedInput.length());
//         int32_t nativeEnd = inputMap.isValid() ?
//                 inputMap->elementAti(inString.length()) : inString.length()+rangeStart;
//         normalizedMap->addElement(nativeEnd, status);

//         inputMap = std::move(normalizedMap);
//         inString = std::move(normalizedInput);
//     }

//     int32_t numCodePts = inString.countChar32();
//     if (numCodePts != inString.length()) {
//         // There are supplementary characters in the input.
//         // The dictionary will produce boundary positions in terms of code point indexes,
//         //   not in terms of code unit string indexes.
//         // Use the inputMap mechanism to take care of this in addition to indexing differences
//         //    from normalization and/or UTF-8 input.
//         UBool hadExistingMap = inputMap.isValid();
//         if (!hadExistingMap) {
//             inputMap.adoptInsteadAndCheckErrorCode(new UVector32(status), status);
//             if (U_FAILURE(status)) {
//                 return 0;
//             }
//         }
//         int32_t cpIdx = 0;
//         for (int32_t cuIdx = 0; ; cuIdx = inString.moveIndex32(cuIdx, 1)) {
//             U_ASSERT(cuIdx >= cpIdx);
//             if (hadExistingMap) {
//                 inputMap->setElementAt(inputMap->elementAti(cuIdx), cpIdx);
//             } else {
//                 inputMap->addElement(cuIdx+rangeStart, status);
//             }
//             cpIdx++;
//             if (cuIdx == inString.length()) {
//                break;
//             }
//         }
//     }

//     // bestSnlp[i] is the snlp of the best segmentation of the first i
//     // code points in the range to be matched.
//     UVector32 bestSnlp(numCodePts + 1, status);
//     bestSnlp.addElement(0, status);
//     for(int32_t i = 1; i <= numCodePts; i++) {
//         bestSnlp.addElement(kuint32max, status);
//     }


//     // prev[i] is the index of the last CJK code point in the previous word in
//     // the best segmentation of the first i characters.
//     UVector32 prev(numCodePts + 1, status);
//     for(int32_t i = 0; i <= numCodePts; i++){
//         prev.addElement(-1, status);
//     }

//     const int32_t maxWordSize = 20;
//     UVector32 values(numCodePts, status);
//     values.setSize(numCodePts);
//     UVector32 lengths(numCodePts, status);
//     lengths.setSize(numCodePts);

//     UText fu = UTEXT_INITIALIZER;
//     utext_openUnicodeString(&fu, &inString, &status);

//     // Dynamic programming to find the best segmentation.

//     // In outer loop, i  is the code point index,
//     //                ix is the corresponding string (code unit) index.
//     //    They differ when the string contains supplementary characters.
//     int32_t ix = 0;
//     bool is_prev_katakana = false;
//     for (int32_t i = 0;  i < numCodePts;  ++i, ix = inString.moveIndex32(ix, 1)) {
//         if ((uint32_t)bestSnlp.elementAti(i) == kuint32max) {
//             continue;
//         }

//         int32_t count;
//         utext_setNativeIndex(&fu, ix);
//         count = fDictionary->matches(&fu, maxWordSize, numCodePts,
//                              NULL, lengths.getBuffer(), values.getBuffer(), NULL);
//                              // Note: lengths is filled with code point lengths
//                              //       The NULL parameter is the ignored code unit lengths.

//         // if there are no single character matches found in the dictionary
//         // starting with this character, treat character as a 1-character word
//         // with the highest value possible, i.e. the least likely to occur.
//         // Exclude Korean characters from this treatment, as they should be left
//         // together by default.
//         if ((count == 0 || lengths.elementAti(0) != 1) &&
//                 !fHangulWordSet.contains(inString.char32At(ix))) {
//             values.setElementAt(maxSnlp, count);   // 255
//             lengths.setElementAt(1, count++);
//         }

//         for (int32_t j = 0; j < count; j++) {
//             uint32_t newSnlp = (uint32_t)bestSnlp.elementAti(i) + (uint32_t)values.elementAti(j);
//             int32_t ln_j_i = lengths.elementAti(j) + i;
//             if (newSnlp < (uint32_t)bestSnlp.elementAti(ln_j_i)) {
//                 bestSnlp.setElementAt(newSnlp, ln_j_i);
//                 prev.setElementAt(i, ln_j_i);
//             }
//         }

//         // In Japanese,
//         // Katakana word in single character is pretty rare. So we apply
//         // the following heuristic to Katakana: any continuous run of Katakana
//         // characters is considered a candidate word with a default cost
//         // specified in the katakanaCost table according to its length.

//         bool is_katakana = isKatakana(inString.char32At(ix));
//         int32_t katakanaRunLength = 1;
//         if (!is_prev_katakana && is_katakana) {
//             int32_t j = inString.moveIndex32(ix, 1);
//             // Find the end of the continuous run of Katakana characters
//             while (j < inString.length() && katakanaRunLength < kMaxKatakanaGroupLength &&
//                     isKatakana(inString.char32At(j))) {
//                 j = inString.moveIndex32(j, 1);
//                 katakanaRunLength++;
//             }
//             if (katakanaRunLength < kMaxKatakanaGroupLength) {
//                 uint32_t newSnlp = bestSnlp.elementAti(i) + getKatakanaCost(katakanaRunLength);
//                 if (newSnlp < (uint32_t)bestSnlp.elementAti(i+katakanaRunLength)) {
//                     bestSnlp.setElementAt(newSnlp, i+katakanaRunLength);
//                     prev.setElementAt(i, i+katakanaRunLength);  // prev[j] = i;
//                 }
//             }
//         }
//         is_prev_katakana = is_katakana;
//     }
//     utext_close(&fu);

//     // Start pushing the optimal offset index into t_boundary (t for tentative).
//     // prev[numCodePts] is guaranteed to be meaningful.
//     // We'll first push in the reverse order, i.e.,
//     // t_boundary[0] = numCodePts, and afterwards do a swap.
//     UVector32 t_boundary(numCodePts+1, status);

//     int32_t numBreaks = 0;
//     // No segmentation found, set boundary to end of range
//     if ((uint32_t)bestSnlp.elementAti(numCodePts) == kuint32max) {
//         t_boundary.addElement(numCodePts, status);
//         numBreaks++;
//     } else {
//         for (int32_t i = numCodePts; i > 0; i = prev.elementAti(i)) {
//             t_boundary.addElement(i, status);
//             numBreaks++;
//         }
//         U_ASSERT(prev.elementAti(t_boundary.elementAti(numBreaks - 1)) == 0);
//     }

//     // Add a break for the start of the dictionary range if there is not one
//     // there already.
//     if (foundBreaks.size() == 0 || foundBreaks.peeki() < rangeStart) {
//         t_boundary.addElement(0, status);
//         numBreaks++;
//     }

//     // Now that we're done, convert positions in t_boundary[] (indices in
//     // the normalized input string) back to indices in the original input UText
//     // while reversing t_boundary and pushing values to foundBreaks.
//     int32_t prevCPPos = -1;
//     int32_t prevUTextPos = -1;
//     for (int32_t i = numBreaks-1; i >= 0; i--) {
//         int32_t cpPos = t_boundary.elementAti(i);
//         U_ASSERT(cpPos > prevCPPos);
//         int32_t utextPos =  inputMap.isValid() ? inputMap->elementAti(cpPos) : cpPos + rangeStart;
//         U_ASSERT(utextPos >= prevUTextPos);
//         if (utextPos > prevUTextPos) {
//             // Boundaries are added to foundBreaks output in ascending order.
//             U_ASSERT(foundBreaks.size() == 0 || foundBreaks.peeki() < utextPos);
//             foundBreaks.push(utextPos, status);
//         } else {
//             // Normalization expanded the input text, the dictionary found a boundary
//             // within the expansion, giving two boundaries with the same index in the
//             // original text. Ignore the second. See ticket #12918.
//             --numBreaks;
//         }
//         prevCPPos = cpPos;
//         prevUTextPos = utextPos;
//     }
//     (void)prevCPPos; // suppress compiler warnings about unused variable

//     // inString goes out of scope
//     // inputMap goes out of scope
//     return numBreaks;
// }
// #endif

// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2014-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* dictionarydata.h
*
* created on: 2012may31
* created by: Markus W. Scherer & Maxime Serrano
*/

// #include "dictionarydata.h"
// #include "unicode/ucharstrie.h"
// #include "unicode/bytestrie.h"
// #include "unicode/udata.h"
// #include "cmemory.h"

// #if !UCONFIG_NO_BREAK_ITERATION

// U_NAMESPACE_BEGIN

// const int32_t  DictionaryData::TRIE_TYPE_BYTES = 0;
// const int32_t  DictionaryData::TRIE_TYPE_UCHARS = 1;
// const int32_t  DictionaryData::TRIE_TYPE_MASK = 7;
// const int32_t  DictionaryData::TRIE_HAS_VALUES = 8;

// const int32_t  DictionaryData::TRANSFORM_NONE = 0;
// const int32_t  DictionaryData::TRANSFORM_TYPE_OFFSET = 0x1000000;
// const int32_t  DictionaryData::TRANSFORM_TYPE_MASK = 0x7f000000;
// const int32_t  DictionaryData::TRANSFORM_OFFSET_MASK = 0x1fffff;

// DictionaryMatcher::~DictionaryMatcher() {
// }

// UCharsDictionaryMatcher::~UCharsDictionaryMatcher() {
//     udata_close(file);
// }

// int32_t UCharsDictionaryMatcher::getType() const {
//     return DictionaryData::TRIE_TYPE_UCHARS;
// }

// int32_t UCharsDictionaryMatcher::matches(UText *text, int32_t maxLength, int32_t limit,
//                             int32_t *lengths, int32_t *cpLengths, int32_t *values,
//                             int32_t *prefix) const {

//     UCharsTrie uct(characters);
//     int32_t startingTextIndex = (int32_t)utext_getNativeIndex(text);
//     int32_t wordCount = 0;
//     int32_t codePointsMatched = 0;

//     for (UChar32 c = utext_next32(text); c >= 0; c=utext_next32(text)) {
//         UStringTrieResult result = (codePointsMatched == 0) ? uct.first(c) : uct.next(c);
//         int32_t lengthMatched = (int32_t)utext_getNativeIndex(text) - startingTextIndex;
//         codePointsMatched += 1;
//         if (USTRINGTRIE_HAS_VALUE(result)) {
//             if (wordCount < limit) {
//                 if (values != NULL) {
//                     values[wordCount] = uct.getValue();
//                 }
//                 if (lengths != NULL) {
//                     lengths[wordCount] = lengthMatched;
//                 }
//                 if (cpLengths != NULL) {
//                     cpLengths[wordCount] = codePointsMatched;
//                 }
//                 ++wordCount;
//             }
//             if (result == USTRINGTRIE_FINAL_VALUE) {
//                 break;
//             }
//         }
//         else if (result == USTRINGTRIE_NO_MATCH) {
//             break;
//         }
//         if (lengthMatched >= maxLength) {
//             break;
//         }
//     }

//     if (prefix != NULL) {
//         *prefix = codePointsMatched;
//     }
//     return wordCount;
// }

// BytesDictionaryMatcher::~BytesDictionaryMatcher() {
//     udata_close(file);
// }

// UChar32 BytesDictionaryMatcher::transform(UChar32 c) const {
//     if ((transformConstant & DictionaryData::TRANSFORM_TYPE_MASK) == DictionaryData::TRANSFORM_TYPE_OFFSET) {
//         if (c == 0x200D) {
//             return 0xFF;
//         } else if (c == 0x200C) {
//             return 0xFE;
//         }
//         int32_t delta = c - (transformConstant & DictionaryData::TRANSFORM_OFFSET_MASK);
//         if (delta < 0 || 0xFD < delta) {
//             return U_SENTINEL;
//         }
//         return (UChar32)delta;
//     }
//     return c;
// }

// int32_t BytesDictionaryMatcher::getType() const {
//     return DictionaryData::TRIE_TYPE_BYTES;
// }

// int32_t BytesDictionaryMatcher::matches(UText *text, int32_t maxLength, int32_t limit,
//                             int32_t *lengths, int32_t *cpLengths, int32_t *values,
//                             int32_t *prefix) const {
//     BytesTrie bt(characters);
//     int32_t startingTextIndex = (int32_t)utext_getNativeIndex(text);
//     int32_t wordCount = 0;
//     int32_t codePointsMatched = 0;

//     for (UChar32 c = utext_next32(text); c >= 0; c=utext_next32(text)) {
//         UStringTrieResult result = (codePointsMatched == 0) ? bt.first(transform(c)) : bt.next(transform(c));
//         int32_t lengthMatched = (int32_t)utext_getNativeIndex(text) - startingTextIndex;
//         codePointsMatched += 1;
//         if (USTRINGTRIE_HAS_VALUE(result)) {
//             if (wordCount < limit) {
//                 if (values != NULL) {
//                     values[wordCount] = bt.getValue();
//                 }
//                 if (lengths != NULL) {
//                     lengths[wordCount] = lengthMatched;
//                 }
//                 if (cpLengths != NULL) {
//                     cpLengths[wordCount] = codePointsMatched;
//                 }
//                 ++wordCount;
//             }
//             if (result == USTRINGTRIE_FINAL_VALUE) {
//                 break;
//             }
//         }
//         else if (result == USTRINGTRIE_NO_MATCH) {
//             break;
//         }
//         if (lengthMatched >= maxLength) {
//             break;
//         }
//     }

//     if (prefix != NULL) {
//         *prefix = codePointsMatched;
//     }
//     return wordCount;
// }


// U_NAMESPACE_END

// U_NAMESPACE_USE

// U_CAPI int32_t U_EXPORT2
// udict_swap(const UDataSwapper *ds, const void *inData, int32_t length,
//            void *outData, UErrorCode *pErrorCode) {
//     const UDataInfo *pInfo;
//     int32_t headerSize;
//     const uint8_t *inBytes;
//     uint8_t *outBytes;
//     const int32_t *inIndexes;
//     int32_t indexes[DictionaryData::IX_COUNT];
//     int32_t i, offset, size;

//     headerSize = udata_swapDataHeader(ds, inData, length, outData, pErrorCode);
//     if (pErrorCode == NULL || U_FAILURE(*pErrorCode)) return 0;
//     pInfo = (const UDataInfo *)((const char *)inData + 4);
//     if (!(pInfo->dataFormat[0] == 0x44 &&
//           pInfo->dataFormat[1] == 0x69 &&
//           pInfo->dataFormat[2] == 0x63 &&
//           pInfo->dataFormat[3] == 0x74 &&
//           pInfo->formatVersion[0] == 1)) {
//         udata_printError(ds, "udict_swap(): data format %02x.%02x.%02x.%02x (format version %02x) is not recognized as dictionary data\n",
//                          pInfo->dataFormat[0], pInfo->dataFormat[1], pInfo->dataFormat[2], pInfo->dataFormat[3], pInfo->formatVersion[0]);
//         *pErrorCode = U_UNSUPPORTED_ERROR;
//         return 0;
//     }

//     inBytes = (const uint8_t *)inData + headerSize;
//     outBytes = (uint8_t *)outData + headerSize;

//     inIndexes = (const int32_t *)inBytes;
//     if (length >= 0) {
//         length -= headerSize;
//         if (length < (int32_t)(sizeof(indexes))) {
//             udata_printError(ds, "udict_swap(): too few bytes (%d after header) for dictionary data\n", length);
//             *pErrorCode = U_INDEX_OUTOFBOUNDS_ERROR;
//             return 0;
//         }
//     }

//     for (i = 0; i < DictionaryData::IX_COUNT; i++) {
//         indexes[i] = udata_readInt32(ds, inIndexes[i]);
//     }

//     size = indexes[DictionaryData::IX_TOTAL_SIZE];

//     if (length >= 0) {
//         if (length < size) {
//             udata_printError(ds, "udict_swap(): too few bytes (%d after header) for all of dictionary data\n", length);
//             *pErrorCode = U_INDEX_OUTOFBOUNDS_ERROR;
//             return 0;
//         }

//         if (inBytes != outBytes) {
//             uprv_memcpy(outBytes, inBytes, size);
//         }

//         offset = 0;
//         ds->swapArray32(ds, inBytes, sizeof(indexes), outBytes, pErrorCode);
//         offset = (int32_t)sizeof(indexes);
//         int32_t trieType = indexes[DictionaryData::IX_TRIE_TYPE] & DictionaryData::TRIE_TYPE_MASK;
//         int32_t nextOffset = indexes[DictionaryData::IX_RESERVED1_OFFSET];

//         if (trieType == DictionaryData::TRIE_TYPE_UCHARS) {
//             ds->swapArray16(ds, inBytes + offset, nextOffset - offset, outBytes + offset, pErrorCode);
//         } else if (trieType == DictionaryData::TRIE_TYPE_BYTES) {
//             // nothing to do
//         } else {
//             udata_printError(ds, "udict_swap(): unknown trie type!\n");
//             *pErrorCode = U_UNSUPPORTED_ERROR;
//             return 0;
//         }

//         // these next two sections are empty in the current format,
//         // but may be used later.
//         offset = nextOffset;
//         nextOffset = indexes[DictionaryData::IX_RESERVED2_OFFSET];
//         offset = nextOffset;
//         nextOffset = indexes[DictionaryData::IX_TOTAL_SIZE];
//         offset = nextOffset;
//     }
//     return headerSize + size;
// }
// #endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*******************************************************************************
* Copyright (C) 2008, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File DTINTRV.CPP 
*
*******************************************************************************
*/



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2008-2009, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File DTINTRV.H 
*
*******************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API



/**
 * \file
 * \brief C++ API: Date Interval data type
 */

U_NAMESPACE_BEGIN


/**
 * This class represents a date interval.
 * It is a pair of UDate representing from UDate 1 to UDate 2.
 * @stable ICU 4.0
**/
class U_COMMON_API DateInterval : public UObject {
public:

    /** 
     * Construct a DateInterval given a from date and a to date.
     * @param fromDate  The from date in date interval.
     * @param toDate    The to date in date interval.
     * @stable ICU 4.0
     */
    DateInterval(UDate fromDate, UDate toDate);

    /**
     * destructor
     * @stable ICU 4.0
     */
    virtual ~DateInterval();
 
    /** 
     * Get the from date.
     * @return  the from date in dateInterval.
     * @stable ICU 4.0
     */
    inline UDate getFromDate() const;

    /** 
     * Get the to date.
     * @return  the to date in dateInterval.
     * @stable ICU 4.0
     */
    inline UDate getToDate() const;


    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       derived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 4.0
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 4.0
     */
    virtual UClassID getDynamicClassID(void) const;

    
    /**
     * Copy constructor.
     * @stable ICU 4.0
     */
    DateInterval(const DateInterval& other);

    /**
     * Default assignment operator
     * @stable ICU 4.0
     */
    DateInterval& operator=(const DateInterval&);

    /**
     * Equality operator.
     * @return TRUE if the two DateIntervals are the same
     * @stable ICU 4.0
     */
    virtual UBool operator==(const DateInterval& other) const;

    /**
     * Non-equality operator
     * @return TRUE if the two DateIntervals are not the same
     * @stable ICU 4.0
     */
    inline UBool operator!=(const DateInterval& other) const;


    /**
     * clone this object. 
     * The caller owns the result and should delete it when done.
     * @return a cloned DateInterval
     * @stable ICU 4.0
     */
     virtual DateInterval* clone() const;

private:
    /** 
     * Default constructor, not implemented.
     */
    DateInterval();

    UDate fromDate;
    UDate toDate;

} ;// end class DateInterval


inline UDate 
DateInterval::getFromDate() const { 
    return fromDate; 
}


inline UDate 
DateInterval::getToDate() const { 
    return toDate; 
}


inline UBool 
DateInterval::operator!=(const DateInterval& other) const { 
    return ( !operator==(other) );
}


U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */




U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DateInterval)

//DateInterval::DateInterval(){}


DateInterval::DateInterval(UDate from, UDate to)
:   fromDate(from),
    toDate(to)
{}


DateInterval::~DateInterval(){}


DateInterval::DateInterval(const DateInterval& other)
: UObject(other) {
    *this = other;
}   


DateInterval&
DateInterval::operator=(const DateInterval& other) {
    if ( this != &other ) {
        fromDate = other.fromDate;
        toDate = other.toDate;
    }
    return *this;
}


DateInterval* 
DateInterval::clone() const {
    return new DateInterval(*this);
}


UBool 
DateInterval::operator==(const DateInterval& other) const { 
    return ( fromDate == other.fromDate && toDate == other.toDate );
}


U_NAMESPACE_END

// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// edits.cpp
// created: 2017feb08 Markus W. Scherer






// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 **********************************************************************
 *   Copyright (c) 2001-2011, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 **********************************************************************
 *   Date        Name        Description
 *   11/19/2001  aliu        Creation.
 **********************************************************************
 */






//--------------------------------------------------------------------
// class ICU_Utility
// i18n utility functions, scoped into the class ICU_Utility.
//--------------------------------------------------------------------

U_NAMESPACE_BEGIN

class UnicodeMatcher;

class U_COMMON_API ICU_Utility /* not : public UObject because all methods are static */ {
 public:

    /**
     * Append a number to the given UnicodeString in the given radix.
     * Standard digits '0'-'9' are used and letters 'A'-'Z' for
     * radices 11 through 36.
     * @param result the digits of the number are appended here
     * @param n the number to be converted to digits; may be negative.
     * If negative, a '-' is prepended to the digits.
     * @param radix a radix from 2 to 36 inclusive.
     * @param minDigits the minimum number of digits, not including
     * any '-', to produce.  Values less than 2 have no effect.  One
     * digit is always emitted regardless of this parameter.
     * @return a reference to result
     */
    static UnicodeString& appendNumber(UnicodeString& result, int32_t n,
                                       int32_t radix = 10,
                                       int32_t minDigits = 1);

    /** Returns a bogus UnicodeString by value. */
    static inline UnicodeString makeBogusString() {
        UnicodeString result;
        result.setToBogus();
        return result;
    }

    /**
     * Return true if the character is NOT printable ASCII.
     *
     * This method should really be in UnicodeString (or similar).  For
     * now, we implement it here and share it with friend classes.
     */
    static UBool isUnprintable(UChar32 c);

    /**
     * Escape unprintable characters using \uxxxx notation for U+0000 to
     * U+FFFF and \Uxxxxxxxx for U+10000 and above.  If the character is
     * printable ASCII, then do nothing and return FALSE.  Otherwise,
     * append the escaped notation and return TRUE.
     */
    static UBool escapeUnprintable(UnicodeString& result, UChar32 c);

    /**
     * Returns the index of a character, ignoring quoted text.
     * For example, in the string "abc'hide'h", the 'h' in "hide" will not be
     * found by a search for 'h'.
     * @param text text to be searched
     * @param start the beginning index, inclusive; <code>0 <= start
     * <= limit</code>.
     * @param limit the ending index, exclusive; <code>start <= limit
     * <= text.length()</code>.
     * @param c character to search for
     * @return Offset of the first instance of c, or -1 if not found.
     */
//?FOR FUTURE USE.  DISABLE FOR NOW for coverage reasons.
//    static int32_t quotedIndexOf(const UnicodeString& text,
//                                 int32_t start, int32_t limit,
//                                 UChar c);

    /**
     * Skip over a sequence of zero or more white space characters at pos.
     * @param advance if true, advance pos to the first non-white-space
     * character at or after pos, or str.length(), if there is none.
     * Otherwise leave pos unchanged.
     * @return the index of the first non-white-space character at or
     * after pos, or str.length(), if there is none.
     */
    static int32_t skipWhitespace(const UnicodeString& str, int32_t& pos,
                                  UBool advance = FALSE);

    /**
     * Skip over Pattern_White_Space in a Replaceable.
     * Skipping may be done in the forward or
     * reverse direction.  In either case, the leftmost index will be
     * inclusive, and the rightmost index will be exclusive.  That is,
     * given a range defined as [start, limit), the call
     * skipWhitespace(text, start, limit) will advance start past leading
     * whitespace, whereas the call skipWhitespace(text, limit, start),
     * will back up limit past trailing whitespace.
     * @param text the text to be analyzed
     * @param pos either the start or limit of a range of 'text', to skip
     * leading or trailing whitespace, respectively
     * @param stop either the limit or start of a range of 'text', to skip
     * leading or trailing whitespace, respectively
     * @return the new start or limit, depending on what was passed in to
     * 'pos'
     */
//?FOR FUTURE USE.  DISABLE FOR NOW for coverage reasons.
//?    static int32_t skipWhitespace(const Replaceable& text,
//?                                  int32_t pos, int32_t stop);

    /**
     * Parse a single non-whitespace character 'ch', optionally
     * preceded by whitespace.
     * @param id the string to be parsed
     * @param pos INPUT-OUTPUT parameter.  On input, pos[0] is the
     * offset of the first character to be parsed.  On output, pos[0]
     * is the index after the last parsed character.  If the parse
     * fails, pos[0] will be unchanged.
     * @param ch the non-whitespace character to be parsed.
     * @return true if 'ch' is seen preceded by zero or more
     * whitespace characters.
     */
    static UBool parseChar(const UnicodeString& id, int32_t& pos, UChar ch);

    /**
     * Parse a pattern string starting at offset pos.  Keywords are
     * matched case-insensitively.  Spaces may be skipped and may be
     * optional or required.  Integer values may be parsed, and if
     * they are, they will be returned in the given array.  If
     * successful, the offset of the next non-space character is
     * returned.  On failure, -1 is returned.
     * @param pattern must only contain lowercase characters, which
     * will match their uppercase equivalents as well.  A space
     * character matches one or more required spaces.  A '~' character
     * matches zero or more optional spaces.  A '#' character matches
     * an integer and stores it in parsedInts, which the caller must
     * ensure has enough capacity.
     * @param parsedInts array to receive parsed integers.  Caller
     * must ensure that parsedInts.length is >= the number of '#'
     * signs in 'pattern'.
     * @return the position after the last character parsed, or -1 if
     * the parse failed
     */
    static int32_t parsePattern(const UnicodeString& rule, int32_t pos, int32_t limit,
                                const UnicodeString& pattern, int32_t* parsedInts);
        
    /**
     * Parse a pattern string within the given Replaceable and a parsing
     * pattern.  Characters are matched literally and case-sensitively
     * except for the following special characters:
     *
     * ~  zero or more Pattern_White_Space chars
     *
     * If end of pattern is reached with all matches along the way,
     * pos is advanced to the first unparsed index and returned.
     * Otherwise -1 is returned.
     * @param pat pattern that controls parsing
     * @param text text to be parsed, starting at index
     * @param index offset to first character to parse
     * @param limit offset after last character to parse
     * @return index after last parsed character, or -1 on parse failure.
     */
    static int32_t parsePattern(const UnicodeString& pat,
                                const Replaceable& text,
                                int32_t index,
                                int32_t limit);

    /**
     * Parse an integer at pos, either of the form \d+ or of the form
     * 0x[0-9A-Fa-f]+ or 0[0-7]+, that is, in standard decimal, hex,
     * or octal format.
     * @param pos INPUT-OUTPUT parameter.  On input, the index of the first
     * character to parse.  On output, the index of the character after the
     * last parsed character.
     */
    static int32_t parseInteger(const UnicodeString& rule, int32_t& pos, int32_t limit);

    /**
     * Parse an integer at pos using only ASCII digits.
     * Base 10 only.
     * @param pos INPUT-OUTPUT parameter.  On input, the index of the first
     * character to parse.  On output, the index of the character after the
     * last parsed character.
     */
    static int32_t parseAsciiInteger(const UnicodeString& str, int32_t& pos);

    /**
     * Parse a Unicode identifier from the given string at the given
     * position.  Return the identifier, or an empty string if there
     * is no identifier.
     * @param str the string to parse
     * @param pos INPUT-OUPUT parameter.  On INPUT, pos is the
     * first character to examine.  It must be less than str.length(),
     * and it must not point to a whitespace character.  That is, must
     * have pos < str.length() and
     * !UCharacter::isWhitespace(str.char32At(pos)).  On
     * OUTPUT, the position after the last parsed character.
     * @return the Unicode identifier, or an empty string if there is
     * no valid identifier at pos.
     */
    static UnicodeString parseUnicodeIdentifier(const UnicodeString& str, int32_t& pos);

    /**
     * Parse an unsigned 31-bit integer at the given offset.  Use
     * UCharacter.digit() to parse individual characters into digits.
     * @param text the text to be parsed
     * @param pos INPUT-OUTPUT parameter.  On entry, pos is the
     * offset within text at which to start parsing; it should point
     * to a valid digit.  On exit, pos is the offset after the last
     * parsed character.  If the parse failed, it will be unchanged on
     * exit.  Must be >= 0 on entry.
     * @param radix the radix in which to parse; must be >= 2 and <=
     * 36.
     * @return a non-negative parsed number, or -1 upon parse failure.
     * Parse fails if there are no digits, that is, if pos does not
     * point to a valid digit on entry, or if the number to be parsed
     * does not fit into a 31-bit unsigned integer.
     */
    static int32_t parseNumber(const UnicodeString& text,
                               int32_t& pos, int8_t radix);

    static void appendToRule(UnicodeString& rule,
                             UChar32 c,
                             UBool isLiteral,
                             UBool escapeUnprintable,
                             UnicodeString& quoteBuf);
    
    static void appendToRule(UnicodeString& rule,
                             const UnicodeString& text,
                             UBool isLiteral,
                             UBool escapeUnprintable,
                             UnicodeString& quoteBuf);

    static void appendToRule(UnicodeString& rule,
                             const UnicodeMatcher* matcher,
                             UBool escapeUnprintable,
                             UnicodeString& quoteBuf);

private:
    // do not instantiate
    ICU_Utility();
};

U_NAMESPACE_END



U_NAMESPACE_BEGIN

namespace {

// 0000uuuuuuuuuuuu records u+1 unchanged text units.
const int32_t MAX_UNCHANGED_LENGTH = 0x1000;
const int32_t MAX_UNCHANGED = MAX_UNCHANGED_LENGTH - 1;

// 0mmmnnnccccccccc with m=1..6 records ccc+1 replacements of m:n text units.
const int32_t MAX_SHORT_CHANGE_OLD_LENGTH = 6;
const int32_t MAX_SHORT_CHANGE_NEW_LENGTH = 7;
const int32_t SHORT_CHANGE_NUM_MASK = 0x1ff;
const int32_t MAX_SHORT_CHANGE = 0x6fff;

// 0111mmmmmmnnnnnn records a replacement of m text units with n.
// m or n = 61: actual length follows in the next edits array unit.
// m or n = 62..63: actual length follows in the next two edits array units.
// Bit 30 of the actual length is in the head unit.
// Trailing units have bit 15 set.
const int32_t LENGTH_IN_1TRAIL = 61;
const int32_t LENGTH_IN_2TRAIL = 62;

}  // namespace

void Edits::releaseArray() U_NOEXCEPT {
    if (array != stackArray) {
        uprv_free(array);
    }
}

Edits &Edits::copyArray(const Edits &other) {
    if (U_FAILURE(errorCode_)) {
        length = delta = numChanges = 0;
        return *this;
    }
    if (length > capacity) {
        uint16_t *newArray = (uint16_t *)uprv_malloc((size_t)length * 2);
        if (newArray == nullptr) {
            length = delta = numChanges = 0;
            errorCode_ = U_MEMORY_ALLOCATION_ERROR;
            return *this;
        }
        releaseArray();
        array = newArray;
        capacity = length;
    }
    if (length > 0) {
        uprv_memcpy(array, other.array, (size_t)length * 2);
    }
    return *this;
}

Edits &Edits::moveArray(Edits &src) U_NOEXCEPT {
    if (U_FAILURE(errorCode_)) {
        length = delta = numChanges = 0;
        return *this;
    }
    releaseArray();
    if (length > STACK_CAPACITY) {
        array = src.array;
        capacity = src.capacity;
        src.array = src.stackArray;
        src.capacity = STACK_CAPACITY;
        src.reset();
        return *this;
    }
    array = stackArray;
    capacity = STACK_CAPACITY;
    if (length > 0) {
        uprv_memcpy(array, src.array, (size_t)length * 2);
    }
    return *this;
}

Edits &Edits::operator=(const Edits &other) {
    length = other.length;
    delta = other.delta;
    numChanges = other.numChanges;
    errorCode_ = other.errorCode_;
    return copyArray(other);
}

Edits &Edits::operator=(Edits &&src) U_NOEXCEPT {
    length = src.length;
    delta = src.delta;
    numChanges = src.numChanges;
    errorCode_ = src.errorCode_;
    return moveArray(src);
}

Edits::~Edits() {
    releaseArray();
}

void Edits::reset() U_NOEXCEPT {
    length = delta = numChanges = 0;
    errorCode_ = U_ZERO_ERROR;
}

void Edits::addUnchanged(int32_t unchangedLength) {
    if(U_FAILURE(errorCode_) || unchangedLength == 0) { return; }
    if(unchangedLength < 0) {
        errorCode_ = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    // Merge into previous unchanged-text record, if any.
    int32_t last = lastUnit();
    if(last < MAX_UNCHANGED) {
        int32_t remaining = MAX_UNCHANGED - last;
        if (remaining >= unchangedLength) {
            setLastUnit(last + unchangedLength);
            return;
        }
        setLastUnit(MAX_UNCHANGED);
        unchangedLength -= remaining;
    }
    // Split large lengths into multiple units.
    while(unchangedLength >= MAX_UNCHANGED_LENGTH) {
        append(MAX_UNCHANGED);
        unchangedLength -= MAX_UNCHANGED_LENGTH;
    }
    // Write a small (remaining) length.
    if(unchangedLength > 0) {
        append(unchangedLength - 1);
    }
}

void Edits::addReplace(int32_t oldLength, int32_t newLength) {
    if(U_FAILURE(errorCode_)) { return; }
    if(oldLength < 0 || newLength < 0) {
        errorCode_ = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if (oldLength == 0 && newLength == 0) {
        return;
    }
    ++numChanges;
    int32_t newDelta = newLength - oldLength;
    if (newDelta != 0) {
        if ((newDelta > 0 && delta >= 0 && newDelta > (INT32_MAX - delta)) ||
                (newDelta < 0 && delta < 0 && newDelta < (INT32_MIN - delta))) {
            // Integer overflow or underflow.
            errorCode_ = U_INDEX_OUTOFBOUNDS_ERROR;
            return;
        }
        delta += newDelta;
    }

    if(0 < oldLength && oldLength <= MAX_SHORT_CHANGE_OLD_LENGTH &&
            newLength <= MAX_SHORT_CHANGE_NEW_LENGTH) {
        // Merge into previous same-lengths short-replacement record, if any.
        int32_t u = (oldLength << 12) | (newLength << 9);
        int32_t last = lastUnit();
        if(MAX_UNCHANGED < last && last < MAX_SHORT_CHANGE &&
                (last & ~SHORT_CHANGE_NUM_MASK) == u &&
                (last & SHORT_CHANGE_NUM_MASK) < SHORT_CHANGE_NUM_MASK) {
            setLastUnit(last + 1);
            return;
        }
        append(u);
        return;
    }

    int32_t head = 0x7000;
    if (oldLength < LENGTH_IN_1TRAIL && newLength < LENGTH_IN_1TRAIL) {
        head |= oldLength << 6;
        head |= newLength;
        append(head);
    } else if ((capacity - length) >= 5 || growArray()) {
        int32_t limit = length + 1;
        if(oldLength < LENGTH_IN_1TRAIL) {
            head |= oldLength << 6;
        } else if(oldLength <= 0x7fff) {
            head |= LENGTH_IN_1TRAIL << 6;
            array[limit++] = (uint16_t)(0x8000 | oldLength);
        } else {
            head |= (LENGTH_IN_2TRAIL + (oldLength >> 30)) << 6;
            array[limit++] = (uint16_t)(0x8000 | (oldLength >> 15));
            array[limit++] = (uint16_t)(0x8000 | oldLength);
        }
        if(newLength < LENGTH_IN_1TRAIL) {
            head |= newLength;
        } else if(newLength <= 0x7fff) {
            head |= LENGTH_IN_1TRAIL;
            array[limit++] = (uint16_t)(0x8000 | newLength);
        } else {
            head |= LENGTH_IN_2TRAIL + (newLength >> 30);
            array[limit++] = (uint16_t)(0x8000 | (newLength >> 15));
            array[limit++] = (uint16_t)(0x8000 | newLength);
        }
        array[length] = (uint16_t)head;
        length = limit;
    }
}

void Edits::append(int32_t r) {
    if(length < capacity || growArray()) {
        array[length++] = (uint16_t)r;
    }
}

UBool Edits::growArray() {
    int32_t newCapacity;
    if (array == stackArray) {
        newCapacity = 2000;
    } else if (capacity == INT32_MAX) {
        // Not U_BUFFER_OVERFLOW_ERROR because that could be confused on a string transform API
        // with a result-string-buffer overflow.
        errorCode_ = U_INDEX_OUTOFBOUNDS_ERROR;
        return FALSE;
    } else if (capacity >= (INT32_MAX / 2)) {
        newCapacity = INT32_MAX;
    } else {
        newCapacity = 2 * capacity;
    }
    // Grow by at least 5 units so that a maximal change record will fit.
    if ((newCapacity - capacity) < 5) {
        errorCode_ = U_INDEX_OUTOFBOUNDS_ERROR;
        return FALSE;
    }
    uint16_t *newArray = (uint16_t *)uprv_malloc((size_t)newCapacity * 2);
    if (newArray == NULL) {
        errorCode_ = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    uprv_memcpy(newArray, array, (size_t)length * 2);
    releaseArray();
    array = newArray;
    capacity = newCapacity;
    return TRUE;
}

UBool Edits::copyErrorTo(UErrorCode &outErrorCode) const {
    if (U_FAILURE(outErrorCode)) { return TRUE; }
    if (U_SUCCESS(errorCode_)) { return FALSE; }
    outErrorCode = errorCode_;
    return TRUE;
}

Edits &Edits::mergeAndAppend(const Edits &ab, const Edits &bc, UErrorCode &errorCode) {
    if (copyErrorTo(errorCode)) { return *this; }
    // Picture string a --(Edits ab)--> string b --(Edits bc)--> string c.
    // Parallel iteration over both Edits.
    Iterator abIter = ab.getFineIterator();
    Iterator bcIter = bc.getFineIterator();
    UBool abHasNext = TRUE, bcHasNext = TRUE;
    // Copy iterator state into local variables, so that we can modify and subdivide spans.
    // ab old & new length, bc old & new length
    int32_t aLength = 0, ab_bLength = 0, bc_bLength = 0, cLength = 0;
    // When we have different-intermediate-length changes, we accumulate a larger change.
    int32_t pending_aLength = 0, pending_cLength = 0;
    for (;;) {
        // At this point, for each of the two iterators:
        // Either we are done with the locally cached current edit,
        // and its intermediate-string length has been reset,
        // or we will continue to work with a truncated remainder of this edit.
        //
        // If the current edit is done, and the iterator has not yet reached the end,
        // then we fetch the next edit. This is true for at least one of the iterators.
        //
        // Normally it does not matter whether we fetch from ab and then bc or vice versa.
        // However, the result is observably different when
        // ab deletions meet bc insertions at the same intermediate-string index.
        // Some users expect the bc insertions to come first, so we fetch from bc first.
        if (bc_bLength == 0) {
            if (bcHasNext && (bcHasNext = bcIter.next(errorCode)) != 0) {
                bc_bLength = bcIter.oldLength();
                cLength = bcIter.newLength();
                if (bc_bLength == 0) {
                    // insertion
                    if (ab_bLength == 0 || !abIter.hasChange()) {
                        addReplace(pending_aLength, pending_cLength + cLength);
                        pending_aLength = pending_cLength = 0;
                    } else {
                        pending_cLength += cLength;
                    }
                    continue;
                }
            }
            // else see if the other iterator is done, too.
        }
        if (ab_bLength == 0) {
            if (abHasNext && (abHasNext = abIter.next(errorCode)) != 0) {
                aLength = abIter.oldLength();
                ab_bLength = abIter.newLength();
                if (ab_bLength == 0) {
                    // deletion
                    if (bc_bLength == bcIter.oldLength() || !bcIter.hasChange()) {
                        addReplace(pending_aLength + aLength, pending_cLength);
                        pending_aLength = pending_cLength = 0;
                    } else {
                        pending_aLength += aLength;
                    }
                    continue;
                }
            } else if (bc_bLength == 0) {
                // Both iterators are done at the same time:
                // The intermediate-string lengths match.
                break;
            } else {
                // The ab output string is shorter than the bc input string.
                if (!copyErrorTo(errorCode)) {
                    errorCode = U_ILLEGAL_ARGUMENT_ERROR;
                }
                return *this;
            }
        }
        if (bc_bLength == 0) {
            // The bc input string is shorter than the ab output string.
            if (!copyErrorTo(errorCode)) {
                errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            }
            return *this;
        }
        //  Done fetching: ab_bLength > 0 && bc_bLength > 0

        // The current state has two parts:
        // - Past: We accumulate a longer ac edit in the "pending" variables.
        // - Current: We have copies of the current ab/bc edits in local variables.
        //   At least one side is newly fetched.
        //   One side might be a truncated remainder of an edit we fetched earlier.

        if (!abIter.hasChange() && !bcIter.hasChange()) {
            // An unchanged span all the way from string a to string c.
            if (pending_aLength != 0 || pending_cLength != 0) {
                addReplace(pending_aLength, pending_cLength);
                pending_aLength = pending_cLength = 0;
            }
            int32_t unchangedLength = aLength <= cLength ? aLength : cLength;
            addUnchanged(unchangedLength);
            ab_bLength = aLength -= unchangedLength;
            bc_bLength = cLength -= unchangedLength;
            // At least one of the unchanged spans is now empty.
            continue;
        }
        if (!abIter.hasChange() && bcIter.hasChange()) {
            // Unchanged a->b but changed b->c.
            if (ab_bLength >= bc_bLength) {
                // Split the longer unchanged span into change + remainder.
                addReplace(pending_aLength + bc_bLength, pending_cLength + cLength);
                pending_aLength = pending_cLength = 0;
                aLength = ab_bLength -= bc_bLength;
                bc_bLength = 0;
                continue;
            }
            // Handle the shorter unchanged span below like a change.
        } else if (abIter.hasChange() && !bcIter.hasChange()) {
            // Changed a->b and then unchanged b->c.
            if (ab_bLength <= bc_bLength) {
                // Split the longer unchanged span into change + remainder.
                addReplace(pending_aLength + aLength, pending_cLength + ab_bLength);
                pending_aLength = pending_cLength = 0;
                cLength = bc_bLength -= ab_bLength;
                ab_bLength = 0;
                continue;
            }
            // Handle the shorter unchanged span below like a change.
        } else {  // both abIter.hasChange() && bcIter.hasChange()
            if (ab_bLength == bc_bLength) {
                // Changes on both sides up to the same position. Emit & reset.
                addReplace(pending_aLength + aLength, pending_cLength + cLength);
                pending_aLength = pending_cLength = 0;
                ab_bLength = bc_bLength = 0;
                continue;
            }
        }
        // Accumulate the a->c change, reset the shorter side,
        // keep a remainder of the longer one.
        pending_aLength += aLength;
        pending_cLength += cLength;
        if (ab_bLength < bc_bLength) {
            bc_bLength -= ab_bLength;
            cLength = ab_bLength = 0;
        } else {  // ab_bLength > bc_bLength
            ab_bLength -= bc_bLength;
            aLength = bc_bLength = 0;
        }
    }
    if (pending_aLength != 0 || pending_cLength != 0) {
        addReplace(pending_aLength, pending_cLength);
    }
    copyErrorTo(errorCode);
    return *this;
}

Edits::Iterator::Iterator(const uint16_t *a, int32_t len, UBool oc, UBool crs) :
        array(a), index(0), length(len), remaining(0),
        onlyChanges_(oc), coarse(crs),
        dir(0), changed(FALSE), oldLength_(0), newLength_(0),
        srcIndex(0), replIndex(0), destIndex(0) {}

int32_t Edits::Iterator::readLength(int32_t head) {
    if (head < LENGTH_IN_1TRAIL) {
        return head;
    } else if (head < LENGTH_IN_2TRAIL) {
        U_ASSERT(index < length);
        U_ASSERT(array[index] >= 0x8000);
        return array[index++] & 0x7fff;
    } else {
        U_ASSERT((index + 2) <= length);
        U_ASSERT(array[index] >= 0x8000);
        U_ASSERT(array[index + 1] >= 0x8000);
        int32_t len = ((head & 1) << 30) |
                ((int32_t)(array[index] & 0x7fff) << 15) |
                (array[index + 1] & 0x7fff);
        index += 2;
        return len;
    }
}

void Edits::Iterator::updateNextIndexes() {
    srcIndex += oldLength_;
    if (changed) {
        replIndex += newLength_;
    }
    destIndex += newLength_;
}

void Edits::Iterator::updatePreviousIndexes() {
    srcIndex -= oldLength_;
    if (changed) {
        replIndex -= newLength_;
    }
    destIndex -= newLength_;
}

UBool Edits::Iterator::noNext() {
    // No change before or beyond the string.
    dir = 0;
    changed = FALSE;
    oldLength_ = newLength_ = 0;
    return FALSE;
}

UBool Edits::Iterator::next(UBool onlyChanges, UErrorCode &errorCode) {
    // Forward iteration: Update the string indexes to the limit of the current span,
    // and post-increment-read array units to assemble a new span.
    // Leaves the array index one after the last unit of that span.
    if (U_FAILURE(errorCode)) { return FALSE; }
    // We have an errorCode in case we need to start guarding against integer overflows.
    // It is also convenient for caller loops if we bail out when an error was set elsewhere.
    if (dir > 0) {
        updateNextIndexes();
    } else {
        if (dir < 0) {
            // Turn around from previous() to next().
            // Post-increment-read the same span again.
            if (remaining > 0) {
                // Fine-grained iterator:
                // Stay on the current one of a sequence of compressed changes.
                ++index;  // next() rests on the index after the sequence unit.
                dir = 1;
                return TRUE;
            }
        }
        dir = 1;
    }
    if (remaining >= 1) {
        // Fine-grained iterator: Continue a sequence of compressed changes.
        if (remaining > 1) {
            --remaining;
            return TRUE;
        }
        remaining = 0;
    }
    if (index >= length) {
        return noNext();
    }
    int32_t u = array[index++];
    if (u <= MAX_UNCHANGED) {
        // Combine adjacent unchanged ranges.
        changed = FALSE;
        oldLength_ = u + 1;
        while (index < length && (u = array[index]) <= MAX_UNCHANGED) {
            ++index;
            oldLength_ += u + 1;
        }
        newLength_ = oldLength_;
        if (onlyChanges) {
            updateNextIndexes();
            if (index >= length) {
                return noNext();
            }
            // already fetched u > MAX_UNCHANGED at index
            ++index;
        } else {
            return TRUE;
        }
    }
    changed = TRUE;
    if (u <= MAX_SHORT_CHANGE) {
        int32_t oldLen = u >> 12;
        int32_t newLen = (u >> 9) & MAX_SHORT_CHANGE_NEW_LENGTH;
        int32_t num = (u & SHORT_CHANGE_NUM_MASK) + 1;
        if (coarse) {
            oldLength_ = num * oldLen;
            newLength_ = num * newLen;
        } else {
            // Split a sequence of changes that was compressed into one unit.
            oldLength_ = oldLen;
            newLength_ = newLen;
            if (num > 1) {
                remaining = num;  // This is the first of two or more changes.
            }
            return TRUE;
        }
    } else {
        U_ASSERT(u <= 0x7fff);
        oldLength_ = readLength((u >> 6) & 0x3f);
        newLength_ = readLength(u & 0x3f);
        if (!coarse) {
            return TRUE;
        }
    }
    // Combine adjacent changes.
    while (index < length && (u = array[index]) > MAX_UNCHANGED) {
        ++index;
        if (u <= MAX_SHORT_CHANGE) {
            int32_t num = (u & SHORT_CHANGE_NUM_MASK) + 1;
            oldLength_ += (u >> 12) * num;
            newLength_ += ((u >> 9) & MAX_SHORT_CHANGE_NEW_LENGTH) * num;
        } else {
            U_ASSERT(u <= 0x7fff);
            oldLength_ += readLength((u >> 6) & 0x3f);
            newLength_ += readLength(u & 0x3f);
        }
    }
    return TRUE;
}

UBool Edits::Iterator::previous(UErrorCode &errorCode) {
    // Backward iteration: Pre-decrement-read array units to assemble a new span,
    // then update the string indexes to the start of that span.
    // Leaves the array index on the head unit of that span.
    if (U_FAILURE(errorCode)) { return FALSE; }
    // We have an errorCode in case we need to start guarding against integer overflows.
    // It is also convenient for caller loops if we bail out when an error was set elsewhere.
    if (dir >= 0) {
        if (dir > 0) {
            // Turn around from next() to previous().
            // Set the string indexes to the span limit and
            // pre-decrement-read the same span again.
            if (remaining > 0) {
                // Fine-grained iterator:
                // Stay on the current one of a sequence of compressed changes.
                --index;  // previous() rests on the sequence unit.
                dir = -1;
                return TRUE;
            }
            updateNextIndexes();
        }
        dir = -1;
    }
    if (remaining > 0) {
        // Fine-grained iterator: Continue a sequence of compressed changes.
        int32_t u = array[index];
        U_ASSERT(MAX_UNCHANGED < u && u <= MAX_SHORT_CHANGE);
        if (remaining <= (u & SHORT_CHANGE_NUM_MASK)) {
            ++remaining;
            updatePreviousIndexes();
            return TRUE;
        }
        remaining = 0;
    }
    if (index <= 0) {
        return noNext();
    }
    int32_t u = array[--index];
    if (u <= MAX_UNCHANGED) {
        // Combine adjacent unchanged ranges.
        changed = FALSE;
        oldLength_ = u + 1;
        while (index > 0 && (u = array[index - 1]) <= MAX_UNCHANGED) {
            --index;
            oldLength_ += u + 1;
        }
        newLength_ = oldLength_;
        // No need to handle onlyChanges as long as previous() is called only from findIndex().
        updatePreviousIndexes();
        return TRUE;
    }
    changed = TRUE;
    if (u <= MAX_SHORT_CHANGE) {
        int32_t oldLen = u >> 12;
        int32_t newLen = (u >> 9) & MAX_SHORT_CHANGE_NEW_LENGTH;
        int32_t num = (u & SHORT_CHANGE_NUM_MASK) + 1;
        if (coarse) {
            oldLength_ = num * oldLen;
            newLength_ = num * newLen;
        } else {
            // Split a sequence of changes that was compressed into one unit.
            oldLength_ = oldLen;
            newLength_ = newLen;
            if (num > 1) {
                remaining = 1;  // This is the last of two or more changes.
            }
            updatePreviousIndexes();
            return TRUE;
        }
    } else {
        if (u <= 0x7fff) {
            // The change is encoded in u alone.
            oldLength_ = readLength((u >> 6) & 0x3f);
            newLength_ = readLength(u & 0x3f);
        } else {
            // Back up to the head of the change, read the lengths,
            // and reset the index to the head again.
            U_ASSERT(index > 0);
            while ((u = array[--index]) > 0x7fff) {}
            U_ASSERT(u > MAX_SHORT_CHANGE);
            int32_t headIndex = index++;
            oldLength_ = readLength((u >> 6) & 0x3f);
            newLength_ = readLength(u & 0x3f);
            index = headIndex;
        }
        if (!coarse) {
            updatePreviousIndexes();
            return TRUE;
        }
    }
    // Combine adjacent changes.
    while (index > 0 && (u = array[index - 1]) > MAX_UNCHANGED) {
        --index;
        if (u <= MAX_SHORT_CHANGE) {
            int32_t num = (u & SHORT_CHANGE_NUM_MASK) + 1;
            oldLength_ += (u >> 12) * num;
            newLength_ += ((u >> 9) & MAX_SHORT_CHANGE_NEW_LENGTH) * num;
        } else if (u <= 0x7fff) {
            // Read the lengths, and reset the index to the head again.
            int32_t headIndex = index++;
            oldLength_ += readLength((u >> 6) & 0x3f);
            newLength_ += readLength(u & 0x3f);
            index = headIndex;
        }
    }
    updatePreviousIndexes();
    return TRUE;
}

int32_t Edits::Iterator::findIndex(int32_t i, UBool findSource, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode) || i < 0) { return -1; }
    int32_t spanStart, spanLength;
    if (findSource) {  // find source index
        spanStart = srcIndex;
        spanLength = oldLength_;
    } else {  // find destination index
        spanStart = destIndex;
        spanLength = newLength_;
    }
    if (i < spanStart) {
        if (i >= (spanStart / 2)) {
            // Search backwards.
            for (;;) {
                UBool hasPrevious = previous(errorCode);
                U_ASSERT(hasPrevious);  // because i>=0 and the first span starts at 0
                (void)hasPrevious;  // avoid unused-variable warning
                spanStart = findSource ? srcIndex : destIndex;
                if (i >= spanStart) {
                    // The index is in the current span.
                    return 0;
                }
                if (remaining > 0) {
                    // Is the index in one of the remaining compressed edits?
                    // spanStart is the start of the current span, first of the remaining ones.
                    spanLength = findSource ? oldLength_ : newLength_;
                    int32_t u = array[index];
                    U_ASSERT(MAX_UNCHANGED < u && u <= MAX_SHORT_CHANGE);
                    int32_t num = (u & SHORT_CHANGE_NUM_MASK) + 1 - remaining;
                    int32_t len = num * spanLength;
                    if (i >= (spanStart - len)) {
                        int32_t n = ((spanStart - i - 1) / spanLength) + 1;
                        // 1 <= n <= num
                        srcIndex -= n * oldLength_;
                        replIndex -= n * newLength_;
                        destIndex -= n * newLength_;
                        remaining += n;
                        return 0;
                    }
                    // Skip all of these edits at once.
                    srcIndex -= num * oldLength_;
                    replIndex -= num * newLength_;
                    destIndex -= num * newLength_;
                    remaining = 0;
                }
            }
        }
        // Reset the iterator to the start.
        dir = 0;
        index = remaining = oldLength_ = newLength_ = srcIndex = replIndex = destIndex = 0;
    } else if (i < (spanStart + spanLength)) {
        // The index is in the current span.
        return 0;
    }
    while (next(FALSE, errorCode)) {
        if (findSource) {
            spanStart = srcIndex;
            spanLength = oldLength_;
        } else {
            spanStart = destIndex;
            spanLength = newLength_;
        }
        if (i < (spanStart + spanLength)) {
            // The index is in the current span.
            return 0;
        }
        if (remaining > 1) {
            // Is the index in one of the remaining compressed edits?
            // spanStart is the start of the current span, first of the remaining ones.
            int32_t len = remaining * spanLength;
            if (i < (spanStart + len)) {
                int32_t n = (i - spanStart) / spanLength;  // 1 <= n <= remaining - 1
                srcIndex += n * oldLength_;
                replIndex += n * newLength_;
                destIndex += n * newLength_;
                remaining -= n;
                return 0;
            }
            // Make next() skip all of these edits at once.
            oldLength_ *= remaining;
            newLength_ *= remaining;
            remaining = 0;
        }
    }
    return 1;
}

int32_t Edits::Iterator::destinationIndexFromSourceIndex(int32_t i, UErrorCode &errorCode) {
    int32_t where = findIndex(i, TRUE, errorCode);
    if (where < 0) {
        // Error or before the string.
        return 0;
    }
    if (where > 0 || i == srcIndex) {
        // At or after string length, or at start of the found span.
        return destIndex;
    }
    if (changed) {
        // In a change span, map to its end.
        return destIndex + newLength_;
    } else {
        // In an unchanged span, offset 1:1 within it.
        return destIndex + (i - srcIndex);
    }
}

int32_t Edits::Iterator::sourceIndexFromDestinationIndex(int32_t i, UErrorCode &errorCode) {
    int32_t where = findIndex(i, FALSE, errorCode);
    if (where < 0) {
        // Error or before the string.
        return 0;
    }
    if (where > 0 || i == destIndex) {
        // At or after string length, or at start of the found span.
        return srcIndex;
    }
    if (changed) {
        // In a change span, map to its end.
        return srcIndex + oldLength_;
    } else {
        // In an unchanged span, offset within it.
        return srcIndex + (i - destIndex);
    }
}

UnicodeString& Edits::Iterator::toString(UnicodeString& sb) const {
    sb.append(u"{ src[", -1);
    ICU_Utility::appendNumber(sb, srcIndex);
    sb.append(u"..", -1);
    ICU_Utility::appendNumber(sb, srcIndex + oldLength_);
    if (changed) {
        sb.append(u"] ⇝ dest[", -1);
    } else {
        sb.append(u"] ≡ dest[", -1);
    }
    ICU_Utility::appendNumber(sb, destIndex);
    sb.append(u"..", -1);
    ICU_Utility::appendNumber(sb, destIndex + newLength_);
    if (changed) {
        sb.append(u"], repl[", -1);
        ICU_Utility::appendNumber(sb, replIndex);
        sb.append(u"..", -1);
        ICU_Utility::appendNumber(sb, replIndex + newLength_);
        sb.append(u"] }", -1);
    } else {
        sb.append(u"] (no-change) }", -1);
    }
    return sb;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2009-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  errorcode.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2009mar10
*   created by: Markus W. Scherer
*/


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2009-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  errorcode.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2009mar10
*   created by: Markus W. Scherer
*/


/**
 * \file 
 * \brief C++ API: ErrorCode class intended to make it easier to use
 *                 ICU C and C++ APIs from C++ user code.
 */



#if U_SHOW_CPLUSPLUS_API



U_NAMESPACE_BEGIN

/**
 * Wrapper class for UErrorCode, with conversion operators for direct use
 * in ICU C and C++ APIs.
 * Intended to be used as a base class, where a subclass overrides
 * the handleFailure() function so that it throws an exception,
 * does an assert(), logs an error, etc.
 * This is not an abstract base class. This class can be used and instantiated
 * by itself, although it will be more useful when subclassed.
 *
 * Features:
 * - The constructor initializes the internal UErrorCode to U_ZERO_ERROR,
 *   removing one common source of errors.
 * - Same use in C APIs taking a UErrorCode * (pointer)
 *   and C++ taking UErrorCode & (reference) via conversion operators.
 * - Possible automatic checking for success when it goes out of scope.
 *
 * Note: For automatic checking for success in the destructor, a subclass
 * must implement such logic in its own destructor because the base class
 * destructor cannot call a subclass function (like handleFailure()).
 * The ErrorCode base class destructor does nothing.
 *
 * Note also: While it is possible for a destructor to throw an exception,
 * it is generally unsafe to do so. This means that in a subclass the destructor
 * and the handleFailure() function may need to take different actions.
 *
 * Sample code:
 * \code
 *   class IcuErrorCode: public icu::ErrorCode {
 *   public:
 *     virtual ~IcuErrorCode() {  // should be defined in .cpp as "key function"
 *       // Safe because our handleFailure() does not throw exceptions.
 *       if(isFailure()) { handleFailure(); }
 *     }
 *   protected:
 *     virtual void handleFailure() const {
 *       log_failure(u_errorName(errorCode));
 *       exit(errorCode);
 *     }
 *   };
 *   IcuErrorCode error_code;
 *   UConverter *cnv = ucnv_open("Shift-JIS", error_code);
 *   length = ucnv_fromUChars(dest, capacity, src, length, error_code);
 *   ucnv_close(cnv);
 *   // IcuErrorCode destructor checks for success.
 * \endcode
 *
 * @stable ICU 4.2
 */
class U_COMMON_API ErrorCode: public UMemory {
public:
    /**
     * Default constructor. Initializes its UErrorCode to U_ZERO_ERROR.
     * @stable ICU 4.2
     */
    ErrorCode() : errorCode(U_ZERO_ERROR) {}
    /** Destructor, does nothing. See class documentation for details. @stable ICU 4.2 */
    virtual ~ErrorCode();
    /** Conversion operator, returns a reference. @stable ICU 4.2 */
    operator UErrorCode & () { return errorCode; }
    /** Conversion operator, returns a pointer. @stable ICU 4.2 */
    operator UErrorCode * () { return &errorCode; }
    /** Tests for U_SUCCESS(). @stable ICU 4.2 */
    UBool isSuccess() const { return U_SUCCESS(errorCode); }
    /** Tests for U_FAILURE(). @stable ICU 4.2 */
    UBool isFailure() const { return U_FAILURE(errorCode); }
    /** Returns the UErrorCode value. @stable ICU 4.2 */
    UErrorCode get() const { return errorCode; }
    /** Sets the UErrorCode value. @stable ICU 4.2 */
    void set(UErrorCode value) { errorCode=value; }
    /** Returns the UErrorCode value and resets it to U_ZERO_ERROR. @stable ICU 4.2 */
    UErrorCode reset();
    /**
     * Asserts isSuccess().
     * In other words, this method checks for a failure code,
     * and the base class handles it like this:
     * \code
     *   if(isFailure()) { handleFailure(); }
     * \endcode
     * @stable ICU 4.4
     */
    void assertSuccess() const;
    /**
     * Return a string for the UErrorCode value.
     * The string will be the same as the name of the error code constant
     * in the UErrorCode enum.
     * @stable ICU 4.4
     */
    const char* errorName() const;

protected:
    /**
     * Internal UErrorCode, accessible to subclasses.
     * @stable ICU 4.2
     */
    UErrorCode errorCode;
    /**
     * Called by assertSuccess() if isFailure() is true.
     * A subclass should override this function to deal with a failure code:
     * Throw an exception, log an error, terminate the program, or similar.
     * @stable ICU 4.2
     */
    virtual void handleFailure() const {}
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */



U_NAMESPACE_BEGIN

ErrorCode::~ErrorCode() {}

UErrorCode ErrorCode::reset() {
    UErrorCode code = errorCode;
    errorCode = U_ZERO_ERROR;
    return code;
}

void ErrorCode::assertSuccess() const {
    if(isFailure()) {
        handleFailure();
    }
}

const char* ErrorCode::errorName() const {
  return u_errorName(errorCode);
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2014-2015, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/


// #if !UCONFIG_NO_BREAK_ITERATION && !UCONFIG_NO_FILTERED_BREAK_ITERATION

// #include "cmemory.h"

// #include "unicode/filteredbrk.h"
// #include "unicode/ucharstriebuilder.h"
// #include "unicode/ures.h"

// #include "uresimp.h" // ures_getByKeyWithFallback
// #include "ubrkimpl.h" // U_ICUDATA_BRKITR
// #include "uvector.h"
// #include "cmemory.h"

// U_NAMESPACE_BEGIN

// #ifndef FB_DEBUG
// #define FB_DEBUG 0
// #endif

// #if FB_DEBUG
// #include <stdio.h>
// static void _fb_trace(const char *m, const UnicodeString *s, UBool b, int32_t d, const char *f, int l) {
//   char buf[2048];
//   if(s) {
//     s->extract(0,s->length(),buf,2048);
//   } else {
//     strcpy(buf,"NULL");
//   }
//   fprintf(stderr,"%s:%d: %s. s='%s'(%p), b=%c, d=%d\n",
//           f, l, m, buf, (const void*)s, b?'T':'F',(int)d);
// }

// #define FB_TRACE(m,s,b,d) _fb_trace(m,s,b,d,__FILE__,__LINE__)
// #else
// #define FB_TRACE(m,s,b,d)
// #endif

// /**
//  * Used with sortedInsert()
//  */
// static int8_t U_CALLCONV compareUnicodeString(UElement t1, UElement t2) {
//     const UnicodeString &a = *(const UnicodeString*)t1.pointer;
//     const UnicodeString &b = *(const UnicodeString*)t2.pointer;
//     return a.compare(b);
// }

// /**
//  * A UVector which implements a set of strings.
//  */
// class U_COMMON_API UStringSet : public UVector {
//  public:
//   UStringSet(UErrorCode &status) : UVector(uprv_deleteUObject,
//                                            uhash_compareUnicodeString,
//                                            1,
//                                            status) {}
//   virtual ~UStringSet();
//   /**
//    * Is this UnicodeSet contained?
//    */
//   inline UBool contains(const UnicodeString& s) {
//     return contains((void*) &s);
//   }
//   using UVector::contains;
//   /**
//    * Return the ith UnicodeString alias
//    */
//   inline const UnicodeString* getStringAt(int32_t i) const {
//     return (const UnicodeString*)elementAt(i);
//   }
//   /**
//    * Adopt the UnicodeString if not already contained.
//    * Caller no longer owns the pointer in any case.
//    * @return true if adopted successfully, false otherwise (error, or else duplicate)
//    */
//   inline UBool adopt(UnicodeString *str, UErrorCode &status) {
//     if(U_FAILURE(status) || contains(*str)) {
//       delete str;
//       return false;
//     } else {
//       sortedInsert(str, compareUnicodeString, status);
//       if(U_FAILURE(status)) {
//         delete str;
//         return false;
//       }
//       return true;
//     }
//   }
//   /**
//    * Add by value.
//    * @return true if successfully adopted.
//    */
//   inline UBool add(const UnicodeString& str, UErrorCode &status) {
//     if(U_FAILURE(status)) return false;
//     UnicodeString *t = new UnicodeString(str);
//     if(t==NULL) {
//       status = U_MEMORY_ALLOCATION_ERROR; return false;
//     }
//     return adopt(t, status);
//   }
//   /**
//    * Remove this string.
//    * @return true if successfully removed, false otherwise (error, or else it wasn't there)
//    */
//   inline UBool remove(const UnicodeString &s, UErrorCode &status) {
//     if(U_FAILURE(status)) return false;
//     return removeElement((void*) &s);
//   }
// };

// /**
//  * Virtual, won't be inlined
//  */
// UStringSet::~UStringSet() {}

// /* ----------------------------------------------------------- */


// /* Filtered Break constants */
// static const int32_t kPARTIAL = (1<<0); //< partial - need to run through forward trie
// static const int32_t kMATCH   = (1<<1); //< exact match - skip this one.
// static const int32_t kSuppressInReverse = (1<<0);
// static const int32_t kAddToForward = (1<<1);
// static const UChar   kFULLSTOP = 0x002E; // '.'

// /**
//  * Shared data for SimpleFilteredSentenceBreakIterator
//  */
// class SimpleFilteredSentenceBreakData : public UMemory {
// public:
//   SimpleFilteredSentenceBreakData(UCharsTrie *forwards, UCharsTrie *backwards )
//       : fForwardsPartialTrie(forwards), fBackwardsTrie(backwards), refcount(1) { }
//   SimpleFilteredSentenceBreakData *incr() { refcount++;  return this; }
//   SimpleFilteredSentenceBreakData *decr() { if((--refcount) <= 0) delete this; return 0; }
//   virtual ~SimpleFilteredSentenceBreakData();

//   LocalPointer<UCharsTrie>    fForwardsPartialTrie; //  Has ".a" for "a.M."
//   LocalPointer<UCharsTrie>    fBackwardsTrie; //  i.e. ".srM" for Mrs.
//   int32_t                     refcount;
// };

// SimpleFilteredSentenceBreakData::~SimpleFilteredSentenceBreakData() {}

// /**
//  * Concrete implementation
//  */
// class SimpleFilteredSentenceBreakIterator : public BreakIterator {
// public:
//   SimpleFilteredSentenceBreakIterator(BreakIterator *adopt, UCharsTrie *forwards, UCharsTrie *backwards, UErrorCode &status);
//   SimpleFilteredSentenceBreakIterator(const SimpleFilteredSentenceBreakIterator& other);
//   virtual ~SimpleFilteredSentenceBreakIterator();
// private:
//   SimpleFilteredSentenceBreakData *fData;
//   LocalPointer<BreakIterator> fDelegate;
//   LocalUTextPointer           fText;

//   /* -- subclass interface -- */
// public:
//   /* -- cloning and other subclass stuff -- */
//   virtual BreakIterator *  createBufferClone(void * /*stackBuffer*/,
//                                              int32_t &/*BufferSize*/,
//                                              UErrorCode &status) {
//     // for now - always deep clone
//     status = U_SAFECLONE_ALLOCATED_WARNING;
//     return clone();
//   }
//   virtual SimpleFilteredSentenceBreakIterator* clone() const { return new SimpleFilteredSentenceBreakIterator(*this); }
//   virtual UClassID getDynamicClassID(void) const { return NULL; }
//   virtual UBool operator==(const BreakIterator& o) const { if(this==&o) return true; return false; }

//   /* -- text modifying -- */
//   virtual void setText(UText *text, UErrorCode &status) { fDelegate->setText(text,status); }
//   virtual BreakIterator &refreshInputText(UText *input, UErrorCode &status) { fDelegate->refreshInputText(input,status); return *this; }
//   virtual void adoptText(CharacterIterator* it) { fDelegate->adoptText(it); }
//   virtual void setText(const UnicodeString &text) { fDelegate->setText(text); }

//   /* -- other functions that are just delegated -- */
//   virtual UText *getUText(UText *fillIn, UErrorCode &status) const { return fDelegate->getUText(fillIn,status); }
//   virtual CharacterIterator& getText(void) const { return fDelegate->getText(); }

//   /* -- ITERATION -- */
//   virtual int32_t first(void);
//   virtual int32_t preceding(int32_t offset);
//   virtual int32_t previous(void);
//   virtual UBool isBoundary(int32_t offset);
//   virtual int32_t current(void) const { return fDelegate->current(); } // we keep the delegate current, so this should be correct.

//   virtual int32_t next(void);

//   virtual int32_t next(int32_t n);
//   virtual int32_t following(int32_t offset);
//   virtual int32_t last(void);

// private:
//     /**
//      * Given that the fDelegate has already given its "initial" answer,
//      * find the NEXT actual (non-excepted) break.
//      * @param n initial position from delegate
//      * @return new break position or UBRK_DONE
//      */
//     int32_t internalNext(int32_t n);
//     /**
//      * Given that the fDelegate has already given its "initial" answer,
//      * find the PREV actual (non-excepted) break.
//      * @param n initial position from delegate
//      * @return new break position or UBRK_DONE
//      */
//     int32_t internalPrev(int32_t n);
//     /**
//      * set up the UText with the value of the fDelegate.
//      * Call this before calling breakExceptionAt.
//      * May be able to avoid excess calls
//      */
//     void resetState(UErrorCode &status);
//     /**
//      * Is there a match  (exception) at this spot?
//      */
//     enum EFBMatchResult { kNoExceptionHere, kExceptionHere };
//     /**
//      * Determine if there is an exception at this spot
//      * @param n spot to check
//      * @return kNoExceptionHere or kExceptionHere
//      **/
//     enum EFBMatchResult breakExceptionAt(int32_t n);
// };

// SimpleFilteredSentenceBreakIterator::SimpleFilteredSentenceBreakIterator(const SimpleFilteredSentenceBreakIterator& other)
//   : BreakIterator(other), fData(other.fData->incr()), fDelegate(other.fDelegate->clone())
// {
// }


// SimpleFilteredSentenceBreakIterator::SimpleFilteredSentenceBreakIterator(BreakIterator *adopt, UCharsTrie *forwards, UCharsTrie *backwards, UErrorCode &status) :
//   BreakIterator(adopt->getLocale(ULOC_VALID_LOCALE,status),adopt->getLocale(ULOC_ACTUAL_LOCALE,status)),
//   fData(new SimpleFilteredSentenceBreakData(forwards, backwards)),
//   fDelegate(adopt)
// {
//   // all set..
// }

// SimpleFilteredSentenceBreakIterator::~SimpleFilteredSentenceBreakIterator() {
//     fData = fData->decr();
// }

// void SimpleFilteredSentenceBreakIterator::resetState(UErrorCode &status) {
//   fText.adoptInstead(fDelegate->getUText(fText.orphan(), status));
// }

// SimpleFilteredSentenceBreakIterator::EFBMatchResult
// SimpleFilteredSentenceBreakIterator::breakExceptionAt(int32_t n) {
//     int64_t bestPosn = -1;
//     int32_t bestValue = -1;
//     // loops while 'n' points to an exception.
//     utext_setNativeIndex(fText.getAlias(), n); // from n..
//     fData->fBackwardsTrie->reset();
//     UChar32 uch;

//     //if(debug2) u_printf(" n@ %d\n", n);
//     // Assume a space is following the '.'  (so we handle the case:  "Mr. /Brown")
//     if((uch=utext_previous32(fText.getAlias()))==(UChar32)0x0020) {  // TODO: skip a class of chars here??
//       // TODO only do this the 1st time?
//       //if(debug2) u_printf("skipping prev: |%C| \n", (UChar)uch);
//     } else {
//       //if(debug2) u_printf("not skipping prev: |%C| \n", (UChar)uch);
//       uch = utext_next32(fText.getAlias());
//       //if(debug2) u_printf(" -> : |%C| \n", (UChar)uch);
//     }

//     UStringTrieResult r = USTRINGTRIE_INTERMEDIATE_VALUE;

//     while((uch=utext_previous32(fText.getAlias()))!=U_SENTINEL  &&   // more to consume backwards and..
//           USTRINGTRIE_HAS_NEXT(r=fData->fBackwardsTrie->nextForCodePoint(uch))) {// more in the trie
//       if(USTRINGTRIE_HAS_VALUE(r)) { // remember the best match so far
//         bestPosn = utext_getNativeIndex(fText.getAlias());
//         bestValue = fData->fBackwardsTrie->getValue();
//       }
//       //if(debug2) u_printf("rev< /%C/ cont?%d @%d\n", (UChar)uch, r, utext_getNativeIndex(fText.getAlias()));
//     }

//     if(USTRINGTRIE_MATCHES(r)) { // exact match?
//       //if(debug2) u_printf("rev<?/%C/?end of seq.. r=%d, bestPosn=%d, bestValue=%d\n", (UChar)uch, r, bestPosn, bestValue);
//       bestValue = fData->fBackwardsTrie->getValue();
//       bestPosn = utext_getNativeIndex(fText.getAlias());
//       //if(debug2) u_printf("rev<+/%C/+end of seq.. r=%d, bestPosn=%d, bestValue=%d\n", (UChar)uch, r, bestPosn, bestValue);
//     }

//     if(bestPosn>=0) {
//       //if(debug2) u_printf("rev< /%C/ end of seq.. r=%d, bestPosn=%d, bestValue=%d\n", (UChar)uch, r, bestPosn, bestValue);

//       //if(USTRINGTRIE_MATCHES(r)) {  // matched - so, now what?
//       //int32_t bestValue = fBackwardsTrie->getValue();
//       ////if(debug2) u_printf("rev< /%C/ matched, skip..%d  bestValue=%d\n", (UChar)uch, r, bestValue);

//       if(bestValue == kMATCH) { // exact match!
//         //if(debug2) u_printf(" exact backward match\n");
//         return kExceptionHere; // See if the next is another exception.
//       } else if(bestValue == kPARTIAL
//                 && fData->fForwardsPartialTrie.isValid()) { // make sure there's a forward trie
//         //if(debug2) u_printf(" partial backward match\n");
//         // We matched the "Ph." in "Ph.D." - now we need to run everything through the forwards trie
//         // to see if it matches something going forward.
//         fData->fForwardsPartialTrie->reset();
//         UStringTrieResult rfwd = USTRINGTRIE_INTERMEDIATE_VALUE;
//         utext_setNativeIndex(fText.getAlias(), bestPosn); // hope that's close ..
//         //if(debug2) u_printf("Retrying at %d\n", bestPosn);
//         while((uch=utext_next32(fText.getAlias()))!=U_SENTINEL &&
//               USTRINGTRIE_HAS_NEXT(rfwd=fData->fForwardsPartialTrie->nextForCodePoint(uch))) {
//           //if(debug2) u_printf("fwd> /%C/ cont?%d @%d\n", (UChar)uch, rfwd, utext_getNativeIndex(fText.getAlias()));
//         }
//         if(USTRINGTRIE_MATCHES(rfwd)) {
//           //if(debug2) u_printf("fwd> /%C/ == forward match!\n", (UChar)uch);
//           // only full matches here, nothing to check
//           // skip the next:
//             return kExceptionHere;
//         } else {
//           //if(debug2) u_printf("fwd> /%C/ no match.\n", (UChar)uch);
//           // no match (no exception) -return the 'underlying' break
//           return kNoExceptionHere;
//         }
//       } else {
//         return kNoExceptionHere; // internal error and/or no forwards trie
//       }
//     } else {
//       //if(debug2) u_printf("rev< /%C/ .. no match..%d\n", (UChar)uch, r);  // no best match
//       return kNoExceptionHere; // No match - so exit. Not an exception.
//     }
// }

// // the workhorse single next.
// int32_t
// SimpleFilteredSentenceBreakIterator::internalNext(int32_t n) {
//   if(n == UBRK_DONE || // at end  or
//     fData->fBackwardsTrie.isNull()) { // .. no backwards table loaded == no exceptions
//       return n;
//   }
//   // OK, do we need to break here?
//   UErrorCode status = U_ZERO_ERROR;
//   // refresh text
//   resetState(status);
//   if(U_FAILURE(status)) return UBRK_DONE; // bail out
//   int64_t utextLen = utext_nativeLength(fText.getAlias());

//   //if(debug2) u_printf("str, native len=%d\n", utext_nativeLength(fText.getAlias()));
//   while (n != UBRK_DONE && n != utextLen) { // outer loop runs once per underlying break (from fDelegate).
//     SimpleFilteredSentenceBreakIterator::EFBMatchResult m = breakExceptionAt(n);

//     switch(m) {
//     case kExceptionHere:
//       n = fDelegate->next(); // skip this one. Find the next lowerlevel break.
//       continue;

//     default:
//     case kNoExceptionHere:
//       return n;
//     }
//   }
//   return n;
// }

// int32_t
// SimpleFilteredSentenceBreakIterator::internalPrev(int32_t n) {
//   if(n == 0 || n == UBRK_DONE || // at end  or
//     fData->fBackwardsTrie.isNull()) { // .. no backwards table loaded == no exceptions
//       return n;
//   }
//   // OK, do we need to break here?
//   UErrorCode status = U_ZERO_ERROR;
//   // refresh text
//   resetState(status);
//   if(U_FAILURE(status)) return UBRK_DONE; // bail out

//   //if(debug2) u_printf("str, native len=%d\n", utext_nativeLength(fText.getAlias()));
//   while (n != UBRK_DONE && n != 0) { // outer loop runs once per underlying break (from fDelegate).
//     SimpleFilteredSentenceBreakIterator::EFBMatchResult m = breakExceptionAt(n);

//     switch(m) {
//     case kExceptionHere:
//       n = fDelegate->previous(); // skip this one. Find the next lowerlevel break.
//       continue;

//     default:
//     case kNoExceptionHere:
//       return n;
//     }
//   }
//   return n;
// }


// int32_t
// SimpleFilteredSentenceBreakIterator::next() {
//   return internalNext(fDelegate->next());
// }

// int32_t
// SimpleFilteredSentenceBreakIterator::first(void) {
//   // Don't suppress a break opportunity at the beginning of text.
//   return fDelegate->first();
// }

// int32_t
// SimpleFilteredSentenceBreakIterator::preceding(int32_t offset) {
//   return internalPrev(fDelegate->preceding(offset));
// }

// int32_t
// SimpleFilteredSentenceBreakIterator::previous(void) {
//   return internalPrev(fDelegate->previous());
// }

// UBool SimpleFilteredSentenceBreakIterator::isBoundary(int32_t offset) {
//   if (!fDelegate->isBoundary(offset)) return false; // no break to suppress

//   if (fData->fBackwardsTrie.isNull()) return true; // no data = no suppressions

//   UErrorCode status = U_ZERO_ERROR;
//   resetState(status);

//   SimpleFilteredSentenceBreakIterator::EFBMatchResult m = breakExceptionAt(offset);

//   switch(m) {
//   case kExceptionHere:
//     return false;
//   default:
//   case kNoExceptionHere:
//     return true;
//   }
// }

// int32_t
// SimpleFilteredSentenceBreakIterator::next(int32_t offset) {
//   return internalNext(fDelegate->next(offset));
// }

// int32_t
// SimpleFilteredSentenceBreakIterator::following(int32_t offset) {
//   return internalNext(fDelegate->following(offset));
// }

// int32_t
// SimpleFilteredSentenceBreakIterator::last(void) {
//   // Don't suppress a break opportunity at the end of text.
//   return fDelegate->last();
// }


// /**
//  * Concrete implementation of builder class.
//  */
// class U_COMMON_API SimpleFilteredBreakIteratorBuilder : public FilteredBreakIteratorBuilder {
// public:
//   virtual ~SimpleFilteredBreakIteratorBuilder();
//   SimpleFilteredBreakIteratorBuilder(const Locale &fromLocale, UErrorCode &status);
//   SimpleFilteredBreakIteratorBuilder(UErrorCode &status);
//   virtual UBool suppressBreakAfter(const UnicodeString& exception, UErrorCode& status);
//   virtual UBool unsuppressBreakAfter(const UnicodeString& exception, UErrorCode& status);
//   virtual BreakIterator *build(BreakIterator* adoptBreakIterator, UErrorCode& status);
// private:
//   UStringSet fSet;
// };

// SimpleFilteredBreakIteratorBuilder::~SimpleFilteredBreakIteratorBuilder()
// {
// }

// SimpleFilteredBreakIteratorBuilder::SimpleFilteredBreakIteratorBuilder(UErrorCode &status)
//   : fSet(status)
// {
// }

// SimpleFilteredBreakIteratorBuilder::SimpleFilteredBreakIteratorBuilder(const Locale &fromLocale, UErrorCode &status)
//   : fSet(status)
// {
//   if(U_SUCCESS(status)) {
//     UErrorCode subStatus = U_ZERO_ERROR;
//     LocalUResourceBundlePointer b(ures_open(U_ICUDATA_BRKITR, fromLocale.getBaseName(), &subStatus));
//     if (U_FAILURE(subStatus) || (subStatus == U_USING_DEFAULT_WARNING) ) {
//       status = subStatus; // copy the failing status
// #if FB_DEBUG
//       fprintf(stderr, "open BUNDLE %s : %s, %s\n", fromLocale.getBaseName(), "[exit]", u_errorName(status));
// #endif
//       return;  // leaves the builder empty, if you try to use it.
//     }
//     LocalUResourceBundlePointer exceptions(ures_getByKeyWithFallback(b.getAlias(), "exceptions", NULL, &subStatus));
//     if (U_FAILURE(subStatus) || (subStatus == U_USING_DEFAULT_WARNING) ) {
//       status = subStatus; // copy the failing status
// #if FB_DEBUG
//       fprintf(stderr, "open EXCEPTIONS %s : %s, %s\n", fromLocale.getBaseName(), "[exit]", u_errorName(status));
// #endif
//       return;  // leaves the builder empty, if you try to use it.
//     }
//     LocalUResourceBundlePointer breaks(ures_getByKeyWithFallback(exceptions.getAlias(), "SentenceBreak", NULL, &subStatus));

// #if FB_DEBUG
//     {
//       UErrorCode subsub = subStatus;
//       fprintf(stderr, "open SentenceBreak %s => %s, %s\n", fromLocale.getBaseName(), ures_getLocale(breaks.getAlias(), &subsub), u_errorName(subStatus));
//     }
// #endif

//     if (U_FAILURE(subStatus) || (subStatus == U_USING_DEFAULT_WARNING) ) {
//       status = subStatus; // copy the failing status
// #if FB_DEBUG
//       fprintf(stderr, "open %s : %s, %s\n", fromLocale.getBaseName(), "[exit]", u_errorName(status));
// #endif
//       return;  // leaves the builder empty, if you try to use it.
//     }

//     LocalUResourceBundlePointer strs;
//     subStatus = status; // Pick up inherited warning status now
//     do {
//       strs.adoptInstead(ures_getNextResource(breaks.getAlias(), strs.orphan(), &subStatus));
//       if(strs.isValid() && U_SUCCESS(subStatus)) {
//         UnicodeString str(ures_getUnicodeString(strs.getAlias(), &status));
//         suppressBreakAfter(str, status); // load the string
//       }
//     } while (strs.isValid() && U_SUCCESS(subStatus));
//     if(U_FAILURE(subStatus)&&subStatus!=U_INDEX_OUTOFBOUNDS_ERROR&&U_SUCCESS(status)) {
//       status = subStatus;
//     }
//   }
// }

// UBool
// SimpleFilteredBreakIteratorBuilder::suppressBreakAfter(const UnicodeString& exception, UErrorCode& status)
// {
//   UBool r = fSet.add(exception, status);
//   FB_TRACE("suppressBreakAfter",&exception,r,0);
//   return r;
// }

// UBool
// SimpleFilteredBreakIteratorBuilder::unsuppressBreakAfter(const UnicodeString& exception, UErrorCode& status)
// {
//   UBool r = fSet.remove(exception, status);
//   FB_TRACE("unsuppressBreakAfter",&exception,r,0);
//   return r;
// }

// /**
//  * Jitterbug 2974: MSVC has a bug whereby new X[0] behaves badly.
//  * Work around this.
//  *
//  * Note: "new UnicodeString[subCount]" ends up calling global operator new
//  * on MSVC2012 for some reason.
//  */
// static inline UnicodeString* newUnicodeStringArray(size_t count) {
//     return new UnicodeString[count ? count : 1];
// }

// BreakIterator *
// SimpleFilteredBreakIteratorBuilder::build(BreakIterator* adoptBreakIterator, UErrorCode& status) {
//   LocalPointer<BreakIterator> adopt(adoptBreakIterator);

//   LocalPointer<UCharsTrieBuilder> builder(new UCharsTrieBuilder(status), status);
//   LocalPointer<UCharsTrieBuilder> builder2(new UCharsTrieBuilder(status), status);
//   if(U_FAILURE(status)) {
//     return NULL;
//   }

//   int32_t revCount = 0;
//   int32_t fwdCount = 0;

//   int32_t subCount = fSet.size();

//   UnicodeString *ustrs_ptr = newUnicodeStringArray(subCount);

//   LocalArray<UnicodeString> ustrs(ustrs_ptr);

//   LocalMemory<int> partials;
//   partials.allocateInsteadAndReset(subCount);

//   LocalPointer<UCharsTrie>    backwardsTrie; //  i.e. ".srM" for Mrs.
//   LocalPointer<UCharsTrie>    forwardsPartialTrie; //  Has ".a" for "a.M."

//   int n=0;
//   for ( int32_t i = 0;
//         i<fSet.size();
//         i++) {
//     const UnicodeString *abbr = fSet.getStringAt(i);
//     if(abbr) {
//       FB_TRACE("build",abbr,TRUE,i);
//       ustrs[n] = *abbr; // copy by value
//       FB_TRACE("ustrs[n]",&ustrs[n],TRUE,i);
//     } else {
//       FB_TRACE("build",abbr,FALSE,i);
//       status = U_MEMORY_ALLOCATION_ERROR;
//       return NULL;
//     }
//     partials[n] = 0; // default: not partial
//     n++;
//   }
//   // first pass - find partials.
//   for(int i=0;i<subCount;i++) {
//     int nn = ustrs[i].indexOf(kFULLSTOP); // TODO: non-'.' abbreviations
//     if(nn>-1 && (nn+1)!=ustrs[i].length()) {
//       FB_TRACE("partial",&ustrs[i],FALSE,i);
//       // is partial.
//       // is it unique?
//       int sameAs = -1;
//       for(int j=0;j<subCount;j++) {
//         if(j==i) continue;
//         if(ustrs[i].compare(0,nn+1,ustrs[j],0,nn+1)==0) {
//           FB_TRACE("prefix",&ustrs[j],FALSE,nn+1);
//           //UBool otherIsPartial = ((nn+1)!=ustrs[j].length());  // true if ustrs[j] doesn't end at nn
//           if(partials[j]==0) { // hasn't been processed yet
//             partials[j] = kSuppressInReverse | kAddToForward;
//             FB_TRACE("suppressing",&ustrs[j],FALSE,j);
//           } else if(partials[j] & kSuppressInReverse) {
//             sameAs = j; // the other entry is already in the reverse table.
//           }
//         }
//       }
//       FB_TRACE("for partial same-",&ustrs[i],FALSE,sameAs);
//       FB_TRACE(" == partial #",&ustrs[i],FALSE,partials[i]);
//       UnicodeString prefix(ustrs[i], 0, nn+1);
//       if(sameAs == -1 && partials[i] == 0) {
//         // first one - add the prefix to the reverse table.
//         prefix.reverse();
//         builder->add(prefix, kPARTIAL, status);
//         revCount++;
//         FB_TRACE("Added partial",&prefix,FALSE, i);
//         FB_TRACE(u_errorName(status),&ustrs[i],FALSE,i);
//         partials[i] = kSuppressInReverse | kAddToForward;
//       } else {
//         FB_TRACE("NOT adding partial",&prefix,FALSE, i);
//         FB_TRACE(u_errorName(status),&ustrs[i],FALSE,i);
//       }
//     }
//   }
//   for(int i=0;i<subCount;i++) {
//     if(partials[i]==0) {
//       ustrs[i].reverse();
//       builder->add(ustrs[i], kMATCH, status);
//       revCount++;
//       FB_TRACE(u_errorName(status), &ustrs[i], FALSE, i);
//     } else {
//       FB_TRACE("Adding fwd",&ustrs[i], FALSE, i);

//       // an optimization would be to only add the portion after the '.'
//       // for example, for "Ph.D." we store ".hP" in the reverse table. We could just store "D." in the forward,
//       // instead of "Ph.D." since we already know the "Ph." part is a match.
//       // would need the trie to be able to hold 0-length strings, though.
//       builder2->add(ustrs[i], kMATCH, status); // forward
//       fwdCount++;
//       //ustrs[i].reverse();
//       ////if(debug2) u_printf("SUPPRESS- not Added(%d):  /%S/ status=%s\n",partials[i], ustrs[i].getTerminatedBuffer(), u_errorName(status));
//     }
//   }
//   FB_TRACE("AbbrCount",NULL,FALSE, subCount);

//   if(revCount>0) {
//     backwardsTrie.adoptInstead(builder->build(USTRINGTRIE_BUILD_FAST, status));
//     if(U_FAILURE(status)) {
//       FB_TRACE(u_errorName(status),NULL,FALSE, -1);
//       return NULL;
//     }
//   }

//   if(fwdCount>0) {
//     forwardsPartialTrie.adoptInstead(builder2->build(USTRINGTRIE_BUILD_FAST, status));
//     if(U_FAILURE(status)) {
//       FB_TRACE(u_errorName(status),NULL,FALSE, -1);
//       return NULL;
//     }
//   }

//   return new SimpleFilteredSentenceBreakIterator(adopt.orphan(), forwardsPartialTrie.orphan(), backwardsTrie.orphan(), status);
// }


// // ----------- Base class implementation

// FilteredBreakIteratorBuilder::FilteredBreakIteratorBuilder() {
// }

// FilteredBreakIteratorBuilder::~FilteredBreakIteratorBuilder() {
// }

// FilteredBreakIteratorBuilder *
// FilteredBreakIteratorBuilder::createInstance(const Locale& where, UErrorCode& status) {
//   if(U_FAILURE(status)) return NULL;
//   LocalPointer<FilteredBreakIteratorBuilder> ret(new SimpleFilteredBreakIteratorBuilder(where, status), status);
//   return (U_SUCCESS(status))? ret.orphan(): NULL;
// }

// FilteredBreakIteratorBuilder *
// FilteredBreakIteratorBuilder::createInstance(UErrorCode &status) {
//   return createEmptyInstance(status);
// }

// FilteredBreakIteratorBuilder *
// FilteredBreakIteratorBuilder::createEmptyInstance(UErrorCode& status) {
//   if(U_FAILURE(status)) return NULL;
//   LocalPointer<FilteredBreakIteratorBuilder> ret(new SimpleFilteredBreakIteratorBuilder(status), status);
//   return (U_SUCCESS(status))? ret.orphan(): NULL;
// }

// U_NAMESPACE_END

// #endif //#if !UCONFIG_NO_BREAK_ITERATION && !UCONFIG_NO_FILTERED_BREAK_ITERATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2009-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  filterednormalizer2.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2009dec10
*   created by: Markus W. Scherer
*/



#if !UCONFIG_NO_NORMALIZATION







// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1997-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  cpputils.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*/






/*==========================================================================*/
/* Array copy utility functions */
/*==========================================================================*/

// static
// inline void uprv_arrayCopy(const double* src, double* dst, int32_t count)
// { uprv_memcpy(dst, src, (size_t)count * sizeof(*src)); }

// static
// inline void uprv_arrayCopy(const double* src, int32_t srcStart,
//               double* dst, int32_t dstStart, int32_t count)
// { uprv_memcpy(dst+dstStart, src+srcStart, (size_t)count * sizeof(*src)); }

static
inline void uprv_arrayCopy(const int8_t* src, int8_t* dst, int32_t count)
    { uprv_memcpy(dst, src, (size_t)count * sizeof(*src)); }

// static
// inline void uprv_arrayCopy(const int8_t* src, int32_t srcStart,
//               int8_t* dst, int32_t dstStart, int32_t count)
// { uprv_memcpy(dst+dstStart, src+srcStart, (size_t)count * sizeof(*src)); }

// static
// inline void uprv_arrayCopy(const int16_t* src, int16_t* dst, int32_t count)
// { uprv_memcpy(dst, src, (size_t)count * sizeof(*src)); }

// static
// inline void uprv_arrayCopy(const int16_t* src, int32_t srcStart,
//               int16_t* dst, int32_t dstStart, int32_t count)
// { uprv_memcpy(dst+dstStart, src+srcStart, (size_t)count * sizeof(*src)); }

static
inline void uprv_arrayCopy(const int32_t* src, int32_t* dst, int32_t count)
{ uprv_memcpy(dst, src, (size_t)count * sizeof(*src)); }

// static
// inline void uprv_arrayCopy(const int32_t* src, int32_t srcStart,
//               int32_t* dst, int32_t dstStart, int32_t count)
// { uprv_memcpy(dst+dstStart, src+srcStart, (size_t)count * sizeof(*src)); }

// static
// inline void
// uprv_arrayCopy(const UChar *src, int32_t srcStart,
//         UChar *dst, int32_t dstStart, int32_t count)
// { uprv_memcpy(dst+dstStart, src+srcStart, (size_t)count * sizeof(*src)); }

/**
 * Copy an array of UnicodeString OBJECTS (not pointers).
 * @internal
 */
static inline void
uprv_arrayCopy(const icu::UnicodeString *src, icu::UnicodeString *dst, int32_t count)
{ while(count-- > 0) *dst++ = *src++; }

/**
 * Copy an array of UnicodeString OBJECTS (not pointers).
 * @internal
 */
static inline void
uprv_arrayCopy(const icu::UnicodeString *src, int32_t srcStart,
               icu::UnicodeString *dst, int32_t dstStart, int32_t count)
{ uprv_arrayCopy(src+srcStart, dst+dstStart, count); }

/**
 * Checks that the string is readable and writable.
 * Sets U_ILLEGAL_ARGUMENT_ERROR if the string isBogus() or has an open getBuffer().
 */
inline void
uprv_checkCanGetBuffer(const icu::UnicodeString &s, UErrorCode &errorCode) {
    if(U_SUCCESS(errorCode) && s.isBogus()) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
    }
}



U_NAMESPACE_BEGIN

FilteredNormalizer2::~FilteredNormalizer2() {}

UnicodeString &
FilteredNormalizer2::normalize(const UnicodeString &src,
                               UnicodeString &dest,
                               UErrorCode &errorCode) const {
    uprv_checkCanGetBuffer(src, errorCode);
    if(U_FAILURE(errorCode)) {
        dest.setToBogus();
        return dest;
    }
    if(&dest==&src) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return dest;
    }
    dest.remove();
    return normalize(src, dest, USET_SPAN_SIMPLE, errorCode);
}

// Internal: No argument checking, and appends to dest.
// Pass as input spanCondition the one that is likely to yield a non-zero
// span length at the start of src.
// For set=[:age=3.2:], since almost all common characters were in Unicode 3.2,
// USET_SPAN_SIMPLE should be passed in for the start of src
// and USET_SPAN_NOT_CONTAINED should be passed in if we continue after
// an in-filter prefix.
UnicodeString &
FilteredNormalizer2::normalize(const UnicodeString &src,
                               UnicodeString &dest,
                               USetSpanCondition spanCondition,
                               UErrorCode &errorCode) const {
    UnicodeString tempDest;  // Don't throw away destination buffer between iterations.
    for(int32_t prevSpanLimit=0; prevSpanLimit<src.length();) {
        int32_t spanLimit=set.span(src, prevSpanLimit, spanCondition);
        int32_t spanLength=spanLimit-prevSpanLimit;
        if(spanCondition==USET_SPAN_NOT_CONTAINED) {
            if(spanLength!=0) {
                dest.append(src, prevSpanLimit, spanLength);
            }
            spanCondition=USET_SPAN_SIMPLE;
        } else {
            if(spanLength!=0) {
                // Not norm2.normalizeSecondAndAppend() because we do not want
                // to modify the non-filter part of dest.
                dest.append(norm2.normalize(src.tempSubStringBetween(prevSpanLimit, spanLimit),
                                            tempDest, errorCode));
                if(U_FAILURE(errorCode)) {
                    break;
                }
            }
            spanCondition=USET_SPAN_NOT_CONTAINED;
        }
        prevSpanLimit=spanLimit;
    }
    return dest;
}

void
FilteredNormalizer2::normalizeUTF8(uint32_t options, StringPiece src, ByteSink &sink,
                                   Edits *edits, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) {
        return;
    }
    if (edits != nullptr && (options & U_EDITS_NO_RESET) == 0) {
        edits->reset();
    }
    options |= U_EDITS_NO_RESET;  // Do not reset for each span.
    normalizeUTF8(options, src.data(), src.length(), sink, edits, USET_SPAN_SIMPLE, errorCode);
}

void
FilteredNormalizer2::normalizeUTF8(uint32_t options, const char *src, int32_t length,
                                   ByteSink &sink, Edits *edits,
                                   USetSpanCondition spanCondition,
                                   UErrorCode &errorCode) const {
    while (length > 0) {
        int32_t spanLength = set.spanUTF8(src, length, spanCondition);
        if (spanCondition == USET_SPAN_NOT_CONTAINED) {
            if (spanLength != 0) {
                if (edits != nullptr) {
                    edits->addUnchanged(spanLength);
                }
                if ((options & U_OMIT_UNCHANGED_TEXT) == 0) {
                    sink.Append(src, spanLength);
                }
            }
            spanCondition = USET_SPAN_SIMPLE;
        } else {
            if (spanLength != 0) {
                // Not norm2.normalizeSecondAndAppend() because we do not want
                // to modify the non-filter part of dest.
                norm2.normalizeUTF8(options, StringPiece(src, spanLength), sink, edits, errorCode);
                if (U_FAILURE(errorCode)) {
                    break;
                }
            }
            spanCondition = USET_SPAN_NOT_CONTAINED;
        }
        src += spanLength;
        length -= spanLength;
    }
}

UnicodeString &
FilteredNormalizer2::normalizeSecondAndAppend(UnicodeString &first,
                                              const UnicodeString &second,
                                              UErrorCode &errorCode) const {
    return normalizeSecondAndAppend(first, second, TRUE, errorCode);
}

UnicodeString &
FilteredNormalizer2::append(UnicodeString &first,
                            const UnicodeString &second,
                            UErrorCode &errorCode) const {
    return normalizeSecondAndAppend(first, second, FALSE, errorCode);
}

UnicodeString &
FilteredNormalizer2::normalizeSecondAndAppend(UnicodeString &first,
                                              const UnicodeString &second,
                                              UBool doNormalize,
                                              UErrorCode &errorCode) const {
    uprv_checkCanGetBuffer(first, errorCode);
    uprv_checkCanGetBuffer(second, errorCode);
    if(U_FAILURE(errorCode)) {
        return first;
    }
    if(&first==&second) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return first;
    }
    if(first.isEmpty()) {
        if(doNormalize) {
            return normalize(second, first, errorCode);
        } else {
            return first=second;
        }
    }
    // merge the in-filter suffix of the first string with the in-filter prefix of the second
    int32_t prefixLimit=set.span(second, 0, USET_SPAN_SIMPLE);
    if(prefixLimit!=0) {
        UnicodeString prefix(second.tempSubString(0, prefixLimit));
        int32_t suffixStart=set.spanBack(first, INT32_MAX, USET_SPAN_SIMPLE);
        if(suffixStart==0) {
            if(doNormalize) {
                norm2.normalizeSecondAndAppend(first, prefix, errorCode);
            } else {
                norm2.append(first, prefix, errorCode);
            }
        } else {
            UnicodeString middle(first, suffixStart, INT32_MAX);
            if(doNormalize) {
                norm2.normalizeSecondAndAppend(middle, prefix, errorCode);
            } else {
                norm2.append(middle, prefix, errorCode);
            }
            first.replace(suffixStart, INT32_MAX, middle);
        }
    }
    if(prefixLimit<second.length()) {
        UnicodeString rest(second.tempSubString(prefixLimit, INT32_MAX));
        if(doNormalize) {
            normalize(rest, first, USET_SPAN_NOT_CONTAINED, errorCode);
        } else {
            first.append(rest);
        }
    }
    return first;
}

UBool
FilteredNormalizer2::getDecomposition(UChar32 c, UnicodeString &decomposition) const {
    return set.contains(c) && norm2.getDecomposition(c, decomposition);
}

UBool
FilteredNormalizer2::getRawDecomposition(UChar32 c, UnicodeString &decomposition) const {
    return set.contains(c) && norm2.getRawDecomposition(c, decomposition);
}

UChar32
FilteredNormalizer2::composePair(UChar32 a, UChar32 b) const {
    return (set.contains(a) && set.contains(b)) ? norm2.composePair(a, b) : U_SENTINEL;
}

uint8_t
FilteredNormalizer2::getCombiningClass(UChar32 c) const {
    return set.contains(c) ? norm2.getCombiningClass(c) : 0;
}

UBool
FilteredNormalizer2::isNormalized(const UnicodeString &s, UErrorCode &errorCode) const {
    uprv_checkCanGetBuffer(s, errorCode);
    if(U_FAILURE(errorCode)) {
        return FALSE;
    }
    USetSpanCondition spanCondition=USET_SPAN_SIMPLE;
    for(int32_t prevSpanLimit=0; prevSpanLimit<s.length();) {
        int32_t spanLimit=set.span(s, prevSpanLimit, spanCondition);
        if(spanCondition==USET_SPAN_NOT_CONTAINED) {
            spanCondition=USET_SPAN_SIMPLE;
        } else {
            if( !norm2.isNormalized(s.tempSubStringBetween(prevSpanLimit, spanLimit), errorCode) ||
                U_FAILURE(errorCode)
            ) {
                return FALSE;
            }
            spanCondition=USET_SPAN_NOT_CONTAINED;
        }
        prevSpanLimit=spanLimit;
    }
    return TRUE;
}

UBool
FilteredNormalizer2::isNormalizedUTF8(StringPiece sp, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return FALSE;
    }
    const char *s = sp.data();
    int32_t length = sp.length();
    USetSpanCondition spanCondition = USET_SPAN_SIMPLE;
    while (length > 0) {
        int32_t spanLength = set.spanUTF8(s, length, spanCondition);
        if (spanCondition == USET_SPAN_NOT_CONTAINED) {
            spanCondition = USET_SPAN_SIMPLE;
        } else {
            if (!norm2.isNormalizedUTF8(StringPiece(s, spanLength), errorCode) ||
                    U_FAILURE(errorCode)) {
                return FALSE;
            }
            spanCondition = USET_SPAN_NOT_CONTAINED;
        }
        s += spanLength;
        length -= spanLength;
    }
    return TRUE;
}

UNormalizationCheckResult
FilteredNormalizer2::quickCheck(const UnicodeString &s, UErrorCode &errorCode) const {
    uprv_checkCanGetBuffer(s, errorCode);
    if(U_FAILURE(errorCode)) {
        return UNORM_MAYBE;
    }
    UNormalizationCheckResult result=UNORM_YES;
    USetSpanCondition spanCondition=USET_SPAN_SIMPLE;
    for(int32_t prevSpanLimit=0; prevSpanLimit<s.length();) {
        int32_t spanLimit=set.span(s, prevSpanLimit, spanCondition);
        if(spanCondition==USET_SPAN_NOT_CONTAINED) {
            spanCondition=USET_SPAN_SIMPLE;
        } else {
            UNormalizationCheckResult qcResult=
                norm2.quickCheck(s.tempSubStringBetween(prevSpanLimit, spanLimit), errorCode);
            if(U_FAILURE(errorCode) || qcResult==UNORM_NO) {
                return qcResult;
            } else if(qcResult==UNORM_MAYBE) {
                result=qcResult;
            }
            spanCondition=USET_SPAN_NOT_CONTAINED;
        }
        prevSpanLimit=spanLimit;
    }
    return result;
}

int32_t
FilteredNormalizer2::spanQuickCheckYes(const UnicodeString &s, UErrorCode &errorCode) const {
    uprv_checkCanGetBuffer(s, errorCode);
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    USetSpanCondition spanCondition=USET_SPAN_SIMPLE;
    for(int32_t prevSpanLimit=0; prevSpanLimit<s.length();) {
        int32_t spanLimit=set.span(s, prevSpanLimit, spanCondition);
        if(spanCondition==USET_SPAN_NOT_CONTAINED) {
            spanCondition=USET_SPAN_SIMPLE;
        } else {
            int32_t yesLimit=
                prevSpanLimit+
                norm2.spanQuickCheckYes(
                    s.tempSubStringBetween(prevSpanLimit, spanLimit), errorCode);
            if(U_FAILURE(errorCode) || yesLimit<spanLimit) {
                return yesLimit;
            }
            spanCondition=USET_SPAN_NOT_CONTAINED;
        }
        prevSpanLimit=spanLimit;
    }
    return s.length();
}

UBool
FilteredNormalizer2::hasBoundaryBefore(UChar32 c) const {
    return !set.contains(c) || norm2.hasBoundaryBefore(c);
}

UBool
FilteredNormalizer2::hasBoundaryAfter(UChar32 c) const {
    return !set.contains(c) || norm2.hasBoundaryAfter(c);
}

UBool
FilteredNormalizer2::isInert(UChar32 c) const {
    return !set.contains(c) || norm2.isInert(c);
}

U_NAMESPACE_END

// C API ------------------------------------------------------------------- ***

U_NAMESPACE_USE

U_CAPI UNormalizer2 * U_EXPORT2
unorm2_openFiltered(const UNormalizer2 *norm2, const USet *filterSet, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if(filterSet==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    Normalizer2 *fn2=new FilteredNormalizer2(*(Normalizer2 *)norm2,
                                             *UnicodeSet::fromUSet(filterSet));
    if(fn2==NULL) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
    }
    return (UNormalizer2 *)fn2;
}

#endif  // !UCONFIG_NO_NORMALIZATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2009-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*/


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2009-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*/


/**
 * \file
 * \brief C API: access to ICU Data Version number
 */




/**
 * @stable ICU 49
 */
#define U_ICU_VERSION_BUNDLE "icuver"

/**
 * @stable ICU 49
 */
#define U_ICU_DATA_KEY "DataVersion"

/**
 * Retrieves the data version from icuver and stores it in dataVersionFillin.
 * 
 * @param dataVersionFillin icuver data version information to be filled in if not-null
 * @param status stores the error code from the calls to resource bundle
 * 
 * @stable ICU 49
 */
U_STABLE void U_EXPORT2 u_getDataVersion(UVersionInfo dataVersionFillin, UErrorCode *status);



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2000-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 1999-2016, International Business Machines
*                Corporation and others. All Rights Reserved.
******************************************************************************
*   file name:  uresdata.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999dec08
*   created by: Markus W. Scherer
*   06/24/02    weiv        Added support for resource sharing
*/



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  udata.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999oct25
*   created by: Markus W. Scherer
*/





U_CDECL_BEGIN

/**
 * \file
 * \brief C API: Data loading interface
 *
 * <h2>Information about data loading interface</h2>
 *
 * This API is used to find and efficiently load data for ICU and applications
 * using ICU. It provides an abstract interface that specifies a data type and
 * name to find and load the data. Normally this API is used by other ICU APIs
 * to load required data out of the ICU data library, but it can be used to
 * load data out of other places.
 *
 * See the User Guide Data Management chapter.
 */
 
#ifndef U_HIDE_INTERNAL_API
/**
 * Character used to separate package names from tree names 
 * @internal ICU 3.0
 */
#define U_TREE_SEPARATOR '-'

/**
 * String used to separate package names from tree names 
 * @internal ICU 3.0
 */
#define U_TREE_SEPARATOR_STRING "-"

/**
 * Character used to separate parts of entry names
 * @internal ICU 3.0
 */
#define U_TREE_ENTRY_SEP_CHAR '/'

/**
 * String used to separate parts of entry names
 * @internal ICU 3.0
 */
#define U_TREE_ENTRY_SEP_STRING "/"

/**
 * Alias for standard ICU data 
 * @internal ICU 3.0
 */
#define U_ICUDATA_ALIAS "ICUDATA"

#endif /* U_HIDE_INTERNAL_API */

/**
 * UDataInfo contains the properties about the requested data.
 * This is meta data.
 *
 * <p>This structure may grow in the future, indicated by the
 * <code>size</code> field.</p>
 *
 * <p>ICU data must be at least 8-aligned, and should be 16-aligned.
 * The UDataInfo struct begins 4 bytes after the start of the data item,
 * so it is 4-aligned.
 *
 * <p>The platform data property fields help determine if a data
 * file can be efficiently used on a given machine.
 * The particular fields are of importance only if the data
 * is affected by the properties - if there is integer data
 * with word sizes > 1 byte, char* text, or UChar* text.</p>
 *
 * <p>The implementation for the <code>udata_open[Choice]()</code>
 * functions may reject data based on the value in <code>isBigEndian</code>.
 * No other field is used by the <code>udata</code> API implementation.</p>
 *
 * <p>The <code>dataFormat</code> may be used to identify
 * the kind of data, e.g. a converter table.</p>
 *
 * <p>The <code>formatVersion</code> field should be used to
 * make sure that the format can be interpreted.
 * It may be a good idea to check only for the one or two highest
 * of the version elements to allow the data memory to
 * get more or somewhat rearranged contents, for as long
 * as the using code can still interpret the older contents.</p>
 *
 * <p>The <code>dataVersion</code> field is intended to be a
 * common place to store the source version of the data;
 * for data from the Unicode character database, this could
 * reflect the Unicode version.</p>
 *
 * @stable ICU 2.0
 */
typedef struct {
    /** sizeof(UDataInfo)
     *  @stable ICU 2.0 */
    uint16_t size;

    /** unused, set to 0 
     *  @stable ICU 2.0*/
    uint16_t reservedWord;

    /* platform data properties */
    /** 0 for little-endian machine, 1 for big-endian
     *  @stable ICU 2.0 */
    uint8_t isBigEndian;

    /** see U_CHARSET_FAMILY values in utypes.h 
     *  @stable ICU 2.0*/
    uint8_t charsetFamily;

    /** sizeof(UChar), one of { 1, 2, 4 } 
     *  @stable ICU 2.0*/
    uint8_t sizeofUChar;

    /** unused, set to 0 
     *  @stable ICU 2.0*/
    uint8_t reservedByte;

    /** data format identifier 
     *  @stable ICU 2.0*/
    uint8_t dataFormat[4];

    /** versions: [0] major [1] minor [2] milli [3] micro 
     *  @stable ICU 2.0*/
    uint8_t formatVersion[4];

    /** versions: [0] major [1] minor [2] milli [3] micro 
     *  @stable ICU 2.0*/
    uint8_t dataVersion[4];
} UDataInfo;

/* API for reading data -----------------------------------------------------*/

/**
 * Forward declaration of the data memory type.
 * @stable ICU 2.0
 */
typedef struct UDataMemory UDataMemory;

/**
 * Callback function for udata_openChoice().
 * @param context parameter passed into <code>udata_openChoice()</code>.
 * @param type The type of the data as passed into <code>udata_openChoice()</code>.
 *             It may be <code>NULL</code>.
 * @param name The name of the data as passed into <code>udata_openChoice()</code>.
 * @param pInfo A pointer to the <code>UDataInfo</code> structure
 *              of data that has been loaded and will be returned
 *              by <code>udata_openChoice()</code> if this function
 *              returns <code>TRUE</code>.
 * @return TRUE if the current data memory is acceptable
 * @stable ICU 2.0
 */
typedef UBool U_CALLCONV
UDataMemoryIsAcceptable(void *context,
                        const char *type, const char *name,
                        const UDataInfo *pInfo);


/**
 * Convenience function.
 * This function works the same as <code>udata_openChoice</code>
 * except that any data that matches the type and name
 * is assumed to be acceptable.
 * @param path Specifies an absolute path and/or a basename for the
 *             finding of the data in the file system.
 *             <code>NULL</code> for ICU data.
 * @param type A string that specifies the type of data to be loaded.
 *             For example, resource bundles are loaded with type "res",
 *             conversion tables with type "cnv".
 *             This may be <code>NULL</code> or empty.
 * @param name A string that specifies the name of the data.
 * @param pErrorCode An ICU UErrorCode parameter. It must not be <code>NULL</code>.
 * @return A pointer (handle) to a data memory object, or <code>NULL</code>
 *         if an error occurs. Call <code>udata_getMemory()</code>
 *         to get a pointer to the actual data.
 *
 * @see udata_openChoice
 * @stable ICU 2.0
 */
U_STABLE UDataMemory * U_EXPORT2
udata_open(const char *path, const char *type, const char *name,
           UErrorCode *pErrorCode);

/**
 * Data loading function.
 * This function is used to find and load efficiently data for
 * ICU and applications using ICU.
 * It provides an abstract interface that allows to specify a data
 * type and name to find and load the data.
 *
 * <p>The implementation depends on platform properties and user preferences
 * and may involve loading shared libraries (DLLs), mapping
 * files into memory, or fopen()/fread() files.
 * It may also involve using static memory or database queries etc.
 * Several or all data items may be combined into one entity
 * (DLL, memory-mappable file).</p>
 *
 * <p>The data is always preceded by a header that includes
 * a <code>UDataInfo</code> structure.
 * The caller's <code>isAcceptable()</code> function is called to make
 * sure that the data is useful. It may be called several times if it
 * rejects the data and there is more than one location with data
 * matching the type and name.</p>
 *
 * <p>If <code>path==NULL</code>, then ICU data is loaded.
 * Otherwise, it is separated into a basename and a basename-less directory string.
 * The basename is used as the data package name, and the directory is
 * logically prepended to the ICU data directory string.</p>
 *
 * <p>For details about ICU data loading see the User Guide
 * Data Management chapter. (http://icu-project.org/userguide/icudata.html)</p>
 *
 * @param path Specifies an absolute path and/or a basename for the
 *             finding of the data in the file system.
 *             <code>NULL</code> for ICU data.
 * @param type A string that specifies the type of data to be loaded.
 *             For example, resource bundles are loaded with type "res",
 *             conversion tables with type "cnv".
 *             This may be <code>NULL</code> or empty.
 * @param name A string that specifies the name of the data.
 * @param isAcceptable This function is called to verify that loaded data
 *                     is useful for the client code. If it returns FALSE
 *                     for all data items, then <code>udata_openChoice()</code>
 *                     will return with an error.
 * @param context Arbitrary parameter to be passed into isAcceptable.
 * @param pErrorCode An ICU UErrorCode parameter. It must not be <code>NULL</code>.
 * @return A pointer (handle) to a data memory object, or <code>NULL</code>
 *         if an error occurs. Call <code>udata_getMemory()</code>
 *         to get a pointer to the actual data.
 * @stable ICU 2.0
 */
U_STABLE UDataMemory * U_EXPORT2
udata_openChoice(const char *path, const char *type, const char *name,
                 UDataMemoryIsAcceptable *isAcceptable, void *context,
                 UErrorCode *pErrorCode);

/**
 * Close the data memory.
 * This function must be called to allow the system to
 * release resources associated with this data memory.
 * @param pData The pointer to data memory object
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
udata_close(UDataMemory *pData);

/**
 * Get the pointer to the actual data inside the data memory.
 * The data is read-only.
 *
 * ICU data must be at least 8-aligned, and should be 16-aligned.
 *
 * @param pData The pointer to data memory object
 * @stable ICU 2.0
 */
U_STABLE const void * U_EXPORT2
udata_getMemory(UDataMemory *pData);

/**
 * Get the information from the data memory header.
 * This allows to get access to the header containing
 * platform data properties etc. which is not part of
 * the data itself and can therefore not be accessed
 * via the pointer that <code>udata_getMemory()</code> returns.
 *
 * @param pData pointer to the data memory object
 * @param pInfo pointer to a UDataInfo object;
 *              its <code>size</code> field must be set correctly,
 *              typically to <code>sizeof(UDataInfo)</code>.
 *
 * <code>*pInfo</code> will be filled with the UDataInfo structure
 * in the data memory object. If this structure is smaller than
 * <code>pInfo->size</code>, then the <code>size</code> will be
 * adjusted and only part of the structure will be filled.
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
udata_getInfo(UDataMemory *pData, UDataInfo *pInfo);

/**
 * This function bypasses the normal ICU data loading process and
 * allows you to force ICU's system data to come out of a user-specified
 * area in memory.
 *
 * ICU data must be at least 8-aligned, and should be 16-aligned.
 * See http://userguide.icu-project.org/icudata
 *
 * The format of this data is that of the icu common data file, as is
 * generated by the pkgdata tool with mode=common or mode=dll.
 * You can read in a whole common mode file and pass the address to the start of the
 * data, or (with the appropriate link options) pass in the pointer to
 * the data that has been loaded from a dll by the operating system,
 * as shown in this code:
 *
 *       extern const char U_IMPORT U_ICUDATA_ENTRY_POINT [];
 *        // U_ICUDATA_ENTRY_POINT is same as entry point specified to pkgdata tool
 *       UErrorCode  status = U_ZERO_ERROR;
 *
 *       udata_setCommonData(&U_ICUDATA_ENTRY_POINT, &status);
 *
 * It is important that the declaration be as above. The entry point
 * must not be declared as an extern void*.
 *
 * Starting with ICU 4.4, it is possible to set several data packages,
 * one per call to this function.
 * udata_open() will look for data in the multiple data packages in the order
 * in which they were set.
 * The position of the linked-in or default-name ICU .data package in the
 * search list depends on when the first data item is loaded that is not contained
 * in the already explicitly set packages.
 * If data was loaded implicitly before the first call to this function
 * (for example, via opening a converter, constructing a UnicodeString
 * from default-codepage data, using formatting or collation APIs, etc.),
 * then the default data will be first in the list.
 *
 * This function has no effect on application (non ICU) data.  See udata_setAppData()
 * for similar functionality for application data.
 *
 * @param data pointer to ICU common data
 * @param err outgoing error status <code>U_USING_DEFAULT_WARNING, U_UNSUPPORTED_ERROR</code>
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
udata_setCommonData(const void *data, UErrorCode *err);


/**
 * This function bypasses the normal ICU data loading process for application-specific
 * data and allows you to force the it to come out of a user-specified
 * pointer.
 *
 * ICU data must be at least 8-aligned, and should be 16-aligned.
 * See http://userguide.icu-project.org/icudata
 *
 * The format of this data is that of the icu common data file, like 'icudt26l.dat'
 * or the corresponding shared library (DLL) file.
 * The application must read in or otherwise construct an image of the data and then
 * pass the address of it to this function.
 *
 *
 * Warning:  setAppData will set a U_USING_DEFAULT_WARNING code if
 *           data with the specifed path that has already been opened, or
 *           if setAppData with the same path has already been called.
 *           Any such calls to setAppData will have no effect.
 *
 *
 * @param packageName the package name by which the application will refer
 *             to (open) this data
 * @param data pointer to the data
 * @param err outgoing error status <code>U_USING_DEFAULT_WARNING, U_UNSUPPORTED_ERROR</code>
 * @see udata_setCommonData
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
udata_setAppData(const char *packageName, const void *data, UErrorCode *err);

/**
 * Possible settings for udata_setFileAccess()
 * @see udata_setFileAccess
 * @stable ICU 3.4
 */
typedef enum UDataFileAccess {
    /** ICU looks for data in single files first, then in packages. (default) @stable ICU 3.4 */
    UDATA_FILES_FIRST,
    /** An alias for the default access mode. @stable ICU 3.4 */
    UDATA_DEFAULT_ACCESS = UDATA_FILES_FIRST,
    /** ICU only loads data from packages, not from single files. @stable ICU 3.4 */
    UDATA_ONLY_PACKAGES,
    /** ICU loads data from packages first, and only from single files
        if the data cannot be found in a package. @stable ICU 3.4 */
    UDATA_PACKAGES_FIRST,
    /** ICU does not access the file system for data loading. @stable ICU 3.4 */
    UDATA_NO_FILES,
#ifndef U_HIDE_DEPRECATED_API
    /**
     * Number of real UDataFileAccess values.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UDATA_FILE_ACCESS_COUNT
#endif  // U_HIDE_DEPRECATED_API
} UDataFileAccess;

/**
 * This function may be called to control how ICU loads data. It must be called
 * before any ICU data is loaded, including application data loaded with 
 * ures/ResourceBundle or udata APIs. This function is not multithread safe.  
 * The results of calling it while other threads are loading data are undefined.
 * @param access The type of file access to be used
 * @param status Error code.
 * @see UDataFileAccess
 * @stable ICU 3.4 
 */
U_STABLE void U_EXPORT2
udata_setFileAccess(UDataFileAccess access, UErrorCode *status);

U_CDECL_END

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUDataMemoryPointer
 * "Smart pointer" class, closes a UDataMemory via udata_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUDataMemoryPointer, UDataMemory, udata_close);

U_NAMESPACE_END

#endif  // U_SHOW_CPLUSPLUS_API






/**
 * Numeric constants for internal-only types of resource items.
 * These must use different numeric values than UResType constants
 * because they are used together.
 * Internal types are never returned by ures_getType().
 */
typedef enum {
    /** Include a negative value so that the compiler uses the same int type as for UResType. */
    URES_INTERNAL_NONE=-1,

    /** Resource type constant for tables with 32-bit count, key offsets and values. */
    URES_TABLE32=4,

    /**
     * Resource type constant for tables with 16-bit count, key offsets and values.
     * All values are URES_STRING_V2 strings.
     */
    URES_TABLE16=5,

    /** Resource type constant for 16-bit Unicode strings in formatVersion 2. */
    URES_STRING_V2=6,

    /**
     * Resource type constant for arrays with 16-bit count and values.
     * All values are URES_STRING_V2 strings.
     */
    URES_ARRAY16=9

    /* Resource type 15 is not defined but effectively used by RES_BOGUS=0xffffffff. */
} UResInternalType;

/*
 * A Resource is a 32-bit value that has 2 bit fields:
 * 31..28   4-bit type, see enum below
 * 27..0    28-bit four-byte-offset or value according to the type
 */
typedef uint32_t Resource;

#define RES_BOGUS 0xffffffff
#define RES_MAX_OFFSET 0x0fffffff

#define RES_GET_TYPE(res) ((int32_t)((res)>>28UL))
#define RES_GET_OFFSET(res) ((res)&0x0fffffff)
#define RES_GET_POINTER(pRoot, res) ((pRoot)+RES_GET_OFFSET(res))

/* get signed and unsigned integer values directly from the Resource handle
 * NOTE: For proper logging, please use the res_getInt() constexpr
 */
#if U_SIGNED_RIGHT_SHIFT_IS_ARITHMETIC
#   define RES_GET_INT_NO_TRACE(res) (((int32_t)((res)<<4L))>>4L)
#else
#   define RES_GET_INT_NO_TRACE(res) (int32_t)(((res)&0x08000000) ? (res)|0xf0000000 : (res)&0x07ffffff)
#endif

#define RES_GET_UINT_NO_TRACE(res) ((res)&0x0fffffff)

#define URES_IS_ARRAY(type) ((int32_t)(type)==URES_ARRAY || (int32_t)(type)==URES_ARRAY16)
#define URES_IS_TABLE(type) ((int32_t)(type)==URES_TABLE || (int32_t)(type)==URES_TABLE16 || (int32_t)(type)==URES_TABLE32)
#define URES_IS_CONTAINER(type) (URES_IS_TABLE(type) || URES_IS_ARRAY(type))

#define URES_MAKE_RESOURCE(type, offset) (((Resource)(type)<<28)|(Resource)(offset))
#define URES_MAKE_EMPTY_RESOURCE(type) ((Resource)(type)<<28)

/* indexes[] value names; indexes are generally 32-bit (Resource) indexes */
enum {
    /**
     * [0] contains the length of indexes[]
     * which is at most URES_INDEX_TOP of the latest format version
     *
     * formatVersion==1: all bits contain the length of indexes[]
     *   but the length is much less than 0xff;
     * formatVersion>1:
     *   only bits  7..0 contain the length of indexes[],
     *        bits 31..8 are reserved and set to 0
     * formatVersion>=3:
     *        bits 31..8 poolStringIndexLimit bits 23..0
     */
    URES_INDEX_LENGTH,
    /**
     * [1] contains the top of the key strings,
     *     same as the bottom of resources or UTF-16 strings, rounded up
     */
    URES_INDEX_KEYS_TOP,
    /** [2] contains the top of all resources */
    URES_INDEX_RESOURCES_TOP,
    /**
     * [3] contains the top of the bundle,
     *     in case it were ever different from [2]
     */
    URES_INDEX_BUNDLE_TOP,
    /** [4] max. length of any table */
    URES_INDEX_MAX_TABLE_LENGTH,
    /**
     * [5] attributes bit set, see URES_ATT_* (new in formatVersion 1.2)
     *
     * formatVersion>=3:
     *   bits 31..16 poolStringIndex16Limit
     *   bits 15..12 poolStringIndexLimit bits 27..24
     */
    URES_INDEX_ATTRIBUTES,
    /**
     * [6] top of the 16-bit units (UTF-16 string v2 UChars, URES_TABLE16, URES_ARRAY16),
     *     rounded up (new in formatVersion 2.0, ICU 4.4)
     */
    URES_INDEX_16BIT_TOP,
    /** [7] checksum of the pool bundle (new in formatVersion 2.0, ICU 4.4) */
    URES_INDEX_POOL_CHECKSUM,
    URES_INDEX_TOP
};

/*
 * Nofallback attribute, attribute bit 0 in indexes[URES_INDEX_ATTRIBUTES].
 * New in formatVersion 1.2 (ICU 3.6).
 *
 * If set, then this resource bundle is a standalone bundle.
 * If not set, then the bundle participates in locale fallback, eventually
 * all the way to the root bundle.
 * If indexes[] is missing or too short, then the attribute cannot be determined
 * reliably. Dependency checking should ignore such bundles, and loading should
 * use fallbacks.
 */
#define URES_ATT_NO_FALLBACK 1

/*
 * Attributes for bundles that are, or use, a pool bundle.
 * A pool bundle provides key strings that are shared among several other bundles
 * to reduce their total size.
 * New in formatVersion 2 (ICU 4.4).
 */
#define URES_ATT_IS_POOL_BUNDLE 2
#define URES_ATT_USES_POOL_BUNDLE 4

/*
 * File format for .res resource bundle files
 *
 * ICU 56: New in formatVersion 3 compared with 2: -------------
 *
 * Resource bundles can optionally use shared string-v2 values
 * stored in the pool bundle.
 * If so, then the indexes[] contain two new values
 * in previously-unused bits of existing indexes[] slots:
 * - poolStringIndexLimit:
 *     String-v2 offsets (in 32-bit Resource words) below this limit
 *     point to pool bundle string-v2 values.
 * - poolStringIndex16Limit:
 *     Resource16 string-v2 offsets below this limit
 *     point to pool bundle string-v2 values.
 * Guarantee: poolStringIndex16Limit <= poolStringIndexLimit
 *
 * The local bundle's poolStringIndexLimit is greater than
 * any pool bundle string index used in the local bundle.
 * The poolStringIndexLimit should not be greater than
 * the maximum possible pool bundle string index.
 *
 * The maximum possible pool bundle string index is the index to the last non-NUL
 * pool string character, due to suffix sharing.
 *
 * In the pool bundle, there is no structure that lists the strings.
 * (The root resource is an empty Table.)
 * If the strings need to be enumerated (as genrb --usePoolBundle does),
 * then iterate through the pool bundle's 16-bit-units array from the beginning.
 * Stop at the end of the array, or when an explicit or implicit string length
 * would lead beyond the end of the array,
 * or when an apparent string is not NUL-terminated.
 * (Future genrb version might terminate the strings with
 * what looks like a large explicit string length.)
 *
 * ICU 4.4: New in formatVersion 2 compared with 1.3: -------------
 *
 * Three new resource types -- String-v2, Table16 and Array16 -- have their
 * values stored in a new array of 16-bit units between the table key strings
 * and the start of the other resources.
 *
 * genrb eliminates duplicates among Unicode string-v2 values.
 * Multiple Unicode strings may use the same offset and string data,
 * or a short string may point to the suffix of a longer string. ("Suffix sharing")
 * For example, one string "abc" may be reused for another string "bc" by pointing
 * to the second character. (Short strings-v2 are NUL-terminated
 * and not preceded by an explicit length value.)
 *
 * It is allowed for all resource types to share values.
 * The swapper code (ures_swap()) has been modified so that it swaps each item
 * exactly once.
 *
 * A resource bundle may use a special pool bundle. Some or all of the table key strings
 * of the using-bundle are omitted, and the key string offsets for such key strings refer
 * to offsets in the pool bundle.
 * The using-bundle's and the pool-bundle's indexes[URES_INDEX_POOL_CHECKSUM] values
 * must match.
 * Two bits in indexes[URES_INDEX_ATTRIBUTES] indicate whether a resource bundle
 * is or uses a pool bundle.
 *
 * Table key strings must be compared in ASCII order, even if they are not
 * stored in ASCII.
 *
 * New in formatVersion 1.3 compared with 1.2: -------------
 *
 * genrb eliminates duplicates among key strings.
 * Multiple table items may share one key string, or one item may point
 * to the suffix of another's key string. ("Suffix sharing")
 * For example, one key "abc" may be reused for another key "bc" by pointing
 * to the second character. (Key strings are NUL-terminated.)
 *
 * -------------
 *
 * An ICU4C resource bundle file (.res) is a binary, memory-mappable file
 * with nested, hierarchical data structures.
 * It physically contains the following:
 *
 *   Resource root; -- 32-bit Resource item, root item for this bundle's tree;
 *                     currently, the root item must be a table or table32 resource item
 *   int32_t indexes[indexes[0]]; -- array of indexes for friendly
 *                                   reading and swapping; see URES_INDEX_* above
 *                                   new in formatVersion 1.1 (ICU 2.8)
 *   char keys[]; -- characters for key strings
 *                   (formatVersion 1.0: up to 65k of characters; 1.1: <2G)
 *                   (minus the space for root and indexes[]),
 *                   which consist of invariant characters (ASCII/EBCDIC) and are NUL-terminated;
 *                   padded to multiple of 4 bytes for 4-alignment of the following data
 *   uint16_t 16BitUnits[]; -- resources that are stored entirely as sequences of 16-bit units
 *                             (new in formatVersion 2/ICU 4.4)
 *                             data is indexed by the offset values in 16-bit resource types,
 *                             with offset 0 pointing to the beginning of this array;
 *                             there is a 0 at offset 0, for empty resources;
 *                             padded to multiple of 4 bytes for 4-alignment of the following data
 *   data; -- data directly and indirectly indexed by the root item;
 *            the structure is determined by walking the tree
 *
 * Each resource bundle item has a 32-bit Resource handle (see typedef above)
 * which contains the item type number in its upper 4 bits (31..28) and either
 * an offset or a direct value in its lower 28 bits (27..0).
 * The order of items is undefined and only determined by walking the tree.
 * Leaves of the tree may be stored first or last or anywhere in between,
 * and it is in theory possible to have unreferenced holes in the file.
 *
 * 16-bit-unit values:
 * Starting with formatVersion 2/ICU 4.4, some resources are stored in a special
 * array of 16-bit units. Each resource value is a sequence of 16-bit units,
 * with no per-resource padding to a 4-byte boundary.
 * 16-bit container types (Table16 and Array16) contain Resource16 values
 * which are offsets to String-v2 resources in the same 16-bit-units array.
 *
 * Direct values:
 * - Empty Unicode strings have an offset value of 0 in the Resource handle itself.
 * - Starting with formatVersion 2/ICU 4.4, an offset value of 0 for
 *   _any_ resource type indicates an empty value.
 * - Integer values are 28-bit values stored in the Resource handle itself;
 *   the interpretation of unsigned vs. signed integers is up to the application.
 *
 * All other types and values use 28-bit offsets to point to the item's data.
 * The offset is an index to the first 32-bit word of the value, relative to the
 * start of the resource data (i.e., the root item handle is at offset 0).
 * To get byte offsets, the offset is multiplied by 4 (or shifted left by 2 bits).
 * All resource item values are 4-aligned.
 *
 * New in formatVersion 2/ICU 4.4: Some types use offsets into the 16-bit-units array,
 * indexing 16-bit units in that array.
 *
 * The structures (memory layouts) for the values for each item type are listed
 * in the table below.
 *
 * Nested, hierarchical structures: -------------
 *
 * Table items contain key-value pairs where the keys are offsets to char * key strings.
 * The values of these pairs are either Resource handles or
 * offsets into the 16-bit-units array, depending on the table type.
 *
 * Array items are simple vectors of Resource handles,
 * or of offsets into the 16-bit-units array, depending on the array type.
 *
 * Table key string offsets: -------
 *
 * Key string offsets are relative to the start of the resource data (of the root handle),
 * i.e., the first string has an offset of 4+sizeof(indexes).
 * (After the 4-byte root handle and after the indexes array.)
 *
 * If the resource bundle uses a pool bundle, then some key strings are stored
 * in the pool bundle rather than in the local bundle itself.
 * - In a Table or Table16, the 16-bit key string offset is local if it is
 *   less than indexes[URES_INDEX_KEYS_TOP]<<2.
 *   Otherwise, subtract indexes[URES_INDEX_KEYS_TOP]<<2 to get the offset into
 *   the pool bundle key strings.
 * - In a Table32, the 32-bit key string offset is local if it is non-negative.
 *   Otherwise, reset bit 31 to get the pool key string offset.
 *
 * Unlike the local offset, the pool key offset is relative to
 * the start of the key strings, not to the start of the bundle.
 *
 * An alias item is special (and new in ICU 2.4): --------------
 *
 * Its memory layout is just like for a UnicodeString, but at runtime it resolves to
 * another resource bundle's item according to the path in the string.
 * This is used to share items across bundles that are in different lookup/fallback
 * chains (e.g., large collation data among zh_TW and zh_HK).
 * This saves space (for large items) and maintenance effort (less duplication of data).
 *
 * --------------------------------------------------------------------------
 *
 * Resource types:
 *
 * Most resources have their values stored at four-byte offsets from the start
 * of the resource data. These values are at least 4-aligned.
 * Some resource values are stored directly in the offset field of the Resource itself.
 * See UResType in unicode/ures.h for enumeration constants for Resource types.
 *
 * Some resources have their values stored as sequences of 16-bit units,
 * at 2-byte offsets from the start of a contiguous 16-bit-unit array between
 * the table key strings and the other resources. (new in formatVersion 2/ICU 4.4)
 * At offset 0 of that array is a 16-bit zero value for empty 16-bit resources.
 *
 * Resource16 values in Table16 and Array16 are 16-bit offsets to String-v2
 * resources, with the offsets relative to the start of the 16-bit-units array.
 * Starting with formatVersion 3/ICU 56, if offset<poolStringIndex16Limit
 * then use the pool bundle's 16-bit-units array,
 * otherwise subtract that limit and use the local 16-bit-units array.
 *
 * Type Name            Memory layout of values
 *                      (in parentheses: scalar, non-offset values)
 *
 * 0  Unicode String:   int32_t length, UChar[length], (UChar)0, (padding)
 *                  or  (empty string ("") if offset==0)
 * 1  Binary:           int32_t length, uint8_t[length], (padding)
 *                      - the start of the bytes is 16-aligned -
 * 2  Table:            uint16_t count, uint16_t keyStringOffsets[count], (uint16_t padding), Resource[count]
 * 3  Alias:            (physically same value layout as string, new in ICU 2.4)
 * 4  Table32:          int32_t count, int32_t keyStringOffsets[count], Resource[count]
 *                      (new in formatVersion 1.1/ICU 2.8)
 * 5  Table16:          uint16_t count, uint16_t keyStringOffsets[count], Resource16[count]
 *                      (stored in the 16-bit-units array; new in formatVersion 2/ICU 4.4)
 * 6  Unicode String-v2:UChar[length], (UChar)0; length determined by the first UChar:
 *                      - if first is not a trail surrogate, then the length is implicit
 *                        and u_strlen() needs to be called
 *                      - if first<0xdfef then length=first&0x3ff (and skip first)
 *                      - if first<0xdfff then length=((first-0xdfef)<<16) | second UChar
 *                      - if first==0xdfff then length=((second UChar)<<16) | third UChar
 *                      (stored in the 16-bit-units array; new in formatVersion 2/ICU 4.4)
 *
 *                      Starting with formatVersion 3/ICU 56, if offset<poolStringIndexLimit
 *                      then use the pool bundle's 16-bit-units array,
 *                      otherwise subtract that limit and use the local 16-bit-units array.
 *                      (Note different limits for Resource16 vs. Resource.)
 *
 * 7  Integer:          (28-bit offset is integer value)
 * 8  Array:            int32_t count, Resource[count]
 * 9  Array16:          uint16_t count, Resource16[count]
 *                      (stored in the 16-bit-units array; new in formatVersion 2/ICU 4.4)
 * 14 Integer Vector:   int32_t length, int32_t[length]
 * 15 Reserved:         This value denotes special purpose resources and is for internal use.
 *
 * Note that there are 3 types with data vector values:
 * - Vectors of 8-bit bytes stored as type Binary.
 * - Vectors of 16-bit words stored as type Unicode String or Unicode String-v2
 *                     (no value restrictions, all values 0..ffff allowed!).
 * - Vectors of 32-bit words stored as type Integer Vector.
 */

/*
 * Structure for a single, memory-mapped ResourceBundle.
 */
typedef struct ResourceData {
    UDataMemory *data;
    const int32_t *pRoot;
    const uint16_t *p16BitUnits;
    const char *poolBundleKeys;
    Resource rootRes;
    int32_t localKeyLimit;
    const uint16_t *poolBundleStrings;
    int32_t poolStringIndexLimit;
    int32_t poolStringIndex16Limit;
    UBool noFallback; /* see URES_ATT_NO_FALLBACK */
    UBool isPoolBundle;
    UBool usesPoolBundle;
    UBool useNativeStrcmp;
} ResourceData;

/*
 * Read a resource bundle from memory.
 */
U_INTERNAL void U_EXPORT2
res_read(ResourceData *pResData,
         const UDataInfo *pInfo, const void *inBytes, int32_t length,
         UErrorCode *errorCode);

/*
 * Load a resource bundle file.
 * The ResourceData structure must be allocated externally.
 */
U_CFUNC void
res_load(ResourceData *pResData,
         const char *path, const char *name, UErrorCode *errorCode);

/*
 * Release a resource bundle file.
 * This does not release the ResourceData structure itself.
 */
U_CFUNC void
res_unload(ResourceData *pResData);

U_INTERNAL UResType U_EXPORT2
res_getPublicType(Resource res);

///////////////////////////////////////////////////////////////////////////
// To enable tracing, use the inline versions of the res_get* functions. //
///////////////////////////////////////////////////////////////////////////

/*
 * Return a pointer to a zero-terminated, const UChar* string
 * and set its length in *pLength.
 * Returns NULL if not found.
 */
U_INTERNAL const UChar * U_EXPORT2
res_getStringNoTrace(const ResourceData *pResData, Resource res, int32_t *pLength);

U_INTERNAL const uint8_t * U_EXPORT2
res_getBinaryNoTrace(const ResourceData *pResData, Resource res, int32_t *pLength);

U_INTERNAL const int32_t * U_EXPORT2
res_getIntVectorNoTrace(const ResourceData *pResData, Resource res, int32_t *pLength);

U_INTERNAL const UChar * U_EXPORT2
res_getAlias(const ResourceData *pResData, Resource res, int32_t *pLength);

U_INTERNAL Resource U_EXPORT2
res_getResource(const ResourceData *pResData, const char *key);

U_INTERNAL int32_t U_EXPORT2
res_countArrayItems(const ResourceData *pResData, Resource res);

U_INTERNAL Resource U_EXPORT2
res_getArrayItem(const ResourceData *pResData, Resource array, int32_t indexS);

U_INTERNAL Resource U_EXPORT2
res_getTableItemByIndex(const ResourceData *pResData, Resource table, int32_t indexS, const char ** key);

U_INTERNAL Resource U_EXPORT2
res_getTableItemByKey(const ResourceData *pResData, Resource table, int32_t *indexS, const char* * key);

/**
 * Iterates over the path and stops when a scalar resource is found.
 * Follows aliases.
 * Modifies the contents of *path (replacing separators with NULs),
 * and also moves *path forward while it finds items.
 *
 * @param path input: "CollationElements/Sequence" or "zoneStrings/3/2" etc.;
 *             output: points to the part that has not yet been processed
 */
U_CFUNC Resource res_findResource(const ResourceData *pResData, Resource r,
                                  char** path, const char** key);

#ifdef __cplusplus

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2015-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* resource.h
*
* created on: 2015nov04
* created by: Markus W. Scherer
*/


/**
 * \file
 * \brief ICU resource bundle key and value types.
 */

// Note: Ported from ICU4J class UResource and its nested classes,
// but the C++ classes are separate, not nested.

// We use the Resource prefix for C++ classes, as usual.
// The UResource prefix would be used for C types.




// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html




#if U_ENABLE_TRACING

struct UResourceBundle;

U_NAMESPACE_BEGIN

class CharString;

/**
 * Instances of this class store information used to trace reads from resource
 * bundles when ICU is built with --enable-tracing.
 *
 * All arguments of type const UResourceBundle*, const char*, and
 * const ResourceTracer& are stored as pointers. The caller must retain
 * ownership for the lifetime of this ResourceTracer.
 *
 * Exported as U_COMMON_API for Windows because it is a value field
 * in other exported types.
 */
class U_COMMON_API ResourceTracer {
public:
    ResourceTracer() :
        fResB(nullptr),
        fParent(nullptr),
        fKey(nullptr),
        fIndex(-1) {}

    ResourceTracer(const UResourceBundle* resB) :
        fResB(resB),
        fParent(nullptr),
        fKey(nullptr),
        fIndex(-1) {}

    ResourceTracer(const UResourceBundle* resB, const char* key) :
        fResB(resB),
        fParent(nullptr),
        fKey(key),
        fIndex(-1) {}

    ResourceTracer(const UResourceBundle* resB, int32_t index) :
        fResB(resB),
        fParent(nullptr),
        fKey(nullptr),
        fIndex(index) {}

    ResourceTracer(const ResourceTracer& parent, const char* key) :
        fResB(nullptr),
        fParent(&parent),
        fKey(key),
        fIndex(-1) {}

    ResourceTracer(const ResourceTracer& parent, int32_t index) :
        fResB(nullptr),
        fParent(&parent),
        fKey(nullptr),
        fIndex(index) {}

    ~ResourceTracer();

    void trace(const char* type) const;
    void traceOpen() const;

    /**
     * Calls trace() if the resB or parent provided to the constructor was
     * non-null; otherwise, does nothing.
     */
    void maybeTrace(const char* type) const {
        if (fResB || fParent) {
            trace(type);
        }
    }

private:
    const UResourceBundle* fResB;
    const ResourceTracer* fParent;
    const char* fKey;
    int32_t fIndex;

    CharString& getFilePath(CharString& output, UErrorCode& status) const;

    CharString& getResPath(CharString& output, UErrorCode& status) const;
};

/**
 * This class provides methods to trace data file reads when ICU is built
 * with --enable-tracing.
 */
class FileTracer {
public:
    static void traceOpen(const char* path, const char* type, const char* name);

private:
    static void traceOpenDataFile(const char* path, const char* type, const char* name);
    static void traceOpenResFile(const char* path, const char* name);
};

U_NAMESPACE_END

#else // U_ENABLE_TRACING

U_NAMESPACE_BEGIN

/**
 * Default trivial implementation when --enable-tracing is not used.
 */
class U_COMMON_API ResourceTracer {
public:
    ResourceTracer() {}

    ResourceTracer(const void*) {}

    ResourceTracer(const void*, const char*) {}

    ResourceTracer(const void*, int32_t) {}

    ResourceTracer(const ResourceTracer&, const char*) {}

    ResourceTracer(const ResourceTracer&, int32_t) {}

    void trace(const char*) const {}

    void traceOpen() const {}

    void maybeTrace(const char*) const {}
};

/**
 * Default trivial implementation when --enable-tracing is not used.
 */
class FileTracer {
public:
    static void traceOpen(const char*, const char*, const char*) {}
};

U_NAMESPACE_END

#endif // U_ENABLE_TRACING



struct ResourceData;

U_NAMESPACE_BEGIN

class ResourceValue;

// Note: In C++, we use const char * pointers for keys,
// rather than an abstraction like Java UResource.Key.

/**
 * Interface for iterating over a resource bundle array resource.
 */
class U_COMMON_API ResourceArray {
public:
    /** Constructs an empty array object. */
    ResourceArray() : items16(NULL), items32(NULL), length(0) {}

    /** Only for implementation use. @internal */
    ResourceArray(const uint16_t *i16, const uint32_t *i32, int32_t len,
                  const ResourceTracer& traceInfo) :
            items16(i16), items32(i32), length(len),
            fTraceInfo(traceInfo) {}

    /**
     * @return The number of items in the array resource.
     */
    int32_t getSize() const { return length; }
    /**
     * @param i Array item index.
     * @param value Output-only, receives the value of the i'th item.
     * @return TRUE if i is non-negative and less than getSize().
     */
    UBool getValue(int32_t i, ResourceValue &value) const;

    /** Only for implementation use. @internal */
    uint32_t internalGetResource(const ResourceData *pResData, int32_t i) const;

private:
    const uint16_t *items16;
    const uint32_t *items32;
    int32_t length;
    ResourceTracer fTraceInfo;
};

/**
 * Interface for iterating over a resource bundle table resource.
 */
class U_COMMON_API ResourceTable {
public:
    /** Constructs an empty table object. */
    ResourceTable() : keys16(NULL), keys32(NULL), items16(NULL), items32(NULL), length(0) {}

    /** Only for implementation use. @internal */
    ResourceTable(const uint16_t *k16, const int32_t *k32,
                  const uint16_t *i16, const uint32_t *i32, int32_t len,
                  const ResourceTracer& traceInfo) :
            keys16(k16), keys32(k32), items16(i16), items32(i32), length(len),
            fTraceInfo(traceInfo) {}

    /**
     * @return The number of items in the array resource.
     */
    int32_t getSize() const { return length; }
    /**
     * @param i Table item index.
     * @param key Output-only, receives the key of the i'th item.
     * @param value Output-only, receives the value of the i'th item.
     * @return TRUE if i is non-negative and less than getSize().
     */
    UBool getKeyAndValue(int32_t i, const char *&key, ResourceValue &value) const;

    /**
     * @param key Key string to find in the table.
     * @param value Output-only, receives the value of the item with that key.
     * @return TRUE if the table contains the key.
     */
    UBool findValue(const char *key, ResourceValue &value) const;

private:
    const uint16_t *keys16;
    const int32_t *keys32;
    const uint16_t *items16;
    const uint32_t *items32;
    int32_t length;
    ResourceTracer fTraceInfo;
};

/**
 * Represents a resource bundle item's value.
 * Avoids object creations as much as possible.
 * Mutable, not thread-safe.
 */
class U_COMMON_API ResourceValue : public UObject {
public:
    virtual ~ResourceValue();

    /**
     * @return ICU resource type, for example, URES_STRING
     */
    virtual UResType getType() const = 0;

    /**
     * Sets U_RESOURCE_TYPE_MISMATCH if this is not a string resource.
     *
     * @see ures_getString()
     */
    virtual const UChar *getString(int32_t &length, UErrorCode &errorCode) const = 0;

    inline UnicodeString getUnicodeString(UErrorCode &errorCode) const {
        int32_t len = 0;
        const UChar *r = getString(len, errorCode);
        return UnicodeString(TRUE, r, len);
    }

    /**
     * Sets U_RESOURCE_TYPE_MISMATCH if this is not an alias resource.
     */
    virtual const UChar *getAliasString(int32_t &length, UErrorCode &errorCode) const = 0;

    inline UnicodeString getAliasUnicodeString(UErrorCode &errorCode) const {
        int32_t len = 0;
        const UChar *r = getAliasString(len, errorCode);
        return UnicodeString(TRUE, r, len);
    }

    /**
     * Sets U_RESOURCE_TYPE_MISMATCH if this is not an integer resource.
     *
     * @see ures_getInt()
     */
    virtual int32_t getInt(UErrorCode &errorCode) const = 0;

    /**
     * Sets U_RESOURCE_TYPE_MISMATCH if this is not an integer resource.
     *
     * @see ures_getUInt()
     */
    virtual uint32_t getUInt(UErrorCode &errorCode) const = 0;

    /**
     * Sets U_RESOURCE_TYPE_MISMATCH if this is not an intvector resource.
     *
     * @see ures_getIntVector()
     */
    virtual const int32_t *getIntVector(int32_t &length, UErrorCode &errorCode) const = 0;

    /**
     * Sets U_RESOURCE_TYPE_MISMATCH if this is not a binary-blob resource.
     *
     * @see ures_getBinary()
     */
    virtual const uint8_t *getBinary(int32_t &length, UErrorCode &errorCode) const = 0;

    /**
     * Sets U_RESOURCE_TYPE_MISMATCH if this is not an array resource
     */
    virtual ResourceArray getArray(UErrorCode &errorCode) const = 0;

    /**
     * Sets U_RESOURCE_TYPE_MISMATCH if this is not a table resource
     */
    virtual ResourceTable getTable(UErrorCode &errorCode) const = 0;

    /**
     * Is this a no-fallback/no-inheritance marker string?
     * Such a marker is used for
     * CLDR no-fallback data values of (three empty-set symbols)=={2205, 2205, 2205}
     * when enumerating tables with fallback from the specific resource bundle to root.
     *
     * @return TRUE if this is a no-inheritance marker string
     */
    virtual UBool isNoInheritanceMarker() const = 0;

    /**
     * Sets the dest strings from the string values in this array resource.
     *
     * @return the number of strings in this array resource.
     *     If greater than capacity, then an overflow error is set.
     *
     * Sets U_RESOURCE_TYPE_MISMATCH if this is not an array resource
     *     or if any of the array items is not a string
     */
    virtual int32_t getStringArray(UnicodeString *dest, int32_t capacity,
                                   UErrorCode &errorCode) const = 0;

    /**
     * Same as
     * <pre>
     * if (getType() == URES_STRING) {
     *     return new String[] { getString(); }
     * } else {
     *     return getStringArray();
     * }
     * </pre>
     *
     * Sets U_RESOURCE_TYPE_MISMATCH if this is
     *     neither a string resource nor an array resource containing strings
     * @see getString()
     * @see getStringArray()
     */
    virtual int32_t getStringArrayOrStringAsArray(UnicodeString *dest, int32_t capacity,
                                                  UErrorCode &errorCode) const = 0;

    /**
     * Same as
     * <pre>
     * if (getType() == URES_STRING) {
     *     return getString();
     * } else {
     *     return getStringArray()[0];
     * }
     * </pre>
     *
     * Sets U_RESOURCE_TYPE_MISMATCH if this is
     *     neither a string resource nor an array resource containing strings
     * @see getString()
     * @see getStringArray()
     */
    virtual UnicodeString getStringOrFirstOfArray(UErrorCode &errorCode) const = 0;

protected:
    ResourceValue() {}

private:
    ResourceValue(const ResourceValue &);  // no copy constructor
    ResourceValue &operator=(const ResourceValue &);  // no assignment operator
};

/**
 * Sink for ICU resource bundle contents.
 */
class U_COMMON_API ResourceSink : public UObject {
public:
    ResourceSink() {}
    virtual ~ResourceSink();

    /**
     * Called once for each bundle (child-parent-...-root).
     * The value is normally an array or table resource,
     * and implementations of this method normally iterate over the
     * tree of resource items stored there.
     *
     * @param key The key string of the enumeration-start resource.
     *     Empty if the enumeration starts at the top level of the bundle.
     * @param value Call getArray() or getTable() as appropriate.
     *     Then reuse for output values from Array and Table getters.
     * @param noFallback true if the bundle has no parent;
     *     that is, its top-level table has the nofallback attribute,
     *     or it is the root bundle of a locale tree.
     */
    virtual void put(const char *key, ResourceValue &value, UBool noFallback,
                     UErrorCode &errorCode) = 0;

private:
    ResourceSink(const ResourceSink &);  // no copy constructor
    ResourceSink &operator=(const ResourceSink &);  // no assignment operator
};

U_NAMESPACE_END




U_NAMESPACE_BEGIN

inline const UChar* res_getString(const ResourceTracer& traceInfo,
        const ResourceData *pResData, Resource res, int32_t *pLength) {
    traceInfo.trace("string");
    return res_getStringNoTrace(pResData, res, pLength);
}

inline const uint8_t* res_getBinary(const ResourceTracer& traceInfo,
        const ResourceData *pResData, Resource res, int32_t *pLength) {
    traceInfo.trace("binary");
    return res_getBinaryNoTrace(pResData, res, pLength);
}

inline const int32_t* res_getIntVector(const ResourceTracer& traceInfo,
        const ResourceData *pResData, Resource res, int32_t *pLength) {
    traceInfo.trace("intvector");
    return res_getIntVectorNoTrace(pResData, res, pLength);
}

inline int32_t res_getInt(const ResourceTracer& traceInfo, Resource res) {
    traceInfo.trace("int");
    return RES_GET_INT_NO_TRACE(res);
}

inline uint32_t res_getUInt(const ResourceTracer& traceInfo, Resource res) {
    traceInfo.trace("uint");
    return RES_GET_UINT_NO_TRACE(res);
}

class ResourceDataValue : public ResourceValue {
public:
    ResourceDataValue() :
        res(static_cast<Resource>(URES_NONE)),
        fTraceInfo() {}
    virtual ~ResourceDataValue();

    void setData(const ResourceData *data) {
        resData = *data;
    }

    void setResource(Resource r, ResourceTracer&& traceInfo) {
        res = r;
        fTraceInfo = traceInfo;
    }

    const ResourceData &getData() const { return resData; }
    virtual UResType getType() const;
    virtual const UChar *getString(int32_t &length, UErrorCode &errorCode) const;
    virtual const UChar *getAliasString(int32_t &length, UErrorCode &errorCode) const;
    virtual int32_t getInt(UErrorCode &errorCode) const;
    virtual uint32_t getUInt(UErrorCode &errorCode) const;
    virtual const int32_t *getIntVector(int32_t &length, UErrorCode &errorCode) const;
    virtual const uint8_t *getBinary(int32_t &length, UErrorCode &errorCode) const;
    virtual ResourceArray getArray(UErrorCode &errorCode) const;
    virtual ResourceTable getTable(UErrorCode &errorCode) const;
    virtual UBool isNoInheritanceMarker() const;
    virtual int32_t getStringArray(UnicodeString *dest, int32_t capacity,
                                   UErrorCode &errorCode) const;
    virtual int32_t getStringArrayOrStringAsArray(UnicodeString *dest, int32_t capacity,
                                                  UErrorCode &errorCode) const;
    virtual UnicodeString getStringOrFirstOfArray(UErrorCode &errorCode) const;

private:
    // TODO(ICU-20769): If UResourceBundle.fResData becomes a pointer,
    // then remove this value field again and just store a pResData pointer.
    ResourceData resData;
    Resource res;
    ResourceTracer fTraceInfo;
};

U_NAMESPACE_END

#endif  /* __cplusplus */

/**
 * Swap an ICU resource bundle. See udataswp.h.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
ures_swap(const UDataSwapper *ds,
          const void *inData, int32_t length, void *outData,
          UErrorCode *pErrorCode);



#define kRootLocaleName         "root"
#define kPoolBundleName         "pool"

/*
 The default minor version and the version separator must be exactly one
 character long.
*/

#define kDefaultMinorVersion    "0"
#define kVersionSeparator       "."
#define kVersionTag             "Version"

#define MAGIC1 19700503
#define MAGIC2 19641227

#define URES_MAX_ALIAS_LEVEL 256
#define URES_MAX_BUFFER_SIZE 256

#define EMPTY_SET 0x2205

struct UResourceDataEntry;
typedef struct UResourceDataEntry UResourceDataEntry;

/*
 * Note: If we wanted to make this structure smaller, then we could try
 * to use one UResourceDataEntry pointer for fAlias and fPool, with a separate
 * flag to distinguish whether this struct is for a real bundle with a pool,
 * or for an alias entry for which we won't use the pool after loading.
 */
struct UResourceDataEntry {
    char *fName; /* name of the locale for bundle - still to decide whether it is original or fallback */
    char *fPath; /* path to bundle - used for distinguishing between resources with the same name */
    UResourceDataEntry *fParent; /*next resource in fallback chain*/
    UResourceDataEntry *fAlias;
    UResourceDataEntry *fPool;
    ResourceData fData; /* data for low level access */
    char fNameBuffer[3]; /* A small buffer of free space for fName. The free space is due to struct padding. */
    uint32_t fCountExisting; /* how much is this resource used */
    UErrorCode fBogus;
    /* int32_t fHashKey;*/ /* for faster access in the hashtable */
};

#define RES_BUFSIZE 64
#define RES_PATH_SEPARATOR   '/'
#define RES_PATH_SEPARATOR_S   "/"

struct UResourceBundle {
    const char *fKey; /*tag*/
    UResourceDataEntry *fData; /*for low-level access*/
    char *fVersion;
    UResourceDataEntry *fTopLevelData; /* for getting the valid locale */
    char *fResPath; /* full path to the resource: "zh_TW/CollationElements/Sequence" */
    // TODO(ICU-20769): Try to change the by-value fResData into a pointer,
    // with the struct in only one place for each bundle.
    // Also replace class ResourceDataValue.resData with a pResData pointer again.
    ResourceData fResData;
    char fResBuf[RES_BUFSIZE];
    int32_t fResPathLen;
    Resource fRes;
    UBool fHasFallback;
    UBool fIsTopLevel;
    uint32_t fMagic1;   /* For determining if it's a stack object */
    uint32_t fMagic2;   /* For determining if it's a stack object */
    int32_t fIndex;
    int32_t fSize;

    /*const UResourceBundle *fParentRes;*/ /* needed to get the actual locale for a child resource */
};

U_CAPI void U_EXPORT2 ures_initStackObject(UResourceBundle* resB);

#ifdef __cplusplus

U_NAMESPACE_BEGIN

/**
 * \class StackUResourceBundle
 * "Smart pointer" like class, closes a UResourceBundle via ures_close().
 *
 * This code:
 *
 *   StackUResourceBundle bundle;
 *   foo(bundle.getAlias());
 *
 * Is equivalent to this code:
 *
 *   UResourceBundle bundle;
 *   ures_initStackObject(&bundle);
 *   foo(&bundle);
 *   ures_close(&bundle);
 *
 * @see LocalUResourceBundlePointer
 * @internal
 */
class U_COMMON_API StackUResourceBundle {
public:
    // No heap allocation. Use only on the stack.
    static void* U_EXPORT2 operator new(size_t) U_NOEXCEPT = delete;
    static void* U_EXPORT2 operator new[](size_t) U_NOEXCEPT = delete;
#if U_HAVE_PLACEMENT_NEW
    static void* U_EXPORT2 operator new(size_t, void*) U_NOEXCEPT = delete;
#endif

    StackUResourceBundle();
    ~StackUResourceBundle();

    UResourceBundle* getAlias() { return &bundle; }

    UResourceBundle& ref() { return bundle; }
    const UResourceBundle& ref() const { return bundle; }

    StackUResourceBundle(const StackUResourceBundle&) = delete;
    StackUResourceBundle& operator=(const StackUResourceBundle&) = delete;

    StackUResourceBundle(StackUResourceBundle&&) = delete;
    StackUResourceBundle& operator=(StackUResourceBundle&&) = delete;

private:
    UResourceBundle bundle;
};

U_NAMESPACE_END

#endif  /* __cplusplus */

/**
 * Opens a resource bundle for the locale;
 * if there is not even a base language bundle, then loads the root bundle;
 * never falls back to the default locale.
 *
 * This is used for algorithms that have good pan-Unicode default behavior,
 * such as case mappings, collation, and segmentation (BreakIterator).
 */
U_CAPI UResourceBundle* U_EXPORT2
ures_openNoDefault(const char* path, const char* localeID, UErrorCode* status);

/* Some getters used by the copy constructor */
U_CFUNC const char* ures_getName(const UResourceBundle* resB);
#ifdef URES_DEBUG
U_CFUNC const char* ures_getPath(const UResourceBundle* resB);
/**
 * If anything was in the RB cache, dump it to the screen.
 * @return TRUE if there was anything into the cache
 */
U_CAPI UBool U_EXPORT2 ures_dumpCacheContents(void);
#endif
/*U_CFUNC void ures_appendResPath(UResourceBundle *resB, const char* toAdd, int32_t lenToAdd);*/
/*U_CFUNC void ures_setResPath(UResourceBundle *resB, const char* toAdd);*/
/*U_CFUNC void ures_freeResPath(UResourceBundle *resB);*/

/* Candidates for export */
U_CFUNC UResourceBundle *ures_copyResb(UResourceBundle *r, const UResourceBundle *original, UErrorCode *status);

/**
 * Returns a resource that can be located using the pathToResource argument. One needs optional package, locale
 * and path inside the locale, for example: "/myData/en/zoneStrings/3". Keys and indexes are supported. Keys
 * need to reference data in named structures, while indexes can reference both named and anonymous resources.
 * Features a fill-in parameter. 
 * 
 * Note, this function does NOT have a syntax for specifying items within a tree.  May want to consider a
 * syntax that delineates between package/tree and resource.  
 *
 * @param pathToResource    a path that will lead to the requested resource
 * @param fillIn            if NULL a new UResourceBundle struct is allocated and must be deleted by the caller.
 *                          Alternatively, you can supply a struct to be filled by this function.
 * @param status            fills in the outgoing error code.
 * @return                  a pointer to a UResourceBundle struct. If fill in param was NULL, caller must delete it
 */
U_CAPI UResourceBundle* U_EXPORT2
ures_findResource(const char* pathToResource, 
                  UResourceBundle *fillIn, UErrorCode *status); 

/**
 * Returns a sub resource that can be located using the pathToResource argument. One needs a path inside 
 * the supplied resource, for example, if you have "en_US" resource bundle opened, you might ask for
 * "zoneStrings/3". Keys and indexes are supported. Keys
 * need to reference data in named structures, while indexes can reference both 
 * named and anonymous resources.
 * Features a fill-in parameter. 
 *
 * @param resourceBundle    a resource
 * @param pathToResource    a path that will lead to the requested resource
 * @param fillIn            if NULL a new UResourceBundle struct is allocated and must be deleted by the caller.
 *                          Alternatively, you can supply a struct to be filled by this function.
 * @param status            fills in the outgoing error code.
 * @return                  a pointer to a UResourceBundle struct. If fill in param was NULL, caller must delete it
 */
U_CAPI UResourceBundle* U_EXPORT2
ures_findSubResource(const UResourceBundle *resB, 
                     char* pathToResource, 
                     UResourceBundle *fillIn, UErrorCode *status);

/**
 * Returns a functionally equivalent locale (considering keywords) for the specified keyword.
 * @param result fillin for the equivalent locale
 * @param resultCapacity capacity of the fillin buffer
 * @param path path to the tree, or NULL for ICU data
 * @param resName top level resource. Example: "collations"
 * @param keyword locale keyword. Example: "collation"
 * @param locid The requested locale
 * @param isAvailable If non-null, pointer to fillin parameter that indicates whether the 
 * requested locale was available. The locale is defined as 'available' if it physically 
 * exists within the specified tree.
 * @param omitDefault if TRUE, omit keyword and value if default. 'de_DE\@collation=standard' -> 'de_DE'
 * @param status error code
 * @return  the actual buffer size needed for the full locale.  If it's greater 
 * than resultCapacity, the returned full name will be truncated and an error code will be returned.
 */
U_CAPI int32_t U_EXPORT2
ures_getFunctionalEquivalent(char *result, int32_t resultCapacity, 
                             const char *path, const char *resName, const char *keyword, const char *locid,
                             UBool *isAvailable, UBool omitDefault, UErrorCode *status);

/**
 * Given a tree path and keyword, return a string enumeration of all possible values for that keyword.
 * @param path path to the tree, or NULL for ICU data
 * @param keyword a particular keyword to consider, must match a top level resource name 
 * within the tree.
 * @param status error code
 */
U_CAPI UEnumeration* U_EXPORT2
ures_getKeywordValues(const char *path, const char *keyword, UErrorCode *status);


/**
 * Get a resource with multi-level fallback. Normally only the top level resources will
 * fallback to its parent. This performs fallback on subresources. For example, when a table
 * is defined in a resource bundle and a parent resource bundle, normally no fallback occurs
 * on the sub-resources because the table is defined in the current resource bundle, but this
 * function can perform fallback on the sub-resources of the table.
 * @param resB              a resource
 * @param inKey             a key associated with the requested resource
 * @param fillIn            if NULL a new UResourceBundle struct is allocated and must be deleted by the caller.
 *                          Alternatively, you can supply a struct to be filled by this function.
 * @param status: fills in the outgoing error code
 *                could be <TT>U_MISSING_RESOURCE_ERROR</TT> if the key is not found
 *                could be a non-failing error 
 *                e.g.: <TT>U_USING_FALLBACK_WARNING</TT>,<TT>U_USING_DEFAULT_WARNING </TT>
 * @return                  a pointer to a UResourceBundle struct. If fill in param was NULL, caller must delete it
 */
U_CAPI UResourceBundle* U_EXPORT2 
ures_getByKeyWithFallback(const UResourceBundle *resB, 
                          const char* inKey, 
                          UResourceBundle *fillIn, 
                          UErrorCode *status);


/**
 * Get a String with multi-level fallback. Normally only the top level resources will
 * fallback to its parent. This performs fallback on subresources. For example, when a table
 * is defined in a resource bundle and a parent resource bundle, normally no fallback occurs
 * on the sub-resources because the table is defined in the current resource bundle, but this
 * function can perform fallback on the sub-resources of the table.
 * @param resB              a resource
 * @param inKey             a key associated with the requested resource
 * @param status: fills in the outgoing error code
 *                could be <TT>U_MISSING_RESOURCE_ERROR</TT> if the key is not found
 *                could be a non-failing error 
 *                e.g.: <TT>U_USING_FALLBACK_WARNING</TT>,<TT>U_USING_DEFAULT_WARNING </TT>
 * @return                  a pointer to a UResourceBundle struct. If fill in param was NULL, caller must delete it
 */
U_CAPI const UChar* U_EXPORT2 
ures_getStringByKeyWithFallback(const UResourceBundle *resB, 
                          const char* inKey,  
                          int32_t* len,
                          UErrorCode *status);

#ifdef __cplusplus

U_CAPI void U_EXPORT2
ures_getValueWithFallback(const UResourceBundle *bundle, const char *path,
                          UResourceBundle *tempFillIn,
                          icu::ResourceDataValue &value, UErrorCode &errorCode);

U_CAPI void U_EXPORT2
ures_getAllItemsWithFallback(const UResourceBundle *bundle, const char *path,
                             icu::ResourceSink &sink, UErrorCode &errorCode);

#endif  /* __cplusplus */

/**
 * Get a version number by key
 * @param resB bundle containing version number
 * @param key the key for the version number
 * @param ver fillin for the version number
 * @param status error code
 */
U_CAPI void U_EXPORT2
ures_getVersionByKey(const UResourceBundle *resB,
                     const char *key,
                     UVersionInfo ver,
                     UErrorCode *status);


/**
 * Internal function.
 * Return the version number associated with this ResourceBundle as a string.
 *
 * @param resourceBundle The resource bundle for which the version is checked.
 * @return  A version number string as specified in the resource bundle or its parent.
 *          The caller does not own this string.
 * @see ures_getVersion
 */
U_CAPI const char* U_EXPORT2 
ures_getVersionNumberInternal(const UResourceBundle *resourceBundle);

/**
 * Return the name of the Locale associated with this ResourceBundle. This API allows
 * you to query for the real locale of the resource. For example, if you requested 
 * "en_US_CALIFORNIA" and only "en_US" bundle exists, "en_US" will be returned. 
 * For subresources, the locale where this resource comes from will be returned.
 * If fallback has occured, getLocale will reflect this.
 *
 * This internal version avoids deprecated-warnings in ICU code.
 *
 * @param resourceBundle resource bundle in question
 * @param status just for catching illegal arguments
 * @return  A Locale name
 */
U_CAPI const char* U_EXPORT2 
ures_getLocaleInternal(const UResourceBundle* resourceBundle, 
               UErrorCode* status);

/**
 * Same as ures_openDirect() but uses the fill-in parameter instead of allocating a new bundle.
 * 
 * @param r The existing UResourceBundle to fill in. If NULL then status will be
 *          set to U_ILLEGAL_ARGUMENT_ERROR.
 * @param packageName   The packageName and locale together point to an ICU udata object,
 *                      as defined by <code> udata_open( packageName, "res", locale, err) </code>
 *                      or equivalent.  Typically, packageName will refer to a (.dat) file, or to
 *                      a package registered with udata_setAppData(). Using a full file or directory
 *                      pathname for packageName is deprecated. If NULL, ICU data will be used.
 * @param locale  specifies the locale for which we want to open the resource
 *                if NULL, the default locale will be used. If strlen(locale) == 0
 *                root locale will be used.
 * @param status The error code.
 * @see ures_openDirect
 * @internal
 */
U_CAPI void U_EXPORT2
ures_openDirectFillIn(UResourceBundle *r,
                      const char *packageName,
                      const char *locale,
                      UErrorCode *status);

 /* for ures_getVersionByKey */

U_CAPI void U_EXPORT2 u_getDataVersion(UVersionInfo dataVersionFillin, UErrorCode *status) {
    UResourceBundle *icudatares = NULL;
    
    if (U_FAILURE(*status)) {
        return;
    }
    
    if (dataVersionFillin != NULL) {
        icudatares = ures_openDirect(NULL, U_ICU_VERSION_BUNDLE , status);
        if (U_SUCCESS(*status)) {
            ures_getVersionByKey(icudatares, U_ICU_DATA_KEY, dataVersionFillin, status);
        }
        ures_close(icudatares);
    }
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2009-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
*  FILE NAME : icuplug.c
*
*   Date         Name        Description
*   10/29/2009   sl          New.
******************************************************************************
*/

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2009-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
*  FILE NAME : icuplug.h
*
*   Date         Name        Description
*   10/29/2009   sl          New.
******************************************************************************
*/

/**
 * \file
 * \brief C API: ICU Plugin API 
 *
 * <h2>C API: ICU Plugin API</h2>
 *
 * <p>C API allowing run-time loadable modules that extend or modify ICU functionality.</p>
 *
 * <h3>Loading and Configuration</h3>
 *
 * <p>At ICU startup time, the environment variable "ICU_PLUGINS" will be 
 * queried for a directory name.  If it is not set, the preprocessor symbol 
 * "DEFAULT_ICU_PLUGINS" will be checked for a default value.</p>
 *
 * <p>Within the above-named directory, the file  "icuplugins##.txt" will be 
 * opened, if present, where ## is the major+minor number of the currently 
 * running ICU (such as, 44 for ICU 4.4, thus icuplugins44.txt)</p>
 *
 * <p>The configuration file has this format:</p>
 *
 * <ul>
 * <li>Hash (#) begins a comment line</li>
 * 
 * <li>Non-comment lines have two or three components:
 * LIBRARYNAME     ENTRYPOINT     [ CONFIGURATION .. ]</li>
 *
 * <li>Tabs or spaces separate the three items.</li>
 *
 * <li>LIBRARYNAME is the name of a shared library, either a short name if 
 * it is on the loader path,  or a full pathname.</li>
 *
 * <li>ENTRYPOINT is the short (undecorated) symbol name of the plugin's 
 * entrypoint, as above.</li>
 *
 * <li>CONFIGURATION is the entire rest of the line . It's passed as-is to 
 * the plugin.</li>
 * </ul>
 *
 * <p>An example configuration file is, in its entirety:</p>
 *
 * \code
 * # this is icuplugins44.txt
 * testplug.dll    myPlugin        hello=world
 * \endcode
 * <p>Plugins are categorized as "high" or "low" level.  Low level are those 
 * which must be run BEFORE high level plugins, and before any operations 
 * which cause ICU to be 'initialized'.  If a plugin is low level but 
 * causes ICU to allocate memory or become initialized, that plugin is said 
 * to cause a 'level change'. </p>
 *
 * <p>At load time, ICU first queries all plugins to determine their level, 
 * then loads all 'low' plugins first, and then loads all 'high' plugins.  
 * Plugins are otherwise loaded in the order listed in the configuration file.</p>
 * 
 * <h3>Implementing a Plugin</h3>
 * \code
 * U_CAPI UPlugTokenReturn U_EXPORT2 
 * myPlugin (UPlugData *plug, UPlugReason reason, UErrorCode *status) {
 *   if(reason==UPLUG_REASON_QUERY) {
 *      uplug_setPlugName(plug, "Simple Plugin");
 *      uplug_setPlugLevel(plug, UPLUG_LEVEL_HIGH);
 *    } else if(reason==UPLUG_REASON_LOAD) {
 *       ... Set up some ICU things here.... 
 *    } else if(reason==UPLUG_REASON_UNLOAD) {
 *       ... unload, clean up ...
 *    }
 *   return UPLUG_TOKEN;
 *  }
 * \endcode
 *
 * <p>The UPlugData*  is an opaque pointer to the plugin-specific data, and is 
 * used in all other API calls.</p>
 *
 * <p>The API contract is:</p>
 * <ol><li>The plugin MUST always return UPLUG_TOKEN as a return value- to 
 * indicate that it is a valid plugin.</li>
 *
 * <li>When the 'reason' parameter is set to UPLUG_REASON_QUERY,  the 
 * plugin MUST call uplug_setPlugLevel() to indicate whether it is a high 
 * level or low level plugin.</li>
 *
 * <li>When the 'reason' parameter is UPLUG_REASON_QUERY, the plugin 
 * SHOULD call uplug_setPlugName to indicate a human readable plugin name.</li></ol>
 * 
 *
 * \internal ICU 4.4 Technology Preview
 */






#if UCONFIG_ENABLE_PLUGINS || defined(U_IN_DOXYGEN)



/* === Basic types === */

#ifndef U_HIDE_INTERNAL_API
/**
 * @{
 * Opaque structure passed to/from a plugin. 
 * use the APIs to access it.
 * @internal ICU 4.4 Technology Preview
 */

struct UPlugData;
typedef struct UPlugData UPlugData;

/** @} */

/**
 * Random Token to identify a valid ICU plugin. Plugins must return this 
 * from the entrypoint.
 * @internal ICU 4.4 Technology Preview
 */
#define UPLUG_TOKEN 0x54762486

/**
 * Max width of names, symbols, and configuration strings
 * @internal ICU 4.4 Technology Preview
 */
#define UPLUG_NAME_MAX              100


/**
 * Return value from a plugin entrypoint. 
 * Must always be set to UPLUG_TOKEN
 * @see UPLUG_TOKEN
 * @internal ICU 4.4 Technology Preview
 */
typedef uint32_t UPlugTokenReturn;

/**
 * Reason code for the entrypoint's call
 * @internal ICU 4.4 Technology Preview
 */
typedef enum {
    UPLUG_REASON_QUERY = 0,     /**< The plugin is being queried for info. **/
    UPLUG_REASON_LOAD = 1,     /**< The plugin is being loaded. **/
    UPLUG_REASON_UNLOAD = 2,   /**< The plugin is being unloaded. **/
    /**
     * Number of known reasons.
     * @internal The numeric value may change over time, see ICU ticket #12420.
     */
    UPLUG_REASON_COUNT
} UPlugReason;


/**
 * Level of plugin loading
 *     INITIAL:  UNKNOWN
 *       QUERY:   INVALID ->  { LOW | HIGH }
 *     ERR -> INVALID
 * @internal ICU 4.4 Technology Preview
 */
typedef enum {
    UPLUG_LEVEL_INVALID = 0,     /**< The plugin is invalid, hasn't called uplug_setLevel, or can't load. **/
    UPLUG_LEVEL_UNKNOWN = 1,     /**< The plugin is waiting to be installed. **/
    UPLUG_LEVEL_LOW     = 2,     /**< The plugin must be called before u_init completes **/
    UPLUG_LEVEL_HIGH    = 3,     /**< The plugin can run at any time. **/
    /**
     * Number of known levels.
     * @internal The numeric value may change over time, see ICU ticket #12420.
     */
    UPLUG_LEVEL_COUNT
} UPlugLevel;

/**
 * Entrypoint for an ICU plugin.
 * @param plug the UPlugData handle. 
 * @param status the plugin's extended status code.
 * @return A valid plugin must return UPLUG_TOKEN
 * @internal ICU 4.4 Technology Preview
 */
typedef UPlugTokenReturn (U_EXPORT2 UPlugEntrypoint) (
                  UPlugData *plug,
                  UPlugReason reason,
                  UErrorCode *status);

/* === Needed for Implementing === */

/**
 * Request that this plugin not be unloaded at cleanup time.
 * This is appropriate for plugins which cannot be cleaned up.
 * @see u_cleanup()
 * @param plug plugin
 * @param dontUnload  set true if this plugin can't be unloaded
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL void U_EXPORT2 
uplug_setPlugNoUnload(UPlugData *plug, UBool dontUnload);

/**
 * Set the level of this plugin.
 * @param plug plugin data handle
 * @param level the level of this plugin
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL void U_EXPORT2
uplug_setPlugLevel(UPlugData *plug, UPlugLevel level);

/**
 * Get the level of this plugin.
 * @param plug plugin data handle
 * @return the level of this plugin
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL UPlugLevel U_EXPORT2
uplug_getPlugLevel(UPlugData *plug);

/**
 * Get the lowest level of plug which can currently load.
 * For example, if UPLUG_LEVEL_LOW is returned, then low level plugins may load
 * if UPLUG_LEVEL_HIGH is returned, then only high level plugins may load.
 * @return the lowest level of plug which can currently load
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL UPlugLevel U_EXPORT2
uplug_getCurrentLevel(void);


/**
 * Get plug load status
 * @return The error code of this plugin's load attempt.
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL UErrorCode U_EXPORT2
uplug_getPlugLoadStatus(UPlugData *plug); 

/**
 * Set the human-readable name of this plugin.
 * @param plug plugin data handle
 * @param name the name of this plugin. The first UPLUG_NAME_MAX characters willi be copied into a new buffer.
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL void U_EXPORT2
uplug_setPlugName(UPlugData *plug, const char *name);

/**
 * Get the human-readable name of this plugin.
 * @param plug plugin data handle
 * @return the name of this plugin
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL const char * U_EXPORT2
uplug_getPlugName(UPlugData *plug);

/**
 * Return the symbol name for this plugin, if known.
 * @param plug plugin data handle
 * @return the symbol name, or NULL
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL const char * U_EXPORT2
uplug_getSymbolName(UPlugData *plug);

/**
 * Return the library name for this plugin, if known.
 * @param plug plugin data handle
 * @param status error code
 * @return the library name, or NULL
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL const char * U_EXPORT2
uplug_getLibraryName(UPlugData *plug, UErrorCode *status);

/**
 * Return the library used for this plugin, if known.
 * Plugins could use this to load data out of their 
 * @param plug plugin data handle
 * @return the library, or NULL
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL void * U_EXPORT2
uplug_getLibrary(UPlugData *plug);

/**
 * Return the plugin-specific context data.
 * @param plug plugin data handle
 * @return the context, or NULL if not set
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL void * U_EXPORT2
uplug_getContext(UPlugData *plug);

/**
 * Set the plugin-specific context data.
 * @param plug plugin data handle
 * @param context new context to set
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL void U_EXPORT2
uplug_setContext(UPlugData *plug, void *context);


/**
 * Get the configuration string, if available.
 * The string is in the platform default codepage.
 * @param plug plugin data handle
 * @return configuration string, or else null.
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL const char * U_EXPORT2
uplug_getConfiguration(UPlugData *plug);

/**
 * Return all currently installed plugins, from newest to oldest
 * Usage Example:
 * \code
 *    UPlugData *plug = NULL;
 *    while(plug=uplug_nextPlug(plug)) {
 *        ... do something with 'plug' ...
 *    }
 * \endcode
 * Not thread safe- do not call while plugs are added or removed.
 * @param prior pass in 'NULL' to get the first (most recent) plug, 
 *  otherwise pass the value returned on a prior call to uplug_nextPlug
 * @return the next oldest plugin, or NULL if no more.
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL UPlugData* U_EXPORT2
uplug_nextPlug(UPlugData *prior);

/**
 * Inject a plugin as if it were loaded from a library.
 * This is useful for testing plugins. 
 * Note that it will have a 'NULL' library pointer associated
 * with it, and therefore no llibrary will be closed at cleanup time.
 * Low level plugins may not be able to load, as ordering can't be enforced.
 * @param entrypoint entrypoint to install
 * @param config user specified configuration string, if available, or NULL.
 * @param status error result
 * @return the new UPlugData associated with this plugin, or NULL if error.
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL UPlugData* U_EXPORT2
uplug_loadPlugFromEntrypoint(UPlugEntrypoint *entrypoint, const char *config, UErrorCode *status);


/**
 * Inject a plugin from a library, as if the information came from a config file.
 * Low level plugins may not be able to load, and ordering can't be enforced.
 * @param libName DLL name to load
 * @param sym symbol of plugin (UPlugEntrypoint function)
 * @param config configuration string, or NULL
 * @param status error result
 * @return the new UPlugData associated with this plugin, or NULL if error.
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL UPlugData* U_EXPORT2
uplug_loadPlugFromLibrary(const char *libName, const char *sym, const char *config, UErrorCode *status);

/**
 * Remove a plugin. 
 * Will request the plugin to be unloaded, and close the library if needed
 * @param plug plugin handle to close
 * @param status error result
 * @internal ICU 4.4 Technology Preview
 */
U_INTERNAL void U_EXPORT2
uplug_removePlug(UPlugData *plug, UErrorCode *status);
#endif  /* U_HIDE_INTERNAL_API */

#endif /* UCONFIG_ENABLE_PLUGINS */




#if UCONFIG_ENABLE_PLUGINS


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2009-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
*  FILE NAME : icuplugimp.h
* 
*  Internal functions for the ICU plugin system
*
*   Date         Name        Description
*   10/29/2009   sl          New.
******************************************************************************
*/





#if UCONFIG_ENABLE_PLUGINS

/*========================*/
/** @{ Library Manipulation  
 */

/**
 * Open a library, adding a reference count if needed.
 * @param libName library name to load
 * @param status error code
 * @return the library pointer, or NULL
 * @internal internal use only
 */
U_INTERNAL void * U_EXPORT2
uplug_openLibrary(const char *libName, UErrorCode *status);

/**
 * Close a library, if its reference count is 0
 * @param lib the library to close
 * @param status error code
 * @internal internal use only
 */
U_INTERNAL void U_EXPORT2
uplug_closeLibrary(void *lib, UErrorCode *status);

/**
 * Get a library's name, or NULL if not found.
 * @param lib the library's name
 * @param status error code
 * @return the library name, or NULL if not found.
 * @internal internal use only
 */
U_INTERNAL  char * U_EXPORT2
uplug_findLibrary(void *lib, UErrorCode *status);

/** @} */

/*========================*/
/** {@ ICU Plugin internal interfaces
 */

/**
 * Initialize the plugins 
 * @param status error result
 * @internal - Internal use only.
 */
U_INTERNAL void U_EXPORT2
uplug_init(UErrorCode *status);

/**
 * Get raw plug N
 * @internal - Internal use only
 */ 
U_INTERNAL UPlugData* U_EXPORT2
uplug_getPlugInternal(int32_t n);

/**
 * Get the name of the plugin file. 
 * @internal - Internal use only.
 */
U_INTERNAL const char* U_EXPORT2
uplug_getPluginFile(void);

/** @} */

#endif






#include <stdio.h>
#ifdef __MVS__  /* defined by z/OS compiler */
#define _POSIX_SOURCE
#include <cics.h> /* 12 Nov 2011 JAM iscics() function */
#endif


using namespace icu;

#ifndef UPLUG_TRACE
#define UPLUG_TRACE 0
#endif

#if UPLUG_TRACE
#include <stdio.h>
#define DBG(x) fprintf(stderr, "%s:%d: ",__FILE__,__LINE__); fprintf x
#endif

/**
 * Internal structure of an ICU plugin. 
 */

struct UPlugData {
  UPlugEntrypoint  *entrypoint; /**< plugin entrypoint */
  uint32_t structSize;    /**< initialized to the size of this structure */
  uint32_t token;         /**< must be U_PLUG_TOKEN */
  void *lib;              /**< plugin library, or NULL */
  char libName[UPLUG_NAME_MAX];   /**< library name */
  char sym[UPLUG_NAME_MAX];        /**< plugin symbol, or NULL */
  char config[UPLUG_NAME_MAX];     /**< configuration data */
  void *context;          /**< user context data */
  char name[UPLUG_NAME_MAX];   /**< name of plugin */
  UPlugLevel  level; /**< level of plugin */
  UBool   awaitingLoad; /**< TRUE if the plugin is awaiting a load call */
  UBool   dontUnload; /**< TRUE if plugin must stay resident (leak plugin and lib) */
  UErrorCode pluginStatus; /**< status code of plugin */
};



#define UPLUG_LIBRARY_INITIAL_COUNT 8
#define UPLUG_PLUGIN_INITIAL_COUNT 12

/**
 * Remove an item
 * @param list the full list
 * @param listSize the number of entries in the list
 * @param memberSize the size of one member
 * @param itemToRemove the item number of the member
 * @return the new listsize 
 */
static int32_t uplug_removeEntryAt(void *list, int32_t listSize, int32_t memberSize, int32_t itemToRemove) {
  uint8_t *bytePtr = (uint8_t *)list;
    
  /* get rid of some bad cases first */
  if(listSize<1) {
    return listSize;
  }
    
  /* is there anything to move? */
  if(listSize > itemToRemove+1) {
    memmove(bytePtr+(itemToRemove*memberSize), bytePtr+((itemToRemove+1)*memberSize), memberSize);
  }
    
  return listSize-1;
}




#if U_ENABLE_DYLOAD
/**
 * Library management. Internal. 
 * @internal
 */
struct UPlugLibrary;

/**
 * Library management. Internal. 
 * @internal
 */
typedef struct UPlugLibrary {
  void *lib;                           /**< library ptr */
  char name[UPLUG_NAME_MAX]; /**< library name */
  uint32_t ref;                        /**< reference count */
} UPlugLibrary;

static UPlugLibrary   staticLibraryList[UPLUG_LIBRARY_INITIAL_COUNT];
static UPlugLibrary * libraryList = staticLibraryList;
static int32_t libraryCount = 0;
static int32_t libraryMax = UPLUG_LIBRARY_INITIAL_COUNT;

/**
 * Search for a library. Doesn't lock
 * @param libName libname to search for
 * @return the library's struct
 */
static int32_t searchForLibraryName(const char *libName) {
  int32_t i;
    
  for(i=0;i<libraryCount;i++) {
    if(!uprv_strcmp(libName, libraryList[i].name)) {
      return i;
    }
  }
  return -1;
}

static int32_t searchForLibrary(void *lib) {
  int32_t i;
    
  for(i=0;i<libraryCount;i++) {
    if(lib==libraryList[i].lib) {
      return i;
    }
  }
  return -1;
}

U_INTERNAL char * U_EXPORT2
uplug_findLibrary(void *lib, UErrorCode *status) {
  int32_t libEnt;
  char *ret = NULL;
  if(U_FAILURE(*status)) {
    return NULL;
  }
  libEnt = searchForLibrary(lib);
  if(libEnt!=-1) { 
    ret = libraryList[libEnt].name;
  } else {
    *status = U_MISSING_RESOURCE_ERROR;
  }
  return ret;
}

U_INTERNAL void * U_EXPORT2
uplug_openLibrary(const char *libName, UErrorCode *status) {
  int32_t libEntry = -1;
  void *lib = NULL;
    
  if(U_FAILURE(*status)) return NULL;

  libEntry = searchForLibraryName(libName);
  if(libEntry == -1) {
    libEntry = libraryCount++;
    if(libraryCount >= libraryMax) {
      /* Ran out of library slots. Statically allocated because we can't depend on allocating memory.. */
      *status = U_MEMORY_ALLOCATION_ERROR;
#if UPLUG_TRACE
      DBG((stderr, "uplug_openLibrary() - out of library slots (max %d)\n", libraryMax));
#endif
      return NULL;
    }
    /* Some operating systems don't want 
       DL operations from multiple threads. */
    libraryList[libEntry].lib = uprv_dl_open(libName, status);
#if UPLUG_TRACE
    DBG((stderr, "uplug_openLibrary(%s,%s) libEntry %d, lib %p\n", libName, u_errorName(*status), libEntry, lib));
#endif
        
    if(libraryList[libEntry].lib == NULL || U_FAILURE(*status)) {
      /* cleanup. */
      libraryList[libEntry].lib = NULL; /* failure with open */
      libraryList[libEntry].name[0] = 0;
#if UPLUG_TRACE
      DBG((stderr, "uplug_openLibrary(%s,%s) libEntry %d, lib %p\n", libName, u_errorName(*status), libEntry, lib));
#endif
      /* no need to free - just won't increase the count. */
      libraryCount--;
    } else { /* is it still there? */
      /* link it in */
      uprv_strncpy(libraryList[libEntry].name,libName,UPLUG_NAME_MAX);
      libraryList[libEntry].ref=1;
      lib = libraryList[libEntry].lib;
    }

  } else {
    lib = libraryList[libEntry].lib;
    libraryList[libEntry].ref++;
  }
  return lib;
}

U_INTERNAL void U_EXPORT2
uplug_closeLibrary(void *lib, UErrorCode *status) {
  int32_t i;
    
#if UPLUG_TRACE
  DBG((stderr, "uplug_closeLibrary(%p,%s) list %p\n", lib, u_errorName(*status), (void*)libraryList));
#endif
  if(U_FAILURE(*status)) return;
    
  for(i=0;i<libraryCount;i++) {
    if(lib==libraryList[i].lib) {
      if(--(libraryList[i].ref) == 0) {
        uprv_dl_close(libraryList[i].lib, status);
        libraryCount = uplug_removeEntryAt(libraryList, libraryCount, sizeof(*libraryList), i);
      }
      return;
    }
  }
  *status = U_INTERNAL_PROGRAM_ERROR; /* could not find the entry! */
}

#endif

static UPlugData pluginList[UPLUG_PLUGIN_INITIAL_COUNT];
static int32_t pluginCount = 0;



  
static int32_t uplug_pluginNumber(UPlugData* d) {
  UPlugData *pastPlug = &pluginList[pluginCount];
  if(d<=pluginList) {
    return 0;
  } else if(d>=pastPlug) {
    return pluginCount;
  } else {
    return (d-pluginList)/sizeof(pluginList[0]);
  }
}


U_CAPI UPlugData * U_EXPORT2
uplug_nextPlug(UPlugData *prior) {
  if(prior==NULL) {
    return pluginList;
  } else {
    UPlugData *nextPlug = &prior[1];
    UPlugData *pastPlug = &pluginList[pluginCount];
    
    if(nextPlug>=pastPlug) {
      return NULL;
    } else {
      return nextPlug;
    }
  }
}



/**
 * Call the plugin with some params
 */
static void uplug_callPlug(UPlugData *plug, UPlugReason reason, UErrorCode *status) {
  UPlugTokenReturn token;
  if(plug==NULL||U_FAILURE(*status)) {
    return;
  }
  token = (*(plug->entrypoint))(plug, reason, status);
  if(token!=UPLUG_TOKEN) {
    *status = U_INTERNAL_PROGRAM_ERROR;
  }
}


static void uplug_unloadPlug(UPlugData *plug, UErrorCode *status) {
  if(plug->awaitingLoad) {  /* shouldn't happen. Plugin hasn'tbeen loaded yet.*/
    *status = U_INTERNAL_PROGRAM_ERROR;
    return; 
  }
  if(U_SUCCESS(plug->pluginStatus)) {
    /* Don't unload a plug which has a failing load status - means it didn't actually load. */
    uplug_callPlug(plug, UPLUG_REASON_UNLOAD, status);
  }
}

static void uplug_queryPlug(UPlugData *plug, UErrorCode *status) {
  if(!plug->awaitingLoad || !(plug->level == UPLUG_LEVEL_UNKNOWN) ) {  /* shouldn't happen. Plugin hasn'tbeen loaded yet.*/
    *status = U_INTERNAL_PROGRAM_ERROR;
    return; 
  }
  plug->level = UPLUG_LEVEL_INVALID;
  uplug_callPlug(plug, UPLUG_REASON_QUERY, status);
  if(U_SUCCESS(*status)) { 
    if(plug->level == UPLUG_LEVEL_INVALID) {
      plug->pluginStatus = U_PLUGIN_DIDNT_SET_LEVEL;
      plug->awaitingLoad = FALSE;
    }
  } else {
    plug->pluginStatus = U_INTERNAL_PROGRAM_ERROR;
    plug->awaitingLoad = FALSE;
  }
}


static void uplug_loadPlug(UPlugData *plug, UErrorCode *status) {
  if(U_FAILURE(*status)) {
    return;
  }
  if(!plug->awaitingLoad || (plug->level < UPLUG_LEVEL_LOW) ) {  /* shouldn't happen. Plugin hasn'tbeen loaded yet.*/
    *status = U_INTERNAL_PROGRAM_ERROR;
    return;
  }
  uplug_callPlug(plug, UPLUG_REASON_LOAD, status);
  plug->awaitingLoad = FALSE;
  if(!U_SUCCESS(*status)) {
    plug->pluginStatus = U_INTERNAL_PROGRAM_ERROR;
  }
}

static UPlugData *uplug_allocateEmptyPlug(UErrorCode *status)
{
  UPlugData *plug = NULL;

  if(U_FAILURE(*status)) {
    return NULL;
  }

  if(pluginCount == UPLUG_PLUGIN_INITIAL_COUNT) {
    *status = U_MEMORY_ALLOCATION_ERROR;
    return NULL;
  }

  plug = &pluginList[pluginCount++];

  plug->token = UPLUG_TOKEN;
  plug->structSize = sizeof(UPlugData);
  plug->name[0]=0;
  plug->level = UPLUG_LEVEL_UNKNOWN; /* initialize to null state */
  plug->awaitingLoad = TRUE;
  plug->dontUnload = FALSE;
  plug->pluginStatus = U_ZERO_ERROR;
  plug->libName[0] = 0;
  plug->config[0]=0;
  plug->sym[0]=0;
  plug->lib=NULL;
  plug->entrypoint=NULL;


  return plug;
}

static UPlugData *uplug_allocatePlug(UPlugEntrypoint *entrypoint, const char *config, void *lib, const char *symName,
                                     UErrorCode *status) {
  UPlugData *plug = uplug_allocateEmptyPlug(status);
  if(U_FAILURE(*status)) {
    return NULL;
  }

  if(config!=NULL) {
    uprv_strncpy(plug->config, config, UPLUG_NAME_MAX);
  } else {
    plug->config[0] = 0;
  }
    
  if(symName!=NULL) {
    uprv_strncpy(plug->sym, symName, UPLUG_NAME_MAX);
  } else {
    plug->sym[0] = 0;
  }
    
  plug->entrypoint = entrypoint;
  plug->lib = lib;
  uplug_queryPlug(plug, status);
    
  return plug;
}

static void uplug_deallocatePlug(UPlugData *plug, UErrorCode *status) {
  UErrorCode subStatus = U_ZERO_ERROR;
  if(!plug->dontUnload) {
#if U_ENABLE_DYLOAD
    uplug_closeLibrary(plug->lib, &subStatus);
#endif
  }
  plug->lib = NULL;
  if(U_SUCCESS(*status) && U_FAILURE(subStatus)) {
    *status = subStatus;
  }
  /* shift plugins up and decrement count. */
  if(U_SUCCESS(*status)) {
    /* all ok- remove. */
    pluginCount = uplug_removeEntryAt(pluginList, pluginCount, sizeof(plug[0]), uplug_pluginNumber(plug));
  } else {
    /* not ok- leave as a message. */
    plug->awaitingLoad=FALSE;
    plug->entrypoint=0;
    plug->dontUnload=TRUE;
  }
}

static void uplug_doUnloadPlug(UPlugData *plugToRemove, UErrorCode *status) {
  if(plugToRemove != NULL) {
    uplug_unloadPlug(plugToRemove, status);
    uplug_deallocatePlug(plugToRemove, status);
  }
}

U_CAPI void U_EXPORT2
uplug_removePlug(UPlugData *plug, UErrorCode *status)  {
  UPlugData *cursor = NULL;
  UPlugData *plugToRemove = NULL;
  if(U_FAILURE(*status)) return;
    
  for(cursor=pluginList;cursor!=NULL;) {
    if(cursor==plug) {
      plugToRemove = plug;
      cursor=NULL;
    } else {
      cursor = uplug_nextPlug(cursor);
    }
  }
    
  uplug_doUnloadPlug(plugToRemove, status);
}




U_CAPI void U_EXPORT2 
uplug_setPlugNoUnload(UPlugData *data, UBool dontUnload)
{
  data->dontUnload = dontUnload;
}


U_CAPI void U_EXPORT2
uplug_setPlugLevel(UPlugData *data, UPlugLevel level) {
  data->level = level;
}


U_CAPI UPlugLevel U_EXPORT2
uplug_getPlugLevel(UPlugData *data) {
  return data->level;
}


U_CAPI void U_EXPORT2
uplug_setPlugName(UPlugData *data, const char *name) {
  uprv_strncpy(data->name, name, UPLUG_NAME_MAX);
}


U_CAPI const char * U_EXPORT2
uplug_getPlugName(UPlugData *data) {
  return data->name;
}


U_CAPI const char * U_EXPORT2
uplug_getSymbolName(UPlugData *data) {
  return data->sym;
}

U_CAPI const char * U_EXPORT2
uplug_getLibraryName(UPlugData *data, UErrorCode *status) {
  if(data->libName[0]) {
    return data->libName;
  } else {
#if U_ENABLE_DYLOAD
    return uplug_findLibrary(data->lib, status);
#else
    return NULL;
#endif
  }
}

U_CAPI void * U_EXPORT2
uplug_getLibrary(UPlugData *data) {
  return data->lib;
}

U_CAPI void * U_EXPORT2
uplug_getContext(UPlugData *data) {
  return data->context;
}


U_CAPI void U_EXPORT2
uplug_setContext(UPlugData *data, void *context) {
  data->context = context;
}

U_CAPI const char* U_EXPORT2
uplug_getConfiguration(UPlugData *data) {
  return data->config;
}

U_INTERNAL UPlugData* U_EXPORT2
uplug_getPlugInternal(int32_t n) { 
  if(n <0 || n >= pluginCount) {
    return NULL;
  } else { 
    return &(pluginList[n]);
  }
}


U_CAPI UErrorCode U_EXPORT2
uplug_getPlugLoadStatus(UPlugData *plug) {
  return plug->pluginStatus;
}




/**
 * Initialize a plugin fron an entrypoint and library - but don't load it.
 */
static UPlugData* uplug_initPlugFromEntrypointAndLibrary(UPlugEntrypoint *entrypoint, const char *config, void *lib, const char *sym,
                                                         UErrorCode *status) {
  UPlugData *plug = NULL;

  plug = uplug_allocatePlug(entrypoint, config, lib, sym, status);

  if(U_SUCCESS(*status)) {
    return plug;
  } else {
    uplug_deallocatePlug(plug, status);
    return NULL;
  }
}

U_CAPI UPlugData* U_EXPORT2
uplug_loadPlugFromEntrypoint(UPlugEntrypoint *entrypoint, const char *config, UErrorCode *status) {
  UPlugData* plug = uplug_initPlugFromEntrypointAndLibrary(entrypoint, config, NULL, NULL, status);
  uplug_loadPlug(plug, status);
  return plug;
}

#if U_ENABLE_DYLOAD

static UPlugData* 
uplug_initErrorPlug(const char *libName, const char *sym, const char *config, const char *nameOrError, UErrorCode loadStatus, UErrorCode *status)
{
  UPlugData *plug = uplug_allocateEmptyPlug(status);
  if(U_FAILURE(*status)) return NULL;

  plug->pluginStatus = loadStatus;
  plug->awaitingLoad = FALSE; /* Won't load. */
  plug->dontUnload = TRUE; /* cannot unload. */

  if(sym!=NULL) {
    uprv_strncpy(plug->sym, sym, UPLUG_NAME_MAX);
  }

  if(libName!=NULL) {
    uprv_strncpy(plug->libName, libName, UPLUG_NAME_MAX);
  }

  if(nameOrError!=NULL) {
    uprv_strncpy(plug->name, nameOrError, UPLUG_NAME_MAX);
  }

  if(config!=NULL) {
    uprv_strncpy(plug->config, config, UPLUG_NAME_MAX);
  }

  return plug;
}

/**
 * Fetch a plugin from DLL, and then initialize it from a library- but don't load it.
 */
static UPlugData* 
uplug_initPlugFromLibrary(const char *libName, const char *sym, const char *config, UErrorCode *status) {
  void *lib = NULL;
  UPlugData *plug = NULL;
  if(U_FAILURE(*status)) { return NULL; }
  lib = uplug_openLibrary(libName, status);
  if(lib!=NULL && U_SUCCESS(*status)) {
    UPlugEntrypoint *entrypoint = NULL;
    entrypoint = (UPlugEntrypoint*)uprv_dlsym_func(lib, sym, status);

    if(entrypoint!=NULL&&U_SUCCESS(*status)) {
      plug = uplug_initPlugFromEntrypointAndLibrary(entrypoint, config, lib, sym, status);
      if(plug!=NULL&&U_SUCCESS(*status)) {
        plug->lib = lib; /* plug takes ownership of library */
        lib = NULL; /* library is now owned by plugin. */
      }
    } else {
      UErrorCode subStatus = U_ZERO_ERROR;
      plug = uplug_initErrorPlug(libName,sym,config,"ERROR: Could not load entrypoint",(lib==NULL)?U_MISSING_RESOURCE_ERROR:*status,&subStatus);
    }
    if(lib!=NULL) { /* still need to close the lib */
      UErrorCode subStatus = U_ZERO_ERROR;
      uplug_closeLibrary(lib, &subStatus); /* don't care here */
    }
  } else {
    UErrorCode subStatus = U_ZERO_ERROR;
    plug = uplug_initErrorPlug(libName,sym,config,"ERROR: could not load library",(lib==NULL)?U_MISSING_RESOURCE_ERROR:*status,&subStatus);
  }
  return plug;
}

U_CAPI UPlugData* U_EXPORT2
uplug_loadPlugFromLibrary(const char *libName, const char *sym, const char *config, UErrorCode *status) { 
  UPlugData *plug = NULL;
  if(U_FAILURE(*status)) { return NULL; }
  plug = uplug_initPlugFromLibrary(libName, sym, config, status);
  uplug_loadPlug(plug, status);

  return plug;
}

#endif

static UPlugLevel gCurrentLevel = UPLUG_LEVEL_LOW;

U_CAPI UPlugLevel U_EXPORT2 uplug_getCurrentLevel() {
  return gCurrentLevel;
}

static UBool U_CALLCONV uplug_cleanup(void)
{
  int32_t i;
    
  UPlugData *pluginToRemove;
  /* cleanup plugs */
  for(i=0;i<pluginCount;i++) {
    UErrorCode subStatus = U_ZERO_ERROR;
    pluginToRemove = &pluginList[i];
    /* unload and deallocate */
    uplug_doUnloadPlug(pluginToRemove, &subStatus);
  }
  /* close other held libs? */
  gCurrentLevel = UPLUG_LEVEL_LOW;
  return TRUE;
}

#if U_ENABLE_DYLOAD

static void uplug_loadWaitingPlugs(UErrorCode *status) {
  int32_t i;
  UPlugLevel currentLevel = uplug_getCurrentLevel();
    
  if(U_FAILURE(*status)) {
    return;
  }
#if UPLUG_TRACE
  DBG((stderr,  "uplug_loadWaitingPlugs() Level: %d\n", currentLevel));
#endif
  /* pass #1: low level plugs */
  for(i=0;i<pluginCount;i++) {
    UErrorCode subStatus = U_ZERO_ERROR;
    UPlugData *pluginToLoad = &pluginList[i];
    if(pluginToLoad->awaitingLoad) {
      if(pluginToLoad->level == UPLUG_LEVEL_LOW) {
        if(currentLevel > UPLUG_LEVEL_LOW) {
          pluginToLoad->pluginStatus = U_PLUGIN_TOO_HIGH;
        } else {
          UPlugLevel newLevel;
          uplug_loadPlug(pluginToLoad, &subStatus);
          newLevel = uplug_getCurrentLevel();
          if(newLevel > currentLevel) {
            pluginToLoad->pluginStatus = U_PLUGIN_CHANGED_LEVEL_WARNING;
            currentLevel = newLevel;
          }
        }
        pluginToLoad->awaitingLoad = FALSE;
      } 
    }
  }    
  for(i=0;i<pluginCount;i++) {
    UErrorCode subStatus = U_ZERO_ERROR;
    UPlugData *pluginToLoad = &pluginList[i];
        
    if(pluginToLoad->awaitingLoad) {
      if(pluginToLoad->level == UPLUG_LEVEL_INVALID) { 
        pluginToLoad->pluginStatus = U_PLUGIN_DIDNT_SET_LEVEL;
      } else if(pluginToLoad->level == UPLUG_LEVEL_UNKNOWN) {
        pluginToLoad->pluginStatus = U_INTERNAL_PROGRAM_ERROR;
      } else {
        uplug_loadPlug(pluginToLoad, &subStatus);
      }
      pluginToLoad->awaitingLoad = FALSE;
    }
  }
    
#if UPLUG_TRACE
  DBG((stderr,  " Done Loading Plugs. Level: %d\n", (int32_t)uplug_getCurrentLevel()));
#endif
}

/* Name of the plugin config file */
static char plugin_file[2048] = "";
#endif

U_INTERNAL const char* U_EXPORT2
uplug_getPluginFile() {
#if U_ENABLE_DYLOAD && !UCONFIG_NO_FILE_IO
  return plugin_file;
#else
  return NULL;
#endif
}


//  uplug_init()  is called first thing from u_init().

U_CAPI void U_EXPORT2
uplug_init(UErrorCode *status) {
#if !U_ENABLE_DYLOAD
  (void)status; /* unused */
#elif !UCONFIG_NO_FILE_IO
  CharString plugin_dir;
  const char *env = getenv("ICU_PLUGINS");

  if(U_FAILURE(*status)) return;
  if(env != NULL) {
    plugin_dir.append(env, -1, *status);
  }
  if(U_FAILURE(*status)) return;

#if defined(DEFAULT_ICU_PLUGINS) 
  if(plugin_dir.isEmpty()) {
    plugin_dir.append(DEFAULT_ICU_PLUGINS, -1, *status);
  }
#endif

#if UPLUG_TRACE
  DBG((stderr, "ICU_PLUGINS=%s\n", plugin_dir.data()));
#endif

  if(!plugin_dir.isEmpty()) {
    FILE *f;
        
    CharString pluginFile;
#ifdef OS390BATCH
/* There are potentially a lot of ways to implement a plugin directory on OS390/zOS  */
/* Keeping in mind that unauthorized file access is logged, monitored, and enforced  */
/* I've chosen to open a DDNAME if BATCH and leave it alone for (presumably) UNIX    */
/* System Services.  Alternative techniques might be allocating a member in          */
/* SYS1.PARMLIB or setting an environment variable "ICU_PLUGIN_PATH" (?).  The       */
/* DDNAME can be connected to a file in the HFS if need be.                          */

    pluginFile.append("//DD:ICUPLUG", -1, *status);        /* JAM 20 Oct 2011 */
#else
    pluginFile.append(plugin_dir, *status);
    pluginFile.append(U_FILE_SEP_STRING, -1, *status);
    pluginFile.append("icuplugins", -1, *status);
    pluginFile.append(U_ICU_VERSION_SHORT, -1, *status);
    pluginFile.append(".txt", -1, *status);
#endif

#if UPLUG_TRACE
    DBG((stderr, "status=%s\n", u_errorName(*status)));
#endif

    if(U_FAILURE(*status)) {
      return;
    }
    if((size_t)pluginFile.length() > (sizeof(plugin_file)-1)) {
      *status = U_BUFFER_OVERFLOW_ERROR;
#if UPLUG_TRACE
      DBG((stderr, "status=%s\n", u_errorName(*status)));
#endif
      return;
    }
    
    /* plugin_file is not used for processing - it is only used 
       so that uplug_getPluginFile() works (i.e. icuinfo)
    */
    uprv_strncpy(plugin_file, pluginFile.data(), sizeof(plugin_file));
        
#if UPLUG_TRACE
    DBG((stderr, "pluginfile= %s len %d/%d\n", plugin_file, (int)strlen(plugin_file), (int)sizeof(plugin_file)));
#endif
        
#ifdef __MVS__
    if (iscics()) /* 12 Nov 2011 JAM */
    {
        f = NULL;
    }
    else
#endif
    {
        f = fopen(pluginFile.data(), "r");
    }

    if(f != NULL) {
      char linebuf[1024];
      char *p, *libName=NULL, *symName=NULL, *config=NULL;
      int32_t line = 0;
            
            
      while(fgets(linebuf,1023,f)) {
        line++;

        if(!*linebuf || *linebuf=='#') {
          continue;
        } else {
          p = linebuf;
          while(*p&&isspace((int)*p))
            p++;
          if(!*p || *p=='#') continue;
          libName = p;
          while(*p&&!isspace((int)*p)) {
            p++;
          }
          if(!*p || *p=='#') continue; /* no tab after libname */
          *p=0; /* end of libname */
          p++;
          while(*p&&isspace((int)*p)) {
            p++;
          }
          if(!*p||*p=='#') continue; /* no symname after libname +tab */
          symName = p;
          while(*p&&!isspace((int)*p)) {
            p++;
          }
                    
          if(*p) { /* has config */
            *p=0;
            ++p;
            while(*p&&isspace((int)*p)) {
              p++;
            }
            if(*p) {
              config = p;
            }
          }
                    
          /* chop whitespace at the end of the config */
          if(config!=NULL&&*config!=0) {
            p = config+strlen(config);
            while(p>config&&isspace((int)*(--p))) {
              *p=0;
            }
          }
                
          /* OK, we're good. */
          { 
            UErrorCode subStatus = U_ZERO_ERROR;
            UPlugData *plug = uplug_initPlugFromLibrary(libName, symName, config, &subStatus);
            if(U_FAILURE(subStatus) && U_SUCCESS(*status)) {
              *status = subStatus;
            }
#if UPLUG_TRACE
            DBG((stderr, "PLUGIN libName=[%s], sym=[%s], config=[%s]\n", libName, symName, config));
            DBG((stderr, " -> %p, %s\n", (void*)plug, u_errorName(subStatus)));
#else
            (void)plug; /* unused */
#endif
          }
        }
      }
      fclose(f);
    } else {
#if UPLUG_TRACE
      DBG((stderr, "Can't open plugin file %s\n", plugin_file));
#endif
    }
  }
  uplug_loadWaitingPlugs(status);
#endif /* U_ENABLE_DYLOAD */
  gCurrentLevel = UPLUG_LEVEL_HIGH;
  ucln_registerCleanup(UCLN_UPLUG, uplug_cleanup);
}

#endif


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* loadednormalizer2impl.cpp
*
* created on: 2014sep03
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_NORMALIZATION









// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* norm2allmodes.h
*
* created on: 2014sep07
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_NORMALIZATION








U_NAMESPACE_BEGIN

// Intermediate class:
// Has Normalizer2Impl and does boilerplate argument checking and setup.
class Normalizer2WithImpl : public Normalizer2 {
public:
    Normalizer2WithImpl(const Normalizer2Impl &ni) : impl(ni) {}
    virtual ~Normalizer2WithImpl();

    // normalize
    virtual UnicodeString &
    normalize(const UnicodeString &src,
              UnicodeString &dest,
              UErrorCode &errorCode) const {
        if(U_FAILURE(errorCode)) {
            dest.setToBogus();
            return dest;
        }
        const UChar *sArray=src.getBuffer();
        if(&dest==&src || sArray==NULL) {
            errorCode=U_ILLEGAL_ARGUMENT_ERROR;
            dest.setToBogus();
            return dest;
        }
        dest.remove();
        ReorderingBuffer buffer(impl, dest);
        if(buffer.init(src.length(), errorCode)) {
            normalize(sArray, sArray+src.length(), buffer, errorCode);
        }
        return dest;
    }
    virtual void
    normalize(const UChar *src, const UChar *limit,
              ReorderingBuffer &buffer, UErrorCode &errorCode) const = 0;

    // normalize and append
    virtual UnicodeString &
    normalizeSecondAndAppend(UnicodeString &first,
                             const UnicodeString &second,
                             UErrorCode &errorCode) const {
        return normalizeSecondAndAppend(first, second, TRUE, errorCode);
    }
    virtual UnicodeString &
    append(UnicodeString &first,
           const UnicodeString &second,
           UErrorCode &errorCode) const {
        return normalizeSecondAndAppend(first, second, FALSE, errorCode);
    }
    UnicodeString &
    normalizeSecondAndAppend(UnicodeString &first,
                             const UnicodeString &second,
                             UBool doNormalize,
                             UErrorCode &errorCode) const {
        uprv_checkCanGetBuffer(first, errorCode);
        if(U_FAILURE(errorCode)) {
            return first;
        }
        const UChar *secondArray=second.getBuffer();
        if(&first==&second || secondArray==NULL) {
            errorCode=U_ILLEGAL_ARGUMENT_ERROR;
            return first;
        }
        int32_t firstLength=first.length();
        UnicodeString safeMiddle;
        {
            ReorderingBuffer buffer(impl, first);
            if(buffer.init(firstLength+second.length(), errorCode)) {
                normalizeAndAppend(secondArray, secondArray+second.length(), doNormalize,
                                   safeMiddle, buffer, errorCode);
            }
        }  // The ReorderingBuffer destructor finalizes the first string.
        if(U_FAILURE(errorCode)) {
            // Restore the modified suffix of the first string.
            first.replace(firstLength-safeMiddle.length(), 0x7fffffff, safeMiddle);
        }
        return first;
    }
    virtual void
    normalizeAndAppend(const UChar *src, const UChar *limit, UBool doNormalize,
                       UnicodeString &safeMiddle,
                       ReorderingBuffer &buffer, UErrorCode &errorCode) const = 0;
    virtual UBool
    getDecomposition(UChar32 c, UnicodeString &decomposition) const {
        UChar buffer[4];
        int32_t length;
        const UChar *d=impl.getDecomposition(c, buffer, length);
        if(d==NULL) {
            return FALSE;
        }
        if(d==buffer) {
            decomposition.setTo(buffer, length);  // copy the string (Jamos from Hangul syllable c)
        } else {
            decomposition.setTo(FALSE, d, length);  // read-only alias
        }
        return TRUE;
    }
    virtual UBool
    getRawDecomposition(UChar32 c, UnicodeString &decomposition) const {
        UChar buffer[30];
        int32_t length;
        const UChar *d=impl.getRawDecomposition(c, buffer, length);
        if(d==NULL) {
            return FALSE;
        }
        if(d==buffer) {
            decomposition.setTo(buffer, length);  // copy the string (algorithmic decomposition)
        } else {
            decomposition.setTo(FALSE, d, length);  // read-only alias
        }
        return TRUE;
    }
    virtual UChar32
    composePair(UChar32 a, UChar32 b) const {
        return impl.composePair(a, b);
    }

    virtual uint8_t
    getCombiningClass(UChar32 c) const {
        return impl.getCC(impl.getNorm16(c));
    }

    // quick checks
    virtual UBool
    isNormalized(const UnicodeString &s, UErrorCode &errorCode) const {
        if(U_FAILURE(errorCode)) {
            return FALSE;
        }
        const UChar *sArray=s.getBuffer();
        if(sArray==NULL) {
            errorCode=U_ILLEGAL_ARGUMENT_ERROR;
            return FALSE;
        }
        const UChar *sLimit=sArray+s.length();
        return sLimit==spanQuickCheckYes(sArray, sLimit, errorCode);
    }
    virtual UNormalizationCheckResult
    quickCheck(const UnicodeString &s, UErrorCode &errorCode) const {
        return Normalizer2WithImpl::isNormalized(s, errorCode) ? UNORM_YES : UNORM_NO;
    }
    virtual int32_t
    spanQuickCheckYes(const UnicodeString &s, UErrorCode &errorCode) const {
        if(U_FAILURE(errorCode)) {
            return 0;
        }
        const UChar *sArray=s.getBuffer();
        if(sArray==NULL) {
            errorCode=U_ILLEGAL_ARGUMENT_ERROR;
            return 0;
        }
        return (int32_t)(spanQuickCheckYes(sArray, sArray+s.length(), errorCode)-sArray);
    }
    virtual const UChar *
    spanQuickCheckYes(const UChar *src, const UChar *limit, UErrorCode &errorCode) const = 0;

    virtual UNormalizationCheckResult getQuickCheck(UChar32) const {
        return UNORM_YES;
    }

    const Normalizer2Impl &impl;
};

class DecomposeNormalizer2 : public Normalizer2WithImpl {
public:
    DecomposeNormalizer2(const Normalizer2Impl &ni) : Normalizer2WithImpl(ni) {}
    virtual ~DecomposeNormalizer2();

private:
    virtual void
    normalize(const UChar *src, const UChar *limit,
              ReorderingBuffer &buffer, UErrorCode &errorCode) const {
        impl.decompose(src, limit, &buffer, errorCode);
    }
    using Normalizer2WithImpl::normalize;  // Avoid warning about hiding base class function.
    virtual void
    normalizeAndAppend(const UChar *src, const UChar *limit, UBool doNormalize,
                       UnicodeString &safeMiddle,
                       ReorderingBuffer &buffer, UErrorCode &errorCode) const {
        impl.decomposeAndAppend(src, limit, doNormalize, safeMiddle, buffer, errorCode);
    }
    virtual const UChar *
    spanQuickCheckYes(const UChar *src, const UChar *limit, UErrorCode &errorCode) const {
        return impl.decompose(src, limit, NULL, errorCode);
    }
    using Normalizer2WithImpl::spanQuickCheckYes;  // Avoid warning about hiding base class function.
    virtual UNormalizationCheckResult getQuickCheck(UChar32 c) const {
        return impl.isDecompYes(impl.getNorm16(c)) ? UNORM_YES : UNORM_NO;
    }
    virtual UBool hasBoundaryBefore(UChar32 c) const { return impl.hasDecompBoundaryBefore(c); }
    virtual UBool hasBoundaryAfter(UChar32 c) const { return impl.hasDecompBoundaryAfter(c); }
    virtual UBool isInert(UChar32 c) const { return impl.isDecompInert(c); }
};

class ComposeNormalizer2 : public Normalizer2WithImpl {
public:
    ComposeNormalizer2(const Normalizer2Impl &ni, UBool fcc) :
        Normalizer2WithImpl(ni), onlyContiguous(fcc) {}
    virtual ~ComposeNormalizer2();

private:
    virtual void
    normalize(const UChar *src, const UChar *limit,
              ReorderingBuffer &buffer, UErrorCode &errorCode) const U_OVERRIDE {
        impl.compose(src, limit, onlyContiguous, TRUE, buffer, errorCode);
    }
    using Normalizer2WithImpl::normalize;  // Avoid warning about hiding base class function.

    void
    normalizeUTF8(uint32_t options, StringPiece src, ByteSink &sink,
                  Edits *edits, UErrorCode &errorCode) const U_OVERRIDE {
        if (U_FAILURE(errorCode)) {
            return;
        }
        if (edits != nullptr && (options & U_EDITS_NO_RESET) == 0) {
            edits->reset();
        }
        const uint8_t *s = reinterpret_cast<const uint8_t *>(src.data());
        impl.composeUTF8(options, onlyContiguous, s, s + src.length(),
                         &sink, edits, errorCode);
        sink.Flush();
    }

    virtual void
    normalizeAndAppend(const UChar *src, const UChar *limit, UBool doNormalize,
                       UnicodeString &safeMiddle,
                       ReorderingBuffer &buffer, UErrorCode &errorCode) const U_OVERRIDE {
        impl.composeAndAppend(src, limit, doNormalize, onlyContiguous, safeMiddle, buffer, errorCode);
    }

    virtual UBool
    isNormalized(const UnicodeString &s, UErrorCode &errorCode) const U_OVERRIDE {
        if(U_FAILURE(errorCode)) {
            return FALSE;
        }
        const UChar *sArray=s.getBuffer();
        if(sArray==NULL) {
            errorCode=U_ILLEGAL_ARGUMENT_ERROR;
            return FALSE;
        }
        UnicodeString temp;
        ReorderingBuffer buffer(impl, temp);
        if(!buffer.init(5, errorCode)) {  // small destCapacity for substring normalization
            return FALSE;
        }
        return impl.compose(sArray, sArray+s.length(), onlyContiguous, FALSE, buffer, errorCode);
    }
    virtual UBool
    isNormalizedUTF8(StringPiece sp, UErrorCode &errorCode) const U_OVERRIDE {
        if(U_FAILURE(errorCode)) {
            return FALSE;
        }
        const uint8_t *s = reinterpret_cast<const uint8_t *>(sp.data());
        return impl.composeUTF8(0, onlyContiguous, s, s + sp.length(), nullptr, nullptr, errorCode);
    }
    virtual UNormalizationCheckResult
    quickCheck(const UnicodeString &s, UErrorCode &errorCode) const U_OVERRIDE {
        if(U_FAILURE(errorCode)) {
            return UNORM_MAYBE;
        }
        const UChar *sArray=s.getBuffer();
        if(sArray==NULL) {
            errorCode=U_ILLEGAL_ARGUMENT_ERROR;
            return UNORM_MAYBE;
        }
        UNormalizationCheckResult qcResult=UNORM_YES;
        impl.composeQuickCheck(sArray, sArray+s.length(), onlyContiguous, &qcResult);
        return qcResult;
    }
    virtual const UChar *
    spanQuickCheckYes(const UChar *src, const UChar *limit, UErrorCode &) const U_OVERRIDE {
        return impl.composeQuickCheck(src, limit, onlyContiguous, NULL);
    }
    using Normalizer2WithImpl::spanQuickCheckYes;  // Avoid warning about hiding base class function.
    virtual UNormalizationCheckResult getQuickCheck(UChar32 c) const U_OVERRIDE {
        return impl.getCompQuickCheck(impl.getNorm16(c));
    }
    virtual UBool hasBoundaryBefore(UChar32 c) const U_OVERRIDE {
        return impl.hasCompBoundaryBefore(c);
    }
    virtual UBool hasBoundaryAfter(UChar32 c) const U_OVERRIDE {
        return impl.hasCompBoundaryAfter(c, onlyContiguous);
    }
    virtual UBool isInert(UChar32 c) const U_OVERRIDE {
        return impl.isCompInert(c, onlyContiguous);
    }

    const UBool onlyContiguous;
};

class FCDNormalizer2 : public Normalizer2WithImpl {
public:
    FCDNormalizer2(const Normalizer2Impl &ni) : Normalizer2WithImpl(ni) {}
    virtual ~FCDNormalizer2();

private:
    virtual void
    normalize(const UChar *src, const UChar *limit,
              ReorderingBuffer &buffer, UErrorCode &errorCode) const {
        impl.makeFCD(src, limit, &buffer, errorCode);
    }
    using Normalizer2WithImpl::normalize;  // Avoid warning about hiding base class function.
    virtual void
    normalizeAndAppend(const UChar *src, const UChar *limit, UBool doNormalize,
                       UnicodeString &safeMiddle,
                       ReorderingBuffer &buffer, UErrorCode &errorCode) const {
        impl.makeFCDAndAppend(src, limit, doNormalize, safeMiddle, buffer, errorCode);
    }
    virtual const UChar *
    spanQuickCheckYes(const UChar *src, const UChar *limit, UErrorCode &errorCode) const {
        return impl.makeFCD(src, limit, NULL, errorCode);
    }
    using Normalizer2WithImpl::spanQuickCheckYes;  // Avoid warning about hiding base class function.
    virtual UBool hasBoundaryBefore(UChar32 c) const { return impl.hasFCDBoundaryBefore(c); }
    virtual UBool hasBoundaryAfter(UChar32 c) const { return impl.hasFCDBoundaryAfter(c); }
    virtual UBool isInert(UChar32 c) const { return impl.isFCDInert(c); }
};

struct Norm2AllModes : public UMemory {
    Norm2AllModes(Normalizer2Impl *i)
            : impl(i), comp(*i, FALSE), decomp(*i), fcd(*i), fcc(*i, TRUE) {}
    ~Norm2AllModes();

    static Norm2AllModes *createInstance(Normalizer2Impl *impl, UErrorCode &errorCode);
    static Norm2AllModes *createNFCInstance(UErrorCode &errorCode);
    static Norm2AllModes *createInstance(const char *packageName,
                                         const char *name,
                                         UErrorCode &errorCode);

    static const Norm2AllModes *getNFCInstance(UErrorCode &errorCode);
    static const Norm2AllModes *getNFKCInstance(UErrorCode &errorCode);
    static const Norm2AllModes *getNFKC_CFInstance(UErrorCode &errorCode);

    Normalizer2Impl *impl;
    ComposeNormalizer2 comp;
    DecomposeNormalizer2 decomp;
    FCDNormalizer2 fcd;
    ComposeNormalizer2 fcc;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_NORMALIZATION






U_NAMESPACE_BEGIN

class LoadedNormalizer2Impl : public Normalizer2Impl {
public:
    LoadedNormalizer2Impl() : memory(NULL), ownedTrie(NULL) {}
    virtual ~LoadedNormalizer2Impl();

    void load(const char *packageName, const char *name, UErrorCode &errorCode);

private:
    static UBool U_CALLCONV
    isAcceptable(void *context, const char *type, const char *name, const UDataInfo *pInfo);

    UDataMemory *memory;
    UCPTrie *ownedTrie;
};

LoadedNormalizer2Impl::~LoadedNormalizer2Impl() {
    udata_close(memory);
    ucptrie_close(ownedTrie);
}

UBool U_CALLCONV
LoadedNormalizer2Impl::isAcceptable(void * /*context*/,
                                    const char * /* type */, const char * /*name*/,
                                    const UDataInfo *pInfo) {
    if(
        pInfo->size>=20 &&
        pInfo->isBigEndian==U_IS_BIG_ENDIAN &&
        pInfo->charsetFamily==U_CHARSET_FAMILY &&
        pInfo->dataFormat[0]==0x4e &&    /* dataFormat="Nrm2" */
        pInfo->dataFormat[1]==0x72 &&
        pInfo->dataFormat[2]==0x6d &&
        pInfo->dataFormat[3]==0x32 &&
        pInfo->formatVersion[0]==4
    ) {
        // Normalizer2Impl *me=(Normalizer2Impl *)context;
        // uprv_memcpy(me->dataVersion, pInfo->dataVersion, 4);
        return TRUE;
    } else {
        return FALSE;
    }
}

void
LoadedNormalizer2Impl::load(const char *packageName, const char *name, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    memory=udata_openChoice(packageName, "nrm", name, isAcceptable, this, &errorCode);
    if(U_FAILURE(errorCode)) {
        return;
    }
    const uint8_t *inBytes=(const uint8_t *)udata_getMemory(memory);
    const int32_t *inIndexes=(const int32_t *)inBytes;
    int32_t indexesLength=inIndexes[IX_NORM_TRIE_OFFSET]/4;
    if(indexesLength<=IX_MIN_LCCC_CP) {
        errorCode=U_INVALID_FORMAT_ERROR;  // Not enough indexes.
        return;
    }

    int32_t offset=inIndexes[IX_NORM_TRIE_OFFSET];
    int32_t nextOffset=inIndexes[IX_EXTRA_DATA_OFFSET];
    ownedTrie=ucptrie_openFromBinary(UCPTRIE_TYPE_FAST, UCPTRIE_VALUE_BITS_16,
                                     inBytes+offset, nextOffset-offset, NULL,
                                     &errorCode);
    if(U_FAILURE(errorCode)) {
        return;
    }

    offset=nextOffset;
    nextOffset=inIndexes[IX_SMALL_FCD_OFFSET];
    const uint16_t *inExtraData=(const uint16_t *)(inBytes+offset);

    // smallFCD: new in formatVersion 2
    offset=nextOffset;
    const uint8_t *inSmallFCD=inBytes+offset;

    init(inIndexes, ownedTrie, inExtraData, inSmallFCD);
}

// instance cache ---------------------------------------------------------- ***

Norm2AllModes *
Norm2AllModes::createInstance(const char *packageName,
                              const char *name,
                              UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    LoadedNormalizer2Impl *impl=new LoadedNormalizer2Impl;
    if(impl==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    impl->load(packageName, name, errorCode);
    return createInstance(impl, errorCode);
}

U_CDECL_BEGIN
static UBool U_CALLCONV uprv_loaded_normalizer2_cleanup();
U_CDECL_END

#if !NORM2_HARDCODE_NFC_DATA
static Norm2AllModes *nfcSingleton;
static icu::UInitOnce nfcInitOnce = U_INITONCE_INITIALIZER;
#endif

static Norm2AllModes *nfkcSingleton;
static icu::UInitOnce nfkcInitOnce = U_INITONCE_INITIALIZER;

static Norm2AllModes *nfkc_cfSingleton;
static icu::UInitOnce nfkc_cfInitOnce = U_INITONCE_INITIALIZER;

static UHashtable    *cache=NULL;

// UInitOnce singleton initialization function
static void U_CALLCONV initSingletons(const char *what, UErrorCode &errorCode) {
#if !NORM2_HARDCODE_NFC_DATA
    if (uprv_strcmp(what, "nfc") == 0) {
        nfcSingleton    = Norm2AllModes::createInstance(NULL, "nfc", errorCode);
    } else
#endif
    if (uprv_strcmp(what, "nfkc") == 0) {
        nfkcSingleton    = Norm2AllModes::createInstance(NULL, "nfkc", errorCode);
    } else if (uprv_strcmp(what, "nfkc_cf") == 0) {
        nfkc_cfSingleton = Norm2AllModes::createInstance(NULL, "nfkc_cf", errorCode);
    } else {
        UPRV_UNREACHABLE;   // Unknown singleton
    }
    ucln_common_registerCleanup(UCLN_COMMON_LOADED_NORMALIZER2, uprv_loaded_normalizer2_cleanup);
}

U_CDECL_BEGIN

static void U_CALLCONV deleteNorm2AllModes(void *allModes) {
    delete (Norm2AllModes *)allModes;
}

static UBool U_CALLCONV uprv_loaded_normalizer2_cleanup() {
#if !NORM2_HARDCODE_NFC_DATA
    delete nfcSingleton;
    nfcSingleton = NULL;
    nfcInitOnce.reset();
#endif

    delete nfkcSingleton;
    nfkcSingleton = NULL;
    nfkcInitOnce.reset();

    delete nfkc_cfSingleton;
    nfkc_cfSingleton = NULL;
    nfkc_cfInitOnce.reset();

    uhash_close(cache);
    cache=NULL;
    return TRUE;
}

U_CDECL_END

#if !NORM2_HARDCODE_NFC_DATA
const Norm2AllModes *
Norm2AllModes::getNFCInstance(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return NULL; }
    umtx_initOnce(nfcInitOnce, &initSingletons, "nfc", errorCode);
    return nfcSingleton;
}
#endif

const Norm2AllModes *
Norm2AllModes::getNFKCInstance(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return NULL; }
    umtx_initOnce(nfkcInitOnce, &initSingletons, "nfkc", errorCode);
    return nfkcSingleton;
}

const Norm2AllModes *
Norm2AllModes::getNFKC_CFInstance(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return NULL; }
    umtx_initOnce(nfkc_cfInitOnce, &initSingletons, "nfkc_cf", errorCode);
    return nfkc_cfSingleton;
}

#if !NORM2_HARDCODE_NFC_DATA
const Normalizer2 *
Normalizer2::getNFCInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFCInstance(errorCode);
    return allModes!=NULL ? &allModes->comp : NULL;
}

const Normalizer2 *
Normalizer2::getNFDInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFCInstance(errorCode);
    return allModes!=NULL ? &allModes->decomp : NULL;
}

const Normalizer2 *Normalizer2Factory::getFCDInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFCInstance(errorCode);
    return allModes!=NULL ? &allModes->fcd : NULL;
}

const Normalizer2 *Normalizer2Factory::getFCCInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFCInstance(errorCode);
    return allModes!=NULL ? &allModes->fcc : NULL;
}

const Normalizer2Impl *
Normalizer2Factory::getNFCImpl(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFCInstance(errorCode);
    return allModes!=NULL ? allModes->impl : NULL;
}
#endif

const Normalizer2 *
Normalizer2::getNFKCInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFKCInstance(errorCode);
    return allModes!=NULL ? &allModes->comp : NULL;
}

const Normalizer2 *
Normalizer2::getNFKDInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFKCInstance(errorCode);
    return allModes!=NULL ? &allModes->decomp : NULL;
}

const Normalizer2 *
Normalizer2::getNFKCCasefoldInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFKC_CFInstance(errorCode);
    return allModes!=NULL ? &allModes->comp : NULL;
}

const Normalizer2 *
Normalizer2::getInstance(const char *packageName,
                         const char *name,
                         UNormalization2Mode mode,
                         UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    if(name==NULL || *name==0) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    const Norm2AllModes *allModes=NULL;
    if(packageName==NULL) {
        if(0==uprv_strcmp(name, "nfc")) {
            allModes=Norm2AllModes::getNFCInstance(errorCode);
        } else if(0==uprv_strcmp(name, "nfkc")) {
            allModes=Norm2AllModes::getNFKCInstance(errorCode);
        } else if(0==uprv_strcmp(name, "nfkc_cf")) {
            allModes=Norm2AllModes::getNFKC_CFInstance(errorCode);
        }
    }
    if(allModes==NULL && U_SUCCESS(errorCode)) {
        {
            Mutex lock;
            if(cache!=NULL) {
                allModes=(Norm2AllModes *)uhash_get(cache, name);
            }
        }
        if(allModes==NULL) {
            ucln_common_registerCleanup(UCLN_COMMON_LOADED_NORMALIZER2, uprv_loaded_normalizer2_cleanup);
            LocalPointer<Norm2AllModes> localAllModes(
                Norm2AllModes::createInstance(packageName, name, errorCode));
            if(U_SUCCESS(errorCode)) {
                Mutex lock;
                if(cache==NULL) {
                    cache=uhash_open(uhash_hashChars, uhash_compareChars, NULL, &errorCode);
                    if(U_FAILURE(errorCode)) {
                        return NULL;
                    }
                    uhash_setKeyDeleter(cache, uprv_free);
                    uhash_setValueDeleter(cache, deleteNorm2AllModes);
                }
                void *temp=uhash_get(cache, name);
                if(temp==NULL) {
                    int32_t keyLength= static_cast<int32_t>(uprv_strlen(name)+1);
                    char *nameCopy=(char *)uprv_malloc(keyLength);
                    if(nameCopy==NULL) {
                        errorCode=U_MEMORY_ALLOCATION_ERROR;
                        return NULL;
                    }
                    uprv_memcpy(nameCopy, name, keyLength);
                    allModes=localAllModes.getAlias();
                    uhash_put(cache, nameCopy, localAllModes.orphan(), &errorCode);
                } else {
                    // race condition
                    allModes=(Norm2AllModes *)temp;
                }
            }
        }
    }
    if(allModes!=NULL && U_SUCCESS(errorCode)) {
        switch(mode) {
        case UNORM2_COMPOSE:
            return &allModes->comp;
        case UNORM2_DECOMPOSE:
            return &allModes->decomp;
        case UNORM2_FCD:
            return &allModes->fcd;
        case UNORM2_COMPOSE_CONTIGUOUS:
            return &allModes->fcc;
        default:
            break;  // do nothing
        }
    }
    return NULL;
}

const Normalizer2 *
Normalizer2Factory::getInstance(UNormalizationMode mode, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    switch(mode) {
    case UNORM_NFD:
        return Normalizer2::getNFDInstance(errorCode);
    case UNORM_NFKD:
        return Normalizer2::getNFKDInstance(errorCode);
    case UNORM_NFC:
        return Normalizer2::getNFCInstance(errorCode);
    case UNORM_NFKC:
        return Normalizer2::getNFKCInstance(errorCode);
    case UNORM_FCD:
        return getFCDInstance(errorCode);
    default:  // UNORM_NONE
        return getNoopInstance(errorCode);
    }
}

const Normalizer2Impl *
Normalizer2Factory::getNFKCImpl(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFKCInstance(errorCode);
    return allModes!=NULL ? allModes->impl : NULL;
}

const Normalizer2Impl *
Normalizer2Factory::getNFKC_CFImpl(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFKC_CFInstance(errorCode);
    return allModes!=NULL ? allModes->impl : NULL;
}

U_NAMESPACE_END

// C API ------------------------------------------------------------------- ***

U_NAMESPACE_USE

U_CAPI const UNormalizer2 * U_EXPORT2
unorm2_getNFKCInstance(UErrorCode *pErrorCode) {
    return (const UNormalizer2 *)Normalizer2::getNFKCInstance(*pErrorCode);
}

U_CAPI const UNormalizer2 * U_EXPORT2
unorm2_getNFKDInstance(UErrorCode *pErrorCode) {
    return (const UNormalizer2 *)Normalizer2::getNFKDInstance(*pErrorCode);
}

U_CAPI const UNormalizer2 * U_EXPORT2
unorm2_getNFKCCasefoldInstance(UErrorCode *pErrorCode) {
    return (const UNormalizer2 *)Normalizer2::getNFKCCasefoldInstance(*pErrorCode);
}

U_CAPI const UNormalizer2 * U_EXPORT2
unorm2_getInstance(const char *packageName,
                   const char *name,
                   UNormalization2Mode mode,
                   UErrorCode *pErrorCode) {
    return (const UNormalizer2 *)Normalizer2::getInstance(packageName, name, mode, *pErrorCode);
}

U_CFUNC UNormalizationCheckResult
unorm_getQuickCheck(UChar32 c, UNormalizationMode mode) {
    if(mode<=UNORM_NONE || UNORM_FCD<=mode) {
        return UNORM_YES;
    }
    UErrorCode errorCode=U_ZERO_ERROR;
    const Normalizer2 *norm2=Normalizer2Factory::getInstance(mode, errorCode);
    if(U_SUCCESS(errorCode)) {
        return ((const Normalizer2WithImpl *)norm2)->getQuickCheck(c);
    } else {
        return UNORM_MAYBE;
    }
}

#endif  // !UCONFIG_NO_NORMALIZATION
// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

#include <utility>

  // CharStringByteSink


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2004-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/





/**
 * Create an iterator over the specified keywords list
 * @param keywordList double-null terminated list. Will be copied.
 * @param keywordListSize size in bytes of keywordList
 * @param status err code
 * @return enumeration (owned by caller) of the keyword list.
 * @internal ICU 3.0
 */
U_CAPI UEnumeration* U_EXPORT2
uloc_openKeywordList(const char *keywordList, int32_t keywordListSize, UErrorCode* status);

/**
 * Look up a resource bundle table item with fallback on the table level.
 * This is accessible so it can be called by C++ code.
 */
U_CAPI const UChar * U_EXPORT2
uloc_getTableStringWithFallback(
    const char *path,
    const char *locale,
    const char *tableKey,
    const char *subTableKey,
    const char *itemKey,
    int32_t *pLength,
    UErrorCode *pErrorCode);

/*returns TRUE if a is an ID separator FALSE otherwise*/
#define _isIDSeparator(a) (a == '_' || a == '-')

U_CFUNC const char* 
uloc_getCurrentCountryID(const char* oldID);

U_CFUNC const char* 
uloc_getCurrentLanguageID(const char* oldID);

U_CFUNC int32_t
ulocimp_getLanguage(const char *localeID,
                    char *language, int32_t languageCapacity,
                    const char **pEnd);

U_CFUNC int32_t
ulocimp_getScript(const char *localeID,
                   char *script, int32_t scriptCapacity,
                   const char **pEnd);

U_CFUNC int32_t
ulocimp_getCountry(const char *localeID,
                   char *country, int32_t countryCapacity,
                   const char **pEnd);

/**
 * Writes a well-formed language tag for this locale ID.
 *
 * **Note**: When `strict` is FALSE, any locale fields which do not satisfy the
 * BCP47 syntax requirement will be omitted from the result.  When `strict` is
 * TRUE, this function sets U_ILLEGAL_ARGUMENT_ERROR to the `err` if any locale
 * fields do not satisfy the BCP47 syntax requirement.
 *
 * @param localeID  the input locale ID
 * @param sink      the output sink receiving the BCP47 language
 *                  tag for this Locale.
 * @param strict    boolean value indicating if the function returns
 *                  an error for an ill-formed input locale ID.
 * @param err       error information if receiving the language
 *                  tag failed.
 * @return          The length of the BCP47 language tag.
 *
 * @internal ICU 64
 */
U_STABLE void U_EXPORT2
ulocimp_toLanguageTag(const char* localeID,
                      icu::ByteSink& sink,
                      UBool strict,
                      UErrorCode* err);

/**
 * Returns a locale ID for the specified BCP47 language tag string.
 * If the specified language tag contains any ill-formed subtags,
 * the first such subtag and all following subtags are ignored.
 * <p>
 * This implements the 'Language-Tag' production of BCP47, and so
 * supports grandfathered (regular and irregular) as well as private
 * use language tags.  Private use tags are represented as 'x-whatever',
 * and grandfathered tags are converted to their canonical replacements
 * where they exist.  Note that a few grandfathered tags have no modern
 * replacement, these will be converted using the fallback described in
 * the first paragraph, so some information might be lost.
 * @param langtag   the input BCP47 language tag.
 * @param tagLen    the length of langtag, or -1 to call uprv_strlen().
 * @param sink      the output sink receiving a locale ID for the
 *                  specified BCP47 language tag.
 * @param parsedLength  if not NULL, successfully parsed length
 *                      for the input language tag is set.
 * @param err       error information if receiving the locald ID
 *                  failed.
 * @internal ICU 63
 */
U_CAPI void U_EXPORT2
ulocimp_forLanguageTag(const char* langtag,
                       int32_t tagLen,
                       icu::ByteSink& sink,
                       int32_t* parsedLength,
                       UErrorCode* err);

/**
 * Get the region to use for supplemental data lookup. Uses
 * (1) any region specified by locale tag "rg"; if none then
 * (2) any unicode_region_tag in the locale ID; if none then
 * (3) if inferRegion is TRUE, the region suggested by
 * getLikelySubtags on the localeID.
 * If no region is found, returns length 0.
 * 
 * @param localeID
 *     The complete locale ID (with keywords) from which
 *     to get the region to use for supplemental data.
 * @param inferRegion
 *     If TRUE, will try to infer region from localeID if
 *     no other region is found.
 * @param region
 *     Buffer in which to put the region ID found; should
 *     have a capacity at least ULOC_COUNTRY_CAPACITY. 
 * @param regionCapacity
 *     The actual capacity of the region buffer.
 * @param status
 *     Pointer to in/out UErrorCode value for latest status.
 * @return
 *     The length of any region code found, or 0 if none.
 * @internal ICU 57
 */
U_CAPI int32_t U_EXPORT2
ulocimp_getRegionForSupplementalData(const char *localeID, UBool inferRegion,
                                     char *region, int32_t regionCapacity, UErrorCode* status);

/**
 * Add the likely subtags for a provided locale ID, per the algorithm described
 * in the following CLDR technical report:
 *
 *   http://www.unicode.org/reports/tr35/#Likely_Subtags
 *
 * If localeID is already in the maximal form, or there is no data available
 * for maximization, it will be copied to the output buffer.  For example,
 * "und-Zzzz" cannot be maximized, since there is no reasonable maximization.
 *
 * Examples:
 *
 * "en" maximizes to "en_Latn_US"
 *
 * "de" maximizes to "de_Latn_US"
 *
 * "sr" maximizes to "sr_Cyrl_RS"
 *
 * "sh" maximizes to "sr_Latn_RS" (Note this will not reverse.)
 *
 * "zh_Hani" maximizes to "zh_Hans_CN" (Note this will not reverse.)
 *
 * @param localeID The locale to maximize
 * @param sink The output sink receiving the maximized locale
 * @param err Error information if maximizing the locale failed.  If the length
 * of the localeID and the null-terminator is greater than the maximum allowed size,
 * or the localeId is not well-formed, the error code is U_ILLEGAL_ARGUMENT_ERROR.
 * @internal ICU 64
 */
U_STABLE void U_EXPORT2
ulocimp_addLikelySubtags(const char* localeID,
                         icu::ByteSink& sink,
                         UErrorCode* err);

/**
 * Minimize the subtags for a provided locale ID, per the algorithm described
 * in the following CLDR technical report:
 *
 *   http://www.unicode.org/reports/tr35/#Likely_Subtags
 *
 * If localeID is already in the minimal form, or there is no data available
 * for minimization, it will be copied to the output buffer.  Since the
 * minimization algorithm relies on proper maximization, see the comments
 * for ulocimp_addLikelySubtags for reasons why there might not be any data.
 *
 * Examples:
 *
 * "en_Latn_US" minimizes to "en"
 *
 * "de_Latn_US" minimizes to "de"
 *
 * "sr_Cyrl_RS" minimizes to "sr"
 *
 * "zh_Hant_TW" minimizes to "zh_TW" (The region is preferred to the
 * script, and minimizing to "zh" would imply "zh_Hans_CN".)
 *
 * @param localeID The locale to minimize
 * @param sink The output sink receiving the maximized locale
 * @param err Error information if minimizing the locale failed.  If the length
 * of the localeID and the null-terminator is greater than the maximum allowed size,
 * or the localeId is not well-formed, the error code is U_ILLEGAL_ARGUMENT_ERROR.
 * @internal ICU 64
 */
U_STABLE void U_EXPORT2
ulocimp_minimizeSubtags(const char* localeID,
                        icu::ByteSink& sink,
                        UErrorCode* err);

U_CAPI const char * U_EXPORT2
locale_getKeywordsStart(const char *localeID);

U_CFUNC UBool
ultag_isExtensionSubtags(const char* s, int32_t len);

U_CFUNC UBool
ultag_isLanguageSubtag(const char* s, int32_t len);

U_CFUNC UBool
ultag_isPrivateuseValueSubtags(const char* s, int32_t len);

U_CFUNC UBool
ultag_isRegionSubtag(const char* s, int32_t len);

U_CFUNC UBool
ultag_isScriptSubtag(const char* s, int32_t len);

U_CFUNC UBool
ultag_isTransformedExtensionSubtags(const char* s, int32_t len);

U_CFUNC UBool
ultag_isUnicodeExtensionSubtags(const char* s, int32_t len);

U_CFUNC UBool
ultag_isUnicodeLocaleAttribute(const char* s, int32_t len);

U_CFUNC UBool
ultag_isUnicodeLocaleAttributes(const char* s, int32_t len);

U_CFUNC UBool
ultag_isUnicodeLocaleKey(const char* s, int32_t len);

U_CFUNC UBool
ultag_isUnicodeLocaleType(const char* s, int32_t len);

U_CFUNC UBool
ultag_isVariantSubtags(const char* s, int32_t len);

U_CFUNC const char*
ulocimp_toBcpKey(const char* key);

U_CFUNC const char*
ulocimp_toLegacyKey(const char* key);

U_CFUNC const char*
ulocimp_toBcpType(const char* key, const char* type, UBool* isKnownKey, UBool* isSpecialType);

U_CFUNC const char*
ulocimp_toLegacyType(const char* key, const char* type, UBool* isKnownKey, UBool* isSpecialType);


// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License



#if U_SHOW_CPLUSPLUS_API


// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License

// localematcher.h
// created: 2019may08 Markus W. Scherer




#if U_SHOW_CPLUSPLUS_API





/**
 * \file
 * \brief C++ API: Locale matcher: User's desired locales vs. application's supported locales.
 */

#ifndef U_HIDE_DRAFT_API

/**
 * Builder option for whether the language subtag or the script subtag is most important.
 *
 * @see Builder#setFavorSubtag(FavorSubtag)
 * @draft ICU 65
 */
enum ULocMatchFavorSubtag {
    /**
     * Language differences are most important, then script differences, then region differences.
     * (This is the default behavior.)
     *
     * @draft ICU 65
     */
    ULOCMATCH_FAVOR_LANGUAGE,
    /**
     * Makes script differences matter relatively more than language differences.
     *
     * @draft ICU 65
     */
    ULOCMATCH_FAVOR_SCRIPT
};
#ifndef U_IN_DOXYGEN
typedef enum ULocMatchFavorSubtag ULocMatchFavorSubtag;
#endif

/**
 * Builder option for whether all desired locales are treated equally or
 * earlier ones are preferred.
 *
 * @see Builder#setDemotionPerDesiredLocale(Demotion)
 * @draft ICU 65
 */
enum ULocMatchDemotion {
    /**
     * All desired locales are treated equally.
     *
     * @draft ICU 65
     */
    ULOCMATCH_DEMOTION_NONE,
    /**
     * Earlier desired locales are preferred.
     *
     * <p>From each desired locale to the next,
     * the distance to any supported locale is increased by an additional amount
     * which is at least as large as most region mismatches.
     * A later desired locale has to have a better match with some supported locale
     * due to more than merely having the same region subtag.
     *
     * <p>For example: <code>Supported={en, sv}  desired=[en-GB, sv]</code>
     * yields <code>Result(en-GB, en)</code> because
     * with the demotion of sv its perfect match is no better than
     * the region distance between the earlier desired locale en-GB and en=en-US.
     *
     * <p>Notes:
     * <ul>
     *   <li>In some cases, language and/or script differences can be as small as
     *       the typical region difference. (Example: sr-Latn vs. sr-Cyrl)
     *   <li>It is possible for certain region differences to be larger than usual,
     *       and larger than the demotion.
     *       (As of CLDR 35 there is no such case, but
     *        this is possible in future versions of the data.)
     * </ul>
     *
     * @draft ICU 65
     */
    ULOCMATCH_DEMOTION_REGION
};
#ifndef U_IN_DOXYGEN
typedef enum ULocMatchDemotion ULocMatchDemotion;
#endif

struct UHashtable;

U_NAMESPACE_BEGIN

struct LSR;

class LocaleDistance;
class LocaleLsrIterator;
class UVector;
class XLikelySubtags;

/**
 * Immutable class that picks the best match between a user's desired locales and
 * an application's supported locales.
 * Movable but not copyable.
 *
 * <p>Example:
 * <pre>
 * UErrorCode errorCode = U_ZERO_ERROR;
 * LocaleMatcher matcher = LocaleMatcher::Builder().setSupportedLocales("fr, en-GB, en").build(errorCode);
 * Locale *bestSupported = matcher.getBestLocale(Locale.US, errorCode);  // "en"
 * </pre>
 *
 * <p>A matcher takes into account when languages are close to one another,
 * such as Danish and Norwegian,
 * and when regional variants are close, like en-GB and en-AU as opposed to en-US.
 *
 * <p>If there are multiple supported locales with the same (language, script, region)
 * likely subtags, then the current implementation returns the first of those locales.
 * It ignores variant subtags (except for pseudolocale variants) and extensions.
 * This may change in future versions.
 *
 * <p>For example, the current implementation does not distinguish between
 * de, de-DE, de-Latn, de-1901, de-u-co-phonebk.
 *
 * <p>If you prefer one equivalent locale over another, then provide only the preferred one,
 * or place it earlier in the list of supported locales.
 *
 * <p>Otherwise, the order of supported locales may have no effect on the best-match results.
 * The current implementation compares each desired locale with supported locales
 * in the following order:
 * 1. Default locale, if supported;
 * 2. CLDR "paradigm locales" like en-GB and es-419;
 * 3. other supported locales.
 * This may change in future versions.
 *
 * <p>Often a product will just need one matcher instance, built with the languages
 * that it supports. However, it may want multiple instances with different
 * default languages based on additional information, such as the domain.
 *
 * <p>This class is not intended for public subclassing.
 *
 * @draft ICU 65
 */
class U_COMMON_API LocaleMatcher : public UMemory {
public:
    /**
     * Data for the best-matching pair of a desired and a supported locale.
     * Movable but not copyable.
     *
     * @draft ICU 65
     */
    class U_COMMON_API Result : public UMemory {
    public:
        /**
         * Move constructor; might modify the source.
         * This object will have the same contents that the source object had.
         *
         * @param src Result to move contents from.
         * @draft ICU 65
         */
        Result(Result &&src) U_NOEXCEPT;

        /**
         * Destructor.
         *
         * @draft ICU 65
         */
        ~Result();

        /**
         * Move assignment; might modify the source.
         * This object will have the same contents that the source object had.
         *
         * @param src Result to move contents from.
         * @draft ICU 65
         */
        Result &operator=(Result &&src) U_NOEXCEPT;

        /**
         * Returns the best-matching desired locale.
         * nullptr if the list of desired locales is empty or if none matched well enough.
         *
         * @return the best-matching desired locale, or nullptr.
         * @draft ICU 65
         */
        inline const Locale *getDesiredLocale() const { return desiredLocale; }

        /**
         * Returns the best-matching supported locale.
         * If none matched well enough, this is the default locale.
         * The default locale is nullptr if the list of supported locales is empty and
         * no explicit default locale is set.
         *
         * @return the best-matching supported locale, or nullptr.
         * @draft ICU 65
         */
        inline const Locale *getSupportedLocale() const { return supportedLocale; }

        /**
         * Returns the index of the best-matching desired locale in the input Iterable order.
         * -1 if the list of desired locales is empty or if none matched well enough.
         *
         * @return the index of the best-matching desired locale, or -1.
         * @draft ICU 65
         */
        inline int32_t getDesiredIndex() const { return desiredIndex; }

        /**
         * Returns the index of the best-matching supported locale in the
         * constructor’s or builder’s input order (“set” Collection plus “added” locales).
         * If the matcher was built from a locale list string, then the iteration order is that
         * of a LocalePriorityList built from the same string.
         * -1 if the list of supported locales is empty or if none matched well enough.
         *
         * @return the index of the best-matching supported locale, or -1.
         * @draft ICU 65
         */
        inline int32_t getSupportedIndex() const { return supportedIndex; }

        /**
         * Takes the best-matching supported locale and adds relevant fields of the
         * best-matching desired locale, such as the -t- and -u- extensions.
         * May replace some fields of the supported locale.
         * The result is the locale that should be used for date and number formatting, collation, etc.
         * Returns the root locale if getSupportedLocale() returns nullptr.
         *
         * <p>Example: desired=ar-SA-u-nu-latn, supported=ar-EG, resolved locale=ar-SA-u-nu-latn
         *
         * @return a locale combining the best-matching desired and supported locales.
         * @draft ICU 65
         */
        Locale makeResolvedLocale(UErrorCode &errorCode) const;

    private:
        Result(const Locale *desired, const Locale *supported,
               int32_t desIndex, int32_t suppIndex, UBool owned) :
                desiredLocale(desired), supportedLocale(supported),
                desiredIndex(desIndex), supportedIndex(suppIndex),
                desiredIsOwned(owned) {}

        Result(const Result &other) = delete;
        Result &operator=(const Result &other) = delete;

        const Locale *desiredLocale;
        const Locale *supportedLocale;
        int32_t desiredIndex;
        int32_t supportedIndex;
        UBool desiredIsOwned;

        friend class LocaleMatcher;
    };

    /**
     * LocaleMatcher builder.
     * Movable but not copyable.
     *
     * @see LocaleMatcher#builder()
     * @draft ICU 65
     */
    class U_COMMON_API Builder : public UMemory {
    public:
        /**
         * Constructs a builder used in chaining parameters for building a LocaleMatcher.
         *
         * @return a new Builder object
         * @draft ICU 65
         */
        Builder() {}

        /**
         * Move constructor; might modify the source.
         * This builder will have the same contents that the source builder had.
         *
         * @param src Builder to move contents from.
         * @draft ICU 65
         */
        Builder(Builder &&src) U_NOEXCEPT;

        /**
         * Destructor.
         *
         * @draft ICU 65
         */
        ~Builder();

        /**
         * Move assignment; might modify the source.
         * This builder will have the same contents that the source builder had.
         *
         * @param src Builder to move contents from.
         * @draft ICU 65
         */
        Builder &operator=(Builder &&src) U_NOEXCEPT;

        /**
         * Parses an Accept-Language string
         * (<a href="https://tools.ietf.org/html/rfc2616#section-14.4">RFC 2616 Section 14.4</a>),
         * such as "af, en, fr;q=0.9", and sets the supported locales accordingly.
         * Allows whitespace in more places but does not allow "*".
         * Clears any previously set/added supported locales first.
         *
         * @param locales the Accept-Language string of locales to set
         * @return this Builder object
         * @draft ICU 65
         */
        Builder &setSupportedLocalesFromListString(StringPiece locales);

        /**
         * Copies the supported locales, preserving iteration order.
         * Clears any previously set/added supported locales first.
         * Duplicates are allowed, and are not removed.
         *
         * @param locales the list of locale
         * @return this Builder object
         * @draft ICU 65
         */
        Builder &setSupportedLocales(Locale::Iterator &locales);

        /**
         * Copies the supported locales from the begin/end range, preserving iteration order.
         * Clears any previously set/added supported locales first.
         * Duplicates are allowed, and are not removed.
         *
         * Each of the iterator parameter values must be an
         * input iterator whose value is convertible to const Locale &.
         *
         * @param begin Start of range.
         * @param end Exclusive end of range.
         * @return this Builder object
         * @draft ICU 65
         */
        template<typename Iter>
        Builder &setSupportedLocales(Iter begin, Iter end) {
            if (U_FAILURE(errorCode_)) { return *this; }
            clearSupportedLocales();
            while (begin != end) {
                addSupportedLocale(*begin++);
            }
            return *this;
        }

        /**
         * Copies the supported locales from the begin/end range, preserving iteration order.
         * Calls the converter to convert each *begin to a Locale or const Locale &.
         * Clears any previously set/added supported locales first.
         * Duplicates are allowed, and are not removed.
         *
         * Each of the iterator parameter values must be an
         * input iterator whose value is convertible to const Locale &.
         *
         * @param begin Start of range.
         * @param end Exclusive end of range.
         * @param converter Converter from *begin to const Locale & or compatible.
         * @return this Builder object
         * @draft ICU 65
         */
        template<typename Iter, typename Conv>
        Builder &setSupportedLocalesViaConverter(Iter begin, Iter end, Conv converter) {
            if (U_FAILURE(errorCode_)) { return *this; }
            clearSupportedLocales();
            while (begin != end) {
                addSupportedLocale(converter(*begin++));
            }
            return *this;
        }

        /**
         * Adds another supported locale.
         * Duplicates are allowed, and are not removed.
         *
         * @param locale another locale
         * @return this Builder object
         * @draft ICU 65
         */
        Builder &addSupportedLocale(const Locale &locale);

        /**
         * Sets the default locale; if nullptr, or if it is not set explicitly,
         * then the first supported locale is used as the default locale.
         *
         * @param defaultLocale the default locale (will be copied)
         * @return this Builder object
         * @draft ICU 65
         */
        Builder &setDefaultLocale(const Locale *defaultLocale);

        /**
         * If ULOCMATCH_FAVOR_SCRIPT, then the language differences are smaller than script
         * differences.
         * This is used in situations (such as maps) where
         * it is better to fall back to the same script than a similar language.
         *
         * @param subtag the subtag to favor
         * @return this Builder object
         * @draft ICU 65
         */
        Builder &setFavorSubtag(ULocMatchFavorSubtag subtag);

        /**
         * Option for whether all desired locales are treated equally or
         * earlier ones are preferred (this is the default).
         *
         * @param demotion the demotion per desired locale to set.
         * @return this Builder object
         * @draft ICU 65
         */
        Builder &setDemotionPerDesiredLocale(ULocMatchDemotion demotion);

        /**
         * Sets the UErrorCode if an error occurred while setting parameters.
         * Preserves older error codes in the outErrorCode.
         *
         * @param outErrorCode Set to an error code if it does not contain one already
         *                  and an error occurred while setting parameters.
         *                  Otherwise unchanged.
         * @return TRUE if U_FAILURE(outErrorCode)
         * @draft ICU 65
         */
        UBool copyErrorTo(UErrorCode &outErrorCode) const;

        /**
         * Builds and returns a new locale matcher.
         * This builder can continue to be used.
         *
         * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
         *                  or else the function returns immediately. Check for U_FAILURE()
         *                  on output or use with function chaining. (See User Guide for details.)
         * @return new LocaleMatcher.
         * @draft ICU 65
         */
        LocaleMatcher build(UErrorCode &errorCode) const;

    private:
        friend class LocaleMatcher;

        Builder(const Builder &other) = delete;
        Builder &operator=(const Builder &other) = delete;

        void clearSupportedLocales();
        bool ensureSupportedLocaleVector();

        UErrorCode errorCode_ = U_ZERO_ERROR;
        UVector *supportedLocales_ = nullptr;
        int32_t thresholdDistance_ = -1;
        ULocMatchDemotion demotion_ = ULOCMATCH_DEMOTION_REGION;
        Locale *defaultLocale_ = nullptr;
        ULocMatchFavorSubtag favor_ = ULOCMATCH_FAVOR_LANGUAGE;
    };

    // FYI No public LocaleMatcher constructors in C++; use the Builder.

    /**
     * Move copy constructor; might modify the source.
     * This matcher will have the same settings that the source matcher had.
     * @param src source matcher
     * @draft ICU 65
     */
    LocaleMatcher(LocaleMatcher &&src) U_NOEXCEPT;

    /**
     * Destructor.
     * @draft ICU 65
     */
    ~LocaleMatcher();

    /**
     * Move assignment operator; might modify the source.
     * This matcher will have the same settings that the source matcher had.
     * The behavior is undefined if *this and src are the same object.
     * @param src source matcher
     * @return *this
     * @draft ICU 65
     */
    LocaleMatcher &operator=(LocaleMatcher &&src) U_NOEXCEPT;

    /**
     * Returns the supported locale which best matches the desired locale.
     *
     * @param desiredLocale Typically a user's language.
     * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
     *                  or else the function returns immediately. Check for U_FAILURE()
     *                  on output or use with function chaining. (See User Guide for details.)
     * @return the best-matching supported locale.
     * @draft ICU 65
     */
    const Locale *getBestMatch(const Locale &desiredLocale, UErrorCode &errorCode) const;

    /**
     * Returns the supported locale which best matches one of the desired locales.
     *
     * @param desiredLocales Typically a user's languages, in order of preference (descending).
     * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
     *                  or else the function returns immediately. Check for U_FAILURE()
     *                  on output or use with function chaining. (See User Guide for details.)
     * @return the best-matching supported locale.
     * @draft ICU 65
     */
    const Locale *getBestMatch(Locale::Iterator &desiredLocales, UErrorCode &errorCode) const;

    /**
     * Parses an Accept-Language string
     * (<a href="https://tools.ietf.org/html/rfc2616#section-14.4">RFC 2616 Section 14.4</a>),
     * such as "af, en, fr;q=0.9",
     * and returns the supported locale which best matches one of the desired locales.
     * Allows whitespace in more places but does not allow "*".
     *
     * @param desiredLocaleList Typically a user's languages, as an Accept-Language string.
     * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
     *                  or else the function returns immediately. Check for U_FAILURE()
     *                  on output or use with function chaining. (See User Guide for details.)
     * @return the best-matching supported locale.
     * @draft ICU 65
     */
    const Locale *getBestMatchForListString(StringPiece desiredLocaleList, UErrorCode &errorCode) const;

    /**
     * Returns the best match between the desired locale and the supported locales.
     * If the result's desired locale is not nullptr, then it is the address of the input locale.
     * It has not been cloned.
     *
     * @param desiredLocale Typically a user's language.
     * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
     *                  or else the function returns immediately. Check for U_FAILURE()
     *                  on output or use with function chaining. (See User Guide for details.)
     * @return the best-matching pair of the desired and a supported locale.
     * @draft ICU 65
     */
    Result getBestMatchResult(const Locale &desiredLocale, UErrorCode &errorCode) const;

    /**
     * Returns the best match between the desired and supported locales.
     * If the result's desired locale is not nullptr, then it is a clone of
     * the best-matching desired locale. The Result object owns the clone.
     *
     * @param desiredLocales Typically a user's languages, in order of preference (descending).
     * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
     *                  or else the function returns immediately. Check for U_FAILURE()
     *                  on output or use with function chaining. (See User Guide for details.)
     * @return the best-matching pair of a desired and a supported locale.
     * @draft ICU 65
     */
    Result getBestMatchResult(Locale::Iterator &desiredLocales, UErrorCode &errorCode) const;

#ifndef U_HIDE_INTERNAL_API
    /**
     * Returns a fraction between 0 and 1, where 1 means that the languages are a
     * perfect match, and 0 means that they are completely different.
     *
     * <p>This is mostly an implementation detail, and the precise values may change over time.
     * The implementation may use either the maximized forms or the others ones, or both.
     * The implementation may or may not rely on the forms to be consistent with each other.
     *
     * <p>Callers should construct and use a matcher rather than match pairs of locales directly.
     *
     * @param desired Desired locale.
     * @param supported Supported locale.
     * @param errorCode ICU error code. Its input value must pass the U_SUCCESS() test,
     *                  or else the function returns immediately. Check for U_FAILURE()
     *                  on output or use with function chaining. (See User Guide for details.)
     * @return value between 0 and 1, inclusive.
     * @internal (has a known user)
     */
    double internalMatch(const Locale &desired, const Locale &supported, UErrorCode &errorCode) const;
#endif  // U_HIDE_INTERNAL_API

private:
    LocaleMatcher(const Builder &builder, UErrorCode &errorCode);
    LocaleMatcher(const LocaleMatcher &other) = delete;
    LocaleMatcher &operator=(const LocaleMatcher &other) = delete;

    int32_t getBestSuppIndex(LSR desiredLSR, LocaleLsrIterator *remainingIter, UErrorCode &errorCode) const;

    const XLikelySubtags &likelySubtags;
    const LocaleDistance &localeDistance;
    int32_t thresholdDistance;
    int32_t demotionPerDesiredLocale;
    ULocMatchFavorSubtag favorSubtag;

    // These are in input order.
    const Locale ** supportedLocales;
    LSR *lsrs;
    int32_t supportedLocalesLength;
    // These are in preference order: 1. Default locale 2. paradigm locales 3. others.
    UHashtable *supportedLsrToIndex;  // Map<LSR, Integer> stores index+1 because 0 is "not found"
    // Array versions of the supportedLsrToIndex keys and values.
    // The distance lookup loops over the supportedLSRs and returns the index of the best match.
    const LSR **supportedLSRs;
    int32_t *supportedIndexes;
    int32_t supportedLSRsLength;
    Locale *ownedDefaultLocale;
    const Locale *defaultLocale;
    int32_t defaultLocaleIndex;
};

U_NAMESPACE_END

#endif  // U_HIDE_DRAFT_API
#endif  // U_SHOW_CPLUSPLUS_API




#ifndef U_HIDE_DRAFT_API
/**
 * \file
 * \brief C++ API: Builder API for Locale
 */

U_NAMESPACE_BEGIN
class CharString;

/**
 * <code>LocaleBuilder</code> is used to build instances of <code>Locale</code>
 * from values configured by the setters.  Unlike the <code>Locale</code>
 * constructors, the <code>LocaleBuilder</code> checks if a value configured by a
 * setter satisfies the syntax requirements defined by the <code>Locale</code>
 * class.  A <code>Locale</code> object created by a <code>LocaleBuilder</code> is
 * well-formed and can be transformed to a well-formed IETF BCP 47 language tag
 * without losing information.
 *
 * <p>The following example shows how to create a <code>Locale</code> object
 * with the <code>LocaleBuilder</code>.
 * <blockquote>
 * <pre>
 *     UErrorCode status = U_ZERO_ERROR;
 *     Locale aLocale = LocaleBuilder()
 *                          .setLanguage("sr")
 *                          .setScript("Latn")
 *                          .setRegion("RS")
 *                          .build(status);
 *     if (U_SUCCESS(status)) {
 *       // ...
 *     }
 * </pre>
 * </blockquote>
 *
 * <p>LocaleBuilders can be reused; <code>clear()</code> resets all
 * fields to their default values.
 *
 * <p>LocaleBuilder tracks errors in an internal UErrorCode. For all setters,
 * except setLanguageTag and setLocale, LocaleBuilder will return immediately
 * if the internal UErrorCode is in error state.
 * To reset internal state and error code, call clear method.
 * The setLanguageTag and setLocale method will first clear the internal
 * UErrorCode, then track the error of the validation of the input parameter
 * into the internal UErrorCode.
 *
 * @draft ICU 64
 */
class U_COMMON_API LocaleBuilder : public UObject {
public:
    /**
     * Constructs an empty LocaleBuilder. The default value of all
     * fields, extensions, and private use information is the
     * empty string.
     *
     * @draft ICU 64
     */
    LocaleBuilder();

    /**
     * Destructor
     * @draft ICU 64
     */
    virtual ~LocaleBuilder();

    /**
     * Resets the <code>LocaleBuilder</code> to match the provided
     * <code>locale</code>.  Existing state is discarded.
     *
     * <p>All fields of the locale must be well-formed.
     * <p>This method clears the internal UErrorCode.
     *
     * @param locale the locale
     * @return This builder.
     *
     * @draft ICU 64
     */
    LocaleBuilder& setLocale(const Locale& locale);

    /**
     * Resets the LocaleBuilder to match the provided
     * [Unicode Locale Identifier](http://www.unicode.org/reports/tr35/tr35.html#unicode_locale_id) .
     * Discards the existing state. the empty string cause the builder to be
     * reset, like {@link #clear}.  Grandfathered tags are converted to their
     * canonical form before being processed.  Otherwise, the <code>language
     * tag</code> must be well-formed, or else the build() method will later
     * report an U_ILLEGAL_ARGUMENT_ERROR.
     *
     * <p>This method clears the internal UErrorCode.
     *
     * @param tag the language tag, defined as
     *   [unicode_locale_id](http://www.unicode.org/reports/tr35/tr35.html#unicode_locale_id).
     * @return This builder.
     * @draft ICU 64
     */
    LocaleBuilder& setLanguageTag(StringPiece tag);

    /**
     * Sets the language.  If <code>language</code> is the empty string, the
     * language in this <code>LocaleBuilder</code> is removed. Otherwise, the
     * <code>language</code> must be well-formed, or else the build() method will
     * later report an U_ILLEGAL_ARGUMENT_ERROR.
     *
     * <p>The syntax of language value is defined as
     * [unicode_language_subtag](http://www.unicode.org/reports/tr35/tr35.html#unicode_language_subtag).
     *
     * @param language the language
     * @return This builder.
     * @draft ICU 64
     */
    LocaleBuilder& setLanguage(StringPiece language);

    /**
     * Sets the script. If <code>script</code> is the empty string, the script in
     * this <code>LocaleBuilder</code> is removed.
     * Otherwise, the <code>script</code> must be well-formed, or else the build()
     * method will later report an U_ILLEGAL_ARGUMENT_ERROR.
     *
     * <p>The script value is a four-letter script code as
     * [unicode_script_subtag](http://www.unicode.org/reports/tr35/tr35.html#unicode_script_subtag)
     * defined by ISO 15924
     *
     * @param script the script
     * @return This builder.
     * @draft ICU 64
     */
    LocaleBuilder& setScript(StringPiece script);

    /**
     * Sets the region.  If region is the empty string, the region in this
     * <code>LocaleBuilder</code> is removed. Otherwise, the <code>region</code>
     * must be well-formed, or else the build() method will later report an
     * U_ILLEGAL_ARGUMENT_ERROR.
     *
     * <p>The region value is defined by
     *  [unicode_region_subtag](http://www.unicode.org/reports/tr35/tr35.html#unicode_region_subtag)
     * as a two-letter ISO 3166 code or a three-digit UN M.49 area code.
     *
     * <p>The region value in the <code>Locale</code> created by the
     * <code>LocaleBuilder</code> is always normalized to upper case.
     *
     * @param region the region
     * @return This builder.
     * @draft ICU 64
     */
    LocaleBuilder& setRegion(StringPiece region);

    /**
     * Sets the variant.  If variant is the empty string, the variant in this
     * <code>LocaleBuilder</code> is removed.  Otherwise, the <code>variant</code>
     * must be well-formed, or else the build() method will later report an
     * U_ILLEGAL_ARGUMENT_ERROR.
     *
     * <p><b>Note:</b> This method checks if <code>variant</code>
     * satisfies the
     * [unicode_variant_subtag](http://www.unicode.org/reports/tr35/tr35.html#unicode_variant_subtag)
     * syntax requirements, and normalizes the value to lowercase letters. However,
     * the <code>Locale</code> class does not impose any syntactic
     * restriction on variant. To set an ill-formed variant, use a Locale constructor.
     * If there are multiple unicode_variant_subtag, the caller must concatenate
     * them with '-' as separator (ex: "foobar-fibar").
     *
     * @param variant the variant
     * @return This builder.
     * @draft ICU 64
     */
    LocaleBuilder& setVariant(StringPiece variant);

    /**
     * Sets the extension for the given key. If the value is the empty string,
     * the extension is removed.  Otherwise, the <code>key</code> and
     * <code>value</code> must be well-formed, or else the build() method will
     * later report an U_ILLEGAL_ARGUMENT_ERROR.
     *
     * <p><b>Note:</b> The key ('u') is used for the Unicode locale extension.
     * Setting a value for this key replaces any existing Unicode locale key/type
     * pairs with those defined in the extension.
     *
     * <p><b>Note:</b> The key ('x') is used for the private use code. To be
     * well-formed, the value for this key needs only to have subtags of one to
     * eight alphanumeric characters, not two to eight as in the general case.
     *
     * @param key the extension key
     * @param value the extension value
     * @return This builder.
     * @draft ICU 64
     */
    LocaleBuilder& setExtension(char key, StringPiece value);

    /**
     * Sets the Unicode locale keyword type for the given key. If the type
     * StringPiece is constructed with a nullptr, the keyword is removed.
     * If the type is the empty string, the keyword is set without type subtags.
     * Otherwise, the key and type must be well-formed, or else the build()
     * method will later report an U_ILLEGAL_ARGUMENT_ERROR.
     *
     * <p>Keys and types are converted to lower case.
     *
     * <p><b>Note</b>:Setting the 'u' extension via {@link #setExtension}
     * replaces all Unicode locale keywords with those defined in the
     * extension.
     *
     * @param key the Unicode locale key
     * @param type the Unicode locale type
     * @return This builder.
     * @draft ICU 64
     */
    LocaleBuilder& setUnicodeLocaleKeyword(
        StringPiece key, StringPiece type);

    /**
     * Adds a unicode locale attribute, if not already present, otherwise
     * has no effect.  The attribute must not be empty string and must be
     * well-formed or U_ILLEGAL_ARGUMENT_ERROR will be set to status
     * during the build() call.
     *
     * @param attribute the attribute
     * @return This builder.
     * @draft ICU 64
     */
    LocaleBuilder& addUnicodeLocaleAttribute(StringPiece attribute);

    /**
     * Removes a unicode locale attribute, if present, otherwise has no
     * effect.  The attribute must not be empty string and must be well-formed
     * or U_ILLEGAL_ARGUMENT_ERROR will be set to status during the build() call.
     *
     * <p>Attribute comparison for removal is case-insensitive.
     *
     * @param attribute the attribute
     * @return This builder.
     * @draft ICU 64
     */
    LocaleBuilder& removeUnicodeLocaleAttribute(StringPiece attribute);

    /**
     * Resets the builder to its initial, empty state.
     * <p>This method clears the internal UErrorCode.
     *
     * @return this builder
     * @draft ICU 64
     */
    LocaleBuilder& clear();

    /**
     * Resets the extensions to their initial, empty state.
     * Language, script, region and variant are unchanged.
     *
     * @return this builder
     * @draft ICU 64
     */
    LocaleBuilder& clearExtensions();

    /**
     * Returns an instance of <code>Locale</code> created from the fields set
     * on this builder.
     * If any set methods or during the build() call require memory allocation
     * but fail U_MEMORY_ALLOCATION_ERROR will be set to status.
     * If any of the fields set by the setters are not well-formed, the status
     * will be set to U_ILLEGAL_ARGUMENT_ERROR. The state of the builder will
     * not change after the build() call and the caller is free to keep using
     * the same builder to build more locales.
     *
     * @return a new Locale
     * @draft ICU 64
     */
    Locale build(UErrorCode& status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Sets the UErrorCode if an error occurred while recording sets.
     * Preserves older error codes in the outErrorCode.
     * @param outErrorCode Set to an error code that occurred while setting subtags.
     *                  Unchanged if there is no such error or if outErrorCode
     *                  already contained an error.
     * @return TRUE if U_FAILURE(outErrorCode)
     * @draft ICU 65
     */
    UBool copyErrorTo(UErrorCode &outErrorCode) const;
#endif  /* U_HIDE_DRAFT_API */

private:
    friend class LocaleMatcher::Result;

    void copyExtensionsFrom(const Locale& src, UErrorCode& errorCode);

    UErrorCode status_;
    char language_[9];
    char script_[5];
    char region_[4];
    CharString *variant_;  // Pointer not object so we need not #include internal charstr.h.
    icu::Locale *extensions_;  // Pointer not object. Storage for all other fields.

};

U_NAMESPACE_END

#endif  // U_HIDE_DRAFT_API

#endif /* U_SHOW_CPLUSPLUS_API */




U_NAMESPACE_BEGIN

#define UPRV_ISDIGIT(c) (((c) >= '0') && ((c) <= '9'))
#define UPRV_ISALPHANUM(c) (uprv_isASCIILetter(c) || UPRV_ISDIGIT(c) )

const char* kAttributeKey = "attribute";

static bool _isExtensionSubtags(char key, const char* s, int32_t len) {
    switch (uprv_tolower(key)) {
        case 'u':
            return ultag_isUnicodeExtensionSubtags(s, len);
        case 't':
            return ultag_isTransformedExtensionSubtags(s, len);
        case 'x':
            return ultag_isPrivateuseValueSubtags(s, len);
        default:
            return ultag_isExtensionSubtags(s, len);
    }
}

LocaleBuilder::LocaleBuilder() : UObject(), status_(U_ZERO_ERROR), language_(),
    script_(), region_(), variant_(nullptr), extensions_(nullptr)
{
    language_[0] = 0;
    script_[0] = 0;
    region_[0] = 0;
}

LocaleBuilder::~LocaleBuilder()
{
    delete variant_;
    delete extensions_;
}

LocaleBuilder& LocaleBuilder::setLocale(const Locale& locale)
{
    clear();
    setLanguage(locale.getLanguage());
    setScript(locale.getScript());
    setRegion(locale.getCountry());
    setVariant(locale.getVariant());
    extensions_ = locale.clone();
    if (extensions_ == nullptr) {
        status_ = U_MEMORY_ALLOCATION_ERROR;
    }
    return *this;
}

LocaleBuilder& LocaleBuilder::setLanguageTag(StringPiece tag)
{
    Locale l = Locale::forLanguageTag(tag, status_);
    if (U_FAILURE(status_)) { return *this; }
    // Because setLocale will reset status_ we need to return
    // first if we have error in forLanguageTag.
    setLocale(l);
    return *this;
}

static void setField(StringPiece input, char* dest, UErrorCode& errorCode,
                     UBool (*test)(const char*, int32_t)) {
    if (U_FAILURE(errorCode)) { return; }
    if (input.empty()) {
        dest[0] = '\0';
    } else if (test(input.data(), input.length())) {
        uprv_memcpy(dest, input.data(), input.length());
        dest[input.length()] = '\0';
    } else {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

LocaleBuilder& LocaleBuilder::setLanguage(StringPiece language)
{
    setField(language, language_, status_, &ultag_isLanguageSubtag);
    return *this;
}

LocaleBuilder& LocaleBuilder::setScript(StringPiece script)
{
    setField(script, script_, status_, &ultag_isScriptSubtag);
    return *this;
}

LocaleBuilder& LocaleBuilder::setRegion(StringPiece region)
{
    setField(region, region_, status_, &ultag_isRegionSubtag);
    return *this;
}

static void transform(char* data, int32_t len) {
    for (int32_t i = 0; i < len; i++, data++) {
        if (*data == '_') {
            *data = '-';
        } else {
            *data = uprv_tolower(*data);
        }
    }
}

LocaleBuilder& LocaleBuilder::setVariant(StringPiece variant)
{
    if (U_FAILURE(status_)) { return *this; }
    if (variant.empty()) {
        delete variant_;
        variant_ = nullptr;
        return *this;
    }
    CharString* new_variant = new CharString(variant, status_);
    if (U_FAILURE(status_)) { return *this; }
    if (new_variant == nullptr) {
        status_ = U_MEMORY_ALLOCATION_ERROR;
        return *this;
    }
    transform(new_variant->data(), new_variant->length());
    if (!ultag_isVariantSubtags(new_variant->data(), new_variant->length())) {
        delete new_variant;
        status_ = U_ILLEGAL_ARGUMENT_ERROR;
        return *this;
    }
    delete variant_;
    variant_ = new_variant;
    return *this;
}

static bool
_isKeywordValue(const char* key, const char* value, int32_t value_len)
{
    if (key[1] == '\0') {
        // one char key
        return (UPRV_ISALPHANUM(uprv_tolower(key[0])) &&
                _isExtensionSubtags(key[0], value, value_len));
    } else if (uprv_strcmp(key, kAttributeKey) == 0) {
        // unicode attributes
        return ultag_isUnicodeLocaleAttributes(value, value_len);
    }
    // otherwise: unicode extension value
    // We need to convert from legacy key/value to unicode
    // key/value
    const char* unicode_locale_key = uloc_toUnicodeLocaleKey(key);
    const char* unicode_locale_type = uloc_toUnicodeLocaleType(key, value);

    return unicode_locale_key && unicode_locale_type &&
           ultag_isUnicodeLocaleKey(unicode_locale_key, -1) &&
           ultag_isUnicodeLocaleType(unicode_locale_type, -1);
}

static void
_copyExtensions(const Locale& from, icu::StringEnumeration *keywords,
                Locale& to, bool validate, UErrorCode& errorCode)
{
    if (U_FAILURE(errorCode)) { return; }
    LocalPointer<icu::StringEnumeration> ownedKeywords;
    if (keywords == nullptr) {
        ownedKeywords.adoptInstead(from.createKeywords(errorCode));
        if (U_FAILURE(errorCode) || ownedKeywords.isNull()) { return; }
        keywords = ownedKeywords.getAlias();
    }
    const char* key;
    while ((key = keywords->next(nullptr, errorCode)) != nullptr) {
        CharString value;
        CharStringByteSink sink(&value);
        from.getKeywordValue(key, sink, errorCode);
        if (U_FAILURE(errorCode)) { return; }
        if (uprv_strcmp(key, kAttributeKey) == 0) {
            transform(value.data(), value.length());
        }
        if (validate &&
            !_isKeywordValue(key, value.data(), value.length())) {
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        to.setKeywordValue(key, value.data(), errorCode);
        if (U_FAILURE(errorCode)) { return; }
    }
}

void static
_clearUAttributesAndKeyType(Locale& locale, UErrorCode& errorCode)
{
    // Clear Unicode attributes
    locale.setKeywordValue(kAttributeKey, "", errorCode);

    // Clear all Unicode keyword values
    LocalPointer<icu::StringEnumeration> iter(locale.createUnicodeKeywords(errorCode));
    if (U_FAILURE(errorCode) || iter.isNull()) { return; }
    const char* key;
    while ((key = iter->next(nullptr, errorCode)) != nullptr) {
        locale.setUnicodeKeywordValue(key, nullptr, errorCode);
    }
}

static void
_setUnicodeExtensions(Locale& locale, const CharString& value, UErrorCode& errorCode)
{
    // Add the unicode extensions to extensions_
    CharString locale_str("und-u-", errorCode);
    locale_str.append(value, errorCode);
    _copyExtensions(
        Locale::forLanguageTag(locale_str.data(), errorCode), nullptr,
        locale, false, errorCode);
}

LocaleBuilder& LocaleBuilder::setExtension(char key, StringPiece value)
{
    if (U_FAILURE(status_)) { return *this; }
    if (!UPRV_ISALPHANUM(key)) {
        status_ = U_ILLEGAL_ARGUMENT_ERROR;
        return *this;
    }
    CharString value_str(value, status_);
    if (U_FAILURE(status_)) { return *this; }
    transform(value_str.data(), value_str.length());
    if (!value_str.isEmpty() &&
            !_isExtensionSubtags(key, value_str.data(), value_str.length())) {
        status_ = U_ILLEGAL_ARGUMENT_ERROR;
        return *this;
    }
    if (extensions_ == nullptr) {
        extensions_ = new Locale();
        if (extensions_ == nullptr) {
            status_ = U_MEMORY_ALLOCATION_ERROR;
            return *this;
        }
    }
    if (uprv_tolower(key) != 'u') {
        // for t, x and others extension.
        extensions_->setKeywordValue(StringPiece(&key, 1), value_str.data(),
                                     status_);
        return *this;
    }
    _clearUAttributesAndKeyType(*extensions_, status_);
    if (U_FAILURE(status_)) { return *this; }
    if (!value.empty()) {
        _setUnicodeExtensions(*extensions_, value_str, status_);
    }
    return *this;
}

LocaleBuilder& LocaleBuilder::setUnicodeLocaleKeyword(
      StringPiece key, StringPiece type)
{
    if (U_FAILURE(status_)) { return *this; }
    if (!ultag_isUnicodeLocaleKey(key.data(), key.length()) ||
            (!type.empty() &&
                 !ultag_isUnicodeLocaleType(type.data(), type.length()))) {
      status_ = U_ILLEGAL_ARGUMENT_ERROR;
      return *this;
    }
    if (extensions_ == nullptr) {
        extensions_ = new Locale();
    }
    if (extensions_ == nullptr) {
        status_ = U_MEMORY_ALLOCATION_ERROR;
        return *this;
    }
    extensions_->setUnicodeKeywordValue(key, type, status_);
    return *this;
}

LocaleBuilder& LocaleBuilder::addUnicodeLocaleAttribute(
    StringPiece value)
{
    CharString value_str(value, status_);
    if (U_FAILURE(status_)) { return *this; }
    transform(value_str.data(), value_str.length());
    if (!ultag_isUnicodeLocaleAttribute(value_str.data(), value_str.length())) {
        status_ = U_ILLEGAL_ARGUMENT_ERROR;
        return *this;
    }
    if (extensions_ == nullptr) {
        extensions_ = new Locale();
        if (extensions_ == nullptr) {
            status_ = U_MEMORY_ALLOCATION_ERROR;
            return *this;
        }
        extensions_->setKeywordValue(kAttributeKey, value_str.data(), status_);
        return *this;
    }

    CharString attributes;
    CharStringByteSink sink(&attributes);
    UErrorCode localErrorCode = U_ZERO_ERROR;
    extensions_->getKeywordValue(kAttributeKey, sink, localErrorCode);
    if (U_FAILURE(localErrorCode)) {
        CharString new_attributes(value_str.data(), status_);
        // No attributes, set the attribute.
        extensions_->setKeywordValue(kAttributeKey, new_attributes.data(), status_);
        return *this;
    }

    transform(attributes.data(),attributes.length());
    const char* start = attributes.data();
    const char* limit = attributes.data() + attributes.length();
    CharString new_attributes;
    bool inserted = false;
    while (start < limit) {
        if (!inserted) {
            int cmp = uprv_strcmp(start, value_str.data());
            if (cmp == 0) { return *this; }  // Found it in attributes: Just return
            if (cmp > 0) {
                if (!new_attributes.isEmpty()) new_attributes.append('_', status_);
                new_attributes.append(value_str.data(), status_);
                inserted = true;
            }
        }
        if (!new_attributes.isEmpty()) {
            new_attributes.append('_', status_);
        }
        new_attributes.append(start, status_);
        start += uprv_strlen(start) + 1;
    }
    if (!inserted) {
        if (!new_attributes.isEmpty()) {
            new_attributes.append('_', status_);
        }
        new_attributes.append(value_str.data(), status_);
    }
    // Not yet in the attributes, set the attribute.
    extensions_->setKeywordValue(kAttributeKey, new_attributes.data(), status_);
    return *this;
}

LocaleBuilder& LocaleBuilder::removeUnicodeLocaleAttribute(
    StringPiece value)
{
    CharString value_str(value, status_);
    if (U_FAILURE(status_)) { return *this; }
    transform(value_str.data(), value_str.length());
    if (!ultag_isUnicodeLocaleAttribute(value_str.data(), value_str.length())) {
        status_ = U_ILLEGAL_ARGUMENT_ERROR;
        return *this;
    }
    if (extensions_ == nullptr) { return *this; }
    UErrorCode localErrorCode = U_ZERO_ERROR;
    CharString attributes;
    CharStringByteSink sink(&attributes);
    extensions_->getKeywordValue(kAttributeKey, sink, localErrorCode);
    // get failure, just return
    if (U_FAILURE(localErrorCode)) { return *this; }
    // Do not have any attributes, just return.
    if (attributes.isEmpty()) { return *this; }

    char* p = attributes.data();
    // Replace null terminiator in place for _ and - so later
    // we can use uprv_strcmp to compare.
    for (int32_t i = 0; i < attributes.length(); i++, p++) {
        *p = (*p == '_' || *p == '-') ? '\0' : uprv_tolower(*p);
    }

    const char* start = attributes.data();
    const char* limit = attributes.data() + attributes.length();
    CharString new_attributes;
    bool found = false;
    while (start < limit) {
        if (uprv_strcmp(start, value_str.data()) == 0) {
            found = true;
        } else {
            if (!new_attributes.isEmpty()) {
                new_attributes.append('_', status_);
            }
            new_attributes.append(start, status_);
        }
        start += uprv_strlen(start) + 1;
    }
    // Found the value in attributes, set the attribute.
    if (found) {
        extensions_->setKeywordValue(kAttributeKey, new_attributes.data(), status_);
    }
    return *this;
}

LocaleBuilder& LocaleBuilder::clear()
{
    status_ = U_ZERO_ERROR;
    language_[0] = 0;
    script_[0] = 0;
    region_[0] = 0;
    delete variant_;
    variant_ = nullptr;
    clearExtensions();
    return *this;
}

LocaleBuilder& LocaleBuilder::clearExtensions()
{
    delete extensions_;
    extensions_ = nullptr;
    return *this;
}

Locale makeBogusLocale() {
  Locale bogus;
  bogus.setToBogus();
  return bogus;
}

void LocaleBuilder::copyExtensionsFrom(const Locale& src, UErrorCode& errorCode)
{
    if (U_FAILURE(errorCode)) { return; }
    LocalPointer<icu::StringEnumeration> keywords(src.createKeywords(errorCode));
    if (U_FAILURE(errorCode) || keywords.isNull() || keywords->count(errorCode) == 0) {
        // Error, or no extensions to copy.
        return;
    }
    if (extensions_ == nullptr) {
        extensions_ = new Locale();
        if (extensions_ == nullptr) {
            status_ = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
    }
    _copyExtensions(src, keywords.getAlias(), *extensions_, false, errorCode);
}

Locale LocaleBuilder::build(UErrorCode& errorCode)
{
    if (U_FAILURE(errorCode)) {
        return makeBogusLocale();
    }
    if (U_FAILURE(status_)) {
        errorCode = status_;
        return makeBogusLocale();
    }
    CharString locale_str(language_, errorCode);
    if (uprv_strlen(script_) > 0) {
        locale_str.append('-', errorCode).append(StringPiece(script_), errorCode);
    }
    if (uprv_strlen(region_) > 0) {
        locale_str.append('-', errorCode).append(StringPiece(region_), errorCode);
    }
    if (variant_ != nullptr) {
        locale_str.append('-', errorCode).append(StringPiece(variant_->data()), errorCode);
    }
    if (U_FAILURE(errorCode)) {
        return makeBogusLocale();
    }
    Locale product(locale_str.data());
    if (extensions_ != nullptr) {
        _copyExtensions(*extensions_, nullptr, product, true, errorCode);
    }
    if (U_FAILURE(errorCode)) {
        return makeBogusLocale();
    }
    return product;
}

UBool LocaleBuilder::copyErrorTo(UErrorCode &outErrorCode) const {
    if (U_FAILURE(outErrorCode)) {
        // Do not overwrite the older error code
        return TRUE;
    }
    outErrorCode = status_;
    return U_FAILURE(outErrorCode);
}

U_NAMESPACE_END
// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License

// localematcher.cpp
// created: 2019may08 Markus W. Scherer









// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License

// localeprioritylist.h
// created: 2019jul11 Markus W. Scherer







struct UHashtable;

U_NAMESPACE_BEGIN

struct LocaleAndWeightArray;

/**
 * Parses a list of locales from an accept-language string.
 * We are a bit more lenient than the spec:
 * We accept extra whitespace in more places, empty range fields,
 * and any number of qvalue fraction digits.
 *
 * https://tools.ietf.org/html/rfc2616#section-14.4
 * 14.4 Accept-Language
 *
 *        Accept-Language = "Accept-Language" ":"
 *                          1#( language-range [ ";" "q" "=" qvalue ] )
 *        language-range  = ( ( 1*8ALPHA *( "-" 1*8ALPHA ) ) | "*" )
 *
 *    Each language-range MAY be given an associated quality value which
 *    represents an estimate of the user's preference for the languages
 *    specified by that range. The quality value defaults to "q=1". For
 *    example,
 *
 *        Accept-Language: da, en-gb;q=0.8, en;q=0.7
 *
 * https://tools.ietf.org/html/rfc2616#section-3.9
 * 3.9 Quality Values
 *
 *    HTTP content negotiation (section 12) uses short "floating point"
 *    numbers to indicate the relative importance ("weight") of various
 *    negotiable parameters.  A weight is normalized to a real number in
 *    the range 0 through 1, where 0 is the minimum and 1 the maximum
 *    value. If a parameter has a quality value of 0, then content with
 *    this parameter is `not acceptable' for the client. HTTP/1.1
 *    applications MUST NOT generate more than three digits after the
 *    decimal point. User configuration of these values SHOULD also be
 *    limited in this fashion.
 *
 *        qvalue         = ( "0" [ "." 0*3DIGIT ] )
 *                       | ( "1" [ "." 0*3("0") ] )
 */
class U_COMMON_API LocalePriorityList : public UMemory {
public:
    class Iterator : public Locale::Iterator {
    public:
        UBool hasNext() const override { return count < length; }

        const Locale &next() override {
            for(;;) {
                const Locale *locale = list.localeAt(index++);
                if (locale != nullptr) {
                    ++count;
                    return *locale;
                }
            }
        }

    private:
        friend class LocalePriorityList;

        Iterator(const LocalePriorityList &list) : list(list), length(list.getLength()) {}

        const LocalePriorityList &list;
        int32_t index = 0;
        int32_t count = 0;
        const int32_t length;
    };

    LocalePriorityList(StringPiece s, UErrorCode &errorCode);

    ~LocalePriorityList();

    int32_t getLength() const { return listLength - numRemoved; }

    int32_t getLengthIncludingRemoved() const { return listLength; }

    Iterator iterator() const { return Iterator(*this); }

    const Locale *localeAt(int32_t i) const;

    Locale *orphanLocaleAt(int32_t i);

private:
    LocalePriorityList(const LocalePriorityList &) = delete;
    LocalePriorityList &operator=(const LocalePriorityList &) = delete;

    bool add(const Locale &locale, int32_t weight, UErrorCode &errorCode);

    void sort(UErrorCode &errorCode);

    LocaleAndWeightArray *list = nullptr;
    int32_t listLength = 0;
    int32_t numRemoved = 0;
    bool hasWeights = false;  // other than 1.0
    UHashtable *map = nullptr;
};

U_NAMESPACE_END


// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License

// loclikelysubtags.h
// created: 2019may08 Markus W. Scherer


#include <utility>





// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License

// lsr.h
// created: 2019may08 Markus W. Scherer






U_NAMESPACE_BEGIN

struct LSR final : public UMemory {
    static constexpr int32_t REGION_INDEX_LIMIT = 1001 + 26 * 26;

    const char *language;
    const char *script;
    const char *region;
    char *owned = nullptr;
    /** Index for region, 0 if ill-formed. @see indexForRegion */
    int32_t regionIndex = 0;
    /** Only set for LSRs that will be used in a hash table. */
    int32_t hashCode = 0;

    LSR() : language("und"), script(""), region("") {}

    /** Constructor which aliases all subtag pointers. */
    LSR(const char *lang, const char *scr, const char *r) :
            language(lang),  script(scr), region(r),
            regionIndex(indexForRegion(region)) {}
    /**
     * Constructor which prepends the prefix to the language and script,
     * copies those into owned memory, and aliases the region.
     */
    LSR(char prefix, const char *lang, const char *scr, const char *r, UErrorCode &errorCode);
    LSR(LSR &&other) U_NOEXCEPT;
    LSR(const LSR &other) = delete;
    inline ~LSR() {
        // Pure inline code for almost all instances.
        if (owned != nullptr) {
            deleteOwned();
        }
    }

    LSR &operator=(LSR &&other) U_NOEXCEPT;
    LSR &operator=(const LSR &other) = delete;

    /**
     * Returns a positive index (>0) for a well-formed region code.
     * Do not rely on a particular region->index mapping; it may change.
     * Returns 0 for ill-formed strings.
     */
    static int32_t indexForRegion(const char *region);

    UBool operator==(const LSR &other) const;

    inline UBool operator!=(const LSR &other) const {
        return !operator==(other);
    }

    LSR &setHashCode();

private:
    void deleteOwned();
};

U_NAMESPACE_END




U_NAMESPACE_BEGIN

struct XLikelySubtagsData;

/**
 * Map of const char * keys & values.
 * Stores pointers as is: Does not own/copy/adopt/release strings.
 */
class CharStringMap final : public UMemory {
public:
    /** Constructs an unusable non-map. */
    CharStringMap() : map(nullptr) {}
    CharStringMap(int32_t size, UErrorCode &errorCode) {
        map = uhash_openSize(uhash_hashChars, uhash_compareChars, uhash_compareChars,
                             size, &errorCode);
    }
    CharStringMap(CharStringMap &&other) U_NOEXCEPT : map(other.map) {
        other.map = nullptr;
    }
    CharStringMap(const CharStringMap &other) = delete;
    ~CharStringMap() {
        uhash_close(map);
    }

    CharStringMap &operator=(CharStringMap &&other) U_NOEXCEPT {
        map = other.map;
        other.map = nullptr;
        return *this;
    }
    CharStringMap &operator=(const CharStringMap &other) = delete;

    const char *get(const char *key) const { return static_cast<const char *>(uhash_get(map, key)); }
    void put(const char *key, const char *value, UErrorCode &errorCode) {
        uhash_put(map, const_cast<char *>(key), const_cast<char *>(value), &errorCode);
    }

private:
    UHashtable *map;
};

struct LocaleDistanceData {
    LocaleDistanceData() = default;
    LocaleDistanceData(LocaleDistanceData &&data);
    ~LocaleDistanceData();

    const uint8_t *distanceTrieBytes = nullptr;
    const uint8_t *regionToPartitions = nullptr;
    const char **partitions = nullptr;
    const LSR *paradigms = nullptr;
    int32_t paradigmsLength = 0;
    const int32_t *distances = nullptr;

private:
    LocaleDistanceData &operator=(const LocaleDistanceData &) = delete;
};

// TODO(ICU-20777): Rename to just LikelySubtags.
class XLikelySubtags final : public UMemory {
public:
    ~XLikelySubtags();

    static constexpr int32_t SKIP_SCRIPT = 1;

    // VisibleForTesting
    static const XLikelySubtags *getSingleton(UErrorCode &errorCode);

    // VisibleForTesting
    LSR makeMaximizedLsrFrom(const Locale &locale, UErrorCode &errorCode) const;

    // TODO(ICU-20777): Switch Locale/uloc_ likely-subtags API from the old code
    // in loclikely.cpp to this new code, including activating this
    // minimizeSubtags() function. The LocaleMatcher does not minimize.
#if 0
    LSR minimizeSubtags(const char *languageIn, const char *scriptIn, const char *regionIn,
                        ULocale.Minimize fieldToFavor, UErrorCode &errorCode) const;
#endif

    // visible for LocaleDistance
    const LocaleDistanceData &getDistanceData() const { return distanceData; }

private:
    XLikelySubtags(XLikelySubtagsData &data);
    XLikelySubtags(const XLikelySubtags &other) = delete;
    XLikelySubtags &operator=(const XLikelySubtags &other) = delete;

    static void initLikelySubtags(UErrorCode &errorCode);

    LSR makeMaximizedLsr(const char *language, const char *script, const char *region,
                         const char *variant, UErrorCode &errorCode) const;

    /**
     * Raw access to addLikelySubtags. Input must be in canonical format, eg "en", not "eng" or "EN".
     */
    LSR maximize(const char *language, const char *script, const char *region) const;

    static int32_t trieNext(BytesTrie &iter, const char *s, int32_t i);

    UResourceBundle *langInfoBundle;
    // We could store the strings by value, except that if there were few enough strings,
    // moving the contents could copy it to a different array,
    // invalidating the pointers stored in the maps.
    CharString *strings;
    CharStringMap languageAliases;
    CharStringMap regionAliases;

    // The trie maps each lang+script+region (encoded in ASCII) to an index into lsrs.
    // There is also a trie value for each intermediate lang and lang+script.
    // '*' is used instead of "und", "Zzzz"/"" and "ZZ"/"".
    BytesTrie trie;
    uint64_t trieUndState;
    uint64_t trieUndZzzzState;
    int32_t defaultLsrIndex;
    uint64_t trieFirstLetterStates[26];
    const LSR *lsrs;
#if U_DEBUG
    int32_t lsrsLength;
#endif

    // distance/matcher data: see comment in XLikelySubtagsData::load()
    LocaleDistanceData distanceData;
};

U_NAMESPACE_END


// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License

// locdistance.h
// created: 2019may08 Markus W. Scherer









U_NAMESPACE_BEGIN

struct LocaleDistanceData;

/**
 * Offline-built data for LocaleMatcher.
 * Mostly but not only the data for mapping locales to their maximized forms.
 */
class LocaleDistance final : public UMemory {
public:
    static const LocaleDistance *getSingleton(UErrorCode &errorCode);

    /**
     * Finds the supported LSR with the smallest distance from the desired one.
     * Equivalent LSR subtags must be normalized into a canonical form.
     *
     * <p>Returns the index of the lowest-distance supported LSR in bits 31..8
     * (negative if none has a distance below the threshold),
     * and its distance (0..ABOVE_THRESHOLD) in bits 7..0.
     */
    int32_t getBestIndexAndDistance(const LSR &desired,
                                    const LSR **supportedLSRs, int32_t supportedLSRsLength,
                                    int32_t threshold, ULocMatchFavorSubtag favorSubtag) const;

    int32_t getParadigmLSRsLength() const { return paradigmLSRsLength; }

    UBool isParadigmLSR(const LSR &lsr) const;

    int32_t getDefaultScriptDistance() const {
        return defaultScriptDistance;
    }

    int32_t getDefaultDemotionPerDesiredLocale() const {
        return defaultDemotionPerDesiredLocale;
    }

private:
    LocaleDistance(const LocaleDistanceData &data);
    LocaleDistance(const LocaleDistance &other) = delete;
    LocaleDistance &operator=(const LocaleDistance &other) = delete;

    static void initLocaleDistance(UErrorCode &errorCode);

    static int32_t getDesSuppScriptDistance(BytesTrie &iter, uint64_t startState,
                                            const char *desired, const char *supported);

    static int32_t getRegionPartitionsDistance(
        BytesTrie &iter, uint64_t startState,
        const char *desiredPartitions, const char *supportedPartitions,
        int32_t threshold);

    static int32_t getFallbackRegionDistance(BytesTrie &iter, uint64_t startState);

    static int32_t trieNext(BytesTrie &iter, const char *s, bool wantValue);

    const char *partitionsForRegion(const LSR &lsr) const {
        // ill-formed region -> one non-matching string
        int32_t pIndex = regionToPartitionsIndex[lsr.regionIndex];
        return partitionArrays[pIndex];
    }

    int32_t getDefaultRegionDistance() const {
        return defaultRegionDistance;
    }

    // The trie maps each dlang+slang+dscript+sscript+dregion+sregion
    // (encoded in ASCII with bit 7 set on the last character of each subtag) to a distance.
    // There is also a trie value for each subsequence of whole subtags.
    // One '*' is used for a (desired, supported) pair of "und", "Zzzz"/"", or "ZZ"/"".
    BytesTrie trie;

    /**
     * Maps each region to zero or more single-character partitions.
     */
    const uint8_t *regionToPartitionsIndex;
    const char **partitionArrays;

    /**
     * Used to get the paradigm region for a cluster, if there is one.
     */
    const LSR *paradigmLSRs;
    int32_t paradigmLSRsLength;

    int32_t defaultLanguageDistance;
    int32_t defaultScriptDistance;
    int32_t defaultRegionDistance;
    int32_t minRegionDistance;
    int32_t defaultDemotionPerDesiredLocale;
};

U_NAMESPACE_END





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1999-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*   Date        Name        Description
*   10/22/99    alan        Creation.  This is an internal header.
*                           It should not be exported.
**********************************************************************
*/








U_NAMESPACE_BEGIN

/**
 * <p>Ultralightweight C++ implementation of a <tt>void*</tt> vector
 * that is (mostly) compatible with java.util.Vector.
 *
 * <p>This is a very simple implementation, written to satisfy an
 * immediate porting need.  As such, it is not completely fleshed out,
 * and it aims for simplicity and conformity.  Nonetheless, it serves
 * its purpose (porting code from java that uses java.util.Vector)
 * well, and it could be easily made into a more robust vector class.
 *
 * <p><b>Design notes</b>
 *
 * <p>There is index bounds checking, but little is done about it.  If
 * indices are out of bounds, either nothing happens, or zero is
 * returned.  We <em>do</em> avoid indexing off into the weeds.
 *
 * <p>There is detection of out of memory, but the handling is very
 * coarse-grained -- similar to UnicodeString's protocol, but even
 * coarser.  The class contains <em>one static flag</em> that is set
 * when any call to <tt>new</tt> returns zero.  This allows the caller
 * to use several vectors and make just one check at the end to see if
 * a memory failure occurred.  This is more efficient than making a
 * check after each call on each vector when doing many operations on
 * multiple vectors.  The single static flag works best when memory
 * failures are infrequent, and when recovery options are limited or
 * nonexistent.
 *
 * <p>Since we don't have garbage collection, UVector was given the
 * option to <em>own</em>its contents.  To employ this, set a deleter
 * function.  The deleter is called on a void* pointer when that
 * pointer is released by the vector, either when the vector itself is
 * destructed, or when a call to setElementAt() overwrites an element,
 * or when a call to remove() or one of its variants explicitly
 * removes an element.  If no deleter is set, or the deleter is set to
 * zero, then it is assumed that the caller will delete elements as
 * needed.
 *
 * <p>In order to implement methods such as contains() and indexOf(),
 * UVector needs a way to compare objects for equality.  To do so, it
 * uses a comparison function, or "comparer."  If the comparer is not
 * set, or is set to zero, then all such methods will act as if the
 * vector contains no element.  That is, indexOf() will always return
 * -1, contains() will always return FALSE, etc.
 *
 * <p><b>To do</b>
 *
 * <p>Improve the handling of index out of bounds errors.
 *
 * @author Alan Liu
 */
class U_COMMON_API UVector : public UObject {
    // NOTE: UVector uses the UHashKey (union of void* and int32_t) as
    // its basic storage type.  It uses UElementsAreEqual as its
    // comparison function.  It uses UObjectDeleter as its deleter
    // function.  These are named for hashtables, but used here as-is
    // rather than duplicating the type.  This allows sharing of
    // support functions.

private:
    int32_t count;

    int32_t capacity;

    UElement* elements;

    UObjectDeleter *deleter;

    UElementsAreEqual *comparer;

public:
    UVector(UErrorCode &status);

    UVector(int32_t initialCapacity, UErrorCode &status);

    UVector(UObjectDeleter *d, UElementsAreEqual *c, UErrorCode &status);

    UVector(UObjectDeleter *d, UElementsAreEqual *c, int32_t initialCapacity, UErrorCode &status);

    virtual ~UVector();

    /**
     * Assign this object to another (make this a copy of 'other').
     * Use the 'assign' function to assign each element.
     */
    void assign(const UVector& other, UElementAssigner *assign, UErrorCode &ec);

    /**
     * Compare this vector with another.  They will be considered
     * equal if they are of the same size and all elements are equal,
     * as compared using this object's comparer.
     */
    UBool operator==(const UVector& other);

    /**
     * Equivalent to !operator==()
     */
    inline UBool operator!=(const UVector& other);

    //------------------------------------------------------------
    // java.util.Vector API
    //------------------------------------------------------------

    void addElement(void* obj, UErrorCode &status);

    void addElement(int32_t elem, UErrorCode &status);

    void setElementAt(void* obj, int32_t index);

    void setElementAt(int32_t elem, int32_t index);

    void insertElementAt(void* obj, int32_t index, UErrorCode &status);

    void insertElementAt(int32_t elem, int32_t index, UErrorCode &status);
    
    void* elementAt(int32_t index) const;

    int32_t elementAti(int32_t index) const;

    UBool equals(const UVector &other) const;

    inline void* firstElement(void) const;

    inline void* lastElement(void) const;

    inline int32_t lastElementi(void) const;

    int32_t indexOf(void* obj, int32_t startIndex = 0) const;

    int32_t indexOf(int32_t obj, int32_t startIndex = 0) const;

    inline UBool contains(void* obj) const;

    inline UBool contains(int32_t obj) const;

    UBool containsAll(const UVector& other) const;

    UBool removeAll(const UVector& other);

    UBool retainAll(const UVector& other);

    void removeElementAt(int32_t index);

    UBool removeElement(void* obj);

    void removeAllElements();

    inline int32_t size(void) const;

    inline UBool isEmpty(void) const;

    UBool ensureCapacity(int32_t minimumCapacity, UErrorCode &status);

    /**
     * Change the size of this vector as follows: If newSize is
     * smaller, then truncate the array, possibly deleting held
     * elements for i >= newSize.  If newSize is larger, grow the
     * array, filling in new slots with NULL.
     */
    void setSize(int32_t newSize, UErrorCode &status);

    /**
     * Fill in the given array with all elements of this vector.
     */
    void** toArray(void** result) const;

    //------------------------------------------------------------
    // New API
    //------------------------------------------------------------

    UObjectDeleter *setDeleter(UObjectDeleter *d);

    UElementsAreEqual *setComparer(UElementsAreEqual *c);

    inline void* operator[](int32_t index) const;

    /**
     * Removes the element at the given index from this vector and
     * transfer ownership of it to the caller.  After this call, the
     * caller owns the result and must delete it and the vector entry
     * at 'index' is removed, shifting all subsequent entries back by
     * one index and shortening the size of the vector by one.  If the
     * index is out of range or if there is no item at the given index
     * then 0 is returned and the vector is unchanged.
     */
    void* orphanElementAt(int32_t index);

    /**
     * Returns true if this vector contains none of the elements
     * of the given vector.
     * @param other vector to be checked for containment
     * @return true if the test condition is met
     */
    UBool containsNone(const UVector& other) const;

    /**
     * Insert the given object into this vector at its sorted position
     * as defined by 'compare'.  The current elements are assumed to
     * be sorted already.
     */
    void sortedInsert(void* obj, UElementComparator *compare, UErrorCode& ec);

    /**
     * Insert the given integer into this vector at its sorted position
     * as defined by 'compare'.  The current elements are assumed to
     * be sorted already.
     */
    void sortedInsert(int32_t obj, UElementComparator *compare, UErrorCode& ec);

    /**
     * Sort the contents of the vector, assuming that the contents of the
     * vector are of type int32_t.
     */
    void sorti(UErrorCode &ec);

    /**
      * Sort the contents of this vector, using a caller-supplied function
      * to do the comparisons.  (It's confusing that
      *  UVector's UElementComparator function is different from the
      *  UComparator function type defined in uarrsort.h)
      */
    void sort(UElementComparator *compare, UErrorCode &ec);

    /**
     * Stable sort the contents of this vector using a caller-supplied function
     * of type UComparator to do the comparison.  Provides more flexibility
     * than UVector::sort() because an additional user parameter can be passed to
     * the comparison function.
     */
    void sortWithUComparator(UComparator *compare, const void *context, UErrorCode &ec);

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     */
    virtual UClassID getDynamicClassID() const;

private:
    void _init(int32_t initialCapacity, UErrorCode &status);

    int32_t indexOf(UElement key, int32_t startIndex = 0, int8_t hint = 0) const;

    void sortedInsert(UElement e, UElementComparator *compare, UErrorCode& ec);

    // Disallow
    UVector(const UVector&);

    // Disallow
    UVector& operator=(const UVector&);

};


/**
 * <p>Ultralightweight C++ implementation of a <tt>void*</tt> stack
 * that is (mostly) compatible with java.util.Stack.  As in java, this
 * is merely a paper thin layer around UVector.  See the UVector
 * documentation for further information.
 *
 * <p><b>Design notes</b>
 *
 * <p>The element at index <tt>n-1</tt> is (of course) the top of the
 * stack.
 *
 * <p>The poorly named <tt>empty()</tt> method doesn't empty the
 * stack; it determines if the stack is empty.
 *
 * @author Alan Liu
 */
class U_COMMON_API UStack : public UVector {
public:
    UStack(UErrorCode &status);

    UStack(int32_t initialCapacity, UErrorCode &status);

    UStack(UObjectDeleter *d, UElementsAreEqual *c, UErrorCode &status);

    UStack(UObjectDeleter *d, UElementsAreEqual *c, int32_t initialCapacity, UErrorCode &status);

    virtual ~UStack();

    // It's okay not to have a virtual destructor (in UVector)
    // because UStack has no special cleanup to do.

    inline UBool empty(void) const;

    inline void* peek(void) const;

    inline int32_t peeki(void) const;
    
    void* pop(void);
    
    int32_t popi(void);
    
    inline void* push(void* obj, UErrorCode &status);

    inline int32_t push(int32_t i, UErrorCode &status);

    /*
    If the object o occurs as an item in this stack,
    this method returns the 1-based distance from the top of the stack.
    */
    int32_t search(void* obj) const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     */
    virtual UClassID getDynamicClassID() const;

private:
    // Disallow
    UStack(const UStack&);

    // Disallow
    UStack& operator=(const UStack&);
};


// UVector inlines

inline int32_t UVector::size(void) const {
    return count;
}

inline UBool UVector::isEmpty(void) const {
    return count == 0;
}

inline UBool UVector::contains(void* obj) const {
    return indexOf(obj) >= 0;
}

inline UBool UVector::contains(int32_t obj) const {
    return indexOf(obj) >= 0;
}

inline void* UVector::firstElement(void) const {
    return elementAt(0);
}

inline void* UVector::lastElement(void) const {
    return elementAt(count-1);
}

inline int32_t UVector::lastElementi(void) const {
    return elementAti(count-1);
}

inline void* UVector::operator[](int32_t index) const {
    return elementAt(index);
}

inline UBool UVector::operator!=(const UVector& other) {
    return !operator==(other);
}

// UStack inlines

inline UBool UStack::empty(void) const {
    return isEmpty();
}

inline void* UStack::peek(void) const {
    return lastElement();
}

inline int32_t UStack::peeki(void) const {
    return lastElementi();
}

inline void* UStack::push(void* obj, UErrorCode &status) {
    addElement(obj, status);
    return obj;
}

inline int32_t UStack::push(int32_t i, UErrorCode &status) {
    addElement(i, status);
    return i;
}

U_NAMESPACE_END



#define UND_LSR LSR("und", "", "")

/**
 * Indicator for the lifetime of desired-locale objects passed into the LocaleMatcher.
 *
 * @draft ICU 65
 */
enum ULocMatchLifetime {
    /**
     * Locale objects are temporary.
     * The matcher will make a copy of a locale that will be used beyond one function call.
     *
     * @draft ICU 65
     */
    ULOCMATCH_TEMPORARY_LOCALES,
    /**
     * Locale objects are stored at least as long as the matcher is used.
     * The matcher will keep only a pointer to a locale that will be used beyond one function call,
     * avoiding a copy.
     *
     * @draft ICU 65
     */
    ULOCMATCH_STORED_LOCALES  // TODO: permanent? cached? clone?
};
#ifndef U_IN_DOXYGEN
typedef enum ULocMatchLifetime ULocMatchLifetime;
#endif

U_NAMESPACE_BEGIN

LocaleMatcher::Result::Result(LocaleMatcher::Result &&src) U_NOEXCEPT :
        desiredLocale(src.desiredLocale),
        supportedLocale(src.supportedLocale),
        desiredIndex(src.desiredIndex),
        supportedIndex(src.supportedIndex),
        desiredIsOwned(src.desiredIsOwned) {
    if (desiredIsOwned) {
        src.desiredLocale = nullptr;
        src.desiredIndex = -1;
        src.desiredIsOwned = FALSE;
    }
}

LocaleMatcher::Result::~Result() {
    if (desiredIsOwned) {
        delete desiredLocale;
    }
}

LocaleMatcher::Result &LocaleMatcher::Result::operator=(LocaleMatcher::Result &&src) U_NOEXCEPT {
    this->~Result();

    desiredLocale = src.desiredLocale;
    supportedLocale = src.supportedLocale;
    desiredIndex = src.desiredIndex;
    supportedIndex = src.supportedIndex;
    desiredIsOwned = src.desiredIsOwned;

    if (desiredIsOwned) {
        src.desiredLocale = nullptr;
        src.desiredIndex = -1;
        src.desiredIsOwned = FALSE;
    }
    return *this;
}

Locale LocaleMatcher::Result::makeResolvedLocale(UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode) || supportedLocale == nullptr) {
        return Locale::getRoot();
    }
    const Locale *bestDesired = getDesiredLocale();
    if (bestDesired == nullptr || *supportedLocale == *bestDesired) {
        return *supportedLocale;
    }
    LocaleBuilder b;
    b.setLocale(*supportedLocale);

    // Copy the region from bestDesired, if there is one.
    const char *region = bestDesired->getCountry();
    if (*region != 0) {
        b.setRegion(region);
    }

    // Copy the variants from bestDesired, if there are any.
    // Note that this will override any supportedLocale variants.
    // For example, "sco-ulster-fonipa" + "...-fonupa" => "sco-fonupa" (replacing ulster).
    const char *variants = bestDesired->getVariant();
    if (*variants != 0) {
        b.setVariant(variants);
    }

    // Copy the extensions from bestDesired, if there are any.
    // C++ note: The following note, copied from Java, may not be true,
    // as long as C++ copies by legacy ICU keyword, not by extension singleton.
    // Note that this will override any supportedLocale extensions.
    // For example, "th-u-nu-latn-ca-buddhist" + "...-u-nu-native" => "th-u-nu-native"
    // (replacing calendar).
    b.copyExtensionsFrom(*bestDesired, errorCode);
    return b.build(errorCode);
}

LocaleMatcher::Builder::Builder(LocaleMatcher::Builder &&src) U_NOEXCEPT :
        errorCode_(src.errorCode_),
        supportedLocales_(src.supportedLocales_),
        thresholdDistance_(src.thresholdDistance_),
        demotion_(src.demotion_),
        defaultLocale_(src.defaultLocale_),
        favor_(src.favor_) {
    src.supportedLocales_ = nullptr;
    src.defaultLocale_ = nullptr;
}

LocaleMatcher::Builder::~Builder() {
    delete supportedLocales_;
    delete defaultLocale_;
}

LocaleMatcher::Builder &LocaleMatcher::Builder::operator=(LocaleMatcher::Builder &&src) U_NOEXCEPT {
    this->~Builder();

    errorCode_ = src.errorCode_;
    supportedLocales_ = src.supportedLocales_;
    thresholdDistance_ = src.thresholdDistance_;
    demotion_ = src.demotion_;
    defaultLocale_ = src.defaultLocale_;
    favor_ = src.favor_;

    src.supportedLocales_ = nullptr;
    src.defaultLocale_ = nullptr;
    return *this;
}

void LocaleMatcher::Builder::clearSupportedLocales() {
    if (supportedLocales_ != nullptr) {
        supportedLocales_->removeAllElements();
    }
}

bool LocaleMatcher::Builder::ensureSupportedLocaleVector() {
    if (U_FAILURE(errorCode_)) { return false; }
    if (supportedLocales_ != nullptr) { return true; }
    supportedLocales_ = new UVector(uprv_deleteUObject, nullptr, errorCode_);
    if (U_FAILURE(errorCode_)) { return false; }
    if (supportedLocales_ == nullptr) {
        errorCode_ = U_MEMORY_ALLOCATION_ERROR;
        return false;
    }
    return true;
}

LocaleMatcher::Builder &LocaleMatcher::Builder::setSupportedLocalesFromListString(
        StringPiece locales) {
    LocalePriorityList list(locales, errorCode_);
    if (U_FAILURE(errorCode_)) { return *this; }
    clearSupportedLocales();
    if (!ensureSupportedLocaleVector()) { return *this; }
    int32_t length = list.getLengthIncludingRemoved();
    for (int32_t i = 0; i < length; ++i) {
        Locale *locale = list.orphanLocaleAt(i);
        if (locale == nullptr) { continue; }
        supportedLocales_->addElement(locale, errorCode_);
        if (U_FAILURE(errorCode_)) {
            delete locale;
            break;
        }
    }
    return *this;
}

LocaleMatcher::Builder &LocaleMatcher::Builder::setSupportedLocales(Locale::Iterator &locales) {
    if (U_FAILURE(errorCode_)) { return *this; }
    clearSupportedLocales();
    if (!ensureSupportedLocaleVector()) { return *this; }
    while (locales.hasNext()) {
        const Locale &locale = locales.next();
        Locale *clone = locale.clone();
        if (clone == nullptr) {
            errorCode_ = U_MEMORY_ALLOCATION_ERROR;
            break;
        }
        supportedLocales_->addElement(clone, errorCode_);
        if (U_FAILURE(errorCode_)) {
            delete clone;
            break;
        }
    }
    return *this;
}

LocaleMatcher::Builder &LocaleMatcher::Builder::addSupportedLocale(const Locale &locale) {
    if (!ensureSupportedLocaleVector()) { return *this; }
    Locale *clone = locale.clone();
    if (clone == nullptr) {
        errorCode_ = U_MEMORY_ALLOCATION_ERROR;
        return *this;
    }
    supportedLocales_->addElement(clone, errorCode_);
    if (U_FAILURE(errorCode_)) {
        delete clone;
    }
    return *this;
}

LocaleMatcher::Builder &LocaleMatcher::Builder::setDefaultLocale(const Locale *defaultLocale) {
    if (U_FAILURE(errorCode_)) { return *this; }
    Locale *clone = nullptr;
    if (defaultLocale != nullptr) {
        clone = defaultLocale->clone();
        if (clone == nullptr) {
            errorCode_ = U_MEMORY_ALLOCATION_ERROR;
            return *this;
        }
    }
    delete defaultLocale_;
    defaultLocale_ = clone;
    return *this;
}

LocaleMatcher::Builder &LocaleMatcher::Builder::setFavorSubtag(ULocMatchFavorSubtag subtag) {
    if (U_FAILURE(errorCode_)) { return *this; }
    favor_ = subtag;
    return *this;
}

LocaleMatcher::Builder &LocaleMatcher::Builder::setDemotionPerDesiredLocale(ULocMatchDemotion demotion) {
    if (U_FAILURE(errorCode_)) { return *this; }
    demotion_ = demotion;
    return *this;
}

#if 0
/**
 * <i>Internal only!</i>
 *
 * @param thresholdDistance the thresholdDistance to set, with -1 = default
 * @return this Builder object
 * @internal
 * @deprecated This API is ICU internal only.
 */
@Deprecated
LocaleMatcher::Builder &LocaleMatcher::Builder::internalSetThresholdDistance(int32_t thresholdDistance) {
    if (U_FAILURE(errorCode_)) { return *this; }
    if (thresholdDistance > 100) {
        thresholdDistance = 100;
    }
    thresholdDistance_ = thresholdDistance;
    return *this;
}
#endif

UBool LocaleMatcher::Builder::copyErrorTo(UErrorCode &outErrorCode) const {
    if (U_FAILURE(outErrorCode)) { return TRUE; }
    if (U_SUCCESS(errorCode_)) { return FALSE; }
    outErrorCode = errorCode_;
    return TRUE;
}

LocaleMatcher LocaleMatcher::Builder::build(UErrorCode &errorCode) const {
    if (U_SUCCESS(errorCode) && U_FAILURE(errorCode_)) {
        errorCode = errorCode_;
    }
    return LocaleMatcher(*this, errorCode);
}

namespace {

LSR getMaximalLsrOrUnd(const XLikelySubtags &likelySubtags, const Locale &locale,
                       UErrorCode &errorCode) {
    if (U_FAILURE(errorCode) || locale.isBogus() || *locale.getName() == 0 /* "und" */) {
        return UND_LSR;
    } else {
        return likelySubtags.makeMaximizedLsrFrom(locale, errorCode);
    }
}

int32_t hashLSR(const UHashTok token) {
    const LSR *lsr = static_cast<const LSR *>(token.pointer);
    return lsr->hashCode;
}

UBool compareLSRs(const UHashTok t1, const UHashTok t2) {
    const LSR *lsr1 = static_cast<const LSR *>(t1.pointer);
    const LSR *lsr2 = static_cast<const LSR *>(t2.pointer);
    return *lsr1 == *lsr2;
}

bool putIfAbsent(UHashtable *lsrToIndex, const LSR &lsr, int32_t i, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return false; }
    U_ASSERT(i > 0);
    int32_t index = uhash_geti(lsrToIndex, &lsr);
    if (index != 0) {
        return false;
    } else {
        uhash_puti(lsrToIndex, const_cast<LSR *>(&lsr), i, &errorCode);
        return U_SUCCESS(errorCode);
    }
}

}  // namespace

LocaleMatcher::LocaleMatcher(const Builder &builder, UErrorCode &errorCode) :
        likelySubtags(*XLikelySubtags::getSingleton(errorCode)),
        localeDistance(*LocaleDistance::getSingleton(errorCode)),
        thresholdDistance(builder.thresholdDistance_),
        demotionPerDesiredLocale(0),
        favorSubtag(builder.favor_),
        supportedLocales(nullptr), lsrs(nullptr), supportedLocalesLength(0),
        supportedLsrToIndex(nullptr),
        supportedLSRs(nullptr), supportedIndexes(nullptr), supportedLSRsLength(0),
        ownedDefaultLocale(nullptr), defaultLocale(nullptr), defaultLocaleIndex(-1) {
    if (U_FAILURE(errorCode)) { return; }
    if (thresholdDistance < 0) {
        thresholdDistance = localeDistance.getDefaultScriptDistance();
    }
    supportedLocalesLength = builder.supportedLocales_ != nullptr ?
        builder.supportedLocales_->size() : 0;
    const Locale *def = builder.defaultLocale_;
    int32_t idef = -1;
    if (supportedLocalesLength > 0) {
        // Store the supported locales in input order,
        // so that when different types are used (e.g., language tag strings)
        // we can return those by parallel index.
        supportedLocales = static_cast<const Locale **>(
            uprv_malloc(supportedLocalesLength * sizeof(const Locale *)));
        // Supported LRSs in input order.
        // In C++, we store these permanently to simplify ownership management
        // in the hash tables. Duplicate LSRs (if any) are unused overhead.
        lsrs = new LSR[supportedLocalesLength];
        if (supportedLocales == nullptr || lsrs == nullptr) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        // If the constructor fails partway, we need null pointers for destructibility.
        uprv_memset(supportedLocales, 0, supportedLocalesLength * sizeof(const Locale *));
        // Also find the first supported locale whose LSR is
        // the same as that for the default locale.
        LSR builderDefaultLSR;
        const LSR *defLSR = nullptr;
        if (def != nullptr) {
            builderDefaultLSR = getMaximalLsrOrUnd(likelySubtags, *def, errorCode);
            if (U_FAILURE(errorCode)) { return; }
            defLSR = &builderDefaultLSR;
        }
        for (int32_t i = 0; i < supportedLocalesLength; ++i) {
            const Locale &locale = *static_cast<Locale *>(builder.supportedLocales_->elementAt(i));
            supportedLocales[i] = locale.clone();
            if (supportedLocales[i] == nullptr) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            const Locale &supportedLocale = *supportedLocales[i];
            LSR &lsr = lsrs[i] = getMaximalLsrOrUnd(likelySubtags, supportedLocale, errorCode);
            lsr.setHashCode();
            if (U_FAILURE(errorCode)) { return; }
            if (idef < 0 && defLSR != nullptr && lsr == *defLSR) {
                idef = i;
                defLSR = &lsr;  // owned pointer to put into supportedLsrToIndex
                if (*def == supportedLocale) {
                    def = &supportedLocale;  // owned pointer to keep
                }
            }
        }

        // We need an unordered map from LSR to first supported locale with that LSR,
        // and an ordered list of (LSR, supported index).
        // We insert the supported locales in the following order:
        // 1. Default locale, if it is supported.
        // 2. Priority locales (aka "paradigm locales") in builder order.
        // 3. Remaining locales in builder order.
        // In Java, we use a LinkedHashMap for both map & ordered lists.
        // In C++, we use separate structures.
        // We over-allocate arrays of LSRs and indexes for simplicity.
        // We reserve slots at the array starts for the default and paradigm locales,
        // plus enough for all supported locales.
        // If there are few paradigm locales and few duplicate supported LSRs,
        // then the amount of wasted space is small.
        supportedLsrToIndex = uhash_openSize(hashLSR, compareLSRs, uhash_compareLong,
                                             supportedLocalesLength, &errorCode);
        if (U_FAILURE(errorCode)) { return; }
        int32_t paradigmLimit = 1 + localeDistance.getParadigmLSRsLength();
        int32_t suppLSRsCapacity = paradigmLimit + supportedLocalesLength;
        supportedLSRs = static_cast<const LSR **>(
            uprv_malloc(suppLSRsCapacity * sizeof(const LSR *)));
        supportedIndexes = static_cast<int32_t *>(
            uprv_malloc(suppLSRsCapacity * sizeof(int32_t)));
        if (supportedLSRs == nullptr || supportedIndexes == nullptr) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        int32_t paradigmIndex = 0;
        int32_t otherIndex = paradigmLimit;
        if (idef >= 0) {
            uhash_puti(supportedLsrToIndex, const_cast<LSR *>(defLSR), idef + 1, &errorCode);
            supportedLSRs[0] = defLSR;
            supportedIndexes[0] = idef;
            paradigmIndex = 1;
        }
        for (int32_t i = 0; i < supportedLocalesLength; ++i) {
            if (i == idef) { continue; }
            const Locale &locale = *supportedLocales[i];
            const LSR &lsr = lsrs[i];
            if (defLSR == nullptr) {
                U_ASSERT(i == 0);
                def = &locale;
                defLSR = &lsr;
                idef = 0;
                uhash_puti(supportedLsrToIndex, const_cast<LSR *>(&lsr), 0 + 1, &errorCode);
                supportedLSRs[0] = &lsr;
                supportedIndexes[0] = 0;
                paradigmIndex = 1;
            } else if (idef >= 0 && lsr == *defLSR) {
                // lsr == *defLSR means that this supported locale is
                // a duplicate of the default locale.
                // Either an explicit default locale is supported, and we added it before the loop,
                // or there is no explicit default locale, and this is
                // a duplicate of the first supported locale.
                // In both cases, idef >= 0 now, so otherwise we can skip the comparison.
                // For a duplicate, putIfAbsent() is a no-op, so nothing to do.
            } else {
                if (putIfAbsent(supportedLsrToIndex, lsr, i + 1, errorCode)) {
                    if (localeDistance.isParadigmLSR(lsr)) {
                        supportedLSRs[paradigmIndex] = &lsr;
                        supportedIndexes[paradigmIndex++] = i;
                    } else {
                        supportedLSRs[otherIndex] = &lsr;
                        supportedIndexes[otherIndex++] = i;
                    }
                }
            }
            if (U_FAILURE(errorCode)) { return; }
        }
        // Squeeze out unused array slots.
        if (paradigmIndex < paradigmLimit && paradigmLimit < otherIndex) {
            uprv_memmove(supportedLSRs + paradigmIndex, supportedLSRs + paradigmLimit,
                         (otherIndex - paradigmLimit) * sizeof(const LSR *));
            uprv_memmove(supportedIndexes + paradigmIndex, supportedIndexes + paradigmLimit,
                         (otherIndex - paradigmLimit) * sizeof(int32_t));
        }
        supportedLSRsLength = otherIndex - (paradigmLimit - paradigmIndex);
    }

    if (def != nullptr && (idef < 0 || def != supportedLocales[idef])) {
        ownedDefaultLocale = def->clone();
        if (ownedDefaultLocale == nullptr) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        def = ownedDefaultLocale;
    }
    defaultLocale = def;
    defaultLocaleIndex = idef;

    if (builder.demotion_ == ULOCMATCH_DEMOTION_REGION) {
        demotionPerDesiredLocale = localeDistance.getDefaultDemotionPerDesiredLocale();
    }
}

LocaleMatcher::LocaleMatcher(LocaleMatcher &&src) U_NOEXCEPT :
        likelySubtags(src.likelySubtags),
        localeDistance(src.localeDistance),
        thresholdDistance(src.thresholdDistance),
        demotionPerDesiredLocale(src.demotionPerDesiredLocale),
        favorSubtag(src.favorSubtag),
        supportedLocales(src.supportedLocales), lsrs(src.lsrs),
        supportedLocalesLength(src.supportedLocalesLength),
        supportedLsrToIndex(src.supportedLsrToIndex),
        supportedLSRs(src.supportedLSRs),
        supportedIndexes(src.supportedIndexes),
        supportedLSRsLength(src.supportedLSRsLength),
        ownedDefaultLocale(src.ownedDefaultLocale), defaultLocale(src.defaultLocale),
        defaultLocaleIndex(src.defaultLocaleIndex) {
    src.supportedLocales = nullptr;
    src.lsrs = nullptr;
    src.supportedLocalesLength = 0;
    src.supportedLsrToIndex = nullptr;
    src.supportedLSRs = nullptr;
    src.supportedIndexes = nullptr;
    src.supportedLSRsLength = 0;
    src.ownedDefaultLocale = nullptr;
    src.defaultLocale = nullptr;
    src.defaultLocaleIndex = -1;
}

LocaleMatcher::~LocaleMatcher() {
    for (int32_t i = 0; i < supportedLocalesLength; ++i) {
        delete supportedLocales[i];
    }
    uprv_free(supportedLocales);
    delete[] lsrs;
    uhash_close(supportedLsrToIndex);
    uprv_free(supportedLSRs);
    uprv_free(supportedIndexes);
    delete ownedDefaultLocale;
}

LocaleMatcher &LocaleMatcher::operator=(LocaleMatcher &&src) U_NOEXCEPT {
    this->~LocaleMatcher();

    thresholdDistance = src.thresholdDistance;
    demotionPerDesiredLocale = src.demotionPerDesiredLocale;
    favorSubtag = src.favorSubtag;
    supportedLocales = src.supportedLocales;
    lsrs = src.lsrs;
    supportedLocalesLength = src.supportedLocalesLength;
    supportedLsrToIndex = src.supportedLsrToIndex;
    supportedLSRs = src.supportedLSRs;
    supportedIndexes = src.supportedIndexes;
    supportedLSRsLength = src.supportedLSRsLength;
    ownedDefaultLocale = src.ownedDefaultLocale;
    defaultLocale = src.defaultLocale;
    defaultLocaleIndex = src.defaultLocaleIndex;

    src.supportedLocales = nullptr;
    src.lsrs = nullptr;
    src.supportedLocalesLength = 0;
    src.supportedLsrToIndex = nullptr;
    src.supportedLSRs = nullptr;
    src.supportedIndexes = nullptr;
    src.supportedLSRsLength = 0;
    src.ownedDefaultLocale = nullptr;
    src.defaultLocale = nullptr;
    src.defaultLocaleIndex = -1;
    return *this;
}

class LocaleLsrIterator {
public:
    LocaleLsrIterator(const XLikelySubtags &likelySubtags, Locale::Iterator &locales,
                      ULocMatchLifetime lifetime) :
            likelySubtags(likelySubtags), locales(locales), lifetime(lifetime) {}

    ~LocaleLsrIterator() {
        if (lifetime == ULOCMATCH_TEMPORARY_LOCALES) {
            delete remembered;
        }
    }

    bool hasNext() const {
        return locales.hasNext();
    }

    LSR next(UErrorCode &errorCode) {
        current = &locales.next();
        return getMaximalLsrOrUnd(likelySubtags, *current, errorCode);
    }

    void rememberCurrent(int32_t desiredIndex, UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return; }
        bestDesiredIndex = desiredIndex;
        if (lifetime == ULOCMATCH_STORED_LOCALES) {
            remembered = current;
        } else {
            // ULOCMATCH_TEMPORARY_LOCALES
            delete remembered;
            remembered = new Locale(*current);
            if (remembered == nullptr) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
            }
        }
    }

    const Locale *orphanRemembered() {
        const Locale *rem = remembered;
        remembered = nullptr;
        return rem;
    }

    int32_t getBestDesiredIndex() const {
        return bestDesiredIndex;
    }

private:
    const XLikelySubtags &likelySubtags;
    Locale::Iterator &locales;
    ULocMatchLifetime lifetime;
    const Locale *current = nullptr, *remembered = nullptr;
    int32_t bestDesiredIndex = -1;
};

const Locale *LocaleMatcher::getBestMatch(const Locale &desiredLocale, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) { return nullptr; }
    int32_t suppIndex = getBestSuppIndex(
        getMaximalLsrOrUnd(likelySubtags, desiredLocale, errorCode),
        nullptr, errorCode);
    return U_SUCCESS(errorCode) && suppIndex >= 0 ? supportedLocales[suppIndex] : defaultLocale;
}

const Locale *LocaleMatcher::getBestMatch(Locale::Iterator &desiredLocales,
                                          UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) { return nullptr; }
    if (!desiredLocales.hasNext()) {
        return defaultLocale;
    }
    LocaleLsrIterator lsrIter(likelySubtags, desiredLocales, ULOCMATCH_TEMPORARY_LOCALES);
    int32_t suppIndex = getBestSuppIndex(lsrIter.next(errorCode), &lsrIter, errorCode);
    return U_SUCCESS(errorCode) && suppIndex >= 0 ? supportedLocales[suppIndex] : defaultLocale;
}

const Locale *LocaleMatcher::getBestMatchForListString(
        StringPiece desiredLocaleList, UErrorCode &errorCode) const {
    LocalePriorityList list(desiredLocaleList, errorCode);
    LocalePriorityList::Iterator iter = list.iterator();
    return getBestMatch(iter, errorCode);
}

LocaleMatcher::Result LocaleMatcher::getBestMatchResult(
        const Locale &desiredLocale, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) {
        return Result(nullptr, defaultLocale, -1, defaultLocaleIndex, FALSE);
    }
    int32_t suppIndex = getBestSuppIndex(
        getMaximalLsrOrUnd(likelySubtags, desiredLocale, errorCode),
        nullptr, errorCode);
    if (U_FAILURE(errorCode) || suppIndex < 0) {
        return Result(nullptr, defaultLocale, -1, defaultLocaleIndex, FALSE);
    } else {
        return Result(&desiredLocale, supportedLocales[suppIndex], 0, suppIndex, FALSE);
    }
}

LocaleMatcher::Result LocaleMatcher::getBestMatchResult(
        Locale::Iterator &desiredLocales, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode) || !desiredLocales.hasNext()) {
        return Result(nullptr, defaultLocale, -1, defaultLocaleIndex, FALSE);
    }
    LocaleLsrIterator lsrIter(likelySubtags, desiredLocales, ULOCMATCH_TEMPORARY_LOCALES);
    int32_t suppIndex = getBestSuppIndex(lsrIter.next(errorCode), &lsrIter, errorCode);
    if (U_FAILURE(errorCode) || suppIndex < 0) {
        return Result(nullptr, defaultLocale, -1, defaultLocaleIndex, FALSE);
    } else {
        return Result(lsrIter.orphanRemembered(), supportedLocales[suppIndex],
                      lsrIter.getBestDesiredIndex(), suppIndex, TRUE);
    }
}

int32_t LocaleMatcher::getBestSuppIndex(LSR desiredLSR, LocaleLsrIterator *remainingIter,
                                        UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) { return -1; }
    int32_t desiredIndex = 0;
    int32_t bestSupportedLsrIndex = -1;
    for (int32_t bestDistance = thresholdDistance;;) {
        // Quick check for exact maximized LSR.
        // Returns suppIndex+1 where 0 means not found.
        if (supportedLsrToIndex != nullptr) {
            desiredLSR.setHashCode();
            int32_t index = uhash_geti(supportedLsrToIndex, &desiredLSR);
            if (index != 0) {
                int32_t suppIndex = index - 1;
                if (remainingIter != nullptr) {
                    remainingIter->rememberCurrent(desiredIndex, errorCode);
                }
                return suppIndex;
            }
        }
        int32_t bestIndexAndDistance = localeDistance.getBestIndexAndDistance(
                desiredLSR, supportedLSRs, supportedLSRsLength, bestDistance, favorSubtag);
        if (bestIndexAndDistance >= 0) {
            bestDistance = bestIndexAndDistance & 0xff;
            if (remainingIter != nullptr) {
                remainingIter->rememberCurrent(desiredIndex, errorCode);
                if (U_FAILURE(errorCode)) { return -1; }
            }
            bestSupportedLsrIndex = bestIndexAndDistance >= 0 ? bestIndexAndDistance >> 8 : -1;
        }
        if ((bestDistance -= demotionPerDesiredLocale) <= 0) {
            break;
        }
        if (remainingIter == nullptr || !remainingIter->hasNext()) {
            break;
        }
        desiredLSR = remainingIter->next(errorCode);
        if (U_FAILURE(errorCode)) { return -1; }
        ++desiredIndex;
    }
    if (bestSupportedLsrIndex < 0) {
        // no good match
        return -1;
    }
    return supportedIndexes[bestSupportedLsrIndex];
}

double LocaleMatcher::internalMatch(const Locale &desired, const Locale &supported, UErrorCode &errorCode) const {
    // Returns the inverse of the distance: That is, 1-distance(desired, supported).
    LSR suppLSR = getMaximalLsrOrUnd(likelySubtags, supported, errorCode);
    if (U_FAILURE(errorCode)) { return 0; }
    const LSR *pSuppLSR = &suppLSR;
    int32_t distance = localeDistance.getBestIndexAndDistance(
            getMaximalLsrOrUnd(likelySubtags, desired, errorCode),
            &pSuppLSR, 1,
            thresholdDistance, favorSubtag) & 0xff;
    return (100 - distance) / 100.0;
}

U_NAMESPACE_END

// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License

// localeprioritylist.cpp
// created: 2019jul11 Markus W. Scherer













U_NAMESPACE_BEGIN

struct LocaleAndWeight {
    Locale *locale;
    int32_t weight;  // 0..1000 = 0.0..1.0
    int32_t index;  // force stable sort

    int32_t compare(const LocaleAndWeight &other) const {
        int32_t diff = other.weight - weight;  // descending: other-this
        if (diff != 0) { return diff; }
        return index - other.index;
    }
};

/**
 * Nothing but a wrapper over a MaybeStackArray of LocaleAndWeight.
 *
 * This wrapper exists (and is not in an anonymous namespace)
 * so that we can forward-declare it in the header file and
 * don't have to expose the MaybeStackArray specialization and
 * the LocaleAndWeight to code (like the test) that #includes localeprioritylist.h.
 * Also, otherwise we would have to do a platform-specific
 * template export declaration of some kind for the MaybeStackArray specialization
 * to be properly exported from the common DLL.
 */
struct LocaleAndWeightArray : public UMemory {
    MaybeStackArray<LocaleAndWeight, 20> array;
};

namespace {

int32_t hashLocale(const UHashTok token) {
    auto *locale = static_cast<const Locale *>(token.pointer);
    return locale->hashCode();
}

UBool compareLocales(const UHashTok t1, const UHashTok t2) {
    auto *l1 = static_cast<const Locale *>(t1.pointer);
    auto *l2 = static_cast<const Locale *>(t2.pointer);
    return *l1 == *l2;
}

constexpr int32_t WEIGHT_ONE = 1000;

int32_t U_CALLCONV
compareLocaleAndWeight(const void * /*context*/, const void *left, const void *right) {
    return static_cast<const LocaleAndWeight *>(left)->
        compare(*static_cast<const LocaleAndWeight *>(right));
}

const char *skipSpaces(const char *p, const char *limit) {
    while (p < limit && *p == ' ') { ++p; }
    return p;
}

int32_t findTagLength(const char *p, const char *limit) {
    // Look for accept-language delimiters.
    // Leave other validation up to the Locale constructor.
    const char *q;
    for (q = p; q < limit; ++q) {
        char c = *q;
        if (c == ' ' || c == ',' || c == ';') { break; }
    }
    return static_cast<int32_t>(q - p);
}

/**
 * Parses and returns a qvalue weight in millis.
 * Advances p to after the parsed substring.
 * Returns a negative value if parsing fails.
 */
int32_t parseWeight(const char *&p, const char *limit) {
    p = skipSpaces(p, limit);
    char c;
    if (p == limit || ((c = *p) != '0' && c != '1')) { return -1; }
    int32_t weight = (c - '0') * 1000;
    if (++p == limit || *p != '.') { return weight; }
    int32_t multiplier = 100;
    while (++p != limit && '0' <= (c = *p) && c <= '9') {
        c -= '0';
        if (multiplier > 0) {
            weight += c * multiplier;
            multiplier /= 10;
        } else if (multiplier == 0) {
            // round up
            if (c >= 5) { ++weight; }
            multiplier = -1;
        }  // else ignore further fraction digits
    }
    return weight <= WEIGHT_ONE ? weight : -1;  // bad if > 1.0
}

}  // namespace

LocalePriorityList::LocalePriorityList(StringPiece s, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return; }
    list = new LocaleAndWeightArray();
    if (list == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    const char *p = s.data();
    const char *limit = p + s.length();
    while ((p = skipSpaces(p, limit)) != limit) {
        if (*p == ',') {  // empty range field
            ++p;
            continue;
        }
        int32_t tagLength = findTagLength(p, limit);
        if (tagLength == 0) {
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        CharString tag(p, tagLength, errorCode);
        if (U_FAILURE(errorCode)) { return; }
        Locale locale = Locale(tag.data());
        if (locale.isBogus()) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        int32_t weight = WEIGHT_ONE;
        if ((p = skipSpaces(p + tagLength, limit)) != limit && *p == ';') {
            if ((p = skipSpaces(p + 1, limit)) == limit || *p != 'q' ||
                    (p = skipSpaces(p + 1, limit)) == limit || *p != '=' ||
                    (++p, (weight = parseWeight(p, limit)) < 0)) {
                errorCode = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
            p = skipSpaces(p, limit);
        }
        if (p != limit && *p != ',') {  // trailing junk
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        add(locale, weight, errorCode);
        if (p == limit) { break; }
        ++p;
    }
    sort(errorCode);
}

LocalePriorityList::~LocalePriorityList() {
    if (list != nullptr) {
        for (int32_t i = 0; i < listLength; ++i) {
            delete list->array[i].locale;
        }
        delete list;
    }
    uhash_close(map);
}

const Locale *LocalePriorityList::localeAt(int32_t i) const {
    return list->array[i].locale;
}

Locale *LocalePriorityList::orphanLocaleAt(int32_t i) {
    if (list == nullptr) { return nullptr; }
    LocaleAndWeight &lw = list->array[i];
    Locale *l = lw.locale;
    lw.locale = nullptr;
    return l;
}

bool LocalePriorityList::add(const Locale &locale, int32_t weight, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return false; }
    if (map == nullptr) {
        if (weight <= 0) { return true; }  // do not add q=0
        map = uhash_open(hashLocale, compareLocales, uhash_compareLong, &errorCode);
        if (U_FAILURE(errorCode)) { return false; }
    }
    LocalPointer<Locale> clone;
    int32_t index = uhash_geti(map, &locale);
    if (index != 0) {
        // Duplicate: Remove the old item and append it anew.
        LocaleAndWeight &lw = list->array[index - 1];
        clone.adoptInstead(lw.locale);
        lw.locale = nullptr;
        lw.weight = 0;
        ++numRemoved;
    }
    if (weight <= 0) {  // do not add q=0
        if (index != 0) {
            // Not strictly necessary but cleaner.
            uhash_removei(map, &locale);
        }
        return true;
    }
    if (clone.isNull()) {
        clone.adoptInstead(locale.clone());
        if (clone.isNull() || (clone->isBogus() && !locale.isBogus())) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return false;
        }
    }
    if (listLength == list->array.getCapacity()) {
        int32_t newCapacity = listLength < 50 ? 100 : 4 * listLength;
        if (list->array.resize(newCapacity, listLength) == nullptr) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return false;
        }
    }
    uhash_puti(map, clone.getAlias(), listLength + 1, &errorCode);
    if (U_FAILURE(errorCode)) { return false; }
    LocaleAndWeight &lw = list->array[listLength];
    lw.locale = clone.orphan();
    lw.weight = weight;
    lw.index = listLength++;
    if (weight < WEIGHT_ONE) { hasWeights = true; }
    U_ASSERT(uhash_count(map) == getLength());
    return true;
}

void LocalePriorityList::sort(UErrorCode &errorCode) {
    // Sort by descending weights if there is a mix of weights.
    // The comparator forces a stable sort via the item index.
    if (U_FAILURE(errorCode) || getLength() <= 1 || !hasWeights) { return; }
    uprv_sortArray(list->array.getAlias(), listLength, sizeof(LocaleAndWeight),
                   compareLocaleAndWeight, nullptr, FALSE, &errorCode);
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1997-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  locavailable.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010feb25
*   created by: Markus W. Scherer
*
*   Code for available locales, separated out from other .cpp files
*   that then do not depend on resource bundle code and res_index bundles.
*/













// C++ API ----------------------------------------------------------------- ***

U_NAMESPACE_BEGIN

static icu::Locale*  gLocAvailable_availableLocaleList = NULL;
static int32_t  gLocAvailable_availableLocaleListCount;
static icu::UInitOnce gLocAvailableInitOnceLocale = U_INITONCE_INITIALIZER;

U_NAMESPACE_END

U_CDECL_BEGIN

static UBool U_CALLCONV locale_available_cleanup(void)
{
    U_NAMESPACE_USE

    if (gLocAvailable_availableLocaleList) {
        delete []gLocAvailable_availableLocaleList;
        gLocAvailable_availableLocaleList = NULL;
    }
    gLocAvailable_availableLocaleListCount = 0;
    gLocAvailableInitOnceLocale.reset();

    return TRUE;
}

U_CDECL_END

U_NAMESPACE_BEGIN

void U_CALLCONV locale_available_init() {
    // This function is a friend of class Locale.
    // This function is only invoked via umtx_initOnce().

    // for now, there is a hardcoded list, so just walk through that list and set it up.
    //  Note: this function is a friend of class Locale.
    gLocAvailable_availableLocaleListCount = uloc_countAvailable();
    if(gLocAvailable_availableLocaleListCount) {
       gLocAvailable_availableLocaleList = new Locale[gLocAvailable_availableLocaleListCount];
    }
    if (gLocAvailable_availableLocaleList == NULL) {
        gLocAvailable_availableLocaleListCount= 0;
    }
    for (int32_t locCount=gLocAvailable_availableLocaleListCount-1; locCount>=0; --locCount) {
        gLocAvailable_availableLocaleList[locCount].setFromPOSIXID(uloc_getAvailable(locCount));
    }
    ucln_common_registerCleanup(UCLN_COMMON_LOCALE_AVAILABLE, locale_available_cleanup);
}

const Locale* U_EXPORT2
Locale::getAvailableLocales(int32_t& count)
{
    umtx_initOnce(gLocAvailableInitOnceLocale, &locale_available_init);
    count = gLocAvailable_availableLocaleListCount;
    return gLocAvailable_availableLocaleList;
}


U_NAMESPACE_END

// C API ------------------------------------------------------------------- ***

U_NAMESPACE_USE

/* ### Constants **************************************************/

namespace {

// Enough capacity for the two lists in the res_index.res file
const char** gAvailableLocaleNames[2] = {};
int32_t gAvailableLocaleCounts[2] = {};
icu::UInitOnce ginstalledLocalesInitOnce = U_INITONCE_INITIALIZER;

class AvailableLocalesSink : public ResourceSink {
  public:
    void put(const char *key, ResourceValue &value, UBool /*noFallback*/, UErrorCode &status) U_OVERRIDE {
        ResourceTable resIndexTable = value.getTable(status);
        if (U_FAILURE(status)) {
            return;
        }
        for (int32_t i = 0; resIndexTable.getKeyAndValue(i, key, value); ++i) {
            ULocAvailableType type;
            if (uprv_strcmp(key, "InstalledLocales") == 0) {
                type = ULOC_AVAILABLE_DEFAULT;
            } else if (uprv_strcmp(key, "AliasLocales") == 0) {
                type = ULOC_AVAILABLE_ONLY_LEGACY_ALIASES;
            } else {
                // CLDRVersion, etc.
                continue;
            }
            ResourceTable availableLocalesTable = value.getTable(status);
            if (U_FAILURE(status)) {
                return;
            }
            gAvailableLocaleCounts[type] = availableLocalesTable.getSize();
            gAvailableLocaleNames[type] = static_cast<const char**>(
                uprv_malloc(gAvailableLocaleCounts[type] * sizeof(const char*)));
            if (gAvailableLocaleNames[type] == nullptr) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            for (int32_t j = 0; availableLocalesTable.getKeyAndValue(j, key, value); ++j) {
                gAvailableLocaleNames[type][j] = key;
            }
        }
    }
};

class AvailableLocalesStringEnumeration : public StringEnumeration {
  public:
    AvailableLocalesStringEnumeration(ULocAvailableType type) : fType(type) {
    }

    const char* next(int32_t *resultLength, UErrorCode&) override {
        ULocAvailableType actualType = fType;
        int32_t actualIndex = fIndex++;

        // If the "combined" list was requested, resolve that now
        if (fType == ULOC_AVAILABLE_WITH_LEGACY_ALIASES) {
            int32_t defaultLocalesCount = gAvailableLocaleCounts[ULOC_AVAILABLE_DEFAULT];
            if (actualIndex < defaultLocalesCount) {
                actualType = ULOC_AVAILABLE_DEFAULT;
            } else {
                actualIndex -= defaultLocalesCount;
                actualType = ULOC_AVAILABLE_ONLY_LEGACY_ALIASES;
            }
        }

        // Return the requested string
        int32_t count = gAvailableLocaleCounts[actualType];
        const char* result;
        if (actualIndex < count) {
            result = gAvailableLocaleNames[actualType][actualIndex];
            if (resultLength != nullptr) {
                *resultLength = static_cast<int32_t>(uprv_strlen(result));
            }
        } else {
            result = nullptr;
            if (resultLength != nullptr) {
                *resultLength = 0;
            }
        }
        return result;
    }

    void reset(UErrorCode&) override {
        fIndex = 0;
    }

    int32_t count(UErrorCode&) const override {
        if (fType == ULOC_AVAILABLE_WITH_LEGACY_ALIASES) {
            return gAvailableLocaleCounts[ULOC_AVAILABLE_DEFAULT]
                + gAvailableLocaleCounts[ULOC_AVAILABLE_ONLY_LEGACY_ALIASES];
        } else {
            return gAvailableLocaleCounts[fType];
        }
    }

  private:
    ULocAvailableType fType;
    int32_t fIndex = 0;
};

/* ### Get available **************************************************/

static UBool U_CALLCONV uloc_cleanup(void) {
    for (int32_t i = 0; i < UPRV_LENGTHOF(gAvailableLocaleNames); i++) {
        uprv_free(gAvailableLocaleNames[i]);
        gAvailableLocaleNames[i] = nullptr;
        gAvailableLocaleCounts[i] = 0;
    }
    ginstalledLocalesInitOnce.reset();
    return TRUE;
}

// Load Installed Locales. This function will be called exactly once
//   via the initOnce mechanism.

static void U_CALLCONV loadInstalledLocales(UErrorCode& status) {
    ucln_common_registerCleanup(UCLN_COMMON_ULOC, uloc_cleanup);

    icu::LocalUResourceBundlePointer rb(ures_openDirect(NULL, "res_index", &status));
    AvailableLocalesSink sink;
    ures_getAllItemsWithFallback(rb.getAlias(), "", sink, status);
}

void _load_installedLocales(UErrorCode& status) {
    umtx_initOnce(ginstalledLocalesInitOnce, &loadInstalledLocales, status);
}

} // namespace

U_CAPI const char* U_EXPORT2
uloc_getAvailable(int32_t offset) {
    icu::ErrorCode status;
    _load_installedLocales(status);
    if (status.isFailure()) {
        return nullptr;
    }
    if (offset > gAvailableLocaleCounts[0]) {
        // *status = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    return gAvailableLocaleNames[0][offset];
}

U_CAPI int32_t  U_EXPORT2
uloc_countAvailable() {
    icu::ErrorCode status;
    _load_installedLocales(status);
    if (status.isFailure()) {
        return 0;
    }
    return gAvailableLocaleCounts[0];
}

U_CAPI UEnumeration* U_EXPORT2
uloc_openAvailableByType(ULocAvailableType type, UErrorCode* status) {
    if (U_FAILURE(*status)) {
        return nullptr;
    }
    if (type < 0 || type >= ULOC_AVAILABLE_COUNT) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    _load_installedLocales(*status);
    if (U_FAILURE(*status)) {
        return nullptr;
    }
    LocalPointer<AvailableLocalesStringEnumeration> result(
        new AvailableLocalesStringEnumeration(type), *status);
    if (U_FAILURE(*status)) {
        return nullptr;
    }
    return uenum_openFromStringEnumeration(result.orphan(), status);
}

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: January 16 2004
* Since: ICU 2.8
**********************************************************************
*/
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: January 16 2004
* Since: ICU 2.8
**********************************************************************
*/




/**
 * Macro to declare a locale LocaleBased wrapper object for the given
 * object, which must have two members named `validLocale' and
 * `actualLocale' of size ULOC_FULLNAME_CAPACITY
 */
#define U_LOCALE_BASED(varname, objname) \
  LocaleBased varname((objname).validLocale, (objname).actualLocale)

U_NAMESPACE_BEGIN

/**
 * A utility class that unifies the implementation of getLocale() by
 * various ICU services.  This class is likely to be removed in the
 * ICU 3.0 time frame in favor of an integrated approach with the
 * services framework.
 * @since ICU 2.8
 */
class U_COMMON_API LocaleBased : public UMemory {

 public:

    /**
     * Construct a LocaleBased wrapper around the two pointers.  These
     * will be aliased for the lifetime of this object.
     */
    inline LocaleBased(char* validAlias, char* actualAlias);

    /**
     * Construct a LocaleBased wrapper around the two const pointers.
     * These will be aliased for the lifetime of this object.
     */
    inline LocaleBased(const char* validAlias, const char* actualAlias);

    /**
     * Return locale meta-data for the service object wrapped by this
     * object.  Either the valid or the actual locale may be
     * retrieved.
     * @param type either ULOC_VALID_LOCALE or ULOC_ACTUAL_LOCALE
     * @param status input-output error code
     * @return the indicated locale
     */
    Locale getLocale(ULocDataLocaleType type, UErrorCode& status) const;

    /**
     * Return the locale ID for the service object wrapped by this
     * object.  Either the valid or the actual locale may be
     * retrieved.
     * @param type either ULOC_VALID_LOCALE or ULOC_ACTUAL_LOCALE
     * @param status input-output error code
     * @return the indicated locale ID
     */
    const char* getLocaleID(ULocDataLocaleType type, UErrorCode& status) const;

    /**
     * Set the locale meta-data for the service object wrapped by this
     * object.  If either parameter is zero, it is ignored.
     * @param valid the ID of the valid locale
     * @param actual the ID of the actual locale
     */
    void setLocaleIDs(const char* valid, const char* actual);

    /**
     * Set the locale meta-data for the service object wrapped by this
     * object.
     * @param valid the ID of the valid locale
     * @param actual the ID of the actual locale
     */
    void setLocaleIDs(const Locale& valid, const Locale& actual);

 private:

    char* valid;
    
    char* actual;
};

inline LocaleBased::LocaleBased(char* validAlias, char* actualAlias) :
    valid(validAlias), actual(actualAlias) {
}

inline LocaleBased::LocaleBased(const char* validAlias,
                                const char* actualAlias) :
    // ugh: cast away const
    valid((char*)validAlias), actual((char*)actualAlias) {
}

U_NAMESPACE_END




U_NAMESPACE_BEGIN

Locale LocaleBased::getLocale(ULocDataLocaleType type, UErrorCode& status) const {
    const char* id = getLocaleID(type, status);
    return Locale((id != 0) ? id : "");
}

const char* LocaleBased::getLocaleID(ULocDataLocaleType type, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return NULL;
    }

    switch(type) {
    case ULOC_VALID_LOCALE:
        return valid;
    case ULOC_ACTUAL_LOCALE:
        return actual;
    default:
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
}

void LocaleBased::setLocaleIDs(const char* validID, const char* actualID) {
    if (validID != 0) {
      uprv_strncpy(valid, validID, ULOC_FULLNAME_CAPACITY);
      valid[ULOC_FULLNAME_CAPACITY-1] = 0; // always terminate
    }
    if (actualID != 0) {
      uprv_strncpy(actual, actualID, ULOC_FULLNAME_CAPACITY);
      actual[ULOC_FULLNAME_CAPACITY-1] = 0; // always terminate
    }
}

void LocaleBased::setLocaleIDs(const Locale& validID, const Locale& actualID) {
  uprv_strcpy(valid, validID.getName());
  uprv_strcpy(actual, actualID.getName());
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  locdispnames.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010feb25
*   created by: Markus W. Scherer
*
*   Code for locale display names, separated out from other .cpp files
*   that then do not depend on resource bundle code and display name data.
*/


// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// ********************************************************************************
// *   Copyright (C) 1997-2016, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// ********************************************************************************
// *
// * File brkiter.h
// *
// * Modification History:
// *
// *   Date        Name        Description
// *   02/18/97    aliu        Added typedef for TextCount.  Made DONE const.
// *   05/07/97    aliu        Fixed DLL declaration.
// *   07/09/97    jfitz       Renamed BreakIterator and interface synced with JDK
// *   08/11/98    helena      Sync-up JDK1.2.
// *   01/13/2000  helena      Added UErrorCode parameter to createXXXInstance methods.
// ********************************************************************************
// */

// #ifndef BRKITER_H
// #define BRKITER_H

// #include "unicode/utypes.h"

// /**
//  * \file
//  * \brief C++ API: Break Iterator.
//  */

// #include "unicode/utypes.h"
// #include "unicode/uobject.h"
// #include "unicode/unistr.h"
// #include "unicode/chariter.h"
// 
// #include "unicode/ubrk.h"
// #include "unicode/strenum.h"
// #include "unicode/utext.h"
// #include "unicode/umisc.h"

// #if U_SHOW_CPLUSPLUS_API

// #if UCONFIG_NO_BREAK_ITERATION

// U_NAMESPACE_BEGIN

// /*
//  * Allow the declaration of APIs with pointers to BreakIterator
//  * even when break iteration is removed from the build.
//  */
// class BreakIterator;

// U_NAMESPACE_END

// #else

// U_NAMESPACE_BEGIN

// /**
//  * The BreakIterator class implements methods for finding the location
//  * of boundaries in text. BreakIterator is an abstract base class.
//  * Instances of BreakIterator maintain a current position and scan over
//  * text returning the index of characters where boundaries occur.
//  * <p>
//  * Line boundary analysis determines where a text string can be broken
//  * when line-wrapping. The mechanism correctly handles punctuation and
//  * hyphenated words.
//  * <p>
//  * Sentence boundary analysis allows selection with correct
//  * interpretation of periods within numbers and abbreviations, and
//  * trailing punctuation marks such as quotation marks and parentheses.
//  * <p>
//  * Word boundary analysis is used by search and replace functions, as
//  * well as within text editing applications that allow the user to
//  * select words with a double click. Word selection provides correct
//  * interpretation of punctuation marks within and following
//  * words. Characters that are not part of a word, such as symbols or
//  * punctuation marks, have word-breaks on both sides.
//  * <p>
//  * Character boundary analysis allows users to interact with
//  * characters as they expect to, for example, when moving the cursor
//  * through a text string. Character boundary analysis provides correct
//  * navigation of through character strings, regardless of how the
//  * character is stored.  For example, an accented character might be
//  * stored as a base character and a diacritical mark. What users
//  * consider to be a character can differ between languages.
//  * <p>
//  * The text boundary positions are found according to the rules
//  * described in Unicode Standard Annex #29, Text Boundaries, and
//  * Unicode Standard Annex #14, Line Breaking Properties.  These
//  * are available at http://www.unicode.org/reports/tr14/ and
//  * http://www.unicode.org/reports/tr29/.
//  * <p>
//  * In addition to the C++ API defined in this header file, a
//  * plain C API with equivalent functionality is defined in the
//  * file ubrk.h
//  * <p>
//  * Code snippets illustrating the use of the Break Iterator APIs
//  * are available in the ICU User Guide,
//  * http://icu-project.org/userguide/boundaryAnalysis.html
//  * and in the sample program icu/source/samples/break/break.cpp
//  *
//  */
// class U_COMMON_API BreakIterator : public UObject {
// public:
//     /**
//      *  destructor
//      *  @stable ICU 2.0
//      */
//     virtual ~BreakIterator();

//     /**
//      * Return true if another object is semantically equal to this
//      * one. The other object should be an instance of the same subclass of
//      * BreakIterator. Objects of different subclasses are considered
//      * unequal.
//      * <P>
//      * Return true if this BreakIterator is at the same position in the
//      * same text, and is the same class and type (word, line, etc.) of
//      * BreakIterator, as the argument.  Text is considered the same if
//      * it contains the same characters, it need not be the same
//      * object, and styles are not considered.
//      * @stable ICU 2.0
//      */
//     virtual UBool operator==(const BreakIterator&) const = 0;

//     /**
//      * Returns the complement of the result of operator==
//      * @param rhs The BreakIterator to be compared for inequality
//      * @return the complement of the result of operator==
//      * @stable ICU 2.0
//      */
//     UBool operator!=(const BreakIterator& rhs) const { return !operator==(rhs); }

//     /**
//      * Return a polymorphic copy of this object.  This is an abstract
//      * method which subclasses implement.
//      * @stable ICU 2.0
//      */
//     virtual BreakIterator* clone() const = 0;

//     /**
//      * Return a polymorphic class ID for this object. Different subclasses
//      * will return distinct unequal values.
//      * @stable ICU 2.0
//      */
//     virtual UClassID getDynamicClassID(void) const = 0;

//     /**
//      * Return a CharacterIterator over the text being analyzed.
//      * @stable ICU 2.0
//      */
//     virtual CharacterIterator& getText(void) const = 0;


//     /**
//       *  Get a UText for the text being analyzed.
//       *  The returned UText is a shallow clone of the UText used internally
//       *  by the break iterator implementation.  It can safely be used to
//       *  access the text without impacting any break iterator operations,
//       *  but the underlying text itself must not be altered.
//       *
//       * @param fillIn A UText to be filled in.  If NULL, a new UText will be
//       *           allocated to hold the result.
//       * @param status receives any error codes.
//       * @return   The current UText for this break iterator.  If an input
//       *           UText was provided, it will always be returned.
//       * @stable ICU 3.4
//       */
//      virtual UText *getUText(UText *fillIn, UErrorCode &status) const = 0;

//     /**
//      * Change the text over which this operates. The text boundary is
//      * reset to the start.
//      *
//      * The BreakIterator will retain a reference to the supplied string.
//      * The caller must not modify or delete the text while the BreakIterator
//      * retains the reference.
//      *
//      * @param text The UnicodeString used to change the text.
//      * @stable ICU 2.0
//      */
//     virtual void  setText(const UnicodeString &text) = 0;

//     /**
//      * Reset the break iterator to operate over the text represented by
//      * the UText.  The iterator position is reset to the start.
//      *
//      * This function makes a shallow clone of the supplied UText.  This means
//      * that the caller is free to immediately close or otherwise reuse the
//      * Utext that was passed as a parameter, but that the underlying text itself
//      * must not be altered while being referenced by the break iterator.
//      *
//      * All index positions returned by break iterator functions are
//      * native indices from the UText. For example, when breaking UTF-8
//      * encoded text, the break positions returned by next(), previous(), etc.
//      * will be UTF-8 string indices, not UTF-16 positions.
//      *
//      * @param text The UText used to change the text.
//      * @param status receives any error codes.
//      * @stable ICU 3.4
//      */
//     virtual void  setText(UText *text, UErrorCode &status) = 0;

//     /**
//      * Change the text over which this operates. The text boundary is
//      * reset to the start.
//      * Note that setText(UText *) provides similar functionality to this function,
//      * and is more efficient.
//      * @param it The CharacterIterator used to change the text.
//      * @stable ICU 2.0
//      */
//     virtual void  adoptText(CharacterIterator* it) = 0;

//     enum {
//         /**
//          * DONE is returned by previous() and next() after all valid
//          * boundaries have been returned.
//          * @stable ICU 2.0
//          */
//         DONE = (int32_t)-1
//     };

//     /**
//      * Sets the current iteration position to the beginning of the text, position zero.
//      * @return The offset of the beginning of the text, zero.
//      * @stable ICU 2.0
//      */
//     virtual int32_t first(void) = 0;

//     /**
//      * Set the iterator position to the index immediately BEYOND the last character in the text being scanned.
//      * @return The index immediately BEYOND the last character in the text being scanned.
//      * @stable ICU 2.0
//      */
//     virtual int32_t last(void) = 0;

//     /**
//      * Set the iterator position to the boundary preceding the current boundary.
//      * @return The character index of the previous text boundary or DONE if all
//      * boundaries have been returned.
//      * @stable ICU 2.0
//      */
//     virtual int32_t previous(void) = 0;

//     /**
//      * Advance the iterator to the boundary following the current boundary.
//      * @return The character index of the next text boundary or DONE if all
//      * boundaries have been returned.
//      * @stable ICU 2.0
//      */
//     virtual int32_t next(void) = 0;

//     /**
//      * Return character index of the current iterator position within the text.
//      * @return The boundary most recently returned.
//      * @stable ICU 2.0
//      */
//     virtual int32_t current(void) const = 0;

//     /**
//      * Advance the iterator to the first boundary following the specified offset.
//      * The value returned is always greater than the offset or
//      * the value BreakIterator.DONE
//      * @param offset the offset to begin scanning.
//      * @return The first boundary after the specified offset.
//      * @stable ICU 2.0
//      */
//     virtual int32_t following(int32_t offset) = 0;

//     /**
//      * Set the iterator position to the first boundary preceding the specified offset.
//      * The value returned is always smaller than the offset or
//      * the value BreakIterator.DONE
//      * @param offset the offset to begin scanning.
//      * @return The first boundary before the specified offset.
//      * @stable ICU 2.0
//      */
//     virtual int32_t preceding(int32_t offset) = 0;

//     /**
//      * Return true if the specified position is a boundary position.
//      * As a side effect, the current position of the iterator is set
//      * to the first boundary position at or following the specified offset.
//      * @param offset the offset to check.
//      * @return True if "offset" is a boundary position.
//      * @stable ICU 2.0
//      */
//     virtual UBool isBoundary(int32_t offset) = 0;

//     /**
//      * Set the iterator position to the nth boundary from the current boundary
//      * @param n the number of boundaries to move by.  A value of 0
//      * does nothing.  Negative values move to previous boundaries
//      * and positive values move to later boundaries.
//      * @return The new iterator position, or
//      * DONE if there are fewer than |n| boundaries in the specified direction.
//      * @stable ICU 2.0
//      */
//     virtual int32_t next(int32_t n) = 0;

//    /**
//      * For RuleBasedBreakIterators, return the status tag from the break rule
//      * that determined the boundary at the current iteration position.
//      * <p>
//      * For break iterator types that do not support a rule status,
//      * a default value of 0 is returned.
//      * <p>
//      * @return the status from the break rule that determined the boundary at
//      *         the current iteration position.
//      * @see RuleBaseBreakIterator::getRuleStatus()
//      * @see UWordBreak
//      * @stable ICU 52
//      */
//     virtual int32_t getRuleStatus() const;

//    /**
//     * For RuleBasedBreakIterators, get the status (tag) values from the break rule(s)
//     * that determined the boundary at the current iteration position.
//     * <p>
//     * For break iterator types that do not support rule status,
//     * no values are returned.
//     * <p>
//     * The returned status value(s) are stored into an array provided by the caller.
//     * The values are stored in sorted (ascending) order.
//     * If the capacity of the output array is insufficient to hold the data,
//     *  the output will be truncated to the available length, and a
//     *  U_BUFFER_OVERFLOW_ERROR will be signaled.
//     * <p>
//     * @see RuleBaseBreakIterator::getRuleStatusVec
//     *
//     * @param fillInVec an array to be filled in with the status values.
//     * @param capacity  the length of the supplied vector.  A length of zero causes
//     *                  the function to return the number of status values, in the
//     *                  normal way, without attempting to store any values.
//     * @param status    receives error codes.
//     * @return          The number of rule status values from rules that determined
//     *                  the boundary at the current iteration position.
//     *                  In the event of a U_BUFFER_OVERFLOW_ERROR, the return value
//     *                  is the total number of status values that were available,
//     *                  not the reduced number that were actually returned.
//     * @see getRuleStatus
//     * @stable ICU 52
//     */
//     virtual int32_t getRuleStatusVec(int32_t *fillInVec, int32_t capacity, UErrorCode &status);

//     /**
//      * Create BreakIterator for word-breaks using the given locale.
//      * Returns an instance of a BreakIterator implementing word breaks.
//      * WordBreak is useful for word selection (ex. double click)
//      * @param where the locale.
//      * @param status the error code
//      * @return A BreakIterator for word-breaks.  The UErrorCode& status
//      * parameter is used to return status information to the user.
//      * To check whether the construction succeeded or not, you should check
//      * the value of U_SUCCESS(err).  If you wish more detailed information, you
//      * can check for informational error results which still indicate success.
//      * U_USING_FALLBACK_WARNING indicates that a fall back locale was used.  For
//      * example, 'de_CH' was requested, but nothing was found there, so 'de' was
//      * used.  U_USING_DEFAULT_WARNING indicates that the default locale data was
//      * used; neither the requested locale nor any of its fall back locales
//      * could be found.
//      * The caller owns the returned object and is responsible for deleting it.
//      * @stable ICU 2.0
//      */
//     static BreakIterator* U_EXPORT2
//     createWordInstance(const Locale& where, UErrorCode& status);

//     /**
//      * Create BreakIterator for line-breaks using specified locale.
//      * Returns an instance of a BreakIterator implementing line breaks. Line
//      * breaks are logically possible line breaks, actual line breaks are
//      * usually determined based on display width.
//      * LineBreak is useful for word wrapping text.
//      * @param where the locale.
//      * @param status The error code.
//      * @return A BreakIterator for line-breaks.  The UErrorCode& status
//      * parameter is used to return status information to the user.
//      * To check whether the construction succeeded or not, you should check
//      * the value of U_SUCCESS(err).  If you wish more detailed information, you
//      * can check for informational error results which still indicate success.
//      * U_USING_FALLBACK_WARNING indicates that a fall back locale was used.  For
//      * example, 'de_CH' was requested, but nothing was found there, so 'de' was
//      * used.  U_USING_DEFAULT_WARNING indicates that the default locale data was
//      * used; neither the requested locale nor any of its fall back locales
//      * could be found.
//      * The caller owns the returned object and is responsible for deleting it.
//      * @stable ICU 2.0
//      */
//     static BreakIterator* U_EXPORT2
//     createLineInstance(const Locale& where, UErrorCode& status);

//     /**
//      * Create BreakIterator for character-breaks using specified locale
//      * Returns an instance of a BreakIterator implementing character breaks.
//      * Character breaks are boundaries of combining character sequences.
//      * @param where the locale.
//      * @param status The error code.
//      * @return A BreakIterator for character-breaks.  The UErrorCode& status
//      * parameter is used to return status information to the user.
//      * To check whether the construction succeeded or not, you should check
//      * the value of U_SUCCESS(err).  If you wish more detailed information, you
//      * can check for informational error results which still indicate success.
//      * U_USING_FALLBACK_WARNING indicates that a fall back locale was used.  For
//      * example, 'de_CH' was requested, but nothing was found there, so 'de' was
//      * used.  U_USING_DEFAULT_WARNING indicates that the default locale data was
//      * used; neither the requested locale nor any of its fall back locales
//      * could be found.
//      * The caller owns the returned object and is responsible for deleting it.
//      * @stable ICU 2.0
//      */
//     static BreakIterator* U_EXPORT2
//     createCharacterInstance(const Locale& where, UErrorCode& status);

//     /**
//      * Create BreakIterator for sentence-breaks using specified locale
//      * Returns an instance of a BreakIterator implementing sentence breaks.
//      * @param where the locale.
//      * @param status The error code.
//      * @return A BreakIterator for sentence-breaks.  The UErrorCode& status
//      * parameter is used to return status information to the user.
//      * To check whether the construction succeeded or not, you should check
//      * the value of U_SUCCESS(err).  If you wish more detailed information, you
//      * can check for informational error results which still indicate success.
//      * U_USING_FALLBACK_WARNING indicates that a fall back locale was used.  For
//      * example, 'de_CH' was requested, but nothing was found there, so 'de' was
//      * used.  U_USING_DEFAULT_WARNING indicates that the default locale data was
//      * used; neither the requested locale nor any of its fall back locales
//      * could be found.
//      * The caller owns the returned object and is responsible for deleting it.
//      * @stable ICU 2.0
//      */
//     static BreakIterator* U_EXPORT2
//     createSentenceInstance(const Locale& where, UErrorCode& status);

// #ifndef U_HIDE_DEPRECATED_API
//     /**
//      * Create BreakIterator for title-casing breaks using the specified locale
//      * Returns an instance of a BreakIterator implementing title breaks.
//      * The iterator returned locates title boundaries as described for
//      * Unicode 3.2 only. For Unicode 4.0 and above title boundary iteration,
//      * please use a word boundary iterator. See {@link #createWordInstance }.
//      *
//      * @param where the locale.
//      * @param status The error code.
//      * @return A BreakIterator for title-breaks.  The UErrorCode& status
//      * parameter is used to return status information to the user.
//      * To check whether the construction succeeded or not, you should check
//      * the value of U_SUCCESS(err).  If you wish more detailed information, you
//      * can check for informational error results which still indicate success.
//      * U_USING_FALLBACK_WARNING indicates that a fall back locale was used.  For
//      * example, 'de_CH' was requested, but nothing was found there, so 'de' was
//      * used.  U_USING_DEFAULT_WARNING indicates that the default locale data was
//      * used; neither the requested locale nor any of its fall back locales
//      * could be found.
//      * The caller owns the returned object and is responsible for deleting it.
//      * @deprecated ICU 64 Use createWordInstance instead.
//      */
//     static BreakIterator* U_EXPORT2
//     createTitleInstance(const Locale& where, UErrorCode& status);
// #endif /* U_HIDE_DEPRECATED_API */

//     /**
//      * Get the set of Locales for which TextBoundaries are installed.
//      * <p><b>Note:</b> this will not return locales added through the register
//      * call. To see the registered locales too, use the getAvailableLocales
//      * function that returns a StringEnumeration object </p>
//      * @param count the output parameter of number of elements in the locale list
//      * @return available locales
//      * @stable ICU 2.0
//      */
//     static const Locale* U_EXPORT2 getAvailableLocales(int32_t& count);

//     /**
//      * Get name of the object for the desired Locale, in the desired language.
//      * @param objectLocale must be from getAvailableLocales.
//      * @param displayLocale specifies the desired locale for output.
//      * @param name the fill-in parameter of the return value
//      * Uses best match.
//      * @return user-displayable name
//      * @stable ICU 2.0
//      */
//     static UnicodeString& U_EXPORT2 getDisplayName(const Locale& objectLocale,
//                                          const Locale& displayLocale,
//                                          UnicodeString& name);

//     /**
//      * Get name of the object for the desired Locale, in the language of the
//      * default locale.
//      * @param objectLocale must be from getMatchingLocales
//      * @param name the fill-in parameter of the return value
//      * @return user-displayable name
//      * @stable ICU 2.0
//      */
//     static UnicodeString& U_EXPORT2 getDisplayName(const Locale& objectLocale,
//                                          UnicodeString& name);

// #ifndef U_FORCE_HIDE_DEPRECATED_API
//     /**
//      * Deprecated functionality. Use clone() instead.
//      *
//      * Thread safe client-buffer-based cloning operation
//      *    Do NOT call delete on a safeclone, since 'new' is not used to create it.
//      * @param stackBuffer user allocated space for the new clone. If NULL new memory will be allocated.
//      * If buffer is not large enough, new memory will be allocated.
//      * @param BufferSize reference to size of allocated space.
//      * If BufferSize == 0, a sufficient size for use in cloning will
//      * be returned ('pre-flighting')
//      * If BufferSize is not enough for a stack-based safe clone,
//      * new memory will be allocated.
//      * @param status to indicate whether the operation went on smoothly or there were errors
//      *  An informational status value, U_SAFECLONE_ALLOCATED_ERROR, is used if any allocations were
//      *  necessary.
//      * @return pointer to the new clone
//      *
//      * @deprecated ICU 52. Use clone() instead.
//      */
//     virtual BreakIterator *  createBufferClone(void *stackBuffer,
//                                                int32_t &BufferSize,
//                                                UErrorCode &status) = 0;
// #endif  // U_FORCE_HIDE_DEPRECATED_API

// #ifndef U_HIDE_DEPRECATED_API

//     /**
//      *   Determine whether the BreakIterator was created in user memory by
//      *   createBufferClone(), and thus should not be deleted.  Such objects
//      *   must be closed by an explicit call to the destructor (not delete).
//      * @deprecated ICU 52. Always delete the BreakIterator.
//      */
//     inline UBool isBufferClone(void);

// #endif /* U_HIDE_DEPRECATED_API */

// #if !UCONFIG_NO_SERVICE
//     /**
//      * Register a new break iterator of the indicated kind, to use in the given locale.
//      * The break iterator will be adopted.  Clones of the iterator will be returned
//      * if a request for a break iterator of the given kind matches or falls back to
//      * this locale.
//      * Because ICU may choose to cache BreakIterators internally, this must
//      * be called at application startup, prior to any calls to
//      * BreakIterator::createXXXInstance to avoid undefined behavior.
//      * @param toAdopt the BreakIterator instance to be adopted
//      * @param locale the Locale for which this instance is to be registered
//      * @param kind the type of iterator for which this instance is to be registered
//      * @param status the in/out status code, no special meanings are assigned
//      * @return a registry key that can be used to unregister this instance
//      * @stable ICU 2.4
//      */
//     static URegistryKey U_EXPORT2 registerInstance(BreakIterator* toAdopt,
//                                         const Locale& locale,
//                                         UBreakIteratorType kind,
//                                         UErrorCode& status);

//     /**
//      * Unregister a previously-registered BreakIterator using the key returned from the
//      * register call.  Key becomes invalid after a successful call and should not be used again.
//      * The BreakIterator corresponding to the key will be deleted.
//      * Because ICU may choose to cache BreakIterators internally, this should
//      * be called during application shutdown, after all calls to
//      * BreakIterator::createXXXInstance to avoid undefined behavior.
//      * @param key the registry key returned by a previous call to registerInstance
//      * @param status the in/out status code, no special meanings are assigned
//      * @return TRUE if the iterator for the key was successfully unregistered
//      * @stable ICU 2.4
//      */
//     static UBool U_EXPORT2 unregister(URegistryKey key, UErrorCode& status);

//     /**
//      * Return a StringEnumeration over the locales available at the time of the call,
//      * including registered locales.
//      * @return a StringEnumeration over the locales available at the time of the call
//      * @stable ICU 2.4
//      */
//     static StringEnumeration* U_EXPORT2 getAvailableLocales(void);
// #endif

//     /**
//      * Returns the locale for this break iterator. Two flavors are available: valid and
//      * actual locale.
//      * @stable ICU 2.8
//      */
//     Locale getLocale(ULocDataLocaleType type, UErrorCode& status) const;

// #ifndef U_HIDE_INTERNAL_API
//     /** Get the locale for this break iterator object. You can choose between valid and actual locale.
//      *  @param type type of the locale we're looking for (valid or actual)
//      *  @param status error code for the operation
//      *  @return the locale
//      *  @internal
//      */
//     const char *getLocaleID(ULocDataLocaleType type, UErrorCode& status) const;
// #endif  /* U_HIDE_INTERNAL_API */

//     /**
//      *  Set the subject text string upon which the break iterator is operating
//      *  without changing any other aspect of the matching state.
//      *  The new and previous text strings must have the same content.
//      *
//      *  This function is intended for use in environments where ICU is operating on
//      *  strings that may move around in memory.  It provides a mechanism for notifying
//      *  ICU that the string has been relocated, and providing a new UText to access the
//      *  string in its new position.
//      *
//      *  Note that the break iterator implementation never copies the underlying text
//      *  of a string being processed, but always operates directly on the original text
//      *  provided by the user. Refreshing simply drops the references to the old text
//      *  and replaces them with references to the new.
//      *
//      *  Caution:  this function is normally used only by very specialized,
//      *  system-level code.  One example use case is with garbage collection that moves
//      *  the text in memory.
//      *
//      * @param input      The new (moved) text string.
//      * @param status     Receives errors detected by this function.
//      * @return           *this
//      *
//      * @stable ICU 49
//      */
//     virtual BreakIterator &refreshInputText(UText *input, UErrorCode &status) = 0;

//  private:
//     static BreakIterator* buildInstance(const Locale& loc, const char *type, UErrorCode& status);
//     static BreakIterator* createInstance(const Locale& loc, int32_t kind, UErrorCode& status);
//     static BreakIterator* makeInstance(const Locale& loc, int32_t kind, UErrorCode& status);

//     friend class ICUBreakIteratorFactory;
//     friend class ICUBreakIteratorService;

// protected:
//     // Do not enclose protected default/copy constructors with #ifndef U_HIDE_INTERNAL_API
//     // or else the compiler will create a public ones.
//     /** @internal */
//     BreakIterator();
//     /** @internal */
//     BreakIterator (const BreakIterator &other);
// #ifndef U_HIDE_INTERNAL_API
//     /** @internal */
//     BreakIterator (const Locale& valid, const Locale &actual);
//     /** @internal. Assignment Operator, used by RuleBasedBreakIterator. */
//     BreakIterator &operator = (const BreakIterator &other);
// #endif  /* U_HIDE_INTERNAL_API */

// private:

//     /** @internal (private) */
//     char actualLocale[ULOC_FULLNAME_CAPACITY];
//     char validLocale[ULOC_FULLNAME_CAPACITY];
// };

// #ifndef U_HIDE_DEPRECATED_API

// inline UBool BreakIterator::isBufferClone()
// {
//     return FALSE;
// }

// #endif /* U_HIDE_DEPRECATED_API */

// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */

// #endif /* U_SHOW_CPLUSPLUS_API */

// #endif // BRKITER_H
// //eof











// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2009-2014 International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/





U_CDECL_BEGIN


#define U_ICUDATA_LANG U_ICUDATA_NAME U_TREE_SEPARATOR_STRING "lang"
#define U_ICUDATA_REGION U_ICUDATA_NAME U_TREE_SEPARATOR_STRING "region"
#define U_ICUDATA_CURR U_ICUDATA_NAME U_TREE_SEPARATOR_STRING "curr"
#define U_ICUDATA_ZONE U_ICUDATA_NAME U_TREE_SEPARATOR_STRING "zone"
#define U_ICUDATA_UNIT U_ICUDATA_NAME U_TREE_SEPARATOR_STRING "unit"

U_CDECL_END




// C++ API ----------------------------------------------------------------- ***

U_NAMESPACE_BEGIN

UnicodeString&
Locale::getDisplayLanguage(UnicodeString& dispLang) const
{
    return this->getDisplayLanguage(getDefault(), dispLang);
}

/*We cannot make any assumptions on the size of the output display strings
* Yet, since we are calling through to a C API, we need to set limits on
* buffer size. For all the following getDisplay functions we first attempt
* to fill up a stack allocated buffer. If it is to small we heap allocated
* the exact buffer we need copy it to the UnicodeString and delete it*/

UnicodeString&
Locale::getDisplayLanguage(const Locale &displayLocale,
                           UnicodeString &result) const {
    UChar *buffer;
    UErrorCode errorCode=U_ZERO_ERROR;
    int32_t length;

    buffer=result.getBuffer(ULOC_FULLNAME_CAPACITY);
    if(buffer==0) {
        result.truncate(0);
        return result;
    }

    length=uloc_getDisplayLanguage(fullName, displayLocale.fullName,
                                   buffer, result.getCapacity(),
                                   &errorCode);
    result.releaseBuffer(U_SUCCESS(errorCode) ? length : 0);

    if(errorCode==U_BUFFER_OVERFLOW_ERROR) {
        buffer=result.getBuffer(length);
        if(buffer==0) {
            result.truncate(0);
            return result;
        }
        errorCode=U_ZERO_ERROR;
        length=uloc_getDisplayLanguage(fullName, displayLocale.fullName,
                                       buffer, result.getCapacity(),
                                       &errorCode);
        result.releaseBuffer(U_SUCCESS(errorCode) ? length : 0);
    }

    return result;
}

UnicodeString&
Locale::getDisplayScript(UnicodeString& dispScript) const
{
    return this->getDisplayScript(getDefault(), dispScript);
}

UnicodeString&
Locale::getDisplayScript(const Locale &displayLocale,
                          UnicodeString &result) const {
    UChar *buffer;
    UErrorCode errorCode=U_ZERO_ERROR;
    int32_t length;

    buffer=result.getBuffer(ULOC_FULLNAME_CAPACITY);
    if(buffer==0) {
        result.truncate(0);
        return result;
    }

    length=uloc_getDisplayScript(fullName, displayLocale.fullName,
                                  buffer, result.getCapacity(),
                                  &errorCode);
    result.releaseBuffer(U_SUCCESS(errorCode) ? length : 0);

    if(errorCode==U_BUFFER_OVERFLOW_ERROR) {
        buffer=result.getBuffer(length);
        if(buffer==0) {
            result.truncate(0);
            return result;
        }
        errorCode=U_ZERO_ERROR;
        length=uloc_getDisplayScript(fullName, displayLocale.fullName,
                                      buffer, result.getCapacity(),
                                      &errorCode);
        result.releaseBuffer(U_SUCCESS(errorCode) ? length : 0);
    }

    return result;
}

UnicodeString&
Locale::getDisplayCountry(UnicodeString& dispCntry) const
{
    return this->getDisplayCountry(getDefault(), dispCntry);
}

UnicodeString&
Locale::getDisplayCountry(const Locale &displayLocale,
                          UnicodeString &result) const {
    UChar *buffer;
    UErrorCode errorCode=U_ZERO_ERROR;
    int32_t length;

    buffer=result.getBuffer(ULOC_FULLNAME_CAPACITY);
    if(buffer==0) {
        result.truncate(0);
        return result;
    }

    length=uloc_getDisplayCountry(fullName, displayLocale.fullName,
                                  buffer, result.getCapacity(),
                                  &errorCode);
    result.releaseBuffer(U_SUCCESS(errorCode) ? length : 0);

    if(errorCode==U_BUFFER_OVERFLOW_ERROR) {
        buffer=result.getBuffer(length);
        if(buffer==0) {
            result.truncate(0);
            return result;
        }
        errorCode=U_ZERO_ERROR;
        length=uloc_getDisplayCountry(fullName, displayLocale.fullName,
                                      buffer, result.getCapacity(),
                                      &errorCode);
        result.releaseBuffer(U_SUCCESS(errorCode) ? length : 0);
    }

    return result;
}

UnicodeString&
Locale::getDisplayVariant(UnicodeString& dispVar) const
{
    return this->getDisplayVariant(getDefault(), dispVar);
}

UnicodeString&
Locale::getDisplayVariant(const Locale &displayLocale,
                          UnicodeString &result) const {
    UChar *buffer;
    UErrorCode errorCode=U_ZERO_ERROR;
    int32_t length;

    buffer=result.getBuffer(ULOC_FULLNAME_CAPACITY);
    if(buffer==0) {
        result.truncate(0);
        return result;
    }

    length=uloc_getDisplayVariant(fullName, displayLocale.fullName,
                                  buffer, result.getCapacity(),
                                  &errorCode);
    result.releaseBuffer(U_SUCCESS(errorCode) ? length : 0);

    if(errorCode==U_BUFFER_OVERFLOW_ERROR) {
        buffer=result.getBuffer(length);
        if(buffer==0) {
            result.truncate(0);
            return result;
        }
        errorCode=U_ZERO_ERROR;
        length=uloc_getDisplayVariant(fullName, displayLocale.fullName,
                                      buffer, result.getCapacity(),
                                      &errorCode);
        result.releaseBuffer(U_SUCCESS(errorCode) ? length : 0);
    }

    return result;
}

UnicodeString&
Locale::getDisplayName( UnicodeString& name ) const
{
    return this->getDisplayName(getDefault(), name);
}

UnicodeString&
Locale::getDisplayName(const Locale &displayLocale,
                       UnicodeString &result) const {
    UChar *buffer;
    UErrorCode errorCode=U_ZERO_ERROR;
    int32_t length;

    buffer=result.getBuffer(ULOC_FULLNAME_CAPACITY);
    if(buffer==0) {
        result.truncate(0);
        return result;
    }

    length=uloc_getDisplayName(fullName, displayLocale.fullName,
                               buffer, result.getCapacity(),
                               &errorCode);
    result.releaseBuffer(U_SUCCESS(errorCode) ? length : 0);

    if(errorCode==U_BUFFER_OVERFLOW_ERROR) {
        buffer=result.getBuffer(length);
        if(buffer==0) {
            result.truncate(0);
            return result;
        }
        errorCode=U_ZERO_ERROR;
        length=uloc_getDisplayName(fullName, displayLocale.fullName,
                                   buffer, result.getCapacity(),
                                   &errorCode);
        result.releaseBuffer(U_SUCCESS(errorCode) ? length : 0);
    }

    return result;
}

#if ! UCONFIG_NO_BREAK_ITERATION

// -------------------------------------
// Gets the objectLocale display name in the default locale language.
UnicodeString& U_EXPORT2
BreakIterator::getDisplayName(const Locale& objectLocale,
                             UnicodeString& name)
{
    return objectLocale.getDisplayName(name);
}

// -------------------------------------
// Gets the objectLocale display name in the displayLocale language.
UnicodeString& U_EXPORT2
BreakIterator::getDisplayName(const Locale& objectLocale,
                             const Locale& displayLocale,
                             UnicodeString& name)
{
    return objectLocale.getDisplayName(displayLocale, name);
}

#endif


U_NAMESPACE_END

// C API ------------------------------------------------------------------- ***

U_NAMESPACE_USE

/* ### Constants **************************************************/

/* These strings describe the resources we attempt to load from
 the locale ResourceBundle data file.*/
static const char _kLanguages[]       = "Languages";
static const char _kScripts[]         = "Scripts";
static const char _kScriptsStandAlone[] = "Scripts%stand-alone";
static const char _kCountries[]       = "Countries";
static const char _kVariants[]        = "Variants";
static const char _kKeys[]            = "Keys";
static const char _kTypes[]           = "Types";
//static const char _kRootName[]        = "root";
static const char _kCurrency[]        = "currency";
static const char _kCurrencies[]      = "Currencies";
static const char _kLocaleDisplayPattern[] = "localeDisplayPattern";
static const char _kPattern[]         = "pattern";
static const char _kSeparator[]       = "separator";

/* ### Display name **************************************************/

static int32_t
_getStringOrCopyKey(const char *path, const char *locale,
                    const char *tableKey, 
                    const char* subTableKey,
                    const char *itemKey,
                    const char *substitute,
                    UChar *dest, int32_t destCapacity,
                    UErrorCode *pErrorCode) {
    const UChar *s = NULL;
    int32_t length = 0;

    if(itemKey==NULL) {
        /* top-level item: normal resource bundle access */
        icu::LocalUResourceBundlePointer rb(ures_open(path, locale, pErrorCode));

        if(U_SUCCESS(*pErrorCode)) {
            s=ures_getStringByKey(rb.getAlias(), tableKey, &length, pErrorCode);
            /* see comment about closing rb near "return item;" in _res_getTableStringWithFallback() */
        }
    } else {
        /* Language code should not be a number. If it is, set the error code. */
        if (!uprv_strncmp(tableKey, "Languages", 9) && uprv_strtol(itemKey, NULL, 10)) {
            *pErrorCode = U_MISSING_RESOURCE_ERROR;
        } else {
            /* second-level item, use special fallback */
            s=uloc_getTableStringWithFallback(path, locale,
                                               tableKey, 
                                               subTableKey,
                                               itemKey,
                                               &length,
                                               pErrorCode);
        }
    }

    if(U_SUCCESS(*pErrorCode)) {
        int32_t copyLength=uprv_min(length, destCapacity);
        if(copyLength>0 && s != NULL) {
            u_memcpy(dest, s, copyLength);
        }
    } else {
        /* no string from a resource bundle: convert the substitute */
        length=(int32_t)uprv_strlen(substitute);
        u_charsToUChars(substitute, dest, uprv_min(length, destCapacity));
        *pErrorCode=U_USING_DEFAULT_WARNING;
    }

    return u_terminateUChars(dest, destCapacity, length, pErrorCode);
}

typedef  int32_t U_CALLCONV UDisplayNameGetter(const char *, char *, int32_t, UErrorCode *);

static int32_t
_getDisplayNameForComponent(const char *locale,
                            const char *displayLocale,
                            UChar *dest, int32_t destCapacity,
                            UDisplayNameGetter *getter,
                            const char *tag,
                            UErrorCode *pErrorCode) {
    char localeBuffer[ULOC_FULLNAME_CAPACITY*4];
    int32_t length;
    UErrorCode localStatus;
    const char* root = NULL;

    /* argument checking */
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    if(destCapacity<0 || (destCapacity>0 && dest==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    localStatus = U_ZERO_ERROR;
    length=(*getter)(locale, localeBuffer, sizeof(localeBuffer), &localStatus);
    if(U_FAILURE(localStatus) || localStatus==U_STRING_NOT_TERMINATED_WARNING) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    if(length==0) {
        // For the display name, we treat this as unknown language (ICU-20273).
        if (getter == uloc_getLanguage) {
            uprv_strcpy(localeBuffer, "und");
        } else {
            return u_terminateUChars(dest, destCapacity, 0, pErrorCode);
        }
    }

    root = tag == _kCountries ? U_ICUDATA_REGION : U_ICUDATA_LANG;

    return _getStringOrCopyKey(root, displayLocale,
                               tag, NULL, localeBuffer,
                               localeBuffer,
                               dest, destCapacity,
                               pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uloc_getDisplayLanguage(const char *locale,
                        const char *displayLocale,
                        UChar *dest, int32_t destCapacity,
                        UErrorCode *pErrorCode) {
    return _getDisplayNameForComponent(locale, displayLocale, dest, destCapacity,
                uloc_getLanguage, _kLanguages, pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uloc_getDisplayScript(const char* locale,
                      const char* displayLocale,
                      UChar *dest, int32_t destCapacity,
                      UErrorCode *pErrorCode)
{
	UErrorCode err = U_ZERO_ERROR;
	int32_t res = _getDisplayNameForComponent(locale, displayLocale, dest, destCapacity,
                uloc_getScript, _kScriptsStandAlone, &err);
	
	if ( err == U_USING_DEFAULT_WARNING ) {
        return _getDisplayNameForComponent(locale, displayLocale, dest, destCapacity,
                    uloc_getScript, _kScripts, pErrorCode);
	} else {
		*pErrorCode = err;
		return res;
	}
}

U_INTERNAL int32_t U_EXPORT2
uloc_getDisplayScriptInContext(const char* locale,
                      const char* displayLocale,
                      UChar *dest, int32_t destCapacity,
                      UErrorCode *pErrorCode)
{
    return _getDisplayNameForComponent(locale, displayLocale, dest, destCapacity,
                    uloc_getScript, _kScripts, pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uloc_getDisplayCountry(const char *locale,
                       const char *displayLocale,
                       UChar *dest, int32_t destCapacity,
                       UErrorCode *pErrorCode) {
    return _getDisplayNameForComponent(locale, displayLocale, dest, destCapacity,
                uloc_getCountry, _kCountries, pErrorCode);
}

/*
 * TODO separate variant1_variant2_variant3...
 * by getting each tag's display string and concatenating them with ", "
 * in between - similar to uloc_getDisplayName()
 */
U_CAPI int32_t U_EXPORT2
uloc_getDisplayVariant(const char *locale,
                       const char *displayLocale,
                       UChar *dest, int32_t destCapacity,
                       UErrorCode *pErrorCode) {
    return _getDisplayNameForComponent(locale, displayLocale, dest, destCapacity,
                uloc_getVariant, _kVariants, pErrorCode);
}

/* Instead of having a separate pass for 'special' patterns, reintegrate the two
 * so we don't get bitten by preflight bugs again.  We can be reasonably efficient
 * without two separate code paths, this code isn't that performance-critical.
 *
 * This code is general enough to deal with patterns that have a prefix or swap the
 * language and remainder components, since we gave developers enough rope to do such
 * things if they futz with the pattern data.  But since we don't give them a way to
 * specify a pattern for arbitrary combinations of components, there's not much use in
 * that.  I don't think our data includes such patterns, the only variable I know if is
 * whether there is a space before the open paren, or not.  Oh, and zh uses different
 * chars than the standard open/close paren (which ja and ko use, btw).
 */
U_CAPI int32_t U_EXPORT2
uloc_getDisplayName(const char *locale,
                    const char *displayLocale,
                    UChar *dest, int32_t destCapacity,
                    UErrorCode *pErrorCode)
{
    static const UChar defaultSeparator[9] = { 0x007b, 0x0030, 0x007d, 0x002c, 0x0020, 0x007b, 0x0031, 0x007d, 0x0000 }; /* "{0}, {1}" */
    static const UChar sub0[4] = { 0x007b, 0x0030, 0x007d , 0x0000 } ; /* {0} */
    static const UChar sub1[4] = { 0x007b, 0x0031, 0x007d , 0x0000 } ; /* {1} */
    static const int32_t subLen = 3;
    static const UChar defaultPattern[10] = {
        0x007b, 0x0030, 0x007d, 0x0020, 0x0028, 0x007b, 0x0031, 0x007d, 0x0029, 0x0000
    }; /* {0} ({1}) */
    static const int32_t defaultPatLen = 9;
    static const int32_t defaultSub0Pos = 0;
    static const int32_t defaultSub1Pos = 5;

    int32_t length; /* of formatted result */

    const UChar *separator;
    int32_t sepLen = 0;
    const UChar *pattern;
    int32_t patLen = 0;
    int32_t sub0Pos, sub1Pos;
    
    UChar formatOpenParen         = 0x0028; // (
    UChar formatReplaceOpenParen  = 0x005B; // [
    UChar formatCloseParen        = 0x0029; // )
    UChar formatReplaceCloseParen = 0x005D; // ]

    UBool haveLang = TRUE; /* assume true, set false if we find we don't have
                              a lang component in the locale */
    UBool haveRest = TRUE; /* assume true, set false if we find we don't have
                              any other component in the locale */
    UBool retry = FALSE; /* set true if we need to retry, see below */

    int32_t langi = 0; /* index of the language substitution (0 or 1), virtually always 0 */

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    if(destCapacity<0 || (destCapacity>0 && dest==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    {
        UErrorCode status = U_ZERO_ERROR;

        icu::LocalUResourceBundlePointer locbundle(
                ures_open(U_ICUDATA_LANG, displayLocale, &status));
        icu::LocalUResourceBundlePointer dspbundle(
                ures_getByKeyWithFallback(locbundle.getAlias(), _kLocaleDisplayPattern, NULL, &status));

        separator=ures_getStringByKeyWithFallback(dspbundle.getAlias(), _kSeparator, &sepLen, &status);
        pattern=ures_getStringByKeyWithFallback(dspbundle.getAlias(), _kPattern, &patLen, &status);
    }

    /* If we couldn't find any data, then use the defaults */
    if(sepLen == 0) {
       separator = defaultSeparator;
    }
    /* #10244: Even though separator is now a pattern, it is awkward to handle it as such
     * here since we are trying to build the display string in place in the dest buffer,
     * and to handle it as a pattern would entail having separate storage for the
     * substrings that need to be combined (the first of which may be the result of
     * previous such combinations). So for now we continue to treat the portion between
     * {0} and {1} as a string to be appended when joining substrings, ignoring anything
     * that is before {0} or after {1} (no existing separator pattern has any such thing).
     * This is similar to how pattern is handled below.
     */
    {
        UChar *p0=u_strstr(separator, sub0);
        UChar *p1=u_strstr(separator, sub1);
        if (p0==NULL || p1==NULL || p1<p0) {
            *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
            return 0;
        }
        separator = (const UChar *)p0 + subLen;
        sepLen = static_cast<int32_t>(p1 - separator);
    }

    if(patLen==0 || (patLen==defaultPatLen && !u_strncmp(pattern, defaultPattern, patLen))) {
        pattern=defaultPattern;
        patLen=defaultPatLen;
        sub0Pos=defaultSub0Pos;
        sub1Pos=defaultSub1Pos;
        // use default formatOpenParen etc. set above
    } else { /* non-default pattern */
        UChar *p0=u_strstr(pattern, sub0);
        UChar *p1=u_strstr(pattern, sub1);
        if (p0==NULL || p1==NULL) {
            *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
            return 0;
        }
        sub0Pos = static_cast<int32_t>(p0-pattern);
        sub1Pos = static_cast<int32_t>(p1-pattern);
        if (sub1Pos < sub0Pos) { /* a very odd pattern */
            int32_t t=sub0Pos; sub0Pos=sub1Pos; sub1Pos=t;
            langi=1;
        }
        if (u_strchr(pattern, 0xFF08) != NULL) {
            formatOpenParen         = 0xFF08; // fullwidth (
            formatReplaceOpenParen  = 0xFF3B; // fullwidth [
            formatCloseParen        = 0xFF09; // fullwidth )
            formatReplaceCloseParen = 0xFF3D; // fullwidth ]
        }
    }

    /* We loop here because there is one case in which after the first pass we could need to
     * reextract the data.  If there's initial padding before the first element, we put in
     * the padding and then write that element.  If it turns out there's no second element,
     * we didn't need the padding.  If we do need the data (no preflight), and the first element
     * would have fit but for the padding, we need to reextract.  In this case (only) we
     * adjust the parameters so padding is not added, and repeat.
     */
    do {
        UChar* p=dest;
        int32_t patPos=0; /* position in the pattern, used for non-substitution portions */
        int32_t langLen=0; /* length of language substitution */
        int32_t langPos=0; /* position in output of language substitution */
        int32_t restLen=0; /* length of 'everything else' substitution */
        int32_t restPos=0; /* position in output of 'everything else' substitution */
        icu::LocalUEnumerationPointer kenum; /* keyword enumeration */

        /* prefix of pattern, extremely likely to be empty */
        if(sub0Pos) {
            if(destCapacity >= sub0Pos) {
                while (patPos < sub0Pos) {
                    *p++ = pattern[patPos++];
                }
            } else {
                patPos=sub0Pos;
            }
            length=sub0Pos;
        } else {
            length=0;
        }

        for(int32_t subi=0,resti=0;subi<2;) { /* iterate through patterns 0 and 1*/
            UBool subdone = FALSE; /* set true when ready to move to next substitution */

            /* prep p and cap for calls to get display components, pin cap to 0 since
               they complain if cap is negative */
            int32_t cap=destCapacity-length;
            if (cap <= 0) {
                cap=0;
            } else {
                p=dest+length;
            }

            if (subi == langi) { /* {0}*/
                if(haveLang) {
                    langPos=length;
                    langLen=uloc_getDisplayLanguage(locale, displayLocale, p, cap, pErrorCode);
                    length+=langLen;
                    haveLang=langLen>0;
                }
                subdone=TRUE;
            } else { /* {1} */
                if(!haveRest) {
                    subdone=TRUE;
                } else {
                    int32_t len; /* length of component (plus other stuff) we just fetched */
                    switch(resti++) {
                        case 0:
                            restPos=length;
                            len=uloc_getDisplayScriptInContext(locale, displayLocale, p, cap, pErrorCode);
                            break;
                        case 1:
                            len=uloc_getDisplayCountry(locale, displayLocale, p, cap, pErrorCode);
                            break;
                        case 2:
                            len=uloc_getDisplayVariant(locale, displayLocale, p, cap, pErrorCode);
                            break;
                        case 3:
                            kenum.adoptInstead(uloc_openKeywords(locale, pErrorCode));
                            U_FALLTHROUGH;
                        default: {
                            const char* kw=uenum_next(kenum.getAlias(), &len, pErrorCode);
                            if (kw == NULL) {
                                len=0; /* mark that we didn't add a component */
                                subdone=TRUE;
                            } else {
                                /* incorporating this behavior into the loop made it even more complex,
                                   so just special case it here */
                                len = uloc_getDisplayKeyword(kw, displayLocale, p, cap, pErrorCode);
                                if(len) {
                                    if(len < cap) {
                                        p[len]=0x3d; /* '=', assume we'll need it */
                                    }
                                    len+=1;

                                    /* adjust for call to get keyword */
                                    cap-=len;
                                    if(cap <= 0) {
                                        cap=0;
                                    } else {
                                        p+=len;
                                    }
                                }
                                /* reset for call below */
                                if(*pErrorCode == U_BUFFER_OVERFLOW_ERROR) {
                                    *pErrorCode=U_ZERO_ERROR;
                                }
                                int32_t vlen = uloc_getDisplayKeywordValue(locale, kw, displayLocale,
                                                                           p, cap, pErrorCode);
                                if(len) {
                                    if(vlen==0) {
                                        --len; /* remove unneeded '=' */
                                    }
                                    /* restore cap and p to what they were at start */
                                    cap=destCapacity-length;
                                    if(cap <= 0) {
                                        cap=0;
                                    } else {
                                        p=dest+length;
                                    }
                                }
                                len+=vlen; /* total we added for key + '=' + value */
                            }
                        } break;
                    } /* end switch */

                    if (len>0) {
                        /* we addeed a component, so add separator and write it if there's room. */
                        if(len+sepLen<=cap) {
                            const UChar * plimit = p + len;
                            for (; p < plimit; p++) {
                                if (*p == formatOpenParen) {
                                    *p = formatReplaceOpenParen;
                                } else if (*p == formatCloseParen) {
                                    *p = formatReplaceCloseParen;
                                }
                            }
                            for(int32_t i=0;i<sepLen;++i) {
                                *p++=separator[i];
                            }
                        }
                        length+=len+sepLen;
                    } else if(subdone) {
                        /* remove separator if we added it */
                        if (length!=restPos) {
                            length-=sepLen;
                        }
                        restLen=length-restPos;
                        haveRest=restLen>0;
                    }
                }
            }

            if(*pErrorCode == U_BUFFER_OVERFLOW_ERROR) {
                *pErrorCode=U_ZERO_ERROR;
            }

            if(subdone) {
                if(haveLang && haveRest) {
                    /* append internal portion of pattern, the first time,
                       or last portion of pattern the second time */
                    int32_t padLen;
                    patPos+=subLen;
                    padLen=(subi==0 ? sub1Pos : patLen)-patPos;
                    if(length+padLen < destCapacity) {
                        p=dest+length;
                        for(int32_t i=0;i<padLen;++i) {
                            *p++=pattern[patPos++];
                        }
                    } else {
                        patPos+=padLen;
                    }
                    length+=padLen;
                } else if(subi==0) {
                    /* don't have first component, reset for second component */
                    sub0Pos=0;
                    length=0;
                } else if(length>0) {
                    /* true length is the length of just the component we got. */
                    length=haveLang?langLen:restLen;
                    if(dest && sub0Pos!=0) {
                        if (sub0Pos+length<=destCapacity) {
                            /* first component not at start of result,
                               but we have full component in buffer. */
                            u_memmove(dest, dest+(haveLang?langPos:restPos), length);
                        } else {
                            /* would have fit, but didn't because of pattern prefix. */
                            sub0Pos=0; /* stops initial padding (and a second retry,
                                          so we won't end up here again) */
                            retry=TRUE;
                        }
                    }
                }

                ++subi; /* move on to next substitution */
            }
        }
    } while(retry);

    return u_terminateUChars(dest, destCapacity, length, pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uloc_getDisplayKeyword(const char* keyword,
                       const char* displayLocale,
                       UChar* dest,
                       int32_t destCapacity,
                       UErrorCode* status){

    /* argument checking */
    if(status==NULL || U_FAILURE(*status)) {
        return 0;
    }

    if(destCapacity<0 || (destCapacity>0 && dest==NULL)) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }


    /* pass itemKey=NULL to look for a top-level item */
    return _getStringOrCopyKey(U_ICUDATA_LANG, displayLocale,
                               _kKeys, NULL, 
                               keyword, 
                               keyword,      
                               dest, destCapacity,
                               status);

}


#define UCURRENCY_DISPLAY_NAME_INDEX 1

U_CAPI int32_t U_EXPORT2
uloc_getDisplayKeywordValue(   const char* locale,
                               const char* keyword,
                               const char* displayLocale,
                               UChar* dest,
                               int32_t destCapacity,
                               UErrorCode* status){


    char keywordValue[ULOC_FULLNAME_CAPACITY*4];
    int32_t capacity = ULOC_FULLNAME_CAPACITY*4;
    int32_t keywordValueLen =0;

    /* argument checking */
    if(status==NULL || U_FAILURE(*status)) {
        return 0;
    }

    if(destCapacity<0 || (destCapacity>0 && dest==NULL)) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* get the keyword value */
    keywordValue[0]=0;
    keywordValueLen = uloc_getKeywordValue(locale, keyword, keywordValue, capacity, status);
    if (*status == U_STRING_NOT_TERMINATED_WARNING)
      *status = U_BUFFER_OVERFLOW_ERROR;

    /* 
     * if the keyword is equal to currency .. then to get the display name 
     * we need to do the fallback ourselves
     */
    if(uprv_stricmp(keyword, _kCurrency)==0){

        int32_t dispNameLen = 0;
        const UChar *dispName = NULL;

        icu::LocalUResourceBundlePointer bundle(
                ures_open(U_ICUDATA_CURR, displayLocale, status));
        icu::LocalUResourceBundlePointer currencies(
                ures_getByKey(bundle.getAlias(), _kCurrencies, NULL, status));
        icu::LocalUResourceBundlePointer currency(
                ures_getByKeyWithFallback(currencies.getAlias(), keywordValue, NULL, status));

        dispName = ures_getStringByIndex(currency.getAlias(), UCURRENCY_DISPLAY_NAME_INDEX, &dispNameLen, status);

        if(U_FAILURE(*status)){
            if(*status == U_MISSING_RESOURCE_ERROR){
                /* we just want to write the value over if nothing is available */
                *status = U_USING_DEFAULT_WARNING;
            }else{
                return 0;
            }
        }

        /* now copy the dispName over if not NULL */
        if(dispName != NULL){
            if(dispNameLen <= destCapacity){
                u_memcpy(dest, dispName, dispNameLen);
                return u_terminateUChars(dest, destCapacity, dispNameLen, status);
            }else{
                *status = U_BUFFER_OVERFLOW_ERROR;
                return dispNameLen;
            }
        }else{
            /* we have not found the display name for the value .. just copy over */
            if(keywordValueLen <= destCapacity){
                u_charsToUChars(keywordValue, dest, keywordValueLen);
                return u_terminateUChars(dest, destCapacity, keywordValueLen, status);
            }else{
                 *status = U_BUFFER_OVERFLOW_ERROR;
                return keywordValueLen;
            }
        }

        
    }else{

        return _getStringOrCopyKey(U_ICUDATA_LANG, displayLocale,
                                   _kTypes, keyword, 
                                   keywordValue,
                                   keywordValue,
                                   dest, destCapacity,
                                   status);
    }
}
// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License

// locdistance.cpp
// created: 2019may08 Markus W. Scherer















U_NAMESPACE_BEGIN

namespace {

/**
 * Bit flag used on the last character of a subtag in the trie.
 * Must be set consistently by the builder and the lookup code.
 */
constexpr int32_t END_OF_SUBTAG = 0x80;
/** Distance value bit flag, set by the builder. */
constexpr int32_t DISTANCE_SKIP_SCRIPT = 0x80;
/** Distance value bit flag, set by trieNext(). */
constexpr int32_t DISTANCE_IS_FINAL = 0x100;
constexpr int32_t DISTANCE_IS_FINAL_OR_SKIP_SCRIPT = DISTANCE_IS_FINAL | DISTANCE_SKIP_SCRIPT;

constexpr int32_t ABOVE_THRESHOLD = 100;

// Indexes into array of distances.
enum {
    IX_DEF_LANG_DISTANCE,
    IX_DEF_SCRIPT_DISTANCE,
    IX_DEF_REGION_DISTANCE,
    IX_MIN_REGION_DISTANCE,
    IX_LIMIT
};

LocaleDistance *gLocaleDistance = nullptr;
UInitOnce locdistance_gInitOnce = U_INITONCE_INITIALIZER;

UBool U_CALLCONV locdistance_cleanup() {
    delete gLocaleDistance;
    gLocaleDistance = nullptr;
    locdistance_gInitOnce.reset();
    return TRUE;
}

}  // namespace

void U_CALLCONV LocaleDistance::initLocaleDistance(UErrorCode &errorCode) {
    // This function is invoked only via umtx_initOnce().
    U_ASSERT(gLocaleDistance == nullptr);
    const XLikelySubtags &likely = *XLikelySubtags::getSingleton(errorCode);
    if (U_FAILURE(errorCode)) { return; }
    const LocaleDistanceData &data = likely.getDistanceData();
    if (data.distanceTrieBytes == nullptr ||
            data.regionToPartitions == nullptr || data.partitions == nullptr ||
            // ok if no paradigms
            data.distances == nullptr) {
        errorCode = U_MISSING_RESOURCE_ERROR;
        return;
    }
    gLocaleDistance = new LocaleDistance(data);
    if (gLocaleDistance == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    ucln_common_registerCleanup(UCLN_COMMON_LOCALE_DISTANCE, locdistance_cleanup);
}

const LocaleDistance *LocaleDistance::getSingleton(UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return nullptr; }
    umtx_initOnce(locdistance_gInitOnce, &LocaleDistance::initLocaleDistance, errorCode);
    return gLocaleDistance;
}

LocaleDistance::LocaleDistance(const LocaleDistanceData &data) :
        trie(data.distanceTrieBytes),
        regionToPartitionsIndex(data.regionToPartitions), partitionArrays(data.partitions),
        paradigmLSRs(data.paradigms), paradigmLSRsLength(data.paradigmsLength),
        defaultLanguageDistance(data.distances[IX_DEF_LANG_DISTANCE]),
        defaultScriptDistance(data.distances[IX_DEF_SCRIPT_DISTANCE]),
        defaultRegionDistance(data.distances[IX_DEF_REGION_DISTANCE]),
        minRegionDistance(data.distances[IX_MIN_REGION_DISTANCE]) {
    // For the default demotion value, use the
    // default region distance between unrelated Englishes.
    // Thus, unless demotion is turned off,
    // a mere region difference for one desired locale
    // is as good as a perfect match for the next following desired locale.
    // As of CLDR 36, we have <languageMatch desired="en_*_*" supported="en_*_*" distance="5"/>.
    LSR en("en", "Latn", "US");
    LSR enGB("en", "Latn", "GB");
    const LSR *p_enGB = &enGB;
    defaultDemotionPerDesiredLocale = getBestIndexAndDistance(en, &p_enGB, 1,
            50, ULOCMATCH_FAVOR_LANGUAGE) & 0xff;
}

int32_t LocaleDistance::getBestIndexAndDistance(
        const LSR &desired,
        const LSR **supportedLSRs, int32_t supportedLSRsLength,
        int32_t threshold, ULocMatchFavorSubtag favorSubtag) const {
    BytesTrie iter(trie);
    // Look up the desired language only once for all supported LSRs.
    // Its "distance" is either a match point value of 0, or a non-match negative value.
    // Note: The data builder verifies that there are no <*, supported> or <desired, *> rules.
    int32_t desLangDistance = trieNext(iter, desired.language, false);
    uint64_t desLangState = desLangDistance >= 0 && supportedLSRsLength > 1 ? iter.getState64() : 0;
    // Index of the supported LSR with the lowest distance.
    int32_t bestIndex = -1;
    for (int32_t slIndex = 0; slIndex < supportedLSRsLength; ++slIndex) {
        const LSR &supported = *supportedLSRs[slIndex];
        bool star = false;
        int32_t distance = desLangDistance;
        if (distance >= 0) {
            U_ASSERT((distance & DISTANCE_IS_FINAL) == 0);
            if (slIndex != 0) {
                iter.resetToState64(desLangState);
            }
            distance = trieNext(iter, supported.language, true);
        }
        // Note: The data builder verifies that there are no rules with "any" (*) language and
        // real (non *) script or region subtags.
        // This means that if the lookup for either language fails we can use
        // the default distances without further lookups.
        int32_t flags;
        if (distance >= 0) {
            flags = distance & DISTANCE_IS_FINAL_OR_SKIP_SCRIPT;
            distance &= ~DISTANCE_IS_FINAL_OR_SKIP_SCRIPT;
        } else {  // <*, *>
            if (uprv_strcmp(desired.language, supported.language) == 0) {
                distance = 0;
            } else {
                distance = defaultLanguageDistance;
            }
            flags = 0;
            star = true;
        }
        U_ASSERT(0 <= distance && distance <= 100);
        // We implement "favor subtag" by reducing the language subtag distance
        // (unscientifically reducing it to a quarter of the normal value),
        // so that the script distance is relatively more important.
        // For example, given a default language distance of 80, we reduce it to 20,
        // which is below the default threshold of 50, which is the default script distance.
        if (favorSubtag == ULOCMATCH_FAVOR_SCRIPT) {
            distance >>= 2;
        }
        if (distance >= threshold) {
            continue;
        }

        int32_t scriptDistance;
        if (star || flags != 0) {
            if (uprv_strcmp(desired.script, supported.script) == 0) {
                scriptDistance = 0;
            } else {
                scriptDistance = defaultScriptDistance;
            }
        } else {
            scriptDistance = getDesSuppScriptDistance(iter, iter.getState64(),
                    desired.script, supported.script);
            flags = scriptDistance & DISTANCE_IS_FINAL;
            scriptDistance &= ~DISTANCE_IS_FINAL;
        }
        distance += scriptDistance;
        if (distance >= threshold) {
            continue;
        }

        if (uprv_strcmp(desired.region, supported.region) == 0) {
            // regionDistance = 0
        } else if (star || (flags & DISTANCE_IS_FINAL) != 0) {
            distance += defaultRegionDistance;
        } else {
            int32_t remainingThreshold = threshold - distance;
            if (minRegionDistance >= remainingThreshold) {
                continue;
            }

            // From here on we know the regions are not equal.
            // Map each region to zero or more partitions. (zero = one non-matching string)
            // (Each array of single-character partition strings is encoded as one string.)
            // If either side has more than one, then we find the maximum distance.
            // This could be optimized by adding some more structure, but probably not worth it.
            distance += getRegionPartitionsDistance(
                    iter, iter.getState64(),
                    partitionsForRegion(desired),
                    partitionsForRegion(supported),
                    remainingThreshold);
        }
        if (distance < threshold) {
            if (distance == 0) {
                return slIndex << 8;
            }
            bestIndex = slIndex;
            threshold = distance;
        }
    }
    return bestIndex >= 0 ? (bestIndex << 8) | threshold : 0xffffff00 | ABOVE_THRESHOLD;
}

int32_t LocaleDistance::getDesSuppScriptDistance(
        BytesTrie &iter, uint64_t startState, const char *desired, const char *supported) {
    // Note: The data builder verifies that there are no <*, supported> or <desired, *> rules.
    int32_t distance = trieNext(iter, desired, false);
    if (distance >= 0) {
        distance = trieNext(iter, supported, true);
    }
    if (distance < 0) {
        UStringTrieResult result = iter.resetToState64(startState).next(u'*');  // <*, *>
        U_ASSERT(USTRINGTRIE_HAS_VALUE(result));
        if (uprv_strcmp(desired, supported) == 0) {
            distance = 0;  // same script
        } else {
            distance = iter.getValue();
            U_ASSERT(distance >= 0);
        }
        if (result == USTRINGTRIE_FINAL_VALUE) {
            distance |= DISTANCE_IS_FINAL;
        }
    }
    return distance;
}

int32_t LocaleDistance::getRegionPartitionsDistance(
        BytesTrie &iter, uint64_t startState,
        const char *desiredPartitions, const char *supportedPartitions, int32_t threshold) {
    char desired = *desiredPartitions++;
    char supported = *supportedPartitions++;
    U_ASSERT(desired != 0 && supported != 0);
    // See if we have single desired/supported partitions, from NUL-terminated
    // partition strings without explicit length.
    bool suppLengthGt1 = *supportedPartitions != 0;  // gt1: more than 1 character
    // equivalent to: if (desLength == 1 && suppLength == 1)
    if (*desiredPartitions == 0 && !suppLengthGt1) {
        // Fastpath for single desired/supported partitions.
        UStringTrieResult result = iter.next(uprv_invCharToAscii(desired) | END_OF_SUBTAG);
        if (USTRINGTRIE_HAS_NEXT(result)) {
            result = iter.next(uprv_invCharToAscii(supported) | END_OF_SUBTAG);
            if (USTRINGTRIE_HAS_VALUE(result)) {
                return iter.getValue();
            }
        }
        return getFallbackRegionDistance(iter, startState);
    }

    const char *supportedStart = supportedPartitions - 1;  // for restart of inner loop
    int32_t regionDistance = 0;
    // Fall back to * only once, not for each pair of partition strings.
    bool star = false;
    for (;;) {
        // Look up each desired-partition string only once,
        // not for each (desired, supported) pair.
        UStringTrieResult result = iter.next(uprv_invCharToAscii(desired) | END_OF_SUBTAG);
        if (USTRINGTRIE_HAS_NEXT(result)) {
            uint64_t desState = suppLengthGt1 ? iter.getState64() : 0;
            for (;;) {
                result = iter.next(uprv_invCharToAscii(supported) | END_OF_SUBTAG);
                int32_t d;
                if (USTRINGTRIE_HAS_VALUE(result)) {
                    d = iter.getValue();
                } else if (star) {
                    d = 0;
                } else {
                    d = getFallbackRegionDistance(iter, startState);
                    star = true;
                }
                if (d >= threshold) {
                    return d;
                } else if (regionDistance < d) {
                    regionDistance = d;
                }
                if ((supported = *supportedPartitions++) != 0) {
                    iter.resetToState64(desState);
                } else {
                    break;
                }
            }
        } else if (!star) {
            int32_t d = getFallbackRegionDistance(iter, startState);
            if (d >= threshold) {
                return d;
            } else if (regionDistance < d) {
                regionDistance = d;
            }
            star = true;
        }
        if ((desired = *desiredPartitions++) != 0) {
            iter.resetToState64(startState);
            supportedPartitions = supportedStart;
            supported = *supportedPartitions++;
        } else {
            break;
        }
    }
    return regionDistance;
}

int32_t LocaleDistance::getFallbackRegionDistance(BytesTrie &iter, uint64_t startState) {
#if U_DEBUG
    UStringTrieResult result =
#endif
    iter.resetToState64(startState).next(u'*');  // <*, *>
    U_ASSERT(USTRINGTRIE_HAS_VALUE(result));
    int32_t distance = iter.getValue();
    U_ASSERT(distance >= 0);
    return distance;
}

int32_t LocaleDistance::trieNext(BytesTrie &iter, const char *s, bool wantValue) {
    uint8_t c;
    if ((c = *s) == 0) {
        return -1;  // no empty subtags in the distance data
    }
    for (;;) {
        c = uprv_invCharToAscii(c);
        // EBCDIC: If *s is not an invariant character,
        // then c is now 0 and will simply not match anything, which is harmless.
        uint8_t next = *++s;
        if (next != 0) {
            if (!USTRINGTRIE_HAS_NEXT(iter.next(c))) {
                return -1;
            }
        } else {
            // last character of this subtag
            UStringTrieResult result = iter.next(c | END_OF_SUBTAG);
            if (wantValue) {
                if (USTRINGTRIE_HAS_VALUE(result)) {
                    int32_t value = iter.getValue();
                    if (result == USTRINGTRIE_FINAL_VALUE) {
                        value |= DISTANCE_IS_FINAL;
                    }
                    return value;
                }
            } else {
                if (USTRINGTRIE_HAS_NEXT(result)) {
                    return 0;
                }
            }
            return -1;
        }
        c = next;
    }
}

UBool LocaleDistance::isParadigmLSR(const LSR &lsr) const {
    // Linear search for a very short list (length 6 as of 2019).
    // If there are many paradigm LSRs we should use a hash set.
    U_ASSERT(paradigmLSRsLength <= 15);
    for (int32_t i = 0; i < paradigmLSRsLength; ++i) {
        if (lsr == paradigmLSRs[i]) { return true; }
    }
    return false;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2010-2016, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Provides display names of Locale and its components.
 */

#if !UCONFIG_NO_FORMATTING




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2016, International Business Machines Corporation and
*   others.  All Rights Reserved.
*******************************************************************************
*/


/**
 * \file
 * \brief C API: Provides display names of Locale ids and their components.
 */




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2014-2016, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/




#if !UCONFIG_NO_FORMATTING

/**
 * \file
 * \brief C API: Display context types (enum values)
 */

/**
 * Display context types, for getting values of a particular setting.
 * Note, the specific numeric values are internal and may change.
 * @stable ICU 51
 */
enum UDisplayContextType {
    /**
     * Type to retrieve the dialect handling setting, e.g.
     * UDISPCTX_STANDARD_NAMES or UDISPCTX_DIALECT_NAMES.
     * @stable ICU 51
     */
    UDISPCTX_TYPE_DIALECT_HANDLING = 0,
    /**
     * Type to retrieve the capitalization context setting, e.g.
     * UDISPCTX_CAPITALIZATION_NONE, UDISPCTX_CAPITALIZATION_FOR_MIDDLE_OF_SENTENCE,
     * UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE, etc.
     * @stable ICU 51
     */
    UDISPCTX_TYPE_CAPITALIZATION = 1,
    /**
     * Type to retrieve the display length setting, e.g.
     * UDISPCTX_LENGTH_FULL, UDISPCTX_LENGTH_SHORT.
     * @stable ICU 54
     */
    UDISPCTX_TYPE_DISPLAY_LENGTH = 2,
    /**
     * Type to retrieve the substitute handling setting, e.g.
     * UDISPCTX_SUBSTITUTE, UDISPCTX_NO_SUBSTITUTE.
     * @stable ICU 58
     */
    UDISPCTX_TYPE_SUBSTITUTE_HANDLING = 3
};
/**
*  @stable ICU 51
*/
typedef enum UDisplayContextType UDisplayContextType;

/**
 * Display context settings.
 * Note, the specific numeric values are internal and may change.
 * @stable ICU 51
 */
enum UDisplayContext {
    /**
     * ================================
     * DIALECT_HANDLING can be set to one of UDISPCTX_STANDARD_NAMES or
     * UDISPCTX_DIALECT_NAMES. Use UDisplayContextType UDISPCTX_TYPE_DIALECT_HANDLING
     * to get the value.
     */
    /**
     * A possible setting for DIALECT_HANDLING:
     * use standard names when generating a locale name,
     * e.g. en_GB displays as 'English (United Kingdom)'.
     * @stable ICU 51
     */
    UDISPCTX_STANDARD_NAMES = (UDISPCTX_TYPE_DIALECT_HANDLING<<8) + 0,
    /**
     * A possible setting for DIALECT_HANDLING:
     * use dialect names, when generating a locale name,
     * e.g. en_GB displays as 'British English'.
     * @stable ICU 51
     */
    UDISPCTX_DIALECT_NAMES = (UDISPCTX_TYPE_DIALECT_HANDLING<<8) + 1,
    /**
     * ================================
     * CAPITALIZATION can be set to one of UDISPCTX_CAPITALIZATION_NONE,
     * UDISPCTX_CAPITALIZATION_FOR_MIDDLE_OF_SENTENCE,
     * UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE,
     * UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU, or
     * UDISPCTX_CAPITALIZATION_FOR_STANDALONE.
     * Use UDisplayContextType UDISPCTX_TYPE_CAPITALIZATION to get the value.
     */
    /**
     * The capitalization context to be used is unknown (this is the default value).
     * @stable ICU 51
     */
    UDISPCTX_CAPITALIZATION_NONE = (UDISPCTX_TYPE_CAPITALIZATION<<8) + 0,
    /**
     * The capitalization context if a date, date symbol or display name is to be
     * formatted with capitalization appropriate for the middle of a sentence.
     * @stable ICU 51
     */
    UDISPCTX_CAPITALIZATION_FOR_MIDDLE_OF_SENTENCE = (UDISPCTX_TYPE_CAPITALIZATION<<8) + 1,
    /**
     * The capitalization context if a date, date symbol or display name is to be
     * formatted with capitalization appropriate for the beginning of a sentence.
     * @stable ICU 51
     */
    UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE = (UDISPCTX_TYPE_CAPITALIZATION<<8) + 2,
    /**
     * The capitalization context if a date, date symbol or display name is to be
     * formatted with capitalization appropriate for a user-interface list or menu item.
     * @stable ICU 51
     */
    UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU = (UDISPCTX_TYPE_CAPITALIZATION<<8) + 3,
    /**
     * The capitalization context if a date, date symbol or display name is to be
     * formatted with capitalization appropriate for stand-alone usage such as an
     * isolated name on a calendar page.
     * @stable ICU 51
     */
    UDISPCTX_CAPITALIZATION_FOR_STANDALONE = (UDISPCTX_TYPE_CAPITALIZATION<<8) + 4,
    /**
     * ================================
     * DISPLAY_LENGTH can be set to one of UDISPCTX_LENGTH_FULL or
     * UDISPCTX_LENGTH_SHORT. Use UDisplayContextType UDISPCTX_TYPE_DISPLAY_LENGTH
     * to get the value.
     */
    /**
     * A possible setting for DISPLAY_LENGTH:
     * use full names when generating a locale name,
     * e.g. "United States" for US.
     * @stable ICU 54
     */
    UDISPCTX_LENGTH_FULL = (UDISPCTX_TYPE_DISPLAY_LENGTH<<8) + 0,
    /**
     * A possible setting for DISPLAY_LENGTH:
     * use short names when generating a locale name,
     * e.g. "U.S." for US.
     * @stable ICU 54
     */
    UDISPCTX_LENGTH_SHORT = (UDISPCTX_TYPE_DISPLAY_LENGTH<<8) + 1,
    /**
     * ================================
     * SUBSTITUTE_HANDLING can be set to one of UDISPCTX_SUBSTITUTE or
     * UDISPCTX_NO_SUBSTITUTE. Use UDisplayContextType UDISPCTX_TYPE_SUBSTITUTE_HANDLING
     * to get the value.
     */
    /**
     * A possible setting for SUBSTITUTE_HANDLING:
     * Returns a fallback value (e.g., the input code) when no data is available.
     * This is the default value.
     * @stable ICU 58
     */
    UDISPCTX_SUBSTITUTE = (UDISPCTX_TYPE_SUBSTITUTE_HANDLING<<8) + 0,
    /**
     * A possible setting for SUBSTITUTE_HANDLING:
     * Returns a null value when no data is available.
     * @stable ICU 58
     */
    UDISPCTX_NO_SUBSTITUTE = (UDISPCTX_TYPE_SUBSTITUTE_HANDLING<<8) + 1

};
/**
*  @stable ICU 51
*/
typedef enum UDisplayContext UDisplayContext;

#endif /* #if !UCONFIG_NO_FORMATTING */



/**
 * Enum used in LocaleDisplayNames::createInstance.
 * @stable ICU 4.4
 */
typedef enum {
    /**
     * Use standard names when generating a locale name,
     * e.g. en_GB displays as 'English (United Kingdom)'.
     * @stable ICU 4.4
     */
    ULDN_STANDARD_NAMES = 0,
    /**
     * Use dialect names, when generating a locale name,
     * e.g. en_GB displays as 'British English'.
     * @stable ICU 4.4
     */
    ULDN_DIALECT_NAMES
} UDialectHandling;

/**
 * Opaque C service object type for the locale display names API
 * @stable ICU 4.4
 */
struct ULocaleDisplayNames;

/** 
 * C typedef for struct ULocaleDisplayNames. 
 * @stable ICU 4.4 
 */
typedef struct ULocaleDisplayNames ULocaleDisplayNames;  

#if !UCONFIG_NO_FORMATTING

/**
 * Returns an instance of LocaleDisplayNames that returns names
 * formatted for the provided locale, using the provided
 * dialectHandling.  The usual value for dialectHandling is
 * ULOC_STANDARD_NAMES.
 *
 * @param locale the display locale 
 * @param dialectHandling how to select names for locales 
 * @return a ULocaleDisplayNames instance 
 * @param pErrorCode the status code
 * @stable ICU 4.4
 */
U_STABLE ULocaleDisplayNames * U_EXPORT2
uldn_open(const char * locale,
          UDialectHandling dialectHandling,
          UErrorCode *pErrorCode);

/**
 * Closes a ULocaleDisplayNames instance obtained from uldn_open().
 * @param ldn the ULocaleDisplayNames instance to be closed
 * @stable ICU 4.4
 */
U_STABLE void U_EXPORT2
uldn_close(ULocaleDisplayNames *ldn);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalULocaleDisplayNamesPointer
 * "Smart pointer" class, closes a ULocaleDisplayNames via uldn_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalULocaleDisplayNamesPointer, ULocaleDisplayNames, uldn_close);

U_NAMESPACE_END

#endif

/* getters for state */

/**
 * Returns the locale used to determine the display names. This is
 * not necessarily the same locale passed to {@link #uldn_open}.
 * @param ldn the LocaleDisplayNames instance
 * @return the display locale 
 * @stable ICU 4.4
 */
U_STABLE const char * U_EXPORT2
uldn_getLocale(const ULocaleDisplayNames *ldn);

/**
 * Returns the dialect handling used in the display names.
 * @param ldn the LocaleDisplayNames instance
 * @return the dialect handling enum
 * @stable ICU 4.4
 */
U_STABLE UDialectHandling U_EXPORT2
uldn_getDialectHandling(const ULocaleDisplayNames *ldn);

/* names for entire locales */

/**
 * Returns the display name of the provided locale.
 * @param ldn the LocaleDisplayNames instance
 * @param locale the locale whose display name to return
 * @param result receives the display name
 * @param maxResultSize the size of the result buffer
 * @param pErrorCode the status code
 * @return the actual buffer size needed for the display name.  If it's
 * greater than maxResultSize, the returned name will be truncated.
 * @stable ICU 4.4
 */
U_STABLE int32_t U_EXPORT2
uldn_localeDisplayName(const ULocaleDisplayNames *ldn,
                       const char *locale,
                       UChar *result,
                       int32_t maxResultSize,
                       UErrorCode *pErrorCode);

/* names for components of a locale */

/**
 * Returns the display name of the provided language code.
 * @param ldn the LocaleDisplayNames instance
 * @param lang the language code whose display name to return
 * @param result receives the display name
 * @param maxResultSize the size of the result buffer
 * @param pErrorCode the status code
 * @return the actual buffer size needed for the display name.  If it's
 * greater than maxResultSize, the returned name will be truncated.
 * @stable ICU 4.4
 */
U_STABLE int32_t U_EXPORT2
uldn_languageDisplayName(const ULocaleDisplayNames *ldn,
                         const char *lang,
                         UChar *result,
                         int32_t maxResultSize,
                         UErrorCode *pErrorCode);

/**
 * Returns the display name of the provided script.
 * @param ldn the LocaleDisplayNames instance
 * @param script the script whose display name to return
 * @param result receives the display name
 * @param maxResultSize the size of the result buffer
 * @param pErrorCode the status code
 * @return the actual buffer size needed for the display name.  If it's
 * greater than maxResultSize, the returned name will be truncated.
 * @stable ICU 4.4
 */
U_STABLE int32_t U_EXPORT2
uldn_scriptDisplayName(const ULocaleDisplayNames *ldn,
                       const char *script,
                       UChar *result,
                       int32_t maxResultSize,
                       UErrorCode *pErrorCode);

/**
 * Returns the display name of the provided script code.
 * @param ldn the LocaleDisplayNames instance
 * @param scriptCode the script code whose display name to return
 * @param result receives the display name
 * @param maxResultSize the size of the result buffer
 * @param pErrorCode the status code
 * @return the actual buffer size needed for the display name.  If it's
 * greater than maxResultSize, the returned name will be truncated.
 * @stable ICU 4.4
 */
U_STABLE int32_t U_EXPORT2
uldn_scriptCodeDisplayName(const ULocaleDisplayNames *ldn,
                           UScriptCode scriptCode,
                           UChar *result,
                           int32_t maxResultSize,
                           UErrorCode *pErrorCode);

/**
 * Returns the display name of the provided region code.
 * @param ldn the LocaleDisplayNames instance
 * @param region the region code whose display name to return
 * @param result receives the display name
 * @param maxResultSize the size of the result buffer
 * @param pErrorCode the status code
 * @return the actual buffer size needed for the display name.  If it's
 * greater than maxResultSize, the returned name will be truncated.
 * @stable ICU 4.4
 */
U_STABLE int32_t U_EXPORT2
uldn_regionDisplayName(const ULocaleDisplayNames *ldn,
                       const char *region,
                       UChar *result,
                       int32_t maxResultSize,
                       UErrorCode *pErrorCode);

/**
 * Returns the display name of the provided variant
 * @param ldn the LocaleDisplayNames instance
 * @param variant the variant whose display name to return
 * @param result receives the display name
 * @param maxResultSize the size of the result buffer
 * @param pErrorCode the status code
 * @return the actual buffer size needed for the display name.  If it's
 * greater than maxResultSize, the returned name will be truncated.
 * @stable ICU 4.4
 */
U_STABLE int32_t U_EXPORT2
uldn_variantDisplayName(const ULocaleDisplayNames *ldn,
                        const char *variant,
                        UChar *result,
                        int32_t maxResultSize,
                        UErrorCode *pErrorCode);

/**
 * Returns the display name of the provided locale key
 * @param ldn the LocaleDisplayNames instance
 * @param key the locale key whose display name to return
 * @param result receives the display name
 * @param maxResultSize the size of the result buffer
 * @param pErrorCode the status code
 * @return the actual buffer size needed for the display name.  If it's
 * greater than maxResultSize, the returned name will be truncated.
 * @stable ICU 4.4
 */
U_STABLE int32_t U_EXPORT2
uldn_keyDisplayName(const ULocaleDisplayNames *ldn,
                    const char *key,
                    UChar *result,
                    int32_t maxResultSize,
                    UErrorCode *pErrorCode);

/**
 * Returns the display name of the provided value (used with the provided key).
 * @param ldn the LocaleDisplayNames instance
 * @param key the locale key
 * @param value the locale key's value
 * @param result receives the display name
 * @param maxResultSize the size of the result buffer
 * @param pErrorCode the status code
 * @return the actual buffer size needed for the display name.  If it's
 * greater than maxResultSize, the returned name will be truncated.
 * @stable ICU 4.4
 */
U_STABLE int32_t U_EXPORT2
uldn_keyValueDisplayName(const ULocaleDisplayNames *ldn,
                         const char *key,
                         const char *value,
                         UChar *result,
                         int32_t maxResultSize,
                         UErrorCode *pErrorCode);

/**
* Returns an instance of LocaleDisplayNames that returns names formatted
* for the provided locale, using the provided UDisplayContext settings.
*
* @param locale The display locale 
* @param contexts List of one or more context settings (e.g. for dialect
*               handling, capitalization, etc.
* @param length Number of items in the contexts list
* @param pErrorCode Pointer to UErrorCode input/output status. If at entry this indicates
*               a failure status, the function will do nothing; otherwise this will be
*               updated with any new status from the function. 
* @return a ULocaleDisplayNames instance 
* @stable ICU 51
*/
U_STABLE ULocaleDisplayNames * U_EXPORT2
uldn_openForContext(const char * locale, UDisplayContext *contexts,
                    int32_t length, UErrorCode *pErrorCode);

/**
* Returns the UDisplayContext value for the specified UDisplayContextType.
* @param ldn the ULocaleDisplayNames instance
* @param type the UDisplayContextType whose value to return
* @param pErrorCode Pointer to UErrorCode input/output status. If at entry this indicates
*               a failure status, the function will do nothing; otherwise this will be
*               updated with any new status from the function. 
* @return the UDisplayContextValue for the specified type.
* @stable ICU 51
*/
U_STABLE UDisplayContext U_EXPORT2
uldn_getContext(const ULocaleDisplayNames *ldn, UDisplayContextType type,
                UErrorCode *pErrorCode);

#endif  /* !UCONFIG_NO_FORMATTING */



U_NAMESPACE_BEGIN

/**
 * Returns display names of Locales and components of Locales. For
 * more information on language, script, region, variant, key, and
 * values, see Locale.
 * @stable ICU 4.4
 */
class U_COMMON_API LocaleDisplayNames : public UObject {
public:
    /**
     * Destructor.
     * @stable ICU 4.4
     */
    virtual ~LocaleDisplayNames();

    /**
     * Convenience overload of
     * {@link #createInstance(const Locale& locale, UDialectHandling dialectHandling)}
     * that specifies STANDARD dialect handling.
     * @param locale the display locale
     * @return a LocaleDisplayNames instance
     * @stable ICU 4.4
     */
    inline static LocaleDisplayNames* U_EXPORT2 createInstance(const Locale& locale);

    /**
     * Returns an instance of LocaleDisplayNames that returns names
     * formatted for the provided locale, using the provided
     * dialectHandling.
     *
     * @param locale the display locale
     * @param dialectHandling how to select names for locales
     * @return a LocaleDisplayNames instance
     * @stable ICU 4.4
     */
    static LocaleDisplayNames* U_EXPORT2 createInstance(const Locale& locale,
                            UDialectHandling dialectHandling);

    /**
     * Returns an instance of LocaleDisplayNames that returns names formatted
     * for the provided locale, using the provided UDisplayContext settings.
     *
     * @param locale the display locale
     * @param contexts List of one or more context settings (e.g. for dialect
     *               handling, capitalization, etc.
     * @param length Number of items in the contexts list
     * @return a LocaleDisplayNames instance
     * @stable ICU 51
     */
    static LocaleDisplayNames* U_EXPORT2 createInstance(const Locale& locale,
                            UDisplayContext *contexts, int32_t length);

    // getters for state
    /**
     * Returns the locale used to determine the display names. This is
     * not necessarily the same locale passed to {@link #createInstance}.
     * @return the display locale
     * @stable ICU 4.4
     */
    virtual const Locale& getLocale() const = 0;

    /**
     * Returns the dialect handling used in the display names.
     * @return the dialect handling enum
     * @stable ICU 4.4
     */
    virtual UDialectHandling getDialectHandling() const = 0;

    /**
     * Returns the UDisplayContext value for the specified UDisplayContextType.
     * @param type the UDisplayContextType whose value to return
     * @return the UDisplayContext for the specified type.
     * @stable ICU 51
     */
    virtual UDisplayContext getContext(UDisplayContextType type) const = 0;

    // names for entire locales
    /**
     * Returns the display name of the provided locale.
     * @param locale the locale whose display name to return
     * @param result receives the locale's display name
     * @return the display name of the provided locale
     * @stable ICU 4.4
     */
    virtual UnicodeString& localeDisplayName(const Locale& locale,
                         UnicodeString& result) const = 0;

    /**
     * Returns the display name of the provided locale id.
     * @param localeId the id of the locale whose display name to return
     * @param result receives the locale's display name
     * @return the display name of the provided locale
     * @stable ICU 4.4
     */
    virtual UnicodeString& localeDisplayName(const char* localeId,
                         UnicodeString& result) const = 0;

    // names for components of a locale id
    /**
     * Returns the display name of the provided language code.
     * @param lang the language code
     * @param result receives the language code's display name
     * @return the display name of the provided language code
     * @stable ICU 4.4
     */
    virtual UnicodeString& languageDisplayName(const char* lang,
                           UnicodeString& result) const = 0;

    /**
     * Returns the display name of the provided script code.
     * @param script the script code
     * @param result receives the script code's display name
     * @return the display name of the provided script code
     * @stable ICU 4.4
     */
    virtual UnicodeString& scriptDisplayName(const char* script,
                         UnicodeString& result) const = 0;

    /**
     * Returns the display name of the provided script code.
     * @param scriptCode the script code number
     * @param result receives the script code's display name
     * @return the display name of the provided script code
     * @stable ICU 4.4
     */
    virtual UnicodeString& scriptDisplayName(UScriptCode scriptCode,
                         UnicodeString& result) const = 0;

    /**
     * Returns the display name of the provided region code.
     * @param region the region code
     * @param result receives the region code's display name
     * @return the display name of the provided region code
     * @stable ICU 4.4
     */
    virtual UnicodeString& regionDisplayName(const char* region,
                         UnicodeString& result) const = 0;

    /**
     * Returns the display name of the provided variant.
     * @param variant the variant string
     * @param result receives the variant's display name
     * @return the display name of the provided variant
     * @stable ICU 4.4
     */
    virtual UnicodeString& variantDisplayName(const char* variant,
                          UnicodeString& result) const = 0;

    /**
     * Returns the display name of the provided locale key.
     * @param key the locale key name
     * @param result receives the locale key's display name
     * @return the display name of the provided locale key
     * @stable ICU 4.4
     */
    virtual UnicodeString& keyDisplayName(const char* key,
                      UnicodeString& result) const = 0;

    /**
     * Returns the display name of the provided value (used with the provided key).
     * @param key the locale key name
     * @param value the locale key's value
     * @param result receives the value's display name
     * @return the display name of the provided value
     * @stable ICU 4.4
     */
    virtual UnicodeString& keyValueDisplayName(const char* key, const char* value,
                           UnicodeString& result) const = 0;
};

inline LocaleDisplayNames* LocaleDisplayNames::createInstance(const Locale& locale) {
  return LocaleDisplayNames::createInstance(locale, ULDN_STANDARD_NAMES);
}

U_NAMESPACE_END

#endif

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2014-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* simpleformatter.h
*/


/**
 * \file
 * \brief C++ API: Simple formatter, minimal subset of MessageFormat.
 */



#if U_SHOW_CPLUSPLUS_API



U_NAMESPACE_BEGIN

// Forward declaration:
namespace number {
namespace impl {
class SimpleModifier;
}
}

/**
 * Formats simple patterns like "{1} was born in {0}".
 * Minimal subset of MessageFormat; fast, simple, minimal dependencies.
 * Supports only numbered arguments with no type nor style parameters,
 * and formats only string values.
 * Quoting via ASCII apostrophe compatible with ICU MessageFormat default behavior.
 *
 * Factory methods set error codes for syntax errors
 * and for too few or too many arguments/placeholders.
 *
 * SimpleFormatter objects are thread-safe except for assignment and applying new patterns.
 *
 * Example:
 * <pre>
 * UErrorCode errorCode = U_ZERO_ERROR;
 * SimpleFormatter fmt("{1} '{born}' in {0}", errorCode);
 * UnicodeString result;
 *
 * // Output: "paul {born} in england"
 * fmt.format("england", "paul", result, errorCode);
 * </pre>
 *
 * This class is not intended for public subclassing.
 *
 * @see MessageFormat
 * @see UMessagePatternApostropheMode
 * @stable ICU 57
 */
class U_COMMON_API SimpleFormatter U_FINAL : public UMemory {
public:
    /**
     * Default constructor.
     * @stable ICU 57
     */
    SimpleFormatter() : compiledPattern((char16_t)0) {}

    /**
     * Constructs a formatter from the pattern string.
     *
     * @param pattern The pattern string.
     * @param errorCode ICU error code in/out parameter.
     *                  Must fulfill U_SUCCESS before the function call.
     *                  Set to U_ILLEGAL_ARGUMENT_ERROR for bad argument syntax.
     * @stable ICU 57
     */
    SimpleFormatter(const UnicodeString& pattern, UErrorCode &errorCode) {
        applyPattern(pattern, errorCode);
    }

    /**
     * Constructs a formatter from the pattern string.
     * The number of arguments checked against the given limits is the
     * highest argument number plus one, not the number of occurrences of arguments.
     *
     * @param pattern The pattern string.
     * @param min The pattern must have at least this many arguments.
     * @param max The pattern must have at most this many arguments.
     * @param errorCode ICU error code in/out parameter.
     *                  Must fulfill U_SUCCESS before the function call.
     *                  Set to U_ILLEGAL_ARGUMENT_ERROR for bad argument syntax and
     *                  too few or too many arguments.
     * @stable ICU 57
     */
    SimpleFormatter(const UnicodeString& pattern, int32_t min, int32_t max,
                    UErrorCode &errorCode) {
        applyPatternMinMaxArguments(pattern, min, max, errorCode);
    }

    /**
     * Copy constructor.
     * @stable ICU 57
     */
    SimpleFormatter(const SimpleFormatter& other)
            : compiledPattern(other.compiledPattern) {}

    /**
     * Assignment operator.
     * @stable ICU 57
     */
    SimpleFormatter &operator=(const SimpleFormatter& other);

    /**
     * Destructor.
     * @stable ICU 57
     */
    ~SimpleFormatter();

    /**
     * Changes this object according to the new pattern.
     *
     * @param pattern The pattern string.
     * @param errorCode ICU error code in/out parameter.
     *                  Must fulfill U_SUCCESS before the function call.
     *                  Set to U_ILLEGAL_ARGUMENT_ERROR for bad argument syntax.
     * @return TRUE if U_SUCCESS(errorCode).
     * @stable ICU 57
     */
    UBool applyPattern(const UnicodeString &pattern, UErrorCode &errorCode) {
        return applyPatternMinMaxArguments(pattern, 0, INT32_MAX, errorCode);
    }

    /**
     * Changes this object according to the new pattern.
     * The number of arguments checked against the given limits is the
     * highest argument number plus one, not the number of occurrences of arguments.
     *
     * @param pattern The pattern string.
     * @param min The pattern must have at least this many arguments.
     * @param max The pattern must have at most this many arguments.
     * @param errorCode ICU error code in/out parameter.
     *                  Must fulfill U_SUCCESS before the function call.
     *                  Set to U_ILLEGAL_ARGUMENT_ERROR for bad argument syntax and
     *                  too few or too many arguments.
     * @return TRUE if U_SUCCESS(errorCode).
     * @stable ICU 57
     */
    UBool applyPatternMinMaxArguments(const UnicodeString &pattern,
                                      int32_t min, int32_t max, UErrorCode &errorCode);

    /**
     * @return The max argument number + 1.
     * @stable ICU 57
     */
    int32_t getArgumentLimit() const {
        return getArgumentLimit(compiledPattern.getBuffer(), compiledPattern.length());
    }

    /**
     * Formats the given value, appending to the appendTo builder.
     * The argument value must not be the same object as appendTo.
     * getArgumentLimit() must be at most 1.
     *
     * @param value0 Value for argument {0}.
     * @param appendTo Gets the formatted pattern and value appended.
     * @param errorCode ICU error code in/out parameter.
     *                  Must fulfill U_SUCCESS before the function call.
     * @return appendTo
     * @stable ICU 57
     */
    UnicodeString &format(
            const UnicodeString &value0,
            UnicodeString &appendTo, UErrorCode &errorCode) const;

    /**
     * Formats the given values, appending to the appendTo builder.
     * An argument value must not be the same object as appendTo.
     * getArgumentLimit() must be at most 2.
     *
     * @param value0 Value for argument {0}.
     * @param value1 Value for argument {1}.
     * @param appendTo Gets the formatted pattern and values appended.
     * @param errorCode ICU error code in/out parameter.
     *                  Must fulfill U_SUCCESS before the function call.
     * @return appendTo
     * @stable ICU 57
     */
    UnicodeString &format(
            const UnicodeString &value0,
            const UnicodeString &value1,
            UnicodeString &appendTo, UErrorCode &errorCode) const;

    /**
     * Formats the given values, appending to the appendTo builder.
     * An argument value must not be the same object as appendTo.
     * getArgumentLimit() must be at most 3.
     *
     * @param value0 Value for argument {0}.
     * @param value1 Value for argument {1}.
     * @param value2 Value for argument {2}.
     * @param appendTo Gets the formatted pattern and values appended.
     * @param errorCode ICU error code in/out parameter.
     *                  Must fulfill U_SUCCESS before the function call.
     * @return appendTo
     * @stable ICU 57
     */
    UnicodeString &format(
            const UnicodeString &value0,
            const UnicodeString &value1,
            const UnicodeString &value2,
            UnicodeString &appendTo, UErrorCode &errorCode) const;

    /**
     * Formats the given values, appending to the appendTo string.
     *
     * @param values The argument values.
     *               An argument value must not be the same object as appendTo.
     *               Can be NULL if valuesLength==getArgumentLimit()==0.
     * @param valuesLength The length of the values array.
     *                     Must be at least getArgumentLimit().
     * @param appendTo Gets the formatted pattern and values appended.
     * @param offsets offsets[i] receives the offset of where
     *                values[i] replaced pattern argument {i}.
     *                Can be shorter or longer than values. Can be NULL if offsetsLength==0.
     *                If there is no {i} in the pattern, then offsets[i] is set to -1.
     * @param offsetsLength The length of the offsets array.
     * @param errorCode ICU error code in/out parameter.
     *                  Must fulfill U_SUCCESS before the function call.
     * @return appendTo
     * @stable ICU 57
     */
    UnicodeString &formatAndAppend(
            const UnicodeString *const *values, int32_t valuesLength,
            UnicodeString &appendTo,
            int32_t *offsets, int32_t offsetsLength, UErrorCode &errorCode) const;

    /**
     * Formats the given values, replacing the contents of the result string.
     * May optimize by actually appending to the result if it is the same object
     * as the value corresponding to the initial argument in the pattern.
     *
     * @param values The argument values.
     *               An argument value may be the same object as result.
     *               Can be NULL if valuesLength==getArgumentLimit()==0.
     * @param valuesLength The length of the values array.
     *                     Must be at least getArgumentLimit().
     * @param result Gets its contents replaced by the formatted pattern and values.
     * @param offsets offsets[i] receives the offset of where
     *                values[i] replaced pattern argument {i}.
     *                Can be shorter or longer than values. Can be NULL if offsetsLength==0.
     *                If there is no {i} in the pattern, then offsets[i] is set to -1.
     * @param offsetsLength The length of the offsets array.
     * @param errorCode ICU error code in/out parameter.
     *                  Must fulfill U_SUCCESS before the function call.
     * @return result
     * @stable ICU 57
     */
    UnicodeString &formatAndReplace(
            const UnicodeString *const *values, int32_t valuesLength,
            UnicodeString &result,
            int32_t *offsets, int32_t offsetsLength, UErrorCode &errorCode) const;

    /**
     * Returns the pattern text with none of the arguments.
     * Like formatting with all-empty string values.
     * @stable ICU 57
     */
    UnicodeString getTextWithNoArguments() const {
        return getTextWithNoArguments(
            compiledPattern.getBuffer(),
            compiledPattern.length(),
            nullptr,
            0);
    }

#ifndef U_HIDE_INTERNAL_API
    /**
     * Returns the pattern text with none of the arguments.
     * Like formatting with all-empty string values.
     *
     * TODO(ICU-20406): Replace this with an Iterator interface.
     *
     * @param offsets offsets[i] receives the offset of where {i} was located
     *                before it was replaced by an empty string.
     *                For example, "a{0}b{1}" produces offset 1 for i=0 and 2 for i=1.
     *                Can be nullptr if offsetsLength==0.
     *                If there is no {i} in the pattern, then offsets[i] is set to -1.
     * @param offsetsLength The length of the offsets array.
     *
     * @internal
     */
    UnicodeString getTextWithNoArguments(int32_t *offsets, int32_t offsetsLength) const {
        return getTextWithNoArguments(
            compiledPattern.getBuffer(),
            compiledPattern.length(),
            offsets,
            offsetsLength);
    }
#endif // U_HIDE_INTERNAL_API

private:
    /**
     * Binary representation of the compiled pattern.
     * Index 0: One more than the highest argument number.
     * Followed by zero or more arguments or literal-text segments.
     *
     * An argument is stored as its number, less than ARG_NUM_LIMIT.
     * A literal-text segment is stored as its length (at least 1) offset by ARG_NUM_LIMIT,
     * followed by that many chars.
     */
    UnicodeString compiledPattern;

    static inline int32_t getArgumentLimit(const char16_t *compiledPattern,
                                              int32_t compiledPatternLength) {
        return compiledPatternLength == 0 ? 0 : compiledPattern[0];
    }

    static UnicodeString getTextWithNoArguments(
        const char16_t *compiledPattern,
        int32_t compiledPatternLength,
        int32_t *offsets,
        int32_t offsetsLength);

    static UnicodeString &format(
            const char16_t *compiledPattern, int32_t compiledPatternLength,
            const UnicodeString *const *values,
            UnicodeString &result, const UnicodeString *resultCopy, UBool forbidResultAsValue,
            int32_t *offsets, int32_t offsetsLength,
            UErrorCode &errorCode);

    // Give access to internals to SimpleModifier for number formatting
    friend class number::impl::SimpleModifier;
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2005-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ucasemap.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2005may06
*   created by: Markus W. Scherer
*
*   Case mapping service object and functions using it.
*/







/**
 * \file
 * \brief C API: Unicode case mapping functions using a UCaseMap service object.
 *
 * The service object takes care of memory allocations, data loading, and setup
 * for the attributes, as usual.
 *
 * Currently, the functionality provided here does not overlap with uchar.h
 * and ustring.h, except for ucasemap_toTitle().
 *
 * ucasemap_utf8XYZ() functions operate directly on UTF-8 strings.
 */

/**
 * UCaseMap is an opaque service object for newer ICU case mapping functions.
 * Older functions did not use a service object.
 * @stable ICU 3.4
 */
struct UCaseMap;
typedef struct UCaseMap UCaseMap; /**< C typedef for struct UCaseMap. @stable ICU 3.4 */

/**
 * Open a UCaseMap service object for a locale and a set of options.
 * The locale ID and options are preprocessed so that functions using the
 * service object need not process them in each call.
 *
 * @param locale ICU locale ID, used for language-dependent
 *               upper-/lower-/title-casing according to the Unicode standard.
 *               Usual semantics: ""=root, NULL=default locale, etc.
 * @param options Options bit set, used for case folding and string comparisons.
 *                Same flags as for u_foldCase(), u_strFoldCase(),
 *                u_strCaseCompare(), etc.
 *                Use 0 or U_FOLD_CASE_DEFAULT for default behavior.
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                   which must not indicate a failure before the function call.
 * @return Pointer to a UCaseMap service object, if successful.
 *
 * @see U_FOLD_CASE_DEFAULT
 * @see U_FOLD_CASE_EXCLUDE_SPECIAL_I
 * @see U_TITLECASE_NO_LOWERCASE
 * @see U_TITLECASE_NO_BREAK_ADJUSTMENT
 * @stable ICU 3.4
 */
U_STABLE UCaseMap * U_EXPORT2
ucasemap_open(const char *locale, uint32_t options, UErrorCode *pErrorCode);

/**
 * Close a UCaseMap service object.
 * @param csm Object to be closed.
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
ucasemap_close(UCaseMap *csm);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUCaseMapPointer
 * "Smart pointer" class, closes a UCaseMap via ucasemap_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUCaseMapPointer, UCaseMap, ucasemap_close);

U_NAMESPACE_END

#endif

/**
 * Get the locale ID that is used for language-dependent case mappings.
 * @param csm UCaseMap service object.
 * @return locale ID
 * @stable ICU 3.4
 */
U_STABLE const char * U_EXPORT2
ucasemap_getLocale(const UCaseMap *csm);

/**
 * Get the options bit set that is used for case folding and string comparisons.
 * @param csm UCaseMap service object.
 * @return options bit set
 * @stable ICU 3.4
 */
U_STABLE uint32_t U_EXPORT2
ucasemap_getOptions(const UCaseMap *csm);

/**
 * Set the locale ID that is used for language-dependent case mappings.
 *
 * @param csm UCaseMap service object.
 * @param locale Locale ID, see ucasemap_open().
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                   which must not indicate a failure before the function call.
 *
 * @see ucasemap_open
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
ucasemap_setLocale(UCaseMap *csm, const char *locale, UErrorCode *pErrorCode);

/**
 * Set the options bit set that is used for case folding and string comparisons.
 *
 * @param csm UCaseMap service object.
 * @param options Options bit set, see ucasemap_open().
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                   which must not indicate a failure before the function call.
 *
 * @see ucasemap_open
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
ucasemap_setOptions(UCaseMap *csm, uint32_t options, UErrorCode *pErrorCode);

#if !UCONFIG_NO_BREAK_ITERATION

/**
 * Get the break iterator that is used for titlecasing.
 * Do not modify the returned break iterator.
 * @param csm UCaseMap service object.
 * @return titlecasing break iterator
 * @stable ICU 3.8
 */
U_STABLE const UBreakIterator * U_EXPORT2
ucasemap_getBreakIterator(const UCaseMap *csm);

/**
 * Set the break iterator that is used for titlecasing.
 * The UCaseMap service object releases a previously set break iterator
 * and "adopts" this new one, taking ownership of it.
 * It will be released in a subsequent call to ucasemap_setBreakIterator()
 * or ucasemap_close().
 *
 * Break iterator operations are not thread-safe. Therefore, titlecasing
 * functions use non-const UCaseMap objects. It is not possible to titlecase
 * strings concurrently using the same UCaseMap.
 *
 * @param csm UCaseMap service object.
 * @param iterToAdopt Break iterator to be adopted for titlecasing.
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                   which must not indicate a failure before the function call.
 *
 * @see ucasemap_toTitle
 * @see ucasemap_utf8ToTitle
 * @stable ICU 3.8
 */
U_STABLE void U_EXPORT2
ucasemap_setBreakIterator(UCaseMap *csm, UBreakIterator *iterToAdopt, UErrorCode *pErrorCode);

/**
 * Titlecase a UTF-16 string. This function is almost a duplicate of u_strToTitle(),
 * except that it takes ucasemap_setOptions() into account and has performance
 * advantages from being able to use a UCaseMap object for multiple case mapping
 * operations, saving setup time.
 *
 * Casing is locale-dependent and context-sensitive.
 * Titlecasing uses a break iterator to find the first characters of words
 * that are to be titlecased. It titlecases those characters and lowercases
 * all others. (This can be modified with ucasemap_setOptions().)
 *
 * Note: This function takes a non-const UCaseMap pointer because it will
 * open a default break iterator if no break iterator was set yet,
 * and effectively call ucasemap_setBreakIterator();
 * also because the break iterator is stateful and will be modified during
 * the iteration.
 *
 * The titlecase break iterator can be provided to customize for arbitrary
 * styles, using rules and dictionaries beyond the standard iterators.
 * The standard titlecase iterator for the root locale implements the
 * algorithm of Unicode TR 21.
 *
 * This function uses only the setText(), first() and next() methods of the
 * provided break iterator.
 *
 * The result may be longer or shorter than the original.
 * The source string and the destination buffer must not overlap.
 *
 * @param csm       UCaseMap service object. This pointer is non-const!
 *                  See the note above for details.
 * @param dest      A buffer for the result string. The result will be NUL-terminated if
 *                  the buffer is large enough.
 *                  The contents is undefined in case of failure.
 * @param destCapacity The size of the buffer (number of UChars). If it is 0, then
 *                  dest may be NULL and the function will only return the length of the result
 *                  without writing any of the result string.
 * @param src       The original string.
 * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return The length of the result string, if successful - or in case of a buffer overflow,
 *         in which case it will be greater than destCapacity.
 *
 * @see u_strToTitle
 * @stable ICU 3.8
 */
U_STABLE int32_t U_EXPORT2
ucasemap_toTitle(UCaseMap *csm,
                 UChar *dest, int32_t destCapacity,
                 const UChar *src, int32_t srcLength,
                 UErrorCode *pErrorCode);

#endif  // UCONFIG_NO_BREAK_ITERATION

/**
 * Lowercase the characters in a UTF-8 string.
 * Casing is locale-dependent and context-sensitive.
 * The result may be longer or shorter than the original.
 * The source string and the destination buffer must not overlap.
 *
 * @param csm       UCaseMap service object.
 * @param dest      A buffer for the result string. The result will be NUL-terminated if
 *                  the buffer is large enough.
 *                  The contents is undefined in case of failure.
 * @param destCapacity The size of the buffer (number of bytes). If it is 0, then
 *                  dest may be NULL and the function will only return the length of the result
 *                  without writing any of the result string.
 * @param src       The original string.
 * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return The length of the result string, if successful - or in case of a buffer overflow,
 *         in which case it will be greater than destCapacity.
 *
 * @see u_strToLower
 * @stable ICU 3.4
 */
U_STABLE int32_t U_EXPORT2
ucasemap_utf8ToLower(const UCaseMap *csm,
                     char *dest, int32_t destCapacity,
                     const char *src, int32_t srcLength,
                     UErrorCode *pErrorCode);

/**
 * Uppercase the characters in a UTF-8 string.
 * Casing is locale-dependent and context-sensitive.
 * The result may be longer or shorter than the original.
 * The source string and the destination buffer must not overlap.
 *
 * @param csm       UCaseMap service object.
 * @param dest      A buffer for the result string. The result will be NUL-terminated if
 *                  the buffer is large enough.
 *                  The contents is undefined in case of failure.
 * @param destCapacity The size of the buffer (number of bytes). If it is 0, then
 *                  dest may be NULL and the function will only return the length of the result
 *                  without writing any of the result string.
 * @param src       The original string.
 * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return The length of the result string, if successful - or in case of a buffer overflow,
 *         in which case it will be greater than destCapacity.
 *
 * @see u_strToUpper
 * @stable ICU 3.4
 */
U_STABLE int32_t U_EXPORT2
ucasemap_utf8ToUpper(const UCaseMap *csm,
                     char *dest, int32_t destCapacity,
                     const char *src, int32_t srcLength,
                     UErrorCode *pErrorCode);

#if !UCONFIG_NO_BREAK_ITERATION

/**
 * Titlecase a UTF-8 string.
 * Casing is locale-dependent and context-sensitive.
 * Titlecasing uses a break iterator to find the first characters of words
 * that are to be titlecased. It titlecases those characters and lowercases
 * all others. (This can be modified with ucasemap_setOptions().)
 *
 * Note: This function takes a non-const UCaseMap pointer because it will
 * open a default break iterator if no break iterator was set yet,
 * and effectively call ucasemap_setBreakIterator();
 * also because the break iterator is stateful and will be modified during
 * the iteration.
 *
 * The titlecase break iterator can be provided to customize for arbitrary
 * styles, using rules and dictionaries beyond the standard iterators.
 * The standard titlecase iterator for the root locale implements the
 * algorithm of Unicode TR 21.
 *
 * This function uses only the setUText(), first(), next() and close() methods of the
 * provided break iterator.
 *
 * The result may be longer or shorter than the original.
 * The source string and the destination buffer must not overlap.
 *
 * @param csm       UCaseMap service object. This pointer is non-const!
 *                  See the note above for details.
 * @param dest      A buffer for the result string. The result will be NUL-terminated if
 *                  the buffer is large enough.
 *                  The contents is undefined in case of failure.
 * @param destCapacity The size of the buffer (number of bytes). If it is 0, then
 *                  dest may be NULL and the function will only return the length of the result
 *                  without writing any of the result string.
 * @param src       The original string.
 * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return The length of the result string, if successful - or in case of a buffer overflow,
 *         in which case it will be greater than destCapacity.
 *
 * @see u_strToTitle
 * @see U_TITLECASE_NO_LOWERCASE
 * @see U_TITLECASE_NO_BREAK_ADJUSTMENT
 * @stable ICU 3.8
 */
U_STABLE int32_t U_EXPORT2
ucasemap_utf8ToTitle(UCaseMap *csm,
                    char *dest, int32_t destCapacity,
                    const char *src, int32_t srcLength,
                    UErrorCode *pErrorCode);

#endif

/**
 * Case-folds the characters in a UTF-8 string.
 *
 * Case-folding is locale-independent and not context-sensitive,
 * but there is an option for whether to include or exclude mappings for dotted I
 * and dotless i that are marked with 'T' in CaseFolding.txt.
 *
 * The result may be longer or shorter than the original.
 * The source string and the destination buffer must not overlap.
 *
 * @param csm       UCaseMap service object.
 * @param dest      A buffer for the result string. The result will be NUL-terminated if
 *                  the buffer is large enough.
 *                  The contents is undefined in case of failure.
 * @param destCapacity The size of the buffer (number of bytes). If it is 0, then
 *                  dest may be NULL and the function will only return the length of the result
 *                  without writing any of the result string.
 * @param src       The original string.
 * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
 * @param pErrorCode Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return The length of the result string, if successful - or in case of a buffer overflow,
 *         in which case it will be greater than destCapacity.
 *
 * @see u_strFoldCase
 * @see ucasemap_setOptions
 * @see U_FOLD_CASE_DEFAULT
 * @see U_FOLD_CASE_EXCLUDE_SPECIAL_I
 * @stable ICU 3.8
 */
U_STABLE int32_t U_EXPORT2
ucasemap_utf8FoldCase(const UCaseMap *csm,
                      char *dest, int32_t destCapacity,
                      const char *src, int32_t srcLength,
                      UErrorCode *pErrorCode);





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2002-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
*/




/**
 * \file 
 * \brief C API: Encapsulates information about a currency.
 *
 * The ucurr API encapsulates information about a currency, as defined by
 * ISO 4217.  A currency is represented by a 3-character string
 * containing its ISO 4217 code.  This API can return various data
 * necessary the proper display of a currency:
 *
 * <ul><li>A display symbol, for a specific locale
 * <li>The number of fraction digits to display
 * <li>A rounding increment
 * </ul>
 *
 * The <tt>DecimalFormat</tt> class uses these data to display
 * currencies.
 * @author Alan Liu
 * @since ICU 2.2
 */

#if !UCONFIG_NO_FORMATTING

/**
 * Currency Usage used for Decimal Format
 * @stable ICU 54
 */
enum UCurrencyUsage {
    /**
     * a setting to specify currency usage which determines currency digit
     * and rounding for standard usage, for example: "50.00 NT$"
     * used as DEFAULT value
     * @stable ICU 54
     */
    UCURR_USAGE_STANDARD=0,
    /**
     * a setting to specify currency usage which determines currency digit
     * and rounding for cash usage, for example: "50 NT$"
     * @stable ICU 54
     */
    UCURR_USAGE_CASH=1,
#ifndef U_HIDE_DEPRECATED_API
    /**
     * One higher than the last enum UCurrencyUsage constant.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UCURR_USAGE_COUNT=2
#endif  // U_HIDE_DEPRECATED_API
};
/** Currency Usage used for Decimal Format */
typedef enum UCurrencyUsage UCurrencyUsage; 

/**
 * Finds a currency code for the given locale.
 * @param locale the locale for which to retrieve a currency code. 
 *               Currency can be specified by the "currency" keyword
 *               in which case it overrides the default currency code
 * @param buff   fill in buffer. Can be NULL for preflighting.
 * @param buffCapacity capacity of the fill in buffer. Can be 0 for
 *               preflighting. If it is non-zero, the buff parameter
 *               must not be NULL.
 * @param ec error code
 * @return length of the currency string. It should always be 3. If 0,
 *                currency couldn't be found or the input values are 
 *                invalid. 
 * @stable ICU 2.8
 */
U_STABLE int32_t U_EXPORT2
ucurr_forLocale(const char* locale,
                UChar* buff,
                int32_t buffCapacity,
                UErrorCode* ec);

/**
 * Selector constants for ucurr_getName().
 *
 * @see ucurr_getName
 * @stable ICU 2.6
 */
typedef enum UCurrNameStyle {
    /**
     * Selector for ucurr_getName indicating a symbolic name for a
     * currency, such as "$" for USD.
     * @stable ICU 2.6
     */
    UCURR_SYMBOL_NAME,

    /**
     * Selector for ucurr_getName indicating the long name for a
     * currency, such as "US Dollar" for USD.
     * @stable ICU 2.6
     */
    UCURR_LONG_NAME,

    /**
     * Selector for getName() indicating the narrow currency symbol.
     * The narrow currency symbol is similar to the regular currency
     * symbol, but it always takes the shortest form: for example,
     * "$" instead of "US$" for USD in en-CA.
     *
     * @stable ICU 61
     */
    UCURR_NARROW_SYMBOL_NAME
} UCurrNameStyle;

#if !UCONFIG_NO_SERVICE
/**
 * @stable ICU 2.6
 */
typedef const void* UCurrRegistryKey;

/**
 * Register an (existing) ISO 4217 currency code for the given locale.
 * Only the country code and the two variants EURO and PRE_EURO are
 * recognized.
 * @param isoCode the three-letter ISO 4217 currency code
 * @param locale  the locale for which to register this currency code
 * @param status the in/out status code
 * @return a registry key that can be used to unregister this currency code, or NULL
 * if there was an error.
 * @stable ICU 2.6
 */
U_STABLE UCurrRegistryKey U_EXPORT2
ucurr_register(const UChar* isoCode, 
                   const char* locale,  
                   UErrorCode* status);
/**
 * Unregister the previously-registered currency definitions using the
 * URegistryKey returned from ucurr_register.  Key becomes invalid after
 * a successful call and should not be used again.  Any currency 
 * that might have been hidden by the original ucurr_register call is 
 * restored.
 * @param key the registry key returned by a previous call to ucurr_register
 * @param status the in/out status code, no special meanings are assigned
 * @return TRUE if the currency for this key was successfully unregistered
 * @stable ICU 2.6
 */
U_STABLE UBool U_EXPORT2
ucurr_unregister(UCurrRegistryKey key, UErrorCode* status);
#endif /* UCONFIG_NO_SERVICE */

/**
 * Returns the display name for the given currency in the
 * given locale.  For example, the display name for the USD
 * currency object in the en_US locale is "$".
 * @param currency null-terminated 3-letter ISO 4217 code
 * @param locale locale in which to display currency
 * @param nameStyle selector for which kind of name to return
 * @param isChoiceFormat always set to FALSE, or can be NULL;
 *     display names are static strings;
 *     since ICU 4.4, ChoiceFormat patterns are no longer supported
 * @param len fill-in parameter to receive length of result
 * @param ec error code
 * @return pointer to display string of 'len' UChars.  If the resource
 * data contains no entry for 'currency', then 'currency' itself is
 * returned.
 * @stable ICU 2.6
 */
U_STABLE const UChar* U_EXPORT2
ucurr_getName(const UChar* currency,
              const char* locale,
              UCurrNameStyle nameStyle,
              UBool* isChoiceFormat,
              int32_t* len,
              UErrorCode* ec);

/**
 * Returns the plural name for the given currency in the
 * given locale.  For example, the plural name for the USD
 * currency object in the en_US locale is "US dollar" or "US dollars".
 * @param currency null-terminated 3-letter ISO 4217 code
 * @param locale locale in which to display currency
 * @param isChoiceFormat always set to FALSE, or can be NULL;
 *     display names are static strings;
 *     since ICU 4.4, ChoiceFormat patterns are no longer supported
 * @param pluralCount plural count
 * @param len fill-in parameter to receive length of result
 * @param ec error code
 * @return pointer to display string of 'len' UChars.  If the resource
 * data contains no entry for 'currency', then 'currency' itself is
 * returned.
 * @stable ICU 4.2
 */
U_STABLE const UChar* U_EXPORT2
ucurr_getPluralName(const UChar* currency,
                    const char* locale,
                    UBool* isChoiceFormat,
                    const char* pluralCount,
                    int32_t* len,
                    UErrorCode* ec);

/**
 * Returns the number of the number of fraction digits that should
 * be displayed for the given currency.
 * This is equivalent to ucurr_getDefaultFractionDigitsForUsage(currency,UCURR_USAGE_STANDARD,ec);
 *
 * Important: The number of fraction digits for a given currency is NOT
 * guaranteed to be constant across versions of ICU or CLDR. For example,
 * do NOT use this value as a mechanism for deciding the magnitude used
 * to store currency values in a database. You should use this value for
 * display purposes only.
 *
 * @param currency null-terminated 3-letter ISO 4217 code
 * @param ec input-output error code
 * @return a non-negative number of fraction digits to be
 * displayed, or 0 if there is an error
 * @stable ICU 3.0
 */
U_STABLE int32_t U_EXPORT2
ucurr_getDefaultFractionDigits(const UChar* currency,
                               UErrorCode* ec);

/**
 * Returns the number of the number of fraction digits that should
 * be displayed for the given currency with usage.
 *
 * Important: The number of fraction digits for a given currency is NOT
 * guaranteed to be constant across versions of ICU or CLDR. For example,
 * do NOT use this value as a mechanism for deciding the magnitude used
 * to store currency values in a database. You should use this value for
 * display purposes only.
 *
 * @param currency null-terminated 3-letter ISO 4217 code
 * @param usage enum usage for the currency
 * @param ec input-output error code
 * @return a non-negative number of fraction digits to be
 * displayed, or 0 if there is an error
 * @stable ICU 54
 */
U_STABLE int32_t U_EXPORT2
ucurr_getDefaultFractionDigitsForUsage(const UChar* currency, 
                                       const UCurrencyUsage usage,
                                       UErrorCode* ec);

/**
 * Returns the rounding increment for the given currency, or 0.0 if no
 * rounding is done by the currency.
 * This is equivalent to ucurr_getRoundingIncrementForUsage(currency,UCURR_USAGE_STANDARD,ec);
 * @param currency null-terminated 3-letter ISO 4217 code
 * @param ec input-output error code
 * @return the non-negative rounding increment, or 0.0 if none,
 * or 0.0 if there is an error
 * @stable ICU 3.0
 */
U_STABLE double U_EXPORT2
ucurr_getRoundingIncrement(const UChar* currency,
                           UErrorCode* ec);

/**
 * Returns the rounding increment for the given currency, or 0.0 if no
 * rounding is done by the currency given usage.
 * @param currency null-terminated 3-letter ISO 4217 code
 * @param usage enum usage for the currency
 * @param ec input-output error code
 * @return the non-negative rounding increment, or 0.0 if none,
 * or 0.0 if there is an error
 * @stable ICU 54
 */
U_STABLE double U_EXPORT2
ucurr_getRoundingIncrementForUsage(const UChar* currency,
                                   const UCurrencyUsage usage,
                                   UErrorCode* ec);

/**
 * Selector constants for ucurr_openCurrencies().
 *
 * @see ucurr_openCurrencies
 * @stable ICU 3.2
 */
typedef enum UCurrCurrencyType {
    /**
     * Select all ISO-4217 currency codes.
     * @stable ICU 3.2
     */
    UCURR_ALL = INT32_MAX,
    /**
     * Select only ISO-4217 commonly used currency codes.
     * These currencies can be found in common use, and they usually have
     * bank notes or coins associated with the currency code.
     * This does not include fund codes, precious metals and other
     * various ISO-4217 codes limited to special financial products.
     * @stable ICU 3.2
     */
    UCURR_COMMON = 1,
    /**
     * Select ISO-4217 uncommon currency codes.
     * These codes respresent fund codes, precious metals and other
     * various ISO-4217 codes limited to special financial products.
     * A fund code is a monetary resource associated with a currency.
     * @stable ICU 3.2
     */
    UCURR_UNCOMMON = 2,
    /**
     * Select only deprecated ISO-4217 codes.
     * These codes are no longer in general public use.
     * @stable ICU 3.2
     */
    UCURR_DEPRECATED = 4,
    /**
     * Select only non-deprecated ISO-4217 codes.
     * These codes are in general public use.
     * @stable ICU 3.2
     */
    UCURR_NON_DEPRECATED = 8
} UCurrCurrencyType;

/**
 * Provides a UEnumeration object for listing ISO-4217 codes.
 * @param currType You can use one of several UCurrCurrencyType values for this
 *      variable. You can also | (or) them together to get a specific list of
 *      currencies. Most people will want to use the (UCURR_COMMON|UCURR_NON_DEPRECATED) value to
 *      get a list of current currencies.
 * @param pErrorCode Error code
 * @stable ICU 3.2
 */
U_STABLE UEnumeration * U_EXPORT2
ucurr_openISOCurrencies(uint32_t currType, UErrorCode *pErrorCode);

/**
  * Queries if the given ISO 4217 3-letter code is available on the specified date range. 
  * 
  * Note: For checking availability of a currency on a specific date, specify the date on both 'from' and 'to' 
  * 
  * When 'from' is U_DATE_MIN and 'to' is U_DATE_MAX, this method checks if the specified currency is available any time. 
  * If 'from' and 'to' are same UDate value, this method checks if the specified currency is available on that date.
  * 
  * @param isoCode 
  *            The ISO 4217 3-letter code. 
  * 
  * @param from 
  *            The lower bound of the date range, inclusive. When 'from' is U_DATE_MIN, check the availability 
  *            of the currency any date before 'to' 
  * 
  * @param to 
  *            The upper bound of the date range, inclusive. When 'to' is U_DATE_MAX, check the availability of 
  *            the currency any date after 'from' 
  * 
  * @param errorCode 
  *            ICU error code 
   * 
  * @return TRUE if the given ISO 4217 3-letter code is supported on the specified date range. 
  * 
  * @stable ICU 4.8 
  */ 
U_STABLE UBool U_EXPORT2
ucurr_isAvailable(const UChar* isoCode, 
             UDate from, 
             UDate to, 
             UErrorCode* errorCode);

/** 
 * Finds the number of valid currency codes for the
 * given locale and date.
 * @param locale the locale for which to retrieve the
 *               currency count.
 * @param date   the date for which to retrieve the
 *               currency count for the given locale.
 * @param ec     error code
 * @return       the number of currency codes for the
 *               given locale and date.  If 0, currency
 *               codes couldn't be found for the input
 *               values are invalid.
 * @stable ICU 4.0
 */
U_STABLE int32_t U_EXPORT2
ucurr_countCurrencies(const char* locale, 
                 UDate date, 
                 UErrorCode* ec); 

/** 
 * Finds a currency code for the given locale and date 
 * @param locale the locale for which to retrieve a currency code.  
 *               Currency can be specified by the "currency" keyword 
 *               in which case it overrides the default currency code 
 * @param date   the date for which to retrieve a currency code for 
 *               the given locale. 
 * @param index  the index within the available list of currency codes
 *               for the given locale on the given date.
 * @param buff   fill in buffer. Can be NULL for preflighting. 
 * @param buffCapacity capacity of the fill in buffer. Can be 0 for 
 *               preflighting. If it is non-zero, the buff parameter 
 *               must not be NULL. 
 * @param ec     error code 
 * @return       length of the currency string. It should always be 3. 
 *               If 0, currency couldn't be found or the input values are  
 *               invalid.  
 * @stable ICU 4.0 
 */ 
U_STABLE int32_t U_EXPORT2 
ucurr_forLocaleAndDate(const char* locale, 
                UDate date, 
                int32_t index,
                UChar* buff, 
                int32_t buffCapacity, 
                UErrorCode* ec); 

/**
 * Given a key and a locale, returns an array of string values in a preferred
 * order that would make a difference. These are all and only those values where
 * the open (creation) of the service with the locale formed from the input locale
 * plus input keyword and that value has different behavior than creation with the
 * input locale alone.
 * @param key           one of the keys supported by this service.  For now, only
 *                      "currency" is supported.
 * @param locale        the locale
 * @param commonlyUsed  if set to true it will return only commonly used values
 *                      with the given locale in preferred order.  Otherwise,
 *                      it will return all the available values for the locale.
 * @param status error status
 * @return a string enumeration over keyword values for the given key and the locale.
 * @stable ICU 4.2
 */
U_STABLE UEnumeration* U_EXPORT2
ucurr_getKeywordValuesForLocale(const char* key,
                                const char* locale,
                                UBool commonlyUsed,
                                UErrorCode* status);

/**
 * Returns the ISO 4217 numeric code for the currency.
 * <p>Note: If the ISO 4217 numeric code is not assigned for the currency or
 * the currency is unknown, this function returns 0.
 *
 * @param currency null-terminated 3-letter ISO 4217 code
 * @return The ISO 4217 numeric code of the currency
 * @stable ICU 49
 */
U_STABLE int32_t U_EXPORT2
ucurr_getNumericCode(const UChar* currency);

#endif /* #if !UCONFIG_NO_FORMATTING */










#include <stdarg.h>

/**
 * Concatenate a number of null-terminated strings to buffer, leaving a
 * null-terminated string.  The last argument should be the null pointer.
 * Return the length of the string in the buffer, not counting the trailing
 * null.  Return -1 if there is an error (buffer is null, or buflen < 1).
 */
static int32_t ncat(char *buffer, uint32_t buflen, ...) {
  va_list args;
  char *str;
  char *p = buffer;
  const char* e = buffer + buflen - 1;

  if (buffer == NULL || buflen < 1) {
    return -1;
  }

  va_start(args, buflen);
  while ((str = va_arg(args, char *)) != 0) {
    char c;
    while (p != e && (c = *str++) != 0) {
      *p++ = c;
    }
  }
  *p = 0;
  va_end(args);

  return static_cast<int32_t>(p - buffer);
}

U_NAMESPACE_BEGIN

////////////////////////////////////////////////////////////////////////////////////////////////////

// Access resource data for locale components.
// Wrap code in uloc.c for now.
class ICUDataTable {
    const char* path;
    Locale locale;

public:
    ICUDataTable(const char* path, const Locale& locale);
    ~ICUDataTable();

    const Locale& getLocale();

    UnicodeString& get(const char* tableKey, const char* itemKey,
                        UnicodeString& result) const;
    UnicodeString& get(const char* tableKey, const char* subTableKey, const char* itemKey,
                        UnicodeString& result) const;

    UnicodeString& getNoFallback(const char* tableKey, const char* itemKey,
                                UnicodeString &result) const;
    UnicodeString& getNoFallback(const char* tableKey, const char* subTableKey, const char* itemKey,
                                UnicodeString &result) const;
};

inline UnicodeString &
ICUDataTable::get(const char* tableKey, const char* itemKey, UnicodeString& result) const {
    return get(tableKey, NULL, itemKey, result);
}

inline UnicodeString &
ICUDataTable::getNoFallback(const char* tableKey, const char* itemKey, UnicodeString& result) const {
    return getNoFallback(tableKey, NULL, itemKey, result);
}

ICUDataTable::ICUDataTable(const char* path, const Locale& locale)
    : path(NULL), locale(Locale::getRoot())
{
  if (path) {
    int32_t len = static_cast<int32_t>(uprv_strlen(path));
    this->path = (const char*) uprv_malloc(len + 1);
    if (this->path) {
      uprv_strcpy((char *)this->path, path);
      this->locale = locale;
    }
  }
}

ICUDataTable::~ICUDataTable() {
  if (path) {
    uprv_free((void*) path);
    path = NULL;
  }
}

const Locale&
ICUDataTable::getLocale() {
  return locale;
}

UnicodeString &
ICUDataTable::get(const char* tableKey, const char* subTableKey, const char* itemKey,
                  UnicodeString &result) const {
  UErrorCode status = U_ZERO_ERROR;
  int32_t len = 0;

  const UChar *s = uloc_getTableStringWithFallback(path, locale.getName(),
                                                   tableKey, subTableKey, itemKey,
                                                   &len, &status);
  if (U_SUCCESS(status) && len > 0) {
    return result.setTo(s, len);
  }
  return result.setTo(UnicodeString(itemKey, -1, US_INV));
}

UnicodeString &
ICUDataTable::getNoFallback(const char* tableKey, const char* subTableKey, const char* itemKey,
                            UnicodeString& result) const {
  UErrorCode status = U_ZERO_ERROR;
  int32_t len = 0;

  const UChar *s = uloc_getTableStringWithFallback(path, locale.getName(),
                                                   tableKey, subTableKey, itemKey,
                                                   &len, &status);
  if (U_SUCCESS(status)) {
    return result.setTo(s, len);
  }

  result.setToBogus();
  return result;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

LocaleDisplayNames::~LocaleDisplayNames() {}

////////////////////////////////////////////////////////////////////////////////////////////////////

#if 0  // currently unused

class DefaultLocaleDisplayNames : public LocaleDisplayNames {
  UDialectHandling dialectHandling;

public:
  // constructor
  DefaultLocaleDisplayNames(UDialectHandling dialectHandling);

  virtual ~DefaultLocaleDisplayNames();

  virtual const Locale& getLocale() const;
  virtual UDialectHandling getDialectHandling() const;

  virtual UnicodeString& localeDisplayName(const Locale& locale,
                                           UnicodeString& result) const;
  virtual UnicodeString& localeDisplayName(const char* localeId,
                                           UnicodeString& result) const;
  virtual UnicodeString& languageDisplayName(const char* lang,
                                             UnicodeString& result) const;
  virtual UnicodeString& scriptDisplayName(const char* script,
                                           UnicodeString& result) const;
  virtual UnicodeString& scriptDisplayName(UScriptCode scriptCode,
                                           UnicodeString& result) const;
  virtual UnicodeString& regionDisplayName(const char* region,
                                           UnicodeString& result) const;
  virtual UnicodeString& variantDisplayName(const char* variant,
                                            UnicodeString& result) const;
  virtual UnicodeString& keyDisplayName(const char* key,
                                        UnicodeString& result) const;
  virtual UnicodeString& keyValueDisplayName(const char* key,
                                             const char* value,
                                             UnicodeString& result) const;
};

DefaultLocaleDisplayNames::DefaultLocaleDisplayNames(UDialectHandling dialectHandling)
    : dialectHandling(dialectHandling) {
}

DefaultLocaleDisplayNames::~DefaultLocaleDisplayNames() {
}

const Locale&
DefaultLocaleDisplayNames::getLocale() const {
  return Locale::getRoot();
}

UDialectHandling
DefaultLocaleDisplayNames::getDialectHandling() const {
  return dialectHandling;
}

UnicodeString&
DefaultLocaleDisplayNames::localeDisplayName(const Locale& locale,
                                             UnicodeString& result) const {
  return result = UnicodeString(locale.getName(), -1, US_INV);
}

UnicodeString&
DefaultLocaleDisplayNames::localeDisplayName(const char* localeId,
                                             UnicodeString& result) const {
  return result = UnicodeString(localeId, -1, US_INV);
}

UnicodeString&
DefaultLocaleDisplayNames::languageDisplayName(const char* lang,
                                               UnicodeString& result) const {
  return result = UnicodeString(lang, -1, US_INV);
}

UnicodeString&
DefaultLocaleDisplayNames::scriptDisplayName(const char* script,
                                             UnicodeString& result) const {
  return result = UnicodeString(script, -1, US_INV);
}

UnicodeString&
DefaultLocaleDisplayNames::scriptDisplayName(UScriptCode scriptCode,
                                             UnicodeString& result) const {
  const char* name = uscript_getName(scriptCode);
  if (name) {
    return result = UnicodeString(name, -1, US_INV);
  }
  return result.remove();
}

UnicodeString&
DefaultLocaleDisplayNames::regionDisplayName(const char* region,
                                             UnicodeString& result) const {
  return result = UnicodeString(region, -1, US_INV);
}

UnicodeString&
DefaultLocaleDisplayNames::variantDisplayName(const char* variant,
                                              UnicodeString& result) const {
  return result = UnicodeString(variant, -1, US_INV);
}

UnicodeString&
DefaultLocaleDisplayNames::keyDisplayName(const char* key,
                                          UnicodeString& result) const {
  return result = UnicodeString(key, -1, US_INV);
}

UnicodeString&
DefaultLocaleDisplayNames::keyValueDisplayName(const char* /* key */,
                                               const char* value,
                                               UnicodeString& result) const {
  return result = UnicodeString(value, -1, US_INV);
}

#endif  // currently unused class DefaultLocaleDisplayNames

////////////////////////////////////////////////////////////////////////////////////////////////////

class LocaleDisplayNamesImpl : public LocaleDisplayNames {
    Locale locale;
    UDialectHandling dialectHandling;
    ICUDataTable langData;
    ICUDataTable regionData;
    SimpleFormatter separatorFormat;
    SimpleFormatter format;
    SimpleFormatter keyTypeFormat;
    UDisplayContext capitalizationContext;
public:
#if !UCONFIG_NO_BREAK_ITERATION
    BreakIterator* capitalizationBrkIter;
#else
    UObject* capitalizationBrkIter;
#endif
private:
    UnicodeString formatOpenParen;
    UnicodeString formatReplaceOpenParen;
    UnicodeString formatCloseParen;
    UnicodeString formatReplaceCloseParen;
    UDisplayContext nameLength;
    UDisplayContext substitute;

    // Constants for capitalization context usage types.
    enum CapContextUsage {
        kCapContextUsageLanguage,
        kCapContextUsageScript,
        kCapContextUsageTerritory,
        kCapContextUsageVariant,
        kCapContextUsageKey,
        kCapContextUsageKeyValue,
        kCapContextUsageCount
    };
    // Capitalization transforms. For each usage type, indicates whether to titlecase for
    // the context specified in capitalizationContext (which we know at construction time)
     UBool fCapitalization[kCapContextUsageCount];

public:
    // constructor
    LocaleDisplayNamesImpl(const Locale& locale, UDialectHandling dialectHandling);
    LocaleDisplayNamesImpl(const Locale& locale, UDisplayContext *contexts, int32_t length);
    virtual ~LocaleDisplayNamesImpl();

    virtual const Locale& getLocale() const;
    virtual UDialectHandling getDialectHandling() const;
    virtual UDisplayContext getContext(UDisplayContextType type) const;

    virtual UnicodeString& localeDisplayName(const Locale& locale,
                                                UnicodeString& result) const;
    virtual UnicodeString& localeDisplayName(const char* localeId,
                                                UnicodeString& result) const;
    virtual UnicodeString& languageDisplayName(const char* lang,
                                               UnicodeString& result) const;
    virtual UnicodeString& scriptDisplayName(const char* script,
                                                UnicodeString& result) const;
    virtual UnicodeString& scriptDisplayName(UScriptCode scriptCode,
                                                UnicodeString& result) const;
    virtual UnicodeString& regionDisplayName(const char* region,
                                                UnicodeString& result) const;
    virtual UnicodeString& variantDisplayName(const char* variant,
                                                UnicodeString& result) const;
    virtual UnicodeString& keyDisplayName(const char* key,
                                                UnicodeString& result) const;
    virtual UnicodeString& keyValueDisplayName(const char* key,
                                                const char* value,
                                                UnicodeString& result) const;
private:
    UnicodeString& localeIdName(const char* localeId,
                                UnicodeString& result, bool substitute) const;
    UnicodeString& appendWithSep(UnicodeString& buffer, const UnicodeString& src) const;
    UnicodeString& adjustForUsageAndContext(CapContextUsage usage, UnicodeString& result) const;
    UnicodeString& scriptDisplayName(const char* script, UnicodeString& result, UBool skipAdjust) const;
    UnicodeString& regionDisplayName(const char* region, UnicodeString& result, UBool skipAdjust) const;
    UnicodeString& variantDisplayName(const char* variant, UnicodeString& result, UBool skipAdjust) const;
    UnicodeString& keyDisplayName(const char* key, UnicodeString& result, UBool skipAdjust) const;
    UnicodeString& keyValueDisplayName(const char* key, const char* value,
                                        UnicodeString& result, UBool skipAdjust) const;
    void initialize(void);

    struct CapitalizationContextSink;
};

LocaleDisplayNamesImpl::LocaleDisplayNamesImpl(const Locale& locale,
                                               UDialectHandling dialectHandling)
    : dialectHandling(dialectHandling)
    , langData(U_ICUDATA_LANG, locale)
    , regionData(U_ICUDATA_REGION, locale)
    , capitalizationContext(UDISPCTX_CAPITALIZATION_NONE)
    , capitalizationBrkIter(NULL)
    , nameLength(UDISPCTX_LENGTH_FULL)
    , substitute(UDISPCTX_SUBSTITUTE)
{
    initialize();
}

LocaleDisplayNamesImpl::LocaleDisplayNamesImpl(const Locale& locale,
                                               UDisplayContext *contexts, int32_t length)
    : dialectHandling(ULDN_STANDARD_NAMES)
    , langData(U_ICUDATA_LANG, locale)
    , regionData(U_ICUDATA_REGION, locale)
    , capitalizationContext(UDISPCTX_CAPITALIZATION_NONE)
    , capitalizationBrkIter(NULL)
    , nameLength(UDISPCTX_LENGTH_FULL)
    , substitute(UDISPCTX_SUBSTITUTE)
{
    while (length-- > 0) {
        UDisplayContext value = *contexts++;
        UDisplayContextType selector = (UDisplayContextType)((uint32_t)value >> 8);
        switch (selector) {
            case UDISPCTX_TYPE_DIALECT_HANDLING:
                dialectHandling = (UDialectHandling)value;
                break;
            case UDISPCTX_TYPE_CAPITALIZATION:
                capitalizationContext = value;
                break;
            case UDISPCTX_TYPE_DISPLAY_LENGTH:
                nameLength = value;
                break;
            case UDISPCTX_TYPE_SUBSTITUTE_HANDLING:
                substitute = value;
                break;
            default:
                break;
        }
    }
    initialize();
}

struct LocaleDisplayNamesImpl::CapitalizationContextSink : public ResourceSink {
    UBool hasCapitalizationUsage;
    LocaleDisplayNamesImpl& parent;

    CapitalizationContextSink(LocaleDisplayNamesImpl& _parent)
      : hasCapitalizationUsage(FALSE), parent(_parent) {}
    virtual ~CapitalizationContextSink();

    virtual void put(const char *key, ResourceValue &value, UBool /*noFallback*/,
            UErrorCode &errorCode) {
        ResourceTable contexts = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }
        for (int i = 0; contexts.getKeyAndValue(i, key, value); ++i) {

            CapContextUsage usageEnum;
            if (uprv_strcmp(key, "key") == 0) {
                usageEnum = kCapContextUsageKey;
            } else if (uprv_strcmp(key, "keyValue") == 0) {
                usageEnum = kCapContextUsageKeyValue;
            } else if (uprv_strcmp(key, "languages") == 0) {
                usageEnum = kCapContextUsageLanguage;
            } else if (uprv_strcmp(key, "script") == 0) {
                usageEnum = kCapContextUsageScript;
            } else if (uprv_strcmp(key, "territory") == 0) {
                usageEnum = kCapContextUsageTerritory;
            } else if (uprv_strcmp(key, "variant") == 0) {
                usageEnum = kCapContextUsageVariant;
            } else {
                continue;
            }

            int32_t len = 0;
            const int32_t* intVector = value.getIntVector(len, errorCode);
            if (U_FAILURE(errorCode)) { return; }
            if (len < 2) { continue; }

            int32_t titlecaseInt = (parent.capitalizationContext == UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU) ? intVector[0] : intVector[1];
            if (titlecaseInt == 0) { continue; }

            parent.fCapitalization[usageEnum] = TRUE;
            hasCapitalizationUsage = TRUE;
        }
    }
};

// Virtual destructors must be defined out of line.
LocaleDisplayNamesImpl::CapitalizationContextSink::~CapitalizationContextSink() {}

void
LocaleDisplayNamesImpl::initialize(void) {
    LocaleDisplayNamesImpl *nonConstThis = (LocaleDisplayNamesImpl *)this;
    nonConstThis->locale = langData.getLocale() == Locale::getRoot()
        ? regionData.getLocale()
        : langData.getLocale();

    UnicodeString sep;
    langData.getNoFallback("localeDisplayPattern", "separator", sep);
    if (sep.isBogus()) {
        sep = UnicodeString("{0}, {1}", -1, US_INV);
    }
    UErrorCode status = U_ZERO_ERROR;
    separatorFormat.applyPatternMinMaxArguments(sep, 2, 2, status);

    UnicodeString pattern;
    langData.getNoFallback("localeDisplayPattern", "pattern", pattern);
    if (pattern.isBogus()) {
        pattern = UnicodeString("{0} ({1})", -1, US_INV);
    }
    format.applyPatternMinMaxArguments(pattern, 2, 2, status);
    if (pattern.indexOf((UChar)0xFF08) >= 0) {
        formatOpenParen.setTo((UChar)0xFF08);         // fullwidth (
        formatReplaceOpenParen.setTo((UChar)0xFF3B);  // fullwidth [
        formatCloseParen.setTo((UChar)0xFF09);        // fullwidth )
        formatReplaceCloseParen.setTo((UChar)0xFF3D); // fullwidth ]
    } else {
        formatOpenParen.setTo((UChar)0x0028);         // (
        formatReplaceOpenParen.setTo((UChar)0x005B);  // [
        formatCloseParen.setTo((UChar)0x0029);        // )
        formatReplaceCloseParen.setTo((UChar)0x005D); // ]
    }

    UnicodeString ktPattern;
    langData.get("localeDisplayPattern", "keyTypePattern", ktPattern);
    if (ktPattern.isBogus()) {
        ktPattern = UnicodeString("{0}={1}", -1, US_INV);
    }
    keyTypeFormat.applyPatternMinMaxArguments(ktPattern, 2, 2, status);

    uprv_memset(fCapitalization, 0, sizeof(fCapitalization));
#if !UCONFIG_NO_BREAK_ITERATION
    // Only get the context data if we need it! This is a const object so we know now...
    // Also check whether we will need a break iterator (depends on the data)
    UBool needBrkIter = FALSE;
    if (capitalizationContext == UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU || capitalizationContext == UDISPCTX_CAPITALIZATION_FOR_STANDALONE) {
        LocalUResourceBundlePointer resource(ures_open(NULL, locale.getName(), &status));
        if (U_FAILURE(status)) { return; }
        CapitalizationContextSink sink(*this);
        ures_getAllItemsWithFallback(resource.getAlias(), "contextTransforms", sink, status);
        if (status == U_MISSING_RESOURCE_ERROR) {
            // Silently ignore.  Not every locale has contextTransforms.
            status = U_ZERO_ERROR;
        } else if (U_FAILURE(status)) {
            return;
        }
        needBrkIter = sink.hasCapitalizationUsage;
    }
    // Get a sentence break iterator if we will need it
    if (needBrkIter || capitalizationContext == UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE) {
        status = U_ZERO_ERROR;
        capitalizationBrkIter = BreakIterator::createSentenceInstance(locale, status);
        if (U_FAILURE(status)) {
            delete capitalizationBrkIter;
            capitalizationBrkIter = NULL;
        }
    }
#endif
}

LocaleDisplayNamesImpl::~LocaleDisplayNamesImpl() {
#if !UCONFIG_NO_BREAK_ITERATION
    delete capitalizationBrkIter;
#endif
}

const Locale&
LocaleDisplayNamesImpl::getLocale() const {
    return locale;
}

UDialectHandling
LocaleDisplayNamesImpl::getDialectHandling() const {
    return dialectHandling;
}

UDisplayContext
LocaleDisplayNamesImpl::getContext(UDisplayContextType type) const {
    switch (type) {
        case UDISPCTX_TYPE_DIALECT_HANDLING:
            return (UDisplayContext)dialectHandling;
        case UDISPCTX_TYPE_CAPITALIZATION:
            return capitalizationContext;
        case UDISPCTX_TYPE_DISPLAY_LENGTH:
            return nameLength;
        case UDISPCTX_TYPE_SUBSTITUTE_HANDLING:
            return substitute;
        default:
            break;
    }
    return (UDisplayContext)0;
}

UnicodeString&
LocaleDisplayNamesImpl::adjustForUsageAndContext(CapContextUsage usage,
                                                UnicodeString& result) const {
#if !UCONFIG_NO_BREAK_ITERATION
    // check to see whether we need to titlecase result
    if ( result.length() > 0 && u_islower(result.char32At(0)) && capitalizationBrkIter!= NULL &&
          ( capitalizationContext==UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE || fCapitalization[usage] ) ) {
        // note fCapitalization[usage] won't be set unless capitalizationContext is UI_LIST_OR_MENU or STANDALONE
        static UMutex capitalizationBrkIterLock;
        Mutex lock(&capitalizationBrkIterLock);
        result.toTitle(capitalizationBrkIter, locale, U_TITLECASE_NO_LOWERCASE | U_TITLECASE_NO_BREAK_ADJUSTMENT);
    }
#endif
    return result;
}

UnicodeString&
LocaleDisplayNamesImpl::localeDisplayName(const Locale& loc,
                                          UnicodeString& result) const {
  if (loc.isBogus()) {
    result.setToBogus();
    return result;
  }
  UnicodeString resultName;

  const char* lang = loc.getLanguage();
  if (uprv_strlen(lang) == 0) {
    lang = "root";
  }
  const char* script = loc.getScript();
  const char* country = loc.getCountry();
  const char* variant = loc.getVariant();

  UBool hasScript = uprv_strlen(script) > 0;
  UBool hasCountry = uprv_strlen(country) > 0;
  UBool hasVariant = uprv_strlen(variant) > 0;

  if (dialectHandling == ULDN_DIALECT_NAMES) {
    char buffer[ULOC_FULLNAME_CAPACITY];
    do { // loop construct is so we can break early out of search
      if (hasScript && hasCountry) {
        ncat(buffer, ULOC_FULLNAME_CAPACITY, lang, "_", script, "_", country, (char *)0);
        localeIdName(buffer, resultName, false);
        if (!resultName.isBogus()) {
          hasScript = FALSE;
          hasCountry = FALSE;
          break;
        }
      }
      if (hasScript) {
        ncat(buffer, ULOC_FULLNAME_CAPACITY, lang, "_", script, (char *)0);
        localeIdName(buffer, resultName, false);
        if (!resultName.isBogus()) {
          hasScript = FALSE;
          break;
        }
      }
      if (hasCountry) {
        ncat(buffer, ULOC_FULLNAME_CAPACITY, lang, "_", country, (char*)0);
        localeIdName(buffer, resultName, false);
        if (!resultName.isBogus()) {
          hasCountry = FALSE;
          break;
        }
      }
    } while (FALSE);
  }
  if (resultName.isBogus() || resultName.isEmpty()) {
    localeIdName(lang, resultName, substitute == UDISPCTX_SUBSTITUTE);
    if (resultName.isBogus()) {
      result.setToBogus();
      return result;
    }
  }

  UnicodeString resultRemainder;
  UnicodeString temp;
  UErrorCode status = U_ZERO_ERROR;

  if (hasScript) {
    UnicodeString script_str = scriptDisplayName(script, temp, TRUE);
    if (script_str.isBogus()) {
      result.setToBogus();
      return result;
    }
    resultRemainder.append(script_str);
  }
  if (hasCountry) {
    UnicodeString region_str = regionDisplayName(country, temp, TRUE);
    if (region_str.isBogus()) {
      result.setToBogus();
      return result;
    }
    appendWithSep(resultRemainder, region_str);
  }
  if (hasVariant) {
    UnicodeString variant_str = variantDisplayName(variant, temp, TRUE);
    if (variant_str.isBogus()) {
      result.setToBogus();
      return result;
    }
    appendWithSep(resultRemainder, variant_str);
  }
  resultRemainder.findAndReplace(formatOpenParen, formatReplaceOpenParen);
  resultRemainder.findAndReplace(formatCloseParen, formatReplaceCloseParen);

  LocalPointer<StringEnumeration> e(loc.createKeywords(status));
  if (e.isValid() && U_SUCCESS(status)) {
    UnicodeString temp2;
    char value[ULOC_KEYWORD_AND_VALUES_CAPACITY]; // sigh, no ULOC_VALUE_CAPACITY
    const char* key;
    while ((key = e->next((int32_t *)0, status)) != NULL) {
      value[0] = 0;
      loc.getKeywordValue(key, value, ULOC_KEYWORD_AND_VALUES_CAPACITY, status);
      if (U_FAILURE(status) || status == U_STRING_NOT_TERMINATED_WARNING) {
        return result;
      }
      keyDisplayName(key, temp, TRUE);
      temp.findAndReplace(formatOpenParen, formatReplaceOpenParen);
      temp.findAndReplace(formatCloseParen, formatReplaceCloseParen);
      keyValueDisplayName(key, value, temp2, TRUE);
      temp2.findAndReplace(formatOpenParen, formatReplaceOpenParen);
      temp2.findAndReplace(formatCloseParen, formatReplaceCloseParen);
      if (temp2 != UnicodeString(value, -1, US_INV)) {
        appendWithSep(resultRemainder, temp2);
      } else if (temp != UnicodeString(key, -1, US_INV)) {
        UnicodeString temp3;
        keyTypeFormat.format(temp, temp2, temp3, status);
        appendWithSep(resultRemainder, temp3);
      } else {
        appendWithSep(resultRemainder, temp)
          .append((UChar)0x3d /* = */)
          .append(temp2);
      }
    }
  }

  if (!resultRemainder.isEmpty()) {
    format.format(resultName, resultRemainder, result.remove(), status);
    return adjustForUsageAndContext(kCapContextUsageLanguage, result);
  }

  result = resultName;
  return adjustForUsageAndContext(kCapContextUsageLanguage, result);
}

UnicodeString&
LocaleDisplayNamesImpl::appendWithSep(UnicodeString& buffer, const UnicodeString& src) const {
    if (buffer.isEmpty()) {
        buffer.setTo(src);
    } else {
        const UnicodeString *values[2] = { &buffer, &src };
        UErrorCode status = U_ZERO_ERROR;
        separatorFormat.formatAndReplace(values, 2, buffer, NULL, 0, status);
    }
    return buffer;
}

UnicodeString&
LocaleDisplayNamesImpl::localeDisplayName(const char* localeId,
                                          UnicodeString& result) const {
    return localeDisplayName(Locale(localeId), result);
}

// private
UnicodeString&
LocaleDisplayNamesImpl::localeIdName(const char* localeId,
                                     UnicodeString& result, bool substitute) const {
    if (nameLength == UDISPCTX_LENGTH_SHORT) {
        langData.getNoFallback("Languages%short", localeId, result);
        if (!result.isBogus()) {
            return result;
        }
    }
    if (substitute) {
        return langData.get("Languages", localeId, result);
    } else {
        return langData.getNoFallback("Languages", localeId, result);
    }
}

UnicodeString&
LocaleDisplayNamesImpl::languageDisplayName(const char* lang,
                                            UnicodeString& result) const {
    if (uprv_strcmp("root", lang) == 0 || uprv_strchr(lang, '_') != NULL) {
        return result = UnicodeString(lang, -1, US_INV);
    }
    if (nameLength == UDISPCTX_LENGTH_SHORT) {
        langData.getNoFallback("Languages%short", lang, result);
        if (!result.isBogus()) {
            return adjustForUsageAndContext(kCapContextUsageLanguage, result);
        }
    }
    if (substitute == UDISPCTX_SUBSTITUTE) {
        langData.get("Languages", lang, result);
    } else {
        langData.getNoFallback("Languages", lang, result);
    }
    return adjustForUsageAndContext(kCapContextUsageLanguage, result);
}

UnicodeString&
LocaleDisplayNamesImpl::scriptDisplayName(const char* script,
                                          UnicodeString& result,
                                          UBool skipAdjust) const {
    if (nameLength == UDISPCTX_LENGTH_SHORT) {
        langData.getNoFallback("Scripts%short", script, result);
        if (!result.isBogus()) {
            return skipAdjust? result: adjustForUsageAndContext(kCapContextUsageScript, result);
        }
    }
    if (substitute == UDISPCTX_SUBSTITUTE) {
        langData.get("Scripts", script, result);
    } else {
        langData.getNoFallback("Scripts", script, result);
    }
    return skipAdjust? result: adjustForUsageAndContext(kCapContextUsageScript, result);
}

UnicodeString&
LocaleDisplayNamesImpl::scriptDisplayName(const char* script,
                                          UnicodeString& result) const {
    return scriptDisplayName(script, result, FALSE);
}

UnicodeString&
LocaleDisplayNamesImpl::scriptDisplayName(UScriptCode scriptCode,
                                          UnicodeString& result) const {
    return scriptDisplayName(uscript_getName(scriptCode), result, FALSE);
}

UnicodeString&
LocaleDisplayNamesImpl::regionDisplayName(const char* region,
                                          UnicodeString& result,
                                          UBool skipAdjust) const {
    if (nameLength == UDISPCTX_LENGTH_SHORT) {
         regionData.getNoFallback("Countries%short", region, result);
        if (!result.isBogus()) {
            return skipAdjust? result: adjustForUsageAndContext(kCapContextUsageTerritory, result);
        }
    }
    if (substitute == UDISPCTX_SUBSTITUTE) {
        regionData.get("Countries", region, result);
    } else {
        regionData.getNoFallback("Countries", region, result);
    }
    return skipAdjust? result: adjustForUsageAndContext(kCapContextUsageTerritory, result);
}

UnicodeString&
LocaleDisplayNamesImpl::regionDisplayName(const char* region,
                                          UnicodeString& result) const {
    return regionDisplayName(region, result, FALSE);
}


UnicodeString&
LocaleDisplayNamesImpl::variantDisplayName(const char* variant,
                                           UnicodeString& result,
                                           UBool skipAdjust) const {
    // don't have a resource for short variant names
    if (substitute == UDISPCTX_SUBSTITUTE) {
        langData.get("Variants", variant, result);
    } else {
        langData.getNoFallback("Variants", variant, result);
    }
    return skipAdjust? result: adjustForUsageAndContext(kCapContextUsageVariant, result);
}

UnicodeString&
LocaleDisplayNamesImpl::variantDisplayName(const char* variant,
                                           UnicodeString& result) const {
    return variantDisplayName(variant, result, FALSE);
}

UnicodeString&
LocaleDisplayNamesImpl::keyDisplayName(const char* key,
                                       UnicodeString& result,
                                       UBool skipAdjust) const {
    // don't have a resource for short key names
    if (substitute == UDISPCTX_SUBSTITUTE) {
        langData.get("Keys", key, result);
    } else {
        langData.getNoFallback("Keys", key, result);
    }
    return skipAdjust? result: adjustForUsageAndContext(kCapContextUsageKey, result);
}

UnicodeString&
LocaleDisplayNamesImpl::keyDisplayName(const char* key,
                                       UnicodeString& result) const {
    return keyDisplayName(key, result, FALSE);
}

UnicodeString&
LocaleDisplayNamesImpl::keyValueDisplayName(const char* key,
                                            const char* value,
                                            UnicodeString& result,
                                            UBool skipAdjust) const {
    if (uprv_strcmp(key, "currency") == 0) {
        // ICU4C does not have ICU4J CurrencyDisplayInfo equivalent for now.
        UErrorCode sts = U_ZERO_ERROR;
        UnicodeString ustrValue(value, -1, US_INV);
        int32_t len;
        const UChar *currencyName = ucurr_getName(ustrValue.getTerminatedBuffer(),
            locale.getBaseName(), UCURR_LONG_NAME, nullptr /* isChoiceFormat */, &len, &sts);
        if (U_FAILURE(sts)) {
            // Return the value as is on failure
            result = ustrValue;
            return result;
        }
        result.setTo(currencyName, len);
        return skipAdjust? result: adjustForUsageAndContext(kCapContextUsageKeyValue, result);
    }

    if (nameLength == UDISPCTX_LENGTH_SHORT) {
        langData.getNoFallback("Types%short", key, value, result);
        if (!result.isBogus()) {
            return skipAdjust? result: adjustForUsageAndContext(kCapContextUsageKeyValue, result);
        }
    }
    if (substitute == UDISPCTX_SUBSTITUTE) {
        langData.get("Types", key, value, result);
    } else {
        langData.getNoFallback("Types", key, value, result);
    }
    return skipAdjust? result: adjustForUsageAndContext(kCapContextUsageKeyValue, result);
}

UnicodeString&
LocaleDisplayNamesImpl::keyValueDisplayName(const char* key,
                                            const char* value,
                                            UnicodeString& result) const {
    return keyValueDisplayName(key, value, result, FALSE);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
LocaleDisplayNames*
LocaleDisplayNames::createInstance(const Locale& locale,
                                   UDialectHandling dialectHandling) {
    return new LocaleDisplayNamesImpl(locale, dialectHandling);
}

LocaleDisplayNames*
LocaleDisplayNames::createInstance(const Locale& locale,
                                   UDisplayContext *contexts, int32_t length) {
    if (contexts == NULL) {
        length = 0;
    }
    return new LocaleDisplayNamesImpl(locale, contexts, length);
}

U_NAMESPACE_END

////////////////////////////////////////////////////////////////////////////////////////////////////

U_NAMESPACE_USE

U_CAPI ULocaleDisplayNames * U_EXPORT2
uldn_open(const char * locale,
          UDialectHandling dialectHandling,
          UErrorCode *pErrorCode) {
  if (U_FAILURE(*pErrorCode)) {
    return 0;
  }
  if (locale == NULL) {
    locale = uloc_getDefault();
  }
  return (ULocaleDisplayNames *)LocaleDisplayNames::createInstance(Locale(locale), dialectHandling);
}

U_CAPI ULocaleDisplayNames * U_EXPORT2
uldn_openForContext(const char * locale,
                    UDisplayContext *contexts, int32_t length,
                    UErrorCode *pErrorCode) {
  if (U_FAILURE(*pErrorCode)) {
    return 0;
  }
  if (locale == NULL) {
    locale = uloc_getDefault();
  }
  return (ULocaleDisplayNames *)LocaleDisplayNames::createInstance(Locale(locale), contexts, length);
}


U_CAPI void U_EXPORT2
uldn_close(ULocaleDisplayNames *ldn) {
  delete (LocaleDisplayNames *)ldn;
}

U_CAPI const char * U_EXPORT2
uldn_getLocale(const ULocaleDisplayNames *ldn) {
  if (ldn) {
    return ((const LocaleDisplayNames *)ldn)->getLocale().getName();
  }
  return NULL;
}

U_CAPI UDialectHandling U_EXPORT2
uldn_getDialectHandling(const ULocaleDisplayNames *ldn) {
  if (ldn) {
    return ((const LocaleDisplayNames *)ldn)->getDialectHandling();
  }
  return ULDN_STANDARD_NAMES;
}

U_CAPI UDisplayContext U_EXPORT2
uldn_getContext(const ULocaleDisplayNames *ldn,
              UDisplayContextType type,
              UErrorCode *pErrorCode) {
  if (U_FAILURE(*pErrorCode)) {
    return (UDisplayContext)0;
  }
  return ((const LocaleDisplayNames *)ldn)->getContext(type);
}

U_CAPI int32_t U_EXPORT2
uldn_localeDisplayName(const ULocaleDisplayNames *ldn,
                       const char *locale,
                       UChar *result,
                       int32_t maxResultSize,
                       UErrorCode *pErrorCode) {
  if (U_FAILURE(*pErrorCode)) {
    return 0;
  }
  if (ldn == NULL || locale == NULL || (result == NULL && maxResultSize > 0) || maxResultSize < 0) {
    *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
    return 0;
  }
  UnicodeString temp(result, 0, maxResultSize);
  ((const LocaleDisplayNames *)ldn)->localeDisplayName(locale, temp);
  if (temp.isBogus()) {
    *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
    return 0;
  }
  return temp.extract(result, maxResultSize, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uldn_languageDisplayName(const ULocaleDisplayNames *ldn,
                         const char *lang,
                         UChar *result,
                         int32_t maxResultSize,
                         UErrorCode *pErrorCode) {
  if (U_FAILURE(*pErrorCode)) {
    return 0;
  }
  if (ldn == NULL || lang == NULL || (result == NULL && maxResultSize > 0) || maxResultSize < 0) {
    *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
    return 0;
  }
  UnicodeString temp(result, 0, maxResultSize);
  ((const LocaleDisplayNames *)ldn)->languageDisplayName(lang, temp);
  return temp.extract(result, maxResultSize, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uldn_scriptDisplayName(const ULocaleDisplayNames *ldn,
                       const char *script,
                       UChar *result,
                       int32_t maxResultSize,
                       UErrorCode *pErrorCode) {
  if (U_FAILURE(*pErrorCode)) {
    return 0;
  }
  if (ldn == NULL || script == NULL || (result == NULL && maxResultSize > 0) || maxResultSize < 0) {
    *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
    return 0;
  }
  UnicodeString temp(result, 0, maxResultSize);
  ((const LocaleDisplayNames *)ldn)->scriptDisplayName(script, temp);
  return temp.extract(result, maxResultSize, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uldn_scriptCodeDisplayName(const ULocaleDisplayNames *ldn,
                           UScriptCode scriptCode,
                           UChar *result,
                           int32_t maxResultSize,
                           UErrorCode *pErrorCode) {
  return uldn_scriptDisplayName(ldn, uscript_getName(scriptCode), result, maxResultSize, pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uldn_regionDisplayName(const ULocaleDisplayNames *ldn,
                       const char *region,
                       UChar *result,
                       int32_t maxResultSize,
                       UErrorCode *pErrorCode) {
  if (U_FAILURE(*pErrorCode)) {
    return 0;
  }
  if (ldn == NULL || region == NULL || (result == NULL && maxResultSize > 0) || maxResultSize < 0) {
    *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
    return 0;
  }
  UnicodeString temp(result, 0, maxResultSize);
  ((const LocaleDisplayNames *)ldn)->regionDisplayName(region, temp);
  return temp.extract(result, maxResultSize, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uldn_variantDisplayName(const ULocaleDisplayNames *ldn,
                        const char *variant,
                        UChar *result,
                        int32_t maxResultSize,
                        UErrorCode *pErrorCode) {
  if (U_FAILURE(*pErrorCode)) {
    return 0;
  }
  if (ldn == NULL || variant == NULL || (result == NULL && maxResultSize > 0) || maxResultSize < 0) {
    *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
    return 0;
  }
  UnicodeString temp(result, 0, maxResultSize);
  ((const LocaleDisplayNames *)ldn)->variantDisplayName(variant, temp);
  return temp.extract(result, maxResultSize, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uldn_keyDisplayName(const ULocaleDisplayNames *ldn,
                    const char *key,
                    UChar *result,
                    int32_t maxResultSize,
                    UErrorCode *pErrorCode) {
  if (U_FAILURE(*pErrorCode)) {
    return 0;
  }
  if (ldn == NULL || key == NULL || (result == NULL && maxResultSize > 0) || maxResultSize < 0) {
    *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
    return 0;
  }
  UnicodeString temp(result, 0, maxResultSize);
  ((const LocaleDisplayNames *)ldn)->keyDisplayName(key, temp);
  return temp.extract(result, maxResultSize, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uldn_keyValueDisplayName(const ULocaleDisplayNames *ldn,
                         const char *key,
                         const char *value,
                         UChar *result,
                         int32_t maxResultSize,
                         UErrorCode *pErrorCode) {
  if (U_FAILURE(*pErrorCode)) {
    return 0;
  }
  if (ldn == NULL || key == NULL || value == NULL || (result == NULL && maxResultSize > 0)
      || maxResultSize < 0) {
    *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
    return 0;
  }
  UnicodeString temp(result, 0, maxResultSize);
  ((const LocaleDisplayNames *)ldn)->keyValueDisplayName(key, value, temp);
  return temp.extract(result, maxResultSize, *pErrorCode);
}

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 **********************************************************************
 *   Copyright (C) 1997-2016, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 **********************************************************************
*
* File locid.cpp
*
* Created by: Richard Gillam
*
* Modification History:
*
*   Date        Name        Description
*   02/11/97    aliu        Changed gLocPath to fgDataDirectory and added
*                           methods to get and set it.
*   04/02/97    aliu        Made operator!= inline; fixed return value
*                           of getName().
*   04/15/97    aliu        Cleanup for AIX/Win32.
*   04/24/97    aliu        Numerous changes per code review.
*   08/18/98    stephen     Changed getDisplayName()
*                           Added SIMPLIFIED_CHINESE, TRADITIONAL_CHINESE
*                           Added getISOCountries(), getISOLanguages(),
*                           getLanguagesForCountry()
*   03/16/99    bertrand    rehaul.
*   07/21/99    stephen     Added U_CFUNC setDefault
*   11/09/99    weiv        Added const char * getName() const;
*   04/12/00    srl         removing unicodestring api's and cached hash code
*   08/10/01    grhoten     Change the static Locales to accessor functions
******************************************************************************
*/

#include <utility>




















U_CDECL_BEGIN
static UBool U_CALLCONV locale_cleanup(void);
U_CDECL_END

U_NAMESPACE_BEGIN

static Locale   *gLocaleCache = NULL;
static UInitOnce gLocaleCacheInitOnce = U_INITONCE_INITIALIZER;

// gDefaultLocaleMutex protects all access to gDefaultLocalesHashT and gDefaultLocale.
static UMutex gDefaultLocaleMutex;
static UHashtable *gDefaultLocalesHashT = NULL;
static Locale *gDefaultLocale = NULL;

/**
 * \def ULOC_STRING_LIMIT
 * strings beyond this value crash in CharString
 */
#define ULOC_STRING_LIMIT 357913941

U_NAMESPACE_END

typedef enum ELocalePos {
    eENGLISH,
    eFRENCH,
    eGERMAN,
    eITALIAN,
    eJAPANESE,
    eKOREAN,
    eCHINESE,

    eFRANCE,
    eGERMANY,
    eITALY,
    eJAPAN,
    eKOREA,
    eCHINA,      /* Alias for PRC */
    eTAIWAN,
    eUK,
    eUS,
    eCANADA,
    eCANADA_FRENCH,
    eROOT,


    //eDEFAULT,
    eMAX_LOCALES
} ELocalePos;

U_CFUNC int32_t locale_getKeywords(const char *localeID,
            char prev,
            char *keywords, int32_t keywordCapacity,
            char *values, int32_t valuesCapacity, int32_t *valLen,
            UBool valuesToo,
            UErrorCode *status);

U_CDECL_BEGIN
//
// Deleter function for Locales owned by the default Locale hash table/
//
static void U_CALLCONV
deleteLocale(void *obj) {
    delete (icu::Locale *) obj;
}

static UBool U_CALLCONV locale_cleanup(void)
{
    U_NAMESPACE_USE

    delete [] gLocaleCache;
    gLocaleCache = NULL;
    gLocaleCacheInitOnce.reset();

    if (gDefaultLocalesHashT) {
        uhash_close(gDefaultLocalesHashT);   // Automatically deletes all elements, using deleter func.
        gDefaultLocalesHashT = NULL;
    }
    gDefaultLocale = NULL;
    return TRUE;
}


static void U_CALLCONV locale_init(UErrorCode &status) {
    U_NAMESPACE_USE

    U_ASSERT(gLocaleCache == NULL);
    gLocaleCache = new Locale[(int)eMAX_LOCALES];
    if (gLocaleCache == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    ucln_common_registerCleanup(UCLN_COMMON_LOCALE, locale_cleanup);
    gLocaleCache[eROOT]          = Locale("");
    gLocaleCache[eENGLISH]       = Locale("en");
    gLocaleCache[eFRENCH]        = Locale("fr");
    gLocaleCache[eGERMAN]        = Locale("de");
    gLocaleCache[eITALIAN]       = Locale("it");
    gLocaleCache[eJAPANESE]      = Locale("ja");
    gLocaleCache[eKOREAN]        = Locale("ko");
    gLocaleCache[eCHINESE]       = Locale("zh");
    gLocaleCache[eFRANCE]        = Locale("fr", "FR");
    gLocaleCache[eGERMANY]       = Locale("de", "DE");
    gLocaleCache[eITALY]         = Locale("it", "IT");
    gLocaleCache[eJAPAN]         = Locale("ja", "JP");
    gLocaleCache[eKOREA]         = Locale("ko", "KR");
    gLocaleCache[eCHINA]         = Locale("zh", "CN");
    gLocaleCache[eTAIWAN]        = Locale("zh", "TW");
    gLocaleCache[eUK]            = Locale("en", "GB");
    gLocaleCache[eUS]            = Locale("en", "US");
    gLocaleCache[eCANADA]        = Locale("en", "CA");
    gLocaleCache[eCANADA_FRENCH] = Locale("fr", "CA");
}

U_CDECL_END

U_NAMESPACE_BEGIN

Locale *locale_set_default_internal(const char *id, UErrorCode& status) {
    // Synchronize this entire function.
    Mutex lock(&gDefaultLocaleMutex);

    UBool canonicalize = FALSE;

    // If given a NULL string for the locale id, grab the default
    //   name from the system.
    //   (Different from most other locale APIs, where a null name means use
    //    the current ICU default locale.)
    if (id == NULL) {
        id = uprv_getDefaultLocaleID();   // This function not thread safe? TODO: verify.
        canonicalize = TRUE; // always canonicalize host ID
    }

    char localeNameBuf[512];

    if (canonicalize) {
        uloc_canonicalize(id, localeNameBuf, sizeof(localeNameBuf)-1, &status);
    } else {
        uloc_getName(id, localeNameBuf, sizeof(localeNameBuf)-1, &status);
    }
    localeNameBuf[sizeof(localeNameBuf)-1] = 0;  // Force null termination in event of
                                                 //   a long name filling the buffer.
                                                 //   (long names are truncated.)
                                                 //
    if (U_FAILURE(status)) {
        return gDefaultLocale;
    }

    if (gDefaultLocalesHashT == NULL) {
        gDefaultLocalesHashT = uhash_open(uhash_hashChars, uhash_compareChars, NULL, &status);
        if (U_FAILURE(status)) {
            return gDefaultLocale;
        }
        uhash_setValueDeleter(gDefaultLocalesHashT, deleteLocale);
        ucln_common_registerCleanup(UCLN_COMMON_LOCALE, locale_cleanup);
    }

    Locale *newDefault = (Locale *)uhash_get(gDefaultLocalesHashT, localeNameBuf);
    if (newDefault == NULL) {
        newDefault = new Locale(Locale::eBOGUS);
        if (newDefault == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return gDefaultLocale;
        }
        newDefault->init(localeNameBuf, FALSE);
        uhash_put(gDefaultLocalesHashT, (char*) newDefault->getName(), newDefault, &status);
        if (U_FAILURE(status)) {
            return gDefaultLocale;
        }
    }
    gDefaultLocale = newDefault;
    return gDefaultLocale;
}

U_NAMESPACE_END

/* sfb 07/21/99 */
U_CFUNC void
locale_set_default(const char *id)
{
    U_NAMESPACE_USE
    UErrorCode status = U_ZERO_ERROR;
    locale_set_default_internal(id, status);
}
/* end */

U_CFUNC const char *
locale_get_default(void)
{
    U_NAMESPACE_USE
    return Locale::getDefault().getName();
}


U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(Locale)

/*Character separating the posix id fields*/
// '_'
// In the platform codepage.
#define SEP_CHAR '_'

Locale::~Locale()
{
    if (baseName != fullName) {
        uprv_free(baseName);
    }
    baseName = NULL;
    /*if fullName is on the heap, we free it*/
    if (fullName != fullNameBuffer)
    {
        uprv_free(fullName);
        fullName = NULL;
    }
}

Locale::Locale()
    : UObject(), fullName(fullNameBuffer), baseName(NULL)
{
    init(NULL, FALSE);
}

/*
 * Internal constructor to allow construction of a locale object with
 *   NO side effects.   (Default constructor tries to get
 *   the default locale.)
 */
Locale::Locale(Locale::ELocaleType)
    : UObject(), fullName(fullNameBuffer), baseName(NULL)
{
    setToBogus();
}


Locale::Locale( const   char * newLanguage,
                const   char * newCountry,
                const   char * newVariant,
                const   char * newKeywords)
    : UObject(), fullName(fullNameBuffer), baseName(NULL)
{
    if( (newLanguage==NULL) && (newCountry == NULL) && (newVariant == NULL) )
    {
        init(NULL, FALSE); /* shortcut */
    }
    else
    {
        UErrorCode status = U_ZERO_ERROR;
        int32_t size = 0;
        int32_t lsize = 0;
        int32_t csize = 0;
        int32_t vsize = 0;
        int32_t ksize = 0;

        // Calculate the size of the resulting string.

        // Language
        if ( newLanguage != NULL )
        {
            lsize = (int32_t)uprv_strlen(newLanguage);
            if ( lsize < 0 || lsize > ULOC_STRING_LIMIT ) { // int32 wrap
                setToBogus();
                return;
            }
            size = lsize;
        }

        CharString togo(newLanguage, lsize, status); // start with newLanguage

        // _Country
        if ( newCountry != NULL )
        {
            csize = (int32_t)uprv_strlen(newCountry);
            if ( csize < 0 || csize > ULOC_STRING_LIMIT ) { // int32 wrap
                setToBogus();
                return;
            }
            size += csize;
        }

        // _Variant
        if ( newVariant != NULL )
        {
            // remove leading _'s
            while(newVariant[0] == SEP_CHAR)
            {
                newVariant++;
            }

            // remove trailing _'s
            vsize = (int32_t)uprv_strlen(newVariant);
            if ( vsize < 0 || vsize > ULOC_STRING_LIMIT ) { // int32 wrap
                setToBogus();
                return;
            }
            while( (vsize>1) && (newVariant[vsize-1] == SEP_CHAR) )
            {
                vsize--;
            }
        }

        if( vsize > 0 )
        {
            size += vsize;
        }

        // Separator rules:
        if ( vsize > 0 )
        {
            size += 2;  // at least: __v
        }
        else if ( csize > 0 )
        {
            size += 1;  // at least: _v
        }

        if ( newKeywords != NULL)
        {
            ksize = (int32_t)uprv_strlen(newKeywords);
            if ( ksize < 0 || ksize > ULOC_STRING_LIMIT ) {
              setToBogus();
              return;
            }
            size += ksize + 1;
        }

        //  NOW we have the full locale string..
        // Now, copy it back.

        // newLanguage is already copied

        if ( ( vsize != 0 ) || (csize != 0) )  // at least:  __v
        {                                      //            ^
            togo.append(SEP_CHAR, status);
        }

        if ( csize != 0 )
        {
            togo.append(newCountry, status);
        }

        if ( vsize != 0)
        {
            togo.append(SEP_CHAR, status)
                .append(newVariant, vsize, status);
        }

        if ( ksize != 0)
        {
            if (uprv_strchr(newKeywords, '=')) {
                togo.append('@', status); /* keyword parsing */
            }
            else {
                togo.append('_', status); /* Variant parsing with a script */
                if ( vsize == 0) {
                    togo.append('_', status); /* No country found */
                }
            }
            togo.append(newKeywords, status);
        }

        if (U_FAILURE(status)) {
            // Something went wrong with appending, etc.
            setToBogus();
            return;
        }
        // Parse it, because for example 'language' might really be a complete
        // string.
        init(togo.data(), FALSE);
    }
}

Locale::Locale(const Locale &other)
    : UObject(other), fullName(fullNameBuffer), baseName(NULL)
{
    *this = other;
}

Locale::Locale(Locale&& other) U_NOEXCEPT
    : UObject(other), fullName(fullNameBuffer), baseName(fullName) {
  *this = std::move(other);
}

Locale& Locale::operator=(const Locale& other) {
    if (this == &other) {
        return *this;
    }

    setToBogus();

    if (other.fullName == other.fullNameBuffer) {
        uprv_strcpy(fullNameBuffer, other.fullNameBuffer);
    } else if (other.fullName == nullptr) {
        fullName = nullptr;
    } else {
        fullName = uprv_strdup(other.fullName);
        if (fullName == nullptr) return *this;
    }

    if (other.baseName == other.fullName) {
        baseName = fullName;
    } else if (other.baseName != nullptr) {
        baseName = uprv_strdup(other.baseName);
        if (baseName == nullptr) return *this;
    }

    uprv_strcpy(language, other.language);
    uprv_strcpy(script, other.script);
    uprv_strcpy(country, other.country);

    variantBegin = other.variantBegin;
    fIsBogus = other.fIsBogus;

    return *this;
}

Locale& Locale::operator=(Locale&& other) U_NOEXCEPT {
    if (baseName != fullName) uprv_free(baseName);
    if (fullName != fullNameBuffer) uprv_free(fullName);

    if (other.fullName == other.fullNameBuffer) {
        uprv_strcpy(fullNameBuffer, other.fullNameBuffer);
        fullName = fullNameBuffer;
    } else {
        fullName = other.fullName;
    }

    if (other.baseName == other.fullName) {
        baseName = fullName;
    } else {
        baseName = other.baseName;
    }

    uprv_strcpy(language, other.language);
    uprv_strcpy(script, other.script);
    uprv_strcpy(country, other.country);

    variantBegin = other.variantBegin;
    fIsBogus = other.fIsBogus;

    other.baseName = other.fullName = other.fullNameBuffer;

    return *this;
}

Locale *
Locale::clone() const {
    return new Locale(*this);
}

UBool
Locale::operator==( const   Locale& other) const
{
    return (uprv_strcmp(other.fullName, fullName) == 0);
}

#define ISASCIIALPHA(c) (((c) >= 'a' && (c) <= 'z') || ((c) >= 'A' && (c) <= 'Z'))

/*This function initializes a Locale from a C locale ID*/
Locale& Locale::init(const char* localeID, UBool canonicalize)
{
    fIsBogus = FALSE;
    /* Free our current storage */
    if (baseName != fullName) {
        uprv_free(baseName);
    }
    baseName = NULL;
    if(fullName != fullNameBuffer) {
        uprv_free(fullName);
        fullName = fullNameBuffer;
    }

    // not a loop:
    // just an easy way to have a common error-exit
    // without goto and without another function
    do {
        char *separator;
        char *field[5] = {0};
        int32_t fieldLen[5] = {0};
        int32_t fieldIdx;
        int32_t variantField;
        int32_t length;
        UErrorCode err;

        if(localeID == NULL) {
            // not an error, just set the default locale
            return *this = getDefault();
        }

        /* preset all fields to empty */
        language[0] = script[0] = country[0] = 0;

        // "canonicalize" the locale ID to ICU/Java format
        err = U_ZERO_ERROR;
        length = canonicalize ?
            uloc_canonicalize(localeID, fullName, sizeof(fullNameBuffer), &err) :
            uloc_getName(localeID, fullName, sizeof(fullNameBuffer), &err);

        if(err == U_BUFFER_OVERFLOW_ERROR || length >= (int32_t)sizeof(fullNameBuffer)) {
            /*Go to heap for the fullName if necessary*/
            fullName = (char *)uprv_malloc(sizeof(char)*(length + 1));
            if(fullName == 0) {
                fullName = fullNameBuffer;
                break; // error: out of memory
            }
            err = U_ZERO_ERROR;
            length = canonicalize ?
                uloc_canonicalize(localeID, fullName, length+1, &err) :
                uloc_getName(localeID, fullName, length+1, &err);
        }
        if(U_FAILURE(err) || err == U_STRING_NOT_TERMINATED_WARNING) {
            /* should never occur */
            break;
        }

        variantBegin = length;

        /* after uloc_getName/canonicalize() we know that only '_' are separators */
        /* But _ could also appeared in timezone such as "en@timezone=America/Los_Angeles" */
        separator = field[0] = fullName;
        fieldIdx = 1;
        char* at = uprv_strchr(fullName, '@');
        while ((separator = uprv_strchr(field[fieldIdx-1], SEP_CHAR)) != 0 &&
               fieldIdx < UPRV_LENGTHOF(field)-1 &&
               (at == nullptr || separator < at)) {
            field[fieldIdx] = separator + 1;
            fieldLen[fieldIdx-1] = (int32_t)(separator - field[fieldIdx-1]);
            fieldIdx++;
        }
        // variant may contain @foo or .foo POSIX cruft; remove it
        separator = uprv_strchr(field[fieldIdx-1], '@');
        char* sep2 = uprv_strchr(field[fieldIdx-1], '.');
        if (separator!=NULL || sep2!=NULL) {
            if (separator==NULL || (sep2!=NULL && separator > sep2)) {
                separator = sep2;
            }
            fieldLen[fieldIdx-1] = (int32_t)(separator - field[fieldIdx-1]);
        } else {
            fieldLen[fieldIdx-1] = length - (int32_t)(field[fieldIdx-1] - fullName);
        }

        if (fieldLen[0] >= (int32_t)(sizeof(language)))
        {
            break; // error: the language field is too long
        }

        variantField = 1; /* Usually the 2nd one, except when a script or country is also used. */
        if (fieldLen[0] > 0) {
            /* We have a language */
            uprv_memcpy(language, fullName, fieldLen[0]);
            language[fieldLen[0]] = 0;
        }
        if (fieldLen[1] == 4 && ISASCIIALPHA(field[1][0]) &&
                ISASCIIALPHA(field[1][1]) && ISASCIIALPHA(field[1][2]) &&
                ISASCIIALPHA(field[1][3])) {
            /* We have at least a script */
            uprv_memcpy(script, field[1], fieldLen[1]);
            script[fieldLen[1]] = 0;
            variantField++;
        }

        if (fieldLen[variantField] == 2 || fieldLen[variantField] == 3) {
            /* We have a country */
            uprv_memcpy(country, field[variantField], fieldLen[variantField]);
            country[fieldLen[variantField]] = 0;
            variantField++;
        } else if (fieldLen[variantField] == 0) {
            variantField++; /* script or country empty but variant in next field (i.e. en__POSIX) */
        }

        if (fieldLen[variantField] > 0) {
            /* We have a variant */
            variantBegin = (int32_t)(field[variantField] - fullName);
        }

        err = U_ZERO_ERROR;
        initBaseName(err);
        if (U_FAILURE(err)) {
            break;
        }

        // successful end of init()
        return *this;
    } while(0); /*loop doesn't iterate*/

    // when an error occurs, then set this object to "bogus" (there is no UErrorCode here)
    setToBogus();

    return *this;
}

/*
 * Set up the base name.
 * If there are no key words, it's exactly the full name.
 * If key words exist, it's the full name truncated at the '@' character.
 * Need to set up both at init() and after setting a keyword.
 */
void
Locale::initBaseName(UErrorCode &status) {
    if (U_FAILURE(status)) {
        return;
    }
    U_ASSERT(baseName==NULL || baseName==fullName);
    const char *atPtr = uprv_strchr(fullName, '@');
    const char *eqPtr = uprv_strchr(fullName, '=');
    if (atPtr && eqPtr && atPtr < eqPtr) {
        // Key words exist.
        int32_t baseNameLength = (int32_t)(atPtr - fullName);
        baseName = (char *)uprv_malloc(baseNameLength + 1);
        if (baseName == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        uprv_strncpy(baseName, fullName, baseNameLength);
        baseName[baseNameLength] = 0;

        // The original computation of variantBegin leaves it equal to the length
        // of fullName if there is no variant.  It should instead be
        // the length of the baseName.
        if (variantBegin > baseNameLength) {
            variantBegin = baseNameLength;
        }
    } else {
        baseName = fullName;
    }
}


int32_t
Locale::hashCode() const
{
    return ustr_hashCharsN(fullName, static_cast<int32_t>(uprv_strlen(fullName)));
}

void
Locale::setToBogus() {
    /* Free our current storage */
    if(baseName != fullName) {
        uprv_free(baseName);
    }
    baseName = NULL;
    if(fullName != fullNameBuffer) {
        uprv_free(fullName);
        fullName = fullNameBuffer;
    }
    *fullNameBuffer = 0;
    *language = 0;
    *script = 0;
    *country = 0;
    fIsBogus = TRUE;
    variantBegin = 0;
}

const Locale& U_EXPORT2
Locale::getDefault()
{
    {
        Mutex lock(&gDefaultLocaleMutex);
        if (gDefaultLocale != NULL) {
            return *gDefaultLocale;
        }
    }
    UErrorCode status = U_ZERO_ERROR;
    return *locale_set_default_internal(NULL, status);
}



void U_EXPORT2
Locale::setDefault( const   Locale&     newLocale,
                            UErrorCode&  status)
{
    if (U_FAILURE(status)) {
        return;
    }

    /* Set the default from the full name string of the supplied locale.
     * This is a convenient way to access the default locale caching mechanisms.
     */
    const char *localeID = newLocale.getName();
    locale_set_default_internal(localeID, status);
}

void
Locale::addLikelySubtags(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }

    CharString maximizedLocaleID;
    {
        CharStringByteSink sink(&maximizedLocaleID);
        ulocimp_addLikelySubtags(fullName, sink, &status);
    }

    if (U_FAILURE(status)) {
        return;
    }

    init(maximizedLocaleID.data(), /*canonicalize=*/FALSE);
    if (isBogus()) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

void
Locale::minimizeSubtags(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }

    CharString minimizedLocaleID;
    {
        CharStringByteSink sink(&minimizedLocaleID);
        ulocimp_minimizeSubtags(fullName, sink, &status);
    }

    if (U_FAILURE(status)) {
        return;
    }

    init(minimizedLocaleID.data(), /*canonicalize=*/FALSE);
    if (isBogus()) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

Locale U_EXPORT2
Locale::forLanguageTag(StringPiece tag, UErrorCode& status)
{
    Locale result(Locale::eBOGUS);

    if (U_FAILURE(status)) {
        return result;
    }

    // If a BCP-47 language tag is passed as the language parameter to the
    // normal Locale constructor, it will actually fall back to invoking
    // uloc_forLanguageTag() to parse it if it somehow is able to detect that
    // the string actually is BCP-47. This works well for things like strings
    // using BCP-47 extensions, but it does not at all work for things like
    // BCP-47 grandfathered tags (eg. "en-GB-oed") which are possible to also
    // interpret as ICU locale IDs and because of that won't trigger the BCP-47
    // parsing. Therefore the code here explicitly calls uloc_forLanguageTag()
    // and then Locale::init(), instead of just calling the normal constructor.

    CharString localeID;
    int32_t parsedLength;
    {
        CharStringByteSink sink(&localeID);
        ulocimp_forLanguageTag(
                tag.data(),
                tag.length(),
                sink,
                &parsedLength,
                &status);
    }

    if (U_FAILURE(status)) {
        return result;
    }

    if (parsedLength != tag.size()) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return result;
    }

    result.init(localeID.data(), /*canonicalize=*/FALSE);
    if (result.isBogus()) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return result;
}

void
Locale::toLanguageTag(ByteSink& sink, UErrorCode& status) const
{
    if (U_FAILURE(status)) {
        return;
    }

    if (fIsBogus) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    ulocimp_toLanguageTag(fullName, sink, /*strict=*/FALSE, &status);
}

Locale U_EXPORT2
Locale::createFromName (const char *name)
{
    if (name) {
        Locale l("");
        l.init(name, FALSE);
        return l;
    }
    else {
        return getDefault();
    }
}

Locale U_EXPORT2
Locale::createCanonical(const char* name) {
    Locale loc("");
    loc.init(name, TRUE);
    return loc;
}

const char *
Locale::getISO3Language() const
{
    return uloc_getISO3Language(fullName);
}


const char *
Locale::getISO3Country() const
{
    return uloc_getISO3Country(fullName);
}

/**
 * Return the LCID value as specified in the "LocaleID" resource for this
 * locale.  The LocaleID must be expressed as a hexadecimal number, from
 * one to four digits.  If the LocaleID resource is not present, or is
 * in an incorrect format, 0 is returned.  The LocaleID is for use in
 * Windows (it is an LCID), but is available on all platforms.
 */
uint32_t
Locale::getLCID() const
{
    return uloc_getLCID(fullName);
}

const char* const* U_EXPORT2 Locale::getISOCountries()
{
    return uloc_getISOCountries();
}

const char* const* U_EXPORT2 Locale::getISOLanguages()
{
    return uloc_getISOLanguages();
}

// Set the locale's data based on a posix id.
void Locale::setFromPOSIXID(const char *posixID)
{
    init(posixID, TRUE);
}

const Locale & U_EXPORT2
Locale::getRoot(void)
{
    return getLocale(eROOT);
}

const Locale & U_EXPORT2
Locale::getEnglish(void)
{
    return getLocale(eENGLISH);
}

const Locale & U_EXPORT2
Locale::getFrench(void)
{
    return getLocale(eFRENCH);
}

const Locale & U_EXPORT2
Locale::getGerman(void)
{
    return getLocale(eGERMAN);
}

const Locale & U_EXPORT2
Locale::getItalian(void)
{
    return getLocale(eITALIAN);
}

const Locale & U_EXPORT2
Locale::getJapanese(void)
{
    return getLocale(eJAPANESE);
}

const Locale & U_EXPORT2
Locale::getKorean(void)
{
    return getLocale(eKOREAN);
}

const Locale & U_EXPORT2
Locale::getChinese(void)
{
    return getLocale(eCHINESE);
}

const Locale & U_EXPORT2
Locale::getSimplifiedChinese(void)
{
    return getLocale(eCHINA);
}

const Locale & U_EXPORT2
Locale::getTraditionalChinese(void)
{
    return getLocale(eTAIWAN);
}


const Locale & U_EXPORT2
Locale::getFrance(void)
{
    return getLocale(eFRANCE);
}

const Locale & U_EXPORT2
Locale::getGermany(void)
{
    return getLocale(eGERMANY);
}

const Locale & U_EXPORT2
Locale::getItaly(void)
{
    return getLocale(eITALY);
}

const Locale & U_EXPORT2
Locale::getJapan(void)
{
    return getLocale(eJAPAN);
}

const Locale & U_EXPORT2
Locale::getKorea(void)
{
    return getLocale(eKOREA);
}

const Locale & U_EXPORT2
Locale::getChina(void)
{
    return getLocale(eCHINA);
}

const Locale & U_EXPORT2
Locale::getPRC(void)
{
    return getLocale(eCHINA);
}

const Locale & U_EXPORT2
Locale::getTaiwan(void)
{
    return getLocale(eTAIWAN);
}

const Locale & U_EXPORT2
Locale::getUK(void)
{
    return getLocale(eUK);
}

const Locale & U_EXPORT2
Locale::getUS(void)
{
    return getLocale(eUS);
}

const Locale & U_EXPORT2
Locale::getCanada(void)
{
    return getLocale(eCANADA);
}

const Locale & U_EXPORT2
Locale::getCanadaFrench(void)
{
    return getLocale(eCANADA_FRENCH);
}

const Locale &
Locale::getLocale(int locid)
{
    Locale *localeCache = getLocaleCache();
    U_ASSERT((locid < eMAX_LOCALES)&&(locid>=0));
    if (localeCache == NULL) {
        // Failure allocating the locale cache.
        //   The best we can do is return a NULL reference.
        locid = 0;
    }
    return localeCache[locid]; /*operating on NULL*/
}

/*
This function is defined this way in order to get around static
initialization and static destruction.
 */
Locale *
Locale::getLocaleCache(void)
{
    UErrorCode status = U_ZERO_ERROR;
    umtx_initOnce(gLocaleCacheInitOnce, locale_init, status);
    return gLocaleCache;
}

class KeywordEnumeration : public StringEnumeration {
private:
    char *keywords;
    char *current;
    int32_t length;
    UnicodeString currUSKey;
    static const char fgClassID;/* Warning this is used beyond the typical RTTI usage. */

public:
    static UClassID U_EXPORT2 getStaticClassID(void) { return (UClassID)&fgClassID; }
    virtual UClassID getDynamicClassID(void) const { return getStaticClassID(); }
public:
    KeywordEnumeration(const char *keys, int32_t keywordLen, int32_t currentIndex, UErrorCode &status)
        : keywords((char *)&fgClassID), current((char *)&fgClassID), length(0) {
        if(U_SUCCESS(status) && keywordLen != 0) {
            if(keys == NULL || keywordLen < 0) {
                status = U_ILLEGAL_ARGUMENT_ERROR;
            } else {
                keywords = (char *)uprv_malloc(keywordLen+1);
                if (keywords == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                }
                else {
                    uprv_memcpy(keywords, keys, keywordLen);
                    keywords[keywordLen] = 0;
                    current = keywords + currentIndex;
                    length = keywordLen;
                }
            }
        }
    }

    virtual ~KeywordEnumeration();

    virtual StringEnumeration * clone() const
    {
        UErrorCode status = U_ZERO_ERROR;
        return new KeywordEnumeration(keywords, length, (int32_t)(current - keywords), status);
    }

    virtual int32_t count(UErrorCode &/*status*/) const {
        char *kw = keywords;
        int32_t result = 0;
        while(*kw) {
            result++;
            kw += uprv_strlen(kw)+1;
        }
        return result;
    }

    virtual const char* next(int32_t* resultLength, UErrorCode& status) {
        const char* result;
        int32_t len;
        if(U_SUCCESS(status) && *current != 0) {
            result = current;
            len = (int32_t)uprv_strlen(current);
            current += len+1;
            if(resultLength != NULL) {
                *resultLength = len;
            }
        } else {
            if(resultLength != NULL) {
                *resultLength = 0;
            }
            result = NULL;
        }
        return result;
    }

    virtual const UnicodeString* snext(UErrorCode& status) {
        int32_t resultLength = 0;
        const char *s = next(&resultLength, status);
        return setChars(s, resultLength, status);
    }

    virtual void reset(UErrorCode& /*status*/) {
        current = keywords;
    }
};

const char KeywordEnumeration::fgClassID = '\0';

KeywordEnumeration::~KeywordEnumeration() {
    uprv_free(keywords);
}

// A wrapper around KeywordEnumeration that calls uloc_toUnicodeLocaleKey() in
// the next() method for each keyword before returning it.
class UnicodeKeywordEnumeration : public KeywordEnumeration {
public:
    using KeywordEnumeration::KeywordEnumeration;
    virtual ~UnicodeKeywordEnumeration();

    virtual const char* next(int32_t* resultLength, UErrorCode& status) {
        const char* legacy_key = KeywordEnumeration::next(nullptr, status);
        if (U_SUCCESS(status) && legacy_key != nullptr) {
            const char* key = uloc_toUnicodeLocaleKey(legacy_key);
            if (key == nullptr) {
                status = U_ILLEGAL_ARGUMENT_ERROR;
            } else {
                if (resultLength != nullptr) {
                    *resultLength = static_cast<int32_t>(uprv_strlen(key));
                }
                return key;
            }
        }
        if (resultLength != nullptr) *resultLength = 0;
        return nullptr;
    }
};

// Out-of-line virtual destructor to serve as the "key function".
UnicodeKeywordEnumeration::~UnicodeKeywordEnumeration() = default;

StringEnumeration *
Locale::createKeywords(UErrorCode &status) const
{
    char keywords[256];
    int32_t keywordCapacity = sizeof keywords;
    StringEnumeration *result = NULL;

    if (U_FAILURE(status)) {
        return result;
    }

    const char* variantStart = uprv_strchr(fullName, '@');
    const char* assignment = uprv_strchr(fullName, '=');
    if(variantStart) {
        if(assignment > variantStart) {
            int32_t keyLen = locale_getKeywords(variantStart+1, '@', keywords, keywordCapacity, NULL, 0, NULL, FALSE, &status);
            if(U_SUCCESS(status) && keyLen) {
                result = new KeywordEnumeration(keywords, keyLen, 0, status);
                if (!result) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                }
            }
        } else {
            status = U_INVALID_FORMAT_ERROR;
        }
    }
    return result;
}

StringEnumeration *
Locale::createUnicodeKeywords(UErrorCode &status) const
{
    char keywords[256];
    int32_t keywordCapacity = sizeof keywords;
    StringEnumeration *result = NULL;

    if (U_FAILURE(status)) {
        return result;
    }

    const char* variantStart = uprv_strchr(fullName, '@');
    const char* assignment = uprv_strchr(fullName, '=');
    if(variantStart) {
        if(assignment > variantStart) {
            int32_t keyLen = locale_getKeywords(variantStart+1, '@', keywords, keywordCapacity, NULL, 0, NULL, FALSE, &status);
            if(U_SUCCESS(status) && keyLen) {
                result = new UnicodeKeywordEnumeration(keywords, keyLen, 0, status);
                if (!result) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                }
            }
        } else {
            status = U_INVALID_FORMAT_ERROR;
        }
    }
    return result;
}

int32_t
Locale::getKeywordValue(const char* keywordName, char *buffer, int32_t bufLen, UErrorCode &status) const
{
    return uloc_getKeywordValue(fullName, keywordName, buffer, bufLen, &status);
}

void
Locale::getKeywordValue(StringPiece keywordName, ByteSink& sink, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }

    if (fIsBogus) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    // TODO: Remove the need for a const char* to a NUL terminated buffer.
    const CharString keywordName_nul(keywordName, status);
    if (U_FAILURE(status)) {
        return;
    }

    LocalMemory<char> scratch;
    int32_t scratch_capacity = 16;  // Arbitrarily chosen default size.

    char* buffer;
    int32_t result_capacity, reslen;

    for (;;) {
        if (scratch.allocateInsteadAndReset(scratch_capacity) == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }

        buffer = sink.GetAppendBuffer(
                /*min_capacity=*/scratch_capacity,
                /*desired_capacity_hint=*/scratch_capacity,
                scratch.getAlias(),
                scratch_capacity,
                &result_capacity);

        reslen = uloc_getKeywordValue(
                fullName,
                keywordName_nul.data(),
                buffer,
                result_capacity,
                &status);

        if (status != U_BUFFER_OVERFLOW_ERROR) {
            break;
        }

        scratch_capacity = reslen;
        status = U_ZERO_ERROR;
    }

    if (U_FAILURE(status)) {
        return;
    }

    sink.Append(buffer, reslen);
    if (status == U_STRING_NOT_TERMINATED_WARNING) {
        status = U_ZERO_ERROR;  // Terminators not used.
    }
}

void
Locale::getUnicodeKeywordValue(StringPiece keywordName,
                               ByteSink& sink,
                               UErrorCode& status) const {
    // TODO: Remove the need for a const char* to a NUL terminated buffer.
    const CharString keywordName_nul(keywordName, status);
    if (U_FAILURE(status)) {
        return;
    }

    const char* legacy_key = uloc_toLegacyKey(keywordName_nul.data());

    if (legacy_key == nullptr) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    CharString legacy_value;
    {
        CharStringByteSink sink(&legacy_value);
        getKeywordValue(legacy_key, sink, status);
    }

    if (U_FAILURE(status)) {
        return;
    }

    const char* unicode_value = uloc_toUnicodeLocaleType(
            keywordName_nul.data(), legacy_value.data());

    if (unicode_value == nullptr) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    sink.Append(unicode_value, static_cast<int32_t>(uprv_strlen(unicode_value)));
}

void
Locale::setKeywordValue(const char* keywordName, const char* keywordValue, UErrorCode &status)
{
    if (U_FAILURE(status)) {
        return;
    }
    int32_t bufferLength = uprv_max((int32_t)(uprv_strlen(fullName) + 1), ULOC_FULLNAME_CAPACITY);
    int32_t newLength = uloc_setKeywordValue(keywordName, keywordValue, fullName,
                                             bufferLength, &status) + 1;
    /* Handle the case the current buffer is not enough to hold the new id */
    if (status == U_BUFFER_OVERFLOW_ERROR) {
        U_ASSERT(newLength > bufferLength);
        char* newFullName = (char *)uprv_malloc(newLength);
        if (newFullName == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        uprv_strcpy(newFullName, fullName);
        if (fullName != fullNameBuffer) {
            // if full Name is already on the heap, need to free it.
            uprv_free(fullName);
        }
        fullName = newFullName;
        status = U_ZERO_ERROR;
        uloc_setKeywordValue(keywordName, keywordValue, fullName, newLength, &status);
    } else {
        U_ASSERT(newLength <= bufferLength);
    }
    if (U_SUCCESS(status) && baseName == fullName) {
        // May have added the first keyword, meaning that the fullName is no longer also the baseName.
        initBaseName(status);
    }
}

void
Locale::setKeywordValue(StringPiece keywordName,
                        StringPiece keywordValue,
                        UErrorCode& status) {
    // TODO: Remove the need for a const char* to a NUL terminated buffer.
    const CharString keywordName_nul(keywordName, status);
    const CharString keywordValue_nul(keywordValue, status);
    setKeywordValue(keywordName_nul.data(), keywordValue_nul.data(), status);
}

void
Locale::setUnicodeKeywordValue(StringPiece keywordName,
                               StringPiece keywordValue,
                               UErrorCode& status) {
    // TODO: Remove the need for a const char* to a NUL terminated buffer.
    const CharString keywordName_nul(keywordName, status);
    const CharString keywordValue_nul(keywordValue, status);

    if (U_FAILURE(status)) {
        return;
    }

    const char* legacy_key = uloc_toLegacyKey(keywordName_nul.data());

    if (legacy_key == nullptr) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    const char* legacy_value = nullptr;

    if (!keywordValue_nul.isEmpty()) {
        legacy_value =
            uloc_toLegacyType(keywordName_nul.data(), keywordValue_nul.data());

        if (legacy_value == nullptr) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
    }

    setKeywordValue(legacy_key, legacy_value, status);
}

const char *
Locale::getBaseName() const {
    return baseName;
}

Locale::Iterator::~Iterator() = default;

//eof
U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  loclikely.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010feb25
*   created by: Markus W. Scherer
*
*   Code for likely and minimized locale subtags, separated out from other .cpp files
*   that then do not depend on resource bundle code and likely-subtags data.
*/
















/**
 * These are the canonical strings for unknown languages, scripts and regions.
 **/
static const char* const unknownLanguage = "und";
static const char* const unknownScript = "Zzzz";
static const char* const unknownRegion = "ZZ";

/**
 * This function looks for the localeID in the likelySubtags resource.
 *
 * @param localeID The tag to find.
 * @param buffer A buffer to hold the matching entry
 * @param bufferLength The length of the output buffer
 * @return A pointer to "buffer" if found, or a null pointer if not.
 */
static const char*  U_CALLCONV
findLikelySubtags(const char* localeID,
                  char* buffer,
                  int32_t bufferLength,
                  UErrorCode* err) {
    const char* result = NULL;

    if (!U_FAILURE(*err)) {
        int32_t resLen = 0;
        const UChar* s = NULL;
        UErrorCode tmpErr = U_ZERO_ERROR;
        icu::LocalUResourceBundlePointer subtags(ures_openDirect(NULL, "likelySubtags", &tmpErr));
        if (U_SUCCESS(tmpErr)) {
            icu::CharString und;
            if (localeID != NULL) {
                if (*localeID == '\0') {
                    localeID = unknownLanguage;
                } else if (*localeID == '_') {
                    und.append(unknownLanguage, *err);
                    und.append(localeID, *err);
                    if (U_FAILURE(*err)) {
                        return NULL;
                    }
                    localeID = und.data();
                }
            }
            s = ures_getStringByKey(subtags.getAlias(), localeID, &resLen, &tmpErr);

            if (U_FAILURE(tmpErr)) {
                /*
                 * If a resource is missing, it's not really an error, it's
                 * just that we don't have any data for that particular locale ID.
                 */
                if (tmpErr != U_MISSING_RESOURCE_ERROR) {
                    *err = tmpErr;
                }
            }
            else if (resLen >= bufferLength) {
                /* The buffer should never overflow. */
                *err = U_INTERNAL_PROGRAM_ERROR;
            }
            else {
                u_UCharsToChars(s, buffer, resLen + 1);
                if (resLen >= 3 &&
                    uprv_strnicmp(buffer, unknownLanguage, 3) == 0 &&
                    (resLen == 3 || buffer[3] == '_')) {
                    uprv_memmove(buffer, buffer + 3, resLen - 3 + 1);
                }
                result = buffer;
            }
        } else {
            *err = tmpErr;
        }
    }

    return result;
}

/**
 * Append a tag to a buffer, adding the separator if necessary.  The buffer
 * must be large enough to contain the resulting tag plus any separator
 * necessary. The tag must not be a zero-length string.
 *
 * @param tag The tag to add.
 * @param tagLength The length of the tag.
 * @param buffer The output buffer.
 * @param bufferLength The length of the output buffer.  This is an input/ouput parameter.
 **/
static void U_CALLCONV
appendTag(
    const char* tag,
    int32_t tagLength,
    char* buffer,
    int32_t* bufferLength,
    UBool withSeparator) {

    if (withSeparator) {
        buffer[*bufferLength] = '_';
        ++(*bufferLength);
    }

    uprv_memmove(
        &buffer[*bufferLength],
        tag,
        tagLength);

    *bufferLength += tagLength;
}

/**
 * Create a tag string from the supplied parameters.  The lang, script and region
 * parameters may be NULL pointers. If they are, their corresponding length parameters
 * must be less than or equal to 0.
 *
 * If any of the language, script or region parameters are empty, and the alternateTags
 * parameter is not NULL, it will be parsed for potential language, script and region tags
 * to be used when constructing the new tag.  If the alternateTags parameter is NULL, or
 * it contains no language tag, the default tag for the unknown language is used.
 *
 * If the length of the new string exceeds the capacity of the output buffer, 
 * the function copies as many bytes to the output buffer as it can, and returns
 * the error U_BUFFER_OVERFLOW_ERROR.
 *
 * If an illegal argument is provided, the function returns the error
 * U_ILLEGAL_ARGUMENT_ERROR.
 *
 * Note that this function can return the warning U_STRING_NOT_TERMINATED_WARNING if
 * the tag string fits in the output buffer, but the null terminator doesn't.
 *
 * @param lang The language tag to use.
 * @param langLength The length of the language tag.
 * @param script The script tag to use.
 * @param scriptLength The length of the script tag.
 * @param region The region tag to use.
 * @param regionLength The length of the region tag.
 * @param trailing Any trailing data to append to the new tag.
 * @param trailingLength The length of the trailing data.
 * @param alternateTags A string containing any alternate tags.
 * @param sink The output sink receiving the tag string.
 * @param err A pointer to a UErrorCode for error reporting.
 **/
static void U_CALLCONV
createTagStringWithAlternates(
    const char* lang,
    int32_t langLength,
    const char* script,
    int32_t scriptLength,
    const char* region,
    int32_t regionLength,
    const char* trailing,
    int32_t trailingLength,
    const char* alternateTags,
    icu::ByteSink& sink,
    UErrorCode* err) {

    if (U_FAILURE(*err)) {
        goto error;
    }
    else if (langLength >= ULOC_LANG_CAPACITY ||
             scriptLength >= ULOC_SCRIPT_CAPACITY ||
             regionLength >= ULOC_COUNTRY_CAPACITY) {
        goto error;
    }
    else {
        /**
         * ULOC_FULLNAME_CAPACITY will provide enough capacity
         * that we can build a string that contains the language,
         * script and region code without worrying about overrunning
         * the user-supplied buffer.
         **/
        char tagBuffer[ULOC_FULLNAME_CAPACITY];
        int32_t tagLength = 0;
        UBool regionAppended = FALSE;

        if (langLength > 0) {
            appendTag(
                lang,
                langLength,
                tagBuffer,
                &tagLength,
                /*withSeparator=*/FALSE);
        }
        else if (alternateTags == NULL) {
            /*
             * Use the empty string for an unknown language, if
             * we found no language.
             */
        }
        else {
            /*
             * Parse the alternateTags string for the language.
             */
            char alternateLang[ULOC_LANG_CAPACITY];
            int32_t alternateLangLength = sizeof(alternateLang);

            alternateLangLength =
                uloc_getLanguage(
                    alternateTags,
                    alternateLang,
                    alternateLangLength,
                    err);
            if(U_FAILURE(*err) ||
                alternateLangLength >= ULOC_LANG_CAPACITY) {
                goto error;
            }
            else if (alternateLangLength == 0) {
                /*
                 * Use the empty string for an unknown language, if
                 * we found no language.
                 */
            }
            else {
                appendTag(
                    alternateLang,
                    alternateLangLength,
                    tagBuffer,
                    &tagLength,
                    /*withSeparator=*/FALSE);
            }
        }

        if (scriptLength > 0) {
            appendTag(
                script,
                scriptLength,
                tagBuffer,
                &tagLength,
                /*withSeparator=*/TRUE);
        }
        else if (alternateTags != NULL) {
            /*
             * Parse the alternateTags string for the script.
             */
            char alternateScript[ULOC_SCRIPT_CAPACITY];

            const int32_t alternateScriptLength =
                uloc_getScript(
                    alternateTags,
                    alternateScript,
                    sizeof(alternateScript),
                    err);

            if (U_FAILURE(*err) ||
                alternateScriptLength >= ULOC_SCRIPT_CAPACITY) {
                goto error;
            }
            else if (alternateScriptLength > 0) {
                appendTag(
                    alternateScript,
                    alternateScriptLength,
                    tagBuffer,
                    &tagLength,
                    /*withSeparator=*/TRUE);
            }
        }

        if (regionLength > 0) {
            appendTag(
                region,
                regionLength,
                tagBuffer,
                &tagLength,
                /*withSeparator=*/TRUE);

            regionAppended = TRUE;
        }
        else if (alternateTags != NULL) {
            /*
             * Parse the alternateTags string for the region.
             */
            char alternateRegion[ULOC_COUNTRY_CAPACITY];

            const int32_t alternateRegionLength =
                uloc_getCountry(
                    alternateTags,
                    alternateRegion,
                    sizeof(alternateRegion),
                    err);
            if (U_FAILURE(*err) ||
                alternateRegionLength >= ULOC_COUNTRY_CAPACITY) {
                goto error;
            }
            else if (alternateRegionLength > 0) {
                appendTag(
                    alternateRegion,
                    alternateRegionLength,
                    tagBuffer,
                    &tagLength,
                    /*withSeparator=*/TRUE);

                regionAppended = TRUE;
            }
        }

        /**
         * Copy the partial tag from our internal buffer to the supplied
         * target.
         **/
        sink.Append(tagBuffer, tagLength);

        if (trailingLength > 0) {
            if (*trailing != '@') {
                sink.Append("_", 1);
                if (!regionAppended) {
                    /* extra separator is required */
                    sink.Append("_", 1);
                }
            }

            /*
             * Copy the trailing data into the supplied buffer.
             */
            sink.Append(trailing, trailingLength);
        }

        return;
    }

error:

    /**
     * An overflow indicates the locale ID passed in
     * is ill-formed.  If we got here, and there was
     * no previous error, it's an implicit overflow.
     **/
    if (*err ==  U_BUFFER_OVERFLOW_ERROR ||
        U_SUCCESS(*err)) {
        *err = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

/**
 * Create a tag string from the supplied parameters.  The lang, script and region
 * parameters may be NULL pointers. If they are, their corresponding length parameters
 * must be less than or equal to 0.  If the lang parameter is an empty string, the
 * default value for an unknown language is written to the output buffer.
 *
 * If the length of the new string exceeds the capacity of the output buffer, 
 * the function copies as many bytes to the output buffer as it can, and returns
 * the error U_BUFFER_OVERFLOW_ERROR.
 *
 * If an illegal argument is provided, the function returns the error
 * U_ILLEGAL_ARGUMENT_ERROR.
 *
 * @param lang The language tag to use.
 * @param langLength The length of the language tag.
 * @param script The script tag to use.
 * @param scriptLength The length of the script tag.
 * @param region The region tag to use.
 * @param regionLength The length of the region tag.
 * @param trailing Any trailing data to append to the new tag.
 * @param trailingLength The length of the trailing data.
 * @param sink The output sink receiving the tag string.
 * @param err A pointer to a UErrorCode for error reporting.
 **/
static void U_CALLCONV
createTagString(
    const char* lang,
    int32_t langLength,
    const char* script,
    int32_t scriptLength,
    const char* region,
    int32_t regionLength,
    const char* trailing,
    int32_t trailingLength,
    icu::ByteSink& sink,
    UErrorCode* err)
{
    createTagStringWithAlternates(
                lang,
                langLength,
                script,
                scriptLength,
                region,
                regionLength,
                trailing,
                trailingLength,
                NULL,
                sink,
                err);
}

/**
 * Parse the language, script, and region subtags from a tag string, and copy the
 * results into the corresponding output parameters. The buffers are null-terminated,
 * unless overflow occurs.
 *
 * The langLength, scriptLength, and regionLength parameters are input/output
 * parameters, and must contain the capacity of their corresponding buffers on
 * input.  On output, they will contain the actual length of the buffers, not
 * including the null terminator.
 *
 * If the length of any of the output subtags exceeds the capacity of the corresponding
 * buffer, the function copies as many bytes to the output buffer as it can, and returns
 * the error U_BUFFER_OVERFLOW_ERROR.  It will not parse any more subtags once overflow
 * occurs.
 *
 * If an illegal argument is provided, the function returns the error
 * U_ILLEGAL_ARGUMENT_ERROR.
 *
 * @param localeID The locale ID to parse.
 * @param lang The language tag buffer.
 * @param langLength The length of the language tag.
 * @param script The script tag buffer.
 * @param scriptLength The length of the script tag.
 * @param region The region tag buffer.
 * @param regionLength The length of the region tag.
 * @param err A pointer to a UErrorCode for error reporting.
 * @return The number of chars of the localeID parameter consumed.
 **/
static int32_t U_CALLCONV
parseTagString(
    const char* localeID,
    char* lang,
    int32_t* langLength,
    char* script,
    int32_t* scriptLength,
    char* region,
    int32_t* regionLength,
    UErrorCode* err)
{
    const char* position = localeID;
    int32_t subtagLength = 0;

    if(U_FAILURE(*err) ||
       localeID == NULL ||
       lang == NULL ||
       langLength == NULL ||
       script == NULL ||
       scriptLength == NULL ||
       region == NULL ||
       regionLength == NULL) {
        goto error;
    }

    subtagLength = ulocimp_getLanguage(position, lang, *langLength, &position);
    u_terminateChars(lang, *langLength, subtagLength, err);

    /*
     * Note that we explicit consider U_STRING_NOT_TERMINATED_WARNING
     * to be an error, because it indicates the user-supplied tag is
     * not well-formed.
     */
    if(U_FAILURE(*err)) {
        goto error;
    }

    *langLength = subtagLength;

    /*
     * If no language was present, use the empty string instead.
     * Otherwise, move past any separator.
     */
    if (_isIDSeparator(*position)) {
        ++position;
    }

    subtagLength = ulocimp_getScript(position, script, *scriptLength, &position);
    u_terminateChars(script, *scriptLength, subtagLength, err);

    if(U_FAILURE(*err)) {
        goto error;
    }

    *scriptLength = subtagLength;

    if (*scriptLength > 0) {
        if (uprv_strnicmp(script, unknownScript, *scriptLength) == 0) {
            /**
             * If the script part is the "unknown" script, then don't return it.
             **/
            *scriptLength = 0;
        }

        /*
         * Move past any separator.
         */
        if (_isIDSeparator(*position)) {
            ++position;
        }    
    }

    subtagLength = ulocimp_getCountry(position, region, *regionLength, &position);
    u_terminateChars(region, *regionLength, subtagLength, err);

    if(U_FAILURE(*err)) {
        goto error;
    }

    *regionLength = subtagLength;

    if (*regionLength > 0) {
        if (uprv_strnicmp(region, unknownRegion, *regionLength) == 0) {
            /**
             * If the region part is the "unknown" region, then don't return it.
             **/
            *regionLength = 0;
        }
    } else if (*position != 0 && *position != '@') {
        /* back up over consumed trailing separator */
        --position;
    }

exit:

    return (int32_t)(position - localeID);

error:

    /**
     * If we get here, we have no explicit error, it's the result of an
     * illegal argument.
     **/
    if (!U_FAILURE(*err)) {
        *err = U_ILLEGAL_ARGUMENT_ERROR;
    }

    goto exit;
}

static UBool U_CALLCONV
createLikelySubtagsString(
    const char* lang,
    int32_t langLength,
    const char* script,
    int32_t scriptLength,
    const char* region,
    int32_t regionLength,
    const char* variants,
    int32_t variantsLength,
    icu::ByteSink& sink,
    UErrorCode* err) {
    /**
     * ULOC_FULLNAME_CAPACITY will provide enough capacity
     * that we can build a string that contains the language,
     * script and region code without worrying about overrunning
     * the user-supplied buffer.
     **/
    char likelySubtagsBuffer[ULOC_FULLNAME_CAPACITY];

    if(U_FAILURE(*err)) {
        goto error;
    }

    /**
     * Try the language with the script and region first.
     **/
    if (scriptLength > 0 && regionLength > 0) {

        const char* likelySubtags = NULL;

        icu::CharString tagBuffer;
        {
            icu::CharStringByteSink sink(&tagBuffer);
            createTagString(
                lang,
                langLength,
                script,
                scriptLength,
                region,
                regionLength,
                NULL,
                0,
                sink,
                err);
        }
        if(U_FAILURE(*err)) {
            goto error;
        }

        likelySubtags =
            findLikelySubtags(
                tagBuffer.data(),
                likelySubtagsBuffer,
                sizeof(likelySubtagsBuffer),
                err);
        if(U_FAILURE(*err)) {
            goto error;
        }

        if (likelySubtags != NULL) {
            /* Always use the language tag from the
               maximal string, since it may be more
               specific than the one provided. */
            createTagStringWithAlternates(
                        NULL,
                        0,
                        NULL,
                        0,
                        NULL,
                        0,
                        variants,
                        variantsLength,
                        likelySubtags,
                        sink,
                        err);
            return TRUE;
        }
    }

    /**
     * Try the language with just the script.
     **/
    if (scriptLength > 0) {

        const char* likelySubtags = NULL;

        icu::CharString tagBuffer;
        {
            icu::CharStringByteSink sink(&tagBuffer);
            createTagString(
                lang,
                langLength,
                script,
                scriptLength,
                NULL,
                0,
                NULL,
                0,
                sink,
                err);
        }
        if(U_FAILURE(*err)) {
            goto error;
        }

        likelySubtags =
            findLikelySubtags(
                tagBuffer.data(),
                likelySubtagsBuffer,
                sizeof(likelySubtagsBuffer),
                err);
        if(U_FAILURE(*err)) {
            goto error;
        }

        if (likelySubtags != NULL) {
            /* Always use the language tag from the
               maximal string, since it may be more
               specific than the one provided. */
            createTagStringWithAlternates(
                        NULL,
                        0,
                        NULL,
                        0,
                        region,
                        regionLength,
                        variants,
                        variantsLength,
                        likelySubtags,
                        sink,
                        err);
            return TRUE;
        }
    }

    /**
     * Try the language with just the region.
     **/
    if (regionLength > 0) {

        const char* likelySubtags = NULL;

        icu::CharString tagBuffer;
        {
            icu::CharStringByteSink sink(&tagBuffer);
            createTagString(
                lang,
                langLength,
                NULL,
                0,
                region,
                regionLength,
                NULL,
                0,
                sink,
                err);
        }
        if(U_FAILURE(*err)) {
            goto error;
        }

        likelySubtags =
            findLikelySubtags(
                tagBuffer.data(),
                likelySubtagsBuffer,
                sizeof(likelySubtagsBuffer),
                err);
        if(U_FAILURE(*err)) {
            goto error;
        }

        if (likelySubtags != NULL) {
            /* Always use the language tag from the
               maximal string, since it may be more
               specific than the one provided. */
            createTagStringWithAlternates(
                        NULL,
                        0,
                        script,
                        scriptLength,
                        NULL,
                        0,
                        variants,
                        variantsLength,
                        likelySubtags,
                        sink,
                        err);
            return TRUE;
        }
    }

    /**
     * Finally, try just the language.
     **/
    {
        const char* likelySubtags = NULL;

        icu::CharString tagBuffer;
        {
            icu::CharStringByteSink sink(&tagBuffer);
            createTagString(
                lang,
                langLength,
                NULL,
                0,
                NULL,
                0,
                NULL,
                0,
                sink,
                err);
        }
        if(U_FAILURE(*err)) {
            goto error;
        }

        likelySubtags =
            findLikelySubtags(
                tagBuffer.data(),
                likelySubtagsBuffer,
                sizeof(likelySubtagsBuffer),
                err);
        if(U_FAILURE(*err)) {
            goto error;
        }

        if (likelySubtags != NULL) {
            /* Always use the language tag from the
               maximal string, since it may be more
               specific than the one provided. */
            createTagStringWithAlternates(
                        NULL,
                        0,
                        script,
                        scriptLength,
                        region,
                        regionLength,
                        variants,
                        variantsLength,
                        likelySubtags,
                        sink,
                        err);
            return TRUE;
        }
    }

    return FALSE;

error:

    if (!U_FAILURE(*err)) {
        *err = U_ILLEGAL_ARGUMENT_ERROR;
    }

    return FALSE;
}

#define CHECK_TRAILING_VARIANT_SIZE(trailing, trailingLength) UPRV_BLOCK_MACRO_BEGIN { \
    int32_t count = 0; \
    int32_t i; \
    for (i = 0; i < trailingLength; i++) { \
        if (trailing[i] == '-' || trailing[i] == '_') { \
            count = 0; \
            if (count > 8) { \
                goto error; \
            } \
        } else if (trailing[i] == '@') { \
            break; \
        } else if (count > 8) { \
            goto error; \
        } else { \
            count++; \
        } \
    } \
} UPRV_BLOCK_MACRO_END

static void
_uloc_addLikelySubtags(const char* localeID,
                       icu::ByteSink& sink,
                       UErrorCode* err) {
    char lang[ULOC_LANG_CAPACITY];
    int32_t langLength = sizeof(lang);
    char script[ULOC_SCRIPT_CAPACITY];
    int32_t scriptLength = sizeof(script);
    char region[ULOC_COUNTRY_CAPACITY];
    int32_t regionLength = sizeof(region);
    const char* trailing = "";
    int32_t trailingLength = 0;
    int32_t trailingIndex = 0;
    UBool success = FALSE;

    if(U_FAILURE(*err)) {
        goto error;
    }
    if (localeID == NULL) {
        goto error;
    }

    trailingIndex = parseTagString(
        localeID,
        lang,
        &langLength,
        script,
        &scriptLength,
        region,
        &regionLength,
        err);
    if(U_FAILURE(*err)) {
        /* Overflow indicates an illegal argument error */
        if (*err == U_BUFFER_OVERFLOW_ERROR) {
            *err = U_ILLEGAL_ARGUMENT_ERROR;
        }

        goto error;
    }

    /* Find the length of the trailing portion. */
    while (_isIDSeparator(localeID[trailingIndex])) {
        trailingIndex++;
    }
    trailing = &localeID[trailingIndex];
    trailingLength = (int32_t)uprv_strlen(trailing);

    CHECK_TRAILING_VARIANT_SIZE(trailing, trailingLength);

    success =
        createLikelySubtagsString(
            lang,
            langLength,
            script,
            scriptLength,
            region,
            regionLength,
            trailing,
            trailingLength,
            sink,
            err);

    if (!success) {
        const int32_t localIDLength = (int32_t)uprv_strlen(localeID);

        /*
         * If we get here, we need to return localeID.
         */
        sink.Append(localeID, localIDLength);
    }

    return;

error:

    if (!U_FAILURE(*err)) {
        *err = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

static void
_uloc_minimizeSubtags(const char* localeID,
                      icu::ByteSink& sink,
                      UErrorCode* err) {
    icu::CharString maximizedTagBuffer;

    char lang[ULOC_LANG_CAPACITY];
    int32_t langLength = sizeof(lang);
    char script[ULOC_SCRIPT_CAPACITY];
    int32_t scriptLength = sizeof(script);
    char region[ULOC_COUNTRY_CAPACITY];
    int32_t regionLength = sizeof(region);
    const char* trailing = "";
    int32_t trailingLength = 0;
    int32_t trailingIndex = 0;

    if(U_FAILURE(*err)) {
        goto error;
    }
    else if (localeID == NULL) {
        goto error;
    }

    trailingIndex =
        parseTagString(
            localeID,
            lang,
            &langLength,
            script,
            &scriptLength,
            region,
            &regionLength,
            err);
    if(U_FAILURE(*err)) {

        /* Overflow indicates an illegal argument error */
        if (*err == U_BUFFER_OVERFLOW_ERROR) {
            *err = U_ILLEGAL_ARGUMENT_ERROR;
        }

        goto error;
    }

    /* Find the spot where the variants or the keywords begin, if any. */
    while (_isIDSeparator(localeID[trailingIndex])) {
        trailingIndex++;
    }
    trailing = &localeID[trailingIndex];
    trailingLength = (int32_t)uprv_strlen(trailing);

    CHECK_TRAILING_VARIANT_SIZE(trailing, trailingLength);

    {
        icu::CharString base;
        {
            icu::CharStringByteSink sink(&base);
            createTagString(
                lang,
                langLength,
                script,
                scriptLength,
                region,
                regionLength,
                NULL,
                0,
                sink,
                err);
        }

        /**
         * First, we need to first get the maximization
         * from AddLikelySubtags.
         **/
        {
            icu::CharStringByteSink sink(&maximizedTagBuffer);
            ulocimp_addLikelySubtags(base.data(), sink, err);
        }
    }

    if(U_FAILURE(*err)) {
        goto error;
    }

    /**
     * Start first with just the language.
     **/
    {
        icu::CharString tagBuffer;
        {
            icu::CharStringByteSink sink(&tagBuffer);
            createLikelySubtagsString(
                lang,
                langLength,
                NULL,
                0,
                NULL,
                0,
                NULL,
                0,
                sink,
                err);
        }

        if(U_FAILURE(*err)) {
            goto error;
        }
        else if (!tagBuffer.isEmpty() && uprv_strnicmp(
                    maximizedTagBuffer.data(),
                    tagBuffer.data(),
                    tagBuffer.length()) == 0) {

            createTagString(
                        lang,
                        langLength,
                        NULL,
                        0,
                        NULL,
                        0,
                        trailing,
                        trailingLength,
                        sink,
                        err);
            return;
        }
    }

    /**
     * Next, try the language and region.
     **/
    if (regionLength > 0) {

        icu::CharString tagBuffer;
        {
            icu::CharStringByteSink sink(&tagBuffer);
            createLikelySubtagsString(
                lang,
                langLength,
                NULL,
                0,
                region,
                regionLength,
                NULL,
                0,
                sink,
                err);
        }

        if(U_FAILURE(*err)) {
            goto error;
        }
        else if (uprv_strnicmp(
                    maximizedTagBuffer.data(),
                    tagBuffer.data(),
                    tagBuffer.length()) == 0) {

            createTagString(
                        lang,
                        langLength,
                        NULL,
                        0,
                        region,
                        regionLength,
                        trailing,
                        trailingLength,
                        sink,
                        err);
            return;
        }
    }

    /**
     * Finally, try the language and script.  This is our last chance,
     * since trying with all three subtags would only yield the
     * maximal version that we already have.
     **/
    if (scriptLength > 0 && regionLength > 0) {
        icu::CharString tagBuffer;
        {
            icu::CharStringByteSink sink(&tagBuffer);
            createLikelySubtagsString(
                lang,
                langLength,
                script,
                scriptLength,
                NULL,
                0,
                NULL,
                0,
                sink,
                err);
        }

        if(U_FAILURE(*err)) {
            goto error;
        }
        else if (uprv_strnicmp(
                    maximizedTagBuffer.data(),
                    tagBuffer.data(),
                    tagBuffer.length()) == 0) {

            createTagString(
                        lang,
                        langLength,
                        script,
                        scriptLength,
                        NULL,
                        0,
                        trailing,
                        trailingLength,
                        sink,
                        err);
            return;
        }
    }

    {
        /**
         * If we got here, return the locale ID parameter.
         **/
        const int32_t localeIDLength = (int32_t)uprv_strlen(localeID);
        sink.Append(localeID, localeIDLength);
        return;
    }

error:

    if (!U_FAILURE(*err)) {
        *err = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

static UBool
do_canonicalize(const char*    localeID,
         char* buffer,
         int32_t bufferCapacity,
         UErrorCode* err)
{
    uloc_canonicalize(
        localeID,
        buffer,
        bufferCapacity,
        err);

    if (*err == U_STRING_NOT_TERMINATED_WARNING ||
        *err == U_BUFFER_OVERFLOW_ERROR) {
        *err = U_ILLEGAL_ARGUMENT_ERROR;

        return FALSE;
    }
    else if (U_FAILURE(*err)) {

        return FALSE;
    }
    else {
        return TRUE;
    }
}

U_CAPI int32_t U_EXPORT2
uloc_addLikelySubtags(const char* localeID,
                      char* maximizedLocaleID,
                      int32_t maximizedLocaleIDCapacity,
                      UErrorCode* status) {
    if (U_FAILURE(*status)) {
        return 0;
    }

    icu::CheckedArrayByteSink sink(
            maximizedLocaleID, maximizedLocaleIDCapacity);

    ulocimp_addLikelySubtags(localeID, sink, status);
    int32_t reslen = sink.NumberOfBytesAppended();

    if (U_FAILURE(*status)) {
        return sink.Overflowed() ? reslen : -1;
    }

    if (sink.Overflowed()) {
        *status = U_BUFFER_OVERFLOW_ERROR;
    } else {
        u_terminateChars(
                maximizedLocaleID, maximizedLocaleIDCapacity, reslen, status);
    }

    return reslen;
}

U_CAPI void U_EXPORT2
ulocimp_addLikelySubtags(const char* localeID,
                         icu::ByteSink& sink,
                         UErrorCode* status) {
    char localeBuffer[ULOC_FULLNAME_CAPACITY];

    if (do_canonicalize(localeID, localeBuffer, sizeof localeBuffer, status)) {
        _uloc_addLikelySubtags(localeBuffer, sink, status);
    }
}

U_CAPI int32_t U_EXPORT2
uloc_minimizeSubtags(const char* localeID,
                     char* minimizedLocaleID,
                     int32_t minimizedLocaleIDCapacity,
                     UErrorCode* status) {
    if (U_FAILURE(*status)) {
        return 0;
    }

    icu::CheckedArrayByteSink sink(
            minimizedLocaleID, minimizedLocaleIDCapacity);

    ulocimp_minimizeSubtags(localeID, sink, status);
    int32_t reslen = sink.NumberOfBytesAppended();

    if (U_FAILURE(*status)) {
        return sink.Overflowed() ? reslen : -1;
    }

    if (sink.Overflowed()) {
        *status = U_BUFFER_OVERFLOW_ERROR;
    } else {
        u_terminateChars(
                minimizedLocaleID, minimizedLocaleIDCapacity, reslen, status);
    }

    return reslen;
}

U_CAPI void U_EXPORT2
ulocimp_minimizeSubtags(const char* localeID,
                        icu::ByteSink& sink,
                        UErrorCode* status) {
    char localeBuffer[ULOC_FULLNAME_CAPACITY];

    if (do_canonicalize(localeID, localeBuffer, sizeof localeBuffer, status)) {
        _uloc_minimizeSubtags(localeBuffer, sink, status);
    }
}

// Pairs of (language subtag, + or -) for finding out fast if common languages
// are LTR (minus) or RTL (plus).
static const char LANG_DIR_STRING[] =
        "root-en-es-pt-zh-ja-ko-de-fr-it-ar+he+fa+ru-nl-pl-th-tr-";

// Implemented here because this calls ulocimp_addLikelySubtags().
U_CAPI UBool U_EXPORT2
uloc_isRightToLeft(const char *locale) {
    UErrorCode errorCode = U_ZERO_ERROR;
    char script[8];
    int32_t scriptLength = uloc_getScript(locale, script, UPRV_LENGTHOF(script), &errorCode);
    if (U_FAILURE(errorCode) || errorCode == U_STRING_NOT_TERMINATED_WARNING ||
            scriptLength == 0) {
        // Fastpath: We know the likely scripts and their writing direction
        // for some common languages.
        errorCode = U_ZERO_ERROR;
        char lang[8];
        int32_t langLength = uloc_getLanguage(locale, lang, UPRV_LENGTHOF(lang), &errorCode);
        if (U_FAILURE(errorCode) || errorCode == U_STRING_NOT_TERMINATED_WARNING) {
            return FALSE;
        }
        if (langLength > 0) {
            const char* langPtr = uprv_strstr(LANG_DIR_STRING, lang);
            if (langPtr != NULL) {
                switch (langPtr[langLength]) {
                case '-': return FALSE;
                case '+': return TRUE;
                default: break;  // partial match of a longer code
                }
            }
        }
        // Otherwise, find the likely script.
        errorCode = U_ZERO_ERROR;
        icu::CharString likely;
        {
            icu::CharStringByteSink sink(&likely);
            ulocimp_addLikelySubtags(locale, sink, &errorCode);
        }
        if (U_FAILURE(errorCode) || errorCode == U_STRING_NOT_TERMINATED_WARNING) {
            return FALSE;
        }
        scriptLength = uloc_getScript(likely.data(), script, UPRV_LENGTHOF(script), &errorCode);
        if (U_FAILURE(errorCode) || errorCode == U_STRING_NOT_TERMINATED_WARNING ||
                scriptLength == 0) {
            return FALSE;
        }
    }
    UScriptCode scriptCode = (UScriptCode)u_getPropertyValueEnum(UCHAR_SCRIPT, script);
    return uscript_isRightToLeft(scriptCode);
}

U_NAMESPACE_BEGIN

UBool
Locale::isRightToLeft() const {
    return uloc_isRightToLeft(getBaseName());
}

U_NAMESPACE_END

// The following must at least allow for rg key value (6) plus terminator (1).
#define ULOC_RG_BUFLEN 8

U_CAPI int32_t U_EXPORT2
ulocimp_getRegionForSupplementalData(const char *localeID, UBool inferRegion,
                                     char *region, int32_t regionCapacity, UErrorCode* status) {
    if (U_FAILURE(*status)) {
        return 0;
    }
    char rgBuf[ULOC_RG_BUFLEN];
    UErrorCode rgStatus = U_ZERO_ERROR;

    // First check for rg keyword value
    int32_t rgLen = uloc_getKeywordValue(localeID, "rg", rgBuf, ULOC_RG_BUFLEN, &rgStatus);
    if (U_FAILURE(rgStatus) || rgLen != 6) {
        rgLen = 0;
    } else {
        // rgBuf guaranteed to be zero terminated here, with text len 6
        char *rgPtr = rgBuf;
        for (; *rgPtr!= 0; rgPtr++) {
            *rgPtr = uprv_toupper(*rgPtr);
        }
        rgLen = (uprv_strcmp(rgBuf+2, "ZZZZ") == 0)? 2: 0;
    }

    if (rgLen == 0) {
        // No valid rg keyword value, try for unicode_region_subtag
        rgLen = uloc_getCountry(localeID, rgBuf, ULOC_RG_BUFLEN, status);
        if (U_FAILURE(*status)) {
            rgLen = 0;
        } else if (rgLen == 0 && inferRegion) {
            // no unicode_region_subtag but inferRegion TRUE, try likely subtags
            rgStatus = U_ZERO_ERROR;
            icu::CharString locBuf;
            {
                icu::CharStringByteSink sink(&locBuf);
                ulocimp_addLikelySubtags(localeID, sink, &rgStatus);
            }
            if (U_SUCCESS(rgStatus)) {
                rgLen = uloc_getCountry(locBuf.data(), rgBuf, ULOC_RG_BUFLEN, status);
                if (U_FAILURE(*status)) {
                    rgLen = 0;
                }
            }
        }
    }

    rgBuf[rgLen] = 0;
    uprv_strncpy(region, rgBuf, regionCapacity);
    return u_terminateChars(region, regionCapacity, rgLen, status);
}

// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License

// loclikelysubtags.cpp
// created: 2019may08 Markus W. Scherer

#include <utility>


















U_NAMESPACE_BEGIN

namespace {

constexpr char PSEUDO_ACCENTS_PREFIX = '\'';  // -XA, -PSACCENT
constexpr char PSEUDO_BIDI_PREFIX = '+';  // -XB, -PSBIDI
constexpr char PSEUDO_CRACKED_PREFIX = ',';  // -XC, -PSCRACK

}  // namespace

/**
 * Stores NUL-terminated strings with duplicate elimination.
 * Checks for unique UTF-16 string pointers and converts to invariant characters.
 */
class UniqueCharStrings {
public:
    UniqueCharStrings(UErrorCode &errorCode) : strings(nullptr) {
        uhash_init(&map, uhash_hashUChars, uhash_compareUChars, uhash_compareLong, &errorCode);
        if (U_FAILURE(errorCode)) { return; }
        strings = new CharString();
        if (strings == nullptr) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
        }
    }
    ~UniqueCharStrings() {
        uhash_close(&map);
        delete strings;
    }

    /** Returns/orphans the CharString that contains all strings. */
    CharString *orphanCharStrings() {
        CharString *result = strings;
        strings = nullptr;
        return result;
    }

    /** Adds a string and returns a unique number for it. */
    int32_t add(const UnicodeString &s, UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return 0; }
        if (isFrozen) {
            errorCode = U_NO_WRITE_PERMISSION;
            return 0;
        }
        // The string points into the resource bundle.
        const char16_t *p = s.getBuffer();
        int32_t oldIndex = uhash_geti(&map, p);
        if (oldIndex != 0) {  // found duplicate
            return oldIndex;
        }
        // Explicit NUL terminator for the previous string.
        // The strings object is also terminated with one implicit NUL.
        strings->append(0, errorCode);
        int32_t newIndex = strings->length();
        strings->appendInvariantChars(s, errorCode);
        uhash_puti(&map, const_cast<char16_t *>(p), newIndex, &errorCode);
        return newIndex;
    }

    void freeze() { isFrozen = true; }

    /**
     * Returns a string pointer for its unique number, if this object is frozen.
     * Otherwise nullptr.
     */
    const char *get(int32_t i) const {
        U_ASSERT(isFrozen);
        return isFrozen && i > 0 ? strings->data() + i : nullptr;
    }

private:
    UHashtable map;
    CharString *strings;
    bool isFrozen = false;
};

LocaleDistanceData::LocaleDistanceData(LocaleDistanceData &&data) :
        distanceTrieBytes(data.distanceTrieBytes),
        regionToPartitions(data.regionToPartitions),
        partitions(data.partitions),
        paradigms(data.paradigms), paradigmsLength(data.paradigmsLength),
        distances(data.distances) {
    data.partitions = nullptr;
    data.paradigms = nullptr;
}

LocaleDistanceData::~LocaleDistanceData() {
    uprv_free(partitions);
    delete[] paradigms;
}

// TODO(ICU-20777): Rename to just LikelySubtagsData.
struct XLikelySubtagsData {
    UResourceBundle *langInfoBundle = nullptr;
    UniqueCharStrings strings;
    CharStringMap languageAliases;
    CharStringMap regionAliases;
    const uint8_t *trieBytes = nullptr;
    LSR *lsrs = nullptr;
    int32_t lsrsLength = 0;

    LocaleDistanceData distanceData;

    XLikelySubtagsData(UErrorCode &errorCode) : strings(errorCode) {}

    ~XLikelySubtagsData() {
        ures_close(langInfoBundle);
        delete[] lsrs;
    }

    void load(UErrorCode &errorCode) {
        langInfoBundle = ures_openDirect(nullptr, "langInfo", &errorCode);
        if (U_FAILURE(errorCode)) { return; }
        StackUResourceBundle stackTempBundle;
        ResourceDataValue value;
        ures_getValueWithFallback(langInfoBundle, "likely", stackTempBundle.getAlias(),
                                  value, errorCode);
        ResourceTable likelyTable = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }

        // Read all strings in the resource bundle and convert them to invariant char *.
        LocalMemory<int32_t> languageIndexes, regionIndexes, lsrSubtagIndexes;
        int32_t languagesLength = 0, regionsLength = 0, lsrSubtagsLength = 0;
        if (!readStrings(likelyTable, "languageAliases", value,
                         languageIndexes, languagesLength, errorCode) ||
                !readStrings(likelyTable, "regionAliases", value,
                             regionIndexes, regionsLength, errorCode) ||
                !readStrings(likelyTable, "lsrs", value,
                             lsrSubtagIndexes,lsrSubtagsLength, errorCode)) {
            return;
        }
        if ((languagesLength & 1) != 0 ||
                (regionsLength & 1) != 0 ||
                (lsrSubtagsLength % 3) != 0) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        if (lsrSubtagsLength == 0) {
            errorCode = U_MISSING_RESOURCE_ERROR;
            return;
        }

        if (!likelyTable.findValue("trie", value)) {
            errorCode = U_MISSING_RESOURCE_ERROR;
            return;
        }
        int32_t length;
        trieBytes = value.getBinary(length, errorCode);
        if (U_FAILURE(errorCode)) { return; }

        // Also read distance/matcher data if available,
        // to open & keep only one resource bundle pointer
        // and to use one single UniqueCharStrings.
        UErrorCode matchErrorCode = U_ZERO_ERROR;
        ures_getValueWithFallback(langInfoBundle, "match", stackTempBundle.getAlias(),
                                  value, matchErrorCode);
        LocalMemory<int32_t> partitionIndexes, paradigmSubtagIndexes;
        int32_t partitionsLength = 0, paradigmSubtagsLength = 0;
        if (U_SUCCESS(matchErrorCode)) {
            ResourceTable matchTable = value.getTable(errorCode);
            if (U_FAILURE(errorCode)) { return; }

            if (matchTable.findValue("trie", value)) {
                distanceData.distanceTrieBytes = value.getBinary(length, errorCode);
                if (U_FAILURE(errorCode)) { return; }
            }

            if (matchTable.findValue("regionToPartitions", value)) {
                distanceData.regionToPartitions = value.getBinary(length, errorCode);
                if (U_FAILURE(errorCode)) { return; }
                if (length < LSR::REGION_INDEX_LIMIT) {
                    errorCode = U_INVALID_FORMAT_ERROR;
                    return;
                }
            }

            if (!readStrings(matchTable, "partitions", value,
                             partitionIndexes, partitionsLength, errorCode) ||
                    !readStrings(matchTable, "paradigms", value,
                                 paradigmSubtagIndexes, paradigmSubtagsLength, errorCode)) {
                return;
            }
            if ((paradigmSubtagsLength % 3) != 0) {
                errorCode = U_INVALID_FORMAT_ERROR;
                return;
            }

            if (matchTable.findValue("distances", value)) {
                distanceData.distances = value.getIntVector(length, errorCode);
                if (U_FAILURE(errorCode)) { return; }
                if (length < 4) {  // LocaleDistance IX_LIMIT
                    errorCode = U_INVALID_FORMAT_ERROR;
                    return;
                }
            }
        } else if (matchErrorCode == U_MISSING_RESOURCE_ERROR) {
            // ok for likely subtags
        } else {  // error other than missing resource
            errorCode = matchErrorCode;
            return;
        }

        // Fetch & store invariant-character versions of strings
        // only after we have collected and de-duplicated all of them.
        strings.freeze();

        languageAliases = CharStringMap(languagesLength / 2, errorCode);
        for (int32_t i = 0; i < languagesLength; i += 2) {
            languageAliases.put(strings.get(languageIndexes[i]),
                                strings.get(languageIndexes[i + 1]), errorCode);
        }

        regionAliases = CharStringMap(regionsLength / 2, errorCode);
        for (int32_t i = 0; i < regionsLength; i += 2) {
            regionAliases.put(strings.get(regionIndexes[i]),
                              strings.get(regionIndexes[i + 1]), errorCode);
        }
        if (U_FAILURE(errorCode)) { return; }

        lsrsLength = lsrSubtagsLength / 3;
        lsrs = new LSR[lsrsLength];
        if (lsrs == nullptr) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        for (int32_t i = 0, j = 0; i < lsrSubtagsLength; i += 3, ++j) {
            lsrs[j] = LSR(strings.get(lsrSubtagIndexes[i]),
                          strings.get(lsrSubtagIndexes[i + 1]),
                          strings.get(lsrSubtagIndexes[i + 2]));
        }

        if (partitionsLength > 0) {
            distanceData.partitions = static_cast<const char **>(
                uprv_malloc(partitionsLength * sizeof(const char *)));
            if (distanceData.partitions == nullptr) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            for (int32_t i = 0; i < partitionsLength; ++i) {
                distanceData.partitions[i] = strings.get(partitionIndexes[i]);
            }
        }

        if (paradigmSubtagsLength > 0) {
            distanceData.paradigmsLength = paradigmSubtagsLength / 3;
            LSR *paradigms = new LSR[distanceData.paradigmsLength];
            if (paradigms == nullptr) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            for (int32_t i = 0, j = 0; i < paradigmSubtagsLength; i += 3, ++j) {
                paradigms[j] = LSR(strings.get(paradigmSubtagIndexes[i]),
                                   strings.get(paradigmSubtagIndexes[i + 1]),
                                   strings.get(paradigmSubtagIndexes[i + 2]));
            }
            distanceData.paradigms = paradigms;
        }
    }

private:
    bool readStrings(const ResourceTable &table, const char *key, ResourceValue &value,
                     LocalMemory<int32_t> &indexes, int32_t &length, UErrorCode &errorCode) {
        if (table.findValue(key, value)) {
            ResourceArray stringArray = value.getArray(errorCode);
            if (U_FAILURE(errorCode)) { return false; }
            length = stringArray.getSize();
            if (length == 0) { return true; }
            int32_t *rawIndexes = indexes.allocateInsteadAndCopy(length);
            if (rawIndexes == nullptr) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return false;
            }
            for (int i = 0; i < length; ++i) {
                stringArray.getValue(i, value);  // returns TRUE because i < length
                rawIndexes[i] = strings.add(value.getUnicodeString(errorCode), errorCode);
                if (U_FAILURE(errorCode)) { return false; }
            }
        }
        return true;
    }
};

namespace {

XLikelySubtags *gLikelySubtags = nullptr;
UInitOnce loclikelysubtags_gInitOnce = U_INITONCE_INITIALIZER;

UBool U_CALLCONV loclikelysubtags_cleanup() {
    delete gLikelySubtags;
    gLikelySubtags = nullptr;
    loclikelysubtags_gInitOnce.reset();
    return TRUE;
}

}  // namespace

void U_CALLCONV XLikelySubtags::initLikelySubtags(UErrorCode &errorCode) {
    // This function is invoked only via umtx_initOnce().
    U_ASSERT(gLikelySubtags == nullptr);
    XLikelySubtagsData data(errorCode);
    data.load(errorCode);
    if (U_FAILURE(errorCode)) { return; }
    gLikelySubtags = new XLikelySubtags(data);
    if (gLikelySubtags == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    ucln_common_registerCleanup(UCLN_COMMON_LIKELY_SUBTAGS, loclikelysubtags_cleanup);
}

const XLikelySubtags *XLikelySubtags::getSingleton(UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return nullptr; }
    umtx_initOnce(loclikelysubtags_gInitOnce, &XLikelySubtags::initLikelySubtags, errorCode);
    return gLikelySubtags;
}

XLikelySubtags::XLikelySubtags(XLikelySubtagsData &data) :
        langInfoBundle(data.langInfoBundle),
        strings(data.strings.orphanCharStrings()),
        languageAliases(std::move(data.languageAliases)),
        regionAliases(std::move(data.regionAliases)),
        trie(data.trieBytes),
        lsrs(data.lsrs),
#if U_DEBUG
        lsrsLength(data.lsrsLength),
#endif
        distanceData(std::move(data.distanceData)) {
    data.langInfoBundle = nullptr;
    data.lsrs = nullptr;

    // Cache the result of looking up language="und" encoded as "*", and "und-Zzzz" ("**").
    UStringTrieResult result = trie.next(u'*');
    U_ASSERT(USTRINGTRIE_HAS_NEXT(result));
    trieUndState = trie.getState64();
    result = trie.next(u'*');
    U_ASSERT(USTRINGTRIE_HAS_NEXT(result));
    trieUndZzzzState = trie.getState64();
    result = trie.next(u'*');
    U_ASSERT(USTRINGTRIE_HAS_VALUE(result));
    defaultLsrIndex = trie.getValue();
    trie.reset();

    for (char16_t c = u'a'; c <= u'z'; ++c) {
        result = trie.next(c);
        if (result == USTRINGTRIE_NO_VALUE) {
            trieFirstLetterStates[c - u'a'] = trie.getState64();
        }
        trie.reset();
    }
}

XLikelySubtags::~XLikelySubtags() {
    ures_close(langInfoBundle);
    delete strings;
    delete[] lsrs;
}

LSR XLikelySubtags::makeMaximizedLsrFrom(const Locale &locale, UErrorCode &errorCode) const {
    const char *name = locale.getName();
    if (uprv_isAtSign(name[0]) && name[1] == 'x' && name[2] == '=') {  // name.startsWith("@x=")
        // Private use language tag x-subtag-subtag...
        return LSR(name, "", "");
    }
    return makeMaximizedLsr(locale.getLanguage(), locale.getScript(), locale.getCountry(),
                            locale.getVariant(), errorCode);
}

namespace {

const char *getCanonical(const CharStringMap &aliases, const char *alias) {
    const char *canonical = aliases.get(alias);
    return canonical == nullptr ? alias : canonical;
}

}  // namespace

LSR XLikelySubtags::makeMaximizedLsr(const char *language, const char *script, const char *region,
                                     const char *variant, UErrorCode &errorCode) const {
    // Handle pseudolocales like en-XA, ar-XB, fr-PSCRACK.
    // They should match only themselves,
    // not other locales with what looks like the same language and script subtags.
    char c1;
    if (region[0] == 'X' && (c1 = region[1]) != 0 && region[2] == 0) {
        switch (c1) {
        case 'A':
            return LSR(PSEUDO_ACCENTS_PREFIX, language, script, region, errorCode);
        case 'B':
            return LSR(PSEUDO_BIDI_PREFIX, language, script, region, errorCode);
        case 'C':
            return LSR(PSEUDO_CRACKED_PREFIX, language, script, region, errorCode);
        default:  // normal locale
            break;
        }
    }

    if (variant[0] == 'P' && variant[1] == 'S') {
        if (uprv_strcmp(variant, "PSACCENT") == 0) {
            return LSR(PSEUDO_ACCENTS_PREFIX, language, script,
                       *region == 0 ? "XA" : region, errorCode);
        } else if (uprv_strcmp(variant, "PSBIDI") == 0) {
            return LSR(PSEUDO_BIDI_PREFIX, language, script,
                       *region == 0 ? "XB" : region, errorCode);
        } else if (uprv_strcmp(variant, "PSCRACK") == 0) {
            return LSR(PSEUDO_CRACKED_PREFIX, language, script,
                       *region == 0 ? "XC" : region, errorCode);
        }
        // else normal locale
    }

    language = getCanonical(languageAliases, language);
    // (We have no script mappings.)
    region = getCanonical(regionAliases, region);
    return maximize(language, script, region);
}

LSR XLikelySubtags::maximize(const char *language, const char *script, const char *region) const {
    if (uprv_strcmp(language, "und") == 0) {
        language = "";
    }
    if (uprv_strcmp(script, "Zzzz") == 0) {
        script = "";
    }
    if (uprv_strcmp(region, "ZZ") == 0) {
        region = "";
    }
    if (*script != 0 && *region != 0 && *language != 0) {
        return LSR(language, script, region);  // already maximized
    }

    uint32_t retainOldMask = 0;
    BytesTrie iter(trie);
    uint64_t state;
    int32_t value;
    // Small optimization: Array lookup for first language letter.
    int32_t c0;
    if (0 <= (c0 = uprv_lowerOrdinal(language[0])) && c0 <= 25 &&
            language[1] != 0 &&  // language.length() >= 2
            (state = trieFirstLetterStates[c0]) != 0) {
        value = trieNext(iter.resetToState64(state), language, 1);
    } else {
        value = trieNext(iter, language, 0);
    }
    if (value >= 0) {
        if (*language != 0) {
            retainOldMask |= 4;
        }
        state = iter.getState64();
    } else {
        retainOldMask |= 4;
        iter.resetToState64(trieUndState);  // "und" ("*")
        state = 0;
    }

    if (value > 0) {
        // Intermediate or final value from just language.
        if (value == SKIP_SCRIPT) {
            value = 0;
        }
        if (*script != 0) {
            retainOldMask |= 2;
        }
    } else {
        value = trieNext(iter, script, 0);
        if (value >= 0) {
            if (*script != 0) {
                retainOldMask |= 2;
            }
            state = iter.getState64();
        } else {
            retainOldMask |= 2;
            if (state == 0) {
                iter.resetToState64(trieUndZzzzState);  // "und-Zzzz" ("**")
            } else {
                iter.resetToState64(state);
                value = trieNext(iter, "", 0);
                U_ASSERT(value >= 0);
                state = iter.getState64();
            }
        }
    }

    if (value > 0) {
        // Final value from just language or language+script.
        if (*region != 0) {
            retainOldMask |= 1;
        }
    } else {
        value = trieNext(iter, region, 0);
        if (value >= 0) {
            if (*region != 0) {
                retainOldMask |= 1;
            }
        } else {
            retainOldMask |= 1;
            if (state == 0) {
                value = defaultLsrIndex;
            } else {
                iter.resetToState64(state);
                value = trieNext(iter, "", 0);
                U_ASSERT(value > 0);
            }
        }
    }
    U_ASSERT(value < lsrsLength);
    const LSR &result = lsrs[value];

    if (*language == 0) {
        language = "und";
    }

    if (retainOldMask == 0) {
        // Quickly return a copy of the lookup-result LSR
        // without new allocation of the subtags.
        return LSR(result.language, result.script, result.region);
    }
    if ((retainOldMask & 4) == 0) {
        language = result.language;
    }
    if ((retainOldMask & 2) == 0) {
        script = result.script;
    }
    if ((retainOldMask & 1) == 0) {
        region = result.region;
    }
    return LSR(language, script, region);
}

int32_t XLikelySubtags::trieNext(BytesTrie &iter, const char *s, int32_t i) {
    UStringTrieResult result;
    uint8_t c;
    if ((c = s[i]) == 0) {
        result = iter.next(u'*');
    } else {
        for (;;) {
            c = uprv_invCharToAscii(c);
            // EBCDIC: If s[i] is not an invariant character,
            // then c is now 0 and will simply not match anything, which is harmless.
            uint8_t next = s[++i];
            if (next != 0) {
                if (!USTRINGTRIE_HAS_NEXT(iter.next(c))) {
                    return -1;
                }
            } else {
                // last character of this subtag
                result = iter.next(c | 0x80);
                break;
            }
            c = next;
        }
    }
    switch (result) {
    case USTRINGTRIE_NO_MATCH: return -1;
    case USTRINGTRIE_NO_VALUE: return 0;
    case USTRINGTRIE_INTERMEDIATE_VALUE:
        U_ASSERT(iter.getValue() == SKIP_SCRIPT);
        return SKIP_SCRIPT;
    case USTRINGTRIE_FINAL_VALUE: return iter.getValue();
    default: return -1;
    }
}

// TODO(ICU-20777): Switch Locale/uloc_ likely-subtags API from the old code
// in loclikely.cpp to this new code, including activating this
// minimizeSubtags() function. The LocaleMatcher does not minimize.
#if 0
LSR XLikelySubtags::minimizeSubtags(const char *languageIn, const char *scriptIn,
                                    const char *regionIn, ULocale.Minimize fieldToFavor,
                                    UErrorCode &errorCode) const {
    LSR result = maximize(languageIn, scriptIn, regionIn);

    // We could try just a series of checks, like:
    // LSR result2 = addLikelySubtags(languageIn, "", "");
    // if result.equals(result2) return result2;
    // However, we can optimize 2 of the cases:
    //   (languageIn, "", "")
    //   (languageIn, "", regionIn)

    // value00 = lookup(result.language, "", "")
    BytesTrie iter = new BytesTrie(trie);
    int value = trieNext(iter, result.language, 0);
    U_ASSERT(value >= 0);
    if (value == 0) {
        value = trieNext(iter, "", 0);
        U_ASSERT(value >= 0);
        if (value == 0) {
            value = trieNext(iter, "", 0);
        }
    }
    U_ASSERT(value > 0);
    LSR value00 = lsrs[value];
    boolean favorRegionOk = false;
    if (result.script.equals(value00.script)) { //script is default
        if (result.region.equals(value00.region)) {
            return new LSR(result.language, "", "");
        } else if (fieldToFavor == ULocale.Minimize.FAVOR_REGION) {
            return new LSR(result.language, "", result.region);
        } else {
            favorRegionOk = true;
        }
    }

    // The last case is not as easy to optimize.
    // Maybe do later, but for now use the straightforward code.
    LSR result2 = maximize(languageIn, scriptIn, "");
    if (result2.equals(result)) {
        return new LSR(result.language, result.script, "");
    } else if (favorRegionOk) {
        return new LSR(result.language, "", result.region);
    }
    return result;
}
#endif

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1997-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  loclikely.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010feb25
*   created by: Markus W. Scherer
*
*   Code for miscellaneous locale-related resource bundle data access,
*   separated out from other .cpp files
*   that then do not depend on resource bundle code and this data.
*/









/*
 * Lookup a resource bundle table item with fallback on the table level.
 * Regular resource bundle lookups perform fallback to parent locale bundles
 * and eventually the root bundle, but only for top-level items.
 * This function takes the name of a top-level table and of an item in that table
 * and performs a lookup of both, falling back until a bundle contains a table
 * with this item.
 *
 * Note: Only the opening of entire bundles falls back through the default locale
 * before root. Once a bundle is open, item lookups do not go through the
 * default locale because that would result in a mix of languages that is
 * unpredictable to the programmer and most likely useless.
 */
U_CAPI const UChar * U_EXPORT2
uloc_getTableStringWithFallback(const char *path, const char *locale,
                              const char *tableKey, const char *subTableKey,
                              const char *itemKey,
                              int32_t *pLength,
                              UErrorCode *pErrorCode)
{
/*    char localeBuffer[ULOC_FULLNAME_CAPACITY*4];*/
    const UChar *item=NULL;
    UErrorCode errorCode;
    char explicitFallbackName[ULOC_FULLNAME_CAPACITY] = {0};

    /*
     * open the bundle for the current locale
     * this falls back through the locale's chain to root
     */
    errorCode=U_ZERO_ERROR;
    icu::LocalUResourceBundlePointer rb(ures_open(path, locale, &errorCode));

    if(U_FAILURE(errorCode)) {
        /* total failure, not even root could be opened */
        *pErrorCode=errorCode;
        return NULL;
    } else if(errorCode==U_USING_DEFAULT_WARNING ||
                (errorCode==U_USING_FALLBACK_WARNING && *pErrorCode!=U_USING_DEFAULT_WARNING)
    ) {
        /* set the "strongest" error code (success->fallback->default->failure) */
        *pErrorCode=errorCode;
    }

    for(;;){
        icu::StackUResourceBundle table;
        icu::StackUResourceBundle subTable;
        ures_getByKeyWithFallback(rb.getAlias(), tableKey, table.getAlias(), &errorCode);

        if (subTableKey != NULL) {
            /*
            ures_getByKeyWithFallback(table.getAlias(), subTableKey, subTable.getAlias(), &errorCode);
            item = ures_getStringByKeyWithFallback(subTable.getAlias(), itemKey, pLength, &errorCode);
            if(U_FAILURE(errorCode)){
                *pErrorCode = errorCode;
            }
            
            break;*/
            
            ures_getByKeyWithFallback(table.getAlias(), subTableKey, table.getAlias(), &errorCode);
        }
        if(U_SUCCESS(errorCode)){
            item = ures_getStringByKeyWithFallback(table.getAlias(), itemKey, pLength, &errorCode);
            if(U_FAILURE(errorCode)){
                const char* replacement = NULL;
                *pErrorCode = errorCode; /*save the errorCode*/
                errorCode = U_ZERO_ERROR;
                /* may be a deprecated code */
                if(uprv_strcmp(tableKey, "Countries")==0){
                    replacement =  uloc_getCurrentCountryID(itemKey);
                }else if(uprv_strcmp(tableKey, "Languages")==0){
                    replacement =  uloc_getCurrentLanguageID(itemKey);
                }
                /*pointer comparison is ok since uloc_getCurrentCountryID & uloc_getCurrentLanguageID return the key itself is replacement is not found*/
                if(replacement!=NULL && itemKey != replacement){
                    item = ures_getStringByKeyWithFallback(table.getAlias(), replacement, pLength, &errorCode);
                    if(U_SUCCESS(errorCode)){
                        *pErrorCode = errorCode;
                        break;
                    }
                }
            }else{
                break;
            }
        }
        
        if(U_FAILURE(errorCode)){    

            /* still can't figure out ?.. try the fallback mechanism */
            int32_t len = 0;
            const UChar* fallbackLocale =  NULL;
            *pErrorCode = errorCode;
            errorCode = U_ZERO_ERROR;

            fallbackLocale = ures_getStringByKeyWithFallback(table.getAlias(), "Fallback", &len, &errorCode);
            if(U_FAILURE(errorCode)){
               *pErrorCode = errorCode;
                break;
            }
            
            u_UCharsToChars(fallbackLocale, explicitFallbackName, len);
            
            /* guard against recursive fallback */
            if(uprv_strcmp(explicitFallbackName, locale)==0){
                *pErrorCode = U_INTERNAL_PROGRAM_ERROR;
                break;
            }
            rb.adoptInstead(ures_open(path, explicitFallbackName, &errorCode));
            if(U_FAILURE(errorCode)){
                *pErrorCode = errorCode;
                break;
            }
            /* succeeded in opening the fallback bundle .. continue and try to fetch the item */
        }else{
            break;
        }
    }

    return item;
}

static ULayoutType
_uloc_getOrientationHelper(const char* localeId,
                           const char* key,
                           UErrorCode *status)
{
    ULayoutType result = ULOC_LAYOUT_UNKNOWN;

    if (!U_FAILURE(*status)) {
        int32_t length = 0;
        char localeBuffer[ULOC_FULLNAME_CAPACITY];

        uloc_canonicalize(localeId, localeBuffer, sizeof(localeBuffer), status);

        if (!U_FAILURE(*status)) {
            const UChar* const value =
                uloc_getTableStringWithFallback(
                    NULL,
                    localeBuffer,
                    "layout",
                    NULL,
                    key,
                    &length,
                    status);

            if (!U_FAILURE(*status) && length != 0) {
                switch(value[0])
                {
                case 0x0062: /* 'b' */
                    result = ULOC_LAYOUT_BTT;
                    break;
                case 0x006C: /* 'l' */
                    result = ULOC_LAYOUT_LTR;
                    break;
                case 0x0072: /* 'r' */
                    result = ULOC_LAYOUT_RTL;
                    break;
                case 0x0074: /* 't' */
                    result = ULOC_LAYOUT_TTB;
                    break;
                default:
                    *status = U_INTERNAL_PROGRAM_ERROR;
                    break;
                }
            }
        }
    }

    return result;
}

U_CAPI ULayoutType U_EXPORT2
uloc_getCharacterOrientation(const char* localeId,
                             UErrorCode *status)
{
    return _uloc_getOrientationHelper(localeId, "characters", status);
}

/**
 * Get the layout line orientation for the specified locale.
 * 
 * @param localeID locale name
 * @param status Error status
 * @return an enum indicating the layout orientation for lines.
 */
U_CAPI ULayoutType U_EXPORT2
uloc_getLineOrientation(const char* localeId,
                        UErrorCode *status)
{
    return _uloc_getOrientationHelper(localeId, "lines", status);
}
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
//  *******************************************************************************
//  * Copyright (C) 2002-2014, International Business Machines Corporation and
//  * others. All Rights Reserved.
//  *******************************************************************************
//  */
// #include "unicode/utypes.h"

// #if !UCONFIG_NO_SERVICE || !UCONFIG_NO_TRANSLITERATION

// #include "unicode/resbund.h"
// #include "unicode/uenum.h"
// #include "cmemory.h"
// #include "ustrfmt.h"
// #include "locutil.h"
// #include "charstr.h"
// #include "ucln_cmn.h"
// #include "uassert.h"
// #include "umutex.h"

// // see LocaleUtility::getAvailableLocaleNames
// static icu::UInitOnce   LocaleUtilityInitOnce = U_INITONCE_INITIALIZER;
// static icu::Hashtable * LocaleUtility_cache = NULL;

// #define UNDERSCORE_CHAR ((UChar)0x005f)
// #define AT_SIGN_CHAR    ((UChar)64)
// #define PERIOD_CHAR     ((UChar)46)

// /*
//  ******************************************************************
//  */

// /**
//  * Release all static memory held by Locale Utility.
//  */
// U_CDECL_BEGIN
// static UBool U_CALLCONV service_cleanup(void) {
//     if (LocaleUtility_cache) {
//         delete LocaleUtility_cache;
//         LocaleUtility_cache = NULL;
//     }
//     return TRUE;
// }


// static void U_CALLCONV locale_utility_init(UErrorCode &status) {
//     using namespace icu;
//     U_ASSERT(LocaleUtility_cache == NULL);
//     ucln_common_registerCleanup(UCLN_COMMON_SERVICE, service_cleanup);
//     LocaleUtility_cache = new Hashtable(status);
//     if (U_FAILURE(status)) {
//         delete LocaleUtility_cache;
//         LocaleUtility_cache = NULL;
//         return;
//     }
//     if (LocaleUtility_cache == NULL) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
//     LocaleUtility_cache->setValueDeleter(uhash_deleteHashtable);
// }

// U_CDECL_END

// U_NAMESPACE_BEGIN

// UnicodeString&
// LocaleUtility::canonicalLocaleString(const UnicodeString* id, UnicodeString& result)
// {
//   if (id == NULL) {
//     result.setToBogus();
//   } else {
//     // Fix case only (no other changes) up to the first '@' or '.' or
//     // end of string, whichever comes first.  In 3.0 I changed this to
//     // stop at first '@' or '.'.  It used to run out to the end of
//     // string.  My fix makes the tests pass but is probably
//     // structurally incorrect.  See below.  [alan 3.0]

//     // TODO: Doug, you might want to revise this...
//     result = *id;
//     int32_t i = 0;
//     int32_t end = result.indexOf(AT_SIGN_CHAR);
//     int32_t n = result.indexOf(PERIOD_CHAR);
//     if (n >= 0 && n < end) {
//         end = n;
//     }
//     if (end < 0) {
//         end = result.length();
//     }
//     n = result.indexOf(UNDERSCORE_CHAR);
//     if (n < 0) {
//       n = end;
//     }
//     for (; i < n; ++i) {
//       UChar c = result.charAt(i);
//       if (c >= 0x0041 && c <= 0x005a) {
//         c += 0x20;
//         result.setCharAt(i, c);
//       }
//     }
//     for (n = end; i < n; ++i) {
//       UChar c = result.charAt(i);
//       if (c >= 0x0061 && c <= 0x007a) {
//         c -= 0x20;
//         result.setCharAt(i, c);
//       }
//     }
//   }
//   return result;

// #if 0
//     // This code does a proper full level 2 canonicalization of id.
//     // It's nasty to go from UChar to char to char to UChar -- but
//     // that's what you have to do to use the uloc_canonicalize
//     // function on UnicodeStrings.

//     // I ended up doing the alternate fix (see above) not for
//     // performance reasons, although performance will certainly be
//     // better, but because doing a full level 2 canonicalization
//     // causes some tests to fail.  [alan 3.0]

//     // TODO: Doug, you might want to revisit this...
//     result.setToBogus();
//     if (id != 0) {
//         int32_t buflen = id->length() + 8; // space for NUL
//         char* buf = (char*) uprv_malloc(buflen);
//         char* canon = (buf == 0) ? 0 : (char*) uprv_malloc(buflen);
//         if (buf != 0 && canon != 0) {
//             U_ASSERT(id->extract(0, INT32_MAX, buf, buflen) < buflen);
//             UErrorCode ec = U_ZERO_ERROR;
//             uloc_canonicalize(buf, canon, buflen, &ec);
//             if (U_SUCCESS(ec)) {
//                 result = UnicodeString(canon);
//             }
//         }
//         uprv_free(buf);
//         uprv_free(canon);
//     }
//     return result;
// #endif
// }

// Locale&
// LocaleUtility::initLocaleFromName(const UnicodeString& id, Locale& result)
// {
//     enum { BUFLEN = 128 }; // larger than ever needed

//     if (id.isBogus() || id.length() >= BUFLEN) {
//         result.setToBogus();
//     } else {
//         /*
//          * We need to convert from a UnicodeString to char * in order to
//          * create a Locale.
//          *
//          * Problem: Locale ID strings may contain '@' which is a variant
//          * character and cannot be handled by invariant-character conversion.
//          *
//          * Hack: Since ICU code can handle locale IDs with multiple encodings
//          * of '@' (at least for EBCDIC; it's not known to be a problem for
//          * ASCII-based systems),
//          * we use regular invariant-character conversion for everything else
//          * and manually convert U+0040 into a compiler-char-constant '@'.
//          * While this compilation-time constant may not match the runtime
//          * encoding of '@', it should be one of the encodings which ICU
//          * recognizes.
//          *
//          * There should be only at most one '@' in a locale ID.
//          */
//         char buffer[BUFLEN];
//         int32_t prev, i;
//         prev = 0;
//         for(;;) {
//             i = id.indexOf((UChar)0x40, prev);
//             if(i < 0) {
//                 // no @ between prev and the rest of the string
//                 id.extract(prev, INT32_MAX, buffer + prev, BUFLEN - prev, US_INV);
//                 break; // done
//             } else {
//                 // normal invariant-character conversion for text between @s
//                 id.extract(prev, i - prev, buffer + prev, BUFLEN - prev, US_INV);
//                 // manually "convert" U+0040 at id[i] into '@' at buffer[i]
//                 buffer[i] = '@';
//                 prev = i + 1;
//             }
//         }
//         result = Locale::createFromName(buffer);
//     }
//     return result;
// }

// UnicodeString&
// LocaleUtility::initNameFromLocale(const Locale& locale, UnicodeString& result)
// {
//     if (locale.isBogus()) {
//         result.setToBogus();
//     } else {
//         result.append(UnicodeString(locale.getName(), -1, US_INV));
//     }
//     return result;
// }

// const Hashtable*
// LocaleUtility::getAvailableLocaleNames(const UnicodeString& bundleID)
// {
//     // LocaleUtility_cache is a hash-of-hashes.  The top-level keys
//     // are path strings ('bundleID') passed to
//     // ures_openAvailableLocales.  The top-level values are
//     // second-level hashes.  The second-level keys are result strings
//     // from ures_openAvailableLocales.  The second-level values are
//     // garbage ((void*)1 or other random pointer).

//     UErrorCode status = U_ZERO_ERROR;
//     umtx_initOnce(LocaleUtilityInitOnce, locale_utility_init, status);
//     Hashtable *cache = LocaleUtility_cache;
//     if (cache == NULL) {
//         // Catastrophic failure.
//         return NULL;
//     }

//     Hashtable* htp;
//     umtx_lock(NULL);
//     htp = (Hashtable*) cache->get(bundleID);
//     umtx_unlock(NULL);

//     if (htp == NULL) {
//         htp = new Hashtable(status);
//         if (htp && U_SUCCESS(status)) {
//             CharString cbundleID;
//             cbundleID.appendInvariantChars(bundleID, status);
//             const char* path = cbundleID.isEmpty() ? NULL : cbundleID.data();
//             icu::LocalUEnumerationPointer uenum(ures_openAvailableLocales(path, &status));
//             for (;;) {
//                 const UChar* id = uenum_unext(uenum.getAlias(), NULL, &status);
//                 if (id == NULL) {
//                     break;
//                 }
//                 htp->put(UnicodeString(id), (void*)htp, status);
//             }
//             if (U_FAILURE(status)) {
//                 delete htp;
//                 return NULL;
//             }
//             umtx_lock(NULL);
//             Hashtable *t = static_cast<Hashtable *>(cache->get(bundleID));
//             if (t != NULL) {
//                 // Another thread raced through this code, creating the cache entry first.
//                 // Discard ours and return theirs.
//                 umtx_unlock(NULL);
//                 delete htp;
//                 htp = t;
//             } else {
//                 cache->put(bundleID, (void*)htp, status);
//                 umtx_unlock(NULL);
//             }
//         }
//     }
//     return htp;
// }

// UBool
// LocaleUtility::isFallbackOf(const UnicodeString& root, const UnicodeString& child)
// {
//     return child.indexOf(root) == 0 &&
//       (child.length() == root.length() ||
//        child.charAt(root.length()) == UNDERSCORE_CHAR);
// }

// U_NAMESPACE_END

// /* !UCONFIG_NO_SERVICE */
// #endif


// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html#License

// lsr.cpp
// created: 2019may08 Markus W. Scherer









U_NAMESPACE_BEGIN

LSR::LSR(char prefix, const char *lang, const char *scr, const char *r, UErrorCode &errorCode) :
        language(nullptr), script(nullptr), region(r),
        regionIndex(indexForRegion(region)) {
    if (U_SUCCESS(errorCode)) {
        CharString langScript;
        langScript.append(prefix, errorCode).append(lang, errorCode).append('\0', errorCode);
        int32_t scriptOffset = langScript.length();
        langScript.append(prefix, errorCode).append(scr, errorCode);
        owned = langScript.cloneData(errorCode);
        if (U_SUCCESS(errorCode)) {
            language = owned;
            script = owned + scriptOffset;
        }
    }
}

LSR::LSR(LSR &&other) U_NOEXCEPT :
        language(other.language), script(other.script), region(other.region), owned(other.owned),
        regionIndex(other.regionIndex), hashCode(other.hashCode) {
    if (owned != nullptr) {
        other.language = other.script = "";
        other.owned = nullptr;
        other.hashCode = 0;
    }
}

void LSR::deleteOwned() {
    uprv_free(owned);
}

LSR &LSR::operator=(LSR &&other) U_NOEXCEPT {
    this->~LSR();
    language = other.language;
    script = other.script;
    region = other.region;
    regionIndex = other.regionIndex;
    owned = other.owned;
    hashCode = other.hashCode;
    if (owned != nullptr) {
        other.language = other.script = "";
        other.owned = nullptr;
        other.hashCode = 0;
    }
    return *this;
}

UBool LSR::operator==(const LSR &other) const {
    return
        uprv_strcmp(language, other.language) == 0 &&
        uprv_strcmp(script, other.script) == 0 &&
        regionIndex == other.regionIndex &&
        // Compare regions if both are ill-formed (and their indexes are 0).
        (regionIndex > 0 || uprv_strcmp(region, other.region) == 0);
}

int32_t LSR::indexForRegion(const char *region) {
    int32_t c = region[0];
    int32_t a = c - '0';
    if (0 <= a && a <= 9) {  // digits: "419"
        int32_t b = region[1] - '0';
        if (b < 0 || 9 < b) { return 0; }
        c = region[2] - '0';
        if (c < 0 || 9 < c || region[3] != 0) { return 0; }
        return (10 * a + b) * 10 + c + 1;
    } else {  // letters: "DE"
        a = uprv_upperOrdinal(c);
        if (a < 0 || 25 < a) { return 0; }
        int32_t b = uprv_upperOrdinal(region[1]);
        if (b < 0 || 25 < b || region[2] != 0) { return 0; }
        return 26 * a + b + 1001;
    }
    return 0;
}

LSR &LSR::setHashCode() {
    if (hashCode == 0) {
        hashCode =
            (ustr_hashCharsN(language, static_cast<int32_t>(uprv_strlen(language))) * 37 +
            ustr_hashCharsN(script, static_cast<int32_t>(uprv_strlen(script)))) * 37 +
            regionIndex;
    }
    return *this;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2011-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  messagepattern.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011mar14
*   created by: Markus W. Scherer
*/



#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2011-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  messagepattern.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011mar14
*   created by: Markus W. Scherer
*/


/**
 * \file
 * \brief C++ API: MessagePattern class: Parses and represents ICU MessageFormat patterns.
 */



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING




/**
 * Mode for when an apostrophe starts quoted literal text for MessageFormat output.
 * The default is DOUBLE_OPTIONAL unless overridden via uconfig.h
 * (UCONFIG_MSGPAT_DEFAULT_APOSTROPHE_MODE).
 * <p>
 * A pair of adjacent apostrophes always results in a single apostrophe in the output,
 * even when the pair is between two single, text-quoting apostrophes.
 * <p>
 * The following table shows examples of desired MessageFormat.format() output
 * with the pattern strings that yield that output.
 * <p>
 * <table>
 *   <tr>
 *     <th>Desired output</th>
 *     <th>DOUBLE_OPTIONAL</th>
 *     <th>DOUBLE_REQUIRED</th>
 *   </tr>
 *   <tr>
 *     <td>I see {many}</td>
 *     <td>I see '{many}'</td>
 *     <td>(same)</td>
 *   </tr>
 *   <tr>
 *     <td>I said {'Wow!'}</td>
 *     <td>I said '{''Wow!''}'</td>
 *     <td>(same)</td>
 *   </tr>
 *   <tr>
 *     <td>I don't know</td>
 *     <td>I don't know OR<br> I don''t know</td>
 *     <td>I don''t know</td>
 *   </tr>
 * </table>
 * @stable ICU 4.8
 * @see UCONFIG_MSGPAT_DEFAULT_APOSTROPHE_MODE
 */
enum UMessagePatternApostropheMode {
    /**
     * A literal apostrophe is represented by
     * either a single or a double apostrophe pattern character.
     * Within a MessageFormat pattern, a single apostrophe only starts quoted literal text
     * if it immediately precedes a curly brace {},
     * or a pipe symbol | if inside a choice format,
     * or a pound symbol # if inside a plural format.
     * <p>
     * This is the default behavior starting with ICU 4.8.
     * @stable ICU 4.8
     */
    UMSGPAT_APOS_DOUBLE_OPTIONAL,
    /**
     * A literal apostrophe must be represented by
     * a double apostrophe pattern character.
     * A single apostrophe always starts quoted literal text.
     * <p>
     * This is the behavior of ICU 4.6 and earlier, and of the JDK.
     * @stable ICU 4.8
     */
    UMSGPAT_APOS_DOUBLE_REQUIRED
};
/**
 * @stable ICU 4.8
 */
typedef enum UMessagePatternApostropheMode UMessagePatternApostropheMode;

/**
 * MessagePattern::Part type constants.
 * @stable ICU 4.8
 */
enum UMessagePatternPartType {
    /**
     * Start of a message pattern (main or nested).
     * The length is 0 for the top-level message
     * and for a choice argument sub-message, otherwise 1 for the '{'.
     * The value indicates the nesting level, starting with 0 for the main message.
     * <p>
     * There is always a later MSG_LIMIT part.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_MSG_START,
    /**
     * End of a message pattern (main or nested).
     * The length is 0 for the top-level message and
     * the last sub-message of a choice argument,
     * otherwise 1 for the '}' or (in a choice argument style) the '|'.
     * The value indicates the nesting level, starting with 0 for the main message.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_MSG_LIMIT,
    /**
     * Indicates a substring of the pattern string which is to be skipped when formatting.
     * For example, an apostrophe that begins or ends quoted text
     * would be indicated with such a part.
     * The value is undefined and currently always 0.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_SKIP_SYNTAX,
    /**
     * Indicates that a syntax character needs to be inserted for auto-quoting.
     * The length is 0.
     * The value is the character code of the insertion character. (U+0027=APOSTROPHE)
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_INSERT_CHAR,
    /**
     * Indicates a syntactic (non-escaped) # symbol in a plural variant.
     * When formatting, replace this part's substring with the
     * (value-offset) for the plural argument value.
     * The value is undefined and currently always 0.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_REPLACE_NUMBER,
    /**
     * Start of an argument.
     * The length is 1 for the '{'.
     * The value is the ordinal value of the ArgType. Use getArgType().
     * <p>
     * This part is followed by either an ARG_NUMBER or ARG_NAME,
     * followed by optional argument sub-parts (see UMessagePatternArgType constants)
     * and finally an ARG_LIMIT part.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_ARG_START,
    /**
     * End of an argument.
     * The length is 1 for the '}'.
     * The value is the ordinal value of the ArgType. Use getArgType().
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_ARG_LIMIT,
    /**
     * The argument number, provided by the value.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_ARG_NUMBER,
    /**
     * The argument name.
     * The value is undefined and currently always 0.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_ARG_NAME,
    /**
     * The argument type.
     * The value is undefined and currently always 0.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_ARG_TYPE,
    /**
     * The argument style text.
     * The value is undefined and currently always 0.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_ARG_STYLE,
    /**
     * A selector substring in a "complex" argument style.
     * The value is undefined and currently always 0.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_ARG_SELECTOR,
    /**
     * An integer value, for example the offset or an explicit selector value
     * in a PluralFormat style.
     * The part value is the integer value.
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_ARG_INT,
    /**
     * A numeric value, for example the offset or an explicit selector value
     * in a PluralFormat style.
     * The part value is an index into an internal array of numeric values;
     * use getNumericValue().
     * @stable ICU 4.8
     */
    UMSGPAT_PART_TYPE_ARG_DOUBLE
};
/**
 * @stable ICU 4.8
 */
typedef enum UMessagePatternPartType UMessagePatternPartType;

/**
 * Argument type constants.
 * Returned by Part.getArgType() for ARG_START and ARG_LIMIT parts.
 *
 * Messages nested inside an argument are each delimited by MSG_START and MSG_LIMIT,
 * with a nesting level one greater than the surrounding message.
 * @stable ICU 4.8
 */
enum UMessagePatternArgType {
    /**
     * The argument has no specified type.
     * @stable ICU 4.8
     */
    UMSGPAT_ARG_TYPE_NONE,
    /**
     * The argument has a "simple" type which is provided by the ARG_TYPE part.
     * An ARG_STYLE part might follow that.
     * @stable ICU 4.8
     */
    UMSGPAT_ARG_TYPE_SIMPLE,
    /**
     * The argument is a ChoiceFormat with one or more
     * ((ARG_INT | ARG_DOUBLE), ARG_SELECTOR, message) tuples.
     * @stable ICU 4.8
     */
    UMSGPAT_ARG_TYPE_CHOICE,
    /**
     * The argument is a cardinal-number PluralFormat with an optional ARG_INT or ARG_DOUBLE offset
     * (e.g., offset:1)
     * and one or more (ARG_SELECTOR [explicit-value] message) tuples.
     * If the selector has an explicit value (e.g., =2), then
     * that value is provided by the ARG_INT or ARG_DOUBLE part preceding the message.
     * Otherwise the message immediately follows the ARG_SELECTOR.
     * @stable ICU 4.8
     */
    UMSGPAT_ARG_TYPE_PLURAL,
    /**
     * The argument is a SelectFormat with one or more (ARG_SELECTOR, message) pairs.
     * @stable ICU 4.8
     */
    UMSGPAT_ARG_TYPE_SELECT,
    /**
     * The argument is an ordinal-number PluralFormat
     * with the same style parts sequence and semantics as UMSGPAT_ARG_TYPE_PLURAL.
     * @stable ICU 50
     */
    UMSGPAT_ARG_TYPE_SELECTORDINAL
};
/**
 * @stable ICU 4.8
 */
typedef enum UMessagePatternArgType UMessagePatternArgType;

/**
 * \def UMSGPAT_ARG_TYPE_HAS_PLURAL_STYLE
 * Returns TRUE if the argument type has a plural style part sequence and semantics,
 * for example UMSGPAT_ARG_TYPE_PLURAL and UMSGPAT_ARG_TYPE_SELECTORDINAL.
 * @stable ICU 50
 */
#define UMSGPAT_ARG_TYPE_HAS_PLURAL_STYLE(argType) \
    ((argType)==UMSGPAT_ARG_TYPE_PLURAL || (argType)==UMSGPAT_ARG_TYPE_SELECTORDINAL)

enum {
    /**
     * Return value from MessagePattern.validateArgumentName() for when
     * the string is a valid "pattern identifier" but not a number.
     * @stable ICU 4.8
     */
    UMSGPAT_ARG_NAME_NOT_NUMBER=-1,

    /**
     * Return value from MessagePattern.validateArgumentName() for when
     * the string is invalid.
     * It might not be a valid "pattern identifier",
     * or it have only ASCII digits but there is a leading zero or the number is too large.
     * @stable ICU 4.8
     */
    UMSGPAT_ARG_NAME_NOT_VALID=-2
};

/**
 * Special value that is returned by getNumericValue(Part) when no
 * numeric value is defined for a part.
 * @see MessagePattern.getNumericValue()
 * @stable ICU 4.8
 */
#define UMSGPAT_NO_NUMERIC_VALUE ((double)(-123456789))

U_NAMESPACE_BEGIN

class MessagePatternDoubleList;
class MessagePatternPartsList;

/**
 * Parses and represents ICU MessageFormat patterns.
 * Also handles patterns for ChoiceFormat, PluralFormat and SelectFormat.
 * Used in the implementations of those classes as well as in tools
 * for message validation, translation and format conversion.
 * <p>
 * The parser handles all syntax relevant for identifying message arguments.
 * This includes "complex" arguments whose style strings contain
 * nested MessageFormat pattern substrings.
 * For "simple" arguments (with no nested MessageFormat pattern substrings),
 * the argument style is not parsed any further.
 * <p>
 * The parser handles named and numbered message arguments and allows both in one message.
 * <p>
 * Once a pattern has been parsed successfully, iterate through the parsed data
 * with countParts(), getPart() and related methods.
 * <p>
 * The data logically represents a parse tree, but is stored and accessed
 * as a list of "parts" for fast and simple parsing and to minimize object allocations.
 * Arguments and nested messages are best handled via recursion.
 * For every _START "part", MessagePattern.getLimitPartIndex() efficiently returns
 * the index of the corresponding _LIMIT "part".
 * <p>
 * List of "parts":
 * <pre>
 * message = MSG_START (SKIP_SYNTAX | INSERT_CHAR | REPLACE_NUMBER | argument)* MSG_LIMIT
 * argument = noneArg | simpleArg | complexArg
 * complexArg = choiceArg | pluralArg | selectArg
 *
 * noneArg = ARG_START.NONE (ARG_NAME | ARG_NUMBER) ARG_LIMIT.NONE
 * simpleArg = ARG_START.SIMPLE (ARG_NAME | ARG_NUMBER) ARG_TYPE [ARG_STYLE] ARG_LIMIT.SIMPLE
 * choiceArg = ARG_START.CHOICE (ARG_NAME | ARG_NUMBER) choiceStyle ARG_LIMIT.CHOICE
 * pluralArg = ARG_START.PLURAL (ARG_NAME | ARG_NUMBER) pluralStyle ARG_LIMIT.PLURAL
 * selectArg = ARG_START.SELECT (ARG_NAME | ARG_NUMBER) selectStyle ARG_LIMIT.SELECT
 *
 * choiceStyle = ((ARG_INT | ARG_DOUBLE) ARG_SELECTOR message)+
 * pluralStyle = [ARG_INT | ARG_DOUBLE] (ARG_SELECTOR [ARG_INT | ARG_DOUBLE] message)+
 * selectStyle = (ARG_SELECTOR message)+
 * </pre>
 * <ul>
 *   <li>Literal output text is not represented directly by "parts" but accessed
 *       between parts of a message, from one part's getLimit() to the next part's getIndex().
 *   <li><code>ARG_START.CHOICE</code> stands for an ARG_START Part with ArgType CHOICE.
 *   <li>In the choiceStyle, the ARG_SELECTOR has the '<', the '#' or
 *       the less-than-or-equal-to sign (U+2264).
 *   <li>In the pluralStyle, the first, optional numeric Part has the "offset:" value.
 *       The optional numeric Part between each (ARG_SELECTOR, message) pair
 *       is the value of an explicit-number selector like "=2",
 *       otherwise the selector is a non-numeric identifier.
 *   <li>The REPLACE_NUMBER Part can occur only in an immediate sub-message of the pluralStyle.
 * </ul>
 * <p>
 * This class is not intended for public subclassing.
 *
 * @stable ICU 4.8
 */
class U_COMMON_API MessagePattern : public UObject {
public:
    /**
     * Constructs an empty MessagePattern with default UMessagePatternApostropheMode.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @stable ICU 4.8
     */
    MessagePattern(UErrorCode &errorCode);

    /**
     * Constructs an empty MessagePattern.
     * @param mode Explicit UMessagePatternApostropheMode.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @stable ICU 4.8
     */
    MessagePattern(UMessagePatternApostropheMode mode, UErrorCode &errorCode);

    /**
     * Constructs a MessagePattern with default UMessagePatternApostropheMode and
     * parses the MessageFormat pattern string.
     * @param pattern a MessageFormat pattern string
     * @param parseError Struct to receive information on the position
     *                   of an error within the pattern.
     *                   Can be NULL.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * TODO: turn @throws into UErrorCode specifics?
     * @throws IllegalArgumentException for syntax errors in the pattern string
     * @throws IndexOutOfBoundsException if certain limits are exceeded
     *         (e.g., argument number too high, argument name too long, etc.)
     * @throws NumberFormatException if a number could not be parsed
     * @stable ICU 4.8
     */
    MessagePattern(const UnicodeString &pattern, UParseError *parseError, UErrorCode &errorCode);

    /**
     * Copy constructor.
     * @param other Object to copy.
     * @stable ICU 4.8
     */
    MessagePattern(const MessagePattern &other);

    /**
     * Assignment operator.
     * @param other Object to copy.
     * @return *this=other
     * @stable ICU 4.8
     */
    MessagePattern &operator=(const MessagePattern &other);

    /**
     * Destructor.
     * @stable ICU 4.8
     */
    virtual ~MessagePattern();

    /**
     * Parses a MessageFormat pattern string.
     * @param pattern a MessageFormat pattern string
     * @param parseError Struct to receive information on the position
     *                   of an error within the pattern.
     *                   Can be NULL.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return *this
     * @throws IllegalArgumentException for syntax errors in the pattern string
     * @throws IndexOutOfBoundsException if certain limits are exceeded
     *         (e.g., argument number too high, argument name too long, etc.)
     * @throws NumberFormatException if a number could not be parsed
     * @stable ICU 4.8
     */
    MessagePattern &parse(const UnicodeString &pattern,
                          UParseError *parseError, UErrorCode &errorCode);

    /**
     * Parses a ChoiceFormat pattern string.
     * @param pattern a ChoiceFormat pattern string
     * @param parseError Struct to receive information on the position
     *                   of an error within the pattern.
     *                   Can be NULL.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return *this
     * @throws IllegalArgumentException for syntax errors in the pattern string
     * @throws IndexOutOfBoundsException if certain limits are exceeded
     *         (e.g., argument number too high, argument name too long, etc.)
     * @throws NumberFormatException if a number could not be parsed
     * @stable ICU 4.8
     */
    MessagePattern &parseChoiceStyle(const UnicodeString &pattern,
                                     UParseError *parseError, UErrorCode &errorCode);

    /**
     * Parses a PluralFormat pattern string.
     * @param pattern a PluralFormat pattern string
     * @param parseError Struct to receive information on the position
     *                   of an error within the pattern.
     *                   Can be NULL.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return *this
     * @throws IllegalArgumentException for syntax errors in the pattern string
     * @throws IndexOutOfBoundsException if certain limits are exceeded
     *         (e.g., argument number too high, argument name too long, etc.)
     * @throws NumberFormatException if a number could not be parsed
     * @stable ICU 4.8
     */
    MessagePattern &parsePluralStyle(const UnicodeString &pattern,
                                     UParseError *parseError, UErrorCode &errorCode);

    /**
     * Parses a SelectFormat pattern string.
     * @param pattern a SelectFormat pattern string
     * @param parseError Struct to receive information on the position
     *                   of an error within the pattern.
     *                   Can be NULL.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return *this
     * @throws IllegalArgumentException for syntax errors in the pattern string
     * @throws IndexOutOfBoundsException if certain limits are exceeded
     *         (e.g., argument number too high, argument name too long, etc.)
     * @throws NumberFormatException if a number could not be parsed
     * @stable ICU 4.8
     */
    MessagePattern &parseSelectStyle(const UnicodeString &pattern,
                                     UParseError *parseError, UErrorCode &errorCode);

    /**
     * Clears this MessagePattern.
     * countParts() will return 0.
     * @stable ICU 4.8
     */
    void clear();

    /**
     * Clears this MessagePattern and sets the UMessagePatternApostropheMode.
     * countParts() will return 0.
     * @param mode The new UMessagePatternApostropheMode.
     * @stable ICU 4.8
     */
    void clearPatternAndSetApostropheMode(UMessagePatternApostropheMode mode) {
        clear();
        aposMode=mode;
    }

    /**
     * @param other another object to compare with.
     * @return TRUE if this object is equivalent to the other one.
     * @stable ICU 4.8
     */
    UBool operator==(const MessagePattern &other) const;

    /**
     * @param other another object to compare with.
     * @return FALSE if this object is equivalent to the other one.
     * @stable ICU 4.8
     */
    inline UBool operator!=(const MessagePattern &other) const {
        return !operator==(other);
    }

    /**
     * @return A hash code for this object.
     * @stable ICU 4.8
     */
    int32_t hashCode() const;

    /**
     * @return this instance's UMessagePatternApostropheMode.
     * @stable ICU 4.8
     */
    UMessagePatternApostropheMode getApostropheMode() const {
        return aposMode;
    }

    // Java has package-private jdkAposMode() here.
    // In C++, this is declared in the MessageImpl class.

    /**
     * @return the parsed pattern string (null if none was parsed).
     * @stable ICU 4.8
     */
    const UnicodeString &getPatternString() const {
        return msg;
    }

    /**
     * Does the parsed pattern have named arguments like {first_name}?
     * @return TRUE if the parsed pattern has at least one named argument.
     * @stable ICU 4.8
     */
    UBool hasNamedArguments() const {
        return hasArgNames;
    }

    /**
     * Does the parsed pattern have numbered arguments like {2}?
     * @return TRUE if the parsed pattern has at least one numbered argument.
     * @stable ICU 4.8
     */
    UBool hasNumberedArguments() const {
        return hasArgNumbers;
    }

    /**
     * Validates and parses an argument name or argument number string.
     * An argument name must be a "pattern identifier", that is, it must contain
     * no Unicode Pattern_Syntax or Pattern_White_Space characters.
     * If it only contains ASCII digits, then it must be a small integer with no leading zero.
     * @param name Input string.
     * @return &gt;=0 if the name is a valid number,
     *         ARG_NAME_NOT_NUMBER (-1) if it is a "pattern identifier" but not all ASCII digits,
     *         ARG_NAME_NOT_VALID (-2) if it is neither.
     * @stable ICU 4.8
     */
    static int32_t validateArgumentName(const UnicodeString &name);

    /**
     * Returns a version of the parsed pattern string where each ASCII apostrophe
     * is doubled (escaped) if it is not already, and if it is not interpreted as quoting syntax.
     * <p>
     * For example, this turns "I don't '{know}' {gender,select,female{h''er}other{h'im}}."
     * into "I don''t '{know}' {gender,select,female{h''er}other{h''im}}."
     * @return the deep-auto-quoted version of the parsed pattern string.
     * @see MessageFormat.autoQuoteApostrophe()
     * @stable ICU 4.8
     */
    UnicodeString autoQuoteApostropheDeep() const;

    class Part;

    /**
     * Returns the number of "parts" created by parsing the pattern string.
     * Returns 0 if no pattern has been parsed or clear() was called.
     * @return the number of pattern parts.
     * @stable ICU 4.8
     */
    int32_t countParts() const {
        return partsLength;
    }

    /**
     * Gets the i-th pattern "part".
     * @param i The index of the Part data. (0..countParts()-1)
     * @return the i-th pattern "part".
     * @stable ICU 4.8
     */
    const Part &getPart(int32_t i) const {
        return parts[i];
    }

    /**
     * Returns the UMessagePatternPartType of the i-th pattern "part".
     * Convenience method for getPart(i).getType().
     * @param i The index of the Part data. (0..countParts()-1)
     * @return The UMessagePatternPartType of the i-th Part.
     * @stable ICU 4.8
     */
    UMessagePatternPartType getPartType(int32_t i) const {
        return getPart(i).type;
    }

    /**
     * Returns the pattern index of the specified pattern "part".
     * Convenience method for getPart(partIndex).getIndex().
     * @param partIndex The index of the Part data. (0..countParts()-1)
     * @return The pattern index of this Part.
     * @stable ICU 4.8
     */
    int32_t getPatternIndex(int32_t partIndex) const {
        return getPart(partIndex).index;
    }

    /**
     * Returns the substring of the pattern string indicated by the Part.
     * Convenience method for getPatternString().substring(part.getIndex(), part.getLimit()).
     * @param part a part of this MessagePattern.
     * @return the substring associated with part.
     * @stable ICU 4.8
     */
    UnicodeString getSubstring(const Part &part) const {
        return msg.tempSubString(part.index, part.length);
    }

    /**
     * Compares the part's substring with the input string s.
     * @param part a part of this MessagePattern.
     * @param s a string.
     * @return TRUE if getSubstring(part).equals(s).
     * @stable ICU 4.8
     */
    UBool partSubstringMatches(const Part &part, const UnicodeString &s) const {
        return 0==msg.compare(part.index, part.length, s);
    }

    /**
     * Returns the numeric value associated with an ARG_INT or ARG_DOUBLE.
     * @param part a part of this MessagePattern.
     * @return the part's numeric value, or UMSGPAT_NO_NUMERIC_VALUE if this is not a numeric part.
     * @stable ICU 4.8
     */
    double getNumericValue(const Part &part) const;

    /**
     * Returns the "offset:" value of a PluralFormat argument, or 0 if none is specified.
     * @param pluralStart the index of the first PluralFormat argument style part. (0..countParts()-1)
     * @return the "offset:" value.
     * @stable ICU 4.8
     */
    double getPluralOffset(int32_t pluralStart) const;

    /**
     * Returns the index of the ARG|MSG_LIMIT part corresponding to the ARG|MSG_START at start.
     * @param start The index of some Part data (0..countParts()-1);
     *        this Part should be of Type ARG_START or MSG_START.
     * @return The first i>start where getPart(i).getType()==ARG|MSG_LIMIT at the same nesting level,
     *         or start itself if getPartType(msgStart)!=ARG|MSG_START.
     * @stable ICU 4.8
     */
    int32_t getLimitPartIndex(int32_t start) const {
        int32_t limit=getPart(start).limitPartIndex;
        if(limit<start) {
            return start;
        }
        return limit;
    }

    /**
     * A message pattern "part", representing a pattern parsing event.
     * There is a part for the start and end of a message or argument,
     * for quoting and escaping of and with ASCII apostrophes,
     * and for syntax elements of "complex" arguments.
     * @stable ICU 4.8
     */
    class Part : public UMemory {
    public:
        /**
         * Default constructor, do not use.
         * @internal
         */
        Part() {}

        /**
         * Returns the type of this part.
         * @return the part type.
         * @stable ICU 4.8
         */
        UMessagePatternPartType getType() const {
            return type;
        }

        /**
         * Returns the pattern string index associated with this Part.
         * @return this part's pattern string index.
         * @stable ICU 4.8
         */
        int32_t getIndex() const {
            return index;
        }

        /**
         * Returns the length of the pattern substring associated with this Part.
         * This is 0 for some parts.
         * @return this part's pattern substring length.
         * @stable ICU 4.8
         */
        int32_t getLength() const {
            return length;
        }

        /**
         * Returns the pattern string limit (exclusive-end) index associated with this Part.
         * Convenience method for getIndex()+getLength().
         * @return this part's pattern string limit index, same as getIndex()+getLength().
         * @stable ICU 4.8
         */
        int32_t getLimit() const {
            return index+length;
        }

        /**
         * Returns a value associated with this part.
         * See the documentation of each part type for details.
         * @return the part value.
         * @stable ICU 4.8
         */
        int32_t getValue() const {
            return value;
        }

        /**
         * Returns the argument type if this part is of type ARG_START or ARG_LIMIT,
         * otherwise UMSGPAT_ARG_TYPE_NONE.
         * @return the argument type for this part.
         * @stable ICU 4.8
         */
        UMessagePatternArgType getArgType() const {
            UMessagePatternPartType msgType=getType();
            if(msgType ==UMSGPAT_PART_TYPE_ARG_START || msgType ==UMSGPAT_PART_TYPE_ARG_LIMIT) {
                return (UMessagePatternArgType)value;
            } else {
                return UMSGPAT_ARG_TYPE_NONE;
            }
        }

        /**
         * Indicates whether the Part type has a numeric value.
         * If so, then that numeric value can be retrieved via MessagePattern.getNumericValue().
         * @param type The Part type to be tested.
         * @return TRUE if the Part type has a numeric value.
         * @stable ICU 4.8
         */
        static UBool hasNumericValue(UMessagePatternPartType type) {
            return type==UMSGPAT_PART_TYPE_ARG_INT || type==UMSGPAT_PART_TYPE_ARG_DOUBLE;
        }

        /**
         * @param other another object to compare with.
         * @return TRUE if this object is equivalent to the other one.
         * @stable ICU 4.8
         */
        UBool operator==(const Part &other) const;

        /**
         * @param other another object to compare with.
         * @return FALSE if this object is equivalent to the other one.
         * @stable ICU 4.8
         */
        inline UBool operator!=(const Part &other) const {
            return !operator==(other);
        }

        /**
         * @return A hash code for this object.
         * @stable ICU 4.8
         */
        int32_t hashCode() const {
            return ((type*37+index)*37+length)*37+value;
        }

    private:
        friend class MessagePattern;

        static const int32_t MAX_LENGTH=0xffff;
        static const int32_t MAX_VALUE=0x7fff;

        // Some fields are not final because they are modified during pattern parsing.
        // After pattern parsing, the parts are effectively immutable.
        UMessagePatternPartType type;
        int32_t index;
        uint16_t length;
        int16_t value;
        int32_t limitPartIndex;
    };

private:
    void preParse(const UnicodeString &pattern, UParseError *parseError, UErrorCode &errorCode);

    void postParse();

    int32_t parseMessage(int32_t index, int32_t msgStartLength,
                         int32_t nestingLevel, UMessagePatternArgType parentType,
                         UParseError *parseError, UErrorCode &errorCode);

    int32_t parseArg(int32_t index, int32_t argStartLength, int32_t nestingLevel,
                     UParseError *parseError, UErrorCode &errorCode);

    int32_t parseSimpleStyle(int32_t index, UParseError *parseError, UErrorCode &errorCode);

    int32_t parseChoiceStyle(int32_t index, int32_t nestingLevel,
                             UParseError *parseError, UErrorCode &errorCode);

    int32_t parsePluralOrSelectStyle(UMessagePatternArgType argType, int32_t index, int32_t nestingLevel,
                                     UParseError *parseError, UErrorCode &errorCode);

    /**
     * Validates and parses an argument name or argument number string.
     * This internal method assumes that the input substring is a "pattern identifier".
     * @return &gt;=0 if the name is a valid number,
     *         ARG_NAME_NOT_NUMBER (-1) if it is a "pattern identifier" but not all ASCII digits,
     *         ARG_NAME_NOT_VALID (-2) if it is neither.
     * @see #validateArgumentName(String)
     */
    static int32_t parseArgNumber(const UnicodeString &s, int32_t start, int32_t limit);

    int32_t parseArgNumber(int32_t start, int32_t limit) {
        return parseArgNumber(msg, start, limit);
    }

    /**
     * Parses a number from the specified message substring.
     * @param start start index into the message string
     * @param limit limit index into the message string, must be start<limit
     * @param allowInfinity TRUE if U+221E is allowed (for ChoiceFormat)
     * @param parseError
     * @param errorCode
     */
    void parseDouble(int32_t start, int32_t limit, UBool allowInfinity,
                     UParseError *parseError, UErrorCode &errorCode);

    // Java has package-private appendReducedApostrophes() here.
    // In C++, this is declared in the MessageImpl class.

    int32_t skipWhiteSpace(int32_t index);

    int32_t skipIdentifier(int32_t index);

    /**
     * Skips a sequence of characters that could occur in a double value.
     * Does not fully parse or validate the value.
     */
    int32_t skipDouble(int32_t index);

    static UBool isArgTypeChar(UChar32 c);

    UBool isChoice(int32_t index);

    UBool isPlural(int32_t index);

    UBool isSelect(int32_t index);

    UBool isOrdinal(int32_t index);

    /**
     * @return TRUE if we are inside a MessageFormat (sub-)pattern,
     *         as opposed to inside a top-level choice/plural/select pattern.
     */
    UBool inMessageFormatPattern(int32_t nestingLevel);

    /**
     * @return TRUE if we are in a MessageFormat sub-pattern
     *         of a top-level ChoiceFormat pattern.
     */
    UBool inTopLevelChoiceMessage(int32_t nestingLevel, UMessagePatternArgType parentType);

    void addPart(UMessagePatternPartType type, int32_t index, int32_t length,
                 int32_t value, UErrorCode &errorCode);

    void addLimitPart(int32_t start,
                      UMessagePatternPartType type, int32_t index, int32_t length,
                      int32_t value, UErrorCode &errorCode);

    void addArgDoublePart(double numericValue, int32_t start, int32_t length, UErrorCode &errorCode);

    void setParseError(UParseError *parseError, int32_t index);

    UBool init(UErrorCode &errorCode);
    UBool copyStorage(const MessagePattern &other, UErrorCode &errorCode);

    UMessagePatternApostropheMode aposMode;
    UnicodeString msg;
    // ArrayList<Part> parts=new ArrayList<Part>();
    MessagePatternPartsList *partsList;
    Part *parts;
    int32_t partsLength;
    // ArrayList<Double> numericValues;
    MessagePatternDoubleList *numericValuesList;
    double *numericValues;
    int32_t numericValuesLength;
    UBool hasArgNames;
    UBool hasArgNumbers;
    UBool needsAutoQuoting;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_FORMATTING

#endif /* U_SHOW_CPLUSPLUS_API */






// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  messageimpl.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011apr04
*   created by: Markus W. Scherer
*/




#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

/**
 * Helper functions for use of MessagePattern.
 * In Java, these are package-private methods in MessagePattern itself.
 * In C++, they are declared here and implemented in messagepattern.cpp.
 */
class U_COMMON_API MessageImpl {
public:
    /**
     * @return TRUE if getApostropheMode()==UMSGPAT_APOS_DOUBLE_REQUIRED
     */
    static UBool jdkAposMode(const MessagePattern &msgPattern) {
        return msgPattern.getApostropheMode()==UMSGPAT_APOS_DOUBLE_REQUIRED;
    }

    /**
     * Appends the s[start, limit[ substring to sb, but with only half of the apostrophes
     * according to JDK pattern behavior.
     */
    static void appendReducedApostrophes(const UnicodeString &s, int32_t start, int32_t limit,
                                         UnicodeString &sb);

    /**
     * Appends the sub-message to the result string.
     * Omits SKIP_SYNTAX and appends whole arguments using appendReducedApostrophes().
     */
    static UnicodeString &appendSubMessageWithoutSkipSyntax(const MessagePattern &msgPattern,
                                                            int32_t msgStart,
                                                            UnicodeString &result);

private:
    MessageImpl();  // no constructor: all static methods
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  patternprops.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011mar13
*   created by: Markus W. Scherer
*/





U_NAMESPACE_BEGIN

/**
 * Implements the immutable Unicode properties Pattern_Syntax and Pattern_White_Space.
 * Hardcodes these properties, does not load data, does not depend on other ICU classes.
 * <p>
 * Note: Both properties include ASCII as well as non-ASCII, non-Latin-1 code points,
 * and both properties only include BMP code points (no supplementary ones).
 * Pattern_Syntax includes some unassigned code points.
 * <p>
 * [:Pattern_White_Space:] =
 *   [\u0009-\u000D\ \u0085\u200E\u200F\u2028\u2029]
 * <p>
 * [:Pattern_Syntax:] =
 *   [!-/\:-@\[-\^`\{-~\u00A1-\u00A7\u00A9\u00AB\u00AC\u00AE
 *    \u00B0\u00B1\u00B6\u00BB\u00BF\u00D7\u00F7
 *    \u2010-\u2027\u2030-\u203E\u2041-\u2053\u2055-\u205E
 *    \u2190-\u245F\u2500-\u2775\u2794-\u2BFF\u2E00-\u2E7F
 *    \u3001-\u3003\u3008-\u3020\u3030\uFD3E\uFD3F\uFE45\uFE46]
 * @author mscherer
 */
class U_COMMON_API PatternProps {
public:
    /**
     * @return TRUE if c is a Pattern_Syntax code point.
     */
    static UBool isSyntax(UChar32 c);

    /**
     * @return TRUE if c is a Pattern_Syntax or Pattern_White_Space code point.
     */
    static UBool isSyntaxOrWhiteSpace(UChar32 c);

    /**
     * @return TRUE if c is a Pattern_White_Space character.
     */
    static UBool isWhiteSpace(UChar32 c);

    /**
     * Skips over Pattern_White_Space starting at s.
     * @return The smallest pointer at or after s with a non-white space character.
     */
    static const UChar *skipWhiteSpace(const UChar *s, int32_t length);

    /**
     * Skips over Pattern_White_Space starting at index start in s.
     * @return The smallest index at or after start with a non-white space character.
     */
    static int32_t skipWhiteSpace(const UnicodeString &s, int32_t start);

    /**
     * @return s except with leading and trailing Pattern_White_Space removed and length adjusted.
     */
    static const UChar *trimWhiteSpace(const UChar *s, int32_t &length);

    /**
     * Tests whether the string contains a "pattern identifier", that is,
     * whether it contains only non-Pattern_White_Space, non-Pattern_Syntax characters.
     * @return TRUE if there are no Pattern_White_Space or Pattern_Syntax characters in s.
     */
    static UBool isIdentifier(const UChar *s, int32_t length);

    /**
     * Skips over a "pattern identifier" starting at index s.
     * @return The smallest pointer at or after s with
     *         a Pattern_White_Space or Pattern_Syntax character.
     */
    static const UChar *skipIdentifier(const UChar *s, int32_t length);

private:
    PatternProps();  // no constructor: all static methods
};

U_NAMESPACE_END





U_NAMESPACE_BEGIN

// Unicode character/code point constants ---------------------------------- ***

static const UChar u_pound=0x23;
static const UChar u_apos=0x27;
static const UChar u_plus=0x2B;
static const UChar u_comma=0x2C;
static const UChar u_minus=0x2D;
static const UChar u_dot=0x2E;
static const UChar u_colon=0x3A;
static const UChar u_lessThan=0x3C;
static const UChar u_equal=0x3D;
static const UChar u_A=0x41;
static const UChar u_C=0x43;
static const UChar u_D=0x44;
static const UChar u_E=0x45;
static const UChar u_H=0x48;
static const UChar u_I=0x49;
static const UChar u_L=0x4C;
static const UChar u_N=0x4E;
static const UChar u_O=0x4F;
static const UChar u_P=0x50;
static const UChar u_R=0x52;
static const UChar u_S=0x53;
static const UChar u_T=0x54;
static const UChar u_U=0x55;
static const UChar u_Z=0x5A;
static const UChar u_a=0x61;
static const UChar u_c=0x63;
static const UChar u_d=0x64;
static const UChar u_e=0x65;
static const UChar u_f=0x66;
static const UChar u_h=0x68;
static const UChar u_i=0x69;
static const UChar u_l=0x6C;
static const UChar u_n=0x6E;
static const UChar u_o=0x6F;
static const UChar u_p=0x70;
static const UChar u_r=0x72;
static const UChar u_s=0x73;
static const UChar u_t=0x74;
static const UChar u_u=0x75;
static const UChar u_z=0x7A;
static const UChar u_leftCurlyBrace=0x7B;
static const UChar u_pipe=0x7C;
static const UChar u_rightCurlyBrace=0x7D;
static const UChar u_lessOrEqual=0x2264;  // U+2264 is <=

static const UChar kOffsetColon[]={  // "offset:"
    u_o, u_f, u_f, u_s, u_e, u_t, u_colon
};

static const UChar kOther[]={  // "other"
    u_o, u_t, u_h, u_e, u_r
};

// MessagePatternList ------------------------------------------------------ ***

template<typename T, int32_t stackCapacity>
class MessagePatternList : public UMemory {
public:
    MessagePatternList() {}
    void copyFrom(const MessagePatternList<T, stackCapacity> &other,
                  int32_t length,
                  UErrorCode &errorCode);
    UBool ensureCapacityForOneMore(int32_t oldLength, UErrorCode &errorCode);
    UBool equals(const MessagePatternList<T, stackCapacity> &other, int32_t length) const {
        for(int32_t i=0; i<length; ++i) {
            if(a[i]!=other.a[i]) { return FALSE; }
        }
        return TRUE;
    }

    MaybeStackArray<T, stackCapacity> a;
};

template<typename T, int32_t stackCapacity>
void
MessagePatternList<T, stackCapacity>::copyFrom(
        const MessagePatternList<T, stackCapacity> &other,
        int32_t length,
        UErrorCode &errorCode) {
    if(U_SUCCESS(errorCode) && length>0) {
        if(length>a.getCapacity() && NULL==a.resize(length)) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        uprv_memcpy(a.getAlias(), other.a.getAlias(), (size_t)length*sizeof(T));
    }
}

template<typename T, int32_t stackCapacity>
UBool
MessagePatternList<T, stackCapacity>::ensureCapacityForOneMore(int32_t oldLength, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return FALSE;
    }
    if(a.getCapacity()>oldLength || a.resize(2*oldLength, oldLength)!=NULL) {
        return TRUE;
    }
    errorCode=U_MEMORY_ALLOCATION_ERROR;
    return FALSE;
}

// MessagePatternList specializations -------------------------------------- ***

class MessagePatternDoubleList : public MessagePatternList<double, 8> {
};

class MessagePatternPartsList : public MessagePatternList<MessagePattern::Part, 32> {
};

// MessagePattern constructors etc. ---------------------------------------- ***

MessagePattern::MessagePattern(UErrorCode &errorCode)
        : aposMode(UCONFIG_MSGPAT_DEFAULT_APOSTROPHE_MODE),
          partsList(NULL), parts(NULL), partsLength(0),
          numericValuesList(NULL), numericValues(NULL), numericValuesLength(0),
          hasArgNames(FALSE), hasArgNumbers(FALSE), needsAutoQuoting(FALSE) {
    init(errorCode);
}

MessagePattern::MessagePattern(UMessagePatternApostropheMode mode, UErrorCode &errorCode)
        : aposMode(mode),
          partsList(NULL), parts(NULL), partsLength(0),
          numericValuesList(NULL), numericValues(NULL), numericValuesLength(0),
          hasArgNames(FALSE), hasArgNumbers(FALSE), needsAutoQuoting(FALSE) {
    init(errorCode);
}

MessagePattern::MessagePattern(const UnicodeString &pattern, UParseError *parseError, UErrorCode &errorCode)
        : aposMode(UCONFIG_MSGPAT_DEFAULT_APOSTROPHE_MODE),
          partsList(NULL), parts(NULL), partsLength(0),
          numericValuesList(NULL), numericValues(NULL), numericValuesLength(0),
          hasArgNames(FALSE), hasArgNumbers(FALSE), needsAutoQuoting(FALSE) {
    if(init(errorCode)) {
        parse(pattern, parseError, errorCode);
    }
}

UBool
MessagePattern::init(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return FALSE;
    }
    partsList=new MessagePatternPartsList();
    if(partsList==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    parts=partsList->a.getAlias();
    return TRUE;
}

MessagePattern::MessagePattern(const MessagePattern &other)
        : UObject(other), aposMode(other.aposMode), msg(other.msg),
          partsList(NULL), parts(NULL), partsLength(0),
          numericValuesList(NULL), numericValues(NULL), numericValuesLength(0),
          hasArgNames(other.hasArgNames), hasArgNumbers(other.hasArgNumbers),
          needsAutoQuoting(other.needsAutoQuoting) {
    UErrorCode errorCode=U_ZERO_ERROR;
    if(!copyStorage(other, errorCode)) {
        clear();
    }
}

MessagePattern &
MessagePattern::operator=(const MessagePattern &other) {
    if(this==&other) {
        return *this;
    }
    aposMode=other.aposMode;
    msg=other.msg;
    hasArgNames=other.hasArgNames;
    hasArgNumbers=other.hasArgNumbers;
    needsAutoQuoting=other.needsAutoQuoting;
    UErrorCode errorCode=U_ZERO_ERROR;
    if(!copyStorage(other, errorCode)) {
        clear();
    }
    return *this;
}

UBool
MessagePattern::copyStorage(const MessagePattern &other, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return FALSE;
    }
    parts=NULL;
    partsLength=0;
    numericValues=NULL;
    numericValuesLength=0;
    if(partsList==NULL) {
        partsList=new MessagePatternPartsList();
        if(partsList==NULL) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            return FALSE;
        }
        parts=partsList->a.getAlias();
    }
    if(other.partsLength>0) {
        partsList->copyFrom(*other.partsList, other.partsLength, errorCode);
        if(U_FAILURE(errorCode)) {
            return FALSE;
        }
        parts=partsList->a.getAlias();
        partsLength=other.partsLength;
    }
    if(other.numericValuesLength>0) {
        if(numericValuesList==NULL) {
            numericValuesList=new MessagePatternDoubleList();
            if(numericValuesList==NULL) {
                errorCode=U_MEMORY_ALLOCATION_ERROR;
                return FALSE;
            }
            numericValues=numericValuesList->a.getAlias();
        }
        numericValuesList->copyFrom(
            *other.numericValuesList, other.numericValuesLength, errorCode);
        if(U_FAILURE(errorCode)) {
            return FALSE;
        }
        numericValues=numericValuesList->a.getAlias();
        numericValuesLength=other.numericValuesLength;
    }
    return TRUE;
}

MessagePattern::~MessagePattern() {
    delete partsList;
    delete numericValuesList;
}

// MessagePattern API ------------------------------------------------------ ***

MessagePattern &
MessagePattern::parse(const UnicodeString &pattern, UParseError *parseError, UErrorCode &errorCode) {
    preParse(pattern, parseError, errorCode);
    parseMessage(0, 0, 0, UMSGPAT_ARG_TYPE_NONE, parseError, errorCode);
    postParse();
    return *this;
}

MessagePattern &
MessagePattern::parseChoiceStyle(const UnicodeString &pattern,
                                 UParseError *parseError, UErrorCode &errorCode) {
    preParse(pattern, parseError, errorCode);
    parseChoiceStyle(0, 0, parseError, errorCode);
    postParse();
    return *this;
}

MessagePattern &
MessagePattern::parsePluralStyle(const UnicodeString &pattern,
                                 UParseError *parseError, UErrorCode &errorCode) {
    preParse(pattern, parseError, errorCode);
    parsePluralOrSelectStyle(UMSGPAT_ARG_TYPE_PLURAL, 0, 0, parseError, errorCode);
    postParse();
    return *this;
}

MessagePattern &
MessagePattern::parseSelectStyle(const UnicodeString &pattern,
                                 UParseError *parseError, UErrorCode &errorCode) {
    preParse(pattern, parseError, errorCode);
    parsePluralOrSelectStyle(UMSGPAT_ARG_TYPE_SELECT, 0, 0, parseError, errorCode);
    postParse();
    return *this;
}

void
MessagePattern::clear() {
    // Mostly the same as preParse().
    msg.remove();
    hasArgNames=hasArgNumbers=FALSE;
    needsAutoQuoting=FALSE;
    partsLength=0;
    numericValuesLength=0;
}

UBool
MessagePattern::operator==(const MessagePattern &other) const {
    if(this==&other) {
        return TRUE;
    }
    return
        aposMode==other.aposMode &&
        msg==other.msg &&
        // parts.equals(o.parts)
        partsLength==other.partsLength &&
        (partsLength==0 || partsList->equals(*other.partsList, partsLength));
    // No need to compare numericValues if msg and parts are the same.
}

int32_t
MessagePattern::hashCode() const {
    int32_t hash=(aposMode*37+msg.hashCode())*37+partsLength;
    for(int32_t i=0; i<partsLength; ++i) {
        hash=hash*37+parts[i].hashCode();
    }
    return hash;
}

int32_t
MessagePattern::validateArgumentName(const UnicodeString &name) {
    if(!PatternProps::isIdentifier(name.getBuffer(), name.length())) {
        return UMSGPAT_ARG_NAME_NOT_VALID;
    }
    return parseArgNumber(name, 0, name.length());
}

UnicodeString
MessagePattern::autoQuoteApostropheDeep() const {
    if(!needsAutoQuoting) {
        return msg;
    }
    UnicodeString modified(msg);
    // Iterate backward so that the insertion indexes do not change.
    int32_t count=countParts();
    for(int32_t i=count; i>0;) {
        const Part &part=getPart(--i);
        if(part.getType()==UMSGPAT_PART_TYPE_INSERT_CHAR) {
           modified.insert(part.index, (UChar)part.value);
        }
    }
    return modified;
}

double
MessagePattern::getNumericValue(const Part &part) const {
    UMessagePatternPartType type=part.type;
    if(type==UMSGPAT_PART_TYPE_ARG_INT) {
        return part.value;
    } else if(type==UMSGPAT_PART_TYPE_ARG_DOUBLE) {
        return numericValues[part.value];
    } else {
        return UMSGPAT_NO_NUMERIC_VALUE;
    }
}

/**
  * Returns the "offset:" value of a PluralFormat argument, or 0 if none is specified.
  * @param pluralStart the index of the first PluralFormat argument style part. (0..countParts()-1)
  * @return the "offset:" value.
  * @draft ICU 4.8
  */
double
MessagePattern::getPluralOffset(int32_t pluralStart) const {
    const Part &part=getPart(pluralStart);
    if(Part::hasNumericValue(part.type)) {
        return getNumericValue(part);
    } else {
        return 0;
    }
}

// MessagePattern::Part ---------------------------------------------------- ***

UBool
MessagePattern::Part::operator==(const Part &other) const {
    if(this==&other) {
        return TRUE;
    }
    return
        type==other.type &&
        index==other.index &&
        length==other.length &&
        value==other.value &&
        limitPartIndex==other.limitPartIndex;
}

// MessagePattern parser --------------------------------------------------- ***

void
MessagePattern::preParse(const UnicodeString &pattern, UParseError *parseError, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    if(parseError!=NULL) {
        parseError->line=0;
        parseError->offset=0;
        parseError->preContext[0]=0;
        parseError->postContext[0]=0;
    }
    msg=pattern;
    hasArgNames=hasArgNumbers=FALSE;
    needsAutoQuoting=FALSE;
    partsLength=0;
    numericValuesLength=0;
}

void
MessagePattern::postParse() {
    if(partsList!=NULL) {
        parts=partsList->a.getAlias();
    }
    if(numericValuesList!=NULL) {
        numericValues=numericValuesList->a.getAlias();
    }
}

int32_t
MessagePattern::parseMessage(int32_t index, int32_t msgStartLength,
                             int32_t nestingLevel, UMessagePatternArgType parentType,
                             UParseError *parseError, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    if(nestingLevel>Part::MAX_VALUE) {
        errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }
    int32_t msgStart=partsLength;
    addPart(UMSGPAT_PART_TYPE_MSG_START, index, msgStartLength, nestingLevel, errorCode);
    index+=msgStartLength;
    for(;;) {  // while(index<msg.length()) with U_FAILURE(errorCode) check
        if(U_FAILURE(errorCode)) {
            return 0;
        }
        if(index>=msg.length()) {
            break;
        }
        UChar c=msg.charAt(index++);
        if(c==u_apos) {
            if(index==msg.length()) {
                // The apostrophe is the last character in the pattern. 
                // Add a Part for auto-quoting.
                addPart(UMSGPAT_PART_TYPE_INSERT_CHAR, index, 0,
                        u_apos, errorCode);  // value=char to be inserted
                needsAutoQuoting=TRUE;
            } else {
                c=msg.charAt(index);
                if(c==u_apos) {
                    // double apostrophe, skip the second one
                    addPart(UMSGPAT_PART_TYPE_SKIP_SYNTAX, index++, 1, 0, errorCode);
                } else if(
                    aposMode==UMSGPAT_APOS_DOUBLE_REQUIRED ||
                    c==u_leftCurlyBrace || c==u_rightCurlyBrace ||
                    (parentType==UMSGPAT_ARG_TYPE_CHOICE && c==u_pipe) ||
                    (UMSGPAT_ARG_TYPE_HAS_PLURAL_STYLE(parentType) && c==u_pound)
                ) {
                    // skip the quote-starting apostrophe
                    addPart(UMSGPAT_PART_TYPE_SKIP_SYNTAX, index-1, 1, 0, errorCode);
                    // find the end of the quoted literal text
                    for(;;) {
                        index=msg.indexOf(u_apos, index+1);
                        if(index>=0) {
                            if(/*(index+1)<msg.length() &&*/ msg.charAt(index+1)==u_apos) {
                                // double apostrophe inside quoted literal text
                                // still encodes a single apostrophe, skip the second one
                                addPart(UMSGPAT_PART_TYPE_SKIP_SYNTAX, ++index, 1, 0, errorCode);
                            } else {
                                // skip the quote-ending apostrophe
                                addPart(UMSGPAT_PART_TYPE_SKIP_SYNTAX, index++, 1, 0, errorCode);
                                break;
                            }
                        } else {
                            // The quoted text reaches to the end of the of the message.
                            index=msg.length();
                            // Add a Part for auto-quoting.
                            addPart(UMSGPAT_PART_TYPE_INSERT_CHAR, index, 0,
                                    u_apos, errorCode);  // value=char to be inserted
                            needsAutoQuoting=TRUE;
                            break;
                        }
                    }
                } else {
                    // Interpret the apostrophe as literal text.
                    // Add a Part for auto-quoting.
                    addPart(UMSGPAT_PART_TYPE_INSERT_CHAR, index, 0,
                            u_apos, errorCode);  // value=char to be inserted
                    needsAutoQuoting=TRUE;
                }
            }
        } else if(UMSGPAT_ARG_TYPE_HAS_PLURAL_STYLE(parentType) && c==u_pound) {
            // The unquoted # in a plural message fragment will be replaced
            // with the (number-offset).
            addPart(UMSGPAT_PART_TYPE_REPLACE_NUMBER, index-1, 1, 0, errorCode);
        } else if(c==u_leftCurlyBrace) {
            index=parseArg(index-1, 1, nestingLevel, parseError, errorCode);
        } else if((nestingLevel>0 && c==u_rightCurlyBrace) ||
                  (parentType==UMSGPAT_ARG_TYPE_CHOICE && c==u_pipe)) {
            // Finish the message before the terminator.
            // In a choice style, report the "}" substring only for the following ARG_LIMIT,
            // not for this MSG_LIMIT.
            int32_t limitLength=(parentType==UMSGPAT_ARG_TYPE_CHOICE && c==u_rightCurlyBrace) ? 0 : 1;
            addLimitPart(msgStart, UMSGPAT_PART_TYPE_MSG_LIMIT, index-1, limitLength,
                         nestingLevel, errorCode);
            if(parentType==UMSGPAT_ARG_TYPE_CHOICE) {
                // Let the choice style parser see the '}' or '|'.
                return index-1;
            } else {
                // continue parsing after the '}'
                return index;
            }
        }  // else: c is part of literal text
    }
    if(nestingLevel>0 && !inTopLevelChoiceMessage(nestingLevel, parentType)) {
        setParseError(parseError, 0);  // Unmatched '{' braces in message.
        errorCode=U_UNMATCHED_BRACES;
        return 0;
    }
    addLimitPart(msgStart, UMSGPAT_PART_TYPE_MSG_LIMIT, index, 0, nestingLevel, errorCode);
    return index;
}

int32_t
MessagePattern::parseArg(int32_t index, int32_t argStartLength, int32_t nestingLevel,
                         UParseError *parseError, UErrorCode &errorCode) {
    int32_t argStart=partsLength;
    UMessagePatternArgType argType=UMSGPAT_ARG_TYPE_NONE;
    addPart(UMSGPAT_PART_TYPE_ARG_START, index, argStartLength, argType, errorCode);
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    int32_t nameIndex=index=skipWhiteSpace(index+argStartLength);
    if(index==msg.length()) {
        setParseError(parseError, 0);  // Unmatched '{' braces in message.
        errorCode=U_UNMATCHED_BRACES;
        return 0;
    }
    // parse argument name or number
    index=skipIdentifier(index);
    int32_t number=parseArgNumber(nameIndex, index);
    if(number>=0) {
        int32_t length=index-nameIndex;
        if(length>Part::MAX_LENGTH || number>Part::MAX_VALUE) {
            setParseError(parseError, nameIndex);  // Argument number too large.
            errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }
        hasArgNumbers=TRUE;
        addPart(UMSGPAT_PART_TYPE_ARG_NUMBER, nameIndex, length, number, errorCode);
    } else if(number==UMSGPAT_ARG_NAME_NOT_NUMBER) {
        int32_t length=index-nameIndex;
        if(length>Part::MAX_LENGTH) {
            setParseError(parseError, nameIndex);  // Argument name too long.
            errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }
        hasArgNames=TRUE;
        addPart(UMSGPAT_PART_TYPE_ARG_NAME, nameIndex, length, 0, errorCode);
    } else {  // number<-1 (ARG_NAME_NOT_VALID)
        setParseError(parseError, nameIndex);  // Bad argument syntax.
        errorCode=U_PATTERN_SYNTAX_ERROR;
        return 0;
    }
    index=skipWhiteSpace(index);
    if(index==msg.length()) {
        setParseError(parseError, 0);  // Unmatched '{' braces in message.
        errorCode=U_UNMATCHED_BRACES;
        return 0;
    }
    UChar c=msg.charAt(index);
    if(c==u_rightCurlyBrace) {
        // all done
    } else if(c!=u_comma) {
        setParseError(parseError, nameIndex);  // Bad argument syntax.
        errorCode=U_PATTERN_SYNTAX_ERROR;
        return 0;
    } else /* ',' */ {
        // parse argument type: case-sensitive a-zA-Z
        int32_t typeIndex=index=skipWhiteSpace(index+1);
        while(index<msg.length() && isArgTypeChar(msg.charAt(index))) {
            ++index;
        }
        int32_t length=index-typeIndex;
        index=skipWhiteSpace(index);
        if(index==msg.length()) {
            setParseError(parseError, 0);  // Unmatched '{' braces in message.
            errorCode=U_UNMATCHED_BRACES;
            return 0;
        }
        if(length==0 || ((c=msg.charAt(index))!=u_comma && c!=u_rightCurlyBrace)) {
            setParseError(parseError, nameIndex);  // Bad argument syntax.
            errorCode=U_PATTERN_SYNTAX_ERROR;
            return 0;
        }
        if(length>Part::MAX_LENGTH) {
            setParseError(parseError, nameIndex);  // Argument type name too long.
            errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }
        argType=UMSGPAT_ARG_TYPE_SIMPLE;
        if(length==6) {
            // case-insensitive comparisons for complex-type names
            if(isChoice(typeIndex)) {
                argType=UMSGPAT_ARG_TYPE_CHOICE;
            } else if(isPlural(typeIndex)) {
                argType=UMSGPAT_ARG_TYPE_PLURAL;
            } else if(isSelect(typeIndex)) {
                argType=UMSGPAT_ARG_TYPE_SELECT;
            }
        } else if(length==13) {
            if(isSelect(typeIndex) && isOrdinal(typeIndex+6)) {
                argType=UMSGPAT_ARG_TYPE_SELECTORDINAL;
            }
        }
        // change the ARG_START type from NONE to argType
        partsList->a[argStart].value=(int16_t)argType;
        if(argType==UMSGPAT_ARG_TYPE_SIMPLE) {
            addPart(UMSGPAT_PART_TYPE_ARG_TYPE, typeIndex, length, 0, errorCode);
        }
        // look for an argument style (pattern)
        if(c==u_rightCurlyBrace) {
            if(argType!=UMSGPAT_ARG_TYPE_SIMPLE) {
                setParseError(parseError, nameIndex);  // No style field for complex argument.
                errorCode=U_PATTERN_SYNTAX_ERROR;
                return 0;
            }
        } else /* ',' */ {
            ++index;
            if(argType==UMSGPAT_ARG_TYPE_SIMPLE) {
                index=parseSimpleStyle(index, parseError, errorCode);
            } else if(argType==UMSGPAT_ARG_TYPE_CHOICE) {
                index=parseChoiceStyle(index, nestingLevel, parseError, errorCode);
            } else {
                index=parsePluralOrSelectStyle(argType, index, nestingLevel, parseError, errorCode);
            }
        }
    }
    // Argument parsing stopped on the '}'.
    addLimitPart(argStart, UMSGPAT_PART_TYPE_ARG_LIMIT, index, 1, argType, errorCode);
    return index+1;
}

int32_t
MessagePattern::parseSimpleStyle(int32_t index, UParseError *parseError, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    int32_t start=index;
    int32_t nestedBraces=0;
    while(index<msg.length()) {
        UChar c=msg.charAt(index++);
        if(c==u_apos) {
            // Treat apostrophe as quoting but include it in the style part.
            // Find the end of the quoted literal text.
            index=msg.indexOf(u_apos, index);
            if(index<0) {
                // Quoted literal argument style text reaches to the end of the message.
                setParseError(parseError, start);
                errorCode=U_PATTERN_SYNTAX_ERROR;
                return 0;
            }
            // skip the quote-ending apostrophe
            ++index;
        } else if(c==u_leftCurlyBrace) {
            ++nestedBraces;
        } else if(c==u_rightCurlyBrace) {
            if(nestedBraces>0) {
                --nestedBraces;
            } else {
                int32_t length=--index-start;
                if(length>Part::MAX_LENGTH) {
                    setParseError(parseError, start);  // Argument style text too long.
                    errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                    return 0;
                }
                addPart(UMSGPAT_PART_TYPE_ARG_STYLE, start, length, 0, errorCode);
                return index;
            }
        }  // c is part of literal text
    }
    setParseError(parseError, 0);  // Unmatched '{' braces in message.
    errorCode=U_UNMATCHED_BRACES;
    return 0;
}

int32_t
MessagePattern::parseChoiceStyle(int32_t index, int32_t nestingLevel,
                                 UParseError *parseError, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    int32_t start=index;
    index=skipWhiteSpace(index);
    if(index==msg.length() || msg.charAt(index)==u_rightCurlyBrace) {
        setParseError(parseError, 0);  // Missing choice argument pattern.
        errorCode=U_PATTERN_SYNTAX_ERROR;
        return 0;
    }
    for(;;) {
        // The choice argument style contains |-separated (number, separator, message) triples.
        // Parse the number.
        int32_t numberIndex=index;
        index=skipDouble(index);
        int32_t length=index-numberIndex;
        if(length==0) {
            setParseError(parseError, start);  // Bad choice pattern syntax.
            errorCode=U_PATTERN_SYNTAX_ERROR;
            return 0;
        }
        if(length>Part::MAX_LENGTH) {
            setParseError(parseError, numberIndex);  // Choice number too long.
            errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }
        parseDouble(numberIndex, index, TRUE, parseError, errorCode);  // adds ARG_INT or ARG_DOUBLE
        if(U_FAILURE(errorCode)) {
            return 0;
        }
        // Parse the separator.
        index=skipWhiteSpace(index);
        if(index==msg.length()) {
            setParseError(parseError, start);  // Bad choice pattern syntax.
            errorCode=U_PATTERN_SYNTAX_ERROR;
            return 0;
        }
        UChar c=msg.charAt(index);
        if(!(c==u_pound || c==u_lessThan || c==u_lessOrEqual)) {  // U+2264 is <=
            setParseError(parseError, start);  // Expected choice separator (#<\u2264) instead of c.
            errorCode=U_PATTERN_SYNTAX_ERROR;
            return 0;
        }
        addPart(UMSGPAT_PART_TYPE_ARG_SELECTOR, index, 1, 0, errorCode);
        // Parse the message fragment.
        index=parseMessage(++index, 0, nestingLevel+1, UMSGPAT_ARG_TYPE_CHOICE, parseError, errorCode);
        if(U_FAILURE(errorCode)) {
            return 0;
        }
        // parseMessage(..., CHOICE) returns the index of the terminator, or msg.length().
        if(index==msg.length()) {
            return index;
        }
        if(msg.charAt(index)==u_rightCurlyBrace) {
            if(!inMessageFormatPattern(nestingLevel)) {
                setParseError(parseError, start);  // Bad choice pattern syntax.
                errorCode=U_PATTERN_SYNTAX_ERROR;
                return 0;
            }
            return index;
        }  // else the terminator is '|'
        index=skipWhiteSpace(index+1);
    }
}

int32_t
MessagePattern::parsePluralOrSelectStyle(UMessagePatternArgType argType,
                                         int32_t index, int32_t nestingLevel,
                                         UParseError *parseError, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    int32_t start=index;
    UBool isEmpty=TRUE;
    UBool hasOther=FALSE;
    for(;;) {
        // First, collect the selector looking for a small set of terminators.
        // It would be a little faster to consider the syntax of each possible
        // token right here, but that makes the code too complicated.
        index=skipWhiteSpace(index);
        UBool eos=index==msg.length();
        if(eos || msg.charAt(index)==u_rightCurlyBrace) {
            if(eos==inMessageFormatPattern(nestingLevel)) {
                setParseError(parseError, start);  // Bad plural/select pattern syntax.
                errorCode=U_PATTERN_SYNTAX_ERROR;
                return 0;
            }
            if(!hasOther) {
                setParseError(parseError, 0);  // Missing 'other' keyword in plural/select pattern.
                errorCode=U_DEFAULT_KEYWORD_MISSING;
                return 0;
            }
            return index;
        }
        int32_t selectorIndex=index;
        if(UMSGPAT_ARG_TYPE_HAS_PLURAL_STYLE(argType) && msg.charAt(selectorIndex)==u_equal) {
            // explicit-value plural selector: =double
            index=skipDouble(index+1);
            int32_t length=index-selectorIndex;
            if(length==1) {
                setParseError(parseError, start);  // Bad plural/select pattern syntax.
                errorCode=U_PATTERN_SYNTAX_ERROR;
                return 0;
            }
            if(length>Part::MAX_LENGTH) {
                setParseError(parseError, selectorIndex);  // Argument selector too long.
                errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                return 0;
            }
            addPart(UMSGPAT_PART_TYPE_ARG_SELECTOR, selectorIndex, length, 0, errorCode);
            parseDouble(selectorIndex+1, index, FALSE,
                        parseError, errorCode);  // adds ARG_INT or ARG_DOUBLE
        } else {
            index=skipIdentifier(index);
            int32_t length=index-selectorIndex;
            if(length==0) {
                setParseError(parseError, start);  // Bad plural/select pattern syntax.
                errorCode=U_PATTERN_SYNTAX_ERROR;
                return 0;
            }
            // Note: The ':' in "offset:" is just beyond the skipIdentifier() range.
            if( UMSGPAT_ARG_TYPE_HAS_PLURAL_STYLE(argType) && length==6 && index<msg.length() &&
                0==msg.compare(selectorIndex, 7, kOffsetColon, 0, 7)
            ) {
                // plural offset, not a selector
                if(!isEmpty) {
                    // Plural argument 'offset:' (if present) must precede key-message pairs.
                    setParseError(parseError, start);
                    errorCode=U_PATTERN_SYNTAX_ERROR;
                    return 0;
                }
                // allow whitespace between offset: and its value
                int32_t valueIndex=skipWhiteSpace(index+1);  // The ':' is at index.
                index=skipDouble(valueIndex);
                if(index==valueIndex) {
                    setParseError(parseError, start);  // Missing value for plural 'offset:'.
                    errorCode=U_PATTERN_SYNTAX_ERROR;
                    return 0;
                }
                if((index-valueIndex)>Part::MAX_LENGTH) {
                    setParseError(parseError, valueIndex);  // Plural offset value too long.
                    errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                    return 0;
                }
                parseDouble(valueIndex, index, FALSE,
                            parseError, errorCode);  // adds ARG_INT or ARG_DOUBLE
                if(U_FAILURE(errorCode)) {
                    return 0;
                }
                isEmpty=FALSE;
                continue;  // no message fragment after the offset
            } else {
                // normal selector word
                if(length>Part::MAX_LENGTH) {
                    setParseError(parseError, selectorIndex);  // Argument selector too long.
                    errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                    return 0;
                }
                addPart(UMSGPAT_PART_TYPE_ARG_SELECTOR, selectorIndex, length, 0, errorCode);
                if(0==msg.compare(selectorIndex, length, kOther, 0, 5)) {
                    hasOther=TRUE;
                }
            }
        }
        if(U_FAILURE(errorCode)) {
            return 0;
        }

        // parse the message fragment following the selector
        index=skipWhiteSpace(index);
        if(index==msg.length() || msg.charAt(index)!=u_leftCurlyBrace) {
            setParseError(parseError, selectorIndex);  // No message fragment after plural/select selector.
            errorCode=U_PATTERN_SYNTAX_ERROR;
            return 0;
        }
        index=parseMessage(index, 1, nestingLevel+1, argType, parseError, errorCode);
        if(U_FAILURE(errorCode)) {
            return 0;
        }
        isEmpty=FALSE;
    }
}

int32_t
MessagePattern::parseArgNumber(const UnicodeString &s, int32_t start, int32_t limit) {
    // If the identifier contains only ASCII digits, then it is an argument _number_
    // and must not have leading zeros (except "0" itself).
    // Otherwise it is an argument _name_.
    if(start>=limit) {
        return UMSGPAT_ARG_NAME_NOT_VALID;
    }
    int32_t number;
    // Defer numeric errors until we know there are only digits.
    UBool badNumber;
    UChar c=s.charAt(start++);
    if(c==0x30) {
        if(start==limit) {
            return 0;
        } else {
            number=0;
            badNumber=TRUE;  // leading zero
        }
    } else if(0x31<=c && c<=0x39) {
        number=c-0x30;
        badNumber=FALSE;
    } else {
        return UMSGPAT_ARG_NAME_NOT_NUMBER;
    }
    while(start<limit) {
        c=s.charAt(start++);
        if(0x30<=c && c<=0x39) {
            if(number>=INT32_MAX/10) {
                badNumber=TRUE;  // overflow
            }
            number=number*10+(c-0x30);
        } else {
            return UMSGPAT_ARG_NAME_NOT_NUMBER;
        }
    }
    // There are only ASCII digits.
    if(badNumber) {
        return UMSGPAT_ARG_NAME_NOT_VALID;
    } else {
        return number;
    }
}

void
MessagePattern::parseDouble(int32_t start, int32_t limit, UBool allowInfinity,
                            UParseError *parseError, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    U_ASSERT(start<limit);
    // fake loop for easy exit and single throw statement
    for(;;) { /*loop doesn't iterate*/
        // fast path for small integers and infinity
        int32_t value=0;
        int32_t isNegative=0;  // not boolean so that we can easily add it to value
        int32_t index=start;
        UChar c=msg.charAt(index++);
        if(c==u_minus) {
            isNegative=1;
            if(index==limit) {
                break;  // no number
            }
            c=msg.charAt(index++);
        } else if(c==u_plus) {
            if(index==limit) {
                break;  // no number
            }
            c=msg.charAt(index++);
        }
        if(c==0x221e) {  // infinity
            if(allowInfinity && index==limit) {
                double infinity=uprv_getInfinity();
                addArgDoublePart(
                    isNegative!=0 ? -infinity : infinity,
                    start, limit-start, errorCode);
                return;
            } else {
                break;
            }
        }
        // try to parse the number as a small integer but fall back to a double
        while('0'<=c && c<='9') {
            value=value*10+(c-'0');
            if(value>(Part::MAX_VALUE+isNegative)) {
                break;  // not a small-enough integer
            }
            if(index==limit) {
                addPart(UMSGPAT_PART_TYPE_ARG_INT, start, limit-start,
                        isNegative!=0 ? -value : value, errorCode);
                return;
            }
            c=msg.charAt(index++);
        }
        // Let Double.parseDouble() throw a NumberFormatException.
        char numberChars[128];
        int32_t capacity=(int32_t)sizeof(numberChars);
        int32_t length=limit-start;
        if(length>=capacity) {
            break;  // number too long
        }
        msg.extract(start, length, numberChars, capacity, US_INV);
        if((int32_t)uprv_strlen(numberChars)<length) {
            break;  // contains non-invariant character that was turned into NUL
        }
        char *end;
        double numericValue=uprv_strtod(numberChars, &end);
        if(end!=(numberChars+length)) {
            break;  // parsing error
        }
        addArgDoublePart(numericValue, start, length, errorCode);
        return;
    }
    setParseError(parseError, start /*, limit*/);  // Bad syntax for numeric value.
    errorCode=U_PATTERN_SYNTAX_ERROR;
    return;
}

int32_t
MessagePattern::skipWhiteSpace(int32_t index) {
    const UChar *s=msg.getBuffer();
    int32_t msgLength=msg.length();
    const UChar *t=PatternProps::skipWhiteSpace(s+index, msgLength-index);
    return (int32_t)(t-s);
}

int32_t
MessagePattern::skipIdentifier(int32_t index) {
    const UChar *s=msg.getBuffer();
    int32_t msgLength=msg.length();
    const UChar *t=PatternProps::skipIdentifier(s+index, msgLength-index);
    return (int32_t)(t-s);
}

int32_t
MessagePattern::skipDouble(int32_t index) {
    int32_t msgLength=msg.length();
    while(index<msgLength) {
        UChar c=msg.charAt(index);
        // U+221E: Allow the infinity symbol, for ChoiceFormat patterns.
        if((c<0x30 && c!=u_plus && c!=u_minus && c!=u_dot) || (c>0x39 && c!=u_e && c!=u_E && c!=0x221e)) {
            break;
        }
        ++index;
    }
    return index;
}

UBool
MessagePattern::isArgTypeChar(UChar32 c) {
    return (u_a<=c && c<=u_z) || (u_A<=c && c<=u_Z);
}

UBool
MessagePattern::isChoice(int32_t index) {
    UChar c;
    return
        ((c=msg.charAt(index++))==u_c || c==u_C) &&
        ((c=msg.charAt(index++))==u_h || c==u_H) &&
        ((c=msg.charAt(index++))==u_o || c==u_O) &&
        ((c=msg.charAt(index++))==u_i || c==u_I) &&
        ((c=msg.charAt(index++))==u_c || c==u_C) &&
        ((c=msg.charAt(index))==u_e || c==u_E);
}

UBool
MessagePattern::isPlural(int32_t index) {
    UChar c;
    return
        ((c=msg.charAt(index++))==u_p || c==u_P) &&
        ((c=msg.charAt(index++))==u_l || c==u_L) &&
        ((c=msg.charAt(index++))==u_u || c==u_U) &&
        ((c=msg.charAt(index++))==u_r || c==u_R) &&
        ((c=msg.charAt(index++))==u_a || c==u_A) &&
        ((c=msg.charAt(index))==u_l || c==u_L);
}

UBool
MessagePattern::isSelect(int32_t index) {
    UChar c;
    return
        ((c=msg.charAt(index++))==u_s || c==u_S) &&
        ((c=msg.charAt(index++))==u_e || c==u_E) &&
        ((c=msg.charAt(index++))==u_l || c==u_L) &&
        ((c=msg.charAt(index++))==u_e || c==u_E) &&
        ((c=msg.charAt(index++))==u_c || c==u_C) &&
        ((c=msg.charAt(index))==u_t || c==u_T);
}

UBool
MessagePattern::isOrdinal(int32_t index) {
    UChar c;
    return
        ((c=msg.charAt(index++))==u_o || c==u_O) &&
        ((c=msg.charAt(index++))==u_r || c==u_R) &&
        ((c=msg.charAt(index++))==u_d || c==u_D) &&
        ((c=msg.charAt(index++))==u_i || c==u_I) &&
        ((c=msg.charAt(index++))==u_n || c==u_N) &&
        ((c=msg.charAt(index++))==u_a || c==u_A) &&
        ((c=msg.charAt(index))==u_l || c==u_L);
}

UBool
MessagePattern::inMessageFormatPattern(int32_t nestingLevel) {
    return nestingLevel>0 || partsList->a[0].type==UMSGPAT_PART_TYPE_MSG_START;
}

UBool
MessagePattern::inTopLevelChoiceMessage(int32_t nestingLevel, UMessagePatternArgType parentType) {
    return
        nestingLevel==1 &&
        parentType==UMSGPAT_ARG_TYPE_CHOICE &&
        partsList->a[0].type!=UMSGPAT_PART_TYPE_MSG_START;
}

void
MessagePattern::addPart(UMessagePatternPartType type, int32_t index, int32_t length,
                        int32_t value, UErrorCode &errorCode) {
    if(partsList->ensureCapacityForOneMore(partsLength, errorCode)) {
        Part &part=partsList->a[partsLength++];
        part.type=type;
        part.index=index;
        part.length=(uint16_t)length;
        part.value=(int16_t)value;
        part.limitPartIndex=0;
    }
}

void
MessagePattern::addLimitPart(int32_t start,
                             UMessagePatternPartType type, int32_t index, int32_t length,
                             int32_t value, UErrorCode &errorCode) {
    partsList->a[start].limitPartIndex=partsLength;
    addPart(type, index, length, value, errorCode);
}

void
MessagePattern::addArgDoublePart(double numericValue, int32_t start, int32_t length,
                                 UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    int32_t numericIndex=numericValuesLength;
    if(numericValuesList==NULL) {
        numericValuesList=new MessagePatternDoubleList();
        if(numericValuesList==NULL) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            return;
        }
    } else if(!numericValuesList->ensureCapacityForOneMore(numericValuesLength, errorCode)) {
        return;
    } else {
        if(numericIndex>Part::MAX_VALUE) {
            errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return;
        }
    }
    numericValuesList->a[numericValuesLength++]=numericValue;
    addPart(UMSGPAT_PART_TYPE_ARG_DOUBLE, start, length, numericIndex, errorCode);
}

void
MessagePattern::setParseError(UParseError *parseError, int32_t index) {
    if(parseError==NULL) {
        return;
    }
    parseError->offset=index;

    // Set preContext to some of msg before index.
    // Avoid splitting a surrogate pair.
    int32_t length=index;
    if(length>=U_PARSE_CONTEXT_LEN) {
        length=U_PARSE_CONTEXT_LEN-1;
        if(length>0 && U16_IS_TRAIL(msg[index-length])) {
            --length;
        }
    }
    msg.extract(index-length, length, parseError->preContext);
    parseError->preContext[length]=0;

    // Set postContext to some of msg starting at index.
    length=msg.length()-index;
    if(length>=U_PARSE_CONTEXT_LEN) {
        length=U_PARSE_CONTEXT_LEN-1;
        if(length>0 && U16_IS_LEAD(msg[index+length-1])) {
            --length;
        }
    }
    msg.extract(index, length, parseError->postContext);
    parseError->postContext[length]=0;
}

// MessageImpl ------------------------------------------------------------- ***

void
MessageImpl::appendReducedApostrophes(const UnicodeString &s, int32_t start, int32_t limit,
                                      UnicodeString &sb) {
    int32_t doubleApos=-1;
    for(;;) {
        int32_t i=s.indexOf(u_apos, start);
        if(i<0 || i>=limit) {
            sb.append(s, start, limit-start);
            break;
        }
        if(i==doubleApos) {
            // Double apostrophe at start-1 and start==i, append one.
            sb.append(u_apos);
            ++start;
            doubleApos=-1;
        } else {
            // Append text between apostrophes and skip this one.
            sb.append(s, start, i-start);
            doubleApos=start=i+1;
        }
    }
}

// Ported from second half of ICU4J SelectFormat.format(String).
UnicodeString &
MessageImpl::appendSubMessageWithoutSkipSyntax(const MessagePattern &msgPattern,
                                               int32_t msgStart,
                                               UnicodeString &result) {
    const UnicodeString &msgString=msgPattern.getPatternString();
    int32_t prevIndex=msgPattern.getPart(msgStart).getLimit();
    for(int32_t i=msgStart;;) {
        const MessagePattern::Part &part=msgPattern.getPart(++i);
        UMessagePatternPartType type=part.getType();
        int32_t index=part.getIndex();
        if(type==UMSGPAT_PART_TYPE_MSG_LIMIT) {
            return result.append(msgString, prevIndex, index-prevIndex);
        } else if(type==UMSGPAT_PART_TYPE_SKIP_SYNTAX) {
            result.append(msgString, prevIndex, index-prevIndex);
            prevIndex=part.getLimit();
        } else if(type==UMSGPAT_PART_TYPE_ARG_START) {
            result.append(msgString, prevIndex, index-prevIndex);
            prevIndex=index;
            i=msgPattern.getLimitPartIndex(i);
            index=msgPattern.getPart(i).getLimit();
            appendReducedApostrophes(msgString, prevIndex, index, result);
            prevIndex=index;
        }
    }
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_FORMATTING
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2009-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  normalizer2.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2009nov22
*   created by: Markus W. Scherer
*/



#if !UCONFIG_NO_NORMALIZATION













using icu::Normalizer2Impl;

#if NORM2_HARDCODE_NFC_DATA
// NFC/NFD data machine-generated by gennorm2 --csource
#define INCLUDED_FROM_NORMALIZER2_CPP
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// Copyright (C) 1999-2016, International Business Machines
// Corporation and others.  All Rights Reserved.
//
// file name: norm2_nfc_data.h
//
// machine-generated by: icu/source/tools/gennorm2/n2builder.cpp


#ifdef INCLUDED_FROM_NORMALIZER2_CPP

// static const UVersionInfo norm2_nfc_data_formatVersion={4,0,0,0};
// static const UVersionInfo norm2_nfc_data_dataVersion={0xd,0,0,0};

static const int32_t norm2_nfc_data_indexes[Normalizer2Impl::IX_COUNT]={
0x50,0x4bac,0x8814,0x8914,0x8914,0x8914,0x8914,0x8914,0xc0,0x300,0xae2,0x29e0,0x3c66,0xfc00,0x1288,0x3b9c,
0x3c34,0x3c66,0x300,0
};

static const uint16_t norm2_nfc_data_trieIndex[1746]={
0,0x40,0x7b,0xbb,0xfb,0x13a,0x17a,0x1b2,0x1f2,0x226,0x254,0x226,0x294,0x2d4,0x313,0x353,
0x393,0x3d2,0x40f,0x44e,0x226,0x226,0x488,0x4c8,0x4f8,0x530,0x226,0x570,0x59f,0x5de,0x226,0x5f3,
0x631,0x65f,0x226,0x68c,0x6cc,0x709,0x729,0x768,0x7a7,0x7e4,0x803,0x840,0x729,0x879,0x8a7,0x8e6,
0x226,0x920,0x937,0x977,0x98e,0x9cd,0x226,0xa03,0xa23,0xa5e,0xa6a,0xaa5,0xacd,0xb0a,0xb4a,0xb84,
0xb9f,0x226,0xbda,0x226,0xc1a,0xc39,0xc6f,0xcac,0x226,0x226,0x226,0x226,0x226,0xccf,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0xcfb,0x226,0x226,0xd30,
0x226,0x226,0xd4e,0x226,0xd78,0x226,0x226,0x226,0xdb4,0xdd4,0xe14,0xe53,0xe8e,0xece,0xf02,0xf2e,
0x808,0x226,0x226,0xf62,0x226,0x226,0x226,0xfa2,0xfe2,0x1022,0x1062,0x10a2,0x10e2,0x1122,0x1162,0x11a2,
0x11e2,0x226,0x226,0x1212,0x1243,0x226,0x1273,0x12a6,0x12e3,0x1322,0x1362,0x1398,0x13c6,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x13f1,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0xcbd,0x226,0x140e,0x226,0x144e,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x148e,0x14c8,0x1506,0x1546,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x1585,0x15c3,0x15e3,0x226,0x226,0x226,0x226,
0x161d,0x226,0x226,0x1645,0x1677,0x16a5,0x80c,0x16b8,0x226,0x226,0x16c8,0x1708,0x226,0x226,0x226,0x1420,
0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,
0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,
0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,
0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,
0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,
0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,
0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,
0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,
0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,
0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,
0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x175c,0x1748,0x1750,0x1758,0x1760,0x174c,0x1754,0x1794,0x226,
0x17d4,0x180f,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x184f,0x188f,0x18cf,0x190f,0x194f,0x198f,0x19cf,0x1a0f,0x1a32,0x1a72,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x1a92,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x655,0x664,0x67c,0x69b,0x6b0,0x6b0,0x6b0,0x6b4,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0xbda,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x54f,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x40c,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x1ac5,0x226,0x226,0x1ad5,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0xdc6,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x1ae5,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x15d6,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x1aef,0x54f,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x7eb,0x226,0x226,
0x9ba,0x226,0x1aff,0x1b0c,0x1b18,0x226,0x226,0x226,0x226,0x414,0x226,0x1b23,0x1b33,0x226,0x226,0x226,
0x7e0,0x226,0x226,0x226,0x226,0x1b43,0x226,0x226,0x226,0x1b4e,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x1b55,0x226,0x226,0x226,0x226,0x1b60,0x1b6f,0x8f6,0x1b7d,0x412,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x1b8b,0x798,0x226,0x226,0x226,0x226,0x226,0x1b9b,0x1baa,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x8d6,0x1bb2,0x1bc2,0x226,
0x226,0x226,0x9ba,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x1bcc,0x226,0x226,0x226,0x226,0x226,
0x226,0x7e6,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x1bc9,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x1bdc,
0x7e0,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x84d,0x226,0x226,0x226,0x7ed,0x7ea,
0x226,0x226,0x226,0x226,0x7e8,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x9ba,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0xbd4,0x226,0x226,0x226,
0x226,0x7ea,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x1bec,0x226,0x226,0x226,0xefb,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x1bfc,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x1bfe,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x1c0d,0x1c1d,0x1c2b,0x1c38,0x226,0x1c44,0x1c52,0x1c62,0x226,0x226,
0x226,0x226,0xcea,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x1c72,0x1c7a,
0x1c88,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0xefb,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x4fc,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x1c98,0x226,0x226,0x226,0x226,0x226,0x226,0x1ca4,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x1cb4,0x1cc4,0x1cd4,0x1ce4,0x1cf4,0x1d04,0x1d14,0x1d24,0x1d34,0x1d44,0x1d54,
0x1d64,0x1d74,0x1d84,0x1d94,0x1da4,0x1db4,0x1dc4,0x1dd4,0x1de4,0x1df4,0x1e04,0x1e14,0x1e24,0x1e34,0x1e44,0x1e54,
0x1e64,0x1e74,0x1e84,0x1e94,0x1ea4,0x1eb4,0x1ec4,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,0x226,
0x226,0x226,0x226,0x226,0x226,0x408,0x428,0xc4,0xc4,0xc4,0x448,0x457,0x46d,0x489,0x4a6,0x4c2,
0x4df,0x4fc,0x51b,0x538,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,
0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0x552,0xc4,0x566,0xc4,0xc4,0xc4,0xc4,
0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,
0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0x586,0xc4,0xc4,0xc4,0xc4,0xc4,
0xc4,0xc4,0xc4,0x591,0x5ae,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0x5ce,0x5e2,0xc4,0xc4,0x5f5,
0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,
0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,0xc4,
0x615,0x635
};

static const uint16_t norm2_nfc_data_trieData[7892]={
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,4,8,0xc,1,
1,0x10,0x50,0x5c,0x70,0x88,0xcc,0xd0,0xec,0x108,0x144,0x148,0x15c,0x174,0x180,0x1a4,
0x1e4,1,0x1ec,0x20c,0x228,0x244,0x290,0x298,0x2b0,0x2b8,0x2dc,1,1,1,1,1,
1,0x2f4,0x334,0x340,0x354,0x36c,0x3b0,0x3b4,0x3d0,0x3f0,0x428,0x430,0x444,0x45c,0x468,0x48c,
0x4cc,1,0x4d4,0x4f4,0x510,0x530,0x57c,0x584,0x5a0,0x5a8,0x5d0,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,0x5e8,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,0x128a,0x1290,0xae4,0x1296,0xafa,
0xb04,0x5f4,0xb0e,0x129c,0x12a2,0xb18,0x12a8,0x12ae,0x12b4,0x12ba,0xb2e,1,0x12c0,0x12c6,0x12cc,0xb38,
0xb4e,0xb60,1,0x5fc,0x12d2,0x12d8,0x12de,0xb6a,0x12e4,1,1,0x12ea,0x12f0,0xb80,0x12f6,0xb96,
0xba0,0x600,0xbaa,0x12fc,0x1302,0xbb4,0x1308,0x130e,0x1314,0x131a,0xbca,1,0x1320,0x1326,0x132c,0xbd4,
0xbea,0xbfc,1,0x608,0x1332,0x1338,0x133e,0xc06,0x1344,1,0x134a,0x1350,0x1356,0xc1c,0xc32,0x135d,
0x1363,0x1368,0x136e,0x1374,0x137a,0x1380,0x1386,0x138c,0x1392,0x1398,0x139e,1,1,0xc48,0xc56,0x13a4,
0x13aa,0x13b0,0x13b6,0x13bd,0x13c3,0x13c8,0x13ce,0x13d4,0x13da,0x13e0,0x13e6,0x13ec,0x13f2,0x13f9,0x13ff,0x1404,
0x140a,1,1,0x1410,0x1416,0x141c,0x1422,0x1428,0x142e,0x1435,0x143b,0x1440,1,1,1,0x1447,
0x144d,0x1453,0x1459,1,0x145e,0x1464,0x146b,0x1471,0x1476,0x147c,1,1,1,0x1482,0x1488,0x148f,
0x1495,0x149a,0x14a0,1,1,1,0xc64,0xc72,0x14a6,0x14ac,0x14b2,0x14b8,1,1,0x14be,0x14c4,
0x14cb,0x14d1,0x14d6,0x14dc,0xc80,0xc8a,0x14e2,0x14e8,0x14ef,0x14f5,0xc94,0xc9e,0x14fb,0x1501,0x1506,0x150c,
1,1,0xca8,0xcb2,0xcbc,0xcc6,0x1512,0x1518,0x151e,0x1524,0x152a,0x1530,0x1537,0x153d,0x1542,0x1548,
0x154e,0x1554,0x155a,0x1560,0x1566,0x156c,0x1572,0x1578,0x157e,0x60c,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,0xcd0,0xcea,1,1,1,1,
1,1,1,1,1,1,1,1,1,0xd04,0xd1e,1,1,1,1,1,
1,0x610,1,1,1,1,1,1,1,1,1,1,1,1,1,0x1584,
0x158a,0x1590,0x1596,0x159c,0x15a2,0x15a8,0x15ae,0x15b6,0x15c0,0x15ca,0x15d4,0x15de,0x15e8,0x15f2,0x15fc,1,
0x1606,0x1610,0x161a,0x1624,0x162d,0x1633,1,1,0x1638,0x163e,0x1644,0x164a,0xd38,0xd42,0x1653,0x165d,
0x1665,0x166b,0x1671,1,1,1,0x1676,0x167c,1,1,0x1682,0x1688,0x1690,0x169a,0x16a3,0x16a9,
0x16af,0x16b5,0x16ba,0x16c0,0x16c6,0x16cc,0x16d2,0x16d8,0x16de,0x16e4,0x16ea,0x16f0,0x16f6,0x16fc,0x1702,0x1708,
0x170e,0x1714,0x171a,0x1720,0x1726,0x172c,0x1732,0x1738,0x173e,0x1744,0x174a,0x1750,0x1756,0x175c,1,1,
0x1762,0x1768,1,1,1,1,1,1,0xd4c,0xd56,0xd60,0xd6a,0x1770,0x177a,0x1784,0x178e,
0xd74,0xd7e,0x1798,0x17a2,0x17aa,0x17b0,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,0x614,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,0xfdcc,0xfdcc,0xfdcc,0xfdcc,0xfdcc,0xffcc,0xfdcc,0xfdcc,0xfdcc,0xfdcc,0xfdcc,0xfdcc,
0xfdcc,0xffcc,0xffcc,0xfdcc,0xffcc,0xfdcc,0xffcc,0xfdcc,0xfdcc,0xffd0,0xffb8,0xffb8,0xffb8,0xffb8,0xffd0,0xfdb0,
0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,0xff94,0xff94,0xfdb8,0xfdb8,0xfdb8,0xfdb8,0xfd94,0xfd94,0xffb8,0xffb8,0xffb8,
0xffb8,0xfdb8,0xfdb8,0xffb8,0xfdb8,0xfdb8,0xffb8,0xffb8,0xfe02,0xfe02,0xfe02,0xfe02,0xfc02,0xffb8,0xffb8,0xffb8,
0xffb8,0xffcc,0xffcc,0xffcc,0x3c36,0x3c3c,0xfdcc,0x3c42,0x3c48,0xfde0,0xffcc,0xffb8,0xffb8,0xffb8,0xffcc,0xffcc,
0xffcc,0xffb8,0xffb8,1,0xffcc,0xffcc,0xffcc,0xffb8,0xffb8,0xffb8,0xffb8,0xffcc,0xffd0,0xffb8,0xffb8,0xffcc,
0xffd2,0xffd4,0xffd4,0xffd2,0xffd4,0xffd4,0xffd2,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,0xffcc,1,1,1,1,0x29e1,1,1,1,1,1,1,1,
1,1,0x29e5,1,1,1,1,1,0x17b7,0x17bd,0x29e9,0x17c3,0x17c9,0x17cf,1,0x17d5,
1,0x17db,0x17e1,0x17e9,0x618,1,1,1,0x634,1,0x644,1,0x658,1,1,1,
1,1,0x674,1,0x684,1,1,1,0x688,1,1,1,0x6a0,0x17f1,0x17f7,0xd88,
0x17fd,0xd92,0x1803,0x180b,0x6b4,1,1,1,0x6d4,1,0x6e4,1,0x6fc,1,1,1,
1,1,0x71c,1,0x72c,1,1,1,0x734,1,1,1,0x754,0xd9c,0xdae,0x1813,
0x1819,0xdc0,1,1,1,0x76c,0x181f,0x1825,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,0x182b,0x1831,1,0x1837,1,1,0x774,0x183d,1,1,1,1,0x1843,
0x1849,0x184f,1,0x778,1,1,0x780,1,0x784,0x790,0x798,0x79c,0x1855,0x7ac,1,1,
1,0x7b0,1,1,1,1,0x7b4,1,1,1,0x7c4,1,1,1,0x7c8,1,
0x7cc,1,1,0x7d0,1,1,0x7d8,1,0x7dc,0x7e8,0x7f0,0x7f4,0x185b,0x804,1,1,
1,0x808,1,1,1,0x80c,1,1,1,0x81c,1,1,1,0x820,1,0x824,
1,1,0x1861,0x1867,1,0x186d,1,1,0x828,0x1873,1,1,1,1,0x1879,0x187f,
0x1885,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,0x82c,0x830,0x188b,0x1891,1,1,1,1,1,1,
1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0x1897,
0x189d,1,1,1,1,1,1,1,1,1,1,1,1,1,0x18a3,0x18a9,
0x18af,0x18b5,1,1,0x18bb,0x18c1,0x834,0x838,0x18c7,0x18cd,0x18d3,0x18d9,0x18df,0x18e5,1,1,
0x18eb,0x18f1,0x18f7,0x18fd,0x1903,0x1909,0x83c,0x840,0x190f,0x1915,0x191b,0x1921,0x1927,0x192d,0x1933,0x1939,
0x193f,0x1945,0x194b,0x1951,1,1,0x1957,0x195d,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,0xffb8,0xffcc,0xffcc,0xffcc,0xffcc,0xffb8,0xffcc,
0xffcc,0xffcc,0xffbc,0xffb8,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,
0xffcc,0xffcc,0xffb8,0xffcc,0xffcc,0xffbc,0xffc8,0xffcc,0xfe14,0xfe16,0xfe18,0xfe1a,0xfe1c,0xfe1e,0xfe20,0xfe22,
0xfe24,0xfe26,0xfe26,0xfe28,0xfe2a,0xfe2c,1,0xfe2e,1,0xfe30,0xfe32,1,0xffcc,0xffb8,1,0xfe24,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,
0xfe3c,0xfe3e,0xfe40,1,1,1,1,1,1,1,0x1962,0x1968,0x196f,0x1975,0x197b,0x844,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0x850,1,0x854,0xfe36,0xfe38,0xfe3a,0xfe3c,0xfe3e,
0xfe40,0xfe42,0xfe44,0xfdcc,0xfdcc,0xfdb8,0xffb8,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffb8,0xffcc,0xffcc,0xffb8,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
0xfe46,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
0x1981,0x858,0x1987,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0x85c,0x198d,1,0x860,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,1,0xffcc,
0xffcc,0xffcc,0xffcc,0xffb8,0xffcc,1,1,0xffcc,0xffcc,1,0xffb8,0xffcc,0xffcc,0xffb8,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
0xfe48,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0xffcc,
0xffb8,0xffcc,0xffcc,0xffb8,0xffcc,0xffcc,0xffb8,0xffb8,0xffb8,0xffcc,0xffb8,0xffb8,0xffcc,0xffb8,0xffcc,0xffcc,
0xffb8,0xffcc,0xffb8,0xffcc,0xffb8,0xffcc,0xffb8,0xffcc,0xffcc,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffb8,0xffcc,1,1,1,1,1,1,1,1,1,
0xffb8,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,0xffcc,0xffcc,0xffcc,0xffcc,1,0xffcc,0xffcc,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,0xffcc,0xffcc,0xffcc,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0xffb8,0xffb8,0xffb8,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0xffb8,
0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,0xffb8,
0xffcc,0xffcc,0xffb8,0xffcc,0xffcc,0xffb8,0xffcc,0xffcc,0xffcc,0xffb8,0xffb8,0xffb8,0xfe36,0xfe38,0xfe3a,0xffcc,
0xffcc,0xffcc,0xffb8,0xffcc,0xffcc,0xffb8,0xffb8,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,0x864,0x1993,1,1,1,1,1,1,0x868,0x1999,1,0x86c,
0x199f,1,1,1,1,1,1,1,0xfc0e,1,1,1,1,1,1,1,
1,1,1,1,1,1,0xfe12,1,1,1,0xffcc,0xffb8,0xffcc,0xffcc,1,1,
1,0x29ec,0x29f2,0x29f8,0x29fe,0x2a04,0x2a0a,0x2a10,0x2a16,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,0xfe0e,1,0xfc00,1,1,1,1,1,1,1,0x870,
1,1,1,0x19a5,0x19ab,0xfe12,1,1,1,1,1,1,1,1,1,0xfc00,
1,1,1,1,0x2a1c,0x2a22,1,0x2a28,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,0xffcc,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,0x2a2e,1,1,0x2a34,1,1,
1,1,1,0xfe0e,1,1,1,1,1,1,1,1,1,1,1,1,
1,0xfe12,1,1,1,1,1,1,1,1,1,1,1,0x2a3a,0x2a40,0x2a46,
1,1,0x2a4c,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0xfe0e,
1,1,1,1,1,1,1,1,1,1,1,1,1,0xfe12,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
0x878,0x19b1,1,1,0x19b7,0x19bd,0xfe12,1,1,1,1,1,1,1,1,0xfc00,
0xfc00,1,1,1,1,0x2a52,0x2a58,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,0x884,1,0x19c3,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,0xfc00,1,1,1,1,1,1,0x888,0x890,1,1,
0x19c9,0x19cf,0x19d5,0xfe12,1,1,1,1,1,1,1,1,1,0xfc00,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,0x894,1,0x19db,1,1,1,1,0xfe12,1,1,
1,1,1,1,1,0xfea8,0xfcb6,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,0xfe0e,1,1,0x898,0x19e1,1,0xfc00,1,1,1,0x89c,0x19e7,0x19ed,
1,0xdca,0x19f5,1,0xfe12,1,1,1,1,1,1,1,0xfc00,0xfc00,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,0xfe12,0xfe12,1,0xfc00,1,1,1,
1,1,1,0x8a8,0x8b0,1,1,0x19fd,0x1a03,0x1a09,0xfe12,1,1,1,1,1,
1,1,1,1,0xfc00,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,0xfc12,1,1,
1,1,0xfc00,1,1,1,1,1,1,1,1,1,0x8b4,0x1a0f,1,0xdd4,
0x1a17,0x1a1f,0xfc00,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,0xfece,0xfece,0xfe12,1,1,
1,1,1,1,1,1,0xfed6,0xfed6,0xfed6,0xfed6,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xfeec,0xfeec,0xfe12,1,1,1,1,1,1,1,1,0xfef4,0xfef4,0xfef4,
0xfef4,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,0xffb8,0xffb8,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xffb8,1,0xffb8,1,0xffb0,1,1,1,1,1,1,0x2a5f,1,1,
1,1,1,1,1,1,1,0x2a65,1,1,1,1,0x2a6b,1,1,1,
1,0x2a71,1,1,1,1,0x2a77,1,1,1,1,1,1,1,1,1,
1,1,1,0x2a7d,1,1,1,1,1,1,1,0xff02,0xff04,0x3c50,0xff08,0x3c58,
0x2a82,1,0x2a88,1,0xff04,0xff04,0xff04,0xff04,1,1,0xff04,0x3c60,0xffcc,0xffcc,0xfe12,1,
0xffcc,0xffcc,1,1,1,1,1,1,1,1,1,1,1,0x2a8f,1,1,
1,1,1,1,1,1,1,0x2a95,1,1,1,1,0x2a9b,1,1,1,
1,0x2aa1,1,1,1,1,0x2aa7,1,1,1,1,1,1,1,1,1,
1,1,1,0x2aad,1,1,1,1,1,1,0xffb8,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,0x8c0,0x1a25,1,1,1,1,1,1,1,0xfc00,1,1,
1,1,1,1,1,1,0xfe0e,1,0xfe12,0xfe12,1,1,1,1,1,1,
1,1,1,1,1,1,1,0xffb8,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,
0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,
0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,0xfe00,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,0xffcc,0xffcc,0xffcc,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0xfe12,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0xfe12,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xfe12,1,1,1,1,1,1,1,1,1,1,0xffcc,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,0xffc8,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,0xffbc,0xffcc,0xffb8,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,0xffcc,0xffb8,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,0xfe12,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,
0xffcc,1,1,0xffb8,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,0xffcc,
0xffcc,0xffb8,1,0xffb8,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,0x8c4,0x1a2b,0x8c8,0x1a31,0x8cc,0x1a37,0x8d0,0x1a3d,0x8d4,0x1a43,1,1,0x8d8,
0x1a49,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xfe0e,0xfc00,1,1,1,1,0x8dc,0x1a4f,0x8e0,0x1a55,0x8e4,0x8e8,0x1a5b,0x1a61,
0x8ec,0x1a67,0xfe12,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,0xffcc,0xffb8,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,
0xffcc,0xffcc,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,0xfe12,0xfe12,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,0xfe0e,1,1,1,1,1,1,1,1,1,1,1,
0xfe12,0xfe12,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xffcc,0xffcc,0xffcc,1,0xfe02,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,0xffcc,0xffcc,0xffb8,0xffb8,
0xffb8,0xffb8,0xffcc,1,0xfe02,0xfe02,0xfe02,0xfe02,0xfe02,0xfe02,0xfe02,1,1,1,1,0xffb8,
1,1,1,1,1,1,0xffcc,1,1,1,0xffcc,0xffcc,1,1,1,1,
1,1,0xffcc,0xffcc,0xffb8,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffb8,0xffcc,0xffcc,0xffd4,
0xffac,0xffb8,0xff94,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffd0,0xffc8,0xffc8,0xffb8,1,0xffcc,0xffd2,0xffb8,
0xffcc,0xffb8,0x1a6c,0x1a72,0x1a78,0x1a7e,0x1a85,0x1a8b,0x1a91,0x1a97,0x1a9f,0x1aa9,0x1ab0,0x1ab6,0x1abc,0x1ac2,
0x1ac8,0x1ace,0x1ad5,0x1adb,0x1ae0,0x1ae6,0x1aee,0x1af8,0x1b02,0x1b0c,0x1b14,0x1b1a,0x1b20,0x1b26,0x1b2f,0x1b39,
0x1b41,0x1b47,0x1b4c,0x1b52,0x1b58,0x1b5e,0x1b64,0x1b6a,0x1b70,0x1b76,0x1b7d,0x1b83,0x1b88,0x1b8e,0x1b94,0x1b9a,
0x1ba2,0x1bac,0x1bb4,0x1bba,0x1bc0,0x1bc6,0x1bcc,0x1bd2,0xdde,0xde8,0x1bda,0x1be4,0x1bec,0x1bf2,0x1bf8,0x1bfe,
0x1c04,0x1c0a,0x1c10,0x1c16,0x1c1d,0x1c23,0x1c28,0x1c2e,0x1c34,0x1c3a,0x1c40,0x1c46,0x1c4c,0x1c52,0x1c5a,0x1c64,
0x1c6e,0x1c78,0x1c82,0x1c8c,0x1c96,0x1ca0,0x1ca9,0x1caf,0x1cb5,0x1cbb,0x1cc0,0x1cc6,0xdf2,0xdfc,0x1cce,0x1cd8,
0x1ce0,0x1ce6,0x1cec,0x1cf2,0xe06,0xe10,0x1cfa,0x1d04,0x1d0e,0x1d18,0x1d22,0x1d2c,0x1d34,0x1d3a,0x1d40,0x1d46,
0x1d4c,0x1d52,0x1d58,0x1d5e,0x1d64,0x1d6a,0x1d70,0x1d76,0x1d7c,0x1d82,0x1d8a,0x1d94,0x1d9e,0x1da8,0x1db0,0x1db6,
0x1dbd,0x1dc3,0x1dc8,0x1dce,0x1dd4,0x1dda,0x1de0,0x1de6,0x1dec,0x1df2,0x1df9,0x1dff,0x1e05,0x1e0b,0x1e11,0x1e17,
0x1e1c,0x1e22,0x1e28,0x1e2e,0x1e35,0x1e3b,0x1e41,0x1e47,0x1e4c,0x1e52,0x1e58,0x1e5e,1,0x1e65,1,1,
1,1,0xe1a,0xe28,0x1e6a,0x1e70,0x1e78,0x1e82,0x1e8c,0x1e96,0x1ea0,0x1eaa,0x1eb4,0x1ebe,0x1ec8,0x1ed2,
0x1edc,0x1ee6,0x1ef0,0x1efa,0x1f04,0x1f0e,0x1f18,0x1f22,0x1f2c,0x1f36,0xe36,0xe40,0x1f3e,0x1f44,0x1f4a,0x1f50,
0x1f58,0x1f62,0x1f6c,0x1f76,0x1f80,0x1f8a,0x1f94,0x1f9e,0x1fa8,0x1fb2,0x1fba,0x1fc0,0x1fc6,0x1fcc,0xe4a,0xe54,
0x1fd2,0x1fd8,0x1fe0,0x1fea,0x1ff4,0x1ffe,0x2008,0x2012,0x201c,0x2026,0x2030,0x203a,0x2044,0x204e,0x2058,0x2062,
0x206c,0x2076,0x2080,0x208a,0x2094,0x209e,0x20a6,0x20ac,0x20b2,0x20b8,0x20c0,0x20ca,0x20d4,0x20de,0x20e8,0x20f2,
0x20fc,0x2106,0x2110,0x211a,0x2122,0x2128,0x212f,0x2135,0x213a,0x2140,0x2146,0x214c,1,1,1,1,
1,1,0xe5e,0xe74,0xe8c,0xe9a,0xea8,0xeb6,0xec4,0xed2,0xede,0xef4,0xf0c,0xf1a,0xf28,0xf36,
0xf44,0xf52,0xf5e,0xf6c,0x2155,0x215f,0x2169,0x2173,1,1,0xf7a,0xf88,0x217d,0x2187,0x2191,0x219b,
1,1,0xf96,0xfac,0xfc4,0xfd2,0xfe0,0xfee,0xffc,0x100a,0x1016,0x102c,0x1044,0x1052,0x1060,0x106e,
0x107c,0x108a,0x1096,0x10a8,0x21a5,0x21af,0x21b9,0x21c3,0x21cd,0x21d7,0x10ba,0x10cc,0x21e1,0x21eb,0x21f5,0x21ff,
0x2209,0x2213,0x10de,0x10ec,0x221d,0x2227,0x2231,0x223b,1,1,0x10fa,0x1108,0x2245,0x224f,0x2259,0x2263,
1,1,0x1116,0x1128,0x226d,0x2277,0x2281,0x228b,0x2295,0x229f,1,0x113a,1,0x22a9,1,0x22b3,
1,0x22bd,0x114c,0x1162,0x117a,0x1188,0x1196,0x11a4,0x11b2,0x11c0,0x11cc,0x11e2,0x11fa,0x1208,0x1216,0x1224,
0x1232,0x1240,0x124c,0x3b9e,0x22c5,0x3ba6,0x1256,0x3bae,0x22cb,0x3bb6,0x22d1,0x3bbe,0x22d7,0x3bc6,0x1260,0x3bce,
1,1,0x22de,0x22e8,0x22f7,0x2307,0x2317,0x2327,0x2337,0x2347,0x2352,0x235c,0x236b,0x237b,0x238b,0x239b,
0x23ab,0x23bb,0x23c6,0x23d0,0x23df,0x23ef,0x23ff,0x240f,0x241f,0x242f,0x243a,0x2444,0x2453,0x2463,0x2473,0x2483,
0x2493,0x24a3,0x24ae,0x24b8,0x24c7,0x24d7,0x24e7,0x24f7,0x2507,0x2517,0x2522,0x252c,0x253b,0x254b,0x255b,0x256b,
0x257b,0x258b,0x2595,0x259b,0x25a3,0x25aa,0x25b3,1,0x126a,0x25bd,0x25c5,0x25cb,0x25d1,0x3bd6,0x25d6,1,
0x2ab2,0x8f0,1,0x25dd,0x25e5,0x25ec,0x25f5,1,0x1274,0x25ff,0x2607,0x3bde,0x260d,0x3be6,0x2612,0x2619,
0x261f,0x2625,0x262b,0x2631,0x2639,0x3bf0,1,1,0x2641,0x2649,0x2651,0x2657,0x265d,0x3bfa,1,0x2663,
0x2669,0x266f,0x2675,0x267b,0x2683,0x3c04,0x268b,0x2691,0x2697,0x269f,0x26a7,0x26ad,0x26b3,0x3c0e,0x26b9,0x26bf,
0x3c16,0x2ab7,1,1,0x26c7,0x26ce,0x26d7,1,0x127e,0x26e1,0x26e9,0x3c1e,0x26ef,0x3c26,0x26f4,0x2abb,
0x8fc,1,0xfa09,0xfa09,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xffcc,0xffcc,0xfe02,0xfe02,0xffcc,0xffcc,0xffcc,0xffcc,0xfe02,0xfe02,0xfe02,0xffcc,0xffcc,1,
1,1,1,0xffcc,1,1,1,0xfe02,0xfe02,0xffcc,0xffb8,0xffcc,0xfe02,0xfe02,0xffb8,0xffb8,
0xffb8,0xffb8,0xffcc,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,0x2abe,1,1,1,0x2ac2,0x3c2e,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,0x908,1,0x90c,1,0x910,1,1,1,1,1,0x26fb,0x2701,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,0x2707,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,0x270d,0x2713,0x2719,0x914,1,0x918,1,0x91c,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,0x920,0x271f,1,1,1,0x924,0x2725,1,0x928,0x272b,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,0x92c,0x2731,0x930,0x2737,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0x934,
1,1,1,0x273d,1,0x938,0x2743,0x93c,1,0x2749,0x940,0x274f,1,1,1,0x944,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0x2755,0x948,0x275b,1,0x94c,0x950,1,1,1,1,1,1,1,0x2761,
0x2767,0x276d,0x2773,0x2779,0x954,0x958,0x277f,0x2785,0x95c,0x960,0x278b,0x2791,0x964,0x968,0x96c,0x970,
1,1,0x2797,0x279d,0x974,0x978,0x27a3,0x27a9,0x97c,0x980,0x27af,0x27b5,1,1,1,1,
1,1,1,0x984,0x988,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,0x98c,1,1,1,1,1,0x990,0x994,1,0x998,0x27bb,0x27c1,
0x27c7,0x27cd,1,1,0x99c,0x9a0,0x9a4,0x9a8,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0x27d3,0x27d9,0x27df,0x27e5,1,1,1,1,
1,1,0x27eb,0x27f1,0x27f7,0x27fd,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0x2ac7,
0x2acb,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,0x2acf,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,0xfe12,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0xffb4,0xffc8,0xffd0,0xffbc,0xffc0,0xffc0,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0x9ac,1,
1,1,1,0x9b0,0x2803,0x9b4,0x2809,0x9b8,0x280f,0x9bc,0x2815,0x9c0,0x281b,0x9c4,0x2821,0x9c8,
0x2827,0x9cc,0x282d,0x9d0,0x2833,0x9d4,0x2839,0x9d8,0x283f,0x9dc,0x2845,1,0x9e0,0x284b,0x9e4,0x2851,
0x9e8,0x2857,1,1,1,1,1,0x9ec,0x285d,0x2863,0x9f4,0x2869,0x286f,0x9fc,0x2875,0x287b,
0xa04,0x2881,0x2887,0xa0c,0x288d,0x2893,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,0x2899,1,1,1,1,0xfc10,
0xfc10,1,1,0xa14,0x289f,1,1,1,1,1,1,1,0xa18,1,1,1,
1,0xa1c,0x28a5,0xa20,0x28ab,0xa24,0x28b1,0xa28,0x28b7,0xa2c,0x28bd,0xa30,0x28c3,0xa34,0x28c9,0xa38,
0x28cf,0xa3c,0x28d5,0xa40,0x28db,0xa44,0x28e1,0xa48,0x28e7,1,0xa4c,0x28ed,0xa50,0x28f3,0xa54,0x28f9,
1,1,1,1,1,0xa58,0x28ff,0x2905,0xa60,0x290b,0x2911,0xa68,0x2917,0x291d,0xa70,0x2923,
0x2929,0xa78,0x292f,0x2935,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,0xa80,0xa84,0xa88,0xa8c,1,0x293b,1,1,0x2941,0x2947,0x294d,
0x2953,1,1,0xa90,0x2959,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,0xffcc,1,1,1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,0xffcc,0xffcc,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,0xffcc,0xffcc,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,0xfe12,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,0xfe12,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xffb8,0xffb8,0xffb8,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0xfe12,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0xffcc,1,0xffcc,0xffcc,0xffb8,1,1,0xffcc,
0xffcc,1,1,1,1,1,0xffcc,0xffcc,1,0xffcc,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0xfe12,1,
1,1,1,1,1,1,1,1,0xae2,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,
0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,
0x1289,0x1289,0x1289,0x1289,0xae2,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,
0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,
0xae2,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,
0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0xae2,0x1289,0x1289,0x1289,
0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,
0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,0x1289,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,0x3c66,1,0x3c66,0x3c66,0x3c66,0x3c66,0x3c66,0x3c66,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0x3c66,0x3c66,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,0x3c66,1,1,1,1,0x3c66,1,1,1,0x3c66,1,0x3c66,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,0x3b97,1,0x2ad5,
0x2ad9,0x2add,0x2ae1,0x2ae5,0x2ae9,0x2aed,0x2af1,0x2af1,0x2af5,0x2af9,0x2afd,0x2b01,0x2b05,0x2b09,0x2b0d,0x2b11,
0x2b15,0x2b19,0x2b1d,0x2b21,0x2b25,0x2b29,0x2b2d,0x2b31,0x2b35,0x2b39,0x2b3d,0x2b41,0x2b45,0x2b49,0x2b4d,0x2b51,
0x2b55,0x2b59,0x2b5d,0x2b61,0x2b65,0x2b69,0x2b6d,0x2b71,0x2b75,0x2b79,0x2b7d,0x2b81,0x2b85,0x2b89,0x2b8d,0x2b91,
0x2b95,0x2b99,0x2b9d,0x2ba1,0x2ba5,0x2ba9,0x2bad,0x2bb1,0x2bb5,0x2bb9,0x2bbd,0x2bc1,0x2bc5,0x2bc9,0x2bcd,0x2bd1,
0x2bd5,0x2bd9,0x2bdd,0x2be1,0x2be5,0x2be9,0x2bed,0x2bf1,0x2bf5,0x2bf9,0x2bfd,0x2c01,0x2c05,0x2c09,0x2c0d,0x2c11,
0x2c15,0x2c19,0x2c1d,0x2c21,0x2c25,0x2c29,0x2c2d,0x2c31,0x2c35,0x2c39,0x2c3d,0x2b21,0x2c41,0x2c45,0x2c49,0x2c4d,
0x2c51,0x2c55,0x2c59,0x2c5d,0x2c61,0x2c65,0x2c69,0x2c6d,0x2c71,0x2c75,0x2c79,0x2c7d,0x2c81,0x2c85,0x2c89,0x2c8d,
0x2c91,0x2c95,0x2c99,0x2c9d,0x2ca1,0x2ca5,0x2ca9,0x2cad,0x2cb1,0x2cb5,0x2cb9,0x2cbd,0x2cc1,0x2cc5,0x2cc9,0x2ccd,
0x2cd1,0x2cd5,0x2cd9,0x2cdd,0x2ce1,0x2ce5,0x2ce9,0x2ced,0x2cf1,0x2cf5,0x2cf9,0x2cfd,0x2d01,0x2d05,0x2d09,0x2d0d,
0x2d11,0x2d15,0x2d19,0x2d1d,0x2d21,0x2d25,0x2d29,0x2d2d,0x2d31,0x2d35,0x2d39,0x2d3d,0x2d41,0x2d45,0x2d49,0x2d4d,
0x2c89,0x2d51,0x2d55,0x2d59,0x2d5d,0x2d61,0x2d65,0x2d69,0x2d6d,0x2c49,0x2d71,0x2d75,0x2d79,0x2d7d,0x2d81,0x2d85,
0x2d89,0x2d8d,0x2d91,0x2d95,0x2d99,0x2d9d,0x2da1,0x2da5,0x2da9,0x2dad,0x2db1,0x2db5,0x2db9,0x2dbd,0x2b21,0x2dc1,
0x2dc5,0x2dc9,0x2dcd,0x2dd1,0x2dd5,0x2dd9,0x2ddd,0x2de1,0x2de5,0x2de9,0x2ded,0x2df1,0x2df5,0x2df9,0x2dfd,0x2e01,
0x2e05,0x2e09,0x2e0d,0x2e11,0x2e15,0x2e19,0x2e1d,0x2e21,0x2e25,0x2e29,0x2c51,0x2e2d,0x2e31,0x2e35,0x2e39,0x2e3d,
0x2e41,0x2e45,0x2e49,0x2e4d,0x2e51,0x2e55,0x2e59,0x2e5d,0x2e61,0x2e65,0x2e69,0x2e6d,0x2e71,0x2e75,0x2e79,0x2e7d,
0x2e81,0x2e85,0x2e89,0x2e8d,0x2e91,0x2e95,0x2e99,0x2e9d,0x2ea1,0x2ea5,0x2ea9,0x2ead,0x2eb1,0x2eb5,0x2eb9,0x2ebd,
0x2ec1,0x2ec5,0x2ec9,0x2ecd,0x2ed1,0x2ed5,0x2ed9,0x2edd,0x2ee1,0x2ee5,0x2ee9,0x2eed,0x2ef1,1,1,0x2ef5,
1,0x2ef9,1,1,0x2efd,0x2f01,0x2f05,0x2f09,0x2f0d,0x2f11,0x2f15,0x2f19,0x2f1d,0x2f21,1,0x2f25,
1,0x2f29,1,1,0x2f2d,0x2f31,1,1,1,0x2f35,0x2f39,0x2f3d,0x2f41,0x2f45,0x2f49,0x2f4d,
0x2f51,0x2f55,0x2f59,0x2f5d,0x2f61,0x2f65,0x2f69,0x2f6d,0x2f71,0x2f75,0x2f79,0x2f7d,0x2f81,0x2f85,0x2f89,0x2f8d,
0x2f91,0x2f95,0x2f99,0x2f9d,0x2fa1,0x2fa5,0x2fa9,0x2fad,0x2fb1,0x2fb5,0x2fb9,0x2fbd,0x2fc1,0x2fc5,0x2fc9,0x2fcd,
0x2fd1,0x2fd5,0x2fd9,0x2fdd,0x2fe1,0x2fe5,0x2d25,0x2fe9,0x2fed,0x2ff1,0x2ff5,0x2ff9,0x2ffd,0x2ffd,0x3001,0x3005,
0x3009,0x300d,0x3011,0x3015,0x3019,0x301d,0x2f2d,0x3021,0x3025,0x3029,0x302d,0x3031,0x3037,1,1,0x303b,
0x303f,0x3043,0x3047,0x304b,0x304f,0x3053,0x3057,0x2f65,0x305b,0x305f,0x3063,0x2ef5,0x3067,0x306b,0x306f,0x3073,
0x3077,0x307b,0x307f,0x3083,0x3087,0x308b,0x308f,0x3093,0x2f89,0x3097,0x2f8d,0x309b,0x309f,0x30a3,0x30a7,0x30ab,
0x2ef9,0x2b75,0x30af,0x30b3,0x30b7,0x2c8d,0x2de9,0x30bb,0x30bf,0x2fa9,0x30c3,0x2fad,0x30c7,0x30cb,0x30cf,0x2f01,
0x30d3,0x30d7,0x30db,0x30df,0x30e3,0x2f05,0x30e7,0x30eb,0x30ef,0x30f3,0x30f7,0x30fb,0x2fe5,0x30ff,0x3103,0x2d25,
0x3107,0x2ff5,0x310b,0x310f,0x3113,0x3117,0x311b,0x3009,0x311f,0x2f29,0x3123,0x300d,0x2c41,0x3127,0x3011,0x312b,
0x3019,0x312f,0x3133,0x3137,0x313b,0x313f,0x3021,0x2f19,0x3143,0x3025,0x3147,0x3029,0x314b,0x2af1,0x314f,0x3155,
0x315b,0x3161,0x3165,0x3169,0x316d,0x3173,0x3179,0x317f,0x3183,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0x3186,
0xfe34,0x318c,1,1,1,1,1,1,1,1,1,1,0x3192,0x3198,0x31a0,0x31aa,
0x31b2,0x31b8,0x31be,0x31c4,0x31ca,0x31d0,0x31d6,0x31dc,0x31e2,1,0x31e8,0x31ee,0x31f4,0x31fa,0x3200,1,
0x3206,1,0x320c,0x3212,1,0x3218,0x321e,1,0x3224,0x322a,0x3230,0x3236,0x323c,0x3242,0x3248,0x324e,
0x3254,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,
0xffcc,0xffcc,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xffb8,1,0xffcc,1,1,1,1,1,1,1,1,0xffcc,0xfe02,0xffb8,
1,1,1,1,0xfe12,1,1,1,1,0xffcc,0xffcc,0xffcc,0xffcc,1,1,1,
1,1,1,1,1,0xffb8,0xffb8,0xffcc,0xffcc,0xffcc,0xffb8,0xffcc,0xffb8,0xffb8,0xffb8,1,
1,1,1,1,1,1,1,1,0xa94,0x295f,0xa9a,0x2969,1,1,1,1,
1,0xaa0,1,1,1,1,1,0x2973,1,1,1,1,1,1,1,1,
1,0xfe12,0xfc0e,1,1,1,1,1,1,1,0xfc00,1,1,1,1,1,
1,0x297d,0x2987,1,0xaa6,0xaac,0xfe12,0xfe12,1,1,1,1,1,1,1,1,
1,1,1,0xfe12,1,1,1,1,1,1,1,1,1,0xfe0e,1,1,
1,1,1,0xfe12,0xfe0e,1,1,1,1,1,1,1,1,1,0xfe0e,0xfe12,
1,1,1,1,1,1,1,1,1,1,1,0xfe0e,0xfe0e,1,0xfc00,1,
1,1,1,1,1,1,0xab2,1,1,1,0x2991,0x299b,0xfe12,1,1,1,
1,1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,1,1,0xfe12,1,1,
1,0xfe0e,1,1,1,1,1,1,1,1,1,0xfc00,1,1,1,1,
1,1,1,1,0xabe,0xfc00,0x29a5,0x29af,0xfc00,0x29b9,1,1,0xfe12,0xfe0e,1,1,
1,1,1,1,1,1,1,1,1,1,0xad0,0xad6,0x29c3,0x29cd,1,1,
1,0xfe12,0xfe0e,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xfe12,0xfe0e,1,1,1,1,1,1,1,1,0xfc00,1,1,1,
1,0xadc,1,1,0x29d7,1,1,1,1,0xfe12,0xfe12,1,0xfe02,0xfe02,0xfe02,0xfe02,
0xfe02,1,1,1,1,1,1,1,1,1,1,1,0xfe0c,0xfe0c,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,0xfe02,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,0x325a,0x3264,0x3278,0x3290,0x32a8,
0x32c0,0x32d8,0xffb0,0xffb0,0xfe02,0xfe02,0xfe02,1,1,1,0xffc4,0xffb0,0xffb0,0xffb0,1,1,
1,1,1,1,1,1,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,1,1,0xffcc,0xffcc,0xffcc,
0xffcc,0xffcc,0xffb8,0xffb8,1,1,1,1,1,1,1,1,1,1,0xffcc,0xffcc,
0xffcc,0xffcc,1,1,1,1,1,1,1,1,1,1,1,0x32e6,0x32f0,0x3304,
0x331c,0x3334,0x334c,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,1,0xffcc,0xffcc,1,0xffcc,0xffcc,
0xffcc,0xffcc,0xffcc,1,1,1,1,1,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,0xffb8,1,
1,1,1,1,1,1,1,1,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xffcc,0xfe0e,1,
1,1,1,1,0x335b,0x335f,0x3363,0x3367,0x336d,0x2f4d,0x3371,0x3375,0x3379,0x337d,0x2f51,0x3381,
0x3385,0x3389,0x2f55,0x338f,0x3393,0x3397,0x339b,0x33a1,0x33a5,0x33a9,0x33ad,0x33b3,0x33b7,0x33bb,0x33bf,0x303f,
0x33c3,0x33c9,0x33cd,0x33d1,0x33d5,0x33d9,0x33dd,0x33e1,0x33e5,0x3053,0x2f59,0x2f5d,0x3057,0x33e9,0x33ed,0x2c59,
0x33f1,0x2f61,0x33f5,0x33f9,0x33fd,0x3401,0x3401,0x3401,0x3405,0x340b,0x340f,0x3413,0x3417,0x341d,0x3421,0x3425,
0x3429,0x342d,0x3431,0x3435,0x3439,0x343d,0x3441,0x3445,0x3449,0x344d,0x344d,0x305f,0x3451,0x3455,0x3459,0x345d,
0x2f69,0x3461,0x3465,0x3469,0x2ebd,0x346d,0x3471,0x3475,0x3479,0x347d,0x3481,0x3485,0x3489,0x348d,0x3493,0x3497,
0x349b,0x349f,0x34a3,0x34a7,0x34ab,0x34b1,0x34b7,0x34bb,0x34bf,0x34c3,0x34c7,0x34cb,0x34cf,0x34d3,0x34d7,0x34d7,
0x34db,0x34e1,0x34e5,0x2c49,0x34e9,0x34ed,0x34f3,0x34f7,0x34fb,0x34ff,0x3503,0x3507,0x2f7d,0x350b,0x350f,0x3513,
0x3519,0x351d,0x3523,0x3527,0x352b,0x352f,0x3533,0x3537,0x353b,0x353f,0x3543,0x3547,0x354b,0x354f,0x3555,0x3559,
0x355d,0x3561,0x2b71,0x3565,0x356b,0x356f,0x356f,0x3575,0x3579,0x3579,0x357d,0x3581,0x3587,0x358d,0x3591,0x3595,
0x3599,0x359d,0x35a1,0x35a5,0x35a9,0x35ad,0x35b1,0x2f81,0x35b5,0x35bb,0x35bf,0x35c3,0x308f,0x35c3,0x35c7,0x2f89,
0x35cb,0x35cf,0x35d3,0x35d7,0x2f8d,0x2b05,0x35db,0x35df,0x35e3,0x35e7,0x35eb,0x35ef,0x35f3,0x35f9,0x35fd,0x3601,
0x3605,0x3609,0x360d,0x3613,0x3617,0x361b,0x361f,0x3623,0x3627,0x362b,0x362f,0x3633,0x2f91,0x3637,0x363b,0x3641,
0x3645,0x3649,0x364d,0x2f99,0x3651,0x3655,0x3659,0x365d,0x3661,0x3665,0x3669,0x366d,0x2b75,0x30af,0x3671,0x3675,
0x3679,0x367d,0x3683,0x3687,0x368b,0x368f,0x2f9d,0x3693,0x3699,0x369d,0x36a1,0x3161,0x36a5,0x36a9,0x36ad,0x36b1,
0x36b5,0x36bb,0x36bf,0x36c3,0x36c7,0x36cd,0x36d1,0x36d5,0x36d9,0x2c8d,0x36dd,0x36e1,0x36e7,0x36ed,0x36f3,0x36f7,
0x36fd,0x3701,0x3705,0x3709,0x370d,0x2fa1,0x2de9,0x3711,0x3715,0x3719,0x371d,0x3723,0x3727,0x372b,0x372f,0x30bf,
0x3733,0x3737,0x373d,0x3741,0x3745,0x374b,0x3751,0x3755,0x30c3,0x3759,0x375d,0x3761,0x3765,0x3769,0x376d,0x3771,
0x3777,0x377b,0x3781,0x3785,0x378b,0x30cb,0x378f,0x3793,0x3799,0x379d,0x37a1,0x37a7,0x37ad,0x37b1,0x37b5,0x37b9,
0x37bd,0x37bd,0x37c1,0x37c5,0x30d3,0x37c9,0x37cd,0x37d1,0x37d5,0x37d9,0x37df,0x37e3,0x2c55,0x37e9,0x37ef,0x37f3,
0x37f9,0x37ff,0x3805,0x3809,0x30eb,0x380d,0x3813,0x3819,0x381f,0x3825,0x3829,0x3829,0x30ef,0x3169,0x382d,0x3831,
0x3835,0x3839,0x383f,0x2bbd,0x30f7,0x3843,0x3847,0x2fcd,0x384d,0x3853,0x2f15,0x3859,0x385d,0x2fdd,0x3861,0x3865,
0x3869,0x386f,0x386f,0x3875,0x3879,0x387d,0x3883,0x3887,0x388b,0x388f,0x3895,0x3899,0x389d,0x38a1,0x38a5,0x38a9,
0x38af,0x38b3,0x38b7,0x38bb,0x38bf,0x38c3,0x38c7,0x38cd,0x38d3,0x38d7,0x38dd,0x38e1,0x38e7,0x38eb,0x2ff5,0x38ef,
0x38f5,0x38fb,0x38ff,0x3905,0x3909,0x390f,0x3913,0x3917,0x391b,0x391f,0x3923,0x3927,0x392d,0x3933,0x3939,0x3575,
0x393f,0x3943,0x3947,0x394b,0x394f,0x3953,0x3957,0x395b,0x395f,0x3963,0x3967,0x396b,0x2c9d,0x3971,0x3975,0x3979,
0x397d,0x3981,0x3985,0x3001,0x3989,0x398d,0x3991,0x3995,0x3999,0x399f,0x39a5,0x39ab,0x39af,0x39b3,0x39b7,0x39bb,
0x39c1,0x39c5,0x39cb,0x39cf,0x39d3,0x39d9,0x39df,0x39e3,0x2ba9,0x39e7,0x39eb,0x39ef,0x39f3,0x39f7,0x39fb,0x3113,
0x39ff,0x3a03,0x3a07,0x3a0b,0x3a0f,0x3a13,0x3a17,0x3a1b,0x3a1f,0x3a23,0x3a29,0x3a2d,0x3a31,0x3a35,0x3a39,0x3a3d,
0x3a43,0x3a49,0x3a4d,0x3a51,0x3127,0x312b,0x3a55,0x3a59,0x3a5f,0x3a63,0x3a67,0x3a6b,0x3a6f,0x3a75,0x3a7b,0x3a7f,
0x3a83,0x3a87,0x3a8d,0x312f,0x3a91,0x3a97,0x3a9d,0x3aa1,0x3aa5,0x3aa9,0x3aaf,0x3ab3,0x3ab7,0x3abb,0x3abf,0x3ac3,
0x3ac7,0x3acb,0x3ad1,0x3ad5,0x3ad9,0x3add,0x3ae3,0x3ae7,0x3aeb,0x3aef,0x3af3,0x3af9,0x3aff,0x3b03,0x3b07,0x3b0b,
0x3b11,0x3b15,0x3147,0x3147,0x3b1b,0x3b1f,0x3b25,0x3b29,0x3b2d,0x3b31,0x3b35,0x3b39,0x3b3d,0x3b41,0x314b,0x3b47,
0x3b4b,0x3b4f,0x3b53,0x3b57,0x3b5b,0x3b61,0x3b65,0x3b6b,0x3b71,0x3b77,0x3b7b,0x3b7f,0x3b83,0x3b87,0x3b8b,0x3b8f,
0x3b93,0x3b97,1,1
};

static const UCPTrie norm2_nfc_data_trie={
    norm2_nfc_data_trieIndex,
    { norm2_nfc_data_trieData },
    1746, 7892,
    0x2fc00, 0x30,
    0, 0,
    0, 0,
    0xc4, 0x226,
    0x1,
};

static const uint16_t norm2_nfc_data_extraData[7732]={
0xffff,0xffff,0x8670,0x44dc,0x8670,0x44c0,0x8670,0x44de,0x600,0x180,0x602,0x182,0x604,0x185,0x606,0x186,
0x608,0x200,0x60c,0x205,0x60e,0x44d,0x610,0x189,0x612,0x3d44,0x614,0x18b,0x618,0x39a,0x61e,0x400,
0x622,0x404,0x646,0x3d41,0x64a,0x3c00,0x8650,0x208,0x60e,0x3c04,0x646,0x3c08,0x8662,0x3c0c,0x602,0x20c,
0x604,0x210,0x60e,0x214,0x618,0x218,0x864e,0x18f,0x60e,0x3c14,0x618,0x21c,0x646,0x3c18,0x64e,0x3c20,
0x65a,0x3c24,0x8662,0x3c1c,0x600,0x190,0x602,0x192,0x604,0x195,0x606,0x3d78,0x608,0x225,0x60c,0x228,
0x60e,0x22c,0x610,0x196,0x612,0x3d74,0x618,0x234,0x61e,0x408,0x622,0x40c,0x646,0x3d71,0x64e,0x451,
0x650,0x230,0x65a,0x3c30,0x8660,0x3c34,0x860e,0x3c3c,0x602,0x3e8,0x604,0x238,0x608,0x3c40,0x60c,0x23c,
0x60e,0x240,0x618,0x3cc,0x864e,0x244,0x604,0x248,0x60e,0x3c44,0x610,0x3c4c,0x618,0x43c,0x646,0x3c48,
0x64e,0x3c50,0x865c,0x3c54,0x600,0x198,0x602,0x19a,0x604,0x19c,0x606,0x250,0x608,0x254,0x60c,0x258,
0x60e,0x260,0x610,0x19f,0x612,0x3d90,0x618,0x39e,0x61e,0x410,0x622,0x414,0x646,0x3d94,0x650,0x25c,
0x8660,0x3c58,0x8604,0x268,0x602,0x3c60,0x618,0x3d0,0x646,0x3c64,0x64e,0x26c,0x8662,0x3c68,0x602,0x272,
0x618,0x27a,0x646,0x3c6d,0x64e,0x276,0x65a,0x3c78,0x8662,0x3c74,0x602,0x3c7c,0x60e,0x3c80,0x8646,0x3c84,
0x600,0x3f0,0x602,0x286,0x606,0x1a2,0x60e,0x3c88,0x618,0x28e,0x646,0x3c8c,0x64e,0x28a,0x65a,0x3c94,
0x8662,0x3c90,0x600,0x1a4,0x602,0x1a6,0x604,0x1a9,0x606,0x1ab,0x608,0x299,0x60c,0x29c,0x60e,0x45d,
0x610,0x1ad,0x612,0x3d9c,0x616,0x2a0,0x618,0x3a2,0x61e,0x418,0x622,0x41c,0x636,0x341,0x646,0x3d99,
0x8650,0x3d5,0x602,0x3ca8,0x860e,0x3cac,0x602,0x2a8,0x60e,0x3cb0,0x618,0x2b0,0x61e,0x420,0x622,0x424,
0x646,0x3cb5,0x64e,0x2ac,0x8662,0x3cbc,0x602,0x2b5,0x604,0x2b8,0x60e,0x3cc0,0x618,0x2c1,0x646,0x3cc5,
0x64c,0x430,0x864e,0x2bc,0x60e,0x3cd4,0x618,0x2c8,0x646,0x3cd8,0x64c,0x434,0x64e,0x2c4,0x65a,0x3ce0,
0x8662,0x3cdc,0x600,0x1b2,0x602,0x1b4,0x604,0x1b6,0x606,0x2d1,0x608,0x2d5,0x60c,0x2d8,0x610,0x1b9,
0x612,0x3dcc,0x614,0x2dc,0x616,0x2e0,0x618,0x3a6,0x61e,0x428,0x622,0x42c,0x636,0x35f,0x646,0x3dc8,
0x648,0x3ce4,0x650,0x2e4,0x65a,0x3cec,0x8660,0x3ce8,0x606,0x3cf8,0x8646,0x3cfc,0x600,0x3d00,0x602,0x3d04,
0x604,0x2e8,0x60e,0x3d0c,0x610,0x3d08,0x8646,0x3d10,0x60e,0x3d14,0x8610,0x3d18,0x600,0x3de4,0x602,0x1ba,
0x604,0x2ec,0x606,0x3df0,0x608,0x464,0x60e,0x3d1c,0x610,0x2f0,0x612,0x3dec,0x8646,0x3de8,0x602,0x2f2,
0x604,0x3d20,0x60e,0x2f6,0x618,0x2fa,0x646,0x3d24,0x8662,0x3d28,0x600,0x1c0,0x602,0x1c2,0x604,0x1c5,
0x606,0x1c6,0x608,0x202,0x60c,0x207,0x60e,0x44f,0x610,0x1c9,0x612,0x3d46,0x614,0x1cb,0x618,0x39c,
0x61e,0x402,0x622,0x406,0x646,0x3d43,0x64a,0x3c02,0x8650,0x20a,0x60e,0x3c06,0x646,0x3c0a,0x8662,0x3c0e,
0x602,0x20e,0x604,0x212,0x60e,0x216,0x618,0x21a,0x864e,0x1cf,0x60e,0x3c16,0x618,0x21e,0x646,0x3c1a,
0x64e,0x3c22,0x65a,0x3c26,0x8662,0x3c1e,0x600,0x1d0,0x602,0x1d2,0x604,0x1d5,0x606,0x3d7a,0x608,0x227,
0x60c,0x22a,0x60e,0x22e,0x610,0x1d6,0x612,0x3d76,0x618,0x236,0x61e,0x40a,0x622,0x40e,0x646,0x3d73,
0x64e,0x453,0x650,0x232,0x65a,0x3c32,0x8660,0x3c36,0x860e,0x3c3e,0x602,0x3ea,0x604,0x23a,0x608,0x3c42,
0x60c,0x23e,0x60e,0x242,0x618,0x3ce,0x864e,0x246,0x604,0x24a,0x60e,0x3c46,0x610,0x3c4e,0x618,0x43e,
0x646,0x3c4a,0x64e,0x3c52,0x65c,0x3c56,0x8662,0x3d2c,0x600,0x1d8,0x602,0x1da,0x604,0x1dc,0x606,0x252,
0x608,0x256,0x60c,0x25a,0x610,0x1df,0x612,0x3d92,0x618,0x3a0,0x61e,0x412,0x622,0x416,0x646,0x3d96,
0x650,0x25e,0x8660,0x3c5a,0x604,0x26a,0x8618,0x3e0,0x602,0x3c62,0x618,0x3d2,0x646,0x3c66,0x64e,0x26e,
0x8662,0x3c6a,0x602,0x274,0x618,0x27c,0x646,0x3c6f,0x64e,0x278,0x65a,0x3c7a,0x8662,0x3c76,0x602,0x3c7e,
0x60e,0x3c82,0x8646,0x3c86,0x600,0x3f2,0x602,0x288,0x606,0x1e2,0x60e,0x3c8a,0x618,0x290,0x646,0x3c8e,
0x64e,0x28c,0x65a,0x3c96,0x8662,0x3c92,0x600,0x1e4,0x602,0x1e6,0x604,0x1e9,0x606,0x1eb,0x608,0x29b,
0x60c,0x29e,0x60e,0x45f,0x610,0x1ed,0x612,0x3d9e,0x616,0x2a2,0x618,0x3a4,0x61e,0x41a,0x622,0x41e,
0x636,0x343,0x646,0x3d9b,0x8650,0x3d7,0x602,0x3caa,0x860e,0x3cae,0x602,0x2aa,0x60e,0x3cb2,0x618,0x2b2,
0x61e,0x422,0x622,0x426,0x646,0x3cb7,0x64e,0x2ae,0x8662,0x3cbe,0x602,0x2b7,0x604,0x2ba,0x60e,0x3cc2,
0x618,0x2c3,0x646,0x3cc7,0x64c,0x432,0x864e,0x2be,0x60e,0x3cd6,0x610,0x3d2e,0x618,0x2ca,0x646,0x3cda,
0x64c,0x436,0x64e,0x2c6,0x65a,0x3ce2,0x8662,0x3cde,0x600,0x1f2,0x602,0x1f4,0x604,0x1f6,0x606,0x2d3,
0x608,0x2d7,0x60c,0x2da,0x610,0x1f9,0x612,0x3dce,0x614,0x2de,0x616,0x2e2,0x618,0x3a8,0x61e,0x42a,
0x622,0x42e,0x636,0x361,0x646,0x3dca,0x648,0x3ce6,0x650,0x2e6,0x65a,0x3cee,0x8660,0x3cea,0x606,0x3cfa,
0x8646,0x3cfe,0x600,0x3d02,0x602,0x3d06,0x604,0x2ea,0x60e,0x3d0e,0x610,0x3d0a,0x614,0x3d30,0x8646,0x3d12,
0x60e,0x3d16,0x8610,0x3d1a,0x600,0x3de6,0x602,0x1fa,0x604,0x2ee,0x606,0x3df2,0x608,0x466,0x60e,0x3d1e,
0x610,0x1fe,0x612,0x3dee,0x614,0x3d32,0x8646,0x3dea,0x602,0x2f4,0x604,0x3d22,0x60e,0x2f8,0x618,0x2fc,
0x646,0x3d26,0x8662,0x3d2a,0x600,0x3fda,0x602,0x70a,0x8684,0x3f82,0x602,0x3f8,0x8608,0x3c4,0x8602,0x3fc,
0x602,0x3fa,0x8608,0x3c6,0x8602,0x3fe,0x860e,0x3d36,0x8618,0x3dc,0x8618,0x3de,0x600,0x3f74,0x602,0x70c,
0x608,0x3f72,0x60c,0x3f70,0x626,0x3e11,0x628,0x3e13,0x868a,0x3f78,0x600,0x3f90,0x602,0x710,0x626,0x3e31,
0x8628,0x3e33,0x600,0x3f94,0x602,0x712,0x626,0x3e51,0x628,0x3e53,0x868a,0x3f98,0x600,0x3fb4,0x602,0x714,
0x608,0x3fb2,0x60c,0x3fb0,0x610,0x754,0x626,0x3e71,0x8628,0x3e73,0x600,0x3ff0,0x602,0x718,0x626,0x3e91,
0x8628,0x3e93,0x8628,0x3fd8,0x600,0x3fd4,0x602,0x71c,0x608,0x3fd2,0x60c,0x3fd0,0x610,0x756,0x8628,0x3eb3,
0x600,0x3ff4,0x602,0x71e,0x626,0x3ed1,0x628,0x3ed3,0x868a,0x3ff8,0x600,0x3ee1,0x602,0x759,0x608,0x3f62,
0x60c,0x3f60,0x626,0x3e01,0x628,0x3e03,0x684,0x3f6d,0x868a,0x3f66,0x600,0x3ee4,0x602,0x75a,0x626,0x3e21,
0x8628,0x3e23,0x600,0x3ee9,0x602,0x75d,0x626,0x3e41,0x628,0x3e43,0x684,0x3f8d,0x868a,0x3f86,0x600,0x3eec,
0x602,0x75e,0x608,0x3fa2,0x60c,0x3fa0,0x610,0x795,0x626,0x3e61,0x628,0x3e63,0x8684,0x3fac,0x600,0x3ef0,
0x602,0x798,0x626,0x3e81,0x8628,0x3e83,0x626,0x3fc8,0x8628,0x3fca,0x600,0x3ef4,0x602,0x79a,0x608,0x3fc2,
0x60c,0x3fc0,0x610,0x797,0x626,0x3ea1,0x628,0x3ea3,0x8684,0x3fcc,0x600,0x3ef9,0x602,0x79d,0x626,0x3ec1,
0x628,0x3ec3,0x684,0x3fed,0x868a,0x3fe6,0x602,0x7a6,0x8610,0x7a8,0x8610,0x80e,0x60c,0x9a0,0x8610,0x9a4,
0x8602,0x806,0x600,0x800,0x60c,0x9ac,0x8610,0x802,0x60c,0x982,0x8610,0x9b8,0x8610,0x9bc,0x600,0x81a,
0x608,0x9c4,0x60c,0x832,0x8610,0x9c8,0x8602,0x818,0x8610,0x9cc,0x608,0x9dc,0x60c,0x81c,0x610,0x9e0,
0x8616,0x9e4,0x8610,0x9e8,0x8610,0x9f0,0x8610,0x9d8,0x60c,0x9a2,0x8610,0x9a6,0x8602,0x8a6,0x600,0x8a0,
0x60c,0x9ae,0x8610,0x8a2,0x60c,0x984,0x8610,0x9ba,0x8610,0x9be,0x600,0x8ba,0x608,0x9c6,0x60c,0x872,
0x8610,0x9ca,0x8602,0x8b8,0x8610,0x9ce,0x608,0x9de,0x60c,0x8bc,0x610,0x9e2,0x8616,0x9e6,0x8610,0x9ea,
0x8610,0x9f2,0x8610,0x9da,0x8610,0x8ae,0x861e,0x8ec,0x861e,0x8ee,0x8610,0x9b4,0x8610,0x9b6,0x8610,0x9d4,
0x8610,0x9d6,0xca6,0xc44,0xca8,0xc46,0x8caa,0xc4a,0x8ca8,0xc48,0x8ca8,0xc4c,0x8ca8,0xd84,0x8ca8,0xda6,
0x8ca8,0xd80,0x9278,0x1252,0x9278,0x1262,0x9278,0x1268,0x137c,0x1396,0x93ae,0x1398,0x167c,0x1696,0x16ac,0x1690,
0x96ae,0x1698,0x97ae,0x1728,0x177c,0x1794,0x97ae,0x1798,0x977c,0x1796,0x98ac,0x1890,0x99aa,0x1980,0x1984,0x1995,
0x19aa,0x198e,0x99ac,0x1990,0x1a7c,0x1a94,0x9aae,0x1a98,0x9a7c,0x1a96,0x1b94,0x1bb4,0x1b9e,0x1bb9,0x9bbe,0x1bbc,
0xa05c,0x204c,0xb66a,0x360c,0xb66a,0x3610,0xb66a,0x3614,0xb66a,0x3618,0xb66a,0x361c,0xb66a,0x3624,0xb66a,0x3676,
0xb66a,0x367a,0xb66a,0x3680,0xb66a,0x3682,0xb66a,0x3686,0x600,0x3f9a,0x602,0x3f9c,0x8684,0x3f9e,0x600,0x3fba,
0x602,0x3fbc,0x8684,0x3fbe,0x8670,0x4334,0x8670,0x4336,0x8670,0x435c,0x8670,0x439a,0x8670,0x439e,0x8670,0x439c,
0x8670,0x4408,0x8670,0x4412,0x8670,0x4418,0x8670,0x4448,0x8670,0x444c,0x8670,0x4482,0x8670,0x4488,0x8670,0x448e,
0x8670,0x4492,0x8670,0x44da,0x8670,0x44c4,0x8670,0x44e0,0x8670,0x44e2,0x8670,0x44e8,0x8670,0x44ea,0x8670,0x44f0,
0x8670,0x44f2,0x8670,0x4500,0x8670,0x4502,0x8670,0x45c0,0x8670,0x45c2,0x8670,0x4508,0x8670,0x450a,0x8670,0x4510,
0x8670,0x4512,0x8670,0x45c4,0x8670,0x45c6,0x8670,0x4558,0x8670,0x455a,0x8670,0x455c,0x8670,0x455e,0x8670,0x45d4,
0x8670,0x45d6,0x8670,0x45d8,0x8670,0x45da,0xe132,0x6128,0xe132,0x6098,0xe132,0x609c,0xe132,0x60a0,0xe132,0x60a4,
0xe132,0x60a8,0xe132,0x60ac,0xe132,0x60b0,0xe132,0x60b4,0xe132,0x60b8,0xe132,0x60bc,0xe132,0x60c0,0xe132,0x60c4,
0xe132,0x60ca,0xe132,0x60ce,0xe132,0x60d2,0x6132,0x60e0,0xe134,0x60e2,0x6132,0x60e6,0xe134,0x60e8,0x6132,0x60ec,
0xe134,0x60ee,0x6132,0x60f2,0xe134,0x60f4,0x6132,0x60f8,0xe134,0x60fa,0xe132,0x613c,0xe132,0x61e8,0xe132,0x6158,
0xe132,0x615c,0xe132,0x6160,0xe132,0x6164,0xe132,0x6168,0xe132,0x616c,0xe132,0x6170,0xe132,0x6174,0xe132,0x6178,
0xe132,0x617c,0xe132,0x6180,0xe132,0x6184,0xe132,0x618a,0xe132,0x618e,0xe132,0x6192,0x6132,0x61a0,0xe134,0x61a2,
0x6132,0x61a6,0xe134,0x61a8,0x6132,0x61ac,0xe134,0x61ae,0x6132,0x61b2,0xe134,0x61b4,0x6132,0x61b8,0xe134,0x61ba,
0xe132,0x61ee,0xe132,0x61f0,0xe132,0x61f2,0xe132,0x61f4,0xe132,0x61fc,0xb489,0x2e82,0x2134,0xb489,0x2e82,0x2138,
0xb489,0x2e82,0x2156,0xb489,0x49c2,0x225c,0xb489,0x49c2,0x225e,0x3489,0xcf82,0x2696,0xb489,0xd5c2,0x2698,0x348b,
0x2c02,0x2978,0x348b,0x2e82,0x2976,0xb48b,0x2f42,0x297c,0xb48b,0x6bc2,0x2b74,0xb48b,0x6bc2,0x2b76,0xb48d,0x4c02,
0x3270,2,0xe602,0x41,0x302,0x600,0x3d4c,0x602,0x3d48,0x606,0x3d54,0x8612,0x3d50,0xe602,0x41,0x308,
0x8608,0x3bc,0xe602,0x41,0x30a,0x8602,0x3f4,0xca02,0x43,0x327,0x8602,0x3c10,0xe602,0x45,0x302,0x600,
0x3d80,0x602,0x3d7c,0x606,0x3d88,0x8612,0x3d84,0xe602,0x49,0x308,0x8602,0x3c5c,0xe602,0x4f,0x302,0x600,
0x3da4,0x602,0x3da0,0x606,0x3dac,0x8612,0x3da8,0xe602,0x4f,0x303,0x602,0x3c98,0x608,0x458,0x8610,0x3c9c,
0xe602,0x4f,0x308,0x8608,0x454,0xe602,0x55,0x308,0x600,0x3b6,0x602,0x3ae,0x608,0x3aa,0x8618,0x3b2,
0xe602,0x61,0x302,0x600,0x3d4e,0x602,0x3d4a,0x606,0x3d56,0x8612,0x3d52,0xe602,0x61,0x308,0x8608,0x3be,
0xe602,0x61,0x30a,0x8602,0x3f6,0xca02,0x63,0x327,0x8602,0x3c12,0xe602,0x65,0x302,0x600,0x3d82,0x602,
0x3d7e,0x606,0x3d8a,0x8612,0x3d86,0xe602,0x69,0x308,0x8602,0x3c5e,0xe602,0x6f,0x302,0x600,0x3da6,0x602,
0x3da2,0x606,0x3dae,0x8612,0x3daa,0xe602,0x6f,0x303,0x602,0x3c9a,0x608,0x45a,0x8610,0x3c9e,0xe602,0x6f,
0x308,0x8608,0x456,0xe602,0x75,0x308,0x600,0x3b8,0x602,0x3b0,0x608,0x3ac,0x8618,0x3b4,0xe602,0x41,
0x306,0x600,0x3d60,0x602,0x3d5c,0x606,0x3d68,0x8612,0x3d64,0xe602,0x61,0x306,0x600,0x3d62,0x602,0x3d5e,
0x606,0x3d6a,0x8612,0x3d66,0xe602,0x45,0x304,0x600,0x3c28,0x8602,0x3c2c,0xe602,0x65,0x304,0x600,0x3c2a,
0x8602,0x3c2e,0xe602,0x4f,0x304,0x600,0x3ca0,0x8602,0x3ca4,0xe602,0x6f,0x304,0x600,0x3ca2,0x8602,0x3ca6,
0xe602,0x53,0x301,0x860e,0x3cc8,0xe602,0x73,0x301,0x860e,0x3cca,0xe602,0x53,0x30c,0x860e,0x3ccc,0xe602,
0x73,0x30c,0x860e,0x3cce,0xe602,0x55,0x303,0x8602,0x3cf0,0xe602,0x75,0x303,0x8602,0x3cf2,0xe602,0x55,
0x304,0x8610,0x3cf4,0xe602,0x75,0x304,0x8610,0x3cf6,0xd802,0x4f,0x31b,0x600,0x3db8,0x602,0x3db4,0x606,
0x3dc0,0x612,0x3dbc,0x8646,0x3dc4,0xd802,0x6f,0x31b,0x600,0x3dba,0x602,0x3db6,0x606,0x3dc2,0x612,0x3dbe,
0x8646,0x3dc6,0xd802,0x55,0x31b,0x600,0x3dd4,0x602,0x3dd0,0x606,0x3ddc,0x612,0x3dd8,0x8646,0x3de0,0xd802,
0x75,0x31b,0x600,0x3dd6,0x602,0x3dd2,0x606,0x3dde,0x612,0x3dda,0x8646,0x3de2,0xca02,0x4f,0x328,0x8608,
0x3d8,0xca02,0x6f,0x328,0x8608,0x3da,0xe602,0x41,0x307,0x8608,0x3c0,0xe602,0x61,0x307,0x8608,0x3c2,
0xca02,0x45,0x327,0x860c,0x3c38,0xca02,0x65,0x327,0x860c,0x3c3a,0xe602,0x4f,0x307,0x8608,0x460,0xe602,
0x6f,0x307,0x8608,0x462,0xe602,0x3b1,0x301,0x868a,0x3f68,0xe602,0x3b7,0x301,0x868a,0x3f88,0xe602,0x3b9,
0x308,0x600,0x3fa4,0x602,0x720,0x8684,0x3fae,0xe602,0x3c5,0x308,0x600,0x3fc4,0x602,0x760,0x8684,0x3fce,
0xe602,0x3c9,0x301,0x868a,0x3fe8,2,0xcc6,0xcc2,0x99aa,0x1996,2,0xdd9,0xdcf,0x9b94,0x1bba,0xdc02,
0x4c,0x323,0x8608,0x3c70,0xdc02,0x6c,0x323,0x8608,0x3c72,0xdc02,0x52,0x323,0x8608,0x3cb8,0xdc02,0x72,
0x323,0x8608,0x3cba,0xdc02,0x53,0x323,0x860e,0x3cd0,0xdc02,0x73,0x323,0x860e,0x3cd2,0xdc02,0x41,0x323,
0x604,0x3d58,0x860c,0x3d6c,0xdc02,0x61,0x323,0x604,0x3d5a,0x860c,0x3d6e,0xdc02,0x45,0x323,0x8604,0x3d8c,
0xdc02,0x65,0x323,0x8604,0x3d8e,0xdc02,0x4f,0x323,0x8604,0x3db0,0xdc02,0x6f,0x323,0x8604,0x3db2,0xe602,
0x3b1,0x313,0x600,0x3e05,0x602,0x3e09,0x684,0x3e0d,0x868a,0x3f00,0xe602,0x3b1,0x314,0x600,0x3e07,0x602,
0x3e0b,0x684,0x3e0f,0x868a,0x3f02,0x1f00,0xe643,0x3b1,0x313,0x300,0x868a,0x3f04,0x1f01,0xe643,0x3b1,0x314,
0x300,0x868a,0x3f06,0x1f00,0xe643,0x3b1,0x313,0x301,0x868a,0x3f08,0x1f01,0xe643,0x3b1,0x314,0x301,0x868a,
0x3f0a,0x1f00,0xe643,0x3b1,0x313,0x342,0x868a,0x3f0c,0x1f01,0xe643,0x3b1,0x314,0x342,0x868a,0x3f0e,0xe602,
0x391,0x313,0x600,0x3e15,0x602,0x3e19,0x684,0x3e1d,0x868a,0x3f10,0xe602,0x391,0x314,0x600,0x3e17,0x602,
0x3e1b,0x684,0x3e1f,0x868a,0x3f12,0x1f08,0xe643,0x391,0x313,0x300,0x868a,0x3f14,0x1f09,0xe643,0x391,0x314,
0x300,0x868a,0x3f16,0x1f08,0xe643,0x391,0x313,0x301,0x868a,0x3f18,0x1f09,0xe643,0x391,0x314,0x301,0x868a,
0x3f1a,0x1f08,0xe643,0x391,0x313,0x342,0x868a,0x3f1c,0x1f09,0xe643,0x391,0x314,0x342,0x868a,0x3f1e,0xe602,
0x3b5,0x313,0x600,0x3e24,0x8602,0x3e28,0xe602,0x3b5,0x314,0x600,0x3e26,0x8602,0x3e2a,0xe602,0x395,0x313,
0x600,0x3e34,0x8602,0x3e38,0xe602,0x395,0x314,0x600,0x3e36,0x8602,0x3e3a,0xe602,0x3b7,0x313,0x600,0x3e45,
0x602,0x3e49,0x684,0x3e4d,0x868a,0x3f20,0xe602,0x3b7,0x314,0x600,0x3e47,0x602,0x3e4b,0x684,0x3e4f,0x868a,
0x3f22,0x1f20,0xe643,0x3b7,0x313,0x300,0x868a,0x3f24,0x1f21,0xe643,0x3b7,0x314,0x300,0x868a,0x3f26,0x1f20,
0xe643,0x3b7,0x313,0x301,0x868a,0x3f28,0x1f21,0xe643,0x3b7,0x314,0x301,0x868a,0x3f2a,0x1f20,0xe643,0x3b7,
0x313,0x342,0x868a,0x3f2c,0x1f21,0xe643,0x3b7,0x314,0x342,0x868a,0x3f2e,0xe602,0x397,0x313,0x600,0x3e55,
0x602,0x3e59,0x684,0x3e5d,0x868a,0x3f30,0xe602,0x397,0x314,0x600,0x3e57,0x602,0x3e5b,0x684,0x3e5f,0x868a,
0x3f32,0x1f28,0xe643,0x397,0x313,0x300,0x868a,0x3f34,0x1f29,0xe643,0x397,0x314,0x300,0x868a,0x3f36,0x1f28,
0xe643,0x397,0x313,0x301,0x868a,0x3f38,0x1f29,0xe643,0x397,0x314,0x301,0x868a,0x3f3a,0x1f28,0xe643,0x397,
0x313,0x342,0x868a,0x3f3c,0x1f29,0xe643,0x397,0x314,0x342,0x868a,0x3f3e,0xe602,0x3b9,0x313,0x600,0x3e64,
0x602,0x3e68,0x8684,0x3e6c,0xe602,0x3b9,0x314,0x600,0x3e66,0x602,0x3e6a,0x8684,0x3e6e,0xe602,0x399,0x313,
0x600,0x3e74,0x602,0x3e78,0x8684,0x3e7c,0xe602,0x399,0x314,0x600,0x3e76,0x602,0x3e7a,0x8684,0x3e7e,0xe602,
0x3bf,0x313,0x600,0x3e84,0x8602,0x3e88,0xe602,0x3bf,0x314,0x600,0x3e86,0x8602,0x3e8a,0xe602,0x39f,0x313,
0x600,0x3e94,0x8602,0x3e98,0xe602,0x39f,0x314,0x600,0x3e96,0x8602,0x3e9a,0xe602,0x3c5,0x313,0x600,0x3ea4,
0x602,0x3ea8,0x8684,0x3eac,0xe602,0x3c5,0x314,0x600,0x3ea6,0x602,0x3eaa,0x8684,0x3eae,0xe602,0x3a5,0x314,
0x600,0x3eb6,0x602,0x3eba,0x8684,0x3ebe,0xe602,0x3c9,0x313,0x600,0x3ec5,0x602,0x3ec9,0x684,0x3ecd,0x868a,
0x3f40,0xe602,0x3c9,0x314,0x600,0x3ec7,0x602,0x3ecb,0x684,0x3ecf,0x868a,0x3f42,0x1f60,0xe643,0x3c9,0x313,
0x300,0x868a,0x3f44,0x1f61,0xe643,0x3c9,0x314,0x300,0x868a,0x3f46,0x1f60,0xe643,0x3c9,0x313,0x301,0x868a,
0x3f48,0x1f61,0xe643,0x3c9,0x314,0x301,0x868a,0x3f4a,0x1f60,0xe643,0x3c9,0x313,0x342,0x868a,0x3f4c,0x1f61,
0xe643,0x3c9,0x314,0x342,0x868a,0x3f4e,0xe602,0x3a9,0x313,0x600,0x3ed5,0x602,0x3ed9,0x684,0x3edd,0x868a,
0x3f50,0xe602,0x3a9,0x314,0x600,0x3ed7,0x602,0x3edb,0x684,0x3edf,0x868a,0x3f52,0x1f68,0xe643,0x3a9,0x313,
0x300,0x868a,0x3f54,0x1f69,0xe643,0x3a9,0x314,0x300,0x868a,0x3f56,0x1f68,0xe643,0x3a9,0x313,0x301,0x868a,
0x3f58,0x1f69,0xe643,0x3a9,0x314,0x301,0x868a,0x3f5a,0x1f68,0xe643,0x3a9,0x313,0x342,0x868a,0x3f5c,0x1f69,
0xe643,0x3a9,0x314,0x342,0x868a,0x3f5e,0xe602,0x3b1,0x300,0x868a,0x3f64,0xe602,0x3b7,0x300,0x868a,0x3f84,
0xe602,0x3c9,0x300,0x868a,0x3fe4,0xe602,0x3b1,0x342,0x868a,0x3f6e,0xe602,0x3b7,0x342,0x868a,0x3f8e,0xe602,
0x3c9,0x342,0x868a,0x3fee,3,0xe602,0x41,0x300,0xe602,0x41,0x301,0xe602,0x41,0x303,0xe602,0x45,
0x300,0xe602,0x45,0x301,0xe602,0x45,0x308,0xe602,0x49,0x300,0xe602,0x49,0x301,0xe602,0x49,0x302,
0xe602,0x4e,0x303,0xe602,0x4f,0x300,0xe602,0x4f,0x301,0xe602,0x55,0x300,0xe602,0x55,0x301,0xe602,
0x55,0x302,0xe602,0x59,0x301,0xe602,0x61,0x300,0xe602,0x61,0x301,0xe602,0x61,0x303,0xe602,0x65,
0x300,0xe602,0x65,0x301,0xe602,0x65,0x308,0xe602,0x69,0x300,0xe602,0x69,0x301,0xe602,0x69,0x302,
0xe602,0x6e,0x303,0xe602,0x6f,0x300,0xe602,0x6f,0x301,0xe602,0x75,0x300,0xe602,0x75,0x301,0xe602,
0x75,0x302,0xe602,0x79,0x301,0xe602,0x79,0x308,0xe602,0x41,0x304,0xe602,0x61,0x304,0xca02,0x41,
0x328,0xca02,0x61,0x328,0xe602,0x43,0x301,0xe602,0x63,0x301,0xe602,0x43,0x302,0xe602,0x63,0x302,
0xe602,0x43,0x307,0xe602,0x63,0x307,0xe602,0x43,0x30c,0xe602,0x63,0x30c,0xe602,0x44,0x30c,0xe602,
0x64,0x30c,0xe602,0x45,0x306,0xe602,0x65,0x306,0xe602,0x45,0x307,0xe602,0x65,0x307,0xca02,0x45,
0x328,0xca02,0x65,0x328,0xe602,0x45,0x30c,0xe602,0x65,0x30c,0xe602,0x47,0x302,0xe602,0x67,0x302,
0xe602,0x47,0x306,0xe602,0x67,0x306,0xe602,0x47,0x307,0xe602,0x67,0x307,0xca02,0x47,0x327,0xca02,
0x67,0x327,0xe602,0x48,0x302,0xe602,0x68,0x302,0xe602,0x49,0x303,0xe602,0x69,0x303,0xe602,0x49,
0x304,0xe602,0x69,0x304,0xe602,0x49,0x306,0xe602,0x69,0x306,0xca02,0x49,0x328,0xca02,0x69,0x328,
0xe602,0x49,0x307,0xe602,0x4a,0x302,0xe602,0x6a,0x302,0xca02,0x4b,0x327,0xca02,0x6b,0x327,0xe602,
0x4c,0x301,0xe602,0x6c,0x301,0xca02,0x4c,0x327,0xca02,0x6c,0x327,0xe602,0x4c,0x30c,0xe602,0x6c,
0x30c,0xe602,0x4e,0x301,0xe602,0x6e,0x301,0xca02,0x4e,0x327,0xca02,0x6e,0x327,0xe602,0x4e,0x30c,
0xe602,0x6e,0x30c,0xe602,0x4f,0x306,0xe602,0x6f,0x306,0xe602,0x4f,0x30b,0xe602,0x6f,0x30b,0xe602,
0x52,0x301,0xe602,0x72,0x301,0xca02,0x52,0x327,0xca02,0x72,0x327,0xe602,0x52,0x30c,0xe602,0x72,
0x30c,0xe602,0x53,0x302,0xe602,0x73,0x302,0xca02,0x53,0x327,0xca02,0x73,0x327,0xca02,0x54,0x327,
0xca02,0x74,0x327,0xe602,0x54,0x30c,0xe602,0x74,0x30c,0xe602,0x55,0x306,0xe602,0x75,0x306,0xe602,
0x55,0x30a,0xe602,0x75,0x30a,0xe602,0x55,0x30b,0xe602,0x75,0x30b,0xca02,0x55,0x328,0xca02,0x75,
0x328,0xe602,0x57,0x302,0xe602,0x77,0x302,0xe602,0x59,0x302,0xe602,0x79,0x302,0xe602,0x59,0x308,
0xe602,0x5a,0x301,0xe602,0x7a,0x301,0xe602,0x5a,0x307,0xe602,0x7a,0x307,0xe602,0x5a,0x30c,0xe602,
0x7a,0x30c,0xe602,0x41,0x30c,0xe602,0x61,0x30c,0xe602,0x49,0x30c,0xe602,0x69,0x30c,0xe602,0x4f,
0x30c,0xe602,0x6f,0x30c,0xe602,0x55,0x30c,0xe602,0x75,0x30c,0xdc,0xe643,0x55,0x308,0x304,0xfc,
0xe643,0x75,0x308,0x304,0xdc,0xe643,0x55,0x308,0x301,0xfc,0xe643,0x75,0x308,0x301,0xdc,0xe643,
0x55,0x308,0x30c,0xfc,0xe643,0x75,0x308,0x30c,0xdc,0xe643,0x55,0x308,0x300,0xfc,0xe643,0x75,
0x308,0x300,0xc4,0xe643,0x41,0x308,0x304,0xe4,0xe643,0x61,0x308,0x304,0x226,0xe643,0x41,0x307,
0x304,0x227,0xe643,0x61,0x307,0x304,0xe602,0xc6,0x304,0xe602,0xe6,0x304,0xe602,0x47,0x30c,0xe602,
0x67,0x30c,0xe602,0x4b,0x30c,0xe602,0x6b,0x30c,0x1ea,0xe643,0x4f,0x328,0x304,0x1eb,0xe643,0x6f,
0x328,0x304,0xe602,0x1b7,0x30c,0xe602,0x292,0x30c,0xe602,0x6a,0x30c,0xe602,0x47,0x301,0xe602,0x67,
0x301,0xe602,0x4e,0x300,0xe602,0x6e,0x300,0xc5,0xe643,0x41,0x30a,0x301,0xe5,0xe643,0x61,0x30a,
0x301,0xe602,0xc6,0x301,0xe602,0xe6,0x301,0xe602,0xd8,0x301,0xe602,0xf8,0x301,0xe602,0x41,0x30f,
0xe602,0x61,0x30f,0xe602,0x41,0x311,0xe602,0x61,0x311,0xe602,0x45,0x30f,0xe602,0x65,0x30f,0xe602,
0x45,0x311,0xe602,0x65,0x311,0xe602,0x49,0x30f,0xe602,0x69,0x30f,0xe602,0x49,0x311,0xe602,0x69,
0x311,0xe602,0x4f,0x30f,0xe602,0x6f,0x30f,0xe602,0x4f,0x311,0xe602,0x6f,0x311,0xe602,0x52,0x30f,
0xe602,0x72,0x30f,0xe602,0x52,0x311,0xe602,0x72,0x311,0xe602,0x55,0x30f,0xe602,0x75,0x30f,0xe602,
0x55,0x311,0xe602,0x75,0x311,0xdc02,0x53,0x326,0xdc02,0x73,0x326,0xdc02,0x54,0x326,0xdc02,0x74,
0x326,0xe602,0x48,0x30c,0xe602,0x68,0x30c,0xd6,0xe643,0x4f,0x308,0x304,0xf6,0xe643,0x6f,0x308,
0x304,0xd5,0xe643,0x4f,0x303,0x304,0xf5,0xe643,0x6f,0x303,0x304,0x22e,0xe643,0x4f,0x307,0x304,
0x22f,0xe643,0x6f,0x307,0x304,0xe602,0x59,0x304,0xe602,0x79,0x304,0xe602,0xa8,0x301,0xe602,0x391,
0x301,0xe602,0x395,0x301,0xe602,0x397,0x301,0xe602,0x399,0x301,0xe602,0x39f,0x301,0xe602,0x3a5,0x301,
0xe602,0x3a9,0x301,0x3ca,0xe643,0x3b9,0x308,0x301,0xe602,0x399,0x308,0xe602,0x3a5,0x308,0xe602,0x3b5,
0x301,0xe602,0x3b9,0x301,0x3cb,0xe643,0x3c5,0x308,0x301,0xe602,0x3bf,0x301,0xe602,0x3c5,0x301,0xe602,
0x3d2,0x301,0xe602,0x3d2,0x308,0xe602,0x415,0x300,0xe602,0x415,0x308,0xe602,0x413,0x301,0xe602,0x406,
0x308,0xe602,0x41a,0x301,0xe602,0x418,0x300,0xe602,0x423,0x306,0xe602,0x418,0x306,0xe602,0x438,0x306,
0xe602,0x435,0x300,0xe602,0x435,0x308,0xe602,0x433,0x301,0xe602,0x456,0x308,0xe602,0x43a,0x301,0xe602,
0x438,0x300,0xe602,0x443,0x306,0xe602,0x474,0x30f,0xe602,0x475,0x30f,0xe602,0x416,0x306,0xe602,0x436,
0x306,0xe602,0x410,0x306,0xe602,0x430,0x306,0xe602,0x410,0x308,0xe602,0x430,0x308,0xe602,0x415,0x306,
0xe602,0x435,0x306,0xe602,0x4d8,0x308,0xe602,0x4d9,0x308,0xe602,0x416,0x308,0xe602,0x436,0x308,0xe602,
0x417,0x308,0xe602,0x437,0x308,0xe602,0x418,0x304,0xe602,0x438,0x304,0xe602,0x418,0x308,0xe602,0x438,
0x308,0xe602,0x41e,0x308,0xe602,0x43e,0x308,0xe602,0x4e8,0x308,0xe602,0x4e9,0x308,0xe602,0x42d,0x308,
0xe602,0x44d,0x308,0xe602,0x423,0x304,0xe602,0x443,0x304,0xe602,0x423,0x308,0xe602,0x443,0x308,0xe602,
0x423,0x30b,0xe602,0x443,0x30b,0xe602,0x427,0x308,0xe602,0x447,0x308,0xe602,0x42b,0x308,0xe602,0x44b,
0x308,0xe602,0x627,0x653,0xe602,0x627,0x654,0xe602,0x648,0x654,0xdc02,0x627,0x655,0xe602,0x64a,0x654,
0xe602,0x6d5,0x654,0xe602,0x6c1,0x654,0xe602,0x6d2,0x654,0x702,0x928,0x93c,0x702,0x930,0x93c,0x702,
0x933,0x93c,2,0x9c7,0x9be,2,0x9c7,0x9d7,2,0xb47,0xb56,2,0xb47,0xb3e,2,0xb47,
0xb57,2,0xb92,0xbd7,2,0xbc6,0xbbe,2,0xbc7,0xbbe,2,0xbc6,0xbd7,0x5b02,0xc46,0xc56,
2,0xcbf,0xcd5,2,0xcc6,0xcd5,2,0xcc6,0xcd6,0xcca,0x43,0xcc6,0xcc2,0xcd5,2,0xd46,
0xd3e,2,0xd47,0xd3e,2,0xd46,0xd57,0x902,0xdd9,0xdca,0xddc,0x943,0xdd9,0xdcf,0xdca,2,
0xdd9,0xddf,2,0x1025,0x102e,2,0x1b05,0x1b35,2,0x1b07,0x1b35,2,0x1b09,0x1b35,2,0x1b0b,
0x1b35,2,0x1b0d,0x1b35,2,0x1b11,0x1b35,2,0x1b3a,0x1b35,2,0x1b3c,0x1b35,2,0x1b3e,0x1b35,
2,0x1b3f,0x1b35,2,0x1b42,0x1b35,0xdc02,0x41,0x325,0xdc02,0x61,0x325,0xe602,0x42,0x307,0xe602,
0x62,0x307,0xdc02,0x42,0x323,0xdc02,0x62,0x323,0xdc02,0x42,0x331,0xdc02,0x62,0x331,0xc7,0xe643,
0x43,0x327,0x301,0xe7,0xe643,0x63,0x327,0x301,0xe602,0x44,0x307,0xe602,0x64,0x307,0xdc02,0x44,
0x323,0xdc02,0x64,0x323,0xdc02,0x44,0x331,0xdc02,0x64,0x331,0xca02,0x44,0x327,0xca02,0x64,0x327,
0xdc02,0x44,0x32d,0xdc02,0x64,0x32d,0x112,0xe643,0x45,0x304,0x300,0x113,0xe643,0x65,0x304,0x300,
0x112,0xe643,0x45,0x304,0x301,0x113,0xe643,0x65,0x304,0x301,0xdc02,0x45,0x32d,0xdc02,0x65,0x32d,
0xdc02,0x45,0x330,0xdc02,0x65,0x330,0x228,0xe643,0x45,0x327,0x306,0x229,0xe643,0x65,0x327,0x306,
0xe602,0x46,0x307,0xe602,0x66,0x307,0xe602,0x47,0x304,0xe602,0x67,0x304,0xe602,0x48,0x307,0xe602,
0x68,0x307,0xdc02,0x48,0x323,0xdc02,0x68,0x323,0xe602,0x48,0x308,0xe602,0x68,0x308,0xca02,0x48,
0x327,0xca02,0x68,0x327,0xdc02,0x48,0x32e,0xdc02,0x68,0x32e,0xdc02,0x49,0x330,0xdc02,0x69,0x330,
0xcf,0xe643,0x49,0x308,0x301,0xef,0xe643,0x69,0x308,0x301,0xe602,0x4b,0x301,0xe602,0x6b,0x301,
0xdc02,0x4b,0x323,0xdc02,0x6b,0x323,0xdc02,0x4b,0x331,0xdc02,0x6b,0x331,0x1e36,0xe643,0x4c,0x323,
0x304,0x1e37,0xe643,0x6c,0x323,0x304,0xdc02,0x4c,0x331,0xdc02,0x6c,0x331,0xdc02,0x4c,0x32d,0xdc02,
0x6c,0x32d,0xe602,0x4d,0x301,0xe602,0x6d,0x301,0xe602,0x4d,0x307,0xe602,0x6d,0x307,0xdc02,0x4d,
0x323,0xdc02,0x6d,0x323,0xe602,0x4e,0x307,0xe602,0x6e,0x307,0xdc02,0x4e,0x323,0xdc02,0x6e,0x323,
0xdc02,0x4e,0x331,0xdc02,0x6e,0x331,0xdc02,0x4e,0x32d,0xdc02,0x6e,0x32d,0xd5,0xe643,0x4f,0x303,
0x301,0xf5,0xe643,0x6f,0x303,0x301,0xd5,0xe643,0x4f,0x303,0x308,0xf5,0xe643,0x6f,0x303,0x308,
0x14c,0xe643,0x4f,0x304,0x300,0x14d,0xe643,0x6f,0x304,0x300,0x14c,0xe643,0x4f,0x304,0x301,0x14d,
0xe643,0x6f,0x304,0x301,0xe602,0x50,0x301,0xe602,0x70,0x301,0xe602,0x50,0x307,0xe602,0x70,0x307,
0xe602,0x52,0x307,0xe602,0x72,0x307,0x1e5a,0xe643,0x52,0x323,0x304,0x1e5b,0xe643,0x72,0x323,0x304,
0xdc02,0x52,0x331,0xdc02,0x72,0x331,0xe602,0x53,0x307,0xe602,0x73,0x307,0x15a,0xe643,0x53,0x301,
0x307,0x15b,0xe643,0x73,0x301,0x307,0x160,0xe643,0x53,0x30c,0x307,0x161,0xe643,0x73,0x30c,0x307,
0x1e62,0xe643,0x53,0x323,0x307,0x1e63,0xe643,0x73,0x323,0x307,0xe602,0x54,0x307,0xe602,0x74,0x307,
0xdc02,0x54,0x323,0xdc02,0x74,0x323,0xdc02,0x54,0x331,0xdc02,0x74,0x331,0xdc02,0x54,0x32d,0xdc02,
0x74,0x32d,0xdc02,0x55,0x324,0xdc02,0x75,0x324,0xdc02,0x55,0x330,0xdc02,0x75,0x330,0xdc02,0x55,
0x32d,0xdc02,0x75,0x32d,0x168,0xe643,0x55,0x303,0x301,0x169,0xe643,0x75,0x303,0x301,0x16a,0xe643,
0x55,0x304,0x308,0x16b,0xe643,0x75,0x304,0x308,0xe602,0x56,0x303,0xe602,0x76,0x303,0xdc02,0x56,
0x323,0xdc02,0x76,0x323,0xe602,0x57,0x300,0xe602,0x77,0x300,0xe602,0x57,0x301,0xe602,0x77,0x301,
0xe602,0x57,0x308,0xe602,0x77,0x308,0xe602,0x57,0x307,0xe602,0x77,0x307,0xdc02,0x57,0x323,0xdc02,
0x77,0x323,0xe602,0x58,0x307,0xe602,0x78,0x307,0xe602,0x58,0x308,0xe602,0x78,0x308,0xe602,0x59,
0x307,0xe602,0x79,0x307,0xe602,0x5a,0x302,0xe602,0x7a,0x302,0xdc02,0x5a,0x323,0xdc02,0x7a,0x323,
0xdc02,0x5a,0x331,0xdc02,0x7a,0x331,0xdc02,0x68,0x331,0xe602,0x74,0x308,0xe602,0x77,0x30a,0xe602,
0x79,0x30a,0xe602,0x17f,0x307,0xe602,0x41,0x309,0xe602,0x61,0x309,0xc2,0xe643,0x41,0x302,0x301,
0xe2,0xe643,0x61,0x302,0x301,0xc2,0xe643,0x41,0x302,0x300,0xe2,0xe643,0x61,0x302,0x300,0xc2,
0xe643,0x41,0x302,0x309,0xe2,0xe643,0x61,0x302,0x309,0xc2,0xe643,0x41,0x302,0x303,0xe2,0xe643,
0x61,0x302,0x303,0x1ea0,0xe643,0x41,0x323,0x302,0x1ea1,0xe643,0x61,0x323,0x302,0x102,0xe643,0x41,
0x306,0x301,0x103,0xe643,0x61,0x306,0x301,0x102,0xe643,0x41,0x306,0x300,0x103,0xe643,0x61,0x306,
0x300,0x102,0xe643,0x41,0x306,0x309,0x103,0xe643,0x61,0x306,0x309,0x102,0xe643,0x41,0x306,0x303,
0x103,0xe643,0x61,0x306,0x303,0x1ea0,0xe643,0x41,0x323,0x306,0x1ea1,0xe643,0x61,0x323,0x306,0xe602,
0x45,0x309,0xe602,0x65,0x309,0xe602,0x45,0x303,0xe602,0x65,0x303,0xca,0xe643,0x45,0x302,0x301,
0xea,0xe643,0x65,0x302,0x301,0xca,0xe643,0x45,0x302,0x300,0xea,0xe643,0x65,0x302,0x300,0xca,
0xe643,0x45,0x302,0x309,0xea,0xe643,0x65,0x302,0x309,0xca,0xe643,0x45,0x302,0x303,0xea,0xe643,
0x65,0x302,0x303,0x1eb8,0xe643,0x45,0x323,0x302,0x1eb9,0xe643,0x65,0x323,0x302,0xe602,0x49,0x309,
0xe602,0x69,0x309,0xdc02,0x49,0x323,0xdc02,0x69,0x323,0xe602,0x4f,0x309,0xe602,0x6f,0x309,0xd4,
0xe643,0x4f,0x302,0x301,0xf4,0xe643,0x6f,0x302,0x301,0xd4,0xe643,0x4f,0x302,0x300,0xf4,0xe643,
0x6f,0x302,0x300,0xd4,0xe643,0x4f,0x302,0x309,0xf4,0xe643,0x6f,0x302,0x309,0xd4,0xe643,0x4f,
0x302,0x303,0xf4,0xe643,0x6f,0x302,0x303,0x1ecc,0xe643,0x4f,0x323,0x302,0x1ecd,0xe643,0x6f,0x323,
0x302,0x1a0,0xe643,0x4f,0x31b,0x301,0x1a1,0xe643,0x6f,0x31b,0x301,0x1a0,0xe643,0x4f,0x31b,0x300,
0x1a1,0xe643,0x6f,0x31b,0x300,0x1a0,0xe643,0x4f,0x31b,0x309,0x1a1,0xe643,0x6f,0x31b,0x309,0x1a0,
0xe643,0x4f,0x31b,0x303,0x1a1,0xe643,0x6f,0x31b,0x303,0x1a0,0xdc43,0x4f,0x31b,0x323,0x1a1,0xdc43,
0x6f,0x31b,0x323,0xdc02,0x55,0x323,0xdc02,0x75,0x323,0xe602,0x55,0x309,0xe602,0x75,0x309,0x1af,
0xe643,0x55,0x31b,0x301,0x1b0,0xe643,0x75,0x31b,0x301,0x1af,0xe643,0x55,0x31b,0x300,0x1b0,0xe643,
0x75,0x31b,0x300,0x1af,0xe643,0x55,0x31b,0x309,0x1b0,0xe643,0x75,0x31b,0x309,0x1af,0xe643,0x55,
0x31b,0x303,0x1b0,0xe643,0x75,0x31b,0x303,0x1af,0xdc43,0x55,0x31b,0x323,0x1b0,0xdc43,0x75,0x31b,
0x323,0xe602,0x59,0x300,0xe602,0x79,0x300,0xdc02,0x59,0x323,0xdc02,0x79,0x323,0xe602,0x59,0x309,
0xe602,0x79,0x309,0xe602,0x59,0x303,0xe602,0x79,0x303,0x1f10,0xe643,0x3b5,0x313,0x300,0x1f11,0xe643,
0x3b5,0x314,0x300,0x1f10,0xe643,0x3b5,0x313,0x301,0x1f11,0xe643,0x3b5,0x314,0x301,0x1f18,0xe643,0x395,
0x313,0x300,0x1f19,0xe643,0x395,0x314,0x300,0x1f18,0xe643,0x395,0x313,0x301,0x1f19,0xe643,0x395,0x314,
0x301,0x1f30,0xe643,0x3b9,0x313,0x300,0x1f31,0xe643,0x3b9,0x314,0x300,0x1f30,0xe643,0x3b9,0x313,0x301,
0x1f31,0xe643,0x3b9,0x314,0x301,0x1f30,0xe643,0x3b9,0x313,0x342,0x1f31,0xe643,0x3b9,0x314,0x342,0x1f38,
0xe643,0x399,0x313,0x300,0x1f39,0xe643,0x399,0x314,0x300,0x1f38,0xe643,0x399,0x313,0x301,0x1f39,0xe643,
0x399,0x314,0x301,0x1f38,0xe643,0x399,0x313,0x342,0x1f39,0xe643,0x399,0x314,0x342,0x1f40,0xe643,0x3bf,
0x313,0x300,0x1f41,0xe643,0x3bf,0x314,0x300,0x1f40,0xe643,0x3bf,0x313,0x301,0x1f41,0xe643,0x3bf,0x314,
0x301,0x1f48,0xe643,0x39f,0x313,0x300,0x1f49,0xe643,0x39f,0x314,0x300,0x1f48,0xe643,0x39f,0x313,0x301,
0x1f49,0xe643,0x39f,0x314,0x301,0x1f50,0xe643,0x3c5,0x313,0x300,0x1f51,0xe643,0x3c5,0x314,0x300,0x1f50,
0xe643,0x3c5,0x313,0x301,0x1f51,0xe643,0x3c5,0x314,0x301,0x1f50,0xe643,0x3c5,0x313,0x342,0x1f51,0xe643,
0x3c5,0x314,0x342,0x1f59,0xe643,0x3a5,0x314,0x300,0x1f59,0xe643,0x3a5,0x314,0x301,0x1f59,0xe643,0x3a5,
0x314,0x342,0xe602,0x3b5,0x300,0xe602,0x3b9,0x300,0xe602,0x3bf,0x300,0xe602,0x3c5,0x300,0x1f00,0xf043,
0x3b1,0x313,0x345,0x1f01,0xf043,0x3b1,0x314,0x345,0x1f02,0x345,2,0xf044,0x3b1,0x313,0x300,0x345,
0x1f03,0x345,2,0xf044,0x3b1,0x314,0x300,0x345,0x1f04,0x345,2,0xf044,0x3b1,0x313,0x301,0x345,
0x1f05,0x345,2,0xf044,0x3b1,0x314,0x301,0x345,0x1f06,0x345,2,0xf044,0x3b1,0x313,0x342,0x345,
0x1f07,0x345,2,0xf044,0x3b1,0x314,0x342,0x345,0x1f08,0xf043,0x391,0x313,0x345,0x1f09,0xf043,0x391,
0x314,0x345,0x1f0a,0x345,2,0xf044,0x391,0x313,0x300,0x345,0x1f0b,0x345,2,0xf044,0x391,0x314,
0x300,0x345,0x1f0c,0x345,2,0xf044,0x391,0x313,0x301,0x345,0x1f0d,0x345,2,0xf044,0x391,0x314,
0x301,0x345,0x1f0e,0x345,2,0xf044,0x391,0x313,0x342,0x345,0x1f0f,0x345,2,0xf044,0x391,0x314,
0x342,0x345,0x1f20,0xf043,0x3b7,0x313,0x345,0x1f21,0xf043,0x3b7,0x314,0x345,0x1f22,0x345,2,0xf044,
0x3b7,0x313,0x300,0x345,0x1f23,0x345,2,0xf044,0x3b7,0x314,0x300,0x345,0x1f24,0x345,2,0xf044,
0x3b7,0x313,0x301,0x345,0x1f25,0x345,2,0xf044,0x3b7,0x314,0x301,0x345,0x1f26,0x345,2,0xf044,
0x3b7,0x313,0x342,0x345,0x1f27,0x345,2,0xf044,0x3b7,0x314,0x342,0x345,0x1f28,0xf043,0x397,0x313,
0x345,0x1f29,0xf043,0x397,0x314,0x345,0x1f2a,0x345,2,0xf044,0x397,0x313,0x300,0x345,0x1f2b,0x345,
2,0xf044,0x397,0x314,0x300,0x345,0x1f2c,0x345,2,0xf044,0x397,0x313,0x301,0x345,0x1f2d,0x345,
2,0xf044,0x397,0x314,0x301,0x345,0x1f2e,0x345,2,0xf044,0x397,0x313,0x342,0x345,0x1f2f,0x345,
2,0xf044,0x397,0x314,0x342,0x345,0x1f60,0xf043,0x3c9,0x313,0x345,0x1f61,0xf043,0x3c9,0x314,0x345,
0x1f62,0x345,2,0xf044,0x3c9,0x313,0x300,0x345,0x1f63,0x345,2,0xf044,0x3c9,0x314,0x300,0x345,
0x1f64,0x345,2,0xf044,0x3c9,0x313,0x301,0x345,0x1f65,0x345,2,0xf044,0x3c9,0x314,0x301,0x345,
0x1f66,0x345,2,0xf044,0x3c9,0x313,0x342,0x345,0x1f67,0x345,2,0xf044,0x3c9,0x314,0x342,0x345,
0x1f68,0xf043,0x3a9,0x313,0x345,0x1f69,0xf043,0x3a9,0x314,0x345,0x1f6a,0x345,2,0xf044,0x3a9,0x313,
0x300,0x345,0x1f6b,0x345,2,0xf044,0x3a9,0x314,0x300,0x345,0x1f6c,0x345,2,0xf044,0x3a9,0x313,
0x301,0x345,0x1f6d,0x345,2,0xf044,0x3a9,0x314,0x301,0x345,0x1f6e,0x345,2,0xf044,0x3a9,0x313,
0x342,0x345,0x1f6f,0x345,2,0xf044,0x3a9,0x314,0x342,0x345,0xe602,0x3b1,0x306,0xe602,0x3b1,0x304,
0x1f70,0xf043,0x3b1,0x300,0x345,0xf002,0x3b1,0x345,0x3ac,0xf043,0x3b1,0x301,0x345,0x1fb6,0xf043,0x3b1,
0x342,0x345,0xe602,0x391,0x306,0xe602,0x391,0x304,0xe602,0x391,0x300,0xf002,0x391,0x345,0xe602,0xa8,
0x342,0x1f74,0xf043,0x3b7,0x300,0x345,0xf002,0x3b7,0x345,0x3ae,0xf043,0x3b7,0x301,0x345,0x1fc6,0xf043,
0x3b7,0x342,0x345,0xe602,0x395,0x300,0xe602,0x397,0x300,0xf002,0x397,0x345,0xe602,0x1fbf,0x300,0xe602,
0x1fbf,0x301,0xe602,0x1fbf,0x342,0xe602,0x3b9,0x306,0xe602,0x3b9,0x304,0x3ca,0xe643,0x3b9,0x308,0x300,
0xe602,0x3b9,0x342,0x3ca,0xe643,0x3b9,0x308,0x342,0xe602,0x399,0x306,0xe602,0x399,0x304,0xe602,0x399,
0x300,0xe602,0x1ffe,0x300,0xe602,0x1ffe,0x301,0xe602,0x1ffe,0x342,0xe602,0x3c5,0x306,0xe602,0x3c5,0x304,
0x3cb,0xe643,0x3c5,0x308,0x300,0xe602,0x3c1,0x313,0xe602,0x3c1,0x314,0xe602,0x3c5,0x342,0x3cb,0xe643,
0x3c5,0x308,0x342,0xe602,0x3a5,0x306,0xe602,0x3a5,0x304,0xe602,0x3a5,0x300,0xe602,0x3a1,0x314,0xe602,
0xa8,0x300,0x1f7c,0xf043,0x3c9,0x300,0x345,0xf002,0x3c9,0x345,0x3ce,0xf043,0x3c9,0x301,0x345,0x1ff6,
0xf043,0x3c9,0x342,0x345,0xe602,0x39f,0x300,0xe602,0x3a9,0x300,0xf002,0x3a9,0x345,0x102,0x2190,0x338,
0x102,0x2192,0x338,0x102,0x2194,0x338,0x102,0x21d0,0x338,0x102,0x21d4,0x338,0x102,0x21d2,0x338,0x102,
0x2203,0x338,0x102,0x2208,0x338,0x102,0x220b,0x338,0x102,0x2223,0x338,0x102,0x2225,0x338,0x102,0x223c,
0x338,0x102,0x2243,0x338,0x102,0x2245,0x338,0x102,0x2248,0x338,0x102,0x3d,0x338,0x102,0x2261,0x338,
0x102,0x224d,0x338,0x102,0x3c,0x338,0x102,0x3e,0x338,0x102,0x2264,0x338,0x102,0x2265,0x338,0x102,
0x2272,0x338,0x102,0x2273,0x338,0x102,0x2276,0x338,0x102,0x2277,0x338,0x102,0x227a,0x338,0x102,0x227b,
0x338,0x102,0x2282,0x338,0x102,0x2283,0x338,0x102,0x2286,0x338,0x102,0x2287,0x338,0x102,0x22a2,0x338,
0x102,0x22a8,0x338,0x102,0x22a9,0x338,0x102,0x22ab,0x338,0x102,0x227c,0x338,0x102,0x227d,0x338,0x102,
0x2291,0x338,0x102,0x2292,0x338,0x102,0x22b2,0x338,0x102,0x22b3,0x338,0x102,0x22b4,0x338,0x102,0x22b5,
0x338,0x802,0x304b,0x3099,0x802,0x304d,0x3099,0x802,0x304f,0x3099,0x802,0x3051,0x3099,0x802,0x3053,0x3099,
0x802,0x3055,0x3099,0x802,0x3057,0x3099,0x802,0x3059,0x3099,0x802,0x305b,0x3099,0x802,0x305d,0x3099,0x802,
0x305f,0x3099,0x802,0x3061,0x3099,0x802,0x3064,0x3099,0x802,0x3066,0x3099,0x802,0x3068,0x3099,0x802,0x306f,
0x3099,0x802,0x306f,0x309a,0x802,0x3072,0x3099,0x802,0x3072,0x309a,0x802,0x3075,0x3099,0x802,0x3075,0x309a,
0x802,0x3078,0x3099,0x802,0x3078,0x309a,0x802,0x307b,0x3099,0x802,0x307b,0x309a,0x802,0x3046,0x3099,0x802,
0x309d,0x3099,0x802,0x30ab,0x3099,0x802,0x30ad,0x3099,0x802,0x30af,0x3099,0x802,0x30b1,0x3099,0x802,0x30b3,
0x3099,0x802,0x30b5,0x3099,0x802,0x30b7,0x3099,0x802,0x30b9,0x3099,0x802,0x30bb,0x3099,0x802,0x30bd,0x3099,
0x802,0x30bf,0x3099,0x802,0x30c1,0x3099,0x802,0x30c4,0x3099,0x802,0x30c6,0x3099,0x802,0x30c8,0x3099,0x802,
0x30cf,0x3099,0x802,0x30cf,0x309a,0x802,0x30d2,0x3099,0x802,0x30d2,0x309a,0x802,0x30d5,0x3099,0x802,0x30d5,
0x309a,0x802,0x30d8,0x3099,0x802,0x30d8,0x309a,0x802,0x30db,0x3099,0x802,0x30db,0x309a,0x802,0x30a6,0x3099,
0x802,0x30ef,0x3099,0x802,0x30f0,0x3099,0x802,0x30f1,0x3099,0x802,0x30f2,0x3099,0x802,0x30fd,0x3099,0x704,
0xd804,0xdc99,0xd804,0xdcba,0x704,0xd804,0xdc9b,0xd804,0xdcba,0x704,0xd804,0xdca5,0xd804,0xdcba,4,0xd804,
0xdd31,0xd804,0xdd27,4,0xd804,0xdd32,0xd804,0xdd27,4,0xd804,0xdf47,0xd804,0xdf3e,4,0xd804,0xdf47,
0xd804,0xdf57,4,0xd805,0xdcb9,0xd805,0xdcba,4,0xd805,0xdcb9,0xd805,0xdcb0,4,0xd805,0xdcb9,0xd805,
0xdcbd,4,0xd805,0xddb8,0xd805,0xddaf,4,0xd805,0xddb9,0xd805,0xddaf,4,0xd806,0xdd35,0xd806,0xdd30,
1,0x2b9,1,0x3b,1,0xb7,0x702,0x915,0x93c,0x702,0x916,0x93c,0x702,0x917,0x93c,0x702,
0x91c,0x93c,0x702,0x921,0x93c,0x702,0x922,0x93c,0x702,0x92b,0x93c,0x702,0x92f,0x93c,0x702,0x9a1,
0x9bc,0x702,0x9a2,0x9bc,0x702,0x9af,0x9bc,0x702,0xa32,0xa3c,0x702,0xa38,0xa3c,0x702,0xa16,0xa3c,
0x702,0xa17,0xa3c,0x702,0xa1c,0xa3c,0x702,0xa2b,0xa3c,0x702,0xb21,0xb3c,0x702,0xb22,0xb3c,2,
0xf42,0xfb7,2,0xf4c,0xfb7,2,0xf51,0xfb7,2,0xf56,0xfb7,2,0xf5b,0xfb7,2,0xf40,
0xfb5,0x8202,0xfb2,0xf80,0x8202,0xfb3,0xf80,2,0xf92,0xfb7,2,0xf9c,0xfb7,2,0xfa1,0xfb7,
2,0xfa6,0xfb7,2,0xfab,0xfb7,2,0xf90,0xfb5,1,0x3b9,1,0x60,1,0xb4,1,
0x3a9,1,0x4b,1,0x3008,1,0x3009,0x102,0x2add,0x338,1,0x8c48,1,0x66f4,1,0x8eca,
1,0x8cc8,1,0x6ed1,1,0x4e32,1,0x53e5,1,0x9f9c,1,0x5951,1,0x91d1,1,0x5587,
1,0x5948,1,0x61f6,1,0x7669,1,0x7f85,1,0x863f,1,0x87ba,1,0x88f8,1,0x908f,
1,0x6a02,1,0x6d1b,1,0x70d9,1,0x73de,1,0x843d,1,0x916a,1,0x99f1,1,0x4e82,
1,0x5375,1,0x6b04,1,0x721b,1,0x862d,1,0x9e1e,1,0x5d50,1,0x6feb,1,0x85cd,
1,0x8964,1,0x62c9,1,0x81d8,1,0x881f,1,0x5eca,1,0x6717,1,0x6d6a,1,0x72fc,
1,0x90ce,1,0x4f86,1,0x51b7,1,0x52de,1,0x64c4,1,0x6ad3,1,0x7210,1,0x76e7,
1,0x8001,1,0x8606,1,0x865c,1,0x8def,1,0x9732,1,0x9b6f,1,0x9dfa,1,0x788c,
1,0x797f,1,0x7da0,1,0x83c9,1,0x9304,1,0x9e7f,1,0x8ad6,1,0x58df,1,0x5f04,
1,0x7c60,1,0x807e,1,0x7262,1,0x78ca,1,0x8cc2,1,0x96f7,1,0x58d8,1,0x5c62,
1,0x6a13,1,0x6dda,1,0x6f0f,1,0x7d2f,1,0x7e37,1,0x964b,1,0x52d2,1,0x808b,
1,0x51dc,1,0x51cc,1,0x7a1c,1,0x7dbe,1,0x83f1,1,0x9675,1,0x8b80,1,0x62cf,
1,0x8afe,1,0x4e39,1,0x5be7,1,0x6012,1,0x7387,1,0x7570,1,0x5317,1,0x78fb,
1,0x4fbf,1,0x5fa9,1,0x4e0d,1,0x6ccc,1,0x6578,1,0x7d22,1,0x53c3,1,0x585e,
1,0x7701,1,0x8449,1,0x8aaa,1,0x6bba,1,0x8fb0,1,0x6c88,1,0x62fe,1,0x82e5,
1,0x63a0,1,0x7565,1,0x4eae,1,0x5169,1,0x51c9,1,0x6881,1,0x7ce7,1,0x826f,
1,0x8ad2,1,0x91cf,1,0x52f5,1,0x5442,1,0x5973,1,0x5eec,1,0x65c5,1,0x6ffe,
1,0x792a,1,0x95ad,1,0x9a6a,1,0x9e97,1,0x9ece,1,0x529b,1,0x66c6,1,0x6b77,
1,0x8f62,1,0x5e74,1,0x6190,1,0x6200,1,0x649a,1,0x6f23,1,0x7149,1,0x7489,
1,0x79ca,1,0x7df4,1,0x806f,1,0x8f26,1,0x84ee,1,0x9023,1,0x934a,1,0x5217,
1,0x52a3,1,0x54bd,1,0x70c8,1,0x88c2,1,0x5ec9,1,0x5ff5,1,0x637b,1,0x6bae,
1,0x7c3e,1,0x7375,1,0x4ee4,1,0x56f9,1,0x5dba,1,0x601c,1,0x73b2,1,0x7469,
1,0x7f9a,1,0x8046,1,0x9234,1,0x96f6,1,0x9748,1,0x9818,1,0x4f8b,1,0x79ae,
1,0x91b4,1,0x96b8,1,0x60e1,1,0x4e86,1,0x50da,1,0x5bee,1,0x5c3f,1,0x6599,
1,0x71ce,1,0x7642,1,0x84fc,1,0x907c,1,0x9f8d,1,0x6688,1,0x962e,1,0x5289,
1,0x677b,1,0x67f3,1,0x6d41,1,0x6e9c,1,0x7409,1,0x7559,1,0x786b,1,0x7d10,
1,0x985e,1,0x516d,1,0x622e,1,0x9678,1,0x502b,1,0x5d19,1,0x6dea,1,0x8f2a,
1,0x5f8b,1,0x6144,1,0x6817,1,0x9686,1,0x5229,1,0x540f,1,0x5c65,1,0x6613,
1,0x674e,1,0x68a8,1,0x6ce5,1,0x7406,1,0x75e2,1,0x7f79,1,0x88cf,1,0x88e1,
1,0x91cc,1,0x96e2,1,0x533f,1,0x6eba,1,0x541d,1,0x71d0,1,0x7498,1,0x85fa,
1,0x96a3,1,0x9c57,1,0x9e9f,1,0x6797,1,0x6dcb,1,0x81e8,1,0x7acb,1,0x7b20,
1,0x7c92,1,0x72c0,1,0x7099,1,0x8b58,1,0x4ec0,1,0x8336,1,0x523a,1,0x5207,
1,0x5ea6,1,0x62d3,1,0x7cd6,1,0x5b85,1,0x6d1e,1,0x66b4,1,0x8f3b,1,0x884c,
1,0x964d,1,0x898b,1,0x5ed3,1,0x5140,1,0x55c0,1,0x585a,1,0x6674,1,0x51de,
1,0x732a,1,0x76ca,1,0x793c,1,0x795e,1,0x7965,1,0x798f,1,0x9756,1,0x7cbe,
1,0x7fbd,1,0x8612,1,0x8af8,1,0x9038,1,0x90fd,1,0x98ef,1,0x98fc,1,0x9928,
1,0x9db4,1,0x90de,1,0x96b7,1,0x4fae,1,0x50e7,1,0x514d,1,0x52c9,1,0x52e4,
1,0x5351,1,0x559d,1,0x5606,1,0x5668,1,0x5840,1,0x58a8,1,0x5c64,1,0x5c6e,
1,0x6094,1,0x6168,1,0x618e,1,0x61f2,1,0x654f,1,0x65e2,1,0x6691,1,0x6885,
1,0x6d77,1,0x6e1a,1,0x6f22,1,0x716e,1,0x722b,1,0x7422,1,0x7891,1,0x793e,
1,0x7949,1,0x7948,1,0x7950,1,0x7956,1,0x795d,1,0x798d,1,0x798e,1,0x7a40,
1,0x7a81,1,0x7bc0,1,0x7e09,1,0x7e41,1,0x7f72,1,0x8005,1,0x81ed,1,0x8279,
1,0x8457,1,0x8910,1,0x8996,1,0x8b01,1,0x8b39,1,0x8cd3,1,0x8d08,1,0x8fb6,
1,0x96e3,1,0x97ff,1,0x983b,1,0x6075,2,0xd850,0xdeee,1,0x8218,1,0x4e26,1,
0x51b5,1,0x5168,1,0x4f80,1,0x5145,1,0x5180,1,0x52c7,1,0x52fa,1,0x5555,1,
0x5599,1,0x55e2,1,0x58b3,1,0x5944,1,0x5954,1,0x5a62,1,0x5b28,1,0x5ed2,1,
0x5ed9,1,0x5f69,1,0x5fad,1,0x60d8,1,0x614e,1,0x6108,1,0x6160,1,0x6234,1,
0x63c4,1,0x641c,1,0x6452,1,0x6556,1,0x671b,1,0x6756,1,0x6b79,1,0x6edb,1,
0x6ecb,1,0x701e,1,0x77a7,1,0x7235,1,0x72af,1,0x7471,1,0x7506,1,0x753b,1,
0x761d,1,0x761f,1,0x76db,1,0x76f4,1,0x774a,1,0x7740,1,0x78cc,1,0x7ab1,1,
0x7c7b,1,0x7d5b,1,0x7f3e,1,0x8352,1,0x83ef,1,0x8779,1,0x8941,1,0x8986,1,
0x8abf,1,0x8acb,1,0x8aed,1,0x8b8a,1,0x8f38,1,0x9072,1,0x9199,1,0x9276,1,
0x967c,1,0x97db,1,0x980b,1,0x9b12,2,0xd84a,0xdc4a,2,0xd84a,0xdc44,2,0xd84c,0xdfd5,
1,0x3b9d,1,0x4018,1,0x4039,2,0xd854,0xde49,2,0xd857,0xdcd0,2,0xd85f,0xded3,1,
0x9f43,1,0x9f8e,0xe02,0x5d9,0x5b4,0x1102,0x5f2,0x5b7,0x1802,0x5e9,0x5c1,0x1902,0x5e9,0x5c2,0xfb49,
0x1843,0x5e9,0x5bc,0x5c1,0xfb49,0x1943,0x5e9,0x5bc,0x5c2,0x1102,0x5d0,0x5b7,0x1202,0x5d0,0x5b8,0x1502,
0x5d0,0x5bc,0x1502,0x5d1,0x5bc,0x1502,0x5d2,0x5bc,0x1502,0x5d3,0x5bc,0x1502,0x5d4,0x5bc,0x1502,0x5d5,
0x5bc,0x1502,0x5d6,0x5bc,0x1502,0x5d8,0x5bc,0x1502,0x5d9,0x5bc,0x1502,0x5da,0x5bc,0x1502,0x5db,0x5bc,
0x1502,0x5dc,0x5bc,0x1502,0x5de,0x5bc,0x1502,0x5e0,0x5bc,0x1502,0x5e1,0x5bc,0x1502,0x5e3,0x5bc,0x1502,
0x5e4,0x5bc,0x1502,0x5e6,0x5bc,0x1502,0x5e7,0x5bc,0x1502,0x5e8,0x5bc,0x1502,0x5e9,0x5bc,0x1502,0x5ea,
0x5bc,0x1302,0x5d5,0x5b9,0x1702,0x5d1,0x5bf,0x1702,0x5db,0x5bf,0x1702,0x5e4,0x5bf,0xd804,0xd834,0xdd57,
0xd834,0xdd65,0xd804,0xd834,0xdd58,0xd834,0xdd65,0xd834,0xdd5f,0xd834,0xdd6e,4,0xd846,0xd834,0xdd58,0xd834,
0xdd65,0xd834,0xdd6e,0xd834,0xdd5f,0xd834,0xdd6f,4,0xd846,0xd834,0xdd58,0xd834,0xdd65,0xd834,0xdd6f,0xd834,
0xdd5f,0xd834,0xdd70,4,0xd846,0xd834,0xdd58,0xd834,0xdd65,0xd834,0xdd70,0xd834,0xdd5f,0xd834,0xdd71,4,
0xd846,0xd834,0xdd58,0xd834,0xdd65,0xd834,0xdd71,0xd834,0xdd5f,0xd834,0xdd72,4,0xd846,0xd834,0xdd58,0xd834,
0xdd65,0xd834,0xdd72,0xd804,0xd834,0xddb9,0xd834,0xdd65,0xd804,0xd834,0xddba,0xd834,0xdd65,0xd834,0xddbb,0xd834,
0xdd6e,4,0xd846,0xd834,0xddb9,0xd834,0xdd65,0xd834,0xdd6e,0xd834,0xddbc,0xd834,0xdd6e,4,0xd846,0xd834,
0xddba,0xd834,0xdd65,0xd834,0xdd6e,0xd834,0xddbb,0xd834,0xdd6f,4,0xd846,0xd834,0xddb9,0xd834,0xdd65,0xd834,
0xdd6f,0xd834,0xddbc,0xd834,0xdd6f,4,0xd846,0xd834,0xddba,0xd834,0xdd65,0xd834,0xdd6f,1,0x4e3d,1,
0x4e38,1,0x4e41,2,0xd840,0xdd22,1,0x4f60,1,0x4fbb,1,0x5002,1,0x507a,1,0x5099,
1,0x50cf,1,0x349e,2,0xd841,0xde3a,1,0x5154,1,0x5164,1,0x5177,2,0xd841,0xdd1c,
1,0x34b9,1,0x5167,1,0x518d,2,0xd841,0xdd4b,1,0x5197,1,0x51a4,1,0x4ecc,1,
0x51ac,2,0xd864,0xdddf,1,0x51f5,1,0x5203,1,0x34df,1,0x523b,1,0x5246,1,0x5272,
1,0x5277,1,0x3515,1,0x5305,1,0x5306,1,0x5349,1,0x535a,1,0x5373,1,0x537d,
1,0x537f,2,0xd842,0xde2c,1,0x7070,1,0x53ca,1,0x53df,2,0xd842,0xdf63,1,0x53eb,
1,0x53f1,1,0x5406,1,0x549e,1,0x5438,1,0x5448,1,0x5468,1,0x54a2,1,0x54f6,
1,0x5510,1,0x5553,1,0x5563,1,0x5584,1,0x55ab,1,0x55b3,1,0x55c2,1,0x5716,
1,0x5717,1,0x5651,1,0x5674,1,0x58ee,1,0x57ce,1,0x57f4,1,0x580d,1,0x578b,
1,0x5832,1,0x5831,1,0x58ac,2,0xd845,0xdce4,1,0x58f2,1,0x58f7,1,0x5906,1,
0x591a,1,0x5922,1,0x5962,2,0xd845,0xdea8,2,0xd845,0xdeea,1,0x59ec,1,0x5a1b,1,
0x5a27,1,0x59d8,1,0x5a66,1,0x36ee,1,0x36fc,1,0x5b08,1,0x5b3e,2,0xd846,0xddc8,
1,0x5bc3,1,0x5bd8,1,0x5bf3,2,0xd846,0xdf18,1,0x5bff,1,0x5c06,1,0x5f53,1,
0x5c22,1,0x3781,1,0x5c60,1,0x5cc0,1,0x5c8d,2,0xd847,0xdde4,1,0x5d43,2,0xd847,
0xdde6,1,0x5d6e,1,0x5d6b,1,0x5d7c,1,0x5de1,1,0x5de2,1,0x382f,1,0x5dfd,1,
0x5e28,1,0x5e3d,1,0x5e69,1,0x3862,2,0xd848,0xdd83,1,0x387c,1,0x5eb0,1,0x5eb3,
1,0x5eb6,2,0xd868,0xdf92,1,0x5efe,2,0xd848,0xdf31,1,0x8201,1,0x5f22,1,0x38c7,
2,0xd84c,0xdeb8,2,0xd858,0xddda,1,0x5f62,1,0x5f6b,1,0x38e3,1,0x5f9a,1,0x5fcd,
1,0x5fd7,1,0x5ff9,1,0x6081,1,0x393a,1,0x391c,2,0xd849,0xded4,1,0x60c7,1,
0x6148,1,0x614c,1,0x617a,1,0x61b2,1,0x61a4,1,0x61af,1,0x61de,1,0x6210,1,
0x621b,1,0x625d,1,0x62b1,1,0x62d4,1,0x6350,2,0xd84a,0xdf0c,1,0x633d,1,0x62fc,
1,0x6368,1,0x6383,1,0x63e4,2,0xd84a,0xdff1,1,0x6422,1,0x63c5,1,0x63a9,1,
0x3a2e,1,0x6469,1,0x647e,1,0x649d,1,0x6477,1,0x3a6c,1,0x656c,2,0xd84c,0xdc0a,
1,0x65e3,1,0x66f8,1,0x6649,1,0x3b19,1,0x3b08,1,0x3ae4,1,0x5192,1,0x5195,
1,0x6700,1,0x669c,1,0x80ad,1,0x43d9,1,0x6721,1,0x675e,1,0x6753,2,0xd84c,
0xdfc3,1,0x3b49,1,0x67fa,1,0x6785,1,0x6852,2,0xd84d,0xdc6d,1,0x688e,1,0x681f,
1,0x6914,1,0x6942,1,0x69a3,1,0x69ea,1,0x6aa8,2,0xd84d,0xdea3,1,0x6adb,1,
0x3c18,1,0x6b21,2,0xd84e,0xdca7,1,0x6b54,1,0x3c4e,1,0x6b72,1,0x6b9f,1,0x6bbb,
2,0xd84e,0xde8d,2,0xd847,0xdd0b,2,0xd84e,0xdefa,1,0x6c4e,2,0xd84f,0xdcbc,1,0x6cbf,
1,0x6ccd,1,0x6c67,1,0x6d16,1,0x6d3e,1,0x6d69,1,0x6d78,1,0x6d85,2,0xd84f,
0xdd1e,1,0x6d34,1,0x6e2f,1,0x6e6e,1,0x3d33,1,0x6ec7,2,0xd84f,0xded1,1,0x6df9,
1,0x6f6e,2,0xd84f,0xdf5e,2,0xd84f,0xdf8e,1,0x6fc6,1,0x7039,1,0x701b,1,0x3d96,
1,0x704a,1,0x707d,1,0x7077,1,0x70ad,2,0xd841,0xdd25,1,0x7145,2,0xd850,0xde63,
1,0x719c,2,0xd850,0xdfab,1,0x7228,1,0x7250,2,0xd851,0xde08,1,0x7280,1,0x7295,
2,0xd851,0xdf35,2,0xd852,0xdc14,1,0x737a,1,0x738b,1,0x3eac,1,0x73a5,1,0x3eb8,
1,0x7447,1,0x745c,1,0x7485,1,0x74ca,1,0x3f1b,1,0x7524,2,0xd853,0xdc36,1,
0x753e,2,0xd853,0xdc92,2,0xd848,0xdd9f,1,0x7610,2,0xd853,0xdfa1,2,0xd853,0xdfb8,2,
0xd854,0xdc44,1,0x3ffc,1,0x4008,2,0xd854,0xdcf3,2,0xd854,0xdcf2,2,0xd854,0xdd19,2,
0xd854,0xdd33,1,0x771e,1,0x771f,1,0x778b,1,0x4046,1,0x4096,2,0xd855,0xdc1d,1,
0x784e,1,0x40e3,2,0xd855,0xde26,2,0xd855,0xde9a,2,0xd855,0xdec5,1,0x79eb,1,0x412f,
1,0x7a4a,1,0x7a4f,2,0xd856,0xdd7c,2,0xd856,0xdea7,1,0x7aee,1,0x4202,2,0xd856,
0xdfab,1,0x7bc6,1,0x7bc9,1,0x4227,2,0xd857,0xdc80,1,0x7cd2,1,0x42a0,1,0x7ce8,
1,0x7ce3,1,0x7d00,2,0xd857,0xdf86,1,0x7d63,1,0x4301,1,0x7dc7,1,0x7e02,1,
0x7e45,1,0x4334,2,0xd858,0xde28,2,0xd858,0xde47,1,0x4359,2,0xd858,0xded9,1,0x7f7a,
2,0xd858,0xdf3e,1,0x7f95,1,0x7ffa,2,0xd859,0xdcda,2,0xd859,0xdd23,1,0x8060,2,
0xd859,0xdda8,1,0x8070,2,0xd84c,0xdf5f,1,0x43d5,1,0x80b2,1,0x8103,1,0x440b,1,
0x813e,1,0x5ab5,2,0xd859,0xdfa7,2,0xd859,0xdfb5,2,0xd84c,0xdf93,2,0xd84c,0xdf9c,1,
0x8204,1,0x8f9e,1,0x446b,1,0x8291,1,0x828b,1,0x829d,1,0x52b3,1,0x82b1,1,
0x82b3,1,0x82bd,1,0x82e6,2,0xd85a,0xdf3c,1,0x831d,1,0x8363,1,0x83ad,1,0x8323,
1,0x83bd,1,0x83e7,1,0x8353,1,0x83ca,1,0x83cc,1,0x83dc,2,0xd85b,0xdc36,2,
0xd85b,0xdd6b,2,0xd85b,0xdcd5,1,0x452b,1,0x84f1,1,0x84f3,1,0x8516,2,0xd85c,0xdfca,
1,0x8564,2,0xd85b,0xdf2c,1,0x455d,1,0x4561,2,0xd85b,0xdfb1,2,0xd85c,0xdcd2,1,
0x456b,1,0x8650,1,0x8667,1,0x8669,1,0x86a9,1,0x8688,1,0x870e,1,0x86e2,1,
0x8728,1,0x876b,1,0x8786,1,0x45d7,1,0x87e1,1,0x8801,1,0x45f9,1,0x8860,1,
0x8863,2,0xd85d,0xde67,1,0x88d7,1,0x88de,1,0x4635,1,0x88fa,1,0x34bb,2,0xd85e,
0xdcae,2,0xd85e,0xdd66,1,0x46be,1,0x46c7,1,0x8aa0,1,0x8c55,2,0xd85f,0xdca8,1,
0x8cab,1,0x8cc1,1,0x8d1b,1,0x8d77,2,0xd85f,0xdf2f,2,0xd842,0xdc04,1,0x8dcb,1,
0x8dbc,1,0x8df0,2,0xd842,0xdcde,1,0x8ed4,2,0xd861,0xddd2,2,0xd861,0xdded,1,0x9094,
1,0x90f1,1,0x9111,2,0xd861,0xdf2e,1,0x911b,1,0x9238,1,0x92d7,1,0x92d8,1,
0x927c,1,0x93f9,1,0x9415,2,0xd862,0xdffa,1,0x958b,1,0x4995,1,0x95b7,2,0xd863,
0xdd77,1,0x49e6,1,0x96c3,1,0x5db2,1,0x9723,2,0xd864,0xdd45,2,0xd864,0xde1a,1,
0x4a6e,1,0x4a76,1,0x97e0,2,0xd865,0xdc0a,1,0x4ab2,2,0xd865,0xdc96,1,0x9829,2,
0xd865,0xddb6,1,0x98e2,1,0x4b33,1,0x9929,1,0x99a7,1,0x99c2,1,0x99fe,1,0x4bce,
2,0xd866,0xdf30,1,0x9c40,1,0x9cfd,1,0x4cce,1,0x4ced,1,0x9d67,2,0xd868,0xdcce,
1,0x4cf8,2,0xd868,0xdd05,2,0xd868,0xde0e,2,0xd868,0xde91,1,0x9ebb,1,0x4d56,1,
0x9ef9,1,0x9efe,1,0x9f05,1,0x9f0f,1,0x9f16,1,0x9f3b,2,0xd869,0xde00,0x3ac,0xe642,
0x3b1,0x301,0x3ad,0xe642,0x3b5,0x301,0x3ae,0xe642,0x3b7,0x301,0x3af,0xe642,0x3b9,0x301,0x3cc,0xe642,
0x3bf,0x301,0x3cd,0xe642,0x3c5,0x301,0x3ce,0xe642,0x3c9,0x301,0x386,0xe642,0x391,0x301,0x388,0xe642,
0x395,0x301,0x389,0xe642,0x397,0x301,0x390,1,0xe643,0x3b9,0x308,0x301,0x38a,0xe642,0x399,0x301,
0x3b0,1,0xe643,0x3c5,0x308,0x301,0x38e,0xe642,0x3a5,0x301,0x385,0xe642,0xa8,0x301,0x38c,0xe642,
0x39f,0x301,0x38f,0xe642,0x3a9,0x301,0xc5,0xe642,0x41,0x30a,0xe6e6,0xe681,0x300,0xe6e6,0xe681,0x301,
0xe6e6,0xe681,0x313,0xe6e6,0xe682,0x308,0x301,0x8100,0x8282,0xf71,0xf72,0x8100,0x8482,0xf71,0xf74,0x8100,
0x8282,0xf71,0xf80,0
};

static const uint8_t norm2_nfc_data_smallFCD[256]={
0xc0,0xef,3,0x7f,0xdf,0x70,0xcf,0x87,0xc7,0xe6,0x66,0x46,0x64,0x46,0x66,0x5b,
0x12,0,0,4,0,0,0,0x43,0x20,2,0x69,0xae,0xc2,0xc0,0xff,0xff,
0xc0,0x72,0xbf,0,0,0,0,0,0,0,0x40,0,0x80,0x88,0,0,
0xfe,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0x98,0,0xc3,0x66,0xe0,0x80,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,3,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,7,0,0,2,0
};

#endif  // INCLUDED_FROM_NORMALIZER2_CPP

#endif

U_NAMESPACE_BEGIN

// Public API dispatch via Normalizer2 subclasses -------------------------- ***

Normalizer2::~Normalizer2() {}

void
Normalizer2::normalizeUTF8(uint32_t /*options*/, StringPiece src, ByteSink &sink,
                           Edits *edits, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) {
        return;
    }
    if (edits != nullptr) {
        errorCode = U_UNSUPPORTED_ERROR;
        return;
    }
    UnicodeString src16 = UnicodeString::fromUTF8(src);
    normalize(src16, errorCode).toUTF8(sink);
}

UBool
Normalizer2::getRawDecomposition(UChar32, UnicodeString &) const {
    return FALSE;
}

UChar32
Normalizer2::composePair(UChar32, UChar32) const {
    return U_SENTINEL;
}

uint8_t
Normalizer2::getCombiningClass(UChar32 /*c*/) const {
    return 0;
}

UBool
Normalizer2::isNormalizedUTF8(StringPiece s, UErrorCode &errorCode) const {
    return U_SUCCESS(errorCode) && isNormalized(UnicodeString::fromUTF8(s), errorCode);
}

// Normalizer2 implementation for the old UNORM_NONE.
class NoopNormalizer2 : public Normalizer2 {
    virtual ~NoopNormalizer2();

    virtual UnicodeString &
    normalize(const UnicodeString &src,
              UnicodeString &dest,
              UErrorCode &errorCode) const U_OVERRIDE {
        if(U_SUCCESS(errorCode)) {
            if(&dest!=&src) {
                dest=src;
            } else {
                errorCode=U_ILLEGAL_ARGUMENT_ERROR;
            }
        }
        return dest;
    }
    virtual void
    normalizeUTF8(uint32_t options, StringPiece src, ByteSink &sink,
                  Edits *edits, UErrorCode &errorCode) const U_OVERRIDE {
        if(U_SUCCESS(errorCode)) {
            if (edits != nullptr) {
                if ((options & U_EDITS_NO_RESET) == 0) {
                    edits->reset();
                }
                edits->addUnchanged(src.length());
            }
            if ((options & U_OMIT_UNCHANGED_TEXT) == 0) {
                sink.Append(src.data(), src.length());
            }
            sink.Flush();
        }
    }

    virtual UnicodeString &
    normalizeSecondAndAppend(UnicodeString &first,
                             const UnicodeString &second,
                             UErrorCode &errorCode) const U_OVERRIDE {
        if(U_SUCCESS(errorCode)) {
            if(&first!=&second) {
                first.append(second);
            } else {
                errorCode=U_ILLEGAL_ARGUMENT_ERROR;
            }
        }
        return first;
    }
    virtual UnicodeString &
    append(UnicodeString &first,
           const UnicodeString &second,
           UErrorCode &errorCode) const U_OVERRIDE {
        if(U_SUCCESS(errorCode)) {
            if(&first!=&second) {
                first.append(second);
            } else {
                errorCode=U_ILLEGAL_ARGUMENT_ERROR;
            }
        }
        return first;
    }
    virtual UBool
    getDecomposition(UChar32, UnicodeString &) const U_OVERRIDE {
        return FALSE;
    }
    // No need to U_OVERRIDE the default getRawDecomposition().
    virtual UBool
    isNormalized(const UnicodeString &, UErrorCode &errorCode) const U_OVERRIDE {
        return U_SUCCESS(errorCode);
    }
    virtual UBool
    isNormalizedUTF8(StringPiece, UErrorCode &errorCode) const U_OVERRIDE {
        return U_SUCCESS(errorCode);
    }
    virtual UNormalizationCheckResult
    quickCheck(const UnicodeString &, UErrorCode &) const U_OVERRIDE {
        return UNORM_YES;
    }
    virtual int32_t
    spanQuickCheckYes(const UnicodeString &s, UErrorCode &) const U_OVERRIDE {
        return s.length();
    }
    virtual UBool hasBoundaryBefore(UChar32) const U_OVERRIDE { return TRUE; }
    virtual UBool hasBoundaryAfter(UChar32) const U_OVERRIDE { return TRUE; }
    virtual UBool isInert(UChar32) const U_OVERRIDE { return TRUE; }
};

NoopNormalizer2::~NoopNormalizer2() {}

Normalizer2WithImpl::~Normalizer2WithImpl() {}

DecomposeNormalizer2::~DecomposeNormalizer2() {}

ComposeNormalizer2::~ComposeNormalizer2() {}

FCDNormalizer2::~FCDNormalizer2() {}

// instance cache ---------------------------------------------------------- ***

U_CDECL_BEGIN
static UBool U_CALLCONV uprv_normalizer2_cleanup();
U_CDECL_END

static Normalizer2   *noopSingleton;
static icu::UInitOnce noopInitOnce = U_INITONCE_INITIALIZER;

static void U_CALLCONV initNoopSingleton(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    noopSingleton=new NoopNormalizer2;
    if(noopSingleton==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    ucln_common_registerCleanup(UCLN_COMMON_NORMALIZER2, uprv_normalizer2_cleanup);
}

const Normalizer2 *Normalizer2Factory::getNoopInstance(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return NULL; }
    umtx_initOnce(noopInitOnce, &initNoopSingleton, errorCode);
    return noopSingleton;
}

const Normalizer2Impl *
Normalizer2Factory::getImpl(const Normalizer2 *norm2) {
    return &((Normalizer2WithImpl *)norm2)->impl;
}

Norm2AllModes::~Norm2AllModes() {
    delete impl;
}

Norm2AllModes *
Norm2AllModes::createInstance(Normalizer2Impl *impl, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        delete impl;
        return NULL;
    }
    Norm2AllModes *allModes=new Norm2AllModes(impl);
    if(allModes==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        delete impl;
        return NULL;
    }
    return allModes;
}

#if NORM2_HARDCODE_NFC_DATA
Norm2AllModes *
Norm2AllModes::createNFCInstance(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    Normalizer2Impl *impl=new Normalizer2Impl;
    if(impl==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    impl->init(norm2_nfc_data_indexes, &norm2_nfc_data_trie,
               norm2_nfc_data_extraData, norm2_nfc_data_smallFCD);
    return createInstance(impl, errorCode);
}

static Norm2AllModes *nfcSingleton;

static icu::UInitOnce nfcInitOnce = U_INITONCE_INITIALIZER;

static void U_CALLCONV initNFCSingleton(UErrorCode &errorCode) {
    nfcSingleton=Norm2AllModes::createNFCInstance(errorCode);
    ucln_common_registerCleanup(UCLN_COMMON_NORMALIZER2, uprv_normalizer2_cleanup);
}

const Norm2AllModes *
Norm2AllModes::getNFCInstance(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return NULL; }
    umtx_initOnce(nfcInitOnce, &initNFCSingleton, errorCode);
    return nfcSingleton;
}

const Normalizer2 *
Normalizer2::getNFCInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFCInstance(errorCode);
    return allModes!=NULL ? &allModes->comp : NULL;
}

const Normalizer2 *
Normalizer2::getNFDInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFCInstance(errorCode);
    return allModes!=NULL ? &allModes->decomp : NULL;
}

const Normalizer2 *Normalizer2Factory::getFCDInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFCInstance(errorCode);
    return allModes!=NULL ? &allModes->fcd : NULL;
}

const Normalizer2 *Normalizer2Factory::getFCCInstance(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFCInstance(errorCode);
    return allModes!=NULL ? &allModes->fcc : NULL;
}

const Normalizer2Impl *
Normalizer2Factory::getNFCImpl(UErrorCode &errorCode) {
    const Norm2AllModes *allModes=Norm2AllModes::getNFCInstance(errorCode);
    return allModes!=NULL ? allModes->impl : NULL;
}
#endif  // NORM2_HARDCODE_NFC_DATA

U_CDECL_BEGIN

static UBool U_CALLCONV uprv_normalizer2_cleanup() {
    delete noopSingleton;
    noopSingleton = NULL;
    noopInitOnce.reset(); 
#if NORM2_HARDCODE_NFC_DATA
    delete nfcSingleton;
    nfcSingleton = NULL;
    nfcInitOnce.reset(); 
#endif
    return TRUE;
}

U_CDECL_END

U_NAMESPACE_END

// C API ------------------------------------------------------------------- ***

U_NAMESPACE_USE

U_CAPI const UNormalizer2 * U_EXPORT2
unorm2_getNFCInstance(UErrorCode *pErrorCode) {
    return (const UNormalizer2 *)Normalizer2::getNFCInstance(*pErrorCode);
}

U_CAPI const UNormalizer2 * U_EXPORT2
unorm2_getNFDInstance(UErrorCode *pErrorCode) {
    return (const UNormalizer2 *)Normalizer2::getNFDInstance(*pErrorCode);
}

U_CAPI void U_EXPORT2
unorm2_close(UNormalizer2 *norm2) {
    delete (Normalizer2 *)norm2;
}

U_CAPI int32_t U_EXPORT2
unorm2_normalize(const UNormalizer2 *norm2,
                 const UChar *src, int32_t length,
                 UChar *dest, int32_t capacity,
                 UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if( (src==NULL ? length!=0 : length<-1) ||
        (dest==NULL ? capacity!=0 : capacity<0) ||
        (src==dest && src!=NULL)
    ) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString destString(dest, 0, capacity);
    // length==0: Nothing to do, and n2wi->normalize(NULL, NULL, buffer, ...) would crash.
    if(length!=0) {
        const Normalizer2 *n2=(const Normalizer2 *)norm2;
        const Normalizer2WithImpl *n2wi=dynamic_cast<const Normalizer2WithImpl *>(n2);
        if(n2wi!=NULL) {
            // Avoid duplicate argument checking and support NUL-terminated src.
            ReorderingBuffer buffer(n2wi->impl, destString);
            if(buffer.init(length, *pErrorCode)) {
                n2wi->normalize(src, length>=0 ? src+length : NULL, buffer, *pErrorCode);
            }
        } else {
            UnicodeString srcString(length<0, src, length);
            n2->normalize(srcString, destString, *pErrorCode);
        }
    }
    return destString.extract(dest, capacity, *pErrorCode);
}

static int32_t
normalizeSecondAndAppend(const UNormalizer2 *norm2,
                         UChar *first, int32_t firstLength, int32_t firstCapacity,
                         const UChar *second, int32_t secondLength,
                         UBool doNormalize,
                         UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if( (second==NULL ? secondLength!=0 : secondLength<-1) ||
        (first==NULL ? (firstCapacity!=0 || firstLength!=0) :
                       (firstCapacity<0 || firstLength<-1)) ||
        (first==second && first!=NULL)
    ) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString firstString(first, firstLength, firstCapacity);
    firstLength=firstString.length();  // In case it was -1.
    // secondLength==0: Nothing to do, and n2wi->normalizeAndAppend(NULL, NULL, buffer, ...) would crash.
    if(secondLength!=0) {
        const Normalizer2 *n2=(const Normalizer2 *)norm2;
        const Normalizer2WithImpl *n2wi=dynamic_cast<const Normalizer2WithImpl *>(n2);
        if(n2wi!=NULL) {
            // Avoid duplicate argument checking and support NUL-terminated src.
            UnicodeString safeMiddle;
            {
                ReorderingBuffer buffer(n2wi->impl, firstString);
                if(buffer.init(firstLength+secondLength+1, *pErrorCode)) {  // destCapacity>=-1
                    n2wi->normalizeAndAppend(second, secondLength>=0 ? second+secondLength : NULL,
                                             doNormalize, safeMiddle, buffer, *pErrorCode);
                }
            }  // The ReorderingBuffer destructor finalizes firstString.
            if(U_FAILURE(*pErrorCode) || firstString.length()>firstCapacity) {
                // Restore the modified suffix of the first string.
                // This does not restore first[] array contents between firstLength and firstCapacity.
                // (That might be uninitialized memory, as far as we know.)
                if(first!=NULL) { /* don't dereference NULL */
                  safeMiddle.extract(0, 0x7fffffff, first+firstLength-safeMiddle.length());
                  if(firstLength<firstCapacity) {
                    first[firstLength]=0;  // NUL-terminate in case it was originally.
                  }
                }
            }
        } else {
            UnicodeString secondString(secondLength<0, second, secondLength);
            if(doNormalize) {
                n2->normalizeSecondAndAppend(firstString, secondString, *pErrorCode);
            } else {
                n2->append(firstString, secondString, *pErrorCode);
            }
        }
    }
    return firstString.extract(first, firstCapacity, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
unorm2_normalizeSecondAndAppend(const UNormalizer2 *norm2,
                                UChar *first, int32_t firstLength, int32_t firstCapacity,
                                const UChar *second, int32_t secondLength,
                                UErrorCode *pErrorCode) {
    return normalizeSecondAndAppend(norm2,
                                    first, firstLength, firstCapacity,
                                    second, secondLength,
                                    TRUE, pErrorCode);
}

U_CAPI int32_t U_EXPORT2
unorm2_append(const UNormalizer2 *norm2,
              UChar *first, int32_t firstLength, int32_t firstCapacity,
              const UChar *second, int32_t secondLength,
              UErrorCode *pErrorCode) {
    return normalizeSecondAndAppend(norm2,
                                    first, firstLength, firstCapacity,
                                    second, secondLength,
                                    FALSE, pErrorCode);
}

U_CAPI int32_t U_EXPORT2
unorm2_getDecomposition(const UNormalizer2 *norm2,
                        UChar32 c, UChar *decomposition, int32_t capacity,
                        UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(decomposition==NULL ? capacity!=0 : capacity<0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString destString(decomposition, 0, capacity);
    if(reinterpret_cast<const Normalizer2 *>(norm2)->getDecomposition(c, destString)) {
        return destString.extract(decomposition, capacity, *pErrorCode);
    } else {
        return -1;
    }
}

U_CAPI int32_t U_EXPORT2
unorm2_getRawDecomposition(const UNormalizer2 *norm2,
                           UChar32 c, UChar *decomposition, int32_t capacity,
                           UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(decomposition==NULL ? capacity!=0 : capacity<0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString destString(decomposition, 0, capacity);
    if(reinterpret_cast<const Normalizer2 *>(norm2)->getRawDecomposition(c, destString)) {
        return destString.extract(decomposition, capacity, *pErrorCode);
    } else {
        return -1;
    }
}

U_CAPI UChar32 U_EXPORT2
unorm2_composePair(const UNormalizer2 *norm2, UChar32 a, UChar32 b) {
    return reinterpret_cast<const Normalizer2 *>(norm2)->composePair(a, b);
}

U_CAPI uint8_t U_EXPORT2
unorm2_getCombiningClass(const UNormalizer2 *norm2, UChar32 c) {
    return reinterpret_cast<const Normalizer2 *>(norm2)->getCombiningClass(c);
}

U_CAPI UBool U_EXPORT2
unorm2_isNormalized(const UNormalizer2 *norm2,
                    const UChar *s, int32_t length,
                    UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if((s==NULL && length!=0) || length<-1) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString sString(length<0, s, length);
    return ((const Normalizer2 *)norm2)->isNormalized(sString, *pErrorCode);
}

U_CAPI UNormalizationCheckResult U_EXPORT2
unorm2_quickCheck(const UNormalizer2 *norm2,
                  const UChar *s, int32_t length,
                  UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return UNORM_NO;
    }
    if((s==NULL && length!=0) || length<-1) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return UNORM_NO;
    }
    UnicodeString sString(length<0, s, length);
    return ((const Normalizer2 *)norm2)->quickCheck(sString, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
unorm2_spanQuickCheckYes(const UNormalizer2 *norm2,
                         const UChar *s, int32_t length,
                         UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if((s==NULL && length!=0) || length<-1) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString sString(length<0, s, length);
    return ((const Normalizer2 *)norm2)->spanQuickCheckYes(sString, *pErrorCode);
}

U_CAPI UBool U_EXPORT2
unorm2_hasBoundaryBefore(const UNormalizer2 *norm2, UChar32 c) {
    return ((const Normalizer2 *)norm2)->hasBoundaryBefore(c);
}

U_CAPI UBool U_EXPORT2
unorm2_hasBoundaryAfter(const UNormalizer2 *norm2, UChar32 c) {
    return ((const Normalizer2 *)norm2)->hasBoundaryAfter(c);
}

U_CAPI UBool U_EXPORT2
unorm2_isInert(const UNormalizer2 *norm2, UChar32 c) {
    return ((const Normalizer2 *)norm2)->isInert(c);
}

// Some properties APIs ---------------------------------------------------- ***

U_CAPI uint8_t U_EXPORT2
u_getCombiningClass(UChar32 c) {
    UErrorCode errorCode=U_ZERO_ERROR;
    const Normalizer2 *nfd=Normalizer2::getNFDInstance(errorCode);
    if(U_SUCCESS(errorCode)) {
        return nfd->getCombiningClass(c);
    } else {
        return 0;
    }
}

U_CFUNC uint16_t
unorm_getFCD16(UChar32 c) {
    UErrorCode errorCode=U_ZERO_ERROR;
    const Normalizer2Impl *impl=Normalizer2Factory::getNFCImpl(errorCode);
    if(U_SUCCESS(errorCode)) {
        return impl->getFCD16(c);
    } else {
        return 0;
    }
}

#endif  // !UCONFIG_NO_NORMALIZATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2009-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  normalizer2impl.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2009nov22
*   created by: Markus W. Scherer
*/

// #define UCPTRIE_DEBUG



#if !UCONFIG_NO_NORMALIZATION

















// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// ucptrie_impl.h (modified from utrie2_impl.h)
// created: 2017dec29 Markus W. Scherer



#ifdef UCPTRIE_DEBUG

#endif

// UCPTrie signature values, in platform endianness and opposite endianness.
// The UCPTrie signature ASCII byte values spell "Tri3".
#define UCPTRIE_SIG     0x54726933
#define UCPTRIE_OE_SIG  0x33697254

/**
 * Header data for the binary, memory-mappable representation of a UCPTrie/CodePointTrie.
 * @internal
 */
struct UCPTrieHeader {
    /** "Tri3" in big-endian US-ASCII (0x54726933) */
    uint32_t signature;

    /**
     * Options bit field:
     * Bits 15..12: Data length bits 19..16.
     * Bits 11..8: Data null block offset bits 19..16.
     * Bits 7..6: UCPTrieType
     * Bits 5..3: Reserved (0).
     * Bits 2..0: UCPTrieValueWidth
     */
    uint16_t options;

    /** Total length of the index tables. */
    uint16_t indexLength;

    /** Data length bits 15..0. */
    uint16_t dataLength;

    /** Index-3 null block offset, 0x7fff or 0xffff if none. */
    uint16_t index3NullOffset;

    /** Data null block offset bits 15..0, 0xfffff if none. */
    uint16_t dataNullOffset;

    /**
     * First code point of the single-value range ending with U+10ffff,
     * rounded up and then shifted right by UCPTRIE_SHIFT_2.
     */
    uint16_t shiftedHighStart;
};

/**
 * Constants for use with UCPTrieHeader.options.
 * @internal
 */
enum {
    UCPTRIE_OPTIONS_DATA_LENGTH_MASK = 0xf000,
    UCPTRIE_OPTIONS_DATA_NULL_OFFSET_MASK = 0xf00,
    UCPTRIE_OPTIONS_RESERVED_MASK = 0x38,
    UCPTRIE_OPTIONS_VALUE_BITS_MASK = 7,
    /**
     * Value for index3NullOffset which indicates that there is no index-3 null block.
     * Bit 15 is unused for this value because this bit is used if the index-3 contains
     * 18-bit indexes.
     */
    UCPTRIE_NO_INDEX3_NULL_OFFSET = 0x7fff,
    UCPTRIE_NO_DATA_NULL_OFFSET = 0xfffff
};

// Internal constants.
enum {
    /** The length of the BMP index table. 1024=0x400 */
    UCPTRIE_BMP_INDEX_LENGTH = 0x10000 >> UCPTRIE_FAST_SHIFT,

    UCPTRIE_SMALL_LIMIT = 0x1000,
    UCPTRIE_SMALL_INDEX_LENGTH = UCPTRIE_SMALL_LIMIT >> UCPTRIE_FAST_SHIFT,

    /** Shift size for getting the index-3 table offset. */
    UCPTRIE_SHIFT_3 = 4,

    /** Shift size for getting the index-2 table offset. */
    UCPTRIE_SHIFT_2 = 5 + UCPTRIE_SHIFT_3,

    /** Shift size for getting the index-1 table offset. */
    UCPTRIE_SHIFT_1 = 5 + UCPTRIE_SHIFT_2,

    /**
     * Difference between two shift sizes,
     * for getting an index-2 offset from an index-3 offset. 5=9-4
     */
    UCPTRIE_SHIFT_2_3 = UCPTRIE_SHIFT_2 - UCPTRIE_SHIFT_3,

    /**
     * Difference between two shift sizes,
     * for getting an index-1 offset from an index-2 offset. 5=14-9
     */
    UCPTRIE_SHIFT_1_2 = UCPTRIE_SHIFT_1 - UCPTRIE_SHIFT_2,

    /**
     * Number of index-1 entries for the BMP. (4)
     * This part of the index-1 table is omitted from the serialized form.
     */
    UCPTRIE_OMITTED_BMP_INDEX_1_LENGTH = 0x10000 >> UCPTRIE_SHIFT_1,

    /** Number of entries in an index-2 block. 32=0x20 */
    UCPTRIE_INDEX_2_BLOCK_LENGTH = 1 << UCPTRIE_SHIFT_1_2,

    /** Mask for getting the lower bits for the in-index-2-block offset. */
    UCPTRIE_INDEX_2_MASK = UCPTRIE_INDEX_2_BLOCK_LENGTH - 1,

    /** Number of code points per index-2 table entry. 512=0x200 */
    UCPTRIE_CP_PER_INDEX_2_ENTRY = 1 << UCPTRIE_SHIFT_2,

    /** Number of entries in an index-3 block. 32=0x20 */
    UCPTRIE_INDEX_3_BLOCK_LENGTH = 1 << UCPTRIE_SHIFT_2_3,

    /** Mask for getting the lower bits for the in-index-3-block offset. */
    UCPTRIE_INDEX_3_MASK = UCPTRIE_INDEX_3_BLOCK_LENGTH - 1,

    /** Number of entries in a small data block. 16=0x10 */
    UCPTRIE_SMALL_DATA_BLOCK_LENGTH = 1 << UCPTRIE_SHIFT_3,

    /** Mask for getting the lower bits for the in-small-data-block offset. */
    UCPTRIE_SMALL_DATA_MASK = UCPTRIE_SMALL_DATA_BLOCK_LENGTH - 1
};

typedef UChar32
UCPTrieGetRange(const void *trie, UChar32 start,
                UCPMapValueFilter *filter, const void *context, uint32_t *pValue);

U_CFUNC UChar32
ucptrie_internalGetRange(UCPTrieGetRange *getRange,
                         const void *trie, UChar32 start,
                         UCPMapRangeOption option, uint32_t surrogateValue,
                         UCPMapValueFilter *filter, const void *context, uint32_t *pValue);

#ifdef UCPTRIE_DEBUG
U_CFUNC void
ucptrie_printLengths(const UCPTrie *trie, const char *which);

U_CFUNC void umutablecptrie_setName(UMutableCPTrie *builder, const char *name);
#endif

/*
 * Format of the binary, memory-mappable representation of a UCPTrie/CodePointTrie.
 * For overview information see http://site.icu-project.org/design/struct/utrie
 *
 * The binary trie data should be 32-bit-aligned.
 * The overall layout is:
 *
 * UCPTrieHeader header; -- 16 bytes, see struct definition above
 * uint16_t index[header.indexLength];
 * uintXY_t data[header.dataLength];
 *
 * The trie data array is an array of uint16_t, uint32_t, or uint8_t,
 * specified via the UCPTrieValueWidth when building the trie.
 * The data array is 32-bit-aligned for uint32_t, otherwise 16-bit-aligned.
 * The overall length of the trie data is a multiple of 4 bytes.
 * (Padding is added at the end of the index array and/or near the end of the data array as needed.)
 *
 * The length of the data array (dataLength) is stored as an integer split across two fields
 * of the header struct (high bits in header.options).
 *
 * The trie type can be "fast" or "small" which determines the index structure,
 * specified via the UCPTrieType when building the trie.
 *
 * The type and valueWidth are stored in the header.options.
 * There are reserved type and valueWidth values, and reserved header.options bits.
 * They could be used in future format extensions.
 * Code reading the trie structure must fail with an error when unknown values or options are set.
 *
 * Values for ASCII character (U+0000..U+007F) can always be found at the start of the data array.
 *
 * Values for code points below a type-specific fast-indexing limit are found via two-stage lookup.
 * For a "fast" trie, the limit is the BMP/supplementary boundary at U+10000.
 * For a "small" trie, the limit is UCPTRIE_SMALL_MAX+1=U+1000.
 *
 * All code points in the range highStart..U+10FFFF map to a single highValue
 * which is stored at the second-to-last position of the data array.
 * (See UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET.)
 * The highStart value is header.shiftedHighStart<<UCPTRIE_SHIFT_2.
 * (UCPTRIE_SHIFT_2=9)
 *
 * Values for code points fast_limit..highStart-1 are found via four-stage lookup.
 * The data block size is smaller for this range than for the fast range.
 * This together with more index stages with small blocks makes this range
 * more easily compactable.
 *
 * There is also a trie error value stored at the last position of the data array.
 * (See UCPTRIE_ERROR_VALUE_NEG_DATA_OFFSET.)
 * It is intended to be returned for inputs that are not Unicode code points
 * (outside U+0000..U+10FFFF), or in string processing for ill-formed input
 * (unpaired surrogate in UTF-16, ill-formed UTF-8 subsequence).
 *
 * For a "fast" trie:
 *
 * The index array starts with the BMP index table for BMP code point lookup.
 * Its length is 1024=0x400.
 *
 * The supplementary index-1 table follows the BMP index table.
 * Variable length, for code points up to highStart-1.
 * Maximum length 64=0x40=0x100000>>UCPTRIE_SHIFT_1.
 * (For 0x100000 supplementary code points U+10000..U+10ffff.)
 *
 * After this index-1 table follow the variable-length index-3 and index-2 tables.
 *
 * The supplementary index tables are omitted completely
 * if there is only BMP data (highStart<=U+10000).
 *
 * For a "small" trie:
 *
 * The index array starts with a fast-index table for lookup of code points U+0000..U+0FFF.
 *
 * The "supplementary" index tables are always stored.
 * The index-1 table starts from U+0000, its maximum length is 68=0x44=0x110000>>UCPTRIE_SHIFT_1.
 *
 * For both trie types:
 *
 * The last index-2 block may be a partial block, storing indexes only for code points
 * below highStart.
 *
 * Lookup for ASCII code point c:
 *
 * Linear access from the start of the data array.
 *
 * value = data[c];
 *
 * Lookup for fast-range code point c:
 *
 * Shift the code point right by UCPTRIE_FAST_SHIFT=6 bits,
 * fetch the index array value at that offset,
 * add the lower code point bits, index into the data array.
 *
 * value = data[index[c>>6] + (c&0x3f)];
 *
 * (This works for ASCII as well.)
 *
 * Lookup for small-range code point c below highStart:
 *
 * Split the code point into four bit fields using several sets of shifts & masks
 * to read consecutive values from the index-1, index-2, index-3 and data tables.
 *
 * If all of the data block offsets in an index-3 block fit within 16 bits (up to 0xffff),
 * then the data block offsets are stored directly as uint16_t.
 *
 * Otherwise (this is very unusual but possible), the index-2 entry for the index-3 block
 * has bit 15 (0x8000) set, and each set of 8 index-3 entries is preceded by
 * an additional uint16_t word. Data block offsets are 18 bits wide, with the top 2 bits stored
 * in the additional word.
 *
 * See ucptrie_internalSmallIndex() for details.
 *
 * (In a "small" trie, this works for ASCII and below-fast_limit code points as well.)
 *
 * Compaction:
 *
 * Multiple code point ranges ("blocks") that are aligned on certain boundaries
 * (determined by the shifting/bit fields of code points) and
 * map to the same data values normally share a single subsequence of the data array.
 * Data blocks can also overlap partially.
 * (Depending on the builder code finding duplicate and overlapping blocks.)
 *
 * Iteration over same-value ranges:
 *
 * Range iteration (ucptrie_getRange()) walks the structure from a start code point
 * until some code point is found that maps to a different value;
 * the end of the returned range is just before that.
 *
 * The header.dataNullOffset (split across two header fields, high bits in header.options)
 * is the offset of a widely shared data block filled with one single value.
 * It helps quickly skip over large ranges of data with that value.
 * The builder must ensure that if the start of any data block (fast or small)
 * matches the dataNullOffset, then the whole block must be filled with the null value.
 * Special care must be taken if there is no fast null data block
 * but a small one, which is shorter, and it matches the *start* of some fast data block.
 *
 * Similarly, the header.index3NullOffset is the index-array offset of an index-3 block
 * where all index entries point to the dataNullOffset.
 * If there is no such data or index-3 block, then these offsets are set to
 * values that cannot be reached (data offset out of range/reserved index offset),
 * normally UCPTRIE_NO_DATA_NULL_OFFSET or UCPTRIE_NO_INDEX3_NULL_OFFSET respectively.
 */





U_NAMESPACE_BEGIN

namespace {

/**
 * UTF-8 lead byte for minNoMaybeCP.
 * Can be lower than the actual lead byte for c.
 * Typically U+0300 for NFC/NFD, U+00A0 for NFKC/NFKD, U+0041 for NFKC_Casefold.
 */
inline uint8_t leadByteForCP(UChar32 c) {
    if (c <= 0x7f) {
        return (uint8_t)c;
    } else if (c <= 0x7ff) {
        return (uint8_t)(0xc0+(c>>6));
    } else {
        // Should not occur because ccc(U+0300)!=0.
        return 0xe0;
    }
}

/**
 * Returns the code point from one single well-formed UTF-8 byte sequence
 * between cpStart and cpLimit.
 *
 * Trie UTF-8 macros do not assemble whole code points (for efficiency).
 * When we do need the code point, we call this function.
 * We should not need it for normalization-inert data (norm16==0).
 * Illegal sequences yield the error value norm16==0 just like real normalization-inert code points.
 */
UChar32 codePointFromValidUTF8(const uint8_t *cpStart, const uint8_t *cpLimit) {
    // Similar to U8_NEXT_UNSAFE(s, i, c).
    U_ASSERT(cpStart < cpLimit);
    uint8_t c = *cpStart;
    switch(cpLimit-cpStart) {
    case 1:
        return c;
    case 2:
        return ((c&0x1f)<<6) | (cpStart[1]&0x3f);
    case 3:
        // no need for (c&0xf) because the upper bits are truncated after <<12 in the cast to (UChar)
        return (UChar)((c<<12) | ((cpStart[1]&0x3f)<<6) | (cpStart[2]&0x3f));
    case 4:
        return ((c&7)<<18) | ((cpStart[1]&0x3f)<<12) | ((cpStart[2]&0x3f)<<6) | (cpStart[3]&0x3f);
    default:
        UPRV_UNREACHABLE;  // Should not occur.
    }
}

/**
 * Returns the last code point in [start, p[ if it is valid and in U+1000..U+D7FF.
 * Otherwise returns a negative value.
 */
UChar32 previousHangulOrJamo(const uint8_t *start, const uint8_t *p) {
    if ((p - start) >= 3) {
        p -= 3;
        uint8_t l = *p;
        uint8_t t1, t2;
        if (0xe1 <= l && l <= 0xed &&
                (t1 = (uint8_t)(p[1] - 0x80)) <= 0x3f &&
                (t2 = (uint8_t)(p[2] - 0x80)) <= 0x3f &&
                (l < 0xed || t1 <= 0x1f)) {
            return ((l & 0xf) << 12) | (t1 << 6) | t2;
        }
    }
    return U_SENTINEL;
}

/**
 * Returns the offset from the Jamo T base if [src, limit[ starts with a single Jamo T code point.
 * Otherwise returns a negative value.
 */
int32_t getJamoTMinusBase(const uint8_t *src, const uint8_t *limit) {
    // Jamo T: E1 86 A8..E1 87 82
    if ((limit - src) >= 3 && *src == 0xe1) {
        if (src[1] == 0x86) {
            uint8_t t = src[2];
            // The first Jamo T is U+11A8 but JAMO_T_BASE is 11A7.
            // Offset 0 does not correspond to any conjoining Jamo.
            if (0xa8 <= t && t <= 0xbf) {
                return t - 0xa7;
            }
        } else if (src[1] == 0x87) {
            uint8_t t = src[2];
            if ((int8_t)t <= (int8_t)0x82u) {
                return t - (0xa7 - 0x40);
            }
        }
    }
    return -1;
}

void
appendCodePointDelta(const uint8_t *cpStart, const uint8_t *cpLimit, int32_t delta,
                     ByteSink &sink, Edits *edits) {
    char buffer[U8_MAX_LENGTH];
    int32_t length;
    int32_t cpLength = (int32_t)(cpLimit - cpStart);
    if (cpLength == 1) {
        // The builder makes ASCII map to ASCII.
        buffer[0] = (uint8_t)(*cpStart + delta);
        length = 1;
    } else {
        int32_t trail = *(cpLimit-1) + delta;
        if (0x80 <= trail && trail <= 0xbf) {
            // The delta only changes the last trail byte.
            --cpLimit;
            length = 0;
            do { buffer[length++] = *cpStart++; } while (cpStart < cpLimit);
            buffer[length++] = (uint8_t)trail;
        } else {
            // Decode the code point, add the delta, re-encode.
            UChar32 c = codePointFromValidUTF8(cpStart, cpLimit) + delta;
            length = 0;
            U8_APPEND_UNSAFE(buffer, length, c);
        }
    }
    if (edits != nullptr) {
        edits->addReplace(cpLength, length);
    }
    sink.Append(buffer, length);
}

}  // namespace

// ReorderingBuffer -------------------------------------------------------- ***

ReorderingBuffer::ReorderingBuffer(const Normalizer2Impl &ni, UnicodeString &dest,
                                   UErrorCode &errorCode) :
        impl(ni), str(dest),
        start(str.getBuffer(8)), reorderStart(start), limit(start),
        remainingCapacity(str.getCapacity()), lastCC(0) {
    if (start == nullptr && U_SUCCESS(errorCode)) {
        // getBuffer() already did str.setToBogus()
        errorCode = U_MEMORY_ALLOCATION_ERROR;
    }
}

UBool ReorderingBuffer::init(int32_t destCapacity, UErrorCode &errorCode) {
    int32_t length=str.length();
    start=str.getBuffer(destCapacity);
    if(start==NULL) {
        // getBuffer() already did str.setToBogus()
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    limit=start+length;
    remainingCapacity=str.getCapacity()-length;
    reorderStart=start;
    if(start==limit) {
        lastCC=0;
    } else {
        setIterator();
        lastCC=previousCC();
        // Set reorderStart after the last code point with cc<=1 if there is one.
        if(lastCC>1) {
            while(previousCC()>1) {}
        }
        reorderStart=codePointLimit;
    }
    return TRUE;
}

UBool ReorderingBuffer::equals(const UChar *otherStart, const UChar *otherLimit) const {
    int32_t length=(int32_t)(limit-start);
    return
        length==(int32_t)(otherLimit-otherStart) &&
        0==u_memcmp(start, otherStart, length);
}

UBool ReorderingBuffer::equals(const uint8_t *otherStart, const uint8_t *otherLimit) const {
    U_ASSERT((otherLimit - otherStart) <= INT32_MAX);  // ensured by caller
    int32_t length = (int32_t)(limit - start);
    int32_t otherLength = (int32_t)(otherLimit - otherStart);
    // For equal strings, UTF-8 is at least as long as UTF-16, and at most three times as long.
    if (otherLength < length || (otherLength / 3) > length) {
        return FALSE;
    }
    // Compare valid strings from between normalization boundaries.
    // (Invalid sequences are normalization-inert.)
    for (int32_t i = 0, j = 0;;) {
        if (i >= length) {
            return j >= otherLength;
        } else if (j >= otherLength) {
            return FALSE;
        }
        // Not at the end of either string yet.
        UChar32 c, other;
        U16_NEXT_UNSAFE(start, i, c);
        U8_NEXT_UNSAFE(otherStart, j, other);
        if (c != other) {
            return FALSE;
        }
    }
}

UBool ReorderingBuffer::appendSupplementary(UChar32 c, uint8_t cc, UErrorCode &errorCode) {
    if(remainingCapacity<2 && !resize(2, errorCode)) {
        return FALSE;
    }
    if(lastCC<=cc || cc==0) {
        limit[0]=U16_LEAD(c);
        limit[1]=U16_TRAIL(c);
        limit+=2;
        lastCC=cc;
        if(cc<=1) {
            reorderStart=limit;
        }
    } else {
        insert(c, cc);
    }
    remainingCapacity-=2;
    return TRUE;
}

UBool ReorderingBuffer::append(const UChar *s, int32_t length, UBool isNFD,
                               uint8_t leadCC, uint8_t trailCC,
                               UErrorCode &errorCode) {
    if(length==0) {
        return TRUE;
    }
    if(remainingCapacity<length && !resize(length, errorCode)) {
        return FALSE;
    }
    remainingCapacity-=length;
    if(lastCC<=leadCC || leadCC==0) {
        if(trailCC<=1) {
            reorderStart=limit+length;
        } else if(leadCC<=1) {
            reorderStart=limit+1;  // Ok if not a code point boundary.
        }
        const UChar *sLimit=s+length;
        do { *limit++=*s++; } while(s!=sLimit);
        lastCC=trailCC;
    } else {
        int32_t i=0;
        UChar32 c;
        U16_NEXT(s, i, length, c);
        insert(c, leadCC);  // insert first code point
        while(i<length) {
            U16_NEXT(s, i, length, c);
            if(i<length) {
                if (isNFD) {
                    leadCC = Normalizer2Impl::getCCFromYesOrMaybe(impl.getRawNorm16(c));
                } else {
                    leadCC = impl.getCC(impl.getNorm16(c));
                }
            } else {
                leadCC=trailCC;
            }
            append(c, leadCC, errorCode);
        }
    }
    return TRUE;
}

UBool ReorderingBuffer::appendZeroCC(UChar32 c, UErrorCode &errorCode) {
    int32_t cpLength=U16_LENGTH(c);
    if(remainingCapacity<cpLength && !resize(cpLength, errorCode)) {
        return FALSE;
    }
    remainingCapacity-=cpLength;
    if(cpLength==1) {
        *limit++=(UChar)c;
    } else {
        limit[0]=U16_LEAD(c);
        limit[1]=U16_TRAIL(c);
        limit+=2;
    }
    lastCC=0;
    reorderStart=limit;
    return TRUE;
}

UBool ReorderingBuffer::appendZeroCC(const UChar *s, const UChar *sLimit, UErrorCode &errorCode) {
    if(s==sLimit) {
        return TRUE;
    }
    int32_t length=(int32_t)(sLimit-s);
    if(remainingCapacity<length && !resize(length, errorCode)) {
        return FALSE;
    }
    u_memcpy(limit, s, length);
    limit+=length;
    remainingCapacity-=length;
    lastCC=0;
    reorderStart=limit;
    return TRUE;
}

void ReorderingBuffer::remove() {
    reorderStart=limit=start;
    remainingCapacity=str.getCapacity();
    lastCC=0;
}

void ReorderingBuffer::removeSuffix(int32_t suffixLength) {
    if(suffixLength<(limit-start)) {
        limit-=suffixLength;
        remainingCapacity+=suffixLength;
    } else {
        limit=start;
        remainingCapacity=str.getCapacity();
    }
    lastCC=0;
    reorderStart=limit;
}

UBool ReorderingBuffer::resize(int32_t appendLength, UErrorCode &errorCode) {
    int32_t reorderStartIndex=(int32_t)(reorderStart-start);
    int32_t length=(int32_t)(limit-start);
    str.releaseBuffer(length);
    int32_t newCapacity=length+appendLength;
    int32_t doubleCapacity=2*str.getCapacity();
    if(newCapacity<doubleCapacity) {
        newCapacity=doubleCapacity;
    }
    if(newCapacity<256) {
        newCapacity=256;
    }
    start=str.getBuffer(newCapacity);
    if(start==NULL) {
        // getBuffer() already did str.setToBogus()
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    reorderStart=start+reorderStartIndex;
    limit=start+length;
    remainingCapacity=str.getCapacity()-length;
    return TRUE;
}

void ReorderingBuffer::skipPrevious() {
    codePointLimit=codePointStart;
    UChar c=*--codePointStart;
    if(U16_IS_TRAIL(c) && start<codePointStart && U16_IS_LEAD(*(codePointStart-1))) {
        --codePointStart;
    }
}

uint8_t ReorderingBuffer::previousCC() {
    codePointLimit=codePointStart;
    if(reorderStart>=codePointStart) {
        return 0;
    }
    UChar32 c=*--codePointStart;
    UChar c2;
    if(U16_IS_TRAIL(c) && start<codePointStart && U16_IS_LEAD(c2=*(codePointStart-1))) {
        --codePointStart;
        c=U16_GET_SUPPLEMENTARY(c2, c);
    }
    return impl.getCCFromYesOrMaybeCP(c);
}

// Inserts c somewhere before the last character.
// Requires 0<cc<lastCC which implies reorderStart<limit.
void ReorderingBuffer::insert(UChar32 c, uint8_t cc) {
    for(setIterator(), skipPrevious(); previousCC()>cc;) {}
    // insert c at codePointLimit, after the character with prevCC<=cc
    UChar *q=limit;
    UChar *r=limit+=U16_LENGTH(c);
    do {
        *--r=*--q;
    } while(codePointLimit!=q);
    writeCodePoint(q, c);
    if(cc<=1) {
        reorderStart=r;
    }
}

// Normalizer2Impl --------------------------------------------------------- ***

struct CanonIterData : public UMemory {
    CanonIterData(UErrorCode &errorCode);
    ~CanonIterData();
    void addToStartSet(UChar32 origin, UChar32 decompLead, UErrorCode &errorCode);
    UMutableCPTrie *mutableTrie;
    UCPTrie *trie;
    UVector canonStartSets;  // contains UnicodeSet *
};

Normalizer2Impl::~Normalizer2Impl() {
    delete fCanonIterData;
}

void
Normalizer2Impl::init(const int32_t *inIndexes, const UCPTrie *inTrie,
                      const uint16_t *inExtraData, const uint8_t *inSmallFCD) {
    minDecompNoCP = static_cast<UChar>(inIndexes[IX_MIN_DECOMP_NO_CP]);
    minCompNoMaybeCP = static_cast<UChar>(inIndexes[IX_MIN_COMP_NO_MAYBE_CP]);
    minLcccCP = static_cast<UChar>(inIndexes[IX_MIN_LCCC_CP]);

    minYesNo = static_cast<uint16_t>(inIndexes[IX_MIN_YES_NO]);
    minYesNoMappingsOnly = static_cast<uint16_t>(inIndexes[IX_MIN_YES_NO_MAPPINGS_ONLY]);
    minNoNo = static_cast<uint16_t>(inIndexes[IX_MIN_NO_NO]);
    minNoNoCompBoundaryBefore = static_cast<uint16_t>(inIndexes[IX_MIN_NO_NO_COMP_BOUNDARY_BEFORE]);
    minNoNoCompNoMaybeCC = static_cast<uint16_t>(inIndexes[IX_MIN_NO_NO_COMP_NO_MAYBE_CC]);
    minNoNoEmpty = static_cast<uint16_t>(inIndexes[IX_MIN_NO_NO_EMPTY]);
    limitNoNo = static_cast<uint16_t>(inIndexes[IX_LIMIT_NO_NO]);
    minMaybeYes = static_cast<uint16_t>(inIndexes[IX_MIN_MAYBE_YES]);
    U_ASSERT((minMaybeYes & 7) == 0);  // 8-aligned for noNoDelta bit fields
    centerNoNoDelta = (minMaybeYes >> DELTA_SHIFT) - MAX_DELTA - 1;

    normTrie=inTrie;

    maybeYesCompositions=inExtraData;
    extraData=maybeYesCompositions+((MIN_NORMAL_MAYBE_YES-minMaybeYes)>>OFFSET_SHIFT);

    smallFCD=inSmallFCD;
}

U_CDECL_BEGIN

static uint32_t U_CALLCONV
segmentStarterMapper(const void * /*context*/, uint32_t value) {
    return value&CANON_NOT_SEGMENT_STARTER;
}

U_CDECL_END

void
Normalizer2Impl::addLcccChars(UnicodeSet &set) const {
    UChar32 start = 0, end;
    uint32_t norm16;
    while ((end = ucptrie_getRange(normTrie, start, UCPMAP_RANGE_FIXED_LEAD_SURROGATES, INERT,
                                   nullptr, nullptr, &norm16)) >= 0) {
        if (norm16 > Normalizer2Impl::MIN_NORMAL_MAYBE_YES &&
                norm16 != Normalizer2Impl::JAMO_VT) {
            set.add(start, end);
        } else if (minNoNoCompNoMaybeCC <= norm16 && norm16 < limitNoNo) {
            uint16_t fcd16 = getFCD16(start);
            if (fcd16 > 0xff) { set.add(start, end); }
        }
        start = end + 1;
    }
}

void
Normalizer2Impl::addPropertyStarts(const USetAdder *sa, UErrorCode & /*errorCode*/) const {
    // Add the start code point of each same-value range of the trie.
    UChar32 start = 0, end;
    uint32_t value;
    while ((end = ucptrie_getRange(normTrie, start, UCPMAP_RANGE_FIXED_LEAD_SURROGATES, INERT,
                                   nullptr, nullptr, &value)) >= 0) {
        sa->add(sa->set, start);
        if (start != end && isAlgorithmicNoNo((uint16_t)value) &&
                (value & Normalizer2Impl::DELTA_TCCC_MASK) > Normalizer2Impl::DELTA_TCCC_1) {
            // Range of code points with same-norm16-value algorithmic decompositions.
            // They might have different non-zero FCD16 values.
            uint16_t prevFCD16 = getFCD16(start);
            while (++start <= end) {
                uint16_t fcd16 = getFCD16(start);
                if (fcd16 != prevFCD16) {
                    sa->add(sa->set, start);
                    prevFCD16 = fcd16;
                }
            }
        }
        start = end + 1;
    }

    /* add Hangul LV syllables and LV+1 because of skippables */
    for(UChar c=Hangul::HANGUL_BASE; c<Hangul::HANGUL_LIMIT; c+=Hangul::JAMO_T_COUNT) {
        sa->add(sa->set, c);
        sa->add(sa->set, c+1);
    }
    sa->add(sa->set, Hangul::HANGUL_LIMIT); /* add Hangul+1 to continue with other properties */
}

void
Normalizer2Impl::addCanonIterPropertyStarts(const USetAdder *sa, UErrorCode &errorCode) const {
    // Add the start code point of each same-value range of the canonical iterator data trie.
    if (!ensureCanonIterData(errorCode)) { return; }
    // Currently only used for the SEGMENT_STARTER property.
    UChar32 start = 0, end;
    uint32_t value;
    while ((end = ucptrie_getRange(fCanonIterData->trie, start, UCPMAP_RANGE_NORMAL, 0,
                                   segmentStarterMapper, nullptr, &value)) >= 0) {
        sa->add(sa->set, start);
        start = end + 1;
    }
}

const UChar *
Normalizer2Impl::copyLowPrefixFromNulTerminated(const UChar *src,
                                                UChar32 minNeedDataCP,
                                                ReorderingBuffer *buffer,
                                                UErrorCode &errorCode) const {
    // Make some effort to support NUL-terminated strings reasonably.
    // Take the part of the fast quick check loop that does not look up
    // data and check the first part of the string.
    // After this prefix, determine the string length to simplify the rest
    // of the code.
    const UChar *prevSrc=src;
    UChar c;
    while((c=*src++)<minNeedDataCP && c!=0) {}
    // Back out the last character for full processing.
    // Copy this prefix.
    if(--src!=prevSrc) {
        if(buffer!=NULL) {
            buffer->appendZeroCC(prevSrc, src, errorCode);
        }
    }
    return src;
}

UnicodeString &
Normalizer2Impl::decompose(const UnicodeString &src, UnicodeString &dest,
                           UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        dest.setToBogus();
        return dest;
    }
    const UChar *sArray=src.getBuffer();
    if(&dest==&src || sArray==NULL) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        dest.setToBogus();
        return dest;
    }
    decompose(sArray, sArray+src.length(), dest, src.length(), errorCode);
    return dest;
}

void
Normalizer2Impl::decompose(const UChar *src, const UChar *limit,
                           UnicodeString &dest,
                           int32_t destLengthEstimate,
                           UErrorCode &errorCode) const {
    if(destLengthEstimate<0 && limit!=NULL) {
        destLengthEstimate=(int32_t)(limit-src);
    }
    dest.remove();
    ReorderingBuffer buffer(*this, dest);
    if(buffer.init(destLengthEstimate, errorCode)) {
        decompose(src, limit, &buffer, errorCode);
    }
}

// Dual functionality:
// buffer!=NULL: normalize
// buffer==NULL: isNormalized/spanQuickCheckYes
const UChar *
Normalizer2Impl::decompose(const UChar *src, const UChar *limit,
                           ReorderingBuffer *buffer,
                           UErrorCode &errorCode) const {
    UChar32 minNoCP=minDecompNoCP;
    if(limit==NULL) {
        src=copyLowPrefixFromNulTerminated(src, minNoCP, buffer, errorCode);
        if(U_FAILURE(errorCode)) {
            return src;
        }
        limit=u_strchr(src, 0);
    }

    const UChar *prevSrc;
    UChar32 c=0;
    uint16_t norm16=0;

    // only for quick check
    const UChar *prevBoundary=src;
    uint8_t prevCC=0;

    for(;;) {
        // count code units below the minimum or with irrelevant data for the quick check
        for(prevSrc=src; src!=limit;) {
            if( (c=*src)<minNoCP ||
                isMostDecompYesAndZeroCC(norm16=UCPTRIE_FAST_BMP_GET(normTrie, UCPTRIE_16, c))
            ) {
                ++src;
            } else if(!U16_IS_LEAD(c)) {
                break;
            } else {
                UChar c2;
                if((src+1)!=limit && U16_IS_TRAIL(c2=src[1])) {
                    c=U16_GET_SUPPLEMENTARY(c, c2);
                    norm16=UCPTRIE_FAST_SUPP_GET(normTrie, UCPTRIE_16, c);
                    if(isMostDecompYesAndZeroCC(norm16)) {
                        src+=2;
                    } else {
                        break;
                    }
                } else {
                    ++src;  // unpaired lead surrogate: inert
                }
            }
        }
        // copy these code units all at once
        if(src!=prevSrc) {
            if(buffer!=NULL) {
                if(!buffer->appendZeroCC(prevSrc, src, errorCode)) {
                    break;
                }
            } else {
                prevCC=0;
                prevBoundary=src;
            }
        }
        if(src==limit) {
            break;
        }

        // Check one above-minimum, relevant code point.
        src+=U16_LENGTH(c);
        if(buffer!=NULL) {
            if(!decompose(c, norm16, *buffer, errorCode)) {
                break;
            }
        } else {
            if(isDecompYes(norm16)) {
                uint8_t cc=getCCFromYesOrMaybe(norm16);
                if(prevCC<=cc || cc==0) {
                    prevCC=cc;
                    if(cc<=1) {
                        prevBoundary=src;
                    }
                    continue;
                }
            }
            return prevBoundary;  // "no" or cc out of order
        }
    }
    return src;
}

// Decompose a short piece of text which is likely to contain characters that
// fail the quick check loop and/or where the quick check loop's overhead
// is unlikely to be amortized.
// Called by the compose() and makeFCD() implementations.
const UChar *
Normalizer2Impl::decomposeShort(const UChar *src, const UChar *limit,
                                UBool stopAtCompBoundary, UBool onlyContiguous,
                                ReorderingBuffer &buffer, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) {
        return nullptr;
    }
    while(src<limit) {
        if (stopAtCompBoundary && *src < minCompNoMaybeCP) {
            return src;
        }
        const UChar *prevSrc = src;
        UChar32 c;
        uint16_t norm16;
        UCPTRIE_FAST_U16_NEXT(normTrie, UCPTRIE_16, src, limit, c, norm16);
        if (stopAtCompBoundary && norm16HasCompBoundaryBefore(norm16)) {
            return prevSrc;
        }
        if(!decompose(c, norm16, buffer, errorCode)) {
            return nullptr;
        }
        if (stopAtCompBoundary && norm16HasCompBoundaryAfter(norm16, onlyContiguous)) {
            return src;
        }
    }
    return src;
}

UBool Normalizer2Impl::decompose(UChar32 c, uint16_t norm16,
                                 ReorderingBuffer &buffer,
                                 UErrorCode &errorCode) const {
    // get the decomposition and the lead and trail cc's
    if (norm16 >= limitNoNo) {
        if (isMaybeOrNonZeroCC(norm16)) {
            return buffer.append(c, getCCFromYesOrMaybe(norm16), errorCode);
        }
        // Maps to an isCompYesAndZeroCC.
        c=mapAlgorithmic(c, norm16);
        norm16=getRawNorm16(c);
    }
    if (norm16 < minYesNo) {
        // c does not decompose
        return buffer.append(c, 0, errorCode);
    } else if(isHangulLV(norm16) || isHangulLVT(norm16)) {
        // Hangul syllable: decompose algorithmically
        UChar jamos[3];
        return buffer.appendZeroCC(jamos, jamos+Hangul::decompose(c, jamos), errorCode);
    }
    // c decomposes, get everything from the variable-length extra data
    const uint16_t *mapping=getMapping(norm16);
    uint16_t firstUnit=*mapping;
    int32_t length=firstUnit&MAPPING_LENGTH_MASK;
    uint8_t leadCC, trailCC;
    trailCC=(uint8_t)(firstUnit>>8);
    if(firstUnit&MAPPING_HAS_CCC_LCCC_WORD) {
        leadCC=(uint8_t)(*(mapping-1)>>8);
    } else {
        leadCC=0;
    }
    return buffer.append((const UChar *)mapping+1, length, TRUE, leadCC, trailCC, errorCode);
}

const uint8_t *
Normalizer2Impl::decomposeShort(const uint8_t *src, const uint8_t *limit,
                                UBool stopAtCompBoundary, UBool onlyContiguous,
                                ReorderingBuffer &buffer, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) {
        return nullptr;
    }
    while (src < limit) {
        const uint8_t *prevSrc = src;
        uint16_t norm16;
        UCPTRIE_FAST_U8_NEXT(normTrie, UCPTRIE_16, src, limit, norm16);
        // Get the decomposition and the lead and trail cc's.
        UChar32 c = U_SENTINEL;
        if (norm16 >= limitNoNo) {
            if (isMaybeOrNonZeroCC(norm16)) {
                // No boundaries around this character.
                c = codePointFromValidUTF8(prevSrc, src);
                if (!buffer.append(c, getCCFromYesOrMaybe(norm16), errorCode)) {
                    return nullptr;
                }
                continue;
            }
            // Maps to an isCompYesAndZeroCC.
            if (stopAtCompBoundary) {
                return prevSrc;
            }
            c = codePointFromValidUTF8(prevSrc, src);
            c = mapAlgorithmic(c, norm16);
            norm16 = getRawNorm16(c);
        } else if (stopAtCompBoundary && norm16 < minNoNoCompNoMaybeCC) {
            return prevSrc;
        }
        // norm16!=INERT guarantees that [prevSrc, src[ is valid UTF-8.
        // We do not see invalid UTF-8 here because
        // its norm16==INERT is normalization-inert,
        // so it gets copied unchanged in the fast path,
        // and we stop the slow path where invalid UTF-8 begins.
        U_ASSERT(norm16 != INERT);
        if (norm16 < minYesNo) {
            if (c < 0) {
                c = codePointFromValidUTF8(prevSrc, src);
            }
            // does not decompose
            if (!buffer.append(c, 0, errorCode)) {
                return nullptr;
            }
        } else if (isHangulLV(norm16) || isHangulLVT(norm16)) {
            // Hangul syllable: decompose algorithmically
            if (c < 0) {
                c = codePointFromValidUTF8(prevSrc, src);
            }
            char16_t jamos[3];
            if (!buffer.appendZeroCC(jamos, jamos+Hangul::decompose(c, jamos), errorCode)) {
                return nullptr;
            }
        } else {
            // The character decomposes, get everything from the variable-length extra data.
            const uint16_t *mapping = getMapping(norm16);
            uint16_t firstUnit = *mapping;
            int32_t length = firstUnit & MAPPING_LENGTH_MASK;
            uint8_t trailCC = (uint8_t)(firstUnit >> 8);
            uint8_t leadCC;
            if (firstUnit & MAPPING_HAS_CCC_LCCC_WORD) {
                leadCC = (uint8_t)(*(mapping-1) >> 8);
            } else {
                leadCC = 0;
            }
            if (!buffer.append((const char16_t *)mapping+1, length, TRUE, leadCC, trailCC, errorCode)) {
                return nullptr;
            }
        }
        if (stopAtCompBoundary && norm16HasCompBoundaryAfter(norm16, onlyContiguous)) {
            return src;
        }
    }
    return src;
}

const UChar *
Normalizer2Impl::getDecomposition(UChar32 c, UChar buffer[4], int32_t &length) const {
    uint16_t norm16;
    if(c<minDecompNoCP || isMaybeOrNonZeroCC(norm16=getNorm16(c))) {
        // c does not decompose
        return nullptr;
    }
    const UChar *decomp = nullptr;
    if(isDecompNoAlgorithmic(norm16)) {
        // Maps to an isCompYesAndZeroCC.
        c=mapAlgorithmic(c, norm16);
        decomp=buffer;
        length=0;
        U16_APPEND_UNSAFE(buffer, length, c);
        // The mapping might decompose further.
        norm16 = getRawNorm16(c);
    }
    if (norm16 < minYesNo) {
        return decomp;
    } else if(isHangulLV(norm16) || isHangulLVT(norm16)) {
        // Hangul syllable: decompose algorithmically
        length=Hangul::decompose(c, buffer);
        return buffer;
    }
    // c decomposes, get everything from the variable-length extra data
    const uint16_t *mapping=getMapping(norm16);
    length=*mapping&MAPPING_LENGTH_MASK;
    return (const UChar *)mapping+1;
}

// The capacity of the buffer must be 30=MAPPING_LENGTH_MASK-1
// so that a raw mapping fits that consists of one unit ("rm0")
// plus all but the first two code units of the normal mapping.
// The maximum length of a normal mapping is 31=MAPPING_LENGTH_MASK.
const UChar *
Normalizer2Impl::getRawDecomposition(UChar32 c, UChar buffer[30], int32_t &length) const {
    uint16_t norm16;
    if(c<minDecompNoCP || isDecompYes(norm16=getNorm16(c))) {
        // c does not decompose
        return NULL;
    } else if(isHangulLV(norm16) || isHangulLVT(norm16)) {
        // Hangul syllable: decompose algorithmically
        Hangul::getRawDecomposition(c, buffer);
        length=2;
        return buffer;
    } else if(isDecompNoAlgorithmic(norm16)) {
        c=mapAlgorithmic(c, norm16);
        length=0;
        U16_APPEND_UNSAFE(buffer, length, c);
        return buffer;
    }
    // c decomposes, get everything from the variable-length extra data
    const uint16_t *mapping=getMapping(norm16);
    uint16_t firstUnit=*mapping;
    int32_t mLength=firstUnit&MAPPING_LENGTH_MASK;  // length of normal mapping
    if(firstUnit&MAPPING_HAS_RAW_MAPPING) {
        // Read the raw mapping from before the firstUnit and before the optional ccc/lccc word.
        // Bit 7=MAPPING_HAS_CCC_LCCC_WORD
        const uint16_t *rawMapping=mapping-((firstUnit>>7)&1)-1;
        uint16_t rm0=*rawMapping;
        if(rm0<=MAPPING_LENGTH_MASK) {
            length=rm0;
            return (const UChar *)rawMapping-rm0;
        } else {
            // Copy the normal mapping and replace its first two code units with rm0.
            buffer[0]=(UChar)rm0;
            u_memcpy(buffer+1, (const UChar *)mapping+1+2, mLength-2);
            length=mLength-1;
            return buffer;
        }
    } else {
        length=mLength;
        return (const UChar *)mapping+1;
    }
}

void Normalizer2Impl::decomposeAndAppend(const UChar *src, const UChar *limit,
                                         UBool doDecompose,
                                         UnicodeString &safeMiddle,
                                         ReorderingBuffer &buffer,
                                         UErrorCode &errorCode) const {
    buffer.copyReorderableSuffixTo(safeMiddle);
    if(doDecompose) {
        decompose(src, limit, &buffer, errorCode);
        return;
    }
    // Just merge the strings at the boundary.
    bool isFirst = true;
    uint8_t firstCC = 0, prevCC = 0, cc;
    const UChar *p = src;
    while (p != limit) {
        const UChar *codePointStart = p;
        UChar32 c;
        uint16_t norm16;
        UCPTRIE_FAST_U16_NEXT(normTrie, UCPTRIE_16, p, limit, c, norm16);
        if ((cc = getCC(norm16)) == 0) {
            p = codePointStart;
            break;
        }
        if (isFirst) {
            firstCC = cc;
            isFirst = false;
        }
        prevCC = cc;
    }
    if(limit==NULL) {  // appendZeroCC() needs limit!=NULL
        limit=u_strchr(p, 0);
    }

    if (buffer.append(src, (int32_t)(p - src), FALSE, firstCC, prevCC, errorCode)) {
        buffer.appendZeroCC(p, limit, errorCode);
    }
}

UBool Normalizer2Impl::hasDecompBoundaryBefore(UChar32 c) const {
    return c < minLcccCP || (c <= 0xffff && !singleLeadMightHaveNonZeroFCD16(c)) ||
        norm16HasDecompBoundaryBefore(getNorm16(c));
}

UBool Normalizer2Impl::norm16HasDecompBoundaryBefore(uint16_t norm16) const {
    if (norm16 < minNoNoCompNoMaybeCC) {
        return TRUE;
    }
    if (norm16 >= limitNoNo) {
        return norm16 <= MIN_NORMAL_MAYBE_YES || norm16 == JAMO_VT;
    }
    // c decomposes, get everything from the variable-length extra data
    const uint16_t *mapping=getMapping(norm16);
    uint16_t firstUnit=*mapping;
    // TRUE if leadCC==0 (hasFCDBoundaryBefore())
    return (firstUnit&MAPPING_HAS_CCC_LCCC_WORD)==0 || (*(mapping-1)&0xff00)==0;
}

UBool Normalizer2Impl::hasDecompBoundaryAfter(UChar32 c) const {
    if (c < minDecompNoCP) {
        return TRUE;
    }
    if (c <= 0xffff && !singleLeadMightHaveNonZeroFCD16(c)) {
        return TRUE;
    }
    return norm16HasDecompBoundaryAfter(getNorm16(c));
}

UBool Normalizer2Impl::norm16HasDecompBoundaryAfter(uint16_t norm16) const {
    if(norm16 <= minYesNo || isHangulLVT(norm16)) {
        return TRUE;
    }
    if (norm16 >= limitNoNo) {
        if (isMaybeOrNonZeroCC(norm16)) {
            return norm16 <= MIN_NORMAL_MAYBE_YES || norm16 == JAMO_VT;
        }
        // Maps to an isCompYesAndZeroCC.
        return (norm16 & DELTA_TCCC_MASK) <= DELTA_TCCC_1;
    }
    // c decomposes, get everything from the variable-length extra data
    const uint16_t *mapping=getMapping(norm16);
    uint16_t firstUnit=*mapping;
    // decomp after-boundary: same as hasFCDBoundaryAfter(),
    // fcd16<=1 || trailCC==0
    if(firstUnit>0x1ff) {
        return FALSE;  // trailCC>1
    }
    if(firstUnit<=0xff) {
        return TRUE;  // trailCC==0
    }
    // if(trailCC==1) test leadCC==0, same as checking for before-boundary
    // TRUE if leadCC==0 (hasFCDBoundaryBefore())
    return (firstUnit&MAPPING_HAS_CCC_LCCC_WORD)==0 || (*(mapping-1)&0xff00)==0;
}

/*
 * Finds the recomposition result for
 * a forward-combining "lead" character,
 * specified with a pointer to its compositions list,
 * and a backward-combining "trail" character.
 *
 * If the lead and trail characters combine, then this function returns
 * the following "compositeAndFwd" value:
 * Bits 21..1  composite character
 * Bit      0  set if the composite is a forward-combining starter
 * otherwise it returns -1.
 *
 * The compositions list has (trail, compositeAndFwd) pair entries,
 * encoded as either pairs or triples of 16-bit units.
 * The last entry has the high bit of its first unit set.
 *
 * The list is sorted by ascending trail characters (there are no duplicates).
 * A linear search is used.
 *
 * See normalizer2impl.h for a more detailed description
 * of the compositions list format.
 */
int32_t Normalizer2Impl::combine(const uint16_t *list, UChar32 trail) {
    uint16_t key1, firstUnit;
    if(trail<COMP_1_TRAIL_LIMIT) {
        // trail character is 0..33FF
        // result entry may have 2 or 3 units
        key1=(uint16_t)(trail<<1);
        while(key1>(firstUnit=*list)) {
            list+=2+(firstUnit&COMP_1_TRIPLE);
        }
        if(key1==(firstUnit&COMP_1_TRAIL_MASK)) {
            if(firstUnit&COMP_1_TRIPLE) {
                return ((int32_t)list[1]<<16)|list[2];
            } else {
                return list[1];
            }
        }
    } else {
        // trail character is 3400..10FFFF
        // result entry has 3 units
        key1=(uint16_t)(COMP_1_TRAIL_LIMIT+
                        (((trail>>COMP_1_TRAIL_SHIFT))&
                          ~COMP_1_TRIPLE));
        uint16_t key2=(uint16_t)(trail<<COMP_2_TRAIL_SHIFT);
        uint16_t secondUnit;
        for(;;) {
            if(key1>(firstUnit=*list)) {
                list+=2+(firstUnit&COMP_1_TRIPLE);
            } else if(key1==(firstUnit&COMP_1_TRAIL_MASK)) {
                if(key2>(secondUnit=list[1])) {
                    if(firstUnit&COMP_1_LAST_TUPLE) {
                        break;
                    } else {
                        list+=3;
                    }
                } else if(key2==(secondUnit&COMP_2_TRAIL_MASK)) {
                    return ((int32_t)(secondUnit&~COMP_2_TRAIL_MASK)<<16)|list[2];
                } else {
                    break;
                }
            } else {
                break;
            }
        }
    }
    return -1;
}

/**
  * @param list some character's compositions list
  * @param set recursively receives the composites from these compositions
  */
void Normalizer2Impl::addComposites(const uint16_t *list, UnicodeSet &set) const {
    uint16_t firstUnit;
    int32_t compositeAndFwd;
    do {
        firstUnit=*list;
        if((firstUnit&COMP_1_TRIPLE)==0) {
            compositeAndFwd=list[1];
            list+=2;
        } else {
            compositeAndFwd=(((int32_t)list[1]&~COMP_2_TRAIL_MASK)<<16)|list[2];
            list+=3;
        }
        UChar32 composite=compositeAndFwd>>1;
        if((compositeAndFwd&1)!=0) {
            addComposites(getCompositionsListForComposite(getRawNorm16(composite)), set);
        }
        set.add(composite);
    } while((firstUnit&COMP_1_LAST_TUPLE)==0);
}

/*
 * Recomposes the buffer text starting at recomposeStartIndex
 * (which is in NFD - decomposed and canonically ordered),
 * and truncates the buffer contents.
 *
 * Note that recomposition never lengthens the text:
 * Any character consists of either one or two code units;
 * a composition may contain at most one more code unit than the original starter,
 * while the combining mark that is removed has at least one code unit.
 */
void Normalizer2Impl::recompose(ReorderingBuffer &buffer, int32_t recomposeStartIndex,
                                UBool onlyContiguous) const {
    UChar *p=buffer.getStart()+recomposeStartIndex;
    UChar *limit=buffer.getLimit();
    if(p==limit) {
        return;
    }

    UChar *starter, *pRemove, *q, *r;
    const uint16_t *compositionsList;
    UChar32 c, compositeAndFwd;
    uint16_t norm16;
    uint8_t cc, prevCC;
    UBool starterIsSupplementary;

    // Some of the following variables are not used until we have a forward-combining starter
    // and are only initialized now to avoid compiler warnings.
    compositionsList=NULL;  // used as indicator for whether we have a forward-combining starter
    starter=NULL;
    starterIsSupplementary=FALSE;
    prevCC=0;

    for(;;) {
        UCPTRIE_FAST_U16_NEXT(normTrie, UCPTRIE_16, p, limit, c, norm16);
        cc=getCCFromYesOrMaybe(norm16);
        if( // this character combines backward and
            isMaybe(norm16) &&
            // we have seen a starter that combines forward and
            compositionsList!=NULL &&
            // the backward-combining character is not blocked
            (prevCC<cc || prevCC==0)
        ) {
            if(isJamoVT(norm16)) {
                // c is a Jamo V/T, see if we can compose it with the previous character.
                if(c<Hangul::JAMO_T_BASE) {
                    // c is a Jamo Vowel, compose with previous Jamo L and following Jamo T.
                    UChar prev=(UChar)(*starter-Hangul::JAMO_L_BASE);
                    if(prev<Hangul::JAMO_L_COUNT) {
                        pRemove=p-1;
                        UChar syllable=(UChar)
                            (Hangul::HANGUL_BASE+
                             (prev*Hangul::JAMO_V_COUNT+(c-Hangul::JAMO_V_BASE))*
                             Hangul::JAMO_T_COUNT);
                        UChar t;
                        if(p!=limit && (t=(UChar)(*p-Hangul::JAMO_T_BASE))<Hangul::JAMO_T_COUNT) {
                            ++p;
                            syllable+=t;  // The next character was a Jamo T.
                        }
                        *starter=syllable;
                        // remove the Jamo V/T
                        q=pRemove;
                        r=p;
                        while(r<limit) {
                            *q++=*r++;
                        }
                        limit=q;
                        p=pRemove;
                    }
                }
                /*
                 * No "else" for Jamo T:
                 * Since the input is in NFD, there are no Hangul LV syllables that
                 * a Jamo T could combine with.
                 * All Jamo Ts are combined above when handling Jamo Vs.
                 */
                if(p==limit) {
                    break;
                }
                compositionsList=NULL;
                continue;
            } else if((compositeAndFwd=combine(compositionsList, c))>=0) {
                // The starter and the combining mark (c) do combine.
                UChar32 composite=compositeAndFwd>>1;

                // Replace the starter with the composite, remove the combining mark.
                pRemove=p-U16_LENGTH(c);  // pRemove & p: start & limit of the combining mark
                if(starterIsSupplementary) {
                    if(U_IS_SUPPLEMENTARY(composite)) {
                        // both are supplementary
                        starter[0]=U16_LEAD(composite);
                        starter[1]=U16_TRAIL(composite);
                    } else {
                        *starter=(UChar)composite;
                        // The composite is shorter than the starter,
                        // move the intermediate characters forward one.
                        starterIsSupplementary=FALSE;
                        q=starter+1;
                        r=q+1;
                        while(r<pRemove) {
                            *q++=*r++;
                        }
                        --pRemove;
                    }
                } else if(U_IS_SUPPLEMENTARY(composite)) {
                    // The composite is longer than the starter,
                    // move the intermediate characters back one.
                    starterIsSupplementary=TRUE;
                    ++starter;  // temporarily increment for the loop boundary
                    q=pRemove;
                    r=++pRemove;
                    while(starter<q) {
                        *--r=*--q;
                    }
                    *starter=U16_TRAIL(composite);
                    *--starter=U16_LEAD(composite);  // undo the temporary increment
                } else {
                    // both are on the BMP
                    *starter=(UChar)composite;
                }

                /* remove the combining mark by moving the following text over it */
                if(pRemove<p) {
                    q=pRemove;
                    r=p;
                    while(r<limit) {
                        *q++=*r++;
                    }
                    limit=q;
                    p=pRemove;
                }
                // Keep prevCC because we removed the combining mark.

                if(p==limit) {
                    break;
                }
                // Is the composite a starter that combines forward?
                if(compositeAndFwd&1) {
                    compositionsList=
                        getCompositionsListForComposite(getRawNorm16(composite));
                } else {
                    compositionsList=NULL;
                }

                // We combined; continue with looking for compositions.
                continue;
            }
        }

        // no combination this time
        prevCC=cc;
        if(p==limit) {
            break;
        }

        // If c did not combine, then check if it is a starter.
        if(cc==0) {
            // Found a new starter.
            if((compositionsList=getCompositionsListForDecompYes(norm16))!=NULL) {
                // It may combine with something, prepare for it.
                if(U_IS_BMP(c)) {
                    starterIsSupplementary=FALSE;
                    starter=p-1;
                } else {
                    starterIsSupplementary=TRUE;
                    starter=p-2;
                }
            }
        } else if(onlyContiguous) {
            // FCC: no discontiguous compositions; any intervening character blocks.
            compositionsList=NULL;
        }
    }
    buffer.setReorderingLimit(limit);
}

UChar32
Normalizer2Impl::composePair(UChar32 a, UChar32 b) const {
    uint16_t norm16=getNorm16(a);  // maps an out-of-range 'a' to inert norm16
    const uint16_t *list;
    if(isInert(norm16)) {
        return U_SENTINEL;
    } else if(norm16<minYesNoMappingsOnly) {
        // a combines forward.
        if(isJamoL(norm16)) {
            b-=Hangul::JAMO_V_BASE;
            if(0<=b && b<Hangul::JAMO_V_COUNT) {
                return
                    (Hangul::HANGUL_BASE+
                     ((a-Hangul::JAMO_L_BASE)*Hangul::JAMO_V_COUNT+b)*
                     Hangul::JAMO_T_COUNT);
            } else {
                return U_SENTINEL;
            }
        } else if(isHangulLV(norm16)) {
            b-=Hangul::JAMO_T_BASE;
            if(0<b && b<Hangul::JAMO_T_COUNT) {  // not b==0!
                return a+b;
            } else {
                return U_SENTINEL;
            }
        } else {
            // 'a' has a compositions list in extraData
            list=getMapping(norm16);
            if(norm16>minYesNo) {  // composite 'a' has both mapping & compositions list
                list+=  // mapping pointer
                    1+  // +1 to skip the first unit with the mapping length
                    (*list&MAPPING_LENGTH_MASK);  // + mapping length
            }
        }
    } else if(norm16<minMaybeYes || MIN_NORMAL_MAYBE_YES<=norm16) {
        return U_SENTINEL;
    } else {
        list=getCompositionsListForMaybe(norm16);
    }
    if(b<0 || 0x10ffff<b) {  // combine(list, b) requires a valid code point b
        return U_SENTINEL;
    }
#if U_SIGNED_RIGHT_SHIFT_IS_ARITHMETIC
    return combine(list, b)>>1;
#else
    int32_t compositeAndFwd=combine(list, b);
    return compositeAndFwd>=0 ? compositeAndFwd>>1 : U_SENTINEL;
#endif
}

// Very similar to composeQuickCheck(): Make the same changes in both places if relevant.
// doCompose: normalize
// !doCompose: isNormalized (buffer must be empty and initialized)
UBool
Normalizer2Impl::compose(const UChar *src, const UChar *limit,
                         UBool onlyContiguous,
                         UBool doCompose,
                         ReorderingBuffer &buffer,
                         UErrorCode &errorCode) const {
    const UChar *prevBoundary=src;
    UChar32 minNoMaybeCP=minCompNoMaybeCP;
    if(limit==NULL) {
        src=copyLowPrefixFromNulTerminated(src, minNoMaybeCP,
                                           doCompose ? &buffer : NULL,
                                           errorCode);
        if(U_FAILURE(errorCode)) {
            return FALSE;
        }
        limit=u_strchr(src, 0);
        if (prevBoundary != src) {
            if (hasCompBoundaryAfter(*(src-1), onlyContiguous)) {
                prevBoundary = src;
            } else {
                buffer.removeSuffix(1);
                prevBoundary = --src;
            }
        }
    }

    for (;;) {
        // Fast path: Scan over a sequence of characters below the minimum "no or maybe" code point,
        // or with (compYes && ccc==0) properties.
        const UChar *prevSrc;
        UChar32 c = 0;
        uint16_t norm16 = 0;
        for (;;) {
            if (src == limit) {
                if (prevBoundary != limit && doCompose) {
                    buffer.appendZeroCC(prevBoundary, limit, errorCode);
                }
                return TRUE;
            }
            if( (c=*src)<minNoMaybeCP ||
                isCompYesAndZeroCC(norm16=UCPTRIE_FAST_BMP_GET(normTrie, UCPTRIE_16, c))
            ) {
                ++src;
            } else {
                prevSrc = src++;
                if(!U16_IS_LEAD(c)) {
                    break;
                } else {
                    UChar c2;
                    if(src!=limit && U16_IS_TRAIL(c2=*src)) {
                        ++src;
                        c=U16_GET_SUPPLEMENTARY(c, c2);
                        norm16=UCPTRIE_FAST_SUPP_GET(normTrie, UCPTRIE_16, c);
                        if(!isCompYesAndZeroCC(norm16)) {
                            break;
                        }
                    }
                }
            }
        }
        // isCompYesAndZeroCC(norm16) is false, that is, norm16>=minNoNo.
        // The current character is either a "noNo" (has a mapping)
        // or a "maybeYes" (combines backward)
        // or a "yesYes" with ccc!=0.
        // It is not a Hangul syllable or Jamo L because those have "yes" properties.

        // Medium-fast path: Handle cases that do not require full decomposition and recomposition.
        if (!isMaybeOrNonZeroCC(norm16)) {  // minNoNo <= norm16 < minMaybeYes
            if (!doCompose) {
                return FALSE;
            }
            // Fast path for mapping a character that is immediately surrounded by boundaries.
            // In this case, we need not decompose around the current character.
            if (isDecompNoAlgorithmic(norm16)) {
                // Maps to a single isCompYesAndZeroCC character
                // which also implies hasCompBoundaryBefore.
                if (norm16HasCompBoundaryAfter(norm16, onlyContiguous) ||
                        hasCompBoundaryBefore(src, limit)) {
                    if (prevBoundary != prevSrc && !buffer.appendZeroCC(prevBoundary, prevSrc, errorCode)) {
                        break;
                    }
                    if(!buffer.append(mapAlgorithmic(c, norm16), 0, errorCode)) {
                        break;
                    }
                    prevBoundary = src;
                    continue;
                }
            } else if (norm16 < minNoNoCompBoundaryBefore) {
                // The mapping is comp-normalized which also implies hasCompBoundaryBefore.
                if (norm16HasCompBoundaryAfter(norm16, onlyContiguous) ||
                        hasCompBoundaryBefore(src, limit)) {
                    if (prevBoundary != prevSrc && !buffer.appendZeroCC(prevBoundary, prevSrc, errorCode)) {
                        break;
                    }
                    const UChar *mapping = reinterpret_cast<const UChar *>(getMapping(norm16));
                    int32_t length = *mapping++ & MAPPING_LENGTH_MASK;
                    if(!buffer.appendZeroCC(mapping, mapping + length, errorCode)) {
                        break;
                    }
                    prevBoundary = src;
                    continue;
                }
            } else if (norm16 >= minNoNoEmpty) {
                // The current character maps to nothing.
                // Simply omit it from the output if there is a boundary before _or_ after it.
                // The character itself implies no boundaries.
                if (hasCompBoundaryBefore(src, limit) ||
                        hasCompBoundaryAfter(prevBoundary, prevSrc, onlyContiguous)) {
                    if (prevBoundary != prevSrc && !buffer.appendZeroCC(prevBoundary, prevSrc, errorCode)) {
                        break;
                    }
                    prevBoundary = src;
                    continue;
                }
            }
            // Other "noNo" type, or need to examine more text around this character:
            // Fall through to the slow path.
        } else if (isJamoVT(norm16) && prevBoundary != prevSrc) {
            UChar prev=*(prevSrc-1);
            if(c<Hangul::JAMO_T_BASE) {
                // The current character is a Jamo Vowel,
                // compose with previous Jamo L and following Jamo T.
                UChar l = (UChar)(prev-Hangul::JAMO_L_BASE);
                if(l<Hangul::JAMO_L_COUNT) {
                    if (!doCompose) {
                        return FALSE;
                    }
                    int32_t t;
                    if (src != limit &&
                            0 < (t = ((int32_t)*src - Hangul::JAMO_T_BASE)) &&
                            t < Hangul::JAMO_T_COUNT) {
                        // The next character is a Jamo T.
                        ++src;
                    } else if (hasCompBoundaryBefore(src, limit)) {
                        // No Jamo T follows, not even via decomposition.
                        t = 0;
                    } else {
                        t = -1;
                    }
                    if (t >= 0) {
                        UChar32 syllable = Hangul::HANGUL_BASE +
                            (l*Hangul::JAMO_V_COUNT + (c-Hangul::JAMO_V_BASE)) *
                            Hangul::JAMO_T_COUNT + t;
                        --prevSrc;  // Replace the Jamo L as well.
                        if (prevBoundary != prevSrc && !buffer.appendZeroCC(prevBoundary, prevSrc, errorCode)) {
                            break;
                        }
                        if(!buffer.appendBMP((UChar)syllable, 0, errorCode)) {
                            break;
                        }
                        prevBoundary = src;
                        continue;
                    }
                    // If we see L+V+x where x!=T then we drop to the slow path,
                    // decompose and recompose.
                    // This is to deal with NFKC finding normal L and V but a
                    // compatibility variant of a T.
                    // We need to either fully compose that combination here
                    // (which would complicate the code and may not work with strange custom data)
                    // or use the slow path.
                }
            } else if (Hangul::isHangulLV(prev)) {
                // The current character is a Jamo Trailing consonant,
                // compose with previous Hangul LV that does not contain a Jamo T.
                if (!doCompose) {
                    return FALSE;
                }
                UChar32 syllable = prev + c - Hangul::JAMO_T_BASE;
                --prevSrc;  // Replace the Hangul LV as well.
                if (prevBoundary != prevSrc && !buffer.appendZeroCC(prevBoundary, prevSrc, errorCode)) {
                    break;
                }
                if(!buffer.appendBMP((UChar)syllable, 0, errorCode)) {
                    break;
                }
                prevBoundary = src;
                continue;
            }
            // No matching context, or may need to decompose surrounding text first:
            // Fall through to the slow path.
        } else if (norm16 > JAMO_VT) {  // norm16 >= MIN_YES_YES_WITH_CC
            // One or more combining marks that do not combine-back:
            // Check for canonical order, copy unchanged if ok and
            // if followed by a character with a boundary-before.
            uint8_t cc = getCCFromNormalYesOrMaybe(norm16);  // cc!=0
            if (onlyContiguous /* FCC */ && getPreviousTrailCC(prevBoundary, prevSrc) > cc) {
                // Fails FCD test, need to decompose and contiguously recompose.
                if (!doCompose) {
                    return FALSE;
                }
            } else {
                // If !onlyContiguous (not FCC), then we ignore the tccc of
                // the previous character which passed the quick check "yes && ccc==0" test.
                const UChar *nextSrc;
                uint16_t n16;
                for (;;) {
                    if (src == limit) {
                        if (doCompose) {
                            buffer.appendZeroCC(prevBoundary, limit, errorCode);
                        }
                        return TRUE;
                    }
                    uint8_t prevCC = cc;
                    nextSrc = src;
                    UCPTRIE_FAST_U16_NEXT(normTrie, UCPTRIE_16, nextSrc, limit, c, n16);
                    if (n16 >= MIN_YES_YES_WITH_CC) {
                        cc = getCCFromNormalYesOrMaybe(n16);
                        if (prevCC > cc) {
                            if (!doCompose) {
                                return FALSE;
                            }
                            break;
                        }
                    } else {
                        break;
                    }
                    src = nextSrc;
                }
                // src is after the last in-order combining mark.
                // If there is a boundary here, then we continue with no change.
                if (norm16HasCompBoundaryBefore(n16)) {
                    if (isCompYesAndZeroCC(n16)) {
                        src = nextSrc;
                    }
                    continue;
                }
                // Use the slow path. There is no boundary in [prevSrc, src[.
            }
        }

        // Slow path: Find the nearest boundaries around the current character,
        // decompose and recompose.
        if (prevBoundary != prevSrc && !norm16HasCompBoundaryBefore(norm16)) {
            const UChar *p = prevSrc;
            UCPTRIE_FAST_U16_PREV(normTrie, UCPTRIE_16, prevBoundary, p, c, norm16);
            if (!norm16HasCompBoundaryAfter(norm16, onlyContiguous)) {
                prevSrc = p;
            }
        }
        if (doCompose && prevBoundary != prevSrc && !buffer.appendZeroCC(prevBoundary, prevSrc, errorCode)) {
            break;
        }
        int32_t recomposeStartIndex=buffer.length();
        // We know there is not a boundary here.
        decomposeShort(prevSrc, src, FALSE /* !stopAtCompBoundary */, onlyContiguous,
                       buffer, errorCode);
        // Decompose until the next boundary.
        src = decomposeShort(src, limit, TRUE /* stopAtCompBoundary */, onlyContiguous,
                             buffer, errorCode);
        if (U_FAILURE(errorCode)) {
            break;
        }
        if ((src - prevSrc) > INT32_MAX) {  // guard before buffer.equals()
            errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
            return TRUE;
        }
        recompose(buffer, recomposeStartIndex, onlyContiguous);
        if(!doCompose) {
            if(!buffer.equals(prevSrc, src)) {
                return FALSE;
            }
            buffer.remove();
        }
        prevBoundary=src;
    }
    return TRUE;
}

// Very similar to compose(): Make the same changes in both places if relevant.
// pQCResult==NULL: spanQuickCheckYes
// pQCResult!=NULL: quickCheck (*pQCResult must be UNORM_YES)
const UChar *
Normalizer2Impl::composeQuickCheck(const UChar *src, const UChar *limit,
                                   UBool onlyContiguous,
                                   UNormalizationCheckResult *pQCResult) const {
    const UChar *prevBoundary=src;
    UChar32 minNoMaybeCP=minCompNoMaybeCP;
    if(limit==NULL) {
        UErrorCode errorCode=U_ZERO_ERROR;
        src=copyLowPrefixFromNulTerminated(src, minNoMaybeCP, NULL, errorCode);
        limit=u_strchr(src, 0);
        if (prevBoundary != src) {
            if (hasCompBoundaryAfter(*(src-1), onlyContiguous)) {
                prevBoundary = src;
            } else {
                prevBoundary = --src;
            }
        }
    }

    for(;;) {
        // Fast path: Scan over a sequence of characters below the minimum "no or maybe" code point,
        // or with (compYes && ccc==0) properties.
        const UChar *prevSrc;
        UChar32 c = 0;
        uint16_t norm16 = 0;
        for (;;) {
            if(src==limit) {
                return src;
            }
            if( (c=*src)<minNoMaybeCP ||
                isCompYesAndZeroCC(norm16=UCPTRIE_FAST_BMP_GET(normTrie, UCPTRIE_16, c))
            ) {
                ++src;
            } else {
                prevSrc = src++;
                if(!U16_IS_LEAD(c)) {
                    break;
                } else {
                    UChar c2;
                    if(src!=limit && U16_IS_TRAIL(c2=*src)) {
                        ++src;
                        c=U16_GET_SUPPLEMENTARY(c, c2);
                        norm16=UCPTRIE_FAST_SUPP_GET(normTrie, UCPTRIE_16, c);
                        if(!isCompYesAndZeroCC(norm16)) {
                            break;
                        }
                    }
                }
            }
        }
        // isCompYesAndZeroCC(norm16) is false, that is, norm16>=minNoNo.
        // The current character is either a "noNo" (has a mapping)
        // or a "maybeYes" (combines backward)
        // or a "yesYes" with ccc!=0.
        // It is not a Hangul syllable or Jamo L because those have "yes" properties.

        uint16_t prevNorm16 = INERT;
        if (prevBoundary != prevSrc) {
            if (norm16HasCompBoundaryBefore(norm16)) {
                prevBoundary = prevSrc;
            } else {
                const UChar *p = prevSrc;
                uint16_t n16;
                UCPTRIE_FAST_U16_PREV(normTrie, UCPTRIE_16, prevBoundary, p, c, n16);
                if (norm16HasCompBoundaryAfter(n16, onlyContiguous)) {
                    prevBoundary = prevSrc;
                } else {
                    prevBoundary = p;
                    prevNorm16 = n16;
                }
            }
        }

        if(isMaybeOrNonZeroCC(norm16)) {
            uint8_t cc=getCCFromYesOrMaybe(norm16);
            if (onlyContiguous /* FCC */ && cc != 0 &&
                    getTrailCCFromCompYesAndZeroCC(prevNorm16) > cc) {
                // The [prevBoundary..prevSrc[ character
                // passed the quick check "yes && ccc==0" test
                // but is out of canonical order with the current combining mark.
            } else {
                // If !onlyContiguous (not FCC), then we ignore the tccc of
                // the previous character which passed the quick check "yes && ccc==0" test.
                const UChar *nextSrc;
                for (;;) {
                    if (norm16 < MIN_YES_YES_WITH_CC) {
                        if (pQCResult != nullptr) {
                            *pQCResult = UNORM_MAYBE;
                        } else {
                            return prevBoundary;
                        }
                    }
                    if (src == limit) {
                        return src;
                    }
                    uint8_t prevCC = cc;
                    nextSrc = src;
                    UCPTRIE_FAST_U16_NEXT(normTrie, UCPTRIE_16, nextSrc, limit, c, norm16);
                    if (isMaybeOrNonZeroCC(norm16)) {
                        cc = getCCFromYesOrMaybe(norm16);
                        if (!(prevCC <= cc || cc == 0)) {
                            break;
                        }
                    } else {
                        break;
                    }
                    src = nextSrc;
                }
                // src is after the last in-order combining mark.
                if (isCompYesAndZeroCC(norm16)) {
                    prevBoundary = src;
                    src = nextSrc;
                    continue;
                }
            }
        }
        if(pQCResult!=NULL) {
            *pQCResult=UNORM_NO;
        }
        return prevBoundary;
    }
}

void Normalizer2Impl::composeAndAppend(const UChar *src, const UChar *limit,
                                       UBool doCompose,
                                       UBool onlyContiguous,
                                       UnicodeString &safeMiddle,
                                       ReorderingBuffer &buffer,
                                       UErrorCode &errorCode) const {
    if(!buffer.isEmpty()) {
        const UChar *firstStarterInSrc=findNextCompBoundary(src, limit, onlyContiguous);
        if(src!=firstStarterInSrc) {
            const UChar *lastStarterInDest=findPreviousCompBoundary(buffer.getStart(),
                                                                    buffer.getLimit(), onlyContiguous);
            int32_t destSuffixLength=(int32_t)(buffer.getLimit()-lastStarterInDest);
            UnicodeString middle(lastStarterInDest, destSuffixLength);
            buffer.removeSuffix(destSuffixLength);
            safeMiddle=middle;
            middle.append(src, (int32_t)(firstStarterInSrc-src));
            const UChar *middleStart=middle.getBuffer();
            compose(middleStart, middleStart+middle.length(), onlyContiguous,
                    TRUE, buffer, errorCode);
            if(U_FAILURE(errorCode)) {
                return;
            }
            src=firstStarterInSrc;
        }
    }
    if(doCompose) {
        compose(src, limit, onlyContiguous, TRUE, buffer, errorCode);
    } else {
        if(limit==NULL) {  // appendZeroCC() needs limit!=NULL
            limit=u_strchr(src, 0);
        }
        buffer.appendZeroCC(src, limit, errorCode);
    }
}

UBool
Normalizer2Impl::composeUTF8(uint32_t options, UBool onlyContiguous,
                             const uint8_t *src, const uint8_t *limit,
                             ByteSink *sink, Edits *edits, UErrorCode &errorCode) const {
    U_ASSERT(limit != nullptr);
    UnicodeString s16;
    uint8_t minNoMaybeLead = leadByteForCP(minCompNoMaybeCP);
    const uint8_t *prevBoundary = src;

    for (;;) {
        // Fast path: Scan over a sequence of characters below the minimum "no or maybe" code point,
        // or with (compYes && ccc==0) properties.
        const uint8_t *prevSrc;
        uint16_t norm16 = 0;
        for (;;) {
            if (src == limit) {
                if (prevBoundary != limit && sink != nullptr) {
                    ByteSinkUtil::appendUnchanged(prevBoundary, limit,
                                                  *sink, options, edits, errorCode);
                }
                return TRUE;
            }
            if (*src < minNoMaybeLead) {
                ++src;
            } else {
                prevSrc = src;
                UCPTRIE_FAST_U8_NEXT(normTrie, UCPTRIE_16, src, limit, norm16);
                if (!isCompYesAndZeroCC(norm16)) {
                    break;
                }
            }
        }
        // isCompYesAndZeroCC(norm16) is false, that is, norm16>=minNoNo.
        // The current character is either a "noNo" (has a mapping)
        // or a "maybeYes" (combines backward)
        // or a "yesYes" with ccc!=0.
        // It is not a Hangul syllable or Jamo L because those have "yes" properties.

        // Medium-fast path: Handle cases that do not require full decomposition and recomposition.
        if (!isMaybeOrNonZeroCC(norm16)) {  // minNoNo <= norm16 < minMaybeYes
            if (sink == nullptr) {
                return FALSE;
            }
            // Fast path for mapping a character that is immediately surrounded by boundaries.
            // In this case, we need not decompose around the current character.
            if (isDecompNoAlgorithmic(norm16)) {
                // Maps to a single isCompYesAndZeroCC character
                // which also implies hasCompBoundaryBefore.
                if (norm16HasCompBoundaryAfter(norm16, onlyContiguous) ||
                        hasCompBoundaryBefore(src, limit)) {
                    if (prevBoundary != prevSrc &&
                            !ByteSinkUtil::appendUnchanged(prevBoundary, prevSrc,
                                                           *sink, options, edits, errorCode)) {
                        break;
                    }
                    appendCodePointDelta(prevSrc, src, getAlgorithmicDelta(norm16), *sink, edits);
                    prevBoundary = src;
                    continue;
                }
            } else if (norm16 < minNoNoCompBoundaryBefore) {
                // The mapping is comp-normalized which also implies hasCompBoundaryBefore.
                if (norm16HasCompBoundaryAfter(norm16, onlyContiguous) ||
                        hasCompBoundaryBefore(src, limit)) {
                    if (prevBoundary != prevSrc &&
                            !ByteSinkUtil::appendUnchanged(prevBoundary, prevSrc,
                                                           *sink, options, edits, errorCode)) {
                        break;
                    }
                    const uint16_t *mapping = getMapping(norm16);
                    int32_t length = *mapping++ & MAPPING_LENGTH_MASK;
                    if (!ByteSinkUtil::appendChange(prevSrc, src, (const UChar *)mapping, length,
                                                    *sink, edits, errorCode)) {
                        break;
                    }
                    prevBoundary = src;
                    continue;
                }
            } else if (norm16 >= minNoNoEmpty) {
                // The current character maps to nothing.
                // Simply omit it from the output if there is a boundary before _or_ after it.
                // The character itself implies no boundaries.
                if (hasCompBoundaryBefore(src, limit) ||
                        hasCompBoundaryAfter(prevBoundary, prevSrc, onlyContiguous)) {
                    if (prevBoundary != prevSrc &&
                            !ByteSinkUtil::appendUnchanged(prevBoundary, prevSrc,
                                                           *sink, options, edits, errorCode)) {
                        break;
                    }
                    if (edits != nullptr) {
                        edits->addReplace((int32_t)(src - prevSrc), 0);
                    }
                    prevBoundary = src;
                    continue;
                }
            }
            // Other "noNo" type, or need to examine more text around this character:
            // Fall through to the slow path.
        } else if (isJamoVT(norm16)) {
            // Jamo L: E1 84 80..92
            // Jamo V: E1 85 A1..B5
            // Jamo T: E1 86 A8..E1 87 82
            U_ASSERT((src - prevSrc) == 3 && *prevSrc == 0xe1);
            UChar32 prev = previousHangulOrJamo(prevBoundary, prevSrc);
            if (prevSrc[1] == 0x85) {
                // The current character is a Jamo Vowel,
                // compose with previous Jamo L and following Jamo T.
                UChar32 l = prev - Hangul::JAMO_L_BASE;
                if ((uint32_t)l < Hangul::JAMO_L_COUNT) {
                    if (sink == nullptr) {
                        return FALSE;
                    }
                    int32_t t = getJamoTMinusBase(src, limit);
                    if (t >= 0) {
                        // The next character is a Jamo T.
                        src += 3;
                    } else if (hasCompBoundaryBefore(src, limit)) {
                        // No Jamo T follows, not even via decomposition.
                        t = 0;
                    }
                    if (t >= 0) {
                        UChar32 syllable = Hangul::HANGUL_BASE +
                            (l*Hangul::JAMO_V_COUNT + (prevSrc[2]-0xa1)) *
                            Hangul::JAMO_T_COUNT + t;
                        prevSrc -= 3;  // Replace the Jamo L as well.
                        if (prevBoundary != prevSrc &&
                                !ByteSinkUtil::appendUnchanged(prevBoundary, prevSrc,
                                                               *sink, options, edits, errorCode)) {
                            break;
                        }
                        ByteSinkUtil::appendCodePoint(prevSrc, src, syllable, *sink, edits);
                        prevBoundary = src;
                        continue;
                    }
                    // If we see L+V+x where x!=T then we drop to the slow path,
                    // decompose and recompose.
                    // This is to deal with NFKC finding normal L and V but a
                    // compatibility variant of a T.
                    // We need to either fully compose that combination here
                    // (which would complicate the code and may not work with strange custom data)
                    // or use the slow path.
                }
            } else if (Hangul::isHangulLV(prev)) {
                // The current character is a Jamo Trailing consonant,
                // compose with previous Hangul LV that does not contain a Jamo T.
                if (sink == nullptr) {
                    return FALSE;
                }
                UChar32 syllable = prev + getJamoTMinusBase(prevSrc, src);
                prevSrc -= 3;  // Replace the Hangul LV as well.
                if (prevBoundary != prevSrc &&
                        !ByteSinkUtil::appendUnchanged(prevBoundary, prevSrc,
                                                       *sink, options, edits, errorCode)) {
                    break;
                }
                ByteSinkUtil::appendCodePoint(prevSrc, src, syllable, *sink, edits);
                prevBoundary = src;
                continue;
            }
            // No matching context, or may need to decompose surrounding text first:
            // Fall through to the slow path.
        } else if (norm16 > JAMO_VT) {  // norm16 >= MIN_YES_YES_WITH_CC
            // One or more combining marks that do not combine-back:
            // Check for canonical order, copy unchanged if ok and
            // if followed by a character with a boundary-before.
            uint8_t cc = getCCFromNormalYesOrMaybe(norm16);  // cc!=0
            if (onlyContiguous /* FCC */ && getPreviousTrailCC(prevBoundary, prevSrc) > cc) {
                // Fails FCD test, need to decompose and contiguously recompose.
                if (sink == nullptr) {
                    return FALSE;
                }
            } else {
                // If !onlyContiguous (not FCC), then we ignore the tccc of
                // the previous character which passed the quick check "yes && ccc==0" test.
                const uint8_t *nextSrc;
                uint16_t n16;
                for (;;) {
                    if (src == limit) {
                        if (sink != nullptr) {
                            ByteSinkUtil::appendUnchanged(prevBoundary, limit,
                                                          *sink, options, edits, errorCode);
                        }
                        return TRUE;
                    }
                    uint8_t prevCC = cc;
                    nextSrc = src;
                    UCPTRIE_FAST_U8_NEXT(normTrie, UCPTRIE_16, nextSrc, limit, n16);
                    if (n16 >= MIN_YES_YES_WITH_CC) {
                        cc = getCCFromNormalYesOrMaybe(n16);
                        if (prevCC > cc) {
                            if (sink == nullptr) {
                                return FALSE;
                            }
                            break;
                        }
                    } else {
                        break;
                    }
                    src = nextSrc;
                }
                // src is after the last in-order combining mark.
                // If there is a boundary here, then we continue with no change.
                if (norm16HasCompBoundaryBefore(n16)) {
                    if (isCompYesAndZeroCC(n16)) {
                        src = nextSrc;
                    }
                    continue;
                }
                // Use the slow path. There is no boundary in [prevSrc, src[.
            }
        }

        // Slow path: Find the nearest boundaries around the current character,
        // decompose and recompose.
        if (prevBoundary != prevSrc && !norm16HasCompBoundaryBefore(norm16)) {
            const uint8_t *p = prevSrc;
            UCPTRIE_FAST_U8_PREV(normTrie, UCPTRIE_16, prevBoundary, p, norm16);
            if (!norm16HasCompBoundaryAfter(norm16, onlyContiguous)) {
                prevSrc = p;
            }
        }
        ReorderingBuffer buffer(*this, s16, errorCode);
        if (U_FAILURE(errorCode)) {
            break;
        }
        // We know there is not a boundary here.
        decomposeShort(prevSrc, src, FALSE /* !stopAtCompBoundary */, onlyContiguous,
                       buffer, errorCode);
        // Decompose until the next boundary.
        src = decomposeShort(src, limit, TRUE /* stopAtCompBoundary */, onlyContiguous,
                             buffer, errorCode);
        if (U_FAILURE(errorCode)) {
            break;
        }
        if ((src - prevSrc) > INT32_MAX) {  // guard before buffer.equals()
            errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
            return TRUE;
        }
        recompose(buffer, 0, onlyContiguous);
        if (!buffer.equals(prevSrc, src)) {
            if (sink == nullptr) {
                return FALSE;
            }
            if (prevBoundary != prevSrc &&
                    !ByteSinkUtil::appendUnchanged(prevBoundary, prevSrc,
                                                   *sink, options, edits, errorCode)) {
                break;
            }
            if (!ByteSinkUtil::appendChange(prevSrc, src, buffer.getStart(), buffer.length(),
                                            *sink, edits, errorCode)) {
                break;
            }
            prevBoundary = src;
        }
    }
    return TRUE;
}

UBool Normalizer2Impl::hasCompBoundaryBefore(const UChar *src, const UChar *limit) const {
    if (src == limit || *src < minCompNoMaybeCP) {
        return TRUE;
    }
    UChar32 c;
    uint16_t norm16;
    UCPTRIE_FAST_U16_NEXT(normTrie, UCPTRIE_16, src, limit, c, norm16);
    return norm16HasCompBoundaryBefore(norm16);
}

UBool Normalizer2Impl::hasCompBoundaryBefore(const uint8_t *src, const uint8_t *limit) const {
    if (src == limit) {
        return TRUE;
    }
    uint16_t norm16;
    UCPTRIE_FAST_U8_NEXT(normTrie, UCPTRIE_16, src, limit, norm16);
    return norm16HasCompBoundaryBefore(norm16);
}

UBool Normalizer2Impl::hasCompBoundaryAfter(const UChar *start, const UChar *p,
                                            UBool onlyContiguous) const {
    if (start == p) {
        return TRUE;
    }
    UChar32 c;
    uint16_t norm16;
    UCPTRIE_FAST_U16_PREV(normTrie, UCPTRIE_16, start, p, c, norm16);
    return norm16HasCompBoundaryAfter(norm16, onlyContiguous);
}

UBool Normalizer2Impl::hasCompBoundaryAfter(const uint8_t *start, const uint8_t *p,
                                            UBool onlyContiguous) const {
    if (start == p) {
        return TRUE;
    }
    uint16_t norm16;
    UCPTRIE_FAST_U8_PREV(normTrie, UCPTRIE_16, start, p, norm16);
    return norm16HasCompBoundaryAfter(norm16, onlyContiguous);
}

const UChar *Normalizer2Impl::findPreviousCompBoundary(const UChar *start, const UChar *p,
                                                       UBool onlyContiguous) const {
    while (p != start) {
        const UChar *codePointLimit = p;
        UChar32 c;
        uint16_t norm16;
        UCPTRIE_FAST_U16_PREV(normTrie, UCPTRIE_16, start, p, c, norm16);
        if (norm16HasCompBoundaryAfter(norm16, onlyContiguous)) {
            return codePointLimit;
        }
        if (hasCompBoundaryBefore(c, norm16)) {
            return p;
        }
    }
    return p;
}

const UChar *Normalizer2Impl::findNextCompBoundary(const UChar *p, const UChar *limit,
                                                   UBool onlyContiguous) const {
    while (p != limit) {
        const UChar *codePointStart = p;
        UChar32 c;
        uint16_t norm16;
        UCPTRIE_FAST_U16_NEXT(normTrie, UCPTRIE_16, p, limit, c, norm16);
        if (hasCompBoundaryBefore(c, norm16)) {
            return codePointStart;
        }
        if (norm16HasCompBoundaryAfter(norm16, onlyContiguous)) {
            return p;
        }
    }
    return p;
}

uint8_t Normalizer2Impl::getPreviousTrailCC(const UChar *start, const UChar *p) const {
    if (start == p) {
        return 0;
    }
    int32_t i = (int32_t)(p - start);
    UChar32 c;
    U16_PREV(start, 0, i, c);
    return (uint8_t)getFCD16(c);
}

uint8_t Normalizer2Impl::getPreviousTrailCC(const uint8_t *start, const uint8_t *p) const {
    if (start == p) {
        return 0;
    }
    int32_t i = (int32_t)(p - start);
    UChar32 c;
    U8_PREV(start, 0, i, c);
    return (uint8_t)getFCD16(c);
}

// Note: normalizer2impl.cpp r30982 (2011-nov-27)
// still had getFCDTrie() which built and cached an FCD trie.
// That provided faster access to FCD data than getFCD16FromNormData()
// but required synchronization and consumed some 10kB of heap memory
// in any process that uses FCD (e.g., via collation).
// minDecompNoCP etc. and smallFCD[] are intended to help with any loss of performance,
// at least for ASCII & CJK.

// Ticket 20907 - The optimizer in MSVC/Visual Studio versions below 16.4 has trouble with this
// function on Windows ARM64. As a work-around, we disable optimizations for this function.
// This work-around could/should be removed once the following versions of Visual Studio are no
// longer supported: All versions of VS2017, and versions of VS2019 below 16.4.
#if (defined(_MSC_VER) && (defined(_M_ARM64)) && (_MSC_VER < 1924))
#pragma optimize( "", off )
#endif
// Gets the FCD value from the regular normalization data.
uint16_t Normalizer2Impl::getFCD16FromNormData(UChar32 c) const {
    uint16_t norm16=getNorm16(c);
    if (norm16 >= limitNoNo) {
        if(norm16>=MIN_NORMAL_MAYBE_YES) {
            // combining mark
            norm16=getCCFromNormalYesOrMaybe(norm16);
            return norm16|(norm16<<8);
        } else if(norm16>=minMaybeYes) {
            return 0;
        } else {  // isDecompNoAlgorithmic(norm16)
            uint16_t deltaTrailCC = norm16 & DELTA_TCCC_MASK;
            if (deltaTrailCC <= DELTA_TCCC_1) {
                return deltaTrailCC >> OFFSET_SHIFT;
            }
            // Maps to an isCompYesAndZeroCC.
            c=mapAlgorithmic(c, norm16);
            norm16=getRawNorm16(c);
        }
    }
    if(norm16<=minYesNo || isHangulLVT(norm16)) {
        // no decomposition or Hangul syllable, all zeros
        return 0;
    }
    // c decomposes, get everything from the variable-length extra data
    const uint16_t *mapping=getMapping(norm16);
    uint16_t firstUnit=*mapping;
    norm16=firstUnit>>8;  // tccc
    if(firstUnit&MAPPING_HAS_CCC_LCCC_WORD) {
        norm16|=*(mapping-1)&0xff00;  // lccc
    }
    return norm16;
}
#if (defined(_MSC_VER) && (defined(_M_ARM64)) && (_MSC_VER < 1924))
#pragma optimize( "", on )
#endif

// Dual functionality:
// buffer!=NULL: normalize
// buffer==NULL: isNormalized/quickCheck/spanQuickCheckYes
const UChar *
Normalizer2Impl::makeFCD(const UChar *src, const UChar *limit,
                         ReorderingBuffer *buffer,
                         UErrorCode &errorCode) const {
    // Tracks the last FCD-safe boundary, before lccc=0 or after properly-ordered tccc<=1.
    // Similar to the prevBoundary in the compose() implementation.
    const UChar *prevBoundary=src;
    int32_t prevFCD16=0;
    if(limit==NULL) {
        src=copyLowPrefixFromNulTerminated(src, minLcccCP, buffer, errorCode);
        if(U_FAILURE(errorCode)) {
            return src;
        }
        if(prevBoundary<src) {
            prevBoundary=src;
            // We know that the previous character's lccc==0.
            // Fetching the fcd16 value was deferred for this below-U+0300 code point.
            prevFCD16=getFCD16(*(src-1));
            if(prevFCD16>1) {
                --prevBoundary;
            }
        }
        limit=u_strchr(src, 0);
    }

    // Note: In this function we use buffer->appendZeroCC() because we track
    // the lead and trail combining classes here, rather than leaving it to
    // the ReorderingBuffer.
    // The exception is the call to decomposeShort() which uses the buffer
    // in the normal way.

    const UChar *prevSrc;
    UChar32 c=0;
    uint16_t fcd16=0;

    for(;;) {
        // count code units with lccc==0
        for(prevSrc=src; src!=limit;) {
            if((c=*src)<minLcccCP) {
                prevFCD16=~c;
                ++src;
            } else if(!singleLeadMightHaveNonZeroFCD16(c)) {
                prevFCD16=0;
                ++src;
            } else {
                if(U16_IS_LEAD(c)) {
                    UChar c2;
                    if((src+1)!=limit && U16_IS_TRAIL(c2=src[1])) {
                        c=U16_GET_SUPPLEMENTARY(c, c2);
                    }
                }
                if((fcd16=getFCD16FromNormData(c))<=0xff) {
                    prevFCD16=fcd16;
                    src+=U16_LENGTH(c);
                } else {
                    break;
                }
            }
        }
        // copy these code units all at once
        if(src!=prevSrc) {
            if(buffer!=NULL && !buffer->appendZeroCC(prevSrc, src, errorCode)) {
                break;
            }
            if(src==limit) {
                break;
            }
            prevBoundary=src;
            // We know that the previous character's lccc==0.
            if(prevFCD16<0) {
                // Fetching the fcd16 value was deferred for this below-minLcccCP code point.
                UChar32 prev=~prevFCD16;
                if(prev<minDecompNoCP) {
                    prevFCD16=0;
                } else {
                    prevFCD16=getFCD16FromNormData(prev);
                    if(prevFCD16>1) {
                        --prevBoundary;
                    }
                }
            } else {
                const UChar *p=src-1;
                if(U16_IS_TRAIL(*p) && prevSrc<p && U16_IS_LEAD(*(p-1))) {
                    --p;
                    // Need to fetch the previous character's FCD value because
                    // prevFCD16 was just for the trail surrogate code point.
                    prevFCD16=getFCD16FromNormData(U16_GET_SUPPLEMENTARY(p[0], p[1]));
                    // Still known to have lccc==0 because its lead surrogate unit had lccc==0.
                }
                if(prevFCD16>1) {
                    prevBoundary=p;
                }
            }
            // The start of the current character (c).
            prevSrc=src;
        } else if(src==limit) {
            break;
        }

        src+=U16_LENGTH(c);
        // The current character (c) at [prevSrc..src[ has a non-zero lead combining class.
        // Check for proper order, and decompose locally if necessary.
        if((prevFCD16&0xff)<=(fcd16>>8)) {
            // proper order: prev tccc <= current lccc
            if((fcd16&0xff)<=1) {
                prevBoundary=src;
            }
            if(buffer!=NULL && !buffer->appendZeroCC(c, errorCode)) {
                break;
            }
            prevFCD16=fcd16;
            continue;
        } else if(buffer==NULL) {
            return prevBoundary;  // quick check "no"
        } else {
            /*
             * Back out the part of the source that we copied or appended
             * already but is now going to be decomposed.
             * prevSrc is set to after what was copied/appended.
             */
            buffer->removeSuffix((int32_t)(prevSrc-prevBoundary));
            /*
             * Find the part of the source that needs to be decomposed,
             * up to the next safe boundary.
             */
            src=findNextFCDBoundary(src, limit);
            /*
             * The source text does not fulfill the conditions for FCD.
             * Decompose and reorder a limited piece of the text.
             */
            decomposeShort(prevBoundary, src, FALSE, FALSE, *buffer, errorCode);
            if (U_FAILURE(errorCode)) {
                break;
            }
            prevBoundary=src;
            prevFCD16=0;
        }
    }
    return src;
}

void Normalizer2Impl::makeFCDAndAppend(const UChar *src, const UChar *limit,
                                       UBool doMakeFCD,
                                       UnicodeString &safeMiddle,
                                       ReorderingBuffer &buffer,
                                       UErrorCode &errorCode) const {
    if(!buffer.isEmpty()) {
        const UChar *firstBoundaryInSrc=findNextFCDBoundary(src, limit);
        if(src!=firstBoundaryInSrc) {
            const UChar *lastBoundaryInDest=findPreviousFCDBoundary(buffer.getStart(),
                                                                    buffer.getLimit());
            int32_t destSuffixLength=(int32_t)(buffer.getLimit()-lastBoundaryInDest);
            UnicodeString middle(lastBoundaryInDest, destSuffixLength);
            buffer.removeSuffix(destSuffixLength);
            safeMiddle=middle;
            middle.append(src, (int32_t)(firstBoundaryInSrc-src));
            const UChar *middleStart=middle.getBuffer();
            makeFCD(middleStart, middleStart+middle.length(), &buffer, errorCode);
            if(U_FAILURE(errorCode)) {
                return;
            }
            src=firstBoundaryInSrc;
        }
    }
    if(doMakeFCD) {
        makeFCD(src, limit, &buffer, errorCode);
    } else {
        if(limit==NULL) {  // appendZeroCC() needs limit!=NULL
            limit=u_strchr(src, 0);
        }
        buffer.appendZeroCC(src, limit, errorCode);
    }
}

const UChar *Normalizer2Impl::findPreviousFCDBoundary(const UChar *start, const UChar *p) const {
    while(start<p) {
        const UChar *codePointLimit = p;
        UChar32 c;
        uint16_t norm16;
        UCPTRIE_FAST_U16_PREV(normTrie, UCPTRIE_16, start, p, c, norm16);
        if (c < minDecompNoCP || norm16HasDecompBoundaryAfter(norm16)) {
            return codePointLimit;
        }
        if (norm16HasDecompBoundaryBefore(norm16)) {
            return p;
        }
    }
    return p;
}

const UChar *Normalizer2Impl::findNextFCDBoundary(const UChar *p, const UChar *limit) const {
    while(p<limit) {
        const UChar *codePointStart=p;
        UChar32 c;
        uint16_t norm16;
        UCPTRIE_FAST_U16_NEXT(normTrie, UCPTRIE_16, p, limit, c, norm16);
        if (c < minLcccCP || norm16HasDecompBoundaryBefore(norm16)) {
            return codePointStart;
        }
        if (norm16HasDecompBoundaryAfter(norm16)) {
            return p;
        }
    }
    return p;
}

// CanonicalIterator data -------------------------------------------------- ***

CanonIterData::CanonIterData(UErrorCode &errorCode) :
        mutableTrie(umutablecptrie_open(0, 0, &errorCode)), trie(nullptr),
        canonStartSets(uprv_deleteUObject, NULL, errorCode) {}

CanonIterData::~CanonIterData() {
    umutablecptrie_close(mutableTrie);
    ucptrie_close(trie);
}

void CanonIterData::addToStartSet(UChar32 origin, UChar32 decompLead, UErrorCode &errorCode) {
    uint32_t canonValue = umutablecptrie_get(mutableTrie, decompLead);
    if((canonValue&(CANON_HAS_SET|CANON_VALUE_MASK))==0 && origin!=0) {
        // origin is the first character whose decomposition starts with
        // the character for which we are setting the value.
        umutablecptrie_set(mutableTrie, decompLead, canonValue|origin, &errorCode);
    } else {
        // origin is not the first character, or it is U+0000.
        UnicodeSet *set;
        if((canonValue&CANON_HAS_SET)==0) {
            set=new UnicodeSet;
            if(set==NULL) {
                errorCode=U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            UChar32 firstOrigin=(UChar32)(canonValue&CANON_VALUE_MASK);
            canonValue=(canonValue&~CANON_VALUE_MASK)|CANON_HAS_SET|(uint32_t)canonStartSets.size();
            umutablecptrie_set(mutableTrie, decompLead, canonValue, &errorCode);
            canonStartSets.addElement(set, errorCode);
            if(firstOrigin!=0) {
                set->add(firstOrigin);
            }
        } else {
            set=(UnicodeSet *)canonStartSets[(int32_t)(canonValue&CANON_VALUE_MASK)];
        }
        set->add(origin);
    }
}

// C++ class for friend access to private Normalizer2Impl members.
class InitCanonIterData {
public:
    static void doInit(Normalizer2Impl *impl, UErrorCode &errorCode);
};

U_CDECL_BEGIN

// UInitOnce instantiation function for CanonIterData
static void U_CALLCONV
initCanonIterData(Normalizer2Impl *impl, UErrorCode &errorCode) {
    InitCanonIterData::doInit(impl, errorCode);
}

U_CDECL_END

void InitCanonIterData::doInit(Normalizer2Impl *impl, UErrorCode &errorCode) {
    U_ASSERT(impl->fCanonIterData == NULL);
    impl->fCanonIterData = new CanonIterData(errorCode);
    if (impl->fCanonIterData == NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
    }
    if (U_SUCCESS(errorCode)) {
        UChar32 start = 0, end;
        uint32_t value;
        while ((end = ucptrie_getRange(impl->normTrie, start,
                                       UCPMAP_RANGE_FIXED_LEAD_SURROGATES, Normalizer2Impl::INERT,
                                       nullptr, nullptr, &value)) >= 0) {
            // Call Normalizer2Impl::makeCanonIterDataFromNorm16() for a range of same-norm16 characters.
            if (value != Normalizer2Impl::INERT) {
                impl->makeCanonIterDataFromNorm16(start, end, value, *impl->fCanonIterData, errorCode);
            }
            start = end + 1;
        }
#ifdef UCPTRIE_DEBUG
        umutablecptrie_setName(impl->fCanonIterData->mutableTrie, "CanonIterData");
#endif
        impl->fCanonIterData->trie = umutablecptrie_buildImmutable(
            impl->fCanonIterData->mutableTrie, UCPTRIE_TYPE_SMALL, UCPTRIE_VALUE_BITS_32, &errorCode);
        umutablecptrie_close(impl->fCanonIterData->mutableTrie);
        impl->fCanonIterData->mutableTrie = nullptr;
    }
    if (U_FAILURE(errorCode)) {
        delete impl->fCanonIterData;
        impl->fCanonIterData = NULL;
    }
}

void Normalizer2Impl::makeCanonIterDataFromNorm16(UChar32 start, UChar32 end, const uint16_t norm16,
                                                  CanonIterData &newData,
                                                  UErrorCode &errorCode) const {
    if(isInert(norm16) || (minYesNo<=norm16 && norm16<minNoNo)) {
        // Inert, or 2-way mapping (including Hangul syllable).
        // We do not write a canonStartSet for any yesNo character.
        // Composites from 2-way mappings are added at runtime from the
        // starter's compositions list, and the other characters in
        // 2-way mappings get CANON_NOT_SEGMENT_STARTER set because they are
        // "maybe" characters.
        return;
    }
    for(UChar32 c=start; c<=end; ++c) {
        uint32_t oldValue = umutablecptrie_get(newData.mutableTrie, c);
        uint32_t newValue=oldValue;
        if(isMaybeOrNonZeroCC(norm16)) {
            // not a segment starter if it occurs in a decomposition or has cc!=0
            newValue|=CANON_NOT_SEGMENT_STARTER;
            if(norm16<MIN_NORMAL_MAYBE_YES) {
                newValue|=CANON_HAS_COMPOSITIONS;
            }
        } else if(norm16<minYesNo) {
            newValue|=CANON_HAS_COMPOSITIONS;
        } else {
            // c has a one-way decomposition
            UChar32 c2=c;
            // Do not modify the whole-range norm16 value.
            uint16_t norm16_2=norm16;
            if (isDecompNoAlgorithmic(norm16_2)) {
                // Maps to an isCompYesAndZeroCC.
                c2 = mapAlgorithmic(c2, norm16_2);
                norm16_2 = getRawNorm16(c2);
                // No compatibility mappings for the CanonicalIterator.
                U_ASSERT(!(isHangulLV(norm16_2) || isHangulLVT(norm16_2)));
            }
            if (norm16_2 > minYesNo) {
                // c decomposes, get everything from the variable-length extra data
                const uint16_t *mapping=getMapping(norm16_2);
                uint16_t firstUnit=*mapping;
                int32_t length=firstUnit&MAPPING_LENGTH_MASK;
                if((firstUnit&MAPPING_HAS_CCC_LCCC_WORD)!=0) {
                    if(c==c2 && (*(mapping-1)&0xff)!=0) {
                        newValue|=CANON_NOT_SEGMENT_STARTER;  // original c has cc!=0
                    }
                }
                // Skip empty mappings (no characters in the decomposition).
                if(length!=0) {
                    ++mapping;  // skip over the firstUnit
                    // add c to first code point's start set
                    int32_t i=0;
                    U16_NEXT_UNSAFE(mapping, i, c2);
                    newData.addToStartSet(c, c2, errorCode);
                    // Set CANON_NOT_SEGMENT_STARTER for each remaining code point of a
                    // one-way mapping. A 2-way mapping is possible here after
                    // intermediate algorithmic mapping.
                    if(norm16_2>=minNoNo) {
                        while(i<length) {
                            U16_NEXT_UNSAFE(mapping, i, c2);
                            uint32_t c2Value = umutablecptrie_get(newData.mutableTrie, c2);
                            if((c2Value&CANON_NOT_SEGMENT_STARTER)==0) {
                                umutablecptrie_set(newData.mutableTrie, c2,
                                                   c2Value|CANON_NOT_SEGMENT_STARTER, &errorCode);
                            }
                        }
                    }
                }
            } else {
                // c decomposed to c2 algorithmically; c has cc==0
                newData.addToStartSet(c, c2, errorCode);
            }
        }
        if(newValue!=oldValue) {
            umutablecptrie_set(newData.mutableTrie, c, newValue, &errorCode);
        }
    }
}

UBool Normalizer2Impl::ensureCanonIterData(UErrorCode &errorCode) const {
    // Logically const: Synchronized instantiation.
    Normalizer2Impl *me=const_cast<Normalizer2Impl *>(this);
    umtx_initOnce(me->fCanonIterDataInitOnce, &initCanonIterData, me, errorCode);
    return U_SUCCESS(errorCode);
}

int32_t Normalizer2Impl::getCanonValue(UChar32 c) const {
    return (int32_t)ucptrie_get(fCanonIterData->trie, c);
}

const UnicodeSet &Normalizer2Impl::getCanonStartSet(int32_t n) const {
    return *(const UnicodeSet *)fCanonIterData->canonStartSets[n];
}

UBool Normalizer2Impl::isCanonSegmentStarter(UChar32 c) const {
    return getCanonValue(c)>=0;
}

UBool Normalizer2Impl::getCanonStartSet(UChar32 c, UnicodeSet &set) const {
    int32_t canonValue=getCanonValue(c)&~CANON_NOT_SEGMENT_STARTER;
    if(canonValue==0) {
        return FALSE;
    }
    set.clear();
    int32_t value=canonValue&CANON_VALUE_MASK;
    if((canonValue&CANON_HAS_SET)!=0) {
        set.addAll(getCanonStartSet(value));
    } else if(value!=0) {
        set.add(value);
    }
    if((canonValue&CANON_HAS_COMPOSITIONS)!=0) {
        uint16_t norm16=getRawNorm16(c);
        if(norm16==JAMO_L) {
            UChar32 syllable=
                (UChar32)(Hangul::HANGUL_BASE+(c-Hangul::JAMO_L_BASE)*Hangul::JAMO_VT_COUNT);
            set.add(syllable, syllable+Hangul::JAMO_VT_COUNT-1);
        } else {
            addComposites(getCompositionsList(norm16), set);
        }
    }
    return TRUE;
}

U_NAMESPACE_END

// Normalizer2 data swapping ----------------------------------------------- ***

U_NAMESPACE_USE

U_CAPI int32_t U_EXPORT2
unorm2_swap(const UDataSwapper *ds,
            const void *inData, int32_t length, void *outData,
            UErrorCode *pErrorCode) {
    const UDataInfo *pInfo;
    int32_t headerSize;

    const uint8_t *inBytes;
    uint8_t *outBytes;

    const int32_t *inIndexes;
    int32_t indexes[Normalizer2Impl::IX_TOTAL_SIZE+1];

    int32_t i, offset, nextOffset, size;

    /* udata_swapDataHeader checks the arguments */
    headerSize=udata_swapDataHeader(ds, inData, length, outData, pErrorCode);
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* check data format and format version */
    pInfo=(const UDataInfo *)((const char *)inData+4);
    uint8_t formatVersion0=pInfo->formatVersion[0];
    if(!(
        pInfo->dataFormat[0]==0x4e &&   /* dataFormat="Nrm2" */
        pInfo->dataFormat[1]==0x72 &&
        pInfo->dataFormat[2]==0x6d &&
        pInfo->dataFormat[3]==0x32 &&
        (1<=formatVersion0 && formatVersion0<=4)
    )) {
        udata_printError(ds, "unorm2_swap(): data format %02x.%02x.%02x.%02x (format version %02x) is not recognized as Normalizer2 data\n",
                         pInfo->dataFormat[0], pInfo->dataFormat[1],
                         pInfo->dataFormat[2], pInfo->dataFormat[3],
                         pInfo->formatVersion[0]);
        *pErrorCode=U_UNSUPPORTED_ERROR;
        return 0;
    }

    inBytes=(const uint8_t *)inData+headerSize;
    outBytes=(uint8_t *)outData+headerSize;

    inIndexes=(const int32_t *)inBytes;
    int32_t minIndexesLength;
    if(formatVersion0==1) {
        minIndexesLength=Normalizer2Impl::IX_MIN_MAYBE_YES+1;
    } else if(formatVersion0==2) {
        minIndexesLength=Normalizer2Impl::IX_MIN_YES_NO_MAPPINGS_ONLY+1;
    } else {
        minIndexesLength=Normalizer2Impl::IX_MIN_LCCC_CP+1;
    }

    if(length>=0) {
        length-=headerSize;
        if(length<minIndexesLength*4) {
            udata_printError(ds, "unorm2_swap(): too few bytes (%d after header) for Normalizer2 data\n",
                             length);
            *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }
    }

    /* read the first few indexes */
    for(i=0; i<UPRV_LENGTHOF(indexes); ++i) {
        indexes[i]=udata_readInt32(ds, inIndexes[i]);
    }

    /* get the total length of the data */
    size=indexes[Normalizer2Impl::IX_TOTAL_SIZE];

    if(length>=0) {
        if(length<size) {
            udata_printError(ds, "unorm2_swap(): too few bytes (%d after header) for all of Normalizer2 data\n",
                             length);
            *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }

        /* copy the data for inaccessible bytes */
        if(inBytes!=outBytes) {
            uprv_memcpy(outBytes, inBytes, size);
        }

        offset=0;

        /* swap the int32_t indexes[] */
        nextOffset=indexes[Normalizer2Impl::IX_NORM_TRIE_OFFSET];
        ds->swapArray32(ds, inBytes, nextOffset-offset, outBytes, pErrorCode);
        offset=nextOffset;

        /* swap the trie */
        nextOffset=indexes[Normalizer2Impl::IX_EXTRA_DATA_OFFSET];
        utrie_swapAnyVersion(ds, inBytes+offset, nextOffset-offset, outBytes+offset, pErrorCode);
        offset=nextOffset;

        /* swap the uint16_t extraData[] */
        nextOffset=indexes[Normalizer2Impl::IX_SMALL_FCD_OFFSET];
        ds->swapArray16(ds, inBytes+offset, nextOffset-offset, outBytes+offset, pErrorCode);
        offset=nextOffset;

        /* no need to swap the uint8_t smallFCD[] (new in formatVersion 2) */
        nextOffset=indexes[Normalizer2Impl::IX_SMALL_FCD_OFFSET+1];
        offset=nextOffset;

        U_ASSERT(offset==size);
    }

    return headerSize+size;
}

#endif  // !UCONFIG_NO_NORMALIZATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *************************************************************************
 * COPYRIGHT: 
 * Copyright (c) 1996-2012, International Business Machines Corporation and
 * others. All Rights Reserved.
 *************************************************************************
 */



#if !UCONFIG_NO_NORMALIZATION




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1998-2005, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File schriter.h
*
* Modification History:
*
*   Date        Name        Description
*  05/05/99     stephen     Cleaned up.
******************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1998-2005, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/




#if U_SHOW_CPLUSPLUS_API



/**
 * \file 
 * \brief C++ API: char16_t Character Iterator
 */
 
U_NAMESPACE_BEGIN

/**
 * A concrete subclass of CharacterIterator that iterates over the
 * characters (code units or code points) in a char16_t array.
 * It's possible not only to create an
 * iterator that iterates over an entire char16_t array, but also to
 * create one that iterates over only a subrange of a char16_t array
 * (iterators over different subranges of the same char16_t array don't
 * compare equal).
 * @see CharacterIterator
 * @see ForwardCharacterIterator
 * @stable ICU 2.0
 */
class U_COMMON_API UCharCharacterIterator : public CharacterIterator {
public:
  /**
   * Create an iterator over the char16_t array referred to by "textPtr".
   * The iteration range is 0 to <code>length-1</code>.
   * text is only aliased, not adopted (the
   * destructor will not delete it).
   * @param textPtr The char16_t array to be iterated over
   * @param length The length of the char16_t array
   * @stable ICU 2.0
   */
  UCharCharacterIterator(ConstChar16Ptr textPtr, int32_t length);

  /**
   * Create an iterator over the char16_t array referred to by "textPtr".
   * The iteration range is 0 to <code>length-1</code>.
   * text is only aliased, not adopted (the
   * destructor will not delete it).
   * The starting
   * position is specified by "position". If "position" is outside the valid
   * iteration range, the behavior of this object is undefined.
   * @param textPtr The char16_t array to be iteratd over
   * @param length The length of the char16_t array
   * @param position The starting position of the iteration
   * @stable ICU 2.0
   */
  UCharCharacterIterator(ConstChar16Ptr textPtr, int32_t length,
                         int32_t position);

  /**
   * Create an iterator over the char16_t array referred to by "textPtr".
   * The iteration range is 0 to <code>end-1</code>.
   * text is only aliased, not adopted (the
   * destructor will not delete it).
   * The starting
   * position is specified by "position". If begin and end do not
   * form a valid iteration range or "position" is outside the valid
   * iteration range, the behavior of this object is undefined.
   * @param textPtr The char16_t array to be iterated over
   * @param length The length of the char16_t array
   * @param textBegin  The begin position of the iteration range
   * @param textEnd    The end position of the iteration range
   * @param position    The starting position of the iteration
   * @stable ICU 2.0
   */
  UCharCharacterIterator(ConstChar16Ptr textPtr, int32_t length,
                         int32_t textBegin,
                         int32_t textEnd,
                         int32_t position);

  /**
   * Copy constructor.  The new iterator iterates over the same range
   * of the same string as "that", and its initial position is the
   * same as "that"'s current position.
   * @param that The UCharCharacterIterator to be copied
   * @stable ICU 2.0
   */
  UCharCharacterIterator(const UCharCharacterIterator&  that);

  /**
   * Destructor.
   * @stable ICU 2.0
   */
  virtual ~UCharCharacterIterator();

  /**
   * Assignment operator.  *this is altered to iterate over the sane
   * range of the same string as "that", and refers to the same
   * character within that string as "that" does.
   * @param that The object to be copied
   * @return the newly created object
   * @stable ICU 2.0
   */
  UCharCharacterIterator&
  operator=(const UCharCharacterIterator&    that);

  /**
   * Returns true if the iterators iterate over the same range of the
   * same string and are pointing at the same character.
   * @param that The ForwardCharacterIterator used to be compared for equality
   * @return true if the iterators iterate over the same range of the
   * same string and are pointing at the same character.
   * @stable ICU 2.0
   */
  virtual UBool          operator==(const ForwardCharacterIterator& that) const;

  /**
   * Generates a hash code for this iterator.
   * @return the hash code.
   * @stable ICU 2.0
   */
  virtual int32_t         hashCode(void) const;

  /**
   * Returns a new UCharCharacterIterator referring to the same
   * character in the same range of the same string as this one.  The
   * caller must delete the new iterator.
   * @return the CharacterIterator newly created
   * @stable ICU 2.0
   */
  virtual UCharCharacterIterator* clone() const;

  /**
   * Sets the iterator to refer to the first code unit in its
   * iteration range, and returns that code unit.
   * This can be used to begin an iteration with next().
   * @return the first code unit in its iteration range.
   * @stable ICU 2.0
   */
  virtual char16_t         first(void);

  /**
   * Sets the iterator to refer to the first code unit in its
   * iteration range, returns that code unit, and moves the position
   * to the second code unit. This is an alternative to setToStart()
   * for forward iteration with nextPostInc().
   * @return the first code unit in its iteration range
   * @stable ICU 2.0
   */
  virtual char16_t         firstPostInc(void);

  /**
   * Sets the iterator to refer to the first code point in its
   * iteration range, and returns that code unit,
   * This can be used to begin an iteration with next32().
   * Note that an iteration with next32PostInc(), beginning with,
   * e.g., setToStart() or firstPostInc(), is more efficient.
   * @return the first code point in its iteration range
   * @stable ICU 2.0
   */
  virtual UChar32       first32(void);

  /**
   * Sets the iterator to refer to the first code point in its
   * iteration range, returns that code point, and moves the position
   * to the second code point. This is an alternative to setToStart()
   * for forward iteration with next32PostInc().
   * @return the first code point in its iteration range.
   * @stable ICU 2.0
   */
  virtual UChar32       first32PostInc(void);

  /**
   * Sets the iterator to refer to the last code unit in its
   * iteration range, and returns that code unit.
   * This can be used to begin an iteration with previous().
   * @return the last code unit in its iteration range.
   * @stable ICU 2.0
   */
  virtual char16_t         last(void);

  /**
   * Sets the iterator to refer to the last code point in its
   * iteration range, and returns that code unit.
   * This can be used to begin an iteration with previous32().
   * @return the last code point in its iteration range.
   * @stable ICU 2.0
   */
  virtual UChar32       last32(void);

  /**
   * Sets the iterator to refer to the "position"-th code unit
   * in the text-storage object the iterator refers to, and
   * returns that code unit.
   * @param position the position within the text-storage object
   * @return the code unit
   * @stable ICU 2.0
   */
  virtual char16_t         setIndex(int32_t position);

  /**
   * Sets the iterator to refer to the beginning of the code point
   * that contains the "position"-th code unit
   * in the text-storage object the iterator refers to, and
   * returns that code point.
   * The current position is adjusted to the beginning of the code point
   * (its first code unit).
   * @param position the position within the text-storage object
   * @return the code unit
   * @stable ICU 2.0
   */
  virtual UChar32       setIndex32(int32_t position);

  /**
   * Returns the code unit the iterator currently refers to.
   * @return the code unit the iterator currently refers to.
   * @stable ICU 2.0
   */
  virtual char16_t         current(void) const;

  /**
   * Returns the code point the iterator currently refers to.
   * @return the code point the iterator currently refers to.
   * @stable ICU 2.0
   */
  virtual UChar32       current32(void) const;

  /**
   * Advances to the next code unit in the iteration range (toward
   * endIndex()), and returns that code unit.  If there are no more
   * code units to return, returns DONE.
   * @return the next code unit in the iteration range.
   * @stable ICU 2.0
   */
  virtual char16_t         next(void);

  /**
   * Gets the current code unit for returning and advances to the next code unit
   * in the iteration range
   * (toward endIndex()).  If there are
   * no more code units to return, returns DONE.
   * @return the current code unit.
   * @stable ICU 2.0
   */
  virtual char16_t         nextPostInc(void);

  /**
   * Advances to the next code point in the iteration range (toward
   * endIndex()), and returns that code point.  If there are no more
   * code points to return, returns DONE.
   * Note that iteration with "pre-increment" semantics is less
   * efficient than iteration with "post-increment" semantics
   * that is provided by next32PostInc().
   * @return the next code point in the iteration range.
   * @stable ICU 2.0
   */
  virtual UChar32       next32(void);

  /**
   * Gets the current code point for returning and advances to the next code point
   * in the iteration range
   * (toward endIndex()).  If there are
   * no more code points to return, returns DONE.
   * @return the current point.
   * @stable ICU 2.0
   */
  virtual UChar32       next32PostInc(void);

  /**
   * Returns FALSE if there are no more code units or code points
   * at or after the current position in the iteration range.
   * This is used with nextPostInc() or next32PostInc() in forward
   * iteration.
   * @return FALSE if there are no more code units or code points
   * at or after the current position in the iteration range.
   * @stable ICU 2.0
   */
  virtual UBool        hasNext();

  /**
   * Advances to the previous code unit in the iteration range (toward
   * startIndex()), and returns that code unit.  If there are no more
   * code units to return, returns DONE.
   * @return the previous code unit in the iteration range.
   * @stable ICU 2.0
   */
  virtual char16_t         previous(void);

  /**
   * Advances to the previous code point in the iteration range (toward
   * startIndex()), and returns that code point.  If there are no more
   * code points to return, returns DONE.
   * @return the previous code point in the iteration range.
   * @stable ICU 2.0
   */
  virtual UChar32       previous32(void);

  /**
   * Returns FALSE if there are no more code units or code points
   * before the current position in the iteration range.
   * This is used with previous() or previous32() in backward
   * iteration.
   * @return FALSE if there are no more code units or code points
   * before the current position in the iteration range.
   * @stable ICU 2.0
   */
  virtual UBool        hasPrevious();

  /**
   * Moves the current position relative to the start or end of the
   * iteration range, or relative to the current position itself.
   * The movement is expressed in numbers of code units forward
   * or backward by specifying a positive or negative delta.
   * @param delta the position relative to origin. A positive delta means forward;
   * a negative delta means backward.
   * @param origin Origin enumeration {kStart, kCurrent, kEnd}
   * @return the new position
   * @stable ICU 2.0
   */
  virtual int32_t      move(int32_t delta, EOrigin origin);

  /**
   * Moves the current position relative to the start or end of the
   * iteration range, or relative to the current position itself.
   * The movement is expressed in numbers of code points forward
   * or backward by specifying a positive or negative delta.
   * @param delta the position relative to origin. A positive delta means forward;
   * a negative delta means backward.
   * @param origin Origin enumeration {kStart, kCurrent, kEnd}
   * @return the new position
   * @stable ICU 2.0
   */
#ifdef move32
   // One of the system headers right now is sometimes defining a conflicting macro we don't use
#undef move32
#endif
  virtual int32_t      move32(int32_t delta, EOrigin origin);

  /**
   * Sets the iterator to iterate over a new range of text
   * @stable ICU 2.0
   */
  void setText(ConstChar16Ptr newText, int32_t newTextLength);

  /**
   * Copies the char16_t array under iteration into the UnicodeString
   * referred to by "result".  Even if this iterator iterates across
   * only a part of this string, the whole string is copied.
   * @param result Receives a copy of the text under iteration.
   * @stable ICU 2.0
   */
  virtual void            getText(UnicodeString& result);

  /**
   * Return a class ID for this class (not really public)
   * @return a class ID for this class
   * @stable ICU 2.0
   */
  static UClassID         U_EXPORT2 getStaticClassID(void);

  /**
   * Return a class ID for this object (not really public)
   * @return a class ID for this object.
   * @stable ICU 2.0
   */
  virtual UClassID        getDynamicClassID(void) const;

protected:
  /**
   * Protected constructor
   * @stable ICU 2.0
   */
  UCharCharacterIterator();
  /**
   * Protected member text
   * @stable ICU 2.0
   */
  const char16_t*            text;

};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */



/**
 * \file 
 * \brief C++ API: String Character Iterator
 */
 
U_NAMESPACE_BEGIN
/**
 * A concrete subclass of CharacterIterator that iterates over the
 * characters (code units or code points) in a UnicodeString.
 * It's possible not only to create an
 * iterator that iterates over an entire UnicodeString, but also to
 * create one that iterates over only a subrange of a UnicodeString
 * (iterators over different subranges of the same UnicodeString don't
 * compare equal).
 * @see CharacterIterator
 * @see ForwardCharacterIterator
 * @stable ICU 2.0
 */
class U_COMMON_API StringCharacterIterator : public UCharCharacterIterator {
public:
  /**
   * Create an iterator over the UnicodeString referred to by "textStr".
   * The UnicodeString object is copied.
   * The iteration range is the whole string, and the starting position is 0.
   * @param textStr The unicode string used to create an iterator
   * @stable ICU 2.0
   */
  StringCharacterIterator(const UnicodeString& textStr);

  /**
   * Create an iterator over the UnicodeString referred to by "textStr".
   * The iteration range is the whole string, and the starting
   * position is specified by "textPos".  If "textPos" is outside the valid
   * iteration range, the behavior of this object is undefined.
   * @param textStr The unicode string used to create an iterator
   * @param textPos The starting position of the iteration
   * @stable ICU 2.0
   */
  StringCharacterIterator(const UnicodeString&    textStr,
              int32_t              textPos);

  /**
   * Create an iterator over the UnicodeString referred to by "textStr".
   * The UnicodeString object is copied.
   * The iteration range begins with the code unit specified by
   * "textBegin" and ends with the code unit BEFORE the code unit specified
   * by "textEnd".  The starting position is specified by "textPos".  If
   * "textBegin" and "textEnd" don't form a valid range on "text" (i.e.,
   * textBegin >= textEnd or either is negative or greater than text.size()),
   * or "textPos" is outside the range defined by "textBegin" and "textEnd",
   * the behavior of this iterator is undefined.
   * @param textStr    The unicode string used to create the StringCharacterIterator
   * @param textBegin  The begin position of the iteration range
   * @param textEnd    The end position of the iteration range
   * @param textPos    The starting position of the iteration
   * @stable ICU 2.0
   */
  StringCharacterIterator(const UnicodeString&    textStr,
              int32_t              textBegin,
              int32_t              textEnd,
              int32_t              textPos);

  /**
   * Copy constructor.  The new iterator iterates over the same range
   * of the same string as "that", and its initial position is the
   * same as "that"'s current position.
   * The UnicodeString object in "that" is copied.
   * @param that The StringCharacterIterator to be copied
   * @stable ICU 2.0
   */
  StringCharacterIterator(const StringCharacterIterator&  that);

  /**
   * Destructor.
   * @stable ICU 2.0
   */
  virtual ~StringCharacterIterator();

  /**
   * Assignment operator.  *this is altered to iterate over the same
   * range of the same string as "that", and refers to the same
   * character within that string as "that" does.
   * @param that The object to be copied.
   * @return the newly created object.
   * @stable ICU 2.0
   */
  StringCharacterIterator&
  operator=(const StringCharacterIterator&    that);

  /**
   * Returns true if the iterators iterate over the same range of the
   * same string and are pointing at the same character.
   * @param that The ForwardCharacterIterator to be compared for equality
   * @return true if the iterators iterate over the same range of the
   * same string and are pointing at the same character.
   * @stable ICU 2.0
   */
  virtual UBool          operator==(const ForwardCharacterIterator& that) const;

  /**
   * Returns a new StringCharacterIterator referring to the same
   * character in the same range of the same string as this one.  The
   * caller must delete the new iterator.
   * @return the newly cloned object.
   * @stable ICU 2.0
   */
  virtual StringCharacterIterator* clone() const;

  /**
   * Sets the iterator to iterate over the provided string.
   * @param newText The string to be iterated over
   * @stable ICU 2.0
   */
  void setText(const UnicodeString& newText);

  /**
   * Copies the UnicodeString under iteration into the UnicodeString
   * referred to by "result".  Even if this iterator iterates across
   * only a part of this string, the whole string is copied.
   * @param result Receives a copy of the text under iteration.
   * @stable ICU 2.0
   */
  virtual void            getText(UnicodeString& result);

  /**
   * Return a class ID for this object (not really public)
   * @return a class ID for this object.
   * @stable ICU 2.0
   */
  virtual UClassID         getDynamicClassID(void) const;

  /**
   * Return a class ID for this class (not really public)
   * @return a class ID for this class
   * @stable ICU 2.0
   */
  static UClassID   U_EXPORT2 getStaticClassID(void);

protected:
  /**
   * Default constructor, iteration over empty string.
   * @stable ICU 2.0
   */
  StringCharacterIterator();

  /**
   * Sets the iterator to iterate over the provided string.
   * @param newText The string to be iterated over
   * @param newTextLength The length of the String
   * @stable ICU 2.0
   */
  void setText(const char16_t* newText, int32_t newTextLength);

  /**
   * Copy of the iterated string object.
   * @stable ICU 2.0
   */
  UnicodeString            text;

};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ********************************************************************
 * COPYRIGHT:
 * Copyright (c) 1996-2015, International Business Machines Corporation and
 * others. All Rights Reserved.
 ********************************************************************
 */




#if U_SHOW_CPLUSPLUS_API

/**
 * \file 
 * \brief C++ API: Unicode Normalization
 */
 
#if !UCONFIG_NO_NORMALIZATION







U_NAMESPACE_BEGIN
/**
 * Old Unicode normalization API.
 *
 * This API has been replaced by the Normalizer2 class and is only available
 * for backward compatibility. This class simply delegates to the Normalizer2 class.
 * There is one exception: The new API does not provide a replacement for Normalizer::compare().
 *
 * The Normalizer class supports the standard normalization forms described in
 * <a href="http://www.unicode.org/unicode/reports/tr15/" target="unicode">
 * Unicode Standard Annex #15: Unicode Normalization Forms</a>.
 *
 * The Normalizer class consists of two parts:
 * - static functions that normalize strings or test if strings are normalized
 * - a Normalizer object is an iterator that takes any kind of text and
 *   provides iteration over its normalized form
 *
 * The Normalizer class is not suitable for subclassing.
 *
 * For basic information about normalization forms and details about the C API
 * please see the documentation in unorm.h.
 *
 * The iterator API with the Normalizer constructors and the non-static functions
 * use a CharacterIterator as input. It is possible to pass a string which
 * is then internally wrapped in a CharacterIterator.
 * The input text is not normalized all at once, but incrementally where needed
 * (providing efficient random access).
 * This allows to pass in a large text but spend only a small amount of time
 * normalizing a small part of that text.
 * However, if the entire text is normalized, then the iterator will be
 * slower than normalizing the entire text at once and iterating over the result.
 * A possible use of the Normalizer iterator is also to report an index into the
 * original text that is close to where the normalized characters come from.
 *
 * <em>Important:</em> The iterator API was cleaned up significantly for ICU 2.0.
 * The earlier implementation reported the getIndex() inconsistently,
 * and previous() could not be used after setIndex(), next(), first(), and current().
 *
 * Normalizer allows to start normalizing from anywhere in the input text by
 * calling setIndexOnly(), first(), or last().
 * Without calling any of these, the iterator will start at the beginning of the text.
 *
 * At any time, next() returns the next normalized code point (UChar32),
 * with post-increment semantics (like CharacterIterator::next32PostInc()).
 * previous() returns the previous normalized code point (UChar32),
 * with pre-decrement semantics (like CharacterIterator::previous32()).
 *
 * current() returns the current code point
 * (respectively the one at the newly set index) without moving
 * the getIndex(). Note that if the text at the current position
 * needs to be normalized, then these functions will do that.
 * (This is why current() is not const.)
 * It is more efficient to call setIndexOnly() instead, which does not
 * normalize.
 *
 * getIndex() always refers to the position in the input text where the normalized
 * code points are returned from. It does not always change with each returned
 * code point.
 * The code point that is returned from any of the functions
 * corresponds to text at or after getIndex(), according to the
 * function's iteration semantics (post-increment or pre-decrement).
 *
 * next() returns a code point from at or after the getIndex()
 * from before the next() call. After the next() call, the getIndex()
 * might have moved to where the next code point will be returned from
 * (from a next() or current() call).
 * This is semantically equivalent to array access with array[index++]
 * (post-increment semantics).
 *
 * previous() returns a code point from at or after the getIndex()
 * from after the previous() call.
 * This is semantically equivalent to array access with array[--index]
 * (pre-decrement semantics).
 *
 * Internally, the Normalizer iterator normalizes a small piece of text
 * starting at the getIndex() and ending at a following "safe" index.
 * The normalized results is stored in an internal string buffer, and
 * the code points are iterated from there.
 * With multiple iteration calls, this is repeated until the next piece
 * of text needs to be normalized, and the getIndex() needs to be moved.
 *
 * The following "safe" index, the internal buffer, and the secondary
 * iteration index into that buffer are not exposed on the API.
 * This also means that it is currently not practical to return to
 * a particular, arbitrary position in the text because one would need to
 * know, and be able to set, in addition to the getIndex(), at least also the
 * current index into the internal buffer.
 * It is currently only possible to observe when getIndex() changes
 * (with careful consideration of the iteration semantics),
 * at which time the internal index will be 0.
 * For example, if getIndex() is different after next() than before it,
 * then the internal index is 0 and one can return to this getIndex()
 * later with setIndexOnly().
 *
 * Note: While the setIndex() and getIndex() refer to indices in the
 * underlying Unicode input text, the next() and previous() methods
 * iterate through characters in the normalized output.
 * This means that there is not necessarily a one-to-one correspondence
 * between characters returned by next() and previous() and the indices
 * passed to and returned from setIndex() and getIndex().
 * It is for this reason that Normalizer does not implement the CharacterIterator interface.
 *
 * @author Laura Werner, Mark Davis, Markus Scherer
 * @stable ICU 2.0
 */
class U_COMMON_API Normalizer : public UObject {
public:
#ifndef U_HIDE_DEPRECATED_API
  /**
   * If DONE is returned from an iteration function that returns a code point,
   * then there are no more normalization results available.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  enum {
      DONE=0xffff
  };

  // Constructors

  /**
   * Creates a new <code>Normalizer</code> object for iterating over the
   * normalized form of a given string.
   * <p>
   * @param str   The string to be normalized.  The normalization
   *              will start at the beginning of the string.
   *
   * @param mode  The normalization mode.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  Normalizer(const UnicodeString& str, UNormalizationMode mode);

  /**
   * Creates a new <code>Normalizer</code> object for iterating over the
   * normalized form of a given string.
   * <p>
   * @param str   The string to be normalized.  The normalization
   *              will start at the beginning of the string.
   *
   * @param length Length of the string, or -1 if NUL-terminated.
   * @param mode  The normalization mode.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  Normalizer(ConstChar16Ptr str, int32_t length, UNormalizationMode mode);

  /**
   * Creates a new <code>Normalizer</code> object for iterating over the
   * normalized form of the given text.
   * <p>
   * @param iter  The input text to be normalized.  The normalization
   *              will start at the beginning of the string.
   *
   * @param mode  The normalization mode.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  Normalizer(const CharacterIterator& iter, UNormalizationMode mode);
#endif  /* U_HIDE_DEPRECATED_API */

#ifndef U_FORCE_HIDE_DEPRECATED_API
  /**
   * Copy constructor.
   * @param copy The object to be copied.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  Normalizer(const Normalizer& copy);

  /**
   * Destructor
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  virtual ~Normalizer();
#endif  // U_FORCE_HIDE_DEPRECATED_API

  //-------------------------------------------------------------------------
  // Static utility methods
  //-------------------------------------------------------------------------

#ifndef U_HIDE_DEPRECATED_API
  /**
   * Normalizes a <code>UnicodeString</code> according to the specified normalization mode.
   * This is a wrapper for unorm_normalize(), using UnicodeString's.
   *
   * The <code>options</code> parameter specifies which optional
   * <code>Normalizer</code> features are to be enabled for this operation.
   *
   * @param source    the input string to be normalized.
   * @param mode      the normalization mode
   * @param options   the optional features to be enabled (0 for no options)
   * @param result    The normalized string (on output).
   * @param status    The error code.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  static void U_EXPORT2 normalize(const UnicodeString& source,
                        UNormalizationMode mode, int32_t options,
                        UnicodeString& result,
                        UErrorCode &status);

  /**
   * Compose a <code>UnicodeString</code>.
   * This is equivalent to normalize() with mode UNORM_NFC or UNORM_NFKC.
   * This is a wrapper for unorm_normalize(), using UnicodeString's.
   *
   * The <code>options</code> parameter specifies which optional
   * <code>Normalizer</code> features are to be enabled for this operation.
   *
   * @param source    the string to be composed.
   * @param compat    Perform compatibility decomposition before composition.
   *                  If this argument is <code>FALSE</code>, only canonical
   *                  decomposition will be performed.
   * @param options   the optional features to be enabled (0 for no options)
   * @param result    The composed string (on output).
   * @param status    The error code.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  static void U_EXPORT2 compose(const UnicodeString& source,
                      UBool compat, int32_t options,
                      UnicodeString& result,
                      UErrorCode &status);

  /**
   * Static method to decompose a <code>UnicodeString</code>.
   * This is equivalent to normalize() with mode UNORM_NFD or UNORM_NFKD.
   * This is a wrapper for unorm_normalize(), using UnicodeString's.
   *
   * The <code>options</code> parameter specifies which optional
   * <code>Normalizer</code> features are to be enabled for this operation.
   *
   * @param source    the string to be decomposed.
   * @param compat    Perform compatibility decomposition.
   *                  If this argument is <code>FALSE</code>, only canonical
   *                  decomposition will be performed.
   * @param options   the optional features to be enabled (0 for no options)
   * @param result    The decomposed string (on output).
   * @param status    The error code.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  static void U_EXPORT2 decompose(const UnicodeString& source,
                        UBool compat, int32_t options,
                        UnicodeString& result,
                        UErrorCode &status);

  /**
   * Performing quick check on a string, to quickly determine if the string is
   * in a particular normalization format.
   * This is a wrapper for unorm_quickCheck(), using a UnicodeString.
   *
   * Three types of result can be returned UNORM_YES, UNORM_NO or
   * UNORM_MAYBE. Result UNORM_YES indicates that the argument
   * string is in the desired normalized format, UNORM_NO determines that
   * argument string is not in the desired normalized format. A
   * UNORM_MAYBE result indicates that a more thorough check is required,
   * the user may have to put the string in its normalized form and compare the
   * results.
   * @param source       string for determining if it is in a normalized format
   * @param mode         normalization format
   * @param status A reference to a UErrorCode to receive any errors
   * @return UNORM_YES, UNORM_NO or UNORM_MAYBE
   *
   * @see isNormalized
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  static inline UNormalizationCheckResult
  quickCheck(const UnicodeString &source, UNormalizationMode mode, UErrorCode &status);

  /**
   * Performing quick check on a string; same as the other version of quickCheck
   * but takes an extra options parameter like most normalization functions.
   *
   * @param source       string for determining if it is in a normalized format
   * @param mode         normalization format
   * @param options      the optional features to be enabled (0 for no options)
   * @param status A reference to a UErrorCode to receive any errors
   * @return UNORM_YES, UNORM_NO or UNORM_MAYBE
   *
   * @see isNormalized
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  static UNormalizationCheckResult
  quickCheck(const UnicodeString &source, UNormalizationMode mode, int32_t options, UErrorCode &status);

  /**
   * Test if a string is in a given normalization form.
   * This is semantically equivalent to source.equals(normalize(source, mode)) .
   *
   * Unlike unorm_quickCheck(), this function returns a definitive result,
   * never a "maybe".
   * For NFD, NFKD, and FCD, both functions work exactly the same.
   * For NFC and NFKC where quickCheck may return "maybe", this function will
   * perform further tests to arrive at a TRUE/FALSE result.
   *
   * @param src        String that is to be tested if it is in a normalization format.
   * @param mode       Which normalization form to test for.
   * @param errorCode  ICU error code in/out parameter.
   *                   Must fulfill U_SUCCESS before the function call.
   * @return Boolean value indicating whether the source string is in the
   *         "mode" normalization form.
   *
   * @see quickCheck
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  static inline UBool
  isNormalized(const UnicodeString &src, UNormalizationMode mode, UErrorCode &errorCode);

  /**
   * Test if a string is in a given normalization form; same as the other version of isNormalized
   * but takes an extra options parameter like most normalization functions.
   *
   * @param src        String that is to be tested if it is in a normalization format.
   * @param mode       Which normalization form to test for.
   * @param options      the optional features to be enabled (0 for no options)
   * @param errorCode  ICU error code in/out parameter.
   *                   Must fulfill U_SUCCESS before the function call.
   * @return Boolean value indicating whether the source string is in the
   *         "mode" normalization form.
   *
   * @see quickCheck
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  static UBool
  isNormalized(const UnicodeString &src, UNormalizationMode mode, int32_t options, UErrorCode &errorCode);

  /**
   * Concatenate normalized strings, making sure that the result is normalized as well.
   *
   * If both the left and the right strings are in
   * the normalization form according to "mode/options",
   * then the result will be
   *
   * \code
   *     dest=normalize(left+right, mode, options)
   * \endcode
   *
   * For details see unorm_concatenate in unorm.h.
   *
   * @param left Left source string.
   * @param right Right source string.
   * @param result The output string.
   * @param mode The normalization mode.
   * @param options A bit set of normalization options.
   * @param errorCode ICU error code in/out parameter.
   *                   Must fulfill U_SUCCESS before the function call.
   * @return result
   *
   * @see unorm_concatenate
   * @see normalize
   * @see unorm_next
   * @see unorm_previous
   *
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  static UnicodeString &
  U_EXPORT2 concatenate(const UnicodeString &left, const UnicodeString &right,
              UnicodeString &result,
              UNormalizationMode mode, int32_t options,
              UErrorCode &errorCode);
#endif  /* U_HIDE_DEPRECATED_API */

  /**
   * Compare two strings for canonical equivalence.
   * Further options include case-insensitive comparison and
   * code point order (as opposed to code unit order).
   *
   * Canonical equivalence between two strings is defined as their normalized
   * forms (NFD or NFC) being identical.
   * This function compares strings incrementally instead of normalizing
   * (and optionally case-folding) both strings entirely,
   * improving performance significantly.
   *
   * Bulk normalization is only necessary if the strings do not fulfill the FCD
   * conditions. Only in this case, and only if the strings are relatively long,
   * is memory allocated temporarily.
   * For FCD strings and short non-FCD strings there is no memory allocation.
   *
   * Semantically, this is equivalent to
   *   strcmp[CodePointOrder](NFD(foldCase(s1)), NFD(foldCase(s2)))
   * where code point order and foldCase are all optional.
   *
   * UAX 21 2.5 Caseless Matching specifies that for a canonical caseless match
   * the case folding must be performed first, then the normalization.
   *
   * @param s1 First source string.
   * @param s2 Second source string.
   *
   * @param options A bit set of options:
   *   - U_FOLD_CASE_DEFAULT or 0 is used for default options:
   *     Case-sensitive comparison in code unit order, and the input strings
   *     are quick-checked for FCD.
   *
   *   - UNORM_INPUT_IS_FCD
   *     Set if the caller knows that both s1 and s2 fulfill the FCD conditions.
   *     If not set, the function will quickCheck for FCD
   *     and normalize if necessary.
   *
   *   - U_COMPARE_CODE_POINT_ORDER
   *     Set to choose code point order instead of code unit order
   *     (see u_strCompare for details).
   *
   *   - U_COMPARE_IGNORE_CASE
   *     Set to compare strings case-insensitively using case folding,
   *     instead of case-sensitively.
   *     If set, then the following case folding options are used.
   *
   *   - Options as used with case-insensitive comparisons, currently:
   *
   *   - U_FOLD_CASE_EXCLUDE_SPECIAL_I
   *    (see u_strCaseCompare for details)
   *
   *   - regular normalization options shifted left by UNORM_COMPARE_NORM_OPTIONS_SHIFT
   *
   * @param errorCode ICU error code in/out parameter.
   *                  Must fulfill U_SUCCESS before the function call.
   * @return <0 or 0 or >0 as usual for string comparisons
   *
   * @see unorm_compare
   * @see normalize
   * @see UNORM_FCD
   * @see u_strCompare
   * @see u_strCaseCompare
   *
   * @stable ICU 2.2
   */
  static inline int32_t
  compare(const UnicodeString &s1, const UnicodeString &s2,
          uint32_t options,
          UErrorCode &errorCode);

#ifndef U_HIDE_DEPRECATED_API
  //-------------------------------------------------------------------------
  // Iteration API
  //-------------------------------------------------------------------------

  /**
   * Return the current character in the normalized text.
   * current() may need to normalize some text at getIndex().
   * The getIndex() is not changed.
   *
   * @return the current normalized code point
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  UChar32              current(void);

  /**
   * Return the first character in the normalized text.
   * This is equivalent to setIndexOnly(startIndex()) followed by next().
   * (Post-increment semantics.)
   *
   * @return the first normalized code point
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  UChar32              first(void);

  /**
   * Return the last character in the normalized text.
   * This is equivalent to setIndexOnly(endIndex()) followed by previous().
   * (Pre-decrement semantics.)
   *
   * @return the last normalized code point
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  UChar32              last(void);

  /**
   * Return the next character in the normalized text.
   * (Post-increment semantics.)
   * If the end of the text has already been reached, DONE is returned.
   * The DONE value could be confused with a U+FFFF non-character code point
   * in the text. If this is possible, you can test getIndex()<endIndex()
   * before calling next(), or (getIndex()<endIndex() || last()!=DONE)
   * after calling next(). (Calling last() will change the iterator state!)
   *
   * The C API unorm_next() is more efficient and does not have this ambiguity.
   *
   * @return the next normalized code point
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  UChar32              next(void);

  /**
   * Return the previous character in the normalized text and decrement.
   * (Pre-decrement semantics.)
   * If the beginning of the text has already been reached, DONE is returned.
   * The DONE value could be confused with a U+FFFF non-character code point
   * in the text. If this is possible, you can test
   * (getIndex()>startIndex() || first()!=DONE). (Calling first() will change
   * the iterator state!)
   *
   * The C API unorm_previous() is more efficient and does not have this ambiguity.
   *
   * @return the previous normalized code point
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  UChar32              previous(void);

  /**
   * Set the iteration position in the input text that is being normalized,
   * without any immediate normalization.
   * After setIndexOnly(), getIndex() will return the same index that is
   * specified here.
   *
   * @param index the desired index in the input text.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  void                 setIndexOnly(int32_t index);

  /**
   * Reset the index to the beginning of the text.
   * This is equivalent to setIndexOnly(startIndex)).
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  void                reset(void);

  /**
   * Retrieve the current iteration position in the input text that is
   * being normalized.
   *
   * A following call to next() will return a normalized code point from
   * the input text at or after this index.
   *
   * After a call to previous(), getIndex() will point at or before the
   * position in the input text where the normalized code point
   * was returned from with previous().
   *
   * @return the current index in the input text
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  int32_t            getIndex(void) const;

  /**
   * Retrieve the index of the start of the input text. This is the begin index
   * of the <code>CharacterIterator</code> or the start (i.e. index 0) of the string
   * over which this <code>Normalizer</code> is iterating.
   *
   * @return the smallest index in the input text where the Normalizer operates
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  int32_t            startIndex(void) const;

  /**
   * Retrieve the index of the end of the input text. This is the end index
   * of the <code>CharacterIterator</code> or the length of the string
   * over which this <code>Normalizer</code> is iterating.
   * This end index is exclusive, i.e., the Normalizer operates only on characters
   * before this index.
   *
   * @return the first index in the input text where the Normalizer does not operate
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  int32_t            endIndex(void) const;

  /**
   * Returns TRUE when both iterators refer to the same character in the same
   * input text.
   *
   * @param that a Normalizer object to compare this one to
   * @return comparison result
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  UBool        operator==(const Normalizer& that) const;

  /**
   * Returns FALSE when both iterators refer to the same character in the same
   * input text.
   *
   * @param that a Normalizer object to compare this one to
   * @return comparison result
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  inline UBool        operator!=(const Normalizer& that) const;

  /**
   * Returns a pointer to a new Normalizer that is a clone of this one.
   * The caller is responsible for deleting the new clone.
   * @return a pointer to a new Normalizer
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  Normalizer*        clone() const;

  /**
   * Generates a hash code for this iterator.
   *
   * @return the hash code
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  int32_t                hashCode(void) const;

  //-------------------------------------------------------------------------
  // Property access methods
  //-------------------------------------------------------------------------

  /**
   * Set the normalization mode for this object.
   * <p>
   * <b>Note:</b>If the normalization mode is changed while iterating
   * over a string, calls to {@link #next() } and {@link #previous() } may
   * return previously buffers characters in the old normalization mode
   * until the iteration is able to re-sync at the next base character.
   * It is safest to call {@link #setIndexOnly }, {@link #reset() },
   * {@link #setText }, {@link #first() },
   * {@link #last() }, etc. after calling <code>setMode</code>.
   * <p>
   * @param newMode the new mode for this <code>Normalizer</code>.
   * @see #getUMode
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  void setMode(UNormalizationMode newMode);

  /**
   * Return the normalization mode for this object.
   *
   * This is an unusual name because there used to be a getMode() that
   * returned a different type.
   *
   * @return the mode for this <code>Normalizer</code>
   * @see #setMode
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  UNormalizationMode getUMode(void) const;

  /**
   * Set options that affect this <code>Normalizer</code>'s operation.
   * Options do not change the basic composition or decomposition operation
   * that is being performed, but they control whether
   * certain optional portions of the operation are done.
   * Currently the only available option is obsolete.
   *
   * It is possible to specify multiple options that are all turned on or off.
   *
   * @param   option  the option(s) whose value is/are to be set.
   * @param   value   the new setting for the option.  Use <code>TRUE</code> to
   *                  turn the option(s) on and <code>FALSE</code> to turn it/them off.
   *
   * @see #getOption
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  void setOption(int32_t option,
         UBool value);

  /**
   * Determine whether an option is turned on or off.
   * If multiple options are specified, then the result is TRUE if any
   * of them are set.
   * <p>
   * @param option the option(s) that are to be checked
   * @return TRUE if any of the option(s) are set
   * @see #setOption
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  UBool getOption(int32_t option) const;

  /**
   * Set the input text over which this <code>Normalizer</code> will iterate.
   * The iteration position is set to the beginning.
   *
   * @param newText a string that replaces the current input text
   * @param status a UErrorCode
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  void setText(const UnicodeString& newText,
           UErrorCode &status);

  /**
   * Set the input text over which this <code>Normalizer</code> will iterate.
   * The iteration position is set to the beginning.
   *
   * @param newText a CharacterIterator object that replaces the current input text
   * @param status a UErrorCode
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  void setText(const CharacterIterator& newText,
           UErrorCode &status);

  /**
   * Set the input text over which this <code>Normalizer</code> will iterate.
   * The iteration position is set to the beginning.
   *
   * @param newText a string that replaces the current input text
   * @param length the length of the string, or -1 if NUL-terminated
   * @param status a UErrorCode
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  void setText(ConstChar16Ptr newText,
                    int32_t length,
            UErrorCode &status);
  /**
   * Copies the input text into the UnicodeString argument.
   *
   * @param result Receives a copy of the text under iteration.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  void            getText(UnicodeString&  result);

  /**
   * ICU "poor man's RTTI", returns a UClassID for this class.
   * @returns a UClassID for this class.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  static UClassID U_EXPORT2 getStaticClassID();
#endif  /* U_HIDE_DEPRECATED_API */

#ifndef U_FORCE_HIDE_DEPRECATED_API
  /**
   * ICU "poor man's RTTI", returns a UClassID for the actual class.
   * @return a UClassID for the actual class.
   * @deprecated ICU 56 Use Normalizer2 instead.
   */
  virtual UClassID getDynamicClassID() const;
#endif  // U_FORCE_HIDE_DEPRECATED_API

private:
  //-------------------------------------------------------------------------
  // Private functions
  //-------------------------------------------------------------------------

  Normalizer(); // default constructor not implemented
  Normalizer &operator=(const Normalizer &that); // assignment operator not implemented

  // Private utility methods for iteration
  // For documentation, see the source code
  UBool nextNormalize();
  UBool previousNormalize();

  void    init();
  void    clearBuffer(void);

  //-------------------------------------------------------------------------
  // Private data
  //-------------------------------------------------------------------------

  FilteredNormalizer2*fFilteredNorm2;  // owned if not NULL
  const Normalizer2  *fNorm2;  // not owned; may be equal to fFilteredNorm2
  UNormalizationMode  fUMode;  // deprecated
  int32_t             fOptions;

  // The input text and our position in it
  CharacterIterator  *text;

  // The normalization buffer is the result of normalization
  // of the source in [currentIndex..nextIndex[ .
  int32_t         currentIndex, nextIndex;

  // A buffer for holding intermediate results
  UnicodeString       buffer;
  int32_t         bufferPos;
};

//-------------------------------------------------------------------------
// Inline implementations
//-------------------------------------------------------------------------

#ifndef U_HIDE_DEPRECATED_API
inline UBool
Normalizer::operator!= (const Normalizer& other) const
{ return ! operator==(other); }

inline UNormalizationCheckResult
Normalizer::quickCheck(const UnicodeString& source,
                       UNormalizationMode mode,
                       UErrorCode &status) {
    return quickCheck(source, mode, 0, status);
}

inline UBool
Normalizer::isNormalized(const UnicodeString& source,
                         UNormalizationMode mode,
                         UErrorCode &status) {
    return isNormalized(source, mode, 0, status);
}
#endif  /* U_HIDE_DEPRECATED_API */

inline int32_t
Normalizer::compare(const UnicodeString &s1, const UnicodeString &s2,
                    uint32_t options,
                    UErrorCode &errorCode) {
  // all argument checking is done in unorm_compare
  return unorm_compare(toUCharPtr(s1.getBuffer()), s1.length(),
                       toUCharPtr(s2.getBuffer()), s2.length(),
                       options,
                       &errorCode);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_NORMALIZATION */

#endif // NORMLZR_H





  // for uniset_getUnicode32Instance()

#if defined(move32)
 // System can define move32 intrinsics, but the char iters define move32 method
 // using same undef trick in headers, so undef here to re-enable the method.
#undef move32
#endif

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(Normalizer)

//-------------------------------------------------------------------------
// Constructors and other boilerplate
//-------------------------------------------------------------------------

Normalizer::Normalizer(const UnicodeString& str, UNormalizationMode mode) :
    UObject(), fFilteredNorm2(NULL), fNorm2(NULL), fUMode(mode), fOptions(0),
    text(new StringCharacterIterator(str)),
    currentIndex(0), nextIndex(0),
    buffer(), bufferPos(0)
{
    init();
}

Normalizer::Normalizer(ConstChar16Ptr str, int32_t length, UNormalizationMode mode) :
    UObject(), fFilteredNorm2(NULL), fNorm2(NULL), fUMode(mode), fOptions(0),
    text(new UCharCharacterIterator(str, length)),
    currentIndex(0), nextIndex(0),
    buffer(), bufferPos(0)
{
    init();
}

Normalizer::Normalizer(const CharacterIterator& iter, UNormalizationMode mode) :
    UObject(), fFilteredNorm2(NULL), fNorm2(NULL), fUMode(mode), fOptions(0),
    text(iter.clone()),
    currentIndex(0), nextIndex(0),
    buffer(), bufferPos(0)
{
    init();
}

Normalizer::Normalizer(const Normalizer &copy) :
    UObject(copy), fFilteredNorm2(NULL), fNorm2(NULL), fUMode(copy.fUMode), fOptions(copy.fOptions),
    text(copy.text->clone()),
    currentIndex(copy.currentIndex), nextIndex(copy.nextIndex),
    buffer(copy.buffer), bufferPos(copy.bufferPos)
{
    init();
}

void
Normalizer::init() {
    UErrorCode errorCode=U_ZERO_ERROR;
    fNorm2=Normalizer2Factory::getInstance(fUMode, errorCode);
    if(fOptions&UNORM_UNICODE_3_2) {
        delete fFilteredNorm2;
        fNorm2=fFilteredNorm2=
            new FilteredNormalizer2(*fNorm2, *uniset_getUnicode32Instance(errorCode));
    }
    if(U_FAILURE(errorCode)) {
        errorCode=U_ZERO_ERROR;
        fNorm2=Normalizer2Factory::getNoopInstance(errorCode);
    }
}

Normalizer::~Normalizer()
{
    delete fFilteredNorm2;
    delete text;
}

Normalizer* 
Normalizer::clone() const
{
    return new Normalizer(*this);
}

/**
 * Generates a hash code for this iterator.
 */
int32_t Normalizer::hashCode() const
{
    return text->hashCode() + fUMode + fOptions + buffer.hashCode() + bufferPos + currentIndex + nextIndex;
}
    
UBool Normalizer::operator==(const Normalizer& that) const
{
    return
        this==&that ||
        (fUMode==that.fUMode &&
        fOptions==that.fOptions &&
        *text==*that.text &&
        buffer==that.buffer &&
        bufferPos==that.bufferPos &&
        nextIndex==that.nextIndex);
}

//-------------------------------------------------------------------------
// Static utility methods
//-------------------------------------------------------------------------

void U_EXPORT2
Normalizer::normalize(const UnicodeString& source, 
                      UNormalizationMode mode, int32_t options,
                      UnicodeString& result, 
                      UErrorCode &status) {
    if(source.isBogus() || U_FAILURE(status)) {
        result.setToBogus();
        if(U_SUCCESS(status)) {
            status=U_ILLEGAL_ARGUMENT_ERROR;
        }
    } else {
        UnicodeString localDest;
        UnicodeString *dest;

        if(&source!=&result) {
            dest=&result;
        } else {
            // the source and result strings are the same object, use a temporary one
            dest=&localDest;
        }
        const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, status);
        if(U_SUCCESS(status)) {
            if(options&UNORM_UNICODE_3_2) {
                FilteredNormalizer2(*n2, *uniset_getUnicode32Instance(status)).
                    normalize(source, *dest, status);
            } else {
                n2->normalize(source, *dest, status);
            }
        }
        if(dest==&localDest && U_SUCCESS(status)) {
            result=*dest;
        }
    }
}

void U_EXPORT2
Normalizer::compose(const UnicodeString& source, 
                    UBool compat, int32_t options,
                    UnicodeString& result, 
                    UErrorCode &status) {
    normalize(source, compat ? UNORM_NFKC : UNORM_NFC, options, result, status);
}

void U_EXPORT2
Normalizer::decompose(const UnicodeString& source, 
                      UBool compat, int32_t options,
                      UnicodeString& result, 
                      UErrorCode &status) {
    normalize(source, compat ? UNORM_NFKD : UNORM_NFD, options, result, status);
}

UNormalizationCheckResult
Normalizer::quickCheck(const UnicodeString& source,
                       UNormalizationMode mode, int32_t options,
                       UErrorCode &status) {
    const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, status);
    if(U_SUCCESS(status)) {
        if(options&UNORM_UNICODE_3_2) {
            return FilteredNormalizer2(*n2, *uniset_getUnicode32Instance(status)).
                quickCheck(source, status);
        } else {
            return n2->quickCheck(source, status);
        }
    } else {
        return UNORM_MAYBE;
    }
}

UBool
Normalizer::isNormalized(const UnicodeString& source,
                         UNormalizationMode mode, int32_t options,
                         UErrorCode &status) {
    const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, status);
    if(U_SUCCESS(status)) {
        if(options&UNORM_UNICODE_3_2) {
            return FilteredNormalizer2(*n2, *uniset_getUnicode32Instance(status)).
                isNormalized(source, status);
        } else {
            return n2->isNormalized(source, status);
        }
    } else {
        return FALSE;
    }
}

UnicodeString & U_EXPORT2
Normalizer::concatenate(const UnicodeString &left, const UnicodeString &right,
                        UnicodeString &result,
                        UNormalizationMode mode, int32_t options,
                        UErrorCode &errorCode) {
    if(left.isBogus() || right.isBogus() || U_FAILURE(errorCode)) {
        result.setToBogus();
        if(U_SUCCESS(errorCode)) {
            errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        }
    } else {
        UnicodeString localDest;
        UnicodeString *dest;

        if(&right!=&result) {
            dest=&result;
        } else {
            // the right and result strings are the same object, use a temporary one
            dest=&localDest;
        }
        *dest=left;
        const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, errorCode);
        if(U_SUCCESS(errorCode)) {
            if(options&UNORM_UNICODE_3_2) {
                FilteredNormalizer2(*n2, *uniset_getUnicode32Instance(errorCode)).
                    append(*dest, right, errorCode);
            } else {
                n2->append(*dest, right, errorCode);
            }
        }
        if(dest==&localDest && U_SUCCESS(errorCode)) {
            result=*dest;
        }
    }
    return result;
}

//-------------------------------------------------------------------------
// Iteration API
//-------------------------------------------------------------------------

/**
 * Return the current character in the normalized text.
 */
UChar32 Normalizer::current() {
    if(bufferPos<buffer.length() || nextNormalize()) {
        return buffer.char32At(bufferPos);
    } else {
        return DONE;
    }
}

/**
 * Return the next character in the normalized text and advance
 * the iteration position by one.  If the end
 * of the text has already been reached, {@link #DONE} is returned.
 */
UChar32 Normalizer::next() {
    if(bufferPos<buffer.length() ||  nextNormalize()) {
        UChar32 c=buffer.char32At(bufferPos);
        bufferPos+=U16_LENGTH(c);
        return c;
    } else {
        return DONE;
    }
}

/**
 * Return the previous character in the normalized text and decrement
 * the iteration position by one.  If the beginning
 * of the text has already been reached, {@link #DONE} is returned.
 */
UChar32 Normalizer::previous() {
    if(bufferPos>0 || previousNormalize()) {
        UChar32 c=buffer.char32At(bufferPos-1);
        bufferPos-=U16_LENGTH(c);
        return c;
    } else {
        return DONE;
    }
}

void Normalizer::reset() {
    currentIndex=nextIndex=text->setToStart();
    clearBuffer();
}

void
Normalizer::setIndexOnly(int32_t index) {
    text->setIndex(index);  // pins index
    currentIndex=nextIndex=text->getIndex();
    clearBuffer();
}

/**
 * Return the first character in the normalized text.  This resets
 * the <tt>Normalizer's</tt> position to the beginning of the text.
 */
UChar32 Normalizer::first() {
    reset();
    return next();
}

/**
 * Return the last character in the normalized text.  This resets
 * the <tt>Normalizer's</tt> position to be just before the
 * the input text corresponding to that normalized character.
 */
UChar32 Normalizer::last() {
    currentIndex=nextIndex=text->setToEnd();
    clearBuffer();
    return previous();
}

/**
 * Retrieve the current iteration position in the input text that is
 * being normalized.  This method is useful in applications such as
 * searching, where you need to be able to determine the position in
 * the input text that corresponds to a given normalized output character.
 * <p>
 * <b>Note:</b> This method sets the position in the <em>input</em>, while
 * {@link #next} and {@link #previous} iterate through characters in the
 * <em>output</em>.  This means that there is not necessarily a one-to-one
 * correspondence between characters returned by <tt>next</tt> and
 * <tt>previous</tt> and the indices passed to and returned from
 * <tt>setIndex</tt> and {@link #getIndex}.
 *
 */
int32_t Normalizer::getIndex() const {
    if(bufferPos<buffer.length()) {
        return currentIndex;
    } else {
        return nextIndex;
    }
}

/**
 * Retrieve the index of the start of the input text.  This is the begin index
 * of the <tt>CharacterIterator</tt> or the start (i.e. 0) of the <tt>String</tt>
 * over which this <tt>Normalizer</tt> is iterating
 */
int32_t Normalizer::startIndex() const {
    return text->startIndex();
}

/**
 * Retrieve the index of the end of the input text.  This is the end index
 * of the <tt>CharacterIterator</tt> or the length of the <tt>String</tt>
 * over which this <tt>Normalizer</tt> is iterating
 */
int32_t Normalizer::endIndex() const {
    return text->endIndex();
}

//-------------------------------------------------------------------------
// Property access methods
//-------------------------------------------------------------------------

void
Normalizer::setMode(UNormalizationMode newMode) 
{
    fUMode = newMode;
    init();
}

UNormalizationMode
Normalizer::getUMode() const
{
    return fUMode;
}

void
Normalizer::setOption(int32_t option, 
                      UBool value) 
{
    if (value) {
        fOptions |= option;
    } else {
        fOptions &= (~option);
    }
    init();
}

UBool
Normalizer::getOption(int32_t option) const
{
    return (fOptions & option) != 0;
}

/**
 * Set the input text over which this <tt>Normalizer</tt> will iterate.
 * The iteration position is set to the beginning of the input text.
 */
void
Normalizer::setText(const UnicodeString& newText, 
                    UErrorCode &status)
{
    if (U_FAILURE(status)) {
        return;
    }
    CharacterIterator *newIter = new StringCharacterIterator(newText);
    if (newIter == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    delete text;
    text = newIter;
    reset();
}

/**
 * Set the input text over which this <tt>Normalizer</tt> will iterate.
 * The iteration position is set to the beginning of the string.
 */
void
Normalizer::setText(const CharacterIterator& newText, 
                    UErrorCode &status) 
{
    if (U_FAILURE(status)) {
        return;
    }
    CharacterIterator *newIter = newText.clone();
    if (newIter == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    delete text;
    text = newIter;
    reset();
}

void
Normalizer::setText(ConstChar16Ptr newText,
                    int32_t length,
                    UErrorCode &status)
{
    if (U_FAILURE(status)) {
        return;
    }
    CharacterIterator *newIter = new UCharCharacterIterator(newText, length);
    if (newIter == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    delete text;
    text = newIter;
    reset();
}

/**
 * Copies the text under iteration into the UnicodeString referred to by "result".
 * @param result Receives a copy of the text under iteration.
 */
void
Normalizer::getText(UnicodeString&  result) 
{
    text->getText(result);
}

//-------------------------------------------------------------------------
// Private utility methods
//-------------------------------------------------------------------------

void Normalizer::clearBuffer() {
    buffer.remove();
    bufferPos=0;
}

UBool
Normalizer::nextNormalize() {
    clearBuffer();
    currentIndex=nextIndex;
    text->setIndex(nextIndex);
    if(!text->hasNext()) {
        return FALSE;
    }
    // Skip at least one character so we make progress.
    UnicodeString segment(text->next32PostInc());
    while(text->hasNext()) {
        UChar32 c;
        if(fNorm2->hasBoundaryBefore(c=text->next32PostInc())) {
            text->move32(-1, CharacterIterator::kCurrent);
            break;
        }
        segment.append(c);
    }
    nextIndex=text->getIndex();
    UErrorCode errorCode=U_ZERO_ERROR;
    fNorm2->normalize(segment, buffer, errorCode);
    return U_SUCCESS(errorCode) && !buffer.isEmpty();
}

UBool
Normalizer::previousNormalize() {
    clearBuffer();
    nextIndex=currentIndex;
    text->setIndex(currentIndex);
    if(!text->hasPrevious()) {
        return FALSE;
    }
    UnicodeString segment;
    while(text->hasPrevious()) {
        UChar32 c=text->previous32();
        segment.insert(0, c);
        if(fNorm2->hasBoundaryBefore(c)) {
            break;
        }
    }
    currentIndex=text->getIndex();
    UErrorCode errorCode=U_ZERO_ERROR;
    fNorm2->normalize(segment, buffer, errorCode);
    bufferPos=buffer.length();
    return U_SUCCESS(errorCode) && !buffer.isEmpty();
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_NORMALIZATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2003-2003, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
* Copyright (C) 1997-2005, International Business Machines Corporation and others. All Rights Reserved.
*******************************************************************************
*
* File PARSEPOS.H
*
* Modification History:
*
*   Date        Name        Description
*   07/09/97    helena      Converted from java.
*   07/17/98    stephen     Added errorIndex support.
*   05/11/99    stephen     Cleaned up.
*******************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API



 
U_NAMESPACE_BEGIN

/**
 * \file
 * \brief C++ API: Canonical Iterator
 */
/** 
 * <code>ParsePosition</code> is a simple class used by <code>Format</code>
 * and its subclasses to keep track of the current position during parsing.
 * The <code>parseObject</code> method in the various <code>Format</code>
 * classes requires a <code>ParsePosition</code> object as an argument.
 *
 * <p>
 * By design, as you parse through a string with different formats,
 * you can use the same <code>ParsePosition</code>, since the index parameter
 * records the current position.
 *
 * The ParsePosition class is not suitable for subclassing.
 *
 * @version     1.3 10/30/97
 * @author      Mark Davis, Helena Shih
 * @see         java.text.Format
 */

class U_COMMON_API ParsePosition : public UObject {
public:
    /**
     * Default constructor, the index starts with 0 as default.
     * @stable ICU 2.0
     */
    ParsePosition()
        : UObject(),
        index(0),
        errorIndex(-1)
      {}

    /**
     * Create a new ParsePosition with the given initial index.
     * @param newIndex the new text offset.
     * @stable ICU 2.0
     */
    ParsePosition(int32_t newIndex)
        : UObject(),
        index(newIndex),
        errorIndex(-1)
      {}

    /**
     * Copy constructor
     * @param copy the object to be copied from.
     * @stable ICU 2.0
     */
    ParsePosition(const ParsePosition& copy)
        : UObject(copy),
        index(copy.index),
        errorIndex(copy.errorIndex)
      {}

    /**
     * Destructor
     * @stable ICU 2.0
     */
    virtual ~ParsePosition();

    /**
     * Assignment operator
     * @stable ICU 2.0
     */
    inline ParsePosition&      operator=(const ParsePosition& copy);

    /**
     * Equality operator.
     * @return TRUE if the two parse positions are equal, FALSE otherwise.
     * @stable ICU 2.0
     */
    inline UBool              operator==(const ParsePosition& that) const;

    /**
     * Equality operator.
     * @return TRUE if the two parse positions are not equal, FALSE otherwise.
     * @stable ICU 2.0
     */
    inline UBool              operator!=(const ParsePosition& that) const;

    /**
     * Clone this object.
     * Clones can be used concurrently in multiple threads.
     * If an error occurs, then NULL is returned.
     * The caller must delete the clone.
     *
     * @return a clone of this object
     *
     * @see getDynamicClassID
     * @stable ICU 2.8
     */
    ParsePosition *clone() const;

    /**
     * Retrieve the current parse position.  On input to a parse method, this
     * is the index of the character at which parsing will begin; on output, it
     * is the index of the character following the last character parsed.
     * @return the current index.
     * @stable ICU 2.0
     */
    inline int32_t getIndex(void) const;

    /**
     * Set the current parse position.
     * @param index the new index.
     * @stable ICU 2.0
     */
    inline void setIndex(int32_t index);

    /**
     * Set the index at which a parse error occurred.  Formatters
     * should set this before returning an error code from their
     * parseObject method.  The default value is -1 if this is not
     * set.
     * @stable ICU 2.0
     */
    inline void setErrorIndex(int32_t ei);

    /**
     * Retrieve the index at which an error occurred, or -1 if the
     * error index has not been set.
     * @stable ICU 2.0
     */
    inline int32_t getErrorIndex(void) const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 2.2
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 2.2
     */
    virtual UClassID getDynamicClassID() const;

private:
    /**
     * Input: the place you start parsing.
     * <br>Output: position where the parse stopped.
     * This is designed to be used serially,
     * with each call setting index up for the next one.
     */
    int32_t index;

    /**
     * The index at which a parse error occurred.
     */
    int32_t errorIndex;

};

inline ParsePosition&
ParsePosition::operator=(const ParsePosition& copy)
{
  index = copy.index;
  errorIndex = copy.errorIndex;
  return *this;
}

inline UBool
ParsePosition::operator==(const ParsePosition& copy) const
{
  if(index != copy.index || errorIndex != copy.errorIndex)
  return FALSE;
  else
  return TRUE;
}

inline UBool
ParsePosition::operator!=(const ParsePosition& copy) const
{
  return !operator==(copy);
}

inline int32_t
ParsePosition::getIndex() const
{
  return index;
}

inline void
ParsePosition::setIndex(int32_t offset)
{
  this->index = offset;
}

inline int32_t
ParsePosition::getErrorIndex() const
{
  return errorIndex;
}

inline void
ParsePosition::setErrorIndex(int32_t ei)
{
  this->errorIndex = ei;
}
U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */



U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(ParsePosition)

ParsePosition::~ParsePosition() {}

ParsePosition *
ParsePosition::clone() const {
    return new ParsePosition(*this);
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  patternprops.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011mar13
*   created by: Markus W. Scherer
*/




U_NAMESPACE_BEGIN

/*
 * One byte per Latin-1 character.
 * Bit 0 is set if either Pattern property is true,
 * bit 1 if Pattern_Syntax is true,
 * bit 2 if Pattern_White_Space is true.
 * That is, Pattern_Syntax is encoded as 3 and Pattern_White_Space as 5.
 */
static const uint8_t latin1[256]={
    // WS: 9..D
    0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 5, 5, 5, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    // WS: 20  Syntax: 21..2F
    5, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    // Syntax: 3A..40
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 3, 3, 3, 3, 3,
    3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    // Syntax: 5B..5E
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 3, 3, 3, 0,
    // Syntax: 60
    3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    // Syntax: 7B..7E
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 3, 3, 3, 0,
    // WS: 85
    0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    // Syntax: A1..A7, A9, AB, AC, AE
    0, 3, 3, 3, 3, 3, 3, 3, 0, 3, 0, 3, 3, 0, 3, 0,
    // Syntax: B0, B1, B6, BB, BF
    3, 3, 0, 0, 0, 0, 3, 0, 0, 0, 0, 3, 0, 0, 0, 3,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    // Syntax: D7
    0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    // Syntax: F7
    0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0
};

/*
 * One byte per 32 characters from U+2000..U+303F indexing into
 * a small table of 32-bit data words.
 * The first two data words are all-zeros and all-ones.
 */
static const uint8_t index2000[130]={
    2, 3, 4, 0, 0, 0, 0, 0,  // 20xx
    0, 0, 0, 0, 5, 1, 1, 1,  // 21xx
    1, 1, 1, 1, 1, 1, 1, 1,  // 22xx
    1, 1, 1, 1, 1, 1, 1, 1,  // 23xx
    1, 1, 1, 0, 0, 0, 0, 0,  // 24xx
    1, 1, 1, 1, 1, 1, 1, 1,  // 25xx
    1, 1, 1, 1, 1, 1, 1, 1,  // 26xx
    1, 1, 1, 6, 7, 1, 1, 1,  // 27xx
    1, 1, 1, 1, 1, 1, 1, 1,  // 28xx
    1, 1, 1, 1, 1, 1, 1, 1,  // 29xx
    1, 1, 1, 1, 1, 1, 1, 1,  // 2Axx
    1, 1, 1, 1, 1, 1, 1, 1,  // 2Bxx
    0, 0, 0, 0, 0, 0, 0, 0,  // 2Cxx
    0, 0, 0, 0, 0, 0, 0, 0,  // 2Dxx
    1, 1, 1, 1, 0, 0, 0, 0,  // 2Exx
    0, 0, 0, 0, 0, 0, 0, 0,  // 2Fxx
    8, 9  // 3000..303F
};

/*
 * One 32-bit integer per 32 characters. Ranges of all-false and all-true
 * are mapped to the first two values, other ranges map to appropriate bit patterns.
 */
static const uint32_t syntax2000[]={
    0,
    0xffffffff,
    0xffff0000,  // 2: 2010..201F
    0x7fff00ff,  // 3: 2020..2027, 2030..203E
    0x7feffffe,  // 4: 2041..2053, 2055..205E
    0xffff0000,  // 5: 2190..219F
    0x003fffff,  // 6: 2760..2775
    0xfff00000,  // 7: 2794..279F
    0xffffff0e,  // 8: 3001..3003, 3008..301F
    0x00010001   // 9: 3020, 3030
};

/*
 * Same as syntax2000, but with additional bits set for the
 * Pattern_White_Space characters 200E 200F 2028 2029.
 */
static const uint32_t syntaxOrWhiteSpace2000[]={
    0,
    0xffffffff,
    0xffffc000,  // 2: 200E..201F
    0x7fff03ff,  // 3: 2020..2029, 2030..203E
    0x7feffffe,  // 4: 2041..2053, 2055..205E
    0xffff0000,  // 5: 2190..219F
    0x003fffff,  // 6: 2760..2775
    0xfff00000,  // 7: 2794..279F
    0xffffff0e,  // 8: 3001..3003, 3008..301F
    0x00010001   // 9: 3020, 3030
};

UBool
PatternProps::isSyntax(UChar32 c) {
    if(c<0) {
        return FALSE;
    } else if(c<=0xff) {
        return (UBool)(latin1[c]>>1)&1;
    } else if(c<0x2010) {
        return FALSE;
    } else if(c<=0x3030) {
        uint32_t bits=syntax2000[index2000[(c-0x2000)>>5]];
        return (UBool)((bits>>(c&0x1f))&1);
    } else if(0xfd3e<=c && c<=0xfe46) {
        return c<=0xfd3f || 0xfe45<=c;
    } else {
        return FALSE;
    }
}

UBool
PatternProps::isSyntaxOrWhiteSpace(UChar32 c) {
    if(c<0) {
        return FALSE;
    } else if(c<=0xff) {
        return (UBool)(latin1[c]&1);
    } else if(c<0x200e) {
        return FALSE;
    } else if(c<=0x3030) {
        uint32_t bits=syntaxOrWhiteSpace2000[index2000[(c-0x2000)>>5]];
        return (UBool)((bits>>(c&0x1f))&1);
    } else if(0xfd3e<=c && c<=0xfe46) {
        return c<=0xfd3f || 0xfe45<=c;
    } else {
        return FALSE;
    }
}

UBool
PatternProps::isWhiteSpace(UChar32 c) {
    if(c<0) {
        return FALSE;
    } else if(c<=0xff) {
        return (UBool)(latin1[c]>>2)&1;
    } else if(0x200e<=c && c<=0x2029) {
        return c<=0x200f || 0x2028<=c;
    } else {
        return FALSE;
    }
}

const UChar *
PatternProps::skipWhiteSpace(const UChar *s, int32_t length) {
    while(length>0 && isWhiteSpace(*s)) {
        ++s;
        --length;
    }
    return s;
}

int32_t
PatternProps::skipWhiteSpace(const UnicodeString& s, int32_t start) {
    int32_t i = start;
    int32_t length = s.length();
    while(i<length && isWhiteSpace(s.charAt(i))) {
        ++i;
    }
    return i;
}

const UChar *
PatternProps::trimWhiteSpace(const UChar *s, int32_t &length) {
    if(length<=0 || (!isWhiteSpace(s[0]) && !isWhiteSpace(s[length-1]))) {
        return s;
    }
    int32_t start=0;
    int32_t limit=length;
    while(start<limit && isWhiteSpace(s[start])) {
        ++start;
    }
    if(start<limit) {
        // There is non-white space at start; we will not move limit below that,
        // so we need not test start<limit in the loop.
        while(isWhiteSpace(s[limit-1])) {
            --limit;
        }
    }
    length=limit-start;
    return s+start;
}

UBool
PatternProps::isIdentifier(const UChar *s, int32_t length) {
    if(length<=0) {
        return FALSE;
    }
    const UChar *limit=s+length;
    do {
        if(isSyntaxOrWhiteSpace(*s++)) {
            return FALSE;
        }
    } while(s<limit);
    return TRUE;
}

const UChar *
PatternProps::skipIdentifier(const UChar *s, int32_t length) {
    while(length>0 && !isSyntaxOrWhiteSpace(*s)) {
        ++s;
        --length;
    }
    return s;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 * Copyright (C) 2015, International Business Machines Corporation and
 * others. All Rights Reserved.
 */




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2015, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*
* File pluralmap.h - PluralMap class that maps plural categories to values.
******************************************************************************
*/





U_NAMESPACE_BEGIN

class UnicodeString;

class U_COMMON_API PluralMapBase : public UMemory {
public:
    /**
     * The names of all the plural categories. NONE is not an actual plural
     * category, but rather represents the absense of a plural category.
     */
    enum Category {
        NONE = -1,
        OTHER,
        ZERO,
        ONE,
        TWO,
        FEW,
        MANY,
        CATEGORY_COUNT
    };

    /**
     * Converts a category name such as "zero", "one", "two", "few", "many"
     * or "other" to a category enum. Returns NONE for an unrecognized
     * category name.
     */
    static Category toCategory(const char *categoryName);

    /**
     * Converts a category name such as "zero", "one", "two", "few", "many"
     * or "other" to a category enum.  Returns NONE for urecongized
     * category name.
     */
    static Category toCategory(const UnicodeString &categoryName);

    /**
     * Converts a category to a name.
     * Passing NONE or CATEGORY_COUNT for category returns NULL.
     */
    static const char *getCategoryName(Category category);
};

/**
 * A Map of plural categories to values. It maintains ownership of the
 * values.
 *
 * Type T is the value type. T must provide the followng:
 * 1) Default constructor
 * 2) Copy constructor
 * 3) Assignment operator
 * 4) Must extend UMemory
 */
template<typename T>
class PluralMap : public PluralMapBase {
public:
    /**
     * Other category is maps to a copy of the default value.
     */
    PluralMap() : fOtherVariant() {
        initializeNew();
    }

    /**
     * Other category is mapped to otherVariant.
     */
    PluralMap(const T &otherVariant) : fOtherVariant(otherVariant) {
        initializeNew();
    }

    PluralMap(const PluralMap<T> &other) : fOtherVariant(other.fOtherVariant) {
        fVariants[0] = &fOtherVariant;
        for (int32_t i = 1; i < UPRV_LENGTHOF(fVariants); ++i) {
            fVariants[i] = other.fVariants[i] ?
                    new T(*other.fVariants[i]) : NULL;
        }
    }

    PluralMap<T> &operator=(const PluralMap<T> &other) {
        if (this == &other) {
            return *this;
        }
        for (int32_t i = 0; i < UPRV_LENGTHOF(fVariants); ++i) {
            if (fVariants[i] != NULL && other.fVariants[i] != NULL) {
                *fVariants[i] = *other.fVariants[i];
            } else if (fVariants[i] != NULL) {
                delete fVariants[i];
                fVariants[i] = NULL;
            } else if (other.fVariants[i] != NULL) {
                fVariants[i] = new T(*other.fVariants[i]);
            } else {
                // do nothing
            }
        }
        return *this;
    }

    ~PluralMap() {
        for (int32_t i = 1; i < UPRV_LENGTHOF(fVariants); ++i) {
            delete fVariants[i];
        }
    }

    /**
     * Removes all mappings and makes 'other' point to the default value.
     */
    void clear() {
        *fVariants[0] = T();
        for (int32_t i = 1; i < UPRV_LENGTHOF(fVariants); ++i) {
            delete fVariants[i];
            fVariants[i] = NULL;
        }
    }

    /**
     * Iterates through the mappings in this instance, set index to NONE
     * prior to using. Call next repeatedly to get the values until it
     * returns NULL. Each time next returns, caller may pass index
     * to getCategoryName() to get the name of the plural category.
     * When this function returns NULL, index is CATEGORY_COUNT
     */
    const T *next(Category &index) const {
        int32_t idx = index;
        ++idx;
        for (; idx < UPRV_LENGTHOF(fVariants); ++idx) {
            if (fVariants[idx] != NULL) {
                index = static_cast<Category>(idx);
                return fVariants[idx];
            }
        }
        index = static_cast<Category>(idx);
        return NULL;
    }

    /**
     * non const version of next.
     */
    T *nextMutable(Category &index) {
        const T *result = next(index);
        return const_cast<T *>(result);
    }

    /**
     * Returns the 'other' variant.
     * Same as calling get(OTHER).
     */
    const T &getOther() const {
        return get(OTHER);
    }

    /**
     * Returns the value associated with a category.
     * If no value found, or v is NONE or CATEGORY_COUNT, falls
     * back to returning the value for the 'other' category.
     */
    const T &get(Category v) const {
        int32_t index = v;
        if (index < 0 || index >= UPRV_LENGTHOF(fVariants) || fVariants[index] == NULL) {
            return *fVariants[0];
        }
        return *fVariants[index];
    }

    /**
     * Convenience routine to get the value by category name. Otherwise
     * works just like get(Category).
     */
    const T &get(const char *category) const {
        return get(toCategory(category));
    }

    /**
     * Convenience routine to get the value by category name as a
     * UnicodeString. Otherwise works just like get(category).
     */
    const T &get(const UnicodeString &category) const {
        return get(toCategory(category));
    }

    /**
     * Returns a pointer to the value associated with a category
     * that caller can safely modify. If the value was defaulting to the 'other'
     * variant because no explicit value was stored, this method creates a
     * new value using the default constructor at the returned pointer.
     *
     * @param category the category with the value to change.
     * @param status error returned here if index is NONE or CATEGORY_COUNT
     *  or memory could not be allocated, or any other error happens.
     */
    T *getMutable(
            Category category,
            UErrorCode &status) {
        return getMutable(category, NULL, status);
    }

    /**
     * Convenience routine to get a mutable pointer to a value by category name.
     * Otherwise works just like getMutable(Category, UErrorCode &).
     * reports an error if the category name is invalid.
     */
    T *getMutable(
            const char *category,
            UErrorCode &status) {
        return getMutable(toCategory(category), NULL, status);
    }

    /**
     * Just like getMutable(Category, UErrorCode &) but copies defaultValue to
     * returned pointer if it was defaulting to the 'other' variant
     * because no explicit value was stored.
     */
    T *getMutableWithDefault(
            Category category,
            const T &defaultValue,
            UErrorCode &status) {
        return getMutable(category, &defaultValue, status);
    }

    /**
     * Returns TRUE if this object equals rhs.
     */
    UBool equals(
            const PluralMap<T> &rhs,
            UBool (*eqFunc)(const T &, const T &)) const {
        for (int32_t i = 0; i < UPRV_LENGTHOF(fVariants); ++i) {
            if (fVariants[i] == rhs.fVariants[i]) {
                continue;
            }
            if (fVariants[i] == NULL || rhs.fVariants[i] == NULL) {
                return FALSE;
            }
            if (!eqFunc(*fVariants[i], *rhs.fVariants[i])) {
                return FALSE;
            }
        }
        return TRUE;
    }

private:
    T fOtherVariant;
    T* fVariants[6];

    T *getMutable(
            Category category,
            const T *defaultValue,
            UErrorCode &status) {
        if (U_FAILURE(status)) {
            return NULL;
        }
        int32_t index = category;
        if (index < 0 || index >= UPRV_LENGTHOF(fVariants)) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return NULL;
        }
        if (fVariants[index] == NULL) {
            fVariants[index] = defaultValue == NULL ?
                    new T() : new T(*defaultValue);
        }
        if (!fVariants[index]) {
            status = U_MEMORY_ALLOCATION_ERROR;
        }
        return fVariants[index];
    }

    void initializeNew() {
        fVariants[0] = &fOtherVariant;
        for (int32_t i = 1; i < UPRV_LENGTHOF(fVariants); ++i) {
            fVariants[i] = NULL;
        }
    }
};

U_NAMESPACE_END



U_NAMESPACE_BEGIN

static const char * const gPluralForms[] = {
        "other", "zero", "one", "two", "few", "many"};

PluralMapBase::Category
PluralMapBase::toCategory(const char *pluralForm) {
    for (int32_t i = 0; i < UPRV_LENGTHOF(gPluralForms); ++i) {
        if (uprv_strcmp(pluralForm, gPluralForms[i]) == 0) {
            return static_cast<Category>(i);
        }
    }
    return NONE;
}

PluralMapBase::Category
PluralMapBase::toCategory(const UnicodeString &pluralForm) {
    CharString cCategory;
    UErrorCode status = U_ZERO_ERROR;
    cCategory.appendInvariantChars(pluralForm, status);    
    return U_FAILURE(status) ? NONE : toCategory(cCategory.data());
}

const char *PluralMapBase::getCategoryName(Category c) {
    int32_t index = c;
    return (index < 0 || index >= UPRV_LENGTHOF(gPluralForms)) ?
            NULL : gPluralForms[index];
}


U_NAMESPACE_END

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2002-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: October 30 2002
* Since: ICU 2.4
* 2010nov19 Markus Scherer  Rewrite for formatVersion 2.
**********************************************************************
*/
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2002-2011, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: October 30 2002
* Since: ICU 2.4
* 2010nov19 Markus Scherer  Rewrite for formatVersion 2.
**********************************************************************
*/







/*
 * This header defines the in-memory layout of the property names data
 * structure representing the UCD data files PropertyAliases.txt and
 * PropertyValueAliases.txt.  It is used by:
 *   propname.cpp - reads data
 *   genpname     - creates data
 */

/* low-level char * property name comparison -------------------------------- */

U_CDECL_BEGIN

/**
 * \var uprv_comparePropertyNames
 * Unicode property names and property value names are compared "loosely".
 *
 * UCD.html 4.0.1 says:
 *   For all property names, property value names, and for property values for
 *   Enumerated, Binary, or Catalog properties, use the following
 *   loose matching rule:
 *
 *   LM3. Ignore case, whitespace, underscore ('_'), and hyphens.
 *
 * This function does just that, for (char *) name strings.
 * It is almost identical to ucnv_compareNames() but also ignores
 * C0 White_Space characters (U+0009..U+000d, and U+0085 on EBCDIC).
 *
 * @internal
 */

U_CAPI int32_t U_EXPORT2
uprv_compareASCIIPropertyNames(const char *name1, const char *name2);

U_CAPI int32_t U_EXPORT2
uprv_compareEBCDICPropertyNames(const char *name1, const char *name2);

#if U_CHARSET_FAMILY==U_ASCII_FAMILY
#   define uprv_comparePropertyNames uprv_compareASCIIPropertyNames
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
#   define uprv_comparePropertyNames uprv_compareEBCDICPropertyNames
#else
#   error U_CHARSET_FAMILY is not valid
#endif

U_CDECL_END

/* UDataMemory structure and signatures ------------------------------------- */

#define PNAME_DATA_NAME "pnames"
#define PNAME_DATA_TYPE "icu"

/* Fields in UDataInfo: */

/* PNAME_SIG[] is encoded as numeric literals for compatibility with the HP compiler */
#define PNAME_SIG_0 ((uint8_t)0x70) /* p */
#define PNAME_SIG_1 ((uint8_t)0x6E) /* n */
#define PNAME_SIG_2 ((uint8_t)0x61) /* a */
#define PNAME_SIG_3 ((uint8_t)0x6D) /* m */

U_NAMESPACE_BEGIN

class PropNameData {
public:
    enum {
        // Byte offsets from the start of the data, after the generic header.
        IX_VALUE_MAPS_OFFSET,
        IX_BYTE_TRIES_OFFSET,
        IX_NAME_GROUPS_OFFSET,
        IX_RESERVED3_OFFSET,
        IX_RESERVED4_OFFSET,
        IX_TOTAL_SIZE,

        // Other values.
        IX_MAX_NAME_LENGTH,
        IX_RESERVED7,
        IX_COUNT
    };

    static const char *getPropertyName(int32_t property, int32_t nameChoice);
    static const char *getPropertyValueName(int32_t property, int32_t value, int32_t nameChoice);

    static int32_t getPropertyEnum(const char *alias);
    static int32_t getPropertyValueEnum(int32_t property, const char *alias);

private:
    static int32_t findProperty(int32_t property);
    static int32_t findPropertyValueNameGroup(int32_t valueMapIndex, int32_t value);
    static const char *getName(const char *nameGroup, int32_t nameIndex);
    static UBool containsName(BytesTrie &trie, const char *name);

    static int32_t getPropertyOrValueEnum(int32_t bytesTrieOffset, const char *alias);

    static const int32_t indexes[];
    static const int32_t valueMaps[];
    static const uint8_t bytesTries[];
    static const char nameGroups[];
};

/*
 * pnames.icu formatVersion 2
 *
 * formatVersion 2 is new in ICU 4.8.
 * In ICU 4.8, the pnames.icu data file is used only in ICU4J.
 * ICU4C 4.8 has the same data structures hardcoded in source/common/propname_data.h.
 *
 * For documentation of pnames.icu formatVersion 1 see ICU4C 4.6 (2010-dec-01)
 * or earlier versions of this header file (source/common/propname.h).
 *
 * The pnames.icu begins with the standard ICU DataHeader/UDataInfo.
 * After that:
 *
 * int32_t indexes[8];
 *
 *      (See the PropNameData::IX_... constants.)
 *
 *      The first 6 indexes are byte offsets from the beginning of the data
 *      (beginning of indexes[]) to following structures.
 *      The length of each structure is the difference between its offset
 *      and the next one.
 *      All offsets are filled in: Where there is no data between two offsets,
 *      those two offsets are the same.
 *      The last offset (indexes[PropNameData::IX_TOTAL_SIZE]) indicates the
 *      total number of bytes in the file. (Not counting the standard headers.)
 *
 *      The sixth index (indexes[PropNameData::IX_MAX_NAME_LENGTH]) has the
 *      maximum length of any Unicode property (or property value) alias.
 *      (Without normalization, that is, including underscores etc.)
 *
 * int32_t valueMaps[];
 *
 *      The valueMaps[] begins with a map from UProperty enums to properties,
 *      followed by the per-property value maps from property values to names,
 *      for those properties that have named values.
 *      (Binary & enumerated, plus General_Category_Mask.)
 *
 *      valueMaps[0] contains the number of UProperty enum ranges.
 *      For each range:
 *        int32_t start, limit -- first and last+1 UProperty enum of a dense range
 *        Followed by (limit-start) pairs of
 *          int32_t nameGroupOffset;
 *            Offset into nameGroups[] for the property's names/aliases.
 *          int32_t valueMapIndex;
 *            Offset of the property's value map in the valueMaps[] array.
 *            If the valueMapIndex is 0, then the property does not have named values.
 *
 *      For each property's value map:
 *      int32_t bytesTrieOffset; -- Offset into bytesTries[] for name->value mapping.
 *      int32_t numRanges;
 *        If numRanges is in the range 1..15, then that many ranges of values follow.
 *        Per range:
 *          int32_t start, limit -- first and last+1 UProperty enum of a range
 *          Followed by (limit-start) entries of
 *            int32_t nameGroupOffset;
 *              Offset into nameGroups[] for the property value's names/aliases.
 *              If the nameGroupOffset is 0, then this is not a named value for this property.
 *              (That is, the ranges need not be dense.)
 *        If numRanges is >=0x10, then (numRanges-0x10) sorted values
 *        and then (numRanges-0x10) corresponding nameGroupOffsets follow.
 *        Values are sorted as signed integers.
 *        In this case, the set of values is dense; no nameGroupOffset will be 0.
 *
 *      For both properties and property values, ranges are sorted by their start/limit values.
 *
 * uint8_t bytesTries[];
 *
 *      This is a sequence of BytesTrie structures, byte-serialized tries for
 *      mapping from names/aliases to values.
 *      The first one maps from property names/aliases to UProperty enum constants.
 *      The following ones are indexed by property value map bytesTrieOffsets
 *      for mapping each property's names/aliases to their property values.
 *
 * char nameGroups[];
 *
 *      This is a sequence of property name groups.
 *      Each group is a list of names/aliases (invariant-character strings) for
 *      one property or property value, in the order of UCharNameChoice.
 *      The first byte of each group is the number of names in the group.
 *      It is followed by that many NUL-terminated strings.
 *      The first string is for the short name; if there is no short name,
 *      then the first string is empty.
 *      The second string is the long name. Further strings are additional aliases.
 *
 *      The first name group is for a property rather than a property value,
 *      so that a nameGroupOffset of 0 can be used to indicate "no value"
 *      in a property's sparse value ranges.
 */

U_NAMESPACE_END











#define INCLUDED_FROM_PROPNAME_CPP
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// Copyright (C) 1999-2016, International Business Machines
// Corporation and others.  All Rights Reserved.
//
// file name: propname_data.h
//
// machine-generated by: icu/tools/unicode/c/genprops/pnamesbuilder.cpp


#ifdef INCLUDED_FROM_PROPNAME_CPP

U_NAMESPACE_BEGIN

const int32_t PropNameData::indexes[8]={0x20,0x15b8,0x5048,0xa69a,0xa69a,0xa69a,0x2f,0};

const int32_t PropNameData::valueMaps[1382]={
6,0,0x41,0,0xe3,0x368,0xe3,0x37e,0xe3,0x393,0xe3,0x3a9,0xe3,0x3b4,0xe3,0x3d5,
0xe3,0x3e5,0xe3,0x3f4,0xe3,0x402,0xe3,0x426,0xe3,0x43d,0xe3,0x455,0xe3,0x46c,0xe3,0x47b,
0xe3,0x48a,0xe3,0x49b,0xe3,0x4a9,0xe3,0x4bb,0xe3,0x4d5,0xe3,0x4f0,0xe3,0x505,0xe3,0x522,
0xe3,0x533,0xe3,0x53e,0xe3,0x55d,0xe3,0x573,0xe3,0x584,0xe3,0x594,0xe3,0x5af,0xe3,0x5c8,
0xe3,0x5d9,0xe3,0x5f3,0xe3,0x606,0xe3,0x616,0xe3,0x630,0xe3,0x649,0xe3,0x660,0xe3,0x674,
0xe3,0x68a,0xe3,0x69e,0xe3,0x6b4,0xe3,0x6ce,0xe3,0x6e6,0xe3,0x702,0xe3,0x70a,0xe3,0x712,
0xe3,0x71a,0xe3,0x722,0xe3,0x72b,0xe3,0x738,0xe3,0x74b,0xe3,0x768,0xe3,0x785,0xe3,0x7a2,
0xe3,0x7c0,0xe3,0x7de,0xe3,0x802,0xe3,0x80f,0xe3,0x829,0xe3,0x83e,0xe3,0x859,0xe3,0x870,
0xe3,0x887,0xe3,0x8a9,0xe3,0x1000,0x1019,0x8c8,0x15f,0xae8,0x17a,0x2f11,0xe9,0x2f30,0x2b3,0x306e,
0x2c9,0x30c8,0x2d3,0x3325,0x2f5,0x3c20,0x35f,0x3c90,0x369,0x3f2a,0x398,0x3f68,0x3a0,0x4a5b,0x465,0x4ad9,
0x46f,0x4afe,0x475,0x4b18,0x47b,0x4b39,0x482,0x4b53,0xe9,0x4b78,0xe9,0x4b9e,0x489,0x4c48,0x49f,0x4cc1,
0x4b2,0x4d73,0x4cd,0x4daa,0x4d4,0x4f8a,0x4e8,0x540a,0x510,0x2000,0x2001,0x5469,0x518,0x3000,0x3001,0x54f5,
0,0x4000,0x400e,0x5507,0,0x5510,0,0x552a,0,0x553b,0,0x554c,0,0x5562,0,0x556b,
0,0x5588,0,0x55a6,0,0x55c4,0,0x55e2,0,0x55f8,0,0x560c,0,0x5622,0,0x7000,
0x7001,0x563b,0,0x7d6,0x12,0,1,0x12,0x20,0x7f4,0x4a,0,1,6,7,8,
9,0xa,0xb,0xc,0xd,0xe,0xf,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,
0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0x24,0x54,0x5b,0x67,0x6b,
0x76,0x7a,0x81,0x82,0x84,0x85,0xc8,0xca,0xd6,0xd8,0xda,0xdc,0xde,0xe0,0xe2,0xe4,
0xe6,0xe8,0xe9,0xea,0xf0,0x2e,0x40,0x4c,0x5e,0x68,0x79,0x84,0x91,0x9e,0xab,0xb8,
0xc5,0xd2,0xdf,0xec,0xf9,0x106,0x113,0x120,0x12d,0x13a,0x147,0x154,0x161,0x16e,0x17b,0x188,
0x195,0x1a2,0x1af,0x1bc,0x1c9,0x1d6,0x1e3,0x1f0,0x1fd,0x20c,0x21b,0x22a,0x239,0x248,0x257,0x266,
0x275,0x28f,0x2a3,0x2b7,0x2d2,0x2e1,0x2ea,0x2fa,0x302,0x30b,0x31a,0x323,0x333,0x344,0x355,0x995,
1,0,0x17,0x8d7,0x8e8,0x8f9,0x90d,0x924,0x93c,0x94e,0x963,0x97a,0x98f,0x99f,0x9b1,0x9ce,
0x9ea,0x9fc,0xa19,0xa35,0xa51,0xa66,0xa7b,0xa95,0xab0,0xacb,0xb37,1,0,0x135,0xaf3,0xb00,
0xb13,0xb3b,0xb59,0xb77,0xb8f,0xbba,0xbe4,0xbfc,0xc0f,0xc22,0xc31,0xc40,0xc4f,0xc5e,0xc75,0xc86,
0xc99,0xcac,0xcb9,0xcc6,0xcd5,0xce6,0xcfb,0xd0c,0xd17,0xd20,0xd31,0xd42,0xd55,0xd67,0xd7a,0xd8d,
0xdcc,0xdd9,0xde6,0xdf3,0xe08,0xe38,0xe52,0xe73,0xe9e,0xec1,0xf1f,0xf46,0xf61,0xf70,0xf97,0xfbf,
0xfe2,0x1005,0x102f,0x1048,0x1067,0x108a,0x10ae,0x10c1,0x10db,0x1105,0x111d,0x1145,0x116e,0x1181,0x1194,0x11a7,
0x11ce,0x11dd,0x11fd,0x122b,0x1249,0x1277,0x1293,0x12ae,0x12c7,0x12e0,0x1301,0x1331,0x1350,0x1372,0x13a6,0x13d3,
0x1418,0x1439,0x1463,0x1484,0x14ad,0x14c0,0x14f3,0x150a,0x1519,0x152a,0x1555,0x156c,0x159d,0x15cb,0x160e,0x1619,
0x1652,0x1663,0x1674,0x1681,0x1694,0x16ce,0x16f2,0x1716,0x1750,0x1788,0x17b3,0x17cb,0x17f7,0x1823,0x1830,0x183f,
0x185c,0x187e,0x18ac,0x18cc,0x18f3,0x191a,0x1939,0x194c,0x195d,0x196e,0x1993,0x19b8,0x19df,0x1a13,0x1a40,0x1a5e,
0x1a71,0x1a8a,0x1ac3,0x1ad2,0x1af2,0x1b14,0x1b36,0x1b4d,0x1b64,0x1b91,0x1baa,0x1bc3,0x1bf4,0x1c1e,0x1c39,0x1c4c,
0x1c6b,0x1c74,0x1c87,0x1ca5,0x1cc3,0x1cd6,0x1ced,0x1d02,0x1d37,0x1d5b,0x1d70,0x1d7f,0x1d92,0x1db6,0x1dbf,0x1de3,
0x1dfa,0x1e0d,0x1e1c,0x1e27,0x1e48,0x1e60,0x1e6f,0x1e7e,0x1e8d,0x1ea4,0x1eb9,0x1ece,0x1f07,0x1f1a,0x1f36,0x1f41,
0x1f4e,0x1f7c,0x1fa0,0x1fc3,0x1fd6,0x1ff8,0x200b,0x2026,0x2049,0x206c,0x2091,0x20a2,0x20d1,0x20fe,0x2115,0x2130,
0x213f,0x216a,0x21a2,0x21dc,0x220a,0x221b,0x2228,0x224c,0x225b,0x2277,0x2291,0x22ae,0x22e6,0x22fb,0x2328,0x2347,
0x2375,0x2395,0x23c9,0x23d8,0x2402,0x2425,0x2450,0x245b,0x246c,0x2487,0x24ab,0x24b8,0x24cd,0x24f4,0x251f,0x2556,
0x2569,0x257a,0x25aa,0x25bb,0x25ca,0x25df,0x25fd,0x2610,0x2623,0x263a,0x2657,0x2662,0x266b,0x268d,0x26a2,0x26c7,
0x26de,0x2707,0x2722,0x2737,0x2750,0x2771,0x27a6,0x27b7,0x27e8,0x280c,0x281d,0x2836,0x2841,0x286e,0x2890,0x28be,
0x28f1,0x2900,0x2911,0x292e,0x2970,0x2997,0x29a4,0x29b9,0x29dd,0x2a03,0x2a3c,0x2a4d,0x2a71,0x2a7c,0x2a89,0x2a98,
0x2abd,0x2aeb,0x2b07,0x2b24,0x2b31,0x2b42,0x2b60,0x2b83,0x2ba0,0x2bad,0x2bcd,0x2bea,0x2c0b,0x2c34,0x2c45,0x2c64,
0x2c7d,0x2c96,0x2ca7,0x2cf0,0x2d01,0x2d1a,0x2d49,0x2d76,0x2d9b,0x2ddd,0x2df9,0x2e08,0x2e1f,0x2e4d,0x2e66,0x2e8f,
0x2ea9,0x2ee4,0x2f02,0x1e85,1,0,0x12,0x2f47,0x2f57,0x2f6a,0x2f7a,0x2f8a,0x2f99,0x2fa9,0x2fbb,0x2fce,
0x2fe0,0x2ff0,0x3000,0x300f,0x301e,0x302e,0x303b,0x304a,0x305e,0x1f43,1,0,6,0x3083,0x308e,0x309b,
0x30a8,0x30b5,0x30c0,0x1f87,1,0,0x1e,0x30dd,0x30ec,0x3101,0x3116,0x312b,0x313f,0x3150,0x3164,0x3177,
0x3188,0x31a1,0x31b3,0x31c4,0x31d8,0x31eb,0x3203,0x3215,0x3220,0x3230,0x323e,0x3253,0x3268,0x327e,0x3298,0x32ae,
0x32be,0x32d2,0x32e6,0x32f7,0x330f,0x21b2,1,0,0x66,0x3337,0x335a,0x3363,0x3370,0x337b,0x3384,0x338f,
0x3398,0x33b1,0x33b6,0x33bf,0x33dc,0x33e5,0x33f2,0x33fb,0x341f,0x3426,0x342f,0x3442,0x344d,0x3456,0x3461,0x347a,
0x3483,0x3492,0x349d,0x34a6,0x34b1,0x34ba,0x34c1,0x34ca,0x34d5,0x34de,0x34f7,0x3500,0x350d,0x3518,0x3529,0x3534,
0x3549,0x3560,0x3569,0x3572,0x358b,0x3596,0x359f,0x35a8,0x35bf,0x35dc,0x35e7,0x35f8,0x3603,0x360a,0x3617,0x3624,
0x3651,0x3666,0x366f,0x368a,0x36ad,0x36ce,0x36ef,0x3714,0x373b,0x375c,0x377f,0x37a0,0x37c7,0x37e8,0x380d,0x382c,
0x384b,0x386a,0x3887,0x38a8,0x38c9,0x38ec,0x3911,0x3930,0x394f,0x3970,0x3997,0x39bc,0x39db,0x39fc,0x3a1f,0x3a3a,
0x3a53,0x3a6e,0x3a87,0x3aa4,0x3abf,0x3adc,0x3afb,0x3b18,0x3b35,0x3b54,0x3b71,0x3b8c,0x3ba9,0x3bc6,0x3bf9,0x24f7,
1,0,6,0x3c31,0x3c40,0x3c50,0x3c60,0x3c70,0x3c81,0x2555,1,0,0x2b,0x3c9f,0x3cab,0x3cb9,
0x3cc8,0x3cd7,0x3ce7,0x3cf8,0x3d0c,0x3d21,0x3d37,0x3d4a,0x3d5e,0x3d6e,0x3d77,0x3d82,0x3d92,0x3dae,0x3dc0,0x3dce,
0x3ddd,0x3de9,0x3dfe,0x3e12,0x3e25,0x3e33,0x3e47,0x3e55,0x3e5f,0x3e71,0x3e7d,0x3e8b,0x3e9b,0x3ea2,0x3ea9,0x3eb0,
0x3eb7,0x3ebe,0x3ed4,0x3ef5,0x870,0x3f07,0x3f12,0x3f21,0x27ae,1,0,4,0x3f3b,0x3f46,0x3f52,0x3f5c,
0x27d4,1,0,0xc1,0x3f73,0x3f80,0x3f95,0x3fa2,0x3fb1,0x3fbf,0x3fce,0x3fdd,0x3fef,0x3ffe,0x400c,0x401d,
0x402c,0x403b,0x4048,0x4054,0x4063,0x4072,0x407c,0x4089,0x4096,0x40a5,0x40b3,0x40c2,0x40ce,0x40d8,0x40e4,0x40f4,
0x4104,0x4112,0x411e,0x412f,0x413b,0x4147,0x4155,0x4162,0x416e,0x417b,0xd0c,0x4188,0x4196,0x41b0,0x41b9,0x41c7,
0x41d5,0x41e1,0x41f0,0x41fe,0x420c,0x4218,0x4227,0x4235,0x4243,0x4250,0x425f,0x427a,0x4289,0x429a,0x42ab,0x42be,
0x42d0,0x42df,0x42f1,0x4300,0x430c,0x4317,0x1e1c,0x4324,0x432f,0x433a,0x4345,0x4350,0x436b,0x4376,0x4381,0x438c,
0x439f,0x43b3,0x43be,0x43cd,0x43dc,0x43e7,0x43f2,0x43ff,0x440e,0x441c,0x4427,0x4442,0x444c,0x445d,0x446e,0x447d,
0x448e,0x4499,0x44a4,0x44af,0x44ba,0x44c5,0x44d0,0x44db,0x44e5,0x44f0,0x4500,0x450b,0x4519,0x4526,0x4531,0x4540,
0x454d,0x455a,0x4569,0x4576,0x4587,0x4599,0x45a9,0x45b4,0x45c7,0x45de,0x45ec,0x45f9,0x4604,0x4611,0x4622,0x463e,
0x4654,0x465f,0x467c,0x468c,0x469b,0x46a6,0x46b1,0x1f36,0x46bd,0x46c8,0x46e0,0x46f0,0x46ff,0x470d,0x471b,0x4726,
0x4731,0x4745,0x475c,0x4774,0x4784,0x4794,0x47a4,0x47b6,0x47c1,0x47cc,0x47d6,0x47e2,0x47f0,0x4803,0x480f,0x481c,
0x4827,0x4843,0x4850,0x485e,0x4877,0x2836,0x4886,0x2657,0x4893,0x48a1,0x48b3,0x48c1,0x48cd,0x48dd,0x2a71,0x48eb,
0x48f7,0x4902,0x490d,0x4918,0x492c,0x493a,0x4951,0x495d,0x4971,0x497f,0x4991,0x49a7,0x49b5,0x49c7,0x49d5,0x49f2,
0x4a04,0x4a11,0x4a22,0x4a34,0x4a4e,0x31cc,1,0,6,0x4a75,0x4a88,0x4a98,0x4aa6,0x4ab7,0x4ac7,0x3228,
0x12,0,1,0x4af1,0x4af7,0x3235,0x12,0,1,0x4af1,0x4af7,0x3242,1,0,3,0x4af1,
0x4af7,0x4b30,0x3258,1,0,3,0x4af1,0x4af7,0x4b30,0x326e,1,0,0x12,0x4bba,0x4bc4,0x4bd0,
0x4bd7,0x4be2,0x4be7,0x4bee,0x4bf5,0x4bfe,0x4c03,0x4c08,0x4c18,0x870,0x3f07,0x4c24,0x3f12,0x4c34,0x3f21,0x3317,
1,0,0xf,0x4bba,0x4c5b,0x4c65,0x4c6f,0x4c7a,0x3ddd,0x4c84,0x4c90,0x4c98,0x4c9f,0x4ca9,0x4bd0,0x4bd7,
0x4be7,0x4cb3,0x339e,1,0,0x17,0x4bba,0x4cd0,0x4c6f,0x4cdc,0x4ce9,0x4cf7,0x3ddd,0x4d02,0x4bd0,0x4d13,
0x4be7,0x4d22,0x4d30,0x870,0x3ef5,0x4d3c,0x4d4d,0x3f07,0x4c24,0x3f12,0x4c34,0x3f21,0x4d5e,0x34bb,1,0,
3,0x4d91,0x4d99,0x4da1,0x34d4,1,0,0x10,0x4dca,0x4dd1,0x4de0,0x4e01,0x4e24,0x4e2f,0x4e4e,0x4e65,
0x4e72,0x4e7b,0x4e9a,0x4ecd,0x4ee8,0x4f17,0x4f34,0x4f59,0x356d,1,0,0x24,0x4fa8,0x4fb5,0x4fc8,0x4fd5,
0x5002,0x5027,0x503c,0x505b,0x507c,0x50a9,0x50e2,0x5105,0x5128,0x5155,0x518a,0x51b1,0x51da,0x5211,0x5240,0x5261,
0x5286,0x5295,0x52b8,0x52cf,0x52dc,0x52eb,0x5308,0x5321,0x5344,0x5369,0x5382,0x5397,0x53a6,0x53b7,0x53c4,0x53e5,
0x373d,1,0,4,0x5423,0x542e,0x5446,0x545e,0x3779,0x36,1,2,4,8,0xe,0x10,
0x20,0x3e,0x40,0x80,0x100,0x1c0,0x200,0x400,0x800,0xe00,0x1000,0x2000,0x4000,0x7000,0x8000,0x10000,
0x20000,0x40000,0x78001,0x80000,0x100000,0x200000,0x400000,0x800000,0x1000000,0x2000000,0x4000000,0x8000000,0xf000000,0x10000000,0x20000000,0x30f80000,
0x30dd,0x30ec,0x3101,0x3116,0x5497,0x312b,0x313f,0x548d,0x3150,0x3164,0x3177,0x54a8,0x3188,0x31a1,0x31b3,0x54bf,
0x31c4,0x31d8,0x31eb,0x54e8,0x3203,0x3215,0x3220,0x3230,0x5484,0x323e,0x3253,0x3268,0x327e,0x3298,0x32ae,0x32be,
0x32d2,0x32e6,0x54de,0x32f7,0x330f,0x54c9
};

const uint8_t PropNameData::bytesTries[14992]={
0,0x15,0x6d,0xc3,0x78,0x73,0xc2,0x12,0x76,0x7a,0x76,0x6a,0x77,0xa2,0x52,0x78,
1,0x64,0x50,0x69,0x10,0x64,1,0x63,0x30,0x73,0x62,0x13,0x74,0x61,0x72,0x74,
0x63,0x60,0x16,0x6f,0x6e,0x74,0x69,0x6e,0x75,0x65,0x61,0x13,0x69,0x67,0x69,0x74,
0x81,3,0x61,0x2e,0x65,0x4c,0x6f,0xc3,0x18,0x73,0x69,0x1e,0x72,0x69,0x61,0x74,
0x69,0x6f,0x6e,0x73,0x65,0x6c,0x65,0x63,0x74,0x6f,0x72,0x69,0x10,0x72,0x1f,0x74,
0x69,0x63,0x61,0x6c,0x6f,0x72,0x69,0x65,0x6e,0x74,0x61,0x74,0x69,0x6f,0x6e,0xc3,
0x18,3,0x62,0xc3,0x14,0x68,0x32,0x6f,0x42,0x73,0x13,0x70,0x61,0x63,0x65,0x5f,
0x17,0x69,0x74,0x65,0x73,0x70,0x61,0x63,0x65,0x5f,0x16,0x72,0x64,0x62,0x72,0x65,
0x61,0x6b,0xc3,0x14,0x73,0xa2,0x49,0x74,0xa4,0x3b,0x75,3,0x63,0xd9,0x40,0xc,
0x69,0x52,0x6e,0x58,0x70,0x12,0x70,0x65,0x72,0x5c,0x13,0x63,0x61,0x73,0x65,0x5c,
0x16,0x6d,0x61,0x70,0x70,0x69,0x6e,0x67,0xd9,0x40,0xc,0x12,0x64,0x65,0x6f,0x5b,
0x10,0x69,1,0x63,0x3e,0x66,0x1b,0x69,0x65,0x64,0x69,0x64,0x65,0x6f,0x67,0x72,
0x61,0x70,0x68,0x5b,0x17,0x6f,0x64,0x65,0x31,0x6e,0x61,0x6d,0x65,0xd9,0x40,0xb,
0xa,0x69,0x84,0x70,0x19,0x70,0x30,0x74,0x36,0x75,0x10,0x63,0xd9,0x40,9,0x12,
0x61,0x63,0x65,0x5f,1,0x63,0xd9,0x40,8,0x65,0x11,0x72,0x6d,0x67,0x69,0x3c,
0x6c,0xa2,0x5f,0x6f,0x17,0x66,0x74,0x64,0x6f,0x74,0x74,0x65,0x64,0x57,0x13,0x6d,
0x70,0x6c,0x65,3,0x63,0x50,0x6c,0x68,0x74,0x8a,0x75,0x1e,0x70,0x70,0x65,0x72,
0x63,0x61,0x73,0x65,0x6d,0x61,0x70,0x70,0x69,0x6e,0x67,0xd9,0x40,9,0x19,0x61,
0x73,0x65,0x66,0x6f,0x6c,0x64,0x69,0x6e,0x67,0xd9,0x40,6,0x1e,0x6f,0x77,0x65,
0x72,0x63,0x61,0x73,0x65,0x6d,0x61,0x70,0x70,0x69,0x6e,0x67,0xd9,0x40,7,0x1e,
0x69,0x74,0x6c,0x65,0x63,0x61,0x73,0x65,0x6d,0x61,0x70,0x70,0x69,0x6e,0x67,0xd9,
0x40,8,0x10,0x63,0xd9,0x40,7,0x62,0xc3,0x13,0x63,0x34,0x64,0x57,0x65,0x6e,
0x66,0x10,0x63,0xd9,0x40,6,0xc2,0xa,2,0x66,0xd9,0x40,6,0x72,0x28,0x78,
0xd9,0x70,0,0x12,0x69,0x70,0x74,0xc2,0xa,0x19,0x65,0x78,0x74,0x65,0x6e,0x73,
0x69,0x6f,0x6e,0x73,0xd9,0x70,0,1,0x67,0x6a,0x6e,1,0x73,0x54,0x74,0x13,
0x65,0x6e,0x63,0x65,1,0x62,0x34,0x74,0x16,0x65,0x72,0x6d,0x69,0x6e,0x61,0x6c,
0x67,0x13,0x72,0x65,0x61,0x6b,0xc3,0x13,0x14,0x69,0x74,0x69,0x76,0x65,0x65,1,
0x6d,0x2e,0x73,0x13,0x74,0x61,0x72,0x74,0x73,0x19,0x65,0x6e,0x74,0x73,0x74,0x61,
0x72,0x74,0x65,0x72,0x73,3,0x63,0x66,0x65,0x72,0x69,0x98,0x72,0x19,0x61,0x69,
0x6c,0x63,0x61,0x6e,0x6f,0x6e,0x69,0x63,0x1f,0x61,0x6c,0x63,0x6f,0x6d,0x62,0x69,
0x6e,0x69,0x6e,0x67,0x63,0x6c,0x61,0x73,0x73,0xc3,0x11,0xd8,0x40,0xa,0x11,0x63,
0x63,0xc3,0x11,0x11,0x72,0x6d,0x58,0x1e,0x69,0x6e,0x61,0x6c,0x70,0x75,0x6e,0x63,
0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0x59,0x1d,0x74,0x6c,0x65,0x63,0x61,0x73,0x65,
0x6d,0x61,0x70,0x70,0x69,0x6e,0x67,0xd9,0x40,0xa,0x6d,0x70,0x6e,0x76,0x70,0xa2,
0xf1,0x71,0xa4,0x43,0x72,2,0x61,0x28,0x65,0x32,0x69,0x9d,0x14,0x64,0x69,0x63,
0x61,0x6c,0x55,0x1e,0x67,0x69,0x6f,0x6e,0x61,0x6c,0x69,0x6e,0x64,0x69,0x63,0x61,
0x74,0x6f,0x72,0x9d,0x12,0x61,0x74,0x68,0x4f,6,0x6f,0x39,0x6f,0x32,0x74,0xc3,
9,0x75,0x54,0x76,0xd9,0x30,0,0x12,0x6e,0x63,0x68,0x1f,0x61,0x72,0x61,0x63,
0x74,0x65,0x72,0x63,0x6f,0x64,0x65,0x70,0x6f,0x69,0x6e,0x74,0x51,0x14,0x6d,0x65,
0x72,0x69,0x63,1,0x74,0x32,0x76,0x13,0x61,0x6c,0x75,0x65,0xd9,0x30,0,0x12,
0x79,0x70,0x65,0xc3,9,0x61,0xa2,0x77,0x63,0xa2,0x82,0x66,2,0x63,0x98,0x64,
0xa2,0x53,0x6b,1,0x63,0x56,0x64,1,0x69,0x42,0x71,1,0x63,0xc3,0xd,0x75,
0x17,0x69,0x63,0x6b,0x63,0x68,0x65,0x63,0x6b,0xc3,0xd,0x13,0x6e,0x65,0x72,0x74,
0x6d,1,0x69,0x42,0x71,1,0x63,0xc3,0xf,0x75,0x17,0x69,0x63,0x6b,0x63,0x68,
0x65,0x63,0x6b,0xc3,0xf,0x13,0x6e,0x65,0x72,0x74,0x71,1,0x69,0x42,0x71,1,
0x63,0xc3,0xe,0x75,0x17,0x69,0x63,0x6b,0x63,0x68,0x65,0x63,0x6b,0xc3,0xe,0x13,
0x6e,0x65,0x72,0x74,0x6f,1,0x69,0x42,0x71,1,0x63,0xc3,0xc,0x75,0x17,0x69,
0x63,0x6b,0x63,0x68,0x65,0x63,0x6b,0xc3,0xc,0x13,0x6e,0x65,0x72,0x74,0x6b,0xd8,
0x40,5,1,0x31,0xd9,0x40,0xb,0x6d,0x10,0x65,0xd9,0x40,5,0x12,0x68,0x61,
0x72,0x51,2,0x61,0x6c,0x63,0xa2,0x4c,0x72,1,0x65,0x2a,0x69,0x11,0x6e,0x74,
0x7f,0x16,0x70,0x65,0x6e,0x64,0x65,0x64,0x63,0x1f,0x6f,0x6e,0x63,0x61,0x74,0x65,
0x6e,0x61,0x74,0x69,0x6f,0x6e,0x6d,0x61,0x72,0x6b,0x9f,0x10,0x74,2,0x73,0x2c,
0x74,0x30,0x77,0x10,0x73,0x77,0x11,0x79,0x6e,0x75,0x12,0x65,0x72,0x6e,1,0x73,
0x38,0x77,0x18,0x68,0x69,0x74,0x65,0x73,0x70,0x61,0x63,0x65,0x77,0x14,0x79,0x6e,
0x74,0x61,0x78,0x75,0x10,0x6d,0x9f,1,0x6d,0x3c,0x75,0x1a,0x6f,0x74,0x61,0x74,
0x69,0x6f,0x6e,0x6d,0x61,0x72,0x6b,0x53,0x12,0x61,0x72,0x6b,0x53,0x66,0xc1,0xf8,
0x69,0xc1,0x3c,0x69,0xa2,0x6f,0x6a,0xa4,9,0x6c,4,0x62,0xc3,8,0x63,0x8c,
0x65,0x98,0x69,0xa2,0x56,0x6f,2,0x65,0x4b,0x67,0x4c,0x77,0x11,0x65,0x72,0x4c,
0x13,0x63,0x61,0x73,0x65,0x4c,0x16,0x6d,0x61,0x70,0x70,0x69,0x6e,0x67,0xd9,0x40,
4,0x11,0x69,0x63,0x1f,0x61,0x6c,0x6f,0x72,0x64,0x65,0x72,0x65,0x78,0x63,0x65,
0x70,0x74,0x69,0x6f,0x6e,0x4b,0xd8,0x40,4,0x11,0x63,0x63,0xc3,0x10,0x18,0x61,
0x64,0x63,0x61,0x6e,0x6f,0x6e,0x69,0x63,0x1f,0x61,0x6c,0x63,0x6f,0x6d,0x62,0x69,
0x6e,0x69,0x6e,0x67,0x63,0x6c,0x61,0x73,0x73,0xc3,0x10,0x16,0x6e,0x65,0x62,0x72,
0x65,0x61,0x6b,0xc3,8,2,0x64,0x4a,0x6e,0xa2,0x5b,0x73,1,0x63,0xd9,0x40,
3,0x6f,0x16,0x63,0x6f,0x6d,0x6d,0x65,0x6e,0x74,0xd9,0x40,3,2,0x63,0x80,
0x65,0x90,0x73,0x40,1,0x62,0x52,0x74,0x46,1,0x61,0x40,0x72,0x1c,0x69,0x6e,
0x61,0x72,0x79,0x6f,0x70,0x65,0x72,0x61,0x74,0x6f,0x72,0x47,0x11,0x72,0x74,0x41,
0x44,0x1c,0x69,0x6e,0x61,0x72,0x79,0x6f,0x70,0x65,0x72,0x61,0x74,0x6f,0x72,0x45,
0x3e,0x16,0x6f,0x6e,0x74,0x69,0x6e,0x75,0x65,0x3f,0x10,0x6f,0x42,0x16,0x67,0x72,
0x61,0x70,0x68,0x69,0x63,0x43,2,0x64,0x2e,0x70,0x86,0x73,0x10,0x63,0xc3,0x17,
0x11,0x69,0x63,1,0x70,0x46,0x73,0x1e,0x79,0x6c,0x6c,0x61,0x62,0x69,0x63,0x63,
0x61,0x74,0x65,0x67,0x6f,0x72,0x79,0xc3,0x17,0x10,0x6f,0x1f,0x73,0x69,0x74,0x69,
0x6f,0x6e,0x61,0x6c,0x63,0x61,0x74,0x65,0x67,0x6f,0x72,0x79,0xc3,0x16,0x10,0x63,
0xc3,0x16,2,0x67,0xc3,6,0x6f,0x26,0x74,0xc3,7,0x11,0x69,0x6e,1,0x63,
0x4a,0x69,0x11,0x6e,0x67,1,0x67,0x2e,0x74,0x12,0x79,0x70,0x65,0xc3,7,0x13,
0x72,0x6f,0x75,0x70,0xc3,6,0x48,0x15,0x6f,0x6e,0x74,0x72,0x6f,0x6c,0x49,0x66,
0x86,0x67,0xa2,0x4a,0x68,3,0x61,0x36,0x65,0x58,0x73,0x68,0x79,0x13,0x70,0x68,
0x65,0x6e,0x3d,0x1f,0x6e,0x67,0x75,0x6c,0x73,0x79,0x6c,0x6c,0x61,0x62,0x6c,0x65,
0x74,0x79,0x70,0x65,0xc3,0xb,0x10,0x78,0x3a,0x14,0x64,0x69,0x67,0x69,0x74,0x3b,
0x10,0x74,0xc3,0xb,0x16,0x75,0x6c,0x6c,0x63,0x6f,0x6d,0x70,0x1f,0x6f,0x73,0x69,
0x74,0x69,0x6f,0x6e,0x65,0x78,0x63,0x6c,0x75,0x73,0x69,0x6f,0x6e,0x33,2,0x63,
0xa2,0x44,0x65,0xa2,0x4b,0x72,3,0x61,0x34,0x62,0x84,0x65,0x8a,0x6c,0x12,0x69,
0x6e,0x6b,0x39,0x11,0x70,0x68,0x7c,0x12,0x65,0x6d,0x65,3,0x62,0x5e,0x63,0x30,
0x65,0x48,0x6c,0x12,0x69,0x6e,0x6b,0x39,0x1a,0x6c,0x75,0x73,0x74,0x65,0x72,0x62,
0x72,0x65,0x61,0x6b,0xc3,0x12,0x14,0x78,0x74,0x65,0x6e,0x64,0x37,0x12,0x61,0x73,
0x65,0x35,0x11,0x78,0x74,0x37,0xc2,5,1,0x62,0xc3,0x12,0x6d,0xd9,0x20,0,
0x1c,0x6e,0x65,0x72,0x61,0x6c,0x63,0x61,0x74,0x65,0x67,0x6f,0x72,0x79,0xc2,5,
0x13,0x6d,0x61,0x73,0x6b,0xd9,0x20,0,0x61,0xa2,0x90,0x62,0xa2,0xbe,0x63,0xa4,
0x30,0x64,0xa4,0xfd,0x65,5,0x6d,0x63,0x6d,0x6e,0x70,0xa2,0x59,0x78,0x10,0x74,
0x30,1,0x65,0x2c,0x70,0x12,0x69,0x63,0x74,0xa1,0x12,0x6e,0x64,0x65,1,0x64,
0x24,0x72,0x31,0x1b,0x70,0x69,0x63,0x74,0x6f,0x67,0x72,0x61,0x70,0x68,0x69,0x63,
0xa1,0x10,0x6f,1,0x64,0x97,0x6a,0x10,0x69,0x92,2,0x63,0x40,0x6d,0x50,0x70,
0x1a,0x72,0x65,0x73,0x65,0x6e,0x74,0x61,0x74,0x69,0x6f,0x6e,0x95,0x17,0x6f,0x6d,
0x70,0x6f,0x6e,0x65,0x6e,0x74,0x9b,0x16,0x6f,0x64,0x69,0x66,0x69,0x65,0x72,0x96,
0x13,0x62,0x61,0x73,0x65,0x99,0x12,0x72,0x65,0x73,0x95,0x61,0x30,0x62,0x4e,0x63,
0x12,0x6f,0x6d,0x70,0x9b,0xc2,4,0x1b,0x73,0x74,0x61,0x73,0x69,0x61,0x6e,0x77,
0x69,0x64,0x74,0x68,0xc3,4,0x12,0x61,0x73,0x65,0x99,3,0x67,0x44,0x68,0x4a,
0x6c,0x4e,0x73,0x1a,0x63,0x69,0x69,0x68,0x65,0x78,0x64,0x69,0x67,0x69,0x74,0x23,
0x10,0x65,0xd9,0x40,0,0x11,0x65,0x78,0x23,1,0x6e,0x38,0x70,0x11,0x68,0x61,
0x20,0x14,0x62,0x65,0x74,0x69,0x63,0x21,0x11,0x75,0x6d,0x79,4,0x63,0xc3,0,
0x69,0x3e,0x6c,0xa2,0x57,0x6d,0xa2,0x64,0x70,1,0x62,0xd9,0x40,0xd,0x74,0xc3,
0x15,0x11,0x64,0x69,2,0x63,0x54,0x6d,0x74,0x70,0x1b,0x61,0x69,0x72,0x65,0x64,
0x62,0x72,0x61,0x63,0x6b,0x65,0x74,0xd8,0x40,0xd,0x13,0x74,0x79,0x70,0x65,0xc3,
0x15,0x24,1,0x6c,0x30,0x6f,0x14,0x6e,0x74,0x72,0x6f,0x6c,0x25,0x12,0x61,0x73,
0x73,0xc3,0,0x26,0x14,0x69,0x72,0x72,0x6f,0x72,1,0x65,0x38,0x69,0x16,0x6e,
0x67,0x67,0x6c,0x79,0x70,0x68,0xd9,0x40,1,0x10,0x64,0x27,2,0x61,0x32,0x6b,
0xc3,1,0x6f,0x11,0x63,0x6b,0xc3,1,0x11,0x6e,0x6b,0x7b,0x10,0x67,0xd9,0x40,
1,6,0x68,0x7c,0x68,0x54,0x69,0x85,0x6f,0xa2,0x6f,0x77,4,0x63,0x30,0x6b,
0x36,0x6c,0x87,0x74,0x8b,0x75,0x89,1,0x66,0x8d,0x6d,0x8f,0x11,0x63,0x66,0x91,
0x18,0x61,0x6e,0x67,0x65,0x73,0x77,0x68,0x65,0x6e,4,0x63,0x44,0x6c,0x6c,0x6e,
0x7e,0x74,0x98,0x75,0x18,0x70,0x70,0x65,0x72,0x63,0x61,0x73,0x65,0x64,0x89,0x12,
0x61,0x73,0x65,1,0x66,0x30,0x6d,0x14,0x61,0x70,0x70,0x65,0x64,0x8f,0x14,0x6f,
0x6c,0x64,0x65,0x64,0x8d,0x18,0x6f,0x77,0x65,0x72,0x63,0x61,0x73,0x65,0x64,0x87,
0x1c,0x66,0x6b,0x63,0x63,0x61,0x73,0x65,0x66,0x6f,0x6c,0x64,0x65,0x64,0x91,0x18,
0x69,0x74,0x6c,0x65,0x63,0x61,0x73,0x65,0x64,0x8b,0x13,0x6d,0x70,0x65,0x78,0x33,
0x61,0x2e,0x63,0xa2,0x48,0x66,0xd9,0x40,2,1,0x6e,0x72,0x73,0x10,0x65,3,
0x64,0x83,0x66,0x3a,0x69,0x4a,0x73,0x17,0x65,0x6e,0x73,0x69,0x74,0x69,0x76,0x65,
0x65,0x15,0x6f,0x6c,0x64,0x69,0x6e,0x67,0xd9,0x40,2,0x17,0x67,0x6e,0x6f,0x72,
0x61,0x62,0x6c,0x65,0x85,0x13,0x6f,0x6e,0x69,0x63,0x1f,0x61,0x6c,0x63,0x6f,0x6d,
0x62,0x69,0x6e,0x69,0x6e,0x67,0x63,0x6c,0x61,0x73,0x73,0xc3,2,0x10,0x63,0xc3,
2,3,0x61,0x30,0x65,0x34,0x69,0xa2,0x41,0x74,0xc3,3,0x11,0x73,0x68,0x29,
2,0x63,0x3a,0x66,0x58,0x70,0x2c,0x16,0x72,0x65,0x63,0x61,0x74,0x65,0x64,0x2d,
0x1d,0x6f,0x6d,0x70,0x6f,0x73,0x69,0x74,0x69,0x6f,0x6e,0x74,0x79,0x70,0x65,0xc3,
3,0x15,0x61,0x75,0x6c,0x74,0x69,0x67,0x1f,0x6e,0x6f,0x72,0x61,0x62,0x6c,0x65,
0x63,0x6f,0x64,0x65,0x70,0x6f,0x69,0x6e,0x74,0x2b,0x2a,0x10,0x61,0x2e,0x15,0x63,
0x72,0x69,0x74,0x69,0x63,0x2f,3,0x66,0x34,0x6e,0x3e,0x74,0x42,0x79,0x22,0x11,
0x65,0x73,0x23,0x20,0x13,0x61,0x6c,0x73,0x65,0x21,0x20,0x10,0x6f,0x21,0x22,0x12,
0x72,0x75,0x65,0x23,0xb,0x6b,0x5b,0x6f,0x23,0x6f,0x3c,0x72,0x4c,0x76,1,0x69,
0x24,0x72,0x33,0x13,0x72,0x61,0x6d,0x61,0x33,0x10,0x76,0x22,0x14,0x65,0x72,0x6c,
0x61,0x79,0x23,0xa2,0xe2,0x13,0x69,0x67,0x68,0x74,0xa3,0xe2,0x6b,0x58,0x6c,0x74,
0x6e,3,0x6b,0x2f,0x6f,0x30,0x72,0x21,0x75,0x12,0x6b,0x74,0x61,0x2f,0x19,0x74,
0x72,0x65,0x6f,0x72,0x64,0x65,0x72,0x65,0x64,0x21,1,0x61,0x24,0x76,0x31,0x18,
0x6e,0x61,0x76,0x6f,0x69,0x63,0x69,0x6e,0x67,0x31,0xa2,0xe0,0x12,0x65,0x66,0x74,
0xa3,0xe0,0x64,0x45,0x64,0x4e,0x68,0x88,0x69,1,0x6f,0x26,0x73,0xa3,0xf0,0x1a,
0x74,0x61,0x73,0x75,0x62,0x73,0x63,0x72,0x69,0x70,0x74,0xa3,0xf0,2,0x61,0xa3,
0xea,0x62,0xa3,0xe9,0x6f,0x13,0x75,0x62,0x6c,0x65,1,0x61,0x30,0x62,0x13,0x65,
0x6c,0x6f,0x77,0xa3,0xe9,0x13,0x62,0x6f,0x76,0x65,0xa3,0xea,0x12,0x61,0x6e,0x72,
0x2c,0x15,0x65,0x61,0x64,0x69,0x6e,0x67,0x2d,0x61,0xa2,0x7b,0x62,0xa2,0xd4,0x63,
0x11,0x63,0x63,4,0x31,0x3c,0x32,0xa2,0x42,0x33,0xa2,0x56,0x38,0xa2,0x64,0x39,
0x10,0x31,0xa3,0x5b,9,0x35,0xa,0x35,0x3f,0x36,0x41,0x37,0x43,0x38,0x45,0x39,
0x47,0x30,0x30,0x31,0x3c,0x32,0x42,0x33,0x4e,0x34,0x3d,0x34,1,0x33,0xa3,0x67,
0x37,0xa3,0x6b,0x36,0x10,0x38,0xa3,0x76,0x38,1,0x32,0xa3,0x7a,0x39,0xa3,0x81,
0x3a,2,0x30,0xa3,0x82,0x32,0xa3,0x84,0x33,0xa3,0x85,9,0x35,0xa,0x35,0x53,
0x36,0x55,0x37,0x57,0x38,0x59,0x39,0x5b,0x30,0x49,0x31,0x4b,0x32,0x4d,0x33,0x4f,
0x34,0x51,6,0x33,8,0x33,0x63,0x34,0x65,0x35,0x67,0x36,0x69,0x30,0x5d,0x31,
0x5f,0x32,0x61,0x10,0x34,0xa3,0x54,0xa2,0xe6,3,0x62,0xa0,0x6c,0xa3,0xe4,0x72,
0xa3,0xe8,0x74,2,0x61,0x74,0x62,0x7c,0x74,0x14,0x61,0x63,0x68,0x65,0x64,1,
0x61,0x3e,0x62,0x13,0x65,0x6c,0x6f,0x77,0xa2,0xca,0x13,0x6c,0x65,0x66,0x74,0xa3,
0xc8,0x13,0x62,0x6f,0x76,0x65,0xa2,0xd6,0x14,0x72,0x69,0x67,0x68,0x74,0xa3,0xd8,
0xa2,0xd6,0x10,0x72,0xa3,0xd8,0xa2,0xca,0x10,0x6c,0xa3,0xc8,0x12,0x6f,0x76,0x65,
0xa2,0xe6,1,0x6c,0x30,0x72,0x13,0x69,0x67,0x68,0x74,0xa3,0xe8,0x12,0x65,0x66,
0x74,0xa3,0xe4,0xa2,0xdc,2,0x65,0x2c,0x6c,0xa3,0xda,0x72,0xa3,0xde,0x12,0x6c,
0x6f,0x77,0xa2,0xdc,1,0x6c,0x30,0x72,0x13,0x69,0x67,0x68,0x74,0xa3,0xde,0x12,
0x65,0x66,0x74,0xa3,0xda,0xb,0x6e,0xc0,0xca,0x72,0x5f,0x72,0x46,0x73,0xa2,0x48,
0x77,1,0x68,0x24,0x73,0x33,0x17,0x69,0x74,0x65,0x73,0x70,0x61,0x63,0x65,0x33,
0x22,1,0x69,0x30,0x6c,2,0x65,0x3d,0x69,0x4b,0x6f,0x3f,0x18,0x67,0x68,0x74,
0x74,0x6f,0x6c,0x65,0x66,0x74,0x22,2,0x65,0x38,0x69,0x48,0x6f,0x16,0x76,0x65,
0x72,0x72,0x69,0x64,0x65,0x3f,0x17,0x6d,0x62,0x65,0x64,0x64,0x69,0x6e,0x67,0x3d,
0x15,0x73,0x6f,0x6c,0x61,0x74,0x65,0x4b,0x30,0x1e,0x65,0x67,0x6d,0x65,0x6e,0x74,
0x73,0x65,0x70,0x61,0x72,0x61,0x74,0x6f,0x72,0x31,0x6e,0xa2,0x41,0x6f,0xa2,0x53,
0x70,2,0x61,0x66,0x64,0x86,0x6f,0x1b,0x70,0x64,0x69,0x72,0x65,0x63,0x74,0x69,
0x6f,0x6e,0x61,0x6c,1,0x66,0x32,0x69,0x15,0x73,0x6f,0x6c,0x61,0x74,0x65,0x4d,
0x14,0x6f,0x72,0x6d,0x61,0x74,0x41,0x1f,0x72,0x61,0x67,0x72,0x61,0x70,0x68,0x73,
0x65,0x70,0x61,0x72,0x61,0x74,0x6f,0x72,0x2f,1,0x66,0x41,0x69,0x4d,1,0x6f,
0x28,0x73,0x10,0x6d,0x43,0x1b,0x6e,0x73,0x70,0x61,0x63,0x69,0x6e,0x67,0x6d,0x61,
0x72,0x6b,0x43,1,0x6e,0x35,0x74,0x19,0x68,0x65,0x72,0x6e,0x65,0x75,0x74,0x72,
0x61,0x6c,0x35,0x65,0x88,0x65,0x98,0x66,0xa2,0x6a,0x6c,0x20,1,0x65,0x30,0x72,
2,0x65,0x37,0x69,0x49,0x6f,0x39,0x18,0x66,0x74,0x74,0x6f,0x72,0x69,0x67,0x68,
0x74,0x20,2,0x65,0x38,0x69,0x48,0x6f,0x16,0x76,0x65,0x72,0x72,0x69,0x64,0x65,
0x39,0x17,0x6d,0x62,0x65,0x64,0x64,0x69,0x6e,0x67,0x37,0x15,0x73,0x6f,0x6c,0x61,
0x74,0x65,0x49,3,0x6e,0x25,0x73,0x27,0x74,0x29,0x75,0x15,0x72,0x6f,0x70,0x65,
0x61,0x6e,2,0x6e,0x3c,0x73,0x46,0x74,0x18,0x65,0x72,0x6d,0x69,0x6e,0x61,0x74,
0x6f,0x72,0x29,0x14,0x75,0x6d,0x62,0x65,0x72,0x25,0x17,0x65,0x70,0x61,0x72,0x61,
0x74,0x6f,0x72,0x27,1,0x69,0x28,0x73,0x10,0x69,0x47,0x1f,0x72,0x73,0x74,0x73,
0x74,0x72,0x6f,0x6e,0x67,0x69,0x73,0x6f,0x6c,0x61,0x74,0x65,0x47,0x61,0x4e,0x62,
0x84,0x63,1,0x6f,0x24,0x73,0x2d,0x1c,0x6d,0x6d,0x6f,0x6e,0x73,0x65,0x70,0x61,
0x72,0x61,0x74,0x6f,0x72,0x2d,2,0x6c,0x3b,0x6e,0x2b,0x72,0x13,0x61,0x62,0x69,
0x63,1,0x6c,0x30,0x6e,0x14,0x75,0x6d,0x62,0x65,0x72,0x2b,0x14,0x65,0x74,0x74,
0x65,0x72,0x3b,0x2e,1,0x6e,0x45,0x6f,0x1c,0x75,0x6e,0x64,0x61,0x72,0x79,0x6e,
0x65,0x75,0x74,0x72,0x61,0x6c,0x45,0,0x16,0x6d,0xc8,0xc8,0x74,0xc1,0xee,0x77,
0x6a,0x77,0x48,0x79,0x70,0x7a,0x1d,0x61,0x6e,0x61,0x62,0x61,0x7a,0x61,0x72,0x73,
0x71,0x75,0x61,0x72,0x65,0xa5,0x18,0x10,0x61,1,0x6e,0x36,0x72,0x16,0x61,0x6e,
0x67,0x63,0x69,0x74,0x69,0xa3,0xfc,0x12,0x63,0x68,0x6f,0xa5,0x2c,1,0x65,0x88,
0x69,2,0x6a,0x3c,0x72,0x68,0x73,0x17,0x79,0x6c,0x6c,0x61,0x62,0x6c,0x65,0x73,
0xa3,0x48,0x12,0x69,0x6e,0x67,0xa2,0x74,0x1e,0x68,0x65,0x78,0x61,0x67,0x72,0x61,
0x6d,0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,0xa3,0x74,0x16,0x61,0x64,0x69,0x63,0x61,
0x6c,0x73,0xa3,0x49,0x13,0x7a,0x69,0x64,0x69,0xa5,0x34,0x74,0xa2,0x59,0x75,0xa4,
0x35,0x76,2,0x61,0x36,0x65,0x7a,0x73,0xa2,0x6c,0x12,0x73,0x75,0x70,0xa3,0x7d,
1,0x69,0xa3,0x9f,0x72,0x1e,0x69,0x61,0x74,0x69,0x6f,0x6e,0x73,0x65,0x6c,0x65,
0x63,0x74,0x6f,0x72,0x73,0xa2,0x6c,0x19,0x73,0x75,0x70,0x70,0x6c,0x65,0x6d,0x65,
0x6e,0x74,0xa3,0x7d,1,0x64,0x3c,0x72,0x19,0x74,0x69,0x63,0x61,0x6c,0x66,0x6f,
0x72,0x6d,0x73,0xa3,0x91,0x14,0x69,0x63,0x65,0x78,0x74,0xa2,0xaf,0x16,0x65,0x6e,
0x73,0x69,0x6f,0x6e,0x73,0xa3,0xaf,4,0x61,0x68,0x65,0xa2,0xad,0x68,0xa2,0xb0,
0x69,0xa2,0xb8,0x72,0x1c,0x61,0x6e,0x73,0x70,0x6f,0x72,0x74,0x61,0x6e,0x64,0x6d,
0x61,0x70,0xa2,0xcf,0x16,0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,0xa3,0xcf,4,0x67,
0x7e,0x69,0xa2,0x41,0x6b,0xa2,0x6a,0x6d,0xa2,0x6c,0x6e,0x12,0x67,0x75,0x74,0xa4,
0x10,1,0x63,0x40,0x73,0x11,0x75,0x70,0xa4,0x33,0x16,0x70,0x6c,0x65,0x6d,0x65,
0x6e,0x74,0xa5,0x33,0x18,0x6f,0x6d,0x70,0x6f,0x6e,0x65,0x6e,0x74,0x73,0xa5,0x11,
2,0x61,0x2a,0x62,0x32,0x73,0xa3,0x60,0x12,0x6c,0x6f,0x67,0xa3,0x62,0x13,0x61,
0x6e,0x77,0x61,0xa3,0x65,3,0x6c,0x52,0x74,0x56,0x76,0x5e,0x78,0x16,0x75,0x61,
0x6e,0x6a,0x69,0x6e,0x67,0xa2,0x7c,0x16,0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,0xa3,
0x7c,0x10,0x65,0xa3,0x70,0x12,0x68,0x61,0x6d,0xa3,0xae,0x12,0x69,0x65,0x74,0xa3,
0xb7,0x11,0x72,0x69,0xa3,0xdc,0x11,0x69,0x6c,0x48,0x12,0x73,0x75,0x70,0xa4,0x2b,
0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa5,0x2b,0x13,0x6c,0x75,0x67,0x75,0x4b,
0x10,0x61,1,0x61,0x24,0x69,0x53,0x11,0x6e,0x61,0x3d,2,0x62,0x34,0x66,0x3c,
0x72,0x13,0x68,0x75,0x74,0x61,0xa3,0xfb,0x13,0x65,0x74,0x61,0x6e,0x57,0x14,0x69,
0x6e,0x61,0x67,0x68,0xa3,0x90,2,0x63,0x82,0x67,0x92,0x6e,0x1f,0x69,0x66,0x69,
0x65,0x64,0x63,0x61,0x6e,0x61,0x64,0x69,0x61,0x6e,0x61,0x62,0x6f,0x1f,0x72,0x69,
0x67,0x69,0x6e,0x61,0x6c,0x73,0x79,0x6c,0x6c,0x61,0x62,0x69,0x63,0x73,0x62,0x17,
0x65,0x78,0x74,0x65,0x6e,0x64,0x65,0x64,0xa3,0xad,0x11,0x61,0x73,0x62,0x12,0x65,
0x78,0x74,0xa3,0xad,0x15,0x61,0x72,0x69,0x74,0x69,0x63,0xa3,0x78,0x70,0xc3,0x4b,
0x70,0xa6,0x61,0x72,0xa8,0x1d,0x73,7,0x6f,0xc1,0xbe,0x6f,0xa2,0x69,0x70,0xa2,
0x85,0x75,0xa2,0xa4,0x79,2,0x6c,0x50,0x6d,0x62,0x72,0x12,0x69,0x61,0x63,0x3a,
0x12,0x73,0x75,0x70,0xa4,0x17,0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa5,0x17,
0x17,0x6f,0x74,0x69,0x6e,0x61,0x67,0x72,0x69,0xa3,0x8f,0x13,0x62,0x6f,0x6c,0x73,
1,0x61,0x4c,0x66,0x10,0x6f,0x1f,0x72,0x6c,0x65,0x67,0x61,0x63,0x79,0x63,0x6f,
0x6d,0x70,0x75,0x74,0x69,0x6e,0x67,0xa5,0x32,0x1f,0x6e,0x64,0x70,0x69,0x63,0x74,
0x6f,0x67,0x72,0x61,0x70,0x68,0x73,0x65,0x78,0x74,1,0x61,0xa5,0x2a,0x65,0x14,
0x6e,0x64,0x65,0x64,0x61,0xa5,0x2a,2,0x67,0x34,0x72,0x3e,0x79,0x13,0x6f,0x6d,
0x62,0x6f,0xa5,0x16,0x13,0x64,0x69,0x61,0x6e,0xa5,0x23,0x17,0x61,0x73,0x6f,0x6d,
0x70,0x65,0x6e,0x67,0xa3,0xda,1,0x61,0x32,0x65,0x14,0x63,0x69,0x61,0x6c,0x73,
0xa3,0x56,0x12,0x63,0x69,0x6e,0x1f,0x67,0x6d,0x6f,0x64,0x69,0x66,0x69,0x65,0x72,
0x6c,0x65,0x74,0x74,0x65,0x72,0x73,0x2d,2,0x6e,0x48,0x70,0x76,0x74,0x1d,0x74,
0x6f,0x6e,0x73,0x69,0x67,0x6e,0x77,0x72,0x69,0x74,0x69,0x6e,0x67,0xa5,6,0x15,
0x64,0x61,0x6e,0x65,0x73,0x65,0xa2,0x9b,0x12,0x73,0x75,0x70,0xa2,0xdb,0x16,0x70,
0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa3,0xdb,4,0x61,0xa2,0xa8,0x65,0x5c,0x6d,0x9e,
0x70,0xa2,0x4b,0x73,0x13,0x79,0x6d,0x62,0x6f,0x1f,0x6c,0x73,0x61,0x6e,0x64,0x70,
0x69,0x63,0x74,0x6f,0x67,0x72,0x61,0x70,0x68,0x73,0xa5,5,0x10,0x72,1,0x61,
0x4e,0x73,0x12,0x63,0x72,0x69,0x1f,0x70,0x74,0x73,0x61,0x6e,0x64,0x73,0x75,0x62,
0x73,0x63,0x72,0x69,0x70,0x74,0x73,0x73,0x14,0x6e,0x64,0x73,0x75,0x62,0x73,0x1b,
0x61,0x74,0x68,0x6f,0x70,0x65,0x72,0x61,0x74,0x6f,0x72,0x73,0xa3,0x6a,1,0x6c,
0x40,0x75,1,0x61,0x6e,0x6e,0x17,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0xa3,
0x8e,0x15,0x65,0x6d,0x65,0x6e,0x74,0x61,1,0x6c,0x50,0x72,0x1e,0x79,0x70,0x72,
0x69,0x76,0x61,0x74,0x65,0x75,0x73,0x65,0x61,0x72,0x65,0x61,1,0x61,0xa3,0x6d,
0x62,0xa3,0x6e,3,0x61,0x5c,0x6d,0x78,0x70,0xa2,0x41,0x73,0x13,0x79,0x6d,0x62,
0x6f,0x1f,0x6c,0x73,0x61,0x6e,0x64,0x70,0x69,0x63,0x74,0x6f,0x67,0x72,0x61,0x70,
0x68,0x73,0xa5,5,0x14,0x72,0x72,0x6f,0x77,0x73,2,0x61,0xa3,0x67,0x62,0xa3,
0x68,0x63,0xa3,0xfa,0x13,0x61,0x74,0x68,0x65,0x1f,0x6d,0x61,0x74,0x69,0x63,0x61,
0x6c,0x6f,0x70,0x65,0x72,0x61,0x74,0x6f,0x72,0x73,0xa3,0x6a,0x19,0x75,0x6e,0x63,
0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0xa3,0x8e,0x61,0x88,0x68,0xa2,0x48,0x69,0xa2,
0x71,0x6d,0x12,0x61,0x6c,0x6c,1,0x66,0x46,0x6b,0x15,0x61,0x6e,0x61,0x65,0x78,
0x74,0xa4,0x29,0x15,0x65,0x6e,0x73,0x69,0x6f,0x6e,0xa5,0x29,0x12,0x6f,0x72,0x6d,
1,0x73,0xa3,0x54,0x76,0x16,0x61,0x72,0x69,0x61,0x6e,0x74,0x73,0xa3,0x54,1,
0x6d,0x36,0x75,0x16,0x72,0x61,0x73,0x68,0x74,0x72,0x61,0xa3,0xa1,0x15,0x61,0x72,
0x69,0x74,0x61,0x6e,0xa3,0xac,1,0x61,0x52,0x6f,0x13,0x72,0x74,0x68,0x61,0x1f,
0x6e,0x64,0x66,0x6f,0x72,0x6d,0x61,0x74,0x63,0x6f,0x6e,0x74,0x72,0x6f,0x6c,0x73,
0xa3,0xf7,1,0x72,0x2e,0x76,0x12,0x69,0x61,0x6e,0xa3,0x79,0x12,0x61,0x64,0x61,
0xa3,0xd9,1,0x64,0x50,0x6e,0x13,0x68,0x61,0x6c,0x61,0x50,0x1d,0x61,0x72,0x63,
0x68,0x61,0x69,0x63,0x6e,0x75,0x6d,0x62,0x65,0x72,0x73,0xa3,0xf9,0x13,0x64,0x68,
0x61,0x6d,0xa3,0xf8,5,0x72,0x35,0x72,0x44,0x73,0x64,0x75,1,0x61,0xa3,0x4e,
0x6e,0x17,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0x71,0x17,0x69,0x76,0x61,0x74,
0x65,0x75,0x73,0x65,0xa2,0x4e,0x13,0x61,0x72,0x65,0x61,0xa3,0x4e,0x1b,0x61,0x6c,
0x74,0x65,0x72,0x70,0x61,0x68,0x6c,0x61,0x76,0x69,0xa3,0xf6,0x61,0x40,0x68,0x82,
0x6c,0x19,0x61,0x79,0x69,0x6e,0x67,0x63,0x61,0x72,0x64,0x73,0xa3,0xcc,2,0x68,
0x38,0x6c,0x4a,0x75,0x15,0x63,0x69,0x6e,0x68,0x61,0x75,0xa3,0xf5,0x17,0x61,0x77,
0x68,0x68,0x6d,0x6f,0x6e,0x67,0xa3,0xf3,0x15,0x6d,0x79,0x72,0x65,0x6e,0x65,0xa3,
0xf4,1,0x61,0x8e,0x6f,1,0x65,0x74,0x6e,0x16,0x65,0x74,0x69,0x63,0x65,0x78,
0x74,0xa2,0x72,1,0x65,0x2c,0x73,0x11,0x75,0x70,0xa3,0x8d,0x15,0x6e,0x73,0x69,
0x6f,0x6e,0x73,0xa2,0x72,0x19,0x73,0x75,0x70,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,
0xa3,0x8d,0x15,0x6e,0x69,0x63,0x69,0x61,0x6e,0xa3,0x97,1,0x67,0x3e,0x69,0x13,
0x73,0x74,0x6f,0x73,0xa2,0xa6,0x13,0x64,0x69,0x73,0x63,0xa3,0xa6,0x12,0x73,0x70,
0x61,0xa3,0x96,1,0x65,0x5c,0x75,1,0x6d,0x2a,0x6e,0x11,0x69,0x63,0x67,0x10,
0x69,0xa2,0xc0,0x1d,0x6e,0x75,0x6d,0x65,0x72,0x61,0x6c,0x73,0x79,0x6d,0x62,0x6f,
0x6c,0x73,0xa3,0xc0,0x13,0x6a,0x61,0x6e,0x67,0xa3,0xa3,0x6d,0xa2,0xe6,0x6e,0xa8,
0x19,0x6f,6,0x70,0x63,0x70,0x56,0x72,0x8a,0x73,0xa2,0x4c,0x74,0x10,0x74,0x1f,
0x6f,0x6d,0x61,0x6e,0x73,0x69,0x79,0x61,0x71,0x6e,0x75,0x6d,0x62,0x65,0x72,0x73,
0xa5,0x28,0x18,0x74,0x69,0x63,0x61,0x6c,0x63,0x68,0x61,0x72,0x1f,0x61,0x63,0x74,
0x65,0x72,0x72,0x65,0x63,0x6f,0x67,0x6e,0x69,0x74,0x69,0x6f,0x6e,0x85,1,0x69,
0x46,0x6e,0x1e,0x61,0x6d,0x65,0x6e,0x74,0x61,0x6c,0x64,0x69,0x6e,0x67,0x62,0x61,
0x74,0x73,0xa3,0xf2,0x11,0x79,0x61,0x47,1,0x61,0x30,0x6d,0x13,0x61,0x6e,0x79,
0x61,0xa3,0x7a,0x11,0x67,0x65,0xa5,0xf,0x63,0xa2,0x71,0x67,0xa2,0x71,0x6c,1,
0x63,0xa2,0x62,0x64,5,0x70,0x38,0x70,0x36,0x73,0x56,0x74,0x14,0x75,0x72,0x6b,
0x69,0x63,0xa3,0xbf,0x11,0x65,0x72,1,0x6d,0x2e,0x73,0x12,0x69,0x61,0x6e,0xa3,
0x8c,0x11,0x69,0x63,0xa3,0xf1,0x10,0x6f,1,0x67,0x3a,0x75,0x18,0x74,0x68,0x61,
0x72,0x61,0x62,0x69,0x61,0x6e,0xa3,0xbb,0x13,0x64,0x69,0x61,0x6e,0xa5,0x22,0x68,
0x42,0x69,0x54,0x6e,0x1a,0x6f,0x72,0x74,0x68,0x61,0x72,0x61,0x62,0x69,0x61,0x6e,
0xa3,0xf0,0x17,0x75,0x6e,0x67,0x61,0x72,0x69,0x61,0x6e,0xa5,4,0x14,0x74,0x61,
0x6c,0x69,0x63,0xa3,0x58,0x13,0x68,0x69,0x6b,0x69,0xa3,0x9d,0x10,0x72,0x85,0x12,
0x68,0x61,0x6d,0x65,6,0x6f,0x86,0x6f,0x6c,0x72,0xa2,0x61,0x75,0xa2,0x62,0x79,
0x14,0x61,0x6e,0x6d,0x61,0x72,0x58,0x12,0x65,0x78,0x74,2,0x61,0xa3,0xb6,0x62,
0xa3,0xee,0x65,0x13,0x6e,0x64,0x65,0x64,1,0x61,0xa3,0xb6,0x62,0xa3,0xee,1,
0x64,0x52,0x6e,0x15,0x67,0x6f,0x6c,0x69,0x61,0x6e,0x6a,0x12,0x73,0x75,0x70,0xa4,
0xd,0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa5,0xd,0x10,0x69,0xa2,0xec,0x13,
0x66,0x69,0x65,0x72,1,0x6c,0x3c,0x74,0x19,0x6f,0x6e,0x65,0x6c,0x65,0x74,0x74,
0x65,0x72,0x73,0xa3,0x8a,0x15,0x65,0x74,0x74,0x65,0x72,0x73,0x2d,0x10,0x6f,0xa3,
0xed,1,0x6c,0x44,0x73,0x11,0x69,0x63,0xa2,0x5c,0x18,0x61,0x6c,0x73,0x79,0x6d,
0x62,0x6f,0x6c,0x73,0xa3,0x5c,0x13,0x74,0x61,0x6e,0x69,0xa5,3,0x61,0xa2,0x9b,
0x65,0xa4,0x4c,0x69,1,0x61,0xa2,0x8f,0x73,0x10,0x63,5,0x70,0x18,0x70,0xa2,
0x71,0x73,0x36,0x74,0x17,0x65,0x63,0x68,0x6e,0x69,0x63,0x61,0x6c,0x81,0x15,0x79,
0x6d,0x62,0x6f,0x6c,0x73,0x8f,0x61,0xa2,0x66,0x65,0x46,0x6d,0x19,0x61,0x74,0x68,
0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,1,0x61,0xa3,0x66,0x62,0xa3,0x69,0x17,0x6c,
0x6c,0x61,0x6e,0x65,0x6f,0x75,0x73,2,0x6d,0x3a,0x73,0x6c,0x74,0x17,0x65,0x63,
0x68,0x6e,0x69,0x63,0x61,0x6c,0x81,0x11,0x61,0x74,0x1f,0x68,0x65,0x6d,0x61,0x74,
0x69,0x63,0x61,0x6c,0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,1,0x61,0xa3,0x66,0x62,
0xa3,0x69,0x15,0x79,0x6d,0x62,0x6f,0x6c,0x73,0x8e,0x12,0x61,0x6e,0x64,1,0x61,
0x3c,0x70,0x19,0x69,0x63,0x74,0x6f,0x67,0x72,0x61,0x70,0x68,0x73,0xa3,0xcd,0x14,
0x72,0x72,0x6f,0x77,0x73,0xa3,0x73,0x10,0x6f,0xa3,0xd8,7,0x72,0x6f,0x72,0x44,
0x73,0x4e,0x74,0x62,0x79,0x19,0x61,0x6e,0x6e,0x75,0x6d,0x65,0x72,0x61,0x6c,0x73,
0xa5,0x20,0x13,0x63,0x68,0x65,0x6e,0xa5,0xc,0x18,0x61,0x72,0x61,0x6d,0x67,0x6f,
0x6e,0x64,0x69,0xa5,0x14,0x10,0x68,2,0x61,0x3a,0x65,0x4a,0x6f,0x17,0x70,0x65,
0x72,0x61,0x74,0x6f,0x72,0x73,0x7f,0x16,0x6c,0x70,0x68,0x61,0x6e,0x75,0x6d,0xa3,
0x5d,0x16,0x6d,0x61,0x74,0x69,0x63,0x61,0x6c,1,0x61,0x36,0x6f,0x17,0x70,0x65,
0x72,0x61,0x74,0x6f,0x72,0x73,0x7f,0x11,0x6c,0x70,0x1f,0x68,0x61,0x6e,0x75,0x6d,
0x65,0x72,0x69,0x63,0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,0xa3,0x5d,0x68,0x50,0x6b,
0x7e,0x6c,0x88,0x6e,1,0x64,0x34,0x69,0x15,0x63,0x68,0x61,0x65,0x61,0x6e,0xa3,
0xea,0x12,0x61,0x69,0x63,0xa3,0xc6,1,0x61,0x3e,0x6a,0x12,0x6f,0x6e,0x67,0xa2,
0xaa,0x14,0x74,0x69,0x6c,0x65,0x73,0xa3,0xaa,0x13,0x6a,0x61,0x6e,0x69,0xa3,0xe9,
0x13,0x61,0x73,0x61,0x72,0xa5,0x1f,0x15,0x61,0x79,0x61,0x6c,0x61,0x6d,0x4f,3,
0x64,0x6c,0x65,0x7e,0x6e,0xa2,0x47,0x72,0x14,0x6f,0x69,0x74,0x69,0x63,1,0x63,
0x3c,0x68,0x19,0x69,0x65,0x72,0x6f,0x67,0x6c,0x79,0x70,0x68,0x73,0xa3,0xd7,0x15,
0x75,0x72,0x73,0x69,0x76,0x65,0xa3,0xd6,0x17,0x65,0x66,0x61,0x69,0x64,0x72,0x69,
0x6e,0xa5,0x21,0x17,0x74,0x65,0x69,0x6d,0x61,0x79,0x65,0x6b,0xa2,0xb8,0x12,0x65,
0x78,0x74,0xa2,0xd5,0x16,0x65,0x6e,0x73,0x69,0x6f,0x6e,0x73,0xa3,0xd5,0x18,0x64,
0x65,0x6b,0x69,0x6b,0x61,0x6b,0x75,0x69,0xa3,0xeb,6,0x6b,0x3b,0x6b,0x56,0x6f,
0x5a,0x75,0x64,0x79,0x11,0x69,0x61,0x1f,0x6b,0x65,0x6e,0x67,0x70,0x75,0x61,0x63,
0x68,0x75,0x65,0x68,0x6d,0x6f,0x6e,0x67,0xa5,0x27,0x10,0x6f,0xa3,0x92,0x14,0x62,
0x6c,0x6f,0x63,0x6b,0x21,1,0x6d,0x2c,0x73,0x11,0x68,0x75,0xa5,0x15,0x17,0x62,
0x65,0x72,0x66,0x6f,0x72,0x6d,0x73,0x7b,0x61,0x44,0x62,0x21,0x65,0x10,0x77,1,
0x61,0xa5,0xe,0x74,0x14,0x61,0x69,0x6c,0x75,0x65,0xa3,0x8b,1,0x62,0x38,0x6e,
0x17,0x64,0x69,0x6e,0x61,0x67,0x61,0x72,0x69,0xa5,0x26,0x15,0x61,0x74,0x61,0x65,
0x61,0x6e,0xa3,0xef,0x67,0xc4,0x32,0x6a,0xc1,0xb9,0x6a,0xa2,0xd5,0x6b,0xa2,0xee,
0x6c,4,0x61,0x54,0x65,0xa2,0x61,0x69,0xa2,0x78,0x6f,0xa2,0xb7,0x79,1,0x63,
0x2e,0x64,0x12,0x69,0x61,0x6e,0xa3,0xa9,0x12,0x69,0x61,0x6e,0xa3,0xa7,1,0x6f,
0x55,0x74,0x11,0x69,0x6e,1,0x31,0x82,0x65,0x11,0x78,0x74,4,0x61,0x5c,0x62,
0x29,0x63,0xa3,0x94,0x64,0xa3,0x95,0x65,0xa2,0xe7,0x13,0x6e,0x64,0x65,0x64,4,
0x61,0x36,0x62,0x29,0x63,0xa3,0x94,0x64,0xa3,0x95,0x65,0xa3,0xe7,0x26,0x18,0x64,
0x64,0x69,0x74,0x69,0x6f,0x6e,0x61,0x6c,0x6d,0x24,0x12,0x73,0x75,0x70,0x24,0x16,
0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0x25,1,0x70,0x42,0x74,0x1d,0x74,0x65,0x72,
0x6c,0x69,0x6b,0x65,0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,0x79,0x12,0x63,0x68,0x61,
0xa3,0x9c,2,0x6d,0x4e,0x6e,0x54,0x73,0x10,0x75,0xa2,0xb0,0x12,0x73,0x75,0x70,
0xa4,0x31,0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa5,0x31,0x11,0x62,0x75,0xa3,
0x6f,0x12,0x65,0x61,0x72,1,0x61,0xa3,0xe8,0x62,1,0x69,0x38,0x73,0x17,0x79,
0x6c,0x6c,0x61,0x62,0x61,0x72,0x79,0xa3,0x75,0x17,0x64,0x65,0x6f,0x67,0x72,0x61,
0x6d,0x73,0xa3,0x76,0x1a,0x77,0x73,0x75,0x72,0x72,0x6f,0x67,0x61,0x74,0x65,0x73,
0xa3,0x4d,0x10,0x61,1,0x6d,0x32,0x76,0x14,0x61,0x6e,0x65,0x73,0x65,0xa3,0xb5,
0x10,0x6f,0x5c,0x12,0x65,0x78,0x74,1,0x61,0xa3,0xb4,0x62,0xa3,0xb9,1,0x61,
0xa2,0x43,0x68,4,0x61,0x40,0x69,0x50,0x6d,0x6e,0x6f,0x86,0x75,0x15,0x64,0x61,
0x77,0x61,0x64,0x69,0xa3,0xe6,0x16,0x72,0x6f,0x73,0x68,0x74,0x68,0x69,0xa3,0x89,
0x1d,0x74,0x61,0x6e,0x73,0x6d,0x61,0x6c,0x6c,0x73,0x63,0x72,0x69,0x70,0x74,0xa5,
0x30,0x11,0x65,0x72,0x68,0x16,0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,0xa3,0x71,0x12,
0x6a,0x6b,0x69,0xa3,0xe5,3,0x69,0x3a,0x6e,0x42,0x74,0xa2,0x51,0x79,0x13,0x61,
0x68,0x6c,0x69,0xa3,0xa2,0x12,0x74,0x68,0x69,0xa3,0xc1,3,0x61,0x34,0x62,0x76,
0x67,0x7c,0x6e,0x12,0x61,0x64,0x61,0x4d,1,0x65,0x40,0x73,0x11,0x75,0x70,0xa2,
0xcb,0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa3,0xcb,0x11,0x78,0x74,1,0x61,
0xa5,0x13,0x65,0x14,0x6e,0x64,0x65,0x64,0x61,0xa5,0x13,0x11,0x75,0x6e,0xa3,0x42,
0x11,0x78,0x69,0x96,0x17,0x72,0x61,0x64,0x69,0x63,0x61,0x6c,0x73,0x97,0x14,0x61,
0x6b,0x61,0x6e,0x61,0x9e,1,0x65,0x4c,0x70,0x10,0x68,0x1f,0x6f,0x6e,0x65,0x74,
0x69,0x63,0x65,0x78,0x74,0x65,0x6e,0x73,0x69,0x6f,0x6e,0x73,0xa3,0x6b,0x11,0x78,
0x74,0xa3,0x6b,0x67,0xa2,0xb5,0x68,0xa4,0x84,0x69,3,0x64,0x4c,0x6d,0xa2,0x55,
0x6e,0xa2,0x62,0x70,0x13,0x61,0x65,0x78,0x74,0x2a,0x16,0x65,0x6e,0x73,0x69,0x6f,
0x6e,0x73,0x2b,1,0x63,0x99,0x65,0x17,0x6f,0x67,0x72,0x61,0x70,0x68,0x69,0x63,
1,0x64,0x56,0x73,0x15,0x79,0x6d,0x62,0x6f,0x6c,0x73,0xa4,0xb,0x1d,0x61,0x6e,
0x64,0x70,0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0xa5,0xb,0x13,0x65,
0x73,0x63,0x72,0x1f,0x69,0x70,0x74,0x69,0x6f,0x6e,0x63,0x68,0x61,0x72,0x61,0x63,
0x74,0x65,0x72,0x73,0x99,0x1c,0x70,0x65,0x72,0x69,0x61,0x6c,0x61,0x72,0x61,0x6d,
0x61,0x69,0x63,0xa3,0xba,1,0x64,0x62,0x73,0x1b,0x63,0x72,0x69,0x70,0x74,0x69,
0x6f,0x6e,0x61,0x6c,0x70,0x61,1,0x68,0x32,0x72,0x14,0x74,0x68,0x69,0x61,0x6e,
0xa3,0xbd,0x13,0x6c,0x61,0x76,0x69,0xa3,0xbe,0x11,0x69,0x63,1,0x6e,0x3e,0x73,
0x1a,0x69,0x79,0x61,0x71,0x6e,0x75,0x6d,0x62,0x65,0x72,0x73,0xa5,0x1e,0x19,0x75,
0x6d,0x62,0x65,0x72,0x66,0x6f,0x72,0x6d,0x73,0xa3,0xb2,4,0x65,0x74,0x6c,0xa2,
0x82,0x6f,0xa2,0x9a,0x72,0xa2,0x9e,0x75,2,0x6a,0x34,0x6e,0x3e,0x72,0x14,0x6d,
0x75,0x6b,0x68,0x69,0x43,0x14,0x61,0x72,0x61,0x74,0x69,0x45,0x18,0x6a,0x61,0x6c,
0x61,0x67,0x6f,0x6e,0x64,0x69,0xa5,0x1c,1,0x6e,0xa2,0x46,0x6f,1,0x6d,0x6e,
0x72,0x13,0x67,0x69,0x61,0x6e,0x5a,1,0x65,0x40,0x73,0x11,0x75,0x70,0xa2,0x87,
0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa3,0x87,0x11,0x78,0x74,0xa4,0x1b,0x14,
0x65,0x6e,0x64,0x65,0x64,0xa5,0x1b,0x1a,0x65,0x74,0x72,0x69,0x63,0x73,0x68,0x61,
0x70,0x65,0x73,0x8c,0x12,0x65,0x78,0x74,0xa2,0xe3,0x14,0x65,0x6e,0x64,0x65,0x64,
0xa3,0xe3,0x1e,0x65,0x72,0x61,0x6c,0x70,0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,
0x6f,0x6e,0x71,0x17,0x61,0x67,0x6f,0x6c,0x69,0x74,0x69,0x63,0xa2,0x88,0x12,0x73,
0x75,0x70,0xa4,0xa,0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa5,0xa,0x13,0x74,
0x68,0x69,0x63,0xa3,0x59,1,0x61,0x5c,0x65,0x11,0x65,0x6b,0x30,1,0x61,0x38,
0x65,0x11,0x78,0x74,0x6e,0x14,0x65,0x6e,0x64,0x65,0x64,0x6f,0x17,0x6e,0x64,0x63,
0x6f,0x70,0x74,0x69,0x63,0x31,0x13,0x6e,0x74,0x68,0x61,0xa3,0xe4,2,0x61,0xa2,
0x48,0x65,0xa2,0xdf,0x69,1,0x67,0x30,0x72,0x14,0x61,0x67,0x61,0x6e,0x61,0x9d,
0x10,0x68,1,0x70,0x3a,0x73,0x18,0x75,0x72,0x72,0x6f,0x67,0x61,0x74,0x65,0x73,
0xa3,0x4b,1,0x72,0x3c,0x75,0x19,0x73,0x75,0x72,0x72,0x6f,0x67,0x61,0x74,0x65,
0x73,0xa3,0x4c,0x11,0x69,0x76,0x1f,0x61,0x74,0x65,0x75,0x73,0x65,0x73,0x75,0x72,
0x72,0x6f,0x67,0x61,0x74,0x65,0x73,0xa3,0x4c,2,0x6c,0x32,0x6e,0x9a,0x74,0x12,
0x72,0x61,0x6e,0xa5,2,0x10,0x66,2,0x61,0x58,0x6d,0x70,0x77,0x14,0x69,0x64,
0x74,0x68,0x61,0x1f,0x6e,0x64,0x66,0x75,0x6c,0x6c,0x77,0x69,0x64,0x74,0x68,0x66,
0x6f,0x72,0x6d,0x73,0xa3,0x57,0x1a,0x6e,0x64,0x66,0x75,0x6c,0x6c,0x66,0x6f,0x72,
0x6d,0x73,0xa3,0x57,0x13,0x61,0x72,0x6b,0x73,0xa3,0x52,2,0x67,0x34,0x69,0xa2,
0x45,0x75,0x12,0x6e,0x6f,0x6f,0xa3,0x63,0x11,0x75,0x6c,0xa2,0x4a,2,0x63,0x3c,
0x6a,0x5e,0x73,0x17,0x79,0x6c,0x6c,0x61,0x62,0x6c,0x65,0x73,0xa3,0x4a,0x1f,0x6f,
0x6d,0x70,0x61,0x74,0x69,0x62,0x69,0x6c,0x69,0x74,0x79,0x6a,0x61,0x6d,0x6f,0xa3,
0x41,0x12,0x61,0x6d,0x6f,0x5c,0x17,0x65,0x78,0x74,0x65,0x6e,0x64,0x65,0x64,1,
0x61,0xa3,0xb4,0x62,0xa3,0xb9,0x19,0x66,0x69,0x72,0x6f,0x68,0x69,0x6e,0x67,0x79,
0x61,0xa5,0x1d,0x13,0x62,0x72,0x65,0x77,0x37,0x61,0xa4,5,0x62,0xa6,0x45,0x63,
0xa8,0x1a,0x64,0xac,0xb8,0x65,5,0x6d,0xa2,0x6d,0x86,0x6e,0x96,0x74,0x15,0x68,
0x69,0x6f,0x70,0x69,0x63,0x5e,1,0x65,0x40,0x73,0x11,0x75,0x70,0xa2,0x86,0x16,
0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa3,0x86,0x11,0x78,0x74,0xa2,0x85,1,0x61,
0xa3,0xc8,0x65,0x13,0x6e,0x64,0x65,0x64,0xa2,0x85,0x10,0x61,0xa3,0xc8,0x16,0x6f,
0x74,0x69,0x63,0x6f,0x6e,0x73,0xa3,0xce,0x15,0x63,0x6c,0x6f,0x73,0x65,0x64,2,
0x61,0x5a,0x63,0x9e,0x69,0x1c,0x64,0x65,0x6f,0x67,0x72,0x61,0x70,0x68,0x69,0x63,
0x73,0x75,0x70,0xa2,0xc4,0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa3,0xc4,0x16,
0x6c,0x70,0x68,0x61,0x6e,0x75,0x6d,0x86,1,0x65,0x2c,0x73,0x11,0x75,0x70,0xa3,
0xc3,0x13,0x72,0x69,0x63,0x73,0x86,0x18,0x75,0x70,0x70,0x6c,0x65,0x6d,0x65,0x6e,
0x74,0xa3,0xc3,0x11,0x6a,0x6b,0xa2,0x44,0x1f,0x6c,0x65,0x74,0x74,0x65,0x72,0x73,
0x61,0x6e,0x64,0x6d,0x6f,0x6e,0x74,0x68,0x73,0xa3,0x44,0x61,0x4a,0x67,0x76,0x6c,
1,0x62,0x30,0x79,0x13,0x6d,0x61,0x69,0x63,0xa5,0x25,0x13,0x61,0x73,0x61,0x6e,
0xa3,0xe2,0x13,0x72,0x6c,0x79,0x64,0x1f,0x79,0x6e,0x61,0x73,0x74,0x69,0x63,0x63,
0x75,0x6e,0x65,0x69,0x66,0x6f,0x72,0x6d,0xa5,1,0x1f,0x79,0x70,0x74,0x69,0x61,
0x6e,0x68,0x69,0x65,0x72,0x6f,0x67,0x6c,0x79,0x70,0x68,1,0x66,0x26,0x73,0xa3,
0xc2,0x1c,0x6f,0x72,0x6d,0x61,0x74,0x63,0x6f,0x6e,0x74,0x72,0x6f,0x6c,0x73,0xa5,
0x24,7,0x6e,0xc0,0xe5,0x6e,0x3e,0x72,0xa2,0x5d,0x73,0xa2,0xd8,0x76,0x14,0x65,
0x73,0x74,0x61,0x6e,0xa3,0xbc,1,0x61,0x92,0x63,0x13,0x69,0x65,0x6e,0x74,1,
0x67,0x34,0x73,0x15,0x79,0x6d,0x62,0x6f,0x6c,0x73,0xa3,0xa5,0x13,0x72,0x65,0x65,
0x6b,1,0x6d,0x34,0x6e,0x15,0x75,0x6d,0x62,0x65,0x72,0x73,0xa3,0x7f,0x13,0x75,
0x73,0x69,0x63,0xa2,0x7e,0x19,0x61,0x6c,0x6e,0x6f,0x74,0x61,0x74,0x69,0x6f,0x6e,
0xa3,0x7e,0x10,0x74,0x1f,0x6f,0x6c,0x69,0x61,0x6e,0x68,0x69,0x65,0x72,0x6f,0x67,
0x6c,0x79,0x70,0x68,0x73,0xa3,0xfe,2,0x61,0x32,0x6d,0xa2,0x71,0x72,0x12,0x6f,
0x77,0x73,0x7d,0x12,0x62,0x69,0x63,0x38,3,0x65,0x4a,0x6d,0x66,0x70,0xa2,0x43,
0x73,0x11,0x75,0x70,0xa2,0x80,0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa3,0x80,
0x11,0x78,0x74,1,0x61,0xa3,0xd2,0x65,0x14,0x6e,0x64,0x65,0x64,0x61,0xa3,0xd2,
0x12,0x61,0x74,0x68,0xa2,0xd3,0x18,0x65,0x6d,0x61,0x74,0x69,0x63,0x61,0x6c,0x61,
0x1f,0x6c,0x70,0x68,0x61,0x62,0x65,0x74,0x69,0x63,0x73,0x79,0x6d,0x62,0x6f,0x6c,
0x73,0xa3,0xd3,1,0x66,0x42,0x72,0x1e,0x65,0x73,0x65,0x6e,0x74,0x61,0x74,0x69,
0x6f,0x6e,0x66,0x6f,0x72,0x6d,0x73,1,0x61,0xa3,0x51,0x62,0xa3,0x55,0x14,0x65,
0x6e,0x69,0x61,0x6e,0x35,0x12,0x63,0x69,0x69,0x23,0x64,0x9e,0x65,0xa2,0x42,0x68,
0xa2,0x4d,0x6c,1,0x63,0x62,0x70,0x17,0x68,0x61,0x62,0x65,0x74,0x69,0x63,0x70,
1,0x66,0xa3,0x50,0x72,0x1e,0x65,0x73,0x65,0x6e,0x74,0x61,0x74,0x69,0x6f,0x6e,
0x66,0x6f,0x72,0x6d,0x73,0xa3,0x50,0x16,0x68,0x65,0x6d,0x69,0x63,0x61,0x6c,0xa2,
0xd0,0x16,0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,0xa3,0xd0,0x12,0x6c,0x61,0x6d,0xa5,
7,0x1a,0x67,0x65,0x61,0x6e,0x6e,0x75,0x6d,0x62,0x65,0x72,0x73,0xa3,0x77,0x11,
0x6f,0x6d,0xa3,0xfd,7,0x6f,0x71,0x6f,0x64,0x72,0xa2,0x41,0x75,0xa2,0x58,0x79,
0x1b,0x7a,0x61,0x6e,0x74,0x69,0x6e,0x65,0x6d,0x75,0x73,0x69,0x63,0xa2,0x5b,0x18,
0x61,0x6c,0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,0xa3,0x5b,1,0x70,0x34,0x78,0x16,
0x64,0x72,0x61,0x77,0x69,0x6e,0x67,0x89,0x14,0x6f,0x6d,0x6f,0x66,0x6f,0xa0,0x12,
0x65,0x78,0x74,0xa2,0x43,0x14,0x65,0x6e,0x64,0x65,0x64,0xa3,0x43,0x10,0x61,1,
0x68,0x40,0x69,0x12,0x6c,0x6c,0x65,0x92,0x17,0x70,0x61,0x74,0x74,0x65,0x72,0x6e,
0x73,0x93,0x11,0x6d,0x69,0xa3,0xc9,1,0x67,0x2c,0x68,0x11,0x69,0x64,0xa3,0x64,
0x14,0x69,0x6e,0x65,0x73,0x65,0xa3,0x81,0x61,0x48,0x65,0xa2,0x4e,0x68,0xa2,0x52,
0x6c,0x1a,0x6f,0x63,0x6b,0x65,0x6c,0x65,0x6d,0x65,0x6e,0x74,0x73,0x8b,3,0x6c,
0x34,0x6d,0x40,0x73,0x66,0x74,0x11,0x61,0x6b,0xa3,0xc7,0x14,0x69,0x6e,0x65,0x73,
0x65,0xa3,0x93,0x11,0x75,0x6d,0xa2,0xb1,0x12,0x73,0x75,0x70,0xa2,0xca,0x16,0x70,
0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa3,0xca,1,0x69,0x30,0x73,0x13,0x61,0x76,0x61,
0x68,0xa3,0xdd,0x15,0x63,0x6c,0x61,0x74,0x69,0x6e,0x23,0x14,0x6e,0x67,0x61,0x6c,
0x69,0x41,0x16,0x61,0x69,0x6b,0x73,0x75,0x6b,0x69,0xa5,8,5,0x6f,0xc1,0x4c,
0x6f,0xa2,0x55,0x75,0xa4,0x10,0x79,1,0x70,0x9c,0x72,0x14,0x69,0x6c,0x6c,0x69,
0x63,0x32,1,0x65,0x4c,0x73,0x11,0x75,0x70,0xa2,0x61,0x16,0x70,0x6c,0x65,0x6d,
0x65,0x6e,0x74,0xa2,0x61,0x12,0x61,0x72,0x79,0xa3,0x61,0x11,0x78,0x74,3,0x61,
0xa3,0x9e,0x62,0xa3,0xa0,0x63,0xa5,9,0x65,0x13,0x6e,0x64,0x65,0x64,2,0x61,
0xa3,0x9e,0x62,0xa3,0xa0,0x63,0xa5,9,0x1c,0x72,0x69,0x6f,0x74,0x73,0x79,0x6c,
0x6c,0x61,0x62,0x61,0x72,0x79,0xa3,0x7b,3,0x6d,0x5a,0x6e,0xa2,0x95,0x70,0xa2,
0xa0,0x75,0x17,0x6e,0x74,0x69,0x6e,0x67,0x72,0x6f,0x64,0xa2,0x9a,0x17,0x6e,0x75,
0x6d,0x65,0x72,0x61,0x6c,0x73,0xa3,0x9a,2,0x62,0x3a,0x6d,0xa2,0x5f,0x70,0x15,
0x61,0x74,0x6a,0x61,0x6d,0x6f,0xa3,0x41,0x14,0x69,0x6e,0x69,0x6e,0x67,2,0x64,
0x46,0x68,0x9e,0x6d,0x1d,0x61,0x72,0x6b,0x73,0x66,0x6f,0x72,0x73,0x79,0x6d,0x62,
0x6f,0x6c,0x73,0x77,0x1e,0x69,0x61,0x63,0x72,0x69,0x74,0x69,0x63,0x61,0x6c,0x6d,
0x61,0x72,0x6b,0x73,0x2e,2,0x65,0x40,0x66,0xa6,0x41,0x73,0x18,0x75,0x70,0x70,
0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa3,0x83,0x16,0x78,0x74,0x65,0x6e,0x64,0x65,0x64,
0xa3,0xe0,0x17,0x61,0x6c,0x66,0x6d,0x61,0x72,0x6b,0x73,0xa3,0x52,0x11,0x6f,0x6e,
0x1f,0x69,0x6e,0x64,0x69,0x63,0x6e,0x75,0x6d,0x62,0x65,0x72,0x66,0x6f,0x72,0x6d,
0x73,0xa3,0xb2,0x1b,0x74,0x72,0x6f,0x6c,0x70,0x69,0x63,0x74,0x75,0x72,0x65,0x73,
0x83,0x12,0x74,0x69,0x63,0xa2,0x84,0x1b,0x65,0x70,0x61,0x63,0x74,0x6e,0x75,0x6d,
0x62,0x65,0x72,0x73,0xa3,0xdf,1,0x6e,0x3e,0x72,0x1b,0x72,0x65,0x6e,0x63,0x79,
0x73,0x79,0x6d,0x62,0x6f,0x6c,0x73,0x75,0x15,0x65,0x69,0x66,0x6f,0x72,0x6d,0xa2,
0x98,0x16,0x6e,0x75,0x6d,0x62,0x65,0x72,0x73,0xa2,0x99,0x1d,0x61,0x6e,0x64,0x70,
0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0xa3,0x99,0x61,0xa2,0xe1,0x68,
0xa4,0xb,0x6a,0x10,0x6b,0xa2,0x47,4,0x63,0x8c,0x65,0xa2,0x80,0x72,0xa2,0x98,
0x73,0xa2,0xaa,0x75,0x1f,0x6e,0x69,0x66,0x69,0x65,0x64,0x69,0x64,0x65,0x6f,0x67,
0x72,0x61,0x70,0x68,0x73,0xa2,0x47,0x18,0x65,0x78,0x74,0x65,0x6e,0x73,0x69,0x6f,
0x6e,6,0x64,0x6b,0x64,0xa3,0xd1,0x65,0xa5,0,0x66,0xa5,0x12,0x67,0xa5,0x2e,
0x14,0x6f,0x6d,0x70,0x61,0x74,0xa2,0x45,1,0x66,0x96,0x69,1,0x62,0x44,0x64,
0x17,0x65,0x6f,0x67,0x72,0x61,0x70,0x68,0x73,0xa2,0x4f,0x12,0x73,0x75,0x70,0xa3,
0x5f,0x14,0x69,0x6c,0x69,0x74,0x79,0xa2,0x45,1,0x66,0x54,0x69,0x18,0x64,0x65,
0x6f,0x67,0x72,0x61,0x70,0x68,0x73,0xa2,0x4f,0x19,0x73,0x75,0x70,0x70,0x6c,0x65,
0x6d,0x65,0x6e,0x74,0xa3,0x5f,0x13,0x6f,0x72,0x6d,0x73,0xa3,0x53,0x11,0x78,0x74,
6,0x64,0xc,0x64,0xa3,0xd1,0x65,0xa5,0,0x66,0xa5,0x12,0x67,0xa5,0x2e,0x61,
0xa3,0x46,0x62,0xa3,0x5e,0x63,0xa3,0xc5,0x19,0x61,0x64,0x69,0x63,0x61,0x6c,0x73,
0x73,0x75,0x70,0x94,0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0x95,1,0x74,0x50,
0x79,0x14,0x6d,0x62,0x6f,0x6c,0x73,0x9a,0x1d,0x61,0x6e,0x64,0x70,0x75,0x6e,0x63,
0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0x9b,0x14,0x72,0x6f,0x6b,0x65,0x73,0xa3,0x82,
2,0x6e,0x48,0x72,0x64,0x75,0x1d,0x63,0x61,0x73,0x69,0x61,0x6e,0x61,0x6c,0x62,
0x61,0x6e,0x69,0x61,0x6e,0xa3,0xde,0x1d,0x61,0x64,0x69,0x61,0x6e,0x73,0x79,0x6c,
0x6c,0x61,0x62,0x69,0x63,0x73,0x63,0x12,0x69,0x61,0x6e,0xa3,0xa8,2,0x61,0x3a,
0x65,0x4c,0x6f,0x16,0x72,0x61,0x73,0x6d,0x69,0x61,0x6e,0xa5,0x2d,1,0x6b,0x26,
0x6d,0xa3,0xa4,0x11,0x6d,0x61,0xa3,0xd4,1,0x72,0x38,0x73,0x17,0x73,0x73,0x79,
0x6d,0x62,0x6f,0x6c,0x73,0xa5,0x19,0x13,0x6f,0x6b,0x65,0x65,0x60,0x12,0x73,0x75,
0x70,0xa2,0xff,0x16,0x70,0x6c,0x65,0x6d,0x65,0x6e,0x74,0xa3,0xff,3,0x65,0x3e,
0x69,0x7e,0x6f,0xa2,0x69,0x75,0x15,0x70,0x6c,0x6f,0x79,0x61,0x6e,0xa3,0xe1,1,
0x73,0x50,0x76,0x16,0x61,0x6e,0x61,0x67,0x61,0x72,0x69,0x3e,0x12,0x65,0x78,0x74,
0xa2,0xb3,0x14,0x65,0x6e,0x64,0x65,0x64,0xa3,0xb3,0x13,0x65,0x72,0x65,0x74,0xa3,
0x5a,2,0x61,0x3a,0x6e,0x82,0x76,0x16,0x65,0x73,0x61,0x6b,0x75,0x72,0x75,0xa5,
0x2f,0x18,0x63,0x72,0x69,0x74,0x69,0x63,0x61,0x6c,0x73,0x2e,2,0x65,0x30,0x66,
0x36,0x73,0x11,0x75,0x70,0xa3,0x83,0x11,0x78,0x74,0xa3,0xe0,0x18,0x6f,0x72,0x73,
0x79,0x6d,0x62,0x6f,0x6c,0x73,0x77,0x14,0x67,0x62,0x61,0x74,0x73,0x91,1,0x67,
0x3e,0x6d,0x12,0x69,0x6e,0x6f,0xa2,0xab,0x14,0x74,0x69,0x6c,0x65,0x73,0xa3,0xab,
0x11,0x72,0x61,0xa5,0x1a,8,0x6d,0x5f,0x6d,0x3a,0x6e,0x48,0x73,0x7a,0x76,0xa2,
0x4b,0x77,0x12,0x69,0x64,0x65,0x43,0x11,0x65,0x64,0x32,0x12,0x69,0x61,0x6c,0x33,
2,0x61,0x40,0x62,0x37,0x6f,1,0x62,0x28,0x6e,0x10,0x65,0x21,0x13,0x72,0x65,
0x61,0x6b,0x37,0x10,0x72,0x34,0x12,0x72,0x6f,0x77,0x35,2,0x6d,0x38,0x71,0x46,
0x75,1,0x62,0x3d,0x70,0x3e,0x11,0x65,0x72,0x3f,1,0x61,0x24,0x6c,0x39,0x11,
0x6c,0x6c,0x39,1,0x72,0x3b,0x75,0x12,0x61,0x72,0x65,0x3b,0x12,0x65,0x72,0x74,
0x40,0x13,0x69,0x63,0x61,0x6c,0x41,0x63,0x58,0x65,0x92,0x66,0x96,0x69,1,0x6e,
0x36,0x73,0x10,0x6f,0x30,0x14,0x6c,0x61,0x74,0x65,0x64,0x31,0x11,0x69,0x74,0x2e,
0x12,0x69,0x61,0x6c,0x2f,2,0x61,0x36,0x69,0x48,0x6f,0x10,0x6d,0x24,0x12,0x70,
0x61,0x74,0x25,0x10,0x6e,0x22,0x15,0x6f,0x6e,0x69,0x63,0x61,0x6c,0x23,0x13,0x72,
0x63,0x6c,0x65,0x27,0x11,0x6e,0x63,0x27,2,0x69,0x3a,0x6f,0x44,0x72,0x10,0x61,
0x2c,0x14,0x63,0x74,0x69,0x6f,0x6e,0x2d,0x10,0x6e,0x28,0x11,0x61,0x6c,0x29,0x11,
0x6e,0x74,0x2b,4,0x61,0x3a,0x66,0x4c,0x68,0x5e,0x6e,0x70,0x77,0x2a,0x12,0x69,
0x64,0x65,0x2b,0x22,0x17,0x6d,0x62,0x69,0x67,0x75,0x6f,0x75,0x73,0x23,0x26,0x17,
0x75,0x6c,0x6c,0x77,0x69,0x64,0x74,0x68,0x27,0x24,0x17,0x61,0x6c,0x66,0x77,0x69,
0x64,0x74,0x68,0x25,0x20,1,0x61,0x30,0x65,0x14,0x75,0x74,0x72,0x61,0x6c,0x21,
0x28,0x13,0x72,0x72,0x6f,0x77,0x29,0xd,0x6e,0xc0,0xfb,0x73,0x6d,0x73,0x3a,0x74,
0x98,0x75,0xa2,0x49,0x7a,2,0x6c,0x3b,0x70,0x3d,0x73,0x39,5,0x6f,0x28,0x6f,
0x57,0x70,0x34,0x75,0x16,0x72,0x72,0x6f,0x67,0x61,0x74,0x65,0x45,0x11,0x61,0x63,
1,0x65,0x32,0x69,0x15,0x6e,0x67,0x6d,0x61,0x72,0x6b,0x31,0x18,0x73,0x65,0x70,
0x61,0x72,0x61,0x74,0x6f,0x72,0x39,0x63,0x53,0x6b,0x55,0x6d,0x51,0x1d,0x69,0x74,
0x6c,0x65,0x63,0x61,0x73,0x65,0x6c,0x65,0x74,0x74,0x65,0x72,0x27,1,0x6e,0x40,
0x70,0x1c,0x70,0x65,0x72,0x63,0x61,0x73,0x65,0x6c,0x65,0x74,0x74,0x65,0x72,0x23,
0x17,0x61,0x73,0x73,0x69,0x67,0x6e,0x65,0x64,0x21,0x6e,0x8a,0x6f,0xa2,0x47,0x70,
8,0x66,0x14,0x66,0x5b,0x69,0x59,0x6f,0x4f,0x72,0x24,0x73,0x49,0x17,0x69,0x76,
0x61,0x74,0x65,0x75,0x73,0x65,0x43,0x61,0x2c,0x63,0x4d,0x64,0x47,0x65,0x4b,0x1f,
0x72,0x61,0x67,0x72,0x61,0x70,0x68,0x73,0x65,0x70,0x61,0x72,0x61,0x74,0x6f,0x72,
0x3d,2,0x64,0x33,0x6c,0x35,0x6f,0x36,0x1b,0x6e,0x73,0x70,0x61,0x63,0x69,0x6e,
0x67,0x6d,0x61,0x72,0x6b,0x2d,1,0x70,0x7c,0x74,0x12,0x68,0x65,0x72,3,0x6c,
0x38,0x6e,0x42,0x70,0x4c,0x73,0x14,0x79,0x6d,0x62,0x6f,0x6c,0x57,0x14,0x65,0x74,
0x74,0x65,0x72,0x2b,0x14,0x75,0x6d,0x62,0x65,0x72,0x37,0x19,0x75,0x6e,0x63,0x74,
0x75,0x61,0x74,0x69,0x6f,0x6e,0x4f,0x1c,0x65,0x6e,0x70,0x75,0x6e,0x63,0x74,0x75,
0x61,0x74,0x69,0x6f,0x6e,0x49,0x66,0x9e,0x66,0x88,0x69,0xa2,0x4b,0x6c,0xa2,0x5c,
0x6d,4,0x61,0x60,0x63,0x31,0x65,0x2f,0x6e,0x2d,0x6f,0x15,0x64,0x69,0x66,0x69,
0x65,0x72,1,0x6c,0x30,0x73,0x14,0x79,0x6d,0x62,0x6f,0x6c,0x55,0x14,0x65,0x74,
0x74,0x65,0x72,0x29,0x17,0x74,0x68,0x73,0x79,0x6d,0x62,0x6f,0x6c,0x51,1,0x69,
0x2e,0x6f,0x13,0x72,0x6d,0x61,0x74,0x41,0x1d,0x6e,0x61,0x6c,0x70,0x75,0x6e,0x63,
0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0x5b,0x10,0x6e,0x1f,0x69,0x74,0x69,0x61,0x6c,
0x70,0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0x59,6,0x6d,0x18,0x6d,
0x29,0x6f,0x28,0x74,0x27,0x75,0x23,0x2a,0x1c,0x77,0x65,0x72,0x63,0x61,0x73,0x65,
0x6c,0x65,0x74,0x74,0x65,0x72,0x25,0x65,0x28,0x69,0x3c,0x6c,0x25,0x19,0x74,0x74,
0x65,0x72,0x6e,0x75,0x6d,0x62,0x65,0x72,0x35,0x1a,0x6e,0x65,0x73,0x65,0x70,0x61,
0x72,0x61,0x74,0x6f,0x72,0x3b,0x63,0x44,0x64,0xa2,0x60,0x65,0x1b,0x6e,0x63,0x6c,
0x6f,0x73,0x69,0x6e,0x67,0x6d,0x61,0x72,0x6b,0x2f,6,0x6e,0x39,0x6e,0x46,0x6f,
0x4e,0x73,0x45,0x75,0x1b,0x72,0x72,0x65,0x6e,0x63,0x79,0x73,0x79,0x6d,0x62,0x6f,
0x6c,0x53,0x20,0x12,0x74,0x72,0x6c,0x3f,0x42,0x10,0x6e,1,0x6e,0x2c,0x74,0x12,
0x72,0x6f,0x6c,0x3f,0x1f,0x65,0x63,0x74,0x6f,0x72,0x70,0x75,0x6e,0x63,0x74,0x75,
0x61,0x74,0x69,0x6f,0x6e,0x4d,0x63,0x3f,0x66,0x41,0x6c,0x1d,0x6f,0x73,0x65,0x70,
0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0x4b,2,0x61,0x30,0x65,0x4a,
0x69,0x12,0x67,0x69,0x74,0x33,0x1c,0x73,0x68,0x70,0x75,0x6e,0x63,0x74,0x75,0x61,
0x74,0x69,0x6f,0x6e,0x47,0x1a,0x63,0x69,0x6d,0x61,0x6c,0x6e,0x75,0x6d,0x62,0x65,
0x72,0x33,0,0x12,0x6d,0xc2,0x3f,0x73,0xa1,0x73,0x4e,0x74,0xa2,0x56,0x77,0xa2,
0x72,0x79,0xa2,0x73,0x7a,1,0x61,0x2c,0x68,0x12,0x61,0x69,0x6e,0x8b,0x11,0x69,
0x6e,0x85,5,0x74,0x22,0x74,0x38,0x77,0x4c,0x79,0x16,0x72,0x69,0x61,0x63,0x77,
0x61,0x77,0x6f,0x18,0x72,0x61,0x69,0x67,0x68,0x74,0x77,0x61,0x77,0xa3,0x55,0x15,
0x61,0x73,0x68,0x6b,0x61,0x66,0x6d,0x61,0x2e,0x65,0x38,0x68,0x11,0x69,0x6e,0x6b,
0x10,0x64,0x62,0x11,0x68,0x65,0x65,1,0x65,0x2e,0x6d,0x13,0x6b,0x61,0x74,0x68,
0x69,0x10,0x6e,0x67,1,0x61,0x4e,0x65,1,0x68,0x28,0x74,0x10,0x68,0x77,0x16,
0x6d,0x61,0x72,0x62,0x75,0x74,0x61,0x74,0x13,0x67,0x6f,0x61,0x6c,0x3d,1,0x68,
0x71,0x77,0x73,0x11,0x61,0x77,0x79,1,0x65,0x32,0x75,0x11,0x64,0x68,0x80,0x11,
0x68,0x65,0x83,0x10,0x68,0x7a,1,0x62,0x34,0x77,0x16,0x69,0x74,0x68,0x74,0x61,
0x69,0x6c,0x7f,0x14,0x61,0x72,0x72,0x65,0x65,0x7d,0x6d,0x6c,0x6e,0xa4,0x6b,0x70,
0xa4,0x88,0x71,0xa4,0x88,0x72,1,0x65,0x38,0x6f,0x18,0x68,0x69,0x6e,0x67,0x79,
0x61,0x79,0x65,0x68,0x93,1,0x68,0x5f,0x76,0x16,0x65,0x72,0x73,0x65,0x64,0x70,
0x65,0x61,2,0x61,0x2e,0x65,0xa4,0x3e,0x69,0x10,0x6d,0x53,1,0x6c,0xa2,0xe7,
0x6e,0x16,0x69,0x63,0x68,0x61,0x65,0x61,0x6e,0,0x12,0x6e,0x76,0x73,0x51,0x73,
0x3e,0x74,0x5c,0x77,0xa0,0x79,0xa2,0x42,0x7a,0x13,0x61,0x79,0x69,0x6e,0xa3,0x54,
0x10,0x61,1,0x64,0x2e,0x6d,0x12,0x65,0x6b,0x68,0xa3,0x4c,0x11,0x68,0x65,0xa3,
0x4b,3,0x61,0x38,0x65,0x3c,0x68,0x4a,0x77,0x13,0x65,0x6e,0x74,0x79,0xa3,0x51,
0x10,0x77,0xa3,0x4d,1,0x6e,0xa3,0x4e,0x74,0x10,0x68,0xa3,0x4f,0x14,0x61,0x6d,
0x65,0x64,0x68,0xa3,0x50,0x11,0x61,0x77,0xa3,0x52,0x12,0x6f,0x64,0x68,0xa3,0x53,
0x6e,0x3a,0x6f,0x40,0x70,0x46,0x71,0x4a,0x72,0x12,0x65,0x73,0x68,0xa3,0x4a,0x11,
0x75,0x6e,0xa3,0x46,0x11,0x6e,0x65,0xa3,0x47,0x10,0x65,0xa3,0x48,0x12,0x6f,0x70,
0x68,0xa3,0x49,0x67,0x33,0x67,0x38,0x68,0x40,0x6b,0x5e,0x6c,0x66,0x6d,0x11,0x65,
0x6d,0xa3,0x45,0x13,0x69,0x6d,0x65,0x6c,0xa1,1,0x65,0x32,0x75,0x14,0x6e,0x64,
0x72,0x65,0x64,0xa3,0x42,0x11,0x74,0x68,0xa3,0x41,0x12,0x61,0x70,0x68,0xa3,0x43,
0x14,0x61,0x6d,0x65,0x64,0x68,0xa3,0x44,0x61,0x34,0x62,0x4a,0x64,0x50,0x66,0x12,
0x69,0x76,0x65,0x9f,1,0x6c,0x2a,0x79,0x11,0x69,0x6e,0x97,0x12,0x65,0x70,0x68,
0x95,0x12,0x65,0x74,0x68,0x99,1,0x61,0x30,0x68,0x14,0x61,0x6d,0x65,0x64,0x68,
0x9d,0x13,0x6c,0x65,0x74,0x68,0x9b,0x15,0x61,0x79,0x61,0x6c,0x61,0x6d,6,0x6e,
0x2c,0x6e,0x34,0x72,0x5e,0x73,0x62,0x74,0x11,0x74,0x61,0xa3,0x63,2,0x67,0x2e,
0x6e,0x32,0x79,0x10,0x61,0xa3,0x60,0x10,0x61,0xa3,0x5d,1,0x61,0xa3,0x5e,0x6e,
0x10,0x61,0xa3,0x5f,0x10,0x61,0xa3,0x61,0x11,0x73,0x61,0xa3,0x62,0x62,0x3c,0x6a,
0x42,0x6c,0x10,0x6c,1,0x61,0xa3,0x5b,0x6c,0x10,0x61,0xa3,0x5c,0x11,0x68,0x61,
0xa3,0x59,0x10,0x61,0xa3,0x5a,0x11,0x65,0x6d,0x51,2,0x6f,0x2c,0x75,0x50,0x79,
0x10,0x61,0x91,1,0x6a,0x28,0x6f,0x10,0x6e,0x55,0x1a,0x6f,0x69,0x6e,0x69,0x6e,
0x67,0x67,0x72,0x6f,0x75,0x70,0x21,0x10,0x6e,0x57,0x10,0x65,0x59,0x10,0x61,1,
0x66,0x5b,0x70,0x10,0x68,0x5d,0x66,0x9a,0x66,0x42,0x67,0x7a,0x68,0x8a,0x6b,0xa2,
0x75,0x6c,0x11,0x61,0x6d,0x4c,0x12,0x61,0x64,0x68,0x4f,2,0x61,0x3e,0x65,0x4a,
0x69,0x19,0x6e,0x61,0x6c,0x73,0x65,0x6d,0x6b,0x61,0x74,0x68,0x35,0x15,0x72,0x73,
0x69,0x79,0x65,0x68,0x8f,0x86,0x10,0x68,0x33,0x10,0x61,1,0x66,0x37,0x6d,0x11,
0x61,0x6c,0x39,1,0x61,0x40,0x65,0x3e,1,0x68,0x28,0x74,0x10,0x68,0x45,0x40,
0x13,0x67,0x6f,0x61,0x6c,0x43,2,0x68,0x3b,0x6d,0x5c,0x6e,0x1a,0x69,0x66,0x69,
0x72,0x6f,0x68,0x69,0x6e,0x67,0x79,0x61,1,0x6b,0x2a,0x70,0x10,0x61,0xa3,0x65,
0x15,0x69,0x6e,0x6e,0x61,0x79,0x61,0xa3,0x64,0x1a,0x7a,0x61,0x6f,0x6e,0x68,0x65,
0x68,0x67,0x6f,0x61,0x6c,0x3d,2,0x61,0x3a,0x68,0x44,0x6e,0x17,0x6f,0x74,0x74,
0x65,0x64,0x68,0x65,0x68,0x4b,1,0x66,0x47,0x70,0x10,0x68,0x49,0x12,0x61,0x70,
0x68,0x89,0x61,0x2e,0x62,0x8a,0x64,0xa2,0x51,0x65,0x31,2,0x66,0x3c,0x69,0x70,
0x6c,1,0x61,0x28,0x65,0x10,0x66,0x27,0x11,0x70,0x68,0x25,0x14,0x72,0x69,0x63,
0x61,0x6e,2,0x66,0x30,0x6e,0x36,0x71,0x11,0x61,0x66,0xa3,0x58,0x11,0x65,0x68,
0xa3,0x56,0x12,0x6f,0x6f,0x6e,0xa3,0x57,0x10,0x6e,0x23,1,0x65,0x4a,0x75,0x10,
0x72,0x1f,0x75,0x73,0x68,0x61,0x73,0x6b,0x69,0x79,0x65,0x68,0x62,0x61,0x72,0x72,
0x65,0x65,0x8d,1,0x68,0x29,0x74,0x10,0x68,0x2b,0x11,0x61,0x6c,0x2c,0x16,0x61,
0x74,0x68,0x72,0x69,0x73,0x68,0x2f,7,0x6e,0x2e,0x6e,0x2c,0x72,0x3e,0x74,0x56,
0x75,0x21,0x18,0x6f,0x6e,0x6a,0x6f,0x69,0x6e,0x69,0x6e,0x67,0x21,0x28,0x1a,0x69,
0x67,0x68,0x74,0x6a,0x6f,0x69,0x6e,0x69,0x6e,0x67,0x29,0x2a,0x19,0x72,0x61,0x6e,
0x73,0x70,0x61,0x72,0x65,0x6e,0x74,0x2b,0x63,0x23,0x64,0x40,0x6a,0x56,0x6c,0x26,
0x19,0x65,0x66,0x74,0x6a,0x6f,0x69,0x6e,0x69,0x6e,0x67,0x27,0x24,0x19,0x75,0x61,
0x6c,0x6a,0x6f,0x69,0x6e,0x69,0x6e,0x67,0x25,0x19,0x6f,0x69,0x6e,0x63,0x61,0x75,
0x73,0x69,0x6e,0x67,0x23,0,0x13,0x6e,0xc0,0xd0,0x73,0x49,0x73,0x48,0x75,0x78,
0x77,0x84,0x78,0x9c,0x7a,0x10,0x77,0x58,1,0x6a,0x75,0x73,0x13,0x70,0x61,0x63,
0x65,0x59,4,0x61,0x51,0x67,0x53,0x70,0x28,0x75,0x30,0x79,0x57,0x54,0x12,0x61,
0x63,0x65,0x55,0x16,0x72,0x72,0x6f,0x67,0x61,0x74,0x65,0x53,0x15,0x6e,0x6b,0x6e,
0x6f,0x77,0x6e,0x21,1,0x6a,0x5d,0x6f,0x17,0x72,0x64,0x6a,0x6f,0x69,0x6e,0x65,
0x72,0x5d,0x10,0x78,0x21,0x6e,0x60,0x6f,0xa2,0x41,0x70,0xa2,0x50,0x71,0xa2,0x6e,
0x72,1,0x65,0x24,0x69,0x6f,0x1e,0x67,0x69,0x6f,0x6e,0x61,0x6c,0x69,0x6e,0x64,
0x69,0x63,0x61,0x74,0x6f,0x72,0x6f,4,0x65,0x3e,0x6c,0x5b,0x6f,0x46,0x73,0x45,
0x75,0x46,0x14,0x6d,0x65,0x72,0x69,0x63,0x47,0x15,0x78,0x74,0x6c,0x69,0x6e,0x65,
0x5b,0x17,0x6e,0x73,0x74,0x61,0x72,0x74,0x65,0x72,0x45,0x10,0x70,0x48,0x1c,0x65,
0x6e,0x70,0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0x49,1,0x6f,0x3e,
0x72,0x4c,0x1a,0x65,0x66,0x69,0x78,0x6e,0x75,0x6d,0x65,0x72,0x69,0x63,0x4d,0x4a,
0x1b,0x73,0x74,0x66,0x69,0x78,0x6e,0x75,0x6d,0x65,0x72,0x69,0x63,0x4b,0x10,0x75,
0x4e,0x16,0x6f,0x74,0x61,0x74,0x69,0x6f,0x6e,0x4f,0x68,0x7b,0x68,0x50,0x69,0x86,
0x6a,0xa2,0x61,0x6c,0xa2,0x65,0x6d,0x1c,0x61,0x6e,0x64,0x61,0x74,0x6f,0x72,0x79,
0x62,0x72,0x65,0x61,0x6b,0x2d,4,0x32,0x5f,0x33,0x61,0x65,0x34,0x6c,0x6d,0x79,
0x3a,0x13,0x70,0x68,0x65,0x6e,0x3b,0x19,0x62,0x72,0x65,0x77,0x6c,0x65,0x74,0x74,
0x65,0x72,0x6d,2,0x64,0x28,0x6e,0x3c,0x73,0x41,0x3c,0x18,0x65,0x6f,0x67,0x72,
0x61,0x70,0x68,0x69,0x63,0x3d,0x3e,1,0x66,0x3e,0x73,0x11,0x65,0x70,1,0x61,
0x22,0x65,0x14,0x72,0x61,0x62,0x6c,0x65,0x3f,0x18,0x69,0x78,0x6e,0x75,0x6d,0x65,
0x72,0x69,0x63,0x41,2,0x6c,0x63,0x74,0x65,0x76,0x67,1,0x66,0x43,0x69,0x15,
0x6e,0x65,0x66,0x65,0x65,0x64,0x43,0x61,0x40,0x62,0x70,0x63,0xa2,0x55,0x65,0xa2,
0xdb,0x67,0x10,0x6c,0x38,0x11,0x75,0x65,0x39,2,0x69,0x23,0x6c,0x34,0x6d,0x16,
0x62,0x69,0x67,0x75,0x6f,0x75,0x73,0x23,0x24,0x17,0x70,0x68,0x61,0x62,0x65,0x74,
0x69,0x63,0x25,4,0x32,0x27,0x61,0x29,0x62,0x2b,0x6b,0x2d,0x72,0x12,0x65,0x61,
0x6b,2,0x61,0x36,0x62,0x3e,0x73,0x15,0x79,0x6d,0x62,0x6f,0x6c,0x73,0x57,0x13,
0x66,0x74,0x65,0x72,0x29,1,0x65,0x2a,0x6f,0x11,0x74,0x68,0x27,0x13,0x66,0x6f,
0x72,0x65,0x2b,7,0x6d,0x51,0x6d,0x33,0x6f,0x28,0x70,0x69,0x72,0x35,1,0x6d,
0x76,0x6e,1,0x64,0x3c,0x74,0x1a,0x69,0x6e,0x67,0x65,0x6e,0x74,0x62,0x72,0x65,
0x61,0x6b,0x2f,0x15,0x69,0x74,0x69,0x6f,0x6e,0x61,0x1f,0x6c,0x6a,0x61,0x70,0x61,
0x6e,0x65,0x73,0x65,0x73,0x74,0x61,0x72,0x74,0x65,0x72,0x6b,1,0x62,0x3a,0x70,
0x19,0x6c,0x65,0x78,0x63,0x6f,0x6e,0x74,0x65,0x78,0x74,0x51,0x18,0x69,0x6e,0x69,
0x6e,0x67,0x6d,0x61,0x72,0x6b,0x33,0x61,0x6a,0x62,0x2f,0x6a,0x6b,0x6c,0x30,0x13,
0x6f,0x73,0x65,0x70,1,0x61,0x38,0x75,0x18,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,
0x6f,0x6e,0x31,0x18,0x72,0x65,0x6e,0x74,0x68,0x65,0x73,0x69,0x73,0x69,0x1b,0x72,
0x72,0x69,0x61,0x67,0x65,0x72,0x65,0x74,0x75,0x72,0x6e,0x35,2,0x62,0x3e,0x6d,
0x46,0x78,0x36,0x18,0x63,0x6c,0x61,0x6d,0x61,0x74,0x69,0x6f,0x6e,0x37,0x70,0x12,
0x61,0x73,0x65,0x71,0x72,0x16,0x6f,0x64,0x69,0x66,0x69,0x65,0x72,0x73,1,0x64,
0x42,0x6e,1,0x6f,0x32,0x75,0x26,0x14,0x6d,0x65,0x72,0x69,0x63,0x27,0x11,0x6e,
0x65,0x21,1,0x65,0x2e,0x69,0x24,0x12,0x67,0x69,0x74,0x25,0x22,0x14,0x63,0x69,
0x6d,0x61,0x6c,0x23,0,0x18,0x6e,0xc4,0x2a,0x74,0xc1,0x6d,0x77,0x96,0x77,0xa2,
0x4c,0x78,0xa2,0x70,0x79,0xa2,0x7a,0x7a,6,0x73,0x1e,0x73,0x34,0x78,0x42,0x79,
0x48,0x7a,0x11,0x7a,0x7a,0xa3,0x67,0x10,0x79,1,0x65,0xa3,0xae,0x6d,0xa3,0x81,
0x11,0x78,0x78,0xa3,0x66,0x11,0x79,0x79,0x21,0x61,0x30,0x69,0x58,0x6d,0x11,0x74,
0x68,0xa3,0x80,0x10,0x6e,1,0x61,0x26,0x62,0xa3,0xb1,0x1a,0x62,0x61,0x7a,0x61,
0x72,0x73,0x71,0x75,0x61,0x72,0x65,0xa3,0xb1,0x11,0x6e,0x68,0x23,2,0x61,0x30,
0x63,0x5a,0x6f,0x11,0x6c,0x65,0xa3,0x9b,1,0x6e,0x3c,0x72,0x10,0x61,0xa2,0x92,
0x15,0x6e,0x67,0x63,0x69,0x74,0x69,0xa3,0x92,0x12,0x63,0x68,0x6f,0xa3,0xbc,0x11,
0x68,0x6f,0xa3,0xbc,1,0x70,0x2c,0x73,0x11,0x75,0x78,0xa3,0x65,0x11,0x65,0x6f,
0x9b,1,0x65,0x2c,0x69,0x72,0x11,0x69,0x69,0x73,0x11,0x7a,0x69,0xa2,0xc0,0x11,
0x64,0x69,0xa3,0xc0,0x74,0x4a,0x75,0xa2,0xba,0x76,1,0x61,0x2c,0x69,0x11,0x73,
0x70,0xa3,0x64,0x10,0x69,0xa2,0x63,0x10,0x69,0xa3,0x63,5,0x67,0x36,0x67,0x68,
0x68,0x6c,0x69,2,0x62,0x3a,0x66,0x4a,0x72,0x10,0x68,0xa2,0x9e,0x12,0x75,0x74,
0x61,0xa3,0x9e,1,0x65,0x24,0x74,0x6f,0x12,0x74,0x61,0x6e,0x6f,0x14,0x69,0x6e,
0x61,0x67,0x68,0x99,0x11,0x6c,0x67,0x75,0x10,0x61,1,0x61,0x24,0x69,0x6d,0x6a,
0x11,0x6e,0x61,0x6b,0x61,0x30,0x65,0xa2,0x5b,0x66,0x11,0x6e,0x67,0x99,6,0x6c,
0x21,0x6c,0x32,0x6d,0x38,0x6e,0x44,0x76,0x10,0x74,0xa3,0x7f,1,0x65,0x89,0x75,
0x97,1,0x69,0x24,0x6c,0x67,0x10,0x6c,0x67,0x10,0x67,0xa2,0x9a,0x11,0x75,0x74,
0xa3,0x9a,0x67,0x36,0x69,0x52,0x6b,0x10,0x72,0xa2,0x99,0x10,0x69,0xa3,0x99,1,
0x61,0x30,0x62,0x7a,0x13,0x61,0x6e,0x77,0x61,0x7b,0x12,0x6c,0x6f,0x67,0x75,2,
0x6c,0x32,0x74,0x34,0x76,0x12,0x69,0x65,0x74,0xa3,0x7f,0x10,0x65,0x89,0x12,0x68,
0x61,0x6d,0xa3,0x6a,1,0x6c,0x2a,0x6e,0x10,0x67,0xa3,0x62,0x10,0x75,0x68,0x11,
0x67,0x75,0x69,1,0x67,0x32,0x6e,0x14,0x6b,0x6e,0x6f,0x77,0x6e,0xa3,0x67,0x11,
0x61,0x72,0x8a,0x13,0x69,0x74,0x69,0x63,0x8b,0x71,0xc1,0x13,0x71,0xa2,0xde,0x72,
0xa2,0xe3,0x73,6,0x69,0x8a,0x69,0x72,0x6f,0xa2,0x4c,0x75,0xa2,0x75,0x79,1,
0x6c,0x46,0x72,4,0x63,0x65,0x65,0xa3,0x5f,0x69,0x2c,0x6a,0xa3,0x60,0x6e,0xa3,
0x61,0x11,0x61,0x63,0x65,0x10,0x6f,0x94,0x16,0x74,0x69,0x6e,0x61,0x67,0x72,0x69,
0x95,2,0x64,0x3c,0x67,0x4c,0x6e,1,0x64,0xa3,0x91,0x68,0x62,0x12,0x61,0x6c,
0x61,0x63,0x10,0x64,0xa2,0xa6,0x12,0x68,0x61,0x6d,0xa3,0xa6,0x17,0x6e,0x77,0x72,
0x69,0x74,0x69,0x6e,0x67,0xa3,0x70,2,0x67,0x3a,0x72,0x52,0x79,0x10,0x6f,0xa2,
0xb0,0x12,0x6d,0x62,0x6f,0xa3,0xb0,1,0x64,0x26,0x6f,0xa3,0xb8,0xa2,0xb7,0x12,
0x69,0x61,0x6e,0xa3,0xb7,0x10,0x61,0xa2,0x98,0x16,0x73,0x6f,0x6d,0x70,0x65,0x6e,
0x67,0xa3,0x98,0x11,0x6e,0x64,0xa2,0x71,0x14,0x61,0x6e,0x65,0x73,0x65,0xa3,0x71,
0x61,0x5c,0x67,0xa2,0x43,0x68,1,0x61,0x2a,0x72,0x10,0x64,0xa3,0x97,2,0x72,
0x28,0x76,0x30,0x77,0x87,0x12,0x61,0x64,0x61,0xa3,0x97,0x12,0x69,0x61,0x6e,0x87,
2,0x6d,0x40,0x72,0x58,0x75,0x10,0x72,0xa2,0x6f,0x15,0x61,0x73,0x68,0x74,0x72,
0x61,0xa3,0x6f,1,0x61,0x26,0x72,0xa3,0x7e,0x14,0x72,0x69,0x74,0x61,0x6e,0xa3,
0x7e,1,0x61,0xa3,0x5e,0x62,0xa3,0x85,0x11,0x6e,0x77,0xa3,0x70,0x11,0x61,0x61,
1,0x63,0x2f,0x69,0x23,3,0x65,0x3e,0x6a,0x48,0x6f,0x4e,0x75,0x10,0x6e,1,
0x69,0x24,0x72,0x61,0x10,0x63,0x61,0x13,0x6a,0x61,0x6e,0x67,0xa3,0x6e,0x11,0x6e,
0x67,0xa3,0x6e,1,0x68,0x2a,0x72,0x10,0x6f,0xa3,0x5d,0x10,0x67,0xa3,0xb6,0x6e,
0xa2,0x83,0x6f,0xa2,0xf2,0x70,5,0x6c,0x1e,0x6c,0x44,0x72,0x4a,0x73,0x1b,0x61,
0x6c,0x74,0x65,0x72,0x70,0x61,0x68,0x6c,0x61,0x76,0x69,0xa3,0x7b,0x11,0x72,0x64,
0xa3,0x5c,0x11,0x74,0x69,0xa3,0x7d,0x61,0x7c,0x65,0xa2,0x54,0x68,3,0x61,0x3e,
0x6c,0x4e,0x6e,0x5e,0x6f,0x16,0x65,0x6e,0x69,0x63,0x69,0x61,0x6e,0xa3,0x5b,0x10,
0x67,0xa2,0x5a,0x12,0x73,0x70,0x61,0xa3,0x5a,2,0x69,0xa3,0x7a,0x70,0xa3,0x7b,
0x76,0xa3,0x7c,0x10,0x78,0xa3,0x5b,2,0x68,0x3e,0x6c,0x50,0x75,0x10,0x63,0xa2,
0xa5,0x14,0x69,0x6e,0x68,0x61,0x75,0xa3,0xa5,0x17,0x61,0x77,0x68,0x68,0x6d,0x6f,
0x6e,0x67,0xa3,0x4b,0x10,0x6d,0xa2,0x90,0x14,0x79,0x72,0x65,0x6e,0x65,0xa3,0x90,
0x11,0x72,0x6d,0xa3,0x59,6,0x6b,0x36,0x6b,0x56,0x73,0x6e,0x75,0x74,0x79,0x11,
0x69,0x61,0x1f,0x6b,0x65,0x6e,0x67,0x70,0x75,0x61,0x63,0x68,0x75,0x65,0x68,0x6d,
0x6f,0x6e,0x67,0xa3,0xba,1,0x67,0x2e,0x6f,0xa2,0x57,0x10,0x6f,0xa3,0x57,0x10,
0x62,0xa3,0x84,0x11,0x68,0x75,0xa3,0x96,0x12,0x73,0x68,0x75,0xa3,0x96,0x61,0x42,
0x62,0x80,0x65,0x10,0x77,1,0x61,0xa3,0xaa,0x74,0x14,0x61,0x69,0x6c,0x75,0x65,
0x97,2,0x62,0x2e,0x6e,0x3c,0x72,0x10,0x62,0xa3,0x8e,0x15,0x61,0x74,0x61,0x65,
0x61,0x6e,0xa3,0x8f,0x10,0x64,0xa2,0xbb,0x16,0x69,0x6e,0x61,0x67,0x61,0x72,0x69,
0xa3,0xbb,0x11,0x61,0x74,0xa3,0x8f,3,0x67,0x5a,0x6c,0x6c,0x72,0xa2,0x93,0x73,
2,0x61,0x36,0x67,0x3c,0x6d,0x10,0x61,0x84,0x12,0x6e,0x79,0x61,0x85,0x11,0x67,
0x65,0xa3,0xab,0x10,0x65,0xa3,0xab,1,0x61,0x2a,0x68,0x11,0x61,0x6d,0x5b,0x10,
0x6d,0x5b,1,0x63,0xa2,0x60,0x64,5,0x70,0x37,0x70,0x36,0x73,0x54,0x74,0x14,
0x75,0x72,0x6b,0x69,0x63,0xa3,0x58,0x11,0x65,0x72,1,0x6d,0x2c,0x73,0x12,0x69,
0x61,0x6e,0x9b,0x11,0x69,0x63,0xa3,0x59,0x10,0x6f,1,0x67,0x3a,0x75,0x18,0x74,
0x68,0x61,0x72,0x61,0x62,0x69,0x61,0x6e,0xa3,0x85,0x13,0x64,0x69,0x61,0x6e,0xa3,
0xb8,0x68,0x42,0x69,0x54,0x6e,0x1a,0x6f,0x72,0x74,0x68,0x61,0x72,0x61,0x62,0x69,
0x61,0x6e,0xa3,0x8e,0x17,0x75,0x6e,0x67,0x61,0x72,0x69,0x61,0x6e,0xa3,0x4c,0x14,
0x74,0x61,0x6c,0x69,0x63,0x5d,1,0x68,0x26,0x6b,0xa3,0x6d,0x12,0x69,0x6b,0x69,
0xa3,0x6d,2,0x69,0x2c,0x6b,0x30,0x79,0x10,0x61,0x5f,0x11,0x79,0x61,0x5f,0x10,
0x68,0xa3,0x58,0x68,0xc3,0xd,0x6b,0xc2,0x24,0x6b,0xa4,0x17,0x6c,0xa4,0xb2,0x6d,
8,0x6f,0x46,0x6f,0x48,0x72,0x74,0x74,0x80,0x75,0x86,0x79,1,0x61,0x28,0x6d,
0x10,0x72,0x59,0x13,0x6e,0x6d,0x61,0x72,0x59,2,0x64,0x2e,0x6e,0x32,0x6f,0x10,
0x6e,0xa3,0x72,0x10,0x69,0xa3,0xa3,0x10,0x67,0x56,0x14,0x6f,0x6c,0x69,0x61,0x6e,
0x57,0x10,0x6f,0xa2,0x95,0x10,0x6f,0xa3,0x95,0x11,0x65,0x69,0xa3,0x73,0x11,0x6c,
0x74,0xa2,0xa4,0x12,0x61,0x6e,0x69,0xa3,0xa4,0x61,0x36,0x65,0xa2,0x67,0x69,0xa2,
0xbd,0x6c,0x11,0x79,0x6d,0x55,6,0x6e,0x38,0x6e,0x32,0x72,0x5c,0x73,0x6c,0x79,
0x10,0x61,0xa3,0x55,1,0x64,0x38,0x69,0xa2,0x79,0x15,0x63,0x68,0x61,0x65,0x61,
0x6e,0xa3,0x79,0xa2,0x54,0x12,0x61,0x69,0x63,0xa3,0x54,0x10,0x63,0xa2,0xa9,0x12,
0x68,0x65,0x6e,0xa3,0xa9,0x18,0x61,0x72,0x61,0x6d,0x67,0x6f,0x6e,0x64,0x69,0xa3,
0xaf,0x68,0x36,0x6b,0x4c,0x6c,0x15,0x61,0x79,0x61,0x6c,0x61,0x6d,0x55,1,0x61,
0x26,0x6a,0xa3,0xa0,0x13,0x6a,0x61,0x6e,0x69,0xa3,0xa0,0x10,0x61,0xa2,0xb4,0x12,
0x73,0x61,0x72,0xa3,0xb4,3,0x64,0x78,0x65,0x94,0x6e,0xa2,0x42,0x72,1,0x63,
0xa3,0x8d,0x6f,0xa2,0x56,0x13,0x69,0x74,0x69,0x63,1,0x63,0x3c,0x68,0x19,0x69,
0x65,0x72,0x6f,0x67,0x6c,0x79,0x70,0x68,0x73,0xa3,0x56,0x15,0x75,0x72,0x73,0x69,
0x76,0x65,0xa3,0x8d,1,0x65,0x26,0x66,0xa3,0xb5,0x16,0x66,0x61,0x69,0x64,0x72,
0x69,0x6e,0xa3,0xb5,0x17,0x74,0x65,0x69,0x6d,0x61,0x79,0x65,0x6b,0xa3,0x73,0x10,
0x64,0xa2,0x8c,0x17,0x65,0x6b,0x69,0x6b,0x61,0x6b,0x75,0x69,0xa3,0x8c,0x11,0x61,
0x6f,0xa3,0x5c,6,0x6e,0x1a,0x6e,0x34,0x6f,0x38,0x70,0x3e,0x74,0x11,0x68,0x69,
0xa3,0x78,0x11,0x64,0x61,0x4b,0x11,0x72,0x65,0xa3,0x77,0x11,0x65,0x6c,0xa3,0x8a,
0x61,0x30,0x68,0x9a,0x69,0x11,0x74,0x73,0xa3,0xbf,4,0x69,0x3c,0x6c,0x44,0x6e,
0x48,0x74,0x56,0x79,0x13,0x61,0x68,0x6c,0x69,0xa3,0x4f,0x12,0x74,0x68,0x69,0xa3,
0x78,0x10,0x69,0xa3,0x4f,1,0x61,0x4d,0x6e,0x12,0x61,0x64,0x61,0x4b,0x14,0x61,
0x6b,0x61,0x6e,0x61,0x4c,0x19,0x6f,0x72,0x68,0x69,0x72,0x61,0x67,0x61,0x6e,0x61,
0x8d,4,0x61,0x40,0x69,0x52,0x6d,0x70,0x6f,0x7c,0x75,0x15,0x64,0x61,0x77,0x61,
0x64,0x69,0xa3,0x91,0x10,0x72,0x92,0x15,0x6f,0x73,0x68,0x74,0x68,0x69,0x93,0x1d,
0x74,0x61,0x6e,0x73,0x6d,0x61,0x6c,0x6c,0x73,0x63,0x72,0x69,0x70,0x74,0xa3,0xbf,
1,0x65,0x24,0x72,0x4f,0x10,0x72,0x4f,0x10,0x6a,0xa2,0x9d,0x11,0x6b,0x69,0xa3,
0x9d,4,0x61,0x5c,0x65,0x90,0x69,0xa0,0x6f,0xa2,0x5d,0x79,1,0x63,0x34,0x64,
0x10,0x69,0xa2,0x6c,0x11,0x61,0x6e,0xa3,0x6c,0x10,0x69,0xa2,0x6b,0x11,0x61,0x6e,
0xa3,0x6b,2,0x6e,0x42,0x6f,0x46,0x74,3,0x66,0xa3,0x50,0x67,0xa3,0x51,0x69,
0x24,0x6e,0x53,0x10,0x6e,0x53,0x10,0x61,0xa3,0x6a,0x50,0x10,0x6f,0x51,0x11,0x70,
0x63,0xa2,0x52,0x11,0x68,0x61,0xa3,0x52,2,0x6d,0x2e,0x6e,0x36,0x73,0x10,0x75,
0xa3,0x83,0x10,0x62,0x80,0x10,0x75,0x81,2,0x61,0xa3,0x53,0x62,0x83,0x65,0x11,
0x61,0x72,1,0x61,0xa3,0x53,0x62,0x83,0x11,0x6d,0x61,0xa3,0x8b,0x68,0x6e,0x69,
0xa2,0x95,0x6a,2,0x61,0x30,0x70,0x52,0x75,0x11,0x72,0x63,0xa3,0x94,1,0x6d,
0x38,0x76,0x10,0x61,0xa2,0x4e,0x13,0x6e,0x65,0x73,0x65,0xa3,0x4e,0x10,0x6f,0xa3,
0xad,0x11,0x61,0x6e,0xa3,0x69,6,0x6c,0x1e,0x6c,0x34,0x6d,0x3a,0x72,0x48,0x75,
0x11,0x6e,0x67,0xa3,0x4c,0x11,0x75,0x77,0xa3,0x9c,0x10,0x6e,1,0x67,0xa3,0x4b,
0x70,0xa3,0xba,0x11,0x6b,0x74,0x8d,0x61,0x3c,0x65,0xa2,0x43,0x69,0x11,0x72,0x61,
0x48,0x13,0x67,0x61,0x6e,0x61,0x49,1,0x6e,0x34,0x74,0x10,0x72,0xa2,0xa2,0x11,
0x61,0x6e,0xa3,0xa2,0x42,6,0x6f,0xe,0x6f,0x77,0x73,0xa3,0x49,0x74,0xa3,0x4a,
0x75,0x12,0x6e,0x6f,0x6f,0x77,0x62,0xa3,0xac,0x67,0x3e,0x69,0x42,0x19,0x66,0x69,
0x72,0x6f,0x68,0x69,0x6e,0x67,0x79,0x61,0xa3,0xb6,0x44,0x11,0x75,0x6c,0x45,0x11,
0x62,0x72,0x46,0x11,0x65,0x77,0x47,2,0x6d,0x2e,0x6e,0x4a,0x74,0x11,0x61,0x6c,
0x5d,0x1c,0x70,0x65,0x72,0x69,0x61,0x6c,0x61,0x72,0x61,0x6d,0x61,0x69,0x63,0xa3,
0x74,2,0x64,0x66,0x68,0x6a,0x73,0x1b,0x63,0x72,0x69,0x70,0x74,0x69,0x6f,0x6e,
0x61,0x6c,0x70,0x61,1,0x68,0x32,0x72,0x14,0x74,0x68,0x69,0x61,0x6e,0xa3,0x7d,
0x13,0x6c,0x61,0x76,0x69,0xa3,0x7a,0x10,0x73,0xa3,0x4d,0x15,0x65,0x72,0x69,0x74,
0x65,0x64,0x23,0x64,0xc1,0xd,0x64,0xa2,0x7a,0x65,0xa2,0xc1,0x67,4,0x65,0x82,
0x6c,0x9a,0x6f,0xa2,0x46,0x72,0xa2,0x55,0x75,2,0x6a,0x3c,0x6e,0x4e,0x72,1,
0x6d,0x24,0x75,0x41,0x13,0x75,0x6b,0x68,0x69,0x41,1,0x61,0x24,0x72,0x3f,0x13,
0x72,0x61,0x74,0x69,0x3f,0x18,0x6a,0x61,0x6c,0x61,0x67,0x6f,0x6e,0x64,0x69,0xa3,
0xb3,0x10,0x6f,1,0x6b,0xa3,0x48,0x72,0x38,0x13,0x67,0x69,0x61,0x6e,0x39,0x11,
0x61,0x67,0x90,0x15,0x6f,0x6c,0x69,0x74,0x69,0x63,0x91,1,0x6e,0x30,0x74,0x10,
0x68,0x3a,0x11,0x69,0x63,0x3b,1,0x67,0xa3,0xb3,0x6d,0xa3,0xaf,1,0x61,0x32,
0x65,1,0x65,0x24,0x6b,0x3d,0x10,0x6b,0x3d,0x10,0x6e,0xa2,0x89,0x12,0x74,0x68,
0x61,0xa3,0x89,4,0x65,0x46,0x69,0x6c,0x6f,0x8c,0x73,0x9a,0x75,0x11,0x70,0x6c,
0xa2,0x87,0x13,0x6f,0x79,0x61,0x6e,0xa3,0x87,1,0x73,0x38,0x76,0x10,0x61,0x34,
0x15,0x6e,0x61,0x67,0x61,0x72,0x69,0x35,0x13,0x65,0x72,0x65,0x74,0x33,1,0x61,
0x36,0x76,0x16,0x65,0x73,0x61,0x6b,0x75,0x72,0x75,0xa3,0xbe,0x10,0x6b,0xa3,0xbe,
0x11,0x67,0x72,0xa2,0xb2,0x10,0x61,0xa3,0xb2,0x11,0x72,0x74,0x33,2,0x67,0x3a,
0x6c,0x72,0x74,0x11,0x68,0x69,0x36,0x13,0x6f,0x70,0x69,0x63,0x37,0x10,0x79,2,
0x64,0xa3,0x45,0x68,0xa3,0x46,0x70,0xa2,0x47,0x1e,0x74,0x69,0x61,0x6e,0x68,0x69,
0x65,0x72,0x6f,0x67,0x6c,0x79,0x70,0x68,0x73,0xa3,0x47,1,0x62,0x36,0x79,0x10,
0x6d,0xa2,0xb9,0x12,0x61,0x69,0x63,0xa3,0xb9,0x10,0x61,0xa2,0x88,0x12,0x73,0x61,
0x6e,0xa3,0x88,0x61,0xa2,0xb4,0x62,0xa4,0x19,0x63,6,0x6f,0x3d,0x6f,0x5a,0x70,
0x76,0x75,0x7a,0x79,1,0x70,0x3e,0x72,2,0x69,0x2a,0x6c,0x31,0x73,0xa3,0x44,
0x13,0x6c,0x6c,0x69,0x63,0x31,0x13,0x72,0x69,0x6f,0x74,0x7f,1,0x6d,0x30,0x70,
0x10,0x74,0x2e,0x11,0x69,0x63,0x2f,0x12,0x6d,0x6f,0x6e,0x21,0x11,0x72,0x74,0x7f,
0x16,0x6e,0x65,0x69,0x66,0x6f,0x72,0x6d,0xa3,0x65,0x61,0x32,0x68,0xa2,0x41,0x69,
0x11,0x72,0x74,0xa3,0x43,3,0x6b,0x4c,0x6e,0x50,0x72,0x76,0x75,0x1d,0x63,0x61,
0x73,0x69,0x61,0x6e,0x61,0x6c,0x62,0x61,0x6e,0x69,0x61,0x6e,0xa3,0x9f,0x10,0x6d,
0xa3,0x76,1,0x61,0x24,0x73,0x71,0x1d,0x64,0x69,0x61,0x6e,0x61,0x62,0x6f,0x72,
0x69,0x67,0x69,0x6e,0x61,0x6c,0x71,0x10,0x69,0xa2,0x68,0x11,0x61,0x6e,0xa3,0x68,
3,0x61,0x32,0x65,0x44,0x6f,0x52,0x72,0x10,0x73,0xa3,0xbd,1,0x6b,0x26,0x6d,
0xa3,0x42,0x11,0x6d,0x61,0xa3,0x76,0x10,0x72,0x2c,0x13,0x6f,0x6b,0x65,0x65,0x2d,
0x16,0x72,0x61,0x73,0x6d,0x69,0x61,0x6e,0xa3,0xbd,6,0x68,0x4a,0x68,0x48,0x6e,
0x4e,0x72,0x76,0x76,1,0x65,0x2a,0x73,0x10,0x74,0xa3,0x75,0x13,0x73,0x74,0x61,
0x6e,0xa3,0x75,0x11,0x6f,0x6d,0xa3,0xa1,0x11,0x61,0x74,0x1f,0x6f,0x6c,0x69,0x61,
0x6e,0x68,0x69,0x65,0x72,0x6f,0x67,0x6c,0x79,0x70,0x68,0x73,0xa3,0x9c,1,0x61,
0x3e,0x6d,2,0x65,0x2a,0x69,0xa3,0x74,0x6e,0x27,0x13,0x6e,0x69,0x61,0x6e,0x27,
0x10,0x62,0x24,0x11,0x69,0x63,0x25,0x64,0x30,0x66,0x44,0x67,0x11,0x68,0x62,0xa3,
0x9f,0x10,0x6c,1,0x61,0x26,0x6d,0xa3,0xa7,0x10,0x6d,0xa3,0xa7,0x11,0x61,0x6b,
0xa3,0x93,6,0x6c,0x3c,0x6c,0x52,0x6f,0x56,0x72,0x66,0x75,1,0x67,0x30,0x68,
1,0x64,0x79,0x69,0x10,0x64,0x79,0x10,0x69,0x8e,0x13,0x6e,0x65,0x73,0x65,0x8f,
0x11,0x69,0x73,0xa1,0x11,0x70,0x6f,0x2a,0x13,0x6d,0x6f,0x66,0x6f,0x2b,0x10,0x61,
1,0x68,0x2e,0x69,0x7c,0x12,0x6c,0x6c,0x65,0x7d,0xa2,0x41,0x11,0x6d,0x69,0xa3,
0x41,0x61,0x48,0x65,0x9c,0x68,1,0x61,0x2a,0x6b,0x10,0x73,0xa3,0xa8,0x15,0x69,
0x6b,0x73,0x75,0x6b,0x69,0xa3,0xa8,3,0x6c,0x3a,0x6d,0x48,0x73,0x54,0x74,1,
0x61,0x24,0x6b,0x9f,0x10,0x6b,0x9f,0x10,0x69,0x9c,0x13,0x6e,0x65,0x73,0x65,0x9d,
0x10,0x75,0xa2,0x82,0x10,0x6d,0xa3,0x82,0x10,0x73,0xa2,0x86,0x13,0x61,0x76,0x61,
0x68,0xa3,0x86,0x11,0x6e,0x67,0x28,0x12,0x61,0x6c,0x69,0x29,3,0x6c,0x42,0x6e,
0x90,0x74,0xa2,0x46,0x76,0x24,0x17,0x6f,0x77,0x65,0x6c,0x6a,0x61,0x6d,0x6f,0x25,
0x22,1,0x65,0x54,0x76,0x28,1,0x73,0x38,0x74,0x2a,0x17,0x73,0x79,0x6c,0x6c,
0x61,0x62,0x6c,0x65,0x2b,0x16,0x79,0x6c,0x6c,0x61,0x62,0x6c,0x65,0x29,0x18,0x61,
0x64,0x69,0x6e,0x67,0x6a,0x61,0x6d,0x6f,0x23,1,0x61,0x21,0x6f,0x1a,0x74,0x61,
0x70,0x70,0x6c,0x69,0x63,0x61,0x62,0x6c,0x65,0x21,0x26,0x1a,0x72,0x61,0x69,0x6c,
0x69,0x6e,0x67,0x6a,0x61,0x6d,0x6f,0x27,1,0x6e,0x2c,0x79,0x22,0x11,0x65,0x73,
0x23,0x20,0x10,0x6f,0x21,1,0x6e,0x2c,0x79,0x22,0x11,0x65,0x73,0x23,0x20,0x10,
0x6f,0x21,2,0x6d,0x30,0x6e,0x3a,0x79,0x22,0x11,0x65,0x73,0x23,0x24,0x13,0x61,
0x79,0x62,0x65,0x25,0x20,0x10,0x6f,0x21,2,0x6d,0x30,0x6e,0x3a,0x79,0x22,0x11,
0x65,0x73,0x23,0x24,0x13,0x61,0x79,0x62,0x65,0x25,0x20,0x10,0x6f,0x21,0xb,0x72,
0x39,0x76,0xc,0x76,0x33,0x78,0x2a,0x7a,0x11,0x77,0x6a,0x43,0x10,0x78,0x21,0x72,
0x28,0x73,0x50,0x74,0x31,1,0x65,0x24,0x69,0x39,0x1e,0x67,0x69,0x6f,0x6e,0x61,
0x6c,0x69,0x6e,0x64,0x69,0x63,0x61,0x74,0x6f,0x72,0x39,1,0x6d,0x35,0x70,0x18,
0x61,0x63,0x69,0x6e,0x67,0x6d,0x61,0x72,0x6b,0x35,0x6c,0x1f,0x6c,0x3c,0x6f,0x4a,
0x70,1,0x70,0x37,0x72,0x14,0x65,0x70,0x65,0x6e,0x64,0x37,0x28,1,0x66,0x2b,
0x76,0x2c,0x10,0x74,0x2f,0x13,0x74,0x68,0x65,0x72,0x21,0x63,0x4c,0x65,0x64,0x67,
1,0x61,0x3a,0x6c,0x19,0x75,0x65,0x61,0x66,0x74,0x65,0x72,0x7a,0x77,0x6a,0x41,
0x10,0x7a,0x41,2,0x6e,0x23,0x6f,0x24,0x72,0x25,0x14,0x6e,0x74,0x72,0x6f,0x6c,
0x23,2,0x62,0x34,0x6d,0x4e,0x78,0x26,0x13,0x74,0x65,0x6e,0x64,0x27,0x3a,1,
0x61,0x24,0x67,0x3d,0x11,0x73,0x65,0x3a,0x12,0x67,0x61,0x7a,0x3d,0x3e,0x16,0x6f,
0x64,0x69,0x66,0x69,0x65,0x72,0x3f,9,0x6e,0x4a,0x6e,0x34,0x6f,0x44,0x73,0x60,
0x75,0x94,0x78,0x10,0x78,0x21,0x10,0x75,0x2a,0x14,0x6d,0x65,0x72,0x69,0x63,0x2b,
1,0x6c,0x2c,0x74,0x12,0x68,0x65,0x72,0x21,0x14,0x65,0x74,0x74,0x65,0x72,0x2d,
3,0x63,0x36,0x65,0x46,0x70,0x31,0x74,0x32,0x12,0x65,0x72,0x6d,0x33,0x3c,0x16,
0x6f,0x6e,0x74,0x69,0x6e,0x75,0x65,0x3d,0x2e,0x10,0x70,0x2f,0x10,0x70,0x34,0x12,
0x70,0x65,0x72,0x35,0x61,0x46,0x63,0x52,0x65,0x64,0x66,0x72,0x6c,2,0x65,0x2d,
0x66,0x3b,0x6f,0x28,0x12,0x77,0x65,0x72,0x29,0x10,0x74,0x22,0x12,0x65,0x72,0x6d,
0x23,1,0x6c,0x24,0x72,0x37,0x24,0x12,0x6f,0x73,0x65,0x25,0x10,0x78,0x38,0x13,
0x74,0x65,0x6e,0x64,0x39,0x10,0x6f,0x26,0x13,0x72,0x6d,0x61,0x74,0x27,0,0x10,
0x6c,0x88,0x72,0x40,0x72,0x36,0x73,0x5e,0x77,0x7a,0x78,0x8a,0x7a,0x11,0x77,0x6a,
0x4b,1,0x65,0x24,0x69,0x3b,0x1e,0x67,0x69,0x6f,0x6e,0x61,0x6c,0x69,0x6e,0x64,
0x69,0x63,0x61,0x74,0x6f,0x72,0x3b,1,0x69,0x24,0x71,0x3f,0x18,0x6e,0x67,0x6c,
0x65,0x71,0x75,0x6f,0x74,0x65,0x3f,0x17,0x73,0x65,0x67,0x73,0x70,0x61,0x63,0x65,
0x4d,0x10,0x78,0x21,0x6c,0x36,0x6d,0x3c,0x6e,0x76,0x6f,0x13,0x74,0x68,0x65,0x72,
0x21,1,0x65,0x23,0x66,0x35,3,0x62,0x37,0x69,0x28,0x6c,0x29,0x6e,0x2b,0x10,
0x64,1,0x6c,0x34,0x6e,0x11,0x75,0x6d,0x2a,0x12,0x6c,0x65,0x74,0x37,0x14,0x65,
0x74,0x74,0x65,0x72,0x29,2,0x65,0x36,0x6c,0x39,0x75,0x2c,0x14,0x6d,0x65,0x72,
0x69,0x63,0x2d,0x14,0x77,0x6c,0x69,0x6e,0x65,0x39,0x66,0x3f,0x66,0x40,0x67,0x4e,
0x68,0x70,0x6b,0x10,0x61,0x26,0x15,0x74,0x61,0x6b,0x61,0x6e,0x61,0x27,0x10,0x6f,
0x24,0x13,0x72,0x6d,0x61,0x74,0x25,1,0x61,0x3a,0x6c,0x19,0x75,0x65,0x61,0x66,
0x74,0x65,0x72,0x7a,0x77,0x6a,0x49,0x10,0x7a,0x49,1,0x65,0x24,0x6c,0x3d,0x19,
0x62,0x72,0x65,0x77,0x6c,0x65,0x74,0x74,0x65,0x72,0x3d,0x61,0x86,0x63,0x92,0x64,
0x94,0x65,2,0x62,0x44,0x6d,0x5e,0x78,0x2e,0x13,0x74,0x65,0x6e,0x64,0x32,0x15,
0x6e,0x75,0x6d,0x6c,0x65,0x74,0x2f,0x42,1,0x61,0x24,0x67,0x45,0x11,0x73,0x65,
0x42,0x12,0x67,0x61,0x7a,0x45,0x46,0x16,0x6f,0x64,0x69,0x66,0x69,0x65,0x72,0x47,
0x15,0x6c,0x65,0x74,0x74,0x65,0x72,0x23,0x10,0x72,0x31,1,0x6f,0x24,0x71,0x41,
0x18,0x75,0x62,0x6c,0x65,0x71,0x75,0x6f,0x74,0x65,0x41,2,0x63,0x32,0x6e,0x3c,
0x6f,0x22,0x12,0x70,0x65,0x6e,0x23,0x24,0x13,0x6c,0x6f,0x73,0x65,0x25,0x20,0x12,
0x6f,0x6e,0x65,0x21,6,0x6f,0x65,0x6f,0x4a,0x72,0x5c,0x74,0x64,0x76,0x1d,0x69,
0x73,0x75,0x61,0x6c,0x6f,0x72,0x64,0x65,0x72,0x6c,0x65,0x66,0x74,0x3d,0x18,0x76,
0x65,0x72,0x73,0x74,0x72,0x75,0x63,0x6b,0x2d,0x13,0x69,0x67,0x68,0x74,0x2f,0x11,
0x6f,0x70,0x30,0x12,0x61,0x6e,0x64,2,0x62,0x32,0x6c,0x62,0x72,0x13,0x69,0x67,
0x68,0x74,0x3b,0x14,0x6f,0x74,0x74,0x6f,0x6d,0x32,0x12,0x61,0x6e,0x64,1,0x6c,
0x2e,0x72,0x13,0x69,0x67,0x68,0x74,0x35,0x12,0x65,0x66,0x74,0x3f,0x12,0x65,0x66,
0x74,0x36,0x17,0x61,0x6e,0x64,0x72,0x69,0x67,0x68,0x74,0x39,0x62,0x2c,0x6c,0x5c,
0x6e,0x10,0x61,0x21,0x14,0x6f,0x74,0x74,0x6f,0x6d,0x22,0x12,0x61,0x6e,0x64,1,
0x6c,0x2e,0x72,0x13,0x69,0x67,0x68,0x74,0x27,0x12,0x65,0x66,0x74,0x25,0x12,0x65,
0x66,0x74,0x28,0x17,0x61,0x6e,0x64,0x72,0x69,0x67,0x68,0x74,0x2b,0xd,0x6e,0xaa,
0x72,0x70,0x72,0x92,0x73,0xa2,0x46,0x74,0xa2,0x54,0x76,1,0x69,0x60,0x6f,0x12,
0x77,0x65,0x6c,0x62,1,0x64,0x3a,0x69,0x19,0x6e,0x64,0x65,0x70,0x65,0x6e,0x64,
0x65,0x6e,0x74,0x67,0x17,0x65,0x70,0x65,0x6e,0x64,0x65,0x6e,0x74,0x65,1,0x72,
0x2e,0x73,0x13,0x61,0x72,0x67,0x61,0x61,0x12,0x61,0x6d,0x61,0x5f,0x1d,0x65,0x67,
0x69,0x73,0x74,0x65,0x72,0x73,0x68,0x69,0x66,0x74,0x65,0x72,0x57,0x1e,0x79,0x6c,
0x6c,0x61,0x62,0x6c,0x65,0x6d,0x6f,0x64,0x69,0x66,0x69,0x65,0x72,0x59,0x12,0x6f,
0x6e,0x65,1,0x6c,0x2c,0x6d,0x12,0x61,0x72,0x6b,0x5d,0x14,0x65,0x74,0x74,0x65,
0x72,0x5b,0x6e,0x3c,0x6f,0x7c,0x70,0x18,0x75,0x72,0x65,0x6b,0x69,0x6c,0x6c,0x65,
0x72,0x55,1,0x6f,0x4c,0x75,1,0x6b,0x3c,0x6d,0x12,0x62,0x65,0x72,0x50,0x15,
0x6a,0x6f,0x69,0x6e,0x65,0x72,0x53,0x11,0x74,0x61,0x4f,0x16,0x6e,0x6a,0x6f,0x69,
0x6e,0x65,0x72,0x4d,0x13,0x74,0x68,0x65,0x72,0x21,0x67,0x3e,0x67,0x4a,0x69,0x64,
0x6a,0x82,0x6d,0x1d,0x6f,0x64,0x69,0x66,0x79,0x69,0x6e,0x67,0x6c,0x65,0x74,0x74,
0x65,0x72,0x4b,0x1c,0x65,0x6d,0x69,0x6e,0x61,0x74,0x69,0x6f,0x6e,0x6d,0x61,0x72,
0x6b,0x45,0x1e,0x6e,0x76,0x69,0x73,0x69,0x62,0x6c,0x65,0x73,0x74,0x61,0x63,0x6b,
0x65,0x72,0x47,0x14,0x6f,0x69,0x6e,0x65,0x72,0x49,0x61,0xa2,0xba,0x62,0xa2,0xc0,
0x63,1,0x61,0xa2,0xa2,0x6f,0x16,0x6e,0x73,0x6f,0x6e,0x61,0x6e,0x74,0x2a,8,
0x6b,0x67,0x6b,0x48,0x6d,0x52,0x70,0x5c,0x73,0xa2,0x42,0x77,0x19,0x69,0x74,0x68,
0x73,0x74,0x61,0x63,0x6b,0x65,0x72,0x43,0x14,0x69,0x6c,0x6c,0x65,0x72,0x35,0x14,
0x65,0x64,0x69,0x61,0x6c,0x37,1,0x6c,0x52,0x72,0x10,0x65,1,0x63,0x2e,0x66,
0x13,0x69,0x78,0x65,0x64,0x3d,0x19,0x65,0x64,0x69,0x6e,0x67,0x72,0x65,0x70,0x68,
0x61,0x3b,0x18,0x61,0x63,0x65,0x68,0x6f,0x6c,0x64,0x65,0x72,0x39,0x10,0x75,1,
0x62,0x3e,0x63,0x1b,0x63,0x65,0x65,0x64,0x69,0x6e,0x67,0x72,0x65,0x70,0x68,0x61,
0x41,0x15,0x6a,0x6f,0x69,0x6e,0x65,0x64,0x3f,0x64,0x4c,0x66,0x52,0x68,0x5a,0x69,
0x1e,0x6e,0x69,0x74,0x69,0x61,0x6c,0x70,0x6f,0x73,0x74,0x66,0x69,0x78,0x65,0x64,
0x33,0x12,0x65,0x61,0x64,0x2d,0x13,0x69,0x6e,0x61,0x6c,0x2f,0x18,0x65,0x61,0x64,
0x6c,0x65,0x74,0x74,0x65,0x72,0x31,0x1d,0x6e,0x74,0x69,0x6c,0x6c,0x61,0x74,0x69,
0x6f,0x6e,0x6d,0x61,0x72,0x6b,0x29,0x16,0x76,0x61,0x67,0x72,0x61,0x68,0x61,0x23,
1,0x69,0x4a,0x72,0x10,0x61,0x1f,0x68,0x6d,0x69,0x6a,0x6f,0x69,0x6e,0x69,0x6e,
0x67,0x6e,0x75,0x6d,0x62,0x65,0x72,0x27,0x12,0x6e,0x64,0x75,0x25,2,0x72,0x38,
0x74,0x46,0x75,0x26,0x15,0x70,0x72,0x69,0x67,0x68,0x74,0x27,0x20,0x15,0x6f,0x74,
0x61,0x74,0x65,0x64,0x21,1,0x72,0x24,0x75,0x25,0x22,0x18,0x61,0x6e,0x73,0x66,
0x6f,0x72,0x6d,0x65,0x64,1,0x72,0x32,0x75,0x15,0x70,0x72,0x69,0x67,0x68,0x74,
0x25,0x15,0x6f,0x74,0x61,0x74,0x65,0x64,0x23,0xd,0x6e,0xc1,0x86,0x73,0xa8,0x73,
0x4c,0x74,0xa2,0x76,0x75,0xa2,0x83,0x7a,0xd8,0x70,0,2,0x6c,0xd9,0x20,0,
0x70,0xd9,0x40,0,0x73,0xc3,0,0xfe,0xf,0,0,0,7,0x6f,0x3c,0x6f,
0xff,8,0,0,0,0x70,0x3a,0x75,0x6e,0x79,0x13,0x6d,0x62,0x6f,0x6c,0xff,
0xf,0,0,0,0x11,0x61,0x63,1,0x65,0x34,0x69,0x15,0x6e,0x67,0x6d,0x61,
0x72,0x6b,0xa5,0,0x18,0x73,0x65,0x70,0x61,0x72,0x61,0x74,0x6f,0x72,0xc3,0,
0x16,0x72,0x72,0x6f,0x67,0x61,0x74,0x65,0xe1,0,0,0x63,0xff,2,0,0,
0,0x65,0x38,0x6b,0xff,4,0,0,0,0x6d,0xff,1,0,0,0,0x16,
0x70,0x61,0x72,0x61,0x74,0x6f,0x72,0xd9,0x70,0,0x1d,0x69,0x74,0x6c,0x65,0x63,
0x61,0x73,0x65,0x6c,0x65,0x74,0x74,0x65,0x72,0x31,1,0x6e,0x40,0x70,0x1c,0x70,
0x65,0x72,0x63,0x61,0x73,0x65,0x6c,0x65,0x74,0x74,0x65,0x72,0x25,0x17,0x61,0x73,
0x73,0x69,0x67,0x6e,0x65,0x64,0x23,0x6e,0xa2,0x69,0x6f,0xa2,0x89,0x70,0xfe,0x30,
0xf8,0,0,9,0x69,0x33,0x69,0xff,0x10,0,0,0,0x6f,0xfd,0x80,0,
0,0x72,0x54,0x73,0xf9,0,0,0x75,0x12,0x6e,0x63,0x74,0xfe,0x30,0xf8,0,
0,0x15,0x75,0x61,0x74,0x69,0x6f,0x6e,0xff,0x30,0xf8,0,0,0x17,0x69,0x76,
0x61,0x74,0x65,0x75,0x73,0x65,0xdd,0,0,0x61,0x48,0x63,0xfd,0x40,0,0,
0x64,0xe9,0,0,0x65,0xfd,0x20,0,0,0x66,0xff,0x20,0,0,0,0x1f,
0x72,0x61,0x67,0x72,0x61,0x70,0x68,0x73,0x65,0x70,0x61,0x72,0x61,0x74,0x6f,0x72,
0xd9,0x40,0,0xbe,0,3,0x64,0xa7,0,0x6c,0xab,0,0x6f,0x30,0x75,0x13,
0x6d,0x62,0x65,0x72,0xbf,0,0xb2,0,0x1b,0x6e,0x73,0x70,0x61,0x63,0x69,0x6e,
0x67,0x6d,0x61,0x72,0x6b,0xa1,1,0x70,0x92,0x74,0x12,0x68,0x65,0x72,0xe6,0x80,
1,3,0x6c,0x40,0x6e,0x4a,0x70,0x56,0x73,0x14,0x79,0x6d,0x62,0x6f,0x6c,0xff,
8,0,0,0,0x14,0x65,0x74,0x74,0x65,0x72,0x61,0x14,0x75,0x6d,0x62,0x65,
0x72,0xb3,0,0x19,0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0xfd,0x80,
0,0,0x1c,0x65,0x6e,0x70,0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,
0xf9,0,0,0x66,0xc0,0xc4,0x66,0xa2,0x47,0x69,0xa2,0x64,0x6c,0xa2,0x79,0x6d,
0xa4,0xc0,4,0x61,0x6c,0x63,0xa5,0,0x65,0xa3,0x80,0x6e,0xa1,0x6f,0x15,0x64,
0x69,0x66,0x69,0x65,0x72,1,0x6c,0x38,0x73,0x14,0x79,0x6d,0x62,0x6f,0x6c,0xff,
4,0,0,0,0x14,0x65,0x74,0x74,0x65,0x72,0x41,1,0x72,0x3c,0x74,0x16,
0x68,0x73,0x79,0x6d,0x62,0x6f,0x6c,0xff,1,0,0,0,0x10,0x6b,0xa5,0xc0,
1,0x69,0x32,0x6f,0x13,0x72,0x6d,0x61,0x74,0xdb,0,0,0x1d,0x6e,0x61,0x6c,
0x70,0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0xff,0x20,0,0,0,
0x10,0x6e,0x1f,0x69,0x74,0x69,0x61,0x6c,0x70,0x75,0x6e,0x63,0x74,0x75,0x61,0x74,
0x69,0x6f,0x6e,0xff,0x10,0,0,0,0x9c,7,0x6d,0x18,0x6d,0x41,0x6f,0x28,
0x74,0x31,0x75,0x25,0x60,0x1c,0x77,0x65,0x72,0x63,0x61,0x73,0x65,0x6c,0x65,0x74,
0x74,0x65,0x72,0x29,0x63,0x3d,0x65,0x28,0x69,0x42,0x6c,0x29,0x13,0x74,0x74,0x65,
0x72,0x9c,0x15,0x6e,0x75,0x6d,0x62,0x65,0x72,0xab,0,0x1a,0x6e,0x65,0x73,0x65,
0x70,0x61,0x72,0x61,0x74,0x6f,0x72,0xd9,0x20,0,0x63,0x46,0x64,0xa2,0x96,0x65,
0x1b,0x6e,0x63,0x6c,0x6f,0x73,0x69,0x6e,0x67,0x6d,0x61,0x72,0x6b,0xa3,0x80,0xe6,
0x80,1,7,0x6e,0x57,0x6e,0x52,0x6f,0x5e,0x73,0xe1,0,0,0x75,0x1b,0x72,
0x72,0x65,0x6e,0x63,0x79,0x73,0x79,0x6d,0x62,0x6f,0x6c,0xff,2,0,0,0,
0x22,0x12,0x74,0x72,0x6c,0xd9,0x80,0,0xdc,0,0,1,0x6d,0x62,0x6e,1,
0x6e,0x30,0x74,0x12,0x72,0x6f,0x6c,0xd9,0x80,0,0x1f,0x65,0x63,0x74,0x6f,0x72,
0x70,0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0xfd,0x40,0,0,0x19,
0x62,0x69,0x6e,0x69,0x6e,0x67,0x6d,0x61,0x72,0x6b,0xa5,0xc0,0x61,0x58,0x63,0xd9,
0x80,0,0x66,0xdb,0,0,0x6c,0x1d,0x6f,0x73,0x65,0x70,0x75,0x6e,0x63,0x74,
0x75,0x61,0x74,0x69,0x6f,0x6e,0xfd,0x20,0,0,0x18,0x73,0x65,0x64,0x6c,0x65,
0x74,0x74,0x65,0x72,0x3d,2,0x61,0x32,0x65,0x50,0x69,0x12,0x67,0x69,0x74,0xa7,
0,0x1c,0x73,0x68,0x70,0x75,0x6e,0x63,0x74,0x75,0x61,0x74,0x69,0x6f,0x6e,0xe9,
0,0,0x1a,0x63,0x69,0x6d,0x61,0x6c,0x6e,0x75,0x6d,0x62,0x65,0x72,0xa7,0
};

const char PropNameData::nameGroups[22098]={
2,'A','l','p','h','a',0,'A','l','p','h','a','b','e','t','i','c',0,
4,'N',0,'N','o',0,'F',0,'F','a','l','s','e',0,4,'Y',0,'Y','e','s',0,'T',0,'T','r','u','e',0,
2,'N','R',0,'N','o','t','_','R','e','o','r','d','e','r','e','d',0,
2,'O','V',0,'O','v','e','r','l','a','y',0,2,'H','A','N','R',0,'H','a','n','_','R','e','a','d','i','n','g',0,
2,'N','K',0,'N','u','k','t','a',0,2,'K','V',0,'K','a','n','a','_','V','o','i','c','i','n','g',0,
2,'V','R',0,'V','i','r','a','m','a',0,2,'C','C','C','1','0',0,'C','C','C','1','0',0,
2,'C','C','C','1','1',0,'C','C','C','1','1',0,2,'C','C','C','1','2',0,'C','C','C','1','2',0,
2,'C','C','C','1','3',0,'C','C','C','1','3',0,2,'C','C','C','1','4',0,'C','C','C','1','4',0,
2,'C','C','C','1','5',0,'C','C','C','1','5',0,2,'C','C','C','1','6',0,'C','C','C','1','6',0,
2,'C','C','C','1','7',0,'C','C','C','1','7',0,2,'C','C','C','1','8',0,'C','C','C','1','8',0,
2,'C','C','C','1','9',0,'C','C','C','1','9',0,2,'C','C','C','2','0',0,'C','C','C','2','0',0,
2,'C','C','C','2','1',0,'C','C','C','2','1',0,2,'C','C','C','2','2',0,'C','C','C','2','2',0,
2,'C','C','C','2','3',0,'C','C','C','2','3',0,2,'C','C','C','2','4',0,'C','C','C','2','4',0,
2,'C','C','C','2','5',0,'C','C','C','2','5',0,2,'C','C','C','2','6',0,'C','C','C','2','6',0,
2,'C','C','C','2','7',0,'C','C','C','2','7',0,2,'C','C','C','2','8',0,'C','C','C','2','8',0,
2,'C','C','C','2','9',0,'C','C','C','2','9',0,2,'C','C','C','3','0',0,'C','C','C','3','0',0,
2,'C','C','C','3','1',0,'C','C','C','3','1',0,2,'C','C','C','3','2',0,'C','C','C','3','2',0,
2,'C','C','C','3','3',0,'C','C','C','3','3',0,2,'C','C','C','3','4',0,'C','C','C','3','4',0,
2,'C','C','C','3','5',0,'C','C','C','3','5',0,2,'C','C','C','3','6',0,'C','C','C','3','6',0,
2,'C','C','C','8','4',0,'C','C','C','8','4',0,2,'C','C','C','9','1',0,'C','C','C','9','1',0,
2,'C','C','C','1','0','3',0,'C','C','C','1','0','3',0,2,'C','C','C','1','0','7',0,'C','C','C','1','0','7',0,
2,'C','C','C','1','1','8',0,'C','C','C','1','1','8',0,2,'C','C','C','1','2','2',0,'C','C','C','1','2','2',0,
2,'C','C','C','1','2','9',0,'C','C','C','1','2','9',0,2,'C','C','C','1','3','0',0,'C','C','C','1','3','0',0,
2,'C','C','C','1','3','2',0,'C','C','C','1','3','2',0,2,'C','C','C','1','3','3',0,'C','C','C','1','3','3',0,
2,'A','T','B','L',0,'A','t','t','a','c','h','e','d','_','B','e','l','o','w','_','L','e','f','t',0,
2,'A','T','B',0,'A','t','t','a','c','h','e','d','_','B','e','l','o','w',0,
2,'A','T','A',0,'A','t','t','a','c','h','e','d','_','A','b','o','v','e',0,
2,'A','T','A','R',0,'A','t','t','a','c','h','e','d','_','A','b','o','v','e','_','R','i','g','h','t',0,
2,'B','L',0,'B','e','l','o','w','_','L','e','f','t',0,2,'B',0,'B','e','l','o','w',0,
2,'B','R',0,'B','e','l','o','w','_','R','i','g','h','t',0,
2,'L',0,'L','e','f','t',0,2,'R',0,'R','i','g','h','t',0,
2,'A','L',0,'A','b','o','v','e','_','L','e','f','t',0,2,'A',0,'A','b','o','v','e',0,
2,'A','R',0,'A','b','o','v','e','_','R','i','g','h','t',0,
2,'D','B',0,'D','o','u','b','l','e','_','B','e','l','o','w',0,
2,'D','A',0,'D','o','u','b','l','e','_','A','b','o','v','e',0,
2,'I','S',0,'I','o','t','a','_','S','u','b','s','c','r','i','p','t',0,
2,'A','H','e','x',0,'A','S','C','I','I','_','H','e','x','_','D','i','g','i','t',0,
2,'B','i','d','i','_','C',0,'B','i','d','i','_','C','o','n','t','r','o','l',0,
2,'B','i','d','i','_','M',0,'B','i','d','i','_','M','i','r','r','o','r','e','d',0,
2,'D','a','s','h',0,'D','a','s','h',0,2,'D','I',0,'D','e','f','a','u','l','t','_','I','g','n','o','r','a','b','l','e',
'_','C','o','d','e','_','P','o','i','n','t',0,2,'D','e','p',0,'D','e','p','r','e','c','a','t','e','d',0,
2,'D','i','a',0,'D','i','a','c','r','i','t','i','c',0,2,'E','x','t',0,'E','x','t','e','n','d','e','r',0,
2,'C','o','m','p','_','E','x',0,'F','u','l','l','_','C','o','m','p','o','s','i','t','i','o','n','_','E','x','c','l','u','s',
'i','o','n',0,2,'G','r','_','B','a','s','e',0,'G','r','a','p','h','e','m','e','_','B','a','s','e',0,
2,'G','r','_','E','x','t',0,'G','r','a','p','h','e','m','e','_','E','x','t','e','n','d',0,
2,'G','r','_','L','i','n','k',0,'G','r','a','p','h','e','m','e','_','L','i','n','k',0,
2,'H','e','x',0,'H','e','x','_','D','i','g','i','t',0,2,'H','y','p','h','e','n',0,'H','y','p','h','e','n',0,
2,'I','D','C',0,'I','D','_','C','o','n','t','i','n','u','e',0,
2,'I','D','S',0,'I','D','_','S','t','a','r','t',0,2,'I','d','e','o',0,'I','d','e','o','g','r','a','p','h','i','c',0,
2,'I','D','S','B',0,'I','D','S','_','B','i','n','a','r','y','_','O','p','e','r','a','t','o','r',0,
2,'I','D','S','T',0,'I','D','S','_','T','r','i','n','a','r','y','_','O','p','e','r','a','t','o','r',0,
2,'J','o','i','n','_','C',0,'J','o','i','n','_','C','o','n','t','r','o','l',0,
2,'L','O','E',0,'L','o','g','i','c','a','l','_','O','r','d','e','r','_','E','x','c','e','p','t','i','o','n',0,
2,'L','o','w','e','r',0,'L','o','w','e','r','c','a','s','e',0,
2,'M','a','t','h',0,'M','a','t','h',0,2,'N','C','h','a','r',0,'N','o','n','c','h','a','r','a','c','t','e','r','_','C',
'o','d','e','_','P','o','i','n','t',0,2,'Q','M','a','r','k',0,'Q','u','o','t','a','t','i','o','n','_','M','a','r','k',0,
2,'R','a','d','i','c','a','l',0,'R','a','d','i','c','a','l',0,
2,'S','D',0,'S','o','f','t','_','D','o','t','t','e','d',0,
2,'T','e','r','m',0,'T','e','r','m','i','n','a','l','_','P','u','n','c','t','u','a','t','i','o','n',0,
2,'U','I','d','e','o',0,'U','n','i','f','i','e','d','_','I','d','e','o','g','r','a','p','h',0,
2,'U','p','p','e','r',0,'U','p','p','e','r','c','a','s','e',0,
3,'W','S','p','a','c','e',0,'W','h','i','t','e','_','S','p','a','c','e',0,'s','p','a','c','e',0,
2,'X','I','D','C',0,'X','I','D','_','C','o','n','t','i','n','u','e',0,
2,'X','I','D','S',0,'X','I','D','_','S','t','a','r','t',0,
2,'S','e','n','s','i','t','i','v','e',0,'C','a','s','e','_','S','e','n','s','i','t','i','v','e',0,
2,'S','T','e','r','m',0,'S','e','n','t','e','n','c','e','_','T','e','r','m','i','n','a','l',0,
2,'V','S',0,'V','a','r','i','a','t','i','o','n','_','S','e','l','e','c','t','o','r',0,
2,'n','f','d','i','n','e','r','t',0,'N','F','D','_','I','n','e','r','t',0,
2,'n','f','k','d','i','n','e','r','t',0,'N','F','K','D','_','I','n','e','r','t',0,
2,'n','f','c','i','n','e','r','t',0,'N','F','C','_','I','n','e','r','t',0,
2,'n','f','k','c','i','n','e','r','t',0,'N','F','K','C','_','I','n','e','r','t',0,
2,'s','e','g','s','t','a','r','t',0,'S','e','g','m','e','n','t','_','S','t','a','r','t','e','r',0,
2,'P','a','t','_','S','y','n',0,'P','a','t','t','e','r','n','_','S','y','n','t','a','x',0,
2,'P','a','t','_','W','S',0,'P','a','t','t','e','r','n','_','W','h','i','t','e','_','S','p','a','c','e',0,
2,0,'a','l','n','u','m',0,2,0,'b','l','a','n','k',0,
2,0,'g','r','a','p','h',0,2,0,'p','r','i','n','t',0,
2,0,'x','d','i','g','i','t',0,2,'C','a','s','e','d',0,'C','a','s','e','d',0,
2,'C','I',0,'C','a','s','e','_','I','g','n','o','r','a','b','l','e',0,
2,'C','W','L',0,'C','h','a','n','g','e','s','_','W','h','e','n','_','L','o','w','e','r','c','a','s','e','d',0,
2,'C','W','U',0,'C','h','a','n','g','e','s','_','W','h','e','n','_','U','p','p','e','r','c','a','s','e','d',0,
2,'C','W','T',0,'C','h','a','n','g','e','s','_','W','h','e','n','_','T','i','t','l','e','c','a','s','e','d',0,
2,'C','W','C','F',0,'C','h','a','n','g','e','s','_','W','h','e','n','_','C','a','s','e','f','o','l','d','e','d',0,
2,'C','W','C','M',0,'C','h','a','n','g','e','s','_','W','h','e','n','_','C','a','s','e','m','a','p','p','e','d',0,
2,'C','W','K','C','F',0,'C','h','a','n','g','e','s','_','W','h','e','n','_','N','F','K','C','_','C','a','s','e','f','o','l',
'd','e','d',0,2,'E','m','o','j','i',0,'E','m','o','j','i',0,
2,'E','P','r','e','s',0,'E','m','o','j','i','_','P','r','e','s','e','n','t','a','t','i','o','n',0,
2,'E','M','o','d',0,'E','m','o','j','i','_','M','o','d','i','f','i','e','r',0,
2,'E','B','a','s','e',0,'E','m','o','j','i','_','M','o','d','i','f','i','e','r','_','B','a','s','e',0,
2,'E','C','o','m','p',0,'E','m','o','j','i','_','C','o','m','p','o','n','e','n','t',0,
2,'R','I',0,'R','e','g','i','o','n','a','l','_','I','n','d','i','c','a','t','o','r',0,
2,'P','C','M',0,'P','r','e','p','e','n','d','e','d','_','C','o','n','c','a','t','e','n','a','t','i','o','n','_','M','a','r',
'k',0,2,'E','x','t','P','i','c','t',0,'E','x','t','e','n','d','e','d','_','P','i','c','t','o','g','r','a','p','h','i','c',
0,2,'b','c',0,'B','i','d','i','_','C','l','a','s','s',0,
2,'L',0,'L','e','f','t','_','T','o','_','R','i','g','h','t',0,
2,'R',0,'R','i','g','h','t','_','T','o','_','L','e','f','t',0,
2,'E','N',0,'E','u','r','o','p','e','a','n','_','N','u','m','b','e','r',0,
2,'E','S',0,'E','u','r','o','p','e','a','n','_','S','e','p','a','r','a','t','o','r',0,
2,'E','T',0,'E','u','r','o','p','e','a','n','_','T','e','r','m','i','n','a','t','o','r',0,
2,'A','N',0,'A','r','a','b','i','c','_','N','u','m','b','e','r',0,
2,'C','S',0,'C','o','m','m','o','n','_','S','e','p','a','r','a','t','o','r',0,
2,'B',0,'P','a','r','a','g','r','a','p','h','_','S','e','p','a','r','a','t','o','r',0,
2,'S',0,'S','e','g','m','e','n','t','_','S','e','p','a','r','a','t','o','r',0,
2,'W','S',0,'W','h','i','t','e','_','S','p','a','c','e',0,
2,'O','N',0,'O','t','h','e','r','_','N','e','u','t','r','a','l',0,
2,'L','R','E',0,'L','e','f','t','_','T','o','_','R','i','g','h','t','_','E','m','b','e','d','d','i','n','g',0,
2,'L','R','O',0,'L','e','f','t','_','T','o','_','R','i','g','h','t','_','O','v','e','r','r','i','d','e',0,
2,'A','L',0,'A','r','a','b','i','c','_','L','e','t','t','e','r',0,
2,'R','L','E',0,'R','i','g','h','t','_','T','o','_','L','e','f','t','_','E','m','b','e','d','d','i','n','g',0,
2,'R','L','O',0,'R','i','g','h','t','_','T','o','_','L','e','f','t','_','O','v','e','r','r','i','d','e',0,
2,'P','D','F',0,'P','o','p','_','D','i','r','e','c','t','i','o','n','a','l','_','F','o','r','m','a','t',0,
2,'N','S','M',0,'N','o','n','s','p','a','c','i','n','g','_','M','a','r','k',0,
2,'B','N',0,'B','o','u','n','d','a','r','y','_','N','e','u','t','r','a','l',0,
2,'F','S','I',0,'F','i','r','s','t','_','S','t','r','o','n','g','_','I','s','o','l','a','t','e',0,
2,'L','R','I',0,'L','e','f','t','_','T','o','_','R','i','g','h','t','_','I','s','o','l','a','t','e',0,
2,'R','L','I',0,'R','i','g','h','t','_','T','o','_','L','e','f','t','_','I','s','o','l','a','t','e',0,
2,'P','D','I',0,'P','o','p','_','D','i','r','e','c','t','i','o','n','a','l','_','I','s','o','l','a','t','e',0,
2,'b','l','k',0,'B','l','o','c','k',0,2,'N','B',0,'N','o','_','B','l','o','c','k',0,
2,'A','S','C','I','I',0,'B','a','s','i','c','_','L','a','t','i','n',0,
3,'L','a','t','i','n','_','1','_','S','u','p',0,'L','a','t','i','n','_','1','_','S','u','p','p','l','e','m','e','n','t',0,
'L','a','t','i','n','_','1',0,2,'L','a','t','i','n','_','E','x','t','_','A',0,'L','a','t','i','n','_','E','x','t','e','n',
'd','e','d','_','A',0,2,'L','a','t','i','n','_','E','x','t','_','B',0,'L','a','t','i','n','_','E','x','t','e','n','d','e',
'd','_','B',0,2,'I','P','A','_','E','x','t',0,'I','P','A','_','E','x','t','e','n','s','i','o','n','s',0,
2,'M','o','d','i','f','i','e','r','_','L','e','t','t','e','r','s',0,'S','p','a','c','i','n','g','_','M','o','d','i','f','i',
'e','r','_','L','e','t','t','e','r','s',0,2,'D','i','a','c','r','i','t','i','c','a','l','s',0,
'C','o','m','b','i','n','i','n','g','_','D','i','a','c','r','i','t','i','c','a','l','_','M','a','r','k','s',0,
2,'G','r','e','e','k',0,'G','r','e','e','k','_','A','n','d','_','C','o','p','t','i','c',0,
2,'C','y','r','i','l','l','i','c',0,'C','y','r','i','l','l','i','c',0,
2,'A','r','m','e','n','i','a','n',0,'A','r','m','e','n','i','a','n',0,
2,'H','e','b','r','e','w',0,'H','e','b','r','e','w',0,2,'A','r','a','b','i','c',0,'A','r','a','b','i','c',0,
2,'S','y','r','i','a','c',0,'S','y','r','i','a','c',0,2,'T','h','a','a','n','a',0,'T','h','a','a','n','a',0,
2,'D','e','v','a','n','a','g','a','r','i',0,'D','e','v','a','n','a','g','a','r','i',0,
2,'B','e','n','g','a','l','i',0,'B','e','n','g','a','l','i',0,
2,'G','u','r','m','u','k','h','i',0,'G','u','r','m','u','k','h','i',0,
2,'G','u','j','a','r','a','t','i',0,'G','u','j','a','r','a','t','i',0,
2,'O','r','i','y','a',0,'O','r','i','y','a',0,2,'T','a','m','i','l',0,'T','a','m','i','l',0,
2,'T','e','l','u','g','u',0,'T','e','l','u','g','u',0,2,'K','a','n','n','a','d','a',0,
'K','a','n','n','a','d','a',0,2,'M','a','l','a','y','a','l','a','m',0,'M','a','l','a','y','a','l','a','m',0,
2,'S','i','n','h','a','l','a',0,'S','i','n','h','a','l','a',0,
2,'T','h','a','i',0,'T','h','a','i',0,2,'L','a','o',0,'L','a','o',0,
2,'T','i','b','e','t','a','n',0,'T','i','b','e','t','a','n',0,
2,'M','y','a','n','m','a','r',0,'M','y','a','n','m','a','r',0,
2,'G','e','o','r','g','i','a','n',0,'G','e','o','r','g','i','a','n',0,
2,'J','a','m','o',0,'H','a','n','g','u','l','_','J','a','m','o',0,
2,'E','t','h','i','o','p','i','c',0,'E','t','h','i','o','p','i','c',0,
2,'C','h','e','r','o','k','e','e',0,'C','h','e','r','o','k','e','e',0,
3,'U','C','A','S',0,'U','n','i','f','i','e','d','_','C','a','n','a','d','i','a','n','_','A','b','o','r','i','g','i','n','a',
'l','_','S','y','l','l','a','b','i','c','s',0,'C','a','n','a','d','i','a','n','_','S','y','l','l','a','b','i','c','s',0,
2,'O','g','h','a','m',0,'O','g','h','a','m',0,2,'R','u','n','i','c',0,'R','u','n','i','c',0,
2,'K','h','m','e','r',0,'K','h','m','e','r',0,2,'M','o','n','g','o','l','i','a','n',0,
'M','o','n','g','o','l','i','a','n',0,2,'L','a','t','i','n','_','E','x','t','_','A','d','d','i','t','i','o','n','a','l',0,
'L','a','t','i','n','_','E','x','t','e','n','d','e','d','_','A','d','d','i','t','i','o','n','a','l',0,
2,'G','r','e','e','k','_','E','x','t',0,'G','r','e','e','k','_','E','x','t','e','n','d','e','d',0,
2,'P','u','n','c','t','u','a','t','i','o','n',0,'G','e','n','e','r','a','l','_','P','u','n','c','t','u','a','t','i','o','n',
0,2,'S','u','p','e','r','_','A','n','d','_','S','u','b',0,'S','u','p','e','r','s','c','r','i','p','t','s','_','A','n','d',
'_','S','u','b','s','c','r','i','p','t','s',0,2,'C','u','r','r','e','n','c','y','_','S','y','m','b','o','l','s',0,
'C','u','r','r','e','n','c','y','_','S','y','m','b','o','l','s',0,
3,'D','i','a','c','r','i','t','i','c','a','l','s','_','F','o','r','_','S','y','m','b','o','l','s',0,
'C','o','m','b','i','n','i','n','g','_','D','i','a','c','r','i','t','i','c','a','l','_','M','a','r','k','s','_','F','o','r','_',
'S','y','m','b','o','l','s',0,'C','o','m','b','i','n','i','n','g','_','M','a','r','k','s','_','F','o','r','_','S','y','m','b',
'o','l','s',0,2,'L','e','t','t','e','r','l','i','k','e','_','S','y','m','b','o','l','s',0,
'L','e','t','t','e','r','l','i','k','e','_','S','y','m','b','o','l','s',0,
2,'N','u','m','b','e','r','_','F','o','r','m','s',0,'N','u','m','b','e','r','_','F','o','r','m','s',0,
2,'A','r','r','o','w','s',0,'A','r','r','o','w','s',0,2,'M','a','t','h','_','O','p','e','r','a','t','o','r','s',0,
'M','a','t','h','e','m','a','t','i','c','a','l','_','O','p','e','r','a','t','o','r','s',0,
2,'M','i','s','c','_','T','e','c','h','n','i','c','a','l',0,'M','i','s','c','e','l','l','a','n','e','o','u','s','_','T','e',
'c','h','n','i','c','a','l',0,2,'C','o','n','t','r','o','l','_','P','i','c','t','u','r','e','s',0,
'C','o','n','t','r','o','l','_','P','i','c','t','u','r','e','s',0,
2,'O','C','R',0,'O','p','t','i','c','a','l','_','C','h','a','r','a','c','t','e','r','_','R','e','c','o','g','n','i','t','i',
'o','n',0,2,'E','n','c','l','o','s','e','d','_','A','l','p','h','a','n','u','m',0,'E','n','c','l','o','s','e','d','_','A',
'l','p','h','a','n','u','m','e','r','i','c','s',0,2,'B','o','x','_','D','r','a','w','i','n','g',0,
'B','o','x','_','D','r','a','w','i','n','g',0,2,'B','l','o','c','k','_','E','l','e','m','e','n','t','s',0,
'B','l','o','c','k','_','E','l','e','m','e','n','t','s',0,2,'G','e','o','m','e','t','r','i','c','_','S','h','a','p','e','s',
0,'G','e','o','m','e','t','r','i','c','_','S','h','a','p','e','s',0,
2,'M','i','s','c','_','S','y','m','b','o','l','s',0,'M','i','s','c','e','l','l','a','n','e','o','u','s','_','S','y','m','b',
'o','l','s',0,2,'D','i','n','g','b','a','t','s',0,'D','i','n','g','b','a','t','s',0,
2,'B','r','a','i','l','l','e',0,'B','r','a','i','l','l','e','_','P','a','t','t','e','r','n','s',0,
2,'C','J','K','_','R','a','d','i','c','a','l','s','_','S','u','p',0,'C','J','K','_','R','a','d','i','c','a','l','s','_','S',
'u','p','p','l','e','m','e','n','t',0,2,'K','a','n','g','x','i',0,'K','a','n','g','x','i','_','R','a','d','i','c','a','l',
's',0,2,'I','D','C',0,'I','d','e','o','g','r','a','p','h','i','c','_','D','e','s','c','r','i','p','t','i','o','n','_','C',
'h','a','r','a','c','t','e','r','s',0,2,'C','J','K','_','S','y','m','b','o','l','s',0,'C','J','K','_','S','y','m','b','o',
'l','s','_','A','n','d','_','P','u','n','c','t','u','a','t','i','o','n',0,
2,'H','i','r','a','g','a','n','a',0,'H','i','r','a','g','a','n','a',0,
2,'K','a','t','a','k','a','n','a',0,'K','a','t','a','k','a','n','a',0,
2,'B','o','p','o','m','o','f','o',0,'B','o','p','o','m','o','f','o',0,
2,'C','o','m','p','a','t','_','J','a','m','o',0,'H','a','n','g','u','l','_','C','o','m','p','a','t','i','b','i','l','i','t',
'y','_','J','a','m','o',0,2,'K','a','n','b','u','n',0,'K','a','n','b','u','n',0,
2,'B','o','p','o','m','o','f','o','_','E','x','t',0,'B','o','p','o','m','o','f','o','_','E','x','t','e','n','d','e','d',0,
2,'E','n','c','l','o','s','e','d','_','C','J','K',0,'E','n','c','l','o','s','e','d','_','C','J','K','_','L','e','t','t','e',
'r','s','_','A','n','d','_','M','o','n','t','h','s',0,2,'C','J','K','_','C','o','m','p','a','t',0,
'C','J','K','_','C','o','m','p','a','t','i','b','i','l','i','t','y',0,
2,'C','J','K','_','E','x','t','_','A',0,'C','J','K','_','U','n','i','f','i','e','d','_','I','d','e','o','g','r','a','p','h',
's','_','E','x','t','e','n','s','i','o','n','_','A',0,2,'C','J','K',0,'C','J','K','_','U','n','i','f','i','e','d','_','I',
'd','e','o','g','r','a','p','h','s',0,2,'Y','i','_','S','y','l','l','a','b','l','e','s',0,
'Y','i','_','S','y','l','l','a','b','l','e','s',0,2,'Y','i','_','R','a','d','i','c','a','l','s',0,
'Y','i','_','R','a','d','i','c','a','l','s',0,2,'H','a','n','g','u','l',0,'H','a','n','g','u','l','_','S','y','l','l','a',
'b','l','e','s',0,2,'H','i','g','h','_','S','u','r','r','o','g','a','t','e','s',0,'H','i','g','h','_','S','u','r','r','o',
'g','a','t','e','s',0,2,'H','i','g','h','_','P','U','_','S','u','r','r','o','g','a','t','e','s',0,
'H','i','g','h','_','P','r','i','v','a','t','e','_','U','s','e','_','S','u','r','r','o','g','a','t','e','s',0,
2,'L','o','w','_','S','u','r','r','o','g','a','t','e','s',0,'L','o','w','_','S','u','r','r','o','g','a','t','e','s',0,
3,'P','U','A',0,'P','r','i','v','a','t','e','_','U','s','e','_','A','r','e','a',0,'P','r','i','v','a','t','e','_','U','s',
'e',0,2,'C','J','K','_','C','o','m','p','a','t','_','I','d','e','o','g','r','a','p','h','s',0,
'C','J','K','_','C','o','m','p','a','t','i','b','i','l','i','t','y','_','I','d','e','o','g','r','a','p','h','s',0,
2,'A','l','p','h','a','b','e','t','i','c','_','P','F',0,'A','l','p','h','a','b','e','t','i','c','_','P','r','e','s','e','n',
't','a','t','i','o','n','_','F','o','r','m','s',0,3,'A','r','a','b','i','c','_','P','F','_','A',0,
'A','r','a','b','i','c','_','P','r','e','s','e','n','t','a','t','i','o','n','_','F','o','r','m','s','_','A',0,
'A','r','a','b','i','c','_','P','r','e','s','e','n','t','a','t','i','o','n','_','F','o','r','m','s','-','A',0,
2,'H','a','l','f','_','M','a','r','k','s',0,'C','o','m','b','i','n','i','n','g','_','H','a','l','f','_','M','a','r','k','s',
0,2,'C','J','K','_','C','o','m','p','a','t','_','F','o','r','m','s',0,'C','J','K','_','C','o','m','p','a','t','i','b','i',
'l','i','t','y','_','F','o','r','m','s',0,2,'S','m','a','l','l','_','F','o','r','m','s',0,
'S','m','a','l','l','_','F','o','r','m','_','V','a','r','i','a','n','t','s',0,
2,'A','r','a','b','i','c','_','P','F','_','B',0,'A','r','a','b','i','c','_','P','r','e','s','e','n','t','a','t','i','o','n',
'_','F','o','r','m','s','_','B',0,2,'S','p','e','c','i','a','l','s',0,'S','p','e','c','i','a','l','s',0,
2,'H','a','l','f','_','A','n','d','_','F','u','l','l','_','F','o','r','m','s',0,'H','a','l','f','w','i','d','t','h','_','A',
'n','d','_','F','u','l','l','w','i','d','t','h','_','F','o','r','m','s',0,
2,'O','l','d','_','I','t','a','l','i','c',0,'O','l','d','_','I','t','a','l','i','c',0,
2,'G','o','t','h','i','c',0,'G','o','t','h','i','c',0,2,'D','e','s','e','r','e','t',0,
'D','e','s','e','r','e','t',0,2,'B','y','z','a','n','t','i','n','e','_','M','u','s','i','c',0,
'B','y','z','a','n','t','i','n','e','_','M','u','s','i','c','a','l','_','S','y','m','b','o','l','s',0,
2,'M','u','s','i','c',0,'M','u','s','i','c','a','l','_','S','y','m','b','o','l','s',0,
2,'M','a','t','h','_','A','l','p','h','a','n','u','m',0,'M','a','t','h','e','m','a','t','i','c','a','l','_','A','l','p','h',
'a','n','u','m','e','r','i','c','_','S','y','m','b','o','l','s',0,
2,'C','J','K','_','E','x','t','_','B',0,'C','J','K','_','U','n','i','f','i','e','d','_','I','d','e','o','g','r','a','p','h',
's','_','E','x','t','e','n','s','i','o','n','_','B',0,2,'C','J','K','_','C','o','m','p','a','t','_','I','d','e','o','g','r',
'a','p','h','s','_','S','u','p',0,'C','J','K','_','C','o','m','p','a','t','i','b','i','l','i','t','y','_','I','d','e','o','g',
'r','a','p','h','s','_','S','u','p','p','l','e','m','e','n','t',0,
2,'T','a','g','s',0,'T','a','g','s',0,3,'C','y','r','i','l','l','i','c','_','S','u','p',0,
'C','y','r','i','l','l','i','c','_','S','u','p','p','l','e','m','e','n','t',0,'C','y','r','i','l','l','i','c','_','S','u','p',
'p','l','e','m','e','n','t','a','r','y',0,2,'T','a','g','a','l','o','g',0,'T','a','g','a','l','o','g',0,
2,'H','a','n','u','n','o','o',0,'H','a','n','u','n','o','o',0,
2,'B','u','h','i','d',0,'B','u','h','i','d',0,2,'T','a','g','b','a','n','w','a',0,'T','a','g','b','a','n','w','a',0,
2,'M','i','s','c','_','M','a','t','h','_','S','y','m','b','o','l','s','_','A',0,'M','i','s','c','e','l','l','a','n','e','o',
'u','s','_','M','a','t','h','e','m','a','t','i','c','a','l','_','S','y','m','b','o','l','s','_','A',0,
2,'S','u','p','_','A','r','r','o','w','s','_','A',0,'S','u','p','p','l','e','m','e','n','t','a','l','_','A','r','r','o','w',
's','_','A',0,2,'S','u','p','_','A','r','r','o','w','s','_','B',0,'S','u','p','p','l','e','m','e','n','t','a','l','_','A',
'r','r','o','w','s','_','B',0,2,'M','i','s','c','_','M','a','t','h','_','S','y','m','b','o','l','s','_','B',0,
'M','i','s','c','e','l','l','a','n','e','o','u','s','_','M','a','t','h','e','m','a','t','i','c','a','l','_','S','y','m','b','o',
'l','s','_','B',0,2,'S','u','p','_','M','a','t','h','_','O','p','e','r','a','t','o','r','s',0,
'S','u','p','p','l','e','m','e','n','t','a','l','_','M','a','t','h','e','m','a','t','i','c','a','l','_','O','p','e','r','a','t',
'o','r','s',0,2,'K','a','t','a','k','a','n','a','_','E','x','t',0,'K','a','t','a','k','a','n','a','_','P','h','o','n','e',
't','i','c','_','E','x','t','e','n','s','i','o','n','s',0,2,'V','S',0,'V','a','r','i','a','t','i','o','n','_','S','e','l',
'e','c','t','o','r','s',0,2,'S','u','p','_','P','U','A','_','A',0,'S','u','p','p','l','e','m','e','n','t','a','r','y','_',
'P','r','i','v','a','t','e','_','U','s','e','_','A','r','e','a','_','A',0,
2,'S','u','p','_','P','U','A','_','B',0,'S','u','p','p','l','e','m','e','n','t','a','r','y','_','P','r','i','v','a','t','e',
'_','U','s','e','_','A','r','e','a','_','B',0,2,'L','i','m','b','u',0,'L','i','m','b','u',0,
2,'T','a','i','_','L','e',0,'T','a','i','_','L','e',0,2,'K','h','m','e','r','_','S','y','m','b','o','l','s',0,
'K','h','m','e','r','_','S','y','m','b','o','l','s',0,2,'P','h','o','n','e','t','i','c','_','E','x','t',0,
'P','h','o','n','e','t','i','c','_','E','x','t','e','n','s','i','o','n','s',0,
2,'M','i','s','c','_','A','r','r','o','w','s',0,'M','i','s','c','e','l','l','a','n','e','o','u','s','_','S','y','m','b','o',
'l','s','_','A','n','d','_','A','r','r','o','w','s',0,2,'Y','i','j','i','n','g',0,'Y','i','j','i','n','g','_','H','e','x',
'a','g','r','a','m','_','S','y','m','b','o','l','s',0,2,'L','i','n','e','a','r','_','B','_','S','y','l','l','a','b','a','r',
'y',0,'L','i','n','e','a','r','_','B','_','S','y','l','l','a','b','a','r','y',0,
2,'L','i','n','e','a','r','_','B','_','I','d','e','o','g','r','a','m','s',0,'L','i','n','e','a','r','_','B','_','I','d','e',
'o','g','r','a','m','s',0,2,'A','e','g','e','a','n','_','N','u','m','b','e','r','s',0,'A','e','g','e','a','n','_','N','u',
'm','b','e','r','s',0,2,'U','g','a','r','i','t','i','c',0,'U','g','a','r','i','t','i','c',0,
2,'S','h','a','v','i','a','n',0,'S','h','a','v','i','a','n',0,
2,'O','s','m','a','n','y','a',0,'O','s','m','a','n','y','a',0,
2,'C','y','p','r','i','o','t','_','S','y','l','l','a','b','a','r','y',0,'C','y','p','r','i','o','t','_','S','y','l','l','a',
'b','a','r','y',0,2,'T','a','i','_','X','u','a','n','_','J','i','n','g',0,'T','a','i','_','X','u','a','n','_','J','i','n',
'g','_','S','y','m','b','o','l','s',0,2,'V','S','_','S','u','p',0,'V','a','r','i','a','t','i','o','n','_','S','e','l','e',
'c','t','o','r','s','_','S','u','p','p','l','e','m','e','n','t',0,
2,'A','n','c','i','e','n','t','_','G','r','e','e','k','_','M','u','s','i','c',0,'A','n','c','i','e','n','t','_','G','r','e',
'e','k','_','M','u','s','i','c','a','l','_','N','o','t','a','t','i','o','n',0,
2,'A','n','c','i','e','n','t','_','G','r','e','e','k','_','N','u','m','b','e','r','s',0,'A','n','c','i','e','n','t','_','G',
'r','e','e','k','_','N','u','m','b','e','r','s',0,2,'A','r','a','b','i','c','_','S','u','p',0,
'A','r','a','b','i','c','_','S','u','p','p','l','e','m','e','n','t',0,
2,'B','u','g','i','n','e','s','e',0,'B','u','g','i','n','e','s','e',0,
2,'C','J','K','_','S','t','r','o','k','e','s',0,'C','J','K','_','S','t','r','o','k','e','s',0,
2,'D','i','a','c','r','i','t','i','c','a','l','s','_','S','u','p',0,'C','o','m','b','i','n','i','n','g','_','D','i','a','c',
'r','i','t','i','c','a','l','_','M','a','r','k','s','_','S','u','p','p','l','e','m','e','n','t',0,
2,'C','o','p','t','i','c',0,'C','o','p','t','i','c',0,2,'E','t','h','i','o','p','i','c','_','E','x','t',0,
'E','t','h','i','o','p','i','c','_','E','x','t','e','n','d','e','d',0,
2,'E','t','h','i','o','p','i','c','_','S','u','p',0,'E','t','h','i','o','p','i','c','_','S','u','p','p','l','e','m','e','n',
't',0,2,'G','e','o','r','g','i','a','n','_','S','u','p',0,'G','e','o','r','g','i','a','n','_','S','u','p','p','l','e','m',
'e','n','t',0,2,'G','l','a','g','o','l','i','t','i','c',0,'G','l','a','g','o','l','i','t','i','c',0,
2,'K','h','a','r','o','s','h','t','h','i',0,'K','h','a','r','o','s','h','t','h','i',0,
2,'M','o','d','i','f','i','e','r','_','T','o','n','e','_','L','e','t','t','e','r','s',0,'M','o','d','i','f','i','e','r','_',
'T','o','n','e','_','L','e','t','t','e','r','s',0,2,'N','e','w','_','T','a','i','_','L','u','e',0,
'N','e','w','_','T','a','i','_','L','u','e',0,2,'O','l','d','_','P','e','r','s','i','a','n',0,
'O','l','d','_','P','e','r','s','i','a','n',0,2,'P','h','o','n','e','t','i','c','_','E','x','t','_','S','u','p',0,
'P','h','o','n','e','t','i','c','_','E','x','t','e','n','s','i','o','n','s','_','S','u','p','p','l','e','m','e','n','t',0,
2,'S','u','p','_','P','u','n','c','t','u','a','t','i','o','n',0,'S','u','p','p','l','e','m','e','n','t','a','l','_','P','u',
'n','c','t','u','a','t','i','o','n',0,2,'S','y','l','o','t','i','_','N','a','g','r','i',0,
'S','y','l','o','t','i','_','N','a','g','r','i',0,2,'T','i','f','i','n','a','g','h',0,'T','i','f','i','n','a','g','h',0,
2,'V','e','r','t','i','c','a','l','_','F','o','r','m','s',0,'V','e','r','t','i','c','a','l','_','F','o','r','m','s',0,
2,'N','K','o',0,'N','K','o',0,2,'B','a','l','i','n','e','s','e',0,'B','a','l','i','n','e','s','e',0,
2,'L','a','t','i','n','_','E','x','t','_','C',0,'L','a','t','i','n','_','E','x','t','e','n','d','e','d','_','C',0,
2,'L','a','t','i','n','_','E','x','t','_','D',0,'L','a','t','i','n','_','E','x','t','e','n','d','e','d','_','D',0,
2,'P','h','a','g','s','_','P','a',0,'P','h','a','g','s','_','P','a',0,
2,'P','h','o','e','n','i','c','i','a','n',0,'P','h','o','e','n','i','c','i','a','n',0,
2,'C','u','n','e','i','f','o','r','m',0,'C','u','n','e','i','f','o','r','m',0,
2,'C','u','n','e','i','f','o','r','m','_','N','u','m','b','e','r','s',0,'C','u','n','e','i','f','o','r','m','_','N','u','m',
'b','e','r','s','_','A','n','d','_','P','u','n','c','t','u','a','t','i','o','n',0,
2,'C','o','u','n','t','i','n','g','_','R','o','d',0,'C','o','u','n','t','i','n','g','_','R','o','d','_','N','u','m','e','r',
'a','l','s',0,2,'S','u','n','d','a','n','e','s','e',0,'S','u','n','d','a','n','e','s','e',0,
2,'L','e','p','c','h','a',0,'L','e','p','c','h','a',0,2,'O','l','_','C','h','i','k','i',0,
'O','l','_','C','h','i','k','i',0,2,'C','y','r','i','l','l','i','c','_','E','x','t','_','A',0,
'C','y','r','i','l','l','i','c','_','E','x','t','e','n','d','e','d','_','A',0,
2,'V','a','i',0,'V','a','i',0,2,'C','y','r','i','l','l','i','c','_','E','x','t','_','B',0,
'C','y','r','i','l','l','i','c','_','E','x','t','e','n','d','e','d','_','B',0,
2,'S','a','u','r','a','s','h','t','r','a',0,'S','a','u','r','a','s','h','t','r','a',0,
2,'K','a','y','a','h','_','L','i',0,'K','a','y','a','h','_','L','i',0,
2,'R','e','j','a','n','g',0,'R','e','j','a','n','g',0,2,'C','h','a','m',0,'C','h','a','m',0,
2,'A','n','c','i','e','n','t','_','S','y','m','b','o','l','s',0,'A','n','c','i','e','n','t','_','S','y','m','b','o','l','s',
0,2,'P','h','a','i','s','t','o','s',0,'P','h','a','i','s','t','o','s','_','D','i','s','c',0,
2,'L','y','c','i','a','n',0,'L','y','c','i','a','n',0,2,'C','a','r','i','a','n',0,'C','a','r','i','a','n',0,
2,'L','y','d','i','a','n',0,'L','y','d','i','a','n',0,2,'M','a','h','j','o','n','g',0,
'M','a','h','j','o','n','g','_','T','i','l','e','s',0,2,'D','o','m','i','n','o',0,'D','o','m','i','n','o','_','T','i','l',
'e','s',0,2,'S','a','m','a','r','i','t','a','n',0,'S','a','m','a','r','i','t','a','n',0,
2,'U','C','A','S','_','E','x','t',0,'U','n','i','f','i','e','d','_','C','a','n','a','d','i','a','n','_','A','b','o','r','i',
'g','i','n','a','l','_','S','y','l','l','a','b','i','c','s','_','E','x','t','e','n','d','e','d',0,
2,'T','a','i','_','T','h','a','m',0,'T','a','i','_','T','h','a','m',0,
2,'V','e','d','i','c','_','E','x','t',0,'V','e','d','i','c','_','E','x','t','e','n','s','i','o','n','s',0,
2,'L','i','s','u',0,'L','i','s','u',0,2,'B','a','m','u','m',0,'B','a','m','u','m',0,
2,'I','n','d','i','c','_','N','u','m','b','e','r','_','F','o','r','m','s',0,'C','o','m','m','o','n','_','I','n','d','i','c',
'_','N','u','m','b','e','r','_','F','o','r','m','s',0,2,'D','e','v','a','n','a','g','a','r','i','_','E','x','t',0,
'D','e','v','a','n','a','g','a','r','i','_','E','x','t','e','n','d','e','d',0,
2,'J','a','m','o','_','E','x','t','_','A',0,'H','a','n','g','u','l','_','J','a','m','o','_','E','x','t','e','n','d','e','d',
'_','A',0,2,'J','a','v','a','n','e','s','e',0,'J','a','v','a','n','e','s','e',0,
2,'M','y','a','n','m','a','r','_','E','x','t','_','A',0,'M','y','a','n','m','a','r','_','E','x','t','e','n','d','e','d','_',
'A',0,2,'T','a','i','_','V','i','e','t',0,'T','a','i','_','V','i','e','t',0,
2,'M','e','e','t','e','i','_','M','a','y','e','k',0,'M','e','e','t','e','i','_','M','a','y','e','k',0,
2,'J','a','m','o','_','E','x','t','_','B',0,'H','a','n','g','u','l','_','J','a','m','o','_','E','x','t','e','n','d','e','d',
'_','B',0,2,'I','m','p','e','r','i','a','l','_','A','r','a','m','a','i','c',0,'I','m','p','e','r','i','a','l','_','A','r',
'a','m','a','i','c',0,2,'O','l','d','_','S','o','u','t','h','_','A','r','a','b','i','a','n',0,
'O','l','d','_','S','o','u','t','h','_','A','r','a','b','i','a','n',0,
2,'A','v','e','s','t','a','n',0,'A','v','e','s','t','a','n',0,
2,'I','n','s','c','r','i','p','t','i','o','n','a','l','_','P','a','r','t','h','i','a','n',0,
'I','n','s','c','r','i','p','t','i','o','n','a','l','_','P','a','r','t','h','i','a','n',0,
2,'I','n','s','c','r','i','p','t','i','o','n','a','l','_','P','a','h','l','a','v','i',0,'I','n','s','c','r','i','p','t','i',
'o','n','a','l','_','P','a','h','l','a','v','i',0,2,'O','l','d','_','T','u','r','k','i','c',0,
'O','l','d','_','T','u','r','k','i','c',0,2,'R','u','m','i',0,'R','u','m','i','_','N','u','m','e','r','a','l','_','S','y',
'm','b','o','l','s',0,2,'K','a','i','t','h','i',0,'K','a','i','t','h','i',0,
2,'E','g','y','p','t','i','a','n','_','H','i','e','r','o','g','l','y','p','h','s',0,'E','g','y','p','t','i','a','n','_','H',
'i','e','r','o','g','l','y','p','h','s',0,2,'E','n','c','l','o','s','e','d','_','A','l','p','h','a','n','u','m','_','S','u',
'p',0,'E','n','c','l','o','s','e','d','_','A','l','p','h','a','n','u','m','e','r','i','c','_','S','u','p','p','l','e','m','e',
'n','t',0,2,'E','n','c','l','o','s','e','d','_','I','d','e','o','g','r','a','p','h','i','c','_','S','u','p',0,
'E','n','c','l','o','s','e','d','_','I','d','e','o','g','r','a','p','h','i','c','_','S','u','p','p','l','e','m','e','n','t',0,
2,'C','J','K','_','E','x','t','_','C',0,'C','J','K','_','U','n','i','f','i','e','d','_','I','d','e','o','g','r','a','p','h',
's','_','E','x','t','e','n','s','i','o','n','_','C',0,2,'M','a','n','d','a','i','c',0,'M','a','n','d','a','i','c',0,
2,'B','a','t','a','k',0,'B','a','t','a','k',0,2,'E','t','h','i','o','p','i','c','_','E','x','t','_','A',0,
'E','t','h','i','o','p','i','c','_','E','x','t','e','n','d','e','d','_','A',0,
2,'B','r','a','h','m','i',0,'B','r','a','h','m','i',0,2,'B','a','m','u','m','_','S','u','p',0,
'B','a','m','u','m','_','S','u','p','p','l','e','m','e','n','t',0,
2,'K','a','n','a','_','S','u','p',0,'K','a','n','a','_','S','u','p','p','l','e','m','e','n','t',0,
2,'P','l','a','y','i','n','g','_','C','a','r','d','s',0,'P','l','a','y','i','n','g','_','C','a','r','d','s',0,
2,'M','i','s','c','_','P','i','c','t','o','g','r','a','p','h','s',0,'M','i','s','c','e','l','l','a','n','e','o','u','s','_',
'S','y','m','b','o','l','s','_','A','n','d','_','P','i','c','t','o','g','r','a','p','h','s',0,
2,'E','m','o','t','i','c','o','n','s',0,'E','m','o','t','i','c','o','n','s',0,
2,'T','r','a','n','s','p','o','r','t','_','A','n','d','_','M','a','p',0,'T','r','a','n','s','p','o','r','t','_','A','n','d',
'_','M','a','p','_','S','y','m','b','o','l','s',0,2,'A','l','c','h','e','m','i','c','a','l',0,
'A','l','c','h','e','m','i','c','a','l','_','S','y','m','b','o','l','s',0,
2,'C','J','K','_','E','x','t','_','D',0,'C','J','K','_','U','n','i','f','i','e','d','_','I','d','e','o','g','r','a','p','h',
's','_','E','x','t','e','n','s','i','o','n','_','D',0,2,'A','r','a','b','i','c','_','E','x','t','_','A',0,
'A','r','a','b','i','c','_','E','x','t','e','n','d','e','d','_','A',0,
2,'A','r','a','b','i','c','_','M','a','t','h',0,'A','r','a','b','i','c','_','M','a','t','h','e','m','a','t','i','c','a','l',
'_','A','l','p','h','a','b','e','t','i','c','_','S','y','m','b','o','l','s',0,
2,'C','h','a','k','m','a',0,'C','h','a','k','m','a',0,2,'M','e','e','t','e','i','_','M','a','y','e','k','_','E','x','t',
0,'M','e','e','t','e','i','_','M','a','y','e','k','_','E','x','t','e','n','s','i','o','n','s',0,
2,'M','e','r','o','i','t','i','c','_','C','u','r','s','i','v','e',0,'M','e','r','o','i','t','i','c','_','C','u','r','s','i',
'v','e',0,2,'M','e','r','o','i','t','i','c','_','H','i','e','r','o','g','l','y','p','h','s',0,
'M','e','r','o','i','t','i','c','_','H','i','e','r','o','g','l','y','p','h','s',0,
2,'M','i','a','o',0,'M','i','a','o',0,2,'S','h','a','r','a','d','a',0,'S','h','a','r','a','d','a',0,
2,'S','o','r','a','_','S','o','m','p','e','n','g',0,'S','o','r','a','_','S','o','m','p','e','n','g',0,
2,'S','u','n','d','a','n','e','s','e','_','S','u','p',0,'S','u','n','d','a','n','e','s','e','_','S','u','p','p','l','e','m',
'e','n','t',0,2,'T','a','k','r','i',0,'T','a','k','r','i',0,
2,'B','a','s','s','a','_','V','a','h',0,'B','a','s','s','a','_','V','a','h',0,
2,'C','a','u','c','a','s','i','a','n','_','A','l','b','a','n','i','a','n',0,'C','a','u','c','a','s','i','a','n','_','A','l',
'b','a','n','i','a','n',0,2,'C','o','p','t','i','c','_','E','p','a','c','t','_','N','u','m','b','e','r','s',0,
'C','o','p','t','i','c','_','E','p','a','c','t','_','N','u','m','b','e','r','s',0,
2,'D','i','a','c','r','i','t','i','c','a','l','s','_','E','x','t',0,'C','o','m','b','i','n','i','n','g','_','D','i','a','c',
'r','i','t','i','c','a','l','_','M','a','r','k','s','_','E','x','t','e','n','d','e','d',0,
2,'D','u','p','l','o','y','a','n',0,'D','u','p','l','o','y','a','n',0,
2,'E','l','b','a','s','a','n',0,'E','l','b','a','s','a','n',0,
2,'G','e','o','m','e','t','r','i','c','_','S','h','a','p','e','s','_','E','x','t',0,'G','e','o','m','e','t','r','i','c','_',
'S','h','a','p','e','s','_','E','x','t','e','n','d','e','d',0,
2,'G','r','a','n','t','h','a',0,'G','r','a','n','t','h','a',0,
2,'K','h','o','j','k','i',0,'K','h','o','j','k','i',0,2,'K','h','u','d','a','w','a','d','i',0,
'K','h','u','d','a','w','a','d','i',0,2,'L','a','t','i','n','_','E','x','t','_','E',0,'L','a','t','i','n','_','E','x','t',
'e','n','d','e','d','_','E',0,2,'L','i','n','e','a','r','_','A',0,'L','i','n','e','a','r','_','A',0,
2,'M','a','h','a','j','a','n','i',0,'M','a','h','a','j','a','n','i',0,
2,'M','a','n','i','c','h','a','e','a','n',0,'M','a','n','i','c','h','a','e','a','n',0,
2,'M','e','n','d','e','_','K','i','k','a','k','u','i',0,'M','e','n','d','e','_','K','i','k','a','k','u','i',0,
2,'M','o','d','i',0,'M','o','d','i',0,2,'M','r','o',0,'M','r','o',0,
2,'M','y','a','n','m','a','r','_','E','x','t','_','B',0,'M','y','a','n','m','a','r','_','E','x','t','e','n','d','e','d','_',
'B',0,2,'N','a','b','a','t','a','e','a','n',0,'N','a','b','a','t','a','e','a','n',0,
2,'O','l','d','_','N','o','r','t','h','_','A','r','a','b','i','a','n',0,'O','l','d','_','N','o','r','t','h','_','A','r','a',
'b','i','a','n',0,2,'O','l','d','_','P','e','r','m','i','c',0,'O','l','d','_','P','e','r','m','i','c',0,
2,'O','r','n','a','m','e','n','t','a','l','_','D','i','n','g','b','a','t','s',0,'O','r','n','a','m','e','n','t','a','l','_',
'D','i','n','g','b','a','t','s',0,2,'P','a','h','a','w','h','_','H','m','o','n','g',0,'P','a','h','a','w','h','_','H','m',
'o','n','g',0,2,'P','a','l','m','y','r','e','n','e',0,'P','a','l','m','y','r','e','n','e',0,
2,'P','a','u','_','C','i','n','_','H','a','u',0,'P','a','u','_','C','i','n','_','H','a','u',0,
2,'P','s','a','l','t','e','r','_','P','a','h','l','a','v','i',0,'P','s','a','l','t','e','r','_','P','a','h','l','a','v','i',
0,2,'S','h','o','r','t','h','a','n','d','_','F','o','r','m','a','t','_','C','o','n','t','r','o','l','s',0,
'S','h','o','r','t','h','a','n','d','_','F','o','r','m','a','t','_','C','o','n','t','r','o','l','s',0,
2,'S','i','d','d','h','a','m',0,'S','i','d','d','h','a','m',0,
2,'S','i','n','h','a','l','a','_','A','r','c','h','a','i','c','_','N','u','m','b','e','r','s',0,
'S','i','n','h','a','l','a','_','A','r','c','h','a','i','c','_','N','u','m','b','e','r','s',0,
2,'S','u','p','_','A','r','r','o','w','s','_','C',0,'S','u','p','p','l','e','m','e','n','t','a','l','_','A','r','r','o','w',
's','_','C',0,2,'T','i','r','h','u','t','a',0,'T','i','r','h','u','t','a',0,
2,'W','a','r','a','n','g','_','C','i','t','i',0,'W','a','r','a','n','g','_','C','i','t','i',0,
2,'A','h','o','m',0,'A','h','o','m',0,2,'A','n','a','t','o','l','i','a','n','_','H','i','e','r','o','g','l','y','p','h',
's',0,'A','n','a','t','o','l','i','a','n','_','H','i','e','r','o','g','l','y','p','h','s',0,
2,'C','h','e','r','o','k','e','e','_','S','u','p',0,'C','h','e','r','o','k','e','e','_','S','u','p','p','l','e','m','e','n',
't',0,2,'C','J','K','_','E','x','t','_','E',0,'C','J','K','_','U','n','i','f','i','e','d','_','I','d','e','o','g','r','a',
'p','h','s','_','E','x','t','e','n','s','i','o','n','_','E',0,
2,'E','a','r','l','y','_','D','y','n','a','s','t','i','c','_','C','u','n','e','i','f','o','r','m',0,
'E','a','r','l','y','_','D','y','n','a','s','t','i','c','_','C','u','n','e','i','f','o','r','m',0,
2,'H','a','t','r','a','n',0,'H','a','t','r','a','n',0,2,'M','u','l','t','a','n','i',0,
'M','u','l','t','a','n','i',0,2,'O','l','d','_','H','u','n','g','a','r','i','a','n',0,'O','l','d','_','H','u','n','g','a',
'r','i','a','n',0,2,'S','u','p','_','S','y','m','b','o','l','s','_','A','n','d','_','P','i','c','t','o','g','r','a','p','h',
's',0,'S','u','p','p','l','e','m','e','n','t','a','l','_','S','y','m','b','o','l','s','_','A','n','d','_','P','i','c','t','o',
'g','r','a','p','h','s',0,2,'S','u','t','t','o','n','_','S','i','g','n','W','r','i','t','i','n','g',0,
'S','u','t','t','o','n','_','S','i','g','n','W','r','i','t','i','n','g',0,
2,'A','d','l','a','m',0,'A','d','l','a','m',0,2,'B','h','a','i','k','s','u','k','i',0,
'B','h','a','i','k','s','u','k','i',0,2,'C','y','r','i','l','l','i','c','_','E','x','t','_','C',0,
'C','y','r','i','l','l','i','c','_','E','x','t','e','n','d','e','d','_','C',0,
2,'G','l','a','g','o','l','i','t','i','c','_','S','u','p',0,'G','l','a','g','o','l','i','t','i','c','_','S','u','p','p','l',
'e','m','e','n','t',0,2,'I','d','e','o','g','r','a','p','h','i','c','_','S','y','m','b','o','l','s',0,
'I','d','e','o','g','r','a','p','h','i','c','_','S','y','m','b','o','l','s','_','A','n','d','_','P','u','n','c','t','u','a','t',
'i','o','n',0,2,'M','a','r','c','h','e','n',0,'M','a','r','c','h','e','n',0,
2,'M','o','n','g','o','l','i','a','n','_','S','u','p',0,'M','o','n','g','o','l','i','a','n','_','S','u','p','p','l','e','m',
'e','n','t',0,2,'N','e','w','a',0,'N','e','w','a',0,2,'O','s','a','g','e',0,'O','s','a','g','e',0,
2,'T','a','n','g','u','t',0,'T','a','n','g','u','t',0,2,'T','a','n','g','u','t','_','C','o','m','p','o','n','e','n','t',
's',0,'T','a','n','g','u','t','_','C','o','m','p','o','n','e','n','t','s',0,
2,'C','J','K','_','E','x','t','_','F',0,'C','J','K','_','U','n','i','f','i','e','d','_','I','d','e','o','g','r','a','p','h',
's','_','E','x','t','e','n','s','i','o','n','_','F',0,2,'K','a','n','a','_','E','x','t','_','A',0,
'K','a','n','a','_','E','x','t','e','n','d','e','d','_','A',0,
2,'M','a','s','a','r','a','m','_','G','o','n','d','i',0,'M','a','s','a','r','a','m','_','G','o','n','d','i',0,
2,'N','u','s','h','u',0,'N','u','s','h','u',0,2,'S','o','y','o','m','b','o',0,'S','o','y','o','m','b','o',0,
2,'S','y','r','i','a','c','_','S','u','p',0,'S','y','r','i','a','c','_','S','u','p','p','l','e','m','e','n','t',0,
2,'Z','a','n','a','b','a','z','a','r','_','S','q','u','a','r','e',0,'Z','a','n','a','b','a','z','a','r','_','S','q','u','a',
'r','e',0,2,'C','h','e','s','s','_','S','y','m','b','o','l','s',0,'C','h','e','s','s','_','S','y','m','b','o','l','s',0,
2,'D','o','g','r','a',0,'D','o','g','r','a',0,2,'G','e','o','r','g','i','a','n','_','E','x','t',0,
'G','e','o','r','g','i','a','n','_','E','x','t','e','n','d','e','d',0,
2,'G','u','n','j','a','l','a','_','G','o','n','d','i',0,'G','u','n','j','a','l','a','_','G','o','n','d','i',0,
2,'H','a','n','i','f','i','_','R','o','h','i','n','g','y','a',0,'H','a','n','i','f','i','_','R','o','h','i','n','g','y','a',
0,2,'I','n','d','i','c','_','S','i','y','a','q','_','N','u','m','b','e','r','s',0,'I','n','d','i','c','_','S','i','y','a',
'q','_','N','u','m','b','e','r','s',0,2,'M','a','k','a','s','a','r',0,'M','a','k','a','s','a','r',0,
2,'M','a','y','a','n','_','N','u','m','e','r','a','l','s',0,'M','a','y','a','n','_','N','u','m','e','r','a','l','s',0,
2,'M','e','d','e','f','a','i','d','r','i','n',0,'M','e','d','e','f','a','i','d','r','i','n',0,
2,'O','l','d','_','S','o','g','d','i','a','n',0,'O','l','d','_','S','o','g','d','i','a','n',0,
2,'S','o','g','d','i','a','n',0,'S','o','g','d','i','a','n',0,
2,'E','g','y','p','t','i','a','n','_','H','i','e','r','o','g','l','y','p','h','_','F','o','r','m','a','t','_','C','o','n','t',
'r','o','l','s',0,'E','g','y','p','t','i','a','n','_','H','i','e','r','o','g','l','y','p','h','_','F','o','r','m','a','t','_',
'C','o','n','t','r','o','l','s',0,2,'E','l','y','m','a','i','c',0,'E','l','y','m','a','i','c',0,
2,'N','a','n','d','i','n','a','g','a','r','i',0,'N','a','n','d','i','n','a','g','a','r','i',0,
2,'N','y','i','a','k','e','n','g','_','P','u','a','c','h','u','e','_','H','m','o','n','g',0,
'N','y','i','a','k','e','n','g','_','P','u','a','c','h','u','e','_','H','m','o','n','g',0,
2,'O','t','t','o','m','a','n','_','S','i','y','a','q','_','N','u','m','b','e','r','s',0,'O','t','t','o','m','a','n','_','S',
'i','y','a','q','_','N','u','m','b','e','r','s',0,2,'S','m','a','l','l','_','K','a','n','a','_','E','x','t',0,
'S','m','a','l','l','_','K','a','n','a','_','E','x','t','e','n','s','i','o','n',0,
2,'S','y','m','b','o','l','s','_','A','n','d','_','P','i','c','t','o','g','r','a','p','h','s','_','E','x','t','_','A',0,
'S','y','m','b','o','l','s','_','A','n','d','_','P','i','c','t','o','g','r','a','p','h','s','_','E','x','t','e','n','d','e','d',
'_','A',0,2,'T','a','m','i','l','_','S','u','p',0,'T','a','m','i','l','_','S','u','p','p','l','e','m','e','n','t',0,
2,'W','a','n','c','h','o',0,'W','a','n','c','h','o',0,2,'C','h','o','r','a','s','m','i','a','n',0,
'C','h','o','r','a','s','m','i','a','n',0,2,'C','J','K','_','E','x','t','_','G',0,'C','J','K','_','U','n','i','f','i','e',
'd','_','I','d','e','o','g','r','a','p','h','s','_','E','x','t','e','n','s','i','o','n','_','G',0,
2,'D','i','v','e','s','_','A','k','u','r','u',0,'D','i','v','e','s','_','A','k','u','r','u',0,
2,'K','h','i','t','a','n','_','S','m','a','l','l','_','S','c','r','i','p','t',0,'K','h','i','t','a','n','_','S','m','a','l',
'l','_','S','c','r','i','p','t',0,2,'L','i','s','u','_','S','u','p',0,'L','i','s','u','_','S','u','p','p','l','e','m','e',
'n','t',0,2,'S','y','m','b','o','l','s','_','F','o','r','_','L','e','g','a','c','y','_','C','o','m','p','u','t','i','n','g',
0,'S','y','m','b','o','l','s','_','F','o','r','_','L','e','g','a','c','y','_','C','o','m','p','u','t','i','n','g',0,
2,'T','a','n','g','u','t','_','S','u','p',0,'T','a','n','g','u','t','_','S','u','p','p','l','e','m','e','n','t',0,
2,'Y','e','z','i','d','i',0,'Y','e','z','i','d','i',0,2,'c','c','c',0,'C','a','n','o','n','i','c','a','l','_','C','o',
'm','b','i','n','i','n','g','_','C','l','a','s','s',0,2,'d','t',0,'D','e','c','o','m','p','o','s','i','t','i','o','n','_',
'T','y','p','e',0,3,'N','o','n','e',0,'N','o','n','e',0,'n','o','n','e',0,
3,'C','a','n',0,'C','a','n','o','n','i','c','a','l',0,'c','a','n',0,
3,'C','o','m',0,'C','o','m','p','a','t',0,'c','o','m',0,
3,'E','n','c',0,'C','i','r','c','l','e',0,'e','n','c',0,
3,'F','i','n',0,'F','i','n','a','l',0,'f','i','n',0,3,'F','o','n','t',0,'F','o','n','t',0,
'f','o','n','t',0,3,'F','r','a',0,'F','r','a','c','t','i','o','n',0,'f','r','a',0,
3,'I','n','i','t',0,'I','n','i','t','i','a','l',0,'i','n','i','t',0,
3,'I','s','o',0,'I','s','o','l','a','t','e','d',0,'i','s','o',0,
3,'M','e','d',0,'M','e','d','i','a','l',0,'m','e','d',0,
3,'N','a','r',0,'N','a','r','r','o','w',0,'n','a','r',0,
3,'N','b',0,'N','o','b','r','e','a','k',0,'n','b',0,3,'S','m','l',0,'S','m','a','l','l',0,
's','m','l',0,3,'S','q','r',0,'S','q','u','a','r','e',0,'s','q','r',0,
3,'S','u','b',0,'S','u','b',0,'s','u','b',0,3,'S','u','p',0,'S','u','p','e','r',0,
's','u','p',0,3,'V','e','r','t',0,'V','e','r','t','i','c','a','l',0,'v','e','r','t',0,
3,'W','i','d','e',0,'W','i','d','e',0,'w','i','d','e',0,
2,'e','a',0,'E','a','s','t','_','A','s','i','a','n','_','W','i','d','t','h',0,
2,'N',0,'N','e','u','t','r','a','l',0,2,'A',0,'A','m','b','i','g','u','o','u','s',0,
2,'H',0,'H','a','l','f','w','i','d','t','h',0,2,'F',0,'F','u','l','l','w','i','d','t','h',0,
2,'N','a',0,'N','a','r','r','o','w',0,2,'W',0,'W','i','d','e',0,
2,'g','c',0,'G','e','n','e','r','a','l','_','C','a','t','e','g','o','r','y',0,
2,'C','n',0,'U','n','a','s','s','i','g','n','e','d',0,2,'L','u',0,'U','p','p','e','r','c','a','s','e','_','L','e','t',
't','e','r',0,2,'L','l',0,'L','o','w','e','r','c','a','s','e','_','L','e','t','t','e','r',0,
2,'L','t',0,'T','i','t','l','e','c','a','s','e','_','L','e','t','t','e','r',0,
2,'L','m',0,'M','o','d','i','f','i','e','r','_','L','e','t','t','e','r',0,
2,'L','o',0,'O','t','h','e','r','_','L','e','t','t','e','r',0,
2,'M','n',0,'N','o','n','s','p','a','c','i','n','g','_','M','a','r','k',0,
2,'M','e',0,'E','n','c','l','o','s','i','n','g','_','M','a','r','k',0,
2,'M','c',0,'S','p','a','c','i','n','g','_','M','a','r','k',0,
3,'N','d',0,'D','e','c','i','m','a','l','_','N','u','m','b','e','r',0,'d','i','g','i','t',0,
2,'N','l',0,'L','e','t','t','e','r','_','N','u','m','b','e','r',0,
2,'N','o',0,'O','t','h','e','r','_','N','u','m','b','e','r',0,
2,'Z','s',0,'S','p','a','c','e','_','S','e','p','a','r','a','t','o','r',0,
2,'Z','l',0,'L','i','n','e','_','S','e','p','a','r','a','t','o','r',0,
2,'Z','p',0,'P','a','r','a','g','r','a','p','h','_','S','e','p','a','r','a','t','o','r',0,
3,'C','c',0,'C','o','n','t','r','o','l',0,'c','n','t','r','l',0,
2,'C','f',0,'F','o','r','m','a','t',0,2,'C','o',0,'P','r','i','v','a','t','e','_','U','s','e',0,
2,'C','s',0,'S','u','r','r','o','g','a','t','e',0,2,'P','d',0,'D','a','s','h','_','P','u','n','c','t','u','a','t','i',
'o','n',0,2,'P','s',0,'O','p','e','n','_','P','u','n','c','t','u','a','t','i','o','n',0,
2,'P','e',0,'C','l','o','s','e','_','P','u','n','c','t','u','a','t','i','o','n',0,
2,'P','c',0,'C','o','n','n','e','c','t','o','r','_','P','u','n','c','t','u','a','t','i','o','n',0,
2,'P','o',0,'O','t','h','e','r','_','P','u','n','c','t','u','a','t','i','o','n',0,
2,'S','m',0,'M','a','t','h','_','S','y','m','b','o','l',0,
2,'S','c',0,'C','u','r','r','e','n','c','y','_','S','y','m','b','o','l',0,
2,'S','k',0,'M','o','d','i','f','i','e','r','_','S','y','m','b','o','l',0,
2,'S','o',0,'O','t','h','e','r','_','S','y','m','b','o','l',0,
2,'P','i',0,'I','n','i','t','i','a','l','_','P','u','n','c','t','u','a','t','i','o','n',0,
2,'P','f',0,'F','i','n','a','l','_','P','u','n','c','t','u','a','t','i','o','n',0,
2,'j','g',0,'J','o','i','n','i','n','g','_','G','r','o','u','p',0,
2,'N','o','_','J','o','i','n','i','n','g','_','G','r','o','u','p',0,'N','o','_','J','o','i','n','i','n','g','_','G','r','o',
'u','p',0,2,'A','i','n',0,'A','i','n',0,2,'A','l','a','p','h',0,'A','l','a','p','h',0,
2,'A','l','e','f',0,'A','l','e','f',0,2,'B','e','h',0,'B','e','h',0,
2,'B','e','t','h',0,'B','e','t','h',0,2,'D','a','l',0,'D','a','l',0,
2,'D','a','l','a','t','h','_','R','i','s','h',0,'D','a','l','a','t','h','_','R','i','s','h',0,
2,'E',0,'E',0,2,'F','e','h',0,'F','e','h',0,2,'F','i','n','a','l','_','S','e','m','k','a','t','h',0,
'F','i','n','a','l','_','S','e','m','k','a','t','h',0,2,'G','a','f',0,'G','a','f',0,
2,'G','a','m','a','l',0,'G','a','m','a','l',0,2,'H','a','h',0,'H','a','h',0,
2,'T','e','h','_','M','a','r','b','u','t','a','_','G','o','a','l',0,'H','a','m','z','a','_','O','n','_','H','e','h','_','G',
'o','a','l',0,2,'H','e',0,'H','e',0,2,'H','e','h',0,'H','e','h',0,
2,'H','e','h','_','G','o','a','l',0,'H','e','h','_','G','o','a','l',0,
2,'H','e','t','h',0,'H','e','t','h',0,2,'K','a','f',0,'K','a','f',0,
2,'K','a','p','h',0,'K','a','p','h',0,2,'K','n','o','t','t','e','d','_','H','e','h',0,
'K','n','o','t','t','e','d','_','H','e','h',0,2,'L','a','m',0,'L','a','m',0,
2,'L','a','m','a','d','h',0,'L','a','m','a','d','h',0,2,'M','e','e','m',0,'M','e','e','m',0,
2,'M','i','m',0,'M','i','m',0,2,'N','o','o','n',0,'N','o','o','n',0,
2,'N','u','n',0,'N','u','n',0,2,'P','e',0,'P','e',0,
2,'Q','a','f',0,'Q','a','f',0,2,'Q','a','p','h',0,'Q','a','p','h',0,
2,'R','e','h',0,'R','e','h',0,2,'R','e','v','e','r','s','e','d','_','P','e',0,'R','e','v','e','r','s','e','d','_','P',
'e',0,2,'S','a','d',0,'S','a','d',0,2,'S','a','d','h','e',0,'S','a','d','h','e',0,
2,'S','e','e','n',0,'S','e','e','n',0,2,'S','e','m','k','a','t','h',0,'S','e','m','k','a','t','h',0,
2,'S','h','i','n',0,'S','h','i','n',0,2,'S','w','a','s','h','_','K','a','f',0,'S','w','a','s','h','_','K','a','f',0,
2,'S','y','r','i','a','c','_','W','a','w',0,'S','y','r','i','a','c','_','W','a','w',0,
2,'T','a','h',0,'T','a','h',0,2,'T','a','w',0,'T','a','w',0,
2,'T','e','h','_','M','a','r','b','u','t','a',0,'T','e','h','_','M','a','r','b','u','t','a',0,
2,'T','e','t','h',0,'T','e','t','h',0,2,'W','a','w',0,'W','a','w',0,
2,'Y','e','h',0,'Y','e','h',0,2,'Y','e','h','_','B','a','r','r','e','e',0,'Y','e','h','_','B','a','r','r','e','e',0,
2,'Y','e','h','_','W','i','t','h','_','T','a','i','l',0,'Y','e','h','_','W','i','t','h','_','T','a','i','l',0,
2,'Y','u','d','h',0,'Y','u','d','h',0,2,'Y','u','d','h','_','H','e',0,'Y','u','d','h','_','H','e',0,
2,'Z','a','i','n',0,'Z','a','i','n',0,2,'F','e',0,'F','e',0,
2,'K','h','a','p','h',0,'K','h','a','p','h',0,2,'Z','h','a','i','n',0,'Z','h','a','i','n',0,
2,'B','u','r','u','s','h','a','s','k','i','_','Y','e','h','_','B','a','r','r','e','e',0,'B','u','r','u','s','h','a','s','k',
'i','_','Y','e','h','_','B','a','r','r','e','e',0,2,'F','a','r','s','i','_','Y','e','h',0,
'F','a','r','s','i','_','Y','e','h',0,2,'N','y','a',0,'N','y','a',0,
2,'R','o','h','i','n','g','y','a','_','Y','e','h',0,'R','o','h','i','n','g','y','a','_','Y','e','h',0,
2,'M','a','n','i','c','h','a','e','a','n','_','A','l','e','p','h',0,'M','a','n','i','c','h','a','e','a','n','_','A','l','e',
'p','h',0,2,'M','a','n','i','c','h','a','e','a','n','_','A','y','i','n',0,'M','a','n','i','c','h','a','e','a','n','_','A',
'y','i','n',0,2,'M','a','n','i','c','h','a','e','a','n','_','B','e','t','h',0,'M','a','n','i','c','h','a','e','a','n','_',
'B','e','t','h',0,2,'M','a','n','i','c','h','a','e','a','n','_','D','a','l','e','t','h',0,
'M','a','n','i','c','h','a','e','a','n','_','D','a','l','e','t','h',0,
2,'M','a','n','i','c','h','a','e','a','n','_','D','h','a','m','e','d','h',0,'M','a','n','i','c','h','a','e','a','n','_','D',
'h','a','m','e','d','h',0,2,'M','a','n','i','c','h','a','e','a','n','_','F','i','v','e',0,
'M','a','n','i','c','h','a','e','a','n','_','F','i','v','e',0,
2,'M','a','n','i','c','h','a','e','a','n','_','G','i','m','e','l',0,'M','a','n','i','c','h','a','e','a','n','_','G','i','m',
'e','l',0,2,'M','a','n','i','c','h','a','e','a','n','_','H','e','t','h',0,'M','a','n','i','c','h','a','e','a','n','_','H',
'e','t','h',0,2,'M','a','n','i','c','h','a','e','a','n','_','H','u','n','d','r','e','d',0,
'M','a','n','i','c','h','a','e','a','n','_','H','u','n','d','r','e','d',0,
2,'M','a','n','i','c','h','a','e','a','n','_','K','a','p','h',0,'M','a','n','i','c','h','a','e','a','n','_','K','a','p','h',
0,2,'M','a','n','i','c','h','a','e','a','n','_','L','a','m','e','d','h',0,'M','a','n','i','c','h','a','e','a','n','_','L',
'a','m','e','d','h',0,2,'M','a','n','i','c','h','a','e','a','n','_','M','e','m',0,'M','a','n','i','c','h','a','e','a','n',
'_','M','e','m',0,2,'M','a','n','i','c','h','a','e','a','n','_','N','u','n',0,'M','a','n','i','c','h','a','e','a','n','_',
'N','u','n',0,2,'M','a','n','i','c','h','a','e','a','n','_','O','n','e',0,'M','a','n','i','c','h','a','e','a','n','_','O',
'n','e',0,2,'M','a','n','i','c','h','a','e','a','n','_','P','e',0,'M','a','n','i','c','h','a','e','a','n','_','P','e',0,
2,'M','a','n','i','c','h','a','e','a','n','_','Q','o','p','h',0,'M','a','n','i','c','h','a','e','a','n','_','Q','o','p','h',
0,2,'M','a','n','i','c','h','a','e','a','n','_','R','e','s','h',0,'M','a','n','i','c','h','a','e','a','n','_','R','e','s',
'h',0,2,'M','a','n','i','c','h','a','e','a','n','_','S','a','d','h','e',0,'M','a','n','i','c','h','a','e','a','n','_','S',
'a','d','h','e',0,2,'M','a','n','i','c','h','a','e','a','n','_','S','a','m','e','k','h',0,
'M','a','n','i','c','h','a','e','a','n','_','S','a','m','e','k','h',0,
2,'M','a','n','i','c','h','a','e','a','n','_','T','a','w',0,'M','a','n','i','c','h','a','e','a','n','_','T','a','w',0,
2,'M','a','n','i','c','h','a','e','a','n','_','T','e','n',0,'M','a','n','i','c','h','a','e','a','n','_','T','e','n',0,
2,'M','a','n','i','c','h','a','e','a','n','_','T','e','t','h',0,'M','a','n','i','c','h','a','e','a','n','_','T','e','t','h',
0,2,'M','a','n','i','c','h','a','e','a','n','_','T','h','a','m','e','d','h',0,'M','a','n','i','c','h','a','e','a','n','_',
'T','h','a','m','e','d','h',0,2,'M','a','n','i','c','h','a','e','a','n','_','T','w','e','n','t','y',0,
'M','a','n','i','c','h','a','e','a','n','_','T','w','e','n','t','y',0,
2,'M','a','n','i','c','h','a','e','a','n','_','W','a','w',0,'M','a','n','i','c','h','a','e','a','n','_','W','a','w',0,
2,'M','a','n','i','c','h','a','e','a','n','_','Y','o','d','h',0,'M','a','n','i','c','h','a','e','a','n','_','Y','o','d','h',
0,2,'M','a','n','i','c','h','a','e','a','n','_','Z','a','y','i','n',0,'M','a','n','i','c','h','a','e','a','n','_','Z','a',
'y','i','n',0,2,'S','t','r','a','i','g','h','t','_','W','a','w',0,'S','t','r','a','i','g','h','t','_','W','a','w',0,
2,'A','f','r','i','c','a','n','_','F','e','h',0,'A','f','r','i','c','a','n','_','F','e','h',0,
2,'A','f','r','i','c','a','n','_','N','o','o','n',0,'A','f','r','i','c','a','n','_','N','o','o','n',0,
2,'A','f','r','i','c','a','n','_','Q','a','f',0,'A','f','r','i','c','a','n','_','Q','a','f',0,
2,'M','a','l','a','y','a','l','a','m','_','B','h','a',0,'M','a','l','a','y','a','l','a','m','_','B','h','a',0,
2,'M','a','l','a','y','a','l','a','m','_','J','a',0,'M','a','l','a','y','a','l','a','m','_','J','a',0,
2,'M','a','l','a','y','a','l','a','m','_','L','l','a',0,'M','a','l','a','y','a','l','a','m','_','L','l','a',0,
2,'M','a','l','a','y','a','l','a','m','_','L','l','l','a',0,'M','a','l','a','y','a','l','a','m','_','L','l','l','a',0,
2,'M','a','l','a','y','a','l','a','m','_','N','g','a',0,'M','a','l','a','y','a','l','a','m','_','N','g','a',0,
2,'M','a','l','a','y','a','l','a','m','_','N','n','a',0,'M','a','l','a','y','a','l','a','m','_','N','n','a',0,
2,'M','a','l','a','y','a','l','a','m','_','N','n','n','a',0,'M','a','l','a','y','a','l','a','m','_','N','n','n','a',0,
2,'M','a','l','a','y','a','l','a','m','_','N','y','a',0,'M','a','l','a','y','a','l','a','m','_','N','y','a',0,
2,'M','a','l','a','y','a','l','a','m','_','R','a',0,'M','a','l','a','y','a','l','a','m','_','R','a',0,
2,'M','a','l','a','y','a','l','a','m','_','S','s','a',0,'M','a','l','a','y','a','l','a','m','_','S','s','a',0,
2,'M','a','l','a','y','a','l','a','m','_','T','t','a',0,'M','a','l','a','y','a','l','a','m','_','T','t','a',0,
2,'H','a','n','i','f','i','_','R','o','h','i','n','g','y','a','_','K','i','n','n','a','_','Y','a',0,
'H','a','n','i','f','i','_','R','o','h','i','n','g','y','a','_','K','i','n','n','a','_','Y','a',0,
2,'H','a','n','i','f','i','_','R','o','h','i','n','g','y','a','_','P','a',0,'H','a','n','i','f','i','_','R','o','h','i','n',
'g','y','a','_','P','a',0,2,'j','t',0,'J','o','i','n','i','n','g','_','T','y','p','e',0,
2,'U',0,'N','o','n','_','J','o','i','n','i','n','g',0,2,'C',0,'J','o','i','n','_','C','a','u','s','i','n','g',0,
2,'D',0,'D','u','a','l','_','J','o','i','n','i','n','g',0,
2,'L',0,'L','e','f','t','_','J','o','i','n','i','n','g',0,
2,'R',0,'R','i','g','h','t','_','J','o','i','n','i','n','g',0,
2,'T',0,'T','r','a','n','s','p','a','r','e','n','t',0,2,'l','b',0,'L','i','n','e','_','B','r','e','a','k',0,
2,'X','X',0,'U','n','k','n','o','w','n',0,2,'A','I',0,'A','m','b','i','g','u','o','u','s',0,
2,'A','L',0,'A','l','p','h','a','b','e','t','i','c',0,2,'B','2',0,'B','r','e','a','k','_','B','o','t','h',0,
2,'B','A',0,'B','r','e','a','k','_','A','f','t','e','r',0,
2,'B','B',0,'B','r','e','a','k','_','B','e','f','o','r','e',0,
2,'B','K',0,'M','a','n','d','a','t','o','r','y','_','B','r','e','a','k',0,
2,'C','B',0,'C','o','n','t','i','n','g','e','n','t','_','B','r','e','a','k',0,
2,'C','L',0,'C','l','o','s','e','_','P','u','n','c','t','u','a','t','i','o','n',0,
2,'C','M',0,'C','o','m','b','i','n','i','n','g','_','M','a','r','k',0,
2,'C','R',0,'C','a','r','r','i','a','g','e','_','R','e','t','u','r','n',0,
2,'E','X',0,'E','x','c','l','a','m','a','t','i','o','n',0,
2,'G','L',0,'G','l','u','e',0,2,'H','Y',0,'H','y','p','h','e','n',0,
2,'I','D',0,'I','d','e','o','g','r','a','p','h','i','c',0,
3,'I','N',0,'I','n','s','e','p','a','r','a','b','l','e',0,'I','n','s','e','p','e','r','a','b','l','e',0,
2,'I','S',0,'I','n','f','i','x','_','N','u','m','e','r','i','c',0,
2,'L','F',0,'L','i','n','e','_','F','e','e','d',0,2,'N','S',0,'N','o','n','s','t','a','r','t','e','r',0,
2,'N','U',0,'N','u','m','e','r','i','c',0,2,'O','P',0,'O','p','e','n','_','P','u','n','c','t','u','a','t','i','o','n',
0,2,'P','O',0,'P','o','s','t','f','i','x','_','N','u','m','e','r','i','c',0,
2,'P','R',0,'P','r','e','f','i','x','_','N','u','m','e','r','i','c',0,
2,'Q','U',0,'Q','u','o','t','a','t','i','o','n',0,2,'S','A',0,'C','o','m','p','l','e','x','_','C','o','n','t','e','x',
't',0,2,'S','G',0,'S','u','r','r','o','g','a','t','e',0,
2,'S','P',0,'S','p','a','c','e',0,2,'S','Y',0,'B','r','e','a','k','_','S','y','m','b','o','l','s',0,
2,'Z','W',0,'Z','W','S','p','a','c','e',0,2,'N','L',0,'N','e','x','t','_','L','i','n','e',0,
2,'W','J',0,'W','o','r','d','_','J','o','i','n','e','r',0,
2,'H','2',0,'H','2',0,2,'H','3',0,'H','3',0,2,'J','L',0,'J','L',0,
2,'J','T',0,'J','T',0,2,'J','V',0,'J','V',0,2,'C','P',0,'C','l','o','s','e','_','P','a','r','e','n','t','h','e',
's','i','s',0,2,'C','J',0,'C','o','n','d','i','t','i','o','n','a','l','_','J','a','p','a','n','e','s','e','_','S','t','a',
'r','t','e','r',0,2,'H','L',0,'H','e','b','r','e','w','_','L','e','t','t','e','r',0,
2,'E','B',0,'E','_','B','a','s','e',0,2,'E','M',0,'E','_','M','o','d','i','f','i','e','r',0,
2,'Z','W','J',0,'Z','W','J',0,2,'n','t',0,'N','u','m','e','r','i','c','_','T','y','p','e',0,
2,'N','o','n','e',0,'N','o','n','e',0,2,'D','e',0,'D','e','c','i','m','a','l',0,
2,'D','i',0,'D','i','g','i','t',0,2,'N','u',0,'N','u','m','e','r','i','c',0,
2,'s','c',0,'S','c','r','i','p','t',0,2,'Z','y','y','y',0,'C','o','m','m','o','n',0,
3,'Z','i','n','h',0,'I','n','h','e','r','i','t','e','d',0,'Q','a','a','i',0,
2,'A','r','a','b',0,'A','r','a','b','i','c',0,2,'A','r','m','n',0,'A','r','m','e','n','i','a','n',0,
2,'B','e','n','g',0,'B','e','n','g','a','l','i',0,2,'B','o','p','o',0,'B','o','p','o','m','o','f','o',0,
2,'C','h','e','r',0,'C','h','e','r','o','k','e','e',0,3,'C','o','p','t',0,'C','o','p','t','i','c',0,
'Q','a','a','c',0,2,'C','y','r','l',0,'C','y','r','i','l','l','i','c',0,
2,'D','s','r','t',0,'D','e','s','e','r','e','t',0,2,'D','e','v','a',0,'D','e','v','a','n','a','g','a','r','i',0,
2,'E','t','h','i',0,'E','t','h','i','o','p','i','c',0,2,'G','e','o','r',0,'G','e','o','r','g','i','a','n',0,
2,'G','o','t','h',0,'G','o','t','h','i','c',0,2,'G','r','e','k',0,'G','r','e','e','k',0,
2,'G','u','j','r',0,'G','u','j','a','r','a','t','i',0,2,'G','u','r','u',0,'G','u','r','m','u','k','h','i',0,
2,'H','a','n','i',0,'H','a','n',0,2,'H','a','n','g',0,'H','a','n','g','u','l',0,
2,'H','e','b','r',0,'H','e','b','r','e','w',0,2,'H','i','r','a',0,'H','i','r','a','g','a','n','a',0,
2,'K','n','d','a',0,'K','a','n','n','a','d','a',0,2,'K','a','n','a',0,'K','a','t','a','k','a','n','a',0,
2,'K','h','m','r',0,'K','h','m','e','r',0,2,'L','a','o','o',0,'L','a','o',0,
2,'L','a','t','n',0,'L','a','t','i','n',0,2,'M','l','y','m',0,'M','a','l','a','y','a','l','a','m',0,
2,'M','o','n','g',0,'M','o','n','g','o','l','i','a','n',0,
2,'M','y','m','r',0,'M','y','a','n','m','a','r',0,2,'O','g','a','m',0,'O','g','h','a','m',0,
2,'I','t','a','l',0,'O','l','d','_','I','t','a','l','i','c',0,
2,'O','r','y','a',0,'O','r','i','y','a',0,2,'R','u','n','r',0,'R','u','n','i','c',0,
2,'S','i','n','h',0,'S','i','n','h','a','l','a',0,2,'S','y','r','c',0,'S','y','r','i','a','c',0,
2,'T','a','m','l',0,'T','a','m','i','l',0,2,'T','e','l','u',0,'T','e','l','u','g','u',0,
2,'T','h','a','a',0,'T','h','a','a','n','a',0,2,'T','i','b','t',0,'T','i','b','e','t','a','n',0,
2,'C','a','n','s',0,'C','a','n','a','d','i','a','n','_','A','b','o','r','i','g','i','n','a','l',0,
2,'Y','i','i','i',0,'Y','i',0,2,'T','g','l','g',0,'T','a','g','a','l','o','g',0,
2,'H','a','n','o',0,'H','a','n','u','n','o','o',0,2,'B','u','h','d',0,'B','u','h','i','d',0,
2,'T','a','g','b',0,'T','a','g','b','a','n','w','a',0,2,'B','r','a','i',0,'B','r','a','i','l','l','e',0,
2,'C','p','r','t',0,'C','y','p','r','i','o','t',0,2,'L','i','m','b',0,'L','i','m','b','u',0,
2,'L','i','n','b',0,'L','i','n','e','a','r','_','B',0,2,'O','s','m','a',0,'O','s','m','a','n','y','a',0,
2,'S','h','a','w',0,'S','h','a','v','i','a','n',0,2,'T','a','l','e',0,'T','a','i','_','L','e',0,
2,'U','g','a','r',0,'U','g','a','r','i','t','i','c',0,2,'H','r','k','t',0,'K','a','t','a','k','a','n','a','_','O','r',
'_','H','i','r','a','g','a','n','a',0,2,'B','u','g','i',0,'B','u','g','i','n','e','s','e',0,
2,'G','l','a','g',0,'G','l','a','g','o','l','i','t','i','c',0,
2,'K','h','a','r',0,'K','h','a','r','o','s','h','t','h','i',0,
2,'S','y','l','o',0,'S','y','l','o','t','i','_','N','a','g','r','i',0,
2,'T','a','l','u',0,'N','e','w','_','T','a','i','_','L','u','e',0,
2,'T','f','n','g',0,'T','i','f','i','n','a','g','h',0,2,'X','p','e','o',0,'O','l','d','_','P','e','r','s','i','a','n',
0,2,'B','a','l','i',0,'B','a','l','i','n','e','s','e',0,
2,'B','a','t','k',0,'B','a','t','a','k',0,2,'B','l','i','s',0,'B','l','i','s',0,
2,'B','r','a','h',0,'B','r','a','h','m','i',0,2,'C','i','r','t',0,'C','i','r','t',0,
2,'C','y','r','s',0,'C','y','r','s',0,2,'E','g','y','d',0,'E','g','y','d',0,
2,'E','g','y','h',0,'E','g','y','h',0,2,'E','g','y','p',0,'E','g','y','p','t','i','a','n','_','H','i','e','r','o','g',
'l','y','p','h','s',0,2,'G','e','o','k',0,'G','e','o','k',0,
2,'H','a','n','s',0,'H','a','n','s',0,2,'H','a','n','t',0,'H','a','n','t',0,
2,'H','m','n','g',0,'P','a','h','a','w','h','_','H','m','o','n','g',0,
2,'H','u','n','g',0,'O','l','d','_','H','u','n','g','a','r','i','a','n',0,
2,'I','n','d','s',0,'I','n','d','s',0,2,'J','a','v','a',0,'J','a','v','a','n','e','s','e',0,
2,'K','a','l','i',0,'K','a','y','a','h','_','L','i',0,2,'L','a','t','f',0,'L','a','t','f',0,
2,'L','a','t','g',0,'L','a','t','g',0,2,'L','e','p','c',0,'L','e','p','c','h','a',0,
2,'L','i','n','a',0,'L','i','n','e','a','r','_','A',0,2,'M','a','n','d',0,'M','a','n','d','a','i','c',0,
2,'M','a','y','a',0,'M','a','y','a',0,2,'M','e','r','o',0,'M','e','r','o','i','t','i','c','_','H','i','e','r','o','g',
'l','y','p','h','s',0,2,'N','k','o','o',0,'N','k','o',0,
2,'O','r','k','h',0,'O','l','d','_','T','u','r','k','i','c',0,
2,'P','e','r','m',0,'O','l','d','_','P','e','r','m','i','c',0,
2,'P','h','a','g',0,'P','h','a','g','s','_','P','a',0,2,'P','h','n','x',0,'P','h','o','e','n','i','c','i','a','n',0,
2,'P','l','r','d',0,'M','i','a','o',0,2,'R','o','r','o',0,'R','o','r','o',0,
2,'S','a','r','a',0,'S','a','r','a',0,2,'S','y','r','e',0,'S','y','r','e',0,
2,'S','y','r','j',0,'S','y','r','j',0,2,'S','y','r','n',0,'S','y','r','n',0,
2,'T','e','n','g',0,'T','e','n','g',0,2,'V','a','i','i',0,'V','a','i',0,
2,'V','i','s','p',0,'V','i','s','p',0,2,'X','s','u','x',0,'C','u','n','e','i','f','o','r','m',0,
2,'Z','x','x','x',0,'Z','x','x','x',0,2,'Z','z','z','z',0,'U','n','k','n','o','w','n',0,
2,'C','a','r','i',0,'C','a','r','i','a','n',0,2,'J','p','a','n',0,'J','p','a','n',0,
2,'L','a','n','a',0,'T','a','i','_','T','h','a','m',0,2,'L','y','c','i',0,'L','y','c','i','a','n',0,
2,'L','y','d','i',0,'L','y','d','i','a','n',0,2,'O','l','c','k',0,'O','l','_','C','h','i','k','i',0,
2,'R','j','n','g',0,'R','e','j','a','n','g',0,2,'S','a','u','r',0,'S','a','u','r','a','s','h','t','r','a',0,
2,'S','g','n','w',0,'S','i','g','n','W','r','i','t','i','n','g',0,
2,'S','u','n','d',0,'S','u','n','d','a','n','e','s','e',0,
2,'M','o','o','n',0,'M','o','o','n',0,2,'M','t','e','i',0,'M','e','e','t','e','i','_','M','a','y','e','k',0,
2,'A','r','m','i',0,'I','m','p','e','r','i','a','l','_','A','r','a','m','a','i','c',0,
2,'A','v','s','t',0,'A','v','e','s','t','a','n',0,2,'C','a','k','m',0,'C','h','a','k','m','a',0,
2,'K','o','r','e',0,'K','o','r','e',0,2,'K','t','h','i',0,'K','a','i','t','h','i',0,
2,'M','a','n','i',0,'M','a','n','i','c','h','a','e','a','n',0,
2,'P','h','l','i',0,'I','n','s','c','r','i','p','t','i','o','n','a','l','_','P','a','h','l','a','v','i',0,
2,'P','h','l','p',0,'P','s','a','l','t','e','r','_','P','a','h','l','a','v','i',0,
2,'P','h','l','v',0,'P','h','l','v',0,2,'P','r','t','i',0,'I','n','s','c','r','i','p','t','i','o','n','a','l','_','P',
'a','r','t','h','i','a','n',0,2,'S','a','m','r',0,'S','a','m','a','r','i','t','a','n',0,
2,'T','a','v','t',0,'T','a','i','_','V','i','e','t',0,2,'Z','m','t','h',0,'Z','m','t','h',0,
2,'Z','s','y','m',0,'Z','s','y','m',0,2,'B','a','m','u',0,'B','a','m','u','m',0,
2,'N','k','g','b',0,'N','k','g','b',0,2,'S','a','r','b',0,'O','l','d','_','S','o','u','t','h','_','A','r','a','b','i',
'a','n',0,2,'B','a','s','s',0,'B','a','s','s','a','_','V','a','h',0,
2,'D','u','p','l',0,'D','u','p','l','o','y','a','n',0,2,'E','l','b','a',0,'E','l','b','a','s','a','n',0,
2,'G','r','a','n',0,'G','r','a','n','t','h','a',0,2,'K','p','e','l',0,'K','p','e','l',0,
2,'L','o','m','a',0,'L','o','m','a',0,2,'M','e','n','d',0,'M','e','n','d','e','_','K','i','k','a','k','u','i',0,
2,'M','e','r','c',0,'M','e','r','o','i','t','i','c','_','C','u','r','s','i','v','e',0,
2,'N','a','r','b',0,'O','l','d','_','N','o','r','t','h','_','A','r','a','b','i','a','n',0,
2,'N','b','a','t',0,'N','a','b','a','t','a','e','a','n',0,
2,'P','a','l','m',0,'P','a','l','m','y','r','e','n','e',0,
2,'S','i','n','d',0,'K','h','u','d','a','w','a','d','i',0,
2,'W','a','r','a',0,'W','a','r','a','n','g','_','C','i','t','i',0,
2,'A','f','a','k',0,'A','f','a','k',0,2,'J','u','r','c',0,'J','u','r','c',0,
2,'M','r','o','o',0,'M','r','o',0,2,'N','s','h','u',0,'N','u','s','h','u',0,
2,'S','h','r','d',0,'S','h','a','r','a','d','a',0,2,'S','o','r','a',0,'S','o','r','a','_','S','o','m','p','e','n','g',
0,2,'T','a','k','r',0,'T','a','k','r','i',0,2,'T','a','n','g',0,'T','a','n','g','u','t',0,
2,'W','o','l','e',0,'W','o','l','e',0,2,'H','l','u','w',0,'A','n','a','t','o','l','i','a','n','_','H','i','e','r','o',
'g','l','y','p','h','s',0,2,'K','h','o','j',0,'K','h','o','j','k','i',0,
2,'T','i','r','h',0,'T','i','r','h','u','t','a',0,2,'A','g','h','b',0,'C','a','u','c','a','s','i','a','n','_','A','l',
'b','a','n','i','a','n',0,2,'M','a','h','j',0,'M','a','h','a','j','a','n','i',0,
2,'H','a','t','r',0,'H','a','t','r','a','n',0,2,'M','u','l','t',0,'M','u','l','t','a','n','i',0,
2,'P','a','u','c',0,'P','a','u','_','C','i','n','_','H','a','u',0,
2,'S','i','d','d',0,'S','i','d','d','h','a','m',0,2,'A','d','l','m',0,'A','d','l','a','m',0,
2,'B','h','k','s',0,'B','h','a','i','k','s','u','k','i',0,
2,'M','a','r','c',0,'M','a','r','c','h','e','n',0,2,'O','s','g','e',0,'O','s','a','g','e',0,
2,'H','a','n','b',0,'H','a','n','b',0,2,'J','a','m','o',0,'J','a','m','o',0,
2,'Z','s','y','e',0,'Z','s','y','e',0,2,'G','o','n','m',0,'M','a','s','a','r','a','m','_','G','o','n','d','i',0,
2,'S','o','y','o',0,'S','o','y','o','m','b','o',0,2,'Z','a','n','b',0,'Z','a','n','a','b','a','z','a','r','_','S','q',
'u','a','r','e',0,2,'D','o','g','r',0,'D','o','g','r','a',0,
2,'G','o','n','g',0,'G','u','n','j','a','l','a','_','G','o','n','d','i',0,
2,'M','a','k','a',0,'M','a','k','a','s','a','r',0,2,'M','e','d','f',0,'M','e','d','e','f','a','i','d','r','i','n',0,
2,'R','o','h','g',0,'H','a','n','i','f','i','_','R','o','h','i','n','g','y','a',0,
2,'S','o','g','d',0,'S','o','g','d','i','a','n',0,2,'S','o','g','o',0,'O','l','d','_','S','o','g','d','i','a','n',0,
2,'E','l','y','m',0,'E','l','y','m','a','i','c',0,2,'H','m','n','p',0,'N','y','i','a','k','e','n','g','_','P','u','a',
'c','h','u','e','_','H','m','o','n','g',0,2,'N','a','n','d',0,'N','a','n','d','i','n','a','g','a','r','i',0,
2,'W','c','h','o',0,'W','a','n','c','h','o',0,2,'C','h','r','s',0,'C','h','o','r','a','s','m','i','a','n',0,
2,'D','i','a','k',0,'D','i','v','e','s','_','A','k','u','r','u',0,
2,'K','i','t','s',0,'K','h','i','t','a','n','_','S','m','a','l','l','_','S','c','r','i','p','t',0,
2,'Y','e','z','i',0,'Y','e','z','i','d','i',0,2,'h','s','t',0,'H','a','n','g','u','l','_','S','y','l','l','a','b','l',
'e','_','T','y','p','e',0,2,'N','A',0,'N','o','t','_','A','p','p','l','i','c','a','b','l','e',0,
2,'L',0,'L','e','a','d','i','n','g','_','J','a','m','o',0,
2,'V',0,'V','o','w','e','l','_','J','a','m','o',0,2,'T',0,'T','r','a','i','l','i','n','g','_','J','a','m','o',0,
2,'L','V',0,'L','V','_','S','y','l','l','a','b','l','e',0,
2,'L','V','T',0,'L','V','T','_','S','y','l','l','a','b','l','e',0,
2,'N','F','D','_','Q','C',0,'N','F','D','_','Q','u','i','c','k','_','C','h','e','c','k',0,
2,'N',0,'N','o',0,2,'Y',0,'Y','e','s',0,2,'N','F','K','D','_','Q','C',0,'N','F','K','D','_','Q','u','i','c','k',
'_','C','h','e','c','k',0,2,'N','F','C','_','Q','C',0,'N','F','C','_','Q','u','i','c','k','_','C','h','e','c','k',0,
2,'M',0,'M','a','y','b','e',0,2,'N','F','K','C','_','Q','C',0,'N','F','K','C','_','Q','u','i','c','k','_','C','h','e',
'c','k',0,2,'l','c','c','c',0,'L','e','a','d','_','C','a','n','o','n','i','c','a','l','_','C','o','m','b','i','n','i','n',
'g','_','C','l','a','s','s',0,2,'t','c','c','c',0,'T','r','a','i','l','_','C','a','n','o','n','i','c','a','l','_','C','o',
'm','b','i','n','i','n','g','_','C','l','a','s','s',0,2,'G','C','B',0,'G','r','a','p','h','e','m','e','_','C','l','u','s',
't','e','r','_','B','r','e','a','k',0,2,'X','X',0,'O','t','h','e','r',0,
2,'C','N',0,'C','o','n','t','r','o','l',0,2,'C','R',0,'C','R',0,
2,'E','X',0,'E','x','t','e','n','d',0,2,'L',0,'L',0,
2,'L','F',0,'L','F',0,2,'L','V',0,'L','V',0,2,'L','V','T',0,'L','V','T',0,
2,'T',0,'T',0,2,'V',0,'V',0,2,'S','M',0,'S','p','a','c','i','n','g','M','a','r','k',0,
2,'P','P',0,'P','r','e','p','e','n','d',0,2,'E','B','G',0,'E','_','B','a','s','e','_','G','A','Z',0,
2,'G','A','Z',0,'G','l','u','e','_','A','f','t','e','r','_','Z','w','j',0,
2,'S','B',0,'S','e','n','t','e','n','c','e','_','B','r','e','a','k',0,
2,'A','T',0,'A','T','e','r','m',0,2,'C','L',0,'C','l','o','s','e',0,
2,'F','O',0,'F','o','r','m','a','t',0,2,'L','O',0,'L','o','w','e','r',0,
2,'L','E',0,'O','L','e','t','t','e','r',0,2,'S','E',0,'S','e','p',0,
2,'S','P',0,'S','p',0,2,'S','T',0,'S','T','e','r','m',0,
2,'U','P',0,'U','p','p','e','r',0,2,'S','C',0,'S','C','o','n','t','i','n','u','e',0,
2,'W','B',0,'W','o','r','d','_','B','r','e','a','k',0,2,'L','E',0,'A','L','e','t','t','e','r',0,
2,'K','A',0,'K','a','t','a','k','a','n','a',0,2,'M','L',0,'M','i','d','L','e','t','t','e','r',0,
2,'M','N',0,'M','i','d','N','u','m',0,2,'E','X',0,'E','x','t','e','n','d','N','u','m','L','e','t',0,
2,'E','x','t','e','n','d',0,'E','x','t','e','n','d',0,2,'M','B',0,'M','i','d','N','u','m','L','e','t',0,
2,'N','L',0,'N','e','w','l','i','n','e',0,2,'S','Q',0,'S','i','n','g','l','e','_','Q','u','o','t','e',0,
2,'D','Q',0,'D','o','u','b','l','e','_','Q','u','o','t','e',0,
2,'W','S','e','g','S','p','a','c','e',0,'W','S','e','g','S','p','a','c','e',0,
2,'b','p','t',0,'B','i','d','i','_','P','a','i','r','e','d','_','B','r','a','c','k','e','t','_','T','y','p','e',0,
2,'n',0,'N','o','n','e',0,2,'o',0,'O','p','e','n',0,
2,'c',0,'C','l','o','s','e',0,2,'I','n','P','C',0,'I','n','d','i','c','_','P','o','s','i','t','i','o','n','a','l','_',
'C','a','t','e','g','o','r','y',0,2,'N','A',0,'N','A',0,
2,'B','o','t','t','o','m',0,'B','o','t','t','o','m',0,2,'B','o','t','t','o','m','_','A','n','d','_','L','e','f','t',0,
'B','o','t','t','o','m','_','A','n','d','_','L','e','f','t',0,
2,'B','o','t','t','o','m','_','A','n','d','_','R','i','g','h','t',0,'B','o','t','t','o','m','_','A','n','d','_','R','i','g',
'h','t',0,2,'L','e','f','t',0,'L','e','f','t',0,2,'L','e','f','t','_','A','n','d','_','R','i','g','h','t',0,
'L','e','f','t','_','A','n','d','_','R','i','g','h','t',0,2,'O','v','e','r','s','t','r','u','c','k',0,
'O','v','e','r','s','t','r','u','c','k',0,2,'R','i','g','h','t',0,'R','i','g','h','t',0,
2,'T','o','p',0,'T','o','p',0,2,'T','o','p','_','A','n','d','_','B','o','t','t','o','m',0,
'T','o','p','_','A','n','d','_','B','o','t','t','o','m',0,2,'T','o','p','_','A','n','d','_','B','o','t','t','o','m','_','A',
'n','d','_','R','i','g','h','t',0,'T','o','p','_','A','n','d','_','B','o','t','t','o','m','_','A','n','d','_','R','i','g','h',
't',0,2,'T','o','p','_','A','n','d','_','L','e','f','t',0,'T','o','p','_','A','n','d','_','L','e','f','t',0,
2,'T','o','p','_','A','n','d','_','L','e','f','t','_','A','n','d','_','R','i','g','h','t',0,
'T','o','p','_','A','n','d','_','L','e','f','t','_','A','n','d','_','R','i','g','h','t',0,
2,'T','o','p','_','A','n','d','_','R','i','g','h','t',0,'T','o','p','_','A','n','d','_','R','i','g','h','t',0,
2,'V','i','s','u','a','l','_','O','r','d','e','r','_','L','e','f','t',0,'V','i','s','u','a','l','_','O','r','d','e','r','_',
'L','e','f','t',0,2,'T','o','p','_','A','n','d','_','B','o','t','t','o','m','_','A','n','d','_','L','e','f','t',0,
'T','o','p','_','A','n','d','_','B','o','t','t','o','m','_','A','n','d','_','L','e','f','t',0,
2,'I','n','S','C',0,'I','n','d','i','c','_','S','y','l','l','a','b','i','c','_','C','a','t','e','g','o','r','y',0,
2,'O','t','h','e','r',0,'O','t','h','e','r',0,2,'A','v','a','g','r','a','h','a',0,'A','v','a','g','r','a','h','a',0,
2,'B','i','n','d','u',0,'B','i','n','d','u',0,2,'B','r','a','h','m','i','_','J','o','i','n','i','n','g','_','N','u','m',
'b','e','r',0,'B','r','a','h','m','i','_','J','o','i','n','i','n','g','_','N','u','m','b','e','r',0,
2,'C','a','n','t','i','l','l','a','t','i','o','n','_','M','a','r','k',0,'C','a','n','t','i','l','l','a','t','i','o','n','_',
'M','a','r','k',0,2,'C','o','n','s','o','n','a','n','t',0,'C','o','n','s','o','n','a','n','t',0,
2,'C','o','n','s','o','n','a','n','t','_','D','e','a','d',0,'C','o','n','s','o','n','a','n','t','_','D','e','a','d',0,
2,'C','o','n','s','o','n','a','n','t','_','F','i','n','a','l',0,'C','o','n','s','o','n','a','n','t','_','F','i','n','a','l',
0,2,'C','o','n','s','o','n','a','n','t','_','H','e','a','d','_','L','e','t','t','e','r',0,
'C','o','n','s','o','n','a','n','t','_','H','e','a','d','_','L','e','t','t','e','r',0,
2,'C','o','n','s','o','n','a','n','t','_','I','n','i','t','i','a','l','_','P','o','s','t','f','i','x','e','d',0,
'C','o','n','s','o','n','a','n','t','_','I','n','i','t','i','a','l','_','P','o','s','t','f','i','x','e','d',0,
2,'C','o','n','s','o','n','a','n','t','_','K','i','l','l','e','r',0,'C','o','n','s','o','n','a','n','t','_','K','i','l','l',
'e','r',0,2,'C','o','n','s','o','n','a','n','t','_','M','e','d','i','a','l',0,'C','o','n','s','o','n','a','n','t','_','M',
'e','d','i','a','l',0,2,'C','o','n','s','o','n','a','n','t','_','P','l','a','c','e','h','o','l','d','e','r',0,
'C','o','n','s','o','n','a','n','t','_','P','l','a','c','e','h','o','l','d','e','r',0,
2,'C','o','n','s','o','n','a','n','t','_','P','r','e','c','e','d','i','n','g','_','R','e','p','h','a',0,
'C','o','n','s','o','n','a','n','t','_','P','r','e','c','e','d','i','n','g','_','R','e','p','h','a',0,
2,'C','o','n','s','o','n','a','n','t','_','P','r','e','f','i','x','e','d',0,'C','o','n','s','o','n','a','n','t','_','P','r',
'e','f','i','x','e','d',0,2,'C','o','n','s','o','n','a','n','t','_','S','u','b','j','o','i','n','e','d',0,
'C','o','n','s','o','n','a','n','t','_','S','u','b','j','o','i','n','e','d',0,
2,'C','o','n','s','o','n','a','n','t','_','S','u','c','c','e','e','d','i','n','g','_','R','e','p','h','a',0,
'C','o','n','s','o','n','a','n','t','_','S','u','c','c','e','e','d','i','n','g','_','R','e','p','h','a',0,
2,'C','o','n','s','o','n','a','n','t','_','W','i','t','h','_','S','t','a','c','k','e','r',0,
'C','o','n','s','o','n','a','n','t','_','W','i','t','h','_','S','t','a','c','k','e','r',0,
2,'G','e','m','i','n','a','t','i','o','n','_','M','a','r','k',0,'G','e','m','i','n','a','t','i','o','n','_','M','a','r','k',
0,2,'I','n','v','i','s','i','b','l','e','_','S','t','a','c','k','e','r',0,'I','n','v','i','s','i','b','l','e','_','S','t',
'a','c','k','e','r',0,2,'J','o','i','n','e','r',0,'J','o','i','n','e','r',0,
2,'M','o','d','i','f','y','i','n','g','_','L','e','t','t','e','r',0,'M','o','d','i','f','y','i','n','g','_','L','e','t','t',
'e','r',0,2,'N','o','n','_','J','o','i','n','e','r',0,'N','o','n','_','J','o','i','n','e','r',0,
2,'N','u','k','t','a',0,'N','u','k','t','a',0,2,'N','u','m','b','e','r',0,'N','u','m','b','e','r',0,
2,'N','u','m','b','e','r','_','J','o','i','n','e','r',0,'N','u','m','b','e','r','_','J','o','i','n','e','r',0,
2,'P','u','r','e','_','K','i','l','l','e','r',0,'P','u','r','e','_','K','i','l','l','e','r',0,
2,'R','e','g','i','s','t','e','r','_','S','h','i','f','t','e','r',0,'R','e','g','i','s','t','e','r','_','S','h','i','f','t',
'e','r',0,2,'S','y','l','l','a','b','l','e','_','M','o','d','i','f','i','e','r',0,'S','y','l','l','a','b','l','e','_','M',
'o','d','i','f','i','e','r',0,2,'T','o','n','e','_','L','e','t','t','e','r',0,'T','o','n','e','_','L','e','t','t','e','r',
0,2,'T','o','n','e','_','M','a','r','k',0,'T','o','n','e','_','M','a','r','k',0,
2,'V','i','r','a','m','a',0,'V','i','r','a','m','a',0,2,'V','i','s','a','r','g','a',0,
'V','i','s','a','r','g','a',0,2,'V','o','w','e','l',0,'V','o','w','e','l',0,
2,'V','o','w','e','l','_','D','e','p','e','n','d','e','n','t',0,'V','o','w','e','l','_','D','e','p','e','n','d','e','n','t',
0,2,'V','o','w','e','l','_','I','n','d','e','p','e','n','d','e','n','t',0,'V','o','w','e','l','_','I','n','d','e','p','e',
'n','d','e','n','t',0,2,'v','o',0,'V','e','r','t','i','c','a','l','_','O','r','i','e','n','t','a','t','i','o','n',0,
2,'R',0,'R','o','t','a','t','e','d',0,2,'T','r',0,'T','r','a','n','s','f','o','r','m','e','d','_','R','o','t','a','t',
'e','d',0,2,'T','u',0,'T','r','a','n','s','f','o','r','m','e','d','_','U','p','r','i','g','h','t',0,
2,'U',0,'U','p','r','i','g','h','t',0,2,'g','c','m',0,'G','e','n','e','r','a','l','_','C','a','t','e','g','o','r','y',
'_','M','a','s','k',0,2,'C',0,'O','t','h','e','r',0,2,'L',0,'L','e','t','t','e','r',0,
2,'L','C',0,'C','a','s','e','d','_','L','e','t','t','e','r',0,
3,'M',0,'M','a','r','k',0,'C','o','m','b','i','n','i','n','g','_','M','a','r','k',0,
2,'N',0,'N','u','m','b','e','r',0,3,'P',0,'P','u','n','c','t','u','a','t','i','o','n',0,
'p','u','n','c','t',0,2,'S',0,'S','y','m','b','o','l',0,
2,'Z',0,'S','e','p','a','r','a','t','o','r',0,2,'n','v',0,'N','u','m','e','r','i','c','_','V','a','l','u','e',0,
2,'a','g','e',0,'A','g','e',0,2,'b','m','g',0,'B','i','d','i','_','M','i','r','r','o','r','i','n','g','_','G','l','y',
'p','h',0,2,'c','f',0,'C','a','s','e','_','F','o','l','d','i','n','g',0,
2,'i','s','c',0,'I','S','O','_','C','o','m','m','e','n','t',0,
2,'l','c',0,'L','o','w','e','r','c','a','s','e','_','M','a','p','p','i','n','g',0,
2,'n','a',0,'N','a','m','e',0,3,'s','c','f',0,'S','i','m','p','l','e','_','C','a','s','e','_','F','o','l','d','i','n',
'g',0,'s','f','c',0,2,'s','l','c',0,'S','i','m','p','l','e','_','L','o','w','e','r','c','a','s','e','_','M','a','p','p',
'i','n','g',0,2,'s','t','c',0,'S','i','m','p','l','e','_','T','i','t','l','e','c','a','s','e','_','M','a','p','p','i','n',
'g',0,2,'s','u','c',0,'S','i','m','p','l','e','_','U','p','p','e','r','c','a','s','e','_','M','a','p','p','i','n','g',0,
2,'t','c',0,'T','i','t','l','e','c','a','s','e','_','M','a','p','p','i','n','g',0,
2,'n','a','1',0,'U','n','i','c','o','d','e','_','1','_','N','a','m','e',0,
2,'u','c',0,'U','p','p','e','r','c','a','s','e','_','M','a','p','p','i','n','g',0,
2,'b','p','b',0,'B','i','d','i','_','P','a','i','r','e','d','_','B','r','a','c','k','e','t',0,
2,'s','c','x',0,'S','c','r','i','p','t','_','E','x','t','e','n','s','i','o','n','s',0
};

U_NAMESPACE_END

#endif  // INCLUDED_FROM_PROPNAME_CPP


U_CDECL_BEGIN

/**
 * Get the next non-ignorable ASCII character from a property name
 * and lowercases it.
 * @return ((advance count for the name)<<8)|character
 */
static inline int32_t
getASCIIPropertyNameChar(const char *name) {
    int32_t i;
    char c;

    /* Ignore delimiters '-', '_', and ASCII White_Space */
    for(i=0;
        (c=name[i++])==0x2d || c==0x5f ||
        c==0x20 || (0x09<=c && c<=0x0d);
    ) {}

    if(c!=0) {
        return (i<<8)|(uint8_t)uprv_asciitolower((char)c);
    } else {
        return i<<8;
    }
}

/**
 * Get the next non-ignorable EBCDIC character from a property name
 * and lowercases it.
 * @return ((advance count for the name)<<8)|character
 */
static inline int32_t
getEBCDICPropertyNameChar(const char *name) {
    int32_t i;
    char c;

    /* Ignore delimiters '-', '_', and EBCDIC White_Space */
    for(i=0;
        (c=name[i++])==0x60 || c==0x6d ||
        c==0x40 || c==0x05 || c==0x15 || c==0x25 || c==0x0b || c==0x0c || c==0x0d;
    ) {}

    if(c!=0) {
        return (i<<8)|(uint8_t)uprv_ebcdictolower((char)c);
    } else {
        return i<<8;
    }
}

/**
 * Unicode property names and property value names are compared "loosely".
 *
 * UCD.html 4.0.1 says:
 *   For all property names, property value names, and for property values for
 *   Enumerated, Binary, or Catalog properties, use the following
 *   loose matching rule:
 *
 *   LM3. Ignore case, whitespace, underscore ('_'), and hyphens.
 *
 * This function does just that, for (char *) name strings.
 * It is almost identical to ucnv_compareNames() but also ignores
 * C0 White_Space characters (U+0009..U+000d, and U+0085 on EBCDIC).
 *
 * @internal
 */

U_CAPI int32_t U_EXPORT2
uprv_compareASCIIPropertyNames(const char *name1, const char *name2) {
    int32_t rc, r1, r2;

    for(;;) {
        r1=getASCIIPropertyNameChar(name1);
        r2=getASCIIPropertyNameChar(name2);

        /* If we reach the ends of both strings then they match */
        if(((r1|r2)&0xff)==0) {
            return 0;
        }

        /* Compare the lowercased characters */
        if(r1!=r2) {
            rc=(r1&0xff)-(r2&0xff);
            if(rc!=0) {
                return rc;
            }
        }

        name1+=r1>>8;
        name2+=r2>>8;
    }
}

U_CAPI int32_t U_EXPORT2
uprv_compareEBCDICPropertyNames(const char *name1, const char *name2) {
    int32_t rc, r1, r2;

    for(;;) {
        r1=getEBCDICPropertyNameChar(name1);
        r2=getEBCDICPropertyNameChar(name2);

        /* If we reach the ends of both strings then they match */
        if(((r1|r2)&0xff)==0) {
            return 0;
        }

        /* Compare the lowercased characters */
        if(r1!=r2) {
            rc=(r1&0xff)-(r2&0xff);
            if(rc!=0) {
                return rc;
            }
        }

        name1+=r1>>8;
        name2+=r2>>8;
    }
}

U_CDECL_END

U_NAMESPACE_BEGIN

int32_t PropNameData::findProperty(int32_t property) {
    int32_t i=1;  // valueMaps index, initially after numRanges
    for(int32_t numRanges=valueMaps[0]; numRanges>0; --numRanges) {
        // Read and skip the start and limit of this range.
        int32_t start=valueMaps[i];
        int32_t limit=valueMaps[i+1];
        i+=2;
        if(property<start) {
            break;
        }
        if(property<limit) {
            return i+(property-start)*2;
        }
        i+=(limit-start)*2;  // Skip all entries for this range.
    }
    return 0;
}

int32_t PropNameData::findPropertyValueNameGroup(int32_t valueMapIndex, int32_t value) {
    if(valueMapIndex==0) {
        return 0;  // The property does not have named values.
    }
    ++valueMapIndex;  // Skip the BytesTrie offset.
    int32_t numRanges=valueMaps[valueMapIndex++];
    if(numRanges<0x10) {
        // Ranges of values.
        for(; numRanges>0; --numRanges) {
            // Read and skip the start and limit of this range.
            int32_t start=valueMaps[valueMapIndex];
            int32_t limit=valueMaps[valueMapIndex+1];
            valueMapIndex+=2;
            if(value<start) {
                break;
            }
            if(value<limit) {
                return valueMaps[valueMapIndex+value-start];
            }
            valueMapIndex+=limit-start;  // Skip all entries for this range.
        }
    } else {
        // List of values.
        int32_t valuesStart=valueMapIndex;
        int32_t nameGroupOffsetsStart=valueMapIndex+numRanges-0x10;
        do {
            int32_t v=valueMaps[valueMapIndex];
            if(value<v) {
                break;
            }
            if(value==v) {
                return valueMaps[nameGroupOffsetsStart+valueMapIndex-valuesStart];
            }
        } while(++valueMapIndex<nameGroupOffsetsStart);
    }
    return 0;
}

const char *PropNameData::getName(const char *nameGroup, int32_t nameIndex) {
    int32_t numNames=*nameGroup++;
    if(nameIndex<0 || numNames<=nameIndex) {
        return NULL;
    }
    // Skip nameIndex names.
    for(; nameIndex>0; --nameIndex) {
        nameGroup=uprv_strchr(nameGroup, 0)+1;
    }
    if(*nameGroup==0) {
        return NULL;  // no name (Property[Value]Aliases.txt has "n/a")
    }
    return nameGroup;
}

UBool PropNameData::containsName(BytesTrie &trie, const char *name) {
    if(name==NULL) {
        return FALSE;
    }
    UStringTrieResult result=USTRINGTRIE_NO_VALUE;
    char c;
    while((c=*name++)!=0) {
        c=uprv_invCharToLowercaseAscii(c);
        // Ignore delimiters '-', '_', and ASCII White_Space.
        if(c==0x2d || c==0x5f || c==0x20 || (0x09<=c && c<=0x0d)) {
            continue;
        }
        if(!USTRINGTRIE_HAS_NEXT(result)) {
            return FALSE;
        }
        result=trie.next((uint8_t)c);
    }
    return USTRINGTRIE_HAS_VALUE(result);
}

const char *PropNameData::getPropertyName(int32_t property, int32_t nameChoice) {
    int32_t valueMapIndex=findProperty(property);
    if(valueMapIndex==0) {
        return NULL;  // Not a known property.
    }
    return getName(nameGroups+valueMaps[valueMapIndex], nameChoice);
}

const char *PropNameData::getPropertyValueName(int32_t property, int32_t value, int32_t nameChoice) {
    int32_t valueMapIndex=findProperty(property);
    if(valueMapIndex==0) {
        return NULL;  // Not a known property.
    }
    int32_t nameGroupOffset=findPropertyValueNameGroup(valueMaps[valueMapIndex+1], value);
    if(nameGroupOffset==0) {
        return NULL;
    }
    return getName(nameGroups+nameGroupOffset, nameChoice);
}

int32_t PropNameData::getPropertyOrValueEnum(int32_t bytesTrieOffset, const char *alias) {
    BytesTrie trie(bytesTries+bytesTrieOffset);
    if(containsName(trie, alias)) {
        return trie.getValue();
    } else {
        return UCHAR_INVALID_CODE;
    }
}

int32_t PropNameData::getPropertyEnum(const char *alias) {
    return getPropertyOrValueEnum(0, alias);
}

int32_t PropNameData::getPropertyValueEnum(int32_t property, const char *alias) {
    int32_t valueMapIndex=findProperty(property);
    if(valueMapIndex==0) {
        return UCHAR_INVALID_CODE;  // Not a known property.
    }
    valueMapIndex=valueMaps[valueMapIndex+1];
    if(valueMapIndex==0) {
        return UCHAR_INVALID_CODE;  // The property does not have named values.
    }
    // valueMapIndex is the start of the property's valueMap,
    // where the first word is the BytesTrie offset.
    return getPropertyOrValueEnum(valueMaps[valueMapIndex], alias);
}
U_NAMESPACE_END

//----------------------------------------------------------------------
// Public API implementation

U_CAPI const char* U_EXPORT2
u_getPropertyName(UProperty property,
                  UPropertyNameChoice nameChoice) {
    U_NAMESPACE_USE
    return PropNameData::getPropertyName(property, nameChoice);
}

U_CAPI UProperty U_EXPORT2
u_getPropertyEnum(const char* alias) {
    U_NAMESPACE_USE
    return (UProperty)PropNameData::getPropertyEnum(alias);
}

U_CAPI const char* U_EXPORT2
u_getPropertyValueName(UProperty property,
                       int32_t value,
                       UPropertyNameChoice nameChoice) {
    U_NAMESPACE_USE
    return PropNameData::getPropertyValueName(property, value, nameChoice);
}

U_CAPI int32_t U_EXPORT2
u_getPropertyValueEnum(UProperty property,
                       const char* alias) {
    U_NAMESPACE_USE
    return PropNameData::getPropertyValueEnum(property, alias);
}

U_CAPI const char*  U_EXPORT2
uscript_getName(UScriptCode scriptCode){
    return u_getPropertyValueName(UCHAR_SCRIPT, scriptCode,
                                  U_LONG_PROPERTY_NAME);
}

U_CAPI const char*  U_EXPORT2
uscript_getShortName(UScriptCode scriptCode){
    return u_getPropertyValueName(UCHAR_SCRIPT, scriptCode,
                                  U_SHORT_PROPERTY_NAME);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2002-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  propsvec.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2002feb22
*   created by: Markus W. Scherer
*
*   Store bits (Unicode character properties) in bit set vectors.
*/

#include <stdlib.h>


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2001-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  utrie.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001nov08
*   created by: Markus W. Scherer
*/





U_CDECL_BEGIN

/**
 * \file
 *
 * This is a common implementation of a "folded" trie.
 * It is a kind of compressed, serializable table of 16- or 32-bit values associated with
 * Unicode code points (0..0x10ffff).
 *
 * This implementation is optimized for getting values while walking forward
 * through a UTF-16 string.
 * Therefore, the simplest and fastest access macros are the
 * _FROM_LEAD() and _FROM_OFFSET_TRAIL() macros.
 *
 * The _FROM_BMP() macros are a little more complicated; they get values
 * even for lead surrogate code _points_, while the _FROM_LEAD() macros
 * get special "folded" values for lead surrogate code _units_ if
 * there is relevant data associated with them.
 * From such a folded value, an offset needs to be extracted to supply
 * to the _FROM_OFFSET_TRAIL() macros.
 *
 * Most of the more complex (and more convenient) functions/macros call a callback function
 * to get that offset from the folded value for a lead surrogate unit.
 */

/**
 * Trie constants, defining shift widths, index array lengths, etc.
 */
enum {
    /** Shift size for shifting right the input index. 1..9 */
    UTRIE_SHIFT=5,

    /** Number of data values in a stage 2 (data array) block. 2, 4, 8, .., 0x200 */
    UTRIE_DATA_BLOCK_LENGTH=1<<UTRIE_SHIFT,

    /** Mask for getting the lower bits from the input index. */
    UTRIE_MASK=UTRIE_DATA_BLOCK_LENGTH-1,

    /**
     * Lead surrogate code points' index displacement in the index array.
     * 0x10000-0xd800=0x2800
     */
    UTRIE_LEAD_INDEX_DISP=0x2800>>UTRIE_SHIFT,

    /**
     * Shift size for shifting left the index array values.
     * Increases possible data size with 16-bit index values at the cost
     * of compactability.
     * This requires blocks of stage 2 data to be aligned by UTRIE_DATA_GRANULARITY.
     * 0..UTRIE_SHIFT
     */
    UTRIE_INDEX_SHIFT=2,

    /** The alignment size of a stage 2 data block. Also the granularity for compaction. */
    UTRIE_DATA_GRANULARITY=1<<UTRIE_INDEX_SHIFT,

    /** Number of bits of a trail surrogate that are used in index table lookups. */
    UTRIE_SURROGATE_BLOCK_BITS=10-UTRIE_SHIFT,

    /**
     * Number of index (stage 1) entries per lead surrogate.
     * Same as number of index entries for 1024 trail surrogates,
     * ==0x400>>UTRIE_SHIFT
     */
    UTRIE_SURROGATE_BLOCK_COUNT=(1<<UTRIE_SURROGATE_BLOCK_BITS),

    /** Length of the BMP portion of the index (stage 1) array. */
    UTRIE_BMP_INDEX_LENGTH=0x10000>>UTRIE_SHIFT
};

/**
 * Length of the index (stage 1) array before folding.
 * Maximum number of Unicode code points (0x110000) shifted right by UTRIE_SHIFT.
 */
#define UTRIE_MAX_INDEX_LENGTH (0x110000>>UTRIE_SHIFT)

/**
 * Maximum length of the runtime data (stage 2) array.
 * Limited by 16-bit index values that are left-shifted by UTRIE_INDEX_SHIFT.
 */
#define UTRIE_MAX_DATA_LENGTH (0x10000<<UTRIE_INDEX_SHIFT)

/**
 * Maximum length of the build-time data (stage 2) array.
 * The maximum length is 0x110000+UTRIE_DATA_BLOCK_LENGTH+0x400.
 * (Number of Unicode code points + one all-initial-value block +
 *  possible duplicate entries for 1024 lead surrogates.)
 */
#define UTRIE_MAX_BUILD_TIME_DATA_LENGTH (0x110000+UTRIE_DATA_BLOCK_LENGTH+0x400)

/**
 * Number of bytes for a dummy trie.
 * A dummy trie is an empty runtime trie, used when a real data trie cannot
 * be loaded.
 * The number of bytes works for Latin-1-linear tries with 32-bit data
 * (worst case).
 *
 * Calculation:
 *   BMP index + 1 index block for lead surrogate code points +
 *   Latin-1-linear array + 1 data block for lead surrogate code points
 *
 * Latin-1: if(UTRIE_SHIFT<=8) { 256 } else { included in first data block }
 *
 * @see utrie_unserializeDummy
 */
#define UTRIE_DUMMY_SIZE ((UTRIE_BMP_INDEX_LENGTH+UTRIE_SURROGATE_BLOCK_COUNT)*2+(UTRIE_SHIFT<=8?256:UTRIE_DATA_BLOCK_LENGTH)*4+UTRIE_DATA_BLOCK_LENGTH*4)

/**
 * Runtime UTrie callback function.
 * Extract from a lead surrogate's data the
 * index array offset of the indexes for that lead surrogate.
 *
 * @param data data value for a surrogate from the trie, including the folding offset
 * @return offset>=UTRIE_BMP_INDEX_LENGTH, or 0 if there is no data for the lead surrogate
 */
typedef int32_t U_CALLCONV
UTrieGetFoldingOffset(uint32_t data);

/**
 * Run-time Trie structure.
 *
 * Either the data table is 16 bits wide and accessed via the index
 * pointer, with each index item increased by indexLength;
 * in this case, data32==NULL.
 *
 * Or the data table is 32 bits wide and accessed via the data32 pointer.
 */
struct UTrie {
    const uint16_t *index;
    const uint32_t *data32; /* NULL if 16b data is used via index */

    /**
     * This function is not used in _FROM_LEAD, _FROM_BMP, and _FROM_OFFSET_TRAIL macros.
     * If convenience macros like _GET16 or _NEXT32 are used, this function must be set.
     *
     * utrie_unserialize() sets a default function which simply returns
     * the lead surrogate's value itself - which is the inverse of the default
     * folding function used by utrie_serialize().
     *
     * @see UTrieGetFoldingOffset
     */
    UTrieGetFoldingOffset *getFoldingOffset;

    int32_t indexLength, dataLength;
    uint32_t initialValue;
    UBool isLatin1Linear;
};

#ifndef __UTRIE2_H__
typedef struct UTrie UTrie;
#endif

/** Internal trie getter from an offset (0 if c16 is a BMP/lead units) and a 16-bit unit */
#define _UTRIE_GET_RAW(trie, data, offset, c16) \
    (trie)->data[ \
        ((int32_t)((trie)->index[(offset)+((c16)>>UTRIE_SHIFT)])<<UTRIE_INDEX_SHIFT)+ \
        ((c16)&UTRIE_MASK) \
    ]

/** Internal trie getter from a pair of surrogates */
#define _UTRIE_GET_FROM_PAIR(trie, data, c, c2, result, resultType) UPRV_BLOCK_MACRO_BEGIN { \
    int32_t __offset; \
\
    /* get data for lead surrogate */ \
    (result)=_UTRIE_GET_RAW((trie), data, 0, (c)); \
    __offset=(trie)->getFoldingOffset(result); \
\
    /* get the real data from the folded lead/trail units */ \
    if(__offset>0) { \
        (result)=_UTRIE_GET_RAW((trie), data, __offset, (c2)&0x3ff); \
    } else { \
        (result)=(resultType)((trie)->initialValue); \
    } \
} UPRV_BLOCK_MACRO_END

/** Internal trie getter from a BMP code point, treating a lead surrogate as a normal code point */
#define _UTRIE_GET_FROM_BMP(trie, data, c16) \
    _UTRIE_GET_RAW(trie, data, 0xd800<=(c16) && (c16)<=0xdbff ? UTRIE_LEAD_INDEX_DISP : 0, c16)

/**
 * Internal trie getter from a code point.
 * Could be faster(?) but longer with
 *   if((c32)<=0xd7ff) { (result)=_UTRIE_GET_RAW(trie, data, 0, c32); }
 */
#define _UTRIE_GET(trie, data, c32, result, resultType) UPRV_BLOCK_MACRO_BEGIN { \
    if((uint32_t)(c32)<=0xffff) { \
        /* BMP code points */ \
        (result)=_UTRIE_GET_FROM_BMP(trie, data, c32); \
    } else if((uint32_t)(c32)<=0x10ffff) { \
        /* supplementary code point */ \
        UChar __lead16=U16_LEAD(c32); \
        _UTRIE_GET_FROM_PAIR(trie, data, __lead16, c32, result, resultType); \
    } else { \
        /* out of range */ \
        (result)=(resultType)((trie)->initialValue); \
    } \
} UPRV_BLOCK_MACRO_END

/** Internal next-post-increment: get the next code point (c, c2) and its data */
#define _UTRIE_NEXT(trie, data, src, limit, c, c2, result, resultType) UPRV_BLOCK_MACRO_BEGIN { \
    (c)=*(src)++; \
    if(!U16_IS_LEAD(c)) { \
        (c2)=0; \
        (result)=_UTRIE_GET_RAW((trie), data, 0, (c)); \
    } else if((src)!=(limit) && U16_IS_TRAIL((c2)=*(src))) { \
        ++(src); \
        _UTRIE_GET_FROM_PAIR((trie), data, (c), (c2), (result), resultType); \
    } else { \
        /* unpaired lead surrogate code point */ \
        (c2)=0; \
        (result)=_UTRIE_GET_RAW((trie), data, UTRIE_LEAD_INDEX_DISP, (c)); \
    } \
} UPRV_BLOCK_MACRO_END

/** Internal previous: get the previous code point (c, c2) and its data */
#define _UTRIE_PREVIOUS(trie, data, start, src, c, c2, result, resultType) UPRV_BLOCK_MACRO_BEGIN { \
    (c)=*--(src); \
    if(!U16_IS_SURROGATE(c)) { \
        (c2)=0; \
        (result)=_UTRIE_GET_RAW((trie), data, 0, (c)); \
    } else if(!U16_IS_SURROGATE_LEAD(c)) { \
        /* trail surrogate */ \
        if((start)!=(src) && U16_IS_LEAD((c2)=*((src)-1))) { \
            --(src); \
            (result)=(c); (c)=(c2); (c2)=(UChar)(result); /* swap c, c2 */ \
            _UTRIE_GET_FROM_PAIR((trie), data, (c), (c2), (result), resultType); \
        } else { \
            /* unpaired trail surrogate code point */ \
            (c2)=0; \
            (result)=_UTRIE_GET_RAW((trie), data, 0, (c)); \
        } \
    } else { \
        /* unpaired lead surrogate code point */ \
        (c2)=0; \
        (result)=_UTRIE_GET_RAW((trie), data, UTRIE_LEAD_INDEX_DISP, (c)); \
    } \
} UPRV_BLOCK_MACRO_END

/* Public UTrie API ---------------------------------------------------------*/

/**
 * Get a pointer to the contiguous part of the data array
 * for the Latin-1 range (U+0000..U+00ff).
 * Must be used only if the Latin-1 range is in fact linear
 * (trie->isLatin1Linear).
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @return (const uint16_t *) pointer to values for Latin-1 code points
 */
#define UTRIE_GET16_LATIN1(trie) ((trie)->index+(trie)->indexLength+UTRIE_DATA_BLOCK_LENGTH)

/**
 * Get a pointer to the contiguous part of the data array
 * for the Latin-1 range (U+0000..U+00ff).
 * Must be used only if the Latin-1 range is in fact linear
 * (trie->isLatin1Linear).
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @return (const uint32_t *) pointer to values for Latin-1 code points
 */
#define UTRIE_GET32_LATIN1(trie) ((trie)->data32+UTRIE_DATA_BLOCK_LENGTH)

/**
 * Get a 16-bit trie value from a BMP code point (UChar, <=U+ffff).
 * c16 may be a lead surrogate, which may have a value including a folding offset.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param c16 (UChar, in) the input BMP code point
 * @return (uint16_t) trie lookup result
 */
#define UTRIE_GET16_FROM_LEAD(trie, c16) _UTRIE_GET_RAW(trie, index, 0, c16)

/**
 * Get a 32-bit trie value from a BMP code point (UChar, <=U+ffff).
 * c16 may be a lead surrogate, which may have a value including a folding offset.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param c16 (UChar, in) the input BMP code point
 * @return (uint32_t) trie lookup result
 */
#define UTRIE_GET32_FROM_LEAD(trie, c16) _UTRIE_GET_RAW(trie, data32, 0, c16)

/**
 * Get a 16-bit trie value from a BMP code point (UChar, <=U+ffff).
 * Even lead surrogate code points are treated as normal code points,
 * with unfolded values that may differ from _FROM_LEAD() macro results for them.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param c16 (UChar, in) the input BMP code point
 * @return (uint16_t) trie lookup result
 */
#define UTRIE_GET16_FROM_BMP(trie, c16) _UTRIE_GET_FROM_BMP(trie, index, c16)

/**
 * Get a 32-bit trie value from a BMP code point (UChar, <=U+ffff).
 * Even lead surrogate code points are treated as normal code points,
 * with unfolded values that may differ from _FROM_LEAD() macro results for them.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param c16 (UChar, in) the input BMP code point
 * @return (uint32_t) trie lookup result
 */
#define UTRIE_GET32_FROM_BMP(trie, c16) _UTRIE_GET_FROM_BMP(trie, data32, c16)

/**
 * Get a 16-bit trie value from a code point.
 * Even lead surrogate code points are treated as normal code points,
 * with unfolded values that may differ from _FROM_LEAD() macro results for them.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param c32 (UChar32, in) the input code point
 * @param result (uint16_t, out) uint16_t variable for the trie lookup result
 */
#define UTRIE_GET16(trie, c32, result) _UTRIE_GET(trie, index, c32, result, uint16_t)

/**
 * Get a 32-bit trie value from a code point.
 * Even lead surrogate code points are treated as normal code points,
 * with unfolded values that may differ from _FROM_LEAD() macro results for them.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param c32 (UChar32, in) the input code point
 * @param result (uint32_t, out) uint32_t variable for the trie lookup result
 */
#define UTRIE_GET32(trie, c32, result) _UTRIE_GET(trie, data32, c32, result, uint32_t)

/**
 * Get the next code point (c, c2), post-increment src,
 * and get a 16-bit value from the trie.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param src (const UChar *, in/out) the source text pointer
 * @param limit (const UChar *, in) the limit pointer for the text, or NULL
 * @param c (UChar, out) variable for the BMP or lead code unit
 * @param c2 (UChar, out) variable for 0 or the trail code unit
 * @param result (uint16_t, out) uint16_t variable for the trie lookup result
 */
#define UTRIE_NEXT16(trie, src, limit, c, c2, result) _UTRIE_NEXT(trie, index, src, limit, c, c2, result, uint16_t)

/**
 * Get the next code point (c, c2), post-increment src,
 * and get a 32-bit value from the trie.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param src (const UChar *, in/out) the source text pointer
 * @param limit (const UChar *, in) the limit pointer for the text, or NULL
 * @param c (UChar, out) variable for the BMP or lead code unit
 * @param c2 (UChar, out) variable for 0 or the trail code unit
 * @param result (uint32_t, out) uint32_t variable for the trie lookup result
 */
#define UTRIE_NEXT32(trie, src, limit, c, c2, result) _UTRIE_NEXT(trie, data32, src, limit, c, c2, result, uint32_t)

/**
 * Get the previous code point (c, c2), pre-decrement src,
 * and get a 16-bit value from the trie.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param start (const UChar *, in) the start pointer for the text, or NULL
 * @param src (const UChar *, in/out) the source text pointer
 * @param c (UChar, out) variable for the BMP or lead code unit
 * @param c2 (UChar, out) variable for 0 or the trail code unit
 * @param result (uint16_t, out) uint16_t variable for the trie lookup result
 */
#define UTRIE_PREVIOUS16(trie, start, src, c, c2, result) _UTRIE_PREVIOUS(trie, index, start, src, c, c2, result, uint16_t)

/**
 * Get the previous code point (c, c2), pre-decrement src,
 * and get a 32-bit value from the trie.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param start (const UChar *, in) the start pointer for the text, or NULL
 * @param src (const UChar *, in/out) the source text pointer
 * @param c (UChar, out) variable for the BMP or lead code unit
 * @param c2 (UChar, out) variable for 0 or the trail code unit
 * @param result (uint32_t, out) uint32_t variable for the trie lookup result
 */
#define UTRIE_PREVIOUS32(trie, start, src, c, c2, result) _UTRIE_PREVIOUS(trie, data32, start, src, c, c2, result, uint32_t)

/**
 * Get a 16-bit trie value from a pair of surrogates.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param c (UChar, in) a lead surrogate
 * @param c2 (UChar, in) a trail surrogate
 * @param result (uint16_t, out) uint16_t variable for the trie lookup result
 */
#define UTRIE_GET16_FROM_PAIR(trie, c, c2, result) _UTRIE_GET_FROM_PAIR(trie, index, c, c2, result, uint16_t)

/**
 * Get a 32-bit trie value from a pair of surrogates.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param c (UChar, in) a lead surrogate
 * @param c2 (UChar, in) a trail surrogate
 * @param result (uint32_t, out) uint32_t variable for the trie lookup result
 */
#define UTRIE_GET32_FROM_PAIR(trie, c, c2, result) _UTRIE_GET_FROM_PAIR(trie, data32, c, c2, result, uint32_t)

/**
 * Get a 16-bit trie value from a folding offset (from the value of a lead surrogate)
 * and a trail surrogate.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param offset (int32_t, in) the folding offset from the value of a lead surrogate
 * @param c2 (UChar, in) a trail surrogate (only the 10 low bits are significant)
 * @return (uint16_t) trie lookup result
 */
#define UTRIE_GET16_FROM_OFFSET_TRAIL(trie, offset, c2) _UTRIE_GET_RAW(trie, index, offset, (c2)&0x3ff)

/**
 * Get a 32-bit trie value from a folding offset (from the value of a lead surrogate)
 * and a trail surrogate.
 *
 * @param trie (const UTrie *, in) a pointer to the runtime trie structure
 * @param offset (int32_t, in) the folding offset from the value of a lead surrogate
 * @param c2 (UChar, in) a trail surrogate (only the 10 low bits are significant)
 * @return (uint32_t) trie lookup result
 */
#define UTRIE_GET32_FROM_OFFSET_TRAIL(trie, offset, c2) _UTRIE_GET_RAW(trie, data32, offset, (c2)&0x3ff)

/* enumeration callback types */

/**
 * Callback from utrie_enum(), extracts a uint32_t value from a
 * trie value. This value will be passed on to the UTrieEnumRange function.
 *
 * @param context an opaque pointer, as passed into utrie_enum()
 * @param value a value from the trie
 * @return the value that is to be passed on to the UTrieEnumRange function
 */
typedef uint32_t U_CALLCONV
UTrieEnumValue(const void *context, uint32_t value);

/**
 * Callback from utrie_enum(), is called for each contiguous range
 * of code points with the same value as retrieved from the trie and
 * transformed by the UTrieEnumValue function.
 *
 * The callback function can stop the enumeration by returning FALSE.
 *
 * @param context an opaque pointer, as passed into utrie_enum()
 * @param start the first code point in a contiguous range with value
 * @param limit one past the last code point in a contiguous range with value
 * @param value the value that is set for all code points in [start..limit[
 * @return FALSE to stop the enumeration
 */
typedef UBool U_CALLCONV
UTrieEnumRange(const void *context, UChar32 start, UChar32 limit, uint32_t value);

/**
 * Enumerate efficiently all values in a trie.
 * For each entry in the trie, the value to be delivered is passed through
 * the UTrieEnumValue function.
 * The value is unchanged if that function pointer is NULL.
 *
 * For each contiguous range of code points with a given value,
 * the UTrieEnumRange function is called.
 *
 * @param trie a pointer to the runtime trie structure
 * @param enumValue a pointer to a function that may transform the trie entry value,
 *                  or NULL if the values from the trie are to be used directly
 * @param enumRange a pointer to a function that is called for each contiguous range
 *                  of code points with the same value
 * @param context an opaque pointer that is passed on to the callback functions
 */
U_CAPI void U_EXPORT2
utrie_enum(const UTrie *trie,
           UTrieEnumValue *enumValue, UTrieEnumRange *enumRange, const void *context);

/**
 * Unserialize a trie from 32-bit-aligned memory.
 * Inverse of utrie_serialize().
 * Fills the UTrie runtime trie structure with the settings for the trie data.
 *
 * @param trie a pointer to the runtime trie structure
 * @param data a pointer to 32-bit-aligned memory containing trie data
 * @param length the number of bytes available at data
 * @param pErrorCode an in/out ICU UErrorCode
 * @return the number of bytes at data taken up by the trie data
 */
U_CAPI int32_t U_EXPORT2
utrie_unserialize(UTrie *trie, const void *data, int32_t length, UErrorCode *pErrorCode);

/**
 * "Unserialize" a dummy trie.
 * A dummy trie is an empty runtime trie, used when a real data trie cannot
 * be loaded.
 *
 * The input memory is filled so that the trie always returns the initialValue,
 * or the leadUnitValue for lead surrogate code points.
 * The Latin-1 part is always set up to be linear.
 *
 * @param trie a pointer to the runtime trie structure
 * @param data a pointer to 32-bit-aligned memory to be filled with the dummy trie data
 * @param length the number of bytes available at data (recommended to use UTRIE_DUMMY_SIZE)
 * @param initialValue the initial value that is set for all code points
 * @param leadUnitValue the value for lead surrogate code _units_ that do not
 *                      have associated supplementary data
 * @param pErrorCode an in/out ICU UErrorCode
 *
 * @see UTRIE_DUMMY_SIZE
 * @see utrie_open
 */
U_CAPI int32_t U_EXPORT2
utrie_unserializeDummy(UTrie *trie,
                       void *data, int32_t length,
                       uint32_t initialValue, uint32_t leadUnitValue,
                       UBool make16BitTrie,
                       UErrorCode *pErrorCode);

/**
 * Default implementation for UTrie.getFoldingOffset, set automatically by
 * utrie_unserialize().
 * Simply returns the lead surrogate's value itself - which is the inverse
 * of the default folding function used by utrie_serialize().
 * Exported for static const UTrie structures.
 *
 * @see UTrieGetFoldingOffset
 */
U_CAPI int32_t U_EXPORT2
utrie_defaultGetFoldingOffset(uint32_t data);

/* Building a trie ----------------------------------------------------------*/

/**
 * Build-time trie structure.
 * Opaque definition, here only to make fillIn parameters possible
 * for utrie_open() and utrie_clone().
 */
struct UNewTrie {
    /**
     * Index values at build-time are 32 bits wide for easier processing.
     * Bit 31 is set if the data block is used by multiple index values (from utrie_setRange()).
     */
    int32_t index[UTRIE_MAX_INDEX_LENGTH+UTRIE_SURROGATE_BLOCK_COUNT];
    uint32_t *data;

    uint32_t leadUnitValue;
    int32_t indexLength, dataCapacity, dataLength;
    UBool isAllocated, isDataAllocated;
    UBool isLatin1Linear, isCompacted;

    /**
     * Map of adjusted indexes, used in utrie_compact().
     * Maps from original indexes to new ones.
     */
    int32_t map[UTRIE_MAX_BUILD_TIME_DATA_LENGTH>>UTRIE_SHIFT];
};

typedef struct UNewTrie UNewTrie;

/**
 * Build-time trie callback function, used with utrie_serialize().
 * This function calculates a lead surrogate's value including a folding offset
 * from the 1024 supplementary code points [start..start+1024[ .
 * It is U+10000 <= start <= U+10fc00 and (start&0x3ff)==0.
 *
 * The folding offset is provided by the caller.
 * It is offset=UTRIE_BMP_INDEX_LENGTH+n*UTRIE_SURROGATE_BLOCK_COUNT with n=0..1023.
 * Instead of the offset itself, n can be stored in 10 bits -
 * or fewer if it can be assumed that few lead surrogates have associated data.
 *
 * The returned value must be
 * - not zero if and only if there is relevant data
 *   for the corresponding 1024 supplementary code points
 * - such that UTrie.getFoldingOffset(UNewTrieGetFoldedValue(..., offset))==offset
 *
 * @return a folded value, or 0 if there is no relevant data for the lead surrogate.
 */
typedef uint32_t U_CALLCONV
UNewTrieGetFoldedValue(UNewTrie *trie, UChar32 start, int32_t offset);

/**
 * Open a build-time trie structure.
 * The size of the build-time data array is specified to avoid allocating a large
 * array in all cases. The array itself can also be passed in.
 *
 * Although the trie is never fully expanded to a linear array, especially when
 * utrie_setRange32() is used, the data array could be large during build time.
 * The maximum length is
 * UTRIE_MAX_BUILD_TIME_DATA_LENGTH=0x110000+UTRIE_DATA_BLOCK_LENGTH+0x400.
 * (Number of Unicode code points + one all-initial-value block +
 *  possible duplicate entries for 1024 lead surrogates.)
 * (UTRIE_DATA_BLOCK_LENGTH<=0x200 in all cases.)
 *
 * @param fillIn a pointer to a UNewTrie structure to be initialized (will not be released), or
 *               NULL if one is to be allocated
 * @param aliasData a pointer to a data array to be used (will not be released), or
 *                  NULL if one is to be allocated
 * @param maxDataLength the capacity of aliasData (if not NULL) or
 *                      the length of the data array to be allocated
 * @param initialValue the initial value that is set for all code points
 * @param leadUnitValue the value for lead surrogate code _units_ that do not
 *                      have associated supplementary data
 * @param latin1Linear a flag indicating whether the Latin-1 range is to be allocated and
 *                     kept in a linear, contiguous part of the data array
 * @return a pointer to the initialized fillIn or the allocated and initialized new UNewTrie
 */
U_CAPI UNewTrie * U_EXPORT2
utrie_open(UNewTrie *fillIn,
           uint32_t *aliasData, int32_t maxDataLength,
           uint32_t initialValue, uint32_t leadUnitValue,
           UBool latin1Linear);

/**
 * Clone a build-time trie structure with all entries.
 *
 * @param fillIn like in utrie_open()
 * @param other the build-time trie structure to clone
 * @param aliasData like in utrie_open(),
 *                  used if aliasDataLength>=(capacity of other's data array)
 * @param aliasDataLength the length of aliasData
 * @return a pointer to the initialized fillIn or the allocated and initialized new UNewTrie
 */
U_CAPI UNewTrie * U_EXPORT2
utrie_clone(UNewTrie *fillIn, const UNewTrie *other, uint32_t *aliasData, int32_t aliasDataLength);

/**
 * Close a build-time trie structure, and release memory
 * that was allocated by utrie_open() or utrie_clone().
 *
 * @param trie the build-time trie
 */
U_CAPI void U_EXPORT2
utrie_close(UNewTrie *trie);

/**
 * Get the data array of a build-time trie.
 * The data may be modified, but entries that are equal before
 * must still be equal after modification.
 *
 * @param trie the build-time trie
 * @param pLength (out) a pointer to a variable that receives the number
 *                of entries in the data array
 * @return the data array
 */
U_CAPI uint32_t * U_EXPORT2
utrie_getData(UNewTrie *trie, int32_t *pLength);

/**
 * Set a value for a code point.
 *
 * @param trie the build-time trie
 * @param c the code point
 * @param value the value
 * @return FALSE if a failure occurred (illegal argument or data array overrun)
 */
U_CAPI UBool U_EXPORT2
utrie_set32(UNewTrie *trie, UChar32 c, uint32_t value);

/**
 * Get a value from a code point as stored in the build-time trie.
 *
 * @param trie the build-time trie
 * @param c the code point
 * @param pInBlockZero if not NULL, then *pInBlockZero is set to TRUE
 *                     iff the value is retrieved from block 0;
 *                     block 0 is the all-initial-value initial block
 * @return the value
 */
U_CAPI uint32_t U_EXPORT2
utrie_get32(UNewTrie *trie, UChar32 c, UBool *pInBlockZero);

/**
 * Set a value in a range of code points [start..limit[.
 * All code points c with start<=c<limit will get the value if
 * overwrite is TRUE or if the old value is 0.
 *
 * @param trie the build-time trie
 * @param start the first code point to get the value
 * @param limit one past the last code point to get the value
 * @param value the value
 * @param overwrite flag for whether old non-initial values are to be overwritten
 * @return FALSE if a failure occurred (illegal argument or data array overrun)
 */
U_CAPI UBool U_EXPORT2
utrie_setRange32(UNewTrie *trie, UChar32 start, UChar32 limit, uint32_t value, UBool overwrite);

/**
 * Compact the build-time trie after all values are set, and then
 * serialize it into 32-bit aligned memory.
 *
 * After this, the trie can only be serizalized again and/or closed;
 * no further values can be added.
 *
 * @see utrie_unserialize()
 *
 * @param trie the build-time trie
 * @param data a pointer to 32-bit-aligned memory for the trie data
 * @param capacity the number of bytes available at data
 * @param getFoldedValue a callback function that calculates the value for
 *                       a lead surrogate from all of its supplementary code points
 *                       and the folding offset;
 *                       if NULL, then a default function is used which returns just
 *                       the input offset when there are any non-initial-value entries
 * @param reduceTo16Bits flag for whether the values are to be reduced to a
 *                       width of 16 bits for serialization and runtime
 * @param pErrorCode a UErrorCode argument; among other possible error codes:
 * - U_BUFFER_OVERFLOW_ERROR if the data storage block is too small for serialization
 * - U_MEMORY_ALLOCATION_ERROR if the trie data array is too small
 * - U_INDEX_OUTOFBOUNDS_ERROR if the index or data arrays are too long after compaction for serialization
 *
 * @return the number of bytes written for the trie
 */
U_CAPI int32_t U_EXPORT2
utrie_serialize(UNewTrie *trie, void *data, int32_t capacity,
                UNewTrieGetFoldedValue *getFoldedValue,
                UBool reduceTo16Bits,
                UErrorCode *pErrorCode);

/* serialization ------------------------------------------------------------ */

// UTrie signature values, in platform endianness and opposite endianness.
// The UTrie signature ASCII byte values spell "Trie".
#define UTRIE_SIG       0x54726965
#define UTRIE_OE_SIG    0x65697254

/**
 * Trie data structure in serialized form:
 *
 * UTrieHeader header;
 * uint16_t index[header.indexLength];
 * uint16_t data[header.dataLength];
 * @internal
 */
typedef struct UTrieHeader {
    /** "Trie" in big-endian US-ASCII (0x54726965) */
    uint32_t signature;

    /**
     * options bit field:
     *     9    1=Latin-1 data is stored linearly at data+UTRIE_DATA_BLOCK_LENGTH
     *     8    0=16-bit data, 1=32-bit data
     *  7..4    UTRIE_INDEX_SHIFT   // 0..UTRIE_SHIFT
     *  3..0    UTRIE_SHIFT         // 1..9
     */
    uint32_t options;

    /** indexLength is a multiple of UTRIE_SURROGATE_BLOCK_COUNT */
    int32_t indexLength;

    /** dataLength>=UTRIE_DATA_BLOCK_LENGTH */
    int32_t dataLength;
} UTrieHeader;

/**
 * Constants for use with UTrieHeader.options.
 * @internal
 */
enum {
    /** Mask to get the UTRIE_SHIFT value from options. */
    UTRIE_OPTIONS_SHIFT_MASK=0xf,

    /** Shift options right this much to get the UTRIE_INDEX_SHIFT value. */
    UTRIE_OPTIONS_INDEX_SHIFT=4,

    /** If set, then the data (stage 2) array is 32 bits wide. */
    UTRIE_OPTIONS_DATA_IS_32_BIT=0x100,

    /**
     * If set, then Latin-1 data (for U+0000..U+00ff) is stored in the data (stage 2) array
     * as a simple, linear array at data+UTRIE_DATA_BLOCK_LENGTH.
     */
    UTRIE_OPTIONS_LATIN1_IS_LINEAR=0x200
};

U_CDECL_END




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2002-2010, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  propsvec.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2002feb22
*   created by: Markus W. Scherer
*
*   Store bits (Unicode character properties) in bit set vectors.
*/






U_CDECL_BEGIN

/**
 * Unicode Properties Vectors associated with code point ranges.
 *
 * Rows of uint32_t integers in a contiguous array store
 * the range limits and the properties vectors.
 *
 * Logically, each row has a certain number of uint32_t values,
 * which is set via the upvec_open() "columns" parameter.
 *
 * Internally, two additional columns are stored.
 * In each internal row,
 * row[0] contains the start code point and
 * row[1] contains the limit code point,
 * which is the start of the next range.
 *
 * Initially, there is only one "normal" row for
 * range [0..0x110000[ with values 0.
 * There are additional rows for special purposes, see UPVEC_FIRST_SPECIAL_CP.
 *
 * It would be possible to store only one range boundary per row,
 * but self-contained rows allow to later sort them by contents.
 */
struct UPropsVectors;
typedef struct UPropsVectors UPropsVectors;

/*
 * Special pseudo code points for storing the initialValue and the errorValue,
 * which are used to initialize a UTrie2 or similar.
 */
#define UPVEC_FIRST_SPECIAL_CP 0x110000
#define UPVEC_INITIAL_VALUE_CP 0x110000
#define UPVEC_ERROR_VALUE_CP 0x110001
#define UPVEC_MAX_CP 0x110001

/*
 * Special pseudo code point used in upvec_compact() signalling the end of
 * delivering special values and the beginning of delivering real ones.
 * Stable value, unlike UPVEC_MAX_CP which might grow over time.
 */
#define UPVEC_START_REAL_VALUES_CP 0x200000

/*
 * Open a UPropsVectors object.
 * @param columns Number of value integers (uint32_t) per row.
 */
U_CAPI UPropsVectors * U_EXPORT2
upvec_open(int32_t columns, UErrorCode *pErrorCode);

U_CAPI void U_EXPORT2
upvec_close(UPropsVectors *pv);

/*
 * In rows for code points [start..end], select the column,
 * reset the mask bits and set the value bits (ANDed with the mask).
 *
 * Will set U_NO_WRITE_PERMISSION if called after upvec_compact().
 */
U_CAPI void U_EXPORT2
upvec_setValue(UPropsVectors *pv,
               UChar32 start, UChar32 end,
               int32_t column,
               uint32_t value, uint32_t mask,
               UErrorCode *pErrorCode);

/*
 * Logically const but must not be used on the same pv concurrently!
 * Always returns 0 if called after upvec_compact().
 */
U_CAPI uint32_t U_EXPORT2
upvec_getValue(const UPropsVectors *pv, UChar32 c, int32_t column);

/*
 * pRangeStart and pRangeEnd can be NULL.
 * @return NULL if rowIndex out of range and for illegal arguments,
 *         or if called after upvec_compact()
 */
U_CAPI uint32_t * U_EXPORT2
upvec_getRow(const UPropsVectors *pv, int32_t rowIndex,
             UChar32 *pRangeStart, UChar32 *pRangeEnd);

/*
 * Compact the vectors:
 * - modify the memory
 * - keep only unique vectors
 * - store them contiguously from the beginning of the memory
 * - for each (non-unique) row, call the handler function
 *
 * The handler's rowIndex is the index of the row in the compacted
 * memory block.
 * (Therefore, it starts at 0 increases in increments of the columns value.)
 *
 * In a first phase, only special values are delivered (each exactly once),
 * with start==end both equalling a special pseudo code point.
 * Then the handler is called once more with start==end==UPVEC_START_REAL_VALUES_CP
 * where rowIndex is the length of the compacted array,
 * and the row is arbitrary (but not NULL).
 * Then, in the second phase, the handler is called for each row of real values.
 */
typedef void U_CALLCONV
UPVecCompactHandler(void *context,
                    UChar32 start, UChar32 end,
                    int32_t rowIndex, uint32_t *row, int32_t columns,
                    UErrorCode *pErrorCode);

U_CAPI void U_EXPORT2
upvec_compact(UPropsVectors *pv, UPVecCompactHandler *handler, void *context, UErrorCode *pErrorCode);

/*
 * Get the vectors array after calling upvec_compact().
 * The caller must not modify nor release the returned array.
 * Returns NULL if called before upvec_compact().
 */
U_CAPI const uint32_t * U_EXPORT2
upvec_getArray(const UPropsVectors *pv, int32_t *pRows, int32_t *pColumns);

/*
 * Get a clone of the vectors array after calling upvec_compact().
 * The caller owns the returned array and must uprv_free() it.
 * Returns NULL if called before upvec_compact().
 */
U_CAPI uint32_t * U_EXPORT2
upvec_cloneArray(const UPropsVectors *pv,
                 int32_t *pRows, int32_t *pColumns, UErrorCode *pErrorCode);

/*
 * Call upvec_compact(), create a 16-bit UTrie2 with indexes into the compacted
 * vectors array, and freeze the trie.
 */
U_CAPI UTrie2 * U_EXPORT2
upvec_compactToUTrie2WithRowIndexes(UPropsVectors *pv, UErrorCode *pErrorCode);

struct UPVecToUTrie2Context {
    UTrie2 *trie;
    int32_t initialValue;
    int32_t errorValue;
    int32_t maxValue;
};
typedef struct UPVecToUTrie2Context UPVecToUTrie2Context;

/* context=UPVecToUTrie2Context, creates the trie and stores the rowIndex values */
U_CAPI void U_CALLCONV
upvec_compactToUTrie2Handler(void *context,
                             UChar32 start, UChar32 end,
                             int32_t rowIndex, uint32_t *row, int32_t columns,
                             UErrorCode *pErrorCode);

U_CDECL_END




struct UPropsVectors {
    uint32_t *v;
    int32_t columns;  /* number of columns, plus two for start & limit values */
    int32_t maxRows;
    int32_t rows;
    int32_t prevRow;  /* search optimization: remember last row seen */
    UBool isCompacted;
};

#define UPVEC_INITIAL_ROWS (1<<12)
#define UPVEC_MEDIUM_ROWS ((int32_t)1<<16)
#define UPVEC_MAX_ROWS (UPVEC_MAX_CP+1)

U_CAPI UPropsVectors * U_EXPORT2
upvec_open(int32_t columns, UErrorCode *pErrorCode) {
    UPropsVectors *pv;
    uint32_t *v, *row;
    uint32_t cp;

    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if(columns<1) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    columns+=2; /* count range start and limit columns */

    pv=(UPropsVectors *)uprv_malloc(sizeof(UPropsVectors));
    v=(uint32_t *)uprv_malloc(UPVEC_INITIAL_ROWS*columns*4);
    if(pv==NULL || v==NULL) {
        uprv_free(pv);
        uprv_free(v);
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    uprv_memset(pv, 0, sizeof(UPropsVectors));
    pv->v=v;
    pv->columns=columns;
    pv->maxRows=UPVEC_INITIAL_ROWS;
    pv->rows=2+(UPVEC_MAX_CP-UPVEC_FIRST_SPECIAL_CP);

    /* set the all-Unicode row and the special-value rows */
    row=pv->v;
    uprv_memset(row, 0, pv->rows*columns*4);
    row[0]=0;
    row[1]=0x110000;
    row+=columns;
    for(cp=UPVEC_FIRST_SPECIAL_CP; cp<=UPVEC_MAX_CP; ++cp) {
        row[0]=cp;
        row[1]=cp+1;
        row+=columns;
    }
    return pv;
}

U_CAPI void U_EXPORT2
upvec_close(UPropsVectors *pv) {
    if(pv!=NULL) {
        uprv_free(pv->v);
        uprv_free(pv);
    }
}

static uint32_t *
_findRow(UPropsVectors *pv, UChar32 rangeStart) {
    uint32_t *row;
    int32_t columns, i, start, limit, prevRow;

    columns=pv->columns;
    limit=pv->rows;
    prevRow=pv->prevRow;

    /* check the vicinity of the last-seen row (start searching with an unrolled loop) */
    row=pv->v+prevRow*columns;
    if(rangeStart>=(UChar32)row[0]) {
        if(rangeStart<(UChar32)row[1]) {
            /* same row as last seen */
            return row;
        } else if(rangeStart<(UChar32)(row+=columns)[1]) {
            /* next row after the last one */
            pv->prevRow=prevRow+1;
            return row;
        } else if(rangeStart<(UChar32)(row+=columns)[1]) {
            /* second row after the last one */
            pv->prevRow=prevRow+2;
            return row;
        } else if((rangeStart-(UChar32)row[1])<10) {
            /* we are close, continue looping */
            prevRow+=2;
            do {
                ++prevRow;
                row+=columns;
            } while(rangeStart>=(UChar32)row[1]);
            pv->prevRow=prevRow;
            return row;
        }
    } else if(rangeStart<(UChar32)pv->v[1]) {
        /* the very first row */
        pv->prevRow=0;
        return pv->v;
    }

    /* do a binary search for the start of the range */
    start=0;
    while(start<limit-1) {
        i=(start+limit)/2;
        row=pv->v+i*columns;
        if(rangeStart<(UChar32)row[0]) {
            limit=i;
        } else if(rangeStart<(UChar32)row[1]) {
            pv->prevRow=i;
            return row;
        } else {
            start=i;
        }
    }

    /* must be found because all ranges together always cover all of Unicode */
    pv->prevRow=start;
    return pv->v+start*columns;
}

U_CAPI void U_EXPORT2
upvec_setValue(UPropsVectors *pv,
               UChar32 start, UChar32 end,
               int32_t column,
               uint32_t value, uint32_t mask,
               UErrorCode *pErrorCode) {
    uint32_t *firstRow, *lastRow;
    int32_t columns;
    UChar32 limit;
    UBool splitFirstRow, splitLastRow;

    /* argument checking */
    if(U_FAILURE(*pErrorCode)) {
        return;
    }
    if( pv==NULL ||
        start<0 || start>end || end>UPVEC_MAX_CP ||
        column<0 || column>=(pv->columns-2)
    ) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if(pv->isCompacted) {
        *pErrorCode=U_NO_WRITE_PERMISSION;
        return;
    }
    limit=end+1;

    /* initialize */
    columns=pv->columns;
    column+=2; /* skip range start and limit columns */
    value&=mask;

    /* find the rows whose ranges overlap with the input range */

    /* find the first and last rows, always successful */
    firstRow=_findRow(pv, start);
    lastRow=_findRow(pv, end);

    /*
     * Rows need to be split if they partially overlap with the
     * input range (only possible for the first and last rows)
     * and if their value differs from the input value.
     */
    splitFirstRow= (UBool)(start!=(UChar32)firstRow[0] && value!=(firstRow[column]&mask));
    splitLastRow= (UBool)(limit!=(UChar32)lastRow[1] && value!=(lastRow[column]&mask));

    /* split first/last rows if necessary */
    if(splitFirstRow || splitLastRow) {
        int32_t count, rows;

        rows=pv->rows;
        if((rows+splitFirstRow+splitLastRow)>pv->maxRows) {
            uint32_t *newVectors;
            int32_t newMaxRows;

            if(pv->maxRows<UPVEC_MEDIUM_ROWS) {
                newMaxRows=UPVEC_MEDIUM_ROWS;
            } else if(pv->maxRows<UPVEC_MAX_ROWS) {
                newMaxRows=UPVEC_MAX_ROWS;
            } else {
                /* Implementation bug, or UPVEC_MAX_ROWS too low. */
                *pErrorCode=U_INTERNAL_PROGRAM_ERROR;
                return;
            }
            newVectors=(uint32_t *)uprv_malloc(newMaxRows*columns*4);
            if(newVectors==NULL) {
                *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            uprv_memcpy(newVectors, pv->v, (size_t)rows*columns*4);
            firstRow=newVectors+(firstRow-pv->v);
            lastRow=newVectors+(lastRow-pv->v);
            uprv_free(pv->v);
            pv->v=newVectors;
            pv->maxRows=newMaxRows;
        }

        /* count the number of row cells to move after the last row, and move them */
        count = (int32_t)((pv->v+rows*columns)-(lastRow+columns));
        if(count>0) {
            uprv_memmove(
                lastRow+(1+splitFirstRow+splitLastRow)*columns,
                lastRow+columns,
                count*4);
        }
        pv->rows=rows+splitFirstRow+splitLastRow;

        /* split the first row, and move the firstRow pointer to the second part */
        if(splitFirstRow) {
            /* copy all affected rows up one and move the lastRow pointer */
            count = (int32_t)((lastRow-firstRow)+columns);
            uprv_memmove(firstRow+columns, firstRow, (size_t)count*4);
            lastRow+=columns;

            /* split the range and move the firstRow pointer */
            firstRow[1]=firstRow[columns]=(uint32_t)start;
            firstRow+=columns;
        }

        /* split the last row */
        if(splitLastRow) {
            /* copy the last row data */
            uprv_memcpy(lastRow+columns, lastRow, (size_t)columns*4);

            /* split the range and move the firstRow pointer */
            lastRow[1]=lastRow[columns]=(uint32_t)limit;
        }
    }

    /* set the "row last seen" to the last row for the range */
    pv->prevRow=(int32_t)((lastRow-(pv->v))/columns);

    /* set the input value in all remaining rows */
    firstRow+=column;
    lastRow+=column;
    mask=~mask;
    for(;;) {
        *firstRow=(*firstRow&mask)|value;
        if(firstRow==lastRow) {
            break;
        }
        firstRow+=columns;
    }
}

U_CAPI uint32_t U_EXPORT2
upvec_getValue(const UPropsVectors *pv, UChar32 c, int32_t column) {
    uint32_t *row;
    UPropsVectors *ncpv;

    if(pv->isCompacted || c<0 || c>UPVEC_MAX_CP || column<0 || column>=(pv->columns-2)) {
        return 0;
    }
    ncpv=(UPropsVectors *)pv;
    row=_findRow(ncpv, c);
    return row[2+column];
}

U_CAPI uint32_t * U_EXPORT2
upvec_getRow(const UPropsVectors *pv, int32_t rowIndex,
             UChar32 *pRangeStart, UChar32 *pRangeEnd) {
    uint32_t *row;
    int32_t columns;

    if(pv->isCompacted || rowIndex<0 || rowIndex>=pv->rows) {
        return NULL;
    }

    columns=pv->columns;
    row=pv->v+rowIndex*columns;
    if(pRangeStart!=NULL) {
        *pRangeStart=(UChar32)row[0];
    }
    if(pRangeEnd!=NULL) {
        *pRangeEnd=(UChar32)row[1]-1;
    }
    return row+2;
}

static int32_t U_CALLCONV
upvec_compareRows(const void *context, const void *l, const void *r) {
    const uint32_t *left=(const uint32_t *)l, *right=(const uint32_t *)r;
    const UPropsVectors *pv=(const UPropsVectors *)context;
    int32_t i, count, columns;

    count=columns=pv->columns; /* includes start/limit columns */

    /* start comparing after start/limit but wrap around to them */
    i=2;
    do {
        if(left[i]!=right[i]) {
            return left[i]<right[i] ? -1 : 1;
        }
        if(++i==columns) {
            i=0;
        }
    } while(--count>0);

    return 0;
}

U_CAPI void U_EXPORT2
upvec_compact(UPropsVectors *pv, UPVecCompactHandler *handler, void *context, UErrorCode *pErrorCode) {
    uint32_t *row;
    int32_t i, columns, valueColumns, rows, count;
    UChar32 start, limit;

    /* argument checking */
    if(U_FAILURE(*pErrorCode)) {
        return;
    }
    if(handler==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if(pv->isCompacted) {
        return;
    }

    /* Set the flag now: Sorting and compacting destroys the builder data structure. */
    pv->isCompacted=TRUE;

    rows=pv->rows;
    columns=pv->columns;
    U_ASSERT(columns>=3); /* upvec_open asserts this */
    valueColumns=columns-2; /* not counting start & limit */

    /* sort the properties vectors to find unique vector values */
    uprv_sortArray(pv->v, rows, columns*4,
                   upvec_compareRows, pv, FALSE, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        return;
    }

    /*
     * Find and set the special values.
     * This has to do almost the same work as the compaction below,
     * to find the indexes where the special-value rows will move.
     */
    row=pv->v;
    count=-valueColumns;
    for(i=0; i<rows; ++i) {
        start=(UChar32)row[0];

        /* count a new values vector if it is different from the current one */
        if(count<0 || 0!=uprv_memcmp(row+2, row-valueColumns, valueColumns*4)) {
            count+=valueColumns;
        }

        if(start>=UPVEC_FIRST_SPECIAL_CP) {
            handler(context, start, start, count, row+2, valueColumns, pErrorCode);
            if(U_FAILURE(*pErrorCode)) {
                return;
            }
        }

        row+=columns;
    }

    /* count is at the beginning of the last vector, add valueColumns to include that last vector */
    count+=valueColumns;

    /* Call the handler once more to signal the start of delivering real values. */
    handler(context, UPVEC_START_REAL_VALUES_CP, UPVEC_START_REAL_VALUES_CP,
            count, row-valueColumns, valueColumns, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        return;
    }

    /*
     * Move vector contents up to a contiguous array with only unique
     * vector values, and call the handler function for each vector.
     *
     * This destroys the Properties Vector structure and replaces it
     * with an array of just vector values.
     */
    row=pv->v;
    count=-valueColumns;
    for(i=0; i<rows; ++i) {
        /* fetch these first before memmove() may overwrite them */
        start=(UChar32)row[0];
        limit=(UChar32)row[1];

        /* add a new values vector if it is different from the current one */
        if(count<0 || 0!=uprv_memcmp(row+2, pv->v+count, valueColumns*4)) {
            count+=valueColumns;
            uprv_memmove(pv->v+count, row+2, (size_t)valueColumns*4);
        }

        if(start<UPVEC_FIRST_SPECIAL_CP) {
            handler(context, start, limit-1, count, pv->v+count, valueColumns, pErrorCode);
            if(U_FAILURE(*pErrorCode)) {
                return;
            }
        }

        row+=columns;
    }

    /* count is at the beginning of the last vector, add one to include that last vector */
    pv->rows=count/valueColumns+1;
}

U_CAPI const uint32_t * U_EXPORT2
upvec_getArray(const UPropsVectors *pv, int32_t *pRows, int32_t *pColumns) {
    if(!pv->isCompacted) {
        return NULL;
    }
    if(pRows!=NULL) {
        *pRows=pv->rows;
    }
    if(pColumns!=NULL) {
        *pColumns=pv->columns-2;
    }
    return pv->v;
}

U_CAPI uint32_t * U_EXPORT2
upvec_cloneArray(const UPropsVectors *pv,
                 int32_t *pRows, int32_t *pColumns, UErrorCode *pErrorCode) {
    uint32_t *clonedArray;
    int32_t byteLength;

    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if(!pv->isCompacted) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    byteLength=pv->rows*(pv->columns-2)*4;
    clonedArray=(uint32_t *)uprv_malloc(byteLength);
    if(clonedArray==NULL) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    uprv_memcpy(clonedArray, pv->v, byteLength);
    if(pRows!=NULL) {
        *pRows=pv->rows;
    }
    if(pColumns!=NULL) {
        *pColumns=pv->columns-2;
    }
    return clonedArray;
}

U_CAPI UTrie2 * U_EXPORT2
upvec_compactToUTrie2WithRowIndexes(UPropsVectors *pv, UErrorCode *pErrorCode) {
    UPVecToUTrie2Context toUTrie2={ NULL, 0, 0, 0 };
    upvec_compact(pv, upvec_compactToUTrie2Handler, &toUTrie2, pErrorCode);
    utrie2_freeze(toUTrie2.trie, UTRIE2_16_VALUE_BITS, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        utrie2_close(toUTrie2.trie);
        toUTrie2.trie=NULL;
    }
    return toUTrie2.trie;
}

/*
 * TODO(markus): Add upvec_16BitsToUTrie2() function that enumerates all rows, extracts
 * some 16-bit field and builds and returns a UTrie2.
 */

U_CAPI void U_CALLCONV
upvec_compactToUTrie2Handler(void *context,
                             UChar32 start, UChar32 end,
                             int32_t rowIndex, uint32_t *row, int32_t columns,
                             UErrorCode *pErrorCode) {
    (void)row;
    (void)columns;
    UPVecToUTrie2Context *toUTrie2=(UPVecToUTrie2Context *)context;
    if(start<UPVEC_FIRST_SPECIAL_CP) {
        utrie2_setRange32(toUTrie2->trie, start, end, (uint32_t)rowIndex, TRUE, pErrorCode);
    } else {
        switch(start) {
        case UPVEC_INITIAL_VALUE_CP:
            toUTrie2->initialValue=rowIndex;
            break;
        case UPVEC_ERROR_VALUE_CP:
            toUTrie2->errorValue=rowIndex;
            break;
        case UPVEC_START_REAL_VALUES_CP:
            toUTrie2->maxValue=rowIndex;
            if(rowIndex>0xffff) {
                /* too many rows for a 16-bit trie */
                *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            } else {
                toUTrie2->trie=utrie2_open(toUTrie2->initialValue,
                                           toUTrie2->errorValue, pErrorCode);
            }
            break;
        default:
            break;
        }
    }
}
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// *******************************************************************************
// *
// *   Copyright (C) 2002-2011, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *
// *******************************************************************************
// *   file name:  punycode.cpp
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2002jan31
// *   created by: Markus W. Scherer
// */


// /* This ICU code derived from: */
// /*
// punycode.c 0.4.0 (2001-Nov-17-Sat)
// http://www.cs.berkeley.edu/~amc/idn/
// Adam M. Costello
// http://www.nicemice.net/amc/

// Disclaimer and license

//     Regarding this entire document or any portion of it (including
//     the pseudocode and C code), the author makes no guarantees and
//     is not responsible for any damage resulting from its use.  The
//     author grants irrevocable permission to anyone to use, modify,
//     and distribute it in any way that does not diminish the rights
//     of anyone else to use, modify, and distribute it, provided that
//     redistributed derivative works do not contain misleading author or
//     version information.  Derivative works need not be licensed under
//     similar terms.
// */
// /*
//  * ICU modifications:
//  * - ICU data types and coding conventions
//  * - ICU string buffer handling with implicit source lengths
//  *   and destination preflighting
//  * - UTF-16 handling
//  */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_IDNA

// #include "unicode/ustring.h"
// #include "unicode/utf.h"
// #include "unicode/utf16.h"
// #include "ustr_imp.h"
// #include "cstring.h"
// #include "cmemory.h"
// #include "punycode.h"
// #include "uassert.h"


// /* Punycode ----------------------------------------------------------------- */

// /* Punycode parameters for Bootstring */
// #define BASE            36
// #define TMIN            1
// #define TMAX            26
// #define SKEW            38
// #define DAMP            700
// #define INITIAL_BIAS    72
// #define INITIAL_N       0x80

// /* "Basic" Unicode/ASCII code points */
// #define _HYPHEN         0X2d
// #define DELIMITER       _HYPHEN

// #define _ZERO_          0X30
// #define _NINE           0x39

// #define _SMALL_A        0X61
// #define _SMALL_Z        0X7a

// #define _CAPITAL_A      0X41
// #define _CAPITAL_Z      0X5a

// #define IS_BASIC(c) ((c)<0x80)
// #define IS_BASIC_UPPERCASE(c) (_CAPITAL_A<=(c) && (c)<=_CAPITAL_Z)

// /**
//  * digitToBasic() returns the basic code point whose value
//  * (when used for representing integers) is d, which must be in the
//  * range 0 to BASE-1. The lowercase form is used unless the uppercase flag is
//  * nonzero, in which case the uppercase form is used.
//  */
// static inline char
// digitToBasic(int32_t digit, UBool uppercase) {
//     /*  0..25 map to ASCII a..z or A..Z */
//     /* 26..35 map to ASCII 0..9         */
//     if(digit<26) {
//         if(uppercase) {
//             return (char)(_CAPITAL_A+digit);
//         } else {
//             return (char)(_SMALL_A+digit);
//         }
//     } else {
//         return (char)((_ZERO_-26)+digit);
//     }
// }

// /**
//  * basicToDigit[] contains the numeric value of a basic code
//  * point (for use in representing integers) in the range 0 to
//  * BASE-1, or -1 if b is does not represent a value.
//  */
// static const int8_t
// basicToDigit[256]={
//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,

//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
//     26, 27, 28, 29, 30, 31, 32, 33, 34, 35, -1, -1, -1, -1, -1, -1,

//     -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
//     15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,

//     -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
//     15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,

//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,

//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,

//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,

//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
//     -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
// };

// static inline char
// asciiCaseMap(char b, UBool uppercase) {
//     if(uppercase) {
//         if(_SMALL_A<=b && b<=_SMALL_Z) {
//             b-=(_SMALL_A-_CAPITAL_A);
//         }
//     } else {
//         if(_CAPITAL_A<=b && b<=_CAPITAL_Z) {
//             b+=(_SMALL_A-_CAPITAL_A);
//         }
//     }
//     return b;
// }

// /* Punycode-specific Bootstring code ---------------------------------------- */

// /*
//  * The following code omits the {parts} of the pseudo-algorithm in the spec
//  * that are not used with the Punycode parameter set.
//  */

// /* Bias adaptation function. */
// static int32_t
// adaptBias(int32_t delta, int32_t length, UBool firstTime) {
//     int32_t count;

//     if(firstTime) {
//         delta/=DAMP;
//     } else {
//         delta/=2;
//     }

//     delta+=delta/length;
//     for(count=0; delta>((BASE-TMIN)*TMAX)/2; count+=BASE) {
//         delta/=(BASE-TMIN);
//     }

//     return count+(((BASE-TMIN+1)*delta)/(delta+SKEW));
// }

// #define MAX_CP_COUNT    200

// U_CFUNC int32_t
// u_strToPunycode(const UChar *src, int32_t srcLength,
//                 UChar *dest, int32_t destCapacity,
//                 const UBool *caseFlags,
//                 UErrorCode *pErrorCode) {

//     int32_t cpBuffer[MAX_CP_COUNT];
//     int32_t n, delta, handledCPCount, basicLength, destLength, bias, j, m, q, k, t, srcCPCount;
//     UChar c, c2;

//     /* argument checking */
//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     if(src==NULL || srcLength<-1 || (dest==NULL && destCapacity!=0)) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     /*
//      * Handle the basic code points and
//      * convert extended ones to UTF-32 in cpBuffer (caseFlag in sign bit):
//      */
//     srcCPCount=destLength=0;
//     if(srcLength==-1) {
//         /* NUL-terminated input */
//         for(j=0; /* no condition */; ++j) {
//             if((c=src[j])==0) {
//                 break;
//             }
//             if(srcCPCount==MAX_CP_COUNT) {
//                 /* too many input code points */
//                 *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//                 return 0;
//             }
//             if(IS_BASIC(c)) {
//                 cpBuffer[srcCPCount++]=0;
//                 if(destLength<destCapacity) {
//                     dest[destLength]=
//                         caseFlags!=NULL ?
//                             asciiCaseMap((char)c, caseFlags[j]) :
//                             (char)c;
//                 }
//                 ++destLength;
//             } else {
//                 n=(caseFlags!=NULL && caseFlags[j])<<31L;
//                 if(U16_IS_SINGLE(c)) {
//                     n|=c;
//                 } else if(U16_IS_LEAD(c) && U16_IS_TRAIL(c2=src[j+1])) {
//                     ++j;
//                     n|=(int32_t)U16_GET_SUPPLEMENTARY(c, c2);
//                 } else {
//                     /* error: unmatched surrogate */
//                     *pErrorCode=U_INVALID_CHAR_FOUND;
//                     return 0;
//                 }
//                 cpBuffer[srcCPCount++]=n;
//             }
//         }
//     } else {
//         /* length-specified input */
//         for(j=0; j<srcLength; ++j) {
//             if(srcCPCount==MAX_CP_COUNT) {
//                 /* too many input code points */
//                 *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//                 return 0;
//             }
//             c=src[j];
//             if(IS_BASIC(c)) {
//                 cpBuffer[srcCPCount++]=0;
//                 if(destLength<destCapacity) {
//                     dest[destLength]=
//                         caseFlags!=NULL ?
//                             asciiCaseMap((char)c, caseFlags[j]) :
//                             (char)c;
//                 }
//                 ++destLength;
//             } else {
//                 n=(caseFlags!=NULL && caseFlags[j])<<31L;
//                 if(U16_IS_SINGLE(c)) {
//                     n|=c;
//                 } else if(U16_IS_LEAD(c) && (j+1)<srcLength && U16_IS_TRAIL(c2=src[j+1])) {
//                     ++j;
//                     n|=(int32_t)U16_GET_SUPPLEMENTARY(c, c2);
//                 } else {
//                     /* error: unmatched surrogate */
//                     *pErrorCode=U_INVALID_CHAR_FOUND;
//                     return 0;
//                 }
//                 cpBuffer[srcCPCount++]=n;
//             }
//         }
//     }

//     /* Finish the basic string - if it is not empty - with a delimiter. */
//     basicLength=destLength;
//     if(basicLength>0) {
//         if(destLength<destCapacity) {
//             dest[destLength]=DELIMITER;
//         }
//         ++destLength;
//     }

//     /*
//      * handledCPCount is the number of code points that have been handled
//      * basicLength is the number of basic code points
//      * destLength is the number of chars that have been output
//      */

//     /* Initialize the state: */
//     n=INITIAL_N;
//     delta=0;
//     bias=INITIAL_BIAS;

//     /* Main encoding loop: */
//     for(handledCPCount=basicLength; handledCPCount<srcCPCount; /* no op */) {
//         /*
//          * All non-basic code points < n have been handled already.
//          * Find the next larger one:
//          */
//         for(m=0x7fffffff, j=0; j<srcCPCount; ++j) {
//             q=cpBuffer[j]&0x7fffffff; /* remove case flag from the sign bit */
//             if(n<=q && q<m) {
//                 m=q;
//             }
//         }

//         /*
//          * Increase delta enough to advance the decoder's
//          * <n,i> state to <m,0>, but guard against overflow:
//          */
//         if(m-n>(0x7fffffff-MAX_CP_COUNT-delta)/(handledCPCount+1)) {
//             *pErrorCode=U_INTERNAL_PROGRAM_ERROR;
//             return 0;
//         }
//         delta+=(m-n)*(handledCPCount+1);
//         n=m;

//         /* Encode a sequence of same code points n */
//         for(j=0; j<srcCPCount; ++j) {
//             q=cpBuffer[j]&0x7fffffff; /* remove case flag from the sign bit */
//             if(q<n) {
//                 ++delta;
//             } else if(q==n) {
//                 /* Represent delta as a generalized variable-length integer: */
//                 for(q=delta, k=BASE; /* no condition */; k+=BASE) {

//                     /** RAM: comment out the old code for conformance with draft-ietf-idn-punycode-03.txt

//                     t=k-bias;
//                     if(t<TMIN) {
//                         t=TMIN;
//                     } else if(t>TMAX) {
//                         t=TMAX;
//                     }
//                     */

//                     t=k-bias;
//                     if(t<TMIN) {
//                         t=TMIN;
//                     } else if(k>=(bias+TMAX)) {
//                         t=TMAX;
//                     }

//                     if(q<t) {
//                         break;
//                     }

//                     if(destLength<destCapacity) {
//                         dest[destLength]=digitToBasic(t+(q-t)%(BASE-t), 0);
//                     }
//                     ++destLength;
//                     q=(q-t)/(BASE-t);
//                 }

//                 if(destLength<destCapacity) {
//                     dest[destLength]=digitToBasic(q, (UBool)(cpBuffer[j]<0));
//                 }
//                 ++destLength;
//                 bias=adaptBias(delta, handledCPCount+1, (UBool)(handledCPCount==basicLength));
//                 delta=0;
//                 ++handledCPCount;
//             }
//         }

//         ++delta;
//         ++n;
//     }

//     return u_terminateUChars(dest, destCapacity, destLength, pErrorCode);
// }

// U_CFUNC int32_t
// u_strFromPunycode(const UChar *src, int32_t srcLength,
//                   UChar *dest, int32_t destCapacity,
//                   UBool *caseFlags,
//                   UErrorCode *pErrorCode) {
//     int32_t n, destLength, i, bias, basicLength, j, in, oldi, w, k, digit, t,
//             destCPCount, firstSupplementaryIndex, cpLength;
//     UChar b;

//     /* argument checking */
//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     if(src==NULL || srcLength<-1 || (dest==NULL && destCapacity!=0)) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     if(srcLength==-1) {
//         srcLength=u_strlen(src);
//     }

//     /*
//      * Handle the basic code points:
//      * Let basicLength be the number of input code points
//      * before the last delimiter, or 0 if there is none,
//      * then copy the first basicLength code points to the output.
//      *
//      * The two following loops iterate backward.
//      */
//     for(j=srcLength; j>0;) {
//         if(src[--j]==DELIMITER) {
//             break;
//         }
//     }
//     destLength=basicLength=destCPCount=j;
//     U_ASSERT(destLength>=0);

//     while(j>0) {
//         b=src[--j];
//         if(!IS_BASIC(b)) {
//             *pErrorCode=U_INVALID_CHAR_FOUND;
//             return 0;
//         }

//         if(j<destCapacity) {
//             dest[j]=(UChar)b;

//             if(caseFlags!=NULL) {
//                 caseFlags[j]=IS_BASIC_UPPERCASE(b);
//             }
//         }
//     }

//     /* Initialize the state: */
//     n=INITIAL_N;
//     i=0;
//     bias=INITIAL_BIAS;
//     firstSupplementaryIndex=1000000000;

//     /*
//      * Main decoding loop:
//      * Start just after the last delimiter if any
//      * basic code points were copied; start at the beginning otherwise.
//      */
//     for(in=basicLength>0 ? basicLength+1 : 0; in<srcLength; /* no op */) {
//         /*
//          * in is the index of the next character to be consumed, and
//          * destCPCount is the number of code points in the output array.
//          *
//          * Decode a generalized variable-length integer into delta,
//          * which gets added to i.  The overflow checking is easier
//          * if we increase i as we go, then subtract off its starting
//          * value at the end to obtain delta.
//          */
//         for(oldi=i, w=1, k=BASE; /* no condition */; k+=BASE) {
//             if(in>=srcLength) {
//                 *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 return 0;
//             }

//             digit=basicToDigit[(uint8_t)src[in++]];
//             if(digit<0) {
//                 *pErrorCode=U_INVALID_CHAR_FOUND;
//                 return 0;
//             }
//             if(digit>(0x7fffffff-i)/w) {
//                 /* integer overflow */
//                 *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 return 0;
//             }

//             i+=digit*w;
//             /** RAM: comment out the old code for conformance with draft-ietf-idn-punycode-03.txt
//             t=k-bias;
//             if(t<TMIN) {
//                 t=TMIN;
//             } else if(t>TMAX) {
//                 t=TMAX;
//             }
//             */
//             t=k-bias;
//             if(t<TMIN) {
//                 t=TMIN;
//             } else if(k>=(bias+TMAX)) {
//                 t=TMAX;
//             }
//             if(digit<t) {
//                 break;
//             }

//             if(w>0x7fffffff/(BASE-t)) {
//                 /* integer overflow */
//                 *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 return 0;
//             }
//             w*=BASE-t;
//         }

//         /*
//          * Modification from sample code:
//          * Increments destCPCount here,
//          * where needed instead of in for() loop tail.
//          */
//         ++destCPCount;
//         bias=adaptBias(i-oldi, destCPCount, (UBool)(oldi==0));

//         /*
//          * i was supposed to wrap around from (incremented) destCPCount to 0,
//          * incrementing n each time, so we'll fix that now:
//          */
//         if(i/destCPCount>(0x7fffffff-n)) {
//             /* integer overflow */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//             return 0;
//         }

//         n+=i/destCPCount;
//         i%=destCPCount;
//         /* not needed for Punycode: */
//         /* if (decode_digit(n) <= BASE) return punycode_invalid_input; */

//         if(n>0x10ffff || U_IS_SURROGATE(n)) {
//             /* Unicode code point overflow */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//             return 0;
//         }

//         /* Insert n at position i of the output: */
//         cpLength=U16_LENGTH(n);
//         if(dest!=NULL && ((destLength+cpLength)<=destCapacity)) {
//             int32_t codeUnitIndex;

//             /*
//              * Handle indexes when supplementary code points are present.
//              *
//              * In almost all cases, there will be only BMP code points before i
//              * and even in the entire string.
//              * This is handled with the same efficiency as with UTF-32.
//              *
//              * Only the rare cases with supplementary code points are handled
//              * more slowly - but not too bad since this is an insertion anyway.
//              */
//             if(i<=firstSupplementaryIndex) {
//                 codeUnitIndex=i;
//                 if(cpLength>1) {
//                     firstSupplementaryIndex=codeUnitIndex;
//                 } else {
//                     ++firstSupplementaryIndex;
//                 }
//             } else {
//                 codeUnitIndex=firstSupplementaryIndex;
//                 U16_FWD_N(dest, codeUnitIndex, destLength, i-codeUnitIndex);
//             }

//             /* use the UChar index codeUnitIndex instead of the code point index i */
//             if(codeUnitIndex<destLength) {
//                 uprv_memmove(dest+codeUnitIndex+cpLength,
//                              dest+codeUnitIndex,
//                              (destLength-codeUnitIndex)*U_SIZEOF_UCHAR);
//                 if(caseFlags!=NULL) {
//                     uprv_memmove(caseFlags+codeUnitIndex+cpLength,
//                                  caseFlags+codeUnitIndex,
//                                  destLength-codeUnitIndex);
//                 }
//             }
//             if(cpLength==1) {
//                 /* BMP, insert one code unit */
//                 dest[codeUnitIndex]=(UChar)n;
//             } else {
//                 /* supplementary character, insert two code units */
//                 dest[codeUnitIndex]=U16_LEAD(n);
//                 dest[codeUnitIndex+1]=U16_TRAIL(n);
//             }
//             if(caseFlags!=NULL) {
//                 /* Case of last character determines uppercase flag: */
//                 caseFlags[codeUnitIndex]=IS_BASIC_UPPERCASE(src[in-1]);
//                 if(cpLength==2) {
//                     caseFlags[codeUnitIndex+1]=FALSE;
//                 }
//             }
//         }
//         destLength+=cpLength;
//         U_ASSERT(destLength>=0);
//         ++i;
//     }

//     return u_terminateUChars(dest, destCapacity, destLength, pErrorCode);
// }

// /* ### check notes on overflow handling - only necessary if not IDNA? are these Punycode functions to be public? */

// #endif /* #if !UCONFIG_NO_IDNA */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
***************************************************************************
*   Copyright (C) 1999-2016 International Business Machines Corporation
*   and others. All rights reserved.
***************************************************************************
*/
//
//  file:  rbbi.cpp  Contains the implementation of the rule based break iterator
//                   runtime engine and the API implementation for
//                   class RuleBasedBreakIterator
//

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2012-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*/


// Windows header <typeinfo> does not define 'exception' in 'std' namespace.
// Therefore, a project using ICU cannot be compiled with _HAS_EXCEPTIONS
// set to 0 on Windows with Visual Studio. To work around that, we have to
// include <exception> explicitly and add using statement below.
// Whenever 'typeid' is used, this header has to be included
// instead of <typeinfo>.
// Visual Studio 10 emits warning 4275 with this change. If you compile
// with exception disabled, you have to suppress warning 4275.
#if defined(_MSC_VER) && _HAS_EXCEPTIONS == 0
#include <exception>
using std::exception;
#endif
#if defined(__GLIBCXX__)
namespace std { class type_info; } // WORKAROUND: http://llvm.org/bugs/show_bug.cgi?id=13364
#endif
#include <typeinfo>  // for 'typeid' to work

  // for 'typeid' to work



// #if !UCONFIG_NO_BREAK_ITERATION

// #include <cinttypes>

// #include "unicode/rbbi.h"
// #include "unicode/schriter.h"
// #include "unicode/uchriter.h"
// #include "unicode/uclean.h"
// #include "unicode/udata.h"

// #include "brkeng.h"
// #include "ucln_cmn.h"
// #include "cmemory.h"
// #include "cstring.h"
// #include "localsvc.h"
// #include "rbbidata.h"
// #include "rbbi_cache.h"
// #include "rbbirb.h"
// #include "uassert.h"
// #include "umutex.h"
// #include "uvectr32.h"

// #ifdef RBBI_DEBUG
// static UBool gTrace = FALSE;
// #endif

// U_NAMESPACE_BEGIN

// // The state number of the starting state
// constexpr int32_t START_STATE = 1;

// // The state-transition value indicating "stop"
// constexpr int32_t STOP_STATE = 0;


// UOBJECT_DEFINE_RTTI_IMPLEMENTATION(RuleBasedBreakIterator)


// //=======================================================================
// // constructors
// //=======================================================================

// /**
//  * Constructs a RuleBasedBreakIterator that uses the already-created
//  * tables object that is passed in as a parameter.
//  */
// RuleBasedBreakIterator::RuleBasedBreakIterator(RBBIDataHeader* data, UErrorCode &status)
//  : fSCharIter(UnicodeString())
// {
//     init(status);
//     fData = new RBBIDataWrapper(data, status); // status checked in constructor
//     if (U_FAILURE(status)) {return;}
//     if(fData == 0) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
// }

// //
// //  Construct from precompiled binary rules (tables).  This constructor is public API,
// //  taking the rules as a (const uint8_t *) to match the type produced by getBinaryRules().
// //
// RuleBasedBreakIterator::RuleBasedBreakIterator(const uint8_t *compiledRules,
//                        uint32_t       ruleLength,
//                        UErrorCode     &status)
//  : fSCharIter(UnicodeString())
// {
//     init(status);
//     if (U_FAILURE(status)) {
//         return;
//     }
//     if (compiledRules == NULL || ruleLength < sizeof(RBBIDataHeader)) {
//         status = U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }
//     const RBBIDataHeader *data = (const RBBIDataHeader *)compiledRules;
//     if (data->fLength > ruleLength) {
//         status = U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }
//     fData = new RBBIDataWrapper(data, RBBIDataWrapper::kDontAdopt, status);
//     if (U_FAILURE(status)) {return;}
//     if(fData == 0) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
// }


// //-------------------------------------------------------------------------------
// //
// //   Constructor   from a UDataMemory handle to precompiled break rules
// //                 stored in an ICU data file.
// //
// //-------------------------------------------------------------------------------
// RuleBasedBreakIterator::RuleBasedBreakIterator(UDataMemory* udm, UErrorCode &status)
//  : fSCharIter(UnicodeString())
// {
//     init(status);
//     fData = new RBBIDataWrapper(udm, status); // status checked in constructor
//     if (U_FAILURE(status)) {return;}
//     if(fData == 0) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
// }



// //-------------------------------------------------------------------------------
// //
// //   Constructor       from a set of rules supplied as a string.
// //
// //-------------------------------------------------------------------------------
// RuleBasedBreakIterator::RuleBasedBreakIterator( const UnicodeString  &rules,
//                                                 UParseError          &parseError,
//                                                 UErrorCode           &status)
//  : fSCharIter(UnicodeString())
// {
//     init(status);
//     if (U_FAILURE(status)) {return;}
//     RuleBasedBreakIterator *bi = (RuleBasedBreakIterator *)
//         RBBIRuleBuilder::createRuleBasedBreakIterator(rules, &parseError, status);
//     // Note:  This is a bit awkward.  The RBBI ruleBuilder has a factory method that
//     //        creates and returns a complete RBBI.  From here, in a constructor, we
//     //        can't just return the object created by the builder factory, hence
//     //        the assignment of the factory created object to "this".
//     if (U_SUCCESS(status)) {
//         *this = *bi;
//         delete bi;
//     }
// }


// //-------------------------------------------------------------------------------
// //
// // Default Constructor.      Create an empty shell that can be set up later.
// //                           Used when creating a RuleBasedBreakIterator from a set
// //                           of rules.
// //-------------------------------------------------------------------------------
// RuleBasedBreakIterator::RuleBasedBreakIterator()
//  : fSCharIter(UnicodeString())
// {
//     UErrorCode status = U_ZERO_ERROR;
//     init(status);
// }


// //-------------------------------------------------------------------------------
// //
// //   Copy constructor.  Will produce a break iterator with the same behavior,
// //                      and which iterates over the same text, as the one passed in.
// //
// //-------------------------------------------------------------------------------
// RuleBasedBreakIterator::RuleBasedBreakIterator(const RuleBasedBreakIterator& other)
// : BreakIterator(other),
//   fSCharIter(UnicodeString())
// {
//     UErrorCode status = U_ZERO_ERROR;
//     this->init(status);
//     *this = other;
// }


// /**
//  * Destructor
//  */
// RuleBasedBreakIterator::~RuleBasedBreakIterator() {
//     if (fCharIter != &fSCharIter) {
//         // fCharIter was adopted from the outside.
//         delete fCharIter;
//     }
//     fCharIter = NULL;

//     utext_close(&fText);

//     if (fData != NULL) {
//         fData->removeReference();
//         fData = NULL;
//     }
//     delete fBreakCache;
//     fBreakCache = NULL;

//     delete fDictionaryCache;
//     fDictionaryCache = NULL;

//     delete fLanguageBreakEngines;
//     fLanguageBreakEngines = NULL;

//     delete fUnhandledBreakEngine;
//     fUnhandledBreakEngine = NULL;
// }

// /**
//  * Assignment operator.  Sets this iterator to have the same behavior,
//  * and iterate over the same text, as the one passed in.
//  */
// RuleBasedBreakIterator&
// RuleBasedBreakIterator::operator=(const RuleBasedBreakIterator& that) {
//     if (this == &that) {
//         return *this;
//     }
//     BreakIterator::operator=(that);

//     if (fLanguageBreakEngines != NULL) {
//         delete fLanguageBreakEngines;
//         fLanguageBreakEngines = NULL;   // Just rebuild for now
//     }
//     // TODO: clone fLanguageBreakEngines from "that"
//     UErrorCode status = U_ZERO_ERROR;
//     utext_clone(&fText, &that.fText, FALSE, TRUE, &status);

//     if (fCharIter != &fSCharIter) {
//         delete fCharIter;
//     }
//     fCharIter = &fSCharIter;

//     if (that.fCharIter != NULL && that.fCharIter != &that.fSCharIter) {
//         // This is a little bit tricky - it will intially appear that
//         //  this->fCharIter is adopted, even if that->fCharIter was
//         //  not adopted.  That's ok.
//         fCharIter = that.fCharIter->clone();
//     }
//     fSCharIter = that.fSCharIter;
//     if (fCharIter == NULL) {
//         fCharIter = &fSCharIter;
//     }

//     if (fData != NULL) {
//         fData->removeReference();
//         fData = NULL;
//     }
//     if (that.fData != NULL) {
//         fData = that.fData->addReference();
//     }

//     fPosition = that.fPosition;
//     fRuleStatusIndex = that.fRuleStatusIndex;
//     fDone = that.fDone;

//     // TODO: both the dictionary and the main cache need to be copied.
//     //       Current position could be within a dictionary range. Trying to continue
//     //       the iteration without the caches present would go to the rules, with
//     //       the assumption that the current position is on a rule boundary.
//     fBreakCache->reset(fPosition, fRuleStatusIndex);
//     fDictionaryCache->reset();

//     return *this;
// }



// //-----------------------------------------------------------------------------
// //
// //    init()      Shared initialization routine.   Used by all the constructors.
// //                Initializes all fields, leaving the object in a consistent state.
// //
// //-----------------------------------------------------------------------------
// void RuleBasedBreakIterator::init(UErrorCode &status) {
//     fCharIter             = NULL;
//     fData                 = NULL;
//     fPosition             = 0;
//     fRuleStatusIndex      = 0;
//     fDone                 = false;
//     fDictionaryCharCount  = 0;
//     fLanguageBreakEngines = NULL;
//     fUnhandledBreakEngine = NULL;
//     fBreakCache           = NULL;
//     fDictionaryCache      = NULL;

//     // Note: IBM xlC is unable to assign or initialize member fText from UTEXT_INITIALIZER.
//     // fText                 = UTEXT_INITIALIZER;
//     static const UText initializedUText = UTEXT_INITIALIZER;
//     uprv_memcpy(&fText, &initializedUText, sizeof(UText));

//    if (U_FAILURE(status)) {
//         return;
//     }

//     utext_openUChars(&fText, NULL, 0, &status);
//     fDictionaryCache = new DictionaryCache(this, status);
//     fBreakCache      = new BreakCache(this, status);
//     if (U_SUCCESS(status) && (fDictionaryCache == NULL || fBreakCache == NULL)) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//     }

// #ifdef RBBI_DEBUG
//     static UBool debugInitDone = FALSE;
//     if (debugInitDone == FALSE) {
//         char *debugEnv = getenv("U_RBBIDEBUG");
//         if (debugEnv && uprv_strstr(debugEnv, "trace")) {
//             gTrace = TRUE;
//         }
//         debugInitDone = TRUE;
//     }
// #endif
// }



// //-----------------------------------------------------------------------------
// //
// //    clone - Returns a newly-constructed RuleBasedBreakIterator with the same
// //            behavior, and iterating over the same text, as this one.
// //            Virtual function: does the right thing with subclasses.
// //
// //-----------------------------------------------------------------------------
// RuleBasedBreakIterator*
// RuleBasedBreakIterator::clone() const {
//     return new RuleBasedBreakIterator(*this);
// }

// /**
//  * Equality operator.  Returns TRUE if both BreakIterators are of the
//  * same class, have the same behavior, and iterate over the same text.
//  */
// UBool
// RuleBasedBreakIterator::operator==(const BreakIterator& that) const {
//     if (typeid(*this) != typeid(that)) {
//         return FALSE;
//     }
//     if (this == &that) {
//         return TRUE;
//     }

//     // The base class BreakIterator carries no state that participates in equality,
//     // and does not implement an equality function that would otherwise be
//     // checked at this point.

//     const RuleBasedBreakIterator& that2 = (const RuleBasedBreakIterator&) that;

//     if (!utext_equals(&fText, &that2.fText)) {
//         // The two break iterators are operating on different text,
//         //   or have a different iteration position.
//         //   Note that fText's position is always the same as the break iterator's position.
//         return FALSE;
//     }

//     if (!(fPosition == that2.fPosition &&
//             fRuleStatusIndex == that2.fRuleStatusIndex &&
//             fDone == that2.fDone)) {
//         return FALSE;
//     }

//     if (that2.fData == fData ||
//         (fData != NULL && that2.fData != NULL && *that2.fData == *fData)) {
//             // The two break iterators are using the same rules.
//             return TRUE;
//         }
//     return FALSE;
// }

// /**
//  * Compute a hash code for this BreakIterator
//  * @return A hash code
//  */
// int32_t
// RuleBasedBreakIterator::hashCode(void) const {
//     int32_t   hash = 0;
//     if (fData != NULL) {
//         hash = fData->hashCode();
//     }
//     return hash;
// }


// void RuleBasedBreakIterator::setText(UText *ut, UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return;
//     }
//     fBreakCache->reset();
//     fDictionaryCache->reset();
//     utext_clone(&fText, ut, FALSE, TRUE, &status);

//     // Set up a dummy CharacterIterator to be returned if anyone
//     //   calls getText().  With input from UText, there is no reasonable
//     //   way to return a characterIterator over the actual input text.
//     //   Return one over an empty string instead - this is the closest
//     //   we can come to signaling a failure.
//     //   (GetText() is obsolete, this failure is sort of OK)
//     fSCharIter.setText(UnicodeString());

//     if (fCharIter != &fSCharIter) {
//         // existing fCharIter was adopted from the outside.  Delete it now.
//         delete fCharIter;
//     }
//     fCharIter = &fSCharIter;

//     this->first();
// }


// UText *RuleBasedBreakIterator::getUText(UText *fillIn, UErrorCode &status) const {
//     UText *result = utext_clone(fillIn, &fText, FALSE, TRUE, &status);
//     return result;
// }


// //=======================================================================
// // BreakIterator overrides
// //=======================================================================

// /**
//  * Return a CharacterIterator over the text being analyzed.
//  */
// CharacterIterator&
// RuleBasedBreakIterator::getText() const {
//     return *fCharIter;
// }

// /**
//  * Set the iterator to analyze a new piece of text.  This function resets
//  * the current iteration position to the beginning of the text.
//  * @param newText An iterator over the text to analyze.
//  */
// void
// RuleBasedBreakIterator::adoptText(CharacterIterator* newText) {
//     // If we are holding a CharacterIterator adopted from a
//     //   previous call to this function, delete it now.
//     if (fCharIter != &fSCharIter) {
//         delete fCharIter;
//     }

//     fCharIter = newText;
//     UErrorCode status = U_ZERO_ERROR;
//     fBreakCache->reset();
//     fDictionaryCache->reset();
//     if (newText==NULL || newText->startIndex() != 0) {
//         // startIndex !=0 wants to be an error, but there's no way to report it.
//         // Make the iterator text be an empty string.
//         utext_openUChars(&fText, NULL, 0, &status);
//     } else {
//         utext_openCharacterIterator(&fText, newText, &status);
//     }
//     this->first();
// }

// /**
//  * Set the iterator to analyze a new piece of text.  This function resets
//  * the current iteration position to the beginning of the text.
//  * @param newText An iterator over the text to analyze.
//  */
// void
// RuleBasedBreakIterator::setText(const UnicodeString& newText) {
//     UErrorCode status = U_ZERO_ERROR;
//     fBreakCache->reset();
//     fDictionaryCache->reset();
//     utext_openConstUnicodeString(&fText, &newText, &status);

//     // Set up a character iterator on the string.
//     //   Needed in case someone calls getText().
//     //  Can not, unfortunately, do this lazily on the (probably never)
//     //  call to getText(), because getText is const.
//     fSCharIter.setText(newText);

//     if (fCharIter != &fSCharIter) {
//         // old fCharIter was adopted from the outside.  Delete it.
//         delete fCharIter;
//     }
//     fCharIter = &fSCharIter;

//     this->first();
// }


// /**
//  *  Provide a new UText for the input text.  Must reference text with contents identical
//  *  to the original.
//  *  Intended for use with text data originating in Java (garbage collected) environments
//  *  where the data may be moved in memory at arbitrary times.
//  */
// RuleBasedBreakIterator &RuleBasedBreakIterator::refreshInputText(UText *input, UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return *this;
//     }
//     if (input == NULL) {
//         status = U_ILLEGAL_ARGUMENT_ERROR;
//         return *this;
//     }
//     int64_t pos = utext_getNativeIndex(&fText);
//     //  Shallow read-only clone of the new UText into the existing input UText
//     utext_clone(&fText, input, FALSE, TRUE, &status);
//     if (U_FAILURE(status)) {
//         return *this;
//     }
//     utext_setNativeIndex(&fText, pos);
//     if (utext_getNativeIndex(&fText) != pos) {
//         // Sanity check.  The new input utext is supposed to have the exact same
//         // contents as the old.  If we can't set to the same position, it doesn't.
//         // The contents underlying the old utext might be invalid at this point,
//         // so it's not safe to check directly.
//         status = U_ILLEGAL_ARGUMENT_ERROR;
//     }
//     return *this;
// }


// /**
//  * Sets the current iteration position to the beginning of the text, position zero.
//  * @return The new iterator position, which is zero.
//  */
// int32_t RuleBasedBreakIterator::first(void) {
//     UErrorCode status = U_ZERO_ERROR;
//     if (!fBreakCache->seek(0)) {
//         fBreakCache->populateNear(0, status);
//     }
//     fBreakCache->current();
//     U_ASSERT(fPosition == 0);
//     return 0;
// }

// /**
//  * Sets the current iteration position to the end of the text.
//  * @return The text's past-the-end offset.
//  */
// int32_t RuleBasedBreakIterator::last(void) {
//     int32_t endPos = (int32_t)utext_nativeLength(&fText);
//     UBool endShouldBeBoundary = isBoundary(endPos);      // Has side effect of setting iterator position.
//     (void)endShouldBeBoundary;
//     U_ASSERT(endShouldBeBoundary);
//     U_ASSERT(fPosition == endPos);
//     return endPos;
// }

// /**
//  * Advances the iterator either forward or backward the specified number of steps.
//  * Negative values move backward, and positive values move forward.  This is
//  * equivalent to repeatedly calling next() or previous().
//  * @param n The number of steps to move.  The sign indicates the direction
//  * (negative is backwards, and positive is forwards).
//  * @return The character offset of the boundary position n boundaries away from
//  * the current one.
//  */
// int32_t RuleBasedBreakIterator::next(int32_t n) {
//     int32_t result = 0;
//     if (n > 0) {
//         for (; n > 0 && result != UBRK_DONE; --n) {
//             result = next();
//         }
//     } else if (n < 0) {
//         for (; n < 0 && result != UBRK_DONE; ++n) {
//             result = previous();
//         }
//     } else {
//         result = current();
//     }
//     return result;
// }

// /**
//  * Advances the iterator to the next boundary position.
//  * @return The position of the first boundary after this one.
//  */
// int32_t RuleBasedBreakIterator::next(void) {
//     fBreakCache->next();
//     return fDone ? UBRK_DONE : fPosition;
// }

// /**
//  * Move the iterator backwards, to the boundary preceding the current one.
//  *
//  *         Starts from the current position within fText.
//  *         Starting position need not be on a boundary.
//  *
//  * @return The position of the boundary position immediately preceding the starting position.
//  */
// int32_t RuleBasedBreakIterator::previous(void) {
//     UErrorCode status = U_ZERO_ERROR;
//     fBreakCache->previous(status);
//     return fDone ? UBRK_DONE : fPosition;
// }

// /**
//  * Sets the iterator to refer to the first boundary position following
//  * the specified position.
//  * @param startPos The position from which to begin searching for a break position.
//  * @return The position of the first break after the current position.
//  */
// int32_t RuleBasedBreakIterator::following(int32_t startPos) {
//     // if the supplied position is before the beginning, return the
//     // text's starting offset
//     if (startPos < 0) {
//         return first();
//     }

//     // Move requested offset to a code point start. It might be on a trail surrogate,
//     // or on a trail byte if the input is UTF-8. Or it may be beyond the end of the text.
//     utext_setNativeIndex(&fText, startPos);
//     startPos = (int32_t)utext_getNativeIndex(&fText);

//     UErrorCode status = U_ZERO_ERROR;
//     fBreakCache->following(startPos, status);
//     return fDone ? UBRK_DONE : fPosition;
// }

// /**
//  * Sets the iterator to refer to the last boundary position before the
//  * specified position.
//  * @param offset The position to begin searching for a break from.
//  * @return The position of the last boundary before the starting position.
//  */
// int32_t RuleBasedBreakIterator::preceding(int32_t offset) {
//     if (offset > utext_nativeLength(&fText)) {
//         return last();
//     }

//     // Move requested offset to a code point start. It might be on a trail surrogate,
//     // or on a trail byte if the input is UTF-8.

//     utext_setNativeIndex(&fText, offset);
//     int32_t adjustedOffset = static_cast<int32_t>(utext_getNativeIndex(&fText));

//     UErrorCode status = U_ZERO_ERROR;
//     fBreakCache->preceding(adjustedOffset, status);
//     return fDone ? UBRK_DONE : fPosition;
// }

// /**
//  * Returns true if the specfied position is a boundary position.  As a side
//  * effect, leaves the iterator pointing to the first boundary position at
//  * or after "offset".
//  *
//  * @param offset the offset to check.
//  * @return True if "offset" is a boundary position.
//  */
// UBool RuleBasedBreakIterator::isBoundary(int32_t offset) {
//     // out-of-range indexes are never boundary positions
//     if (offset < 0) {
//         first();       // For side effects on current position, tag values.
//         return FALSE;
//     }

//     // Adjust offset to be on a code point boundary and not beyond the end of the text.
//     // Note that isBoundary() is always false for offsets that are not on code point boundaries.
//     // But we still need the side effect of leaving iteration at the following boundary.

//     utext_setNativeIndex(&fText, offset);
//     int32_t adjustedOffset = static_cast<int32_t>(utext_getNativeIndex(&fText));

//     bool result = false;
//     UErrorCode status = U_ZERO_ERROR;
//     if (fBreakCache->seek(adjustedOffset) || fBreakCache->populateNear(adjustedOffset, status)) {
//         result = (fBreakCache->current() == offset);
//     }

//     if (result && adjustedOffset < offset && utext_char32At(&fText, offset) == U_SENTINEL) {
//         // Original offset is beyond the end of the text. Return FALSE, it's not a boundary,
//         // but the iteration position remains set to the end of the text, which is a boundary.
//         return FALSE;
//     }
//     if (!result) {
//         // Not on a boundary. isBoundary() must leave iterator on the following boundary.
//         // Cache->seek(), above, left us on the preceding boundary, so advance one.
//         next();
//     }
//     return result;
// }


// /**
//  * Returns the current iteration position.
//  * @return The current iteration position.
//  */
// int32_t RuleBasedBreakIterator::current(void) const {
//     return fPosition;
// }


// //=======================================================================
// // implementation
// //=======================================================================

// //
// // RBBIRunMode  -  the state machine runs an extra iteration at the beginning and end
// //                 of user text.  A variable with this enum type keeps track of where we
// //                 are.  The state machine only fetches user input while in the RUN mode.
// //
// enum RBBIRunMode {
//     RBBI_START,     // state machine processing is before first char of input
//     RBBI_RUN,       // state machine processing is in the user text
//     RBBI_END        // state machine processing is after end of user text.
// };


// // Map from look-ahead break states (corresponds to rules) to boundary positions.
// // Allows multiple lookahead break rules to be in flight at the same time.
// //
// // This is a temporary approach for ICU 57. A better fix is to make the look-ahead numbers
// // in the state table be sequential, then we can just index an array. And the
// // table could also tell us in advance how big that array needs to be.
// //
// // Before ICU 57 there was just a single simple variable for a look-ahead match that
// // was in progress. Two rules at once did not work.

// static const int32_t kMaxLookaheads = 8;
// struct LookAheadResults {
//     int32_t    fUsedSlotLimit;
//     int32_t    fPositions[8];
//     int16_t    fKeys[8];

//     LookAheadResults() : fUsedSlotLimit(0), fPositions(), fKeys() {}

//     int32_t getPosition(int16_t key) {
//         for (int32_t i=0; i<fUsedSlotLimit; ++i) {
//             if (fKeys[i] == key) {
//                 return fPositions[i];
//             }
//         }
//         UPRV_UNREACHABLE;
//     }

//     void setPosition(int16_t key, int32_t position) {
//         int32_t i;
//         for (i=0; i<fUsedSlotLimit; ++i) {
//             if (fKeys[i] == key) {
//                 fPositions[i] = position;
//                 return;
//             }
//         }
//         if (i >= kMaxLookaheads) {
//             UPRV_UNREACHABLE;
//         }
//         fKeys[i] = key;
//         fPositions[i] = position;
//         U_ASSERT(fUsedSlotLimit == i);
//         fUsedSlotLimit = i + 1;
//     }
// };


// //-----------------------------------------------------------------------------------
// //
// //  handleNext()
// //     Run the state machine to find a boundary
// //
// //-----------------------------------------------------------------------------------
// int32_t RuleBasedBreakIterator::handleNext() {
//     int32_t             state;
//     uint16_t            category        = 0;
//     RBBIRunMode         mode;

//     RBBIStateTableRow  *row;
//     UChar32             c;
//     LookAheadResults    lookAheadMatches;
//     int32_t             result             = 0;
//     int32_t             initialPosition    = 0;
//     const RBBIStateTable *statetable       = fData->fForwardTable;
//     const char         *tableData          = statetable->fTableData;
//     uint32_t            tableRowLen        = statetable->fRowLen;
//     #ifdef RBBI_DEBUG
//         if (gTrace) {
//             RBBIDebugPuts("Handle Next   pos   char  state category");
//         }
//     #endif

//     // handleNext alway sets the break tag value.
//     // Set the default for it.
//     fRuleStatusIndex = 0;

//     fDictionaryCharCount = 0;

//     // if we're already at the end of the text, return DONE.
//     initialPosition = fPosition;
//     UTEXT_SETNATIVEINDEX(&fText, initialPosition);
//     result          = initialPosition;
//     c               = UTEXT_NEXT32(&fText);
//     if (c==U_SENTINEL) {
//         fDone = TRUE;
//         return UBRK_DONE;
//     }

//     //  Set the initial state for the state machine
//     state = START_STATE;
//     row = (RBBIStateTableRow *)
//             //(statetable->fTableData + (statetable->fRowLen * state));
//             (tableData + tableRowLen * state);


//     mode     = RBBI_RUN;
//     if (statetable->fFlags & RBBI_BOF_REQUIRED) {
//         category = 2;
//         mode     = RBBI_START;
//     }


//     // loop until we reach the end of the text or transition to state 0
//     //
//     for (;;) {
//         if (c == U_SENTINEL) {
//             // Reached end of input string.
//             if (mode == RBBI_END) {
//                 // We have already run the loop one last time with the
//                 //   character set to the psueudo {eof} value.  Now it is time
//                 //   to unconditionally bail out.
//                 break;
//             }
//             // Run the loop one last time with the fake end-of-input character category.
//             mode = RBBI_END;
//             category = 1;
//         }

//         //
//         // Get the char category.  An incoming category of 1 or 2 means that
//         //      we are preset for doing the beginning or end of input, and
//         //      that we shouldn't get a category from an actual text input character.
//         //
//         if (mode == RBBI_RUN) {
//             // look up the current character's character category, which tells us
//             // which column in the state table to look at.
//             // Note:  the 16 in UTRIE_GET16 refers to the size of the data being returned,
//             //        not the size of the character going in, which is a UChar32.
//             //
//             category = UTRIE2_GET16(fData->fTrie, c);

//             // Check the dictionary bit in the character's category.
//             //    Counter is only used by dictionary based iteration.
//             //    Chars that need to be handled by a dictionary have a flag bit set
//             //    in their category values.
//             //
//             if ((category & 0x4000) != 0)  {
//                 fDictionaryCharCount++;
//                 //  And off the dictionary flag bit.
//                 category &= ~0x4000;
//             }
//         }

//        #ifdef RBBI_DEBUG
//             if (gTrace) {
//                 RBBIDebugPrintf("             %4" PRId64 "   ", utext_getNativeIndex(&fText));
//                 if (0x20<=c && c<0x7f) {
//                     RBBIDebugPrintf("\"%c\"  ", c);
//                 } else {
//                     RBBIDebugPrintf("%5x  ", c);
//                 }
//                 RBBIDebugPrintf("%3d  %3d\n", state, category);
//             }
//         #endif

//         // State Transition - move machine to its next state
//         //

//         // fNextState is a variable-length array.
//         U_ASSERT(category<fData->fHeader->fCatCount);
//         state = row->fNextState[category];  /*Not accessing beyond memory*/
//         row = (RBBIStateTableRow *)
//             // (statetable->fTableData + (statetable->fRowLen * state));
//             (tableData + tableRowLen * state);


//         if (row->fAccepting == -1) {
//             // Match found, common case.
//             if (mode != RBBI_START) {
//                 result = (int32_t)UTEXT_GETNATIVEINDEX(&fText);
//             }
//             fRuleStatusIndex = row->fTagIdx;   // Remember the break status (tag) values.
//         }

//         int16_t completedRule = row->fAccepting;
//         if (completedRule > 0) {
//             // Lookahead match is completed.
//             int32_t lookaheadResult = lookAheadMatches.getPosition(completedRule);
//             if (lookaheadResult >= 0) {
//                 fRuleStatusIndex = row->fTagIdx;
//                 fPosition = lookaheadResult;
//                 return lookaheadResult;
//             }
//         }
//         int16_t rule = row->fLookAhead;
//         if (rule != 0) {
//             // At the position of a '/' in a look-ahead match. Record it.
//             int32_t  pos = (int32_t)UTEXT_GETNATIVEINDEX(&fText);
//             lookAheadMatches.setPosition(rule, pos);
//         }

//         if (state == STOP_STATE) {
//             // This is the normal exit from the lookup state machine.
//             // We have advanced through the string until it is certain that no
//             //   longer match is possible, no matter what characters follow.
//             break;
//         }

//         // Advance to the next character.
//         // If this is a beginning-of-input loop iteration, don't advance
//         //    the input position.  The next iteration will be processing the
//         //    first real input character.
//         if (mode == RBBI_RUN) {
//             c = UTEXT_NEXT32(&fText);
//         } else {
//             if (mode == RBBI_START) {
//                 mode = RBBI_RUN;
//             }
//         }
//     }

//     // The state machine is done.  Check whether it found a match...

//     // If the iterator failed to advance in the match engine, force it ahead by one.
//     //   (This really indicates a defect in the break rules.  They should always match
//     //    at least one character.)
//     if (result == initialPosition) {
//         utext_setNativeIndex(&fText, initialPosition);
//         utext_next32(&fText);
//         result = (int32_t)utext_getNativeIndex(&fText);
//         fRuleStatusIndex = 0;
//     }

//     // Leave the iterator at our result position.
//     fPosition = result;
//     #ifdef RBBI_DEBUG
//         if (gTrace) {
//             RBBIDebugPrintf("result = %d\n\n", result);
//         }
//     #endif
//     return result;
// }


// //-----------------------------------------------------------------------------------
// //
// //  handleSafePrevious()
// //
// //      Iterate backwards using the safe reverse rules.
// //      The logic of this function is similar to handleNext(), but simpler
// //      because the safe table does not require as many options.
// //
// //-----------------------------------------------------------------------------------
// int32_t RuleBasedBreakIterator::handleSafePrevious(int32_t fromPosition) {
//     int32_t             state;
//     uint16_t            category        = 0;
//     RBBIStateTableRow  *row;
//     UChar32             c;
//     int32_t             result          = 0;

//     const RBBIStateTable *stateTable = fData->fReverseTable;
//     UTEXT_SETNATIVEINDEX(&fText, fromPosition);
//     #ifdef RBBI_DEBUG
//         if (gTrace) {
//             RBBIDebugPuts("Handle Previous   pos   char  state category");
//         }
//     #endif

//     // if we're already at the start of the text, return DONE.
//     if (fData == NULL || UTEXT_GETNATIVEINDEX(&fText)==0) {
//         return BreakIterator::DONE;
//     }

//     //  Set the initial state for the state machine
//     c = UTEXT_PREVIOUS32(&fText);
//     state = START_STATE;
//     row = (RBBIStateTableRow *)
//             (stateTable->fTableData + (stateTable->fRowLen * state));

//     // loop until we reach the start of the text or transition to state 0
//     //
//     for (; c != U_SENTINEL; c = UTEXT_PREVIOUS32(&fText)) {

//         // look up the current character's character category, which tells us
//         // which column in the state table to look at.
//         // Note:  the 16 in UTRIE_GET16 refers to the size of the data being returned,
//         //        not the size of the character going in, which is a UChar32.
//         //
//         //  And off the dictionary flag bit. For reverse iteration it is not used.
//         category = UTRIE2_GET16(fData->fTrie, c);
//         category &= ~0x4000;

//         #ifdef RBBI_DEBUG
//             if (gTrace) {
//                 RBBIDebugPrintf("             %4d   ", (int32_t)utext_getNativeIndex(&fText));
//                 if (0x20<=c && c<0x7f) {
//                     RBBIDebugPrintf("\"%c\"  ", c);
//                 } else {
//                     RBBIDebugPrintf("%5x  ", c);
//                 }
//                 RBBIDebugPrintf("%3d  %3d\n", state, category);
//             }
//         #endif

//         // State Transition - move machine to its next state
//         //
//         // fNextState is a variable-length array.
//         U_ASSERT(category<fData->fHeader->fCatCount);
//         state = row->fNextState[category];  /*Not accessing beyond memory*/
//         row = (RBBIStateTableRow *)
//             (stateTable->fTableData + (stateTable->fRowLen * state));

//         if (state == STOP_STATE) {
//             // This is the normal exit from the lookup state machine.
//             // Transistion to state zero means we have found a safe point.
//             break;
//         }
//     }

//     // The state machine is done.  Check whether it found a match...
//     result = (int32_t)UTEXT_GETNATIVEINDEX(&fText);
//     #ifdef RBBI_DEBUG
//         if (gTrace) {
//             RBBIDebugPrintf("result = %d\n\n", result);
//         }
//     #endif
//     return result;
// }

// //-------------------------------------------------------------------------------
// //
// //   getRuleStatus()   Return the break rule tag associated with the current
// //                     iterator position.  If the iterator arrived at its current
// //                     position by iterating forwards, the value will have been
// //                     cached by the handleNext() function.
// //
// //-------------------------------------------------------------------------------

// int32_t  RuleBasedBreakIterator::getRuleStatus() const {

//     // fLastRuleStatusIndex indexes to the start of the appropriate status record
//     //                                                 (the number of status values.)
//     //   This function returns the last (largest) of the array of status values.
//     int32_t  idx = fRuleStatusIndex + fData->fRuleStatusTable[fRuleStatusIndex];
//     int32_t  tagVal = fData->fRuleStatusTable[idx];

//     return tagVal;
// }


// int32_t RuleBasedBreakIterator::getRuleStatusVec(
//              int32_t *fillInVec, int32_t capacity, UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return 0;
//     }

//     int32_t  numVals = fData->fRuleStatusTable[fRuleStatusIndex];
//     int32_t  numValsToCopy = numVals;
//     if (numVals > capacity) {
//         status = U_BUFFER_OVERFLOW_ERROR;
//         numValsToCopy = capacity;
//     }
//     int i;
//     for (i=0; i<numValsToCopy; i++) {
//         fillInVec[i] = fData->fRuleStatusTable[fRuleStatusIndex + i + 1];
//     }
//     return numVals;
// }



// //-------------------------------------------------------------------------------
// //
// //   getBinaryRules        Access to the compiled form of the rules,
// //                         for use by build system tools that save the data
// //                         for standard iterator types.
// //
// //-------------------------------------------------------------------------------
// const uint8_t  *RuleBasedBreakIterator::getBinaryRules(uint32_t &length) {
//     const uint8_t  *retPtr = NULL;
//     length = 0;

//     if (fData != NULL) {
//         retPtr = (const uint8_t *)fData->fHeader;
//         length = fData->fHeader->fLength;
//     }
//     return retPtr;
// }


// RuleBasedBreakIterator *RuleBasedBreakIterator::createBufferClone(
//         void * /*stackBuffer*/, int32_t &bufferSize, UErrorCode &status) {
//     if (U_FAILURE(status)){
//         return NULL;
//     }

//     if (bufferSize == 0) {
//         bufferSize = 1;  // preflighting for deprecated functionality
//         return NULL;
//     }

//     BreakIterator *clonedBI = clone();
//     if (clonedBI == NULL) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//     } else {
//         status = U_SAFECLONE_ALLOCATED_WARNING;
//     }
//     return (RuleBasedBreakIterator *)clonedBI;
// }

// U_NAMESPACE_END


// static icu::UStack *gLanguageBreakFactories = nullptr;
// static const icu::UnicodeString *gEmptyString = nullptr;
// static icu::UInitOnce gLanguageBreakFactoriesInitOnce = U_INITONCE_INITIALIZER;
// static icu::UInitOnce gRBBIInitOnce = U_INITONCE_INITIALIZER;

// /**
//  * Release all static memory held by breakiterator.
//  */
// U_CDECL_BEGIN
// static UBool U_CALLCONV rbbi_cleanup(void) {
//     delete gLanguageBreakFactories;
//     gLanguageBreakFactories = nullptr;
//     delete gEmptyString;
//     gEmptyString = nullptr;
//     gLanguageBreakFactoriesInitOnce.reset();
//     gRBBIInitOnce.reset();
//     return TRUE;
// }
// U_CDECL_END

// U_CDECL_BEGIN
// static void U_CALLCONV _deleteFactory(void *obj) {
//     delete (icu::LanguageBreakFactory *) obj;
// }
// U_CDECL_END
// U_NAMESPACE_BEGIN

// static void U_CALLCONV rbbiInit() {
//     gEmptyString = new UnicodeString();
//     ucln_common_registerCleanup(UCLN_COMMON_RBBI, rbbi_cleanup);
// }

// static void U_CALLCONV initLanguageFactories() {
//     UErrorCode status = U_ZERO_ERROR;
//     U_ASSERT(gLanguageBreakFactories == NULL);
//     gLanguageBreakFactories = new UStack(_deleteFactory, NULL, status);
//     if (gLanguageBreakFactories != NULL && U_SUCCESS(status)) {
//         ICULanguageBreakFactory *builtIn = new ICULanguageBreakFactory(status);
//         gLanguageBreakFactories->push(builtIn, status);
// #ifdef U_LOCAL_SERVICE_HOOK
//         LanguageBreakFactory *extra = (LanguageBreakFactory *)uprv_svc_hook("languageBreakFactory", &status);
//         if (extra != NULL) {
//             gLanguageBreakFactories->push(extra, status);
//         }
// #endif
//     }
//     ucln_common_registerCleanup(UCLN_COMMON_RBBI, rbbi_cleanup);
// }


// static const LanguageBreakEngine*
// getLanguageBreakEngineFromFactory(UChar32 c)
// {
//     umtx_initOnce(gLanguageBreakFactoriesInitOnce, &initLanguageFactories);
//     if (gLanguageBreakFactories == NULL) {
//         return NULL;
//     }

//     int32_t i = gLanguageBreakFactories->size();
//     const LanguageBreakEngine *lbe = NULL;
//     while (--i >= 0) {
//         LanguageBreakFactory *factory = (LanguageBreakFactory *)(gLanguageBreakFactories->elementAt(i));
//         lbe = factory->getEngineFor(c);
//         if (lbe != NULL) {
//             break;
//         }
//     }
//     return lbe;
// }


// //-------------------------------------------------------------------------------
// //
// //  getLanguageBreakEngine  Find an appropriate LanguageBreakEngine for the
// //                          the character c.
// //
// //-------------------------------------------------------------------------------
// const LanguageBreakEngine *
// RuleBasedBreakIterator::getLanguageBreakEngine(UChar32 c) {
//     const LanguageBreakEngine *lbe = NULL;
//     UErrorCode status = U_ZERO_ERROR;

//     if (fLanguageBreakEngines == NULL) {
//         fLanguageBreakEngines = new UStack(status);
//         if (fLanguageBreakEngines == NULL || U_FAILURE(status)) {
//             delete fLanguageBreakEngines;
//             fLanguageBreakEngines = 0;
//             return NULL;
//         }
//     }

//     int32_t i = fLanguageBreakEngines->size();
//     while (--i >= 0) {
//         lbe = (const LanguageBreakEngine *)(fLanguageBreakEngines->elementAt(i));
//         if (lbe->handles(c)) {
//             return lbe;
//         }
//     }

//     // No existing dictionary took the character. See if a factory wants to
//     // give us a new LanguageBreakEngine for this character.
//     lbe = getLanguageBreakEngineFromFactory(c);

//     // If we got one, use it and push it on our stack.
//     if (lbe != NULL) {
//         fLanguageBreakEngines->push((void *)lbe, status);
//         // Even if we can't remember it, we can keep looking it up, so
//         // return it even if the push fails.
//         return lbe;
//     }

//     // No engine is forthcoming for this character. Add it to the
//     // reject set. Create the reject break engine if needed.
//     if (fUnhandledBreakEngine == NULL) {
//         fUnhandledBreakEngine = new UnhandledEngine(status);
//         if (U_SUCCESS(status) && fUnhandledBreakEngine == NULL) {
//             status = U_MEMORY_ALLOCATION_ERROR;
//             return nullptr;
//         }
//         // Put it last so that scripts for which we have an engine get tried
//         // first.
//         fLanguageBreakEngines->insertElementAt(fUnhandledBreakEngine, 0, status);
//         // If we can't insert it, or creation failed, get rid of it
//         if (U_FAILURE(status)) {
//             delete fUnhandledBreakEngine;
//             fUnhandledBreakEngine = 0;
//             return NULL;
//         }
//     }

//     // Tell the reject engine about the character; at its discretion, it may
//     // add more than just the one character.
//     fUnhandledBreakEngine->handleCharacter(c);

//     return fUnhandledBreakEngine;
// }

// void RuleBasedBreakIterator::dumpCache() {
//     fBreakCache->dumpCache();
// }

// void RuleBasedBreakIterator::dumpTables() {
//     fData->printData();
// }

// /**
//  * Returns the description used to create this iterator
//  */

// const UnicodeString&
// RuleBasedBreakIterator::getRules() const {
//     if (fData != NULL) {
//         return fData->getRuleSourceString();
//     } else {
//         umtx_initOnce(gRBBIInitOnce, &rbbiInit);
//         return *gEmptyString;
//     }
// }

// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */
// Copyright (C) 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// file: rbbi_cache.cpp



// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/ubrk.h"
// #include "unicode/rbbi.h"

// #include "rbbi_cache.h"

// #include "brkeng.h"
// #include "cmemory.h"
// #include "rbbidata.h"
// #include "rbbirb.h"
// #include "uassert.h"
// #include "uvectr32.h"

// U_NAMESPACE_BEGIN

// /*
//  * DictionaryCache implementation
//  */

// RuleBasedBreakIterator::DictionaryCache::DictionaryCache(RuleBasedBreakIterator *bi, UErrorCode &status) :
//         fBI(bi), fBreaks(status), fPositionInCache(-1),
//         fStart(0), fLimit(0), fFirstRuleStatusIndex(0), fOtherRuleStatusIndex(0) {
// }

// RuleBasedBreakIterator::DictionaryCache::~DictionaryCache() {
// }

// void RuleBasedBreakIterator::DictionaryCache::reset() {
//     fPositionInCache = -1;
//     fStart = 0;
//     fLimit = 0;
//     fFirstRuleStatusIndex = 0;
//     fOtherRuleStatusIndex = 0;
//     fBreaks.removeAllElements();
// }

// UBool RuleBasedBreakIterator::DictionaryCache::following(int32_t fromPos, int32_t *result, int32_t *statusIndex) {
//     if (fromPos >= fLimit || fromPos < fStart) {
//         fPositionInCache = -1;
//         return FALSE;
//     }

//     // Sequential iteration, move from previous boundary to the following

//     int32_t r = 0;
//     if (fPositionInCache >= 0 && fPositionInCache < fBreaks.size() && fBreaks.elementAti(fPositionInCache) == fromPos) {
//         ++fPositionInCache;
//         if (fPositionInCache >= fBreaks.size()) {
//             fPositionInCache = -1;
//             return FALSE;
//         }
//         r = fBreaks.elementAti(fPositionInCache);
//         U_ASSERT(r > fromPos);
//         *result = r;
//         *statusIndex = fOtherRuleStatusIndex;
//         return TRUE;
//     }

//     // Random indexing. Linear search for the boundary following the given position.

//     for (fPositionInCache = 0; fPositionInCache < fBreaks.size(); ++fPositionInCache) {
//         r= fBreaks.elementAti(fPositionInCache);
//         if (r > fromPos) {
//             *result = r;
//             *statusIndex = fOtherRuleStatusIndex;
//             return TRUE;
//         }
//     }
//     UPRV_UNREACHABLE;
// }


// UBool RuleBasedBreakIterator::DictionaryCache::preceding(int32_t fromPos, int32_t *result, int32_t *statusIndex) {
//     if (fromPos <= fStart || fromPos > fLimit) {
//         fPositionInCache = -1;
//         return FALSE;
//     }

//     if (fromPos == fLimit) {
//         fPositionInCache = fBreaks.size() - 1;
//         if (fPositionInCache >= 0) {
//             U_ASSERT(fBreaks.elementAti(fPositionInCache) == fromPos);
//         }
//     }

//     int32_t r;
//     if (fPositionInCache > 0 && fPositionInCache < fBreaks.size() && fBreaks.elementAti(fPositionInCache) == fromPos) {
//         --fPositionInCache;
//         r = fBreaks.elementAti(fPositionInCache);
//         U_ASSERT(r < fromPos);
//         *result = r;
//         *statusIndex = ( r== fStart) ? fFirstRuleStatusIndex : fOtherRuleStatusIndex;
//         return TRUE;
//     }

//     if (fPositionInCache == 0) {
//         fPositionInCache = -1;
//         return FALSE;
//     }

//     for (fPositionInCache = fBreaks.size()-1; fPositionInCache >= 0; --fPositionInCache) {
//         r = fBreaks.elementAti(fPositionInCache);
//         if (r < fromPos) {
//             *result = r;
//             *statusIndex = ( r == fStart) ? fFirstRuleStatusIndex : fOtherRuleStatusIndex;
//             return TRUE;
//         }
//     }
//     UPRV_UNREACHABLE;
// }

// void RuleBasedBreakIterator::DictionaryCache::populateDictionary(int32_t startPos, int32_t endPos,
//                                        int32_t firstRuleStatus, int32_t otherRuleStatus) {
//     if ((endPos - startPos) <= 1) {
//         return;
//     }

//     reset();
//     fFirstRuleStatusIndex = firstRuleStatus;
//     fOtherRuleStatusIndex = otherRuleStatus;

//     int32_t rangeStart = startPos;
//     int32_t rangeEnd = endPos;

//     uint16_t    category;
//     int32_t     current;
//     UErrorCode  status = U_ZERO_ERROR;
//     int32_t     foundBreakCount = 0;
//     UText      *text = &fBI->fText;

//     // Loop through the text, looking for ranges of dictionary characters.
//     // For each span, find the appropriate break engine, and ask it to find
//     // any breaks within the span.

//     utext_setNativeIndex(text, rangeStart);
//     UChar32     c = utext_current32(text);
//     category = UTRIE2_GET16(fBI->fData->fTrie, c);

//     while(U_SUCCESS(status)) {
//         while((current = (int32_t)UTEXT_GETNATIVEINDEX(text)) < rangeEnd && (category & 0x4000) == 0) {
//             utext_next32(text);           // TODO: cleaner loop structure.
//             c = utext_current32(text);
//             category = UTRIE2_GET16(fBI->fData->fTrie, c);
//         }
//         if (current >= rangeEnd) {
//             break;
//         }

//         // We now have a dictionary character. Get the appropriate language object
//         // to deal with it.
//         const LanguageBreakEngine *lbe = fBI->getLanguageBreakEngine(c);

//         // Ask the language object if there are any breaks. It will add them to the cache and
//         // leave the text pointer on the other side of its range, ready to search for the next one.
//         if (lbe != NULL) {
//             foundBreakCount += lbe->findBreaks(text, rangeStart, rangeEnd, fBreaks);
//         }

//         // Reload the loop variables for the next go-round
//         c = utext_current32(text);
//         category = UTRIE2_GET16(fBI->fData->fTrie, c);
//     }

//     // If we found breaks, ensure that the first and last entries are
//     // the original starting and ending position. And initialize the
//     // cache iteration position to the first entry.

//     // printf("foundBreakCount = %d\n", foundBreakCount);
//     if (foundBreakCount > 0) {
//         U_ASSERT(foundBreakCount == fBreaks.size());
//         if (startPos < fBreaks.elementAti(0)) {
//             // The dictionary did not place a boundary at the start of the segment of text.
//             // Add one now. This should not commonly happen, but it would be easy for interactions
//             // of the rules for dictionary segments and the break engine implementations to
//             // inadvertently cause it. Cover it here, just in case.
//             fBreaks.insertElementAt(startPos, 0, status);
//         }
//         if (endPos > fBreaks.peeki()) {
//             fBreaks.push(endPos, status);
//         }
//         fPositionInCache = 0;
//         // Note: Dictionary matching may extend beyond the original limit.
//         fStart = fBreaks.elementAti(0);
//         fLimit = fBreaks.peeki();
//     } else {
//         // there were no language-based breaks, even though the segment contained
//         // dictionary characters. Subsequent attempts to fetch boundaries from the dictionary cache
//         // for this range will fail, and the calling code will fall back to the rule based boundaries.
//     }
// }


// /*
//  *   BreakCache implemetation
//  */

// RuleBasedBreakIterator::BreakCache::BreakCache(RuleBasedBreakIterator *bi, UErrorCode &status) :
//         fBI(bi), fSideBuffer(status) {
//     reset();
// }


// RuleBasedBreakIterator::BreakCache::~BreakCache() {
// }


// void RuleBasedBreakIterator::BreakCache::reset(int32_t pos, int32_t ruleStatus) {
//     fStartBufIdx = 0;
//     fEndBufIdx = 0;
//     fTextIdx = pos;
//     fBufIdx = 0;
//     fBoundaries[0] = pos;
//     fStatuses[0] = (uint16_t)ruleStatus;
// }


// int32_t  RuleBasedBreakIterator::BreakCache::current() {
//     fBI->fPosition = fTextIdx;
//     fBI->fRuleStatusIndex = fStatuses[fBufIdx];
//     fBI->fDone = FALSE;
//     return fTextIdx;
// }


// void RuleBasedBreakIterator::BreakCache::following(int32_t startPos, UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return;
//     }
//     if (startPos == fTextIdx || seek(startPos) || populateNear(startPos, status)) {
//         // startPos is in the cache. Do a next() from that position.
//         // TODO: an awkward set of interactions with bi->fDone
//         //       seek() does not clear it; it can't because of interactions with populateNear().
//         //       next() does not clear it in the fast-path case, where everything matters. Maybe it should.
//         //       So clear it here, for the case where seek() succeeded on an iterator that had previously run off the end.
//         fBI->fDone = false;
//         next();
//     }
//     return;
// }


// void RuleBasedBreakIterator::BreakCache::preceding(int32_t startPos, UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return;
//     }
//     if (startPos == fTextIdx || seek(startPos) || populateNear(startPos, status)) {
//         if (startPos == fTextIdx) {
//             previous(status);
//         } else {
//             // seek() leaves the BreakCache positioned at the preceding boundary
//             //        if the requested position is between two bounaries.
//             // current() pushes the BreakCache position out to the BreakIterator itself.
//             U_ASSERT(startPos > fTextIdx);
//             current();
//         }
//     }
//     return;
// }


// /*
//  * Out-of-line code for BreakCache::next().
//  * Cache does not already contain the boundary
//  */
// void RuleBasedBreakIterator::BreakCache::nextOL() {
//     fBI->fDone = !populateFollowing();
//     fBI->fPosition = fTextIdx;
//     fBI->fRuleStatusIndex = fStatuses[fBufIdx];
//     return;
// }


// void RuleBasedBreakIterator::BreakCache::previous(UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return;
//     }
//     int32_t initialBufIdx = fBufIdx;
//     if (fBufIdx == fStartBufIdx) {
//         // At start of cache. Prepend to it.
//         populatePreceding(status);
//     } else {
//         // Cache already holds the next boundary
//         fBufIdx = modChunkSize(fBufIdx - 1);
//         fTextIdx = fBoundaries[fBufIdx];
//     }
//     fBI->fDone = (fBufIdx == initialBufIdx);
//     fBI->fPosition = fTextIdx;
//     fBI->fRuleStatusIndex = fStatuses[fBufIdx];
//     return;
// }


// UBool RuleBasedBreakIterator::BreakCache::seek(int32_t pos) {
//     if (pos < fBoundaries[fStartBufIdx] || pos > fBoundaries[fEndBufIdx]) {
//         return FALSE;
//     }
//     if (pos == fBoundaries[fStartBufIdx]) {
//         // Common case: seek(0), from BreakIterator::first()
//         fBufIdx = fStartBufIdx;
//         fTextIdx = fBoundaries[fBufIdx];
//         return TRUE;
//     }
//     if (pos == fBoundaries[fEndBufIdx]) {
//         fBufIdx = fEndBufIdx;
//         fTextIdx = fBoundaries[fBufIdx];
//         return TRUE;
//     }

//     int32_t min = fStartBufIdx;
//     int32_t max = fEndBufIdx;
//     while (min != max) {
//         int32_t probe = (min + max + (min>max ? CACHE_SIZE : 0)) / 2;
//         probe = modChunkSize(probe);
//         if (fBoundaries[probe] > pos) {
//             max = probe;
//         } else {
//             min = modChunkSize(probe + 1);
//         }
//     }
//     U_ASSERT(fBoundaries[max] > pos);
//     fBufIdx = modChunkSize(max - 1);
//     fTextIdx = fBoundaries[fBufIdx];
//     U_ASSERT(fTextIdx <= pos);
//     return TRUE;
// }


// UBool RuleBasedBreakIterator::BreakCache::populateNear(int32_t position, UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return FALSE;
//     }
//     U_ASSERT(position < fBoundaries[fStartBufIdx] || position > fBoundaries[fEndBufIdx]);

//     // Find a boundary somewhere in the vicinity of the requested position.
//     // Depending on the safe rules and the text data, it could be either before, at, or after
//     // the requested position.


//     // If the requested position is not near already cached positions, clear the existing cache,
//     // find a near-by boundary and begin new cache contents there.

//     if ((position < fBoundaries[fStartBufIdx] - 15) || position > (fBoundaries[fEndBufIdx] + 15)) {
//         int32_t aBoundary = 0;
//         int32_t ruleStatusIndex = 0;
//         if (position > 20) {
//             int32_t backupPos = fBI->handleSafePrevious(position);

//             if (backupPos > 0) {
//                 // Advance to the boundary following the backup position.
//                 // There is a complication: the safe reverse rules identify pairs of code points
//                 // that are safe. If advancing from the safe point moves forwards by less than
//                 // two code points, we need to advance one more time to ensure that the boundary
//                 // is good, including a correct rules status value.
//                 //
//                 fBI->fPosition = backupPos;
//                 aBoundary = fBI->handleNext();
//                 if (aBoundary <= backupPos + 4) {
//                     // +4 is a quick test for possibly having advanced only one codepoint.
//                     // Four being the length of the longest potential code point, a supplementary in UTF-8
//                     utext_setNativeIndex(&fBI->fText, aBoundary);
//                     if (backupPos == utext_getPreviousNativeIndex(&fBI->fText)) {
//                         // The initial handleNext() only advanced by a single code point. Go again.
//                         aBoundary = fBI->handleNext();   // Safe rules identify safe pairs.
//                     }
//                 }
//                 ruleStatusIndex = fBI->fRuleStatusIndex;
//             }
//         }
//         reset(aBoundary, ruleStatusIndex);        // Reset cache to hold aBoundary as a single starting point.
//     }

//     // Fill in boundaries between existing cache content and the new requested position.

//     if (fBoundaries[fEndBufIdx] < position) {
//         // The last position in the cache precedes the requested position.
//         // Add following position(s) to the cache.
//         while (fBoundaries[fEndBufIdx] < position) {
//             if (!populateFollowing()) {
//                 UPRV_UNREACHABLE;
//             }
//         }
//         fBufIdx = fEndBufIdx;                      // Set iterator position to the end of the buffer.
//         fTextIdx = fBoundaries[fBufIdx];           // Required because populateFollowing may add extra boundaries.
//         while (fTextIdx > position) {              // Move backwards to a position at or preceding the requested pos.
//             previous(status);
//         }
//         return true;
//     }

//     if (fBoundaries[fStartBufIdx] > position) {
//         // The first position in the cache is beyond the requested position.
//         // back up more until we get a boundary <= the requested position.
//         while (fBoundaries[fStartBufIdx] > position) {
//             populatePreceding(status);
//         }
//         fBufIdx = fStartBufIdx;                    // Set iterator position to the start of the buffer.
//         fTextIdx = fBoundaries[fBufIdx];           // Required because populatePreceding may add extra boundaries.
//         while (fTextIdx < position) {              // Move forwards to a position at or following the requested pos.
//             next();
//         }
//         if (fTextIdx > position) {
//             // If position is not itself a boundary, the next() loop above will overshoot.
//             // Back up one, leaving cache position at the boundary preceding the requested position.
//             previous(status);
//         }
//         return true;
//     }

//     U_ASSERT(fTextIdx == position);
//     return true;
// }



// UBool RuleBasedBreakIterator::BreakCache::populateFollowing() {
//     int32_t fromPosition = fBoundaries[fEndBufIdx];
//     int32_t fromRuleStatusIdx = fStatuses[fEndBufIdx];
//     int32_t pos = 0;
//     int32_t ruleStatusIdx = 0;

//     if (fBI->fDictionaryCache->following(fromPosition, &pos, &ruleStatusIdx)) {
//         addFollowing(pos, ruleStatusIdx, UpdateCachePosition);
//         return TRUE;
//     }

//     fBI->fPosition = fromPosition;
//     pos = fBI->handleNext();
//     if (pos == UBRK_DONE) {
//         return FALSE;
//     }

//     ruleStatusIdx = fBI->fRuleStatusIndex;
//     if (fBI->fDictionaryCharCount > 0) {
//         // The text segment obtained from the rules includes dictionary characters.
//         // Subdivide it, with subdivided results going into the dictionary cache.
//         fBI->fDictionaryCache->populateDictionary(fromPosition, pos, fromRuleStatusIdx, ruleStatusIdx);
//         if (fBI->fDictionaryCache->following(fromPosition, &pos, &ruleStatusIdx)) {
//             addFollowing(pos, ruleStatusIdx, UpdateCachePosition);
//             return TRUE;
//             // TODO: may want to move a sizable chunk of dictionary cache to break cache at this point.
//             //       But be careful with interactions with populateNear().
//         }
//     }

//     // Rule based segment did not include dictionary characters.
//     // Or, it did contain dictionary chars, but the dictionary segmenter didn't handle them,
//     //    meaning that we didn't take the return, above.
//     // Add its end point to the cache.
//     addFollowing(pos, ruleStatusIdx, UpdateCachePosition);

//     // Add several non-dictionary boundaries at this point, to optimize straight forward iteration.
//     //    (subsequent calls to BreakIterator::next() will take the fast path, getting cached results.
//     //
//     for (int count=0; count<6; ++count) {
//         pos = fBI->handleNext();
//         if (pos == UBRK_DONE || fBI->fDictionaryCharCount > 0) {
//             break;
//         }
//         addFollowing(pos, fBI->fRuleStatusIndex, RetainCachePosition);
//     }

//     return TRUE;
// }


// UBool RuleBasedBreakIterator::BreakCache::populatePreceding(UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return FALSE;
//     }

//     int32_t fromPosition = fBoundaries[fStartBufIdx];
//     if (fromPosition == 0) {
//         return FALSE;
//     }

//     int32_t position = 0;
//     int32_t positionStatusIdx = 0;

//     if (fBI->fDictionaryCache->preceding(fromPosition, &position, &positionStatusIdx)) {
//         addPreceding(position, positionStatusIdx, UpdateCachePosition);
//         return TRUE;
//     }

//     int32_t backupPosition = fromPosition;

//     // Find a boundary somewhere preceding the first already-cached boundary
//     do {
//         backupPosition = backupPosition - 30;
//         if (backupPosition <= 0) {
//             backupPosition = 0;
//         } else {
//             backupPosition = fBI->handleSafePrevious(backupPosition);
//         }
//         if (backupPosition == UBRK_DONE || backupPosition == 0) {
//             position = 0;
//             positionStatusIdx = 0;
//         } else {
//             // Advance to the boundary following the backup position.
//             // There is a complication: the safe reverse rules identify pairs of code points
//             // that are safe. If advancing from the safe point moves forwards by less than
//             // two code points, we need to advance one more time to ensure that the boundary
//             // is good, including a correct rules status value.
//             //
//             fBI->fPosition = backupPosition;
//             position = fBI->handleNext();
//             if (position <= backupPosition + 4) {
//                 // +4 is a quick test for possibly having advanced only one codepoint.
//                 // Four being the length of the longest potential code point, a supplementary in UTF-8
//                 utext_setNativeIndex(&fBI->fText, position);
//                 if (backupPosition == utext_getPreviousNativeIndex(&fBI->fText)) {
//                     // The initial handleNext() only advanced by a single code point. Go again.
//                     position = fBI->handleNext();   // Safe rules identify safe pairs.
//                 }
//             }
//             positionStatusIdx = fBI->fRuleStatusIndex;
//         }
//     } while (position >= fromPosition);

//     // Find boundaries between the one we just located and the first already-cached boundary
//     // Put them in a side buffer, because we don't yet know where they will fall in the circular cache buffer..

//     fSideBuffer.removeAllElements();
//     fSideBuffer.addElement(position, status);
//     fSideBuffer.addElement(positionStatusIdx, status);

//     do {
//         int32_t prevPosition = fBI->fPosition = position;
//         int32_t prevStatusIdx = positionStatusIdx;
//         position = fBI->handleNext();
//         positionStatusIdx = fBI->fRuleStatusIndex;
//         if (position == UBRK_DONE) {
//             break;
//         }

//         UBool segmentHandledByDictionary = FALSE;
//         if (fBI->fDictionaryCharCount != 0) {
//             // Segment from the rules includes dictionary characters.
//             // Subdivide it, with subdivided results going into the dictionary cache.
//             int32_t dictSegEndPosition = position;
//             fBI->fDictionaryCache->populateDictionary(prevPosition, dictSegEndPosition, prevStatusIdx, positionStatusIdx);
//             while (fBI->fDictionaryCache->following(prevPosition, &position, &positionStatusIdx)) {
//                 segmentHandledByDictionary = true;
//                 U_ASSERT(position > prevPosition);
//                 if (position >= fromPosition) {
//                     break;
//                 }
//                 U_ASSERT(position <= dictSegEndPosition);
//                 fSideBuffer.addElement(position, status);
//                 fSideBuffer.addElement(positionStatusIdx, status);
//                 prevPosition = position;
//             }
//             U_ASSERT(position==dictSegEndPosition || position>=fromPosition);
//         }

//         if (!segmentHandledByDictionary && position < fromPosition) {
//             fSideBuffer.addElement(position, status);
//             fSideBuffer.addElement(positionStatusIdx, status);
//         }
//     } while (position < fromPosition);

//     // Move boundaries from the side buffer to the main circular buffer.
//     UBool success = FALSE;
//     if (!fSideBuffer.isEmpty()) {
//         positionStatusIdx = fSideBuffer.popi();
//         position = fSideBuffer.popi();
//         addPreceding(position, positionStatusIdx, UpdateCachePosition);
//         success = TRUE;
//     }

//     while (!fSideBuffer.isEmpty()) {
//         positionStatusIdx = fSideBuffer.popi();
//         position = fSideBuffer.popi();
//         if (!addPreceding(position, positionStatusIdx, RetainCachePosition)) {
//             // No space in circular buffer to hold a new preceding result while
//             // also retaining the current cache (iteration) position.
//             // Bailing out is safe; the cache will refill again if needed.
//             break;
//         }
//     }

//     return success;
// }


// void RuleBasedBreakIterator::BreakCache::addFollowing(int32_t position, int32_t ruleStatusIdx, UpdatePositionValues update) {
//     U_ASSERT(position > fBoundaries[fEndBufIdx]);
//     U_ASSERT(ruleStatusIdx <= UINT16_MAX);
//     int32_t nextIdx = modChunkSize(fEndBufIdx + 1);
//     if (nextIdx == fStartBufIdx) {
//         fStartBufIdx = modChunkSize(fStartBufIdx + 6);    // TODO: experiment. Probably revert to 1.
//     }
//     fBoundaries[nextIdx] = position;
//     fStatuses[nextIdx] = static_cast<uint16_t>(ruleStatusIdx);
//     fEndBufIdx = nextIdx;
//     if (update == UpdateCachePosition) {
//         // Set current position to the newly added boundary.
//         fBufIdx = nextIdx;
//         fTextIdx = position;
//     } else {
//         // Retaining the original cache position.
//         // Check if the added boundary wraps around the buffer, and would over-write the original position.
//         // It's the responsibility of callers of this function to not add too many.
//         U_ASSERT(nextIdx != fBufIdx);
//     }
// }

// bool RuleBasedBreakIterator::BreakCache::addPreceding(int32_t position, int32_t ruleStatusIdx, UpdatePositionValues update) {
//     U_ASSERT(position < fBoundaries[fStartBufIdx]);
//     U_ASSERT(ruleStatusIdx <= UINT16_MAX);
//     int32_t nextIdx = modChunkSize(fStartBufIdx - 1);
//     if (nextIdx == fEndBufIdx) {
//         if (fBufIdx == fEndBufIdx && update == RetainCachePosition) {
//             // Failure. The insertion of the new boundary would claim the buffer position that is the
//             // current iteration position. And we also want to retain the current iteration position.
//             // (The buffer is already completely full of entries that precede the iteration position.)
//             return false;
//         }
//         fEndBufIdx = modChunkSize(fEndBufIdx - 1);
//     }
//     fBoundaries[nextIdx] = position;
//     fStatuses[nextIdx] = static_cast<uint16_t>(ruleStatusIdx);
//     fStartBufIdx = nextIdx;
//     if (update == UpdateCachePosition) {
//         fBufIdx = nextIdx;
//         fTextIdx = position;
//     }
//     return true;
// }


// void RuleBasedBreakIterator::BreakCache::dumpCache() {
// #ifdef RBBI_DEBUG
//     RBBIDebugPrintf("fTextIdx:%d   fBufIdx:%d\n", fTextIdx, fBufIdx);
//     for (int32_t i=fStartBufIdx; ; i=modChunkSize(i+1)) {
//         RBBIDebugPrintf("%d  %d\n", i, fBoundaries[i]);
//         if (i == fEndBufIdx) {
//             break;
//         }
//     }
// #endif
// }

// U_NAMESPACE_END

// #endif // #if !UCONFIG_NO_BREAK_ITERATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
***************************************************************************
*   Copyright (C) 1999-2014 International Business Machines Corporation   *
*   and others. All rights reserved.                                      *
***************************************************************************
*/



// #if !UCONFIG_NO_BREAK_ITERATION

// 
// #include "rbbidata.h"
// #include "rbbirb.h"
// #include "utrie2.h"
// #include "udatamem.h"
// #include "cmemory.h"
// #include "cstring.h"
// #include "umutex.h"

// #include "uassert.h"


// U_NAMESPACE_BEGIN

// //-----------------------------------------------------------------------------
// //
// //    Constructors.
// //
// //-----------------------------------------------------------------------------
// RBBIDataWrapper::RBBIDataWrapper(const RBBIDataHeader *data, UErrorCode &status) {
//     init0();
//     init(data, status);
// }

// RBBIDataWrapper::RBBIDataWrapper(const RBBIDataHeader *data, enum EDontAdopt, UErrorCode &status) {
//     init0();
//     init(data, status);
//     fDontFreeData = TRUE;
// }

// RBBIDataWrapper::RBBIDataWrapper(UDataMemory* udm, UErrorCode &status) {
//     init0();
//     if (U_FAILURE(status)) {
//         return;
//     }
//     const DataHeader *dh = udm->pHeader;
//     int32_t headerSize = dh->dataHeader.headerSize;
//     if (  !(headerSize >= 20 &&
//             dh->info.isBigEndian == U_IS_BIG_ENDIAN &&
//             dh->info.charsetFamily == U_CHARSET_FAMILY &&
//             dh->info.dataFormat[0] == 0x42 &&  // dataFormat="Brk "
//             dh->info.dataFormat[1] == 0x72 &&
//             dh->info.dataFormat[2] == 0x6b &&
//             dh->info.dataFormat[3] == 0x20 &&
//             isDataVersionAcceptable(dh->info.formatVersion))
//         ) {
//         status = U_INVALID_FORMAT_ERROR;
//         return;
//     }
//     const char *dataAsBytes = reinterpret_cast<const char *>(dh);
//     const RBBIDataHeader *rbbidh = reinterpret_cast<const RBBIDataHeader *>(dataAsBytes + headerSize);
//     init(rbbidh, status);
//     fUDataMem = udm;
// }

// UBool RBBIDataWrapper::isDataVersionAcceptable(const UVersionInfo version) {
//     return RBBI_DATA_FORMAT_VERSION[0] == version[0];
// }


// //-----------------------------------------------------------------------------
// //
// //    init().   Does most of the work of construction, shared between the
// //              constructors.
// //
// //-----------------------------------------------------------------------------
// void RBBIDataWrapper::init0() {
//     fHeader = NULL;
//     fForwardTable = NULL;
//     fReverseTable = NULL;
//     fRuleSource   = NULL;
//     fRuleStatusTable = NULL;
//     fTrie         = NULL;
//     fUDataMem     = NULL;
//     fRefCount     = 0;
//     fDontFreeData = TRUE;
// }

// void RBBIDataWrapper::init(const RBBIDataHeader *data, UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return;
//     }
//     fHeader = data;
//     if (fHeader->fMagic != 0xb1a0 || !isDataVersionAcceptable(fHeader->fFormatVersion)) {
//         status = U_INVALID_FORMAT_ERROR;
//         return;
//     }
//     // Note: in ICU version 3.2 and earlier, there was a formatVersion 1
//     //       that is no longer supported.  At that time fFormatVersion was
//     //       an int32_t field, rather than an array of 4 bytes.

//     fDontFreeData = FALSE;
//     if (data->fFTableLen != 0) {
//         fForwardTable = (RBBIStateTable *)((char *)data + fHeader->fFTable);
//     }
//     if (data->fRTableLen != 0) {
//         fReverseTable = (RBBIStateTable *)((char *)data + fHeader->fRTable);
//     }

//     fTrie = utrie2_openFromSerialized(UTRIE2_16_VALUE_BITS,
//                                       (uint8_t *)data + fHeader->fTrie,
//                                       fHeader->fTrieLen,
//                                       NULL,           // *actual length
//                                       &status);
//     if (U_FAILURE(status)) {
//         return;
//     }

//     fRuleSource   = (UChar *)((char *)data + fHeader->fRuleSource);
//     fRuleString.setTo(TRUE, fRuleSource, -1);
//     U_ASSERT(data->fRuleSourceLen > 0);

//     fRuleStatusTable = (int32_t *)((char *)data + fHeader->fStatusTable);
//     fStatusMaxIdx    = data->fStatusTableLen / sizeof(int32_t);

//     fRefCount = 1;

// #ifdef RBBI_DEBUG
//     char *debugEnv = getenv("U_RBBIDEBUG");
//     if (debugEnv && uprv_strstr(debugEnv, "data")) {this->printData();}
// #endif
// }


// //-----------------------------------------------------------------------------
// //
// //    Destructor.     Don't call this - use removeReference() instead.
// //
// //-----------------------------------------------------------------------------
// RBBIDataWrapper::~RBBIDataWrapper() {
//     U_ASSERT(fRefCount == 0);
//     utrie2_close(fTrie);
//     fTrie = NULL;
//     if (fUDataMem) {
//         udata_close(fUDataMem);
//     } else if (!fDontFreeData) {
//         uprv_free((void *)fHeader);
//     }
// }



// //-----------------------------------------------------------------------------
// //
// //   Operator ==    Consider two RBBIDataWrappers to be equal if they
// //                  refer to the same underlying data.  Although
// //                  the data wrappers are normally shared between
// //                  iterator instances, it's possible to independently
// //                  open the same data twice, and get two instances, which
// //                  should still be ==.
// //
// //-----------------------------------------------------------------------------
// UBool RBBIDataWrapper::operator ==(const RBBIDataWrapper &other) const {
//     if (fHeader == other.fHeader) {
//         return TRUE;
//     }
//     if (fHeader->fLength != other.fHeader->fLength) {
//         return FALSE;
//     }
//     if (uprv_memcmp(fHeader, other.fHeader, fHeader->fLength) == 0) {
//         return TRUE;
//     }
//     return FALSE;
// }

// int32_t  RBBIDataWrapper::hashCode() {
//     return fHeader->fFTableLen;
// }



// //-----------------------------------------------------------------------------
// //
// //    Reference Counting.   A single RBBIDataWrapper object is shared among
// //                          however many RulesBasedBreakIterator instances are
// //                          referencing the same data.
// //
// //-----------------------------------------------------------------------------
// void RBBIDataWrapper::removeReference() {
//     if (umtx_atomic_dec(&fRefCount) == 0) {
//         delete this;
//     }
// }


// RBBIDataWrapper *RBBIDataWrapper::addReference() {
//    umtx_atomic_inc(&fRefCount);
//    return this;
// }



// //-----------------------------------------------------------------------------
// //
// //  getRuleSourceString
// //
// //-----------------------------------------------------------------------------
// const UnicodeString &RBBIDataWrapper::getRuleSourceString() const {
//     return fRuleString;
// }


// //-----------------------------------------------------------------------------
// //
// //  print   -  debugging function to dump the runtime data tables.
// //
// //-----------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void  RBBIDataWrapper::printTable(const char *heading, const RBBIStateTable *table) {
//     uint32_t   c;
//     uint32_t   s;

//     RBBIDebugPrintf("   %s\n", heading);

//     RBBIDebugPrintf("State |  Acc  LA TagIx");
//     for (c=0; c<fHeader->fCatCount; c++) {RBBIDebugPrintf("%3d ", c);}
//     RBBIDebugPrintf("\n------|---------------"); for (c=0;c<fHeader->fCatCount; c++) {
//         RBBIDebugPrintf("----");
//     }
//     RBBIDebugPrintf("\n");

//     if (table == NULL) {
//         RBBIDebugPrintf("         N U L L   T A B L E\n\n");
//         return;
//     }
//     for (s=0; s<table->fNumStates; s++) {
//         RBBIStateTableRow *row = (RBBIStateTableRow *)
//                                   (table->fTableData + (table->fRowLen * s));
//         RBBIDebugPrintf("%4d  |  %3d %3d %3d ", s, row->fAccepting, row->fLookAhead, row->fTagIdx);
//         for (c=0; c<fHeader->fCatCount; c++)  {
//             RBBIDebugPrintf("%3d ", row->fNextState[c]);
//         }
//         RBBIDebugPrintf("\n");
//     }
//     RBBIDebugPrintf("\n");
// }
// #endif


// void  RBBIDataWrapper::printData() {
// #ifdef RBBI_DEBUG
//     RBBIDebugPrintf("RBBI Data at %p\n", (void *)fHeader);
//     RBBIDebugPrintf("   Version = {%d %d %d %d}\n", fHeader->fFormatVersion[0], fHeader->fFormatVersion[1],
//                                                     fHeader->fFormatVersion[2], fHeader->fFormatVersion[3]);
//     RBBIDebugPrintf("   total length of data  = %d\n", fHeader->fLength);
//     RBBIDebugPrintf("   number of character categories = %d\n\n", fHeader->fCatCount);

//     printTable("Forward State Transition Table", fForwardTable);
//     printTable("Reverse State Transition Table", fReverseTable);

//     RBBIDebugPrintf("\nOrignal Rules source:\n");
//     for (int32_t c=0; fRuleSource[c] != 0; c++) {
//         RBBIDebugPrintf("%c", fRuleSource[c]);
//     }
//     RBBIDebugPrintf("\n\n");
// #endif
// }


// U_NAMESPACE_END
// U_NAMESPACE_USE

// //-----------------------------------------------------------------------------
// //
// //  ubrk_swap   -  byte swap and char encoding swap of RBBI data
// //
// //-----------------------------------------------------------------------------

// U_CAPI int32_t U_EXPORT2
// ubrk_swap(const UDataSwapper *ds, const void *inData, int32_t length, void *outData,
//            UErrorCode *status) {

//     if (status == NULL || U_FAILURE(*status)) {
//         return 0;
//     }
//     if(ds==NULL || inData==NULL || length<-1 || (length>0 && outData==NULL)) {
//         *status=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     //
//     //  Check that the data header is for for break data.
//     //    (Header contents are defined in genbrk.cpp)
//     //
//     const UDataInfo *pInfo = (const UDataInfo *)((const char *)inData+4);
//     if(!(  pInfo->dataFormat[0]==0x42 &&   /* dataFormat="Brk " */
//            pInfo->dataFormat[1]==0x72 &&
//            pInfo->dataFormat[2]==0x6b &&
//            pInfo->dataFormat[3]==0x20 &&
//            RBBIDataWrapper::isDataVersionAcceptable(pInfo->formatVersion) )) {
//         udata_printError(ds, "ubrk_swap(): data format %02x.%02x.%02x.%02x (format version %02x) is not recognized\n",
//                          pInfo->dataFormat[0], pInfo->dataFormat[1],
//                          pInfo->dataFormat[2], pInfo->dataFormat[3],
//                          pInfo->formatVersion[0]);
//         *status=U_UNSUPPORTED_ERROR;
//         return 0;
//     }

//     //
//     // Swap the data header.  (This is the generic ICU Data Header, not the RBBI Specific
//     //                         RBBIDataHeader).  This swap also conveniently gets us
//     //                         the size of the ICU d.h., which lets us locate the start
//     //                         of the RBBI specific data.
//     //
//     int32_t headerSize=udata_swapDataHeader(ds, inData, length, outData, status);


//     //
//     // Get the RRBI Data Header, and check that it appears to be OK.
//     //
//     const uint8_t  *inBytes =(const uint8_t *)inData+headerSize;
//     RBBIDataHeader *rbbiDH = (RBBIDataHeader *)inBytes;
//     if (ds->readUInt32(rbbiDH->fMagic) != 0xb1a0 ||
//             !RBBIDataWrapper::isDataVersionAcceptable(rbbiDH->fFormatVersion) ||
//             ds->readUInt32(rbbiDH->fLength)  <  sizeof(RBBIDataHeader)) {
//         udata_printError(ds, "ubrk_swap(): RBBI Data header is invalid.\n");
//         *status=U_UNSUPPORTED_ERROR;
//         return 0;
//     }

//     //
//     // Prefight operation?  Just return the size
//     //
//     int32_t breakDataLength = ds->readUInt32(rbbiDH->fLength);
//     int32_t totalSize = headerSize + breakDataLength;
//     if (length < 0) {
//         return totalSize;
//     }

//     //
//     // Check that length passed in is consistent with length from RBBI data header.
//     //
//     if (length < totalSize) {
//         udata_printError(ds, "ubrk_swap(): too few bytes (%d after ICU Data header) for break data.\n",
//                             breakDataLength);
//         *status=U_INDEX_OUTOFBOUNDS_ERROR;
//         return 0;
//         }


//     //
//     // Swap the Data.  Do the data itself first, then the RBBI Data Header, because
//     //                 we need to reference the header to locate the data, and an
//     //                 inplace swap of the header leaves it unusable.
//     //
//     uint8_t         *outBytes = (uint8_t *)outData + headerSize;
//     RBBIDataHeader  *outputDH = (RBBIDataHeader *)outBytes;

//     int32_t   tableStartOffset;
//     int32_t   tableLength;

//     //
//     // If not swapping in place, zero out the output buffer before starting.
//     //    Individual tables and other data items within are aligned to 8 byte boundaries
//     //    when originally created.  Any unused space between items needs to be zero.
//     //
//     if (inBytes != outBytes) {
//         uprv_memset(outBytes, 0, breakDataLength);
//     }

//     //
//     // Each state table begins with several 32 bit fields.  Calculate the size
//     //   in bytes of these.
//     //
//     int32_t         topSize = offsetof(RBBIStateTable, fTableData);

//     // Forward state table.
//     tableStartOffset = ds->readUInt32(rbbiDH->fFTable);
//     tableLength      = ds->readUInt32(rbbiDH->fFTableLen);

//     if (tableLength > 0) {
//         ds->swapArray32(ds, inBytes+tableStartOffset, topSize,
//                             outBytes+tableStartOffset, status);
//         ds->swapArray16(ds, inBytes+tableStartOffset+topSize, tableLength-topSize,
//                             outBytes+tableStartOffset+topSize, status);
//     }

//     // Reverse state table.  Same layout as forward table, above.
//     tableStartOffset = ds->readUInt32(rbbiDH->fRTable);
//     tableLength      = ds->readUInt32(rbbiDH->fRTableLen);

//     if (tableLength > 0) {
//         ds->swapArray32(ds, inBytes+tableStartOffset, topSize,
//                             outBytes+tableStartOffset, status);
//         ds->swapArray16(ds, inBytes+tableStartOffset+topSize, tableLength-topSize,
//                             outBytes+tableStartOffset+topSize, status);
//     }

//     // Trie table for character categories
//     utrie2_swap(ds, inBytes+ds->readUInt32(rbbiDH->fTrie), ds->readUInt32(rbbiDH->fTrieLen),
//                     outBytes+ds->readUInt32(rbbiDH->fTrie), status);

//     // Source Rules Text.  It's UChar data
//     ds->swapArray16(ds, inBytes+ds->readUInt32(rbbiDH->fRuleSource), ds->readUInt32(rbbiDH->fRuleSourceLen),
//                         outBytes+ds->readUInt32(rbbiDH->fRuleSource), status);

//     // Table of rule status values.  It's all int_32 values
//     ds->swapArray32(ds, inBytes+ds->readUInt32(rbbiDH->fStatusTable), ds->readUInt32(rbbiDH->fStatusTableLen),
//                         outBytes+ds->readUInt32(rbbiDH->fStatusTable), status);

//     // And, last, the header.
//     //   It is all int32_t values except for fFormataVersion, which is an array of four bytes.
//     //   Swap the whole thing as int32_t, then re-swap the one field.
//     //
//     ds->swapArray32(ds, inBytes, sizeof(RBBIDataHeader), outBytes, status);
//     ds->swapArray32(ds, outputDH->fFormatVersion, 4, outputDH->fFormatVersion, status);

//     return totalSize;
// }


// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
***************************************************************************
*   Copyright (C) 2002-2016 International Business Machines Corporation   *
*   and others. All rights reserved.                                      *
***************************************************************************
*/

//
//  File:  rbbinode.cpp
//
//         Implementation of class RBBINode, which represents a node in the
//         tree generated when parsing the Rules Based Break Iterator rules.
//
//         This "Class" is actually closer to a struct.
//         Code using it is expected to directly access fields much of the time.
//



// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/unistr.h"
// #include "unicode/uniset.h"
// #include "unicode/uchar.h"
// #include "unicode/parsepos.h"

// #include "cstr.h"
// #include "uvector.h"

// #include "rbbirb.h"
// #include "rbbinode.h"

// #include "uassert.h"


// U_NAMESPACE_BEGIN

// #ifdef RBBI_DEBUG
// static int  gLastSerial = 0;
// #endif


// //-------------------------------------------------------------------------
// //
// //    Constructor.   Just set the fields to reasonable default values.
// //
// //-------------------------------------------------------------------------
// RBBINode::RBBINode(NodeType t) : UMemory() {
// #ifdef RBBI_DEBUG
//     fSerialNum    = ++gLastSerial;
// #endif
//     fType         = t;
//     fParent       = NULL;
//     fLeftChild    = NULL;
//     fRightChild   = NULL;
//     fInputSet     = NULL;
//     fFirstPos     = 0;
//     fLastPos      = 0;
//     fNullable     = FALSE;
//     fLookAheadEnd = FALSE;
//     fRuleRoot     = FALSE;
//     fChainIn      = FALSE;
//     fVal          = 0;
//     fPrecedence   = precZero;

//     UErrorCode     status = U_ZERO_ERROR;
//     fFirstPosSet  = new UVector(status);  // TODO - get a real status from somewhere
//     fLastPosSet   = new UVector(status);
//     fFollowPos    = new UVector(status);
//     if      (t==opCat)    {fPrecedence = precOpCat;}
//     else if (t==opOr)     {fPrecedence = precOpOr;}
//     else if (t==opStart)  {fPrecedence = precStart;}
//     else if (t==opLParen) {fPrecedence = precLParen;}

// }


// RBBINode::RBBINode(const RBBINode &other) : UMemory(other) {
// #ifdef RBBI_DEBUG
//     fSerialNum   = ++gLastSerial;
// #endif
//     fType        = other.fType;
//     fParent      = NULL;
//     fLeftChild   = NULL;
//     fRightChild  = NULL;
//     fInputSet    = other.fInputSet;
//     fPrecedence  = other.fPrecedence;
//     fText        = other.fText;
//     fFirstPos    = other.fFirstPos;
//     fLastPos     = other.fLastPos;
//     fNullable    = other.fNullable;
//     fVal         = other.fVal;
//     fRuleRoot    = FALSE;
//     fChainIn     = other.fChainIn;
//     UErrorCode     status = U_ZERO_ERROR;
//     fFirstPosSet = new UVector(status);   // TODO - get a real status from somewhere
//     fLastPosSet  = new UVector(status);
//     fFollowPos   = new UVector(status);
// }


// //-------------------------------------------------------------------------
// //
// //    Destructor.   Deletes both this node AND any child nodes,
// //                  except in the case of variable reference nodes.  For
// //                  these, the l. child points back to the definition, which
// //                  is common for all references to the variable, meaning
// //                  it can't be deleted here.
// //
// //-------------------------------------------------------------------------
// RBBINode::~RBBINode() {
//     // printf("deleting node %8x   serial %4d\n", this, this->fSerialNum);
//     delete fInputSet;
//     fInputSet = NULL;

//     switch (this->fType) {
//     case varRef:
//     case setRef:
//         // for these node types, multiple instances point to the same "children"
//         // Storage ownership of children handled elsewhere.  Don't delete here.
//         break;

//     default:
//         delete        fLeftChild;
//         fLeftChild =   NULL;
//         delete        fRightChild;
//         fRightChild = NULL;
//     }


//     delete fFirstPosSet;
//     delete fLastPosSet;
//     delete fFollowPos;

// }


// //-------------------------------------------------------------------------
// //
// //    cloneTree     Make a copy of the subtree rooted at this node.
// //                  Discard any variable references encountered along the way,
// //                  and replace with copies of the variable's definitions.
// //                  Used to replicate the expression underneath variable
// //                  references in preparation for generating the DFA tables.
// //
// //-------------------------------------------------------------------------
// RBBINode *RBBINode::cloneTree() {
//     RBBINode    *n;

//     if (fType == RBBINode::varRef) {
//         // If the current node is a variable reference, skip over it
//         //   and clone the definition of the variable instead.
//         n = fLeftChild->cloneTree();
//     } else if (fType == RBBINode::uset) {
//         n = this;
//     } else {
//         n = new RBBINode(*this);
//         // Check for null pointer.
//         if (n != NULL) {
//             if (fLeftChild != NULL) {
//                 n->fLeftChild          = fLeftChild->cloneTree();
//                 n->fLeftChild->fParent = n;
//             }
//             if (fRightChild != NULL) {
//                 n->fRightChild          = fRightChild->cloneTree();
//                 n->fRightChild->fParent = n;
//             }
//         }
//     }
//     return n;
// }



// //-------------------------------------------------------------------------
// //
// //   flattenVariables   Walk a parse tree, replacing any variable
// //                      references with a copy of the variable's definition.
// //                      Aside from variables, the tree is not changed.
// //
// //                      Return the root of the tree.  If the root was not a variable
// //                      reference, it remains unchanged - the root we started with
// //                      is the root we return.  If, however, the root was a variable
// //                      reference, the root of the newly cloned replacement tree will
// //                      be returned, and the original tree deleted.
// //
// //                      This function works by recursively walking the tree
// //                      without doing anything until a variable reference is
// //                      found, then calling cloneTree() at that point.  Any
// //                      nested references are handled by cloneTree(), not here.
// //
// //-------------------------------------------------------------------------
// RBBINode *RBBINode::flattenVariables() {
//     if (fType == varRef) {
//         RBBINode *retNode  = fLeftChild->cloneTree();
//         if (retNode != NULL) {
//             retNode->fRuleRoot = this->fRuleRoot;
//             retNode->fChainIn  = this->fChainIn;
//         }
//         delete this;   // TODO: undefined behavior. Fix.
//         return retNode;
//     }

//     if (fLeftChild != NULL) {
//         fLeftChild = fLeftChild->flattenVariables();
//         fLeftChild->fParent  = this;
//     }
//     if (fRightChild != NULL) {
//         fRightChild = fRightChild->flattenVariables();
//         fRightChild->fParent = this;
//     }
//     return this;
// }


// //-------------------------------------------------------------------------
// //
// //  flattenSets    Walk the parse tree, replacing any nodes of type setRef
// //                 with a copy of the expression tree for the set.  A set's
// //                 equivalent expression tree is precomputed and saved as
// //                 the left child of the uset node.
// //
// //-------------------------------------------------------------------------
// void RBBINode::flattenSets() {
//     U_ASSERT(fType != setRef);

//     if (fLeftChild != NULL) {
//         if (fLeftChild->fType==setRef) {
//             RBBINode *setRefNode = fLeftChild;
//             RBBINode *usetNode   = setRefNode->fLeftChild;
//             RBBINode *replTree   = usetNode->fLeftChild;
//             fLeftChild           = replTree->cloneTree();
//             fLeftChild->fParent  = this;
//             delete setRefNode;
//         } else {
//             fLeftChild->flattenSets();
//         }
//     }

//     if (fRightChild != NULL) {
//         if (fRightChild->fType==setRef) {
//             RBBINode *setRefNode = fRightChild;
//             RBBINode *usetNode   = setRefNode->fLeftChild;
//             RBBINode *replTree   = usetNode->fLeftChild;
//             fRightChild           = replTree->cloneTree();
//             fRightChild->fParent  = this;
//             delete setRefNode;
//         } else {
//             fRightChild->flattenSets();
//         }
//     }
// }



// //-------------------------------------------------------------------------
// //
// //   findNodes()     Locate all the nodes of the specified type, starting
// //                   at the specified root.
// //
// //-------------------------------------------------------------------------
// void   RBBINode::findNodes(UVector *dest, RBBINode::NodeType kind, UErrorCode &status) {
//     /* test for buffer overflows */
//     if (U_FAILURE(status)) {
//         return;
//     }
//     if (fType == kind) {
//         dest->addElement(this, status);
//     }
//     if (fLeftChild != NULL) {
//         fLeftChild->findNodes(dest, kind, status);
//     }
//     if (fRightChild != NULL) {
//         fRightChild->findNodes(dest, kind, status);
//     }
// }


// //-------------------------------------------------------------------------
// //
// //    print.         Print out a single node, for debugging.
// //
// //-------------------------------------------------------------------------
// #ifdef RBBI_DEBUG

// static int32_t serial(const RBBINode *node) {
//     return (node == NULL? -1 : node->fSerialNum);
// }


// void RBBINode::printNode(const RBBINode *node) {
//     static const char * const nodeTypeNames[] = {
//                 "setRef",
//                 "uset",
//                 "varRef",
//                 "leafChar",
//                 "lookAhead",
//                 "tag",
//                 "endMark",
//                 "opStart",
//                 "opCat",
//                 "opOr",
//                 "opStar",
//                 "opPlus",
//                 "opQuestion",
//                 "opBreak",
//                 "opReverse",
//                 "opLParen"
//     };

//     if (node==NULL) {
//         RBBIDebugPrintf("%10p", (void *)node);
//     } else {
//         RBBIDebugPrintf("%10p %5d %12s %c%c  %5d       %5d     %5d       %6d     %d ",
//             (void *)node, node->fSerialNum, nodeTypeNames[node->fType],
//             node->fRuleRoot?'R':' ', node->fChainIn?'C':' ',
//             serial(node->fLeftChild), serial(node->fRightChild), serial(node->fParent),
//             node->fFirstPos, node->fVal);
//         if (node->fType == varRef) {
//             RBBI_DEBUG_printUnicodeString(node->fText);
//         }
//     }
//     RBBIDebugPrintf("\n");
// }
// #endif


// #ifdef RBBI_DEBUG
// U_CFUNC void RBBI_DEBUG_printUnicodeString(const UnicodeString &s, int minWidth) {
//     RBBIDebugPrintf("%*s", minWidth, CStr(s)());
// }
// #endif


// //-------------------------------------------------------------------------
// //
// //    print.         Print out the tree of nodes rooted at "this"
// //
// //-------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void RBBINode::printNodeHeader() {
//     RBBIDebugPrintf(" Address   serial        type     LeftChild  RightChild   Parent   position value\n");
// }

// void RBBINode::printTree(const RBBINode *node, UBool printHeading) {
//     if (printHeading) {
//         printNodeHeader();
//     }
//     printNode(node);
//     if (node != NULL) {
//         // Only dump the definition under a variable reference if asked to.
//         // Unconditinally dump children of all other node types.
//         if (node->fType != varRef) {
//             if (node->fLeftChild != NULL) {
//                 printTree(node->fLeftChild, FALSE);
//             }

//             if (node->fRightChild != NULL) {
//                 printTree(node->fRightChild, FALSE);
//             }
//         }
//     }
// }
// #endif



// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
//  file:  rbbirb.cpp
//
//  Copyright (C) 2002-2011, International Business Machines Corporation and others.
//  All Rights Reserved.
//
//  This file contains the RBBIRuleBuilder class implementation.  This is the main class for
//    building (compiling) break rules into the tables required by the runtime
//    RBBI engine.
//



// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/brkiter.h"
// #include "unicode/rbbi.h"
// #include "unicode/ubrk.h"
// #include "unicode/unistr.h"
// #include "unicode/uniset.h"
// #include "unicode/uchar.h"
// #include "unicode/uchriter.h"
// #include "unicode/parsepos.h"
// #include "unicode/parseerr.h"

// #include "cmemory.h"
// #include "cstring.h"
// #include "rbbirb.h"
// #include "rbbinode.h"
// #include "rbbiscan.h"
// #include "rbbisetb.h"
// #include "rbbitblb.h"
// #include "rbbidata.h"
// #include "uassert.h"


// U_NAMESPACE_BEGIN


// //----------------------------------------------------------------------------------------
// //
// //  Constructor.
// //
// //----------------------------------------------------------------------------------------
// RBBIRuleBuilder::RBBIRuleBuilder(const UnicodeString   &rules,
//                                        UParseError     *parseErr,
//                                        UErrorCode      &status)
//  : fRules(rules), fStrippedRules(rules)
// {
//     fStatus = &status; // status is checked below
//     fParseError = parseErr;
//     fDebugEnv   = NULL;
// #ifdef RBBI_DEBUG
//     fDebugEnv   = getenv("U_RBBIDEBUG");
// #endif


//     fForwardTree        = NULL;
//     fReverseTree        = NULL;
//     fSafeFwdTree        = NULL;
//     fSafeRevTree        = NULL;
//     fDefaultTree        = &fForwardTree;
//     fForwardTable       = NULL;
//     fRuleStatusVals     = NULL;
//     fChainRules         = FALSE;
//     fLBCMNoChain        = FALSE;
//     fLookAheadHardBreak = FALSE;
//     fUSetNodes          = NULL;
//     fRuleStatusVals     = NULL;
//     fScanner            = NULL;
//     fSetBuilder         = NULL;
//     if (parseErr) {
//         uprv_memset(parseErr, 0, sizeof(UParseError));
//     }

//     if (U_FAILURE(status)) {
//         return;
//     }

//     fUSetNodes          = new UVector(status); // bcos status gets overwritten here
//     fRuleStatusVals     = new UVector(status);
//     fScanner            = new RBBIRuleScanner(this);
//     fSetBuilder         = new RBBISetBuilder(this);
//     if (U_FAILURE(status)) {
//         return;
//     }
//     if(fSetBuilder == 0 || fScanner == 0 || fUSetNodes == 0 || fRuleStatusVals == 0) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//     }
// }



// //----------------------------------------------------------------------------------------
// //
// //  Destructor
// //
// //----------------------------------------------------------------------------------------
// RBBIRuleBuilder::~RBBIRuleBuilder() {

//     int        i;
//     for (i=0; ; i++) {
//         RBBINode *n = (RBBINode *)fUSetNodes->elementAt(i);
//         if (n==NULL) {
//             break;
//         }
//         delete n;
//     }

//     delete fUSetNodes;
//     delete fSetBuilder;
//     delete fForwardTable;
//     delete fForwardTree;
//     delete fReverseTree;
//     delete fSafeFwdTree;
//     delete fSafeRevTree;
//     delete fScanner;
//     delete fRuleStatusVals;
// }





// //----------------------------------------------------------------------------------------
// //
// //   flattenData() -  Collect up the compiled RBBI rule data and put it into
// //                    the format for saving in ICU data files,
// //                    which is also the format needed by the RBBI runtime engine.
// //
// //----------------------------------------------------------------------------------------
// static int32_t align8(int32_t i) {return (i+7) & 0xfffffff8;}

// RBBIDataHeader *RBBIRuleBuilder::flattenData() {
//     int32_t    i;

//     if (U_FAILURE(*fStatus)) {
//         return NULL;
//     }

//     // Remove whitespace from the rules to make it smaller.
//     // The rule parser has already removed comments.
//     fStrippedRules = fScanner->stripRules(fStrippedRules);

//     // Calculate the size of each section in the data.
//     //   Sizes here are padded up to a multiple of 8 for better memory alignment.
//     //   Sections sizes actually stored in the header are for the actual data
//     //     without the padding.
//     //
//     int32_t headerSize        = align8(sizeof(RBBIDataHeader));
//     int32_t forwardTableSize  = align8(fForwardTable->getTableSize());
//     int32_t reverseTableSize  = align8(fForwardTable->getSafeTableSize());
//     int32_t trieSize          = align8(fSetBuilder->getTrieSize());
//     int32_t statusTableSize   = align8(fRuleStatusVals->size() * sizeof(int32_t));
//     int32_t rulesSize         = align8((fStrippedRules.length()+1) * sizeof(UChar));

//     int32_t         totalSize = headerSize
//                                 + forwardTableSize
//                                 + reverseTableSize
//                                 + statusTableSize + trieSize + rulesSize;

//     RBBIDataHeader  *data     = (RBBIDataHeader *)uprv_malloc(totalSize);
//     if (data == NULL) {
//         *fStatus = U_MEMORY_ALLOCATION_ERROR;
//         return NULL;
//     }
//     uprv_memset(data, 0, totalSize);


//     data->fMagic            = 0xb1a0;
//     data->fFormatVersion[0] = RBBI_DATA_FORMAT_VERSION[0];
//     data->fFormatVersion[1] = RBBI_DATA_FORMAT_VERSION[1];
//     data->fFormatVersion[2] = RBBI_DATA_FORMAT_VERSION[2];
//     data->fFormatVersion[3] = RBBI_DATA_FORMAT_VERSION[3];
//     data->fLength           = totalSize;
//     data->fCatCount         = fSetBuilder->getNumCharCategories();

//     data->fFTable        = headerSize;
//     data->fFTableLen     = forwardTableSize;

//     data->fRTable        = data->fFTable  + data->fFTableLen;
//     data->fRTableLen     = reverseTableSize;

//     data->fTrie          = data->fRTable + data->fRTableLen;
//     data->fTrieLen       = fSetBuilder->getTrieSize();
//     data->fStatusTable   = data->fTrie    + trieSize;
//     data->fStatusTableLen= statusTableSize;
//     data->fRuleSource    = data->fStatusTable + statusTableSize;
//     data->fRuleSourceLen = fStrippedRules.length() * sizeof(UChar);

//     uprv_memset(data->fReserved, 0, sizeof(data->fReserved));

//     fForwardTable->exportTable((uint8_t *)data + data->fFTable);
//     fForwardTable->exportSafeTable((uint8_t *)data + data->fRTable);
//     fSetBuilder->serializeTrie ((uint8_t *)data + data->fTrie);

//     int32_t *ruleStatusTable = (int32_t *)((uint8_t *)data + data->fStatusTable);
//     for (i=0; i<fRuleStatusVals->size(); i++) {
//         ruleStatusTable[i] = fRuleStatusVals->elementAti(i);
//     }

//     fStrippedRules.extract((UChar *)((uint8_t *)data+data->fRuleSource), rulesSize/2+1, *fStatus);

//     return data;
// }


// //----------------------------------------------------------------------------------------
// //
// //  createRuleBasedBreakIterator    construct from source rules that are passed in
// //                                  in a UnicodeString
// //
// //----------------------------------------------------------------------------------------
// BreakIterator *
// RBBIRuleBuilder::createRuleBasedBreakIterator( const UnicodeString    &rules,
//                                     UParseError      *parseError,
//                                     UErrorCode       &status)
// {
//     //
//     // Read the input rules, generate a parse tree, symbol table,
//     // and list of all Unicode Sets referenced by the rules.
//     //
//     RBBIRuleBuilder  builder(rules, parseError, status);
//     if (U_FAILURE(status)) { // status checked here bcos build below doesn't
//         return NULL;
//     }

//     RBBIDataHeader *data = builder.build(status);

//     if (U_FAILURE(status)) {
//         return nullptr;
//     }

//     //
//     //  Create a break iterator from the compiled rules.
//     //     (Identical to creation from stored pre-compiled rules)
//     //
//     // status is checked after init in construction.
//     RuleBasedBreakIterator *This = new RuleBasedBreakIterator(data, status);
//     if (U_FAILURE(status)) {
//         delete This;
//         This = NULL;
//     }
//     else if(This == NULL) { // test for NULL
//         status = U_MEMORY_ALLOCATION_ERROR;
//     }
//     return This;
// }

// RBBIDataHeader *RBBIRuleBuilder::build(UErrorCode &status) {
//     if (U_FAILURE(status)) {
//         return nullptr;
//     }

//     fScanner->parse();
//     if (U_FAILURE(status)) {
//         return nullptr;
//     }

//     //
//     // UnicodeSet processing.
//     //    Munge the Unicode Sets to create a set of character categories.
//     //    Generate the mapping tables (TRIE) from input code points to
//     //    the character categories.
//     //
//     fSetBuilder->buildRanges();

//     //
//     //   Generate the DFA state transition table.
//     //
//     fForwardTable = new RBBITableBuilder(this, &fForwardTree, status);
//     if (fForwardTable == nullptr) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//         return nullptr;
//     }

//     fForwardTable->buildForwardTable();
//     optimizeTables();
//     fForwardTable->buildSafeReverseTable(status);


// #ifdef RBBI_DEBUG
//     if (fDebugEnv && uprv_strstr(fDebugEnv, "states")) {
//         fForwardTable->printStates();
//         fForwardTable->printRuleStatusTable();
//         fForwardTable->printReverseTable();
//     }
// #endif

//     fSetBuilder->buildTrie();

//     //
//     //   Package up the compiled data into a memory image
//     //      in the run-time format.
//     //
//     RBBIDataHeader *data = flattenData(); // returns NULL if error
//     if (U_FAILURE(status)) {
//         return nullptr;
//     }
//     return data;
// }

// void RBBIRuleBuilder::optimizeTables() {
//     bool didSomething;
//     do {
//         didSomething = false;

//         // Begin looking for duplicates with char class 3.
//         // Classes 0, 1 and 2 are special; they are unused, {bof} and {eof} respectively,
//         // and should not have other categories merged into them.
//         IntPair duplPair = {3, 0};
//         while (fForwardTable->findDuplCharClassFrom(&duplPair)) {
//             fSetBuilder->mergeCategories(duplPair);
//             fForwardTable->removeColumn(duplPair.second);
//             didSomething = true;
//         }

//         while (fForwardTable->removeDuplicateStates() > 0) {
//             didSomething = true;
//         }
//     } while (didSomething);
// }

// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// //
// //  file:  rbbiscan.cpp
// //
// //  Copyright (C) 2002-2016, International Business Machines Corporation and others.
// //  All Rights Reserved.
// //
// //  This file contains the Rule Based Break Iterator Rule Builder functions for
// //   scanning the rules and assembling a parse tree.  This is the first phase
// //   of compiling the rules.
// //
// //  The overall of the rules is managed by class RBBIRuleBuilder, which will
// //  create and use an instance of this class as part of the process.
// //

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/unistr.h"
// #include "unicode/uniset.h"
// #include "unicode/uchar.h"
// #include "unicode/uchriter.h"
// #include "unicode/parsepos.h"
// #include "unicode/parseerr.h"
// #include "cmemory.h"
// #include "cstring.h"

// #include "rbbirpt.h"   // Contains state table for the rbbi rules parser.
//                        //   generated by a Perl script.
// #include "rbbirb.h"
// #include "rbbinode.h"
// #include "rbbiscan.h"
// #include "rbbitblb.h"

// #include "uassert.h"

// //------------------------------------------------------------------------------
// //
// // Unicode Set init strings for each of the character classes needed for parsing a rule file.
// //               (Initialized with hex values for portability to EBCDIC based machines.
// //                Really ugly, but there's no good way to avoid it.)
// //
// //              The sets are referred to by name in the rbbirpt.txt, which is the
// //              source form of the state transition table for the RBBI rule parser.
// //
// //------------------------------------------------------------------------------
// static const UChar gRuleSet_rule_char_pattern[]       = {
//  // Characters that may appear as literals in patterns without escaping or quoting.
//  //   [    ^      [    \     p     {      Z     }     \     u    0      0    2      0
//     0x5b, 0x5e, 0x5b, 0x5c, 0x70, 0x7b, 0x5a, 0x7d, 0x5c, 0x75, 0x30, 0x30, 0x32, 0x30,
//  //   -    \      u    0     0     7      f     ]     -     [    \      p
//     0x2d, 0x5c, 0x75, 0x30, 0x30, 0x37, 0x66, 0x5d, 0x2d, 0x5b, 0x5c, 0x70,
//  //   {     L     }    ]     -     [      \     p     {     N    }      ]     ]
//     0x7b, 0x4c, 0x7d, 0x5d, 0x2d, 0x5b, 0x5c, 0x70, 0x7b, 0x4e, 0x7d, 0x5d, 0x5d, 0};

// static const UChar gRuleSet_name_char_pattern[]       = {
// //    [    _      \    p     {     L      }     \     p     {    N      }     ]
//     0x5b, 0x5f, 0x5c, 0x70, 0x7b, 0x4c, 0x7d, 0x5c, 0x70, 0x7b, 0x4e, 0x7d, 0x5d, 0};

// static const UChar gRuleSet_digit_char_pattern[] = {
// //    [    0      -    9     ]
//     0x5b, 0x30, 0x2d, 0x39, 0x5d, 0};

// static const UChar gRuleSet_name_start_char_pattern[] = {
// //    [    _      \    p     {     L      }     ]
//     0x5b, 0x5f, 0x5c, 0x70, 0x7b, 0x4c, 0x7d, 0x5d, 0 };

// static const UChar kAny[] = {0x61, 0x6e, 0x79, 0x00};  // "any"


// U_CDECL_BEGIN
// static void U_CALLCONV RBBISetTable_deleter(void *p) {
//     icu::RBBISetTableEl *px = (icu::RBBISetTableEl *)p;
//     delete px->key;
//     // Note:  px->val is owned by the linked list "fSetsListHead" in scanner.
//     //        Don't delete the value nodes here.
//     uprv_free(px);
// }
// U_CDECL_END

// U_NAMESPACE_BEGIN

// //------------------------------------------------------------------------------
// //
// //  Constructor.
// //
// //------------------------------------------------------------------------------
// RBBIRuleScanner::RBBIRuleScanner(RBBIRuleBuilder *rb)
// {
//     fRB                 = rb;
//     fScanIndex          = 0;
//     fNextIndex          = 0;
//     fQuoteMode          = FALSE;
//     fLineNum            = 1;
//     fCharNum            = 0;
//     fLastChar           = 0;

//     fStateTable         = NULL;
//     fStack[0]           = 0;
//     fStackPtr           = 0;
//     fNodeStack[0]       = NULL;
//     fNodeStackPtr       = 0;

//     fReverseRule        = FALSE;
//     fLookAheadRule      = FALSE;
//     fNoChainInRule      = FALSE;

//     fSymbolTable        = NULL;
//     fSetTable           = NULL;
//     fRuleNum            = 0;
//     fOptionStart        = 0;

//     // Do not check status until after all critical fields are sufficiently initialized
//     //   that the destructor can run cleanly.
//     if (U_FAILURE(*rb->fStatus)) {
//         return;
//     }

//     //
//     //  Set up the constant Unicode Sets.
//     //     Note:  These could be made static, lazily initialized, and shared among
//     //            all instances of RBBIRuleScanners.  BUT this is quite a bit simpler,
//     //            and the time to build these few sets should be small compared to a
//     //            full break iterator build.
//     fRuleSets[kRuleSet_rule_char-128]
//         = UnicodeSet(UnicodeString(gRuleSet_rule_char_pattern),       *rb->fStatus);
//     // fRuleSets[kRuleSet_white_space-128] = [:Pattern_White_Space:]
//     fRuleSets[kRuleSet_white_space-128].
//         add(9, 0xd).add(0x20).add(0x85).add(0x200e, 0x200f).add(0x2028, 0x2029);
//     fRuleSets[kRuleSet_name_char-128]
//         = UnicodeSet(UnicodeString(gRuleSet_name_char_pattern),       *rb->fStatus);
//     fRuleSets[kRuleSet_name_start_char-128]
//         = UnicodeSet(UnicodeString(gRuleSet_name_start_char_pattern), *rb->fStatus);
//     fRuleSets[kRuleSet_digit_char-128]
//         = UnicodeSet(UnicodeString(gRuleSet_digit_char_pattern),      *rb->fStatus);
//     if (*rb->fStatus == U_ILLEGAL_ARGUMENT_ERROR) {
//         // This case happens if ICU's data is missing.  UnicodeSet tries to look up property
//         //   names from the init string, can't find them, and claims an illegal argument.
//         //   Change the error so that the actual problem will be clearer to users.
//         *rb->fStatus = U_BRK_INIT_ERROR;
//     }
//     if (U_FAILURE(*rb->fStatus)) {
//         return;
//     }

//     fSymbolTable = new RBBISymbolTable(this, rb->fRules, *rb->fStatus);
//     if (fSymbolTable == NULL) {
//         *rb->fStatus = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
//     fSetTable    = uhash_open(uhash_hashUnicodeString, uhash_compareUnicodeString, NULL, rb->fStatus);
//     if (U_FAILURE(*rb->fStatus)) {
//         return;
//     }
//     uhash_setValueDeleter(fSetTable, RBBISetTable_deleter);
// }



// //------------------------------------------------------------------------------
// //
// //  Destructor
// //
// //------------------------------------------------------------------------------
// RBBIRuleScanner::~RBBIRuleScanner() {
//     delete fSymbolTable;
//     if (fSetTable != NULL) {
//          uhash_close(fSetTable);
//          fSetTable = NULL;

//     }


//     // Node Stack.
//     //   Normally has one entry, which is the entire parse tree for the rules.
//     //   If errors occured, there may be additional subtrees left on the stack.
//     while (fNodeStackPtr > 0) {
//         delete fNodeStack[fNodeStackPtr];
//         fNodeStackPtr--;
//     }

// }

// //------------------------------------------------------------------------------
// //
// //  doParseAction        Do some action during rule parsing.
// //                       Called by the parse state machine.
// //                       Actions build the parse tree and Unicode Sets,
// //                       and maintain the parse stack for nested expressions.
// //
// //                       TODO:  unify EParseAction and RBBI_RuleParseAction enum types.
// //                              They represent exactly the same thing.  They're separate
// //                              only to work around enum forward declaration restrictions
// //                              in some compilers, while at the same time avoiding multiple
// //                              definitions problems.  I'm sure that there's a better way.
// //
// //------------------------------------------------------------------------------
// UBool RBBIRuleScanner::doParseActions(int32_t action)
// {
//     RBBINode *n       = NULL;

//     UBool   returnVal = TRUE;

//     switch (action) {

//     case doExprStart:
//         pushNewNode(RBBINode::opStart);
//         fRuleNum++;
//         break;


//     case doNoChain:
//         // Scanned a '^' while on the rule start state.
//         fNoChainInRule = TRUE;
//         break;


//     case doExprOrOperator:
//         {
//             fixOpStack(RBBINode::precOpCat);
//             RBBINode  *operandNode = fNodeStack[fNodeStackPtr--];
//             RBBINode  *orNode      = pushNewNode(RBBINode::opOr);
//             if (U_FAILURE(*fRB->fStatus)) {
//                 break;
//             }
//             orNode->fLeftChild     = operandNode;
//             operandNode->fParent   = orNode;
//         }
//         break;

//     case doExprCatOperator:
//         // concatenation operator.
//         // For the implicit concatenation of adjacent terms in an expression that are
//         //   not separated by any other operator.  Action is invoked between the
//         //   actions for the two terms.
//         {
//             fixOpStack(RBBINode::precOpCat);
//             RBBINode  *operandNode = fNodeStack[fNodeStackPtr--];
//             RBBINode  *catNode     = pushNewNode(RBBINode::opCat);
//             if (U_FAILURE(*fRB->fStatus)) {
//                 break;
//             }
//             catNode->fLeftChild    = operandNode;
//             operandNode->fParent   = catNode;
//         }
//         break;

//     case doLParen:
//         // Open Paren.
//         //   The openParen node is a dummy operation type with a low precedence,
//         //     which has the affect of ensuring that any real binary op that
//         //     follows within the parens binds more tightly to the operands than
//         //     stuff outside of the parens.
//         pushNewNode(RBBINode::opLParen);
//         break;

//     case doExprRParen:
//         fixOpStack(RBBINode::precLParen);
//         break;

//     case doNOP:
//         break;

//     case doStartAssign:
//         // We've just scanned "$variable = "
//         // The top of the node stack has the $variable ref node.

//         // Save the start position of the RHS text in the StartExpression node
//         //   that precedes the $variableReference node on the stack.
//         //   This will eventually be used when saving the full $variable replacement
//         //   text as a string.
//         n = fNodeStack[fNodeStackPtr-1];
//         n->fFirstPos = fNextIndex;              // move past the '='

//         // Push a new start-of-expression node; needed to keep parse of the
//         //   RHS expression happy.
//         pushNewNode(RBBINode::opStart);
//         break;




//     case doEndAssign:
//         {
//             // We have reached the end of an assignement statement.
//             //   Current scan char is the ';' that terminates the assignment.

//             // Terminate expression, leaves expression parse tree rooted in TOS node.
//             fixOpStack(RBBINode::precStart);

//             RBBINode *startExprNode  = fNodeStack[fNodeStackPtr-2];
//             RBBINode *varRefNode     = fNodeStack[fNodeStackPtr-1];
//             RBBINode *RHSExprNode    = fNodeStack[fNodeStackPtr];

//             // Save original text of right side of assignment, excluding the terminating ';'
//             //  in the root of the node for the right-hand-side expression.
//             RHSExprNode->fFirstPos = startExprNode->fFirstPos;
//             RHSExprNode->fLastPos  = fScanIndex;
//             fRB->fRules.extractBetween(RHSExprNode->fFirstPos, RHSExprNode->fLastPos, RHSExprNode->fText);

//             // Expression parse tree becomes l. child of the $variable reference node.
//             varRefNode->fLeftChild = RHSExprNode;
//             RHSExprNode->fParent   = varRefNode;

//             // Make a symbol table entry for the $variableRef node.
//             fSymbolTable->addEntry(varRefNode->fText, varRefNode, *fRB->fStatus);
//             if (U_FAILURE(*fRB->fStatus)) {
//                 // This is a round-about way to get the parse position set
//                 //  so that duplicate symbols error messages include a line number.
//                 UErrorCode t = *fRB->fStatus;
//                 *fRB->fStatus = U_ZERO_ERROR;
//                 error(t);
//             }

//             // Clean up the stack.
//             delete startExprNode;
//             fNodeStackPtr-=3;
//             break;
//         }

//     case doEndOfRule:
//         {
//         fixOpStack(RBBINode::precStart);      // Terminate expression, leaves expression
//         if (U_FAILURE(*fRB->fStatus)) {       //   parse tree rooted in TOS node.
//             break;
//         }
// #ifdef RBBI_DEBUG
//         if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "rtree")) {printNodeStack("end of rule");}
// #endif
//         U_ASSERT(fNodeStackPtr == 1);
//         RBBINode *thisRule = fNodeStack[fNodeStackPtr];

//         // If this rule includes a look-ahead '/', add a endMark node to the
//         //   expression tree.
//         if (fLookAheadRule) {
//             RBBINode  *endNode        = pushNewNode(RBBINode::endMark);
//             RBBINode  *catNode        = pushNewNode(RBBINode::opCat);
//             if (U_FAILURE(*fRB->fStatus)) {
//                 break;
//             }
//             fNodeStackPtr -= 2;
//             catNode->fLeftChild       = thisRule;
//             catNode->fRightChild      = endNode;
//             fNodeStack[fNodeStackPtr] = catNode;
//             endNode->fVal             = fRuleNum;
//             endNode->fLookAheadEnd    = TRUE;
//             thisRule                  = catNode;

//             // TODO: Disable chaining out of look-ahead (hard break) rules.
//             //   The break on rule match is forced, so there is no point in building up
//             //   the state table to chain into another rule for a longer match.
//         }

//         // Mark this node as being the root of a rule.
//         thisRule->fRuleRoot = TRUE;

//         // Flag if chaining into this rule is wanted.
//         //
//         if (fRB->fChainRules &&         // If rule chaining is enabled globally via !!chain
//                 !fNoChainInRule) {      //     and no '^' chain-in inhibit was on this rule
//             thisRule->fChainIn = TRUE;
//         }


//         // All rule expressions are ORed together.
//         // The ';' that terminates an expression really just functions as a '|' with
//         //   a low operator prededence.
//         //
//         // Each of the four sets of rules are collected separately.
//         //  (forward, reverse, safe_forward, safe_reverse)
//         //  OR this rule into the appropriate group of them.
//         //
//         RBBINode **destRules = (fReverseRule? &fRB->fSafeRevTree : fRB->fDefaultTree);

//         if (*destRules != NULL) {
//             // This is not the first rule encounted.
//             // OR previous stuff  (from *destRules)
//             // with the current rule expression (on the Node Stack)
//             //  with the resulting OR expression going to *destRules
//             //
//                        thisRule    = fNodeStack[fNodeStackPtr];
//             RBBINode  *prevRules   = *destRules;
//             RBBINode  *orNode      = pushNewNode(RBBINode::opOr);
//             if (U_FAILURE(*fRB->fStatus)) {
//                 break;
//             }
//             orNode->fLeftChild     = prevRules;
//             prevRules->fParent     = orNode;
//             orNode->fRightChild    = thisRule;
//             thisRule->fParent      = orNode;
//             *destRules             = orNode;
//         }
//         else
//         {
//             // This is the first rule encountered (for this direction).
//             // Just move its parse tree from the stack to *destRules.
//             *destRules = fNodeStack[fNodeStackPtr];
//         }
//         fReverseRule   = FALSE;   // in preparation for the next rule.
//         fLookAheadRule = FALSE;
//         fNoChainInRule = FALSE;
//         fNodeStackPtr  = 0;
//         }
//         break;


//     case doRuleError:
//         error(U_BRK_RULE_SYNTAX);
//         returnVal = FALSE;
//         break;


//     case doVariableNameExpectedErr:
//         error(U_BRK_RULE_SYNTAX);
//         break;


//     //
//     //  Unary operands  + ? *
//     //    These all appear after the operand to which they apply.
//     //    When we hit one, the operand (may be a whole sub expression)
//     //    will be on the top of the stack.
//     //    Unary Operator becomes TOS, with the old TOS as its one child.
//     case doUnaryOpPlus:
//         {
//             RBBINode  *operandNode = fNodeStack[fNodeStackPtr--];
//             RBBINode  *plusNode    = pushNewNode(RBBINode::opPlus);
//             if (U_FAILURE(*fRB->fStatus)) {
//                 break;
//             }
//             plusNode->fLeftChild   = operandNode;
//             operandNode->fParent   = plusNode;
//         }
//         break;

//     case doUnaryOpQuestion:
//         {
//             RBBINode  *operandNode = fNodeStack[fNodeStackPtr--];
//             RBBINode  *qNode       = pushNewNode(RBBINode::opQuestion);
//             if (U_FAILURE(*fRB->fStatus)) {
//                 break;
//             }
//             qNode->fLeftChild      = operandNode;
//             operandNode->fParent   = qNode;
//         }
//         break;

//     case doUnaryOpStar:
//         {
//             RBBINode  *operandNode = fNodeStack[fNodeStackPtr--];
//             RBBINode  *starNode    = pushNewNode(RBBINode::opStar);
//             if (U_FAILURE(*fRB->fStatus)) {
//                 break;
//             }
//             starNode->fLeftChild   = operandNode;
//             operandNode->fParent   = starNode;
//         }
//         break;

//     case doRuleChar:
//         // A "Rule Character" is any single character that is a literal part
//         // of the regular expression.  Like a, b and c in the expression "(abc*) | [:L:]"
//         // These are pretty uncommon in break rules; the terms are more commonly
//         //  sets.  To keep things uniform, treat these characters like as
//         // sets that just happen to contain only one character.
//         {
//             n = pushNewNode(RBBINode::setRef);
//             if (U_FAILURE(*fRB->fStatus)) {
//                 break;
//             }
//             findSetFor(UnicodeString(fC.fChar), n);
//             n->fFirstPos = fScanIndex;
//             n->fLastPos  = fNextIndex;
//             fRB->fRules.extractBetween(n->fFirstPos, n->fLastPos, n->fText);
//             break;
//         }

//     case doDotAny:
//         // scanned a ".", meaning match any single character.
//         {
//             n = pushNewNode(RBBINode::setRef);
//             if (U_FAILURE(*fRB->fStatus)) {
//                 break;
//             }
//             findSetFor(UnicodeString(TRUE, kAny, 3), n);
//             n->fFirstPos = fScanIndex;
//             n->fLastPos  = fNextIndex;
//             fRB->fRules.extractBetween(n->fFirstPos, n->fLastPos, n->fText);
//             break;
//         }

//     case doSlash:
//         // Scanned a '/', which identifies a look-ahead break position in a rule.
//         n = pushNewNode(RBBINode::lookAhead);
//         if (U_FAILURE(*fRB->fStatus)) {
//             break;
//         }
//         n->fVal      = fRuleNum;
//         n->fFirstPos = fScanIndex;
//         n->fLastPos  = fNextIndex;
//         fRB->fRules.extractBetween(n->fFirstPos, n->fLastPos, n->fText);
//         fLookAheadRule = TRUE;
//         break;


//     case doStartTagValue:
//         // Scanned a '{', the opening delimiter for a tag value within a rule.
//         n = pushNewNode(RBBINode::tag);
//         if (U_FAILURE(*fRB->fStatus)) {
//             break;
//         }
//         n->fVal      = 0;
//         n->fFirstPos = fScanIndex;
//         n->fLastPos  = fNextIndex;
//         break;

//     case doTagDigit:
//         // Just scanned a decimal digit that's part of a tag value
//         {
//             n = fNodeStack[fNodeStackPtr];
//             uint32_t v = u_charDigitValue(fC.fChar);
//             U_ASSERT(v < 10);
//             n->fVal = n->fVal*10 + v;
//             break;
//         }

//     case doTagValue:
//         n = fNodeStack[fNodeStackPtr];
//         n->fLastPos = fNextIndex;
//         fRB->fRules.extractBetween(n->fFirstPos, n->fLastPos, n->fText);
//         break;

//     case doTagExpectedError:
//         error(U_BRK_MALFORMED_RULE_TAG);
//         returnVal = FALSE;
//         break;

//     case doOptionStart:
//         // Scanning a !!option.   At the start of string.
//         fOptionStart = fScanIndex;
//         break;

//     case doOptionEnd:
//         {
//             UnicodeString opt(fRB->fRules, fOptionStart, fScanIndex-fOptionStart);
//             if (opt == UNICODE_STRING("chain", 5)) {
//                 fRB->fChainRules = TRUE;
//             } else if (opt == UNICODE_STRING("LBCMNoChain", 11)) {
//                 fRB->fLBCMNoChain = TRUE;
//             } else if (opt == UNICODE_STRING("forward", 7)) {
//                 fRB->fDefaultTree   = &fRB->fForwardTree;
//             } else if (opt == UNICODE_STRING("reverse", 7)) {
//                 fRB->fDefaultTree   = &fRB->fReverseTree;
//             } else if (opt == UNICODE_STRING("safe_forward", 12)) {
//                 fRB->fDefaultTree   = &fRB->fSafeFwdTree;
//             } else if (opt == UNICODE_STRING("safe_reverse", 12)) {
//                 fRB->fDefaultTree   = &fRB->fSafeRevTree;
//             } else if (opt == UNICODE_STRING("lookAheadHardBreak", 18)) {
//                 fRB->fLookAheadHardBreak = TRUE;
//             } else if (opt == UNICODE_STRING("quoted_literals_only", 20)) {
//                 fRuleSets[kRuleSet_rule_char-128].clear();
//             } else if (opt == UNICODE_STRING("unquoted_literals",  17)) {
//                 fRuleSets[kRuleSet_rule_char-128].applyPattern(UnicodeString(gRuleSet_rule_char_pattern), *fRB->fStatus);
//             } else {
//                 error(U_BRK_UNRECOGNIZED_OPTION);
//             }
//         }
//         break;

//     case doReverseDir:
//         fReverseRule = TRUE;
//         break;

//     case doStartVariableName:
//         n = pushNewNode(RBBINode::varRef);
//         if (U_FAILURE(*fRB->fStatus)) {
//             break;
//         }
//         n->fFirstPos = fScanIndex;
//         break;

//     case doEndVariableName:
//         n = fNodeStack[fNodeStackPtr];
//         if (n==NULL || n->fType != RBBINode::varRef) {
//             error(U_BRK_INTERNAL_ERROR);
//             break;
//         }
//         n->fLastPos = fScanIndex;
//         fRB->fRules.extractBetween(n->fFirstPos+1, n->fLastPos, n->fText);
//         // Look the newly scanned name up in the symbol table
//         //   If there's an entry, set the l. child of the var ref to the replacement expression.
//         //   (We also pass through here when scanning assignments, but no harm is done, other
//         //    than a slight wasted effort that seems hard to avoid.  Lookup will be null)
//         n->fLeftChild = fSymbolTable->lookupNode(n->fText);
//         break;

//     case doCheckVarDef:
//         n = fNodeStack[fNodeStackPtr];
//         if (n->fLeftChild == NULL) {
//             error(U_BRK_UNDEFINED_VARIABLE);
//             returnVal = FALSE;
//         }
//         break;

//     case doExprFinished:
//         break;

//     case doRuleErrorAssignExpr:
//         error(U_BRK_ASSIGN_ERROR);
//         returnVal = FALSE;
//         break;

//     case doExit:
//         returnVal = FALSE;
//         break;

//     case doScanUnicodeSet:
//         scanSet();
//         break;

//     default:
//         error(U_BRK_INTERNAL_ERROR);
//         returnVal = FALSE;
//         break;
//     }
//     return returnVal && U_SUCCESS(*fRB->fStatus);
// }




// //------------------------------------------------------------------------------
// //
// //  Error         Report a rule parse error.
// //                Only report it if no previous error has been recorded.
// //
// //------------------------------------------------------------------------------
// void RBBIRuleScanner::error(UErrorCode e) {
//     if (U_SUCCESS(*fRB->fStatus)) {
//         *fRB->fStatus = e;
//         if (fRB->fParseError) {
//             fRB->fParseError->line  = fLineNum;
//             fRB->fParseError->offset = fCharNum;
//             fRB->fParseError->preContext[0] = 0;
//             fRB->fParseError->postContext[0] = 0;
//         }
//     }
// }




// //------------------------------------------------------------------------------
// //
// //  fixOpStack   The parse stack holds partially assembled chunks of the parse tree.
// //               An entry on the stack may be as small as a single setRef node,
// //               or as large as the parse tree
// //               for an entire expression (this will be the one item left on the stack
// //               when the parsing of an RBBI rule completes.
// //
// //               This function is called when a binary operator is encountered.
// //               It looks back up the stack for operators that are not yet associated
// //               with a right operand, and if the precedence of the stacked operator >=
// //               the precedence of the current operator, binds the operand left,
// //               to the previously encountered operator.
// //
// //------------------------------------------------------------------------------
// void RBBIRuleScanner::fixOpStack(RBBINode::OpPrecedence p) {
//     RBBINode *n;
//     // printNodeStack("entering fixOpStack()");
//     for (;;) {
//         n = fNodeStack[fNodeStackPtr-1];   // an operator node
//         if (n->fPrecedence == 0) {
//             RBBIDebugPuts("RBBIRuleScanner::fixOpStack, bad operator node");
//             error(U_BRK_INTERNAL_ERROR);
//             return;
//         }

//         if (n->fPrecedence < p || n->fPrecedence <= RBBINode::precLParen) {
//             // The most recent operand goes with the current operator,
//             //   not with the previously stacked one.
//             break;
//         }
//             // Stack operator is a binary op  ( '|' or concatenation)
//             //   TOS operand becomes right child of this operator.
//             //   Resulting subexpression becomes the TOS operand.
//             n->fRightChild = fNodeStack[fNodeStackPtr];
//             fNodeStack[fNodeStackPtr]->fParent = n;
//             fNodeStackPtr--;
//         // printNodeStack("looping in fixOpStack()   ");
//     }

//     if (p <= RBBINode::precLParen) {
//         // Scan is at a right paren or end of expression.
//         //  The scanned item must match the stack, or else there was an error.
//         //  Discard the left paren (or start expr) node from the stack,
//             //  leaving the completed (sub)expression as TOS.
//             if (n->fPrecedence != p) {
//                 // Right paren encountered matched start of expression node, or
//                 // end of expression matched with a left paren node.
//                 error(U_BRK_MISMATCHED_PAREN);
//             }
//             fNodeStack[fNodeStackPtr-1] = fNodeStack[fNodeStackPtr];
//             fNodeStackPtr--;
//             // Delete the now-discarded LParen or Start node.
//             delete n;
//     }
//     // printNodeStack("leaving fixOpStack()");
// }




// //------------------------------------------------------------------------------
// //
// //   findSetFor    given a UnicodeString,
// //                  - find the corresponding Unicode Set  (uset node)
// //                         (create one if necessary)
// //                  - Set fLeftChild of the caller's node (should be a setRef node)
// //                         to the uset node
// //                 Maintain a hash table of uset nodes, so the same one is always used
// //                    for the same string.
// //                 If a "to adopt" set is provided and we haven't seen this key before,
// //                    add the provided set to the hash table.
// //                 If the string is one (32 bit) char in length, the set contains
// //                    just one element which is the char in question.
// //                 If the string is "any", return a set containing all chars.
// //
// //------------------------------------------------------------------------------
// void RBBIRuleScanner::findSetFor(const UnicodeString &s, RBBINode *node, UnicodeSet *setToAdopt) {

//     RBBISetTableEl   *el;

//     // First check whether we've already cached a set for this string.
//     // If so, just use the cached set in the new node.
//     //   delete any set provided by the caller, since we own it.
//     el = (RBBISetTableEl *)uhash_get(fSetTable, &s);
//     if (el != NULL) {
//         delete setToAdopt;
//         node->fLeftChild = el->val;
//         U_ASSERT(node->fLeftChild->fType == RBBINode::uset);
//         return;
//     }

//     // Haven't seen this set before.
//     // If the caller didn't provide us with a prebuilt set,
//     //   create a new UnicodeSet now.
//     if (setToAdopt == NULL) {
//         if (s.compare(kAny, -1) == 0) {
//             setToAdopt = new UnicodeSet(0x000000, 0x10ffff);
//         } else {
//             UChar32 c;
//             c = s.char32At(0);
//             setToAdopt = new UnicodeSet(c, c);
//         }
//     }

//     //
//     // Make a new uset node to refer to this UnicodeSet
//     // This new uset node becomes the child of the caller's setReference node.
//     //
//     RBBINode *usetNode    = new RBBINode(RBBINode::uset);
//     if (usetNode == NULL) {
//         error(U_MEMORY_ALLOCATION_ERROR);
//         return;
//     }
//     usetNode->fInputSet   = setToAdopt;
//     usetNode->fParent     = node;
//     node->fLeftChild      = usetNode;
//     usetNode->fText = s;


//     //
//     // Add the new uset node to the list of all uset nodes.
//     //
//     fRB->fUSetNodes->addElement(usetNode, *fRB->fStatus);


//     //
//     // Add the new set to the set hash table.
//     //
//     el      = (RBBISetTableEl *)uprv_malloc(sizeof(RBBISetTableEl));
//     UnicodeString *tkey = new UnicodeString(s);
//     if (tkey == NULL || el == NULL || setToAdopt == NULL) {
//         // Delete to avoid memory leak
//         delete tkey;
//         tkey = NULL;
//         uprv_free(el);
//         el = NULL;
//         delete setToAdopt;
//         setToAdopt = NULL;

//         error(U_MEMORY_ALLOCATION_ERROR);
//         return;
//     }
//     el->key = tkey;
//     el->val = usetNode;
//     uhash_put(fSetTable, el->key, el, fRB->fStatus);

//     return;
// }



// //
// //  Assorted Unicode character constants.
// //     Numeric because there is no portable way to enter them as literals.
// //     (Think EBCDIC).
// //
// static const UChar      chCR        = 0x0d;      // New lines, for terminating comments.
// static const UChar      chLF        = 0x0a;
// static const UChar      chNEL       = 0x85;      //    NEL newline variant
// static const UChar      chLS        = 0x2028;    //    Unicode Line Separator
// static const UChar      chApos      = 0x27;      //  single quote, for quoted chars.
// static const UChar      chPound     = 0x23;      // '#', introduces a comment.
// static const UChar      chBackSlash = 0x5c;      // '\'  introduces a char escape
// static const UChar      chLParen    = 0x28;
// static const UChar      chRParen    = 0x29;


// //------------------------------------------------------------------------------
// //
// //  stripRules    Return a rules string without extra spaces.
// //                (Comments are removed separately, during rule parsing.)
// //
// //------------------------------------------------------------------------------
// UnicodeString RBBIRuleScanner::stripRules(const UnicodeString &rules) {
//     UnicodeString strippedRules;
//     int32_t rulesLength = rules.length();
//     bool skippingSpaces = false;

//     for (int32_t idx=0; idx<rulesLength; idx = rules.moveIndex32(idx, 1)) {
//         UChar32 cp = rules.char32At(idx);
//         bool whiteSpace = u_hasBinaryProperty(cp, UCHAR_PATTERN_WHITE_SPACE);
//         if (skippingSpaces && whiteSpace) {
//             continue;
//         }
//         strippedRules.append(cp);
//         skippingSpaces = whiteSpace;
//     }
//     return strippedRules;
// }


// //------------------------------------------------------------------------------
// //
// //  nextCharLL    Low Level Next Char from rule input source.
// //                Get a char from the input character iterator,
// //                keep track of input position for error reporting.
// //
// //------------------------------------------------------------------------------
// UChar32  RBBIRuleScanner::nextCharLL() {
//     UChar32  ch;

//     if (fNextIndex >= fRB->fRules.length()) {
//         return (UChar32)-1;
//     }
//     ch         = fRB->fRules.char32At(fNextIndex);
//     fNextIndex = fRB->fRules.moveIndex32(fNextIndex, 1);

//     if (ch == chCR ||
//         ch == chNEL ||
//         ch == chLS   ||
//         (ch == chLF && fLastChar != chCR)) {
//         // Character is starting a new line.  Bump up the line number, and
//         //  reset the column to 0.
//         fLineNum++;
//         fCharNum=0;
//         if (fQuoteMode) {
//             error(U_BRK_NEW_LINE_IN_QUOTED_STRING);
//             fQuoteMode = FALSE;
//         }
//     }
//     else {
//         // Character is not starting a new line.  Except in the case of a
//         //   LF following a CR, increment the column position.
//         if (ch != chLF) {
//             fCharNum++;
//         }
//     }
//     fLastChar = ch;
//     return ch;
// }


// //------------------------------------------------------------------------------
// //
// //   nextChar     for rules scanning.  At this level, we handle stripping
// //                out comments and processing backslash character escapes.
// //                The rest of the rules grammar is handled at the next level up.
// //
// //------------------------------------------------------------------------------
// void RBBIRuleScanner::nextChar(RBBIRuleChar &c) {

//     // Unicode Character constants needed for the processing done by nextChar(),
//     //   in hex because literals wont work on EBCDIC machines.

//     fScanIndex = fNextIndex;
//     c.fChar    = nextCharLL();
//     c.fEscaped = FALSE;

//     //
//     //  check for '' sequence.
//     //  These are recognized in all contexts, whether in quoted text or not.
//     //
//     if (c.fChar == chApos) {
//         if (fRB->fRules.char32At(fNextIndex) == chApos) {
//             c.fChar    = nextCharLL();        // get nextChar officially so character counts
//             c.fEscaped = TRUE;                //   stay correct.
//         }
//         else
//         {
//             // Single quote, by itself.
//             //   Toggle quoting mode.
//             //   Return either '('  or ')', because quotes cause a grouping of the quoted text.
//             fQuoteMode = !fQuoteMode;
//             if (fQuoteMode == TRUE) {
//                 c.fChar = chLParen;
//             } else {
//                 c.fChar = chRParen;
//             }
//             c.fEscaped = FALSE;      // The paren that we return is not escaped.
//             return;
//         }
//     }

//     if (fQuoteMode) {
//         c.fEscaped = TRUE;
//     }
//     else
//     {
//         // We are not in a 'quoted region' of the source.
//         //
//         if (c.fChar == chPound) {
//             // Start of a comment.  Consume the rest of it.
//             //  The new-line char that terminates the comment is always returned.
//             //  It will be treated as white-space, and serves to break up anything
//             //    that might otherwise incorrectly clump together with a comment in
//             //    the middle (a variable name, for example.)
//             int32_t commentStart = fScanIndex;
//             for (;;) {
//                 c.fChar = nextCharLL();
//                 if (c.fChar == (UChar32)-1 ||  // EOF
//                     c.fChar == chCR     ||
//                     c.fChar == chLF     ||
//                     c.fChar == chNEL    ||
//                     c.fChar == chLS)       {break;}
//             }
//             for (int32_t i=commentStart; i<fNextIndex-1; ++i) {
//                 fRB->fStrippedRules.setCharAt(i, u' ');
//             }
//         }
//         if (c.fChar == (UChar32)-1) {
//             return;
//         }

//         //
//         //  check for backslash escaped characters.
//         //  Use UnicodeString::unescapeAt() to handle them.
//         //
//         if (c.fChar == chBackSlash) {
//             c.fEscaped = TRUE;
//             int32_t startX = fNextIndex;
//             c.fChar = fRB->fRules.unescapeAt(fNextIndex);
//             if (fNextIndex == startX) {
//                 error(U_BRK_HEX_DIGITS_EXPECTED);
//             }
//             fCharNum += fNextIndex-startX;
//         }
//     }
//     // putc(c.fChar, stdout);
// }

// //------------------------------------------------------------------------------
// //
// //  Parse RBBI rules.   The state machine for rules parsing is here.
// //                      The state tables are hand-written in the file rbbirpt.txt,
// //                      and converted to the form used here by a perl
// //                      script rbbicst.pl
// //
// //------------------------------------------------------------------------------
// void RBBIRuleScanner::parse() {
//     uint16_t                state;
//     const RBBIRuleTableEl  *tableEl;

//     if (U_FAILURE(*fRB->fStatus)) {
//         return;
//     }

//     state = 1;
//     nextChar(fC);
//     //
//     // Main loop for the rule parsing state machine.
//     //   Runs once per state transition.
//     //   Each time through optionally performs, depending on the state table,
//     //      - an advance to the the next input char
//     //      - an action to be performed.
//     //      - pushing or popping a state to/from the local state return stack.
//     //
//     for (;;) {
//         //  Bail out if anything has gone wrong.
//         //  RBBI rule file parsing stops on the first error encountered.
//         if (U_FAILURE(*fRB->fStatus)) {
//             break;
//         }

//         // Quit if state == 0.  This is the normal way to exit the state machine.
//         //
//         if (state == 0) {
//             break;
//         }

//         // Find the state table element that matches the input char from the rule, or the
//         //    class of the input character.  Start with the first table row for this
//         //    state, then linearly scan forward until we find a row that matches the
//         //    character.  The last row for each state always matches all characters, so
//         //    the search will stop there, if not before.
//         //
//         tableEl = &gRuleParseStateTable[state];
//         #ifdef RBBI_DEBUG
//             if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "scan")) {
//                 RBBIDebugPrintf("char, line, col = (\'%c\', %d, %d)    state=%s ",
//                     fC.fChar, fLineNum, fCharNum, RBBIRuleStateNames[state]);
//             }
//         #endif

//         for (;;) {
//             #ifdef RBBI_DEBUG
//                 if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "scan")) { RBBIDebugPrintf("."); fflush(stdout);}
//             #endif
//             if (tableEl->fCharClass < 127 && fC.fEscaped == FALSE &&   tableEl->fCharClass == fC.fChar) {
//                 // Table row specified an individual character, not a set, and
//                 //   the input character is not escaped, and
//                 //   the input character matched it.
//                 break;
//             }
//             if (tableEl->fCharClass == 255) {
//                 // Table row specified default, match anything character class.
//                 break;
//             }
//             if (tableEl->fCharClass == 254 && fC.fEscaped)  {
//                 // Table row specified "escaped" and the char was escaped.
//                 break;
//             }
//             if (tableEl->fCharClass == 253 && fC.fEscaped &&
//                 (fC.fChar == 0x50 || fC.fChar == 0x70 ))  {
//                 // Table row specified "escaped P" and the char is either 'p' or 'P'.
//                 break;
//             }
//             if (tableEl->fCharClass == 252 && fC.fChar == (UChar32)-1)  {
//                 // Table row specified eof and we hit eof on the input.
//                 break;
//             }

//             if (tableEl->fCharClass >= 128 && tableEl->fCharClass < 240 &&   // Table specs a char class &&
//                 fC.fEscaped == FALSE &&                                      //   char is not escaped &&
//                 fC.fChar != (UChar32)-1) {                                   //   char is not EOF
//                 U_ASSERT((tableEl->fCharClass-128) < UPRV_LENGTHOF(fRuleSets));
//                 if (fRuleSets[tableEl->fCharClass-128].contains(fC.fChar)) {
//                     // Table row specified a character class, or set of characters,
//                     //   and the current char matches it.
//                     break;
//                 }
//             }

//             // No match on this row, advance to the next  row for this state,
//             tableEl++;
//         }
//         if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "scan")) { RBBIDebugPuts("");}

//         //
//         // We've found the row of the state table that matches the current input
//         //   character from the rules string.
//         // Perform any action specified  by this row in the state table.
//         if (doParseActions((int32_t)tableEl->fAction) == FALSE) {
//             // Break out of the state machine loop if the
//             //   the action signalled some kind of error, or
//             //   the action was to exit, occurs on normal end-of-rules-input.
//             break;
//         }

//         if (tableEl->fPushState != 0) {
//             fStackPtr++;
//             if (fStackPtr >= kStackSize) {
//                 error(U_BRK_INTERNAL_ERROR);
//                 RBBIDebugPuts("RBBIRuleScanner::parse() - state stack overflow.");
//                 fStackPtr--;
//             }
//             fStack[fStackPtr] = tableEl->fPushState;
//         }

//         if (tableEl->fNextChar) {
//             nextChar(fC);
//         }

//         // Get the next state from the table entry, or from the
//         //   state stack if the next state was specified as "pop".
//         if (tableEl->fNextState != 255) {
//             state = tableEl->fNextState;
//         } else {
//             state = fStack[fStackPtr];
//             fStackPtr--;
//             if (fStackPtr < 0) {
//                 error(U_BRK_INTERNAL_ERROR);
//                 RBBIDebugPuts("RBBIRuleScanner::parse() - state stack underflow.");
//                 fStackPtr++;
//             }
//         }

//     }

//     if (U_FAILURE(*fRB->fStatus)) {
//         return;
//     }

//     // If there are no forward rules set an error.
//     //
//     if (fRB->fForwardTree == NULL) {
//         error(U_BRK_RULE_SYNTAX);
//         return;
//     }

//     //
//     // Parsing of the input RBBI rules is complete.
//     // We now have a parse tree for the rule expressions
//     // and a list of all UnicodeSets that are referenced.
//     //
// #ifdef RBBI_DEBUG
//     if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "symbols")) {fSymbolTable->rbbiSymtablePrint();}
//     if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "ptree")) {
//         RBBIDebugPrintf("Completed Forward Rules Parse Tree...\n");
//         RBBINode::printTree(fRB->fForwardTree, TRUE);
//         RBBIDebugPrintf("\nCompleted Reverse Rules Parse Tree...\n");
//         RBBINode::printTree(fRB->fReverseTree, TRUE);
//         RBBIDebugPrintf("\nCompleted Safe Point Forward Rules Parse Tree...\n");
//         RBBINode::printTree(fRB->fSafeFwdTree, TRUE);
//         RBBIDebugPrintf("\nCompleted Safe Point Reverse Rules Parse Tree...\n");
//         RBBINode::printTree(fRB->fSafeRevTree, TRUE);
//     }
// #endif
// }


// //------------------------------------------------------------------------------
// //
// //  printNodeStack     for debugging...
// //
// //------------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void RBBIRuleScanner::printNodeStack(const char *title) {
//     int i;
//     RBBIDebugPrintf("%s.  Dumping node stack...\n", title);
//     for (i=fNodeStackPtr; i>0; i--) {RBBINode::printTree(fNodeStack[i], TRUE);}
// }
// #endif




// //------------------------------------------------------------------------------
// //
// //  pushNewNode   create a new RBBINode of the specified type and push it
// //                onto the stack of nodes.
// //
// //------------------------------------------------------------------------------
// RBBINode  *RBBIRuleScanner::pushNewNode(RBBINode::NodeType  t) {
//     if (U_FAILURE(*fRB->fStatus)) {
//         return NULL;
//     }
//     if (fNodeStackPtr >= kStackSize - 1) {
//         error(U_BRK_RULE_SYNTAX);
//         RBBIDebugPuts("RBBIRuleScanner::pushNewNode - stack overflow.");
//         return NULL;
//     }
//     fNodeStackPtr++;
//     fNodeStack[fNodeStackPtr] = new RBBINode(t);
//     if (fNodeStack[fNodeStackPtr] == NULL) {
//         *fRB->fStatus = U_MEMORY_ALLOCATION_ERROR;
//     }
//     return fNodeStack[fNodeStackPtr];
// }



// //------------------------------------------------------------------------------
// //
// //  scanSet    Construct a UnicodeSet from the text at the current scan
// //             position.  Advance the scan position to the first character
// //             after the set.
// //
// //             A new RBBI setref node referring to the set is pushed onto the node
// //             stack.
// //
// //             The scan position is normally under the control of the state machine
// //             that controls rule parsing.  UnicodeSets, however, are parsed by
// //             the UnicodeSet constructor, not by the RBBI rule parser.
// //
// //------------------------------------------------------------------------------
// void RBBIRuleScanner::scanSet() {
//     UnicodeSet    *uset;
//     ParsePosition  pos;
//     int            startPos;
//     int            i;

//     if (U_FAILURE(*fRB->fStatus)) {
//         return;
//     }

//     pos.setIndex(fScanIndex);
//     startPos = fScanIndex;
//     UErrorCode localStatus = U_ZERO_ERROR;
//     uset = new UnicodeSet();
//     if (uset == NULL) {
//         localStatus = U_MEMORY_ALLOCATION_ERROR;
//     } else {
//         uset->applyPatternIgnoreSpace(fRB->fRules, pos, fSymbolTable, localStatus);
//     }
//     if (U_FAILURE(localStatus)) {
//         //  TODO:  Get more accurate position of the error from UnicodeSet's return info.
//         //         UnicodeSet appears to not be reporting correctly at this time.
//         #ifdef RBBI_DEBUG
//             RBBIDebugPrintf("UnicodeSet parse postion.ErrorIndex = %d\n", pos.getIndex());
//         #endif
//         error(localStatus);
//         delete uset;
//         return;
//     }

//     // Verify that the set contains at least one code point.
//     //
//     U_ASSERT(uset!=NULL);
//     if (uset->isEmpty()) {
//         // This set is empty.
//         //  Make it an error, because it almost certainly is not what the user wanted.
//         //  Also, avoids having to think about corner cases in the tree manipulation code
//         //   that occurs later on.
//         error(U_BRK_RULE_EMPTY_SET);
//         delete uset;
//         return;
//     }


//     // Advance the RBBI parse postion over the UnicodeSet pattern.
//     //   Don't just set fScanIndex because the line/char positions maintained
//     //   for error reporting would be thrown off.
//     i = pos.getIndex();
//     for (;;) {
//         if (fNextIndex >= i) {
//             break;
//         }
//         nextCharLL();
//     }

//     if (U_SUCCESS(*fRB->fStatus)) {
//         RBBINode         *n;

//         n = pushNewNode(RBBINode::setRef);
//         if (U_FAILURE(*fRB->fStatus)) {
//             return;
//         }
//         n->fFirstPos = startPos;
//         n->fLastPos  = fNextIndex;
//         fRB->fRules.extractBetween(n->fFirstPos, n->fLastPos, n->fText);
//         //  findSetFor() serves several purposes here:
//         //     - Adopts storage for the UnicodeSet, will be responsible for deleting.
//         //     - Mantains collection of all sets in use, needed later for establishing
//         //          character categories for run time engine.
//         //     - Eliminates mulitiple instances of the same set.
//         //     - Creates a new uset node if necessary (if this isn't a duplicate.)
//         findSetFor(n->fText, n, uset);
//     }

// }

// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// //
// //  rbbisetb.cpp
// //
// /*
// ***************************************************************************
// *   Copyright (C) 2002-2008 International Business Machines Corporation   *
// *   and others. All rights reserved.                                      *
// ***************************************************************************
// */
// //
// //  RBBISetBuilder   Handles processing of Unicode Sets from RBBI rules
// //                   (part of the rule building process.)
// //
// //      Starting with the rules parse tree from the scanner,
// //
// //                   -  Enumerate the set of UnicodeSets that are referenced
// //                      by the RBBI rules.
// //                   -  compute a set of non-overlapping character ranges
// //                      with all characters within a range belonging to the same
// //                      set of input uniocde sets.
// //                   -  Derive a set of non-overlapping UnicodeSet (like things)
// //                      that will correspond to columns in the state table for
// //                      the RBBI execution engine.  All characters within one
// //                      of these sets belong to the same set of the original
// //                      UnicodeSets from the user's rules.
// //                   -  construct the trie table that maps input characters
// //                      to the index of the matching non-overlapping set of set from
// //                      the previous step.
// //

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/uniset.h"
// #include "utrie2.h"
// #include "uvector.h"
// #include "uassert.h"
// #include "cmemory.h"
// #include "cstring.h"

// #include "rbbisetb.h"
// #include "rbbinode.h"

// U_NAMESPACE_BEGIN

// //------------------------------------------------------------------------
// //
// //   Constructor
// //
// //------------------------------------------------------------------------
// RBBISetBuilder::RBBISetBuilder(RBBIRuleBuilder *rb)
// {
//     fRB             = rb;
//     fStatus         = rb->fStatus;
//     fRangeList      = 0;
//     fTrie           = 0;
//     fTrieSize       = 0;
//     fGroupCount     = 0;
//     fSawBOF         = FALSE;
// }


// //------------------------------------------------------------------------
// //
// //   Destructor
// //
// //------------------------------------------------------------------------
// RBBISetBuilder::~RBBISetBuilder()
// {
//     RangeDescriptor   *nextRangeDesc;

//     // Walk through & delete the linked list of RangeDescriptors
//     for (nextRangeDesc = fRangeList; nextRangeDesc!=NULL;) {
//         RangeDescriptor *r = nextRangeDesc;
//         nextRangeDesc      = r->fNext;
//         delete r;
//     }

//     utrie2_close(fTrie);
// }




// //------------------------------------------------------------------------
// //
// //   build          Build the list of non-overlapping character ranges
// //                  from the Unicode Sets.
// //
// //------------------------------------------------------------------------
// void RBBISetBuilder::buildRanges() {
//     RBBINode        *usetNode;
//     RangeDescriptor *rlRange;

//     if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "usets")) {printSets();}

//     //
//     //  Initialize the process by creating a single range encompassing all characters
//     //  that is in no sets.
//     //
//     fRangeList                = new RangeDescriptor(*fStatus); // will check for status here
//     if (fRangeList == NULL) {
//         *fStatus = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
//     fRangeList->fStartChar    = 0;
//     fRangeList->fEndChar      = 0x10ffff;

//     if (U_FAILURE(*fStatus)) {
//         return;
//     }

//     //
//     //  Find the set of non-overlapping ranges of characters
//     //
//     int  ni;
//     for (ni=0; ; ni++) {        // Loop over each of the UnicodeSets encountered in the input rules
//         usetNode = (RBBINode *)this->fRB->fUSetNodes->elementAt(ni);
//         if (usetNode==NULL) {
//             break;
//         }

//         UnicodeSet      *inputSet             = usetNode->fInputSet;
//         int32_t          inputSetRangeCount   = inputSet->getRangeCount();
//         int              inputSetRangeIndex   = 0;
//                          rlRange              = fRangeList;

//         for (;;) {
//             if (inputSetRangeIndex >= inputSetRangeCount) {
//                 break;
//             }
//             UChar32      inputSetRangeBegin  = inputSet->getRangeStart(inputSetRangeIndex);
//             UChar32      inputSetRangeEnd    = inputSet->getRangeEnd(inputSetRangeIndex);

//             // skip over ranges from the range list that are completely
//             //   below the current range from the input unicode set.
//             while (rlRange->fEndChar < inputSetRangeBegin) {
//                 rlRange = rlRange->fNext;
//             }

//             // If the start of the range from the range list is before with
//             //   the start of the range from the unicode set, split the range list range
//             //   in two, with one part being before (wholly outside of) the unicode set
//             //   and the other containing the rest.
//             //   Then continue the loop; the post-split current range will then be skipped
//             //     over
//             if (rlRange->fStartChar < inputSetRangeBegin) {
//                 rlRange->split(inputSetRangeBegin, *fStatus);
//                 if (U_FAILURE(*fStatus)) {
//                     return;
//                 }
//                 continue;
//             }

//             // Same thing at the end of the ranges...
//             // If the end of the range from the range list doesn't coincide with
//             //   the end of the range from the unicode set, split the range list
//             //   range in two.  The first part of the split range will be
//             //   wholly inside the Unicode set.
//             if (rlRange->fEndChar > inputSetRangeEnd) {
//                 rlRange->split(inputSetRangeEnd+1, *fStatus);
//                 if (U_FAILURE(*fStatus)) {
//                     return;
//                 }
//             }

//             // The current rlRange is now entirely within the UnicodeSet range.
//             // Add this unicode set to the list of sets for this rlRange
//             if (rlRange->fIncludesSets->indexOf(usetNode) == -1) {
//                 rlRange->fIncludesSets->addElement(usetNode, *fStatus);
//                 if (U_FAILURE(*fStatus)) {
//                     return;
//                 }
//             }

//             // Advance over ranges that we are finished with.
//             if (inputSetRangeEnd == rlRange->fEndChar) {
//                 inputSetRangeIndex++;
//             }
//             rlRange = rlRange->fNext;
//         }
//     }

//     if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "range")) { printRanges();}

//     //
//     //  Group the above ranges, with each group consisting of one or more
//     //    ranges that are in exactly the same set of original UnicodeSets.
//     //    The groups are numbered, and these group numbers are the set of
//     //    input symbols recognized by the run-time state machine.
//     //
//     //    Numbering: # 0  (state table column 0) is unused.
//     //               # 1  is reserved - table column 1 is for end-of-input
//     //               # 2  is reserved - table column 2 is for beginning-in-input
//     //               # 3  is the first range list.
//     //
//     RangeDescriptor *rlSearchRange;
//     for (rlRange = fRangeList; rlRange!=0; rlRange=rlRange->fNext) {
//         for (rlSearchRange=fRangeList; rlSearchRange != rlRange; rlSearchRange=rlSearchRange->fNext) {
//             if (rlRange->fIncludesSets->equals(*rlSearchRange->fIncludesSets)) {
//                 rlRange->fNum = rlSearchRange->fNum;
//                 break;
//             }
//         }
//         if (rlRange->fNum == 0) {
//             fGroupCount ++;
//             rlRange->fNum = fGroupCount+2;
//             rlRange->setDictionaryFlag();
//             addValToSets(rlRange->fIncludesSets, fGroupCount+2);
//         }
//     }

//     // Handle input sets that contain the special string {eof}.
//     //   Column 1 of the state table is reserved for EOF on input.
//     //   Column 2 is reserved for before-the-start-input.
//     //            (This column can be optimized away later if there are no rule
//     //             references to {bof}.)
//     //   Add this column value (1 or 2) to the equivalent expression
//     //     subtree for each UnicodeSet that contains the string {eof}
//     //   Because {bof} and {eof} are not a characters in the normal sense,
//     //   they doesn't affect the computation of ranges or TRIE.
//     static const UChar eofUString[] = {0x65, 0x6f, 0x66, 0};
//     static const UChar bofUString[] = {0x62, 0x6f, 0x66, 0};

//     UnicodeString eofString(eofUString);
//     UnicodeString bofString(bofUString);
//     for (ni=0; ; ni++) {        // Loop over each of the UnicodeSets encountered in the input rules
//         usetNode = (RBBINode *)this->fRB->fUSetNodes->elementAt(ni);
//         if (usetNode==NULL) {
//             break;
//         }
//         UnicodeSet      *inputSet = usetNode->fInputSet;
//         if (inputSet->contains(eofString)) {
//             addValToSet(usetNode, 1);
//         }
//         if (inputSet->contains(bofString)) {
//             addValToSet(usetNode, 2);
//             fSawBOF = TRUE;
//         }
//     }


//     if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "rgroup")) {printRangeGroups();}
//     if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "esets")) {printSets();}
// }


// //
// // Build the Trie table for mapping UChar32 values to the corresponding
// // range group number.
// //
// void RBBISetBuilder::buildTrie() {
//     RangeDescriptor *rlRange;

//     fTrie = utrie2_open(0,       //  Initial value for all code points.
//                         0,       //  Error value for out-of-range input.
//                         fStatus);

//     for (rlRange = fRangeList; rlRange!=0 && U_SUCCESS(*fStatus); rlRange=rlRange->fNext) {
//         utrie2_setRange32(fTrie,
//                           rlRange->fStartChar,     // Range start
//                           rlRange->fEndChar,       // Range end (inclusive)
//                           rlRange->fNum,           // value for range
//                           TRUE,                    // Overwrite previously written values
//                           fStatus);
//     }
// }


// void RBBISetBuilder::mergeCategories(IntPair categories) {
//     U_ASSERT(categories.first >= 1);
//     U_ASSERT(categories.second > categories.first);
//     for (RangeDescriptor *rd = fRangeList; rd != nullptr; rd = rd->fNext) {
//         int32_t rangeNum = rd->fNum & ~DICT_BIT;
//         int32_t rangeDict = rd->fNum & DICT_BIT;
//         if (rangeNum == categories.second) {
//             rd->fNum = categories.first | rangeDict;
//         } else if (rangeNum > categories.second) {
//             rd->fNum--;
//         }
//     }
//     --fGroupCount;
// }


// //-----------------------------------------------------------------------------------
// //
// //  getTrieSize()    Return the size that will be required to serialize the Trie.
// //
// //-----------------------------------------------------------------------------------
// int32_t RBBISetBuilder::getTrieSize()  {
//     if (U_FAILURE(*fStatus)) {
//         return 0;
//     }
//     utrie2_freeze(fTrie, UTRIE2_16_VALUE_BITS, fStatus);
//     fTrieSize  = utrie2_serialize(fTrie,
//                                   NULL,                // Buffer
//                                   0,                   // Capacity
//                                   fStatus);
//     if (*fStatus == U_BUFFER_OVERFLOW_ERROR) {
//         *fStatus = U_ZERO_ERROR;
//     }
//     // RBBIDebugPrintf("Trie table size is %d\n", trieSize);
//     return fTrieSize;
// }


// //-----------------------------------------------------------------------------------
// //
// //  serializeTrie()   Put the serialized trie at the specified address.
// //                    Trust the caller to have given us enough memory.
// //                    getTrieSize() MUST be called first.
// //
// //-----------------------------------------------------------------------------------
// void RBBISetBuilder::serializeTrie(uint8_t *where) {
//     utrie2_serialize(fTrie,
//                      where,                   // Buffer
//                      fTrieSize,               // Capacity
//                      fStatus);
// }

// //------------------------------------------------------------------------
// //
// //  addValToSets     Add a runtime-mapped input value to each uset from a
// //                   list of uset nodes. (val corresponds to a state table column.)
// //                   For each of the original Unicode sets - which correspond
// //                   directly to uset nodes - a logically equivalent expression
// //                   is constructed in terms of the remapped runtime input
// //                   symbol set.  This function adds one runtime input symbol to
// //                   a list of sets.
// //
// //                   The "logically equivalent expression" is the tree for an
// //                   or-ing together of all of the symbols that go into the set.
// //
// //------------------------------------------------------------------------
// void  RBBISetBuilder::addValToSets(UVector *sets, uint32_t val) {
//     int32_t       ix;

//     for (ix=0; ix<sets->size(); ix++) {
//         RBBINode *usetNode = (RBBINode *)sets->elementAt(ix);
//         addValToSet(usetNode, val);
//     }
// }

// void  RBBISetBuilder::addValToSet(RBBINode *usetNode, uint32_t val) {
//     RBBINode *leafNode = new RBBINode(RBBINode::leafChar);
//     if (leafNode == NULL) {
//         *fStatus = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
//     leafNode->fVal = (unsigned short)val;
//     if (usetNode->fLeftChild == NULL) {
//         usetNode->fLeftChild = leafNode;
//         leafNode->fParent    = usetNode;
//     } else {
//         // There are already input symbols present for this set.
//         // Set up an OR node, with the previous stuff as the left child
//         //   and the new value as the right child.
//         RBBINode *orNode = new RBBINode(RBBINode::opOr);
//         if (orNode == NULL) {
//             *fStatus = U_MEMORY_ALLOCATION_ERROR;
//             return;
//         }
//         orNode->fLeftChild  = usetNode->fLeftChild;
//         orNode->fRightChild = leafNode;
//         orNode->fLeftChild->fParent  = orNode;
//         orNode->fRightChild->fParent = orNode;
//         usetNode->fLeftChild = orNode;
//         orNode->fParent = usetNode;
//     }
// }


// //------------------------------------------------------------------------
// //
// //   getNumCharCategories
// //
// //------------------------------------------------------------------------
// int32_t  RBBISetBuilder::getNumCharCategories() const {
//     return fGroupCount + 3;
// }


// //------------------------------------------------------------------------
// //
// //   sawBOF
// //
// //------------------------------------------------------------------------
// UBool  RBBISetBuilder::sawBOF() const {
//     return fSawBOF;
// }


// //------------------------------------------------------------------------
// //
// //   getFirstChar      Given a runtime RBBI character category, find
// //                     the first UChar32 that is in the set of chars
// //                     in the category.
// //------------------------------------------------------------------------
// UChar32  RBBISetBuilder::getFirstChar(int32_t category) const {
//     RangeDescriptor   *rlRange;
//     UChar32            retVal = (UChar32)-1;
//     for (rlRange = fRangeList; rlRange!=0; rlRange=rlRange->fNext) {
//         if (rlRange->fNum == category) {
//             retVal = rlRange->fStartChar;
//             break;
//         }
//     }
//     return retVal;
// }



// //------------------------------------------------------------------------
// //
// //   printRanges        A debugging function.
// //                      dump out all of the range definitions.
// //
// //------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void RBBISetBuilder::printRanges() {
//     RangeDescriptor       *rlRange;
//     int                    i;

//     RBBIDebugPrintf("\n\n Nonoverlapping Ranges ...\n");
//     for (rlRange = fRangeList; rlRange!=0; rlRange=rlRange->fNext) {
//         RBBIDebugPrintf("%2i  %4x-%4x  ", rlRange->fNum, rlRange->fStartChar, rlRange->fEndChar);

//         for (i=0; i<rlRange->fIncludesSets->size(); i++) {
//             RBBINode       *usetNode    = (RBBINode *)rlRange->fIncludesSets->elementAt(i);
//             UnicodeString   setName = UNICODE_STRING("anon", 4);
//             RBBINode       *setRef = usetNode->fParent;
//             if (setRef != NULL) {
//                 RBBINode *varRef = setRef->fParent;
//                 if (varRef != NULL  &&  varRef->fType == RBBINode::varRef) {
//                     setName = varRef->fText;
//                 }
//             }
//             RBBI_DEBUG_printUnicodeString(setName); RBBIDebugPrintf("  ");
//         }
//         RBBIDebugPrintf("\n");
//     }
// }
// #endif


// //------------------------------------------------------------------------
// //
// //   printRangeGroups     A debugging function.
// //                        dump out all of the range groups.
// //
// //------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void RBBISetBuilder::printRangeGroups() {
//     RangeDescriptor       *rlRange;
//     RangeDescriptor       *tRange;
//     int                    i;
//     int                    lastPrintedGroupNum = 0;

//     RBBIDebugPrintf("\nRanges grouped by Unicode Set Membership...\n");
//     for (rlRange = fRangeList; rlRange!=0; rlRange=rlRange->fNext) {
//         int groupNum = rlRange->fNum & 0xbfff;
//         if (groupNum > lastPrintedGroupNum) {
//             lastPrintedGroupNum = groupNum;
//             RBBIDebugPrintf("%2i  ", groupNum);

//             if (rlRange->fNum & DICT_BIT) { RBBIDebugPrintf(" <DICT> ");}

//             for (i=0; i<rlRange->fIncludesSets->size(); i++) {
//                 RBBINode       *usetNode    = (RBBINode *)rlRange->fIncludesSets->elementAt(i);
//                 UnicodeString   setName = UNICODE_STRING("anon", 4);
//                 RBBINode       *setRef = usetNode->fParent;
//                 if (setRef != NULL) {
//                     RBBINode *varRef = setRef->fParent;
//                     if (varRef != NULL  &&  varRef->fType == RBBINode::varRef) {
//                         setName = varRef->fText;
//                     }
//                 }
//                 RBBI_DEBUG_printUnicodeString(setName); RBBIDebugPrintf(" ");
//             }

//             i = 0;
//             for (tRange = rlRange; tRange != 0; tRange = tRange->fNext) {
//                 if (tRange->fNum == rlRange->fNum) {
//                     if (i++ % 5 == 0) {
//                         RBBIDebugPrintf("\n    ");
//                     }
//                     RBBIDebugPrintf("  %05x-%05x", tRange->fStartChar, tRange->fEndChar);
//                 }
//             }
//             RBBIDebugPrintf("\n");
//         }
//     }
//     RBBIDebugPrintf("\n");
// }
// #endif


// //------------------------------------------------------------------------
// //
// //   printSets          A debugging function.
// //                      dump out all of the set definitions.
// //
// //------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void RBBISetBuilder::printSets() {
//     int                   i;

//     RBBIDebugPrintf("\n\nUnicode Sets List\n------------------\n");
//     for (i=0; ; i++) {
//         RBBINode        *usetNode;
//         RBBINode        *setRef;
//         RBBINode        *varRef;
//         UnicodeString    setName;

//         usetNode = (RBBINode *)fRB->fUSetNodes->elementAt(i);
//         if (usetNode == NULL) {
//             break;
//         }

//         RBBIDebugPrintf("%3d    ", i);
//         setName = UNICODE_STRING("anonymous", 9);
//         setRef = usetNode->fParent;
//         if (setRef != NULL) {
//             varRef = setRef->fParent;
//             if (varRef != NULL  &&  varRef->fType == RBBINode::varRef) {
//                 setName = varRef->fText;
//             }
//         }
//         RBBI_DEBUG_printUnicodeString(setName);
//         RBBIDebugPrintf("   ");
//         RBBI_DEBUG_printUnicodeString(usetNode->fText);
//         RBBIDebugPrintf("\n");
//         if (usetNode->fLeftChild != NULL) {
//             RBBINode::printTree(usetNode->fLeftChild, TRUE);
//         }
//     }
//     RBBIDebugPrintf("\n");
// }
// #endif



// //-------------------------------------------------------------------------------------
// //
// //  RangeDescriptor copy constructor
// //
// //-------------------------------------------------------------------------------------

// RangeDescriptor::RangeDescriptor(const RangeDescriptor &other, UErrorCode &status) {
//     int  i;

//     this->fStartChar    = other.fStartChar;
//     this->fEndChar      = other.fEndChar;
//     this->fNum          = other.fNum;
//     this->fNext         = NULL;
//     UErrorCode oldstatus = status;
//     this->fIncludesSets = new UVector(status);
//     if (U_FAILURE(oldstatus)) {
//         status = oldstatus;
//     }
//     if (U_FAILURE(status)) {
//         return;
//     }
//     /* test for NULL */
//     if (this->fIncludesSets == 0) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }

//     for (i=0; i<other.fIncludesSets->size(); i++) {
//         this->fIncludesSets->addElement(other.fIncludesSets->elementAt(i), status);
//     }
// }


// //-------------------------------------------------------------------------------------
// //
// //  RangeDesriptor default constructor
// //
// //-------------------------------------------------------------------------------------
// RangeDescriptor::RangeDescriptor(UErrorCode &status) {
//     this->fStartChar    = 0;
//     this->fEndChar      = 0;
//     this->fNum          = 0;
//     this->fNext         = NULL;
//     UErrorCode oldstatus = status;
//     this->fIncludesSets = new UVector(status);
//     if (U_FAILURE(oldstatus)) {
//         status = oldstatus;
//     }
//     if (U_FAILURE(status)) {
//         return;
//     }
//     /* test for NULL */
//     if(this->fIncludesSets == 0) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }

// }


// //-------------------------------------------------------------------------------------
// //
// //  RangeDesriptor Destructor
// //
// //-------------------------------------------------------------------------------------
// RangeDescriptor::~RangeDescriptor() {
//     delete  fIncludesSets;
//     fIncludesSets = NULL;
// }

// //-------------------------------------------------------------------------------------
// //
// //  RangeDesriptor::split()
// //
// //-------------------------------------------------------------------------------------
// void RangeDescriptor::split(UChar32 where, UErrorCode &status) {
//     U_ASSERT(where>fStartChar && where<=fEndChar);
//     RangeDescriptor *nr = new RangeDescriptor(*this, status);
//     if(nr == 0) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
//     if (U_FAILURE(status)) {
//         delete nr;
//         return;
//     }
//     //  RangeDescriptor copy constructor copies all fields.
//     //  Only need to update those that are different after the split.
//     nr->fStartChar = where;
//     this->fEndChar = where-1;
//     nr->fNext      = this->fNext;
//     this->fNext    = nr;
// }


// //-------------------------------------------------------------------------------------
// //
// //   RangeDescriptor::setDictionaryFlag
// //
// //            Character Category Numbers that include characters from
// //            the original Unicode Set named "dictionary" have bit 14
// //            set to 1.  The RBBI runtime engine uses this to trigger
// //            use of the word dictionary.
// //
// //            This function looks through the Unicode Sets that it
// //            (the range) includes, and sets the bit in fNum when
// //            "dictionary" is among them.
// //
// //            TODO:  a faster way would be to find the set node for
// //                   "dictionary" just once, rather than looking it
// //                   up by name every time.
// //
// //-------------------------------------------------------------------------------------
// void RangeDescriptor::setDictionaryFlag() {
//     int i;

//     static const char16_t *dictionary = u"dictionary";
//     for (i=0; i<fIncludesSets->size(); i++) {
//         RBBINode *usetNode  = (RBBINode *)fIncludesSets->elementAt(i);
//         RBBINode *setRef = usetNode->fParent;
//         if (setRef != nullptr) {
//             RBBINode *varRef = setRef->fParent;
//             if (varRef && varRef->fType == RBBINode::varRef) {
//                 const UnicodeString *setName = &varRef->fText;
//                 if (setName->compare(dictionary, -1) == 0) {
//                     fNum |= RBBISetBuilder::DICT_BIT;
//                     break;
//                 }
//             }
//         }
//     }
// }



// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// //
// //  file:  rbbistbl.cpp    Implementation of the ICU RBBISymbolTable class
// //
// /*
// ***************************************************************************
// *   Copyright (C) 2002-2014 International Business Machines Corporation
// *   and others. All rights reserved.
// ***************************************************************************
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/unistr.h"
// #include "unicode/uniset.h"
// #include "unicode/uchar.h"
// #include "unicode/parsepos.h"

// #include "cstr.h"
// #include "rbbinode.h"
// #include "rbbirb.h"
// #include "umutex.h"


// //
// //  RBBISymbolTableEntry_deleter    Used by the UHashTable to delete the contents
// //                                  when the hash table is deleted.
// //
// U_CDECL_BEGIN
// static void U_CALLCONV RBBISymbolTableEntry_deleter(void *p) {
//     icu::RBBISymbolTableEntry *px = (icu::RBBISymbolTableEntry *)p;
//     delete px;
// }
// U_CDECL_END



// U_NAMESPACE_BEGIN

// RBBISymbolTable::RBBISymbolTable(RBBIRuleScanner *rs, const UnicodeString &rules, UErrorCode &status)
//     :fRules(rules), fRuleScanner(rs), ffffString(UChar(0xffff))
// {
//     fHashTable       = NULL;
//     fCachedSetLookup = NULL;

//     fHashTable = uhash_open(uhash_hashUnicodeString, uhash_compareUnicodeString, NULL, &status);
//     // uhash_open checks status
//     if (U_FAILURE(status)) {
//         return;
//     }
//     uhash_setValueDeleter(fHashTable, RBBISymbolTableEntry_deleter);
// }



// RBBISymbolTable::~RBBISymbolTable()
// {
//     uhash_close(fHashTable);
// }


// //
// //  RBBISymbolTable::lookup       This function from the abstract symbol table inteface
// //                                looks up a variable name and returns a UnicodeString
// //                                containing the substitution text.
// //
// //                                The variable name does NOT include the leading $.
// //
// const UnicodeString  *RBBISymbolTable::lookup(const UnicodeString& s) const
// {
//     RBBISymbolTableEntry  *el;
//     RBBINode              *varRefNode;
//     RBBINode              *exprNode;
//     RBBINode              *usetNode;
//     const UnicodeString   *retString;
//     RBBISymbolTable       *This = (RBBISymbolTable *)this;   // cast off const

//     el = (RBBISymbolTableEntry *)uhash_get(fHashTable, &s);
//     if (el == NULL) {
//         return NULL;
//     }

//     varRefNode = el->val;
//     exprNode   = varRefNode->fLeftChild;     // Root node of expression for variable
//     if (exprNode->fType == RBBINode::setRef) {
//         // The $variable refers to a single UnicodeSet
//         //   return the ffffString, which will subsequently be interpreted as a
//         //   stand-in character for the set by RBBISymbolTable::lookupMatcher()
//         usetNode = exprNode->fLeftChild;
//         This->fCachedSetLookup = usetNode->fInputSet;
//         retString = &ffffString;
//     }
//     else
//     {
//         // The variable refers to something other than just a set.
//         // return the original source string for the expression
//         retString = &exprNode->fText;
//         This->fCachedSetLookup = NULL;
//     }
//     return retString;
// }



// //
// //  RBBISymbolTable::lookupMatcher   This function from the abstract symbol table
// //                                   interface maps a single stand-in character to a
// //                                   pointer to a Unicode Set.   The Unicode Set code uses this
// //                                   mechanism to get all references to the same $variable
// //                                   name to refer to a single common Unicode Set instance.
// //
// //    This implementation cheats a little, and does not maintain a map of stand-in chars
// //    to sets.  Instead, it takes advantage of the fact that  the UnicodeSet
// //    constructor will always call this function right after calling lookup(),
// //    and we just need to remember what set to return between these two calls.
// const UnicodeFunctor *RBBISymbolTable::lookupMatcher(UChar32 ch) const
// {
//     UnicodeSet *retVal = NULL;
//     RBBISymbolTable *This = (RBBISymbolTable *)this;   // cast off const
//     if (ch == 0xffff) {
//         retVal = fCachedSetLookup;
//         This->fCachedSetLookup = 0;
//     }
//     return retVal;
// }

// //
// // RBBISymbolTable::parseReference   This function from the abstract symbol table interface
// //                                   looks for a $variable name in the source text.
// //                                   It does not look it up, only scans for it.
// //                                   It is used by the UnicodeSet parser.
// //
// //                                   This implementation is lifted pretty much verbatim
// //                                   from the rules based transliterator implementation.
// //                                   I didn't see an obvious way of sharing it.
// //
// UnicodeString   RBBISymbolTable::parseReference(const UnicodeString& text,
//                                                 ParsePosition& pos, int32_t limit) const
// {
//     int32_t start = pos.getIndex();
//     int32_t i = start;
//     UnicodeString result;
//     while (i < limit) {
//         UChar c = text.charAt(i);
//         if ((i==start && !u_isIDStart(c)) || !u_isIDPart(c)) {
//             break;
//         }
//         ++i;
//     }
//     if (i == start) { // No valid name chars
//         return result; // Indicate failure with empty string
//     }
//     pos.setIndex(i);
//     text.extractBetween(start, i, result);
//     return result;
// }



// //
// // RBBISymbolTable::lookupNode      Given a key (a variable name), return the
// //                                  corresponding RBBI Node.  If there is no entry
// //                                  in the table for this name, return NULL.
// //
// RBBINode       *RBBISymbolTable::lookupNode(const UnicodeString &key) const{

//     RBBINode             *retNode = NULL;
//     RBBISymbolTableEntry *el;

//     el = (RBBISymbolTableEntry *)uhash_get(fHashTable, &key);
//     if (el != NULL) {
//         retNode = el->val;
//     }
//     return retNode;
// }


// //
// //    RBBISymbolTable::addEntry     Add a new entry to the symbol table.
// //                                  Indicate an error if the name already exists -
// //                                    this will only occur in the case of duplicate
// //                                    variable assignments.
// //
// void            RBBISymbolTable::addEntry  (const UnicodeString &key, RBBINode *val, UErrorCode &err) {
//     RBBISymbolTableEntry *e;
//     /* test for buffer overflows */
//     if (U_FAILURE(err)) {
//         return;
//     }
//     e = (RBBISymbolTableEntry *)uhash_get(fHashTable, &key);
//     if (e != NULL) {
//         err = U_BRK_VARIABLE_REDFINITION;
//         return;
//     }

//     e = new RBBISymbolTableEntry;
//     if (e == NULL) {
//         err = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
//     e->key = key;
//     e->val = val;
//     uhash_put( fHashTable, &e->key, e, &err);
// }


// RBBISymbolTableEntry::RBBISymbolTableEntry() : UMemory(), key(), val(NULL) {}

// RBBISymbolTableEntry::~RBBISymbolTableEntry() {
//     // The "val" of a symbol table entry is a variable reference node.
//     // The l. child of the val is the rhs expression from the assignment.
//     // Unlike other node types, children of variable reference nodes are not
//     //    automatically recursively deleted.  We do it manually here.
//     delete val->fLeftChild;
//     val->fLeftChild = NULL;

//     delete  val;

//     // Note: the key UnicodeString is destructed by virtue of being in the object by value.
// }


// //
// //  RBBISymbolTable::print    Debugging function, dump out the symbol table contents.
// //
// #ifdef RBBI_DEBUG
// void RBBISymbolTable::rbbiSymtablePrint() const {
//     RBBIDebugPrintf("Variable Definitions Symbol Table\n"
//            "Name                  Node         serial  String Val\n"
//            "-------------------------------------------------------------------\n");

//     int32_t pos = UHASH_FIRST;
//     const UHashElement  *e   = NULL;
//     for (;;) {
//         e = uhash_nextElement(fHashTable,  &pos);
//         if (e == NULL ) {
//             break;
//         }
//         RBBISymbolTableEntry  *s   = (RBBISymbolTableEntry *)e->value.pointer;

//         RBBIDebugPrintf("%-19s   %8p %7d ", CStr(s->key)(), (void *)s->val, s->val->fSerialNum);
//         RBBIDebugPrintf(" %s\n", CStr(s->val->fLeftChild->fText)());
//     }

//     RBBIDebugPrintf("\nParsed Variable Definitions\n");
//     pos = -1;
//     for (;;) {
//         e = uhash_nextElement(fHashTable,  &pos);
//         if (e == NULL ) {
//             break;
//         }
//         RBBISymbolTableEntry  *s   = (RBBISymbolTableEntry *)e->value.pointer;
//         RBBIDebugPrintf("%s\n", CStr(s->key)());
//         RBBINode::printTree(s->val, TRUE);
//         RBBINode::printTree(s->val->fLeftChild, FALSE);
//         RBBIDebugPrintf("\n");
//     }
// }
// #endif





// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (c) 2002-2016, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// **********************************************************************
// */
// //
// //  rbbitblb.cpp
// //


// #include "unicode/utypes.h"

// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/unistr.h"
// #include "rbbitblb.h"
// #include "rbbirb.h"
// #include "rbbisetb.h"
// #include "rbbidata.h"
// #include "cstring.h"
// #include "uassert.h"
// #include "uvectr32.h"
// #include "cmemory.h"

// U_NAMESPACE_BEGIN

// RBBITableBuilder::RBBITableBuilder(RBBIRuleBuilder *rb, RBBINode **rootNode, UErrorCode &status) :
//         fRB(rb),
//         fTree(*rootNode),
//         fStatus(&status),
//         fDStates(nullptr),
//         fSafeTable(nullptr) {
//     if (U_FAILURE(status)) {
//         return;
//     }
//     // fDStates is UVector<RBBIStateDescriptor *>
//     fDStates = new UVector(status);
//     if (U_SUCCESS(status) && fDStates == nullptr ) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//     }
// }



// RBBITableBuilder::~RBBITableBuilder() {
//     int i;
//     for (i=0; i<fDStates->size(); i++) {
//         delete (RBBIStateDescriptor *)fDStates->elementAt(i);
//     }
//     delete fDStates;
//     delete fSafeTable;
// }


// //-----------------------------------------------------------------------------
// //
// //   RBBITableBuilder::buildForwardTable  -  This is the main function for building
// //                               the DFA state transition table from the RBBI rules parse tree.
// //
// //-----------------------------------------------------------------------------
// void  RBBITableBuilder::buildForwardTable() {

//     if (U_FAILURE(*fStatus)) {
//         return;
//     }

//     // If there were no rules, just return.  This situation can easily arise
//     //   for the reverse rules.
//     if (fTree==NULL) {
//         return;
//     }

//     //
//     // Walk through the tree, replacing any references to $variables with a copy of the
//     //   parse tree for the substition expression.
//     //
//     fTree = fTree->flattenVariables();
// #ifdef RBBI_DEBUG
//     if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "ftree")) {
//         RBBIDebugPuts("\nParse tree after flattening variable references.");
//         RBBINode::printTree(fTree, TRUE);
//     }
// #endif

//     //
//     // If the rules contained any references to {bof}
//     //   add a {bof} <cat> <former root of tree> to the
//     //   tree.  Means that all matches must start out with the
//     //   {bof} fake character.
//     //
//     if (fRB->fSetBuilder->sawBOF()) {
//         RBBINode *bofTop    = new RBBINode(RBBINode::opCat);
//         RBBINode *bofLeaf   = new RBBINode(RBBINode::leafChar);
//         // Delete and exit if memory allocation failed.
//         if (bofTop == NULL || bofLeaf == NULL) {
//             *fStatus = U_MEMORY_ALLOCATION_ERROR;
//             delete bofTop;
//             delete bofLeaf;
//             return;
//         }
//         bofTop->fLeftChild  = bofLeaf;
//         bofTop->fRightChild = fTree;
//         bofLeaf->fParent    = bofTop;
//         bofLeaf->fVal       = 2;      // Reserved value for {bof}.
//         fTree               = bofTop;
//     }

//     //
//     // Add a unique right-end marker to the expression.
//     //   Appears as a cat-node, left child being the original tree,
//     //   right child being the end marker.
//     //
//     RBBINode *cn = new RBBINode(RBBINode::opCat);
//     // Exit if memory allocation failed.
//     if (cn == NULL) {
//         *fStatus = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
//     cn->fLeftChild = fTree;
//     fTree->fParent = cn;
//     cn->fRightChild = new RBBINode(RBBINode::endMark);
//     // Delete and exit if memory allocation failed.
//     if (cn->fRightChild == NULL) {
//         *fStatus = U_MEMORY_ALLOCATION_ERROR;
//         delete cn;
//         return;
//     }
//     cn->fRightChild->fParent = cn;
//     fTree = cn;

//     //
//     //  Replace all references to UnicodeSets with the tree for the equivalent
//     //      expression.
//     //
//     fTree->flattenSets();
// #ifdef RBBI_DEBUG
//     if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "stree")) {
//         RBBIDebugPuts("\nParse tree after flattening Unicode Set references.");
//         RBBINode::printTree(fTree, TRUE);
//     }
// #endif


//     //
//     // calculate the functions nullable, firstpos, lastpos and followpos on
//     // nodes in the parse tree.
//     //    See the alogrithm description in Aho.
//     //    Understanding how this works by looking at the code alone will be
//     //       nearly impossible.
//     //
//     calcNullable(fTree);
//     calcFirstPos(fTree);
//     calcLastPos(fTree);
//     calcFollowPos(fTree);
//     if (fRB->fDebugEnv && uprv_strstr(fRB->fDebugEnv, "pos")) {
//         RBBIDebugPuts("\n");
//         printPosSets(fTree);
//     }

//     //
//     //  For "chained" rules, modify the followPos sets
//     //
//     if (fRB->fChainRules) {
//         calcChainedFollowPos(fTree);
//     }

//     //
//     //  BOF (start of input) test fixup.
//     //
//     if (fRB->fSetBuilder->sawBOF()) {
//         bofFixup();
//     }

//     //
//     // Build the DFA state transition tables.
//     //
//     buildStateTable();
//     flagAcceptingStates();
//     flagLookAheadStates();
//     flagTaggedStates();

//     //
//     // Update the global table of rule status {tag} values
//     // The rule builder has a global vector of status values that are common
//     //    for all tables.  Merge the ones from this table into the global set.
//     //
//     mergeRuleStatusVals();
// }



// //-----------------------------------------------------------------------------
// //
// //   calcNullable.    Impossible to explain succinctly.  See Aho, section 3.9
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::calcNullable(RBBINode *n) {
//     if (n == NULL) {
//         return;
//     }
//     if (n->fType == RBBINode::setRef ||
//         n->fType == RBBINode::endMark ) {
//         // These are non-empty leaf node types.
//         n->fNullable = FALSE;
//         return;
//     }

//     if (n->fType == RBBINode::lookAhead || n->fType == RBBINode::tag) {
//         // Lookahead marker node.  It's a leaf, so no recursion on children.
//         // It's nullable because it does not match any literal text from the input stream.
//         n->fNullable = TRUE;
//         return;
//     }


//     // The node is not a leaf.
//     //  Calculate nullable on its children.
//     calcNullable(n->fLeftChild);
//     calcNullable(n->fRightChild);

//     // Apply functions from table 3.40 in Aho
//     if (n->fType == RBBINode::opOr) {
//         n->fNullable = n->fLeftChild->fNullable || n->fRightChild->fNullable;
//     }
//     else if (n->fType == RBBINode::opCat) {
//         n->fNullable = n->fLeftChild->fNullable && n->fRightChild->fNullable;
//     }
//     else if (n->fType == RBBINode::opStar || n->fType == RBBINode::opQuestion) {
//         n->fNullable = TRUE;
//     }
//     else {
//         n->fNullable = FALSE;
//     }
// }




// //-----------------------------------------------------------------------------
// //
// //   calcFirstPos.    Impossible to explain succinctly.  See Aho, section 3.9
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::calcFirstPos(RBBINode *n) {
//     if (n == NULL) {
//         return;
//     }
//     if (n->fType == RBBINode::leafChar  ||
//         n->fType == RBBINode::endMark   ||
//         n->fType == RBBINode::lookAhead ||
//         n->fType == RBBINode::tag) {
//         // These are non-empty leaf node types.
//         // Note: In order to maintain the sort invariant on the set,
//         // this function should only be called on a node whose set is
//         // empty to start with.
//         n->fFirstPosSet->addElement(n, *fStatus);
//         return;
//     }

//     // The node is not a leaf.
//     //  Calculate firstPos on its children.
//     calcFirstPos(n->fLeftChild);
//     calcFirstPos(n->fRightChild);

//     // Apply functions from table 3.40 in Aho
//     if (n->fType == RBBINode::opOr) {
//         setAdd(n->fFirstPosSet, n->fLeftChild->fFirstPosSet);
//         setAdd(n->fFirstPosSet, n->fRightChild->fFirstPosSet);
//     }
//     else if (n->fType == RBBINode::opCat) {
//         setAdd(n->fFirstPosSet, n->fLeftChild->fFirstPosSet);
//         if (n->fLeftChild->fNullable) {
//             setAdd(n->fFirstPosSet, n->fRightChild->fFirstPosSet);
//         }
//     }
//     else if (n->fType == RBBINode::opStar ||
//              n->fType == RBBINode::opQuestion ||
//              n->fType == RBBINode::opPlus) {
//         setAdd(n->fFirstPosSet, n->fLeftChild->fFirstPosSet);
//     }
// }



// //-----------------------------------------------------------------------------
// //
// //   calcLastPos.    Impossible to explain succinctly.  See Aho, section 3.9
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::calcLastPos(RBBINode *n) {
//     if (n == NULL) {
//         return;
//     }
//     if (n->fType == RBBINode::leafChar  ||
//         n->fType == RBBINode::endMark   ||
//         n->fType == RBBINode::lookAhead ||
//         n->fType == RBBINode::tag) {
//         // These are non-empty leaf node types.
//         // Note: In order to maintain the sort invariant on the set,
//         // this function should only be called on a node whose set is
//         // empty to start with.
//         n->fLastPosSet->addElement(n, *fStatus);
//         return;
//     }

//     // The node is not a leaf.
//     //  Calculate lastPos on its children.
//     calcLastPos(n->fLeftChild);
//     calcLastPos(n->fRightChild);

//     // Apply functions from table 3.40 in Aho
//     if (n->fType == RBBINode::opOr) {
//         setAdd(n->fLastPosSet, n->fLeftChild->fLastPosSet);
//         setAdd(n->fLastPosSet, n->fRightChild->fLastPosSet);
//     }
//     else if (n->fType == RBBINode::opCat) {
//         setAdd(n->fLastPosSet, n->fRightChild->fLastPosSet);
//         if (n->fRightChild->fNullable) {
//             setAdd(n->fLastPosSet, n->fLeftChild->fLastPosSet);
//         }
//     }
//     else if (n->fType == RBBINode::opStar     ||
//              n->fType == RBBINode::opQuestion ||
//              n->fType == RBBINode::opPlus) {
//         setAdd(n->fLastPosSet, n->fLeftChild->fLastPosSet);
//     }
// }



// //-----------------------------------------------------------------------------
// //
// //   calcFollowPos.    Impossible to explain succinctly.  See Aho, section 3.9
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::calcFollowPos(RBBINode *n) {
//     if (n == NULL ||
//         n->fType == RBBINode::leafChar ||
//         n->fType == RBBINode::endMark) {
//         return;
//     }

//     calcFollowPos(n->fLeftChild);
//     calcFollowPos(n->fRightChild);

//     // Aho rule #1
//     if (n->fType == RBBINode::opCat) {
//         RBBINode *i;   // is 'i' in Aho's description
//         uint32_t     ix;

//         UVector *LastPosOfLeftChild = n->fLeftChild->fLastPosSet;

//         for (ix=0; ix<(uint32_t)LastPosOfLeftChild->size(); ix++) {
//             i = (RBBINode *)LastPosOfLeftChild->elementAt(ix);
//             setAdd(i->fFollowPos, n->fRightChild->fFirstPosSet);
//         }
//     }

//     // Aho rule #2
//     if (n->fType == RBBINode::opStar ||
//         n->fType == RBBINode::opPlus) {
//         RBBINode   *i;  // again, n and i are the names from Aho's description.
//         uint32_t    ix;

//         for (ix=0; ix<(uint32_t)n->fLastPosSet->size(); ix++) {
//             i = (RBBINode *)n->fLastPosSet->elementAt(ix);
//             setAdd(i->fFollowPos, n->fFirstPosSet);
//         }
//     }



// }

// //-----------------------------------------------------------------------------
// //
// //    addRuleRootNodes    Recursively walk a parse tree, adding all nodes flagged
// //                        as roots of a rule to a destination vector.
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::addRuleRootNodes(UVector *dest, RBBINode *node) {
//     if (node == NULL || U_FAILURE(*fStatus)) {
//         return;
//     }
//     if (node->fRuleRoot) {
//         dest->addElement(node, *fStatus);
//         // Note: rules cannot nest. If we found a rule start node,
//         //       no child node can also be a start node.
//         return;
//     }
//     addRuleRootNodes(dest, node->fLeftChild);
//     addRuleRootNodes(dest, node->fRightChild);
// }

// //-----------------------------------------------------------------------------
// //
// //   calcChainedFollowPos.    Modify the previously calculated followPos sets
// //                            to implement rule chaining.  NOT described by Aho
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::calcChainedFollowPos(RBBINode *tree) {

//     UVector         endMarkerNodes(*fStatus);
//     UVector         leafNodes(*fStatus);
//     int32_t         i;

//     if (U_FAILURE(*fStatus)) {
//         return;
//     }

//     // get a list of all endmarker nodes.
//     tree->findNodes(&endMarkerNodes, RBBINode::endMark, *fStatus);

//     // get a list all leaf nodes
//     tree->findNodes(&leafNodes, RBBINode::leafChar, *fStatus);
//     if (U_FAILURE(*fStatus)) {
//         return;
//     }

//     // Collect all leaf nodes that can start matches for rules
//     // with inbound chaining enabled, which is the union of the
//     // firstPosition sets from each of the rule root nodes.

//     UVector ruleRootNodes(*fStatus);
//     addRuleRootNodes(&ruleRootNodes, tree);

//     UVector matchStartNodes(*fStatus);
//     for (int j=0; j<ruleRootNodes.size(); ++j) {
//         RBBINode *node = static_cast<RBBINode *>(ruleRootNodes.elementAt(j));
//         if (node->fChainIn) {
//             setAdd(&matchStartNodes, node->fFirstPosSet);
//         }
//     }
//     if (U_FAILURE(*fStatus)) {
//         return;
//     }

//     int32_t  endNodeIx;
//     int32_t  startNodeIx;

//     for (endNodeIx=0; endNodeIx<leafNodes.size(); endNodeIx++) {
//         RBBINode *tNode   = (RBBINode *)leafNodes.elementAt(endNodeIx);
//         RBBINode *endNode = NULL;

//         // Identify leaf nodes that correspond to overall rule match positions.
//         //   These include an endMarkerNode in their followPos sets.
//         for (i=0; i<endMarkerNodes.size(); i++) {
//             if (tNode->fFollowPos->contains(endMarkerNodes.elementAt(i))) {
//                 endNode = tNode;
//                 break;
//             }
//         }
//         if (endNode == NULL) {
//             // node wasn't an end node.  Try again with the next.
//             continue;
//         }

//         // We've got a node that can end a match.

//         // Line Break Specific hack:  If this node's val correspond to the $CM char class,
//         //                            don't chain from it.
//         // TODO:  Add rule syntax for this behavior, get specifics out of here and
//         //        into the rule file.
//         if (fRB->fLBCMNoChain) {
//             UChar32 c = this->fRB->fSetBuilder->getFirstChar(endNode->fVal);
//             if (c != -1) {
//                 // c == -1 occurs with sets containing only the {eof} marker string.
//                 ULineBreak cLBProp = (ULineBreak)u_getIntPropertyValue(c, UCHAR_LINE_BREAK);
//                 if (cLBProp == U_LB_COMBINING_MARK) {
//                     continue;
//                 }
//             }
//         }


//         // Now iterate over the nodes that can start a match, looking for ones
//         //   with the same char class as our ending node.
//         RBBINode *startNode;
//         for (startNodeIx = 0; startNodeIx<matchStartNodes.size(); startNodeIx++) {
//             startNode = (RBBINode *)matchStartNodes.elementAt(startNodeIx);
//             if (startNode->fType != RBBINode::leafChar) {
//                 continue;
//             }

//             if (endNode->fVal == startNode->fVal) {
//                 // The end val (character class) of one possible match is the
//                 //   same as the start of another.

//                 // Add all nodes from the followPos of the start node to the
//                 //  followPos set of the end node, which will have the effect of
//                 //  letting matches transition from a match state at endNode
//                 //  to the second char of a match starting with startNode.
//                 setAdd(endNode->fFollowPos, startNode->fFollowPos);
//             }
//         }
//     }
// }


// //-----------------------------------------------------------------------------
// //
// //   bofFixup.    Fixup for state tables that include {bof} beginning of input testing.
// //                Do an swizzle similar to chaining, modifying the followPos set of
// //                the bofNode to include the followPos nodes from other {bot} nodes
// //                scattered through the tree.
// //
// //                This function has much in common with calcChainedFollowPos().
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::bofFixup() {

//     if (U_FAILURE(*fStatus)) {
//         return;
//     }

//     //   The parse tree looks like this ...
//     //         fTree root  --->       <cat>
//     //                               /     \       .
//     //                            <cat>   <#end node>
//     //                           /     \  .
//     //                     <bofNode>   rest
//     //                               of tree
//     //
//     //    We will be adding things to the followPos set of the <bofNode>
//     //
//     RBBINode  *bofNode = fTree->fLeftChild->fLeftChild;
//     U_ASSERT(bofNode->fType == RBBINode::leafChar);
//     U_ASSERT(bofNode->fVal == 2);

//     // Get all nodes that can be the start a match of the user-written rules
//     //  (excluding the fake bofNode)
//     //  We want the nodes that can start a match in the
//     //     part labeled "rest of tree"
//     //
//     UVector *matchStartNodes = fTree->fLeftChild->fRightChild->fFirstPosSet;

//     RBBINode *startNode;
//     int       startNodeIx;
//     for (startNodeIx = 0; startNodeIx<matchStartNodes->size(); startNodeIx++) {
//         startNode = (RBBINode *)matchStartNodes->elementAt(startNodeIx);
//         if (startNode->fType != RBBINode::leafChar) {
//             continue;
//         }

//         if (startNode->fVal == bofNode->fVal) {
//             //  We found a leaf node corresponding to a {bof} that was
//             //    explicitly written into a rule.
//             //  Add everything from the followPos set of this node to the
//             //    followPos set of the fake bofNode at the start of the tree.
//             //
//             setAdd(bofNode->fFollowPos, startNode->fFollowPos);
//         }
//     }
// }

// //-----------------------------------------------------------------------------
// //
// //   buildStateTable()    Determine the set of runtime DFA states and the
// //                        transition tables for these states, by the algorithm
// //                        of fig. 3.44 in Aho.
// //
// //                        Most of the comments are quotes of Aho's psuedo-code.
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::buildStateTable() {
//     if (U_FAILURE(*fStatus)) {
//         return;
//     }
//     RBBIStateDescriptor *failState;
//     // Set it to NULL to avoid uninitialized warning
//     RBBIStateDescriptor *initialState = NULL;
//     //
//     // Add a dummy state 0 - the stop state.  Not from Aho.
//     int      lastInputSymbol = fRB->fSetBuilder->getNumCharCategories() - 1;
//     failState = new RBBIStateDescriptor(lastInputSymbol, fStatus);
//     if (failState == NULL) {
//         *fStatus = U_MEMORY_ALLOCATION_ERROR;
//         goto ExitBuildSTdeleteall;
//     }
//     failState->fPositions = new UVector(*fStatus);
//     if (failState->fPositions == NULL) {
//         *fStatus = U_MEMORY_ALLOCATION_ERROR;
//     }
//     if (failState->fPositions == NULL || U_FAILURE(*fStatus)) {
//         goto ExitBuildSTdeleteall;
//     }
//     fDStates->addElement(failState, *fStatus);
//     if (U_FAILURE(*fStatus)) {
//         goto ExitBuildSTdeleteall;
//     }

//     // initially, the only unmarked state in Dstates is firstpos(root),
//     //       where toot is the root of the syntax tree for (r)#;
//     initialState = new RBBIStateDescriptor(lastInputSymbol, fStatus);
//     if (initialState == NULL) {
//         *fStatus = U_MEMORY_ALLOCATION_ERROR;
//     }
//     if (U_FAILURE(*fStatus)) {
//         goto ExitBuildSTdeleteall;
//     }
//     initialState->fPositions = new UVector(*fStatus);
//     if (initialState->fPositions == NULL) {
//         *fStatus = U_MEMORY_ALLOCATION_ERROR;
//     }
//     if (U_FAILURE(*fStatus)) {
//         goto ExitBuildSTdeleteall;
//     }
//     setAdd(initialState->fPositions, fTree->fFirstPosSet);
//     fDStates->addElement(initialState, *fStatus);
//     if (U_FAILURE(*fStatus)) {
//         goto ExitBuildSTdeleteall;
//     }

//     // while there is an unmarked state T in Dstates do begin
//     for (;;) {
//         RBBIStateDescriptor *T = NULL;
//         int32_t              tx;
//         for (tx=1; tx<fDStates->size(); tx++) {
//             RBBIStateDescriptor *temp;
//             temp = (RBBIStateDescriptor *)fDStates->elementAt(tx);
//             if (temp->fMarked == FALSE) {
//                 T = temp;
//                 break;
//             }
//         }
//         if (T == NULL) {
//             break;
//         }

//         // mark T;
//         T->fMarked = TRUE;

//         // for each input symbol a do begin
//         int32_t  a;
//         for (a = 1; a<=lastInputSymbol; a++) {
//             // let U be the set of positions that are in followpos(p)
//             //    for some position p in T
//             //    such that the symbol at position p is a;
//             UVector    *U = NULL;
//             RBBINode   *p;
//             int32_t     px;
//             for (px=0; px<T->fPositions->size(); px++) {
//                 p = (RBBINode *)T->fPositions->elementAt(px);
//                 if ((p->fType == RBBINode::leafChar) &&  (p->fVal == a)) {
//                     if (U == NULL) {
//                         U = new UVector(*fStatus);
//                         if (U == NULL) {
//                         	*fStatus = U_MEMORY_ALLOCATION_ERROR;
//                         	goto ExitBuildSTdeleteall;
//                         }
//                     }
//                     setAdd(U, p->fFollowPos);
//                 }
//             }

//             // if U is not empty and not in DStates then
//             int32_t  ux = 0;
//             UBool    UinDstates = FALSE;
//             if (U != NULL) {
//                 U_ASSERT(U->size() > 0);
//                 int  ix;
//                 for (ix=0; ix<fDStates->size(); ix++) {
//                     RBBIStateDescriptor *temp2;
//                     temp2 = (RBBIStateDescriptor *)fDStates->elementAt(ix);
//                     if (setEquals(U, temp2->fPositions)) {
//                         delete U;
//                         U  = temp2->fPositions;
//                         ux = ix;
//                         UinDstates = TRUE;
//                         break;
//                     }
//                 }

//                 // Add U as an unmarked state to Dstates
//                 if (!UinDstates)
//                 {
//                     RBBIStateDescriptor *newState = new RBBIStateDescriptor(lastInputSymbol, fStatus);
//                     if (newState == NULL) {
//                     	*fStatus = U_MEMORY_ALLOCATION_ERROR;
//                     }
//                     if (U_FAILURE(*fStatus)) {
//                         goto ExitBuildSTdeleteall;
//                     }
//                     newState->fPositions = U;
//                     fDStates->addElement(newState, *fStatus);
//                     if (U_FAILURE(*fStatus)) {
//                         return;
//                     }
//                     ux = fDStates->size()-1;
//                 }

//                 // Dtran[T, a] := U;
//                 T->fDtran->setElementAt(ux, a);
//             }
//         }
//     }
//     return;
//     // delete local pointers only if error occured.
// ExitBuildSTdeleteall:
//     delete initialState;
//     delete failState;
// }



// //-----------------------------------------------------------------------------
// //
// //   flagAcceptingStates    Identify accepting states.
// //                          First get a list of all of the end marker nodes.
// //                          Then, for each state s,
// //                              if s contains one of the end marker nodes in its list of tree positions then
// //                                  s is an accepting state.
// //
// //-----------------------------------------------------------------------------
// void     RBBITableBuilder::flagAcceptingStates() {
//     if (U_FAILURE(*fStatus)) {
//         return;
//     }
//     UVector     endMarkerNodes(*fStatus);
//     RBBINode    *endMarker;
//     int32_t     i;
//     int32_t     n;

//     if (U_FAILURE(*fStatus)) {
//         return;
//     }

//     fTree->findNodes(&endMarkerNodes, RBBINode::endMark, *fStatus);
//     if (U_FAILURE(*fStatus)) {
//         return;
//     }

//     for (i=0; i<endMarkerNodes.size(); i++) {
//         endMarker = (RBBINode *)endMarkerNodes.elementAt(i);
//         for (n=0; n<fDStates->size(); n++) {
//             RBBIStateDescriptor *sd = (RBBIStateDescriptor *)fDStates->elementAt(n);
//             if (sd->fPositions->indexOf(endMarker) >= 0) {
//                 // Any non-zero value for fAccepting means this is an accepting node.
//                 // The value is what will be returned to the user as the break status.
//                 // If no other value was specified, force it to -1.

//                 if (sd->fAccepting==0) {
//                     // State hasn't been marked as accepting yet.  Do it now.
//                     sd->fAccepting = endMarker->fVal;
//                     if (sd->fAccepting == 0) {
//                         sd->fAccepting = -1;
//                     }
//                 }
//                 if (sd->fAccepting==-1 && endMarker->fVal != 0) {
//                     // Both lookahead and non-lookahead accepting for this state.
//                     // Favor the look-ahead.  Expedient for line break.
//                     // TODO:  need a more elegant resolution for conflicting rules.
//                     sd->fAccepting = endMarker->fVal;
//                 }
//                 // implicit else:
//                 // if sd->fAccepting already had a value other than 0 or -1, leave it be.

//                 // If the end marker node is from a look-ahead rule, set
//                 //   the fLookAhead field for this state also.
//                 if (endMarker->fLookAheadEnd) {
//                     // TODO:  don't change value if already set?
//                     // TODO:  allow for more than one active look-ahead rule in engine.
//                     //        Make value here an index to a side array in engine?
//                     sd->fLookAhead = sd->fAccepting;
//                 }
//             }
//         }
//     }
// }


// //-----------------------------------------------------------------------------
// //
// //    flagLookAheadStates   Very similar to flagAcceptingStates, above.
// //
// //-----------------------------------------------------------------------------
// void     RBBITableBuilder::flagLookAheadStates() {
//     if (U_FAILURE(*fStatus)) {
//         return;
//     }
//     UVector     lookAheadNodes(*fStatus);
//     RBBINode    *lookAheadNode;
//     int32_t     i;
//     int32_t     n;

//     fTree->findNodes(&lookAheadNodes, RBBINode::lookAhead, *fStatus);
//     if (U_FAILURE(*fStatus)) {
//         return;
//     }
//     for (i=0; i<lookAheadNodes.size(); i++) {
//         lookAheadNode = (RBBINode *)lookAheadNodes.elementAt(i);

//         for (n=0; n<fDStates->size(); n++) {
//             RBBIStateDescriptor *sd = (RBBIStateDescriptor *)fDStates->elementAt(n);
//             if (sd->fPositions->indexOf(lookAheadNode) >= 0) {
//                 sd->fLookAhead = lookAheadNode->fVal;
//             }
//         }
//     }
// }




// //-----------------------------------------------------------------------------
// //
// //    flagTaggedStates
// //
// //-----------------------------------------------------------------------------
// void     RBBITableBuilder::flagTaggedStates() {
//     if (U_FAILURE(*fStatus)) {
//         return;
//     }
//     UVector     tagNodes(*fStatus);
//     RBBINode    *tagNode;
//     int32_t     i;
//     int32_t     n;

//     if (U_FAILURE(*fStatus)) {
//         return;
//     }
//     fTree->findNodes(&tagNodes, RBBINode::tag, *fStatus);
//     if (U_FAILURE(*fStatus)) {
//         return;
//     }
//     for (i=0; i<tagNodes.size(); i++) {                   // For each tag node t (all of 'em)
//         tagNode = (RBBINode *)tagNodes.elementAt(i);

//         for (n=0; n<fDStates->size(); n++) {              //    For each state  s (row in the state table)
//             RBBIStateDescriptor *sd = (RBBIStateDescriptor *)fDStates->elementAt(n);
//             if (sd->fPositions->indexOf(tagNode) >= 0) {  //       if  s include the tag node t
//                 sortedAdd(&sd->fTagVals, tagNode->fVal);
//             }
//         }
//     }
// }




// //-----------------------------------------------------------------------------
// //
// //  mergeRuleStatusVals
// //
// //      Update the global table of rule status {tag} values
// //      The rule builder has a global vector of status values that are common
// //      for all tables.  Merge the ones from this table into the global set.
// //
// //-----------------------------------------------------------------------------
// void  RBBITableBuilder::mergeRuleStatusVals() {
//     //
//     //  The basic outline of what happens here is this...
//     //
//     //    for each state in this state table
//     //       if the status tag list for this state is in the global statuses list
//     //           record where and
//     //           continue with the next state
//     //       else
//     //           add the tag list for this state to the global list.
//     //
//     int i;
//     int n;

//     // Pre-set a single tag of {0} into the table.
//     //   We will need this as a default, for rule sets with no explicit tagging.
//     if (fRB->fRuleStatusVals->size() == 0) {
//         fRB->fRuleStatusVals->addElement(1, *fStatus);  // Num of statuses in group
//         fRB->fRuleStatusVals->addElement((int32_t)0, *fStatus);  //   and our single status of zero
//     }

//     //    For each state
//     for (n=0; n<fDStates->size(); n++) {
//         RBBIStateDescriptor *sd = (RBBIStateDescriptor *)fDStates->elementAt(n);
//         UVector *thisStatesTagValues = sd->fTagVals;
//         if (thisStatesTagValues == NULL) {
//             // No tag values are explicitly associated with this state.
//             //   Set the default tag value.
//             sd->fTagsIdx = 0;
//             continue;
//         }

//         // There are tag(s) associated with this state.
//         //   fTagsIdx will be the index into the global tag list for this state's tag values.
//         //   Initial value of -1 flags that we haven't got it set yet.
//         sd->fTagsIdx = -1;
//         int32_t  thisTagGroupStart = 0;   // indexes into the global rule status vals list
//         int32_t  nextTagGroupStart = 0;

//         // Loop runs once per group of tags in the global list
//         while (nextTagGroupStart < fRB->fRuleStatusVals->size()) {
//             thisTagGroupStart = nextTagGroupStart;
//             nextTagGroupStart += fRB->fRuleStatusVals->elementAti(thisTagGroupStart) + 1;
//             if (thisStatesTagValues->size() != fRB->fRuleStatusVals->elementAti(thisTagGroupStart)) {
//                 // The number of tags for this state is different from
//                 //    the number of tags in this group from the global list.
//                 //    Continue with the next group from the global list.
//                 continue;
//             }
//             // The lengths match, go ahead and compare the actual tag values
//             //    between this state and the group from the global list.
//             for (i=0; i<thisStatesTagValues->size(); i++) {
//                 if (thisStatesTagValues->elementAti(i) !=
//                     fRB->fRuleStatusVals->elementAti(thisTagGroupStart + 1 + i) ) {
//                     // Mismatch.
//                     break;
//                 }
//             }

//             if (i == thisStatesTagValues->size()) {
//                 // We found a set of tag values in the global list that match
//                 //   those for this state.  Use them.
//                 sd->fTagsIdx = thisTagGroupStart;
//                 break;
//             }
//         }

//         if (sd->fTagsIdx == -1) {
//             // No suitable entry in the global tag list already.  Add one
//             sd->fTagsIdx = fRB->fRuleStatusVals->size();
//             fRB->fRuleStatusVals->addElement(thisStatesTagValues->size(), *fStatus);
//             for (i=0; i<thisStatesTagValues->size(); i++) {
//                 fRB->fRuleStatusVals->addElement(thisStatesTagValues->elementAti(i), *fStatus);
//             }
//         }
//     }
// }







// //-----------------------------------------------------------------------------
// //
// //  sortedAdd  Add a value to a vector of sorted values (ints).
// //             Do not replicate entries; if the value is already there, do not
// //                add a second one.
// //             Lazily create the vector if it does not already exist.
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::sortedAdd(UVector **vector, int32_t val) {
//     int32_t i;

//     if (*vector == NULL) {
//         *vector = new UVector(*fStatus);
//     }
//     if (*vector == NULL || U_FAILURE(*fStatus)) {
//         return;
//     }
//     UVector *vec = *vector;
//     int32_t  vSize = vec->size();
//     for (i=0; i<vSize; i++) {
//         int32_t valAtI = vec->elementAti(i);
//         if (valAtI == val) {
//             // The value is already in the vector.  Don't add it again.
//             return;
//         }
//         if (valAtI > val) {
//             break;
//         }
//     }
//     vec->insertElementAt(val, i, *fStatus);
// }



// //-----------------------------------------------------------------------------
// //
// //  setAdd     Set operation on UVector
// //             dest = dest union source
// //             Elements may only appear once and must be sorted.
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::setAdd(UVector *dest, UVector *source) {
//     int32_t destOriginalSize = dest->size();
//     int32_t sourceSize       = source->size();
//     int32_t di           = 0;
//     MaybeStackArray<void *, 16> destArray, sourceArray;  // Handle small cases without malloc
//     void **destPtr, **sourcePtr;
//     void **destLim, **sourceLim;

//     if (destOriginalSize > destArray.getCapacity()) {
//         if (destArray.resize(destOriginalSize) == NULL) {
//             return;
//         }
//     }
//     destPtr = destArray.getAlias();
//     destLim = destPtr + destOriginalSize;  // destArray.getArrayLimit()?

//     if (sourceSize > sourceArray.getCapacity()) {
//         if (sourceArray.resize(sourceSize) == NULL) {
//             return;
//         }
//     }
//     sourcePtr = sourceArray.getAlias();
//     sourceLim = sourcePtr + sourceSize;  // sourceArray.getArrayLimit()?

//     // Avoid multiple "get element" calls by getting the contents into arrays
//     (void) dest->toArray(destPtr);
//     (void) source->toArray(sourcePtr);

//     dest->setSize(sourceSize+destOriginalSize, *fStatus);

//     while (sourcePtr < sourceLim && destPtr < destLim) {
//         if (*destPtr == *sourcePtr) {
//             dest->setElementAt(*sourcePtr++, di++);
//             destPtr++;
//         }
//         // This check is required for machines with segmented memory, like i5/OS.
//         // Direct pointer comparison is not recommended.
//         else if (uprv_memcmp(destPtr, sourcePtr, sizeof(void *)) < 0) {
//             dest->setElementAt(*destPtr++, di++);
//         }
//         else { /* *sourcePtr < *destPtr */
//             dest->setElementAt(*sourcePtr++, di++);
//         }
//     }

//     // At most one of these two cleanup loops will execute
//     while (destPtr < destLim) {
//         dest->setElementAt(*destPtr++, di++);
//     }
//     while (sourcePtr < sourceLim) {
//         dest->setElementAt(*sourcePtr++, di++);
//     }

//     dest->setSize(di, *fStatus);
// }



// //-----------------------------------------------------------------------------
// //
// //  setEqual    Set operation on UVector.
// //              Compare for equality.
// //              Elements must be sorted.
// //
// //-----------------------------------------------------------------------------
// UBool RBBITableBuilder::setEquals(UVector *a, UVector *b) {
//     return a->equals(*b);
// }


// //-----------------------------------------------------------------------------
// //
// //  printPosSets   Debug function.  Dump Nullable, firstpos, lastpos and followpos
// //                 for each node in the tree.
// //
// //-----------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void RBBITableBuilder::printPosSets(RBBINode *n) {
//     if (n==NULL) {
//         return;
//     }
//     printf("\n");
//     RBBINode::printNodeHeader();
//     RBBINode::printNode(n);
//     RBBIDebugPrintf("         Nullable:  %s\n", n->fNullable?"TRUE":"FALSE");

//     RBBIDebugPrintf("         firstpos:  ");
//     printSet(n->fFirstPosSet);

//     RBBIDebugPrintf("         lastpos:   ");
//     printSet(n->fLastPosSet);

//     RBBIDebugPrintf("         followpos: ");
//     printSet(n->fFollowPos);

//     printPosSets(n->fLeftChild);
//     printPosSets(n->fRightChild);
// }
// #endif

// //
// //    findDuplCharClassFrom()
// //
// bool RBBITableBuilder::findDuplCharClassFrom(IntPair *categories) {
//     int32_t numStates = fDStates->size();
//     int32_t numCols = fRB->fSetBuilder->getNumCharCategories();

//     for (; categories->first < numCols-1; categories->first++) {
//         for (categories->second=categories->first+1; categories->second < numCols; categories->second++) {
//             // Initialized to different values to prevent returning true if numStates = 0 (implies no duplicates).
//             uint16_t table_base = 0;
//             uint16_t table_dupl = 1;
//             for (int32_t state=0; state<numStates; state++) {
//                 RBBIStateDescriptor *sd = (RBBIStateDescriptor *)fDStates->elementAt(state);
//                 table_base = (uint16_t)sd->fDtran->elementAti(categories->first);
//                 table_dupl = (uint16_t)sd->fDtran->elementAti(categories->second);
//                 if (table_base != table_dupl) {
//                     break;
//                 }
//             }
//             if (table_base == table_dupl) {
//                 return true;
//             }
//         }
//     }
//     return false;
// }


// //
// //    removeColumn()
// //
// void RBBITableBuilder::removeColumn(int32_t column) {
//     int32_t numStates = fDStates->size();
//     for (int32_t state=0; state<numStates; state++) {
//         RBBIStateDescriptor *sd = (RBBIStateDescriptor *)fDStates->elementAt(state);
//         U_ASSERT(column < sd->fDtran->size());
//         sd->fDtran->removeElementAt(column);
//     }
// }

// /*
//  * findDuplicateState
//  */
// bool RBBITableBuilder::findDuplicateState(IntPair *states) {
//     int32_t numStates = fDStates->size();
//     int32_t numCols = fRB->fSetBuilder->getNumCharCategories();

//     for (; states->first<numStates-1; states->first++) {
//         RBBIStateDescriptor *firstSD = (RBBIStateDescriptor *)fDStates->elementAt(states->first);
//         for (states->second=states->first+1; states->second<numStates; states->second++) {
//             RBBIStateDescriptor *duplSD = (RBBIStateDescriptor *)fDStates->elementAt(states->second);
//             if (firstSD->fAccepting != duplSD->fAccepting ||
//                 firstSD->fLookAhead != duplSD->fLookAhead ||
//                 firstSD->fTagsIdx   != duplSD->fTagsIdx) {
//                 continue;
//             }
//             bool rowsMatch = true;
//             for (int32_t col=0; col < numCols; ++col) {
//                 int32_t firstVal = firstSD->fDtran->elementAti(col);
//                 int32_t duplVal = duplSD->fDtran->elementAti(col);
//                 if (!((firstVal == duplVal) ||
//                         ((firstVal == states->first || firstVal == states->second) &&
//                         (duplVal  == states->first || duplVal  == states->second)))) {
//                     rowsMatch = false;
//                     break;
//                 }
//             }
//             if (rowsMatch) {
//                 return true;
//             }
//         }
//     }
//     return false;
// }


// bool RBBITableBuilder::findDuplicateSafeState(IntPair *states) {
//     int32_t numStates = fSafeTable->size();

//     for (; states->first<numStates-1; states->first++) {
//         UnicodeString *firstRow = static_cast<UnicodeString *>(fSafeTable->elementAt(states->first));
//         for (states->second=states->first+1; states->second<numStates; states->second++) {
//             UnicodeString *duplRow = static_cast<UnicodeString *>(fSafeTable->elementAt(states->second));
//             bool rowsMatch = true;
//             int32_t numCols = firstRow->length();
//             for (int32_t col=0; col < numCols; ++col) {
//                 int32_t firstVal = firstRow->charAt(col);
//                 int32_t duplVal = duplRow->charAt(col);
//                 if (!((firstVal == duplVal) ||
//                         ((firstVal == states->first || firstVal == states->second) &&
//                         (duplVal  == states->first || duplVal  == states->second)))) {
//                     rowsMatch = false;
//                     break;
//                 }
//             }
//             if (rowsMatch) {
//                 return true;
//             }
//         }
//     }
//     return false;
// }


// void RBBITableBuilder::removeState(IntPair duplStates) {
//     const int32_t keepState = duplStates.first;
//     const int32_t duplState = duplStates.second;
//     U_ASSERT(keepState < duplState);
//     U_ASSERT(duplState < fDStates->size());

//     RBBIStateDescriptor *duplSD = (RBBIStateDescriptor *)fDStates->elementAt(duplState);
//     fDStates->removeElementAt(duplState);
//     delete duplSD;

//     int32_t numStates = fDStates->size();
//     int32_t numCols = fRB->fSetBuilder->getNumCharCategories();
//     for (int32_t state=0; state<numStates; ++state) {
//         RBBIStateDescriptor *sd = (RBBIStateDescriptor *)fDStates->elementAt(state);
//         for (int32_t col=0; col<numCols; col++) {
//             int32_t existingVal = sd->fDtran->elementAti(col);
//             int32_t newVal = existingVal;
//             if (existingVal == duplState) {
//                 newVal = keepState;
//             } else if (existingVal > duplState) {
//                 newVal = existingVal - 1;
//             }
//             sd->fDtran->setElementAt(newVal, col);
//         }
//         if (sd->fAccepting == duplState) {
//             sd->fAccepting = keepState;
//         } else if (sd->fAccepting > duplState) {
//             sd->fAccepting--;
//         }
//         if (sd->fLookAhead == duplState) {
//             sd->fLookAhead = keepState;
//         } else if (sd->fLookAhead > duplState) {
//             sd->fLookAhead--;
//         }
//     }
// }

// void RBBITableBuilder::removeSafeState(IntPair duplStates) {
//     const int32_t keepState = duplStates.first;
//     const int32_t duplState = duplStates.second;
//     U_ASSERT(keepState < duplState);
//     U_ASSERT(duplState < fSafeTable->size());

//     fSafeTable->removeElementAt(duplState);   // Note that fSafeTable has a deleter function
//                                               // and will auto-delete the removed element.
//     int32_t numStates = fSafeTable->size();
//     for (int32_t state=0; state<numStates; ++state) {
//         UnicodeString *sd = (UnicodeString *)fSafeTable->elementAt(state);
//         int32_t numCols = sd->length();
//         for (int32_t col=0; col<numCols; col++) {
//             int32_t existingVal = sd->charAt(col);
//             int32_t newVal = existingVal;
//             if (existingVal == duplState) {
//                 newVal = keepState;
//             } else if (existingVal > duplState) {
//                 newVal = existingVal - 1;
//             }
//             sd->setCharAt(col, static_cast<char16_t>(newVal));
//         }
//     }
// }


// /*
//  * RemoveDuplicateStates
//  */
// int32_t RBBITableBuilder::removeDuplicateStates() {
//     IntPair dupls = {3, 0};
//     int32_t numStatesRemoved = 0;

//     while (findDuplicateState(&dupls)) {
//         // printf("Removing duplicate states (%d, %d)\n", dupls.first, dupls.second);
//         removeState(dupls);
//         ++numStatesRemoved;
//     }
//     return numStatesRemoved;
// }


// //-----------------------------------------------------------------------------
// //
// //   getTableSize()    Calculate the size of the runtime form of this
// //                     state transition table.
// //
// //-----------------------------------------------------------------------------
// int32_t  RBBITableBuilder::getTableSize() const {
//     int32_t    size = 0;
//     int32_t    numRows;
//     int32_t    numCols;
//     int32_t    rowSize;

//     if (fTree == NULL) {
//         return 0;
//     }

//     size    = offsetof(RBBIStateTable, fTableData);    // The header, with no rows to the table.

//     numRows = fDStates->size();
//     numCols = fRB->fSetBuilder->getNumCharCategories();

//     rowSize = offsetof(RBBIStateTableRow, fNextState) + sizeof(uint16_t)*numCols;
//     size   += numRows * rowSize;
//     return size;
// }


// //-----------------------------------------------------------------------------
// //
// //   exportTable()    export the state transition table in the format required
// //                    by the runtime engine.  getTableSize() bytes of memory
// //                    must be available at the output address "where".
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::exportTable(void *where) {
//     RBBIStateTable    *table = (RBBIStateTable *)where;
//     uint32_t           state;
//     int                col;

//     if (U_FAILURE(*fStatus) || fTree == NULL) {
//         return;
//     }

//     int32_t catCount = fRB->fSetBuilder->getNumCharCategories();
//     if (catCount > 0x7fff ||
//         fDStates->size() > 0x7fff) {
//         *fStatus = U_BRK_INTERNAL_ERROR;
//         return;
//     }

//     table->fRowLen    = offsetof(RBBIStateTableRow, fNextState) + sizeof(uint16_t) * catCount;
//     table->fNumStates = fDStates->size();
//     table->fFlags     = 0;
//     if (fRB->fLookAheadHardBreak) {
//         table->fFlags  |= RBBI_LOOKAHEAD_HARD_BREAK;
//     }
//     if (fRB->fSetBuilder->sawBOF()) {
//         table->fFlags  |= RBBI_BOF_REQUIRED;
//     }
//     table->fReserved  = 0;

//     for (state=0; state<table->fNumStates; state++) {
//         RBBIStateDescriptor *sd = (RBBIStateDescriptor *)fDStates->elementAt(state);
//         RBBIStateTableRow   *row = (RBBIStateTableRow *)(table->fTableData + state*table->fRowLen);
//         U_ASSERT (-32768 < sd->fAccepting && sd->fAccepting <= 32767);
//         U_ASSERT (-32768 < sd->fLookAhead && sd->fLookAhead <= 32767);
//         row->fAccepting = (int16_t)sd->fAccepting;
//         row->fLookAhead = (int16_t)sd->fLookAhead;
//         row->fTagIdx    = (int16_t)sd->fTagsIdx;
//         for (col=0; col<catCount; col++) {
//             row->fNextState[col] = (uint16_t)sd->fDtran->elementAti(col);
//         }
//     }
// }


// /**
//  *   Synthesize a safe state table from the main state table.
//  */
// void RBBITableBuilder::buildSafeReverseTable(UErrorCode &status) {
//     // The safe table creation has three steps:

//     // 1. Identifiy pairs of character classes that are "safe." Safe means that boundaries
//     // following the pair do not depend on context or state before the pair. To test
//     // whether a pair is safe, run it through the main forward state table, starting
//     // from each state. If the the final state is the same, no matter what the starting state,
//     // the pair is safe.
//     //
//     // 2. Build a state table that recognizes the safe pairs. It's similar to their
//     // forward table, with a column for each input character [class], and a row for
//     // each state. Row 1 is the start state, and row 0 is the stop state. Initially
//     // create an additional state for each input character category; being in
//     // one of these states means that the character has been seen, and is potentially
//     // the first of a pair. In each of these rows, the entry for the second character
//     // of a safe pair is set to the stop state (0), indicating that a match was found.
//     // All other table entries are set to the state corresponding the current input
//     // character, allowing that charcter to be the of a start following pair.
//     //
//     // Because the safe rules are to be run in reverse, moving backwards in the text,
//     // the first and second pair categories are swapped when building the table.
//     //
//     // 3. Compress the table. There are typically many rows (states) that are
//     // equivalent - that have zeroes (match completed) in the same columns -
//     // and can be folded together.

//     // Each safe pair is stored as two UChars in the safePair string.
//     UnicodeString safePairs;

//     int32_t numCharClasses = fRB->fSetBuilder->getNumCharCategories();
//     int32_t numStates = fDStates->size();

//     for (int32_t c1=0; c1<numCharClasses; ++c1) {
//         for (int32_t c2=0; c2 < numCharClasses; ++c2) {
//             int32_t wantedEndState = -1;
//             int32_t endState = 0;
//             for (int32_t startState = 1; startState < numStates; ++startState) {
//                 RBBIStateDescriptor *startStateD = static_cast<RBBIStateDescriptor *>(fDStates->elementAt(startState));
//                 int32_t s2 = startStateD->fDtran->elementAti(c1);
//                 RBBIStateDescriptor *s2StateD = static_cast<RBBIStateDescriptor *>(fDStates->elementAt(s2));
//                 endState = s2StateD->fDtran->elementAti(c2);
//                 if (wantedEndState < 0) {
//                     wantedEndState = endState;
//                 } else {
//                     if (wantedEndState != endState) {
//                         break;
//                     }
//                 }
//             }
//             if (wantedEndState == endState) {
//                 safePairs.append((char16_t)c1);
//                 safePairs.append((char16_t)c2);
//                 // printf("(%d, %d) ", c1, c2);
//             }
//         }
//         // printf("\n");
//     }

//     // Populate the initial safe table.
//     // The table as a whole is UVector<UnicodeString>
//     // Each row is represented by a UnicodeString, being used as a Vector<int16>.
//     // Row 0 is the stop state.
//     // Row 1 is the start sate.
//     // Row 2 and beyond are other states, initially one per char class, but
//     //   after initial construction, many of the states will be combined, compacting the table.
//     // The String holds the nextState data only. The four leading fields of a row, fAccepting,
//     // fLookAhead, etc. are not needed for the safe table, and are omitted at this stage of building.

//     U_ASSERT(fSafeTable == nullptr);
//     fSafeTable = new UVector(uprv_deleteUObject, uhash_compareUnicodeString, numCharClasses + 2, status);
//     for (int32_t row=0; row<numCharClasses + 2; ++row) {
//         fSafeTable->addElement(new UnicodeString(numCharClasses, 0, numCharClasses+4), status);
//     }

//     // From the start state, each input char class transitions to the state for that input.
//     UnicodeString &startState = *static_cast<UnicodeString *>(fSafeTable->elementAt(1));
//     for (int32_t charClass=0; charClass < numCharClasses; ++charClass) {
//         // Note: +2 for the start & stop state.
//         startState.setCharAt(charClass, static_cast<char16_t>(charClass+2));
//     }

//     // Initially make every other state table row look like the start state row,
//     for (int32_t row=2; row<numCharClasses+2; ++row) {
//         UnicodeString &rowState = *static_cast<UnicodeString *>(fSafeTable->elementAt(row));
//         rowState = startState;   // UnicodeString assignment, copies contents.
//     }

//     // Run through the safe pairs, set the next state to zero when pair has been seen.
//     // Zero being the stop state, meaning we found a safe point.
//     for (int32_t pairIdx=0; pairIdx<safePairs.length(); pairIdx+=2) {
//         int32_t c1 = safePairs.charAt(pairIdx);
//         int32_t c2 = safePairs.charAt(pairIdx + 1);

//         UnicodeString &rowState = *static_cast<UnicodeString *>(fSafeTable->elementAt(c2 + 2));
//         rowState.setCharAt(c1, 0);
//     }

//     // Remove duplicate or redundant rows from the table.
//     IntPair states = {1, 0};
//     while (findDuplicateSafeState(&states)) {
//         // printf("Removing duplicate safe states (%d, %d)\n", states.first, states.second);
//         removeSafeState(states);
//     }
// }


// //-----------------------------------------------------------------------------
// //
// //   getSafeTableSize()    Calculate the size of the runtime form of this
// //                         safe state table.
// //
// //-----------------------------------------------------------------------------
// int32_t  RBBITableBuilder::getSafeTableSize() const {
//     int32_t    size = 0;
//     int32_t    numRows;
//     int32_t    numCols;
//     int32_t    rowSize;

//     if (fSafeTable == nullptr) {
//         return 0;
//     }

//     size    = offsetof(RBBIStateTable, fTableData);    // The header, with no rows to the table.

//     numRows = fSafeTable->size();
//     numCols = fRB->fSetBuilder->getNumCharCategories();

//     rowSize = offsetof(RBBIStateTableRow, fNextState) + sizeof(uint16_t)*numCols;
//     size   += numRows * rowSize;
//     return size;
// }


// //-----------------------------------------------------------------------------
// //
// //   exportSafeTable()   export the state transition table in the format required
// //                       by the runtime engine.  getTableSize() bytes of memory
// //                       must be available at the output address "where".
// //
// //-----------------------------------------------------------------------------
// void RBBITableBuilder::exportSafeTable(void *where) {
//     RBBIStateTable    *table = (RBBIStateTable *)where;
//     uint32_t           state;
//     int                col;

//     if (U_FAILURE(*fStatus) || fSafeTable == nullptr) {
//         return;
//     }

//     int32_t catCount = fRB->fSetBuilder->getNumCharCategories();
//     if (catCount > 0x7fff ||
//             fSafeTable->size() > 0x7fff) {
//         *fStatus = U_BRK_INTERNAL_ERROR;
//         return;
//     }

//     table->fRowLen    = offsetof(RBBIStateTableRow, fNextState) + sizeof(uint16_t) * catCount;
//     table->fNumStates = fSafeTable->size();
//     table->fFlags     = 0;
//     table->fReserved  = 0;

//     for (state=0; state<table->fNumStates; state++) {
//         UnicodeString *rowString = (UnicodeString *)fSafeTable->elementAt(state);
//         RBBIStateTableRow   *row = (RBBIStateTableRow *)(table->fTableData + state*table->fRowLen);
//         row->fAccepting = 0;
//         row->fLookAhead = 0;
//         row->fTagIdx    = 0;
//         row->fReserved  = 0;
//         for (col=0; col<catCount; col++) {
//             row->fNextState[col] = rowString->charAt(col);
//         }
//     }
// }




// //-----------------------------------------------------------------------------
// //
// //   printSet    Debug function.   Print the contents of a UVector
// //
// //-----------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void RBBITableBuilder::printSet(UVector *s) {
//     int32_t  i;
//     for (i=0; i<s->size(); i++) {
//         const RBBINode *v = static_cast<const RBBINode *>(s->elementAt(i));
//         RBBIDebugPrintf("%5d", v==NULL? -1 : v->fSerialNum);
//     }
//     RBBIDebugPrintf("\n");
// }
// #endif


// //-----------------------------------------------------------------------------
// //
// //   printStates    Debug Function.  Dump the fully constructed state transition table.
// //
// //-----------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void RBBITableBuilder::printStates() {
//     int     c;    // input "character"
//     int     n;    // state number

//     RBBIDebugPrintf("state |           i n p u t     s y m b o l s \n");
//     RBBIDebugPrintf("      | Acc  LA    Tag");
//     for (c=0; c<fRB->fSetBuilder->getNumCharCategories(); c++) {
//         RBBIDebugPrintf(" %2d", c);
//     }
//     RBBIDebugPrintf("\n");
//     RBBIDebugPrintf("      |---------------");
//     for (c=0; c<fRB->fSetBuilder->getNumCharCategories(); c++) {
//         RBBIDebugPrintf("---");
//     }
//     RBBIDebugPrintf("\n");

//     for (n=0; n<fDStates->size(); n++) {
//         RBBIStateDescriptor *sd = (RBBIStateDescriptor *)fDStates->elementAt(n);
//         RBBIDebugPrintf("  %3d | " , n);
//         RBBIDebugPrintf("%3d %3d %5d ", sd->fAccepting, sd->fLookAhead, sd->fTagsIdx);
//         for (c=0; c<fRB->fSetBuilder->getNumCharCategories(); c++) {
//             RBBIDebugPrintf(" %2d", sd->fDtran->elementAti(c));
//         }
//         RBBIDebugPrintf("\n");
//     }
//     RBBIDebugPrintf("\n\n");
// }
// #endif


// //-----------------------------------------------------------------------------
// //
// //   printSafeTable    Debug Function.  Dump the fully constructed safe table.
// //
// //-----------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void RBBITableBuilder::printReverseTable() {
//     int     c;    // input "character"
//     int     n;    // state number

//     RBBIDebugPrintf("    Safe Reverse Table \n");
//     if (fSafeTable == nullptr) {
//         RBBIDebugPrintf("   --- nullptr ---\n");
//         return;
//     }
//     RBBIDebugPrintf("state |           i n p u t     s y m b o l s \n");
//     RBBIDebugPrintf("      | Acc  LA    Tag");
//     for (c=0; c<fRB->fSetBuilder->getNumCharCategories(); c++) {
//         RBBIDebugPrintf(" %2d", c);
//     }
//     RBBIDebugPrintf("\n");
//     RBBIDebugPrintf("      |---------------");
//     for (c=0; c<fRB->fSetBuilder->getNumCharCategories(); c++) {
//         RBBIDebugPrintf("---");
//     }
//     RBBIDebugPrintf("\n");

//     for (n=0; n<fSafeTable->size(); n++) {
//         UnicodeString *rowString = (UnicodeString *)fSafeTable->elementAt(n);
//         RBBIDebugPrintf("  %3d | " , n);
//         RBBIDebugPrintf("%3d %3d %5d ", 0, 0, 0);  // Accepting, LookAhead, Tags
//         for (c=0; c<fRB->fSetBuilder->getNumCharCategories(); c++) {
//             RBBIDebugPrintf(" %2d", rowString->charAt(c));
//         }
//         RBBIDebugPrintf("\n");
//     }
//     RBBIDebugPrintf("\n\n");
// }
// #endif



// //-----------------------------------------------------------------------------
// //
// //   printRuleStatusTable    Debug Function.  Dump the common rule status table
// //
// //-----------------------------------------------------------------------------
// #ifdef RBBI_DEBUG
// void RBBITableBuilder::printRuleStatusTable() {
//     int32_t  thisRecord = 0;
//     int32_t  nextRecord = 0;
//     int      i;
//     UVector  *tbl = fRB->fRuleStatusVals;

//     RBBIDebugPrintf("index |  tags \n");
//     RBBIDebugPrintf("-------------------\n");

//     while (nextRecord < tbl->size()) {
//         thisRecord = nextRecord;
//         nextRecord = thisRecord + tbl->elementAti(thisRecord) + 1;
//         RBBIDebugPrintf("%4d   ", thisRecord);
//         for (i=thisRecord+1; i<nextRecord; i++) {
//             RBBIDebugPrintf("  %5d", tbl->elementAti(i));
//         }
//         RBBIDebugPrintf("\n");
//     }
//     RBBIDebugPrintf("\n\n");
// }
// #endif


// //-----------------------------------------------------------------------------
// //
// //   RBBIStateDescriptor     Methods.  This is a very struct-like class
// //                           Most access is directly to the fields.
// //
// //-----------------------------------------------------------------------------

// RBBIStateDescriptor::RBBIStateDescriptor(int lastInputSymbol, UErrorCode *fStatus) {
//     fMarked    = FALSE;
//     fAccepting = 0;
//     fLookAhead = 0;
//     fTagsIdx   = 0;
//     fTagVals   = NULL;
//     fPositions = NULL;
//     fDtran     = NULL;

//     fDtran     = new UVector32(lastInputSymbol+1, *fStatus);
//     if (U_FAILURE(*fStatus)) {
//         return;
//     }
//     if (fDtran == NULL) {
//         *fStatus = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
//     fDtran->setSize(lastInputSymbol+1);    // fDtran needs to be pre-sized.
//                                            //   It is indexed by input symbols, and will
//                                            //   hold  the next state number for each
//                                            //   symbol.
// }


// RBBIStateDescriptor::~RBBIStateDescriptor() {
//     delete       fPositions;
//     delete       fDtran;
//     delete       fTagVals;
//     fPositions = NULL;
//     fDtran     = NULL;
//     fTagVals   = NULL;
// }

// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1997-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*
* File resbund.cpp
*
* Modification History:
*
*   Date        Name        Description
*   02/05/97    aliu        Fixed bug in chopLocale.  Added scanForLocaleInFile
*                           based on code taken from scanForLocale.  Added
*                           constructor which attempts to read resource bundle
*                           from a specific file, without searching other files.
*   02/11/97    aliu        Added UErrorCode return values to constructors. Fixed
*                           infinite loops in scanForFile and scanForLocale.
*                           Modified getRawResourceData to not delete storage in
*                           localeData and resourceData which it doesn't own.
*                           Added Mac compatibility #ifdefs for tellp() and
*                           ios::nocreate.
*   03/04/97    aliu        Modified to use ExpandingDataSink objects instead of
*                           the highly inefficient ostrstream objects.
*   03/13/97    aliu        Rewrote to load in entire resource bundle and store
*                           it as a Hashtable of ResourceBundleData objects.
*                           Added state table to govern parsing of files.
*                           Modified to load locale index out of new file distinct
*                           from default.txt.
*   03/25/97    aliu        Modified to support 2-d arrays, needed for timezone data.
*                           Added support for custom file suffixes.  Again, needed
*                           to support timezone data.  Improved error handling to
*                           detect duplicate tags and subtags.
*   04/07/97    aliu        Fixed bug in getHashtableForLocale().  Fixed handling
*                           of failing UErrorCode values on entry to API methods.
*                           Fixed bugs in getArrayItem() for negative indices.
*   04/29/97    aliu        Update to use new Hashtable deletion protocol.
*   05/06/97    aliu        Flattened kTransitionTable for HP compiler.
*                           Fixed usage of CharString.
* 06/11/99      stephen     Removed parsing of .txt files.
*                           Reworked to use new binary format.
*                           Cleaned up.
* 06/14/99      stephen     Removed methods taking a filename suffix.
* 06/22/99      stephen     Added missing T_FileStream_close in parse()
* 11/09/99      weiv        Added getLocale(), rewritten constructForLocale()
* March 2000    weiv        complete overhaul.
******************************************************************************
*/


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1996-2013, International Business Machines Corporation
*   and others.  All Rights Reserved.
*
******************************************************************************
*
* File resbund.h
*
*   CREATED BY
*       Richard Gillam
*
* Modification History:
*
*   Date        Name        Description
*   2/5/97      aliu        Added scanForLocaleInFile.  Added
*                           constructor which attempts to read resource bundle
*                           from a specific file, without searching other files.
*   2/11/97     aliu        Added UErrorCode return values to constructors.  Fixed
*                           infinite loops in scanForFile and scanForLocale.
*                           Modified getRawResourceData to not delete storage
*                           in localeData and resourceData which it doesn't own.
*                           Added Mac compatibility #ifdefs for tellp() and
*                           ios::nocreate.
*   2/18/97     helena      Updated with 100% documentation coverage.
*   3/13/97     aliu        Rewrote to load in entire resource bundle and store
*                           it as a Hashtable of ResourceBundleData objects.
*                           Added state table to govern parsing of files.
*                           Modified to load locale index out of new file
*                           distinct from default.txt.
*   3/25/97     aliu        Modified to support 2-d arrays, needed for timezone
*                           data. Added support for custom file suffixes.  Again,
*                           needed to support timezone data.
*   4/7/97      aliu        Cleaned up.
* 03/02/99      stephen     Removed dependency on FILE*.
* 03/29/99      helena      Merged Bertrand and Stephen's changes.
* 06/11/99      stephen     Removed parsing of .txt files.
*                           Reworked to use new binary format.
*                           Cleaned up.
* 06/14/99      stephen     Removed methods taking a filename suffix.
* 11/09/99      weiv        Added getLocale(), fRealLocale, removed fRealLocaleID
******************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API






/**
 * \file 
 * \brief C++ API: Resource Bundle
 */
 
U_NAMESPACE_BEGIN

/**
 * A class representing a collection of resource information pertaining to a given
 * locale. A resource bundle provides a way of accessing locale- specfic information in
 * a data file. You create a resource bundle that manages the resources for a given
 * locale and then ask it for individual resources.
 * <P>
 * Resource bundles in ICU4C are currently defined using text files which conform to the following
 * <a href="http://source.icu-project.org/repos/icu/icuhtml/trunk/design/bnf_rb.txt">BNF definition</a>.
 * More on resource bundle concepts and syntax can be found in the
 * <a href="http://icu-project.org/userguide/ResourceManagement.html">Users Guide</a>.
 * <P>
 *
 * The ResourceBundle class is not suitable for subclassing.
 *
 * @stable ICU 2.0
 */
class U_COMMON_API ResourceBundle : public UObject {
public:
    /**
     * Constructor
     *
     * @param packageName   The packageName and locale together point to an ICU udata object, 
     *                      as defined by <code> udata_open( packageName, "res", locale, err) </code> 
     *                      or equivalent.  Typically, packageName will refer to a (.dat) file, or to
     *                      a package registered with udata_setAppData(). Using a full file or directory
     *                      pathname for packageName is deprecated.
     * @param locale  This is the locale this resource bundle is for. To get resources
     *                for the French locale, for example, you would create a
     *                ResourceBundle passing Locale::FRENCH for the "locale" parameter,
     *                and all subsequent calls to that resource bundle will return
     *                resources that pertain to the French locale. If the caller doesn't
     *                pass a locale parameter, the default locale for the system (as
     *                returned by Locale::getDefault()) will be used.
     * @param err     The Error Code.
     * The UErrorCode& err parameter is used to return status information to the user. To
     * check whether the construction succeeded or not, you should check the value of
     * U_SUCCESS(err). If you wish more detailed information, you can check for
     * informational error results which still indicate success. U_USING_FALLBACK_WARNING
     * indicates that a fall back locale was used. For example, 'de_CH' was requested,
     * but nothing was found there, so 'de' was used. U_USING_DEFAULT_WARNING indicates that
     * the default locale data was used; neither the requested locale nor any of its
     * fall back locales could be found.
     * @stable ICU 2.0
     */
    ResourceBundle(const UnicodeString&    packageName,
                   const Locale&           locale,
                   UErrorCode&              err);

    /**
     * Construct a resource bundle for the default bundle in the specified package.
     *
     * @param packageName   The packageName and locale together point to an ICU udata object, 
     *                      as defined by <code> udata_open( packageName, "res", locale, err) </code> 
     *                      or equivalent.  Typically, packageName will refer to a (.dat) file, or to
     *                      a package registered with udata_setAppData(). Using a full file or directory
     *                      pathname for packageName is deprecated.
     * @param err A UErrorCode value
     * @stable ICU 2.0
     */
    ResourceBundle(const UnicodeString&    packageName,
                   UErrorCode&              err);

    /**
     * Construct a resource bundle for the ICU default bundle.
     *
     * @param err A UErrorCode value
     * @stable ICU 2.0
     */
    ResourceBundle(UErrorCode &err);

    /**
     * Standard constructor, constructs a resource bundle for the locale-specific
     * bundle in the specified package.
     *
     * @param packageName   The packageName and locale together point to an ICU udata object, 
     *                      as defined by <code> udata_open( packageName, "res", locale, err) </code> 
     *                      or equivalent.  Typically, packageName will refer to a (.dat) file, or to
     *                      a package registered with udata_setAppData(). Using a full file or directory
     *                      pathname for packageName is deprecated.
     *                      NULL is used to refer to ICU data.
     * @param locale The locale for which to open a resource bundle.
     * @param err A UErrorCode value
     * @stable ICU 2.0
     */
    ResourceBundle(const char* packageName,
                   const Locale& locale,
                   UErrorCode& err);

    /**
     * Copy constructor.
     *
     * @param original The resource bundle to copy.
     * @stable ICU 2.0
     */
    ResourceBundle(const ResourceBundle &original);

    /**
     * Constructor from a C UResourceBundle. The resource bundle is
     * copied and not adopted. ures_close will still need to be used on the
     * original resource bundle.
     *
     * @param res A pointer to the C resource bundle.
     * @param status A UErrorCode value.
     * @stable ICU 2.0
     */
    ResourceBundle(UResourceBundle *res,
                   UErrorCode &status);

    /**
     * Assignment operator.
     *
     * @param other The resource bundle to copy.
     * @stable ICU 2.0
     */
    ResourceBundle&
      operator=(const ResourceBundle& other);

    /** Destructor.
     * @stable ICU 2.0
     */
    virtual ~ResourceBundle();

    /**
     * Clone this object.
     * Clones can be used concurrently in multiple threads.
     * If an error occurs, then NULL is returned.
     * The caller must delete the clone.
     *
     * @return a clone of this object
     *
     * @see getDynamicClassID
     * @stable ICU 2.8
     */
    ResourceBundle *clone() const;

    /**
     * Returns the size of a resource. Size for scalar types is always 1, and for vector/table types is
     * the number of child resources.
     * @warning Integer array is treated as a scalar type. There are no
     *          APIs to access individual members of an integer array. It
     *          is always returned as a whole.
     *
     * @return number of resources in a given resource.
     * @stable ICU 2.0
     */
    int32_t
      getSize(void) const;

    /**
     * returns a string from a string resource type
     *
     * @param status  fills in the outgoing error code
     *                could be <TT>U_MISSING_RESOURCE_ERROR</TT> if the key is not found
     *                could be a warning
     *                e.g.: <TT>U_USING_FALLBACK_WARNING</TT>,<TT>U_USING_DEFAULT_WARNING </TT>
     * @return a pointer to a zero-terminated char16_t array which lives in a memory mapped/DLL file.
     * @stable ICU 2.0
     */
    UnicodeString
      getString(UErrorCode& status) const;

    /**
     * returns a binary data from a resource. Can be used at most primitive resource types (binaries,
     * strings, ints)
     *
     * @param len     fills in the length of resulting byte chunk
     * @param status  fills in the outgoing error code
     *                could be <TT>U_MISSING_RESOURCE_ERROR</TT> if the key is not found
     *                could be a warning
     *                e.g.: <TT>U_USING_FALLBACK_WARNING</TT>,<TT>U_USING_DEFAULT_WARNING </TT>
     * @return a pointer to a chunk of unsigned bytes which live in a memory mapped/DLL file.
     * @stable ICU 2.0
     */
    const uint8_t*
      getBinary(int32_t& len, UErrorCode& status) const;


    /**
     * returns an integer vector from a resource.
     *
     * @param len     fills in the length of resulting integer vector
     * @param status  fills in the outgoing error code
     *                could be <TT>U_MISSING_RESOURCE_ERROR</TT> if the key is not found
     *                could be a warning
     *                e.g.: <TT>U_USING_FALLBACK_WARNING</TT>,<TT>U_USING_DEFAULT_WARNING </TT>
     * @return a pointer to a vector of integers that lives in a memory mapped/DLL file.
     * @stable ICU 2.0
     */
    const int32_t*
      getIntVector(int32_t& len, UErrorCode& status) const;

    /**
     * returns an unsigned integer from a resource.
     * This integer is originally 28 bits.
     *
     * @param status  fills in the outgoing error code
     *                could be <TT>U_MISSING_RESOURCE_ERROR</TT> if the key is not found
     *                could be a warning
     *                e.g.: <TT>U_USING_FALLBACK_WARNING</TT>,<TT>U_USING_DEFAULT_WARNING </TT>
     * @return an unsigned integer value
     * @stable ICU 2.0
     */
    uint32_t
      getUInt(UErrorCode& status) const;

    /**
     * returns a signed integer from a resource.
     * This integer is originally 28 bit and the sign gets propagated.
     *
     * @param status  fills in the outgoing error code
     *                could be <TT>U_MISSING_RESOURCE_ERROR</TT> if the key is not found
     *                could be a warning
     *                e.g.: <TT>U_USING_FALLBACK_WARNING</TT>,<TT>U_USING_DEFAULT_WARNING </TT>
     * @return a signed integer value
     * @stable ICU 2.0
     */
    int32_t
      getInt(UErrorCode& status) const;

    /**
     * Checks whether the resource has another element to iterate over.
     *
     * @return TRUE if there are more elements, FALSE if there is no more elements
     * @stable ICU 2.0
     */
    UBool
      hasNext(void) const;

    /**
     * Resets the internal context of a resource so that iteration starts from the first element.
     *
     * @stable ICU 2.0
     */
    void
      resetIterator(void);

    /**
     * Returns the key associated with this resource. Not all the resources have a key - only
     * those that are members of a table.
     *
     * @return a key associated to this resource, or NULL if it doesn't have a key
     * @stable ICU 2.0
     */
    const char*
      getKey(void) const;

    /**
     * Gets the locale ID of the resource bundle as a string.
     * Same as getLocale().getName() .
     *
     * @return the locale ID of the resource bundle as a string
     * @stable ICU 2.0
     */
    const char*
      getName(void) const;


    /**
     * Returns the type of a resource. Available types are defined in enum UResType
     *
     * @return type of the given resource.
     * @stable ICU 2.0
     */
    UResType
      getType(void) const;

    /**
     * Returns the next resource in a given resource or NULL if there are no more resources
     *
     * @param status            fills in the outgoing error code
     * @return                  ResourceBundle object.
     * @stable ICU 2.0
     */
    ResourceBundle
      getNext(UErrorCode& status);

    /**
     * Returns the next string in a resource or NULL if there are no more resources
     * to iterate over.
     *
     * @param status            fills in the outgoing error code
     * @return an UnicodeString object.
     * @stable ICU 2.0
     */
    UnicodeString
      getNextString(UErrorCode& status);

    /**
     * Returns the next string in a resource or NULL if there are no more resources
     * to iterate over.
     *
     * @param key               fill in for key associated with this string
     * @param status            fills in the outgoing error code
     * @return an UnicodeString object.
     * @stable ICU 2.0
     */
    UnicodeString
      getNextString(const char ** key,
                    UErrorCode& status);

    /**
     * Returns the resource in a resource at the specified index.
     *
     * @param index             an index to the wanted resource.
     * @param status            fills in the outgoing error code
     * @return                  ResourceBundle object. If there is an error, resource is invalid.
     * @stable ICU 2.0
     */
    ResourceBundle
      get(int32_t index,
          UErrorCode& status) const;

    /**
     * Returns the string in a given resource at the specified index.
     *
     * @param index             an index to the wanted string.
     * @param status            fills in the outgoing error code
     * @return                  an UnicodeString object. If there is an error, string is bogus
     * @stable ICU 2.0
     */
    UnicodeString
      getStringEx(int32_t index,
                  UErrorCode& status) const;

    /**
     * Returns a resource in a resource that has a given key. This procedure works only with table
     * resources.
     *
     * @param key               a key associated with the wanted resource
     * @param status            fills in the outgoing error code.
     * @return                  ResourceBundle object. If there is an error, resource is invalid.
     * @stable ICU 2.0
     */
    ResourceBundle
      get(const char* key,
          UErrorCode& status) const;

    /**
     * Returns a string in a resource that has a given key. This procedure works only with table
     * resources.
     *
     * @param key               a key associated with the wanted string
     * @param status            fills in the outgoing error code
     * @return                  an UnicodeString object. If there is an error, string is bogus
     * @stable ICU 2.0
     */
    UnicodeString
      getStringEx(const char* key,
                  UErrorCode& status) const;

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Return the version number associated with this ResourceBundle as a string. Please
     * use getVersion, as this method is going to be deprecated.
     *
     * @return  A version number string as specified in the resource bundle or its parent.
     *          The caller does not own this string.
     * @see getVersion
     * @deprecated ICU 2.8 Use getVersion instead.
     */
    const char*
      getVersionNumber(void) const;
#endif  /* U_HIDE_DEPRECATED_API */

    /**
     * Return the version number associated with this ResourceBundle as a UVersionInfo array.
     *
     * @param versionInfo A UVersionInfo array that is filled with the version number
     *                    as specified in the resource bundle or its parent.
     * @stable ICU 2.0
     */
    void
      getVersion(UVersionInfo versionInfo) const;

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Return the Locale associated with this ResourceBundle.
     *
     * @return a Locale object
     * @deprecated ICU 2.8 Use getLocale(ULocDataLocaleType type, UErrorCode &status) overload instead.
     */
    const Locale&
      getLocale(void) const;
#endif  /* U_HIDE_DEPRECATED_API */

    /**
     * Return the Locale associated with this ResourceBundle.
     * @param type You can choose between requested, valid and actual
     *             locale. For description see the definition of
     *             ULocDataLocaleType in uloc.h
     * @param status just for catching illegal arguments
     *
     * @return a Locale object
     * @stable ICU 2.8
     */
    const Locale
      getLocale(ULocDataLocaleType type, UErrorCode &status) const;
#ifndef U_HIDE_INTERNAL_API
    /**
     * This API implements multilevel fallback
     * @internal
     */
    ResourceBundle
        getWithFallback(const char* key, UErrorCode& status);
#endif  /* U_HIDE_INTERNAL_API */
    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 2.2
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 2.2
     */
    static UClassID U_EXPORT2 getStaticClassID();

private:
    ResourceBundle(); // default constructor not implemented

    UResourceBundle *fResource;
    void constructForLocale(const UnicodeString& path, const Locale& locale, UErrorCode& error);
    Locale *fLocale;
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */










U_NAMESPACE_BEGIN

/*-----------------------------------------------------------------------------
 * Implementation Notes
 *
 * Resource bundles are read in once, and thereafter cached.
 * ResourceBundle statically keeps track of which files have been
 * read, so we are guaranteed that each file is read at most once.
 * Resource bundles can be loaded from different data directories and
 * will be treated as distinct, even if they are for the same locale.
 *
 * Resource bundles are lightweight objects, which have pointers to
 * one or more shared Hashtable objects containing all the data.
 * Copying would be cheap, but there is no copy constructor, since
 * there wasn't one in the original API.
 *
 * The ResourceBundle parsing mechanism is implemented as a transition
 * network, for easy maintenance and modification.  The network is
 * implemented as a matrix (instead of in code) to make this even
 * easier.  The matrix contains Transition objects.  Each Transition
 * object describes a destination node and an action to take before
 * moving to the destination node.  The source node is encoded by the
 * index of the object in the array that contains it.  The pieces
 * needed to understand the transition network are the enums for node
 * IDs and actions, the parse() method, which walks through the
 * network and implements the actions, and the network itself.  The
 * network guarantees certain conditions, for example, that a new
 * resource will not be closed until one has been opened first; or
 * that data will not be stored into a TaggedList until a TaggedList
 * has been created.  Nonetheless, the code in parse() does some
 * consistency checks as it runs the network, and fails with an
 * U_INTERNAL_PROGRAM_ERROR if one of these checks fails.  If the input
 * data has a bad format, an U_INVALID_FORMAT_ERROR is returned.  If you
 * see an U_INTERNAL_PROGRAM_ERROR the transition matrix has a bug in
 * it.
 *
 * Old functionality of multiple locales in a single file is still
 * supported.  For this reason, LOCALE names override FILE names.  If
 * data for en_US is located in the en.txt file, once it is loaded,
 * the code will not care where it came from (other than remembering
 * which directory it came from).  However, if there is an en_US
 * resource in en_US.txt, that will take precedence.  There is no
 * limit to the number or type of resources that can be stored in a
 * file, however, files are only searched in a specific way.  If
 * en_US_CA is requested, then first en_US_CA.txt is searched, then
 * en_US.txt, then en.txt, then default.txt.  So it only makes sense
 * to put certain locales in certain files.  In this example, it would
 * be logical to put en_US_CA, en_US, and en into the en.txt file,
 * since they would be found there if asked for.  The extreme example
 * is to place all locale resources into default.txt, which should
 * also work.
 *
 * Inheritance is implemented.  For example, xx_YY_zz inherits as
 * follows: xx_YY_zz, xx_YY, xx, default.  Inheritance is implemented
 * as an array of hashtables.  There will be from 1 to 4 hashtables in
 * the array.
 *
 * Fallback files are implemented.  The fallback pattern is Language
 * Country Variant (LCV) -> LC -> L.  Fallback is first done for the
 * requested locale.  Then it is done for the default locale, as
 * returned by Locale::getDefault().  Then the special file
 * default.txt is searched for the default locale.  The overall FILE
 * fallback path is LCV -> LC -> L -> dLCV -> dLC -> dL -> default.
 *
 * Note that although file name searching includes the default locale,
 * once a ResourceBundle object is constructed, the inheritance path
 * no longer includes the default locale.  The path is LCV -> LC -> L
 * -> default.
 *
 * File parsing is lazy.  Nothing is parsed unless it is called for by
 * someone.  So when a ResourceBundle for xx_YY_zz is constructed,
 * only that locale is parsed (along with anything else in the same
 * file).  Later, if the FooBar tag is asked for, and if it isn't
 * found in xx_YY_zz, then xx_YY.txt will be parsed and checked, and
 * so forth, until the chain is exhausted or the tag is found.
 *
 * Thread-safety is implemented around caches, both the cache that
 * stores all the resouce data, and the cache that stores flags
 * indicating whether or not a file has been visited.  These caches
 * delete their storage at static cleanup time, when the process
 * quits.
 *
 * ResourceBundle supports TableCollation as a special case.  This
 * involves having special ResourceBundle objects which DO own their
 * data, since we don't want large collation rule strings in the
 * ResourceBundle cache (these are already cached in the
 * TableCollation cache).  TableCollation files (.ctx files) have the
 * same format as normal resource data files, with a different
 * interpretation, from the standpoint of ResourceBundle.  .ctx files
 * are loaded into otherwise ordinary ResourceBundle objects.  They
 * don't inherit (that's implemented by TableCollation) and they own
 * their data (as mentioned above).  However, they still support
 * possible multiple locales in a single .ctx file.  (This is in
 * practice a bad idea, since you only want the one locale you're
 * looking for, and only one tag will be present
 * ("CollationElements"), so you don't need an inheritance chain of
 * multiple locales.)  Up to 4 locale resources will be loaded from a
 * .ctx file; everything after the first 4 is ignored (parsed and
 * deleted).  (Normal .txt files have no limit.)  Instead of being
 * loaded into the cache, and then looked up as needed, the locale
 * resources are read straight into the ResourceBundle object.
 *
 * The Index, which used to reside in default.txt, has been moved to a
 * new file, index.txt.  This file contains a slightly modified format
 * with the addition of the "InstalledLocales" tag; it looks like:
 *
 * Index {
 *   InstalledLocales {
 *     ar
 *     ..
 *     zh_TW
 *   }
 * }
 */
//-----------------------------------------------------------------------------

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(ResourceBundle)

ResourceBundle::ResourceBundle(UErrorCode &err)
                                :UObject(), fLocale(NULL)
{
    fResource = ures_open(0, Locale::getDefault().getName(), &err);
}

ResourceBundle::ResourceBundle(const ResourceBundle &other)
                              :UObject(other), fLocale(NULL)
{
    UErrorCode status = U_ZERO_ERROR;

    if (other.fResource) {
        fResource = ures_copyResb(0, other.fResource, &status);
    } else {
        /* Copying a bad resource bundle */
        fResource = NULL;
    }
}

ResourceBundle::ResourceBundle(UResourceBundle *res, UErrorCode& err)
                               :UObject(), fLocale(NULL)
{
    if (res) {
        fResource = ures_copyResb(0, res, &err);
    } else {
        /* Copying a bad resource bundle */
        fResource = NULL;
    }
}

ResourceBundle::ResourceBundle(const char* path, const Locale& locale, UErrorCode& err)
                               :UObject(), fLocale(NULL)
{
    fResource = ures_open(path, locale.getName(), &err);
}


ResourceBundle& ResourceBundle::operator=(const ResourceBundle& other)
{
    if(this == &other) {
        return *this;
    }
    if(fResource != 0) {
        ures_close(fResource);
        fResource = NULL;
    }
    if (fLocale != NULL) {
        delete fLocale;
        fLocale = NULL;
    }
    UErrorCode status = U_ZERO_ERROR;
    if (other.fResource) {
        fResource = ures_copyResb(0, other.fResource, &status);
    } else {
        /* Copying a bad resource bundle */
        fResource = NULL;
    }
    return *this;
}

ResourceBundle::~ResourceBundle()
{
    if(fResource != 0) {
        ures_close(fResource);
    }
    if(fLocale != NULL) {
      delete(fLocale);
    }
}

ResourceBundle *
ResourceBundle::clone() const {
    return new ResourceBundle(*this);
}

UnicodeString ResourceBundle::getString(UErrorCode& status) const {
    int32_t len = 0;
    const UChar *r = ures_getString(fResource, &len, &status);
    return UnicodeString(TRUE, r, len);
}

const uint8_t *ResourceBundle::getBinary(int32_t& len, UErrorCode& status) const {
    return ures_getBinary(fResource, &len, &status);
}

const int32_t *ResourceBundle::getIntVector(int32_t& len, UErrorCode& status) const {
    return ures_getIntVector(fResource, &len, &status);
}

uint32_t ResourceBundle::getUInt(UErrorCode& status) const {
    return ures_getUInt(fResource, &status);
}

int32_t ResourceBundle::getInt(UErrorCode& status) const {
    return ures_getInt(fResource, &status);
}

const char *ResourceBundle::getName(void) const {
    return ures_getName(fResource);
}

const char *ResourceBundle::getKey(void) const {
    return ures_getKey(fResource);
}

UResType ResourceBundle::getType(void) const {
    return ures_getType(fResource);
}

int32_t ResourceBundle::getSize(void) const {
    return ures_getSize(fResource);
}

UBool ResourceBundle::hasNext(void) const {
    return ures_hasNext(fResource);
}

void ResourceBundle::resetIterator(void) {
    ures_resetIterator(fResource);
}

ResourceBundle ResourceBundle::getNext(UErrorCode& status) {
    UResourceBundle r;

    ures_initStackObject(&r);
    ures_getNextResource(fResource, &r, &status);
    ResourceBundle res(&r, status);
    if (U_SUCCESS(status)) {
        ures_close(&r);
    }
    return res;
}

UnicodeString ResourceBundle::getNextString(UErrorCode& status) {
    int32_t len = 0;
    const UChar* r = ures_getNextString(fResource, &len, 0, &status);
    return UnicodeString(TRUE, r, len);
}

UnicodeString ResourceBundle::getNextString(const char ** key, UErrorCode& status) {
    int32_t len = 0;
    const UChar* r = ures_getNextString(fResource, &len, key, &status);
    return UnicodeString(TRUE, r, len);
}

ResourceBundle ResourceBundle::get(int32_t indexR, UErrorCode& status) const {
    UResourceBundle r;

    ures_initStackObject(&r);
    ures_getByIndex(fResource, indexR, &r, &status);
    ResourceBundle res(&r, status);
    if (U_SUCCESS(status)) {
        ures_close(&r);
    }
    return res;
}

UnicodeString ResourceBundle::getStringEx(int32_t indexS, UErrorCode& status) const {
    int32_t len = 0;
    const UChar* r = ures_getStringByIndex(fResource, indexS, &len, &status);
    return UnicodeString(TRUE, r, len);
}

ResourceBundle ResourceBundle::get(const char* key, UErrorCode& status) const {
    UResourceBundle r;

    ures_initStackObject(&r);
    ures_getByKey(fResource, key, &r, &status);
    ResourceBundle res(&r, status);
    if (U_SUCCESS(status)) {
        ures_close(&r);
    }
    return res;
}

ResourceBundle ResourceBundle::getWithFallback(const char* key, UErrorCode& status){
    UResourceBundle r;
    ures_initStackObject(&r);
    ures_getByKeyWithFallback(fResource, key, &r, &status);
    ResourceBundle res(&r, status);
    if(U_SUCCESS(status)){
        ures_close(&r);
    }
    return res;
}
UnicodeString ResourceBundle::getStringEx(const char* key, UErrorCode& status) const {
    int32_t len = 0;
    const UChar* r = ures_getStringByKey(fResource, key, &len, &status);
    return UnicodeString(TRUE, r, len);
}

const char*
ResourceBundle::getVersionNumber()  const
{
    return ures_getVersionNumberInternal(fResource);
}

void ResourceBundle::getVersion(UVersionInfo versionInfo) const {
    ures_getVersion(fResource, versionInfo);
}

const Locale &ResourceBundle::getLocale(void) const {
    static UMutex gLocaleLock;
    Mutex lock(&gLocaleLock);
    if (fLocale != NULL) {
        return *fLocale;
    }
    UErrorCode status = U_ZERO_ERROR;
    const char *localeName = ures_getLocaleInternal(fResource, &status);
    ResourceBundle *ncThis = const_cast<ResourceBundle *>(this);
    ncThis->fLocale = new Locale(localeName);
    return ncThis->fLocale != NULL ? *ncThis->fLocale : Locale::getDefault();
}

const Locale ResourceBundle::getLocale(ULocDataLocaleType type, UErrorCode &status) const
{
  return ures_getLocaleByType(fResource, type, &status);
}

U_NAMESPACE_END
//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1997-2006, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  resbund_cnv.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004aug25
*   created by: Markus W. Scherer
*
*   Character conversion functions moved here from resbund.cpp
*/





U_NAMESPACE_BEGIN

ResourceBundle::ResourceBundle( const UnicodeString&    path,
                                const Locale&           locale,
                                UErrorCode&              error)
                                :UObject(), fLocale(NULL)
{
    constructForLocale(path, locale, error);
}

ResourceBundle::ResourceBundle( const UnicodeString&    path,
                                UErrorCode&              error)
                                :UObject(), fLocale(NULL)
{
    constructForLocale(path, Locale::getDefault(), error);
}

void 
ResourceBundle::constructForLocale(const UnicodeString& path,
                                   const Locale& locale,
                                   UErrorCode& error)
{
    if (path.isEmpty()) {
        fResource = ures_open(NULL, locale.getName(), &error);
    }
    else {
        UnicodeString nullTerminatedPath(path);
        nullTerminatedPath.append((UChar)0);
        fResource = ures_openU(nullTerminatedPath.getBuffer(), locale.getName(), &error);
    }
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2015-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* resource.cpp
*
* created on: 2015nov04
* created by: Markus W. Scherer
*/



U_NAMESPACE_BEGIN

ResourceValue::~ResourceValue() {}

ResourceSink::~ResourceSink() {}

U_NAMESPACE_END
// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



// #if U_ENABLE_TRACING

// #include "restrace.h"
// #include "charstr.h"
// #include "cstring.h"
// #include "utracimp.h"
// #include "uresimp.h"
// #include "uassert.h"
// #include "util.h"

// U_NAMESPACE_BEGIN

// ResourceTracer::~ResourceTracer() = default;

// void ResourceTracer::trace(const char* resType) const {
//     U_ASSERT(fResB || fParent);
//     UTRACE_ENTRY(UTRACE_UDATA_RESOURCE);
//     UErrorCode status = U_ZERO_ERROR;

//     CharString filePath;
//     getFilePath(filePath, status);

//     CharString resPath;
//     getResPath(resPath, status);

//     // The longest type ("intvector") is 9 chars
//     const char kSpaces[] = "         ";
//     CharString format;
//     format.append(kSpaces, sizeof(kSpaces) - 1 - uprv_strlen(resType), status);
//     format.append("(%s) %s @ %s", status);

//     UTRACE_DATA3(UTRACE_VERBOSE,
//         format.data(),
//         resType,
//         filePath.data(),
//         resPath.data());
//     UTRACE_EXIT_STATUS(status);
// }

// void ResourceTracer::traceOpen() const {
//     U_ASSERT(fResB);
//     UTRACE_ENTRY(UTRACE_UDATA_BUNDLE);
//     UErrorCode status = U_ZERO_ERROR;

//     CharString filePath;
//     UTRACE_DATA1(UTRACE_VERBOSE, "%s", getFilePath(filePath, status).data());
//     UTRACE_EXIT_STATUS(status);
// }

// CharString& ResourceTracer::getFilePath(CharString& output, UErrorCode& status) const {
//     if (fResB) {
//         output.append(fResB->fData->fPath, status);
//         output.append('/', status);
//         output.append(fResB->fData->fName, status);
//         output.append(".res", status);
//     } else {
//         fParent->getFilePath(output, status);
//     }
//     return output;
// }

// CharString& ResourceTracer::getResPath(CharString& output, UErrorCode& status) const {
//     if (fResB) {
//         output.append('/', status);
//         output.append(fResB->fResPath, status);
//         // removing the trailing /
//         U_ASSERT(output[output.length()-1] == '/');
//         output.truncate(output.length()-1);
//     } else {
//         fParent->getResPath(output, status);
//     }
//     if (fKey) {
//         output.append('/', status);
//         output.append(fKey, status);
//     }
//     if (fIndex != -1) {
//         output.append('[', status);
//         UnicodeString indexString;
//         ICU_Utility::appendNumber(indexString, fIndex);
//         output.appendInvariantChars(indexString, status);
//         output.append(']', status);
//     }
//     return output;
// }

// void FileTracer::traceOpen(const char* path, const char* type, const char* name) {
//     if (uprv_strcmp(type, "res") == 0) {
//         traceOpenResFile(path, name);
//     } else {
//         traceOpenDataFile(path, type, name);
//     }
// }

// void FileTracer::traceOpenDataFile(const char* path, const char* type, const char* name) {
//     UTRACE_ENTRY(UTRACE_UDATA_DATA_FILE);
//     UErrorCode status = U_ZERO_ERROR;

//     CharString filePath;
//     filePath.append(path, status);
//     filePath.append('/', status);
//     filePath.append(name, status);
//     filePath.append('.', status);
//     filePath.append(type, status);

//     UTRACE_DATA1(UTRACE_VERBOSE, "%s", filePath.data());
//     UTRACE_EXIT_STATUS(status);
// }

// void FileTracer::traceOpenResFile(const char* path, const char* name) {
//     UTRACE_ENTRY(UTRACE_UDATA_RES_FILE);
//     UErrorCode status = U_ZERO_ERROR;

//     CharString filePath;
//     filePath.append(path, status);
//     filePath.append('/', status);
//     filePath.append(name, status);
//     filePath.append(".res", status);

//     UTRACE_DATA1(UTRACE_VERBOSE, "%s", filePath.data());
//     UTRACE_EXIT_STATUS(status);
// }

// U_NAMESPACE_END

// #endif // U_ENABLE_TRACING
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2003-2011, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: September 24 2003
* Since: ICU 2.8
**********************************************************************
*/
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2003-2011, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: September 24 2003
* Since: ICU 2.8
**********************************************************************
*/



U_NAMESPACE_BEGIN

class UnicodeString;
class ParsePosition;
class SymbolTable;

/**
 * An iterator that returns 32-bit code points.  This class is deliberately
 * <em>not</em> related to any of the ICU character iterator classes
 * in order to minimize complexity.
 * @author Alan Liu
 * @since ICU 2.8
 */
class RuleCharacterIterator : public UMemory {

    // TODO: Ideas for later.  (Do not implement if not needed, lest the
    // code coverage numbers go down due to unused methods.)
    // 1. Add a copy constructor, operator==() method.
    // 2. Rather than return DONE, throw an exception if the end
    // is reached -- this is an alternate usage model, probably not useful.

private:
    /**
     * Text being iterated.
     */    
    const UnicodeString& text;

    /**
     * Position of iterator.
     */
    ParsePosition& pos;

    /**
     * Symbol table used to parse and dereference variables.  May be 0.
     */
    const SymbolTable* sym;
    
    /**
     * Current variable expansion, or 0 if none.
     */
    const UnicodeString* buf;

    /**
     * Position within buf.  Meaningless if buf == 0.
     */
    int32_t bufPos;

public:
    /**
     * Value returned when there are no more characters to iterate.
     */
    enum { DONE = -1 };

    /**
     * Bitmask option to enable parsing of variable names.  If (options &
     * PARSE_VARIABLES) != 0, then an embedded variable will be expanded to
     * its value.  Variables are parsed using the SymbolTable API.
     */
    enum { PARSE_VARIABLES = 1 };

    /**
     * Bitmask option to enable parsing of escape sequences.  If (options &
     * PARSE_ESCAPES) != 0, then an embedded escape sequence will be expanded
     * to its value.  Escapes are parsed using Utility.unescapeAt().
     */
    enum { PARSE_ESCAPES   = 2 };

    /**
     * Bitmask option to enable skipping of whitespace.  If (options &
     * SKIP_WHITESPACE) != 0, then Pattern_White_Space characters will be silently
     * skipped, as if they were not present in the input.
     */
    enum { SKIP_WHITESPACE = 4 };

    /**
     * Constructs an iterator over the given text, starting at the given
     * position.
     * @param text the text to be iterated
     * @param sym the symbol table, or null if there is none.  If sym is null,
     * then variables will not be deferenced, even if the PARSE_VARIABLES
     * option is set.
     * @param pos upon input, the index of the next character to return.  If a
     * variable has been dereferenced, then pos will <em>not</em> increment as
     * characters of the variable value are iterated.
     */
    RuleCharacterIterator(const UnicodeString& text, const SymbolTable* sym,
                          ParsePosition& pos);
    
    /**
     * Returns true if this iterator has no more characters to return.
     */
    UBool atEnd() const;

    /**
     * Returns the next character using the given options, or DONE if there
     * are no more characters, and advance the position to the next
     * character.
     * @param options one or more of the following options, bitwise-OR-ed
     * together: PARSE_VARIABLES, PARSE_ESCAPES, SKIP_WHITESPACE.
     * @param isEscaped output parameter set to TRUE if the character
     * was escaped
     * @param ec input-output error code.  An error will only be set by
     * this routing if options includes PARSE_VARIABLES and an unknown
     * variable name is seen, or if options includes PARSE_ESCAPES and
     * an invalid escape sequence is seen.
     * @return the current 32-bit code point, or DONE
     */
    UChar32 next(int32_t options, UBool& isEscaped, UErrorCode& ec);

    /**
     * Returns true if this iterator is currently within a variable expansion.
     */
    inline UBool inVariable() const;

    /**
     * An opaque object representing the position of a RuleCharacterIterator.
     */
    struct Pos : public UMemory {
    private:
        const UnicodeString* buf;
        int32_t pos;
        int32_t bufPos;
        friend class RuleCharacterIterator;
    };

    /**
     * Sets an object which, when later passed to setPos(), will
     * restore this iterator's position.  Usage idiom:
     *
     * RuleCharacterIterator iterator = ...;
     * RuleCharacterIterator::Pos pos;
     * iterator.getPos(pos);
     * for (;;) {
     *   iterator.getPos(pos);
     *   int c = iterator.next(...);
     *   ...
     * }
     * iterator.setPos(pos);
     *
     * @param p a position object to be set to this iterator's
     * current position.
     */
    void getPos(Pos& p) const;

    /**
     * Restores this iterator to the position it had when getPos()
     * set the given object.
     * @param p a position object previously set by getPos()
     */
    void setPos(const Pos& p);

    /**
     * Skips ahead past any ignored characters, as indicated by the given
     * options.  This is useful in conjunction with the lookahead() method.
     *
     * Currently, this only has an effect for SKIP_WHITESPACE.
     * @param options one or more of the following options, bitwise-OR-ed
     * together: PARSE_VARIABLES, PARSE_ESCAPES, SKIP_WHITESPACE.
     */
    void skipIgnored(int32_t options);

    /**
     * Returns a string containing the remainder of the characters to be
     * returned by this iterator, without any option processing.  If the
     * iterator is currently within a variable expansion, this will only
     * extend to the end of the variable expansion.  This method is provided
     * so that iterators may interoperate with string-based APIs.  The typical
     * sequence of calls is to call skipIgnored(), then call lookahead(), then
     * parse the string returned by lookahead(), then call jumpahead() to
     * resynchronize the iterator.
     * @param result a string to receive the characters to be returned
     * by future calls to next()
     * @param maxLookAhead The maximum to copy into the result.
     * @return a reference to result
     */
    UnicodeString& lookahead(UnicodeString& result, int32_t maxLookAhead = -1) const;

    /**
     * Advances the position by the given number of 16-bit code units.
     * This is useful in conjunction with the lookahead() method.
     * @param count the number of 16-bit code units to jump over
     */
    void jumpahead(int32_t count);

    /**
     * Returns a string representation of this object, consisting of the
     * characters being iterated, with a '|' marking the current position.
     * Position within an expanded variable is <em>not</em> indicated.
     * @param result output parameter to receive a string
     * representation of this object
     */
//    UnicodeString& toString(UnicodeString& result) const;
    
private:
    /**
     * Returns the current 32-bit code point without parsing escapes, parsing
     * variables, or skipping whitespace.
     * @return the current 32-bit code point
     */
    UChar32 _current() const;
    
    /**
     * Advances the position by the given amount.
     * @param count the number of 16-bit code units to advance past
     */
    void _advance(int32_t count);
};

inline UBool RuleCharacterIterator::inVariable() const {
    return buf != 0;
}

U_NAMESPACE_END



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (c) 2000-2005, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*   Date        Name        Description
*   02/04/00    aliu        Creation.
**********************************************************************
*/



#if U_SHOW_CPLUSPLUS_API



/**
 * \file 
 * \brief C++ API: An interface that defines both lookup protocol and parsing of
 * symbolic names.
 */
 
U_NAMESPACE_BEGIN

class ParsePosition;
class UnicodeFunctor;
class UnicodeSet;
class UnicodeString;

/**
 * An interface that defines both lookup protocol and parsing of
 * symbolic names.
 *
 * <p>A symbol table maintains two kinds of mappings.  The first is
 * between symbolic names and their values.  For example, if the
 * variable with the name "start" is set to the value "alpha"
 * (perhaps, though not necessarily, through an expression such as
 * "$start=alpha"), then the call lookup("start") will return the
 * char[] array ['a', 'l', 'p', 'h', 'a'].
 *
 * <p>The second kind of mapping is between character values and
 * UnicodeMatcher objects.  This is used by RuleBasedTransliterator,
 * which uses characters in the private use area to represent objects
 * such as UnicodeSets.  If U+E015 is mapped to the UnicodeSet [a-z],
 * then lookupMatcher(0xE015) will return the UnicodeSet [a-z].
 *
 * <p>Finally, a symbol table defines parsing behavior for symbolic
 * names.  All symbolic names start with the SYMBOL_REF character.
 * When a parser encounters this character, it calls parseReference()
 * with the position immediately following the SYMBOL_REF.  The symbol
 * table parses the name, if there is one, and returns it.
 *
 * @stable ICU 2.8
 */
class U_COMMON_API SymbolTable /* not : public UObject because this is an interface/mixin class */ {
public:

    /**
     * The character preceding a symbol reference name.
     * @stable ICU 2.8
     */
    enum { SYMBOL_REF = 0x0024 /*$*/ };

    /**
     * Destructor.
     * @stable ICU 2.8
     */
    virtual ~SymbolTable();

    /**
     * Lookup the characters associated with this string and return it.
     * Return <tt>NULL</tt> if no such name exists.  The resultant
     * string may have length zero.
     * @param s the symbolic name to lookup
     * @return a string containing the name's value, or <tt>NULL</tt> if
     * there is no mapping for s.
     * @stable ICU 2.8
     */
    virtual const UnicodeString* lookup(const UnicodeString& s) const = 0;

    /**
     * Lookup the UnicodeMatcher associated with the given character, and
     * return it.  Return <tt>NULL</tt> if not found.
     * @param ch a 32-bit code point from 0 to 0x10FFFF inclusive.
     * @return the UnicodeMatcher object represented by the given
     * character, or NULL if there is no mapping for ch.
     * @stable ICU 2.8
     */
    virtual const UnicodeFunctor* lookupMatcher(UChar32 ch) const = 0;

    /**
     * Parse a symbol reference name from the given string, starting
     * at the given position.  If no valid symbol reference name is
     * found, return the empty string and leave pos unchanged.  That is, if the
     * character at pos cannot start a name, or if pos is at or after
     * text.length(), then return an empty string.  This indicates an
     * isolated SYMBOL_REF character.
     * @param text the text to parse for the name
     * @param pos on entry, the index of the first character to parse.
     * This is the character following the SYMBOL_REF character.  On
     * exit, the index after the last parsed character.  If the parse
     * failed, pos is unchanged on exit.
     * @param limit the index after the last character to be parsed.
     * @return the parsed name, or an empty string if there is no
     * valid symbolic name at the given position.
     * @stable ICU 2.8
     */
    virtual UnicodeString parseReference(const UnicodeString& text,
                                         ParsePosition& pos, int32_t limit) const = 0;
};
U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */






/* \U87654321 or \ud800\udc00 */
#define MAX_U_NOTATION_LEN 12

U_NAMESPACE_BEGIN

RuleCharacterIterator::RuleCharacterIterator(const UnicodeString& theText, const SymbolTable* theSym,
                      ParsePosition& thePos) :
    text(theText),
    pos(thePos),
    sym(theSym),
    buf(0),
    bufPos(0)
{}

UBool RuleCharacterIterator::atEnd() const {
    return buf == 0 && pos.getIndex() == text.length();
}

UChar32 RuleCharacterIterator::next(int32_t options, UBool& isEscaped, UErrorCode& ec) {
    if (U_FAILURE(ec)) return DONE;

    UChar32 c = DONE;
    isEscaped = FALSE;

    for (;;) {
        c = _current();
        _advance(U16_LENGTH(c));

        if (c == SymbolTable::SYMBOL_REF && buf == 0 &&
            (options & PARSE_VARIABLES) != 0 && sym != 0) {
            UnicodeString name = sym->parseReference(text, pos, text.length());
            // If name is empty there was an isolated SYMBOL_REF;
            // return it.  Caller must be prepared for this.
            if (name.length() == 0) {
                break;
            }
            bufPos = 0;
            buf = sym->lookup(name);
            if (buf == 0) {
                ec = U_UNDEFINED_VARIABLE;
                return DONE;
            }
            // Handle empty variable value
            if (buf->length() == 0) {
                buf = 0;
            }
            continue;
        }

        if ((options & SKIP_WHITESPACE) != 0 && PatternProps::isWhiteSpace(c)) {
            continue;
        }

        if (c == 0x5C /*'\\'*/ && (options & PARSE_ESCAPES) != 0) {
            UnicodeString tempEscape;
            int32_t offset = 0;
            c = lookahead(tempEscape, MAX_U_NOTATION_LEN).unescapeAt(offset);
            jumpahead(offset);
            isEscaped = TRUE;
            if (c < 0) {
                ec = U_MALFORMED_UNICODE_ESCAPE;
                return DONE;
            }
        }

        break;
    }

    return c;
}

void RuleCharacterIterator::getPos(RuleCharacterIterator::Pos& p) const {
    p.buf = buf;
    p.pos = pos.getIndex();
    p.bufPos = bufPos;
}

void RuleCharacterIterator::setPos(const RuleCharacterIterator::Pos& p) {
    buf = p.buf;
    pos.setIndex(p.pos);
    bufPos = p.bufPos;
}

void RuleCharacterIterator::skipIgnored(int32_t options) {
    if ((options & SKIP_WHITESPACE) != 0) {
        for (;;) {
            UChar32 a = _current();
            if (!PatternProps::isWhiteSpace(a)) break;
            _advance(U16_LENGTH(a));
        }
    }
}

UnicodeString& RuleCharacterIterator::lookahead(UnicodeString& result, int32_t maxLookAhead) const {
    if (maxLookAhead < 0) {
        maxLookAhead = 0x7FFFFFFF;
    }
    if (buf != 0) {
        buf->extract(bufPos, maxLookAhead, result);
    } else {
        text.extract(pos.getIndex(), maxLookAhead, result);
    }
    return result;
}

void RuleCharacterIterator::jumpahead(int32_t count) {
    _advance(count);
}

/*
UnicodeString& RuleCharacterIterator::toString(UnicodeString& result) const {
    int32_t b = pos.getIndex();
    text.extract(0, b, result);
    return result.append((UChar) 0x7C).append(text, b, 0x7FFFFFFF); // Insert '|' at index
}
*/

UChar32 RuleCharacterIterator::_current() const {
    if (buf != 0) {
        return buf->char32At(bufPos);
    } else {
        int i = pos.getIndex();
        return (i < text.length()) ? text.char32At(i) : (UChar32)DONE;
    }
}

void RuleCharacterIterator::_advance(int32_t count) {
    if (buf != 0) {
        bufPos += count;
        if (bufPos == buf->length()) {
            buf = 0;
        }
    } else {
        pos.setIndex(pos.getIndex() + count);
        if (pos.getIndex() > text.length()) {
            pos.setIndex(text.length());
        }
    }
}

U_NAMESPACE_END

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 1998-2012, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*
* File schriter.cpp
*
* Modification History:
*
*   Date        Name        Description
*  05/05/99     stephen     Cleaned up.
******************************************************************************
*/

  // for 'typeid' to work




U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(StringCharacterIterator)

StringCharacterIterator::StringCharacterIterator()
  : UCharCharacterIterator(),
    text()
{
  // NEVER DEFAULT CONSTRUCT!
}

StringCharacterIterator::StringCharacterIterator(const UnicodeString& textStr)
  : UCharCharacterIterator(textStr.getBuffer(), textStr.length()),
    text(textStr)
{
    // we had set the input parameter's array, now we need to set our copy's array
    UCharCharacterIterator::text = this->text.getBuffer();
}

StringCharacterIterator::StringCharacterIterator(const UnicodeString& textStr,
                                                 int32_t textPos)
  : UCharCharacterIterator(textStr.getBuffer(), textStr.length(), textPos),
    text(textStr)
{
    // we had set the input parameter's array, now we need to set our copy's array
    UCharCharacterIterator::text = this->text.getBuffer();
}

StringCharacterIterator::StringCharacterIterator(const UnicodeString& textStr,
                                                 int32_t textBegin,
                                                 int32_t textEnd,
                                                 int32_t textPos)
  : UCharCharacterIterator(textStr.getBuffer(), textStr.length(), textBegin, textEnd, textPos),
    text(textStr)
{
    // we had set the input parameter's array, now we need to set our copy's array
    UCharCharacterIterator::text = this->text.getBuffer();
}

StringCharacterIterator::StringCharacterIterator(const StringCharacterIterator& that)
  : UCharCharacterIterator(that),
    text(that.text)
{
    // we had set the input parameter's array, now we need to set our copy's array
    UCharCharacterIterator::text = this->text.getBuffer();
}

StringCharacterIterator::~StringCharacterIterator() {
}

StringCharacterIterator&
StringCharacterIterator::operator=(const StringCharacterIterator& that) {
    UCharCharacterIterator::operator=(that);
    text = that.text;
    // we had set the input parameter's array, now we need to set our copy's array
    UCharCharacterIterator::text = this->text.getBuffer();
    return *this;
}

UBool
StringCharacterIterator::operator==(const ForwardCharacterIterator& that) const {
    if (this == &that) {
        return TRUE;
    }

    // do not call UCharCharacterIterator::operator==()
    // because that checks for array pointer equality
    // while we compare UnicodeString objects

    if (typeid(*this) != typeid(that)) {
        return FALSE;
    }

    StringCharacterIterator&    realThat = (StringCharacterIterator&)that;

    return text == realThat.text
        && pos == realThat.pos
        && begin == realThat.begin
        && end == realThat.end;
}

StringCharacterIterator*
StringCharacterIterator::clone() const {
    return new StringCharacterIterator(*this);
}

void
StringCharacterIterator::setText(const UnicodeString& newText) {
    text = newText;
    UCharCharacterIterator::setText(text.getBuffer(), text.length());
}

void
StringCharacterIterator::getText(UnicodeString& result) {
    result = text;
}
U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
*******************************************************************************
* Copyright (C) 2001-2014, International Business Machines Corporation.
* All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_SERVICE

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
 *******************************************************************************
 * Copyright (C) 2001-2011, International Business Machines Corporation.       *
 * All Rights Reserved.                                                        *
 *******************************************************************************
 */




#if UCONFIG_NO_SERVICE

U_NAMESPACE_BEGIN

/*
 * Allow the declaration of APIs with pointers to ICUService
 * even when service is removed from the build.
 */
class ICUService;

U_NAMESPACE_END

#else







// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
 *******************************************************************************
 * Copyright (C) 2001-2014, International Business Machines Corporation and    *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 */



#if UCONFIG_NO_SERVICE

U_NAMESPACE_BEGIN

/*
 * Allow the declaration of APIs with pointers to BreakIterator
 * even when break iteration is removed from the build.
 */
class ICUNotifier;

U_NAMESPACE_END

#else







U_NAMESPACE_BEGIN

class U_COMMON_API EventListener : public UObject {
public: 
    virtual ~EventListener();

public:
    static UClassID U_EXPORT2 getStaticClassID();

    virtual UClassID getDynamicClassID() const;

public:
#ifdef SERVICE_DEBUG
    virtual UnicodeString& debug(UnicodeString& result) const {
      return debugClass(result);
    }

    virtual UnicodeString& debugClass(UnicodeString& result) const {
      return result.append((UnicodeString)"Key");
    }
#endif
};

/**
 * <p>Abstract implementation of a notification facility.  Clients add
 * EventListeners with addListener and remove them with removeListener.
 * Notifiers call notifyChanged when they wish to notify listeners.
 * This queues the listener list on the notification thread, which
 * eventually dequeues the list and calls notifyListener on each
 * listener in the list.</p>
 *
 * <p>Subclasses override acceptsListener and notifyListener 
 * to add type-safe notification.  AcceptsListener should return
 * true if the listener is of the appropriate type; ICUNotifier
 * itself will ensure the listener is non-null and that the
 * identical listener is not already registered with the Notifier.
 * NotifyListener should cast the listener to the appropriate 
 * type and call the appropriate method on the listener.
 */

class U_COMMON_API ICUNotifier : public UMemory  {
private: UVector* listeners;
         
public: 
    ICUNotifier(void);
    
    virtual ~ICUNotifier(void);
    
    /**
     * Add a listener to be notified when notifyChanged is called.
     * The listener must not be null. AcceptsListener must return
     * true for the listener.  Attempts to concurrently
     * register the identical listener more than once will be
     * silently ignored.  
     */
    virtual void addListener(const EventListener* l, UErrorCode& status);
    
    /**
     * Stop notifying this listener.  The listener must
     * not be null.  Attemps to remove a listener that is
     * not registered will be silently ignored.
     */
    virtual void removeListener(const EventListener* l, UErrorCode& status);
    
    /**
     * ICU doesn't spawn its own threads.  All listeners are notified in
     * the thread of the caller.  Misbehaved listeners can therefore
     * indefinitely block the calling thread.  Callers should beware of
     * deadlock situations.  
     */
    virtual void notifyChanged(void);
    
protected: 
    /**
     * Subclasses implement this to return TRUE if the listener is
     * of the appropriate type.
     */
    virtual UBool acceptsListener(const EventListener& l) const = 0;
    
    /**
     * Subclasses implement this to notify the listener.
     */
    virtual void notifyListener(EventListener& l) const = 0;
};

U_NAMESPACE_END

/* UCONFIG_NO_SERVICE */
#endif

/* ICUNOTIF_H */


class ICUServiceTest;

U_NAMESPACE_BEGIN

class ICUServiceKey;
class ICUServiceFactory;
class SimpleFactory;
class ServiceListener;
class ICUService;

class DNCache;

/*******************************************************************
 * ICUServiceKey
 */

/**
 * <p>ICUServiceKeys are used to communicate with factories to
 * generate an instance of the service.  ICUServiceKeys define how
 * ids are canonicalized, provide both a current id and a current
 * descriptor to use in querying the cache and factories, and
 * determine the fallback strategy.</p>
 *
 * <p>ICUServiceKeys provide both a currentDescriptor and a currentID.
 * The descriptor contains an optional prefix, followed by '/'
 * and the currentID.  Factories that handle complex keys,
 * for example number format factories that generate multiple
 * kinds of formatters for the same locale, use the descriptor 
 * to provide a fully unique identifier for the service object, 
 * while using the currentID (in this case, the locale string),
 * as the visible IDs that can be localized.</p>
 *
 * <p>The default implementation of ICUServiceKey has no fallbacks and
 * has no custom descriptors.</p> 
 */
class U_COMMON_API ICUServiceKey : public UObject {
 private: 
  const UnicodeString _id;

 protected:
  static const UChar PREFIX_DELIMITER;

 public:

  /**
   * <p>Construct a key from an id.</p>
   *
   * @param id the ID from which to construct the key.
   */
  ICUServiceKey(const UnicodeString& id);

  /**
   * <p>Virtual destructor.</p>
   */
  virtual ~ICUServiceKey();

 /**
  * <p>Return the original ID used to construct this key.</p>
  *
  * @return the ID used to construct this key.
  */
  virtual const UnicodeString& getID() const;

 /**
  * <p>Return the canonical version of the original ID.  This implementation
  * appends the original ID to result.  Result is returned as a convenience.</p>
  *
  * @param result the output parameter to which the id will be appended.
  * @return the modified result.
  */
  virtual UnicodeString& canonicalID(UnicodeString& result) const;

 /**
  * <p>Return the (canonical) current ID.  This implementation appends
  * the canonical ID to result.  Result is returned as a convenience.</p>
  *
  * @param result the output parameter to which the current id will be appended.
  * @return the modified result.  
  */
  virtual UnicodeString& currentID(UnicodeString& result) const;

 /**
  * <p>Return the current descriptor.  This implementation appends
  * the current descriptor to result.  Result is returned as a convenience.</p>
  *
  * <p>The current descriptor is used to fully
  * identify an instance of the service in the cache.  A
  * factory may handle all descriptors for an ID, or just a
  * particular descriptor.  The factory can either parse the
  * descriptor or use custom API on the key in order to
  * instantiate the service.</p>
  *
  * @param result the output parameter to which the current id will be appended.
  * @return the modified result.  
  */
  virtual UnicodeString& currentDescriptor(UnicodeString& result) const;

 /**
  * <p>If the key has a fallback, modify the key and return true,
  * otherwise return false.  The current ID will change if there
  * is a fallback.  No currentIDs should be repeated, and fallback
  * must eventually return false.  This implementation has no fallbacks
  * and always returns false.</p>
  *
  * @return TRUE if the ICUServiceKey changed to a valid fallback value.
  */
  virtual UBool fallback();

 /**
  * <p>Return TRUE if a key created from id matches, or would eventually
  * fallback to match, the canonical ID of this ICUServiceKey.</p>
  *
  * @param id the id to test.
  * @return TRUE if this ICUServiceKey's canonical ID is a fallback of id.
  */
  virtual UBool isFallbackOf(const UnicodeString& id) const;

 /**
  * <p>Return the prefix.  This implementation leaves result unchanged.
  * Result is returned as a convenience.</p>
  *
  * @param result the output parameter to which the prefix will be appended.
  * @return the modified result.
  */
  virtual UnicodeString& prefix(UnicodeString& result) const;

 /**
  * <p>A utility to parse the prefix out of a descriptor string.  Only
  * the (undelimited) prefix, if any, remains in result.  Result is returned as a 
  * convenience.</p>
  *
  * @param result an input/output parameter that on entry is a descriptor, and 
  * on exit is the prefix of that descriptor.
  * @return the modified result.
  */
  static UnicodeString& parsePrefix(UnicodeString& result);

  /**
  * <p>A utility to parse the suffix out of a descriptor string.  Only
  * the (undelimited) suffix, if any, remains in result.  Result is returned as a 
  * convenience.</p>
  *
  * @param result an input/output parameter that on entry is a descriptor, and 
  * on exit is the suffix of that descriptor.
  * @return the modified result.
  */
  static UnicodeString& parseSuffix(UnicodeString& result);

public:
  /**
   * UObject RTTI boilerplate.
   */
  static UClassID U_EXPORT2 getStaticClassID();

  /**
   * UObject RTTI boilerplate.
   */
  virtual UClassID getDynamicClassID() const;

#ifdef SERVICE_DEBUG
 public:
  virtual UnicodeString& debug(UnicodeString& result) const;
  virtual UnicodeString& debugClass(UnicodeString& result) const;
#endif

};

 /*******************************************************************
  * ICUServiceFactory
  */

 /**
  * <p>An implementing ICUServiceFactory generates the service objects maintained by the
  * service.  A factory generates a service object from a key,
  * updates id->factory mappings, and returns the display name for
  * a supported id.</p>
  */
class U_COMMON_API ICUServiceFactory : public UObject {
 public:
    virtual ~ICUServiceFactory();

    /**
     * <p>Create a service object from the key, if this factory
     * supports the key.  Otherwise, return NULL.</p>
     *
     * <p>If the factory supports the key, then it can call
     * the service's getKey(ICUServiceKey, String[], ICUServiceFactory) method
     * passing itself as the factory to get the object that
     * the service would have created prior to the factory's
     * registration with the service.  This can change the
     * key, so any information required from the key should
     * be extracted before making such a callback.</p>
     *
     * @param key the service key.
     * @param service the service with which this factory is registered.
     * @param status the error code status.
     * @return the service object, or NULL if the factory does not support the key.
     */
    virtual UObject* create(const ICUServiceKey& key, const ICUService* service, UErrorCode& status) const = 0;

    /**
     * <p>Update result to reflect the IDs (not descriptors) that this
     * factory publicly handles.  Result contains mappings from ID to
     * factory.  On entry it will contain all (visible) mappings from
     * previously-registered factories.</p>
     *
     * <p>This function, together with getDisplayName, are used to
     * support ICUService::getDisplayNames.  The factory determines
     * which IDs (of those it supports) it will make visible, and of
     * those, which it will provide localized display names for.  In
     * most cases it will register mappings from all IDs it supports
     * to itself.</p>
     *
     * @param result the mapping table to update.
     * @param status the error code status.
     */
    virtual void updateVisibleIDs(Hashtable& result, UErrorCode& status) const = 0;

    /**
     * <p>Return, in result, the display name of the id in the provided locale.
     * This is an id, not a descriptor.  If the id is 
     * not visible, sets result to bogus.  If the
     * incoming result is bogus, it remains bogus.  Result is returned as a
     * convenience.  Results are not defined if id is not one supported by this
         * factory.</p>
     *
     * @param id a visible id supported by this factory.
     * @param locale the locale for which to generate the corresponding localized display name.
     * @param result output parameter to hold the display name.
     * @return result.
     */
    virtual UnicodeString& getDisplayName(const UnicodeString& id, const Locale& locale, UnicodeString& result) const = 0;
};

/*
 ******************************************************************
 */

 /**
  * <p>A default implementation of factory.  This provides default
  * implementations for subclasses, and implements a singleton
  * factory that matches a single ID and returns a single
  * (possibly deferred-initialized) instance.  This implements
  * updateVisibleIDs to add a mapping from its ID to itself
  * if visible is true, or to remove any existing mapping
  * for its ID if visible is false.  No localization of display
  * names is performed.</p>
  */
class U_COMMON_API SimpleFactory : public ICUServiceFactory {
 protected:
  UObject* _instance;
  const UnicodeString _id;
  const UBool _visible;

 public:
  /**
   * <p>Construct a SimpleFactory that maps a single ID to a single 
   * service instance.  If visible is TRUE, the ID will be visible.
   * The instance must not be NULL.  The SimpleFactory will adopt
   * the instance, which must not be changed subsequent to this call.</p>
   *
   * @param instanceToAdopt the service instance to adopt.
   * @param id the ID to assign to this service instance.
   * @param visible if TRUE, the ID will be visible.
   */
  SimpleFactory(UObject* instanceToAdopt, const UnicodeString& id, UBool visible = TRUE);

  /**
   * <p>Destructor.</p>
   */
  virtual ~SimpleFactory();

  /**
   * <p>This implementation returns a clone of the service instance if the factory's ID is equal to
   * the key's currentID.  Service and prefix are ignored.</p>
   *
   * @param key the service key.
   * @param service the service with which this factory is registered.
   * @param status the error code status.
   * @return the service object, or NULL if the factory does not support the key.
   */
  virtual UObject* create(const ICUServiceKey& key, const ICUService* service, UErrorCode& status) const;

  /**
   * <p>This implementation adds a mapping from ID -> this to result if visible is TRUE, 
   * otherwise it removes ID from result.</p>
   *
   * @param result the mapping table to update.
   * @param status the error code status.
   */
  virtual void updateVisibleIDs(Hashtable& result, UErrorCode& status) const;

  /**
   * <p>This implementation returns the factory ID if it equals id and visible is TRUE,
   * otherwise it returns the empty string.  (This implementation provides
   * no localized id information.)</p>
   *
   * @param id a visible id supported by this factory.
   * @param locale the locale for which to generate the corresponding localized display name.
   * @param result output parameter to hold the display name.
   * @return result.
   */
  virtual UnicodeString& getDisplayName(const UnicodeString& id, const Locale& locale, UnicodeString& result) const;

public:
 /**
  * UObject RTTI boilerplate.
  */
  static UClassID U_EXPORT2 getStaticClassID();

 /**
  * UObject RTTI boilerplate.
  */
  virtual UClassID getDynamicClassID() const;

#ifdef SERVICE_DEBUG
 public:
  virtual UnicodeString& debug(UnicodeString& toAppendTo) const;
  virtual UnicodeString& debugClass(UnicodeString& toAppendTo) const;
#endif

};

/*
 ******************************************************************
 */

/**
 * <p>ServiceListener is the listener that ICUService provides by default.
 * ICUService will notifiy this listener when factories are added to
 * or removed from the service.  Subclasses can provide
 * different listener interfaces that extend EventListener, and modify
 * acceptsListener and notifyListener as appropriate.</p>
 */
class U_COMMON_API ServiceListener : public EventListener {
public:
    virtual ~ServiceListener();

    /**
     * <p>This method is called when the service changes. At the time of the
     * call this listener is registered with the service.  It must
     * not modify the notifier in the context of this call.</p>
     * 
     * @param service the service that changed.
     */
    virtual void serviceChanged(const ICUService& service) const = 0;
    
public:
    /**
     * UObject RTTI boilerplate.
     */
    static UClassID U_EXPORT2 getStaticClassID();
    
    /**
     * UObject RTTI boilerplate.
     */
    virtual UClassID getDynamicClassID() const;
    
};

/*
 ******************************************************************
 */

/**
 * <p>A StringPair holds a displayName/ID pair.  ICUService uses it
 * as the array elements returned by getDisplayNames.
 */
class U_COMMON_API StringPair : public UMemory {
public:
  /**
   * <p>The display name of the pair.</p>
   */
  const UnicodeString displayName;

  /**
   * <p>The ID of the pair.</p>
   */
  const UnicodeString id;

  /**
   * <p>Creates a string pair from a displayName and an ID.</p>
   *
   * @param displayName the displayName.
   * @param id the ID.
   * @param status the error code status.
   * @return a StringPair if the creation was successful, otherwise NULL.
   */
  static StringPair* create(const UnicodeString& displayName, 
                            const UnicodeString& id,
                            UErrorCode& status);

  /**
   * <p>Return TRUE if either string of the pair is bogus.</p>
   * @return TRUE if either string of the pair is bogus.
   */
  UBool isBogus() const;

private:
  StringPair(const UnicodeString& displayName, const UnicodeString& id);
};

/*******************************************************************
 * ICUService
 */

 /**
 * <p>A Service provides access to service objects that implement a
 * particular service, e.g. transliterators.  Users provide a String
 * id (for example, a locale string) to the service, and get back an
 * object for that id.  Service objects can be any kind of object.  A
 * new service object is returned for each query. The caller is
 * responsible for deleting it.</p>
 *
 * <p>Services 'canonicalize' the query ID and use the canonical ID to
 * query for the service.  The service also defines a mechanism to
 * 'fallback' the ID multiple times.  Clients can optionally request
 * the actual ID that was matched by a query when they use an ID to
 * retrieve a service object.</p>
 *
 * <p>Service objects are instantiated by ICUServiceFactory objects
 * registered with the service.  The service queries each
 * ICUServiceFactory in turn, from most recently registered to
 * earliest registered, until one returns a service object.  If none
 * responds with a service object, a fallback ID is generated, and the
 * process repeats until a service object is returned or until the ID
 * has no further fallbacks.</p>
 *
 * <p>In ICU 2.4, UObject (the base class of service instances) does
 * not define a polymorphic clone function.  ICUService uses clones to
 * manage ownership.  Thus, for now, ICUService defines an abstract
 * method, cloneInstance, that clients must implement to create clones
 * of the service instances.  This may change in future releases of
 * ICU.</p>
 *
 * <p>ICUServiceFactories can be dynamically registered and
 * unregistered with the service.  When registered, an
 * ICUServiceFactory is installed at the head of the factory list, and
 * so gets 'first crack' at any keys or fallback keys.  When
 * unregistered, it is removed from the service and can no longer be
 * located through it.  Service objects generated by this factory and
 * held by the client are unaffected.</p>
 *
 * <p>If a service has variants (e.g., the different variants of
 * BreakIterator) an ICUServiceFactory can use the prefix of the
 * ICUServiceKey to determine the variant of a service to generate.
 * If it does not support all variants, it can request
 * previously-registered factories to handle the ones it does not
 * support.</p>
 *
 * <p>ICUService uses ICUServiceKeys to query factories and perform
 * fallback.  The ICUServiceKey defines the canonical form of the ID,
 * and implements the fallback strategy.  Custom ICUServiceKeys can be
 * defined that parse complex IDs into components that
 * ICUServiceFactories can more easily use.  The ICUServiceKey can
 * cache the results of this parsing to save repeated effort.
 * ICUService provides convenience APIs that take UnicodeStrings and
 * generate default ICUServiceKeys for use in querying.</p>
 *
 * <p>ICUService provides API to get the list of IDs publicly
 * supported by the service (although queries aren't restricted to
 * this list).  This list contains only 'simple' IDs, and not fully
 * unique IDs.  ICUServiceFactories are associated with each simple ID
 * and the responsible factory can also return a human-readable
 * localized version of the simple ID, for use in user interfaces.
 * ICUService can also provide an array of the all the localized
 * visible IDs and their corresponding internal IDs.</p>
 *
 * <p>ICUService implements ICUNotifier, so that clients can register
 * to receive notification when factories are added or removed from
 * the service.  ICUService provides a default EventListener
 * subinterface, ServiceListener, which can be registered with the
 * service.  When the service changes, the ServiceListener's
 * serviceChanged method is called with the service as the
 * argument.</p>
 *
 * <p>The ICUService API is both rich and generic, and it is expected
 * that most implementations will statically 'wrap' ICUService to
 * present a more appropriate API-- for example, to declare the type
 * of the objects returned from get, to limit the factories that can
 * be registered with the service, or to define their own listener
 * interface with a custom callback method.  They might also customize
 * ICUService by overriding it, for example, to customize the
 * ICUServiceKey and fallback strategy.  ICULocaleService is a
 * subclass of ICUService that uses Locale names as IDs and uses
 * ICUServiceKeys that implement the standard resource bundle fallback
 * strategy.  Most clients will wish to subclass it instead of
 * ICUService.</p> 
 */
class U_COMMON_API ICUService : public ICUNotifier {
 protected: 
    /**
     * Name useful for debugging.
     */
    const UnicodeString name;

 private:

    /**
     * Timestamp so iterators can be fail-fast.
     */
    uint32_t timestamp;

    /**
     * All the factories registered with this service.
     */
    UVector* factories;

    /**
     * The service cache.
     */
    Hashtable* serviceCache;

    /**
     * The ID cache.
     */
    Hashtable* idCache;

    /**
     * The name cache.
     */
    DNCache* dnCache;

    /**
     * Constructor.
     */
 public:
    /**
     * <p>Construct a new ICUService.</p>
     */
    ICUService();

    /**
     * <p>Construct with a name (useful for debugging).</p>
     *
     * @param name a name to use in debugging.
     */
    ICUService(const UnicodeString& name);

    /**
     * <p>Destructor.</p>
     */
    virtual ~ICUService();

    /**
     * <p>Return the name of this service. This will be the empty string if none was assigned.
     * Returns result as a convenience.</p>
     *
     * @param result an output parameter to contain the name of this service.
     * @return the name of this service.
     */
    UnicodeString& getName(UnicodeString& result) const;

    /**
     * <p>Convenience override for get(ICUServiceKey&, UnicodeString*). This uses
     * createKey to create a key for the provided descriptor.</p>
     *
     * @param descriptor the descriptor.
     * @param status the error code status.
     * @return the service instance, or NULL.
     */
    UObject* get(const UnicodeString& descriptor, UErrorCode& status) const;

    /**
     * <p>Convenience override for get(ICUServiceKey&, UnicodeString*).  This uses
     * createKey to create a key from the provided descriptor.</p>
     *
     * @param descriptor the descriptor.
     * @param actualReturn a pointer to a UnicodeString to hold the matched descriptor, or NULL.
     * @param status the error code status.
     * @return the service instance, or NULL.
     */
    UObject* get(const UnicodeString& descriptor, UnicodeString* actualReturn, UErrorCode& status) const;

    /**
     * <p>Convenience override for get(ICUServiceKey&, UnicodeString*).</p>
     *
     * @param key the key.
     * @param status the error code status.
     * @return the service instance, or NULL.
     */
    UObject* getKey(ICUServiceKey& key, UErrorCode& status) const;

    /**
     * <p>Given a key, return a service object, and, if actualReturn
     * is not NULL, the descriptor with which it was found in the
     * first element of actualReturn.  If no service object matches
     * this key, returns NULL and leaves actualReturn unchanged.</p>
     *
     * <p>This queries the cache using the key's descriptor, and if no
     * object in the cache matches, tries the key on each
     * registered factory, in order.  If none generates a service
     * object for the key, repeats the process with each fallback of
     * the key, until either a factory returns a service object, or the key
     * has no fallback.  If no object is found, the result of handleDefault
     * is returned.</p>
     *
     * <p>Subclasses can override this method to further customize the 
     * result before returning it.
     *
     * @param key the key.
     * @param actualReturn a pointer to a UnicodeString to hold the matched descriptor, or NULL.
     * @param status the error code status.
     * @return the service instance, or NULL.
     */
    virtual UObject* getKey(ICUServiceKey& key, UnicodeString* actualReturn, UErrorCode& status) const;

    /**
     * <p>This version of getKey is only called by ICUServiceFactories within the scope
     * of a previous getKey call, to determine what previously-registered factories would
     * have returned.  For details, see getKey(ICUServiceKey&, UErrorCode&).  Subclasses
     * should not call it directly, but call through one of the other get functions.</p>
     * 
     * @param key the key.
     * @param actualReturn a pointer to a UnicodeString to hold the matched descriptor, or NULL.
     * @param factory the factory making the recursive call.
     * @param status the error code status.
     * @return the service instance, or NULL.
     */
    UObject* getKey(ICUServiceKey& key, UnicodeString* actualReturn, const ICUServiceFactory* factory, UErrorCode& status) const;

    /**
     * <p>Convenience override for getVisibleIDs(String) that passes null
     * as the fallback, thus returning all visible IDs.</p>
     *
     * @param result a vector to hold the returned IDs.
     * @param status the error code status.
     * @return the result vector.
     */
    UVector& getVisibleIDs(UVector& result, UErrorCode& status) const;

    /**
     * <p>Return a snapshot of the visible IDs for this service.  This
     * list will not change as ICUServiceFactories are added or removed, but the
     * supported IDs will, so there is no guarantee that all and only
     * the IDs in the returned list will be visible and supported by the
     * service in subsequent calls.</p>
     *
     * <p>The IDs are returned as pointers to UnicodeStrings.  The
     * caller owns the IDs.  Previous contents of result are discarded before
     * new elements, if any, are added.</p>
     *
     * <p>matchID is passed to createKey to create a key.  If the key
     * is not NULL, its isFallbackOf method is used to filter out IDs
     * that don't match the key or have it as a fallback.</p>
     *
     * @param result a vector to hold the returned IDs.
     * @param matchID an ID used to filter the result, or NULL if all IDs are desired.
     * @param status the error code status.
     * @return the result vector.
     */
    UVector& getVisibleIDs(UVector& result, const UnicodeString* matchID, UErrorCode& status) const;

    /**
     * <p>Convenience override for getDisplayName(const UnicodeString&, const Locale&, UnicodeString&) that
     * uses the current default locale.</p>
     *
     * @param id the ID for which to retrieve the localized displayName.
     * @param result an output parameter to hold the display name.
     * @return the modified result.
     */
    UnicodeString& getDisplayName(const UnicodeString& id, UnicodeString& result) const;

    /**
     * <p>Given a visible ID, return the display name in the requested locale.
     * If there is no directly supported ID corresponding to this ID, result is
     * set to bogus.</p>
     *
     * @param id the ID for which to retrieve the localized displayName.
     * @param result an output parameter to hold the display name.
     * @param locale the locale in which to localize the ID.
     * @return the modified result.
     */
    UnicodeString& getDisplayName(const UnicodeString& id, UnicodeString& result, const Locale& locale) const;

    /**
     * <p>Convenience override of getDisplayNames(const Locale&, const UnicodeString*) that 
     * uses the current default Locale as the locale and NULL for
     * the matchID.</p>
     *
     * @param result a vector to hold the returned displayName/id StringPairs.
     * @param status the error code status.
     * @return the modified result vector.
     */
    UVector& getDisplayNames(UVector& result, UErrorCode& status) const;

    /**
     * <p>Convenience override of getDisplayNames(const Locale&, const UnicodeString*) that 
     * uses NULL for the matchID.</p>
     *
     * @param result a vector to hold the returned displayName/id StringPairs.
     * @param locale the locale in which to localize the ID.
     * @param status the error code status.
     * @return the modified result vector.
     */
    UVector& getDisplayNames(UVector& result, const Locale& locale, UErrorCode& status) const;

    /**
     * <p>Return a snapshot of the mapping from display names to visible
     * IDs for this service.  This set will not change as factories
     * are added or removed, but the supported IDs will, so there is
     * no guarantee that all and only the IDs in the returned map will
     * be visible and supported by the service in subsequent calls,
     * nor is there any guarantee that the current display names match
     * those in the result.</p>
     *
     * <p>The names are returned as pointers to StringPairs, which
     * contain both the displayName and the corresponding ID.  The
     * caller owns the StringPairs.  Previous contents of result are
     * discarded before new elements, if any, are added.</p>
     *
     * <p>matchID is passed to createKey to create a key.  If the key
     * is not NULL, its isFallbackOf method is used to filter out IDs
     * that don't match the key or have it as a fallback.</p>
     *
     * @param result a vector to hold the returned displayName/id StringPairs.
     * @param locale the locale in which to localize the ID.
     * @param matchID an ID used to filter the result, or NULL if all IDs are desired.
     * @param status the error code status.
     * @return the result vector.  */
    UVector& getDisplayNames(UVector& result,
                             const Locale& locale, 
                             const UnicodeString* matchID, 
                             UErrorCode& status) const;

    /**
     * <p>A convenience override of registerInstance(UObject*, const UnicodeString&, UBool)
     * that defaults visible to TRUE.</p>
     *
     * @param objToAdopt the object to register and adopt.
     * @param id the ID to assign to this object.
     * @param status the error code status.
     * @return a registry key that can be passed to unregister to unregister
     * (and discard) this instance.
     */
    URegistryKey registerInstance(UObject* objToAdopt, const UnicodeString& id, UErrorCode& status);

    /**
     * <p>Register a service instance with the provided ID.  The ID will be 
     * canonicalized.  The canonicalized ID will be returned by
     * getVisibleIDs if visible is TRUE.  The service instance will be adopted and
     * must not be modified subsequent to this call.</p>
     *
     * <p>This issues a serviceChanged notification to registered listeners.</p>
     *
     * <p>This implementation wraps the object using
     * createSimpleFactory, and calls registerFactory.</p>
     *
     * @param objToAdopt the object to register and adopt.
     * @param id the ID to assign to this object.
     * @param visible TRUE if getVisibleIDs is to return this ID.
     * @param status the error code status.
     * @return a registry key that can be passed to unregister() to unregister
     * (and discard) this instance.
     */
    virtual URegistryKey registerInstance(UObject* objToAdopt, const UnicodeString& id, UBool visible, UErrorCode& status);

    /**
     * <p>Register an ICUServiceFactory.  Returns a registry key that
     * can be used to unregister the factory.  The factory
     * must not be modified subsequent to this call.  The service owns
     * all registered factories. In case of an error, the factory is
     * deleted.</p>
     *
     * <p>This issues a serviceChanged notification to registered listeners.</p>
     *
     * <p>The default implementation accepts all factories.</p>
     *
     * @param factoryToAdopt the factory to register and adopt.
     * @param status the error code status.
     * @return a registry key that can be passed to unregister to unregister
     * (and discard) this factory.
     */
    virtual URegistryKey registerFactory(ICUServiceFactory* factoryToAdopt, UErrorCode& status);

    /**
     * <p>Unregister a factory using a registry key returned by
     * registerInstance or registerFactory.  After a successful call,
     * the factory will be removed from the service factory list and
     * deleted, and the key becomes invalid.</p>
     *
     * <p>This issues a serviceChanged notification to registered
     * listeners.</p>
     *
     * @param rkey the registry key.
     * @param status the error code status.  
     * @return TRUE if the call successfully unregistered the factory.
     */
    virtual UBool unregister(URegistryKey rkey, UErrorCode& status);

    /**
     * </p>Reset the service to the default factories.  The factory
     * lock is acquired and then reInitializeFactories is called.</p>
     *
     * <p>This issues a serviceChanged notification to registered listeners.</p>
     */
    virtual void reset(void);

    /**
     * <p>Return TRUE if the service is in its default state.</p>
     *
     * <p>The default implementation returns TRUE if there are no 
     * factories registered.</p>
     */
    virtual UBool isDefault(void) const;

    /**
     * <p>Create a key from an ID.  If ID is NULL, returns NULL.</p>
     *
     * <p>The default implementation creates an ICUServiceKey instance.
     * Subclasses can override to define more useful keys appropriate
     * to the factories they accept.</p>
     *
     * @param a pointer to the ID for which to create a default ICUServiceKey.
     * @param status the error code status.
     * @return the ICUServiceKey corresponding to ID, or NULL.
     */
    virtual ICUServiceKey* createKey(const UnicodeString* id, UErrorCode& status) const;

    /**
     * <p>Clone object so that caller can own the copy.  In ICU2.4, UObject doesn't define
     * clone, so we need an instance-aware method that knows how to do this.
     * This is public so factories can call it, but should really be protected.</p>
     *
     * @param instance the service instance to clone.
     * @return a clone of the passed-in instance, or NULL if cloning was unsuccessful.
     */
    virtual UObject* cloneInstance(UObject* instance) const = 0;


    /************************************************************************
     * Subclassing API
     */

 protected:

    /**
     * <p>Create a factory that wraps a single service object.  Called by registerInstance.</p>
     *
     * <p>The default implementation returns an instance of SimpleFactory.</p>
     *
     * @param instanceToAdopt the service instance to adopt.
     * @param id the ID to assign to this service instance.
     * @param visible if TRUE, the ID will be visible.
     * @param status the error code status.
     * @return an instance of ICUServiceFactory that maps this instance to the provided ID.
     */
    virtual ICUServiceFactory* createSimpleFactory(UObject* instanceToAdopt, const UnicodeString& id, UBool visible, UErrorCode& status);

    /**
     * <p>Reinitialize the factory list to its default state.  After this call, isDefault()
     * must return TRUE.</p>
     *
     * <p>This issues a serviceChanged notification to registered listeners.</p>
     *
     * <p>The default implementation clears the factory list.
     * Subclasses can override to provide other default initialization
     * of the factory list.  Subclasses must not call this method
     * directly, since it must only be called while holding write
     * access to the factory list.</p>
     */
    virtual void reInitializeFactories(void);

    /**
     * <p>Default handler for this service if no factory in the factory list
     * handled the key passed to getKey.</p>
     *
     * <p>The default implementation returns NULL.</p>
     *
     * @param key the key.
     * @param actualReturn a pointer to a UnicodeString to hold the matched descriptor, or NULL.
     * @param status the error code status.
     * @return the service instance, or NULL.
     */
    virtual UObject* handleDefault(const ICUServiceKey& key, UnicodeString* actualReturn, UErrorCode& status) const;

    /**
     * <p>Clear caches maintained by this service.</p>
     *
     * <p>Subclasses can override if they implement additional caches
     * that need to be cleared when the service changes.  Subclasses
     * should generally not call this method directly, as it must only
     * be called while synchronized on the factory lock.</p>
     */
    virtual void clearCaches(void);

    /**
     * <p>Return true if the listener is accepted.</p>
     *
     * <p>The default implementation accepts the listener if it is
     * a ServiceListener.  Subclasses can override this to accept
     * different listeners.</p>
     *
     * @param l the listener to test.
     * @return TRUE if the service accepts the listener.
     */
    virtual UBool acceptsListener(const EventListener& l) const;

    /**
     * <p>Notify the listener of a service change.</p>
     *
     * <p>The default implementation assumes a ServiceListener.
     * If acceptsListener has been overridden to accept different
     * listeners, this should be overridden as well.</p>
     *
     * @param l the listener to notify.
     */
    virtual void notifyListener(EventListener& l) const;

    /************************************************************************
     * Utilities for subclasses.
     */

    /**
     * <p>Clear only the service cache.</p>
     *
     * <p>This can be called by subclasses when a change affects the service
     * cache but not the ID caches, e.g., when the default locale changes
     * the resolution of IDs also changes, requiring the cache to be
     * flushed, but not the visible IDs themselves.</p>
     */
    void clearServiceCache(void);

    /**
     * <p>Return a map from visible IDs to factories.
     * This must only be called when the mutex is held.</p>
     *
     * @param status the error code status.
     * @return a Hashtable containing mappings from visible
     * IDs to factories.
     */
    const Hashtable* getVisibleIDMap(UErrorCode& status) const;

    /**
     * <p>Allow subclasses to read the time stamp.</p>
     *
     * @return the timestamp.
     */
    int32_t getTimestamp(void) const;

    /**
     * <p>Return the number of registered factories.</p>
     *
     * @return the number of factories registered at the time of the call.
     */
    int32_t countFactories(void) const;

private:

    friend class ::ICUServiceTest; // give tests access to countFactories.
};

U_NAMESPACE_END

    /* UCONFIG_NO_SERVICE */
#endif

    /* ICUSERV_H */



#undef SERVICE_REFCOUNT

// in case we use the refcount stuff

U_NAMESPACE_BEGIN

/*
******************************************************************
*/

const UChar ICUServiceKey::PREFIX_DELIMITER = 0x002F;   /* '/' */

ICUServiceKey::ICUServiceKey(const UnicodeString& id) 
: _id(id) {
}

ICUServiceKey::~ICUServiceKey() 
{
}

const UnicodeString& 
ICUServiceKey::getID() const 
{
    return _id;
}

UnicodeString& 
ICUServiceKey::canonicalID(UnicodeString& result) const 
{
    return result.append(_id);
}

UnicodeString& 
ICUServiceKey::currentID(UnicodeString& result) const 
{
    return canonicalID(result);
}

UnicodeString& 
ICUServiceKey::currentDescriptor(UnicodeString& result) const 
{
    prefix(result);
    result.append(PREFIX_DELIMITER);
    return currentID(result);
}

UBool 
ICUServiceKey::fallback() 
{
    return FALSE;
}

UBool 
ICUServiceKey::isFallbackOf(const UnicodeString& id) const 
{
    return id == _id;
}

UnicodeString& 
ICUServiceKey::prefix(UnicodeString& result) const 
{
    return result;
}

UnicodeString& 
ICUServiceKey::parsePrefix(UnicodeString& result) 
{
    int32_t n = result.indexOf(PREFIX_DELIMITER);
    if (n < 0) {
        n = 0;
    }
    result.remove(n);
    return result;
}

UnicodeString& 
ICUServiceKey::parseSuffix(UnicodeString& result) 
{
    int32_t n = result.indexOf(PREFIX_DELIMITER);
    if (n >= 0) {
        result.remove(0, n+1);
    }
    return result;
}

#ifdef SERVICE_DEBUG
UnicodeString& 
ICUServiceKey::debug(UnicodeString& result) const 
{
    debugClass(result);
    result.append((UnicodeString)" id: ");
    result.append(_id);
    return result;
}

UnicodeString& 
ICUServiceKey::debugClass(UnicodeString& result) const 
{
    return result.append((UnicodeString)"ICUServiceKey");
}
#endif

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(ICUServiceKey)

/*
******************************************************************
*/

ICUServiceFactory::~ICUServiceFactory() {}

SimpleFactory::SimpleFactory(UObject* instanceToAdopt, const UnicodeString& id, UBool visible) 
: _instance(instanceToAdopt), _id(id), _visible(visible)
{
}

SimpleFactory::~SimpleFactory() 
{
    delete _instance;
}

UObject* 
SimpleFactory::create(const ICUServiceKey& key, const ICUService* service, UErrorCode& status) const 
{
    if (U_SUCCESS(status)) {
        UnicodeString temp;
        if (_id == key.currentID(temp)) {
            return service->cloneInstance(_instance); 
        }
    }
    return NULL;
}

void 
SimpleFactory::updateVisibleIDs(Hashtable& result, UErrorCode& status) const 
{
    if (_visible) {
        result.put(_id, (void*)this, status); // cast away const
    } else {
        result.remove(_id);
    }
}

UnicodeString& 
SimpleFactory::getDisplayName(const UnicodeString& id, const Locale& /* locale */, UnicodeString& result) const 
{
    if (_visible && _id == id) {
        result = _id;
    } else {
        result.setToBogus();
    }
    return result;
}

#ifdef SERVICE_DEBUG
UnicodeString& 
SimpleFactory::debug(UnicodeString& toAppendTo) const 
{
    debugClass(toAppendTo);
    toAppendTo.append((UnicodeString)" id: ");
    toAppendTo.append(_id);
    toAppendTo.append((UnicodeString)", visible: ");
    toAppendTo.append(_visible ? (UnicodeString)"T" : (UnicodeString)"F");
    return toAppendTo;
}

UnicodeString& 
SimpleFactory::debugClass(UnicodeString& toAppendTo) const 
{
    return toAppendTo.append((UnicodeString)"SimpleFactory");
}
#endif

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(SimpleFactory)

/*
******************************************************************
*/

ServiceListener::~ServiceListener() {}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(ServiceListener)

/*
******************************************************************
*/

// Record the actual id for this service in the cache, so we can return it
// even if we succeed later with a different id.
class CacheEntry : public UMemory {
private:
    int32_t refcount;

public:
    UnicodeString actualDescriptor;
    UObject* service;

    /**
    * Releases a reference to the shared resource.
    */
    ~CacheEntry() {
        delete service;
    }

    CacheEntry(const UnicodeString& _actualDescriptor, UObject* _service) 
        : refcount(1), actualDescriptor(_actualDescriptor), service(_service) {
    }

    /**
    * Instantiation creates an initial reference, so don't call this
    * unless you're creating a new pointer to this.  Management of
    * that pointer will have to know how to deal with refcounts.  
    * Return true if the resource has not already been released.
    */
    CacheEntry* ref() {
        ++refcount;
        return this;
    }

    /**
    * Destructions removes a reference, so don't call this unless
    * you're removing pointer to this somewhere.  Management of that
    * pointer will have to know how to deal with refcounts.  Once
    * the refcount drops to zero, the resource is released.  Return
    * false if the resouce has been released.
    */
    CacheEntry* unref() {
        if ((--refcount) == 0) {
            delete this;
            return NULL;
        }
        return this;
    }

    /**
    * Return TRUE if there is at least one reference to this and the
    * resource has not been released.
    */
    UBool isShared() const {
        return refcount > 1;
    }
};

// UObjectDeleter for serviceCache
U_CDECL_BEGIN
static void U_CALLCONV
cacheDeleter(void* obj) {
    U_NAMESPACE_USE ((CacheEntry*)obj)->unref();
}

/**
* Deleter for UObjects
*/
static void U_CALLCONV
deleteUObject(void *obj) {
    U_NAMESPACE_USE delete (UObject*) obj;
}
U_CDECL_END

/*
******************************************************************
*/

class DNCache : public UMemory {
public:
    Hashtable cache;
    const Locale locale;

    DNCache(const Locale& _locale) 
        : cache(), locale(_locale) 
    {
        // cache.setKeyDeleter(uprv_deleteUObject);
    }
};


/*
******************************************************************
*/

StringPair* 
StringPair::create(const UnicodeString& displayName, 
                   const UnicodeString& id,
                   UErrorCode& status)
{
    if (U_SUCCESS(status)) {
        StringPair* sp = new StringPair(displayName, id);
        if (sp == NULL || sp->isBogus()) {
            status = U_MEMORY_ALLOCATION_ERROR;
            delete sp;
            return NULL;
        }
        return sp;
    }
    return NULL;
}

UBool 
StringPair::isBogus() const {
    return displayName.isBogus() || id.isBogus();
}

StringPair::StringPair(const UnicodeString& _displayName, 
                       const UnicodeString& _id)
: displayName(_displayName)
, id(_id)
{
}

U_CDECL_BEGIN
static void U_CALLCONV
userv_deleteStringPair(void *obj) {
    U_NAMESPACE_USE delete (StringPair*) obj;
}
U_CDECL_END

/*
******************************************************************
*/

static UMutex lock;

ICUService::ICUService()
: name()
, timestamp(0)
, factories(NULL)
, serviceCache(NULL)
, idCache(NULL)
, dnCache(NULL)
{
}

ICUService::ICUService(const UnicodeString& newName) 
: name(newName)
, timestamp(0)
, factories(NULL)
, serviceCache(NULL)
, idCache(NULL)
, dnCache(NULL)
{
}

ICUService::~ICUService()
{
    {
        Mutex mutex(&lock);
        clearCaches();
        delete factories;
        factories = NULL;
    }
}

UObject* 
ICUService::get(const UnicodeString& descriptor, UErrorCode& status) const 
{
    return get(descriptor, NULL, status);
}

UObject* 
ICUService::get(const UnicodeString& descriptor, UnicodeString* actualReturn, UErrorCode& status) const 
{
    UObject* result = NULL;
    ICUServiceKey* key = createKey(&descriptor, status);
    if (key) {
        result = getKey(*key, actualReturn, status);
        delete key;
    }
    return result;
}

UObject* 
ICUService::getKey(ICUServiceKey& key, UErrorCode& status) const 
{
    return getKey(key, NULL, status);
}

// this is a vector that subclasses of ICUService can override to further customize the result object
// before returning it.  All other public get functions should call this one.

UObject* 
ICUService::getKey(ICUServiceKey& key, UnicodeString* actualReturn, UErrorCode& status) const 
{
    return getKey(key, actualReturn, NULL, status);
}

// make it possible to call reentrantly on systems that don't have reentrant mutexes.
// we can use this simple approach since we know the situation where we're calling
// reentrantly even without knowing the thread.
class XMutex : public UMemory {
public:
    inline XMutex(UMutex *mutex, UBool reentering) 
        : fMutex(mutex)
        , fActive(!reentering) 
    {
        if (fActive) umtx_lock(fMutex);
    }
    inline ~XMutex() {
        if (fActive) umtx_unlock(fMutex);
    }

private:
    UMutex  *fMutex;
    UBool fActive;
};

struct UVectorDeleter {
    UVector* _obj;
    UVectorDeleter() : _obj(NULL) {}
    ~UVectorDeleter() { delete _obj; }
};

// called only by factories, treat as private
UObject* 
ICUService::getKey(ICUServiceKey& key, UnicodeString* actualReturn, const ICUServiceFactory* factory, UErrorCode& status) const 
{
    if (U_FAILURE(status)) {
        return NULL;
    }

    if (isDefault()) {
        return handleDefault(key, actualReturn, status);
    }

    ICUService* ncthis = (ICUService*)this; // cast away semantic const

    CacheEntry* result = NULL;
    {
        // The factory list can't be modified until we're done, 
        // otherwise we might update the cache with an invalid result.
        // The cache has to stay in synch with the factory list.
        // ICU doesn't have monitors so we can't use rw locks, so 
        // we single-thread everything using this service, for now.

        // if factory is not null, we're calling from within the mutex,
        // and since some unix machines don't have reentrant mutexes we
        // need to make sure not to try to lock it again.
        XMutex mutex(&lock, factory != NULL);

        if (serviceCache == NULL) {
            ncthis->serviceCache = new Hashtable(status);
            if (ncthis->serviceCache == NULL) {
                return NULL;
            }
            if (U_FAILURE(status)) {
                delete serviceCache;
                return NULL;
            }
            serviceCache->setValueDeleter(cacheDeleter);
        }

        UnicodeString currentDescriptor;
        UVectorDeleter cacheDescriptorList;
        UBool putInCache = FALSE;

        int32_t startIndex = 0;
        int32_t limit = factories->size();
        UBool cacheResult = TRUE;

        if (factory != NULL) {
            for (int32_t i = 0; i < limit; ++i) {
                if (factory == (const ICUServiceFactory*)factories->elementAt(i)) {
                    startIndex = i + 1;
                    break;
                }
            }
            if (startIndex == 0) {
                // throw new InternalError("Factory " + factory + "not registered with service: " + this);
                status = U_ILLEGAL_ARGUMENT_ERROR;
                return NULL;
            }
            cacheResult = FALSE;
        }

        do {
            currentDescriptor.remove();
            key.currentDescriptor(currentDescriptor);
            result = (CacheEntry*)serviceCache->get(currentDescriptor);
            if (result != NULL) {
                break;
            }

            // first test of cache failed, so we'll have to update
            // the cache if we eventually succeed-- that is, if we're 
            // going to update the cache at all.
            putInCache = TRUE;

            int32_t index = startIndex;
            while (index < limit) {
                ICUServiceFactory* f = (ICUServiceFactory*)factories->elementAt(index++);
                UObject* service = f->create(key, this, status);
                if (U_FAILURE(status)) {
                    delete service;
                    return NULL;
                }
                if (service != NULL) {
                    result = new CacheEntry(currentDescriptor, service);
                    if (result == NULL) {
                        delete service;
                        status = U_MEMORY_ALLOCATION_ERROR;
                        return NULL;
                    }

                    goto outerEnd;
                }
            }

            // prepare to load the cache with all additional ids that 
            // will resolve to result, assuming we'll succeed.  We
            // don't want to keep querying on an id that's going to
            // fallback to the one that succeeded, we want to hit the
            // cache the first time next goaround.
            if (cacheDescriptorList._obj == NULL) {
                cacheDescriptorList._obj = new UVector(uprv_deleteUObject, NULL, 5, status);
                if (U_FAILURE(status)) {
                    return NULL;
                }
            }
            UnicodeString* idToCache = new UnicodeString(currentDescriptor);
            if (idToCache == NULL || idToCache->isBogus()) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return NULL;
            }

            cacheDescriptorList._obj->addElement(idToCache, status);
            if (U_FAILURE(status)) {
                return NULL;
            }
        } while (key.fallback());
outerEnd:

        if (result != NULL) {
            if (putInCache && cacheResult) {
                serviceCache->put(result->actualDescriptor, result, status);
                if (U_FAILURE(status)) {
                    return NULL;
                }

                if (cacheDescriptorList._obj != NULL) {
                    for (int32_t i = cacheDescriptorList._obj->size(); --i >= 0;) {
                        UnicodeString* desc = (UnicodeString*)cacheDescriptorList._obj->elementAt(i);

                        serviceCache->put(*desc, result, status);
                        if (U_FAILURE(status)) {
                            return NULL;
                        }

                        result->ref();
                        cacheDescriptorList._obj->removeElementAt(i);
                    }
                }
            }

            if (actualReturn != NULL) {
                // strip null prefix
                if (result->actualDescriptor.indexOf((UChar)0x2f) == 0) { // U+002f=slash (/)
                    actualReturn->remove();
                    actualReturn->append(result->actualDescriptor, 
                        1, 
                        result->actualDescriptor.length() - 1);
                } else {
                    *actualReturn = result->actualDescriptor;
                }

                if (actualReturn->isBogus()) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                    delete result;
                    return NULL;
                }
            }

            UObject* service = cloneInstance(result->service);
            if (putInCache && !cacheResult) {
                delete result;
            }
            return service;
        }
    }

    return handleDefault(key, actualReturn, status);
}

UObject* 
ICUService::handleDefault(const ICUServiceKey& /* key */, UnicodeString* /* actualIDReturn */, UErrorCode& /* status */) const 
{
    return NULL;
}

UVector& 
ICUService::getVisibleIDs(UVector& result, UErrorCode& status) const {
    return getVisibleIDs(result, NULL, status);
}

UVector& 
ICUService::getVisibleIDs(UVector& result, const UnicodeString* matchID, UErrorCode& status) const 
{
    result.removeAllElements();

    if (U_FAILURE(status)) {
        return result;
    }

    {
        Mutex mutex(&lock);
        const Hashtable* map = getVisibleIDMap(status);
        if (map != NULL) {
            ICUServiceKey* fallbackKey = createKey(matchID, status);

            for (int32_t pos = UHASH_FIRST;;) {
                const UHashElement* e = map->nextElement(pos);
                if (e == NULL) {
                    break;
                }

                const UnicodeString* id = (const UnicodeString*)e->key.pointer;
                if (fallbackKey != NULL) {
                    if (!fallbackKey->isFallbackOf(*id)) {
                        continue;
                    }
                }

                UnicodeString* idClone = new UnicodeString(*id);
                if (idClone == NULL || idClone->isBogus()) {
                    delete idClone;
                    status = U_MEMORY_ALLOCATION_ERROR;
                    break;
                }
                result.addElement(idClone, status);
                if (U_FAILURE(status)) {
                    delete idClone;
                    break;
                }
            }
            delete fallbackKey;
        }
    }
    if (U_FAILURE(status)) {
        result.removeAllElements();
    }
    return result;
}

const Hashtable* 
ICUService::getVisibleIDMap(UErrorCode& status) const {
    if (U_FAILURE(status)) return NULL;

    // must only be called when lock is already held

    ICUService* ncthis = (ICUService*)this; // cast away semantic const
    if (idCache == NULL) {
        ncthis->idCache = new Hashtable(status);
        if (idCache == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
        } else if (factories != NULL) {
            for (int32_t pos = factories->size(); --pos >= 0;) {
                ICUServiceFactory* f = (ICUServiceFactory*)factories->elementAt(pos);
                f->updateVisibleIDs(*idCache, status);
            }
            if (U_FAILURE(status)) {
                delete idCache;
                ncthis->idCache = NULL;
            }
        }
    }

    return idCache;
}


UnicodeString& 
ICUService::getDisplayName(const UnicodeString& id, UnicodeString& result) const 
{
    return getDisplayName(id, result, Locale::getDefault());
}

UnicodeString& 
ICUService::getDisplayName(const UnicodeString& id, UnicodeString& result, const Locale& locale) const 
{
    {
        UErrorCode status = U_ZERO_ERROR;
        Mutex mutex(&lock);
        const Hashtable* map = getVisibleIDMap(status);
        if (map != NULL) {
            ICUServiceFactory* f = (ICUServiceFactory*)map->get(id);
            if (f != NULL) {
                f->getDisplayName(id, locale, result);
                return result;
            }

            // fallback
            status = U_ZERO_ERROR;
            ICUServiceKey* fallbackKey = createKey(&id, status);
            while (fallbackKey != NULL && fallbackKey->fallback()) {
                UnicodeString us;
                fallbackKey->currentID(us);
                f = (ICUServiceFactory*)map->get(us);
                if (f != NULL) {
                    f->getDisplayName(id, locale, result);
                    delete fallbackKey;
                    return result;
                }
            }
            delete fallbackKey;
        }
    }
    result.setToBogus();
    return result;
}

UVector& 
ICUService::getDisplayNames(UVector& result, UErrorCode& status) const 
{
    return getDisplayNames(result, Locale::getDefault(), NULL, status);
}


UVector& 
ICUService::getDisplayNames(UVector& result, const Locale& locale, UErrorCode& status) const 
{
    return getDisplayNames(result, locale, NULL, status);
}

UVector& 
ICUService::getDisplayNames(UVector& result, 
                            const Locale& locale, 
                            const UnicodeString* matchID, 
                            UErrorCode& status) const 
{
    result.removeAllElements();
    result.setDeleter(userv_deleteStringPair);
    if (U_SUCCESS(status)) {
        ICUService* ncthis = (ICUService*)this; // cast away semantic const
        Mutex mutex(&lock);

        if (dnCache != NULL && dnCache->locale != locale) {
            delete dnCache;
            ncthis->dnCache = NULL;
        }

        if (dnCache == NULL) {
            const Hashtable* m = getVisibleIDMap(status);
            if (U_FAILURE(status)) {
                return result;
            }
            ncthis->dnCache = new DNCache(locale); 
            if (dnCache == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return result;
            }

            int32_t pos = UHASH_FIRST;
            const UHashElement* entry = NULL;
            while ((entry = m->nextElement(pos)) != NULL) {
                const UnicodeString* id = (const UnicodeString*)entry->key.pointer;
                ICUServiceFactory* f = (ICUServiceFactory*)entry->value.pointer;
                UnicodeString dname;
                f->getDisplayName(*id, locale, dname);
                if (dname.isBogus()) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                } else {
                    dnCache->cache.put(dname, (void*)id, status); // share pointer with visibleIDMap
                    if (U_SUCCESS(status)) {
                        continue;
                    }
                }
                delete dnCache;
                ncthis->dnCache = NULL;
                return result;
            }
        }
    }

    ICUServiceKey* matchKey = createKey(matchID, status);
    /* To ensure that all elements in the hashtable are iterated, set pos to -1.
     * nextElement(pos) will skip the position at pos and begin the iteration
     * at the next position, which in this case will be 0.
     */
    int32_t pos = UHASH_FIRST; 
    const UHashElement *entry = NULL;
    while ((entry = dnCache->cache.nextElement(pos)) != NULL) {
        const UnicodeString* id = (const UnicodeString*)entry->value.pointer;
        if (matchKey != NULL && !matchKey->isFallbackOf(*id)) {
            continue;
        }
        const UnicodeString* dn = (const UnicodeString*)entry->key.pointer;
        StringPair* sp = StringPair::create(*id, *dn, status);
        result.addElement(sp, status);
        if (U_FAILURE(status)) {
            result.removeAllElements();
            break;
        }
    }
    delete matchKey;

    return result;
}

URegistryKey
ICUService::registerInstance(UObject* objToAdopt, const UnicodeString& id, UErrorCode& status) 
{
    return registerInstance(objToAdopt, id, TRUE, status);
}

URegistryKey
ICUService::registerInstance(UObject* objToAdopt, const UnicodeString& id, UBool visible, UErrorCode& status) 
{
    ICUServiceKey* key = createKey(&id, status);
    if (key != NULL) {
        UnicodeString canonicalID;
        key->canonicalID(canonicalID);
        delete key;

        ICUServiceFactory* f = createSimpleFactory(objToAdopt, canonicalID, visible, status);
        if (f != NULL) {
            return registerFactory(f, status);
        }
    }
    delete objToAdopt;
    return NULL;
}

ICUServiceFactory* 
ICUService::createSimpleFactory(UObject* objToAdopt, const UnicodeString& id, UBool visible, UErrorCode& status)
{
    if (U_SUCCESS(status)) {
        if ((objToAdopt != NULL) && (!id.isBogus())) {
            return new SimpleFactory(objToAdopt, id, visible);
        }
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return NULL;
}

URegistryKey
ICUService::registerFactory(ICUServiceFactory* factoryToAdopt, UErrorCode& status) 
{
    if (U_SUCCESS(status) && factoryToAdopt != NULL) {
        Mutex mutex(&lock);

        if (factories == NULL) {
            factories = new UVector(deleteUObject, NULL, status);
            if (U_FAILURE(status)) {
                delete factories;
                return NULL;
            }
        }
        factories->insertElementAt(factoryToAdopt, 0, status);
        if (U_SUCCESS(status)) {
            clearCaches();
        } else {
            delete factoryToAdopt;
            factoryToAdopt = NULL;
        }
    }

    if (factoryToAdopt != NULL) {
        notifyChanged();
    }

    return (URegistryKey)factoryToAdopt;
}

UBool 
ICUService::unregister(URegistryKey rkey, UErrorCode& status) 
{
    ICUServiceFactory *factory = (ICUServiceFactory*)rkey;
    UBool result = FALSE;
    if (factory != NULL && factories != NULL) {
        Mutex mutex(&lock);

        if (factories->removeElement(factory)) {
            clearCaches();
            result = TRUE;
        } else {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            delete factory;
        }
    }
    if (result) {
        notifyChanged();
    }
    return result;
}

void 
ICUService::reset() 
{
    {
        Mutex mutex(&lock);
        reInitializeFactories();
        clearCaches();
    }
    notifyChanged();
}

void 
ICUService::reInitializeFactories() 
{
    if (factories != NULL) {
        factories->removeAllElements();
    }
}

UBool 
ICUService::isDefault() const 
{
    return countFactories() == 0;
}

ICUServiceKey* 
ICUService::createKey(const UnicodeString* id, UErrorCode& status) const 
{
    return (U_FAILURE(status) || id == NULL) ? NULL : new ICUServiceKey(*id);
}

void 
ICUService::clearCaches() 
{
    // callers synchronize before use
    ++timestamp;
    delete dnCache;
    dnCache = NULL;
    delete idCache;
    idCache = NULL;
    delete serviceCache; serviceCache = NULL;
}

void 
ICUService::clearServiceCache() 
{
    // callers synchronize before use
    delete serviceCache; serviceCache = NULL;
}

UBool 
ICUService::acceptsListener(const EventListener& l) const 
{
    return dynamic_cast<const ServiceListener*>(&l) != NULL;
}

void 
ICUService::notifyListener(EventListener& l) const 
{
    ((ServiceListener&)l).serviceChanged(*this);
}

UnicodeString&
ICUService::getName(UnicodeString& result) const 
{
    return result.append(name);
}

int32_t 
ICUService::countFactories() const 
{
    return factories == NULL ? 0 : factories->size();
}

int32_t
ICUService::getTimestamp() const
{
    return timestamp;
}

U_NAMESPACE_END

/* UCONFIG_NO_SERVICE */
#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
 *******************************************************************************
 * Copyright (C) 2001-2014, International Business Machines Corporation and    *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 *
 *******************************************************************************
 */


#if !UCONFIG_NO_SERVICE




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
 *******************************************************************************
 * Copyright (C) 2001-2011, International Business Machines Corporation and    *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 *
 *******************************************************************************
 */



#if UCONFIG_NO_SERVICE

U_NAMESPACE_BEGIN

/*
 * Allow the declaration of APIs with pointers to ICUService
 * even when service is removed from the build.
 */
class ICULocaleService;

U_NAMESPACE_END

#else









// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /**
//  *******************************************************************************
//  * Copyright (C) 2002-2005, International Business Machines Corporation and    *
//  * others. All Rights Reserved.                                                *
//  *******************************************************************************
//  *
//  *******************************************************************************
//  */
// #ifndef LOCUTIL_H
// #define LOCUTIL_H

// #include "unicode/utypes.h"
// #include "hash.h"

// #if !UCONFIG_NO_SERVICE || !UCONFIG_NO_TRANSLITERATION


// U_NAMESPACE_BEGIN

// // temporary utility functions, till I know where to find them
// // in header so tests can also access them

// class U_COMMON_API LocaleUtility {
// public:
//   static UnicodeString& canonicalLocaleString(const UnicodeString* id, UnicodeString& result);
//   static Locale& initLocaleFromName(const UnicodeString& id, Locale& result);
//   static UnicodeString& initNameFromLocale(const Locale& locale, UnicodeString& result);
//   static const Hashtable* getAvailableLocaleNames(const UnicodeString& bundleID);
//   static UBool isFallbackOf(const UnicodeString& root, const UnicodeString& child);
// };

// U_NAMESPACE_END


// #endif

// #endif


U_NAMESPACE_BEGIN

class ICULocaleService;

class LocaleKey;
class LocaleKeyFactory;
class SimpleLocaleKeyFactory;
class ServiceListener;

/*
 ******************************************************************
 */

/**
 * A subclass of Key that implements a locale fallback mechanism.
 * The first locale to search for is the locale provided by the
 * client, and the fallback locale to search for is the current
 * default locale.  If a prefix is present, the currentDescriptor
 * includes it before the locale proper, separated by "/".  This
 * is the default key instantiated by ICULocaleService.</p>
 *
 * <p>Canonicalization adjusts the locale string so that the
 * section before the first understore is in lower case, and the rest
 * is in upper case, with no trailing underscores.</p> 
 */

class U_COMMON_API LocaleKey : public ICUServiceKey {
  private: 
    int32_t _kind;
    UnicodeString _primaryID;
    UnicodeString _fallbackID;
    UnicodeString _currentID;

  public:
    enum {
        KIND_ANY = -1
    };

    /**
     * Create a LocaleKey with canonical primary and fallback IDs.
     */
    static LocaleKey* createWithCanonicalFallback(const UnicodeString* primaryID, 
                                                  const UnicodeString* canonicalFallbackID,
                                                  UErrorCode& status);

    /**
     * Create a LocaleKey with canonical primary and fallback IDs.
     */
    static LocaleKey* createWithCanonicalFallback(const UnicodeString* primaryID, 
                                                  const UnicodeString* canonicalFallbackID, 
                                                  int32_t kind,
                                                  UErrorCode& status);

  protected:
    /**
     * PrimaryID is the user's requested locale string,
     * canonicalPrimaryID is this string in canonical form,
     * fallbackID is the current default locale's string in
     * canonical form.
     */
    LocaleKey(const UnicodeString& primaryID, 
              const UnicodeString& canonicalPrimaryID, 
              const UnicodeString* canonicalFallbackID, 
              int32_t kind);

 public:
    /**
     * Append the prefix associated with the kind, or nothing if the kind is KIND_ANY.
     */
    virtual UnicodeString& prefix(UnicodeString& result) const;

    /**
     * Return the kind code associated with this key.
     */
    virtual int32_t kind() const;

    /**
     * Return the canonicalID.
     */
    virtual UnicodeString& canonicalID(UnicodeString& result) const;

    /**
     * Return the currentID.
     */
    virtual UnicodeString& currentID(UnicodeString& result) const;

    /**
     * Return the (canonical) current descriptor, or null if no current id.
     */
    virtual UnicodeString& currentDescriptor(UnicodeString& result) const;

    /**
     * Convenience method to return the locale corresponding to the (canonical) original ID.
     */
    virtual Locale& canonicalLocale(Locale& result) const;

    /**
     * Convenience method to return the locale corresponding to the (canonical) current ID.
     */
    virtual Locale& currentLocale(Locale& result) const;

    /**
     * If the key has a fallback, modify the key and return true,
     * otherwise return false.</p>
     *
     * <p>First falls back through the primary ID, then through
     * the fallbackID.  The final fallback is the empty string,
     * unless the primary id was the empty string, in which case
     * there is no fallback.  
     */
    virtual UBool fallback();

    /**
     * Return true if a key created from id matches, or would eventually
     * fallback to match, the canonical ID of this key.  
     */
    virtual UBool isFallbackOf(const UnicodeString& id) const;
    
 public:
    /**
     * UObject boilerplate.
     */
    static UClassID U_EXPORT2 getStaticClassID();

    virtual UClassID getDynamicClassID() const;

    /**
     * Destructor.
     */
    virtual ~LocaleKey();

#ifdef SERVICE_DEBUG
 public:
    virtual UnicodeString& debug(UnicodeString& result) const;
    virtual UnicodeString& debugClass(UnicodeString& result) const;
#endif

};

/*
 ******************************************************************
 */

/**
 * A subclass of ICUServiceFactory that uses LocaleKeys, and is able to
 * 'cover' more specific locales with more general locales that it
 * supports.  
 *
 * <p>Coverage may be either of the values VISIBLE or INVISIBLE.
 *
 * <p>'Visible' indicates that the specific locale(s) supported by
 * the factory are registered in getSupportedIDs, 'Invisible'
 * indicates that they are not.
 *
 * <p>Localization of visible ids is handled
 * by the handling factory, regardless of kind.
 */
class U_COMMON_API LocaleKeyFactory : public ICUServiceFactory {
protected:
    const UnicodeString _name;
    const int32_t _coverage;

public:
    enum {
        /**
         * Coverage value indicating that the factory makes
         * its locales visible, and does not cover more specific 
         * locales.
         */
        VISIBLE = 0,

        /**
         * Coverage value indicating that the factory does not make
         * its locales visible, and does not cover more specific
         * locales.
         */
        INVISIBLE = 1
    };

    /**
     * Destructor.
     */
    virtual ~LocaleKeyFactory();

protected:
    /**
     * Constructor used by subclasses.
     */
    LocaleKeyFactory(int32_t coverage);

    /**
     * Constructor used by subclasses.
     */
    LocaleKeyFactory(int32_t coverage, const UnicodeString& name);

    /**
     * Implement superclass abstract method.  This checks the currentID of
     * the key against the supported IDs, and passes the canonicalLocale and
     * kind off to handleCreate (which subclasses must implement).
     */
public:
    virtual UObject* create(const ICUServiceKey& key, const ICUService* service, UErrorCode& status) const;

protected:
    virtual UBool handlesKey(const ICUServiceKey& key, UErrorCode& status) const;

public:
    /**
     * Override of superclass method.  This adjusts the result based
     * on the coverage rule for this factory.
     */
    virtual void updateVisibleIDs(Hashtable& result, UErrorCode& status) const;

    /**
     * Return a localized name for the locale represented by id.
     */
    virtual UnicodeString& getDisplayName(const UnicodeString& id, const Locale& locale, UnicodeString& result) const;

protected:
    /**
     * Utility method used by create(ICUServiceKey, ICUService).  Subclasses can implement
     * this instead of create.  The default returns NULL.
     */
    virtual UObject* handleCreate(const Locale& loc, int32_t kind, const ICUService* service, UErrorCode& status) const;

   /**
     * Return true if this id is one the factory supports (visible or 
     * otherwise).
     */
 //   virtual UBool isSupportedID(const UnicodeString& id, UErrorCode& status) const;

   /**
     * Return the set of ids that this factory supports (visible or 
     * otherwise).  This can be called often and might need to be
     * cached if it is expensive to create.
     */
    virtual const Hashtable* getSupportedIDs(UErrorCode& status) const;

public:
    /**
     * UObject boilerplate.
     */
    static UClassID U_EXPORT2 getStaticClassID();

    virtual UClassID getDynamicClassID() const;

#ifdef SERVICE_DEBUG
 public:
    virtual UnicodeString& debug(UnicodeString& result) const;
    virtual UnicodeString& debugClass(UnicodeString& result) const;
#endif

};

/*
 ******************************************************************
 */

/**
 * A LocaleKeyFactory that just returns a single object for a kind/locale.
 */

class U_COMMON_API SimpleLocaleKeyFactory : public LocaleKeyFactory {
 private:
    UObject* _obj;
    UnicodeString _id;
    const int32_t _kind;

 public:
    SimpleLocaleKeyFactory(UObject* objToAdopt, 
                           const UnicodeString& locale, 
                           int32_t kind, 
                           int32_t coverage);

    SimpleLocaleKeyFactory(UObject* objToAdopt, 
                           const Locale& locale, 
                           int32_t kind, 
                           int32_t coverage);

    /**
     * Destructor.
     */
    virtual ~SimpleLocaleKeyFactory();

    /**
     * Override of superclass method.  Returns the service object if kind/locale match.  Service is not used.
     */
    virtual UObject* create(const ICUServiceKey& key, const ICUService* service, UErrorCode& status) const;

    /**
     * Override of superclass method.  This adjusts the result based
     * on the coverage rule for this factory.
     */
    virtual void updateVisibleIDs(Hashtable& result, UErrorCode& status) const;

 protected:
    /**
     * Return true if this id is equal to the locale name.
     */
    //virtual UBool isSupportedID(const UnicodeString& id, UErrorCode& status) const;


public:
    /**
     * UObject boilerplate.
     */
    static UClassID U_EXPORT2 getStaticClassID();

    virtual UClassID getDynamicClassID() const;

#ifdef SERVICE_DEBUG
 public:
    virtual UnicodeString& debug(UnicodeString& result) const;
    virtual UnicodeString& debugClass(UnicodeString& result) const;
#endif

};

/*
 ******************************************************************
 */

/**
 * A LocaleKeyFactory that creates a service based on the ICU locale data.
 * This is a base class for most ICU factories.  Subclasses instantiate it
 * with a constructor that takes a bundle name, which determines the supported
 * IDs.  Subclasses then override handleCreate to create the actual service
 * object.  The default implementation returns a resource bundle.
 */
class U_COMMON_API ICUResourceBundleFactory : public LocaleKeyFactory 
{
 protected:
    UnicodeString _bundleName;

 public:
    /**
     * Convenience constructor that uses the main ICU bundle name.
     */
    ICUResourceBundleFactory();

    /**
     * A service factory based on ICU resource data in resources with
     * the given name.  This should be a 'path' that can be passed to
     * ures_openAvailableLocales, such as U_ICUDATA or U_ICUDATA_COLL.
     * The empty string is equivalent to U_ICUDATA.
     */
    ICUResourceBundleFactory(const UnicodeString& bundleName);

    /**
     * Destructor
     */
    virtual ~ICUResourceBundleFactory();

protected:
    /**
     * Return the supported IDs.  This is the set of all locale names in ICULocaleData.
     */
    virtual const Hashtable* getSupportedIDs(UErrorCode& status) const;

    /**
     * Create the service.  The default implementation returns the resource bundle
     * for the locale, ignoring kind, and service.
     */
    virtual UObject* handleCreate(const Locale& loc, int32_t kind, const ICUService* service, UErrorCode& status) const;

public:
    /**
     * UObject boilerplate.
     */
    static UClassID U_EXPORT2 getStaticClassID();
    virtual UClassID getDynamicClassID() const;


#ifdef SERVICE_DEBUG
 public:
    virtual UnicodeString& debug(UnicodeString& result) const;
    virtual UnicodeString& debugClass(UnicodeString& result) const;
#endif

};

/*
 ******************************************************************
 */

class U_COMMON_API ICULocaleService : public ICUService 
{
 private:
  Locale fallbackLocale;
  UnicodeString fallbackLocaleName;

 public:
  /**
   * Construct an ICULocaleService.
   */
  ICULocaleService();

  /**
   * Construct an ICULocaleService with a name (useful for debugging).
   */
  ICULocaleService(const UnicodeString& name);

  /**
   * Destructor.
   */
  virtual ~ICULocaleService();

#if 0
  // redeclare because of overload resolution rules?
  // no, causes ambiguities since both UnicodeString and Locale have constructors that take a const char*
  // need some compiler flag to remove warnings 
  UObject* get(const UnicodeString& descriptor, UErrorCode& status) const {
    return ICUService::get(descriptor, status);
  }

  UObject* get(const UnicodeString& descriptor, UnicodeString* actualReturn, UErrorCode& status) const {
    return ICUService::get(descriptor, actualReturn, status);
  }
#endif

  /**
   * Convenience override for callers using locales.  This calls
   * get(Locale, int, Locale[]) with KIND_ANY for kind and null for
   * actualReturn.
   */
  UObject* get(const Locale& locale, UErrorCode& status) const;

  /**
   * Convenience override for callers using locales.  This calls
   * get(Locale, int, Locale[]) with a null actualReturn.
   */
  UObject* get(const Locale& locale, int32_t kind, UErrorCode& status) const;

  /**
   * Convenience override for callers using locales. This calls
   * get(Locale, String, Locale[]) with a null kind.
   */
  UObject* get(const Locale& locale, Locale* actualReturn, UErrorCode& status) const;
                   
  /**
   * Convenience override for callers using locales.  This uses
   * createKey(Locale.toString(), kind) to create a key, calls getKey, and then
   * if actualReturn is not null, returns the actualResult from
   * getKey (stripping any prefix) into a Locale.  
   */
  UObject* get(const Locale& locale, int32_t kind, Locale* actualReturn, UErrorCode& status) const;

  /**
   * Convenience override for callers using locales.  This calls
   * registerObject(Object, Locale, int32_t kind, int coverage)
   * passing KIND_ANY for the kind, and VISIBLE for the coverage.
   */
  virtual URegistryKey registerInstance(UObject* objToAdopt, const Locale& locale, UErrorCode& status);

  /**
   * Convenience function for callers using locales.  This calls
   * registerObject(Object, Locale, int kind, int coverage)
   * passing VISIBLE for the coverage.
   */
  virtual URegistryKey registerInstance(UObject* objToAdopt, const Locale& locale, int32_t kind, UErrorCode& status);

  /**
   * Convenience function for callers using locales.  This  instantiates
   * a SimpleLocaleKeyFactory, and registers the factory.
   */
  virtual URegistryKey registerInstance(UObject* objToAdopt, const Locale& locale, int32_t kind, int32_t coverage, UErrorCode& status);


  /**
   * (Stop compiler from complaining about hidden overrides.)
   * Since both UnicodeString and Locale have constructors that take const char*, adding a public
   * method that takes UnicodeString causes ambiguity at call sites that use const char*.
   * We really need a flag that is understood by all compilers that will suppress the warning about
   * hidden overrides.
   */
  virtual URegistryKey registerInstance(UObject* objToAdopt, const UnicodeString& locale, UBool visible, UErrorCode& status);

  /**
   * Convenience method for callers using locales.  This returns the standard
   * service ID enumeration.
   */
  virtual StringEnumeration* getAvailableLocales(void) const;

 protected:

  /**
   * Return the name of the current fallback locale.  If it has changed since this was
   * last accessed, the service cache is cleared.
   */
  const UnicodeString& validateFallbackLocale() const;

  /**
   * Override superclass createKey method.
   */
  virtual ICUServiceKey* createKey(const UnicodeString* id, UErrorCode& status) const;

  /**
   * Additional createKey that takes a kind.
   */
  virtual ICUServiceKey* createKey(const UnicodeString* id, int32_t kind, UErrorCode& status) const;

  friend class ServiceEnumeration;
};

U_NAMESPACE_END

    /* UCONFIG_NO_SERVICE */
#endif

    /* ICULSERV_H */

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2001-2006, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/




U_CAPI int32_t U_EXPORT2
uprv_itou (UChar * buffer, int32_t capacity, uint32_t i, uint32_t radix, int32_t minwidth);







#define UNDERSCORE_CHAR ((UChar)0x005f)
#define AT_SIGN_CHAR    ((UChar)64)
#define PERIOD_CHAR     ((UChar)46)

U_NAMESPACE_BEGIN

LocaleKey*
LocaleKey::createWithCanonicalFallback(const UnicodeString* primaryID,
                                       const UnicodeString* canonicalFallbackID,
                                       UErrorCode& status)
{
    return LocaleKey::createWithCanonicalFallback(primaryID, canonicalFallbackID, KIND_ANY, status);
}

LocaleKey*
LocaleKey::createWithCanonicalFallback(const UnicodeString* primaryID,
                                       const UnicodeString* canonicalFallbackID,
                                       int32_t kind,
                                       UErrorCode& status)
{
    if (primaryID == NULL || U_FAILURE(status)) {
        return NULL;
    }
    UnicodeString canonicalPrimaryID;
    LocaleUtility::canonicalLocaleString(primaryID, canonicalPrimaryID);
    return new LocaleKey(*primaryID, canonicalPrimaryID, canonicalFallbackID, kind);
}

LocaleKey::LocaleKey(const UnicodeString& primaryID,
                     const UnicodeString& canonicalPrimaryID,
                     const UnicodeString* canonicalFallbackID,
                     int32_t kind)
  : ICUServiceKey(primaryID)
  , _kind(kind)
  , _primaryID(canonicalPrimaryID)
  , _fallbackID()
  , _currentID()
{
    _fallbackID.setToBogus();
    if (_primaryID.length() != 0) {
        if (canonicalFallbackID != NULL && _primaryID != *canonicalFallbackID) {
            _fallbackID = *canonicalFallbackID;
        }
    }

    _currentID = _primaryID;
}

LocaleKey::~LocaleKey() {}

UnicodeString&
LocaleKey::prefix(UnicodeString& result) const {
    if (_kind != KIND_ANY) {
        UChar buffer[64];
        uprv_itou(buffer, 64, _kind, 10, 0);
        UnicodeString temp(buffer);
        result.append(temp);
    }
    return result;
}

int32_t
LocaleKey::kind() const {
    return _kind;
}

UnicodeString&
LocaleKey::canonicalID(UnicodeString& result) const {
    return result.append(_primaryID);
}

UnicodeString&
LocaleKey::currentID(UnicodeString& result) const {
    if (!_currentID.isBogus()) {
        result.append(_currentID);
    }
    return result;
}

UnicodeString&
LocaleKey::currentDescriptor(UnicodeString& result) const {
    if (!_currentID.isBogus()) {
        prefix(result).append(PREFIX_DELIMITER).append(_currentID);
    } else {
        result.setToBogus();
    }
    return result;
}

Locale&
LocaleKey::canonicalLocale(Locale& result) const {
    return LocaleUtility::initLocaleFromName(_primaryID, result);
}

Locale&
LocaleKey::currentLocale(Locale& result) const {
    return LocaleUtility::initLocaleFromName(_currentID, result);
}

UBool
LocaleKey::fallback() {
    if (!_currentID.isBogus()) {
        int x = _currentID.lastIndexOf(UNDERSCORE_CHAR);
        if (x != -1) {
            _currentID.remove(x); // truncate current or fallback, whichever we're pointing to
            return TRUE;
        }

        if (!_fallbackID.isBogus()) {
            _currentID = _fallbackID;
            _fallbackID.setToBogus();
            return TRUE;
        }

        if (_currentID.length() > 0) {
            _currentID.remove(0); // completely truncate
            return TRUE;
        }

        _currentID.setToBogus();
    }

    return FALSE;
}

UBool
LocaleKey::isFallbackOf(const UnicodeString& id) const {
    UnicodeString temp(id);
    parseSuffix(temp);
    return temp.indexOf(_primaryID) == 0 &&
        (temp.length() == _primaryID.length() ||
        temp.charAt(_primaryID.length()) == UNDERSCORE_CHAR);
}

#ifdef SERVICE_DEBUG
UnicodeString&
LocaleKey::debug(UnicodeString& result) const
{
    ICUServiceKey::debug(result);
    result.append((UnicodeString)" kind: ");
    result.append(_kind);
    result.append((UnicodeString)" primaryID: ");
    result.append(_primaryID);
    result.append((UnicodeString)" fallbackID: ");
    result.append(_fallbackID);
    result.append((UnicodeString)" currentID: ");
    result.append(_currentID);
    return result;
}

UnicodeString&
LocaleKey::debugClass(UnicodeString& result) const
{
    return result.append((UnicodeString)"LocaleKey ");
}
#endif

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(LocaleKey)

U_NAMESPACE_END

/* !UCONFIG_NO_SERVICE */
#endif


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
 *******************************************************************************
 * Copyright (C) 2001-2014, International Business Machines Corporation and    *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 *
 *******************************************************************************
 */


#if !UCONFIG_NO_SERVICE











#define UNDERSCORE_CHAR ((UChar)0x005f)
#define AT_SIGN_CHAR    ((UChar)64)
#define PERIOD_CHAR     ((UChar)46)


U_NAMESPACE_BEGIN

LocaleKeyFactory::LocaleKeyFactory(int32_t coverage)
  : _name()
  , _coverage(coverage)
{
}

LocaleKeyFactory::LocaleKeyFactory(int32_t coverage, const UnicodeString& name)
  : _name(name)
  , _coverage(coverage)
{
}

LocaleKeyFactory::~LocaleKeyFactory() {
}

UObject*
LocaleKeyFactory::create(const ICUServiceKey& key, const ICUService* service, UErrorCode& status) const {
    if (handlesKey(key, status)) {
        const LocaleKey& lkey = (const LocaleKey&)key;
        int32_t kind = lkey.kind();
        Locale loc;
        lkey.currentLocale(loc);

        return handleCreate(loc, kind, service, status);
    }
    return NULL;
}

UBool
LocaleKeyFactory::handlesKey(const ICUServiceKey& key, UErrorCode& status) const {
    const Hashtable* supported = getSupportedIDs(status);
    if (supported) {
        UnicodeString id;
        key.currentID(id);
        return supported->get(id) != NULL;
    }
    return FALSE;
}

void
LocaleKeyFactory::updateVisibleIDs(Hashtable& result, UErrorCode& status) const {
    const Hashtable* supported = getSupportedIDs(status);
    if (supported) {
        UBool visible = (_coverage & 0x1) == 0;
        const UHashElement* elem = NULL;
        int32_t pos = UHASH_FIRST;
        while ((elem = supported->nextElement(pos)) != NULL) {
            const UnicodeString& id = *((const UnicodeString*)elem->key.pointer);
            if (!visible) {
                result.remove(id);
            } else {
                result.put(id, (void*)this, status); // this is dummy non-void marker used for set semantics
                if (U_FAILURE(status)) {
                    break;
                }
            }
        }
    }
}

UnicodeString&
LocaleKeyFactory::getDisplayName(const UnicodeString& id, const Locale& locale, UnicodeString& result) const {
    if ((_coverage & 0x1) == 0) {
        //UErrorCode status = U_ZERO_ERROR;
        // assume if this is called on us, we support some fallback of this id
        // if (isSupportedID(id, status)) {
            Locale loc;
            LocaleUtility::initLocaleFromName(id, loc);
            return loc.getDisplayName(locale, result);
        // }
    }
    result.setToBogus();
    return result;
}

UObject*
LocaleKeyFactory::handleCreate(const Locale& /* loc */, 
                   int32_t /* kind */, 
                   const ICUService* /* service */, 
                   UErrorCode& /* status */) const {
    return NULL;
}

//UBool
//LocaleKeyFactory::isSupportedID(const UnicodeString& id, UErrorCode& status) const {
//    const Hashtable* ids = getSupportedIDs(status);
//    return ids && ids->get(id);
//}

const Hashtable*
LocaleKeyFactory::getSupportedIDs(UErrorCode& /* status */) const {
    return NULL;
}

#ifdef SERVICE_DEBUG
UnicodeString&
LocaleKeyFactory::debug(UnicodeString& result) const
{
    debugClass(result);
    result.append((UnicodeString)", name: ");
    result.append(_name);
    result.append((UnicodeString)", coverage: ");
    result.append(_coverage);
    return result;
}

UnicodeString&
LocaleKeyFactory::debugClass(UnicodeString& result) const
{
  return result.append((UnicodeString)"LocaleKeyFactory");
}
#endif

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(LocaleKeyFactory)

U_NAMESPACE_END

/* !UCONFIG_NO_SERVICE */
#endif


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
 *******************************************************************************
 * Copyright (C) 2001-2014, International Business Machines Corporation and    *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 *
 *******************************************************************************
 */


#if !UCONFIG_NO_SERVICE









#define UNDERSCORE_CHAR ((UChar)0x005f)
#define AT_SIGN_CHAR    ((UChar)64)
#define PERIOD_CHAR     ((UChar)46)

U_NAMESPACE_BEGIN

ICULocaleService::ICULocaleService()
  : fallbackLocale(Locale::getDefault())
{
}

ICULocaleService::ICULocaleService(const UnicodeString& dname)
  : ICUService(dname)
  , fallbackLocale(Locale::getDefault())
{
}

ICULocaleService::~ICULocaleService()
{
}

UObject*
ICULocaleService::get(const Locale& locale, UErrorCode& status) const
{
    return get(locale, LocaleKey::KIND_ANY, NULL, status);
}

UObject*
ICULocaleService::get(const Locale& locale, int32_t kind, UErrorCode& status) const
{
    return get(locale, kind, NULL, status);
}

UObject*
ICULocaleService::get(const Locale& locale, Locale* actualReturn, UErrorCode& status) const
{
    return get(locale, LocaleKey::KIND_ANY, actualReturn, status);
}

UObject*
ICULocaleService::get(const Locale& locale, int32_t kind, Locale* actualReturn, UErrorCode& status) const
{
    UObject* result = NULL;
    if (U_FAILURE(status)) {
        return result;
    }

    UnicodeString locName(locale.getName(), -1, US_INV);
    if (locName.isBogus()) {
        status = U_MEMORY_ALLOCATION_ERROR;
    } else {
        ICUServiceKey* key = createKey(&locName, kind, status);
        if (key) {
            if (actualReturn == NULL) {
                result = getKey(*key, status);
            } else {
                UnicodeString temp;
                result = getKey(*key, &temp, status);

                if (result != NULL) {
                    key->parseSuffix(temp);
                    LocaleUtility::initLocaleFromName(temp, *actualReturn);
                }
            }
            delete key;
        }
    }
    return result;
}


URegistryKey
ICULocaleService::registerInstance(UObject* objToAdopt, const UnicodeString& locale, 
    UBool visible, UErrorCode& status)
{
    Locale loc;
    LocaleUtility::initLocaleFromName(locale, loc);
    return registerInstance(objToAdopt, loc, LocaleKey::KIND_ANY, 
        visible ? LocaleKeyFactory::VISIBLE : LocaleKeyFactory::INVISIBLE, status);
}

URegistryKey
ICULocaleService::registerInstance(UObject* objToAdopt, const Locale& locale, UErrorCode& status)
{
    return registerInstance(objToAdopt, locale, LocaleKey::KIND_ANY, LocaleKeyFactory::VISIBLE, status);
}

URegistryKey
ICULocaleService::registerInstance(UObject* objToAdopt, const Locale& locale, int32_t kind, UErrorCode& status)
{
    return registerInstance(objToAdopt, locale, kind, LocaleKeyFactory::VISIBLE, status);
}

URegistryKey
ICULocaleService::registerInstance(UObject* objToAdopt, const Locale& locale, int32_t kind, int32_t coverage, UErrorCode& status)
{
    ICUServiceFactory * factory = new SimpleLocaleKeyFactory(objToAdopt, locale, kind, coverage);
    if (factory != NULL) {
        return registerFactory(factory, status);
    }
    delete objToAdopt;
    return NULL;
}

#if 0
URegistryKey
ICULocaleService::registerInstance(UObject* objToAdopt, const UnicodeString& locale, UErrorCode& status)
{
    return registerInstance(objToAdopt, locale, LocaleKey::KIND_ANY, LocaleKeyFactory::VISIBLE, status);
}

URegistryKey
ICULocaleService::registerInstance(UObject* objToAdopt, const UnicodeString& locale, UBool visible, UErrorCode& status)
{
    return registerInstance(objToAdopt, locale, LocaleKey::KIND_ANY,
                            visible ? LocaleKeyFactory::VISIBLE : LocaleKeyFactory::INVISIBLE,
                            status);
}

URegistryKey
ICULocaleService::registerInstance(UObject* objToAdopt, const UnicodeString& locale, int32_t kind, int32_t coverage, UErrorCode& status)
{
    ICUServiceFactory * factory = new SimpleLocaleKeyFactory(objToAdopt, locale, kind, coverage);
    if (factory != NULL) {
        return registerFactory(factory, status);
    }
    delete objToAdopt;
    return NULL;
}
#endif

class ServiceEnumeration : public StringEnumeration {
private:
    const ICULocaleService* _service;
    int32_t _timestamp;
    UVector _ids;
    int32_t _pos;

private:
    ServiceEnumeration(const ICULocaleService* service, UErrorCode &status)
        : _service(service)
        , _timestamp(service->getTimestamp())
        , _ids(uprv_deleteUObject, NULL, status)
        , _pos(0)
    {
        _service->getVisibleIDs(_ids, status);
    }

    ServiceEnumeration(const ServiceEnumeration &other, UErrorCode &status)
        : _service(other._service)
        , _timestamp(other._timestamp)
        , _ids(uprv_deleteUObject, NULL, status)
        , _pos(0)
    {
        if(U_SUCCESS(status)) {
            int32_t i, length;

            length = other._ids.size();
            for(i = 0; i < length; ++i) {
                _ids.addElement(((UnicodeString *)other._ids.elementAt(i))->clone(), status);
            }

            if(U_SUCCESS(status)) {
                _pos = other._pos;
            }
        }
    }

public:
    static ServiceEnumeration* create(const ICULocaleService* service) {
        UErrorCode status = U_ZERO_ERROR;
        ServiceEnumeration* result = new ServiceEnumeration(service, status);
        if (U_SUCCESS(status)) {
            return result;
        }
        delete result;
        return NULL;
    }

    virtual ~ServiceEnumeration();

    virtual StringEnumeration *clone() const {
        UErrorCode status = U_ZERO_ERROR;
        ServiceEnumeration *cl = new ServiceEnumeration(*this, status);
        if(U_FAILURE(status)) {
            delete cl;
            cl = NULL;
        }
        return cl;
    }

    UBool upToDate(UErrorCode& status) const {
        if (U_SUCCESS(status)) {
            if (_timestamp == _service->getTimestamp()) {
                return TRUE;
            }
            status = U_ENUM_OUT_OF_SYNC_ERROR;
        }
        return FALSE;
    }

    virtual int32_t count(UErrorCode& status) const {
        return upToDate(status) ? _ids.size() : 0;
    }

    virtual const UnicodeString* snext(UErrorCode& status) {
        if (upToDate(status) && (_pos < _ids.size())) {
            return (const UnicodeString*)_ids[_pos++];
        }
        return NULL;
    }

    virtual void reset(UErrorCode& status) {
        if (status == U_ENUM_OUT_OF_SYNC_ERROR) {
            status = U_ZERO_ERROR;
        }
        if (U_SUCCESS(status)) {
            _timestamp = _service->getTimestamp();
            _pos = 0;
            _service->getVisibleIDs(_ids, status);
        }
    }

public:
    static UClassID U_EXPORT2 getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
};

ServiceEnumeration::~ServiceEnumeration() {}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(ServiceEnumeration)

StringEnumeration*
ICULocaleService::getAvailableLocales(void) const
{
    return ServiceEnumeration::create(this);
}

const UnicodeString&
ICULocaleService::validateFallbackLocale() const
{
    const Locale&     loc    = Locale::getDefault();
    ICULocaleService* ncThis = (ICULocaleService*)this;
    static UMutex llock;
    {
        Mutex mutex(&llock);
        if (loc != fallbackLocale) {
            ncThis->fallbackLocale = loc;
            LocaleUtility::initNameFromLocale(loc, ncThis->fallbackLocaleName);
            ncThis->clearServiceCache();
        }
    }
    return fallbackLocaleName;
}

ICUServiceKey*
ICULocaleService::createKey(const UnicodeString* id, UErrorCode& status) const
{
    return LocaleKey::createWithCanonicalFallback(id, &validateFallbackLocale(), status);
}

ICUServiceKey*
ICULocaleService::createKey(const UnicodeString* id, int32_t kind, UErrorCode& status) const
{
    return LocaleKey::createWithCanonicalFallback(id, &validateFallbackLocale(), kind, status);
}

U_NAMESPACE_END

/* !UCONFIG_NO_SERVICE */
#endif


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
 *******************************************************************************
 * Copyright (C) 2001-2012, International Business Machines Corporation and    *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 */



#if !UCONFIG_NO_SERVICE


#ifdef NOTIFIER_DEBUG
#include <stdio.h>
#endif

U_NAMESPACE_BEGIN

EventListener::~EventListener() {}
UOBJECT_DEFINE_RTTI_IMPLEMENTATION(EventListener)

static UMutex notifyLock;

ICUNotifier::ICUNotifier(void) 
: listeners(NULL) 
{
}

ICUNotifier::~ICUNotifier(void) {
    {
        Mutex lmx(&notifyLock);
        delete listeners;
        listeners = NULL;
    }
}


void 
ICUNotifier::addListener(const EventListener* l, UErrorCode& status) 
{
    if (U_SUCCESS(status)) {
        if (l == NULL) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }

        if (acceptsListener(*l)) {
            Mutex lmx(&notifyLock);
            if (listeners == NULL) {
                listeners = new UVector(5, status);
            } else {
                for (int i = 0, e = listeners->size(); i < e; ++i) {
                    const EventListener* el = (const EventListener*)(listeners->elementAt(i));
                    if (l == el) {
                        return;
                    }
                }
            }

            listeners->addElement((void*)l, status); // cast away const
        }
#ifdef NOTIFIER_DEBUG
        else {
            fprintf(stderr, "Listener invalid for this notifier.");
            exit(1);
        }
#endif
    }
}

void 
ICUNotifier::removeListener(const EventListener *l, UErrorCode& status) 
{
    if (U_SUCCESS(status)) {
        if (l == NULL) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }

        {
            Mutex lmx(&notifyLock);
            if (listeners != NULL) {
                // identity equality check
                for (int i = 0, e = listeners->size(); i < e; ++i) {
                    const EventListener* el = (const EventListener*)listeners->elementAt(i);
                    if (l == el) {
                        listeners->removeElementAt(i);
                        if (listeners->size() == 0) {
                            delete listeners;
                            listeners = NULL;
                        }
                        return;
                    }
                }
            }
        }
    }
}

void 
ICUNotifier::notifyChanged(void) 
{
    if (listeners != NULL) {
        Mutex lmx(&notifyLock);
        if (listeners != NULL) {
            for (int i = 0, e = listeners->size(); i < e; ++i) {
                EventListener* el = (EventListener*)listeners->elementAt(i);
                notifyListener(*el);
            }
        }
    }
}

U_NAMESPACE_END

/* UCONFIG_NO_SERVICE */
#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
 *******************************************************************************
 * Copyright (C) 2001-2014, International Business Machines Corporation and    *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 *
 *******************************************************************************
 */


#if !UCONFIG_NO_SERVICE











#define UNDERSCORE_CHAR ((UChar)0x005f)
#define AT_SIGN_CHAR    ((UChar)64)
#define PERIOD_CHAR     ((UChar)46)

U_NAMESPACE_BEGIN

ICUResourceBundleFactory::ICUResourceBundleFactory()
  : LocaleKeyFactory(VISIBLE)
  , _bundleName()
{
}

ICUResourceBundleFactory::ICUResourceBundleFactory(const UnicodeString& bundleName)
  : LocaleKeyFactory(VISIBLE)
  , _bundleName(bundleName)
{
}

ICUResourceBundleFactory::~ICUResourceBundleFactory() {}

const Hashtable*
ICUResourceBundleFactory::getSupportedIDs(UErrorCode& status) const
{
    if (U_SUCCESS(status)) {
        return LocaleUtility::getAvailableLocaleNames(_bundleName);
    }
    return NULL;
}

UObject*
ICUResourceBundleFactory::handleCreate(const Locale& loc, int32_t /* kind */, const ICUService* /* service */, UErrorCode& status) const
{
    if (U_SUCCESS(status)) {
        // _bundleName is a package name
        // and should only contain invariant characters
                // ??? is it always true that the max length of the bundle name is 19?
                // who made this change? -- dlf
        char pkg[20];
        int32_t length;
        length=_bundleName.extract(0, INT32_MAX, pkg, (int32_t)sizeof(pkg), US_INV);
        if(length>=(int32_t)sizeof(pkg)) {
            return NULL;
        }
        return new ResourceBundle(pkg, loc, status);
    }
    return NULL;
}

#ifdef SERVICE_DEBUG
UnicodeString&
ICUResourceBundleFactory::debug(UnicodeString& result) const
{
    LocaleKeyFactory::debug(result);
    result.append((UnicodeString)", bundle: ");
    return result.append(_bundleName);
}

UnicodeString&
ICUResourceBundleFactory::debugClass(UnicodeString& result) const
{
    return result.append((UnicodeString)"ICUResourceBundleFactory");
}
#endif

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(ICUResourceBundleFactory)

U_NAMESPACE_END

/* !UCONFIG_NO_SERVICE */
#endif


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/**
 *******************************************************************************
 * Copyright (C) 2001-2014, International Business Machines Corporation and    *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 *
 *******************************************************************************
 */


#if !UCONFIG_NO_SERVICE










#define UNDERSCORE_CHAR ((UChar)0x005f)
#define AT_SIGN_CHAR    ((UChar)64)
#define PERIOD_CHAR     ((UChar)46)

U_NAMESPACE_BEGIN

/*
 ******************************************************************
 */

SimpleLocaleKeyFactory::SimpleLocaleKeyFactory(UObject* objToAdopt,
                                               const UnicodeString& locale,
                                               int32_t kind,
                                               int32_t coverage)
  : LocaleKeyFactory(coverage)
  , _obj(objToAdopt)
  , _id(locale)
  , _kind(kind)
{
}

SimpleLocaleKeyFactory::SimpleLocaleKeyFactory(UObject* objToAdopt,
                                               const Locale& locale,
                                               int32_t kind,
                                               int32_t coverage)
  : LocaleKeyFactory(coverage)
  , _obj(objToAdopt)
  , _id()
  , _kind(kind)
{
    LocaleUtility::initNameFromLocale(locale, _id);
}

SimpleLocaleKeyFactory::~SimpleLocaleKeyFactory()
{
  delete _obj;
  _obj = NULL;
}

UObject*
SimpleLocaleKeyFactory::create(const ICUServiceKey& key, const ICUService* service, UErrorCode& status) const
{
    if (U_SUCCESS(status)) {
        const LocaleKey& lkey = (const LocaleKey&)key;
        if (_kind == LocaleKey::KIND_ANY || _kind == lkey.kind()) {
            UnicodeString keyID;
            lkey.currentID(keyID);
            if (_id == keyID) {
                return service->cloneInstance(_obj);
            }
        }
    }
    return NULL;
}

//UBool
//SimpleLocaleKeyFactory::isSupportedID(const UnicodeString& id, UErrorCode& /* status */) const
//{
//    return id == _id;
//}

void
SimpleLocaleKeyFactory::updateVisibleIDs(Hashtable& result, UErrorCode& status) const
{
    if (U_SUCCESS(status)) {
        if (_coverage & 0x1) {
            result.remove(_id);
        } else {
            result.put(_id, (void*)this, status);
        }
    }
}

#ifdef SERVICE_DEBUG
UnicodeString&
SimpleLocaleKeyFactory::debug(UnicodeString& result) const
{
    LocaleKeyFactory::debug(result);
    result.append((UnicodeString)", id: ");
    result.append(_id);
    result.append((UnicodeString)", kind: ");
    result.append(_kind);
    return result;
}

UnicodeString&
SimpleLocaleKeyFactory::debugClass(UnicodeString& result) const
{
    return result.append((UnicodeString)"SimpleLocaleKeyFactory");
}
#endif

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(SimpleLocaleKeyFactory)

U_NAMESPACE_END

/* !UCONFIG_NO_SERVICE */
#endif


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2015, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* sharedobject.cpp
*/
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2015-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* sharedobject.h
*/






U_NAMESPACE_BEGIN

class SharedObject;

/**
 * Base class for unified cache exposing enough methods to SharedObject
 * instances to allow their addRef() and removeRef() methods to
 * update cache metrics. No other part of ICU, except for SharedObject,
 * should directly call the methods of this base class.
 */
class U_COMMON_API UnifiedCacheBase : public UObject {
public:
    UnifiedCacheBase() { }

    /**
     * Notify the cache implementation that an object was seen transitioning to
     * zero hard references. The cache may use this to keep track the number of
     * unreferenced SharedObjects, and to trigger evictions.
     */
    virtual void handleUnreferencedObject() const = 0;

    virtual ~UnifiedCacheBase();
private:
    UnifiedCacheBase(const UnifiedCacheBase &);
    UnifiedCacheBase &operator=(const UnifiedCacheBase &);
};

/**
 * Base class for shared, reference-counted, auto-deleted objects.
 * Subclasses can be immutable.
 * If they are mutable, then they must implement their copy constructor
 * so that copyOnWrite() works.
 *
 * Either stack-allocate, use LocalPointer, or use addRef()/removeRef().
 * Sharing requires reference-counting.
 */
class U_COMMON_API SharedObject : public UObject {
public:
    /** Initializes totalRefCount, softRefCount to 0. */
    SharedObject() :
            softRefCount(0),
            hardRefCount(0),
            cachePtr(NULL) {}

    /** Initializes totalRefCount, softRefCount to 0. */
    SharedObject(const SharedObject &other) :
            UObject(other),
            softRefCount(0),
            hardRefCount(0),
            cachePtr(NULL) {}

    virtual ~SharedObject();

    /**
     * Increments the number of hard references to this object. Thread-safe.
     * Not for use from within the Unified Cache implementation.
     */
    void addRef() const;

    /**
     * Decrements the number of hard references to this object, and
     * arrange for possible cache-eviction and/or deletion if ref
     * count goes to zero. Thread-safe.
     * 
     * Not for use from within the UnifiedCache implementation.
     */
    void removeRef() const;

    /**
     * Returns the number of hard references for this object.
     * Uses a memory barrier.
     */
    int32_t getRefCount() const;

    /**
     * If noHardReferences() == TRUE then this object has no hard references.
     * Must be called only from within the internals of UnifiedCache.
     */
    inline UBool noHardReferences() const { return getRefCount() == 0; }

    /**
     * If hasHardReferences() == TRUE then this object has hard references.
     * Must be called only from within the internals of UnifiedCache.
     */
    inline UBool hasHardReferences() const { return getRefCount() != 0; }

    /**
     * Deletes this object if it has no references.
     * Available for non-cached SharedObjects only. Ownership of cached objects
     * is with the UnifiedCache, which is solely responsible for eviction and deletion.
     */
    void deleteIfZeroRefCount() const;

        
    /**
     * Returns a writable version of ptr.
     * If there is exactly one owner, then ptr itself is returned as a
     *  non-const pointer.
     * If there are multiple owners, then ptr is replaced with a 
     * copy-constructed clone,
     * and that is returned.
     * Returns NULL if cloning failed.
     *
     * T must be a subclass of SharedObject.
     */
    template<typename T>
    static T *copyOnWrite(const T *&ptr) {
        const T *p = ptr;
        if(p->getRefCount() <= 1) { return const_cast<T *>(p); }
        T *p2 = new T(*p);
        if(p2 == NULL) { return NULL; }
        p->removeRef();
        ptr = p2;
        p2->addRef();
        return p2;
    }

    /**
     * Makes dest an owner of the object pointed to by src while adjusting
     * reference counts and deleting the previous object dest pointed to
     * if necessary. Before this call is made, dest must either be NULL or
     * be included in the reference count of the object it points to. 
     *
     * T must be a subclass of SharedObject.
     */
    template<typename T>
    static void copyPtr(const T *src, const T *&dest) {
        if(src != dest) {
            if(dest != NULL) { dest->removeRef(); }
            dest = src;
            if(src != NULL) { src->addRef(); }
        }
    }

    /**
     * Equivalent to copyPtr(NULL, dest).
     */
    template<typename T>
    static void clearPtr(const T *&ptr) {
        if (ptr != NULL) {
            ptr->removeRef();
            ptr = NULL;
        }
    }

private:
    /**
     * The number of references from the UnifiedCache, which is
     * the number of times that the sharedObject is stored as a hash table value.
     * For use by UnifiedCache implementation code only.
     * All access is synchronized by UnifiedCache's gCacheMutex
     */
    mutable int32_t softRefCount;
    friend class UnifiedCache;

    /**
     * Reference count, excluding references from within the UnifiedCache implementation.
     */
    mutable u_atomic_int32_t hardRefCount;
    
    mutable const UnifiedCacheBase *cachePtr;

};

U_NAMESPACE_END





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2015, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*
* File UNIFIEDCACHE.H - The ICU Unified cache.
******************************************************************************
*/


  // for 'typeid' to work








struct UHashtable;
struct UHashElement;

U_NAMESPACE_BEGIN

class UnifiedCache;

/**
 * A base class for all cache keys.
 */
class U_COMMON_API CacheKeyBase : public UObject {
 public:
   CacheKeyBase() : fCreationStatus(U_ZERO_ERROR), fIsMaster(FALSE) {}

   /**
    * Copy constructor. Needed to support cloning.
    */
   CacheKeyBase(const CacheKeyBase &other)
           : UObject(other), fCreationStatus(other.fCreationStatus), fIsMaster(FALSE) { }
   virtual ~CacheKeyBase();

   /**
    * Returns the hash code for this object.
    */
   virtual int32_t hashCode() const = 0;

   /**
    * Clones this object polymorphically. Caller owns returned value.
    */
   virtual CacheKeyBase *clone() const = 0;

   /**
    * Equality operator.
    */
   virtual UBool operator == (const CacheKeyBase &other) const = 0;

   /**
    * Create a new object for this key. Called by cache on cache miss.
    * createObject must add a reference to the object it returns. Note
    * that getting an object from the cache and returning it without calling
    * removeRef on it satisfies this requirement. It can also return NULL
    * and set status to an error.
    *
    * @param creationContext the context in which the object is being
    *                        created. May be NULL.
    * @param status          Implementations can return a failure here.
    *                        In addition, implementations may return a
    *                        non NULL object and set a warning status.
    */
   virtual const SharedObject *createObject(
           const void *creationContext, UErrorCode &status) const = 0;

   /**
    * Writes a description of this key to buffer and returns buffer. Written
    * description is NULL terminated.
    */
   virtual char *writeDescription(char *buffer, int32_t bufSize) const = 0;

   /**
    * Inequality operator.
    */
   UBool operator != (const CacheKeyBase &other) const {
       return !(*this == other);
   }
 private:
   mutable UErrorCode fCreationStatus;
   mutable UBool fIsMaster;
   friend class UnifiedCache;
};



/**
 * Templated version of CacheKeyBase.
 * A key of type LocaleCacheKey<T> maps to a value of type T.
 */
template<typename T>
class CacheKey : public CacheKeyBase {
 public:
   virtual ~CacheKey() { }
   /**
    * The template parameter, T, determines the hash code returned.
    */
   virtual int32_t hashCode() const {
       const char *s = typeid(T).name();
       return ustr_hashCharsN(s, static_cast<int32_t>(uprv_strlen(s)));
   }

   /**
    * Use the value type, T,  as the description.
    */
   virtual char *writeDescription(char *buffer, int32_t bufLen) const {
       const char *s = typeid(T).name();
       uprv_strncpy(buffer, s, bufLen);
       buffer[bufLen - 1] = 0;
       return buffer;
   }

   /**
    * Two objects are equal if they are of the same type.
    */
   virtual UBool operator == (const CacheKeyBase &other) const {
       return typeid(*this) == typeid(other);
   }
};

/**
 * Cache key based on locale.
 * A key of type LocaleCacheKey<T> maps to a value of type T.
 */
template<typename T>
class LocaleCacheKey : public CacheKey<T> {
 protected:
   Locale   fLoc;
 public:
   LocaleCacheKey(const Locale &loc) : fLoc(loc) {}
   LocaleCacheKey(const LocaleCacheKey<T> &other)
           : CacheKey<T>(other), fLoc(other.fLoc) { }
   virtual ~LocaleCacheKey() { }
   virtual int32_t hashCode() const {
       return (int32_t)(37u * (uint32_t)CacheKey<T>::hashCode() + (uint32_t)fLoc.hashCode());
   }
   virtual UBool operator == (const CacheKeyBase &other) const {
       // reflexive
       if (this == &other) {
           return TRUE;
       }
       if (!CacheKey<T>::operator == (other)) {
           return FALSE;
       }
       // We know this and other are of same class because operator== on
       // CacheKey returned true.
       const LocaleCacheKey<T> *fOther =
               static_cast<const LocaleCacheKey<T> *>(&other);
       return fLoc == fOther->fLoc;
   }
   virtual CacheKeyBase *clone() const {
       return new LocaleCacheKey<T>(*this);
   }
   virtual const T *createObject(
           const void *creationContext, UErrorCode &status) const;
   /**
    * Use the locale id as the description.
    */
   virtual char *writeDescription(char *buffer, int32_t bufLen) const {
       const char *s = fLoc.getName();
       uprv_strncpy(buffer, s, bufLen);
       buffer[bufLen - 1] = 0;
       return buffer;
   }

};

/**
 * The unified cache. A singleton type.
 * Design doc here:
 * https://docs.google.com/document/d/1RwGQJs4N4tawNbf809iYDRCvXoMKqDJihxzYt1ysmd8/edit?usp=sharing
 */
class U_COMMON_API UnifiedCache : public UnifiedCacheBase {
 public:
   /**
    * @internal
    * Do not call directly. Instead use UnifiedCache::getInstance() as
    * there should be only one UnifiedCache in an application.
    */
   UnifiedCache(UErrorCode &status);

   /**
    * Return a pointer to the global cache instance.
    */
   static UnifiedCache *getInstance(UErrorCode &status);

   /**
    * Fetches a value from the cache by key. Equivalent to
    * get(key, NULL, ptr, status);
    */
   template<typename T>
   void get(
           const CacheKey<T>& key,
           const T *&ptr,
           UErrorCode &status) const {
       get(key, NULL, ptr, status);
   }

   /**
    * Fetches value from the cache by key.
    *
    * @param key             the cache key.
    * @param creationContext passed verbatim to createObject method of key
    * @param ptr             On entry, ptr must be NULL or be included if
    *                        the reference count of the object it points
    *                        to. On exit, ptr points to the fetched object
    *                        from the cache or is left unchanged on
    *                        failure. Caller must call removeRef on ptr
    *                        if set to a non NULL value.
    * @param status          Any error returned here. May be set to a
    *                        warning value even if ptr is set.
    */
   template<typename T>
   void get(
           const CacheKey<T>& key,
           const void *creationContext,
           const T *&ptr,
           UErrorCode &status) const {
       if (U_FAILURE(status)) {
           return;
       }
       UErrorCode creationStatus = U_ZERO_ERROR;
       const SharedObject *value = NULL;
       _get(key, value, creationContext, creationStatus);
       const T *tvalue = (const T *) value;
       if (U_SUCCESS(creationStatus)) {
           SharedObject::copyPtr(tvalue, ptr);
       }
       SharedObject::clearPtr(tvalue);
       // Take care not to overwrite a warning status passed in with
       // another warning or U_ZERO_ERROR.
       if (status == U_ZERO_ERROR || U_FAILURE(creationStatus)) {
           status = creationStatus;
       }
   }

#ifdef UNIFIED_CACHE_DEBUG
   /**
    * Dumps the contents of this cache to standard error. Used for testing of
    * cache only.
    */
   void dumpContents() const;
#endif

   /**
    * Convenience method to get a value of type T from cache for a
    * particular locale with creationContext == NULL.
    * @param loc    the locale
    * @param ptr    On entry, must be NULL or included in the ref count
    *               of the object to which it points.
    *               On exit, fetched value stored here or is left
    *               unchanged on failure. Caller must call removeRef on
    *               ptr if set to a non NULL value.
    * @param status Any error returned here. May be set to a
    *               warning value even if ptr is set.
    */
   template<typename T>
   static void getByLocale(
           const Locale &loc, const T *&ptr, UErrorCode &status) {
       const UnifiedCache *cache = getInstance(status);
       if (U_FAILURE(status)) {
           return;
       }
       cache->get(LocaleCacheKey<T>(loc), ptr, status);
   }

#ifdef UNIFIED_CACHE_DEBUG
   /**
    * Dumps the cache contents to stderr. For testing only.
    */
   static void dump();
#endif

   /**
    * Returns the number of keys in this cache. For testing only.
    */
   int32_t keyCount() const;

   /**
    * Removes any values from cache that are not referenced outside
    * the cache.
    */
   void flush() const;

   /**
    * Configures at what point evcition of unused entries will begin.
    * Eviction is triggered whenever the number of evictable keys exeeds
    * BOTH count AND (number of in-use items) * (percentageOfInUseItems / 100).
    * Once the number of unused entries drops below one of these,
    * eviction ceases. Because eviction happens incrementally,
    * the actual unused entry count may exceed both these numbers
    * from time to time.
    *
    * A cache entry is defined as unused if it is not essential to guarantee
    * that for a given key X, the cache returns the same reference to the
    * same value as long as the client already holds a reference to that
    * value.
    *
    * If this method is never called, the default settings are 1000 and 100%.
    *
    * Although this method is thread-safe, it is designed to be called at
    * application startup. If it is called in the middle of execution, it
    * will have no immediate effect on the cache. However over time, the
    * cache will perform eviction slices in an attempt to honor the new
    * settings.
    *
    * If a client already holds references to many different unique values
    * in the cache such that the number of those unique values far exeeds
    * "count" then the cache may not be able to maintain this maximum.
    * However, if this happens, the cache still guarantees that the number of
    * unused entries will remain only a small percentage of the total cache
    * size.
    *
    * If the parameters passed are negative, setEvctionPolicy sets status to
    * U_ILLEGAL_ARGUMENT_ERROR.
    */
   void setEvictionPolicy(
           int32_t count, int32_t percentageOfInUseItems, UErrorCode &status);


   /**
    * Returns how many entries have been auto evicted during the lifetime
    * of this cache. This only includes auto evicted entries, not
    * entries evicted because of a call to flush().
    */
   int64_t autoEvictedCount() const;

   /**
    * Returns the unused entry count in this cache. For testing only,
    * Regular clients will not need this.
    */
   int32_t unusedCount() const;

   virtual void handleUnreferencedObject() const;
   virtual ~UnifiedCache();

 private:
   UHashtable *fHashtable;
   mutable int32_t fEvictPos;
   mutable int32_t fNumValuesTotal;
   mutable int32_t fNumValuesInUse;
   int32_t fMaxUnused;
   int32_t fMaxPercentageOfInUse;
   mutable int64_t fAutoEvictedCount;
   SharedObject *fNoValue;

   UnifiedCache(const UnifiedCache &other);
   UnifiedCache &operator=(const UnifiedCache &other);

   /**
    * Flushes the contents of the cache. If cache values hold references to other
    * cache values then _flush should be called in a loop until it returns FALSE.
    *
    * On entry, gCacheMutex must be held.
    * On exit, those values with are evictable are flushed.
    *
    *  @param all if false flush evictable items only, which are those with no external
    *                    references, plus those that can be safely recreated.<br>
    *            if true, flush all elements. Any values (sharedObjects) with remaining
    *                     hard (external) references are not deleted, but are detached from
    *                     the cache, so that a subsequent removeRefs can delete them.
    *                     _flush is not thread safe when all is true.
    *   @return TRUE if any value in cache was flushed or FALSE otherwise.
    */
   UBool _flush(UBool all) const;

   /**
    * Gets value out of cache.
    * On entry. gCacheMutex must not be held. value must be NULL. status
    * must be U_ZERO_ERROR.
    * On exit. value and status set to what is in cache at key or on cache
    * miss the key's createObject() is called and value and status are set to
    * the result of that. In this latter case, best effort is made to add the
    * value and status to the cache. If createObject() fails to create a value,
    * fNoValue is stored in cache, and value is set to NULL. Caller must call
    * removeRef on value if non NULL.
    */
   void _get(
           const CacheKeyBase &key,
           const SharedObject *&value,
           const void *creationContext,
           UErrorCode &status) const;

    /**
     * Attempts to fetch value and status for key from cache.
     * On entry, gCacheMutex must not be held value must be NULL and status must
     * be U_ZERO_ERROR.
     * On exit, either returns FALSE (In this
     * case caller should try to create the object) or returns TRUE with value
     * pointing to the fetched value and status set to fetched status. When
     * FALSE is returned status may be set to failure if an in progress hash
     * entry could not be made but value will remain unchanged. When TRUE is
     * returned, caller must call removeRef() on value.
     */
    UBool _poll(
            const CacheKeyBase &key,
            const SharedObject *&value,
            UErrorCode &status) const;

    /**
     * Places a new value and creationStatus in the cache for the given key.
     * On entry, gCacheMutex must be held. key must not exist in the cache.
     * On exit, value and creation status placed under key. Soft reference added
     * to value on successful add. On error sets status.
     */
    void _putNew(
        const CacheKeyBase &key,
        const SharedObject *value,
        const UErrorCode creationStatus,
        UErrorCode &status) const;

    /**
     * Places value and status at key if there is no value at key or if cache
     * entry for key is in progress. Otherwise, it leaves the current value and
     * status there.
     *
     * On entry. gCacheMutex must not be held. Value must be
     * included in the reference count of the object to which it points.
     *
     * On exit, value and status are changed to what was already in the cache if
     * something was there and not in progress. Otherwise, value and status are left
     * unchanged in which case they are placed in the cache on a best-effort basis.
     * Caller must call removeRef() on value.
     */
   void _putIfAbsentAndGet(
           const CacheKeyBase &key,
           const SharedObject *&value,
           UErrorCode &status) const;

    /**
     * Returns the next element in the cache round robin style.
     * Returns nullptr if the cache is empty.
     * On entry, gCacheMutex must be held.
     */
    const UHashElement *_nextElement() const;

   /**
    * Return the number of cache items that would need to be evicted
    * to bring usage into conformance with eviction policy.
    *
    * An item corresponds to an entry in the hash table, a hash table element.
    *
    * On entry, gCacheMutex must be held.
    */
   int32_t _computeCountOfItemsToEvict() const;

   /**
    * Run an eviction slice.
    * On entry, gCacheMutex must be held.
    * _runEvictionSlice runs a slice of the evict pipeline by examining the next
    * 10 entries in the cache round robin style evicting them if they are eligible.
    */
   void _runEvictionSlice() const;

   /**
    * Register a master cache entry. A master key is the first key to create
    * a given  SharedObject value. Subsequent keys whose create function
    * produce referneces to an already existing SharedObject are not masters -
    * they can be evicted and subsequently recreated.
    *
    * On entry, gCacheMutex must be held.
    * On exit, items in use count incremented, entry is marked as a master
    * entry, and value registered with cache so that subsequent calls to
    * addRef() and removeRef() on it correctly interact with the cache.
    */
   void _registerMaster(const CacheKeyBase *theKey, const SharedObject *value) const;

   /**
    * Store a value and creation error status in given hash entry.
    * On entry, gCacheMutex must be held. Hash entry element must be in progress.
    * value must be non NULL.
    * On Exit, soft reference added to value. value and status stored in hash
    * entry. Soft reference removed from previous stored value. Waiting
    * threads notified.
    */
   void _put(
           const UHashElement *element,
           const SharedObject *value,
           const UErrorCode status) const;
    /**
     * Remove a soft reference, and delete the SharedObject if no references remain.
     * To be used from within the UnifiedCache implementation only.
     * gCacheMutex must be held by caller.
     * @param value the SharedObject to be acted on.
     */
   void removeSoftRef(const SharedObject *value) const;

   /**
    * Increment the hard reference count of the given SharedObject.
    * gCacheMutex must be held by the caller.
    * Update numValuesEvictable on transitions between zero and one reference.
    *
    * @param value The SharedObject to be referenced.
    * @return the hard reference count after the addition.
    */
   int32_t addHardRef(const SharedObject *value) const;

  /**
    * Decrement the hard reference count of the given SharedObject.
    * gCacheMutex must be held by the caller.
    * Update numValuesEvictable on transitions between one and zero reference.
    *
    * @param value The SharedObject to be referenced.
    * @return the hard reference count after the removal.
    */
   int32_t removeHardRef(const SharedObject *value) const;


#ifdef UNIFIED_CACHE_DEBUG
   void _dumpContents() const;
#endif

   /**
    *  Fetch value and error code from a particular hash entry.
    *  On entry, gCacheMutex must be held. value must be either NULL or must be
    *  included in the ref count of the object to which it points.
    *  On exit, value and status set to what is in the hash entry. Caller must
    *  eventually call removeRef on value.
    *  If hash entry is in progress, value will be set to gNoValue and status will
    *  be set to U_ZERO_ERROR.
    */
   void _fetch(const UHashElement *element, const SharedObject *&value,
                       UErrorCode &status) const;

    /**
     * Determine if given hash entry is in progress.
     * On entry, gCacheMutex must be held.
     */
   UBool _inProgress(const UHashElement *element) const;

   /**
    * Determine if given hash entry is in progress.
    * On entry, gCacheMutex must be held.
    */
   UBool _inProgress(const SharedObject *theValue, UErrorCode creationStatus) const;

   /**
    * Determine if given hash entry is eligible for eviction.
    * On entry, gCacheMutex must be held.
    */
   UBool _isEvictable(const UHashElement *element) const;
};

U_NAMESPACE_END



U_NAMESPACE_BEGIN

SharedObject::~SharedObject() {}

UnifiedCacheBase::~UnifiedCacheBase() {}

void
SharedObject::addRef() const {
    umtx_atomic_inc(&hardRefCount);
}

// removeRef Decrement the reference count and delete if it is zero.
//           Note that SharedObjects with a non-null cachePtr are owned by the
//           unified cache, and the cache will be responsible for the actual deletion.
//           The deletion could be as soon as immediately following the
//           update to the reference count, if another thread is running
//           a cache eviction cycle concurrently.
//           NO ACCESS TO *this PERMITTED AFTER REFERENCE COUNT == 0 for cached objects.
//           THE OBJECT MAY ALREADY BE GONE.
void
SharedObject::removeRef() const {
    const UnifiedCacheBase *cache = this->cachePtr;
    int32_t updatedRefCount = umtx_atomic_dec(&hardRefCount);
    U_ASSERT(updatedRefCount >= 0);
    if (updatedRefCount == 0) {
        if (cache) {
            cache->handleUnreferencedObject();
        } else {
            delete this;
        }
    }
}


int32_t
SharedObject::getRefCount() const {
    return umtx_loadAcquire(hardRefCount);
}

void
SharedObject::deleteIfZeroRefCount() const {
    if (this->cachePtr == nullptr && getRefCount() == 0) {
        delete this;
    }
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2014-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* simpleformatter.cpp
*/






U_NAMESPACE_BEGIN

namespace {

/**
 * Argument numbers must be smaller than this limit.
 * Text segment lengths are offset by this much.
 * This is currently the only unused char value in compiled patterns,
 * except it is the maximum value of the first unit (max arg +1).
 */
const int32_t SIMPLE_FORMATTER_ARG_NUM_LIMIT = 0x100;
/**
 * Initial and maximum char/UChar value set for a text segment.
 * Segment length char values are from SIMPLE_FORMATTER_ARG_NUM_LIMIT+1 to this value here.
 * Normally 0xffff, but can be as small as SIMPLE_FORMATTER_ARG_NUM_LIMIT+1 for testing.
 */
const UChar SEGMENT_LENGTH_PLACEHOLDER_CHAR = 0xffff;
/**
 * Maximum length of a text segment. Longer segments are split into shorter ones.
 */
const int32_t MAX_SEGMENT_LENGTH = SEGMENT_LENGTH_PLACEHOLDER_CHAR - SIMPLE_FORMATTER_ARG_NUM_LIMIT;

enum {
    APOS = 0x27,
    DIGIT_ZERO = 0x30,
    DIGIT_ONE = 0x31,
    DIGIT_NINE = 0x39,
    OPEN_BRACE = 0x7b,
    CLOSE_BRACE = 0x7d
};

inline UBool isInvalidArray(const void *array, int32_t length) {
   return (length < 0 || (array == NULL && length != 0));
}

}  // namespace

SimpleFormatter &SimpleFormatter::operator=(const SimpleFormatter& other) {
    if (this == &other) {
        return *this;
    }
    compiledPattern = other.compiledPattern;
    return *this;
}

SimpleFormatter::~SimpleFormatter() {}

UBool SimpleFormatter::applyPatternMinMaxArguments(
        const UnicodeString &pattern,
        int32_t min, int32_t max,
        UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) {
        return FALSE;
    }
    // Parse consistent with MessagePattern, but
    // - support only simple numbered arguments
    // - build a simple binary structure into the result string
    const UChar *patternBuffer = pattern.getBuffer();
    int32_t patternLength = pattern.length();
    // Reserve the first char for the number of arguments.
    compiledPattern.setTo((UChar)0);
    int32_t textLength = 0;
    int32_t maxArg = -1;
    UBool inQuote = FALSE;
    for (int32_t i = 0; i < patternLength;) {
        UChar c = patternBuffer[i++];
        if (c == APOS) {
            if (i < patternLength && (c = patternBuffer[i]) == APOS) {
                // double apostrophe, skip the second one
                ++i;
            } else if (inQuote) {
                // skip the quote-ending apostrophe
                inQuote = FALSE;
                continue;
            } else if (c == OPEN_BRACE || c == CLOSE_BRACE) {
                // Skip the quote-starting apostrophe, find the end of the quoted literal text.
                ++i;
                inQuote = TRUE;
            } else {
                // The apostrophe is part of literal text.
                c = APOS;
            }
        } else if (!inQuote && c == OPEN_BRACE) {
            if (textLength > 0) {
                compiledPattern.setCharAt(compiledPattern.length() - textLength - 1,
                                          (UChar)(SIMPLE_FORMATTER_ARG_NUM_LIMIT + textLength));
                textLength = 0;
            }
            int32_t argNumber;
            if ((i + 1) < patternLength &&
                    0 <= (argNumber = patternBuffer[i] - DIGIT_ZERO) && argNumber <= 9 &&
                    patternBuffer[i + 1] == CLOSE_BRACE) {
                i += 2;
            } else {
                // Multi-digit argument number (no leading zero) or syntax error.
                // MessagePattern permits PatternProps.skipWhiteSpace(pattern, index)
                // around the number, but this class does not.
                argNumber = -1;
                if (i < patternLength && DIGIT_ONE <= (c = patternBuffer[i++]) && c <= DIGIT_NINE) {
                    argNumber = c - DIGIT_ZERO;
                    while (i < patternLength &&
                            DIGIT_ZERO <= (c = patternBuffer[i++]) && c <= DIGIT_NINE) {
                        argNumber = argNumber * 10 + (c - DIGIT_ZERO);
                        if (argNumber >= SIMPLE_FORMATTER_ARG_NUM_LIMIT) {
                            break;
                        }
                    }
                }
                if (argNumber < 0 || c != CLOSE_BRACE) {
                    errorCode = U_ILLEGAL_ARGUMENT_ERROR;
                    return FALSE;
                }
            }
            if (argNumber > maxArg) {
                maxArg = argNumber;
            }
            compiledPattern.append((UChar)argNumber);
            continue;
        }  // else: c is part of literal text
        // Append c and track the literal-text segment length.
        if (textLength == 0) {
            // Reserve a char for the length of a new text segment, preset the maximum length.
            compiledPattern.append(SEGMENT_LENGTH_PLACEHOLDER_CHAR);
        }
        compiledPattern.append(c);
        if (++textLength == MAX_SEGMENT_LENGTH) {
            textLength = 0;
        }
    }
    if (textLength > 0) {
        compiledPattern.setCharAt(compiledPattern.length() - textLength - 1,
                                  (UChar)(SIMPLE_FORMATTER_ARG_NUM_LIMIT + textLength));
    }
    int32_t argCount = maxArg + 1;
    if (argCount < min || max < argCount) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
    compiledPattern.setCharAt(0, (UChar)argCount);
    return TRUE;
}

UnicodeString& SimpleFormatter::format(
        const UnicodeString &value0,
        UnicodeString &appendTo, UErrorCode &errorCode) const {
    const UnicodeString *values[] = { &value0 };
    return formatAndAppend(values, 1, appendTo, NULL, 0, errorCode);
}

UnicodeString& SimpleFormatter::format(
        const UnicodeString &value0,
        const UnicodeString &value1,
        UnicodeString &appendTo, UErrorCode &errorCode) const {
    const UnicodeString *values[] = { &value0, &value1 };
    return formatAndAppend(values, 2, appendTo, NULL, 0, errorCode);
}

UnicodeString& SimpleFormatter::format(
        const UnicodeString &value0,
        const UnicodeString &value1,
        const UnicodeString &value2,
        UnicodeString &appendTo, UErrorCode &errorCode) const {
    const UnicodeString *values[] = { &value0, &value1, &value2 };
    return formatAndAppend(values, 3, appendTo, NULL, 0, errorCode);
}

UnicodeString& SimpleFormatter::formatAndAppend(
        const UnicodeString *const *values, int32_t valuesLength,
        UnicodeString &appendTo,
        int32_t *offsets, int32_t offsetsLength, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) {
        return appendTo;
    }
    if (isInvalidArray(values, valuesLength) || isInvalidArray(offsets, offsetsLength) ||
            valuesLength < getArgumentLimit()) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return appendTo;
    }
    return format(compiledPattern.getBuffer(), compiledPattern.length(), values,
                  appendTo, NULL, TRUE,
                  offsets, offsetsLength, errorCode);
}

UnicodeString &SimpleFormatter::formatAndReplace(
        const UnicodeString *const *values, int32_t valuesLength,
        UnicodeString &result,
        int32_t *offsets, int32_t offsetsLength, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) {
        return result;
    }
    if (isInvalidArray(values, valuesLength) || isInvalidArray(offsets, offsetsLength)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return result;
    }
    const UChar *cp = compiledPattern.getBuffer();
    int32_t cpLength = compiledPattern.length();
    if (valuesLength < getArgumentLimit(cp, cpLength)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return result;
    }

    // If the pattern starts with an argument whose value is the same object
    // as the result, then we keep the result contents and append to it.
    // Otherwise we replace its contents.
    int32_t firstArg = -1;
    // If any non-initial argument value is the same object as the result,
    // then we first copy its contents and use that instead while formatting.
    UnicodeString resultCopy;
    if (getArgumentLimit(cp, cpLength) > 0) {
        for (int32_t i = 1; i < cpLength;) {
            int32_t n = cp[i++];
            if (n < SIMPLE_FORMATTER_ARG_NUM_LIMIT) {
                if (values[n] == &result) {
                    if (i == 2) {
                        firstArg = n;
                    } else if (resultCopy.isEmpty() && !result.isEmpty()) {
                        resultCopy = result;
                    }
                }
            } else {
                i += n - SIMPLE_FORMATTER_ARG_NUM_LIMIT;
            }
        }
    }
    if (firstArg < 0) {
        result.remove();
    }
    return format(cp, cpLength, values,
                  result, &resultCopy, FALSE,
                  offsets, offsetsLength, errorCode);
}

UnicodeString SimpleFormatter::getTextWithNoArguments(
        const UChar *compiledPattern,
        int32_t compiledPatternLength,
        int32_t* offsets,
        int32_t offsetsLength) {
    for (int32_t i = 0; i < offsetsLength; i++) {
        offsets[i] = -1;
    }
    int32_t capacity = compiledPatternLength - 1 -
            getArgumentLimit(compiledPattern, compiledPatternLength);
    UnicodeString sb(capacity, 0, 0);  // Java: StringBuilder
    for (int32_t i = 1; i < compiledPatternLength;) {
        int32_t n = compiledPattern[i++];
        if (n > SIMPLE_FORMATTER_ARG_NUM_LIMIT) {
            n -= SIMPLE_FORMATTER_ARG_NUM_LIMIT;
            sb.append(compiledPattern + i, n);
            i += n;
        } else if (n < offsetsLength) {
            offsets[n] = sb.length();
        }
    }
    return sb;
}

UnicodeString &SimpleFormatter::format(
        const UChar *compiledPattern, int32_t compiledPatternLength,
        const UnicodeString *const *values,
        UnicodeString &result, const UnicodeString *resultCopy, UBool forbidResultAsValue,
        int32_t *offsets, int32_t offsetsLength,
        UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) {
        return result;
    }
    for (int32_t i = 0; i < offsetsLength; i++) {
        offsets[i] = -1;
    }
    for (int32_t i = 1; i < compiledPatternLength;) {
        int32_t n = compiledPattern[i++];
        if (n < SIMPLE_FORMATTER_ARG_NUM_LIMIT) {
            const UnicodeString *value = values[n];
            if (value == NULL) {
                errorCode = U_ILLEGAL_ARGUMENT_ERROR;
                return result;
            }
            if (value == &result) {
                if (forbidResultAsValue) {
                    errorCode = U_ILLEGAL_ARGUMENT_ERROR;
                    return result;
                }
                if (i == 2) {
                    // We are appending to result which is also the first value object.
                    if (n < offsetsLength) {
                        offsets[n] = 0;
                    }
                } else {
                    if (n < offsetsLength) {
                        offsets[n] = result.length();
                    }
                    result.append(*resultCopy);
                }
            } else {
                if (n < offsetsLength) {
                    offsets[n] = result.length();
                }
                result.append(*value);
            }
        } else {
            int32_t length = n - SIMPLE_FORMATTER_ARG_NUM_LIMIT;
            result.append(compiledPattern + i, length);
            i += length;
        }
    }
    return result;
}

U_NAMESPACE_END
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// This file contains utilities to deal with static-allocated UnicodeSets.
//
// Common use case: you write a "private static final" UnicodeSet in Java, and
// want something similarly easy in C++.  Originally written for number
// parsing, but this header can be used for other applications.
//
// Main entrypoint: `unisets::get(unisets::MY_SET_ID_HERE)`
//
// This file is in common instead of i18n because it is needed by ucurr.cpp.
//
// Author: sffc



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN
namespace unisets {

enum Key {
    // UNISET_KEY_NONE is used to indicate null in chooseFrom().
    // UNISET_KEY_EMPTY is used to get an empty UnicodeSet.
    UNISET_KEY_NONE = -1,
    UNISET_KEY_EMPTY = 0,

    // Ignorables
    UNISET_KEY_DEFAULT_IGNORABLES,
    UNISET_KEY_STRICT_IGNORABLES,

    // Separators
    // Notes:
    // - UNISET_KEY_COMMA is a superset of UNISET_KEY_STRICT_COMMA
    // - UNISET_KEY_PERIOD is a superset of SCRICT_UNISET_KEY_PERIOD
    // - UNISET_KEY_ALL_SEPARATORS is the union of UNISET_KEY_COMMA, UNISET_KEY_PERIOD, and UNISET_KEY_OTHER_GROUPING_SEPARATORS
    // - UNISET_KEY_STRICT_ALL_SEPARATORS is the union of UNISET_KEY_STRICT_COMMA, UNISET_KEY_STRICT_PERIOD, and OTHER_GRP_SEPARATORS
    UNISET_KEY_COMMA,
    UNISET_KEY_PERIOD,
    UNISET_KEY_STRICT_COMMA,
    UNISET_KEY_STRICT_PERIOD,
    UNISET_KEY_APOSTROPHE_SIGN,
    UNISET_KEY_OTHER_GROUPING_SEPARATORS,
    UNISET_KEY_ALL_SEPARATORS,
    UNISET_KEY_STRICT_ALL_SEPARATORS,

    // Symbols
    UNISET_KEY_MINUS_SIGN,
    UNISET_KEY_PLUS_SIGN,
    UNISET_KEY_PERCENT_SIGN,
    UNISET_KEY_PERMILLE_SIGN,
    UNISET_KEY_INFINITY_SIGN,

    // Currency Symbols
    UNISET_KEY_DOLLAR_SIGN,
    UNISET_KEY_POUND_SIGN,
    UNISET_KEY_RUPEE_SIGN,
    UNISET_KEY_YEN_SIGN,
    UNISET_KEY_WON_SIGN,

    // Other
    UNISET_KEY_DIGITS,

    // Combined Separators with Digits (for lead code points)
    UNISET_KEY_DIGITS_OR_ALL_SEPARATORS,
    UNISET_KEY_DIGITS_OR_STRICT_ALL_SEPARATORS,

    // The number of elements in the enum.
    UNISET_KEY_UNISETS_KEY_COUNT
};

/**
 * Gets the static-allocated UnicodeSet according to the provided key. The
 * pointer will be deleted during u_cleanup(); the caller should NOT delete it.
 *
 * Exported as U_COMMON_API for ucurr.cpp
 *
 * This method is always safe and OK to chain: in the case of a memory or other
 * error, it returns an empty set from static memory.
 *
 * Example:
 *
 *     UBool hasIgnorables = unisets::get(unisets::UNISET_KEY_DEFAULT_IGNORABLES)->contains(...);
 *
 * @param key The desired UnicodeSet according to the enum in this file.
 * @return The requested UnicodeSet. Guaranteed to be frozen and non-null, but
 *         may be empty if an error occurred during data loading.
 */
U_COMMON_API const UnicodeSet* get(Key key);

/**
 * Checks if the UnicodeSet given by key1 contains the given string.
 *
 * Exported as U_COMMON_API for numparse_decimal.cpp
 *
 * @param str The string to check.
 * @param key1 The set to check.
 * @return key1 if the set contains str, or UNISET_KEY_NONE if not.
 */
U_COMMON_API Key chooseFrom(UnicodeString str, Key key1);

/**
 * Checks if the UnicodeSet given by either key1 or key2 contains the string.
 *
 * Exported as U_COMMON_API for numparse_decimal.cpp
 *
 * @param str The string to check.
 * @param key1 The first set to check.
 * @param key2 The second set to check.
 * @return key1 if that set contains str; key2 if that set contains str; or
 *         UNISET_KEY_NONE if neither set contains str.
 */
U_COMMON_API Key chooseFrom(UnicodeString str, Key key1, Key key2);

// TODO: Load these from data: ICU-20108
// Unused in C++:
// Key chooseCurrency(UnicodeString str);
// Used instead:
static const struct {
    Key key;
    UChar32 exemplar;
} kCurrencyEntries[] = {
	{UNISET_KEY_DOLLAR_SIGN, u'\x24'}, // u'$'
	{UNISET_KEY_POUND_SIGN, u'\xa3'}, // u'£'
	{UNISET_KEY_RUPEE_SIGN, u'\x20b9'}, // u'₹'
	{UNISET_KEY_YEN_SIGN, u'\xa5'}, // u'¥'
	{UNISET_KEY_WON_SIGN, u'\x20a9'}, // u'₩'
};

} // namespace unisets
U_NAMESPACE_END

#endif //__STATIC_UNICODE_SETS_H__








using namespace icu;
using namespace icu::unisets;


namespace {

UnicodeSet* gUnicodeSets[UNISET_KEY_UNISETS_KEY_COUNT] = {};

// Save the empty instance in static memory to have well-defined behavior if a
// regular UnicodeSet cannot be allocated.
alignas(UnicodeSet)
char gEmptyUnicodeSet[sizeof(UnicodeSet)];

// Whether the gEmptyUnicodeSet is initialized and ready to use.
UBool gEmptyUnicodeSetInitialized = FALSE;

inline UnicodeSet* getImpl(Key key) {
    UnicodeSet* candidate = gUnicodeSets[key];
    if (candidate == nullptr) {
        return reinterpret_cast<UnicodeSet*>(gEmptyUnicodeSet);
    }
    return candidate;
}

UnicodeSet* computeUnion(Key k1, Key k2) {
    UnicodeSet* result = new UnicodeSet();
    if (result == nullptr) {
        return nullptr;
    }
    result->addAll(*getImpl(k1));
    result->addAll(*getImpl(k2));
    result->freeze();
    return result;
}

UnicodeSet* computeUnion(Key k1, Key k2, Key k3) {
    UnicodeSet* result = new UnicodeSet();
    if (result == nullptr) {
        return nullptr;
    }
    result->addAll(*getImpl(k1));
    result->addAll(*getImpl(k2));
    result->addAll(*getImpl(k3));
    result->freeze();
    return result;
}


void saveSet(Key key, const UnicodeString& unicodeSetPattern, UErrorCode& status) {
    // assert unicodeSets.get(key) == null;
    gUnicodeSets[key] = new UnicodeSet(unicodeSetPattern, status);
}

class ParseDataSink : public ResourceSink {
  public:
    void put(const char* key, ResourceValue& value, UBool /*noFallback*/, UErrorCode& status) U_OVERRIDE {
        ResourceTable contextsTable = value.getTable(status);
        if (U_FAILURE(status)) { return; }
        for (int i = 0; contextsTable.getKeyAndValue(i, key, value); i++) {
            if (uprv_strcmp(key, "date") == 0) {
                // ignore
            } else {
                ResourceTable strictnessTable = value.getTable(status);
                if (U_FAILURE(status)) { return; }
                for (int j = 0; strictnessTable.getKeyAndValue(j, key, value); j++) {
                    bool isLenient = (uprv_strcmp(key, "lenient") == 0);
                    ResourceArray array = value.getArray(status);
                    if (U_FAILURE(status)) { return; }
                    for (int k = 0; k < array.getSize(); k++) {
                        array.getValue(k, value);
                        UnicodeString str = value.getUnicodeString(status);
                        if (U_FAILURE(status)) { return; }
                        // There is both lenient and strict data for comma/period,
                        // but not for any of the other symbols.
                        if (str.indexOf(u'.') != -1) {
                            saveSet(isLenient ? UNISET_KEY_PERIOD : UNISET_KEY_STRICT_PERIOD, str, status);
                        } else if (str.indexOf(u',') != -1) {
                            saveSet(isLenient ? UNISET_KEY_COMMA : UNISET_KEY_STRICT_COMMA, str, status);
                        } else if (str.indexOf(u'+') != -1) {
                            saveSet(UNISET_KEY_PLUS_SIGN, str, status);
                        } else if (str.indexOf(u'-') != -1) {
                            saveSet(UNISET_KEY_MINUS_SIGN, str, status);
                        } else if (str.indexOf(u'$') != -1) {
                            saveSet(UNISET_KEY_DOLLAR_SIGN, str, status);
                        } else if (str.indexOf(u'\xa3') != -1) {
                            saveSet(UNISET_KEY_POUND_SIGN, str, status);
                        } else if (str.indexOf(u'\x20b9') != -1) {
                            saveSet(UNISET_KEY_RUPEE_SIGN, str, status);
                        } else if (str.indexOf(u'\xa5') != -1) {
                            saveSet(UNISET_KEY_YEN_SIGN, str, status);
                        } else if (str.indexOf(u'\x20a9') != -1) {
                            saveSet(UNISET_KEY_WON_SIGN, str, status);
                        } else if (str.indexOf(u'%') != -1) {
                            saveSet(UNISET_KEY_PERCENT_SIGN, str, status);
                        } else if (str.indexOf(u'\x2030') != -1) {
                            saveSet(UNISET_KEY_PERMILLE_SIGN, str, status);
                        } else if (str.indexOf(u'\x2019') != -1) {
                            saveSet(UNISET_KEY_APOSTROPHE_SIGN, str, status);
                        } else {
                            // Unknown class of parse lenients
                            // TODO(ICU-20428): Make ICU automatically accept new classes?
                            U_ASSERT(FALSE);
                        }
                        if (U_FAILURE(status)) { return; }
                    }
                }
            }
        }
    }
};


icu::UInitOnce gNumberParseUniSetsInitOnce = U_INITONCE_INITIALIZER;

UBool U_CALLCONV cleanupNumberParseUniSets() {
    if (gEmptyUnicodeSetInitialized) {
        reinterpret_cast<UnicodeSet*>(gEmptyUnicodeSet)->~UnicodeSet();
        gEmptyUnicodeSetInitialized = FALSE;
    }
    for (int32_t i = 0; i < UNISET_KEY_UNISETS_KEY_COUNT; i++) {
        delete gUnicodeSets[i];
        gUnicodeSets[i] = nullptr;
    }
    gNumberParseUniSetsInitOnce.reset();
    return TRUE;
}

void U_CALLCONV initNumberParseUniSets(UErrorCode& status) {
    ucln_common_registerCleanup(UCLN_COMMON_NUMPARSE_UNISETS, cleanupNumberParseUniSets);

    // Initialize the empty instance for well-defined fallback behavior
    new(gEmptyUnicodeSet) UnicodeSet();
    reinterpret_cast<UnicodeSet*>(gEmptyUnicodeSet)->freeze();
    gEmptyUnicodeSetInitialized = TRUE;

    // These sets were decided after discussion with icu-design@. See tickets #13084 and #13309.
    // Zs+TAB is "horizontal whitespace" according to UTS #18 (blank property).
    gUnicodeSets[UNISET_KEY_DEFAULT_IGNORABLES] = new UnicodeSet(
            u"[[:Zs:][\\u0009][:Bidi_Control:][:Variation_Selector:]]", status);
    gUnicodeSets[UNISET_KEY_STRICT_IGNORABLES] = new UnicodeSet(u"[[:Bidi_Control:]]", status);

    LocalUResourceBundlePointer rb(ures_open(nullptr, "root", &status));
    if (U_FAILURE(status)) { return; }
    ParseDataSink sink;
    ures_getAllItemsWithFallback(rb.getAlias(), "parse", sink, status);
    if (U_FAILURE(status)) { return; }

    // NOTE: It is OK for these assertions to fail if there was a no-data build.
    U_ASSERT(gUnicodeSets[UNISET_KEY_COMMA] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_STRICT_COMMA] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_PERIOD] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_STRICT_PERIOD] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_APOSTROPHE_SIGN] != nullptr);

    LocalPointer<UnicodeSet> otherGrouping(new UnicodeSet(
        u"[٬‘\\u0020\\u00A0\\u2000-\\u200A\\u202F\\u205F\\u3000]",
        status
    ), status);
    if (U_FAILURE(status)) { return; }
    otherGrouping->addAll(*gUnicodeSets[UNISET_KEY_APOSTROPHE_SIGN]);
    gUnicodeSets[UNISET_KEY_OTHER_GROUPING_SEPARATORS] = otherGrouping.orphan();
    gUnicodeSets[UNISET_KEY_ALL_SEPARATORS] = computeUnion(UNISET_KEY_COMMA, UNISET_KEY_PERIOD, UNISET_KEY_OTHER_GROUPING_SEPARATORS);
    gUnicodeSets[UNISET_KEY_STRICT_ALL_SEPARATORS] = computeUnion(
            UNISET_KEY_STRICT_COMMA, UNISET_KEY_STRICT_PERIOD, UNISET_KEY_OTHER_GROUPING_SEPARATORS);

    U_ASSERT(gUnicodeSets[UNISET_KEY_MINUS_SIGN] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_PLUS_SIGN] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_PERCENT_SIGN] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_PERMILLE_SIGN] != nullptr);

    gUnicodeSets[UNISET_KEY_INFINITY_SIGN] = new UnicodeSet(u"[∞]", status);
    if (U_FAILURE(status)) { return; }

    U_ASSERT(gUnicodeSets[UNISET_KEY_DOLLAR_SIGN] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_POUND_SIGN] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_RUPEE_SIGN] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_YEN_SIGN] != nullptr);
    U_ASSERT(gUnicodeSets[UNISET_KEY_WON_SIGN] != nullptr);

    gUnicodeSets[UNISET_KEY_DIGITS] = new UnicodeSet(u"[:digit:]", status);
    if (U_FAILURE(status)) { return; }
    gUnicodeSets[UNISET_KEY_DIGITS_OR_ALL_SEPARATORS] = computeUnion(UNISET_KEY_DIGITS, UNISET_KEY_ALL_SEPARATORS);
    gUnicodeSets[UNISET_KEY_DIGITS_OR_STRICT_ALL_SEPARATORS] = computeUnion(UNISET_KEY_DIGITS, UNISET_KEY_STRICT_ALL_SEPARATORS);

    for (auto* uniset : gUnicodeSets) {
        if (uniset != nullptr) {
            uniset->freeze();
        }
    }
}

}

const UnicodeSet* unisets::get(Key key) {
    UErrorCode localStatus = U_ZERO_ERROR;
    umtx_initOnce(gNumberParseUniSetsInitOnce, &initNumberParseUniSets, localStatus);
    if (U_FAILURE(localStatus)) {
        return reinterpret_cast<UnicodeSet*>(gEmptyUnicodeSet);
    }
    return getImpl(key);
}

Key unisets::chooseFrom(UnicodeString str, Key key1) {
    return get(key1)->contains(str) ? key1 : UNISET_KEY_NONE;
}

Key unisets::chooseFrom(UnicodeString str, Key key1, Key key2) {
    return get(key1)->contains(str) ? key1 : chooseFrom(str, key2);
}

//Key unisets::chooseCurrency(UnicodeString str) {
//    if (get(UNISET_KEY_DOLLAR_SIGN)->contains(str)) {
//        return UNISET_KEY_DOLLAR_SIGN;
//    } else if (get(UNISET_KEY_POUND_SIGN)->contains(str)) {
//        return UNISET_KEY_POUND_SIGN;
//    } else if (get(UNISET_KEY_RUPEE_SIGN)->contains(str)) {
//        return UNISET_KEY_RUPEE_SIGN;
//    } else if (get(UNISET_KEY_YEN_SIGN)->contains(str)) {
//        return UNISET_KEY_YEN_SIGN;
//    } else {
//        return UNISET_KEY_NONE;
//    }
//}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
// Copyright (C) 2009-2013, International Business Machines
// Corporation and others. All Rights Reserved.
//
// Copyright 2004 and onwards Google Inc.
//
// Author: wilsonh@google.com (Wilson Hsieh)
//






U_NAMESPACE_BEGIN

StringPiece::StringPiece(const char* str)
    : ptr_(str), length_((str == NULL) ? 0 : static_cast<int32_t>(uprv_strlen(str))) { }

StringPiece::StringPiece(const StringPiece& x, int32_t pos) {
  if (pos < 0) {
    pos = 0;
  } else if (pos > x.length_) {
    pos = x.length_;
  }
  ptr_ = x.ptr_ + pos;
  length_ = x.length_ - pos;
}

StringPiece::StringPiece(const StringPiece& x, int32_t pos, int32_t len) {
  if (pos < 0) {
    pos = 0;
  } else if (pos > x.length_) {
    pos = x.length_;
  }
  if (len < 0) {
    len = 0;
  } else if (len > x.length_ - pos) {
    len = x.length_ - pos;
  }
  ptr_ = x.ptr_ + pos;
  length_ = len;
}

void StringPiece::set(const char* str) {
  ptr_ = str;
  if (str != NULL)
    length_ = static_cast<int32_t>(uprv_strlen(str));
  else
    length_ = 0;
}

U_EXPORT UBool U_EXPORT2
operator==(const StringPiece& x, const StringPiece& y) {
  int32_t len = x.size();
  if (len != y.size()) {
    return false;
  }
  if (len == 0) {
    return true;
  }
  const char* p = x.data();
  const char* p2 = y.data();
  // Test last byte in case strings share large common prefix
  --len;
  if (p[len] != p2[len]) return false;
  // At this point we can, but don't have to, ignore the last byte.
  return uprv_memcmp(p, p2, len) == 0;
}


const int32_t StringPiece::npos = 0x7fffffff;

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  stringtriebuilder.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010dec24
*   created by: Markus W. Scherer
*/

  // for 'typeid' to work





U_CDECL_BEGIN

static int32_t U_CALLCONV
hashStringTrieNode(const UHashTok key) {
    return icu::StringTrieBuilder::hashNode(key.pointer);
}

static UBool U_CALLCONV
equalStringTrieNodes(const UHashTok key1, const UHashTok key2) {
    return icu::StringTrieBuilder::equalNodes(key1.pointer, key2.pointer);
}

U_CDECL_END

U_NAMESPACE_BEGIN

StringTrieBuilder::StringTrieBuilder() : nodes(NULL) {}

StringTrieBuilder::~StringTrieBuilder() {
    deleteCompactBuilder();
}

void
StringTrieBuilder::createCompactBuilder(int32_t sizeGuess, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    nodes=uhash_openSize(hashStringTrieNode, equalStringTrieNodes, NULL,
                         sizeGuess, &errorCode);
    if(U_SUCCESS(errorCode)) {
        if(nodes==NULL) {
          errorCode=U_MEMORY_ALLOCATION_ERROR;
        } else {
          uhash_setKeyDeleter(nodes, uprv_deleteUObject);
        }
    }
}

void
StringTrieBuilder::deleteCompactBuilder() {
    uhash_close(nodes);
    nodes=NULL;
}

void
StringTrieBuilder::build(UStringTrieBuildOption buildOption, int32_t elementsLength,
                       UErrorCode &errorCode) {
    if(buildOption==USTRINGTRIE_BUILD_FAST) {
        writeNode(0, elementsLength, 0);
    } else /* USTRINGTRIE_BUILD_SMALL */ {
        createCompactBuilder(2*elementsLength, errorCode);
        Node *root=makeNode(0, elementsLength, 0, errorCode);
        if(U_SUCCESS(errorCode)) {
            root->markRightEdgesFirst(-1);
            root->write(*this);
        }
        deleteCompactBuilder();
    }
}

// Requires start<limit,
// and all strings of the [start..limit[ elements must be sorted and
// have a common prefix of length unitIndex.
int32_t
StringTrieBuilder::writeNode(int32_t start, int32_t limit, int32_t unitIndex) {
    UBool hasValue=FALSE;
    int32_t value=0;
    int32_t type;
    if(unitIndex==getElementStringLength(start)) {
        // An intermediate or final value.
        value=getElementValue(start++);
        if(start==limit) {
            return writeValueAndFinal(value, TRUE);  // final-value node
        }
        hasValue=TRUE;
    }
    // Now all [start..limit[ strings are longer than unitIndex.
    int32_t minUnit=getElementUnit(start, unitIndex);
    int32_t maxUnit=getElementUnit(limit-1, unitIndex);
    if(minUnit==maxUnit) {
        // Linear-match node: All strings have the same character at unitIndex.
        int32_t lastUnitIndex=getLimitOfLinearMatch(start, limit-1, unitIndex);
        writeNode(start, limit, lastUnitIndex);
        // Break the linear-match sequence into chunks of at most kMaxLinearMatchLength.
        int32_t length=lastUnitIndex-unitIndex;
        int32_t maxLinearMatchLength=getMaxLinearMatchLength();
        while(length>maxLinearMatchLength) {
            lastUnitIndex-=maxLinearMatchLength;
            length-=maxLinearMatchLength;
            writeElementUnits(start, lastUnitIndex, maxLinearMatchLength);
            write(getMinLinearMatch()+maxLinearMatchLength-1);
        }
        writeElementUnits(start, unitIndex, length);
        type=getMinLinearMatch()+length-1;
    } else {
        // Branch node.
        int32_t length=countElementUnits(start, limit, unitIndex);
        // length>=2 because minUnit!=maxUnit.
        writeBranchSubNode(start, limit, unitIndex, length);
        if(--length<getMinLinearMatch()) {
            type=length;
        } else {
            write(length);
            type=0;
        }
    }
    return writeValueAndType(hasValue, value, type);
}

// start<limit && all strings longer than unitIndex &&
// length different units at unitIndex
int32_t
StringTrieBuilder::writeBranchSubNode(int32_t start, int32_t limit, int32_t unitIndex, int32_t length) {
    UChar middleUnits[kMaxSplitBranchLevels];
    int32_t lessThan[kMaxSplitBranchLevels];
    int32_t ltLength=0;
    while(length>getMaxBranchLinearSubNodeLength()) {
        // Branch on the middle unit.
        // First, find the middle unit.
        int32_t i=skipElementsBySomeUnits(start, unitIndex, length/2);
        // Encode the less-than branch first.
        middleUnits[ltLength]=getElementUnit(i, unitIndex);  // middle unit
        lessThan[ltLength]=writeBranchSubNode(start, i, unitIndex, length/2);
        ++ltLength;
        // Continue for the greater-or-equal branch.
        start=i;
        length=length-length/2;
    }
    // For each unit, find its elements array start and whether it has a final value.
    int32_t starts[kMaxBranchLinearSubNodeLength];
    UBool isFinal[kMaxBranchLinearSubNodeLength-1];
    int32_t unitNumber=0;
    do {
        int32_t i=starts[unitNumber]=start;
        UChar unit=getElementUnit(i++, unitIndex);
        i=indexOfElementWithNextUnit(i, unitIndex, unit);
        isFinal[unitNumber]= start==i-1 && unitIndex+1==getElementStringLength(start);
        start=i;
    } while(++unitNumber<length-1);
    // unitNumber==length-1, and the maxUnit elements range is [start..limit[
    starts[unitNumber]=start;

    // Write the sub-nodes in reverse order: The jump lengths are deltas from
    // after their own positions, so if we wrote the minUnit sub-node first,
    // then its jump delta would be larger.
    // Instead we write the minUnit sub-node last, for a shorter delta.
    int32_t jumpTargets[kMaxBranchLinearSubNodeLength-1];
    do {
        --unitNumber;
        if(!isFinal[unitNumber]) {
            jumpTargets[unitNumber]=writeNode(starts[unitNumber], starts[unitNumber+1], unitIndex+1);
        }
    } while(unitNumber>0);
    // The maxUnit sub-node is written as the very last one because we do
    // not jump for it at all.
    unitNumber=length-1;
    writeNode(start, limit, unitIndex+1);
    int32_t offset=write(getElementUnit(start, unitIndex));
    // Write the rest of this node's unit-value pairs.
    while(--unitNumber>=0) {
        start=starts[unitNumber];
        int32_t value;
        if(isFinal[unitNumber]) {
            // Write the final value for the one string ending with this unit.
            value=getElementValue(start);
        } else {
            // Write the delta to the start position of the sub-node.
            value=offset-jumpTargets[unitNumber];
        }
        writeValueAndFinal(value, isFinal[unitNumber]);
        offset=write(getElementUnit(start, unitIndex));
    }
    // Write the split-branch nodes.
    while(ltLength>0) {
        --ltLength;
        writeDeltaTo(lessThan[ltLength]);
        offset=write(middleUnits[ltLength]);
    }
    return offset;
}

// Requires start<limit,
// and all strings of the [start..limit[ elements must be sorted and
// have a common prefix of length unitIndex.
StringTrieBuilder::Node *
StringTrieBuilder::makeNode(int32_t start, int32_t limit, int32_t unitIndex, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    UBool hasValue=FALSE;
    int32_t value=0;
    if(unitIndex==getElementStringLength(start)) {
        // An intermediate or final value.
        value=getElementValue(start++);
        if(start==limit) {
            return registerFinalValue(value, errorCode);
        }
        hasValue=TRUE;
    }
    Node *node;
    // Now all [start..limit[ strings are longer than unitIndex.
    int32_t minUnit=getElementUnit(start, unitIndex);
    int32_t maxUnit=getElementUnit(limit-1, unitIndex);
    if(minUnit==maxUnit) {
        // Linear-match node: All strings have the same character at unitIndex.
        int32_t lastUnitIndex=getLimitOfLinearMatch(start, limit-1, unitIndex);
        Node *nextNode=makeNode(start, limit, lastUnitIndex, errorCode);
        // Break the linear-match sequence into chunks of at most kMaxLinearMatchLength.
        int32_t length=lastUnitIndex-unitIndex;
        int32_t maxLinearMatchLength=getMaxLinearMatchLength();
        while(length>maxLinearMatchLength) {
            lastUnitIndex-=maxLinearMatchLength;
            length-=maxLinearMatchLength;
            node=createLinearMatchNode(start, lastUnitIndex, maxLinearMatchLength, nextNode);
            nextNode=registerNode(node, errorCode);
        }
        node=createLinearMatchNode(start, unitIndex, length, nextNode);
    } else {
        // Branch node.
        int32_t length=countElementUnits(start, limit, unitIndex);
        // length>=2 because minUnit!=maxUnit.
        Node *subNode=makeBranchSubNode(start, limit, unitIndex, length, errorCode);
        node=new BranchHeadNode(length, subNode);
    }
    if(hasValue && node!=NULL) {
        if(matchNodesCanHaveValues()) {
            ((ValueNode *)node)->setValue(value);
        } else {
            node=new IntermediateValueNode(value, registerNode(node, errorCode));
        }
    }
    return registerNode(node, errorCode);
}

// start<limit && all strings longer than unitIndex &&
// length different units at unitIndex
StringTrieBuilder::Node *
StringTrieBuilder::makeBranchSubNode(int32_t start, int32_t limit, int32_t unitIndex,
                                   int32_t length, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    UChar middleUnits[kMaxSplitBranchLevels];
    Node *lessThan[kMaxSplitBranchLevels];
    int32_t ltLength=0;
    while(length>getMaxBranchLinearSubNodeLength()) {
        // Branch on the middle unit.
        // First, find the middle unit.
        int32_t i=skipElementsBySomeUnits(start, unitIndex, length/2);
        // Create the less-than branch.
        middleUnits[ltLength]=getElementUnit(i, unitIndex);  // middle unit
        lessThan[ltLength]=makeBranchSubNode(start, i, unitIndex, length/2, errorCode);
        ++ltLength;
        // Continue for the greater-or-equal branch.
        start=i;
        length=length-length/2;
    }
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    ListBranchNode *listNode=new ListBranchNode();
    if(listNode==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    // For each unit, find its elements array start and whether it has a final value.
    int32_t unitNumber=0;
    do {
        int32_t i=start;
        UChar unit=getElementUnit(i++, unitIndex);
        i=indexOfElementWithNextUnit(i, unitIndex, unit);
        if(start==i-1 && unitIndex+1==getElementStringLength(start)) {
            listNode->add(unit, getElementValue(start));
        } else {
            listNode->add(unit, makeNode(start, i, unitIndex+1, errorCode));
        }
        start=i;
    } while(++unitNumber<length-1);
    // unitNumber==length-1, and the maxUnit elements range is [start..limit[
    UChar unit=getElementUnit(start, unitIndex);
    if(start==limit-1 && unitIndex+1==getElementStringLength(start)) {
        listNode->add(unit, getElementValue(start));
    } else {
        listNode->add(unit, makeNode(start, limit, unitIndex+1, errorCode));
    }
    Node *node=registerNode(listNode, errorCode);
    // Create the split-branch nodes.
    while(ltLength>0) {
        --ltLength;
        node=registerNode(
            new SplitBranchNode(middleUnits[ltLength], lessThan[ltLength], node), errorCode);
    }
    return node;
}

StringTrieBuilder::Node *
StringTrieBuilder::registerNode(Node *newNode, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        delete newNode;
        return NULL;
    }
    if(newNode==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    const UHashElement *old=uhash_find(nodes, newNode);
    if(old!=NULL) {
        delete newNode;
        return (Node *)old->key.pointer;
    }
    // If uhash_puti() returns a non-zero value from an equivalent, previously
    // registered node, then uhash_find() failed to find that and we will leak newNode.
#if U_DEBUG
    int32_t oldValue=  // Only in debug mode to avoid a compiler warning about unused oldValue.
#endif
    uhash_puti(nodes, newNode, 1, &errorCode);
    U_ASSERT(oldValue==0);
    if(U_FAILURE(errorCode)) {
        delete newNode;
        return NULL;
    }
    return newNode;
}

StringTrieBuilder::Node *
StringTrieBuilder::registerFinalValue(int32_t value, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    FinalValueNode key(value);
    const UHashElement *old=uhash_find(nodes, &key);
    if(old!=NULL) {
        return (Node *)old->key.pointer;
    }
    Node *newNode=new FinalValueNode(value);
    if(newNode==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    // If uhash_puti() returns a non-zero value from an equivalent, previously
    // registered node, then uhash_find() failed to find that and we will leak newNode.
#if U_DEBUG
    int32_t oldValue=  // Only in debug mode to avoid a compiler warning about unused oldValue.
#endif
    uhash_puti(nodes, newNode, 1, &errorCode);
    U_ASSERT(oldValue==0);
    if(U_FAILURE(errorCode)) {
        delete newNode;
        return NULL;
    }
    return newNode;
}

int32_t
StringTrieBuilder::hashNode(const void *node) {
    return ((const Node *)node)->hashCode();
}

UBool
StringTrieBuilder::equalNodes(const void *left, const void *right) {
    return *(const Node *)left==*(const Node *)right;
}

UBool
StringTrieBuilder::Node::operator==(const Node &other) const {
    return this==&other || (typeid(*this)==typeid(other) && hash==other.hash);
}

int32_t
StringTrieBuilder::Node::markRightEdgesFirst(int32_t edgeNumber) {
    if(offset==0) {
        offset=edgeNumber;
    }
    return edgeNumber;
}

UBool
StringTrieBuilder::FinalValueNode::operator==(const Node &other) const {
    if(this==&other) {
        return TRUE;
    }
    if(!Node::operator==(other)) {
        return FALSE;
    }
    const FinalValueNode &o=(const FinalValueNode &)other;
    return value==o.value;
}

void
StringTrieBuilder::FinalValueNode::write(StringTrieBuilder &builder) {
    offset=builder.writeValueAndFinal(value, TRUE);
}

UBool
StringTrieBuilder::ValueNode::operator==(const Node &other) const {
    if(this==&other) {
        return TRUE;
    }
    if(!Node::operator==(other)) {
        return FALSE;
    }
    const ValueNode &o=(const ValueNode &)other;
    return hasValue==o.hasValue && (!hasValue || value==o.value);
}

UBool
StringTrieBuilder::IntermediateValueNode::operator==(const Node &other) const {
    if(this==&other) {
        return TRUE;
    }
    if(!ValueNode::operator==(other)) {
        return FALSE;
    }
    const IntermediateValueNode &o=(const IntermediateValueNode &)other;
    return next==o.next;
}

int32_t
StringTrieBuilder::IntermediateValueNode::markRightEdgesFirst(int32_t edgeNumber) {
    if(offset==0) {
        offset=edgeNumber=next->markRightEdgesFirst(edgeNumber);
    }
    return edgeNumber;
}

void
StringTrieBuilder::IntermediateValueNode::write(StringTrieBuilder &builder) {
    next->write(builder);
    offset=builder.writeValueAndFinal(value, FALSE);
}

UBool
StringTrieBuilder::LinearMatchNode::operator==(const Node &other) const {
    if(this==&other) {
        return TRUE;
    }
    if(!ValueNode::operator==(other)) {
        return FALSE;
    }
    const LinearMatchNode &o=(const LinearMatchNode &)other;
    return length==o.length && next==o.next;
}

int32_t
StringTrieBuilder::LinearMatchNode::markRightEdgesFirst(int32_t edgeNumber) {
    if(offset==0) {
        offset=edgeNumber=next->markRightEdgesFirst(edgeNumber);
    }
    return edgeNumber;
}

UBool
StringTrieBuilder::ListBranchNode::operator==(const Node &other) const {
    if(this==&other) {
        return TRUE;
    }
    if(!Node::operator==(other)) {
        return FALSE;
    }
    const ListBranchNode &o=(const ListBranchNode &)other;
    for(int32_t i=0; i<length; ++i) {
        if(units[i]!=o.units[i] || values[i]!=o.values[i] || equal[i]!=o.equal[i]) {
            return FALSE;
        }
    }
    return TRUE;
}

int32_t
StringTrieBuilder::ListBranchNode::markRightEdgesFirst(int32_t edgeNumber) {
    if(offset==0) {
        firstEdgeNumber=edgeNumber;
        int32_t step=0;
        int32_t i=length;
        do {
            Node *edge=equal[--i];
            if(edge!=NULL) {
                edgeNumber=edge->markRightEdgesFirst(edgeNumber-step);
            }
            // For all but the rightmost edge, decrement the edge number.
            step=1;
        } while(i>0);
        offset=edgeNumber;
    }
    return edgeNumber;
}

void
StringTrieBuilder::ListBranchNode::write(StringTrieBuilder &builder) {
    // Write the sub-nodes in reverse order: The jump lengths are deltas from
    // after their own positions, so if we wrote the minUnit sub-node first,
    // then its jump delta would be larger.
    // Instead we write the minUnit sub-node last, for a shorter delta.
    int32_t unitNumber=length-1;
    Node *rightEdge=equal[unitNumber];
    int32_t rightEdgeNumber= rightEdge==NULL ? firstEdgeNumber : rightEdge->getOffset();
    do {
        --unitNumber;
        if(equal[unitNumber]!=NULL) {
            equal[unitNumber]->writeUnlessInsideRightEdge(firstEdgeNumber, rightEdgeNumber, builder);
        }
    } while(unitNumber>0);
    // The maxUnit sub-node is written as the very last one because we do
    // not jump for it at all.
    unitNumber=length-1;
    if(rightEdge==NULL) {
        builder.writeValueAndFinal(values[unitNumber], TRUE);
    } else {
        rightEdge->write(builder);
    }
    offset=builder.write(units[unitNumber]);
    // Write the rest of this node's unit-value pairs.
    while(--unitNumber>=0) {
        int32_t value;
        UBool isFinal;
        if(equal[unitNumber]==NULL) {
            // Write the final value for the one string ending with this unit.
            value=values[unitNumber];
            isFinal=TRUE;
        } else {
            // Write the delta to the start position of the sub-node.
            U_ASSERT(equal[unitNumber]->getOffset()>0);
            value=offset-equal[unitNumber]->getOffset();
            isFinal=FALSE;
        }
        builder.writeValueAndFinal(value, isFinal);
        offset=builder.write(units[unitNumber]);
    }
}

UBool
StringTrieBuilder::SplitBranchNode::operator==(const Node &other) const {
    if(this==&other) {
        return TRUE;
    }
    if(!Node::operator==(other)) {
        return FALSE;
    }
    const SplitBranchNode &o=(const SplitBranchNode &)other;
    return unit==o.unit && lessThan==o.lessThan && greaterOrEqual==o.greaterOrEqual;
}

int32_t
StringTrieBuilder::SplitBranchNode::markRightEdgesFirst(int32_t edgeNumber) {
    if(offset==0) {
        firstEdgeNumber=edgeNumber;
        edgeNumber=greaterOrEqual->markRightEdgesFirst(edgeNumber);
        offset=edgeNumber=lessThan->markRightEdgesFirst(edgeNumber-1);
    }
    return edgeNumber;
}

void
StringTrieBuilder::SplitBranchNode::write(StringTrieBuilder &builder) {
    // Encode the less-than branch first.
    lessThan->writeUnlessInsideRightEdge(firstEdgeNumber, greaterOrEqual->getOffset(), builder);
    // Encode the greater-or-equal branch last because we do not jump for it at all.
    greaterOrEqual->write(builder);
    // Write this node.
    U_ASSERT(lessThan->getOffset()>0);
    builder.writeDeltaTo(lessThan->getOffset());  // less-than
    offset=builder.write(unit);
}

UBool
StringTrieBuilder::BranchHeadNode::operator==(const Node &other) const {
    if(this==&other) {
        return TRUE;
    }
    if(!ValueNode::operator==(other)) {
        return FALSE;
    }
    const BranchHeadNode &o=(const BranchHeadNode &)other;
    return length==o.length && next==o.next;
}

int32_t
StringTrieBuilder::BranchHeadNode::markRightEdgesFirst(int32_t edgeNumber) {
    if(offset==0) {
        offset=edgeNumber=next->markRightEdgesFirst(edgeNumber);
    }
    return edgeNumber;
}

void
StringTrieBuilder::BranchHeadNode::write(StringTrieBuilder &builder) {
    next->write(builder);
    if(length<=builder.getMinLinearMatch()) {
        offset=builder.writeValueAndType(hasValue, value, length-1);
    } else {
        builder.write(length-1);
        offset=builder.writeValueAndType(hasValue, value, 0);
    }
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2003-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uarrsort.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2003aug04
*   created by: Markus W. Scherer
*
*   Internal function for sorting arrays.
*/





enum {
    /**
     * "from Knuth"
     *
     * A binary search over 8 items performs 4 comparisons:
     * log2(8)=3 to subdivide, +1 to check for equality.
     * A linear search over 8 items on average also performs 4 comparisons.
     */
    MIN_QSORT=9,
    STACK_ITEM_SIZE=200
};

static constexpr int32_t sizeInMaxAlignTs(int32_t sizeInBytes) {
    return (sizeInBytes + sizeof(max_align_t) - 1) / sizeof(max_align_t);
}

/* UComparator convenience implementations ---------------------------------- */

U_CAPI int32_t U_EXPORT2
uprv_uint16Comparator(const void *context, const void *left, const void *right) {
    (void)context;
    return (int32_t)*(const uint16_t *)left - (int32_t)*(const uint16_t *)right;
}

U_CAPI int32_t U_EXPORT2
uprv_int32Comparator(const void *context, const void *left, const void *right) {
    (void)context;
    return *(const int32_t *)left - *(const int32_t *)right;
}

U_CAPI int32_t U_EXPORT2
uprv_uint32Comparator(const void *context, const void *left, const void *right) {
    (void)context;
    uint32_t l=*(const uint32_t *)left, r=*(const uint32_t *)right;

    /* compare directly because (l-r) would overflow the int32_t result */
    if(l<r) {
        return -1;
    } else if(l==r) {
        return 0;
    } else /* l>r */ {
        return 1;
    }
}

/* Insertion sort using binary search --------------------------------------- */

U_CAPI int32_t U_EXPORT2
uprv_stableBinarySearch(char *array, int32_t limit, void *item, int32_t itemSize,
                        UComparator *cmp, const void *context) {
    int32_t start=0;
    UBool found=FALSE;

    /* Binary search until we get down to a tiny sub-array. */
    while((limit-start)>=MIN_QSORT) {
        int32_t i=(start+limit)/2;
        int32_t diff=cmp(context, item, array+i*itemSize);
        if(diff==0) {
            /*
             * Found the item. We look for the *last* occurrence of such
             * an item, for stable sorting.
             * If we knew that there will be only few equal items,
             * we could break now and enter the linear search.
             * However, if there are many equal items, then it should be
             * faster to continue with the binary search.
             * It seems likely that we either have all unique items
             * (where found will never become TRUE in the insertion sort)
             * or potentially many duplicates.
             */
            found=TRUE;
            start=i+1;
        } else if(diff<0) {
            limit=i;
        } else {
            start=i;
        }
    }

    /* Linear search over the remaining tiny sub-array. */
    while(start<limit) {
        int32_t diff=cmp(context, item, array+start*itemSize);
        if(diff==0) {
            found=TRUE;
        } else if(diff<0) {
            break;
        }
        ++start;
    }
    return found ? (start-1) : ~start;
}

static void
doInsertionSort(char *array, int32_t length, int32_t itemSize,
                UComparator *cmp, const void *context, void *pv) {
    int32_t j;

    for(j=1; j<length; ++j) {
        char *item=array+j*itemSize;
        int32_t insertionPoint=uprv_stableBinarySearch(array, j, item, itemSize, cmp, context);
        if(insertionPoint<0) {
            insertionPoint=~insertionPoint;
        } else {
            ++insertionPoint;  /* one past the last equal item */
        }
        if(insertionPoint<j) {
            char *dest=array+insertionPoint*itemSize;
            uprv_memcpy(pv, item, itemSize);  /* v=array[j] */
            uprv_memmove(dest+itemSize, dest, (j-insertionPoint)*(size_t)itemSize);
            uprv_memcpy(dest, pv, itemSize);  /* array[insertionPoint]=v */
        }
    }
}

static void
insertionSort(char *array, int32_t length, int32_t itemSize,
              UComparator *cmp, const void *context, UErrorCode *pErrorCode) {

    icu::MaybeStackArray<max_align_t, sizeInMaxAlignTs(STACK_ITEM_SIZE)> v;
    if (sizeInMaxAlignTs(itemSize) > v.getCapacity() &&
            v.resize(sizeInMaxAlignTs(itemSize)) == nullptr) {
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    doInsertionSort(array, length, itemSize, cmp, context, v.getAlias());
}

/* QuickSort ---------------------------------------------------------------- */

/*
 * This implementation is semi-recursive:
 * It recurses for the smaller sub-array to shorten the recursion depth,
 * and loops for the larger sub-array.
 *
 * Loosely after QuickSort algorithms in
 * Niklaus Wirth
 * Algorithmen und Datenstrukturen mit Modula-2
 * B.G. Teubner Stuttgart
 * 4. Auflage 1986
 * ISBN 3-519-02260-5
 */
static void
subQuickSort(char *array, int32_t start, int32_t limit, int32_t itemSize,
             UComparator *cmp, const void *context,
             void *px, void *pw) {
    int32_t left, right;

    /* start and left are inclusive, limit and right are exclusive */
    do {
        if((start+MIN_QSORT)>=limit) {
            doInsertionSort(array+start*itemSize, limit-start, itemSize, cmp, context, px);
            break;
        }

        left=start;
        right=limit;

        /* x=array[middle] */
        uprv_memcpy(px, array+(size_t)((start+limit)/2)*itemSize, itemSize);

        do {
            while(/* array[left]<x */
                  cmp(context, array+left*itemSize, px)<0
            ) {
                ++left;
            }
            while(/* x<array[right-1] */
                  cmp(context, px, array+(right-1)*itemSize)<0
            ) {
                --right;
            }

            /* swap array[left] and array[right-1] via w; ++left; --right */
            if(left<right) {
                --right;

                if(left<right) {
                    uprv_memcpy(pw, array+(size_t)left*itemSize, itemSize);
                    uprv_memcpy(array+(size_t)left*itemSize, array+(size_t)right*itemSize, itemSize);
                    uprv_memcpy(array+(size_t)right*itemSize, pw, itemSize);
                }

                ++left;
            }
        } while(left<right);

        /* sort sub-arrays */
        if((right-start)<(limit-left)) {
            /* sort [start..right[ */
            if(start<(right-1)) {
                subQuickSort(array, start, right, itemSize, cmp, context, px, pw);
            }

            /* sort [left..limit[ */
            start=left;
        } else {
            /* sort [left..limit[ */
            if(left<(limit-1)) {
                subQuickSort(array, left, limit, itemSize, cmp, context, px, pw);
            }

            /* sort [start..right[ */
            limit=right;
        }
    } while(start<(limit-1));
}

static void
quickSort(char *array, int32_t length, int32_t itemSize,
            UComparator *cmp, const void *context, UErrorCode *pErrorCode) {
    /* allocate two intermediate item variables (x and w) */
    icu::MaybeStackArray<max_align_t, sizeInMaxAlignTs(STACK_ITEM_SIZE) * 2> xw;
    if(sizeInMaxAlignTs(itemSize)*2 > xw.getCapacity() &&
            xw.resize(sizeInMaxAlignTs(itemSize) * 2) == nullptr) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    subQuickSort(array, 0, length, itemSize, cmp, context,
                 xw.getAlias(), xw.getAlias() + sizeInMaxAlignTs(itemSize));
}

/* uprv_sortArray() API ----------------------------------------------------- */

/*
 * Check arguments, select an appropriate implementation,
 * cast the array to char * so that array+i*itemSize works.
 */
U_CAPI void U_EXPORT2
uprv_sortArray(void *array, int32_t length, int32_t itemSize,
               UComparator *cmp, const void *context,
               UBool sortStable, UErrorCode *pErrorCode) {
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return;
    }
    if((length>0 && array==NULL) || length<0 || itemSize<=0 || cmp==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    if(length<=1) {
        return;
    } else if(length<MIN_QSORT || sortStable) {
        insertionSort((char *)array, length, itemSize, cmp, context, pErrorCode);
    } else {
        quickSort((char *)array, length, itemSize, cmp, context, pErrorCode);
    }
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  ubidi.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999jul27
*   created by: Markus W. Scherer, updated by Matitiahu Allouche
*
*/





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  ubidi.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999jul27
*   created by: Markus W. Scherer, updated by Matitiahu Allouche
*/






/**
 *\file
 * \brief C API: Bidi algorithm
 *
 * <h2>Bidi algorithm for ICU</h2>
 *
 * This is an implementation of the Unicode Bidirectional Algorithm.
 * The algorithm is defined in the
 * <a href="http://www.unicode.org/unicode/reports/tr9/">Unicode Standard Annex #9</a>.<p>
 *
 * Note: Libraries that perform a bidirectional algorithm and
 * reorder strings accordingly are sometimes called "Storage Layout Engines".
 * ICU's Bidi and shaping (u_shapeArabic()) APIs can be used at the core of such
 * "Storage Layout Engines".
 *
 * <h3>General remarks about the API:</h3>
 *
 * In functions with an error code parameter,
 * the <code>pErrorCode</code> pointer must be valid
 * and the value that it points to must not indicate a failure before
 * the function call. Otherwise, the function returns immediately.
 * After the function call, the value indicates success or failure.<p>
 *
 * The &quot;limit&quot; of a sequence of characters is the position just after their
 * last character, i.e., one more than that position.<p>
 *
 * Some of the API functions provide access to &quot;runs&quot;.
 * Such a &quot;run&quot; is defined as a sequence of characters
 * that are at the same embedding level
 * after performing the Bidi algorithm.<p>
 *
 * @author Markus W. Scherer
 * @version 1.0
 *
 *
 * <h4> Sample code for the ICU Bidi API </h4>
 *
 * <h5>Rendering a paragraph with the ICU Bidi API</h5>
 *
 * This is (hypothetical) sample code that illustrates
 * how the ICU Bidi API could be used to render a paragraph of text.
 * Rendering code depends highly on the graphics system,
 * therefore this sample code must make a lot of assumptions,
 * which may or may not match any existing graphics system's properties.
 *
 * <p>The basic assumptions are:</p>
 * <ul>
 * <li>Rendering is done from left to right on a horizontal line.</li>
 * <li>A run of single-style, unidirectional text can be rendered at once.</li>
 * <li>Such a run of text is passed to the graphics system with
 *     characters (code units) in logical order.</li>
 * <li>The line-breaking algorithm is very complicated
 *     and Locale-dependent -
 *     and therefore its implementation omitted from this sample code.</li>
 * </ul>
 *
 * <pre>
 * \code
 *#include "unicode/ubidi.h"
 *
 *typedef enum {
 *     styleNormal=0, styleSelected=1,
 *     styleBold=2, styleItalics=4,
 *     styleSuper=8, styleSub=16
 *} Style;
 *
 *typedef struct { int32_t limit; Style style; } StyleRun;
 *
 *int getTextWidth(const UChar *text, int32_t start, int32_t limit,
 *                  const StyleRun *styleRuns, int styleRunCount);
 *
 * // set *pLimit and *pStyleRunLimit for a line
 * // from text[start] and from styleRuns[styleRunStart]
 * // using ubidi_getLogicalRun(para, ...)
 *void getLineBreak(const UChar *text, int32_t start, int32_t *pLimit,
 *                  UBiDi *para,
 *                  const StyleRun *styleRuns, int styleRunStart, int *pStyleRunLimit,
 *                  int *pLineWidth);
 *
 * // render runs on a line sequentially, always from left to right
 *
 * // prepare rendering a new line
 * void startLine(UBiDiDirection textDirection, int lineWidth);
 *
 * // render a run of text and advance to the right by the run width
 * // the text[start..limit-1] is always in logical order
 * void renderRun(const UChar *text, int32_t start, int32_t limit,
 *               UBiDiDirection textDirection, Style style);
 *
 * // We could compute a cross-product
 * // from the style runs with the directional runs
 * // and then reorder it.
 * // Instead, here we iterate over each run type
 * // and render the intersections -
 * // with shortcuts in simple (and common) cases.
 * // renderParagraph() is the main function.
 *
 * // render a directional run with
 * // (possibly) multiple style runs intersecting with it
 * void renderDirectionalRun(const UChar *text,
 *                           int32_t start, int32_t limit,
 *                           UBiDiDirection direction,
 *                           const StyleRun *styleRuns, int styleRunCount) {
 *     int i;
 *
 *     // iterate over style runs
 *     if(direction==UBIDI_LTR) {
 *         int styleLimit;
 *
 *         for(i=0; i<styleRunCount; ++i) {
 *             styleLimit=styleRun[i].limit;
 *             if(start<styleLimit) {
 *                 if(styleLimit>limit) { styleLimit=limit; }
 *                 renderRun(text, start, styleLimit,
 *                           direction, styleRun[i].style);
 *                 if(styleLimit==limit) { break; }
 *                 start=styleLimit;
 *             }
 *         }
 *     } else {
 *         int styleStart;
 *
 *         for(i=styleRunCount-1; i>=0; --i) {
 *             if(i>0) {
 *                 styleStart=styleRun[i-1].limit;
 *             } else {
 *                 styleStart=0;
 *             }
 *             if(limit>=styleStart) {
 *                 if(styleStart<start) { styleStart=start; }
 *                 renderRun(text, styleStart, limit,
 *                           direction, styleRun[i].style);
 *                 if(styleStart==start) { break; }
 *                 limit=styleStart;
 *             }
 *         }
 *     }
 * }
 *
 * // the line object represents text[start..limit-1]
 * void renderLine(UBiDi *line, const UChar *text,
 *                 int32_t start, int32_t limit,
 *                 const StyleRun *styleRuns, int styleRunCount) {
 *     UBiDiDirection direction=ubidi_getDirection(line);
 *     if(direction!=UBIDI_MIXED) {
 *         // unidirectional
 *         if(styleRunCount<=1) {
 *             renderRun(text, start, limit, direction, styleRuns[0].style);
 *         } else {
 *             renderDirectionalRun(text, start, limit,
 *                                  direction, styleRuns, styleRunCount);
 *         }
 *     } else {
 *         // mixed-directional
 *         int32_t count, i, length;
 *         UBiDiLevel level;
 *
 *         count=ubidi_countRuns(para, pErrorCode);
 *         if(U_SUCCESS(*pErrorCode)) {
 *             if(styleRunCount<=1) {
 *                 Style style=styleRuns[0].style;
 *
 *                 // iterate over directional runs
 *                for(i=0; i<count; ++i) {
 *                    direction=ubidi_getVisualRun(para, i, &start, &length);
 *                     renderRun(text, start, start+length, direction, style);
 *                }
 *             } else {
 *                 int32_t j;
 *
 *                 // iterate over both directional and style runs
 *                 for(i=0; i<count; ++i) {
 *                     direction=ubidi_getVisualRun(line, i, &start, &length);
 *                     renderDirectionalRun(text, start, start+length,
 *                                          direction, styleRuns, styleRunCount);
 *                 }
 *             }
 *         }
 *     }
 * }
 *
 *void renderParagraph(const UChar *text, int32_t length,
 *                     UBiDiDirection textDirection,
 *                      const StyleRun *styleRuns, int styleRunCount,
 *                      int lineWidth,
 *                      UErrorCode *pErrorCode) {
 *     UBiDi *para;
 *
 *     if(pErrorCode==NULL || U_FAILURE(*pErrorCode) || length<=0) {
 *         return;
 *     }
 *
 *     para=ubidi_openSized(length, 0, pErrorCode);
 *     if(para==NULL) { return; }
 *
 *     ubidi_setPara(para, text, length,
 *                   textDirection ? UBIDI_DEFAULT_RTL : UBIDI_DEFAULT_LTR,
 *                   NULL, pErrorCode);
 *     if(U_SUCCESS(*pErrorCode)) {
 *         UBiDiLevel paraLevel=1&ubidi_getParaLevel(para);
 *         StyleRun styleRun={ length, styleNormal };
 *         int width;
 *
 *         if(styleRuns==NULL || styleRunCount<=0) {
 *            styleRunCount=1;
 *             styleRuns=&styleRun;
 *         }
 *
 *        // assume styleRuns[styleRunCount-1].limit>=length
 *
 *         width=getTextWidth(text, 0, length, styleRuns, styleRunCount);
 *         if(width<=lineWidth) {
 *             // everything fits onto one line
 *
 *            // prepare rendering a new line from either left or right
 *             startLine(paraLevel, width);
 *
 *             renderLine(para, text, 0, length,
 *                        styleRuns, styleRunCount);
 *         } else {
 *             UBiDi *line;
 *
 *             // we need to render several lines
 *             line=ubidi_openSized(length, 0, pErrorCode);
 *             if(line!=NULL) {
 *                 int32_t start=0, limit;
 *                 int styleRunStart=0, styleRunLimit;
 *
 *                 for(;;) {
 *                     limit=length;
 *                     styleRunLimit=styleRunCount;
 *                     getLineBreak(text, start, &limit, para,
 *                                  styleRuns, styleRunStart, &styleRunLimit,
 *                                 &width);
 *                     ubidi_setLine(para, start, limit, line, pErrorCode);
 *                     if(U_SUCCESS(*pErrorCode)) {
 *                         // prepare rendering a new line
 *                         // from either left or right
 *                         startLine(paraLevel, width);
 *
 *                         renderLine(line, text, start, limit,
 *                                    styleRuns+styleRunStart,
 *                                    styleRunLimit-styleRunStart);
 *                     }
 *                     if(limit==length) { break; }
 *                     start=limit;
 *                     styleRunStart=styleRunLimit-1;
 *                     if(start>=styleRuns[styleRunStart].limit) {
 *                         ++styleRunStart;
 *                     }
 *                 }
 *
 *                 ubidi_close(line);
 *             }
 *        }
 *    }
 *
 *     ubidi_close(para);
 *}
 *\endcode
 * </pre>
 */

/*DOCXX_TAG*/
/*@{*/

/**
 * UBiDiLevel is the type of the level values in this
 * Bidi implementation.
 * It holds an embedding level and indicates the visual direction
 * by its bit&nbsp;0 (even/odd value).<p>
 *
 * It can also hold non-level values for the
 * <code>paraLevel</code> and <code>embeddingLevels</code>
 * arguments of <code>ubidi_setPara()</code>; there:
 * <ul>
 * <li>bit&nbsp;7 of an <code>embeddingLevels[]</code>
 * value indicates whether the using application is
 * specifying the level of a character to <i>override</i> whatever the
 * Bidi implementation would resolve it to.</li>
 * <li><code>paraLevel</code> can be set to the
 * pseudo-level values <code>UBIDI_DEFAULT_LTR</code>
 * and <code>UBIDI_DEFAULT_RTL</code>.</li>
 * </ul>
 *
 * @see ubidi_setPara
 *
 * <p>The related constants are not real, valid level values.
 * <code>UBIDI_DEFAULT_XXX</code> can be used to specify
 * a default for the paragraph level for
 * when the <code>ubidi_setPara()</code> function
 * shall determine it but there is no
 * strongly typed character in the input.<p>
 *
 * Note that the value for <code>UBIDI_DEFAULT_LTR</code> is even
 * and the one for <code>UBIDI_DEFAULT_RTL</code> is odd,
 * just like with normal LTR and RTL level values -
 * these special values are designed that way. Also, the implementation
 * assumes that UBIDI_MAX_EXPLICIT_LEVEL is odd.
 *
 * Note: The numeric values of the related constants will not change:
 * They are tied to the use of 7-bit byte values (plus the override bit)
 * and of the UBiDiLevel=uint8_t data type in this API.
 *
 * @see UBIDI_DEFAULT_LTR
 * @see UBIDI_DEFAULT_RTL
 * @see UBIDI_LEVEL_OVERRIDE
 * @see UBIDI_MAX_EXPLICIT_LEVEL
 * @stable ICU 2.0
 */
typedef uint8_t UBiDiLevel;

/** Paragraph level setting.<p>
 *
 * Constant indicating that the base direction depends on the first strong
 * directional character in the text according to the Unicode Bidirectional
 * Algorithm. If no strong directional character is present,
 * then set the paragraph level to 0 (left-to-right).<p>
 *
 * If this value is used in conjunction with reordering modes
 * <code>UBIDI_REORDER_INVERSE_LIKE_DIRECT</code> or
 * <code>UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL</code>, the text to reorder
 * is assumed to be visual LTR, and the text after reordering is required
 * to be the corresponding logical string with appropriate contextual
 * direction. The direction of the result string will be RTL if either
 * the righmost or leftmost strong character of the source text is RTL
 * or Arabic Letter, the direction will be LTR otherwise.<p>
 *
 * If reordering option <code>UBIDI_OPTION_INSERT_MARKS</code> is set, an RLM may
 * be added at the beginning of the result string to ensure round trip
 * (that the result string, when reordered back to visual, will produce
 * the original source text).
 * @see UBIDI_REORDER_INVERSE_LIKE_DIRECT
 * @see UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL
 * @stable ICU 2.0
 */
#define UBIDI_DEFAULT_LTR 0xfe

/** Paragraph level setting.<p>
 *
 * Constant indicating that the base direction depends on the first strong
 * directional character in the text according to the Unicode Bidirectional
 * Algorithm. If no strong directional character is present,
 * then set the paragraph level to 1 (right-to-left).<p>
 *
 * If this value is used in conjunction with reordering modes
 * <code>UBIDI_REORDER_INVERSE_LIKE_DIRECT</code> or
 * <code>UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL</code>, the text to reorder
 * is assumed to be visual LTR, and the text after reordering is required
 * to be the corresponding logical string with appropriate contextual
 * direction. The direction of the result string will be RTL if either
 * the righmost or leftmost strong character of the source text is RTL
 * or Arabic Letter, or if the text contains no strong character;
 * the direction will be LTR otherwise.<p>
 *
 * If reordering option <code>UBIDI_OPTION_INSERT_MARKS</code> is set, an RLM may
 * be added at the beginning of the result string to ensure round trip
 * (that the result string, when reordered back to visual, will produce
 * the original source text).
 * @see UBIDI_REORDER_INVERSE_LIKE_DIRECT
 * @see UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL
 * @stable ICU 2.0
 */
#define UBIDI_DEFAULT_RTL 0xff

/**
 * Maximum explicit embedding level.
 * Same as the max_depth value in the
 * <a href="http://www.unicode.org/reports/tr9/#BD2">Unicode Bidirectional Algorithm</a>.
 * (The maximum resolved level can be up to <code>UBIDI_MAX_EXPLICIT_LEVEL+1</code>).
 * @stable ICU 2.0
 */
#define UBIDI_MAX_EXPLICIT_LEVEL 125

/** Bit flag for level input.
 *  Overrides directional properties.
 * @stable ICU 2.0
 */
#define UBIDI_LEVEL_OVERRIDE 0x80

/**
 * Special value which can be returned by the mapping functions when a logical
 * index has no corresponding visual index or vice-versa. This may happen
 * for the logical-to-visual mapping of a Bidi control when option
 * <code>#UBIDI_OPTION_REMOVE_CONTROLS</code> is specified. This can also happen
 * for the visual-to-logical mapping of a Bidi mark (LRM or RLM) inserted
 * by option <code>#UBIDI_OPTION_INSERT_MARKS</code>.
 * @see ubidi_getVisualIndex
 * @see ubidi_getVisualMap
 * @see ubidi_getLogicalIndex
 * @see ubidi_getLogicalMap
 * @stable ICU 3.6
 */
#define UBIDI_MAP_NOWHERE   (-1)

/**
 * <code>UBiDiDirection</code> values indicate the text direction.
 * @stable ICU 2.0
 */
enum UBiDiDirection {
  /** Left-to-right text. This is a 0 value.
   * <ul>
   * <li>As return value for <code>ubidi_getDirection()</code>, it means
   *     that the source string contains no right-to-left characters, or
   *     that the source string is empty and the paragraph level is even.
   * <li> As return value for <code>ubidi_getBaseDirection()</code>, it
   *      means that the first strong character of the source string has
   *      a left-to-right direction.
   * </ul>
   * @stable ICU 2.0
   */
  UBIDI_LTR,
  /** Right-to-left text. This is a 1 value.
   * <ul>
   * <li>As return value for <code>ubidi_getDirection()</code>, it means
   *     that the source string contains no left-to-right characters, or
   *     that the source string is empty and the paragraph level is odd.
   * <li> As return value for <code>ubidi_getBaseDirection()</code>, it
   *      means that the first strong character of the source string has
   *      a right-to-left direction.
   * </ul>
   * @stable ICU 2.0
   */
  UBIDI_RTL,
  /** Mixed-directional text.
   * <p>As return value for <code>ubidi_getDirection()</code>, it means
   *    that the source string contains both left-to-right and
   *    right-to-left characters.
   * @stable ICU 2.0
   */
  UBIDI_MIXED,
  /** No strongly directional text.
   * <p>As return value for <code>ubidi_getBaseDirection()</code>, it means
   *    that the source string is missing or empty, or contains neither left-to-right
   *    nor right-to-left characters.
   * @stable ICU 4.6
   */
  UBIDI_NEUTRAL
};

/** @stable ICU 2.0 */
typedef enum UBiDiDirection UBiDiDirection;

/**
 * Forward declaration of the <code>UBiDi</code> structure for the declaration of
 * the API functions. Its fields are implementation-specific.<p>
 * This structure holds information about a paragraph (or multiple paragraphs)
 * of text with Bidi-algorithm-related details, or about one line of
 * such a paragraph.<p>
 * Reordering can be done on a line, or on one or more paragraphs which are
 * then interpreted each as one single line.
 * @stable ICU 2.0
 */
struct UBiDi;

/** @stable ICU 2.0 */
typedef struct UBiDi UBiDi;

/**
 * Allocate a <code>UBiDi</code> structure.
 * Such an object is initially empty. It is assigned
 * the Bidi properties of a piece of text containing one or more paragraphs
 * by <code>ubidi_setPara()</code>
 * or the Bidi properties of a line within a paragraph by
 * <code>ubidi_setLine()</code>.<p>
 * This object can be reused for as long as it is not deallocated
 * by calling <code>ubidi_close()</code>.<p>
 * <code>ubidi_setPara()</code> and <code>ubidi_setLine()</code> will allocate
 * additional memory for internal structures as necessary.
 *
 * @return An empty <code>UBiDi</code> object.
 * @stable ICU 2.0
 */
U_STABLE UBiDi * U_EXPORT2
ubidi_open(void);

/**
 * Allocate a <code>UBiDi</code> structure with preallocated memory
 * for internal structures.
 * This function provides a <code>UBiDi</code> object like <code>ubidi_open()</code>
 * with no arguments, but it also preallocates memory for internal structures
 * according to the sizings supplied by the caller.<p>
 * Subsequent functions will not allocate any more memory, and are thus
 * guaranteed not to fail because of lack of memory.<p>
 * The preallocation can be limited to some of the internal memory
 * by setting some values to 0 here. That means that if, e.g.,
 * <code>maxRunCount</code> cannot be reasonably predetermined and should not
 * be set to <code>maxLength</code> (the only failproof value) to avoid
 * wasting memory, then <code>maxRunCount</code> could be set to 0 here
 * and the internal structures that are associated with it will be allocated
 * on demand, just like with <code>ubidi_open()</code>.
 *
 * @param maxLength is the maximum text or line length that internal memory
 *        will be preallocated for. An attempt to associate this object with a
 *        longer text will fail, unless this value is 0, which leaves the allocation
 *        up to the implementation.
 *
 * @param maxRunCount is the maximum anticipated number of same-level runs
 *        that internal memory will be preallocated for. An attempt to access
 *        visual runs on an object that was not preallocated for as many runs
 *        as the text was actually resolved to will fail,
 *        unless this value is 0, which leaves the allocation up to the implementation.<br><br>
 *        The number of runs depends on the actual text and maybe anywhere between
 *        1 and <code>maxLength</code>. It is typically small.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @return An empty <code>UBiDi</code> object with preallocated memory.
 * @stable ICU 2.0
 */
U_STABLE UBiDi * U_EXPORT2
ubidi_openSized(int32_t maxLength, int32_t maxRunCount, UErrorCode *pErrorCode);

/**
 * <code>ubidi_close()</code> must be called to free the memory
 * associated with a UBiDi object.<p>
 *
 * <strong>Important: </strong>
 * A parent <code>UBiDi</code> object must not be destroyed or reused if
 * it still has children.
 * If a <code>UBiDi</code> object has become the <i>child</i>
 * of another one (its <i>parent</i>) by calling
 * <code>ubidi_setLine()</code>, then the child object must
 * be destroyed (closed) or reused (by calling
 * <code>ubidi_setPara()</code> or <code>ubidi_setLine()</code>)
 * before the parent object.
 *
 * @param pBiDi is a <code>UBiDi</code> object.
 *
 * @see ubidi_setPara
 * @see ubidi_setLine
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
ubidi_close(UBiDi *pBiDi);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUBiDiPointer
 * "Smart pointer" class, closes a UBiDi via ubidi_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUBiDiPointer, UBiDi, ubidi_close);

U_NAMESPACE_END

#endif

/**
 * Modify the operation of the Bidi algorithm such that it
 * approximates an "inverse Bidi" algorithm. This function
 * must be called before <code>ubidi_setPara()</code>.
 *
 * <p>The normal operation of the Bidi algorithm as described
 * in the Unicode Technical Report is to take text stored in logical
 * (keyboard, typing) order and to determine the reordering of it for visual
 * rendering.
 * Some legacy systems store text in visual order, and for operations
 * with standard, Unicode-based algorithms, the text needs to be transformed
 * to logical order. This is effectively the inverse algorithm of the
 * described Bidi algorithm. Note that there is no standard algorithm for
 * this "inverse Bidi" and that the current implementation provides only an
 * approximation of "inverse Bidi".</p>
 *
 * <p>With <code>isInverse</code> set to <code>TRUE</code>,
 * this function changes the behavior of some of the subsequent functions
 * in a way that they can be used for the inverse Bidi algorithm.
 * Specifically, runs of text with numeric characters will be treated in a
 * special way and may need to be surrounded with LRM characters when they are
 * written in reordered sequence.</p>
 *
 * <p>Output runs should be retrieved using <code>ubidi_getVisualRun()</code>.
 * Since the actual input for "inverse Bidi" is visually ordered text and
 * <code>ubidi_getVisualRun()</code> gets the reordered runs, these are actually
 * the runs of the logically ordered output.</p>
 *
 * <p>Calling this function with argument <code>isInverse</code> set to
 * <code>TRUE</code> is equivalent to calling
 * <code>ubidi_setReorderingMode</code> with argument
 * <code>reorderingMode</code>
 * set to <code>#UBIDI_REORDER_INVERSE_NUMBERS_AS_L</code>.<br>
 * Calling this function with argument <code>isInverse</code> set to
 * <code>FALSE</code> is equivalent to calling
 * <code>ubidi_setReorderingMode</code> with argument
 * <code>reorderingMode</code>
 * set to <code>#UBIDI_REORDER_DEFAULT</code>.
 *
 * @param pBiDi is a <code>UBiDi</code> object.
 *
 * @param isInverse specifies "forward" or "inverse" Bidi operation.
 *
 * @see ubidi_setPara
 * @see ubidi_writeReordered
 * @see ubidi_setReorderingMode
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
ubidi_setInverse(UBiDi *pBiDi, UBool isInverse);

/**
 * Is this Bidi object set to perform the inverse Bidi algorithm?
 * <p>Note: calling this function after setting the reordering mode with
 * <code>ubidi_setReorderingMode</code> will return <code>TRUE</code> if the
 * reordering mode was set to <code>#UBIDI_REORDER_INVERSE_NUMBERS_AS_L</code>,
 * <code>FALSE</code> for all other values.</p>
 *
 * @param pBiDi is a <code>UBiDi</code> object.
 * @return TRUE if the Bidi object is set to perform the inverse Bidi algorithm
 * by handling numbers as L.
 *
 * @see ubidi_setInverse
 * @see ubidi_setReorderingMode
 * @stable ICU 2.0
 */

U_STABLE UBool U_EXPORT2
ubidi_isInverse(UBiDi *pBiDi);

/**
 * Specify whether block separators must be allocated level zero,
 * so that successive paragraphs will progress from left to right.
 * This function must be called before <code>ubidi_setPara()</code>.
 * Paragraph separators (B) may appear in the text.  Setting them to level zero
 * means that all paragraph separators (including one possibly appearing
 * in the last text position) are kept in the reordered text after the text
 * that they follow in the source text.
 * When this feature is not enabled, a paragraph separator at the last
 * position of the text before reordering will go to the first position
 * of the reordered text when the paragraph level is odd.
 *
 * @param pBiDi is a <code>UBiDi</code> object.
 *
 * @param orderParagraphsLTR specifies whether paragraph separators (B) must
 * receive level 0, so that successive paragraphs progress from left to right.
 *
 * @see ubidi_setPara
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
ubidi_orderParagraphsLTR(UBiDi *pBiDi, UBool orderParagraphsLTR);

/**
 * Is this Bidi object set to allocate level 0 to block separators so that
 * successive paragraphs progress from left to right?
 *
 * @param pBiDi is a <code>UBiDi</code> object.
 * @return TRUE if the Bidi object is set to allocate level 0 to block
 *         separators.
 *
 * @see ubidi_orderParagraphsLTR
 * @stable ICU 3.4
 */
U_STABLE UBool U_EXPORT2
ubidi_isOrderParagraphsLTR(UBiDi *pBiDi);

/**
 * <code>UBiDiReorderingMode</code> values indicate which variant of the Bidi
 * algorithm to use.
 *
 * @see ubidi_setReorderingMode
 * @stable ICU 3.6
 */
typedef enum UBiDiReorderingMode {
    /** Regular Logical to Visual Bidi algorithm according to Unicode.
      * This is a 0 value.
      * @stable ICU 3.6 */
    UBIDI_REORDER_DEFAULT = 0,
    /** Logical to Visual algorithm which handles numbers in a way which
      * mimics the behavior of Windows XP.
      * @stable ICU 3.6 */
    UBIDI_REORDER_NUMBERS_SPECIAL,
    /** Logical to Visual algorithm grouping numbers with adjacent R characters
      * (reversible algorithm).
      * @stable ICU 3.6 */
    UBIDI_REORDER_GROUP_NUMBERS_WITH_R,
    /** Reorder runs only to transform a Logical LTR string to the Logical RTL
      * string with the same display, or vice-versa.<br>
      * If this mode is set together with option
      * <code>#UBIDI_OPTION_INSERT_MARKS</code>, some Bidi controls in the source
      * text may be removed and other controls may be added to produce the
      * minimum combination which has the required display.
      * @stable ICU 3.6 */
    UBIDI_REORDER_RUNS_ONLY,
    /** Visual to Logical algorithm which handles numbers like L
      * (same algorithm as selected by <code>ubidi_setInverse(TRUE)</code>.
      * @see ubidi_setInverse
      * @stable ICU 3.6 */
    UBIDI_REORDER_INVERSE_NUMBERS_AS_L,
    /** Visual to Logical algorithm equivalent to the regular Logical to Visual
      * algorithm.
      * @stable ICU 3.6 */
    UBIDI_REORDER_INVERSE_LIKE_DIRECT,
    /** Inverse Bidi (Visual to Logical) algorithm for the
      * <code>UBIDI_REORDER_NUMBERS_SPECIAL</code> Bidi algorithm.
      * @stable ICU 3.6 */
    UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL,
#ifndef U_HIDE_DEPRECATED_API
    /**
     * Number of values for reordering mode.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UBIDI_REORDER_COUNT
#endif  // U_HIDE_DEPRECATED_API
} UBiDiReorderingMode;

/**
 * Modify the operation of the Bidi algorithm such that it implements some
 * variant to the basic Bidi algorithm or approximates an "inverse Bidi"
 * algorithm, depending on different values of the "reordering mode".
 * This function must be called before <code>ubidi_setPara()</code>, and stays
 * in effect until called again with a different argument.
 *
 * <p>The normal operation of the Bidi algorithm as described
 * in the Unicode Standard Annex #9 is to take text stored in logical
 * (keyboard, typing) order and to determine how to reorder it for visual
 * rendering.</p>
 *
 * <p>With the reordering mode set to a value other than
 * <code>#UBIDI_REORDER_DEFAULT</code>, this function changes the behavior of
 * some of the subsequent functions in a way such that they implement an
 * inverse Bidi algorithm or some other algorithm variants.</p>
 *
 * <p>Some legacy systems store text in visual order, and for operations
 * with standard, Unicode-based algorithms, the text needs to be transformed
 * into logical order. This is effectively the inverse algorithm of the
 * described Bidi algorithm. Note that there is no standard algorithm for
 * this "inverse Bidi", so a number of variants are implemented here.</p>
 *
 * <p>In other cases, it may be desirable to emulate some variant of the
 * Logical to Visual algorithm (e.g. one used in MS Windows), or perform a
 * Logical to Logical transformation.</p>
 *
 * <ul>
 * <li>When the reordering mode is set to <code>#UBIDI_REORDER_DEFAULT</code>,
 * the standard Bidi Logical to Visual algorithm is applied.</li>
 *
 * <li>When the reordering mode is set to
 * <code>#UBIDI_REORDER_NUMBERS_SPECIAL</code>,
 * the algorithm used to perform Bidi transformations when calling
 * <code>ubidi_setPara</code> should approximate the algorithm used in
 * Microsoft Windows XP rather than strictly conform to the Unicode Bidi
 * algorithm.
 * <br>
 * The differences between the basic algorithm and the algorithm addressed
 * by this option are as follows:
 * <ul>
 *   <li>Within text at an even embedding level, the sequence "123AB"
 *   (where AB represent R or AL letters) is transformed to "123BA" by the
 *   Unicode algorithm and to "BA123" by the Windows algorithm.</li>
 *   <li>Arabic-Indic numbers (AN) are handled by the Windows algorithm just
 *   like regular numbers (EN).</li>
 * </ul></li>
 *
 * <li>When the reordering mode is set to
 * <code>#UBIDI_REORDER_GROUP_NUMBERS_WITH_R</code>,
 * numbers located between LTR text and RTL text are associated with the RTL
 * text. For instance, an LTR paragraph with content "abc 123 DEF" (where
 * upper case letters represent RTL characters) will be transformed to
 * "abc FED 123" (and not "abc 123 FED"), "DEF 123 abc" will be transformed
 * to "123 FED abc" and "123 FED abc" will be transformed to "DEF 123 abc".
 * This makes the algorithm reversible and makes it useful when round trip
 * (from visual to logical and back to visual) must be achieved without
 * adding LRM characters. However, this is a variation from the standard
 * Unicode Bidi algorithm.<br>
 * The source text should not contain Bidi control characters other than LRM
 * or RLM.</li>
 *
 * <li>When the reordering mode is set to
 * <code>#UBIDI_REORDER_RUNS_ONLY</code>,
 * a "Logical to Logical" transformation must be performed:
 * <ul>
 * <li>If the default text level of the source text (argument <code>paraLevel</code>
 * in <code>ubidi_setPara</code>) is even, the source text will be handled as
 * LTR logical text and will be transformed to the RTL logical text which has
 * the same LTR visual display.</li>
 * <li>If the default level of the source text is odd, the source text
 * will be handled as RTL logical text and will be transformed to the
 * LTR logical text which has the same LTR visual display.</li>
 * </ul>
 * This mode may be needed when logical text which is basically Arabic or
 * Hebrew, with possible included numbers or phrases in English, has to be
 * displayed as if it had an even embedding level (this can happen if the
 * displaying application treats all text as if it was basically LTR).
 * <br>
 * This mode may also be needed in the reverse case, when logical text which is
 * basically English, with possible included phrases in Arabic or Hebrew, has to
 * be displayed as if it had an odd embedding level.
 * <br>
 * Both cases could be handled by adding LRE or RLE at the head of the text,
 * if the display subsystem supports these formatting controls. If it does not,
 * the problem may be handled by transforming the source text in this mode
 * before displaying it, so that it will be displayed properly.<br>
 * The source text should not contain Bidi control characters other than LRM
 * or RLM.</li>
 *
 * <li>When the reordering mode is set to
 * <code>#UBIDI_REORDER_INVERSE_NUMBERS_AS_L</code>, an "inverse Bidi" algorithm
 * is applied.
 * Runs of text with numeric characters will be treated like LTR letters and
 * may need to be surrounded with LRM characters when they are written in
 * reordered sequence (the option <code>#UBIDI_INSERT_LRM_FOR_NUMERIC</code> can
 * be used with function <code>ubidi_writeReordered</code> to this end. This
 * mode is equivalent to calling <code>ubidi_setInverse()</code> with
 * argument <code>isInverse</code> set to <code>TRUE</code>.</li>
 *
 * <li>When the reordering mode is set to
 * <code>#UBIDI_REORDER_INVERSE_LIKE_DIRECT</code>, the "direct" Logical to Visual
 * Bidi algorithm is used as an approximation of an "inverse Bidi" algorithm.
 * This mode is similar to mode <code>#UBIDI_REORDER_INVERSE_NUMBERS_AS_L</code>
 * but is closer to the regular Bidi algorithm.
 * <br>
 * For example, an LTR paragraph with the content "FED 123 456 CBA" (where
 * upper case represents RTL characters) will be transformed to
 * "ABC 456 123 DEF", as opposed to "DEF 123 456 ABC"
 * with mode <code>UBIDI_REORDER_INVERSE_NUMBERS_AS_L</code>.<br>
 * When used in conjunction with option
 * <code>#UBIDI_OPTION_INSERT_MARKS</code>, this mode generally
 * adds Bidi marks to the output significantly more sparingly than mode
 * <code>#UBIDI_REORDER_INVERSE_NUMBERS_AS_L</code> with option
 * <code>#UBIDI_INSERT_LRM_FOR_NUMERIC</code> in calls to
 * <code>ubidi_writeReordered</code>.</li>
 *
 * <li>When the reordering mode is set to
 * <code>#UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL</code>, the Logical to Visual
 * Bidi algorithm used in Windows XP is used as an approximation of an "inverse Bidi" algorithm.
 * <br>
 * For example, an LTR paragraph with the content "abc FED123" (where
 * upper case represents RTL characters) will be transformed to "abc 123DEF."</li>
 * </ul>
 *
 * <p>In all the reordering modes specifying an "inverse Bidi" algorithm
 * (i.e. those with a name starting with <code>UBIDI_REORDER_INVERSE</code>),
 * output runs should be retrieved using
 * <code>ubidi_getVisualRun()</code>, and the output text with
 * <code>ubidi_writeReordered()</code>. The caller should keep in mind that in
 * "inverse Bidi" modes the input is actually visually ordered text and
 * reordered output returned by <code>ubidi_getVisualRun()</code> or
 * <code>ubidi_writeReordered()</code> are actually runs or character string
 * of logically ordered output.<br>
 * For all the "inverse Bidi" modes, the source text should not contain
 * Bidi control characters other than LRM or RLM.</p>
 *
 * <p>Note that option <code>#UBIDI_OUTPUT_REVERSE</code> of
 * <code>ubidi_writeReordered</code> has no useful meaning and should not be
 * used in conjunction with any value of the reordering mode specifying
 * "inverse Bidi" or with value <code>UBIDI_REORDER_RUNS_ONLY</code>.
 *
 * @param pBiDi is a <code>UBiDi</code> object.
 * @param reorderingMode specifies the required variant of the Bidi algorithm.
 *
 * @see UBiDiReorderingMode
 * @see ubidi_setInverse
 * @see ubidi_setPara
 * @see ubidi_writeReordered
 * @stable ICU 3.6
 */
U_STABLE void U_EXPORT2
ubidi_setReorderingMode(UBiDi *pBiDi, UBiDiReorderingMode reorderingMode);

/**
 * What is the requested reordering mode for a given Bidi object?
 *
 * @param pBiDi is a <code>UBiDi</code> object.
 * @return the current reordering mode of the Bidi object
 * @see ubidi_setReorderingMode
 * @stable ICU 3.6
 */
U_STABLE UBiDiReorderingMode U_EXPORT2
ubidi_getReorderingMode(UBiDi *pBiDi);

/**
 * <code>UBiDiReorderingOption</code> values indicate which options are
 * specified to affect the Bidi algorithm.
 *
 * @see ubidi_setReorderingOptions
 * @stable ICU 3.6
 */
typedef enum UBiDiReorderingOption {
    /**
     * option value for <code>ubidi_setReorderingOptions</code>:
     * disable all the options which can be set with this function
     * @see ubidi_setReorderingOptions
     * @stable ICU 3.6
     */
    UBIDI_OPTION_DEFAULT = 0,

    /**
     * option bit for <code>ubidi_setReorderingOptions</code>:
     * insert Bidi marks (LRM or RLM) when needed to ensure correct result of
     * a reordering to a Logical order
     *
     * <p>This option must be set or reset before calling
     * <code>ubidi_setPara</code>.</p>
     *
     * <p>This option is significant only with reordering modes which generate
     * a result with Logical order, specifically:</p>
     * <ul>
     *   <li><code>#UBIDI_REORDER_RUNS_ONLY</code></li>
     *   <li><code>#UBIDI_REORDER_INVERSE_NUMBERS_AS_L</code></li>
     *   <li><code>#UBIDI_REORDER_INVERSE_LIKE_DIRECT</code></li>
     *   <li><code>#UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL</code></li>
     * </ul>
     *
     * <p>If this option is set in conjunction with reordering mode
     * <code>#UBIDI_REORDER_INVERSE_NUMBERS_AS_L</code> or with calling
     * <code>ubidi_setInverse(TRUE)</code>, it implies
     * option <code>#UBIDI_INSERT_LRM_FOR_NUMERIC</code>
     * in calls to function <code>ubidi_writeReordered()</code>.</p>
     *
     * <p>For other reordering modes, a minimum number of LRM or RLM characters
     * will be added to the source text after reordering it so as to ensure
     * round trip, i.e. when applying the inverse reordering mode on the
     * resulting logical text with removal of Bidi marks
     * (option <code>#UBIDI_OPTION_REMOVE_CONTROLS</code> set before calling
     * <code>ubidi_setPara()</code> or option <code>#UBIDI_REMOVE_BIDI_CONTROLS</code>
     * in <code>ubidi_writeReordered</code>), the result will be identical to the
     * source text in the first transformation.
     *
     * <p>This option will be ignored if specified together with option
     * <code>#UBIDI_OPTION_REMOVE_CONTROLS</code>. It inhibits option
     * <code>UBIDI_REMOVE_BIDI_CONTROLS</code> in calls to function
     * <code>ubidi_writeReordered()</code> and it implies option
     * <code>#UBIDI_INSERT_LRM_FOR_NUMERIC</code> in calls to function
     * <code>ubidi_writeReordered()</code> if the reordering mode is
     * <code>#UBIDI_REORDER_INVERSE_NUMBERS_AS_L</code>.</p>
     *
     * @see ubidi_setReorderingMode
     * @see ubidi_setReorderingOptions
     * @stable ICU 3.6
     */
    UBIDI_OPTION_INSERT_MARKS = 1,

    /**
     * option bit for <code>ubidi_setReorderingOptions</code>:
     * remove Bidi control characters
     *
     * <p>This option must be set or reset before calling
     * <code>ubidi_setPara</code>.</p>
     *
     * <p>This option nullifies option <code>#UBIDI_OPTION_INSERT_MARKS</code>.
     * It inhibits option <code>#UBIDI_INSERT_LRM_FOR_NUMERIC</code> in calls
     * to function <code>ubidi_writeReordered()</code> and it implies option
     * <code>#UBIDI_REMOVE_BIDI_CONTROLS</code> in calls to that function.</p>
     *
     * @see ubidi_setReorderingMode
     * @see ubidi_setReorderingOptions
     * @stable ICU 3.6
     */
    UBIDI_OPTION_REMOVE_CONTROLS = 2,

    /**
     * option bit for <code>ubidi_setReorderingOptions</code>:
     * process the output as part of a stream to be continued
     *
     * <p>This option must be set or reset before calling
     * <code>ubidi_setPara</code>.</p>
     *
     * <p>This option specifies that the caller is interested in processing large
     * text object in parts.
     * The results of the successive calls are expected to be concatenated by the
     * caller. Only the call for the last part will have this option bit off.</p>
     *
     * <p>When this option bit is on, <code>ubidi_setPara()</code> may process
     * less than the full source text in order to truncate the text at a meaningful
     * boundary. The caller should call <code>ubidi_getProcessedLength()</code>
     * immediately after calling <code>ubidi_setPara()</code> in order to
     * determine how much of the source text has been processed.
     * Source text beyond that length should be resubmitted in following calls to
     * <code>ubidi_setPara</code>. The processed length may be less than
     * the length of the source text if a character preceding the last character of
     * the source text constitutes a reasonable boundary (like a block separator)
     * for text to be continued.<br>
     * If the last character of the source text constitutes a reasonable
     * boundary, the whole text will be processed at once.<br>
     * If nowhere in the source text there exists
     * such a reasonable boundary, the processed length will be zero.<br>
     * The caller should check for such an occurrence and do one of the following:
     * <ul><li>submit a larger amount of text with a better chance to include
     *         a reasonable boundary.</li>
     *     <li>resubmit the same text after turning off option
     *         <code>UBIDI_OPTION_STREAMING</code>.</li></ul>
     * In all cases, this option should be turned off before processing the last
     * part of the text.</p>
     *
     * <p>When the <code>UBIDI_OPTION_STREAMING</code> option is used,
     * it is recommended to call <code>ubidi_orderParagraphsLTR()</code> with
     * argument <code>orderParagraphsLTR</code> set to <code>TRUE</code> before
     * calling <code>ubidi_setPara</code> so that later paragraphs may be
     * concatenated to previous paragraphs on the right.</p>
     *
     * @see ubidi_setReorderingMode
     * @see ubidi_setReorderingOptions
     * @see ubidi_getProcessedLength
     * @see ubidi_orderParagraphsLTR
     * @stable ICU 3.6
     */
    UBIDI_OPTION_STREAMING = 4
} UBiDiReorderingOption;

/**
 * Specify which of the reordering options
 * should be applied during Bidi transformations.
 *
 * @param pBiDi is a <code>UBiDi</code> object.
 * @param reorderingOptions is a combination of zero or more of the following
 * options:
 * <code>#UBIDI_OPTION_DEFAULT</code>, <code>#UBIDI_OPTION_INSERT_MARKS</code>,
 * <code>#UBIDI_OPTION_REMOVE_CONTROLS</code>, <code>#UBIDI_OPTION_STREAMING</code>.
 *
 * @see ubidi_getReorderingOptions
 * @stable ICU 3.6
 */
U_STABLE void U_EXPORT2
ubidi_setReorderingOptions(UBiDi *pBiDi, uint32_t reorderingOptions);

/**
 * What are the reordering options applied to a given Bidi object?
 *
 * @param pBiDi is a <code>UBiDi</code> object.
 * @return the current reordering options of the Bidi object
 * @see ubidi_setReorderingOptions
 * @stable ICU 3.6
 */
U_STABLE uint32_t U_EXPORT2
ubidi_getReorderingOptions(UBiDi *pBiDi);

/**
 * Set the context before a call to ubidi_setPara().<p>
 *
 * ubidi_setPara() computes the left-right directionality for a given piece
 * of text which is supplied as one of its arguments. Sometimes this piece
 * of text (the "main text") should be considered in context, because text
 * appearing before ("prologue") and/or after ("epilogue") the main text
 * may affect the result of this computation.<p>
 *
 * This function specifies the prologue and/or the epilogue for the next
 * call to ubidi_setPara(). The characters specified as prologue and
 * epilogue should not be modified by the calling program until the call
 * to ubidi_setPara() has returned. If successive calls to ubidi_setPara()
 * all need specification of a context, ubidi_setContext() must be called
 * before each call to ubidi_setPara(). In other words, a context is not
 * "remembered" after the following successful call to ubidi_setPara().<p>
 *
 * If a call to ubidi_setPara() specifies UBIDI_DEFAULT_LTR or
 * UBIDI_DEFAULT_RTL as paraLevel and is preceded by a call to
 * ubidi_setContext() which specifies a prologue, the paragraph level will
 * be computed taking in consideration the text in the prologue.<p>
 *
 * When ubidi_setPara() is called without a previous call to
 * ubidi_setContext, the main text is handled as if preceded and followed
 * by strong directional characters at the current paragraph level.
 * Calling ubidi_setContext() with specification of a prologue will change
 * this behavior by handling the main text as if preceded by the last
 * strong character appearing in the prologue, if any.
 * Calling ubidi_setContext() with specification of an epilogue will change
 * the behavior of ubidi_setPara() by handling the main text as if followed
 * by the first strong character or digit appearing in the epilogue, if any.<p>
 *
 * Note 1: if <code>ubidi_setContext</code> is called repeatedly without
 *         calling <code>ubidi_setPara</code>, the earlier calls have no effect,
 *         only the last call will be remembered for the next call to
 *         <code>ubidi_setPara</code>.<p>
 *
 * Note 2: calling <code>ubidi_setContext(pBiDi, NULL, 0, NULL, 0, &errorCode)</code>
 *         cancels any previous setting of non-empty prologue or epilogue.
 *         The next call to <code>ubidi_setPara()</code> will process no
 *         prologue or epilogue.<p>
 *
 * Note 3: users must be aware that even after setting the context
 *         before a call to ubidi_setPara() to perform e.g. a logical to visual
 *         transformation, the resulting string may not be identical to what it
 *         would have been if all the text, including prologue and epilogue, had
 *         been processed together.<br>
 * Example (upper case letters represent RTL characters):<br>
 * &nbsp;&nbsp;prologue = "<code>abc DE</code>"<br>
 * &nbsp;&nbsp;epilogue = none<br>
 * &nbsp;&nbsp;main text = "<code>FGH xyz</code>"<br>
 * &nbsp;&nbsp;paraLevel = UBIDI_LTR<br>
 * &nbsp;&nbsp;display without prologue = "<code>HGF xyz</code>"
 *             ("HGF" is adjacent to "xyz")<br>
 * &nbsp;&nbsp;display with prologue = "<code>abc HGFED xyz</code>"
 *             ("HGF" is not adjacent to "xyz")<br>
 *
 * @param pBiDi is a paragraph <code>UBiDi</code> object.
 *
 * @param prologue is a pointer to the text which precedes the text that
 *        will be specified in a coming call to ubidi_setPara().
 *        If there is no prologue to consider, then <code>proLength</code>
 *        must be zero and this pointer can be NULL.
 *
 * @param proLength is the length of the prologue; if <code>proLength==-1</code>
 *        then the prologue must be zero-terminated.
 *        Otherwise proLength must be >= 0. If <code>proLength==0</code>, it means
 *        that there is no prologue to consider.
 *
 * @param epilogue is a pointer to the text which follows the text that
 *        will be specified in a coming call to ubidi_setPara().
 *        If there is no epilogue to consider, then <code>epiLength</code>
 *        must be zero and this pointer can be NULL.
 *
 * @param epiLength is the length of the epilogue; if <code>epiLength==-1</code>
 *        then the epilogue must be zero-terminated.
 *        Otherwise epiLength must be >= 0. If <code>epiLength==0</code>, it means
 *        that there is no epilogue to consider.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @see ubidi_setPara
 * @stable ICU 4.8
 */
U_STABLE void U_EXPORT2
ubidi_setContext(UBiDi *pBiDi,
                 const UChar *prologue, int32_t proLength,
                 const UChar *epilogue, int32_t epiLength,
                 UErrorCode *pErrorCode);

/**
 * Perform the Unicode Bidi algorithm. It is defined in the
 * <a href="http://www.unicode.org/unicode/reports/tr9/">Unicode Standard Annex #9</a>,
 * version 13,
 * also described in The Unicode Standard, Version 4.0 .<p>
 *
 * This function takes a piece of plain text containing one or more paragraphs,
 * with or without externally specified embedding levels from <i>styled</i>
 * text and computes the left-right-directionality of each character.<p>
 *
 * If the entire text is all of the same directionality, then
 * the function may not perform all the steps described by the algorithm,
 * i.e., some levels may not be the same as if all steps were performed.
 * This is not relevant for unidirectional text.<br>
 * For example, in pure LTR text with numbers the numbers would get
 * a resolved level of 2 higher than the surrounding text according to
 * the algorithm. This implementation may set all resolved levels to
 * the same value in such a case.<p>
 *
 * The text can be composed of multiple paragraphs. Occurrence of a block
 * separator in the text terminates a paragraph, and whatever comes next starts
 * a new paragraph. The exception to this rule is when a Carriage Return (CR)
 * is followed by a Line Feed (LF). Both CR and LF are block separators, but
 * in that case, the pair of characters is considered as terminating the
 * preceding paragraph, and a new paragraph will be started by a character
 * coming after the LF.
 *
 * @param pBiDi A <code>UBiDi</code> object allocated with <code>ubidi_open()</code>
 *        which will be set to contain the reordering information,
 *        especially the resolved levels for all the characters in <code>text</code>.
 *
 * @param text is a pointer to the text that the Bidi algorithm will be performed on.
 *        This pointer is stored in the UBiDi object and can be retrieved
 *        with <code>ubidi_getText()</code>.<br>
 *        <strong>Note:</strong> the text must be (at least) <code>length</code> long.
 *
 * @param length is the length of the text; if <code>length==-1</code> then
 *        the text must be zero-terminated.
 *
 * @param paraLevel specifies the default level for the text;
 *        it is typically 0 (LTR) or 1 (RTL).
 *        If the function shall determine the paragraph level from the text,
 *        then <code>paraLevel</code> can be set to
 *        either <code>#UBIDI_DEFAULT_LTR</code>
 *        or <code>#UBIDI_DEFAULT_RTL</code>; if the text contains multiple
 *        paragraphs, the paragraph level shall be determined separately for
 *        each paragraph; if a paragraph does not include any strongly typed
 *        character, then the desired default is used (0 for LTR or 1 for RTL).
 *        Any other value between 0 and <code>#UBIDI_MAX_EXPLICIT_LEVEL</code>
 *        is also valid, with odd levels indicating RTL.
 *
 * @param embeddingLevels (in) may be used to preset the embedding and override levels,
 *        ignoring characters like LRE and PDF in the text.
 *        A level overrides the directional property of its corresponding
 *        (same index) character if the level has the
 *        <code>#UBIDI_LEVEL_OVERRIDE</code> bit set.<br><br>
 *        Aside from that bit, it must be
 *        <code>paraLevel<=embeddingLevels[]<=UBIDI_MAX_EXPLICIT_LEVEL</code>,
 *        except that level 0 is always allowed.
 *        Level 0 for a paragraph separator prevents reordering of paragraphs;
 *        this only works reliably if <code>#UBIDI_LEVEL_OVERRIDE</code>
 *        is also set for paragraph separators.
 *        Level 0 for other characters is treated as a wildcard
 *        and is lifted up to the resolved level of the surrounding paragraph.<br><br>
 *        <strong>Caution: </strong>A copy of this pointer, not of the levels,
 *        will be stored in the <code>UBiDi</code> object;
 *        the <code>embeddingLevels</code> array must not be
 *        deallocated before the <code>UBiDi</code> structure is destroyed or reused,
 *        and the <code>embeddingLevels</code>
 *        should not be modified to avoid unexpected results on subsequent Bidi operations.
 *        However, the <code>ubidi_setPara()</code> and
 *        <code>ubidi_setLine()</code> functions may modify some or all of the levels.<br><br>
 *        After the <code>UBiDi</code> object is reused or destroyed, the caller
 *        must take care of the deallocation of the <code>embeddingLevels</code> array.<br><br>
 *        <strong>Note:</strong> the <code>embeddingLevels</code> array must be
 *        at least <code>length</code> long.
 *        This pointer can be <code>NULL</code> if this
 *        value is not necessary.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
ubidi_setPara(UBiDi *pBiDi, const UChar *text, int32_t length,
              UBiDiLevel paraLevel, UBiDiLevel *embeddingLevels,
              UErrorCode *pErrorCode);

/**
 * <code>ubidi_setLine()</code> sets a <code>UBiDi</code> to
 * contain the reordering information, especially the resolved levels,
 * for all the characters in a line of text. This line of text is
 * specified by referring to a <code>UBiDi</code> object representing
 * this information for a piece of text containing one or more paragraphs,
 * and by specifying a range of indexes in this text.<p>
 * In the new line object, the indexes will range from 0 to <code>limit-start-1</code>.<p>
 *
 * This is used after calling <code>ubidi_setPara()</code>
 * for a piece of text, and after line-breaking on that text.
 * It is not necessary if each paragraph is treated as a single line.<p>
 *
 * After line-breaking, rules (L1) and (L2) for the treatment of
 * trailing WS and for reordering are performed on
 * a <code>UBiDi</code> object that represents a line.<p>
 *
 * <strong>Important: </strong><code>pLineBiDi</code> shares data with
 * <code>pParaBiDi</code>.
 * You must destroy or reuse <code>pLineBiDi</code> before <code>pParaBiDi</code>.
 * In other words, you must destroy or reuse the <code>UBiDi</code> object for a line
 * before the object for its parent paragraph.<p>
 *
 * The text pointer that was stored in <code>pParaBiDi</code> is also copied,
 * and <code>start</code> is added to it so that it points to the beginning of the
 * line for this object.
 *
 * @param pParaBiDi is the parent paragraph object. It must have been set
 * by a successful call to ubidi_setPara.
 *
 * @param start is the line's first index into the text.
 *
 * @param limit is just behind the line's last index into the text
 *        (its last index +1).<br>
 *        It must be <code>0<=start<limit<=</code>containing paragraph limit.
 *        If the specified line crosses a paragraph boundary, the function
 *        will terminate with error code U_ILLEGAL_ARGUMENT_ERROR.
 *
 * @param pLineBiDi is the object that will now represent a line of the text.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @see ubidi_setPara
 * @see ubidi_getProcessedLength
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
ubidi_setLine(const UBiDi *pParaBiDi,
              int32_t start, int32_t limit,
              UBiDi *pLineBiDi,
              UErrorCode *pErrorCode);

/**
 * Get the directionality of the text.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @return a value of <code>UBIDI_LTR</code>, <code>UBIDI_RTL</code>
 *         or <code>UBIDI_MIXED</code>
 *         that indicates if the entire text
 *         represented by this object is unidirectional,
 *         and which direction, or if it is mixed-directional.
 * Note -  The value <code>UBIDI_NEUTRAL</code> is never returned from this method.
 *
 * @see UBiDiDirection
 * @stable ICU 2.0
 */
U_STABLE UBiDiDirection U_EXPORT2
ubidi_getDirection(const UBiDi *pBiDi);

/**
 * Gets the base direction of the text provided according
 * to the Unicode Bidirectional Algorithm. The base direction
 * is derived from the first character in the string with bidirectional
 * character type L, R, or AL. If the first such character has type L,
 * <code>UBIDI_LTR</code> is returned. If the first such character has
 * type R or AL, <code>UBIDI_RTL</code> is returned. If the string does
 * not contain any character of these types, then
 * <code>UBIDI_NEUTRAL</code> is returned.
 *
 * This is a lightweight function for use when only the base direction
 * is needed and no further bidi processing of the text is needed.
 *
 * @param text is a pointer to the text whose base
 *             direction is needed.
 * Note: the text must be (at least) @c length long.
 *
 * @param length is the length of the text;
 *               if <code>length==-1</code> then the text
 *               must be zero-terminated.
 *
 * @return  <code>UBIDI_LTR</code>, <code>UBIDI_RTL</code>,
 *          <code>UBIDI_NEUTRAL</code>
 *
 * @see UBiDiDirection
 * @stable ICU 4.6
 */
U_STABLE UBiDiDirection U_EXPORT2
ubidi_getBaseDirection(const UChar *text,  int32_t length );

/**
 * Get the pointer to the text.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @return The pointer to the text that the UBiDi object was created for.
 *
 * @see ubidi_setPara
 * @see ubidi_setLine
 * @stable ICU 2.0
 */
U_STABLE const UChar * U_EXPORT2
ubidi_getText(const UBiDi *pBiDi);

/**
 * Get the length of the text.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @return The length of the text that the UBiDi object was created for.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
ubidi_getLength(const UBiDi *pBiDi);

/**
 * Get the paragraph level of the text.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @return The paragraph level. If there are multiple paragraphs, their
 *         level may vary if the required paraLevel is UBIDI_DEFAULT_LTR or
 *         UBIDI_DEFAULT_RTL.  In that case, the level of the first paragraph
 *         is returned.
 *
 * @see UBiDiLevel
 * @see ubidi_getParagraph
 * @see ubidi_getParagraphByIndex
 * @stable ICU 2.0
 */
U_STABLE UBiDiLevel U_EXPORT2
ubidi_getParaLevel(const UBiDi *pBiDi);

/**
 * Get the number of paragraphs.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @return The number of paragraphs.
 * @stable ICU 3.4
 */
U_STABLE int32_t U_EXPORT2
ubidi_countParagraphs(UBiDi *pBiDi);

/**
 * Get a paragraph, given a position within the text.
 * This function returns information about a paragraph.<br>
 * Note: if the paragraph index is known, it is more efficient to
 * retrieve the paragraph information using ubidi_getParagraphByIndex().<p>
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @param charIndex is the index of a character within the text, in the
 *        range <code>[0..ubidi_getProcessedLength(pBiDi)-1]</code>.
 *
 * @param pParaStart will receive the index of the first character of the
 *        paragraph in the text.
 *        This pointer can be <code>NULL</code> if this
 *        value is not necessary.
 *
 * @param pParaLimit will receive the limit of the paragraph.
 *        The l-value that you point to here may be the
 *        same expression (variable) as the one for
 *        <code>charIndex</code>.
 *        This pointer can be <code>NULL</code> if this
 *        value is not necessary.
 *
 * @param pParaLevel will receive the level of the paragraph.
 *        This pointer can be <code>NULL</code> if this
 *        value is not necessary.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @return The index of the paragraph containing the specified position.
 *
 * @see ubidi_getProcessedLength
 * @stable ICU 3.4
 */
U_STABLE int32_t U_EXPORT2
ubidi_getParagraph(const UBiDi *pBiDi, int32_t charIndex, int32_t *pParaStart,
                   int32_t *pParaLimit, UBiDiLevel *pParaLevel,
                   UErrorCode *pErrorCode);

/**
 * Get a paragraph, given the index of this paragraph.
 *
 * This function returns information about a paragraph.<p>
 *
 * @param pBiDi is the paragraph <code>UBiDi</code> object.
 *
 * @param paraIndex is the number of the paragraph, in the
 *        range <code>[0..ubidi_countParagraphs(pBiDi)-1]</code>.
 *
 * @param pParaStart will receive the index of the first character of the
 *        paragraph in the text.
 *        This pointer can be <code>NULL</code> if this
 *        value is not necessary.
 *
 * @param pParaLimit will receive the limit of the paragraph.
 *        This pointer can be <code>NULL</code> if this
 *        value is not necessary.
 *
 * @param pParaLevel will receive the level of the paragraph.
 *        This pointer can be <code>NULL</code> if this
 *        value is not necessary.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
ubidi_getParagraphByIndex(const UBiDi *pBiDi, int32_t paraIndex,
                          int32_t *pParaStart, int32_t *pParaLimit,
                          UBiDiLevel *pParaLevel, UErrorCode *pErrorCode);

/**
 * Get the level for one character.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @param charIndex the index of a character. It must be in the range
 *         [0..ubidi_getProcessedLength(pBiDi)].
 *
 * @return The level for the character at charIndex (0 if charIndex is not
 *         in the valid range).
 *
 * @see UBiDiLevel
 * @see ubidi_getProcessedLength
 * @stable ICU 2.0
 */
U_STABLE UBiDiLevel U_EXPORT2
ubidi_getLevelAt(const UBiDi *pBiDi, int32_t charIndex);

/**
 * Get an array of levels for each character.<p>
 *
 * Note that this function may allocate memory under some
 * circumstances, unlike <code>ubidi_getLevelAt()</code>.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object, whose
 *        text length must be strictly positive.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @return The levels array for the text,
 *         or <code>NULL</code> if an error occurs.
 *
 * @see UBiDiLevel
 * @see ubidi_getProcessedLength
 * @stable ICU 2.0
 */
U_STABLE const UBiDiLevel * U_EXPORT2
ubidi_getLevels(UBiDi *pBiDi, UErrorCode *pErrorCode);

/**
 * Get a logical run.
 * This function returns information about a run and is used
 * to retrieve runs in logical order.<p>
 * This is especially useful for line-breaking on a paragraph.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @param logicalPosition is a logical position within the source text.
 *
 * @param pLogicalLimit will receive the limit of the corresponding run.
 *        The l-value that you point to here may be the
 *        same expression (variable) as the one for
 *        <code>logicalPosition</code>.
 *        This pointer can be <code>NULL</code> if this
 *        value is not necessary.
 *
 * @param pLevel will receive the level of the corresponding run.
 *        This pointer can be <code>NULL</code> if this
 *        value is not necessary.
 *
 * @see ubidi_getProcessedLength
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
ubidi_getLogicalRun(const UBiDi *pBiDi, int32_t logicalPosition,
                    int32_t *pLogicalLimit, UBiDiLevel *pLevel);

/**
 * Get the number of runs.
 * This function may invoke the actual reordering on the
 * <code>UBiDi</code> object, after <code>ubidi_setPara()</code>
 * may have resolved only the levels of the text. Therefore,
 * <code>ubidi_countRuns()</code> may have to allocate memory,
 * and may fail doing so.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @return The number of runs.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
ubidi_countRuns(UBiDi *pBiDi, UErrorCode *pErrorCode);

/**
 * Get one run's logical start, length, and directionality,
 * which can be 0 for LTR or 1 for RTL.
 * In an RTL run, the character at the logical start is
 * visually on the right of the displayed run.
 * The length is the number of characters in the run.<p>
 * <code>ubidi_countRuns()</code> should be called
 * before the runs are retrieved.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @param runIndex is the number of the run in visual order, in the
 *        range <code>[0..ubidi_countRuns(pBiDi)-1]</code>.
 *
 * @param pLogicalStart is the first logical character index in the text.
 *        The pointer may be <code>NULL</code> if this index is not needed.
 *
 * @param pLength is the number of characters (at least one) in the run.
 *        The pointer may be <code>NULL</code> if this is not needed.
 *
 * @return the directionality of the run,
 *         <code>UBIDI_LTR==0</code> or <code>UBIDI_RTL==1</code>,
 *         never <code>UBIDI_MIXED</code>,
 *         never <code>UBIDI_NEUTRAL</code>.
 *
 * @see ubidi_countRuns
 *
 * Example:
 * <pre>
 * \code
 * int32_t i, count=ubidi_countRuns(pBiDi),
 *         logicalStart, visualIndex=0, length;
 * for(i=0; i<count; ++i) {
 *    if(UBIDI_LTR==ubidi_getVisualRun(pBiDi, i, &logicalStart, &length)) {
 *         do { // LTR
 *             show_char(text[logicalStart++], visualIndex++);
 *         } while(--length>0);
 *     } else {
 *         logicalStart+=length;  // logicalLimit
 *         do { // RTL
 *             show_char(text[--logicalStart], visualIndex++);
 *         } while(--length>0);
 *     }
 * }
 *\endcode
 * </pre>
 *
 * Note that in right-to-left runs, code like this places
 * second surrogates before first ones (which is generally a bad idea)
 * and combining characters before base characters.
 * <p>
 * Use of <code>ubidi_writeReordered()</code>, optionally with the
 * <code>#UBIDI_KEEP_BASE_COMBINING</code> option, can be considered in order
 * to avoid these issues.
 * @stable ICU 2.0
 */
U_STABLE UBiDiDirection U_EXPORT2
ubidi_getVisualRun(UBiDi *pBiDi, int32_t runIndex,
                   int32_t *pLogicalStart, int32_t *pLength);

/**
 * Get the visual position from a logical text position.
 * If such a mapping is used many times on the same
 * <code>UBiDi</code> object, then calling
 * <code>ubidi_getLogicalMap()</code> is more efficient.<p>
 *
 * The value returned may be <code>#UBIDI_MAP_NOWHERE</code> if there is no
 * visual position because the corresponding text character is a Bidi control
 * removed from output by the option <code>#UBIDI_OPTION_REMOVE_CONTROLS</code>.
 * <p>
 * When the visual output is altered by using options of
 * <code>ubidi_writeReordered()</code> such as <code>UBIDI_INSERT_LRM_FOR_NUMERIC</code>,
 * <code>UBIDI_KEEP_BASE_COMBINING</code>, <code>UBIDI_OUTPUT_REVERSE</code>,
 * <code>UBIDI_REMOVE_BIDI_CONTROLS</code>, the visual position returned may not
 * be correct. It is advised to use, when possible, reordering options
 * such as <code>UBIDI_OPTION_INSERT_MARKS</code> and <code>UBIDI_OPTION_REMOVE_CONTROLS</code>.
 * <p>
 * Note that in right-to-left runs, this mapping places
 * second surrogates before first ones (which is generally a bad idea)
 * and combining characters before base characters.
 * Use of <code>ubidi_writeReordered()</code>, optionally with the
 * <code>#UBIDI_KEEP_BASE_COMBINING</code> option can be considered instead
 * of using the mapping, in order to avoid these issues.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @param logicalIndex is the index of a character in the text.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @return The visual position of this character.
 *
 * @see ubidi_getLogicalMap
 * @see ubidi_getLogicalIndex
 * @see ubidi_getProcessedLength
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
ubidi_getVisualIndex(UBiDi *pBiDi, int32_t logicalIndex, UErrorCode *pErrorCode);

/**
 * Get the logical text position from a visual position.
 * If such a mapping is used many times on the same
 * <code>UBiDi</code> object, then calling
 * <code>ubidi_getVisualMap()</code> is more efficient.<p>
 *
 * The value returned may be <code>#UBIDI_MAP_NOWHERE</code> if there is no
 * logical position because the corresponding text character is a Bidi mark
 * inserted in the output by option <code>#UBIDI_OPTION_INSERT_MARKS</code>.
 * <p>
 * This is the inverse function to <code>ubidi_getVisualIndex()</code>.
 * <p>
 * When the visual output is altered by using options of
 * <code>ubidi_writeReordered()</code> such as <code>UBIDI_INSERT_LRM_FOR_NUMERIC</code>,
 * <code>UBIDI_KEEP_BASE_COMBINING</code>, <code>UBIDI_OUTPUT_REVERSE</code>,
 * <code>UBIDI_REMOVE_BIDI_CONTROLS</code>, the logical position returned may not
 * be correct. It is advised to use, when possible, reordering options
 * such as <code>UBIDI_OPTION_INSERT_MARKS</code> and <code>UBIDI_OPTION_REMOVE_CONTROLS</code>.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @param visualIndex is the visual position of a character.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @return The index of this character in the text.
 *
 * @see ubidi_getVisualMap
 * @see ubidi_getVisualIndex
 * @see ubidi_getResultLength
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
ubidi_getLogicalIndex(UBiDi *pBiDi, int32_t visualIndex, UErrorCode *pErrorCode);

/**
 * Get a logical-to-visual index map (array) for the characters in the UBiDi
 * (paragraph or line) object.
 * <p>
 * Some values in the map may be <code>#UBIDI_MAP_NOWHERE</code> if the
 * corresponding text characters are Bidi controls removed from the visual
 * output by the option <code>#UBIDI_OPTION_REMOVE_CONTROLS</code>.
 * <p>
 * When the visual output is altered by using options of
 * <code>ubidi_writeReordered()</code> such as <code>UBIDI_INSERT_LRM_FOR_NUMERIC</code>,
 * <code>UBIDI_KEEP_BASE_COMBINING</code>, <code>UBIDI_OUTPUT_REVERSE</code>,
 * <code>UBIDI_REMOVE_BIDI_CONTROLS</code>, the visual positions returned may not
 * be correct. It is advised to use, when possible, reordering options
 * such as <code>UBIDI_OPTION_INSERT_MARKS</code> and <code>UBIDI_OPTION_REMOVE_CONTROLS</code>.
 * <p>
 * Note that in right-to-left runs, this mapping places
 * second surrogates before first ones (which is generally a bad idea)
 * and combining characters before base characters.
 * Use of <code>ubidi_writeReordered()</code>, optionally with the
 * <code>#UBIDI_KEEP_BASE_COMBINING</code> option can be considered instead
 * of using the mapping, in order to avoid these issues.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @param indexMap is a pointer to an array of <code>ubidi_getProcessedLength()</code>
 *        indexes which will reflect the reordering of the characters.
 *        If option <code>#UBIDI_OPTION_INSERT_MARKS</code> is set, the number
 *        of elements allocated in <code>indexMap</code> must be no less than
 *        <code>ubidi_getResultLength()</code>.
 *        The array does not need to be initialized.<br><br>
 *        The index map will result in <code>indexMap[logicalIndex]==visualIndex</code>.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @see ubidi_getVisualMap
 * @see ubidi_getVisualIndex
 * @see ubidi_getProcessedLength
 * @see ubidi_getResultLength
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
ubidi_getLogicalMap(UBiDi *pBiDi, int32_t *indexMap, UErrorCode *pErrorCode);

/**
 * Get a visual-to-logical index map (array) for the characters in the UBiDi
 * (paragraph or line) object.
 * <p>
 * Some values in the map may be <code>#UBIDI_MAP_NOWHERE</code> if the
 * corresponding text characters are Bidi marks inserted in the visual output
 * by the option <code>#UBIDI_OPTION_INSERT_MARKS</code>.
 * <p>
 * When the visual output is altered by using options of
 * <code>ubidi_writeReordered()</code> such as <code>UBIDI_INSERT_LRM_FOR_NUMERIC</code>,
 * <code>UBIDI_KEEP_BASE_COMBINING</code>, <code>UBIDI_OUTPUT_REVERSE</code>,
 * <code>UBIDI_REMOVE_BIDI_CONTROLS</code>, the logical positions returned may not
 * be correct. It is advised to use, when possible, reordering options
 * such as <code>UBIDI_OPTION_INSERT_MARKS</code> and <code>UBIDI_OPTION_REMOVE_CONTROLS</code>.
 *
 * @param pBiDi is the paragraph or line <code>UBiDi</code> object.
 *
 * @param indexMap is a pointer to an array of <code>ubidi_getResultLength()</code>
 *        indexes which will reflect the reordering of the characters.
 *        If option <code>#UBIDI_OPTION_REMOVE_CONTROLS</code> is set, the number
 *        of elements allocated in <code>indexMap</code> must be no less than
 *        <code>ubidi_getProcessedLength()</code>.
 *        The array does not need to be initialized.<br><br>
 *        The index map will result in <code>indexMap[visualIndex]==logicalIndex</code>.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @see ubidi_getLogicalMap
 * @see ubidi_getLogicalIndex
 * @see ubidi_getProcessedLength
 * @see ubidi_getResultLength
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
ubidi_getVisualMap(UBiDi *pBiDi, int32_t *indexMap, UErrorCode *pErrorCode);

/**
 * This is a convenience function that does not use a UBiDi object.
 * It is intended to be used for when an application has determined the levels
 * of objects (character sequences) and just needs to have them reordered (L2).
 * This is equivalent to using <code>ubidi_getLogicalMap()</code> on a
 * <code>UBiDi</code> object.
 *
 * @param levels is an array with <code>length</code> levels that have been determined by
 *        the application.
 *
 * @param length is the number of levels in the array, or, semantically,
 *        the number of objects to be reordered.
 *        It must be <code>length>0</code>.
 *
 * @param indexMap is a pointer to an array of <code>length</code>
 *        indexes which will reflect the reordering of the characters.
 *        The array does not need to be initialized.<p>
 *        The index map will result in <code>indexMap[logicalIndex]==visualIndex</code>.
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
ubidi_reorderLogical(const UBiDiLevel *levels, int32_t length, int32_t *indexMap);

/**
 * This is a convenience function that does not use a UBiDi object.
 * It is intended to be used for when an application has determined the levels
 * of objects (character sequences) and just needs to have them reordered (L2).
 * This is equivalent to using <code>ubidi_getVisualMap()</code> on a
 * <code>UBiDi</code> object.
 *
 * @param levels is an array with <code>length</code> levels that have been determined by
 *        the application.
 *
 * @param length is the number of levels in the array, or, semantically,
 *        the number of objects to be reordered.
 *        It must be <code>length>0</code>.
 *
 * @param indexMap is a pointer to an array of <code>length</code>
 *        indexes which will reflect the reordering of the characters.
 *        The array does not need to be initialized.<p>
 *        The index map will result in <code>indexMap[visualIndex]==logicalIndex</code>.
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
ubidi_reorderVisual(const UBiDiLevel *levels, int32_t length, int32_t *indexMap);

/**
 * Invert an index map.
 * The index mapping of the first map is inverted and written to
 * the second one.
 *
 * @param srcMap is an array with <code>length</code> elements
 *        which defines the original mapping from a source array containing
 *        <code>length</code> elements to a destination array.
 *        Some elements of the source array may have no mapping in the
 *        destination array. In that case, their value will be
 *        the special value <code>UBIDI_MAP_NOWHERE</code>.
 *        All elements must be >=0 or equal to <code>UBIDI_MAP_NOWHERE</code>.
 *        Some elements may have a value >= <code>length</code>, if the
 *        destination array has more elements than the source array.
 *        There must be no duplicate indexes (two or more elements with the
 *        same value except <code>UBIDI_MAP_NOWHERE</code>).
 *
 * @param destMap is an array with a number of elements equal to 1 + the highest
 *        value in <code>srcMap</code>.
 *        <code>destMap</code> will be filled with the inverse mapping.
 *        If element with index i in <code>srcMap</code> has a value k different
 *        from <code>UBIDI_MAP_NOWHERE</code>, this means that element i of
 *        the source array maps to element k in the destination array.
 *        The inverse map will have value i in its k-th element.
 *        For all elements of the destination array which do not map to
 *        an element in the source array, the corresponding element in the
 *        inverse map will have a value equal to <code>UBIDI_MAP_NOWHERE</code>.
 *
 * @param length is the length of each array.
 * @see UBIDI_MAP_NOWHERE
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
ubidi_invertMap(const int32_t *srcMap, int32_t *destMap, int32_t length);

/** option flags for ubidi_writeReordered() */

/**
 * option bit for ubidi_writeReordered():
 * keep combining characters after their base characters in RTL runs
 *
 * @see ubidi_writeReordered
 * @stable ICU 2.0
 */
#define UBIDI_KEEP_BASE_COMBINING       1

/**
 * option bit for ubidi_writeReordered():
 * replace characters with the "mirrored" property in RTL runs
 * by their mirror-image mappings
 *
 * @see ubidi_writeReordered
 * @stable ICU 2.0
 */
#define UBIDI_DO_MIRRORING              2

/**
 * option bit for ubidi_writeReordered():
 * surround the run with LRMs if necessary;
 * this is part of the approximate "inverse Bidi" algorithm
 *
 * <p>This option does not imply corresponding adjustment of the index
 * mappings.</p>
 *
 * @see ubidi_setInverse
 * @see ubidi_writeReordered
 * @stable ICU 2.0
 */
#define UBIDI_INSERT_LRM_FOR_NUMERIC    4

/**
 * option bit for ubidi_writeReordered():
 * remove Bidi control characters
 * (this does not affect #UBIDI_INSERT_LRM_FOR_NUMERIC)
 *
 * <p>This option does not imply corresponding adjustment of the index
 * mappings.</p>
 *
 * @see ubidi_writeReordered
 * @stable ICU 2.0
 */
#define UBIDI_REMOVE_BIDI_CONTROLS      8

/**
 * option bit for ubidi_writeReordered():
 * write the output in reverse order
 *
 * <p>This has the same effect as calling <code>ubidi_writeReordered()</code>
 * first without this option, and then calling
 * <code>ubidi_writeReverse()</code> without mirroring.
 * Doing this in the same step is faster and avoids a temporary buffer.
 * An example for using this option is output to a character terminal that
 * is designed for RTL scripts and stores text in reverse order.</p>
 *
 * @see ubidi_writeReordered
 * @stable ICU 2.0
 */
#define UBIDI_OUTPUT_REVERSE            16

/**
 * Get the length of the source text processed by the last call to
 * <code>ubidi_setPara()</code>. This length may be different from the length
 * of the source text if option <code>#UBIDI_OPTION_STREAMING</code>
 * has been set.
 * <br>
 * Note that whenever the length of the text affects the execution or the
 * result of a function, it is the processed length which must be considered,
 * except for <code>ubidi_setPara</code> (which receives unprocessed source
 * text) and <code>ubidi_getLength</code> (which returns the original length
 * of the source text).<br>
 * In particular, the processed length is the one to consider in the following
 * cases:
 * <ul>
 * <li>maximum value of the <code>limit</code> argument of
 * <code>ubidi_setLine</code></li>
 * <li>maximum value of the <code>charIndex</code> argument of
 * <code>ubidi_getParagraph</code></li>
 * <li>maximum value of the <code>charIndex</code> argument of
 * <code>ubidi_getLevelAt</code></li>
 * <li>number of elements in the array returned by <code>ubidi_getLevels</code></li>
 * <li>maximum value of the <code>logicalStart</code> argument of
 * <code>ubidi_getLogicalRun</code></li>
 * <li>maximum value of the <code>logicalIndex</code> argument of
 * <code>ubidi_getVisualIndex</code></li>
 * <li>number of elements filled in the <code>*indexMap</code> argument of
 * <code>ubidi_getLogicalMap</code></li>
 * <li>length of text processed by <code>ubidi_writeReordered</code></li>
 * </ul>
 *
 * @param pBiDi is the paragraph <code>UBiDi</code> object.
 *
 * @return The length of the part of the source text processed by
 *         the last call to <code>ubidi_setPara</code>.
 * @see ubidi_setPara
 * @see UBIDI_OPTION_STREAMING
 * @stable ICU 3.6
 */
U_STABLE int32_t U_EXPORT2
ubidi_getProcessedLength(const UBiDi *pBiDi);

/**
 * Get the length of the reordered text resulting from the last call to
 * <code>ubidi_setPara()</code>. This length may be different from the length
 * of the source text if option <code>#UBIDI_OPTION_INSERT_MARKS</code>
 * or option <code>#UBIDI_OPTION_REMOVE_CONTROLS</code> has been set.
 * <br>
 * This resulting length is the one to consider in the following cases:
 * <ul>
 * <li>maximum value of the <code>visualIndex</code> argument of
 * <code>ubidi_getLogicalIndex</code></li>
 * <li>number of elements of the <code>*indexMap</code> argument of
 * <code>ubidi_getVisualMap</code></li>
 * </ul>
 * Note that this length stays identical to the source text length if
 * Bidi marks are inserted or removed using option bits of
 * <code>ubidi_writeReordered</code>, or if option
 * <code>#UBIDI_REORDER_INVERSE_NUMBERS_AS_L</code> has been set.
 *
 * @param pBiDi is the paragraph <code>UBiDi</code> object.
 *
 * @return The length of the reordered text resulting from
 *         the last call to <code>ubidi_setPara</code>.
 * @see ubidi_setPara
 * @see UBIDI_OPTION_INSERT_MARKS
 * @see UBIDI_OPTION_REMOVE_CONTROLS
 * @stable ICU 3.6
 */
U_STABLE int32_t U_EXPORT2
ubidi_getResultLength(const UBiDi *pBiDi);

U_CDECL_BEGIN

#ifndef U_HIDE_DEPRECATED_API
/**
 * Value returned by <code>UBiDiClassCallback</code> callbacks when
 * there is no need to override the standard Bidi class for a given code point.
 *
 * This constant is deprecated; use u_getIntPropertyMaxValue(UCHAR_BIDI_CLASS)+1 instead.
 *
 * @see UBiDiClassCallback
 * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
 */
#define U_BIDI_CLASS_DEFAULT  U_CHAR_DIRECTION_COUNT
#endif  // U_HIDE_DEPRECATED_API

/**
 * Callback type declaration for overriding default Bidi class values with
 * custom ones.
 * <p>Usually, the function pointer will be propagated to a <code>UBiDi</code>
 * object by calling the <code>ubidi_setClassCallback()</code> function;
 * then the callback will be invoked by the UBA implementation any time the
 * class of a character is to be determined.</p>
 *
 * @param context is a pointer to the callback private data.
 *
 * @param c       is the code point to get a Bidi class for.
 *
 * @return The directional property / Bidi class for the given code point
 *         <code>c</code> if the default class has been overridden, or
 *         <code>u_getIntPropertyMaxValue(UCHAR_BIDI_CLASS)+1</code>
 *         if the standard Bidi class value for <code>c</code> is to be used.
 * @see ubidi_setClassCallback
 * @see ubidi_getClassCallback
 * @stable ICU 3.6
 */
typedef UCharDirection U_CALLCONV
UBiDiClassCallback(const void *context, UChar32 c);

U_CDECL_END

/**
 * Retrieve the Bidi class for a given code point.
 * <p>If a <code>#UBiDiClassCallback</code> callback is defined and returns a
 * value other than <code>u_getIntPropertyMaxValue(UCHAR_BIDI_CLASS)+1</code>,
 * that value is used; otherwise the default class determination mechanism is invoked.</p>
 *
 * @param pBiDi is the paragraph <code>UBiDi</code> object.
 *
 * @param c     is the code point whose Bidi class must be retrieved.
 *
 * @return The Bidi class for character <code>c</code> based
 *         on the given <code>pBiDi</code> instance.
 * @see UBiDiClassCallback
 * @stable ICU 3.6
 */
U_STABLE UCharDirection U_EXPORT2
ubidi_getCustomizedClass(UBiDi *pBiDi, UChar32 c);

/**
 * Set the callback function and callback data used by the UBA
 * implementation for Bidi class determination.
 * <p>This may be useful for assigning Bidi classes to PUA characters, or
 * for special application needs. For instance, an application may want to
 * handle all spaces like L or R characters (according to the base direction)
 * when creating the visual ordering of logical lines which are part of a report
 * organized in columns: there should not be interaction between adjacent
 * cells.<p>
 *
 * @param pBiDi is the paragraph <code>UBiDi</code> object.
 *
 * @param newFn is the new callback function pointer.
 *
 * @param newContext is the new callback context pointer. This can be NULL.
 *
 * @param oldFn fillin: Returns the old callback function pointer. This can be
 *                      NULL.
 *
 * @param oldContext fillin: Returns the old callback's context. This can be
 *                           NULL.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @see ubidi_getClassCallback
 * @stable ICU 3.6
 */
U_STABLE void U_EXPORT2
ubidi_setClassCallback(UBiDi *pBiDi, UBiDiClassCallback *newFn,
                       const void *newContext, UBiDiClassCallback **oldFn,
                       const void **oldContext, UErrorCode *pErrorCode);

/**
 * Get the current callback function used for Bidi class determination.
 *
 * @param pBiDi is the paragraph <code>UBiDi</code> object.
 *
 * @param fn fillin: Returns the callback function pointer.
 *
 * @param context fillin: Returns the callback's private context.
 *
 * @see ubidi_setClassCallback
 * @stable ICU 3.6
 */
U_STABLE void U_EXPORT2
ubidi_getClassCallback(UBiDi *pBiDi, UBiDiClassCallback **fn, const void **context);

/**
 * Take a <code>UBiDi</code> object containing the reordering
 * information for a piece of text (one or more paragraphs) set by
 * <code>ubidi_setPara()</code> or for a line of text set by
 * <code>ubidi_setLine()</code> and write a reordered string to the
 * destination buffer.
 *
 * This function preserves the integrity of characters with multiple
 * code units and (optionally) combining characters.
 * Characters in RTL runs can be replaced by mirror-image characters
 * in the destination buffer. Note that "real" mirroring has
 * to be done in a rendering engine by glyph selection
 * and that for many "mirrored" characters there are no
 * Unicode characters as mirror-image equivalents.
 * There are also options to insert or remove Bidi control
 * characters; see the description of the <code>destSize</code>
 * and <code>options</code> parameters and of the option bit flags.
 *
 * @param pBiDi A pointer to a <code>UBiDi</code> object that
 *              is set by <code>ubidi_setPara()</code> or
 *              <code>ubidi_setLine()</code> and contains the reordering
 *              information for the text that it was defined for,
 *              as well as a pointer to that text.<br><br>
 *              The text was aliased (only the pointer was stored
 *              without copying the contents) and must not have been modified
 *              since the <code>ubidi_setPara()</code> call.
 *
 * @param dest A pointer to where the reordered text is to be copied.
 *             The source text and <code>dest[destSize]</code>
 *             must not overlap.
 *
 * @param destSize The size of the <code>dest</code> buffer,
 *                 in number of UChars.
 *                 If the <code>UBIDI_INSERT_LRM_FOR_NUMERIC</code>
 *                 option is set, then the destination length could be
 *                 as large as
 *                 <code>ubidi_getLength(pBiDi)+2*ubidi_countRuns(pBiDi)</code>.
 *                 If the <code>UBIDI_REMOVE_BIDI_CONTROLS</code> option
 *                 is set, then the destination length may be less than
 *                 <code>ubidi_getLength(pBiDi)</code>.
 *                 If none of these options is set, then the destination length
 *                 will be exactly <code>ubidi_getProcessedLength(pBiDi)</code>.
 *
 * @param options A bit set of options for the reordering that control
 *                how the reordered text is written.
 *                The options include mirroring the characters on a code
 *                point basis and inserting LRM characters, which is used
 *                especially for transforming visually stored text
 *                to logically stored text (although this is still an
 *                imperfect implementation of an "inverse Bidi" algorithm
 *                because it uses the "forward Bidi" algorithm at its core).
 *                The available options are:
 *                <code>#UBIDI_DO_MIRRORING</code>,
 *                <code>#UBIDI_INSERT_LRM_FOR_NUMERIC</code>,
 *                <code>#UBIDI_KEEP_BASE_COMBINING</code>,
 *                <code>#UBIDI_OUTPUT_REVERSE</code>,
 *                <code>#UBIDI_REMOVE_BIDI_CONTROLS</code>
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @return The length of the output string.
 *
 * @see ubidi_getProcessedLength
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
ubidi_writeReordered(UBiDi *pBiDi,
                     UChar *dest, int32_t destSize,
                     uint16_t options,
                     UErrorCode *pErrorCode);

/**
 * Reverse a Right-To-Left run of Unicode text.
 *
 * This function preserves the integrity of characters with multiple
 * code units and (optionally) combining characters.
 * Characters can be replaced by mirror-image characters
 * in the destination buffer. Note that "real" mirroring has
 * to be done in a rendering engine by glyph selection
 * and that for many "mirrored" characters there are no
 * Unicode characters as mirror-image equivalents.
 * There are also options to insert or remove Bidi control
 * characters.
 *
 * This function is the implementation for reversing RTL runs as part
 * of <code>ubidi_writeReordered()</code>. For detailed descriptions
 * of the parameters, see there.
 * Since no Bidi controls are inserted here, the output string length
 * will never exceed <code>srcLength</code>.
 *
 * @see ubidi_writeReordered
 *
 * @param src A pointer to the RTL run text.
 *
 * @param srcLength The length of the RTL run.
 *
 * @param dest A pointer to where the reordered text is to be copied.
 *             <code>src[srcLength]</code> and <code>dest[destSize]</code>
 *             must not overlap.
 *
 * @param destSize The size of the <code>dest</code> buffer,
 *                 in number of UChars.
 *                 If the <code>UBIDI_REMOVE_BIDI_CONTROLS</code> option
 *                 is set, then the destination length may be less than
 *                 <code>srcLength</code>.
 *                 If this option is not set, then the destination length
 *                 will be exactly <code>srcLength</code>.
 *
 * @param options A bit set of options for the reordering that control
 *                how the reordered text is written.
 *                See the <code>options</code> parameter in <code>ubidi_writeReordered()</code>.
 *
 * @param pErrorCode must be a valid pointer to an error code value.
 *
 * @return The length of the output string.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
ubidi_writeReverse(const UChar *src, int32_t srcLength,
                   UChar *dest, int32_t destSize,
                   uint16_t options,
                   UErrorCode *pErrorCode);

/*#define BIDI_SAMPLE_CODE*/
/*@}*/




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  ubidiimp.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999aug06
*   created by: Markus W. Scherer, updated by Matitiahu Allouche
*/







/* miscellaneous definitions ---------------------------------------------- */

typedef uint8_t DirProp;
typedef uint32_t Flags;

/*  Comparing the description of the BiDi algorithm with this implementation
    is easier with the same names for the BiDi types in the code as there.
    See UCharDirection in uchar.h .
*/
enum {
    L=  U_LEFT_TO_RIGHT,                /*  0 */
    R=  U_RIGHT_TO_LEFT,                /*  1 */
    EN= U_EUROPEAN_NUMBER,              /*  2 */
    ES= U_EUROPEAN_NUMBER_SEPARATOR,    /*  3 */
    ET= U_EUROPEAN_NUMBER_TERMINATOR,   /*  4 */
    AN= U_ARABIC_NUMBER,                /*  5 */
    CS= U_COMMON_NUMBER_SEPARATOR,      /*  6 */
    B=  U_BLOCK_SEPARATOR,              /*  7 */
    S=  U_SEGMENT_SEPARATOR,            /*  8 */
    WS= U_WHITE_SPACE_NEUTRAL,          /*  9 */
    ON= U_OTHER_NEUTRAL,                /* 10 */
    LRE=U_LEFT_TO_RIGHT_EMBEDDING,      /* 11 */
    LRO=U_LEFT_TO_RIGHT_OVERRIDE,       /* 12 */
    AL= U_RIGHT_TO_LEFT_ARABIC,         /* 13 */
    RLE=U_RIGHT_TO_LEFT_EMBEDDING,      /* 14 */
    RLO=U_RIGHT_TO_LEFT_OVERRIDE,       /* 15 */
    PDF=U_POP_DIRECTIONAL_FORMAT,       /* 16 */
    NSM=U_DIR_NON_SPACING_MARK,         /* 17 */
    BN= U_BOUNDARY_NEUTRAL,             /* 18 */
    FSI=U_FIRST_STRONG_ISOLATE,         /* 19 */
    LRI=U_LEFT_TO_RIGHT_ISOLATE,        /* 20 */
    RLI=U_RIGHT_TO_LEFT_ISOLATE,        /* 21 */
    PDI=U_POP_DIRECTIONAL_ISOLATE,      /* 22 */
    ENL,    /* EN after W7 */           /* 23 */
    ENR,    /* EN not subject to W7 */  /* 24 */
    dirPropCount
};

/*  Sometimes, bit values are more appropriate
    to deal with directionality properties.
    Abbreviations in these macro names refer to names
    used in the BiDi algorithm.
*/
#define DIRPROP_FLAG(dir) (1UL<<(dir))
#define PURE_DIRPROP(prop)  ((prop)&~0xE0)    ?????????????????????????

/* special flag for multiple runs from explicit embedding codes */
#define DIRPROP_FLAG_MULTI_RUNS (1UL<<31)

/* are there any characters that are LTR or RTL? */
#define MASK_LTR (DIRPROP_FLAG(L)|DIRPROP_FLAG(EN)|DIRPROP_FLAG(ENL)|DIRPROP_FLAG(ENR)|DIRPROP_FLAG(AN)|DIRPROP_FLAG(LRE)|DIRPROP_FLAG(LRO)|DIRPROP_FLAG(LRI))
#define MASK_RTL (DIRPROP_FLAG(R)|DIRPROP_FLAG(AL)|DIRPROP_FLAG(RLE)|DIRPROP_FLAG(RLO)|DIRPROP_FLAG(RLI))
#define MASK_R_AL (DIRPROP_FLAG(R)|DIRPROP_FLAG(AL))
#define MASK_STRONG_EN_AN (DIRPROP_FLAG(L)|DIRPROP_FLAG(R)|DIRPROP_FLAG(AL)|DIRPROP_FLAG(EN)|DIRPROP_FLAG(AN))

/* explicit embedding codes */
#define MASK_EXPLICIT (DIRPROP_FLAG(LRE)|DIRPROP_FLAG(LRO)|DIRPROP_FLAG(RLE)|DIRPROP_FLAG(RLO)|DIRPROP_FLAG(PDF))

/* explicit isolate codes */
#define MASK_ISO (DIRPROP_FLAG(LRI)|DIRPROP_FLAG(RLI)|DIRPROP_FLAG(FSI)|DIRPROP_FLAG(PDI))

#define MASK_BN_EXPLICIT (DIRPROP_FLAG(BN)|MASK_EXPLICIT)

/* paragraph and segment separators */
#define MASK_B_S (DIRPROP_FLAG(B)|DIRPROP_FLAG(S))

/* all types that are counted as White Space or Neutral in some steps */
#define MASK_WS (MASK_B_S|DIRPROP_FLAG(WS)|MASK_BN_EXPLICIT|MASK_ISO)

/* types that are neutrals or could becomes neutrals in (Wn) */
#define MASK_POSSIBLE_N (DIRPROP_FLAG(ON)|DIRPROP_FLAG(CS)|DIRPROP_FLAG(ES)|DIRPROP_FLAG(ET)|MASK_WS)

/*
 *  These types may be changed to "e",
 *  the embedding type (L or R) of the run,
 *  in the BiDi algorithm (N2)
 */
#define MASK_EMBEDDING (DIRPROP_FLAG(NSM)|MASK_POSSIBLE_N)

/* the dirProp's L and R are defined to 0 and 1 values in UCharDirection */
#define GET_LR_FROM_LEVEL(level) ((DirProp)((level)&1))

#define IS_DEFAULT_LEVEL(level) ((level)>=0xfe)

/*
 *  The following bit is used for the directional isolate status.
 *  Stack entries corresponding to isolate sequences are greater than ISOLATE.
 */
#define ISOLATE  0x0100

U_CFUNC UBiDiLevel
ubidi_getParaLevelAtIndex(const UBiDi *pBiDi, int32_t index);

#define GET_PARALEVEL(ubidi, index) \
            ((UBiDiLevel)(!(ubidi)->defaultParaLevel || (index)<(ubidi)->paras[0].limit ? \
                         (ubidi)->paraLevel : ubidi_getParaLevelAtIndex((ubidi), (index))))

/* number of paras entries allocated initially without malloc */
#define SIMPLE_PARAS_COUNT      10
/* number of isolate entries allocated initially without malloc */
#define SIMPLE_ISOLATES_COUNT   5
/* number of isolate run entries for paired brackets allocated initially without malloc */
#define SIMPLE_OPENINGS_COUNT   20

#define UBIDIIMP_CR  0x000D
#define UBIDIIMP_LF  0x000A

/* Run structure for reordering --------------------------------------------- */
enum {
    LRM_BEFORE=1,
    LRM_AFTER=2,
    RLM_BEFORE=4,
    RLM_AFTER=8
};

typedef struct Para {
    int32_t limit;
    int32_t level;
} Para;

enum {                                  /* flags for Opening.flags */
    FOUND_L=DIRPROP_FLAG(L),
    FOUND_R=DIRPROP_FLAG(R)
};

typedef struct Opening {
    int32_t position;                   /* position of opening bracket */
    int32_t match;                      /* matching char or -position of closing bracket */
    int32_t contextPos;                 /* position of last strong char found before opening */
    uint16_t flags;                     /* bits for L or R/AL found within the pair */
    UBiDiDirection contextDir;          /* L or R according to last strong char before opening */
    uint8_t filler;                     /* to complete a nice multiple of 4 chars */
} Opening;

typedef struct IsoRun {
    int32_t  contextPos;                /* position of char determining context */
    uint16_t start;                     /* index of first opening entry for this run */
    uint16_t limit;                     /* index after last opening entry for this run */
    UBiDiLevel level;                   /* level of this run */
    DirProp lastStrong;                 /* bidi class of last strong char found in this run */
    DirProp lastBase;                   /* bidi class of last base char found in this run */
    UBiDiDirection contextDir;          /* L or R to use as context for following openings */
} IsoRun;

typedef struct BracketData {
    UBiDi   *pBiDi;
    /* array of opening entries which should be enough in most cases; no malloc() */
    Opening simpleOpenings[SIMPLE_OPENINGS_COUNT];
    Opening *openings;                  /* pointer to current array of entries */
    int32_t openingsCount;              /* number of allocated entries */
    int32_t isoRunLast;                 /* index of last used entry */
    /* array of nested isolated sequence entries; can never excess UBIDI_MAX_EXPLICIT_LEVEL
       + 1 for index 0, + 1 for before the first isolated sequence */
    IsoRun  isoRuns[UBIDI_MAX_EXPLICIT_LEVEL+2];
    UBool isNumbersSpecial;             /* reordering mode for NUMBERS_SPECIAL */
} BracketData;

typedef struct Isolate {
    int32_t startON;
    int32_t start1;
    int32_t state;
    int16_t stateImp;
} Isolate;

typedef struct Run {
    int32_t logicalStart,   /* first character of the run; b31 indicates even/odd level */
            visualLimit,    /* last visual position of the run +1 */
            insertRemove;   /* if >0, flags for inserting LRM/RLM before/after run,
                               if <0, count of bidi controls within run            */
} Run;

/* in a Run, logicalStart will get this bit set if the run level is odd */
#define INDEX_ODD_BIT (1UL<<31)

#define MAKE_INDEX_ODD_PAIR(index, level) ((index)|((int32_t)((level)&1)<<31))
#define ADD_ODD_BIT_FROM_LEVEL(x, level)  ((x)|=((int32_t)((level)&1)<<31))
#define REMOVE_ODD_BIT(x)                 ((x)&=~INDEX_ODD_BIT)

#define GET_INDEX(x)   ((x)&~INDEX_ODD_BIT)
#define GET_ODD_BIT(x) ((uint32_t)(x)>>31)
#define IS_ODD_RUN(x)  ((UBool)(((x)&INDEX_ODD_BIT)!=0))
#define IS_EVEN_RUN(x) ((UBool)(((x)&INDEX_ODD_BIT)==0))

U_CFUNC UBool
ubidi_getRuns(UBiDi *pBiDi, UErrorCode *pErrorCode);

/** BiDi control code points */
enum {
    ZWNJ_CHAR=0x200c,
    ZWJ_CHAR,
    LRM_CHAR,
    RLM_CHAR,
    LRE_CHAR=0x202a,
    RLE_CHAR,
    PDF_CHAR,
    LRO_CHAR,
    RLO_CHAR,
    LRI_CHAR=0x2066,
    RLI_CHAR,
    FSI_CHAR,
    PDI_CHAR
};

#define IS_BIDI_CONTROL_CHAR(c) (((uint32_t)(c)&0xfffffffc)==ZWNJ_CHAR || (uint32_t)((c)-LRE_CHAR)<5 || (uint32_t)((c)-LRI_CHAR)<4)

/* InsertPoints structure for noting where to put BiDi marks ---------------- */

typedef struct Point {
    int32_t pos;            /* position in text */
    int32_t flag;           /* flag for LRM/RLM, before/after */
} Point;

typedef struct InsertPoints {
    int32_t capacity;       /* number of points allocated */
    int32_t size;           /* number of points used */
    int32_t confirmed;      /* number of points confirmed */
    UErrorCode errorCode;   /* for eventual memory shortage */
    Point *points;          /* pointer to array of points */
} InsertPoints;


/* UBiDi structure ----------------------------------------------------------- */

struct UBiDi {
    /* pointer to parent paragraph object (pointer to self if this object is
     * a paragraph object); set to NULL in a newly opened object; set to a
     * real value after a successful execution of ubidi_setPara or ubidi_setLine
     */
    const UBiDi * pParaBiDi;

    /* alias pointer to the current text */
    const UChar *text;

    /* length of the current text */
    int32_t originalLength;

    /* if the UBIDI_OPTION_STREAMING option is set, this is the length
     * of text actually processed by ubidi_setPara, which may be shorter than
     * the original length.
     * Otherwise, it is identical to the original length.
     */
    int32_t length;

    /* if the UBIDI_OPTION_REMOVE_CONTROLS option is set, and/or
     * marks are allowed to be inserted in one of the reordering mode, the
     * length of the result string may be different from the processed length.
     */
    int32_t resultLength;

    /* memory sizes in bytes */
    int32_t dirPropsSize, levelsSize, openingsSize, parasSize, runsSize, isolatesSize;

    /* allocated memory */
    DirProp *dirPropsMemory;
    UBiDiLevel *levelsMemory;
    Opening *openingsMemory;
    Para *parasMemory;
    Run *runsMemory;
    Isolate *isolatesMemory;

    /* indicators for whether memory may be allocated after ubidi_open() */
    UBool mayAllocateText, mayAllocateRuns;

    /* arrays with one value per text-character */
    DirProp *dirProps;
    UBiDiLevel *levels;

    /* are we performing an approximation of the "inverse BiDi" algorithm? */
    UBool isInverse;

    /* are we using the basic algorithm or its variation? */
    UBiDiReorderingMode reorderingMode;

    /* UBIDI_REORDER_xxx values must be ordered so that all the regular
     * logical to visual modes come first, and all inverse BiDi modes
     * come last.
     */
    #define UBIDI_REORDER_LAST_LOGICAL_TO_VISUAL    UBIDI_REORDER_NUMBERS_SPECIAL

    /* bitmask for reordering options */
    uint32_t reorderingOptions;

    /* must block separators receive level 0? */
    UBool orderParagraphsLTR;

    /* the paragraph level */
    UBiDiLevel paraLevel;
    /* original paraLevel when contextual */
    /* must be one of UBIDI_DEFAULT_xxx or 0 if not contextual */
    UBiDiLevel defaultParaLevel;

    /* context data */
    const UChar *prologue;
    int32_t proLength;
    const UChar *epilogue;
    int32_t epiLength;

    /* the following is set in ubidi_setPara, used in processPropertySeq */
    const struct ImpTabPair * pImpTabPair;  /* pointer to levels state table pair */

    /* the overall paragraph or line directionality - see UBiDiDirection */
    UBiDiDirection direction;

    /* flags is a bit set for which directional properties are in the text */
    Flags flags;

    /* lastArabicPos is index to the last AL in the text, -1 if none */
    int32_t lastArabicPos;

    /* characters after trailingWSStart are WS and are */
    /* implicitly at the paraLevel (rule (L1)) - levels may not reflect that */
    int32_t trailingWSStart;

    /* fields for paragraph handling */
    int32_t paraCount;                  /* set in getDirProps() */
    /* filled in getDirProps() */
    Para *paras;

    /* for relatively short text, we only need a tiny array of paras (no malloc()) */
    Para simpleParas[SIMPLE_PARAS_COUNT];

    /* fields for line reordering */
    int32_t runCount;     /* ==-1: runs not set up yet */
    Run *runs;

    /* for non-mixed text, we only need a tiny array of runs (no malloc()) */
    Run simpleRuns[1];

    /* maximum or current nesting depth of isolate sequences */
    /* Within resolveExplicitLevels() and checkExplicitLevels(), this is the maximal
       nesting encountered.
       Within resolveImplicitLevels(), this is the index of the current isolates
       stack entry. */
    int32_t isolateCount;
    Isolate *isolates;

    /* for simple text, have a small stack (no malloc()) */
    Isolate simpleIsolates[SIMPLE_ISOLATES_COUNT];

    /* for inverse Bidi with insertion of directional marks */
    InsertPoints insertPoints;

    /* for option UBIDI_OPTION_REMOVE_CONTROLS */
    int32_t controlCount;

    /* for Bidi class callback */
    UBiDiClassCallback *fnClassCallback;    /* action pointer */
    const void *coClassCallback;            /* context pointer */
};

#define IS_VALID_PARA(x) ((x) && ((x)->pParaBiDi==(x)))
#define IS_VALID_PARA_OR_LINE(x) ((x) && ((x)->pParaBiDi==(x) || (((x)->pParaBiDi) && (x)->pParaBiDi->pParaBiDi==(x)->pParaBiDi)))

typedef union {
    DirProp *dirPropsMemory;
    UBiDiLevel *levelsMemory;
    Opening *openingsMemory;
    Para *parasMemory;
    Run *runsMemory;
    Isolate *isolatesMemory;
} BidiMemoryForAllocation;

/* Macros for initial checks at function entry */
#define RETURN_IF_NULL_OR_FAILING_ERRCODE(pErrcode, retvalue) UPRV_BLOCK_MACRO_BEGIN { \
    if((pErrcode)==NULL || U_FAILURE(*pErrcode)) return retvalue; \
} UPRV_BLOCK_MACRO_END
#define RETURN_IF_NOT_VALID_PARA(bidi, errcode, retvalue) UPRV_BLOCK_MACRO_BEGIN { \
    if(!IS_VALID_PARA(bidi)) { \
        errcode=U_INVALID_STATE_ERROR; \
        return retvalue; \
    } \
} UPRV_BLOCK_MACRO_END
#define RETURN_IF_NOT_VALID_PARA_OR_LINE(bidi, errcode, retvalue) UPRV_BLOCK_MACRO_BEGIN { \
    if(!IS_VALID_PARA_OR_LINE(bidi)) { \
        errcode=U_INVALID_STATE_ERROR; \
        return retvalue; \
    } \
} UPRV_BLOCK_MACRO_END
#define RETURN_IF_BAD_RANGE(arg, start, limit, errcode, retvalue) UPRV_BLOCK_MACRO_BEGIN { \
    if((arg)<(start) || (arg)>=(limit)) { \
        (errcode)=U_ILLEGAL_ARGUMENT_ERROR; \
        return retvalue; \
    } \
} UPRV_BLOCK_MACRO_END

#define RETURN_VOID_IF_NULL_OR_FAILING_ERRCODE(pErrcode) UPRV_BLOCK_MACRO_BEGIN { \
    if((pErrcode)==NULL || U_FAILURE(*pErrcode)) return; \
} UPRV_BLOCK_MACRO_END
#define RETURN_VOID_IF_NOT_VALID_PARA(bidi, errcode) UPRV_BLOCK_MACRO_BEGIN { \
    if(!IS_VALID_PARA(bidi)) { \
        errcode=U_INVALID_STATE_ERROR; \
        return; \
    } \
} UPRV_BLOCK_MACRO_END
#define RETURN_VOID_IF_NOT_VALID_PARA_OR_LINE(bidi, errcode) UPRV_BLOCK_MACRO_BEGIN { \
    if(!IS_VALID_PARA_OR_LINE(bidi)) { \
        errcode=U_INVALID_STATE_ERROR; \
        return; \
    } \
} UPRV_BLOCK_MACRO_END
#define RETURN_VOID_IF_BAD_RANGE(arg, start, limit, errcode) UPRV_BLOCK_MACRO_BEGIN { \
    if((arg)<(start) || (arg)>=(limit)) { \
        (errcode)=U_ILLEGAL_ARGUMENT_ERROR; \
        return; \
    } \
} UPRV_BLOCK_MACRO_END

/* helper function to (re)allocate memory if allowed */
U_CFUNC UBool
ubidi_getMemory(BidiMemoryForAllocation *pMemory, int32_t *pSize, UBool mayAllocate, int32_t sizeNeeded);

/* helper macros for each allocated array in UBiDi */
#define getDirPropsMemory(pBiDi, length) \
        ubidi_getMemory((BidiMemoryForAllocation *)&(pBiDi)->dirPropsMemory, &(pBiDi)->dirPropsSize, \
                        (pBiDi)->mayAllocateText, (length))

#define getLevelsMemory(pBiDi, length) \
        ubidi_getMemory((BidiMemoryForAllocation *)&(pBiDi)->levelsMemory, &(pBiDi)->levelsSize, \
                        (pBiDi)->mayAllocateText, (length))

#define getRunsMemory(pBiDi, length) \
        ubidi_getMemory((BidiMemoryForAllocation *)&(pBiDi)->runsMemory, &(pBiDi)->runsSize, \
                        (pBiDi)->mayAllocateRuns, (length)*sizeof(Run))

/* additional macros used by ubidi_open() - always allow allocation */
#define getInitialDirPropsMemory(pBiDi, length) \
        ubidi_getMemory((BidiMemoryForAllocation *)&(pBiDi)->dirPropsMemory, &(pBiDi)->dirPropsSize, \
                        TRUE, (length))

#define getInitialLevelsMemory(pBiDi, length) \
        ubidi_getMemory((BidiMemoryForAllocation *)&(pBiDi)->levelsMemory, &(pBiDi)->levelsSize, \
                        TRUE, (length))

#define getInitialOpeningsMemory(pBiDi, length) \
        ubidi_getMemory((BidiMemoryForAllocation *)&(pBiDi)->openingsMemory, &(pBiDi)->openingsSize, \
                        TRUE, (length)*sizeof(Opening))

#define getInitialParasMemory(pBiDi, length) \
        ubidi_getMemory((BidiMemoryForAllocation *)&(pBiDi)->parasMemory, &(pBiDi)->parasSize, \
                        TRUE, (length)*sizeof(Para))

#define getInitialRunsMemory(pBiDi, length) \
        ubidi_getMemory((BidiMemoryForAllocation *)&(pBiDi)->runsMemory, &(pBiDi)->runsSize, \
                        TRUE, (length)*sizeof(Run))

#define getInitialIsolatesMemory(pBiDi, length) \
        ubidi_getMemory((BidiMemoryForAllocation *)&(pBiDi)->isolatesMemory, &(pBiDi)->isolatesSize, \
                        TRUE, (length)*sizeof(Isolate))




/*
 * General implementation notes:
 *
 * Throughout the implementation, there are comments like (W2) that refer to
 * rules of the BiDi algorithm, in this example to the second rule of the
 * resolution of weak types.
 *
 * For handling surrogate pairs, where two UChar's form one "abstract" (or UTF-32)
 * character according to UTF-16, the second UChar gets the directional property of
 * the entire character assigned, while the first one gets a BN, a boundary
 * neutral, type, which is ignored by most of the algorithm according to
 * rule (X9) and the implementation suggestions of the BiDi algorithm.
 *
 * Later, adjustWSLevels() will set the level for each BN to that of the
 * following character (UChar), which results in surrogate pairs getting the
 * same level on each of their surrogates.
 *
 * In a UTF-8 implementation, the same thing could be done: the last byte of
 * a multi-byte sequence would get the "real" property, while all previous
 * bytes of that sequence would get BN.
 *
 * It is not possible to assign all those parts of a character the same real
 * property because this would fail in the resolution of weak types with rules
 * that look at immediately surrounding types.
 *
 * As a related topic, this implementation does not remove Boundary Neutral
 * types from the input, but ignores them wherever this is relevant.
 * For example, the loop for the resolution of the weak types reads
 * types until it finds a non-BN.
 * Also, explicit embedding codes are neither changed into BN nor removed.
 * They are only treated the same way real BNs are.
 * As stated before, adjustWSLevels() takes care of them at the end.
 * For the purpose of conformance, the levels of all these codes
 * do not matter.
 *
 * Note that this implementation modifies the dirProps
 * after the initial setup, when applying X5c (replace FSI by LRI or RLI),
 * X6, N0 (replace paired brackets by L or R).
 *
 * In this implementation, the resolution of weak types (W1 to W6),
 * neutrals (N1 and N2), and the assignment of the resolved level (In)
 * are all done in one single loop, in resolveImplicitLevels().
 * Changes of dirProp values are done on the fly, without writing
 * them back to the dirProps array.
 *
 *
 * This implementation contains code that allows to bypass steps of the
 * algorithm that are not needed on the specific paragraph
 * in order to speed up the most common cases considerably,
 * like text that is entirely LTR, or RTL text without numbers.
 *
 * Most of this is done by setting a bit for each directional property
 * in a flags variable and later checking for whether there are
 * any LTR characters or any RTL characters, or both, whether
 * there are any explicit embedding codes, etc.
 *
 * If the (Xn) steps are performed, then the flags are re-evaluated,
 * because they will then not contain the embedding codes any more
 * and will be adjusted for override codes, so that subsequently
 * more bypassing may be possible than what the initial flags suggested.
 *
 * If the text is not mixed-directional, then the
 * algorithm steps for the weak type resolution are not performed,
 * and all levels are set to the paragraph level.
 *
 * If there are no explicit embedding codes, then the (Xn) steps
 * are not performed.
 *
 * If embedding levels are supplied as a parameter, then all
 * explicit embedding codes are ignored, and the (Xn) steps
 * are not performed.
 *
 * White Space types could get the level of the run they belong to,
 * and are checked with a test of (flags&MASK_EMBEDDING) to
 * consider if the paragraph direction should be considered in
 * the flags variable.
 *
 * If there are no White Space types in the paragraph, then
 * (L1) is not necessary in adjustWSLevels().
 */

/* to avoid some conditional statements, use tiny constant arrays */
static const Flags flagLR[2]={ DIRPROP_FLAG(L), DIRPROP_FLAG(R) };
static const Flags flagE[2]={ DIRPROP_FLAG(LRE), DIRPROP_FLAG(RLE) };
static const Flags flagO[2]={ DIRPROP_FLAG(LRO), DIRPROP_FLAG(RLO) };

#define DIRPROP_FLAG_LR(level) flagLR[(level)&1]
#define DIRPROP_FLAG_E(level)  flagE[(level)&1]
#define DIRPROP_FLAG_O(level)  flagO[(level)&1]

#define DIR_FROM_STRONG(strong) ((strong)==L ? L : R)

#define NO_OVERRIDE(level)  ((level)&~UBIDI_LEVEL_OVERRIDE)

/* UBiDi object management -------------------------------------------------- */

U_CAPI UBiDi * U_EXPORT2
ubidi_open(void)
{
    UErrorCode errorCode=U_ZERO_ERROR;
    return ubidi_openSized(0, 0, &errorCode);
}

U_CAPI UBiDi * U_EXPORT2
ubidi_openSized(int32_t maxLength, int32_t maxRunCount, UErrorCode *pErrorCode) {
    UBiDi *pBiDi;

    /* check the argument values */
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return NULL;
    } else if(maxLength<0 || maxRunCount<0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;    /* invalid arguments */
    }

    /* allocate memory for the object */
    pBiDi=(UBiDi *)uprv_malloc(sizeof(UBiDi));
    if(pBiDi==NULL) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    /* reset the object, all pointers NULL, all flags FALSE, all sizes 0 */
    uprv_memset(pBiDi, 0, sizeof(UBiDi));

    /* allocate memory for arrays as requested */
    if(maxLength>0) {
        if( !getInitialDirPropsMemory(pBiDi, maxLength) ||
            !getInitialLevelsMemory(pBiDi, maxLength)
        ) {
            *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        }
    } else {
        pBiDi->mayAllocateText=TRUE;
    }

    if(maxRunCount>0) {
        if(maxRunCount==1) {
            /* use simpleRuns[] */
            pBiDi->runsSize=sizeof(Run);
        } else if(!getInitialRunsMemory(pBiDi, maxRunCount)) {
            *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        }
    } else {
        pBiDi->mayAllocateRuns=TRUE;
    }

    if(U_SUCCESS(*pErrorCode)) {
        return pBiDi;
    } else {
        ubidi_close(pBiDi);
        return NULL;
    }
}

/*
 * We are allowed to allocate memory if memory==NULL or
 * mayAllocate==TRUE for each array that we need.
 * We also try to grow memory as needed if we
 * allocate it.
 *
 * Assume sizeNeeded>0.
 * If *pMemory!=NULL, then assume *pSize>0.
 *
 * ### this realloc() may unnecessarily copy the old data,
 * which we know we don't need any more;
 * is this the best way to do this??
 */
U_CFUNC UBool
ubidi_getMemory(BidiMemoryForAllocation *bidiMem, int32_t *pSize, UBool mayAllocate, int32_t sizeNeeded) {
    void **pMemory = (void **)bidiMem;
    /* check for existing memory */
    if(*pMemory==NULL) {
        /* we need to allocate memory */
        if(mayAllocate && (*pMemory=uprv_malloc(sizeNeeded))!=NULL) {
            *pSize=sizeNeeded;
            return TRUE;
        } else {
            return FALSE;
        }
    } else {
        if(sizeNeeded<=*pSize) {
            /* there is already enough memory */
            return TRUE;
        }
        else if(!mayAllocate) {
            /* not enough memory, and we must not allocate */
            return FALSE;
        } else {
            /* we try to grow */
            void *memory;
            /* in most cases, we do not need the copy-old-data part of
             * realloc, but it is needed when adding runs using getRunsMemory()
             * in setParaRunsOnly()
             */
            if((memory=uprv_realloc(*pMemory, sizeNeeded))!=NULL) {
                *pMemory=memory;
                *pSize=sizeNeeded;
                return TRUE;
            } else {
                /* we failed to grow */
                return FALSE;
            }
        }
    }
}

U_CAPI void U_EXPORT2
ubidi_close(UBiDi *pBiDi) {
    if(pBiDi!=NULL) {
        pBiDi->pParaBiDi=NULL;          /* in case one tries to reuse this block */
        if(pBiDi->dirPropsMemory!=NULL) {
            uprv_free(pBiDi->dirPropsMemory);
        }
        if(pBiDi->levelsMemory!=NULL) {
            uprv_free(pBiDi->levelsMemory);
        }
        if(pBiDi->openingsMemory!=NULL) {
            uprv_free(pBiDi->openingsMemory);
        }
        if(pBiDi->parasMemory!=NULL) {
            uprv_free(pBiDi->parasMemory);
        }
        if(pBiDi->runsMemory!=NULL) {
            uprv_free(pBiDi->runsMemory);
        }
        if(pBiDi->isolatesMemory!=NULL) {
            uprv_free(pBiDi->isolatesMemory);
        }
        if(pBiDi->insertPoints.points!=NULL) {
            uprv_free(pBiDi->insertPoints.points);
        }

        uprv_free(pBiDi);
    }
}

/* set to approximate "inverse BiDi" ---------------------------------------- */

U_CAPI void U_EXPORT2
ubidi_setInverse(UBiDi *pBiDi, UBool isInverse) {
    if(pBiDi!=NULL) {
        pBiDi->isInverse=isInverse;
        pBiDi->reorderingMode = isInverse ? UBIDI_REORDER_INVERSE_NUMBERS_AS_L
                                          : UBIDI_REORDER_DEFAULT;
    }
}

U_CAPI UBool U_EXPORT2
ubidi_isInverse(UBiDi *pBiDi) {
    if(pBiDi!=NULL) {
        return pBiDi->isInverse;
    } else {
        return FALSE;
    }
}

/* FOOD FOR THOUGHT: currently the reordering modes are a mixture of
 * algorithm for direct BiDi, algorithm for inverse BiDi and the bizarre
 * concept of RUNS_ONLY which is a double operation.
 * It could be advantageous to divide this into 3 concepts:
 * a) Operation: direct / inverse / RUNS_ONLY
 * b) Direct algorithm: default / NUMBERS_SPECIAL / GROUP_NUMBERS_WITH_R
 * c) Inverse algorithm: default / INVERSE_LIKE_DIRECT / NUMBERS_SPECIAL
 * This would allow combinations not possible today like RUNS_ONLY with
 * NUMBERS_SPECIAL.
 * Also allow to set INSERT_MARKS for the direct step of RUNS_ONLY and
 * REMOVE_CONTROLS for the inverse step.
 * Not all combinations would be supported, and probably not all do make sense.
 * This would need to document which ones are supported and what are the
 * fallbacks for unsupported combinations.
 */
U_CAPI void U_EXPORT2
ubidi_setReorderingMode(UBiDi *pBiDi, UBiDiReorderingMode reorderingMode) {
    if ((pBiDi!=NULL) && (reorderingMode >= UBIDI_REORDER_DEFAULT)
                        && (reorderingMode < UBIDI_REORDER_COUNT)) {
        pBiDi->reorderingMode = reorderingMode;
        pBiDi->isInverse = (UBool)(reorderingMode == UBIDI_REORDER_INVERSE_NUMBERS_AS_L);
    }
}

U_CAPI UBiDiReorderingMode U_EXPORT2
ubidi_getReorderingMode(UBiDi *pBiDi) {
    if (pBiDi!=NULL) {
        return pBiDi->reorderingMode;
    } else {
        return UBIDI_REORDER_DEFAULT;
    }
}

U_CAPI void U_EXPORT2
ubidi_setReorderingOptions(UBiDi *pBiDi, uint32_t reorderingOptions) {
    if (reorderingOptions & UBIDI_OPTION_REMOVE_CONTROLS) {
        reorderingOptions&=~UBIDI_OPTION_INSERT_MARKS;
    }
    if (pBiDi!=NULL) {
        pBiDi->reorderingOptions=reorderingOptions;
    }
}

U_CAPI uint32_t U_EXPORT2
ubidi_getReorderingOptions(UBiDi *pBiDi) {
    if (pBiDi!=NULL) {
        return pBiDi->reorderingOptions;
    } else {
        return 0;
    }
}

U_CAPI UBiDiDirection U_EXPORT2
ubidi_getBaseDirection(const UChar *text,
int32_t length){

    int32_t i;
    UChar32 uchar;
    UCharDirection dir;

    if( text==NULL || length<-1 ){
        return UBIDI_NEUTRAL;
    }

    if(length==-1) {
        length=u_strlen(text);
    }

    for( i = 0 ; i < length; ) {
        /* i is incremented by U16_NEXT */
        U16_NEXT(text, i, length, uchar);
        dir = u_charDirection(uchar);
        if( dir == U_LEFT_TO_RIGHT )
                return UBIDI_LTR;
        if( dir == U_RIGHT_TO_LEFT || dir ==U_RIGHT_TO_LEFT_ARABIC )
                return UBIDI_RTL;
    }
    return UBIDI_NEUTRAL;
}

/* perform (P2)..(P3) ------------------------------------------------------- */

/**
 * Returns the directionality of the first strong character
 * after the last B in prologue, if any.
 * Requires prologue!=null.
 */
static DirProp
firstL_R_AL(UBiDi *pBiDi) {
    const UChar *text=pBiDi->prologue;
    int32_t length=pBiDi->proLength;
    int32_t i;
    UChar32 uchar;
    DirProp dirProp, result=ON;
    for(i=0; i<length; ) {
        /* i is incremented by U16_NEXT */
        U16_NEXT(text, i, length, uchar);
        dirProp=(DirProp)ubidi_getCustomizedClass(pBiDi, uchar);
        if(result==ON) {
            if(dirProp==L || dirProp==R || dirProp==AL) {
                result=dirProp;
            }
        } else {
            if(dirProp==B) {
                result=ON;
            }
        }
    }
    return result;
}

/*
 * Check that there are enough entries in the array pointed to by pBiDi->paras
 */
static UBool
checkParaCount(UBiDi *pBiDi) {
    int32_t count=pBiDi->paraCount;
    if(pBiDi->paras==pBiDi->simpleParas) {
        if(count<=SIMPLE_PARAS_COUNT)
            return TRUE;
        if(!getInitialParasMemory(pBiDi, SIMPLE_PARAS_COUNT * 2))
            return FALSE;
        pBiDi->paras=pBiDi->parasMemory;
        uprv_memcpy(pBiDi->parasMemory, pBiDi->simpleParas, SIMPLE_PARAS_COUNT * sizeof(Para));
        return TRUE;
    }
    if(!getInitialParasMemory(pBiDi, count * 2))
        return FALSE;
    pBiDi->paras=pBiDi->parasMemory;
    return TRUE;
}

/*
 * Get the directional properties for the text, calculate the flags bit-set, and
 * determine the paragraph level if necessary (in pBiDi->paras[i].level).
 * FSI initiators are also resolved and their dirProp replaced with LRI or RLI.
 * When encountering an FSI, it is initially replaced with an LRI, which is the
 * default. Only if a strong R or AL is found within its scope will the LRI be
 * replaced by an RLI.
 */
static UBool
getDirProps(UBiDi *pBiDi) {
    const UChar *text=pBiDi->text;
    DirProp *dirProps=pBiDi->dirPropsMemory;    /* pBiDi->dirProps is const */

    int32_t i=0, originalLength=pBiDi->originalLength;
    Flags flags=0;      /* collect all directionalities in the text */
    UChar32 uchar;
    DirProp dirProp=0, defaultParaLevel=0;  /* initialize to avoid compiler warnings */
    UBool isDefaultLevel=IS_DEFAULT_LEVEL(pBiDi->paraLevel);
    /* for inverse BiDi, the default para level is set to RTL if there is a
       strong R or AL character at either end of the text                            */
    UBool isDefaultLevelInverse=isDefaultLevel && (UBool)
            (pBiDi->reorderingMode==UBIDI_REORDER_INVERSE_LIKE_DIRECT ||
             pBiDi->reorderingMode==UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL);
    int32_t lastArabicPos=-1;
    int32_t controlCount=0;
    UBool removeBiDiControls = (UBool)(pBiDi->reorderingOptions &
                                       UBIDI_OPTION_REMOVE_CONTROLS);

    enum State {
         NOT_SEEKING_STRONG,            /* 0: not contextual paraLevel, not after FSI */
         SEEKING_STRONG_FOR_PARA,       /* 1: looking for first strong char in para */
         SEEKING_STRONG_FOR_FSI,        /* 2: looking for first strong after FSI */
         LOOKING_FOR_PDI                /* 3: found strong after FSI, looking for PDI */
    };
    State state;
    DirProp lastStrong=ON;              /* for default level & inverse BiDi */
    /* The following stacks are used to manage isolate sequences. Those
       sequences may be nested, but obviously never more deeply than the
       maximum explicit embedding level.
       lastStack is the index of the last used entry in the stack. A value of -1
       means that there is no open isolate sequence.
       lastStack is reset to -1 on paragraph boundaries. */
    /* The following stack contains the position of the initiator of
       each open isolate sequence */
    int32_t isolateStartStack[UBIDI_MAX_EXPLICIT_LEVEL+1];
    /* The following stack contains the last known state before
       encountering the initiator of an isolate sequence */
    State  previousStateStack[UBIDI_MAX_EXPLICIT_LEVEL+1];
    int32_t stackLast=-1;

    if(pBiDi->reorderingOptions & UBIDI_OPTION_STREAMING)
        pBiDi->length=0;
    defaultParaLevel=pBiDi->paraLevel&1;
    if(isDefaultLevel) {
        pBiDi->paras[0].level=defaultParaLevel;
        lastStrong=defaultParaLevel;
        if(pBiDi->proLength>0 &&                    /* there is a prologue */
           (dirProp=firstL_R_AL(pBiDi))!=ON) {  /* with a strong character */
            if(dirProp==L)
                pBiDi->paras[0].level=0;    /* set the default para level */
            else
                pBiDi->paras[0].level=1;    /* set the default para level */
            state=NOT_SEEKING_STRONG;
        } else {
            state=SEEKING_STRONG_FOR_PARA;
        }
    } else {
        pBiDi->paras[0].level=pBiDi->paraLevel;
        state=NOT_SEEKING_STRONG;
    }
    /* count paragraphs and determine the paragraph level (P2..P3) */
    /*
     * see comment in ubidi.h:
     * the UBIDI_DEFAULT_XXX values are designed so that
     * their bit 0 alone yields the intended default
     */
    for( /* i=0 above */ ; i<originalLength; ) {
        /* i is incremented by U16_NEXT */
        U16_NEXT(text, i, originalLength, uchar);
        flags|=DIRPROP_FLAG(dirProp=(DirProp)ubidi_getCustomizedClass(pBiDi, uchar));
        dirProps[i-1]=dirProp;
        if(uchar>0xffff) {  /* set the lead surrogate's property to BN */
            flags|=DIRPROP_FLAG(BN);
            dirProps[i-2]=BN;
        }
        if(removeBiDiControls && IS_BIDI_CONTROL_CHAR(uchar))
            controlCount++;
        if(dirProp==L) {
            if(state==SEEKING_STRONG_FOR_PARA) {
                pBiDi->paras[pBiDi->paraCount-1].level=0;
                state=NOT_SEEKING_STRONG;
            }
            else if(state==SEEKING_STRONG_FOR_FSI) {
                if(stackLast<=UBIDI_MAX_EXPLICIT_LEVEL) {
                    /* no need for next statement, already set by default */
                    /* dirProps[isolateStartStack[stackLast]]=LRI; */
                    flags|=DIRPROP_FLAG(LRI);
                }
                state=LOOKING_FOR_PDI;
            }
            lastStrong=L;
            continue;
        }
        if(dirProp==R || dirProp==AL) {
            if(state==SEEKING_STRONG_FOR_PARA) {
                pBiDi->paras[pBiDi->paraCount-1].level=1;
                state=NOT_SEEKING_STRONG;
            }
            else if(state==SEEKING_STRONG_FOR_FSI) {
                if(stackLast<=UBIDI_MAX_EXPLICIT_LEVEL) {
                    dirProps[isolateStartStack[stackLast]]=RLI;
                    flags|=DIRPROP_FLAG(RLI);
                }
                state=LOOKING_FOR_PDI;
            }
            lastStrong=R;
            if(dirProp==AL)
                lastArabicPos=i-1;
            continue;
        }
        if(dirProp>=FSI && dirProp<=RLI) {  /* FSI, LRI or RLI */
            stackLast++;
            if(stackLast<=UBIDI_MAX_EXPLICIT_LEVEL) {
                isolateStartStack[stackLast]=i-1;
                previousStateStack[stackLast]=state;
            }
            if(dirProp==FSI) {
                dirProps[i-1]=LRI;      /* default if no strong char */
                state=SEEKING_STRONG_FOR_FSI;
            }
            else
                state=LOOKING_FOR_PDI;
            continue;
        }
        if(dirProp==PDI) {
            if(state==SEEKING_STRONG_FOR_FSI) {
                if(stackLast<=UBIDI_MAX_EXPLICIT_LEVEL) {
                    /* no need for next statement, already set by default */
                    /* dirProps[isolateStartStack[stackLast]]=LRI; */
                    flags|=DIRPROP_FLAG(LRI);
                }
            }
            if(stackLast>=0) {
                if(stackLast<=UBIDI_MAX_EXPLICIT_LEVEL)
                    state=previousStateStack[stackLast];
                stackLast--;
            }
            continue;
        }
        if(dirProp==B) {
            if(i<originalLength && uchar==UBIDIIMP_CR && text[i]==UBIDIIMP_LF) /* do nothing on the CR */
                continue;
            pBiDi->paras[pBiDi->paraCount-1].limit=i;
            if(isDefaultLevelInverse && lastStrong==R)
                pBiDi->paras[pBiDi->paraCount-1].level=1;
            if(pBiDi->reorderingOptions & UBIDI_OPTION_STREAMING) {
                /* When streaming, we only process whole paragraphs
                   thus some updates are only done on paragraph boundaries */
                pBiDi->length=i;        /* i is index to next character */
                pBiDi->controlCount=controlCount;
            }
            if(i<originalLength) {              /* B not last char in text */
                pBiDi->paraCount++;
                if(checkParaCount(pBiDi)==FALSE)    /* not enough memory for a new para entry */
                    return FALSE;
                if(isDefaultLevel) {
                    pBiDi->paras[pBiDi->paraCount-1].level=defaultParaLevel;
                    state=SEEKING_STRONG_FOR_PARA;
                    lastStrong=defaultParaLevel;
                } else {
                    pBiDi->paras[pBiDi->paraCount-1].level=pBiDi->paraLevel;
                    state=NOT_SEEKING_STRONG;
                }
                stackLast=-1;
            }
            continue;
        }
    }
    /* Ignore still open isolate sequences with overflow */
    if(stackLast>UBIDI_MAX_EXPLICIT_LEVEL) {
        stackLast=UBIDI_MAX_EXPLICIT_LEVEL;
        state=SEEKING_STRONG_FOR_FSI;   /* to be on the safe side */
    }
    /* Resolve direction of still unresolved open FSI sequences */
    while(stackLast>=0) {
        if(state==SEEKING_STRONG_FOR_FSI) {
            /* no need for next statement, already set by default */
            /* dirProps[isolateStartStack[stackLast]]=LRI; */
            flags|=DIRPROP_FLAG(LRI);
            break;
        }
        state=previousStateStack[stackLast];
        stackLast--;
    }
    /* When streaming, ignore text after the last paragraph separator */
    if(pBiDi->reorderingOptions & UBIDI_OPTION_STREAMING) {
        if(pBiDi->length<originalLength)
            pBiDi->paraCount--;
    } else {
        pBiDi->paras[pBiDi->paraCount-1].limit=originalLength;
        pBiDi->controlCount=controlCount;
    }
    /* For inverse bidi, default para direction is RTL if there is
       a strong R or AL at either end of the paragraph */
    if(isDefaultLevelInverse && lastStrong==R) {
        pBiDi->paras[pBiDi->paraCount-1].level=1;
    }
    if(isDefaultLevel) {
        pBiDi->paraLevel=static_cast<UBiDiLevel>(pBiDi->paras[0].level);
    }
    /* The following is needed to resolve the text direction for default level
       paragraphs containing no strong character */
    for(i=0; i<pBiDi->paraCount; i++)
        flags|=DIRPROP_FLAG_LR(pBiDi->paras[i].level);

    if(pBiDi->orderParagraphsLTR && (flags&DIRPROP_FLAG(B))) {
        flags|=DIRPROP_FLAG(L);
    }
    pBiDi->flags=flags;
    pBiDi->lastArabicPos=lastArabicPos;
    return TRUE;
}

/* determine the paragraph level at position index */
U_CFUNC UBiDiLevel
ubidi_getParaLevelAtIndex(const UBiDi *pBiDi, int32_t pindex) {
    int32_t i;
    for(i=0; i<pBiDi->paraCount; i++)
        if(pindex<pBiDi->paras[i].limit)
            break;
    if(i>=pBiDi->paraCount)
        i=pBiDi->paraCount-1;
    return (UBiDiLevel)(pBiDi->paras[i].level);
}

/* Functions for handling paired brackets ----------------------------------- */

/* In the isoRuns array, the first entry is used for text outside of any
   isolate sequence.  Higher entries are used for each more deeply nested
   isolate sequence. isoRunLast is the index of the last used entry.  The
   openings array is used to note the data of opening brackets not yet
   matched by a closing bracket, or matched but still susceptible to change
   level.
   Each isoRun entry contains the index of the first and
   one-after-last openings entries for pending opening brackets it
   contains.  The next openings entry to use is the one-after-last of the
   most deeply nested isoRun entry.
   isoRun entries also contain their current embedding level and the last
   encountered strong character, since these will be needed to resolve
   the level of paired brackets.  */

static void
bracketInit(UBiDi *pBiDi, BracketData *bd) {
    bd->pBiDi=pBiDi;
    bd->isoRunLast=0;
    bd->isoRuns[0].start=0;
    bd->isoRuns[0].limit=0;
    bd->isoRuns[0].level=GET_PARALEVEL(pBiDi, 0);
    UBiDiLevel t = GET_PARALEVEL(pBiDi, 0) & 1;
    bd->isoRuns[0].lastStrong = bd->isoRuns[0].lastBase = t;
    bd->isoRuns[0].contextDir = (UBiDiDirection)t;
    bd->isoRuns[0].contextPos=0;
    if(pBiDi->openingsMemory) {
        bd->openings=pBiDi->openingsMemory;
        bd->openingsCount=pBiDi->openingsSize / sizeof(Opening);
    } else {
        bd->openings=bd->simpleOpenings;
        bd->openingsCount=SIMPLE_OPENINGS_COUNT;
    }
    bd->isNumbersSpecial=bd->pBiDi->reorderingMode==UBIDI_REORDER_NUMBERS_SPECIAL ||
                         bd->pBiDi->reorderingMode==UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL;
}

/* paragraph boundary */
static void
bracketProcessB(BracketData *bd, UBiDiLevel level) {
    bd->isoRunLast=0;
    bd->isoRuns[0].limit=0;
    bd->isoRuns[0].level=level;
    bd->isoRuns[0].lastStrong=bd->isoRuns[0].lastBase=level&1;
    bd->isoRuns[0].contextDir=(UBiDiDirection)(level&1);
    bd->isoRuns[0].contextPos=0;
}

/* LRE, LRO, RLE, RLO, PDF */
static void
bracketProcessBoundary(BracketData *bd, int32_t lastCcPos,
                       UBiDiLevel contextLevel, UBiDiLevel embeddingLevel) {
    IsoRun *pLastIsoRun=&bd->isoRuns[bd->isoRunLast];
    DirProp *dirProps=bd->pBiDi->dirProps;
    if(DIRPROP_FLAG(dirProps[lastCcPos])&MASK_ISO)  /* after an isolate */
        return;
    if(NO_OVERRIDE(embeddingLevel)>NO_OVERRIDE(contextLevel))   /* not a PDF */
        contextLevel=embeddingLevel;
    pLastIsoRun->limit=pLastIsoRun->start;
    pLastIsoRun->level=embeddingLevel;
    pLastIsoRun->lastStrong=pLastIsoRun->lastBase=contextLevel&1;
    pLastIsoRun->contextDir=(UBiDiDirection)(contextLevel&1);
    pLastIsoRun->contextPos=(UBiDiDirection)lastCcPos;
}

/* LRI or RLI */
static void
bracketProcessLRI_RLI(BracketData *bd, UBiDiLevel level) {
    IsoRun *pLastIsoRun=&bd->isoRuns[bd->isoRunLast];
    int16_t lastLimit;
    pLastIsoRun->lastBase=ON;
    lastLimit=pLastIsoRun->limit;
    bd->isoRunLast++;
    pLastIsoRun++;
    pLastIsoRun->start=pLastIsoRun->limit=lastLimit;
    pLastIsoRun->level=level;
    pLastIsoRun->lastStrong=pLastIsoRun->lastBase=level&1;
    pLastIsoRun->contextDir=(UBiDiDirection)(level&1);
    pLastIsoRun->contextPos=0;
}

/* PDI */
static void
bracketProcessPDI(BracketData *bd) {
    IsoRun *pLastIsoRun;
    bd->isoRunLast--;
    pLastIsoRun=&bd->isoRuns[bd->isoRunLast];
    pLastIsoRun->lastBase=ON;
}

/* newly found opening bracket: create an openings entry */
static UBool                            /* return TRUE if success */
bracketAddOpening(BracketData *bd, UChar match, int32_t position) {
    IsoRun *pLastIsoRun=&bd->isoRuns[bd->isoRunLast];
    Opening *pOpening;
    if(pLastIsoRun->limit>=bd->openingsCount) {  /* no available new entry */
        UBiDi *pBiDi=bd->pBiDi;
        if(!getInitialOpeningsMemory(pBiDi, pLastIsoRun->limit * 2))
            return FALSE;
        if(bd->openings==bd->simpleOpenings)
            uprv_memcpy(pBiDi->openingsMemory, bd->simpleOpenings,
                        SIMPLE_OPENINGS_COUNT * sizeof(Opening));
        bd->openings=pBiDi->openingsMemory;     /* may have changed */
        bd->openingsCount=pBiDi->openingsSize / sizeof(Opening);
    }
    pOpening=&bd->openings[pLastIsoRun->limit];
    pOpening->position=position;
    pOpening->match=match;
    pOpening->contextDir=pLastIsoRun->contextDir;
    pOpening->contextPos=pLastIsoRun->contextPos;
    pOpening->flags=0;
    pLastIsoRun->limit++;
    return TRUE;
}

/* change N0c1 to N0c2 when a preceding bracket is assigned the embedding level */
static void
fixN0c(BracketData *bd, int32_t openingIndex, int32_t newPropPosition, DirProp newProp) {
    /* This function calls itself recursively */
    IsoRun *pLastIsoRun=&bd->isoRuns[bd->isoRunLast];
    Opening *qOpening;
    DirProp *dirProps=bd->pBiDi->dirProps;
    int32_t k, openingPosition, closingPosition;
    for(k=openingIndex+1, qOpening=&bd->openings[k]; k<pLastIsoRun->limit; k++, qOpening++) {
        if(qOpening->match>=0)      /* not an N0c match */
            continue;
        if(newPropPosition<qOpening->contextPos)
            break;
        if(newPropPosition>=qOpening->position)
            continue;
        if(newProp==qOpening->contextDir)
            break;
        openingPosition=qOpening->position;
        dirProps[openingPosition]=newProp;
        closingPosition=-(qOpening->match);
        dirProps[closingPosition]=newProp;
        qOpening->match=0;                      /* prevent further changes */
        fixN0c(bd, k, openingPosition, newProp);
        fixN0c(bd, k, closingPosition, newProp);
    }
}

/* process closing bracket */
static DirProp              /* return L or R if N0b or N0c, ON if N0d */
bracketProcessClosing(BracketData *bd, int32_t openIdx, int32_t position) {
    IsoRun *pLastIsoRun=&bd->isoRuns[bd->isoRunLast];
    Opening *pOpening, *qOpening;
    UBiDiDirection direction;
    UBool stable;
    DirProp newProp;
    pOpening=&bd->openings[openIdx];
    direction=(UBiDiDirection)(pLastIsoRun->level&1);
    stable=TRUE;            /* assume stable until proved otherwise */

    /* The stable flag is set when brackets are paired and their
       level is resolved and cannot be changed by what will be
       found later in the source string.
       An unstable match can occur only when applying N0c, where
       the resolved level depends on the preceding context, and
       this context may be affected by text occurring later.
       Example: RTL paragraph containing:  abc[(latin) HEBREW]
       When the closing parenthesis is encountered, it appears
       that N0c1 must be applied since 'abc' sets an opposite
       direction context and both parentheses receive level 2.
       However, when the closing square bracket is processed,
       N0b applies because of 'HEBREW' being included within the
       brackets, thus the square brackets are treated like R and
       receive level 1. However, this changes the preceding
       context of the opening parenthesis, and it now appears
       that N0c2 must be applied to the parentheses rather than
       N0c1. */

    if((direction==0 && pOpening->flags&FOUND_L) ||
       (direction==1 && pOpening->flags&FOUND_R)) {                         /* N0b */
        newProp=static_cast<DirProp>(direction);
    }
    else if(pOpening->flags&(FOUND_L|FOUND_R)) {                            /* N0c */
        /* it is stable if there is no containing pair or in
           conditions too complicated and not worth checking */
        stable=(openIdx==pLastIsoRun->start);
        if(direction!=pOpening->contextDir)
            newProp= static_cast<DirProp>(pOpening->contextDir);           /* N0c1 */
        else
            newProp= static_cast<DirProp>(direction);                      /* N0c2 */
    } else {
        /* forget this and any brackets nested within this pair */
        pLastIsoRun->limit= static_cast<uint16_t>(openIdx);
        return ON;                                                          /* N0d */
    }
    bd->pBiDi->dirProps[pOpening->position]=newProp;
    bd->pBiDi->dirProps[position]=newProp;
    /* Update nested N0c pairs that may be affected */
    fixN0c(bd, openIdx, pOpening->position, newProp);
    if(stable) {
        pLastIsoRun->limit= static_cast<uint16_t>(openIdx); /* forget any brackets nested within this pair */
        /* remove lower located synonyms if any */
        while(pLastIsoRun->limit>pLastIsoRun->start &&
              bd->openings[pLastIsoRun->limit-1].position==pOpening->position)
            pLastIsoRun->limit--;
    } else {
        int32_t k;
        pOpening->match=-position;
        /* neutralize lower located synonyms if any */
        k=openIdx-1;
        while(k>=pLastIsoRun->start &&
              bd->openings[k].position==pOpening->position)
            bd->openings[k--].match=0;
        /* neutralize any unmatched opening between the current pair;
           this will also neutralize higher located synonyms if any */
        for(k=openIdx+1; k<pLastIsoRun->limit; k++) {
            qOpening=&bd->openings[k];
            if(qOpening->position>=position)
                break;
            if(qOpening->match>0)
                qOpening->match=0;
        }
    }
    return newProp;
}

/* handle strong characters, digits and candidates for closing brackets */
static UBool                            /* return TRUE if success */
bracketProcessChar(BracketData *bd, int32_t position) {
    IsoRun *pLastIsoRun=&bd->isoRuns[bd->isoRunLast];
    DirProp *dirProps, dirProp, newProp;
    UBiDiLevel level;
    dirProps=bd->pBiDi->dirProps;
    dirProp=dirProps[position];
    if(dirProp==ON) {
        UChar c, match;
        int32_t idx;
        /* First see if it is a matching closing bracket. Hopefully, this is
           more efficient than checking if it is a closing bracket at all */
        c=bd->pBiDi->text[position];
        for(idx=pLastIsoRun->limit-1; idx>=pLastIsoRun->start; idx--) {
            if(bd->openings[idx].match!=c)
                continue;
            /* We have a match */
            newProp=bracketProcessClosing(bd, idx, position);
            if(newProp==ON) {           /* N0d */
                c=0;        /* prevent handling as an opening */
                break;
            }
            pLastIsoRun->lastBase=ON;
            pLastIsoRun->contextDir=(UBiDiDirection)newProp;
            pLastIsoRun->contextPos=position;
            level=bd->pBiDi->levels[position];
            if(level&UBIDI_LEVEL_OVERRIDE) {    /* X4, X5 */
                uint16_t flag;
                int32_t i;
                newProp=level&1;
                pLastIsoRun->lastStrong=newProp;
                flag=DIRPROP_FLAG(newProp);
                for(i=pLastIsoRun->start; i<idx; i++)
                    bd->openings[i].flags|=flag;
                /* matching brackets are not overridden by LRO/RLO */
                bd->pBiDi->levels[position]&=~UBIDI_LEVEL_OVERRIDE;
            }
            /* matching brackets are not overridden by LRO/RLO */
            bd->pBiDi->levels[bd->openings[idx].position]&=~UBIDI_LEVEL_OVERRIDE;
            return TRUE;
        }
        /* We get here only if the ON character is not a matching closing
           bracket or it is a case of N0d */
        /* Now see if it is an opening bracket */
        if(c)
            match= static_cast<UChar>(u_getBidiPairedBracket(c));    /* get the matching char */
        else
            match=0;
        if(match!=c &&                  /* has a matching char */
           ubidi_getPairedBracketType(c)==U_BPT_OPEN) { /* opening bracket */
            /* special case: process synonyms
               create an opening entry for each synonym */
            if(match==0x232A) {     /* RIGHT-POINTING ANGLE BRACKET */
                if(!bracketAddOpening(bd, 0x3009, position))
                    return FALSE;
            }
            else if(match==0x3009) {         /* RIGHT ANGLE BRACKET */
                if(!bracketAddOpening(bd, 0x232A, position))
                    return FALSE;
            }
            if(!bracketAddOpening(bd, match, position))
                return FALSE;
        }
    }
    level=bd->pBiDi->levels[position];
    if(level&UBIDI_LEVEL_OVERRIDE) {    /* X4, X5 */
        newProp=level&1;
        if(dirProp!=S && dirProp!=WS && dirProp!=ON)
            dirProps[position]=newProp;
        pLastIsoRun->lastBase=newProp;
        pLastIsoRun->lastStrong=newProp;
        pLastIsoRun->contextDir=(UBiDiDirection)newProp;
        pLastIsoRun->contextPos=position;
    }
    else if(dirProp<=R || dirProp==AL) {
        newProp= static_cast<DirProp>(DIR_FROM_STRONG(dirProp));
        pLastIsoRun->lastBase=dirProp;
        pLastIsoRun->lastStrong=dirProp;
        pLastIsoRun->contextDir=(UBiDiDirection)newProp;
        pLastIsoRun->contextPos=position;
    }
    else if(dirProp==EN) {
        pLastIsoRun->lastBase=EN;
        if(pLastIsoRun->lastStrong==L) {
            newProp=L;                  /* W7 */
            if(!bd->isNumbersSpecial)
                dirProps[position]=ENL;
            pLastIsoRun->contextDir=(UBiDiDirection)L;
            pLastIsoRun->contextPos=position;
        }
        else {
            newProp=R;                  /* N0 */
            if(pLastIsoRun->lastStrong==AL)
                dirProps[position]=AN;  /* W2 */
            else
                dirProps[position]=ENR;
            pLastIsoRun->contextDir=(UBiDiDirection)R;
            pLastIsoRun->contextPos=position;
        }
    }
    else if(dirProp==AN) {
        newProp=R;                      /* N0 */
        pLastIsoRun->lastBase=AN;
        pLastIsoRun->contextDir=(UBiDiDirection)R;
        pLastIsoRun->contextPos=position;
    }
    else if(dirProp==NSM) {
        /* if the last real char was ON, change NSM to ON so that it
           will stay ON even if the last real char is a bracket which
           may be changed to L or R */
        newProp=pLastIsoRun->lastBase;
        if(newProp==ON)
            dirProps[position]=newProp;
    }
    else {
        newProp=dirProp;
        pLastIsoRun->lastBase=dirProp;
    }
    if(newProp<=R || newProp==AL) {
        int32_t i;
        uint16_t flag=DIRPROP_FLAG(DIR_FROM_STRONG(newProp));
        for(i=pLastIsoRun->start; i<pLastIsoRun->limit; i++)
            if(position>bd->openings[i].position)
                bd->openings[i].flags|=flag;
    }
    return TRUE;
}

/* perform (X1)..(X9) ------------------------------------------------------- */

/* determine if the text is mixed-directional or single-directional */
static UBiDiDirection
directionFromFlags(UBiDi *pBiDi) {
    Flags flags=pBiDi->flags;
    /* if the text contains AN and neutrals, then some neutrals may become RTL */
    if(!(flags&MASK_RTL || ((flags&DIRPROP_FLAG(AN)) && (flags&MASK_POSSIBLE_N)))) {
        return UBIDI_LTR;
    } else if(!(flags&MASK_LTR)) {
        return UBIDI_RTL;
    } else {
        return UBIDI_MIXED;
    }
}

/*
 * Resolve the explicit levels as specified by explicit embedding codes.
 * Recalculate the flags to have them reflect the real properties
 * after taking the explicit embeddings into account.
 *
 * The BiDi algorithm is designed to result in the same behavior whether embedding
 * levels are externally specified (from "styled text", supposedly the preferred
 * method) or set by explicit embedding codes (LRx, RLx, PDF, FSI, PDI) in the plain text.
 * That is why (X9) instructs to remove all not-isolate explicit codes (and BN).
 * However, in a real implementation, the removal of these codes and their index
 * positions in the plain text is undesirable since it would result in
 * reallocated, reindexed text.
 * Instead, this implementation leaves the codes in there and just ignores them
 * in the subsequent processing.
 * In order to get the same reordering behavior, positions with a BN or a not-isolate
 * explicit embedding code just get the same level assigned as the last "real"
 * character.
 *
 * Some implementations, not this one, then overwrite some of these
 * directionality properties at "real" same-level-run boundaries by
 * L or R codes so that the resolution of weak types can be performed on the
 * entire paragraph at once instead of having to parse it once more and
 * perform that resolution on same-level-runs.
 * This limits the scope of the implicit rules in effectively
 * the same way as the run limits.
 *
 * Instead, this implementation does not modify these codes, except for
 * paired brackets whose properties (ON) may be replaced by L or R.
 * On one hand, the paragraph has to be scanned for same-level-runs, but
 * on the other hand, this saves another loop to reset these codes,
 * or saves making and modifying a copy of dirProps[].
 *
 *
 * Note that (Pn) and (Xn) changed significantly from version 4 of the BiDi algorithm.
 *
 *
 * Handling the stack of explicit levels (Xn):
 *
 * With the BiDi stack of explicit levels, as pushed with each
 * LRE, RLE, LRO, RLO, LRI, RLI and FSI and popped with each PDF and PDI,
 * the explicit level must never exceed UBIDI_MAX_EXPLICIT_LEVEL.
 *
 * In order to have a correct push-pop semantics even in the case of overflows,
 * overflow counters and a valid isolate counter are used as described in UAX#9
 * section 3.3.2 "Explicit Levels and Directions".
 *
 * This implementation assumes that UBIDI_MAX_EXPLICIT_LEVEL is odd.
 *
 * Returns normally the direction; -1 if there was a memory shortage
 *
 */
static UBiDiDirection
resolveExplicitLevels(UBiDi *pBiDi, UErrorCode *pErrorCode) {
    DirProp *dirProps=pBiDi->dirProps;
    UBiDiLevel *levels=pBiDi->levels;
    const UChar *text=pBiDi->text;

    int32_t i=0, length=pBiDi->length;
    Flags flags=pBiDi->flags;       /* collect all directionalities in the text */
    DirProp dirProp;
    UBiDiLevel level=GET_PARALEVEL(pBiDi, 0);
    UBiDiDirection direction;
    pBiDi->isolateCount=0;

    if(U_FAILURE(*pErrorCode)) { return UBIDI_LTR; }

    /* determine if the text is mixed-directional or single-directional */
    direction=directionFromFlags(pBiDi);

    /* we may not need to resolve any explicit levels */
    if((direction!=UBIDI_MIXED)) {
        /* not mixed directionality: levels don't matter - trailingWSStart will be 0 */
        return direction;
    }
    if(pBiDi->reorderingMode > UBIDI_REORDER_LAST_LOGICAL_TO_VISUAL) {
        /* inverse BiDi: mixed, but all characters are at the same embedding level */
        /* set all levels to the paragraph level */
        int32_t paraIndex, start, limit;
        for(paraIndex=0; paraIndex<pBiDi->paraCount; paraIndex++) {
            if(paraIndex==0)
                start=0;
            else
                start=pBiDi->paras[paraIndex-1].limit;
            limit=pBiDi->paras[paraIndex].limit;
            level= static_cast<UBiDiLevel>(pBiDi->paras[paraIndex].level);
            for(i=start; i<limit; i++)
                levels[i]=level;
        }
        return direction;   /* no bracket matching for inverse BiDi */
    }
    if(!(flags&(MASK_EXPLICIT|MASK_ISO))) {
        /* no embeddings, set all levels to the paragraph level */
        /* we still have to perform bracket matching */
        int32_t paraIndex, start, limit;
        BracketData bracketData;
        bracketInit(pBiDi, &bracketData);
        for(paraIndex=0; paraIndex<pBiDi->paraCount; paraIndex++) {
            if(paraIndex==0)
                start=0;
            else
                start=pBiDi->paras[paraIndex-1].limit;
            limit=pBiDi->paras[paraIndex].limit;
            level= static_cast<UBiDiLevel>(pBiDi->paras[paraIndex].level);
            for(i=start; i<limit; i++) {
                levels[i]=level;
                dirProp=dirProps[i];
                if(dirProp==BN)
                    continue;
                if(dirProp==B) {
                    if((i+1)<length) {
                        if(text[i]==UBIDIIMP_CR && text[i+1]==UBIDIIMP_LF)
                            continue;   /* skip CR when followed by LF */
                        bracketProcessB(&bracketData, level);
                    }
                    continue;
                }
                if(!bracketProcessChar(&bracketData, i)) {
                    *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
                    return UBIDI_LTR;
                }
            }
        }
        return direction;
    }
    {
        /* continue to perform (Xn) */

        /* (X1) level is set for all codes, embeddingLevel keeps track of the push/pop operations */
        /* both variables may carry the UBIDI_LEVEL_OVERRIDE flag to indicate the override status */
        UBiDiLevel embeddingLevel=level, newLevel;
        UBiDiLevel previousLevel=level;     /* previous level for regular (not CC) characters */
        int32_t lastCcPos=0;                /* index of last effective LRx,RLx, PDx */

        /* The following stack remembers the embedding level and the ISOLATE flag of level runs.
           stackLast points to its current entry. */
        uint16_t stack[UBIDI_MAX_EXPLICIT_LEVEL+2];   /* we never push anything >=UBIDI_MAX_EXPLICIT_LEVEL
                                                        but we need one more entry as base */
        uint32_t stackLast=0;
        int32_t overflowIsolateCount=0;
        int32_t overflowEmbeddingCount=0;
        int32_t validIsolateCount=0;
        BracketData bracketData;
        bracketInit(pBiDi, &bracketData);
        stack[0]=level;     /* initialize base entry to para level, no override, no isolate */

        /* recalculate the flags */
        flags=0;

        for(i=0; i<length; ++i) {
            dirProp=dirProps[i];
            switch(dirProp) {
            case LRE:
            case RLE:
            case LRO:
            case RLO:
                /* (X2, X3, X4, X5) */
                flags|=DIRPROP_FLAG(BN);
                levels[i]=previousLevel;
                if (dirProp==LRE || dirProp==LRO)
                    /* least greater even level */
                    newLevel=(UBiDiLevel)((embeddingLevel+2)&~(UBIDI_LEVEL_OVERRIDE|1));
                else
                    /* least greater odd level */
                    newLevel=(UBiDiLevel)((NO_OVERRIDE(embeddingLevel)+1)|1);
                if(newLevel<=UBIDI_MAX_EXPLICIT_LEVEL && overflowIsolateCount==0 &&
                                                         overflowEmbeddingCount==0) {
                    lastCcPos=i;
                    embeddingLevel=newLevel;
                    if(dirProp==LRO || dirProp==RLO)
                        embeddingLevel|=UBIDI_LEVEL_OVERRIDE;
                    stackLast++;
                    stack[stackLast]=embeddingLevel;
                    /* we don't need to set UBIDI_LEVEL_OVERRIDE off for LRE and RLE
                       since this has already been done for newLevel which is
                       the source for embeddingLevel.
                     */
                } else {
                    if(overflowIsolateCount==0)
                        overflowEmbeddingCount++;
                }
                break;
            case PDF:
                /* (X7) */
                flags|=DIRPROP_FLAG(BN);
                levels[i]=previousLevel;
                /* handle all the overflow cases first */
                if(overflowIsolateCount) {
                    break;
                }
                if(overflowEmbeddingCount) {
                    overflowEmbeddingCount--;
                    break;
                }
                if(stackLast>0 && stack[stackLast]<ISOLATE) {   /* not an isolate entry */
                    lastCcPos=i;
                    stackLast--;
                    embeddingLevel=(UBiDiLevel)stack[stackLast];
                }
                break;
            case LRI:
            case RLI:
                flags|=(DIRPROP_FLAG(ON)|DIRPROP_FLAG_LR(embeddingLevel));
                levels[i]=NO_OVERRIDE(embeddingLevel);
                if(NO_OVERRIDE(embeddingLevel)!=NO_OVERRIDE(previousLevel)) {
                    bracketProcessBoundary(&bracketData, lastCcPos,
                                           previousLevel, embeddingLevel);
                    flags|=DIRPROP_FLAG_MULTI_RUNS;
                }
                previousLevel=embeddingLevel;
                /* (X5a, X5b) */
                if(dirProp==LRI)
                    /* least greater even level */
                    newLevel=(UBiDiLevel)((embeddingLevel+2)&~(UBIDI_LEVEL_OVERRIDE|1));
                else
                    /* least greater odd level */
                    newLevel=(UBiDiLevel)((NO_OVERRIDE(embeddingLevel)+1)|1);
                if(newLevel<=UBIDI_MAX_EXPLICIT_LEVEL && overflowIsolateCount==0 &&
                                                         overflowEmbeddingCount==0) {
                    flags|=DIRPROP_FLAG(dirProp);
                    lastCcPos=i;
                    validIsolateCount++;
                    if(validIsolateCount>pBiDi->isolateCount)
                        pBiDi->isolateCount=validIsolateCount;
                    embeddingLevel=newLevel;
                    /* we can increment stackLast without checking because newLevel
                       will exceed UBIDI_MAX_EXPLICIT_LEVEL before stackLast overflows */
                    stackLast++;
                    stack[stackLast]=embeddingLevel+ISOLATE;
                    bracketProcessLRI_RLI(&bracketData, embeddingLevel);
                } else {
                    /* make it WS so that it is handled by adjustWSLevels() */
                    dirProps[i]=WS;
                    overflowIsolateCount++;
                }
                break;
            case PDI:
                if(NO_OVERRIDE(embeddingLevel)!=NO_OVERRIDE(previousLevel)) {
                    bracketProcessBoundary(&bracketData, lastCcPos,
                                           previousLevel, embeddingLevel);
                    flags|=DIRPROP_FLAG_MULTI_RUNS;
                }
                /* (X6a) */
                if(overflowIsolateCount) {
                    overflowIsolateCount--;
                    /* make it WS so that it is handled by adjustWSLevels() */
                    dirProps[i]=WS;
                }
                else if(validIsolateCount) {
                    flags|=DIRPROP_FLAG(PDI);
                    lastCcPos=i;
                    overflowEmbeddingCount=0;
                    while(stack[stackLast]<ISOLATE) /* pop embedding entries */
                        stackLast--;                /* until the last isolate entry */
                    stackLast--;                    /* pop also the last isolate entry */
                    validIsolateCount--;
                    bracketProcessPDI(&bracketData);
                } else
                    /* make it WS so that it is handled by adjustWSLevels() */
                    dirProps[i]=WS;
                embeddingLevel=(UBiDiLevel)stack[stackLast]&~ISOLATE;
                flags|=(DIRPROP_FLAG(ON)|DIRPROP_FLAG_LR(embeddingLevel));
                previousLevel=embeddingLevel;
                levels[i]=NO_OVERRIDE(embeddingLevel);
                break;
            case B:
                flags|=DIRPROP_FLAG(B);
                levels[i]=GET_PARALEVEL(pBiDi, i);
                if((i+1)<length) {
                    if(text[i]==UBIDIIMP_CR && text[i+1]==UBIDIIMP_LF)
                        break;          /* skip CR when followed by LF */
                    overflowEmbeddingCount=overflowIsolateCount=0;
                    validIsolateCount=0;
                    stackLast=0;
                    previousLevel=embeddingLevel=GET_PARALEVEL(pBiDi, i+1);
                    stack[0]=embeddingLevel; /* initialize base entry to para level, no override, no isolate */
                    bracketProcessB(&bracketData, embeddingLevel);
                }
                break;
            case BN:
                /* BN, LRE, RLE, and PDF are supposed to be removed (X9) */
                /* they will get their levels set correctly in adjustWSLevels() */
                levels[i]=previousLevel;
                flags|=DIRPROP_FLAG(BN);
                break;
            default:
                /* all other types are normal characters and get the "real" level */
                if(NO_OVERRIDE(embeddingLevel)!=NO_OVERRIDE(previousLevel)) {
                    bracketProcessBoundary(&bracketData, lastCcPos,
                                           previousLevel, embeddingLevel);
                    flags|=DIRPROP_FLAG_MULTI_RUNS;
                    if(embeddingLevel&UBIDI_LEVEL_OVERRIDE)
                        flags|=DIRPROP_FLAG_O(embeddingLevel);
                    else
                        flags|=DIRPROP_FLAG_E(embeddingLevel);
                }
                previousLevel=embeddingLevel;
                levels[i]=embeddingLevel;
                if(!bracketProcessChar(&bracketData, i))
                    return (UBiDiDirection)-1;
                /* the dirProp may have been changed in bracketProcessChar() */
                flags|=DIRPROP_FLAG(dirProps[i]);
                break;
            }
        }
        if(flags&MASK_EMBEDDING)
            flags|=DIRPROP_FLAG_LR(pBiDi->paraLevel);
        if(pBiDi->orderParagraphsLTR && (flags&DIRPROP_FLAG(B)))
            flags|=DIRPROP_FLAG(L);
        /* again, determine if the text is mixed-directional or single-directional */
        pBiDi->flags=flags;
        direction=directionFromFlags(pBiDi);
    }
    return direction;
}

/*
 * Use a pre-specified embedding levels array:
 *
 * Adjust the directional properties for overrides (->LEVEL_OVERRIDE),
 * ignore all explicit codes (X9),
 * and check all the preset levels.
 *
 * Recalculate the flags to have them reflect the real properties
 * after taking the explicit embeddings into account.
 */
static UBiDiDirection
checkExplicitLevels(UBiDi *pBiDi, UErrorCode *pErrorCode) {
    DirProp *dirProps=pBiDi->dirProps;
    UBiDiLevel *levels=pBiDi->levels;
    int32_t isolateCount=0;

    int32_t length=pBiDi->length;
    Flags flags=0;  /* collect all directionalities in the text */
    pBiDi->isolateCount=0;

    int32_t currentParaIndex = 0;
    int32_t currentParaLimit = pBiDi->paras[0].limit;
    int32_t currentParaLevel = pBiDi->paraLevel;

    for(int32_t i=0; i<length; ++i) {
        UBiDiLevel level=levels[i];
        DirProp dirProp=dirProps[i];
        if(dirProp==LRI || dirProp==RLI) {
            isolateCount++;
            if(isolateCount>pBiDi->isolateCount)
                pBiDi->isolateCount=isolateCount;
        }
        else if(dirProp==PDI)
            isolateCount--;
        else if(dirProp==B)
            isolateCount=0;

        // optimized version of  int32_t currentParaLevel = GET_PARALEVEL(pBiDi, i);
        if (pBiDi->defaultParaLevel != 0 &&
                i == currentParaLimit && (currentParaIndex + 1) < pBiDi->paraCount) {
            currentParaLevel = pBiDi->paras[++currentParaIndex].level;
            currentParaLimit = pBiDi->paras[currentParaIndex].limit;
        }

        UBiDiLevel overrideFlag = level & UBIDI_LEVEL_OVERRIDE;
        level &= ~UBIDI_LEVEL_OVERRIDE;
        if (level < currentParaLevel || UBIDI_MAX_EXPLICIT_LEVEL < level) {
            if (level == 0) {
                if (dirProp == B) {
                    // Paragraph separators are ok with explicit level 0.
                    // Prevents reordering of paragraphs.
                } else {
                    // Treat explicit level 0 as a wildcard for the paragraph level.
                    // Avoid making the caller guess what the paragraph level would be.
                    level = (UBiDiLevel)currentParaLevel;
                    levels[i] = level | overrideFlag;
                }
            } else {
                // 1 <= level < currentParaLevel or UBIDI_MAX_EXPLICIT_LEVEL < level
                /* level out of bounds */
                *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
                return UBIDI_LTR;
            }
        }
        if (overrideFlag != 0) {
            /* keep the override flag in levels[i] but adjust the flags */
            flags|=DIRPROP_FLAG_O(level);
        } else {
            /* set the flags */
            flags|=DIRPROP_FLAG_E(level)|DIRPROP_FLAG(dirProp);
        }
    }
    if(flags&MASK_EMBEDDING)
        flags|=DIRPROP_FLAG_LR(pBiDi->paraLevel);
    /* determine if the text is mixed-directional or single-directional */
    pBiDi->flags=flags;
    return directionFromFlags(pBiDi);
}

/******************************************************************
 The Properties state machine table
*******************************************************************

 All table cells are 8 bits:
      bits 0..4:  next state
      bits 5..7:  action to perform (if > 0)

 Cells may be of format "n" where n represents the next state
 (except for the rightmost column).
 Cells may also be of format "s(x,y)" where x represents an action
 to perform and y represents the next state.

*******************************************************************
 Definitions and type for properties state table
*******************************************************************
*/
#define IMPTABPROPS_COLUMNS 16
#define IMPTABPROPS_RES (IMPTABPROPS_COLUMNS - 1)
#define GET_STATEPROPS(cell) ((cell)&0x1f)
#define GET_ACTIONPROPS(cell) ((cell)>>5)
#define s(action, newState) ((uint8_t)(newState+(action<<5)))

static const uint8_t groupProp[] =          /* dirProp regrouped */
{
/*  L   R   EN  ES  ET  AN  CS  B   S   WS  ON  LRE LRO AL  RLE RLO PDF NSM BN  FSI LRI RLI PDI ENL ENR */
    0,  1,  2,  7,  8,  3,  9,  6,  5,  4,  4,  10, 10, 12, 10, 10, 10, 11, 10, 4,  4,  4,  4,  13, 14
};
enum { DirProp_L=0, DirProp_R=1, DirProp_EN=2, DirProp_AN=3, DirProp_ON=4, DirProp_S=5, DirProp_B=6 }; /* reduced dirProp */

/******************************************************************

      PROPERTIES  STATE  TABLE

 In table impTabProps,
      - the ON column regroups ON and WS, FSI, RLI, LRI and PDI
      - the BN column regroups BN, LRE, RLE, LRO, RLO, PDF
      - the Res column is the reduced property assigned to a run

 Action 1: process current run1, init new run1
        2: init new run2
        3: process run1, process run2, init new run1
        4: process run1, set run1=run2, init new run2

 Notes:
  1) This table is used in resolveImplicitLevels().
  2) This table triggers actions when there is a change in the Bidi
     property of incoming characters (action 1).
  3) Most such property sequences are processed immediately (in
     fact, passed to processPropertySeq().
  4) However, numbers are assembled as one sequence. This means
     that undefined situations (like CS following digits, until
     it is known if the next char will be a digit) are held until
     following chars define them.
     Example: digits followed by CS, then comes another CS or ON;
              the digits will be processed, then the CS assigned
              as the start of an ON sequence (action 3).
  5) There are cases where more than one sequence must be
     processed, for instance digits followed by CS followed by L:
     the digits must be processed as one sequence, and the CS
     must be processed as an ON sequence, all this before starting
     assembling chars for the opening L sequence.


*/
static const uint8_t impTabProps[][IMPTABPROPS_COLUMNS] =
{
/*                        L ,     R ,    EN ,    AN ,    ON ,     S ,     B ,    ES ,    ET ,    CS ,    BN ,   NSM ,    AL ,   ENL ,   ENR , Res */
/* 0 Init        */ {     1 ,     2 ,     4 ,     5 ,     7 ,    15 ,    17 ,     7 ,     9 ,     7 ,     0 ,     7 ,     3 ,    18 ,    21 , DirProp_ON },
/* 1 L           */ {     1 , s(1,2), s(1,4), s(1,5), s(1,7),s(1,15),s(1,17), s(1,7), s(1,9), s(1,7),     1 ,     1 , s(1,3),s(1,18),s(1,21),  DirProp_L },
/* 2 R           */ { s(1,1),     2 , s(1,4), s(1,5), s(1,7),s(1,15),s(1,17), s(1,7), s(1,9), s(1,7),     2 ,     2 , s(1,3),s(1,18),s(1,21),  DirProp_R },
/* 3 AL          */ { s(1,1), s(1,2), s(1,6), s(1,6), s(1,8),s(1,16),s(1,17), s(1,8), s(1,8), s(1,8),     3 ,     3 ,     3 ,s(1,18),s(1,21),  DirProp_R },
/* 4 EN          */ { s(1,1), s(1,2),     4 , s(1,5), s(1,7),s(1,15),s(1,17),s(2,10),    11 ,s(2,10),     4 ,     4 , s(1,3),    18 ,    21 , DirProp_EN },
/* 5 AN          */ { s(1,1), s(1,2), s(1,4),     5 , s(1,7),s(1,15),s(1,17), s(1,7), s(1,9),s(2,12),     5 ,     5 , s(1,3),s(1,18),s(1,21), DirProp_AN },
/* 6 AL:EN/AN    */ { s(1,1), s(1,2),     6 ,     6 , s(1,8),s(1,16),s(1,17), s(1,8), s(1,8),s(2,13),     6 ,     6 , s(1,3),    18 ,    21 , DirProp_AN },
/* 7 ON          */ { s(1,1), s(1,2), s(1,4), s(1,5),     7 ,s(1,15),s(1,17),     7 ,s(2,14),     7 ,     7 ,     7 , s(1,3),s(1,18),s(1,21), DirProp_ON },
/* 8 AL:ON       */ { s(1,1), s(1,2), s(1,6), s(1,6),     8 ,s(1,16),s(1,17),     8 ,     8 ,     8 ,     8 ,     8 , s(1,3),s(1,18),s(1,21), DirProp_ON },
/* 9 ET          */ { s(1,1), s(1,2),     4 , s(1,5),     7 ,s(1,15),s(1,17),     7 ,     9 ,     7 ,     9 ,     9 , s(1,3),    18 ,    21 , DirProp_ON },
/*10 EN+ES/CS    */ { s(3,1), s(3,2),     4 , s(3,5), s(4,7),s(3,15),s(3,17), s(4,7),s(4,14), s(4,7),    10 , s(4,7), s(3,3),    18 ,    21 , DirProp_EN },
/*11 EN+ET       */ { s(1,1), s(1,2),     4 , s(1,5), s(1,7),s(1,15),s(1,17), s(1,7),    11 , s(1,7),    11 ,    11 , s(1,3),    18 ,    21 , DirProp_EN },
/*12 AN+CS       */ { s(3,1), s(3,2), s(3,4),     5 , s(4,7),s(3,15),s(3,17), s(4,7),s(4,14), s(4,7),    12 , s(4,7), s(3,3),s(3,18),s(3,21), DirProp_AN },
/*13 AL:EN/AN+CS */ { s(3,1), s(3,2),     6 ,     6 , s(4,8),s(3,16),s(3,17), s(4,8), s(4,8), s(4,8),    13 , s(4,8), s(3,3),    18 ,    21 , DirProp_AN },
/*14 ON+ET       */ { s(1,1), s(1,2), s(4,4), s(1,5),     7 ,s(1,15),s(1,17),     7 ,    14 ,     7 ,    14 ,    14 , s(1,3),s(4,18),s(4,21), DirProp_ON },
/*15 S           */ { s(1,1), s(1,2), s(1,4), s(1,5), s(1,7),    15 ,s(1,17), s(1,7), s(1,9), s(1,7),    15 , s(1,7), s(1,3),s(1,18),s(1,21),  DirProp_S },
/*16 AL:S        */ { s(1,1), s(1,2), s(1,6), s(1,6), s(1,8),    16 ,s(1,17), s(1,8), s(1,8), s(1,8),    16 , s(1,8), s(1,3),s(1,18),s(1,21),  DirProp_S },
/*17 B           */ { s(1,1), s(1,2), s(1,4), s(1,5), s(1,7),s(1,15),    17 , s(1,7), s(1,9), s(1,7),    17 , s(1,7), s(1,3),s(1,18),s(1,21),  DirProp_B },
/*18 ENL         */ { s(1,1), s(1,2),    18 , s(1,5), s(1,7),s(1,15),s(1,17),s(2,19),    20 ,s(2,19),    18 ,    18 , s(1,3),    18 ,    21 ,  DirProp_L },
/*19 ENL+ES/CS   */ { s(3,1), s(3,2),    18 , s(3,5), s(4,7),s(3,15),s(3,17), s(4,7),s(4,14), s(4,7),    19 , s(4,7), s(3,3),    18 ,    21 ,  DirProp_L },
/*20 ENL+ET      */ { s(1,1), s(1,2),    18 , s(1,5), s(1,7),s(1,15),s(1,17), s(1,7),    20 , s(1,7),    20 ,    20 , s(1,3),    18 ,    21 ,  DirProp_L },
/*21 ENR         */ { s(1,1), s(1,2),    21 , s(1,5), s(1,7),s(1,15),s(1,17),s(2,22),    23 ,s(2,22),    21 ,    21 , s(1,3),    18 ,    21 , DirProp_AN },
/*22 ENR+ES/CS   */ { s(3,1), s(3,2),    21 , s(3,5), s(4,7),s(3,15),s(3,17), s(4,7),s(4,14), s(4,7),    22 , s(4,7), s(3,3),    18 ,    21 , DirProp_AN },
/*23 ENR+ET      */ { s(1,1), s(1,2),    21 , s(1,5), s(1,7),s(1,15),s(1,17), s(1,7),    23 , s(1,7),    23 ,    23 , s(1,3),    18 ,    21 , DirProp_AN }
};

/*  we must undef macro s because the levels tables have a different
 *  structure (4 bits for action and 4 bits for next state.
 */
#undef s

/******************************************************************
 The levels state machine tables
*******************************************************************

 All table cells are 8 bits:
      bits 0..3:  next state
      bits 4..7:  action to perform (if > 0)

 Cells may be of format "n" where n represents the next state
 (except for the rightmost column).
 Cells may also be of format "s(x,y)" where x represents an action
 to perform and y represents the next state.

 This format limits each table to 16 states each and to 15 actions.

*******************************************************************
 Definitions and type for levels state tables
*******************************************************************
*/
#define IMPTABLEVELS_COLUMNS (DirProp_B + 2)
#define IMPTABLEVELS_RES (IMPTABLEVELS_COLUMNS - 1)
#define GET_STATE(cell) ((cell)&0x0f)
#define GET_ACTION(cell) ((cell)>>4)
#define s(action, newState) ((uint8_t)(newState+(action<<4)))

typedef uint8_t ImpTab[][IMPTABLEVELS_COLUMNS];
typedef uint8_t ImpAct[];

/* FOOD FOR THOUGHT: each ImpTab should have its associated ImpAct,
 * instead of having a pair of ImpTab and a pair of ImpAct.
 */
typedef struct ImpTabPair {
    const void * pImpTab[2];
    const void * pImpAct[2];
} ImpTabPair;

/******************************************************************

      LEVELS  STATE  TABLES

 In all levels state tables,
      - state 0 is the initial state
      - the Res column is the increment to add to the text level
        for this property sequence.

 The impAct arrays for each table of a pair map the local action
 numbers of the table to the total list of actions. For instance,
 action 2 in a given table corresponds to the action number which
 appears in entry [2] of the impAct array for that table.
 The first entry of all impAct arrays must be 0.

 Action 1: init conditional sequence
        2: prepend conditional sequence to current sequence
        3: set ON sequence to new level - 1
        4: init EN/AN/ON sequence
        5: fix EN/AN/ON sequence followed by R
        6: set previous level sequence to level 2

 Notes:
  1) These tables are used in processPropertySeq(). The input
     is property sequences as determined by resolveImplicitLevels.
  2) Most such property sequences are processed immediately
     (levels are assigned).
  3) However, some sequences cannot be assigned a final level till
     one or more following sequences are received. For instance,
     ON following an R sequence within an even-level paragraph.
     If the following sequence is R, the ON sequence will be
     assigned basic run level+1, and so will the R sequence.
  4) S is generally handled like ON, since its level will be fixed
     to paragraph level in adjustWSLevels().

*/

static const ImpTab impTabL_DEFAULT =   /* Even paragraph level */
/*  In this table, conditional sequences receive the lower possible level
    until proven otherwise.
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 : init       */ {     0 ,     1 ,     0 ,     2 ,     0 ,     0 ,     0 ,  0 },
/* 1 : R          */ {     0 ,     1 ,     3 ,     3 , s(1,4), s(1,4),     0 ,  1 },
/* 2 : AN         */ {     0 ,     1 ,     0 ,     2 , s(1,5), s(1,5),     0 ,  2 },
/* 3 : R+EN/AN    */ {     0 ,     1 ,     3 ,     3 , s(1,4), s(1,4),     0 ,  2 },
/* 4 : R+ON       */ {     0 , s(2,1), s(3,3), s(3,3),     4 ,     4 ,     0 ,  0 },
/* 5 : AN+ON      */ {     0 , s(2,1),     0 , s(3,2),     5 ,     5 ,     0 ,  0 }
};
static const ImpTab impTabR_DEFAULT =   /* Odd  paragraph level */
/*  In this table, conditional sequences receive the lower possible level
    until proven otherwise.
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 : init       */ {     1 ,     0 ,     2 ,     2 ,     0 ,     0 ,     0 ,  0 },
/* 1 : L          */ {     1 ,     0 ,     1 ,     3 , s(1,4), s(1,4),     0 ,  1 },
/* 2 : EN/AN      */ {     1 ,     0 ,     2 ,     2 ,     0 ,     0 ,     0 ,  1 },
/* 3 : L+AN       */ {     1 ,     0 ,     1 ,     3 ,     5 ,     5 ,     0 ,  1 },
/* 4 : L+ON       */ { s(2,1),     0 , s(2,1),     3 ,     4 ,     4 ,     0 ,  0 },
/* 5 : L+AN+ON    */ {     1 ,     0 ,     1 ,     3 ,     5 ,     5 ,     0 ,  0 }
};
static const ImpAct impAct0 = {0,1,2,3,4};
static const ImpTabPair impTab_DEFAULT = {{&impTabL_DEFAULT,
                                           &impTabR_DEFAULT},
                                          {&impAct0, &impAct0}};

static const ImpTab impTabL_NUMBERS_SPECIAL =   /* Even paragraph level */
/*  In this table, conditional sequences receive the lower possible level
    until proven otherwise.
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 : init       */ {     0 ,     2 , s(1,1), s(1,1),     0 ,     0 ,     0 ,  0 },
/* 1 : L+EN/AN    */ {     0 , s(4,2),     1 ,     1 ,     0 ,     0 ,     0 ,  0 },
/* 2 : R          */ {     0 ,     2 ,     4 ,     4 , s(1,3), s(1,3),     0 ,  1 },
/* 3 : R+ON       */ {     0 , s(2,2), s(3,4), s(3,4),     3 ,     3 ,     0 ,  0 },
/* 4 : R+EN/AN    */ {     0 ,     2 ,     4 ,     4 , s(1,3), s(1,3),     0 ,  2 }
};
static const ImpTabPair impTab_NUMBERS_SPECIAL = {{&impTabL_NUMBERS_SPECIAL,
                                                   &impTabR_DEFAULT},
                                                  {&impAct0, &impAct0}};

static const ImpTab impTabL_GROUP_NUMBERS_WITH_R =
/*  In this table, EN/AN+ON sequences receive levels as if associated with R
    until proven that there is L or sor/eor on both sides. AN is handled like EN.
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 init         */ {     0 ,     3 , s(1,1), s(1,1),     0 ,     0 ,     0 ,  0 },
/* 1 EN/AN        */ { s(2,0),     3 ,     1 ,     1 ,     2 , s(2,0), s(2,0),  2 },
/* 2 EN/AN+ON     */ { s(2,0),     3 ,     1 ,     1 ,     2 , s(2,0), s(2,0),  1 },
/* 3 R            */ {     0 ,     3 ,     5 ,     5 , s(1,4),     0 ,     0 ,  1 },
/* 4 R+ON         */ { s(2,0),     3 ,     5 ,     5 ,     4 , s(2,0), s(2,0),  1 },
/* 5 R+EN/AN      */ {     0 ,     3 ,     5 ,     5 , s(1,4),     0 ,     0 ,  2 }
};
static const ImpTab impTabR_GROUP_NUMBERS_WITH_R =
/*  In this table, EN/AN+ON sequences receive levels as if associated with R
    until proven that there is L on both sides. AN is handled like EN.
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 init         */ {     2 ,     0 ,     1 ,     1 ,     0 ,     0 ,     0 ,  0 },
/* 1 EN/AN        */ {     2 ,     0 ,     1 ,     1 ,     0 ,     0 ,     0 ,  1 },
/* 2 L            */ {     2 ,     0 , s(1,4), s(1,4), s(1,3),     0 ,     0 ,  1 },
/* 3 L+ON         */ { s(2,2),     0 ,     4 ,     4 ,     3 ,     0 ,     0 ,  0 },
/* 4 L+EN/AN      */ { s(2,2),     0 ,     4 ,     4 ,     3 ,     0 ,     0 ,  1 }
};
static const ImpTabPair impTab_GROUP_NUMBERS_WITH_R = {
                        {&impTabL_GROUP_NUMBERS_WITH_R,
                         &impTabR_GROUP_NUMBERS_WITH_R},
                        {&impAct0, &impAct0}};


static const ImpTab impTabL_INVERSE_NUMBERS_AS_L =
/*  This table is identical to the Default LTR table except that EN and AN are
    handled like L.
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 : init       */ {     0 ,     1 ,     0 ,     0 ,     0 ,     0 ,     0 ,  0 },
/* 1 : R          */ {     0 ,     1 ,     0 ,     0 , s(1,4), s(1,4),     0 ,  1 },
/* 2 : AN         */ {     0 ,     1 ,     0 ,     0 , s(1,5), s(1,5),     0 ,  2 },
/* 3 : R+EN/AN    */ {     0 ,     1 ,     0 ,     0 , s(1,4), s(1,4),     0 ,  2 },
/* 4 : R+ON       */ { s(2,0),     1 , s(2,0), s(2,0),     4 ,     4 , s(2,0),  1 },
/* 5 : AN+ON      */ { s(2,0),     1 , s(2,0), s(2,0),     5 ,     5 , s(2,0),  1 }
};
static const ImpTab impTabR_INVERSE_NUMBERS_AS_L =
/*  This table is identical to the Default RTL table except that EN and AN are
    handled like L.
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 : init       */ {     1 ,     0 ,     1 ,     1 ,     0 ,     0 ,     0 ,  0 },
/* 1 : L          */ {     1 ,     0 ,     1 ,     1 , s(1,4), s(1,4),     0 ,  1 },
/* 2 : EN/AN      */ {     1 ,     0 ,     1 ,     1 ,     0 ,     0 ,     0 ,  1 },
/* 3 : L+AN       */ {     1 ,     0 ,     1 ,     1 ,     5 ,     5 ,     0 ,  1 },
/* 4 : L+ON       */ { s(2,1),     0 , s(2,1), s(2,1),     4 ,     4 ,     0 ,  0 },
/* 5 : L+AN+ON    */ {     1 ,     0 ,     1 ,     1 ,     5 ,     5 ,     0 ,  0 }
};
static const ImpTabPair impTab_INVERSE_NUMBERS_AS_L = {
                        {&impTabL_INVERSE_NUMBERS_AS_L,
                         &impTabR_INVERSE_NUMBERS_AS_L},
                        {&impAct0, &impAct0}};

static const ImpTab impTabR_INVERSE_LIKE_DIRECT =   /* Odd  paragraph level */
/*  In this table, conditional sequences receive the lower possible level
    until proven otherwise.
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 : init       */ {     1 ,     0 ,     2 ,     2 ,     0 ,     0 ,     0 ,  0 },
/* 1 : L          */ {     1 ,     0 ,     1 ,     2 , s(1,3), s(1,3),     0 ,  1 },
/* 2 : EN/AN      */ {     1 ,     0 ,     2 ,     2 ,     0 ,     0 ,     0 ,  1 },
/* 3 : L+ON       */ { s(2,1), s(3,0),     6 ,     4 ,     3 ,     3 , s(3,0),  0 },
/* 4 : L+ON+AN    */ { s(2,1), s(3,0),     6 ,     4 ,     5 ,     5 , s(3,0),  3 },
/* 5 : L+AN+ON    */ { s(2,1), s(3,0),     6 ,     4 ,     5 ,     5 , s(3,0),  2 },
/* 6 : L+ON+EN    */ { s(2,1), s(3,0),     6 ,     4 ,     3 ,     3 , s(3,0),  1 }
};
static const ImpAct impAct1 = {0,1,13,14};
/* FOOD FOR THOUGHT: in LTR table below, check case "JKL 123abc"
 */
static const ImpTabPair impTab_INVERSE_LIKE_DIRECT = {
                        {&impTabL_DEFAULT,
                         &impTabR_INVERSE_LIKE_DIRECT},
                        {&impAct0, &impAct1}};

static const ImpTab impTabL_INVERSE_LIKE_DIRECT_WITH_MARKS =
/*  The case handled in this table is (visually):  R EN L
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 : init       */ {     0 , s(6,3),     0 ,     1 ,     0 ,     0 ,     0 ,  0 },
/* 1 : L+AN       */ {     0 , s(6,3),     0 ,     1 , s(1,2), s(3,0),     0 ,  4 },
/* 2 : L+AN+ON    */ { s(2,0), s(6,3), s(2,0),     1 ,     2 , s(3,0), s(2,0),  3 },
/* 3 : R          */ {     0 , s(6,3), s(5,5), s(5,6), s(1,4), s(3,0),     0 ,  3 },
/* 4 : R+ON       */ { s(3,0), s(4,3), s(5,5), s(5,6),     4 , s(3,0), s(3,0),  3 },
/* 5 : R+EN       */ { s(3,0), s(4,3),     5 , s(5,6), s(1,4), s(3,0), s(3,0),  4 },
/* 6 : R+AN       */ { s(3,0), s(4,3), s(5,5),     6 , s(1,4), s(3,0), s(3,0),  4 }
};
static const ImpTab impTabR_INVERSE_LIKE_DIRECT_WITH_MARKS =
/*  The cases handled in this table are (visually):  R EN L
                                                     R L AN L
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 : init       */ { s(1,3),     0 ,     1 ,     1 ,     0 ,     0 ,     0 ,  0 },
/* 1 : R+EN/AN    */ { s(2,3),     0 ,     1 ,     1 ,     2 , s(4,0),     0 ,  1 },
/* 2 : R+EN/AN+ON */ { s(2,3),     0 ,     1 ,     1 ,     2 , s(4,0),     0 ,  0 },
/* 3 : L          */ {     3 ,     0 ,     3 , s(3,6), s(1,4), s(4,0),     0 ,  1 },
/* 4 : L+ON       */ { s(5,3), s(4,0),     5 , s(3,6),     4 , s(4,0), s(4,0),  0 },
/* 5 : L+ON+EN    */ { s(5,3), s(4,0),     5 , s(3,6),     4 , s(4,0), s(4,0),  1 },
/* 6 : L+AN       */ { s(5,3), s(4,0),     6 ,     6 ,     4 , s(4,0), s(4,0),  3 }
};
static const ImpAct impAct2 = {0,1,2,5,6,7,8};
static const ImpAct impAct3 = {0,1,9,10,11,12};
static const ImpTabPair impTab_INVERSE_LIKE_DIRECT_WITH_MARKS = {
                        {&impTabL_INVERSE_LIKE_DIRECT_WITH_MARKS,
                         &impTabR_INVERSE_LIKE_DIRECT_WITH_MARKS},
                        {&impAct2, &impAct3}};

static const ImpTabPair impTab_INVERSE_FOR_NUMBERS_SPECIAL = {
                        {&impTabL_NUMBERS_SPECIAL,
                         &impTabR_INVERSE_LIKE_DIRECT},
                        {&impAct0, &impAct1}};

static const ImpTab impTabL_INVERSE_FOR_NUMBERS_SPECIAL_WITH_MARKS =
/*  The case handled in this table is (visually):  R EN L
*/
{
/*                         L ,     R ,    EN ,    AN ,    ON ,     S ,     B , Res */
/* 0 : init       */ {     0 , s(6,2),     1 ,     1 ,     0 ,     0 ,     0 ,  0 },
/* 1 : L+EN/AN    */ {     0 , s(6,2),     1 ,     1 ,     0 , s(3,0),     0 ,  4 },
/* 2 : R          */ {     0 , s(6,2), s(5,4), s(5,4), s(1,3), s(3,0),     0 ,  3 },
/* 3 : R+ON       */ { s(3,0), s(4,2), s(5,4), s(5,4),     3 , s(3,0), s(3,0),  3 },
/* 4 : R+EN/AN    */ { s(3,0), s(4,2),     4 ,     4 , s(1,3), s(3,0), s(3,0),  4 }
};
static const ImpTabPair impTab_INVERSE_FOR_NUMBERS_SPECIAL_WITH_MARKS = {
                        {&impTabL_INVERSE_FOR_NUMBERS_SPECIAL_WITH_MARKS,
                         &impTabR_INVERSE_LIKE_DIRECT_WITH_MARKS},
                        {&impAct2, &impAct3}};

#undef s

typedef struct {
    const ImpTab * pImpTab;             /* level table pointer          */
    const ImpAct * pImpAct;             /* action map array             */
    int32_t startON;                    /* start of ON sequence         */
    int32_t startL2EN;                  /* start of level 2 sequence    */
    int32_t lastStrongRTL;              /* index of last found R or AL  */
    int32_t state;                      /* current state                */
    int32_t runStart;                   /* start position of the run    */
    UBiDiLevel runLevel;                /* run level before implicit solving */
} LevState;

/*------------------------------------------------------------------------*/

static void
addPoint(UBiDi *pBiDi, int32_t pos, int32_t flag)
  /* param pos:     position where to insert
     param flag:    one of LRM_BEFORE, LRM_AFTER, RLM_BEFORE, RLM_AFTER
  */
{
#define FIRSTALLOC  10
    Point point;
    InsertPoints * pInsertPoints=&(pBiDi->insertPoints);

    if (pInsertPoints->capacity == 0)
    {
        pInsertPoints->points=static_cast<Point *>(uprv_malloc(sizeof(Point)*FIRSTALLOC));
        if (pInsertPoints->points == NULL)
        {
            pInsertPoints->errorCode=U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        pInsertPoints->capacity=FIRSTALLOC;
    }
    if (pInsertPoints->size >= pInsertPoints->capacity) /* no room for new point */
    {
        Point * savePoints=pInsertPoints->points;
        pInsertPoints->points=static_cast<Point *>(uprv_realloc(pInsertPoints->points,
                                           pInsertPoints->capacity*2*sizeof(Point)));
        if (pInsertPoints->points == NULL)
        {
            pInsertPoints->points=savePoints;
            pInsertPoints->errorCode=U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        else  pInsertPoints->capacity*=2;
    }
    point.pos=pos;
    point.flag=flag;
    pInsertPoints->points[pInsertPoints->size]=point;
    pInsertPoints->size++;
#undef FIRSTALLOC
}

static void
setLevelsOutsideIsolates(UBiDi *pBiDi, int32_t start, int32_t limit, UBiDiLevel level)
{
    DirProp *dirProps=pBiDi->dirProps, dirProp;
    UBiDiLevel *levels=pBiDi->levels;
    int32_t isolateCount=0, k;
    for(k=start; k<limit; k++) {
        dirProp=dirProps[k];
        if(dirProp==PDI)
            isolateCount--;
        if(isolateCount==0)
            levels[k]=level;
        if(dirProp==LRI || dirProp==RLI)
            isolateCount++;
    }
}

/* perform rules (Wn), (Nn), and (In) on a run of the text ------------------ */

/*
 * This implementation of the (Wn) rules applies all rules in one pass.
 * In order to do so, it needs a look-ahead of typically 1 character
 * (except for W5: sequences of ET) and keeps track of changes
 * in a rule Wp that affect a later Wq (p<q).
 *
 * The (Nn) and (In) rules are also performed in that same single loop,
 * but effectively one iteration behind for white space.
 *
 * Since all implicit rules are performed in one step, it is not necessary
 * to actually store the intermediate directional properties in dirProps[].
 */

static void
processPropertySeq(UBiDi *pBiDi, LevState *pLevState, uint8_t _prop,
                   int32_t start, int32_t limit) {
    uint8_t cell, oldStateSeq, actionSeq;
    const ImpTab * pImpTab=pLevState->pImpTab;
    const ImpAct * pImpAct=pLevState->pImpAct;
    UBiDiLevel * levels=pBiDi->levels;
    UBiDiLevel level, addLevel;
    InsertPoints * pInsertPoints;
    int32_t start0, k;

    start0=start;                           /* save original start position */
    oldStateSeq=(uint8_t)pLevState->state;
    cell=(*pImpTab)[oldStateSeq][_prop];
    pLevState->state=GET_STATE(cell);       /* isolate the new state */
    actionSeq=(*pImpAct)[GET_ACTION(cell)]; /* isolate the action */
    addLevel=(*pImpTab)[pLevState->state][IMPTABLEVELS_RES];

    if(actionSeq) {
        switch(actionSeq) {
        case 1:                         /* init ON seq */
            pLevState->startON=start0;
            break;

        case 2:                         /* prepend ON seq to current seq */
            start=pLevState->startON;
            break;

        case 3:                         /* EN/AN after R+ON */
            level=pLevState->runLevel+1;
            setLevelsOutsideIsolates(pBiDi, pLevState->startON, start0, level);
            break;

        case 4:                         /* EN/AN before R for NUMBERS_SPECIAL */
            level=pLevState->runLevel+2;
            setLevelsOutsideIsolates(pBiDi, pLevState->startON, start0, level);
            break;

        case 5:                         /* L or S after possible relevant EN/AN */
            /* check if we had EN after R/AL */
            if (pLevState->startL2EN >= 0) {
                addPoint(pBiDi, pLevState->startL2EN, LRM_BEFORE);
            }
            pLevState->startL2EN=-1;  /* not within previous if since could also be -2 */
            /* check if we had any relevant EN/AN after R/AL */
            pInsertPoints=&(pBiDi->insertPoints);
            if ((pInsertPoints->capacity == 0) ||
                (pInsertPoints->size <= pInsertPoints->confirmed))
            {
                /* nothing, just clean up */
                pLevState->lastStrongRTL=-1;
                /* check if we have a pending conditional segment */
                level=(*pImpTab)[oldStateSeq][IMPTABLEVELS_RES];
                if ((level & 1) && (pLevState->startON > 0)) {  /* after ON */
                    start=pLevState->startON;   /* reset to basic run level */
                }
                if (_prop == DirProp_S)                /* add LRM before S */
                {
                    addPoint(pBiDi, start0, LRM_BEFORE);
                    pInsertPoints->confirmed=pInsertPoints->size;
                }
                break;
            }
            /* reset previous RTL cont to level for LTR text */
            for (k=pLevState->lastStrongRTL+1; k<start0; k++)
            {
                /* reset odd level, leave runLevel+2 as is */
                levels[k]=(levels[k] - 2) & ~1;
            }
            /* mark insert points as confirmed */
            pInsertPoints->confirmed=pInsertPoints->size;
            pLevState->lastStrongRTL=-1;
            if (_prop == DirProp_S)            /* add LRM before S */
            {
                addPoint(pBiDi, start0, LRM_BEFORE);
                pInsertPoints->confirmed=pInsertPoints->size;
            }
            break;

        case 6:                         /* R/AL after possible relevant EN/AN */
            /* just clean up */
            pInsertPoints=&(pBiDi->insertPoints);
            if (pInsertPoints->capacity > 0)
                /* remove all non confirmed insert points */
                pInsertPoints->size=pInsertPoints->confirmed;
            pLevState->startON=-1;
            pLevState->startL2EN=-1;
            pLevState->lastStrongRTL=limit - 1;
            break;

        case 7:                         /* EN/AN after R/AL + possible cont */
            /* check for real AN */
            if ((_prop == DirProp_AN) && (pBiDi->dirProps[start0] == AN) &&
                (pBiDi->reorderingMode!=UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL))
            {
                /* real AN */
                if (pLevState->startL2EN == -1) /* if no relevant EN already found */
                {
                    /* just note the righmost digit as a strong RTL */
                    pLevState->lastStrongRTL=limit - 1;
                    break;
                }
                if (pLevState->startL2EN >= 0)  /* after EN, no AN */
                {
                    addPoint(pBiDi, pLevState->startL2EN, LRM_BEFORE);
                    pLevState->startL2EN=-2;
                }
                /* note AN */
                addPoint(pBiDi, start0, LRM_BEFORE);
                break;
            }
            /* if first EN/AN after R/AL */
            if (pLevState->startL2EN == -1) {
                pLevState->startL2EN=start0;
            }
            break;

        case 8:                         /* note location of latest R/AL */
            pLevState->lastStrongRTL=limit - 1;
            pLevState->startON=-1;
            break;

        case 9:                         /* L after R+ON/EN/AN */
            /* include possible adjacent number on the left */
            for (k=start0-1; k>=0 && !(levels[k]&1); k--);
            if(k>=0) {
                addPoint(pBiDi, k, RLM_BEFORE);             /* add RLM before */
                pInsertPoints=&(pBiDi->insertPoints);
                pInsertPoints->confirmed=pInsertPoints->size;   /* confirm it */
            }
            pLevState->startON=start0;
            break;

        case 10:                        /* AN after L */
            /* AN numbers between L text on both sides may be trouble. */
            /* tentatively bracket with LRMs; will be confirmed if followed by L */
            addPoint(pBiDi, start0, LRM_BEFORE);    /* add LRM before */
            addPoint(pBiDi, start0, LRM_AFTER);     /* add LRM after  */
            break;

        case 11:                        /* R after L+ON/EN/AN */
            /* false alert, infirm LRMs around previous AN */
            pInsertPoints=&(pBiDi->insertPoints);
            pInsertPoints->size=pInsertPoints->confirmed;
            if (_prop == DirProp_S)            /* add RLM before S */
            {
                addPoint(pBiDi, start0, RLM_BEFORE);
                pInsertPoints->confirmed=pInsertPoints->size;
            }
            break;

        case 12:                        /* L after L+ON/AN */
            level=pLevState->runLevel + addLevel;
            for(k=pLevState->startON; k<start0; k++) {
                if (levels[k]<level)
                    levels[k]=level;
            }
            pInsertPoints=&(pBiDi->insertPoints);
            pInsertPoints->confirmed=pInsertPoints->size;   /* confirm inserts */
            pLevState->startON=start0;
            break;

        case 13:                        /* L after L+ON+EN/AN/ON */
            level=pLevState->runLevel;
            for(k=start0-1; k>=pLevState->startON; k--) {
                if(levels[k]==level+3) {
                    while(levels[k]==level+3) {
                        levels[k--]-=2;
                    }
                    while(levels[k]==level) {
                        k--;
                    }
                }
                if(levels[k]==level+2) {
                    levels[k]=level;
                    continue;
                }
                levels[k]=level+1;
            }
            break;

        case 14:                        /* R after L+ON+EN/AN/ON */
            level=pLevState->runLevel+1;
            for(k=start0-1; k>=pLevState->startON; k--) {
                if(levels[k]>level) {
                    levels[k]-=2;
                }
            }
            break;

        default:                        /* we should never get here */
            UPRV_UNREACHABLE;
        }
    }
    if((addLevel) || (start < start0)) {
        level=pLevState->runLevel + addLevel;
        if(start>=pLevState->runStart) {
            for(k=start; k<limit; k++) {
                levels[k]=level;
            }
        } else {
            setLevelsOutsideIsolates(pBiDi, start, limit, level);
        }
    }
}

/**
 * Returns the directionality of the last strong character at the end of the prologue, if any.
 * Requires prologue!=null.
 */
static DirProp
lastL_R_AL(UBiDi *pBiDi) {
    const UChar *text=pBiDi->prologue;
    int32_t length=pBiDi->proLength;
    int32_t i;
    UChar32 uchar;
    DirProp dirProp;
    for(i=length; i>0; ) {
        /* i is decremented by U16_PREV */
        U16_PREV(text, 0, i, uchar);
        dirProp=(DirProp)ubidi_getCustomizedClass(pBiDi, uchar);
        if(dirProp==L) {
            return DirProp_L;
        }
        if(dirProp==R || dirProp==AL) {
            return DirProp_R;
        }
        if(dirProp==B) {
            return DirProp_ON;
        }
    }
    return DirProp_ON;
}

/**
 * Returns the directionality of the first strong character, or digit, in the epilogue, if any.
 * Requires epilogue!=null.
 */
static DirProp
firstL_R_AL_EN_AN(UBiDi *pBiDi) {
    const UChar *text=pBiDi->epilogue;
    int32_t length=pBiDi->epiLength;
    int32_t i;
    UChar32 uchar;
    DirProp dirProp;
    for(i=0; i<length; ) {
        /* i is incremented by U16_NEXT */
        U16_NEXT(text, i, length, uchar);
        dirProp=(DirProp)ubidi_getCustomizedClass(pBiDi, uchar);
        if(dirProp==L) {
            return DirProp_L;
        }
        if(dirProp==R || dirProp==AL) {
            return DirProp_R;
        }
        if(dirProp==EN) {
            return DirProp_EN;
        }
        if(dirProp==AN) {
            return DirProp_AN;
        }
    }
    return DirProp_ON;
}

static void
resolveImplicitLevels(UBiDi *pBiDi,
                      int32_t start, int32_t limit,
                      DirProp sor, DirProp eor) {
    const DirProp *dirProps=pBiDi->dirProps;
    DirProp dirProp;
    LevState levState;
    int32_t i, start1, start2;
    uint16_t oldStateImp, stateImp, actionImp;
    uint8_t gprop, resProp, cell;
    UBool inverseRTL;
    DirProp nextStrongProp=R;
    int32_t nextStrongPos=-1;

    /* check for RTL inverse BiDi mode */
    /* FOOD FOR THOUGHT: in case of RTL inverse BiDi, it would make sense to
     * loop on the text characters from end to start.
     * This would need a different properties state table (at least different
     * actions) and different levels state tables (maybe very similar to the
     * LTR corresponding ones.
     */
    inverseRTL=(UBool)
        ((start<pBiDi->lastArabicPos) && (GET_PARALEVEL(pBiDi, start) & 1) &&
         (pBiDi->reorderingMode==UBIDI_REORDER_INVERSE_LIKE_DIRECT  ||
          pBiDi->reorderingMode==UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL));

    /* initialize for property and levels state tables */
    levState.startL2EN=-1;              /* used for INVERSE_LIKE_DIRECT_WITH_MARKS */
    levState.lastStrongRTL=-1;          /* used for INVERSE_LIKE_DIRECT_WITH_MARKS */
    levState.runStart=start;
    levState.runLevel=pBiDi->levels[start];
    levState.pImpTab=(const ImpTab*)((pBiDi->pImpTabPair)->pImpTab)[levState.runLevel&1];
    levState.pImpAct=(const ImpAct*)((pBiDi->pImpTabPair)->pImpAct)[levState.runLevel&1];
    if(start==0 && pBiDi->proLength>0) {
        DirProp lastStrong=lastL_R_AL(pBiDi);
        if(lastStrong!=DirProp_ON) {
            sor=lastStrong;
        }
    }
    /* The isolates[] entries contain enough information to
       resume the bidi algorithm in the same state as it was
       when it was interrupted by an isolate sequence. */
    if(dirProps[start]==PDI  && pBiDi->isolateCount >= 0) {
        levState.startON=pBiDi->isolates[pBiDi->isolateCount].startON;
        start1=pBiDi->isolates[pBiDi->isolateCount].start1;
        stateImp=pBiDi->isolates[pBiDi->isolateCount].stateImp;
        levState.state=pBiDi->isolates[pBiDi->isolateCount].state;
        pBiDi->isolateCount--;
    } else {
        levState.startON=-1;
        start1=start;
        if(dirProps[start]==NSM)
            stateImp = 1 + sor;
        else
            stateImp=0;
        levState.state=0;
        processPropertySeq(pBiDi, &levState, sor, start, start);
    }
    start2=start;                       /* to make Java compiler happy */

    for(i=start; i<=limit; i++) {
        if(i>=limit) {
            int32_t k;
            for(k=limit-1; k>start&&(DIRPROP_FLAG(dirProps[k])&MASK_BN_EXPLICIT); k--);
            dirProp=dirProps[k];
            if(dirProp==LRI || dirProp==RLI)
                break;      /* no forced closing for sequence ending with LRI/RLI */
            gprop=eor;
        } else {
            DirProp prop, prop1;
            prop=dirProps[i];
            if(prop==B) {
                pBiDi->isolateCount=-1; /* current isolates stack entry == none */
            }
            if(inverseRTL) {
                if(prop==AL) {
                    /* AL before EN does not make it AN */
                    prop=R;
                } else if(prop==EN) {
                    if(nextStrongPos<=i) {
                        /* look for next strong char (L/R/AL) */
                        int32_t j;
                        nextStrongProp=R;   /* set default */
                        nextStrongPos=limit;
                        for(j=i+1; j<limit; j++) {
                            prop1=dirProps[j];
                            if(prop1==L || prop1==R || prop1==AL) {
                                nextStrongProp=prop1;
                                nextStrongPos=j;
                                break;
                            }
                        }
                    }
                    if(nextStrongProp==AL) {
                        prop=AN;
                    }
                }
            }
            gprop=groupProp[prop];
        }
        oldStateImp=stateImp;
        cell=impTabProps[oldStateImp][gprop];
        stateImp=GET_STATEPROPS(cell);      /* isolate the new state */
        actionImp=GET_ACTIONPROPS(cell);    /* isolate the action */
        if((i==limit) && (actionImp==0)) {
            /* there is an unprocessed sequence if its property == eor   */
            actionImp=1;                    /* process the last sequence */
        }
        if(actionImp) {
            resProp=impTabProps[oldStateImp][IMPTABPROPS_RES];
            switch(actionImp) {
            case 1:             /* process current seq1, init new seq1 */
                processPropertySeq(pBiDi, &levState, resProp, start1, i);
                start1=i;
                break;
            case 2:             /* init new seq2 */
                start2=i;
                break;
            case 3:             /* process seq1, process seq2, init new seq1 */
                processPropertySeq(pBiDi, &levState, resProp, start1, start2);
                processPropertySeq(pBiDi, &levState, DirProp_ON, start2, i);
                start1=i;
                break;
            case 4:             /* process seq1, set seq1=seq2, init new seq2 */
                processPropertySeq(pBiDi, &levState, resProp, start1, start2);
                start1=start2;
                start2=i;
                break;
            default:            /* we should never get here */
                UPRV_UNREACHABLE;
            }
        }
    }

    /* flush possible pending sequence, e.g. ON */
    if(limit==pBiDi->length && pBiDi->epiLength>0) {
        DirProp firstStrong=firstL_R_AL_EN_AN(pBiDi);
        if(firstStrong!=DirProp_ON) {
            eor=firstStrong;
        }
    }

    /* look for the last char not a BN or LRE/RLE/LRO/RLO/PDF */
    for(i=limit-1; i>start&&(DIRPROP_FLAG(dirProps[i])&MASK_BN_EXPLICIT); i--);
    dirProp=dirProps[i];
    if((dirProp==LRI || dirProp==RLI) && limit<pBiDi->length) {
        pBiDi->isolateCount++;
        pBiDi->isolates[pBiDi->isolateCount].stateImp=stateImp;
        pBiDi->isolates[pBiDi->isolateCount].state=levState.state;
        pBiDi->isolates[pBiDi->isolateCount].start1=start1;
        pBiDi->isolates[pBiDi->isolateCount].startON=levState.startON;
    }
    else
        processPropertySeq(pBiDi, &levState, eor, limit, limit);
}

/* perform (L1) and (X9) ---------------------------------------------------- */

/*
 * Reset the embedding levels for some non-graphic characters (L1).
 * This function also sets appropriate levels for BN, and
 * explicit embedding types that are supposed to have been removed
 * from the paragraph in (X9).
 */
static void
adjustWSLevels(UBiDi *pBiDi) {
    const DirProp *dirProps=pBiDi->dirProps;
    UBiDiLevel *levels=pBiDi->levels;
    int32_t i;

    if(pBiDi->flags&MASK_WS) {
        UBool orderParagraphsLTR=pBiDi->orderParagraphsLTR;
        Flags flag;

        i=pBiDi->trailingWSStart;
        while(i>0) {
            /* reset a sequence of WS/BN before eop and B/S to the paragraph paraLevel */
            while(i>0 && (flag=DIRPROP_FLAG(dirProps[--i]))&MASK_WS) {
                if(orderParagraphsLTR&&(flag&DIRPROP_FLAG(B))) {
                    levels[i]=0;
                } else {
                    levels[i]=GET_PARALEVEL(pBiDi, i);
                }
            }

            /* reset BN to the next character's paraLevel until B/S, which restarts above loop */
            /* here, i+1 is guaranteed to be <length */
            while(i>0) {
                flag=DIRPROP_FLAG(dirProps[--i]);
                if(flag&MASK_BN_EXPLICIT) {
                    levels[i]=levels[i+1];
                } else if(orderParagraphsLTR&&(flag&DIRPROP_FLAG(B))) {
                    levels[i]=0;
                    break;
                } else if(flag&MASK_B_S) {
                    levels[i]=GET_PARALEVEL(pBiDi, i);
                    break;
                }
            }
        }
    }
}

U_CAPI void U_EXPORT2
ubidi_setContext(UBiDi *pBiDi,
                 const UChar *prologue, int32_t proLength,
                 const UChar *epilogue, int32_t epiLength,
                 UErrorCode *pErrorCode) {
    /* check the argument values */
    RETURN_VOID_IF_NULL_OR_FAILING_ERRCODE(pErrorCode);
    if(pBiDi==NULL || proLength<-1 || epiLength<-1 ||
       (prologue==NULL && proLength!=0) || (epilogue==NULL && epiLength!=0)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    if(proLength==-1) {
        pBiDi->proLength=u_strlen(prologue);
    } else {
        pBiDi->proLength=proLength;
    }
    if(epiLength==-1) {
        pBiDi->epiLength=u_strlen(epilogue);
    } else {
        pBiDi->epiLength=epiLength;
    }
    pBiDi->prologue=prologue;
    pBiDi->epilogue=epilogue;
}

static void
setParaSuccess(UBiDi *pBiDi) {
    pBiDi->proLength=0;                 /* forget the last context */
    pBiDi->epiLength=0;
    pBiDi->pParaBiDi=pBiDi;             /* mark successful setPara */
}

#define BIDI_MIN(x, y)   ((x)<(y) ? (x) : (y))
#define BIDI_ABS(x)      ((x)>=0  ? (x) : (-(x)))

static void
setParaRunsOnly(UBiDi *pBiDi, const UChar *text, int32_t length,
                UBiDiLevel paraLevel, UErrorCode *pErrorCode) {
    int32_t *runsOnlyMemory = NULL;
    int32_t *visualMap;
    UChar *visualText;
    int32_t saveLength, saveTrailingWSStart;
    const UBiDiLevel *levels;
    UBiDiLevel *saveLevels;
    UBiDiDirection saveDirection;
    UBool saveMayAllocateText;
    Run *runs;
    int32_t visualLength, i, j, visualStart, logicalStart,
            runCount, runLength, addedRuns, insertRemove,
            start, limit, step, indexOddBit, logicalPos,
            index0, index1;
    uint32_t saveOptions;

    pBiDi->reorderingMode=UBIDI_REORDER_DEFAULT;
    if(length==0) {
        ubidi_setPara(pBiDi, text, length, paraLevel, NULL, pErrorCode);
        goto cleanup3;
    }
    /* obtain memory for mapping table and visual text */
    runsOnlyMemory=static_cast<int32_t *>(uprv_malloc(length*(sizeof(int32_t)+sizeof(UChar)+sizeof(UBiDiLevel))));
    if(runsOnlyMemory==NULL) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        goto cleanup3;
    }
    visualMap=runsOnlyMemory;
    visualText=(UChar *)&visualMap[length];
    saveLevels=(UBiDiLevel *)&visualText[length];
    saveOptions=pBiDi->reorderingOptions;
    if(saveOptions & UBIDI_OPTION_INSERT_MARKS) {
        pBiDi->reorderingOptions&=~UBIDI_OPTION_INSERT_MARKS;
        pBiDi->reorderingOptions|=UBIDI_OPTION_REMOVE_CONTROLS;
    }
    paraLevel&=1;                       /* accept only 0 or 1 */
    ubidi_setPara(pBiDi, text, length, paraLevel, NULL, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        goto cleanup3;
    }
    /* we cannot access directly pBiDi->levels since it is not yet set if
     * direction is not MIXED
     */
    levels=ubidi_getLevels(pBiDi, pErrorCode);
    uprv_memcpy(saveLevels, levels, (size_t)pBiDi->length*sizeof(UBiDiLevel));
    saveTrailingWSStart=pBiDi->trailingWSStart;
    saveLength=pBiDi->length;
    saveDirection=pBiDi->direction;

    /* FOOD FOR THOUGHT: instead of writing the visual text, we could use
     * the visual map and the dirProps array to drive the second call
     * to ubidi_setPara (but must make provision for possible removal of
     * BiDi controls.  Alternatively, only use the dirProps array via
     * customized classifier callback.
     */
    visualLength=ubidi_writeReordered(pBiDi, visualText, length,
                                      UBIDI_DO_MIRRORING, pErrorCode);
    ubidi_getVisualMap(pBiDi, visualMap, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        goto cleanup2;
    }
    pBiDi->reorderingOptions=saveOptions;

    pBiDi->reorderingMode=UBIDI_REORDER_INVERSE_LIKE_DIRECT;
    paraLevel^=1;
    /* Because what we did with reorderingOptions, visualText may be shorter
     * than the original text. But we don't want the levels memory to be
     * reallocated shorter than the original length, since we need to restore
     * the levels as after the first call to ubidi_setpara() before returning.
     * We will force mayAllocateText to FALSE before the second call to
     * ubidi_setpara(), and will restore it afterwards.
     */
    saveMayAllocateText=pBiDi->mayAllocateText;
    pBiDi->mayAllocateText=FALSE;
    ubidi_setPara(pBiDi, visualText, visualLength, paraLevel, NULL, pErrorCode);
    pBiDi->mayAllocateText=saveMayAllocateText;
    ubidi_getRuns(pBiDi, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        goto cleanup1;
    }
    /* check if some runs must be split, count how many splits */
    addedRuns=0;
    runCount=pBiDi->runCount;
    runs=pBiDi->runs;
    visualStart=0;
    for(i=0; i<runCount; i++, visualStart+=runLength) {
        runLength=runs[i].visualLimit-visualStart;
        if(runLength<2) {
            continue;
        }
        logicalStart=GET_INDEX(runs[i].logicalStart);
        for(j=logicalStart+1; j<logicalStart+runLength; j++) {
            index0=visualMap[j];
            index1=visualMap[j-1];
            if((BIDI_ABS(index0-index1)!=1) || (saveLevels[index0]!=saveLevels[index1])) {
                addedRuns++;
            }
        }
    }
    if(addedRuns) {
        if(getRunsMemory(pBiDi, runCount+addedRuns)) {
            if(runCount==1) {
                /* because we switch from UBiDi.simpleRuns to UBiDi.runs */
                pBiDi->runsMemory[0]=runs[0];
            }
            runs=pBiDi->runs=pBiDi->runsMemory;
            pBiDi->runCount+=addedRuns;
        } else {
            goto cleanup1;
        }
    }
    /* split runs which are not consecutive in source text */
    for(i=runCount-1; i>=0; i--) {
        runLength= i==0 ? runs[0].visualLimit :
                          runs[i].visualLimit-runs[i-1].visualLimit;
        logicalStart=runs[i].logicalStart;
        indexOddBit=GET_ODD_BIT(logicalStart);
        logicalStart=GET_INDEX(logicalStart);
        if(runLength<2) {
            if(addedRuns) {
                runs[i+addedRuns]=runs[i];
            }
            logicalPos=visualMap[logicalStart];
            runs[i+addedRuns].logicalStart=MAKE_INDEX_ODD_PAIR(logicalPos,
                                            saveLevels[logicalPos]^indexOddBit);
            continue;
        }
        if(indexOddBit) {
            start=logicalStart;
            limit=logicalStart+runLength-1;
            step=1;
        } else {
            start=logicalStart+runLength-1;
            limit=logicalStart;
            step=-1;
        }
        for(j=start; j!=limit; j+=step) {
            index0=visualMap[j];
            index1=visualMap[j+step];
            if((BIDI_ABS(index0-index1)!=1) || (saveLevels[index0]!=saveLevels[index1])) {
                logicalPos=BIDI_MIN(visualMap[start], index0);
                runs[i+addedRuns].logicalStart=MAKE_INDEX_ODD_PAIR(logicalPos,
                                            saveLevels[logicalPos]^indexOddBit);
                runs[i+addedRuns].visualLimit=runs[i].visualLimit;
                runs[i].visualLimit-=BIDI_ABS(j-start)+1;
                insertRemove=runs[i].insertRemove&(LRM_AFTER|RLM_AFTER);
                runs[i+addedRuns].insertRemove=insertRemove;
                runs[i].insertRemove&=~insertRemove;
                start=j+step;
                addedRuns--;
            }
        }
        if(addedRuns) {
            runs[i+addedRuns]=runs[i];
        }
        logicalPos=BIDI_MIN(visualMap[start], visualMap[limit]);
        runs[i+addedRuns].logicalStart=MAKE_INDEX_ODD_PAIR(logicalPos,
                                            saveLevels[logicalPos]^indexOddBit);
    }

  cleanup1:
    /* restore initial paraLevel */
    pBiDi->paraLevel^=1;
  cleanup2:
    /* restore real text */
    pBiDi->text=text;
    pBiDi->length=saveLength;
    pBiDi->originalLength=length;
    pBiDi->direction=saveDirection;
    /* the saved levels should never excess levelsSize, but we check anyway */
    if(saveLength>pBiDi->levelsSize) {
        saveLength=pBiDi->levelsSize;
    }
    uprv_memcpy(pBiDi->levels, saveLevels, (size_t)saveLength*sizeof(UBiDiLevel));
    pBiDi->trailingWSStart=saveTrailingWSStart;
    if(pBiDi->runCount>1) {
        pBiDi->direction=UBIDI_MIXED;
    }
  cleanup3:
    /* free memory for mapping table and visual text */
    uprv_free(runsOnlyMemory);

    pBiDi->reorderingMode=UBIDI_REORDER_RUNS_ONLY;
}

/* ubidi_setPara ------------------------------------------------------------ */

U_CAPI void U_EXPORT2
ubidi_setPara(UBiDi *pBiDi, const UChar *text, int32_t length,
              UBiDiLevel paraLevel, UBiDiLevel *embeddingLevels,
              UErrorCode *pErrorCode) {
    UBiDiDirection direction;
    DirProp *dirProps;

    /* check the argument values */
    RETURN_VOID_IF_NULL_OR_FAILING_ERRCODE(pErrorCode);
    if(pBiDi==NULL || text==NULL || length<-1 ||
       (paraLevel>UBIDI_MAX_EXPLICIT_LEVEL && paraLevel<UBIDI_DEFAULT_LTR)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    if(length==-1) {
        length=u_strlen(text);
    }

    /* special treatment for RUNS_ONLY mode */
    if(pBiDi->reorderingMode==UBIDI_REORDER_RUNS_ONLY) {
        setParaRunsOnly(pBiDi, text, length, paraLevel, pErrorCode);
        return;
    }

    /* initialize the UBiDi structure */
    pBiDi->pParaBiDi=NULL;          /* mark unfinished setPara */
    pBiDi->text=text;
    pBiDi->length=pBiDi->originalLength=pBiDi->resultLength=length;
    pBiDi->paraLevel=paraLevel;
    pBiDi->direction=(UBiDiDirection)(paraLevel&1);
    pBiDi->paraCount=1;

    pBiDi->dirProps=NULL;
    pBiDi->levels=NULL;
    pBiDi->runs=NULL;
    pBiDi->insertPoints.size=0;         /* clean up from last call */
    pBiDi->insertPoints.confirmed=0;    /* clean up from last call */

    /*
     * Save the original paraLevel if contextual; otherwise, set to 0.
     */
    pBiDi->defaultParaLevel=IS_DEFAULT_LEVEL(paraLevel);

    if(length==0) {
        /*
         * For an empty paragraph, create a UBiDi object with the paraLevel and
         * the flags and the direction set but without allocating zero-length arrays.
         * There is nothing more to do.
         */
        if(IS_DEFAULT_LEVEL(paraLevel)) {
            pBiDi->paraLevel&=1;
            pBiDi->defaultParaLevel=0;
        }
        pBiDi->flags=DIRPROP_FLAG_LR(paraLevel);
        pBiDi->runCount=0;
        pBiDi->paraCount=0;
        setParaSuccess(pBiDi);          /* mark successful setPara */
        return;
    }

    pBiDi->runCount=-1;

    /* allocate paras memory */
    if(pBiDi->parasMemory)
        pBiDi->paras=pBiDi->parasMemory;
    else
        pBiDi->paras=pBiDi->simpleParas;

    /*
     * Get the directional properties,
     * the flags bit-set, and
     * determine the paragraph level if necessary.
     */
    if(getDirPropsMemory(pBiDi, length)) {
        pBiDi->dirProps=pBiDi->dirPropsMemory;
        if(!getDirProps(pBiDi)) {
            *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
            return;
        }
    } else {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    dirProps=pBiDi->dirProps;
    /* the processed length may have changed if UBIDI_OPTION_STREAMING */
    length= pBiDi->length;
    pBiDi->trailingWSStart=length;  /* the levels[] will reflect the WS run */

    /* are explicit levels specified? */
    if(embeddingLevels==NULL) {
        /* no: determine explicit levels according to the (Xn) rules */\
        if(getLevelsMemory(pBiDi, length)) {
            pBiDi->levels=pBiDi->levelsMemory;
            direction=resolveExplicitLevels(pBiDi, pErrorCode);
            if(U_FAILURE(*pErrorCode)) {
                return;
            }
        } else {
            *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
            return;
        }
    } else {
        /* set BN for all explicit codes, check that all levels are 0 or paraLevel..UBIDI_MAX_EXPLICIT_LEVEL */
        pBiDi->levels=embeddingLevels;
        direction=checkExplicitLevels(pBiDi, pErrorCode);
        if(U_FAILURE(*pErrorCode)) {
            return;
        }
    }

    /* allocate isolate memory */
    if(pBiDi->isolateCount<=SIMPLE_ISOLATES_COUNT)
        pBiDi->isolates=pBiDi->simpleIsolates;
    else
        if((int32_t)(pBiDi->isolateCount*sizeof(Isolate))<=pBiDi->isolatesSize)
            pBiDi->isolates=pBiDi->isolatesMemory;
        else {
            if(getInitialIsolatesMemory(pBiDi, pBiDi->isolateCount)) {
                pBiDi->isolates=pBiDi->isolatesMemory;
            } else {
                *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
                return;
            }
        }
    pBiDi->isolateCount=-1;             /* current isolates stack entry == none */

    /*
     * The steps after (X9) in the UBiDi algorithm are performed only if
     * the paragraph text has mixed directionality!
     */
    pBiDi->direction=direction;
    switch(direction) {
    case UBIDI_LTR:
        /* all levels are implicitly at paraLevel (important for ubidi_getLevels()) */
        pBiDi->trailingWSStart=0;
        break;
    case UBIDI_RTL:
        /* all levels are implicitly at paraLevel (important for ubidi_getLevels()) */
        pBiDi->trailingWSStart=0;
        break;
    default:
        /*
         *  Choose the right implicit state table
         */
        switch(pBiDi->reorderingMode) {
        case UBIDI_REORDER_DEFAULT:
            pBiDi->pImpTabPair=&impTab_DEFAULT;
            break;
        case UBIDI_REORDER_NUMBERS_SPECIAL:
            pBiDi->pImpTabPair=&impTab_NUMBERS_SPECIAL;
            break;
        case UBIDI_REORDER_GROUP_NUMBERS_WITH_R:
            pBiDi->pImpTabPair=&impTab_GROUP_NUMBERS_WITH_R;
            break;
        case UBIDI_REORDER_INVERSE_NUMBERS_AS_L:
            pBiDi->pImpTabPair=&impTab_INVERSE_NUMBERS_AS_L;
            break;
        case UBIDI_REORDER_INVERSE_LIKE_DIRECT:
            if (pBiDi->reorderingOptions & UBIDI_OPTION_INSERT_MARKS) {
                pBiDi->pImpTabPair=&impTab_INVERSE_LIKE_DIRECT_WITH_MARKS;
            } else {
                pBiDi->pImpTabPair=&impTab_INVERSE_LIKE_DIRECT;
            }
            break;
        case UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL:
            if (pBiDi->reorderingOptions & UBIDI_OPTION_INSERT_MARKS) {
                pBiDi->pImpTabPair=&impTab_INVERSE_FOR_NUMBERS_SPECIAL_WITH_MARKS;
            } else {
                pBiDi->pImpTabPair=&impTab_INVERSE_FOR_NUMBERS_SPECIAL;
            }
            break;
        default:
            /* we should never get here */
            UPRV_UNREACHABLE;
        }
        /*
         * If there are no external levels specified and there
         * are no significant explicit level codes in the text,
         * then we can treat the entire paragraph as one run.
         * Otherwise, we need to perform the following rules on runs of
         * the text with the same embedding levels. (X10)
         * "Significant" explicit level codes are ones that actually
         * affect non-BN characters.
         * Examples for "insignificant" ones are empty embeddings
         * LRE-PDF, LRE-RLE-PDF-PDF, etc.
         */
        if(embeddingLevels==NULL && pBiDi->paraCount<=1 &&
                                   !(pBiDi->flags&DIRPROP_FLAG_MULTI_RUNS)) {
            resolveImplicitLevels(pBiDi, 0, length,
                                    GET_LR_FROM_LEVEL(GET_PARALEVEL(pBiDi, 0)),
                                    GET_LR_FROM_LEVEL(GET_PARALEVEL(pBiDi, length-1)));
        } else {
            /* sor, eor: start and end types of same-level-run */
            UBiDiLevel *levels=pBiDi->levels;
            int32_t start, limit=0;
            UBiDiLevel level, nextLevel;
            DirProp sor, eor;

            /* determine the first sor and set eor to it because of the loop body (sor=eor there) */
            level=GET_PARALEVEL(pBiDi, 0);
            nextLevel=levels[0];
            if(level<nextLevel) {
                eor=GET_LR_FROM_LEVEL(nextLevel);
            } else {
                eor=GET_LR_FROM_LEVEL(level);
            }

            do {
                /* determine start and limit of the run (end points just behind the run) */

                /* the values for this run's start are the same as for the previous run's end */
                start=limit;
                level=nextLevel;
                if((start>0) && (dirProps[start-1]==B)) {
                    /* except if this is a new paragraph, then set sor = para level */
                    sor=GET_LR_FROM_LEVEL(GET_PARALEVEL(pBiDi, start));
                } else {
                    sor=eor;
                }

                /* search for the limit of this run */
                while((++limit<length) &&
                      ((levels[limit]==level) ||
                       (DIRPROP_FLAG(dirProps[limit])&MASK_BN_EXPLICIT))) {}

                /* get the correct level of the next run */
                if(limit<length) {
                    nextLevel=levels[limit];
                } else {
                    nextLevel=GET_PARALEVEL(pBiDi, length-1);
                }

                /* determine eor from max(level, nextLevel); sor is last run's eor */
                if(NO_OVERRIDE(level)<NO_OVERRIDE(nextLevel)) {
                    eor=GET_LR_FROM_LEVEL(nextLevel);
                } else {
                    eor=GET_LR_FROM_LEVEL(level);
                }

                /* if the run consists of overridden directional types, then there
                   are no implicit types to be resolved */
                if(!(level&UBIDI_LEVEL_OVERRIDE)) {
                    resolveImplicitLevels(pBiDi, start, limit, sor, eor);
                } else {
                    /* remove the UBIDI_LEVEL_OVERRIDE flags */
                    do {
                        levels[start++]&=~UBIDI_LEVEL_OVERRIDE;
                    } while(start<limit);
                }
            } while(limit<length);
        }
        /* check if we got any memory shortage while adding insert points */
        if (U_FAILURE(pBiDi->insertPoints.errorCode))
        {
            *pErrorCode=pBiDi->insertPoints.errorCode;
            return;
        }
        /* reset the embedding levels for some non-graphic characters (L1), (X9) */
        adjustWSLevels(pBiDi);
        break;
    }
    /* add RLM for inverse Bidi with contextual orientation resolving
     * to RTL which would not round-trip otherwise
     */
    if((pBiDi->defaultParaLevel>0) &&
       (pBiDi->reorderingOptions & UBIDI_OPTION_INSERT_MARKS) &&
       ((pBiDi->reorderingMode==UBIDI_REORDER_INVERSE_LIKE_DIRECT) ||
        (pBiDi->reorderingMode==UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL))) {
        int32_t i, j, start, last;
        UBiDiLevel level;
        DirProp dirProp;
        for(i=0; i<pBiDi->paraCount; i++) {
            last=(pBiDi->paras[i].limit)-1;
            level= static_cast<UBiDiLevel>(pBiDi->paras[i].level);
            if(level==0)
                continue;           /* LTR paragraph */
            start= i==0 ? 0 : pBiDi->paras[i-1].limit;
            for(j=last; j>=start; j--) {
                dirProp=dirProps[j];
                if(dirProp==L) {
                    if(j<last) {
                        while(dirProps[last]==B) {
                            last--;
                        }
                    }
                    addPoint(pBiDi, last, RLM_BEFORE);
                    break;
                }
                if(DIRPROP_FLAG(dirProp) & MASK_R_AL) {
                    break;
                }
            }
        }
    }

    if(pBiDi->reorderingOptions & UBIDI_OPTION_REMOVE_CONTROLS) {
        pBiDi->resultLength -= pBiDi->controlCount;
    } else {
        pBiDi->resultLength += pBiDi->insertPoints.size;
    }
    setParaSuccess(pBiDi);              /* mark successful setPara */
}

U_CAPI void U_EXPORT2
ubidi_orderParagraphsLTR(UBiDi *pBiDi, UBool orderParagraphsLTR) {
    if(pBiDi!=NULL) {
        pBiDi->orderParagraphsLTR=orderParagraphsLTR;
    }
}

U_CAPI UBool U_EXPORT2
ubidi_isOrderParagraphsLTR(UBiDi *pBiDi) {
    if(pBiDi!=NULL) {
        return pBiDi->orderParagraphsLTR;
    } else {
        return FALSE;
    }
}

U_CAPI UBiDiDirection U_EXPORT2
ubidi_getDirection(const UBiDi *pBiDi) {
    if(IS_VALID_PARA_OR_LINE(pBiDi)) {
        return pBiDi->direction;
    } else {
        return UBIDI_LTR;
    }
}

U_CAPI const UChar * U_EXPORT2
ubidi_getText(const UBiDi *pBiDi) {
    if(IS_VALID_PARA_OR_LINE(pBiDi)) {
        return pBiDi->text;
    } else {
        return NULL;
    }
}

U_CAPI int32_t U_EXPORT2
ubidi_getLength(const UBiDi *pBiDi) {
    if(IS_VALID_PARA_OR_LINE(pBiDi)) {
        return pBiDi->originalLength;
    } else {
        return 0;
    }
}

U_CAPI int32_t U_EXPORT2
ubidi_getProcessedLength(const UBiDi *pBiDi) {
    if(IS_VALID_PARA_OR_LINE(pBiDi)) {
        return pBiDi->length;
    } else {
        return 0;
    }
}

U_CAPI int32_t U_EXPORT2
ubidi_getResultLength(const UBiDi *pBiDi) {
    if(IS_VALID_PARA_OR_LINE(pBiDi)) {
        return pBiDi->resultLength;
    } else {
        return 0;
    }
}

/* paragraphs API functions ------------------------------------------------- */

U_CAPI UBiDiLevel U_EXPORT2
ubidi_getParaLevel(const UBiDi *pBiDi) {
    if(IS_VALID_PARA_OR_LINE(pBiDi)) {
        return pBiDi->paraLevel;
    } else {
        return 0;
    }
}

U_CAPI int32_t U_EXPORT2
ubidi_countParagraphs(UBiDi *pBiDi) {
    if(!IS_VALID_PARA_OR_LINE(pBiDi)) {
        return 0;
    } else {
        return pBiDi->paraCount;
    }
}

U_CAPI void U_EXPORT2
ubidi_getParagraphByIndex(const UBiDi *pBiDi, int32_t paraIndex,
                          int32_t *pParaStart, int32_t *pParaLimit,
                          UBiDiLevel *pParaLevel, UErrorCode *pErrorCode) {
    int32_t paraStart;

    /* check the argument values */
    RETURN_VOID_IF_NULL_OR_FAILING_ERRCODE(pErrorCode);
    RETURN_VOID_IF_NOT_VALID_PARA_OR_LINE(pBiDi, *pErrorCode);
    RETURN_VOID_IF_BAD_RANGE(paraIndex, 0, pBiDi->paraCount, *pErrorCode);

    pBiDi=pBiDi->pParaBiDi;             /* get Para object if Line object */
    if(paraIndex) {
        paraStart=pBiDi->paras[paraIndex-1].limit;
    } else {
        paraStart=0;
    }
    if(pParaStart!=NULL) {
        *pParaStart=paraStart;
    }
    if(pParaLimit!=NULL) {
        *pParaLimit=pBiDi->paras[paraIndex].limit;
    }
    if(pParaLevel!=NULL) {
        *pParaLevel=GET_PARALEVEL(pBiDi, paraStart);
    }
}

U_CAPI int32_t U_EXPORT2
ubidi_getParagraph(const UBiDi *pBiDi, int32_t charIndex,
                          int32_t *pParaStart, int32_t *pParaLimit,
                          UBiDiLevel *pParaLevel, UErrorCode *pErrorCode) {
    int32_t paraIndex;

    /* check the argument values */
    /* pErrorCode will be checked by the call to ubidi_getParagraphByIndex */
    RETURN_IF_NULL_OR_FAILING_ERRCODE(pErrorCode, -1);
    RETURN_IF_NOT_VALID_PARA_OR_LINE(pBiDi, *pErrorCode, -1);
    pBiDi=pBiDi->pParaBiDi;             /* get Para object if Line object */
    RETURN_IF_BAD_RANGE(charIndex, 0, pBiDi->length, *pErrorCode, -1);

    for(paraIndex=0; charIndex>=pBiDi->paras[paraIndex].limit; paraIndex++);
    ubidi_getParagraphByIndex(pBiDi, paraIndex, pParaStart, pParaLimit, pParaLevel, pErrorCode);
    return paraIndex;
}

U_CAPI void U_EXPORT2
ubidi_setClassCallback(UBiDi *pBiDi, UBiDiClassCallback *newFn,
                       const void *newContext, UBiDiClassCallback **oldFn,
                       const void **oldContext, UErrorCode *pErrorCode)
{
    RETURN_VOID_IF_NULL_OR_FAILING_ERRCODE(pErrorCode);
    if(pBiDi==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if( oldFn )
    {
        *oldFn = pBiDi->fnClassCallback;
    }
    if( oldContext )
    {
        *oldContext = pBiDi->coClassCallback;
    }
    pBiDi->fnClassCallback = newFn;
    pBiDi->coClassCallback = newContext;
}

U_CAPI void U_EXPORT2
ubidi_getClassCallback(UBiDi *pBiDi, UBiDiClassCallback **fn, const void **context)
{
    if(pBiDi==NULL) {
        return;
    }
    if( fn )
    {
        *fn = pBiDi->fnClassCallback;
    }
    if( context )
    {
        *context = pBiDi->coClassCallback;
    }
}

U_CAPI UCharDirection U_EXPORT2
ubidi_getCustomizedClass(UBiDi *pBiDi, UChar32 c)
{
    UCharDirection dir;

    if( pBiDi->fnClassCallback == NULL ||
        (dir = (*pBiDi->fnClassCallback)(pBiDi->coClassCallback, c)) == U_BIDI_CLASS_DEFAULT )
    {
        dir = ubidi_getClass(c);
    }
    if(dir >= U_CHAR_DIRECTION_COUNT) {
        dir = (UCharDirection)ON;
    }
    return dir;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2004-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ubidi_props.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004dec30
*   created by: Markus W. Scherer
*
*   Low-level Unicode bidi/shaping properties access.
*/



 /* UDataInfo */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************/


/*----------------------------------------------------------------------------------
 *
 *   UCommonData   An abstract interface for dealing with ICU Common Data Files.
 *                 ICU Common Data Files are a grouping of a number of individual
 *                 data items (resources, converters, tables, anything) into a
 *                 single file or dll.  The combined format includes a table of
 *                 contents for locating the individual items by name.
 *
 *                 Two formats for the table of contents are supported, which is
 *                 why there is an abstract inteface involved.
 *
 *                 These functions are part of the ICU internal implementation, and
 *                 are not inteded to be used directly by applications.
 */



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************/

/*----------------------------------------------------------------------------------
 *
 *       Memory mapped file wrappers for use by the ICU Data Implementation
 *
 *           Porting note:  The implementation of these functions is very platform specific.
 *             Not all platforms can do real memory mapping.  Those that can't
 *             still must implement these functions, getting the data into memory using
 *             whatever means are available.
 *
 *            These functions are part of the ICU internal implementation, and
 *            are not inteded to be used directly by applications.
 *
 *----------------------------------------------------------------------------------*/






U_CFUNC UBool uprv_mapFile(UDataMemory *pdm, const char *path, UErrorCode *status);
U_CFUNC void  uprv_unmapFile(UDataMemory *pData);

/* MAP_NONE: no memory mapping, no file access at all */
#define MAP_NONE        0
#define MAP_WIN32       1
#define MAP_POSIX       2
#define MAP_STDIO       3
#define MAP_390DLL      4

#if UCONFIG_NO_FILE_IO
#   define MAP_IMPLEMENTATION MAP_NONE
#elif U_PLATFORM_USES_ONLY_WIN32_API
#   define MAP_IMPLEMENTATION MAP_WIN32
#elif U_HAVE_MMAP || U_PLATFORM == U_PF_OS390
#   if U_PLATFORM == U_PF_OS390 && defined (OS390_STUBDATA)
        /*   No memory mapping for 390 batch mode.  Fake it using dll loading.  */
#       define MAP_IMPLEMENTATION MAP_390DLL
#   else
#       define MAP_IMPLEMENTATION MAP_POSIX
#   endif
#else /* unknown platform, no memory map implementation: use stdio.h and uprv_malloc() instead */
#   define MAP_IMPLEMENTATION MAP_STDIO
#endif




#define COMMON_DATA_NAME U_ICUDATA_NAME

typedef struct  {
    uint16_t    headerSize;
    uint8_t     magic1;
    uint8_t     magic2;
} MappedData;


typedef struct  {
    MappedData  dataHeader;
    UDataInfo   info;
} DataHeader;

typedef struct {
    uint32_t nameOffset;
    uint32_t dataOffset;
} UDataOffsetTOCEntry;

typedef struct {
    uint32_t count;
    /**
     * Variable-length array declared with length 1 to disable bounds checkers.
     * The actual array length is in the count field.
     */
    UDataOffsetTOCEntry entry[1];
} UDataOffsetTOC;

/**
 * Get the header size from a const DataHeader *udh.
 * Handles opposite-endian data.
 *
 * @internal
 */
U_CFUNC uint16_t
udata_getHeaderSize(const DataHeader *udh);

/**
 * Get the UDataInfo.size from a const UDataInfo *info.
 * Handles opposite-endian data.
 *
 * @internal
 */
U_CFUNC uint16_t
udata_getInfoSize(const UDataInfo *info);

U_CDECL_BEGIN
/*
 *  "Virtual" functions for data lookup.
 *  To call one, given a UDataMemory *p, the code looks like this:
 *     p->vFuncs.Lookup(p, tocEntryName, pErrorCode);
 *          (I sure do wish this was written in C++, not C)
 */

typedef const DataHeader *
(U_CALLCONV * LookupFn)(const UDataMemory *pData,
                        const char *tocEntryName,
                        int32_t *pLength,
                        UErrorCode *pErrorCode);

typedef uint32_t
(U_CALLCONV * NumEntriesFn)(const UDataMemory *pData);

U_CDECL_END

typedef struct {
    LookupFn      Lookup;
    NumEntriesFn  NumEntries; 
} commonDataFuncs;


/*
 *  Functions to check whether a UDataMemory refers to memory containing 
 *     a recognizable header and table of contents a Common Data Format
 *
 *     If a valid header and TOC are found,
 *         set the CommonDataFuncs function dispatch vector in the UDataMemory
 *             to point to the right functions for the TOC type.
 *     otherwise
 *         set an errorcode.
 */
U_CFUNC void udata_checkCommonData(UDataMemory *pData, UErrorCode *pErrorCode);

 /* DataHeader */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2010, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************/


/*----------------------------------------------------------------------------------
 *
 *  UDataMemory     A class-like struct that serves as a handle to a piece of memory
 *                  that contains some ICU data (resource, converters, whatever.)
 *
 *                  When an application opens ICU data (with udata_open, for example,
 *                  a UDataMemory * is returned.
 *
 *----------------------------------------------------------------------------------*/




struct UDataMemory {
    const commonDataFuncs  *vFuncs;      /* Function Pointers for accessing TOC             */

    const DataHeader *pHeader;     /* Header of the memory being described by this    */
                                   /*   UDataMemory object.                           */
    const void       *toc;         /* For common memory, table of contents for        */
                                   /*   the pieces within.                            */
    UBool             heapAllocated;  /* True if this UDataMemory Object is on the    */
                                   /*  heap and thus needs to be deleted when closed. */

    void             *mapAddr;     /* For mapped or allocated memory, the start addr. */
                                   /* Only non-null if a close operation should unmap */
                                   /*  the associated data.                           */
    void             *map;         /* Handle, or other data, OS dependent.            */
                                   /* Only non-null if a close operation should unmap */
                                   /*  the associated data, and additional info       */
                                   /*   beyond the mapAddr is needed to do that.      */
    int32_t           length;      /* Length of the data in bytes; -1 if unknown.     */
};

U_CFUNC UDataMemory *UDataMemory_createNewInstance(UErrorCode *pErr);
U_CFUNC void         UDatamemory_assign  (UDataMemory *dest, UDataMemory *source);
U_CFUNC void         UDataMemory_init    (UDataMemory *This);
U_CFUNC UBool        UDataMemory_isLoaded(const UDataMemory *This);
U_CFUNC void         UDataMemory_setData (UDataMemory *This, const void *dataAddr);

U_CFUNC const DataHeader *UDataMemory_normalizeDataPointer(const void *p);

U_CAPI int32_t U_EXPORT2
udata_getLength(const UDataMemory *pData);

U_CAPI const void * U_EXPORT2
udata_getRawMemory(const UDataMemory *pData);








struct UBiDiProps {
    UDataMemory *mem;
    const int32_t *indexes;
    const uint32_t *mirrors;
    const uint8_t *jgArray;
    const uint8_t *jgArray2;

    UTrie2 trie;
    uint8_t formatVersion[4];
};

/* ubidi_props_data.h is machine-generated by genbidi --csource */
#define INCLUDED_FROM_UBIDI_PROPS_C
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// Copyright (C) 1999-2016, International Business Machines
// Corporation and others.  All Rights Reserved.
//
// file name: ubidi_props_data.h
//
// machine-generated by: icu/tools/unicode/c/genprops/bidipropsbuilder.cpp


#ifdef INCLUDED_FROM_UBIDI_PROPS_C

// static const UVersionInfo ubidi_props_dataVersion={0xd,0,0,0};

static const int32_t ubidi_props_indexes[UBIDI_IX_TOP]={0x10,0x67ec,0x6200,0x28,0x620,0x8c8,0x10ac0,0x10d24,0,0,0,0,0,0,0,0x6502b6};

static const uint16_t ubidi_props_trieIndex[12536]={
0x37c,0x384,0x38c,0x394,0x3ac,0x3b4,0x3bc,0x3c4,0x39c,0x3a4,0x39c,0x3a4,0x39c,0x3a4,0x39c,0x3a4,
0x39c,0x3a4,0x39c,0x3a4,0x3ca,0x3d2,0x3da,0x3e2,0x3ea,0x3f2,0x3ee,0x3f6,0x3fe,0x406,0x401,0x409,
0x39c,0x3a4,0x39c,0x3a4,0x411,0x419,0x39c,0x3a4,0x39c,0x3a4,0x39c,0x3a4,0x41f,0x427,0x42f,0x437,
0x43f,0x447,0x44f,0x457,0x45d,0x465,0x46d,0x475,0x47d,0x485,0x48b,0x493,0x49b,0x4a3,0x4ab,0x4b3,
0x4bf,0x4bb,0x4c7,0x4cf,0x431,0x4df,0x4e6,0x4d7,0x4ee,0x4f0,0x4f8,0x500,0x508,0x509,0x511,0x519,
0x521,0x509,0x529,0x52e,0x521,0x509,0x536,0x53e,0x508,0x546,0x54e,0x500,0x556,0x39c,0x55e,0x562,
0x56a,0x56c,0x574,0x57c,0x508,0x584,0x58c,0x500,0x413,0x590,0x511,0x500,0x508,0x39c,0x598,0x39c,
0x39c,0x59e,0x5a6,0x39c,0x39c,0x5aa,0x5b2,0x39c,0x5b6,0x5bd,0x39c,0x5c5,0x5cd,0x5d4,0x555,0x39c,
0x39c,0x5dc,0x5e4,0x5ec,0x5f4,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x5fc,0x39c,0x604,0x39c,0x39c,0x39c,
0x60c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x614,0x39c,0x39c,0x39c,0x61c,0x61c,0x515,0x515,0x39c,0x622,0x62a,0x604,
0x640,0x632,0x632,0x648,0x64f,0x638,0x39c,0x39c,0x39c,0x657,0x65f,0x39c,0x39c,0x39c,0x661,0x669,
0x671,0x39c,0x678,0x680,0x39c,0x688,0x56b,0x39c,0x545,0x690,0x556,0x698,0x413,0x6a0,0x39c,0x6a7,
0x39c,0x6ac,0x39c,0x39c,0x39c,0x39c,0x6b2,0x6ba,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x3ea,0x6c2,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x6ca,0x6d2,0x6d6,
0x6ee,0x6f4,0x6de,0x6e6,0x6fc,0x704,0x708,0x5d7,0x710,0x718,0x720,0x39c,0x728,0x669,0x669,0x669,
0x738,0x740,0x748,0x750,0x755,0x75d,0x765,0x730,0x76d,0x775,0x39c,0x77b,0x782,0x669,0x669,0x669,
0x669,0x582,0x788,0x669,0x790,0x39c,0x39c,0x666,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,
0x669,0x669,0x669,0x669,0x669,0x798,0x669,0x669,0x669,0x669,0x669,0x79e,0x669,0x669,0x7a6,0x7ae,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x669,0x669,0x669,0x669,0x7be,0x7c6,0x7ce,0x7b6,
0x7de,0x7e6,0x7ee,0x7f5,0x7fc,0x804,0x808,0x7d6,0x669,0x669,0x669,0x810,0x816,0x669,0x669,0x81c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x824,0x39c,0x39c,0x39c,0x82c,0x39c,0x39c,0x39c,0x3ea,
0x834,0x83c,0x840,0x39c,0x848,0x669,0x669,0x66c,0x669,0x669,0x669,0x669,0x669,0x669,0x84f,0x855,
0x865,0x85d,0x39c,0x39c,0x86d,0x60c,0x39c,0x3c3,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x669,0x82b,
0x3d1,0x39c,0x875,0x87d,0x39c,0x885,0x88d,0x39c,0x39c,0x39c,0x39c,0x891,0x39c,0x39c,0x661,0x3c2,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x669,0x669,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x875,0x669,0x582,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x898,0x39c,0x39c,0x89d,0x56c,0x39c,0x39c,0x5b8,0x669,0x660,0x39c,0x39c,0x8a5,0x39c,0x39c,0x39c,
0x8ad,0x8b4,0x632,0x8bc,0x39c,0x39c,0x58e,0x8c4,0x39c,0x8cc,0x8d3,0x39c,0x4ee,0x8d8,0x39c,0x507,
0x39c,0x8e0,0x8e8,0x509,0x39c,0x8ec,0x508,0x8f4,0x39c,0x39c,0x39c,0x8fa,0x39c,0x39c,0x39c,0x901,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x915,0x909,0x90d,0x49b,0x49b,0x49b,0x49b,0x49b,
0x49b,0x49b,0x49b,0x49b,0x49b,0x49b,0x49b,0x49b,0x49b,0x91d,0x49b,0x49b,0x49b,0x49b,0x925,0x929,
0x931,0x939,0x93d,0x945,0x49b,0x49b,0x49b,0x949,0x951,0x38c,0x959,0x961,0x39c,0x39c,0x39c,0x969,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0xe70,0xe70,0xeb0,0xef0,0xe70,0xe70,0xe70,0xe70,0xe70,0xe70,0xf28,0xf68,0xfa8,0xfb8,0xff8,0x1004,
0xe70,0xe70,0x1044,0xe70,0xe70,0xe70,0x107c,0x10bc,0x10fc,0x113c,0x1174,0x11b4,0x11f4,0x122c,0x126c,0x12ac,
0xa40,0xa80,0xac0,0xaff,0x1a0,0x1a0,0xb3f,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xb68,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xba8,0x1a0,0x1a0,0xbdd,0xc1d,0xc5d,0xc9d,0xcdd,0xd1d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0xd9d,0xdad,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,
0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0x1a0,0xd5d,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x971,0x39c,0x669,0x669,0x979,0x60c,0x39c,0x501,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x981,0x39c,0x39c,0x39c,0x988,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x990,0x431,0x431,0x431,0x431,0x431,0x431,0x431,
0x998,0x99c,0x431,0x431,0x431,0x431,0x9ac,0x9a4,0x431,0x9b4,0x431,0x431,0x9bc,0x9c2,0x431,0x431,
0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x9d2,0x9ca,0x431,0x431,0x431,0x431,0x431,0x431,
0x431,0x431,0x431,0x9da,0x431,0x9e2,0x431,0x431,0x431,0x9e6,0x9ed,0x9f3,0x431,0x9f7,0x9ff,0x431,
0x508,0xa07,0xa0e,0xa15,0x413,0xa18,0x39c,0x39c,0x4ee,0xa1f,0x39c,0xa25,0x413,0xa2a,0xa32,0x39c,
0x39c,0xa37,0x39c,0x39c,0x39c,0x39c,0x82c,0xa3f,0x413,0x590,0x56b,0xa46,0x39c,0x39c,0x39c,0x39c,
0x39c,0xa07,0xa4e,0x39c,0x39c,0xa56,0xa5e,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0xa62,0xa6a,0x39c,
0x39c,0xa72,0x56b,0xa7a,0x39c,0xa80,0x39c,0x39c,0x5fc,0xa88,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0xa8d,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0xa94,0xa9c,0x39c,0x39c,0x39c,0xa9f,0x56b,0xaa7,
0xaab,0xab3,0x39c,0xaba,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0xac1,0x39c,0x39c,0xacf,0xac9,0x39c,0x39c,0x39c,0xad7,0xadf,0x39c,0xae3,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x592,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0xaf0,0xaeb,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0xaf8,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0xaff,
0x39c,0xb05,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0xa26,0x39c,0xb0b,0x39c,0x39c,0xb13,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x522,0xb1b,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0xb22,0xb2a,0xb30,0x39c,0x39c,0x669,0x669,0xb38,
0x39c,0x39c,0x39c,0x39c,0x39c,0x669,0x669,0x83f,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0xb3a,0x39c,0xb41,0x39c,0xb3d,0x39c,0xb44,0x39c,0xb4c,0xb50,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x3ea,0xb58,0x3ea,
0xb5f,0xb66,0xb6e,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0xb76,0xb7e,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0xb05,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0xb83,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x431,0x431,0x431,
0x431,0x431,0x431,0xb8b,0x431,0xb93,0xb93,0xb9a,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,
0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,
0x90d,0x49b,0x49b,0x431,0x431,0x49b,0x49b,0x9f3,0x431,0x431,0x431,0x431,0x431,0x49b,0x49b,0x49b,
0x49b,0x49b,0x49b,0x49b,0xba2,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x431,0x669,0xbaa,0x669,
0x669,0x66c,0xbaf,0xbb3,0x84f,0xbbb,0x3be,0x39c,0xbc1,0x39c,0xbc6,0x39c,0x39c,0x39c,0x39c,0x39c,
0x779,0x39c,0x39c,0x39c,0x39c,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,
0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,0x669,
0x669,0x669,0x669,0x66b,0x979,0x669,0x669,0x669,0x66c,0x669,0x669,0xbce,0x66e,0xbaa,0x669,0xbd6,
0x669,0xbde,0xbe3,0x39c,0x39c,0x669,0x669,0x669,0xbeb,0x669,0x669,0x798,0x669,0x669,0x669,0x66c,
0xbf2,0xbfa,0xc00,0xc05,0x39c,0x669,0x669,0x669,0x669,0xc0d,0x669,0x788,0xc15,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0xc1c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,
0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0x39c,0xc1c,0xc2c,0xc24,0xc24,
0xc24,0xc2d,0xc2d,0xc2d,0xc2d,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0xc35,0xc2d,0xc2d,0xc2d,
0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,
0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,
0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,
0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0xc2d,0x37b,0x37b,0x37b,
0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,8,7,8,9,7,0x12,0x12,
0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,7,7,7,8,
9,0xa,0xa,4,4,4,0xa,0xa,0x310a,0xf20a,0xa,3,6,3,6,6,
2,2,2,2,2,2,2,2,2,2,6,0xa,0x500a,0xa,0xd00a,0xa,
0xa,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0x510a,0xa,0xd20a,0xa,0xa,
0xa,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0x510a,0xa,0xd20a,0xa,0x12,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x12,0x12,0x12,0x12,0x12,7,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,
0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,
6,0xa,4,4,4,4,0xa,0xa,0xa,0xa,0,0x900a,0xa,0xb2,0xa,0xa,
4,4,2,2,0xa,0,0xa,0xa,0xa,2,0,0x900a,0xa,0xa,0xa,0xa,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0xa,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0xa,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0xa,0xa,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0,0,0,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0xa,0xa,0,0,
0,0,0,0,0,0,0xa,0,0,0,0,0,0xa,0xa,0,0xa,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0xa,0,0,0,0,0,
0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0xa,0,0,0xa,0xa,4,1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,1,0xb1,1,0xb1,0xb1,1,
0xb1,0xb1,1,0xb1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,5,5,5,5,
5,5,0xa,0xa,0xd,4,4,0xd,6,0xd,0xa,0xa,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xd,0x8ad,0xd,0xd,0xd,0x4d,0xd,0x8d,0x8d,
0x8d,0x8d,0x4d,0x8d,0x4d,0x8d,0x4d,0x4d,0x4d,0x4d,0x4d,0x8d,0x8d,0x8d,0x8d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x2d,0x4d,0x4d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x8d,0x4d,0x4d,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,5,5,5,5,
5,5,5,5,5,5,4,5,5,0xd,0x4d,0x4d,0xb1,0x8d,0x8d,0x8d,
0xd,0x8d,0x8d,0x8d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x8d,0x8d,0x8d,0x8d,
0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x4d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x8d,0x4d,0x4d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,
0x4d,0x8d,0x4d,0x8d,0x4d,0x4d,0x8d,0x8d,0xd,0x8d,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,5,0xa,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xd,0xd,0xb1,0xb1,0xa,0xb1,0xb1,
0xb1,0xb1,0x8d,0x8d,2,2,2,2,2,2,2,2,2,2,0x4d,0x4d,
0x4d,0xd,0xd,0x4d,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xad,0x8d,0xb1,0x4d,0x4d,0x4d,0x8d,0x8d,0x8d,0x8d,0x8d,0x4d,0x4d,
0x4d,0x4d,0x8d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x8d,0x4d,0x8d,0x4d,
0x8d,0x4d,0x4d,0x8d,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xd,0xd,0x8d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x4d,0x8d,0x8d,0x8d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x8d,0x8d,0x4d,0x4d,0x4d,0x4d,0x8d,0x4d,0x8d,
0x8d,0x4d,0x4d,0x4d,0x8d,0x8d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xd,0xd,0xd,
0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,1,1,1,1,
1,1,1,1,1,1,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,
0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,
0x41,0x41,0x41,0x41,0x41,0x41,0x41,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
1,1,0xa,0xa,0xa,0xa,0x21,1,1,0xb1,1,1,0xb1,0xb1,0xb1,0xb1,
1,0xb1,0xb1,0xb1,1,0xb1,0xb1,0xb1,0xb1,0xb1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0xb1,0xb1,0xb1,0xb1,1,0xb1,0xb1,0xb1,0xb1,0xb1,0x81,0x41,0x41,0x41,
0x41,0x41,0x81,0x81,0x41,0x81,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,
0x81,0x41,0x81,0x81,0x81,0xb1,0xb1,0xb1,1,1,1,1,0x4d,0xd,0x4d,0x4d,
0x4d,0x4d,0xd,0x8d,0x4d,0x8d,0x8d,0xd,0xd,0xd,0xd,0xd,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,0xb1,0xb1,5,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0x4d,0x4d,0x4d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x8d,0x8d,0x8d,0xd,0x8d,0x4d,0x4d,0x8d,0x8d,0x4d,
0x4d,0xd,0x4d,0x4d,0x4d,0x8d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,
0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0xb1,0,0xb1,0,0,0,
0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0xb1,0,0,
0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,
0,0,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xb1,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,
0,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0xb1,0xb1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,4,4,0,0,0,0,0,0,0,4,
0,0,0xb1,0,0,0xb1,0xb1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0xb1,0xb1,0,0,0,0,0xb1,0xb1,0,0,0xb1,
0xb1,0xb1,0,0,0,0xb1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0xb1,0xb1,0,0,0,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,
0xb1,0,0,0,0,0xb1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0,0,0,0,
0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,
0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xb1,0,0,0xb1,0,0xb1,0xb1,0xb1,0xb1,0,0,0,
0,0,0,0,0,0xb1,0,0,0,0,0,0,0,0xb1,0xb1,0,
0,0,0,0,0,0,0,0,0,0,0xb1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0xb1,0,0,0,0,0,0,0,
0,0,0,0,0,0xb1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,
0xa,4,0xa,0,0,0,0,0,0xb1,0,0,0,0xb1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,
0xb1,0,0,0,0,0,0xb1,0xb1,0xb1,0,0xb1,0xb1,0xb1,0xb1,0,0,
0,0,0,0,0,0xb1,0xb1,0,0,0,0,0,0,0,0,0,
0,0,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0,0,0xa0,
0,0,0,0,0,0,0xa0,0,0,0,0,0,0xb1,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0xb1,0,0,0,0,0,
0,0,0xb1,0xb1,0xb1,0,0xb1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0xb1,0,0,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0,0,0,0,4,0,0,0,0,0,0,0,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0xb1,0,0,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0xb1,0,0xb1,0,0xb1,0x310a,0xf20a,
0x310a,0xf20a,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0,0,0,0,
0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,
0xb1,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0,0,0xb1,0xb1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0xb1,0xb1,0,0,0,0,0xb1,0xb1,
0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0xb1,0,0,0xb1,0xb1,0,0,0,0,0,0,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0,0,0,0,
0xa,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0x310a,0xf20a,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0xb1,0,
0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,
0,0,0,4,0,0xb1,0,0,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,
0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,
0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0xb1,0x40,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x4a,0xa,0xa,0x2a,0xb1,0xb1,0xb1,0x12,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,
0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0,0,0,0,0,0,0,
0,0xb1,0xb1,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,
0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0xb1,0xb1,0xb1,0,
0,0,0,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0xb1,0,
0,0,0,0,0,0xb1,0xb1,0xb1,0,0,0,0,0xa,0,0,0,
0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0,0,0xb1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0xb1,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,
0xb1,0,0xb1,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,
0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0xb1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xb1,0,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0xb1,0xb1,0xb1,0xb1,0,0,0xb1,0xb1,0,0xb1,0xb1,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0xb1,0,0xb1,0xb1,0,0,0,0xb1,0,0xb1,0xb1,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0,0,0xb1,0xb1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0,0,0,0,0xb1,0,0,0,0,0,0,0xb1,0,0,0,
0xb1,0xb1,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0xa,0,0xa,0xa,0xa,0,0,0,0,0,0,
0,0,0,0,0,0xa,0xa,0xa,0,0,0,0,0,0,0,0,
0,0,0,0,0,0xa,0xa,0xa,0,0,0,0,0,0,0,0,
0,0,0,0,0,0xa,0xa,0,0xa,0xa,0xa,0xa,6,0x310a,0xf20a,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,9,0xb2,0xb2,0xb2,0xb2,0xb2,0x12,0x814,0x815,
0x813,0x816,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,2,0,0,0,2,2,2,2,
2,2,3,3,0xa,0x310a,0xf20a,0,9,9,9,9,9,9,9,9,
9,9,9,0xb2,0x412,0x432,0x8a0,0x8a1,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,9,7,0x8ab,0x8ae,0x8b0,0x8ac,0x8af,6,
4,4,4,4,4,0xa,0xa,0xa,0xa,0x300a,0xf00a,0xa,0xa,0xa,0xa,0xa,
2,2,2,2,2,2,2,2,2,2,3,3,0xa,0x310a,0xf20a,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xa,0xa,0,0xa,0xa,0xa,0xa,0,0xa,0xa,0,0,0,0,0,0,
0,0,0,0,0xa,0,0xa,0xa,0xa,0,0,0,0,0,0xa,0xa,
0xa,0xa,0xa,0xa,0,0xa,0,0xa,0,0xa,0,0,0,0,4,0,
0,0,0,0,0,0,0,0,0,0,0xa,0xa,0,0,0,0,
0x100a,0xa,0xa,0xa,0xa,0,0,0,0,0,0xa,0xa,0xa,0xa,0,0,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0,0,0,0,0,0,0,0,0,0xa,0xa,0xa,0,0,0,0,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0xa,0xa,
0x300a,0xf00a,0x900a,0x900a,0x900a,0x100a,0x900a,0x900a,0x100a,0x100a,0x900a,0x900a,0x900a,0x900a,0x900a,0x100a,
0xa,0x100a,0x100a,0x100a,0x100a,0xa,0xa,0xa,0x700a,0x700a,0x700a,0xb00a,0xb00a,0xb00a,0xa,0xa,
0xa,0x100a,3,4,0xa,0x900a,0x100a,0xa,0xa,0xa,0x100a,0x100a,0x100a,0x100a,0xa,0x900a,
0x900a,0x900a,0x900a,0xa,0x900a,0xa,0x100a,0xa,0xa,0xa,0xa,0x100a,0x100a,0x100a,0x100a,0x100a,
0x100a,0x100a,0x100a,0x100a,0xa,0xa,0xa,0xa,0xa,0x100a,0xa,0x100a,0x300a,0xf00a,0x100a,0x100a,
0x100a,0x100a,0x100a,0x900a,0x100a,0x900a,0x100a,0x100a,0x100a,0x100a,0x100a,0x100a,0x900a,0xa,0xa,0xa,
0xa,0xa,0x300a,0xf00a,0x300a,0xf00a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x100a,
0x100a,0xa,0x100a,0xa,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0xa,0xa,0x300a,0xf00a,
0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,
0x100a,0xa,0xa,0x300a,0xf00a,0x300a,0xf00a,0xa,0xa,0xa,0xa,0xa,0x900a,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0x300a,0xf00a,0xa,0xa,0x900a,0x100a,0x900a,0x900a,0x100a,0x900a,
0x100a,0x100a,0x100a,0x100a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x900a,0xa,0xa,0xa,
0xa,0xa,0x100a,0x100a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x300a,0xf00a,0x300a,
0xf00a,0x900a,0xa,0xa,0x300a,0xf00a,0xa,0xa,0xa,0xa,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,
0x300a,0xf00a,0x300a,0xf00a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x310a,0xf20a,0x310a,0xf20a,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0x100a,0x100a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x310a,0xf20a,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,0,0,0,0,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x100a,0xa,0xa,0x300a,0xf00a,0x310a,0xf20a,0xa,
0x300a,0xf00a,0xa,0x500a,0x100a,0xd00a,0xa,0xa,0xa,0xa,0xa,0x100a,0x100a,0x300a,0xf00a,0xa,
0xa,0xa,0xa,0xa,0x900a,0x300a,0xf00a,0xa,0xa,0xa,0x300a,0xf00a,0x300a,0xf00a,0x310a,0xf20a,
0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x100a,0xa,0x100a,0x100a,0x100a,0xa,0xa,
0x300a,0xf00a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x100a,0x900a,0x100a,0x100a,
0x300a,0xf00a,0xa,0xa,0x310a,0xf20a,0xa,0xa,0xa,0xa,0xa,0x310a,0xf20a,0x310a,0xf20a,0x310a,
0xf20a,0x310a,0xf20a,0x310a,0xf20a,0x710a,0x320a,0xf10a,0xb20a,0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0x310a,
0xf20a,0xa,0xa,0x900a,0x100a,0x100a,0x100a,0x100a,0x900a,0xa,0x100a,0x900a,0x300a,0xf00a,0x100a,0x100a,
0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0x900a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x300a,0xf00a,0x100a,0x100a,0x300a,0xf00a,0xa,0xa,
0xa,0x100a,0xa,0xa,0xa,0xa,0x100a,0x300a,0xf00a,0x300a,0xf00a,0xa,0x300a,0xf00a,0xa,0xa,
0x310a,0xf20a,0x310a,0xf20a,0x100a,0xa,0xa,0xa,0xa,0xa,0x100a,0x900a,0x900a,0x900a,0x100a,0xa,
0xa,0xa,0xa,0xa,0x300a,0xf00a,0x900a,0xa,0xa,0xa,0xa,0x100a,0xa,0xa,0xa,0x300a,
0xf00a,0x300a,0xf00a,0x100a,0xa,0x100a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0x100a,0x100a,0x100a,0x100a,0x100a,0x100a,0x100a,0x100a,0x100a,0x100a,0x100a,0x100a,0x100a,0x100a,
0x100a,0x100a,0x100a,0x100a,0x100a,0xa,0x100a,0x100a,0x100a,0x100a,0xa,0xa,0x100a,0xa,0x100a,0xa,
0xa,0x100a,0xa,0x300a,0xf00a,0x300a,0xf00a,0xa,0xa,0xa,0xa,0xa,0x300a,0xf00a,0xa,0xa,
0xa,0xa,0xa,0xa,0x300a,0xf00a,0x100a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x100a,
0x100a,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x300a,0xf00a,0xa,0xa,0xa,0xa,0x100a,0x100a,
0x100a,0x100a,0xa,0x100a,0x100a,0xa,0xa,0x100a,0x100a,0xa,0xa,0xa,0xa,0x300a,0xf00a,0x300a,
0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,
0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,
0xf00a,0x300a,0xf00a,0x100a,0xa,0xa,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0xa,0x300a,
0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,
0xf00a,0x300a,0xf00a,0x300a,0xf00a,0x300a,0xf00a,0xa,0xa,0xa,0xa,0xa,0x100a,0xa,0x900a,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x900a,0xa,
0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0,0,0xb1,
0xb1,0xb1,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,
0xa,0xa,0x300a,0xf00a,0x300a,0xf00a,0xa,0xa,0xa,0x300a,0xf00a,0xa,0x300a,0xf00a,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0x300a,0xf00a,0xa,0xa,
0x300a,0xf00a,0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0,0,0,0,0xa,0,0,0,0,0,0,0,0,0,0xb1,0xb1,
0xb1,0xb1,0,0,0xa,0,0,0,0,0,0xa,0xa,0,0,0,0,
0,0xa,0xa,0xa,9,0xa,0xa,0xa,0xa,0,0,0,0x310a,0xf20a,0x310a,0xf20a,
0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0xa,0xa,0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,0x310a,0xf20a,
0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xa,
0xa,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,
0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0xa,0xa,0xa,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0xb1,0xb1,0xb1,0xb1,0xa,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xa,0xa,0,0,0,0,0,0,0,0,0xa,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0xb1,0,0,0,0xb1,0,0,0,0,0xb1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0xb1,0xb1,0,0xa,0xa,0xa,0xa,0xb1,0,0,0,
0,0,0,0,0,0,0,0,4,4,0,0,0,0,0,0,
0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,
0x40,0x40,0x60,0,0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,0,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,
0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0xb1,0,0,0xb1,0xb1,0xb1,0xb1,0,0,0xb1,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,
0,0xb1,0xb1,0,0,0xb1,0xb1,0,0,0,0,0,0,0,0,0,
0,0,0,0xb1,0,0,0,0,0,0,0,0,0xb1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xb1,0,0xb1,0xb1,0xb1,0,0,0xb1,0xb1,0,0,0,0,0,0xb1,0xb1,
0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0,0,
0,0,0,0,0,0,0xb1,0,0,0,0,0,0,0,0,0,
0,0,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0xb1,0,0,0xb1,0,0,0,
0,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1,1,1,1,1,1,1,1,1,3,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xd,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1,0xb1,1,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xa,0xa,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xd,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,
0x12,0x12,0x12,0x12,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xa,0xd,0xd,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,
0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,6,0xa,6,0,0xa,6,0xa,0xa,0xa,0x310a,0xf20a,0x310a,
0xf20a,0x310a,0xf20a,4,0xa,0xa,3,3,0x300a,0xf00a,0xa,0,0xa,4,4,0xa,
0,0,0,0,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xb2,0,0xa,0xa,4,4,4,0xa,0xa,0x310a,0xf20a,0xa,3,
6,3,6,6,2,2,2,2,2,2,2,2,2,2,6,0xa,
0x500a,0xa,0xd00a,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0x510a,
0xa,0xd20a,0xa,0x310a,0xf20a,0xa,0x310a,0xf20a,0xa,0xa,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,4,4,0xa,0xa,0xa,4,4,0,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0xaa,0xaa,0xaa,
0xa,0xa,0x12,0x12,0,0xa,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0,0,0,0xb1,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0xa,
1,0xb1,0xb1,0xb1,1,0xb1,0xb1,1,1,1,1,1,0xb1,0xb1,0xb1,0xb1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0xb1,0xb1,0xb1,1,1,1,1,0xb1,
0x41,0x81,1,1,0x81,0xb1,0xb1,1,1,1,1,0x41,0x41,0x41,0x41,0x81,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
0x41,0x41,0x41,0x41,0x41,0x81,1,0x81,1,0x81,0x81,1,1,0x61,0x81,0x81,
0x81,0x81,0x81,0x41,0x41,0x41,0x41,0x61,0x41,0x41,0x41,0x41,0x41,0x81,0x41,0x41,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0x41,0x81,0x41,0x81,0x81,0x81,0x41,0x41,0x41,0x81,0x41,0x41,0x81,0x41,0x81,0x81,
0x41,0x81,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,0x81,0x81,0x81,0x81,0x41,0x41,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0x4d,0x4d,0x8d,0x4d,0xb1,0xb1,0xb1,0xb1,
0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,5,5,5,5,5,5,5,5,
5,5,0xd,0xd,0xd,0xd,0xd,0xd,0x6d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,1,1,1,1,1,1,1,1,1,
1,1,1,0xb1,0xb1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0x4d,0x4d,0x4d,0x8d,0x4d,0x4d,0x4d,0x4d,
0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0x4d,0xd,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0x4d,0x4d,0x4d,0x8d,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,0x41,1,0x41,0x41,
0x81,0x81,0x81,1,0x41,0x81,0x81,0x41,0x41,0x81,0x41,0x41,1,0x41,0x81,0x81,
0x41,1,1,1,1,0x81,0x41,0x61,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,
0,0,0,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0xb1,0xb1,0xb1,0xb1,0,0,0xb1,0xb1,0,0,0,0,0,
0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0xb1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0,
0,0xb1,0xb1,0xb1,0xb1,0,0,0xb1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0,0,
0xb1,0,0xb1,0xb1,0,0,0,0,0,0,0xb1,0,0,0,0,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0,0xb1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0xb1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0,0xb1,0,0,0,0,0xb1,0xb1,0,0xb1,0xb1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0,0,
0,0,0,0,0xb1,0xb1,0,0xb1,0xb1,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0,0,0xb1,0,0xb1,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xb1,0,0xb1,0,0,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0,0,0,0,0,0,0,0,
0,0,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0,0xb1,0,
0,0,0,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xb1,0xb1,0xb1,0xb1,0,0,0xb1,0xb1,0,0,0,0,0,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xa0,0xa0,0xb1,0xb1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,
0,0,0,0xb1,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0,0,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,
0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,
0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0,0xa0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0,0xb1,0xb1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0,0,0,0xb1,0,0xb1,0xb1,0,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0,0,
0,0xb1,0,0xb1,0,0,0,0,0,0,0,0,4,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,4,4,4,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xa0,0xa0,0xa0,0xa0,0xa0,0xa0,0xa0,0xa0,0xa0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,
0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0xa,0,
0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0xb2,0xb2,0xb2,0xb2,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0xb1,
0xb1,0xb1,0,0,0,0,0,0,0,0,0,0xb2,0xb2,0xb2,0xb2,0xb2,
0xb2,0xb2,0xb2,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xa,0xa,0xb1,0xb1,0xb1,0xa,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0x100a,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0x100a,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0x100a,0,0,0,0,0,0,0,0,0,0,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,
0,0xb1,0,0,0,0,0,0,0,0,0,0,0xb1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,
0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0,0,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0,0xb1,0xb1,0,0xb1,0xb1,
0xb1,0xb1,0xb1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0xb1,0xb1,0xb1,0xb1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,4,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,1,1,1,1,1,1,1,1,1,0x41,0x41,0x41,0x41,
0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,
0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0x41,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xa1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xa,0xa,0xd,0xd,0xd,0xd,0xd,0xd,
0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xd,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,2,2,2,2,
2,2,2,2,2,2,2,0xa,0xa,0xa,0xa,0xa,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0xa,0xa,
0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0xa,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0xa,0xa,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0,0,0xa,0xa,0xa,0xa,0xa,0,0,0,
0xa,0xa,0xa,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,
0,0,0,0,0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0,0,0,0,
0,0,0,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0xa,0xa,0xa,0,0xa,0xa,0xa,0xa,0xa,0xa,0xa,0xa,
0xa,0xa,0xa,0xa,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,2,2,2,2,2,2,2,2,2,2,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0x12,0x12,
0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,
0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,0xb2,
0x12,0xb2,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,
0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,
0x12,0x12,0x12,0x12,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,
0x12,0x12,0x12,0x12,0,0,0,0
};

static const uint32_t ubidi_props_mirrors[40]={
0x2000ab,0xbb,0x4202215,0x4e0221f,0x3e02220,0x3a02221,0x3c02222,0x4c02224,0x2202243,0x1402245,0x120224c,0x4002298,0x44022a6,0x48022a8,0x46022a9,0x4a022ab,
0x38022b8,0x10022cd,0x2e022f2,0x30022f3,0x32022f4,0x34022f6,0x36022f7,0x24022fa,0x26022fb,0x28022fc,0x2a022fd,0x2c022fe,0x20027dc,0xa0299b,0xc029a0,0x8029a3,
0x16029b8,0x4029f5,0x1802ade,0x1c02ae3,0x1a02ae4,0x1e02ae5,0xe02aee,0x602bfe
};

static const uint8_t ubidi_props_jgArray[680]={
0x2d,0,3,3,0x2c,3,0x2d,3,4,0x2a,4,4,0xd,0xd,0xd,6,
6,0x1f,0x1f,0x23,0x23,0x21,0x21,0x28,0x28,1,1,0xb,0xb,0x37,0x37,0x37,
0,9,0x1d,0x13,0x16,0x18,0x1a,0x10,0x2c,0x2d,0x2d,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0x1d,
0,3,3,3,0,3,0x2c,0x2c,0x2d,4,4,4,4,4,4,4,
4,0xd,0xd,0xd,0xd,0xd,0xd,0xd,6,6,6,6,6,6,6,6,
6,0x1f,0x1f,0x1f,0x1f,0x1f,0x1f,0x1f,0x1f,0x1f,0x23,0x23,0x23,0x21,0x21,0x28,
1,9,9,9,9,9,9,0x1d,0x1d,0xb,0x26,0xb,0x13,0x13,0x13,0xb,
0xb,0xb,0xb,0xb,0xb,0x16,0x16,0x16,0x16,0x1a,0x1a,0x1a,0x1a,0x38,0x15,0xd,
0x2a,0x11,0x11,0xe,0x2c,0x2c,0x2c,0x2c,0x2c,0x2c,0x2c,0x2c,0x37,0x2f,0x37,0x2c,
0x2d,0x2d,0x2e,0x2e,0,0x2a,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,6,0x1f,
0,0,0,0,0,0,0,0,0,0,0x23,0x21,1,0,0,0x15,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
2,0,5,0xc,0xc,7,7,0xf,0x27,0x32,0x12,0x2b,0x2b,0x30,0x31,0x14,
0x17,0x19,0x1b,0x24,0xa,8,0x1c,0x20,0x22,0x1e,7,0x25,0x29,5,0xc,7,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0x35,0x34,0x33,
4,4,4,4,4,4,4,0xd,0xd,6,6,0x1f,0x23,1,1,1,
9,9,0xb,0xb,0xb,0x18,0x18,0x1a,0x1a,0x1a,0x16,0x1f,0x1f,0x23,0xd,0xd,
0x23,0x1f,0xd,3,3,0x37,0x37,0x2d,0x2c,0x2c,0x36,0x36,0xd,0x23,0x23,0x13,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x5d,0x5a,0x60,0x63,0x5e,0x5f,0x59,0x61,0x5b,0x5c,0x62,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
4,4,0xd,0x28,9,0x1d,0x16,0x18,0x2d,0x2d,0x1f,0x2c,0x39,0,6,0x21,
0xb,0x55,0x1f,1,0x13,0,4,4,4,0x1f,0x2d,0x56,0x58,0x57,4,4,
4,0xd,0xb,1,0x58,0xd,0xd,0x16
};

static const uint8_t ubidi_props_jgArray2[612]={
0x3a,0x3c,0x3c,0x40,0x40,0x3d,0,0x52,0,0x54,0x54,0,0,0x41,0x4f,0x53,
0x43,0x43,0x43,0x44,0x3e,0x50,0x45,0x46,0x4c,0x3b,0x3b,0x48,0x48,0x4b,0x49,0x49,
0x49,0x4a,0,0,0x4d,0,0,0,0,0,0,0x47,0x3f,0x4e,0x51,0x42,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0x65,0,0,0,0,0,0,0x65,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0x64,0,0,0x65,0,0x64,0,
0x64,0,0,0x64
};

static const UBiDiProps ubidi_props_singleton={
  NULL,
  ubidi_props_indexes,
  ubidi_props_mirrors,
  ubidi_props_jgArray,
  ubidi_props_jgArray2,
  {
    ubidi_props_trieIndex,
    ubidi_props_trieIndex+3568,
    NULL,
    3568,
    8968,
    0x1a0,
    0xe70,
    0x0,
    0x0,
    0x110000,
    0x30f4,
    NULL, 0, FALSE, FALSE, 0, NULL
  },
  { 2,2,0,0 }
};

#endif  // INCLUDED_FROM_UBIDI_PROPS_C


/* set of property starts for UnicodeSet ------------------------------------ */

static UBool U_CALLCONV
ubidi_props_enumPropertyStartsRange(const void *context, UChar32 start, UChar32 end, uint32_t value) {
    (void)end;
    (void)value;
    /* add the start code point to the USet */
    const USetAdder *sa=(const USetAdder *)context;
    sa->add(sa->set, start);
    return TRUE;
}

U_CFUNC void
ubidi_addPropertyStarts(const USetAdder *sa, UErrorCode *pErrorCode) {
    int32_t i, length;
    UChar32 c, start, limit;

    const uint8_t *jgArray;
    uint8_t prev, jg;

    if(U_FAILURE(*pErrorCode)) {
        return;
    }

    /* add the start code point of each same-value range of the trie */
    utrie2_enum(&ubidi_props_singleton.trie, NULL, ubidi_props_enumPropertyStartsRange, sa);

    /* add the code points from the bidi mirroring table */
    length=ubidi_props_singleton.indexes[UBIDI_IX_MIRROR_LENGTH];
    for(i=0; i<length; ++i) {
        c=UBIDI_GET_MIRROR_CODE_POINT(ubidi_props_singleton.mirrors[i]);
        sa->addRange(sa->set, c, c+1);
    }

    /* add the code points from the Joining_Group array where the value changes */
    start=ubidi_props_singleton.indexes[UBIDI_IX_JG_START];
    limit=ubidi_props_singleton.indexes[UBIDI_IX_JG_LIMIT];
    jgArray=ubidi_props_singleton.jgArray;
    for(;;) {
        prev=0;
        while(start<limit) {
            jg=*jgArray++;
            if(jg!=prev) {
                sa->add(sa->set, start);
                prev=jg;
            }
            ++start;
        }
        if(prev!=0) {
            /* add the limit code point if the last value was not 0 (it is now start==limit) */
            sa->add(sa->set, limit);
        }
        if(limit==ubidi_props_singleton.indexes[UBIDI_IX_JG_LIMIT]) {
            /* switch to the second Joining_Group range */
            start=ubidi_props_singleton.indexes[UBIDI_IX_JG_START2];
            limit=ubidi_props_singleton.indexes[UBIDI_IX_JG_LIMIT2];
            jgArray=ubidi_props_singleton.jgArray2;
        } else {
            break;
        }
    }

    /* add code points with hardcoded properties, plus the ones following them */

    /* (none right now) */
}

/* property access functions ------------------------------------------------ */

U_CFUNC int32_t
ubidi_getMaxValue(UProperty which) {
    int32_t max=ubidi_props_singleton.indexes[UBIDI_MAX_VALUES_INDEX];
    switch(which) {
    case UCHAR_BIDI_CLASS:
        return (max&UBIDI_CLASS_MASK);
    case UCHAR_JOINING_GROUP:
        return (max&UBIDI_MAX_JG_MASK)>>UBIDI_MAX_JG_SHIFT;
    case UCHAR_JOINING_TYPE:
        return (max&UBIDI_JT_MASK)>>UBIDI_JT_SHIFT;
    case UCHAR_BIDI_PAIRED_BRACKET_TYPE:
        return (max&UBIDI_BPT_MASK)>>UBIDI_BPT_SHIFT;
    default:
        return -1; /* undefined */
    }
}

U_CAPI UCharDirection
ubidi_getClass(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ubidi_props_singleton.trie, c);
    return (UCharDirection)UBIDI_GET_CLASS(props);
}

U_CFUNC UBool
ubidi_isMirrored(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ubidi_props_singleton.trie, c);
    return (UBool)UBIDI_GET_FLAG(props, UBIDI_IS_MIRRORED_SHIFT);
}

static UChar32
getMirror(UChar32 c, uint16_t props) {
    int32_t delta=UBIDI_GET_MIRROR_DELTA(props);
    if(delta!=UBIDI_ESC_MIRROR_DELTA) {
        return c+delta;
    } else {
        /* look for mirror code point in the mirrors[] table */
        const uint32_t *mirrors;
        uint32_t m;
        int32_t i, length;
        UChar32 c2;

        mirrors=ubidi_props_singleton.mirrors;
        length=ubidi_props_singleton.indexes[UBIDI_IX_MIRROR_LENGTH];

        /* linear search */
        for(i=0; i<length; ++i) {
            m=mirrors[i];
            c2=UBIDI_GET_MIRROR_CODE_POINT(m);
            if(c==c2) {
                /* found c, return its mirror code point using the index in m */
                return UBIDI_GET_MIRROR_CODE_POINT(mirrors[UBIDI_GET_MIRROR_INDEX(m)]);
            } else if(c<c2) {
                break;
            }
        }

        /* c not found, return it itself */
        return c;
    }
}

U_CFUNC UChar32
ubidi_getMirror(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ubidi_props_singleton.trie, c);
    return getMirror(c, props);
}

U_CFUNC UBool
ubidi_isBidiControl(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ubidi_props_singleton.trie, c);
    return (UBool)UBIDI_GET_FLAG(props, UBIDI_BIDI_CONTROL_SHIFT);
}

U_CFUNC UBool
ubidi_isJoinControl(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ubidi_props_singleton.trie, c);
    return (UBool)UBIDI_GET_FLAG(props, UBIDI_JOIN_CONTROL_SHIFT);
}

U_CFUNC UJoiningType
ubidi_getJoiningType(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ubidi_props_singleton.trie, c);
    return (UJoiningType)((props&UBIDI_JT_MASK)>>UBIDI_JT_SHIFT);
}

U_CFUNC UJoiningGroup
ubidi_getJoiningGroup(UChar32 c) {
    UChar32 start, limit;

    start=ubidi_props_singleton.indexes[UBIDI_IX_JG_START];
    limit=ubidi_props_singleton.indexes[UBIDI_IX_JG_LIMIT];
    if(start<=c && c<limit) {
        return (UJoiningGroup)ubidi_props_singleton.jgArray[c-start];
    }
    start=ubidi_props_singleton.indexes[UBIDI_IX_JG_START2];
    limit=ubidi_props_singleton.indexes[UBIDI_IX_JG_LIMIT2];
    if(start<=c && c<limit) {
        return (UJoiningGroup)ubidi_props_singleton.jgArray2[c-start];
    }
    return U_JG_NO_JOINING_GROUP;
}

U_CFUNC UBidiPairedBracketType
ubidi_getPairedBracketType(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ubidi_props_singleton.trie, c);
    return (UBidiPairedBracketType)((props&UBIDI_BPT_MASK)>>UBIDI_BPT_SHIFT);
}

U_CFUNC UChar32
ubidi_getPairedBracket(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ubidi_props_singleton.trie, c);
    if((props&UBIDI_BPT_MASK)==0) {
        return c;
    } else {
        return getMirror(c, props);
    }
}

/* public API (see uchar.h) ------------------------------------------------- */

U_CFUNC UCharDirection
u_charDirection(UChar32 c) {
    return ubidi_getClass(c);
}

U_CFUNC UBool
u_isMirrored(UChar32 c) {
    return ubidi_isMirrored(c);
}

U_CFUNC UChar32
u_charMirror(UChar32 c) {
    return ubidi_getMirror(c);
}

U_STABLE UChar32 U_EXPORT2
u_getBidiPairedBracket(UChar32 c) {
    return ubidi_getPairedBracket(c);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  ubidiln.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999aug06
*   created by: Markus W. Scherer, updated by Matitiahu Allouche
*/









/*
 * General remarks about the functions in this file:
 *
 * These functions deal with the aspects of potentially mixed-directional
 * text in a single paragraph or in a line of a single paragraph
 * which has already been processed according to
 * the Unicode 6.3 BiDi algorithm as defined in
 * http://www.unicode.org/unicode/reports/tr9/ , version 28,
 * also described in The Unicode Standard, Version 6.3.0 .
 *
 * This means that there is a UBiDi object with a levels
 * and a dirProps array.
 * paraLevel and direction are also set.
 * Only if the length of the text is zero, then levels==dirProps==NULL.
 *
 * The overall directionality of the paragraph
 * or line is used to bypass the reordering steps if possible.
 * Even purely RTL text does not need reordering there because
 * the ubidi_getLogical/VisualIndex() functions can compute the
 * index on the fly in such a case.
 *
 * The implementation of the access to same-level-runs and of the reordering
 * do attempt to provide better performance and less memory usage compared to
 * a direct implementation of especially rule (L2) with an array of
 * one (32-bit) integer per text character.
 *
 * Here, the levels array is scanned as soon as necessary, and a vector of
 * same-level-runs is created. Reordering then is done on this vector.
 * For each run of text positions that were resolved to the same level,
 * only 8 bytes are stored: the first text position of the run and the visual
 * position behind the run after reordering.
 * One sign bit is used to hold the directionality of the run.
 * This is inefficient if there are many very short runs. If the average run
 * length is <2, then this uses more memory.
 *
 * In a further attempt to save memory, the levels array is never changed
 * after all the resolution rules (Xn, Wn, Nn, In).
 * Many functions have to consider the field trailingWSStart:
 * if it is less than length, then there is an implicit trailing run
 * at the paraLevel,
 * which is not reflected in the levels array.
 * This allows a line UBiDi object to use the same levels array as
 * its paragraph parent object.
 *
 * When a UBiDi object is created for a line of a paragraph, then the
 * paragraph's levels and dirProps arrays are reused by way of setting
 * a pointer into them, not by copying. This again saves memory and forbids to
 * change the now shared levels for (L1).
 */

/* handle trailing WS (L1) -------------------------------------------------- */

/*
 * setTrailingWSStart() sets the start index for a trailing
 * run of WS in the line. This is necessary because we do not modify
 * the paragraph's levels array that we just point into.
 * Using trailingWSStart is another form of performing (L1).
 *
 * To make subsequent operations easier, we also include the run
 * before the WS if it is at the paraLevel - we merge the two here.
 *
 * This function is called only from ubidi_setLine(), so pBiDi->paraLevel is
 * set correctly for the line even when contextual multiple paragraphs.
 */
static void
setTrailingWSStart(UBiDi *pBiDi) {
    /* pBiDi->direction!=UBIDI_MIXED */

    const DirProp *dirProps=pBiDi->dirProps;
    UBiDiLevel *levels=pBiDi->levels;
    int32_t start=pBiDi->length;
    UBiDiLevel paraLevel=pBiDi->paraLevel;

    /* If the line is terminated by a block separator, all preceding WS etc...
       are already set to paragraph level.
       Setting trailingWSStart to pBidi->length will avoid changing the
       level of B chars from 0 to paraLevel in ubidi_getLevels when
       orderParagraphsLTR==TRUE.
     */
    if(dirProps[start-1]==B) {
        pBiDi->trailingWSStart=start;   /* currently == pBiDi->length */
        return;
    }
    /* go backwards across all WS, BN, explicit codes */
    while(start>0 && DIRPROP_FLAG(dirProps[start-1])&MASK_WS) {
        --start;
    }

    /* if the WS run can be merged with the previous run then do so here */
    while(start>0 && levels[start-1]==paraLevel) {
        --start;
    }

    pBiDi->trailingWSStart=start;
}

/* ubidi_setLine ------------------------------------------------------------ */

U_CAPI void U_EXPORT2
ubidi_setLine(const UBiDi *pParaBiDi,
              int32_t start, int32_t limit,
              UBiDi *pLineBiDi,
              UErrorCode *pErrorCode) {
    int32_t length;

    /* check the argument values */
    RETURN_VOID_IF_NULL_OR_FAILING_ERRCODE(pErrorCode);
    RETURN_VOID_IF_NOT_VALID_PARA(pParaBiDi, *pErrorCode);
    RETURN_VOID_IF_BAD_RANGE(start, 0, limit, *pErrorCode);
    RETURN_VOID_IF_BAD_RANGE(limit, 0, pParaBiDi->length+1, *pErrorCode);
    if(pLineBiDi==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if(ubidi_getParagraph(pParaBiDi, start, NULL, NULL, NULL, pErrorCode) !=
       ubidi_getParagraph(pParaBiDi, limit-1, NULL, NULL, NULL, pErrorCode)) {
        /* the line crosses a paragraph boundary */
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    /* set the values in pLineBiDi from its pParaBiDi parent */
    pLineBiDi->pParaBiDi=NULL;          /* mark unfinished setLine */
    pLineBiDi->text=pParaBiDi->text+start;
    length=pLineBiDi->length=limit-start;
    pLineBiDi->resultLength=pLineBiDi->originalLength=length;
    pLineBiDi->paraLevel=GET_PARALEVEL(pParaBiDi, start);
    pLineBiDi->paraCount=pParaBiDi->paraCount;
    pLineBiDi->runs=NULL;
    pLineBiDi->flags=0;
    pLineBiDi->reorderingMode=pParaBiDi->reorderingMode;
    pLineBiDi->reorderingOptions=pParaBiDi->reorderingOptions;
    pLineBiDi->controlCount=0;
    if(pParaBiDi->controlCount>0) {
        int32_t j;
        for(j=start; j<limit; j++) {
            if(IS_BIDI_CONTROL_CHAR(pParaBiDi->text[j])) {
                pLineBiDi->controlCount++;
            }
        }
        pLineBiDi->resultLength-=pLineBiDi->controlCount;
    }

    pLineBiDi->dirProps=pParaBiDi->dirProps+start;
    pLineBiDi->levels=pParaBiDi->levels+start;
    pLineBiDi->runCount=-1;

    if(pParaBiDi->direction!=UBIDI_MIXED) {
        /* the parent is already trivial */
        pLineBiDi->direction=pParaBiDi->direction;

        /*
         * The parent's levels are all either
         * implicitly or explicitly ==paraLevel;
         * do the same here.
         */
        if(pParaBiDi->trailingWSStart<=start) {
            pLineBiDi->trailingWSStart=0;
        } else if(pParaBiDi->trailingWSStart<limit) {
            pLineBiDi->trailingWSStart=pParaBiDi->trailingWSStart-start;
        } else {
            pLineBiDi->trailingWSStart=length;
        }
    } else {
        const UBiDiLevel *levels=pLineBiDi->levels;
        int32_t i, trailingWSStart;
        UBiDiLevel level;

        setTrailingWSStart(pLineBiDi);
        trailingWSStart=pLineBiDi->trailingWSStart;

        /* recalculate pLineBiDi->direction */
        if(trailingWSStart==0) {
            /* all levels are at paraLevel */
            pLineBiDi->direction=(UBiDiDirection)(pLineBiDi->paraLevel&1);
        } else {
            /* get the level of the first character */
            level=(UBiDiLevel)(levels[0]&1);

            /* if there is anything of a different level, then the line is mixed */
            if(trailingWSStart<length && (pLineBiDi->paraLevel&1)!=level) {
                /* the trailing WS is at paraLevel, which differs from levels[0] */
                pLineBiDi->direction=UBIDI_MIXED;
            } else {
                /* see if levels[1..trailingWSStart-1] have the same direction as levels[0] and paraLevel */
                i=1;
                for(;;) {
                    if(i==trailingWSStart) {
                        /* the direction values match those in level */
                        pLineBiDi->direction=(UBiDiDirection)level;
                        break;
                    } else if((levels[i]&1)!=level) {
                        pLineBiDi->direction=UBIDI_MIXED;
                        break;
                    }
                    ++i;
                }
            }
        }

        switch(pLineBiDi->direction) {
        case UBIDI_LTR:
            /* make sure paraLevel is even */
            pLineBiDi->paraLevel=(UBiDiLevel)((pLineBiDi->paraLevel+1)&~1);

            /* all levels are implicitly at paraLevel (important for ubidi_getLevels()) */
            pLineBiDi->trailingWSStart=0;
            break;
        case UBIDI_RTL:
            /* make sure paraLevel is odd */
            pLineBiDi->paraLevel|=1;

            /* all levels are implicitly at paraLevel (important for ubidi_getLevels()) */
            pLineBiDi->trailingWSStart=0;
            break;
        default:
            break;
        }
    }
    pLineBiDi->pParaBiDi=pParaBiDi;     /* mark successful setLine */
    return;
}

U_CAPI UBiDiLevel U_EXPORT2
ubidi_getLevelAt(const UBiDi *pBiDi, int32_t charIndex) {
    /* return paraLevel if in the trailing WS run, otherwise the real level */
    if(!IS_VALID_PARA_OR_LINE(pBiDi) || charIndex<0 || pBiDi->length<=charIndex) {
        return 0;
    } else if(pBiDi->direction!=UBIDI_MIXED || charIndex>=pBiDi->trailingWSStart) {
        return GET_PARALEVEL(pBiDi, charIndex);
    } else {
        return pBiDi->levels[charIndex];
    }
}

U_CAPI const UBiDiLevel * U_EXPORT2
ubidi_getLevels(UBiDi *pBiDi, UErrorCode *pErrorCode) {
    int32_t start, length;

    RETURN_IF_NULL_OR_FAILING_ERRCODE(pErrorCode, NULL);
    RETURN_IF_NOT_VALID_PARA_OR_LINE(pBiDi, *pErrorCode, NULL);
    if((length=pBiDi->length)<=0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    if((start=pBiDi->trailingWSStart)==length) {
        /* the current levels array reflects the WS run */
        return pBiDi->levels;
    }

    /*
     * After the previous if(), we know that the levels array
     * has an implicit trailing WS run and therefore does not fully
     * reflect itself all the levels.
     * This must be a UBiDi object for a line, and
     * we need to create a new levels array.
     */
    if(getLevelsMemory(pBiDi, length)) {
        UBiDiLevel *levels=pBiDi->levelsMemory;

        if(start>0 && levels!=pBiDi->levels) {
            uprv_memcpy(levels, pBiDi->levels, start);
        }
        /* pBiDi->paraLevel is ok even if contextual multiple paragraphs,
           since pBidi is a line object                                     */
        uprv_memset(levels+start, pBiDi->paraLevel, length-start);

        /* this new levels array is set for the line and reflects the WS run */
        pBiDi->trailingWSStart=length;
        return pBiDi->levels=levels;
    } else {
        /* out of memory */
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
}

U_CAPI void U_EXPORT2
ubidi_getLogicalRun(const UBiDi *pBiDi, int32_t logicalPosition,
                    int32_t *pLogicalLimit, UBiDiLevel *pLevel) {
    UErrorCode errorCode;
    int32_t runCount, visualStart, logicalLimit, logicalFirst, i;
    Run iRun;

    errorCode=U_ZERO_ERROR;
    RETURN_VOID_IF_BAD_RANGE(logicalPosition, 0, pBiDi->length, errorCode);
    /* ubidi_countRuns will check VALID_PARA_OR_LINE */
    runCount=ubidi_countRuns((UBiDi *)pBiDi, &errorCode);
    if(U_FAILURE(errorCode)) {
        return;
    }
    /* this is done based on runs rather than on levels since levels have
       a special interpretation when UBIDI_REORDER_RUNS_ONLY
     */
    visualStart=logicalLimit=0;
    iRun=pBiDi->runs[0];

    for(i=0; i<runCount; i++) {
        iRun = pBiDi->runs[i];
        logicalFirst=GET_INDEX(iRun.logicalStart);
        logicalLimit=logicalFirst+iRun.visualLimit-visualStart;
        if((logicalPosition>=logicalFirst) &&
           (logicalPosition<logicalLimit)) {
            break;
        }
        visualStart = iRun.visualLimit;
    }
    if(pLogicalLimit) {
        *pLogicalLimit=logicalLimit;
    }
    if(pLevel) {
        if(pBiDi->reorderingMode==UBIDI_REORDER_RUNS_ONLY) {
            *pLevel=(UBiDiLevel)GET_ODD_BIT(iRun.logicalStart);
        }
        else if(pBiDi->direction!=UBIDI_MIXED || logicalPosition>=pBiDi->trailingWSStart) {
            *pLevel=GET_PARALEVEL(pBiDi, logicalPosition);
        } else {
        *pLevel=pBiDi->levels[logicalPosition];
        }
    }
}

/* runs API functions ------------------------------------------------------- */

U_CAPI int32_t U_EXPORT2
ubidi_countRuns(UBiDi *pBiDi, UErrorCode *pErrorCode) {
    RETURN_IF_NULL_OR_FAILING_ERRCODE(pErrorCode, -1);
    RETURN_IF_NOT_VALID_PARA_OR_LINE(pBiDi, *pErrorCode, -1);
    ubidi_getRuns(pBiDi, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        return -1;
    }
    return pBiDi->runCount;
}

U_CAPI UBiDiDirection U_EXPORT2
ubidi_getVisualRun(UBiDi *pBiDi, int32_t runIndex,
                   int32_t *pLogicalStart, int32_t *pLength)
{
    int32_t start;
    UErrorCode errorCode = U_ZERO_ERROR;
    RETURN_IF_NOT_VALID_PARA_OR_LINE(pBiDi, errorCode, UBIDI_LTR);
    ubidi_getRuns(pBiDi, &errorCode);
    if(U_FAILURE(errorCode)) {
        return UBIDI_LTR;
    }
    RETURN_IF_BAD_RANGE(runIndex, 0, pBiDi->runCount, errorCode, UBIDI_LTR);

    start=pBiDi->runs[runIndex].logicalStart;
    if(pLogicalStart!=NULL) {
        *pLogicalStart=GET_INDEX(start);
    }
    if(pLength!=NULL) {
        if(runIndex>0) {
            *pLength=pBiDi->runs[runIndex].visualLimit-
                     pBiDi->runs[runIndex-1].visualLimit;
        } else {
            *pLength=pBiDi->runs[0].visualLimit;
        }
    }
    return (UBiDiDirection)GET_ODD_BIT(start);
}

/* in trivial cases there is only one trivial run; called by ubidi_getRuns() */
static void
getSingleRun(UBiDi *pBiDi, UBiDiLevel level) {
    /* simple, single-run case */
    pBiDi->runs=pBiDi->simpleRuns;
    pBiDi->runCount=1;

    /* fill and reorder the single run */
    pBiDi->runs[0].logicalStart=MAKE_INDEX_ODD_PAIR(0, level);
    pBiDi->runs[0].visualLimit=pBiDi->length;
    pBiDi->runs[0].insertRemove=0;
}

/* reorder the runs array (L2) ---------------------------------------------- */

/*
 * Reorder the same-level runs in the runs array.
 * Here, runCount>1 and maxLevel>=minLevel>=paraLevel.
 * All the visualStart fields=logical start before reordering.
 * The "odd" bits are not set yet.
 *
 * Reordering with this data structure lends itself to some handy shortcuts:
 *
 * Since each run is moved but not modified, and since at the initial maxLevel
 * each sequence of same-level runs consists of only one run each, we
 * don't need to do anything there and can predecrement maxLevel.
 * In many simple cases, the reordering is thus done entirely in the
 * index mapping.
 * Also, reordering occurs only down to the lowest odd level that occurs,
 * which is minLevel|1. However, if the lowest level itself is odd, then
 * in the last reordering the sequence of the runs at this level or higher
 * will be all runs, and we don't need the elaborate loop to search for them.
 * This is covered by ++minLevel instead of minLevel|=1 followed
 * by an extra reorder-all after the reorder-some loop.
 * About a trailing WS run:
 * Such a run would need special treatment because its level is not
 * reflected in levels[] if this is not a paragraph object.
 * Instead, all characters from trailingWSStart on are implicitly at
 * paraLevel.
 * However, for all maxLevel>paraLevel, this run will never be reordered
 * and does not need to be taken into account. maxLevel==paraLevel is only reordered
 * if minLevel==paraLevel is odd, which is done in the extra segment.
 * This means that for the main reordering loop we don't need to consider
 * this run and can --runCount. If it is later part of the all-runs
 * reordering, then runCount is adjusted accordingly.
 */
static void
reorderLine(UBiDi *pBiDi, UBiDiLevel minLevel, UBiDiLevel maxLevel) {
    Run *runs, tempRun;
    UBiDiLevel *levels;
    int32_t firstRun, endRun, limitRun, runCount;

    /* nothing to do? */
    if(maxLevel<=(minLevel|1)) {
        return;
    }

    /*
     * Reorder only down to the lowest odd level
     * and reorder at an odd minLevel in a separate, simpler loop.
     * See comments above for why minLevel is always incremented.
     */
    ++minLevel;

    runs=pBiDi->runs;
    levels=pBiDi->levels;
    runCount=pBiDi->runCount;

    /* do not include the WS run at paraLevel<=old minLevel except in the simple loop */
    if(pBiDi->trailingWSStart<pBiDi->length) {
        --runCount;
    }

    while(--maxLevel>=minLevel) {
        firstRun=0;

        /* loop for all sequences of runs */
        for(;;) {
            /* look for a sequence of runs that are all at >=maxLevel */
            /* look for the first run of such a sequence */
            while(firstRun<runCount && levels[runs[firstRun].logicalStart]<maxLevel) {
                ++firstRun;
            }
            if(firstRun>=runCount) {
                break;  /* no more such runs */
            }

            /* look for the limit run of such a sequence (the run behind it) */
            for(limitRun=firstRun; ++limitRun<runCount && levels[runs[limitRun].logicalStart]>=maxLevel;) {}

            /* Swap the entire sequence of runs from firstRun to limitRun-1. */
            endRun=limitRun-1;
            while(firstRun<endRun) {
                tempRun = runs[firstRun];
                runs[firstRun]=runs[endRun];
                runs[endRun]=tempRun;
                ++firstRun;
                --endRun;
            }

            if(limitRun==runCount) {
                break;  /* no more such runs */
            } else {
                firstRun=limitRun+1;
            }
        }
    }

    /* now do maxLevel==old minLevel (==odd!), see above */
    if(!(minLevel&1)) {
        firstRun=0;

        /* include the trailing WS run in this complete reordering */
        if(pBiDi->trailingWSStart==pBiDi->length) {
            --runCount;
        }

        /* Swap the entire sequence of all runs. (endRun==runCount) */
        while(firstRun<runCount) {
            tempRun=runs[firstRun];
            runs[firstRun]=runs[runCount];
            runs[runCount]=tempRun;
            ++firstRun;
            --runCount;
        }
    }
}

/* compute the runs array --------------------------------------------------- */

static int32_t getRunFromLogicalIndex(UBiDi *pBiDi, int32_t logicalIndex) {
    Run *runs=pBiDi->runs;
    int32_t runCount=pBiDi->runCount, visualStart=0, i, length, logicalStart;

    for(i=0; i<runCount; i++) {
        length=runs[i].visualLimit-visualStart;
        logicalStart=GET_INDEX(runs[i].logicalStart);
        if((logicalIndex>=logicalStart) && (logicalIndex<(logicalStart+length))) {
            return i;
        }
        visualStart+=length;
    }
    /* we should never get here */
    UPRV_UNREACHABLE;
}

/*
 * Compute the runs array from the levels array.
 * After ubidi_getRuns() returns TRUE, runCount is guaranteed to be >0
 * and the runs are reordered.
 * Odd-level runs have visualStart on their visual right edge and
 * they progress visually to the left.
 * If option UBIDI_OPTION_INSERT_MARKS is set, insertRemove will contain the
 * sum of appropriate LRM/RLM_BEFORE/AFTER flags.
 * If option UBIDI_OPTION_REMOVE_CONTROLS is set, insertRemove will contain the
 * negative number of BiDi control characters within this run.
 */
U_CFUNC UBool
ubidi_getRuns(UBiDi *pBiDi, UErrorCode*) {
    /*
     * This method returns immediately if the runs are already set. This
     * includes the case of length==0 (handled in setPara)..
     */
    if (pBiDi->runCount>=0) {
        return TRUE;
    }

    if(pBiDi->direction!=UBIDI_MIXED) {
        /* simple, single-run case - this covers length==0 */
        /* pBiDi->paraLevel is ok even for contextual multiple paragraphs */
        getSingleRun(pBiDi, pBiDi->paraLevel);
    } else /* UBIDI_MIXED, length>0 */ {
        /* mixed directionality */
        int32_t length=pBiDi->length, limit;
        UBiDiLevel *levels=pBiDi->levels;
        int32_t i, runCount;
        UBiDiLevel level=UBIDI_DEFAULT_LTR;   /* initialize with no valid level */
        /*
         * If there are WS characters at the end of the line
         * and the run preceding them has a level different from
         * paraLevel, then they will form their own run at paraLevel (L1).
         * Count them separately.
         * We need some special treatment for this in order to not
         * modify the levels array which a line UBiDi object shares
         * with its paragraph parent and its other line siblings.
         * In other words, for the trailing WS, it may be
         * levels[]!=paraLevel but we have to treat it like it were so.
         */
        limit=pBiDi->trailingWSStart;
        /* count the runs, there is at least one non-WS run, and limit>0 */
        runCount=0;
        for(i=0; i<limit; ++i) {
            /* increment runCount at the start of each run */
            if(levels[i]!=level) {
                ++runCount;
                level=levels[i];
            }
        }

        /*
         * We don't need to see if the last run can be merged with a trailing
         * WS run because setTrailingWSStart() would have done that.
         */
        if(runCount==1 && limit==length) {
            /* There is only one non-WS run and no trailing WS-run. */
            getSingleRun(pBiDi, levels[0]);
        } else /* runCount>1 || limit<length */ {
            /* allocate and set the runs */
            Run *runs;
            int32_t runIndex, start;
            UBiDiLevel minLevel=UBIDI_MAX_EXPLICIT_LEVEL+1, maxLevel=0;

            /* now, count a (non-mergeable) WS run */
            if(limit<length) {
                ++runCount;
            }

            /* runCount>1 */
            if(getRunsMemory(pBiDi, runCount)) {
                runs=pBiDi->runsMemory;
            } else {
                return FALSE;
            }

            /* set the runs */
            /* FOOD FOR THOUGHT: this could be optimized, e.g.:
             * 464->444, 484->444, 575->555, 595->555
             * However, that would take longer. Check also how it would
             * interact with BiDi control removal and inserting Marks.
             */
            runIndex=0;

            /* search for the run limits and initialize visualLimit values with the run lengths */
            i=0;
            do {
                /* prepare this run */
                start=i;
                level=levels[i];
                if(level<minLevel) {
                    minLevel=level;
                }
                if(level>maxLevel) {
                    maxLevel=level;
                }

                /* look for the run limit */
                while(++i<limit && levels[i]==level) {}

                /* i is another run limit */
                runs[runIndex].logicalStart=start;
                runs[runIndex].visualLimit=i-start;
                runs[runIndex].insertRemove=0;
                ++runIndex;
            } while(i<limit);

            if(limit<length) {
                /* there is a separate WS run */
                runs[runIndex].logicalStart=limit;
                runs[runIndex].visualLimit=length-limit;
                /* For the trailing WS run, pBiDi->paraLevel is ok even
                   if contextual multiple paragraphs.                   */
                if(pBiDi->paraLevel<minLevel) {
                    minLevel=pBiDi->paraLevel;
                }
            }

            /* set the object fields */
            pBiDi->runs=runs;
            pBiDi->runCount=runCount;

            reorderLine(pBiDi, minLevel, maxLevel);

            /* now add the direction flags and adjust the visualLimit's to be just that */
            /* this loop will also handle the trailing WS run */
            limit=0;
            for(i=0; i<runCount; ++i) {
                ADD_ODD_BIT_FROM_LEVEL(runs[i].logicalStart, levels[runs[i].logicalStart]);
                limit+=runs[i].visualLimit;
                runs[i].visualLimit=limit;
            }

            /* Set the "odd" bit for the trailing WS run. */
            /* For a RTL paragraph, it will be the *first* run in visual order. */
            /* For the trailing WS run, pBiDi->paraLevel is ok even if
               contextual multiple paragraphs.                          */
            if(runIndex<runCount) {
                int32_t trailingRun = ((pBiDi->paraLevel & 1) != 0)? 0 : runIndex;

                ADD_ODD_BIT_FROM_LEVEL(runs[trailingRun].logicalStart, pBiDi->paraLevel);
            }
        }
    }

    /* handle insert LRM/RLM BEFORE/AFTER run */
    if(pBiDi->insertPoints.size>0) {
        Point *point, *start=pBiDi->insertPoints.points,
                      *limit=start+pBiDi->insertPoints.size;
        int32_t runIndex;
        for(point=start; point<limit; point++) {
            runIndex=getRunFromLogicalIndex(pBiDi, point->pos);
            pBiDi->runs[runIndex].insertRemove|=point->flag;
        }
    }

    /* handle remove BiDi control characters */
    if(pBiDi->controlCount>0) {
        int32_t runIndex;
        const UChar *start=pBiDi->text, *limit=start+pBiDi->length, *pu;
        for(pu=start; pu<limit; pu++) {
            if(IS_BIDI_CONTROL_CHAR(*pu)) {
                runIndex=getRunFromLogicalIndex(pBiDi, (int32_t)(pu-start));
                pBiDi->runs[runIndex].insertRemove--;
            }
        }
    }

    return TRUE;
}

static UBool
prepareReorder(const UBiDiLevel *levels, int32_t length,
               int32_t *indexMap,
               UBiDiLevel *pMinLevel, UBiDiLevel *pMaxLevel) {
    int32_t start;
    UBiDiLevel level, minLevel, maxLevel;

    if(levels==NULL || length<=0) {
        return FALSE;
    }

    /* determine minLevel and maxLevel */
    minLevel=UBIDI_MAX_EXPLICIT_LEVEL+1;
    maxLevel=0;
    for(start=length; start>0;) {
        level=levels[--start];
        if(level>UBIDI_MAX_EXPLICIT_LEVEL+1) {
            return FALSE;
        }
        if(level<minLevel) {
            minLevel=level;
        }
        if(level>maxLevel) {
            maxLevel=level;
        }
    }
    *pMinLevel=minLevel;
    *pMaxLevel=maxLevel;

    /* initialize the index map */
    for(start=length; start>0;) {
        --start;
        indexMap[start]=start;
    }

    return TRUE;
}

/* reorder a line based on a levels array (L2) ------------------------------ */

U_CAPI void U_EXPORT2
ubidi_reorderLogical(const UBiDiLevel *levels, int32_t length, int32_t *indexMap) {
    int32_t start, limit, sumOfSosEos;
    UBiDiLevel minLevel = 0, maxLevel = 0;

    if(indexMap==NULL || !prepareReorder(levels, length, indexMap, &minLevel, &maxLevel)) {
        return;
    }

    /* nothing to do? */
    if(minLevel==maxLevel && (minLevel&1)==0) {
        return;
    }

    /* reorder only down to the lowest odd level */
    minLevel|=1;

    /* loop maxLevel..minLevel */
    do {
        start=0;

        /* loop for all sequences of levels to reorder at the current maxLevel */
        for(;;) {
            /* look for a sequence of levels that are all at >=maxLevel */
            /* look for the first index of such a sequence */
            while(start<length && levels[start]<maxLevel) {
                ++start;
            }
            if(start>=length) {
                break;  /* no more such sequences */
            }

            /* look for the limit of such a sequence (the index behind it) */
            for(limit=start; ++limit<length && levels[limit]>=maxLevel;) {}

            /*
             * sos=start of sequence, eos=end of sequence
             *
             * The closed (inclusive) interval from sos to eos includes all the logical
             * and visual indexes within this sequence. They are logically and
             * visually contiguous and in the same range.
             *
             * For each run, the new visual index=sos+eos-old visual index;
             * we pre-add sos+eos into sumOfSosEos ->
             * new visual index=sumOfSosEos-old visual index;
             */
            sumOfSosEos=start+limit-1;

            /* reorder each index in the sequence */
            do {
                indexMap[start]=sumOfSosEos-indexMap[start];
            } while(++start<limit);

            /* start==limit */
            if(limit==length) {
                break;  /* no more such sequences */
            } else {
                start=limit+1;
            }
        }
    } while(--maxLevel>=minLevel);
}

U_CAPI void U_EXPORT2
ubidi_reorderVisual(const UBiDiLevel *levels, int32_t length, int32_t *indexMap) {
    int32_t start, end, limit, temp;
    UBiDiLevel minLevel = 0, maxLevel = 0;

    if(indexMap==NULL || !prepareReorder(levels, length, indexMap, &minLevel, &maxLevel)) {
        return;
    }

    /* nothing to do? */
    if(minLevel==maxLevel && (minLevel&1)==0) {
        return;
    }

    /* reorder only down to the lowest odd level */
    minLevel|=1;

    /* loop maxLevel..minLevel */
    do {
        start=0;

        /* loop for all sequences of levels to reorder at the current maxLevel */
        for(;;) {
            /* look for a sequence of levels that are all at >=maxLevel */
            /* look for the first index of such a sequence */
            while(start<length && levels[start]<maxLevel) {
                ++start;
            }
            if(start>=length) {
                break;  /* no more such runs */
            }

            /* look for the limit of such a sequence (the index behind it) */
            for(limit=start; ++limit<length && levels[limit]>=maxLevel;) {}

            /*
             * Swap the entire interval of indexes from start to limit-1.
             * We don't need to swap the levels for the purpose of this
             * algorithm: the sequence of levels that we look at does not
             * move anyway.
             */
            end=limit-1;
            while(start<end) {
                temp=indexMap[start];
                indexMap[start]=indexMap[end];
                indexMap[end]=temp;

                ++start;
                --end;
            }

            if(limit==length) {
                break;  /* no more such sequences */
            } else {
                start=limit+1;
            }
        }
    } while(--maxLevel>=minLevel);
}

/* API functions for logical<->visual mapping ------------------------------- */

U_CAPI int32_t U_EXPORT2
ubidi_getVisualIndex(UBiDi *pBiDi, int32_t logicalIndex, UErrorCode *pErrorCode) {
    int32_t visualIndex=UBIDI_MAP_NOWHERE;
    RETURN_IF_NULL_OR_FAILING_ERRCODE(pErrorCode, -1);
    RETURN_IF_NOT_VALID_PARA_OR_LINE(pBiDi, *pErrorCode, -1);
    RETURN_IF_BAD_RANGE(logicalIndex, 0, pBiDi->length, *pErrorCode, -1);

    /* we can do the trivial cases without the runs array */
    switch(pBiDi->direction) {
    case UBIDI_LTR:
        visualIndex=logicalIndex;
        break;
    case UBIDI_RTL:
        visualIndex=pBiDi->length-logicalIndex-1;
        break;
    default:
        if(!ubidi_getRuns(pBiDi, pErrorCode)) {
            *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
            return -1;
        } else {
            Run *runs=pBiDi->runs;
            int32_t i, visualStart=0, offset, length;

            /* linear search for the run, search on the visual runs */
            for(i=0; i<pBiDi->runCount; ++i) {
                length=runs[i].visualLimit-visualStart;
                offset=logicalIndex-GET_INDEX(runs[i].logicalStart);
                if(offset>=0 && offset<length) {
                    if(IS_EVEN_RUN(runs[i].logicalStart)) {
                        /* LTR */
                        visualIndex=visualStart+offset;
                    } else {
                        /* RTL */
                        visualIndex=visualStart+length-offset-1;
                    }
                    break;          /* exit for loop */
                }
                visualStart+=length;
            }
            if(i>=pBiDi->runCount) {
                return UBIDI_MAP_NOWHERE;
            }
        }
    }

    if(pBiDi->insertPoints.size>0) {
        /* add the number of added marks until the calculated visual index */
        Run *runs=pBiDi->runs;
        int32_t i, length, insertRemove;
        int32_t visualStart=0, markFound=0;
        for(i=0; ; i++, visualStart+=length) {
            length=runs[i].visualLimit-visualStart;
            insertRemove=runs[i].insertRemove;
            if(insertRemove & (LRM_BEFORE|RLM_BEFORE)) {
                markFound++;
            }
            /* is it the run containing the visual index? */
            if(visualIndex<runs[i].visualLimit) {
                return visualIndex+markFound;
            }
            if(insertRemove & (LRM_AFTER|RLM_AFTER)) {
                markFound++;
            }
        }
    }
    else if(pBiDi->controlCount>0) {
        /* subtract the number of controls until the calculated visual index */
        Run *runs=pBiDi->runs;
        int32_t i, j, start, limit, length, insertRemove;
        int32_t visualStart=0, controlFound=0;
        UChar uchar=pBiDi->text[logicalIndex];
        /* is the logical index pointing to a control ? */
        if(IS_BIDI_CONTROL_CHAR(uchar)) {
            return UBIDI_MAP_NOWHERE;
        }
        /* loop on runs */
        for(i=0; ; i++, visualStart+=length) {
            length=runs[i].visualLimit-visualStart;
            insertRemove=runs[i].insertRemove;
            /* calculated visual index is beyond this run? */
            if(visualIndex>=runs[i].visualLimit) {
                controlFound-=insertRemove;
                continue;
            }
            /* calculated visual index must be within current run */
            if(insertRemove==0) {
                return visualIndex-controlFound;
            }
            if(IS_EVEN_RUN(runs[i].logicalStart)) {
                /* LTR: check from run start to logical index */
                start=runs[i].logicalStart;
                limit=logicalIndex;
            } else {
                /* RTL: check from logical index to run end */
                start=logicalIndex+1;
                limit=GET_INDEX(runs[i].logicalStart)+length;
            }
            for(j=start; j<limit; j++) {
                uchar=pBiDi->text[j];
                if(IS_BIDI_CONTROL_CHAR(uchar)) {
                    controlFound++;
                }
            }
            return visualIndex-controlFound;
        }
    }

    return visualIndex;
}

U_CAPI int32_t U_EXPORT2
ubidi_getLogicalIndex(UBiDi *pBiDi, int32_t visualIndex, UErrorCode *pErrorCode) {
    Run *runs;
    int32_t i, runCount, start;
    RETURN_IF_NULL_OR_FAILING_ERRCODE(pErrorCode, -1);
    RETURN_IF_NOT_VALID_PARA_OR_LINE(pBiDi, *pErrorCode, -1);
    RETURN_IF_BAD_RANGE(visualIndex, 0, pBiDi->resultLength, *pErrorCode, -1);
    /* we can do the trivial cases without the runs array */
    if(pBiDi->insertPoints.size==0 && pBiDi->controlCount==0) {
        if(pBiDi->direction==UBIDI_LTR) {
            return visualIndex;
        }
        else if(pBiDi->direction==UBIDI_RTL) {
            return pBiDi->length-visualIndex-1;
        }
    }
    if(!ubidi_getRuns(pBiDi, pErrorCode)) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return -1;
    }

    runs=pBiDi->runs;
    runCount=pBiDi->runCount;
    if(pBiDi->insertPoints.size>0) {
        /* handle inserted LRM/RLM */
        int32_t markFound=0, insertRemove;
        int32_t visualStart=0, length;
        runs=pBiDi->runs;
        /* subtract number of marks until visual index */
        for(i=0; ; i++, visualStart+=length) {
            length=runs[i].visualLimit-visualStart;
            insertRemove=runs[i].insertRemove;
            if(insertRemove&(LRM_BEFORE|RLM_BEFORE)) {
                if(visualIndex<=(visualStart+markFound)) {
                    return UBIDI_MAP_NOWHERE;
                }
                markFound++;
            }
            /* is adjusted visual index within this run? */
            if(visualIndex<(runs[i].visualLimit+markFound)) {
                visualIndex-=markFound;
                break;
            }
            if(insertRemove&(LRM_AFTER|RLM_AFTER)) {
                if(visualIndex==(visualStart+length+markFound)) {
                    return UBIDI_MAP_NOWHERE;
                }
                markFound++;
            }
        }
    }
    else if(pBiDi->controlCount>0) {
        /* handle removed BiDi control characters */
        int32_t controlFound=0, insertRemove, length;
        int32_t logicalStart, logicalEnd, visualStart=0, j, k;
        UChar uchar;
        UBool evenRun;
        /* add number of controls until visual index */
        for(i=0; ; i++, visualStart+=length) {
            length=runs[i].visualLimit-visualStart;
            insertRemove=runs[i].insertRemove;
            /* is adjusted visual index beyond current run? */
            if(visualIndex>=(runs[i].visualLimit-controlFound+insertRemove)) {
                controlFound-=insertRemove;
                continue;
            }
            /* adjusted visual index is within current run */
            if(insertRemove==0) {
                visualIndex+=controlFound;
                break;
            }
            /* count non-control chars until visualIndex */
            logicalStart=runs[i].logicalStart;
            evenRun=IS_EVEN_RUN(logicalStart);
            REMOVE_ODD_BIT(logicalStart);
            logicalEnd=logicalStart+length-1;
            for(j=0; j<length; j++) {
                k= evenRun ? logicalStart+j : logicalEnd-j;
                uchar=pBiDi->text[k];
                if(IS_BIDI_CONTROL_CHAR(uchar)) {
                    controlFound++;
                }
                if((visualIndex+controlFound)==(visualStart+j)) {
                    break;
                }
            }
            visualIndex+=controlFound;
            break;
        }
    }
    /* handle all cases */
    if(runCount<=10) {
        /* linear search for the run */
        for(i=0; visualIndex>=runs[i].visualLimit; ++i) {}
    } else {
        /* binary search for the run */
        int32_t begin=0, limit=runCount;

        /* the middle if() is guaranteed to find the run, we don't need a loop limit */
        for(;;) {
            i=(begin+limit)/2;
            if(visualIndex>=runs[i].visualLimit) {
                begin=i+1;
            } else if(i==0 || visualIndex>=runs[i-1].visualLimit) {
                break;
            } else {
                limit=i;
            }
        }
    }

    start=runs[i].logicalStart;
    if(IS_EVEN_RUN(start)) {
        /* LTR */
        /* the offset in runs[i] is visualIndex-runs[i-1].visualLimit */
        if(i>0) {
            visualIndex-=runs[i-1].visualLimit;
        }
        return start+visualIndex;
    } else {
        /* RTL */
        return GET_INDEX(start)+runs[i].visualLimit-visualIndex-1;
    }
}

U_CAPI void U_EXPORT2
ubidi_getLogicalMap(UBiDi *pBiDi, int32_t *indexMap, UErrorCode *pErrorCode) {
    RETURN_VOID_IF_NULL_OR_FAILING_ERRCODE(pErrorCode);
    /* ubidi_countRuns() checks for VALID_PARA_OR_LINE */
    ubidi_countRuns(pBiDi, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        /* no op */
    } else if(indexMap==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
    } else {
        /* fill a logical-to-visual index map using the runs[] */
        int32_t visualStart, visualLimit, i, j, k;
        int32_t logicalStart, logicalLimit;
        Run *runs=pBiDi->runs;
        if (pBiDi->length<=0) {
            return;
        }
        if (pBiDi->length>pBiDi->resultLength) {
            uprv_memset(indexMap, 0xFF, pBiDi->length*sizeof(int32_t));
        }

        visualStart=0;
        for(j=0; j<pBiDi->runCount; ++j) {
            logicalStart=GET_INDEX(runs[j].logicalStart);
            visualLimit=runs[j].visualLimit;
            if(IS_EVEN_RUN(runs[j].logicalStart)) {
                do { /* LTR */
                    indexMap[logicalStart++]=visualStart++;
                } while(visualStart<visualLimit);
            } else {
                logicalStart+=visualLimit-visualStart;  /* logicalLimit */
                do { /* RTL */
                    indexMap[--logicalStart]=visualStart++;
                } while(visualStart<visualLimit);
            }
            /* visualStart==visualLimit; */
        }

        if(pBiDi->insertPoints.size>0) {
            int32_t markFound=0, runCount=pBiDi->runCount;
            int32_t length, insertRemove;
            visualStart=0;
            /* add number of marks found until each index */
            for(i=0; i<runCount; i++, visualStart+=length) {
                length=runs[i].visualLimit-visualStart;
                insertRemove=runs[i].insertRemove;
                if(insertRemove&(LRM_BEFORE|RLM_BEFORE)) {
                    markFound++;
                }
                if(markFound>0) {
                    logicalStart=GET_INDEX(runs[i].logicalStart);
                    logicalLimit=logicalStart+length;
                    for(j=logicalStart; j<logicalLimit; j++) {
                        indexMap[j]+=markFound;
                    }
                }
                if(insertRemove&(LRM_AFTER|RLM_AFTER)) {
                    markFound++;
                }
            }
        }
        else if(pBiDi->controlCount>0) {
            int32_t controlFound=0, runCount=pBiDi->runCount;
            int32_t length, insertRemove;
            UBool evenRun;
            UChar uchar;
            visualStart=0;
            /* subtract number of controls found until each index */
            for(i=0; i<runCount; i++, visualStart+=length) {
                length=runs[i].visualLimit-visualStart;
                insertRemove=runs[i].insertRemove;
                /* no control found within previous runs nor within this run */
                if((controlFound-insertRemove)==0) {
                    continue;
                }
                logicalStart=runs[i].logicalStart;
                evenRun=IS_EVEN_RUN(logicalStart);
                REMOVE_ODD_BIT(logicalStart);
                logicalLimit=logicalStart+length;
                /* if no control within this run */
                if(insertRemove==0) {
                    for(j=logicalStart; j<logicalLimit; j++) {
                        indexMap[j]-=controlFound;
                    }
                    continue;
                }
                for(j=0; j<length; j++) {
                    k= evenRun ? logicalStart+j : logicalLimit-j-1;
                    uchar=pBiDi->text[k];
                    if(IS_BIDI_CONTROL_CHAR(uchar)) {
                        controlFound++;
                        indexMap[k]=UBIDI_MAP_NOWHERE;
                        continue;
                    }
                    indexMap[k]-=controlFound;
                }
            }
        }
    }
}

U_CAPI void U_EXPORT2
ubidi_getVisualMap(UBiDi *pBiDi, int32_t *indexMap, UErrorCode *pErrorCode) {
    RETURN_VOID_IF_NULL_OR_FAILING_ERRCODE(pErrorCode);
    if(indexMap==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    /* ubidi_countRuns() checks for VALID_PARA_OR_LINE */
    ubidi_countRuns(pBiDi, pErrorCode);
    if(U_SUCCESS(*pErrorCode)) {
        /* fill a visual-to-logical index map using the runs[] */
        Run *runs=pBiDi->runs, *runsLimit=runs+pBiDi->runCount;
        int32_t logicalStart, visualStart, visualLimit, *pi=indexMap;

        if (pBiDi->resultLength<=0) {
            return;
        }
        visualStart=0;
        for(; runs<runsLimit; ++runs) {
            logicalStart=runs->logicalStart;
            visualLimit=runs->visualLimit;
            if(IS_EVEN_RUN(logicalStart)) {
                do { /* LTR */
                    *pi++ = logicalStart++;
                } while(++visualStart<visualLimit);
            } else {
                REMOVE_ODD_BIT(logicalStart);
                logicalStart+=visualLimit-visualStart;  /* logicalLimit */
                do { /* RTL */
                    *pi++ = --logicalStart;
                } while(++visualStart<visualLimit);
            }
            /* visualStart==visualLimit; */
        }

        if(pBiDi->insertPoints.size>0) {
            int32_t markFound=0, runCount=pBiDi->runCount;
            int32_t insertRemove, i, j, k;
            runs=pBiDi->runs;
            /* count all inserted marks */
            for(i=0; i<runCount; i++) {
                insertRemove=runs[i].insertRemove;
                if(insertRemove&(LRM_BEFORE|RLM_BEFORE)) {
                    markFound++;
                }
                if(insertRemove&(LRM_AFTER|RLM_AFTER)) {
                    markFound++;
                }
            }
            /* move back indexes by number of preceding marks */
            k=pBiDi->resultLength;
            for(i=runCount-1; i>=0 && markFound>0; i--) {
                insertRemove=runs[i].insertRemove;
                if(insertRemove&(LRM_AFTER|RLM_AFTER)) {
                    indexMap[--k]= UBIDI_MAP_NOWHERE;
                    markFound--;
                }
                visualStart= i>0 ? runs[i-1].visualLimit : 0;
                for(j=runs[i].visualLimit-1; j>=visualStart && markFound>0; j--) {
                    indexMap[--k]=indexMap[j];
                }
                if(insertRemove&(LRM_BEFORE|RLM_BEFORE)) {
                    indexMap[--k]= UBIDI_MAP_NOWHERE;
                    markFound--;
                }
            }
        }
        else if(pBiDi->controlCount>0) {
            int32_t runCount=pBiDi->runCount, logicalEnd;
            int32_t insertRemove, length, i, j, k, m;
            UChar uchar;
            UBool evenRun;
            runs=pBiDi->runs;
            visualStart=0;
            /* move forward indexes by number of preceding controls */
            k=0;
            for(i=0; i<runCount; i++, visualStart+=length) {
                length=runs[i].visualLimit-visualStart;
                insertRemove=runs[i].insertRemove;
                /* if no control found yet, nothing to do in this run */
                if((insertRemove==0)&&(k==visualStart)) {
                    k+=length;
                    continue;
                }
                /* if no control in this run */
                if(insertRemove==0) {
                    visualLimit=runs[i].visualLimit;
                    for(j=visualStart; j<visualLimit; j++) {
                        indexMap[k++]=indexMap[j];
                    }
                    continue;
                }
                logicalStart=runs[i].logicalStart;
                evenRun=IS_EVEN_RUN(logicalStart);
                REMOVE_ODD_BIT(logicalStart);
                logicalEnd=logicalStart+length-1;
                for(j=0; j<length; j++) {
                    m= evenRun ? logicalStart+j : logicalEnd-j;
                    uchar=pBiDi->text[m];
                    if(!IS_BIDI_CONTROL_CHAR(uchar)) {
                        indexMap[k++]=m;
                    }
                }
            }
        }
    }
}

U_CAPI void U_EXPORT2
ubidi_invertMap(const int32_t *srcMap, int32_t *destMap, int32_t length) {
    if(srcMap!=NULL && destMap!=NULL && length>0) {
        const int32_t *pi;
        int32_t destLength=-1, count=0;
        /* find highest value and count positive indexes in srcMap */
        pi=srcMap+length;
        while(pi>srcMap) {
            if(*--pi>destLength) {
                destLength=*pi;
            }
            if(*pi>=0) {
                count++;
            }
        }
        destLength++;           /* add 1 for origin 0 */
        if(count<destLength) {
            /* we must fill unmatched destMap entries with -1 */
            uprv_memset(destMap, 0xFF, destLength*sizeof(int32_t));
        }
        pi=srcMap+length;
        while(length>0) {
            if(*--pi>=0) {
                destMap[*pi]=--length;
            } else {
                --length;
            }
        }
    }
}
/*
******************************************************************************
*
* © 2016 and later: Unicode, Inc. and others.
* License & terms of use: http://www.unicode.org/copyright.html
*
******************************************************************************
*   file name:  ubiditransform.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2016jul24
*   created by: Lina Kemmel
*
*/




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2000-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  ushape.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2000jun29
*   created by: Markus W. Scherer
*/




/**
 * \file
 * \brief C API:  Arabic shaping
 * 
 */

/**
 * Shape Arabic text on a character basis.
 *
 * <p>This function performs basic operations for "shaping" Arabic text. It is most
 * useful for use with legacy data formats and legacy display technology
 * (simple terminals). All operations are performed on Unicode characters.</p>
 *
 * <p>Text-based shaping means that some character code points in the text are
 * replaced by others depending on the context. It transforms one kind of text
 * into another. In comparison, modern displays for Arabic text select
 * appropriate, context-dependent font glyphs for each text element, which means
 * that they transform text into a glyph vector.</p>
 *
 * <p>Text transformations are necessary when modern display technology is not
 * available or when text needs to be transformed to or from legacy formats that
 * use "shaped" characters. Since the Arabic script is cursive, connecting
 * adjacent letters to each other, computers select images for each letter based
 * on the surrounding letters. This usually results in four images per Arabic
 * letter: initial, middle, final, and isolated forms. In Unicode, on the other
 * hand, letters are normally stored abstract, and a display system is expected
 * to select the necessary glyphs. (This makes searching and other text
 * processing easier because the same letter has only one code.) It is possible
 * to mimic this with text transformations because there are characters in
 * Unicode that are rendered as letters with a specific shape
 * (or cursive connectivity). They were included for interoperability with
 * legacy systems and codepages, and for unsophisticated display systems.</p>
 *
 * <p>A second kind of text transformations is supported for Arabic digits:
 * For compatibility with legacy codepages that only include European digits,
 * it is possible to replace one set of digits by another, changing the
 * character code points. These operations can be performed for either
 * Arabic-Indic Digits (U+0660...U+0669) or Eastern (Extended) Arabic-Indic
 * digits (U+06f0...U+06f9).</p>
 *
 * <p>Some replacements may result in more or fewer characters (code points).
 * By default, this means that the destination buffer may receive text with a
 * length different from the source length. Some legacy systems rely on the
 * length of the text to be constant. They expect extra spaces to be added
 * or consumed either next to the affected character or at the end of the
 * text.</p>
 *
 * <p>For details about the available operations, see the description of the
 * <code>U_SHAPE_...</code> options.</p>
 *
 * @param source The input text.
 *
 * @param sourceLength The number of UChars in <code>source</code>.
 *
 * @param dest The destination buffer that will receive the results of the
 *             requested operations. It may be <code>NULL</code> only if
 *             <code>destSize</code> is 0. The source and destination must not
 *             overlap.
 *
 * @param destSize The size (capacity) of the destination buffer in UChars.
 *                 If <code>destSize</code> is 0, then no output is produced,
 *                 but the necessary buffer size is returned ("preflighting").
 *
 * @param options This is a 32-bit set of flags that specify the operations
 *                that are performed on the input text. If no error occurs,
 *                then the result will always be written to the destination
 *                buffer.
 *
 * @param pErrorCode must be a valid pointer to an error code value,
 *        which must not indicate a failure before the function call.
 *
 * @return The number of UChars written to the destination buffer.
 *         If an error occurred, then no output was written, or it may be
 *         incomplete. If <code>U_BUFFER_OVERFLOW_ERROR</code> is set, then
 *         the return value indicates the necessary destination buffer size.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
u_shapeArabic(const UChar *source, int32_t sourceLength,
              UChar *dest, int32_t destSize,
              uint32_t options,
              UErrorCode *pErrorCode);

/**
 * Memory option: allow the result to have a different length than the source.
 * Affects: LamAlef options
 * @stable ICU 2.0
 */
#define U_SHAPE_LENGTH_GROW_SHRINK              0

/**
 * Memory option: allow the result to have a different length than the source.
 * Affects: LamAlef options
 * This option is an alias to U_SHAPE_LENGTH_GROW_SHRINK
 * @stable ICU 4.2
 */
#define U_SHAPE_LAMALEF_RESIZE                  0 

/**
 * Memory option: the result must have the same length as the source.
 * If more room is necessary, then try to consume spaces next to modified characters.
 * @stable ICU 2.0
 */
#define U_SHAPE_LENGTH_FIXED_SPACES_NEAR        1

/**
 * Memory option: the result must have the same length as the source.
 * If more room is necessary, then try to consume spaces next to modified characters.
 * Affects: LamAlef options
 * This option is an alias to U_SHAPE_LENGTH_FIXED_SPACES_NEAR
 * @stable ICU 4.2
 */
#define U_SHAPE_LAMALEF_NEAR                    1 

/**
 * Memory option: the result must have the same length as the source.
 * If more room is necessary, then try to consume spaces at the end of the text.
 * @stable ICU 2.0
 */
#define U_SHAPE_LENGTH_FIXED_SPACES_AT_END      2

/**
 * Memory option: the result must have the same length as the source.
 * If more room is necessary, then try to consume spaces at the end of the text.
 * Affects: LamAlef options
 * This option is an alias to U_SHAPE_LENGTH_FIXED_SPACES_AT_END
 * @stable ICU 4.2
 */
#define U_SHAPE_LAMALEF_END                     2 

/**
 * Memory option: the result must have the same length as the source.
 * If more room is necessary, then try to consume spaces at the beginning of the text.
 * @stable ICU 2.0
 */
#define U_SHAPE_LENGTH_FIXED_SPACES_AT_BEGINNING 3

/**
 * Memory option: the result must have the same length as the source.
 * If more room is necessary, then try to consume spaces at the beginning of the text.
 * Affects: LamAlef options
 * This option is an alias to U_SHAPE_LENGTH_FIXED_SPACES_AT_BEGINNING
 * @stable ICU 4.2
 */
#define U_SHAPE_LAMALEF_BEGIN                    3 


/**
 * Memory option: the result must have the same length as the source.
 * Shaping Mode: For each LAMALEF character found, expand LAMALEF using space at end.
 *               If there is no space at end, use spaces at beginning of the buffer. If there
 *               is no space at beginning of the buffer, use spaces at the near (i.e. the space
 *               after the LAMALEF character).
 *               If there are no spaces found, an error U_NO_SPACE_AVAILABLE (as defined in utypes.h) 
 *               will be set in pErrorCode
 *
 * Deshaping Mode: Perform the same function as the flag equals U_SHAPE_LAMALEF_END. 
 * Affects: LamAlef options
 * @stable ICU 4.2
 */
#define U_SHAPE_LAMALEF_AUTO                     0x10000 

/** Bit mask for memory options. @stable ICU 2.0 */
#define U_SHAPE_LENGTH_MASK                      0x10003 /* Changed old value 3 */


/**
 * Bit mask for LamAlef memory options.
 * @stable ICU 4.2
 */
#define U_SHAPE_LAMALEF_MASK                     0x10003 /* updated */

/** Direction indicator: the source is in logical (keyboard) order. @stable ICU 2.0 */
#define U_SHAPE_TEXT_DIRECTION_LOGICAL          0

/**
 * Direction indicator:
 * the source is in visual RTL order,
 * the rightmost displayed character stored first.
 * This option is an alias to U_SHAPE_TEXT_DIRECTION_LOGICAL
 * @stable ICU 4.2
 */
#define U_SHAPE_TEXT_DIRECTION_VISUAL_RTL       0

/**
 * Direction indicator:
 * the source is in visual LTR order,
 * the leftmost displayed character stored first.
 * @stable ICU 2.0
 */
#define U_SHAPE_TEXT_DIRECTION_VISUAL_LTR       4

/** Bit mask for direction indicators. @stable ICU 2.0 */
#define U_SHAPE_TEXT_DIRECTION_MASK             4


/** Letter shaping option: do not perform letter shaping. @stable ICU 2.0 */
#define U_SHAPE_LETTERS_NOOP                    0

/** Letter shaping option: replace abstract letter characters by "shaped" ones. @stable ICU 2.0 */
#define U_SHAPE_LETTERS_SHAPE                   8

/** Letter shaping option: replace "shaped" letter characters by abstract ones. @stable ICU 2.0 */
#define U_SHAPE_LETTERS_UNSHAPE                 0x10

/**
 * Letter shaping option: replace abstract letter characters by "shaped" ones.
 * The only difference with U_SHAPE_LETTERS_SHAPE is that Tashkeel letters
 * are always "shaped" into the isolated form instead of the medial form
 * (selecting code points from the Arabic Presentation Forms-B block).
 * @stable ICU 2.0
 */
#define U_SHAPE_LETTERS_SHAPE_TASHKEEL_ISOLATED 0x18


/** Bit mask for letter shaping options. @stable ICU 2.0 */
#define U_SHAPE_LETTERS_MASK                        0x18


/** Digit shaping option: do not perform digit shaping. @stable ICU 2.0 */
#define U_SHAPE_DIGITS_NOOP                     0

/**
 * Digit shaping option:
 * Replace European digits (U+0030...) by Arabic-Indic digits.
 * @stable ICU 2.0
 */
#define U_SHAPE_DIGITS_EN2AN                    0x20

/**
 * Digit shaping option:
 * Replace Arabic-Indic digits by European digits (U+0030...).
 * @stable ICU 2.0
 */
#define U_SHAPE_DIGITS_AN2EN                    0x40

/**
 * Digit shaping option:
 * Replace European digits (U+0030...) by Arabic-Indic digits if the most recent
 * strongly directional character is an Arabic letter
 * (<code>u_charDirection()</code> result <code>U_RIGHT_TO_LEFT_ARABIC</code> [AL]).<br>
 * The direction of "preceding" depends on the direction indicator option.
 * For the first characters, the preceding strongly directional character
 * (initial state) is assumed to be not an Arabic letter
 * (it is <code>U_LEFT_TO_RIGHT</code> [L] or <code>U_RIGHT_TO_LEFT</code> [R]).
 * @stable ICU 2.0
 */
#define U_SHAPE_DIGITS_ALEN2AN_INIT_LR          0x60

/**
 * Digit shaping option:
 * Replace European digits (U+0030...) by Arabic-Indic digits if the most recent
 * strongly directional character is an Arabic letter
 * (<code>u_charDirection()</code> result <code>U_RIGHT_TO_LEFT_ARABIC</code> [AL]).<br>
 * The direction of "preceding" depends on the direction indicator option.
 * For the first characters, the preceding strongly directional character
 * (initial state) is assumed to be an Arabic letter.
 * @stable ICU 2.0
 */
#define U_SHAPE_DIGITS_ALEN2AN_INIT_AL          0x80

/** Not a valid option value. May be replaced by a new option. @stable ICU 2.0 */
#define U_SHAPE_DIGITS_RESERVED                 0xa0

/** Bit mask for digit shaping options. @stable ICU 2.0 */
#define U_SHAPE_DIGITS_MASK                     0xe0


/** Digit type option: Use Arabic-Indic digits (U+0660...U+0669). @stable ICU 2.0 */
#define U_SHAPE_DIGIT_TYPE_AN                   0

/** Digit type option: Use Eastern (Extended) Arabic-Indic digits (U+06f0...U+06f9). @stable ICU 2.0 */
#define U_SHAPE_DIGIT_TYPE_AN_EXTENDED          0x100

/** Not a valid option value. May be replaced by a new option. @stable ICU 2.0 */
#define U_SHAPE_DIGIT_TYPE_RESERVED             0x200

/** Bit mask for digit type options. @stable ICU 2.0 */
#define U_SHAPE_DIGIT_TYPE_MASK                 0x300 /* I need to change this from 0x3f00 to 0x300 */

/** 
 * Tashkeel aggregation option:
 * Replaces any combination of U+0651 with one of
 * U+064C, U+064D, U+064E, U+064F, U+0650 with
 * U+FC5E, U+FC5F, U+FC60, U+FC61, U+FC62 consecutively.
 * @stable ICU 3.6
 */
#define U_SHAPE_AGGREGATE_TASHKEEL              0x4000
/** Tashkeel aggregation option: do not aggregate tashkeels. @stable ICU 3.6 */
#define U_SHAPE_AGGREGATE_TASHKEEL_NOOP         0
/** Bit mask for tashkeel aggregation. @stable ICU 3.6 */
#define U_SHAPE_AGGREGATE_TASHKEEL_MASK         0x4000

/** 
 * Presentation form option:
 * Don't replace Arabic Presentation Forms-A and Arabic Presentation Forms-B
 * characters with 0+06xx characters, before shaping.
 * @stable ICU 3.6
 */
#define U_SHAPE_PRESERVE_PRESENTATION           0x8000
/** Presentation form option: 
 * Replace Arabic Presentation Forms-A and Arabic Presentationo Forms-B with 
 * their unshaped correspondants in range 0+06xx, before shaping.
 * @stable ICU 3.6 
 */
#define U_SHAPE_PRESERVE_PRESENTATION_NOOP      0
/** Bit mask for preserve presentation form. @stable ICU 3.6 */
#define U_SHAPE_PRESERVE_PRESENTATION_MASK      0x8000

/* Seen Tail option */ 
/**
 * Memory option: the result must have the same length as the source.
 * Shaping mode: The SEEN family character will expand into two characters using space near 
 *               the SEEN family character(i.e. the space after the character).
 *               If there are no spaces found, an error U_NO_SPACE_AVAILABLE (as defined in utypes.h) 
 *               will be set in pErrorCode
 *
 * De-shaping mode: Any Seen character followed by Tail character will be
 *                  replaced by one cell Seen and a space will replace the Tail.
 * Affects: Seen options
 * @stable ICU 4.2
 */
#define U_SHAPE_SEEN_TWOCELL_NEAR     0x200000

/**
 * Bit mask for Seen memory options. 
 * @stable ICU 4.2
 */
#define U_SHAPE_SEEN_MASK             0x700000

/* YehHamza option */ 
/**
 * Memory option: the result must have the same length as the source.
 * Shaping mode: The YEHHAMZA character will expand into two characters using space near it 
 *              (i.e. the space after the character
 *               If there are no spaces found, an error U_NO_SPACE_AVAILABLE (as defined in utypes.h) 
 *               will be set in pErrorCode
 *
 * De-shaping mode: Any Yeh (final or isolated) character followed by Hamza character will be
 *                  replaced by one cell YehHamza and space will replace the Hamza.
 * Affects: YehHamza options
 * @stable ICU 4.2
 */
#define U_SHAPE_YEHHAMZA_TWOCELL_NEAR      0x1000000


/**
 * Bit mask for YehHamza memory options. 
 * @stable ICU 4.2
 */
#define U_SHAPE_YEHHAMZA_MASK              0x3800000

/* New Tashkeel options */ 
/**
 * Memory option: the result must have the same length as the source.
 * Shaping mode: Tashkeel characters will be replaced by spaces. 
 *               Spaces will be placed at beginning of the buffer
 *
 * De-shaping mode: N/A
 * Affects: Tashkeel options
 * @stable ICU 4.2
 */
#define U_SHAPE_TASHKEEL_BEGIN                      0x40000

/**
 * Memory option: the result must have the same length as the source.
 * Shaping mode: Tashkeel characters will be replaced by spaces. 
 *               Spaces will be placed at end of the buffer
 *
 * De-shaping mode: N/A
 * Affects: Tashkeel options
 * @stable ICU 4.2
 */
#define U_SHAPE_TASHKEEL_END                        0x60000

/**
 * Memory option: allow the result to have a different length than the source.
 * Shaping mode: Tashkeel characters will be removed, buffer length will shrink. 
 * De-shaping mode: N/A 
 *
 * Affect: Tashkeel options
 * @stable ICU 4.2
 */
#define U_SHAPE_TASHKEEL_RESIZE                     0x80000

/**
 * Memory option: the result must have the same length as the source.
 * Shaping mode: Tashkeel characters will be replaced by Tatweel if it is connected to adjacent
 *               characters (i.e. shaped on Tatweel) or replaced by space if it is not connected.
 *
 * De-shaping mode: N/A
 * Affects: YehHamza options
 * @stable ICU 4.2
 */
#define U_SHAPE_TASHKEEL_REPLACE_BY_TATWEEL         0xC0000

/** 
 * Bit mask for Tashkeel replacement with Space or Tatweel memory options. 
 * @stable ICU 4.2
 */
#define U_SHAPE_TASHKEEL_MASK                       0xE0000


/* Space location Control options */ 
/**
 * This option affect the meaning of BEGIN and END options. if this option is not used the default
 * for BEGIN and END will be as following: 
 * The Default (for both Visual LTR, Visual RTL and Logical Text)
 *           1. BEGIN always refers to the start address of physical memory.
 *           2. END always refers to the end address of physical memory.
 *
 * If this option is used it will swap the meaning of BEGIN and END only for Visual LTR text. 
 *
 * The effect on BEGIN and END Memory Options will be as following:
 *    A. BEGIN For Visual LTR text: This will be the beginning (right side) of the visual text(
 *       corresponding to the physical memory address end for Visual LTR text, Same as END in 
 *       default behavior)
 *    B. BEGIN For Logical text: Same as BEGIN in default behavior. 
 *    C. END For Visual LTR text: This will be the end (left side) of the visual text (corresponding
 *       to the physical memory address beginning for Visual LTR text, Same as BEGIN in default behavior.
 *    D. END For Logical text: Same as END in default behavior). 
 * Affects: All LamAlef BEGIN, END and AUTO options.
 * @stable ICU 4.2
 */
#define U_SHAPE_SPACES_RELATIVE_TO_TEXT_BEGIN_END 0x4000000

/**
 * Bit mask for swapping BEGIN and END for Visual LTR text 
 * @stable ICU 4.2
 */
#define U_SHAPE_SPACES_RELATIVE_TO_TEXT_MASK      0x4000000

/**
 * If this option is used, shaping will use the new Unicode code point for TAIL (i.e. 0xFE73). 
 * If this option is not specified (Default), old unofficial Unicode TAIL code point is used (i.e. 0x200B)
 * De-shaping will not use this option as it will always search for both the new Unicode code point for the 
 * TAIL (i.e. 0xFE73) or the old unofficial Unicode TAIL code point (i.e. 0x200B) and de-shape the
 * Seen-Family letter accordingly.
 *
 * Shaping Mode: Only shaping.
 * De-shaping Mode: N/A.
 * Affects: All Seen options
 * @stable ICU 4.8
 */
#define U_SHAPE_TAIL_NEW_UNICODE        0x8000000

/**
 * Bit mask for new Unicode Tail option 
 * @stable ICU 4.8
 */
#define U_SHAPE_TAIL_TYPE_MASK          0x8000000




/*
******************************************************************************
*
* © 2016 and later: Unicode, Inc. and others.
* License & terms of use: http://www.unicode.org/copyright.html
*
******************************************************************************
*   file name:  ubiditransform.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2016jul24
*   created by: Lina Kemmel
*
*/







/**
 * \file
 * \brief Bidi Transformations
 */

/**
 * `UBiDiOrder` indicates the order of text.
 *
 * This bidi transformation engine supports all possible combinations (4 in
 * total) of input and output text order:
 *
 *   - <logical input, visual output>: unless the output direction is RTL, this
 *     corresponds to a normal operation of the Bidi algorithm as described in the
 *     Unicode Technical Report and implemented by `UBiDi` when the
 *     reordering mode is set to `UBIDI_REORDER_DEFAULT`. Visual RTL
 *     mode is not supported by `UBiDi` and is accomplished through
 *     reversing a visual LTR string,
 *
 *   - <visual input, logical output>: unless the input direction is RTL, this
 *     corresponds to an "inverse bidi algorithm" in `UBiDi` with the
 *     reordering mode set to `UBIDI_REORDER_INVERSE_LIKE_DIRECT`.
 *     Visual RTL mode is not not supported by `UBiDi` and is
 *     accomplished through reversing a visual LTR string,
 *
 *   - <logical input, logical output>: if the input and output base directions
 *     mismatch, this corresponds to the `UBiDi` implementation with the
 *     reordering mode set to `UBIDI_REORDER_RUNS_ONLY`; and if the
 *     input and output base directions are identical, the transformation engine
 *     will only handle character mirroring and Arabic shaping operations without
 *     reordering,
 *
 *   - <visual input, visual output>: this reordering mode is not supported by
 *     the `UBiDi` engine; it implies character mirroring, Arabic
 *     shaping, and - if the input/output base directions mismatch -  string
 *     reverse operations.
 * @see ubidi_setInverse
 * @see ubidi_setReorderingMode
 * @see UBIDI_REORDER_DEFAULT
 * @see UBIDI_REORDER_INVERSE_LIKE_DIRECT
 * @see UBIDI_REORDER_RUNS_ONLY
 * @stable ICU 58
 */
typedef enum {
    /** 0: Constant indicating a logical order.
      * This is the default for input text.
      * @stable ICU 58
      */
    UBIDI_LOGICAL = 0,
    /** 1: Constant indicating a visual order.
      * This is a default for output text.
      * @stable ICU 58
      */
    UBIDI_VISUAL
} UBiDiOrder;

/**
 * <code>UBiDiMirroring</code> indicates whether or not characters with the
 * "mirrored" property in RTL runs should be replaced with their mirror-image
 * counterparts.
 * @see UBIDI_DO_MIRRORING
 * @see ubidi_setReorderingOptions
 * @see ubidi_writeReordered
 * @see ubidi_writeReverse
 * @stable ICU 58
 */
typedef enum {
    /** 0: Constant indicating that character mirroring should not be
      * performed.
      * This is the default.
      * @stable ICU 58
      */
    UBIDI_MIRRORING_OFF = 0,
    /** 1: Constant indicating that character mirroring should be performed.
      * This corresponds to calling <code>ubidi_writeReordered</code> or
      * <code>ubidi_writeReverse</code> with the
      * <code>UBIDI_DO_MIRRORING</code> option bit set.
      * @stable ICU 58
      */
    UBIDI_MIRRORING_ON
} UBiDiMirroring;

/**
 * Forward declaration of the <code>UBiDiTransform</code> structure that stores
 * information used by the layout transformation engine.
 * @stable ICU 58
 */
typedef struct UBiDiTransform UBiDiTransform;

/**
 * Performs transformation of text from the bidi layout defined by the input
 * ordering scheme to the bidi layout defined by the output ordering scheme,
 * and applies character mirroring and Arabic shaping operations.<p>
 * In terms of <code>UBiDi</code>, such a transformation implies:
 * <ul>
 * <li>calling <code>ubidi_setReorderingMode</code> as needed (when the
 * reordering mode is other than normal),</li>
 * <li>calling <code>ubidi_setInverse</code> as needed (when text should be
 * transformed from a visual to a logical form),</li>
 * <li>resolving embedding levels of each character in the input text by
 * calling <code>ubidi_setPara</code>,</li>
 * <li>reordering the characters based on the computed embedding levels, also
 * performing character mirroring as needed, and streaming the result to the
 * output, by calling <code>ubidi_writeReordered</code>,</li>
 * <li>performing Arabic digit and letter shaping on the output text by calling
 * <code>u_shapeArabic</code>.</li>
 * </ul>
 * An "ordering scheme" encompasses the base direction and the order of text,
 * and these characteristics must be defined by the caller for both input and
 * output explicitly .<p>
 * There are 36 possible combinations of <input, output> ordering schemes,
 * which are partially supported by <code>UBiDi</code> already. Examples of the
 * currently supported combinations:
 * <ul>
 * <li><Logical LTR, Visual LTR>: this is equivalent to calling
 * <code>ubidi_setPara</code> with <code>paraLevel == UBIDI_LTR</code>,</li>
 * <li><Logical RTL, Visual LTR>: this is equivalent to calling
 * <code>ubidi_setPara</code> with <code>paraLevel == UBIDI_RTL</code>,</li>
 * <li><Logical Default ("Auto") LTR, Visual LTR>: this is equivalent to
 * calling <code>ubidi_setPara</code> with
 * <code>paraLevel == UBIDI_DEFAULT_LTR</code>,</li>
 * <li><Logical Default ("Auto") RTL, Visual LTR>: this is equivalent to
 * calling <code>ubidi_setPara</code> with
 * <code>paraLevel == UBIDI_DEFAULT_RTL</code>,</li>
 * <li><Visual LTR, Logical LTR>: this is equivalent to
 * calling <code>ubidi_setInverse(UBiDi*, TRUE)</code> and then
 * <code>ubidi_setPara</code> with <code>paraLevel == UBIDI_LTR</code>,</li>
 * <li><Visual LTR, Logical RTL>: this is equivalent to
 * calling <code>ubidi_setInverse(UBiDi*, TRUE)</code> and then
 * <code>ubidi_setPara</code> with <code>paraLevel == UBIDI_RTL</code>.</li>
 * </ul>
 * All combinations that involve the Visual RTL scheme are unsupported by
 * <code>UBiDi</code>, for instance:
 * <ul>
 * <li><Logical LTR, Visual RTL>,</li>
 * <li><Visual RTL, Logical RTL>.</li>
 * </ul>
 * <p>Example of usage of the transformation engine:<br>
 * <pre>
 * \code
 * UChar text1[] = {'a', 'b', 'c', 0x0625, '1', 0};
 * UChar text2[] = {'a', 'b', 'c', 0x0625, '1', 0};
 * UErrorCode errorCode = U_ZERO_ERROR;
 * // Run a transformation.
 * ubiditransform_transform(pBidiTransform,
 *          text1, -1, text2, -1,
 *          UBIDI_LTR, UBIDI_VISUAL,
 *          UBIDI_RTL, UBIDI_LOGICAL,
 *          UBIDI_MIRRORING_OFF,
 *          U_SHAPE_DIGITS_AN2EN | U_SHAPE_DIGIT_TYPE_AN_EXTENDED,
 *          &errorCode);
 * // Do something with text2.
 *  text2[4] = '2';
 * // Run a reverse transformation.
 * ubiditransform_transform(pBidiTransform,
 *          text2, -1, text1, -1,
 *          UBIDI_RTL, UBIDI_LOGICAL,
 *          UBIDI_LTR, UBIDI_VISUAL,
 *          UBIDI_MIRRORING_OFF,
 *          U_SHAPE_DIGITS_EN2AN | U_SHAPE_DIGIT_TYPE_AN_EXTENDED,
 *          &errorCode);
 *\endcode
 * </pre>
 * </p>
 *
 * @param pBiDiTransform A pointer to a <code>UBiDiTransform</code> object
 *        allocated with <code>ubiditransform_open()</code> or
 *        <code>NULL</code>.<p>
 *        This object serves for one-time setup to amortize initialization
 *        overheads. Use of this object is not thread-safe. All other threads
 *        should allocate a new <code>UBiDiTransform</code> object by calling
 *        <code>ubiditransform_open()</code> before using it. Alternatively,
 *        a caller can set this parameter to <code>NULL</code>, in which case
 *        the object will be allocated by the engine on the fly.</p>
 * @param src A pointer to the text that the Bidi layout transformations will
 *        be performed on.
 *        <p><strong>Note:</strong> the text must be (at least)
 *        <code>srcLength</code> long.</p>
 * @param srcLength The length of the text, in number of UChars. If
 *        <code>length == -1</code> then the text must be zero-terminated.
 * @param dest A pointer to where the processed text is to be copied.
 * @param destSize The size of the <code>dest</code> buffer, in number of
 *        UChars. If the <code>U_SHAPE_LETTERS_UNSHAPE</code> option is set,
 *        then the destination length could be as large as
 *        <code>srcLength * 2</code>. Otherwise, the destination length will
 *        not exceed <code>srcLength</code>. If the caller reserves the last
 *        position for zero-termination, it should be excluded from
 *        <code>destSize</code>.
 *        <p><code>destSize == -1</code> is allowed and makes sense when
 *        <code>dest</code> was holds some meaningful value, e.g. that of
 *        <code>src</code>. In this case <code>dest</code> must be
 *        zero-terminated.</p>
 * @param inParaLevel A base embedding level of the input as defined in
 *        <code>ubidi_setPara</code> documentation for the
 *        <code>paraLevel</code> parameter.
 * @param inOrder An order of the input, which can be one of the
 *        <code>UBiDiOrder</code> values.
 * @param outParaLevel A base embedding level of the output as defined in
 *        <code>ubidi_setPara</code> documentation for the
 *        <code>paraLevel</code> parameter.
 * @param outOrder An order of the output, which can be one of the
 *        <code>UBiDiOrder</code> values.
 * @param doMirroring Indicates whether or not to perform character mirroring,
 *        and can accept one of the <code>UBiDiMirroring</code> values.
 * @param shapingOptions Arabic digit and letter shaping options defined in the
 *        ushape.h documentation.
 *        <p><strong>Note:</strong> Direction indicator options are computed by
 *        the transformation engine based on the effective ordering schemes, so
 *        user-defined direction indicators will be ignored.</p>
 * @param pErrorCode A pointer to an error code value.
 *
 * @return The destination length, i.e. the number of UChars written to
 *         <code>dest</code>. If the transformation fails, the return value
 *         will be 0 (and the error code will be written to
 *         <code>pErrorCode</code>).
 *
 * @see UBiDiLevel
 * @see UBiDiOrder
 * @see UBiDiMirroring
 * @see ubidi_setPara
 * @see u_shapeArabic
 * @stable ICU 58
 */
U_STABLE uint32_t U_EXPORT2
ubiditransform_transform(UBiDiTransform *pBiDiTransform,
            const UChar *src, int32_t srcLength,
            UChar *dest, int32_t destSize,
            UBiDiLevel inParaLevel, UBiDiOrder inOrder,
            UBiDiLevel outParaLevel, UBiDiOrder outOrder,
            UBiDiMirroring doMirroring, uint32_t shapingOptions,
            UErrorCode *pErrorCode);

/**
 * Allocates a <code>UBiDiTransform</code> object. This object can be reused,
 * e.g. with different ordering schemes, mirroring or shaping options.<p>
 * <strong>Note:</strong>The object can only be reused in the same thread.
 * All other threads should allocate a new <code>UBiDiTransform</code> object
 * before using it.<p>
 * Example of usage:<p>
 * <pre>
 * \code
 * UErrorCode errorCode = U_ZERO_ERROR;
 * // Open a new UBiDiTransform.
 * UBiDiTransform* transform = ubiditransform_open(&errorCode);
 * // Run a transformation.
 * ubiditransform_transform(transform,
 *          text1, -1, text2, -1,
 *          UBIDI_RTL, UBIDI_LOGICAL,
 *          UBIDI_LTR, UBIDI_VISUAL,
 *          UBIDI_MIRRORING_ON,
 *          U_SHAPE_DIGITS_EN2AN,
 *          &errorCode);
 * // Do something with the output text and invoke another transformation using
 * //   that text as input.
 * ubiditransform_transform(transform,
 *          text2, -1, text3, -1,
 *          UBIDI_LTR, UBIDI_VISUAL,
 *          UBIDI_RTL, UBIDI_VISUAL,
 *          UBIDI_MIRRORING_ON,
 *          0, &errorCode);
 *\endcode
 * </pre>
 * <p>
 * The <code>UBiDiTransform</code> object must be deallocated by calling
 * <code>ubiditransform_close()</code>.
 *
 * @return An empty <code>UBiDiTransform</code> object.
 * @stable ICU 58
 */
U_STABLE UBiDiTransform* U_EXPORT2
ubiditransform_open(UErrorCode *pErrorCode);

/**
 * Deallocates the given <code>UBiDiTransform</code> object.
 * @stable ICU 58
 */
U_STABLE void U_EXPORT2
ubiditransform_close(UBiDiTransform *pBidiTransform);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUBiDiTransformPointer
 * "Smart pointer" class, closes a UBiDiTransform via ubiditransform_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 58
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUBiDiTransformPointer, UBiDiTransform, ubiditransform_close);

U_NAMESPACE_END

#endif



/* Some convenience defines */
#define LTR                     UBIDI_LTR
#define RTL                     UBIDI_RTL
#define LOGICAL                 UBIDI_LOGICAL
#define VISUAL                  UBIDI_VISUAL
#define SHAPE_LOGICAL           U_SHAPE_TEXT_DIRECTION_LOGICAL
#define SHAPE_VISUAL            U_SHAPE_TEXT_DIRECTION_VISUAL_LTR

#define CHECK_LEN(STR, LEN, ERROR) UPRV_BLOCK_MACRO_BEGIN { \
    if (LEN == 0) return 0; \
    if (LEN < -1) { *(ERROR) = U_ILLEGAL_ARGUMENT_ERROR; return 0; } \
    if (LEN == -1) LEN = u_strlen(STR); \
} UPRV_BLOCK_MACRO_END

#define MAX_ACTIONS     7

/**
 * Typedef for a pointer to a function, which performs some operation (such as
 * reordering, setting "inverse" mode, character mirroring, etc.). Return value
 * indicates whether the text was changed in the course of this operation or
 * not.
 */
typedef UBool (*UBiDiAction)(UBiDiTransform *, UErrorCode *);

/**
 * Structure that holds a predefined reordering scheme, including the following
 * information:
 * <ul>
 * <li>an input base direction,</li>
 * <li>an input order,</li>
 * <li>an output base direction,</li>
 * <li>an output order,</li>
 * <li>a digit shaping direction,</li>
 * <li>a letter shaping direction,</li>
 * <li>a base direction that should be applied when the reordering engine is
 *     invoked (which can not always be derived from the caller-defined
 *     options),</li>
 * <li>an array of pointers to functions that accomplish the bidi layout
 *     transformation.</li>
 * </ul>
 */
typedef struct {
    UBiDiLevel        inLevel;               /* input level */
    UBiDiOrder        inOrder;               /* input order */
    UBiDiLevel        outLevel;              /* output level */
    UBiDiOrder        outOrder;              /* output order */
    uint32_t          digitsDir;             /* digit shaping direction */
    uint32_t          lettersDir;            /* letter shaping direction */
    UBiDiLevel        baseLevel;             /* paragraph level to be used with setPara */
    const UBiDiAction actions[MAX_ACTIONS];  /* array of pointers to functions carrying out the transformation */
} ReorderingScheme;

struct UBiDiTransform {
    UBiDi                   *pBidi;             /* pointer to a UBiDi object */
    const ReorderingScheme  *pActiveScheme;     /* effective reordering scheme */
    UChar                   *src;               /* input text */
    UChar                   *dest;              /* output text */
    uint32_t                srcLength;          /* input text length - not really needed as we are zero-terminated and can u_strlen */
    uint32_t                srcSize;            /* input text capacity excluding the trailing zero */
    uint32_t                destSize;           /* output text capacity */
    uint32_t                *pDestLength;       /* number of UChars written to dest */
    uint32_t                reorderingOptions;  /* reordering options - currently only suppot DO_MIRRORING */
    uint32_t                digits;             /* digit option for ArabicShaping */
    uint32_t                letters;            /* letter option for ArabicShaping */
};

U_DRAFT UBiDiTransform* U_EXPORT2
ubiditransform_open(UErrorCode *pErrorCode)
{
    UBiDiTransform *pBiDiTransform = NULL;
    if (U_SUCCESS(*pErrorCode)) {
        pBiDiTransform = (UBiDiTransform*) uprv_calloc(1, sizeof(UBiDiTransform));
        if (pBiDiTransform == NULL) {
            *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
        }
    }
    return pBiDiTransform;
}

U_DRAFT void U_EXPORT2
ubiditransform_close(UBiDiTransform *pBiDiTransform)
{
    if (pBiDiTransform != NULL) {
        if (pBiDiTransform->pBidi != NULL) {
            ubidi_close(pBiDiTransform->pBidi);
        }
        if (pBiDiTransform->src != NULL) {
            uprv_free(pBiDiTransform->src);
        }
        uprv_free(pBiDiTransform);
    }
}

/**
 * Performs Bidi resolution of text.
 * 
 * @param pTransform Pointer to the <code>UBiDiTransform</code> structure.
 * @param pErrorCode Pointer to the error code value.
 *
 * @return Whether or not this function modifies the text. Besides the return
 * value, the caller should also check <code>U_SUCCESS(*pErrorCode)</code>.
 */
static UBool
action_resolve(UBiDiTransform *pTransform, UErrorCode *pErrorCode)
{
    ubidi_setPara(pTransform->pBidi, pTransform->src, pTransform->srcLength,
            pTransform->pActiveScheme->baseLevel, NULL, pErrorCode);
    return FALSE;
}

/**
 * Performs basic reordering of text (Logical -> Visual LTR).
 * 
 * @param pTransform Pointer to the <code>UBiDiTransform</code> structure.
 * @param pErrorCode Pointer to the error code value.
 *
 * @return Whether or not this function modifies the text. Besides the return
 * value, the caller should also check <code>U_SUCCESS(*pErrorCode)</code>.
 */
static UBool
action_reorder(UBiDiTransform *pTransform, UErrorCode *pErrorCode)
{
    ubidi_writeReordered(pTransform->pBidi, pTransform->dest, pTransform->destSize,
            static_cast<uint16_t>(pTransform->reorderingOptions), pErrorCode);

    *pTransform->pDestLength = pTransform->srcLength;
    pTransform->reorderingOptions = UBIDI_REORDER_DEFAULT;
    return TRUE;
}

/**
 * Sets "inverse" mode on the <code>UBiDi</code> object.
 * 
 * @param pTransform Pointer to the <code>UBiDiTransform</code> structure.
 * @param pErrorCode Pointer to the error code value.
 *
 * @return Whether or not this function modifies the text. Besides the return
 * value, the caller should also check <code>U_SUCCESS(*pErrorCode)</code>.
 */
static UBool
action_setInverse(UBiDiTransform *pTransform, UErrorCode *pErrorCode)
{
    (void)pErrorCode;
    ubidi_setInverse(pTransform->pBidi, TRUE);
    ubidi_setReorderingMode(pTransform->pBidi, UBIDI_REORDER_INVERSE_LIKE_DIRECT);
    return FALSE;
}

/**
 * Sets "runs only" reordering mode indicating a Logical LTR <-> Logical RTL
 * transformation.
 * 
 * @param pTransform Pointer to the <code>UBiDiTransform</code> structure.
 * @param pErrorCode Pointer to the error code value.
 *
 * @return Whether or not this function modifies the text. Besides the return
 * value, the caller should also check <code>U_SUCCESS(*pErrorCode)</code>.
 */
static UBool
action_setRunsOnly(UBiDiTransform *pTransform, UErrorCode *pErrorCode)
{
    (void)pErrorCode;
    ubidi_setReorderingMode(pTransform->pBidi, UBIDI_REORDER_RUNS_ONLY);
    return FALSE;
}

/**
 * Performs string reverse.
 * 
 * @param pTransform Pointer to the <code>UBiDiTransform</code> structure.
 * @param pErrorCode Pointer to the error code value.
 *
 * @return Whether or not this function modifies the text. Besides the return
 * value, the caller should also check <code>U_SUCCESS(*pErrorCode)</code>.
 */
static UBool
action_reverse(UBiDiTransform *pTransform, UErrorCode *pErrorCode)
{
    ubidi_writeReverse(pTransform->src, pTransform->srcLength,
            pTransform->dest, pTransform->destSize,
            UBIDI_REORDER_DEFAULT, pErrorCode);
    *pTransform->pDestLength = pTransform->srcLength;
    return TRUE;
}

/**
 * Applies a new value to the text that serves as input at the current
 * processing step. This value is identical to the original one when we begin
 * the processing, but usually changes as the transformation progresses.
 * 
 * @param pTransform A pointer to the <code>UBiDiTransform</code> structure.
 * @param newSrc A pointer whose value is to be used as input text.
 * @param newLength A length of the new text in <code>UChar</code>s.
 * @param newSize A new source capacity in <code>UChar</code>s.
 * @param pErrorCode Pointer to the error code value.
 */
static void
updateSrc(UBiDiTransform *pTransform, const UChar *newSrc, uint32_t newLength,
        uint32_t newSize, UErrorCode *pErrorCode)
{
    if (newSize < newLength) {
        *pErrorCode = U_BUFFER_OVERFLOW_ERROR;
        return;
    }
    if (newSize > pTransform->srcSize) {
        newSize += 50; // allocate slightly more than needed right now
        if (pTransform->src != NULL) {
            uprv_free(pTransform->src);
            pTransform->src = NULL;
        }
        pTransform->src = (UChar *)uprv_malloc(newSize * sizeof(UChar));
        if (pTransform->src == NULL) {
            *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
            //pTransform->srcLength = pTransform->srcSize = 0;
            return;
        }
        pTransform->srcSize = newSize;
    }
    u_strncpy(pTransform->src, newSrc, newLength);
    pTransform->srcLength = u_terminateUChars(pTransform->src,
    		pTransform->srcSize, newLength, pErrorCode);
}

/**
 * Calls a lower level shaping function.
 * 
 * @param pTransform Pointer to the <code>UBiDiTransform</code> structure.
 * @param options Shaping options.
 * @param pErrorCode Pointer to the error code value.
 */
static void
doShape(UBiDiTransform *pTransform, uint32_t options, UErrorCode *pErrorCode)
{
    *pTransform->pDestLength = u_shapeArabic(pTransform->src,
            pTransform->srcLength, pTransform->dest, pTransform->destSize,
            options, pErrorCode);
}

/**
 * Performs digit and letter shaping.
 * 
 * @param pTransform Pointer to the <code>UBiDiTransform</code> structure.
 * @param pErrorCode Pointer to the error code value.
 *
 * @return Whether or not this function modifies the text. Besides the return
 * value, the caller should also check <code>U_SUCCESS(*pErrorCode)</code>.
 */
static UBool
action_shapeArabic(UBiDiTransform *pTransform, UErrorCode *pErrorCode)
{
    if ((pTransform->letters | pTransform->digits) == 0) {
        return FALSE;
    }
    if (pTransform->pActiveScheme->lettersDir == pTransform->pActiveScheme->digitsDir) {
        doShape(pTransform, pTransform->letters | pTransform->digits | pTransform->pActiveScheme->lettersDir,
                pErrorCode);
    } else {
        doShape(pTransform, pTransform->digits | pTransform->pActiveScheme->digitsDir, pErrorCode);
        if (U_SUCCESS(*pErrorCode)) {
            updateSrc(pTransform, pTransform->dest, *pTransform->pDestLength,
                    *pTransform->pDestLength, pErrorCode);
            doShape(pTransform, pTransform->letters | pTransform->pActiveScheme->lettersDir,
                    pErrorCode);
        }
    }
    return TRUE;
}

/**
 * Performs character mirroring.
 * 
 * @param pTransform Pointer to the <code>UBiDiTransform</code> structure.
 * @param pErrorCode Pointer to the error code value.
 *
 * @return Whether or not this function modifies the text. Besides the return
 * value, the caller should also check <code>U_SUCCESS(*pErrorCode)</code>.
 */
static UBool
action_mirror(UBiDiTransform *pTransform, UErrorCode *pErrorCode)
{
    UChar32 c;
    uint32_t i = 0, j = 0;
    if (0 == (pTransform->reorderingOptions & UBIDI_DO_MIRRORING)) {
        return FALSE;
    }
    if (pTransform->destSize < pTransform->srcLength) {
        *pErrorCode = U_BUFFER_OVERFLOW_ERROR;
        return FALSE;
    }
    do {
        UBool isOdd = ubidi_getLevelAt(pTransform->pBidi, i) & 1;
        U16_NEXT(pTransform->src, i, pTransform->srcLength, c); 
        U16_APPEND_UNSAFE(pTransform->dest, j, isOdd ? u_charMirror(c) : c);
    } while (i < pTransform->srcLength);
    
    *pTransform->pDestLength = pTransform->srcLength;
    pTransform->reorderingOptions = UBIDI_REORDER_DEFAULT;
    return TRUE;
}

/**
 * All possible reordering schemes.
 *
 */
static const ReorderingScheme Schemes[] =
{
    /* 0: Logical LTR => Visual LTR */
    {LTR, LOGICAL, LTR, VISUAL, SHAPE_LOGICAL, SHAPE_LOGICAL, LTR,
            {action_shapeArabic, action_resolve, action_reorder, NULL}},
    /* 1: Logical RTL => Visual LTR */
    {RTL, LOGICAL, LTR, VISUAL, SHAPE_LOGICAL, SHAPE_VISUAL, RTL,
            {action_resolve, action_reorder, action_shapeArabic, NULL}},
    /* 2: Logical LTR => Visual RTL */
    {LTR, LOGICAL, RTL, VISUAL, SHAPE_LOGICAL, SHAPE_LOGICAL, LTR,
            {action_shapeArabic, action_resolve, action_reorder, action_reverse, NULL}},
    /* 3: Logical RTL => Visual RTL */
    {RTL, LOGICAL, RTL, VISUAL, SHAPE_LOGICAL, SHAPE_VISUAL, RTL,
            {action_resolve, action_reorder, action_shapeArabic, action_reverse, NULL}},
    /* 4: Visual LTR => Logical RTL */
    {LTR, VISUAL, RTL, LOGICAL, SHAPE_LOGICAL, SHAPE_VISUAL, RTL,
            {action_shapeArabic, action_setInverse, action_resolve, action_reorder, NULL}},
    /* 5: Visual RTL => Logical RTL */
    {RTL, VISUAL, RTL, LOGICAL, SHAPE_LOGICAL, SHAPE_VISUAL, RTL,
            {action_reverse, action_shapeArabic, action_setInverse, action_resolve, action_reorder, NULL}},
    /* 6: Visual LTR => Logical LTR */
    {LTR, VISUAL, LTR, LOGICAL, SHAPE_LOGICAL, SHAPE_LOGICAL, LTR,
            {action_setInverse, action_resolve, action_reorder, action_shapeArabic, NULL}},
    /* 7: Visual RTL => Logical LTR */
    {RTL, VISUAL, LTR, LOGICAL, SHAPE_LOGICAL, SHAPE_LOGICAL, LTR,
            {action_reverse, action_setInverse, action_resolve, action_reorder, action_shapeArabic, NULL}},
    /* 8: Logical LTR => Logical RTL */
    {LTR, LOGICAL, RTL, LOGICAL, SHAPE_LOGICAL, SHAPE_LOGICAL, LTR,
            {action_shapeArabic, action_resolve, action_mirror, action_setRunsOnly, action_resolve, action_reorder, NULL}},
    /* 9: Logical RTL => Logical LTR */
    {RTL, LOGICAL, LTR, LOGICAL, SHAPE_LOGICAL, SHAPE_LOGICAL, RTL,
            {action_resolve, action_mirror, action_setRunsOnly, action_resolve, action_reorder, action_shapeArabic, NULL}},
    /* 10: Visual LTR => Visual RTL */
    {LTR, VISUAL, RTL, VISUAL, SHAPE_LOGICAL, SHAPE_VISUAL, LTR,
            {action_shapeArabic, action_setInverse, action_resolve, action_mirror, action_reverse, NULL}},
    /* 11: Visual RTL => Visual LTR */
    {RTL, VISUAL, LTR, VISUAL, SHAPE_LOGICAL, SHAPE_VISUAL, LTR,
            {action_reverse, action_shapeArabic, action_setInverse, action_resolve, action_mirror, NULL}},
    /* 12: Logical LTR => Logical LTR */
    {LTR, LOGICAL, LTR, LOGICAL, SHAPE_LOGICAL, SHAPE_LOGICAL, LTR,
            {action_resolve, action_mirror, action_shapeArabic, NULL}},
    /* 13: Logical RTL => Logical RTL */
    {RTL, LOGICAL, RTL, LOGICAL, SHAPE_VISUAL, SHAPE_LOGICAL, RTL,
            {action_resolve, action_mirror, action_shapeArabic, NULL}},
    /* 14: Visual LTR => Visual LTR */
    {LTR, VISUAL, LTR, VISUAL, SHAPE_LOGICAL, SHAPE_VISUAL, LTR,
            {action_resolve, action_mirror, action_shapeArabic, NULL}},
    /* 15: Visual RTL => Visual RTL */
    {RTL, VISUAL, RTL, VISUAL, SHAPE_LOGICAL, SHAPE_VISUAL, LTR,
            {action_reverse, action_resolve, action_mirror, action_shapeArabic, action_reverse, NULL}}
};

static const uint32_t nSchemes = sizeof(Schemes) / sizeof(*Schemes);

/**
 * When the direction option is <code>UBIDI_DEFAULT_LTR</code> or
 * <code>UBIDI_DEFAULT_RTL</code>, resolve the base direction according to that
 * of the first strong bidi character.
 */
static void
resolveBaseDirection(const UChar *text, uint32_t length,
        UBiDiLevel *pInLevel, UBiDiLevel *pOutLevel)
{
    switch (*pInLevel) {
        case UBIDI_DEFAULT_LTR:
        case UBIDI_DEFAULT_RTL: {
            UBiDiLevel level = static_cast<UBiDiLevel>(ubidi_getBaseDirection(text, length));
            *pInLevel = static_cast<UBiDiLevel>(level != UBIDI_NEUTRAL) ? level
                    : *pInLevel == UBIDI_DEFAULT_RTL ? static_cast<UBiDiLevel>(RTL) : static_cast<UBiDiLevel>(LTR);
            break;
        }
        default:
            *pInLevel &= 1;
            break;
    }
    switch (*pOutLevel) {
        case UBIDI_DEFAULT_LTR:
        case UBIDI_DEFAULT_RTL:
            *pOutLevel = *pInLevel;
            break;
        default:
            *pOutLevel &= 1;
            break;
    }
}

/**
 * Finds a valid <code>ReorderingScheme</code> matching the
 * caller-defined scheme.
 * 
 * @return A valid <code>ReorderingScheme</code> object or NULL
 */
static const ReorderingScheme*
findMatchingScheme(UBiDiLevel inLevel, UBiDiLevel outLevel,
        UBiDiOrder inOrder, UBiDiOrder outOrder)
{
    uint32_t i;
    for (i = 0; i < nSchemes; i++) {
        const ReorderingScheme *pScheme = Schemes + i;
        if (inLevel == pScheme->inLevel && outLevel == pScheme->outLevel
                && inOrder == pScheme->inOrder && outOrder == pScheme->outOrder) {
            return pScheme;
        }
    }
    return NULL;
}

U_DRAFT uint32_t U_EXPORT2
ubiditransform_transform(UBiDiTransform *pBiDiTransform,
            const UChar *src, int32_t srcLength,
            UChar *dest, int32_t destSize,
            UBiDiLevel inParaLevel, UBiDiOrder inOrder,
            UBiDiLevel outParaLevel, UBiDiOrder outOrder,
            UBiDiMirroring doMirroring, uint32_t shapingOptions,
            UErrorCode *pErrorCode)
{
    uint32_t destLength = 0;
    UBool textChanged = FALSE;
    const UBiDiTransform *pOrigTransform = pBiDiTransform;
    const UBiDiAction *action = NULL;

    if (U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if (src == NULL || dest == NULL) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    CHECK_LEN(src, srcLength, pErrorCode);
    CHECK_LEN(dest, destSize, pErrorCode);

    if (pBiDiTransform == NULL) {
        pBiDiTransform = ubiditransform_open(pErrorCode);
        if (U_FAILURE(*pErrorCode)) {
            return 0;
        }
    }
    /* Current limitation: in multiple paragraphs will be resolved according
       to the 1st paragraph */
    resolveBaseDirection(src, srcLength, &inParaLevel, &outParaLevel);

    pBiDiTransform->pActiveScheme = findMatchingScheme(inParaLevel, outParaLevel,
            inOrder, outOrder);
    if (pBiDiTransform->pActiveScheme == NULL) {
        goto cleanup;
    }
    pBiDiTransform->reorderingOptions = doMirroring ? UBIDI_DO_MIRRORING
            : UBIDI_REORDER_DEFAULT;

    /* Ignore TEXT_DIRECTION_* flags, as we apply our own depending on the text
       scheme at the time shaping is invoked. */
    shapingOptions &= ~U_SHAPE_TEXT_DIRECTION_MASK;
    pBiDiTransform->digits = shapingOptions & ~U_SHAPE_LETTERS_MASK;
    pBiDiTransform->letters = shapingOptions & ~U_SHAPE_DIGITS_MASK;

    updateSrc(pBiDiTransform, src, srcLength, destSize > srcLength ? destSize : srcLength, pErrorCode);
    if (U_FAILURE(*pErrorCode)) {
        goto cleanup;
    }
    if (pBiDiTransform->pBidi == NULL) {
        pBiDiTransform->pBidi = ubidi_openSized(0, 0, pErrorCode);
        if (U_FAILURE(*pErrorCode)) {
            goto cleanup;
        }
    }
    pBiDiTransform->dest = dest;
    pBiDiTransform->destSize = destSize;
    pBiDiTransform->pDestLength = &destLength;

    /* Checking for U_SUCCESS() within the loop to bail out on first failure. */
    for (action = pBiDiTransform->pActiveScheme->actions; *action && U_SUCCESS(*pErrorCode); action++) {
        if ((*action)(pBiDiTransform, pErrorCode)) {
            if (action + 1) {
                updateSrc(pBiDiTransform, pBiDiTransform->dest, *pBiDiTransform->pDestLength,
                        *pBiDiTransform->pDestLength, pErrorCode);
            }
            textChanged = TRUE;
        }
    }
    ubidi_setInverse(pBiDiTransform->pBidi, FALSE);

    if (!textChanged && U_SUCCESS(*pErrorCode)) {
        /* Text was not changed - just copy src to dest */
        if (destSize < srcLength) {
            *pErrorCode = U_BUFFER_OVERFLOW_ERROR;
        } else {
            u_strncpy(dest, src, srcLength);
            destLength = srcLength;
        }
    }
cleanup:
    if (pOrigTransform != pBiDiTransform) {
        ubiditransform_close(pBiDiTransform);
    } else {
        pBiDiTransform->dest = NULL;
        pBiDiTransform->pDestLength = NULL;
        pBiDiTransform->srcLength = 0;
        pBiDiTransform->destSize = 0;
    }
    return U_FAILURE(*pErrorCode) ? 0 : destLength;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2000-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  ubidiwrt.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999aug06
*   created by: Markus W. Scherer, updated by Matitiahu Allouche
*
* This file contains implementations for BiDi functions that use
* the core algorithm and core API to write reordered text.
*/










/*
 * The function implementations in this file are designed
 * for UTF-16 and UTF-32, not for UTF-8.
 *
 * Assumptions that are not true for UTF-8:
 * - Any code point always needs the same number of code units
 *   ("minimum-length-problem" of UTF-8)
 * - The BiDi control characters need only one code unit each
 *
 * Further assumptions for all UTFs:
 * - u_charMirror(c) needs the same number of code units as c
 */
#if defined(UTF_SIZE) && UTF_SIZE==8
# error reimplement ubidi_writeReordered() for UTF-8, see comment above
#endif

#define IS_COMBINING(type) ((1UL<<(type))&(1UL<<U_NON_SPACING_MARK|1UL<<U_COMBINING_SPACING_MARK|1UL<<U_ENCLOSING_MARK))

/*
 * When we have UBIDI_OUTPUT_REVERSE set on ubidi_writeReordered(), then we
 * semantically write RTL runs in reverse and later reverse them again.
 * Instead, we actually write them in forward order to begin with.
 * However, if the RTL run was to be mirrored, we need to mirror here now
 * since the implicit second reversal must not do it.
 * It looks strange to do mirroring in LTR output, but it is only because
 * we are writing RTL output in reverse.
 */
static int32_t
doWriteForward(const UChar *src, int32_t srcLength,
               UChar *dest, int32_t destSize,
               uint16_t options,
               UErrorCode *pErrorCode) {
    /* optimize for several combinations of options */
    switch(options&(UBIDI_REMOVE_BIDI_CONTROLS|UBIDI_DO_MIRRORING)) {
    case 0: {
        /* simply copy the LTR run to the destination */
        int32_t length=srcLength;
        if(destSize<length) {
            *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
            return srcLength;
        }
        do {
            *dest++=*src++;
        } while(--length>0);
        return srcLength;
    }
    case UBIDI_DO_MIRRORING: {
        /* do mirroring */
        int32_t i=0, j=0;
        UChar32 c;

        if(destSize<srcLength) {
            *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
            return srcLength;
        }
        do {
            U16_NEXT(src, i, srcLength, c);
            c=u_charMirror(c);
            U16_APPEND_UNSAFE(dest, j, c);
        } while(i<srcLength);
        return srcLength;
    }
    case UBIDI_REMOVE_BIDI_CONTROLS: {
        /* copy the LTR run and remove any BiDi control characters */
        int32_t remaining=destSize;
        UChar c;
        do {
            c=*src++;
            if(!IS_BIDI_CONTROL_CHAR(c)) {
                if(--remaining<0) {
                    *pErrorCode=U_BUFFER_OVERFLOW_ERROR;

                    /* preflight the length */
                    while(--srcLength>0) {
                        c=*src++;
                        if(!IS_BIDI_CONTROL_CHAR(c)) {
                            --remaining;
                        }
                    }
                    return destSize-remaining;
                }
                *dest++=c;
            }
        } while(--srcLength>0);
        return destSize-remaining;
    }
    default: {
        /* remove BiDi control characters and do mirroring */
        int32_t remaining=destSize;
        int32_t i, j=0;
        UChar32 c;
        do {
            i=0;
            U16_NEXT(src, i, srcLength, c);
            src+=i;
            srcLength-=i;
            if(!IS_BIDI_CONTROL_CHAR(c)) {
                remaining-=i;
                if(remaining<0) {
                    *pErrorCode=U_BUFFER_OVERFLOW_ERROR;

                    /* preflight the length */
                    while(srcLength>0) {
                        c=*src++;
                        if(!IS_BIDI_CONTROL_CHAR(c)) {
                            --remaining;
                        }
                        --srcLength;
                    }
                    return destSize-remaining;
                }
                c=u_charMirror(c);
                U16_APPEND_UNSAFE(dest, j, c);
            }
        } while(srcLength>0);
        return j;
    }
    } /* end of switch */
}

static int32_t
doWriteReverse(const UChar *src, int32_t srcLength,
               UChar *dest, int32_t destSize,
               uint16_t options,
               UErrorCode *pErrorCode) {
    /*
     * RTL run -
     *
     * RTL runs need to be copied to the destination in reverse order
     * of code points, not code units, to keep Unicode characters intact.
     *
     * The general strategy for this is to read the source text
     * in backward order, collect all code units for a code point
     * (and optionally following combining characters, see below),
     * and copy all these code units in ascending order
     * to the destination for this run.
     *
     * Several options request whether combining characters
     * should be kept after their base characters,
     * whether BiDi control characters should be removed, and
     * whether characters should be replaced by their mirror-image
     * equivalent Unicode characters.
     */
    int32_t i, j;
    UChar32 c;

    /* optimize for several combinations of options */
    switch(options&(UBIDI_REMOVE_BIDI_CONTROLS|UBIDI_DO_MIRRORING|UBIDI_KEEP_BASE_COMBINING)) {
    case 0:
        /*
         * With none of the "complicated" options set, the destination
         * run will have the same length as the source run,
         * and there is no mirroring and no keeping combining characters
         * with their base characters.
         */
        if(destSize<srcLength) {
            *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
            return srcLength;
        }
        destSize=srcLength;

        /* preserve character integrity */
        do {
            /* i is always after the last code unit known to need to be kept in this segment */
            i=srcLength;

            /* collect code units for one base character */
            U16_BACK_1(src, 0, srcLength);

            /* copy this base character */
            j=srcLength;
            do {
                *dest++=src[j++];
            } while(j<i);
        } while(srcLength>0);
        break;
    case UBIDI_KEEP_BASE_COMBINING:
        /*
         * Here, too, the destination
         * run will have the same length as the source run,
         * and there is no mirroring.
         * We do need to keep combining characters with their base characters.
         */
        if(destSize<srcLength) {
            *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
            return srcLength;
        }
        destSize=srcLength;

        /* preserve character integrity */
        do {
            /* i is always after the last code unit known to need to be kept in this segment */
            i=srcLength;

            /* collect code units and modifier letters for one base character */
            do {
                U16_PREV(src, 0, srcLength, c);
            } while(srcLength>0 && IS_COMBINING(u_charType(c)));

            /* copy this "user character" */
            j=srcLength;
            do {
                *dest++=src[j++];
            } while(j<i);
        } while(srcLength>0);
        break;
    default:
        /*
         * With several "complicated" options set, this is the most
         * general and the slowest copying of an RTL run.
         * We will do mirroring, remove BiDi controls, and
         * keep combining characters with their base characters
         * as requested.
         */
        if(!(options&UBIDI_REMOVE_BIDI_CONTROLS)) {
            i=srcLength;
        } else {
            /* we need to find out the destination length of the run,
               which will not include the BiDi control characters */
            int32_t length=srcLength;
            UChar ch;

            i=0;
            do {
                ch=*src++;
                if(!IS_BIDI_CONTROL_CHAR(ch)) {
                    ++i;
                }
            } while(--length>0);
            src-=srcLength;
        }

        if(destSize<i) {
            *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
            return i;
        }
        destSize=i;

        /* preserve character integrity */
        do {
            /* i is always after the last code unit known to need to be kept in this segment */
            i=srcLength;

            /* collect code units for one base character */
            U16_PREV(src, 0, srcLength, c);
            if(options&UBIDI_KEEP_BASE_COMBINING) {
                /* collect modifier letters for this base character */
                while(srcLength>0 && IS_COMBINING(u_charType(c))) {
                    U16_PREV(src, 0, srcLength, c);
                }
            }

            if(options&UBIDI_REMOVE_BIDI_CONTROLS && IS_BIDI_CONTROL_CHAR(c)) {
                /* do not copy this BiDi control character */
                continue;
            }

            /* copy this "user character" */
            j=srcLength;
            if(options&UBIDI_DO_MIRRORING) {
                /* mirror only the base character */
                int32_t k=0;
                c=u_charMirror(c);
                U16_APPEND_UNSAFE(dest, k, c);
                dest+=k;
                j+=k;
            }
            while(j<i) {
                *dest++=src[j++];
            }
        } while(srcLength>0);
        break;
    } /* end of switch */

    return destSize;
}

U_CAPI int32_t U_EXPORT2
ubidi_writeReverse(const UChar *src, int32_t srcLength,
                   UChar *dest, int32_t destSize,
                   uint16_t options,
                   UErrorCode *pErrorCode) {
    int32_t destLength;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* more error checking */
    if( src==NULL || srcLength<-1 ||
        destSize<0 || (destSize>0 && dest==NULL))
    {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* do input and output overlap? */
    if( dest!=NULL &&
        ((src>=dest && src<dest+destSize) ||
         (dest>=src && dest<src+srcLength)))
    {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if(srcLength==-1) {
        srcLength=u_strlen(src);
    }
    if(srcLength>0) {
        destLength=doWriteReverse(src, srcLength, dest, destSize, options, pErrorCode);
    } else {
        /* nothing to do */
        destLength=0;
    }

    return u_terminateUChars(dest, destSize, destLength, pErrorCode);
}

// Ticket 20907 - The optimizer in MSVC/Visual Studio versions below 16.4 has trouble with this
// function on Windows ARM64. As a work-around, we disable optimizations for this function.
// This work-around could/should be removed once the following versions of Visual Studio are no
// longer supported: All versions of VS2017, and versions of VS2019 below 16.4.
#if (defined(_MSC_VER) && (defined(_M_ARM64)) && (_MSC_VER < 1924))
#pragma optimize( "", off )
#endif
U_CAPI int32_t U_EXPORT2
ubidi_writeReordered(UBiDi *pBiDi,
                     UChar *dest, int32_t destSize,
                     uint16_t options,
                     UErrorCode *pErrorCode) {
    const UChar *text;
    UChar *saveDest;
    int32_t length, destCapacity;
    int32_t run, runCount, logicalStart, runLength;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* more error checking */
    if( pBiDi==NULL ||
        (text=pBiDi->text)==NULL || (length=pBiDi->length)<0 ||
        destSize<0 || (destSize>0 && dest==NULL))
    {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* do input and output overlap? */
    if( dest!=NULL &&
        ((text>=dest && text<dest+destSize) ||
         (dest>=text && dest<text+pBiDi->originalLength)))
    {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if(length==0) {
        /* nothing to do */
        return u_terminateUChars(dest, destSize, 0, pErrorCode);
    }

    runCount=ubidi_countRuns(pBiDi, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* destSize shrinks, later destination length=destCapacity-destSize */
    saveDest=dest;
    destCapacity=destSize;

    /*
     * Option "insert marks" implies UBIDI_INSERT_LRM_FOR_NUMERIC if the
     * reordering mode (checked below) is appropriate.
     */
    if(pBiDi->reorderingOptions & UBIDI_OPTION_INSERT_MARKS) {
        options|=UBIDI_INSERT_LRM_FOR_NUMERIC;
        options&=~UBIDI_REMOVE_BIDI_CONTROLS;
    }
    /*
     * Option "remove controls" implies UBIDI_REMOVE_BIDI_CONTROLS
     * and cancels UBIDI_INSERT_LRM_FOR_NUMERIC.
     */
    if(pBiDi->reorderingOptions & UBIDI_OPTION_REMOVE_CONTROLS) {
        options|=UBIDI_REMOVE_BIDI_CONTROLS;
        options&=~UBIDI_INSERT_LRM_FOR_NUMERIC;
    }
    /*
     * If we do not perform the "inverse BiDi" algorithm, then we
     * don't need to insert any LRMs, and don't need to test for it.
     */
    if((pBiDi->reorderingMode != UBIDI_REORDER_INVERSE_NUMBERS_AS_L) &&
       (pBiDi->reorderingMode != UBIDI_REORDER_INVERSE_LIKE_DIRECT)  &&
       (pBiDi->reorderingMode != UBIDI_REORDER_INVERSE_FOR_NUMBERS_SPECIAL) &&
       (pBiDi->reorderingMode != UBIDI_REORDER_RUNS_ONLY)) {
        options&=~UBIDI_INSERT_LRM_FOR_NUMERIC;
    }
    /*
     * Iterate through all visual runs and copy the run text segments to
     * the destination, according to the options.
     *
     * The tests for where to insert LRMs ignore the fact that there may be
     * BN codes or non-BMP code points at the beginning and end of a run;
     * they may insert LRMs unnecessarily but the tests are faster this way
     * (this would have to be improved for UTF-8).
     *
     * Note that the only errors that are set by doWriteXY() are buffer overflow
     * errors. Ignore them until the end, and continue for preflighting.
     */
    if(!(options&UBIDI_OUTPUT_REVERSE)) {
        /* forward output */
        if(!(options&UBIDI_INSERT_LRM_FOR_NUMERIC)) {
            /* do not insert BiDi controls */
            for(run=0; run<runCount; ++run) {
                if(UBIDI_LTR==ubidi_getVisualRun(pBiDi, run, &logicalStart, &runLength)) {
                    runLength=doWriteForward(text+logicalStart, runLength,
                                             dest, destSize,
                                             (uint16_t)(options&~UBIDI_DO_MIRRORING), pErrorCode);
                } else {
                    runLength=doWriteReverse(text+logicalStart, runLength,
                                             dest, destSize,
                                             options, pErrorCode);
                }
                if(dest!=NULL) {
                  dest+=runLength;
                }
                destSize-=runLength;
            }
        } else {
            /* insert BiDi controls for "inverse BiDi" */
            const DirProp *dirProps=pBiDi->dirProps;
            const UChar *src;
            UChar uc;
            UBiDiDirection dir;
            int32_t markFlag;

            for(run=0; run<runCount; ++run) {
                dir=ubidi_getVisualRun(pBiDi, run, &logicalStart, &runLength);
                src=text+logicalStart;
                /* check if something relevant in insertPoints */
                markFlag=pBiDi->runs[run].insertRemove;
                if(markFlag<0) {        /* BiDi controls count */
                    markFlag=0;
                }

                if(UBIDI_LTR==dir) {
                    if((pBiDi->isInverse) &&
                       (/*run>0 &&*/ dirProps[logicalStart]!=L)) {
                        markFlag |= LRM_BEFORE;
                    }
                    if (markFlag & LRM_BEFORE) {
                        uc=LRM_CHAR;
                    }
                    else if (markFlag & RLM_BEFORE) {
                        uc=RLM_CHAR;
                    }
                    else  uc=0;
                    if(uc) {
                        if(destSize>0) {
                            *dest++=uc;
                        }
                        --destSize;
                    }

                    runLength=doWriteForward(src, runLength,
                                             dest, destSize,
                                             (uint16_t)(options&~UBIDI_DO_MIRRORING), pErrorCode);
                    if(dest!=NULL) {
                      dest+=runLength;
                    }
                    destSize-=runLength;

                    if((pBiDi->isInverse) &&
                       (/*run<runCount-1 &&*/ dirProps[logicalStart+runLength-1]!=L)) {
                        markFlag |= LRM_AFTER;
                    }
                    if (markFlag & LRM_AFTER) {
                        uc=LRM_CHAR;
                    }
                    else if (markFlag & RLM_AFTER) {
                        uc=RLM_CHAR;
                    }
                    else  uc=0;
                    if(uc) {
                        if(destSize>0) {
                            *dest++=uc;
                        }
                        --destSize;
                    }
                } else {                /* RTL run */
                    if((pBiDi->isInverse) &&
                       (/*run>0 &&*/ !(MASK_R_AL&DIRPROP_FLAG(dirProps[logicalStart+runLength-1])))) {
                        markFlag |= RLM_BEFORE;
                    }
                    if (markFlag & LRM_BEFORE) {
                        uc=LRM_CHAR;
                    }
                    else if (markFlag & RLM_BEFORE) {
                        uc=RLM_CHAR;
                    }
                    else  uc=0;
                    if(uc) {
                        if(destSize>0) {
                            *dest++=uc;
                        }
                        --destSize;
                    }

                    runLength=doWriteReverse(src, runLength,
                                             dest, destSize,
                                             options, pErrorCode);
                    if(dest!=NULL) {
                      dest+=runLength;
                    }
                    destSize-=runLength;

                    if((pBiDi->isInverse) &&
                       (/*run<runCount-1 &&*/ !(MASK_R_AL&DIRPROP_FLAG(dirProps[logicalStart])))) {
                        markFlag |= RLM_AFTER;
                    }
                    if (markFlag & LRM_AFTER) {
                        uc=LRM_CHAR;
                    }
                    else if (markFlag & RLM_AFTER) {
                        uc=RLM_CHAR;
                    }
                    else  uc=0;
                    if(uc) {
                        if(destSize>0) {
                            *dest++=uc;
                        }
                        --destSize;
                    }
                }
            }
        }
    } else {
        /* reverse output */
        if(!(options&UBIDI_INSERT_LRM_FOR_NUMERIC)) {
            /* do not insert BiDi controls */
            for(run=runCount; --run>=0;) {
                if(UBIDI_LTR==ubidi_getVisualRun(pBiDi, run, &logicalStart, &runLength)) {
                    runLength=doWriteReverse(text+logicalStart, runLength,
                                             dest, destSize,
                                             (uint16_t)(options&~UBIDI_DO_MIRRORING), pErrorCode);
                } else {
                    runLength=doWriteForward(text+logicalStart, runLength,
                                             dest, destSize,
                                             options, pErrorCode);
                }
                if(dest!=NULL) {
                  dest+=runLength;
                }
                destSize-=runLength;
            }
        } else {
            /* insert BiDi controls for "inverse BiDi" */
            const DirProp *dirProps=pBiDi->dirProps;
            const UChar *src;
            UBiDiDirection dir;

            for(run=runCount; --run>=0;) {
                /* reverse output */
                dir=ubidi_getVisualRun(pBiDi, run, &logicalStart, &runLength);
                src=text+logicalStart;

                if(UBIDI_LTR==dir) {
                    if(/*run<runCount-1 &&*/ dirProps[logicalStart+runLength-1]!=L) {
                        if(destSize>0) {
                            *dest++=LRM_CHAR;
                        }
                        --destSize;
                    }

                    runLength=doWriteReverse(src, runLength,
                                             dest, destSize,
                                             (uint16_t)(options&~UBIDI_DO_MIRRORING), pErrorCode);
                    if(dest!=NULL) {
                      dest+=runLength;
                    }
                    destSize-=runLength;

                    if(/*run>0 &&*/ dirProps[logicalStart]!=L) {
                        if(destSize>0) {
                            *dest++=LRM_CHAR;
                        }
                        --destSize;
                    }
                } else {
                    if(/*run<runCount-1 &&*/ !(MASK_R_AL&DIRPROP_FLAG(dirProps[logicalStart]))) {
                        if(destSize>0) {
                            *dest++=RLM_CHAR;
                        }
                        --destSize;
                    }

                    runLength=doWriteForward(src, runLength,
                                             dest, destSize,
                                             options, pErrorCode);
                    if(dest!=NULL) {
                      dest+=runLength;
                    }
                    destSize-=runLength;

                    if(/*run>0 &&*/ !(MASK_R_AL&DIRPROP_FLAG(dirProps[logicalStart+runLength-1]))) {
                        if(destSize>0) {
                            *dest++=RLM_CHAR;
                        }
                        --destSize;
                    }
                }
            }
        }
    }

    return u_terminateUChars(saveDest, destCapacity, destCapacity-destSize, pErrorCode);
}
#if (defined(_MSC_VER) && (defined(_M_ARM64)) && (_MSC_VER < 1924))
#pragma optimize( "", on )
#endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// ********************************************************************************
// *   Copyright (C) 1996-2015, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// ********************************************************************************
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/ubrk.h"

// #include "unicode/brkiter.h"
// #include "unicode/uloc.h"
// #include "unicode/ustring.h"
// #include "unicode/uchriter.h"
// #include "unicode/rbbi.h"
// #include "rbbirb.h"
// #include "uassert.h"
// #include "cmemory.h"

// U_NAMESPACE_USE

// //------------------------------------------------------------------------------
// //
// //    ubrk_open      Create a canned type of break iterator based on type (word, line, etc.)
// //                   and locale.
// //
// //------------------------------------------------------------------------------
// U_CAPI UBreakIterator* U_EXPORT2
// ubrk_open(UBreakIteratorType type,
//       const char *locale,
//       const UChar *text,
//       int32_t textLength,
//       UErrorCode *status)
// {

//   if(U_FAILURE(*status)) return 0;

//   BreakIterator *result = 0;

//   switch(type) {

//   case UBRK_CHARACTER:
//     result = BreakIterator::createCharacterInstance(Locale(locale), *status);
//     break;

//   case UBRK_WORD:
//     result = BreakIterator::createWordInstance(Locale(locale), *status);
//     break;

//   case UBRK_LINE:
//     result = BreakIterator::createLineInstance(Locale(locale), *status);
//     break;

//   case UBRK_SENTENCE:
//     result = BreakIterator::createSentenceInstance(Locale(locale), *status);
//     break;

//   case UBRK_TITLE:
//     result = BreakIterator::createTitleInstance(Locale(locale), *status);
//     break;

//   default:
//     *status = U_ILLEGAL_ARGUMENT_ERROR;
//   }

//   // check for allocation error
//   if (U_FAILURE(*status)) {
//      return 0;
//   }
//   if(result == 0) {
//     *status = U_MEMORY_ALLOCATION_ERROR;
//     return 0;
//   }


//   UBreakIterator *uBI = (UBreakIterator *)result;
//   if (text != NULL) {
//       ubrk_setText(uBI, text, textLength, status);
//   }
//   return uBI;
// }



// //------------------------------------------------------------------------------
// //
// //   ubrk_openRules      open a break iterator from a set of break rules.
// //                       Invokes the rule builder.
// //
// //------------------------------------------------------------------------------
// U_CAPI UBreakIterator* U_EXPORT2
// ubrk_openRules(  const UChar        *rules,
//                        int32_t       rulesLength,
//                  const UChar        *text,
//                        int32_t       textLength,
//                        UParseError  *parseErr,
//                        UErrorCode   *status)  {

//     if (status == NULL || U_FAILURE(*status)){
//         return 0;
//     }

//     BreakIterator *result = 0;
//     UnicodeString ruleString(rules, rulesLength);
//     result = RBBIRuleBuilder::createRuleBasedBreakIterator(ruleString, parseErr, *status);
//     if(U_FAILURE(*status)) {
//         return 0;
//     }

//     UBreakIterator *uBI = (UBreakIterator *)result;
//     if (text != NULL) {
//         ubrk_setText(uBI, text, textLength, status);
//     }
//     return uBI;
// }


// U_CAPI UBreakIterator* U_EXPORT2
// ubrk_openBinaryRules(const uint8_t *binaryRules, int32_t rulesLength,
//                      const UChar *  text, int32_t textLength,
//                      UErrorCode *   status)
// {
//     if (U_FAILURE(*status)) {
//         return NULL;
//     }
//     if (rulesLength < 0) {
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return NULL;
//     }
//     LocalPointer<RuleBasedBreakIterator> lpRBBI(new RuleBasedBreakIterator(binaryRules, rulesLength, *status), *status);
//     if (U_FAILURE(*status)) {
//         return NULL;
//     }
//     UBreakIterator *uBI = reinterpret_cast<UBreakIterator *>(lpRBBI.orphan());
//     if (text != NULL) {
//         ubrk_setText(uBI, text, textLength, status);
//     }
//     return uBI;
// }


// U_CAPI UBreakIterator * U_EXPORT2
// ubrk_safeClone(
//           const UBreakIterator *bi,
//           void * /*stackBuffer*/,
//           int32_t *pBufferSize,
//           UErrorCode *status)
// {
//     if (status == NULL || U_FAILURE(*status)){
//         return NULL;
//     }
//     if (bi == NULL) {
//        *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return NULL;
//     }
//     if (pBufferSize != NULL) {
//         int32_t inputSize = *pBufferSize;
//         *pBufferSize = 1;
//         if (inputSize == 0) {
//             return NULL;  // preflighting for deprecated functionality
//         }
//     }
//     BreakIterator *newBI = ((BreakIterator *)bi)->clone();
//     if (newBI == NULL) {
//         *status = U_MEMORY_ALLOCATION_ERROR;
//     } else {
//         *status = U_SAFECLONE_ALLOCATED_WARNING;
//     }
//     return (UBreakIterator *)newBI;
// }



// U_CAPI void U_EXPORT2
// ubrk_close(UBreakIterator *bi)
// {
//     delete (BreakIterator *)bi;
// }

// U_CAPI void U_EXPORT2
// ubrk_setText(UBreakIterator* bi,
//              const UChar*    text,
//              int32_t         textLength,
//              UErrorCode*     status)
// {
//     UText  ut = UTEXT_INITIALIZER;
//     utext_openUChars(&ut, text, textLength, status);
//     ((BreakIterator*)bi)->setText(&ut, *status);
//     // A stack allocated UText wrapping a UChar * string
//     //   can be dumped without explicitly closing it.
// }



// U_CAPI void U_EXPORT2
// ubrk_setUText(UBreakIterator *bi,
//              UText          *text,
//              UErrorCode     *status)
// {
//   ((BreakIterator*)bi)->setText(text, *status);
// }





// U_CAPI int32_t U_EXPORT2
// ubrk_current(const UBreakIterator *bi)
// {

//   return ((BreakIterator*)bi)->current();
// }

// U_CAPI int32_t U_EXPORT2
// ubrk_next(UBreakIterator *bi)
// {

//   return ((BreakIterator*)bi)->next();
// }

// U_CAPI int32_t U_EXPORT2
// ubrk_previous(UBreakIterator *bi)
// {

//   return ((BreakIterator*)bi)->previous();
// }

// U_CAPI int32_t U_EXPORT2
// ubrk_first(UBreakIterator *bi)
// {

//   return ((BreakIterator*)bi)->first();
// }

// U_CAPI int32_t U_EXPORT2
// ubrk_last(UBreakIterator *bi)
// {

//   return ((BreakIterator*)bi)->last();
// }

// U_CAPI int32_t U_EXPORT2
// ubrk_preceding(UBreakIterator *bi,
//            int32_t offset)
// {

//   return ((BreakIterator*)bi)->preceding(offset);
// }

// U_CAPI int32_t U_EXPORT2
// ubrk_following(UBreakIterator *bi,
//            int32_t offset)
// {

//   return ((BreakIterator*)bi)->following(offset);
// }

// U_CAPI const char* U_EXPORT2
// ubrk_getAvailable(int32_t index)
// {

//   return uloc_getAvailable(index);
// }

// U_CAPI int32_t U_EXPORT2
// ubrk_countAvailable()
// {

//   return uloc_countAvailable();
// }


// U_CAPI  UBool U_EXPORT2
// ubrk_isBoundary(UBreakIterator *bi, int32_t offset)
// {
//     return ((BreakIterator*)bi)->isBoundary(offset);
// }


// U_CAPI  int32_t U_EXPORT2
// ubrk_getRuleStatus(UBreakIterator *bi)
// {
//     return ((BreakIterator*)bi)->getRuleStatus();
// }

// U_CAPI  int32_t U_EXPORT2
// ubrk_getRuleStatusVec(UBreakIterator *bi, int32_t *fillInVec, int32_t capacity, UErrorCode *status)
// {
//     return ((BreakIterator*)bi)->getRuleStatusVec(fillInVec, capacity, *status);
// }


// U_CAPI const char* U_EXPORT2
// ubrk_getLocaleByType(const UBreakIterator *bi,
//                      ULocDataLocaleType type,
//                      UErrorCode* status)
// {
//     if (bi == NULL) {
//         if (U_SUCCESS(*status)) {
//             *status = U_ILLEGAL_ARGUMENT_ERROR;
//         }
//         return NULL;
//     }
//     return ((BreakIterator*)bi)->getLocaleID(type, *status);
// }


// U_CAPI void U_EXPORT2
// ubrk_refreshUText(UBreakIterator *bi,
//                        UText          *text,
//                        UErrorCode     *status)
// {
//     BreakIterator *bii = reinterpret_cast<BreakIterator *>(bi);
//     bii->refreshInputText(text, *status);
// }

// U_CAPI int32_t U_EXPORT2
// ubrk_getBinaryRules(UBreakIterator *bi,
//                     uint8_t *       binaryRules, int32_t rulesCapacity,
//                     UErrorCode *    status)
// {
//     if (U_FAILURE(*status)) {
//         return 0;
//     }
//     if ((binaryRules == NULL && rulesCapacity > 0) || rulesCapacity < 0) {
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }
//     RuleBasedBreakIterator* rbbi;
//     if ((rbbi = dynamic_cast<RuleBasedBreakIterator*>(reinterpret_cast<BreakIterator*>(bi))) == NULL) {
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }
//     uint32_t rulesLength;
//     const uint8_t * returnedRules = rbbi->getBinaryRules(rulesLength);
//     if (rulesLength > INT32_MAX) {
//         *status = U_INDEX_OUTOFBOUNDS_ERROR;
//         return 0;
//     }
//     if (binaryRules != NULL) { // if not preflighting
//         // Here we know rulesLength <= INT32_MAX and rulesCapacity >= 0, can cast safely
//         if ((int32_t)rulesLength > rulesCapacity) {
//             *status = U_BUFFER_OVERFLOW_ERROR;
//         } else {
//             uprv_memcpy(binaryRules, returnedRules, rulesLength);
//         }
//     }
//     return (int32_t)rulesLength;
// }


// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2004-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ucase.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004aug30
*   created by: Markus W. Scherer
*
*   Low-level Unicode character/string case mapping code.
*   Much code moved here (and modified) from uchar.c.
*/




 /* UDataInfo */

 /* DataHeader */







struct UCaseProps {
    UDataMemory *mem;
    const int32_t *indexes;
    const uint16_t *exceptions;
    const uint16_t *unfold;

    UTrie2 trie;
    uint8_t formatVersion[4];
};

/* ucase_props_data.h is machine-generated by gencase --csource */
#define INCLUDED_FROM_UCASE_CPP
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// Copyright (C) 1999-2016, International Business Machines
// Corporation and others.  All Rights Reserved.
//
// file name: ucase_props_data.h
//
// machine-generated by: icu/tools/unicode/c/genprops/casepropsbuilder.cpp


#ifdef INCLUDED_FROM_UCASE_CPP

// static const UVersionInfo ucase_props_dataVersion={0xd,0,0,0};

static const int32_t ucase_props_indexes[UCASE_IX_TOP]={0x10,0x70ca,0x6098,0x687,0x172,0,0,0,0,0,0,0,0,0,0,3};

static const uint16_t ucase_props_trieIndex[12356]={
0x336,0x33e,0x346,0x34e,0x35c,0x364,0x36c,0x374,0x37c,0x384,0x38b,0x393,0x39b,0x3a3,0x3ab,0x3b3,
0x3b9,0x3c1,0x3c9,0x3d1,0x3d9,0x3e1,0x3e9,0x3f1,0x3f9,0x401,0x409,0x411,0x419,0x421,0x429,0x431,
0x439,0x441,0x449,0x451,0x459,0x461,0x469,0x471,0x46d,0x475,0x47a,0x482,0x489,0x491,0x499,0x4a1,
0x4a9,0x4b1,0x4b9,0x4c1,0x355,0x35d,0x4c6,0x4ce,0x4d3,0x4db,0x4e3,0x4eb,0x4ea,0x4f2,0x4f7,0x4ff,
0x507,0x50e,0x512,0x355,0x355,0x355,0x519,0x521,0x529,0x52b,0x533,0x53b,0x53f,0x540,0x548,0x550,
0x558,0x540,0x560,0x565,0x558,0x540,0x56d,0x575,0x53f,0x57d,0x585,0x58d,0x595,0x355,0x59d,0x355,
0x5a5,0x4ec,0x5ad,0x58d,0x53f,0x57d,0x5b4,0x58d,0x5bc,0x5be,0x548,0x58d,0x53f,0x355,0x5c6,0x355,
0x355,0x5cc,0x5d3,0x355,0x355,0x5d7,0x5df,0x355,0x5e3,0x5ea,0x355,0x5f1,0x5f9,0x600,0x608,0x355,
0x355,0x60d,0x615,0x61d,0x625,0x62d,0x634,0x63c,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x644,0x355,0x355,0x654,0x654,0x64c,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x65c,0x65c,0x54c,0x54c,0x355,0x662,0x66a,0x355,
0x672,0x355,0x67a,0x355,0x681,0x687,0x355,0x355,0x355,0x68f,0x355,0x355,0x355,0x355,0x355,0x355,
0x696,0x355,0x69d,0x6a5,0x355,0x6ad,0x6b5,0x355,0x57c,0x6b8,0x6c0,0x6c6,0x5bc,0x6ce,0x355,0x6d5,
0x355,0x6da,0x355,0x6e0,0x6e8,0x6ec,0x6f4,0x6fc,0x704,0x709,0x70c,0x714,0x724,0x71c,0x734,0x72c,
0x37c,0x73c,0x37c,0x744,0x747,0x37c,0x74f,0x37c,0x757,0x75f,0x767,0x76f,0x777,0x77f,0x787,0x78f,
0x797,0x79e,0x355,0x7a6,0x7ae,0x355,0x7b6,0x7be,0x7c6,0x7ce,0x7d6,0x7de,0x7e6,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x7e9,0x7ef,0x7f5,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x7fd,0x802,0x806,0x80e,0x37c,0x37c,0x37c,0x816,0x81e,0x825,0x355,0x82a,0x355,0x355,0x355,0x832,
0x355,0x677,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x53e,0x83a,0x355,0x355,0x841,0x355,0x355,0x849,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x851,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x6e0,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x857,0x355,0x85f,0x864,0x86c,0x355,0x355,0x874,0x87c,0x884,0x37c,0x889,0x891,0x897,0x89f,0x8a2,
0x8aa,0x8b1,0x355,0x355,0x355,0x355,0x8b8,0x8c0,0x355,0x8c8,0x8cf,0x355,0x529,0x8d4,0x8dc,0x681,
0x355,0x8e2,0x8ea,0x8ee,0x355,0x8f6,0x8fe,0x906,0x355,0x90c,0x910,0x918,0x928,0x920,0x355,0x930,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x938,0x355,0x355,0x355,0x355,0x940,0x5bc,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x945,0x94d,0x951,0x355,0x355,0x355,0x355,0x338,0x33e,0x959,0x961,0x968,0x4ec,0x355,0x355,0x970,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0xd58,0xd58,0xd70,0xdb0,0xdf0,0xe2c,0xe6c,0xeac,0xee4,0xf24,0xf64,0xfa4,0xfe4,0x1024,0x1064,0x10a4,
0x10e4,0x1124,0x1164,0x11a4,0x11b4,0x11e8,0x1224,0x1264,0x12a4,0x12e4,0xd54,0x1318,0x134c,0x138c,0x13a8,0x13dc,
0x9e1,0xa11,0xa51,0xa90,0x188,0x188,0xac8,0x188,0x188,0x188,0x188,0x188,0x188,0xaf1,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0xb31,0x188,0x188,0xb66,0xba5,0xbe5,0xc1f,0xc56,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,0x188,
0xc96,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x977,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x6b5,0x355,0x355,0x355,0x97f,0x355,0x355,0x355,
0x355,0x987,0x98d,0x991,0x355,0x355,0x995,0x999,0x99f,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x9a7,0x9ab,0x355,0x355,0x355,0x355,0x355,0x9b3,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x9bb,0x9bf,0x9c7,0x9cb,0x355,0x9d2,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x9d8,0x355,0x355,0x355,0x355,0x9df,0x355,0x355,0x355,0x355,
0x355,0x53f,0x9e4,0x9eb,0x5bd,0x5bc,0x9ef,0x53c,0x355,0x9f7,0x9fe,0x355,0xa04,0x5bc,0xa09,0xa11,
0x355,0x355,0xa16,0x355,0x355,0x355,0x355,0x338,0xa1e,0x5bc,0x5be,0xa26,0xa2d,0x355,0x355,0x355,
0x355,0x355,0x9e4,0xa35,0x355,0x355,0xa3d,0xa45,0x355,0x355,0x355,0x355,0x355,0x355,0xa49,0xa51,
0x355,0x355,0xa59,0x4b0,0x355,0x355,0xa61,0x355,0x355,0xa67,0xa6f,0x355,0x355,0x355,0x355,0x355,
0x355,0xa74,0x355,0x355,0x355,0xa7c,0xa84,0x355,0x355,0xa8c,0xa94,0x355,0x355,0x355,0xa97,0x6b5,
0xa9f,0xaa3,0xaab,0x355,0xab2,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0xab9,0x355,0x355,0x940,0xac1,0x355,0x355,0x355,0xac7,0xacf,0x355,0xad3,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0xad9,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0xadf,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0xae6,0x355,0xaec,0x57c,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0xa7c,0xa84,0x355,0x355,0x355,0x355,0x355,0x355,0x677,0x355,0xaf2,0x355,0x355,
0xafa,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0xaff,0x57c,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0xb07,0xb0f,0xb15,0x355,0x355,0x355,0x355,0xb1d,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0xb25,0xb2d,0xb32,0xb38,0xb40,0xb48,0xb50,0xb29,0xb58,0xb60,
0xb68,0xb6f,0xb2a,0xb25,0xb2d,0xb28,0xb38,0xb2b,0xb26,0xb77,0xb29,0xb7f,0xb87,0xb8f,0xb96,0xb82,
0xb8a,0xb92,0xb99,0xb85,0xba1,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x87c,0xba9,0x87c,0xbb0,0xbb7,0xbbf,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0xbc7,0xbcf,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0xbd3,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x9d0,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0xbdb,0x355,0xbe3,0xbeb,0xbf2,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0xb21,
0xbfa,0xbfa,0xc00,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x9f9,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x53f,0x87c,0x87c,0x87c,0x355,0x355,0x355,0x355,0x87c,0x87c,
0x87c,0x87c,0x87c,0x87c,0x87c,0xc08,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,0x355,
0x355,0x355,0x355,0x355,0x355,0x355,0x335,0x335,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,
0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,
0,0,4,0,0,0,0,0,0,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,
0x1012,0xa,0x5a,0x7a,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0xba,0x1012,0x1012,0x1012,0x1012,
0x1012,0x1012,0x1012,0,0,0,4,0,4,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,
0xf011,0xf9,0xf031,0x149,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0x189,0xf011,0xf011,0xf011,0xf011,
0xf011,0xf011,0xf011,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,4,0,1,0,0,4,0,4,
0,0,0,0,4,0x1c9,0,4,4,0,1,0,0,0,0,0,
0x1012,0x1012,0x1012,0x1012,0x1012,0x1fa,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x5a,0x5a,0x1012,0x1012,
0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x239,
0xf011,0xf011,0xf011,0xf011,0xf011,0x2d9,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,
0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0x3c91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x31a,0xff91,0x92,0xff91,0x92,0xff91,0x31a,0xffb1,
0x33a,0x389,0x92,0xff91,0x92,0xff91,0x92,0xff91,1,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,
0xff91,0x92,0xff91,0x92,0xff91,0x3d9,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0xc392,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x459,0x6191,0x6912,0x92,0xff91,
0x92,0xff91,0x6712,0x92,0xff91,0x6692,0x6692,0x92,0xff91,1,0x2792,0x6512,0x6592,0x92,0xff91,0x6692,
0x6792,0x3091,0x6992,0x6892,0x92,0xff91,0x5191,1,0x6992,0x6a92,0x4111,0x6b12,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x6d12,0x92,0xff91,0x6d12,1,1,0x92,0xff91,0x6d12,0x92,0xff91,0x6c92,0x6c92,0x92,
0xff91,0x92,0xff91,0x6d92,0x92,0xff91,1,0,0x92,0xff91,1,0x1c11,0,0,0,0,
0x48a,0x4bb,0x4f9,0x52a,0x55b,0x599,0x5ca,0x5fb,0x639,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,
0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0xd891,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x669,0x6ea,0x71b,0x759,
0x92,0xff91,0xcf92,0xe412,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0xbf12,1,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,1,1,1,1,1,1,0x78a,0x92,
0xff91,0xae92,0x7aa,0x7c9,0x7c9,0x92,0xff91,0x9e92,0x2292,0x2392,0x92,0xff91,0x92,0xffb1,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x7e9,0x809,0x829,0x9711,0x9911,1,0x9991,0x9991,1,0x9b11,1,0x9a91,
0x849,1,1,1,0x9991,0x869,1,0x9891,1,0x889,0x8a9,1,0x97b1,0x9691,0x8a9,0x8c9,
0x8e9,1,1,0x9691,1,0x909,0x9591,1,1,0x9511,1,1,1,1,1,1,
1,0x929,1,1,0x9311,1,0x949,0x9311,1,1,1,0x969,0x9311,0xdd91,0x9391,0x9391,
0xdc91,1,1,1,1,1,0x9291,1,0,1,1,1,1,1,1,1,
1,0x989,0x9a9,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,5,5,0x25,5,5,5,5,5,5,4,4,4,
0x14,4,0x14,4,5,5,4,4,4,4,4,4,4,4,4,4,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,4,4,5,5,5,5,5,4,4,4,4,4,4,4,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,4,4,0x54,0x54,0x44,0x44,0x44,0x44,0x44,0x9cc,0x54,0x44,0x54,0x44,
0x54,0x44,0x44,0x44,0x44,0x44,0x44,0x54,0x44,0x64,0x64,0x64,0x64,0x64,0x64,0x64,
0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,
0x64,0x64,0x64,0x64,0x64,0x74,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,
0x64,0x44,0x44,0x44,0x44,0x44,0x54,0x44,0x44,0x9dd,0x44,0x64,0x64,0x64,0x44,0x44,
0x44,0x64,0x64,4,0x44,0x44,0x44,0x64,0x64,0x64,0x64,0x44,0x64,0x64,0x64,0x44,
0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,
0x44,0x44,0x44,0x44,0x92,0xff91,0x92,0xff91,4,4,0x92,0xff91,0,0,5,0x4111,
0x4111,0x4111,0,0x3a12,0,0,0,0,4,4,0x1312,4,0x1292,0x1292,0x1292,0,
0x2012,0,0x1f92,0x1f92,0xa29,0x1012,0xafa,0x1012,0x1012,0xb3a,0x1012,0x1012,0xb7a,0xbca,0xc1a,0x1012,
0xc5a,0x1012,0x1012,0x1012,0xc9a,0xcda,0,0xd1a,0x1012,0x1012,0xd5a,0x1012,0x1012,0xd9a,0x1012,0x1012,
0xed11,0xed91,0xed91,0xed91,0xdd9,0xf011,0xea9,0xf011,0xf011,0xee9,0xf011,0xf011,0xf29,0xf79,0xfc9,0xf011,
0x1009,0xf011,0xf011,0xf011,0x1049,0x1089,0x10c9,0x10f9,0xf011,0xf011,0x1139,0xf011,0xf011,0x1179,0xf011,0xf011,
0xe011,0xe091,0xe091,0x412,0x11b9,0x11e9,2,2,2,0x1239,0x1269,0xfc11,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x1299,0x12c9,0x391,0xc631,0x12fa,0x1349,0,0x92,0xff91,0xfc92,0x92,0xff91,
1,0xbf12,0xbf12,0xbf12,0x2812,0x2812,0x2812,0x2812,0x2812,0x2812,0x2812,0x2812,0x2812,0x2812,0x2812,0x2812,
0x2812,0x2812,0x2812,0x2812,0x1012,0x1012,0x137a,0x1012,0x13ba,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,
0x1012,0x1012,0x13fa,0x1012,0x1012,0x143a,0x147a,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x14ca,0x1012,
0x1012,0x1012,0x1012,0x1012,0xf011,0xf011,0x1509,0xf011,0x1549,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,
0xf011,0xf011,0x1589,0xf011,0xf011,0x15c9,0x1609,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0x1659,0xf011,
0xf011,0xf011,0xf011,0xf011,0xd811,0xd811,0xd811,0xd811,0xd811,0xd811,0xd831,0xd811,0xd831,0xd811,0xd811,0xd811,
0xd811,0xd811,0xd811,0xd811,0x92,0xff91,0x169a,0x16d9,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0,0x44,0x44,0x44,0x44,0x44,4,4,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x792,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,
0xff91,0x92,0xff91,0xf891,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,
0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0,
0,4,0,0,0,0,0,4,1,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,
0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,
0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0x1719,1,0,0,0,
0,0,0,0,0,0x64,0x44,0x44,0x44,0x44,0x64,0x44,0x44,0x44,0x64,0x64,
0x44,0x44,0x44,0x44,0x44,0x44,0x64,0x64,0x64,0x64,0x64,0x64,0x44,0x44,0x64,0x44,
0x44,0x64,0x64,0x44,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,
0x64,0x64,0,0x64,0,0x64,0x64,0,0x44,0x64,0,0x64,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,
0,0,0,0,4,4,4,4,4,4,0,0,0,0,0,0,
0,0,0,0,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x64,0x64,0x64,0,
4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0x64,
0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x44,0x44,0x64,0x64,0x44,0x44,0x44,0x44,0x44,
0x64,0x44,0x44,0x64,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x64,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0x44,0x44,
0x44,0x44,0x44,0x44,0x44,4,0,0x44,0x44,0x44,0x44,0x64,0x44,4,4,0x44,
0x44,0,0x64,0x44,0x44,0x64,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,4,0,0x64,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x44,0x64,0x44,0x44,
0x64,0x44,0x44,0x64,0x64,0x64,0x44,0x64,0x64,0x44,0x64,0x44,0x44,0x44,0x64,0x44,
0x64,0x44,0x64,0x44,0x64,0x44,0x44,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,4,
4,4,4,4,4,4,4,4,4,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x64,0x44,
4,4,0,0,0,0,4,0,0,0x64,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0x44,0x44,0x44,0x44,4,0x44,0x44,0x44,0x44,0x44,4,0x44,0x44,0x44,
4,0x44,0x44,0x44,0x44,0x44,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0x64,0x64,0x64,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0x64,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,
0x44,0x44,0x44,0x44,0x44,0x44,4,0x64,0x44,0x44,0x64,0x44,0x44,0x64,0x44,0x44,
0x44,0x64,0x64,0x64,0x64,0x64,0x64,0x44,0x44,0x44,0x64,0x44,0x44,0x64,0x64,0x44,
0x44,0x44,0x44,0x44,4,4,4,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,4,0,0x64,0,0,0,0,4,4,4,
4,4,4,4,4,0,0,0,0,0x64,0,0,0,0x44,0x64,0x44,
0x44,4,4,4,0,0,0,0,0,0,0,0,0,0,4,4,
0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x64,0,0,0,
0,4,4,4,4,0,0,0,0,0,0,0,0,0x64,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,4,4,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0x44,0,
0,4,4,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,4,4,0,0,0,0,4,4,0,0,4,4,0x64,0,0,
0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,4,4,0,0,0,4,0,0,0,0,0,0,
0,0,0,0,0,4,4,4,4,4,0,4,4,0,0,0,
0,0x64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,4,4,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,4,
4,4,4,4,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x64,0,0,4,0,4,4,4,4,0,0,0,0,0,0,0,
0,0x64,0,0,0,0,0,0,0,4,4,0,0,0,0,0,
0,0,0,0,0,0,4,4,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,
0,0x64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,4,0,0,0,4,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,4,0,0,0,0,0,4,4,4,0,4,4,
4,0x64,0,0,0,0,0,0,0,0x64,0x64,0,0,0,0,0,
0,0,0,0,0,0,4,0,0,0,0,0,4,0x64,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
4,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0x64,0x64,0,0,0,0,0,0,0,0,0,0,0,
0,0,0x64,0,0,0,0,0,0,0,4,4,4,0,4,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,4,0,0,4,4,4,4,0x64,0x64,0x64,0,0,0,0,0,
0,0,4,4,0x64,0x64,0x64,0x64,4,4,4,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,
4,4,4,4,0x64,0x64,0x64,4,4,0,0,0,0,0,0,0,
0,0,4,0,0x64,0x64,0x64,0x64,4,4,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x64,0x64,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0x64,0,0x64,
0,0x64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0x64,0x64,4,0x64,4,4,4,4,4,0x64,0x64,
0x64,0x64,4,0,0x64,4,0x44,0x44,0x64,0,0x44,0x44,0,0,0,0,
0,4,4,4,4,4,4,4,4,4,4,4,0,4,4,4,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,4,4,4,4,4,4,4,4,4,4,4,0,0,0,
0,0,0,0,0,0,0x64,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,4,4,4,4,0,4,4,4,4,4,0x64,0,0x64,0x64,0,
0,4,4,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,4,4,0,0,
0,0,4,4,4,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,4,4,4,4,0,0,0,0,0,0,0,
0,0,0,0,0,0,4,0,0,4,4,0,0,0,0,0,
0,0x64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,4,0,0,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,
0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,
0x179a,0x179a,0x179a,0x179a,0x179a,0x179a,0,0x179a,0,0,0,0,0,0x179a,0,0,
0x17b9,0x17e9,0x1819,0x1849,0x1879,0x18a9,0x18d9,0x1909,0x1939,0x1969,0x1999,0x19c9,0x19f9,0x1a29,0x1a59,0x1a89,
0x1ab9,0x1ae9,0x1b19,0x1b49,0x1b79,0x1ba9,0x1bd9,0x1c09,0x1c39,0x1c69,0x1c99,0x1cc9,0x1cf9,0x1d29,0x1d59,0x1d89,
0x1db9,0x1de9,0x1e19,0x1e49,0x1e79,0x1ea9,0x1ed9,0x1f09,0x1f39,0x1f69,0x1f99,0,4,0x1fc9,0x1ff9,0x2029,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0x44,0x44,0x44,
0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,
0x207a,0x207a,0x207a,0x207a,0x207a,0x207a,0,0,0x2099,0x20c9,0x20f9,0x2129,0x2159,0x2189,0,0,
0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,
0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,0x205a,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,4,4,0x64,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,4,4,0,4,
4,4,4,4,4,4,0,0,0,0,0,0,0,0,4,0,
0,4,4,4,4,4,4,4,4,4,0x64,4,0,0,0,4,
0,0,0,0,0,0x44,0,0,0,0,0,0,0,0,0,0,
0,0,0,4,4,4,4,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,4,4,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0x64,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,4,4,4,0,
0,0,0,4,4,0,0,0,0,0,0,0,0,0,4,0,
0,0,0,0,0,0x64,0x44,0x64,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0x44,
0x64,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,4,0,4,4,4,4,
4,4,4,0,0x64,0,4,0,0,4,4,4,4,4,4,4,
4,0,0,0,0,0,0,4,4,0x44,0x44,0x44,0x44,0x44,0x44,0x44,
0x44,0,0,0x64,0,0,0,0,0,0,0,4,0,0,0,0,
0,0,0,0,0x44,0x44,0x44,0x44,0x44,0x64,0x64,0x64,0x64,0x64,0x64,0x44,
0x44,0x64,4,0x64,0x64,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x64,0,4,4,4,4,4,0,4,0,0,0,
0,0,4,0,0x60,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0x44,0x64,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,4,4,4,4,0,0,
4,4,0x60,0x64,4,4,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0x64,0,4,4,0,0,
0,4,0,4,4,4,0x60,0x60,0,0,0,0,0,0,0,0,
0,0,0,0,4,4,4,4,4,4,4,4,0,0,4,0x64,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,4,4,4,4,4,4,0,0,
0x21b9,0x21e9,0x2219,0x2249,0x2279,0x22c9,0x2319,0x2349,0x2379,0,0,0,0,0,0,0,
0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,
0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0x23aa,0,0,0x23aa,0x23aa,0x23aa,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x44,0x44,0x44,0,0x64,0x64,0x64,0x64,0x64,0x64,0x44,0x44,0x64,0x64,0x64,0x64,
0x44,0,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0,0,0,0,0x64,0,0,
0,0,0,0,0x44,0,0,0,0x44,0x44,0,0,0,0,0,0,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,0x25,5,5,5,5,5,5,5,5,1,1,1,1,1,
1,1,1,1,1,1,1,1,5,0x23c9,1,1,1,0x23e9,1,1,
5,5,5,5,0x25,5,5,5,0x25,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0x2409,1,
1,1,1,1,1,1,0x21,1,1,1,1,5,5,5,5,5,
0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,
0x44,0x44,0x44,0x44,0x44,0x44,0x64,0x64,0x64,0x64,0,0x44,0x64,0x64,0x44,0x64,
0x44,0x44,0x64,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x64,0x44,0x44,0x64,0x64,0x64,
0x64,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xffb1,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x242a,0x2469,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x24a9,0x2529,0x25a9,0x2629,0x26a9,0x2729,1,1,0x275a,1,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xffb1,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x411,0x411,0x411,0x411,
0x411,0x411,0x411,0x411,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0x411,0x411,0x411,0x411,
0x411,0x411,0,0,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0,0,0x411,0x411,0x411,0x411,
0x411,0x411,0x411,0x411,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0x411,0x411,0x411,0x411,
0x411,0x411,0x411,0x411,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0x411,0x411,0x411,0x411,
0x411,0x411,0,0,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0,0,0x27a9,0x411,0x2829,0x411,
0x28d9,0x411,0x2989,0x411,0,0xfc12,0,0xfc12,0,0xfc12,0,0xfc12,0x411,0x411,0x411,0x411,
0x411,0x411,0x411,0x411,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0xfc12,0x2511,0x2511,0x2b11,0x2b11,
0x2b11,0x2b11,0x3211,0x3211,0x4011,0x4011,0x3811,0x3811,0x3f11,0x3f11,0,0,0x2a39,0x2aa9,0x2b19,0x2b89,
0x2bf9,0x2c69,0x2cd9,0x2d49,0x2dbb,0x2e2b,0x2e9b,0x2f0b,0x2f7b,0x2feb,0x305b,0x30cb,0x3139,0x31a9,0x3219,0x3289,
0x32f9,0x3369,0x33d9,0x3449,0x34bb,0x352b,0x359b,0x360b,0x367b,0x36eb,0x375b,0x37cb,0x3839,0x38a9,0x3919,0x3989,
0x39f9,0x3a69,0x3ad9,0x3b49,0x3bbb,0x3c2b,0x3c9b,0x3d0b,0x3d7b,0x3deb,0x3e5b,0x3ecb,0x411,0x411,0x3f39,0x3fb9,
0x4029,0,0x40a9,0x4129,0xfc12,0xfc12,0xdb12,0xdb12,0x41db,4,0x4249,4,4,4,0x4299,0x4319,
0x4389,0,0x4409,0x4489,0xd512,0xd512,0xd512,0xd512,0x453b,4,4,4,0x411,0x411,0x45a9,0x4659,
0,0,0x4729,0x47a9,0xfc12,0xfc12,0xce12,0xce12,0,4,4,4,0x411,0x411,0x4859,0x4909,
0x49d9,0x391,0x4a59,0x4ad9,0xfc12,0xfc12,0xc812,0xc812,0xfc92,4,4,4,0,0,0x4b89,0x4c09,
0x4c79,0,0x4cf9,0x4d79,0xc012,0xc012,0xc112,0xc112,0x4e2b,4,4,0,0,0,0,0,
0,0,0,0,0,0,0,4,4,4,4,4,0,0,0,0,
0,0,0,0,4,4,0,0,0,0,0,0,4,0,0,4,
0,0,4,4,4,4,4,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,4,4,4,4,4,0,4,4,
4,4,4,4,4,4,4,4,0,0x25,0,0,0,0,0,0,
0,0,0,0,0,0,0,5,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x44,0x44,0x64,0x64,0x44,0x44,0x44,0x44,
0x64,0x64,0x64,0x44,0x44,4,4,4,4,0x44,4,4,4,0x64,0x64,0x44,
0x64,0x44,0x64,0x64,0x64,0x64,0x64,0x64,0x44,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,2,
0,0,1,2,2,2,1,1,2,2,2,1,0,2,0,0,
0,2,2,2,2,2,0,0,0,0,0,0,2,0,0x4e9a,0,
2,0,0x4eda,0x4f1a,2,2,0,1,2,2,0xe12,2,1,0,0,0,
0,1,0,0,1,1,2,2,0,0,0,0,0,2,1,1,
0x21,0x21,0,0,0,0,0xf211,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x812,0x812,0x812,0x812,0x812,0x812,0x812,0x812,
0x812,0x812,0x812,0x812,0x812,0x812,0x812,0x812,0xf811,0xf811,0xf811,0xf811,0xf811,0xf811,0xf811,0xf811,
0xf811,0xf811,0xf811,0xf811,0xf811,0xf811,0xf811,0xf811,0,0,0,0x92,0xff91,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0xd12,0xd12,0xd12,0xd12,0xd12,0xd12,
0xd12,0xd12,0xd12,0xd12,0xd12,0xd12,0xd12,0xd12,0xd12,0xd12,0xd12,0xd12,0xf311,0xf311,0xf311,0xf311,
0xf311,0xf311,0xf311,0xf311,0xf311,0xf311,0xf311,0xf311,0xf311,0xf311,0xf311,0xf311,0xf311,0xf311,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,
0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,
0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0x1812,0,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,
0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,
0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0xe811,0,0x92,0xff91,0x4f5a,0x4f7a,0x4f9a,0x4fb9,0x4fd9,0x92,
0xff91,0x92,0xff91,0x92,0xff91,0x4ffa,0x501a,0x503a,0x505a,1,0x92,0xff91,1,0x92,0xff91,1,
1,1,1,1,0x25,5,0x507a,0x507a,0x92,0xff91,0x92,0xff91,1,0,0,0,
0,0,0,0x92,0xff91,0x92,0xff91,0x44,0x44,0x44,0x92,0xff91,0,0,0,0,
0,0,0,0,0,0,0,0,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,
0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,
0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0x5099,0,0x5099,0,0,0,0,
0,0x5099,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0x64,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,
0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,
0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0,0,0,0,0,0,0,0,
0,0,0x64,0x64,0x64,0x64,0x60,0x60,0,4,4,4,4,4,0,0,
0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0x64,0x64,4,
4,4,4,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
4,4,4,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x50ba,0x50f9,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0,0x44,
4,4,4,0,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0,4,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,5,5,0x44,0x44,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x44,0x44,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
1,1,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,5,1,1,1,1,1,1,1,1,0x92,0xff91,0x92,
0xff91,0x513a,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,4,4,4,0x92,
0xff91,0x515a,1,0,0x92,0xff91,0x92,0xff91,0x1811,1,0x92,0xff91,0x92,0xff91,0x92,0xff91,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x517a,0x519a,0x51ba,0x51da,0x517a,1,0x51fa,0x521a,0x523a,0x525a,
0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0x92,0xff91,0,0,0x92,0xff91,
0xe812,0x527a,0x529a,0x92,0xff91,0x92,0xff91,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0x92,0xff91,0,
5,5,1,0,0,0,0,0,0,0,4,0,0,0,0x64,0,
0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,4,4,0,0,0,0,0,
0x64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x64,4,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,
0x44,0x44,0,0,0,0,0,0,0,0,0,0,0,0,0,4,
0,0,0,0,0,0,4,4,4,4,4,0x64,0x64,0x64,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,4,4,4,4,4,4,4,4,4,4,4,0,0x60,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0x64,0,0,4,4,4,4,0,0,4,4,0,0,
0x60,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,4,4,4,4,4,4,0,0,4,4,0,0,4,4,0,
0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,
0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,
0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x44,0,0x44,0x44,0x64,0,0,0x44,
0x44,0,0,0,0,0,0x44,0x44,0,0x44,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,4,4,0,0,0,0,0,4,4,0,0x64,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,0x52b9,1,1,1,1,1,1,1,4,5,5,5,5,
1,1,1,1,1,1,1,1,1,4,4,4,0,0,0,0,
0x52d9,0x5309,0x5339,0x5369,0x5399,0x53c9,0x53f9,0x5429,0x5459,0x5489,0x54b9,0x54e9,0x5519,0x5549,0x5579,0x55a9,
0x5bd9,0x5c09,0x5c39,0x5c69,0x5c99,0x5cc9,0x5cf9,0x5d29,0x5d59,0x5d89,0x5db9,0x5de9,0x5e19,0x5e49,0x5e79,0x5ea9,
0x5ed9,0x5f09,0x5f39,0x5f69,0x5f99,0x5fc9,0x5ff9,0x6029,0x6059,0x6089,0x60b9,0x60e9,0x6119,0x6149,0x6179,0x61a9,
0x55d9,0x5609,0x5639,0x5669,0x5699,0x56c9,0x56f9,0x5729,0x5759,0x5789,0x57b9,0x57e9,0x5819,0x5849,0x5879,0x58a9,
0x58d9,0x5909,0x5939,0x5969,0x5999,0x59c9,0x59f9,0x5a29,0x5a59,0x5a89,0x5ab9,0x5ae9,0x5b19,0x5b49,0x5b79,0x5ba9,
0,0,0,0,0,4,0,0,4,0,0,0,0,0x64,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x61d9,0x6259,0x62d9,0x6359,0x6409,0x64b9,0x6559,0,0,0,0,0,0,0,0,0,
0,0,0,0x65f9,0x6679,0x66f9,0x6779,0x67f9,0,0,0,0,0,0,0x64,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,4,4,0,0,0,4,0,0,0,0,0,0,0,0,
0,0,0,0,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x64,0x64,0x64,0x64,0x64,
0x64,0x64,0x44,0x44,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,4,0,0,4,0,0,0,0,0,0,
0,0,0,0,0,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,
0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0,
0,0,4,0,4,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,
0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,4,4,4,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0x64,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0x44,0x44,0x44,0x44,0x44,0,0,0,0,0,0x1412,0x1412,0x1412,0x1412,
0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,
0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0xec11,0xec11,0xec11,0xec11,
0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,
0xec11,0xec11,0xec11,0xec11,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,
0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0x1412,0,0,0,0,0xec11,0xec11,0xec11,0xec11,
0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,
0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0xec11,0,0,0,0,0,4,4,4,
0,4,4,0,0,0,0,0,4,0x64,4,0x44,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x44,0x64,0x64,0,0,0,0,0x64,0,0,0,0,
0,0x44,0x64,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x2012,0x2012,0x2012,0x2012,
0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,
0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0x2012,0,
0,0,0,0,0,0,0,0,0,0,0,0,0xe011,0xe011,0xe011,0xe011,
0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,
0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0xe011,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x44,0x44,0x44,0x44,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0x44,0x44,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0x64,0x64,0x44,0x44,0x44,0x64,0x44,0x64,0x64,0x64,0x64,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,4,4,4,4,4,4,4,4,
4,4,0x64,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,
4,4,4,0,0,0x64,0x64,0,0,4,0,0,0x44,0x44,0x44,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,
4,4,4,4,0,4,4,4,4,4,4,0x64,0x64,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0x64,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,4,4,4,4,4,4,
4,4,4,0,0x60,0,0,0,0,0,0,0,0,4,0x64,4,
4,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,4,4,4,0,0,4,0x60,0x64,4,
0,0,0,0,0,0,4,0,0,0,0,4,4,4,4,4,
4,0x64,0x64,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,
0,0,0,0,0,0x60,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0x44,0x44,0x44,0x44,0x44,0x44,
0x44,0,0,0,0x44,0x44,0x44,0x44,0x44,0,0,0,0,0,0,0,
0,0,0,0,0,0,0x64,4,4,0,0x64,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0x44,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,4,4,4,4,4,4,0,4,0,
0,0,0,4,4,0,0x64,0x64,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,4,4,4,4,0,0,0,0,0,0,
4,4,0,0x64,0x64,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
4,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,4,4,4,4,4,4,4,4,0,
0,4,0,0x64,0,0,0,0,0,0,0,0,0,0,0,4,
0,4,0,0,4,4,4,4,4,4,0x60,0x64,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,4,4,4,0,0,4,4,
4,4,0,4,4,4,4,0x64,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,
4,4,4,4,4,4,4,4,0,0x64,0x64,0,0,0,0,0,
0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,
0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,0x1012,
0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,
0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,0xf011,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,4,4,0x60,0x64,0,
0,0,0,0x64,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
4,4,4,4,0,0,4,4,0,0,0,0,0,4,4,4,
4,4,4,4,4,4,4,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,
0x64,4,4,4,4,0,0,4,4,4,4,0,0,0,0,0,
0,0,0,0x64,0,0,0,0,0,0,0,0,0,4,4,4,
4,4,4,0,0,4,4,4,0,0,0,0,0,0,0,0,
0,0,4,4,4,4,4,4,4,4,4,4,4,4,4,0,
4,0x64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,4,4,4,4,4,4,4,0,4,4,4,4,
4,4,0,0x64,4,4,4,4,4,4,4,4,0,0,4,4,
4,4,4,4,4,0,4,4,0,4,4,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,4,4,4,
4,4,4,0,0,0,4,0,4,4,0,4,4,4,0x64,4,
0x64,0x64,0,4,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,4,4,0,0,
0,4,0,0x64,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,4,4,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,4,4,4,4,
4,4,4,4,4,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x64,0x64,0x64,0x64,0x64,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x44,0x44,0x44,0x44,0x44,0x44,0x44,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,4,4,4,4,4,4,4,4,4,
4,4,4,4,4,4,4,4,4,4,0,4,4,0,0,0,
0,0,0,0,0,0,0,0,0x60,0x60,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,4,0x64,0,0,0,0,0,
0,0x60,0x60,0x64,0x64,0x64,0,0,0,0x60,0x60,0x60,0x60,0x60,0x60,4,
4,4,4,4,4,4,4,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0x64,0,
0,0x44,0x44,0x44,0x44,0x44,0x64,0x64,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0x44,0x44,
0x44,0x44,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0x44,0x44,0x44,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,1,1,
1,1,1,1,1,1,0x21,0x21,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,1,1,1,1,1,1,1,0,0x21,0x21,
1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,1,1,1,1,1,1,1,1,0x21,0x21,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,2,0,2,2,
0,0,2,0,0,2,2,0,0,2,2,2,2,0,2,2,
2,2,2,2,2,2,1,1,1,1,0,1,0,1,0x21,0x21,
1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
1,1,1,1,2,2,0,2,2,2,2,0,0,2,2,2,
2,2,2,2,2,0,2,2,2,2,2,2,2,0,1,1,
1,1,1,1,1,1,0x21,0x21,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,2,2,0,2,2,2,2,0,
2,2,2,2,2,0,2,0,0,0,2,2,2,2,2,2,
2,0,1,1,1,1,1,1,1,1,0x21,0x21,1,1,1,1,
1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,1,1,1,1,
1,1,0,0,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,0,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,0,1,1,1,1,1,1,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,0,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,
1,1,1,1,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,1,1,1,0,1,1,1,1,1,1,2,1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,4,4,4,4,4,4,4,4,4,0,0,0,0,4,
4,4,4,4,4,4,4,4,4,4,4,4,4,0,0,0,
0,0,0,0,0,4,0,0,0,0,0,0,0,0,0,0,
4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,4,4,4,4,4,0,4,4,4,
4,4,4,4,4,4,4,4,4,4,4,4,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x44,0x44,0x44,0x44,
0x44,0x44,0x44,0,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,
0x44,0x44,0x44,0x44,0x44,0,0,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0,0x44,
0x44,0,0x44,0x44,0x44,0x44,0x44,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x44,0x44,0x44,0x44,
0x44,0x44,0x44,4,4,4,4,4,4,4,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x64,0x64,0x64,0x64,
0x64,0x64,0x64,0,0,0,0,0,0,0,0,0,0x1112,0x1112,0x1112,0x1112,
0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,
0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0x1112,0xef11,0xef11,
0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,
0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0xef11,0x44,0x44,0x44,0x44,
0x44,0x44,0x64,4,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,
2,2,0,0,0,0,0,0,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0
};

static const uint16_t ucase_props_exceptions[1671]={
0xc850,0x20,2,0x130,0x131,0x4810,0x20,0x841,0x6b,1,0x212a,0x841,0x73,1,0x17f,0x5c50,
0x20,2,0x130,0x131,0x844,0x4b,1,0x212a,0x844,0x53,1,0x17f,0x806,0x3bc,0x39c,0x841,
0xe5,1,0x212b,0x8c0,1,0x2220,0x73,0x73,0x53,0x53,0x53,0x73,0x1e9e,0x844,0xc5,1,
0x212b,0x4810,1,0xce50,0xc7,2,0x49,0x131,0x844,0x49,2,0x69,0x130,0x880,0x2220,0x2bc,
0x6e,0x2bc,0x4e,0x2bc,0x4e,0x806,0x73,0x53,0x809,0x1c6,0x1c5,0x80d,0x1c6,0x1c4,0x1c5,0x80c,
0x1c4,0x1c5,0x809,0x1c9,0x1c8,0x80d,0x1c9,0x1c7,0x1c8,0x80c,0x1c7,0x1c8,0x809,0x1cc,0x1cb,0x80d,
0x1cc,0x1ca,0x1cb,0x80c,0x1ca,0x1cb,0x880,0x2220,0x6a,0x30c,0x4a,0x30c,0x4a,0x30c,0x809,0x1f3,
0x1f2,0x80d,0x1f3,0x1f1,0x1f2,0x80c,0x1f1,0x1f2,0x810,0x2a2b,0x810,0x2a28,0x810,0x2a3f,0x810,0x2a1f,
0x810,0x2a1c,0x810,0x2a1e,0x810,0xa54f,0x810,0xa54b,0x810,0xa528,0x810,0xa544,0x810,0x29f7,0x810,0xa541,
0x810,0x29fd,0x810,0x29e7,0x810,0xa543,0x810,0xa52a,0x1810,0xa515,0x810,0xa512,0x6800,0x3846,0x3b9,0x399,
1,0x1fbe,0x8c0,1,0x3330,0x3b9,0x308,0x301,0x399,0x308,0x301,0x399,0x308,0x301,0x1fd3,0x841,
0x3b2,1,0x3d0,0x841,0x3b5,1,0x3f5,0x841,0x3b8,2,0x3d1,0x3f4,0x841,0x3b9,2,0x345,
0x1fbe,0x841,0x3ba,1,0x3f0,0x841,0x3bc,1,0xb5,0x841,0x3c0,1,0x3d6,0x841,0x3c1,1,
0x3f1,0x4850,0x20,1,0x3c2,0x841,0x3c6,1,0x3d5,0x841,0x3c9,1,0x2126,0x8c0,1,0x3330,
0x3c5,0x308,0x301,0x3a5,0x308,0x301,0x3a5,0x308,0x301,0x1fe3,0x844,0x392,1,0x3d0,0x844,0x395,
1,0x3f5,0x844,0x398,2,0x3d1,0x3f4,0x844,0x399,2,0x345,0x1fbe,0x844,0x39a,1,0x3f0,
0x844,0x39c,1,0xb5,0x844,0x3a0,1,0x3d6,0x844,0x3a1,1,0x3f1,0x806,0x3c3,0x3a3,0x844,
0x3a3,1,0x3c2,0x844,0x3a6,1,0x3d5,0x844,0x3a9,1,0x2126,0x806,0x3b2,0x392,0x846,0x3b8,
0x398,1,0x3f4,0x806,0x3c6,0x3a6,0x806,0x3c0,0x3a0,0x806,0x3ba,0x39a,0x806,0x3c1,0x3a1,0x841,
0x3b8,2,0x398,0x3d1,0x806,0x3b5,0x395,0x841,0x432,1,0x1c80,0x841,0x434,1,0x1c81,0x841,
0x43e,1,0x1c82,0x841,0x441,1,0x1c83,0x841,0x442,2,0x1c84,0x1c85,0x841,0x44a,1,0x1c86,
0x844,0x412,1,0x1c80,0x844,0x414,1,0x1c81,0x844,0x41e,1,0x1c82,0x844,0x421,1,0x1c83,
0x844,0x422,2,0x1c84,0x1c85,0x844,0x42a,1,0x1c86,0x841,0x463,1,0x1c87,0x844,0x462,1,
0x1c87,0x880,0x2220,0x565,0x582,0x535,0x552,0x535,0x582,0x810,0x1c60,0x80c,0x1c90,0x10d0,0x80c,0x1c91,
0x10d1,0x80c,0x1c92,0x10d2,0x80c,0x1c93,0x10d3,0x80c,0x1c94,0x10d4,0x80c,0x1c95,0x10d5,0x80c,0x1c96,0x10d6,
0x80c,0x1c97,0x10d7,0x80c,0x1c98,0x10d8,0x80c,0x1c99,0x10d9,0x80c,0x1c9a,0x10da,0x80c,0x1c9b,0x10db,0x80c,
0x1c9c,0x10dc,0x80c,0x1c9d,0x10dd,0x80c,0x1c9e,0x10de,0x80c,0x1c9f,0x10df,0x80c,0x1ca0,0x10e0,0x80c,0x1ca1,
0x10e1,0x80c,0x1ca2,0x10e2,0x80c,0x1ca3,0x10e3,0x80c,0x1ca4,0x10e4,0x80c,0x1ca5,0x10e5,0x80c,0x1ca6,0x10e6,
0x80c,0x1ca7,0x10e7,0x80c,0x1ca8,0x10e8,0x80c,0x1ca9,0x10e9,0x80c,0x1caa,0x10ea,0x80c,0x1cab,0x10eb,0x80c,
0x1cac,0x10ec,0x80c,0x1cad,0x10ed,0x80c,0x1cae,0x10ee,0x80c,0x1caf,0x10ef,0x80c,0x1cb0,0x10f0,0x80c,0x1cb1,
0x10f1,0x80c,0x1cb2,0x10f2,0x80c,0x1cb3,0x10f3,0x80c,0x1cb4,0x10f4,0x80c,0x1cb5,0x10f5,0x80c,0x1cb6,0x10f6,
0x80c,0x1cb7,0x10f7,0x80c,0x1cb8,0x10f8,0x80c,0x1cb9,0x10f9,0x80c,0x1cba,0x10fa,0x80c,0x1cbd,0x10fd,0x80c,
0x1cbe,0x10fe,0x80c,0x1cbf,0x10ff,0xa10,0x97d0,0xa10,8,0x806,0x13f0,0x13f0,0x806,0x13f1,0x13f1,0x806,
0x13f2,0x13f2,0x806,0x13f3,0x13f3,0x806,0x13f4,0x13f4,0x806,0x13f5,0x13f5,0x806,0x432,0x412,0x806,0x434,
0x414,0x806,0x43e,0x41e,0x806,0x441,0x421,0x846,0x442,0x422,1,0x1c85,0x846,0x442,0x422,1,
0x1c84,0x806,0x44a,0x42a,0x806,0x463,0x462,0x806,0xa64b,0xa64a,0xc10,0xbc0,0x810,0x8a04,0x810,0xee6,
0x810,0x8a38,0x841,0x1e61,1,0x1e9b,0x844,0x1e60,1,0x1e9b,0x880,0x2220,0x68,0x331,0x48,0x331,
0x48,0x331,0x880,0x2220,0x74,0x308,0x54,0x308,0x54,0x308,0x880,0x2220,0x77,0x30a,0x57,0x30a,
0x57,0x30a,0x880,0x2220,0x79,0x30a,0x59,0x30a,0x59,0x30a,0x880,0x2220,0x61,0x2be,0x41,0x2be,
0x41,0x2be,0x806,0x1e61,0x1e60,0xc90,0x1dbf,0x20,0x73,0x73,0x880,0x2220,0x3c5,0x313,0x3a5,0x313,
0x3a5,0x313,0x880,0x3330,0x3c5,0x313,0x300,0x3a5,0x313,0x300,0x3a5,0x313,0x300,0x880,0x3330,0x3c5,
0x313,0x301,0x3a5,0x313,0x301,0x3a5,0x313,0x301,0x880,0x3330,0x3c5,0x313,0x342,0x3a5,0x313,0x342,
0x3a5,0x313,0x342,0x890,8,0x220,0x1f00,0x3b9,0x1f08,0x399,0x890,8,0x220,0x1f01,0x3b9,0x1f09,
0x399,0x890,8,0x220,0x1f02,0x3b9,0x1f0a,0x399,0x890,8,0x220,0x1f03,0x3b9,0x1f0b,0x399,0x890,
8,0x220,0x1f04,0x3b9,0x1f0c,0x399,0x890,8,0x220,0x1f05,0x3b9,0x1f0d,0x399,0x890,8,0x220,
0x1f06,0x3b9,0x1f0e,0x399,0x890,8,0x220,0x1f07,0x3b9,0x1f0f,0x399,0xc90,8,0x220,0x1f00,0x3b9,
0x1f08,0x399,0xc90,8,0x220,0x1f01,0x3b9,0x1f09,0x399,0xc90,8,0x220,0x1f02,0x3b9,0x1f0a,0x399,
0xc90,8,0x220,0x1f03,0x3b9,0x1f0b,0x399,0xc90,8,0x220,0x1f04,0x3b9,0x1f0c,0x399,0xc90,8,
0x220,0x1f05,0x3b9,0x1f0d,0x399,0xc90,8,0x220,0x1f06,0x3b9,0x1f0e,0x399,0xc90,8,0x220,0x1f07,
0x3b9,0x1f0f,0x399,0x890,8,0x220,0x1f20,0x3b9,0x1f28,0x399,0x890,8,0x220,0x1f21,0x3b9,0x1f29,
0x399,0x890,8,0x220,0x1f22,0x3b9,0x1f2a,0x399,0x890,8,0x220,0x1f23,0x3b9,0x1f2b,0x399,0x890,
8,0x220,0x1f24,0x3b9,0x1f2c,0x399,0x890,8,0x220,0x1f25,0x3b9,0x1f2d,0x399,0x890,8,0x220,
0x1f26,0x3b9,0x1f2e,0x399,0x890,8,0x220,0x1f27,0x3b9,0x1f2f,0x399,0xc90,8,0x220,0x1f20,0x3b9,
0x1f28,0x399,0xc90,8,0x220,0x1f21,0x3b9,0x1f29,0x399,0xc90,8,0x220,0x1f22,0x3b9,0x1f2a,0x399,
0xc90,8,0x220,0x1f23,0x3b9,0x1f2b,0x399,0xc90,8,0x220,0x1f24,0x3b9,0x1f2c,0x399,0xc90,8,
0x220,0x1f25,0x3b9,0x1f2d,0x399,0xc90,8,0x220,0x1f26,0x3b9,0x1f2e,0x399,0xc90,8,0x220,0x1f27,
0x3b9,0x1f2f,0x399,0x890,8,0x220,0x1f60,0x3b9,0x1f68,0x399,0x890,8,0x220,0x1f61,0x3b9,0x1f69,
0x399,0x890,8,0x220,0x1f62,0x3b9,0x1f6a,0x399,0x890,8,0x220,0x1f63,0x3b9,0x1f6b,0x399,0x890,
8,0x220,0x1f64,0x3b9,0x1f6c,0x399,0x890,8,0x220,0x1f65,0x3b9,0x1f6d,0x399,0x890,8,0x220,
0x1f66,0x3b9,0x1f6e,0x399,0x890,8,0x220,0x1f67,0x3b9,0x1f6f,0x399,0xc90,8,0x220,0x1f60,0x3b9,
0x1f68,0x399,0xc90,8,0x220,0x1f61,0x3b9,0x1f69,0x399,0xc90,8,0x220,0x1f62,0x3b9,0x1f6a,0x399,
0xc90,8,0x220,0x1f63,0x3b9,0x1f6b,0x399,0xc90,8,0x220,0x1f64,0x3b9,0x1f6c,0x399,0xc90,8,
0x220,0x1f65,0x3b9,0x1f6d,0x399,0xc90,8,0x220,0x1f66,0x3b9,0x1f6e,0x399,0xc90,8,0x220,0x1f67,
0x3b9,0x1f6f,0x399,0x880,0x2220,0x1f70,0x3b9,0x1fba,0x399,0x1fba,0x345,0x890,9,0x220,0x3b1,0x3b9,
0x391,0x399,0x880,0x2220,0x3ac,0x3b9,0x386,0x399,0x386,0x345,0x880,0x2220,0x3b1,0x342,0x391,0x342,
0x391,0x342,0x880,0x3330,0x3b1,0x342,0x3b9,0x391,0x342,0x399,0x391,0x342,0x345,0xc90,9,0x220,
0x3b1,0x3b9,0x391,0x399,0x846,0x3b9,0x399,1,0x345,0x880,0x2220,0x1f74,0x3b9,0x1fca,0x399,0x1fca,
0x345,0x890,9,0x220,0x3b7,0x3b9,0x397,0x399,0x880,0x2220,0x3ae,0x3b9,0x389,0x399,0x389,0x345,
0x880,0x2220,0x3b7,0x342,0x397,0x342,0x397,0x342,0x880,0x3330,0x3b7,0x342,0x3b9,0x397,0x342,0x399,
0x397,0x342,0x345,0xc90,9,0x220,0x3b7,0x3b9,0x397,0x399,0x880,0x3330,0x3b9,0x308,0x300,0x399,
0x308,0x300,0x399,0x308,0x300,0x8c0,1,0x3330,0x3b9,0x308,0x301,0x399,0x308,0x301,0x399,0x308,
0x301,0x390,0x880,0x2220,0x3b9,0x342,0x399,0x342,0x399,0x342,0x880,0x3330,0x3b9,0x308,0x342,0x399,
0x308,0x342,0x399,0x308,0x342,0x880,0x3330,0x3c5,0x308,0x300,0x3a5,0x308,0x300,0x3a5,0x308,0x300,
0x8c0,1,0x3330,0x3c5,0x308,0x301,0x3a5,0x308,0x301,0x3a5,0x308,0x301,0x3b0,0x880,0x2220,0x3c1,
0x313,0x3a1,0x313,0x3a1,0x313,0x880,0x2220,0x3c5,0x342,0x3a5,0x342,0x3a5,0x342,0x880,0x3330,0x3c5,
0x308,0x342,0x3a5,0x308,0x342,0x3a5,0x308,0x342,0x880,0x2220,0x1f7c,0x3b9,0x1ffa,0x399,0x1ffa,0x345,
0x890,9,0x220,0x3c9,0x3b9,0x3a9,0x399,0x880,0x2220,0x3ce,0x3b9,0x38f,0x399,0x38f,0x345,0x880,
0x2220,0x3c9,0x342,0x3a9,0x342,0x3a9,0x342,0x880,0x3330,0x3c9,0x342,0x3b9,0x3a9,0x342,0x399,0x3a9,
0x342,0x345,0xc90,9,0x220,0x3c9,0x3b9,0x3a9,0x399,0xc50,0x1d5d,1,0x3a9,0xc50,0x20bf,1,
0x4b,0xc50,0x2046,1,0xc5,0xc10,0x29f7,0xc10,0xee6,0xc10,0x29e7,0xc10,0x2a2b,0xc10,0x2a28,0xc10,
0x2a1c,0xc10,0x29fd,0xc10,0x2a1f,0xc10,0x2a1e,0xc10,0x2a3f,0xc10,0x1c60,0x841,0xa64b,1,0x1c88,0x844,
0xa64a,1,0x1c88,0xc10,0x8a04,0xc10,0xa528,0xc10,0xa544,0xc10,0xa54f,0xc10,0xa54b,0xc10,0xa541,0xc10,
0xa512,0xc10,0xa52a,0xc10,0xa515,0x810,0x3a0,0xc10,0xa543,0xc10,0x8a38,0xc10,0x3a0,0x806,0x13a0,0x13a0,
0x806,0x13a1,0x13a1,0x806,0x13a2,0x13a2,0x806,0x13a3,0x13a3,0x806,0x13a4,0x13a4,0x806,0x13a5,0x13a5,0x806,
0x13a6,0x13a6,0x806,0x13a7,0x13a7,0x806,0x13a8,0x13a8,0x806,0x13a9,0x13a9,0x806,0x13aa,0x13aa,0x806,0x13ab,
0x13ab,0x806,0x13ac,0x13ac,0x806,0x13ad,0x13ad,0x806,0x13ae,0x13ae,0x806,0x13af,0x13af,0x806,0x13b0,0x13b0,
0x806,0x13b1,0x13b1,0x806,0x13b2,0x13b2,0x806,0x13b3,0x13b3,0x806,0x13b4,0x13b4,0x806,0x13b5,0x13b5,0x806,
0x13b6,0x13b6,0x806,0x13b7,0x13b7,0x806,0x13b8,0x13b8,0x806,0x13b9,0x13b9,0x806,0x13ba,0x13ba,0x806,0x13bb,
0x13bb,0x806,0x13bc,0x13bc,0x806,0x13bd,0x13bd,0x806,0x13be,0x13be,0x806,0x13bf,0x13bf,0x806,0x13c0,0x13c0,
0x806,0x13c1,0x13c1,0x806,0x13c2,0x13c2,0x806,0x13c3,0x13c3,0x806,0x13c4,0x13c4,0x806,0x13c5,0x13c5,0x806,
0x13c6,0x13c6,0x806,0x13c7,0x13c7,0x806,0x13c8,0x13c8,0x806,0x13c9,0x13c9,0x806,0x13ca,0x13ca,0x806,0x13cb,
0x13cb,0x806,0x13cc,0x13cc,0x806,0x13cd,0x13cd,0x806,0x13ce,0x13ce,0x806,0x13cf,0x13cf,0x806,0x13d0,0x13d0,
0x806,0x13d1,0x13d1,0x806,0x13d2,0x13d2,0x806,0x13d3,0x13d3,0x806,0x13d4,0x13d4,0x806,0x13d5,0x13d5,0x806,
0x13d6,0x13d6,0x806,0x13d7,0x13d7,0x806,0x13d8,0x13d8,0x806,0x13d9,0x13d9,0x806,0x13da,0x13da,0x806,0x13db,
0x13db,0x806,0x13dc,0x13dc,0x806,0x13dd,0x13dd,0x806,0x13de,0x13de,0x806,0x13df,0x13df,0x806,0x13e0,0x13e0,
0x806,0x13e1,0x13e1,0x806,0x13e2,0x13e2,0x806,0x13e3,0x13e3,0x806,0x13e4,0x13e4,0x806,0x13e5,0x13e5,0x806,
0x13e6,0x13e6,0x806,0x13e7,0x13e7,0x806,0x13e8,0x13e8,0x806,0x13e9,0x13e9,0x806,0x13ea,0x13ea,0x806,0x13eb,
0x13eb,0x806,0x13ec,0x13ec,0x806,0x13ed,0x13ed,0x806,0x13ee,0x13ee,0x806,0x13ef,0x13ef,0x880,0x2220,0x66,
0x66,0x46,0x46,0x46,0x66,0x880,0x2220,0x66,0x69,0x46,0x49,0x46,0x69,0x880,0x2220,0x66,
0x6c,0x46,0x4c,0x46,0x6c,0x880,0x3330,0x66,0x66,0x69,0x46,0x46,0x49,0x46,0x66,0x69,
0x880,0x3330,0x66,0x66,0x6c,0x46,0x46,0x4c,0x46,0x66,0x6c,0x8c0,1,0x2220,0x73,0x74,
0x53,0x54,0x53,0x74,0xfb06,0x8c0,1,0x2220,0x73,0x74,0x53,0x54,0x53,0x74,0xfb05,0x880,
0x2220,0x574,0x576,0x544,0x546,0x544,0x576,0x880,0x2220,0x574,0x565,0x544,0x535,0x544,0x565,0x880,
0x2220,0x574,0x56b,0x544,0x53b,0x544,0x56b,0x880,0x2220,0x57e,0x576,0x54e,0x546,0x54e,0x576,0x880,
0x2220,0x574,0x56d,0x544,0x53d,0x544,0x56d
};

static const uint16_t ucase_props_unfold[370]={
0x49,5,3,0,0,0x61,0x2be,0,0x1e9a,0,0x66,0x66,0,0xfb00,0,0x66,
0x66,0x69,0xfb03,0,0x66,0x66,0x6c,0xfb04,0,0x66,0x69,0,0xfb01,0,0x66,0x6c,
0,0xfb02,0,0x68,0x331,0,0x1e96,0,0x69,0x307,0,0x130,0,0x6a,0x30c,0,
0x1f0,0,0x73,0x73,0,0xdf,0x1e9e,0x73,0x74,0,0xfb05,0xfb06,0x74,0x308,0,0x1e97,
0,0x77,0x30a,0,0x1e98,0,0x79,0x30a,0,0x1e99,0,0x2bc,0x6e,0,0x149,0,
0x3ac,0x3b9,0,0x1fb4,0,0x3ae,0x3b9,0,0x1fc4,0,0x3b1,0x342,0,0x1fb6,0,0x3b1,
0x342,0x3b9,0x1fb7,0,0x3b1,0x3b9,0,0x1fb3,0x1fbc,0x3b7,0x342,0,0x1fc6,0,0x3b7,0x342,
0x3b9,0x1fc7,0,0x3b7,0x3b9,0,0x1fc3,0x1fcc,0x3b9,0x308,0x300,0x1fd2,0,0x3b9,0x308,0x301,
0x390,0x1fd3,0x3b9,0x308,0x342,0x1fd7,0,0x3b9,0x342,0,0x1fd6,0,0x3c1,0x313,0,0x1fe4,
0,0x3c5,0x308,0x300,0x1fe2,0,0x3c5,0x308,0x301,0x3b0,0x1fe3,0x3c5,0x308,0x342,0x1fe7,0,
0x3c5,0x313,0,0x1f50,0,0x3c5,0x313,0x300,0x1f52,0,0x3c5,0x313,0x301,0x1f54,0,0x3c5,
0x313,0x342,0x1f56,0,0x3c5,0x342,0,0x1fe6,0,0x3c9,0x342,0,0x1ff6,0,0x3c9,0x342,
0x3b9,0x1ff7,0,0x3c9,0x3b9,0,0x1ff3,0x1ffc,0x3ce,0x3b9,0,0x1ff4,0,0x565,0x582,0,
0x587,0,0x574,0x565,0,0xfb14,0,0x574,0x56b,0,0xfb15,0,0x574,0x56d,0,0xfb17,
0,0x574,0x576,0,0xfb13,0,0x57e,0x576,0,0xfb16,0,0x1f00,0x3b9,0,0x1f80,0x1f88,
0x1f01,0x3b9,0,0x1f81,0x1f89,0x1f02,0x3b9,0,0x1f82,0x1f8a,0x1f03,0x3b9,0,0x1f83,0x1f8b,0x1f04,
0x3b9,0,0x1f84,0x1f8c,0x1f05,0x3b9,0,0x1f85,0x1f8d,0x1f06,0x3b9,0,0x1f86,0x1f8e,0x1f07,0x3b9,
0,0x1f87,0x1f8f,0x1f20,0x3b9,0,0x1f90,0x1f98,0x1f21,0x3b9,0,0x1f91,0x1f99,0x1f22,0x3b9,0,
0x1f92,0x1f9a,0x1f23,0x3b9,0,0x1f93,0x1f9b,0x1f24,0x3b9,0,0x1f94,0x1f9c,0x1f25,0x3b9,0,0x1f95,
0x1f9d,0x1f26,0x3b9,0,0x1f96,0x1f9e,0x1f27,0x3b9,0,0x1f97,0x1f9f,0x1f60,0x3b9,0,0x1fa0,0x1fa8,
0x1f61,0x3b9,0,0x1fa1,0x1fa9,0x1f62,0x3b9,0,0x1fa2,0x1faa,0x1f63,0x3b9,0,0x1fa3,0x1fab,0x1f64,
0x3b9,0,0x1fa4,0x1fac,0x1f65,0x3b9,0,0x1fa5,0x1fad,0x1f66,0x3b9,0,0x1fa6,0x1fae,0x1f67,0x3b9,
0,0x1fa7,0x1faf,0x1f70,0x3b9,0,0x1fb2,0,0x1f74,0x3b9,0,0x1fc2,0,0x1f7c,0x3b9,0,
0x1ff2,0
};

static const UCaseProps ucase_props_singleton={
  NULL,
  ucase_props_indexes,
  ucase_props_exceptions,
  ucase_props_unfold,
  {
    ucase_props_trieIndex,
    ucase_props_trieIndex+3288,
    NULL,
    3288,
    9068,
    0x188,
    0xd54,
    0x0,
    0x0,
    0xe0800,
    0x3040,
    NULL, 0, FALSE, FALSE, 0, NULL
  },
  { 4,0,0,0 }
};

#endif  // INCLUDED_FROM_UCASE_CPP


/* set of property starts for UnicodeSet ------------------------------------ */

static UBool U_CALLCONV
ucase_enumPropertyStartsRange(const void *context, UChar32 start, UChar32 /*end*/, uint32_t /*value*/) {
    /* add the start code point to the USet */
    const USetAdder *sa=(const USetAdder *)context;
    sa->add(sa->set, start);
    return TRUE;
}

U_CFUNC void U_EXPORT2
ucase_addPropertyStarts(const USetAdder *sa, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return;
    }

    /* add the start code point of each same-value range of the trie */
    utrie2_enum(&ucase_props_singleton.trie, NULL, ucase_enumPropertyStartsRange, sa);

    /* add code points with hardcoded properties, plus the ones following them */

    /* (none right now, see comment below) */

    /*
     * Omit code points with hardcoded specialcasing properties
     * because we do not build property UnicodeSets for them right now.
     */
}

/* data access primitives --------------------------------------------------- */

U_CFUNC const UTrie2 * U_EXPORT2
ucase_getTrie() {
    return &ucase_props_singleton.trie;
}

#define GET_EXCEPTIONS(csp, props) ((csp)->exceptions+((props)>>UCASE_EXC_SHIFT))

/* number of bits in an 8-bit integer value */
static const uint8_t flagsOffset[256]={
    0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};

#define HAS_SLOT(flags, idx) ((flags)&(1<<(idx)))
#define SLOT_OFFSET(flags, idx) flagsOffset[(flags)&((1<<(idx))-1)]

/*
 * Get the value of an optional-value slot where HAS_SLOT(excWord, idx).
 *
 * @param excWord (in) initial exceptions word
 * @param idx (in) desired slot index
 * @param pExc16 (in/out) const uint16_t * after excWord=*pExc16++;
 *               moved to the last uint16_t of the value, use +1 for beginning of next slot
 * @param value (out) int32_t or uint32_t output if hasSlot, otherwise not modified
 */
#define GET_SLOT_VALUE(excWord, idx, pExc16, value) UPRV_BLOCK_MACRO_BEGIN { \
    if(((excWord)&UCASE_EXC_DOUBLE_SLOTS)==0) { \
        (pExc16)+=SLOT_OFFSET(excWord, idx); \
        (value)=*pExc16; \
    } else { \
        (pExc16)+=2*SLOT_OFFSET(excWord, idx); \
        (value)=*pExc16++; \
        (value)=((value)<<16)|*pExc16; \
    } \
} UPRV_BLOCK_MACRO_END

/* simple case mappings ----------------------------------------------------- */

U_CAPI UChar32 U_EXPORT2
ucase_tolower(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    if(!UCASE_HAS_EXCEPTION(props)) {
        if(UCASE_IS_UPPER_OR_TITLE(props)) {
            c+=UCASE_GET_DELTA(props);
        }
    } else {
        const uint16_t *pe=GET_EXCEPTIONS(&ucase_props_singleton, props);
        uint16_t excWord=*pe++;
        if(HAS_SLOT(excWord, UCASE_EXC_DELTA) && UCASE_IS_UPPER_OR_TITLE(props)) {
            int32_t delta;
            GET_SLOT_VALUE(excWord, UCASE_EXC_DELTA, pe, delta);
            return (excWord&UCASE_EXC_DELTA_IS_NEGATIVE)==0 ? c+delta : c-delta;
        }
        if(HAS_SLOT(excWord, UCASE_EXC_LOWER)) {
            GET_SLOT_VALUE(excWord, UCASE_EXC_LOWER, pe, c);
        }
    }
    return c;
}

U_CAPI UChar32 U_EXPORT2
ucase_toupper(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    if(!UCASE_HAS_EXCEPTION(props)) {
        if(UCASE_GET_TYPE(props)==UCASE_LOWER) {
            c+=UCASE_GET_DELTA(props);
        }
    } else {
        const uint16_t *pe=GET_EXCEPTIONS(&ucase_props_singleton, props);
        uint16_t excWord=*pe++;
        if(HAS_SLOT(excWord, UCASE_EXC_DELTA) && UCASE_GET_TYPE(props)==UCASE_LOWER) {
            int32_t delta;
            GET_SLOT_VALUE(excWord, UCASE_EXC_DELTA, pe, delta);
            return (excWord&UCASE_EXC_DELTA_IS_NEGATIVE)==0 ? c+delta : c-delta;
        }
        if(HAS_SLOT(excWord, UCASE_EXC_UPPER)) {
            GET_SLOT_VALUE(excWord, UCASE_EXC_UPPER, pe, c);
        }
    }
    return c;
}

U_CAPI UChar32 U_EXPORT2
ucase_totitle(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    if(!UCASE_HAS_EXCEPTION(props)) {
        if(UCASE_GET_TYPE(props)==UCASE_LOWER) {
            c+=UCASE_GET_DELTA(props);
        }
    } else {
        const uint16_t *pe=GET_EXCEPTIONS(&ucase_props_singleton, props);
        uint16_t excWord=*pe++;
        if(HAS_SLOT(excWord, UCASE_EXC_DELTA) && UCASE_GET_TYPE(props)==UCASE_LOWER) {
            int32_t delta;
            GET_SLOT_VALUE(excWord, UCASE_EXC_DELTA, pe, delta);
            return (excWord&UCASE_EXC_DELTA_IS_NEGATIVE)==0 ? c+delta : c-delta;
        }
        int32_t idx;
        if(HAS_SLOT(excWord, UCASE_EXC_TITLE)) {
            idx=UCASE_EXC_TITLE;
        } else if(HAS_SLOT(excWord, UCASE_EXC_UPPER)) {
            idx=UCASE_EXC_UPPER;
        } else {
            return c;
        }
        GET_SLOT_VALUE(excWord, idx, pe, c);
    }
    return c;
}

static const UChar iDot[2] = { 0x69, 0x307 };
static const UChar jDot[2] = { 0x6a, 0x307 };
static const UChar iOgonekDot[3] = { 0x12f, 0x307 };
static const UChar iDotGrave[3] = { 0x69, 0x307, 0x300 };
static const UChar iDotAcute[3] = { 0x69, 0x307, 0x301 };
static const UChar iDotTilde[3] = { 0x69, 0x307, 0x303 };


U_CFUNC void U_EXPORT2
ucase_addCaseClosure(UChar32 c, const USetAdder *sa) {
    uint16_t props;

    /*
     * Hardcode the case closure of i and its relatives and ignore the
     * data file data for these characters.
     * The Turkic dotless i and dotted I with their case mapping conditions
     * and case folding option make the related characters behave specially.
     * This code matches their closure behavior to their case folding behavior.
     */

    switch(c) {
    case 0x49:
        /* regular i and I are in one equivalence class */
        sa->add(sa->set, 0x69);
        return;
    case 0x69:
        sa->add(sa->set, 0x49);
        return;
    case 0x130:
        /* dotted I is in a class with <0069 0307> (for canonical equivalence with <0049 0307>) */
        sa->addString(sa->set, iDot, 2);
        return;
    case 0x131:
        /* dotless i is in a class by itself */
        return;
    default:
        /* otherwise use the data file data */
        break;
    }

    props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    if(!UCASE_HAS_EXCEPTION(props)) {
        if(UCASE_GET_TYPE(props)!=UCASE_NONE) {
            /* add the one simple case mapping, no matter what type it is */
            int32_t delta=UCASE_GET_DELTA(props);
            if(delta!=0) {
                sa->add(sa->set, c+delta);
            }
        }
    } else {
        /*
         * c has exceptions, so there may be multiple simple and/or
         * full case mappings. Add them all.
         */
        const uint16_t *pe0, *pe=GET_EXCEPTIONS(&ucase_props_singleton, props);
        const UChar *closure;
        uint16_t excWord=*pe++;
        int32_t idx, closureLength, fullLength, length;

        pe0=pe;

        /* add all simple case mappings */
        for(idx=UCASE_EXC_LOWER; idx<=UCASE_EXC_TITLE; ++idx) {
            if(HAS_SLOT(excWord, idx)) {
                pe=pe0;
                GET_SLOT_VALUE(excWord, idx, pe, c);
                sa->add(sa->set, c);
            }
        }
        if(HAS_SLOT(excWord, UCASE_EXC_DELTA)) {
            pe=pe0;
            int32_t delta;
            GET_SLOT_VALUE(excWord, UCASE_EXC_DELTA, pe, delta);
            sa->add(sa->set, (excWord&UCASE_EXC_DELTA_IS_NEGATIVE)==0 ? c+delta : c-delta);
        }

        /* get the closure string pointer & length */
        if(HAS_SLOT(excWord, UCASE_EXC_CLOSURE)) {
            pe=pe0;
            GET_SLOT_VALUE(excWord, UCASE_EXC_CLOSURE, pe, closureLength);
            closureLength&=UCASE_CLOSURE_MAX_LENGTH; /* higher bits are reserved */
            closure=(const UChar *)pe+1; /* behind this slot, unless there are full case mappings */
        } else {
            closureLength=0;
            closure=NULL;
        }

        /* add the full case folding */
        if(HAS_SLOT(excWord, UCASE_EXC_FULL_MAPPINGS)) {
            pe=pe0;
            GET_SLOT_VALUE(excWord, UCASE_EXC_FULL_MAPPINGS, pe, fullLength);

            /* start of full case mapping strings */
            ++pe;

            fullLength&=0xffff; /* bits 16 and higher are reserved */

            /* skip the lowercase result string */
            pe+=fullLength&UCASE_FULL_LOWER;
            fullLength>>=4;

            /* add the full case folding string */
            length=fullLength&0xf;
            if(length!=0) {
                sa->addString(sa->set, (const UChar *)pe, length);
                pe+=length;
            }

            /* skip the uppercase and titlecase strings */
            fullLength>>=4;
            pe+=fullLength&0xf;
            fullLength>>=4;
            pe+=fullLength;

            closure=(const UChar *)pe; /* behind full case mappings */
        }

        /* add each code point in the closure string */
        for(idx=0; idx<closureLength;) {
            U16_NEXT_UNSAFE(closure, idx, c);
            sa->add(sa->set, c);
        }
    }
}

/*
 * compare s, which has a length, with t, which has a maximum length or is NUL-terminated
 * must be length>0 and max>0 and length<=max
 */
static inline int32_t
strcmpMax(const UChar *s, int32_t length, const UChar *t, int32_t max) {
    int32_t c1, c2;

    max-=length; /* we require length<=max, so no need to decrement max in the loop */
    do {
        c1=*s++;
        c2=*t++;
        if(c2==0) {
            return 1; /* reached the end of t but not of s */
        }
        c1-=c2;
        if(c1!=0) {
            return c1; /* return difference result */
        }
    } while(--length>0);
    /* ends with length==0 */

    if(max==0 || *t==0) {
        return 0; /* equal to length of both strings */
    } else {
        return -max; /* return lengh difference */
    }
}

U_CFUNC UBool U_EXPORT2
ucase_addStringCaseClosure(const UChar *s, int32_t length, const USetAdder *sa) {
    int32_t i, start, limit, result, unfoldRows, unfoldRowWidth, unfoldStringWidth;

    if(ucase_props_singleton.unfold==NULL || s==NULL) {
        return FALSE; /* no reverse case folding data, or no string */
    }
    if(length<=1) {
        /* the string is too short to find any match */
        /*
         * more precise would be:
         * if(!u_strHasMoreChar32Than(s, length, 1))
         * but this does not make much practical difference because
         * a single supplementary code point would just not be found
         */
        return FALSE;
    }

    const uint16_t *unfold=ucase_props_singleton.unfold;
    unfoldRows=unfold[UCASE_UNFOLD_ROWS];
    unfoldRowWidth=unfold[UCASE_UNFOLD_ROW_WIDTH];
    unfoldStringWidth=unfold[UCASE_UNFOLD_STRING_WIDTH];
    unfold+=unfoldRowWidth;

    if(length>unfoldStringWidth) {
        /* the string is too long to find any match */
        return FALSE;
    }

    /* do a binary search for the string */
    start=0;
    limit=unfoldRows;
    while(start<limit) {
        i=(start+limit)/2;
        const UChar *p=reinterpret_cast<const UChar *>(unfold+(i*unfoldRowWidth));
        result=strcmpMax(s, length, p, unfoldStringWidth);

        if(result==0) {
            /* found the string: add each code point, and its case closure */
            UChar32 c;

            for(i=unfoldStringWidth; i<unfoldRowWidth && p[i]!=0;) {
                U16_NEXT_UNSAFE(p, i, c);
                sa->add(sa->set, c);
                ucase_addCaseClosure(c, sa);
            }
            return TRUE;
        } else if(result<0) {
            limit=i;
        } else /* result>0 */ {
            start=i+1;
        }
    }

    return FALSE; /* string not found */
}

U_NAMESPACE_BEGIN

FullCaseFoldingIterator::FullCaseFoldingIterator()
        : unfold(reinterpret_cast<const UChar *>(ucase_props_singleton.unfold)),
          unfoldRows(unfold[UCASE_UNFOLD_ROWS]),
          unfoldRowWidth(unfold[UCASE_UNFOLD_ROW_WIDTH]),
          unfoldStringWidth(unfold[UCASE_UNFOLD_STRING_WIDTH]),
          currentRow(0),
          rowCpIndex(unfoldStringWidth) {
    unfold+=unfoldRowWidth;
}

UChar32
FullCaseFoldingIterator::next(UnicodeString &full) {
    // Advance past the last-delivered code point.
    const UChar *p=unfold+(currentRow*unfoldRowWidth);
    if(rowCpIndex>=unfoldRowWidth || p[rowCpIndex]==0) {
        ++currentRow;
        p+=unfoldRowWidth;
        rowCpIndex=unfoldStringWidth;
    }
    if(currentRow>=unfoldRows) { return U_SENTINEL; }
    // Set "full" to the NUL-terminated string in the first unfold column.
    int32_t length=unfoldStringWidth;
    while(length>0 && p[length-1]==0) { --length; }
    full.setTo(FALSE, p, length);
    // Return the code point.
    UChar32 c;
    U16_NEXT_UNSAFE(p, rowCpIndex, c);
    return c;
}

namespace LatinCase {

const int8_t TO_LOWER_NORMAL[LIMIT] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
    32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, EXC, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
    32, 32, 32, 32, 32, 32, 32, 0, 32, 32, 32, 32, 32, 32, 32, EXC,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    EXC, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1,

    0, 1, 0, 1, 0, 1, 0, 1, 0, EXC, 1, 0, 1, 0, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, -121, 1, 0, 1, 0, 1, 0, EXC
};

const int8_t TO_LOWER_TR_LT[LIMIT] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 32, 32, 32, 32, 32, 32, 32, 32, EXC, EXC, 32, 32, 32, 32, 32,
    32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, EXC, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, EXC, EXC, 32, 32,
    32, 32, 32, 32, 32, 32, 32, 0, 32, 32, 32, 32, 32, 32, 32, EXC,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, EXC, 0, 1, 0, 1, 0, EXC, 0,
    EXC, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1,

    0, 1, 0, 1, 0, 1, 0, 1, 0, EXC, 1, 0, 1, 0, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, -121, 1, 0, 1, 0, 1, 0, EXC
};

const int8_t TO_UPPER_NORMAL[LIMIT] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32,
    -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, EXC, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, EXC,
    -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32,
    -32, -32, -32, -32, -32, -32, -32, 0, -32, -32, -32, -32, -32, -32, -32, 121,

    0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1,
    0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1,
    0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1,
    0, EXC, 0, -1, 0, -1, 0, -1, 0, 0, -1, 0, -1, 0, -1, 0,

    -1, 0, -1, 0, -1, 0, -1, 0, -1, EXC, 0, -1, 0, -1, 0, -1,
    0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1,
    0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1,
    0, -1, 0, -1, 0, -1, 0, -1, 0, 0, -1, 0, -1, 0, -1, EXC
};

const int8_t TO_UPPER_TR[LIMIT] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, -32, -32, -32, -32, -32, -32, -32, -32, EXC, -32, -32, -32, -32, -32, -32,
    -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, EXC, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, EXC,
    -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32, -32,
    -32, -32, -32, -32, -32, -32, -32, 0, -32, -32, -32, -32, -32, -32, -32, 121,

    0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1,
    0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1,
    0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1,
    0, EXC, 0, -1, 0, -1, 0, -1, 0, 0, -1, 0, -1, 0, -1, 0,

    -1, 0, -1, 0, -1, 0, -1, 0, -1, EXC, 0, -1, 0, -1, 0, -1,
    0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1,
    0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1, 0, -1,
    0, -1, 0, -1, 0, -1, 0, -1, 0, 0, -1, 0, -1, 0, -1, EXC
};

}  // namespace LatinCase

U_NAMESPACE_END

/** @return UCASE_NONE, UCASE_LOWER, UCASE_UPPER, UCASE_TITLE */
U_CAPI int32_t U_EXPORT2
ucase_getType(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    return UCASE_GET_TYPE(props);
}

/** @return same as ucase_getType() and set bit 2 if c is case-ignorable */
U_CAPI int32_t U_EXPORT2
ucase_getTypeOrIgnorable(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    return UCASE_GET_TYPE_AND_IGNORABLE(props);
}

/** @return UCASE_NO_DOT, UCASE_SOFT_DOTTED, UCASE_ABOVE, UCASE_OTHER_ACCENT */
static inline int32_t
getDotType(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    if(!UCASE_HAS_EXCEPTION(props)) {
        return props&UCASE_DOT_MASK;
    } else {
        const uint16_t *pe=GET_EXCEPTIONS(&ucase_props_singleton, props);
        return (*pe>>UCASE_EXC_DOT_SHIFT)&UCASE_DOT_MASK;
    }
}

U_CAPI UBool U_EXPORT2
ucase_isSoftDotted(UChar32 c) {
    return (UBool)(getDotType(c)==UCASE_SOFT_DOTTED);
}

U_CAPI UBool U_EXPORT2
ucase_isCaseSensitive(UChar32 c) {
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    if(!UCASE_HAS_EXCEPTION(props)) {
        return (UBool)((props&UCASE_SENSITIVE)!=0);
    } else {
        const uint16_t *pe=GET_EXCEPTIONS(&ucase_props_singleton, props);
        return (UBool)((*pe&UCASE_EXC_SENSITIVE)!=0);
    }
}

/* string casing ------------------------------------------------------------ */

/*
 * These internal functions form the core of string case mappings.
 * They map single code points to result code points or strings and take
 * all necessary conditions (context, locale ID, options) into account.
 *
 * They do not iterate over the source or write to the destination
 * so that the same functions are useful for non-standard string storage,
 * such as in a Replaceable (for Transliterator) or UTF-8/32 strings etc.
 * For the same reason, the "surrounding text" context is passed in as a
 * UCaseContextIterator which does not make any assumptions about
 * the underlying storage.
 *
 * This section contains helper functions that check for conditions
 * in the input text surrounding the current code point
 * according to SpecialCasing.txt.
 *
 * Each helper function gets the index
 * - after the current code point if it looks at following text
 * - before the current code point if it looks at preceding text
 *
 * Unicode 3.2 UAX 21 "Case Mappings" defines the conditions as follows:
 *
 * Final_Sigma
 *   C is preceded by a sequence consisting of
 *     a cased letter and a case-ignorable sequence,
 *   and C is not followed by a sequence consisting of
 *     an ignorable sequence and then a cased letter.
 *
 * More_Above
 *   C is followed by one or more characters of combining class 230 (ABOVE)
 *   in the combining character sequence.
 *
 * After_Soft_Dotted
 *   The last preceding character with combining class of zero before C
 *   was Soft_Dotted,
 *   and there is no intervening combining character class 230 (ABOVE).
 *
 * Before_Dot
 *   C is followed by combining dot above (U+0307).
 *   Any sequence of characters with a combining class that is neither 0 nor 230
 *   may intervene between the current character and the combining dot above.
 *
 * The erratum from 2002-10-31 adds the condition
 *
 * After_I
 *   The last preceding base character was an uppercase I, and there is no
 *   intervening combining character class 230 (ABOVE).
 *
 *   (See Jitterbug 2344 and the comments on After_I below.)
 *
 * Helper definitions in Unicode 3.2 UAX 21:
 *
 * D1. A character C is defined to be cased
 *     if it meets any of the following criteria:
 *
 *   - The general category of C is Titlecase Letter (Lt)
 *   - In [CoreProps], C has one of the properties Uppercase, or Lowercase
 *   - Given D = NFD(C), then it is not the case that:
 *     D = UCD_lower(D) = UCD_upper(D) = UCD_title(D)
 *     (This third criterium does not add any characters to the list
 *      for Unicode 3.2. Ignored.)
 *
 * D2. A character C is defined to be case-ignorable
 *     if it meets either of the following criteria:
 *
 *   - The general category of C is
 *     Nonspacing Mark (Mn), or Enclosing Mark (Me), or Format Control (Cf), or
 *     Letter Modifier (Lm), or Symbol Modifier (Sk)
 *   - C is one of the following characters
 *     U+0027 APOSTROPHE
 *     U+00AD SOFT HYPHEN (SHY)
 *     U+2019 RIGHT SINGLE QUOTATION MARK
 *            (the preferred character for apostrophe)
 *
 * D3. A case-ignorable sequence is a sequence of
 *     zero or more case-ignorable characters.
 */

#define is_d(c) ((c)=='d' || (c)=='D')
#define is_e(c) ((c)=='e' || (c)=='E')
#define is_i(c) ((c)=='i' || (c)=='I')
#define is_l(c) ((c)=='l' || (c)=='L')
#define is_r(c) ((c)=='r' || (c)=='R')
#define is_t(c) ((c)=='t' || (c)=='T')
#define is_u(c) ((c)=='u' || (c)=='U')
#define is_z(c) ((c)=='z' || (c)=='Z')

/* separator? */
#define is_sep(c) ((c)=='_' || (c)=='-' || (c)==0)

/**
 * Requires non-NULL locale ID but otherwise does the equivalent of
 * checking for language codes as if uloc_getLanguage() were called:
 * Accepts both 2- and 3-letter codes and accepts case variants.
 */
U_CFUNC int32_t
ucase_getCaseLocale(const char *locale) {
    /*
     * This function used to use uloc_getLanguage(), but the current code
     * removes the dependency of this low-level code on uloc implementation code
     * and is faster because not the whole locale ID has to be
     * examined and copied/transformed.
     *
     * Because this code does not want to depend on uloc, the caller must
     * pass in a non-NULL locale, i.e., may need to call uloc_getDefault().
     */
    char c=*locale++;
    // Fastpath for English "en" which is often used for default (=root locale) case mappings,
    // and for Chinese "zh": Very common but no special case mapping behavior.
    // Then check lowercase vs. uppercase to reduce the number of comparisons
    // for other locales without special behavior.
    if(c=='e') {
        /* el or ell? */
        c=*locale++;
        if(is_l(c)) {
            c=*locale++;
            if(is_l(c)) {
                c=*locale;
            }
            if(is_sep(c)) {
                return UCASE_LOC_GREEK;
            }
        }
        // en, es, ... -> root
    } else if(c=='z') {
        return UCASE_LOC_ROOT;
#if U_CHARSET_FAMILY==U_ASCII_FAMILY
    } else if(c>='a') {  // ASCII a-z = 0x61..0x7a, after A-Z
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
    } else if(c<='z') {  // EBCDIC a-z = 0x81..0xa9 with two gaps, before A-Z
#else
#   error Unknown charset family!
#endif
        // lowercase c
        if(c=='t') {
            /* tr or tur? */
            c=*locale++;
            if(is_u(c)) {
                c=*locale++;
            }
            if(is_r(c)) {
                c=*locale;
                if(is_sep(c)) {
                    return UCASE_LOC_TURKISH;
                }
            }
        } else if(c=='a') {
            /* az or aze? */
            c=*locale++;
            if(is_z(c)) {
                c=*locale++;
                if(is_e(c)) {
                    c=*locale;
                }
                if(is_sep(c)) {
                    return UCASE_LOC_TURKISH;
                }
            }
        } else if(c=='l') {
            /* lt or lit? */
            c=*locale++;
            if(is_i(c)) {
                c=*locale++;
            }
            if(is_t(c)) {
                c=*locale;
                if(is_sep(c)) {
                    return UCASE_LOC_LITHUANIAN;
                }
            }
        } else if(c=='n') {
            /* nl or nld? */
            c=*locale++;
            if(is_l(c)) {
                c=*locale++;
                if(is_d(c)) {
                    c=*locale;
                }
                if(is_sep(c)) {
                    return UCASE_LOC_DUTCH;
                }
            }
        }
    } else {
        // uppercase c
        // Same code as for lowercase c but also check for 'E'.
        if(c=='T') {
            /* tr or tur? */
            c=*locale++;
            if(is_u(c)) {
                c=*locale++;
            }
            if(is_r(c)) {
                c=*locale;
                if(is_sep(c)) {
                    return UCASE_LOC_TURKISH;
                }
            }
        } else if(c=='A') {
            /* az or aze? */
            c=*locale++;
            if(is_z(c)) {
                c=*locale++;
                if(is_e(c)) {
                    c=*locale;
                }
                if(is_sep(c)) {
                    return UCASE_LOC_TURKISH;
                }
            }
        } else if(c=='L') {
            /* lt or lit? */
            c=*locale++;
            if(is_i(c)) {
                c=*locale++;
            }
            if(is_t(c)) {
                c=*locale;
                if(is_sep(c)) {
                    return UCASE_LOC_LITHUANIAN;
                }
            }
        } else if(c=='E') {
            /* el or ell? */
            c=*locale++;
            if(is_l(c)) {
                c=*locale++;
                if(is_l(c)) {
                    c=*locale;
                }
                if(is_sep(c)) {
                    return UCASE_LOC_GREEK;
                }
            }
        } else if(c=='N') {
            /* nl or nld? */
            c=*locale++;
            if(is_l(c)) {
                c=*locale++;
                if(is_d(c)) {
                    c=*locale;
                }
                if(is_sep(c)) {
                    return UCASE_LOC_DUTCH;
                }
            }
        }
    }
    return UCASE_LOC_ROOT;
}

/*
 * Is followed by
 *   {case-ignorable}* cased
 * ?
 * (dir determines looking forward/backward)
 * If a character is case-ignorable, it is skipped regardless of whether
 * it is also cased or not.
 */
static UBool
isFollowedByCasedLetter(UCaseContextIterator *iter, void *context, int8_t dir) {
    UChar32 c;

    if(iter==NULL) {
        return FALSE;
    }

    for(/* dir!=0 sets direction */; (c=iter(context, dir))>=0; dir=0) {
        int32_t type=ucase_getTypeOrIgnorable(c);
        if(type&4) {
            /* case-ignorable, continue with the loop */
        } else if(type!=UCASE_NONE) {
            return TRUE; /* followed by cased letter */
        } else {
            return FALSE; /* uncased and not case-ignorable */
        }
    }

    return FALSE; /* not followed by cased letter */
}

/* Is preceded by Soft_Dotted character with no intervening cc=230 ? */
static UBool
isPrecededBySoftDotted(UCaseContextIterator *iter, void *context) {
    UChar32 c;
    int32_t dotType;
    int8_t dir;

    if(iter==NULL) {
        return FALSE;
    }

    for(dir=-1; (c=iter(context, dir))>=0; dir=0) {
        dotType=getDotType(c);
        if(dotType==UCASE_SOFT_DOTTED) {
            return TRUE; /* preceded by TYPE_i */
        } else if(dotType!=UCASE_OTHER_ACCENT) {
            return FALSE; /* preceded by different base character (not TYPE_i), or intervening cc==230 */
        }
    }

    return FALSE; /* not preceded by TYPE_i */
}

/*
 * See Jitterbug 2344:
 * The condition After_I for Turkic-lowercasing of U+0307 combining dot above
 * is checked in ICU 2.0, 2.1, 2.6 but was not in 2.2 & 2.4 because
 * we made those releases compatible with Unicode 3.2 which had not fixed
 * a related bug in SpecialCasing.txt.
 *
 * From the Jitterbug 2344 text:
 * ... this bug is listed as a Unicode erratum
 * from 2002-10-31 at http://www.unicode.org/uni2errata/UnicodeErrata.html
 * <quote>
 * There are two errors in SpecialCasing.txt.
 * 1. Missing semicolons on two lines. ... [irrelevant for ICU]
 * 2. An incorrect context definition. Correct as follows:
 * < 0307; ; 0307; 0307; tr After_Soft_Dotted; # COMBINING DOT ABOVE
 * < 0307; ; 0307; 0307; az After_Soft_Dotted; # COMBINING DOT ABOVE
 * ---
 * > 0307; ; 0307; 0307; tr After_I; # COMBINING DOT ABOVE
 * > 0307; ; 0307; 0307; az After_I; # COMBINING DOT ABOVE
 * where the context After_I is defined as:
 * The last preceding base character was an uppercase I, and there is no
 * intervening combining character class 230 (ABOVE).
 * </quote>
 *
 * Note that SpecialCasing.txt even in Unicode 3.2 described the condition as:
 *
 * # When lowercasing, remove dot_above in the sequence I + dot_above, which will turn into i.
 * # This matches the behavior of the canonically equivalent I-dot_above
 *
 * See also the description in this place in older versions of uchar.c (revision 1.100).
 *
 * Markus W. Scherer 2003-feb-15
 */

/* Is preceded by base character 'I' with no intervening cc=230 ? */
static UBool
isPrecededBy_I(UCaseContextIterator *iter, void *context) {
    UChar32 c;
    int32_t dotType;
    int8_t dir;

    if(iter==NULL) {
        return FALSE;
    }

    for(dir=-1; (c=iter(context, dir))>=0; dir=0) {
        if(c==0x49) {
            return TRUE; /* preceded by I */
        }
        dotType=getDotType(c);
        if(dotType!=UCASE_OTHER_ACCENT) {
            return FALSE; /* preceded by different base character (not I), or intervening cc==230 */
        }
    }

    return FALSE; /* not preceded by I */
}

/* Is followed by one or more cc==230 ? */
static UBool
isFollowedByMoreAbove(UCaseContextIterator *iter, void *context) {
    UChar32 c;
    int32_t dotType;
    int8_t dir;

    if(iter==NULL) {
        return FALSE;
    }

    for(dir=1; (c=iter(context, dir))>=0; dir=0) {
        dotType=getDotType(c);
        if(dotType==UCASE_ABOVE) {
            return TRUE; /* at least one cc==230 following */
        } else if(dotType!=UCASE_OTHER_ACCENT) {
            return FALSE; /* next base character, no more cc==230 following */
        }
    }

    return FALSE; /* no more cc==230 following */
}

/* Is followed by a dot above (without cc==230 in between) ? */
static UBool
isFollowedByDotAbove(UCaseContextIterator *iter, void *context) {
    UChar32 c;
    int32_t dotType;
    int8_t dir;

    if(iter==NULL) {
        return FALSE;
    }

    for(dir=1; (c=iter(context, dir))>=0; dir=0) {
        if(c==0x307) {
            return TRUE;
        }
        dotType=getDotType(c);
        if(dotType!=UCASE_OTHER_ACCENT) {
            return FALSE; /* next base character or cc==230 in between */
        }
    }

    return FALSE; /* no dot above following */
}

U_CAPI int32_t U_EXPORT2
ucase_toFullLower(UChar32 c,
                  UCaseContextIterator *iter, void *context,
                  const UChar **pString,
                  int32_t loc) {
    // The sign of the result has meaning, input must be non-negative so that it can be returned as is.
    U_ASSERT(c >= 0);
    UChar32 result=c;
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    if(!UCASE_HAS_EXCEPTION(props)) {
        if(UCASE_IS_UPPER_OR_TITLE(props)) {
            result=c+UCASE_GET_DELTA(props);
        }
    } else {
        const uint16_t *pe=GET_EXCEPTIONS(&ucase_props_singleton, props), *pe2;
        uint16_t excWord=*pe++;
        int32_t full;

        pe2=pe;

        if(excWord&UCASE_EXC_CONDITIONAL_SPECIAL) {
            /* use hardcoded conditions and mappings */

            /*
             * Test for conditional mappings first
             *   (otherwise the unconditional default mappings are always taken),
             * then test for characters that have unconditional mappings in SpecialCasing.txt,
             * then get the UnicodeData.txt mappings.
             */
            if( loc==UCASE_LOC_LITHUANIAN &&
                    /* base characters, find accents above */
                    (((c==0x49 || c==0x4a || c==0x12e) &&
                        isFollowedByMoreAbove(iter, context)) ||
                    /* precomposed with accent above, no need to find one */
                    (c==0xcc || c==0xcd || c==0x128))
            ) {
                /*
                    # Lithuanian

                    # Lithuanian retains the dot in a lowercase i when followed by accents.

                    # Introduce an explicit dot above when lowercasing capital I's and J's
                    # whenever there are more accents above.
                    # (of the accents used in Lithuanian: grave, acute, tilde above, and ogonek)

                    0049; 0069 0307; 0049; 0049; lt More_Above; # LATIN CAPITAL LETTER I
                    004A; 006A 0307; 004A; 004A; lt More_Above; # LATIN CAPITAL LETTER J
                    012E; 012F 0307; 012E; 012E; lt More_Above; # LATIN CAPITAL LETTER I WITH OGONEK
                    00CC; 0069 0307 0300; 00CC; 00CC; lt; # LATIN CAPITAL LETTER I WITH GRAVE
                    00CD; 0069 0307 0301; 00CD; 00CD; lt; # LATIN CAPITAL LETTER I WITH ACUTE
                    0128; 0069 0307 0303; 0128; 0128; lt; # LATIN CAPITAL LETTER I WITH TILDE
                 */
                switch(c) {
                case 0x49:  /* LATIN CAPITAL LETTER I */
                    *pString=iDot;
                    return 2;
                case 0x4a:  /* LATIN CAPITAL LETTER J */
                    *pString=jDot;
                    return 2;
                case 0x12e: /* LATIN CAPITAL LETTER I WITH OGONEK */
                    *pString=iOgonekDot;
                    return 2;
                case 0xcc:  /* LATIN CAPITAL LETTER I WITH GRAVE */
                    *pString=iDotGrave;
                    return 3;
                case 0xcd:  /* LATIN CAPITAL LETTER I WITH ACUTE */
                    *pString=iDotAcute;
                    return 3;
                case 0x128: /* LATIN CAPITAL LETTER I WITH TILDE */
                    *pString=iDotTilde;
                    return 3;
                default:
                    return 0; /* will not occur */
                }
            /* # Turkish and Azeri */
            } else if(loc==UCASE_LOC_TURKISH && c==0x130) {
                /*
                    # I and i-dotless; I-dot and i are case pairs in Turkish and Azeri
                    # The following rules handle those cases.

                    0130; 0069; 0130; 0130; tr # LATIN CAPITAL LETTER I WITH DOT ABOVE
                    0130; 0069; 0130; 0130; az # LATIN CAPITAL LETTER I WITH DOT ABOVE
                 */
                return 0x69;
            } else if(loc==UCASE_LOC_TURKISH && c==0x307 && isPrecededBy_I(iter, context)) {
                /*
                    # When lowercasing, remove dot_above in the sequence I + dot_above, which will turn into i.
                    # This matches the behavior of the canonically equivalent I-dot_above

                    0307; ; 0307; 0307; tr After_I; # COMBINING DOT ABOVE
                    0307; ; 0307; 0307; az After_I; # COMBINING DOT ABOVE
                 */
                *pString=nullptr;
                return 0; /* remove the dot (continue without output) */
            } else if(loc==UCASE_LOC_TURKISH && c==0x49 && !isFollowedByDotAbove(iter, context)) {
                /*
                    # When lowercasing, unless an I is before a dot_above, it turns into a dotless i.

                    0049; 0131; 0049; 0049; tr Not_Before_Dot; # LATIN CAPITAL LETTER I
                    0049; 0131; 0049; 0049; az Not_Before_Dot; # LATIN CAPITAL LETTER I
                 */
                return 0x131;
            } else if(c==0x130) {
                /*
                    # Preserve canonical equivalence for I with dot. Turkic is handled below.

                    0130; 0069 0307; 0130; 0130; # LATIN CAPITAL LETTER I WITH DOT ABOVE
                 */
                *pString=iDot;
                return 2;
            } else if(  c==0x3a3 &&
                        !isFollowedByCasedLetter(iter, context, 1) &&
                        isFollowedByCasedLetter(iter, context, -1) /* -1=preceded */
            ) {
                /* greek capital sigma maps depending on surrounding cased letters (see SpecialCasing.txt) */
                /*
                    # Special case for final form of sigma

                    03A3; 03C2; 03A3; 03A3; Final_Sigma; # GREEK CAPITAL LETTER SIGMA
                 */
                return 0x3c2; /* greek small final sigma */
            } else {
                /* no known conditional special case mapping, use a normal mapping */
            }
        } else if(HAS_SLOT(excWord, UCASE_EXC_FULL_MAPPINGS)) {
            GET_SLOT_VALUE(excWord, UCASE_EXC_FULL_MAPPINGS, pe, full);
            full&=UCASE_FULL_LOWER;
            if(full!=0) {
                /* set the output pointer to the lowercase mapping */
                *pString=reinterpret_cast<const UChar *>(pe+1);

                /* return the string length */
                return full;
            }
        }

        if(HAS_SLOT(excWord, UCASE_EXC_DELTA) && UCASE_IS_UPPER_OR_TITLE(props)) {
            int32_t delta;
            GET_SLOT_VALUE(excWord, UCASE_EXC_DELTA, pe2, delta);
            return (excWord&UCASE_EXC_DELTA_IS_NEGATIVE)==0 ? c+delta : c-delta;
        }
        if(HAS_SLOT(excWord, UCASE_EXC_LOWER)) {
            GET_SLOT_VALUE(excWord, UCASE_EXC_LOWER, pe2, result);
        }
    }

    return (result==c) ? ~result : result;
}

/* internal */
static int32_t
toUpperOrTitle(UChar32 c,
               UCaseContextIterator *iter, void *context,
               const UChar **pString,
               int32_t loc,
               UBool upperNotTitle) {
    // The sign of the result has meaning, input must be non-negative so that it can be returned as is.
    U_ASSERT(c >= 0);
    UChar32 result=c;
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    if(!UCASE_HAS_EXCEPTION(props)) {
        if(UCASE_GET_TYPE(props)==UCASE_LOWER) {
            result=c+UCASE_GET_DELTA(props);
        }
    } else {
        const uint16_t *pe=GET_EXCEPTIONS(&ucase_props_singleton, props), *pe2;
        uint16_t excWord=*pe++;
        int32_t full, idx;

        pe2=pe;

        if(excWord&UCASE_EXC_CONDITIONAL_SPECIAL) {
            /* use hardcoded conditions and mappings */
            if(loc==UCASE_LOC_TURKISH && c==0x69) {
                /*
                    # Turkish and Azeri

                    # I and i-dotless; I-dot and i are case pairs in Turkish and Azeri
                    # The following rules handle those cases.

                    # When uppercasing, i turns into a dotted capital I

                    0069; 0069; 0130; 0130; tr; # LATIN SMALL LETTER I
                    0069; 0069; 0130; 0130; az; # LATIN SMALL LETTER I
                */
                return 0x130;
            } else if(loc==UCASE_LOC_LITHUANIAN && c==0x307 && isPrecededBySoftDotted(iter, context)) {
                /*
                    # Lithuanian

                    # Lithuanian retains the dot in a lowercase i when followed by accents.

                    # Remove DOT ABOVE after "i" with upper or titlecase

                    0307; 0307; ; ; lt After_Soft_Dotted; # COMBINING DOT ABOVE
                 */
                *pString=nullptr;
                return 0; /* remove the dot (continue without output) */
            } else {
                /* no known conditional special case mapping, use a normal mapping */
            }
        } else if(HAS_SLOT(excWord, UCASE_EXC_FULL_MAPPINGS)) {
            GET_SLOT_VALUE(excWord, UCASE_EXC_FULL_MAPPINGS, pe, full);

            /* start of full case mapping strings */
            ++pe;

            /* skip the lowercase and case-folding result strings */
            pe+=full&UCASE_FULL_LOWER;
            full>>=4;
            pe+=full&0xf;
            full>>=4;

            if(upperNotTitle) {
                full&=0xf;
            } else {
                /* skip the uppercase result string */
                pe+=full&0xf;
                full=(full>>4)&0xf;
            }

            if(full!=0) {
                /* set the output pointer to the result string */
                *pString=reinterpret_cast<const UChar *>(pe);

                /* return the string length */
                return full;
            }
        }

        if(HAS_SLOT(excWord, UCASE_EXC_DELTA) && UCASE_GET_TYPE(props)==UCASE_LOWER) {
            int32_t delta;
            GET_SLOT_VALUE(excWord, UCASE_EXC_DELTA, pe2, delta);
            return (excWord&UCASE_EXC_DELTA_IS_NEGATIVE)==0 ? c+delta : c-delta;
        }
        if(!upperNotTitle && HAS_SLOT(excWord, UCASE_EXC_TITLE)) {
            idx=UCASE_EXC_TITLE;
        } else if(HAS_SLOT(excWord, UCASE_EXC_UPPER)) {
            /* here, titlecase is same as uppercase */
            idx=UCASE_EXC_UPPER;
        } else {
            return ~c;
        }
        GET_SLOT_VALUE(excWord, idx, pe2, result);
    }

    return (result==c) ? ~result : result;
}

U_CAPI int32_t U_EXPORT2
ucase_toFullUpper(UChar32 c,
                  UCaseContextIterator *iter, void *context,
                  const UChar **pString,
                  int32_t caseLocale) {
    return toUpperOrTitle(c, iter, context, pString, caseLocale, TRUE);
}

U_CAPI int32_t U_EXPORT2
ucase_toFullTitle(UChar32 c,
                  UCaseContextIterator *iter, void *context,
                  const UChar **pString,
                  int32_t caseLocale) {
    return toUpperOrTitle(c, iter, context, pString, caseLocale, FALSE);
}

/* case folding ------------------------------------------------------------- */

/*
 * Case folding is similar to lowercasing.
 * The result may be a simple mapping, i.e., a single code point, or
 * a full mapping, i.e., a string.
 * If the case folding for a code point is the same as its simple (1:1) lowercase mapping,
 * then only the lowercase mapping is stored.
 *
 * Some special cases are hardcoded because their conditions cannot be
 * parsed and processed from CaseFolding.txt.
 *
 * Unicode 3.2 CaseFolding.txt specifies for its status field:

# C: common case folding, common mappings shared by both simple and full mappings.
# F: full case folding, mappings that cause strings to grow in length. Multiple characters are separated by spaces.
# S: simple case folding, mappings to single characters where different from F.
# T: special case for uppercase I and dotted uppercase I
#    - For non-Turkic languages, this mapping is normally not used.
#    - For Turkic languages (tr, az), this mapping can be used instead of the normal mapping for these characters.
#
# Usage:
#  A. To do a simple case folding, use the mappings with status C + S.
#  B. To do a full case folding, use the mappings with status C + F.
#
#    The mappings with status T can be used or omitted depending on the desired case-folding
#    behavior. (The default option is to exclude them.)

 * Unicode 3.2 has 'T' mappings as follows:

0049; T; 0131; # LATIN CAPITAL LETTER I
0130; T; 0069; # LATIN CAPITAL LETTER I WITH DOT ABOVE

 * while the default mappings for these code points are:

0049; C; 0069; # LATIN CAPITAL LETTER I
0130; F; 0069 0307; # LATIN CAPITAL LETTER I WITH DOT ABOVE

 * U+0130 has no simple case folding (simple-case-folds to itself).
 */

/* return the simple case folding mapping for c */
U_CAPI UChar32 U_EXPORT2
ucase_fold(UChar32 c, uint32_t options) {
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    if(!UCASE_HAS_EXCEPTION(props)) {
        if(UCASE_IS_UPPER_OR_TITLE(props)) {
            c+=UCASE_GET_DELTA(props);
        }
    } else {
        const uint16_t *pe=GET_EXCEPTIONS(&ucase_props_singleton, props);
        uint16_t excWord=*pe++;
        int32_t idx;
        if(excWord&UCASE_EXC_CONDITIONAL_FOLD) {
            /* special case folding mappings, hardcoded */
            if((options&_FOLD_CASE_OPTIONS_MASK)==U_FOLD_CASE_DEFAULT) {
                /* default mappings */
                if(c==0x49) {
                    /* 0049; C; 0069; # LATIN CAPITAL LETTER I */
                    return 0x69;
                } else if(c==0x130) {
                    /* no simple case folding for U+0130 */
                    return c;
                }
            } else {
                /* Turkic mappings */
                if(c==0x49) {
                    /* 0049; T; 0131; # LATIN CAPITAL LETTER I */
                    return 0x131;
                } else if(c==0x130) {
                    /* 0130; T; 0069; # LATIN CAPITAL LETTER I WITH DOT ABOVE */
                    return 0x69;
                }
            }
        }
        if((excWord&UCASE_EXC_NO_SIMPLE_CASE_FOLDING)!=0) {
            return c;
        }
        if(HAS_SLOT(excWord, UCASE_EXC_DELTA) && UCASE_IS_UPPER_OR_TITLE(props)) {
            int32_t delta;
            GET_SLOT_VALUE(excWord, UCASE_EXC_DELTA, pe, delta);
            return (excWord&UCASE_EXC_DELTA_IS_NEGATIVE)==0 ? c+delta : c-delta;
        }
        if(HAS_SLOT(excWord, UCASE_EXC_FOLD)) {
            idx=UCASE_EXC_FOLD;
        } else if(HAS_SLOT(excWord, UCASE_EXC_LOWER)) {
            idx=UCASE_EXC_LOWER;
        } else {
            return c;
        }
        GET_SLOT_VALUE(excWord, idx, pe, c);
    }
    return c;
}

/*
 * Issue for canonical caseless match (UAX #21):
 * Turkic casefolding (using "T" mappings in CaseFolding.txt) does not preserve
 * canonical equivalence, unlike default-option casefolding.
 * For example, I-grave and I + grave fold to strings that are not canonically
 * equivalent.
 * For more details, see the comment in unorm_compare() in unorm.cpp
 * and the intermediate prototype changes for Jitterbug 2021.
 * (For example, revision 1.104 of uchar.c and 1.4 of CaseFolding.txt.)
 *
 * This did not get fixed because it appears that it is not possible to fix
 * it for uppercase and lowercase characters (I-grave vs. i-grave)
 * together in a way that they still fold to common result strings.
 */

U_CAPI int32_t U_EXPORT2
ucase_toFullFolding(UChar32 c,
                    const UChar **pString,
                    uint32_t options) {
    // The sign of the result has meaning, input must be non-negative so that it can be returned as is.
    U_ASSERT(c >= 0);
    UChar32 result=c;
    uint16_t props=UTRIE2_GET16(&ucase_props_singleton.trie, c);
    if(!UCASE_HAS_EXCEPTION(props)) {
        if(UCASE_IS_UPPER_OR_TITLE(props)) {
            result=c+UCASE_GET_DELTA(props);
        }
    } else {
        const uint16_t *pe=GET_EXCEPTIONS(&ucase_props_singleton, props), *pe2;
        uint16_t excWord=*pe++;
        int32_t full, idx;

        pe2=pe;

        if(excWord&UCASE_EXC_CONDITIONAL_FOLD) {
            /* use hardcoded conditions and mappings */
            if((options&_FOLD_CASE_OPTIONS_MASK)==U_FOLD_CASE_DEFAULT) {
                /* default mappings */
                if(c==0x49) {
                    /* 0049; C; 0069; # LATIN CAPITAL LETTER I */
                    return 0x69;
                } else if(c==0x130) {
                    /* 0130; F; 0069 0307; # LATIN CAPITAL LETTER I WITH DOT ABOVE */
                    *pString=iDot;
                    return 2;
                }
            } else {
                /* Turkic mappings */
                if(c==0x49) {
                    /* 0049; T; 0131; # LATIN CAPITAL LETTER I */
                    return 0x131;
                } else if(c==0x130) {
                    /* 0130; T; 0069; # LATIN CAPITAL LETTER I WITH DOT ABOVE */
                    return 0x69;
                }
            }
        } else if(HAS_SLOT(excWord, UCASE_EXC_FULL_MAPPINGS)) {
            GET_SLOT_VALUE(excWord, UCASE_EXC_FULL_MAPPINGS, pe, full);

            /* start of full case mapping strings */
            ++pe;

            /* skip the lowercase result string */
            pe+=full&UCASE_FULL_LOWER;
            full=(full>>4)&0xf;

            if(full!=0) {
                /* set the output pointer to the result string */
                *pString=reinterpret_cast<const UChar *>(pe);

                /* return the string length */
                return full;
            }
        }

        if((excWord&UCASE_EXC_NO_SIMPLE_CASE_FOLDING)!=0) {
            return ~c;
        }
        if(HAS_SLOT(excWord, UCASE_EXC_DELTA) && UCASE_IS_UPPER_OR_TITLE(props)) {
            int32_t delta;
            GET_SLOT_VALUE(excWord, UCASE_EXC_DELTA, pe2, delta);
            return (excWord&UCASE_EXC_DELTA_IS_NEGATIVE)==0 ? c+delta : c-delta;
        }
        if(HAS_SLOT(excWord, UCASE_EXC_FOLD)) {
            idx=UCASE_EXC_FOLD;
        } else if(HAS_SLOT(excWord, UCASE_EXC_LOWER)) {
            idx=UCASE_EXC_LOWER;
        } else {
            return ~c;
        }
        GET_SLOT_VALUE(excWord, idx, pe2, result);
    }

    return (result==c) ? ~result : result;
}

/* case mapping properties API ---------------------------------------------- */

/* public API (see uchar.h) */

U_CAPI UBool U_EXPORT2
u_isULowercase(UChar32 c) {
    return (UBool)(UCASE_LOWER==ucase_getType(c));
}

U_CAPI UBool U_EXPORT2
u_isUUppercase(UChar32 c) {
    return (UBool)(UCASE_UPPER==ucase_getType(c));
}

/* Transforms the Unicode character to its lower case equivalent.*/
U_CAPI UChar32 U_EXPORT2
u_tolower(UChar32 c) {
    return ucase_tolower(c);
}

/* Transforms the Unicode character to its upper case equivalent.*/
U_CAPI UChar32 U_EXPORT2
u_toupper(UChar32 c) {
    return ucase_toupper(c);
}

/* Transforms the Unicode character to its title case equivalent.*/
U_CAPI UChar32 U_EXPORT2
u_totitle(UChar32 c) {
    return ucase_totitle(c);
}

/* return the simple case folding mapping for c */
U_CAPI UChar32 U_EXPORT2
u_foldCase(UChar32 c, uint32_t options) {
    return ucase_fold(c, options);
}

U_CFUNC int32_t U_EXPORT2
ucase_hasBinaryProperty(UChar32 c, UProperty which) {
    /* case mapping properties */
    const UChar *resultString;
    switch(which) {
    case UCHAR_LOWERCASE:
        return (UBool)(UCASE_LOWER==ucase_getType(c));
    case UCHAR_UPPERCASE:
        return (UBool)(UCASE_UPPER==ucase_getType(c));
    case UCHAR_SOFT_DOTTED:
        return ucase_isSoftDotted(c);
    case UCHAR_CASE_SENSITIVE:
        return ucase_isCaseSensitive(c);
    case UCHAR_CASED:
        return (UBool)(UCASE_NONE!=ucase_getType(c));
    case UCHAR_CASE_IGNORABLE:
        return (UBool)(ucase_getTypeOrIgnorable(c)>>2);
    /*
     * Note: The following Changes_When_Xyz are defined as testing whether
     * the NFD form of the input changes when Xyz-case-mapped.
     * However, this simpler implementation of these properties,
     * ignoring NFD, passes the tests.
     * The implementation needs to be changed if the tests start failing.
     * When that happens, optimizations should be used to work with the
     * per-single-code point ucase_toFullXyz() functions unless
     * the NFD form has more than one code point,
     * and the property starts set needs to be the union of the
     * start sets for normalization and case mappings.
     */
    case UCHAR_CHANGES_WHEN_LOWERCASED:
        return (UBool)(ucase_toFullLower(c, NULL, NULL, &resultString, UCASE_LOC_ROOT)>=0);
    case UCHAR_CHANGES_WHEN_UPPERCASED:
        return (UBool)(ucase_toFullUpper(c, NULL, NULL, &resultString, UCASE_LOC_ROOT)>=0);
    case UCHAR_CHANGES_WHEN_TITLECASED:
        return (UBool)(ucase_toFullTitle(c, NULL, NULL, &resultString, UCASE_LOC_ROOT)>=0);
    /* case UCHAR_CHANGES_WHEN_CASEFOLDED: -- in uprops.c */
    case UCHAR_CHANGES_WHEN_CASEMAPPED:
        return (UBool)(
            ucase_toFullLower(c, NULL, NULL, &resultString, UCASE_LOC_ROOT)>=0 ||
            ucase_toFullUpper(c, NULL, NULL, &resultString, UCASE_LOC_ROOT)>=0 ||
            ucase_toFullTitle(c, NULL, NULL, &resultString, UCASE_LOC_ROOT)>=0);
    default:
        return FALSE;
    }
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2005-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ucasemap.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2005may06
*   created by: Markus W. Scherer
*
*   Case mapping service object and functions using it.
*/




// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// casemap.h
// created: 2017jan12 Markus W. Scherer




#if U_SHOW_CPLUSPLUS_API




/**
 * \file
 * \brief C++ API: Low-level C++ case mapping functions.
 */

U_NAMESPACE_BEGIN

class BreakIterator;
class ByteSink;
class Edits;

/**
 * Low-level C++ case mapping functions.
 *
 * @stable ICU 59
 */
class U_COMMON_API CaseMap U_FINAL : public UMemory {
public:
    /**
     * Lowercases a UTF-16 string and optionally records edits.
     * Casing is locale-dependent and context-sensitive.
     * The result may be longer or shorter than the original.
     * The source string and the destination buffer must not overlap.
     *
     * @param locale    The locale ID. ("" = root locale, NULL = default locale.)
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT and U_EDITS_NO_RESET.
     * @param src       The original string.
     * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
     * @param dest      A buffer for the result string. The result will be NUL-terminated if
     *                  the buffer is large enough.
     *                  The contents is undefined in case of failure.
     * @param destCapacity The size of the buffer (number of char16_ts). If it is 0, then
     *                  dest may be NULL and the function will only return the length of the result
     *                  without writing any of the result string.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     * @return The length of the result string, if successful.
     *         When the result would be longer than destCapacity,
     *         the full length is returned and a U_BUFFER_OVERFLOW_ERROR is set.
     *
     * @see u_strToLower
     * @stable ICU 59
     */
     static int32_t toLower(
            const char *locale, uint32_t options,
            const char16_t *src, int32_t srcLength,
            char16_t *dest, int32_t destCapacity, Edits *edits,
            UErrorCode &errorCode);

    /**
     * Uppercases a UTF-16 string and optionally records edits.
     * Casing is locale-dependent and context-sensitive.
     * The result may be longer or shorter than the original.
     * The source string and the destination buffer must not overlap.
     *
     * @param locale    The locale ID. ("" = root locale, NULL = default locale.)
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT and U_EDITS_NO_RESET.
     * @param src       The original string.
     * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
     * @param dest      A buffer for the result string. The result will be NUL-terminated if
     *                  the buffer is large enough.
     *                  The contents is undefined in case of failure.
     * @param destCapacity The size of the buffer (number of char16_ts). If it is 0, then
     *                  dest may be NULL and the function will only return the length of the result
     *                  without writing any of the result string.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     * @return The length of the result string, if successful.
     *         When the result would be longer than destCapacity,
     *         the full length is returned and a U_BUFFER_OVERFLOW_ERROR is set.
     *
     * @see u_strToUpper
     * @stable ICU 59
     */
    static int32_t toUpper(
            const char *locale, uint32_t options,
            const char16_t *src, int32_t srcLength,
            char16_t *dest, int32_t destCapacity, Edits *edits,
            UErrorCode &errorCode);

#if !UCONFIG_NO_BREAK_ITERATION

    /**
     * Titlecases a UTF-16 string and optionally records edits.
     * Casing is locale-dependent and context-sensitive.
     * The result may be longer or shorter than the original.
     * The source string and the destination buffer must not overlap.
     *
     * Titlecasing uses a break iterator to find the first characters of words
     * that are to be titlecased. It titlecases those characters and lowercases
     * all others. (This can be modified with options bits.)
     *
     * @param locale    The locale ID. ("" = root locale, NULL = default locale.)
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT, U_EDITS_NO_RESET,
     *                  U_TITLECASE_NO_LOWERCASE,
     *                  U_TITLECASE_NO_BREAK_ADJUSTMENT, U_TITLECASE_ADJUST_TO_CASED,
     *                  U_TITLECASE_WHOLE_STRING, U_TITLECASE_SENTENCES.
     * @param iter      A break iterator to find the first characters of words that are to be titlecased.
     *                  It is set to the source string (setText())
     *                  and used one or more times for iteration (first() and next()).
     *                  If NULL, then a word break iterator for the locale is used
     *                  (or something equivalent).
     * @param src       The original string.
     * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
     * @param dest      A buffer for the result string. The result will be NUL-terminated if
     *                  the buffer is large enough.
     *                  The contents is undefined in case of failure.
     * @param destCapacity The size of the buffer (number of char16_ts). If it is 0, then
     *                  dest may be NULL and the function will only return the length of the result
     *                  without writing any of the result string.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     * @return The length of the result string, if successful.
     *         When the result would be longer than destCapacity,
     *         the full length is returned and a U_BUFFER_OVERFLOW_ERROR is set.
     *
     * @see u_strToTitle
     * @see ucasemap_toTitle
     * @stable ICU 59
     */
    static int32_t toTitle(
            const char *locale, uint32_t options, BreakIterator *iter,
            const char16_t *src, int32_t srcLength,
            char16_t *dest, int32_t destCapacity, Edits *edits,
            UErrorCode &errorCode);

#endif  // UCONFIG_NO_BREAK_ITERATION

    /**
     * Case-folds a UTF-16 string and optionally records edits.
     *
     * Case folding is locale-independent and not context-sensitive,
     * but there is an option for whether to include or exclude mappings for dotted I
     * and dotless i that are marked with 'T' in CaseFolding.txt.
     *
     * The result may be longer or shorter than the original.
     * The source string and the destination buffer must not overlap.
     *
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT, U_EDITS_NO_RESET,
     *                  U_FOLD_CASE_DEFAULT, U_FOLD_CASE_EXCLUDE_SPECIAL_I.
     * @param src       The original string.
     * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
     * @param dest      A buffer for the result string. The result will be NUL-terminated if
     *                  the buffer is large enough.
     *                  The contents is undefined in case of failure.
     * @param destCapacity The size of the buffer (number of char16_ts). If it is 0, then
     *                  dest may be NULL and the function will only return the length of the result
     *                  without writing any of the result string.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     * @return The length of the result string, if successful.
     *         When the result would be longer than destCapacity,
     *         the full length is returned and a U_BUFFER_OVERFLOW_ERROR is set.
     *
     * @see u_strFoldCase
     * @stable ICU 59
     */
    static int32_t fold(
            uint32_t options,
            const char16_t *src, int32_t srcLength,
            char16_t *dest, int32_t destCapacity, Edits *edits,
            UErrorCode &errorCode);

    /**
     * Lowercases a UTF-8 string and optionally records edits.
     * Casing is locale-dependent and context-sensitive.
     * The result may be longer or shorter than the original.
     *
     * @param locale    The locale ID. ("" = root locale, NULL = default locale.)
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT and U_EDITS_NO_RESET.
     * @param src       The original string.
     * @param sink      A ByteSink to which the result string is written.
     *                  sink.Flush() is called at the end.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     *
     * @see ucasemap_utf8ToLower
     * @stable ICU 60
     */
    static void utf8ToLower(
            const char *locale, uint32_t options,
            StringPiece src, ByteSink &sink, Edits *edits,
            UErrorCode &errorCode);

    /**
     * Uppercases a UTF-8 string and optionally records edits.
     * Casing is locale-dependent and context-sensitive.
     * The result may be longer or shorter than the original.
     *
     * @param locale    The locale ID. ("" = root locale, NULL = default locale.)
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT and U_EDITS_NO_RESET.
     * @param src       The original string.
     * @param sink      A ByteSink to which the result string is written.
     *                  sink.Flush() is called at the end.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     *
     * @see ucasemap_utf8ToUpper
     * @stable ICU 60
     */
    static void utf8ToUpper(
            const char *locale, uint32_t options,
            StringPiece src, ByteSink &sink, Edits *edits,
            UErrorCode &errorCode);

#if !UCONFIG_NO_BREAK_ITERATION

    /**
     * Titlecases a UTF-8 string and optionally records edits.
     * Casing is locale-dependent and context-sensitive.
     * The result may be longer or shorter than the original.
     *
     * Titlecasing uses a break iterator to find the first characters of words
     * that are to be titlecased. It titlecases those characters and lowercases
     * all others. (This can be modified with options bits.)
     *
     * @param locale    The locale ID. ("" = root locale, NULL = default locale.)
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT, U_EDITS_NO_RESET,
     *                  U_TITLECASE_NO_LOWERCASE,
     *                  U_TITLECASE_NO_BREAK_ADJUSTMENT, U_TITLECASE_ADJUST_TO_CASED,
     *                  U_TITLECASE_WHOLE_STRING, U_TITLECASE_SENTENCES.
     * @param iter      A break iterator to find the first characters of words that are to be titlecased.
     *                  It is set to the source string (setUText())
     *                  and used one or more times for iteration (first() and next()).
     *                  If NULL, then a word break iterator for the locale is used
     *                  (or something equivalent).
     * @param src       The original string.
     * @param sink      A ByteSink to which the result string is written.
     *                  sink.Flush() is called at the end.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     *
     * @see ucasemap_utf8ToTitle
     * @stable ICU 60
     */
    static void utf8ToTitle(
            const char *locale, uint32_t options, BreakIterator *iter,
            StringPiece src, ByteSink &sink, Edits *edits,
            UErrorCode &errorCode);

#endif  // UCONFIG_NO_BREAK_ITERATION

    /**
     * Case-folds a UTF-8 string and optionally records edits.
     *
     * Case folding is locale-independent and not context-sensitive,
     * but there is an option for whether to include or exclude mappings for dotted I
     * and dotless i that are marked with 'T' in CaseFolding.txt.
     *
     * The result may be longer or shorter than the original.
     *
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT and U_EDITS_NO_RESET.
     * @param src       The original string.
     * @param sink      A ByteSink to which the result string is written.
     *                  sink.Flush() is called at the end.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     *
     * @see ucasemap_utf8FoldCase
     * @stable ICU 60
     */
    static void utf8Fold(
            uint32_t options,
            StringPiece src, ByteSink &sink, Edits *edits,
            UErrorCode &errorCode);

    /**
     * Lowercases a UTF-8 string and optionally records edits.
     * Casing is locale-dependent and context-sensitive.
     * The result may be longer or shorter than the original.
     * The source string and the destination buffer must not overlap.
     *
     * @param locale    The locale ID. ("" = root locale, NULL = default locale.)
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT and U_EDITS_NO_RESET.
     * @param src       The original string.
     * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
     * @param dest      A buffer for the result string. The result will be NUL-terminated if
     *                  the buffer is large enough.
     *                  The contents is undefined in case of failure.
     * @param destCapacity The size of the buffer (number of bytes). If it is 0, then
     *                  dest may be NULL and the function will only return the length of the result
     *                  without writing any of the result string.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     * @return The length of the result string, if successful.
     *         When the result would be longer than destCapacity,
     *         the full length is returned and a U_BUFFER_OVERFLOW_ERROR is set.
     *
     * @see ucasemap_utf8ToLower
     * @stable ICU 59
     */
    static int32_t utf8ToLower(
            const char *locale, uint32_t options,
            const char *src, int32_t srcLength,
            char *dest, int32_t destCapacity, Edits *edits,
            UErrorCode &errorCode);

    /**
     * Uppercases a UTF-8 string and optionally records edits.
     * Casing is locale-dependent and context-sensitive.
     * The result may be longer or shorter than the original.
     * The source string and the destination buffer must not overlap.
     *
     * @param locale    The locale ID. ("" = root locale, NULL = default locale.)
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT and U_EDITS_NO_RESET.
     * @param src       The original string.
     * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
     * @param dest      A buffer for the result string. The result will be NUL-terminated if
     *                  the buffer is large enough.
     *                  The contents is undefined in case of failure.
     * @param destCapacity The size of the buffer (number of bytes). If it is 0, then
     *                  dest may be NULL and the function will only return the length of the result
     *                  without writing any of the result string.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     * @return The length of the result string, if successful.
     *         When the result would be longer than destCapacity,
     *         the full length is returned and a U_BUFFER_OVERFLOW_ERROR is set.
     *
     * @see ucasemap_utf8ToUpper
     * @stable ICU 59
     */
    static int32_t utf8ToUpper(
            const char *locale, uint32_t options,
            const char *src, int32_t srcLength,
            char *dest, int32_t destCapacity, Edits *edits,
            UErrorCode &errorCode);

#if !UCONFIG_NO_BREAK_ITERATION

    /**
     * Titlecases a UTF-8 string and optionally records edits.
     * Casing is locale-dependent and context-sensitive.
     * The result may be longer or shorter than the original.
     * The source string and the destination buffer must not overlap.
     *
     * Titlecasing uses a break iterator to find the first characters of words
     * that are to be titlecased. It titlecases those characters and lowercases
     * all others. (This can be modified with options bits.)
     *
     * @param locale    The locale ID. ("" = root locale, NULL = default locale.)
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT, U_EDITS_NO_RESET,
     *                  U_TITLECASE_NO_LOWERCASE,
     *                  U_TITLECASE_NO_BREAK_ADJUSTMENT, U_TITLECASE_ADJUST_TO_CASED,
     *                  U_TITLECASE_WHOLE_STRING, U_TITLECASE_SENTENCES.
     * @param iter      A break iterator to find the first characters of words that are to be titlecased.
     *                  It is set to the source string (setUText())
     *                  and used one or more times for iteration (first() and next()).
     *                  If NULL, then a word break iterator for the locale is used
     *                  (or something equivalent).
     * @param src       The original string.
     * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
     * @param dest      A buffer for the result string. The result will be NUL-terminated if
     *                  the buffer is large enough.
     *                  The contents is undefined in case of failure.
     * @param destCapacity The size of the buffer (number of bytes). If it is 0, then
     *                  dest may be NULL and the function will only return the length of the result
     *                  without writing any of the result string.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     * @return The length of the result string, if successful.
     *         When the result would be longer than destCapacity,
     *         the full length is returned and a U_BUFFER_OVERFLOW_ERROR is set.
     *
     * @see ucasemap_utf8ToTitle
     * @stable ICU 59
     */
    static int32_t utf8ToTitle(
            const char *locale, uint32_t options, BreakIterator *iter,
            const char *src, int32_t srcLength,
            char *dest, int32_t destCapacity, Edits *edits,
            UErrorCode &errorCode);

#endif  // UCONFIG_NO_BREAK_ITERATION

    /**
     * Case-folds a UTF-8 string and optionally records edits.
     *
     * Case folding is locale-independent and not context-sensitive,
     * but there is an option for whether to include or exclude mappings for dotted I
     * and dotless i that are marked with 'T' in CaseFolding.txt.
     *
     * The result may be longer or shorter than the original.
     * The source string and the destination buffer must not overlap.
     *
     * @param options   Options bit set, usually 0. See U_OMIT_UNCHANGED_TEXT, U_EDITS_NO_RESET,
     *                  U_FOLD_CASE_DEFAULT, U_FOLD_CASE_EXCLUDE_SPECIAL_I.
     * @param src       The original string.
     * @param srcLength The length of the original string. If -1, then src must be NUL-terminated.
     * @param dest      A buffer for the result string. The result will be NUL-terminated if
     *                  the buffer is large enough.
     *                  The contents is undefined in case of failure.
     * @param destCapacity The size of the buffer (number of bytes). If it is 0, then
     *                  dest may be NULL and the function will only return the length of the result
     *                  without writing any of the result string.
     * @param edits     Records edits for index mapping, working with styled text,
     *                  and getting only changes (if any).
     *                  The Edits contents is undefined if any error occurs.
     *                  This function calls edits->reset() first unless
     *                  options includes U_EDITS_NO_RESET. edits can be NULL.
     * @param errorCode Reference to an in/out error code value
     *                  which must not indicate a failure before the function call.
     * @return The length of the result string, if successful.
     *         When the result would be longer than destCapacity,
     *         the full length is returned and a U_BUFFER_OVERFLOW_ERROR is set.
     *
     * @see ucasemap_utf8FoldCase
     * @stable ICU 59
     */
    static int32_t utf8Fold(
            uint32_t options,
            const char *src, int32_t srcLength,
            char *dest, int32_t destCapacity, Edits *edits,
            UErrorCode &errorCode);

private:
    CaseMap() = delete;
    CaseMap(const CaseMap &other) = delete;
    CaseMap &operator=(const CaseMap &other) = delete;
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */





// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// ******************************************************************************
// * Copyright (C) 1996-2015, International Business Machines Corporation and others.
// * All Rights Reserved.
// ******************************************************************************
// */

// #ifndef UBRK_H
// #define UBRK_H

// #include "unicode/utypes.h"
// 
// // © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2004-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  utext.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004oct06
*   created by: Markus W. Scherer
*/


/**
 * \file
 * \brief C API: Abstract Unicode Text API
 *
 * The Text Access API provides a means to allow text that is stored in alternative
 * formats to work with ICU services.  ICU normally operates on text that is
 * stored in UTF-16 format, in (UChar *) arrays for the C APIs or as type
 * UnicodeString for C++ APIs.
 *
 * ICU Text Access allows other formats, such as UTF-8 or non-contiguous
 * UTF-16 strings, to be placed in a UText wrapper and then passed to ICU services.
 *
 * There are three general classes of usage for UText:
 *
 *     Application Level Use.  This is the simplest usage - applications would
 *     use one of the utext_open() functions on their input text, and pass
 *     the resulting UText to the desired ICU service.
 *
 *     Second is usage in ICU Services, such as break iteration, that will need to
 *     operate on input presented to them as a UText.  These implementations
 *     will need to use the iteration and related UText functions to gain
 *     access to the actual text.
 *
 *     The third class of UText users are "text providers."  These are the
 *     UText implementations for the various text storage formats.  An application
 *     or system with a unique text storage format can implement a set of
 *     UText provider functions for that format, which will then allow
 *     ICU services to operate on that format.
 *
 *
 * <em>Iterating over text</em>
 *
 * Here is sample code for a forward iteration over the contents of a UText
 *
 * \code
 *    UChar32  c;
 *    UText    *ut = whatever();
 *
 *    for (c=utext_next32From(ut, 0); c>=0; c=utext_next32(ut)) {
 *       // do whatever with the codepoint c here.
 *    }
 * \endcode
 *
 * And here is similar code to iterate in the reverse direction, from the end
 * of the text towards the beginning.
 *
 * \code
 *    UChar32  c;
 *    UText    *ut = whatever();
 *    int      textLength = utext_nativeLength(ut);
 *    for (c=utext_previous32From(ut, textLength); c>=0; c=utext_previous32(ut)) {
 *       // do whatever with the codepoint c here.
 *    }
 * \endcode
 *
 * <em>Characters and Indexing</em>
 *
 * Indexing into text by UText functions is nearly always in terms of the native
 * indexing of the underlying text storage.  The storage format could be UTF-8
 * or UTF-32, for example.  When coding to the UText access API, no assumptions
 * can be made regarding the size of characters, or how far an index
 * may move when iterating between characters.
 *
 * All indices supplied to UText functions are pinned to the length of the
 * text.  An out-of-bounds index is not considered to be an error, but is
 * adjusted to be in the range  0 <= index <= length of input text.
 *
 *
 * When an index position is returned from a UText function, it will be
 * a native index to the underlying text.  In the case of multi-unit characters,
 * it will  always refer to the first position of the character,
 * never to the interior.  This is essentially the same thing as saying that
 * a returned index will always point to a boundary between characters.
 *
 * When a native index is supplied to a UText function, all indices that
 * refer to any part of a multi-unit character representation are considered
 * to be equivalent.  In the case of multi-unit characters, an incoming index
 * will be logically normalized to refer to the start of the character.
 *
 * It is possible to test whether a native index is on a code point boundary
 * by doing a utext_setNativeIndex() followed by a utext_getNativeIndex().
 * If the index is returned unchanged, it was on a code point boundary.  If
 * an adjusted index is returned, the original index referred to the
 * interior of a character.
 *
 * <em>Conventions for calling UText functions</em>
 *
 * Most UText access functions have as their first parameter a (UText *) pointer,
 * which specifies the UText to be used.  Unless otherwise noted, the
 * pointer must refer to a valid, open UText.  Attempting to
 * use a closed UText or passing a NULL pointer is a programming error and
 * will produce undefined results or NULL pointer exceptions.
 *
 * The UText_Open family of functions can either open an existing (closed)
 * UText, or heap allocate a new UText.  Here is sample code for creating
 * a stack-allocated UText.
 *
 * \code
 *    char     *s = whatever();  // A utf-8 string
 *    U_ErrorCode status = U_ZERO_ERROR;
 *    UText    ut = UTEXT_INITIALIZER;
 *    utext_openUTF8(ut, s, -1, &status);
 *    if (U_FAILURE(status)) {
 *        // error handling
 *    } else {
 *        // work with the UText
 *    }
 * \endcode
 *
 * Any existing UText passed to an open function _must_ have been initialized,
 * either by the UTEXT_INITIALIZER, or by having been originally heap-allocated
 * by an open function.  Passing NULL will cause the open function to
 * heap-allocate and fully initialize a new UText.
 *
 */





// #if U_SHOW_CPLUSPLUS_API




// #endif


U_CDECL_BEGIN

struct UText;
typedef struct UText UText; /**< C typedef for struct UText. @stable ICU 3.6 */


/***************************************************************************************
 *
 *   C Functions for creating UText wrappers around various kinds of text strings.
 *
 ****************************************************************************************/


/**
  * Close function for UText instances.
  * Cleans up, releases any resources being held by an open UText.
  * <p>
  *   If the UText was originally allocated by one of the utext_open functions,
  *   the storage associated with the utext will also be freed.
  *   If the UText storage originated with the application, as it would with
  *   a local or static instance, the storage will not be deleted.
  *
  *   An open UText can be reset to refer to new string by using one of the utext_open()
  *   functions without first closing the UText.
  *
  * @param ut  The UText to be closed.
  * @return    NULL if the UText struct was deleted by the close.  If the UText struct
  *            was originally provided by the caller to the open function, it is
  *            returned by this function, and may be safely used again in
  *            a subsequent utext_open.
  *
  * @stable ICU 3.4
  */
U_STABLE UText * U_EXPORT2
utext_close(UText *ut);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUTextPointer
 * "Smart pointer" class, closes a UText via utext_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUTextPointer, UText, utext_close);

U_NAMESPACE_END

#endif

/**
 * Open a read-only UText implementation for UTF-8 strings.
 *
 * \htmlonly
 * Any invalid UTF-8 in the input will be handled in this way:
 * a sequence of bytes that has the form of a truncated, but otherwise valid,
 * UTF-8 sequence will be replaced by a single unicode replacement character, \uFFFD.
 * Any other illegal bytes will each be replaced by a \uFFFD.
 * \endhtmlonly
 *
 * @param ut     Pointer to a UText struct.  If NULL, a new UText will be created.
 *               If non-NULL, must refer to an initialized UText struct, which will then
 *               be reset to reference the specified UTF-8 string.
 * @param s      A UTF-8 string.  Must not be NULL.
 * @param length The length of the UTF-8 string in bytes, or -1 if the string is
 *               zero terminated.
 * @param status Errors are returned here.
 * @return       A pointer to the UText.  If a pre-allocated UText was provided, it
 *               will always be used and returned.
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openUTF8(UText *ut, const char *s, int64_t length, UErrorCode *status);


/**
 * Open a read-only UText for UChar * string.
 *
 * @param ut     Pointer to a UText struct.  If NULL, a new UText will be created.
 *               If non-NULL, must refer to an initialized UText struct, which will then
 *               be reset to reference the specified UChar string.
 * @param s      A UChar (UTF-16) string
 * @param length The number of UChars in the input string, or -1 if the string is
 *               zero terminated.
 * @param status Errors are returned here.
 * @return       A pointer to the UText.  If a pre-allocated UText was provided, it
 *               will always be used and returned.
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openUChars(UText *ut, const UChar *s, int64_t length, UErrorCode *status);


#if U_SHOW_CPLUSPLUS_API
/**
 * Open a writable UText for a non-const UnicodeString.
 *
 * @param ut      Pointer to a UText struct.  If NULL, a new UText will be created.
 *                 If non-NULL, must refer to an initialized UText struct, which will then
 *                 be reset to reference the specified input string.
 * @param s       A UnicodeString.
 * @param status Errors are returned here.
 * @return        Pointer to the UText.  If a UText was supplied as input, this
 *                 will always be used and returned.
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openUnicodeString(UText *ut, icu::UnicodeString *s, UErrorCode *status);


/**
 * Open a UText for a const UnicodeString.   The resulting UText will not be writable.
 *
 * @param ut    Pointer to a UText struct.  If NULL, a new UText will be created.
 *               If non-NULL, must refer to an initialized UText struct, which will then
 *               be reset to reference the specified input string.
 * @param s      A const UnicodeString to be wrapped.
 * @param status Errors are returned here.
 * @return       Pointer to the UText.  If a UText was supplied as input, this
 *               will always be used and returned.
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openConstUnicodeString(UText *ut, const icu::UnicodeString *s, UErrorCode *status);


/**
 * Open a writable UText implementation for an ICU Replaceable object.
 * @param ut    Pointer to a UText struct.  If NULL, a new UText will be created.
 *               If non-NULL, must refer to an already existing UText, which will then
 *               be reset to reference the specified replaceable text.
 * @param rep    A Replaceable text object.
 * @param status Errors are returned here.
 * @return       Pointer to the UText.  If a UText was supplied as input, this
 *               will always be used and returned.
 * @see Replaceable
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openReplaceable(UText *ut, icu::Replaceable *rep, UErrorCode *status);

/**
 * Open a  UText implementation over an ICU CharacterIterator.
 * @param ut    Pointer to a UText struct.  If NULL, a new UText will be created.
 *               If non-NULL, must refer to an already existing UText, which will then
 *               be reset to reference the specified replaceable text.
 * @param ci     A Character Iterator.
 * @param status Errors are returned here.
 * @return       Pointer to the UText.  If a UText was supplied as input, this
 *               will always be used and returned.
 * @see Replaceable
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openCharacterIterator(UText *ut, icu::CharacterIterator *ci, UErrorCode *status);

#endif


/**
  *  Clone a UText.  This is much like opening a UText where the source text is itself
  *  another UText.
  *
  *  A deep clone will copy both the UText data structures and the underlying text.
  *  The original and cloned UText will operate completely independently; modifications
  *  made to the text in one will not affect the other.  Text providers are not
  *  required to support deep clones.  The user of clone() must check the status return
  *  and be prepared to handle failures.
  *
  *  The standard UText implementations for UTF8, UChar *, UnicodeString and
  *  Replaceable all support deep cloning.
  *
  *  The UText returned from a deep clone will be writable, assuming that the text
  *  provider is able to support writing, even if the source UText had been made
  *  non-writable by means of UText_freeze().
  *
  *  A shallow clone replicates only the UText data structures; it does not make
  *  a copy of the underlying text.  Shallow clones can be used as an efficient way to
  *  have multiple iterators active in a single text string that is not being
  *  modified.
  *
  *  A shallow clone operation will not fail, barring truly exceptional conditions such
  *  as memory allocation failures.
  *
  *  Shallow UText clones should be avoided if the UText functions that modify the
  *  text are expected to be used, either on the original or the cloned UText.
  *  Any such modifications  can cause unpredictable behavior.  Read Only
  *  shallow clones provide some protection against errors of this type by
  *  disabling text modification via the cloned UText.
  *
  *  A shallow clone made with the readOnly parameter == FALSE will preserve the
  *  utext_isWritable() state of the source object.  Note, however, that
  *  write operations must be avoided while more than one UText exists that refer
  *  to the same underlying text.
  *
  *  A UText and its clone may be safely concurrently accessed by separate threads.
  *  This is true for read access only with shallow clones, and for both read and
  *  write access with deep clones.
  *  It is the responsibility of the Text Provider to ensure that this thread safety
  *  constraint is met.
  *
  *  @param dest   A UText struct to be filled in with the result of the clone operation,
  *                or NULL if the clone function should heap-allocate a new UText struct.
  *                If non-NULL, must refer to an already existing UText, which will then
  *                be reset to become the clone.
  *  @param src    The UText to be cloned.
  *  @param deep   TRUE to request a deep clone, FALSE for a shallow clone.
  *  @param readOnly TRUE to request that the cloned UText have read only access to the
  *                underlying text.

  *  @param status Errors are returned here.  For deep clones, U_UNSUPPORTED_ERROR
  *                will be returned if the text provider is unable to clone the
  *                original text.
  *  @return       The newly created clone, or NULL if the clone operation failed.
  *  @stable ICU 3.4
  */
U_STABLE UText * U_EXPORT2
utext_clone(UText *dest, const UText *src, UBool deep, UBool readOnly, UErrorCode *status);


/**
  *  Compare two UText objects for equality.
  *  UTexts are equal if they are iterating over the same text, and
  *    have the same iteration position within the text.
  *    If either or both of the parameters are NULL, the comparison is FALSE.
  *
  *  @param a   The first of the two UTexts to compare.
  *  @param b   The other UText to be compared.
  *  @return    TRUE if the two UTexts are equal.
  *  @stable ICU 3.6
  */
U_STABLE UBool U_EXPORT2
utext_equals(const UText *a, const UText *b);


/*****************************************************************************
 *
 *   Functions to work with the text represented by a UText wrapper
 *
 *****************************************************************************/

/**
  * Get the length of the text.  Depending on the characteristics
  * of the underlying text representation, this may be expensive.
  * @see  utext_isLengthExpensive()
  *
  *
  * @param ut  the text to be accessed.
  * @return the length of the text, expressed in native units.
  *
  * @stable ICU 3.4
  */
U_STABLE int64_t U_EXPORT2
utext_nativeLength(UText *ut);

/**
 *  Return TRUE if calculating the length of the text could be expensive.
 *  Finding the length of NUL terminated strings is considered to be expensive.
 *
 *  Note that the value of this function may change
 *  as the result of other operations on a UText.
 *  Once the length of a string has been discovered, it will no longer
 *  be expensive to report it.
 *
 * @param ut the text to be accessed.
 * @return TRUE if determining the length of the text could be time consuming.
 * @stable ICU 3.4
 */
U_STABLE UBool U_EXPORT2
utext_isLengthExpensive(const UText *ut);

/**
 * Returns the code point at the requested index,
 * or U_SENTINEL (-1) if it is out of bounds.
 *
 * If the specified index points to the interior of a multi-unit
 * character - one of the trail bytes of a UTF-8 sequence, for example -
 * the complete code point will be returned.
 *
 * The iteration position will be set to the start of the returned code point.
 *
 * This function is roughly equivalent to the sequence
 *    utext_setNativeIndex(index);
 *    utext_current32();
 * (There is a subtle difference if the index is out of bounds by being less than zero -
 * utext_setNativeIndex(negative value) sets the index to zero, after which utext_current()
 * will return the char at zero.  utext_char32At(negative index), on the other hand, will
 * return the U_SENTINEL value of -1.)
 *
 * @param ut the text to be accessed
 * @param nativeIndex the native index of the character to be accessed.  If the index points
 *        to other than the first unit of a multi-unit character, it will be adjusted
 *        to the start of the character.
 * @return the code point at the specified index.
 * @stable ICU 3.4
 */
U_STABLE UChar32 U_EXPORT2
utext_char32At(UText *ut, int64_t nativeIndex);


/**
 *
 * Get the code point at the current iteration position,
 * or U_SENTINEL (-1) if the iteration has reached the end of
 * the input text.
 *
 * @param ut the text to be accessed.
 * @return the Unicode code point at the current iterator position.
 * @stable ICU 3.4
 */
U_STABLE UChar32 U_EXPORT2
utext_current32(UText *ut);


/**
 * Get the code point at the current iteration position of the UText, and
 * advance the position to the first index following the character.
 *
 * If the position is at the end of the text (the index following
 * the last character, which is also the length of the text),
 * return U_SENTINEL (-1) and do not advance the index.
 *
 * This is a post-increment operation.
 *
 * An inline macro version of this function, UTEXT_NEXT32(),
 * is available for performance critical use.
 *
 * @param ut the text to be accessed.
 * @return the Unicode code point at the iteration position.
 * @see UTEXT_NEXT32
 * @stable ICU 3.4
 */
U_STABLE UChar32 U_EXPORT2
utext_next32(UText *ut);


/**
 *  Move the iterator position to the character (code point) whose
 *  index precedes the current position, and return that character.
 *  This is a pre-decrement operation.
 *
 *  If the initial position is at the start of the text (index of 0)
 *  return U_SENTINEL (-1), and leave the position unchanged.
 *
 *  An inline macro version of this function, UTEXT_PREVIOUS32(),
 *  is available for performance critical use.
 *
 *  @param ut the text to be accessed.
 *  @return the previous UChar32 code point, or U_SENTINEL (-1)
 *          if the iteration has reached the start of the text.
 *  @see UTEXT_PREVIOUS32
 *  @stable ICU 3.4
 */
U_STABLE UChar32 U_EXPORT2
utext_previous32(UText *ut);


/**
  * Set the iteration index and return the code point at that index.
  * Leave the iteration index at the start of the following code point.
  *
  * This function is the most efficient and convenient way to
  * begin a forward iteration.  The results are identical to the those
  * from the sequence
  * \code
  *    utext_setIndex();
  *    utext_next32();
  * \endcode
  *
  *  @param ut the text to be accessed.
  *  @param nativeIndex Iteration index, in the native units of the text provider.
  *  @return Code point which starts at or before index,
  *         or U_SENTINEL (-1) if it is out of bounds.
  * @stable ICU 3.4
  */
U_STABLE UChar32 U_EXPORT2
utext_next32From(UText *ut, int64_t nativeIndex);



/**
  * Set the iteration index, and return the code point preceding the
  * one specified by the initial index.  Leave the iteration position
  * at the start of the returned code point.
  *
  * This function is the most efficient and convenient way to
  * begin a backwards iteration.
  *
  * @param ut the text to be accessed.
  * @param nativeIndex Iteration index in the native units of the text provider.
  * @return Code point preceding the one at the initial index,
  *         or U_SENTINEL (-1) if it is out of bounds.
  *
  * @stable ICU 3.4
  */
U_STABLE UChar32 U_EXPORT2
utext_previous32From(UText *ut, int64_t nativeIndex);

/**
  * Get the current iterator position, which can range from 0 to
  * the length of the text.
  * The position is a native index into the input text, in whatever format it
  * may have (possibly UTF-8 for example), and may not always be the same as
  * the corresponding UChar (UTF-16) index.
  * The returned position will always be aligned to a code point boundary.
  *
  * @param ut the text to be accessed.
  * @return the current index position, in the native units of the text provider.
  * @stable ICU 3.4
  */
U_STABLE int64_t U_EXPORT2
utext_getNativeIndex(const UText *ut);

/**
 * Set the current iteration position to the nearest code point
 * boundary at or preceding the specified index.
 * The index is in the native units of the original input text.
 * If the index is out of range, it will be pinned to be within
 * the range of the input text.
 * <p>
 * It will usually be more efficient to begin an iteration
 * using the functions utext_next32From() or utext_previous32From()
 * rather than setIndex().
 * <p>
 * Moving the index position to an adjacent character is best done
 * with utext_next32(), utext_previous32() or utext_moveIndex32().
 * Attempting to do direct arithmetic on the index position is
 * complicated by the fact that the size (in native units) of a
 * character depends on the underlying representation of the character
 * (UTF-8, UTF-16, UTF-32, arbitrary codepage), and is not
 * easily knowable.
 *
 * @param ut the text to be accessed.
 * @param nativeIndex the native unit index of the new iteration position.
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
utext_setNativeIndex(UText *ut, int64_t nativeIndex);

/**
 * Move the iterator position by delta code points.  The number of code points
 * is a signed number; a negative delta will move the iterator backwards,
 * towards the start of the text.
 * <p>
 * The index is moved by <code>delta</code> code points
 * forward or backward, but no further backward than to 0 and
 * no further forward than to utext_nativeLength().
 * The resulting index value will be in between 0 and length, inclusive.
 *
 * @param ut the text to be accessed.
 * @param delta the signed number of code points to move the iteration position.
 * @return TRUE if the position could be moved the requested number of positions while
 *              staying within the range [0 - text length].
 * @stable ICU 3.4
 */
U_STABLE UBool U_EXPORT2
utext_moveIndex32(UText *ut, int32_t delta);

/**
 * Get the native index of the character preceding the current position.
 * If the iteration position is already at the start of the text, zero
 * is returned.
 * The value returned is the same as that obtained from the following sequence,
 * but without the side effect of changing the iteration position.
 *
 * \code
 *    UText  *ut = whatever;
 *      ...
 *    utext_previous(ut)
 *    utext_getNativeIndex(ut);
 * \endcode
 *
 * This function is most useful during forwards iteration, where it will get the
 *   native index of the character most recently returned from utext_next().
 *
 * @param ut the text to be accessed
 * @return the native index of the character preceding the current index position,
 *         or zero if the current position is at the start of the text.
 * @stable ICU 3.6
 */
U_STABLE int64_t U_EXPORT2
utext_getPreviousNativeIndex(UText *ut);


/**
 *
 * Extract text from a UText into a UChar buffer.  The range of text to be extracted
 * is specified in the native indices of the UText provider.  These may not necessarily
 * be UTF-16 indices.
 * <p>
 * The size (number of 16 bit UChars) of the data to be extracted is returned.  The
 * full number of UChars is returned, even when the extracted text is truncated
 * because the specified buffer size is too small.
 * <p>
 * The extracted string will (if you are a user) / must (if you are a text provider)
 * be NUL-terminated if there is sufficient space in the destination buffer.  This
 * terminating NUL is not included in the returned length.
 * <p>
 * The iteration index is left at the position following the last extracted character.
 *
 * @param  ut    the UText from which to extract data.
 * @param  nativeStart the native index of the first character to extract.\
 *               If the specified index is out of range,
 *               it will be pinned to be within 0 <= index <= textLength
 * @param  nativeLimit the native string index of the position following the last
 *               character to extract.  If the specified index is out of range,
 *               it will be pinned to be within 0 <= index <= textLength.
 *               nativeLimit must be >= nativeStart.
 * @param  dest  the UChar (UTF-16) buffer into which the extracted text is placed
 * @param  destCapacity  The size, in UChars, of the destination buffer.  May be zero
 *               for precomputing the required size.
 * @param  status receives any error status.
 *         U_BUFFER_OVERFLOW_ERROR: the extracted text was truncated because the
 *         buffer was too small.  Returns number of UChars for preflighting.
 * @return Number of UChars in the data to be extracted.  Does not include a trailing NUL.
 *
 * @stable ICU 3.4
 */
U_STABLE int32_t U_EXPORT2
utext_extract(UText *ut,
             int64_t nativeStart, int64_t nativeLimit,
             UChar *dest, int32_t destCapacity,
             UErrorCode *status);



/************************************************************************************
 *
 *  #define inline versions of selected performance-critical text access functions
 *          Caution:  do not use auto increment++ or decrement-- expressions
 *                    as parameters to these macros.
 *
 *          For most use, where there is no extreme performance constraint, the
 *          normal, non-inline functions are a better choice.  The resulting code
 *          will be smaller, and, if the need ever arises, easier to debug.
 *
 *          These are implemented as #defines rather than real functions
 *          because there is no fully portable way to do inline functions in plain C.
 *
 ************************************************************************************/

#ifndef U_HIDE_INTERNAL_API
/**
 * inline version of utext_current32(), for performance-critical situations.
 *
 * Get the code point at the current iteration position of the UText.
 * Returns U_SENTINEL (-1) if the position is at the end of the
 * text.
 *
 * @internal ICU 4.4 technology preview
 */
#define UTEXT_CURRENT32(ut)  \
    ((ut)->chunkOffset < (ut)->chunkLength && ((ut)->chunkContents)[(ut)->chunkOffset]<0xd800 ? \
    ((ut)->chunkContents)[((ut)->chunkOffset)] : utext_current32(ut))
#endif  /* U_HIDE_INTERNAL_API */

/**
 * inline version of utext_next32(), for performance-critical situations.
 *
 * Get the code point at the current iteration position of the UText, and
 * advance the position to the first index following the character.
 * This is a post-increment operation.
 * Returns U_SENTINEL (-1) if the position is at the end of the
 * text.
 *
 * @stable ICU 3.4
 */
#define UTEXT_NEXT32(ut)  \
    ((ut)->chunkOffset < (ut)->chunkLength && ((ut)->chunkContents)[(ut)->chunkOffset]<0xd800 ? \
    ((ut)->chunkContents)[((ut)->chunkOffset)++] : utext_next32(ut))

/**
 * inline version of utext_previous32(), for performance-critical situations.
 *
 *  Move the iterator position to the character (code point) whose
 *  index precedes the current position, and return that character.
 *  This is a pre-decrement operation.
 *  Returns U_SENTINEL (-1) if the position is at the start of the  text.
 *
 * @stable ICU 3.4
 */
#define UTEXT_PREVIOUS32(ut)  \
    ((ut)->chunkOffset > 0 && \
     (ut)->chunkContents[(ut)->chunkOffset-1] < 0xd800 ? \
          (ut)->chunkContents[--((ut)->chunkOffset)]  :  utext_previous32(ut))

/**
  *  inline version of utext_getNativeIndex(), for performance-critical situations.
  *
  * Get the current iterator position, which can range from 0 to
  * the length of the text.
  * The position is a native index into the input text, in whatever format it
  * may have (possibly UTF-8 for example), and may not always be the same as
  * the corresponding UChar (UTF-16) index.
  * The returned position will always be aligned to a code point boundary.
  *
  * @stable ICU 3.6
  */
#define UTEXT_GETNATIVEINDEX(ut)                       \
    ((ut)->chunkOffset <= (ut)->nativeIndexingLimit?   \
        (ut)->chunkNativeStart+(ut)->chunkOffset :     \
        (ut)->pFuncs->mapOffsetToNative(ut))

/**
  *  inline version of utext_setNativeIndex(), for performance-critical situations.
  *
  * Set the current iteration position to the nearest code point
  * boundary at or preceding the specified index.
  * The index is in the native units of the original input text.
  * If the index is out of range, it will be pinned to be within
  * the range of the input text.
  *
  * @stable ICU 3.8
  */
#define UTEXT_SETNATIVEINDEX(ut, ix) UPRV_BLOCK_MACRO_BEGIN { \
    int64_t __offset = (ix) - (ut)->chunkNativeStart; \
    if (__offset>=0 && __offset<(int64_t)(ut)->nativeIndexingLimit && (ut)->chunkContents[__offset]<0xdc00) { \
        (ut)->chunkOffset=(int32_t)__offset; \
    } else { \
        utext_setNativeIndex((ut), (ix)); \
    } \
} UPRV_BLOCK_MACRO_END



/************************************************************************************
 *
 *   Functions related to writing or modifying the text.
 *   These will work only with modifiable UTexts.  Attempting to
 *   modify a read-only UText will return an error status.
 *
 ************************************************************************************/


/**
 *  Return TRUE if the text can be written (modified) with utext_replace() or
 *  utext_copy().  For the text to be writable, the text provider must
 *  be of a type that supports writing and the UText must not be frozen.
 *
 *  Attempting to modify text when utext_isWriteable() is FALSE will fail -
 *  the text will not be modified, and an error will be returned from the function
 *  that attempted the modification.
 *
 * @param  ut   the UText to be tested.
 * @return TRUE if the text is modifiable.
 *
 * @see    utext_freeze()
 * @see    utext_replace()
 * @see    utext_copy()
 * @stable ICU 3.4
 *
 */
U_STABLE UBool U_EXPORT2
utext_isWritable(const UText *ut);


/**
  * Test whether there is meta data associated with the text.
  * @see Replaceable::hasMetaData()
  *
  * @param ut The UText to be tested
  * @return TRUE if the underlying text includes meta data.
  * @stable ICU 3.4
  */
U_STABLE UBool U_EXPORT2
utext_hasMetaData(const UText *ut);


/**
 * Replace a range of the original text with a replacement text.
 *
 * Leaves the current iteration position at the position following the
 *  newly inserted replacement text.
 *
 * This function is only available on UText types that support writing,
 * that is, ones where utext_isWritable() returns TRUE.
 *
 * When using this function, there should be only a single UText opened onto the
 * underlying native text string.  Behavior after a replace operation
 * on a UText is undefined for any other additional UTexts that refer to the
 * modified string.
 *
 * @param ut               the UText representing the text to be operated on.
 * @param nativeStart      the native index of the start of the region to be replaced
 * @param nativeLimit      the native index of the character following the region to be replaced.
 * @param replacementText  pointer to the replacement text
 * @param replacementLength length of the replacement text, or -1 if the text is NUL terminated.
 * @param status           receives any error status.  Possible errors include
 *                         U_NO_WRITE_PERMISSION
 *
 * @return The signed number of (native) storage units by which
 *         the length of the text expanded or contracted.
 *
 * @stable ICU 3.4
 */
U_STABLE int32_t U_EXPORT2
utext_replace(UText *ut,
             int64_t nativeStart, int64_t nativeLimit,
             const UChar *replacementText, int32_t replacementLength,
             UErrorCode *status);



/**
 *
 * Copy or move a substring from one position to another within the text,
 * while retaining any metadata associated with the text.
 * This function is used to duplicate or reorder substrings.
 * The destination index must not overlap the source range.
 *
 * The text to be copied or moved is inserted at destIndex;
 * it does not replace or overwrite any existing text.
 *
 * The iteration position is left following the newly inserted text
 * at the destination position.
 *
 * This function is only available on UText types that support writing,
 * that is, ones where utext_isWritable() returns TRUE.
 *
 * When using this function, there should be only a single UText opened onto the
 * underlying native text string.  Behavior after a copy operation
 * on a UText is undefined in any other additional UTexts that refer to the
 * modified string.
 *
 * @param ut           The UText representing the text to be operated on.
 * @param nativeStart  The native index of the start of the region to be copied or moved
 * @param nativeLimit  The native index of the character position following the region
 *                     to be copied.
 * @param destIndex    The native destination index to which the source substring is
 *                     copied or moved.
 * @param move         If TRUE, then the substring is moved, not copied/duplicated.
 * @param status       receives any error status.  Possible errors include U_NO_WRITE_PERMISSION
 *
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
utext_copy(UText *ut,
          int64_t nativeStart, int64_t nativeLimit,
          int64_t destIndex,
          UBool move,
          UErrorCode *status);


/**
  *  <p>
  *  Freeze a UText.  This prevents any modification to the underlying text itself
  *  by means of functions operating on this UText.
  *  </p>
  *  <p>
  *  Once frozen, a UText can not be unfrozen.  The intent is to ensure
  *  that a the text underlying a frozen UText wrapper cannot be modified via that UText.
  *  </p>
  *  <p>
  *  Caution:  freezing a UText will disable changes made via the specific
  *   frozen UText wrapper only; it will not have any effect on the ability to
  *   directly modify the text by bypassing the UText.  Any such backdoor modifications
  *   are always an error while UText access is occurring because the underlying
  *   text can get out of sync with UText's buffering.
  *  </p>
  *
  *  @param ut  The UText to be frozen.
  *  @see   utext_isWritable()
  *  @stable ICU 3.6
  */
U_STABLE void U_EXPORT2
utext_freeze(UText *ut);


/**
 * UText provider properties (bit field indexes).
 *
 * @see UText
 * @stable ICU 3.4
 */
enum {
    /**
     * It is potentially time consuming for the provider to determine the length of the text.
     * @stable ICU 3.4
     */
    UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE = 1,
    /**
     * Text chunks remain valid and usable until the text object is modified or
     * deleted, not just until the next time the access() function is called
     * (which is the default).
     * @stable ICU 3.4
     */
    UTEXT_PROVIDER_STABLE_CHUNKS = 2,
    /**
     * The provider supports modifying the text via the replace() and copy()
     * functions.
     * @see Replaceable
     * @stable ICU 3.4
     */
    UTEXT_PROVIDER_WRITABLE = 3,
    /**
     * There is meta data associated with the text.
     * @see Replaceable::hasMetaData()
     * @stable ICU 3.4
     */
    UTEXT_PROVIDER_HAS_META_DATA = 4,
    /**
     * Text provider owns the text storage.
     *  Generally occurs as the result of a deep clone of the UText.
     *  When closing the UText, the associated text must
     *  also be closed/deleted/freed/ whatever is appropriate.
     * @stable ICU 3.6
     */
     UTEXT_PROVIDER_OWNS_TEXT = 5
};

/**
  * Function type declaration for UText.clone().
  *
  *  clone a UText.  Much like opening a UText where the source text is itself
  *  another UText.
  *
  *  A deep clone will copy both the UText data structures and the underlying text.
  *  The original and cloned UText will operate completely independently; modifications
  *  made to the text in one will not effect the other.  Text providers are not
  *  required to support deep clones.  The user of clone() must check the status return
  *  and be prepared to handle failures.
  *
  *  A shallow clone replicates only the UText data structures; it does not make
  *  a copy of the underlying text.  Shallow clones can be used as an efficient way to
  *  have multiple iterators active in a single text string that is not being
  *  modified.
  *
  *  A shallow clone operation must not fail except for truly exceptional conditions such
  *  as memory allocation failures.
  *
  *  A UText and its clone may be safely concurrently accessed by separate threads.
  *  This is true for both shallow and deep clones.
  *  It is the responsibility of the Text Provider to ensure that this thread safety
  *  constraint is met.

  *
  *  @param dest   A UText struct to be filled in with the result of the clone operation,
  *                or NULL if the clone function should heap-allocate a new UText struct.
  *  @param src    The UText to be cloned.
  *  @param deep   TRUE to request a deep clone, FALSE for a shallow clone.
  *  @param status Errors are returned here.  For deep clones, U_UNSUPPORTED_ERROR
  *                should be returned if the text provider is unable to clone the
  *                original text.
  *  @return       The newly created clone, or NULL if the clone operation failed.
  *
  * @stable ICU 3.4
  */
typedef UText * U_CALLCONV
UTextClone(UText *dest, const UText *src, UBool deep, UErrorCode *status);


/**
 * Function type declaration for UText.nativeLength().
 *
 * @param ut the UText to get the length of.
 * @return the length, in the native units of the original text string.
 * @see UText
 * @stable ICU 3.4
 */
typedef int64_t U_CALLCONV
UTextNativeLength(UText *ut);

/**
 * Function type declaration for UText.access().  Get the description of the text chunk
 *  containing the text at a requested native index.  The UText's iteration
 *  position will be left at the requested index.  If the index is out
 *  of bounds, the iteration position will be left at the start or end
 *  of the string, as appropriate.
 *
 *  Chunks must begin and end on code point boundaries.  A single code point
 *  comprised of multiple storage units must never span a chunk boundary.
 *
 *
 * @param ut          the UText being accessed.
 * @param nativeIndex Requested index of the text to be accessed.
 * @param forward     If TRUE, then the returned chunk must contain text
 *                    starting from the index, so that start<=index<limit.
 *                    If FALSE, then the returned chunk must contain text
 *                    before the index, so that start<index<=limit.
 * @return            True if the requested index could be accessed.  The chunk
 *                    will contain the requested text.
 *                    False value if a chunk cannot be accessed
 *                    (the requested index is out of bounds).
 *
 * @see UText
 * @stable ICU 3.4
 */
typedef UBool U_CALLCONV
UTextAccess(UText *ut, int64_t nativeIndex, UBool forward);

/**
 * Function type declaration for UText.extract().
 *
 * Extract text from a UText into a UChar buffer.  The range of text to be extracted
 * is specified in the native indices of the UText provider.  These may not necessarily
 * be UTF-16 indices.
 * <p>
 * The size (number of 16 bit UChars) in the data to be extracted is returned.  The
 * full amount is returned, even when the specified buffer size is smaller.
 * <p>
 * The extracted string will (if you are a user) / must (if you are a text provider)
 * be NUL-terminated if there is sufficient space in the destination buffer.
 *
 * @param  ut            the UText from which to extract data.
 * @param  nativeStart   the native index of the first character to extract.
 * @param  nativeLimit   the native string index of the position following the last
 *                       character to extract.
 * @param  dest          the UChar (UTF-16) buffer into which the extracted text is placed
 * @param  destCapacity  The size, in UChars, of the destination buffer.  May be zero
 *                       for precomputing the required size.
 * @param  status        receives any error status.
 *                       If U_BUFFER_OVERFLOW_ERROR: Returns number of UChars for
 *                       preflighting.
 * @return Number of UChars in the data.  Does not include a trailing NUL.
 *
 * @stable ICU 3.4
 */
typedef int32_t U_CALLCONV
UTextExtract(UText *ut,
             int64_t nativeStart, int64_t nativeLimit,
             UChar *dest, int32_t destCapacity,
             UErrorCode *status);

/**
 * Function type declaration for UText.replace().
 *
 * Replace a range of the original text with a replacement text.
 *
 * Leaves the current iteration position at the position following the
 *  newly inserted replacement text.
 *
 * This function need only be implemented on UText types that support writing.
 *
 * When using this function, there should be only a single UText opened onto the
 * underlying native text string.  The function is responsible for updating the
 * text chunk within the UText to reflect the updated iteration position,
 * taking into account any changes to the underlying string's structure caused
 * by the replace operation.
 *
 * @param ut               the UText representing the text to be operated on.
 * @param nativeStart      the index of the start of the region to be replaced
 * @param nativeLimit      the index of the character following the region to be replaced.
 * @param replacementText  pointer to the replacement text
 * @param replacmentLength length of the replacement text in UChars, or -1 if the text is NUL terminated.
 * @param status           receives any error status.  Possible errors include
 *                         U_NO_WRITE_PERMISSION
 *
 * @return The signed number of (native) storage units by which
 *         the length of the text expanded or contracted.
 *
 * @stable ICU 3.4
 */
typedef int32_t U_CALLCONV
UTextReplace(UText *ut,
             int64_t nativeStart, int64_t nativeLimit,
             const UChar *replacementText, int32_t replacmentLength,
             UErrorCode *status);

/**
 * Function type declaration for UText.copy().
 *
 * Copy or move a substring from one position to another within the text,
 * while retaining any metadata associated with the text.
 * This function is used to duplicate or reorder substrings.
 * The destination index must not overlap the source range.
 *
 * The text to be copied or moved is inserted at destIndex;
 * it does not replace or overwrite any existing text.
 *
 * This function need only be implemented for UText types that support writing.
 *
 * When using this function, there should be only a single UText opened onto the
 * underlying native text string.  The function is responsible for updating the
 * text chunk within the UText to reflect the updated iteration position,
 * taking into account any changes to the underlying string's structure caused
 * by the replace operation.
 *
 * @param ut           The UText representing the text to be operated on.
 * @param nativeStart  The index of the start of the region to be copied or moved
 * @param nativeLimit  The index of the character following the region to be replaced.
 * @param nativeDest   The destination index to which the source substring is copied or moved.
 * @param move         If TRUE, then the substring is moved, not copied/duplicated.
 * @param status       receives any error status.  Possible errors include U_NO_WRITE_PERMISSION
 *
 * @stable ICU 3.4
 */
typedef void U_CALLCONV
UTextCopy(UText *ut,
          int64_t nativeStart, int64_t nativeLimit,
          int64_t nativeDest,
          UBool move,
          UErrorCode *status);

/**
 * Function type declaration for UText.mapOffsetToNative().
 * Map from the current UChar offset within the current text chunk to
 *  the corresponding native index in the original source text.
 *
 * This is required only for text providers that do not use native UTF-16 indexes.
 *
 * @param ut     the UText.
 * @return Absolute (native) index corresponding to chunkOffset in the current chunk.
 *         The returned native index should always be to a code point boundary.
 *
 * @stable ICU 3.4
 */
typedef int64_t U_CALLCONV
UTextMapOffsetToNative(const UText *ut);

/**
 * Function type declaration for UText.mapIndexToUTF16().
 * Map from a native index to a UChar offset within a text chunk.
 * Behavior is undefined if the native index does not fall within the
 *   current chunk.
 *
 * This function is required only for text providers that do not use native UTF-16 indexes.
 *
 * @param ut          The UText containing the text chunk.
 * @param nativeIndex Absolute (native) text index, chunk->start<=index<=chunk->limit.
 * @return            Chunk-relative UTF-16 offset corresponding to the specified native
 *                    index.
 *
 * @stable ICU 3.4
 */
typedef int32_t U_CALLCONV
UTextMapNativeIndexToUTF16(const UText *ut, int64_t nativeIndex);


/**
 * Function type declaration for UText.utextClose().
 *
 * A Text Provider close function is only required for provider types that make
 *  allocations in their open function (or other functions) that must be
 *  cleaned when the UText is closed.
 *
 * The allocation of the UText struct itself and any "extra" storage
 * associated with the UText is handled by the common UText implementation
 * and does not require provider specific cleanup in a close function.
 *
 * Most UText provider implementations do not need to implement this function.
 *
 * @param ut A UText object to be closed.
 *
 * @stable ICU 3.4
 */
typedef void U_CALLCONV
UTextClose(UText *ut);


/**
  *   (public)  Function dispatch table for UText.
  *             Conceptually very much like a C++ Virtual Function Table.
  *             This struct defines the organization of the table.
  *             Each text provider implementation must provide an
  *              actual table that is initialized with the appropriate functions
  *              for the type of text being handled.
  *   @stable ICU 3.6
  */
struct UTextFuncs {
    /**
     *   (public)  Function table size, sizeof(UTextFuncs)
     *             Intended for use should the table grow to accommodate added
     *             functions in the future, to allow tests for older format
     *             function tables that do not contain the extensions.
     *
     *             Fields are placed for optimal alignment on
     *             32/64/128-bit-pointer machines, by normally grouping together
     *             4 32-bit fields,
     *             4 pointers,
     *             2 64-bit fields
     *             in sequence.
     *   @stable ICU 3.6
     */
    int32_t       tableSize;

    /**
      *   (private)  Alignment padding.
      *              Do not use, reserved for use by the UText framework only.
      *   @internal
      */
    int32_t       reserved1, /** @internal */ reserved2, /** @internal */ reserved3;


    /**
     * (public) Function pointer for UTextClone
     *
     * @see UTextClone
     * @stable ICU 3.6
     */
    UTextClone *clone;

    /**
     * (public) function pointer for UTextLength
     * May be expensive to compute!
     *
     * @see UTextLength
     * @stable ICU 3.6
     */
    UTextNativeLength *nativeLength;

    /**
     * (public) Function pointer for UTextAccess.
     *
     * @see UTextAccess
     * @stable ICU 3.6
     */
    UTextAccess *access;

    /**
     * (public) Function pointer for UTextExtract.
     *
     * @see UTextExtract
     * @stable ICU 3.6
     */
    UTextExtract *extract;

    /**
     * (public) Function pointer for UTextReplace.
     *
     * @see UTextReplace
     * @stable ICU 3.6
     */
    UTextReplace *replace;

    /**
     * (public) Function pointer for UTextCopy.
     *
     * @see UTextCopy
     * @stable ICU 3.6
     */
    UTextCopy *copy;

    /**
     * (public) Function pointer for UTextMapOffsetToNative.
     *
     * @see UTextMapOffsetToNative
     * @stable ICU 3.6
     */
    UTextMapOffsetToNative *mapOffsetToNative;

    /**
     * (public) Function pointer for UTextMapNativeIndexToUTF16.
     *
     * @see UTextMapNativeIndexToUTF16
     * @stable ICU 3.6
     */
    UTextMapNativeIndexToUTF16 *mapNativeIndexToUTF16;

    /**
     * (public) Function pointer for UTextClose.
      *
      * @see UTextClose
      * @stable ICU 3.6
      */
    UTextClose  *close;

    /**
      * (private)  Spare function pointer
      * @internal
      */
    UTextClose  *spare1;

    /**
      * (private)  Spare function pointer
      * @internal
      */
    UTextClose  *spare2;

    /**
      * (private)  Spare function pointer
      * @internal
      */
    UTextClose  *spare3;

};
/**
 * Function dispatch table for UText
 * @see UTextFuncs
 */
typedef struct UTextFuncs UTextFuncs;

 /**
  *   UText struct.  Provides the interface between the generic UText access code
  *                  and the UText provider code that works on specific kinds of
  *                  text  (UTF-8, noncontiguous UTF-16, whatever.)
  *
  *                  Applications that are using predefined types of text providers
  *                  to pass text data to ICU services will have no need to view the
  *                  internals of the UText structs that they open.
  *
  * @stable ICU 3.6
  */
struct UText {
    /**
     *     (private)  Magic.  Used to help detect when UText functions are handed
     *                        invalid or uninitialized UText structs.
     *                        utext_openXYZ() functions take an initialized,
     *                        but not necessarily open, UText struct as an
     *                        optional fill-in parameter.  This magic field
     *                        is used to check for that initialization.
     *                        Text provider close functions must NOT clear
     *                        the magic field because that would prevent
     *                        reuse of the UText struct.
     * @internal
     */
    uint32_t       magic;


    /**
     *     (private)  Flags for managing the allocation and freeing of
     *                memory associated with this UText.
     * @internal
     */
    int32_t        flags;


    /**
      *  Text provider properties.  This set of flags is maintained by the
      *                             text provider implementation.
      *  @stable ICU 3.4
      */
    int32_t         providerProperties;

    /**
     * (public) sizeOfStruct=sizeof(UText)
     * Allows possible backward compatible extension.
     *
     * @stable ICU 3.4
     */
    int32_t         sizeOfStruct;

    /* ------ 16 byte alignment boundary -----------  */


    /**
      *  (protected) Native index of the first character position following
      *              the current chunk.
      *  @stable ICU 3.6
      */
    int64_t         chunkNativeLimit;

    /**
     *   (protected)  Size in bytes of the extra space (pExtra).
     *  @stable ICU 3.4
     */
    int32_t        extraSize;

    /**
      *    (protected) The highest chunk offset where native indexing and
      *    chunk (UTF-16) indexing correspond.  For UTF-16 sources, value
      *    will be equal to chunkLength.
      *
      *    @stable ICU 3.6
      */
    int32_t         nativeIndexingLimit;

    /* ---- 16 byte alignment boundary------ */

    /**
     *  (protected) Native index of the first character in the text chunk.
     *  @stable ICU 3.6
     */
    int64_t         chunkNativeStart;

    /**
     *  (protected) Current iteration position within the text chunk (UTF-16 buffer).
     *  This is the index to the character that will be returned by utext_next32().
     *  @stable ICU 3.6
     */
    int32_t         chunkOffset;

    /**
     *  (protected) Length the text chunk (UTF-16 buffer), in UChars.
     *  @stable ICU 3.6
     */
    int32_t         chunkLength;

    /* ---- 16  byte alignment boundary-- */


    /**
     *  (protected)  pointer to a chunk of text in UTF-16 format.
     *  May refer either to original storage of the source of the text, or
     *  if conversion was required, to a buffer owned by the UText.
     *  @stable ICU 3.6
     */
    const UChar    *chunkContents;

     /**
      * (public)     Pointer to Dispatch table for accessing functions for this UText.
      * @stable ICU 3.6
      */
    const UTextFuncs     *pFuncs;

    /**
     *  (protected)  Pointer to additional space requested by the
     *               text provider during the utext_open operation.
     * @stable ICU 3.4
     */
    void          *pExtra;

    /**
     * (protected) Pointer to string or text-containing object or similar.
     * This is the source of the text that this UText is wrapping, in a format
     *  that is known to the text provider functions.
     * @stable ICU 3.4
     */
    const void   *context;

    /* --- 16 byte alignment boundary--- */

    /**
     * (protected) Pointer fields available for use by the text provider.
     * Not used by UText common code.
     * @stable ICU 3.6
     */
    const void     *p;
    /**
     * (protected) Pointer fields available for use by the text provider.
     * Not used by UText common code.
     * @stable ICU 3.6
     */
    const void     *q;
     /**
     * (protected) Pointer fields available for use by the text provider.
     * Not used by UText common code.
     * @stable ICU 3.6
      */
    const void     *r;

    /**
      *  Private field reserved for future use by the UText framework
      *     itself.  This is not to be touched by the text providers.
      * @internal ICU 3.4
      */
    void           *privP;


    /* --- 16 byte alignment boundary--- */


    /**
      * (protected) Integer field reserved for use by the text provider.
      * Not used by the UText framework, or by the client (user) of the UText.
      * @stable ICU 3.4
      */
    int64_t         a;

    /**
      * (protected) Integer field reserved for use by the text provider.
      * Not used by the UText framework, or by the client (user) of the UText.
      * @stable ICU 3.4
      */
    int32_t         b;

    /**
      * (protected) Integer field reserved for use by the text provider.
      * Not used by the UText framework, or by the client (user) of the UText.
      * @stable ICU 3.4
      */
    int32_t         c;

    /*  ---- 16 byte alignment boundary---- */


    /**
      *  Private field reserved for future use by the UText framework
      *     itself.  This is not to be touched by the text providers.
      * @internal ICU 3.4
      */
    int64_t         privA;
    /**
      *  Private field reserved for future use by the UText framework
      *     itself.  This is not to be touched by the text providers.
      * @internal ICU 3.4
      */
    int32_t         privB;
    /**
      *  Private field reserved for future use by the UText framework
      *     itself.  This is not to be touched by the text providers.
      * @internal ICU 3.4
      */
    int32_t         privC;
};


/**
 *  Common function for use by Text Provider implementations to allocate and/or initialize
 *  a new UText struct.  To be called in the implementation of utext_open() functions.
 *  If the supplied UText parameter is null, a new UText struct will be allocated on the heap.
 *  If the supplied UText is already open, the provider's close function will be called
 *  so that the struct can be reused by the open that is in progress.
 *
 * @param ut   pointer to a UText struct to be re-used, or null if a new UText
 *             should be allocated.
 * @param extraSpace The amount of additional space to be allocated as part
 *             of this UText, for use by types of providers that require
 *             additional storage.
 * @param status Errors are returned here.
 * @return pointer to the UText, allocated if necessary, with extra space set up if requested.
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_setup(UText *ut, int32_t extraSpace, UErrorCode *status);

// do not use #ifndef U_HIDE_INTERNAL_API around the following!
/**
  * @internal
  *  Value used to help identify correctly initialized UText structs.
  *  Note:  must be publicly visible so that UTEXT_INITIALIZER can access it.
  */
enum {
    UTEXT_MAGIC = 0x345ad82c
};

/**
 * initializer to be used with local (stack) instances of a UText
 *  struct.  UText structs must be initialized before passing
 *  them to one of the utext_open functions.
 *
 * @stable ICU 3.6
 */
#define UTEXT_INITIALIZER {                                        \
                  UTEXT_MAGIC,          /* magic                */ \
                  0,                    /* flags                */ \
                  0,                    /* providerProps        */ \
                  sizeof(UText),        /* sizeOfStruct         */ \
                  0,                    /* chunkNativeLimit     */ \
                  0,                    /* extraSize            */ \
                  0,                    /* nativeIndexingLimit  */ \
                  0,                    /* chunkNativeStart     */ \
                  0,                    /* chunkOffset          */ \
                  0,                    /* chunkLength          */ \
                  NULL,                 /* chunkContents        */ \
                  NULL,                 /* pFuncs               */ \
                  NULL,                 /* pExtra               */ \
                  NULL,                 /* context              */ \
                  NULL, NULL, NULL,     /* p, q, r              */ \
                  NULL,                 /* privP                */ \
                  0, 0, 0,              /* a, b, c              */ \
                  0, 0, 0               /* privA,B,C,           */ \
                  }


U_CDECL_END




// #include "unicode/localpointer.h"

// /**
//  * A text-break iterator.
//  *  For usage in C programs.
//  */
// #ifndef UBRK_TYPEDEF_UBREAK_ITERATOR
// #   define UBRK_TYPEDEF_UBREAK_ITERATOR
//     /**
//      *  Opaque type representing an ICU Break iterator object.
//      *  @stable ICU 2.0
//      */
//     typedef struct UBreakIterator UBreakIterator;
// #endif

// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/parseerr.h"

// /**
//  * \file
//  * \brief C API: BreakIterator
//  *
//  * <h2> BreakIterator C API </h2>
//  *
//  * The BreakIterator C API defines  methods for finding the location
//  * of boundaries in text. Pointer to a UBreakIterator maintain a
//  * current position and scan over text returning the index of characters
//  * where boundaries occur.
//  * <p>
//  * Line boundary analysis determines where a text string can be broken
//  * when line-wrapping. The mechanism correctly handles punctuation and
//  * hyphenated words.
//  * <p>
//  * Note: The locale keyword "lb" can be used to modify line break
//  * behavior according to the CSS level 3 line-break options, see
//  * <http://dev.w3.org/csswg/css-text/#line-breaking>. For example:
//  * "ja@lb=strict", "zh@lb=loose".
//  * <p>
//  * Sentence boundary analysis allows selection with correct
//  * interpretation of periods within numbers and abbreviations, and
//  * trailing punctuation marks such as quotation marks and parentheses.
//  * <p>
//  * Note: The locale keyword "ss" can be used to enable use of
//  * segmentation suppression data (preventing breaks in English after
//  * abbreviations such as "Mr." or "Est.", for example), as follows:
//  * "en@ss=standard".
//  * <p>
//  * Word boundary analysis is used by search and replace functions, as
//  * well as within text editing applications that allow the user to
//  * select words with a double click. Word selection provides correct
//  * interpretation of punctuation marks within and following
//  * words. Characters that are not part of a word, such as symbols or
//  * punctuation marks, have word-breaks on both sides.
//  * <p>
//  * Character boundary analysis identifies the boundaries of
//  * "Extended Grapheme Clusters", which are groupings of codepoints
//  * that should be treated as character-like units for many text operations.
//  * Please see Unicode Standard Annex #29, Unicode Text Segmentation,
//  * http://www.unicode.org/reports/tr29/ for additional information
//  * on grapheme clusters and guidelines on their use.
//  * <p>
//  * Title boundary analysis locates all positions,
//  * typically starts of words, that should be set to Title Case
//  * when title casing the text.
//  * <p>
//  * The text boundary positions are found according to the rules
//  * described in Unicode Standard Annex #29, Text Boundaries, and
//  * Unicode Standard Annex #14, Line Breaking Properties.  These
//  * are available at http://www.unicode.org/reports/tr14/ and
//  * http://www.unicode.org/reports/tr29/.
//  * <p>
//  * In addition to the plain C API defined in this header file, an
//  * object oriented C++ API with equivalent functionality is defined in the
//  * file brkiter.h.
//  * <p>
//  * Code snippets illustrating the use of the Break Iterator APIs
//  * are available in the ICU User Guide,
//  * http://icu-project.org/userguide/boundaryAnalysis.html
//  * and in the sample program icu/source/samples/break/break.cpp
//  */

// /** The possible types of text boundaries.  @stable ICU 2.0 */
// typedef enum UBreakIteratorType {
//   /** Character breaks  @stable ICU 2.0 */
//   UBRK_CHARACTER = 0,
//   /** Word breaks @stable ICU 2.0 */
//   UBRK_WORD = 1,
//   /** Line breaks @stable ICU 2.0 */
//   UBRK_LINE = 2,
//   /** Sentence breaks @stable ICU 2.0 */
//   UBRK_SENTENCE = 3,

// #ifndef U_HIDE_DEPRECATED_API
//   /**
//    * Title Case breaks
//    * The iterator created using this type locates title boundaries as described for
//    * Unicode 3.2 only. For Unicode 4.0 and above title boundary iteration,
//    * please use Word Boundary iterator.
//    *
//    * @deprecated ICU 2.8 Use the word break iterator for titlecasing for Unicode 4 and later.
//    */
//   UBRK_TITLE = 4,
//     /**
//      * One more than the highest normal UBreakIteratorType value.
//      * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
//      */
//     UBRK_COUNT = 5
// #endif  // U_HIDE_DEPRECATED_API
// } UBreakIteratorType;

// /** Value indicating all text boundaries have been returned.
//  *  @stable ICU 2.0
//  */
// #define UBRK_DONE ((int32_t) -1)


// /**
//  *  Enum constants for the word break tags returned by
//  *  getRuleStatus().  A range of values is defined for each category of
//  *  word, to allow for further subdivisions of a category in future releases.
//  *  Applications should check for tag values falling within the range, rather
//  *  than for single individual values.
//  *
//  * The numeric values of all of these constants are stable (will not change).
//  *
//  * @stable ICU 2.2
// */
// typedef enum UWordBreak {
//     /** Tag value for "words" that do not fit into any of other categories.
//      *  Includes spaces and most punctuation. */
//     UBRK_WORD_NONE           = 0,
//     /** Upper bound for tags for uncategorized words. */
//     UBRK_WORD_NONE_LIMIT     = 100,
//     /** Tag value for words that appear to be numbers, lower limit.    */
//     UBRK_WORD_NUMBER         = 100,
//     /** Tag value for words that appear to be numbers, upper limit.    */
//     UBRK_WORD_NUMBER_LIMIT   = 200,
//     /** Tag value for words that contain letters, excluding
//      *  hiragana, katakana or ideographic characters, lower limit.    */
//     UBRK_WORD_LETTER         = 200,
//     /** Tag value for words containing letters, upper limit  */
//     UBRK_WORD_LETTER_LIMIT   = 300,
//     /** Tag value for words containing kana characters, lower limit */
//     UBRK_WORD_KANA           = 300,
//     /** Tag value for words containing kana characters, upper limit */
//     UBRK_WORD_KANA_LIMIT     = 400,
//     /** Tag value for words containing ideographic characters, lower limit */
//     UBRK_WORD_IDEO           = 400,
//     /** Tag value for words containing ideographic characters, upper limit */
//     UBRK_WORD_IDEO_LIMIT     = 500
// } UWordBreak;

// /**
//  *  Enum constants for the line break tags returned by getRuleStatus().
//  *  A range of values is defined for each category of
//  *  word, to allow for further subdivisions of a category in future releases.
//  *  Applications should check for tag values falling within the range, rather
//  *  than for single individual values.
//  *
//  * The numeric values of all of these constants are stable (will not change).
//  *
//  * @stable ICU 2.8
// */
// typedef enum ULineBreakTag {
//     /** Tag value for soft line breaks, positions at which a line break
//       *  is acceptable but not required                */
//     UBRK_LINE_SOFT            = 0,
//     /** Upper bound for soft line breaks.              */
//     UBRK_LINE_SOFT_LIMIT      = 100,
//     /** Tag value for a hard, or mandatory line break  */
//     UBRK_LINE_HARD            = 100,
//     /** Upper bound for hard line breaks.              */
//     UBRK_LINE_HARD_LIMIT      = 200
// } ULineBreakTag;



// /**
//  *  Enum constants for the sentence break tags returned by getRuleStatus().
//  *  A range of values is defined for each category of
//  *  sentence, to allow for further subdivisions of a category in future releases.
//  *  Applications should check for tag values falling within the range, rather
//  *  than for single individual values.
//  *
//  * The numeric values of all of these constants are stable (will not change).
//  *
//  * @stable ICU 2.8
// */
// typedef enum USentenceBreakTag {
//     /** Tag value for for sentences  ending with a sentence terminator
//       * ('.', '?', '!', etc.) character, possibly followed by a
//       * hard separator (CR, LF, PS, etc.)
//       */
//     UBRK_SENTENCE_TERM       = 0,
//     /** Upper bound for tags for sentences ended by sentence terminators.    */
//     UBRK_SENTENCE_TERM_LIMIT = 100,
//     /** Tag value for for sentences that do not contain an ending
//       * sentence terminator ('.', '?', '!', etc.) character, but
//       * are ended only by a hard separator (CR, LF, PS, etc.) or end of input.
//       */
//     UBRK_SENTENCE_SEP        = 100,
//     /** Upper bound for tags for sentences ended by a separator.              */
//     UBRK_SENTENCE_SEP_LIMIT  = 200
//     /** Tag value for a hard, or mandatory line break  */
// } USentenceBreakTag;


// /**
//  * Open a new UBreakIterator for locating text boundaries for a specified locale.
//  * A UBreakIterator may be used for detecting character, line, word,
//  * and sentence breaks in text.
//  * @param type The type of UBreakIterator to open: one of UBRK_CHARACTER, UBRK_WORD,
//  * UBRK_LINE, UBRK_SENTENCE
//  * @param locale The locale specifying the text-breaking conventions. Note that
//  * locale keys such as "lb" and "ss" may be used to modify text break behavior,
//  * see general discussion of BreakIterator C API.
//  * @param text The text to be iterated over. May be null, in which case ubrk_setText() is
//  *        used to specify the text to be iterated.
//  * @param textLength The number of characters in text, or -1 if null-terminated.
//  * @param status A UErrorCode to receive any errors.
//  * @return A UBreakIterator for the specified locale.
//  * @see ubrk_openRules
//  * @stable ICU 2.0
//  */
// U_STABLE UBreakIterator* U_EXPORT2
// ubrk_open(UBreakIteratorType type,
//       const char *locale,
//       const UChar *text,
//       int32_t textLength,
//       UErrorCode *status);

// /**
//  * Open a new UBreakIterator for locating text boundaries using specified breaking rules.
//  * The rule syntax is ... (TBD)
//  * @param rules A set of rules specifying the text breaking conventions.
//  * @param rulesLength The number of characters in rules, or -1 if null-terminated.
//  * @param text The text to be iterated over.  May be null, in which case ubrk_setText() is
//  *        used to specify the text to be iterated.
//  * @param textLength The number of characters in text, or -1 if null-terminated.
//  * @param parseErr   Receives position and context information for any syntax errors
//  *                   detected while parsing the rules.
//  * @param status A UErrorCode to receive any errors.
//  * @return A UBreakIterator for the specified rules.
//  * @see ubrk_open
//  * @stable ICU 2.2
//  */
// U_STABLE UBreakIterator* U_EXPORT2
// ubrk_openRules(const UChar     *rules,
//                int32_t         rulesLength,
//                const UChar     *text,
//                int32_t          textLength,
//                UParseError     *parseErr,
//                UErrorCode      *status);

// /**
//  * Open a new UBreakIterator for locating text boundaries using precompiled binary rules.
//  * Opening a UBreakIterator this way is substantially faster than using ubrk_openRules.
//  * Binary rules may be obtained using ubrk_getBinaryRules. The compiled rules are not
//  * compatible across different major versions of ICU, nor across platforms of different
//  * endianness or different base character set family (ASCII vs EBCDIC).
//  * @param binaryRules A set of compiled binary rules specifying the text breaking
//  *                    conventions. Ownership of the storage containing the compiled
//  *                    rules remains with the caller of this function. The compiled
//  *                    rules must not be modified or deleted during the life of the
//  *                    break iterator.
//  * @param rulesLength The length of binaryRules in bytes; must be >= 0.
//  * @param text        The text to be iterated over.  May be null, in which case
//  *                    ubrk_setText() is used to specify the text to be iterated.
//  * @param textLength  The number of characters in text, or -1 if null-terminated.
//  * @param status      Pointer to UErrorCode to receive any errors.
//  * @return            UBreakIterator for the specified rules.
//  * @see ubrk_getBinaryRules
//  * @stable ICU 59
//  */
// U_STABLE UBreakIterator* U_EXPORT2
// ubrk_openBinaryRules(const uint8_t *binaryRules, int32_t rulesLength,
//                      const UChar *  text, int32_t textLength,
//                      UErrorCode *   status);

// /**
//  * Thread safe cloning operation
//  * @param bi iterator to be cloned
//  * @param stackBuffer <em>Deprecated functionality as of ICU 52, use NULL.</em><br>
//  *  user allocated space for the new clone. If NULL new memory will be allocated.
//  *  If buffer is not large enough, new memory will be allocated.
//  *  Clients can use the U_BRK_SAFECLONE_BUFFERSIZE.
//  * @param pBufferSize <em>Deprecated functionality as of ICU 52, use NULL or 1.</em><br>
//  *  pointer to size of allocated space.
//  *  If *pBufferSize == 0, a sufficient size for use in cloning will
//  *  be returned ('pre-flighting')
//  *  If *pBufferSize is not enough for a stack-based safe clone,
//  *  new memory will be allocated.
//  * @param status to indicate whether the operation went on smoothly or there were errors
//  *  An informational status value, U_SAFECLONE_ALLOCATED_ERROR, is used if any allocations were necessary.
//  * @return pointer to the new clone
//  * @stable ICU 2.0
//  */
// U_STABLE UBreakIterator * U_EXPORT2
// ubrk_safeClone(
//           const UBreakIterator *bi,
//           void *stackBuffer,
//           int32_t *pBufferSize,
//           UErrorCode *status);

// #ifndef U_HIDE_DEPRECATED_API

// /**
//   * A recommended size (in bytes) for the memory buffer to be passed to ubrk_saveClone().
//   * @deprecated ICU 52. Do not rely on ubrk_safeClone() cloning into any provided buffer.
//   */
// #define U_BRK_SAFECLONE_BUFFERSIZE 1

// #endif /* U_HIDE_DEPRECATED_API */

// /**
// * Close a UBreakIterator.
// * Once closed, a UBreakIterator may no longer be used.
// * @param bi The break iterator to close.
//  * @stable ICU 2.0
// */
// U_STABLE void U_EXPORT2
// ubrk_close(UBreakIterator *bi);

// #if U_SHOW_CPLUSPLUS_API

// U_NAMESPACE_BEGIN

// /**
//  * \class LocalUBreakIteratorPointer
//  * "Smart pointer" class, closes a UBreakIterator via ubrk_close().
//  * For most methods see the LocalPointerBase base class.
//  *
//  * @see LocalPointerBase
//  * @see LocalPointer
//  * @stable ICU 4.4
//  */
// U_DEFINE_LOCAL_OPEN_POINTER(LocalUBreakIteratorPointer, UBreakIterator, ubrk_close);

// U_NAMESPACE_END

// #endif

// /**
//  * Sets an existing iterator to point to a new piece of text.
//  * The break iterator retains a pointer to the supplied text.
//  * The caller must not modify or delete the text while the BreakIterator
//  * retains the reference.
//  *
//  * @param bi The iterator to use
//  * @param text The text to be set
//  * @param textLength The length of the text
//  * @param status The error code
//  * @stable ICU 2.0
//  */
// U_STABLE void U_EXPORT2
// ubrk_setText(UBreakIterator* bi,
//              const UChar*    text,
//              int32_t         textLength,
//              UErrorCode*     status);


// /**
//  * Sets an existing iterator to point to a new piece of text.
//  *
//  * All index positions returned by break iterator functions are
//  * native indices from the UText. For example, when breaking UTF-8
//  * encoded text, the break positions returned by \ref ubrk_next, \ref ubrk_previous, etc.
//  * will be UTF-8 string indices, not UTF-16 positions.
//  *
//  * @param bi The iterator to use
//  * @param text The text to be set.
//  *             This function makes a shallow clone of the supplied UText.  This means
//  *             that the caller is free to immediately close or otherwise reuse the
//  *             UText that was passed as a parameter, but that the underlying text itself
//  *             must not be altered while being referenced by the break iterator.
//  * @param status The error code
//  * @stable ICU 3.4
//  */
// U_STABLE void U_EXPORT2
// ubrk_setUText(UBreakIterator* bi,
//              UText*          text,
//              UErrorCode*     status);



// /**
//  * Determine the most recently-returned text boundary.
//  *
//  * @param bi The break iterator to use.
//  * @return The character index most recently returned by \ref ubrk_next, \ref ubrk_previous,
//  * \ref ubrk_first, or \ref ubrk_last.
//  * @stable ICU 2.0
//  */
// U_STABLE int32_t U_EXPORT2
// ubrk_current(const UBreakIterator *bi);

// /**
//  * Advance the iterator to the boundary following the current boundary.
//  *
//  * @param bi The break iterator to use.
//  * @return The character index of the next text boundary, or UBRK_DONE
//  * if all text boundaries have been returned.
//  * @see ubrk_previous
//  * @stable ICU 2.0
//  */
// U_STABLE int32_t U_EXPORT2
// ubrk_next(UBreakIterator *bi);

// /**
//  * Set the iterator position to the boundary preceding the current boundary.
//  *
//  * @param bi The break iterator to use.
//  * @return The character index of the preceding text boundary, or UBRK_DONE
//  * if all text boundaries have been returned.
//  * @see ubrk_next
//  * @stable ICU 2.0
//  */
// U_STABLE int32_t U_EXPORT2
// ubrk_previous(UBreakIterator *bi);

// /**
//  * Set the iterator position to zero, the start of the text being scanned.
//  * @param bi The break iterator to use.
//  * @return The new iterator position (zero).
//  * @see ubrk_last
//  * @stable ICU 2.0
//  */
// U_STABLE int32_t U_EXPORT2
// ubrk_first(UBreakIterator *bi);

// /**
//  * Set the iterator position to the index immediately <EM>beyond</EM> the last character in the text being scanned.
//  * This is not the same as the last character.
//  * @param bi The break iterator to use.
//  * @return The character offset immediately <EM>beyond</EM> the last character in the
//  * text being scanned.
//  * @see ubrk_first
//  * @stable ICU 2.0
//  */
// U_STABLE int32_t U_EXPORT2
// ubrk_last(UBreakIterator *bi);

// /**
//  * Set the iterator position to the first boundary preceding the specified offset.
//  * The new position is always smaller than offset, or UBRK_DONE.
//  * @param bi The break iterator to use.
//  * @param offset The offset to begin scanning.
//  * @return The text boundary preceding offset, or UBRK_DONE.
//  * @see ubrk_following
//  * @stable ICU 2.0
//  */
// U_STABLE int32_t U_EXPORT2
// ubrk_preceding(UBreakIterator *bi,
//            int32_t offset);

// /**
//  * Advance the iterator to the first boundary following the specified offset.
//  * The value returned is always greater than offset, or UBRK_DONE.
//  * @param bi The break iterator to use.
//  * @param offset The offset to begin scanning.
//  * @return The text boundary following offset, or UBRK_DONE.
//  * @see ubrk_preceding
//  * @stable ICU 2.0
//  */
// U_STABLE int32_t U_EXPORT2
// ubrk_following(UBreakIterator *bi,
//            int32_t offset);

// /**
// * Get a locale for which text breaking information is available.
// * A UBreakIterator in a locale returned by this function will perform the correct
// * text breaking for the locale.
// * @param index The index of the desired locale.
// * @return A locale for which number text breaking information is available, or 0 if none.
// * @see ubrk_countAvailable
// * @stable ICU 2.0
// */
// U_STABLE const char* U_EXPORT2
// ubrk_getAvailable(int32_t index);

// /**
// * Determine how many locales have text breaking information available.
// * This function is most useful as determining the loop ending condition for
// * calls to \ref ubrk_getAvailable.
// * @return The number of locales for which text breaking information is available.
// * @see ubrk_getAvailable
// * @stable ICU 2.0
// */
// U_STABLE int32_t U_EXPORT2
// ubrk_countAvailable(void);


// /**
// * Returns true if the specified position is a boundary position.  As a side
// * effect, leaves the iterator pointing to the first boundary position at
// * or after "offset".
// * @param bi The break iterator to use.
// * @param offset the offset to check.
// * @return True if "offset" is a boundary position.
// * @stable ICU 2.0
// */
// U_STABLE  UBool U_EXPORT2
// ubrk_isBoundary(UBreakIterator *bi, int32_t offset);

// /**
//  * Return the status from the break rule that determined the most recently
//  * returned break position.  The values appear in the rule source
//  * within brackets, {123}, for example.  For rules that do not specify a
//  * status, a default value of 0 is returned.
//  * <p>
//  * For word break iterators, the possible values are defined in enum UWordBreak.
//  * @stable ICU 2.2
//  */
// U_STABLE  int32_t U_EXPORT2
// ubrk_getRuleStatus(UBreakIterator *bi);

// /**
//  * Get the statuses from the break rules that determined the most recently
//  * returned break position.  The values appear in the rule source
//  * within brackets, {123}, for example.  The default status value for rules
//  * that do not explicitly provide one is zero.
//  * <p>
//  * For word break iterators, the possible values are defined in enum UWordBreak.
//  * @param bi        The break iterator to use
//  * @param fillInVec an array to be filled in with the status values.
//  * @param capacity  the length of the supplied vector.  A length of zero causes
//  *                  the function to return the number of status values, in the
//  *                  normal way, without attempting to store any values.
//  * @param status    receives error codes.
//  * @return          The number of rule status values from rules that determined
//  *                  the most recent boundary returned by the break iterator.
//  * @stable ICU 3.0
//  */
// U_STABLE  int32_t U_EXPORT2
// ubrk_getRuleStatusVec(UBreakIterator *bi, int32_t *fillInVec, int32_t capacity, UErrorCode *status);

// /**
//  * Return the locale of the break iterator. You can choose between the valid and
//  * the actual locale.
//  * @param bi break iterator
//  * @param type locale type (valid or actual)
//  * @param status error code
//  * @return locale string
//  * @stable ICU 2.8
//  */
// U_STABLE const char* U_EXPORT2
// ubrk_getLocaleByType(const UBreakIterator *bi, ULocDataLocaleType type, UErrorCode* status);

// /**
//   *  Set the subject text string upon which the break iterator is operating
//   *  without changing any other aspect of the state.
//   *  The new and previous text strings must have the same content.
//   *
//   *  This function is intended for use in environments where ICU is operating on
//   *  strings that may move around in memory.  It provides a mechanism for notifying
//   *  ICU that the string has been relocated, and providing a new UText to access the
//   *  string in its new position.
//   *
//   *  Note that the break iterator never copies the underlying text
//   *  of a string being processed, but always operates directly on the original text
//   *  provided by the user. Refreshing simply drops the references to the old text
//   *  and replaces them with references to the new.
//   *
//   *  Caution:  this function is normally used only by very specialized
//   *            system-level code.   One example use case is with garbage collection
//   *            that moves the text in memory.
//   *
//   * @param bi         The break iterator.
//   * @param text       The new (moved) text string.
//   * @param status     Receives errors detected by this function.
//   *
//   * @stable ICU 49
//   */
// U_STABLE void U_EXPORT2
// ubrk_refreshUText(UBreakIterator *bi,
//                        UText          *text,
//                        UErrorCode     *status);


// /**
//  * Get a compiled binary version of the rules specifying the behavior of a UBreakIterator.
//  * The binary rules may be used with ubrk_openBinaryRules to open a new UBreakIterator
//  * more quickly than using ubrk_openRules. The compiled rules are not compatible across
//  * different major versions of ICU, nor across platforms of different endianness or
//  * different base character set family (ASCII vs EBCDIC). Supports preflighting (with
//  * binaryRules=NULL and rulesCapacity=0) to get the rules length without copying them to
//  * the binaryRules buffer. However, whether preflighting or not, if the actual length
//  * is greater than INT32_MAX, then the function returns 0 and sets *status to
//  * U_INDEX_OUTOFBOUNDS_ERROR.

//  * @param bi            The break iterator to use.
//  * @param binaryRules   Buffer to receive the compiled binary rules; set to NULL for
//  *                      preflighting.
//  * @param rulesCapacity Capacity (in bytes) of the binaryRules buffer; set to 0 for
//  *                      preflighting. Must be >= 0.
//  * @param status        Pointer to UErrorCode to receive any errors, such as
//  *                      U_BUFFER_OVERFLOW_ERROR, U_INDEX_OUTOFBOUNDS_ERROR, or
//  *                      U_ILLEGAL_ARGUMENT_ERROR.
//  * @return              The actual byte length of the binary rules, if <= INT32_MAX;
//  *                      otherwise 0. If not preflighting and this is larger than
//  *                      rulesCapacity, *status will be set to an error.
//  * @see ubrk_openBinaryRules
//  * @stable ICU 59
//  */
// U_STABLE int32_t U_EXPORT2
// ubrk_getBinaryRules(UBreakIterator *bi,
//                     uint8_t *       binaryRules, int32_t rulesCapacity,
//                     UErrorCode *    status);

// #endif /* #if !UCONFIG_NO_BREAK_ITERATION */

// #endif




#if !UCONFIG_NO_BREAK_ITERATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2004-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  utext.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004oct06
*   created by: Markus W. Scherer
*/


/**
 * \file
 * \brief C API: Abstract Unicode Text API
 *
 * The Text Access API provides a means to allow text that is stored in alternative
 * formats to work with ICU services.  ICU normally operates on text that is
 * stored in UTF-16 format, in (UChar *) arrays for the C APIs or as type
 * UnicodeString for C++ APIs.
 *
 * ICU Text Access allows other formats, such as UTF-8 or non-contiguous
 * UTF-16 strings, to be placed in a UText wrapper and then passed to ICU services.
 *
 * There are three general classes of usage for UText:
 *
 *     Application Level Use.  This is the simplest usage - applications would
 *     use one of the utext_open() functions on their input text, and pass
 *     the resulting UText to the desired ICU service.
 *
 *     Second is usage in ICU Services, such as break iteration, that will need to
 *     operate on input presented to them as a UText.  These implementations
 *     will need to use the iteration and related UText functions to gain
 *     access to the actual text.
 *
 *     The third class of UText users are "text providers."  These are the
 *     UText implementations for the various text storage formats.  An application
 *     or system with a unique text storage format can implement a set of
 *     UText provider functions for that format, which will then allow
 *     ICU services to operate on that format.
 *
 *
 * <em>Iterating over text</em>
 *
 * Here is sample code for a forward iteration over the contents of a UText
 *
 * \code
 *    UChar32  c;
 *    UText    *ut = whatever();
 *
 *    for (c=utext_next32From(ut, 0); c>=0; c=utext_next32(ut)) {
 *       // do whatever with the codepoint c here.
 *    }
 * \endcode
 *
 * And here is similar code to iterate in the reverse direction, from the end
 * of the text towards the beginning.
 *
 * \code
 *    UChar32  c;
 *    UText    *ut = whatever();
 *    int      textLength = utext_nativeLength(ut);
 *    for (c=utext_previous32From(ut, textLength); c>=0; c=utext_previous32(ut)) {
 *       // do whatever with the codepoint c here.
 *    }
 * \endcode
 *
 * <em>Characters and Indexing</em>
 *
 * Indexing into text by UText functions is nearly always in terms of the native
 * indexing of the underlying text storage.  The storage format could be UTF-8
 * or UTF-32, for example.  When coding to the UText access API, no assumptions
 * can be made regarding the size of characters, or how far an index
 * may move when iterating between characters.
 *
 * All indices supplied to UText functions are pinned to the length of the
 * text.  An out-of-bounds index is not considered to be an error, but is
 * adjusted to be in the range  0 <= index <= length of input text.
 *
 *
 * When an index position is returned from a UText function, it will be
 * a native index to the underlying text.  In the case of multi-unit characters,
 * it will  always refer to the first position of the character,
 * never to the interior.  This is essentially the same thing as saying that
 * a returned index will always point to a boundary between characters.
 *
 * When a native index is supplied to a UText function, all indices that
 * refer to any part of a multi-unit character representation are considered
 * to be equivalent.  In the case of multi-unit characters, an incoming index
 * will be logically normalized to refer to the start of the character.
 *
 * It is possible to test whether a native index is on a code point boundary
 * by doing a utext_setNativeIndex() followed by a utext_getNativeIndex().
 * If the index is returned unchanged, it was on a code point boundary.  If
 * an adjusted index is returned, the original index referred to the
 * interior of a character.
 *
 * <em>Conventions for calling UText functions</em>
 *
 * Most UText access functions have as their first parameter a (UText *) pointer,
 * which specifies the UText to be used.  Unless otherwise noted, the
 * pointer must refer to a valid, open UText.  Attempting to
 * use a closed UText or passing a NULL pointer is a programming error and
 * will produce undefined results or NULL pointer exceptions.
 *
 * The UText_Open family of functions can either open an existing (closed)
 * UText, or heap allocate a new UText.  Here is sample code for creating
 * a stack-allocated UText.
 *
 * \code
 *    char     *s = whatever();  // A utf-8 string
 *    U_ErrorCode status = U_ZERO_ERROR;
 *    UText    ut = UTEXT_INITIALIZER;
 *    utext_openUTF8(ut, s, -1, &status);
 *    if (U_FAILURE(status)) {
 *        // error handling
 *    } else {
 *        // work with the UText
 *    }
 * \endcode
 *
 * Any existing UText passed to an open function _must_ have been initialized,
 * either by the UTEXT_INITIALIZER, or by having been originally heap-allocated
 * by an open function.  Passing NULL will cause the open function to
 * heap-allocate and fully initialize a new UText.
 *
 */





// #if U_SHOW_CPLUSPLUS_API




// #endif


U_CDECL_BEGIN

struct UText;
typedef struct UText UText; /**< C typedef for struct UText. @stable ICU 3.6 */


/***************************************************************************************
 *
 *   C Functions for creating UText wrappers around various kinds of text strings.
 *
 ****************************************************************************************/


/**
  * Close function for UText instances.
  * Cleans up, releases any resources being held by an open UText.
  * <p>
  *   If the UText was originally allocated by one of the utext_open functions,
  *   the storage associated with the utext will also be freed.
  *   If the UText storage originated with the application, as it would with
  *   a local or static instance, the storage will not be deleted.
  *
  *   An open UText can be reset to refer to new string by using one of the utext_open()
  *   functions without first closing the UText.
  *
  * @param ut  The UText to be closed.
  * @return    NULL if the UText struct was deleted by the close.  If the UText struct
  *            was originally provided by the caller to the open function, it is
  *            returned by this function, and may be safely used again in
  *            a subsequent utext_open.
  *
  * @stable ICU 3.4
  */
U_STABLE UText * U_EXPORT2
utext_close(UText *ut);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUTextPointer
 * "Smart pointer" class, closes a UText via utext_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUTextPointer, UText, utext_close);

U_NAMESPACE_END

#endif

/**
 * Open a read-only UText implementation for UTF-8 strings.
 *
 * \htmlonly
 * Any invalid UTF-8 in the input will be handled in this way:
 * a sequence of bytes that has the form of a truncated, but otherwise valid,
 * UTF-8 sequence will be replaced by a single unicode replacement character, \uFFFD.
 * Any other illegal bytes will each be replaced by a \uFFFD.
 * \endhtmlonly
 *
 * @param ut     Pointer to a UText struct.  If NULL, a new UText will be created.
 *               If non-NULL, must refer to an initialized UText struct, which will then
 *               be reset to reference the specified UTF-8 string.
 * @param s      A UTF-8 string.  Must not be NULL.
 * @param length The length of the UTF-8 string in bytes, or -1 if the string is
 *               zero terminated.
 * @param status Errors are returned here.
 * @return       A pointer to the UText.  If a pre-allocated UText was provided, it
 *               will always be used and returned.
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openUTF8(UText *ut, const char *s, int64_t length, UErrorCode *status);


/**
 * Open a read-only UText for UChar * string.
 *
 * @param ut     Pointer to a UText struct.  If NULL, a new UText will be created.
 *               If non-NULL, must refer to an initialized UText struct, which will then
 *               be reset to reference the specified UChar string.
 * @param s      A UChar (UTF-16) string
 * @param length The number of UChars in the input string, or -1 if the string is
 *               zero terminated.
 * @param status Errors are returned here.
 * @return       A pointer to the UText.  If a pre-allocated UText was provided, it
 *               will always be used and returned.
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openUChars(UText *ut, const UChar *s, int64_t length, UErrorCode *status);


#if U_SHOW_CPLUSPLUS_API
/**
 * Open a writable UText for a non-const UnicodeString.
 *
 * @param ut      Pointer to a UText struct.  If NULL, a new UText will be created.
 *                 If non-NULL, must refer to an initialized UText struct, which will then
 *                 be reset to reference the specified input string.
 * @param s       A UnicodeString.
 * @param status Errors are returned here.
 * @return        Pointer to the UText.  If a UText was supplied as input, this
 *                 will always be used and returned.
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openUnicodeString(UText *ut, icu::UnicodeString *s, UErrorCode *status);


/**
 * Open a UText for a const UnicodeString.   The resulting UText will not be writable.
 *
 * @param ut    Pointer to a UText struct.  If NULL, a new UText will be created.
 *               If non-NULL, must refer to an initialized UText struct, which will then
 *               be reset to reference the specified input string.
 * @param s      A const UnicodeString to be wrapped.
 * @param status Errors are returned here.
 * @return       Pointer to the UText.  If a UText was supplied as input, this
 *               will always be used and returned.
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openConstUnicodeString(UText *ut, const icu::UnicodeString *s, UErrorCode *status);


/**
 * Open a writable UText implementation for an ICU Replaceable object.
 * @param ut    Pointer to a UText struct.  If NULL, a new UText will be created.
 *               If non-NULL, must refer to an already existing UText, which will then
 *               be reset to reference the specified replaceable text.
 * @param rep    A Replaceable text object.
 * @param status Errors are returned here.
 * @return       Pointer to the UText.  If a UText was supplied as input, this
 *               will always be used and returned.
 * @see Replaceable
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openReplaceable(UText *ut, icu::Replaceable *rep, UErrorCode *status);

/**
 * Open a  UText implementation over an ICU CharacterIterator.
 * @param ut    Pointer to a UText struct.  If NULL, a new UText will be created.
 *               If non-NULL, must refer to an already existing UText, which will then
 *               be reset to reference the specified replaceable text.
 * @param ci     A Character Iterator.
 * @param status Errors are returned here.
 * @return       Pointer to the UText.  If a UText was supplied as input, this
 *               will always be used and returned.
 * @see Replaceable
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_openCharacterIterator(UText *ut, icu::CharacterIterator *ci, UErrorCode *status);

#endif


/**
  *  Clone a UText.  This is much like opening a UText where the source text is itself
  *  another UText.
  *
  *  A deep clone will copy both the UText data structures and the underlying text.
  *  The original and cloned UText will operate completely independently; modifications
  *  made to the text in one will not affect the other.  Text providers are not
  *  required to support deep clones.  The user of clone() must check the status return
  *  and be prepared to handle failures.
  *
  *  The standard UText implementations for UTF8, UChar *, UnicodeString and
  *  Replaceable all support deep cloning.
  *
  *  The UText returned from a deep clone will be writable, assuming that the text
  *  provider is able to support writing, even if the source UText had been made
  *  non-writable by means of UText_freeze().
  *
  *  A shallow clone replicates only the UText data structures; it does not make
  *  a copy of the underlying text.  Shallow clones can be used as an efficient way to
  *  have multiple iterators active in a single text string that is not being
  *  modified.
  *
  *  A shallow clone operation will not fail, barring truly exceptional conditions such
  *  as memory allocation failures.
  *
  *  Shallow UText clones should be avoided if the UText functions that modify the
  *  text are expected to be used, either on the original or the cloned UText.
  *  Any such modifications  can cause unpredictable behavior.  Read Only
  *  shallow clones provide some protection against errors of this type by
  *  disabling text modification via the cloned UText.
  *
  *  A shallow clone made with the readOnly parameter == FALSE will preserve the
  *  utext_isWritable() state of the source object.  Note, however, that
  *  write operations must be avoided while more than one UText exists that refer
  *  to the same underlying text.
  *
  *  A UText and its clone may be safely concurrently accessed by separate threads.
  *  This is true for read access only with shallow clones, and for both read and
  *  write access with deep clones.
  *  It is the responsibility of the Text Provider to ensure that this thread safety
  *  constraint is met.
  *
  *  @param dest   A UText struct to be filled in with the result of the clone operation,
  *                or NULL if the clone function should heap-allocate a new UText struct.
  *                If non-NULL, must refer to an already existing UText, which will then
  *                be reset to become the clone.
  *  @param src    The UText to be cloned.
  *  @param deep   TRUE to request a deep clone, FALSE for a shallow clone.
  *  @param readOnly TRUE to request that the cloned UText have read only access to the
  *                underlying text.

  *  @param status Errors are returned here.  For deep clones, U_UNSUPPORTED_ERROR
  *                will be returned if the text provider is unable to clone the
  *                original text.
  *  @return       The newly created clone, or NULL if the clone operation failed.
  *  @stable ICU 3.4
  */
U_STABLE UText * U_EXPORT2
utext_clone(UText *dest, const UText *src, UBool deep, UBool readOnly, UErrorCode *status);


/**
  *  Compare two UText objects for equality.
  *  UTexts are equal if they are iterating over the same text, and
  *    have the same iteration position within the text.
  *    If either or both of the parameters are NULL, the comparison is FALSE.
  *
  *  @param a   The first of the two UTexts to compare.
  *  @param b   The other UText to be compared.
  *  @return    TRUE if the two UTexts are equal.
  *  @stable ICU 3.6
  */
U_STABLE UBool U_EXPORT2
utext_equals(const UText *a, const UText *b);


/*****************************************************************************
 *
 *   Functions to work with the text represented by a UText wrapper
 *
 *****************************************************************************/

/**
  * Get the length of the text.  Depending on the characteristics
  * of the underlying text representation, this may be expensive.
  * @see  utext_isLengthExpensive()
  *
  *
  * @param ut  the text to be accessed.
  * @return the length of the text, expressed in native units.
  *
  * @stable ICU 3.4
  */
U_STABLE int64_t U_EXPORT2
utext_nativeLength(UText *ut);

/**
 *  Return TRUE if calculating the length of the text could be expensive.
 *  Finding the length of NUL terminated strings is considered to be expensive.
 *
 *  Note that the value of this function may change
 *  as the result of other operations on a UText.
 *  Once the length of a string has been discovered, it will no longer
 *  be expensive to report it.
 *
 * @param ut the text to be accessed.
 * @return TRUE if determining the length of the text could be time consuming.
 * @stable ICU 3.4
 */
U_STABLE UBool U_EXPORT2
utext_isLengthExpensive(const UText *ut);

/**
 * Returns the code point at the requested index,
 * or U_SENTINEL (-1) if it is out of bounds.
 *
 * If the specified index points to the interior of a multi-unit
 * character - one of the trail bytes of a UTF-8 sequence, for example -
 * the complete code point will be returned.
 *
 * The iteration position will be set to the start of the returned code point.
 *
 * This function is roughly equivalent to the sequence
 *    utext_setNativeIndex(index);
 *    utext_current32();
 * (There is a subtle difference if the index is out of bounds by being less than zero -
 * utext_setNativeIndex(negative value) sets the index to zero, after which utext_current()
 * will return the char at zero.  utext_char32At(negative index), on the other hand, will
 * return the U_SENTINEL value of -1.)
 *
 * @param ut the text to be accessed
 * @param nativeIndex the native index of the character to be accessed.  If the index points
 *        to other than the first unit of a multi-unit character, it will be adjusted
 *        to the start of the character.
 * @return the code point at the specified index.
 * @stable ICU 3.4
 */
U_STABLE UChar32 U_EXPORT2
utext_char32At(UText *ut, int64_t nativeIndex);


/**
 *
 * Get the code point at the current iteration position,
 * or U_SENTINEL (-1) if the iteration has reached the end of
 * the input text.
 *
 * @param ut the text to be accessed.
 * @return the Unicode code point at the current iterator position.
 * @stable ICU 3.4
 */
U_STABLE UChar32 U_EXPORT2
utext_current32(UText *ut);


/**
 * Get the code point at the current iteration position of the UText, and
 * advance the position to the first index following the character.
 *
 * If the position is at the end of the text (the index following
 * the last character, which is also the length of the text),
 * return U_SENTINEL (-1) and do not advance the index.
 *
 * This is a post-increment operation.
 *
 * An inline macro version of this function, UTEXT_NEXT32(),
 * is available for performance critical use.
 *
 * @param ut the text to be accessed.
 * @return the Unicode code point at the iteration position.
 * @see UTEXT_NEXT32
 * @stable ICU 3.4
 */
U_STABLE UChar32 U_EXPORT2
utext_next32(UText *ut);


/**
 *  Move the iterator position to the character (code point) whose
 *  index precedes the current position, and return that character.
 *  This is a pre-decrement operation.
 *
 *  If the initial position is at the start of the text (index of 0)
 *  return U_SENTINEL (-1), and leave the position unchanged.
 *
 *  An inline macro version of this function, UTEXT_PREVIOUS32(),
 *  is available for performance critical use.
 *
 *  @param ut the text to be accessed.
 *  @return the previous UChar32 code point, or U_SENTINEL (-1)
 *          if the iteration has reached the start of the text.
 *  @see UTEXT_PREVIOUS32
 *  @stable ICU 3.4
 */
U_STABLE UChar32 U_EXPORT2
utext_previous32(UText *ut);


/**
  * Set the iteration index and return the code point at that index.
  * Leave the iteration index at the start of the following code point.
  *
  * This function is the most efficient and convenient way to
  * begin a forward iteration.  The results are identical to the those
  * from the sequence
  * \code
  *    utext_setIndex();
  *    utext_next32();
  * \endcode
  *
  *  @param ut the text to be accessed.
  *  @param nativeIndex Iteration index, in the native units of the text provider.
  *  @return Code point which starts at or before index,
  *         or U_SENTINEL (-1) if it is out of bounds.
  * @stable ICU 3.4
  */
U_STABLE UChar32 U_EXPORT2
utext_next32From(UText *ut, int64_t nativeIndex);



/**
  * Set the iteration index, and return the code point preceding the
  * one specified by the initial index.  Leave the iteration position
  * at the start of the returned code point.
  *
  * This function is the most efficient and convenient way to
  * begin a backwards iteration.
  *
  * @param ut the text to be accessed.
  * @param nativeIndex Iteration index in the native units of the text provider.
  * @return Code point preceding the one at the initial index,
  *         or U_SENTINEL (-1) if it is out of bounds.
  *
  * @stable ICU 3.4
  */
U_STABLE UChar32 U_EXPORT2
utext_previous32From(UText *ut, int64_t nativeIndex);

/**
  * Get the current iterator position, which can range from 0 to
  * the length of the text.
  * The position is a native index into the input text, in whatever format it
  * may have (possibly UTF-8 for example), and may not always be the same as
  * the corresponding UChar (UTF-16) index.
  * The returned position will always be aligned to a code point boundary.
  *
  * @param ut the text to be accessed.
  * @return the current index position, in the native units of the text provider.
  * @stable ICU 3.4
  */
U_STABLE int64_t U_EXPORT2
utext_getNativeIndex(const UText *ut);

/**
 * Set the current iteration position to the nearest code point
 * boundary at or preceding the specified index.
 * The index is in the native units of the original input text.
 * If the index is out of range, it will be pinned to be within
 * the range of the input text.
 * <p>
 * It will usually be more efficient to begin an iteration
 * using the functions utext_next32From() or utext_previous32From()
 * rather than setIndex().
 * <p>
 * Moving the index position to an adjacent character is best done
 * with utext_next32(), utext_previous32() or utext_moveIndex32().
 * Attempting to do direct arithmetic on the index position is
 * complicated by the fact that the size (in native units) of a
 * character depends on the underlying representation of the character
 * (UTF-8, UTF-16, UTF-32, arbitrary codepage), and is not
 * easily knowable.
 *
 * @param ut the text to be accessed.
 * @param nativeIndex the native unit index of the new iteration position.
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
utext_setNativeIndex(UText *ut, int64_t nativeIndex);

/**
 * Move the iterator position by delta code points.  The number of code points
 * is a signed number; a negative delta will move the iterator backwards,
 * towards the start of the text.
 * <p>
 * The index is moved by <code>delta</code> code points
 * forward or backward, but no further backward than to 0 and
 * no further forward than to utext_nativeLength().
 * The resulting index value will be in between 0 and length, inclusive.
 *
 * @param ut the text to be accessed.
 * @param delta the signed number of code points to move the iteration position.
 * @return TRUE if the position could be moved the requested number of positions while
 *              staying within the range [0 - text length].
 * @stable ICU 3.4
 */
U_STABLE UBool U_EXPORT2
utext_moveIndex32(UText *ut, int32_t delta);

/**
 * Get the native index of the character preceding the current position.
 * If the iteration position is already at the start of the text, zero
 * is returned.
 * The value returned is the same as that obtained from the following sequence,
 * but without the side effect of changing the iteration position.
 *
 * \code
 *    UText  *ut = whatever;
 *      ...
 *    utext_previous(ut)
 *    utext_getNativeIndex(ut);
 * \endcode
 *
 * This function is most useful during forwards iteration, where it will get the
 *   native index of the character most recently returned from utext_next().
 *
 * @param ut the text to be accessed
 * @return the native index of the character preceding the current index position,
 *         or zero if the current position is at the start of the text.
 * @stable ICU 3.6
 */
U_STABLE int64_t U_EXPORT2
utext_getPreviousNativeIndex(UText *ut);


/**
 *
 * Extract text from a UText into a UChar buffer.  The range of text to be extracted
 * is specified in the native indices of the UText provider.  These may not necessarily
 * be UTF-16 indices.
 * <p>
 * The size (number of 16 bit UChars) of the data to be extracted is returned.  The
 * full number of UChars is returned, even when the extracted text is truncated
 * because the specified buffer size is too small.
 * <p>
 * The extracted string will (if you are a user) / must (if you are a text provider)
 * be NUL-terminated if there is sufficient space in the destination buffer.  This
 * terminating NUL is not included in the returned length.
 * <p>
 * The iteration index is left at the position following the last extracted character.
 *
 * @param  ut    the UText from which to extract data.
 * @param  nativeStart the native index of the first character to extract.\
 *               If the specified index is out of range,
 *               it will be pinned to be within 0 <= index <= textLength
 * @param  nativeLimit the native string index of the position following the last
 *               character to extract.  If the specified index is out of range,
 *               it will be pinned to be within 0 <= index <= textLength.
 *               nativeLimit must be >= nativeStart.
 * @param  dest  the UChar (UTF-16) buffer into which the extracted text is placed
 * @param  destCapacity  The size, in UChars, of the destination buffer.  May be zero
 *               for precomputing the required size.
 * @param  status receives any error status.
 *         U_BUFFER_OVERFLOW_ERROR: the extracted text was truncated because the
 *         buffer was too small.  Returns number of UChars for preflighting.
 * @return Number of UChars in the data to be extracted.  Does not include a trailing NUL.
 *
 * @stable ICU 3.4
 */
U_STABLE int32_t U_EXPORT2
utext_extract(UText *ut,
             int64_t nativeStart, int64_t nativeLimit,
             UChar *dest, int32_t destCapacity,
             UErrorCode *status);



/************************************************************************************
 *
 *  #define inline versions of selected performance-critical text access functions
 *          Caution:  do not use auto increment++ or decrement-- expressions
 *                    as parameters to these macros.
 *
 *          For most use, where there is no extreme performance constraint, the
 *          normal, non-inline functions are a better choice.  The resulting code
 *          will be smaller, and, if the need ever arises, easier to debug.
 *
 *          These are implemented as #defines rather than real functions
 *          because there is no fully portable way to do inline functions in plain C.
 *
 ************************************************************************************/

#ifndef U_HIDE_INTERNAL_API
/**
 * inline version of utext_current32(), for performance-critical situations.
 *
 * Get the code point at the current iteration position of the UText.
 * Returns U_SENTINEL (-1) if the position is at the end of the
 * text.
 *
 * @internal ICU 4.4 technology preview
 */
#define UTEXT_CURRENT32(ut)  \
    ((ut)->chunkOffset < (ut)->chunkLength && ((ut)->chunkContents)[(ut)->chunkOffset]<0xd800 ? \
    ((ut)->chunkContents)[((ut)->chunkOffset)] : utext_current32(ut))
#endif  /* U_HIDE_INTERNAL_API */

/**
 * inline version of utext_next32(), for performance-critical situations.
 *
 * Get the code point at the current iteration position of the UText, and
 * advance the position to the first index following the character.
 * This is a post-increment operation.
 * Returns U_SENTINEL (-1) if the position is at the end of the
 * text.
 *
 * @stable ICU 3.4
 */
#define UTEXT_NEXT32(ut)  \
    ((ut)->chunkOffset < (ut)->chunkLength && ((ut)->chunkContents)[(ut)->chunkOffset]<0xd800 ? \
    ((ut)->chunkContents)[((ut)->chunkOffset)++] : utext_next32(ut))

/**
 * inline version of utext_previous32(), for performance-critical situations.
 *
 *  Move the iterator position to the character (code point) whose
 *  index precedes the current position, and return that character.
 *  This is a pre-decrement operation.
 *  Returns U_SENTINEL (-1) if the position is at the start of the  text.
 *
 * @stable ICU 3.4
 */
#define UTEXT_PREVIOUS32(ut)  \
    ((ut)->chunkOffset > 0 && \
     (ut)->chunkContents[(ut)->chunkOffset-1] < 0xd800 ? \
          (ut)->chunkContents[--((ut)->chunkOffset)]  :  utext_previous32(ut))

/**
  *  inline version of utext_getNativeIndex(), for performance-critical situations.
  *
  * Get the current iterator position, which can range from 0 to
  * the length of the text.
  * The position is a native index into the input text, in whatever format it
  * may have (possibly UTF-8 for example), and may not always be the same as
  * the corresponding UChar (UTF-16) index.
  * The returned position will always be aligned to a code point boundary.
  *
  * @stable ICU 3.6
  */
#define UTEXT_GETNATIVEINDEX(ut)                       \
    ((ut)->chunkOffset <= (ut)->nativeIndexingLimit?   \
        (ut)->chunkNativeStart+(ut)->chunkOffset :     \
        (ut)->pFuncs->mapOffsetToNative(ut))

/**
  *  inline version of utext_setNativeIndex(), for performance-critical situations.
  *
  * Set the current iteration position to the nearest code point
  * boundary at or preceding the specified index.
  * The index is in the native units of the original input text.
  * If the index is out of range, it will be pinned to be within
  * the range of the input text.
  *
  * @stable ICU 3.8
  */
#define UTEXT_SETNATIVEINDEX(ut, ix) UPRV_BLOCK_MACRO_BEGIN { \
    int64_t __offset = (ix) - (ut)->chunkNativeStart; \
    if (__offset>=0 && __offset<(int64_t)(ut)->nativeIndexingLimit && (ut)->chunkContents[__offset]<0xdc00) { \
        (ut)->chunkOffset=(int32_t)__offset; \
    } else { \
        utext_setNativeIndex((ut), (ix)); \
    } \
} UPRV_BLOCK_MACRO_END



/************************************************************************************
 *
 *   Functions related to writing or modifying the text.
 *   These will work only with modifiable UTexts.  Attempting to
 *   modify a read-only UText will return an error status.
 *
 ************************************************************************************/


/**
 *  Return TRUE if the text can be written (modified) with utext_replace() or
 *  utext_copy().  For the text to be writable, the text provider must
 *  be of a type that supports writing and the UText must not be frozen.
 *
 *  Attempting to modify text when utext_isWriteable() is FALSE will fail -
 *  the text will not be modified, and an error will be returned from the function
 *  that attempted the modification.
 *
 * @param  ut   the UText to be tested.
 * @return TRUE if the text is modifiable.
 *
 * @see    utext_freeze()
 * @see    utext_replace()
 * @see    utext_copy()
 * @stable ICU 3.4
 *
 */
U_STABLE UBool U_EXPORT2
utext_isWritable(const UText *ut);


/**
  * Test whether there is meta data associated with the text.
  * @see Replaceable::hasMetaData()
  *
  * @param ut The UText to be tested
  * @return TRUE if the underlying text includes meta data.
  * @stable ICU 3.4
  */
U_STABLE UBool U_EXPORT2
utext_hasMetaData(const UText *ut);


/**
 * Replace a range of the original text with a replacement text.
 *
 * Leaves the current iteration position at the position following the
 *  newly inserted replacement text.
 *
 * This function is only available on UText types that support writing,
 * that is, ones where utext_isWritable() returns TRUE.
 *
 * When using this function, there should be only a single UText opened onto the
 * underlying native text string.  Behavior after a replace operation
 * on a UText is undefined for any other additional UTexts that refer to the
 * modified string.
 *
 * @param ut               the UText representing the text to be operated on.
 * @param nativeStart      the native index of the start of the region to be replaced
 * @param nativeLimit      the native index of the character following the region to be replaced.
 * @param replacementText  pointer to the replacement text
 * @param replacementLength length of the replacement text, or -1 if the text is NUL terminated.
 * @param status           receives any error status.  Possible errors include
 *                         U_NO_WRITE_PERMISSION
 *
 * @return The signed number of (native) storage units by which
 *         the length of the text expanded or contracted.
 *
 * @stable ICU 3.4
 */
U_STABLE int32_t U_EXPORT2
utext_replace(UText *ut,
             int64_t nativeStart, int64_t nativeLimit,
             const UChar *replacementText, int32_t replacementLength,
             UErrorCode *status);



/**
 *
 * Copy or move a substring from one position to another within the text,
 * while retaining any metadata associated with the text.
 * This function is used to duplicate or reorder substrings.
 * The destination index must not overlap the source range.
 *
 * The text to be copied or moved is inserted at destIndex;
 * it does not replace or overwrite any existing text.
 *
 * The iteration position is left following the newly inserted text
 * at the destination position.
 *
 * This function is only available on UText types that support writing,
 * that is, ones where utext_isWritable() returns TRUE.
 *
 * When using this function, there should be only a single UText opened onto the
 * underlying native text string.  Behavior after a copy operation
 * on a UText is undefined in any other additional UTexts that refer to the
 * modified string.
 *
 * @param ut           The UText representing the text to be operated on.
 * @param nativeStart  The native index of the start of the region to be copied or moved
 * @param nativeLimit  The native index of the character position following the region
 *                     to be copied.
 * @param destIndex    The native destination index to which the source substring is
 *                     copied or moved.
 * @param move         If TRUE, then the substring is moved, not copied/duplicated.
 * @param status       receives any error status.  Possible errors include U_NO_WRITE_PERMISSION
 *
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
utext_copy(UText *ut,
          int64_t nativeStart, int64_t nativeLimit,
          int64_t destIndex,
          UBool move,
          UErrorCode *status);


/**
  *  <p>
  *  Freeze a UText.  This prevents any modification to the underlying text itself
  *  by means of functions operating on this UText.
  *  </p>
  *  <p>
  *  Once frozen, a UText can not be unfrozen.  The intent is to ensure
  *  that a the text underlying a frozen UText wrapper cannot be modified via that UText.
  *  </p>
  *  <p>
  *  Caution:  freezing a UText will disable changes made via the specific
  *   frozen UText wrapper only; it will not have any effect on the ability to
  *   directly modify the text by bypassing the UText.  Any such backdoor modifications
  *   are always an error while UText access is occurring because the underlying
  *   text can get out of sync with UText's buffering.
  *  </p>
  *
  *  @param ut  The UText to be frozen.
  *  @see   utext_isWritable()
  *  @stable ICU 3.6
  */
U_STABLE void U_EXPORT2
utext_freeze(UText *ut);


/**
 * UText provider properties (bit field indexes).
 *
 * @see UText
 * @stable ICU 3.4
 */
enum {
    /**
     * It is potentially time consuming for the provider to determine the length of the text.
     * @stable ICU 3.4
     */
    UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE = 1,
    /**
     * Text chunks remain valid and usable until the text object is modified or
     * deleted, not just until the next time the access() function is called
     * (which is the default).
     * @stable ICU 3.4
     */
    UTEXT_PROVIDER_STABLE_CHUNKS = 2,
    /**
     * The provider supports modifying the text via the replace() and copy()
     * functions.
     * @see Replaceable
     * @stable ICU 3.4
     */
    UTEXT_PROVIDER_WRITABLE = 3,
    /**
     * There is meta data associated with the text.
     * @see Replaceable::hasMetaData()
     * @stable ICU 3.4
     */
    UTEXT_PROVIDER_HAS_META_DATA = 4,
    /**
     * Text provider owns the text storage.
     *  Generally occurs as the result of a deep clone of the UText.
     *  When closing the UText, the associated text must
     *  also be closed/deleted/freed/ whatever is appropriate.
     * @stable ICU 3.6
     */
     UTEXT_PROVIDER_OWNS_TEXT = 5
};

/**
  * Function type declaration for UText.clone().
  *
  *  clone a UText.  Much like opening a UText where the source text is itself
  *  another UText.
  *
  *  A deep clone will copy both the UText data structures and the underlying text.
  *  The original and cloned UText will operate completely independently; modifications
  *  made to the text in one will not effect the other.  Text providers are not
  *  required to support deep clones.  The user of clone() must check the status return
  *  and be prepared to handle failures.
  *
  *  A shallow clone replicates only the UText data structures; it does not make
  *  a copy of the underlying text.  Shallow clones can be used as an efficient way to
  *  have multiple iterators active in a single text string that is not being
  *  modified.
  *
  *  A shallow clone operation must not fail except for truly exceptional conditions such
  *  as memory allocation failures.
  *
  *  A UText and its clone may be safely concurrently accessed by separate threads.
  *  This is true for both shallow and deep clones.
  *  It is the responsibility of the Text Provider to ensure that this thread safety
  *  constraint is met.

  *
  *  @param dest   A UText struct to be filled in with the result of the clone operation,
  *                or NULL if the clone function should heap-allocate a new UText struct.
  *  @param src    The UText to be cloned.
  *  @param deep   TRUE to request a deep clone, FALSE for a shallow clone.
  *  @param status Errors are returned here.  For deep clones, U_UNSUPPORTED_ERROR
  *                should be returned if the text provider is unable to clone the
  *                original text.
  *  @return       The newly created clone, or NULL if the clone operation failed.
  *
  * @stable ICU 3.4
  */
typedef UText * U_CALLCONV
UTextClone(UText *dest, const UText *src, UBool deep, UErrorCode *status);


/**
 * Function type declaration for UText.nativeLength().
 *
 * @param ut the UText to get the length of.
 * @return the length, in the native units of the original text string.
 * @see UText
 * @stable ICU 3.4
 */
typedef int64_t U_CALLCONV
UTextNativeLength(UText *ut);

/**
 * Function type declaration for UText.access().  Get the description of the text chunk
 *  containing the text at a requested native index.  The UText's iteration
 *  position will be left at the requested index.  If the index is out
 *  of bounds, the iteration position will be left at the start or end
 *  of the string, as appropriate.
 *
 *  Chunks must begin and end on code point boundaries.  A single code point
 *  comprised of multiple storage units must never span a chunk boundary.
 *
 *
 * @param ut          the UText being accessed.
 * @param nativeIndex Requested index of the text to be accessed.
 * @param forward     If TRUE, then the returned chunk must contain text
 *                    starting from the index, so that start<=index<limit.
 *                    If FALSE, then the returned chunk must contain text
 *                    before the index, so that start<index<=limit.
 * @return            True if the requested index could be accessed.  The chunk
 *                    will contain the requested text.
 *                    False value if a chunk cannot be accessed
 *                    (the requested index is out of bounds).
 *
 * @see UText
 * @stable ICU 3.4
 */
typedef UBool U_CALLCONV
UTextAccess(UText *ut, int64_t nativeIndex, UBool forward);

/**
 * Function type declaration for UText.extract().
 *
 * Extract text from a UText into a UChar buffer.  The range of text to be extracted
 * is specified in the native indices of the UText provider.  These may not necessarily
 * be UTF-16 indices.
 * <p>
 * The size (number of 16 bit UChars) in the data to be extracted is returned.  The
 * full amount is returned, even when the specified buffer size is smaller.
 * <p>
 * The extracted string will (if you are a user) / must (if you are a text provider)
 * be NUL-terminated if there is sufficient space in the destination buffer.
 *
 * @param  ut            the UText from which to extract data.
 * @param  nativeStart   the native index of the first character to extract.
 * @param  nativeLimit   the native string index of the position following the last
 *                       character to extract.
 * @param  dest          the UChar (UTF-16) buffer into which the extracted text is placed
 * @param  destCapacity  The size, in UChars, of the destination buffer.  May be zero
 *                       for precomputing the required size.
 * @param  status        receives any error status.
 *                       If U_BUFFER_OVERFLOW_ERROR: Returns number of UChars for
 *                       preflighting.
 * @return Number of UChars in the data.  Does not include a trailing NUL.
 *
 * @stable ICU 3.4
 */
typedef int32_t U_CALLCONV
UTextExtract(UText *ut,
             int64_t nativeStart, int64_t nativeLimit,
             UChar *dest, int32_t destCapacity,
             UErrorCode *status);

/**
 * Function type declaration for UText.replace().
 *
 * Replace a range of the original text with a replacement text.
 *
 * Leaves the current iteration position at the position following the
 *  newly inserted replacement text.
 *
 * This function need only be implemented on UText types that support writing.
 *
 * When using this function, there should be only a single UText opened onto the
 * underlying native text string.  The function is responsible for updating the
 * text chunk within the UText to reflect the updated iteration position,
 * taking into account any changes to the underlying string's structure caused
 * by the replace operation.
 *
 * @param ut               the UText representing the text to be operated on.
 * @param nativeStart      the index of the start of the region to be replaced
 * @param nativeLimit      the index of the character following the region to be replaced.
 * @param replacementText  pointer to the replacement text
 * @param replacmentLength length of the replacement text in UChars, or -1 if the text is NUL terminated.
 * @param status           receives any error status.  Possible errors include
 *                         U_NO_WRITE_PERMISSION
 *
 * @return The signed number of (native) storage units by which
 *         the length of the text expanded or contracted.
 *
 * @stable ICU 3.4
 */
typedef int32_t U_CALLCONV
UTextReplace(UText *ut,
             int64_t nativeStart, int64_t nativeLimit,
             const UChar *replacementText, int32_t replacmentLength,
             UErrorCode *status);

/**
 * Function type declaration for UText.copy().
 *
 * Copy or move a substring from one position to another within the text,
 * while retaining any metadata associated with the text.
 * This function is used to duplicate or reorder substrings.
 * The destination index must not overlap the source range.
 *
 * The text to be copied or moved is inserted at destIndex;
 * it does not replace or overwrite any existing text.
 *
 * This function need only be implemented for UText types that support writing.
 *
 * When using this function, there should be only a single UText opened onto the
 * underlying native text string.  The function is responsible for updating the
 * text chunk within the UText to reflect the updated iteration position,
 * taking into account any changes to the underlying string's structure caused
 * by the replace operation.
 *
 * @param ut           The UText representing the text to be operated on.
 * @param nativeStart  The index of the start of the region to be copied or moved
 * @param nativeLimit  The index of the character following the region to be replaced.
 * @param nativeDest   The destination index to which the source substring is copied or moved.
 * @param move         If TRUE, then the substring is moved, not copied/duplicated.
 * @param status       receives any error status.  Possible errors include U_NO_WRITE_PERMISSION
 *
 * @stable ICU 3.4
 */
typedef void U_CALLCONV
UTextCopy(UText *ut,
          int64_t nativeStart, int64_t nativeLimit,
          int64_t nativeDest,
          UBool move,
          UErrorCode *status);

/**
 * Function type declaration for UText.mapOffsetToNative().
 * Map from the current UChar offset within the current text chunk to
 *  the corresponding native index in the original source text.
 *
 * This is required only for text providers that do not use native UTF-16 indexes.
 *
 * @param ut     the UText.
 * @return Absolute (native) index corresponding to chunkOffset in the current chunk.
 *         The returned native index should always be to a code point boundary.
 *
 * @stable ICU 3.4
 */
typedef int64_t U_CALLCONV
UTextMapOffsetToNative(const UText *ut);

/**
 * Function type declaration for UText.mapIndexToUTF16().
 * Map from a native index to a UChar offset within a text chunk.
 * Behavior is undefined if the native index does not fall within the
 *   current chunk.
 *
 * This function is required only for text providers that do not use native UTF-16 indexes.
 *
 * @param ut          The UText containing the text chunk.
 * @param nativeIndex Absolute (native) text index, chunk->start<=index<=chunk->limit.
 * @return            Chunk-relative UTF-16 offset corresponding to the specified native
 *                    index.
 *
 * @stable ICU 3.4
 */
typedef int32_t U_CALLCONV
UTextMapNativeIndexToUTF16(const UText *ut, int64_t nativeIndex);


/**
 * Function type declaration for UText.utextClose().
 *
 * A Text Provider close function is only required for provider types that make
 *  allocations in their open function (or other functions) that must be
 *  cleaned when the UText is closed.
 *
 * The allocation of the UText struct itself and any "extra" storage
 * associated with the UText is handled by the common UText implementation
 * and does not require provider specific cleanup in a close function.
 *
 * Most UText provider implementations do not need to implement this function.
 *
 * @param ut A UText object to be closed.
 *
 * @stable ICU 3.4
 */
typedef void U_CALLCONV
UTextClose(UText *ut);


/**
  *   (public)  Function dispatch table for UText.
  *             Conceptually very much like a C++ Virtual Function Table.
  *             This struct defines the organization of the table.
  *             Each text provider implementation must provide an
  *              actual table that is initialized with the appropriate functions
  *              for the type of text being handled.
  *   @stable ICU 3.6
  */
struct UTextFuncs {
    /**
     *   (public)  Function table size, sizeof(UTextFuncs)
     *             Intended for use should the table grow to accommodate added
     *             functions in the future, to allow tests for older format
     *             function tables that do not contain the extensions.
     *
     *             Fields are placed for optimal alignment on
     *             32/64/128-bit-pointer machines, by normally grouping together
     *             4 32-bit fields,
     *             4 pointers,
     *             2 64-bit fields
     *             in sequence.
     *   @stable ICU 3.6
     */
    int32_t       tableSize;

    /**
      *   (private)  Alignment padding.
      *              Do not use, reserved for use by the UText framework only.
      *   @internal
      */
    int32_t       reserved1, /** @internal */ reserved2, /** @internal */ reserved3;


    /**
     * (public) Function pointer for UTextClone
     *
     * @see UTextClone
     * @stable ICU 3.6
     */
    UTextClone *clone;

    /**
     * (public) function pointer for UTextLength
     * May be expensive to compute!
     *
     * @see UTextLength
     * @stable ICU 3.6
     */
    UTextNativeLength *nativeLength;

    /**
     * (public) Function pointer for UTextAccess.
     *
     * @see UTextAccess
     * @stable ICU 3.6
     */
    UTextAccess *access;

    /**
     * (public) Function pointer for UTextExtract.
     *
     * @see UTextExtract
     * @stable ICU 3.6
     */
    UTextExtract *extract;

    /**
     * (public) Function pointer for UTextReplace.
     *
     * @see UTextReplace
     * @stable ICU 3.6
     */
    UTextReplace *replace;

    /**
     * (public) Function pointer for UTextCopy.
     *
     * @see UTextCopy
     * @stable ICU 3.6
     */
    UTextCopy *copy;

    /**
     * (public) Function pointer for UTextMapOffsetToNative.
     *
     * @see UTextMapOffsetToNative
     * @stable ICU 3.6
     */
    UTextMapOffsetToNative *mapOffsetToNative;

    /**
     * (public) Function pointer for UTextMapNativeIndexToUTF16.
     *
     * @see UTextMapNativeIndexToUTF16
     * @stable ICU 3.6
     */
    UTextMapNativeIndexToUTF16 *mapNativeIndexToUTF16;

    /**
     * (public) Function pointer for UTextClose.
      *
      * @see UTextClose
      * @stable ICU 3.6
      */
    UTextClose  *close;

    /**
      * (private)  Spare function pointer
      * @internal
      */
    UTextClose  *spare1;

    /**
      * (private)  Spare function pointer
      * @internal
      */
    UTextClose  *spare2;

    /**
      * (private)  Spare function pointer
      * @internal
      */
    UTextClose  *spare3;

};
/**
 * Function dispatch table for UText
 * @see UTextFuncs
 */
typedef struct UTextFuncs UTextFuncs;

 /**
  *   UText struct.  Provides the interface between the generic UText access code
  *                  and the UText provider code that works on specific kinds of
  *                  text  (UTF-8, noncontiguous UTF-16, whatever.)
  *
  *                  Applications that are using predefined types of text providers
  *                  to pass text data to ICU services will have no need to view the
  *                  internals of the UText structs that they open.
  *
  * @stable ICU 3.6
  */
struct UText {
    /**
     *     (private)  Magic.  Used to help detect when UText functions are handed
     *                        invalid or uninitialized UText structs.
     *                        utext_openXYZ() functions take an initialized,
     *                        but not necessarily open, UText struct as an
     *                        optional fill-in parameter.  This magic field
     *                        is used to check for that initialization.
     *                        Text provider close functions must NOT clear
     *                        the magic field because that would prevent
     *                        reuse of the UText struct.
     * @internal
     */
    uint32_t       magic;


    /**
     *     (private)  Flags for managing the allocation and freeing of
     *                memory associated with this UText.
     * @internal
     */
    int32_t        flags;


    /**
      *  Text provider properties.  This set of flags is maintained by the
      *                             text provider implementation.
      *  @stable ICU 3.4
      */
    int32_t         providerProperties;

    /**
     * (public) sizeOfStruct=sizeof(UText)
     * Allows possible backward compatible extension.
     *
     * @stable ICU 3.4
     */
    int32_t         sizeOfStruct;

    /* ------ 16 byte alignment boundary -----------  */


    /**
      *  (protected) Native index of the first character position following
      *              the current chunk.
      *  @stable ICU 3.6
      */
    int64_t         chunkNativeLimit;

    /**
     *   (protected)  Size in bytes of the extra space (pExtra).
     *  @stable ICU 3.4
     */
    int32_t        extraSize;

    /**
      *    (protected) The highest chunk offset where native indexing and
      *    chunk (UTF-16) indexing correspond.  For UTF-16 sources, value
      *    will be equal to chunkLength.
      *
      *    @stable ICU 3.6
      */
    int32_t         nativeIndexingLimit;

    /* ---- 16 byte alignment boundary------ */

    /**
     *  (protected) Native index of the first character in the text chunk.
     *  @stable ICU 3.6
     */
    int64_t         chunkNativeStart;

    /**
     *  (protected) Current iteration position within the text chunk (UTF-16 buffer).
     *  This is the index to the character that will be returned by utext_next32().
     *  @stable ICU 3.6
     */
    int32_t         chunkOffset;

    /**
     *  (protected) Length the text chunk (UTF-16 buffer), in UChars.
     *  @stable ICU 3.6
     */
    int32_t         chunkLength;

    /* ---- 16  byte alignment boundary-- */


    /**
     *  (protected)  pointer to a chunk of text in UTF-16 format.
     *  May refer either to original storage of the source of the text, or
     *  if conversion was required, to a buffer owned by the UText.
     *  @stable ICU 3.6
     */
    const UChar    *chunkContents;

     /**
      * (public)     Pointer to Dispatch table for accessing functions for this UText.
      * @stable ICU 3.6
      */
    const UTextFuncs     *pFuncs;

    /**
     *  (protected)  Pointer to additional space requested by the
     *               text provider during the utext_open operation.
     * @stable ICU 3.4
     */
    void          *pExtra;

    /**
     * (protected) Pointer to string or text-containing object or similar.
     * This is the source of the text that this UText is wrapping, in a format
     *  that is known to the text provider functions.
     * @stable ICU 3.4
     */
    const void   *context;

    /* --- 16 byte alignment boundary--- */

    /**
     * (protected) Pointer fields available for use by the text provider.
     * Not used by UText common code.
     * @stable ICU 3.6
     */
    const void     *p;
    /**
     * (protected) Pointer fields available for use by the text provider.
     * Not used by UText common code.
     * @stable ICU 3.6
     */
    const void     *q;
     /**
     * (protected) Pointer fields available for use by the text provider.
     * Not used by UText common code.
     * @stable ICU 3.6
      */
    const void     *r;

    /**
      *  Private field reserved for future use by the UText framework
      *     itself.  This is not to be touched by the text providers.
      * @internal ICU 3.4
      */
    void           *privP;


    /* --- 16 byte alignment boundary--- */


    /**
      * (protected) Integer field reserved for use by the text provider.
      * Not used by the UText framework, or by the client (user) of the UText.
      * @stable ICU 3.4
      */
    int64_t         a;

    /**
      * (protected) Integer field reserved for use by the text provider.
      * Not used by the UText framework, or by the client (user) of the UText.
      * @stable ICU 3.4
      */
    int32_t         b;

    /**
      * (protected) Integer field reserved for use by the text provider.
      * Not used by the UText framework, or by the client (user) of the UText.
      * @stable ICU 3.4
      */
    int32_t         c;

    /*  ---- 16 byte alignment boundary---- */


    /**
      *  Private field reserved for future use by the UText framework
      *     itself.  This is not to be touched by the text providers.
      * @internal ICU 3.4
      */
    int64_t         privA;
    /**
      *  Private field reserved for future use by the UText framework
      *     itself.  This is not to be touched by the text providers.
      * @internal ICU 3.4
      */
    int32_t         privB;
    /**
      *  Private field reserved for future use by the UText framework
      *     itself.  This is not to be touched by the text providers.
      * @internal ICU 3.4
      */
    int32_t         privC;
};


/**
 *  Common function for use by Text Provider implementations to allocate and/or initialize
 *  a new UText struct.  To be called in the implementation of utext_open() functions.
 *  If the supplied UText parameter is null, a new UText struct will be allocated on the heap.
 *  If the supplied UText is already open, the provider's close function will be called
 *  so that the struct can be reused by the open that is in progress.
 *
 * @param ut   pointer to a UText struct to be re-used, or null if a new UText
 *             should be allocated.
 * @param extraSpace The amount of additional space to be allocated as part
 *             of this UText, for use by types of providers that require
 *             additional storage.
 * @param status Errors are returned here.
 * @return pointer to the UText, allocated if necessary, with extra space set up if requested.
 * @stable ICU 3.4
 */
U_STABLE UText * U_EXPORT2
utext_setup(UText *ut, int32_t extraSpace, UErrorCode *status);

// do not use #ifndef U_HIDE_INTERNAL_API around the following!
/**
  * @internal
  *  Value used to help identify correctly initialized UText structs.
  *  Note:  must be publicly visible so that UTEXT_INITIALIZER can access it.
  */
enum {
    UTEXT_MAGIC = 0x345ad82c
};

/**
 * initializer to be used with local (stack) instances of a UText
 *  struct.  UText structs must be initialized before passing
 *  them to one of the utext_open functions.
 *
 * @stable ICU 3.6
 */
#define UTEXT_INITIALIZER {                                        \
                  UTEXT_MAGIC,          /* magic                */ \
                  0,                    /* flags                */ \
                  0,                    /* providerProps        */ \
                  sizeof(UText),        /* sizeOfStruct         */ \
                  0,                    /* chunkNativeLimit     */ \
                  0,                    /* extraSize            */ \
                  0,                    /* nativeIndexingLimit  */ \
                  0,                    /* chunkNativeStart     */ \
                  0,                    /* chunkOffset          */ \
                  0,                    /* chunkLength          */ \
                  NULL,                 /* chunkContents        */ \
                  NULL,                 /* pFuncs               */ \
                  NULL,                 /* pExtra               */ \
                  NULL,                 /* context              */ \
                  NULL, NULL, NULL,     /* p, q, r              */ \
                  NULL,                 /* privP                */ \
                  0, 0, 0,              /* a, b, c              */ \
                  0, 0, 0               /* privA,B,C,           */ \
                  }


U_CDECL_END




#endif








// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// ucasemap_imp.h
// created: 2017feb08 Markus W. Scherer







/**
 * Bit mask for the titlecasing iterator options bit field.
 * Currently only 3 out of 8 values are used:
 * 0 (words), U_TITLECASE_WHOLE_STRING, U_TITLECASE_SENTENCES.
 * See stringoptions.h.
 * @internal
 */
#define U_TITLECASE_ITERATOR_MASK 0xe0

/**
 * Bit mask for the titlecasing index adjustment options bit set.
 * Currently two bits are defined:
 * U_TITLECASE_NO_BREAK_ADJUSTMENT, U_TITLECASE_ADJUST_TO_CASED.
 * See stringoptions.h.
 * @internal
 */
#define U_TITLECASE_ADJUSTMENT_MASK 0x600

/**
 * Internal API, used by u_strcasecmp() etc.
 * Compare strings case-insensitively,
 * in code point order or code unit order.
 */
U_CFUNC int32_t
u_strcmpFold(const UChar *s1, int32_t length1,
             const UChar *s2, int32_t length2,
             uint32_t options,
             UErrorCode *pErrorCode);

/**
 * Internal API, used for detecting length of
 * shared prefix case-insensitively.
 * @param s1            input string 1
 * @param length1       length of string 1, or -1 (NULL terminated)
 * @param s2            input string 2
 * @param length2       length of string 2, or -1 (NULL terminated)
 * @param options       compare options
 * @param matchLen1     (output) length of partial prefix match in s1
 * @param matchLen2     (output) length of partial prefix match in s2
 * @param pErrorCode    receives error status
 */
U_CAPI void
u_caseInsensitivePrefixMatch(const UChar *s1, int32_t length1,
                             const UChar *s2, int32_t length2,
                             uint32_t options,
                             int32_t *matchLen1, int32_t *matchLen2,
                             UErrorCode *pErrorCode);

#ifdef __cplusplus

U_NAMESPACE_BEGIN

class BreakIterator;        // unicode/brkiter.h
class ByteSink;
class Locale;               // unicode/locid.h

/** Returns TRUE if the options are valid. Otherwise FALSE, and sets an error. */
inline UBool ustrcase_checkTitleAdjustmentOptions(uint32_t options, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return FALSE; }
    if ((options & U_TITLECASE_ADJUSTMENT_MASK) == U_TITLECASE_ADJUSTMENT_MASK) {
        // Both options together.
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
    return TRUE;
}

inline UBool ustrcase_isLNS(UChar32 c) {
    // Letter, number, symbol,
    // or a private use code point because those are typically used as letters or numbers.
    // Consider modifier letters only if they are cased.
    const uint32_t LNS = (U_GC_L_MASK|U_GC_N_MASK|U_GC_S_MASK|U_GC_CO_MASK) & ~U_GC_LM_MASK;
    int gc = u_charType(c);
    return (U_MASK(gc) & LNS) != 0 || (gc == U_MODIFIER_LETTER && ucase_getType(c) != UCASE_NONE);
}

#if !UCONFIG_NO_BREAK_ITERATION

/** Returns nullptr if error. Pass in either locale or locID, not both. */
U_CFUNC
BreakIterator *ustrcase_getTitleBreakIterator(
        const Locale *locale, const char *locID, uint32_t options, BreakIterator *iter,
        LocalPointer<BreakIterator> &ownedIter, UErrorCode &errorCode);

#endif

U_NAMESPACE_END

  // for UStringCaseMapper

/*
 * Internal string casing functions implementing
 * ustring.h/ustrcase.cpp and UnicodeString case mapping functions.
 */

struct UCaseMap : public icu::UMemory {
    /** Implements most of ucasemap_open(). */
    UCaseMap(const char *localeID, uint32_t opts, UErrorCode *pErrorCode);
    ~UCaseMap();

#if !UCONFIG_NO_BREAK_ITERATION
    icu::BreakIterator *iter;  /* We adopt the iterator, so we own it. */
#endif
    char locale[32];
    int32_t caseLocale;
    uint32_t options;
};

#if UCONFIG_NO_BREAK_ITERATION
#   define UCASEMAP_BREAK_ITERATOR_PARAM
#   define UCASEMAP_BREAK_ITERATOR_UNUSED
#   define UCASEMAP_BREAK_ITERATOR
#   define UCASEMAP_BREAK_ITERATOR_NULL
#else
#   define UCASEMAP_BREAK_ITERATOR_PARAM icu::BreakIterator *iter,
#   define UCASEMAP_BREAK_ITERATOR_UNUSED icu::BreakIterator *,
#   define UCASEMAP_BREAK_ITERATOR iter,
#   define UCASEMAP_BREAK_ITERATOR_NULL NULL,
#endif

U_CFUNC int32_t
ustrcase_getCaseLocale(const char *locale);

// TODO: swap src / dest if approved for new public api
/** Implements UStringCaseMapper. */
U_CFUNC int32_t U_CALLCONV
ustrcase_internalToLower(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
                         UChar *dest, int32_t destCapacity,
                         const UChar *src, int32_t srcLength,
                         icu::Edits *edits,
                         UErrorCode &errorCode);

/** Implements UStringCaseMapper. */
U_CFUNC int32_t U_CALLCONV
ustrcase_internalToUpper(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
                         UChar *dest, int32_t destCapacity,
                         const UChar *src, int32_t srcLength,
                         icu::Edits *edits,
                         UErrorCode &errorCode);

#if !UCONFIG_NO_BREAK_ITERATION

/** Implements UStringCaseMapper. */
U_CFUNC int32_t U_CALLCONV
ustrcase_internalToTitle(int32_t caseLocale, uint32_t options,
                         icu::BreakIterator *iter,
                         UChar *dest, int32_t destCapacity,
                         const UChar *src, int32_t srcLength,
                         icu::Edits *edits,
                         UErrorCode &errorCode);

#endif

/** Implements UStringCaseMapper. */
U_CFUNC int32_t U_CALLCONV
ustrcase_internalFold(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
                      UChar *dest, int32_t destCapacity,
                      const UChar *src, int32_t srcLength,
                      icu::Edits *edits,
                      UErrorCode &errorCode);

/**
 * Common string case mapping implementation for ucasemap_toXyz() and UnicodeString::toXyz().
 * Implements argument checking.
 */
U_CFUNC int32_t
ustrcase_map(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
             UChar *dest, int32_t destCapacity,
             const UChar *src, int32_t srcLength,
             UStringCaseMapper *stringCaseMapper,
             icu::Edits *edits,
             UErrorCode &errorCode);

/**
 * Common string case mapping implementation for old-fashioned u_strToXyz() functions
 * that allow the source string to overlap the destination buffer.
 * Implements argument checking and internally works with an intermediate buffer if necessary.
 */
U_CFUNC int32_t
ustrcase_mapWithOverlap(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
                        UChar *dest, int32_t destCapacity,
                        const UChar *src, int32_t srcLength,
                        UStringCaseMapper *stringCaseMapper,
                        UErrorCode &errorCode);

/**
 * UTF-8 string case mapping function type, used by ucasemap_mapUTF8().
 * UTF-8 version of UStringCaseMapper.
 * All error checking must be done.
 * The UCaseMap must be fully initialized, with locale and/or iter set as needed.
 */
typedef void U_CALLCONV
UTF8CaseMapper(int32_t caseLocale, uint32_t options,
#if !UCONFIG_NO_BREAK_ITERATION
               icu::BreakIterator *iter,
#endif
               const uint8_t *src, int32_t srcLength,
               icu::ByteSink &sink, icu::Edits *edits,
               UErrorCode &errorCode);

#if !UCONFIG_NO_BREAK_ITERATION

/** Implements UTF8CaseMapper. */
U_CFUNC void U_CALLCONV
ucasemap_internalUTF8ToTitle(int32_t caseLocale, uint32_t options,
        icu::BreakIterator *iter,
        const uint8_t *src, int32_t srcLength,
        icu::ByteSink &sink, icu::Edits *edits,
        UErrorCode &errorCode);

#endif

void
ucasemap_mapUTF8(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
                 const char *src, int32_t srcLength,
                 UTF8CaseMapper *stringCaseMapper,
                 icu::ByteSink &sink, icu::Edits *edits,
                 UErrorCode &errorCode);

/**
 * Implements argument checking and buffer handling
 * for UTF-8 string case mapping as a common function.
 */
int32_t
ucasemap_mapUTF8(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
                 char *dest, int32_t destCapacity,
                 const char *src, int32_t srcLength,
                 UTF8CaseMapper *stringCaseMapper,
                 icu::Edits *edits,
                 UErrorCode &errorCode);

U_NAMESPACE_BEGIN
namespace GreekUpper {

// Data bits.
static const uint32_t UPPER_MASK = 0x3ff;
static const uint32_t HAS_VOWEL = 0x1000;
static const uint32_t HAS_YPOGEGRAMMENI = 0x2000;
static const uint32_t HAS_ACCENT = 0x4000;
static const uint32_t HAS_DIALYTIKA = 0x8000;
// Further bits during data building and processing, not stored in the data map.
static const uint32_t HAS_COMBINING_DIALYTIKA = 0x10000;
static const uint32_t HAS_OTHER_GREEK_DIACRITIC = 0x20000;

static const uint32_t HAS_VOWEL_AND_ACCENT = HAS_VOWEL | HAS_ACCENT;
static const uint32_t HAS_VOWEL_AND_ACCENT_AND_DIALYTIKA =
        HAS_VOWEL_AND_ACCENT | HAS_DIALYTIKA;
static const uint32_t HAS_EITHER_DIALYTIKA = HAS_DIALYTIKA | HAS_COMBINING_DIALYTIKA;

// State bits.
static const uint32_t AFTER_CASED = 1;
static const uint32_t AFTER_VOWEL_WITH_ACCENT = 2;

uint32_t getLetterData(UChar32 c);

/**
 * Returns a non-zero value for each of the Greek combining diacritics
 * listed in The Unicode Standard, version 8, chapter 7.2 Greek,
 * plus some perispomeni look-alikes.
 */
uint32_t getDiacriticData(UChar32 c);

}  // namespace GreekUpper
U_NAMESPACE_END

#endif  // __cplusplus




U_NAMESPACE_USE

/* UCaseMap service object -------------------------------------------------- */

UCaseMap::UCaseMap(const char *localeID, uint32_t opts, UErrorCode *pErrorCode) :
#if !UCONFIG_NO_BREAK_ITERATION
        iter(NULL),
#endif
        caseLocale(UCASE_LOC_UNKNOWN), options(opts) {
    ucasemap_setLocale(this, localeID, pErrorCode);
}

UCaseMap::~UCaseMap() {
#if !UCONFIG_NO_BREAK_ITERATION
    delete iter;
#endif
}

U_CAPI UCaseMap * U_EXPORT2
ucasemap_open(const char *locale, uint32_t options, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    UCaseMap *csm = new UCaseMap(locale, options, pErrorCode);
    if(csm==NULL) {
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    } else if (U_FAILURE(*pErrorCode)) {
        delete csm;
        return NULL;
    }
    return csm;
}

U_CAPI void U_EXPORT2
ucasemap_close(UCaseMap *csm) {
    delete csm;
}

U_CAPI const char * U_EXPORT2
ucasemap_getLocale(const UCaseMap *csm) {
    return csm->locale;
}

U_CAPI uint32_t U_EXPORT2
ucasemap_getOptions(const UCaseMap *csm) {
    return csm->options;
}

U_CAPI void U_EXPORT2
ucasemap_setLocale(UCaseMap *csm, const char *locale, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return;
    }
    if (locale != NULL && *locale == 0) {
        csm->locale[0] = 0;
        csm->caseLocale = UCASE_LOC_ROOT;
        return;
    }

    int32_t length=uloc_getName(locale, csm->locale, (int32_t)sizeof(csm->locale), pErrorCode);
    if(*pErrorCode==U_BUFFER_OVERFLOW_ERROR || length==sizeof(csm->locale)) {
        *pErrorCode=U_ZERO_ERROR;
        /* we only really need the language code for case mappings */
        length=uloc_getLanguage(locale, csm->locale, (int32_t)sizeof(csm->locale), pErrorCode);
    }
    if(length==sizeof(csm->locale)) {
        *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
    }
    if(U_SUCCESS(*pErrorCode)) {
        csm->caseLocale=UCASE_LOC_UNKNOWN;
        csm->caseLocale = ucase_getCaseLocale(csm->locale);
    } else {
        csm->locale[0]=0;
        csm->caseLocale = UCASE_LOC_ROOT;
    }
}

U_CAPI void U_EXPORT2
ucasemap_setOptions(UCaseMap *csm, uint32_t options, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return;
    }
    csm->options=options;
}

/* UTF-8 string case mappings ----------------------------------------------- */

/* TODO(markus): Move to a new, separate utf8case.cpp file. */

namespace {

/* append a full case mapping result, see UCASE_MAX_STRING_LENGTH */
inline UBool
appendResult(int32_t cpLength, int32_t result, const UChar *s,
             ByteSink &sink, uint32_t options, icu::Edits *edits, UErrorCode &errorCode) {
    U_ASSERT(U_SUCCESS(errorCode));

    /* decode the result */
    if(result<0) {
        /* (not) original code point */
        if(edits!=NULL) {
            edits->addUnchanged(cpLength);
        }
        if((options & U_OMIT_UNCHANGED_TEXT) == 0) {
            ByteSinkUtil::appendCodePoint(cpLength, ~result, sink);
        }
    } else {
        if(result<=UCASE_MAX_STRING_LENGTH) {
            // string: "result" is the UTF-16 length
            return ByteSinkUtil::appendChange(cpLength, s, result, sink, edits, errorCode);
        } else {
            ByteSinkUtil::appendCodePoint(cpLength, result, sink, edits);
        }
    }
    return TRUE;
}

// See unicode/utf8.h U8_APPEND_UNSAFE().
uint8_t ucasemap_getTwoByteLead(UChar32 c) { return (uint8_t)((c >> 6) | 0xc0); }
uint8_t ucasemap_getTwoByteTrail(UChar32 c) { return (uint8_t)((c & 0x3f) | 0x80); }

UChar32 U_CALLCONV
utf8_caseContextIterator(void *context, int8_t dir) {
    UCaseContext *csc=(UCaseContext *)context;
    UChar32 c;

    if(dir<0) {
        /* reset for backward iteration */
        csc->index=csc->cpStart;
        csc->dir=dir;
    } else if(dir>0) {
        /* reset for forward iteration */
        csc->index=csc->cpLimit;
        csc->dir=dir;
    } else {
        /* continue current iteration direction */
        dir=csc->dir;
    }

    if(dir<0) {
        if(csc->start<csc->index) {
            U8_PREV((const uint8_t *)csc->p, csc->start, csc->index, c);
            return c;
        }
    } else {
        if(csc->index<csc->limit) {
            U8_NEXT((const uint8_t *)csc->p, csc->index, csc->limit, c);
            return c;
        }
    }
    return U_SENTINEL;
}

/**
 * caseLocale >= 0: Lowercases [srcStart..srcLimit[ but takes context [0..srcLength[ into account.
 * caseLocale < 0: Case-folds [srcStart..srcLimit[.
 */
void toLower(int32_t caseLocale, uint32_t options,
             const uint8_t *src, UCaseContext *csc, int32_t srcStart, int32_t srcLimit,
             icu::ByteSink &sink, icu::Edits *edits, UErrorCode &errorCode) {
    const int8_t *latinToLower;
    if (caseLocale == UCASE_LOC_ROOT ||
            (caseLocale >= 0 ?
                !(caseLocale == UCASE_LOC_TURKISH || caseLocale == UCASE_LOC_LITHUANIAN) :
                (options & _FOLD_CASE_OPTIONS_MASK) == U_FOLD_CASE_DEFAULT)) {
        latinToLower = LatinCase::TO_LOWER_NORMAL;
    } else {
        latinToLower = LatinCase::TO_LOWER_TR_LT;
    }
    const UTrie2 *trie = ucase_getTrie();
    int32_t prev = srcStart;
    int32_t srcIndex = srcStart;
    for (;;) {
        // fast path for simple cases
        int32_t cpStart;
        UChar32 c;
        for (;;) {
            if (U_FAILURE(errorCode) || srcIndex >= srcLimit) {
                c = U_SENTINEL;
                break;
            }
            uint8_t lead = src[srcIndex++];
            if (lead <= 0x7f) {
                int8_t d = latinToLower[lead];
                if (d == LatinCase::EXC) {
                    cpStart = srcIndex - 1;
                    c = lead;
                    break;
                }
                if (d == 0) { continue; }
                ByteSinkUtil::appendUnchanged(src + prev, srcIndex - 1 - prev,
                                              sink, options, edits, errorCode);
                char ascii = (char)(lead + d);
                sink.Append(&ascii, 1);
                if (edits != nullptr) {
                    edits->addReplace(1, 1);
                }
                prev = srcIndex;
                continue;
            } else if (lead < 0xe3) {
                uint8_t t;
                if (0xc2 <= lead && lead <= 0xc5 && srcIndex < srcLimit &&
                        (t = src[srcIndex] - 0x80) <= 0x3f) {
                    // U+0080..U+017F
                    ++srcIndex;
                    c = ((lead - 0xc0) << 6) | t;
                    int8_t d = latinToLower[c];
                    if (d == LatinCase::EXC) {
                        cpStart = srcIndex - 2;
                        break;
                    }
                    if (d == 0) { continue; }
                    ByteSinkUtil::appendUnchanged(src + prev, srcIndex - 2 - prev,
                                                  sink, options, edits, errorCode);
                    ByteSinkUtil::appendTwoBytes(c + d, sink);
                    if (edits != nullptr) {
                        edits->addReplace(2, 2);
                    }
                    prev = srcIndex;
                    continue;
                }
            } else if ((lead <= 0xe9 || lead == 0xeb || lead == 0xec) &&
                    (srcIndex + 2) <= srcLimit &&
                    U8_IS_TRAIL(src[srcIndex]) && U8_IS_TRAIL(src[srcIndex + 1])) {
                // most of CJK: no case mappings
                srcIndex += 2;
                continue;
            }
            cpStart = --srcIndex;
            U8_NEXT(src, srcIndex, srcLimit, c);
            if (c < 0) {
                // ill-formed UTF-8
                continue;
            }
            uint16_t props = UTRIE2_GET16(trie, c);
            if (UCASE_HAS_EXCEPTION(props)) { break; }
            int32_t delta;
            if (!UCASE_IS_UPPER_OR_TITLE(props) || (delta = UCASE_GET_DELTA(props)) == 0) {
                continue;
            }
            ByteSinkUtil::appendUnchanged(src + prev, cpStart - prev,
                                          sink, options, edits, errorCode);
            ByteSinkUtil::appendCodePoint(srcIndex - cpStart, c + delta, sink, edits);
            prev = srcIndex;
        }
        if (c < 0) {
            break;
        }
        // slow path
        const UChar *s;
        if (caseLocale >= 0) {
            csc->cpStart = cpStart;
            csc->cpLimit = srcIndex;
            c = ucase_toFullLower(c, utf8_caseContextIterator, csc, &s, caseLocale);
        } else {
            c = ucase_toFullFolding(c, &s, options);
        }
        if (c >= 0) {
            ByteSinkUtil::appendUnchanged(src + prev, cpStart - prev,
                                          sink, options, edits, errorCode);
            appendResult(srcIndex - cpStart, c, s, sink, options, edits, errorCode);
            prev = srcIndex;
        }
    }
    ByteSinkUtil::appendUnchanged(src + prev, srcIndex - prev,
                                  sink, options, edits, errorCode);
}

void toUpper(int32_t caseLocale, uint32_t options,
             const uint8_t *src, UCaseContext *csc, int32_t srcLength,
             icu::ByteSink &sink, icu::Edits *edits, UErrorCode &errorCode) {
    const int8_t *latinToUpper;
    if (caseLocale == UCASE_LOC_TURKISH) {
        latinToUpper = LatinCase::TO_UPPER_TR;
    } else {
        latinToUpper = LatinCase::TO_UPPER_NORMAL;
    }
    const UTrie2 *trie = ucase_getTrie();
    int32_t prev = 0;
    int32_t srcIndex = 0;
    for (;;) {
        // fast path for simple cases
        int32_t cpStart;
        UChar32 c;
        for (;;) {
            if (U_FAILURE(errorCode) || srcIndex >= srcLength) {
                c = U_SENTINEL;
                break;
            }
            uint8_t lead = src[srcIndex++];
            if (lead <= 0x7f) {
                int8_t d = latinToUpper[lead];
                if (d == LatinCase::EXC) {
                    cpStart = srcIndex - 1;
                    c = lead;
                    break;
                }
                if (d == 0) { continue; }
                ByteSinkUtil::appendUnchanged(src + prev, srcIndex - 1 - prev,
                                              sink, options, edits, errorCode);
                char ascii = (char)(lead + d);
                sink.Append(&ascii, 1);
                if (edits != nullptr) {
                    edits->addReplace(1, 1);
                }
                prev = srcIndex;
                continue;
            } else if (lead < 0xe3) {
                uint8_t t;
                if (0xc2 <= lead && lead <= 0xc5 && srcIndex < srcLength &&
                        (t = src[srcIndex] - 0x80) <= 0x3f) {
                    // U+0080..U+017F
                    ++srcIndex;
                    c = ((lead - 0xc0) << 6) | t;
                    int8_t d = latinToUpper[c];
                    if (d == LatinCase::EXC) {
                        cpStart = srcIndex - 2;
                        break;
                    }
                    if (d == 0) { continue; }
                    ByteSinkUtil::appendUnchanged(src + prev, srcIndex - 2 - prev,
                                                  sink, options, edits, errorCode);
                    ByteSinkUtil::appendTwoBytes(c + d, sink);
                    if (edits != nullptr) {
                        edits->addReplace(2, 2);
                    }
                    prev = srcIndex;
                    continue;
                }
            } else if ((lead <= 0xe9 || lead == 0xeb || lead == 0xec) &&
                    (srcIndex + 2) <= srcLength &&
                    U8_IS_TRAIL(src[srcIndex]) && U8_IS_TRAIL(src[srcIndex + 1])) {
                // most of CJK: no case mappings
                srcIndex += 2;
                continue;
            }
            cpStart = --srcIndex;
            U8_NEXT(src, srcIndex, srcLength, c);
            if (c < 0) {
                // ill-formed UTF-8
                continue;
            }
            uint16_t props = UTRIE2_GET16(trie, c);
            if (UCASE_HAS_EXCEPTION(props)) { break; }
            int32_t delta;
            if (UCASE_GET_TYPE(props) != UCASE_LOWER || (delta = UCASE_GET_DELTA(props)) == 0) {
                continue;
            }
            ByteSinkUtil::appendUnchanged(src + prev, cpStart - prev,
                                          sink, options, edits, errorCode);
            ByteSinkUtil::appendCodePoint(srcIndex - cpStart, c + delta, sink, edits);
            prev = srcIndex;
        }
        if (c < 0) {
            break;
        }
        // slow path
        csc->cpStart = cpStart;
        csc->cpLimit = srcIndex;
        const UChar *s;
        c = ucase_toFullUpper(c, utf8_caseContextIterator, csc, &s, caseLocale);
        if (c >= 0) {
            ByteSinkUtil::appendUnchanged(src + prev, cpStart - prev,
                                          sink, options, edits, errorCode);
            appendResult(srcIndex - cpStart, c, s, sink, options, edits, errorCode);
            prev = srcIndex;
        }
    }
    ByteSinkUtil::appendUnchanged(src + prev, srcIndex - prev,
                                  sink, options, edits, errorCode);
}

}  // namespace

#if !UCONFIG_NO_BREAK_ITERATION

U_CFUNC void U_CALLCONV
ucasemap_internalUTF8ToTitle(
        int32_t caseLocale, uint32_t options, BreakIterator *iter,
        const uint8_t *src, int32_t srcLength,
        ByteSink &sink, icu::Edits *edits,
        UErrorCode &errorCode) {
    if (!ustrcase_checkTitleAdjustmentOptions(options, errorCode)) {
        return;
    }

    /* set up local variables */
    UCaseContext csc=UCASECONTEXT_INITIALIZER;
    csc.p=(void *)src;
    csc.limit=srcLength;
    int32_t prev=0;
    UBool isFirstIndex=TRUE;

    /* titlecasing loop */
    while(prev<srcLength) {
        /* find next index where to titlecase */
        int32_t index;
        if(isFirstIndex) {
            isFirstIndex=FALSE;
            index=iter->first();
        } else {
            index=iter->next();
        }
        if(index==UBRK_DONE || index>srcLength) {
            index=srcLength;
        }

        /*
         * Segment [prev..index[ into 3 parts:
         * a) skipped characters (copy as-is) [prev..titleStart[
         * b) first letter (titlecase)              [titleStart..titleLimit[
         * c) subsequent characters (lowercase)                 [titleLimit..index[
         */
        if(prev<index) {
            /* find and copy skipped characters [prev..titleStart[ */
            int32_t titleStart=prev;
            int32_t titleLimit=prev;
            UChar32 c;
            U8_NEXT(src, titleLimit, index, c);
            if ((options&U_TITLECASE_NO_BREAK_ADJUSTMENT)==0) {
                // Adjust the titlecasing index to the next cased character,
                // or to the next letter/number/symbol/private use.
                // Stop with titleStart<titleLimit<=index
                // if there is a character to be titlecased,
                // or else stop with titleStart==titleLimit==index.
                UBool toCased = (options&U_TITLECASE_ADJUST_TO_CASED) != 0;
                while (toCased ? UCASE_NONE==ucase_getType(c) : !ustrcase_isLNS(c)) {
                    titleStart=titleLimit;
                    if(titleLimit==index) {
                        break;
                    }
                    U8_NEXT(src, titleLimit, index, c);
                }
                if (prev < titleStart) {
                    if (!ByteSinkUtil::appendUnchanged(src+prev, titleStart-prev,
                                                       sink, options, edits, errorCode)) {
                        return;
                    }
                }
            }

            if(titleStart<titleLimit) {
                /* titlecase c which is from [titleStart..titleLimit[ */
                if(c>=0) {
                    csc.cpStart=titleStart;
                    csc.cpLimit=titleLimit;
                    const UChar *s;
                    c=ucase_toFullTitle(c, utf8_caseContextIterator, &csc, &s, caseLocale);
                    if (!appendResult(titleLimit-titleStart, c, s, sink, options, edits, errorCode)) {
                        return;
                    }
                } else {
                    // Malformed UTF-8.
                    if (!ByteSinkUtil::appendUnchanged(src+titleStart, titleLimit-titleStart,
                                                       sink, options, edits, errorCode)) {
                        return;
                    }
                }

                /* Special case Dutch IJ titlecasing */
                if (titleStart+1 < index &&
                        caseLocale == UCASE_LOC_DUTCH &&
                        (src[titleStart] == 0x0049 || src[titleStart] == 0x0069)) {
                    if (src[titleStart+1] == 0x006A) {
                        ByteSinkUtil::appendCodePoint(1, 0x004A, sink, edits);
                        titleLimit++;
                    } else if (src[titleStart+1] == 0x004A) {
                        // Keep the capital J from getting lowercased.
                        if (!ByteSinkUtil::appendUnchanged(src+titleStart+1, 1,
                                                           sink, options, edits, errorCode)) {
                            return;
                        }
                        titleLimit++;
                    }
                }

                /* lowercase [titleLimit..index[ */
                if(titleLimit<index) {
                    if((options&U_TITLECASE_NO_LOWERCASE)==0) {
                        /* Normal operation: Lowercase the rest of the word. */
                        toLower(caseLocale, options,
                                src, &csc, titleLimit, index,
                                sink, edits, errorCode);
                        if(U_FAILURE(errorCode)) {
                            return;
                        }
                    } else {
                        /* Optionally just copy the rest of the word unchanged. */
                        if (!ByteSinkUtil::appendUnchanged(src+titleLimit, index-titleLimit,
                                                           sink, options, edits, errorCode)) {
                            return;
                        }
                    }
                }
            }
        }

        prev=index;
    }
}

#endif

U_NAMESPACE_BEGIN
namespace GreekUpper {

UBool isFollowedByCasedLetter(const uint8_t *s, int32_t i, int32_t length) {
    while (i < length) {
        UChar32 c;
        U8_NEXT(s, i, length, c);
        int32_t type = ucase_getTypeOrIgnorable(c);
        if ((type & UCASE_IGNORABLE) != 0) {
            // Case-ignorable, continue with the loop.
        } else if (type != UCASE_NONE) {
            return TRUE;  // Followed by cased letter.
        } else {
            return FALSE;  // Uncased and not case-ignorable.
        }
    }
    return FALSE;  // Not followed by cased letter.
}

// Keep this consistent with the UTF-16 version in ustrcase.cpp and the Java version in CaseMap.java.
void toUpper(uint32_t options,
             const uint8_t *src, int32_t srcLength,
             ByteSink &sink, Edits *edits,
             UErrorCode &errorCode) {
    uint32_t state = 0;
    for (int32_t i = 0; i < srcLength;) {
        int32_t nextIndex = i;
        UChar32 c;
        U8_NEXT(src, nextIndex, srcLength, c);
        uint32_t nextState = 0;
        int32_t type = ucase_getTypeOrIgnorable(c);
        if ((type & UCASE_IGNORABLE) != 0) {
            // c is case-ignorable
            nextState |= (state & AFTER_CASED);
        } else if (type != UCASE_NONE) {
            // c is cased
            nextState |= AFTER_CASED;
        }
        uint32_t data = getLetterData(c);
        if (data > 0) {
            uint32_t upper = data & UPPER_MASK;
            // Add a dialytika to this iota or ypsilon vowel
            // if we removed a tonos from the previous vowel,
            // and that previous vowel did not also have (or gain) a dialytika.
            // Adding one only to the final vowel in a longer sequence
            // (which does not occur in normal writing) would require lookahead.
            // Set the same flag as for preserving an existing dialytika.
            if ((data & HAS_VOWEL) != 0 && (state & AFTER_VOWEL_WITH_ACCENT) != 0 &&
                    (upper == 0x399 || upper == 0x3A5)) {
                data |= HAS_DIALYTIKA;
            }
            int32_t numYpogegrammeni = 0;  // Map each one to a trailing, spacing, capital iota.
            if ((data & HAS_YPOGEGRAMMENI) != 0) {
                numYpogegrammeni = 1;
            }
            // Skip combining diacritics after this Greek letter.
            int32_t nextNextIndex = nextIndex;
            while (nextIndex < srcLength) {
                UChar32 c2;
                U8_NEXT(src, nextNextIndex, srcLength, c2);
                uint32_t diacriticData = getDiacriticData(c2);
                if (diacriticData != 0) {
                    data |= diacriticData;
                    if ((diacriticData & HAS_YPOGEGRAMMENI) != 0) {
                        ++numYpogegrammeni;
                    }
                    nextIndex = nextNextIndex;
                } else {
                    break;  // not a Greek diacritic
                }
            }
            if ((data & HAS_VOWEL_AND_ACCENT_AND_DIALYTIKA) == HAS_VOWEL_AND_ACCENT) {
                nextState |= AFTER_VOWEL_WITH_ACCENT;
            }
            // Map according to Greek rules.
            UBool addTonos = FALSE;
            if (upper == 0x397 &&
                    (data & HAS_ACCENT) != 0 &&
                    numYpogegrammeni == 0 &&
                    (state & AFTER_CASED) == 0 &&
                    !isFollowedByCasedLetter(src, nextIndex, srcLength)) {
                // Keep disjunctive "or" with (only) a tonos.
                // We use the same "word boundary" conditions as for the Final_Sigma test.
                if (i == nextIndex) {
                    upper = 0x389;  // Preserve the precomposed form.
                } else {
                    addTonos = TRUE;
                }
            } else if ((data & HAS_DIALYTIKA) != 0) {
                // Preserve a vowel with dialytika in precomposed form if it exists.
                if (upper == 0x399) {
                    upper = 0x3AA;
                    data &= ~HAS_EITHER_DIALYTIKA;
                } else if (upper == 0x3A5) {
                    upper = 0x3AB;
                    data &= ~HAS_EITHER_DIALYTIKA;
                }
            }

            UBool change;
            if (edits == nullptr && (options & U_OMIT_UNCHANGED_TEXT) == 0) {
                change = TRUE;  // common, simple usage
            } else {
                // Find out first whether we are changing the text.
                U_ASSERT(0x370 <= upper && upper <= 0x3ff);  // 2-byte UTF-8, main Greek block
                change = (i + 2) > nextIndex ||
                        src[i] != ucasemap_getTwoByteLead(upper) || src[i + 1] != ucasemap_getTwoByteTrail(upper) ||
                        numYpogegrammeni > 0;
                int32_t i2 = i + 2;
                if ((data & HAS_EITHER_DIALYTIKA) != 0) {
                    change |= (i2 + 2) > nextIndex ||
                            src[i2] != (uint8_t)u8"\u0308"[0] ||
                            src[i2 + 1] != (uint8_t)u8"\u0308"[1];
                    i2 += 2;
                }
                if (addTonos) {
                    change |= (i2 + 2) > nextIndex ||
                            src[i2] != (uint8_t)u8"\u0301"[0] ||
                            src[i2 + 1] != (uint8_t)u8"\u0301"[1];
                    i2 += 2;
                }
                int32_t oldLength = nextIndex - i;
                int32_t newLength = (i2 - i) + numYpogegrammeni * 2;  // 2 bytes per U+0399
                change |= oldLength != newLength;
                if (change) {
                    if (edits != NULL) {
                        edits->addReplace(oldLength, newLength);
                    }
                } else {
                    if (edits != NULL) {
                        edits->addUnchanged(oldLength);
                    }
                    // Write unchanged text?
                    change = (options & U_OMIT_UNCHANGED_TEXT) == 0;
                }
            }

            if (change) {
                ByteSinkUtil::appendTwoBytes(upper, sink);
                if ((data & HAS_EITHER_DIALYTIKA) != 0) {
                    sink.Append(reinterpret_cast<const char*>(u8"\u0308"), 2);  // restore or add a dialytika
                }
                if (addTonos) {
                    sink.Append(reinterpret_cast<const char*>(u8"\u0301"), 2);
                }
                while (numYpogegrammeni > 0) {
                    sink.Append(reinterpret_cast<const char*>(u8"\u0399"), 2);
                    --numYpogegrammeni;
                }
            }
        } else if(c>=0) {
            const UChar *s;
            c=ucase_toFullUpper(c, NULL, NULL, &s, UCASE_LOC_GREEK);
            if (!appendResult(nextIndex - i, c, s, sink, options, edits, errorCode)) {
                return;
            }
        } else {
            // Malformed UTF-8.
            if (!ByteSinkUtil::appendUnchanged(src+i, nextIndex-i,
                                               sink, options, edits, errorCode)) {
                return;
            }
        }
        i = nextIndex;
        state = nextState;
    }
}

}  // namespace GreekUpper
U_NAMESPACE_END

static void U_CALLCONV
ucasemap_internalUTF8ToLower(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_UNUSED
                             const uint8_t *src, int32_t srcLength,
                             icu::ByteSink &sink, icu::Edits *edits,
                             UErrorCode &errorCode) {
    UCaseContext csc=UCASECONTEXT_INITIALIZER;
    csc.p=(void *)src;
    csc.limit=srcLength;
    toLower(
        caseLocale, options,
        src, &csc, 0, srcLength,
        sink, edits, errorCode);
}

static void U_CALLCONV
ucasemap_internalUTF8ToUpper(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_UNUSED
                             const uint8_t *src, int32_t srcLength,
                             icu::ByteSink &sink, icu::Edits *edits,
                             UErrorCode &errorCode) {
    if (caseLocale == UCASE_LOC_GREEK) {
        GreekUpper::toUpper(options, src, srcLength, sink, edits, errorCode);
    } else {
        UCaseContext csc=UCASECONTEXT_INITIALIZER;
        csc.p=(void *)src;
        csc.limit=srcLength;
        toUpper(
            caseLocale, options,
            src, &csc, srcLength,
            sink, edits, errorCode);
    }
}

static void U_CALLCONV
ucasemap_internalUTF8Fold(int32_t /* caseLocale */, uint32_t options, UCASEMAP_BREAK_ITERATOR_UNUSED
                          const uint8_t *src, int32_t srcLength,
                          icu::ByteSink &sink, icu::Edits *edits,
                          UErrorCode &errorCode) {
    toLower(
        -1, options,
        src, nullptr, 0, srcLength,
        sink, edits, errorCode);
}

void
ucasemap_mapUTF8(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
                 const char *src, int32_t srcLength,
                 UTF8CaseMapper *stringCaseMapper,
                 icu::ByteSink &sink, icu::Edits *edits,
                 UErrorCode &errorCode) {
    /* check argument values */
    if (U_FAILURE(errorCode)) {
        return;
    }
    if ((src == nullptr && srcLength != 0) || srcLength < -1) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    // Get the string length.
    if (srcLength == -1) {
        srcLength = (int32_t)uprv_strlen((const char *)src);
    }

    if (edits != nullptr && (options & U_EDITS_NO_RESET) == 0) {
        edits->reset();
    }
    stringCaseMapper(caseLocale, options, UCASEMAP_BREAK_ITERATOR
                     (const uint8_t *)src, srcLength, sink, edits, errorCode);
    sink.Flush();
    if (U_SUCCESS(errorCode)) {
        if (edits != nullptr) {
            edits->copyErrorTo(errorCode);
        }
    }
}

int32_t
ucasemap_mapUTF8(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
                 char *dest, int32_t destCapacity,
                 const char *src, int32_t srcLength,
                 UTF8CaseMapper *stringCaseMapper,
                 icu::Edits *edits,
                 UErrorCode &errorCode) {
    /* check argument values */
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    if( destCapacity<0 ||
        (dest==NULL && destCapacity>0) ||
        (src==NULL && srcLength!=0) || srcLength<-1
    ) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* get the string length */
    if(srcLength==-1) {
        srcLength=(int32_t)uprv_strlen((const char *)src);
    }

    /* check for overlapping source and destination */
    if( dest!=NULL &&
        ((src>=dest && src<(dest+destCapacity)) ||
         (dest>=src && dest<(src+srcLength)))
    ) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    CheckedArrayByteSink sink(dest, destCapacity);
    if (edits != nullptr && (options & U_EDITS_NO_RESET) == 0) {
        edits->reset();
    }
    stringCaseMapper(caseLocale, options, UCASEMAP_BREAK_ITERATOR
                     (const uint8_t *)src, srcLength, sink, edits, errorCode);
    sink.Flush();
    if (U_SUCCESS(errorCode)) {
        if (sink.Overflowed()) {
            errorCode = U_BUFFER_OVERFLOW_ERROR;
        } else if (edits != nullptr) {
            edits->copyErrorTo(errorCode);
        }
    }
    return u_terminateChars(dest, destCapacity, sink.NumberOfBytesAppended(), &errorCode);
}

/* public API functions */

U_CAPI int32_t U_EXPORT2
ucasemap_utf8ToLower(const UCaseMap *csm,
                     char *dest, int32_t destCapacity,
                     const char *src, int32_t srcLength,
                     UErrorCode *pErrorCode) {
    return ucasemap_mapUTF8(
        csm->caseLocale, csm->options, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ucasemap_internalUTF8ToLower, NULL, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
ucasemap_utf8ToUpper(const UCaseMap *csm,
                     char *dest, int32_t destCapacity,
                     const char *src, int32_t srcLength,
                     UErrorCode *pErrorCode) {
    return ucasemap_mapUTF8(
        csm->caseLocale, csm->options, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ucasemap_internalUTF8ToUpper, NULL, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
ucasemap_utf8FoldCase(const UCaseMap *csm,
                      char *dest, int32_t destCapacity,
                      const char *src, int32_t srcLength,
                      UErrorCode *pErrorCode) {
    return ucasemap_mapUTF8(
        UCASE_LOC_ROOT, csm->options, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ucasemap_internalUTF8Fold, NULL, *pErrorCode);
}

U_NAMESPACE_BEGIN

void CaseMap::utf8ToLower(
        const char *locale, uint32_t options,
        StringPiece src, ByteSink &sink, Edits *edits,
        UErrorCode &errorCode) {
    ucasemap_mapUTF8(
        ustrcase_getCaseLocale(locale), options, UCASEMAP_BREAK_ITERATOR_NULL
        src.data(), src.length(),
        ucasemap_internalUTF8ToLower, sink, edits, errorCode);
}

void CaseMap::utf8ToUpper(
        const char *locale, uint32_t options,
        StringPiece src, ByteSink &sink, Edits *edits,
        UErrorCode &errorCode) {
    ucasemap_mapUTF8(
        ustrcase_getCaseLocale(locale), options, UCASEMAP_BREAK_ITERATOR_NULL
        src.data(), src.length(),
        ucasemap_internalUTF8ToUpper, sink, edits, errorCode);
}

void CaseMap::utf8Fold(
        uint32_t options,
        StringPiece src, ByteSink &sink, Edits *edits,
        UErrorCode &errorCode) {
    ucasemap_mapUTF8(
        UCASE_LOC_ROOT, options, UCASEMAP_BREAK_ITERATOR_NULL
        src.data(), src.length(),
        ucasemap_internalUTF8Fold, sink, edits, errorCode);
}

int32_t CaseMap::utf8ToLower(
        const char *locale, uint32_t options,
        const char *src, int32_t srcLength,
        char *dest, int32_t destCapacity, Edits *edits,
        UErrorCode &errorCode) {
    return ucasemap_mapUTF8(
        ustrcase_getCaseLocale(locale), options, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ucasemap_internalUTF8ToLower, edits, errorCode);
}

int32_t CaseMap::utf8ToUpper(
        const char *locale, uint32_t options,
        const char *src, int32_t srcLength,
        char *dest, int32_t destCapacity, Edits *edits,
        UErrorCode &errorCode) {
    return ucasemap_mapUTF8(
        ustrcase_getCaseLocale(locale), options, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ucasemap_internalUTF8ToUpper, edits, errorCode);
}

int32_t CaseMap::utf8Fold(
        uint32_t options,
        const char *src, int32_t srcLength,
        char *dest, int32_t destCapacity, Edits *edits,
        UErrorCode &errorCode) {
    return ucasemap_mapUTF8(
        UCASE_LOC_ROOT, options, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ucasemap_internalUTF8Fold, edits, errorCode);
}

U_NAMESPACE_END
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// *******************************************************************************
// *   Copyright (C) 2011, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *******************************************************************************
// *   file name:  ucasemap_titlecase_brkiter.cpp
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2011jun02
// *   created by: Markus W. Scherer
// *
// *   Titlecasing functions that are based on BreakIterator
// *   were moved here to break dependency cycles among parts of the common library.
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/brkiter.h"
// #include "unicode/ubrk.h"
// #include "unicode/casemap.h"
// #include "unicode/ucasemap.h"
// #include "cmemory.h"
// #include "ucase.h"
// #include "ucasemap_imp.h"

// U_NAMESPACE_BEGIN

// void CaseMap::utf8ToTitle(
//         const char *locale, uint32_t options, BreakIterator *iter,
//         StringPiece src, ByteSink &sink, Edits *edits,
//         UErrorCode &errorCode) {
//     if (U_FAILURE(errorCode)) {
//         return;
//     }
//     UText utext = UTEXT_INITIALIZER;
//     utext_openUTF8(&utext, src.data(), src.length(), &errorCode);
//     LocalPointer<BreakIterator> ownedIter;
//     iter = ustrcase_getTitleBreakIterator(nullptr, locale, options, iter, ownedIter, errorCode);
//     if (iter == nullptr) {
//         utext_close(&utext);
//         return;
//     }
//     iter->setText(&utext, errorCode);
//     ucasemap_mapUTF8(
//         ustrcase_getCaseLocale(locale), options, iter,
//         src.data(), src.length(),
//         ucasemap_internalUTF8ToTitle, sink, edits, errorCode);
//     utext_close(&utext);
// }

// int32_t CaseMap::utf8ToTitle(
//         const char *locale, uint32_t options, BreakIterator *iter,
//         const char *src, int32_t srcLength,
//         char *dest, int32_t destCapacity, Edits *edits,
//         UErrorCode &errorCode) {
//     if (U_FAILURE(errorCode)) {
//         return 0;
//     }
//     UText utext=UTEXT_INITIALIZER;
//     utext_openUTF8(&utext, src, srcLength, &errorCode);
//     LocalPointer<BreakIterator> ownedIter;
//     iter = ustrcase_getTitleBreakIterator(nullptr, locale, options, iter, ownedIter, errorCode);
//     if(iter==NULL) {
//         utext_close(&utext);
//         return 0;
//     }
//     iter->setText(&utext, errorCode);
//     int32_t length=ucasemap_mapUTF8(
//         ustrcase_getCaseLocale(locale), options, iter,
//         dest, destCapacity,
//         src, srcLength,
//         ucasemap_internalUTF8ToTitle, edits, errorCode);
//     utext_close(&utext);
//     return length;
// }

// U_NAMESPACE_END

// U_NAMESPACE_USE

// U_CAPI const UBreakIterator * U_EXPORT2
// ucasemap_getBreakIterator(const UCaseMap *csm) {
//     return reinterpret_cast<UBreakIterator *>(csm->iter);
// }

// U_CAPI void U_EXPORT2
// ucasemap_setBreakIterator(UCaseMap *csm, UBreakIterator *iterToAdopt, UErrorCode *pErrorCode) {
//     if(U_FAILURE(*pErrorCode)) {
//         return;
//     }
//     delete csm->iter;
//     csm->iter=reinterpret_cast<BreakIterator *>(iterToAdopt);
// }

// U_CAPI int32_t U_EXPORT2
// ucasemap_utf8ToTitle(UCaseMap *csm,
//                      char *dest, int32_t destCapacity,
//                      const char *src, int32_t srcLength,
//                      UErrorCode *pErrorCode) {
//     if (U_FAILURE(*pErrorCode)) {
//         return 0;
//     }
//     UText utext=UTEXT_INITIALIZER;
//     utext_openUTF8(&utext, (const char *)src, srcLength, pErrorCode);
//     if (U_FAILURE(*pErrorCode)) {
//         return 0;
//     }
//     if(csm->iter==NULL) {
//         LocalPointer<BreakIterator> ownedIter;
//         BreakIterator *iter = ustrcase_getTitleBreakIterator(
//             nullptr, csm->locale, csm->options, nullptr, ownedIter, *pErrorCode);
//         if (iter == nullptr) {
//             utext_close(&utext);
//             return 0;
//         }
//         csm->iter = ownedIter.orphan();
//     }
//     csm->iter->setText(&utext, *pErrorCode);
//     int32_t length=ucasemap_mapUTF8(
//             csm->caseLocale, csm->options, csm->iter,
//             dest, destCapacity,
//             src, srcLength,
//             ucasemap_internalUTF8ToTitle, NULL, *pErrorCode);
//     utext_close(&utext);
//     return length;
// }

// #endif  // !UCONFIG_NO_BREAK_ITERATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2003, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: March 19 2003
* Since: ICU 2.6
**********************************************************************
*/
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2003-2004, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: March 19 2003
* Since: ICU 2.6
**********************************************************************
*/




/**
 * \file
 * \brief C API: Message Catalog Wrappers
 *
 * This C API provides look-alike functions that deliberately resemble
 * the POSIX catopen, catclose, and catgets functions.  The underlying
 * implementation is in terms of ICU resource bundles, rather than
 * POSIX message catalogs.
 *
 * The ICU resource bundles obey standard ICU inheritance policies.
 * To facilitate this, sets and messages are flattened into one tier.
 * This is done by creating resource bundle keys of the form
 * &lt;set_num&gt;%&lt;msg_num&gt; where set_num is the set number and msg_num is
 * the message number, formatted as decimal strings.
 *
 * Example:  Consider a message catalog containing two sets:
 *
 * Set 1: Message 4  = "Good morning."
 *        Message 5  = "Good afternoon."
 *        Message 7  = "Good evening."
 *        Message 8  = "Good night."
 * Set 4: Message 14 = "Please "
 *        Message 19 = "Thank you."
 *        Message 20 = "Sincerely,"
 *
 * The ICU resource bundle source file would, assuming it is named
 * "greet.txt", would look like this:
 *
 * greet
 * {
 *     1%4  { "Good morning." }
 *     1%5  { "Good afternoon." }
 *     1%7  { "Good evening." }
 *     1%8  { "Good night." }
 * 
 *     4%14 { "Please " }
 *     4%19 { "Thank you." }
 *     4%20 { "Sincerely," }
 * }
 *
 * The catgets function is commonly used in combination with functions
 * like printf and strftime.  ICU components like message format can
 * be used instead, although they use a different format syntax.
 * There is an ICU package, icuio, that provides some of
 * the POSIX-style formatting API.
 */

U_CDECL_BEGIN

/**
 * An ICU message catalog descriptor, analogous to nl_catd.
 * 
 * @stable ICU 2.6
 */
typedef UResourceBundle* u_nl_catd;

/**
 * Open and return an ICU message catalog descriptor. The descriptor
 * may be passed to u_catgets() to retrieve localized strings.
 *
 * @param name string containing the full path pointing to the
 * directory where the resources reside followed by the package name
 * e.g. "/usr/resource/my_app/resources/guimessages" on a Unix system.
 * If NULL, ICU default data files will be used.
 *
 * Unlike POSIX, environment variables are not interpolated within the
 * name.
 *
 * @param locale the locale for which we want to open the resource. If
 * NULL, the default ICU locale will be used (see uloc_getDefault). If
 * strlen(locale) == 0, the root locale will be used.
 *
 * @param ec input/output error code. Upon output,
 * U_USING_FALLBACK_WARNING indicates that a fallback locale was
 * used. For example, 'de_CH' was requested, but nothing was found
 * there, so 'de' was used. U_USING_DEFAULT_WARNING indicates that the
 * default locale data or root locale data was used; neither the
 * requested locale nor any of its fallback locales were found.
 *
 * @return a message catalog descriptor that may be passed to
 * u_catgets(). If the ec parameter indicates success, then the caller
 * is responsible for calling u_catclose() to close the message
 * catalog. If the ec parameter indicates failure, then NULL will be
 * returned.
 * 
 * @stable ICU 2.6
 */
U_STABLE u_nl_catd U_EXPORT2
u_catopen(const char* name, const char* locale, UErrorCode* ec);

/**
 * Close an ICU message catalog, given its descriptor.
 *
 * @param catd a message catalog descriptor to be closed. May be NULL,
 * in which case no action is taken.
 * 
 * @stable ICU 2.6
 */
U_STABLE void U_EXPORT2
u_catclose(u_nl_catd catd);

/**
 * Retrieve a localized string from an ICU message catalog.
 *
 * @param catd a message catalog descriptor returned by u_catopen.
 *
 * @param set_num the message catalog set number. Sets need not be
 * numbered consecutively.
 *
 * @param msg_num the message catalog message number within the
 * set. Messages need not be numbered consecutively.
 *
 * @param s the default string. This is returned if the string
 * specified by the set_num and msg_num is not found. It must be
 * zero-terminated.
 *
 * @param len fill-in parameter to receive the length of the result.
 * May be NULL, in which case it is ignored.
 *
 * @param ec input/output error code. May be U_USING_FALLBACK_WARNING
 * or U_USING_DEFAULT_WARNING. U_MISSING_RESOURCE_ERROR indicates that
 * the set_num/msg_num tuple does not specify a valid message string
 * in this catalog.
 *
 * @return a pointer to a zero-terminated UChar array which lives in
 * an internal buffer area, typically a memory mapped/DLL file. The
 * caller must NOT delete this pointer. If the call is unsuccessful
 * for any reason, then s is returned.  This includes the situation in
 * which ec indicates a failing error code upon entry to this
 * function.
 * 
 * @stable ICU 2.6
 */
U_STABLE const UChar* U_EXPORT2
u_catgets(u_nl_catd catd, int32_t set_num, int32_t msg_num,
          const UChar* s,
          int32_t* len, UErrorCode* ec);

U_CDECL_END






/* Separator between set_num and msg_num */
static const char SEPARATOR = '%';

/* Maximum length of a set_num/msg_num key, incl. terminating zero.
 * Longest possible key is "-2147483648%-2147483648" */
#define MAX_KEY_LEN (24)

/**
 * Fill in buffer with a set_num/msg_num key string, given the numeric
 * values. Numeric values must be >= 0. Buffer must be of length
 * MAX_KEY_LEN or more.
 */
static char*
_catkey(char* buffer, int32_t set_num, int32_t msg_num) {
    int32_t i = 0;
    i = T_CString_integerToString(buffer, set_num, 10);
    buffer[i++] = SEPARATOR;
    T_CString_integerToString(buffer+i, msg_num, 10);
    return buffer;
}

U_CAPI u_nl_catd U_EXPORT2
u_catopen(const char* name, const char* locale, UErrorCode* ec) {
    return (u_nl_catd) ures_open(name, locale, ec);
}

U_CAPI void U_EXPORT2
u_catclose(u_nl_catd catd) {
    ures_close((UResourceBundle*) catd); /* may be NULL */
}

U_CAPI const UChar* U_EXPORT2
u_catgets(u_nl_catd catd, int32_t set_num, int32_t msg_num,
          const UChar* s,
          int32_t* len, UErrorCode* ec) {

    char key[MAX_KEY_LEN];
    const UChar* result;

    if (ec == NULL || U_FAILURE(*ec)) {
        goto ucat_ERROR;
    }

    result = ures_getStringByKey((const UResourceBundle*) catd,
                                 _catkey(key, set_num, msg_num),
                                 len, ec);
    if (U_FAILURE(*ec)) {
        goto ucat_ERROR;
    }

    return result;

 ucat_ERROR:
    /* In case of any failure, return s */
    if (len != NULL) {
        *len = u_strlen(s);
    }
    return s;
}

/*eof*/
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 1996-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File UCHAR.C
*
* Modification History:
*
*   Date        Name        Description
*   04/02/97    aliu        Creation.
*   4/15/99     Madhu       Updated all the function definitions for C Implementation
*   5/20/99     Madhu       Added the function u_getVersion()
*   8/19/1999   srl         Upgraded scripts to Unicode3.0
*   11/11/1999  weiv        added u_isalnum(), cleaned comments
*   01/11/2000  helena      Renamed u_getVersion to u_getUnicodeVersion.
*   06/20/2000  helena      OS/400 port changes; mostly typecast.
******************************************************************************
*/













/* uchar_props_data.h is machine-generated by genprops --csource */
#define INCLUDED_FROM_UCHAR_C
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// Copyright (C) 1999-2016, International Business Machines
// Corporation and others.  All Rights Reserved.
//
// file name: uchar_props_data.h
//
// machine-generated by: icu/tools/unicode/c/genprops/corepropsbuilder.cpp


#ifdef INCLUDED_FROM_UCHAR_C

static const UVersionInfo dataVersion={0xd,0,0,0};

static const uint16_t propsTrie_index[22276]={
0x46d,0x475,0x47d,0x485,0x49d,0x4a5,0x4ad,0x4b5,0x4bd,0x4c5,0x4cb,0x4d3,0x4db,0x4e3,0x4eb,0x4f3,
0x4f9,0x501,0x509,0x511,0x514,0x51c,0x524,0x52c,0x534,0x53c,0x538,0x540,0x548,0x550,0x555,0x55d,
0x565,0x56d,0x571,0x579,0x581,0x589,0x591,0x599,0x595,0x59d,0x5a2,0x5aa,0x5b0,0x5b8,0x5c0,0x5c8,
0x5d0,0x5d8,0x5e0,0x5e8,0x5ed,0x5f5,0x5f8,0x600,0x608,0x610,0x616,0x61e,0x61d,0x625,0x62d,0x635,
0x645,0x63d,0x64d,0x655,0x48d,0x665,0x66b,0x65d,0x67b,0x67d,0x685,0x673,0x695,0x69b,0x6a3,0x68d,
0x6b3,0x6b9,0x6c1,0x6ab,0x6d1,0x6d7,0x6df,0x6c9,0x6ef,0x6f5,0x6fd,0x6e7,0x70d,0x715,0x71d,0x705,
0x72d,0x733,0x73b,0x725,0x74b,0x751,0x759,0x743,0x769,0x76e,0x776,0x761,0x786,0x78d,0x795,0x77e,
0x619,0x79d,0x7a5,0x48d,0x7ad,0x7b4,0x7bc,0x48d,0x7c4,0x7cc,0x7d4,0x7d9,0x7e1,0x7e8,0x7f0,0x48d,
0x5d8,0x7f8,0x800,0x808,0x810,0x565,0x820,0x818,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x828,0x5d8,0x830,0x834,0x83c,0x5d8,0x842,0x5d8,0x848,0x850,0x858,0x565,0x565,0x860,
0x868,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x86d,0x875,0x5d8,0x5d8,0x87d,0x885,0x88d,0x895,0x89d,0x5d8,0x8a5,0x8ad,0x8b5,
0x8c5,0x5d8,0x8cd,0x8cf,0x8d7,0x8bd,0x5d8,0x8da,0x8ee,0x8e2,0x8ea,0x8f6,0x5d8,0x8fe,0x904,0x90c,
0x914,0x5d8,0x924,0x92c,0x934,0x91c,0x944,0x48d,0x94c,0x94f,0x957,0x93c,0x967,0x95f,0x5d8,0x96e,
0x5d8,0x97d,0x976,0x985,0x98d,0x991,0x999,0x9a1,0x50d,0x9a9,0x9ac,0x9b2,0x9b9,0x9ac,0x534,0x9c1,
0x4bd,0x4bd,0x4bd,0x4bd,0x9c9,0x4bd,0x4bd,0x4bd,0x9d9,0x9e1,0x9e9,0x9f1,0x9f9,0x9fd,0xa05,0x9d1,
0xa1d,0xa25,0xa0d,0xa15,0xa2d,0xa35,0xa3d,0xa45,0xa5d,0xa4d,0xa55,0xa65,0xa6d,0xa7c,0xa81,0xa74,
0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa91,0xa99,0x90c,0xa9c,0xaa4,0xaab,0xab0,0xab8,
0x90c,0xabf,0xabe,0xacf,0xad2,0x90c,0x90c,0xac7,0x90c,0x90c,0x90c,0x90c,0x90c,0xae1,0xae9,0xad9,
0x90c,0x90c,0x90c,0xaee,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0xaf4,0xafc,0x90c,0xb04,0xb0b,
0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0xa89,0xa89,0xa89,0xa89,0xb13,0xa89,0xb1a,0xb21,
0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0x90c,0xb29,0xb30,0xb34,0xb3a,0x90c,0x90c,0x90c,
0x565,0xb4a,0xb42,0xb52,0x4bd,0x4bd,0x4bd,0xb5a,0x50d,0xb62,0x5d8,0xb68,0xb78,0xb70,0xb70,0x534,
0xb80,0xb88,0xb90,0x48d,0xb98,0x90c,0x90c,0xb9f,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0xba7,0xbad,
0xbbd,0xbb5,0x619,0x5d8,0xbc5,0x868,0x5d8,0xbcd,0xbd5,0xbd9,0x5d8,0x5d8,0xbde,0x5d8,0x90c,0xbe5,
0xab9,0xbed,0xbf3,0x90c,0xbed,0xbfb,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,
0xc03,0x5d8,0x5d8,0x5d8,0xc0b,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0xc11,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xc16,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x90c,0x90c,
0xc1e,0x5d8,0xc21,0x5d8,0xc29,0xc2f,0xc37,0xc3f,0xc44,0x5d8,0x5d8,0xc48,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xc4f,0x5d8,0xc56,0xc5c,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xc64,0x5d8,0x5d8,0x5d8,0xc6c,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xc6e,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xc75,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0xc7c,0x5d8,0x5d8,0x5d8,0xc83,0xc8b,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xc90,0x5d8,0x5d8,0xc98,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xc9c,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xc9f,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xca2,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0xca8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0xcb0,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0xcb5,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xcba,0x5d8,0x5d8,0x5d8,0xcbf,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0xcc7,0xcce,0xcd2,0x5d8,0x5d8,0x5d8,0xcd9,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x8ce,
0xce7,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0xcdf,0x90c,0xcef,0x985,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0xcf4,0xcfc,0x4bd,0xd0c,0xd04,0x5d8,0x5d8,0xd14,0xd1c,0xd2c,0x4bd,0xd31,0xd39,0xd3f,0xd47,0xd24,
0xd4f,0xd57,0x5d8,0xd5f,0xd6f,0xd72,0xd67,0xd7a,0x62d,0xd82,0xd89,0x8ce,0x67b,0xd99,0xd91,0xda1,
0x5d8,0xda9,0xdb1,0xdb9,0x5d8,0xdc1,0xdc9,0xdd1,0xdd9,0xde1,0xde5,0xded,0x50d,0x50d,0x5d8,0xdf5,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xdfd,0xe09,0xe01,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,
0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0x5d8,0x5d8,0x5d8,0xe21,0x5d8,0xcda,0xe28,0xe2d,
0x5d8,0x5d8,0x5d8,0xe35,0x5d8,0x5d8,0x8d9,0x48d,0xe4b,0xe3b,0xe43,0x5d8,0x5d8,0xe53,0xe5b,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xe60,0xe68,0x5d8,0xe6c,0x5d8,0xe72,0xe76,
0xe7e,0xe86,0xe8d,0xe95,0x5d8,0x5d8,0x5d8,0xe9b,0xeb3,0x47d,0xebb,0xec3,0xec8,0x8ee,0xea3,0xeab,
0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,
0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,0xe11,
0x1234,0x1234,0x1274,0x12b4,0x12f4,0x132c,0x136c,0x13ac,0x13e4,0x1424,0x1450,0x1490,0x14d0,0x14e0,0x1520,0x1554,
0x1594,0x15c4,0x1604,0x1644,0x1654,0x1688,0x16c0,0x1700,0x1740,0x1780,0x17b4,0x17e0,0x1820,0x1858,0x1874,0x18b4,
0xa80,0xac0,0xb00,0xb40,0xb80,0xa40,0xbc0,0xa40,0xbe2,0xa40,0xa40,0xa40,0xa40,0xc22,0x1db,0x1db,
0xc62,0xca2,0xa40,0xa40,0xa40,0xa40,0xce2,0xd02,0xa40,0xa40,0xd42,0xd82,0xdc2,0xe02,0xe42,0xe82,
0xec2,0xef9,0x1db,0x1db,0xf1d,0xf51,0x1db,0xf79,0x1db,0x1db,0x1db,0x1db,0xfa6,0x1db,0x1db,0x1db,
0x1db,0x1db,0x1db,0x1db,0xfba,0x1db,0xff2,0x1032,0x1db,0x103d,0x1db,0x1db,0x1db,0x1073,0xa40,0x10b3,
0x1db,0x1db,0x10f3,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0x1133,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,
0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x1173,
0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,
0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x700,0x1173,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0xed0,0xed7,0xedf,0x48d,0x5d8,0x5d8,0x5d8,0xee7,0xef7,0xeef,0xf0e,0xeff,0xf06,0xf16,0xf1a,0xf1e,
0x48d,0x48d,0x48d,0x48d,0x8ce,0x5d8,0xf26,0xf2e,0x5d8,0xf36,0xf3e,0xf42,0xf4a,0x5d8,0xf52,0x48d,
0x565,0x56f,0xf5a,0x5d8,0xf5e,0xf66,0xf76,0xf6e,0x5d8,0xf7e,0x5d8,0xf85,0x48d,0x48d,0x48d,0x48d,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xb78,0x8da,0xe72,0x48d,0x48d,0x48d,0x48d,
0xf95,0xf8d,0xf98,0xfa0,0x8ee,0xfa8,0x48d,0xfb0,0xfb8,0xfc0,0x48d,0x48d,0x5d8,0xfd0,0xfd8,0xfc8,
0xfe8,0xfef,0xfe0,0xff7,0xfff,0x48d,0x100f,0x1007,0x5d8,0x1012,0x101a,0x1022,0x102a,0x1032,0x48d,0x48d,
0x5d8,0x5d8,0x103a,0x48d,0x565,0x1042,0x50d,0x104a,0x5d8,0x1052,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x105a,0x5d8,0x1062,0x48d,0x48d,0x106a,0x1072,0x1079,0x48d,0x48d,0xe68,0x1081,0xb78,
0x1091,0x60e,0x1099,0x1089,0x967,0x10a1,0x10a9,0x10af,0x10c7,0x10b7,0x10bf,0x10cb,0x967,0x10db,0x10d3,0x10e3,
0x10f3,0x10eb,0x48d,0x48d,0x10fa,0x1102,0x630,0x110a,0x111a,0x1120,0x1128,0x1112,0x48d,0x48d,0x48d,0x48d,
0x5d8,0x1130,0x1138,0x1140,0x5d8,0x1148,0x1150,0x48d,0x48d,0x48d,0x48d,0x48d,0x5d8,0x1158,0x1160,0x48d,
0x5d8,0x1168,0x1170,0x1178,0x5d8,0x1188,0x1180,0x48d,0x848,0x1190,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x5d8,0x1198,0x48d,0x48d,0x48d,0x565,0x50d,0x11a0,0x11b0,0x11b6,0x11a8,0x48d,0x48d,0x11c6,0x11ca,0x11be,
0x11e2,0x11d2,0x11da,0x5d8,0x11f2,0x11ea,0x5d8,0x8cf,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x1208,0x120d,0x11fa,0x1202,0x121d,0x1215,0x48d,0x48d,0x122c,0x1230,0x1224,0x1240,0x1238,0x1180,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x1244,0x48d,0x48d,0x48d,0x48d,0x48d,0x124b,0x125b,0x1253,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x8d9,0x48d,0x48d,0x48d,
0x126b,0x1273,0x127b,0x1263,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x1283,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x128b,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x1293,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x8cf,0x8ee,0x129b,0x48d,0x48d,0xe68,0x12a3,0x5d8,0x12b3,0x12bb,0x12c3,0x12ab,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x565,0x50d,0x12cb,0x48d,0x48d,0x48d,0x5d8,0x5d8,0x12d3,0x12d8,0x12de,0x48d,
0x48d,0x12e6,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x12ee,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x8da,0x48d,0x103a,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x8ee,0x48d,0x12f4,0x12fb,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xe01,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x5d8,0x5d8,0x5d8,0x1301,0x1306,0x130e,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0xba7,0x90c,0x1316,0x90c,0x131d,0x1325,0x132b,
0x90c,0x1331,0x90c,0x90c,0x1339,0x48d,0x48d,0x48d,0x48d,0x1341,0x90c,0x90c,0xabb,0x1349,0x48d,0x48d,
0x48d,0x48d,0x1359,0x1360,0x1365,0x136b,0x1373,0x137b,0x1383,0x135d,0x138b,0x1393,0x139b,0x13a0,0x1372,0x1359,
0x1360,0x135c,0x136b,0x13a8,0x135a,0x13ab,0x135d,0x13b3,0x13bb,0x13c3,0x13ca,0x13b6,0x13be,0x13c6,0x13cd,0x13b9,
0x13d5,0x1351,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,
0x90c,0x90c,0x534,0x13e5,0x534,0x13ec,0x13f3,0x13dd,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x13fa,0x1402,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x5d8,0x1412,0x140a,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x5d8,0x141a,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x1422,0x48d,0x565,0x1432,0x142a,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x143a,0x144a,0x1442,0x48d,0x48d,0x145a,0x1452,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x146a,0x1472,0x147a,0x1482,0x148a,0x1492,0x48d,0x1462,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x90c,0x149a,0x90c,0x90c,0xb9f,0x149f,0x14a3,0xba7,0x14ab,0x90c,0x90c,0x90c,0x90c,0xba9,
0x48d,0x14b3,0x14bb,0x14bf,0x14c7,0x14cf,0x48d,0x48d,0x48d,0x48d,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,
0x90c,0x14d7,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,
0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x90c,0x14df,0x14e7,0x90c,0x90c,0x90c,0xb9f,0x90c,0x90c,
0x14ef,0x14f7,0x149a,0x90c,0x14ff,0x90c,0x1507,0x150c,0x48d,0x48d,0x90c,0x90c,0x90c,0x1514,0x90c,0x90c,
0x151b,0x90c,0x90c,0x90c,0xb9f,0x1520,0x1528,0x152e,0x1533,0x48d,0x90c,0x90c,0x90c,0x90c,0x153b,0x90c,
0xabe,0x117c,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x1543,0x5d8,0x5d8,0x154a,0x5d8,0x5d8,0x5d8,0x1552,0x5d8,0x155a,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0xc80,0x5d8,0x5d8,0x1562,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x156a,0x1572,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0xcbf,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x1579,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x1580,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x1587,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0xf5e,0x48d,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x158b,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0xf5e,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x1066,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x1590,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x5d8,0x5d8,0x5d8,0x5d8,0x1598,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0xf5e,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,
0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x5d8,0x655,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x15a8,0x15a0,0x15a0,0x15a0,0x48d,0x48d,0x48d,0x48d,0x534,0x534,0x534,0x534,0x534,
0x534,0x534,0x15b0,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,0x48d,
0x48d,0x48d,0x48d,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,0xe19,
0xe19,0xe19,0x15b8,0x46c,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,
0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,
0xf,0xf,0xf,0xf,0xc,0x17,0x17,0x17,0x19,0x17,0x17,0x17,0x14,0x15,0x17,0x18,
0x17,0x13,0x17,0x17,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x17,0x17,
0x18,0x18,0x18,0x17,0x17,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0x14,
0x17,0x15,0x1a,0x16,0x1a,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,0x14,
0x18,0x15,0x18,0xf,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,
0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,0xf,
0xf,0xf,0xf,0xf,0xc,0x17,0x19,0x19,0x19,0x19,0x1b,0x17,0x1a,0x1b,5,0x1c,
0x18,0x10,0x1b,0x1a,0x1b,0x18,0x34b,0x38b,0x1a,2,0x17,0x17,0x1a,0x30b,5,0x1d,
0x34cb,0x344b,0x3ccb,0x17,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,0x18,1,1,1,1,
1,1,1,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,0x18,2,2,2,2,
2,2,2,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,2,1,2,1,
2,1,2,1,2,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,1,2,1,2,1,2,2,2,1,1,2,
1,2,1,1,2,1,1,1,2,2,1,1,1,1,2,1,
1,2,1,1,1,2,2,2,1,1,2,1,1,2,1,2,
1,2,1,1,2,1,2,2,1,2,1,1,2,1,1,1,
2,1,2,1,1,2,2,5,1,2,2,2,5,5,5,5,
1,3,2,1,3,2,1,3,2,1,2,1,2,1,2,1,
2,1,2,1,2,1,2,1,2,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,2,1,3,2,
1,2,1,1,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,2,2,2,2,2,2,1,1,
2,1,1,2,2,1,2,1,1,1,1,2,1,2,1,2,
1,2,1,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,5,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,0x1a,0x1a,0x1a,0x1a,4,4,4,4,4,4,4,4,4,4,
4,4,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,
4,4,4,4,4,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,4,0x1a,4,0x1a,
0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
1,2,1,2,4,0x1a,1,2,0,0,4,2,2,2,0x17,1,
0,0,0,0,0x1a,0x1a,1,0x17,1,1,1,0,1,0,1,1,
2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,0,1,1,1,1,1,1,1,1,1,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,1,2,2,1,1,1,2,2,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,2,2,2,2,1,2,0x18,1,2,1,1,2,
2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,0x1b,6,6,6,6,6,7,7,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,1,2,1,2,1,2,1,2,1,2,1,
2,1,2,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,0,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,
0,4,0x17,0x17,0x17,0x17,0x17,0x17,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,0x17,0x13,0,0,0x1b,0x1b,0x19,
0,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,0x13,6,
0x17,6,6,0x17,6,6,0x17,6,0,0,0,0,0,0,0,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,0,0,0,0,5,
5,5,5,0x17,0x17,0,0,0,0,0,0,0,0,0,0,0,
0x10,0x10,0x10,0x10,0x10,0x10,0x18,0x18,0x18,0x17,0x17,0x19,0x17,0x17,0x1b,0x1b,
6,6,6,6,6,6,6,6,6,6,6,0x17,0x10,0,0x17,0x17,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
4,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x17,0x17,0x17,0x17,5,5,
6,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0x17,5,6,6,6,6,6,6,6,0x10,0x1b,6,
6,6,6,6,6,4,4,6,6,0x1b,6,6,6,6,5,5,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,5,5,5,0x1b,0x1b,5,
0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0,0x10,
5,6,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,0,0,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,
6,6,6,6,6,5,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,
6,6,6,6,6,6,6,6,4,4,0x1b,0x17,0x17,0x17,4,0,
0,6,0x19,0x19,6,6,6,6,4,6,6,6,4,6,6,6,
6,6,0,0,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,
0x17,0x17,0x17,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,6,6,6,6,4,6,
6,6,6,6,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,
0,0,0x17,0,5,5,5,5,5,5,5,5,5,5,5,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,6,6,0x10,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0,5,5,5,5,5,5,
5,5,5,5,0,0,0,0,0,0,0,0,0,0,0,6,
6,6,6,6,6,6,6,6,6,6,6,6,5,5,6,6,
0x17,0x17,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x17,4,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,8,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,8,
6,5,8,8,8,6,6,6,6,6,6,6,6,8,8,8,
8,6,8,8,5,6,6,6,6,6,6,6,5,5,5,5,
5,5,5,5,5,5,6,6,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,
0x1c9,0x209,0x249,0x289,5,5,0x19,0x19,0x37cb,0x35cb,0x3fcb,0x34cb,0x3ccb,0x94b,0x1b,0x19,
5,0x17,6,0,5,6,8,8,0,5,5,5,5,5,5,5,
5,0,0,5,5,0,0,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,0,5,5,5,5,5,5,5,0,5,0,
0,0,5,5,5,5,0,0,6,5,8,8,8,6,6,6,
6,0,0,8,8,0,0,8,8,6,5,0,0,0,0,0,
0,0,0,8,0,0,0,0,5,5,0,5,0,0,0,0,
0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,6,6,5,5,
5,6,0x17,0,0,0,0,0,0,0,0,0,0,6,6,8,
0,5,5,5,5,5,5,0,0,0,0,5,5,0,0,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,0,5,5,
5,5,5,5,5,0,5,5,0,5,5,0,5,5,0,0,
6,0,8,8,8,6,6,0,0,0,0,6,6,0,0,6,
6,6,0,0,0,6,0,0,0,0,0,0,0,5,5,5,
5,0,5,0,5,5,6,6,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,
0x1c9,0x209,0x249,0x289,0x17,0x19,0,0,0,0,0,0,0,5,6,6,
6,6,6,6,0,6,6,8,0,5,5,5,5,5,5,5,
5,5,0,5,5,5,0,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,0,5,5,5,5,5,5,5,0,5,5,
0,5,5,5,5,5,0,0,6,5,8,8,8,6,6,6,
6,6,0,6,6,8,0,8,8,6,0,0,5,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,5,5,6,6,
0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x1b,5,0x34cb,0x344b,
0x3ccb,0x37cb,0x35cb,0x3fcb,0,0,0,0,0,0,0,0,0,6,8,8,
0,5,5,5,5,5,5,5,5,0,0,5,5,0,0,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,0,5,5,
5,5,5,5,5,0,5,5,0,5,5,5,5,5,0,0,
6,5,8,6,8,6,6,6,6,0,0,8,8,0,0,8,
8,6,0,0,0,0,0,0,0,6,6,8,0,0,0,0,
5,5,0,5,0,0,0,0,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,
0x1c9,0x209,0x249,0x289,0x7cb,0x1e4b,0x784b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x19,0x1b,0,
0,0,0,0,0,0,6,5,0,5,5,5,5,5,5,0,
0,0,5,5,5,0,5,5,5,5,0,0,0,5,5,0,
5,0,5,5,0,0,0,5,5,0,0,0,5,5,5,0,
0,0,5,5,5,5,5,5,5,5,5,5,5,5,0,0,
0,0,8,8,6,8,8,0,0,0,8,8,8,0,8,8,
8,6,0,0,5,0,0,0,0,0,0,8,0,0,0,0,
0,0,0,0,5,5,6,6,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,
0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,0,0x17,0x54b,0x58b,0x5cb,0x60b,
0x58b,0x5cb,0x60b,0x1b,6,8,8,8,6,5,5,5,5,5,5,5,
5,0,5,5,5,0,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,0,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,0,0,0,5,6,6,6,8,8,8,
8,0,6,6,6,0,6,6,6,6,0,0,0,0,0,0,
0,6,6,0,5,5,5,0,0,0,0,0,5,5,6,6,
0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,5,5,0,
0,0,0,0,0,0,0,0,0,0,0,0,5,6,8,8,
0x17,5,5,5,5,5,5,5,5,0,5,5,5,0,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,0,5,5,
5,5,5,5,5,5,5,5,0,5,5,5,5,5,0,0,
6,5,8,6,8,8,8,8,8,0,6,8,8,0,8,8,
6,6,0,0,0,0,0,0,0,8,8,0,0,0,0,0,
0,0,5,0,5,5,6,6,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,
0x1c9,0x209,0x249,0x289,0x7cb,0x1e4b,0x784b,0x34cb,0x344b,0x3ccb,0x37cb,0x35cb,0x3fcb,0x1b,5,5,
5,5,5,5,6,6,8,8,5,5,5,5,5,5,5,5,
5,0,5,5,5,0,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,6,6,5,8,8,8,6,6,6,6,0,8,8,
8,0,8,8,8,6,5,0x1b,0,0,0,0,5,5,5,8,
0xcc0b,0xca0b,0xcb4b,0xc90b,0x364b,0xc94b,0x350b,5,0,0,0,0,0,0,0x49,0x89,
0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,8,8,0x17,0,0,0,
0,0,0,0,0,0,0,0,0,6,8,8,0,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,0,
0,0,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,0,5,5,5,5,5,5,5,5,5,
0,5,0,0,5,5,5,5,5,5,5,0,0,0,6,0,
0,0,0,8,8,8,6,6,6,0,6,0,8,8,8,8,
8,8,8,8,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,6,5,5,6,6,6,6,6,6,6,0,
0,0,0,0x19,5,5,5,5,5,5,4,6,6,6,6,6,
6,6,6,0x17,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x17,0x17,
0,0,0,0,0,5,5,0,5,0,5,5,5,5,5,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0,5,0,5,5,5,5,5,5,5,5,5,
5,6,5,5,6,6,6,6,6,6,6,6,6,5,0,0,
5,5,5,5,5,0,4,0,6,6,6,6,6,6,0,0,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,5,5,5,5,
5,0x1b,0x1b,0x1b,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,
0x17,0x17,0x17,0x1b,0x17,0x1b,0x1b,0x1b,6,6,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x344b,0x3c4b,0x444b,0x4c4b,0x544b,0x5c4b,
0x644b,0x6c4b,0x744b,0x2c4b,0x1b,6,0x1b,6,0x1b,6,0x14,0x15,0x14,0x15,8,8,
5,5,5,5,5,5,5,5,0,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,0,0,0,0,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,8,6,6,6,6,6,0x17,6,6,5,5,5,5,
5,6,6,6,6,6,6,6,6,6,6,6,0,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,0,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,6,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0x1b,0x1b,
0x17,0x17,0x17,0x17,0x17,0x1b,0x1b,0x1b,0x1b,0x17,0x17,0,0,0,0,0,
5,5,5,5,5,5,5,5,5,5,5,8,8,6,6,6,
6,8,6,6,6,6,6,6,8,6,6,8,8,6,6,5,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x17,0x17,0x17,0x17,0x17,0x17,
5,5,5,5,5,5,8,8,6,6,5,5,5,5,6,6,
6,5,8,8,8,5,5,8,8,8,8,8,8,8,5,5,
5,6,6,6,6,5,5,5,5,5,5,5,5,5,5,5,
5,5,6,8,8,6,6,8,8,8,8,8,8,6,5,8,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,8,8,8,6,0x1b,0x1b,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,0x17,4,2,2,2,
1,1,1,1,1,1,0,1,0,0,0,0,0,1,0,0,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
5,5,5,5,5,5,5,5,5,0,5,5,5,5,0,0,
5,5,5,5,5,5,5,0,5,0,5,5,5,5,0,0,
5,5,5,5,5,5,5,5,5,0,5,5,5,5,0,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,0,5,5,5,5,0,0,5,5,5,5,5,5,5,0,
5,0,5,5,5,5,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0,5,5,5,5,0,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,0,0,6,6,6,
0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x30b,0x34b,0x38b,0x3cb,0x40b,0x44b,0x48b,
0x4cb,0x50b,0x7cb,0xa4b,0xccb,0xf4b,0x11cb,0x144b,0x16cb,0x194b,0x1bcb,0x1e4b,0x788b,0,0,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0,0,0,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,0,0,2,2,2,2,2,2,0,0,
0x13,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,0x1b,0x17,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0xc,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x14,
0x15,0,0,0,5,5,5,5,5,5,5,5,5,5,5,0x17,
0x17,0x17,0x98a,0x9ca,0xa0a,5,5,5,5,5,5,5,5,0,0,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,0,5,5,5,5,6,6,6,0,0,0,0,0,0,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,6,6,6,0x17,0x17,0,0,0,0,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,6,6,0,0,0,0,0,0,0,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,0,5,5,5,0,6,6,0,0,0,0,0,0,0,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,6,6,8,6,6,6,6,6,
6,6,8,8,8,8,8,8,8,8,6,8,8,6,6,6,
6,6,6,6,6,6,6,6,0x17,0x17,0x17,4,0x17,0x17,0x17,0x19,
5,6,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,
0,0,0,0,0x54b,0x58b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,0,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,6,5,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0x17,0x17,0x17,0x17,0x17,0x17,0x13,0x17,0x17,0x17,0x17,6,
6,6,0x10,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,
0,0,0,0,5,5,5,4,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,0,0,0,0,0,0,0,5,5,5,5,
5,6,6,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0,0,
0,0,0,0,0,0,0,0,6,6,6,8,8,8,8,6,
6,8,8,8,0,0,0,0,8,8,6,8,8,8,8,8,
8,6,6,6,0,0,0,0,0x1b,0,0,0,0x17,0x17,0x49,0x89,
0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,0,0,5,5,5,5,5,0,0,0,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0,0,0,0,0,0,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x30b,0,0,0,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,6,6,8,8,6,0,0,0x17,0x17,
0x17,0x17,0x17,0x17,0x17,0x17,0x17,4,0x17,0x17,0x17,0x17,0x17,0x17,0,0,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,7,6,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,8,6,8,6,6,6,6,6,6,6,0,
6,8,6,8,8,6,6,6,6,6,6,6,6,8,8,8,
8,8,8,6,6,6,6,6,6,6,6,6,6,0,0,6,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,
0x17,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,6,6,6,6,6,
6,6,6,6,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,
6,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
6,6,6,6,8,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
6,8,6,6,6,6,6,8,6,8,8,8,8,8,6,8,
8,5,5,5,5,5,5,5,0,0,0,0,0x49,0x89,0xc9,0x109,
0x149,0x189,0x1c9,0x209,0x249,0x289,0x17,0x17,0x17,0x17,0x17,0x17,5,8,6,6,
6,6,8,8,6,6,8,6,6,6,5,5,0x49,0x89,0xc9,0x109,
0x149,0x189,0x1c9,0x209,0x249,0x289,5,5,5,5,5,5,6,6,8,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,8,
6,6,8,8,8,6,8,6,6,6,8,8,0,0,0,0,
0,0,0,0,0x17,0x17,0x17,0x17,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,
0x249,0x289,0,0,0,5,5,5,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,
0x249,0x289,5,5,5,5,5,5,8,8,8,8,8,8,8,8,
6,6,6,6,6,6,6,6,8,8,6,6,0,0,0,0x17,
0x17,0x17,0x17,0x17,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,4,4,4,4,
4,4,0x17,0x17,2,2,2,2,2,2,2,2,2,0,0,0,
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,
0,1,1,1,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0,0,0,0,
0,0,0,0,6,6,6,0x17,6,6,6,6,6,6,6,6,
6,6,6,6,6,8,6,6,6,6,6,6,6,5,5,5,
5,6,5,5,5,5,5,5,6,5,5,8,6,6,5,0,
0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
4,4,4,2,2,2,2,2,2,2,2,2,2,2,2,2,
4,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,4,
4,4,4,4,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,0,6,
6,6,6,6,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,2,2,2,2,2,2,
2,2,1,2,2,2,2,2,2,2,2,2,1,1,1,1,
1,0x1a,0x1a,0x1a,0,0,2,2,2,0,2,2,1,1,1,1,
3,0x1a,0x1a,0,2,2,2,2,2,2,2,2,1,1,1,1,
1,1,1,1,2,2,2,2,2,2,0,0,1,1,1,1,
1,1,0,0,2,2,2,2,2,2,2,2,1,1,1,1,
1,1,1,1,2,2,2,2,2,2,2,2,1,1,1,1,
1,1,1,1,2,2,2,2,2,2,0,0,1,1,1,1,
1,1,0,0,2,2,2,2,2,2,2,2,0,1,0,1,
0,1,0,1,2,2,2,2,2,2,2,2,1,1,1,1,
1,1,1,1,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,0,0,2,2,2,2,2,2,2,2,3,3,3,3,
3,3,3,3,2,2,2,2,2,2,2,2,3,3,3,3,
3,3,3,3,2,2,2,2,2,0,2,2,1,1,1,1,
3,0x1a,2,0x1a,0x1a,0x1a,2,2,2,0,2,2,1,1,1,1,
3,0x1a,0x1a,0x1a,2,2,2,2,0,0,2,2,1,1,1,1,
0,0x1a,0x1a,0x1a,0x16,0x17,0x17,0x17,0x18,0x14,0x15,0x17,0x17,0x17,0x17,0x17,
0x17,0x17,0x17,0x17,0x17,0x17,0x18,0x17,0x16,0x17,0x17,0x17,0x17,0x17,0x17,0x17,
0x17,0x17,0x17,0xc,0x10,0x10,0x10,0x10,0x10,0,0x10,0x10,0x10,0x10,0x10,0x10,
0x10,0x10,0x10,0x10,0x2cb,4,0,0,0x3cb,0x40b,0x44b,0x48b,0x4cb,0x50b,0x18,0x18,
0x18,0x14,0x15,4,0xc,0xc,0xc,0xc,0xc,0xc,0xc,0xc,0xc,0xc,0xc,0x10,
0x10,0x10,0x10,0x10,0x13,0x13,0x13,0x13,0x13,0x13,0x17,0x17,0x1c,0x1d,0x14,0x1c,
0x1c,0x1d,0x14,0x1c,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0xd,0xe,0x10,0x10,
0x10,0x10,0x10,0xc,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x1c,0x1d,0x17,
0x17,0x17,0x17,0x16,0x2cb,0x30b,0x34b,0x38b,0x3cb,0x40b,0x44b,0x48b,0x4cb,0x50b,0x18,0x18,
0x18,0x14,0x15,0,4,4,4,4,4,4,4,4,4,4,4,4,
4,0,0,0,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,
0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,0x19,
0x19,0x19,0x19,0x19,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,6,6,6,6,6,6,6,6,6,6,6,6,
6,7,7,7,7,6,7,7,7,6,6,6,6,6,6,6,
6,6,6,6,6,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x1b,0x1b,0x1b,0x1b,1,0x1b,1,0x1b,1,0x1b,1,1,
1,1,0x1b,2,1,1,1,1,2,5,5,5,5,2,0x1b,0x1b,
2,2,1,1,0x18,0x18,0x18,0x18,0x18,1,2,2,2,2,0x1b,0x18,
0x1b,0x1b,2,0x1b,0x358b,0x360b,0x364b,0x348b,0x388b,0x350b,0x390b,0x3d0b,0x410b,0x354b,0x454b,0x35cb,
0x3dcb,0x45cb,0x4dcb,0x58b,0x1b,0x1b,1,0x1b,0x1b,0x1b,0x1b,1,0x1b,0x1b,2,1,
1,1,2,2,1,1,1,2,0x1b,1,0x1b,0x1b,0x18,1,1,1,
1,1,0x1b,0x1b,0x58a,0x5ca,0x60a,0x64a,0x68a,0x6ca,0x70a,0x74a,0x78a,0x7ca,0x80a,0x84a,
0x11ca,0x1e4a,0x980a,0x784a,0x58a,0x5ca,0x60a,0x64a,0x68a,0x6ca,0x70a,0x74a,0x78a,0x7ca,0x80a,0x84a,
0x11ca,0x1e4a,0x980a,0x784a,0x784a,0x984a,0x788a,1,2,0x6ca,0x11ca,0x988a,0x78ca,0x54b,0x1b,0x1b,
0,0,0,0,0x18,0x18,0x18,0x18,0x18,0x1b,0x1b,0x1b,0x1b,0x1b,0x18,0x18,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x18,0x1b,0x1b,0x18,0x1b,0x1b,0x18,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x18,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x18,0x18,0x1b,0x1b,0x18,0x1b,0x18,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x14,0x15,0x14,0x15,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x18,0x18,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x14,0x15,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x18,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x18,0x18,0x18,0x18,
0x18,0x18,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x2cb,0x80b,0x84b,0x88b,0x8cb,0x90b,0x94b,0x98b,0x9cb,0xa0b,
0xa4b,0x30b,0x34b,0x38b,0x3cb,0x40b,0x44b,0x48b,0x4cb,0x50b,0x7cb,0x2cb,0x30b,0x34b,0x38b,0x3cb,
0x40b,0x44b,0x48b,0x4cb,0x50b,0x7cb,0x80b,0x84b,0x88b,0x8cb,0x90b,0x94b,0x98b,0x9cb,0xa0b,0xa4b,
0x30b,0x34b,0x38b,0x3cb,0x40b,0x44b,0x48b,0x4cb,0x50b,0x7cb,0x80b,0x84b,0x88b,0x8cb,0x90b,0x94b,
0x98b,0x9cb,0xa0b,0xa4b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x18,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x18,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x18,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x14,0x15,0x14,0x15,0x14,0x15,0x14,0x15,
0x14,0x15,0x14,0x15,0x14,0x15,0x30b,0x34b,0x38b,0x3cb,0x40b,0x44b,0x48b,0x4cb,0x50b,0x7cb,
0x30b,0x34b,0x38b,0x3cb,0x40b,0x44b,0x48b,0x4cb,0x50b,0x7cb,0x30b,0x34b,0x38b,0x3cb,0x40b,0x44b,
0x48b,0x4cb,0x50b,0x7cb,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x18,0x18,0x18,0x18,0x18,0x14,0x15,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x14,0x15,0x14,0x15,0x14,0x15,0x14,0x15,0x14,0x15,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x14,
0x15,0x14,0x15,0x14,0x15,0x14,0x15,0x14,0x15,0x14,0x15,0x14,0x15,0x14,0x15,0x14,
0x15,0x14,0x15,0x14,0x15,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x14,0x15,0x14,0x15,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x14,0x15,0x18,0x18,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x18,0x1b,0x1b,0x18,0x18,0x18,0x18,0x18,0x18,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,0,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,0,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,1,2,1,1,1,2,2,1,
2,1,2,1,2,1,1,1,1,2,1,2,2,1,2,2,
2,2,2,2,4,4,1,1,1,2,1,2,2,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,1,2,1,2,6,6,6,1,2,0,0,0,0,
0,0x17,0x17,0x17,0x17,0x344b,0x17,0x17,2,2,2,2,2,2,0,2,
0,0,0,0,0,2,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,0,0,0,0,0,0,0,4,
0x17,0,0,0,0,0,0,0,0,0,0,0,0,0,0,6,
5,5,5,5,5,5,5,0,5,5,5,5,5,5,5,0,
5,5,5,5,5,5,5,0,5,5,5,5,5,5,5,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,0,0,0,0,0,0,0,0,0,
0x17,0x17,0x1c,0x1d,0x1c,0x1d,0x17,0x17,0x17,0x1c,0x1d,0x17,0x1c,0x1d,0x17,0x17,
0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x13,0x17,0x17,0x13,0x17,0x1c,0x1d,0x17,0x17,
0x1c,0x1d,0x14,0x15,0x14,0x15,0x14,0x15,0x14,0x15,0x17,0x17,0x17,0x17,0x17,4,
0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x13,0x13,0x17,0x17,0x17,0x17,
0x13,0x17,0x14,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,
0x1b,0x1b,0x17,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0,0,0,0,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0,0,0,0,0x1b,0x58a,0x5ca,0x60a,0x64a,0x68a,0x6ca,0x70a,0x74a,0x78a,6,6,
6,6,8,8,0x13,4,4,4,4,4,0x1b,0x1b,0x7ca,0xa4a,0xcca,4,
5,0x17,0x1b,0x1b,0xc,0x17,0x17,0x17,0x1b,4,5,0x54a,0x14,0x15,0x14,0x15,
0x14,0x15,0x14,0x15,0x14,0x15,0x1b,0x1b,0x14,0x15,0x14,0x15,0x14,0x15,0x14,0x15,
0x13,0x14,0x15,0x15,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,0,0,6,6,0x1a,
0x1a,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x17,
4,4,4,5,0,0,0,0,0,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,0,0x1b,0x1b,0x58b,0x5cb,0x60b,0x64b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0x58b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,0x7cb,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x7cb,0xa4b,0xccb,0xf4b,0x11cb,0x144b,0x16cb,0x194b,0x1b,0xa8b,0xacb,0xb0b,
0xb4b,0xb8b,0xbcb,0xc0b,0xc4b,0xc8b,0xccb,0xd0b,0xd4b,0xd8b,0xdcb,0xe0b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0xe4b,0xe8b,0xecb,
0xf0b,0xf4b,0xf8b,0xfcb,0x100b,0x104b,0x108b,0x10cb,0x110b,0x114b,0x118b,0x11cb,5,5,5,5,
5,0x685,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x5c5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x685,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,0x705,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,0x585,5,5,0x705,5,5,5,0x7885,
5,0x605,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,0x785,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0x5c5,5,5,5,5,5,5,5,0x685,5,0x645,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,0x7985,0x7c5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,0x7845,5,5,5,5,
5,5,5,5,0x605,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,0x685,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0x1e45,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0x7985,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x7a85,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0x5c5,5,0x745,5,0x6c5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,0x7c5,5,0x7845,0xa45,0xcc5,5,5,5,5,5,5,0xf45,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,0x605,0x605,0x605,0x605,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,0x645,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,0x585,5,5,5,5,5,5,5,0x585,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0x585,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0x785,0xa45,5,5,5,5,
5,5,5,5,5,5,5,5,0x585,0x5c5,0x605,5,0x5c5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x7c5,5,
5,5,5,5,5,5,5,5,5,5,5,5,0x745,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,0x705,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x785,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x1e45,5,
5,5,5,5,5,5,0x645,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0x7885,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,0x5c5,5,5,5,5,0x5c5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,0x5c5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,0x7845,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x6c5,5,
5,5,5,5,0x1e45,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0x6c5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0x545,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,4,5,5,5,5,5,5,5,5,5,5,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,4,0x17,0x17,0x17,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,5,5,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,2,1,2,1,2,4,4,6,6,
1,2,1,2,1,2,1,2,1,2,1,2,1,2,5,6,
7,7,7,0x17,6,6,6,6,6,6,6,6,6,6,0x17,4,
5,5,5,5,5,5,0x58a,0x5ca,0x60a,0x64a,0x68a,0x6ca,0x70a,0x74a,0x78a,0x54a,
6,6,0x17,0x17,0x17,0x17,0x17,0x17,0,0,0,0,0,0,0,0,
0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,
0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,4,4,4,4,4,4,4,4,4,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1,2,5,4,4,2,5,5,5,5,5,
0x1a,0x1a,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
2,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
1,2,1,2,4,2,2,2,2,2,2,2,2,1,2,1,
2,1,1,2,1,2,1,2,1,2,1,2,4,0x1a,0x1a,1,
2,1,2,5,1,2,1,2,2,2,1,2,1,2,1,2,
1,2,1,2,1,2,1,1,1,1,1,2,1,1,1,1,
1,2,1,2,1,2,1,2,1,2,1,2,0,0,1,2,
1,1,1,1,2,1,2,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,5,5,6,5,
5,5,6,5,5,5,5,6,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,8,
8,6,6,8,0x1b,0x1b,0x1b,0x1b,6,0,0,0,0x34cb,0x344b,0x3ccb,0x37cb,
0x35cb,0x3fcb,0x1b,0x1b,0x19,0x1b,0,0,0,0,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0x17,0x17,0x17,0x17,0,0,0,0,0,0,0,0,8,8,8,8,
6,6,0,0,0,0,0,0,0,0,0x17,0x17,0x49,0x89,0xc9,0x109,
0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,8,8,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,8,8,8,8,
8,8,8,8,8,8,8,8,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,5,5,5,5,5,5,
0x17,0x17,0x17,5,0x17,5,5,6,5,5,5,5,5,5,6,6,
6,6,6,6,6,6,0x17,0x17,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,
6,6,6,6,6,6,8,8,0,0,0,0,0,0,0,0,
0,0,0,0x17,8,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,
0x17,0x17,0,4,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,
0,0,0x17,0x17,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,6,8,8,6,6,6,6,8,8,
6,6,8,8,5,5,5,5,5,6,4,5,5,5,5,5,
5,5,5,5,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,5,5,
5,5,5,0,5,5,5,5,5,5,5,5,5,6,6,6,
6,6,6,8,8,6,6,8,8,6,6,0,0,0,0,0,
0,0,0,0,5,5,5,6,5,5,5,5,5,5,5,5,
6,8,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,
0x17,0x17,0x17,0x17,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,4,5,5,5,5,5,5,0x1b,0x1b,0x1b,5,8,
6,8,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,6,5,6,6,6,5,5,6,6,5,5,5,
5,5,6,6,5,6,5,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5,
5,4,0x17,0x17,5,5,5,5,5,5,5,5,5,5,5,8,
6,6,8,8,0x17,0x17,5,4,4,8,6,0,0,0,0,0,
0,0,0,0,0,5,5,5,5,5,5,0,0,5,5,5,
5,5,5,0,0,5,5,5,5,5,5,0,0,0,0,0,
0,0,0,0,5,5,5,5,5,5,5,0,5,5,5,5,
5,5,5,0,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,0x1a,
4,4,4,4,2,2,2,2,2,2,2,2,2,4,0x1a,0x1a,
0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,5,5,5,8,8,6,8,8,6,8,8,0x17,
8,6,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,
0,0,0,0,5,5,5,5,0,0,0,0,0,0,0,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0,0,0,0,5,5,5,5,5,5,5,0,0,0,0,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,
0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,0x12,
0x12,0x12,0x12,0x12,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,
0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,
0x11,0x11,0x11,0x11,5,5,5,5,5,5,5,5,5,5,5,0x605,
5,5,5,5,5,5,5,0x7c5,5,5,5,5,0x5c5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,0x6c5,5,0x6c5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,0x7c5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,0x18,5,5,5,5,5,5,5,5,5,5,
5,5,5,0,5,5,5,5,5,0,5,0,5,5,0,5,
5,0,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,2,2,2,2,
2,2,2,0,0,0,0,0,0,0,0,0,0,0,0,2,
2,2,2,2,0,0,0,0,0,5,6,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x1a,0x1a,
0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0x1a,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0x15,0x14,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0,0,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,0x19,0x1b,0,0,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x14,
0x15,0x17,0,0,0,0,0,0,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,0x17,0x13,0x13,0x16,0x16,0x14,0x15,0x14,
0x15,0x14,0x15,0x14,0x15,0x14,0x15,0x14,0x15,0x17,0x17,0x14,0x15,0x17,0x17,0x17,
0x17,0x16,0x16,0x16,0x17,0x17,0x17,0,0x17,0x17,0x17,0x17,0x13,0x14,0x15,0x14,
0x15,0x14,0x15,0x17,0x17,0x17,0x18,0x13,0x18,0x18,0x18,0,0x17,0x19,0x17,0x17,
0,0,0,0,5,5,5,5,5,0,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0,0,0x10,0,0,5,5,
5,5,5,5,0,0,5,5,5,5,5,5,0,0,5,5,
5,5,5,5,0,0,5,5,5,0,0,0,0x19,0x19,0x18,0x1a,
0x1b,0x19,0x19,0,0x1b,0x18,0x18,0x18,0x18,0x1b,0x1b,0,0,0,0,0,
0,0,0,0,0,0x10,0x10,0x10,0x1b,0x1b,0,0,0,0x17,0x17,0x17,
0x19,0x17,0x17,0x17,0x14,0x15,0x17,0x18,0x17,0x13,0x17,0x17,0x49,0x89,0xc9,0x109,
0x149,0x189,0x1c9,0x209,0x249,0x289,0x17,0x17,0x18,0x18,0x18,0x17,0x1a,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,0x14,0x18,0x15,0x18,0x14,0x15,0x17,0x14,0x15,
0x17,0x17,5,5,5,5,5,5,5,5,5,5,4,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,4,4,
5,5,5,5,5,5,5,5,5,5,5,5,0,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,0,5,5,0,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,0,0,0,0,0,0xb00b,0xb80b,0x784b,0x804b,
0x884b,0x904b,0x984b,0xa04b,0xa84b,0xb04b,0xb84b,0x788b,0x808b,0x888b,0x908b,0x988b,0xa08b,0xa88b,0xb08b,0xb88b,
0,0,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x17,0x17,0x17,0,
0,0,0,0x58b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,0x7cb,0xa4b,0xccb,0xf4b,
0x11cb,0x144b,0x16cb,0x194b,0x1bcb,0x1e4b,0x800b,0x880b,0x900b,0x980b,0xa00b,0xa80b,0x7ca,0x7ca,0x7ca,0x7ca,
0x7ca,0xcca,0x11ca,0x11ca,0x11ca,0x11ca,0x1e4a,0x880a,0x980a,0x980a,0x980a,0x980a,0x980a,0x784a,0x984a,0x68a,
0x11ca,0x344b,0x344b,0x388b,0x3ccb,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x54b,0x34cb,0x1b,0x1b,0x1b,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0x34ca,0x344a,0x58a,0x68a,0x11ca,0x980a,0x984a,0x988a,
0x68a,0x7ca,0x11ca,0x1e4a,0x980a,0x784a,0x984a,0x68a,0x7ca,0x11ca,0x1e4a,0x980a,0x784a,0x788a,0x988a,0x7ca,
0x58a,0x58a,0x58a,0x5ca,0x5ca,0x5ca,0x5ca,0x68a,0x1b,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,6,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,6,0x58b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,
0x74b,0x78b,0x7cb,0xa4b,0xccb,0xf4b,0x11cb,0x144b,0x16cb,0x194b,0x1bcb,0x1e4b,0x800b,0x880b,0x900b,0x980b,
0xa00b,0xa80b,0xb00b,0xb80b,0,0,0,0,0x58b,0x68b,0x7cb,0x11cb,0,0,0,0,
0,0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0x1bca,5,5,5,5,5,5,
5,5,0xb80a,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,6,
6,6,6,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,0,0x17,5,5,5,5,0,0,0,0,
5,5,5,5,5,5,5,5,0x17,0x58a,0x5ca,0x7ca,0xa4a,0x1e4a,0,0,
0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,
0x249,0x289,0,0,0,0,0,0,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,0,0,0,0,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,
2,2,2,2,2,2,2,2,5,5,5,5,5,5,5,5,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,0,0,0,0,0,0,0,0,
0,0,0,0x17,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0,5,5,0,0,0,
5,0,0,5,5,5,5,5,5,5,0,0,5,0,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,0,0x17,0x58b,0x5cb,0x60b,0x7cb,0xa4b,0x1e4b,0x784b,0x788b,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,0x1b,0x1b,0x58b,0x5cb,0x60b,0x64b,0x68b,0x7cb,0xa4b,
0,0,0,0,0,0,0,0x58b,0x5cb,0x60b,0x64b,0x64b,0x68b,0x7cb,0xa4b,0x1e4b,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,0,5,5,0,0,0,0,0,0x58b,0x68b,0x7cb,0xa4b,0x1e4b,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,0x58b,0x7cb,0xa4b,0x1e4b,0x5cb,0x60b,0,0,0,0x17,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0,0,0,0,0,0x17,
0xa04b,0xa84b,0xb04b,0xb84b,0x788b,0x808b,0x888b,0x908b,0x988b,0xa08b,0xa88b,0xb08b,0xb88b,0x78cb,0x80cb,0x88cb,
0x90cb,0x98cb,0xa0cb,0xa8cb,0xb0cb,0xb8cb,0x36cb,0x354b,0x34cb,0x348b,0x46cb,0x344b,0x4ecb,0x388b,0x3ccb,0x454b,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,0,0,0,0,0x5ecb,0x344b,5,5,
0x58b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,0x7cb,0xa4b,0xccb,0xf4b,0x11cb,0x144b,0x16cb,
0,0,0x1e4b,0x800b,0x880b,0x900b,0x980b,0xa00b,0xa80b,0xb00b,0xb80b,0x784b,0x804b,0x884b,0x904b,0x984b,
0x30b,0x34b,0x38b,0x3cb,0x7cb,0xa4b,0x1e4b,0x784b,0x344b,0,0,0,0,0,0,0,
0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0,0,0,0,0,0,0,
5,6,6,6,0,6,6,0,0,0,0,0,6,6,6,6,
5,5,5,5,0,5,5,5,0,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,0,0,6,6,6,0,0,0,0,6,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0x58b,0x11cb,0x17,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0x58b,0x7cb,0xa4b,5,5,5,5,
5,6,6,0,0,0,0,0x58b,0x68b,0x7cb,0xa4b,0x1e4b,0x17,0x17,0x17,0x17,
0x17,0x17,0x17,0,0,0,0,0,0,0,0,0,5,5,5,5,
5,5,5,5,0x1b,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0,0,
0,0x17,0x17,0x17,0x17,0x17,0x17,0x17,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0,0,
0x58b,0x5cb,0x60b,0x64b,0x7cb,0xa4b,0x1e4b,0x784b,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,0,0,0,0,0,
0x58b,0x5cb,0x60b,0x64b,0x7cb,0xa4b,0x1e4b,0x784b,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0,0,0,0,0,0,
0,0x17,0x17,0x17,0x17,0,0,0,0,0,0,0,0,0,0,0,
0,0x58b,0x5cb,0x60b,0x64b,0x7cb,0xa4b,0x1e4b,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,
0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,
0,0,0x58b,0x68b,0x7cb,0x11cb,0x1e4b,0x784b,5,5,5,5,6,6,6,6,
0,0,0,0,0,0,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,
0x249,0x289,0,0,0,0,0,0,0x30b,0x34b,0x38b,0x3cb,0x40b,0x44b,0x48b,0x4cb,
0x50b,0x7cb,0xa4b,0xccb,0xf4b,0x11cb,0x144b,0x16cb,0x194b,0x1bcb,0x1e4b,0x800b,0x880b,0x900b,0x980b,0xa00b,
0xa80b,0xb00b,0xb80b,0x344b,0x34cb,0x348b,0x388b,0,5,5,5,5,5,5,5,5,
5,5,0,6,6,0x13,0,0,5,5,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,0x58b,0x5cb,0x60b,0x64b,0x68b,0x7cb,0xa4b,0xccb,0x1e4b,0x344b,5,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,
6,6,6,6,6,0x58b,0x7cb,0xa4b,0x1e4b,0x17,0x17,0x17,0x17,0x17,0,0,
0,0,0,0,5,5,5,5,5,0x58b,0x5cb,0x60b,0x64b,0x7cb,0xa4b,0x1e4b,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x144b,0x16cb,0x194b,0x1bcb,0x1e4b,0x784b,0x49,0x89,0xc9,0x109,0x149,0x189,
0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,6,8,6,8,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,6,6,6,6,6,6,6,0x17,0x17,0x17,0x17,0x17,
0x17,0x17,0,0,0,0,0x30b,0x34b,0x38b,0x3cb,0x40b,0x44b,0x48b,0x4cb,0x50b,0x7cb,
0xa4b,0xccb,0xf4b,0x11cb,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,8,8,8,6,6,6,6,8,8,6,6,0x17,
0x17,0x10,0x17,0x17,0x17,0x17,0,0,0,0,0,0,0,0,0,0,
0,0x10,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,0,0,0,0,0,0,0,0x49,0x89,0xc9,0x109,
0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,5,5,5,5,
5,5,5,6,6,6,6,6,8,6,6,6,6,6,6,6,
6,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x17,0x17,0x17,0x17,
5,8,8,5,0,0,0,0,0,0,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,
0x17,0x17,5,0,0,0,0,0,0,0,0,0,8,5,5,5,
5,0x17,0x17,0x17,0x17,6,6,6,6,0x17,8,6,0x49,0x89,0xc9,0x109,
0x149,0x189,0x1c9,0x209,0x249,0x289,5,0x17,5,0x17,0x17,0x17,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,8,
8,8,6,6,6,6,6,6,6,6,6,8,0,0x58b,0x5cb,0x60b,
0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,0x7cb,0xa4b,0xccb,0xf4b,0x11cb,0x144b,0x16cb,0x194b,0x1bcb,0x1e4b,
0x784b,0,0,0,0,0,0,0,0,0,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,8,8,8,6,6,6,8,8,
6,8,6,6,0x17,0x17,0x17,0x17,0x17,0x17,6,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,0,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,0,
5,0,5,5,5,5,0,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,0,5,5,5,5,5,5,5,5,5,
5,0x17,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,8,8,8,6,6,6,6,6,
6,6,6,0,0,0,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,
0x249,0x289,0,0,0,0,0,0,5,5,8,8,0,0,6,6,
6,6,6,6,6,0,0,0,6,6,6,6,6,0,0,0,
0,0,0,0,0,0,0,0,6,6,8,8,0,5,5,5,
5,5,5,5,5,0,0,5,5,0,0,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0,5,5,5,5,5,5,
5,0,5,5,0,5,5,5,5,5,0,6,6,5,8,8,
6,8,8,8,8,0,0,8,8,0,0,8,8,8,0,0,
5,0,0,0,0,0,0,8,0,0,0,0,0,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,8,8,8,6,6,6,6,6,6,6,6,
8,8,6,6,6,8,6,5,5,5,5,0x17,0x17,0x17,0x17,0x17,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x17,0x17,0,0x17,6,5,
5,5,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
8,8,8,6,6,6,6,6,6,8,6,8,8,8,8,6,
6,8,6,6,5,5,0x17,5,0,0,0,0,0,0,0,0,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,8,
8,8,6,6,6,6,0,0,8,8,8,8,6,6,8,6,
6,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,
0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,5,5,5,5,6,6,0,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
8,8,8,6,6,6,6,6,6,6,6,8,8,6,8,6,
6,0x17,0x17,0x17,5,0,0,0,0,0,0,0,0,0,0,0,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,
0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
5,5,5,5,5,5,5,5,5,5,5,6,8,6,8,8,
6,6,6,6,6,6,8,6,5,0,0,0,0,0,0,0,
8,8,6,6,6,6,8,6,6,6,6,6,0,0,0,0,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x7cb,0xa4b,0x17,0x17,0x17,0x1b,
5,5,5,5,5,5,5,5,5,5,5,5,8,8,8,6,
6,6,6,6,6,6,6,6,8,6,6,0x17,0,0,0,0,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x7cb,0xa4b,0xccb,0xf4b,0x11cb,0x144b,
0x16cb,0x194b,0x1bcb,0,0,0,0,0,0,0,0,0,0,0,0,5,
8,5,8,6,0x17,0x17,0x17,0,0,0,0,0,0,0,0,0,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,
5,5,5,5,5,5,5,0,0,5,0,0,5,5,5,5,
5,5,5,5,0,5,5,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,8,8,8,8,8,8,0,8,
8,0,0,6,6,8,6,5,6,5,0x17,5,8,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
0,0,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,8,8,8,6,6,6,6,
0,0,6,6,8,8,8,8,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,
6,8,5,6,6,6,6,0x17,0x17,0x17,0x17,0x17,0x17,0x17,0x17,6,
0,0,0,0,0,0,0,0,5,6,6,6,6,6,6,8,
8,6,6,6,5,5,5,5,5,6,6,6,6,6,6,6,
6,6,6,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,0x17,0x17,0x17,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,6,6,6,6,6,6,6,6,6,6,6,6,6,8,
6,6,0x17,0x17,0x17,5,0x17,0x17,5,0x17,0x17,0x17,0x17,0x17,0,0,
0,0,0,0,0,0,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,
0x249,0x289,0x58b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,0x7cb,0xa4b,0xccb,0xf4b,0x11cb,
0x144b,0x16cb,0x194b,0x1bcb,0x1e4b,0,0,0,0x17,0x17,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,8,6,6,6,6,6,6,6,0,6,6,6,6,
6,6,8,6,6,6,6,6,6,6,6,6,0,8,6,6,
6,6,6,6,6,8,6,6,8,6,6,0,0,0,0,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0,0,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,5,6,0,0,0,0,0,0,0,0,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,0,0,0,0,
5,5,5,5,5,5,5,0,5,5,0,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,6,6,6,6,6,6,0,0,0,6,0,6,6,0,6,
5,5,5,5,5,5,5,5,5,5,8,8,8,8,8,0,
6,6,0,8,8,6,8,6,5,0,0,0,0,0,0,0,
5,5,5,5,5,5,0,5,5,0,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,6,6,8,8,0x17,0x17,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,5,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x19,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,
0,0,0,0,0,0,0,0,0,0,0,0x17,0xcd0b,0xcc0b,0xcb0b,0xd00b,
0xca0b,0xcf0b,0xcb4b,0xd04b,0xc90b,0x37cb,0x37cb,0x364b,0x35cb,0xc94b,0x3fcb,0x350b,0x34cb,0x344b,0x344b,0x3ccb,
0xcd0b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x19,0x19,0x19,0x34ca,0x354a,0x34ca,0x34ca,
0x344a,0x348a,0x388a,0xf4a,0x11ca,0x64a,0x68a,0x6ca,0x70a,0x74a,0x78a,0,0x17,0x17,0x17,0x17,
0x17,0,0,0,0,0,0,0,0,0,0,0,0x5ca,0x60a,0x64a,0x68a,
0x6ca,0x70a,0x74a,0x78a,0x60a,0x64a,0x68a,0x6ca,0x70a,0x74a,0x78a,0x64a,0x68a,0x6ca,0x70a,0x74a,
0x78a,0x58a,0x5ca,0x60a,0x64a,0x68a,0x6ca,0x70a,0x74a,0x78a,0x58a,0x5ca,0x60a,0x64a,0x68a,0x5ca,
0x60a,0x60a,0x64a,0x68a,0x6ca,0x70a,0x74a,0x78a,0x58a,0x5ca,0x60a,0x60a,0x64a,0x68a,0xc08a,0xc18a,
0x58a,0x5ca,0x60a,0x60a,0x64a,0x68a,0x60a,0x60a,0x64a,0x64a,0x64a,0x64a,0x6ca,0x70a,0x70a,0x70a,
0x74a,0x74a,0x78a,0x78a,0x78a,0x78a,0x5ca,0x60a,0x64a,0x68a,0x6ca,0x58a,0x5ca,0x60a,0x64a,0x64a,
0x68a,0x68a,0x5ca,0x60a,0x58a,0x5ca,0x348a,0x388a,0x454a,0x348a,0x388a,0x35ca,5,5,5,5,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,0,0x10,0x10,0x10,0x10,
0x10,0x10,0x10,0x10,0x10,0,0,0,0,0,0,0,5,5,5,5,
5,5,5,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x49,0x89,0xc9,0x109,
0x149,0x189,0x1c9,0x209,0x249,0x289,0,0,0,0,0x17,0x17,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0,0,6,6,6,6,
6,0x17,0,0,0,0,0,0,0,0,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,
6,6,6,0x17,0x17,0x17,0x17,0x17,0x1b,0x1b,0x1b,0x1b,4,4,4,4,
0x17,0x1b,0,0,0,0,0,0,0,0,0,0,0x49,0x89,0xc9,0x109,
0x149,0x189,0x1c9,0x209,0x249,0x289,0,0x7cb,0x1e4b,0x788b,0x790b,0x798b,0x7a0b,0x7a8b,0,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0,0,0,0,0,5,5,5,0x54b,0x58b,0x5cb,0x60b,
0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,0x7cb,0x80b,0x84b,0x88b,0x8cb,0x90b,0x94b,0x98b,0x9cb,0xa0b,
0x58b,0x5cb,0x60b,0x17,0x17,0x17,0x17,0,0,0,0,0,5,5,5,5,
5,5,5,5,5,5,5,0,0,0,0,6,5,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
0,0,0,0,0,0,0,6,6,6,6,4,4,4,4,4,
4,4,4,4,4,4,4,4,4,4,0x17,4,6,0,0,0,
0,0,0,0,0,0,0,0,8,8,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
5,5,5,0,0,0,0,0,0,0,0,0,0,0,0,0,
5,5,5,5,0,0,0,0,0,0,0,0,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,0,
0,0,0,0,5,5,5,5,5,5,5,5,5,5,5,5,
5,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,0,0,0x1b,6,6,0x17,0x10,0x10,0x10,0x10,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,
0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,8,8,6,6,6,0x1b,0x1b,
0x1b,8,8,8,8,8,8,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,6,
6,6,6,6,6,6,6,0x1b,0x1b,6,6,6,6,6,6,6,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,6,6,6,6,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x1b,0x1b,6,6,6,0x1b,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0x54b,0x58b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,0x7cb,0x80b,
0x84b,0x88b,0x8cb,0x90b,0x94b,0x98b,0x9cb,0xa0b,0,0,0,0,0,0,0,0,
0,0,0,0,0x58b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,0x7cb,0xa4b,0xccb,
0xf4b,0x11cb,0x144b,0x16cb,0x194b,0x1bcb,0x58b,0x5cb,0x60b,0x64b,0x68b,0x58b,0x68b,0,0,0,
0,0,0,0,0x249,0x289,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,
0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x49,0x89,0xc9,0x109,0x149,0x189,
0x1c9,0x209,0x249,0x289,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,2,2,2,2,2,2,2,0,2,2,2,2,2,2,
2,2,2,2,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,1,0,1,1,0,0,1,0,
0,1,1,0,0,1,1,1,1,0,1,1,1,1,1,1,
1,1,2,2,2,2,0,2,0,2,2,2,2,2,2,2,
0,2,2,2,2,2,2,2,2,2,2,2,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,2,2,2,2,
1,1,0,1,1,1,1,0,0,1,1,1,1,1,1,1,
1,0,1,1,1,1,1,1,1,0,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,1,1,0,1,1,1,1,0,1,1,1,1,
1,0,1,0,0,0,1,1,1,1,1,1,1,0,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,2,2,2,2,
2,2,0,0,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,0x18,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,0x18,2,2,2,2,2,2,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,0x18,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,0x18,2,2,
2,2,2,2,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,2,2,2,0x18,2,2,2,2,2,2,1,2,
0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,0x249,0x289,0x49,0x89,0xc9,0x109,
0x149,0x189,0x1c9,0x209,0,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,0x1b,0x1b,0x1b,0x1b,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,6,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
6,0x1b,0x1b,0x17,0x17,0x17,0x17,0x17,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,6,6,6,6,6,6,6,6,0,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
6,0,0,6,6,6,6,6,6,6,0,6,6,0,6,6,
6,6,6,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,
0x249,0x289,0,0,0,0,5,0x1b,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,5,0,0,0,6,6,6,6,6,6,6,4,
4,4,4,4,4,4,0,0,5,5,5,5,5,5,5,5,
5,5,5,5,6,6,6,6,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,
0x249,0x289,0,0,0,0,0,0x19,5,5,5,5,5,0,0,0x58b,
0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,6,6,6,6,6,6,6,0,
0,0,0,0,0,0,0,0,2,2,2,2,6,6,6,6,
6,6,6,4,0,0,0,0,0x49,0x89,0xc9,0x109,0x149,0x189,0x1c9,0x209,
0x249,0x289,0,0,0,0,0x17,0x17,1,1,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0x58b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,
0x74b,0x78b,0x7cb,0xa4b,0xccb,0xf4b,0x11cb,0x144b,0x78cb,0x794b,0x814b,0x58b,0x5cb,0x60b,0x64b,0x68b,
0x6cb,0x70b,0x74b,0x78b,0x1b,0x34cb,0x344b,0x3ccb,0x19,0x58b,0x5cb,0x788b,0x78cb,0,0,0,
0,0,0,0,0,0,0,0,0x16cb,0x194b,0x1bcb,0x1e4b,0x800b,0x880b,0x900b,0x980b,
0xa00b,0xa80b,0xb00b,0xb80b,0x784b,0x804b,0x884b,0x904b,0x984b,0xa04b,0xa84b,0xb04b,0xb84b,0x788b,0x808b,0x888b,
0x908b,0x988b,0xa08b,0xa88b,0xb08b,0xb88b,0x78cb,0x80cb,0x984b,0xa04b,0xa84b,0xb04b,0xb84b,0x788b,0x808b,0x888b,
0x908b,0x988b,0xa08b,0xa88b,0xb08b,0xb88b,0x1b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,0x74b,0x78b,0x7cb,
0x900b,0xa00b,0x804b,0x788b,0x344b,0x354b,0,0,0,0x58b,0x5cb,0x60b,0x64b,0x68b,0x6cb,0x70b,
0x74b,0x78b,0x7cb,0xa4b,0xccb,0xf4b,0x11cb,0x144b,0x16cb,0x194b,0x1bcb,0x1e4b,0x800b,0x880b,0x900b,0x980b,
0xa00b,0xa80b,0xb00b,0xb80b,0x784b,0x804b,0x884b,0x904b,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x18,0x18,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,5,5,5,5,0,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,0,5,5,0,5,0,0,5,
0,5,5,5,5,5,5,5,5,5,5,0,5,5,5,5,
0,5,0,5,0,0,0,0,0,0,5,0,0,0,0,5,
0,5,0,5,0,5,5,5,0,5,5,0,5,0,0,5,
0,5,0,5,0,5,0,5,0,5,5,0,5,0,0,5,
5,5,5,0,5,5,5,5,5,5,5,0,5,5,5,5,
0,5,5,5,5,0,5,0,5,5,5,5,5,5,5,5,
5,5,0,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0,0,0,0,0,5,5,5,0,5,5,5,
5,5,0,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0,0,0,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0,0,0,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x2cb,0x2cb,0x30b,0x34b,
0x38b,0x3cb,0x40b,0x44b,0x48b,0x4cb,0x50b,0x54b,0x54b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0,
0,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0x1b,0x1b,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1a,0x1a,0x1a,0x1a,0x1a,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,
0x1b,0x1b,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,
0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0x1b,0x1b,0x1b,0,0,0,0,0,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0,0,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,
0,0,0,0,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0,0,0,0,0,0,0,0,0,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0,
0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,0x1b,5,0x705,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,0x645,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,0x645,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,0x685,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,0xcc5,5,5,5,5,
5,5,5,5,0xf45,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,0xf45,5,5,5,5,5,5,5,5,5,5,5,
5,5,0x6c5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,0x605,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,0x605,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,0x605,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,0x605,5,5,5,5,
5,5,5,5,5,5,5,5,5,0x645,5,5,5,5,5,5,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
5,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0x785,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,
0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,
0,0x10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,
0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0,0,
0,0,0,0
};

static const UTrie2 propsTrie={
    propsTrie_index,
    propsTrie_index+4532,
    NULL,
    4532,
    17744,
    0xa40,
    0x1234,
    0x0,
    0x0,
    0x110000,
    0x5700,
    NULL, 0, FALSE, FALSE, 0, NULL
};

static const uint16_t propsVectorsTrie_index[31228]={
0x4e8,0x4f0,0x4f8,0x500,0x518,0x520,0x528,0x530,0x538,0x540,0x548,0x550,0x558,0x560,0x568,0x570,
0x577,0x57f,0x587,0x58f,0x592,0x59a,0x5a2,0x5aa,0x5b2,0x5ba,0x5c2,0x5ca,0x5d2,0x5da,0x5e2,0x5ea,
0x5f2,0x5fa,0x601,0x609,0x611,0x619,0x621,0x629,0x631,0x639,0x63e,0x646,0x64d,0x655,0x65d,0x665,
0x66d,0x675,0x67d,0x685,0x68c,0x694,0x69c,0x6a4,0x6ac,0x6b4,0x6bc,0x6c4,0x6cc,0x6d4,0x6dc,0x6e4,
0x1a38,0xd5e,0xe35,0x6ec,0x508,0xe9c,0xea4,0x1bf2,0x1300,0x1310,0x12f8,0x1308,0x7c5,0x7cb,0x7d3,0x7db,
0x7e3,0x7e9,0x7f1,0x7f9,0x801,0x807,0x80f,0x817,0x81f,0x825,0x82d,0x835,0x83d,0x845,0x84d,0x854,
0x85c,0x862,0x86a,0x872,0x87a,0x880,0x888,0x890,0x898,0x1318,0x8a0,0x8a8,0x8b0,0x8b7,0x8bf,0x8c7,
0x8cf,0x8d3,0x8db,0x8e2,0x8ea,0x8f2,0x8fa,0x902,0x162c,0x1634,0x90a,0x912,0x91a,0x922,0x92a,0x931,
0x1692,0x1682,0x168a,0x1973,0x197b,0x1328,0x939,0x1320,0x1572,0x1572,0x1574,0x133c,0x133d,0x1330,0x1332,0x1334,
0x169a,0x169c,0x941,0x169c,0x949,0x94e,0x956,0x16a1,0x95c,0x169c,0x962,0x96a,0xc39,0x16a9,0x16a9,0x972,
0x16b9,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,0x16ba,
0x16ba,0x16ba,0x16ba,0x16b1,0x97a,0x16c2,0x16c2,0x982,0xb59,0xb61,0xb69,0xb71,0x16d2,0x16ca,0x98a,0x992,
0x99a,0x16dc,0x16e4,0x9a2,0x16da,0x9aa,0x1a40,0xd66,0xb79,0xb81,0xb89,0xb8e,0x18e1,0xc6c,0xc73,0x1849,
0xc09,0x1a48,0xd6e,0xd76,0xd7e,0xd86,0xf47,0xf48,0x1939,0x193e,0xca8,0xcb0,0x19af,0x19b7,0x1b11,0xe3d,
0x19bf,0xcf2,0xcfa,0x19c7,0x10f6,0x1196,0xf27,0xd8e,0x1869,0x1851,0x1861,0x1859,0x18f9,0x18f1,0x18b9,0xc19,
0x1345,0x1345,0x1345,0x1345,0x1348,0x1345,0x1345,0x1350,0x9b2,0x1358,0x9b6,0x9be,0x1358,0x9c6,0x9ce,0x9d6,
0x1368,0x1360,0x1370,0x9de,0x9e6,0x1378,0x9ee,0x9f6,0x1380,0x1388,0x1390,0x1398,0x9fe,0x13a0,0x13a7,0x13af,
0x13b7,0x13bf,0x13c7,0x13cf,0x13d7,0x13de,0x13e6,0x13ee,0x13f6,0x13fe,0x1401,0x1403,0x16ec,0x17dc,0x17e2,0x1929,
0x140b,0xa06,0xa0e,0x1525,0x152a,0x152d,0x1535,0x1413,0x153d,0x153d,0x1423,0x141b,0x142b,0x1433,0x143b,0x1443,
0x144b,0x1453,0x145b,0x1463,0x17ea,0x1841,0x1983,0x1ad9,0x1473,0x147a,0x1482,0x148a,0x146b,0x1492,0x17f2,0x17f9,
0x16f4,0x16f4,0x16f4,0x16f4,0x16f4,0x16f4,0x16f4,0x16f4,0x1801,0x1804,0x1801,0x1801,0x180c,0x1813,0x1815,0x181c,
0x1824,0x1828,0x1828,0x182b,0x1828,0x1828,0x1831,0x1828,0x1871,0x1931,0x198b,0xb96,0xb9c,0x1c36,0x1c3e,0x1d15,
0x18d1,0xc49,0xc4d,0x1946,0x18c1,0x18c1,0x18c1,0xc21,0x18c9,0xc41,0x1911,0xc98,0xc29,0xc31,0xc31,0x19cf,
0x1901,0x1993,0xc83,0xc88,0xa16,0x16fc,0x16fc,0xa1e,0x1704,0x1704,0x1704,0x1704,0x1704,0x1704,0xa26,0x6f0,
0x155a,0x157c,0xa2e,0x1584,0xa36,0x158c,0x1594,0x159c,0xa3e,0xa43,0x15a4,0x15ab,0xa48,0x170c,0x1921,0xc11,
0xa50,0x1606,0x160d,0x15b3,0x1615,0x161c,0x15bb,0x15bf,0x15d8,0x15d8,0x15da,0x15c7,0x15cf,0x15cf,0x15d0,0x1624,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,
0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1714,0x1717,0x1879,0x1879,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,
0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e2,0x15e9,0x1a30,0x12b5,
0x171f,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,
0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,0x1725,
0x1725,0x1725,0x1725,0x1725,0xa58,0x172d,0xa60,0x1a50,0x19db,0x19db,0x19db,0x19db,0x19db,0x19db,0x19db,0x19db,
0x19d7,0xd02,0x19eb,0x19e3,0x19ed,0x1a58,0x1a58,0xd96,0x18d9,0x194e,0x19a3,0x19a7,0x199b,0x1b09,0xcb8,0xcbb,
0x1909,0xc90,0x1956,0xcc3,0x19f5,0x19f8,0xd0a,0x1a60,0x1a08,0x1a00,0xd12,0xd9e,0x1a68,0x1a6c,0xda6,0xff0,
0x1a10,0xd1a,0xd22,0x1a74,0x1a84,0x1a7c,0xdae,0xef7,0xe45,0xe4d,0x1c85,0xfa8,0x1d32,0x1d32,0x1a8c,0xdb6,
0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,
0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,
0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,
0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,
0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,
0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,
0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,
0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,
0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,
0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,
0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,
0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,
0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,
0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,
0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,
0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,
0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,
0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,
0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,
0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,
0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,
0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0x167a,0x1674,0x1675,0x1676,0x1677,0x1678,0x1679,0xa68,0xdbe,0xdc1,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,
0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,0x164c,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,
0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x1545,0x15f1,0x15f1,0x15f1,0x15f1,0x15f1,0x15f1,0x15f1,0x15f1,
0x15f6,0x15fe,0x1839,0x12bd,0x1919,0x1919,0x12c1,0x12c8,0xa70,0xa78,0xa80,0x14b2,0x14b9,0x14c1,0xa88,0x14c9,
0x14fa,0x14fa,0x14a2,0x14aa,0x14d1,0x14f1,0x14f2,0x1502,0x14d9,0x149a,0xa90,0x14e1,0xa98,0x14e9,0xaa0,0xaa4,
0xca0,0x150a,0xaac,0xab4,0x1512,0x1518,0x151d,0xabc,0xacc,0x1562,0x156a,0x154d,0x1552,0xad4,0xadc,0xac4,
0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,
0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x163c,0x1644,0x1644,0x1644,0x1644,
0x1420,0x1420,0x1460,0x14a0,0x14e0,0x1520,0x1560,0x15a0,0x15dc,0x161c,0x1648,0x1688,0x16c8,0x1708,0x1748,0x1788,
0x17c8,0x1804,0x1844,0x1884,0x18c4,0x18f8,0x1934,0x1974,0x19b4,0x19f4,0x1a30,0x1a70,0x1ab0,0x1af0,0x1b30,0x1b70,
0xa80,0xac0,0xb00,0xb40,0xb80,0xa40,0xe75,0xa40,0xe97,0xa40,0xa40,0xa40,0xa40,0xbc0,0x12dd,0x12dd,
0xed7,0xc00,0xa40,0xa40,0xa40,0xa40,0xf17,0xc2d,0xa40,0xa40,0xc6d,0xcad,0xced,0xd2d,0xe35,0xda5,
0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,0x121d,
0x121d,0x121d,0x121d,0x121d,0xf57,0x125d,0x1092,0x10d2,0x129d,0x10dd,0x131d,0x131d,0x131d,0xf97,0xfb7,0xff7,
0x135d,0x135d,0x1037,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,
0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0xfb7,0x1052,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0xde5,0xdf5,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,
0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xa40,0xd65,
0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,
0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x119d,0x111d,
0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,
0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x11dd,0x115d,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0xba4,0xbab,0xbb3,0xbbb,0x1881,0x1881,0x1881,0xbc3,0xbcb,0xbce,0x18b1,0x18a9,0xc01,0xd2a,0xd2e,0xd32,
0x508,0x508,0x508,0x508,0xd3a,0x1a18,0xd42,0xf3f,0x1735,0xae4,0xaea,0x1000,0xbd6,0x18e9,0xc7b,0x508,
0x174a,0x173d,0x1742,0x1889,0xbde,0xbe6,0x1134,0x113a,0x1c6d,0xf5d,0x1c5d,0x6f8,0x508,0x508,0x508,0x508,
0x1c8d,0x1c8d,0x1c8d,0x1c8d,0x1c8d,0x1c8d,0x1c8d,0x1c8d,0x1c8d,0xfb0,0xfb8,0xfc0,0x508,0x508,0x508,0x508,
0xbee,0xbf1,0xdc9,0x1cd5,0xff8,0x700,0x508,0x1092,0xccb,0xd4a,0x508,0x508,0x1c02,0xeff,0xf07,0x1d1d,
0xc55,0xc5c,0xc64,0x1a94,0x1cb5,0x508,0x1c95,0xfd0,0x1a9c,0xdd1,0xdd9,0xde1,0x1020,0x708,0x508,0x508,
0x1aa4,0x1aa4,0x710,0x508,0x1d4a,0x10aa,0x1d42,0x10b2,0x1e0e,0x11ac,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0xde9,0x1e66,0x1291,0x508,0x508,0x1e2e,0x11d4,0x11db,0x718,0x508,0x71c,0x1248,0x11e3,
0x1b19,0x1b1b,0xe55,0xe5c,0x1aac,0x1ab4,0xdf1,0xf1f,0x1bfa,0xee7,0xeef,0xfc8,0x1c1a,0x1c1e,0x1c26,0x1040,
0xf93,0xf98,0x724,0x508,0x109a,0x10a2,0x1c7d,0xfa0,0xf75,0xf7b,0xf83,0xf8b,0x508,0x508,0x508,0x508,
0x1daa,0x1da2,0x1124,0x112c,0x1cfd,0x1cf5,0x1068,0x508,0x508,0x508,0x508,0x508,0x1ce5,0x1028,0x1030,0x1038,
0x1cad,0x1ca5,0xfe0,0x111c,0x1c2e,0xf2f,0x72c,0x508,0x1078,0x1080,0x508,0x508,0x508,0x508,0x508,0x508,
0x1e06,0x118e,0x734,0x508,0x508,0x1d0d,0x1d05,0x1070,0x1250,0x1256,0x125e,0x508,0x508,0x11eb,0x11ef,0x11f7,
0x1dde,0x1dd6,0x1176,0x1dce,0x1dc6,0x73c,0x1cdd,0x1018,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x10da,0x10df,0x10e7,0x10ee,0x110e,0x1114,0x508,0x508,0x115a,0x115e,0x1166,0x119e,0x11a4,0x744,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x11bc,0x508,0x508,0x508,0x508,0x508,0x748,0x1e4e,0x1238,
0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,
0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x195e,0x1963,0xcd3,0xcda,0xcda,0xcda,
0x196b,0x196b,0x196b,0xce2,0x1d3a,0x1d3a,0x1d3a,0x1d3a,0x1d3a,0x1d3a,0x750,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,0x1b23,
0x1b23,0xe64,0xfe8,0x758,0x508,0x508,0x75c,0xf37,0x1ccd,0x1cc5,0x1008,0x1010,0x764,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x1e26,0x1e1e,0x11cc,0x508,0x508,0x508,0x1c12,0x1c12,0xf0f,0x1c0a,0xf17,0x508,0x508,0x1106,
0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,
0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dba,0x1dbe,0x1e76,0x1e76,0x1e76,0x1e76,0x1e76,0x1e76,0x1e76,0x1e76,
0x1e76,0x1e76,0x1e76,0x1e76,0x1e76,0x1e76,0x1266,0x126c,0x1286,0x1289,0x1289,0x1289,0x76c,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x1c65,0x1c65,0x1c65,
0xf50,0xf55,0x774,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x1752,0x1752,0x1752,
0x1752,0x1752,0x1752,0x1752,0xaf2,0x1762,0xafa,0x1763,0x175a,0x176b,0x1771,0x1779,0xb02,0x18a1,0x18a1,0x77c,
0x508,0x508,0x508,0x508,0x11c4,0x1891,0x1891,0xbf9,0xcea,0x508,0x508,0x508,0x508,0x17aa,0x17b1,0xb0a,
0x17b4,0xb12,0xb1a,0xb22,0x17ae,0xb2a,0xb32,0xb3a,0x17b3,0x17bb,0x17aa,0x17b1,0x17ad,0x17b4,0x17bc,0x17ab,
0x17b2,0x17ae,0xb41,0x1781,0x1789,0x1790,0x1797,0x1784,0x178c,0x1793,0x179a,0xb49,0x17a2,0x1d62,0x1d62,0x1d62,
0x1d62,0x1d62,0x1d62,0x1d62,0x1d62,0x1d62,0x1d62,0x1d62,0x1d62,0x1d62,0x1d62,0x1d62,0x1d62,0x1d52,0x1d55,0x1d52,
0x1d5c,0x10ca,0x784,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x10fe,0x78c,0x508,
0x508,0x508,0x508,0x508,0x508,0x1e46,0x11ff,0x794,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x1e56,0x1240,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x1c9d,0x1c9d,0x1c9d,
0x1c9d,0x1c9d,0x1c9d,0xfd8,0x508,0x1d9a,0x1d92,0x10d2,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x798,0x1e16,0x11b4,0x508,0x508,0x1207,0x1208,0x7a0,0x508,0x508,0x508,0x508,0x508,0xeac,0xeb4,0xebc,
0xec4,0xecc,0xed4,0xedb,0xedf,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x7a4,0x1048,0x1ced,0x104e,0x1ced,0x1056,0x105b,0x1060,0x1060,0x1d72,0x1d82,0x1d8a,
0x10ba,0x1d7a,0x1e36,0x10c2,0x1dee,0x1e3e,0x1e3e,0x117e,0x1186,0x121f,0x1225,0x122a,0x1230,0x1e5e,0x1e5e,0x1e5e,
0x1e5e,0x1274,0x1e5e,0x127a,0x127e,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,
0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,0x7ac,
0x7ac,0x7ac,0x7ac,0x7ac,0x7ad,0xb51,0x17c4,0x17c4,0x17c4,0x7b5,0x7b5,0x7b5,0x7b5,0x1899,0x1899,0x1899,
0x1899,0x1899,0x1899,0x1899,0x7bd,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,
0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,
0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,
0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,
0x7b5,0x7b5,0x7b5,0x7b5,0x7b5,0x1a20,0xd52,0x1a28,0x1a28,0xd56,0xe6c,0xe74,0xe7c,0x1ae9,0x1ad1,0x1af1,
0x1af9,0x1ae1,0xe01,0xe05,0xe0c,0xe14,0xe1b,0xe23,0xe2b,0xe2d,0xe2d,0xe2d,0xe2d,0x1b5a,0x1b62,0x1b5a,
0x1b68,0x1b70,0x1b3b,0x1b78,0x1b80,0x1b5a,0x1b88,0x1b90,0x1b97,0x1b9f,0x1b43,0x1b5a,0x1ba4,0x1b4b,0x1b52,0x1bac,
0x1bb2,0x1c4e,0x1c55,0x1c46,0x1bba,0x1bc2,0x1bca,0x1bd2,0x1cbd,0x1bda,0x1be2,0xe84,0xe8c,0x1b2b,0x1b2b,0x1b2b,
0xe94,0x1c75,0x1c75,0xf65,0xf6d,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,
0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0x1abe,0x1abc,0x1ac6,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,
0x1ac9,0x1abc,0x1abc,0x1abc,0x1abc,0x1abc,0xdf9,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x1d25,0x1d25,0x1d25,0x1d25,0x1d25,0x1d25,0x1d25,0x1d25,0x1d25,
0x1d25,0x1d25,0x1d25,0x1d25,0x1d25,0x1d2a,0x1d25,0x1d25,0x1d25,0x1088,0x108a,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,
0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,
0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,
0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,
0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1142,0x1b33,0x1de6,0x1de6,0x1de6,0x1de6,0x1de6,0x1de6,0x1de6,0x114a,
0x1152,0x1210,0x1217,0x1dfe,0x1dfe,0x1dfe,0x1dfe,0x1dfe,0x1dfe,0x1dfe,0x1dfe,0x1dfe,0x1dfe,0x1dfe,0x116e,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x508,
0x508,0x508,0x508,0x508,0x508,0x508,0x508,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,
0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,
0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,
0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x12d0,0x1299,0x1b01,
0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,
0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,
0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x12a1,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x17d4,0x17d4,0x17d4,0x17d4,0x17d4,0x17d4,0x17d4,0x17d4,0x17d4,
0x17d4,0x17d4,0x17d4,0x17d4,0x17d4,0x17d4,0x17d4,0x12d8,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x12a5,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,
0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,
0x1e6e,0x12ad,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,0x1299,
0x1299,0x12a5,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,
0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,
0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,
0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x12e0,0x1bea,0x1bea,0x1bea,0x1bea,
0x1bea,0x1bea,0x12e8,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,
0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,
0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,
0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,
0x1d6a,0x1d6a,0x12f0,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1664,0x1664,0x1664,
0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,
0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,
0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,
0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1654,0x166c,0x166c,0x166c,
0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,
0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,
0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,
0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x165c,0x1664,0x1664,0x1664,
0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,
0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,
0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,
0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x1664,0x166c,0x166c,0x166c,
0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,
0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,
0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,
0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x166c,0x17cc,0x17cc,0x17cc,
0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,
0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,
0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,
0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x17cc,0x1b01,0x1b01,0x1b01,
0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,
0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,
0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,
0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1b01,0x1d6a,0x1d6a,0x1d6a,
0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,
0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,
0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,
0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1d6a,0x1db2,0x1db2,0x1db2,
0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,
0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,
0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,
0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1db2,0x1df6,0x1df6,0x1df6,
0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,
0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,
0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,
0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1df6,0x1e6e,0x1e6e,0x1e6e,
0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,
0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,
0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,
0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x1e6e,0x4e7,0x4e7,0x4e7,
0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2ca,0x2d3,0x2cd,0x2cd,0x2d0,0x2c7,0x2c7,
0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,0x2c7,
0x7fb,0x7f5,0x7da,0x7d1,0x7c8,0x7c5,0x7bc,0x7d7,0x7c2,0x7ce,0x7d1,0x7ec,0x7e3,0x7d4,0x7f8,0x7cb,
0x7b9,0x7b9,0x7b9,0x7b9,0x7b9,0x7b9,0x7b9,0x7b9,0x7b9,0x7b9,0x7e0,0x7dd,0x7e6,0x7e6,0x7e6,0x7f5,
0x7bc,0x807,0x807,0x807,0x807,0x807,0x807,0x801,0x801,0x801,0x801,0x801,0x801,0x801,0x801,0x801,
0x801,0x801,0x801,0x801,0x801,0x801,0x801,0x801,0x801,0x801,0x801,0x7c2,0x7c8,0x7ce,0x7f2,0x7b6,
0x7ef,0x804,0x804,0x804,0x804,0x804,0x804,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,
0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7fe,0x7c2,0x7e9,0x7bf,0x7e6,0x2c7,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2e5,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,
0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,0x2d6,
0x2d9,0x651,0x810,0x813,0x657,0x813,0x80d,0x64e,0x645,0x2df,0x663,0x2e2,0x816,0x63c,0x65a,0x80a,
0x654,0x660,0x642,0x642,0x648,0x2dc,0x64e,0x64b,0x645,0x642,0x663,0x2e2,0x63f,0x63f,0x63f,0x651,
0x2eb,0x2eb,0x2eb,0x2eb,0x2eb,0x2eb,0x66c,0x2eb,0x2eb,0x2eb,0x2eb,0x2eb,0x2eb,0x2eb,0x2eb,0x2eb,
0x66c,0x2eb,0x2eb,0x2eb,0x2eb,0x2eb,0x2eb,0x65d,0x66c,0x2eb,0x2eb,0x2eb,0x2eb,0x2eb,0x66c,0x666,
0x669,0x669,0x2e8,0x2e8,0x2e8,0x2e8,0x666,0x2e8,0x669,0x669,0x669,0x2e8,0x669,0x669,0x2e8,0x2e8,
0x666,0x2e8,0x669,0x669,0x2e8,0x2e8,0x2e8,0x65d,0x666,0x669,0x669,0x2e8,0x669,0x2e8,0x666,0x2e8,
0x2f7,0x672,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,
0x2f4,0x66f,0x2f7,0x672,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x672,0x2f7,0x2ee,0x2f7,0x2ee,
0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x678,0x66f,0x2f7,0x2ee,0x2f7,0x672,0x2f7,0x2ee,0x2f7,0x2ee,
0x2f7,0x66f,0x67b,0x675,0x2f7,0x2ee,0x2f7,0x2ee,0x66f,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x67b,
0x675,0x678,0x66f,0x2f7,0x672,0x2f7,0x2ee,0x2f7,0x672,0x67e,0x678,0x66f,0x2f7,0x672,0x2f7,0x2ee,
0x2f7,0x2ee,0x678,0x66f,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,
0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x678,0x66f,0x2f7,0x2ee,0x2f7,0x672,0x2f7,0x2ee,0x2f7,0x2ee,
0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2f7,0x2ee,0x2f7,0x2ee,0x2f7,0x2ee,0x2f1,
0x2fa,0x306,0x306,0x2fa,0x306,0x2fa,0x306,0x306,0x2fa,0x306,0x306,0x306,0x2fa,0x2fa,0x306,0x306,
0x306,0x306,0x2fa,0x306,0x306,0x2fa,0x306,0x306,0x306,0x2fa,0x2fa,0x2fa,0x306,0x306,0x2fa,0x306,
0x309,0x2fd,0x306,0x2fa,0x306,0x2fa,0x306,0x306,0x2fa,0x306,0x2fa,0x2fa,0x306,0x2fa,0x306,0x309,
0x2fd,0x306,0x306,0x306,0x2fa,0x306,0x2fa,0x306,0x306,0x2fa,0x2fa,0x303,0x306,0x2fa,0x2fa,0x2fa,
0x303,0x303,0x303,0x303,0x30c,0x30c,0x300,0x30c,0x30c,0x300,0x30c,0x30c,0x300,0x309,0x681,0x309,
0x681,0x309,0x681,0x309,0x681,0x309,0x681,0x309,0x681,0x309,0x681,0x309,0x681,0x2fa,0x309,0x2fd,
0x309,0x2fd,0x309,0x2fd,0x306,0x2fa,0x309,0x2fd,0x309,0x2fd,0x309,0x2fd,0x309,0x2fd,0x309,0x2fd,
0x2fd,0x30c,0x30c,0x300,0x309,0x2fd,0x9ea,0x9ea,0x9ed,0x9e7,0x309,0x2fd,0x309,0x2fd,0x309,0x2fd,
0x309,0x2fd,0x309,0x2fd,0x309,0x2fd,0x309,0x2fd,0x309,0x2fd,0x309,0x2fd,0x309,0x2fd,0x309,0x2fd,
0x309,0x2fd,0x309,0x2fd,0x9ed,0x9e7,0x9ed,0x9e7,0x9ea,0x9e4,0x9ed,0x9e7,0xbaf,0xcb7,0x9ea,0x9e4,
0x9ea,0x9e4,0x9ed,0x9e7,0x9ed,0x9e7,0x9ed,0x9e7,0x9ed,0x9e7,0x9ed,0x9e7,0x9ed,0x9e7,0x9ed,0x9e7,
0xcb7,0xcb7,0xcb7,0xdb6,0xdb6,0xdb6,0xdb9,0xdb9,0xdb6,0xdb9,0xdb9,0xdb6,0xdb6,0xdb9,0xefa,0xefd,
0xefd,0xefd,0xefd,0xefa,0xefd,0xefa,0xefd,0xefa,0xefd,0xefa,0xefd,0xefa,0x30f,0x684,0x30f,0x30f,
0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x684,0x30f,0x30f,
0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,
0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x312,0x30f,0x30f,0x30f,
0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,0x30f,
0x30f,0x9f0,0x9f0,0x9f0,0x9f0,0x9f0,0xcba,0xcba,0x327,0x327,0x327,0x327,0x327,0x327,0x327,0x327,
0x327,0x31e,0x31e,0x31e,0x31e,0x31e,0x31e,0x31e,0x31b,0x31b,0x318,0x318,0x68a,0x318,0x31e,0x68d,
0x321,0x68d,0x68d,0x68d,0x321,0x68d,0x31e,0x31e,0x690,0x324,0x318,0x318,0x318,0x318,0x318,0x318,
0x687,0x687,0x687,0x687,0x315,0x687,0x318,0xb25,0x327,0x327,0x327,0x327,0x327,0x318,0x318,0x318,
0x318,0x318,0x9f9,0x9f9,0x9f6,0x9f3,0x9f6,0xcbd,0xcbd,0xcbd,0xcbd,0xcbd,0xcbd,0xcbd,0xcbd,0xcbd,
0xcbd,0xcbd,0xcbd,0xcbd,0xcbd,0xcbd,0xcbd,0xcbd,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,
0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,
0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,
0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,
0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x693,0x696,0x696,0x94b,0x696,0x696,0x94e,0xb28,0xb28,
0xb28,0xb28,0xb28,0xb28,0xb28,0xb28,0xb28,0xc6c,0xd83,0xd83,0xd83,0xd83,0xd83,0xd83,0xd83,0xd83,
0xebe,0xebe,0xebe,0xebe,0xec1,0xd86,0xd86,0xd86,0x699,0x699,0xb2b,0xcb4,0xcb4,0xcb4,0xcb4,0xcb4,
0xcb4,0xcb4,0xcb4,0xcb4,0xcb4,0xcb4,0xcb4,0xcb4,0xfa8,0xfa5,0xfa8,0xfa5,0x333,0x33c,0xfa8,0xfa5,
9,9,0x342,0xf00,0xf00,0xf00,0x32a,0x14fd,9,9,9,9,0x33f,0x32d,0x351,0x330,
0x351,0x351,0x351,9,0x351,9,0x351,0x351,0x348,0x69f,0x69f,0x69f,0x69f,0x69f,0x69f,0x69f,
0x69f,0x69f,0x69f,0x69f,0x69f,0x69f,0x69f,0x69f,0x69f,0x69f,9,0x69f,0x69f,0x69f,0x69f,0x69f,
0x69f,0x69f,0x351,0x351,0x348,0x348,0x348,0x348,0x348,0x69c,0x69c,0x69c,0x69c,0x69c,0x69c,0x69c,
0x69c,0x69c,0x69c,0x69c,0x69c,0x69c,0x69c,0x69c,0x69c,0x69c,0x345,0x69c,0x69c,0x69c,0x69c,0x69c,
0x69c,0x69c,0x348,0x348,0x348,0x348,0x348,0xfa8,0x354,0x354,0x357,0x351,0x351,0x354,0x34b,0x9fc,
0xbb8,0xbb5,0x34e,0x9fc,0x34e,0x9fc,0x34e,0x9fc,0x34e,0x9fc,0x339,0x336,0x339,0x336,0x339,0x336,
0x339,0x336,0x339,0x336,0x339,0x336,0x339,0x336,0x354,0x354,0x34b,0x345,0xb67,0xb64,0xbb2,0xcc3,
0xcc0,0xcc6,0xcc3,0xcc0,0xdbc,0xdbf,0xdbf,0xdbf,0xa0b,0x6ab,0x363,0x366,0x363,0x363,0x363,0x366,
0x363,0x363,0x363,0x363,0x366,0xa0b,0x366,0x363,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,
0x6a8,0x6ab,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,
0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a8,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,
0x6a2,0x6a5,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,0x6a2,
0x6a2,0x6a2,0x6a2,0x6a2,0xa05,0x6a5,0x35d,0x360,0x35d,0x35d,0x35d,0x360,0x35d,0x35d,0x35d,0x35d,
0x360,0xa05,0x360,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,
0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x366,0x360,0x363,0x35d,0x363,0x35d,
0x363,0x35d,0x363,0x35d,0x363,0x35d,0x35a,0x957,0x95a,0x93c,0x93c,0x114f,0x9ff,0x9ff,0xbbe,0xbbb,
0xa08,0xa02,0xa08,0xa02,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,
0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,
0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,0x363,0x35d,
0x363,0x35d,0x363,0x35d,0x363,0x366,0x360,0x363,0x35d,0xbbe,0xbbb,0x363,0x35d,0xbbe,0xbbb,0x363,
0x35d,0xbbe,0xbbb,0xf03,0x366,0x360,0x366,0x360,0x363,0x35d,0x366,0x360,0x363,0x35d,0x366,0x360,
0x366,0x360,0x366,0x360,0x363,0x35d,0x366,0x360,0x366,0x360,0x366,0x360,0x363,0x35d,0x366,0x360,
0xa0b,0xa05,0x366,0x360,0x366,0x360,0x366,0x360,0x366,0x360,0xdc5,0xdc2,0x366,0x360,0xf06,0xf03,
0xf06,0xf03,0xf06,0xf03,0xc2d,0xc2a,0xc2d,0xc2a,0xc2d,0xc2a,0xc2d,0xc2a,0xc2d,0xc2a,0xc2d,0xc2a,
0xc2d,0xc2a,0xc2d,0xc2a,0xf33,0xf30,0xf33,0xf30,0x1023,0x1020,0x1023,0x1020,0x1023,0x1020,0x1023,0x1020,
0x1023,0x1020,0x1023,0x1020,0x1023,0x1020,0x1023,0x1020,0x1188,0x1185,0x1371,0x136e,0x1536,0x1533,0x1536,0x1533,
0x1536,0x1533,0x1536,0x1533,0xc,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,
0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0x378,0xc,
0xc,0x37b,0x369,0x369,0x369,0x36f,0x369,0x36c,0x1941,0x372,0x372,0x372,0x372,0x372,0x372,0x372,
0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,
0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x372,0x375,0x1941,0x37e,0xa0e,0xc,
0xc,0x1500,0x1500,0x141c,0xf,0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,
0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,0xdc8,0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,0x97e,
0x97e,0x97e,0x97e,0x97e,0x381,0x381,0x381,0x381,0x381,0x381,0x381,0x381,0x381,0x381,0xf09,0x381,
0x381,0x381,0x38d,0x381,0x384,0x381,0x381,0x390,0x981,0xdcb,0xdce,0xdcb,0xf,0xf,0xf,0xf,
0xf,0xf,0xf,0xf,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,
0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0x393,0xf,
0xf,0xf,0xf,0x1944,0x393,0x393,0x393,0x38a,0x387,0xf,0xf,0xf,0xf,0xf,0xf,0xf,
0xf,0xf,0xf,0xf,0xcdb,0xcdb,0xcdb,0xcdb,0x141f,0x1503,0xfb1,0xfb1,0xfb1,0xfae,0xfae,0xdd4,
0x8c7,0xcd5,0xcd2,0xcd2,0xcc9,0xcc9,0xcc9,0xcc9,0xcc9,0xcc9,0xfab,0xfab,0xfab,0xfab,0xfab,0x8c4,
0x14fa,0x12,0xdd7,0x8ca,0x1338,0x3ae,0x3b1,0x3b1,0x3b1,0x3b1,0x3b1,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,
0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0xfb4,
0xfb4,0xfb4,0xfb4,0xfb4,0x8cd,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x942,
0x942,0x942,0x942,0x942,0x942,0x942,0x942,0xb5e,0xb5e,0xb5e,0xcc9,0xccf,0xccc,0xdd1,0xdd1,0xdd1,
0xdd1,0xdd1,0xdd1,0x1335,0x960,0x960,0x960,0x960,0x960,0x960,0x960,0x960,0x960,0x960,0x3a8,0x3a5,
0x3a2,0x39f,0xbc1,0xbc1,0x93f,0x3ae,0x3ae,0x3ba,0x3ae,0x3b4,0x3b4,0x3b4,0x3b4,0x3ae,0x3ae,0x3ae,
0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,
0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,
0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,
0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0xa14,0xa14,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0xa14,
0x3b1,0x3ae,0x3b1,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0x3ae,0xa14,
0x3ae,0x3ae,0x3ae,0x3b1,0x95d,0x3ae,0x399,0x399,0x399,0x399,0x399,0x399,0x399,0x396,0x39f,0x39c,
0x39c,0x399,0x399,0x399,0x399,0x3b7,0x3b7,0x399,0x399,0x39f,0x39c,0x39c,0x39c,0x399,0xcd8,0xcd8,
0x3ab,0x3ab,0x3ab,0x3ab,0x3ab,0x3ab,0x3ab,0x3ab,0x3ab,0x3ab,0xa14,0xa14,0xa14,0xa11,0xa11,0xcd8,
0xa29,0xa29,0xa29,0xa23,0xa23,0xa23,0xa23,0xa23,0xa23,0xa23,0xa23,0xa20,0xa23,0xa20,0x15,0xa2c,
0xa26,0xa17,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,
0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xa26,0xcde,0xcde,0xcde,
0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,0xa1d,
0xa1a,0xa1a,0xa1a,0xa1a,0xa1a,0xa1a,0xa1a,0xa1a,0xa1a,0xa1a,0xa1a,0x15,0x15,0xcde,0xcde,0xcde,
0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,
0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0xe37,0x1035,0x1035,
0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,0x1035,
0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,
0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,
0xa32,0xa32,0xa32,0xa32,0xa32,0xa32,0xa2f,0xa2f,0xa2f,0xa2f,0xa2f,0xa2f,0xa2f,0xa2f,0xa2f,0xa2f,
0xa2f,0xbc4,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0xf4b,0xf4b,0xf4b,0xf4b,0xf4b,0xf4b,0xf4b,0xf4b,0xf4b,0xf4b,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,
0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,
0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf4e,0xf42,0xf42,0xf42,0xf42,0xf42,
0xf42,0xf42,0xf42,0xf42,0xf51,0xf51,0xf45,0xf45,0xf48,0xf57,0xf54,0x10b,0x10b,0x1968,0x196b,0x196b,
0x18f9,0x18f9,0x18f9,0x18f9,0x18f9,0x18f9,0x18f9,0x18f9,0x18f9,0x18f9,0x18f9,0x252,0x252,0x252,0x252,0x252,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xb37,0xb37,0xb3a,0xb3a,0xb37,0xb37,0xb37,0xb37,0xb37,0xb37,0xb37,0xb37,0x72,0x72,0x72,0x72,
0x15ba,0x15ba,0x15ba,0x15ba,0x1bc,0x1bc,0x1bc,0x1bc,0x1bc,0x1bc,0x1bc,0x1bc,0x1bc,0x1bc,0x1bc,0x15b7,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x1ef,0x1ef,0x1ef,0x1ef,0x1ef,0x1ef,0x1ef,0x1674,0x1674,0x1674,0x1674,0x1674,0x1674,0x1674,0x1674,0x1674,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x1fb,0x1fb,0x1fb,0x1fb,0x1fb,0x1fb,0x1fb,0x1fb,0x1fb,0x16ad,0x16ad,0x16ad,0x16ad,0x16ad,0x16ad,0x16ad,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x174,0x174,0x174,0x174,0x174,0x174,0x174,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x276,0x276,0x276,0x276,0x276,0x276,0x276,0x276,0x276,0x276,0x276,0x276,0x276,0x276,0x276,0x276,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,
0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,0x1d1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x14d6,0x14d6,0x14d6,0x14d6,0x14d6,0x14d6,0x14d6,0x14d6,0x14d6,0x14d6,0x1b6,0x1b6,0x1b6,0x1b6,0x1b6,0x1b6,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,0x25b,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x18e7,0x18ea,0x18ea,0x24f,0x24f,0x24f,0x24f,0x24f,0x24f,0x24f,0x24f,0x24f,0x24f,0x24f,0x24f,0x24f,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x19b6,0x19b6,0x19b6,0x19b6,0x19b6,0x19b6,0x19b6,0x19b6,0x19b6,0x19b6,0x261,0x261,0x261,0x261,0x261,0x261,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x1b1b,0x29d,0x29d,0x29d,0x29d,0x29d,0x29d,0x29d,0x29d,0x29d,0x29d,0x29d,0x29d,0x29d,0x29d,0x29d,
0x17a9,0x17a9,0x17a9,0x17a9,0x216,0x216,0x216,0x216,0x216,0x216,0x216,0x216,0x216,0x216,0x216,0x216,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x165f,0x165f,0x165f,0x165f,0x165f,0x165f,0x165f,0x165f,0x165f,0x165f,0x1e9,0x1e9,0x1e9,0x1e9,0x1665,0x1665,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,
0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x1713,0x1713,0x1713,0x1713,0x1fe,0x1fe,0x1fe,0x1fe,0x1fe,0x1fe,0x1fe,0x1fe,0x1fe,0x1fe,0x1fe,0x1fe,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0xe2e,0xe2e,0xe2b,0xe2b,0xe2b,0xe2e,0xd8,0xd8,0xd8,0xd8,0xd8,0xd8,0xd8,0xd8,0xd8,0xd8,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x225,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x1845,0x1845,0x231,0x1845,0x1845,0x231,0x1845,0x1845,0x1845,0x1845,0x1845,0x231,0x231,0x231,0x231,0x231,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x1a6d,0x1a6d,0x1a6d,0x1a6d,0x1a6d,0x1a6d,0x1a6d,0x1a6d,0x1a6d,0x1a6d,0x282,0x282,0x282,0x282,0x1a70,0x1a6a,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x267,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,
0x285,0x285,0x285,0x285,0x285,0x285,0x285,0x285,0x285,0x285,0x285,0x285,0x285,0x285,0x285,0x285,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0x97b,0x97b,
3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
3,3,0x97b,0x97b,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,
0xd8c,0xd8c,0xd8c,0xd8c,6,6,6,6,6,6,6,6,6,6,6,6,
6,6,6,6,0x1509,0x3d5,0x3e4,0x3e4,0x1b,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,
0x3ea,0x1b,0x1b,0x3ea,0x3ea,0x1b,0x1b,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,
0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x1b,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x3ea,0x1b,0x3ea,0x1b,
0x1b,0x1b,0x3ea,0x3ea,0x3ea,0x3ea,0x1b,0x1b,0x3d8,0xce4,0x3d5,0x3e4,0x3e4,0x3d5,0x3d5,0x3d5,
0x3d5,0x1b,0x1b,0x3e4,0x3e4,0x1b,0x1b,0x3e7,0x3e7,0x3db,0xddd,0x1b,0x1b,0x1b,0x1b,0x1b,
0x1b,0x1b,0x1b,0x3d5,0x1b,0x1b,0x1b,0x1b,0x3ed,0x3ed,0x1b,0x3ed,0x3ea,0x3ea,0x3d5,0x3d5,
0x1b,0x1b,0x966,0x966,0x966,0x966,0x966,0x966,0x966,0x966,0x966,0x966,0x3ea,0x3ea,0x3e1,0x3e1,
0x3de,0x3de,0x3de,0x3de,0x3de,0x3e1,0x3de,0x115e,0x18a2,0x189f,0x1947,0x1b,0x1e,0xce7,0x3f0,0xcea,
0x1e,0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x1e,0x1e,0x1e,0x1e,0x3fc,0x3fc,0x1e,0x1e,0x3fc,
0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x1e,0x3fc,0x3fc,
0x3fc,0x3fc,0x3fc,0x3fc,0x3fc,0x1e,0x3fc,0x3ff,0x1e,0x3fc,0x3ff,0x1e,0x3fc,0x3fc,0x1e,0x1e,
0x3f3,0x1e,0x3f9,0x3f9,0x3f9,0x3f0,0x3f0,0x1e,0x1e,0x1e,0x1e,0x3f0,0x3f0,0x1e,0x1e,0x3f0,
0x3f0,0x3f6,0x1e,0x1e,0x1e,0xfbd,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x3ff,0x3ff,0x3ff,
0x3fc,0x1e,0x3ff,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x969,0x969,0x969,0x969,0x969,0x969,
0x969,0x969,0x969,0x969,0x3f0,0x3f0,0x3fc,0x3fc,0x3fc,0xfbd,0x194a,0x1e,0x1e,0x1e,0x1e,0x1e,
0x1e,0x1e,0x1e,0x1e,0x21,0x402,0x402,0x40b,0x21,0x40e,0x40e,0x40e,0x40e,0x40e,0x40e,0x40e,
0xcf3,0x40e,0x21,0x40e,0x40e,0x40e,0x21,0x40e,0x40e,0x40e,0x40e,0x40e,0x40e,0x40e,0x40e,0x40e,
0x40e,0x40e,0x40e,0x40e,0x40e,0x21,0x40e,0x40e,0x40e,0x40e,0x40e,0x40e,0x40e,0x21,0x40e,0x40e,
0x21,0x40e,0x40e,0x40e,0x40e,0x40e,0x21,0x21,0x405,0x40e,0x40b,0x40b,0x40b,0x402,0x402,0x402,
0x402,0x402,0x21,0x402,0x402,0x40b,0x21,0x40b,0x40b,0x408,0x21,0x21,0x40e,0x21,0x21,0x21,
0x21,0x21,0x21,0x21,0x21,0x21,0x21,0x21,0x21,0x21,0x21,0x21,0x40e,0xcf3,0xced,0xced,
0x21,0x21,0x96c,0x96c,0x96c,0x96c,0x96c,0x96c,0x96c,0x96c,0x96c,0x96c,0x1422,0xcf0,0x21,0x21,
0x21,0x21,0x21,0x21,0x21,0x1725,0x18a5,0x18a5,0x18a5,0x18a8,0x18a8,0x18a8,0x24,0x411,0x420,0x420,
0x24,0x426,0x426,0x426,0x426,0x426,0x426,0x426,0x426,0x24,0x24,0x426,0x426,0x24,0x24,0x426,
0x426,0x426,0x426,0x426,0x426,0x426,0x426,0x426,0x426,0x426,0x426,0x426,0x426,0x24,0x426,0x426,
0x426,0x426,0x426,0x426,0x426,0x24,0x426,0x426,0x24,0xcf6,0x426,0x426,0x426,0x426,0x24,0x24,
0x414,0x426,0x411,0x411,0x420,0x411,0x411,0x411,0xfc0,0x24,0x24,0x420,0x423,0x24,0x24,0x423,
0x423,0x417,0x24,0x24,0x24,0x24,0x24,0x24,0x24,0x1ab5,0x411,0x411,0x24,0x24,0x24,0x24,
0x429,0x429,0x24,0x426,0x426,0x426,0xfc0,0xfc0,0x24,0x24,0x41d,0x41d,0x41d,0x41d,0x41d,0x41d,
0x41d,0x41d,0x41d,0x41d,0x41a,0xcf6,0x1344,0x1344,0x1344,0x1344,0x1344,0x1344,0x24,0x24,0x24,0x24,
0x24,0x24,0x24,0x24,0x27,0x27,0x42c,0x438,0x27,0x438,0x438,0x438,0x438,0x438,0x438,0x27,
0x27,0x27,0x438,0x438,0x438,0x27,0x438,0x438,0x43b,0x438,0x27,0x27,0x27,0x438,0x438,0x27,
0x438,0x27,0x438,0x438,0x27,0x27,0x27,0x438,0x438,0x27,0x27,0x27,0x438,0x438,0x438,0x27,
0x27,0x27,0x438,0x438,0x438,0x438,0x438,0x438,0x438,0x438,0xde0,0x438,0x438,0x438,0x27,0x27,
0x27,0x27,0x42c,0x432,0x42c,0x432,0x432,0x27,0x27,0x27,0x432,0x432,0x432,0x27,0x435,0x435,
0x435,0x42f,0x27,0x27,0xfc3,0x27,0x27,0x27,0x27,0x27,0x27,0x42c,0x27,0x27,0x27,0x27,
0x27,0x27,0x27,0x27,0x27,0x27,0xef7,0x972,0x972,0x972,0x972,0x972,0x972,0x972,0x972,0x972,
0x96f,0x96f,0x96f,0xdb0,0xcf9,0xcf9,0xcf9,0xcf9,0xcf9,0xcfc,0xcf9,0x27,0x27,0x27,0x27,0x27,
0x150c,0x44a,0x44a,0x44a,0x194d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x2a,0x44d,0x44d,
0x44d,0x2a,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,
0x44d,0x2a,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x44d,0x150f,0x44d,0x44d,0x44d,
0x44d,0x44d,0x2a,0x2a,0x2a,0xfcc,0x43e,0x43e,0x43e,0x44a,0x44a,0x44a,0x44a,0x2a,0x43e,0x43e,
0x441,0x2a,0x43e,0x43e,0x43e,0x444,0x2a,0x2a,0x2a,0x2a,0x2a,0x2a,0x2a,0x43e,0x43e,0x2a,
0xfcc,0xfcc,0x1728,0x2a,0x2a,0x2a,0x2a,0x2a,0x44d,0x44d,0xfc6,0xfc6,0x2a,0x2a,0x447,0x447,
0x447,0x447,0x447,0x447,0x447,0x447,0x447,0x447,0x2a,0x2a,0x2a,0x2a,0x2a,0x2a,0x2a,0x1a19,
0xfc9,0xfc9,0xfc9,0xfc9,0xfc9,0xfc9,0xfc9,0xfc9,0x17e5,0x1512,0x456,0x456,0x1950,0x45c,0x45c,0x45c,
0x45c,0x45c,0x45c,0x45c,0x45c,0x2d,0x45c,0x45c,0x45c,0x2d,0x45c,0x45c,0x45c,0x45c,0x45c,0x45c,
0x45c,0x45c,0x45c,0x45c,0x45c,0x45c,0x45c,0x45c,0x45c,0x2d,0x45c,0x45c,0x45c,0x45c,0x45c,0x45c,
0x45c,0x45c,0x45c,0x45c,0x2d,0x45c,0x45c,0x45c,0x45c,0x45c,0x2d,0x2d,0xcff,0xd02,0x456,0x450,
0x459,0x456,0x450,0x456,0x456,0x2d,0x450,0x459,0x459,0x2d,0x459,0x459,0x450,0x453,0x2d,0x2d,
0x2d,0x2d,0x2d,0x2d,0x2d,0x450,0x450,0x2d,0x2d,0x2d,0x2d,0x2d,0x2d,0x2d,0x45c,0x2d,
0x45c,0x45c,0xf0f,0xf0f,0x2d,0x2d,0x975,0x975,0x975,0x975,0x975,0x975,0x975,0x975,0x975,0x975,
0x2d,0xf12,0xf12,0x2d,0x2d,0x2d,0x2d,0x2d,0x2d,0x2d,0x2d,0x2d,0x2d,0x2d,0x2d,0x2d,
0x18ab,0x1515,0x468,0x468,0x1ab8,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x30,0x46e,0x46e,
0x46e,0x30,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,
0x468,0x45f,0x45f,0x45f,0xfcf,0x30,0x468,0x468,0x468,0x30,0x46b,0x46b,0x46b,0x462,0x134a,0x17e8,
0x30,0x30,0x30,0x30,0x17eb,0x17eb,0x17eb,0x45f,0x17e8,0x17e8,0x17e8,0x17e8,0x17e8,0x17e8,0x17e8,0x172b,
0x46e,0x46e,0xfcf,0xfcf,0x30,0x30,0x465,0x465,0x465,0x465,0x465,0x465,0x465,0x465,0x465,0x465,
0xfd2,0xfd2,0xfd2,0xfd2,0xfd2,0xfd2,0x17e8,0x17e8,0x17e8,0xfd5,0xfd8,0xfd8,0xfd8,0xfd8,0xfd8,0xfd8,
0x33,0x1abb,0xa3e,0xa3e,0x33,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,
0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0x33,0x33,0x33,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,
0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0x33,0xa44,
0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0xa44,0x33,0xa44,0x33,0x33,0xa44,0xa44,0xa44,0xa44,
0xa44,0xa44,0xa44,0x33,0x33,0x33,0xa38,0x33,0x33,0x33,0x33,0xa35,0xa3e,0xa3e,0xa35,0xa35,
0xa35,0x33,0xa35,0x33,0xa3e,0xa3e,0xa41,0xa3e,0xa41,0xa41,0xa41,0xa35,0x33,0x33,0x33,0x33,
0x33,0x33,0x1518,0x1518,0x1518,0x1518,0x1518,0x1518,0x1518,0x1518,0x1518,0x1518,0x33,0x33,0xa3e,0xa3e,
0xa3b,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x36,0x489,0x489,0x489,
0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,
0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x489,0x474,0x489,0x486,
0x474,0x474,0x474,0x474,0x474,0x474,0x47a,0x36,0x36,0x36,0x36,0x471,0x48f,0x48f,0x48f,0x48f,
0x48f,0x489,0x48c,0x477,0x477,0x477,0x477,0x477,0x477,0x474,0x477,0x47d,0x483,0x483,0x483,0x483,
0x483,0x483,0x483,0x483,0x483,0x483,0x480,0x480,0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x39,0x49e,0x49e,0x39,0x49e,0x39,0x1a1f,0x49e,
0x49e,0x1a1f,0x49e,0x39,0x1a1f,0x49e,0x1a1f,0x1a1f,0x1a1f,0x1a1f,0x1a1f,0x1a1f,0x49e,0x49e,0x49e,0x49e,
0x1a1f,0x49e,0x49e,0x49e,0x49e,0x49e,0x49e,0x49e,0x1a1f,0x49e,0x49e,0x49e,0x39,0x49e,0x39,0x49e,
0x1a1f,0x1a1f,0x49e,0x49e,0x1a1f,0x49e,0x49e,0x49e,0x49e,0x492,0x49e,0x49b,0x492,0x492,0x492,0x492,
0x492,0x492,0x1a1c,0x492,0x492,0x49e,0x39,0x39,0x4a7,0x4a7,0x4a7,0x4a7,0x4a7,0x39,0x4a4,0x39,
0x495,0x495,0x495,0x495,0x495,0x492,0x39,0x39,0x498,0x498,0x498,0x498,0x498,0x498,0x498,0x498,
0x498,0x498,0x39,0x39,0x4a1,0x4a1,0x1425,0x1425,0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,
0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,
0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x39,0x9b7,0x9b7,0x9b7,0x9ba,0x9b7,0x9b7,0x9b7,0x9b7,
0x3c,0x9b7,0x9b7,0x9b7,0x9b7,0x9ba,0x9b7,0x9b7,0x9b7,0x9b7,0x9ba,0x9b7,0x9b7,0x9b7,0x9b7,0x9ba,
0x9b7,0x9b7,0x9b7,0x9b7,0x9ba,0x9b7,0x9b7,0x9b7,0x9b7,0x9b7,0x9b7,0x9b7,0x9b7,0x9b7,0x9b7,0x9b7,
0x9b7,0x9ba,0xa53,0xfe4,0xfe4,0x3c,0x3c,0x3c,0x3c,0x984,0x984,0x987,0x984,0x987,0x987,0x990,
0x987,0x990,0x984,0x984,0x984,0x984,0x984,0x9b1,0x984,0x987,0x98a,0x98a,0x98d,0x996,0x98a,0x98a,
0x9b7,0x9b7,0x9b7,0x9b7,0x1353,0x134d,0x134d,0x134d,0x984,0x984,0x984,0x987,0x984,0x984,0xa47,0x984,
0x3c,0x984,0x984,0x984,0x984,0x987,0x984,0x984,0x984,0x984,0x987,0x984,0x984,0x984,0x984,0x987,
0x984,0x984,0x984,0x984,0x987,0x984,0xa47,0xa47,0xa47,0x984,0x984,0x984,0x984,0x984,0x984,0x984,
0xa47,0x987,0xa47,0xa47,0xa47,0x3c,0xa50,0xa50,0xa4d,0xa4d,0xa4d,0xa4d,0xa4d,0xa4d,0xa4a,0xa4d,
0xa4d,0xa4d,0xa4d,0xa4d,0xa4d,0x3c,0xfdb,0xa4d,0xde3,0xde3,0xfde,0xfe1,0xfdb,0x1161,0x1161,0x1161,
0x1161,0x1350,0x1350,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,
0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,
0x3c,0x3c,0x3c,0x3c,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x3f,0x142b,0x3f,0x3f,0x3f,0x3f,
0x3f,0x142b,0x3f,0x3f,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,
0x4aa,0x4aa,0x4aa,0x4aa,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xdf2,0xa7d,0x42,0xa7d,0xa7d,
0xa7d,0xa7d,0x42,0x42,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0x42,0xa7d,0x42,0xa7d,0xa7d,
0xa7d,0xa7d,0x42,0x42,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xdf2,0xa7d,0x42,0xa7d,0xa7d,
0xa7d,0xa7d,0x42,0x42,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,
0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xdf2,0xa7d,0x42,0xa7d,0xa7d,0xa7d,0xa7d,0x42,0x42,
0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0x42,0xa7d,0x42,0xa7d,0xa7d,0xa7d,0xa7d,0x42,0x42,
0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xdf2,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0x42,
0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xdf2,
0xa7d,0x42,0xa7d,0xa7d,0xa7d,0xa7d,0x42,0x42,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xdf2,
0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,
0xa7d,0xa7d,0xa7d,0x42,0x42,0x1356,0x1356,0xdec,0xdef,0xa77,0xa80,0xa74,0xa74,0xa74,0xa74,0xa80,
0xa80,0xa7a,0xa7a,0xa7a,0xa7a,0xa7a,0xa7a,0xa7a,0xa7a,0xa7a,0xa71,0xa71,0xa71,0xa71,0xa71,0xa71,
0xa71,0xa71,0xa71,0xa71,0xa71,0x42,0x42,0x42,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,
0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0x1731,0x45,0x45,
0x172e,0x172e,0x172e,0x172e,0x172e,0x172e,0x45,0x45,0xa95,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,
0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,0xa98,
0xa98,0xa98,0xa98,0xa92,0xa8f,0x48,0x48,0x48,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,
0xa9e,0xa9e,0xa9e,0xa9b,0xa9b,0xa9b,0xa9e,0xa9e,0xa9e,0x151b,0x151b,0x151b,0x151b,0x151b,0x151b,0x151b,
0x151b,0x4b,0x4b,0x4b,0x4b,0x4b,0x4b,0x4b,0xabf,0xabf,0xabf,0xabf,0xabf,0xabf,0xaa1,0xabf,
0xabf,0xaa4,0xaa4,0xaa4,0xaa4,0xaa4,0xaa4,0xaa4,0xaa4,0xaa4,0xaa7,0xaa4,0xab6,0xab6,0xab9,0xac2,
0xab0,0xaad,0xab6,0xab3,0xac2,0xd05,0x4e,0x4e,0xabc,0xabc,0xabc,0xabc,0xabc,0xabc,0xabc,0xabc,
0xabc,0xabc,0x4e,0x4e,0x4e,0x4e,0x4e,0x4e,0xd08,0xd08,0xd08,0xd08,0xd08,0xd08,0xd08,0xd08,
0xd08,0xd08,0x4e,0x4e,0x4e,0x4e,0x4e,0x4e,0xad1,0xad1,0xb52,0xb55,0xad7,0xb4f,0xad4,0xad1,
0xada,0xae9,0xadd,0xaec,0xaec,0xaec,0xac8,0x51,0xae0,0xae0,0xae0,0xae0,0xae0,0xae0,0xae0,0xae0,
0xae0,0xae0,0x51,0x51,0x51,0x51,0x51,0x51,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,
0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,
0x1953,0x51,0x51,0x51,0x51,0x51,0x51,0x51,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,
0xae3,0xacb,0x1002,0x51,0x51,0x51,0x51,0x51,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,
0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,
0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x54,0x54,
0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x54,0x54,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,
0x54,0x4ce,0x54,0x4ce,0x54,0x4ce,0x54,0x4ce,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,
0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,
0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x54,0x54,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,
0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x54,0x4cb,0x4cb,
0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4c5,0x4cb,0x4c5,0x4c5,0x4c2,0x4cb,0x4cb,0x4cb,0x54,0x4cb,0x4cb,
0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4c2,0x4c2,0x4c2,0x4cb,0x4cb,0x4cb,0x4cb,0x54,0x54,0x4cb,0x4cb,
0x4ce,0x4ce,0x4ce,0x4ce,0x54,0x4c2,0x4c2,0x4c2,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,
0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4c2,0x4c2,0x4c2,0x54,0x54,0x4cb,0x4cb,0x4cb,0x54,0x4cb,0x4cb,
0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4c8,0x4c5,0x54,0xbca,0xbcd,0xbcd,0xbcd,0x100b,0x57,0x14f7,0x14f7,
0x14f7,0x14f7,0x4d7,0x4d7,0x4d7,0x4d7,0x4d7,0x4d7,0x522,0xbdf,0x5a,0x5a,0x6e1,0x522,0x522,0x522,
0x522,0x522,0x528,0x53a,0x528,0x534,0x52e,0x6e4,0x51f,0x6de,0x6de,0x6de,0x6de,0x51f,0x51f,0x51f,
0x51f,0x51f,0x525,0x537,0x525,0x531,0x52b,0x5a,0xdfb,0xdfb,0xdfb,0xdfb,0xdfb,0x1359,0x1359,0x1359,
0x1359,0x1359,0x1359,0x1359,0x1359,0x5a,0x5a,0x5a,0x5d,0x5d,0x5d,0x5d,0x5d,0x5d,0x5d,0x5d,
0x5d,0x5d,0x5d,0x5d,0x5d,0x5d,0x5d,0x5d,0x549,0x549,0x549,0x549,0x549,0x549,0x549,0x549,
0x549,0x549,0x549,0x549,0x549,0x546,0x546,0x546,0x546,0x549,0xafb,0xafe,0xbe5,0xbeb,0xbeb,0xbe8,
0xbe8,0xbe8,0xbe8,0xe01,0xf15,0xf15,0xf15,0xf15,0x114c,0x60,0x60,0x60,0x60,0x60,0x60,0x60,
0x60,0x60,0x60,0x60,0x60,0x60,0x60,0x60,0x579,0x579,0x579,0xb07,0xf1e,0x1011,0x1011,0x1011,
0x1011,0x12ab,0x1737,0x1737,0x63,0x63,0x63,0x63,0x70b,0x70b,0x70b,0x70b,0x70e,0x70e,0x70e,0x70e,
0x70e,0x70e,0x585,0x585,0x582,0x582,0x582,0x582,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0xb13,0xb13,0x66,
0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,
0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x5af,0x5af,0x5af,0x5af,0x5af,0x5af,0x5af,0x5af,
0x5af,0x5af,0x5af,0x69,0x69,0x69,0x69,0x69,0x69,0x69,0x69,0x69,0x69,0x69,0x69,0x69,
0x69,0x69,0x69,0x69,0x69,0x69,0x69,0x69,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,
0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,
0xb2e,0xb2e,0x6c,0xb2e,0xb2e,0xb2e,0xb2e,0xb31,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,
0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb31,0x6c,0x6c,0x6c,0x6c,
0x6c,0x6c,0x6c,0x6c,0x6c,0x6c,0x6c,0x6c,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,
0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0x6f,0x6f,
0x6f,0x6f,0x6f,0x6f,0x6f,0x6f,0x6f,0x6f,0x75,0x843,0x83d,0x843,0x83d,0x843,0x83d,0x843,
0x83d,0x843,0x83d,0x83d,0x840,0x83d,0x840,0x83d,0x840,0x83d,0x840,0x83d,0x840,0x83d,0x840,0x83d,
0x840,0x83d,0x840,0x83d,0x840,0x83d,0x840,0x83d,0x83d,0x83d,0x83d,0x843,0x83d,0x843,0x83d,0x843,
0x83d,0x83d,0x83d,0x83d,0x83d,0x83d,0x843,0x83d,0x83d,0x83d,0x83d,0x83d,0x840,0xc93,0xc93,0x75,
0x75,0x954,0x954,0x91e,0x91e,0x846,0x849,0xc90,0x78,0x78,0x78,0x78,0x78,0x85b,0x85b,0x85b,
0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,
0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x85b,0x113a,0x191a,0x1a01,0x7b,0x85e,0x85e,0x85e,
0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x7b,
0x927,0x927,0x92a,0x92a,0x92a,0x92a,0x92a,0x92a,0x92a,0x92a,0x92a,0x92a,0x92a,0x92a,0x92a,0x92a,
0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,
0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0x867,0xd98,0xd98,0x7e,
0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0x81,0x81,0x81,
0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,
0xb4c,0xc9c,0xb4c,0xb4c,0xb4c,0xc9c,0xb4c,0x84,0x84,0x84,0x84,0x84,0x84,0x84,0x84,0x84,
0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,
0x9db,0x9db,0x9db,0x9db,0x87,0x87,0x87,0x87,0x87,0x87,0x87,0x87,0x87,0x87,0x87,0x87,
0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,
0x60c,0x60c,0x60c,0x60c,0x60c,0x60c,0x60c,0x8a,0x8a,0x8a,0x8a,0x8a,0x8a,0x8a,0x8a,0x8a,
0x8a,0x8a,0x8a,0x5fa,0x5fa,0x5fa,0x5fa,0x5fa,0x8a,0x8a,0x8a,0x8a,0x8a,0xb1f,0x5fd,0x603,
0x609,0x609,0x609,0x609,0x609,0x609,0x609,0x609,0x609,0x600,0x603,0x603,0x603,0x603,0x603,0x603,
0x603,0x603,0x603,0x603,0x603,0x603,0x603,0x8a,0x603,0x603,0x603,0x603,0x603,0x8a,0x603,0x8a,
0x603,0x603,0x8a,0x603,0x603,0x8a,0x603,0x603,0x603,0x603,0x603,0x603,0x603,0x603,0x603,0x606,
0x61e,0x618,0x61e,0x618,0x61b,0x621,0x61e,0x618,0x61b,0x621,0x61e,0x618,0x61b,0x621,0x61e,0x618,
0x136b,0x136b,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,
0x8d,0x8d,0x8d,0x61e,0x618,0x61b,0x621,0x61e,0x618,0x61e,0x618,0x61e,0x618,0x61e,0x61e,0x618,
0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,
0x61b,0x618,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x618,0x61b,0x618,0x618,0x61b,0x61b,0x618,0x618,
0x618,0x618,0x618,0x61b,0x618,0x618,0x61b,0x618,0x61b,0x61b,0x61b,0x618,0x61b,0x61b,0x61b,0x61b,
0x8d,0x8d,0x61b,0x61b,0x61b,0x61b,0x618,0x618,0x61b,0x618,0x618,0x618,0x618,0x61b,0x618,0x618,
0x618,0x618,0x618,0x61b,0x61b,0x61b,0x618,0x618,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,0x8d,
0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,0xb6a,
0x61e,0x61e,0x978,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x615,0x615,0xc24,0xdb3,0x8d,0x8d,
0x87f,0x891,0x88e,0x891,0x88e,0xcb1,0xcb1,0xda4,0xda1,0x882,0x882,0x882,0x882,0x894,0x894,0x894,
0x8ac,0x8af,0x8be,0x90,0x8b2,0x8b5,0x8c1,0x8c1,0x8a9,0x8a0,0x89a,0x8a0,0x89a,0x8a0,0x89a,0x89d,
0x89d,0x8b8,0x8b8,0x8bb,0x8b8,0x8b8,0x8b8,0x90,0x8b8,0x8a6,0x8a3,0x89d,0x90,0x90,0x90,0x90,
0x62a,0x636,0x62a,0xc27,0x62a,0x93,0x62a,0x636,0x62a,0x636,0x62a,0x636,0x62a,0x636,0x62a,0x636,
0x636,0x633,0x62d,0x630,0x636,0x633,0x62d,0x630,0x636,0x633,0x62d,0x630,0x636,0x633,0x62d,0x633,
0x62d,0x633,0x62d,0x630,0x636,0x633,0x62d,0x633,0x62d,0x633,0x62d,0x633,0x62d,0x93,0x93,0x627,
0x77d,0x780,0x795,0x798,0x777,0x780,0x780,0x99,0x75f,0x762,0x762,0x762,0x762,0x75f,0x75f,0x99,
0x96,0x96,0x96,0x96,0x96,0x96,0x96,0x96,0x96,0xb22,0xb22,0xb22,0x9de,0x759,0x639,0x639,
0x99,0x7a7,0x786,0x777,0x780,0x77d,0x777,0x789,0x77a,0x774,0x777,0x795,0x78c,0x783,0x7a4,0x777,
0x7a1,0x7a1,0x7a1,0x7a1,0x7a1,0x7a1,0x7a1,0x7a1,0x7a1,0x7a1,0x792,0x78f,0x795,0x795,0x795,0x7a7,
0x768,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,
0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x765,0x99,
0x99,0x99,0x765,0x765,0x765,0x765,0x765,0x765,0x99,0x99,0x765,0x765,0x765,0x765,0x765,0x765,
0x99,0x99,0x765,0x765,0x765,0x765,0x765,0x765,0x99,0x99,0x765,0x765,0x765,0x99,0x99,0x99,
0xb6d,0xb6d,0xb6d,0xb6d,0x9c,0x9c,0x9c,0x9c,0x9c,0x9c,0x9c,0x9c,0x9c,0x18b7,0x18b7,0x18b7,
0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,0xb73,
0xb73,0xb73,0xb73,0x9f,0x9f,0x9f,0x9f,0x9f,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,
0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,
0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xa2,0xa2,
0xa2,0xa2,0xa2,0xa2,0xa2,0xa2,0xa2,0xa2,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xa5,
0xa5,0x101d,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,
0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0x173d,0x173d,0x173d,0x173d,0x173d,0x173d,0x173d,0x173d,
0x173d,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,
0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xa5,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,
0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xa8,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xba0,0xba0,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xba0,0xa8,0xba0,0xba0,0xa8,0xa8,0xba0,0xa8,0xa8,0xba0,0xba0,0xa8,
0xa8,0xba0,0xba0,0xba0,0xba0,0xa8,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xb9d,0xb9d,
0xb9d,0xb9d,0xa8,0xb9d,0xa8,0xb9d,0xb9d,0xb9d,0xb9d,0xd29,0xb9d,0xb9d,0xa8,0xb9d,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,
0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xb9d,0xb9d,0xb9d,0xb9d,0xba0,0xba0,0xa8,0xba0,
0xba0,0xba0,0xba0,0xa8,0xa8,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xa8,0xba0,0xba0,
0xba0,0xba0,0xba0,0xba0,0xba0,0xa8,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,
0xba0,0xba0,0xa8,0xba0,0xba0,0xba0,0xba0,0xa8,0xba0,0xba0,0xba0,0xba0,0xba0,0xa8,0xba0,0xa8,
0xa8,0xa8,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xa8,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xe16,0xe16,0xa8,0xa8,0xba0,0xba0,0xba0,0xba0,
0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,
0xba0,0xba0,0xba0,0xba0,0xb9d,0xb9d,0xb9d,0xb97,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xf2d,0xf2a,
0xa8,0xa8,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,
0xb9a,0xb9a,0xb9a,0xb9a,0xab,0xba6,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,
0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,0xab,
0xab,0xab,0xab,0xab,0xc36,0xc36,0xc36,0xc36,0xc36,0xc36,0xc36,0xc36,0xc36,0xc36,0xc36,0xc36,
0xc36,0xae,0xc36,0xc36,0xc36,0xc36,0xc30,0xc30,0xc33,0xae,0xae,0xae,0xae,0xae,0xae,0xae,
0xae,0xae,0xae,0xae,0xc3f,0xc3f,0xc3f,0xc3f,0xc3f,0xc3f,0xc3f,0xc3f,0xc3f,0xc3f,0xc3f,0xc3f,
0xc3f,0xc3f,0xc3f,0xc3f,0xc3f,0xc3f,0xc39,0xc39,0xc3c,0xca5,0xca5,0xb1,0xb1,0xb1,0xb1,0xb1,
0xb1,0xb1,0xb1,0xb1,0xc45,0xc45,0xc45,0xc45,0xc45,0xc45,0xc45,0xc45,0xc45,0xc45,0xc45,0xc45,
0xc45,0xc45,0xc45,0xc45,0xc45,0xc45,0xc42,0xc42,0xb4,0xb4,0xb4,0xb4,0xb4,0xb4,0xb4,0xb4,
0xb4,0xb4,0xb4,0xb4,0xc4b,0xc4b,0xc4b,0xc4b,0xc4b,0xc4b,0xc4b,0xc4b,0xc4b,0xc4b,0xc4b,0xc4b,
0xc4b,0xb7,0xc4b,0xc4b,0xc4b,0xb7,0xc48,0xc48,0xb7,0xb7,0xb7,0xb7,0xb7,0xb7,0xb7,0xb7,
0xb7,0xb7,0xb7,0xb7,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,
0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,0xd3b,
0xd3b,0x1539,0x1539,0xba,0xd2c,0xd2c,0xd2c,0xd38,0xd38,0xd38,0xd38,0xd2c,0xd2c,0xd38,0xd38,0xd38,
0xba,0xba,0xba,0xba,0xd38,0xd38,0xd2c,0xd38,0xd38,0xd38,0xd38,0xd38,0xd38,0xd2f,0xd2f,0xd2f,
0xba,0xba,0xba,0xba,0xd32,0xba,0xba,0xba,0xd3e,0xd3e,0xd35,0xd35,0xd35,0xd35,0xd35,0xd35,
0xd35,0xd35,0xd35,0xd35,0xd41,0xd41,0xd41,0xd41,0xd41,0xd41,0xd41,0xd41,0xd41,0xd41,0xd41,0xd41,
0xd41,0xd41,0xd41,0xd41,0xd41,0xd41,0xbd,0xbd,0xd41,0xd41,0xd41,0xd41,0xd41,0xbd,0xbd,0xbd,
0xbd,0xbd,0xbd,0xbd,0xbd,0xbd,0xbd,0xbd,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,
0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0xc0,0xc0,0x153c,0x153c,
0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,
0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0xc0,0x1abe,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,
0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xc3,0xd68,0xd68,0xd68,
0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,
0xd68,0xd68,0xd68,0xc3,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,
0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xc3,0xd68,0xd68,0xc3,0xd68,0xd68,0xd68,0xd68,0xd68,
0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xc3,0xc3,0xd68,0xd68,0xd68,0xd68,
0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xd68,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,
0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,
0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xc3,0xd6b,0xd6b,0xd6b,0xd6b,
0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,
0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xc6,0xc6,0xc6,0xc6,0xc6,0xdad,0xdad,0xdad,0xc9,
0xc9,0xc9,0xc9,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,
0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xda7,0xc9,0xc9,0xc9,0xdaa,
0xdaa,0xdaa,0xdaa,0xdaa,0xdaa,0xdaa,0xdaa,0xdaa,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,
0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,
0xd71,0xd71,0xd71,0xd71,0xd71,0xd71,0xcc,0xd6e,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,
0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,
0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xd7a,0xcf,0xcf,0xd77,0xd77,0xd77,0xd77,0xd77,0xd77,0xd77,0xd77,
0xd77,0xd77,0xcf,0xcf,0xcf,0xcf,0xcf,0xcf,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,
0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd2,0xd2,
0xd7d,0xd2,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,
0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd7d,0xd2,0xd7d,0xd7d,0xd2,0xd2,0xd2,
0xd7d,0xd2,0xd2,0xd7d,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,
0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd5,0xd5,0xd5,0xd5,0xd5,
0xd5,0xd5,0xd5,0xd5,0xe31,0xe31,0xe31,0xe31,0xe31,0xe31,0xe31,0xe31,0xe31,0xe31,0xe31,0x153f,
0x153f,0x17f1,0x17f1,0xdb,0x1119,0x1119,0x1119,0x1119,0x1119,0x1119,0x1119,0x1119,0x1119,0x1119,0x1119,0x1119,
0x1acd,0x132,0x132,0x132,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,
0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe43,0xe3a,0xe3a,0xe40,0xe40,0xe3a,
0xde,0xde,0xe3d,0xe3d,0x1149,0x1149,0x1149,0x1149,0xe1,0xe1,0xe1,0xe1,0xe1,0xe1,0xe1,0xe1,
0xe1,0xe1,0xe1,0xe1,0xca2,0xca2,0xca2,0xca2,0xca2,0xca2,0xca2,0xca2,0xca2,0xca2,0xca2,0xca2,
0xca2,0xca2,0xca2,0xca2,0x1038,0x1038,0x1038,0x1038,0x1038,0x1038,0x1038,0x1542,0x1542,0x1542,0x1542,0x1542,
0x1542,0x1542,0x1542,0x1542,0x1542,0x1542,0x1542,0x1542,0x1542,0x1545,0x18bd,0x18bd,0x193e,0x18bd,0xe4,0x17f4,
0x1377,0x118b,0xf3c,0xf3c,0xe55,0xe52,0xe55,0xe52,0xe52,0xe49,0xe49,0xe49,0xe49,0xe49,0xe49,0x1194,
0x1191,0x1194,0x1191,0x118e,0x118e,0x118e,0x1434,0x1431,0xe7,0xe7,0xe7,0xe7,0xe7,0xe4f,0xe4c,0xe4c,
0xe4c,0xe49,0xe4f,0xe4c,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,
0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xea,0xea,0xea,0xea,0xea,
0xea,0xea,0xea,0xea,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xea,0xe58,0xe58,0xe58,0xe58,
0xe58,0xe58,0xe58,0xea,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xe58,0xea,0xe58,0xe58,0xe58,0xe58,
0xe58,0xe58,0xe58,0xea,0xe5e,0xe5e,0xe5e,0xe5e,0xe5e,0xe5e,0xe5e,0xe5e,0xe5e,0xe5e,0xe5e,0xe5e,
0xe5e,0xe5e,0xe5e,0xe5e,0xe5b,0xe5b,0xe5b,0xe5b,0xe5b,0xe5b,0xe5b,0xe5b,0xe5b,0xe5b,0xed,0xed,
0xed,0xed,0xed,0xed,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xf0,0x1437,0xf0,0xf0,0xf0,0xf0,
0xf0,0x1437,0xf0,0xf0,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,
0xeb8,0xeb8,0xeb8,0xeb8,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,
0xe67,0xe67,0xe67,0xf3,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,
0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,0xe64,
0xe64,0xe64,0xe64,0xf3,0xe79,0xe6d,0xe6d,0xe6d,0xf6,0xe6d,0xe6d,0xf6,0xf6,0xf6,0xf6,0xf6,
0xe6d,0xe6d,0xe6d,0xe6d,0xe79,0xe79,0xe79,0xe79,0xf6,0xe79,0xe79,0xe79,0xf6,0xe79,0xe79,0xe79,
0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,0xe79,
0xe79,0xe79,0xe79,0xe79,0x195c,0x195c,0xf6,0xf6,0xe6a,0xe6a,0xe6a,0xf6,0xf6,0xf6,0xf6,0xe70,
0xe73,0xe73,0xe73,0xe73,0xe73,0xe73,0xe73,0xe73,0x1959,0xf6,0xf6,0xf6,0xf6,0xf6,0xf6,0xf6,
0xe76,0xe76,0xe76,0xe76,0xe76,0xe76,0xe7c,0xe7c,0xe73,0xf6,0xf6,0xf6,0xf6,0xf6,0xf6,0xf6,
0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0x119a,0x119a,0xf9,0xf9,0xf9,0xf9,
0xe88,0xe88,0xe88,0xe88,0xe88,0xe8b,0xe8b,0xe8b,0xe88,0xe88,0xe8b,0xe88,0xe88,0xe88,0xe88,0xe88,
0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xf9,0xf9,0xf9,0xf9,0xf9,0xf9,0xe85,0xe85,0xe85,0xe85,
0xe85,0xe85,0xe85,0xe85,0xe85,0xe85,0x1197,0xf9,0xf9,0xf9,0xe82,0xe82,0xe91,0xe91,0xe91,0xe91,
0xfc,0xfc,0xfc,0xfc,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe8e,0xe91,0xe91,0xe91,
0xe91,0xe91,0xfc,0xfc,0xfc,0xfc,0xfc,0xfc,0xfc,0xfc,0xfc,0xfc,0x154e,0x1554,0x1551,0x189c,
0x17f7,0x18c0,0x18c0,0x18c0,0x18c0,0x18c0,0x1962,0x195f,0x1965,0x195f,0x1965,0x1a25,0x1ac1,0x1ac1,0x1ac1,0xff,
0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
0xeb5,0xeb5,0xeb5,0xeb2,0xeb2,0xea9,0xea9,0xeb2,0xeaf,0xeaf,0xeaf,0xeaf,0x1ac4,0x102,0x102,0x102,
0x1314,0x1314,0x1314,0x1317,0x1317,0x1317,0x130e,0x130e,0x1311,0x130e,0x156,0x156,0x156,0x156,0x156,0x156,
0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0x1443,0x1443,0x105,0x105,0x105,0x105,0x105,0x105,0x105,0xebb,
0x137d,0x105,0x105,0x105,0x105,0x105,0x105,0x105,0x105,0x105,0x105,0x105,0x105,0x105,0x105,0x137a,
0xc72,0xc72,0xc72,0xc72,0xc72,0xc72,0xc72,0xc72,0xc72,0xc72,0xc72,0xc72,0xc72,0xc72,0xc72,0xc75,
0xee8,0xed9,0xed3,0xee5,0xee2,0xedc,0xedc,0xeeb,0xed6,0xedf,0x108,0x108,0x108,0x108,0x108,0x108,
0xf6f,0xf6f,0xf5a,0xf6f,0xf72,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0x10e,0x10e,0x10e,0x10e,
0xf69,0xf69,0xf69,0xf69,0xf69,0xf69,0xf69,0xf69,0xf69,0xf69,0xf7b,0xf7b,0xf60,0xf66,0xf7b,0xf7b,
0xf63,0xf60,0xf60,0xf60,0xf60,0xf60,0xf60,0xf60,0xf60,0xf60,0xf60,0xf5d,0xf5d,0xf5d,0xf5d,0xf5d,
0xf5d,0xf5d,0xf5d,0xf5d,0xf60,0xf60,0xf60,0xf60,0xf60,0xf60,0xf60,0xf60,0xf60,0x10e,0x10e,0x10e,
0x111,0x111,0x1a2b,0x1a28,0x1a2b,0x1a2b,0x1a2b,0x1aca,0x1ac7,0x1aca,0x1ac7,0x111,0x111,0x111,0x111,0x111,
0x111,0x111,0x111,0x111,0x111,0x111,0x111,0x111,0x111,0x111,0x111,0x111,0x111,0x111,0x111,0x111,
0x111,0x1aca,0x1ac7,0x155d,0x144c,0x144c,0x1380,0x1074,0x1074,0x1074,0x1074,0x1074,0xf8a,0xf8a,0xf8a,0xf8a,
0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,
0xf87,0xf87,0xf8d,0xf8d,0x114,0x114,0x114,0x114,0x114,0x114,0x114,0x114,0xf96,0xf96,0xf96,0xf96,
0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,0xf96,
0xf96,0xf96,0xf90,0xf90,0xf90,0xf90,0x11a3,0x11a3,0x117,0x117,0x117,0xf93,0x1563,0x1563,0x1563,0x1563,
0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,
0x1563,0x1563,0x1563,0x1563,0x1563,0x174c,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,
0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,
0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0x11a,0xf9f,0xf9f,0xf9f,0x1569,0x1569,0x1569,0x1569,0x1569,
0x1569,0x1569,0x1569,0x1569,0x1569,0x1569,0x1569,0x11d,0xf9c,0xf9c,0xf9c,0xf9c,0x1566,0x11d,0x11d,0x11d,
0x11d,0x11d,0x11d,0x11d,0x11d,0x11d,0x11d,0x11d,0xfa2,0xfa2,0xfa2,0xfa2,0xfa2,0xfa2,0xfa2,0xfa2,
0xfa2,0xfa2,0xfa2,0xfa2,0xfa2,0xfa2,0xfa2,0xfa2,0xfa2,0xfa2,0x1974,0x1974,0x1974,0x1974,0x1974,0x1974,
0x1974,0x120,0x120,0x120,0x120,0x120,0x120,0x120,0x109b,0x109b,0x109b,0x109b,0x1098,0x1098,0x1098,0x1098,
0x1098,0x1098,0x1098,0x1098,0x1089,0x1089,0x1089,0x1089,0x1089,0x1089,0x1089,0x1089,0x1098,0x1098,0x108f,0x108c,
0x123,0x123,0x123,0x109e,0x109e,0x1092,0x1092,0x1092,0x1095,0x1095,0x1095,0x1095,0x1095,0x1095,0x1095,0x1095,
0x1095,0x1095,0x123,0x123,0x123,0x109b,0x109b,0x109b,0x10a1,0x10a1,0x10a1,0x10a1,0x10a1,0x10a1,0x10a1,0x10a1,
0x10a1,0x10a1,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10b6,0x10b6,0x10b6,0x10b6,0x10b6,0x10b6,0x10b6,0x10b6,
0x10b6,0x10b6,0x10b9,0x10b9,0x126,0x126,0x126,0x126,0x126,0x126,0x126,0x126,0x126,0x126,0x126,0x126,
0x126,0x126,0x126,0x126,0x126,0x126,0x126,0x126,0x10e0,0x10e0,0x10e0,0x10e0,0x10da,0x17fd,0x129,0x129,
0x129,0x129,0x129,0x129,0x129,0x129,0x10e6,0x10e6,0x10dd,0x10dd,0x10dd,0x10dd,0x10dd,0x10dd,0x10dd,0x10dd,
0x10dd,0x10dd,0x129,0x129,0x129,0x129,0x129,0x129,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x10f8,
0x10f8,0x10f8,0x10f8,0x10f8,0x10f8,0x10f8,0x10f8,0x10f8,0x10f8,0x10f8,0x10fe,0x1101,0x12c,0x12c,0x12c,0x12c,
0x12c,0x12c,0x12c,0x12c,0x12c,0x12c,0x12c,0x10fb,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,
0x1113,0x1107,0x1107,0x1107,0x1107,0x1107,0x1107,0x1110,0x1110,0x1107,0x1107,0x1110,0x1110,0x1107,0x1107,0x12f,
0x12f,0x12f,0x12f,0x12f,0x12f,0x12f,0x12f,0x12f,0x1113,0x1113,0x1113,0x1107,0x1113,0x1113,0x1113,0x1113,
0x1113,0x1113,0x1113,0x1113,0x1107,0x1110,0x12f,0x12f,0x110d,0x110d,0x110d,0x110d,0x110d,0x110d,0x110d,0x110d,
0x110d,0x110d,0x12f,0x12f,0x110a,0x1116,0x1116,0x1116,0x1575,0x132,0x132,0x132,0x132,0x132,0x132,0x132,
0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x132,
0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x132,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,
0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,0x111c,
0x111c,0x111c,0x111c,0x111c,0x111c,0x111f,0x135,0x135,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,
0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,0x1122,
0x1122,0x1122,0x1122,0x1122,0x1122,0x138,0x138,0x138,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,
0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x13b,0x13b,0x13b,0x13b,0x13b,0x13b,0x13b,
0x13b,0x13b,0x13b,0x13b,0x13b,0x13b,0x13b,0x13b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,
0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,0x112b,
0x112b,0x112b,0x13e,0x13e,0x13e,0x13e,0x13e,0x1128,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,
0x112e,0x112e,0x112e,0x112e,0x141,0x141,0x141,0x141,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,
0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x144,0x144,0x144,0x144,
0x144,0x144,0x144,0x144,0x144,0x144,0x144,0x144,0x11a9,0x11a9,0x11a9,0x11a9,0x11b2,0x11a9,0x11a9,0x11a9,
0x11b2,0x11a9,0x11a9,0x11a9,0x11a9,0x11a6,0x147,0x147,0x11af,0x11af,0x11af,0x11af,0x11af,0x11af,0x11af,0x11b5,
0x11af,0x11b5,0x11af,0x11af,0x11af,0x11b5,0x11b5,0x147,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,
0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x14a,0x14a,
0x14a,0x14a,0x14a,0x14a,0x14a,0x14a,0x14a,0x14a,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,
0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d0,0x11bb,0x11d0,
0x11bb,0x11bb,0x11bb,0x11bb,0x11bb,0x11bb,0x11bb,0x14d,0x11c4,0x11cd,0x11bb,0x11cd,0x11cd,0x11bb,0x11bb,0x11bb,
0x11bb,0x11bb,0x11bb,0x11bb,0x11bb,0x11d0,0x11d0,0x11d0,0x11d0,0x11d0,0x11d0,0x11bb,0x11bb,0x11c1,0x11c1,0x11c1,
0x11c1,0x11c1,0x11c1,0x11c1,0x11c1,0x14d,0x14d,0x11be,0x11ca,0x11ca,0x11ca,0x11ca,0x11ca,0x11ca,0x11ca,0x11ca,
0x11ca,0x11ca,0x14d,0x14d,0x14d,0x14d,0x14d,0x14d,0x11ca,0x11ca,0x11ca,0x11ca,0x11ca,0x11ca,0x11ca,0x11ca,
0x11ca,0x11ca,0x14d,0x14d,0x14d,0x14d,0x14d,0x14d,0x11c7,0x11c7,0x11c7,0x11c7,0x11c7,0x11c7,0x11c7,0x11d6,
0x11d9,0x11d9,0x11d9,0x11d9,0x11c7,0x11c7,0x14d,0x14d,0x15c0,0x15c0,0x15c0,0x15c0,0x15c0,0x15c0,0x15c0,0x15c0,
0x15c0,0x15c0,0x15c0,0x15c0,0x15c0,0x15c0,0x15bd,0x1adf,0x1329,0x1302,0x1320,0x1320,0x1320,0x1320,0x1320,0x1320,
0x1320,0x1308,0x1305,0x12fc,0x12fc,0x1326,0x12fc,0x12fc,0x12fc,0x12fc,0x130b,0x14eb,0x14f1,0x14ee,0x14ee,0x193b,
0x1716,0x1716,0x1aac,0x150,0x150,0x150,0x150,0x150,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,
0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11e5,0x11e5,0x11e8,0x11f1,0x11eb,0x11eb,0x11eb,0x11f1,
0x153,0x153,0x153,0x153,0x153,0x153,0x153,0x153,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,
0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,
0x12ea,0x12ea,0x12ea,0x12ea,0x12ea,0x159,0x159,0x159,0x120f,0x1203,0x1203,0x1203,0x1203,0x1203,0x1203,0x1206,
0x1215,0x1215,0x1203,0x1203,0x1203,0x1203,0x15c,0x131a,0x1209,0x1209,0x1209,0x1209,0x1209,0x1209,0x1209,0x1209,
0x1209,0x1209,0x15c,0x15c,0x15c,0x15c,0x1203,0x1203,0x1233,0x1227,0x1233,0x15f,0x15f,0x15f,0x15f,0x15f,
0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,0x15f,
0x15f,0x15f,0x15f,0x1230,0x1230,0x1236,0x122a,0x122d,0x124b,0x124b,0x124b,0x1245,0x1245,0x123c,0x1245,0x1245,
0x123c,0x1245,0x1245,0x124e,0x1248,0x123f,0x162,0x162,0x1242,0x1242,0x1242,0x1242,0x1242,0x1242,0x1242,0x1242,
0x1242,0x1242,0x162,0x162,0x162,0x162,0x162,0x162,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x1254,0x165,
0x165,0x165,0x165,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,
0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,0x1251,
0x165,0x165,0x165,0x165,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,
0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x125d,0x168,0x125a,0x1257,0x1257,0x1257,0x1257,
0x1257,0x1257,0x1257,0x1257,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,
0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x16b,0x16b,0x16b,0x1266,0x1269,0x1269,
0x1269,0x1269,0x1269,0x1269,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,
0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x1272,0x16e,0x16e,0x126f,0x126f,0x126f,0x126f,
0x126f,0x126f,0x126f,0x126f,0x1278,0x1278,0x1278,0x1278,0x1278,0x1278,0x1278,0x1278,0x1278,0x1278,0x1278,0x1278,
0x1278,0x1278,0x1278,0x1278,0x1278,0x1278,0x1278,0x171,0x171,0x171,0x171,0x171,0x1275,0x1275,0x1275,0x1275,
0x1275,0x1275,0x1275,0x1275,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,
0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,0x127e,
0x127e,0x127e,0x127e,0x177,0x1299,0x1299,0x17a,0x17a,0x17a,0x17a,0x17a,0x17a,0x17a,0x17a,0x17a,0x17a,
0x17a,0x197d,0x17a,0x17a,0x14ca,0x14ca,0x14ca,0x14ca,0x14ca,0x14ca,0x14ca,0x14ca,0x14ca,0x14ca,0x14ca,0x14ca,
0x14ca,0x14ca,0x14ca,0x14ca,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,
0x129f,0x129f,0x129f,0x17d,0x1a4f,0x1a4f,0x1a4f,0x1a4f,0x1a4f,0x1a4f,0x1a4f,0x1a52,0x1a4c,0x279,0x279,0x279,
0x279,0x279,0x279,0x279,0x187e,0x187e,0x187e,0x187e,0x187e,0x187e,0x187e,0x187e,0x187e,0x187e,0x187e,0x187e,
0x187e,0x1ad0,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,
0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,0x180,
0x180,0x180,0x180,0x180,0x180,0x180,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,
0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,0x138c,
0x12f6,0x13f5,0x13f2,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,
0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f3,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,
0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f0,0x12f3,0x12f0,0x12f0,0x13f5,0x13f5,
0x13f5,0x13f5,0x13f5,0x13f2,0x13f5,0x13f5,0x13f5,0x1881,0x183,0x183,0x183,0x183,0x12ed,0x12ed,0x12ed,0x12ed,
0x12ed,0x12ed,0x12ed,0x12ed,0x12ed,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x1419,0x1419,0x183,0x183,
0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x1920,0x1920,0x1920,0x1920,
0x1920,0x1920,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,
0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,0x183,
0x183,0x183,0x183,0x183,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,
0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x1395,0x138f,0x138f,0x138f,
0x186,0x186,0x1392,0x186,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x1398,0x13a1,0x139b,0x139b,0x13a1,0x13a1,
0x13a1,0x139b,0x13a1,0x139b,0x139b,0x139b,0x13a4,0x13a4,0x189,0x189,0x189,0x189,0x189,0x189,0x189,0x189,
0x139e,0x139e,0x139e,0x139e,0x18c,0x13aa,0x13aa,0x13aa,0x13aa,0x13aa,0x13aa,0x18c,0x18c,0x13aa,0x13aa,0x13aa,
0x13aa,0x13aa,0x13aa,0x18c,0x18c,0x13aa,0x13aa,0x13aa,0x13aa,0x13aa,0x13aa,0x18c,0x18c,0x18c,0x18c,0x18c,
0x18c,0x18c,0x18c,0x18c,0x13aa,0x13aa,0x13aa,0x13aa,0x13aa,0x13aa,0x13aa,0x18c,0x13aa,0x13aa,0x13aa,0x13aa,
0x13aa,0x13aa,0x13aa,0x18c,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,
0x161d,0x161d,0x161d,0x161d,0x13ad,0x13ad,0x13ad,0x13ad,0x13ad,0x13ad,0x13b0,0x13c2,0x13c2,0x13b6,0x13b6,0x13b6,
0x13b6,0x13b6,0x18f,0x18f,0x18f,0x18f,0x13b3,0x13b3,0x13b3,0x13b3,0x13b3,0x13b3,0x13b3,0x13b3,0x13b3,0x13b3,
0x13b3,0x13b3,0x13b3,0x13b3,0x13b3,0x13b3,0x13b9,0x13b9,0x13b9,0x13b9,0x13b9,0x13b9,0x13b9,0x13b9,0x13b9,0x13b9,
0x18f,0x18f,0x18f,0x18f,0x18f,0x18f,0x18f,0x18f,0x18f,0x18f,0x18f,0x18f,0x18f,0x18f,0x18f,0x1584,
0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,
0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x192,0x192,0x192,0x192,0x192,0x192,0x192,
0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x195,
0x195,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x1587,
0x195,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13fe,
0x195,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,0x13c8,
0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,
0x1587,0x1587,0x1587,0x1587,0x1587,0x1587,0x195,0x195,0x195,0x195,0x195,0x195,0x195,0x195,0x195,0x195,
0x1413,0x1410,0x1410,0x1410,0x1410,0x1410,0x159c,0x159c,0x159c,0x159c,0x159c,0x159f,0x170d,0x159f,0x159f,0x159f,
0x17d9,0x188a,0x188a,0x18c3,0x18c3,0x1a8e,0x1b39,0x1b39,0x198,0x198,0x198,0x198,0x198,0x198,0x198,0x198,
0x159f,0x159f,0x159f,0x159f,0x159f,0x159f,0x159c,0x159c,0x159c,0x159f,0x159c,0x170a,0x170a,0x198,0x198,0x198,
0x159f,0x159c,0x159c,0x159f,0x188a,0x188a,0x188a,0x1926,0x1926,0x1a07,0x1a8e,0x1b39,0x1b39,0x198,0x198,0x198,
0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,
0x13cb,0x13cb,0x13cb,0x13cb,0x19b,0x19b,0x19b,0x19b,0x19b,0x19b,0x19b,0x19b,0x19b,0x19b,0x19b,0x19b,
0x1467,0x15a5,0x1467,0x1467,0x1467,0x1467,0x1467,0x1467,0x1467,0x1467,0x1467,0x1467,0x1467,0x15a5,0x15a5,0x15a5,
0x15a5,0x15a5,0x15a5,0x175e,0x175e,0x19e,0x1809,0x1809,0x1809,0x1809,0x1809,0x1809,0x1809,0x1809,0x1ad3,0x1ad3,
0x1ad3,0x1ad3,0x1ad3,0x1ad3,0x1ad3,0x1ad3,0x1ad3,0x1ad3,0x19e,0x19e,0x19e,0x19e,0x19e,0x19e,0x19e,0x19e,
0x19e,0x19e,0x19e,0x1983,0x1806,0x1806,0x1806,0x1806,0x1806,0x1806,0x1806,0x1806,0x1806,0x1806,0x1806,0x1806,
0x146d,0x146d,0x146d,0x146d,0x1a1,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,
0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,
0x1a1,0x146d,0x146d,0x1a1,0x146d,0x1a1,0x1a1,0x146d,0x1a1,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,
0x146d,0x146d,0x146d,0x1a1,0x146d,0x146d,0x146d,0x146d,0x1a1,0x146d,0x1a1,0x146d,0x1a1,0x1a1,0x1a1,0x1a1,
0x1a1,0x1a1,0x146d,0x1a1,0x1a1,0x1a1,0x1a1,0x146d,0x1a1,0x146d,0x1a1,0x146d,0x1a1,0x146d,0x146d,0x146d,
0x1a1,0x146d,0x146d,0x1a1,0x146d,0x1a1,0x1a1,0x146d,0x1a1,0x146d,0x1a1,0x146d,0x1a1,0x146d,0x1a1,0x146d,
0x1a1,0x146d,0x146d,0x1a1,0x146d,0x1a1,0x1a1,0x146d,0x146d,0x146d,0x146d,0x1a1,0x146d,0x146d,0x146d,0x146d,
0x146d,0x146d,0x146d,0x1a1,0x146d,0x146d,0x146d,0x146d,0x1a1,0x146d,0x146d,0x146d,0x146d,0x1a1,0x146d,0x1a1,
0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x1a1,0x146d,0x146d,0x146d,0x146d,0x146d,
0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x1a1,0x1a1,0x1a1,0x1a1,
0x1a1,0x146d,0x146d,0x146d,0x1a1,0x146d,0x146d,0x146d,0x146d,0x146d,0x1a1,0x146d,0x146d,0x146d,0x146d,0x146d,
0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x146d,0x1a1,0x1a1,0x1a1,0x1a1,
0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,
0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x146a,0x146a,0x1a1,0x1a1,
0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1a1,0x1482,0x1482,0x1482,0x1482,
0x1482,0x1482,0x1482,0x1470,0x1470,0x1470,0x1470,0x1470,0x147f,0x1470,0x1473,0x1473,0x1470,0x1470,0x1470,0x1476,
0x1476,0x1a4,0x147c,0x147c,0x147c,0x147c,0x147c,0x147c,0x147c,0x147c,0x147c,0x147c,0x1479,0x1485,0x1485,0x1485,
0x1989,0x1986,0x1986,0x1ad6,0x1a4,0x1a4,0x1a4,0x1a4,0x1a4,0x1a4,0x1a4,0x1a4,0x162f,0x162f,0x162f,0x162f,
0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x1491,0x1491,0x1491,0x1491,
0x1491,0x1491,0x1491,0x1491,0x1491,0x1491,0x1491,0x148e,0x1488,0x1488,0x148e,0x148e,0x1497,0x1497,0x1491,0x1494,
0x1494,0x148e,0x148b,0x1a7,0x1a7,0x1a7,0x1a7,0x1a7,0x1a7,0x1a7,0x1a7,0x1a7,0x149a,0x149a,0x149a,0x149a,
0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,0x149a,
0x149a,0x149a,0x149a,0x149a,0x1aa,0x1aa,0x1aa,0x1aa,0x1761,0x1761,0x149a,0x149a,0x1761,0x1761,0x1761,0x1761,
0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1aa,0x1aa,0x1761,0x1761,
0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x14a6,0x14a6,0x14a6,0x14a6,
0x14a6,0x1a37,0x1a37,0x1a37,0x1a37,0x1a37,0x1a37,0x1ad,0x1ad,0x1ad,0x1ad,0x1a31,0x14a6,0x14a3,0x14a3,0x14a3,
0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x1a34,0x1a34,0x1a34,0x1a34,
0x1a34,0x1a34,0x1a34,0x1a34,0x1ad,0x1ad,0x1ad,0x1ad,0x1ad,0x1ad,0x1ad,0x14a0,0x14a0,0x14a0,0x14a0,0x14a9,
0x14a9,0x14a9,0x14a9,0x14a9,0x14a9,0x14a9,0x14a9,0x14a9,0x14a9,0x14a9,0x14a9,0x14a9,0x14ca,0x14ca,0x14ca,0x14ca,
0x14ca,0x14ca,0x14ca,0x14ca,0x14ca,0x1b0,0x1b0,0x1b0,0x1b0,0x1b0,0x1b0,0x1b0,0x14c7,0x14c7,0x14c7,0x14c7,
0x14c7,0x14c7,0x14c7,0x14c7,0x14c7,0x14c7,0x1b0,0x1b0,0x1b0,0x1b0,0x1b0,0x1b0,0x14cd,0x14cd,0x14cd,0x14cd,
0x14cd,0x14cd,0x14cd,0x14cd,0x1b3,0x1b3,0x1b3,0x1b3,0x1b3,0x1b3,0x1b3,0x1b3,0x1323,0x1320,0x1323,0x12ff,
0x1320,0x1326,0x1326,0x1329,0x1326,0x1329,0x132c,0x1320,0x1329,0x1329,0x1320,0x1320,0x14df,0x14df,0x14df,0x14df,
0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14d0,0x14d9,0x14d0,0x14d9,0x14d9,0x14d0,0x14d0,0x14d0,0x14d0,
0x14d0,0x14d0,0x14dc,0x14d3,0x1a3a,0x1b6,0x1b6,0x1b6,0x1b6,0x1b6,0x1b6,0x1b6,0x15b1,0x15b1,0x15b1,0x15b1,
0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x15b1,0x1b9,0x1b9,0x15ae,0x15ae,0x15ae,0x15ae,
0x15ae,0x15b4,0x1b9,0x1b9,0x1b9,0x1b9,0x1b9,0x1b9,0x1b9,0x1b9,0x1b9,0x1b9,0x1719,0x1710,0x1710,0x1710,
0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,
0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1710,0x1bf,0x1bf,0x1bf,0x1bf,0x1adf,0x1c2,0x1c2,0x1c2,
0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,
0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,0x1c2,
0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x1c5,0x1c5,0x1c5,0x1c5,0x1c5,
0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x1c5,0x1c5,0x1c5,
0x1c5,0x1c5,0x1c5,0x1c5,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x1c5,0x1c5,
0x15c9,0x15c3,0x15c6,0x15cf,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x1c8,0x1c8,0x1c8,0x1c8,
0x1c8,0x1c8,0x1c8,0x1c8,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,
0x15ba,0x15ba,0x15ba,0x15ba,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,
0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x198c,0x198c,0x198c,0x198c,0x1cb,0x1cb,0x1cb,
0x1cb,0x1cb,0x1cb,0x1cb,0x1a91,0x1a91,0x1a91,0x1a91,0x1a91,0x1a91,0x1a91,0x1a91,0x1a91,0x1a91,0x1a91,0x1a91,
0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,0x1cb,
0x1cb,0x1cb,0x1cb,0x1cb,0x1776,0x171c,0x15de,0x1722,0x1ce,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,
0x15e7,0x1ce,0x1ce,0x15e7,0x15e7,0x1ce,0x1ce,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,
0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x1ce,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x1ce,0x15e7,0x15e7,
0x1ce,0x15e7,0x15e7,0x15e7,0x15e7,0x15e7,0x1ce,0x1a16,0x171f,0x15e7,0x15d8,0x15de,0x15d8,0x15de,0x15de,0x15de,
0x15de,0x1ce,0x1ce,0x15de,0x15de,0x1ce,0x1ce,0x15e1,0x15e1,0x15e4,0x1ce,0x1ce,0x1779,0x1ce,0x1ce,0x1ce,
0x1ce,0x1ce,0x1ce,0x15d8,0x1ce,0x1ce,0x1ce,0x1ce,0x1ce,0x15ea,0x15e7,0x15e7,0x15e7,0x15e7,0x15de,0x15de,
0x1ce,0x1ce,0x15db,0x15db,0x15db,0x15db,0x15db,0x15db,0x15db,0x1ce,0x1ce,0x1ce,0x15db,0x15db,0x15db,0x15db,
0x15db,0x1ce,0x1ce,0x1ce,0x1ce,0x1ce,0x1ce,0x1ce,0x1ce,0x1ce,0x1ce,0x1ce,0x15ff,0x15ff,0x15ff,0x15ff,
0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x1d1,0x15ff,
0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15ff,0x15f9,0x15f9,0x15f9,0x15ed,
0x15ed,0x15ed,0x15f9,0x15f9,0x15ed,0x15fc,0x15f0,0x15ed,0x1602,0x1602,0x15f6,0x1602,0x1602,0x15f3,0x180c,0x1d1,
0x1611,0x1611,0x1611,0x1605,0x1605,0x1605,0x1605,0x1605,0x1605,0x1608,0x160b,0x1d4,0x1d4,0x1d4,0x1d4,0x1d4,
0x160e,0x160e,0x160e,0x160e,0x160e,0x160e,0x160e,0x160e,0x160e,0x160e,0x1d4,0x1d4,0x1d4,0x1d4,0x1d4,0x1d4,
0x177c,0x177c,0x177c,0x177c,0x161d,0x161a,0x1a3d,0x1a3d,0x1ae5,0x1ae8,0x1ae2,0x1ae2,0x1d7,0x1d7,0x1d7,0x1d7,
0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,
0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,
0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,
0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,
0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,
0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,
0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,0x1da,
0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,0x162f,
0x162f,0x162f,0x162f,0x1626,0x1629,0x162c,0x162f,0x1dd,0x1dd,0x1dd,0x1dd,0x1dd,0x1dd,0x1dd,0x1dd,0x1dd,
0x163e,0x163e,0x163e,0x163e,0x163e,0x1632,0x1632,0x1e0,0x1e0,0x1e0,0x1e0,0x1635,0x1635,0x1635,0x1635,0x1635,
0x163b,0x163b,0x163b,0x163b,0x163b,0x163b,0x1638,0x1e0,0x1e0,0x1e0,0x1e0,0x1e0,0x1e0,0x1e0,0x1e0,0x1e0,
0x1647,0x1647,0x1647,0x1647,0x1647,0x1e3,0x1e3,0x1644,0x1644,0x1644,0x1644,0x1644,0x1644,0x1644,0x1644,0x1644,
0x1641,0x1641,0x1641,0x1641,0x1641,0x1641,0x1641,0x1e3,0x1e3,0x1e3,0x1e3,0x1e3,0x1e3,0x1e3,0x1e3,0x1e3,
0x164a,0x165c,0x165c,0x1650,0x1659,0x1e6,0x1e6,0x1e6,0x1e6,0x1e6,0x1e6,0x1e6,0x1e6,0x1e6,0x1e6,0x1e6,
0x1653,0x1653,0x1653,0x1653,0x1653,0x1653,0x1653,0x1653,0x1653,0x1653,0x1e6,0x1e6,0x1e6,0x1e6,0x1e6,0x1e6,
0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,
0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1662,0x1e9,
0x166e,0x166e,0x166e,0x166e,0x166e,0x1668,0x1671,0x166e,0x166e,0x166e,0x166e,0x166e,0x166e,0x166e,0x166e,0x166e,
0x166b,0x166b,0x166b,0x166b,0x166b,0x166b,0x166b,0x166b,0x166b,0x166b,0x166e,0x166e,0x166e,0x166e,0x166e,0x1ec,
0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,
0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1677,0x1ef,
0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,
0x1683,0x1683,0x1683,0x1683,0x1683,0x1683,0x1680,0x1680,0x1680,0x1680,0x1680,0x1f2,0x1f2,0x1f2,0x1f2,0x1f2,
0x169b,0x169b,0x169e,0x169e,0x16a1,0x1692,0x1f5,0x1f5,0x1f5,0x1f5,0x1f5,0x1f5,0x1f5,0x1f5,0x1f5,0x1f5,
0x1698,0x1698,0x1698,0x1698,0x1698,0x1698,0x1698,0x1698,0x1698,0x1698,0x1f5,0x1692,0x1692,0x1692,0x1692,0x1692,
0x1692,0x1692,0x1f5,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,
0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x1f5,0x1f5,0x1f5,0x1f5,0x1f5,0x169b,0x169b,0x169b,
0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,
0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x1f8,0x1f8,0x1f8,0x1f8,0x1f8,0x1f8,0x1f8,
0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,0x16b3,
0x16b3,0x16b3,0x1fb,0x1fb,0x1fb,0x1fb,0x1fb,0x1fb,0x1fb,0x16b0,0x16b0,0x16b0,0x16b0,0x1fb,0x1fb,0x1fb,
0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16b6,
0x16c8,0x16c8,0x16b6,0x16b6,0x16b6,0x16b6,0x201,0x201,0x16c8,0x16c8,0x16cb,0x16cb,0x16b6,0x16b6,0x16c8,0x16bc,
0x16b9,0x16bf,0x16d1,0x16d1,0x16c2,0x16c2,0x16c5,0x16c5,0x16c5,0x16d1,0x1785,0x1785,0x1785,0x1785,0x1785,0x1785,
0x1785,0x1785,0x1785,0x1785,0x1785,0x1785,0x1785,0x1785,0x1782,0x1782,0x1782,0x1782,0x177f,0x177f,0x201,0x201,
0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,
0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,0x201,
0x204,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,
0x16d4,0x16d4,0x16d4,0x16d4,0x16d4,0x204,0x204,0x204,0x204,0x204,0x204,0x204,0x204,0x204,0x204,0x204,
0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x207,0x207,0x207,0x207,
0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,
0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,
0x16d7,0x16d7,0x207,0x207,0x207,0x207,0x207,0x207,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,
0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,
0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x207,0x207,0x1aeb,0x1aeb,0x207,0x207,
0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,
0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,0x207,
0x16da,0x16e9,0x16e0,0x16dd,0x16ef,0x16ef,0x16e3,0x16ef,0x20a,0x20a,0x20a,0x20a,0x20a,0x20a,0x20a,0x20a,
0x16e6,0x16e6,0x16e6,0x16e6,0x16e6,0x16e6,0x16e6,0x16e6,0x16e6,0x16e6,0x20a,0x20a,0x20a,0x20a,0x20a,0x20a,
0x16f5,0x16f5,0x16f5,0x16f5,0x16f5,0x16f5,0x16f5,0x16f5,0x16f5,0x16f5,0x16f2,0x16f2,0x16f2,0x16f2,0x16f2,0x16f2,
0x16f2,0x16f2,0x16f2,0x20d,0x20d,0x20d,0x20d,0x20d,0x20d,0x20d,0x20d,0x20d,0x20d,0x20d,0x20d,0x16fb,
0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,
0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x1797,0x198f,0x210,0x210,0x1788,0x1788,0x1788,
0x1794,0x1794,0x1788,0x1788,0x1788,0x1788,0x1794,0x1788,0x1788,0x1788,0x1788,0x178b,0x210,0x210,0x210,0x210,
0x1791,0x1791,0x1791,0x1791,0x1791,0x1791,0x1791,0x1791,0x1791,0x1791,0x178e,0x178e,0x179a,0x179a,0x179a,0x178e,
0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,
0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,
0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x213,0x17af,0x17af,0x17af,0x17af,0x17af,0x17af,0x17af,0x17af,
0x17af,0x17af,0x17af,0x17af,0x17af,0x17af,0x17af,0x17af,0x17af,0x17af,0x17af,0x219,0x17af,0x17af,0x219,0x219,
0x219,0x219,0x219,0x17ac,0x17ac,0x17ac,0x17ac,0x17ac,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x21c,
0x17b2,0x21c,0x17b2,0x17b2,0x17b2,0x17b2,0x21c,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,
0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x21c,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,0x17b2,
0x17b2,0x17b5,0x21c,0x21c,0x21c,0x21c,0x21c,0x21c,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,
0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,
0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x21f,0x21f,0x21f,0x21f,0x21f,
0x21f,0x21f,0x21f,0x21f,0x21f,0x21f,0x21f,0x21f,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,
0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x21f,0x21f,0x21f,0x21f,0x21f,
0x21f,0x21f,0x17b8,0x17b8,0x17b8,0x17b8,0x17b8,0x17b8,0x1929,0x1929,0x1929,0x1929,0x1929,0x1929,0x1929,0x1929,
0x1929,0x1929,0x1929,0x1929,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a94,0x1b3c,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1b3f,
0x1b3c,0x222,0x1a0a,0x1a94,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x17df,0x1a0a,0x1a0a,0x1a94,0x1a94,0x1a94,0x1a94,0x1a94,
0x1a94,0x1a94,0x1a94,0x1b3c,0x222,0x1a97,0x1a97,0x1a97,0x1929,0x192c,0x192c,0x192c,0x192c,0x192c,0x192c,0x192c,
0x192c,0x192c,0x192c,0x192c,0x192c,0x192c,0x1929,0x1929,0x17c4,0x17c4,0x17c4,0x17c4,0x17c1,0x17c4,0x17c4,0x17c7,
0x17ca,0x17c7,0x17c7,0x17c4,0x225,0x225,0x225,0x225,0x225,0x225,0x225,0x225,0x225,0x225,0x225,0x225,
0x225,0x225,0x225,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x1821,0x1821,0x1821,0x1821,0x1818,0x1818,0x1818,0x1812,
0x1815,0x1815,0x1815,0x1a40,0x228,0x228,0x228,0x228,0x181e,0x181e,0x181e,0x181e,0x181e,0x181e,0x181e,0x181e,
0x181e,0x181e,0x228,0x228,0x228,0x228,0x181b,0x181b,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,
0x183c,0x22b,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,
0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x183c,0x1839,0x1827,0x1827,0x1827,0x1827,
0x1827,0x1827,0x1827,0x22b,0x1827,0x1827,0x1827,0x1827,0x1827,0x1827,0x1839,0x182a,0x183c,0x183f,0x183f,0x1833,
0x1830,0x1830,0x22b,0x22b,0x22b,0x22b,0x22b,0x22b,0x22b,0x22b,0x22b,0x22b,0x1836,0x1836,0x1836,0x1836,
0x1836,0x1836,0x1836,0x1836,0x1836,0x1836,0x182d,0x182d,0x182d,0x182d,0x182d,0x182d,0x182d,0x182d,0x182d,0x182d,
0x182d,0x182d,0x182d,0x182d,0x182d,0x22b,0x22b,0x22b,0x184b,0x184e,0x1854,0x1854,0x1854,0x1854,0x1854,0x1854,
0x1854,0x1854,0x1854,0x1854,0x1854,0x1854,0x1854,0x1854,0x1842,0x1842,0x1842,0x1842,0x1842,0x1842,0x1842,0x1842,
0x1842,0x22e,0x22e,0x22e,0x22e,0x22e,0x22e,0x22e,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,
0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x231,
0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,0x1845,
0x1845,0x231,0x231,0x1845,0x1845,0x1845,0x1845,0x1845,0x1893,0x192f,0x1a9a,0x1a9d,0x1b45,0x234,0x234,0x234,
0x234,0x234,0x234,0x234,0x234,0x234,0x234,0x234,0x1b42,0x1b42,0x234,0x234,0x234,0x234,0x234,0x234,
0x234,0x234,0x234,0x234,0x234,0x234,0x234,0x234,0x1854,0x1854,0x1854,0x1854,0x1854,0x1854,0x1854,0x1854,
0x1854,0x1854,0x1854,0x1854,0x1854,0x1854,0x1854,0x1854,0x237,0x237,0x1848,0x1848,0x1848,0x1848,0x1848,0x1848,
0x1848,0x1848,0x1848,0x1848,0x1848,0x1848,0x1848,0x1848,0x237,0x1851,0x1848,0x1848,0x1848,0x1848,0x1848,0x1848,
0x1848,0x1851,0x1848,0x1848,0x1851,0x1848,0x1848,0x237,0x237,0x237,0x237,0x237,0x237,0x237,0x237,0x237,
0x1857,0x1857,0x1857,0x1857,0x1857,0x1857,0x1857,0x1857,0x1857,0x1857,0x1857,0x1857,0x1857,0x23a,0x23a,0x23a,
0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,0x23a,
0x186f,0x186f,0x1860,0x185a,0x185a,0x186f,0x185d,0x1872,0x1872,0x1872,0x1872,0x1875,0x1875,0x1869,0x1866,0x1863,
0x186c,0x186c,0x186c,0x186c,0x186c,0x186c,0x186c,0x186c,0x186c,0x186c,0x1aee,0x1869,0x23d,0x1863,0x1992,0x1a43,
0x1af1,0x1af1,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,
0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,0x23d,
0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,0x187b,
0x187b,0x187b,0x187b,0x187b,0x240,0x240,0x240,0x240,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,
0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,0x1878,
0x1878,0x1878,0x1878,0x1878,0x240,0x240,0x240,0x240,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,
0x1896,0x1896,0x1896,0x1896,0x1896,0x1a13,0x1a13,0x1a13,0x1a13,0x1a13,0x1aa0,0x1aa0,0x1aa0,0x1aa0,0x1aa0,0x1aa0,
0x243,0x243,0x243,0x243,0x243,0x243,0x243,0x243,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,
0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,
0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x1935,0x246,0x246,0x246,0x246,0x246,0x246,0x246,0x246,0x246,
0x246,0x246,0x246,0x246,0x246,0x246,0x246,0x246,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,
0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x249,
0x18d5,0x18d5,0x249,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,
0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18d5,0x18c9,0x18c9,0x18c9,0x18c9,0x18c9,0x18c9,0x249,
0x249,0x249,0x18c9,0x249,0x18c9,0x18c9,0x249,0x18c9,0x18c9,0x18c9,0x18cc,0x18c9,0x18cf,0x18cf,0x18d8,0x18c9,
0x249,0x249,0x249,0x249,0x249,0x249,0x249,0x249,0x18d2,0x18d2,0x18d2,0x18d2,0x18d2,0x18d2,0x18d2,0x18d2,
0x18d2,0x18d2,0x249,0x249,0x249,0x249,0x249,0x249,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,
0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,
0x1938,0x1938,0x1938,0x1938,0x24c,0x24c,0x24c,0x24c,0x1905,0x1908,0x1917,0x1917,0x1908,0x190b,0x1905,0x1902,
0x255,0x255,0x255,0x255,0x255,0x255,0x255,0x255,0x18f0,0x18db,0x18db,0x18db,0x18db,0x18db,0x18db,0x18ed,
0x18ed,0x18db,0x18db,0x18db,0x18f0,0x18f0,0x18f0,0x18f0,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,
0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x258,0x258,0x258,0x258,
0x258,0x258,0x258,0x258,0x258,0x258,0x258,0x258,0x1998,0x1998,0x1998,0x1998,0x1998,0x1998,0x1998,0x1998,
0x1998,0x1998,0x1998,0x1998,0x1998,0x1998,0x258,0x258,0x1aa9,0x1aa9,0x1aa9,0x1aa9,0x1b4b,0x28b,0x28b,0x28b,
0x1aa9,0x1aa9,0x1aa9,0x28b,0x28b,0x28b,0x28b,0x28b,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,
0x19aa,0x19aa,0x19aa,0x19aa,0x19a7,0x19a7,0x19a7,0x199b,0x199b,0x199b,0x199b,0x199b,0x199b,0x199b,0x199b,0x199b,
0x19a7,0x19a1,0x199e,0x19a4,0x25b,0x25b,0x25b,0x25b,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,
0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,0x19ad,
0x19ad,0x19ad,0x19ad,0x25e,0x25e,0x19ad,0x19ad,0x19ad,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x261,0x19bc,
0x19bc,0x261,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,
0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19bc,0x19b9,0x19b9,0x19b9,0x19b9,0x19b9,0x261,
0x19b0,0x19b0,0x261,0x19b9,0x19b9,0x19b0,0x19b9,0x19b3,0x19bc,0x261,0x261,0x261,0x261,0x261,0x261,0x261,
0x19c5,0x19c5,0x19c8,0x19c8,0x19bf,0x19bf,0x19bf,0x19bf,0x264,0x264,0x264,0x264,0x264,0x264,0x264,0x264,
0x19c2,0x19c2,0x19c2,0x19c2,0x19c2,0x19c2,0x19c2,0x19c2,0x19c2,0x19c2,0x264,0x264,0x264,0x264,0x264,0x264,
0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19ce,0x19cb,0x19cb,0x19cb,
0x19ce,0x19cb,0x19cb,0x19cb,0x19cb,0x267,0x267,0x267,0x267,0x267,0x267,0x267,0x267,0x267,0x267,0x267,
0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,0x19d7,
0x19d7,0x19d7,0x19d7,0x19d1,0x19d1,0x19d4,0x19d4,0x19da,0x19da,0x26a,0x26a,0x26a,0x26a,0x26a,0x26a,0x26a,
0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,0x19dd,
0x19dd,0x19dd,0x19dd,0x19dd,0x26d,0x26d,0x26d,0x26d,0x26d,0x26d,0x26d,0x26d,0x26d,0x26d,0x26d,0x26d,
0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,
0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e0,0x19e3,0x19ec,0x19e0,0x19e0,0x270,0x270,0x270,0x270,0x270,
0x19ef,0x19ef,0x19ef,0x19ef,0x19ef,0x19ef,0x19ef,0x19f2,0x273,0x273,0x273,0x273,0x273,0x273,0x273,0x273,
0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,0x19fb,
0x19fb,0x19fb,0x19f5,0x19f5,0x19f5,0x19f5,0x19f5,0x19f5,0x19f5,0x19f5,0x19f5,0x19f5,0x19f5,0x19f8,0x19f8,0x19f8,
0x19f8,0x19fe,0x19fe,0x19fe,0x19fe,0x19fe,0x276,0x276,0x276,0x276,0x276,0x276,0x1a55,0x1a55,0x1a55,0x1a55,
0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,0x1a55,
0x1a55,0x1a55,0x1a55,0x27c,0x27c,0x27c,0x27c,0x27c,0x27c,0x27c,0x27c,0x27c,0x1a64,0x1a64,0x1a64,0x1a64,
0x1a64,0x1a64,0x1a64,0x1a64,0x27f,0x27f,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,
0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a64,0x1a61,0x1a61,0x1a61,
0x1a58,0x1a58,0x1a58,0x1a58,0x27f,0x27f,0x1a58,0x1a58,0x1a61,0x1a61,0x1a61,0x1a61,0x1a5b,0x1a64,0x1a5e,0x1a64,
0x1a61,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,
0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x27f,0x1a70,0x1a70,0x1a70,0x1a70,
0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x282,0x282,0x282,0x1a67,0x1a67,0x1a67,0x1a67,
0x1a67,0x1a67,0x1a67,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a73,0x1a73,0x282,0x282,0x285,0x1a76,0x1a76,0x1a76,
0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,
0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x1a76,0x285,0x285,
0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,
0x1aa3,0x1aa3,0x1aa3,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,
0x1aa6,0x1aa6,0x1aa6,0x1aa6,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x288,0x1938,0x1938,0x1938,0x1938,
0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1aa9,0x1aa9,0x1aa9,0x1b4b,
0x1b4b,0x1b4b,0x1b4b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x1aa9,0x1aa9,0x1aa9,0x1aa9,
0x1aa9,0x1aa9,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x28b,0x28b,0x28b,
0x28b,0x28b,0x28b,0x28b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x28b,0x28b,0x28b,0x28b,0x28b,
0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x1b4b,0x28b,
0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,
0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,0x28b,
0x1a7f,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,
0x1a79,0x1a79,0x28e,0x28e,0x28e,0x28e,0x28e,0x28e,0x28e,0x28e,0x28e,0x28e,0x28e,0x28e,0x28e,0x1a7c,
0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a82,0x1a82,0x1a82,0x1a82,
0x1a88,0x1a88,0x1a88,0x1a88,0x1a88,0x1a88,0x1a88,0x1a88,0x1a88,0x1a88,0x291,0x291,0x291,0x291,0x291,0x1a85,
0x1af7,0x1af7,0x1af7,0x1af7,0x1af7,0x1af4,0x1af4,0x1af4,0x1af4,0x1af4,0x1af4,0x1af4,0x294,0x294,0x294,0x294,
0x294,0x294,0x294,0x294,0x294,0x294,0x294,0x294,0x294,0x294,0x294,0x294,0x294,0x294,0x294,0x294,
0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,0x297,0x297,0x1b12,0x297,0x297,0x1b12,0x1b12,0x1b12,0x1b12,
0x1b12,0x1b12,0x1b12,0x1b12,0x297,0x1b12,0x1b12,0x297,0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,
0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,0x1b12,0x1afa,0x1b09,0x1b09,0x1b09,0x1b09,0x1b09,0x297,0x1b09,
0x1b0c,0x297,0x297,0x1afa,0x1afa,0x1b0f,0x1b00,0x1b15,0x1b09,0x1b15,0x1b09,0x1afd,0x1b18,0x1b03,0x1b18,0x297,
0x297,0x297,0x297,0x297,0x297,0x297,0x297,0x297,0x1b06,0x1b06,0x1b06,0x1b06,0x1b06,0x1b06,0x1b06,0x1b06,
0x1b06,0x1b06,0x297,0x297,0x297,0x297,0x297,0x297,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,
0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x29a,0x29a,
0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,
0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,0x29a,
0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,
0x1b1e,0x1b1e,0x1b1e,0x2a0,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,
0x1b1e,0x1b1e,0x1b1e,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,
0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x1b21,0x1b21,0x1b21,0x1b21,0x1b21,0x1b21,0x1b21,0x1b21,
0x1b21,0x1b21,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x2a0,0x1b54,0x1b54,0x1b54,0x1b54,0x1b54,0x1b54,0x1b54,0x1b54,
0x1b54,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,
0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,0x2a3,
0x2a3,0x2a3,0x2a3,0x2a3,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x2a6,0x1b24,
0x1b24,0x1b27,0x2a6,0x2a6,0x1b2a,0x1b2a,0x2a6,0x2a6,0x2a6,0x2a6,0x2a6,0x2a6,0x2a6,0x2a6,0x2a6,0x2a6,
0x2a6,0x2a6,0x2a6,0x2a6,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,
0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,
0x2a9,0x2a9,0x2a9,0x2a9,0x1932,0x2c1,0x2c1,0x2c1,0x2c1,0x2c1,0x2c1,0x2c1,0x2c1,0x2c1,0x2c1,0x2c1,
0x2c1,0x2c1,0x2c1,0x2c1,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,
0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,
0x2a9,0x2a9,0x97b,0x97b,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x2c4,
0x2c4,0x2c4,0x2c4,0x2c4,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,0x2a9,
0x2a9,0x2a9,0x2a9,0x2a9,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,0x1a04,
0x1a04,0x1a04,0x1a04,0x1a04,0x1b33,0x1b33,0x1b33,0x1b33,0x1b33,0x1b33,0x1b33,0x1b33,0x1b33,0x1b33,0x1b33,0x1b33,
0x1b33,0x2ac,0x2ac,0x2ac,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0x12e4,
0x12e4,0x12e4,0x2af,0x2af,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,
0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0x2af,0x2af,
0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,
0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,0x2af,
0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,
0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0x1b36,0x1b36,0x1b36,0x1b36,0x1b36,0x1b36,0x1b36,0x2b2,0x2b2,
0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,
0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0x2b5,0x2b5,
0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,
0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x2b8,0x2b8,0x2b8,0x2b8,0x2b8,0x2b8,0x2b8,0x2b8,0x2b8,0x2b8,0x2b8,
0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,
0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x2bb,0x2bb,
0x17dc,0x17dc,0x2be,0x2be,0x2be,0x2be,0x2be,0x2be,0x2be,0x2be,0x2be,0x2be,0x2be,0x2be,0x2be,0x2be,
0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,
0x3cc,0x3c0,0x3c0,0x3c0,0x3c0,0x3c0,0x3c0,0x3c0,0x3c0,0x3cc,0x3cc,0x3cc,0x3cc,0x3c6,0x1158,0x133e,
0x3cf,0x945,0x948,0x3bd,0x3bd,0x1155,0x133b,0x133b,0x3d2,0x3d2,0x3d2,0x3d2,0x3d2,0x3d2,0x3d2,0x3d2,
0x1155,0x3c0,0x3c0,0x3cc,0xce1,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,
0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,
0x3cf,0x3cf,0x3c0,0x3c0,0x8d0,0x8d3,0x963,0x963,0x963,0x963,0x963,0x963,0x963,0x963,0x963,0x963,
0x3c9,0xfba,0xfb7,0x1341,0x1341,0x1341,0x1341,0x1341,0x1506,0x115b,0x115b,0xf0c,0xf0c,0xdda,0xf0c,0xf0c,
0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3d2,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,
0x3cf,0x3d2,0x3cf,0x3cf,0x3d2,0x3cf,0x3cf,0x3cf,0x3cf,0x3cf,0x133b,0x133e,0x3c3,0x3cf,0x3cc,0x3cc,
0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x1347,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,
0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x46e,0x1347,0x18ae,0x18ae,0xfd8,0x45f,0x468,
0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,
0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0x4aa,0xbc7,0xbc7,0xde6,0xde6,0x8d6,0xde9,0x1428,0x1428,0x1428,
0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,
0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,0x4ad,
0x4b3,0x4b3,0x4b3,0x1170,0x1170,0x1170,0x1170,0x1170,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,
0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,
0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x4b0,0x116d,0x116d,0x116d,0x116d,0x116d,0x116d,
0x4b6,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,
0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,0x4b3,
0x4b3,0x4b3,0x4b3,0x4b3,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,
0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,
0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4b9,0x4b9,0x4b9,0x4b9,0x4bc,0x9bd,0x1005,0x1005,0x1008,0x1005,
0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,
0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x4bf,0x4b9,0x1008,0x1005,0x1008,0x1005,0x1008,0x1005,
0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,
0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4cb,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,0x4ce,
0x6ae,0x6ae,0x6b1,0x4e9,0x6bd,0x6ba,0x6ba,0x6b7,0x513,0x513,0x4d1,0x4d1,0x4d1,0x4d1,0x4d1,0xb58,
0x6c0,0x4f5,0x6d8,0x6db,0x50a,0x6c0,0x4f8,0x4f8,0x4e9,0x504,0x504,0x6ae,0x510,0x50d,0x6b4,0x4e3,
0x4da,0x4da,0x4dd,0x4dd,0x4dd,0x4dd,0x4dd,0x4e0,0x4dd,0x4dd,0x4dd,0x4d4,0x51c,0x519,0x516,0x516,
0x6cc,0x4fe,0x4fb,0x6c9,0x6c6,0x6c3,0x6d5,0x4ec,0x6d2,0x6d2,0x501,0x504,0x6cf,0x6cf,0x501,0x504,
0x4e6,0x4e9,0x4e9,0x4e9,0x507,0x4f2,0x4ef,0xbdc,0xaf2,0xaf5,0xaef,0xaef,0xaef,0xaef,0xbd3,0xbd3,
0xbd3,0xbd3,0xbd9,0xd0e,0xd0b,0xdf5,0xdf8,0xbd6,0xdf8,0xdf8,0xdf8,0xdf8,0xdf5,0xdf8,0xdf8,0xbd0,
0x540,0x540,0x540,0x540,0x540,0x540,0x540,0x53d,0x543,0x75c,0x540,0x9c0,0x9e1,0xaf8,0xaf8,0xaf8,
0xbe2,0xbe2,0xdfe,0xdfe,0xdfe,0xdfe,0x1179,0x117c,0x117c,0x135c,0x14f4,0x151e,0x1521,0x1521,0x1734,0x18b1,
0x54f,0x54f,0x567,0x6ea,0x54c,0x6e7,0x54f,0x564,0x54c,0x6ea,0x55e,0x567,0x567,0x567,0x55e,0x55e,
0x567,0x567,0x567,0x6f3,0x54c,0x567,0x6ed,0x54c,0x55b,0x567,0x567,0x567,0x567,0x567,0x54c,0x54c,
0x552,0x6e7,0x6f0,0x54c,0x567,0x54c,0x6f6,0x54c,0x567,0x555,0x56d,0x6f9,0x567,0x567,0x558,0x55e,
0x567,0x567,0x56a,0x567,0x55e,0x561,0x561,0x561,0x561,0xb04,0xb01,0xd11,0xe07,0xbf7,0xbfa,0xbfa,
0xbf4,0xbf1,0xbf1,0xbf1,0xbf1,0xbfa,0xbf7,0xbf7,0xbf7,0xbf7,0xbee,0xbf1,0xe04,0xf18,0xf1b,0x100e,
0x117f,0x117f,0x117f,0x6ff,0x6fc,0x570,0x573,0x573,0x573,0x573,0x573,0x6fc,0x6ff,0x6ff,0x6fc,0x573,
0x705,0x705,0x705,0x705,0x705,0x705,0x705,0x705,0x705,0x705,0x705,0x705,0x57c,0x57c,0x57c,0x57c,
0x702,0x702,0x702,0x702,0x702,0x702,0x702,0x702,0x702,0x702,0x576,0x576,0x576,0x576,0x576,0x576,
0x582,0x582,0x582,0x582,0x582,0x582,0x582,0x582,0x57f,0x588,0x588,0x582,0x582,0x582,0x585,0x57f,
0x582,0x582,0x57f,0x57f,0x57f,0x57f,0x582,0x582,0x708,0x708,0x57f,0x57f,0x582,0x582,0x582,0x582,
0x582,0x582,0x582,0x582,0x582,0x582,0x582,0x582,0x582,0x585,0x585,0x585,0x582,0x582,0x70b,0x582,
0x70b,0x582,0x582,0x582,0x582,0x582,0x582,0x582,0x57f,0x582,0x57f,0x57f,0x57f,0x57f,0x57f,0x57f,
0x582,0x582,0x57f,0x708,0x57f,0x57f,0x57f,0xb0a,0xb0a,0xb0a,0xb0a,0xb0a,0xb0a,0xb0a,0xb0a,0xb0a,
0xbfd,0xbfd,0xbfd,0xbfd,0xbfd,0xbfd,0xbfd,0xbfd,0xbfd,0xbfd,0xbfd,0xbfd,0x711,0x58b,0x711,0x711,
0x58e,0x58b,0x58b,0x711,0x711,0x58e,0x58b,0x711,0x58e,0x58b,0x58b,0x711,0x58b,0x711,0x59a,0x597,
0x58b,0x711,0x58b,0x58b,0x58b,0x58b,0x711,0x58b,0x58b,0x711,0x711,0x711,0x711,0x58b,0x58b,0x711,
0x58e,0x711,0x58e,0x711,0x711,0x711,0x711,0x711,0x717,0x591,0x711,0x591,0x591,0x58b,0x58b,0x58b,
0x711,0x711,0x711,0x711,0x58b,0x58b,0x58b,0x58b,0x711,0x711,0x58b,0x58b,0x58b,0x58e,0x58b,0x58b,
0x58e,0x58b,0x58b,0x58e,0x711,0x58e,0x58b,0x58b,0x711,0x58b,0x58b,0x58b,0x58b,0x58b,0x711,0x58b,
0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x714,0x711,0x58e,0x58b,
0x711,0x711,0x711,0x711,0x58b,0x58b,0x711,0x711,0x58b,0x58e,0x714,0x714,0x58e,0x58e,0x58b,0x58b,
0x58e,0x58e,0x58b,0x58b,0x58e,0x58e,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58e,0x58e,0x711,0x711,
0x58e,0x58e,0x711,0x711,0x58e,0x58e,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,
0x58b,0x711,0x58b,0x58b,0x58b,0x711,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x711,0x58b,0x58b,
0x58b,0x58b,0x58b,0x58b,0x58e,0x58e,0x58e,0x58e,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,
0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x711,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,
0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,
0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58b,0x58e,0x58e,0x58e,0x58e,0x58b,0x58b,0x58b,0x58b,
0x58b,0x58b,0x58e,0x58e,0x58e,0x58e,0x58b,0x594,0x58b,0x58b,0xc00,0xc00,0xc00,0xc00,0xc00,0xc00,
0xc00,0xc00,0xc00,0xc00,0xc00,0xc00,0xc00,0xc00,0x59d,0xb0d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,
0x5a9,0x5a6,0x5a9,0x5a6,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x71a,0x59d,0x59d,0x59d,0x59d,0x59d,
0x59d,0x59d,0x81f,0x81f,0x59d,0x59d,0x59d,0x59d,0x5a3,0x5a3,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,
0x5a0,0x825,0x822,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,
0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,
0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0x59d,0xb0d,0xc06,0xb0d,0xb0d,0xb0d,0x5ac,0x5ac,0x5ac,0x5ac,
0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,
0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x5ac,0x723,0x723,0x723,0x723,
0x723,0x723,0x723,0x723,0x723,0x723,0x5b2,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,
0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xc6f,0xd89,0x72c,0x72c,0x72c,0x72c,
0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,
0x5b5,0x5b8,0x5b8,0x5b8,0x5b8,0x5b8,0x5b8,0x5b8,0x5b8,0x5b8,0x5b8,0x5b8,0x72c,0x72c,0x72c,0x72c,
0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x5b8,0x5b8,0x5b8,0x5b8,0x72c,0x72c,0x72c,0x72c,
0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72f,0x72f,0x72f,0x72f,
0x72f,0x72f,0x72f,0x72f,0x72f,0x72f,0x72f,0x72f,0x72f,0x72f,0x72f,0x72f,0x5bb,0x5bb,0x72f,0x72f,
0x72f,0x72f,0xc09,0xc09,0xc09,0xc09,0xc09,0xc09,0xc09,0xc09,0xc09,0xc09,0x735,0x735,0x5be,0x732,
0x732,0x732,0x732,0x732,0x732,0x732,0x5c1,0x5c1,0x5be,0x5be,0x5c4,0x5c4,0x5c4,0x5c4,0x735,0x735,
0x5c4,0x5c4,0x738,0x735,0x5be,0x5be,0x5be,0x5be,0x735,0x735,0x5c4,0x5c4,0x738,0x735,0x5be,0x5be,
0x5be,0x5be,0x735,0x735,0x732,0x5be,0x5c4,0x735,0x5be,0x5be,0x732,0x735,0x735,0x735,0x5c4,0x5c4,
0x5be,0x5be,0x5be,0x5be,0x5be,0x5be,0x5be,0x5be,0x5be,0x5be,0x5be,0x5be,0x5be,0x5be,0x735,0x732,
0x735,0x732,0x5be,0x5c4,0x5c4,0x5c4,0x5c4,0x5c4,0x5c4,0x5be,0x5be,0x732,0xb16,0xb16,0xb16,0xb16,
0xb16,0xb16,0xb16,0xb16,0xc0c,0xc0c,0xc0c,0xc0f,0xc0f,0xc8a,0xc8a,0xc0c,0x5d3,0x5d3,0x5d3,0x5d3,
0x5d0,0x74a,0x747,0x5ca,0x5ca,0x73b,0x5ca,0x5ca,0x5ca,0x5ca,0x741,0x73b,0x5ca,0x5d0,0x5ca,0x5c7,
0xd92,0xd92,0xc15,0xc15,0xe13,0xb19,0x5cd,0x5cd,0x73e,0x5d6,0x73e,0x5cd,0x5d0,0x5ca,0x5d0,0x5d0,
0x5ca,0x5ca,0x5d0,0x5ca,0x5ca,0x5ca,0x5d0,0x5ca,0x5ca,0x5ca,0x5d0,0x5d0,0x5ca,0x5ca,0x5ca,0x5ca,
0x5ca,0x5ca,0x5ca,0x5ca,0x5d0,0x5d3,0x5d3,0x5cd,0x5ca,0x5ca,0x5ca,0x5ca,0x74d,0x5ca,0x74d,0x5ca,
0x5ca,0x5ca,0x5ca,0x5ca,0x828,0x828,0x828,0x828,0x828,0x828,0x828,0x828,0x828,0x828,0x828,0x828,
0x5ca,0x5ca,0x5ca,0x5ca,0x5ca,0x5ca,0x5ca,0x5ca,0x5ca,0x5ca,0x5ca,0x5d0,0x74d,0x74a,0x5d9,0x74d,
0x73b,0x741,0x5d0,0x73b,0x744,0x73b,0x73b,0x5ca,0x73b,0x74a,0x5d9,0x74a,0xb19,0xb19,0xc18,0xc18,
0xc18,0xc18,0xc18,0xc18,0xc18,0xc18,0xc18,0xc1b,0xc18,0xc18,0xe10,0xec7,0x5dc,0x5dc,0x5dc,0x5dc,
0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,0x5dc,
0x5df,0x13e6,0x13e6,0x13e6,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x152a,0x5eb,0x5f4,0x5eb,
0x5eb,0x13e6,0x5df,0x5df,0x5f4,0x5f4,0x13e9,0x13e9,0x5f7,0x5f7,0x5e8,0x5ee,0x5e8,0x5e8,0x5ee,0x5df,
0x5ee,0x5df,0x5ee,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5ee,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,
0x13e6,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5ee,0x5ee,0x5df,0x5df,0x5df,
0x5df,0x5df,0x5df,0x5df,0x5df,0x753,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5ee,0x5df,0x5df,0x5ee,
0x5df,0x5df,0x5df,0x5df,0x13e6,0x5df,0x13e6,0x5df,0x5df,0x5df,0x5df,0x13e6,0x13e6,0x13e6,0x5df,0x12de,
0x5df,0x5df,0x5df,0x5e5,0x5e5,0x5e5,0x5e5,0x1368,0x1368,0x5df,0x5e2,0x5f1,0x5f4,0x5e8,0x5e8,0x5e8,
0xc21,0xc1e,0xc21,0xc1e,0xc21,0xc1e,0xc21,0xc1e,0xc21,0xc1e,0xc21,0xc1e,0xc21,0xc1e,0x750,0x750,
0x750,0x750,0x750,0x750,0x750,0x750,0x750,0x750,0x5df,0x5ee,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,
0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x13e6,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,
0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x5df,0x13e6,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,
0x618,0x618,0x618,0x618,0x618,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x621,0x621,0x621,0x621,
0x621,0x621,0x621,0x621,0x618,0x61e,0x60f,0x612,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,
0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,
0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x615,0x615,0x615,0x615,0x615,0x615,0x618,0x618,0x618,0x618,
0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,
0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x61b,0x621,0x61e,0x618,0x61b,0x621,0x61e,0x618,
0x61b,0x621,0x61e,0x618,0x61b,0x621,0x61e,0x618,0x61b,0x621,0x61e,0x618,0x61b,0x621,0x61e,0x618,
0x61b,0x621,0x61e,0x618,0x61b,0x621,0x61e,0x618,0x61e,0x618,0x61e,0x618,0x61e,0x618,0x61e,0x618,
0x61e,0x618,0x61e,0x618,0x61b,0x621,0x61e,0x618,0x61b,0x621,0x61e,0x618,0x61b,0x621,0x61e,0x618,
0x61b,0x621,0x61e,0x618,0x61e,0x618,0x61b,0x621,0x61e,0x618,0x61e,0x618,0x61b,0x621,0x61e,0x618,
0x61b,0x621,0x61e,0x618,0x61e,0x618,0x136b,0x136b,0x136b,0x136b,0x136b,0x136b,0x136b,0x136b,0x136b,0x136b,
0x136b,0x136b,0x136b,0x136b,0x61e,0x618,0x61e,0x618,0x61e,0x618,0x61b,0x621,0x61b,0x621,0x61e,0x618,
0x61e,0x618,0x61e,0x618,0x61e,0x618,0x61e,0x618,0x61e,0x618,0x61e,0x618,0x61b,0x61e,0x618,0x61b,
0x61e,0x618,0x61b,0x621,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,
0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x61b,0x61b,0x61b,0x61b,0x61b,
0x61b,0x61b,0x61b,0x61b,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,
0x61e,0x61e,0x61e,0x61e,0x61e,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,
0x618,0x618,0x618,0x618,0x61b,0x61b,0x618,0x61b,0x618,0x61b,0x618,0x618,0x61b,0x618,0x618,0x61b,
0x618,0x61b,0x618,0x618,0x61b,0x618,0x61b,0x61b,0x618,0x618,0x618,0x61b,0x618,0x618,0x618,0x618,
0x618,0x61b,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,
0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x618,0x61b,0x61b,0x618,0x618,0x61b,0x618,0x61b,0x618,
0x618,0x618,0x618,0x618,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,
0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,
0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x61b,0x621,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,
0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,
0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x621,0x621,0x621,0x621,0x621,0x621,0x621,0x621,
0x621,0x621,0x621,0x621,0x621,0x621,0x621,0x621,0x621,0x621,0x621,0x621,0x621,0x61e,0x61e,0x61e,
0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x61e,0x624,0x624,0x624,0x624,0x101a,0x101a,0x101a,0x152d,
0x152d,0x152d,0x152d,0x152d,0x152d,0x152d,0x173a,0x173a,0x885,0x88b,0x88b,0x897,0x897,0x888,0x87f,0x888,
0x87f,0x888,0x87f,0x888,0x87f,0x888,0x87f,0x888,0x633,0x633,0x62d,0x633,0x62d,0x633,0x62d,0x633,
0x62d,0x633,0x62d,0x630,0x636,0x633,0x62d,0x633,0x62d,0x630,0x636,0x633,0x62d,0x633,0x62d,0x630,
0x636,0x633,0x62d,0x630,0x636,0x633,0x62d,0x630,0x636,0x633,0x62d,0x633,0x62d,0x633,0x62d,0x633,
0x62d,0x633,0x62d,0x630,0x636,0x633,0x62d,0x630,0x636,0x633,0x62d,0x630,0x636,0x633,0x62d,0x630,
0x636,0x633,0x62d,0x630,0x636,0x633,0x62d,0x630,0x636,0x633,0x62d,0x630,0x636,0x633,0x62d,0x630,
0x636,0x633,0x62d,0x630,0x720,0x720,0x720,0x720,0x720,0x720,0x720,0x720,0x720,0x720,0x720,0x720,
0x720,0x720,0x720,0x720,0x720,0x720,0x720,0x720,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,
0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,
0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x71d,0x726,0x726,0x726,0x726,0x726,0x726,
0x726,0x726,0x726,0x726,0x726,0x726,0x729,0x726,0x726,0x726,0x726,0x726,0x726,0x726,0x726,0x726,
0x726,0x726,0x726,0x726,0x723,0x723,0x723,0x723,0x723,0x723,0x723,0x723,0x723,0x723,0x723,0x723,
0x723,0x723,0x723,0x723,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,
0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,0x72c,
0x72c,0x72c,0x72c,0x72c,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,
0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,0x756,
0x756,0x756,0x756,0x756,0xc78,0x8e8,0x8e2,0x8df,0x8e5,0x8dc,0x76b,0x76e,0x76e,0x76e,0x76e,0x76e,
0x76e,0x76e,0x76e,0x76e,0x8ee,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,
0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,
0x76b,0x76b,0x76b,0x76b,0x76b,0x76b,0x8eb,0x8eb,0x771,0x8fd,0x900,0x906,0x82b,0x837,0x91b,0x834,
0x8f4,0x8f1,0x8f4,0x8f1,0x8fa,0x8f7,0x8fa,0x8f7,0x8f4,0x8f1,0x831,0x906,0x8f4,0x8f1,0x8f4,0x8f1,
0x8f4,0x8f1,0x8f4,0x8f1,0x909,0x912,0x90f,0x90f,0x777,0x7b3,0x7b3,0x7b3,0x7b3,0x7b3,0x7b3,0x7ad,
0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,0x7ad,
0x7ad,0x7ad,0x7ad,0x77a,0x795,0x774,0x79b,0x79e,0x798,0x7b0,0x7b0,0x7b0,0x7b0,0x7b0,0x7b0,0x7aa,
0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,0x7aa,
0x7aa,0x7aa,0x7aa,0x77a,0x795,0x774,0x795,0xc7b,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,
0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,
0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x819,0x12d8,0x12d8,0x12d8,0x12d8,0x12d8,0x81c,
0x831,0x834,0x834,0x834,0x834,0x834,0x834,0x834,0x834,0x834,0x951,0x951,0x951,0x951,0x83a,0x83a,
0x90c,0x918,0x918,0x918,0x918,0x915,0x82e,0x903,0xb3d,0xb3d,0xb3d,0xc8d,0xcab,0xca8,0xb5b,0x8d9,
0x840,0x83d,0x840,0x843,0x83d,0x840,0x83d,0x840,0x83d,0x840,0x83d,0x83d,0x83d,0x83d,0x83d,0x83d,
0x840,0x840,0x83d,0x840,0x840,0x83d,0x840,0x840,0x83d,0x840,0x840,0x83d,0x840,0x840,0x83d,0x83d,
0xcae,0x852,0x84c,0x852,0x84c,0x852,0x84c,0x852,0x84c,0x852,0x84c,0x84c,0x84f,0x84c,0x84f,0x84c,
0x84f,0x84c,0x84f,0x84c,0x84f,0x84c,0x84f,0x84c,0x84f,0x84c,0x84f,0x84c,0x84f,0x84c,0x84f,0x84c,
0x84f,0x84c,0x84f,0x852,0x84c,0x84f,0x84c,0x84f,0x84c,0x84f,0x84c,0x84c,0x84c,0x84c,0x84c,0x84c,
0x84f,0x84f,0x84c,0x84f,0x84f,0x84c,0x84f,0x84f,0x84c,0x84f,0x84f,0x84c,0x84f,0x84f,0x84c,0x84c,
0x84c,0x84c,0x84c,0x852,0x84c,0x852,0x84c,0x852,0x84c,0x84c,0x84c,0x84c,0x84c,0x84c,0x852,0x84c,
0x84c,0x84c,0x84c,0x84c,0x84f,0x852,0x852,0x84f,0x84f,0x84f,0x84f,0x921,0x924,0x855,0x858,0xc96,
0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,
0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,
0x861,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,
0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x85e,0x86a,0x86a,0x86a,0x86a,
0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,
0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0x86a,0xd9b,0xd9b,0xeca,0x864,0x92d,0x92d,0x92d,0x92d,
0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0xd95,0xd95,0xd95,0xd95,0x86d,0x86d,0x86d,0x86d,
0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,
0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x86d,0x1ab2,0x936,0x936,0x936,0x936,
0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x870,0x870,0x870,
0x870,0x870,0x870,0xd9e,0xd9e,0xd9e,0xd9e,0x939,0x939,0x939,0x939,0x939,0x870,0x870,0x870,0x870,
0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,
0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0x870,0xd9e,0xd9e,
0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,
0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,0x873,
0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,
0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,
0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0x876,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,
0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,0xecd,
0x113d,0x113d,0x113d,0x113d,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,
0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x879,
0x879,0x879,0x879,0x879,0x879,0x879,0x87c,0x87c,0x879,0x87c,0x879,0x87c,0x87c,0x879,0x879,0x879,
0x879,0x879,0x879,0x879,0x879,0x879,0x879,0x87c,0x879,0x87c,0x879,0x87c,0x87c,0x879,0x879,0x87c,
0x87c,0x87c,0x879,0x879,0x879,0x879,0x14e5,0x14e5,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,
0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,
0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,
0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x92d,0x131d,0x131d,0x131d,0x131d,0x12bd,0x12bd,0x12bd,0x12bd,
0x12bd,0x12bd,0x12bd,0x12bd,0xd95,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,
0xc99,0xc99,0xc99,0xc99,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,
0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x933,0x930,0x933,0x930,0x930,
0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,0x930,
0x930,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,0xc99,
0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,
0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0x936,0xd9e,
0x9b7,0x999,0x999,0x999,0x999,0x993,0x999,0x999,0x9ab,0x999,0x999,0x996,0x9a2,0x9a8,0x9a8,0x9a8,
0x9a8,0x9a8,0x9ab,0x993,0x99f,0x993,0x993,0x993,0x98a,0x98a,0x993,0x993,0x993,0x993,0x993,0x993,
0x9ae,0x9ae,0x9ae,0x9ae,0x9ae,0x9ae,0x9ae,0x9ae,0x9ae,0x9ae,0x993,0x993,0x993,0x993,0x993,0x993,
0x993,0x993,0x993,0x993,0x996,0x98a,0x993,0x98a,0x993,0x98a,0x9a5,0x99c,0x9a5,0x99c,0x9b4,0x9b4,
0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,
0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,0x9c3,
0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,
0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,0x9c6,
0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,
0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,0x9c9,
0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,
0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9cc,0x9cc,
0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,
0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9cf,0x9cf,
0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,
0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,0x9d2,
0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,
0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,0x9d5,
0x9d8,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,
0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9d8,0x9db,0x9db,0x9db,
0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,
0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0x9db,0xa68,0xa68,0xfff,0xa68,0xa68,0xa68,0xa6b,0xa68,
0xfff,0xa68,0xa68,0xff6,0xa62,0xa56,0xa56,0xa56,0xa56,0xa65,0xa56,0xfe7,0xfe7,0xfe7,0xa56,0xa59,
0xa62,0xa5c,0xfed,0xff9,0xff9,0xfe7,0xfe7,0xfff,0xb61,0xb61,0xb61,0xb61,0xb61,0xb61,0xb61,0xb61,
0xb61,0xb61,0xa6e,0xa6e,0xa5f,0xa5f,0xa5f,0xa5f,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa65,0xa65,
0xa56,0xa56,0xfff,0xfff,0xfff,0xfff,0xfe7,0xfe7,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,
0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,
0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa68,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xdf2,
0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,
0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,
0xa7d,0xa7d,0xa7d,0xdf2,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,0xa7d,
0xa7d,0xa7d,0xa7d,0xa7d,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,
0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,0xa83,
0xa83,0xa83,0xa83,0xa83,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,
0xa89,0xa86,0xa8c,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0x1176,0x1176,0x1176,0x1176,0x1176,
0x1176,0x1176,0x1176,0x1176,0x1173,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,
0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,
0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa89,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,
0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,
0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xa9e,0xac2,0xac2,0xac2,0xac5,0xac5,0xac2,0xac2,0xac2,
0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xaaa,0xaaa,0xabf,0xaa1,
0xaa1,0xaa1,0xaa1,0xaa1,0xaa1,0xaa1,0xabf,0xabf,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,
0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,
0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xac2,0xae3,0xae3,0xae3,0xae3,0xae3,0xace,0xace,0xae3,
0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,
0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,
0xae3,0xae3,0xae3,0xae6,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,
0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,0xae3,
0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb10,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,
0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xb0d,0xc06,0xc06,0xc06,0xc06,0xc06,
0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,
0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,0xb1c,
0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,
0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,0xb2e,
0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,
0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,0xb34,
0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,
0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0xb40,0x13ec,0x13ec,0x13ec,0x1b2d,0x1b2d,0x1b2d,0x1b2d,0x1b2d,
0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,
0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,0xb43,
0xb43,0xb43,0x1b30,0x1b30,0x1b30,0x1b30,0x1b30,0x1b30,0x1b30,0x1b30,0x1b30,0x1b30,0xb46,0xb46,0xb46,0xb46,
0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,
0xb46,0xb49,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,
0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,0xb46,
0xb46,0xb46,0xb46,0xb46,0xb4c,0xb4c,0xc9c,0xc9c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,
0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xc9c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,0xb4c,
0xb4c,0xb4c,0xb4c,0xb4c,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,
0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,0xb70,
0xb70,0xb70,0xb70,0x1530,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xd26,0xd26,0xb76,0xb76,0xb76,0xb76,
0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,
0xb76,0xb76,0xb76,0xb76,0xb76,0xb76,0xd23,0xd23,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,
0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,
0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,
0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb79,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,
0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,
0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb7c,0xb8b,0xb8b,0xb8b,0xb8b,0xb8b,0xb82,0xb8e,0xb94,
0xb94,0xb94,0xb88,0xb88,0xb88,0xb91,0xb85,0xb85,0xb85,0xb85,0xb85,0xb7f,0xb7f,0xb7f,0xb7f,0xb7f,
0xb7f,0xb7f,0xb7f,0xb94,0xb94,0xb94,0xb94,0xb94,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,
0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,
0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb8b,0xb8b,0xb94,0xb94,0xb94,0xb88,
0xb88,0xb94,0xb94,0xb94,0xb94,0xb94,0xb94,0xb94,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,
0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb94,0xb94,
0xb94,0xb94,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb8b,
0xb8b,0xb8b,0xb8b,0xb8b,0xb8b,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,
0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,0xb88,
0xb88,0xb88,0x173d,0x173d,0xba0,0xb97,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb97,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,
0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xb97,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb97,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xba0,0xba0,0xba0,0xba0,
0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,
0xba0,0xb97,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,
0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,
0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xb9a,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,
0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,
0xba0,0xba0,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,
0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,0xba0,
0xba0,0xba0,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,
0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xb9d,0xba0,0xba0,0xba0,0xba0,
0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,
0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,0xba3,
0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,
0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,0xba9,
0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,
0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,0xbac,
0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,
0xc06,0xc06,0xc06,0xc06,0xc06,0xc06,0xc03,0xc06,0xc03,0xc03,0xc03,0xc03,0xc03,0xc03,0xc03,0xc03,
0xc03,0xc03,0xc03,0xc03,0xc03,0xc03,0xc03,0xd14,0xd17,0xe0a,0xe0a,0xe0a,0xe0a,0xe0a,0xe0a,0xe0a,
0xe0a,0xe0a,0xe0a,0xe0a,0xf24,0xf24,0xf24,0xf24,0xc18,0xc18,0xc18,0xc18,0xc18,0xc18,0xc12,0xc12,
0xc12,0xc12,0xd1a,0xd1a,0xd1a,0xd1a,0xd1a,0xd1a,0xd1d,0xd1d,0xe10,0xec4,0xe10,0xe10,0xe10,0xe10,
0xe0d,0xe10,0xe0d,0xe10,0xe10,0x1014,0x12ae,0x12ae,0xe19,0xe19,0xe19,0xe19,0xe19,0xe1f,0xe1c,0xf36,
0xf36,0xf36,0xf36,0x142e,0x1026,0x142e,0x1374,0x1374,0xc4e,0xc4e,0xc4e,0xc4e,0xc4e,0xc4e,0xc4e,0xc4e,
0xc4e,0xc4e,0xc4e,0xc4e,0xc4e,0xc4e,0xc4e,0xc4e,0xc4e,0xc4e,0xc81,0xc7e,0xc81,0xc7e,0xc81,0xc7e,
0x1137,0x1134,0x102c,0x1029,0xc51,0xc51,0xc51,0xc51,0xc51,0xc51,0xc51,0xc51,0xc51,0xc51,0xc51,0xc51,
0xc51,0xc51,0xc51,0xc51,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,
0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,
0xc54,0xc54,0xc54,0xc54,0xc57,0xc57,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,0xc54,
0xc5a,0xc5a,0xc5a,0xc60,0xc5d,0xc87,0xc84,0xc60,0xc5d,0xc60,0xc5d,0xc60,0xc5d,0xc60,0xc5d,0xc60,
0xc5d,0xc60,0xc5d,0xc60,0xc5d,0xc60,0xc5d,0xc60,0xc5d,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,
0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,
0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc60,0xc5d,0xc60,0xc5d,
0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,
0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc5a,0xc60,0xc5d,0xc5a,0xc5a,
0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc69,0xc63,0xc63,0xc63,
0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,
0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,
0xc69,0xc69,0xc69,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,
0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,0xc63,
0xc66,0xc63,0xc63,0xc63,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,
0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,0xc9f,
0xc9f,0xc9f,0xc9f,0xc9f,0xd20,0xd8f,0xe0d,0xe0d,0xe0d,0xe0d,0xe0d,0xe10,0xe0d,0xe0d,0xec4,0xec4,
0xe0d,0xe0d,0xe0d,0xe0d,0xe10,0xe10,0xf27,0x1014,0x1014,0x1014,0x1014,0x1014,0x1014,0x1014,0x1014,0x1014,
0x1014,0x12db,0x12db,0x12b1,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,
0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,0xd44,
0xd44,0xd44,0xd44,0xd44,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd4a,0xd4a,0xd4a,0xd4a,0xd4a,0xd47,
0xd5c,0xd5c,0xd5c,0xd56,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd56,
0xd5c,0xd5c,0xd5c,0xd5c,0xd50,0xd50,0xd59,0xd59,0xd59,0xd59,0xd4d,0xd4d,0xd4d,0xd4d,0xd4d,0xd53,
0xe25,0xe25,0xe25,0xe25,0xe25,0xe25,0xe25,0xe25,0xe25,0xe25,0xe25,0xe25,0xe22,0xe25,0xe25,0xe25,
0xe25,0xe25,0xe25,0xe25,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,
0xd5c,0xd5c,0xd56,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,0xd5c,
0xd5c,0xd50,0xd50,0xd50,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,
0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,0xd53,
0xd53,0xd53,0xd53,0xd53,0xd5f,0xd5f,0xd5f,0xd5f,0xd5f,0xd62,0xd62,0xd62,0xd5f,0xd5f,0xd5f,0xd5f,
0xd5f,0xd5f,0xe28,0xe28,0xe28,0xe28,0xe28,0xe28,0xf39,0xf39,0xf39,0xf39,0xf39,0xf39,0xf39,0x1140,
0x1140,0x102f,0x102f,0x102f,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,
0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,0xd65,
0xd65,0xd65,0xd65,0xd65,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,
0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,0xd6b,
0xd6b,0xd6b,0xd6b,0xd6b,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,
0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,0xd74,
0xd74,0xd74,0xd74,0xd74,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,
0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,0xd80,
0xd80,0xd80,0xd80,0xd80,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,
0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,0xd8c,
0xd8c,0xd8c,0xd8c,0xd8c,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,
0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,0xe2e,
0xe2e,0xe2e,0xe2e,0xe2e,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,
0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe31,0xe31,0xe31,0xe31,0xe31,0xe31,0xe31,
0xe31,0xe31,0xe31,0xe31,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,
0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,0xe34,
0xe34,0xe34,0xe34,0xe34,0xef4,0xef4,0xe46,0xe46,0xf3c,0xf3c,0xf3c,0xf3c,0xf3c,0xf3c,0xf3c,0x103b,
0x103b,0x103b,0x103b,0x103b,0x1038,0x1038,0x1038,0x1038,0x1038,0x1038,0x1038,0x1038,0x1038,0x1038,0x1038,0x1038,
0x1038,0x1038,0x1038,0x1038,0xe55,0xe52,0xe55,0xe52,0xe55,0xe52,0xe55,0xe52,0xe55,0xe52,0xe55,0xe52,
0xe55,0xe52,0xe55,0xe52,0xe55,0xe52,0xe55,0xe52,0xe55,0xe52,0xe55,0xe52,0xe55,0xe52,0xe55,0xe52,
0xe55,0xe52,0xe55,0xe52,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,
0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,0xe61,
0xe61,0xe61,0xe61,0xe61,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,
0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,0xe67,
0xe67,0xe67,0xe67,0xe67,0xeee,0xeee,0xeee,0xeee,0xeee,0xeee,0xeee,0xeee,0xe7f,0xe7f,0xe7f,0xe7f,
0xe7f,0xe7f,0xe7f,0xe7f,0xe7f,0xe7f,0xe7f,0xe7f,0xe7f,0xe7f,0xe7f,0xf3f,0xf3f,0xf3f,0xf3f,0x103e,
0x103e,0x103e,0x103e,0x103e,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,
0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,0xe88,
0xe88,0xe88,0xe88,0xe88,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,
0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,0xe91,
0xe91,0xe91,0xe91,0xe91,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,
0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,0xe9a,
0xe9a,0xe9a,0xe9a,0xe94,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,
0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe97,0xe9a,
0xe9a,0xe9a,0xe9a,0xe9a,0xea3,0xea3,0xea3,0xea3,0xea3,0xea3,0xea3,0xea3,0xea3,0xea3,0xea3,0xea3,
0xea3,0xea3,0xea0,0xea0,0xea0,0xea0,0xea0,0xea0,0xea0,0xea0,0xe9d,0xea6,0x104a,0x1044,0x1053,0x1041,
0xea3,0xea3,0x1041,0x1041,0xeb5,0xeb5,0xea9,0xeb5,0xeb5,0xeb5,0xeac,0xeb5,0xeb5,0xeb5,0xeb5,0xea9,
0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,0xeb5,
0xeb5,0xeb5,0xeb5,0xeb5,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,
0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,0xeb8,
0xeb8,0xeb8,0xeb8,0xeb8,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,
0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,0xed0,
0xed0,0xed0,0xed0,0xed0,0xef1,0xef1,0xef1,0xef1,0xef1,0xef1,0xef1,0xef1,0xef1,0xef1,0xef1,0xef1,
0xef1,0xef1,0xef1,0xef1,0x1149,0x1149,0x1149,0x1149,0x1149,0x1149,0x1149,0x1149,0x1149,0x1149,0x1149,0x1149,
0x1149,0x1149,0x1149,0x1149,0xf24,0xf24,0xf24,0xf21,0xf21,0xf21,0xf21,0xf21,0x1182,0x13dd,0x13dd,0x13dd,
0x13dd,0x135f,0x135f,0x135f,0x13e0,0x1362,0x1362,0x13e0,0x1524,0x1524,0x1524,0x1524,0x1527,0x1527,0x1527,0x17ee,
0x17ee,0x17ee,0x17ee,0x18b4,0xf39,0xf39,0xf39,0xf39,0x102f,0x102f,0x102f,0x102f,0x102f,0x102f,0x102f,0x102f,
0x102f,0x102f,0x102f,0x102f,0x1032,0x1032,0x1032,0x1032,0x1032,0x1032,0x1032,0x1032,0x1032,0x1032,0x1032,0x1032,
0x1032,0x1032,0x1032,0x1032,0xf5a,0xf5a,0xf5a,0xf5a,0xf6c,0xf75,0xf78,0xf75,0xf78,0xf75,0xf78,0xf75,
0xf78,0xf75,0xf78,0xf75,0xf75,0xf75,0xf78,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,
0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf75,0xf5d,0xf5a,0xf5a,0xf5a,
0xf5a,0xf5a,0xf5a,0xf6f,0xf5a,0xf6f,0xf6c,0xf6c,0xf81,0xf7e,0xf81,0xf81,0xf81,0xf7e,0xf7e,0xf81,
0xf7e,0xf81,0xf7e,0xf81,0xf7e,0x1065,0x1065,0x1065,0x11a0,0x105c,0x1065,0x105c,0xf7e,0xf81,0xf7e,0xf7e,
0x105c,0x105c,0x105c,0x105c,0x105f,0x1062,0x11a0,0x11a0,0xf84,0xf84,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,
0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x106e,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,
0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,
0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,
0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf8a,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,
0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,
0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0xf99,0x1563,0x1563,0x1563,0x1563,0x1563,
0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0x1563,0xf9f,0xf9f,0xf9f,0xf9f,
0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,
0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xf9f,0xfe7,0xfff,0xff6,0xffc,
0xffc,0xfff,0xfff,0xff6,0xff6,0xffc,0xffc,0xffc,0xffc,0xffc,0xfff,0xfff,0xfff,0xfe7,0xfe7,0xfe7,
0xfe7,0xfff,0xfff,0xfff,0xfff,0xfff,0xfff,0xfff,0xfff,0xfff,0xfff,0xfff,0xfff,0xfff,0xfe7,0xff6,
0xff9,0xfe7,0xfe7,0xffc,0xffc,0xffc,0xffc,0xffc,0xffc,0xfea,0xfff,0xffc,0xff3,0xff3,0xff3,0xff3,
0xff3,0xff3,0xff3,0xff3,0xff3,0xff3,0x116a,0x116a,0x1167,0x1164,0xff0,0xff0,0x1017,0x1017,0x1017,0x1017,
0x12db,0x12db,0x12b1,0x12b1,0x12b7,0x12ae,0x12ae,0x12ae,0x12ae,0x12b1,0x13e3,0x12b7,0x12b1,0x12b7,0x12ae,0x12b7,
0x12db,0x12ae,0x12ae,0x12ae,0x12b1,0x12b1,0x12ae,0x12ae,0x12b1,0x12ae,0x12ae,0x12b1,0x1032,0x1032,0x1032,0x1032,
0x1032,0x102f,0x102f,0x1032,0x1032,0x1032,0x1032,0x1032,0x1032,0x153c,0x153c,0x153c,0x1140,0x102f,0x102f,0x102f,
0x102f,0x12e7,0x12c0,0x12c0,0x12c0,0x12c0,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x1050,0x1050,0x104d,0x1047,
0x104d,0x1047,0x104d,0x1047,0x104d,0x1047,0x1044,0x1044,0x1044,0x1044,0x1059,0x1056,0x1044,0x119d,0x143a,0x143d,
0x143d,0x143a,0x143a,0x143a,0x143a,0x143a,0x1440,0x1440,0x1557,0x154b,0x154b,0x1548,0x1077,0x106e,0x1077,0x106e,
0x1077,0x106e,0x1077,0x106e,0x106b,0x1068,0x1068,0x1077,0x106e,0x1383,0x1380,0x1746,0x1383,0x1380,0x1449,0x1446,
0x155a,0x155a,0x1560,0x155a,0x1560,0x155a,0x1560,0x155a,0x1560,0x155a,0x1560,0x155a,0x1077,0x106e,0x1077,0x106e,
0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,
0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x106e,0x1071,0x106e,0x106e,0x106e,
0x106e,0x106e,0x106e,0x106e,0x106e,0x1077,0x106e,0x1077,0x106e,0x1077,0x1077,0x106e,0x107a,0x107a,0x1080,0x1086,
0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,
0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1086,0x1080,0x107a,0x107a,
0x107a,0x107a,0x1080,0x1080,0x107a,0x107a,0x1083,0x1452,0x144f,0x144f,0x1086,0x1086,0x107d,0x107d,0x107d,0x107d,
0x107d,0x107d,0x107d,0x107d,0x107d,0x107d,0x1455,0x1455,0x1455,0x1455,0x1455,0x1455,0x109b,0x109b,0x109b,0x109b,
0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,
0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x109b,0x10a4,0x10a4,0x10a4,0x10a4,
0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,0x10a4,
0x10a4,0x10a4,0x10a4,0x10a4,0x10a7,0x10a7,0x10a7,0x10aa,0x10a7,0x10a7,0x10ad,0x10ad,0x10b0,0x10b0,0x10b0,0x10b0,
0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,
0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b0,0x10b9,0x10b9,0x10b9,0x10b9,
0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10bc,0x10b3,0x10c2,0x10bf,0x10b9,0x10b9,0x10b9,0x10b9,
0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,
0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x10b9,0x1389,0x1386,0x10d4,0x10ce,
0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d1,0x1152,0x10c5,0x10c5,0x10c5,0x10cb,
0x1458,0x1458,0x1458,0x1458,0x1458,0x1458,0x1458,0x1458,0x10c8,0x10c8,0x10cb,0x10d7,0x10d4,0x10ce,0x10d4,0x10ce,
0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,
0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x10d4,0x10ce,0x156f,0x156c,0x156f,0x156c,
0x1572,0x1572,0x174f,0x1458,0x10e0,0x10e0,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,
0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,0x10e3,
0x10e3,0x10e3,0x10e3,0x10e3,0x10e0,0x10e0,0x10e0,0x10e0,0x10e0,0x10e0,0x10e0,0x10e0,0x10e0,0x10e0,0x10e0,0x10e0,
0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10e9,0x10e9,0x10e9,0x10e9,0x10e9,0x10ec,0x10ec,0x10ec,0x1146,0x10f5,
0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,0x1104,
0x10ef,0x10ef,0x10ef,0x10ef,0x10ef,0x10ef,0x10ef,0x10ef,0x10ef,0x10ef,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,
0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,0x10f2,
0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,
0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,0x1113,
0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,
0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,0x1125,
0x112e,0x112e,0x112e,0x112e,0x1143,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,
0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,0x112e,
0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,
0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,0x1131,
0x113d,0x113d,0x113d,0x113d,0x12e1,0x12e1,0x12e1,0x12e1,0x12e1,0x12e1,0x12e1,0x12e1,0x14e2,0x17cd,0x17cd,0x17cd,
0x17cd,0x17cd,0x17cd,0x17cd,0x17cd,0x17cd,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,0x191d,
0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,
0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11b2,0x11a9,0x11a9,0x11ac,0x11ac,0x11b2,0x11a9,0x11a9,0x11a9,0x11a9,0x11a9,
0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,
0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,0x11b8,
0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,
0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,0x11d3,
0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,
0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11df,0x11dc,0x11e2,
0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,
0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,0x11ee,
0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,0x11f4,
0x11f4,0x132f,0x11fa,0x1332,0x11fa,0x11fa,0x11fa,0x11fa,0x11f7,0x11f7,0x11f7,0x11fa,0x1752,0x1755,0x197a,0x1977,
0x11fd,0x11fd,0x11fd,0x120c,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,
0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,0x1212,
0x1212,0x1212,0x1212,0x1200,0x120c,0x120c,0x11fd,0x11fd,0x11fd,0x11fd,0x120c,0x120c,0x11fd,0x11fd,0x120c,0x120c,
0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,
0x1221,0x121e,0x121e,0x121e,0x121e,0x121e,0x121e,0x1218,0x1218,0x1218,0x121e,0x121b,0x1578,0x157b,0x157e,0x157e,
0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,
0x1224,0x1230,0x1224,0x1224,0x1224,0x1239,0x1239,0x1224,0x1224,0x1239,0x1230,0x1239,0x1239,0x1230,0x1224,0x1227,
0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,
0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,0x1230,
0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,
0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,0x124b,
0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,
0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1263,0x1260,0x1260,0x1260,
0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,
0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,0x126c,
0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,
0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,0x127b,
0x1281,0x1281,0x1290,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,
0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1296,0x1293,0x1296,0x1293,0x1293,0x1293,
0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1293,0x1296,0x1293,0x1293,0x1293,0x1293,
0x1290,0x1290,0x1290,0x1284,0x1284,0x1284,0x1284,0x1290,0x1290,0x128a,0x1287,0x128d,0x128d,0x129c,0x1299,0x1299,
0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,
0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,
0x12a5,0x12a5,0x12a5,0x12a2,0x12a2,0x12a2,0x129f,0x129f,0x129f,0x129f,0x12a2,0x129f,0x129f,0x129f,0x12a5,0x12a2,
0x12a5,0x12a2,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,
0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x129f,0x12a5,0x12a2,0x12a2,
0x129f,0x129f,0x129f,0x129f,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c9,
0x12c9,0x12c9,0x12a8,0x1980,0x13d7,0x12d2,0x13d7,0x13d7,0x13d7,0x13d7,0x13d7,0x13d7,0x13d7,0x13d7,0x13d7,0x13d7,
0x13d7,0x12d2,0x13d7,0x12d2,0x12b1,0x12b1,0x1365,0x12ae,0x1365,0x1365,0x1365,0x1365,0x12ae,0x12b4,0x12db,0x12ae,
0x12ae,0x12ae,0x12ae,0x12ae,0x12b4,0x12b7,0x12db,0x12db,0x12b7,0x12db,0x12ae,0x12b7,0x12b7,0x12ba,0x12db,0x12ae,
0x12ae,0x12db,0x12b1,0x12b1,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x12c3,0x12c3,
0x12c3,0x12c3,0x13ef,0x13ce,0x12cc,0x13ef,0x13ef,0x13ef,0x13ef,0x13ef,0x13ef,0x13ef,0x13ef,0x13ef,0x13ef,0x187e,
0x187e,0x187e,0x187e,0x187e,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x1581,
0x1581,0x1ad0,0x1ad0,0x1ad0,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,0x12c6,
0x12c6,0x12c6,0x12c6,0x12c6,0x13d7,0x13d7,0x12d2,0x13d7,0x13d7,0x13d7,0x12d2,0x13d7,0x13d7,0x13d7,0x12cc,0x12cc,
0x12cc,0x12cc,0x12cc,0x13d1,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x12cf,0x13d4,0x13d4,0x13d4,0x13d4,
0x13d4,0x13d4,0x13d4,0x12cf,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x145b,0x145b,
0x1a2e,0x1ad0,0x1ad0,0x1ad0,0x13da,0x13da,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x13d4,0x12cf,0x13d4,0x12cf,
0x12cf,0x13d4,0x13da,0x12d5,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,
0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,0x12f9,
0x12f9,0x12f9,0x12f9,0x12f9,0x1383,0x1380,0x1383,0x1380,0x1383,0x1380,0x1383,0x1380,0x1383,0x1380,0x1449,0x1560,
0x1560,0x1560,0x17fa,0x196e,0x1560,0x1560,0x1749,0x1749,0x1749,0x1743,0x1749,0x1743,0x1971,0x196e,0x1a2b,0x1a28,
0x1a2b,0x1a28,0x1a2b,0x1a28,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,
0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,0x13a7,
0x13a7,0x13a7,0x13a7,0x13a7,0x13bc,0x13ad,0x13bc,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,
0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,0x13bf,
0x13bf,0x13bf,0x13bf,0x13bf,0x13ad,0x13ad,0x13ad,0x13ad,0x13ad,0x13ad,0x13ad,0x13ad,0x13c5,0x13c5,0x13c5,0x13c5,
0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,
0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13c5,0x13cb,0x13cb,0x13cb,0x13cb,
0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,
0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13cb,0x13fb,0x13f8,0x1923,0x1923,
0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,
0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1401,0x1401,0x1404,0x1404,0x1404,0x1404,0x1404,0x1401,0x1404,0x1404,0x1404,0x1401,0x1404,0x1401,0x1404,
0x1401,0x1404,0x1404,0x1404,0x1404,0x1404,0x1407,0x1404,0x1404,0x1404,0x1404,0x1401,0x1404,0x1401,0x1401,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1401,0x1401,0x1401,0x1401,
0x1401,0x1401,0x1401,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1404,0x1404,0x1401,0x1401,0x1401,0x1401,0x1401,0x1401,0x1401,0x1401,0x1401,0x1401,0x1404,0x1404,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1401,0x1401,0x1401,0x1401,0x1401,0x1401,
0x1401,0x1401,0x1401,0x1401,0x1401,0x1401,0x158a,0x158a,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1596,0x1590,0x1590,0x1596,0x1596,0x1596,0x1596,
0x1596,0x1596,0x1596,0x1596,0x1596,0x17d0,0x17d0,0x17d0,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1596,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1596,0x17d0,0x17d0,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1407,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1590,0x1590,0x1596,0x1596,0x1590,0x1596,0x1596,0x1596,0x158d,0x158d,0x1596,0x1596,
0x1404,0x1404,0x1407,0x1407,0x1407,0x1701,0x1404,0x1407,0x1404,0x1404,0x1407,0x1599,0x1599,0x1596,0x1596,0x17d0,
0x17d0,0x17d0,0x17d0,0x17d0,0x1596,0x1596,0x1596,0x1596,0x1596,0x1596,0x1596,0x1596,0x1596,0x1596,0x1596,0x1596,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1590,0x1590,0x1596,0x1701,0x1596,0x1590,0x1596,0x17d0,0x17d0,0x17d0,0x17d3,0x17d3,0x17d3,0x17d3,0x17d3,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1596,
0x1404,0x1596,0x1407,0x1407,0x1404,0x1404,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,
0x1407,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1404,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,0x1407,
0x1407,0x1407,0x1407,0x1407,0x1407,0x1404,0x1404,0x1404,0x1407,0x1404,0x1404,0x1404,0x1404,0x1407,0x1407,0x1407,
0x1404,0x1407,0x1407,0x1407,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1407,0x1404,0x1407,0x1404,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1701,0x1404,0x1404,0x1404,0x1404,0x1596,0x1590,0x17d0,
0x145e,0x145e,0x145e,0x145e,0x158a,0x158a,0x158d,0x158d,0x158d,0x1593,0x1596,0x17d0,0x17d0,0x17d0,0x17d0,0x1758,
0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,0x1404,
0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1596,0x1596,0x1590,0x1590,0x1596,0x1599,0x1599,0x1596,0x1596,
0x1596,0x1596,0x1887,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1596,0x1590,0x1596,0x1590,0x1590,0x1590,0x1590,
0x1596,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1596,0x1590,0x1590,0x1590,0x1596,0x158d,0x158d,0x158d,0x158d,
0x158d,0x158d,0x1596,0x1404,0x1404,0x1404,0x1404,0x1404,0x14e8,0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,
0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,0x14e8,0x140a,0x140a,0x140a,0x14e8,0x140a,0x14e8,
0x140a,0x14e8,0x140a,0x14e8,0x140a,0x140a,0x140a,0x14e8,0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,0x14e8,0x14e8,
0x140a,0x140a,0x140a,0x140a,0x14e8,0x140a,0x14e8,0x14e8,0x140a,0x140a,0x140a,0x140a,0x14e8,0x140a,0x140a,0x140a,
0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,0x140a,0x1707,0x1707,0x17d6,0x17d6,0x140d,0x140d,0x140d,
0x140a,0x140a,0x140a,0x140d,0x140d,0x140d,0x140d,0x140d,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,
0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,
0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,
0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1413,0x1410,0x1410,0x1410,0x1410,
0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1413,0x1413,0x1413,0x1410,
0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1410,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,
0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,
0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1416,0x1803,0x1803,0x1800,0x175b,0x1464,0x1464,0x1464,0x1464,
0x1464,0x1464,0x1461,0x1461,0x1461,0x1461,0x1461,0x1461,0x1464,0x1464,0x1464,0x1464,0x1464,0x1464,0x1464,0x1464,
0x1464,0x1464,0x1464,0x1464,0x1464,0x1464,0x1464,0x15a2,0x1470,0x1470,0x1470,0x1482,0x1482,0x1482,0x1482,0x1482,
0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,
0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x1482,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,
0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,
0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x149d,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,
0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,
0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x14a3,0x1a34,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,
0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,
0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14a6,0x14ac,0x14ac,0x14b8,0x14be,0x14be,0x14be,0x14be,0x14be,
0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,
0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14be,0x14b8,0x14b8,0x14b8,0x14ac,0x14ac,
0x14ac,0x14ac,0x14ac,0x14ac,0x14ac,0x14ac,0x14ac,0x14b8,0x14bb,0x14be,0x14c1,0x14c1,0x14be,0x14c4,0x14c4,0x14af,
0x14b2,0x1764,0x1767,0x1767,0x1767,0x15ab,0x1adc,0x1ad9,0x14b5,0x14b5,0x14b5,0x14b5,0x14b5,0x14b5,0x14b5,0x14b5,
0x14b5,0x14b5,0x15a8,0x176d,0x1770,0x176a,0x1773,0x1773,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,
0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,
0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x14df,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,
0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,
0x153c,0x153c,0x1956,0x1956,0x1956,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,
0x153c,0x1a22,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x153c,0x18ba,0x1956,0x1956,0x1956,0x1956,0x1956,
0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1590,0x1590,0x1596,0x1596,0x1596,0x1590,0x1590,0x1590,
0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1596,0x1596,0x1596,0x158d,0x158d,0x158d,0x158d,
0x158d,0x158d,0x158d,0x158d,0x1596,0x1596,0x1596,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1596,
0x1590,0x1590,0x1596,0x1596,0x1596,0x1596,0x1590,0x1590,0x1599,0x1590,0x1590,0x1590,0x1590,0x1704,0x1704,0x1590,
0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1884,0x1596,0x1590,0x1590,0x1596,0x1590,0x1590,0x1590,
0x1590,0x1590,0x1590,0x1590,0x1590,0x1596,0x1596,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,0x1590,
0x1596,0x1590,0x1590,0x1590,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,
0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,0x15ba,
0x15ba,0x15ba,0x15ba,0x15ba,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,
0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,0x15cc,
0x15cc,0x15cc,0x15cc,0x15cc,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,
0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,0x15d2,
0x15d2,0x15d2,0x15d2,0x15d2,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,
0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,0x15d5,
0x15d5,0x15d5,0x15d5,0x15d5,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,
0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,0x1614,
0x1614,0x1614,0x1614,0x1605,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,
0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x161d,0x1617,
0x1620,0x1620,0x1620,0x1620,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,
0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,0x1623,
0x1623,0x1623,0x1623,0x1623,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x1635,0x163e,0x163e,0x163e,
0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,0x163e,
0x163e,0x163e,0x163e,0x163e,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,
0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,0x1647,
0x1647,0x1647,0x1647,0x1647,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,
0x1659,0x1659,0x1659,0x1659,0x1656,0x1656,0x1656,0x164a,0x164a,0x164a,0x164a,0x164a,0x164a,0x164a,0x164a,0x1656,
0x1656,0x164a,0x1656,0x164d,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,
0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,0x1659,
0x1659,0x1659,0x1659,0x1659,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,
0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,0x167d,
0x167d,0x167a,0x167a,0x167a,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,
0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x1686,0x168c,0x168c,0x168c,0x1689,0x1689,0x1689,
0x1686,0x1686,0x1686,0x1686,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,
0x169b,0x169b,0x169b,0x169b,0x168f,0x168f,0x168f,0x168f,0x168f,0x168f,0x168f,0x16a1,0x16a1,0x1695,0x1692,0x1692,
0x1692,0x1692,0x1692,0x1692,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,
0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,0x169b,
0x169b,0x169b,0x169b,0x169b,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,
0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a7,0x16a4,0x16a4,0x16a4,0x16a4,0x16a4,
0x16a4,0x16a4,0x16a4,0x16a4,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,
0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,0x16aa,
0x16aa,0x16aa,0x16aa,0x16aa,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,
0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,0x16ce,
0x16ce,0x16ce,0x16ce,0x16ce,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,
0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,0x16d7,
0x16d7,0x16d7,0x16d7,0x16d7,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,
0x16ef,0x16ef,0x16ef,0x16ef,0x16da,0x16e9,0x16e9,0x16da,0x16da,0x16da,0x16da,0x16da,0x16da,0x16e9,0x16da,0x16ec,
0x16ec,0x16da,0x16ec,0x16da,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,
0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,0x16ef,
0x16ef,0x16ef,0x16ef,0x16ef,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,
0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,0x16f8,
0x16f8,0x16f8,0x16f8,0x16f8,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,
0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,0x16fe,
0x16fe,0x16fe,0x16fe,0x16fe,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,
0x1740,0x1740,0x1740,0x1740,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,0x1956,
0x1956,0x1956,0x1956,0x1a22,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,
0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,0x1761,
0x1761,0x1761,0x1761,0x1761,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,
0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,
0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x17a3,0x17a0,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,
0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x179d,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,
0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,
0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a6,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,
0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,
0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17a9,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,
0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,
0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17bb,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,
0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,
0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17be,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,
0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,
0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c4,0x17c4,0x17c4,0x17c4,0x17c1,
0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c1,0x17c4,0x17c4,0x17c4,
0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c1,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,
0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,
0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17c4,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,
0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,
0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x17dc,0x18c6,0x18c6,0x18c6,0x18c6,0x18c6,0x18c6,0x18c6,0x18c6,
0x18c6,0x18c6,0x18c6,0x18c6,0x1b3f,0x1a94,0x1a94,0x1a97,0x17df,0x17df,0x17df,0x17df,0x17df,0x17df,0x17df,0x17df,
0x17e2,0x1890,0x1890,0x1890,0x1890,0x1890,0x1890,0x192c,0x17df,0x17df,0x17df,0x17df,0x17df,0x188d,0x188d,0x188d,
0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x1929,0x1929,0x1929,0x1929,0x1929,0x1929,
0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x1890,0x188d,
0x1929,0x1929,0x1929,0x1929,0x1929,0x1929,0x1929,0x1929,0x1890,0x192c,0x192c,0x1890,0x1890,0x1890,0x1890,0x1890,
0x1890,0x1890,0x188d,0x180f,0x1890,0x1890,0x1890,0x1a94,0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x180f,0x188d,
0x188d,0x188d,0x188d,0x188d,0x1929,0x1a0a,0x1a0a,0x1a0a,0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,
0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x188d,0x1929,0x1824,0x1824,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,
0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,
0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1821,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,
0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,
0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1824,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,
0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x186f,0x186f,0x186f,
0x185a,0x185a,0x185a,0x185a,0x185a,0x185a,0x185a,0x185a,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,
0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,
0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1872,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,
0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,
0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1896,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,
0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,
0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1899,0x1b48,0x1b48,0x1b48,0x1b48,0x1b48,
0x1b48,0x1b48,0x1b48,0x1b48,0x1b48,0x1b48,0x1b48,0x1b48,0x18f0,0x18f0,0x18f0,0x18f0,0x1a46,0x1a46,0x18f3,0x18f3,
0x18f3,0x18f3,0x18db,0x18db,0x18db,0x18db,0x18db,0x18db,0x18db,0x18db,0x18db,0x18db,0x18db,0x18db,0x18db,0x18ed,
0x18de,0x18e1,0x18e4,0x18f6,0x18f6,0x1995,0x18e7,0x18e7,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,
0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,
0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x18f0,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,
0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x18fc,0x1902,0x18ff,0x18ff,0x18ff,
0x18ff,0x190e,0x1914,0x18ff,0x18ff,0x18ff,0x18ff,0x190b,0x1911,0x18ff,0x18ff,0x18ff,0x18ff,0x18ff,0x18ff,0x18ff,
0x18ff,0x18ff,0x18ff,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,
0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1911,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,
0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,
0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1923,0x1929,0x1929,0x1929,0x1929,0x1929,0x1929,0x1929,0x1a0a,
0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,
0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1a0a,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,
0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,
0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1932,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,
0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,
0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x1938,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,
0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,
0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19aa,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,
0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,
0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19c5,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,
0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,
0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19cb,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,
0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,
0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e6,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,
0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,
0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19e9,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,
0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,
0x19f2,0x19f2,0x19f2,0x19f2,0x19f2,0x19ef,0x19ef,0x19ef,0x1a0a,0x1a0a,0x1a0a,0x1b3c,0x1b3c,0x1a94,0x1a94,0x1a94,
0x1a94,0x1a94,0x1a94,0x1b3c,0x1b3c,0x1b3c,0x1a94,0x1a94,0x1a0d,0x1a0d,0x1a0d,0x1a0d,0x1a0a,0x1a10,0x1a10,0x1a0a,
0x1a10,0x1a10,0x1a94,0x1a97,0x1a94,0x1a94,0x1a94,0x1a94,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,
0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,
0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a49,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,
0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,
0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a70,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,
0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1aaf,0x1aaf,0x1a79,0x1aaf,0x1a79,0x1a79,0x1a79,0x1a79,
0x1a79,0x1a79,0x1a79,0x1a79,0x1a79,0x1a7f,0x1a7f,0x1a7f,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,
0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,
0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1a8b,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,
0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,
0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b1e,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,
0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,
0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b2a,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,
0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,
0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b4e,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,
0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,
0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0x1b51,0,0,0,0
};

static const UTrie2 propsVectorsTrie={
    propsVectorsTrie_index,
    propsVectorsTrie_index+5024,
    NULL,
    5024,
    26204,
    0xa40,
    0x1420,
    0x0,
    0x0,
    0x110000,
    0x79f8,
    NULL, 0, FALSE, FALSE, 0, NULL
};

static const uint32_t propsVectors[6999]={
0x67,0,0,0x67,0,0x4e00000,0x67,0x80000,0x20,0x867,0,0,0xa67,0,0,0xb67,
0,0,0xc67,0,0,0xd67,0,0,0xe67,0,0,0x1067,0,0,0x1167,0,
0,0x1267,0,0,0x1367,0,0,0x1467,0,0,0x1567,0,0,0x1667,0,0,
0x1767,0,0,0x1867,0,0,0x1967,0,0,0x1a67,0,0,0x1b67,0,0,0x1d67,
0,0,0x1f67,0,0,0x2067,0,0,0x2267,0,0,0x2367,0,0,0x2467,0,
0,0x2567,0,0,0x2767,0,0,0x2867,0x80000,0x20,0x2967,0,0,0x2a67,0,0x1600000,
0x2b67,0,0,0x2d67,0,0,0x3167,0x20000000,0,0x3267,0x20000000,0,0x3a67,0,0,0x3b67,
0,0,0x3c67,0,0,0x3e67,0,0,0x4067,0,0,0x4167,0,0,0x4467,0,
0,0x4867,0,0,0x4967,0,0,0x4a67,0,0,0x5067,0,0,0x5167,0,0,
0x5467,0,0,0x5567,0,0,0x5667,0x80000,0x20,0x5767,0,0,0x5867,0,0,0x5967,
0,0,0x5b67,0,0,0x5c67,0,0,0x5d67,0,0,0x6067,0x80000,0x20,0x6267,0,
0,0x6367,0,0,0x6467,0,0,0x6567,0,0,0x6f67,0,0,0x7067,0,0,
0x7367,0x20000000,0,0x7567,0,0,0x7667,0,0,0x7767,0,0,0x7867,0,0,0x7a67,
0,0,0x7b67,0,0,0x7c67,0,0,0x7e67,0,0,0x7f67,0,0,0x8167,0,
0,0x8267,0,0,0x8367,0,0,0x8467,0,0,0x8567,0,0,0x8667,0,0,
0x8767,0,0,0x8867,0,0,0x8967,0,0,0x8b67,0,0,0x8c67,0,0,0x8e67,
0x20000000,0,0x8f67,0,0,0x9067,0,0,0x9167,0,0,0x9267,0,0,0x9367,0,
0,0x9567,0,0,0x9667,0,0,0x9767,0,0,0x9867,0,0,0x9967,0,0,
0x9a67,0,0,0x9c67,0,0,0x9f67,0,0,0xa167,0,0,0xa367,0,0,0xa467,
0,0,0xa567,0,0,0xa667,0,0,0xa767,0,0,0xa867,0,0,0xa967,0,
0,0xaa67,0,0x4e00000,0xab67,0,0x4e00000,0xac67,0,0,0xad67,0,0,0xae67,0,0,
0xaf67,0,0,0xb167,0,0,0xb267,0,0,0xb467,0,0,0xb567,0,0,0xb767,
0,0,0xb867,0,0,0xb967,0,0,0xba67,0,0,0xbc67,0,0,0xbd67,0,
0,0xbe67,0,0,0xbf67,0,0,0xc067,0,0,0xc167,0,0,0xc267,0,0,
0xc367,0,0x4e00000,0xc467,0,0x4e00000,0xc667,0,0,0xc767,0,0,0xc867,0,0,0xc967,
0,0,0xca67,0,0,0xcc67,0,0x4e00000,0xcf67,0,0x4e00000,0xd067,0,0x4e00000,0xd267,0,
0,0xd367,0,0,0xd467,0,0,0xd567,0,0,0xd667,0,0,0xd867,0,0,
0xda67,0,0,0xdb67,0,0,0xdc67,0,0,0xdd67,0,0,0xde67,0,0,0xdf67,
0,0,0xe067,0,0,0xe167,0,0,0xe267,0,0,0xe367,0,0x4e00000,0xe467,0,
0,0xe567,0,0,0xe667,0,0,0xe767,0,0,0xe867,0,0,0xe967,0,0,
0xea67,0,0,0xeb67,0,0,0xec67,0,0,0xed67,0,0,0xee67,0,0,0xef67,
0,0,0xf167,0,0,0xf367,0,0,0xf567,0,0,0xf667,0,0,0xf767,0,
0,0xf867,0,0,0xf967,0,0,0xfa67,0,0x4e00000,0xfb67,0,0,0xfc67,0,0,
0xfd67,0,0,0xfe67,0,0,0x10167,0,0,0x10267,0,0,0x10367,0,0,0x10467,
0,0,0x10567,0,0x4e00000,0x10667,0,0,0x10767,0,0,0x10867,0,0,0x10967,0,
0,0x10a67,0,0,0x10b67,0,0,0x10c67,0,0,0x10d67,0,0,0x10e67,0,0,
0x10f67,0,0,0x11067,0,0,0x11367,0,0,0x11467,0,0,0x11567,0,0,0x11667,
0,0,0x11767,0,0,0x11867,0,0,0x11967,0,0x4e00000,0x11a67,0,0,0x11b67,0,
0,0x11c67,0,0,0x11d67,0,0,0x11e67,0,0,0x11f67,0,0,0x12067,0,0,
0x12167,0,0,0x12267,0,0,0x12367,0,0,0x12467,0,0,0x12567,0,0,0x12667,
0,0,0x12767,0,0,0x12867,0,0,0x12967,0,0,0x12a67,0,0x4e00000,0x12b67,0,
0,0x12c67,0,0,0x12d67,0,0,0x12f67,0,0,0x13067,0,0,0x13167,0,0,
0x13267,0,0,0x13367,0,0,0x13467,0,0,0xa0067,0,0xe00000,0xa4767,0,0xe00000,0xa4f67,
0,0xe00000,0xa5e67,0,0xe00000,0xa5f67,0,0xe00000,0xac567,0,0xe00000,0xad167,0,0xe00000,0xb0067,0,
0xe00000,0xb1267,0,0xe00000,0xb2e67,0,0xe00000,0x11000100,0,0x900020,0x11000100,0x40000001,0x440020,0x11000100,0x40000001,0x643020,
0x11000100,0x40000001,0xa5a040,0x11000100,0x40000001,0x116a8a0,0x11000200,0,0x900020,0x11000200,0x4000001,0xc4000b,0x11000200,0x7c00100,0x220402,0x11000200,
0x24000000,0x14200000,0x11000200,0x24000008,0x1710000,0x11000200,0x40000001,0x1d3b020,0x11000219,0x7c00100,0x220401,0x11000219,0x7c00100,0x250401,0x11000319,0x7c00100,
0x220401,0x11000319,0x7c00100,0x220402,0x11000319,0x7c00100,0x250400,0x11000319,0x7c00100,0x250401,0x11000419,0x7c00100,0x220400,0x11000419,0x7c00100,0x220401,
0x11000419,0x7c00100,0x220402,0x11000419,0x7c00100,0x230400,0x11000419,0x7c00100,0x250400,0x11000419,0x7c00100,0x250401,0x11000419,0x7c00100,0x250402,0x11000519,
0x7c00100,0x220400,0x11000519,0x7c00100,0x230400,0x11000600,0x4000400,0x200002,0x11000600,0x4000400,0x200400,0x11000600,0x7c00500,0x220400,0x11000600,0x7c00500,
0x230400,0x11000600,0x7c00500,0x530400,0x11000600,0x7c00d00,0x230400,0x11000619,0x7c00500,0x22040f,0x11000800,0x4000010,0x1001401,0x11000800,0x4000400,0x200001,
0x11000800,0x6800010,0x201001,0x11000800,0x7c00500,0x230401,0x11000807,0x7c00100,0x220400,0x11000807,0x7c00100,0x250400,0x1100080e,0x4000400,0x200000,0x1100080e,
0x4000400,0x200002,0x1100080e,0x7000500,0x220402,0x1100080e,0x7c00100,0x220400,0x1100080e,0x7c00100,0x220401,0x1100080e,0x7c00100,0x220402,0x1100080e,0x7c00100,
0x250400,0x1100080e,0x7c00100,0x250401,0x1100080e,0x7c00120,0x220402,0x1100080e,0x7c00120,0x250402,0x11000908,0x4000000,0x200000,0x11000908,0x7c00100,0x220400,
0x11000908,0x7c00100,0x220401,0x11000908,0x7c00100,0x250400,0x11000908,0x7c00100,0x250401,0x11000a03,0x4000000,0x200400,0x11000a03,0x4000000,0x201000,0x11000a03,
0x4000000,0x270000,0x11000a03,0x7c00100,0x220400,0x11000a03,0x7c00100,0x220402,0x11000a03,0x7c00100,0x250400,0x11000a03,0x7c00500,0x230400,0x11000a03,0xc000010,
0x1049400,0x11000b13,0x2802500,0x962460,0x11000b13,0x4000000,0x200000,0x11000b13,0x4000000,0x201000,0x11000b13,0x4000000,0x230400,0x11000b13,0x4000002,0x400000,
0x11000b13,0x4000010,0x200000,0x11000b13,0x7c00100,0x2633800,0x11000c00,0x80000000,0x218960,0x11000c02,0x2802100,0x962460,0x11000c02,0x2802400,0x962460,0x11000c02,
0x4000000,0x200000,0x11000c02,0x4000000,0x1329400,0x11000c02,0x4000000,0x1329800,0x11000c02,0x4000000,0x1500000,0x11000c02,0x6800000,0x1329800,0x11000c02,0x7c00100,
0x230400,0x11000c02,0x7c00100,0x230401,0x11000c02,0x7c00100,0x230402,0x11000c02,0x7c00500,0x230400,0x11000c02,0x7d00100,0x230400,0x11000f01,0x2802400,0x962460,
0x11000f0a,0x2802100,0x962460,0x11000f0a,0x2802400,0x962460,0x11000f0a,0x2806400,0x962460,0x11000f0a,0x4000000,0x200000,0x11000f0a,0x6800100,0x962540,0x11000f0a,
0x7c00100,0x230400,0x11000f0a,0x7c00100,0x230401,0x11001004,0x2802100,0x962460,0x11001004,0x2802400,0x962460,0x11001004,0x2806400,0x962460,0x11001004,0x4000000,
0x200000,0x11001004,0x4000000,0x1500000,0x11001004,0x6800100,0x962540,0x11001004,0x6800100,0x962541,0x11001004,0x7c00100,0x230400,0x11001004,0x7c00100,0x230401,
0x11001110,0x2802100,0x962460,0x11001110,0x2802400,0x962460,0x11001110,0x2806400,0x962460,0x11001110,0x6800100,0x962540,0x11001110,0x7c00100,0x230400,0x11001110,
0x7c00100,0x230401,0x1100120f,0x2802100,0x962460,0x1100120f,0x2802400,0x962460,0x1100120f,0x2806400,0x962460,0x1100120f,0x6800100,0x962540,0x1100120f,0x7c00100,
0x230400,0x1100131f,0x2802100,0x962460,0x1100131f,0x2802400,0x962460,0x1100131f,0x2806400,0x962460,0x1100131f,0x4000000,0x200000,0x1100131f,0x6800000,0x1329800,
0x1100131f,0x6800100,0x962540,0x1100131f,0x6800100,0x962541,0x1100131f,0x7c00100,0x230400,0x1100131f,0x7c00100,0x230401,0x11001423,0x2802100,0x962460,0x11001423,
0x2806400,0x962460,0x11001423,0x6800100,0x962540,0x11001423,0x6800100,0x962541,0x11001423,0x7c00100,0x230400,0x11001423,0x7c00100,0x230401,0x11001524,0x2802100,
0x962460,0x11001524,0x2802100,0x962461,0x11001524,0x2806400,0x962460,0x11001524,0x6800000,0x1329800,0x11001524,0x6800100,0x962540,0x11001524,0x7c00100,0x230400,
0x11001615,0x2802100,0x962460,0x11001615,0x2806400,0x962460,0x11001615,0x6800100,0x962540,0x11001615,0x6800100,0x962541,0x11001615,0x7c00100,0x230400,0x1100171a,
0x2802100,0x962460,0x1100171a,0x2806400,0x962460,0x1100171a,0x6800000,0x1329800,0x1100171a,0x6800100,0x962540,0x1100171a,0x6800100,0x962541,0x1100171a,0x7c00100,
0x230400,0x11001900,0x4000000,0x1600000,0x11001926,0x2802100,0x1862460,0x11001926,0x2802400,0x1862460,0x11001926,0x2806100,0x1862460,0x11001926,0x4000000,0x200000,
0x11001926,0x4000010,0x400000,0x11001926,0x6800000,0x1329800,0x11001926,0x7800100,0x1830142,0x11001926,0x7c00100,0x1830000,0x11001926,0x7c00900,0x1830000,0x11001926,
0x7e00100,0x1830000,0x11001a18,0x2802100,0x1862460,0x11001a18,0x2802400,0x1862460,0x11001a18,0x6800000,0x1329800,0x11001a18,0x7800100,0x1830142,0x11001a18,0x7c00100,
0x1830000,0x11001a18,0x7c00100,0x1830002,0x11001a18,0x7c00900,0x1830000,0x11001a18,0x7e00100,0x1830000,0x11001d0c,0x7c00100,0x230400,0x11001d0c,0x7c00100,0x250400,
0x11001e12,0x7c00100,0x2230500,0x11001e12,0x7c00100,0x2330520,0x11001e12,0x7c80100,0x2330520,0x11002619,0x7c00100,0x220401,0x11002619,0x7c00100,0x220402,0x11002619,
0x7c00100,0x250401,0x1100270e,0x4000400,0x200001,0x1100270e,0x4000400,0x200002,0x1100270e,0x4000400,0x500001,0x1100270e,0x7c00100,0x220401,0x1100270e,0x7c00100,
0x250401,0x11002800,0x80000,0x918820,0x11002800,0x80000,0x1c18020,0x11002800,0x180000,0x918820,0x11002800,0x4000001,0x445801,0x11002800,0x4000001,0x445802,
0x11002800,0x4000001,0xc4000b,0x11002800,0x6800000,0x201c00,0x11002800,0x6800020,0x201c00,0x11002800,0x24000000,0x200000,0x11002800,0x24000000,0x200002,0x11002800,
0x24000000,0x810000,0x11002800,0x24000000,0x1410000,0x11002800,0x24000000,0x1500000,0x11002800,0x24000000,0x1500002,0x11002800,0x24000002,0x400000,0x11002800,0x24000006,
0xc0000b,0x11002800,0x24000008,0x1410000,0x11002800,0x24000008,0x1710000,0x11002800,0x24000020,0x1001400,0x11002800,0x24000020,0x1500002,0x11002800,0x2c000010,0x1248000,
0x11002800,0x2c000010,0x15248002,0x11002800,0x40000001,0x63b020,0x11002800,0x40080000,0x918820,0x11002801,0x80000,0xaa65620,0x11002801,0x82000,0x962460,0x11002900,
0x4000000,0x20000e,0x11002900,0x4000000,0x20000f,0x11002900,0x4000020,0x20000e,0x11002900,0x4000020,0x20000f,0x11002900,0x4000020,0x81000e,0x11002900,0x4000020,
0x81000f,0x11002900,0x4000020,0x141000e,0x11002900,0x4000020,0x141000f,0x11002900,0x4000022,0x20000e,0x11002900,0x4000022,0x20000f,0x11002a00,0x4000000,0x1500000,
0x11002a00,0x4000000,0x1600000,0x11002a00,0x4000000,0x1600002,0x11002b01,0x2000,0x962460,0x11002b01,0x2802020,0x962460,0x11002c00,0x4000000,0x200000,0x11002c00,
0x4000000,0x200002,0x11002c00,0x4000000,0x20000f,0x11002c00,0x4000020,0x200000,0x11002c00,0x7c00000,0x200000,0x11002c00,0x7c00020,0x200000,0x11002c00,0x7c00120,
0x220405,0x11002c00,0x7c00120,0x230402,0x11002c00,0x7c00120,0x250402,0x11002c00,0x7c00120,0x250405,0x11002c19,0x7c00100,0x250400,0x11002c19,0x7c00100,0x250401,
0x11002d00,0x4000000,0x100006,0x11002d00,0x4000000,0x200006,0x11002d19,0x7c00100,0x220402,0x11002d19,0x7c00100,0x230400,0x11002d19,0x7c00100,0x250402,0x11002e00,
0x24000000,0x200000,0x11002e00,0x24000020,0x200000,0x11002e00,0x24000020,0x200001,0x11002e00,0x24000020,0x14200000,0x11002f00,0x24000020,0x200000,0x11002f00,0x24000020,
0x200001,0x11002f00,0x24000020,0x200002,0x11002f00,0x24000020,0xf00000,0x11002f00,0x24000020,0x1600000,0x11002f00,0x24000022,0x1600000,0x11003000,0x24000000,0x200000,
0x11003000,0x24000000,0x14200000,0x11003000,0x24000020,0x200000,0x11003000,0x24000020,0x810000,0x11003000,0x24000020,0x1410000,0x11003100,0x24000000,0x200000,0x11003200,
0x24000000,0x200000,0x11003300,0x4000000,0x100003,0x11003400,0x24000000,0x100000,0x11003400,0x24000000,0x200000,0x11003500,0x24000000,0x200000,0x11003600,0x24000000,
0x200000,0x11003600,0x24000000,0x14200000,0x11003600,0x24000020,0x200000,0x11003700,0x24000000,0x200000,0x11003700,0x24000000,0x4200000,0x11003700,0x24000000,0x4e00000,
0x11003700,0x24000000,0x14200000,0x11003700,0x24000000,0x14e00000,0x11003700,0x24000000,0x96800000,0x11003700,0x24000020,0x4200000,0x11003800,0x4000000,0x100000,0x11003800,
0x24000000,0x200000,0x11003800,0x24000000,0xb00000,0x11003800,0x24000000,0x1710000,0x11003800,0x24000000,0x4200000,0x11003800,0x24000000,0x4e00000,0x11003800,0x24000000,
0x14200000,0x11003800,0x24000000,0x14b00000,0x11003800,0x24000000,0x14e00000,0x11003800,0x24000000,0x96800000,0x11005003,0x7c00100,0x220402,0x11005013,0x2802500,0x962460,
0x11005013,0x4000020,0x200005,0x11005013,0x7c00100,0x2633801,0x11005013,0x7c00100,0x2633802,0x11005013,0x7c00100,0x2633805,0x11005019,0x7c00100,0x220402,0x11005100,
0x24000000,0x810000,0x11005100,0x24000000,0x1410000,0x11005102,0x7000100,0x230408,0x11005102,0x7c00100,0x230404,0x11005102,0x7c00100,0x230407,0x11005102,0x7c00100,
0x230408,0x11005102,0x7c00100,0x230409,0x11005201,0x2802400,0x962460,0x11005500,0x80000,0x1e18820,0x11005502,0x7000100,0x230408,0x11005502,0x7c00100,0x230404,
0x11005502,0x7c00100,0x230407,0x11005502,0x7c00100,0x230408,0x11005502,0x7c00100,0x230409,0x11005667,0x1000,0,0x11020200,0x80004,0x418820,0x11020200,
0x4000000,0x100006,0x11020200,0x4000000,0x10000f,0x11020200,0x4000400,0x100002,0x11020200,0x4000400,0x500002,0x11020200,0x6800c00,0x101000,0x11020200,0x24000000,
0x100000,0x11020200,0x24000000,0x1400000,0x11020200,0x24000000,0x1500000,0x11020200,0x24000000,0x1600000,0x11020200,0x24000000,0x14200000,0x11020200,0x24000020,0x100000,
0x11020200,0x24000020,0x1600000,0x11020219,0x7c00100,0x12040f,0x11020219,0x7c00100,0x220400,0x11020219,0x7c00100,0x220401,0x11020219,0x7c00100,0x250400,0x11020319,
0x7c00100,0x220400,0x11020319,0x7c00100,0x220401,0x11020319,0x7c00100,0x220402,0x11020319,0x7c00100,0x250400,0x11020319,0x7c00100,0x250402,0x11020319,0x7d00100,
0x220402,0x11020419,0x7c00100,0x220401,0x11020519,0x7c00100,0x220400,0x11020600,0x4000400,0x100002,0x11020600,0x4000400,0x200400,0x11020600,0x7c00500,0x130400,
0x11020600,0x7c00d00,0x130400,0x11020701,0x2802400,0x962460,0x11020701,0x2802400,0x962461,0x11020701,0x2802400,0xc62460,0x1102080e,0x7c00100,0x220400,0x1102080e,
0x7c00100,0x250400,0x11020908,0x7c00100,0x220400,0x11020908,0x7c00100,0x220401,0x11020908,0x7c00100,0x250400,0x11020908,0x7c00100,0x250401,0x11022800,0x24000000,
0x100000,0x11022800,0x24000000,0x200000,0x11022800,0x24000000,0x200002,0x11022800,0x24000000,0x401000,0x11022800,0x24000000,0xf00002,0x11022800,0x24000000,0xf0ac02,
0x11022800,0x24000000,0x1500000,0x11022800,0x24000002,0x100000,0x11022800,0x24000002,0x370000,0x11022800,0x24000002,0x470000,0x11022800,0x24000006,0x400000,0x11022800,
0x24000008,0x1710000,0x11022800,0x24000008,0x1712c00,0x11022800,0x24000020,0x100000,0x11022800,0x24000020,0x1500000,0x11022800,0x24000020,0x1500002,0x11022900,0x4000000,
0x10000e,0x11022900,0x4000000,0x10000f,0x11022919,0x7c00100,0x12040f,0x11022c00,0x4000000,0x100002,0x11022c00,0x4000000,0x1500002,0x11022c00,0x4000000,0x1600002,
0x11022c00,0x4000000,0x1410000f,0x11022c00,0x7c00120,0x120405,0x11022c0e,0x7c00100,0x250401,0x11022c19,0x7c00100,0x150401,0x11022d00,0x4000000,0x100006,0x11022d00,
0x4000000,0x200006,0x11022d19,0x7c00100,0x120402,0x11022d19,0x7c00100,0x150402,0x11022e00,0x24000000,0x200000,0x11022e00,0x24000020,0x100000,0x11022e00,0x24000020,
0x14100000,0x11022f00,0x24000020,0x100000,0x11022f00,0x24000020,0x100001,0x11022f00,0x24000020,0x100002,0x11023000,0x24000000,0x100000,0x11023300,0x4000000,0x100002,
0x11023300,0x4000000,0x100003,0x11023300,0x4000100,0x120403,0x11023300,0x4000100,0x150403,0x11023300,0x4000100,0x14150403,0x11023400,0x24000000,0x100000,0x11023500,
0x24000000,0x100000,0x11023600,0x24000000,0x100000,0x11023600,0x24000020,0x100000,0x11023600,0x24000020,0x14100000,0x11023700,0x24000000,0x4100000,0x11023700,0x24000000,
0x4e00000,0x11023700,0x24000000,0x14100000,0x11023700,0x24000000,0x14e00000,0x11023700,0x24000020,0x100000,0x11023700,0x24000020,0x4100000,0x11023700,0x24000020,0x14100000,
0x11023800,0x4000000,0x100000,0x11023800,0x24000000,0x200000,0x11024e67,0,0,0x11025600,0x4000000,0x100000,0x11042a00,0x4000000,0x1600000,0x11045700,
0x4000000,0x20000a,0x11045700,0x4000020,0x20000a,0x11045712,0x7c00100,0xe3040a,0x11045712,0x7c80100,0xe3040a,0x11045716,0x7c00100,0xe30c0a,0x11045716,0x7c00100,
0x2530c0a,0x11063d00,0x4000001,0x445811,0x11065700,0x4000000,0x810011,0x11065700,0x4000000,0xe00011,0x11065700,0x4000000,0x1410011,0x11065700,0x4000000,0x1500011,
0x11065700,0x4000000,0x1600011,0x11065700,0x4000006,0xe70011,0x11065700,0x4000008,0xe00011,0x11065700,0x4000008,0xe02c11,0x11065700,0x4000010,0x871411,0x11065700,
0x4000010,0x1201411,0x11065700,0x4000010,0x1271011,0x11065700,0x4000020,0xe00011,0x11065700,0x4000400,0xe00011,0x11065700,0x4000420,0xe00011,0x11065700,0x6800000,
0xe01c11,0x11065700,0x6800040,0xe29811,0x11065700,0xc000010,0x80ac11,0x11065700,0xc000010,0xb48011,0x11065719,0x7c00100,0xe20411,0x11065719,0x7c00100,0xe50411,
0x11065719,0x7c00140,0xe20411,0x11065719,0x7c00140,0xe50411,0x11080100,0x6800000,0x201c00,0x11080100,0x68000c0,0x19329800,0x11080100,0x24000000,0x200000,0x11080100,
0x24000000,0x810000,0x11080100,0x24000000,0x1410000,0x11080100,0x24000000,0x1500000,0x11080100,0x24000000,0x1600000,0x11080100,0x24000000,0x1b00000,0x11080100,0x24000000,
0x2410000,0x11080100,0x24000000,0x18200000,0x11080100,0x24000006,0xd70000,0x11080100,0x24000008,0x1713c00,0x11080100,0x24000008,0x1714000,0x11080100,0x24000010,0x1001400,
0x11080100,0x24000010,0x1071000,0x11080100,0x24000010,0x1071400,0x11080100,0x24000020,0x200000,0x11080100,0x24000020,0x400000,0x11080100,0x24000020,0x1600000,0x11080100,
0x24000400,0x200000,0x11080100,0x24000420,0x200000,0x11080100,0x2c000010,0xb48000,0x11080100,0x2c000010,0x100ac00,0x11080100,0x44000001,0x1a45800,0x11080119,0x7c00100,
0x220400,0x11080119,0x7c00100,0x250400,0x11080119,0x7c001c0,0x220400,0x11080119,0x7c001c0,0x250400,0x11080200,0x4000400,0x200002,0x11080200,0x24000000,0x200000,
0x11080200,0x24000000,0x1500000,0x11080200,0x24000000,0x1600000,0x11080200,0x24000020,0x200000,0x110a1e12,0x7c00100,0x2130480,0x110a1e12,0x7c80100,0x2130480,0x110a3000,
0x24000000,0x34e00000,0x110a3000,0x24100000,0x810001,0x110a3000,0x24100000,0x1410001,0x110a3700,0x24000000,0x34200000,0x110a3d00,0x4000000,0xe00000,0x110a3d00,0x4000000,
0xe00002,0x110a3d00,0x24000000,0xe00000,0x110a3d11,0x7c00300,0xe30000,0x110a3d11,0x7c00900,0x1230400,0x110a3d12,0x2802400,0x962460,0x110a3e14,0x7c00100,0xe30000,
0x110a3e14,0x7c00100,0xe30001,0x110a3e14,0x7c00100,0x2530000,0x110a3e14,0x7c00900,0x1230000,0x110a3e14,0x7c00900,0x1230001,0x110a3f16,0x7c00100,0xe30c00,0x110a3f16,
0x7c00100,0xe30c01,0x110a3f16,0x7c00100,0x2530c00,0x110a3f16,0x7c00900,0x1230c00,0x110a3f16,0x7c00900,0x1230c01,0x110a4005,0x7c00100,0xe30400,0x110a4112,0x7c00100,
0xe30402,0x110a4112,0x7c80100,0xe30402,0x110a4400,0x4000000,0xe00000,0x110a4412,0x4000000,0xe00002,0x110a4412,0x4000000,0xe00003,0x110a4416,0x4000000,0xe00c03,
0x110a4500,0x4000000,0xe0000d,0x110a4516,0x4000000,0xe00c0d,0x110a4711,0x7c40300,0xe30000,0x110a4f11,0x7c00300,0xe30001,0x110a4f11,0x7c40300,0xe30000,0x110a5300,
0x4000000,0x810010,0x110a5300,0x4000000,0xe00002,0x110a5300,0x4000000,0xe00010,0x110a5300,0x4000000,0x1410010,0x110a5300,0x4000002,0xe70010,0x110a5300,0x4000008,
0x810010,0x110a5300,0x4000008,0x1410010,0x110a5300,0x6800000,0xe01c02,0x110a5300,0x6800000,0xe01c10,0x110a5400,0x4000000,0x81000c,0x110a5400,0x4000000,0xe0000c,
0x110a5400,0x4000000,0x141000c,0x110a5400,0x4000000,0x150000c,0x110a5400,0x4000000,0x160000c,0x110a5400,0x4000002,0xe7000c,0x110a5400,0x4000010,0x87140c,0x110a5400,
0x4000010,0xe7000c,0x110a5400,0x4000010,0x120140c,0x110a5400,0x4000010,0x127100c,0x110a5400,0x4000020,0xe0000c,0x110a5400,0x4000026,0xe7000c,0x110a5400,0xc000010,
0x80ac0c,0x110a5400,0xc000010,0xb4800c,0x11400c0c,0x4000010,0xb00000,0x11400c0c,0x4000010,0x1071400,0x11400c0c,0xc000010,0xb48000,0x11400c16,0x7c00900,0x230400,
0x11400f40,0xc000010,0x448000,0x11400f54,0xc000010,0x448000,0x11401d89,0x4000000,0x200000,0x11403dbf,0x4000000,0xe00000,0x114457b4,0x4000004,0x120000a,0x114457b4,
0x4000008,0x81000a,0x114457b4,0x4000008,0x141000a,0x114457b4,0x4000010,0x87000a,0x114457b4,0xc000010,0x84800a,0x114457bd,0x3802500,0x126246a,0x114457bd,0x7c00d00,
0x2530c0a,0x114a3db4,0x24000000,0x810000,0x114a3db4,0x24000000,0x1410000,0x114a3db4,0x24000008,0x810000,0x114a3db4,0x24000008,0x1410000,0x114a3db4,0x24000010,0x870000,
0x114a3db4,0x2c000010,0x848000,0x114a3dba,0x4000000,0xe00000,0x114a3dba,0x24000000,0xe00000,0x114a3dba,0x24000002,0x1200000,0x114a3dba,0x24000002,0x14e00000,0x114a3dba,
0x24000008,0x810000,0x114a3dba,0x24000008,0x1410000,0x114a3dbd,0x7c00900,0x930c00,0x114a3dbd,0x7c00900,0xe30c00,0x114a3dbf,0x7c00300,0xe30000,0x114a3ebd,0x7000400,
0x1200c02,0x114a3fb4,0x4000004,0x1200000,0x114a3fbd,0x7c00d00,0x2530c00,0x114a42bf,0x4000000,0xe00000,0x114a42bf,0x4000000,0xe0000f,0x114a44bf,0x4000000,0xe00002,
0x114a44bf,0x4000000,0xe00003,0x114a44bf,0x4000000,0x14e00003,0x114a45bf,0x4000000,0xe00002,0x114a45bf,0x4000000,0xe0000d,0x1180090a,0x2802400,0x962460,0x11800c1e,
0x2802100,0x962460,0x11800c1e,0x2802500,0x962460,0x11800f27,0x2802400,0x962460,0x11800f34,0x2802400,0x962460,0x11820700,0x2802400,0x962460,0x11820700,0x2802500,
0x962460,0x118a3dc0,0x2802400,0x962460,0x118a3ebd,0x2802400,0x962460,0x11c00904,0x2802400,0x962460,0x11c00908,0x2802400,0x962460,0x11c00c20,0xc000010,0xb48000,
0x11c00c23,0x6800000,0x1329800,0x11c00f6d,0x6800000,0x1329800,0x11c01072,0x6800000,0x1329800,0x11c01176,0x6800000,0x1329800,0x11c0127a,0x6800000,0x1329800,0x11c0147e,
0x4000000,0x200000,0x11c0147e,0x6800000,0x1329800,0x11c01682,0x6800000,0x1329800,0x11c051fa,0x7c00100,0x230408,0x20000067,0x1000,0,0x20000b13,0x2802400,
0x962460,0x20000b13,0x2802500,0x962460,0x20001b27,0x2802100,0x962460,0x20001b27,0x2802100,0x962461,0x20001b27,0x2802400,0x962460,0x20001b27,0x2806400,0x962460,
0x20001b27,0x2902100,0x962462,0x20001b27,0x4000000,0x200000,0x20001b27,0x4000000,0x400000,0x20001b27,0x4000000,0x500000,0x20001b27,0x4000000,0x810000,0x20001b27,
0x4000000,0xb00000,0x20001b27,0x4000000,0xc0000b,0x20001b27,0x4000000,0x1410000,0x20001b27,0x4000010,0xb00000,0x20001b27,0x4000010,0xc00000,0x20001b27,0x6800000,
0x1329800,0x20001b27,0x6800100,0x462540,0x20001b27,0x6800400,0x962540,0x20001b27,0x7c00100,0x230400,0x20001b27,0x7c00100,0x230401,0x20002619,0x7c00100,0x220401,
0x20002a00,0x4000000,0x1600000,0x20004b67,0,0x1900000,0x20004c67,0,0x1900000,0x20004d67,0,0x1900000,0x20006d67,0x1000,0,0x20006e67,
0x1000,0,0x20026d67,0,0,0x20026e67,0,0,0x200a4a12,0x7c00100,0x1f304c1,0x200a4a12,0x7c00100,0x20304e1,0x21005600,0x4000000,
0x700000,0x21022a00,0x4000000,0x1600000,0x30000419,0x7c00100,0x220400,0x30000419,0x7c00100,0x220401,0x30000419,0x7c00100,0x250400,0x30000419,0x7c00100,0x250401,
0x30000519,0x7c00100,0x220400,0x30000600,0x4000400,0x200400,0x30000600,0x7c00500,0x230400,0x30000605,0x4000400,0x200400,0x3000080e,0x7c00100,0x220400,0x30000908,
0x2000,0x962460,0x30000908,0x7c00100,0x220400,0x30000908,0x7c00100,0x220401,0x30000908,0x7c00100,0x250400,0x30000908,0x7c00100,0x250401,0x30000a03,0x4000006,
0x400400,0x30000c02,0x4000000,0x200000,0x30000c02,0x7c00100,0x230400,0x30000d22,0x2802100,0x962460,0x30000d22,0x2802400,0x962460,0x30000d22,0x2802500,0x962460,
0x30000d22,0x4000000,0x200000,0x30000d22,0x4000010,0x200000,0x30000d22,0x7c00100,0x230400,0x30000d22,0xc000010,0x248000,0x30000d22,0x80000000,0x218960,0x30000e25,
0x2802500,0x962460,0x30000e25,0x7c00100,0x230400,0x30001821,0x2802100,0x962460,0x30001821,0x2806400,0x962460,0x30001821,0x4000000,0x200000,0x30001821,0x6800100,
0x962540,0x30001821,0x6800100,0x962541,0x30001821,0x7c00100,0x230400,0x30001b27,0x2802100,0x962460,0x30001b27,0x2802400,0x962460,0x30001b27,0x4000000,0x200000,
0x30001b27,0x4000000,0x400000,0x30001b27,0x7c00100,0x230400,0x30001c1c,0x2802100,0x1862460,0x30001c1c,0x2802400,0x1862460,0x30001c1c,0x2806400,0x1862460,0x30001c1c,
0x4000000,0x200000,0x30001c1c,0x6800100,0x1862400,0x30001c1c,0x6800100,0x1862540,0x30001c1c,0x7c00100,0x1830000,0x30001c1c,0x7c00100,0x1830001,0x30001c1c,0xc000010,
0x448000,0x30001f0b,0x4000000,0x200000,0x30001f0b,0x4000010,0x200000,0x30001f0b,0x4000010,0x400000,0x30001f0b,0x6800000,0x200000,0x30001f0b,0x7c00100,0x230400,
0x30001f0b,0xc000010,0x248000,0x30002006,0x7c00100,0x250400,0x30002128,0x4000000,0x200000,0x30002128,0x7c00100,0x230400,0x30002128,0xc000010,0x248000,0x3000221d,
0x4000000,0x810000,0x3000221d,0x4000000,0x1410000,0x3000221d,0x4000001,0x445800,0x3000221d,0x7c00100,0x230400,0x30002300,0x4000010,0x400000,0x30002320,0x7c00100,
0x230400,0x30002417,0x2802100,0x1862460,0x30002417,0x2802400,0x1862460,0x30002417,0x2806400,0x1862460,0x30002417,0x2882000,0x1862460,0x30002417,0x4000000,0x200000,
0x30002417,0x4000000,0x400000,0x30002417,0x4000000,0x1600000,0x30002417,0x4000010,0x400000,0x30002417,0x4000010,0x1200000,0x30002417,0x6800000,0x1329800,0x30002417,
0x6800100,0x1862540,0x30002417,0x7c00100,0x1830000,0x30002417,0x7d00100,0x1830000,0x3000251b,0x80000,0xc18820,0x3000251b,0x2802100,0x962460,0x3000251b,0x3c02100,
0x962460,0x3000251b,0x4000000,0x200000,0x3000251b,0x4000006,0x500000,0x3000251b,0x4000010,0x400000,0x3000251b,0x4000010,0xb70000,0x3000251b,0x4000800,0x200000,
0x3000251b,0x6800000,0x1329800,0x3000251b,0x7c00100,0x230400,0x3000251b,0x7c00900,0x230400,0x3000251b,0xc000010,0xb48000,0x3000251b,0x12882000,0x962460,0x30002800,
0x24000000,0x200000,0x30002800,0x2c000010,0x1248002,0x30002800,0x2c000010,0x15248002,0x30002a00,0x4000000,0x1600000,0x30002b01,0x2000,0x962460,0x30002b01,0x2000,
0x8962460,0x30002c00,0x4000000,0x200000,0x30002c00,0x7c00100,0x14220405,0x30002d19,0x7c00100,0x250400,0x30002e00,0x24000000,0x200000,0x30003000,0x24000000,0x200000,
0x30003000,0x24000000,0x4200000,0x30003100,0x24000000,0x200000,0x30003600,0x24000000,0x200000,0x30003700,0x24000000,0x4200000,0x3000392e,0x24000000,0x200000,0x30005013,
0x7c00100,0x2633801,0x30005600,0,0x918820,0x30020600,0x4000400,0x500400,0x30020701,0x2802400,0x962460,0x30020701,0x2802400,0xc62460,0x300a3a11,0x4020000,
0xe00000,0x300a3a11,0x4020000,0xe00002,0x300a3b11,0x4020000,0xe00002,0x300a3c00,0x4008000,0xe00000,0x300a3c00,0x4010000,0xe00000,0x300a3d11,0x7c00300,0xe30002,
0x300a4305,0x7c00100,0xe30400,0x300a4611,0x7c40300,0xe30000,0x300a4829,0x7c00100,0xe30400,0x300a4829,0x7c00900,0x1230400,0x300a4929,0x4000000,0xe00000,0x3040258f,
0x4000010,0x400000,0x3040258f,0x4000010,0xb70000,0x3040258f,0xc000010,0xb48000,0x304028af,0x4000001,0xc41c0b,0x304a3dbf,0x4000000,0xe00000,0x30800c1e,0x2802100,
0x962460,0x30c01c87,0x6800000,0x1329800,0x3100080e,0x7c00120,0x220402,0x3100080e,0x7c00120,0x250402,0x31005167,0x1000,0,0x3100581e,0x4000000,0x200000,
0x3100581e,0x7c00100,0x230400,0x3100590d,0x7c00100,0x230400,0x31005a09,0x7c00100,0x220400,0x31005a09,0x7c00100,0x250400,0x31005b00,0x4000000,0x200000,0x31005c00,
0x80000,0x918820,0x31005c00,0x2802000,0x962460,0x31005c00,0x2802400,0x962460,0x31005c00,0x4000000,0x200000,0x31005c00,0x4000000,0x200001,0x31005c00,0x6800000,
0x962540,0x31005c00,0x6800400,0x962540,0x31005c01,0x2802400,0x962460,0x31005d00,0x4000020,0x200005,0x31005d00,0x6800020,0x1329805,0x31005d00,0x7c00120,0x220405,
0x31005d00,0x7c00120,0x250405,0x31006000,0x82000,0x8962460,0x31006000,0x180000,0x918820,0x310a5e11,0x7c40300,0xe30000,0x310a5f11,0x7c00300,0xe30001,0x32000419,
0x7c00100,0x250400,0x3200080e,0x4000020,0x200000,0x3200080e,0x7c00100,0x220400,0x3200080e,0x7c00100,0x250400,0x32000908,0x7c00100,0x220400,0x32000908,0x7c00100,
0x250400,0x32000c02,0x7c00100,0x230400,0x32000e25,0x7c00100,0x230400,0x32001d0c,0x7c00100,0x230400,0x32002800,0x80000,0x1e18820,0x32002800,0x80020,0x218820,
0x32002800,0x4000001,0x445802,0x32002800,0x24000000,0x200000,0x32002800,0x24000000,0x200002,0x32002800,0x24000020,0x200000,0x32002800,0x2c000010,0x1248002,0x32002919,
0x7c00100,0x22040f,0x32002a00,0x4000000,0x1600000,0x32002b01,0x2000,0x962460,0x32002b01,0x2802000,0x962460,0x32002b01,0x2802020,0x962460,0x32002c00,0x4000000,
0x200000,0x32002c00,0x4000020,0x200000,0x32002c00,0x4000020,0x200005,0x32002c00,0x7c00120,0x220405,0x32002c00,0x7c00120,0x250405,0x32002e00,0x24000020,0x200000,
0x32002f00,0x24000020,0x200000,0x32003000,0x24000000,0x200000,0x32003000,0x24000020,0x200000,0x32003500,0x24000000,0x200000,0x32003600,0x24000020,0x200000,0x32003600,
0x24000020,0x14200000,0x32003700,0x24000000,0x200000,0x32003700,0x24000000,0x4100000,0x32003700,0x24000000,0x4200000,0x32003700,0x24000000,0x14200000,0x32003800,0x24000000,
0x810000,0x32003800,0x24000000,0x1410000,0x32005102,0x4000000,0x1500008,0x32005502,0x7c00100,0x230400,0x32006108,0x7c00100,0x220400,0x32006108,0x7c00100,0x250400,
0x3200622a,0x2802100,0x962460,0x3200622a,0x2806000,0x962460,0x3200622a,0x7c00100,0x230400,0x3200632b,0x2802100,0x962460,0x3200632b,0x2806000,0x962460,0x3200632b,
0x7c00100,0x230400,0x3200642c,0x2802100,0x962460,0x3200642c,0x7c00100,0x230400,0x3200652d,0x2802100,0x962460,0x3200652d,0x7c00100,0x230400,0x32006600,0x24000020,
0x200000,0x32006700,0x24000020,0x200000,0x32006800,0x24000020,0x200000,0x32006800,0x24000020,0x14200000,0x32006900,0x24000020,0x200000,0x32006900,0x24000020,0x810000,
0x32006900,0x24000020,0x1410000,0x32006a00,0x24000020,0x200000,0x32006a00,0x24000020,0x200001,0x32006a00,0x24000020,0x200002,0x32020701,0x2882000,0xc62460,0x32023300,
0x4000000,0x100000,0x32026c01,0x12882000,0x962460,0x32026c01,0x12882000,0x8962460,0x32065700,0x4000000,0x810011,0x32065700,0x4000000,0x1410011,0x32086600,0x24000020,
0x810000,0x32086600,0x24000020,0x1410000,0x32086900,0x24000020,0x810000,0x32086900,0x24000020,0x1410000,0x320a3600,0x24000020,0x34200000,0x320a3d11,0x7c00100,0x1230400,
0x320a3e14,0x7c00100,0xe30010,0x320a3e14,0x7c00100,0x2530000,0x320a3f16,0x7c00100,0xe30c10,0x320a4400,0x4000000,0xe00003,0x320a4929,0x4000000,0xe00000,0x320a4f11,
0x7c00300,0xe30001,0x320a6b16,0x7c00100,0x2530c00,0x3240638b,0xc000010,0x448000,0x324a3dc2,0x4000000,0x14e00000,0x324a3dc2,0x7c00100,0x1230400,0x324a3fbd,0x4000002,
0x1200c00,0x324a53ba,0x24000000,0xe00000,0x32820701,0x2802000,0x962460,0x40000419,0x7c00100,0x220400,0x40000519,0x7c00100,0x220400,0x40000600,0x4000400,0x200400,
0x4000080e,0x7c00100,0x220400,0x4000080e,0x7c00100,0x250400,0x4000080e,0x7c00100,0x250402,0x40000c02,0x2802100,0x962460,0x40000c02,0x2802400,0x962460,0x40000c02,
0x2802500,0x962460,0x40000c02,0x4000000,0x200000,0x40000c02,0x4000000,0x1071400,0x40000c02,0x7c00100,0x230400,0x40000c02,0x80000000,0x218960,0x40000d22,0x7c00100,
0x230400,0x40000f0a,0x7c00100,0x230400,0x40001004,0x7c00100,0x230400,0x40001110,0x2802100,0x962460,0x40001110,0x6800100,0x962540,0x4000120f,0x2802100,0x962460,
0x4000120f,0x4000000,0x1600000,0x4000120f,0x7c00100,0x230400,0x4000131f,0x7c00100,0x230400,0x40001423,0x4000000,0x200000,0x40001423,0x4000000,0x1600000,0x40001615,
0x2802400,0x962460,0x40001615,0x7c00100,0x230400,0x40002417,0x2802400,0x1862460,0x40002417,0x4000000,0x200000,0x40002800,0x6800000,0x201c00,0x40002800,0x24000002,
0x200000,0x40002c00,0x4000000,0x200002,0x40003000,0x24000000,0x14200000,0x40003000,0x24000020,0x200000,0x40003700,0x24000000,0x200000,0x40003700,0x24000000,0x4200000,
0x40003700,0x24000000,0x14200000,0x40005a09,0x7c00100,0x220400,0x40005a09,0x7c00100,0x250400,0x40005d00,0x7c00120,0x220405,0x40006f30,0x2802100,0x962460,0x40006f30,
0x2802400,0x962460,0x40006f30,0x4000000,0x200000,0x40006f30,0x6800000,0x1329800,0x40006f30,0x6800100,0x962540,0x40006f30,0x7c00100,0x230400,0x40006f30,0xc000010,
0xb48000,0x40007034,0x7c00100,0x1830000,0x40007117,0x4000000,0x200000,0x40007208,0x7c00100,0x220400,0x4000720e,0x7c00100,0x220400,0x4000720e,0x7c00500,0x22040e,
0x4000720e,0x7c00500,0x22040f,0x40007219,0x7c00100,0x220400,0x40007219,0x7c00500,0x220400,0x40007219,0x7c00500,0x22040e,0x40007219,0x7c00500,0x22040f,0x40007300,
0x24000000,0x200000,0x40007300,0x24000000,0x14200000,0x40007400,0x4000000,0x200000,0x40007531,0x7c00100,0x230400,0x40007631,0x7c00100,0x230400,0x40007835,0x4000010,
0x400000,0x40007835,0x7c00100,0x230400,0x40007933,0x7c00100,0x230400,0x40007a32,0x6800000,0x1329800,0x40007a32,0x7c00100,0x230400,0x40007b2f,0x7c00100,0x230400,
0x40007c00,0x4000000,0x200000,0x40020701,0x2802400,0x962460,0x40020701,0x2802400,0xc62460,0x40023300,0x4000000,0x200000,0x40027d01,0x12882000,0x962460,0x400a3700,
0x24000000,0x34200000,0x400a3700,0x24000000,0x34e00000,0x400a4400,0x4000000,0xe0000d,0x400a4412,0x4000000,0xe00002,0x400a4412,0x4000000,0xe00003,0x400a4500,0x4000000,
0xe0000d,0x400a5300,0x4000000,0x810010,0x400a5300,0x4000000,0x1410010,0x404077fc,0x4000000,0x200000,0x404077ff,0x4000000,0x200000,0x404077ff,0x4000000,0x400000,
0x40c0147e,0x4000000,0x200000,0x40c051fa,0x4000000,0x200000,0x41000419,0x7c00100,0x220400,0x41000419,0x7c00100,0x250400,0x4100080e,0x7c00100,0x220400,0x4100080e,
0x7c00100,0x250400,0x41000908,0x7c00100,0x220400,0x41000908,0x7c00100,0x250400,0x41000b13,0x2802000,0x962460,0x41000b13,0x2802100,0x962460,0x41000b13,0x4000000,
0xb00000,0x41000c02,0x2802100,0x962460,0x41000c02,0x4000000,0x1500000,0x41000c02,0xc000010,0xb48000,0x41000f0a,0x7c00100,0x230400,0x41001004,0x7c00100,0x230400,
0x41001423,0x7c00100,0x230400,0x41001b27,0x4000000,0x500000,0x41001d0c,0x7c00100,0x230400,0x41001d0c,0x7c00100,0x23040f,0x41001f0b,0x2802400,0x962460,0x41001f0b,
0x4000000,0x200000,0x41001f0b,0x7c00100,0x230400,0x41002800,0x24000000,0x200000,0x41002800,0x24000000,0x400000,0x41002919,0x7c00100,0x22040e,0x41002a00,0x4000000,
0x1600000,0x41002b01,0x2802020,0x962460,0x41002c00,0x4000000,0x200000,0x41002c00,0x7c00120,0x220405,0x41003000,0x24000000,0x200000,0x41003700,0x24000000,0x4200000,
0x41003700,0x24000000,0x14200000,0x41003700,0x24000000,0x14e00000,0x41005d00,0x7c00120,0x220405,0x41006600,0x24000020,0x200000,0x41006600,0x24000020,0x810000,0x41006600,
0x24000020,0x1410000,0x41007208,0x7c00100,0x22040f,0x41007219,0x7c00100,0x220400,0x41007300,0x24000000,0x200000,0x41007e0e,0x2802000,0x962460,0x41007e0e,0x4000000,
0x200000,0x41007f0e,0x4000000,0x200000,0x41007f0e,0x7c00100,0x230400,0x41008002,0x7c00100,0x230400,0x41008137,0x2802100,0x962460,0x41008137,0x4000000,0x200000,
0x41008137,0x6800100,0x962540,0x41008137,0x7c00100,0x230400,0x41008301,0x2802000,0x962460,0x41008407,0x4000000,0x200000,0x41008407,0x4000000,0x400000,0x41008407,
0x4000000,0xb00000,0x41008407,0x7c00100,0x220400,0x41008407,0x7c00100,0x250400,0x4100850b,0x7c00100,0x230400,0x4100860b,0x4000000,0x200000,0x4100860b,0x7c00100,
0x230400,0x4100870c,0x7c00100,0x220400,0x41008838,0x7c00100,0x220400,0x41008838,0x7c00100,0x250400,0x41008939,0x2802000,0x962460,0x41008939,0x2802100,0x962460,
0x41008939,0x2806000,0x962460,0x41008939,0x4000000,0x200000,0x41008939,0x4000000,0x400000,0x41008939,0x7c00100,0x230400,0x41008939,0xc000010,0x448000,0x41008a00,
0x4000400,0x200400,0x41008b3b,0x4000000,0x1800000,0x41008b3b,0x6800000,0x1329800,0x41008b3b,0x7c00100,0x1830000,0x41008b3b,0x7e00100,0x1830000,0x41008c3d,0x4000010,
0x400000,0x41008c3d,0x7c00100,0x230400,0x41008d0e,0x7c00100,0x22040f,0x41008d19,0x7c00100,0x220400,0x41008d19,0x7c00100,0x22040f,0x41008e00,0x24000000,0x200000,
0x41008e00,0x24000000,0x400000,0x41008e00,0x24000000,0x1710000,0x41008e00,0x24000006,0x400000,0x41008f3a,0x2802100,0x962460,0x41008f3a,0x2806000,0x962460,0x41008f3a,
0x4000000,0x200000,0x41008f3a,0x6800100,0x962540,0x41008f3a,0x7c00100,0x230400,0x4100903c,0x7c00100,0x230400,0x4100903c,0x7c00100,0x23040f,0x41020701,0x2802000,
0x962460,0x41020701,0x2802000,0xc62460,0x410a3700,0x24000000,0x34200000,0x410a3700,0x24000000,0x34e00000,0x410a4412,0x4000000,0xe00003,0x410a4711,0x7c40300,0xe30000,
0x410a4f11,0x7c00300,0xe30001,0x410a9100,0x4000000,0x800010,0x410a9100,0x4000000,0x810010,0x410a9100,0x4000000,0x870010,0x410a9100,0x4000000,0xb00010,0x410a9100,
0x4000000,0xf00010,0x410a9100,0x4000000,0x1001410,0x410a9100,0x4000000,0x1071010,0x410a9100,0x4000000,0x1071410,0x410a9100,0x4000000,0x1410010,0x41408ac5,0x4000400,
0x200000,0x414a82bf,0x4000000,0xe00000,0x41808300,0x2802000,0x962460,0x41c0147e,0x6800000,0x1329800,0x50000419,0x7c00100,0x220400,0x50000419,0x7c00100,0x250400,
0x5000080e,0x7c00100,0x220400,0x50000908,0x7c00100,0x220400,0x50000908,0x7c00100,0x250400,0x50000b13,0x2802500,0x962460,0x50000f0a,0x7c00100,0x230400,0x50001615,
0x2802100,0x962460,0x50001615,0x7c00100,0x230400,0x50002b01,0x2802020,0x962460,0x50002c00,0x4000000,0x200000,0x50002c19,0x7c00100,0x220400,0x50002d19,0x7c00100,
0x220400,0x50003000,0x24000000,0x200000,0x50003000,0x24000020,0x200000,0x50003700,0x24000000,0x4200000,0x50005d00,0x7c00120,0x220405,0x50005d00,0x7c00120,0x250405,
0x50006108,0x7c00100,0x220400,0x50006108,0x7c00100,0x250400,0x50006600,0x24000020,0x200000,0x50007300,0x24000000,0x200000,0x50008301,0x2802400,0x962460,0x50008a00,
0x7c00500,0x230400,0x50009257,0x2802400,0x962460,0x50009257,0x4000000,0x200000,0x50009257,0x4000010,0x1071400,0x50009257,0x6800000,0x1329800,0x50009257,0x7c00100,
0x230400,0x50009257,0x7c00500,0x230400,0x50009257,0x7c00900,0x230400,0x50009257,0xc000010,0xb48000,0x5000933e,0x2802100,0x962460,0x5000933e,0x2802400,0x962460,
0x5000933e,0x4000000,0x200000,0x5000933e,0x4000000,0x400000,0x5000933e,0x4000010,0x400000,0x5000933e,0x6800000,0x1329800,0x5000933e,0x6800100,0x962540,0x5000933e,
0x6800100,0x962541,0x5000933e,0x6804400,0x962540,0x5000933e,0x7c00100,0x230400,0x5000933e,0x7c00100,0x230401,0x5000933e,0xc000010,0x448000,0x50009419,0x7c00100,
0x220400,0x50009419,0x7c00100,0x250400,0x50009500,0x4000400,0x200400,0x5000965a,0x4000000,0x500000,0x5000965a,0x7c00100,0x230400,0x5000965a,0xc000010,0xb48000,
0x5000975b,0x4000000,0x200000,0x5000975b,0x4000010,0x400000,0x5000975b,0x7c00100,0x230400,0x50009865,0x7c00100,0x230400,0x50009965,0x4000010,0x400000,0x50009965,
0x7c00100,0x230400,0x50409abf,0x4000000,0x200000,0x5100080e,0x7c00100,0x220400,0x5100080e,0x7c00100,0x250400,0x51000c02,0x2802100,0x962460,0x51000c02,0x4000000,
0x1500000,0x51000c02,0x4000020,0x200000,0x51000c02,0x7c00100,0x230400,0x51000f0a,0x7c00100,0x230400,0x51000f0a,0x7c00500,0x230400,0x51001110,0x2802100,0x962460,
0x5100131f,0x2802100,0x962460,0x51001423,0x7c00100,0x230400,0x51001524,0x2802100,0x962460,0x51001524,0x4000000,0x200000,0x51001524,0x7c00100,0x230400,0x5100171a,
0x2802100,0x962460,0x5100171a,0x4000000,0x200000,0x5100171a,0x4000000,0x1500000,0x5100171a,0x7c00100,0x230400,0x51001b27,0x4000000,0x200000,0x51001b27,0x4000000,
0x400000,0x51001b27,0x4000000,0x500000,0x51001b27,0x7c00100,0x230400,0x51001c1c,0x2802100,0x1862460,0x51001c1c,0x2802500,0x1862460,0x51001c1c,0x2806400,0x1862460,
0x51001c1c,0x4000000,0x1800000,0x51001c1c,0x6800000,0x1329800,0x51001c1c,0x6800100,0x1862400,0x51001c1c,0x6800100,0x1862540,0x51001c1c,0x6800500,0x1862400,0x51001c1c,
0x7c00100,0x1830000,0x5100251b,0x7c00100,0x230400,0x51002619,0x7c00100,0x220400,0x51002619,0x7c00100,0x250400,0x51002800,0x80020,0x218820,0x51002c00,0x4000000,
0x200000,0x51002d19,0x7c00100,0x230400,0x51003700,0x24000000,0x4200000,0x51003700,0x24000000,0x4e00000,0x51005201,0x2802400,0x962460,0x51005c00,0x4000000,0x200000,
0x51006108,0x7c00100,0x220400,0x51006108,0x7c00100,0x250400,0x51006600,0x24000020,0x200000,0x51006600,0x24000020,0x810000,0x51006600,0x24000020,0x1410000,0x51007300,
0x24000000,0x200000,0x51007300,0x24000020,0x200000,0x51008002,0x7c00100,0x230400,0x51008301,0x2802000,0x962460,0x51008301,0x2802400,0x962460,0x51008a00,0x7c00500,
0x230400,0x51008e00,0x24000000,0x200000,0x51008e00,0x24000000,0x400000,0x51008e00,0x24000000,0x810000,0x51008e00,0x24000000,0x1400000,0x51008e00,0x24000000,0x1410000,
0x51008e00,0x24000000,0x1710000,0x51008e00,0x24000002,0x200000,0x51008e00,0x24000500,0x230400,0x51008e00,0x2c000010,0xb48000,0x51009419,0x7c00100,0x220400,0x51009419,
0x7c00100,0x22040e,0x51009419,0x7c00100,0x22040f,0x51009419,0x7c00100,0x250400,0x51009500,0x4000400,0x200400,0x51009500,0x7c00500,0x230400,0x51009519,0x7c00100,
0x220400,0x51009519,0x7c00100,0x22040f,0x51009519,0x7c00100,0x230400,0x51009519,0x7c00100,0x250400,0x51009b71,0x2802100,0x962460,0x51009b71,0x6800000,0x1329800,
0x51009b71,0x6800100,0x962540,0x51009b71,0x6804400,0x962540,0x51009b71,0x7c00100,0x230400,0x51009c52,0x2802100,0x962460,0x51009c52,0x2802400,0x962460,0x51009c52,
0x2802d00,0x962460,0x51009c52,0x4000010,0x400000,0x51009c52,0x6800000,0x1329800,0x51009c52,0x6800100,0x962540,0x51009c52,0x7c00100,0x230400,0x51009c52,0xc000010,
0x448000,0x51009d6d,0x6800000,0x1329800,0x51009d6d,0x7c00100,0x230400,0x51009d6d,0x7c00500,0x230400,0x51009d6d,0x7c00d00,0x230400,0x51009d6d,0xc000010,0x448000,
0x51009e08,0x2802100,0x962460,0x51009f63,0x4000010,0x400000,0x51009f63,0x6800000,0x1329800,0x51009f63,0x7c00100,0x230400,0x51009f63,0x7c00900,0x230400,0x51009f63,
0xc000010,0x448000,0x51009f63,0xc000010,0xb48000,0x5100a008,0x2000,0x962460,0x5100a008,0x2802400,0x962460,0x5100a008,0x4000000,0x200000,0x5100a008,0x7c00100,
0x220400,0x5100a008,0x7c00100,0x230400,0x5100a008,0x7c00100,0x250400,0x5100a008,0x7c00500,0x230400,0x5100a16f,0x2806400,0x962460,0x5100a16f,0x6800000,0x1329800,
0x5100a16f,0x6800100,0x962540,0x5100a16f,0x7c00100,0x230400,0x5100a16f,0xc000010,0x448000,0x5100a24f,0x2802100,0x962460,0x5100a24f,0x2802400,0x962460,0x5100a24f,
0x6800000,0x1329800,0x5100a24f,0x7c00100,0x230400,0x5100a24f,0xc000010,0x448000,0x5100a36e,0x2802100,0x962460,0x5100a36e,0x4000000,0x200000,0x5100a36e,0x6800100,
0x962540,0x5100a36e,0x6804400,0x962540,0x5100a36e,0x7c00100,0x230400,0x5100a442,0x2802100,0x962460,0x5100a442,0x4000000,0x200000,0x5100a442,0x6800000,0x1329800,
0x5100a442,0x6800100,0x962540,0x5100a442,0x7c00100,0x230400,0x5100a442,0xc000010,0x448000,0x5100a500,0x4000000,0x200000,0x5100a600,0x4000000,0x200000,0x5100a601,
0x2802000,0x962460,0x5100a76b,0x7c00100,0x230400,0x5100a868,0x7c00100,0x230400,0x5100a96c,0x4000000,0x200000,0x5100a96c,0x7c00100,0x230400,0x5100aa00,0x4000000,
0x4e00000,0x5100ab00,0x4000000,0x4e00000,0x51086600,0x24000020,0x810000,0x51086600,0x24000020,0x1410000,0x510a4005,0x7c00100,0xe30400,0x510a4711,0x7c40300,0xe30000,
0x510a7300,0x24000000,0x34200000,0x510aaa00,0x4000000,0x34e00000,0x5140a2f3,0x4000400,0x400000,0x514a82bf,0x4000000,0xe00000,0x51802bb1,0x2802000,0x962460,0x51c00908,
0x2802400,0x962460,0x51c0a008,0x2802400,0x962460,0x52000f0a,0x2802100,0x962460,0x52000f0a,0x6800100,0x962540,0x52000f0a,0x7c00100,0x230400,0x52001004,0x4000000,
0x1600000,0x52001b00,0x4000000,0x200000,0x52001c1c,0x2802100,0x1862460,0x52001c1c,0x6800100,0x1862400,0x52001c1c,0x6800500,0x1862400,0x52001e12,0x7c00100,0x2230500,
0x52001e12,0x7c00100,0x2330520,0x52002128,0x4000002,0x400000,0x52002128,0x7c00100,0x230400,0x52002a00,0x4000000,0x1500000,0x52002a00,0x4000000,0x1600000,0x52002d00,
0x4000000,0x200006,0x52003000,0x24000000,0x200000,0x52006108,0x7c00100,0x220400,0x52006108,0x7c00100,0x250400,0x52008301,0x2802400,0x962460,0x52008407,0x2802400,
0x962460,0x52008407,0x7c00100,0x220400,0x52008407,0x7c00100,0x250400,0x52008b3b,0x6800000,0x1800000,0x52008b3b,0x7c00100,0x1830000,0x52008e00,0x24000000,0x400000,
0x52009419,0x7c00100,0x250400,0x5200975b,0x4000000,0x200000,0x5200ac7e,0x2802000,0x962460,0x5200ac7e,0x2802100,0x962460,0x5200ac7e,0x2802400,0x962460,0x5200ac7e,
0x4000010,0x200000,0x5200ac7e,0x7c00100,0x230400,0x5200ac7e,0xc000010,0x248000,0x5200ad28,0x7c00100,0x230400,0x5200ae6a,0x2802100,0x1862460,0x5200ae6a,0x2802400,
0x962460,0x5200ae6a,0x2802400,0x1862460,0x5200ae6a,0x2806000,0x1862460,0x5200ae6a,0x4000000,0x1800000,0x5200ae6a,0x6800000,0x1329800,0x5200ae6a,0x6800100,0x1862400,
0x5200ae6a,0x6800100,0x1862540,0x5200ae6a,0x7c00100,0x1830000,0x5200ae6a,0x7c00900,0x1830000,0x5200ae6a,0xc000010,0x1848000,0x5200b083,0x4000010,0x400000,0x5200b083,
0x7c00100,0x230400,0x5200b083,0xc000010,0x448000,0x5200b182,0x2802400,0x962460,0x5200b182,0x4000000,0x200000,0x5200b182,0x4000010,0x400000,0x5200b182,0x7c00100,
0x230400,0x5200b182,0xc000010,0x448000,0x5200b30a,0x2802400,0x962460,0x5200b30a,0x4000000,0x200000,0x5200b30a,0x7c00100,0x230400,0x5200b54e,0x2802100,0x962460,
0x5200b54e,0x2802400,0x962460,0x5200b54e,0x4000000,0x200000,0x5200b54e,0x4000010,0x400000,0x5200b54e,0x6800000,0x1329800,0x5200b54e,0x6800100,0x962540,0x5200b54e,
0x6804400,0x962540,0x5200b54e,0x7c00100,0x230400,0x5200b54e,0xc000010,0x448000,0x5200b61c,0x4000000,0x1800000,0x5200b61c,0x6800500,0x1862400,0x5200b61c,0x7c00100,
0x1830000,0x5200b61c,0x7c00900,0x1830000,0x5200b77f,0x2802100,0x1862460,0x5200b77f,0x2802400,0x1862460,0x5200b77f,0x4000000,0x1800000,0x5200b77f,0x4000010,0x1800000,
0x5200b77f,0x7c00100,0x1830000,0x5200b77f,0x7c00500,0x1830000,0x5200b77f,0x7c00900,0x1830000,0x5200b77f,0x7e00100,0x1830000,0x5200b873,0x2802100,0x962460,0x5200b873,
0x2806400,0x962460,0x5200b873,0x6800000,0x1329800,0x5200b873,0x6800100,0x962540,0x5200b873,0x6800400,0x962540,0x5200b873,0x7c00100,0x230400,0x5200b873,0xc000010,
0x448000,0x5200b912,0x7c00100,0x2230500,0x5200b912,0x7c00100,0x2330520,0x5200ba74,0x4000000,0x200000,0x5200ba74,0x4000010,0x400000,0x5200ba74,0x7c00100,0x230400,
0x5200bb85,0x4000000,0x200000,0x5200bb85,0x7c00100,0x230400,0x5200bc75,0x4000000,0x400000,0x5200bc75,0x4000010,0x400000,0x5200bc75,0x7c00100,0x230400,0x5200bd7d,
0x4000000,0x200000,0x5200bd7d,0x7c00100,0x230400,0x5200be7a,0x4000000,0x200000,0x5200be7a,0x7c00100,0x230400,0x5200bf58,0x7c00100,0x230400,0x5200c002,0x4000000,
0x200000,0x5200c178,0x2802000,0x962460,0x5200c178,0x2802100,0x962460,0x5200c178,0x2802400,0x962460,0x5200c178,0x2806400,0x962460,0x5200c178,0x4000000,0x200000,
0x5200c178,0x6800100,0x962540,0x5200c178,0x7c00100,0x230400,0x5200c178,0x7c00100,0x230401,0x5200c178,0xc000010,0x448000,0x5200c178,0x80000000,0x218960,0x5200c247,
0x7c00100,0x230400,0x5200c247,0x7c00100,0x830400,0x5200c247,0x7c00100,0x1430400,0x5200c300,0x4000000,0x200003,0x52022d00,0x4000000,0x100006,0x52023700,0x24000000,
0x4100000,0x52023700,0x24000000,0x4e00000,0x52023700,0x24000000,0x14100000,0x52023700,0x24000000,0x14e00000,0x52023700,0x24000000,0x96800000,0x52024400,0x4000000,0x100000,
0x52027300,0x24000000,0x100000,0x5202c300,0x4000000,0x100000,0x5202c300,0x4000000,0x100002,0x5202c300,0x4000000,0x100003,0x5202c300,0x4000000,0x10000d,0x5202c300,
0x4000100,0x150400,0x5202c300,0x4000100,0x15040d,0x5202c300,0x4000100,0x14150400,0x520a1e12,0x7c00100,0x2130480,0x520a3700,0x24000000,0x34e00000,0x520a3800,0x24000000,
0x34100000,0x520a4711,0x7c40300,0xe30000,0x520a4f11,0x7c00300,0xe30001,0x520a7300,0x24000000,0x34100000,0x520ab412,0x7c00100,0x2130480,0x520ac400,0x4000000,0xe00002,
0x520ac400,0x4000000,0xe0000d,0x520ac400,0x4000000,0x34e0000d,0x520ac414,0x4000000,0xe0000d,0x520ac511,0x7c40300,0xe30000,0x5240af91,0x7c00100,0x230400,0x5240af96,
0x4000400,0x200000,0x5240af98,0x6800400,0x962540,0x5240af98,0x7c00100,0x230400,0x5240afa2,0x7c00100,0x230400,0x5240afa4,0x7c00100,0x230400,0x5240b2c7,0x4000000,
0x200000,0x5240b2c7,0x4000000,0x1500000,0x5240b2d2,0x4000000,0x200000,0x5240b2e0,0x4000000,0x200000,0x5240b5f6,0x7c00900,0x230400,0x524a44bf,0x4000000,0xe00003,
0x5280af91,0x2802400,0x962460,0x5280af92,0x2802400,0x962460,0x5280af98,0x2802400,0x962460,0x5280af9a,0x2802400,0x962460,0x5280af9c,0x2802400,0x962460,0x52c0b3ed,
0x2802400,0x962460,0x52c0b3f1,0x7c00100,0x230400,0x60000c02,0x2802100,0x962460,0x60000c02,0x7c00100,0x230400,0x60000f0a,0x2802100,0x962460,0x60000f0a,0x6800100,
0x962540,0x60000f0a,0x7c00100,0x230400,0x6000131f,0x4000000,0x200000,0x6000171a,0x7c00100,0x230400,0x6000171a,0x7c00100,0x230560,0x60001b27,0x2802100,0x962460,
0x60001b27,0x4000000,0xc00000,0x60001b27,0x7c00100,0x230400,0x60001f0b,0x2802400,0x962460,0x60002919,0x7c00100,0x22040e,0x60002a00,0x4000000,0x1600000,0x60003000,
0x24000000,0x14200000,0x60003000,0x24000000,0x14e00000,0x60003700,0x24000000,0x4200000,0x60003800,0x24000000,0x1710000,0x60005102,0x4000000,0x200000,0x60006108,0x7c00100,
0x220400,0x60006108,0x7c00100,0x250400,0x60006600,0x24000020,0x200000,0x60008301,0x2802000,0x962460,0x6000903c,0x2806000,0x962460,0x6000903c,0x4000000,0x400000,
0x60009519,0x7c00100,0x220400,0x60009519,0x7c00100,0x250400,0x6000a008,0x7c00100,0x220400,0x6000a008,0x7c00100,0x250400,0x6000c300,0x4000000,0x3a703580,0x6000c654,
0x2802000,0x962460,0x6000c654,0x4000010,0x200000,0x6000c654,0x7c00100,0x230400,0x6000c73f,0x2802000,0x962460,0x6000c73f,0x2802100,0x962460,0x6000c73f,0x4000000,
0x200000,0x6000c73f,0x6800100,0x962540,0x6000c73f,0x6804000,0x962540,0x6000c73f,0x7c00100,0x230400,0x6000c80b,0x7c00100,0x230400,0x6000c941,0x2802100,0x962460,
0x6000c941,0x2806000,0x962460,0x6000c941,0x4000000,0x200000,0x6000c941,0x4000010,0x200000,0x6000c941,0x6800000,0x1329800,0x6000c941,0x6800100,0x962540,0x6000c941,
0x7c00100,0x230400,0x6000c941,0xc000010,0x448000,0x6000ca82,0x7c00100,0x230400,0x6000cc00,0x4000000,0x4e00000,0x6000d000,0x4000000,0x200000,0x6002c300,0x4000000,
0x100000,0x6002c300,0x4000000,0x10000d,0x6002c300,0x4000100,0x150400,0x6002c300,0x4000100,0x15040d,0x6002c300,0x4000100,0x14150400,0x600a3000,0x24000000,0x34200000,
0x600a3000,0x24000000,0x34e00000,0x600a3700,0x24000000,0x34200000,0x600a3800,0x24000000,0x34200000,0x600a3800,0x24000000,0xb6800000,0x600a4305,0x7c00100,0xe30400,0x600ac300,
0x4000000,0x34100000,0x600ac400,0x4000000,0x14e0000d,0x600ac400,0x4000000,0x34e0000d,0x600acb14,0x7c00100,0xe30000,0x600acb16,0x7c00100,0xe30c00,0x600acc00,0x4000000,
0x34e00000,0x600acd00,0x4000000,0x34200000,0x600acd00,0x4000000,0x34e00000,0x600acd00,0x4000000,0xb6800000,0x600ace00,0x4000000,0x34e00000,0x600ace00,0x4000000,0xb6800000,
0x600acf00,0x4000000,0x34e00000,0x600acf00,0x4000000,0xb6800000,0x600ad111,0x7c40300,0xe30000,0x604ac4bf,0x4000000,0x34e00003,0x61000a03,0x4000000,0x1600000,0x61000c02,
0x80000000,0x218960,0x6100120f,0x4000000,0x200000,0x61001a18,0x7c00100,0x1830000,0x61001d0c,0x7c00100,0x230400,0x61001d0c,0x7c00100,0x250400,0x61006600,0x24000020,
0x200000,0x61008407,0x7c00100,0x220400,0x61008407,0x7c00100,0x250400,0x6100870c,0x7c00100,0x220400,0x61008e00,0x24000000,0x200000,0x61008e00,0x24000000,0x400000,
0x61008e00,0x24000002,0x300000,0x6100903c,0x7c00100,0x230400,0x61009519,0x7c00100,0x220400,0x61009519,0x7c00100,0x250400,0x61009519,0x7c00500,0x22040f,0x61009b71,
0x2802100,0x962460,0x61009b71,0x2806400,0x962460,0x61009b71,0x7c00100,0x230400,0x6100a008,0x2802100,0x962460,0x6100c300,0x4000000,0x20000f,0x6100cd00,0x4000000,
0x200000,0x6100d202,0x2802400,0x962460,0x6100d202,0x2802500,0x962460,0x6100d202,0x7c00100,0x230400,0x6100d302,0x4000020,0x200000,0x6100d302,0x7c00120,0x230405,
0x6100d476,0x2802100,0x962460,0x6100d476,0x2802100,0x962461,0x6100d476,0x2806400,0x962460,0x6100d476,0x4000000,0x400000,0x6100d476,0x6800000,0x1329800,0x6100d476,
0x6800100,0x962540,0x6100d476,0x7c00100,0x230400,0x6100d476,0xc000010,0x448000,0x6100d573,0x2802100,0x962460,0x6100d573,0x2806400,0x962460,0x6100d573,0x6800100,
0x962540,0x6100d573,0x7c00100,0x230400,0x6100d573,0x7c00900,0x230400,0x6100d573,0xc000010,0x448000,0x6100d68d,0x7c00100,0x230400,0x6100d756,0x7c00100,0x230400,
0x6100d85c,0x2802500,0x962460,0x6100d85c,0x6800100,0x962540,0x6100d85c,0x7c00100,0x230400,0x6100d85c,0x7c00500,0x230400,0x6100d997,0x2802100,0x962460,0x6100d997,
0x4000000,0x200000,0x6100d997,0x4000000,0x400000,0x6100d997,0x6800000,0x1329800,0x6100d997,0x6800100,0x962540,0x6100d997,0x6804400,0x962540,0x6100d997,0x7c00100,
0x230400,0x6100d997,0x7c00100,0x230560,0x6100d997,0xc000010,0x448000,0x6100da98,0x6800000,0x1329800,0x6100da98,0x7c00100,0x230400,0x6100db71,0x4000000,0x200000,
0x6100dc99,0x2802100,0x962460,0x6100dc99,0x2802400,0x962460,0x6100dc99,0x6800000,0x1329800,0x6100dc99,0x6800100,0x962540,0x6100dc99,0x6804400,0x962540,0x6100dc99,
0x7c00100,0x230400,0x610a4711,0x7c40300,0xe30000,0x610a4f11,0x7c00300,0xe30001,0x610ace00,0x4000000,0x34e00000,0x6140af96,0x7c00100,0x230400,0x6140af98,0x7c00100,
0x230400,0x6180af93,0x2802400,0x962460,0x62002a00,0x4000000,0x1600000,0x63002800,0x80000,0x918820,0x63c00c14,0x80000,0x918820,0x7000080e,0x7c00100,0x250400,
0x70000a03,0x4000000,0x200000,0x70000c00,0x80000000,0x218960,0x70000f0a,0x7c00100,0x230400,0x70001004,0x7c00100,0x230400,0x70001524,0x2802100,0x962460,0x70001524,
0x7c00100,0x230400,0x70001615,0x2802100,0x962460,0x7000171a,0x2802100,0x962460,0x70001821,0x6800000,0x1329800,0x70002320,0x7c00100,0x230400,0x70002a00,0x4000000,
0x1500000,0x70002a00,0x4000000,0x1600000,0x70003000,0x24000000,0x200000,0x70003000,0x24000000,0x14200000,0x70003800,0x24000000,0x4e00000,0x70005201,0x2802400,0x962460,
0x7000581e,0x7c00100,0x230400,0x70006108,0x7c00100,0x220400,0x70006108,0x7c00100,0x250400,0x70006f30,0x7c00100,0x230400,0x70007300,0x24000000,0x200000,0x70007f0e,
0x4000000,0x200000,0x70008301,0x2802100,0x962460,0x70008301,0x2802400,0x962460,0x70008e00,0x24000000,0x200000,0x70008e00,0x24000000,0x400000,0x70008e00,0x24000002,
0x400000,0x70008e00,0x24000008,0x1410000,0x70008e00,0x24000010,0x400000,0x70008e00,0x2c000010,0x448000,0x70009519,0x7c00100,0x220400,0x70009519,0x7c00100,0x230400,
0x70009519,0x7c00100,0x250400,0x70009865,0x7c00100,0x230400,0x70009965,0x4000010,0x400000,0x70009965,0x7c00100,0x230400,0x7000a008,0x7c00100,0x220400,0x7000a008,
0x7c00100,0x250400,0x7000a008,0x7c00500,0x22040f,0x7000a50e,0x4000000,0x200000,0x7000b61c,0x2802500,0x1862460,0x7000b61c,0x6800500,0x1862400,0x7000b61c,0x7c00100,
0x1830000,0x7000c300,0x4000000,0x100000,0x7000c941,0x2806000,0x962460,0x7000cc00,0x4000000,0x4e00000,0x7000cd00,0x4000000,0x200000,0x7000cd00,0x4000000,0x4200000,
0x7000cd00,0x4000000,0x4e00000,0x7000cd00,0x4000000,0x14200000,0x7000cd00,0x4000000,0x14e00000,0x7000cd00,0x4000000,0x96800000,0x7000cf00,0x4000000,0x4e00000,0x7000cf00,
0x4000000,0x14e00000,0x7000d202,0x2802100,0x962460,0x7000d202,0x7c00100,0x230400,0x7000d997,0x7c00100,0x230400,0x7000d997,0xc000010,0x248000,0x7000dd86,0x2802400,
0x962460,0x7000dd86,0x7c00100,0x230400,0x7000dd86,0xc000010,0x448000,0x7000de9f,0x4000000,0x200000,0x7000de9f,0x7c00100,0x230400,0x7000e001,0x2000,0x962460,
0x7000e001,0x2802400,0x962460,0x7000e187,0x2802000,0x962460,0x7000e187,0x2802100,0x962460,0x7000e187,0x4000000,0x200000,0x7000e187,0x7c00100,0x230400,0x7000e187,
0xc000010,0x448000,0x7000e288,0x7c00100,0x230400,0x7000e300,0x4000000,0x200000,0x7000e489,0x2802100,0x962460,0x7000e489,0x2802400,0x962460,0x7000e489,0x6800100,
0x962540,0x7000e489,0x6800100,0x962541,0x7000e489,0x6804400,0x962540,0x7000e489,0x7c00100,0x230400,0x7000e489,0x7c00900,0x230400,0x7000e59d,0x2802100,0x962460,
0x7000e59d,0x2802400,0x962460,0x7000e59d,0x4000000,0x200000,0x7000e59d,0x4000010,0x200000,0x7000e59d,0x6800100,0x962540,0x7000e59d,0x6804400,0x962540,0x7000e59d,
0x7c00100,0x230400,0x7000e59d,0xc000010,0x448000,0x7000e691,0x2802100,0x962460,0x7000e691,0x2802400,0x962460,0x7000e691,0x2806400,0x962460,0x7000e691,0x6800000,
0x1329800,0x7000e691,0x6800100,0x962540,0x7000e691,0x7c00100,0x230400,0x7000e700,0x4000400,0x200400,0x7000e70e,0x7c00100,0x220400,0x7000e719,0x7c00100,0x220400,
0x7000e719,0x7c00500,0x22040f,0x7000e853,0x7c00100,0x230400,0x7000e9a0,0x2802400,0x962460,0x7000e9a0,0x4000000,0x200000,0x7000e9a0,0x4000000,0x500000,0x7000e9a0,
0x7c00100,0x230400,0x7000ea79,0x2802400,0x962460,0x7000ea79,0x4000000,0x200000,0x7000ea79,0x4000000,0xf00000,0x7000ea79,0x4000010,0x400000,0x7000ea79,0x7c00100,
0x230400,0x7000eb8c,0x2802400,0x962460,0x7000eb8c,0x4000000,0x200000,0x7000eb8c,0x7c00100,0x230400,0x7000eca3,0x2802100,0x962460,0x7000eca3,0x2806400,0x962460,
0x7000eca3,0x4000000,0x200000,0x7000eca3,0x6800000,0x1329800,0x7000eca3,0x6800100,0x962540,0x7000eca3,0x7c00100,0x230400,0x7000eca3,0xc000010,0x448000,0x7000ed95,
0x6800000,0x1329800,0x7000ed95,0x7c00100,0x230400,0x7000ed95,0xc000010,0x448000,0x7000ee1c,0x2802500,0x1862460,0x7000ee1c,0x6800000,0x1329800,0x7000ee1c,0x7c00100,
0x1830000,0x7000ee1c,0x7c00900,0x1830000,0x7000ef8f,0x4000000,0x200000,0x7000ef8f,0x7c00100,0x230400,0x7000f08e,0x4000000,0x200000,0x7000f08e,0x7c00100,0x230400,
0x7000f159,0x2802100,0x962460,0x7000f159,0x7c00100,0x230400,0x7000f200,0x4000000,0x200000,0x7000f200,0x4000000,0x1200000,0x7000f200,0x4000000,0x1710000,0x7000f34b,
0x2802400,0x962460,0x7000f34b,0x4000000,0x200000,0x7000f34b,0x4000010,0x400000,0x7000f34b,0x6800000,0x1329800,0x7000f34b,0x7c00100,0x230400,0x7000f34b,0x7c00900,
0x230400,0x7000f34b,0xc000010,0x448000,0x7000f490,0x4000000,0x200000,0x7000f490,0x7c00100,0x230400,0x7000f5a5,0x7c00100,0x230400,0x7000f67b,0x4000000,0x200000,
0x7000f67b,0x4000010,0x200000,0x7000f67b,0x7c00100,0x230400,0x7000f8a6,0x2802100,0x962460,0x7000f8a6,0x2802400,0x962460,0x7000f8a6,0x2806400,0x962460,0x7000f8a6,
0x4000000,0x500000,0x7000f8a6,0x4000010,0xb00000,0x7000f8a6,0x4000800,0x200000,0x7000f8a6,0x6800100,0x962540,0x7000f8a6,0x6800100,0x962541,0x7000f8a6,0x7c00100,
0x230400,0x7000f8a6,0xc000010,0x448000,0x7000f921,0x4000000,0x200000,0x7000fa00,0x4000000,0x200000,0x7000fb9e,0x2802100,0x962460,0x7000fb9e,0x2802400,0x962460,
0x7000fb9e,0x2806400,0x962460,0x7000fb9e,0x4000000,0x200000,0x7000fb9e,0x6800000,0x1329800,0x7000fb9e,0x6800100,0x962540,0x7000fb9e,0x6800100,0x962541,0x7000fb9e,
0x7c00100,0x230400,0x7000fc92,0x4000000,0x200000,0x7000fc92,0x6800000,0x1329800,0x7000fc92,0x7c00100,0x220400,0x7000fc92,0x7c00100,0x230400,0x7000fc92,0x7c00100,
0x250400,0x700acd00,0x4000000,0x34e00000,0x700acd00,0x4000000,0xb6800000,0x700ace00,0x4000000,0x34e00000,0x700acf00,0x4000000,0x34e00000,0x700acf00,0x4000000,0xb6800000,
0x7050df01,0x4000000,0x200000,0x7050f705,0x80000,0x918820,0x7080af96,0x2802400,0x962460,0x7090df01,0x2802400,0x962460,0x70d0e403,0x2802100,0x962460,0x70d0e403,
0x2802400,0x962460,0x70d0e403,0x6800100,0x962540,0x8000120f,0x7c00100,0x230400,0x80001524,0x7c00100,0x230400,0x8000171a,0x7c00100,0x230400,0x80002006,0x7c00100,
0x220400,0x80002006,0x7c00100,0x250400,0x80002a00,0x4000000,0x1500000,0x80002d00,0x4000000,0x200000,0x80005208,0x2802400,0x962460,0x80005c00,0x4000000,0x200000,
0x80007300,0x24000000,0x200000,0x80009519,0x7c00100,0x220400,0x80009519,0x7c00100,0x230400,0x80009519,0x7c00100,0x250400,0x80009865,0x7c00100,0x230400,0x8000a008,
0x2802100,0x962460,0x8000b30a,0x4000000,0x500000,0x8000b30a,0x7c00100,0x230400,0x8000cd00,0x4000000,0x4e00000,0x8000d202,0x2802500,0x962460,0x8000d202,0x7c00100,
0x230400,0x8000d68d,0x4000000,0x200000,0x8000d997,0x2802000,0x962460,0x8000d997,0x2802400,0x962460,0x8000d997,0x4000000,0x400000,0x8000d997,0x4000000,0x500000,
0x8000d997,0x7c00100,0x230400,0x8000d997,0xc000010,0x448000,0x8000e489,0x2802100,0x962460,0x8000e489,0x7c00100,0x230400,0x8000e719,0x7c00100,0x220400,0x8000f8a6,
0x2802100,0x962460,0x8000f8a6,0x7c00100,0x230400,0x8000f8a6,0xc000010,0x448000,0x8000fda1,0x2802100,0x1862460,0x8000fda1,0x2806400,0x1862460,0x8000fda1,0x4000000,
0x1800000,0x8000fda1,0x6800000,0x1329800,0x8000fda1,0x6800100,0x1862540,0x8000fda1,0x7c00100,0x1830000,0x8000fda1,0xc000010,0x448000,0x8000fe9c,0x7c00100,0x230400,
0x8000fe9c,0x7c00100,0x830400,0x8000fe9c,0x7c00100,0x1430400,0x8000ff06,0x7c00100,0x220400,0x80010165,0x7c00100,0x230400,0x800102a2,0x4000000,0x200000,0x800102a2,
0x7c00100,0x230400,0x800103a4,0x7c00100,0x230400,0x800103a4,0xc000010,0x448000,0x8001044c,0x4000000,0x200000,0x8001044c,0x7c00100,0x220400,0x8001044c,0x7c00100,
0x250400,0x80010670,0x2802000,0x962460,0x80010670,0x4000000,0x200000,0x80010670,0x4000010,0x400000,0x80010670,0xc000010,0x448000,0x800a4711,0x7c40300,0xe30000,
0x800acd00,0x4000000,0x34e00000,0x800acd00,0x4000000,0x7a902460,0x800ace00,0x4000000,0x34e00000,0x800acf00,0x4000000,0x34e00000,0x800b0011,0x7c40300,0xe30000,0x800b0500,
0x4000000,0x34e00000,0x800b0500,0x4000000,0xb6800000,0x90001615,0x7c00100,0x230400,0x9000171a,0x4000000,0x200000,0x9000171a,0x7c00100,0x230400,0x90003000,0x24000000,
0x200000,0x90007f0e,0x4000000,0x200000,0x90008301,0x2802000,0x962460,0x90008e00,0x24000000,0x400000,0x90009519,0x7c00100,0x250400,0x9000a16f,0x2802100,0x962460,
0x9000d200,0x80000000,0x218960,0x9000d202,0x2802000,0x962460,0x9000d202,0x2802100,0x962460,0x9000d202,0x7c00100,0x230400,0x9000e59d,0x2802100,0x962460,0x90010500,
0x4000000,0xe00000,0x900107a7,0x2802100,0x962460,0x900107a7,0x2802400,0x962460,0x900107a7,0x2802c00,0x962460,0x900107a7,0x4000000,0x1400000,0x900107a7,0x6800000,
0x1329800,0x900107a7,0x7c00100,0x220400,0x900107a7,0x7c00100,0x250400,0x900108a8,0x2802100,0x962460,0x900108a8,0x2806400,0x962460,0x900108a8,0x4000000,0x200000,
0x900108a8,0x4000000,0x400000,0x900108a8,0x4000010,0x400000,0x900108a8,0x6800000,0x1329800,0x900108a8,0x6800100,0x962540,0x900108a8,0x7c00100,0x230400,0x900108a8,
0xc000010,0x448000,0x90010908,0x7c00100,0x220400,0x90010a38,0x2802100,0x962460,0x90010ca9,0x2802100,0x962460,0x90010ca9,0x4000000,0x500000,0x90010ca9,0x4000010,
0xb00000,0x90010ca9,0x6800100,0x962540,0x90010ca9,0x7c00100,0x230400,0x90010d1b,0x4000000,0x500000,0x90010eaa,0x2802100,0x962460,0x90010eaa,0x2802400,0x962460,
0x90010eaa,0x2806400,0x962460,0x90010eaa,0x4000000,0x200000,0x90010eaa,0x4000000,0x400000,0x90010eaa,0x4000010,0x400000,0x90010eaa,0x6800000,0x1329800,0x90010eaa,
0x6800100,0x962540,0x90010eaa,0x7c00100,0x230400,0x90010eaa,0xc000010,0x448000,0x90010fab,0x7c00100,0x220400,0x90010fab,0x7c00100,0x250400,0x9002c300,0x4000000,
0x100000,0x900ac400,0x4000000,0xe0000d,0x900acd00,0x4000000,0x34e00000,0x900acd00,0x4000000,0xb6800000,0x900acf00,0x4000000,0x34e00000,0x900b0500,0x4000000,0x34e00000,
0x900b0500,0x4000000,0xb6800000,0x900b0b9a,0x7c00900,0x1230400,0x900b109a,0x7c00300,0xe30000,0x900b119a,0x7c00300,0xe30000,0x90408e06,0x24000000,0x400000,0xa0001004,
0x4000000,0x200000,0xa0001004,0x7c00100,0x230400,0xa000120f,0x2802100,0x962460,0xa000120f,0x2802400,0x962460,0xa000171a,0x2802100,0x962460,0xa000171a,0x2806400,
0x962460,0xa0002a00,0x4000000,0x1600000,0xa0003000,0x24000000,0x200000,0xa000581e,0x7c00100,0x230400,0xa0007300,0x24000000,0x200000,0xa0008301,0x2802400,0x962460,
0xa0008e00,0x24000000,0x400000,0xa000cf00,0x4000000,0x4e00000,0xa0010500,0x4000000,0x200000,0xa00114af,0x2802100,0x962460,0xa00114af,0x2802400,0x962460,0xa00114af,
0x2806400,0x962460,0xa00114af,0x6800000,0x1329800,0xa00114af,0x7c00100,0x230400,0xa00114af,0x7c00100,0x230560,0xa00116b0,0x2802100,0x962460,0xa00116b0,0x2802800,
0x962460,0xa00116b0,0x2806400,0x962460,0xa00116b0,0x4000000,0x400000,0xa00116b0,0x4000000,0x500000,0xa00116b0,0x4000010,0x400000,0xa00116b0,0x6800100,0x962540,
0xa00116b0,0x7c00100,0x230400,0xa00116b0,0x7c00100,0x230560,0xa00116b0,0xc000010,0x448000,0xa0011722,0x7c00100,0x230400,0xa00118b1,0x2802000,0x962460,0xa00118b1,
0x2802100,0x962460,0xa00118b1,0x2806400,0x962460,0xa00118b1,0x4000000,0x200000,0xa00118b1,0x4000000,0x400000,0xa00118b1,0x4000000,0x500000,0xa00118b1,0x6800100,
0x962540,0xa00118b1,0x7c00100,0x230400,0xa00118b1,0x7c00100,0x230560,0xa00118b1,0xc000010,0x448000,0xa00a4005,0x7c00100,0xe30400,0xa00a4711,0x7c40300,0xe30000,
0xa00ac400,0x4000000,0x4e00000,0xa00acb14,0x7c00100,0xe30000,0xa00acf00,0x4000000,0x34e00000,0xa00b0500,0x4000000,0x34e00000,0xa00b0500,0x4000000,0xb6800000,0xa00b0b96,
0x7c00900,0x1230400,0xa00b1211,0x7c40300,0xe30000,0xa00b1314,0x7c00100,0xe30000,0xa00b1596,0x7c00300,0xe30000,0xa040afac,0x6800400,0x962540,0xa08083ad,0x2802400,
0x962460,0xb0000a03,0x7c00100,0x220400,0xb0000b13,0x7c00100,0x2633800,0xb0001004,0x2802000,0x962460,0xb0001110,0x4000000,0x200000,0xb0001524,0x2802000,0x962460,
0xb0001615,0x4000000,0x500000,0xb000251b,0x7c00100,0x230400,0xb0007300,0x24000000,0x200000,0xb0008939,0x4000000,0x200000,0xb0008939,0x7c00100,0x230400,0xb0008e00,
0x24000000,0x200000,0xb0008e00,0x24000000,0x400000,0xb0008e00,0x24000010,0x400000,0xb0009257,0x2802000,0x962460,0xb0009257,0x4000000,0x1600000,0xb0009519,0x7c00100,
0x220400,0xb0009519,0x7c00100,0x250400,0xb0009a00,0x4000000,0x200000,0xb000b30a,0x2802100,0x962460,0xb000b30a,0x7c00100,0x230400,0xb000c178,0x80000000,0x218960,
0xb000c300,0x4000000,0x4200000,0xb000d202,0x2802000,0x962460,0xb000d476,0x6800100,0x962540,0xb000d476,0x7c00100,0x230400,0xb000e300,0x4000000,0x4e00000,0xb000fda1,
0x7c00100,0x1830000,0xb0010eaa,0x2802000,0x962460,0xb00116b0,0x7c00100,0x230400,0xb0011900,0x4000000,0x4e00000,0xb0011ab2,0x2802100,0x962460,0xb0011ab2,0x2802400,
0x962460,0xb0011ab2,0x2806400,0x962460,0xb0011ab2,0x4000000,0x200000,0xb0011ab2,0x6800100,0x962540,0xb0011ab2,0x7c00100,0x230400,0xb0011b0c,0x7c00100,0x230400,
0xb0011cb3,0x2802100,0x962460,0xb0011cb3,0x2806400,0x962460,0xb0011cb3,0x6800000,0x1329800,0xb0011cb3,0x6800100,0x962540,0xb0011cb3,0x7c00100,0x230400,0xb0011db6,
0x2802500,0x962460,0xb0011db6,0x6800000,0x1329800,0xb0011db6,0x7c00100,0x230400,0xb0011db6,0x7c00500,0x230400,0xb0011e00,0x4000000,0x200000,0xb0011e00,0x4000000,
0x1500000,0xb0011fb4,0x2802100,0x962460,0xb0011fb4,0x6800100,0x962540,0xb0011fb4,0x7c00100,0x230400,0xb0011fb4,0xc000010,0x248000,0xb0012000,0x4000000,0x200000,
0xb00121b5,0x4000000,0x200000,0xb00121b5,0x4000010,0x400000,0xb00121b5,0x7c00100,0x220400,0xb00121b5,0x7c00100,0x250400,0xb00121b5,0xc000010,0x448000,0xb00122b8,
0x4000000,0x200000,0xb00122b8,0x7c00100,0x230400,0xb00123b7,0x2802400,0x962460,0xb00123b7,0x4000000,0x200000,0xb00123b7,0x7c00100,0x230400,0xb00123b7,0xc000010,
0x248000,0xb00a4005,0x7c00100,0xe30400,0xb00a4711,0x7c40300,0xe30000,0xb00acf00,0x4000000,0x34e00000,0xb00b0500,0x4000000,0x34e00000,0xb00b0500,0x4000000,0x3ce00000,
0xb00b0500,0x4000000,0xb6800000,0xb00b109a,0x7c00300,0xe30000,0xb080e47c,0x2802000,0x962460,0xc0001524,0x4000000,0x500000,0xc0001a18,0x2806400,0x1862460,0xc0001a18,
0x7c00100,0x1830000,0xc0007300,0x24000000,0x200000,0xc0008e00,0x24000010,0x400000,0xc0009519,0x7c00100,0x220400,0xc0009519,0x7c00100,0x250400,0xc000c300,0x4000000,
0x420000f,0xc000d85c,0x2802100,0x962460,0xc000d85c,0x6800100,0x962540,0xc000d85c,0x7c00100,0x230400,0xc000dc99,0x7c00100,0x230400,0xc000e719,0x7c00100,0x220400,
0xc00107a7,0x7c00100,0x230400,0xc0010eaa,0x7c00100,0x230400,0xc00116b0,0x7c00100,0x230560,0xc0011900,0x4000000,0x4200000,0xc0012447,0,0x818820,0xc0012447,
0,0xc18820,0xc0012447,0,0x1418820,0xc00125b9,0x7c00100,0x230400,0xc00126bb,0x2802100,0x962460,0xc00126bb,0x2806400,0x962460,0xc00126bb,0x4000000,
0x500000,0xc00126bb,0x6800100,0x962540,0xc00126bb,0x7c00100,0x230400,0xc00127ba,0x2802400,0x962460,0xc00127ba,0x4000000,0x200000,0xc00127ba,0x6800000,0x1329800,
0xc00127ba,0x7c00100,0x230400,0xc00127ba,0x7c00900,0x230400,0xc0012800,0x4000000,0x200000,0xc0012b23,0x4000000,0x200000,0xc0012b23,0x4000000,0x400000,0xc0012b23,
0x4000000,0x1500000,0xc0012cbc,0x2802400,0x962460,0xc0012cbc,0x4000000,0x1600000,0xc0012cbc,0x6800000,0x1329800,0xc0012cbc,0x7c00100,0x230400,0xc00acf00,0x4000000,
0x34e00000,0xc00ae300,0x4000000,0x34e00000,0xc00b0500,0x4000000,0x34e00000,0xc00b0500,0x4000000,0xb6800000,0xc00b0b00,0x4000000,0x1200000,0xc00b0b00,0x7c00900,0x1230400,
0xc00b109a,0x7c00300,0xe30000,0xc00b2914,0x7c00100,0x2530000,0xc00b2916,0x7c00100,0x2530c00,0xc00b2a00,0x4000000,0x34e00000,0xc040af53,0x7c00100,0x230400,0xc0c12b7e,
0x4000000,0x200000,0xc14a44bf,0x4000000,0xe0000d,0xd000131f,0x2802c00,0x962460,0xd000171a,0x7c00100,0x230400,0xd0001821,0x2802100,0x962460,0xd0007300,0x24000000,
0x200000,0xd0008e00,0x24000000,0x200000,0xd0008f3a,0x2806000,0x962460,0xd0009519,0x7c00100,0x220400,0xd0009519,0x7c00100,0x250400,0xd000a500,0x4000000,0x200000,
0xd000c300,0x4000000,0x4e00000,0xd000d202,0x7c00100,0x230400,0xd000d476,0x7c00100,0x230400,0xd000d997,0x2802100,0x962460,0xd000d997,0x6800100,0x962540,0xd000e001,
0x2802100,0x962460,0xd000e700,0x4000400,0x200000,0xd000e719,0x7c00100,0x220400,0xd000e719,0x7c00500,0x23040f,0xd000fa00,0x4000000,0x4e00000,0xd0010eaa,0x4000010,
0x400000,0xd0010eaa,0x7c00100,0x230400,0xd0012dbd,0x4000000,0x200000,0xd0012dbd,0x7c00100,0x230400,0xd0012fbe,0x2802100,0x962460,0xd0012fbe,0x2802400,0x962460,
0xd0012fbe,0x2806400,0x962460,0xd0012fbe,0x4000000,0x400000,0xd0012fbe,0x6800000,0x1329800,0xd0012fbe,0x6800100,0x962540,0xd0012fbe,0x6800100,0x962541,0xd0012fbe,
0x6804400,0x962540,0xd0012fbe,0x7c00100,0x230400,0xd0012fbe,0x7c00100,0x230560,0xd0012fbe,0xc000010,0x448000,0xd0013183,0x7c00100,0x230400,0xd0013200,0x4000000,
0x200000,0xd0013200,0x6800000,0x1329805,0xd00134c0,0x2802100,0x962460,0xd00134c0,0x4000002,0x400000,0xd00134c0,0x7c00100,0x230400,0xd00a4305,0x7c00100,0xe30400,
0xd00a4611,0x7c40300,0xe30000,0xd00a4711,0x7c40300,0xe30000,0xd00a5e11,0x7c40300,0xe30000,0xd00acf00,0x4000000,0x34e00000,0xd00b0500,0x4000000,0x34e00000,0xd00b0500,
0x4000000,0xb6800000,0xd00b0b11,0x6800500,0x962540,0xd00b0bbf,0x2802200,0xc62460,0xd00b119a,0x7c00300,0xe30000,0xd00b2a00,0x4000000,0x34e00000,0xd00b2e11,0x7c40300,
0xe30000,0xd00b30bf,0x7c00300,0x230000,0xd00b339a,0x7c00300,0xe30000};

// static const int32_t countPropsVectors=6999;
static const int32_t propsVectorsColumns=3;
static const uint16_t scriptExtensions[262]={
0x800e,0x8019,8,0x8059,8,2,8,0x8038,8,6,8,0x8019,2,0x22,0x25,0xb6,
0x80c0,2,0x22,0x8025,2,0x11,2,0x22,0x54,0x79,0x7b,0xa7,0xb6,0x80b7,2,0x8022,
2,0x25,0x80c0,2,0x20,2,0x80b6,4,0xa,0xf,0x10,0x15,0x19,0x1a,0x1f,0x23,
0x24,0x89,0x97,0x809e,4,0xa,0xf,0x10,0x15,0x19,0x1a,0x1f,0x23,0x24,0x89,0x809e,
4,0xa,0xf,0x10,0x15,0x1a,0x1f,0x21,0x23,0x24,0x3a,0x89,0x91,0x99,0x9e,0xa0,
0xaf,0xb2,0xb3,0x80bb,4,0xa,0xf,0x10,0x15,0x1a,0x1f,0x21,0x23,0x24,0x30,0x3a,
0x89,0x91,0x99,0x9e,0xa0,0xaf,0xb2,0xb3,0x80bb,0xa,0x78,0xa0,0x80b2,0xa,0x69,4,
0x3a,0x8076,4,0x6f,0x10,0x80a4,0x10,0x74,0xf,0x809d,0xf,0x78,0x23,0x8089,0x23,0x7c,
0x15,0x80bb,0x15,0x80,0x1c,0x34,0x8076,0x1c,0x84,0xc,0x8019,0x2a,0x2b,0x2c,0x802d,0x1b,
0x805a,0x800a,4,0xa,0x15,0x8089,0xa,0x8089,4,0x800a,0xa,0x8097,0xa,0x15,0x1a,0x1f,
0x23,0x8024,0xa,0x80bb,4,0xa,0x15,0x1f,0x24,0x89,0x9e,0x80bb,0x8004,8,0x8022,0x19,
0x801b,0xa,0x19,0x8089,5,0x11,0x12,0x14,0x16,0x8029,5,0x11,0x12,0x14,0x8016,0x8011,
5,0x8011,0x11,0x14,0x8016,0x11,0x8019,0xa,0xf,0x10,0x78,0x91,0x99,0x9d,0x9e,0xa0,
0xa3,0x80b2,0xa,0xf,0x10,0x15,0x1a,0x78,0x91,0x99,0x9d,0x9e,0xa0,0xa3,0xb2,0x80bb,
0xa,0xf,0x10,0x15,0x78,0x91,0x99,0x9d,0x9e,0xa0,0xa3,0xb2,0x80bb,0xa,0x98,0xa,
0x8023,0xa,0xef,0x19,0x1c,0x804f,0x37,0x804e,2,0x8025,2,0xf8,0x2f,0x31,0x8053,0x2f,
0x8031,2,0x8007,0x89,0x7c,0x8087};

static const int32_t indexes[UPROPS_INDEX_COUNT]={0x2b96,0x2b96,0x2b96,0x2b96,0x6898,3,0x83ef,0x8472,0x8472,0x8472,0xb34c0,0x2a75a31,0,0,0,0};

#endif  // INCLUDED_FROM_UCHAR_C


/* constants and macros for access to the data ------------------------------ */

/* getting a uint32_t properties word from the data */
#define GET_PROPS(c, result) ((result)=UTRIE2_GET16(&propsTrie, c))

/* API functions ------------------------------------------------------------ */

/* Gets the Unicode character's general category.*/
U_CAPI int8_t U_EXPORT2
u_charType(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (int8_t)GET_CATEGORY(props);
}

/* Enumerate all code points with their general categories. */
struct _EnumTypeCallback {
    UCharEnumTypeRange *enumRange;
    const void *context;
};

static uint32_t U_CALLCONV
_enumTypeValue(const void *context, uint32_t value) {
    (void)context;
    return GET_CATEGORY(value);
}

static UBool U_CALLCONV
_enumTypeRange(const void *context, UChar32 start, UChar32 end, uint32_t value) {
    /* just cast the value to UCharCategory */
    return ((struct _EnumTypeCallback *)context)->
        enumRange(((struct _EnumTypeCallback *)context)->context,
                  start, end+1, (UCharCategory)value);
}

U_CAPI void U_EXPORT2
u_enumCharTypes(UCharEnumTypeRange *enumRange, const void *context) {
    struct _EnumTypeCallback callback;

    if(enumRange==NULL) {
        return;
    }

    callback.enumRange=enumRange;
    callback.context=context;
    utrie2_enum(&propsTrie, _enumTypeValue, _enumTypeRange, &callback);
}

/* Checks if ch is a lower case letter.*/
U_CAPI UBool U_EXPORT2
u_islower(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)(GET_CATEGORY(props)==U_LOWERCASE_LETTER);
}

/* Checks if ch is an upper case letter.*/
U_CAPI UBool U_EXPORT2
u_isupper(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)(GET_CATEGORY(props)==U_UPPERCASE_LETTER);
}

/* Checks if ch is a title case letter; usually upper case letters.*/
U_CAPI UBool U_EXPORT2
u_istitle(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)(GET_CATEGORY(props)==U_TITLECASE_LETTER);
}

/* Checks if ch is a decimal digit. */
U_CAPI UBool U_EXPORT2
u_isdigit(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)(GET_CATEGORY(props)==U_DECIMAL_DIGIT_NUMBER);
}

U_CAPI UBool U_EXPORT2
u_isxdigit(UChar32 c) {
    uint32_t props;

    /* check ASCII and Fullwidth ASCII a-fA-F */
    if(
        (c<=0x66 && c>=0x41 && (c<=0x46 || c>=0x61)) ||
        (c>=0xff21 && c<=0xff46 && (c<=0xff26 || c>=0xff41))
    ) {
        return TRUE;
    }

    GET_PROPS(c, props);
    return (UBool)(GET_CATEGORY(props)==U_DECIMAL_DIGIT_NUMBER);
}

/* Checks if the Unicode character is a letter.*/
U_CAPI UBool U_EXPORT2
u_isalpha(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)((CAT_MASK(props)&U_GC_L_MASK)!=0);
}

U_CAPI UBool U_EXPORT2
u_isUAlphabetic(UChar32 c) {
    return (u_getUnicodeProperties(c, 1)&U_MASK(UPROPS_ALPHABETIC))!=0;
}

/* Checks if c is a letter or a decimal digit */
U_CAPI UBool U_EXPORT2
u_isalnum(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)((CAT_MASK(props)&(U_GC_L_MASK|U_GC_ND_MASK))!=0);
}

/**
 * Checks if c is alphabetic, or a decimal digit; implements UCHAR_POSIX_ALNUM.
 * @internal
 */
U_CFUNC UBool
u_isalnumPOSIX(UChar32 c) {
    return (UBool)(u_isUAlphabetic(c) || u_isdigit(c));
}

/* Checks if ch is a unicode character with assigned character type.*/
U_CAPI UBool U_EXPORT2
u_isdefined(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)(GET_CATEGORY(props)!=0);
}

/* Checks if the Unicode character is a base form character that can take a diacritic.*/
U_CAPI UBool U_EXPORT2
u_isbase(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)((CAT_MASK(props)&(U_GC_L_MASK|U_GC_N_MASK|U_GC_MC_MASK|U_GC_ME_MASK))!=0);
}

/* Checks if the Unicode character is a control character.*/
U_CAPI UBool U_EXPORT2
u_iscntrl(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)((CAT_MASK(props)&(U_GC_CC_MASK|U_GC_CF_MASK|U_GC_ZL_MASK|U_GC_ZP_MASK))!=0);
}

U_CAPI UBool U_EXPORT2
u_isISOControl(UChar32 c) {
    return (uint32_t)c<=0x9f && (c<=0x1f || c>=0x7f);
}

/* Some control characters that are used as space. */
#define IS_THAT_CONTROL_SPACE(c) \
    (c<=0x9f && ((c>=uprops_TAB && c<=uprops_CR) || (c>=0x1c && c <=0x1f) || c==uprops_NL))

/* Java has decided that U+0085 New Line is not whitespace any more. */
#define IS_THAT_ASCII_CONTROL_SPACE(c) \
    (c<=0x1f && c>=uprops_TAB && (c<=uprops_CR || c>=0x1c))

/* Checks if the Unicode character is a space character.*/
U_CAPI UBool U_EXPORT2
u_isspace(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)((CAT_MASK(props)&U_GC_Z_MASK)!=0 || IS_THAT_CONTROL_SPACE(c));
}

U_CAPI UBool U_EXPORT2
u_isJavaSpaceChar(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)((CAT_MASK(props)&U_GC_Z_MASK)!=0);
}

/* Checks if the Unicode character is a whitespace character.*/
U_CAPI UBool U_EXPORT2
u_isWhitespace(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)(
                ((CAT_MASK(props)&U_GC_Z_MASK)!=0 &&
                    c!=uprops_NBSP && c!=uprops_FIGURESP && c!=uprops_NNBSP) || /* exclude no-break spaces */
                IS_THAT_ASCII_CONTROL_SPACE(c)
           );
}

U_CAPI UBool U_EXPORT2
u_isblank(UChar32 c) {
    if((uint32_t)c<=0x9f) {
        return c==9 || c==0x20; /* uprops_TAB or SPACE */
    } else {
        /* Zs */
        uint32_t props;
        GET_PROPS(c, props);
        return (UBool)(GET_CATEGORY(props)==U_SPACE_SEPARATOR);
    }
}

U_CAPI UBool U_EXPORT2
u_isUWhiteSpace(UChar32 c) {
    return (u_getUnicodeProperties(c, 1)&U_MASK(UPROPS_WHITE_SPACE))!=0;
}

/* Checks if the Unicode character is printable.*/
U_CAPI UBool U_EXPORT2
u_isprint(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    /* comparing ==0 returns FALSE for the categories mentioned */
    return (UBool)((CAT_MASK(props)&U_GC_C_MASK)==0);
}

/**
 * Checks if c is in \p{graph}\p{blank} - \p{cntrl}.
 * Implements UCHAR_POSIX_PRINT.
 * @internal
 */
U_CFUNC UBool
u_isprintPOSIX(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    /*
     * The only cntrl character in graph+blank is uprops_TAB (in blank).
     * Here we implement (blank-uprops_TAB)=Zs instead of calling u_isblank().
     */
    return (UBool)((GET_CATEGORY(props)==U_SPACE_SEPARATOR) || u_isgraphPOSIX(c));
}

U_CAPI UBool U_EXPORT2
u_isgraph(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    /* comparing ==0 returns FALSE for the categories mentioned */
    return (UBool)((CAT_MASK(props)&
                    (U_GC_CC_MASK|U_GC_CF_MASK|U_GC_CS_MASK|U_GC_CN_MASK|U_GC_Z_MASK))
                   ==0);
}

/**
 * Checks if c is in
 * [^\p{space}\p{gc=Control}\p{gc=Surrogate}\p{gc=Unassigned}]
 * with space=\p{Whitespace} and Control=Cc.
 * Implements UCHAR_POSIX_GRAPH.
 * @internal
 */
U_CFUNC UBool
u_isgraphPOSIX(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    /* \p{space}\p{gc=Control} == \p{gc=Z}\p{Control} */
    /* comparing ==0 returns FALSE for the categories mentioned */
    return (UBool)((CAT_MASK(props)&
                    (U_GC_CC_MASK|U_GC_CS_MASK|U_GC_CN_MASK|U_GC_Z_MASK))
                   ==0);
}

U_CAPI UBool U_EXPORT2
u_ispunct(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)((CAT_MASK(props)&U_GC_P_MASK)!=0);
}

/* Checks if the Unicode character can start a Unicode identifier.*/
U_CAPI UBool U_EXPORT2
u_isIDStart(UChar32 c) {
    /* same as u_isalpha() */
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)((CAT_MASK(props)&(U_GC_L_MASK|U_GC_NL_MASK))!=0);
}

/* Checks if the Unicode character can be a Unicode identifier part other than starting the
 identifier.*/
U_CAPI UBool U_EXPORT2
u_isIDPart(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)(
           (CAT_MASK(props)&
            (U_GC_ND_MASK|U_GC_NL_MASK|
             U_GC_L_MASK|
             U_GC_PC_MASK|U_GC_MC_MASK|U_GC_MN_MASK)
           )!=0 ||
           u_isIDIgnorable(c));
}

/*Checks if the Unicode character can be ignorable in a Java or Unicode identifier.*/
U_CAPI UBool U_EXPORT2
u_isIDIgnorable(UChar32 c) {
    if(c<=0x9f) {
        return u_isISOControl(c) && !IS_THAT_ASCII_CONTROL_SPACE(c);
    } else {
        uint32_t props;
        GET_PROPS(c, props);
        return (UBool)(GET_CATEGORY(props)==U_FORMAT_CHAR);
    }
}

/*Checks if the Unicode character can start a Java identifier.*/
U_CAPI UBool U_EXPORT2
u_isJavaIDStart(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)((CAT_MASK(props)&(U_GC_L_MASK|U_GC_SC_MASK|U_GC_PC_MASK))!=0);
}

/*Checks if the Unicode character can be a Java identifier part other than starting the
 * identifier.
 */
U_CAPI UBool U_EXPORT2
u_isJavaIDPart(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return (UBool)(
           (CAT_MASK(props)&
            (U_GC_ND_MASK|U_GC_NL_MASK|
             U_GC_L_MASK|
             U_GC_SC_MASK|U_GC_PC_MASK|
             U_GC_MC_MASK|U_GC_MN_MASK)
           )!=0 ||
           u_isIDIgnorable(c));
}

U_CAPI int32_t U_EXPORT2
u_charDigitValue(UChar32 c) {
    uint32_t props;
    int32_t value;
    GET_PROPS(c, props);
    value=(int32_t)GET_NUMERIC_TYPE_VALUE(props)-UPROPS_NTV_DECIMAL_START;
    if(value<=9) {
        return value;
    } else {
        return -1;
    }
}

U_CAPI double U_EXPORT2
u_getNumericValue(UChar32 c) {
    uint32_t props;
    int32_t ntv;
    GET_PROPS(c, props);
    ntv=(int32_t)GET_NUMERIC_TYPE_VALUE(props);

    if(ntv==UPROPS_NTV_NONE) {
        return U_NO_NUMERIC_VALUE;
    } else if(ntv<UPROPS_NTV_DIGIT_START) {
        /* decimal digit */
        return ntv-UPROPS_NTV_DECIMAL_START;
    } else if(ntv<UPROPS_NTV_NUMERIC_START) {
        /* other digit */
        return ntv-UPROPS_NTV_DIGIT_START;
    } else if(ntv<UPROPS_NTV_FRACTION_START) {
        /* small integer */
        return ntv-UPROPS_NTV_NUMERIC_START;
    } else if(ntv<UPROPS_NTV_LARGE_START) {
        /* fraction */
        int32_t numerator=(ntv>>4)-12;
        int32_t denominator=(ntv&0xf)+1;
        return (double)numerator/denominator;
    } else if(ntv<UPROPS_NTV_BASE60_START) {
        /* large, single-significant-digit integer */
        double numValue;
        int32_t mant=(ntv>>5)-14;
        int32_t exp=(ntv&0x1f)+2;
        numValue=mant;

        /* multiply by 10^exp without math.h */
        while(exp>=4) {
            numValue*=10000.;
            exp-=4;
        }
        switch(exp) {
        case 3:
            numValue*=1000.;
            break;
        case 2:
            numValue*=100.;
            break;
        case 1:
            numValue*=10.;
            break;
        case 0:
        default:
            break;
        }

        return numValue;
    } else if(ntv<UPROPS_NTV_FRACTION20_START) {
        /* sexagesimal (base 60) integer */
        int32_t numValue=(ntv>>2)-0xbf;
        int32_t exp=(ntv&3)+1;

        switch(exp) {
        case 4:
            numValue*=60*60*60*60;
            break;
        case 3:
            numValue*=60*60*60;
            break;
        case 2:
            numValue*=60*60;
            break;
        case 1:
            numValue*=60;
            break;
        case 0:
        default:
            break;
        }

        return numValue;
    } else if(ntv<UPROPS_NTV_FRACTION32_START) {
        // fraction-20 e.g. 3/80
        int32_t frac20=ntv-UPROPS_NTV_FRACTION20_START;  // 0..0x17
        int32_t numerator=2*(frac20&3)+1;
        int32_t denominator=20<<(frac20>>2);
        return (double)numerator/denominator;
    } else if(ntv<UPROPS_NTV_RESERVED_START) {
        // fraction-32 e.g. 3/64
        int32_t frac32=ntv-UPROPS_NTV_FRACTION32_START;  // 0..15
        int32_t numerator=2*(frac32&3)+1;
        int32_t denominator=32<<(frac32>>2);
        return (double)numerator/denominator;
    } else {
        /* reserved */
        return U_NO_NUMERIC_VALUE;
    }
}

U_CAPI int32_t U_EXPORT2
u_digit(UChar32 ch, int8_t radix) {
    int8_t value;
    if((uint8_t)(radix-2)<=(36-2)) {
        value=(int8_t)u_charDigitValue(ch);
        if(value<0) {
            /* ch is not a decimal digit, try latin letters */
            if(ch>=0x61 && ch<=0x7A) {
                value=(int8_t)(ch-0x57);  /* ch - 'a' + 10 */
            } else if(ch>=0x41 && ch<=0x5A) {
                value=(int8_t)(ch-0x37);  /* ch - 'A' + 10 */
            } else if(ch>=0xFF41 && ch<=0xFF5A) {
                value=(int8_t)(ch-0xFF37);  /* fullwidth ASCII a-z */
            } else if(ch>=0xFF21 && ch<=0xFF3A) {
                value=(int8_t)(ch-0xFF17);  /* fullwidth ASCII A-Z */
            }
        }
    } else {
        value=-1;   /* invalid radix */
    }
    return (int8_t)((value<radix) ? value : -1);
}

U_CAPI UChar32 U_EXPORT2
u_forDigit(int32_t digit, int8_t radix) {
    if((uint8_t)(radix-2)>(36-2) || (uint32_t)digit>=(uint32_t)radix) {
        return 0;
    } else if(digit<10) {
        return (UChar32)(0x30+digit);
    } else {
        return (UChar32)((0x61-10)+digit);
    }
}

/* miscellaneous, and support for uprops.cpp -------------------------------- */

U_CAPI void U_EXPORT2
u_getUnicodeVersion(UVersionInfo versionArray) {
    if(versionArray!=NULL) {
        uprv_memcpy(versionArray, dataVersion, U_MAX_VERSION_LENGTH);
    }
}

U_CFUNC uint32_t
u_getMainProperties(UChar32 c) {
    uint32_t props;
    GET_PROPS(c, props);
    return props;
}

U_CFUNC uint32_t
u_getUnicodeProperties(UChar32 c, int32_t column) {
    U_ASSERT(column>=0);
    if(column>=propsVectorsColumns) {
        return 0;
    } else {
        uint16_t vecIndex=UTRIE2_GET16(&propsVectorsTrie, c);
        return propsVectors[vecIndex+column];
    }
}

U_CFUNC int32_t
uprv_getMaxValues(int32_t column) {
    switch(column) {
    case 0:
        return indexes[UPROPS_MAX_VALUES_INDEX];
    case 2:
        return indexes[UPROPS_MAX_VALUES_2_INDEX];
    default:
        return 0;
    }
}

U_CAPI void U_EXPORT2
u_charAge(UChar32 c, UVersionInfo versionArray) {
    if(versionArray!=NULL) {
        uint32_t version=u_getUnicodeProperties(c, 0)>>UPROPS_AGE_SHIFT;
        versionArray[0]=(uint8_t)(version>>4);
        versionArray[1]=(uint8_t)(version&0xf);
        versionArray[2]=versionArray[3]=0;
    }
}

U_CAPI UScriptCode U_EXPORT2
uscript_getScript(UChar32 c, UErrorCode *pErrorCode) {
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return USCRIPT_INVALID_CODE;
    }
    if((uint32_t)c>0x10ffff) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return USCRIPT_INVALID_CODE;
    }
    uint32_t scriptX=u_getUnicodeProperties(c, 0)&UPROPS_SCRIPT_X_MASK;
    uint32_t codeOrIndex=uprops_mergeScriptCodeOrIndex(scriptX);
    if(scriptX<UPROPS_SCRIPT_X_WITH_COMMON) {
        return (UScriptCode)codeOrIndex;
    } else if(scriptX<UPROPS_SCRIPT_X_WITH_INHERITED) {
        return USCRIPT_COMMON;
    } else if(scriptX<UPROPS_SCRIPT_X_WITH_OTHER) {
        return USCRIPT_INHERITED;
    } else {
        return (UScriptCode)scriptExtensions[codeOrIndex];
    }
}

U_CAPI UBool U_EXPORT2
uscript_hasScript(UChar32 c, UScriptCode sc) {
    uint32_t scriptX=u_getUnicodeProperties(c, 0)&UPROPS_SCRIPT_X_MASK;
    uint32_t codeOrIndex=uprops_mergeScriptCodeOrIndex(scriptX);
    if(scriptX<UPROPS_SCRIPT_X_WITH_COMMON) {
        return sc==(UScriptCode)codeOrIndex;
    }

    const uint16_t *scx=scriptExtensions+codeOrIndex;
    if(scriptX>=UPROPS_SCRIPT_X_WITH_OTHER) {
        scx=scriptExtensions+scx[1];
    }
    uint32_t sc32=sc;
    if(sc32>0x7fff) {
        /* Guard against bogus input that would make us go past the Script_Extensions terminator. */
        return FALSE;
    }
    while(sc32>*scx) {
        ++scx;
    }
    return sc32==(*scx&0x7fff);
}

U_CAPI int32_t U_EXPORT2
uscript_getScriptExtensions(UChar32 c,
                            UScriptCode *scripts, int32_t capacity,
                            UErrorCode *pErrorCode) {
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(capacity<0 || (capacity>0 && scripts==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    uint32_t scriptX=u_getUnicodeProperties(c, 0)&UPROPS_SCRIPT_X_MASK;
    uint32_t codeOrIndex=uprops_mergeScriptCodeOrIndex(scriptX);
    if(scriptX<UPROPS_SCRIPT_X_WITH_COMMON) {
        if(capacity==0) {
            *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
        } else {
            scripts[0]=(UScriptCode)codeOrIndex;
        }
        return 1;
    }

    const uint16_t *scx=scriptExtensions+codeOrIndex;
    if(scriptX>=UPROPS_SCRIPT_X_WITH_OTHER) {
        scx=scriptExtensions+scx[1];
    }
    int32_t length=0;
    uint16_t sx;
    do {
        sx=*scx++;
        if(length<capacity) {
            scripts[length]=(UScriptCode)(sx&0x7fff);
        }
        ++length;
    } while(sx<0x8000);
    if(length>capacity) {
        *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
    }
    return length;
}

U_CAPI UBlockCode U_EXPORT2
ublock_getCode(UChar32 c) {
    return (UBlockCode)((u_getUnicodeProperties(c, 0)&UPROPS_BLOCK_MASK)>>UPROPS_BLOCK_SHIFT);
}

/* property starts for UnicodeSet ------------------------------------------- */

static UBool U_CALLCONV
uchar_enumPropertyStartsRange(const void *context, UChar32 start, UChar32 end, uint32_t value) {
    /* add the start code point to the USet */
    const USetAdder *sa=(const USetAdder *)context;
    sa->add(sa->set, start);
    (void)end;
    (void)value;
    return TRUE;
}

#define USET_ADD_CP_AND_NEXT(sa, cp) sa->add(sa->set, cp); sa->add(sa->set, cp+1)

U_CFUNC void U_EXPORT2
uchar_addPropertyStarts(const USetAdder *sa, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return;
    }

    /* add the start code point of each same-value range of the main trie */
    utrie2_enum(&propsTrie, NULL, uchar_enumPropertyStartsRange, sa);

    /* add code points with hardcoded properties, plus the ones following them */

    /* add for u_isblank() */
    USET_ADD_CP_AND_NEXT(sa, uprops_TAB);

    /* add for IS_THAT_CONTROL_SPACE() */
    sa->add(sa->set, uprops_CR+1); /* range uprops_TAB..uprops_CR */
    sa->add(sa->set, 0x1c);
    sa->add(sa->set, 0x1f+1);
    USET_ADD_CP_AND_NEXT(sa, uprops_NL);

    /* add for u_isIDIgnorable() what was not added above */
    sa->add(sa->set, uprops_DEL); /* range uprops_DEL..uprops_NBSP-1, uprops_NBSP added below */
    sa->add(sa->set, uprops_HAIRSP);
    sa->add(sa->set, uprops_RLM+1);
    sa->add(sa->set, uprops_INHSWAP);
    sa->add(sa->set, uprops_NOMDIG+1);
    USET_ADD_CP_AND_NEXT(sa, uprops_ZWNBSP);

    /* add no-break spaces for u_isWhitespace() what was not added above */
    USET_ADD_CP_AND_NEXT(sa, uprops_NBSP);
    USET_ADD_CP_AND_NEXT(sa, uprops_FIGURESP);
    USET_ADD_CP_AND_NEXT(sa, uprops_NNBSP);

    /* add for u_digit() */
    sa->add(sa->set, uprops_U_a);
    sa->add(sa->set, uprops_U_z+1);
    sa->add(sa->set, uprops_U_A);
    sa->add(sa->set, uprops_U_Z+1);
    sa->add(sa->set, uprops_U_FW_a);
    sa->add(sa->set, uprops_U_FW_z+1);
    sa->add(sa->set, uprops_U_FW_A);
    sa->add(sa->set, uprops_U_FW_Z+1);

    /* add for u_isxdigit() */
    sa->add(sa->set, uprops_U_f+1);
    sa->add(sa->set, uprops_U_F+1);
    sa->add(sa->set, uprops_U_FW_f+1);
    sa->add(sa->set, uprops_U_FW_F+1);

    /* add for UCHAR_DEFAULT_IGNORABLE_CODE_POINT what was not added above */
    sa->add(sa->set, uprops_WJ); /* range uprops_WJ..uprops_NOMDIG */
    sa->add(sa->set, 0xfff0);
    sa->add(sa->set, 0xfffb+1);
    sa->add(sa->set, 0xe0000);
    sa->add(sa->set, 0xe0fff+1);

    /* add for UCHAR_GRAPHEME_BASE and others */
    USET_ADD_CP_AND_NEXT(sa, uprops_CGJ);
}

U_CFUNC void U_EXPORT2
upropsvec_addPropertyStarts(const USetAdder *sa, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return;
    }

    /* add the start code point of each same-value range of the properties vectors trie */
    utrie2_enum(&propsVectorsTrie, NULL, uchar_enumPropertyStartsRange, sa);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  ucharstrie.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010nov14
*   created by: Markus W. Scherer
*/



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  ucharstrie.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010nov14
*   created by: Markus W. Scherer
*/


/**
 * \file
 * \brief C++ API: Trie for mapping Unicode strings (or 16-bit-unit sequences)
 *                 to integer values.
 */



#if U_SHOW_CPLUSPLUS_API





U_NAMESPACE_BEGIN

class Appendable;
class UCharsTrieBuilder;
class UVector32;

/**
 * Light-weight, non-const reader class for a UCharsTrie.
 * Traverses a char16_t-serialized data structure with minimal state,
 * for mapping strings (16-bit-unit sequences) to non-negative integer values.
 *
 * This class owns the serialized trie data only if it was constructed by
 * the builder's build() method.
 * The public constructor and the copy constructor only alias the data (only copy the pointer).
 * There is no assignment operator.
 *
 * This class is not intended for public subclassing.
 * @stable ICU 4.8
 */
class U_COMMON_API UCharsTrie : public UMemory {
public:
    /**
     * Constructs a UCharsTrie reader instance.
     *
     * The trieUChars must contain a copy of a char16_t sequence from the UCharsTrieBuilder,
     * starting with the first char16_t of that sequence.
     * The UCharsTrie object will not read more char16_ts than
     * the UCharsTrieBuilder generated in the corresponding build() call.
     *
     * The array is not copied/cloned and must not be modified while
     * the UCharsTrie object is in use.
     *
     * @param trieUChars The char16_t array that contains the serialized trie.
     * @stable ICU 4.8
     */
    UCharsTrie(ConstChar16Ptr trieUChars)
            : ownedArray_(NULL), uchars_(trieUChars),
              pos_(uchars_), remainingMatchLength_(-1) {}

    /**
     * Destructor.
     * @stable ICU 4.8
     */
    ~UCharsTrie();

    /**
     * Copy constructor, copies the other trie reader object and its state,
     * but not the char16_t array which will be shared. (Shallow copy.)
     * @param other Another UCharsTrie object.
     * @stable ICU 4.8
     */
    UCharsTrie(const UCharsTrie &other)
            : ownedArray_(NULL), uchars_(other.uchars_),
              pos_(other.pos_), remainingMatchLength_(other.remainingMatchLength_) {}

    /**
     * Resets this trie to its initial state.
     * @return *this
     * @stable ICU 4.8
     */
    UCharsTrie &reset() {
        pos_=uchars_;
        remainingMatchLength_=-1;
        return *this;
    }

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns the state of this trie as a 64-bit integer.
     * The state value is never 0.
     *
     * @return opaque state value
     * @see resetToState64
     * @draft ICU 65
     */
    uint64_t getState64() const {
        return (static_cast<uint64_t>(remainingMatchLength_ + 2) << kState64RemainingShift) |
            (uint64_t)(pos_ - uchars_);
    }

    /**
     * Resets this trie to the saved state.
     * Unlike resetToState(State), the 64-bit state value
     * must be from getState64() from the same trie object or
     * from one initialized the exact same way.
     * Because of no validation, this method is faster.
     *
     * @param state The opaque trie state value from getState64().
     * @return *this
     * @see getState64
     * @see resetToState
     * @see reset
     * @draft ICU 65
     */
    UCharsTrie &resetToState64(uint64_t state) {
        remainingMatchLength_ = static_cast<int32_t>(state >> kState64RemainingShift) - 2;
        pos_ = uchars_ + (state & kState64PosMask);
        return *this;
    }
#endif  /* U_HIDE_DRAFT_API */

    /**
     * UCharsTrie state object, for saving a trie's current state
     * and resetting the trie back to this state later.
     * @stable ICU 4.8
     */
    class State : public UMemory {
    public:
        /**
         * Constructs an empty State.
         * @stable ICU 4.8
         */
        State() { uchars=NULL; }
    private:
        friend class UCharsTrie;

        const char16_t *uchars;
        const char16_t *pos;
        int32_t remainingMatchLength;
    };

    /**
     * Saves the state of this trie.
     * @param state The State object to hold the trie's state.
     * @return *this
     * @see resetToState
     * @stable ICU 4.8
     */
    const UCharsTrie &saveState(State &state) const {
        state.uchars=uchars_;
        state.pos=pos_;
        state.remainingMatchLength=remainingMatchLength_;
        return *this;
    }

    /**
     * Resets this trie to the saved state.
     * If the state object contains no state, or the state of a different trie,
     * then this trie remains unchanged.
     * @param state The State object which holds a saved trie state.
     * @return *this
     * @see saveState
     * @see reset
     * @stable ICU 4.8
     */
    UCharsTrie &resetToState(const State &state) {
        if(uchars_==state.uchars && uchars_!=NULL) {
            pos_=state.pos;
            remainingMatchLength_=state.remainingMatchLength;
        }
        return *this;
    }

    /**
     * Determines whether the string so far matches, whether it has a value,
     * and whether another input char16_t can continue a matching string.
     * @return The match/value Result.
     * @stable ICU 4.8
     */
    UStringTrieResult current() const;

    /**
     * Traverses the trie from the initial state for this input char16_t.
     * Equivalent to reset().next(uchar).
     * @param uchar Input char value. Values below 0 and above 0xffff will never match.
     * @return The match/value Result.
     * @stable ICU 4.8
     */
    inline UStringTrieResult first(int32_t uchar) {
        remainingMatchLength_=-1;
        return nextImpl(uchars_, uchar);
    }

    /**
     * Traverses the trie from the initial state for the
     * one or two UTF-16 code units for this input code point.
     * Equivalent to reset().nextForCodePoint(cp).
     * @param cp A Unicode code point 0..0x10ffff.
     * @return The match/value Result.
     * @stable ICU 4.8
     */
    UStringTrieResult firstForCodePoint(UChar32 cp);

    /**
     * Traverses the trie from the current state for this input char16_t.
     * @param uchar Input char value. Values below 0 and above 0xffff will never match.
     * @return The match/value Result.
     * @stable ICU 4.8
     */
    UStringTrieResult next(int32_t uchar);

    /**
     * Traverses the trie from the current state for the
     * one or two UTF-16 code units for this input code point.
     * @param cp A Unicode code point 0..0x10ffff.
     * @return The match/value Result.
     * @stable ICU 4.8
     */
    UStringTrieResult nextForCodePoint(UChar32 cp);

    /**
     * Traverses the trie from the current state for this string.
     * Equivalent to
     * \code
     * Result result=current();
     * for(each c in s)
     *   if(!USTRINGTRIE_HAS_NEXT(result)) return USTRINGTRIE_NO_MATCH;
     *   result=next(c);
     * return result;
     * \endcode
     * @param s A string. Can be NULL if length is 0.
     * @param length The length of the string. Can be -1 if NUL-terminated.
     * @return The match/value Result.
     * @stable ICU 4.8
     */
    UStringTrieResult next(ConstChar16Ptr s, int32_t length);

    /**
     * Returns a matching string's value if called immediately after
     * current()/first()/next() returned USTRINGTRIE_INTERMEDIATE_VALUE or USTRINGTRIE_FINAL_VALUE.
     * getValue() can be called multiple times.
     *
     * Do not call getValue() after USTRINGTRIE_NO_MATCH or USTRINGTRIE_NO_VALUE!
     * @return The value for the string so far.
     * @stable ICU 4.8
     */
    inline int32_t getValue() const {
        const char16_t *pos=pos_;
        int32_t leadUnit=*pos++;
        // U_ASSERT(leadUnit>=kMinValueLead);
        return leadUnit&kValueIsFinal ?
            readValue(pos, leadUnit&0x7fff) : readNodeValue(pos, leadUnit);
    }

    /**
     * Determines whether all strings reachable from the current state
     * map to the same value.
     * @param uniqueValue Receives the unique value, if this function returns TRUE.
     *                    (output-only)
     * @return TRUE if all strings reachable from the current state
     *         map to the same value.
     * @stable ICU 4.8
     */
    inline UBool hasUniqueValue(int32_t &uniqueValue) const {
        const char16_t *pos=pos_;
        // Skip the rest of a pending linear-match node.
        return pos!=NULL && findUniqueValue(pos+remainingMatchLength_+1, FALSE, uniqueValue);
    }

    /**
     * Finds each char16_t which continues the string from the current state.
     * That is, each char16_t c for which it would be next(c)!=USTRINGTRIE_NO_MATCH now.
     * @param out Each next char16_t is appended to this object.
     * @return the number of char16_ts which continue the string from here
     * @stable ICU 4.8
     */
    int32_t getNextUChars(Appendable &out) const;

    /**
     * Iterator for all of the (string, value) pairs in a UCharsTrie.
     * @stable ICU 4.8
     */
    class U_COMMON_API Iterator : public UMemory {
    public:
        /**
         * Iterates from the root of a char16_t-serialized UCharsTrie.
         * @param trieUChars The trie char16_ts.
         * @param maxStringLength If 0, the iterator returns full strings.
         *                        Otherwise, the iterator returns strings with this maximum length.
         * @param errorCode Standard ICU error code. Its input value must
         *                  pass the U_SUCCESS() test, or else the function returns
         *                  immediately. Check for U_FAILURE() on output or use with
         *                  function chaining. (See User Guide for details.)
         * @stable ICU 4.8
         */
        Iterator(ConstChar16Ptr trieUChars, int32_t maxStringLength, UErrorCode &errorCode);

        /**
         * Iterates from the current state of the specified UCharsTrie.
         * @param trie The trie whose state will be copied for iteration.
         * @param maxStringLength If 0, the iterator returns full strings.
         *                        Otherwise, the iterator returns strings with this maximum length.
         * @param errorCode Standard ICU error code. Its input value must
         *                  pass the U_SUCCESS() test, or else the function returns
         *                  immediately. Check for U_FAILURE() on output or use with
         *                  function chaining. (See User Guide for details.)
         * @stable ICU 4.8
         */
        Iterator(const UCharsTrie &trie, int32_t maxStringLength, UErrorCode &errorCode);

        /**
         * Destructor.
         * @stable ICU 4.8
         */
        ~Iterator();

        /**
         * Resets this iterator to its initial state.
         * @return *this
         * @stable ICU 4.8
         */
        Iterator &reset();

        /**
         * @return TRUE if there are more elements.
         * @stable ICU 4.8
         */
        UBool hasNext() const;

        /**
         * Finds the next (string, value) pair if there is one.
         *
         * If the string is truncated to the maximum length and does not
         * have a real value, then the value is set to -1.
         * In this case, this "not a real value" is indistinguishable from
         * a real value of -1.
         * @param errorCode Standard ICU error code. Its input value must
         *                  pass the U_SUCCESS() test, or else the function returns
         *                  immediately. Check for U_FAILURE() on output or use with
         *                  function chaining. (See User Guide for details.)
         * @return TRUE if there is another element.
         * @stable ICU 4.8
         */
        UBool next(UErrorCode &errorCode);

        /**
         * @return The string for the last successful next().
         * @stable ICU 4.8
         */
        const UnicodeString &getString() const { return str_; }
        /**
         * @return The value for the last successful next().
         * @stable ICU 4.8
         */
        int32_t getValue() const { return value_; }

    private:
        UBool truncateAndStop() {
            pos_=NULL;
            value_=-1;  // no real value for str
            return TRUE;
        }

        const char16_t *branchNext(const char16_t *pos, int32_t length, UErrorCode &errorCode);

        const char16_t *uchars_;
        const char16_t *pos_;
        const char16_t *initialPos_;
        int32_t remainingMatchLength_;
        int32_t initialRemainingMatchLength_;
        UBool skipValue_;  // Skip intermediate value which was already delivered.

        UnicodeString str_;
        int32_t maxLength_;
        int32_t value_;

        // The stack stores pairs of integers for backtracking to another
        // outbound edge of a branch node.
        // The first integer is an offset from uchars_.
        // The second integer has the str_.length() from before the node in bits 15..0,
        // and the remaining branch length in bits 31..16.
        // (We could store the remaining branch length minus 1 in bits 30..16 and not use the sign bit,
        // but the code looks more confusing that way.)
        UVector32 *stack_;
    };

private:
    friend class UCharsTrieBuilder;

    /**
     * Constructs a UCharsTrie reader instance.
     * Unlike the public constructor which just aliases an array,
     * this constructor adopts the builder's array.
     * This constructor is only called by the builder.
     */
    UCharsTrie(char16_t *adoptUChars, const char16_t *trieUChars)
            : ownedArray_(adoptUChars), uchars_(trieUChars),
              pos_(uchars_), remainingMatchLength_(-1) {}

    // No assignment operator.
    UCharsTrie &operator=(const UCharsTrie &other);

    inline void stop() {
        pos_=NULL;
    }

    // Reads a compact 32-bit integer.
    // pos is already after the leadUnit, and the lead unit has bit 15 reset.
    static inline int32_t readValue(const char16_t *pos, int32_t leadUnit) {
        int32_t value;
        if(leadUnit<kMinTwoUnitValueLead) {
            value=leadUnit;
        } else if(leadUnit<kThreeUnitValueLead) {
            value=((leadUnit-kMinTwoUnitValueLead)<<16)|*pos;
        } else {
            value=(pos[0]<<16)|pos[1];
        }
        return value;
    }
    static inline const char16_t *skipValue(const char16_t *pos, int32_t leadUnit) {
        if(leadUnit>=kMinTwoUnitValueLead) {
            if(leadUnit<kThreeUnitValueLead) {
                ++pos;
            } else {
                pos+=2;
            }
        }
        return pos;
    }
    static inline const char16_t *skipValue(const char16_t *pos) {
        int32_t leadUnit=*pos++;
        return skipValue(pos, leadUnit&0x7fff);
    }

    static inline int32_t readNodeValue(const char16_t *pos, int32_t leadUnit) {
        // U_ASSERT(kMinValueLead<=leadUnit && leadUnit<kValueIsFinal);
        int32_t value;
        if(leadUnit<kMinTwoUnitNodeValueLead) {
            value=(leadUnit>>6)-1;
        } else if(leadUnit<kThreeUnitNodeValueLead) {
            value=(((leadUnit&0x7fc0)-kMinTwoUnitNodeValueLead)<<10)|*pos;
        } else {
            value=(pos[0]<<16)|pos[1];
        }
        return value;
    }
    static inline const char16_t *skipNodeValue(const char16_t *pos, int32_t leadUnit) {
        // U_ASSERT(kMinValueLead<=leadUnit && leadUnit<kValueIsFinal);
        if(leadUnit>=kMinTwoUnitNodeValueLead) {
            if(leadUnit<kThreeUnitNodeValueLead) {
                ++pos;
            } else {
                pos+=2;
            }
        }
        return pos;
    }

    static inline const char16_t *jumpByDelta(const char16_t *pos) {
        int32_t delta=*pos++;
        if(delta>=kMinTwoUnitDeltaLead) {
            if(delta==kThreeUnitDeltaLead) {
                delta=(pos[0]<<16)|pos[1];
                pos+=2;
            } else {
                delta=((delta-kMinTwoUnitDeltaLead)<<16)|*pos++;
            }
        }
        return pos+delta;
    }

    static const char16_t *skipDelta(const char16_t *pos) {
        int32_t delta=*pos++;
        if(delta>=kMinTwoUnitDeltaLead) {
            if(delta==kThreeUnitDeltaLead) {
                pos+=2;
            } else {
                ++pos;
            }
        }
        return pos;
    }

    static inline UStringTrieResult valueResult(int32_t node) {
        return (UStringTrieResult)(USTRINGTRIE_INTERMEDIATE_VALUE-(node>>15));
    }

    // Handles a branch node for both next(uchar) and next(string).
    UStringTrieResult branchNext(const char16_t *pos, int32_t length, int32_t uchar);

    // Requires remainingLength_<0.
    UStringTrieResult nextImpl(const char16_t *pos, int32_t uchar);

    // Helper functions for hasUniqueValue().
    // Recursively finds a unique value (or whether there is not a unique one)
    // from a branch.
    static const char16_t *findUniqueValueFromBranch(const char16_t *pos, int32_t length,
                                                  UBool haveUniqueValue, int32_t &uniqueValue);
    // Recursively finds a unique value (or whether there is not a unique one)
    // starting from a position on a node lead unit.
    static UBool findUniqueValue(const char16_t *pos, UBool haveUniqueValue, int32_t &uniqueValue);

    // Helper functions for getNextUChars().
    // getNextUChars() when pos is on a branch node.
    static void getNextBranchUChars(const char16_t *pos, int32_t length, Appendable &out);

    // UCharsTrie data structure
    //
    // The trie consists of a series of char16_t-serialized nodes for incremental
    // Unicode string/char16_t sequence matching. (char16_t=16-bit unsigned integer)
    // The root node is at the beginning of the trie data.
    //
    // Types of nodes are distinguished by their node lead unit ranges.
    // After each node, except a final-value node, another node follows to
    // encode match values or continue matching further units.
    //
    // Node types:
    //  - Final-value node: Stores a 32-bit integer in a compact, variable-length format.
    //    The value is for the string/char16_t sequence so far.
    //  - Match node, optionally with an intermediate value in a different compact format.
    //    The value, if present, is for the string/char16_t sequence so far.
    //
    //  Aside from the value, which uses the node lead unit's high bits:
    //
    //  - Linear-match node: Matches a number of units.
    //  - Branch node: Branches to other nodes according to the current input unit.
    //    The node unit is the length of the branch (number of units to select from)
    //    minus 1. It is followed by a sub-node:
    //    - If the length is at most kMaxBranchLinearSubNodeLength, then
    //      there are length-1 (key, value) pairs and then one more comparison unit.
    //      If one of the key units matches, then the value is either a final value for
    //      the string so far, or a "jump" delta to the next node.
    //      If the last unit matches, then matching continues with the next node.
    //      (Values have the same encoding as final-value nodes.)
    //    - If the length is greater than kMaxBranchLinearSubNodeLength, then
    //      there is one unit and one "jump" delta.
    //      If the input unit is less than the sub-node unit, then "jump" by delta to
    //      the next sub-node which will have a length of length/2.
    //      (The delta has its own compact encoding.)
    //      Otherwise, skip the "jump" delta to the next sub-node
    //      which will have a length of length-length/2.

    // Match-node lead unit values, after masking off intermediate-value bits:

    // 0000..002f: Branch node. If node!=0 then the length is node+1, otherwise
    // the length is one more than the next unit.

    // For a branch sub-node with at most this many entries, we drop down
    // to a linear search.
    static const int32_t kMaxBranchLinearSubNodeLength=5;

    // 0030..003f: Linear-match node, match 1..16 units and continue reading the next node.
    static const int32_t kMinLinearMatch=0x30;
    static const int32_t kMaxLinearMatchLength=0x10;

    // Match-node lead unit bits 14..6 for the optional intermediate value.
    // If these bits are 0, then there is no intermediate value.
    // Otherwise, see the *NodeValue* constants below.
    static const int32_t kMinValueLead=kMinLinearMatch+kMaxLinearMatchLength;  // 0x0040
    static const int32_t kNodeTypeMask=kMinValueLead-1;  // 0x003f

    // A final-value node has bit 15 set.
    static const int32_t kValueIsFinal=0x8000;

    // Compact value: After testing and masking off bit 15, use the following thresholds.
    static const int32_t kMaxOneUnitValue=0x3fff;

    static const int32_t kMinTwoUnitValueLead=kMaxOneUnitValue+1;  // 0x4000
    static const int32_t kThreeUnitValueLead=0x7fff;

    static const int32_t kMaxTwoUnitValue=((kThreeUnitValueLead-kMinTwoUnitValueLead)<<16)-1;  // 0x3ffeffff

    // Compact intermediate-value integer, lead unit shared with a branch or linear-match node.
    static const int32_t kMaxOneUnitNodeValue=0xff;
    static const int32_t kMinTwoUnitNodeValueLead=kMinValueLead+((kMaxOneUnitNodeValue+1)<<6);  // 0x4040
    static const int32_t kThreeUnitNodeValueLead=0x7fc0;

    static const int32_t kMaxTwoUnitNodeValue=
        ((kThreeUnitNodeValueLead-kMinTwoUnitNodeValueLead)<<10)-1;  // 0xfdffff

    // Compact delta integers.
    static const int32_t kMaxOneUnitDelta=0xfbff;
    static const int32_t kMinTwoUnitDeltaLead=kMaxOneUnitDelta+1;  // 0xfc00
    static const int32_t kThreeUnitDeltaLead=0xffff;

    static const int32_t kMaxTwoUnitDelta=((kThreeUnitDeltaLead-kMinTwoUnitDeltaLead)<<16)-1;  // 0x03feffff

    // For getState64():
    // The remainingMatchLength_ is -1..14=(kMaxLinearMatchLength=0x10)-2
    // so we need at least 5 bits for that.
    // We add 2 to store it as a positive value 1..16=kMaxLinearMatchLength.
    static constexpr int32_t kState64RemainingShift = 59;
    static constexpr uint64_t kState64PosMask = (UINT64_C(1) << kState64RemainingShift) - 1;

    char16_t *ownedArray_;

    // Fixed value referencing the UCharsTrie words.
    const char16_t *uchars_;

    // Iterator variables.

    // Pointer to next trie unit to read. NULL if no more matches.
    const char16_t *pos_;
    // Remaining length of a linear-match node, minus 1. Negative if not in such a node.
    int32_t remainingMatchLength_;
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */







U_NAMESPACE_BEGIN

UCharsTrie::~UCharsTrie() {
    uprv_free(ownedArray_);
}

UStringTrieResult
UCharsTrie::current() const {
    const UChar *pos=pos_;
    if(pos==NULL) {
        return USTRINGTRIE_NO_MATCH;
    } else {
        int32_t node;
        return (remainingMatchLength_<0 && (node=*pos)>=kMinValueLead) ?
                valueResult(node) : USTRINGTRIE_NO_VALUE;
    }
}

UStringTrieResult
UCharsTrie::firstForCodePoint(UChar32 cp) {
    return cp<=0xffff ?
        first(cp) :
        (USTRINGTRIE_HAS_NEXT(first(U16_LEAD(cp))) ?
            next(U16_TRAIL(cp)) :
            USTRINGTRIE_NO_MATCH);
}

UStringTrieResult
UCharsTrie::nextForCodePoint(UChar32 cp) {
    return cp<=0xffff ?
        next(cp) :
        (USTRINGTRIE_HAS_NEXT(next(U16_LEAD(cp))) ?
            next(U16_TRAIL(cp)) :
            USTRINGTRIE_NO_MATCH);
}

UStringTrieResult
UCharsTrie::branchNext(const UChar *pos, int32_t length, int32_t uchar) {
    // Branch according to the current unit.
    if(length==0) {
        length=*pos++;
    }
    ++length;
    // The length of the branch is the number of units to select from.
    // The data structure encodes a binary search.
    while(length>kMaxBranchLinearSubNodeLength) {
        if(uchar<*pos++) {
            length>>=1;
            pos=jumpByDelta(pos);
        } else {
            length=length-(length>>1);
            pos=skipDelta(pos);
        }
    }
    // Drop down to linear search for the last few units.
    // length>=2 because the loop body above sees length>kMaxBranchLinearSubNodeLength>=3
    // and divides length by 2.
    do {
        if(uchar==*pos++) {
            UStringTrieResult result;
            int32_t node=*pos;
            if(node&kValueIsFinal) {
                // Leave the final value for getValue() to read.
                result=USTRINGTRIE_FINAL_VALUE;
            } else {
                // Use the non-final value as the jump delta.
                ++pos;
                // int32_t delta=readValue(pos, node);
                int32_t delta;
                if(node<kMinTwoUnitValueLead) {
                    delta=node;
                } else if(node<kThreeUnitValueLead) {
                    delta=((node-kMinTwoUnitValueLead)<<16)|*pos++;
                } else {
                    delta=(pos[0]<<16)|pos[1];
                    pos+=2;
                }
                // end readValue()
                pos+=delta;
                node=*pos;
                result= node>=kMinValueLead ? valueResult(node) : USTRINGTRIE_NO_VALUE;
            }
            pos_=pos;
            return result;
        }
        --length;
        pos=skipValue(pos);
    } while(length>1);
    if(uchar==*pos++) {
        pos_=pos;
        int32_t node=*pos;
        return node>=kMinValueLead ? valueResult(node) : USTRINGTRIE_NO_VALUE;
    } else {
        stop();
        return USTRINGTRIE_NO_MATCH;
    }
}

UStringTrieResult
UCharsTrie::nextImpl(const UChar *pos, int32_t uchar) {
    int32_t node=*pos++;
    for(;;) {
        if(node<kMinLinearMatch) {
            return branchNext(pos, node, uchar);
        } else if(node<kMinValueLead) {
            // Match the first of length+1 units.
            int32_t length=node-kMinLinearMatch;  // Actual match length minus 1.
            if(uchar==*pos++) {
                remainingMatchLength_=--length;
                pos_=pos;
                return (length<0 && (node=*pos)>=kMinValueLead) ?
                        valueResult(node) : USTRINGTRIE_NO_VALUE;
            } else {
                // No match.
                break;
            }
        } else if(node&kValueIsFinal) {
            // No further matching units.
            break;
        } else {
            // Skip intermediate value.
            pos=skipNodeValue(pos, node);
            node&=kNodeTypeMask;
        }
    }
    stop();
    return USTRINGTRIE_NO_MATCH;
}

UStringTrieResult
UCharsTrie::next(int32_t uchar) {
    const UChar *pos=pos_;
    if(pos==NULL) {
        return USTRINGTRIE_NO_MATCH;
    }
    int32_t length=remainingMatchLength_;  // Actual remaining match length minus 1.
    if(length>=0) {
        // Remaining part of a linear-match node.
        if(uchar==*pos++) {
            remainingMatchLength_=--length;
            pos_=pos;
            int32_t node;
            return (length<0 && (node=*pos)>=kMinValueLead) ?
                    valueResult(node) : USTRINGTRIE_NO_VALUE;
        } else {
            stop();
            return USTRINGTRIE_NO_MATCH;
        }
    }
    return nextImpl(pos, uchar);
}

UStringTrieResult
UCharsTrie::next(ConstChar16Ptr ptr, int32_t sLength) {
    const UChar *s=ptr;
    if(sLength<0 ? *s==0 : sLength==0) {
        // Empty input.
        return current();
    }
    const UChar *pos=pos_;
    if(pos==NULL) {
        return USTRINGTRIE_NO_MATCH;
    }
    int32_t length=remainingMatchLength_;  // Actual remaining match length minus 1.
    for(;;) {
        // Fetch the next input unit, if there is one.
        // Continue a linear-match node without rechecking sLength<0.
        int32_t uchar;
        if(sLength<0) {
            for(;;) {
                if((uchar=*s++)==0) {
                    remainingMatchLength_=length;
                    pos_=pos;
                    int32_t node;
                    return (length<0 && (node=*pos)>=kMinValueLead) ?
                            valueResult(node) : USTRINGTRIE_NO_VALUE;
                }
                if(length<0) {
                    remainingMatchLength_=length;
                    break;
                }
                if(uchar!=*pos) {
                    stop();
                    return USTRINGTRIE_NO_MATCH;
                }
                ++pos;
                --length;
            }
        } else {
            for(;;) {
                if(sLength==0) {
                    remainingMatchLength_=length;
                    pos_=pos;
                    int32_t node;
                    return (length<0 && (node=*pos)>=kMinValueLead) ?
                            valueResult(node) : USTRINGTRIE_NO_VALUE;
                }
                uchar=*s++;
                --sLength;
                if(length<0) {
                    remainingMatchLength_=length;
                    break;
                }
                if(uchar!=*pos) {
                    stop();
                    return USTRINGTRIE_NO_MATCH;
                }
                ++pos;
                --length;
            }
        }
        int32_t node=*pos++;
        for(;;) {
            if(node<kMinLinearMatch) {
                UStringTrieResult result=branchNext(pos, node, uchar);
                if(result==USTRINGTRIE_NO_MATCH) {
                    return USTRINGTRIE_NO_MATCH;
                }
                // Fetch the next input unit, if there is one.
                if(sLength<0) {
                    if((uchar=*s++)==0) {
                        return result;
                    }
                } else {
                    if(sLength==0) {
                        return result;
                    }
                    uchar=*s++;
                    --sLength;
                }
                if(result==USTRINGTRIE_FINAL_VALUE) {
                    // No further matching units.
                    stop();
                    return USTRINGTRIE_NO_MATCH;
                }
                pos=pos_;  // branchNext() advanced pos and wrote it to pos_ .
                node=*pos++;
            } else if(node<kMinValueLead) {
                // Match length+1 units.
                length=node-kMinLinearMatch;  // Actual match length minus 1.
                if(uchar!=*pos) {
                    stop();
                    return USTRINGTRIE_NO_MATCH;
                }
                ++pos;
                --length;
                break;
            } else if(node&kValueIsFinal) {
                // No further matching units.
                stop();
                return USTRINGTRIE_NO_MATCH;
            } else {
                // Skip intermediate value.
                pos=skipNodeValue(pos, node);
                node&=kNodeTypeMask;
            }
        }
    }
}

const UChar *
UCharsTrie::findUniqueValueFromBranch(const UChar *pos, int32_t length,
                                      UBool haveUniqueValue, int32_t &uniqueValue) {
    while(length>kMaxBranchLinearSubNodeLength) {
        ++pos;  // ignore the comparison unit
        if(NULL==findUniqueValueFromBranch(jumpByDelta(pos), length>>1, haveUniqueValue, uniqueValue)) {
            return NULL;
        }
        length=length-(length>>1);
        pos=skipDelta(pos);
    }
    do {
        ++pos;  // ignore a comparison unit
        // handle its value
        int32_t node=*pos++;
        UBool isFinal=(UBool)(node>>15);
        node&=0x7fff;
        int32_t value=readValue(pos, node);
        pos=skipValue(pos, node);
        if(isFinal) {
            if(haveUniqueValue) {
                if(value!=uniqueValue) {
                    return NULL;
                }
            } else {
                uniqueValue=value;
                haveUniqueValue=TRUE;
            }
        } else {
            if(!findUniqueValue(pos+value, haveUniqueValue, uniqueValue)) {
                return NULL;
            }
            haveUniqueValue=TRUE;
        }
    } while(--length>1);
    return pos+1;  // ignore the last comparison unit
}

UBool
UCharsTrie::findUniqueValue(const UChar *pos, UBool haveUniqueValue, int32_t &uniqueValue) {
    int32_t node=*pos++;
    for(;;) {
        if(node<kMinLinearMatch) {
            if(node==0) {
                node=*pos++;
            }
            pos=findUniqueValueFromBranch(pos, node+1, haveUniqueValue, uniqueValue);
            if(pos==NULL) {
                return FALSE;
            }
            haveUniqueValue=TRUE;
            node=*pos++;
        } else if(node<kMinValueLead) {
            // linear-match node
            pos+=node-kMinLinearMatch+1;  // Ignore the match units.
            node=*pos++;
        } else {
            UBool isFinal=(UBool)(node>>15);
            int32_t value;
            if(isFinal) {
                value=readValue(pos, node&0x7fff);
            } else {
                value=readNodeValue(pos, node);
            }
            if(haveUniqueValue) {
                if(value!=uniqueValue) {
                    return FALSE;
                }
            } else {
                uniqueValue=value;
                haveUniqueValue=TRUE;
            }
            if(isFinal) {
                return TRUE;
            }
            pos=skipNodeValue(pos, node);
            node&=kNodeTypeMask;
        }
    }
}

int32_t
UCharsTrie::getNextUChars(Appendable &out) const {
    const UChar *pos=pos_;
    if(pos==NULL) {
        return 0;
    }
    if(remainingMatchLength_>=0) {
        out.appendCodeUnit(*pos);  // Next unit of a pending linear-match node.
        return 1;
    }
    int32_t node=*pos++;
    if(node>=kMinValueLead) {
        if(node&kValueIsFinal) {
            return 0;
        } else {
            pos=skipNodeValue(pos, node);
            node&=kNodeTypeMask;
        }
    }
    if(node<kMinLinearMatch) {
        if(node==0) {
            node=*pos++;
        }
        out.reserveAppendCapacity(++node);
        getNextBranchUChars(pos, node, out);
        return node;
    } else {
        // First unit of the linear-match node.
        out.appendCodeUnit(*pos);
        return 1;
    }
}

void
UCharsTrie::getNextBranchUChars(const UChar *pos, int32_t length, Appendable &out) {
    while(length>kMaxBranchLinearSubNodeLength) {
        ++pos;  // ignore the comparison unit
        getNextBranchUChars(jumpByDelta(pos), length>>1, out);
        length=length-(length>>1);
        pos=skipDelta(pos);
    }
    do {
        out.appendCodeUnit(*pos++);
        pos=skipValue(pos);
    } while(--length>1);
    out.appendCodeUnit(*pos);
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  ucharstriebuilder.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010nov14
*   created by: Markus W. Scherer
*/



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  ucharstriebuilder.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010nov14
*   created by: Markus W. Scherer
*/




#if U_SHOW_CPLUSPLUS_API





/**
 * \file
 * \brief C++ API: Builder for icu::UCharsTrie
 */

U_NAMESPACE_BEGIN

class UCharsTrieElement;

/**
 * Builder class for UCharsTrie.
 *
 * This class is not intended for public subclassing.
 * @stable ICU 4.8
 */
class U_COMMON_API UCharsTrieBuilder : public StringTrieBuilder {
public:
    /**
     * Constructs an empty builder.
     * @param errorCode Standard ICU error code.
     * @stable ICU 4.8
     */
    UCharsTrieBuilder(UErrorCode &errorCode);

    /**
     * Destructor.
     * @stable ICU 4.8
     */
    virtual ~UCharsTrieBuilder();

    /**
     * Adds a (string, value) pair.
     * The string must be unique.
     * The string contents will be copied; the builder does not keep
     * a reference to the input UnicodeString or its buffer.
     * @param s The input string.
     * @param value The value associated with this string.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return *this
     * @stable ICU 4.8
     */
    UCharsTrieBuilder &add(const UnicodeString &s, int32_t value, UErrorCode &errorCode);

    /**
     * Builds a UCharsTrie for the add()ed data.
     * Once built, no further data can be add()ed until clear() is called.
     *
     * A UCharsTrie cannot be empty. At least one (string, value) pair
     * must have been add()ed.
     *
     * This method passes ownership of the builder's internal result array to the new trie object.
     * Another call to any build() variant will re-serialize the trie.
     * After clear() has been called, a new array will be used as well.
     * @param buildOption Build option, see UStringTrieBuildOption.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return A new UCharsTrie for the add()ed data.
     * @stable ICU 4.8
     */
    UCharsTrie *build(UStringTrieBuildOption buildOption, UErrorCode &errorCode);

    /**
     * Builds a UCharsTrie for the add()ed data and char16_t-serializes it.
     * Once built, no further data can be add()ed until clear() is called.
     *
     * A UCharsTrie cannot be empty. At least one (string, value) pair
     * must have been add()ed.
     *
     * Multiple calls to buildUnicodeString() set the UnicodeStrings to the
     * builder's same char16_t array, without rebuilding.
     * If buildUnicodeString() is called after build(), the trie will be
     * re-serialized into a new array.
     * If build() is called after buildUnicodeString(), the trie object will become
     * the owner of the previously returned array.
     * After clear() has been called, a new array will be used as well.
     * @param buildOption Build option, see UStringTrieBuildOption.
     * @param result A UnicodeString which will be set to the char16_t-serialized
     *               UCharsTrie for the add()ed data.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return result
     * @stable ICU 4.8
     */
    UnicodeString &buildUnicodeString(UStringTrieBuildOption buildOption, UnicodeString &result,
                                      UErrorCode &errorCode);

    /**
     * Removes all (string, value) pairs.
     * New data can then be add()ed and a new trie can be built.
     * @return *this
     * @stable ICU 4.8
     */
    UCharsTrieBuilder &clear() {
        strings.remove();
        elementsLength=0;
        ucharsLength=0;
        return *this;
    }

private:
    UCharsTrieBuilder(const UCharsTrieBuilder &other);  // no copy constructor
    UCharsTrieBuilder &operator=(const UCharsTrieBuilder &other);  // no assignment operator

    void buildUChars(UStringTrieBuildOption buildOption, UErrorCode &errorCode);

    virtual int32_t getElementStringLength(int32_t i) const;
    virtual char16_t getElementUnit(int32_t i, int32_t unitIndex) const;
    virtual int32_t getElementValue(int32_t i) const;

    virtual int32_t getLimitOfLinearMatch(int32_t first, int32_t last, int32_t unitIndex) const;

    virtual int32_t countElementUnits(int32_t start, int32_t limit, int32_t unitIndex) const;
    virtual int32_t skipElementsBySomeUnits(int32_t i, int32_t unitIndex, int32_t count) const;
    virtual int32_t indexOfElementWithNextUnit(int32_t i, int32_t unitIndex, char16_t unit) const;

    virtual UBool matchNodesCanHaveValues() const { return TRUE; }

    virtual int32_t getMaxBranchLinearSubNodeLength() const { return UCharsTrie::kMaxBranchLinearSubNodeLength; }
    virtual int32_t getMinLinearMatch() const { return UCharsTrie::kMinLinearMatch; }
    virtual int32_t getMaxLinearMatchLength() const { return UCharsTrie::kMaxLinearMatchLength; }

    class UCTLinearMatchNode : public LinearMatchNode {
    public:
        UCTLinearMatchNode(const char16_t *units, int32_t len, Node *nextNode);
        virtual UBool operator==(const Node &other) const;
        virtual void write(StringTrieBuilder &builder);
    private:
        const char16_t *s;
    };

    virtual Node *createLinearMatchNode(int32_t i, int32_t unitIndex, int32_t length,
                                        Node *nextNode) const;

    UBool ensureCapacity(int32_t length);
    virtual int32_t write(int32_t unit);
    int32_t write(const char16_t *s, int32_t length);
    virtual int32_t writeElementUnits(int32_t i, int32_t unitIndex, int32_t length);
    virtual int32_t writeValueAndFinal(int32_t i, UBool isFinal);
    virtual int32_t writeValueAndType(UBool hasValue, int32_t value, int32_t node);
    virtual int32_t writeDeltaTo(int32_t jumpTarget);

    UnicodeString strings;
    UCharsTrieElement *elements;
    int32_t elementsCapacity;
    int32_t elementsLength;

    // char16_t serialization of the trie.
    // Grows from the back: ucharsLength measures from the end of the buffer!
    char16_t *uchars;
    int32_t ucharsCapacity;
    int32_t ucharsLength;
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */










U_NAMESPACE_BEGIN

/*
 * Note: This builder implementation stores (string, value) pairs with full copies
 * of the 16-bit-unit sequences, until the UCharsTrie is built.
 * It might(!) take less memory if we collected the data in a temporary, dynamic trie.
 */

class UCharsTrieElement : public UMemory {
public:
    // Use compiler's default constructor, initializes nothing.

    void setTo(const UnicodeString &s, int32_t val, UnicodeString &strings, UErrorCode &errorCode);

    UnicodeString getString(const UnicodeString &strings) const {
        int32_t length=strings[stringOffset];
        return strings.tempSubString(stringOffset+1, length);
    }
    int32_t getStringLength(const UnicodeString &strings) const {
        return strings[stringOffset];
    }

    UChar charAt(int32_t index, const UnicodeString &strings) const {
        return strings[stringOffset+1+index];
    }

    int32_t getValue() const { return value; }

    int32_t compareStringTo(const UCharsTrieElement &o, const UnicodeString &strings) const;

private:
    // The first strings unit contains the string length.
    // (Compared with a stringLength field here, this saves 2 bytes per string.)
    int32_t stringOffset;
    int32_t value;
};

void
UCharsTrieElement::setTo(const UnicodeString &s, int32_t val,
                         UnicodeString &strings, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    int32_t length=s.length();
    if(length>0xffff) {
        // Too long: We store the length in 1 unit.
        errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return;
    }
    stringOffset=strings.length();
    strings.append((UChar)length);
    value=val;
    strings.append(s);
}

int32_t
UCharsTrieElement::compareStringTo(const UCharsTrieElement &other, const UnicodeString &strings) const {
    return getString(strings).compare(other.getString(strings));
}

UCharsTrieBuilder::UCharsTrieBuilder(UErrorCode & /*errorCode*/)
        : elements(NULL), elementsCapacity(0), elementsLength(0),
          uchars(NULL), ucharsCapacity(0), ucharsLength(0) {}

UCharsTrieBuilder::~UCharsTrieBuilder() {
    delete[] elements;
    uprv_free(uchars);
}

UCharsTrieBuilder &
UCharsTrieBuilder::add(const UnicodeString &s, int32_t value, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return *this;
    }
    if(ucharsLength>0) {
        // Cannot add elements after building.
        errorCode=U_NO_WRITE_PERMISSION;
        return *this;
    }
    if(elementsLength==elementsCapacity) {
        int32_t newCapacity;
        if(elementsCapacity==0) {
            newCapacity=1024;
        } else {
            newCapacity=4*elementsCapacity;
        }
        UCharsTrieElement *newElements=new UCharsTrieElement[newCapacity];
        if(newElements==NULL) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            return *this;
        }
        if(elementsLength>0) {
            uprv_memcpy(newElements, elements, (size_t)elementsLength*sizeof(UCharsTrieElement));
        }
        delete[] elements;
        elements=newElements;
        elementsCapacity=newCapacity;
    }
    elements[elementsLength++].setTo(s, value, strings, errorCode);
    if(U_SUCCESS(errorCode) && strings.isBogus()) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
    }
    return *this;
}

U_CDECL_BEGIN

static int32_t U_CALLCONV
ucharstriebuilder_compareElementStrings(const void *context, const void *left, const void *right) {
    const UnicodeString *strings=static_cast<const UnicodeString *>(context);
    const UCharsTrieElement *leftElement=static_cast<const UCharsTrieElement *>(left);
    const UCharsTrieElement *rightElement=static_cast<const UCharsTrieElement *>(right);
    return leftElement->compareStringTo(*rightElement, *strings);
}

U_CDECL_END

UCharsTrie *
UCharsTrieBuilder::build(UStringTrieBuildOption buildOption, UErrorCode &errorCode) {
    buildUChars(buildOption, errorCode);
    UCharsTrie *newTrie=NULL;
    if(U_SUCCESS(errorCode)) {
        newTrie=new UCharsTrie(uchars, uchars+(ucharsCapacity-ucharsLength));
        if(newTrie==NULL) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
        } else {
            uchars=NULL;  // The new trie now owns the array.
            ucharsCapacity=0;
        }
    }
    return newTrie;
}

UnicodeString &
UCharsTrieBuilder::buildUnicodeString(UStringTrieBuildOption buildOption, UnicodeString &result,
                                      UErrorCode &errorCode) {
    buildUChars(buildOption, errorCode);
    if(U_SUCCESS(errorCode)) {
        result.setTo(FALSE, uchars+(ucharsCapacity-ucharsLength), ucharsLength);
    }
    return result;
}

void
UCharsTrieBuilder::buildUChars(UStringTrieBuildOption buildOption, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    if(uchars!=NULL && ucharsLength>0) {
        // Already built.
        return;
    }
    if(ucharsLength==0) {
        if(elementsLength==0) {
            errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return;
        }
        if(strings.isBogus()) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        uprv_sortArray(elements, elementsLength, (int32_t)sizeof(UCharsTrieElement),
                      ucharstriebuilder_compareElementStrings, &strings,
                      FALSE,  // need not be a stable sort
                      &errorCode);
        if(U_FAILURE(errorCode)) {
            return;
        }
        // Duplicate strings are not allowed.
        UnicodeString prev=elements[0].getString(strings);
        for(int32_t i=1; i<elementsLength; ++i) {
            UnicodeString current=elements[i].getString(strings);
            if(prev==current) {
                errorCode=U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
            prev.fastCopyFrom(current);
        }
    }
    // Create and UChar-serialize the trie for the elements.
    ucharsLength=0;
    int32_t capacity=strings.length();
    if(capacity<1024) {
        capacity=1024;
    }
    if(ucharsCapacity<capacity) {
        uprv_free(uchars);
        uchars=static_cast<UChar *>(uprv_malloc(capacity*2));
        if(uchars==NULL) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            ucharsCapacity=0;
            return;
        }
        ucharsCapacity=capacity;
    }
    StringTrieBuilder::build(buildOption, elementsLength, errorCode);
    if(uchars==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
    }
}

int32_t
UCharsTrieBuilder::getElementStringLength(int32_t i) const {
    return elements[i].getStringLength(strings);
}

UChar
UCharsTrieBuilder::getElementUnit(int32_t i, int32_t unitIndex) const {
    return elements[i].charAt(unitIndex, strings);
}

int32_t
UCharsTrieBuilder::getElementValue(int32_t i) const {
    return elements[i].getValue();
}

int32_t
UCharsTrieBuilder::getLimitOfLinearMatch(int32_t first, int32_t last, int32_t unitIndex) const {
    const UCharsTrieElement &firstElement=elements[first];
    const UCharsTrieElement &lastElement=elements[last];
    int32_t minStringLength=firstElement.getStringLength(strings);
    while(++unitIndex<minStringLength &&
            firstElement.charAt(unitIndex, strings)==
            lastElement.charAt(unitIndex, strings)) {}
    return unitIndex;
}

int32_t
UCharsTrieBuilder::countElementUnits(int32_t start, int32_t limit, int32_t unitIndex) const {
    int32_t length=0;  // Number of different units at unitIndex.
    int32_t i=start;
    do {
        UChar unit=elements[i++].charAt(unitIndex, strings);
        while(i<limit && unit==elements[i].charAt(unitIndex, strings)) {
            ++i;
        }
        ++length;
    } while(i<limit);
    return length;
}

int32_t
UCharsTrieBuilder::skipElementsBySomeUnits(int32_t i, int32_t unitIndex, int32_t count) const {
    do {
        UChar unit=elements[i++].charAt(unitIndex, strings);
        while(unit==elements[i].charAt(unitIndex, strings)) {
            ++i;
        }
    } while(--count>0);
    return i;
}

int32_t
UCharsTrieBuilder::indexOfElementWithNextUnit(int32_t i, int32_t unitIndex, UChar unit) const {
    while(unit==elements[i].charAt(unitIndex, strings)) {
        ++i;
    }
    return i;
}

UCharsTrieBuilder::UCTLinearMatchNode::UCTLinearMatchNode(const UChar *units, int32_t len, Node *nextNode)
        : LinearMatchNode(len, nextNode), s(units) {
    hash=hash*37u+ustr_hashUCharsN(units, len);
}

UBool
UCharsTrieBuilder::UCTLinearMatchNode::operator==(const Node &other) const {
    if(this==&other) {
        return TRUE;
    }
    if(!LinearMatchNode::operator==(other)) {
        return FALSE;
    }
    const UCTLinearMatchNode &o=(const UCTLinearMatchNode &)other;
    return 0==u_memcmp(s, o.s, length);
}

void
UCharsTrieBuilder::UCTLinearMatchNode::write(StringTrieBuilder &builder) {
    UCharsTrieBuilder &b=(UCharsTrieBuilder &)builder;
    next->write(builder);
    b.write(s, length);
    offset=b.writeValueAndType(hasValue, value, b.getMinLinearMatch()+length-1);
}

StringTrieBuilder::Node *
UCharsTrieBuilder::createLinearMatchNode(int32_t i, int32_t unitIndex, int32_t length,
                                         Node *nextNode) const {
    return new UCTLinearMatchNode(
            elements[i].getString(strings).getBuffer()+unitIndex,
            length,
            nextNode);
}

UBool
UCharsTrieBuilder::ensureCapacity(int32_t length) {
    if(uchars==NULL) {
        return FALSE;  // previous memory allocation had failed
    }
    if(length>ucharsCapacity) {
        int32_t newCapacity=ucharsCapacity;
        do {
            newCapacity*=2;
        } while(newCapacity<=length);
        UChar *newUChars=static_cast<UChar *>(uprv_malloc(newCapacity*2));
        if(newUChars==NULL) {
            // unable to allocate memory
            uprv_free(uchars);
            uchars=NULL;
            ucharsCapacity=0;
            return FALSE;
        }
        u_memcpy(newUChars+(newCapacity-ucharsLength),
                 uchars+(ucharsCapacity-ucharsLength), ucharsLength);
        uprv_free(uchars);
        uchars=newUChars;
        ucharsCapacity=newCapacity;
    }
    return TRUE;
}

int32_t
UCharsTrieBuilder::write(int32_t unit) {
    int32_t newLength=ucharsLength+1;
    if(ensureCapacity(newLength)) {
        ucharsLength=newLength;
        uchars[ucharsCapacity-ucharsLength]=(UChar)unit;
    }
    return ucharsLength;
}

int32_t
UCharsTrieBuilder::write(const UChar *s, int32_t length) {
    int32_t newLength=ucharsLength+length;
    if(ensureCapacity(newLength)) {
        ucharsLength=newLength;
        u_memcpy(uchars+(ucharsCapacity-ucharsLength), s, length);
    }
    return ucharsLength;
}

int32_t
UCharsTrieBuilder::writeElementUnits(int32_t i, int32_t unitIndex, int32_t length) {
    return write(elements[i].getString(strings).getBuffer()+unitIndex, length);
}

int32_t
UCharsTrieBuilder::writeValueAndFinal(int32_t i, UBool isFinal) {
    if(0<=i && i<=UCharsTrie::kMaxOneUnitValue) {
        return write(i|(isFinal<<15));
    }
    UChar intUnits[3];
    int32_t length;
    if(i<0 || i>UCharsTrie::kMaxTwoUnitValue) {
        intUnits[0]=(UChar)(UCharsTrie::kThreeUnitValueLead);
        intUnits[1]=(UChar)((uint32_t)i>>16);
        intUnits[2]=(UChar)i;
        length=3;
    // } else if(i<=UCharsTrie::kMaxOneUnitValue) {
    //     intUnits[0]=(UChar)(i);
    //     length=1;
    } else {
        intUnits[0]=(UChar)(UCharsTrie::kMinTwoUnitValueLead+(i>>16));
        intUnits[1]=(UChar)i;
        length=2;
    }
    intUnits[0]=(UChar)(intUnits[0]|(isFinal<<15));
    return write(intUnits, length);
}

int32_t
UCharsTrieBuilder::writeValueAndType(UBool hasValue, int32_t value, int32_t node) {
    if(!hasValue) {
        return write(node);
    }
    UChar intUnits[3];
    int32_t length;
    if(value<0 || value>UCharsTrie::kMaxTwoUnitNodeValue) {
        intUnits[0]=(UChar)(UCharsTrie::kThreeUnitNodeValueLead);
        intUnits[1]=(UChar)((uint32_t)value>>16);
        intUnits[2]=(UChar)value;
        length=3;
    } else if(value<=UCharsTrie::kMaxOneUnitNodeValue) {
        intUnits[0]=(UChar)((value+1)<<6);
        length=1;
    } else {
        intUnits[0]=(UChar)(UCharsTrie::kMinTwoUnitNodeValueLead+((value>>10)&0x7fc0));
        intUnits[1]=(UChar)value;
        length=2;
    }
    intUnits[0]|=(UChar)node;
    return write(intUnits, length);
}

int32_t
UCharsTrieBuilder::writeDeltaTo(int32_t jumpTarget) {
    int32_t i=ucharsLength-jumpTarget;
    U_ASSERT(i>=0);
    if(i<=UCharsTrie::kMaxOneUnitDelta) {
        return write(i);
    }
    UChar intUnits[3];
    int32_t length;
    if(i<=UCharsTrie::kMaxTwoUnitDelta) {
        intUnits[0]=(UChar)(UCharsTrie::kMinTwoUnitDeltaLead+(i>>16));
        length=1;
    } else {
        intUnits[0]=(UChar)(UCharsTrie::kThreeUnitDeltaLead);
        intUnits[1]=(UChar)(i>>16);
        length=2;
    }
    intUnits[length++]=(UChar)i;
    return write(intUnits, length);
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  ucharstrieiterator.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010nov15
*   created by: Markus W. Scherer
*/






U_NAMESPACE_BEGIN

UCharsTrie::Iterator::Iterator(ConstChar16Ptr trieUChars, int32_t maxStringLength,
                               UErrorCode &errorCode)
        : uchars_(trieUChars),
          pos_(uchars_), initialPos_(uchars_),
          remainingMatchLength_(-1), initialRemainingMatchLength_(-1),
          skipValue_(FALSE),
          maxLength_(maxStringLength), value_(0), stack_(NULL) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    // stack_ is a pointer so that it's easy to turn ucharstrie.h into
    // a public API header for which we would want it to depend only on
    // other public headers.
    // Unlike UCharsTrie itself, its Iterator performs memory allocations anyway
    // via the UnicodeString and UVector32 implementations, so this additional
    // cost is minimal.
    stack_=new UVector32(errorCode);
    if(stack_==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
    }
}

UCharsTrie::Iterator::Iterator(const UCharsTrie &trie, int32_t maxStringLength,
                               UErrorCode &errorCode)
        : uchars_(trie.uchars_), pos_(trie.pos_), initialPos_(trie.pos_),
          remainingMatchLength_(trie.remainingMatchLength_),
          initialRemainingMatchLength_(trie.remainingMatchLength_),
          skipValue_(FALSE),
          maxLength_(maxStringLength), value_(0), stack_(NULL) {
    if(U_FAILURE(errorCode)) {
        return;
    }
    stack_=new UVector32(errorCode);
    if(U_FAILURE(errorCode)) {
        return;
    }
    if(stack_==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    int32_t length=remainingMatchLength_;  // Actual remaining match length minus 1.
    if(length>=0) {
        // Pending linear-match node, append remaining UChars to str_.
        ++length;
        if(maxLength_>0 && length>maxLength_) {
            length=maxLength_;  // This will leave remainingMatchLength>=0 as a signal.
        }
        str_.append(pos_, length);
        pos_+=length;
        remainingMatchLength_-=length;
    }
}

UCharsTrie::Iterator::~Iterator() {
    delete stack_;
}

UCharsTrie::Iterator &
UCharsTrie::Iterator::reset() {
    pos_=initialPos_;
    remainingMatchLength_=initialRemainingMatchLength_;
    skipValue_=FALSE;
    int32_t length=remainingMatchLength_+1;  // Remaining match length.
    if(maxLength_>0 && length>maxLength_) {
        length=maxLength_;
    }
    str_.truncate(length);
    pos_+=length;
    remainingMatchLength_-=length;
    stack_->setSize(0);
    return *this;
}

UBool
UCharsTrie::Iterator::hasNext() const { return pos_!=NULL || !stack_->isEmpty(); }

UBool
UCharsTrie::Iterator::next(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return FALSE;
    }
    const UChar *pos=pos_;
    if(pos==NULL) {
        if(stack_->isEmpty()) {
            return FALSE;
        }
        // Pop the state off the stack and continue with the next outbound edge of
        // the branch node.
        int32_t stackSize=stack_->size();
        int32_t length=stack_->elementAti(stackSize-1);
        pos=uchars_+stack_->elementAti(stackSize-2);
        stack_->setSize(stackSize-2);
        str_.truncate(length&0xffff);
        length=(int32_t)((uint32_t)length>>16);
        if(length>1) {
            pos=branchNext(pos, length, errorCode);
            if(pos==NULL) {
                return TRUE;  // Reached a final value.
            }
        } else {
            str_.append(*pos++);
        }
    }
    if(remainingMatchLength_>=0) {
        // We only get here if we started in a pending linear-match node
        // with more than maxLength remaining units.
        return truncateAndStop();
    }
    for(;;) {
        int32_t node=*pos++;
        if(node>=kMinValueLead) {
            if(skipValue_) {
                pos=skipNodeValue(pos, node);
                node&=kNodeTypeMask;
                skipValue_=FALSE;
            } else {
                // Deliver value for the string so far.
                UBool isFinal=(UBool)(node>>15);
                if(isFinal) {
                    value_=readValue(pos, node&0x7fff);
                } else {
                    value_=readNodeValue(pos, node);
                }
                if(isFinal || (maxLength_>0 && str_.length()==maxLength_)) {
                    pos_=NULL;
                } else {
                    // We cannot skip the value right here because it shares its
                    // lead unit with a match node which we have to evaluate
                    // next time.
                    // Instead, keep pos_ on the node lead unit itself.
                    pos_=pos-1;
                    skipValue_=TRUE;
                }
                return TRUE;
            }
        }
        if(maxLength_>0 && str_.length()==maxLength_) {
            return truncateAndStop();
        }
        if(node<kMinLinearMatch) {
            if(node==0) {
                node=*pos++;
            }
            pos=branchNext(pos, node+1, errorCode);
            if(pos==NULL) {
                return TRUE;  // Reached a final value.
            }
        } else {
            // Linear-match node, append length units to str_.
            int32_t length=node-kMinLinearMatch+1;
            if(maxLength_>0 && str_.length()+length>maxLength_) {
                str_.append(pos, maxLength_-str_.length());
                return truncateAndStop();
            }
            str_.append(pos, length);
            pos+=length;
        }
    }
}

// Branch node, needs to take the first outbound edge and push state for the rest.
const UChar *
UCharsTrie::Iterator::branchNext(const UChar *pos, int32_t length, UErrorCode &errorCode) {
    while(length>kMaxBranchLinearSubNodeLength) {
        ++pos;  // ignore the comparison unit
        // Push state for the greater-or-equal edge.
        stack_->addElement((int32_t)(skipDelta(pos)-uchars_), errorCode);
        stack_->addElement(((length-(length>>1))<<16)|str_.length(), errorCode);
        // Follow the less-than edge.
        length>>=1;
        pos=jumpByDelta(pos);
    }
    // List of key-value pairs where values are either final values or jump deltas.
    // Read the first (key, value) pair.
    UChar trieUnit=*pos++;
    int32_t node=*pos++;
    UBool isFinal=(UBool)(node>>15);
    int32_t value=readValue(pos, node&=0x7fff);
    pos=skipValue(pos, node);
    stack_->addElement((int32_t)(pos-uchars_), errorCode);
    stack_->addElement(((length-1)<<16)|str_.length(), errorCode);
    str_.append(trieUnit);
    if(isFinal) {
        pos_=NULL;
        value_=value;
        return NULL;
    } else {
        return pos+value;
    }
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 1998-2012, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*/

  // for 'typeid' to work






U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(UCharCharacterIterator)

UCharCharacterIterator::UCharCharacterIterator()
  : CharacterIterator(),
  text(0)
{
    // never default construct!
}

UCharCharacterIterator::UCharCharacterIterator(ConstChar16Ptr textPtr,
                                               int32_t length)
  : CharacterIterator(textPtr != 0 ? (length>=0 ? length : u_strlen(textPtr)) : 0),
  text(textPtr)
{
}

UCharCharacterIterator::UCharCharacterIterator(ConstChar16Ptr textPtr,
                                               int32_t length,
                                               int32_t position)
  : CharacterIterator(textPtr != 0 ? (length>=0 ? length : u_strlen(textPtr)) : 0, position),
  text(textPtr)
{
}

UCharCharacterIterator::UCharCharacterIterator(ConstChar16Ptr textPtr,
                                               int32_t length,
                                               int32_t textBegin,
                                               int32_t textEnd,
                                               int32_t position)
  : CharacterIterator(textPtr != 0 ? (length>=0 ? length : u_strlen(textPtr)) : 0, textBegin, textEnd, position),
  text(textPtr)
{
}

UCharCharacterIterator::UCharCharacterIterator(const UCharCharacterIterator& that)
: CharacterIterator(that),
  text(that.text)
{
}

UCharCharacterIterator&
UCharCharacterIterator::operator=(const UCharCharacterIterator& that) {
    CharacterIterator::operator=(that);
    text = that.text;
    return *this;
}

UCharCharacterIterator::~UCharCharacterIterator() {
}

UBool
UCharCharacterIterator::operator==(const ForwardCharacterIterator& that) const {
    if (this == &that) {
        return TRUE;
    }
    if (typeid(*this) != typeid(that)) {
        return FALSE;
    }

    UCharCharacterIterator&    realThat = (UCharCharacterIterator&)that;

    return text == realThat.text
        && textLength == realThat.textLength
        && pos == realThat.pos
        && begin == realThat.begin
        && end == realThat.end;
}

int32_t
UCharCharacterIterator::hashCode() const {
    return ustr_hashUCharsN(text, textLength) ^ pos ^ begin ^ end;
}

UCharCharacterIterator*
UCharCharacterIterator::clone() const {
    return new UCharCharacterIterator(*this);
}

UChar
UCharCharacterIterator::first() {
    pos = begin;
    if(pos < end) {
        return text[pos];
    } else {
        return DONE;
    }
}

UChar
UCharCharacterIterator::firstPostInc() {
    pos = begin;
    if(pos < end) {
        return text[pos++];
    } else {
        return DONE;
    }
}

UChar
UCharCharacterIterator::last() {
    pos = end;
    if(pos > begin) {
        return text[--pos];
    } else {
        return DONE;
    }
}

UChar
UCharCharacterIterator::setIndex(int32_t position) {
    if(position < begin) {
        pos = begin;
    } else if(position > end) {
        pos = end;
    } else {
        pos = position;
    }
    if(pos < end) {
        return text[pos];
    } else {
        return DONE;
    }
}

UChar
UCharCharacterIterator::current() const {
    if (pos >= begin && pos < end) {
        return text[pos];
    } else {
        return DONE;
    }
}

UChar
UCharCharacterIterator::next() {
    if (pos + 1 < end) {
        return text[++pos];
    } else {
        /* make current() return DONE */
        pos = end;
        return DONE;
    }
}

UChar
UCharCharacterIterator::nextPostInc() {
    if (pos < end) {
        return text[pos++];
    } else {
        return DONE;
    }
}

UBool
UCharCharacterIterator::hasNext() {
    return (UBool)(pos < end ? TRUE : FALSE);
}

UChar
UCharCharacterIterator::previous() {
    if (pos > begin) {
        return text[--pos];
    } else {
        return DONE;
    }
}

UBool
UCharCharacterIterator::hasPrevious() {
    return (UBool)(pos > begin ? TRUE : FALSE);
}

UChar32
UCharCharacterIterator::first32() {
    pos = begin;
    if(pos < end) {
        int32_t i = pos;
        UChar32 c;
        U16_NEXT(text, i, end, c);
        return c;
    } else {
        return DONE;
    }
}

UChar32
UCharCharacterIterator::first32PostInc() {
    pos = begin;
    if(pos < end) {
        UChar32 c;
        U16_NEXT(text, pos, end, c);
        return c;
    } else {
        return DONE;
    }
}

UChar32
UCharCharacterIterator::last32() {
    pos = end;
    if(pos > begin) {
        UChar32 c;
        U16_PREV(text, begin, pos, c);
        return c;
    } else {
        return DONE;
    }
}

UChar32
UCharCharacterIterator::setIndex32(int32_t position) {
    if(position < begin) {
        position = begin;
    } else if(position > end) {
        position = end;
    }
    if(position < end) {
        U16_SET_CP_START(text, begin, position);
        int32_t i = this->pos = position;
        UChar32 c;
        U16_NEXT(text, i, end, c);
        return c;
    } else {
        this->pos = position;
        return DONE;
    }
}

UChar32
UCharCharacterIterator::current32() const {
    if (pos >= begin && pos < end) {
        UChar32 c;
        U16_GET(text, begin, pos, end, c);
        return c;
    } else {
        return DONE;
    }
}

UChar32
UCharCharacterIterator::next32() {
    if (pos < end) {
        U16_FWD_1(text, pos, end);
        if(pos < end) {
            int32_t i = pos;
            UChar32 c;
            U16_NEXT(text, i, end, c);
            return c;
        }
    }
    /* make current() return DONE */
    pos = end;
    return DONE;
}

UChar32
UCharCharacterIterator::next32PostInc() {
    if (pos < end) {
        UChar32 c;
        U16_NEXT(text, pos, end, c);
        return c;
    } else {
        return DONE;
    }
}

UChar32
UCharCharacterIterator::previous32() {
    if (pos > begin) {
        UChar32 c;
        U16_PREV(text, begin, pos, c);
        return c;
    } else {
        return DONE;
    }
}

int32_t
UCharCharacterIterator::move(int32_t delta, CharacterIterator::EOrigin origin) {
    switch(origin) {
    case kStart:
        pos = begin + delta;
        break;
    case kCurrent:
        pos += delta;
        break;
    case kEnd:
        pos = end + delta;
        break;
    default:
        break;
    }

    if(pos < begin) {
        pos = begin;
    } else if(pos > end) {
        pos = end;
    }

    return pos;
}

int32_t
UCharCharacterIterator::move32(int32_t delta, CharacterIterator::EOrigin origin) {
    // this implementation relies on the "safe" version of the UTF macros
    // (or the trustworthiness of the caller)
    switch(origin) {
    case kStart:
        pos = begin;
        if(delta > 0) {
            U16_FWD_N(text, pos, end, delta);
        }
        break;
    case kCurrent:
        if(delta > 0) {
            U16_FWD_N(text, pos, end, delta);
        } else {
            U16_BACK_N(text, begin, pos, -delta);
        }
        break;
    case kEnd:
        pos = end;
        if(delta < 0) {
            U16_BACK_N(text, begin, pos, -delta);
        }
        break;
    default:
        break;
    }

    return pos;
}

void UCharCharacterIterator::setText(ConstChar16Ptr newText,
                                     int32_t      newTextLength) {
    text = newText;
    if(newText == 0 || newTextLength < 0) {
        newTextLength = 0;
    }
    end = textLength = newTextLength;
    pos = begin = 0;
}

void
UCharCharacterIterator::getText(UnicodeString& result) {
    result = UnicodeString(text, textLength);
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2001-2014, International Business Machines
*                Corporation and others. All Rights Reserved.
******************************************************************************
*   file name:  ucln_cmn.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001July05
*   created by: George Rhoten
*/








// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2003-2009, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  utracimp.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2003aug06
*   created by: Markus W. Scherer
*
*   Internal header for ICU tracing/logging.
*
*
*   Various notes:
*   - using a trace level variable to only call trace functions
*     when the level is sufficient
*   - using the same variable for tracing on/off to never make a function
*     call when off
*   - the function number is put into a local variable by the entry macro
*     and used implicitly to avoid copy&paste/typing mistakes by the developer
*   - the application must call utrace_setFunctions() and pass in
*     implementations for the trace functions
*   - ICU trace macros call ICU functions that route through the function
*     pointers if they have been set;
*     this avoids an indirection at the call site
*     (which would cost more code for another check and for the indirection)
*
*   ### TODO Issues:
*   - Verify that va_list is portable among compilers for the same platform.
*     va_list should be portable because printf() would fail otherwise!
*   - Should enum values like UTraceLevel be passed into int32_t-type arguments,
*     or should enum types be used?
*/


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2003-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  utrace.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2003aug06
*   created by: Markus W. Scherer
*
*   Definitions for ICU tracing/logging.
*
*/


#include <stdarg.h>


/**
 * \file
 * \brief C API:  Definitions for ICU tracing/logging.
 *
 * This provides API for debugging the internals of ICU without the use of
 * a traditional debugger.
 *
 * By default, tracing is disabled in ICU. If you need to debug ICU with
 * tracing, please compile ICU with the --enable-tracing configure option.
 */

U_CDECL_BEGIN

/**
 * Trace severity levels.  Higher levels increase the verbosity of the trace output.
 * @see utrace_setLevel
 * @stable ICU 2.8
 */
typedef enum UTraceLevel {
    /** Disable all tracing  @stable ICU 2.8*/
    UTRACE_OFF=-1,
    /** Trace error conditions only  @stable ICU 2.8*/
    UTRACE_ERROR=0,
    /** Trace errors and warnings  @stable ICU 2.8*/
    UTRACE_WARNING=3,
    /** Trace opens and closes of ICU services  @stable ICU 2.8*/
    UTRACE_OPEN_CLOSE=5,
    /** Trace an intermediate number of ICU operations  @stable ICU 2.8*/
    UTRACE_INFO=7,
    /** Trace the maximum number of ICU operations  @stable ICU 2.8*/
    UTRACE_VERBOSE=9
} UTraceLevel;

/**
 *  These are the ICU functions that will be traced when tracing is enabled.
 *  @stable ICU 2.8
 */
typedef enum UTraceFunctionNumber {
    UTRACE_FUNCTION_START=0,
    UTRACE_U_INIT=UTRACE_FUNCTION_START,
    UTRACE_U_CLEANUP,

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal collation trace location.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UTRACE_FUNCTION_LIMIT,
#endif  // U_HIDE_DEPRECATED_API

    UTRACE_CONVERSION_START=0x1000,
    UTRACE_UCNV_OPEN=UTRACE_CONVERSION_START,
    UTRACE_UCNV_OPEN_PACKAGE,
    UTRACE_UCNV_OPEN_ALGORITHMIC,
    UTRACE_UCNV_CLONE,
    UTRACE_UCNV_CLOSE,
    UTRACE_UCNV_FLUSH_CACHE,
    UTRACE_UCNV_LOAD,
    UTRACE_UCNV_UNLOAD,

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal collation trace location.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UTRACE_CONVERSION_LIMIT,
#endif  // U_HIDE_DEPRECATED_API

    UTRACE_COLLATION_START=0x2000,
    UTRACE_UCOL_OPEN=UTRACE_COLLATION_START,
    UTRACE_UCOL_CLOSE,
    UTRACE_UCOL_STRCOLL,
    UTRACE_UCOL_GET_SORTKEY,
    UTRACE_UCOL_GETLOCALE,
    UTRACE_UCOL_NEXTSORTKEYPART,
    UTRACE_UCOL_STRCOLLITER,
    UTRACE_UCOL_OPEN_FROM_SHORT_STRING,
    UTRACE_UCOL_STRCOLLUTF8, /**< @stable ICU 50 */

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal collation trace location.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UTRACE_COLLATION_LIMIT,
#endif  // U_HIDE_DEPRECATED_API

#ifndef U_HIDE_DRAFT_API

    /**
     * The lowest resource/data location.
     * @draft ICU 65
     */
    UTRACE_UDATA_START=0x3000,

    /**
     * Indicates that a value was read from a resource bundle. Provides three
     * C-style strings to UTraceData: type, file name, and resource path. The
     * possible types are:
     *
     * - "string" (a string value was accessed)
     * - "binary" (a binary value was accessed)
     * - "intvector" (a integer vector value was accessed)
     * - "int" (a signed integer value was accessed)
     * - "uint" (a unsigned integer value was accessed)
     * - "get" (a path was loaded, but the value was not accessed)
     * - "getalias" (a path was loaded, and an alias was resolved)
     *
     * @draft ICU 65
     */
    UTRACE_UDATA_RESOURCE=UTRACE_UDATA_START,

    /**
     * Indicates that a resource bundle was opened.
     *
     * Provides one C-style string to UTraceData: file name.
     * @draft ICU 65
     */
    UTRACE_UDATA_BUNDLE,

    /**
     * Indicates that a data file was opened, but not *.res files.
     *
     * Provides one C-style string to UTraceData: file name.
     *
     * @draft ICU 65
     */
    UTRACE_UDATA_DATA_FILE,

    /**
     * Indicates that a *.res file was opened.
     *
     * This differs from UTRACE_UDATA_BUNDLE because a res file is typically
     * opened only once per application runtime, but the bundle corresponding
     * to that res file may be opened many times.
     *
     * Provides one C-style string to UTraceData: file name.
     *
     * @draft ICU 65
     */
    UTRACE_UDATA_RES_FILE,

#endif  // U_HIDE_DRAFT_API

#ifndef U_HIDE_INTERNAL_API
    /**
     * One more than the highest normal resource/data trace location.
     * @internal The numeric value may change over time, see ICU ticket #12420.
     */
    UTRACE_RES_DATA_LIMIT,
#endif  // U_HIDE_INTERNAL_API

} UTraceFunctionNumber;

/**
 * Setter for the trace level.
 * @param traceLevel A UTraceLevel value.
 * @stable ICU 2.8
 */
U_STABLE void U_EXPORT2
utrace_setLevel(int32_t traceLevel);

/**
 * Getter for the trace level.
 * @return The UTraceLevel value being used by ICU.
 * @stable ICU 2.8
 */
U_STABLE int32_t U_EXPORT2
utrace_getLevel(void);

/* Trace function pointers types  ----------------------------- */

/**
  *  Type signature for the trace function to be called when entering a function.
  *  @param context value supplied at the time the trace functions are set.
  *  @param fnNumber Enum value indicating the ICU function being entered.
  *  @stable ICU 2.8
  */
typedef void U_CALLCONV
UTraceEntry(const void *context, int32_t fnNumber);

/**
  *  Type signature for the trace function to be called when exiting from a function.
  *  @param context value supplied at the time the trace functions are set.
  *  @param fnNumber Enum value indicating the ICU function being exited.
  *  @param fmt     A formatting string that describes the number and types
  *                 of arguments included with the variable args.  The fmt
  *                 string has the same form as the utrace_vformat format
  *                 string.
  *  @param args    A variable arguments list.  Contents are described by
  *                 the fmt parameter.
  *  @see   utrace_vformat
  *  @stable ICU 2.8
  */
typedef void U_CALLCONV
UTraceExit(const void *context, int32_t fnNumber,
           const char *fmt, va_list args);

/**
  *  Type signature for the trace function to be called from within an ICU function
  *  to display data or messages.
  *  @param context  value supplied at the time the trace functions are set.
  *  @param fnNumber Enum value indicating the ICU function being exited.
  *  @param level    The current tracing level
  *  @param fmt      A format string describing the tracing data that is supplied
  *                  as variable args
  *  @param args     The data being traced, passed as variable args.
  *  @stable ICU 2.8
  */
typedef void U_CALLCONV
UTraceData(const void *context, int32_t fnNumber, int32_t level,
           const char *fmt, va_list args);

/**
  *  Set ICU Tracing functions.  Installs application-provided tracing
  *  functions into ICU.  After doing this, subsequent ICU operations
  *  will call back to the installed functions, providing a trace
  *  of the use of ICU.  Passing a NULL pointer for a tracing function
  *  is allowed, and inhibits tracing action at points where that function
  *  would be called.
  *  <p>
  *  Tracing and Threads:  Tracing functions are global to a process, and
  *  will be called in response to ICU operations performed by any
  *  thread.  If tracing of an individual thread is desired, the
  *  tracing functions must themselves filter by checking that the
  *  current thread is the desired thread.
  *
  *  @param context an uninterpreted pointer.  Whatever is passed in
  *                 here will in turn be passed to each of the tracing
  *                 functions UTraceEntry, UTraceExit and UTraceData.
  *                 ICU does not use or alter this pointer.
  *  @param e       Callback function to be called on entry to a
  *                 a traced ICU function.
  *  @param x       Callback function to be called on exit from a
  *                 traced ICU function.
  *  @param d       Callback function to be called from within a
  *                 traced ICU function, for the purpose of providing
  *                 data to the trace.
  *
  *  @stable ICU 2.8
  */
U_STABLE void U_EXPORT2
utrace_setFunctions(const void *context,
                    UTraceEntry *e, UTraceExit *x, UTraceData *d);

/**
  * Get the currently installed ICU tracing functions.   Note that a null function
  *   pointer will be returned if no trace function has been set.
  *
  * @param context  The currently installed tracing context.
  * @param e        The currently installed UTraceEntry function.
  * @param x        The currently installed UTraceExit function.
  * @param d        The currently installed UTraceData function.
  * @stable ICU 2.8
  */
U_STABLE void U_EXPORT2
utrace_getFunctions(const void **context,
                    UTraceEntry **e, UTraceExit **x, UTraceData **d);



/*
 *
 * ICU trace format string syntax
 *
 * Format Strings are passed to UTraceData functions, and define the
 * number and types of the trace data being passed on each call.
 *
 * The UTraceData function, which is supplied by the application,
 * not by ICU, can either forward the trace data (passed via
 * varargs) and the format string back to ICU for formatting into
 * a displayable string, or it can interpret the format itself,
 * and do as it wishes with the trace data.
 *
 *
 * Goals for the format string
 * - basic data output
 * - easy to use for trace programmer
 * - sufficient provision for data types for trace output readability
 * - well-defined types and binary portable APIs
 *
 * Non-goals
 * - printf compatibility
 * - fancy formatting
 * - argument reordering and other internationalization features
 *
 * ICU trace format strings contain plain text with argument inserts,
 * much like standard printf format strings.
 * Each insert begins with a '%', then optionally contains a 'v',
 * then exactly one type character.
 * Two '%' in a row represent a '%' instead of an insert.
 * The trace format strings need not have \n at the end.
 *
 *
 * Types
 * -----
 *
 * Type characters:
 * - c A char character in the default codepage.
 * - s A NUL-terminated char * string in the default codepage.
 * - S A UChar * string.  Requires two params, (ptr, length).  Length=-1 for nul term.
 * - b A byte (8-bit integer).
 * - h A 16-bit integer.  Also a 16 bit Unicode code unit.
 * - d A 32-bit integer.  Also a 20 bit Unicode code point value.
 * - l A 64-bit integer.
 * - p A data pointer.
 *
 * Vectors
 * -------
 *
 * If the 'v' is not specified, then one item of the specified type
 * is passed in.
 * If the 'v' (for "vector") is specified, then a vector of items of the
 * specified type is passed in, via a pointer to the first item
 * and an int32_t value for the length of the vector.
 * Length==-1 means zero or NUL termination.  Works for vectors of all types.
 *
 * Note:  %vS is a vector of (UChar *) strings.  The strings must
 *        be nul terminated as there is no way to provide a
 *        separate length parameter for each string.  The length
 *        parameter (required for all vectors) is the number of
 *        strings, not the length of the strings.
 *
 * Examples
 * --------
 *
 * These examples show the parameters that will be passed to an application's
 *   UTraceData() function for various formats.
 *
 * - the precise formatting is up to the application!
 * - the examples use type casts for arguments only to _show_ the types of
 *   arguments without needing variable declarations in the examples;
 *   the type casts will not be necessary in actual code
 *
 * UTraceDataFunc(context, fnNumber, level,
 *              "There is a character %c in the string %s.",   // Format String
 *              (char)c, (const char *)s);                     // varargs parameters
 * ->   There is a character 0x42 'B' in the string "Bravo".
 *
 * UTraceDataFunc(context, fnNumber, level,
 *              "Vector of bytes %vb vector of chars %vc",
 *              (const uint8_t *)bytes, (int32_t)bytesLength,
 *              (const char *)chars, (int32_t)charsLength);
 * ->  Vector of bytes
 *      42 63 64 3f [4]
 *     vector of chars
 *      "Bcd?"[4]
 *
 * UTraceDataFunc(context, fnNumber, level,
 *              "An int32_t %d and a whole bunch of them %vd",
 *              (int32_t)-5, (const int32_t *)ints, (int32_t)intsLength);
 * ->   An int32_t 0xfffffffb and a whole bunch of them
 *      fffffffb 00000005 0000010a [3]
 *
 */



/**
  *  Trace output Formatter.  An application's UTraceData tracing functions may call
  *                 back to this function to format the trace output in a
  *                 human readable form.  Note that a UTraceData function may choose
  *                 to not format the data;  it could, for example, save it in
  *                 in the raw form it was received (more compact), leaving
  *                 formatting for a later trace analysis tool.
  *  @param outBuf  pointer to a buffer to receive the formatted output.  Output
  *                 will be nul terminated if there is space in the buffer -
  *                 if the length of the requested output < the output buffer size.
  *  @param capacity  Length of the output buffer.
  *  @param indent  Number of spaces to indent the output.  Intended to allow
  *                 data displayed from nested functions to be indented for readability.
  *  @param fmt     Format specification for the data to output
  *  @param args    Data to be formatted.
  *  @return        Length of formatted output, including the terminating NUL.
  *                 If buffer capacity is insufficient, the required capacity is returned.
  *  @stable ICU 2.8
  */
U_STABLE int32_t U_EXPORT2
utrace_vformat(char *outBuf, int32_t capacity,
              int32_t indent, const char *fmt,  va_list args);

/**
  *  Trace output Formatter.  An application's UTraceData tracing functions may call
  *                 this function to format any additional trace data, beyond that
  *                 provided by default, in human readable form with the same
  *                 formatting conventions used by utrace_vformat().
  *  @param outBuf  pointer to a buffer to receive the formatted output.  Output
  *                 will be nul terminated if there is space in the buffer -
  *                 if the length of the requested output < the output buffer size.
  *  @param capacity  Length of the output buffer.
  *  @param indent  Number of spaces to indent the output.  Intended to allow
  *                 data displayed from nested functions to be indented for readability.
  *  @param fmt     Format specification for the data to output
  *  @param ...     Data to be formatted.
  *  @return        Length of formatted output, including the terminating NUL.
  *                 If buffer capacity is insufficient, the required capacity is returned.
  *  @stable ICU 2.8
  */
U_STABLE int32_t U_EXPORT2
utrace_format(char *outBuf, int32_t capacity,
              int32_t indent, const char *fmt,  ...);



/* Trace function numbers --------------------------------------------------- */

/**
 * Get the name of a function from its trace function number.
 *
 * @param fnNumber The trace number for an ICU function.
 * @return The name string for the function.
 *
 * @see UTraceFunctionNumber
 * @stable ICU 2.8
 */
U_STABLE const char * U_EXPORT2
utrace_functionName(int32_t fnNumber);

U_CDECL_END


#include <stdarg.h>

U_CDECL_BEGIN

/**
 *   Traced Function Exit return types.
 *   Flags indicating the number and types of varargs included in a call
 *   to a UTraceExit function.
 *   Bits 0-3:  The function return type.  First variable param.
 *   Bit    4:  Flag for presence of U_ErrorCode status param.
 *   @internal
 */
typedef enum UTraceExitVal {
    /** The traced function returns no value  @internal */
    UTRACE_EXITV_NONE   = 0,
    /** The traced function returns an int32_t, or compatible, type.  @internal */
    UTRACE_EXITV_I32    = 1,
    /** The traced function returns a pointer  @internal */
    UTRACE_EXITV_PTR    = 2,
    /** The traced function returns a UBool  @internal */
    UTRACE_EXITV_BOOL   = 3,
    /** Mask to extract the return type values from a UTraceExitVal  @internal */
    UTRACE_EXITV_MASK   = 0xf,
    /** Bit indicating that the traced function includes a UErrorCode parameter  @internal */
    UTRACE_EXITV_STATUS = 0x10
} UTraceExitVal;

/**
 * Trace function for the entry point of a function.
 * Do not use directly, use UTRACE_ENTRY instead.
 * @param fnNumber The UTraceFunctionNumber for the current function.
 * @internal
 */
U_CAPI void U_EXPORT2
utrace_entry(int32_t fnNumber);

/**
 * Trace function for each exit point of a function.
 * Do not use directly, use UTRACE_EXIT* instead.
 * @param fnNumber The UTraceFunctionNumber for the current function.
 * @param returnType The type of the value returned by the function.
 * @param errorCode The UErrorCode value at function exit. See UTRACE_EXIT.
 * @internal
 */
U_CAPI void U_EXPORT2
utrace_exit(int32_t fnNumber, int32_t returnType, ...);


/**
 * Trace function used inside functions that have a UTRACE_ENTRY() statement.
 * Do not use directly, use UTRACE_DATAX() macros instead.
 *
 * @param utraceFnNumber The number of the current function, from the local
 *        variable of the same name.
 * @param level The trace level for this message.
 * @param fmt The trace format string.
 *
 * @internal
 */
U_CAPI void U_EXPORT2
utrace_data(int32_t utraceFnNumber, int32_t level, const char *fmt, ...);

U_CDECL_END

#if U_ENABLE_TRACING

/**
 * Boolean expression to see if ICU tracing is turned on
 * to at least the specified level.
 * @internal
 */
#define UTRACE_LEVEL(level) (utrace_getLevel()>=(level))

/**
  *  Flag bit in utraceFnNumber, the local variable added to each function
  *  with tracing code to contains the function number.
  *
  *  Set the flag if the function's entry is traced, which will cause the
  *  function's exit to also be traced.  utraceFnNumber is uncoditionally
  *  set at entry, whether or not the entry is traced, so that it will
  *  always be available for error trace output.
  *  @internal
  */
#define UTRACE_TRACED_ENTRY 0x80000000

/**
 * Trace statement for the entry point of a function.
 * Stores the function number in a local variable.
 * In C code, must be placed immediately after the last variable declaration.
 * Must be matched with UTRACE_EXIT() at all function exit points.
 *
 * Tracing should start with UTRACE_ENTRY after checking for
 * U_FAILURE at function entry, so that if a function returns immediately
 * because of a pre-existing error condition, it does not show up in the trace,
 * consistent with ICU's error handling model.
 *
 * @param fnNumber The UTraceFunctionNumber for the current function.
 * @internal
 */
#define UTRACE_ENTRY(fnNumber) \
    int32_t utraceFnNumber=(fnNumber); \
UPRV_BLOCK_MACRO_BEGIN { \
    if(utrace_getLevel()>=UTRACE_INFO) { \
        utrace_entry(fnNumber); \
        utraceFnNumber |= UTRACE_TRACED_ENTRY; \
    } \
} UPRV_BLOCK_MACRO_END


/**
 * Trace statement for the entry point of open and close functions.
 * Produces trace output at a less verbose setting than plain UTRACE_ENTRY
 * Stores the function number in a local variable.
 * In C code, must be placed immediately after the last variable declaration.
 * Must be matched with UTRACE_EXIT() at all function exit points.
 *
 * @param fnNumber The UTraceFunctionNumber for the current function.
 * @internal
 */
#define UTRACE_ENTRY_OC(fnNumber) \
    int32_t utraceFnNumber=(fnNumber); \
UPRV_BLOCK_MACRO_BEGIN { \
    if(utrace_getLevel()>=UTRACE_OPEN_CLOSE) { \
        utrace_entry(fnNumber); \
        utraceFnNumber |= UTRACE_TRACED_ENTRY; \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement for each exit point of a function that has a UTRACE_ENTRY()
 * statement.
 *
 * @param errorCode The function's ICU UErrorCode value at function exit,
 *                  or U_ZERO_ERROR if the function does not use a UErrorCode.
 *                  0==U_ZERO_ERROR indicates success,
 *                  positive values an error (see u_errorName()),
 *                  negative values an informational status.
 *
 * @internal
 */
#define UTRACE_EXIT() UPRV_BLOCK_MACRO_BEGIN { \
    if(utraceFnNumber & UTRACE_TRACED_ENTRY) { \
        utrace_exit(utraceFnNumber & ~UTRACE_TRACED_ENTRY, UTRACE_EXITV_NONE); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement for each exit point of a function that has a UTRACE_ENTRY()
 * statement, and that returns a value.
 *
 * @param val       The function's return value, int32_t or comatible type.
 *
 * @internal
 */
#define UTRACE_EXIT_VALUE(val) UPRV_BLOCK_MACRO_BEGIN { \
    if(utraceFnNumber & UTRACE_TRACED_ENTRY) { \
        utrace_exit(utraceFnNumber & ~UTRACE_TRACED_ENTRY, UTRACE_EXITV_I32, val); \
    } \
} UPRV_BLOCK_MACRO_END

#define UTRACE_EXIT_STATUS(status) UPRV_BLOCK_MACRO_BEGIN { \
    if(utraceFnNumber & UTRACE_TRACED_ENTRY) { \
        utrace_exit(utraceFnNumber & ~UTRACE_TRACED_ENTRY, UTRACE_EXITV_STATUS, status); \
    } \
} UPRV_BLOCK_MACRO_END

#define UTRACE_EXIT_VALUE_STATUS(val, status) UPRV_BLOCK_MACRO_BEGIN { \
    if(utraceFnNumber & UTRACE_TRACED_ENTRY) { \
        utrace_exit(utraceFnNumber & ~UTRACE_TRACED_ENTRY, (UTRACE_EXITV_I32 | UTRACE_EXITV_STATUS), val, status); \
    } \
} UPRV_BLOCK_MACRO_END

#define UTRACE_EXIT_PTR_STATUS(ptr, status) UPRV_BLOCK_MACRO_BEGIN { \
    if(utraceFnNumber & UTRACE_TRACED_ENTRY) { \
        utrace_exit(utraceFnNumber & ~UTRACE_TRACED_ENTRY, (UTRACE_EXITV_PTR | UTRACE_EXITV_STATUS), ptr, status); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement used inside functions that have a UTRACE_ENTRY() statement.
 * Takes no data arguments.
 * The number of arguments for this macro must match the number of inserts
 * in the format string. Vector inserts count as two arguments.
 * Calls utrace_data() if the level is high enough.
 * @internal
 */
#define UTRACE_DATA0(level, fmt) UPRV_BLOCK_MACRO_BEGIN { \
    if(UTRACE_LEVEL(level)) { \
        utrace_data(utraceFnNumber & ~UTRACE_TRACED_ENTRY, (level), (fmt)); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement used inside functions that have a UTRACE_ENTRY() statement.
 * Takes one data argument.
 * The number of arguments for this macro must match the number of inserts
 * in the format string. Vector inserts count as two arguments.
 * Calls utrace_data() if the level is high enough.
 * @internal
 */
#define UTRACE_DATA1(level, fmt, a) UPRV_BLOCK_MACRO_BEGIN { \
    if(UTRACE_LEVEL(level)) { \
        utrace_data(utraceFnNumber & ~UTRACE_TRACED_ENTRY , (level), (fmt), (a)); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement used inside functions that have a UTRACE_ENTRY() statement.
 * Takes two data arguments.
 * The number of arguments for this macro must match the number of inserts
 * in the format string. Vector inserts count as two arguments.
 * Calls utrace_data() if the level is high enough.
 * @internal
 */
#define UTRACE_DATA2(level, fmt, a, b) UPRV_BLOCK_MACRO_BEGIN { \
    if(UTRACE_LEVEL(level)) { \
        utrace_data(utraceFnNumber & ~UTRACE_TRACED_ENTRY , (level), (fmt), (a), (b)); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement used inside functions that have a UTRACE_ENTRY() statement.
 * Takes three data arguments.
 * The number of arguments for this macro must match the number of inserts
 * in the format string. Vector inserts count as two arguments.
 * Calls utrace_data() if the level is high enough.
 * @internal
 */
#define UTRACE_DATA3(level, fmt, a, b, c) UPRV_BLOCK_MACRO_BEGIN { \
    if(UTRACE_LEVEL(level)) { \
        utrace_data(utraceFnNumber & ~UTRACE_TRACED_ENTRY, (level), (fmt), (a), (b), (c)); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement used inside functions that have a UTRACE_ENTRY() statement.
 * Takes four data arguments.
 * The number of arguments for this macro must match the number of inserts
 * in the format string. Vector inserts count as two arguments.
 * Calls utrace_data() if the level is high enough.
 * @internal
 */
#define UTRACE_DATA4(level, fmt, a, b, c, d) UPRV_BLOCK_MACRO_BEGIN { \
    if(UTRACE_LEVEL(level)) { \
        utrace_data(utraceFnNumber & ~UTRACE_TRACED_ENTRY, (level), (fmt), (a), (b), (c), (d)); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement used inside functions that have a UTRACE_ENTRY() statement.
 * Takes five data arguments.
 * The number of arguments for this macro must match the number of inserts
 * in the format string. Vector inserts count as two arguments.
 * Calls utrace_data() if the level is high enough.
 * @internal
 */
#define UTRACE_DATA5(level, fmt, a, b, c, d, e) UPRV_BLOCK_MACRO_BEGIN { \
    if(UTRACE_LEVEL(level)) { \
        utrace_data(utraceFnNumber & ~UTRACE_TRACED_ENTRY, (level), (fmt), (a), (b), (c), (d), (e)); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement used inside functions that have a UTRACE_ENTRY() statement.
 * Takes six data arguments.
 * The number of arguments for this macro must match the number of inserts
 * in the format string. Vector inserts count as two arguments.
 * Calls utrace_data() if the level is high enough.
 * @internal
 */
#define UTRACE_DATA6(level, fmt, a, b, c, d, e, f) UPRV_BLOCK_MACRO_BEGIN { \
    if(UTRACE_LEVEL(level)) { \
        utrace_data(utraceFnNumber & ~UTRACE_TRACED_ENTRY, (level), (fmt), (a), (b), (c), (d), (e), (f)); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement used inside functions that have a UTRACE_ENTRY() statement.
 * Takes seven data arguments.
 * The number of arguments for this macro must match the number of inserts
 * in the format string. Vector inserts count as two arguments.
 * Calls utrace_data() if the level is high enough.
 * @internal
 */
#define UTRACE_DATA7(level, fmt, a, b, c, d, e, f, g) UPRV_BLOCK_MACRO_BEGIN { \
    if(UTRACE_LEVEL(level)) { \
        utrace_data(utraceFnNumber & ~UTRACE_TRACED_ENTRY, (level), (fmt), (a), (b), (c), (d), (e), (f), (g)); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement used inside functions that have a UTRACE_ENTRY() statement.
 * Takes eight data arguments.
 * The number of arguments for this macro must match the number of inserts
 * in the format string. Vector inserts count as two arguments.
 * Calls utrace_data() if the level is high enough.
 * @internal
 */
#define UTRACE_DATA8(level, fmt, a, b, c, d, e, f, g, h) UPRV_BLOCK_MACRO_BEGIN { \
    if(UTRACE_LEVEL(level)) { \
        utrace_data(utraceFnNumber & ~UTRACE_TRACED_ENTRY, (level), (fmt), (a), (b), (c), (d), (e), (f), (g), (h)); \
    } \
} UPRV_BLOCK_MACRO_END

/**
 * Trace statement used inside functions that have a UTRACE_ENTRY() statement.
 * Takes nine data arguments.
 * The number of arguments for this macro must match the number of inserts
 * in the format string. Vector inserts count as two arguments.
 * Calls utrace_data() if the level is high enough.
 * @internal
 */
#define UTRACE_DATA9(level, fmt, a, b, c, d, e, f, g, h, i) UPRV_BLOCK_MACRO_BEGIN { \
    if(UTRACE_LEVEL(level)) { \
        utrace_data(utraceFnNumber & ~UTRACE_TRACED_ENTRY, (level), (fmt), (a), (b), (c), (d), (e), (f), (g), (h), (i)); \
    } \
} UPRV_BLOCK_MACRO_END

#else

/*
 * When tracing is disabled, the following macros become empty
 */

#define UTRACE_LEVEL(level) 0
#define UTRACE_ENTRY(fnNumber)
#define UTRACE_ENTRY_OC(fnNumber)
#define UTRACE_EXIT()
#define UTRACE_EXIT_VALUE(val)
#define UTRACE_EXIT_STATUS(status)
#define UTRACE_EXIT_VALUE_STATUS(val, status)
#define UTRACE_EXIT_PTR_STATUS(ptr, status)
#define UTRACE_DATA0(level, fmt)
#define UTRACE_DATA1(level, fmt, a)
#define UTRACE_DATA2(level, fmt, a, b)
#define UTRACE_DATA3(level, fmt, a, b, c)
#define UTRACE_DATA4(level, fmt, a, b, c, d)
#define UTRACE_DATA5(level, fmt, a, b, c, d, e)
#define UTRACE_DATA6(level, fmt, a, b, c, d, e, f)
#define UTRACE_DATA7(level, fmt, a, b, c, d, e, f, g)
#define UTRACE_DATA8(level, fmt, a, b, c, d, e, f, g, h)
#define UTRACE_DATA9(level, fmt, a, b, c, d, e, f, g, h, i)

#endif




/**  Auto-client for UCLN_COMMON **/
#define UCLN_TYPE_IS_COMMON
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
* Copyright (C) 2009-2011, International Business Machines
*                Corporation and others. All Rights Reserved.
*
******************************************************************************
*   file name:  ucln_imp.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   This file contains the platform specific implementation of per-library cleanup.
*
*/




#include <stdlib.h>

/**
 * Auto cleanup of ICU libraries
 * There are several methods in per library cleanup of icu libraries:
 * 1) Compiler/Platform based cleanup:
 *   a) Windows MSVC uses DllMain()
 *   b) GCC uses destructor function attribute
 *   c) Sun Studio, AIX VA, and HP-UX aCC uses a linker option to set the exit function
 * 2) Using atexit()
 * 3) Implementing own automatic cleanup functions
 *
 * For option 1, ensure that UCLN_NO_AUTO_CLEANUP is set to 0 by using --enable-auto-cleanup
 * configure option or by otherwise setting UCLN_NO_AUTO_CLEANUP to 0
 * For option 2, follow option 1 and also define UCLN_AUTO_ATEXIT
 * For option 3, follow option 1 and also define UCLN_AUTO_LOCAL (see below for more information)
 */

#if !UCLN_NO_AUTO_CLEANUP

/*
 * The following declarations are for when UCLN_AUTO_LOCAL or UCLN_AUTO_ATEXIT
 * are defined. They are commented out because they are static and will be defined
 * later. The information is still here to provide some guidance for the developer
 * who chooses to use UCLN_AUTO_LOCAL.
 */
/**
 * Give the library an opportunity to register an automatic cleanup.
 * This may be called more than once.
 */
/*static void ucln_registerAutomaticCleanup();*/
/**
 * Unregister an automatic cleanup, if possible. Called from cleanup.
 */
/*static void ucln_unRegisterAutomaticCleanup();*/

#ifdef UCLN_TYPE_IS_COMMON
#   define UCLN_CLEAN_ME_UP u_cleanup()
#else
#   define UCLN_CLEAN_ME_UP ucln_cleanupOne(UCLN_TYPE)
#endif

/* ------------ automatic cleanup: registration. Choose ONE ------- */
#if defined(UCLN_AUTO_LOCAL)
/* To use:
 *  1. define UCLN_AUTO_LOCAL,
 *  2. create ucln_local_hook.c containing implementations of
 *           static void ucln_registerAutomaticCleanup()
 *           static void ucln_unRegisterAutomaticCleanup()
 */
#include "ucln_local_hook.c"

#elif defined(UCLN_AUTO_ATEXIT)
/*
 * Use the ANSI C 'atexit' function. Note that this mechanism does not
 * guarantee the order of cleanup relative to other users of ICU!
 */
static UBool gAutoCleanRegistered = FALSE;

static void ucln_atexit_handler()
{
    UCLN_CLEAN_ME_UP;
}

static void ucln_registerAutomaticCleanup()
{
    if(!gAutoCleanRegistered) {
        gAutoCleanRegistered = TRUE;
        atexit(&ucln_atexit_handler);
    }
}

static void ucln_unRegisterAutomaticCleanup () {
}
/* ------------end of automatic cleanup: registration. ------- */

#elif defined (UCLN_FINI)
/**
 * If UCLN_FINI is defined, it is the (versioned, etc) name of a cleanup
 * entrypoint. Add a stub to call ucln_cleanupOne
 * Used on AIX, Solaris, and HP-UX
 */
U_CAPI void U_EXPORT2 UCLN_FINI (void);

U_CAPI void U_EXPORT2 UCLN_FINI ()
{
    /* This function must be defined, if UCLN_FINI is defined, else link error. */
     UCLN_CLEAN_ME_UP;
}

/* Windows: DllMain */
#elif U_PLATFORM_HAS_WIN32_API
/*
 * ICU's own DllMain.
 */

/* these are from putil.c */
/* READ READ READ READ!    Are you getting compilation errors from windows.h?
          Any source file which includes this (ucln_imp.h) header MUST
          be defined with language extensions ON. */
#ifndef WIN32_LEAN_AND_MEAN
#   define WIN32_LEAN_AND_MEAN
#endif
#   define VC_EXTRALEAN
#   define NOUSER
#   define NOSERVICE
#   define NOIME
#   define NOMCX
#   include <windows.h>
/*
 * This is a stub DllMain function with icu specific process handling code.
 */
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
    BOOL status = TRUE;

    switch(fdwReason) {
        case DLL_PROCESS_ATTACH:
             /* ICU does not trap process attach, but must pass these through properly. */
            /* ICU specific process attach could go here */
            break;

        case DLL_PROCESS_DETACH:
            /* Here is the one we actually care about. */

            UCLN_CLEAN_ME_UP;

            break;

        case DLL_THREAD_ATTACH:
            /* ICU does not trap thread attach, but must pass these through properly. */
            /* ICU specific thread attach could go here */
            break;

        case DLL_THREAD_DETACH:
            /* ICU does not trap thread detach, but must pass these through properly. */
            /* ICU specific thread detach could go here */
            break;

    }
    return status;
}

#elif defined(__GNUC__)
/* GCC - use __attribute((destructor)) */
static void ucln_destructor()   __attribute__((destructor)) ;

static void ucln_destructor()
{
    UCLN_CLEAN_ME_UP;
}

#endif

#endif /* UCLN_NO_AUTO_CLEANUP */

// #else
// #error This file can only be included once.


static cleanupFunc *gCommonCleanupFunctions[UCLN_COMMON_COUNT];
static cleanupFunc *gLibCleanupFunctions[UCLN_COMMON];


/************************************************
 The cleanup order is important in this function.
 Please be sure that you have read ucln.h
 ************************************************/
U_CAPI void U_EXPORT2
u_cleanup(void)
{
    UTRACE_ENTRY_OC(UTRACE_U_CLEANUP);
    icu::umtx_lock(NULL);     /* Force a memory barrier, so that we are sure to see   */
    icu::umtx_unlock(NULL);   /*   all state left around by any other threads.        */

    ucln_lib_cleanup();

    cmemory_cleanup();       /* undo any heap functions set by u_setMemoryFunctions(). */
    UTRACE_EXIT();           /* Must be before utrace_cleanup(), which turns off tracing. */
/*#if U_ENABLE_TRACING*/
    utrace_cleanup();
/*#endif*/
}

U_CAPI void U_EXPORT2 ucln_cleanupOne(ECleanupLibraryType libType)
{
    if (gLibCleanupFunctions[libType])
    {
        gLibCleanupFunctions[libType]();
        gLibCleanupFunctions[libType] = NULL;
    }
}

U_CFUNC void
ucln_common_registerCleanup(ECleanupCommonType type,
                            cleanupFunc *func)
{
    // Thread safety messiness: From ticket 10295, calls to registerCleanup() may occur
    // concurrently. Although such cases should be storing the same value, they raise errors
    // from the thread sanity checker. Doing the store within a mutex avoids those.
    // BUT that can trigger a recursive entry into std::call_once() in umutex.cpp when this code,
    // running from the call_once function, tries to grab the ICU global mutex, which
    // re-enters the mutex init path. So, work-around by special casing UCLN_COMMON_MUTEX, not
    // using the ICU global mutex for it.
    //
    // No other point in ICU uses std::call_once().

    U_ASSERT(UCLN_COMMON_START < type && type < UCLN_COMMON_COUNT);
    if (type == UCLN_COMMON_MUTEX) {
        gCommonCleanupFunctions[type] = func;
    } else if (UCLN_COMMON_START < type && type < UCLN_COMMON_COUNT)  {
        icu::Mutex m;     // See ticket 10295 for discussion.
        gCommonCleanupFunctions[type] = func;
    }
#if !UCLN_NO_AUTO_CLEANUP && (defined(UCLN_AUTO_ATEXIT) || defined(UCLN_AUTO_LOCAL))
    ucln_registerAutomaticCleanup();
#endif
}

// Note: ucln_registerCleanup() is called with the ICU global mutex locked.
//       Be aware if adding anything to the function.
//       See ticket 10295 for discussion.

U_CAPI void U_EXPORT2
ucln_registerCleanup(ECleanupLibraryType type,
                     cleanupFunc *func)
{
    U_ASSERT(UCLN_START < type && type < UCLN_COMMON);
    if (UCLN_START < type && type < UCLN_COMMON)
    {
        gLibCleanupFunctions[type] = func;
    }
}

U_CFUNC UBool ucln_lib_cleanup(void) {
    int32_t libType = UCLN_START;
    int32_t commonFunc = UCLN_COMMON_START;

    for (libType++; libType<UCLN_COMMON; libType++) {
        ucln_cleanupOne(static_cast<ECleanupLibraryType>(libType));
    }

    for (commonFunc++; commonFunc<UCLN_COMMON_COUNT; commonFunc++) {
        if (gCommonCleanupFunctions[commonFunc])
        {
            gCommonCleanupFunctions[commonFunc]();
            gCommonCleanupFunctions[commonFunc] = NULL;
        }
    }
#if !UCLN_NO_AUTO_CLEANUP && (defined(UCLN_AUTO_ATEXIT) || defined(UCLN_AUTO_LOCAL))
    ucln_unRegisterAutomaticCleanup();
#endif
    return TRUE;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************/


/*------------------------------------------------------------------------------
 *
 *   UCommonData   An abstract interface for dealing with ICU Common Data Files.
 *                 ICU Common Data Files are a grouping of a number of individual
 *                 data items (resources, converters, tables, anything) into a
 *                 single file or dll.  The combined format includes a table of
 *                 contents for locating the individual items by name.
 *
 *                 Two formats for the table of contents are supported, which is
 *                 why there is an abstract inteface involved.
 *
 */







#if defined(UDATA_DEBUG) || defined(UDATA_DEBUG_DUMP)
#   include <stdio.h>
#endif

U_CFUNC uint16_t
udata_getHeaderSize(const DataHeader *udh) {
    if(udh==NULL) {
        return 0;
    } else if(udh->info.isBigEndian==U_IS_BIG_ENDIAN) {
        /* same endianness */
        return udh->dataHeader.headerSize;
    } else {
        /* opposite endianness */
        uint16_t x=udh->dataHeader.headerSize;
        return (uint16_t)((x<<8)|(x>>8));
    }
}

U_CFUNC uint16_t
udata_getInfoSize(const UDataInfo *info) {
    if(info==NULL) {
        return 0;
    } else if(info->isBigEndian==U_IS_BIG_ENDIAN) {
        /* same endianness */
        return info->size;
    } else {
        /* opposite endianness */
        uint16_t x=info->size;
        return (uint16_t)((x<<8)|(x>>8));
    }
}

/*-----------------------------------------------------------------------------*
 *                                                                             *
 *  Pointer TOCs.   TODO: This form of table-of-contents should be removed     *
 *                  because DLLs must be relocated on loading to correct the   *
 *                  pointer values and this operation makes shared memory      *
 *                  mapping of the data much less likely to work.              *
 *                                                                             *
 *-----------------------------------------------------------------------------*/
typedef struct {
    const char       *entryName;
    const DataHeader *pHeader;
} PointerTOCEntry;


typedef struct  {
    uint32_t          count;
    uint32_t          reserved;
    /**
     * Variable-length array declared with length 1 to disable bounds checkers.
     * The actual array length is in the count field.
     */
    PointerTOCEntry   entry[1];
}  PointerTOC;


/* definition of OffsetTOC struct types moved to ucmndata.h */

/*-----------------------------------------------------------------------------*
 *                                                                             *
 *    entry point lookup implementations                                       *
 *                                                                             *
 *-----------------------------------------------------------------------------*/

#ifndef MIN
#define MIN(a,b) (((a)<(b)) ? (a) : (b))
#endif

/**
 * Compare strings where we know the shared prefix length,
 * and advance the prefix length as we find that the strings share even more characters.
 */
static int32_t
strcmpAfterPrefix(const char *s1, const char *s2, int32_t *pPrefixLength) {
    int32_t pl=*pPrefixLength;
    int32_t cmp=0;
    s1+=pl;
    s2+=pl;
    for(;;) {
        int32_t c1=(uint8_t)*s1++;
        int32_t c2=(uint8_t)*s2++;
        cmp=c1-c2;
        if(cmp!=0 || c1==0) {  /* different or done */
            break;
        }
        ++pl;  /* increment shared same-prefix length */
    }
    *pPrefixLength=pl;
    return cmp;
}

static int32_t
offsetTOCPrefixBinarySearch(const char *s, const char *names,
                            const UDataOffsetTOCEntry *toc, int32_t count) {
    int32_t start=0;
    int32_t limit=count;
    /*
     * Remember the shared prefix between s, start and limit,
     * and don't compare that shared prefix again.
     * The shared prefix should get longer as we narrow the [start, limit[ range.
     */
    int32_t startPrefixLength=0;
    int32_t limitPrefixLength=0;
    if(count==0) {
        return -1;
    }
    /*
     * Prime the prefix lengths so that we don't keep prefixLength at 0 until
     * both the start and limit indexes have moved.
     * At the same time, we find if s is one of the start and (limit-1) names,
     * and if not, exclude them from the actual binary search.
     */
    if(0==strcmpAfterPrefix(s, names+toc[0].nameOffset, &startPrefixLength)) {
        return 0;
    }
    ++start;
    --limit;
    if(0==strcmpAfterPrefix(s, names+toc[limit].nameOffset, &limitPrefixLength)) {
        return limit;
    }
    while(start<limit) {
        int32_t i=(start+limit)/2;
        int32_t prefixLength=MIN(startPrefixLength, limitPrefixLength);
        int32_t cmp=strcmpAfterPrefix(s, names+toc[i].nameOffset, &prefixLength);
        if(cmp<0) {
            limit=i;
            limitPrefixLength=prefixLength;
        } else if(cmp==0) {
            return i;
        } else {
            start=i+1;
            startPrefixLength=prefixLength;
        }
    }
    return -1;
}

static int32_t
pointerTOCPrefixBinarySearch(const char *s, const PointerTOCEntry *toc, int32_t count) {
    int32_t start=0;
    int32_t limit=count;
    /*
     * Remember the shared prefix between s, start and limit,
     * and don't compare that shared prefix again.
     * The shared prefix should get longer as we narrow the [start, limit[ range.
     */
    int32_t startPrefixLength=0;
    int32_t limitPrefixLength=0;
    if(count==0) {
        return -1;
    }
    /*
     * Prime the prefix lengths so that we don't keep prefixLength at 0 until
     * both the start and limit indexes have moved.
     * At the same time, we find if s is one of the start and (limit-1) names,
     * and if not, exclude them from the actual binary search.
     */
    if(0==strcmpAfterPrefix(s, toc[0].entryName, &startPrefixLength)) {
        return 0;
    }
    ++start;
    --limit;
    if(0==strcmpAfterPrefix(s, toc[limit].entryName, &limitPrefixLength)) {
        return limit;
    }
    while(start<limit) {
        int32_t i=(start+limit)/2;
        int32_t prefixLength=MIN(startPrefixLength, limitPrefixLength);
        int32_t cmp=strcmpAfterPrefix(s, toc[i].entryName, &prefixLength);
        if(cmp<0) {
            limit=i;
            limitPrefixLength=prefixLength;
        } else if(cmp==0) {
            return i;
        } else {
            start=i+1;
            startPrefixLength=prefixLength;
        }
    }
    return -1;
}

U_CDECL_BEGIN
static uint32_t U_CALLCONV
offsetTOCEntryCount(const UDataMemory *pData) {
    int32_t          retVal=0;
    const UDataOffsetTOC *toc = (UDataOffsetTOC *)pData->toc;
    if (toc != NULL) {
        retVal = toc->count;
    }
    return retVal;
}

static const DataHeader * U_CALLCONV
offsetTOCLookupFn(const UDataMemory *pData,
                  const char *tocEntryName,
                  int32_t *pLength,
                  UErrorCode *pErrorCode) {
    (void)pErrorCode;
    const UDataOffsetTOC  *toc = (UDataOffsetTOC *)pData->toc;
    if(toc!=NULL) {
        const char *base=(const char *)toc;
        int32_t number, count=(int32_t)toc->count;

        /* perform a binary search for the data in the common data's table of contents */
#if defined (UDATA_DEBUG_DUMP)
        /* list the contents of the TOC each time .. not recommended */
        for(number=0; number<count; ++number) {
            fprintf(stderr, "\tx%d: %s\n", number, &base[toc->entry[number].nameOffset]);
        }
#endif
        number=offsetTOCPrefixBinarySearch(tocEntryName, base, toc->entry, count);
        if(number>=0) {
            /* found it */
            const UDataOffsetTOCEntry *entry=toc->entry+number;
#ifdef UDATA_DEBUG
            fprintf(stderr, "%s: Found.\n", tocEntryName);
#endif
            if((number+1) < count) {
                *pLength = (int32_t)(entry[1].dataOffset - entry->dataOffset);
            } else {
                *pLength = -1;
            }
            return (const DataHeader *)(base+entry->dataOffset);
        } else {
#ifdef UDATA_DEBUG
            fprintf(stderr, "%s: Not found.\n", tocEntryName);
#endif
            return NULL;
        }
    } else {
#ifdef UDATA_DEBUG
        fprintf(stderr, "returning header\n");
#endif

        return pData->pHeader;
    }
}


static uint32_t U_CALLCONV pointerTOCEntryCount(const UDataMemory *pData) {
    const PointerTOC *toc = (PointerTOC *)pData->toc;
    return (uint32_t)((toc != NULL) ? (toc->count) : 0);
}

static const DataHeader * U_CALLCONV pointerTOCLookupFn(const UDataMemory *pData,
                   const char *name,
                   int32_t *pLength,
                   UErrorCode *pErrorCode) {
    (void)pErrorCode;
    if(pData->toc!=NULL) {
        const PointerTOC *toc = (PointerTOC *)pData->toc;
        int32_t number, count=(int32_t)toc->count;

#if defined (UDATA_DEBUG_DUMP)
        /* list the contents of the TOC each time .. not recommended */
        for(number=0; number<count; ++number) {
            fprintf(stderr, "\tx%d: %s\n", number, toc->entry[number].entryName);
        }
#endif
        number=pointerTOCPrefixBinarySearch(name, toc->entry, count);
        if(number>=0) {
            /* found it */
#ifdef UDATA_DEBUG
            fprintf(stderr, "%s: Found.\n", toc->entry[number].entryName);
#endif
            *pLength=-1;
            return UDataMemory_normalizeDataPointer(toc->entry[number].pHeader);
        } else {
#ifdef UDATA_DEBUG
            fprintf(stderr, "%s: Not found.\n", name);
#endif
            return NULL;
        }
    } else {
        return pData->pHeader;
    }
}
U_CDECL_END


static const commonDataFuncs CmnDFuncs = {offsetTOCLookupFn,  offsetTOCEntryCount};
static const commonDataFuncs ToCPFuncs = {pointerTOCLookupFn, pointerTOCEntryCount};



/*----------------------------------------------------------------------*
 *                                                                      *
 *  checkCommonData   Validate the format of a common data file.        *
 *                    Fill in the virtual function ptr based on TOC type *
 *                    If the data is invalid, close the UDataMemory     *
 *                    and set the appropriate error code.               *
 *                                                                      *
 *----------------------------------------------------------------------*/
U_CFUNC void udata_checkCommonData(UDataMemory *udm, UErrorCode *err) {
    if (U_FAILURE(*err)) {
        return;
    }

    if(udm==NULL || udm->pHeader==NULL) {
      *err=U_INVALID_FORMAT_ERROR;
    } else if(!(udm->pHeader->dataHeader.magic1==0xda &&
        udm->pHeader->dataHeader.magic2==0x27 &&
        udm->pHeader->info.isBigEndian==U_IS_BIG_ENDIAN &&
        udm->pHeader->info.charsetFamily==U_CHARSET_FAMILY)
        ) {
        /* header not valid */
        *err=U_INVALID_FORMAT_ERROR;
    }
    else if (udm->pHeader->info.dataFormat[0]==0x43 &&
        udm->pHeader->info.dataFormat[1]==0x6d &&
        udm->pHeader->info.dataFormat[2]==0x6e &&
        udm->pHeader->info.dataFormat[3]==0x44 &&
        udm->pHeader->info.formatVersion[0]==1
        ) {
        /* dataFormat="CmnD" */
        udm->vFuncs = &CmnDFuncs;
        udm->toc=(const char *)udm->pHeader+udata_getHeaderSize(udm->pHeader);
    }
    else if(udm->pHeader->info.dataFormat[0]==0x54 &&
        udm->pHeader->info.dataFormat[1]==0x6f &&
        udm->pHeader->info.dataFormat[2]==0x43 &&
        udm->pHeader->info.dataFormat[3]==0x50 &&
        udm->pHeader->info.formatVersion[0]==1
        ) {
        /* dataFormat="ToCP" */
        udm->vFuncs = &ToCPFuncs;
        udm->toc=(const char *)udm->pHeader+udata_getHeaderSize(udm->pHeader);
    }
    else {
        /* dataFormat not recognized */
        *err=U_INVALID_FORMAT_ERROR;
    }

    if (U_FAILURE(*err)) {
        /* If the data is no good and we memory-mapped it ourselves,
         *  close the memory mapping so it doesn't leak.  Note that this has
         *  no effect on non-memory mapped data, other than clearing fields in udm.
         */
        udata_close(udm);
    }
}

/*
 * TODO: Add a udata_swapPackageHeader() function that swaps an ICU .dat package
 * header but not its sub-items.
 * This function will be needed for automatic runtime swapping.
 * Sub-items should not be swapped to limit the swapping to the parts of the
 * package that are actually used.
 *
 * Since lengths of items are implicit in the order and offsets of their
 * ToC entries, and since offsets are relative to the start of the ToC,
 * a swapped version may need to generate a different data structure
 * with pointers to the original data items and with their lengths
 * (-1 for the last one if it is not known), and maybe even pointers to the
 * swapped versions of the items.
 * These pointers to swapped versions would establish a cache;
 * instead, each open data item could simply own the storage for its swapped
 * data. This fits better with the current design.
 *
 * markus 2003sep18 Jitterbug 2235
 */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// ******************************************************************************
// *
// *   Copyright (C) 1998-2016, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *
// ******************************************************************************
// *
// *  ucnv.c:
// *  Implements APIs for the ICU's codeset conversion library;
// *  mostly calls through internal functions;
// *  created by Bertrand A. Damiba
// *
// * Modification History:
// *
// *   Date        Name        Description
// *   04/04/99    helena      Fixed internal header inclusion.
// *   05/09/00    helena      Added implementation to handle fallback mappings.
// *   06/20/2000  helena      OS/400 port changes; mostly typecast.
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include <memory>

// #include "unicode/ustring.h"
// #include "unicode/ucnv.h"
// #include "unicode/ucnv_err.h"
// #include "unicode/uset.h"
// #include "unicode/utf.h"
// #include "unicode/utf16.h"
// #include "putilimp.h"
// #include "cmemory.h"
// #include "cstring.h"
// #include "uassert.h"
// #include "utracimp.h"
// #include "ustr_imp.h"
// #include "ucnv_imp.h"
// #include "ucnv_cnv.h"
// #include "ucnv_bld.h"

// /* size of intermediate and preflighting buffers in ucnv_convert() */
// #define CHUNK_SIZE 1024

// typedef struct UAmbiguousConverter {
//     const char *name;
//     const UChar variant5c;
// } UAmbiguousConverter;

// static const UAmbiguousConverter ambiguousConverters[]={
//     { "ibm-897_P100-1995", 0xa5 },
//     { "ibm-942_P120-1999", 0xa5 },
//     { "ibm-943_P130-1999", 0xa5 },
//     { "ibm-946_P100-1995", 0xa5 },
//     { "ibm-33722_P120-1999", 0xa5 },
//     { "ibm-1041_P100-1995", 0xa5 },
//     /*{ "ibm-54191_P100-2006", 0xa5 },*/
//     /*{ "ibm-62383_P100-2007", 0xa5 },*/
//     /*{ "ibm-891_P100-1995", 0x20a9 },*/
//     { "ibm-944_P100-1995", 0x20a9 },
//     { "ibm-949_P110-1999", 0x20a9 },
//     { "ibm-1363_P110-1997", 0x20a9 },
//     { "ISO_2022,locale=ko,version=0", 0x20a9 },
//     { "ibm-1088_P100-1995", 0x20a9 }
// };

// /*Calls through createConverter */
// U_CAPI UConverter* U_EXPORT2
// ucnv_open (const char *name,
//                        UErrorCode * err)
// {
//     UConverter *r;

//     if (err == NULL || U_FAILURE (*err)) {
//         return NULL;
//     }

//     r =  ucnv_createConverter(NULL, name, err);
//     return r;
// }

// U_CAPI UConverter* U_EXPORT2
// ucnv_openPackage   (const char *packageName, const char *converterName, UErrorCode * err)
// {
//     return ucnv_createConverterFromPackage(packageName, converterName,  err);
// }

// /*Extracts the UChar* to a char* and calls through createConverter */
// U_CAPI UConverter*   U_EXPORT2
// ucnv_openU (const UChar * name,
//                          UErrorCode * err)
// {
//     char asciiName[UCNV_MAX_CONVERTER_NAME_LENGTH];

//     if (err == NULL || U_FAILURE(*err))
//         return NULL;
//     if (name == NULL)
//         return ucnv_open (NULL, err);
//     if (u_strlen(name) >= UCNV_MAX_CONVERTER_NAME_LENGTH)
//     {
//         *err = U_ILLEGAL_ARGUMENT_ERROR;
//         return NULL;
//     }
//     return ucnv_open(u_austrcpy(asciiName, name), err);
// }

// /* Copy the string that is represented by the UConverterPlatform enum
//  * @param platformString An output buffer
//  * @param platform An enum representing a platform
//  * @return the length of the copied string.
//  */
// static int32_t
// ucnv_copyPlatformString(char *platformString, UConverterPlatform pltfrm)
// {
//     switch (pltfrm)
//     {
//     case UCNV_IBM:
//         uprv_strcpy(platformString, "ibm-");
//         return 4;
//     case UCNV_UNKNOWN:
//         break;
//     }

//     /* default to empty string */
//     *platformString = 0;
//     return 0;
// }

// /*Assumes a $platform-#codepage.$CONVERTER_FILE_EXTENSION scheme and calls
//  *through createConverter*/
// U_CAPI UConverter*   U_EXPORT2
// ucnv_openCCSID (int32_t codepage,
//                 UConverterPlatform platform,
//                 UErrorCode * err)
// {
//     char myName[UCNV_MAX_CONVERTER_NAME_LENGTH];
//     int32_t myNameLen;

//     if (err == NULL || U_FAILURE (*err))
//         return NULL;

//     /* ucnv_copyPlatformString could return "ibm-" or "cp" */
//     myNameLen = ucnv_copyPlatformString(myName, platform);
//     T_CString_integerToString(myName + myNameLen, codepage, 10);

//     return ucnv_createConverter(NULL, myName, err);
// }

// /* Creating a temporary stack-based object that can be used in one thread,
// and created from a converter that is shared across threads.
// */

// U_CAPI UConverter* U_EXPORT2
// ucnv_safeClone(const UConverter* cnv, void *stackBuffer, int32_t *pBufferSize, UErrorCode *status)
// {
//     UConverter *localConverter, *allocatedConverter;
//     int32_t stackBufferSize;
//     int32_t bufferSizeNeeded;
//     UErrorCode cbErr;
//     UConverterToUnicodeArgs toUArgs = {
//         sizeof(UConverterToUnicodeArgs),
//             TRUE,
//             NULL,
//             NULL,
//             NULL,
//             NULL,
//             NULL,
//             NULL
//     };
//     UConverterFromUnicodeArgs fromUArgs = {
//         sizeof(UConverterFromUnicodeArgs),
//             TRUE,
//             NULL,
//             NULL,
//             NULL,
//             NULL,
//             NULL,
//             NULL
//     };

//     UTRACE_ENTRY_OC(UTRACE_UCNV_CLONE);

//     if (status == NULL || U_FAILURE(*status)){
//         UTRACE_EXIT_STATUS(status? *status: U_ILLEGAL_ARGUMENT_ERROR);
//         return NULL;
//     }

//     if (cnv == NULL) {
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         UTRACE_EXIT_STATUS(*status);
//         return NULL;
//     }

//     UTRACE_DATA3(UTRACE_OPEN_CLOSE, "clone converter %s at %p into stackBuffer %p",
//                                     ucnv_getName(cnv, status), cnv, stackBuffer);

//     if (cnv->sharedData->impl->safeClone != NULL) {
//         /* call the custom safeClone function for sizing */
//         bufferSizeNeeded = 0;
//         cnv->sharedData->impl->safeClone(cnv, NULL, &bufferSizeNeeded, status);
//         if (U_FAILURE(*status)) {
//             UTRACE_EXIT_STATUS(*status);
//             return NULL;
//         }
//     }
//     else
//     {
//         /* inherent sizing */
//         bufferSizeNeeded = sizeof(UConverter);
//     }

//     if (pBufferSize == NULL) {
//         stackBufferSize = 1;
//         pBufferSize = &stackBufferSize;
//     } else {
//         stackBufferSize = *pBufferSize;
//         if (stackBufferSize <= 0){ /* 'preflighting' request - set needed size into *pBufferSize */
//             *pBufferSize = bufferSizeNeeded;
//             UTRACE_EXIT_VALUE(bufferSizeNeeded);
//             return NULL;
//         }
//     }

//     /* Adjust (if necessary) the stackBuffer pointer to be aligned correctly for a UConverter.
//      * TODO(Jira ICU-20736) Redo this using std::align() once g++4.9 compatibility is no longer needed.
//      */
//     if (stackBuffer) {
//         uintptr_t p = reinterpret_cast<uintptr_t>(stackBuffer);
//         uintptr_t aligned_p = (p + alignof(UConverter) - 1) & ~(alignof(UConverter) - 1);
//         ptrdiff_t pointerAdjustment = aligned_p - p;
//         if (bufferSizeNeeded + pointerAdjustment <= stackBufferSize) {
//             stackBuffer = reinterpret_cast<void *>(aligned_p);
//             stackBufferSize -= static_cast<int32_t>(pointerAdjustment);
//         } else {
//             /* prevent using the stack buffer but keep the size > 0 so that we do not just preflight */
//             stackBufferSize = 1;
//         }
//     }

//     /* Now, see if we must allocate any memory */
//     if (stackBufferSize < bufferSizeNeeded || stackBuffer == NULL)
//     {
//         /* allocate one here...*/
//         localConverter = allocatedConverter = (UConverter *) uprv_malloc (bufferSizeNeeded);

//         if(localConverter == NULL) {
//             *status = U_MEMORY_ALLOCATION_ERROR;
//             UTRACE_EXIT_STATUS(*status);
//             return NULL;
//         }
//         *status = U_SAFECLONE_ALLOCATED_WARNING;

//         /* record the fact that memory was allocated */
//         *pBufferSize = bufferSizeNeeded;
//     } else {
//         /* just use the stack buffer */
//         localConverter = (UConverter*) stackBuffer;
//         allocatedConverter = NULL;
//     }

//     uprv_memset(localConverter, 0, bufferSizeNeeded);

//     /* Copy initial state */
//     uprv_memcpy(localConverter, cnv, sizeof(UConverter));
//     localConverter->isCopyLocal = localConverter->isExtraLocal = FALSE;

//     /* copy the substitution string */
//     if (cnv->subChars == (uint8_t *)cnv->subUChars) {
//         localConverter->subChars = (uint8_t *)localConverter->subUChars;
//     } else {
//         localConverter->subChars = (uint8_t *)uprv_malloc(UCNV_ERROR_BUFFER_LENGTH * U_SIZEOF_UCHAR);
//         if (localConverter->subChars == NULL) {
//             uprv_free(allocatedConverter);
//             UTRACE_EXIT_STATUS(*status);
//             return NULL;
//         }
//         uprv_memcpy(localConverter->subChars, cnv->subChars, UCNV_ERROR_BUFFER_LENGTH * U_SIZEOF_UCHAR);
//     }

//     /* now either call the safeclone fcn or not */
//     if (cnv->sharedData->impl->safeClone != NULL) {
//         /* call the custom safeClone function */
//         localConverter = cnv->sharedData->impl->safeClone(cnv, localConverter, pBufferSize, status);
//     }

//     if(localConverter==NULL || U_FAILURE(*status)) {
//         if (allocatedConverter != NULL && allocatedConverter->subChars != (uint8_t *)allocatedConverter->subUChars) {
//             uprv_free(allocatedConverter->subChars);
//         }
//         uprv_free(allocatedConverter);
//         UTRACE_EXIT_STATUS(*status);
//         return NULL;
//     }

//     /* increment refcount of shared data if needed */
//     if (cnv->sharedData->isReferenceCounted) {
//         ucnv_incrementRefCount(cnv->sharedData);
//     }

//     if(localConverter == (UConverter*)stackBuffer) {
//         /* we're using user provided data - set to not destroy */
//         localConverter->isCopyLocal = TRUE;
//     }

//     /* allow callback functions to handle any memory allocation */
//     toUArgs.converter = fromUArgs.converter = localConverter;
//     cbErr = U_ZERO_ERROR;
//     cnv->fromCharErrorBehaviour(cnv->toUContext, &toUArgs, NULL, 0, UCNV_CLONE, &cbErr);
//     cbErr = U_ZERO_ERROR;
//     cnv->fromUCharErrorBehaviour(cnv->fromUContext, &fromUArgs, NULL, 0, 0, UCNV_CLONE, &cbErr);

//     UTRACE_EXIT_PTR_STATUS(localConverter, *status);
//     return localConverter;
// }



// /*Decreases the reference counter in the shared immutable section of the object
//  *and frees the mutable part*/

// U_CAPI void  U_EXPORT2
// ucnv_close (UConverter * converter)
// {
//     UErrorCode errorCode = U_ZERO_ERROR;

//     UTRACE_ENTRY_OC(UTRACE_UCNV_CLOSE);

//     if (converter == NULL)
//     {
//         UTRACE_EXIT();
//         return;
//     }

//     UTRACE_DATA3(UTRACE_OPEN_CLOSE, "close converter %s at %p, isCopyLocal=%b",
//         ucnv_getName(converter, &errorCode), converter, converter->isCopyLocal);

//     /* In order to speed up the close, only call the callbacks when they have been changed.
//     This performance check will only work when the callbacks are set within a shared library
//     or from user code that statically links this code. */
//     /* first, notify the callback functions that the converter is closed */
//     if (converter->fromCharErrorBehaviour != UCNV_TO_U_DEFAULT_CALLBACK) {
//         UConverterToUnicodeArgs toUArgs = {
//             sizeof(UConverterToUnicodeArgs),
//                 TRUE,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL
//         };

//         toUArgs.converter = converter;
//         errorCode = U_ZERO_ERROR;
//         converter->fromCharErrorBehaviour(converter->toUContext, &toUArgs, NULL, 0, UCNV_CLOSE, &errorCode);
//     }
//     if (converter->fromUCharErrorBehaviour != UCNV_FROM_U_DEFAULT_CALLBACK) {
//         UConverterFromUnicodeArgs fromUArgs = {
//             sizeof(UConverterFromUnicodeArgs),
//                 TRUE,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL
//         };
//         fromUArgs.converter = converter;
//         errorCode = U_ZERO_ERROR;
//         converter->fromUCharErrorBehaviour(converter->fromUContext, &fromUArgs, NULL, 0, 0, UCNV_CLOSE, &errorCode);
//     }

//     if (converter->sharedData->impl->close != NULL) {
//         converter->sharedData->impl->close(converter);
//     }

//     if (converter->subChars != (uint8_t *)converter->subUChars) {
//         uprv_free(converter->subChars);
//     }

//     if (converter->sharedData->isReferenceCounted) {
//         ucnv_unloadSharedDataIfReady(converter->sharedData);
//     }

//     if(!converter->isCopyLocal){
//         uprv_free(converter);
//     }

//     UTRACE_EXIT();
// }

// /*returns a single Name from the list, will return NULL if out of bounds
//  */
// U_CAPI const char*   U_EXPORT2
// ucnv_getAvailableName (int32_t n)
// {
//     if (0 <= n && n <= 0xffff) {
//         UErrorCode err = U_ZERO_ERROR;
//         const char *name = ucnv_bld_getAvailableConverter((uint16_t)n, &err);
//         if (U_SUCCESS(err)) {
//             return name;
//         }
//     }
//     return NULL;
// }

// U_CAPI int32_t   U_EXPORT2
// ucnv_countAvailable ()
// {
//     UErrorCode err = U_ZERO_ERROR;
//     return ucnv_bld_countAvailableConverters(&err);
// }

// U_CAPI void    U_EXPORT2
// ucnv_getSubstChars (const UConverter * converter,
//                     char *mySubChar,
//                     int8_t * len,
//                     UErrorCode * err)
// {
//     if (U_FAILURE (*err))
//         return;

//     if (converter->subCharLen <= 0) {
//         /* Unicode string or empty string from ucnv_setSubstString(). */
//         *len = 0;
//         return;
//     }

//     if (*len < converter->subCharLen) /*not enough space in subChars */
//     {
//         *err = U_INDEX_OUTOFBOUNDS_ERROR;
//         return;
//     }

//     uprv_memcpy (mySubChar, converter->subChars, converter->subCharLen);   /*fills in the subchars */
//     *len = converter->subCharLen; /*store # of bytes copied to buffer */
// }

// U_CAPI void    U_EXPORT2
// ucnv_setSubstChars (UConverter * converter,
//                     const char *mySubChar,
//                     int8_t len,
//                     UErrorCode * err)
// {
//     if (U_FAILURE (*err))
//         return;

//     /*Makes sure that the subChar is within the codepages char length boundaries */
//     if ((len > converter->sharedData->staticData->maxBytesPerChar)
//      || (len < converter->sharedData->staticData->minBytesPerChar))
//     {
//         *err = U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

//     uprv_memcpy (converter->subChars, mySubChar, len); /*copies the subchars */
//     converter->subCharLen = len;  /*sets the new len */

//     /*
//     * There is currently (2001Feb) no separate API to set/get subChar1.
//     * In order to always have subChar written after it is explicitly set,
//     * we set subChar1 to 0.
//     */
//     converter->subChar1 = 0;

//     return;
// }

// U_CAPI void U_EXPORT2
// ucnv_setSubstString(UConverter *cnv,
//                     const UChar *s,
//                     int32_t length,
//                     UErrorCode *err) {
//     alignas(UConverter) char cloneBuffer[U_CNV_SAFECLONE_BUFFERSIZE];
//     char chars[UCNV_ERROR_BUFFER_LENGTH];

//     UConverter *clone;
//     uint8_t *subChars;
//     int32_t cloneSize, length8;

//     /* Let the following functions check all arguments. */
//     cloneSize = sizeof(cloneBuffer);
//     clone = ucnv_safeClone(cnv, cloneBuffer, &cloneSize, err);
//     ucnv_setFromUCallBack(clone, UCNV_FROM_U_CALLBACK_STOP, NULL, NULL, NULL, err);
//     length8 = ucnv_fromUChars(clone, chars, (int32_t)sizeof(chars), s, length, err);
//     ucnv_close(clone);
//     if (U_FAILURE(*err)) {
//         return;
//     }

//     if (cnv->sharedData->impl->writeSub == NULL
// #if !UCONFIG_NO_LEGACY_CONVERSION
//         || (cnv->sharedData->staticData->conversionType == UCNV_MBCS &&
//          ucnv_MBCSGetType(cnv) != UCNV_EBCDIC_STATEFUL)
// #endif
//     ) {
//         /* The converter is not stateful. Store the charset bytes as a fixed string. */
//         subChars = (uint8_t *)chars;
//     } else {
//         /*
//          * The converter has a non-default writeSub() function, indicating
//          * that it is stateful.
//          * Store the Unicode string for on-the-fly conversion for correct
//          * state handling.
//          */
//         if (length > UCNV_ERROR_BUFFER_LENGTH) {
//             /*
//              * Should not occur. The converter should output at least one byte
//              * per UChar, which means that ucnv_fromUChars() should catch all
//              * overflows.
//              */
//             *err = U_BUFFER_OVERFLOW_ERROR;
//             return;
//         }
//         subChars = (uint8_t *)s;
//         if (length < 0) {
//             length = u_strlen(s);
//         }
//         length8 = length * U_SIZEOF_UCHAR;
//     }

//     /*
//      * For storing the substitution string, select either the small buffer inside
//      * UConverter or allocate a subChars buffer.
//      */
//     if (length8 > UCNV_MAX_SUBCHAR_LEN) {
//         /* Use a separate buffer for the string. Outside UConverter to not make it too large. */
//         if (cnv->subChars == (uint8_t *)cnv->subUChars) {
//             /* Allocate a new buffer for the string. */
//             cnv->subChars = (uint8_t *)uprv_malloc(UCNV_ERROR_BUFFER_LENGTH * U_SIZEOF_UCHAR);
//             if (cnv->subChars == NULL) {
//                 cnv->subChars = (uint8_t *)cnv->subUChars;
//                 *err = U_MEMORY_ALLOCATION_ERROR;
//                 return;
//             }
//             uprv_memset(cnv->subChars, 0, UCNV_ERROR_BUFFER_LENGTH * U_SIZEOF_UCHAR);
//         }
//     }

//     /* Copy the substitution string into the UConverter or its subChars buffer. */
//     if (length8 == 0) {
//         cnv->subCharLen = 0;
//     } else {
//         uprv_memcpy(cnv->subChars, subChars, length8);
//         if (subChars == (uint8_t *)chars) {
//             cnv->subCharLen = (int8_t)length8;
//         } else /* subChars == s */ {
//             cnv->subCharLen = (int8_t)-length;
//         }
//     }

//     /* See comment in ucnv_setSubstChars(). */
//     cnv->subChar1 = 0;
// }

// /*resets the internal states of a converter
//  *goal : have the same behaviour than a freshly created converter
//  */
// static void _reset(UConverter *converter, UConverterResetChoice choice,
//                    UBool callCallback) {
//     if(converter == NULL) {
//         return;
//     }

//     if(callCallback) {
//         /* first, notify the callback functions that the converter is reset */
//         UErrorCode errorCode;

//         if(choice<=UCNV_RESET_TO_UNICODE && converter->fromCharErrorBehaviour != UCNV_TO_U_DEFAULT_CALLBACK) {
//             UConverterToUnicodeArgs toUArgs = {
//                 sizeof(UConverterToUnicodeArgs),
//                 TRUE,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL
//             };
//             toUArgs.converter = converter;
//             errorCode = U_ZERO_ERROR;
//             converter->fromCharErrorBehaviour(converter->toUContext, &toUArgs, NULL, 0, UCNV_RESET, &errorCode);
//         }
//         if(choice!=UCNV_RESET_TO_UNICODE && converter->fromUCharErrorBehaviour != UCNV_FROM_U_DEFAULT_CALLBACK) {
//             UConverterFromUnicodeArgs fromUArgs = {
//                 sizeof(UConverterFromUnicodeArgs),
//                 TRUE,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL,
//                 NULL
//             };
//             fromUArgs.converter = converter;
//             errorCode = U_ZERO_ERROR;
//             converter->fromUCharErrorBehaviour(converter->fromUContext, &fromUArgs, NULL, 0, 0, UCNV_RESET, &errorCode);
//         }
//     }

//     /* now reset the converter itself */
//     if(choice<=UCNV_RESET_TO_UNICODE) {
//         converter->toUnicodeStatus = converter->sharedData->toUnicodeStatus;
//         converter->mode = 0;
//         converter->toULength = 0;
//         converter->invalidCharLength = converter->UCharErrorBufferLength = 0;
//         converter->preToULength = 0;
//     }
//     if(choice!=UCNV_RESET_TO_UNICODE) {
//         converter->fromUnicodeStatus = 0;
//         converter->fromUChar32 = 0;
//         converter->invalidUCharLength = converter->charErrorBufferLength = 0;
//         converter->preFromUFirstCP = U_SENTINEL;
//         converter->preFromULength = 0;
//     }

//     if (converter->sharedData->impl->reset != NULL) {
//         /* call the custom reset function */
//         converter->sharedData->impl->reset(converter, choice);
//     }
// }

// U_CAPI void  U_EXPORT2
// ucnv_reset(UConverter *converter)
// {
//     _reset(converter, UCNV_RESET_BOTH, TRUE);
// }

// U_CAPI void  U_EXPORT2
// ucnv_resetToUnicode(UConverter *converter)
// {
//     _reset(converter, UCNV_RESET_TO_UNICODE, TRUE);
// }

// U_CAPI void  U_EXPORT2
// ucnv_resetFromUnicode(UConverter *converter)
// {
//     _reset(converter, UCNV_RESET_FROM_UNICODE, TRUE);
// }

// U_CAPI int8_t   U_EXPORT2
// ucnv_getMaxCharSize (const UConverter * converter)
// {
//     return converter->maxBytesPerUChar;
// }


// U_CAPI int8_t   U_EXPORT2
// ucnv_getMinCharSize (const UConverter * converter)
// {
//     return converter->sharedData->staticData->minBytesPerChar;
// }

// U_CAPI const char*   U_EXPORT2
// ucnv_getName (const UConverter * converter, UErrorCode * err)

// {
//     if (U_FAILURE (*err))
//         return NULL;
//     if(converter->sharedData->impl->getName){
//         const char* temp= converter->sharedData->impl->getName(converter);
//         if(temp)
//             return temp;
//     }
//     return converter->sharedData->staticData->name;
// }

// U_CAPI int32_t U_EXPORT2
// ucnv_getCCSID(const UConverter * converter,
//               UErrorCode * err)
// {
//     int32_t ccsid;
//     if (U_FAILURE (*err))
//         return -1;

//     ccsid = converter->sharedData->staticData->codepage;
//     if (ccsid == 0) {
//         /* Rare case. This is for cases like gb18030,
//         which doesn't have an IBM canonical name, but does have an IBM alias. */
//         const char *standardName = ucnv_getStandardName(ucnv_getName(converter, err), "IBM", err);
//         if (U_SUCCESS(*err) && standardName) {
//             const char *ccsidStr = uprv_strchr(standardName, '-');
//             if (ccsidStr) {
//                 ccsid = (int32_t)atol(ccsidStr+1);  /* +1 to skip '-' */
//             }
//         }
//     }
//     return ccsid;
// }


// U_CAPI UConverterPlatform   U_EXPORT2
// ucnv_getPlatform (const UConverter * converter,
//                                       UErrorCode * err)
// {
//     if (U_FAILURE (*err))
//         return UCNV_UNKNOWN;

//     return (UConverterPlatform)converter->sharedData->staticData->platform;
// }

// U_CAPI void U_EXPORT2
//     ucnv_getToUCallBack (const UConverter * converter,
//                          UConverterToUCallback *action,
//                          const void **context)
// {
//     *action = converter->fromCharErrorBehaviour;
//     *context = converter->toUContext;
// }

// U_CAPI void U_EXPORT2
//     ucnv_getFromUCallBack (const UConverter * converter,
//                            UConverterFromUCallback *action,
//                            const void **context)
// {
//     *action = converter->fromUCharErrorBehaviour;
//     *context = converter->fromUContext;
// }

// U_CAPI void    U_EXPORT2
// ucnv_setToUCallBack (UConverter * converter,
//                             UConverterToUCallback newAction,
//                             const void* newContext,
//                             UConverterToUCallback *oldAction,
//                             const void** oldContext,
//                             UErrorCode * err)
// {
//     if (U_FAILURE (*err))
//         return;
//     if (oldAction) *oldAction = converter->fromCharErrorBehaviour;
//     converter->fromCharErrorBehaviour = newAction;
//     if (oldContext) *oldContext = converter->toUContext;
//     converter->toUContext = newContext;
// }

// U_CAPI void  U_EXPORT2
// ucnv_setFromUCallBack (UConverter * converter,
//                             UConverterFromUCallback newAction,
//                             const void* newContext,
//                             UConverterFromUCallback *oldAction,
//                             const void** oldContext,
//                             UErrorCode * err)
// {
//     if (U_FAILURE (*err))
//         return;
//     if (oldAction) *oldAction = converter->fromUCharErrorBehaviour;
//     converter->fromUCharErrorBehaviour = newAction;
//     if (oldContext) *oldContext = converter->fromUContext;
//     converter->fromUContext = newContext;
// }

// static void
// _updateOffsets(int32_t *offsets, int32_t length,
//                int32_t sourceIndex, int32_t errorInputLength) {
//     int32_t *limit;
//     int32_t delta, offset;

//     if(sourceIndex>=0) {
//         /*
//          * adjust each offset by adding the previous sourceIndex
//          * minus the length of the input sequence that caused an
//          * error, if any
//          */
//         delta=sourceIndex-errorInputLength;
//     } else {
//         /*
//          * set each offset to -1 because this conversion function
//          * does not handle offsets
//          */
//         delta=-1;
//     }

//     limit=offsets+length;
//     if(delta==0) {
//         /* most common case, nothing to do */
//     } else if(delta>0) {
//         /* add the delta to each offset (but not if the offset is <0) */
//         while(offsets<limit) {
//             offset=*offsets;
//             if(offset>=0) {
//                 *offsets=offset+delta;
//             }
//             ++offsets;
//         }
//     } else /* delta<0 */ {
//         /*
//          * set each offset to -1 because this conversion function
//          * does not handle offsets
//          * or the error input sequence started in a previous buffer
//          */
//         while(offsets<limit) {
//             *offsets++=-1;
//         }
//     }
// }

// /* ucnv_fromUnicode --------------------------------------------------------- */

// /*
//  * Implementation note for m:n conversions
//  *
//  * While collecting source units to find the longest match for m:n conversion,
//  * some source units may need to be stored for a partial match.
//  * When a second buffer does not yield a match on all of the previously stored
//  * source units, then they must be "replayed", i.e., fed back into the converter.
//  *
//  * The code relies on the fact that replaying will not nest -
//  * converting a replay buffer will not result in a replay.
//  * This is because a replay is necessary only after the _continuation_ of a
//  * partial match failed, but a replay buffer is converted as a whole.
//  * It may result in some of its units being stored again for a partial match,
//  * but there will not be a continuation _during_ the replay which could fail.
//  *
//  * It is conceivable that a callback function could call the converter
//  * recursively in a way that causes another replay to be stored, but that
//  * would be an error in the callback function.
//  * Such violations will cause assertion failures in a debug build,
//  * and wrong output, but they will not cause a crash.
//  */

// static void
// _fromUnicodeWithCallback(UConverterFromUnicodeArgs *pArgs, UErrorCode *err) {
//     UConverterFromUnicode fromUnicode;
//     UConverter *cnv;
//     const UChar *s;
//     char *t;
//     int32_t *offsets;
//     int32_t sourceIndex;
//     int32_t errorInputLength;
//     UBool converterSawEndOfInput, calledCallback;

//     /* variables for m:n conversion */
//     UChar replay[UCNV_EXT_MAX_UCHARS];
//     const UChar *realSource, *realSourceLimit;
//     int32_t realSourceIndex;
//     UBool realFlush;

//     cnv=pArgs->converter;
//     s=pArgs->source;
//     t=pArgs->target;
//     offsets=pArgs->offsets;

//     /* get the converter implementation function */
//     sourceIndex=0;
//     if(offsets==NULL) {
//         fromUnicode=cnv->sharedData->impl->fromUnicode;
//     } else {
//         fromUnicode=cnv->sharedData->impl->fromUnicodeWithOffsets;
//         if(fromUnicode==NULL) {
//             /* there is no WithOffsets implementation */
//             fromUnicode=cnv->sharedData->impl->fromUnicode;
//             /* we will write -1 for each offset */
//             sourceIndex=-1;
//         }
//     }

//     if(cnv->preFromULength>=0) {
//         /* normal mode */
//         realSource=NULL;

//         /* avoid compiler warnings - not otherwise necessary, and the values do not matter */
//         realSourceLimit=NULL;
//         realFlush=FALSE;
//         realSourceIndex=0;
//     } else {
//         /*
//          * Previous m:n conversion stored source units from a partial match
//          * and failed to consume all of them.
//          * We need to "replay" them from a temporary buffer and convert them first.
//          */
//         realSource=pArgs->source;
//         realSourceLimit=pArgs->sourceLimit;
//         realFlush=pArgs->flush;
//         realSourceIndex=sourceIndex;

//         uprv_memcpy(replay, cnv->preFromU, -cnv->preFromULength*U_SIZEOF_UCHAR);
//         pArgs->source=replay;
//         pArgs->sourceLimit=replay-cnv->preFromULength;
//         pArgs->flush=FALSE;
//         sourceIndex=-1;

//         cnv->preFromULength=0;
//     }

//     /*
//      * loop for conversion and error handling
//      *
//      * loop {
//      *   convert
//      *   loop {
//      *     update offsets
//      *     handle end of input
//      *     handle errors/call callback
//      *   }
//      * }
//      */
//     for(;;) {
//         if(U_SUCCESS(*err)) {
//             /* convert */
//             fromUnicode(pArgs, err);

//             /*
//              * set a flag for whether the converter
//              * successfully processed the end of the input
//              *
//              * need not check cnv->preFromULength==0 because a replay (<0) will cause
//              * s<sourceLimit before converterSawEndOfInput is checked
//              */
//             converterSawEndOfInput=
//                 (UBool)(U_SUCCESS(*err) &&
//                         pArgs->flush && pArgs->source==pArgs->sourceLimit &&
//                         cnv->fromUChar32==0);
//         } else {
//             /* handle error from ucnv_convertEx() */
//             converterSawEndOfInput=FALSE;
//         }

//         /* no callback called yet for this iteration */
//         calledCallback=FALSE;

//         /* no sourceIndex adjustment for conversion, only for callback output */
//         errorInputLength=0;

//         /*
//          * loop for offsets and error handling
//          *
//          * iterates at most 3 times:
//          * 1. to clean up after the conversion function
//          * 2. after the callback
//          * 3. after the callback again if there was truncated input
//          */
//         for(;;) {
//             /* update offsets if we write any */
//             if(offsets!=NULL) {
//                 int32_t length=(int32_t)(pArgs->target-t);
//                 if(length>0) {
//                     _updateOffsets(offsets, length, sourceIndex, errorInputLength);

//                     /*
//                      * if a converter handles offsets and updates the offsets
//                      * pointer at the end, then pArgs->offset should not change
//                      * here;
//                      * however, some converters do not handle offsets at all
//                      * (sourceIndex<0) or may not update the offsets pointer
//                      */
//                     pArgs->offsets=offsets+=length;
//                 }

//                 if(sourceIndex>=0) {
//                     sourceIndex+=(int32_t)(pArgs->source-s);
//                 }
//             }

//             if(cnv->preFromULength<0) {
//                 /*
//                  * switch the source to new replay units (cannot occur while replaying)
//                  * after offset handling and before end-of-input and callback handling
//                  */
//                 if(realSource==NULL) {
//                     realSource=pArgs->source;
//                     realSourceLimit=pArgs->sourceLimit;
//                     realFlush=pArgs->flush;
//                     realSourceIndex=sourceIndex;

//                     uprv_memcpy(replay, cnv->preFromU, -cnv->preFromULength*U_SIZEOF_UCHAR);
//                     pArgs->source=replay;
//                     pArgs->sourceLimit=replay-cnv->preFromULength;
//                     pArgs->flush=FALSE;
//                     if((sourceIndex+=cnv->preFromULength)<0) {
//                         sourceIndex=-1;
//                     }

//                     cnv->preFromULength=0;
//                 } else {
//                     /* see implementation note before _fromUnicodeWithCallback() */
//                     U_ASSERT(realSource==NULL);
//                     *err=U_INTERNAL_PROGRAM_ERROR;
//                 }
//             }

//             /* update pointers */
//             s=pArgs->source;
//             t=pArgs->target;

//             if(U_SUCCESS(*err)) {
//                 if(s<pArgs->sourceLimit) {
//                     /*
//                      * continue with the conversion loop while there is still input left
//                      * (continue converting by breaking out of only the inner loop)
//                      */
//                     break;
//                 } else if(realSource!=NULL) {
//                     /* switch back from replaying to the real source and continue */
//                     pArgs->source=realSource;
//                     pArgs->sourceLimit=realSourceLimit;
//                     pArgs->flush=realFlush;
//                     sourceIndex=realSourceIndex;

//                     realSource=NULL;
//                     break;
//                 } else if(pArgs->flush && cnv->fromUChar32!=0) {
//                     /*
//                      * the entire input stream is consumed
//                      * and there is a partial, truncated input sequence left
//                      */

//                     /* inject an error and continue with callback handling */
//                     *err=U_TRUNCATED_CHAR_FOUND;
//                     calledCallback=FALSE; /* new error condition */
//                 } else {
//                     /* input consumed */
//                     if(pArgs->flush) {
//                         /*
//                          * return to the conversion loop once more if the flush
//                          * flag is set and the conversion function has not
//                          * successfully processed the end of the input yet
//                          *
//                          * (continue converting by breaking out of only the inner loop)
//                          */
//                         if(!converterSawEndOfInput) {
//                             break;
//                         }

//                         /* reset the converter without calling the callback function */
//                         _reset(cnv, UCNV_RESET_FROM_UNICODE, FALSE);
//                     }

//                     /* done successfully */
//                     return;
//                 }
//             }

//             /* U_FAILURE(*err) */
//             {
//                 UErrorCode e;

//                 if( calledCallback ||
//                     (e=*err)==U_BUFFER_OVERFLOW_ERROR ||
//                     (e!=U_INVALID_CHAR_FOUND &&
//                      e!=U_ILLEGAL_CHAR_FOUND &&
//                      e!=U_TRUNCATED_CHAR_FOUND)
//                 ) {
//                     /*
//                      * the callback did not or cannot resolve the error:
//                      * set output pointers and return
//                      *
//                      * the check for buffer overflow is redundant but it is
//                      * a high-runner case and hopefully documents the intent
//                      * well
//                      *
//                      * if we were replaying, then the replay buffer must be
//                      * copied back into the UConverter
//                      * and the real arguments must be restored
//                      */
//                     if(realSource!=NULL) {
//                         int32_t length;

//                         U_ASSERT(cnv->preFromULength==0);

//                         length=(int32_t)(pArgs->sourceLimit-pArgs->source);
//                         if(length>0) {
//                             u_memcpy(cnv->preFromU, pArgs->source, length);
//                             cnv->preFromULength=(int8_t)-length;
//                         }

//                         pArgs->source=realSource;
//                         pArgs->sourceLimit=realSourceLimit;
//                         pArgs->flush=realFlush;
//                     }

//                     return;
//                 }
//             }

//             /* callback handling */
//             {
//                 UChar32 codePoint;

//                 /* get and write the code point */
//                 codePoint=cnv->fromUChar32;
//                 errorInputLength=0;
//                 U16_APPEND_UNSAFE(cnv->invalidUCharBuffer, errorInputLength, codePoint);
//                 cnv->invalidUCharLength=(int8_t)errorInputLength;

//                 /* set the converter state to deal with the next character */
//                 cnv->fromUChar32=0;

//                 /* call the callback function */
//                 cnv->fromUCharErrorBehaviour(cnv->fromUContext, pArgs,
//                     cnv->invalidUCharBuffer, errorInputLength, codePoint,
//                     *err==U_INVALID_CHAR_FOUND ? UCNV_UNASSIGNED : UCNV_ILLEGAL,
//                     err);
//             }

//             /*
//              * loop back to the offset handling
//              *
//              * this flag will indicate after offset handling
//              * that a callback was called;
//              * if the callback did not resolve the error, then we return
//              */
//             calledCallback=TRUE;
//         }
//     }
// }

// /*
//  * Output the fromUnicode overflow buffer.
//  * Call this function if(cnv->charErrorBufferLength>0).
//  * @return TRUE if overflow
//  */
// static UBool
// ucnv_outputOverflowFromUnicode(UConverter *cnv,
//                                char **target, const char *targetLimit,
//                                int32_t **pOffsets,
//                                UErrorCode *err) {
//     int32_t *offsets;
//     char *overflow, *t;
//     int32_t i, length;

//     t=*target;
//     if(pOffsets!=NULL) {
//         offsets=*pOffsets;
//     } else {
//         offsets=NULL;
//     }

//     overflow=(char *)cnv->charErrorBuffer;
//     length=cnv->charErrorBufferLength;
//     i=0;
//     while(i<length) {
//         if(t==targetLimit) {
//             /* the overflow buffer contains too much, keep the rest */
//             int32_t j=0;

//             do {
//                 overflow[j++]=overflow[i++];
//             } while(i<length);

//             cnv->charErrorBufferLength=(int8_t)j;
//             *target=t;
//             if(offsets!=NULL) {
//                 *pOffsets=offsets;
//             }
//             *err=U_BUFFER_OVERFLOW_ERROR;
//             return TRUE;
//         }

//         /* copy the overflow contents to the target */
//         *t++=overflow[i++];
//         if(offsets!=NULL) {
//             *offsets++=-1; /* no source index available for old output */
//         }
//     }

//     /* the overflow buffer is completely copied to the target */
//     cnv->charErrorBufferLength=0;
//     *target=t;
//     if(offsets!=NULL) {
//         *pOffsets=offsets;
//     }
//     return FALSE;
// }

// U_CAPI void U_EXPORT2
// ucnv_fromUnicode(UConverter *cnv,
//                  char **target, const char *targetLimit,
//                  const UChar **source, const UChar *sourceLimit,
//                  int32_t *offsets,
//                  UBool flush,
//                  UErrorCode *err) {
//     UConverterFromUnicodeArgs args;
//     const UChar *s;
//     char *t;

//     /* check parameters */
//     if(err==NULL || U_FAILURE(*err)) {
//         return;
//     }

//     if(cnv==NULL || target==NULL || source==NULL) {
//         *err=U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

//     s=*source;
//     t=*target;

//     if ((const void *)U_MAX_PTR(sourceLimit) == (const void *)sourceLimit) {
//         /*
//         Prevent code from going into an infinite loop in case we do hit this
//         limit. The limit pointer is expected to be on a UChar * boundary.
//         This also prevents the next argument check from failing.
//         */
//         sourceLimit = (const UChar *)(((const char *)sourceLimit) - 1);
//     }

//     /*
//      * All these conditions should never happen.
//      *
//      * 1) Make sure that the limits are >= to the address source or target
//      *
//      * 2) Make sure that the buffer sizes do not exceed the number range for
//      * int32_t because some functions use the size (in units or bytes)
//      * rather than comparing pointers, and because offsets are int32_t values.
//      *
//      * size_t is guaranteed to be unsigned and large enough for the job.
//      *
//      * Return with an error instead of adjusting the limits because we would
//      * not be able to maintain the semantics that either the source must be
//      * consumed or the target filled (unless an error occurs).
//      * An adjustment would be targetLimit=t+0x7fffffff; for example.
//      *
//      * 3) Make sure that the user didn't incorrectly cast a UChar * pointer
//      * to a char * pointer and provide an incomplete UChar code unit.
//      */
//     if (sourceLimit<s || targetLimit<t ||
//         ((size_t)(sourceLimit-s)>(size_t)0x3fffffff && sourceLimit>s) ||
//         ((size_t)(targetLimit-t)>(size_t)0x7fffffff && targetLimit>t) ||
//         (((const char *)sourceLimit-(const char *)s) & 1) != 0)
//     {
//         *err=U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

//     /* output the target overflow buffer */
//     if( cnv->charErrorBufferLength>0 &&
//         ucnv_outputOverflowFromUnicode(cnv, target, targetLimit, &offsets, err)
//     ) {
//         /* U_BUFFER_OVERFLOW_ERROR */
//         return;
//     }
//     /* *target may have moved, therefore stop using t */

//     if(!flush && s==sourceLimit && cnv->preFromULength>=0) {
//         /* the overflow buffer is emptied and there is no new input: we are done */
//         return;
//     }

//     /*
//      * Do not simply return with a buffer overflow error if
//      * !flush && t==targetLimit
//      * because it is possible that the source will not generate any output.
//      * For example, the skip callback may be called;
//      * it does not output anything.
//      */

//     /* prepare the converter arguments */
//     args.converter=cnv;
//     args.flush=flush;
//     args.offsets=offsets;
//     args.source=s;
//     args.sourceLimit=sourceLimit;
//     args.target=*target;
//     args.targetLimit=targetLimit;
//     args.size=sizeof(args);

//     _fromUnicodeWithCallback(&args, err);

//     *source=args.source;
//     *target=args.target;
// }

// /* ucnv_toUnicode() --------------------------------------------------------- */

// static void
// _toUnicodeWithCallback(UConverterToUnicodeArgs *pArgs, UErrorCode *err) {
//     UConverterToUnicode toUnicode;
//     UConverter *cnv;
//     const char *s;
//     UChar *t;
//     int32_t *offsets;
//     int32_t sourceIndex;
//     int32_t errorInputLength;
//     UBool converterSawEndOfInput, calledCallback;

//     /* variables for m:n conversion */
//     char replay[UCNV_EXT_MAX_BYTES];
//     const char *realSource, *realSourceLimit;
//     int32_t realSourceIndex;
//     UBool realFlush;

//     cnv=pArgs->converter;
//     s=pArgs->source;
//     t=pArgs->target;
//     offsets=pArgs->offsets;

//     /* get the converter implementation function */
//     sourceIndex=0;
//     if(offsets==NULL) {
//         toUnicode=cnv->sharedData->impl->toUnicode;
//     } else {
//         toUnicode=cnv->sharedData->impl->toUnicodeWithOffsets;
//         if(toUnicode==NULL) {
//             /* there is no WithOffsets implementation */
//             toUnicode=cnv->sharedData->impl->toUnicode;
//             /* we will write -1 for each offset */
//             sourceIndex=-1;
//         }
//     }

//     if(cnv->preToULength>=0) {
//         /* normal mode */
//         realSource=NULL;

//         /* avoid compiler warnings - not otherwise necessary, and the values do not matter */
//         realSourceLimit=NULL;
//         realFlush=FALSE;
//         realSourceIndex=0;
//     } else {
//         /*
//          * Previous m:n conversion stored source units from a partial match
//          * and failed to consume all of them.
//          * We need to "replay" them from a temporary buffer and convert them first.
//          */
//         realSource=pArgs->source;
//         realSourceLimit=pArgs->sourceLimit;
//         realFlush=pArgs->flush;
//         realSourceIndex=sourceIndex;

//         uprv_memcpy(replay, cnv->preToU, -cnv->preToULength);
//         pArgs->source=replay;
//         pArgs->sourceLimit=replay-cnv->preToULength;
//         pArgs->flush=FALSE;
//         sourceIndex=-1;

//         cnv->preToULength=0;
//     }

//     /*
//      * loop for conversion and error handling
//      *
//      * loop {
//      *   convert
//      *   loop {
//      *     update offsets
//      *     handle end of input
//      *     handle errors/call callback
//      *   }
//      * }
//      */
//     for(;;) {
//         if(U_SUCCESS(*err)) {
//             /* convert */
//             toUnicode(pArgs, err);

//             /*
//              * set a flag for whether the converter
//              * successfully processed the end of the input
//              *
//              * need not check cnv->preToULength==0 because a replay (<0) will cause
//              * s<sourceLimit before converterSawEndOfInput is checked
//              */
//             converterSawEndOfInput=
//                 (UBool)(U_SUCCESS(*err) &&
//                         pArgs->flush && pArgs->source==pArgs->sourceLimit &&
//                         cnv->toULength==0);
//         } else {
//             /* handle error from getNextUChar() or ucnv_convertEx() */
//             converterSawEndOfInput=FALSE;
//         }

//         /* no callback called yet for this iteration */
//         calledCallback=FALSE;

//         /* no sourceIndex adjustment for conversion, only for callback output */
//         errorInputLength=0;

//         /*
//          * loop for offsets and error handling
//          *
//          * iterates at most 3 times:
//          * 1. to clean up after the conversion function
//          * 2. after the callback
//          * 3. after the callback again if there was truncated input
//          */
//         for(;;) {
//             /* update offsets if we write any */
//             if(offsets!=NULL) {
//                 int32_t length=(int32_t)(pArgs->target-t);
//                 if(length>0) {
//                     _updateOffsets(offsets, length, sourceIndex, errorInputLength);

//                     /*
//                      * if a converter handles offsets and updates the offsets
//                      * pointer at the end, then pArgs->offset should not change
//                      * here;
//                      * however, some converters do not handle offsets at all
//                      * (sourceIndex<0) or may not update the offsets pointer
//                      */
//                     pArgs->offsets=offsets+=length;
//                 }

//                 if(sourceIndex>=0) {
//                     sourceIndex+=(int32_t)(pArgs->source-s);
//                 }
//             }

//             if(cnv->preToULength<0) {
//                 /*
//                  * switch the source to new replay units (cannot occur while replaying)
//                  * after offset handling and before end-of-input and callback handling
//                  */
//                 if(realSource==NULL) {
//                     realSource=pArgs->source;
//                     realSourceLimit=pArgs->sourceLimit;
//                     realFlush=pArgs->flush;
//                     realSourceIndex=sourceIndex;

//                     uprv_memcpy(replay, cnv->preToU, -cnv->preToULength);
//                     pArgs->source=replay;
//                     pArgs->sourceLimit=replay-cnv->preToULength;
//                     pArgs->flush=FALSE;
//                     if((sourceIndex+=cnv->preToULength)<0) {
//                         sourceIndex=-1;
//                     }

//                     cnv->preToULength=0;
//                 } else {
//                     /* see implementation note before _fromUnicodeWithCallback() */
//                     U_ASSERT(realSource==NULL);
//                     *err=U_INTERNAL_PROGRAM_ERROR;
//                 }
//             }

//             /* update pointers */
//             s=pArgs->source;
//             t=pArgs->target;

//             if(U_SUCCESS(*err)) {
//                 if(s<pArgs->sourceLimit) {
//                     /*
//                      * continue with the conversion loop while there is still input left
//                      * (continue converting by breaking out of only the inner loop)
//                      */
//                     break;
//                 } else if(realSource!=NULL) {
//                     /* switch back from replaying to the real source and continue */
//                     pArgs->source=realSource;
//                     pArgs->sourceLimit=realSourceLimit;
//                     pArgs->flush=realFlush;
//                     sourceIndex=realSourceIndex;

//                     realSource=NULL;
//                     break;
//                 } else if(pArgs->flush && cnv->toULength>0) {
//                     /*
//                      * the entire input stream is consumed
//                      * and there is a partial, truncated input sequence left
//                      */

//                     /* inject an error and continue with callback handling */
//                     *err=U_TRUNCATED_CHAR_FOUND;
//                     calledCallback=FALSE; /* new error condition */
//                 } else {
//                     /* input consumed */
//                     if(pArgs->flush) {
//                         /*
//                          * return to the conversion loop once more if the flush
//                          * flag is set and the conversion function has not
//                          * successfully processed the end of the input yet
//                          *
//                          * (continue converting by breaking out of only the inner loop)
//                          */
//                         if(!converterSawEndOfInput) {
//                             break;
//                         }

//                         /* reset the converter without calling the callback function */
//                         _reset(cnv, UCNV_RESET_TO_UNICODE, FALSE);
//                     }

//                     /* done successfully */
//                     return;
//                 }
//             }

//             /* U_FAILURE(*err) */
//             {
//                 UErrorCode e;

//                 if( calledCallback ||
//                     (e=*err)==U_BUFFER_OVERFLOW_ERROR ||
//                     (e!=U_INVALID_CHAR_FOUND &&
//                      e!=U_ILLEGAL_CHAR_FOUND &&
//                      e!=U_TRUNCATED_CHAR_FOUND &&
//                      e!=U_ILLEGAL_ESCAPE_SEQUENCE &&
//                      e!=U_UNSUPPORTED_ESCAPE_SEQUENCE)
//                 ) {
//                     /*
//                      * the callback did not or cannot resolve the error:
//                      * set output pointers and return
//                      *
//                      * the check for buffer overflow is redundant but it is
//                      * a high-runner case and hopefully documents the intent
//                      * well
//                      *
//                      * if we were replaying, then the replay buffer must be
//                      * copied back into the UConverter
//                      * and the real arguments must be restored
//                      */
//                     if(realSource!=NULL) {
//                         int32_t length;

//                         U_ASSERT(cnv->preToULength==0);

//                         length=(int32_t)(pArgs->sourceLimit-pArgs->source);
//                         if(length>0) {
//                             uprv_memcpy(cnv->preToU, pArgs->source, length);
//                             cnv->preToULength=(int8_t)-length;
//                         }

//                         pArgs->source=realSource;
//                         pArgs->sourceLimit=realSourceLimit;
//                         pArgs->flush=realFlush;
//                     }

//                     return;
//                 }
//             }

//             /* copy toUBytes[] to invalidCharBuffer[] */
//             errorInputLength=cnv->invalidCharLength=cnv->toULength;
//             if(errorInputLength>0) {
//                 uprv_memcpy(cnv->invalidCharBuffer, cnv->toUBytes, errorInputLength);
//             }

//             /* set the converter state to deal with the next character */
//             cnv->toULength=0;

//             /* call the callback function */
//             if(cnv->toUCallbackReason==UCNV_ILLEGAL && *err==U_INVALID_CHAR_FOUND) {
//                 cnv->toUCallbackReason = UCNV_UNASSIGNED;
//             }
//             cnv->fromCharErrorBehaviour(cnv->toUContext, pArgs,
//                 cnv->invalidCharBuffer, errorInputLength,
//                 cnv->toUCallbackReason,
//                 err);
//             cnv->toUCallbackReason = UCNV_ILLEGAL; /* reset to default value */

//             /*
//              * loop back to the offset handling
//              *
//              * this flag will indicate after offset handling
//              * that a callback was called;
//              * if the callback did not resolve the error, then we return
//              */
//             calledCallback=TRUE;
//         }
//     }
// }

// /*
//  * Output the toUnicode overflow buffer.
//  * Call this function if(cnv->UCharErrorBufferLength>0).
//  * @return TRUE if overflow
//  */
// static UBool
// ucnv_outputOverflowToUnicode(UConverter *cnv,
//                              UChar **target, const UChar *targetLimit,
//                              int32_t **pOffsets,
//                              UErrorCode *err) {
//     int32_t *offsets;
//     UChar *overflow, *t;
//     int32_t i, length;

//     t=*target;
//     if(pOffsets!=NULL) {
//         offsets=*pOffsets;
//     } else {
//         offsets=NULL;
//     }

//     overflow=cnv->UCharErrorBuffer;
//     length=cnv->UCharErrorBufferLength;
//     i=0;
//     while(i<length) {
//         if(t==targetLimit) {
//             /* the overflow buffer contains too much, keep the rest */
//             int32_t j=0;

//             do {
//                 overflow[j++]=overflow[i++];
//             } while(i<length);

//             cnv->UCharErrorBufferLength=(int8_t)j;
//             *target=t;
//             if(offsets!=NULL) {
//                 *pOffsets=offsets;
//             }
//             *err=U_BUFFER_OVERFLOW_ERROR;
//             return TRUE;
//         }

//         /* copy the overflow contents to the target */
//         *t++=overflow[i++];
//         if(offsets!=NULL) {
//             *offsets++=-1; /* no source index available for old output */
//         }
//     }

//     /* the overflow buffer is completely copied to the target */
//     cnv->UCharErrorBufferLength=0;
//     *target=t;
//     if(offsets!=NULL) {
//         *pOffsets=offsets;
//     }
//     return FALSE;
// }

// U_CAPI void U_EXPORT2
// ucnv_toUnicode(UConverter *cnv,
//                UChar **target, const UChar *targetLimit,
//                const char **source, const char *sourceLimit,
//                int32_t *offsets,
//                UBool flush,
//                UErrorCode *err) {
//     UConverterToUnicodeArgs args;
//     const char *s;
//     UChar *t;

//     /* check parameters */
//     if(err==NULL || U_FAILURE(*err)) {
//         return;
//     }

//     if(cnv==NULL || target==NULL || source==NULL) {
//         *err=U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

//     s=*source;
//     t=*target;

//     if ((const void *)U_MAX_PTR(targetLimit) == (const void *)targetLimit) {
//         /*
//         Prevent code from going into an infinite loop in case we do hit this
//         limit. The limit pointer is expected to be on a UChar * boundary.
//         This also prevents the next argument check from failing.
//         */
//         targetLimit = (const UChar *)(((const char *)targetLimit) - 1);
//     }

//     /*
//      * All these conditions should never happen.
//      *
//      * 1) Make sure that the limits are >= to the address source or target
//      *
//      * 2) Make sure that the buffer sizes do not exceed the number range for
//      * int32_t because some functions use the size (in units or bytes)
//      * rather than comparing pointers, and because offsets are int32_t values.
//      *
//      * size_t is guaranteed to be unsigned and large enough for the job.
//      *
//      * Return with an error instead of adjusting the limits because we would
//      * not be able to maintain the semantics that either the source must be
//      * consumed or the target filled (unless an error occurs).
//      * An adjustment would be sourceLimit=t+0x7fffffff; for example.
//      *
//      * 3) Make sure that the user didn't incorrectly cast a UChar * pointer
//      * to a char * pointer and provide an incomplete UChar code unit.
//      */
//     if (sourceLimit<s || targetLimit<t ||
//         ((size_t)(sourceLimit-s)>(size_t)0x7fffffff && sourceLimit>s) ||
//         ((size_t)(targetLimit-t)>(size_t)0x3fffffff && targetLimit>t) ||
//         (((const char *)targetLimit-(const char *)t) & 1) != 0
//     ) {
//         *err=U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

//     /* output the target overflow buffer */
//     if( cnv->UCharErrorBufferLength>0 &&
//         ucnv_outputOverflowToUnicode(cnv, target, targetLimit, &offsets, err)
//     ) {
//         /* U_BUFFER_OVERFLOW_ERROR */
//         return;
//     }
//     /* *target may have moved, therefore stop using t */

//     if(!flush && s==sourceLimit && cnv->preToULength>=0) {
//         /* the overflow buffer is emptied and there is no new input: we are done */
//         return;
//     }

//     /*
//      * Do not simply return with a buffer overflow error if
//      * !flush && t==targetLimit
//      * because it is possible that the source will not generate any output.
//      * For example, the skip callback may be called;
//      * it does not output anything.
//      */

//     /* prepare the converter arguments */
//     args.converter=cnv;
//     args.flush=flush;
//     args.offsets=offsets;
//     args.source=s;
//     args.sourceLimit=sourceLimit;
//     args.target=*target;
//     args.targetLimit=targetLimit;
//     args.size=sizeof(args);

//     _toUnicodeWithCallback(&args, err);

//     *source=args.source;
//     *target=args.target;
// }

// /* ucnv_to/fromUChars() ----------------------------------------------------- */

// U_CAPI int32_t U_EXPORT2
// ucnv_fromUChars(UConverter *cnv,
//                 char *dest, int32_t destCapacity,
//                 const UChar *src, int32_t srcLength,
//                 UErrorCode *pErrorCode) {
//     const UChar *srcLimit;
//     char *originalDest, *destLimit;
//     int32_t destLength;

//     /* check arguments */
//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     if( cnv==NULL ||
//         destCapacity<0 || (destCapacity>0 && dest==NULL) ||
//         srcLength<-1 || (srcLength!=0 && src==NULL)
//     ) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     /* initialize */
//     ucnv_resetFromUnicode(cnv);
//     originalDest=dest;
//     if(srcLength==-1) {
//         srcLength=u_strlen(src);
//     }
//     if(srcLength>0) {
//         srcLimit=src+srcLength;
//         destCapacity=pinCapacity(dest, destCapacity);
//         destLimit=dest+destCapacity;

//         /* perform the conversion */
//         ucnv_fromUnicode(cnv, &dest, destLimit, &src, srcLimit, 0, TRUE, pErrorCode);
//         destLength=(int32_t)(dest-originalDest);

//         /* if an overflow occurs, then get the preflighting length */
//         if(*pErrorCode==U_BUFFER_OVERFLOW_ERROR) {
//             char buffer[1024];

//             destLimit=buffer+sizeof(buffer);
//             do {
//                 dest=buffer;
//                 *pErrorCode=U_ZERO_ERROR;
//                 ucnv_fromUnicode(cnv, &dest, destLimit, &src, srcLimit, 0, TRUE, pErrorCode);
//                 destLength+=(int32_t)(dest-buffer);
//             } while(*pErrorCode==U_BUFFER_OVERFLOW_ERROR);
//         }
//     } else {
//         destLength=0;
//     }

//     return u_terminateChars(originalDest, destCapacity, destLength, pErrorCode);
// }

// U_CAPI int32_t U_EXPORT2
// ucnv_toUChars(UConverter *cnv,
//               UChar *dest, int32_t destCapacity,
//               const char *src, int32_t srcLength,
//               UErrorCode *pErrorCode) {
//     const char *srcLimit;
//     UChar *originalDest, *destLimit;
//     int32_t destLength;

//     /* check arguments */
//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     if( cnv==NULL ||
//         destCapacity<0 || (destCapacity>0 && dest==NULL) ||
//         srcLength<-1 || (srcLength!=0 && src==NULL))
//     {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     /* initialize */
//     ucnv_resetToUnicode(cnv);
//     originalDest=dest;
//     if(srcLength==-1) {
//         srcLength=(int32_t)uprv_strlen(src);
//     }
//     if(srcLength>0) {
//         srcLimit=src+srcLength;
//         destCapacity=pinCapacity(dest, destCapacity);
//         destLimit=dest+destCapacity;

//         /* perform the conversion */
//         ucnv_toUnicode(cnv, &dest, destLimit, &src, srcLimit, 0, TRUE, pErrorCode);
//         destLength=(int32_t)(dest-originalDest);

//         /* if an overflow occurs, then get the preflighting length */
//         if(*pErrorCode==U_BUFFER_OVERFLOW_ERROR)
//         {
//             UChar buffer[1024];

//             destLimit=buffer+UPRV_LENGTHOF(buffer);
//             do {
//                 dest=buffer;
//                 *pErrorCode=U_ZERO_ERROR;
//                 ucnv_toUnicode(cnv, &dest, destLimit, &src, srcLimit, 0, TRUE, pErrorCode);
//                 destLength+=(int32_t)(dest-buffer);
//             }
//             while(*pErrorCode==U_BUFFER_OVERFLOW_ERROR);
//         }
//     } else {
//         destLength=0;
//     }

//     return u_terminateUChars(originalDest, destCapacity, destLength, pErrorCode);
// }

// /* ucnv_getNextUChar() ------------------------------------------------------ */

// U_CAPI UChar32 U_EXPORT2
// ucnv_getNextUChar(UConverter *cnv,
//                   const char **source, const char *sourceLimit,
//                   UErrorCode *err) {
//     UConverterToUnicodeArgs args;
//     UChar buffer[U16_MAX_LENGTH];
//     const char *s;
//     UChar32 c;
//     int32_t i, length;

//     /* check parameters */
//     if(err==NULL || U_FAILURE(*err)) {
//         return 0xffff;
//     }

//     if(cnv==NULL || source==NULL) {
//         *err=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0xffff;
//     }

//     s=*source;
//     if(sourceLimit<s) {
//         *err=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0xffff;
//     }

//     /*
//      * Make sure that the buffer sizes do not exceed the number range for
//      * int32_t because some functions use the size (in units or bytes)
//      * rather than comparing pointers, and because offsets are int32_t values.
//      *
//      * size_t is guaranteed to be unsigned and large enough for the job.
//      *
//      * Return with an error instead of adjusting the limits because we would
//      * not be able to maintain the semantics that either the source must be
//      * consumed or the target filled (unless an error occurs).
//      * An adjustment would be sourceLimit=t+0x7fffffff; for example.
//      */
//     if(((size_t)(sourceLimit-s)>(size_t)0x7fffffff && sourceLimit>s)) {
//         *err=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0xffff;
//     }

//     c=U_SENTINEL;

//     /* flush the target overflow buffer */
//     if(cnv->UCharErrorBufferLength>0) {
//         UChar *overflow;

//         overflow=cnv->UCharErrorBuffer;
//         i=0;
//         length=cnv->UCharErrorBufferLength;
//         U16_NEXT(overflow, i, length, c);

//         /* move the remaining overflow contents up to the beginning */
//         if((cnv->UCharErrorBufferLength=(int8_t)(length-i))>0) {
//             uprv_memmove(cnv->UCharErrorBuffer, cnv->UCharErrorBuffer+i,
//                          cnv->UCharErrorBufferLength*U_SIZEOF_UCHAR);
//         }

//         if(!U16_IS_LEAD(c) || i<length) {
//             return c;
//         }
//         /*
//          * Continue if the overflow buffer contained only a lead surrogate,
//          * in case the converter outputs single surrogates from complete
//          * input sequences.
//          */
//     }

//     /*
//      * flush==TRUE is implied for ucnv_getNextUChar()
//      *
//      * do not simply return even if s==sourceLimit because the converter may
//      * not have seen flush==TRUE before
//      */

//     /* prepare the converter arguments */
//     args.converter=cnv;
//     args.flush=TRUE;
//     args.offsets=NULL;
//     args.source=s;
//     args.sourceLimit=sourceLimit;
//     args.target=buffer;
//     args.targetLimit=buffer+1;
//     args.size=sizeof(args);

//     if(c<0) {
//         /*
//          * call the native getNextUChar() implementation if we are
//          * at a character boundary (toULength==0)
//          *
//          * unlike with _toUnicode(), getNextUChar() implementations must set
//          * U_TRUNCATED_CHAR_FOUND for truncated input,
//          * in addition to setting toULength/toUBytes[]
//          */
//         if(cnv->toULength==0 && cnv->sharedData->impl->getNextUChar!=NULL) {
//             c=cnv->sharedData->impl->getNextUChar(&args, err);
//             *source=s=args.source;
//             if(*err==U_INDEX_OUTOFBOUNDS_ERROR) {
//                 /* reset the converter without calling the callback function */
//                 _reset(cnv, UCNV_RESET_TO_UNICODE, FALSE);
//                 return 0xffff; /* no output */
//             } else if(U_SUCCESS(*err) && c>=0) {
//                 return c;
//             /*
//              * else fall through to use _toUnicode() because
//              *   UCNV_GET_NEXT_UCHAR_USE_TO_U: the native function did not want to handle it after all
//              *   U_FAILURE: call _toUnicode() for callback handling (do not output c)
//              */
//             }
//         }

//         /* convert to one UChar in buffer[0], or handle getNextUChar() errors */
//         _toUnicodeWithCallback(&args, err);

//         if(*err==U_BUFFER_OVERFLOW_ERROR) {
//             *err=U_ZERO_ERROR;
//         }

//         i=0;
//         length=(int32_t)(args.target-buffer);
//     } else {
//         /* write the lead surrogate from the overflow buffer */
//         buffer[0]=(UChar)c;
//         args.target=buffer+1;
//         i=0;
//         length=1;
//     }

//     /* buffer contents starts at i and ends before length */

//     if(U_FAILURE(*err)) {
//         c=0xffff; /* no output */
//     } else if(length==0) {
//         /* no input or only state changes */
//         *err=U_INDEX_OUTOFBOUNDS_ERROR;
//         /* no need to reset explicitly because _toUnicodeWithCallback() did it */
//         c=0xffff; /* no output */
//     } else {
//         c=buffer[0];
//         i=1;
//         if(!U16_IS_LEAD(c)) {
//             /* consume c=buffer[0], done */
//         } else {
//             /* got a lead surrogate, see if a trail surrogate follows */
//             UChar c2;

//             if(cnv->UCharErrorBufferLength>0) {
//                 /* got overflow output from the conversion */
//                 if(U16_IS_TRAIL(c2=cnv->UCharErrorBuffer[0])) {
//                     /* got a trail surrogate, too */
//                     c=U16_GET_SUPPLEMENTARY(c, c2);

//                     /* move the remaining overflow contents up to the beginning */
//                     if((--cnv->UCharErrorBufferLength)>0) {
//                         uprv_memmove(cnv->UCharErrorBuffer, cnv->UCharErrorBuffer+1,
//                                      cnv->UCharErrorBufferLength*U_SIZEOF_UCHAR);
//                     }
//                 } else {
//                     /* c is an unpaired lead surrogate, just return it */
//                 }
//             } else if(args.source<sourceLimit) {
//                 /* convert once more, to buffer[1] */
//                 args.targetLimit=buffer+2;
//                 _toUnicodeWithCallback(&args, err);
//                 if(*err==U_BUFFER_OVERFLOW_ERROR) {
//                     *err=U_ZERO_ERROR;
//                 }

//                 length=(int32_t)(args.target-buffer);
//                 if(U_SUCCESS(*err) && length==2 && U16_IS_TRAIL(c2=buffer[1])) {
//                     /* got a trail surrogate, too */
//                     c=U16_GET_SUPPLEMENTARY(c, c2);
//                     i=2;
//                 }
//             }
//         }
//     }

//     /*
//      * move leftover output from buffer[i..length[
//      * into the beginning of the overflow buffer
//      */
//     if(i<length) {
//         /* move further overflow back */
//         int32_t delta=length-i;
//         if((length=cnv->UCharErrorBufferLength)>0) {
//             uprv_memmove(cnv->UCharErrorBuffer+delta, cnv->UCharErrorBuffer,
//                          length*U_SIZEOF_UCHAR);
//         }
//         cnv->UCharErrorBufferLength=(int8_t)(length+delta);

//         cnv->UCharErrorBuffer[0]=buffer[i++];
//         if(delta>1) {
//             cnv->UCharErrorBuffer[1]=buffer[i];
//         }
//     }

//     *source=args.source;
//     return c;
// }

// /* ucnv_convert() and siblings ---------------------------------------------- */

// U_CAPI void U_EXPORT2
// ucnv_convertEx(UConverter *targetCnv, UConverter *sourceCnv,
//                char **target, const char *targetLimit,
//                const char **source, const char *sourceLimit,
//                UChar *pivotStart, UChar **pivotSource,
//                UChar **pivotTarget, const UChar *pivotLimit,
//                UBool reset, UBool flush,
//                UErrorCode *pErrorCode) {
//     UChar pivotBuffer[CHUNK_SIZE];
//     const UChar *myPivotSource;
//     UChar *myPivotTarget;
//     const char *s;
//     char *t;

//     UConverterToUnicodeArgs toUArgs;
//     UConverterFromUnicodeArgs fromUArgs;
//     UConverterConvert convert;

//     /* error checking */
//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return;
//     }

//     if( targetCnv==NULL || sourceCnv==NULL ||
//         source==NULL || *source==NULL ||
//         target==NULL || *target==NULL || targetLimit==NULL
//     ) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

//     s=*source;
//     t=*target;
//     if((sourceLimit!=NULL && sourceLimit<s) || targetLimit<t) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

//     /*
//      * Make sure that the buffer sizes do not exceed the number range for
//      * int32_t. See ucnv_toUnicode() for a more detailed comment.
//      */
//     if(
//         (sourceLimit!=NULL && ((size_t)(sourceLimit-s)>(size_t)0x7fffffff && sourceLimit>s)) ||
//         ((size_t)(targetLimit-t)>(size_t)0x7fffffff && targetLimit>t)
//     ) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

//     if(pivotStart==NULL) {
//         if(!flush) {
//             /* streaming conversion requires an explicit pivot buffer */
//             *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//             return;
//         }

//         /* use the stack pivot buffer */
//         myPivotSource=myPivotTarget=pivotStart=pivotBuffer;
//         pivotSource=(UChar **)&myPivotSource;
//         pivotTarget=&myPivotTarget;
//         pivotLimit=pivotBuffer+CHUNK_SIZE;
//     } else if(  pivotStart>=pivotLimit ||
//                 pivotSource==NULL || *pivotSource==NULL ||
//                 pivotTarget==NULL || *pivotTarget==NULL ||
//                 pivotLimit==NULL
//     ) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

//     if(sourceLimit==NULL) {
//         /* get limit of single-byte-NUL-terminated source string */
//         sourceLimit=uprv_strchr(*source, 0);
//     }

//     if(reset) {
//         ucnv_resetToUnicode(sourceCnv);
//         ucnv_resetFromUnicode(targetCnv);
//         *pivotSource=*pivotTarget=pivotStart;
//     } else if(targetCnv->charErrorBufferLength>0) {
//         /* output the targetCnv overflow buffer */
//         if(ucnv_outputOverflowFromUnicode(targetCnv, target, targetLimit, NULL, pErrorCode)) {
//             /* U_BUFFER_OVERFLOW_ERROR */
//             return;
//         }
//         /* *target has moved, therefore stop using t */

//         if( !flush &&
//             targetCnv->preFromULength>=0 && *pivotSource==*pivotTarget &&
//             sourceCnv->UCharErrorBufferLength==0 && sourceCnv->preToULength>=0 && s==sourceLimit
//         ) {
//             /* the fromUnicode overflow buffer is emptied and there is no new input: we are done */
//             return;
//         }
//     }

//     /* Is direct-UTF-8 conversion available? */
//     if( sourceCnv->sharedData->staticData->conversionType==UCNV_UTF8 &&
//         targetCnv->sharedData->impl->fromUTF8!=NULL
//     ) {
//         convert=targetCnv->sharedData->impl->fromUTF8;
//     } else if( targetCnv->sharedData->staticData->conversionType==UCNV_UTF8 &&
//                sourceCnv->sharedData->impl->toUTF8!=NULL
//     ) {
//         convert=sourceCnv->sharedData->impl->toUTF8;
//     } else {
//         convert=NULL;
//     }

//     /*
//      * If direct-UTF-8 conversion is available, then we use a smaller
//      * pivot buffer for error handling and partial matches
//      * so that we quickly return to direct conversion.
//      *
//      * 32 is large enough for UCNV_EXT_MAX_UCHARS and UCNV_ERROR_BUFFER_LENGTH.
//      *
//      * We could reduce the pivot buffer size further, at the cost of
//      * buffer overflows from callbacks.
//      * The pivot buffer should not be smaller than the maximum number of
//      * fromUnicode extension table input UChars
//      * (for m:n conversion, see
//      * targetCnv->sharedData->mbcs.extIndexes[UCNV_EXT_COUNT_UCHARS])
//      * or 2 for surrogate pairs.
//      *
//      * Too small a buffer can cause thrashing between pivoting and direct
//      * conversion, with function call overhead outweighing the benefits
//      * of direct conversion.
//      */
//     if(convert!=NULL && (pivotLimit-pivotStart)>32) {
//         pivotLimit=pivotStart+32;
//     }

//     /* prepare the converter arguments */
//     fromUArgs.converter=targetCnv;
//     fromUArgs.flush=FALSE;
//     fromUArgs.offsets=NULL;
//     fromUArgs.target=*target;
//     fromUArgs.targetLimit=targetLimit;
//     fromUArgs.size=sizeof(fromUArgs);

//     toUArgs.converter=sourceCnv;
//     toUArgs.flush=flush;
//     toUArgs.offsets=NULL;
//     toUArgs.source=s;
//     toUArgs.sourceLimit=sourceLimit;
//     toUArgs.targetLimit=pivotLimit;
//     toUArgs.size=sizeof(toUArgs);

//     /*
//      * TODO: Consider separating this function into two functions,
//      * extracting exactly the conversion loop,
//      * for readability and to reduce the set of visible variables.
//      *
//      * Otherwise stop using s and t from here on.
//      */
//     s=t=NULL;

//     /*
//      * conversion loop
//      *
//      * The sequence of steps in the loop may appear backward,
//      * but the principle is simple:
//      * In the chain of
//      *   source - sourceCnv overflow - pivot - targetCnv overflow - target
//      * empty out later buffers before refilling them from earlier ones.
//      *
//      * The targetCnv overflow buffer is flushed out only once before the loop.
//      */
//     for(;;) {
//         /*
//          * if(pivot not empty or error or replay or flush fromUnicode) {
//          *   fromUnicode(pivot -> target);
//          * }
//          *
//          * For pivoting conversion; and for direct conversion for
//          * error callback handling and flushing the replay buffer.
//          */
//         if( *pivotSource<*pivotTarget ||
//             U_FAILURE(*pErrorCode) ||
//             targetCnv->preFromULength<0 ||
//             fromUArgs.flush
//         ) {
//             fromUArgs.source=*pivotSource;
//             fromUArgs.sourceLimit=*pivotTarget;
//             _fromUnicodeWithCallback(&fromUArgs, pErrorCode);
//             if(U_FAILURE(*pErrorCode)) {
//                 /* target overflow, or conversion error */
//                 *pivotSource=(UChar *)fromUArgs.source;
//                 break;
//             }

//             /*
//              * _fromUnicodeWithCallback() must have consumed the pivot contents
//              * (*pivotSource==*pivotTarget) since it returned with U_SUCCESS()
//              */
//         }

//         /* The pivot buffer is empty; reset it so we start at pivotStart. */
//         *pivotSource=*pivotTarget=pivotStart;

//         /*
//          * if(sourceCnv overflow buffer not empty) {
//          *     move(sourceCnv overflow buffer -> pivot);
//          *     continue;
//          * }
//          */
//         /* output the sourceCnv overflow buffer */
//         if(sourceCnv->UCharErrorBufferLength>0) {
//             if(ucnv_outputOverflowToUnicode(sourceCnv, pivotTarget, pivotLimit, NULL, pErrorCode)) {
//                 /* U_BUFFER_OVERFLOW_ERROR */
//                 *pErrorCode=U_ZERO_ERROR;
//             }
//             continue;
//         }

//         /*
//          * check for end of input and break if done
//          *
//          * Checking both flush and fromUArgs.flush ensures that the converters
//          * have been called with the flush flag set if the ucnv_convertEx()
//          * caller set it.
//          */
//         if( toUArgs.source==sourceLimit &&
//             sourceCnv->preToULength>=0 && sourceCnv->toULength==0 &&
//             (!flush || fromUArgs.flush)
//         ) {
//             /* done successfully */
//             break;
//         }

//         /*
//          * use direct conversion if available
//          * but not if continuing a partial match
//          * or flushing the toUnicode replay buffer
//          */
//         if(convert!=NULL && targetCnv->preFromUFirstCP<0 && sourceCnv->preToULength==0) {
//             if(*pErrorCode==U_USING_DEFAULT_WARNING) {
//                 /* remove a warning that may be set by this function */
//                 *pErrorCode=U_ZERO_ERROR;
//             }
//             convert(&fromUArgs, &toUArgs, pErrorCode);
//             if(*pErrorCode==U_BUFFER_OVERFLOW_ERROR) {
//                 break;
//             } else if(U_FAILURE(*pErrorCode)) {
//                 if(sourceCnv->toULength>0) {
//                     /*
//                      * Fall through to calling _toUnicodeWithCallback()
//                      * for callback handling.
//                      *
//                      * The pivot buffer will be reset with
//                      *   *pivotSource=*pivotTarget=pivotStart;
//                      * which indicates a toUnicode error to the caller
//                      * (*pivotSource==pivotStart shows no pivot UChars consumed).
//                      */
//                 } else {
//                     /*
//                      * Indicate a fromUnicode error to the caller
//                      * (*pivotSource>pivotStart shows some pivot UChars consumed).
//                      */
//                     *pivotSource=*pivotTarget=pivotStart+1;
//                     /*
//                      * Loop around to calling _fromUnicodeWithCallbacks()
//                      * for callback handling.
//                      */
//                     continue;
//                 }
//             } else if(*pErrorCode==U_USING_DEFAULT_WARNING) {
//                 /*
//                  * No error, but the implementation requested to temporarily
//                  * fall back to pivoting.
//                  */
//                 *pErrorCode=U_ZERO_ERROR;
//             /*
//              * The following else branches are almost identical to the end-of-input
//              * handling in _toUnicodeWithCallback().
//              * Avoid calling it just for the end of input.
//              */
//             } else if(flush && sourceCnv->toULength>0) { /* flush==toUArgs.flush */
//                 /*
//                  * the entire input stream is consumed
//                  * and there is a partial, truncated input sequence left
//                  */

//                 /* inject an error and continue with callback handling */
//                 *pErrorCode=U_TRUNCATED_CHAR_FOUND;
//             } else {
//                 /* input consumed */
//                 if(flush) {
//                     /* reset the converters without calling the callback functions */
//                     _reset(sourceCnv, UCNV_RESET_TO_UNICODE, FALSE);
//                     _reset(targetCnv, UCNV_RESET_FROM_UNICODE, FALSE);
//                 }

//                 /* done successfully */
//                 break;
//             }
//         }

//         /*
//          * toUnicode(source -> pivot);
//          *
//          * For pivoting conversion; and for direct conversion for
//          * error callback handling, continuing partial matches
//          * and flushing the replay buffer.
//          *
//          * The pivot buffer is empty and reset.
//          */
//         toUArgs.target=pivotStart; /* ==*pivotTarget */
//         /* toUArgs.targetLimit=pivotLimit; already set before the loop */
//         _toUnicodeWithCallback(&toUArgs, pErrorCode);
//         *pivotTarget=toUArgs.target;
//         if(*pErrorCode==U_BUFFER_OVERFLOW_ERROR) {
//             /* pivot overflow: continue with the conversion loop */
//             *pErrorCode=U_ZERO_ERROR;
//         } else if(U_FAILURE(*pErrorCode) || (!flush && *pivotTarget==pivotStart)) {
//             /* conversion error, or there was nothing left to convert */
//             break;
//         }
//         /*
//          * else:
//          * _toUnicodeWithCallback() wrote into the pivot buffer,
//          * continue with fromUnicode conversion.
//          *
//          * Set the fromUnicode flush flag if we flush and if toUnicode has
//          * processed the end of the input.
//          */
//         if( flush && toUArgs.source==sourceLimit &&
//             sourceCnv->preToULength>=0 &&
//             sourceCnv->UCharErrorBufferLength==0
//         ) {
//             fromUArgs.flush=TRUE;
//         }
//     }

//     /*
//      * The conversion loop is exited when one of the following is true:
//      * - the entire source text has been converted successfully to the target buffer
//      * - a target buffer overflow occurred
//      * - a conversion error occurred
//      */

//     *source=toUArgs.source;
//     *target=fromUArgs.target;

//     /* terminate the target buffer if possible */
//     if(flush && U_SUCCESS(*pErrorCode)) {
//         if(*target!=targetLimit) {
//             **target=0;
//             if(*pErrorCode==U_STRING_NOT_TERMINATED_WARNING) {
//                 *pErrorCode=U_ZERO_ERROR;
//             }
//         } else {
//             *pErrorCode=U_STRING_NOT_TERMINATED_WARNING;
//         }
//     }
// }

// /* internal implementation of ucnv_convert() etc. with preflighting */
// static int32_t
// ucnv_internalConvert(UConverter *outConverter, UConverter *inConverter,
//                      char *target, int32_t targetCapacity,
//                      const char *source, int32_t sourceLength,
//                      UErrorCode *pErrorCode) {
//     UChar pivotBuffer[CHUNK_SIZE];
//     UChar *pivot, *pivot2;

//     char *myTarget;
//     const char *sourceLimit;
//     const char *targetLimit;
//     int32_t targetLength=0;

//     /* set up */
//     if(sourceLength<0) {
//         sourceLimit=uprv_strchr(source, 0);
//     } else {
//         sourceLimit=source+sourceLength;
//     }

//     /* if there is no input data, we're done */
//     if(source==sourceLimit) {
//         return u_terminateChars(target, targetCapacity, 0, pErrorCode);
//     }

//     pivot=pivot2=pivotBuffer;
//     myTarget=target;
//     targetLength=0;

//     if(targetCapacity>0) {
//         /* perform real conversion */
//         targetLimit=target+targetCapacity;
//         ucnv_convertEx(outConverter, inConverter,
//                        &myTarget, targetLimit,
//                        &source, sourceLimit,
//                        pivotBuffer, &pivot, &pivot2, pivotBuffer+CHUNK_SIZE,
//                        FALSE,
//                        TRUE,
//                        pErrorCode);
//         targetLength=(int32_t)(myTarget-target);
//     }

//     /*
//      * If the output buffer is exhausted (or we are only "preflighting"), we need to stop writing
//      * to it but continue the conversion in order to store in targetCapacity
//      * the number of bytes that was required.
//      */
//     if(*pErrorCode==U_BUFFER_OVERFLOW_ERROR || targetCapacity==0)
//     {
//         char targetBuffer[CHUNK_SIZE];

//         targetLimit=targetBuffer+CHUNK_SIZE;
//         do {
//             *pErrorCode=U_ZERO_ERROR;
//             myTarget=targetBuffer;
//             ucnv_convertEx(outConverter, inConverter,
//                            &myTarget, targetLimit,
//                            &source, sourceLimit,
//                            pivotBuffer, &pivot, &pivot2, pivotBuffer+CHUNK_SIZE,
//                            FALSE,
//                            TRUE,
//                            pErrorCode);
//             targetLength+=(int32_t)(myTarget-targetBuffer);
//         } while(*pErrorCode==U_BUFFER_OVERFLOW_ERROR);

//         /* done with preflighting, set warnings and errors as appropriate */
//         return u_terminateChars(target, targetCapacity, targetLength, pErrorCode);
//     }

//     /* no need to call u_terminateChars() because ucnv_convertEx() took care of that */
//     return targetLength;
// }

// U_CAPI int32_t U_EXPORT2
// ucnv_convert(const char *toConverterName, const char *fromConverterName,
//              char *target, int32_t targetCapacity,
//              const char *source, int32_t sourceLength,
//              UErrorCode *pErrorCode) {
//     UConverter in, out; /* stack-allocated */
//     UConverter *inConverter, *outConverter;
//     int32_t targetLength;

//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     if( source==NULL || sourceLength<-1 ||
//         targetCapacity<0 || (targetCapacity>0 && target==NULL)
//     ) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     /* if there is no input data, we're done */
//     if(sourceLength==0 || (sourceLength<0 && *source==0)) {
//         return u_terminateChars(target, targetCapacity, 0, pErrorCode);
//     }

//     /* create the converters */
//     inConverter=ucnv_createConverter(&in, fromConverterName, pErrorCode);
//     if(U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     outConverter=ucnv_createConverter(&out, toConverterName, pErrorCode);
//     if(U_FAILURE(*pErrorCode)) {
//         ucnv_close(inConverter);
//         return 0;
//     }

//     targetLength=ucnv_internalConvert(outConverter, inConverter,
//                                       target, targetCapacity,
//                                       source, sourceLength,
//                                       pErrorCode);

//     ucnv_close(inConverter);
//     ucnv_close(outConverter);

//     return targetLength;
// }

// /* @internal */
// static int32_t
// ucnv_convertAlgorithmic(UBool convertToAlgorithmic,
//                         UConverterType algorithmicType,
//                         UConverter *cnv,
//                         char *target, int32_t targetCapacity,
//                         const char *source, int32_t sourceLength,
//                         UErrorCode *pErrorCode) {
//     UConverter algoConverterStatic; /* stack-allocated */
//     UConverter *algoConverter, *to, *from;
//     int32_t targetLength;

//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     if( cnv==NULL || source==NULL || sourceLength<-1 ||
//         targetCapacity<0 || (targetCapacity>0 && target==NULL)
//     ) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     /* if there is no input data, we're done */
//     if(sourceLength==0 || (sourceLength<0 && *source==0)) {
//         return u_terminateChars(target, targetCapacity, 0, pErrorCode);
//     }

//     /* create the algorithmic converter */
//     algoConverter=ucnv_createAlgorithmicConverter(&algoConverterStatic, algorithmicType,
//                                                   "", 0, pErrorCode);
//     if(U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     /* reset the other converter */
//     if(convertToAlgorithmic) {
//         /* cnv->Unicode->algo */
//         ucnv_resetToUnicode(cnv);
//         to=algoConverter;
//         from=cnv;
//     } else {
//         /* algo->Unicode->cnv */
//         ucnv_resetFromUnicode(cnv);
//         from=algoConverter;
//         to=cnv;
//     }

//     targetLength=ucnv_internalConvert(to, from,
//                                       target, targetCapacity,
//                                       source, sourceLength,
//                                       pErrorCode);

//     ucnv_close(algoConverter);

//     return targetLength;
// }

// U_CAPI int32_t U_EXPORT2
// ucnv_toAlgorithmic(UConverterType algorithmicType,
//                    UConverter *cnv,
//                    char *target, int32_t targetCapacity,
//                    const char *source, int32_t sourceLength,
//                    UErrorCode *pErrorCode) {
//     return ucnv_convertAlgorithmic(TRUE, algorithmicType, cnv,
//                                    target, targetCapacity,
//                                    source, sourceLength,
//                                    pErrorCode);
// }

// U_CAPI int32_t U_EXPORT2
// ucnv_fromAlgorithmic(UConverter *cnv,
//                      UConverterType algorithmicType,
//                      char *target, int32_t targetCapacity,
//                      const char *source, int32_t sourceLength,
//                      UErrorCode *pErrorCode) {
//     return ucnv_convertAlgorithmic(FALSE, algorithmicType, cnv,
//                                    target, targetCapacity,
//                                    source, sourceLength,
//                                    pErrorCode);
// }

// U_CAPI UConverterType  U_EXPORT2
// ucnv_getType(const UConverter* converter)
// {
//     int8_t type = converter->sharedData->staticData->conversionType;
// #if !UCONFIG_NO_LEGACY_CONVERSION
//     if(type == UCNV_MBCS) {
//         return ucnv_MBCSGetType(converter);
//     }
// #endif
//     return (UConverterType)type;
// }

// U_CAPI void  U_EXPORT2
// ucnv_getStarters(const UConverter* converter,
//                  UBool starters[256],
//                  UErrorCode* err)
// {
//     if (err == NULL || U_FAILURE(*err)) {
//         return;
//     }

//     if(converter->sharedData->impl->getStarters != NULL) {
//         converter->sharedData->impl->getStarters(converter, starters, err);
//     } else {
//         *err = U_ILLEGAL_ARGUMENT_ERROR;
//     }
// }

// static const UAmbiguousConverter *ucnv_getAmbiguous(const UConverter *cnv)
// {
//     UErrorCode errorCode;
//     const char *name;
//     int32_t i;

//     if(cnv==NULL) {
//         return NULL;
//     }

//     errorCode=U_ZERO_ERROR;
//     name=ucnv_getName(cnv, &errorCode);
//     if(U_FAILURE(errorCode)) {
//         return NULL;
//     }

//     for(i=0; i<UPRV_LENGTHOF(ambiguousConverters); ++i)
//     {
//         if(0==uprv_strcmp(name, ambiguousConverters[i].name))
//         {
//             return ambiguousConverters+i;
//         }
//     }

//     return NULL;
// }

// U_CAPI void  U_EXPORT2
// ucnv_fixFileSeparator(const UConverter *cnv,
//                       UChar* source,
//                       int32_t sourceLength) {
//     const UAmbiguousConverter *a;
//     int32_t i;
//     UChar variant5c;

//     if(cnv==NULL || source==NULL || sourceLength<=0 || (a=ucnv_getAmbiguous(cnv))==NULL)
//     {
//         return;
//     }

//     variant5c=a->variant5c;
//     for(i=0; i<sourceLength; ++i) {
//         if(source[i]==variant5c) {
//             source[i]=0x5c;
//         }
//     }
// }

// U_CAPI UBool  U_EXPORT2
// ucnv_isAmbiguous(const UConverter *cnv) {
//     return (UBool)(ucnv_getAmbiguous(cnv)!=NULL);
// }

// U_CAPI void  U_EXPORT2
// ucnv_setFallback(UConverter *cnv, UBool usesFallback)
// {
//     cnv->useFallback = usesFallback;
// }

// U_CAPI UBool  U_EXPORT2
// ucnv_usesFallback(const UConverter *cnv)
// {
//     return cnv->useFallback;
// }

// U_CAPI void  U_EXPORT2
// ucnv_getInvalidChars (const UConverter * converter,
//                       char *errBytes,
//                       int8_t * len,
//                       UErrorCode * err)
// {
//     if (err == NULL || U_FAILURE(*err))
//     {
//         return;
//     }
//     if (len == NULL || errBytes == NULL || converter == NULL)
//     {
//         *err = U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }
//     if (*len < converter->invalidCharLength)
//     {
//         *err = U_INDEX_OUTOFBOUNDS_ERROR;
//         return;
//     }
//     if ((*len = converter->invalidCharLength) > 0)
//     {
//         uprv_memcpy (errBytes, converter->invalidCharBuffer, *len);
//     }
// }

// U_CAPI void  U_EXPORT2
// ucnv_getInvalidUChars (const UConverter * converter,
//                        UChar *errChars,
//                        int8_t * len,
//                        UErrorCode * err)
// {
//     if (err == NULL || U_FAILURE(*err))
//     {
//         return;
//     }
//     if (len == NULL || errChars == NULL || converter == NULL)
//     {
//         *err = U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }
//     if (*len < converter->invalidUCharLength)
//     {
//         *err = U_INDEX_OUTOFBOUNDS_ERROR;
//         return;
//     }
//     if ((*len = converter->invalidUCharLength) > 0)
//     {
//         u_memcpy (errChars, converter->invalidUCharBuffer, *len);
//     }
// }

// #define SIG_MAX_LEN 5

// U_CAPI const char* U_EXPORT2
// ucnv_detectUnicodeSignature( const char* source,
//                              int32_t sourceLength,
//                              int32_t* signatureLength,
//                              UErrorCode* pErrorCode) {
//     int32_t dummy;

//     /* initial 0xa5 bytes: make sure that if we read <SIG_MAX_LEN
//      * bytes we don't misdetect something
//      */
//     char start[SIG_MAX_LEN]={ '\xa5', '\xa5', '\xa5', '\xa5', '\xa5' };
//     int i = 0;

//     if((pErrorCode==NULL) || U_FAILURE(*pErrorCode)){
//         return NULL;
//     }

//     if(source == NULL || sourceLength < -1){
//         *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
//         return NULL;
//     }

//     if(signatureLength == NULL) {
//         signatureLength = &dummy;
//     }

//     if(sourceLength==-1){
//         sourceLength=(int32_t)uprv_strlen(source);
//     }


//     while(i<sourceLength&& i<SIG_MAX_LEN){
//         start[i]=source[i];
//         i++;
//     }

//     if(start[0] == '\xFE' && start[1] == '\xFF') {
//         *signatureLength=2;
//         return  "UTF-16BE";
//     } else if(start[0] == '\xFF' && start[1] == '\xFE') {
//         if(start[2] == '\x00' && start[3] =='\x00') {
//             *signatureLength=4;
//             return "UTF-32LE";
//         } else {
//             *signatureLength=2;
//             return  "UTF-16LE";
//         }
//     } else if(start[0] == '\xEF' && start[1] == '\xBB' && start[2] == '\xBF') {
//         *signatureLength=3;
//         return  "UTF-8";
//     } else if(start[0] == '\x00' && start[1] == '\x00' &&
//               start[2] == '\xFE' && start[3]=='\xFF') {
//         *signatureLength=4;
//         return  "UTF-32BE";
//     } else if(start[0] == '\x0E' && start[1] == '\xFE' && start[2] == '\xFF') {
//         *signatureLength=3;
//         return "SCSU";
//     } else if(start[0] == '\xFB' && start[1] == '\xEE' && start[2] == '\x28') {
//         *signatureLength=3;
//         return "BOCU-1";
//     } else if(start[0] == '\x2B' && start[1] == '\x2F' && start[2] == '\x76') {
//         /*
//          * UTF-7: Initial U+FEFF is encoded as +/v8  or  +/v9  or  +/v+  or  +/v/
//          * depending on the second UTF-16 code unit.
//          * Detect the entire, closed Unicode mode sequence +/v8- for only U+FEFF
//          * if it occurs.
//          *
//          * So far we have +/v
//          */
//         if(start[3] == '\x38' && start[4] == '\x2D') {
//             /* 5 bytes +/v8- */
//             *signatureLength=5;
//             return "UTF-7";
//         } else if(start[3] == '\x38' || start[3] == '\x39' || start[3] == '\x2B' || start[3] == '\x2F') {
//             /* 4 bytes +/v8  or  +/v9  or  +/v+  or  +/v/ */
//             *signatureLength=4;
//             return "UTF-7";
//         }
//     }else if(start[0]=='\xDD' && start[1]== '\x73'&& start[2]=='\x66' && start[3]=='\x73'){
//         *signatureLength=4;
//         return "UTF-EBCDIC";
//     }


//     /* no known Unicode signature byte sequence recognized */
//     *signatureLength=0;
//     return NULL;
// }

// U_CAPI int32_t U_EXPORT2
// ucnv_fromUCountPending(const UConverter* cnv, UErrorCode* status)
// {
//     if(status == NULL || U_FAILURE(*status)){
//         return -1;
//     }
//     if(cnv == NULL){
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return -1;
//     }

//     if(cnv->preFromUFirstCP >= 0){
//         return U16_LENGTH(cnv->preFromUFirstCP)+cnv->preFromULength ;
//     }else if(cnv->preFromULength < 0){
//         return -cnv->preFromULength ;
//     }else if(cnv->fromUChar32 > 0){
//         return 1;
//     }
//     return 0;

// }

// U_CAPI int32_t U_EXPORT2
// ucnv_toUCountPending(const UConverter* cnv, UErrorCode* status){

//     if(status == NULL || U_FAILURE(*status)){
//         return -1;
//     }
//     if(cnv == NULL){
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return -1;
//     }

//     if(cnv->preToULength > 0){
//         return cnv->preToULength ;
//     }else if(cnv->preToULength < 0){
//         return -cnv->preToULength;
//     }else if(cnv->toULength > 0){
//         return cnv->toULength;
//     }
//     return 0;
// }

// U_CAPI UBool U_EXPORT2
// ucnv_isFixedWidth(UConverter *cnv, UErrorCode *status){
//     if (U_FAILURE(*status)) {
//         return FALSE;
//     }

//     if (cnv == NULL) {
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return FALSE;
//     }

//     switch (ucnv_getType(cnv)) {
//         case UCNV_SBCS:
//         case UCNV_DBCS:
//         case UCNV_UTF32_BigEndian:
//         case UCNV_UTF32_LittleEndian:
//         case UCNV_UTF32:
//         case UCNV_US_ASCII:
//             return TRUE;
//         default:
//             return FALSE;
//     }
// }
// #endif

// /*
//  * Hey, Emacs, please set the following:
//  *
//  * Local Variables:
//  * indent-tabs-mode: nil
//  * End:
//  *
//  */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 2000-2016, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// **********************************************************************
// *   file name:  ucnv2022.cpp
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2000feb03
// *   created by: Markus W. Scherer
// *
// *   Change history:
// *
// *   06/29/2000  helena  Major rewrite of the callback APIs.
// *   08/08/2000  Ram     Included support for ISO-2022-JP-2
// *                       Changed implementation of toUnicode
// *                       function
// *   08/21/2000  Ram     Added support for ISO-2022-KR
// *   08/29/2000  Ram     Seperated implementation of EBCDIC to
// *                       ucnvebdc.c
// *   09/20/2000  Ram     Added support for ISO-2022-CN
// *                       Added implementations for getNextUChar()
// *                       for specific 2022 country variants.
// *   10/31/2000  Ram     Implemented offsets logic functions
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION && !UCONFIG_NO_LEGACY_CONVERSION

// #include "unicode/ucnv.h"
// #include "unicode/uset.h"
// #include "unicode/ucnv_err.h"
// #include "unicode/ucnv_cb.h"
// #include "unicode/utf16.h"
// #include "ucnv_imp.h"
// #include "ucnv_bld.h"
// #include "ucnv_cnv.h"
// #include "ucnvmbcs.h"
// #include "cstring.h"
// #include "cmemory.h"
// #include "uassert.h"

// #ifdef U_ENABLE_GENERIC_ISO_2022
// /*
//  * I am disabling the generic ISO-2022 converter after proposing to do so on
//  * the icu mailing list two days ago.
//  *
//  * Reasons:
//  * 1. It does not fully support the ISO-2022/ECMA-35 specification with all of
//  *    its designation sequences, single shifts with return to the previous state,
//  *    switch-with-no-return to UTF-16BE or similar, etc.
//  *    This is unlike the language-specific variants like ISO-2022-JP which
//  *    require a much smaller repertoire of ISO-2022 features.
//  *    These variants continue to be supported.
//  * 2. I believe that no one is really using the generic ISO-2022 converter
//  *    but rather always one of the language-specific variants.
//  *    Note that ICU's generic ISO-2022 converter has always output one escape
//  *    sequence followed by UTF-8 for the whole stream.
//  * 3. Switching between subcharsets is extremely slow, because each time
//  *    the previous converter is closed and a new one opened,
//  *    without any kind of caching, least-recently-used list, etc.
//  * 4. The code is currently buggy, and given the above it does not seem
//  *    reasonable to spend the time on maintenance.
//  * 5. ISO-2022 subcharsets should normally be used with 7-bit byte encodings.
//  *    This means, for example, that when ISO-8859-7 is designated, the following
//  *    ISO-2022 bytes 00..7f should be interpreted as ISO-8859-7 bytes 80..ff.
//  *    The ICU ISO-2022 converter does not handle this - and has no information
//  *    about which subconverter would have to be shifted vs. which is designed
//  *    for 7-bit ISO-2022.
//  *
//  * Markus Scherer 2003-dec-03
//  */
// #endif

// #if !UCONFIG_ONLY_HTML_CONVERSION
// static const char SHIFT_IN_STR[]  = "\x0F";
// // static const char SHIFT_OUT_STR[] = "\x0E";
// #endif

// #define CR      0x0D
// #define LF      0x0A
// #define H_TAB   0x09
// #define V_TAB   0x0B
// #define SPACE   0x20

// enum {
//     HWKANA_START=0xff61,
//     HWKANA_END=0xff9f
// };

// /*
//  * 94-character sets with native byte values A1..FE are encoded in ISO 2022
//  * as bytes 21..7E. (Subtract 0x80.)
//  * 96-character sets with native byte values A0..FF are encoded in ISO 2022
//  * as bytes 20..7F. (Subtract 0x80.)
//  * Do not encode C1 control codes with native bytes 80..9F
//  * as bytes 00..1F (C0 control codes).
//  */
// enum {
//     GR94_START=0xa1,
//     GR94_END=0xfe,
//     GR96_START=0xa0,
//     GR96_END=0xff
// };

// /*
//  * ISO 2022 control codes must not be converted from Unicode
//  * because they would mess up the byte stream.
//  * The bit mask 0x0800c000 has bits set at bit positions 0xe, 0xf, 0x1b
//  * corresponding to SO, SI, and ESC.
//  */
// #define IS_2022_CONTROL(c) (((c)<0x20) && (((uint32_t)1<<(c))&0x0800c000)!=0)

// /* for ISO-2022-JP and -CN implementations */
// typedef enum  {
//         /* shared values */
//         INVALID_STATE=-1,
//         ASCII = 0,

//         SS2_STATE=0x10,
//         SS3_STATE,

//         /* JP */
//         ISO8859_1 = 1 ,
//         ISO8859_7 = 2 ,
//         JISX201  = 3,
//         JISX208 = 4,
//         JISX212 = 5,
//         GB2312  =6,
//         KSC5601 =7,
//         HWKANA_7BIT=8,    /* Halfwidth Katakana 7 bit */

//         /* CN */
//         /* the first few enum constants must keep their values because they correspond to myConverterArray[] */
//         GB2312_1=1,
//         ISO_IR_165=2,
//         CNS_11643=3,

//         /*
//          * these are used in StateEnum and ISO2022State variables,
//          * but CNS_11643 must be used to index into myConverterArray[]
//          */
//         CNS_11643_0=0x20,
//         CNS_11643_1,
//         CNS_11643_2,
//         CNS_11643_3,
//         CNS_11643_4,
//         CNS_11643_5,
//         CNS_11643_6,
//         CNS_11643_7
// } StateEnum;

// /* is the StateEnum charset value for a DBCS charset? */
// #if UCONFIG_ONLY_HTML_CONVERSION
// #define IS_JP_DBCS(cs) (JISX208==(cs))
// #else
// #define IS_JP_DBCS(cs) (JISX208<=(cs) && (cs)<=KSC5601)
// #endif

// #define CSM(cs) ((uint16_t)1<<(cs))

// /*
//  * Each of these charset masks (with index x) contains a bit for a charset in exact correspondence
//  * to whether that charset is used in the corresponding version x of ISO_2022,locale=ja,version=x
//  *
//  * Note: The converter uses some leniency:
//  * - The escape sequence ESC ( I for half-width 7-bit Katakana is recognized in
//  *   all versions, not just JIS7 and JIS8.
//  * - ICU does not distinguish between different versions of JIS X 0208.
//  */
// #if UCONFIG_ONLY_HTML_CONVERSION
// enum { MAX_JA_VERSION=0 };
// #else
// enum { MAX_JA_VERSION=4 };
// #endif
// static const uint16_t jpCharsetMasks[MAX_JA_VERSION+1]={
//     CSM(ASCII)|CSM(JISX201)|CSM(JISX208)|CSM(HWKANA_7BIT),
// #if !UCONFIG_ONLY_HTML_CONVERSION
//     CSM(ASCII)|CSM(JISX201)|CSM(JISX208)|CSM(HWKANA_7BIT)|CSM(JISX212),
//     CSM(ASCII)|CSM(JISX201)|CSM(JISX208)|CSM(HWKANA_7BIT)|CSM(JISX212)|CSM(GB2312)|CSM(KSC5601)|CSM(ISO8859_1)|CSM(ISO8859_7),
//     CSM(ASCII)|CSM(JISX201)|CSM(JISX208)|CSM(HWKANA_7BIT)|CSM(JISX212)|CSM(GB2312)|CSM(KSC5601)|CSM(ISO8859_1)|CSM(ISO8859_7),
//     CSM(ASCII)|CSM(JISX201)|CSM(JISX208)|CSM(HWKANA_7BIT)|CSM(JISX212)|CSM(GB2312)|CSM(KSC5601)|CSM(ISO8859_1)|CSM(ISO8859_7)
// #endif
// };

// typedef enum {
//         ASCII1=0,
//         LATIN1,
//         SBCS,
//         DBCS,
//         MBCS,
//         HWKANA
// }Cnv2022Type;

// typedef struct ISO2022State {
//     int8_t cs[4];       /* charset number for SI (G0)/SO (G1)/SS2 (G2)/SS3 (G3) */
//     int8_t g;           /* 0..3 for G0..G3 (SI/SO/SS2/SS3) */
//     int8_t prevG;       /* g before single shift (SS2 or SS3) */
// } ISO2022State;

// #define UCNV_OPTIONS_VERSION_MASK 0xf
// #define UCNV_2022_MAX_CONVERTERS 10

// typedef struct{
//     UConverterSharedData *myConverterArray[UCNV_2022_MAX_CONVERTERS];
//     UConverter *currentConverter;
//     Cnv2022Type currentType;
//     ISO2022State toU2022State, fromU2022State;
//     uint32_t key;
//     uint32_t version;
// #ifdef U_ENABLE_GENERIC_ISO_2022
//     UBool isFirstBuffer;
// #endif
//     UBool isEmptySegment;
//     char name[30];
//     char locale[3];
// }UConverterDataISO2022;

// /* Protos */
// /* ISO-2022 ----------------------------------------------------------------- */

// /*Forward declaration */
// U_CFUNC void U_CALLCONV
// ucnv_fromUnicode_UTF8(UConverterFromUnicodeArgs * args,
//                       UErrorCode * err);
// U_CFUNC void U_CALLCONV
// ucnv_fromUnicode_UTF8_OFFSETS_LOGIC(UConverterFromUnicodeArgs * args,
//                                     UErrorCode * err);

// #define ESC_2022 0x1B /*ESC*/

// typedef enum
// {
//         INVALID_2022 = -1, /*Doesn't correspond to a valid iso 2022 escape sequence*/
//         VALID_NON_TERMINAL_2022 = 0, /*so far corresponds to a valid iso 2022 escape sequence*/
//         VALID_TERMINAL_2022 = 1, /*corresponds to a valid iso 2022 escape sequence*/
//         VALID_MAYBE_TERMINAL_2022 = 2 /*so far matches one iso 2022 escape sequence, but by adding more characters might match another escape sequence*/
// } UCNV_TableStates_2022;

// /*
// * The way these state transition arrays work is:
// * ex : ESC$B is the sequence for JISX208
// *      a) First Iteration: char is ESC
// *          i) Get the value of ESC from normalize_esq_chars_2022[] with int value of ESC as index
// *             int x = normalize_esq_chars_2022[27] which is equal to 1
// *         ii) Search for this value in escSeqStateTable_Key_2022[]
// *             value of x is stored at escSeqStateTable_Key_2022[0]
// *        iii) Save this index as offset
// *         iv) Get state of this sequence from escSeqStateTable_Value_2022[]
// *             escSeqStateTable_Value_2022[offset], which is VALID_NON_TERMINAL_2022
// *     b) Switch on this state and continue to next char
// *          i) Get the value of $ from normalize_esq_chars_2022[] with int value of $ as index
// *             which is normalize_esq_chars_2022[36] == 4
// *         ii) x is currently 1(from above)
// *               x<<=5 -- x is now 32
// *               x+=normalize_esq_chars_2022[36]
// *               now x is 36
// *        iii) Search for this value in escSeqStateTable_Key_2022[]
// *             value of x is stored at escSeqStateTable_Key_2022[2], so offset is 2
// *         iv) Get state of this sequence from escSeqStateTable_Value_2022[]
// *             escSeqStateTable_Value_2022[offset], which is VALID_NON_TERMINAL_2022
// *     c) Switch on this state and continue to next char
// *        i)  Get the value of B from normalize_esq_chars_2022[] with int value of B as index
// *        ii) x is currently 36 (from above)
// *            x<<=5 -- x is now 1152
// *            x+=normalize_esq_chars_2022[66]
// *            now x is 1161
// *       iii) Search for this value in escSeqStateTable_Key_2022[]
// *            value of x is stored at escSeqStateTable_Key_2022[21], so offset is 21
// *        iv) Get state of this sequence from escSeqStateTable_Value_2022[21]
// *            escSeqStateTable_Value_2022[offset], which is VALID_TERMINAL_2022
// *         v) Get the converter name form escSeqStateTable_Result_2022[21] which is JISX208
// */


// /*Below are the 3 arrays depicting a state transition table*/
// static const int8_t normalize_esq_chars_2022[256] = {
// /*       0      1       2       3       4      5       6        7       8       9           */

//          0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,1      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,4      ,7      ,29      ,0
//         ,2     ,24     ,26     ,27     ,0      ,3      ,23     ,6      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,5      ,8      ,9      ,10     ,11     ,12
//         ,13    ,14     ,15     ,16     ,17     ,18     ,19     ,20     ,25     ,28
//         ,0     ,0      ,21     ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,22    ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0      ,0
//         ,0     ,0      ,0      ,0      ,0      ,0
// };

// #ifdef U_ENABLE_GENERIC_ISO_2022
// /*
//  * When the generic ISO-2022 converter is completely removed, not just disabled
//  * per #ifdef, then the following state table and the associated tables that are
//  * dimensioned with MAX_STATES_2022 should be trimmed.
//  *
//  * Especially, VALID_MAYBE_TERMINAL_2022 will not be used any more, and all of
//  * the associated escape sequences starting with ESC ( B should be removed.
//  * This includes the ones with key values 1097 and all of the ones above 1000000.
//  *
//  * For the latter, the tables can simply be truncated.
//  * For the former, since the tables must be kept parallel, it is probably best
//  * to simply duplicate an adjacent table cell, parallel in all tables.
//  *
//  * It may make sense to restructure the tables, especially by using small search
//  * tables for the variants instead of indexing them parallel to the table here.
//  */
// #endif

// #define MAX_STATES_2022 74
// static const int32_t escSeqStateTable_Key_2022[MAX_STATES_2022] = {
// /*   0           1           2           3           4           5           6           7           8           9           */

//      1          ,34         ,36         ,39         ,55         ,57         ,60         ,61         ,1093       ,1096
//     ,1097       ,1098       ,1099       ,1100       ,1101       ,1102       ,1103       ,1104       ,1105       ,1106
//     ,1109       ,1154       ,1157       ,1160       ,1161       ,1176       ,1178       ,1179       ,1254       ,1257
//     ,1768       ,1773       ,1957       ,35105      ,36933      ,36936      ,36937      ,36938      ,36939      ,36940
//     ,36942      ,36943      ,36944      ,36945      ,36946      ,36947      ,36948      ,37640      ,37642      ,37644
//     ,37646      ,37711      ,37744      ,37745      ,37746      ,37747      ,37748      ,40133      ,40136      ,40138
//     ,40139      ,40140      ,40141      ,1123363    ,35947624   ,35947625   ,35947626   ,35947627   ,35947629   ,35947630
//     ,35947631   ,35947635   ,35947636   ,35947638
// };

// #ifdef U_ENABLE_GENERIC_ISO_2022

// static const char* const escSeqStateTable_Result_2022[MAX_STATES_2022] = {
//  /*  0                      1                        2                      3                   4                   5                        6                      7                       8                       9    */

//      NULL                   ,NULL                   ,NULL                   ,NULL               ,NULL               ,NULL                   ,NULL                   ,NULL                   ,"latin1"               ,"latin1"
//     ,"latin1"               ,"ibm-865"              ,"ibm-865"              ,"ibm-865"          ,"ibm-865"          ,"ibm-865"              ,"ibm-865"              ,"JISX0201"             ,"JISX0201"             ,"latin1"
//     ,"latin1"               ,NULL                   ,"JISX-208"             ,"ibm-5478"         ,"JISX-208"         ,NULL                   ,NULL                   ,NULL                   ,NULL                   ,"UTF8"
//     ,"ISO-8859-1"           ,"ISO-8859-7"           ,"JIS-X-208"            ,NULL               ,"ibm-955"          ,"ibm-367"              ,"ibm-952"              ,"ibm-949"              ,"JISX-212"             ,"ibm-1383"
//     ,"ibm-952"              ,"ibm-964"              ,"ibm-964"              ,"ibm-964"          ,"ibm-964"          ,"ibm-964"              ,"ibm-964"              ,"ibm-5478"         ,"ibm-949"              ,"ISO-IR-165"
//     ,"CNS-11643-1992,1"     ,"CNS-11643-1992,2"     ,"CNS-11643-1992,3"     ,"CNS-11643-1992,4" ,"CNS-11643-1992,5" ,"CNS-11643-1992,6"     ,"CNS-11643-1992,7"     ,"UTF16_PlatformEndian" ,"UTF16_PlatformEndian" ,"UTF16_PlatformEndian"
//     ,"UTF16_PlatformEndian" ,"UTF16_PlatformEndian" ,"UTF16_PlatformEndian" ,NULL               ,"latin1"           ,"ibm-912"              ,"ibm-913"              ,"ibm-914"              ,"ibm-813"              ,"ibm-1089"
//     ,"ibm-920"              ,"ibm-915"              ,"ibm-915"              ,"latin1"
// };

// #endif

// static const int8_t escSeqStateTable_Value_2022[MAX_STATES_2022] = {
// /*          0                           1                         2                             3                           4                           5                               6                        7                          8                           9       */
//      VALID_NON_TERMINAL_2022    ,VALID_NON_TERMINAL_2022    ,VALID_NON_TERMINAL_2022    ,VALID_NON_TERMINAL_2022     ,VALID_NON_TERMINAL_2022   ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_NON_TERMINAL_2022    ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022
//     ,VALID_MAYBE_TERMINAL_2022  ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022
//     ,VALID_TERMINAL_2022        ,VALID_NON_TERMINAL_2022    ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_NON_TERMINAL_2022    ,VALID_NON_TERMINAL_2022    ,VALID_NON_TERMINAL_2022    ,VALID_NON_TERMINAL_2022    ,VALID_TERMINAL_2022
//     ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_NON_TERMINAL_2022    ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022
//     ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022
//     ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022
//     ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_NON_TERMINAL_2022    ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022
//     ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022        ,VALID_TERMINAL_2022
// };

// /* Type def for refactoring changeState_2022 code*/
// typedef enum{
// #ifdef U_ENABLE_GENERIC_ISO_2022
//     ISO_2022=0,
// #endif
//     ISO_2022_JP=1,
// #if !UCONFIG_ONLY_HTML_CONVERSION
//     ISO_2022_KR=2,
//     ISO_2022_CN=3
// #endif
// } Variant2022;

// /*********** ISO 2022 Converter Protos ***********/
// static void U_CALLCONV
// _ISO2022Open(UConverter *cnv, UConverterLoadArgs *pArgs, UErrorCode *errorCode);

// static void U_CALLCONV
//  _ISO2022Close(UConverter *converter);

// static void U_CALLCONV
// _ISO2022Reset(UConverter *converter, UConverterResetChoice choice);

// U_CDECL_BEGIN
// static const char * U_CALLCONV
// _ISO2022getName(const UConverter* cnv);
// U_CDECL_END

// static void  U_CALLCONV
// _ISO_2022_WriteSub(UConverterFromUnicodeArgs *args, int32_t offsetIndex, UErrorCode *err);

// U_CDECL_BEGIN
// static UConverter * U_CALLCONV
// _ISO_2022_SafeClone(const UConverter *cnv, void *stackBuffer, int32_t *pBufferSize, UErrorCode *status);

// U_CDECL_END

// #ifdef U_ENABLE_GENERIC_ISO_2022
// static void U_CALLCONV
// T_UConverter_toUnicode_ISO_2022_OFFSETS_LOGIC(UConverterToUnicodeArgs* args, UErrorCode* err);
// #endif

// namespace {

// /*const UConverterSharedData _ISO2022Data;*/
// extern const UConverterSharedData _ISO2022JPData;

// #if !UCONFIG_ONLY_HTML_CONVERSION
// extern const UConverterSharedData _ISO2022KRData;
// extern const UConverterSharedData _ISO2022CNData;
// #endif

// }  // namespace

// /*************** Converter implementations ******************/

// /* The purpose of this function is to get around gcc compiler warnings. */
// static inline void
// fromUWriteUInt8(UConverter *cnv,
//                  const char *bytes, int32_t length,
//                  uint8_t **target, const char *targetLimit,
//                  int32_t **offsets,
//                  int32_t sourceIndex,
//                  UErrorCode *pErrorCode)
// {
//     char *targetChars = (char *)*target;
//     ucnv_fromUWriteBytes(cnv, bytes, length, &targetChars, targetLimit,
//                          offsets, sourceIndex, pErrorCode);
//     *target = (uint8_t*)targetChars;

// }

// static inline void
// setInitialStateToUnicodeKR(UConverter* /*converter*/, UConverterDataISO2022 *myConverterData){
//     if(myConverterData->version == 1) {
//         UConverter *cnv = myConverterData->currentConverter;

//         cnv->toUnicodeStatus=0;     /* offset */
//         cnv->mode=0;                /* state */
//         cnv->toULength=0;           /* byteIndex */
//     }
// }

// static inline void
// setInitialStateFromUnicodeKR(UConverter* converter,UConverterDataISO2022 *myConverterData){
//    /* in ISO-2022-KR the designator sequence appears only once
//     * in a file so we append it only once
//     */
//     if( converter->charErrorBufferLength==0){

//         converter->charErrorBufferLength = 4;
//         converter->charErrorBuffer[0] = 0x1b;
//         converter->charErrorBuffer[1] = 0x24;
//         converter->charErrorBuffer[2] = 0x29;
//         converter->charErrorBuffer[3] = 0x43;
//     }
//     if(myConverterData->version == 1) {
//         UConverter *cnv = myConverterData->currentConverter;

//         cnv->fromUChar32=0;
//         cnv->fromUnicodeStatus=1;   /* prevLength */
//     }
// }

// static void U_CALLCONV
// _ISO2022Open(UConverter *cnv, UConverterLoadArgs *pArgs, UErrorCode *errorCode){

//     char myLocale[6]={' ',' ',' ',' ',' ',' '};

//     cnv->extraInfo = uprv_malloc (sizeof (UConverterDataISO2022));
//     if(cnv->extraInfo != NULL) {
//         UConverterNamePieces stackPieces;
//         UConverterLoadArgs stackArgs=UCNV_LOAD_ARGS_INITIALIZER;
//         UConverterDataISO2022 *myConverterData=(UConverterDataISO2022 *) cnv->extraInfo;
//         uint32_t version;

//         stackArgs.onlyTestIsLoadable = pArgs->onlyTestIsLoadable;

//         uprv_memset(myConverterData, 0, sizeof(UConverterDataISO2022));
//         myConverterData->currentType = ASCII1;
//         cnv->fromUnicodeStatus =FALSE;
//         if(pArgs->locale){
//             uprv_strncpy(myLocale, pArgs->locale, sizeof(myLocale));
//         }
//         version = pArgs->options & UCNV_OPTIONS_VERSION_MASK;
//         myConverterData->version = version;
//         if(myLocale[0]=='j' && (myLocale[1]=='a'|| myLocale[1]=='p') &&
//             (myLocale[2]=='_' || myLocale[2]=='\0'))
//         {
//             /* open the required converters and cache them */
//             if(version>MAX_JA_VERSION) {
//                 // ICU 55 fails to open a converter for an unsupported version.
//                 // Previously, it fell back to version 0, but that would yield
//                 // unexpected behavior.
//                 *errorCode = U_MISSING_RESOURCE_ERROR;
//                 return;
//             }
//             if(jpCharsetMasks[version]&CSM(ISO8859_7)) {
//                 myConverterData->myConverterArray[ISO8859_7] =
//                     ucnv_loadSharedData("ISO8859_7", &stackPieces, &stackArgs, errorCode);
//             }
//             myConverterData->myConverterArray[JISX208] =
//                 ucnv_loadSharedData("Shift-JIS", &stackPieces, &stackArgs, errorCode);
//             if(jpCharsetMasks[version]&CSM(JISX212)) {
//                 myConverterData->myConverterArray[JISX212] =
//                     ucnv_loadSharedData("jisx-212", &stackPieces, &stackArgs, errorCode);
//             }
//             if(jpCharsetMasks[version]&CSM(GB2312)) {
//                 myConverterData->myConverterArray[GB2312] =
//                     ucnv_loadSharedData("ibm-5478", &stackPieces, &stackArgs, errorCode);   /* gb_2312_80-1 */
//             }
//             if(jpCharsetMasks[version]&CSM(KSC5601)) {
//                 myConverterData->myConverterArray[KSC5601] =
//                     ucnv_loadSharedData("ksc_5601", &stackPieces, &stackArgs, errorCode);
//             }

//             /* set the function pointers to appropriate funtions */
//             cnv->sharedData=(UConverterSharedData*)(&_ISO2022JPData);
//             uprv_strcpy(myConverterData->locale,"ja");

//             (void)uprv_strcpy(myConverterData->name,"ISO_2022,locale=ja,version=");
//             size_t len = uprv_strlen(myConverterData->name);
//             myConverterData->name[len]=(char)(myConverterData->version+(int)'0');
//             myConverterData->name[len+1]='\0';
//         }
// #if !UCONFIG_ONLY_HTML_CONVERSION
//         else if(myLocale[0]=='k' && (myLocale[1]=='o'|| myLocale[1]=='r') &&
//             (myLocale[2]=='_' || myLocale[2]=='\0'))
//         {
//             if(version>1) {
//                 // ICU 55 fails to open a converter for an unsupported version.
//                 // Previously, it fell back to version 0, but that would yield
//                 // unexpected behavior.
//                 *errorCode = U_MISSING_RESOURCE_ERROR;
//                 return;
//             }
//             const char *cnvName;
//             if(version==1) {
//                 cnvName="icu-internal-25546";
//             } else {
//                 cnvName="ibm-949";
//                 myConverterData->version=version=0;
//             }
//             if(pArgs->onlyTestIsLoadable) {
//                 ucnv_canCreateConverter(cnvName, errorCode);  /* errorCode carries result */
//                 uprv_free(cnv->extraInfo);
//                 cnv->extraInfo=NULL;
//                 return;
//             } else {
//                 myConverterData->currentConverter=ucnv_open(cnvName, errorCode);
//                 if (U_FAILURE(*errorCode)) {
//                     _ISO2022Close(cnv);
//                     return;
//                 }

//                 if(version==1) {
//                     (void)uprv_strcpy(myConverterData->name,"ISO_2022,locale=ko,version=1");
//                     uprv_memcpy(cnv->subChars, myConverterData->currentConverter->subChars, 4);
//                     cnv->subCharLen = myConverterData->currentConverter->subCharLen;
//                 }else{
//                     (void)uprv_strcpy(myConverterData->name,"ISO_2022,locale=ko,version=0");
//                 }

//                 /* initialize the state variables */
//                 setInitialStateToUnicodeKR(cnv, myConverterData);
//                 setInitialStateFromUnicodeKR(cnv, myConverterData);

//                 /* set the function pointers to appropriate funtions */
//                 cnv->sharedData=(UConverterSharedData*)&_ISO2022KRData;
//                 uprv_strcpy(myConverterData->locale,"ko");
//             }
//         }
//         else if(((myLocale[0]=='z' && myLocale[1]=='h') || (myLocale[0]=='c'&& myLocale[1]=='n'))&&
//             (myLocale[2]=='_' || myLocale[2]=='\0'))
//         {
//             if(version>2) {
//                 // ICU 55 fails to open a converter for an unsupported version.
//                 // Previously, it fell back to version 0, but that would yield
//                 // unexpected behavior.
//                 *errorCode = U_MISSING_RESOURCE_ERROR;
//                 return;
//             }

//             /* open the required converters and cache them */
//             myConverterData->myConverterArray[GB2312_1] =
//                 ucnv_loadSharedData("ibm-5478", &stackPieces, &stackArgs, errorCode);
//             if(version==1) {
//                 myConverterData->myConverterArray[ISO_IR_165] =
//                     ucnv_loadSharedData("iso-ir-165", &stackPieces, &stackArgs, errorCode);
//             }
//             myConverterData->myConverterArray[CNS_11643] =
//                 ucnv_loadSharedData("cns-11643-1992", &stackPieces, &stackArgs, errorCode);


//             /* set the function pointers to appropriate funtions */
//             cnv->sharedData=(UConverterSharedData*)&_ISO2022CNData;
//             uprv_strcpy(myConverterData->locale,"cn");

//             if (version==0){
//                 myConverterData->version = 0;
//                 (void)uprv_strcpy(myConverterData->name,"ISO_2022,locale=zh,version=0");
//             }else if (version==1){
//                 myConverterData->version = 1;
//                 (void)uprv_strcpy(myConverterData->name,"ISO_2022,locale=zh,version=1");
//             }else {
//                 myConverterData->version = 2;
//                 (void)uprv_strcpy(myConverterData->name,"ISO_2022,locale=zh,version=2");
//             }
//         }
// #endif  // !UCONFIG_ONLY_HTML_CONVERSION
//         else{
// #ifdef U_ENABLE_GENERIC_ISO_2022
//             myConverterData->isFirstBuffer = TRUE;

//             /* append the UTF-8 escape sequence */
//             cnv->charErrorBufferLength = 3;
//             cnv->charErrorBuffer[0] = 0x1b;
//             cnv->charErrorBuffer[1] = 0x25;
//             cnv->charErrorBuffer[2] = 0x42;

//             cnv->sharedData=(UConverterSharedData*)&_ISO2022Data;
//             /* initialize the state variables */
//             uprv_strcpy(myConverterData->name,"ISO_2022");
// #else
//             *errorCode = U_MISSING_RESOURCE_ERROR;
//             // Was U_UNSUPPORTED_ERROR but changed in ICU 55 to a more standard
//             // data loading error code.
//             return;
// #endif
//         }

//         cnv->maxBytesPerUChar=cnv->sharedData->staticData->maxBytesPerChar;

//         if(U_FAILURE(*errorCode) || pArgs->onlyTestIsLoadable) {
//             _ISO2022Close(cnv);
//         }
//     } else {
//         *errorCode = U_MEMORY_ALLOCATION_ERROR;
//     }
// }


// static void U_CALLCONV
// _ISO2022Close(UConverter *converter) {
//     UConverterDataISO2022* myData =(UConverterDataISO2022 *) (converter->extraInfo);
//     UConverterSharedData **array = myData->myConverterArray;
//     int32_t i;

//     if (converter->extraInfo != NULL) {
//         /*close the array of converter pointers and free the memory*/
//         for (i=0; i<UCNV_2022_MAX_CONVERTERS; i++) {
//             if(array[i]!=NULL) {
//                 ucnv_unloadSharedDataIfReady(array[i]);
//             }
//         }

//         ucnv_close(myData->currentConverter);

//         if(!converter->isExtraLocal){
//             uprv_free (converter->extraInfo);
//             converter->extraInfo = NULL;
//         }
//     }
// }

// static void U_CALLCONV
// _ISO2022Reset(UConverter *converter, UConverterResetChoice choice) {
//     UConverterDataISO2022 *myConverterData=(UConverterDataISO2022 *) (converter->extraInfo);
//     if(choice<=UCNV_RESET_TO_UNICODE) {
//         uprv_memset(&myConverterData->toU2022State, 0, sizeof(ISO2022State));
//         myConverterData->key = 0;
//         myConverterData->isEmptySegment = FALSE;
//     }
//     if(choice!=UCNV_RESET_TO_UNICODE) {
//         uprv_memset(&myConverterData->fromU2022State, 0, sizeof(ISO2022State));
//     }
// #ifdef U_ENABLE_GENERIC_ISO_2022
//     if(myConverterData->locale[0] == 0){
//         if(choice<=UCNV_RESET_TO_UNICODE) {
//             myConverterData->isFirstBuffer = TRUE;
//             myConverterData->key = 0;
//             if (converter->mode == UCNV_SO){
//                 ucnv_close (myConverterData->currentConverter);
//                 myConverterData->currentConverter=NULL;
//             }
//             converter->mode = UCNV_SI;
//         }
//         if(choice!=UCNV_RESET_TO_UNICODE) {
//             /* re-append UTF-8 escape sequence */
//             converter->charErrorBufferLength = 3;
//             converter->charErrorBuffer[0] = 0x1b;
//             converter->charErrorBuffer[1] = 0x28;
//             converter->charErrorBuffer[2] = 0x42;
//         }
//     }
//     else
// #endif
//     {
//         /* reset the state variables */
//         if(myConverterData->locale[0] == 'k'){
//             if(choice<=UCNV_RESET_TO_UNICODE) {
//                 setInitialStateToUnicodeKR(converter, myConverterData);
//             }
//             if(choice!=UCNV_RESET_TO_UNICODE) {
//                 setInitialStateFromUnicodeKR(converter, myConverterData);
//             }
//         }
//     }
// }

// U_CDECL_BEGIN

// static const char * U_CALLCONV
// _ISO2022getName(const UConverter* cnv){
//     if(cnv->extraInfo){
//         UConverterDataISO2022* myData= (UConverterDataISO2022*)cnv->extraInfo;
//         return myData->name;
//     }
//     return NULL;
// }

// U_CDECL_END


// /*************** to unicode *******************/
// /****************************************************************************
//  * Recognized escape sequences are
//  * <ESC>(B  ASCII
//  * <ESC>.A  ISO-8859-1
//  * <ESC>.F  ISO-8859-7
//  * <ESC>(J  JISX-201
//  * <ESC>(I  JISX-201
//  * <ESC>$B  JISX-208
//  * <ESC>$@  JISX-208
//  * <ESC>$(D JISX-212
//  * <ESC>$A  GB2312
//  * <ESC>$(C KSC5601
//  */
// static const int8_t nextStateToUnicodeJP[MAX_STATES_2022]= {
// /*      0                1               2               3               4               5               6               7               8               9    */
//     INVALID_STATE   ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,SS2_STATE      ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,ASCII          ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,JISX201        ,HWKANA_7BIT    ,JISX201        ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,JISX208        ,GB2312         ,JISX208        ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,ISO8859_1      ,ISO8859_7      ,JISX208        ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,KSC5601        ,JISX212        ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
// };

// #if !UCONFIG_ONLY_HTML_CONVERSION
// /*************** to unicode *******************/
// static const int8_t nextStateToUnicodeCN[MAX_STATES_2022]= {
// /*      0                1               2               3               4               5               6               7               8               9    */
//      INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,SS2_STATE      ,SS3_STATE      ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,GB2312_1       ,INVALID_STATE  ,ISO_IR_165
//     ,CNS_11643_1    ,CNS_11643_2    ,CNS_11643_3    ,CNS_11643_4    ,CNS_11643_5    ,CNS_11643_6    ,CNS_11643_7    ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
//     ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE  ,INVALID_STATE
// };
// #endif


// static UCNV_TableStates_2022
// getKey_2022(char c,int32_t* key,int32_t* offset){
//     int32_t togo;
//     int32_t low = 0;
//     int32_t hi = MAX_STATES_2022;
//     int32_t oldmid=0;

//     togo = normalize_esq_chars_2022[(uint8_t)c];
//     if(togo == 0) {
//         /* not a valid character anywhere in an escape sequence */
//         *key = 0;
//         *offset = 0;
//         return INVALID_2022;
//     }
//     togo = (*key << 5) + togo;

//     while (hi != low)  /*binary search*/{

//         int32_t mid = (hi+low) >> 1; /*Finds median*/

//         if (mid == oldmid)
//             break;

//         if (escSeqStateTable_Key_2022[mid] > togo){
//             hi = mid;
//         }
//         else if (escSeqStateTable_Key_2022[mid] < togo){
//             low = mid;
//         }
//         else /*we found it*/{
//             *key = togo;
//             *offset = mid;
//             return (UCNV_TableStates_2022)escSeqStateTable_Value_2022[mid];
//         }
//         oldmid = mid;

//     }

//     *key = 0;
//     *offset = 0;
//     return INVALID_2022;
// }

// /*runs through a state machine to determine the escape sequence - codepage correspondance
//  */
// static void
// changeState_2022(UConverter* _this,
//                 const char** source,
//                 const char* sourceLimit,
//                 Variant2022 var,
//                 UErrorCode* err){
//     UCNV_TableStates_2022 value;
//     UConverterDataISO2022* myData2022 = ((UConverterDataISO2022*)_this->extraInfo);
//     uint32_t key = myData2022->key;
//     int32_t offset = 0;
//     int8_t initialToULength = _this->toULength;
//     char c;

//     value = VALID_NON_TERMINAL_2022;
//     while (*source < sourceLimit) {
//         c = *(*source)++;
//         _this->toUBytes[_this->toULength++]=(uint8_t)c;
//         value = getKey_2022(c,(int32_t *) &key, &offset);

//         switch (value){

//         case VALID_NON_TERMINAL_2022 :
//             /* continue with the loop */
//             break;

//         case VALID_TERMINAL_2022:
//             key = 0;
//             goto DONE;

//         case INVALID_2022:
//             goto DONE;

//         case VALID_MAYBE_TERMINAL_2022:
// #ifdef U_ENABLE_GENERIC_ISO_2022
//             /* ESC ( B is ambiguous only for ISO_2022 itself */
//             if(var == ISO_2022) {
//                 /* discard toUBytes[] for ESC ( B because this sequence is correct and complete */
//                 _this->toULength = 0;

//                 /* TODO need to indicate that ESC ( B was seen; if failure, then need to replay from source or from MBCS-style replay */

//                 /* continue with the loop */
//                 value = VALID_NON_TERMINAL_2022;
//                 break;
//             } else
// #endif
//             {
//                 /* not ISO_2022 itself, finish here */
//                 value = VALID_TERMINAL_2022;
//                 key = 0;
//                 goto DONE;
//             }
//         }
//     }

// DONE:
//     myData2022->key = key;

//     if (value == VALID_NON_TERMINAL_2022) {
//         /* indicate that the escape sequence is incomplete: key!=0 */
//         return;
//     } else if (value == INVALID_2022 ) {
//         *err = U_ILLEGAL_ESCAPE_SEQUENCE;
//     } else /* value == VALID_TERMINAL_2022 */ {
//         switch(var){
// #ifdef U_ENABLE_GENERIC_ISO_2022
//         case ISO_2022:
//         {
//             const char *chosenConverterName = escSeqStateTable_Result_2022[offset];
//             if(chosenConverterName == NULL) {
//                 /* SS2 or SS3 */
//                 *err = U_UNSUPPORTED_ESCAPE_SEQUENCE;
//                 _this->toUCallbackReason = UCNV_UNASSIGNED;
//                 return;
//             }

//             _this->mode = UCNV_SI;
//             ucnv_close(myData2022->currentConverter);
//             myData2022->currentConverter = myUConverter = ucnv_open(chosenConverterName, err);
//             if(U_SUCCESS(*err)) {
//                 myUConverter->fromCharErrorBehaviour = UCNV_TO_U_CALLBACK_STOP;
//                 _this->mode = UCNV_SO;
//             }
//             break;
//         }
// #endif
//         case ISO_2022_JP:
//             {
//                 StateEnum tempState=(StateEnum)nextStateToUnicodeJP[offset];
//                 switch(tempState) {
//                 case INVALID_STATE:
//                     *err = U_UNSUPPORTED_ESCAPE_SEQUENCE;
//                     break;
//                 case SS2_STATE:
//                     if(myData2022->toU2022State.cs[2]!=0) {
//                         if(myData2022->toU2022State.g<2) {
//                             myData2022->toU2022State.prevG=myData2022->toU2022State.g;
//                         }
//                         myData2022->toU2022State.g=2;
//                     } else {
//                         /* illegal to have SS2 before a matching designator */
//                         *err = U_ILLEGAL_ESCAPE_SEQUENCE;
//                     }
//                     break;
//                 /* case SS3_STATE: not used in ISO-2022-JP-x */
//                 case ISO8859_1:
//                 case ISO8859_7:
//                     if((jpCharsetMasks[myData2022->version] & CSM(tempState)) == 0) {
//                         *err = U_UNSUPPORTED_ESCAPE_SEQUENCE;
//                     } else {
//                         /* G2 charset for SS2 */
//                         myData2022->toU2022State.cs[2]=(int8_t)tempState;
//                     }
//                     break;
//                 default:
//                     if((jpCharsetMasks[myData2022->version] & CSM(tempState)) == 0) {
//                         *err = U_UNSUPPORTED_ESCAPE_SEQUENCE;
//                     } else {
//                         /* G0 charset */
//                         myData2022->toU2022State.cs[0]=(int8_t)tempState;
//                     }
//                     break;
//                 }
//             }
//             break;
// #if !UCONFIG_ONLY_HTML_CONVERSION
//         case ISO_2022_CN:
//             {
//                 StateEnum tempState=(StateEnum)nextStateToUnicodeCN[offset];
//                 switch(tempState) {
//                 case INVALID_STATE:
//                     *err = U_UNSUPPORTED_ESCAPE_SEQUENCE;
//                     break;
//                 case SS2_STATE:
//                     if(myData2022->toU2022State.cs[2]!=0) {
//                         if(myData2022->toU2022State.g<2) {
//                             myData2022->toU2022State.prevG=myData2022->toU2022State.g;
//                         }
//                         myData2022->toU2022State.g=2;
//                     } else {
//                         /* illegal to have SS2 before a matching designator */
//                         *err = U_ILLEGAL_ESCAPE_SEQUENCE;
//                     }
//                     break;
//                 case SS3_STATE:
//                     if(myData2022->toU2022State.cs[3]!=0) {
//                         if(myData2022->toU2022State.g<2) {
//                             myData2022->toU2022State.prevG=myData2022->toU2022State.g;
//                         }
//                         myData2022->toU2022State.g=3;
//                     } else {
//                         /* illegal to have SS3 before a matching designator */
//                         *err = U_ILLEGAL_ESCAPE_SEQUENCE;
//                     }
//                     break;
//                 case ISO_IR_165:
//                     if(myData2022->version==0) {
//                         *err = U_UNSUPPORTED_ESCAPE_SEQUENCE;
//                         break;
//                     }
//                     U_FALLTHROUGH;
//                 case GB2312_1:
//                     U_FALLTHROUGH;
//                 case CNS_11643_1:
//                     myData2022->toU2022State.cs[1]=(int8_t)tempState;
//                     break;
//                 case CNS_11643_2:
//                     myData2022->toU2022State.cs[2]=(int8_t)tempState;
//                     break;
//                 default:
//                     /* other CNS 11643 planes */
//                     if(myData2022->version==0) {
//                         *err = U_UNSUPPORTED_ESCAPE_SEQUENCE;
//                     } else {
//                        myData2022->toU2022State.cs[3]=(int8_t)tempState;
//                     }
//                     break;
//                 }
//             }
//             break;
//         case ISO_2022_KR:
//             if(offset==0x30){
//                 /* nothing to be done, just accept this one escape sequence */
//             } else {
//                 *err = U_UNSUPPORTED_ESCAPE_SEQUENCE;
//             }
//             break;
// #endif  // !UCONFIG_ONLY_HTML_CONVERSION

//         default:
//             *err = U_ILLEGAL_ESCAPE_SEQUENCE;
//             break;
//         }
//     }
//     if(U_SUCCESS(*err)) {
//         _this->toULength = 0;
//     } else if(*err==U_ILLEGAL_ESCAPE_SEQUENCE) {
//         if(_this->toULength>1) {
//             /*
//              * Ticket 5691: consistent illegal sequences:
//              * - We include at least the first byte (ESC) in the illegal sequence.
//              * - If any of the non-initial bytes could be the start of a character,
//              *   we stop the illegal sequence before the first one of those.
//              *   In escape sequences, all following bytes are "printable", that is,
//              *   unless they are completely illegal (>7f in SBCS, outside 21..7e in DBCS),
//              *   they are valid single/lead bytes.
//              *   For simplicity, we always only report the initial ESC byte as the
//              *   illegal sequence and back out all other bytes we looked at.
//              */
//             /* Back out some bytes. */
//             int8_t backOutDistance=_this->toULength-1;
//             int8_t bytesFromThisBuffer=_this->toULength-initialToULength;
//             if(backOutDistance<=bytesFromThisBuffer) {
//                 /* same as initialToULength<=1 */
//                 *source-=backOutDistance;
//             } else {
//                 /* Back out bytes from the previous buffer: Need to replay them. */
//                 _this->preToULength=(int8_t)(bytesFromThisBuffer-backOutDistance);
//                 /* same as -(initialToULength-1) */
//                 /* preToULength is negative! */
//                 uprv_memcpy(_this->preToU, _this->toUBytes+1, -_this->preToULength);
//                 *source-=bytesFromThisBuffer;
//             }
//             _this->toULength=1;
//         }
//     } else if(*err==U_UNSUPPORTED_ESCAPE_SEQUENCE) {
//         _this->toUCallbackReason = UCNV_UNASSIGNED;
//     }
// }

// #if !UCONFIG_ONLY_HTML_CONVERSION
// /*Checks the characters of the buffer against valid 2022 escape sequences
// *if the match we return a pointer to the initial start of the sequence otherwise
// *we return sourceLimit
// */
// /*for 2022 looks ahead in the stream
//  *to determine the longest possible convertible
//  *data stream
//  */
// static inline const char*
// getEndOfBuffer_2022(const char** source,
//                    const char* sourceLimit,
//                    UBool /*flush*/){

//     const char* mySource = *source;

// #ifdef U_ENABLE_GENERIC_ISO_2022
//     if (*source >= sourceLimit)
//         return sourceLimit;

//     do{

//         if (*mySource == ESC_2022){
//             int8_t i;
//             int32_t key = 0;
//             int32_t offset;
//             UCNV_TableStates_2022 value = VALID_NON_TERMINAL_2022;

//             /* Kludge: I could not
//             * figure out the reason for validating an escape sequence
//             * twice - once here and once in changeState_2022().
//             * is it possible to have an ESC character in a ISO2022
//             * byte stream which is valid in a code page? Is it legal?
//             */
//             for (i=0;
//             (mySource+i < sourceLimit)&&(value == VALID_NON_TERMINAL_2022);
//             i++) {
//                 value =  getKey_2022(*(mySource+i), &key, &offset);
//             }
//             if (value > 0 || *mySource==ESC_2022)
//                 return mySource;

//             if ((value == VALID_NON_TERMINAL_2022)&&(!flush) )
//                 return sourceLimit;
//         }
//     }while (++mySource < sourceLimit);

//     return sourceLimit;
// #else
//     while(mySource < sourceLimit && *mySource != ESC_2022) {
//         ++mySource;
//     }
//     return mySource;
// #endif
// }
// #endif

// /* This inline function replicates code in _MBCSFromUChar32() function in ucnvmbcs.c
//  * any future change in _MBCSFromUChar32() function should be reflected here.
//  * @return number of bytes in *value; negative number if fallback; 0 if no mapping
//  */
// static inline int32_t
// MBCS_FROM_UCHAR32_ISO2022(UConverterSharedData* sharedData,
//                                          UChar32 c,
//                                          uint32_t* value,
//                                          UBool useFallback,
//                                          int outputType)
// {
//     const int32_t *cx;
//     const uint16_t *table;
//     uint32_t stage2Entry;
//     uint32_t myValue;
//     int32_t length;
//     const uint8_t *p;
//     /*
//      * TODO(markus): Use and require new, faster MBCS conversion table structures.
//      * Use internal version of ucnv_open() that verifies that the new structures are available,
//      * else U_INTERNAL_PROGRAM_ERROR.
//      */
//     /* BMP-only codepages are stored without stage 1 entries for supplementary code points */
//     if(c<0x10000 || (sharedData->mbcs.unicodeMask&UCNV_HAS_SUPPLEMENTARY)) {
//         table=sharedData->mbcs.fromUnicodeTable;
//         stage2Entry=MBCS_STAGE_2_FROM_U(table, c);
//         /* get the bytes and the length for the output */
//         if(outputType==MBCS_OUTPUT_2){
//             myValue=MBCS_VALUE_2_FROM_STAGE_2(sharedData->mbcs.fromUnicodeBytes, stage2Entry, c);
//             if(myValue<=0xff) {
//                 length=1;
//             } else {
//                 length=2;
//             }
//         } else /* outputType==MBCS_OUTPUT_3 */ {
//             p=MBCS_POINTER_3_FROM_STAGE_2(sharedData->mbcs.fromUnicodeBytes, stage2Entry, c);
//             myValue=((uint32_t)*p<<16)|((uint32_t)p[1]<<8)|p[2];
//             if(myValue<=0xff) {
//                 length=1;
//             } else if(myValue<=0xffff) {
//                 length=2;
//             } else {
//                 length=3;
//             }
//         }
//         /* is this code point assigned, or do we use fallbacks? */
//         if((stage2Entry&(1<<(16+(c&0xf))))!=0) {
//             /* assigned */
//             *value=myValue;
//             return length;
//         } else if(FROM_U_USE_FALLBACK(useFallback, c) && myValue!=0) {
//             /*
//              * We allow a 0 byte output if the "assigned" bit is set for this entry.
//              * There is no way with this data structure for fallback output
//              * to be a zero byte.
//              */
//             *value=myValue;
//             return -length;
//         }
//     }

//     cx=sharedData->mbcs.extIndexes;
//     if(cx!=NULL) {
//         return ucnv_extSimpleMatchFromU(cx, c, value, useFallback);
//     }

//     /* unassigned */
//     return 0;
// }

// /* This inline function replicates code in _MBCSSingleFromUChar32() function in ucnvmbcs.c
//  * any future change in _MBCSSingleFromUChar32() function should be reflected here.
//  * @param retval pointer to output byte
//  * @return 1 roundtrip byte  0 no mapping  -1 fallback byte
//  */
// static inline int32_t
// MBCS_SINGLE_FROM_UCHAR32(UConverterSharedData* sharedData,
//                                        UChar32 c,
//                                        uint32_t* retval,
//                                        UBool useFallback)
// {
//     const uint16_t *table;
//     int32_t value;
//     /* BMP-only codepages are stored without stage 1 entries for supplementary code points */
//     if(c>=0x10000 && !(sharedData->mbcs.unicodeMask&UCNV_HAS_SUPPLEMENTARY)) {
//         return 0;
//     }
//     /* convert the Unicode code point in c into codepage bytes (same as in _MBCSFromUnicodeWithOffsets) */
//     table=sharedData->mbcs.fromUnicodeTable;
//     /* get the byte for the output */
//     value=MBCS_SINGLE_RESULT_FROM_U(table, (uint16_t *)sharedData->mbcs.fromUnicodeBytes, c);
//     /* is this code point assigned, or do we use fallbacks? */
//     *retval=(uint32_t)(value&0xff);
//     if(value>=0xf00) {
//         return 1;  /* roundtrip */
//     } else if(useFallback ? value>=0x800 : value>=0xc00) {
//         return -1;  /* fallback taken */
//     } else {
//         return 0;  /* no mapping */
//     }
// }

// /*
//  * Check that the result is a 2-byte value with each byte in the range A1..FE
//  * (strict EUC DBCS) before accepting it and subtracting 0x80 from each byte
//  * to move it to the ISO 2022 range 21..7E.
//  * Return 0 if out of range.
//  */
// static inline uint32_t
// _2022FromGR94DBCS(uint32_t value) {
//     if( (uint16_t)(value - 0xa1a1) <= (0xfefe - 0xa1a1) &&
//         (uint8_t)(value - 0xa1) <= (0xfe - 0xa1)
//     ) {
//         return value - 0x8080;  /* shift down to 21..7e byte range */
//     } else {
//         return 0;  /* not valid for ISO 2022 */
//     }
// }

// #if 0 /* 5691: Call sites now check for validity. They can just += 0x8080 after that. */
// /*
//  * This method does the reverse of _2022FromGR94DBCS(). Given the 2022 code point, it returns the
//  * 2 byte value that is in the range A1..FE for each byte. Otherwise it returns the 2022 code point
//  * unchanged.
//  */
// static inline uint32_t
// _2022ToGR94DBCS(uint32_t value) {
//     uint32_t returnValue = value + 0x8080;
//     if( (uint16_t)(returnValue - 0xa1a1) <= (0xfefe - 0xa1a1) &&
//         (uint8_t)(returnValue - 0xa1) <= (0xfe - 0xa1)) {
//         return returnValue;
//     } else {
//         return value;
//     }
// }
// #endif

// #ifdef U_ENABLE_GENERIC_ISO_2022

// /**********************************************************************************
// *  ISO-2022 Converter
// *
// *
// */

// static void U_CALLCONV
// T_UConverter_toUnicode_ISO_2022_OFFSETS_LOGIC(UConverterToUnicodeArgs* args,
//                                                            UErrorCode* err){
//     const char* mySourceLimit, *realSourceLimit;
//     const char* sourceStart;
//     const UChar* myTargetStart;
//     UConverter* saveThis;
//     UConverterDataISO2022* myData;
//     int8_t length;

//     saveThis = args->converter;
//     myData=((UConverterDataISO2022*)(saveThis->extraInfo));

//     realSourceLimit = args->sourceLimit;
//     while (args->source < realSourceLimit) {
//         if(myData->key == 0) { /* are we in the middle of an escape sequence? */
//             /*Find the end of the buffer e.g : Next Escape Seq | end of Buffer*/
//             mySourceLimit = getEndOfBuffer_2022(&(args->source), realSourceLimit, args->flush);

//             if(args->source < mySourceLimit) {
//                 if(myData->currentConverter==NULL) {
//                     myData->currentConverter = ucnv_open("ASCII",err);
//                     if(U_FAILURE(*err)){
//                         return;
//                     }

//                     myData->currentConverter->fromCharErrorBehaviour = UCNV_TO_U_CALLBACK_STOP;
//                     saveThis->mode = UCNV_SO;
//                 }

//                 /* convert to before the ESC or until the end of the buffer */
//                 myData->isFirstBuffer=FALSE;
//                 sourceStart = args->source;
//                 myTargetStart = args->target;
//                 args->converter = myData->currentConverter;
//                 ucnv_toUnicode(args->converter,
//                     &args->target,
//                     args->targetLimit,
//                     &args->source,
//                     mySourceLimit,
//                     args->offsets,
//                     (UBool)(args->flush && mySourceLimit == realSourceLimit),
//                     err);
//                 args->converter = saveThis;

//                 if (*err == U_BUFFER_OVERFLOW_ERROR) {
//                     /* move the overflow buffer */
//                     length = saveThis->UCharErrorBufferLength = myData->currentConverter->UCharErrorBufferLength;
//                     myData->currentConverter->UCharErrorBufferLength = 0;
//                     if(length > 0) {
//                         uprv_memcpy(saveThis->UCharErrorBuffer,
//                                     myData->currentConverter->UCharErrorBuffer,
//                                     length*U_SIZEOF_UCHAR);
//                     }
//                     return;
//                 }

//                 /*
//                  * At least one of:
//                  * -Error while converting
//                  * -Done with entire buffer
//                  * -Need to write offsets or update the current offset
//                  *  (leave that up to the code in ucnv.c)
//                  *
//                  * or else we just stopped at an ESC byte and continue with changeState_2022()
//                  */
//                 if (U_FAILURE(*err) ||
//                     (args->source == realSourceLimit) ||
//                     (args->offsets != NULL && (args->target != myTargetStart || args->source != sourceStart) ||
//                     (mySourceLimit < realSourceLimit && myData->currentConverter->toULength > 0))
//                 ) {
//                     /* copy partial or error input for truncated detection and error handling */
//                     if(U_FAILURE(*err)) {
//                         length = saveThis->invalidCharLength = myData->currentConverter->invalidCharLength;
//                         if(length > 0) {
//                             uprv_memcpy(saveThis->invalidCharBuffer, myData->currentConverter->invalidCharBuffer, length);
//                         }
//                     } else {
//                         length = saveThis->toULength = myData->currentConverter->toULength;
//                         if(length > 0) {
//                             uprv_memcpy(saveThis->toUBytes, myData->currentConverter->toUBytes, length);
//                             if(args->source < mySourceLimit) {
//                                 *err = U_TRUNCATED_CHAR_FOUND; /* truncated input before ESC */
//                             }
//                         }
//                     }
//                     return;
//                 }
//             }
//         }

//         sourceStart = args->source;
//         changeState_2022(args->converter,
//                &(args->source),
//                realSourceLimit,
//                ISO_2022,
//                err);
//         if (U_FAILURE(*err) || (args->source != sourceStart && args->offsets != NULL)) {
//             /* let the ucnv.c code update its current offset */
//             return;
//         }
//     }
// }

// #endif

// /*
//  * To Unicode Callback helper function
//  */
// static void
// toUnicodeCallback(UConverter *cnv,
//                   const uint32_t sourceChar, const uint32_t targetUniChar,
//                   UErrorCode* err){
//     if(sourceChar>0xff){
//         cnv->toUBytes[0] = (uint8_t)(sourceChar>>8);
//         cnv->toUBytes[1] = (uint8_t)sourceChar;
//         cnv->toULength = 2;
//     }
//     else{
//         cnv->toUBytes[0] =(char) sourceChar;
//         cnv->toULength = 1;
//     }

//     if(targetUniChar == (missingCharMarker-1/*0xfffe*/)){
//         *err = U_INVALID_CHAR_FOUND;
//     }
//     else{
//         *err = U_ILLEGAL_CHAR_FOUND;
//     }
// }

// /**************************************ISO-2022-JP*************************************************/

// /************************************** IMPORTANT **************************************************
// * The UConverter_fromUnicode_ISO2022_JP converter does not use ucnv_fromUnicode() functions for SBCS,DBCS and
// * MBCS; instead, the values are obtained directly by calling _MBCSFromUChar32().
// * The converter iterates over each Unicode codepoint
// * to obtain the equivalent codepoints from the codepages supported. Since the source buffer is
// * processed one char at a time it would make sense to reduce the extra processing a canned converter
// * would do as far as possible.
// *
// * If the implementation of these macros or structure of sharedData struct change in the future, make
// * sure that ISO-2022 is also changed.
// ***************************************************************************************************
// */

// /***************************************************************************************************
// * Rules for ISO-2022-jp encoding
// * (i)   Escape sequences must be fully contained within a line they should not
// *       span new lines or CRs
// * (ii)  If the last character on a line is represented by two bytes then an ASCII or
// *       JIS-Roman character escape sequence should follow before the line terminates
// * (iii) If the first character on the line is represented by two bytes then a two
// *       byte character escape sequence should precede it
// * (iv)  If no escape sequence is encountered then the characters are ASCII
// * (v)   Latin(ISO-8859-1) and Greek(ISO-8859-7) characters must be designated to G2,
// *       and invoked with SS2 (ESC N).
// * (vi)  If there is any G0 designation in text, there must be a switch to
// *       ASCII or to JIS X 0201-Roman before a space character (but not
// *       necessarily before "ESC 4/14 2/0" or "ESC N ' '") or control
// *       characters such as tab or CRLF.
// * (vi)  Supported encodings:
// *          ASCII, JISX201, JISX208, JISX212, GB2312, KSC5601, ISO-8859-1,ISO-8859-7
// *
// *  source : RFC-1554
// *
// *          JISX201, JISX208,JISX212 : new .cnv data files created
// *          KSC5601 : alias to ibm-949 mapping table
// *          GB2312 : alias to ibm-1386 mapping table
// *          ISO-8859-1 : Algorithmic implemented as LATIN1 case
// *          ISO-8859-7 : alisas to ibm-9409 mapping table
// */

// /* preference order of JP charsets */
// static const StateEnum jpCharsetPref[]={
//     ASCII,
//     JISX201,
//     ISO8859_1,
//     JISX208,
//     ISO8859_7,
//     JISX212,
//     GB2312,
//     KSC5601,
//     HWKANA_7BIT
// };

// /*
//  * The escape sequences must be in order of the enum constants like JISX201  = 3,
//  * not in order of jpCharsetPref[]!
//  */
// static const char escSeqChars[][6] ={
//     "\x1B\x28\x42",         /* <ESC>(B  ASCII       */
//     "\x1B\x2E\x41",         /* <ESC>.A  ISO-8859-1  */
//     "\x1B\x2E\x46",         /* <ESC>.F  ISO-8859-7  */
//     "\x1B\x28\x4A",         /* <ESC>(J  JISX-201    */
//     "\x1B\x24\x42",         /* <ESC>$B  JISX-208    */
//     "\x1B\x24\x28\x44",     /* <ESC>$(D JISX-212    */
//     "\x1B\x24\x41",         /* <ESC>$A  GB2312      */
//     "\x1B\x24\x28\x43",     /* <ESC>$(C KSC5601     */
//     "\x1B\x28\x49"          /* <ESC>(I  HWKANA_7BIT */

// };
// static  const int8_t escSeqCharsLen[] ={
//     3, /* length of <ESC>(B  ASCII       */
//     3, /* length of <ESC>.A  ISO-8859-1  */
//     3, /* length of <ESC>.F  ISO-8859-7  */
//     3, /* length of <ESC>(J  JISX-201    */
//     3, /* length of <ESC>$B  JISX-208    */
//     4, /* length of <ESC>$(D JISX-212    */
//     3, /* length of <ESC>$A  GB2312      */
//     4, /* length of <ESC>$(C KSC5601     */
//     3  /* length of <ESC>(I  HWKANA_7BIT */
// };

// /*
// * The iteration over various code pages works this way:
// * i)   Get the currentState from myConverterData->currentState
// * ii)  Check if the character is mapped to a valid character in the currentState
// *      Yes ->  a) set the initIterState to currentState
// *       b) remain in this state until an invalid character is found
// *      No  ->  a) go to the next code page and find the character
// * iii) Before changing the state increment the current state check if the current state
// *      is equal to the intitIteration state
// *      Yes ->  A character that cannot be represented in any of the supported encodings
// *       break and return a U_INVALID_CHARACTER error
// *      No  ->  Continue and find the character in next code page
// *
// *
// * TODO: Implement a priority technique where the users are allowed to set the priority of code pages
// */

// /* Map 00..7F to Unicode according to JIS X 0201. */
// static inline uint32_t
// jisx201ToU(uint32_t value) {
//     if(value < 0x5c) {
//         return value;
//     } else if(value == 0x5c) {
//         return 0xa5;
//     } else if(value == 0x7e) {
//         return 0x203e;
//     } else /* value <= 0x7f */ {
//         return value;
//     }
// }

// /* Map Unicode to 00..7F according to JIS X 0201. Return U+FFFE if unmappable. */
// static inline uint32_t
// jisx201FromU(uint32_t value) {
//     if(value<=0x7f) {
//         if(value!=0x5c && value!=0x7e) {
//             return value;
//         }
//     } else if(value==0xa5) {
//         return 0x5c;
//     } else if(value==0x203e) {
//         return 0x7e;
//     }
//     return 0xfffe;
// }

// /*
//  * Take a valid Shift-JIS byte pair, check that it is in the range corresponding
//  * to JIS X 0208, and convert it to a pair of 21..7E bytes.
//  * Return 0 if the byte pair is out of range.
//  */
// static inline uint32_t
// _2022FromSJIS(uint32_t value) {
//     uint8_t trail;

//     if(value > 0xEFFC) {
//         return 0;  /* beyond JIS X 0208 */
//     }

//     trail = (uint8_t)value;

//     value &= 0xff00;  /* lead byte */
//     if(value <= 0x9f00) {
//         value -= 0x7000;
//     } else /* 0xe000 <= value <= 0xef00 */ {
//         value -= 0xb000;
//     }
//     value <<= 1;

//     if(trail <= 0x9e) {
//         value -= 0x100;
//         if(trail <= 0x7e) {
//             value |= trail - 0x1f;
//         } else {
//             value |= trail - 0x20;
//         }
//     } else /* trail <= 0xfc */ {
//         value |= trail - 0x7e;
//     }
//     return value;
// }

// /*
//  * Convert a pair of JIS X 0208 21..7E bytes to Shift-JIS.
//  * If either byte is outside 21..7E make sure that the result is not valid
//  * for Shift-JIS so that the converter catches it.
//  * Some invalid byte values already turn into equally invalid Shift-JIS
//  * byte values and need not be tested explicitly.
//  */
// static inline void
// _2022ToSJIS(uint8_t c1, uint8_t c2, char bytes[2]) {
//     if(c1&1) {
//         ++c1;
//         if(c2 <= 0x5f) {
//             c2 += 0x1f;
//         } else if(c2 <= 0x7e) {
//             c2 += 0x20;
//         } else {
//             c2 = 0;  /* invalid */
//         }
//     } else {
//         if((uint8_t)(c2-0x21) <= ((0x7e)-0x21)) {
//             c2 += 0x7e;
//         } else {
//             c2 = 0;  /* invalid */
//         }
//     }
//     c1 >>= 1;
//     if(c1 <= 0x2f) {
//         c1 += 0x70;
//     } else if(c1 <= 0x3f) {
//         c1 += 0xb0;
//     } else {
//         c1 = 0;  /* invalid */
//     }
//     bytes[0] = (char)c1;
//     bytes[1] = (char)c2;
// }

// /*
//  * JIS X 0208 has fallbacks from Unicode half-width Katakana to full-width (DBCS)
//  * Katakana.
//  * Now that we use a Shift-JIS table for JIS X 0208 we need to hardcode these fallbacks
//  * because Shift-JIS roundtrips half-width Katakana to single bytes.
//  * These were the only fallbacks in ICU's jisx-208.ucm file.
//  */
// static const uint16_t hwkana_fb[HWKANA_END - HWKANA_START + 1] = {
//     0x2123,  /* U+FF61 */
//     0x2156,
//     0x2157,
//     0x2122,
//     0x2126,
//     0x2572,
//     0x2521,
//     0x2523,
//     0x2525,
//     0x2527,
//     0x2529,
//     0x2563,
//     0x2565,
//     0x2567,
//     0x2543,
//     0x213C,  /* U+FF70 */
//     0x2522,
//     0x2524,
//     0x2526,
//     0x2528,
//     0x252A,
//     0x252B,
//     0x252D,
//     0x252F,
//     0x2531,
//     0x2533,
//     0x2535,
//     0x2537,
//     0x2539,
//     0x253B,
//     0x253D,
//     0x253F,  /* U+FF80 */
//     0x2541,
//     0x2544,
//     0x2546,
//     0x2548,
//     0x254A,
//     0x254B,
//     0x254C,
//     0x254D,
//     0x254E,
//     0x254F,
//     0x2552,
//     0x2555,
//     0x2558,
//     0x255B,
//     0x255E,
//     0x255F,  /* U+FF90 */
//     0x2560,
//     0x2561,
//     0x2562,
//     0x2564,
//     0x2566,
//     0x2568,
//     0x2569,
//     0x256A,
//     0x256B,
//     0x256C,
//     0x256D,
//     0x256F,
//     0x2573,
//     0x212B,
//     0x212C   /* U+FF9F */
// };

// static void U_CALLCONV
// UConverter_fromUnicode_ISO_2022_JP_OFFSETS_LOGIC(UConverterFromUnicodeArgs* args, UErrorCode* err) {
//     UConverter *cnv = args->converter;
//     UConverterDataISO2022 *converterData;
//     ISO2022State *pFromU2022State;
//     uint8_t *target = (uint8_t *) args->target;
//     const uint8_t *targetLimit = (const uint8_t *) args->targetLimit;
//     const UChar* source = args->source;
//     const UChar* sourceLimit = args->sourceLimit;
//     int32_t* offsets = args->offsets;
//     UChar32 sourceChar;
//     char buffer[8];
//     int32_t len, outLen;
//     int8_t choices[10];
//     int32_t choiceCount;
//     uint32_t targetValue = 0;
//     UBool useFallback;

//     int32_t i;
//     int8_t cs, g;

//     /* set up the state */
//     converterData     = (UConverterDataISO2022*)cnv->extraInfo;
//     pFromU2022State   = &converterData->fromU2022State;

//     choiceCount = 0;

//     /* check if the last codepoint of previous buffer was a lead surrogate*/
//     if((sourceChar = cnv->fromUChar32)!=0 && target< targetLimit) {
//         goto getTrail;
//     }

//     while(source < sourceLimit) {
//         if(target < targetLimit) {

//             sourceChar  = *(source++);
//             /*check if the char is a First surrogate*/
//             if(U16_IS_SURROGATE(sourceChar)) {
//                 if(U16_IS_SURROGATE_LEAD(sourceChar)) {
// getTrail:
//                     /*look ahead to find the trail surrogate*/
//                     if(source < sourceLimit) {
//                         /* test the following code unit */
//                         UChar trail=(UChar) *source;
//                         if(U16_IS_TRAIL(trail)) {
//                             source++;
//                             sourceChar=U16_GET_SUPPLEMENTARY(sourceChar, trail);
//                             cnv->fromUChar32=0x00;
//                             /* convert this supplementary code point */
//                             /* exit this condition tree */
//                         } else {
//                             /* this is an unmatched lead code unit (1st surrogate) */
//                             /* callback(illegal) */
//                             *err=U_ILLEGAL_CHAR_FOUND;
//                             cnv->fromUChar32=sourceChar;
//                             break;
//                         }
//                     } else {
//                         /* no more input */
//                         cnv->fromUChar32=sourceChar;
//                         break;
//                     }
//                 } else {
//                     /* this is an unmatched trail code unit (2nd surrogate) */
//                     /* callback(illegal) */
//                     *err=U_ILLEGAL_CHAR_FOUND;
//                     cnv->fromUChar32=sourceChar;
//                     break;
//                 }
//             }

//             /* do not convert SO/SI/ESC */
//             if(IS_2022_CONTROL(sourceChar)) {
//                 /* callback(illegal) */
//                 *err=U_ILLEGAL_CHAR_FOUND;
//                 cnv->fromUChar32=sourceChar;
//                 break;
//             }

//             /* do the conversion */

//             if(choiceCount == 0) {
//                 uint16_t csm;

//                 /*
//                  * The csm variable keeps track of which charsets are allowed
//                  * and not used yet while building the choices[].
//                  */
//                 csm = jpCharsetMasks[converterData->version];
//                 choiceCount = 0;

//                 /* JIS7/8: try single-byte half-width Katakana before JISX208 */
//                 if(converterData->version == 3 || converterData->version == 4) {
//                     choices[choiceCount++] = (int8_t)HWKANA_7BIT;
//                 }
//                 /* Do not try single-byte half-width Katakana for other versions. */
//                 csm &= ~CSM(HWKANA_7BIT);

//                 /* try the current G0 charset */
//                 choices[choiceCount++] = cs = pFromU2022State->cs[0];
//                 csm &= ~CSM(cs);

//                 /* try the current G2 charset */
//                 if((cs = pFromU2022State->cs[2]) != 0) {
//                     choices[choiceCount++] = cs;
//                     csm &= ~CSM(cs);
//                 }

//                 /* try all the other possible charsets */
//                 for(i = 0; i < UPRV_LENGTHOF(jpCharsetPref); ++i) {
//                     cs = (int8_t)jpCharsetPref[i];
//                     if(CSM(cs) & csm) {
//                         choices[choiceCount++] = cs;
//                         csm &= ~CSM(cs);
//                     }
//                 }
//             }

//             cs = g = 0;
//             /*
//              * len==0: no mapping found yet
//              * len<0: found a fallback result: continue looking for a roundtrip but no further fallbacks
//              * len>0: found a roundtrip result, done
//              */
//             len = 0;
//             /*
//              * We will turn off useFallback after finding a fallback,
//              * but we still get fallbacks from PUA code points as usual.
//              * Therefore, we will also need to check that we don't overwrite
//              * an early fallback with a later one.
//              */
//             useFallback = cnv->useFallback;

//             for(i = 0; i < choiceCount && len <= 0; ++i) {
//                 uint32_t value;
//                 int32_t len2;
//                 int8_t cs0 = choices[i];
//                 switch(cs0) {
//                 case ASCII:
//                     if(sourceChar <= 0x7f) {
//                         targetValue = (uint32_t)sourceChar;
//                         len = 1;
//                         cs = cs0;
//                         g = 0;
//                     }
//                     break;
//                 case ISO8859_1:
//                     if(GR96_START <= sourceChar && sourceChar <= GR96_END) {
//                         targetValue = (uint32_t)sourceChar - 0x80;
//                         len = 1;
//                         cs = cs0;
//                         g = 2;
//                     }
//                     break;
//                 case HWKANA_7BIT:
//                     if((uint32_t)(sourceChar - HWKANA_START) <= (HWKANA_END - HWKANA_START)) {
//                         if(converterData->version==3) {
//                             /* JIS7: use G1 (SO) */
//                             /* Shift U+FF61..U+FF9F to bytes 21..5F. */
//                             targetValue = (uint32_t)(sourceChar - (HWKANA_START - 0x21));
//                             len = 1;
//                             pFromU2022State->cs[1] = cs = cs0; /* do not output an escape sequence */
//                             g = 1;
//                         } else if(converterData->version==4) {
//                             /* JIS8: use 8-bit bytes with any single-byte charset, see escape sequence output below */
//                             /* Shift U+FF61..U+FF9F to bytes A1..DF. */
//                             targetValue = (uint32_t)(sourceChar - (HWKANA_START - 0xa1));
//                             len = 1;

//                             cs = pFromU2022State->cs[0];
//                             if(IS_JP_DBCS(cs)) {
//                                 /* switch from a DBCS charset to JISX201 */
//                                 cs = (int8_t)JISX201;
//                             }
//                             /* else stay in the current G0 charset */
//                             g = 0;
//                         }
//                         /* else do not use HWKANA_7BIT with other versions */
//                     }
//                     break;
//                 case JISX201:
//                     /* G0 SBCS */
//                     value = jisx201FromU(sourceChar);
//                     if(value <= 0x7f) {
//                         targetValue = value;
//                         len = 1;
//                         cs = cs0;
//                         g = 0;
//                         useFallback = FALSE;
//                     }
//                     break;
//                 case JISX208:
//                     /* G0 DBCS from Shift-JIS table */
//                     len2 = MBCS_FROM_UCHAR32_ISO2022(
//                                 converterData->myConverterArray[cs0],
//                                 sourceChar, &value,
//                                 useFallback, MBCS_OUTPUT_2);
//                     if(len2 == 2 || (len2 == -2 && len == 0)) {  /* only accept DBCS: abs(len)==2 */
//                         value = _2022FromSJIS(value);
//                         if(value != 0) {
//                             targetValue = value;
//                             len = len2;
//                             cs = cs0;
//                             g = 0;
//                             useFallback = FALSE;
//                         }
//                     } else if(len == 0 && useFallback &&
//                               (uint32_t)(sourceChar - HWKANA_START) <= (HWKANA_END - HWKANA_START)) {
//                         targetValue = hwkana_fb[sourceChar - HWKANA_START];
//                         len = -2;
//                         cs = cs0;
//                         g = 0;
//                         useFallback = FALSE;
//                     }
//                     break;
//                 case ISO8859_7:
//                     /* G0 SBCS forced to 7-bit output */
//                     len2 = MBCS_SINGLE_FROM_UCHAR32(
//                                 converterData->myConverterArray[cs0],
//                                 sourceChar, &value,
//                                 useFallback);
//                     if(len2 != 0 && !(len2 < 0 && len != 0) && GR96_START <= value && value <= GR96_END) {
//                         targetValue = value - 0x80;
//                         len = len2;
//                         cs = cs0;
//                         g = 2;
//                         useFallback = FALSE;
//                     }
//                     break;
//                 default:
//                     /* G0 DBCS */
//                     len2 = MBCS_FROM_UCHAR32_ISO2022(
//                                 converterData->myConverterArray[cs0],
//                                 sourceChar, &value,
//                                 useFallback, MBCS_OUTPUT_2);
//                     if(len2 == 2 || (len2 == -2 && len == 0)) {  /* only accept DBCS: abs(len)==2 */
//                         if(cs0 == KSC5601) {
//                             /*
//                              * Check for valid bytes for the encoding scheme.
//                              * This is necessary because the sub-converter (windows-949)
//                              * has a broader encoding scheme than is valid for 2022.
//                              */
//                             value = _2022FromGR94DBCS(value);
//                             if(value == 0) {
//                                 break;
//                             }
//                         }
//                         targetValue = value;
//                         len = len2;
//                         cs = cs0;
//                         g = 0;
//                         useFallback = FALSE;
//                     }
//                     break;
//                 }
//             }

//             if(len != 0) {
//                 if(len < 0) {
//                     len = -len;  /* fallback */
//                 }
//                 outLen = 0; /* count output bytes */

//                 /* write SI if necessary (only for JIS7) */
//                 if(pFromU2022State->g == 1 && g == 0) {
//                     buffer[outLen++] = UCNV_SI;
//                     pFromU2022State->g = 0;
//                 }

//                 /* write the designation sequence if necessary */
//                 if(cs != pFromU2022State->cs[g]) {
//                     int32_t escLen = escSeqCharsLen[cs];
//                     uprv_memcpy(buffer + outLen, escSeqChars[cs], escLen);
//                     outLen += escLen;
//                     pFromU2022State->cs[g] = cs;

//                     /* invalidate the choices[] */
//                     choiceCount = 0;
//                 }

//                 /* write the shift sequence if necessary */
//                 if(g != pFromU2022State->g) {
//                     switch(g) {
//                     /* case 0 handled before writing escapes */
//                     case 1:
//                         buffer[outLen++] = UCNV_SO;
//                         pFromU2022State->g = 1;
//                         break;
//                     default: /* case 2 */
//                         buffer[outLen++] = 0x1b;
//                         buffer[outLen++] = 0x4e;
//                         break;
//                     /* no case 3: no SS3 in ISO-2022-JP-x */
//                     }
//                 }

//                 /* write the output bytes */
//                 if(len == 1) {
//                     buffer[outLen++] = (char)targetValue;
//                 } else /* len == 2 */ {
//                     buffer[outLen++] = (char)(targetValue >> 8);
//                     buffer[outLen++] = (char)targetValue;
//                 }
//             } else {
//                 /*
//                  * if we cannot find the character after checking all codepages
//                  * then this is an error
//                  */
//                 *err = U_INVALID_CHAR_FOUND;
//                 cnv->fromUChar32=sourceChar;
//                 break;
//             }

//             if(sourceChar == CR || sourceChar == LF) {
//                 /* reset the G2 state at the end of a line (conversion got us into ASCII or JISX201 already) */
//                 pFromU2022State->cs[2] = 0;
//                 choiceCount = 0;
//             }

//             /* output outLen>0 bytes in buffer[] */
//             if(outLen == 1) {
//                 *target++ = buffer[0];
//                 if(offsets) {
//                     *offsets++ = (int32_t)(source - args->source - 1); /* -1: known to be ASCII */
//                 }
//             } else if(outLen == 2 && (target + 2) <= targetLimit) {
//                 *target++ = buffer[0];
//                 *target++ = buffer[1];
//                 if(offsets) {
//                     int32_t sourceIndex = (int32_t)(source - args->source - U16_LENGTH(sourceChar));
//                     *offsets++ = sourceIndex;
//                     *offsets++ = sourceIndex;
//                 }
//             } else {
//                 fromUWriteUInt8(
//                     cnv,
//                     buffer, outLen,
//                     &target, (const char *)targetLimit,
//                     &offsets, (int32_t)(source - args->source - U16_LENGTH(sourceChar)),
//                     err);
//                 if(U_FAILURE(*err)) {
//                     break;
//                 }
//             }
//         } /* end if(myTargetIndex<myTargetLength) */
//         else{
//             *err =U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }

//     }/* end while(mySourceIndex<mySourceLength) */

//     /*
//      * the end of the input stream and detection of truncated input
//      * are handled by the framework, but for ISO-2022-JP conversion
//      * we need to be in ASCII mode at the very end
//      *
//      * conditions:
//      *   successful
//      *   in SO mode or not in ASCII mode
//      *   end of input and no truncated input
//      */
//     if( U_SUCCESS(*err) &&
//         (pFromU2022State->g!=0 || pFromU2022State->cs[0]!=ASCII) &&
//         args->flush && source>=sourceLimit && cnv->fromUChar32==0
//     ) {
//         int32_t sourceIndex;

//         outLen = 0;

//         if(pFromU2022State->g != 0) {
//             buffer[outLen++] = UCNV_SI;
//             pFromU2022State->g = 0;
//         }

//         if(pFromU2022State->cs[0] != ASCII) {
//             int32_t escLen = escSeqCharsLen[ASCII];
//             uprv_memcpy(buffer + outLen, escSeqChars[ASCII], escLen);
//             outLen += escLen;
//             pFromU2022State->cs[0] = (int8_t)ASCII;
//         }

//         /* get the source index of the last input character */
//         /*
//          * TODO this would be simpler and more reliable if we used a pair
//          * of sourceIndex/prevSourceIndex like in ucnvmbcs.c
//          * so that we could simply use the prevSourceIndex here;
//          * this code gives an incorrect result for the rare case of an unmatched
//          * trail surrogate that is alone in the last buffer of the text stream
//          */
//         sourceIndex=(int32_t)(source-args->source);
//         if(sourceIndex>0) {
//             --sourceIndex;
//             if( U16_IS_TRAIL(args->source[sourceIndex]) &&
//                 (sourceIndex==0 || U16_IS_LEAD(args->source[sourceIndex-1]))
//             ) {
//                 --sourceIndex;
//             }
//         } else {
//             sourceIndex=-1;
//         }

//         fromUWriteUInt8(
//             cnv,
//             buffer, outLen,
//             &target, (const char *)targetLimit,
//             &offsets, sourceIndex,
//             err);
//     }

//     /*save the state and return */
//     args->source = source;
//     args->target = (char*)target;
// }

// /*************** to unicode *******************/

// static void U_CALLCONV
// UConverter_toUnicode_ISO_2022_JP_OFFSETS_LOGIC(UConverterToUnicodeArgs *args,
//                                                UErrorCode* err){
//     char tempBuf[2];
//     const char *mySource = (char *) args->source;
//     UChar *myTarget = args->target;
//     const char *mySourceLimit = args->sourceLimit;
//     uint32_t targetUniChar = 0x0000;
//     uint32_t mySourceChar = 0x0000;
//     uint32_t tmpSourceChar = 0x0000;
//     UConverterDataISO2022* myData;
//     ISO2022State *pToU2022State;
//     StateEnum cs;

//     myData=(UConverterDataISO2022*)(args->converter->extraInfo);
//     pToU2022State = &myData->toU2022State;

//     if(myData->key != 0) {
//         /* continue with a partial escape sequence */
//         goto escape;
//     } else if(args->converter->toULength == 1 && mySource < mySourceLimit && myTarget < args->targetLimit) {
//         /* continue with a partial double-byte character */
//         mySourceChar = args->converter->toUBytes[0];
//         args->converter->toULength = 0;
//         cs = (StateEnum)pToU2022State->cs[pToU2022State->g];
//         targetUniChar = missingCharMarker;
//         goto getTrailByte;
//     }

//     while(mySource < mySourceLimit){

//         targetUniChar =missingCharMarker;

//         if(myTarget < args->targetLimit){

//             mySourceChar= (unsigned char) *mySource++;

//             switch(mySourceChar) {
//             case UCNV_SI:
//                 if(myData->version==3) {
//                     pToU2022State->g=0;
//                     continue;
//                 } else {
//                     /* only JIS7 uses SI/SO, not ISO-2022-JP-x */
//                     myData->isEmptySegment = FALSE;	/* reset this, we have a different error */
//                     break;
//                 }

//             case UCNV_SO:
//                 if(myData->version==3) {
//                     /* JIS7: switch to G1 half-width Katakana */
//                     pToU2022State->cs[1] = (int8_t)HWKANA_7BIT;
//                     pToU2022State->g=1;
//                     continue;
//                 } else {
//                     /* only JIS7 uses SI/SO, not ISO-2022-JP-x */
//                     myData->isEmptySegment = FALSE;	/* reset this, we have a different error */
//                     break;
//                 }

//             case ESC_2022:
//                 mySource--;
// escape:
//                 {
//                     const char * mySourceBefore = mySource;
//                     int8_t toULengthBefore = args->converter->toULength;

//                     changeState_2022(args->converter,&(mySource),
//                         mySourceLimit, ISO_2022_JP,err);

//                     /* If in ISO-2022-JP only and we successully completed an escape sequence, but previous segment was empty, create an error */
//                     if(myData->version==0 && myData->key==0 && U_SUCCESS(*err) && myData->isEmptySegment) {
//                         *err = U_ILLEGAL_ESCAPE_SEQUENCE;
//                         args->converter->toUCallbackReason = UCNV_IRREGULAR;
//                         args->converter->toULength = (int8_t)(toULengthBefore + (mySource - mySourceBefore));
//                     }
//                 }

//                 /* invalid or illegal escape sequence */
//                 if(U_FAILURE(*err)){
//                     args->target = myTarget;
//                     args->source = mySource;
//                     myData->isEmptySegment = FALSE;	/* Reset to avoid future spurious errors */
//                     return;
//                 }
//                 /* If we successfully completed an escape sequence, we begin a new segment, empty so far */
//                 if(myData->key==0) {
//                     myData->isEmptySegment = TRUE;
//                 }
//                 continue;

//             /* ISO-2022-JP does not use single-byte (C1) SS2 and SS3 */

//             case CR:
//             case LF:
//                 /* automatically reset to single-byte mode */
//                 if((StateEnum)pToU2022State->cs[0] != ASCII && (StateEnum)pToU2022State->cs[0] != JISX201) {
//                     pToU2022State->cs[0] = (int8_t)ASCII;
//                 }
//                 pToU2022State->cs[2] = 0;
//                 pToU2022State->g = 0;
//                 U_FALLTHROUGH;
//             default:
//                 /* convert one or two bytes */
//                 myData->isEmptySegment = FALSE;
//                 cs = (StateEnum)pToU2022State->cs[pToU2022State->g];
//                 if( (uint8_t)(mySourceChar - 0xa1) <= (0xdf - 0xa1) && myData->version==4 &&
//                     !IS_JP_DBCS(cs)
//                 ) {
//                     /* 8-bit halfwidth katakana in any single-byte mode for JIS8 */
//                     targetUniChar = mySourceChar + (HWKANA_START - 0xa1);

//                     /* return from a single-shift state to the previous one */
//                     if(pToU2022State->g >= 2) {
//                         pToU2022State->g=pToU2022State->prevG;
//                     }
//                 } else switch(cs) {
//                 case ASCII:
//                     if(mySourceChar <= 0x7f) {
//                         targetUniChar = mySourceChar;
//                     }
//                     break;
//                 case ISO8859_1:
//                     if(mySourceChar <= 0x7f) {
//                         targetUniChar = mySourceChar + 0x80;
//                     }
//                     /* return from a single-shift state to the previous one */
//                     pToU2022State->g=pToU2022State->prevG;
//                     break;
//                 case ISO8859_7:
//                     if(mySourceChar <= 0x7f) {
//                         /* convert mySourceChar+0x80 to use a normal 8-bit table */
//                         targetUniChar =
//                             _MBCS_SINGLE_SIMPLE_GET_NEXT_BMP(
//                                 myData->myConverterArray[cs],
//                                 mySourceChar + 0x80);
//                     }
//                     /* return from a single-shift state to the previous one */
//                     pToU2022State->g=pToU2022State->prevG;
//                     break;
//                 case JISX201:
//                     if(mySourceChar <= 0x7f) {
//                         targetUniChar = jisx201ToU(mySourceChar);
//                     }
//                     break;
//                 case HWKANA_7BIT:
//                     if((uint8_t)(mySourceChar - 0x21) <= (0x5f - 0x21)) {
//                         /* 7-bit halfwidth Katakana */
//                         targetUniChar = mySourceChar + (HWKANA_START - 0x21);
//                     }
//                     break;
//                 default:
//                     /* G0 DBCS */
//                     if(mySource < mySourceLimit) {
//                         int leadIsOk, trailIsOk;
//                         uint8_t trailByte;
// getTrailByte:
//                         trailByte = (uint8_t)*mySource;
//                         /*
//                          * Ticket 5691: consistent illegal sequences:
//                          * - We include at least the first byte in the illegal sequence.
//                          * - If any of the non-initial bytes could be the start of a character,
//                          *   we stop the illegal sequence before the first one of those.
//                          *
//                          * In ISO-2022 DBCS, if the second byte is in the 21..7e range or is
//                          * an ESC/SO/SI, we report only the first byte as the illegal sequence.
//                          * Otherwise we convert or report the pair of bytes.
//                          */
//                         leadIsOk = (uint8_t)(mySourceChar - 0x21) <= (0x7e - 0x21);
//                         trailIsOk = (uint8_t)(trailByte - 0x21) <= (0x7e - 0x21);
//                         if (leadIsOk && trailIsOk) {
//                             ++mySource;
//                             tmpSourceChar = (mySourceChar << 8) | trailByte;
//                             if(cs == JISX208) {
//                                 _2022ToSJIS((uint8_t)mySourceChar, trailByte, tempBuf);
//                                 mySourceChar = tmpSourceChar;
//                             } else {
//                                 /* Copy before we modify tmpSourceChar so toUnicodeCallback() sees the correct bytes. */
//                                 mySourceChar = tmpSourceChar;
//                                 if (cs == KSC5601) {
//                                     tmpSourceChar += 0x8080;  /* = _2022ToGR94DBCS(tmpSourceChar) */
//                                 }
//                                 tempBuf[0] = (char)(tmpSourceChar >> 8);
//                                 tempBuf[1] = (char)(tmpSourceChar);
//                             }
//                             targetUniChar = ucnv_MBCSSimpleGetNextUChar(myData->myConverterArray[cs], tempBuf, 2, FALSE);
//                         } else if (!(trailIsOk || IS_2022_CONTROL(trailByte))) {
//                             /* report a pair of illegal bytes if the second byte is not a DBCS starter */
//                             ++mySource;
//                             /* add another bit so that the code below writes 2 bytes in case of error */
//                             mySourceChar = 0x10000 | (mySourceChar << 8) | trailByte;
//                         }
//                     } else {
//                         args->converter->toUBytes[0] = (uint8_t)mySourceChar;
//                         args->converter->toULength = 1;
//                         goto endloop;
//                     }
//                 }  /* End of inner switch */
//                 break;
//             }  /* End of outer switch */
//             if(targetUniChar < (missingCharMarker-1/*0xfffe*/)){
//                 if(args->offsets){
//                     args->offsets[myTarget - args->target] = (int32_t)(mySource - args->source - (mySourceChar <= 0xff ? 1 : 2));
//                 }
//                 *(myTarget++)=(UChar)targetUniChar;
//             }
//             else if(targetUniChar > missingCharMarker){
//                 /* disassemble the surrogate pair and write to output*/
//                 targetUniChar-=0x0010000;
//                 *myTarget = (UChar)(0xd800+(UChar)(targetUniChar>>10));
//                 if(args->offsets){
//                     args->offsets[myTarget - args->target] = (int32_t)(mySource - args->source - (mySourceChar <= 0xff ? 1 : 2));
//                 }
//                 ++myTarget;
//                 if(myTarget< args->targetLimit){
//                     *myTarget = (UChar)(0xdc00+(UChar)(targetUniChar&0x3ff));
//                     if(args->offsets){
//                         args->offsets[myTarget - args->target] = (int32_t)(mySource - args->source - (mySourceChar <= 0xff ? 1 : 2));
//                     }
//                     ++myTarget;
//                 }else{
//                     args->converter->UCharErrorBuffer[args->converter->UCharErrorBufferLength++]=
//                                     (UChar)(0xdc00+(UChar)(targetUniChar&0x3ff));
//                 }

//             }
//             else{
//                 /* Call the callback function*/
//                 toUnicodeCallback(args->converter,mySourceChar,targetUniChar,err);
//                 break;
//             }
//         }
//         else{    /* goes with "if(myTarget < args->targetLimit)"  way up near top of function */
//             *err =U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }
//     }
// endloop:
//     args->target = myTarget;
//     args->source = mySource;
// }


// #if !UCONFIG_ONLY_HTML_CONVERSION
// /***************************************************************
// *   Rules for ISO-2022-KR encoding
// *   i) The KSC5601 designator sequence should appear only once in a file,
// *      at the begining of a line before any KSC5601 characters. This usually
// *      means that it appears by itself on the first line of the file
// *  ii) There are only 2 shifting sequences SO to shift into double byte mode
// *      and SI to shift into single byte mode
// */
// static void U_CALLCONV
// UConverter_fromUnicode_ISO_2022_KR_OFFSETS_LOGIC_IBM(UConverterFromUnicodeArgs* args, UErrorCode* err){

//     UConverter* saveConv = args->converter;
//     UConverterDataISO2022 *myConverterData=(UConverterDataISO2022*)saveConv->extraInfo;
//     args->converter=myConverterData->currentConverter;

//     myConverterData->currentConverter->fromUChar32 = saveConv->fromUChar32;
//     ucnv_MBCSFromUnicodeWithOffsets(args,err);
//     saveConv->fromUChar32 = myConverterData->currentConverter->fromUChar32;

//     if(*err == U_BUFFER_OVERFLOW_ERROR) {
//         if(myConverterData->currentConverter->charErrorBufferLength > 0) {
//             uprv_memcpy(
//                 saveConv->charErrorBuffer,
//                 myConverterData->currentConverter->charErrorBuffer,
//                 myConverterData->currentConverter->charErrorBufferLength);
//         }
//         saveConv->charErrorBufferLength = myConverterData->currentConverter->charErrorBufferLength;
//         myConverterData->currentConverter->charErrorBufferLength = 0;
//     }
//     args->converter=saveConv;
// }

// static void U_CALLCONV
// UConverter_fromUnicode_ISO_2022_KR_OFFSETS_LOGIC(UConverterFromUnicodeArgs* args, UErrorCode* err){

//     const UChar *source = args->source;
//     const UChar *sourceLimit = args->sourceLimit;
//     unsigned char *target = (unsigned char *) args->target;
//     unsigned char *targetLimit = (unsigned char *) args->targetLimit;
//     int32_t* offsets = args->offsets;
//     uint32_t targetByteUnit = 0x0000;
//     UChar32 sourceChar = 0x0000;
//     UBool isTargetByteDBCS;
//     UBool oldIsTargetByteDBCS;
//     UConverterDataISO2022 *converterData;
//     UConverterSharedData* sharedData;
//     UBool useFallback;
//     int32_t length =0;

//     converterData=(UConverterDataISO2022*)args->converter->extraInfo;
//     /* if the version is 1 then the user is requesting
//      * conversion with ibm-25546 pass the arguments to
//      * MBCS converter and return
//      */
//     if(converterData->version==1){
//         UConverter_fromUnicode_ISO_2022_KR_OFFSETS_LOGIC_IBM(args,err);
//         return;
//     }

//     /* initialize data */
//     sharedData = converterData->currentConverter->sharedData;
//     useFallback = args->converter->useFallback;
//     isTargetByteDBCS=(UBool)args->converter->fromUnicodeStatus;
//     oldIsTargetByteDBCS = isTargetByteDBCS;

//     isTargetByteDBCS   = (UBool) args->converter->fromUnicodeStatus;
//     if((sourceChar = args->converter->fromUChar32)!=0 && target <targetLimit) {
//         goto getTrail;
//     }
//     while(source < sourceLimit){

//         targetByteUnit = missingCharMarker;

//         if(target < (unsigned char*) args->targetLimit){
//             sourceChar = *source++;

//             /* do not convert SO/SI/ESC */
//             if(IS_2022_CONTROL(sourceChar)) {
//                 /* callback(illegal) */
//                 *err=U_ILLEGAL_CHAR_FOUND;
//                 args->converter->fromUChar32=sourceChar;
//                 break;
//             }

//             length = MBCS_FROM_UCHAR32_ISO2022(sharedData,sourceChar,&targetByteUnit,useFallback,MBCS_OUTPUT_2);
//             if(length < 0) {
//                 length = -length;  /* fallback */
//             }
//             /* only DBCS or SBCS characters are expected*/
//             /* DB characters with high bit set to 1 are expected */
//             if( length > 2 || length==0 ||
//                 (length == 1 && targetByteUnit > 0x7f) ||
//                 (length == 2 &&
//                     ((uint16_t)(targetByteUnit - 0xa1a1) > (0xfefe - 0xa1a1) ||
//                     (uint8_t)(targetByteUnit - 0xa1) > (0xfe - 0xa1)))
//             ) {
//                 targetByteUnit=missingCharMarker;
//             }
//             if (targetByteUnit != missingCharMarker){

//                 oldIsTargetByteDBCS = isTargetByteDBCS;
//                 isTargetByteDBCS = (UBool)(targetByteUnit>0x00FF);
//                   /* append the shift sequence */
//                 if (oldIsTargetByteDBCS != isTargetByteDBCS ){

//                     if (isTargetByteDBCS)
//                         *target++ = UCNV_SO;
//                     else
//                         *target++ = UCNV_SI;
//                     if(offsets)
//                         *(offsets++) = (int32_t)(source - args->source-1);
//                 }
//                 /* write the targetUniChar  to target */
//                 if(targetByteUnit <= 0x00FF){
//                     if( target < targetLimit){
//                         *(target++) = (unsigned char) targetByteUnit;
//                         if(offsets){
//                             *(offsets++) = (int32_t)(source - args->source-1);
//                         }

//                     }else{
//                         args->converter->charErrorBuffer[args->converter->charErrorBufferLength++] = (unsigned char) (targetByteUnit);
//                         *err = U_BUFFER_OVERFLOW_ERROR;
//                     }
//                 }else{
//                     if(target < targetLimit){
//                         *(target++) =(unsigned char) ((targetByteUnit>>8) -0x80);
//                         if(offsets){
//                             *(offsets++) = (int32_t)(source - args->source-1);
//                         }
//                         if(target < targetLimit){
//                             *(target++) =(unsigned char) (targetByteUnit -0x80);
//                             if(offsets){
//                                 *(offsets++) = (int32_t)(source - args->source-1);
//                             }
//                         }else{
//                             args->converter->charErrorBuffer[args->converter->charErrorBufferLength++] = (unsigned char) (targetByteUnit -0x80);
//                             *err = U_BUFFER_OVERFLOW_ERROR;
//                         }
//                     }else{
//                         args->converter->charErrorBuffer[args->converter->charErrorBufferLength++] = (unsigned char) ((targetByteUnit>>8) -0x80);
//                         args->converter->charErrorBuffer[args->converter->charErrorBufferLength++] = (unsigned char) (targetByteUnit-0x80);
//                         *err = U_BUFFER_OVERFLOW_ERROR;
//                     }
//                 }

//             }
//             else{
//                 /* oops.. the code point is unassingned
//                  * set the error and reason
//                  */

//                 /*check if the char is a First surrogate*/
//                 if(U16_IS_SURROGATE(sourceChar)) {
//                     if(U16_IS_SURROGATE_LEAD(sourceChar)) {
// getTrail:
//                         /*look ahead to find the trail surrogate*/
//                         if(source <  sourceLimit) {
//                             /* test the following code unit */
//                             UChar trail=(UChar) *source;
//                             if(U16_IS_TRAIL(trail)) {
//                                 source++;
//                                 sourceChar=U16_GET_SUPPLEMENTARY(sourceChar, trail);
//                                 *err = U_INVALID_CHAR_FOUND;
//                                 /* convert this surrogate code point */
//                                 /* exit this condition tree */
//                             } else {
//                                 /* this is an unmatched lead code unit (1st surrogate) */
//                                 /* callback(illegal) */
//                                 *err=U_ILLEGAL_CHAR_FOUND;
//                             }
//                         } else {
//                             /* no more input */
//                             *err = U_ZERO_ERROR;
//                         }
//                     } else {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         *err=U_ILLEGAL_CHAR_FOUND;
//                     }
//                 } else {
//                     /* callback(unassigned) for a BMP code point */
//                     *err = U_INVALID_CHAR_FOUND;
//                 }

//                 args->converter->fromUChar32=sourceChar;
//                 break;
//             }
//         } /* end if(myTargetIndex<myTargetLength) */
//         else{
//             *err =U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }

//     }/* end while(mySourceIndex<mySourceLength) */

//     /*
//      * the end of the input stream and detection of truncated input
//      * are handled by the framework, but for ISO-2022-KR conversion
//      * we need to be in ASCII mode at the very end
//      *
//      * conditions:
//      *   successful
//      *   not in ASCII mode
//      *   end of input and no truncated input
//      */
//     if( U_SUCCESS(*err) &&
//         isTargetByteDBCS &&
//         args->flush && source>=sourceLimit && args->converter->fromUChar32==0
//     ) {
//         int32_t sourceIndex;

//         /* we are switching to ASCII */
//         isTargetByteDBCS=FALSE;

//         /* get the source index of the last input character */
//         /*
//          * TODO this would be simpler and more reliable if we used a pair
//          * of sourceIndex/prevSourceIndex like in ucnvmbcs.c
//          * so that we could simply use the prevSourceIndex here;
//          * this code gives an incorrect result for the rare case of an unmatched
//          * trail surrogate that is alone in the last buffer of the text stream
//          */
//         sourceIndex=(int32_t)(source-args->source);
//         if(sourceIndex>0) {
//             --sourceIndex;
//             if( U16_IS_TRAIL(args->source[sourceIndex]) &&
//                 (sourceIndex==0 || U16_IS_LEAD(args->source[sourceIndex-1]))
//             ) {
//                 --sourceIndex;
//             }
//         } else {
//             sourceIndex=-1;
//         }

//         fromUWriteUInt8(
//             args->converter,
//             SHIFT_IN_STR, 1,
//             &target, (const char *)targetLimit,
//             &offsets, sourceIndex,
//             err);
//     }

//     /*save the state and return */
//     args->source = source;
//     args->target = (char*)target;
//     args->converter->fromUnicodeStatus = (uint32_t)isTargetByteDBCS;
// }

// /************************ To Unicode ***************************************/

// static void U_CALLCONV
// UConverter_toUnicode_ISO_2022_KR_OFFSETS_LOGIC_IBM(UConverterToUnicodeArgs *args,
//                                                             UErrorCode* err){
//     char const* sourceStart;
//     UConverterDataISO2022* myData=(UConverterDataISO2022*)(args->converter->extraInfo);

//     UConverterToUnicodeArgs subArgs;
//     int32_t minArgsSize;

//     /* set up the subconverter arguments */
//     if(args->size<sizeof(UConverterToUnicodeArgs)) {
//         minArgsSize = args->size;
//     } else {
//         minArgsSize = (int32_t)sizeof(UConverterToUnicodeArgs);
//     }

//     uprv_memcpy(&subArgs, args, minArgsSize);
//     subArgs.size = (uint16_t)minArgsSize;
//     subArgs.converter = myData->currentConverter;

//     /* remember the original start of the input for offsets */
//     sourceStart = args->source;

//     if(myData->key != 0) {
//         /* continue with a partial escape sequence */
//         goto escape;
//     }

//     while(U_SUCCESS(*err) && args->source < args->sourceLimit) {
//         /*Find the end of the buffer e.g : Next Escape Seq | end of Buffer*/
//         subArgs.source = args->source;
//         subArgs.sourceLimit = getEndOfBuffer_2022(&(args->source), args->sourceLimit, args->flush);
//         if(subArgs.source != subArgs.sourceLimit) {
//             /*
//              * get the current partial byte sequence
//              *
//              * it needs to be moved between the public and the subconverter
//              * so that the conversion framework, which only sees the public
//              * converter, can handle truncated and illegal input etc.
//              */
//             if(args->converter->toULength > 0) {
//                 uprv_memcpy(subArgs.converter->toUBytes, args->converter->toUBytes, args->converter->toULength);
//             }
//             subArgs.converter->toULength = args->converter->toULength;

//             /*
//              * Convert up to the end of the input, or to before the next escape character.
//              * Does not handle conversion extensions because the preToU[] state etc.
//              * is not copied.
//              */
//             ucnv_MBCSToUnicodeWithOffsets(&subArgs, err);

//             if(args->offsets != NULL && sourceStart != args->source) {
//                 /* update offsets to base them on the actual start of the input */
//                 int32_t *offsets = args->offsets;
//                 UChar *target = args->target;
//                 int32_t delta = (int32_t)(args->source - sourceStart);
//                 while(target < subArgs.target) {
//                     if(*offsets >= 0) {
//                         *offsets += delta;
//                     }
//                     ++offsets;
//                     ++target;
//                 }
//             }
//             args->source = subArgs.source;
//             args->target = subArgs.target;
//             args->offsets = subArgs.offsets;

//             /* copy input/error/overflow buffers */
//             if(subArgs.converter->toULength > 0) {
//                 uprv_memcpy(args->converter->toUBytes, subArgs.converter->toUBytes, subArgs.converter->toULength);
//             }
//             args->converter->toULength = subArgs.converter->toULength;

//             if(*err == U_BUFFER_OVERFLOW_ERROR) {
//                 if(subArgs.converter->UCharErrorBufferLength > 0) {
//                     uprv_memcpy(args->converter->UCharErrorBuffer, subArgs.converter->UCharErrorBuffer,
//                                 subArgs.converter->UCharErrorBufferLength);
//                 }
//                 args->converter->UCharErrorBufferLength=subArgs.converter->UCharErrorBufferLength;
//                 subArgs.converter->UCharErrorBufferLength = 0;
//             }
//         }

//         if (U_FAILURE(*err) || (args->source == args->sourceLimit)) {
//             return;
//         }

// escape:
//         changeState_2022(args->converter,
//                &(args->source),
//                args->sourceLimit,
//                ISO_2022_KR,
//                err);
//     }
// }

// static void U_CALLCONV
// UConverter_toUnicode_ISO_2022_KR_OFFSETS_LOGIC(UConverterToUnicodeArgs *args,
//                                                             UErrorCode* err){
//     char tempBuf[2];
//     const char *mySource = ( char *) args->source;
//     UChar *myTarget = args->target;
//     const char *mySourceLimit = args->sourceLimit;
//     UChar32 targetUniChar = 0x0000;
//     UChar mySourceChar = 0x0000;
//     UConverterDataISO2022* myData;
//     UConverterSharedData* sharedData ;
//     UBool useFallback;

//     myData=(UConverterDataISO2022*)(args->converter->extraInfo);
//     if(myData->version==1){
//         UConverter_toUnicode_ISO_2022_KR_OFFSETS_LOGIC_IBM(args,err);
//         return;
//     }

//     /* initialize state */
//     sharedData = myData->currentConverter->sharedData;
//     useFallback = args->converter->useFallback;

//     if(myData->key != 0) {
//         /* continue with a partial escape sequence */
//         goto escape;
//     } else if(args->converter->toULength == 1 && mySource < mySourceLimit && myTarget < args->targetLimit) {
//         /* continue with a partial double-byte character */
//         mySourceChar = args->converter->toUBytes[0];
//         args->converter->toULength = 0;
//         goto getTrailByte;
//     }

//     while(mySource< mySourceLimit){

//         if(myTarget < args->targetLimit){

//             mySourceChar= (unsigned char) *mySource++;

//             if(mySourceChar==UCNV_SI){
//                 myData->toU2022State.g = 0;
//                 if (myData->isEmptySegment) {
//                     myData->isEmptySegment = FALSE;	/* we are handling it, reset to avoid future spurious errors */
//                     *err = U_ILLEGAL_ESCAPE_SEQUENCE;
//                     args->converter->toUCallbackReason = UCNV_IRREGULAR;
//                     args->converter->toUBytes[0] = (uint8_t)mySourceChar;
//                     args->converter->toULength = 1;
//                     args->target = myTarget;
//                     args->source = mySource;
//                     return;
//                 }
//                 /*consume the source */
//                 continue;
//             }else if(mySourceChar==UCNV_SO){
//                 myData->toU2022State.g = 1;
//                 myData->isEmptySegment = TRUE;	/* Begin a new segment, empty so far */
//                 /*consume the source */
//                 continue;
//             }else if(mySourceChar==ESC_2022){
//                 mySource--;
// escape:
//                 myData->isEmptySegment = FALSE;	/* Any invalid ESC sequences will be detected separately, so just reset this */
//                 changeState_2022(args->converter,&(mySource),
//                                 mySourceLimit, ISO_2022_KR, err);
//                 if(U_FAILURE(*err)){
//                     args->target = myTarget;
//                     args->source = mySource;
//                     return;
//                 }
//                 continue;
//             }

//             myData->isEmptySegment = FALSE;	/* Any invalid char errors will be detected separately, so just reset this */
//             if(myData->toU2022State.g == 1) {
//                 if(mySource < mySourceLimit) {
//                     int leadIsOk, trailIsOk;
//                     uint8_t trailByte;
// getTrailByte:
//                     targetUniChar = missingCharMarker;
//                     trailByte = (uint8_t)*mySource;
//                     /*
//                      * Ticket 5691: consistent illegal sequences:
//                      * - We include at least the first byte in the illegal sequence.
//                      * - If any of the non-initial bytes could be the start of a character,
//                      *   we stop the illegal sequence before the first one of those.
//                      *
//                      * In ISO-2022 DBCS, if the second byte is in the 21..7e range or is
//                      * an ESC/SO/SI, we report only the first byte as the illegal sequence.
//                      * Otherwise we convert or report the pair of bytes.
//                      */
//                     leadIsOk = (uint8_t)(mySourceChar - 0x21) <= (0x7e - 0x21);
//                     trailIsOk = (uint8_t)(trailByte - 0x21) <= (0x7e - 0x21);
//                     if (leadIsOk && trailIsOk) {
//                         ++mySource;
//                         tempBuf[0] = (char)(mySourceChar + 0x80);
//                         tempBuf[1] = (char)(trailByte + 0x80);
//                         targetUniChar = ucnv_MBCSSimpleGetNextUChar(sharedData, tempBuf, 2, useFallback);
//                         mySourceChar = (mySourceChar << 8) | trailByte;
//                     } else if (!(trailIsOk || IS_2022_CONTROL(trailByte))) {
//                         /* report a pair of illegal bytes if the second byte is not a DBCS starter */
//                         ++mySource;
//                         /* add another bit so that the code below writes 2 bytes in case of error */
//                         mySourceChar = static_cast<UChar>(0x10000 | (mySourceChar << 8) | trailByte);
//                     }
//                 } else {
//                     args->converter->toUBytes[0] = (uint8_t)mySourceChar;
//                     args->converter->toULength = 1;
//                     break;
//                 }
//             }
//             else if(mySourceChar <= 0x7f) {
//                 targetUniChar = ucnv_MBCSSimpleGetNextUChar(sharedData, mySource - 1, 1, useFallback);
//             } else {
//                 targetUniChar = 0xffff;
//             }
//             if(targetUniChar < 0xfffe){
//                 if(args->offsets) {
//                     args->offsets[myTarget - args->target] = (int32_t)(mySource - args->source - (mySourceChar <= 0xff ? 1 : 2));
//                 }
//                 *(myTarget++)=(UChar)targetUniChar;
//             }
//             else {
//                 /* Call the callback function*/
//                 toUnicodeCallback(args->converter,mySourceChar,targetUniChar,err);
//                 break;
//             }
//         }
//         else{
//             *err =U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }
//     }
//     args->target = myTarget;
//     args->source = mySource;
// }

// /*************************** END ISO2022-KR *********************************/

// /*************************** ISO-2022-CN *********************************
// *
// * Rules for ISO-2022-CN Encoding:
// * i)   The designator sequence must appear once on a line before any instance
// *      of character set it designates.
// * ii)  If two lines contain characters from the same character set, both lines
// *      must include the designator sequence.
// * iii) Once the designator sequence is known, a shifting sequence has to be found
// *      to invoke the  shifting
// * iv)  All lines start in ASCII and end in ASCII.
// * v)   Four shifting sequences are employed for this purpose:
// *
// *      Sequcence   ASCII Eq    Charsets
// *      ----------  -------    ---------
// *      SI           <SI>        US-ASCII
// *      SO           <SO>        CNS-11643-1992 Plane 1, GB2312, ISO-IR-165
// *      SS2          <ESC>N      CNS-11643-1992 Plane 2
// *      SS3          <ESC>O      CNS-11643-1992 Planes 3-7
// *
// * vi)
// *      SOdesignator  : ESC "$" ")" finalchar_for_SO
// *      SS2designator : ESC "$" "*" finalchar_for_SS2
// *      SS3designator : ESC "$" "+" finalchar_for_SS3
// *
// *      ESC $ ) A       Indicates the bytes following SO are Chinese
// *       characters as defined in GB 2312-80, until
// *       another SOdesignation appears
// *
// *
// *      ESC $ ) E       Indicates the bytes following SO are as defined
// *       in ISO-IR-165 (for details, see section 2.1),
// *       until another SOdesignation appears
// *
// *      ESC $ ) G       Indicates the bytes following SO are as defined
// *       in CNS 11643-plane-1, until another
// *       SOdesignation appears
// *
// *      ESC $ * H       Indicates the two bytes immediately following
// *       SS2 is a Chinese character as defined in CNS
// *       11643-plane-2, until another SS2designation
// *       appears
// *       (Meaning <ESC>N must preceed every 2 byte
// *        sequence.)
// *
// *      ESC $ + I       Indicates the immediate two bytes following SS3
// *       is a Chinese character as defined in CNS
// *       11643-plane-3, until another SS3designation
// *       appears
// *       (Meaning <ESC>O must preceed every 2 byte
// *        sequence.)
// *
// *      ESC $ + J       Indicates the immediate two bytes following SS3
// *       is a Chinese character as defined in CNS
// *       11643-plane-4, until another SS3designation
// *       appears
// *       (In English: <ESC>O must preceed every 2 byte
// *        sequence.)
// *
// *      ESC $ + K       Indicates the immediate two bytes following SS3
// *       is a Chinese character as defined in CNS
// *       11643-plane-5, until another SS3designation
// *       appears
// *
// *      ESC $ + L       Indicates the immediate two bytes following SS3
// *       is a Chinese character as defined in CNS
// *       11643-plane-6, until another SS3designation
// *       appears
// *
// *      ESC $ + M       Indicates the immediate two bytes following SS3
// *       is a Chinese character as defined in CNS
// *       11643-plane-7, until another SS3designation
// *       appears
// *
// *       As in ISO-2022-CN, each line starts in ASCII, and ends in ASCII, and
// *       has its own designation information before any Chinese characters
// *       appear
// *
// */

// /* The following are defined this way to make the strings truly readonly */
// static const char GB_2312_80_STR[] = "\x1B\x24\x29\x41";
// static const char ISO_IR_165_STR[] = "\x1B\x24\x29\x45";
// static const char CNS_11643_1992_Plane_1_STR[] = "\x1B\x24\x29\x47";
// static const char CNS_11643_1992_Plane_2_STR[] = "\x1B\x24\x2A\x48";
// static const char CNS_11643_1992_Plane_3_STR[] = "\x1B\x24\x2B\x49";
// static const char CNS_11643_1992_Plane_4_STR[] = "\x1B\x24\x2B\x4A";
// static const char CNS_11643_1992_Plane_5_STR[] = "\x1B\x24\x2B\x4B";
// static const char CNS_11643_1992_Plane_6_STR[] = "\x1B\x24\x2B\x4C";
// static const char CNS_11643_1992_Plane_7_STR[] = "\x1B\x24\x2B\x4D";

// /********************** ISO2022-CN Data **************************/
// static const char* const escSeqCharsCN[10] ={
//         SHIFT_IN_STR,                   /* 0 ASCII */
//         GB_2312_80_STR,                 /* 1 GB2312_1 */
//         ISO_IR_165_STR,                 /* 2 ISO_IR_165 */
//         CNS_11643_1992_Plane_1_STR,
//         CNS_11643_1992_Plane_2_STR,
//         CNS_11643_1992_Plane_3_STR,
//         CNS_11643_1992_Plane_4_STR,
//         CNS_11643_1992_Plane_5_STR,
//         CNS_11643_1992_Plane_6_STR,
//         CNS_11643_1992_Plane_7_STR
// };

// static void U_CALLCONV
// UConverter_fromUnicode_ISO_2022_CN_OFFSETS_LOGIC(UConverterFromUnicodeArgs* args, UErrorCode* err){
//     UConverter *cnv = args->converter;
//     UConverterDataISO2022 *converterData;
//     ISO2022State *pFromU2022State;
//     uint8_t *target = (uint8_t *) args->target;
//     const uint8_t *targetLimit = (const uint8_t *) args->targetLimit;
//     const UChar* source = args->source;
//     const UChar* sourceLimit = args->sourceLimit;
//     int32_t* offsets = args->offsets;
//     UChar32 sourceChar;
//     char buffer[8];
//     int32_t len;
//     int8_t choices[3];
//     int32_t choiceCount;
//     uint32_t targetValue = 0;
//     UBool useFallback;

//     /* set up the state */
//     converterData     = (UConverterDataISO2022*)cnv->extraInfo;
//     pFromU2022State   = &converterData->fromU2022State;

//     choiceCount = 0;

//     /* check if the last codepoint of previous buffer was a lead surrogate*/
//     if((sourceChar = cnv->fromUChar32)!=0 && target< targetLimit) {
//         goto getTrail;
//     }

//     while( source < sourceLimit){
//         if(target < targetLimit){

//             sourceChar  = *(source++);
//             /*check if the char is a First surrogate*/
//              if(U16_IS_SURROGATE(sourceChar)) {
//                 if(U16_IS_SURROGATE_LEAD(sourceChar)) {
// getTrail:
//                     /*look ahead to find the trail surrogate*/
//                     if(source < sourceLimit) {
//                         /* test the following code unit */
//                         UChar trail=(UChar) *source;
//                         if(U16_IS_TRAIL(trail)) {
//                             source++;
//                             sourceChar=U16_GET_SUPPLEMENTARY(sourceChar, trail);
//                             cnv->fromUChar32=0x00;
//                             /* convert this supplementary code point */
//                             /* exit this condition tree */
//                         } else {
//                             /* this is an unmatched lead code unit (1st surrogate) */
//                             /* callback(illegal) */
//                             *err=U_ILLEGAL_CHAR_FOUND;
//                             cnv->fromUChar32=sourceChar;
//                             break;
//                         }
//                     } else {
//                         /* no more input */
//                         cnv->fromUChar32=sourceChar;
//                         break;
//                     }
//                 } else {
//                     /* this is an unmatched trail code unit (2nd surrogate) */
//                     /* callback(illegal) */
//                     *err=U_ILLEGAL_CHAR_FOUND;
//                     cnv->fromUChar32=sourceChar;
//                     break;
//                 }
//             }

//             /* do the conversion */
//             if(sourceChar <= 0x007f ){
//                 /* do not convert SO/SI/ESC */
//                 if(IS_2022_CONTROL(sourceChar)) {
//                     /* callback(illegal) */
//                     *err=U_ILLEGAL_CHAR_FOUND;
//                     cnv->fromUChar32=sourceChar;
//                     break;
//                 }

//                 /* US-ASCII */
//                 if(pFromU2022State->g == 0) {
//                     buffer[0] = (char)sourceChar;
//                     len = 1;
//                 } else {
//                     buffer[0] = UCNV_SI;
//                     buffer[1] = (char)sourceChar;
//                     len = 2;
//                     pFromU2022State->g = 0;
//                     choiceCount = 0;
//                 }
//                 if(sourceChar == CR || sourceChar == LF) {
//                     /* reset the state at the end of a line */
//                     uprv_memset(pFromU2022State, 0, sizeof(ISO2022State));
//                     choiceCount = 0;
//                 }
//             }
//             else{
//                 /* convert U+0080..U+10ffff */
//                 int32_t i;
//                 int8_t cs, g;

//                 if(choiceCount == 0) {
//                     /* try the current SO/G1 converter first */
//                     choices[0] = pFromU2022State->cs[1];

//                     /* default to GB2312_1 if none is designated yet */
//                     if(choices[0] == 0) {
//                         choices[0] = GB2312_1;
//                     }

//                     if(converterData->version == 0) {
//                         /* ISO-2022-CN */

//                         /* try the other SO/G1 converter; a CNS_11643_1 lookup may result in any plane */
//                         if(choices[0] == GB2312_1) {
//                             choices[1] = (int8_t)CNS_11643_1;
//                         } else {
//                             choices[1] = (int8_t)GB2312_1;
//                         }

//                         choiceCount = 2;
//                     } else if (converterData->version == 1) {
//                         /* ISO-2022-CN-EXT */

//                         /* try one of the other converters */
//                         switch(choices[0]) {
//                         case GB2312_1:
//                             choices[1] = (int8_t)CNS_11643_1;
//                             choices[2] = (int8_t)ISO_IR_165;
//                             break;
//                         case ISO_IR_165:
//                             choices[1] = (int8_t)GB2312_1;
//                             choices[2] = (int8_t)CNS_11643_1;
//                             break;
//                         default: /* CNS_11643_x */
//                             choices[1] = (int8_t)GB2312_1;
//                             choices[2] = (int8_t)ISO_IR_165;
//                             break;
//                         }

//                         choiceCount = 3;
//                     } else {
//                         choices[0] = (int8_t)CNS_11643_1;
//                         choices[1] = (int8_t)GB2312_1;
//                     }
//                 }

//                 cs = g = 0;
//                 /*
//                  * len==0: no mapping found yet
//                  * len<0: found a fallback result: continue looking for a roundtrip but no further fallbacks
//                  * len>0: found a roundtrip result, done
//                  */
//                 len = 0;
//                 /*
//                  * We will turn off useFallback after finding a fallback,
//                  * but we still get fallbacks from PUA code points as usual.
//                  * Therefore, we will also need to check that we don't overwrite
//                  * an early fallback with a later one.
//                  */
//                 useFallback = cnv->useFallback;

//                 for(i = 0; i < choiceCount && len <= 0; ++i) {
//                     int8_t cs0 = choices[i];
//                     if(cs0 > 0) {
//                         uint32_t value;
//                         int32_t len2;
//                         if(cs0 >= CNS_11643_0) {
//                             len2 = MBCS_FROM_UCHAR32_ISO2022(
//                                         converterData->myConverterArray[CNS_11643],
//                                         sourceChar,
//                                         &value,
//                                         useFallback,
//                                         MBCS_OUTPUT_3);
//                             if(len2 == 3 || (len2 == -3 && len == 0)) {
//                                 targetValue = value;
//                                 cs = (int8_t)(CNS_11643_0 + (value >> 16) - 0x80);
//                                 if(len2 >= 0) {
//                                     len = 2;
//                                 } else {
//                                     len = -2;
//                                     useFallback = FALSE;
//                                 }
//                                 if(cs == CNS_11643_1) {
//                                     g = 1;
//                                 } else if(cs == CNS_11643_2) {
//                                     g = 2;
//                                 } else /* plane 3..7 */ if(converterData->version == 1) {
//                                     g = 3;
//                                 } else {
//                                     /* ISO-2022-CN (without -EXT) does not support plane 3..7 */
//                                     len = 0;
//                                 }
//                             }
//                         } else {
//                             /* GB2312_1 or ISO-IR-165 */
//                             U_ASSERT(cs0<UCNV_2022_MAX_CONVERTERS);
//                             len2 = MBCS_FROM_UCHAR32_ISO2022(
//                                         converterData->myConverterArray[cs0],
//                                         sourceChar,
//                                         &value,
//                                         useFallback,
//                                         MBCS_OUTPUT_2);
//                             if(len2 == 2 || (len2 == -2 && len == 0)) {
//                                 targetValue = value;
//                                 len = len2;
//                                 cs = cs0;
//                                 g = 1;
//                                 useFallback = FALSE;
//                             }
//                         }
//                     }
//                 }

//                 if(len != 0) {
//                     len = 0; /* count output bytes; it must have been abs(len) == 2 */

//                     /* write the designation sequence if necessary */
//                     if(cs != pFromU2022State->cs[g]) {
//                         if(cs < CNS_11643) {
//                             uprv_memcpy(buffer, escSeqCharsCN[cs], 4);
//                         } else {
//                             U_ASSERT(cs >= CNS_11643_1);
//                             uprv_memcpy(buffer, escSeqCharsCN[CNS_11643 + (cs - CNS_11643_1)], 4);
//                         }
//                         len = 4;
//                         pFromU2022State->cs[g] = cs;
//                         if(g == 1) {
//                             /* changing the SO/G1 charset invalidates the choices[] */
//                             choiceCount = 0;
//                         }
//                     }

//                     /* write the shift sequence if necessary */
//                     if(g != pFromU2022State->g) {
//                         switch(g) {
//                         case 1:
//                             buffer[len++] = UCNV_SO;

//                             /* set the new state only if it is the locking shift SO/G1, not for SS2 or SS3 */
//                             pFromU2022State->g = 1;
//                             break;
//                         case 2:
//                             buffer[len++] = 0x1b;
//                             buffer[len++] = 0x4e;
//                             break;
//                         default: /* case 3 */
//                             buffer[len++] = 0x1b;
//                             buffer[len++] = 0x4f;
//                             break;
//                         }
//                     }

//                     /* write the two output bytes */
//                     buffer[len++] = (char)(targetValue >> 8);
//                     buffer[len++] = (char)targetValue;
//                 } else {
//                     /* if we cannot find the character after checking all codepages
//                      * then this is an error
//                      */
//                     *err = U_INVALID_CHAR_FOUND;
//                     cnv->fromUChar32=sourceChar;
//                     break;
//                 }
//             }

//             /* output len>0 bytes in buffer[] */
//             if(len == 1) {
//                 *target++ = buffer[0];
//                 if(offsets) {
//                     *offsets++ = (int32_t)(source - args->source - 1); /* -1: known to be ASCII */
//                 }
//             } else if(len == 2 && (target + 2) <= targetLimit) {
//                 *target++ = buffer[0];
//                 *target++ = buffer[1];
//                 if(offsets) {
//                     int32_t sourceIndex = (int32_t)(source - args->source - U16_LENGTH(sourceChar));
//                     *offsets++ = sourceIndex;
//                     *offsets++ = sourceIndex;
//                 }
//             } else {
//                 fromUWriteUInt8(
//                     cnv,
//                     buffer, len,
//                     &target, (const char *)targetLimit,
//                     &offsets, (int32_t)(source - args->source - U16_LENGTH(sourceChar)),
//                     err);
//                 if(U_FAILURE(*err)) {
//                     break;
//                 }
//             }
//         } /* end if(myTargetIndex<myTargetLength) */
//         else{
//             *err =U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }

//     }/* end while(mySourceIndex<mySourceLength) */

//     /*
//      * the end of the input stream and detection of truncated input
//      * are handled by the framework, but for ISO-2022-CN conversion
//      * we need to be in ASCII mode at the very end
//      *
//      * conditions:
//      *   successful
//      *   not in ASCII mode
//      *   end of input and no truncated input
//      */
//     if( U_SUCCESS(*err) &&
//         pFromU2022State->g!=0 &&
//         args->flush && source>=sourceLimit && cnv->fromUChar32==0
//     ) {
//         int32_t sourceIndex;

//         /* we are switching to ASCII */
//         pFromU2022State->g=0;

//         /* get the source index of the last input character */
//         /*
//          * TODO this would be simpler and more reliable if we used a pair
//          * of sourceIndex/prevSourceIndex like in ucnvmbcs.c
//          * so that we could simply use the prevSourceIndex here;
//          * this code gives an incorrect result for the rare case of an unmatched
//          * trail surrogate that is alone in the last buffer of the text stream
//          */
//         sourceIndex=(int32_t)(source-args->source);
//         if(sourceIndex>0) {
//             --sourceIndex;
//             if( U16_IS_TRAIL(args->source[sourceIndex]) &&
//                 (sourceIndex==0 || U16_IS_LEAD(args->source[sourceIndex-1]))
//             ) {
//                 --sourceIndex;
//             }
//         } else {
//             sourceIndex=-1;
//         }

//         fromUWriteUInt8(
//             cnv,
//             SHIFT_IN_STR, 1,
//             &target, (const char *)targetLimit,
//             &offsets, sourceIndex,
//             err);
//     }

//     /*save the state and return */
//     args->source = source;
//     args->target = (char*)target;
// }


// static void U_CALLCONV
// UConverter_toUnicode_ISO_2022_CN_OFFSETS_LOGIC(UConverterToUnicodeArgs *args,
//                                                UErrorCode* err){
//     char tempBuf[3];
//     const char *mySource = (char *) args->source;
//     UChar *myTarget = args->target;
//     const char *mySourceLimit = args->sourceLimit;
//     uint32_t targetUniChar = 0x0000;
//     uint32_t mySourceChar = 0x0000;
//     UConverterDataISO2022* myData;
//     ISO2022State *pToU2022State;

//     myData=(UConverterDataISO2022*)(args->converter->extraInfo);
//     pToU2022State = &myData->toU2022State;

//     if(myData->key != 0) {
//         /* continue with a partial escape sequence */
//         goto escape;
//     } else if(args->converter->toULength == 1 && mySource < mySourceLimit && myTarget < args->targetLimit) {
//         /* continue with a partial double-byte character */
//         mySourceChar = args->converter->toUBytes[0];
//         args->converter->toULength = 0;
//         targetUniChar = missingCharMarker;
//         goto getTrailByte;
//     }

//     while(mySource < mySourceLimit){

//         targetUniChar =missingCharMarker;

//         if(myTarget < args->targetLimit){

//             mySourceChar= (unsigned char) *mySource++;

//             switch(mySourceChar){
//             case UCNV_SI:
//                 pToU2022State->g=0;
//                 if (myData->isEmptySegment) {
//                     myData->isEmptySegment = FALSE;	/* we are handling it, reset to avoid future spurious errors */
//                     *err = U_ILLEGAL_ESCAPE_SEQUENCE;
//                     args->converter->toUCallbackReason = UCNV_IRREGULAR;
//                     args->converter->toUBytes[0] = static_cast<uint8_t>(mySourceChar);
//                     args->converter->toULength = 1;
//                     args->target = myTarget;
//                     args->source = mySource;
//                     return;
//                 }
//                 continue;

//             case UCNV_SO:
//                 if(pToU2022State->cs[1] != 0) {
//                     pToU2022State->g=1;
//                     myData->isEmptySegment = TRUE;	/* Begin a new segment, empty so far */
//                     continue;
//                 } else {
//                     /* illegal to have SO before a matching designator */
//                     myData->isEmptySegment = FALSE;	/* Handling a different error, reset this to avoid future spurious errs */
//                     break;
//                 }

//             case ESC_2022:
//                 mySource--;
// escape:
//                 {
//                     const char * mySourceBefore = mySource;
//                     int8_t toULengthBefore = args->converter->toULength;

//                     changeState_2022(args->converter,&(mySource),
//                         mySourceLimit, ISO_2022_CN,err);

//                     /* After SO there must be at least one character before a designator (designator error handled separately) */
//                     if(myData->key==0 && U_SUCCESS(*err) && myData->isEmptySegment) {
//                         *err = U_ILLEGAL_ESCAPE_SEQUENCE;
//                         args->converter->toUCallbackReason = UCNV_IRREGULAR;
//                         args->converter->toULength = (int8_t)(toULengthBefore + (mySource - mySourceBefore));
//                     }
//                 }

//                 /* invalid or illegal escape sequence */
//                 if(U_FAILURE(*err)){
//                     args->target = myTarget;
//                     args->source = mySource;
//                     myData->isEmptySegment = FALSE;	/* Reset to avoid future spurious errors */
//                     return;
//                 }
//                 continue;

//             /* ISO-2022-CN does not use single-byte (C1) SS2 and SS3 */

//             case CR:
//             case LF:
//                 uprv_memset(pToU2022State, 0, sizeof(ISO2022State));
//                 U_FALLTHROUGH;
//             default:
//                 /* convert one or two bytes */
//                 myData->isEmptySegment = FALSE;
//                 if(pToU2022State->g != 0) {
//                     if(mySource < mySourceLimit) {
//                         UConverterSharedData *cnv;
//                         StateEnum tempState;
//                         int32_t tempBufLen;
//                         int leadIsOk, trailIsOk;
//                         uint8_t trailByte;
// getTrailByte:
//                         trailByte = (uint8_t)*mySource;
//                         /*
//                          * Ticket 5691: consistent illegal sequences:
//                          * - We include at least the first byte in the illegal sequence.
//                          * - If any of the non-initial bytes could be the start of a character,
//                          *   we stop the illegal sequence before the first one of those.
//                          *
//                          * In ISO-2022 DBCS, if the second byte is in the 21..7e range or is
//                          * an ESC/SO/SI, we report only the first byte as the illegal sequence.
//                          * Otherwise we convert or report the pair of bytes.
//                          */
//                         leadIsOk = (uint8_t)(mySourceChar - 0x21) <= (0x7e - 0x21);
//                         trailIsOk = (uint8_t)(trailByte - 0x21) <= (0x7e - 0x21);
//                         if (leadIsOk && trailIsOk) {
//                             ++mySource;
//                             tempState = (StateEnum)pToU2022State->cs[pToU2022State->g];
//                             if(tempState >= CNS_11643_0) {
//                                 cnv = myData->myConverterArray[CNS_11643];
//                                 tempBuf[0] = (char) (0x80+(tempState-CNS_11643_0));
//                                 tempBuf[1] = (char) (mySourceChar);
//                                 tempBuf[2] = (char) trailByte;
//                                 tempBufLen = 3;

//                             }else{
//                                 U_ASSERT(tempState<UCNV_2022_MAX_CONVERTERS);
//                                 cnv = myData->myConverterArray[tempState];
//                                 tempBuf[0] = (char) (mySourceChar);
//                                 tempBuf[1] = (char) trailByte;
//                                 tempBufLen = 2;
//                             }
//                             targetUniChar = ucnv_MBCSSimpleGetNextUChar(cnv, tempBuf, tempBufLen, FALSE);
//                             mySourceChar = (mySourceChar << 8) | trailByte;
//                         } else if (!(trailIsOk || IS_2022_CONTROL(trailByte))) {
//                             /* report a pair of illegal bytes if the second byte is not a DBCS starter */
//                             ++mySource;
//                             /* add another bit so that the code below writes 2 bytes in case of error */
//                             mySourceChar = 0x10000 | (mySourceChar << 8) | trailByte;
//                         }
//                         if(pToU2022State->g>=2) {
//                             /* return from a single-shift state to the previous one */
//                             pToU2022State->g=pToU2022State->prevG;
//                         }
//                     } else {
//                         args->converter->toUBytes[0] = (uint8_t)mySourceChar;
//                         args->converter->toULength = 1;
//                         goto endloop;
//                     }
//                 }
//                 else{
//                     if(mySourceChar <= 0x7f) {
//                         targetUniChar = (UChar) mySourceChar;
//                     }
//                 }
//                 break;
//             }
//             if(targetUniChar < (missingCharMarker-1/*0xfffe*/)){
//                 if(args->offsets){
//                     args->offsets[myTarget - args->target] = (int32_t)(mySource - args->source - (mySourceChar <= 0xff ? 1 : 2));
//                 }
//                 *(myTarget++)=(UChar)targetUniChar;
//             }
//             else if(targetUniChar > missingCharMarker){
//                 /* disassemble the surrogate pair and write to output*/
//                 targetUniChar-=0x0010000;
//                 *myTarget = (UChar)(0xd800+(UChar)(targetUniChar>>10));
//                 if(args->offsets){
//                     args->offsets[myTarget - args->target] = (int32_t)(mySource - args->source - (mySourceChar <= 0xff ? 1 : 2));
//                 }
//                 ++myTarget;
//                 if(myTarget< args->targetLimit){
//                     *myTarget = (UChar)(0xdc00+(UChar)(targetUniChar&0x3ff));
//                     if(args->offsets){
//                         args->offsets[myTarget - args->target] = (int32_t)(mySource - args->source - (mySourceChar <= 0xff ? 1 : 2));
//                     }
//                     ++myTarget;
//                 }else{
//                     args->converter->UCharErrorBuffer[args->converter->UCharErrorBufferLength++]=
//                                     (UChar)(0xdc00+(UChar)(targetUniChar&0x3ff));
//                 }

//             }
//             else{
//                 /* Call the callback function*/
//                 toUnicodeCallback(args->converter,mySourceChar,targetUniChar,err);
//                 break;
//             }
//         }
//         else{
//             *err =U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }
//     }
// endloop:
//     args->target = myTarget;
//     args->source = mySource;
// }
// #endif /* #if !UCONFIG_ONLY_HTML_CONVERSION */

// static void U_CALLCONV
// _ISO_2022_WriteSub(UConverterFromUnicodeArgs *args, int32_t offsetIndex, UErrorCode *err) {
//     UConverter *cnv = args->converter;
//     UConverterDataISO2022 *myConverterData=(UConverterDataISO2022 *) cnv->extraInfo;
//     ISO2022State *pFromU2022State=&myConverterData->fromU2022State;
//     char *p, *subchar;
//     char buffer[8];
//     int32_t length;

//     subchar=(char *)cnv->subChars;
//     length=cnv->subCharLen; /* assume length==1 for most variants */

//     p = buffer;
//     switch(myConverterData->locale[0]){
//     case 'j':
//         {
//             int8_t cs;

//             if(pFromU2022State->g == 1) {
//                 /* JIS7: switch from G1 to G0 */
//                 pFromU2022State->g = 0;
//                 *p++ = UCNV_SI;
//             }

//             cs = pFromU2022State->cs[0];
//             if(cs != ASCII && cs != JISX201) {
//                 /* not in ASCII or JIS X 0201: switch to ASCII */
//                 pFromU2022State->cs[0] = (int8_t)ASCII;
//                 *p++ = '\x1b';
//                 *p++ = '\x28';
//                 *p++ = '\x42';
//             }

//             *p++ = subchar[0];
//             break;
//         }
//     case 'c':
//         if(pFromU2022State->g != 0) {
//             /* not in ASCII mode: switch to ASCII */
//             pFromU2022State->g = 0;
//             *p++ = UCNV_SI;
//         }
//         *p++ = subchar[0];
//         break;
//     case 'k':
//         if(myConverterData->version == 0) {
//             if(length == 1) {
//                 if(args->converter->fromUnicodeStatus) {
//                     /* in DBCS mode: switch to SBCS */
//                     args->converter->fromUnicodeStatus = 0;
//                     *p++ = UCNV_SI;
//                 }
//                 *p++ = subchar[0];
//             } else /* length == 2*/ {
//                 if(!args->converter->fromUnicodeStatus) {
//                     /* in SBCS mode: switch to DBCS */
//                     args->converter->fromUnicodeStatus = 1;
//                     *p++ = UCNV_SO;
//                 }
//                 *p++ = subchar[0];
//                 *p++ = subchar[1];
//             }
//             break;
//         } else {
//             /* save the subconverter's substitution string */
//             uint8_t *currentSubChars = myConverterData->currentConverter->subChars;
//             int8_t currentSubCharLen = myConverterData->currentConverter->subCharLen;

//             /* set our substitution string into the subconverter */
//             myConverterData->currentConverter->subChars = (uint8_t *)subchar;
//             myConverterData->currentConverter->subCharLen = (int8_t)length;

//             /* let the subconverter write the subchar, set/retrieve fromUChar32 state */
//             args->converter = myConverterData->currentConverter;
//             myConverterData->currentConverter->fromUChar32 = cnv->fromUChar32;
//             ucnv_cbFromUWriteSub(args, 0, err);
//             cnv->fromUChar32 = myConverterData->currentConverter->fromUChar32;
//             args->converter = cnv;

//             /* restore the subconverter's substitution string */
//             myConverterData->currentConverter->subChars = currentSubChars;
//             myConverterData->currentConverter->subCharLen = currentSubCharLen;

//             if(*err == U_BUFFER_OVERFLOW_ERROR) {
//                 if(myConverterData->currentConverter->charErrorBufferLength > 0) {
//                     uprv_memcpy(
//                         cnv->charErrorBuffer,
//                         myConverterData->currentConverter->charErrorBuffer,
//                         myConverterData->currentConverter->charErrorBufferLength);
//                 }
//                 cnv->charErrorBufferLength = myConverterData->currentConverter->charErrorBufferLength;
//                 myConverterData->currentConverter->charErrorBufferLength = 0;
//             }
//             return;
//         }
//     default:
//         /* not expected */
//         break;
//     }
//     ucnv_cbFromUWriteBytes(args,
//                            buffer, (int32_t)(p - buffer),
//                            offsetIndex, err);
// }

// /*
//  * Structure for cloning an ISO 2022 converter into a single memory block.
//  */
// struct cloneStruct
// {
//     UConverter cnv;
//     UConverter currentConverter;
//     UConverterDataISO2022 mydata;
// };


// U_CDECL_BEGIN

// static UConverter * U_CALLCONV
// _ISO_2022_SafeClone(
//             const UConverter *cnv,
//             void *stackBuffer,
//             int32_t *pBufferSize,
//             UErrorCode *status)
// {
//     struct cloneStruct * localClone;
//     UConverterDataISO2022 *cnvData;
//     int32_t i, size;

//     if (U_FAILURE(*status)){
//         return nullptr;
//     }

//     if (*pBufferSize == 0) { /* 'preflighting' request - set needed size into *pBufferSize */
//         *pBufferSize = (int32_t)sizeof(struct cloneStruct);
//         return NULL;
//     }

//     cnvData = (UConverterDataISO2022 *)cnv->extraInfo;
//     localClone = (struct cloneStruct *)stackBuffer;

//     /* ucnv.c/ucnv_safeClone() copied the main UConverter already */

//     uprv_memcpy(&localClone->mydata, cnvData, sizeof(UConverterDataISO2022));
//     localClone->cnv.extraInfo = &localClone->mydata; /* set pointer to extra data */
//     localClone->cnv.isExtraLocal = TRUE;

//     /* share the subconverters */

//     if(cnvData->currentConverter != NULL) {
//         size = (int32_t)sizeof(UConverter);
//         localClone->mydata.currentConverter =
//             ucnv_safeClone(cnvData->currentConverter,
//                             &localClone->currentConverter,
//                             &size, status);
//         if(U_FAILURE(*status)) {
//             return NULL;
//         }
//     }

//     for(i=0; i<UCNV_2022_MAX_CONVERTERS; ++i) {
//         if(cnvData->myConverterArray[i] != NULL) {
//             ucnv_incrementRefCount(cnvData->myConverterArray[i]);
//         }
//     }

//     return &localClone->cnv;
// }

// U_CDECL_END

// static void U_CALLCONV
// _ISO_2022_GetUnicodeSet(const UConverter *cnv,
//                     const USetAdder *sa,
//                     UConverterUnicodeSet which,
//                     UErrorCode *pErrorCode)
// {
//     int32_t i;
//     UConverterDataISO2022* cnvData;

//     if (U_FAILURE(*pErrorCode)) {
//         return;
//     }
// #ifdef U_ENABLE_GENERIC_ISO_2022
//     if (cnv->sharedData == &_ISO2022Data) {
//         /* We use UTF-8 in this case */
//         sa->addRange(sa->set, 0, 0xd7FF);
//         sa->addRange(sa->set, 0xE000, 0x10FFFF);
//         return;
//     }
// #endif

//     cnvData = (UConverterDataISO2022*)cnv->extraInfo;

//     /* open a set and initialize it with code points that are algorithmically round-tripped */
//     switch(cnvData->locale[0]){
//     case 'j':
//         /* include JIS X 0201 which is hardcoded */
//         sa->add(sa->set, 0xa5);
//         sa->add(sa->set, 0x203e);
//         if(jpCharsetMasks[cnvData->version]&CSM(ISO8859_1)) {
//             /* include Latin-1 for some variants of JP */
//             sa->addRange(sa->set, 0, 0xff);
//         } else {
//             /* include ASCII for JP */
//             sa->addRange(sa->set, 0, 0x7f);
//         }
//         if(cnvData->version==3 || cnvData->version==4 || which==UCNV_ROUNDTRIP_AND_FALLBACK_SET) {
//             /*
//              * Do not test (jpCharsetMasks[cnvData->version]&CSM(HWKANA_7BIT))!=0
//              * because the bit is on for all JP versions although only versions 3 & 4 (JIS7 & JIS8)
//              * use half-width Katakana.
//              * This is because all ISO-2022-JP variants are lenient in that they accept (in toUnicode)
//              * half-width Katakana via the ESC ( I sequence.
//              * However, we only emit (fromUnicode) half-width Katakana according to the
//              * definition of each variant.
//              *
//              * When including fallbacks,
//              * we need to include half-width Katakana Unicode code points for all JP variants because
//              * JIS X 0208 has hardcoded fallbacks for them (which map to full-width Katakana).
//              */
//             /* include half-width Katakana for JP */
//             sa->addRange(sa->set, HWKANA_START, HWKANA_END);
//         }
//         break;
// #if !UCONFIG_ONLY_HTML_CONVERSION
//     case 'c':
//     case 'z':
//         /* include ASCII for CN */
//         sa->addRange(sa->set, 0, 0x7f);
//         break;
//     case 'k':
//         /* there is only one converter for KR, and it is not in the myConverterArray[] */
//         cnvData->currentConverter->sharedData->impl->getUnicodeSet(
//                 cnvData->currentConverter, sa, which, pErrorCode);
//         /* the loop over myConverterArray[] will simply not find another converter */
//         break;
// #endif
//     default:
//         break;
//     }

// #if 0  /* Replaced by ucnv_MBCSGetFilteredUnicodeSetForUnicode() until we implement ucnv_getUnicodeSet() with reverse fallbacks. */
//             if( (cnvData->locale[0]=='c' || cnvData->locale[0]=='z') &&
//                 cnvData->version==0 && i==CNS_11643
//             ) {
//                 /* special handling for non-EXT ISO-2022-CN: add only code points for CNS planes 1 and 2 */
//                 ucnv_MBCSGetUnicodeSetForBytes(
//                         cnvData->myConverterArray[i],
//                         sa, UCNV_ROUNDTRIP_SET,
//                         0, 0x81, 0x82,
//                         pErrorCode);
//             }
// #endif

//     for (i=0; i<UCNV_2022_MAX_CONVERTERS; i++) {
//         UConverterSetFilter filter;
//         if(cnvData->myConverterArray[i]!=NULL) {
//             if(cnvData->locale[0]=='j' && i==JISX208) {
//                 /*
//                  * Only add code points that map to Shift-JIS codes
//                  * corresponding to JIS X 0208.
//                  */
//                 filter=UCNV_SET_FILTER_SJIS;
// #if !UCONFIG_ONLY_HTML_CONVERSION
//             } else if( (cnvData->locale[0]=='c' || cnvData->locale[0]=='z') &&
//                        cnvData->version==0 && i==CNS_11643) {
//                 /*
//                  * Version-specific for CN:
//                  * CN version 0 does not map CNS planes 3..7 although
//                  * they are all available in the CNS conversion table;
//                  * CN version 1 (-EXT) does map them all.
//                  * The two versions create different Unicode sets.
//                  */
//                 filter=UCNV_SET_FILTER_2022_CN;
//             } else if(i==KSC5601) {
//                 /*
//                  * Some of the KSC 5601 tables (convrtrs.txt has this aliases on multiple tables)
//                  * are broader than GR94.
//                  */
//                 filter=UCNV_SET_FILTER_GR94DBCS;
// #endif
//             } else {
//                 filter=UCNV_SET_FILTER_NONE;
//             }
//             ucnv_MBCSGetFilteredUnicodeSetForUnicode(cnvData->myConverterArray[i], sa, which, filter, pErrorCode);
//         }
//     }

//     /*
//      * ISO 2022 converters must not convert SO/SI/ESC despite what
//      * sub-converters do by themselves.
//      * Remove these characters from the set.
//      */
//     sa->remove(sa->set, 0x0e);
//     sa->remove(sa->set, 0x0f);
//     sa->remove(sa->set, 0x1b);

//     /* ISO 2022 converters do not convert C1 controls either */
//     sa->removeRange(sa->set, 0x80, 0x9f);
// }

// static const UConverterImpl _ISO2022Impl={
//     UCNV_ISO_2022,

//     NULL,
//     NULL,

//     _ISO2022Open,
//     _ISO2022Close,
//     _ISO2022Reset,

// #ifdef U_ENABLE_GENERIC_ISO_2022
//     T_UConverter_toUnicode_ISO_2022_OFFSETS_LOGIC,
//     T_UConverter_toUnicode_ISO_2022_OFFSETS_LOGIC,
//     ucnv_fromUnicode_UTF8,
//     ucnv_fromUnicode_UTF8_OFFSETS_LOGIC,
// #else
//     NULL,
//     NULL,
//     NULL,
//     NULL,
// #endif
//     NULL,

//     NULL,
//     _ISO2022getName,
//     _ISO_2022_WriteSub,
//     _ISO_2022_SafeClone,
//     _ISO_2022_GetUnicodeSet,

//     NULL,
//     NULL
// };
// static const UConverterStaticData _ISO2022StaticData={
//     sizeof(UConverterStaticData),
//     "ISO_2022",
//     2022,
//     UCNV_IBM,
//     UCNV_ISO_2022,
//     1,
//     3, /* max 3 bytes per UChar from UTF-8 (4 bytes from surrogate _pair_) */
//     { 0x1a, 0, 0, 0 },
//     1,
//     FALSE,
//     FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };
// const UConverterSharedData _ISO2022Data=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_ISO2022StaticData, &_ISO2022Impl);

// /*************JP****************/
// static const UConverterImpl _ISO2022JPImpl={
//     UCNV_ISO_2022,

//     NULL,
//     NULL,

//     _ISO2022Open,
//     _ISO2022Close,
//     _ISO2022Reset,

//     UConverter_toUnicode_ISO_2022_JP_OFFSETS_LOGIC,
//     UConverter_toUnicode_ISO_2022_JP_OFFSETS_LOGIC,
//     UConverter_fromUnicode_ISO_2022_JP_OFFSETS_LOGIC,
//     UConverter_fromUnicode_ISO_2022_JP_OFFSETS_LOGIC,
//     NULL,

//     NULL,
//     _ISO2022getName,
//     _ISO_2022_WriteSub,
//     _ISO_2022_SafeClone,
//     _ISO_2022_GetUnicodeSet,

//     NULL,
//     NULL
// };
// static const UConverterStaticData _ISO2022JPStaticData={
//     sizeof(UConverterStaticData),
//     "ISO_2022_JP",
//     0,
//     UCNV_IBM,
//     UCNV_ISO_2022,
//     1,
//     6, /* max 6 bytes per UChar: 4-byte escape sequence + DBCS */
//     { 0x1a, 0, 0, 0 },
//     1,
//     FALSE,
//     FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };

// namespace {

// const UConverterSharedData _ISO2022JPData=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_ISO2022JPStaticData, &_ISO2022JPImpl);

// }  // namespace

// #if !UCONFIG_ONLY_HTML_CONVERSION
// /************* KR ***************/
// static const UConverterImpl _ISO2022KRImpl={
//     UCNV_ISO_2022,

//     NULL,
//     NULL,

//     _ISO2022Open,
//     _ISO2022Close,
//     _ISO2022Reset,

//     UConverter_toUnicode_ISO_2022_KR_OFFSETS_LOGIC,
//     UConverter_toUnicode_ISO_2022_KR_OFFSETS_LOGIC,
//     UConverter_fromUnicode_ISO_2022_KR_OFFSETS_LOGIC,
//     UConverter_fromUnicode_ISO_2022_KR_OFFSETS_LOGIC,
//     NULL,

//     NULL,
//     _ISO2022getName,
//     _ISO_2022_WriteSub,
//     _ISO_2022_SafeClone,
//     _ISO_2022_GetUnicodeSet,

//     NULL,
//     NULL
// };
// static const UConverterStaticData _ISO2022KRStaticData={
//     sizeof(UConverterStaticData),
//     "ISO_2022_KR",
//     0,
//     UCNV_IBM,
//     UCNV_ISO_2022,
//     1,
//     8, /* max 8 bytes per UChar */
//     { 0x1a, 0, 0, 0 },
//     1,
//     FALSE,
//     FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };

// namespace {

// const UConverterSharedData _ISO2022KRData=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_ISO2022KRStaticData, &_ISO2022KRImpl);

// }  // namespace

// /*************** CN ***************/
// static const UConverterImpl _ISO2022CNImpl={

//     UCNV_ISO_2022,

//     NULL,
//     NULL,

//     _ISO2022Open,
//     _ISO2022Close,
//     _ISO2022Reset,

//     UConverter_toUnicode_ISO_2022_CN_OFFSETS_LOGIC,
//     UConverter_toUnicode_ISO_2022_CN_OFFSETS_LOGIC,
//     UConverter_fromUnicode_ISO_2022_CN_OFFSETS_LOGIC,
//     UConverter_fromUnicode_ISO_2022_CN_OFFSETS_LOGIC,
//     NULL,

//     NULL,
//     _ISO2022getName,
//     _ISO_2022_WriteSub,
//     _ISO_2022_SafeClone,
//     _ISO_2022_GetUnicodeSet,

//     NULL,
//     NULL
// };
// static const UConverterStaticData _ISO2022CNStaticData={
//     sizeof(UConverterStaticData),
//     "ISO_2022_CN",
//     0,
//     UCNV_IBM,
//     UCNV_ISO_2022,
//     1,
//     8, /* max 8 bytes per UChar: 4-byte CNS designator + 2 bytes for SS2/SS3 + DBCS */
//     { 0x1a, 0, 0, 0 },
//     1,
//     FALSE,
//     FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };

// namespace {

// const UConverterSharedData _ISO2022CNData=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_ISO2022CNStaticData, &_ISO2022CNImpl);

// }  // namespace
// #endif /* #if !UCONFIG_ONLY_HTML_CONVERSION */

// #endif /* #if !UCONFIG_NO_LEGACY_CONVERSION */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
//  ********************************************************************
//  * COPYRIGHT:
//  * Copyright (c) 1996-2016, International Business Machines Corporation and
//  * others. All Rights Reserved.
//  ********************************************************************
//  *
//  *  ucnv_bld.cpp:
//  *
//  *  Defines functions that are used in the creation/initialization/deletion
//  *  of converters and related structures.
//  *  uses uconv_io.h routines to access disk information
//  *  is used by ucnv.h to implement public API create/delete/flushCache routines
//  * Modification History:
//  *
//  *   Date        Name        Description
//  *
//  *   06/20/2000  helena      OS/400 port changes; mostly typecast.
//  *   06/29/2000  helena      Major rewrite of the callback interface.
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "unicode/putil.h"
// #include "unicode/udata.h"
// #include "unicode/ucnv.h"
// #include "unicode/uloc.h"
// #include "mutex.h"
// #include "putilimp.h"
// #include "uassert.h"
// #include "utracimp.h"
// #include "ucnv_io.h"
// #include "ucnv_bld.h"
// #include "ucnvmbcs.h"
// #include "ucnv_ext.h"
// #include "ucnv_cnv.h"
// #include "ucnv_imp.h"
// #include "uhash.h"
// #include "umutex.h"
// #include "cstring.h"
// #include "cmemory.h"
// #include "ucln_cmn.h"
// #include "ustr_cnv.h"


// #if 0
// #include <stdio.h>
// extern void UCNV_DEBUG_LOG(char *what, char *who, void *p, int l);
// #define UCNV_DEBUG_LOG(x,y,z) UCNV_DEBUG_LOG(x,y,z,__LINE__)
// #else
// # define UCNV_DEBUG_LOG(x,y,z)
// #endif

// static const UConverterSharedData * const
// converterData[UCNV_NUMBER_OF_SUPPORTED_CONVERTER_TYPES]={
//     NULL, NULL,

// #if UCONFIG_NO_LEGACY_CONVERSION
//     NULL,
// #else
//     &_MBCSData,
// #endif

//     &_Latin1Data,
//     &_UTF8Data, &_UTF16BEData, &_UTF16LEData,
// #if UCONFIG_ONLY_HTML_CONVERSION
//     NULL, NULL,
// #else
//     &_UTF32BEData, &_UTF32LEData,
// #endif
//     NULL,

// #if UCONFIG_NO_LEGACY_CONVERSION
//     NULL,
// #else
//     &_ISO2022Data,
// #endif

// #if UCONFIG_NO_LEGACY_CONVERSION || UCONFIG_ONLY_HTML_CONVERSION
//     NULL, NULL, NULL, NULL, NULL, NULL,
//     NULL, NULL, NULL, NULL, NULL, NULL,
//     NULL,
// #else
//     &_LMBCSData1,&_LMBCSData2, &_LMBCSData3, &_LMBCSData4, &_LMBCSData5, &_LMBCSData6,
//     &_LMBCSData8,&_LMBCSData11,&_LMBCSData16,&_LMBCSData17,&_LMBCSData18,&_LMBCSData19,
//     &_HZData,
// #endif

// #if UCONFIG_ONLY_HTML_CONVERSION
//     NULL,
// #else
//     &_SCSUData,
// #endif


// #if UCONFIG_NO_LEGACY_CONVERSION || UCONFIG_ONLY_HTML_CONVERSION
//     NULL,
// #else
//     &_ISCIIData,
// #endif

//     &_ASCIIData,
// #if UCONFIG_ONLY_HTML_CONVERSION
//     NULL, NULL, &_UTF16Data, NULL, NULL, NULL,
// #else
//     &_UTF7Data, &_Bocu1Data, &_UTF16Data, &_UTF32Data, &_CESU8Data, &_IMAPData,
// #endif

// #if UCONFIG_NO_LEGACY_CONVERSION || UCONFIG_ONLY_HTML_CONVERSION
//     NULL,
// #else
//     &_CompoundTextData
// #endif
// };

// /* Please keep this in binary sorted order for getAlgorithmicTypeFromName.
//    Also the name should be in lower case and all spaces, dashes and underscores
//    removed
// */
// static struct {
//   const char *name;
//   const UConverterType type;
// } const cnvNameType[] = {
// #if !UCONFIG_ONLY_HTML_CONVERSION
//   { "bocu1", UCNV_BOCU1 },
//   { "cesu8", UCNV_CESU8 },
// #endif
// #if !UCONFIG_NO_LEGACY_CONVERSION && !UCONFIG_ONLY_HTML_CONVERSION
//   { "hz",UCNV_HZ },
// #endif
// #if !UCONFIG_ONLY_HTML_CONVERSION
//   { "imapmailboxname", UCNV_IMAP_MAILBOX },
// #endif
// #if !UCONFIG_NO_LEGACY_CONVERSION && !UCONFIG_ONLY_HTML_CONVERSION
//   { "iscii", UCNV_ISCII },
// #endif
// #if !UCONFIG_NO_LEGACY_CONVERSION
//   { "iso2022", UCNV_ISO_2022 },
// #endif
//   { "iso88591", UCNV_LATIN_1 },
// #if !UCONFIG_NO_LEGACY_CONVERSION && !UCONFIG_ONLY_HTML_CONVERSION
//   { "lmbcs1", UCNV_LMBCS_1 },
//   { "lmbcs11",UCNV_LMBCS_11 },
//   { "lmbcs16",UCNV_LMBCS_16 },
//   { "lmbcs17",UCNV_LMBCS_17 },
//   { "lmbcs18",UCNV_LMBCS_18 },
//   { "lmbcs19",UCNV_LMBCS_19 },
//   { "lmbcs2", UCNV_LMBCS_2 },
//   { "lmbcs3", UCNV_LMBCS_3 },
//   { "lmbcs4", UCNV_LMBCS_4 },
//   { "lmbcs5", UCNV_LMBCS_5 },
//   { "lmbcs6", UCNV_LMBCS_6 },
//   { "lmbcs8", UCNV_LMBCS_8 },
// #endif
// #if !UCONFIG_ONLY_HTML_CONVERSION
//   { "scsu", UCNV_SCSU },
// #endif
//   { "usascii", UCNV_US_ASCII },
//   { "utf16", UCNV_UTF16 },
//   { "utf16be", UCNV_UTF16_BigEndian },
//   { "utf16le", UCNV_UTF16_LittleEndian },
// #if U_IS_BIG_ENDIAN
//   { "utf16oppositeendian", UCNV_UTF16_LittleEndian },
//   { "utf16platformendian", UCNV_UTF16_BigEndian },
// #else
//   { "utf16oppositeendian", UCNV_UTF16_BigEndian},
//   { "utf16platformendian", UCNV_UTF16_LittleEndian },
// #endif
// #if !UCONFIG_ONLY_HTML_CONVERSION
//   { "utf32", UCNV_UTF32 },
//   { "utf32be", UCNV_UTF32_BigEndian },
//   { "utf32le", UCNV_UTF32_LittleEndian },
// #if U_IS_BIG_ENDIAN
//   { "utf32oppositeendian", UCNV_UTF32_LittleEndian },
//   { "utf32platformendian", UCNV_UTF32_BigEndian },
// #else
//   { "utf32oppositeendian", UCNV_UTF32_BigEndian },
//   { "utf32platformendian", UCNV_UTF32_LittleEndian },
// #endif
// #endif
// #if !UCONFIG_ONLY_HTML_CONVERSION
//   { "utf7", UCNV_UTF7 },
// #endif
//   { "utf8", UCNV_UTF8 },
// #if !UCONFIG_ONLY_HTML_CONVERSION
//   { "x11compoundtext", UCNV_COMPOUND_TEXT}
// #endif
// };


// /*initializes some global variables */
// static UHashtable *SHARED_DATA_HASHTABLE = NULL;
// static icu::UMutex cnvCacheMutex;
// /*  Note:  the global mutex is used for      */
// /*         reference count updates.          */

// static const char **gAvailableConverters = NULL;
// static uint16_t gAvailableConverterCount = 0;
// static icu::UInitOnce gAvailableConvertersInitOnce = U_INITONCE_INITIALIZER;

// #if !U_CHARSET_IS_UTF8

// /* This contains the resolved converter name. So no further alias lookup is needed again. */
// static char gDefaultConverterNameBuffer[UCNV_MAX_CONVERTER_NAME_LENGTH + 1]; /* +1 for NULL */
// static const char *gDefaultConverterName = NULL;

// /*
// If the default converter is an algorithmic converter, this is the cached value.
// We don't cache a full UConverter and clone it because ucnv_clone doesn't have
// less overhead than an algorithmic open. We don't cache non-algorithmic converters
// because ucnv_flushCache must be able to unload the default converter and its table.
// */
// static const UConverterSharedData *gDefaultAlgorithmicSharedData = NULL;

// /* Does gDefaultConverterName have a converter option and require extra parsing? */
// static UBool gDefaultConverterContainsOption;

// #endif  /* !U_CHARSET_IS_UTF8 */

// static const char DATA_TYPE[] = "cnv";

// /* ucnv_flushAvailableConverterCache. This is only called from ucnv_cleanup().
//  *                       If it is ever to be called from elsewhere, synchronization
//  *                       will need to be considered.
//  */
// static void
// ucnv_flushAvailableConverterCache() {
//     gAvailableConverterCount = 0;
//     if (gAvailableConverters) {
//         uprv_free((char **)gAvailableConverters);
//         gAvailableConverters = NULL;
//     }
//     gAvailableConvertersInitOnce.reset();
// }

// /* ucnv_cleanup - delete all storage held by the converter cache, except any  */
// /*                in use by open converters.                                  */
// /*                Not thread safe.                                            */
// /*                Not supported API.                                          */
// static UBool U_CALLCONV ucnv_cleanup(void) {
//     ucnv_flushCache();
//     if (SHARED_DATA_HASHTABLE != NULL && uhash_count(SHARED_DATA_HASHTABLE) == 0) {
//         uhash_close(SHARED_DATA_HASHTABLE);
//         SHARED_DATA_HASHTABLE = NULL;
//     }

//     /* Isn't called from flushCache because other threads may have preexisting references to the table. */
//     ucnv_flushAvailableConverterCache();

// #if !U_CHARSET_IS_UTF8
//     gDefaultConverterName = NULL;
//     gDefaultConverterNameBuffer[0] = 0;
//     gDefaultConverterContainsOption = FALSE;
//     gDefaultAlgorithmicSharedData = NULL;
// #endif

//     return (SHARED_DATA_HASHTABLE == NULL);
// }

// U_CAPI void U_EXPORT2
// ucnv_enableCleanup() {
//     ucln_common_registerCleanup(UCLN_COMMON_UCNV, ucnv_cleanup);
// }

// static UBool U_CALLCONV
// isCnvAcceptable(void * /*context*/,
//                 const char * /*type*/, const char * /*name*/,
//                 const UDataInfo *pInfo) {
//     return (UBool)(
//         pInfo->size>=20 &&
//         pInfo->isBigEndian==U_IS_BIG_ENDIAN &&
//         pInfo->charsetFamily==U_CHARSET_FAMILY &&
//         pInfo->sizeofUChar==U_SIZEOF_UCHAR &&
//         pInfo->dataFormat[0]==0x63 &&   /* dataFormat="cnvt" */
//         pInfo->dataFormat[1]==0x6e &&
//         pInfo->dataFormat[2]==0x76 &&
//         pInfo->dataFormat[3]==0x74 &&
//         pInfo->formatVersion[0]==6);  /* Everything will be version 6 */
// }

// /**
//  * Un flatten shared data from a UDATA..
//  */
// static UConverterSharedData*
// ucnv_data_unFlattenClone(UConverterLoadArgs *pArgs, UDataMemory *pData, UErrorCode *status)
// {
//     /* UDataInfo info; -- necessary only if some converters have different formatVersion */
//     const uint8_t *raw = (const uint8_t *)udata_getMemory(pData);
//     const UConverterStaticData *source = (const UConverterStaticData *) raw;
//     UConverterSharedData *data;
//     UConverterType type = (UConverterType)source->conversionType;

//     if(U_FAILURE(*status))
//         return NULL;

//     if( (uint16_t)type >= UCNV_NUMBER_OF_SUPPORTED_CONVERTER_TYPES ||
//         converterData[type] == NULL ||
//         !converterData[type]->isReferenceCounted ||
//         converterData[type]->referenceCounter != 1 ||
//         source->structSize != sizeof(UConverterStaticData))
//     {
//         *status = U_INVALID_TABLE_FORMAT;
//         return NULL;
//     }

//     data = (UConverterSharedData *)uprv_malloc(sizeof(UConverterSharedData));
//     if(data == NULL) {
//         *status = U_MEMORY_ALLOCATION_ERROR;
//         return NULL;
//     }

//     /* copy initial values from the static structure for this type */
//     uprv_memcpy(data, converterData[type], sizeof(UConverterSharedData));

//     data->staticData = source;

//     data->sharedDataCached = FALSE;

//     /* fill in fields from the loaded data */
//     data->dataMemory = (void*)pData; /* for future use */

//     if(data->impl->load != NULL) {
//         data->impl->load(data, pArgs, raw + source->structSize, status);
//         if(U_FAILURE(*status)) {
//             uprv_free(data);
//             return NULL;
//         }
//     }
//     return data;
// }

// /*Takes an alias name gets an actual converter file name
//  *goes to disk and opens it.
//  *allocates the memory and returns a new UConverter object
//  */
// static UConverterSharedData *createConverterFromFile(UConverterLoadArgs *pArgs, UErrorCode * err)
// {
//     UDataMemory *data;
//     UConverterSharedData *sharedData;

//     UTRACE_ENTRY_OC(UTRACE_UCNV_LOAD);

//     if (U_FAILURE (*err)) {
//         UTRACE_EXIT_STATUS(*err);
//         return NULL;
//     }

//     UTRACE_DATA2(UTRACE_OPEN_CLOSE, "load converter %s from package %s", pArgs->name, pArgs->pkg);

//     data = udata_openChoice(pArgs->pkg, DATA_TYPE, pArgs->name, isCnvAcceptable, NULL, err);
//     if(U_FAILURE(*err))
//     {
//         UTRACE_EXIT_STATUS(*err);
//         return NULL;
//     }

//     sharedData = ucnv_data_unFlattenClone(pArgs, data, err);
//     if(U_FAILURE(*err))
//     {
//         udata_close(data);
//         UTRACE_EXIT_STATUS(*err);
//         return NULL;
//     }

//     /*
//      * TODO Store pkg in a field in the shared data so that delta-only converters
//      * can load base converters from the same package.
//      * If the pkg name is longer than the field, then either do not load the converter
//      * in the first place, or just set the pkg field to "".
//      */

//     UTRACE_EXIT_PTR_STATUS(sharedData, *err);
//     return sharedData;
// }

// /*returns a converter type from a string
//  */
// static const UConverterSharedData *
// getAlgorithmicTypeFromName(const char *realName)
// {
//     uint32_t mid, start, limit;
//     uint32_t lastMid;
//     int result;
//     char strippedName[UCNV_MAX_CONVERTER_NAME_LENGTH];

//     /* Lower case and remove ignoreable characters. */
//     ucnv_io_stripForCompare(strippedName, realName);

//     /* do a binary search for the alias */
//     start = 0;
//     limit = UPRV_LENGTHOF(cnvNameType);
//     mid = limit;
//     lastMid = UINT32_MAX;

//     for (;;) {
//         mid = (uint32_t)((start + limit) / 2);
//         if (lastMid == mid) {   /* Have we moved? */
//             break;  /* We haven't moved, and it wasn't found. */
//         }
//         lastMid = mid;
//         result = uprv_strcmp(strippedName, cnvNameType[mid].name);

//         if (result < 0) {
//             limit = mid;
//         } else if (result > 0) {
//             start = mid;
//         } else {
//             return converterData[cnvNameType[mid].type];
//         }
//     }

//     return NULL;
// }

// /*
// * Based on the number of known converters, this determines how many times larger
// * the shared data hash table should be. When on small platforms, or just a couple
// * of converters are used, this number should be 2. When memory is plentiful, or
// * when ucnv_countAvailable is ever used with a lot of available converters,
// * this should be 4.
// * Larger numbers reduce the number of hash collisions, but use more memory.
// */
// #define UCNV_CACHE_LOAD_FACTOR 2

// /* Puts the shared data in the static hashtable SHARED_DATA_HASHTABLE */
// /*   Will always be called with the cnvCacheMutex alrady being held   */
// /*     by the calling function.                                       */
// /* Stores the shared data in the SHARED_DATA_HASHTABLE
//  * @param data The shared data
//  */
// static void
// ucnv_shareConverterData(UConverterSharedData * data)
// {
//     UErrorCode err = U_ZERO_ERROR;
//     /*Lazy evaluates the Hashtable itself */
//     /*void *sanity = NULL;*/

//     if (SHARED_DATA_HASHTABLE == NULL)
//     {
//         SHARED_DATA_HASHTABLE = uhash_openSize(uhash_hashChars, uhash_compareChars, NULL,
//                             ucnv_io_countKnownConverters(&err)*UCNV_CACHE_LOAD_FACTOR,
//                             &err);
//         ucnv_enableCleanup();

//         if (U_FAILURE(err))
//             return;
//     }

//     /* ### check to see if the element is not already there! */

//     /*
//     sanity =   ucnv_getSharedConverterData (data->staticData->name);
//     if(sanity != NULL)
//     {
//     UCNV_DEBUG_LOG("put:overwrite!",data->staticData->name,sanity);
//     }
//     UCNV_DEBUG_LOG("put:chk",data->staticData->name,sanity);
//     */

//     /* Mark it shared */
//     data->sharedDataCached = TRUE;

//     uhash_put(SHARED_DATA_HASHTABLE,
//             (void*) data->staticData->name, /* Okay to cast away const as long as
//             keyDeleter == NULL */
//             data,
//             &err);
//     UCNV_DEBUG_LOG("put", data->staticData->name,data);

// }

// /*  Look up a converter name in the shared data cache.                    */
// /*    cnvCacheMutex must be held by the caller to protect the hash table. */
// /* gets the shared data from the SHARED_DATA_HASHTABLE (might return NULL if it isn't there)
//  * @param name The name of the shared data
//  * @return the shared data from the SHARED_DATA_HASHTABLE
//  */
// static UConverterSharedData *
// ucnv_getSharedConverterData(const char *name)
// {
//     /*special case when no Table has yet been created we return NULL */
//     if (SHARED_DATA_HASHTABLE == NULL)
//     {
//         return NULL;
//     }
//     else
//     {
//         UConverterSharedData *rc;

//         rc = (UConverterSharedData*)uhash_get(SHARED_DATA_HASHTABLE, name);
//         UCNV_DEBUG_LOG("get",name,rc);
//         return rc;
//     }
// }

// /*frees the string of memory blocks associates with a sharedConverter
//  *if and only if the referenceCounter == 0
//  */
// /* Deletes (frees) the Shared data it's passed. first it checks the referenceCounter to
//  * see if anyone is using it, if not it frees all the memory stemming from sharedConverterData and
//  * returns TRUE,
//  * otherwise returns FALSE
//  * @param sharedConverterData The shared data
//  * @return if not it frees all the memory stemming from sharedConverterData and
//  * returns TRUE, otherwise returns FALSE
//  */
// static UBool
// ucnv_deleteSharedConverterData(UConverterSharedData * deadSharedData)
// {
//     UTRACE_ENTRY_OC(UTRACE_UCNV_UNLOAD);
//     UTRACE_DATA2(UTRACE_OPEN_CLOSE, "unload converter %s shared data %p", deadSharedData->staticData->name, deadSharedData);

//     if (deadSharedData->referenceCounter > 0) {
//         UTRACE_EXIT_VALUE((int32_t)FALSE);
//         return FALSE;
//     }

//     if (deadSharedData->impl->unload != NULL) {
//         deadSharedData->impl->unload(deadSharedData);
//     }

//     if(deadSharedData->dataMemory != NULL)
//     {
//         UDataMemory *data = (UDataMemory*)deadSharedData->dataMemory;
//         udata_close(data);
//     }

//     uprv_free(deadSharedData);

//     UTRACE_EXIT_VALUE((int32_t)TRUE);
//     return TRUE;
// }

// /**
//  * Load a non-algorithmic converter.
//  * If pkg==NULL, then this function must be called inside umtx_lock(&cnvCacheMutex).
//  */
// UConverterSharedData *
// ucnv_load(UConverterLoadArgs *pArgs, UErrorCode *err) {
//     UConverterSharedData *mySharedConverterData;

//     if(err == NULL || U_FAILURE(*err)) {
//         return NULL;
//     }

//     if(pArgs->pkg != NULL && *pArgs->pkg != 0) {
//         /* application-provided converters are not currently cached */
//         return createConverterFromFile(pArgs, err);
//     }

//     mySharedConverterData = ucnv_getSharedConverterData(pArgs->name);
//     if (mySharedConverterData == NULL)
//     {
//         /*Not cached, we need to stream it in from file */
//         mySharedConverterData = createConverterFromFile(pArgs, err);
//         if (U_FAILURE (*err) || (mySharedConverterData == NULL))
//         {
//             return NULL;
//         }
//         else if (!pArgs->onlyTestIsLoadable)
//         {
//             /* share it with other library clients */
//             ucnv_shareConverterData(mySharedConverterData);
//         }
//     }
//     else
//     {
//         /* The data for this converter was already in the cache.            */
//         /* Update the reference counter on the shared data: one more client */
//         mySharedConverterData->referenceCounter++;
//     }

//     return mySharedConverterData;
// }

// /**
//  * Unload a non-algorithmic converter.
//  * It must be sharedData->isReferenceCounted
//  * and this function must be called inside umtx_lock(&cnvCacheMutex).
//  */
// U_CAPI void
// ucnv_unload(UConverterSharedData *sharedData) {
//     if(sharedData != NULL) {
//         if (sharedData->referenceCounter > 0) {
//             sharedData->referenceCounter--;
//         }

//         if((sharedData->referenceCounter <= 0)&&(sharedData->sharedDataCached == FALSE)) {
//             ucnv_deleteSharedConverterData(sharedData);
//         }
//     }
// }

// U_CFUNC void
// ucnv_unloadSharedDataIfReady(UConverterSharedData *sharedData)
// {
//     if(sharedData != NULL && sharedData->isReferenceCounted) {
//         umtx_lock(&cnvCacheMutex);
//         ucnv_unload(sharedData);
//         umtx_unlock(&cnvCacheMutex);
//     }
// }

// U_CFUNC void
// ucnv_incrementRefCount(UConverterSharedData *sharedData)
// {
//     if(sharedData != NULL && sharedData->isReferenceCounted) {
//         umtx_lock(&cnvCacheMutex);
//         sharedData->referenceCounter++;
//         umtx_unlock(&cnvCacheMutex);
//     }
// }

// /*
//  * *pPieces must be initialized.
//  * The name without options will be copied to pPieces->cnvName.
//  * The locale and options will be copied to pPieces only if present in inName,
//  * otherwise the existing values in pPieces remain.
//  * *pArgs will be set to the pPieces values.
//  */
// static void
// parseConverterOptions(const char *inName,
//                       UConverterNamePieces *pPieces,
//                       UConverterLoadArgs *pArgs,
//                       UErrorCode *err)
// {
//     char *cnvName = pPieces->cnvName;
//     char c;
//     int32_t len = 0;

//     pArgs->name=inName;
//     pArgs->locale=pPieces->locale;
//     pArgs->options=pPieces->options;

//     /* copy the converter name itself to cnvName */
//     while((c=*inName)!=0 && c!=UCNV_OPTION_SEP_CHAR) {
//         if (++len>=UCNV_MAX_CONVERTER_NAME_LENGTH) {
//             *err = U_ILLEGAL_ARGUMENT_ERROR;    /* bad name */
//             pPieces->cnvName[0]=0;
//             return;
//         }
//         *cnvName++=c;
//         inName++;
//     }
//     *cnvName=0;
//     pArgs->name=pPieces->cnvName;

//     /* parse options. No more name copying should occur. */
//     while((c=*inName)!=0) {
//         if(c==UCNV_OPTION_SEP_CHAR) {
//             ++inName;
//         }

//         /* inName is behind an option separator */
//         if(uprv_strncmp(inName, "locale=", 7)==0) {
//             /* do not modify locale itself in case we have multiple locale options */
//             char *dest=pPieces->locale;

//             /* copy the locale option value */
//             inName+=7;
//             len=0;
//             while((c=*inName)!=0 && c!=UCNV_OPTION_SEP_CHAR) {
//                 ++inName;

//                 if(++len>=ULOC_FULLNAME_CAPACITY) {
//                     *err=U_ILLEGAL_ARGUMENT_ERROR;    /* bad name */
//                     pPieces->locale[0]=0;
//                     return;
//                 }

//                 *dest++=c;
//             }
//             *dest=0;
//         } else if(uprv_strncmp(inName, "version=", 8)==0) {
//             /* copy the version option value into bits 3..0 of pPieces->options */
//             inName+=8;
//             c=*inName;
//             if(c==0) {
//                 pArgs->options=(pPieces->options&=~UCNV_OPTION_VERSION);
//                 return;
//             } else if((uint8_t)(c-'0')<10) {
//                 pArgs->options=pPieces->options=(pPieces->options&~UCNV_OPTION_VERSION)|(uint32_t)(c-'0');
//                 ++inName;
//             }
//         } else if(uprv_strncmp(inName, "swaplfnl", 8)==0) {
//             inName+=8;
//             pArgs->options=(pPieces->options|=UCNV_OPTION_SWAP_LFNL);
//         /* add processing for new options here with another } else if(uprv_strncmp(inName, "option-name=", XX)==0) { */
//         } else {
//             /* ignore any other options until we define some */
//             while(((c = *inName++) != 0) && (c != UCNV_OPTION_SEP_CHAR)) {
//             }
//             if(c==0) {
//                 return;
//             }
//         }
//     }
// }

// /*Logic determines if the converter is Algorithmic AND/OR cached
//  *depending on that:
//  * -we either go to get data from disk and cache it (Data=TRUE, Cached=False)
//  * -Get it from a Hashtable (Data=X, Cached=TRUE)
//  * -Call dataConverter initializer (Data=TRUE, Cached=TRUE)
//  * -Call AlgorithmicConverter initializer (Data=FALSE, Cached=TRUE)
//  */
// U_CFUNC UConverterSharedData *
// ucnv_loadSharedData(const char *converterName,
//                     UConverterNamePieces *pPieces,
//                     UConverterLoadArgs *pArgs,
//                     UErrorCode * err) {
//     UConverterNamePieces stackPieces;
//     UConverterLoadArgs stackArgs;
//     UConverterSharedData *mySharedConverterData = NULL;
//     UErrorCode internalErrorCode = U_ZERO_ERROR;
//     UBool mayContainOption = TRUE;
//     UBool checkForAlgorithmic = TRUE;

//     if (U_FAILURE (*err)) {
//         return NULL;
//     }

//     if(pPieces == NULL) {
//         if(pArgs != NULL) {
//             /*
//              * Bad: We may set pArgs pointers to stackPieces fields
//              * which will be invalid after this function returns.
//              */
//             *err = U_INTERNAL_PROGRAM_ERROR;
//             return NULL;
//         }
//         pPieces = &stackPieces;
//     }
//     if(pArgs == NULL) {
//         uprv_memset(&stackArgs, 0, sizeof(stackArgs));
//         stackArgs.size = (int32_t)sizeof(stackArgs);
//         pArgs = &stackArgs;
//     }

//     pPieces->cnvName[0] = 0;
//     pPieces->locale[0] = 0;
//     pPieces->options = 0;

//     pArgs->name = converterName;
//     pArgs->locale = pPieces->locale;
//     pArgs->options = pPieces->options;

//     /* In case "name" is NULL we want to open the default converter. */
//     if (converterName == NULL) {
// #if U_CHARSET_IS_UTF8
//         pArgs->name = "UTF-8";
//         return (UConverterSharedData *)converterData[UCNV_UTF8];
// #else
//         /* Call ucnv_getDefaultName first to query the name from the OS. */
//         pArgs->name = ucnv_getDefaultName();
//         if (pArgs->name == NULL) {
//             *err = U_MISSING_RESOURCE_ERROR;
//             return NULL;
//         }
//         mySharedConverterData = (UConverterSharedData *)gDefaultAlgorithmicSharedData;
//         checkForAlgorithmic = FALSE;
//         mayContainOption = gDefaultConverterContainsOption;
//         /* the default converter name is already canonical */
// #endif
//     }
//     else if(UCNV_FAST_IS_UTF8(converterName)) {
//         /* fastpath for UTF-8 */
//         pArgs->name = "UTF-8";
//         return (UConverterSharedData *)converterData[UCNV_UTF8];
//     }
//     else {
//         /* separate the converter name from the options */
//         parseConverterOptions(converterName, pPieces, pArgs, err);
//         if (U_FAILURE(*err)) {
//             /* Very bad name used. */
//             return NULL;
//         }

//         /* get the canonical converter name */
//         pArgs->name = ucnv_io_getConverterName(pArgs->name, &mayContainOption, &internalErrorCode);
//         if (U_FAILURE(internalErrorCode) || pArgs->name == NULL) {
//             /*
//             * set the input name in case the converter was added
//             * without updating the alias table, or when there is no alias table
//             */
//             pArgs->name = pPieces->cnvName;
//         } else if (internalErrorCode == U_AMBIGUOUS_ALIAS_WARNING) {
//             *err = U_AMBIGUOUS_ALIAS_WARNING;
//         }
//     }

//     /* separate the converter name from the options */
//     if(mayContainOption && pArgs->name != pPieces->cnvName) {
//         parseConverterOptions(pArgs->name, pPieces, pArgs, err);
//     }

//     /* get the shared data for an algorithmic converter, if it is one */
//     if (checkForAlgorithmic) {
//         mySharedConverterData = (UConverterSharedData *)getAlgorithmicTypeFromName(pArgs->name);
//     }
//     if (mySharedConverterData == NULL)
//     {
//         /* it is a data-based converter, get its shared data.               */
//         /* Hold the cnvCacheMutex through the whole process of checking the */
//         /*   converter data cache, and adding new entries to the cache      */
//         /*   to prevent other threads from modifying the cache during the   */
//         /*   process.                                                       */
//         pArgs->nestedLoads=1;
//         pArgs->pkg=NULL;

//         umtx_lock(&cnvCacheMutex);
//         mySharedConverterData = ucnv_load(pArgs, err);
//         umtx_unlock(&cnvCacheMutex);
//         if (U_FAILURE (*err) || (mySharedConverterData == NULL))
//         {
//             return NULL;
//         }
//     }

//     return mySharedConverterData;
// }

// U_CAPI UConverter *
// ucnv_createConverter(UConverter *myUConverter, const char *converterName, UErrorCode * err)
// {
//     UConverterNamePieces stackPieces;
//     UConverterLoadArgs stackArgs=UCNV_LOAD_ARGS_INITIALIZER;
//     UConverterSharedData *mySharedConverterData;

//     UTRACE_ENTRY_OC(UTRACE_UCNV_OPEN);

//     if(U_SUCCESS(*err)) {
//         UTRACE_DATA1(UTRACE_OPEN_CLOSE, "open converter %s", converterName);

//         mySharedConverterData = ucnv_loadSharedData(converterName, &stackPieces, &stackArgs, err);

//         myUConverter = ucnv_createConverterFromSharedData(
//             myUConverter, mySharedConverterData,
//             &stackArgs,
//             err);

//         if(U_SUCCESS(*err)) {
//             UTRACE_EXIT_PTR_STATUS(myUConverter, *err);
//             return myUConverter;
//         }
//     }

//     /* exit with error */
//     UTRACE_EXIT_STATUS(*err);
//     return NULL;
// }

// U_CFUNC UBool
// ucnv_canCreateConverter(const char *converterName, UErrorCode *err) {
//     UConverter myUConverter;
//     UConverterNamePieces stackPieces;
//     UConverterLoadArgs stackArgs=UCNV_LOAD_ARGS_INITIALIZER;
//     UConverterSharedData *mySharedConverterData;

//     UTRACE_ENTRY_OC(UTRACE_UCNV_OPEN);

//     if(U_SUCCESS(*err)) {
//         UTRACE_DATA1(UTRACE_OPEN_CLOSE, "test if can open converter %s", converterName);

//         stackArgs.onlyTestIsLoadable=TRUE;
//         mySharedConverterData = ucnv_loadSharedData(converterName, &stackPieces, &stackArgs, err);
//         ucnv_createConverterFromSharedData(
//             &myUConverter, mySharedConverterData,
//             &stackArgs,
//             err);
//         ucnv_unloadSharedDataIfReady(mySharedConverterData);
//     }

//     UTRACE_EXIT_STATUS(*err);
//     return U_SUCCESS(*err);
// }

// UConverter *
// ucnv_createAlgorithmicConverter(UConverter *myUConverter,
//                                 UConverterType type,
//                                 const char *locale, uint32_t options,
//                                 UErrorCode *err) {
//     UConverter *cnv;
//     const UConverterSharedData *sharedData;
//     UConverterLoadArgs stackArgs=UCNV_LOAD_ARGS_INITIALIZER;

//     UTRACE_ENTRY_OC(UTRACE_UCNV_OPEN_ALGORITHMIC);
//     UTRACE_DATA1(UTRACE_OPEN_CLOSE, "open algorithmic converter type %d", (int32_t)type);

//     if(type<0 || UCNV_NUMBER_OF_SUPPORTED_CONVERTER_TYPES<=type) {
//         *err = U_ILLEGAL_ARGUMENT_ERROR;
//         UTRACE_EXIT_STATUS(U_ILLEGAL_ARGUMENT_ERROR);
//         return NULL;
//     }

//     sharedData = converterData[type];
//     if(sharedData == NULL || sharedData->isReferenceCounted) {
//         /* not a valid type, or not an algorithmic converter */
//         *err = U_ILLEGAL_ARGUMENT_ERROR;
//         UTRACE_EXIT_STATUS(U_ILLEGAL_ARGUMENT_ERROR);
//         return NULL;
//     }

//     stackArgs.name = "";
//     stackArgs.options = options;
//     stackArgs.locale=locale;
//     cnv = ucnv_createConverterFromSharedData(
//             myUConverter, (UConverterSharedData *)sharedData,
//             &stackArgs, err);

//     UTRACE_EXIT_PTR_STATUS(cnv, *err);
//     return cnv;
// }

// U_CFUNC UConverter*
// ucnv_createConverterFromPackage(const char *packageName, const char *converterName, UErrorCode * err)
// {
//     UConverter *myUConverter;
//     UConverterSharedData *mySharedConverterData;
//     UConverterNamePieces stackPieces;
//     UConverterLoadArgs stackArgs=UCNV_LOAD_ARGS_INITIALIZER;

//     UTRACE_ENTRY_OC(UTRACE_UCNV_OPEN_PACKAGE);

//     if(U_FAILURE(*err)) {
//         UTRACE_EXIT_STATUS(*err);
//         return NULL;
//     }

//     UTRACE_DATA2(UTRACE_OPEN_CLOSE, "open converter %s from package %s", converterName, packageName);

//     /* first, get the options out of the converterName string */
//     stackPieces.cnvName[0] = 0;
//     stackPieces.locale[0] = 0;
//     stackPieces.options = 0;
//     parseConverterOptions(converterName, &stackPieces, &stackArgs, err);
//     if (U_FAILURE(*err)) {
//         /* Very bad name used. */
//         UTRACE_EXIT_STATUS(*err);
//         return NULL;
//     }
//     stackArgs.nestedLoads=1;
//     stackArgs.pkg=packageName;

//     /* open the data, unflatten the shared structure */
//     mySharedConverterData = createConverterFromFile(&stackArgs, err);

//     if (U_FAILURE(*err)) {
//         UTRACE_EXIT_STATUS(*err);
//         return NULL;
//     }

//     /* create the actual converter */
//     myUConverter = ucnv_createConverterFromSharedData(NULL, mySharedConverterData, &stackArgs, err);

//     if (U_FAILURE(*err)) {
//         ucnv_close(myUConverter);
//         UTRACE_EXIT_STATUS(*err);
//         return NULL;
//     }

//     UTRACE_EXIT_PTR_STATUS(myUConverter, *err);
//     return myUConverter;
// }


// U_CFUNC UConverter*
// ucnv_createConverterFromSharedData(UConverter *myUConverter,
//                                    UConverterSharedData *mySharedConverterData,
//                                    UConverterLoadArgs *pArgs,
//                                    UErrorCode *err)
// {
//     UBool isCopyLocal;

//     if(U_FAILURE(*err)) {
//         ucnv_unloadSharedDataIfReady(mySharedConverterData);
//         return myUConverter;
//     }
//     if(myUConverter == NULL)
//     {
//         myUConverter = (UConverter *) uprv_malloc (sizeof (UConverter));
//         if(myUConverter == NULL)
//         {
//             *err = U_MEMORY_ALLOCATION_ERROR;
//             ucnv_unloadSharedDataIfReady(mySharedConverterData);
//             return NULL;
//         }
//         isCopyLocal = FALSE;
//     } else {
//         isCopyLocal = TRUE;
//     }

//     /* initialize the converter */
//     uprv_memset(myUConverter, 0, sizeof(UConverter));
//     myUConverter->isCopyLocal = isCopyLocal;
//     /*myUConverter->isExtraLocal = FALSE;*/ /* Set by the memset call */
//     myUConverter->sharedData = mySharedConverterData;
//     myUConverter->options = pArgs->options;
//     if(!pArgs->onlyTestIsLoadable) {
//         myUConverter->preFromUFirstCP = U_SENTINEL;
//         myUConverter->fromCharErrorBehaviour = UCNV_TO_U_DEFAULT_CALLBACK;
//         myUConverter->fromUCharErrorBehaviour = UCNV_FROM_U_DEFAULT_CALLBACK;
//         myUConverter->toUnicodeStatus = mySharedConverterData->toUnicodeStatus;
//         myUConverter->maxBytesPerUChar = mySharedConverterData->staticData->maxBytesPerChar;
//         myUConverter->subChar1 = mySharedConverterData->staticData->subChar1;
//         myUConverter->subCharLen = mySharedConverterData->staticData->subCharLen;
//         myUConverter->subChars = (uint8_t *)myUConverter->subUChars;
//         uprv_memcpy(myUConverter->subChars, mySharedConverterData->staticData->subChar, myUConverter->subCharLen);
//         myUConverter->toUCallbackReason = UCNV_ILLEGAL; /* default reason to invoke (*fromCharErrorBehaviour) */
//     }

//     if(mySharedConverterData->impl->open != NULL) {
//         mySharedConverterData->impl->open(myUConverter, pArgs, err);
//         if(U_FAILURE(*err) && !pArgs->onlyTestIsLoadable) {
//             /* don't ucnv_close() if onlyTestIsLoadable because not fully initialized */
//             ucnv_close(myUConverter);
//             return NULL;
//         }
//     }

//     return myUConverter;
// }

// /*Frees all shared immutable objects that aren't referred to (reference count = 0)
//  */
// U_CAPI int32_t U_EXPORT2
// ucnv_flushCache ()
// {
//     UConverterSharedData *mySharedData = NULL;
//     int32_t pos;
//     int32_t tableDeletedNum = 0;
//     const UHashElement *e;
//     /*UErrorCode status = U_ILLEGAL_ARGUMENT_ERROR;*/
//     int32_t i, remaining;

//     UTRACE_ENTRY_OC(UTRACE_UCNV_FLUSH_CACHE);

//     /* Close the default converter without creating a new one so that everything will be flushed. */
//     u_flushDefaultConverter();

//     /*if shared data hasn't even been lazy evaluated yet
//     * return 0
//     */
//     if (SHARED_DATA_HASHTABLE == NULL) {
//         UTRACE_EXIT_VALUE((int32_t)0);
//         return 0;
//     }

//     /*creates an enumeration to iterate through every element in the
//     * table
//     *
//     * Synchronization:  holding cnvCacheMutex will prevent any other thread from
//     *                   accessing or modifying the hash table during the iteration.
//     *                   The reference count of an entry may be decremented by
//     *                   ucnv_close while the iteration is in process, but this is
//     *                   benign.  It can't be incremented (in ucnv_createConverter())
//     *                   because the sequence of looking up in the cache + incrementing
//     *                   is protected by cnvCacheMutex.
//     */
//     umtx_lock(&cnvCacheMutex);
//     /*
//      * double loop: A delta/extension-only converter has a pointer to its base table's
//      * shared data; the first iteration of the outer loop may see the delta converter
//      * before the base converter, and unloading the delta converter may get the base
//      * converter's reference counter down to 0.
//      */
//     i = 0;
//     do {
//         remaining = 0;
//         pos = UHASH_FIRST;
//         while ((e = uhash_nextElement (SHARED_DATA_HASHTABLE, &pos)) != NULL)
//         {
//             mySharedData = (UConverterSharedData *) e->value.pointer;
//             /*deletes only if reference counter == 0 */
//             if (mySharedData->referenceCounter == 0)
//             {
//                 tableDeletedNum++;

//                 UCNV_DEBUG_LOG("del",mySharedData->staticData->name,mySharedData);

//                 uhash_removeElement(SHARED_DATA_HASHTABLE, e);
//                 mySharedData->sharedDataCached = FALSE;
//                 ucnv_deleteSharedConverterData (mySharedData);
//             } else {
//                 ++remaining;
//             }
//         }
//     } while(++i == 1 && remaining > 0);
//     umtx_unlock(&cnvCacheMutex);

//     UTRACE_DATA1(UTRACE_INFO, "ucnv_flushCache() exits with %d converters remaining", remaining);

//     UTRACE_EXIT_VALUE(tableDeletedNum);
//     return tableDeletedNum;
// }

// /* available converters list --------------------------------------------------- */

// static void U_CALLCONV initAvailableConvertersList(UErrorCode &errCode) {
//     U_ASSERT(gAvailableConverterCount == 0);
//     U_ASSERT(gAvailableConverters == NULL);

//     ucnv_enableCleanup();
//     UEnumeration *allConvEnum = ucnv_openAllNames(&errCode);
//     int32_t allConverterCount = uenum_count(allConvEnum, &errCode);
//     if (U_FAILURE(errCode)) {
//         return;
//     }

//     /* We can't have more than "*converterTable" converters to open */
//     gAvailableConverters = (const char **) uprv_malloc(allConverterCount * sizeof(char*));
//     if (!gAvailableConverters) {
//         errCode = U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }

//     /* Open the default converter to make sure that it has first dibs in the hash table. */
//     UErrorCode localStatus = U_ZERO_ERROR;
//     UConverter tempConverter;
//     ucnv_close(ucnv_createConverter(&tempConverter, NULL, &localStatus));

//     gAvailableConverterCount = 0;

//     for (int32_t idx = 0; idx < allConverterCount; idx++) {
//         localStatus = U_ZERO_ERROR;
//         const char *converterName = uenum_next(allConvEnum, NULL, &localStatus);
//         if (ucnv_canCreateConverter(converterName, &localStatus)) {
//             gAvailableConverters[gAvailableConverterCount++] = converterName;
//         }
//     }

//     uenum_close(allConvEnum);
// }


// static UBool haveAvailableConverterList(UErrorCode *pErrorCode) {
//     umtx_initOnce(gAvailableConvertersInitOnce, &initAvailableConvertersList, *pErrorCode);
//     return U_SUCCESS(*pErrorCode);
// }

// U_CFUNC uint16_t
// ucnv_bld_countAvailableConverters(UErrorCode *pErrorCode) {
//     if (haveAvailableConverterList(pErrorCode)) {
//         return gAvailableConverterCount;
//     }
//     return 0;
// }

// U_CFUNC const char *
// ucnv_bld_getAvailableConverter(uint16_t n, UErrorCode *pErrorCode) {
//     if (haveAvailableConverterList(pErrorCode)) {
//         if (n < gAvailableConverterCount) {
//             return gAvailableConverters[n];
//         }
//         *pErrorCode = U_INDEX_OUTOFBOUNDS_ERROR;
//     }
//     return NULL;
// }

// /* default converter name --------------------------------------------------- */

// #if !U_CHARSET_IS_UTF8
// /*
// Copy the canonical converter name.
// ucnv_getDefaultName must be thread safe, which can call this function.

// ucnv_setDefaultName calls this function and it doesn't have to be
// thread safe because there is no reliable/safe way to reset the
// converter in use in all threads. If you did reset the converter, you
// would not be sure that retrieving a default converter for one string
// would be the same type of default converter for a successive string.
// Since the name is a returned via ucnv_getDefaultName without copying,
// you shouldn't be modifying or deleting the string from a separate thread.
// */
// static inline void
// internalSetName(const char *name, UErrorCode *status) {
//     UConverterNamePieces stackPieces;
//     UConverterLoadArgs stackArgs=UCNV_LOAD_ARGS_INITIALIZER;
//     int32_t length=(int32_t)(uprv_strlen(name));
//     UBool containsOption = (UBool)(uprv_strchr(name, UCNV_OPTION_SEP_CHAR) != NULL);
//     const UConverterSharedData *algorithmicSharedData;

//     stackArgs.name = name;
//     if(containsOption) {
//         stackPieces.cnvName[0] = 0;
//         stackPieces.locale[0] = 0;
//         stackPieces.options = 0;
//         parseConverterOptions(name, &stackPieces, &stackArgs, status);
//         if(U_FAILURE(*status)) {
//             return;
//         }
//     }
//     algorithmicSharedData = getAlgorithmicTypeFromName(stackArgs.name);

//     umtx_lock(&cnvCacheMutex);

//     gDefaultAlgorithmicSharedData = algorithmicSharedData;
//     gDefaultConverterContainsOption = containsOption;
//     uprv_memcpy(gDefaultConverterNameBuffer, name, length);
//     gDefaultConverterNameBuffer[length]=0;

//     /* gDefaultConverterName MUST be the last global var set by this function.  */
//     /*    It is the variable checked in ucnv_getDefaultName() to see if initialization is required. */
//     //    But there is nothing here preventing that from being reordered, either by the compiler
//     //             or hardware. I'm adding the mutex to ucnv_getDefaultName for now. UMTX_CHECK is not enough.
//     //             -- Andy
//     gDefaultConverterName = gDefaultConverterNameBuffer;

//     ucnv_enableCleanup();

//     umtx_unlock(&cnvCacheMutex);
// }
// #endif

// /*
//  * In order to be really thread-safe, the get function would have to take
//  * a buffer parameter and copy the current string inside a mutex block.
//  * This implementation only tries to be really thread-safe while
//  * setting the name.
//  * It assumes that setting a pointer is atomic.
//  */

// U_CAPI const char*  U_EXPORT2
// ucnv_getDefaultName() {
// #if U_CHARSET_IS_UTF8
//     return "UTF-8";
// #else
//     /* local variable to be thread-safe */
//     const char *name;

//     /*
//     Concurrent calls to ucnv_getDefaultName must be thread safe,
//     but ucnv_setDefaultName is not thread safe.
//     */
//     {
//         icu::Mutex lock(&cnvCacheMutex);
//         name = gDefaultConverterName;
//     }
//     if(name==NULL) {
//         UErrorCode errorCode = U_ZERO_ERROR;
//         UConverter *cnv = NULL;

//         name = uprv_getDefaultCodepage();

//         /* if the name is there, test it out and get the canonical name with options */
//         if(name != NULL) {
//             cnv = ucnv_open(name, &errorCode);
//             if(U_SUCCESS(errorCode) && cnv != NULL) {
//                 name = ucnv_getName(cnv, &errorCode);
//             }
//         }

//         if(name == NULL || name[0] == 0
//             || U_FAILURE(errorCode) || cnv == NULL
//             || uprv_strlen(name)>=sizeof(gDefaultConverterNameBuffer))
//         {
//             /* Panic time, let's use a fallback. */
// #if (U_CHARSET_FAMILY == U_ASCII_FAMILY)
//             name = "US-ASCII";
//             /* there is no 'algorithmic' converter for EBCDIC */
// #elif U_PLATFORM == U_PF_OS390
//             name = "ibm-1047_P100-1995" UCNV_SWAP_LFNL_OPTION_STRING;
// #else
//             name = "ibm-37_P100-1995";
// #endif
//         }

//         internalSetName(name, &errorCode);

//         /* The close may make the current name go away. */
//         ucnv_close(cnv);
//     }

//     return name;
// #endif
// }

// #if U_CHARSET_IS_UTF8
// U_CAPI void U_EXPORT2 ucnv_setDefaultName(const char *) {}
// #else
// /*
// This function is not thread safe, and it can't be thread safe.
// See internalSetName or the API reference for details.
// */
// U_CAPI void U_EXPORT2
// ucnv_setDefaultName(const char *converterName) {
//     if(converterName==NULL) {
//         /* reset to the default codepage */
//         gDefaultConverterName=NULL;
//     } else {
//         UErrorCode errorCode = U_ZERO_ERROR;
//         UConverter *cnv = NULL;
//         const char *name = NULL;

//         /* if the name is there, test it out and get the canonical name with options */
//         cnv = ucnv_open(converterName, &errorCode);
//         if(U_SUCCESS(errorCode) && cnv != NULL) {
//             name = ucnv_getName(cnv, &errorCode);
//         }

//         if(U_SUCCESS(errorCode) && name!=NULL) {
//             internalSetName(name, &errorCode);
//         }
//         /* else this converter is bad to use. Don't change it to a bad value. */

//         /* The close may make the current name go away. */
//         ucnv_close(cnv);

//         /* reset the converter cache */
//         u_flushDefaultConverter();
//     }
// }
// #endif

// /* data swapping ------------------------------------------------------------ */

// /* most of this might belong more properly into ucnvmbcs.c, but that is so large */

// #if !UCONFIG_NO_LEGACY_CONVERSION

// U_CAPI int32_t U_EXPORT2
// ucnv_swap(const UDataSwapper *ds,
//           const void *inData, int32_t length, void *outData,
//           UErrorCode *pErrorCode) {
//     const UDataInfo *pInfo;
//     int32_t headerSize;

//     const uint8_t *inBytes;
//     uint8_t *outBytes;

//     uint32_t offset, count, staticDataSize;
//     int32_t size;

//     const UConverterStaticData *inStaticData;
//     UConverterStaticData *outStaticData;

//     const _MBCSHeader *inMBCSHeader;
//     _MBCSHeader *outMBCSHeader;
//     _MBCSHeader mbcsHeader;
//     uint32_t mbcsHeaderLength;
//     UBool noFromU=FALSE;

//     uint8_t outputType;

//     int32_t maxFastUChar, mbcsIndexLength;

//     const int32_t *inExtIndexes;
//     int32_t extOffset;

//     /* udata_swapDataHeader checks the arguments */
//     headerSize=udata_swapDataHeader(ds, inData, length, outData, pErrorCode);
//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     /* check data format and format version */
//     pInfo=(const UDataInfo *)((const char *)inData+4);
//     if(!(
//         pInfo->dataFormat[0]==0x63 &&   /* dataFormat="cnvt" */
//         pInfo->dataFormat[1]==0x6e &&
//         pInfo->dataFormat[2]==0x76 &&
//         pInfo->dataFormat[3]==0x74 &&
//         pInfo->formatVersion[0]==6 &&
//         pInfo->formatVersion[1]>=2
//     )) {
//         udata_printError(ds, "ucnv_swap(): data format %02x.%02x.%02x.%02x (format version %02x.%02x) is not recognized as an ICU .cnv conversion table\n",
//                          pInfo->dataFormat[0], pInfo->dataFormat[1],
//                          pInfo->dataFormat[2], pInfo->dataFormat[3],
//                          pInfo->formatVersion[0], pInfo->formatVersion[1]);
//         *pErrorCode=U_UNSUPPORTED_ERROR;
//         return 0;
//     }

//     inBytes=(const uint8_t *)inData+headerSize;
//     outBytes=(uint8_t *)outData+headerSize;

//     /* read the initial UConverterStaticData structure after the UDataInfo header */
//     inStaticData=(const UConverterStaticData *)inBytes;
//     outStaticData=(UConverterStaticData *)outBytes;

//     if(length<0) {
//         staticDataSize=ds->readUInt32(inStaticData->structSize);
//     } else {
//         length-=headerSize;
//         if( length<(int32_t)sizeof(UConverterStaticData) ||
//             (uint32_t)length<(staticDataSize=ds->readUInt32(inStaticData->structSize))
//         ) {
//             udata_printError(ds, "ucnv_swap(): too few bytes (%d after header) for an ICU .cnv conversion table\n",
//                              length);
//             *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//             return 0;
//         }
//     }

//     if(length>=0) {
//         /* swap the static data */
//         if(inStaticData!=outStaticData) {
//             uprv_memcpy(outStaticData, inStaticData, staticDataSize);
//         }

//         ds->swapArray32(ds, &inStaticData->structSize, 4,
//                            &outStaticData->structSize, pErrorCode);
//         ds->swapArray32(ds, &inStaticData->codepage, 4,
//                            &outStaticData->codepage, pErrorCode);

//         ds->swapInvChars(ds, inStaticData->name, (int32_t)uprv_strlen(inStaticData->name),
//                             outStaticData->name, pErrorCode);
//         if(U_FAILURE(*pErrorCode)) {
//             udata_printError(ds, "ucnv_swap(): error swapping converter name\n");
//             return 0;
//         }
//     }

//     inBytes+=staticDataSize;
//     outBytes+=staticDataSize;
//     if(length>=0) {
//         length-=(int32_t)staticDataSize;
//     }

//     /* check for supported conversionType values */
//     if(inStaticData->conversionType==UCNV_MBCS) {
//         /* swap MBCS data */
//         inMBCSHeader=(const _MBCSHeader *)inBytes;
//         outMBCSHeader=(_MBCSHeader *)outBytes;

//         if(0<=length && length<(int32_t)sizeof(_MBCSHeader)) {
//             udata_printError(ds, "ucnv_swap(): too few bytes (%d after headers) for an ICU MBCS .cnv conversion table\n",
//                                 length);
//             *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//             return 0;
//         }
//         if(inMBCSHeader->version[0]==4 && inMBCSHeader->version[1]>=1) {
//             mbcsHeaderLength=MBCS_HEADER_V4_LENGTH;
//         } else if(inMBCSHeader->version[0]==5 && inMBCSHeader->version[1]>=3 &&
//                   ((mbcsHeader.options=ds->readUInt32(inMBCSHeader->options))&
//                    MBCS_OPT_UNKNOWN_INCOMPATIBLE_MASK)==0
//         ) {
//             mbcsHeaderLength=mbcsHeader.options&MBCS_OPT_LENGTH_MASK;
//             noFromU=(UBool)((mbcsHeader.options&MBCS_OPT_NO_FROM_U)!=0);
//         } else {
//             udata_printError(ds, "ucnv_swap(): unsupported _MBCSHeader.version %d.%d\n",
//                              inMBCSHeader->version[0], inMBCSHeader->version[1]);
//             *pErrorCode=U_UNSUPPORTED_ERROR;
//             return 0;
//         }

//         uprv_memcpy(mbcsHeader.version, inMBCSHeader->version, 4);
//         mbcsHeader.countStates=         ds->readUInt32(inMBCSHeader->countStates);
//         mbcsHeader.countToUFallbacks=   ds->readUInt32(inMBCSHeader->countToUFallbacks);
//         mbcsHeader.offsetToUCodeUnits=  ds->readUInt32(inMBCSHeader->offsetToUCodeUnits);
//         mbcsHeader.offsetFromUTable=    ds->readUInt32(inMBCSHeader->offsetFromUTable);
//         mbcsHeader.offsetFromUBytes=    ds->readUInt32(inMBCSHeader->offsetFromUBytes);
//         mbcsHeader.flags=               ds->readUInt32(inMBCSHeader->flags);
//         mbcsHeader.fromUBytesLength=    ds->readUInt32(inMBCSHeader->fromUBytesLength);
//         /* mbcsHeader.options have been read above */

//         extOffset=(int32_t)(mbcsHeader.flags>>8);
//         outputType=(uint8_t)mbcsHeader.flags;
//         if(noFromU && outputType==MBCS_OUTPUT_1) {
//             udata_printError(ds, "ucnv_swap(): unsupported combination of makeconv --small with SBCS\n");
//             *pErrorCode=U_UNSUPPORTED_ERROR;
//             return 0;
//         }

//         /* make sure that the output type is known */
//         switch(outputType) {
//         case MBCS_OUTPUT_1:
//         case MBCS_OUTPUT_2:
//         case MBCS_OUTPUT_3:
//         case MBCS_OUTPUT_4:
//         case MBCS_OUTPUT_3_EUC:
//         case MBCS_OUTPUT_4_EUC:
//         case MBCS_OUTPUT_2_SISO:
//         case MBCS_OUTPUT_EXT_ONLY:
//             /* OK */
//             break;
//         default:
//             udata_printError(ds, "ucnv_swap(): unsupported MBCS output type 0x%x\n",
//                              outputType);
//             *pErrorCode=U_UNSUPPORTED_ERROR;
//             return 0;
//         }

//         /* calculate the length of the MBCS data */

//         /*
//          * utf8Friendly MBCS files (mbcsHeader.version 4.3)
//          * contain an additional mbcsIndex table:
//          *   uint16_t[(maxFastUChar+1)>>6];
//          * where maxFastUChar=((mbcsHeader.version[2]<<8)|0xff).
//          */
//         maxFastUChar=0;
//         mbcsIndexLength=0;
//         if( outputType!=MBCS_OUTPUT_EXT_ONLY && outputType!=MBCS_OUTPUT_1 &&
//             mbcsHeader.version[1]>=3 && (maxFastUChar=mbcsHeader.version[2])!=0
//         ) {
//             maxFastUChar=(maxFastUChar<<8)|0xff;
//             mbcsIndexLength=((maxFastUChar+1)>>6)*2;  /* number of bytes */
//         }

//         if(extOffset==0) {
//             size=(int32_t)(mbcsHeader.offsetFromUBytes+mbcsIndexLength);
//             if(!noFromU) {
//                 size+=(int32_t)mbcsHeader.fromUBytesLength;
//             }

//             /* avoid compiler warnings - not otherwise necessary, and the value does not matter */
//             inExtIndexes=NULL;
//         } else {
//             /* there is extension data after the base data, see ucnv_ext.h */
//             if(length>=0 && length<(extOffset+UCNV_EXT_INDEXES_MIN_LENGTH*4)) {
//                 udata_printError(ds, "ucnv_swap(): too few bytes (%d after headers) for an ICU MBCS .cnv conversion table with extension data\n",
//                                  length);
//                 *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//                 return 0;
//             }

//             inExtIndexes=(const int32_t *)(inBytes+extOffset);
//             size=extOffset+udata_readInt32(ds, inExtIndexes[UCNV_EXT_SIZE]);
//         }

//         if(length>=0) {
//             if(length<size) {
//                 udata_printError(ds, "ucnv_swap(): too few bytes (%d after headers) for an ICU MBCS .cnv conversion table\n",
//                                  length);
//                 *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//                 return 0;
//             }

//             /* copy the data for inaccessible bytes */
//             if(inBytes!=outBytes) {
//                 uprv_memcpy(outBytes, inBytes, size);
//             }

//             /* swap the MBCSHeader, except for the version field */
//             count=mbcsHeaderLength*4;
//             ds->swapArray32(ds, &inMBCSHeader->countStates, count-4,
//                                &outMBCSHeader->countStates, pErrorCode);

//             if(outputType==MBCS_OUTPUT_EXT_ONLY) {
//                 /*
//                  * extension-only file,
//                  * contains a base name instead of normal base table data
//                  */

//                 /* swap the base name, between the header and the extension data */
//                 const char *inBaseName=(const char *)inBytes+count;
//                 char *outBaseName=(char *)outBytes+count;
//                 ds->swapInvChars(ds, inBaseName, (int32_t)uprv_strlen(inBaseName),
//                                     outBaseName, pErrorCode);
//             } else {
//                 /* normal file with base table data */

//                 /* swap the state table, 1kB per state */
//                 offset=count;
//                 count=mbcsHeader.countStates*1024;
//                 ds->swapArray32(ds, inBytes+offset, (int32_t)count,
//                                    outBytes+offset, pErrorCode);

//                 /* swap the toUFallbacks[] */
//                 offset+=count;
//                 count=mbcsHeader.countToUFallbacks*8;
//                 ds->swapArray32(ds, inBytes+offset, (int32_t)count,
//                                    outBytes+offset, pErrorCode);

//                 /* swap the unicodeCodeUnits[] */
//                 offset=mbcsHeader.offsetToUCodeUnits;
//                 count=mbcsHeader.offsetFromUTable-offset;
//                 ds->swapArray16(ds, inBytes+offset, (int32_t)count,
//                                    outBytes+offset, pErrorCode);

//                 /* offset to the stage 1 table, independent of the outputType */
//                 offset=mbcsHeader.offsetFromUTable;

//                 if(outputType==MBCS_OUTPUT_1) {
//                     /* SBCS: swap the fromU tables, all 16 bits wide */
//                     count=(mbcsHeader.offsetFromUBytes-offset)+mbcsHeader.fromUBytesLength;
//                     ds->swapArray16(ds, inBytes+offset, (int32_t)count,
//                                        outBytes+offset, pErrorCode);
//                 } else {
//                     /* otherwise: swap the stage tables separately */

//                     /* stage 1 table: uint16_t[0x440 or 0x40] */
//                     if(inStaticData->unicodeMask&UCNV_HAS_SUPPLEMENTARY) {
//                         count=0x440*2; /* for all of Unicode */
//                     } else {
//                         count=0x40*2; /* only BMP */
//                     }
//                     ds->swapArray16(ds, inBytes+offset, (int32_t)count,
//                                        outBytes+offset, pErrorCode);

//                     /* stage 2 table: uint32_t[] */
//                     offset+=count;
//                     count=mbcsHeader.offsetFromUBytes-offset;
//                     ds->swapArray32(ds, inBytes+offset, (int32_t)count,
//                                        outBytes+offset, pErrorCode);

//                     /* stage 3/result bytes: sometimes uint16_t[] or uint32_t[] */
//                     offset=mbcsHeader.offsetFromUBytes;
//                     count= noFromU ? 0 : mbcsHeader.fromUBytesLength;
//                     switch(outputType) {
//                     case MBCS_OUTPUT_2:
//                     case MBCS_OUTPUT_3_EUC:
//                     case MBCS_OUTPUT_2_SISO:
//                         ds->swapArray16(ds, inBytes+offset, (int32_t)count,
//                                            outBytes+offset, pErrorCode);
//                         break;
//                     case MBCS_OUTPUT_4:
//                         ds->swapArray32(ds, inBytes+offset, (int32_t)count,
//                                            outBytes+offset, pErrorCode);
//                         break;
//                     default:
//                         /* just uint8_t[], nothing to swap */
//                         break;
//                     }

//                     if(mbcsIndexLength!=0) {
//                         offset+=count;
//                         count=mbcsIndexLength;
//                         ds->swapArray16(ds, inBytes+offset, (int32_t)count,
//                                            outBytes+offset, pErrorCode);
//                     }
//                 }
//             }

//             if(extOffset!=0) {
//                 /* swap the extension data */
//                 inBytes+=extOffset;
//                 outBytes+=extOffset;

//                 /* swap toUTable[] */
//                 offset=udata_readInt32(ds, inExtIndexes[UCNV_EXT_TO_U_INDEX]);
//                 length=udata_readInt32(ds, inExtIndexes[UCNV_EXT_TO_U_LENGTH]);
//                 ds->swapArray32(ds, inBytes+offset, length*4, outBytes+offset, pErrorCode);

//                 /* swap toUUChars[] */
//                 offset=udata_readInt32(ds, inExtIndexes[UCNV_EXT_TO_U_UCHARS_INDEX]);
//                 length=udata_readInt32(ds, inExtIndexes[UCNV_EXT_TO_U_UCHARS_LENGTH]);
//                 ds->swapArray16(ds, inBytes+offset, length*2, outBytes+offset, pErrorCode);

//                 /* swap fromUTableUChars[] */
//                 offset=udata_readInt32(ds, inExtIndexes[UCNV_EXT_FROM_U_UCHARS_INDEX]);
//                 length=udata_readInt32(ds, inExtIndexes[UCNV_EXT_FROM_U_LENGTH]);
//                 ds->swapArray16(ds, inBytes+offset, length*2, outBytes+offset, pErrorCode);

//                 /* swap fromUTableValues[] */
//                 offset=udata_readInt32(ds, inExtIndexes[UCNV_EXT_FROM_U_VALUES_INDEX]);
//                 /* same length as for fromUTableUChars[] */
//                 ds->swapArray32(ds, inBytes+offset, length*4, outBytes+offset, pErrorCode);

//                 /* no need to swap fromUBytes[] */

//                 /* swap fromUStage12[] */
//                 offset=udata_readInt32(ds, inExtIndexes[UCNV_EXT_FROM_U_STAGE_12_INDEX]);
//                 length=udata_readInt32(ds, inExtIndexes[UCNV_EXT_FROM_U_STAGE_12_LENGTH]);
//                 ds->swapArray16(ds, inBytes+offset, length*2, outBytes+offset, pErrorCode);

//                 /* swap fromUStage3[] */
//                 offset=udata_readInt32(ds, inExtIndexes[UCNV_EXT_FROM_U_STAGE_3_INDEX]);
//                 length=udata_readInt32(ds, inExtIndexes[UCNV_EXT_FROM_U_STAGE_3_LENGTH]);
//                 ds->swapArray16(ds, inBytes+offset, length*2, outBytes+offset, pErrorCode);

//                 /* swap fromUStage3b[] */
//                 offset=udata_readInt32(ds, inExtIndexes[UCNV_EXT_FROM_U_STAGE_3B_INDEX]);
//                 length=udata_readInt32(ds, inExtIndexes[UCNV_EXT_FROM_U_STAGE_3B_LENGTH]);
//                 ds->swapArray32(ds, inBytes+offset, length*4, outBytes+offset, pErrorCode);

//                 /* swap indexes[] */
//                 length=udata_readInt32(ds, inExtIndexes[UCNV_EXT_INDEXES_LENGTH]);
//                 ds->swapArray32(ds, inBytes, length*4, outBytes, pErrorCode);
//             }
//         }
//     } else {
//         udata_printError(ds, "ucnv_swap(): unknown conversionType=%d!=UCNV_MBCS\n",
//                          inStaticData->conversionType);
//         *pErrorCode=U_UNSUPPORTED_ERROR;
//         return 0;
//     }

//     return headerSize+(int32_t)staticDataSize+size;
// }

// #endif /* #if !UCONFIG_NO_LEGACY_CONVERSION */

// #endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 2000-2006, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// **********************************************************************
//  *  ucnv_cb.c:
//  *  External APIs for the ICU's codeset conversion library
//  *  Helena Shih
//  *
//  * Modification History:
//  *
//  *   Date        Name        Description
//  *   7/28/2000   srl         Implementation
//  */

// /**
//  * @name Character Conversion C API
//  *
//  */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "unicode/ucnv_cb.h"
// #include "ucnv_bld.h"
// #include "ucnv_cnv.h"
// #include "cmemory.h"

// /* need to update the offsets when the target moves. */
// /* Note: Recursion may occur in the cb functions, be sure to update the offsets correctly
// if you don't use ucnv_cbXXX functions.  Make sure you don't use the same callback within
// the same call stack if the complexity arises. */
// U_CAPI void  U_EXPORT2
// ucnv_cbFromUWriteBytes (UConverterFromUnicodeArgs *args,
//                        const char* source,
//                        int32_t length,
//                        int32_t offsetIndex,
//                        UErrorCode * err)
// {
//     if(U_FAILURE(*err)) {
//         return;
//     }

//     ucnv_fromUWriteBytes(
//         args->converter,
//         source, length,
//         &args->target, args->targetLimit,
//         &args->offsets, offsetIndex,
//         err);
// }

// U_CAPI void  U_EXPORT2
// ucnv_cbFromUWriteUChars(UConverterFromUnicodeArgs *args,
//                              const UChar** source,
//                              const UChar*  sourceLimit,
//                              int32_t offsetIndex,
//                              UErrorCode * err)
// {
//     /*
//     This is a fun one.  Recursion can occur - we're basically going to
//     just retry shoving data through the same converter. Note, if you got
//     here through some kind of invalid sequence, you maybe should emit a
//     reset sequence of some kind and/or call ucnv_reset().  Since this
//     IS an actual conversion, take care that you've changed the callback
//     or the data, or you'll get an infinite loop.

//     Please set the err value to something reasonable before calling
//     into this.
//     */

//     char *oldTarget;

//     if(U_FAILURE(*err))
//     {
//         return;
//     }

//     oldTarget = args->target;

//     ucnv_fromUnicode(args->converter,
//         &args->target,
//         args->targetLimit,
//         source,
//         sourceLimit,
//         NULL, /* no offsets */
//         FALSE, /* no flush */
//         err);

//     if(args->offsets)
//     {
//         while (args->target != oldTarget)  /* if it moved at all.. */
//         {
//             *(args->offsets)++ = offsetIndex;
//             oldTarget++;
//         }
//     }

//     /*
//     Note, if you did something like used a Stop subcallback, things would get interesting.
//     In fact, here's where we want to return the partially consumed in-source!
//     */
//     if(*err == U_BUFFER_OVERFLOW_ERROR)
//     /* && (*source < sourceLimit && args->target >= args->targetLimit)
//     -- S. Hrcek */
//     {
//         /* Overflowed the target.  Now, we'll write into the charErrorBuffer.
//         It's a fixed size. If we overflow it... Hmm */
//         char *newTarget;
//         const char *newTargetLimit;
//         UErrorCode err2 = U_ZERO_ERROR;

//         int8_t errBuffLen;

//         errBuffLen  = args->converter->charErrorBufferLength;

//         /* start the new target at the first free slot in the errbuff.. */
//         newTarget = (char *)(args->converter->charErrorBuffer + errBuffLen);

//         newTargetLimit = (char *)(args->converter->charErrorBuffer +
//             sizeof(args->converter->charErrorBuffer));

//         if(newTarget >= newTargetLimit)
//         {
//             *err = U_INTERNAL_PROGRAM_ERROR;
//             return;
//         }

//         /* We're going to tell the converter that the errbuff len is empty.
//         This prevents the existing errbuff from being 'flushed' out onto
//         itself.  If the errbuff is needed by the converter this time,
//         we're hosed - we're out of space! */

//         args->converter->charErrorBufferLength = 0;

//         ucnv_fromUnicode(args->converter,
//                          &newTarget,
//                          newTargetLimit,
//                          source,
//                          sourceLimit,
//                          NULL,
//                          FALSE,
//                          &err2);

//         /* We can go ahead and overwrite the  length here. We know just how
//         to recalculate it. */

//         args->converter->charErrorBufferLength = (int8_t)(
//             newTarget - (char*)args->converter->charErrorBuffer);

//         if((newTarget >= newTargetLimit) || (err2 == U_BUFFER_OVERFLOW_ERROR))
//         {
//             /* now we're REALLY in trouble.
//             Internal program error - callback shouldn't have written this much
//             data!
//             */
//             *err = U_INTERNAL_PROGRAM_ERROR;
//             return;
//         }
//         /*else {*/
//             /* sub errs could be invalid/truncated/illegal chars or w/e.
//             These might want to be passed on up.. But the problem is, we already
//             need to pass U_BUFFER_OVERFLOW_ERROR. That has to override these
//             other errs.. */

//             /*
//             if(U_FAILURE(err2))
//             ??
//             */
//         /*}*/
//     }
// }

// U_CAPI void  U_EXPORT2
// ucnv_cbFromUWriteSub (UConverterFromUnicodeArgs *args,
//                            int32_t offsetIndex,
//                            UErrorCode * err)
// {
//     UConverter *converter;
//     int32_t length;

//     if(U_FAILURE(*err)) {
//         return;
//     }
//     converter = args->converter;
//     length = converter->subCharLen;

//     if(length == 0) {
//         return;
//     }

//     if(length < 0) {
//         /*
//          * Write/convert the substitution string. Its real length is -length.
//          * Unlike the escape callback, we need not change the converter's
//          * callback function because ucnv_setSubstString() verified that
//          * the string can be converted, so we will not get a conversion error
//          * and will not recurse.
//          * At worst we should get a U_BUFFER_OVERFLOW_ERROR.
//          */
//         const UChar *source = (const UChar *)converter->subChars;
//         ucnv_cbFromUWriteUChars(args, &source, source - length, offsetIndex, err);
//         return;
//     }

//     if(converter->sharedData->impl->writeSub!=NULL) {
//         converter->sharedData->impl->writeSub(args, offsetIndex, err);
//     }
//     else if(converter->subChar1!=0 && (uint16_t)converter->invalidUCharBuffer[0]<=(uint16_t)0xffu) {
//         /*
//         TODO: Is this untestable because the MBCS converter has a writeSub function to call
//         and the other converters don't use subChar1?
//         */
//         ucnv_cbFromUWriteBytes(args,
//                                (const char *)&converter->subChar1, 1,
//                                offsetIndex, err);
//     }
//     else {
//         ucnv_cbFromUWriteBytes(args,
//                                (const char *)converter->subChars, length,
//                                offsetIndex, err);
//     }
// }

// U_CAPI void  U_EXPORT2
// ucnv_cbToUWriteUChars (UConverterToUnicodeArgs *args,
//                             const UChar* source,
//                             int32_t length,
//                             int32_t offsetIndex,
//                             UErrorCode * err)
// {
//     if(U_FAILURE(*err)) {
//         return;
//     }

//     ucnv_toUWriteUChars(
//         args->converter,
//         source, length,
//         &args->target, args->targetLimit,
//         &args->offsets, offsetIndex,
//         err);
// }

// U_CAPI void  U_EXPORT2
// ucnv_cbToUWriteSub (UConverterToUnicodeArgs *args,
//                          int32_t offsetIndex,
//                        UErrorCode * err)
// {
//     static const UChar kSubstituteChar1 = 0x1A, kSubstituteChar = 0xFFFD;

//     /* could optimize this case, just one uchar */
//     if(args->converter->invalidCharLength == 1 && args->converter->subChar1 != 0) {
//         ucnv_cbToUWriteUChars(args, &kSubstituteChar1, 1, offsetIndex, err);
//     } else {
//         ucnv_cbToUWriteUChars(args, &kSubstituteChar, 1, offsetIndex, err);
//     }
// }

// #endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// ******************************************************************************
// *
// *   Copyright (C) 1999-2004, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *
// ******************************************************************************
// *
// *   uconv_cnv.c:
// *   Implements all the low level conversion functions
// *   T_UnicodeConverter_{to,from}Unicode_$ConversionType
// *
// *   Change history:
// *
// *   06/29/2000  helena      Major rewrite of the callback APIs.
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "unicode/ucnv_err.h"
// #include "unicode/ucnv.h"
// #include "unicode/uset.h"
// #include "ucnv_cnv.h"
// #include "ucnv_bld.h"
// #include "cmemory.h"

// U_CFUNC void
// ucnv_getCompleteUnicodeSet(const UConverter *cnv,
//                    const USetAdder *sa,
//                    UConverterUnicodeSet which,
//                    UErrorCode *pErrorCode) {
//     (void)cnv;
//     (void)which;
//     (void)pErrorCode;
//     sa->addRange(sa->set, 0, 0x10ffff);
// }

// U_CFUNC void
// ucnv_getNonSurrogateUnicodeSet(const UConverter *cnv,
//                                const USetAdder *sa,
//                                UConverterUnicodeSet which,
//                                UErrorCode *pErrorCode) {
//     (void)cnv;
//     (void)which;
//     (void)pErrorCode;
//     sa->addRange(sa->set, 0, 0xd7ff);
//     sa->addRange(sa->set, 0xe000, 0x10ffff);
// }

// U_CFUNC void
// ucnv_fromUWriteBytes(UConverter *cnv,
//                      const char *bytes, int32_t length,
//                      char **target, const char *targetLimit,
//                      int32_t **offsets,
//                      int32_t sourceIndex,
//                      UErrorCode *pErrorCode) {
//     char *t=*target;
//     int32_t *o;

//     /* write bytes */
//     if(offsets==NULL || (o=*offsets)==NULL) {
//         while(length>0 && t<targetLimit) {
//             *t++=*bytes++;
//             --length;
//         }
//     } else {
//         /* output with offsets */
//         while(length>0 && t<targetLimit) {
//             *t++=*bytes++;
//             *o++=sourceIndex;
//             --length;
//         }
//         *offsets=o;
//     }
//     *target=t;

//     /* write overflow */
//     if(length>0) {
//         if(cnv!=NULL) {
//             t=(char *)cnv->charErrorBuffer;
//             cnv->charErrorBufferLength=(int8_t)length;
//             do {
//                 *t++=(uint8_t)*bytes++;
//             } while(--length>0);
//         }
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//     }
// }

// U_CFUNC void
// ucnv_toUWriteUChars(UConverter *cnv,
//                     const UChar *uchars, int32_t length,
//                     UChar **target, const UChar *targetLimit,
//                     int32_t **offsets,
//                     int32_t sourceIndex,
//                     UErrorCode *pErrorCode) {
//     UChar *t=*target;
//     int32_t *o;

//     /* write UChars */
//     if(offsets==NULL || (o=*offsets)==NULL) {
//         while(length>0 && t<targetLimit) {
//             *t++=*uchars++;
//             --length;
//         }
//     } else {
//         /* output with offsets */
//         while(length>0 && t<targetLimit) {
//             *t++=*uchars++;
//             *o++=sourceIndex;
//             --length;
//         }
//         *offsets=o;
//     }
//     *target=t;

//     /* write overflow */
//     if(length>0) {
//         if(cnv!=NULL) {
//             t=cnv->UCharErrorBuffer;
//             cnv->UCharErrorBufferLength=(int8_t)length;
//             do {
//                 *t++=*uchars++;
//             } while(--length>0);
//         }
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//     }
// }

// U_CFUNC void
// ucnv_toUWriteCodePoint(UConverter *cnv,
//                        UChar32 c,
//                        UChar **target, const UChar *targetLimit,
//                        int32_t **offsets,
//                        int32_t sourceIndex,
//                        UErrorCode *pErrorCode) {
//     UChar *t;
//     int32_t *o;

//     t=*target;

//     if(t<targetLimit) {
//         if(c<=0xffff) {
//             *t++=(UChar)c;
//             c=U_SENTINEL;
//         } else /* c is a supplementary code point */ {
//             *t++=U16_LEAD(c);
//             c=U16_TRAIL(c);
//             if(t<targetLimit) {
//                 *t++=(UChar)c;
//                 c=U_SENTINEL;
//             }
//         }

//         /* write offsets */
//         if(offsets!=NULL && (o=*offsets)!=NULL) {
//             *o++=sourceIndex;
//             if((*target+1)<t) {
//                 *o++=sourceIndex;
//             }
//             *offsets=o;
//         }
//     }

//     *target=t;

//     /* write overflow from c */
//     if(c>=0) {
//         if(cnv!=NULL) {
//             int8_t i=0;
//             U16_APPEND_UNSAFE(cnv->UCharErrorBuffer, i, c);
//             cnv->UCharErrorBufferLength=i;
//         }
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//     }
// }

// #endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// ******************************************************************************
// *
// *   Copyright (C) 1999-2015, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *
// ******************************************************************************
// *
// *
// *  ucnv_io.cpp:
// *  initializes global variables and defines functions pertaining to converter
// *  name resolution aspect of the conversion code.
// *
// *   new implementation:
// *
// *   created on: 1999nov22
// *   created by: Markus W. Scherer
// *
// *   Use the binary cnvalias.icu (created from convrtrs.txt) to work
// *   with aliases for converter names.
// *
// *   Date        Name        Description
// *   11/22/1999  markus      Created
// *   06/28/2002  grhoten     Major overhaul of the converter alias design.
// *                           Now an alias can map to different converters
// *                           depending on the specified standard.
// *******************************************************************************
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "unicode/ucnv.h"
// #include "unicode/udata.h"

// #include "umutex.h"
// #include "uarrsort.h"
// #include "uassert.h"
// #include "udataswp.h"
// #include "cstring.h"
// #include "cmemory.h"
// #include "ucnv_io.h"
// #include "uenumimp.h"
// #include "ucln_cmn.h"

// /* Format of cnvalias.icu -----------------------------------------------------
//  *
//  * cnvalias.icu is a binary, memory-mappable form of convrtrs.txt.
//  * This binary form contains several tables. All indexes are to uint16_t
//  * units, and not to the bytes (uint8_t units). Addressing everything on
//  * 16-bit boundaries allows us to store more information with small index
//  * numbers, which are also 16-bit in size. The majority of the table (except
//  * the string table) are 16-bit numbers.
//  *
//  * First there is the size of the Table of Contents (TOC). The TOC
//  * entries contain the size of each section. In order to find the offset
//  * you just need to sum up the previous offsets.
//  * The TOC length and entries are an array of uint32_t values.
//  * The first section after the TOC starts immediately after the TOC.
//  *
//  * 1) This section contains a list of converters. This list contains indexes
//  * into the string table for the converter name. The index of this list is
//  * also used by other sections, which are mentioned later on.
//  * This list is not sorted.
//  *
//  * 2) This section contains a list of tags. This list contains indexes
//  * into the string table for the tag name. The index of this list is
//  * also used by other sections, which are mentioned later on.
//  * This list is in priority order of standards.
//  *
//  * 3) This section contains a list of sorted unique aliases. This
//  * list contains indexes into the string table for the alias name. The
//  * index of this list is also used by other sections, like the 4th section.
//  * The index for the 3rd and 4th section is used to get the
//  * alias -> converter name mapping. Section 3 and 4 form a two column table.
//  * Some of the most significant bits of each index may contain other
//  * information (see findConverter for details).
//  *
//  * 4) This section contains a list of mapped converter names. Consider this
//  * as a table that maps the 3rd section to the 1st section. This list contains
//  * indexes into the 1st section. The index of this list is the same index in
//  * the 3rd section. There is also some extra information in the high bits of
//  * each converter index in this table. Currently it's only used to say that
//  * an alias mapped to this converter is ambiguous. See UCNV_CONVERTER_INDEX_MASK
//  * and UCNV_AMBIGUOUS_ALIAS_MAP_BIT for more information. This section is
//  * the predigested form of the 5th section so that an alias lookup can be fast.
//  *
//  * 5) This section contains a 2D array with indexes to the 6th section. This
//  * section is the full form of all alias mappings. The column index is the
//  * index into the converter list (column header). The row index is the index
//  * to tag list (row header). This 2D array is the top part a 3D array. The
//  * third dimension is in the 6th section.
//  *
//  * 6) This is blob of variable length arrays. Each array starts with a size,
//  * and is followed by indexes to alias names in the string table. This is
//  * the third dimension to the section 5. No other section should be referencing
//  * this section.
//  *
//  * 7) Starting in ICU 3.6, this can be a UConverterAliasOptions struct. Its
//  * presence indicates that a section 9 exists. UConverterAliasOptions specifies
//  * what type of string normalization is used among other potential things in the
//  * future.
//  *
//  * 8) This is the string table. All strings are indexed on an even address.
//  * There are two reasons for this. First many chip architectures locate strings
//  * faster on even address boundaries. Second, since all indexes are 16-bit
//  * numbers, this string table can be 128KB in size instead of 64KB when we
//  * only have strings starting on an even address.
//  *
//  * 9) When present this is a set of prenormalized strings from section 8. This
//  * table contains normalized strings with the dashes and spaces stripped out,
//  * and all strings lowercased. In the future, the options in section 7 may state
//  * other types of normalization.
//  *
//  * Here is the concept of section 5 and 6. It's a 3D cube. Each tag
//  * has a unique alias among all converters. That same alias can
//  * be mentioned in other standards on different converters,
//  * but only one alias per tag can be unique.
//  *
//  *
//  *              Converter Names (Usually in TR22 form)
//  *           -------------------------------------------.
//  *     T    /                                          /|
//  *     a   /                                          / |
//  *     g  /                                          /  |
//  *     s /                                          /   |
//  *      /                                          /    |
//  *      ------------------------------------------/     |
//  *    A |                                         |     |
//  *    l |                                         |     |
//  *    i |                                         |    /
//  *    a |                                         |   /
//  *    s |                                         |  /
//  *    e |                                         | /
//  *    s |                                         |/
//  *      -------------------------------------------
//  *
//  *
//  *
//  * Here is what it really looks like. It's like swiss cheese.
//  * There are holes. Some converters aren't recognized by
//  * a standard, or they are really old converters that the
//  * standard doesn't recognize anymore.
//  *
//  *              Converter Names (Usually in TR22 form)
//  *           -------------------------------------------.
//  *     T    /##########################################/|
//  *     a   /     #            #                       /#
//  *     g  /  #      ##     ##     ### # ### ### ### #/
//  *     s / #             #####  ####        ##  ## #/#
//  *      / ### # # ##  #  #   #          ### # #   #/##
//  *      ------------------------------------------/# #
//  *    A |### # # ##  #  #   #          ### # #   #|# #
//  *    l |# # #    #     #               ## #     #|# #
//  *    i |# # #    #     #                #       #|#
//  *    a |#                                       #|#
//  *    s |                                        #|#
//  *    e
//  *    s
//  *
//  */

// /**
//  * Used by the UEnumeration API
//  */
// typedef struct UAliasContext {
//     uint32_t listOffset;
//     uint32_t listIdx;
// } UAliasContext;

// static const char DATA_NAME[] = "cnvalias";
// static const char DATA_TYPE[] = "icu";

// static UDataMemory *gAliasData=NULL;
// static icu::UInitOnce gAliasDataInitOnce = U_INITONCE_INITIALIZER;

// enum {
//     tocLengthIndex=0,
//     converterListIndex=1,
//     tagListIndex=2,
//     aliasListIndex=3,
//     untaggedConvArrayIndex=4,
//     taggedAliasArrayIndex=5,
//     taggedAliasListsIndex=6,
//     tableOptionsIndex=7,
//     stringTableIndex=8,
//     normalizedStringTableIndex=9,
//     offsetsCount,    /* length of the swapper's temporary offsets[] */
//     minTocLength=8 /* min. tocLength in the file, does not count the tocLengthIndex! */
// };

// static const UConverterAliasOptions defaultTableOptions = {
//     UCNV_IO_UNNORMALIZED,
//     0 /* containsCnvOptionInfo */
// };
// static UConverterAlias gMainTable;

// #define GET_STRING(idx) (const char *)(gMainTable.stringTable + (idx))
// #define GET_NORMALIZED_STRING(idx) (const char *)(gMainTable.normalizedStringTable + (idx))

// static UBool U_CALLCONV
// isAcceptable(void * /*context*/,
//              const char * /*type*/, const char * /*name*/,
//              const UDataInfo *pInfo) {
//     return (UBool)(
//         pInfo->size>=20 &&
//         pInfo->isBigEndian==U_IS_BIG_ENDIAN &&
//         pInfo->charsetFamily==U_CHARSET_FAMILY &&
//         pInfo->dataFormat[0]==0x43 &&   /* dataFormat="CvAl" */
//         pInfo->dataFormat[1]==0x76 &&
//         pInfo->dataFormat[2]==0x41 &&
//         pInfo->dataFormat[3]==0x6c &&
//         pInfo->formatVersion[0]==3);
// }

// static UBool U_CALLCONV ucnv_io_cleanup(void)
// {
//     if (gAliasData) {
//         udata_close(gAliasData);
//         gAliasData = NULL;
//     }
//     gAliasDataInitOnce.reset();

//     uprv_memset(&gMainTable, 0, sizeof(gMainTable));

//     return TRUE;                   /* Everything was cleaned up */
// }

// static void U_CALLCONV initAliasData(UErrorCode &errCode) {
//     UDataMemory *data;
//     const uint16_t *table;
//     const uint32_t *sectionSizes;
//     uint32_t tableStart;
//     uint32_t currOffset;

//     ucln_common_registerCleanup(UCLN_COMMON_UCNV_IO, ucnv_io_cleanup);

//     U_ASSERT(gAliasData == NULL);
//     data = udata_openChoice(NULL, DATA_TYPE, DATA_NAME, isAcceptable, NULL, &errCode);
//     if(U_FAILURE(errCode)) {
//         return;
//     }

//     sectionSizes = (const uint32_t *)udata_getMemory(data);
//     table = (const uint16_t *)sectionSizes;

//     tableStart      = sectionSizes[0];
//     if (tableStart < minTocLength) {
//         errCode = U_INVALID_FORMAT_ERROR;
//         udata_close(data);
//         return;
//     }
//     gAliasData = data;

//     gMainTable.converterListSize      = sectionSizes[1];
//     gMainTable.tagListSize            = sectionSizes[2];
//     gMainTable.aliasListSize          = sectionSizes[3];
//     gMainTable.untaggedConvArraySize  = sectionSizes[4];
//     gMainTable.taggedAliasArraySize   = sectionSizes[5];
//     gMainTable.taggedAliasListsSize   = sectionSizes[6];
//     gMainTable.optionTableSize        = sectionSizes[7];
//     gMainTable.stringTableSize        = sectionSizes[8];

//     if (tableStart > 8) {
//         gMainTable.normalizedStringTableSize = sectionSizes[9];
//     }

//     currOffset = tableStart * (sizeof(uint32_t)/sizeof(uint16_t)) + (sizeof(uint32_t)/sizeof(uint16_t));
//     gMainTable.converterList = table + currOffset;

//     currOffset += gMainTable.converterListSize;
//     gMainTable.tagList = table + currOffset;

//     currOffset += gMainTable.tagListSize;
//     gMainTable.aliasList = table + currOffset;

//     currOffset += gMainTable.aliasListSize;
//     gMainTable.untaggedConvArray = table + currOffset;

//     currOffset += gMainTable.untaggedConvArraySize;
//     gMainTable.taggedAliasArray = table + currOffset;

//     /* aliasLists is a 1's based array, but it has a padding character */
//     currOffset += gMainTable.taggedAliasArraySize;
//     gMainTable.taggedAliasLists = table + currOffset;

//     currOffset += gMainTable.taggedAliasListsSize;
//     if (gMainTable.optionTableSize > 0
//         && ((const UConverterAliasOptions *)(table + currOffset))->stringNormalizationType < UCNV_IO_NORM_TYPE_COUNT)
//     {
//         /* Faster table */
//         gMainTable.optionTable = (const UConverterAliasOptions *)(table + currOffset);
//     }
//     else {
//         /* Smaller table, or I can't handle this normalization mode!
//         Use the original slower table lookup. */
//         gMainTable.optionTable = &defaultTableOptions;
//     }

//     currOffset += gMainTable.optionTableSize;
//     gMainTable.stringTable = table + currOffset;

//     currOffset += gMainTable.stringTableSize;
//     gMainTable.normalizedStringTable = ((gMainTable.optionTable->stringNormalizationType == UCNV_IO_UNNORMALIZED)
//         ? gMainTable.stringTable : (table + currOffset));
// }


// static UBool
// haveAliasData(UErrorCode *pErrorCode) {
//     umtx_initOnce(gAliasDataInitOnce, &initAliasData, *pErrorCode);
//     return U_SUCCESS(*pErrorCode);
// }

// static inline UBool
// isAlias(const char *alias, UErrorCode *pErrorCode) {
//     if(alias==NULL) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return FALSE;
//     }
//     return (UBool)(*alias!=0);
// }

// static uint32_t getTagNumber(const char *tagname) {
//     if (gMainTable.tagList) {
//         uint32_t tagNum;
//         for (tagNum = 0; tagNum < gMainTable.tagListSize; tagNum++) {
//             if (!uprv_stricmp(GET_STRING(gMainTable.tagList[tagNum]), tagname)) {
//                 return tagNum;
//             }
//         }
//     }

//     return UINT32_MAX;
// }

// /* character types relevant for ucnv_compareNames() */
// enum {
//     UIGNORE,
//     ZERO,
//     NONZERO,
//     MINLETTER /* any values from here on are lowercase letter mappings */
// };

// /* character types for ASCII 00..7F */
// static const uint8_t asciiTypes[128] = {
//     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
//     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
//     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
//     ZERO, NONZERO, NONZERO, NONZERO, NONZERO, NONZERO, NONZERO, NONZERO, NONZERO, NONZERO, 0, 0, 0, 0, 0, 0,
//     0, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
//     0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0, 0, 0, 0, 0,
//     0, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
//     0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0, 0, 0, 0, 0
// };

// #define GET_ASCII_TYPE(c) ((int8_t)(c) >= 0 ? asciiTypes[(uint8_t)c] : (uint8_t)UIGNORE)

// /* character types for EBCDIC 80..FF */
// static const uint8_t ebcdicTypes[128] = {
//     0,    0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0, 0, 0, 0, 0, 0,
//     0,    0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0, 0, 0, 0, 0, 0,
//     0,    0,    0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0, 0, 0, 0, 0, 0,
//     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
//     0,    0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0, 0, 0, 0, 0, 0,
//     0,    0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0, 0, 0, 0, 0, 0,
//     0,    0,    0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0, 0, 0, 0, 0, 0,
//     ZERO, NONZERO, NONZERO, NONZERO, NONZERO, NONZERO, NONZERO, NONZERO, NONZERO, NONZERO, 0, 0, 0, 0, 0, 0
// };

// #define GET_EBCDIC_TYPE(c) ((int8_t)(c) < 0 ? ebcdicTypes[(c)&0x7f] : (uint8_t)UIGNORE)

// #if U_CHARSET_FAMILY==U_ASCII_FAMILY
// #   define GET_CHAR_TYPE(c) GET_ASCII_TYPE(c)
// #elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
// #   define GET_CHAR_TYPE(c) GET_EBCDIC_TYPE(c)
// #else
// #   error U_CHARSET_FAMILY is not valid
// #endif


// /* @see ucnv_compareNames */
// U_CAPI char * U_CALLCONV
// ucnv_io_stripASCIIForCompare(char *dst, const char *name) {
//     char *dstItr = dst;
//     uint8_t type, nextType;
//     char c1;
//     UBool afterDigit = FALSE;

//     while ((c1 = *name++) != 0) {
//         type = GET_ASCII_TYPE(c1);
//         switch (type) {
//         case UIGNORE:
//             afterDigit = FALSE;
//             continue; /* ignore all but letters and digits */
//         case ZERO:
//             if (!afterDigit) {
//                 nextType = GET_ASCII_TYPE(*name);
//                 if (nextType == ZERO || nextType == NONZERO) {
//                     continue; /* ignore leading zero before another digit */
//                 }
//             }
//             break;
//         case NONZERO:
//             afterDigit = TRUE;
//             break;
//         default:
//             c1 = (char)type; /* lowercased letter */
//             afterDigit = FALSE;
//             break;
//         }
//         *dstItr++ = c1;
//     }
//     *dstItr = 0;
//     return dst;
// }

// U_CAPI char * U_CALLCONV
// ucnv_io_stripEBCDICForCompare(char *dst, const char *name) {
//     char *dstItr = dst;
//     uint8_t type, nextType;
//     char c1;
//     UBool afterDigit = FALSE;

//     while ((c1 = *name++) != 0) {
//         type = GET_EBCDIC_TYPE(c1);
//         switch (type) {
//         case UIGNORE:
//             afterDigit = FALSE;
//             continue; /* ignore all but letters and digits */
//         case ZERO:
//             if (!afterDigit) {
//                 nextType = GET_EBCDIC_TYPE(*name);
//                 if (nextType == ZERO || nextType == NONZERO) {
//                     continue; /* ignore leading zero before another digit */
//                 }
//             }
//             break;
//         case NONZERO:
//             afterDigit = TRUE;
//             break;
//         default:
//             c1 = (char)type; /* lowercased letter */
//             afterDigit = FALSE;
//             break;
//         }
//         *dstItr++ = c1;
//     }
//     *dstItr = 0;
//     return dst;
// }

// /**
//  * Do a fuzzy compare of two converter/alias names.
//  * The comparison is case-insensitive, ignores leading zeroes if they are not
//  * followed by further digits, and ignores all but letters and digits.
//  * Thus the strings "UTF-8", "utf_8", "u*T@f08" and "Utf 8" are exactly equivalent.
//  * See section 1.4, Charset Alias Matching in Unicode Technical Standard #22
//  * at http://www.unicode.org/reports/tr22/
//  *
//  * This is a symmetrical (commutative) operation; order of arguments
//  * is insignificant.  This is an important property for sorting the
//  * list (when the list is preprocessed into binary form) and for
//  * performing binary searches on it at run time.
//  *
//  * @param name1 a converter name or alias, zero-terminated
//  * @param name2 a converter name or alias, zero-terminated
//  * @return 0 if the names match, or a negative value if the name1
//  * lexically precedes name2, or a positive value if the name1
//  * lexically follows name2.
//  *
//  * @see ucnv_io_stripForCompare
//  */
// U_CAPI int U_EXPORT2
// ucnv_compareNames(const char *name1, const char *name2) {
//     int rc;
//     uint8_t type, nextType;
//     char c1, c2;
//     UBool afterDigit1 = FALSE, afterDigit2 = FALSE;

//     for (;;) {
//         while ((c1 = *name1++) != 0) {
//             type = GET_CHAR_TYPE(c1);
//             switch (type) {
//             case UIGNORE:
//                 afterDigit1 = FALSE;
//                 continue; /* ignore all but letters and digits */
//             case ZERO:
//                 if (!afterDigit1) {
//                     nextType = GET_CHAR_TYPE(*name1);
//                     if (nextType == ZERO || nextType == NONZERO) {
//                         continue; /* ignore leading zero before another digit */
//                     }
//                 }
//                 break;
//             case NONZERO:
//                 afterDigit1 = TRUE;
//                 break;
//             default:
//                 c1 = (char)type; /* lowercased letter */
//                 afterDigit1 = FALSE;
//                 break;
//             }
//             break; /* deliver c1 */
//         }
//         while ((c2 = *name2++) != 0) {
//             type = GET_CHAR_TYPE(c2);
//             switch (type) {
//             case UIGNORE:
//                 afterDigit2 = FALSE;
//                 continue; /* ignore all but letters and digits */
//             case ZERO:
//                 if (!afterDigit2) {
//                     nextType = GET_CHAR_TYPE(*name2);
//                     if (nextType == ZERO || nextType == NONZERO) {
//                         continue; /* ignore leading zero before another digit */
//                     }
//                 }
//                 break;
//             case NONZERO:
//                 afterDigit2 = TRUE;
//                 break;
//             default:
//                 c2 = (char)type; /* lowercased letter */
//                 afterDigit2 = FALSE;
//                 break;
//             }
//             break; /* deliver c2 */
//         }

//         /* If we reach the ends of both strings then they match */
//         if ((c1|c2)==0) {
//             return 0;
//         }

//         /* Case-insensitive comparison */
//         rc = (int)(unsigned char)c1 - (int)(unsigned char)c2;
//         if (rc != 0) {
//             return rc;
//         }
//     }
// }

// /*
//  * search for an alias
//  * return the converter number index for gConverterList
//  */
// static inline uint32_t
// findConverter(const char *alias, UBool *containsOption, UErrorCode *pErrorCode) {
//     uint32_t mid, start, limit;
//     uint32_t lastMid;
//     int result;
//     int isUnnormalized = (gMainTable.optionTable->stringNormalizationType == UCNV_IO_UNNORMALIZED);
//     char strippedName[UCNV_MAX_CONVERTER_NAME_LENGTH];

//     if (!isUnnormalized) {
//         if (uprv_strlen(alias) >= UCNV_MAX_CONVERTER_NAME_LENGTH) {
//             *pErrorCode = U_BUFFER_OVERFLOW_ERROR;
//             return UINT32_MAX;
//         }

//         /* Lower case and remove ignoreable characters. */
//         ucnv_io_stripForCompare(strippedName, alias);
//         alias = strippedName;
//     }

//     /* do a binary search for the alias */
//     start = 0;
//     limit = gMainTable.untaggedConvArraySize;
//     mid = limit;
//     lastMid = UINT32_MAX;

//     for (;;) {
//         mid = (uint32_t)((start + limit) / 2);
//         if (lastMid == mid) {   /* Have we moved? */
//             break;  /* We haven't moved, and it wasn't found. */
//         }
//         lastMid = mid;
//         if (isUnnormalized) {
//             result = ucnv_compareNames(alias, GET_STRING(gMainTable.aliasList[mid]));
//         }
//         else {
//             result = uprv_strcmp(alias, GET_NORMALIZED_STRING(gMainTable.aliasList[mid]));
//         }

//         if (result < 0) {
//             limit = mid;
//         } else if (result > 0) {
//             start = mid;
//         } else {
//             /* Since the gencnval tool folds duplicates into one entry,
//              * this alias in gAliasList is unique, but different standards
//              * may map an alias to different converters.
//              */
//             if (gMainTable.untaggedConvArray[mid] & UCNV_AMBIGUOUS_ALIAS_MAP_BIT) {
//                 *pErrorCode = U_AMBIGUOUS_ALIAS_WARNING;
//             }
//             /* State whether the canonical converter name contains an option.
//             This information is contained in this list in order to maintain backward & forward compatibility. */
//             if (containsOption) {
//                 UBool containsCnvOptionInfo = (UBool)gMainTable.optionTable->containsCnvOptionInfo;
//                 *containsOption = (UBool)((containsCnvOptionInfo
//                     && ((gMainTable.untaggedConvArray[mid] & UCNV_CONTAINS_OPTION_BIT) != 0))
//                     || !containsCnvOptionInfo);
//             }
//             return gMainTable.untaggedConvArray[mid] & UCNV_CONVERTER_INDEX_MASK;
//         }
//     }

//     return UINT32_MAX;
// }

// /*
//  * Is this alias in this list?
//  * alias and listOffset should be non-NULL.
//  */
// static inline UBool
// isAliasInList(const char *alias, uint32_t listOffset) {
//     if (listOffset) {
//         uint32_t currAlias;
//         uint32_t listCount = gMainTable.taggedAliasLists[listOffset];
//         /* +1 to skip listCount */
//         const uint16_t *currList = gMainTable.taggedAliasLists + listOffset + 1;
//         for (currAlias = 0; currAlias < listCount; currAlias++) {
//             if (currList[currAlias]
//                 && ucnv_compareNames(alias, GET_STRING(currList[currAlias]))==0)
//             {
//                 return TRUE;
//             }
//         }
//     }
//     return FALSE;
// }

// /*
//  * Search for an standard name of an alias (what is the default name
//  * that this standard uses?)
//  * return the listOffset for gTaggedAliasLists. If it's 0,
//  * the it couldn't be found, but the parameters are valid.
//  */
// static uint32_t
// findTaggedAliasListsOffset(const char *alias, const char *standard, UErrorCode *pErrorCode) {
//     uint32_t idx;
//     uint32_t listOffset;
//     uint32_t convNum;
//     UErrorCode myErr = U_ZERO_ERROR;
//     uint32_t tagNum = getTagNumber(standard);

//     /* Make a quick guess. Hopefully they used a TR22 canonical alias. */
//     convNum = findConverter(alias, NULL, &myErr);
//     if (myErr != U_ZERO_ERROR) {
//         *pErrorCode = myErr;
//     }

//     if (tagNum < (gMainTable.tagListSize - UCNV_NUM_HIDDEN_TAGS) && convNum < gMainTable.converterListSize) {
//         listOffset = gMainTable.taggedAliasArray[tagNum*gMainTable.converterListSize + convNum];
//         if (listOffset && gMainTable.taggedAliasLists[listOffset + 1]) {
//             return listOffset;
//         }
//         if (myErr == U_AMBIGUOUS_ALIAS_WARNING) {
//             /* Uh Oh! They used an ambiguous alias.
//                We have to search the whole swiss cheese starting
//                at the highest standard affinity.
//                This may take a while.
//             */
//             for (idx = 0; idx < gMainTable.taggedAliasArraySize; idx++) {
//                 listOffset = gMainTable.taggedAliasArray[idx];
//                 if (listOffset && isAliasInList(alias, listOffset)) {
//                     uint32_t currTagNum = idx/gMainTable.converterListSize;
//                     uint32_t currConvNum = (idx - currTagNum*gMainTable.converterListSize);
//                     uint32_t tempListOffset = gMainTable.taggedAliasArray[tagNum*gMainTable.converterListSize + currConvNum];
//                     if (tempListOffset && gMainTable.taggedAliasLists[tempListOffset + 1]) {
//                         return tempListOffset;
//                     }
//                     /* else keep on looking */
//                     /* We could speed this up by starting on the next row
//                        because an alias is unique per row, right now.
//                        This would change if alias versioning appears. */
//                 }
//             }
//             /* The standard doesn't know about the alias */
//         }
//         /* else no default name */
//         return 0;
//     }
//     /* else converter or tag not found */

//     return UINT32_MAX;
// }

// /* Return the canonical name */
// static uint32_t
// findTaggedConverterNum(const char *alias, const char *standard, UErrorCode *pErrorCode) {
//     uint32_t idx;
//     uint32_t listOffset;
//     uint32_t convNum;
//     UErrorCode myErr = U_ZERO_ERROR;
//     uint32_t tagNum = getTagNumber(standard);

//     /* Make a quick guess. Hopefully they used a TR22 canonical alias. */
//     convNum = findConverter(alias, NULL, &myErr);
//     if (myErr != U_ZERO_ERROR) {
//         *pErrorCode = myErr;
//     }

//     if (tagNum < (gMainTable.tagListSize - UCNV_NUM_HIDDEN_TAGS) && convNum < gMainTable.converterListSize) {
//         listOffset = gMainTable.taggedAliasArray[tagNum*gMainTable.converterListSize + convNum];
//         if (listOffset && isAliasInList(alias, listOffset)) {
//             return convNum;
//         }
//         if (myErr == U_AMBIGUOUS_ALIAS_WARNING) {
//             /* Uh Oh! They used an ambiguous alias.
//                We have to search one slice of the swiss cheese.
//                We search only in the requested tag, not the whole thing.
//                This may take a while.
//             */
//             uint32_t convStart = (tagNum)*gMainTable.converterListSize;
//             uint32_t convLimit = (tagNum+1)*gMainTable.converterListSize;
//             for (idx = convStart; idx < convLimit; idx++) {
//                 listOffset = gMainTable.taggedAliasArray[idx];
//                 if (listOffset && isAliasInList(alias, listOffset)) {
//                     return idx-convStart;
//                 }
//             }
//             /* The standard doesn't know about the alias */
//         }
//         /* else no canonical name */
//     }
//     /* else converter or tag not found */

//     return UINT32_MAX;
// }

// U_CAPI const char *
// ucnv_io_getConverterName(const char *alias, UBool *containsOption, UErrorCode *pErrorCode) {
//     const char *aliasTmp = alias;
//     int32_t i = 0;
//     for (i = 0; i < 2; i++) {
//         if (i == 1) {
//             /*
//              * After the first unsuccess converter lookup, check to see if
//              * the name begins with 'x-'. If it does, strip it off and try
//              * again.  This behaviour is similar to how ICU4J does it.
//              */
//             if (aliasTmp[0] == 'x' && aliasTmp[1] == '-') {
//                 aliasTmp = aliasTmp+2;
//             } else {
//                 break;
//             }
//         }
//         if(haveAliasData(pErrorCode) && isAlias(aliasTmp, pErrorCode)) {
//             uint32_t convNum = findConverter(aliasTmp, containsOption, pErrorCode);
//             if (convNum < gMainTable.converterListSize) {
//                 return GET_STRING(gMainTable.converterList[convNum]);
//             }
//             /* else converter not found */
//         } else {
//             break;
//         }
//     }

//     return NULL;
// }

// U_CDECL_BEGIN


// static int32_t U_CALLCONV
// ucnv_io_countStandardAliases(UEnumeration *enumerator, UErrorCode * /*pErrorCode*/) {
//     int32_t value = 0;
//     UAliasContext *myContext = (UAliasContext *)(enumerator->context);
//     uint32_t listOffset = myContext->listOffset;

//     if (listOffset) {
//         value = gMainTable.taggedAliasLists[listOffset];
//     }
//     return value;
// }

// static const char * U_CALLCONV
// ucnv_io_nextStandardAliases(UEnumeration *enumerator,
//                             int32_t* resultLength,
//                             UErrorCode * /*pErrorCode*/)
// {
//     UAliasContext *myContext = (UAliasContext *)(enumerator->context);
//     uint32_t listOffset = myContext->listOffset;

//     if (listOffset) {
//         uint32_t listCount = gMainTable.taggedAliasLists[listOffset];
//         const uint16_t *currList = gMainTable.taggedAliasLists + listOffset + 1;

//         if (myContext->listIdx < listCount) {
//             const char *myStr = GET_STRING(currList[myContext->listIdx++]);
//             if (resultLength) {
//                 *resultLength = (int32_t)uprv_strlen(myStr);
//             }
//             return myStr;
//         }
//     }
//     /* Either we accessed a zero length list, or we enumerated too far. */
//     if (resultLength) {
//         *resultLength = 0;
//     }
//     return NULL;
// }

// static void U_CALLCONV
// ucnv_io_resetStandardAliases(UEnumeration *enumerator, UErrorCode * /*pErrorCode*/) {
//     ((UAliasContext *)(enumerator->context))->listIdx = 0;
// }

// static void U_CALLCONV
// ucnv_io_closeUEnumeration(UEnumeration *enumerator) {
//     uprv_free(enumerator->context);
//     uprv_free(enumerator);
// }

// U_CDECL_END

// /* Enumerate the aliases for the specified converter and standard tag */
// static const UEnumeration gEnumAliases = {
//     NULL,
//     NULL,
//     ucnv_io_closeUEnumeration,
//     ucnv_io_countStandardAliases,
//     uenum_unextDefault,
//     ucnv_io_nextStandardAliases,
//     ucnv_io_resetStandardAliases
// };

// U_CAPI UEnumeration * U_EXPORT2
// ucnv_openStandardNames(const char *convName,
//                        const char *standard,
//                        UErrorCode *pErrorCode)
// {
//     UEnumeration *myEnum = NULL;
//     if (haveAliasData(pErrorCode) && isAlias(convName, pErrorCode)) {
//         uint32_t listOffset = findTaggedAliasListsOffset(convName, standard, pErrorCode);

//         /* When listOffset == 0, we want to acknowledge that the
//            converter name and standard are okay, but there
//            is nothing to enumerate. */
//         if (listOffset < gMainTable.taggedAliasListsSize) {
//             UAliasContext *myContext;

//             myEnum = static_cast<UEnumeration *>(uprv_malloc(sizeof(UEnumeration)));
//             if (myEnum == NULL) {
//                 *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
//                 return NULL;
//             }
//             uprv_memcpy(myEnum, &gEnumAliases, sizeof(UEnumeration));
//             myContext = static_cast<UAliasContext *>(uprv_malloc(sizeof(UAliasContext)));
//             if (myContext == NULL) {
//                 *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
//                 uprv_free(myEnum);
//                 return NULL;
//             }
//             myContext->listOffset = listOffset;
//             myContext->listIdx = 0;
//             myEnum->context = myContext;
//         }
//         /* else converter or tag not found */
//     }
//     return myEnum;
// }

// static uint16_t
// ucnv_io_countAliases(const char *alias, UErrorCode *pErrorCode) {
//     if(haveAliasData(pErrorCode) && isAlias(alias, pErrorCode)) {
//         uint32_t convNum = findConverter(alias, NULL, pErrorCode);
//         if (convNum < gMainTable.converterListSize) {
//             /* tagListNum - 1 is the ALL tag */
//             int32_t listOffset = gMainTable.taggedAliasArray[(gMainTable.tagListSize - 1)*gMainTable.converterListSize + convNum];

//             if (listOffset) {
//                 return gMainTable.taggedAliasLists[listOffset];
//             }
//             /* else this shouldn't happen. internal program error */
//         }
//         /* else converter not found */
//     }
//     return 0;
// }

// static uint16_t
// ucnv_io_getAliases(const char *alias, uint16_t start, const char **aliases, UErrorCode *pErrorCode) {
//     if(haveAliasData(pErrorCode) && isAlias(alias, pErrorCode)) {
//         uint32_t currAlias;
//         uint32_t convNum = findConverter(alias, NULL, pErrorCode);
//         if (convNum < gMainTable.converterListSize) {
//             /* tagListNum - 1 is the ALL tag */
//             int32_t listOffset = gMainTable.taggedAliasArray[(gMainTable.tagListSize - 1)*gMainTable.converterListSize + convNum];

//             if (listOffset) {
//                 uint32_t listCount = gMainTable.taggedAliasLists[listOffset];
//                 /* +1 to skip listCount */
//                 const uint16_t *currList = gMainTable.taggedAliasLists + listOffset + 1;

//                 for (currAlias = start; currAlias < listCount; currAlias++) {
//                     aliases[currAlias] = GET_STRING(currList[currAlias]);
//                 }
//             }
//             /* else this shouldn't happen. internal program error */
//         }
//         /* else converter not found */
//     }
//     return 0;
// }

// static const char *
// ucnv_io_getAlias(const char *alias, uint16_t n, UErrorCode *pErrorCode) {
//     if(haveAliasData(pErrorCode) && isAlias(alias, pErrorCode)) {
//         uint32_t convNum = findConverter(alias, NULL, pErrorCode);
//         if (convNum < gMainTable.converterListSize) {
//             /* tagListNum - 1 is the ALL tag */
//             int32_t listOffset = gMainTable.taggedAliasArray[(gMainTable.tagListSize - 1)*gMainTable.converterListSize + convNum];

//             if (listOffset) {
//                 uint32_t listCount = gMainTable.taggedAliasLists[listOffset];
//                 /* +1 to skip listCount */
//                 const uint16_t *currList = gMainTable.taggedAliasLists + listOffset + 1;

//                 if (n < listCount)  {
//                     return GET_STRING(currList[n]);
//                 }
//                 *pErrorCode = U_INDEX_OUTOFBOUNDS_ERROR;
//             }
//             /* else this shouldn't happen. internal program error */
//         }
//         /* else converter not found */
//     }
//     return NULL;
// }

// static uint16_t
// ucnv_io_countStandards(UErrorCode *pErrorCode) {
//     if (haveAliasData(pErrorCode)) {
//         /* Don't include the empty list */
//         return (uint16_t)(gMainTable.tagListSize - UCNV_NUM_HIDDEN_TAGS);
//     }

//     return 0;
// }

// U_CAPI const char * U_EXPORT2
// ucnv_getStandard(uint16_t n, UErrorCode *pErrorCode) {
//     if (haveAliasData(pErrorCode)) {
//         if (n < gMainTable.tagListSize - UCNV_NUM_HIDDEN_TAGS) {
//             return GET_STRING(gMainTable.tagList[n]);
//         }
//         *pErrorCode = U_INDEX_OUTOFBOUNDS_ERROR;
//     }

//     return NULL;
// }

// U_CAPI const char * U_EXPORT2
// ucnv_getStandardName(const char *alias, const char *standard, UErrorCode *pErrorCode) {
//     if (haveAliasData(pErrorCode) && isAlias(alias, pErrorCode)) {
//         uint32_t listOffset = findTaggedAliasListsOffset(alias, standard, pErrorCode);

//         if (0 < listOffset && listOffset < gMainTable.taggedAliasListsSize) {
//             const uint16_t *currList = gMainTable.taggedAliasLists + listOffset + 1;

//             /* Get the preferred name from this list */
//             if (currList[0]) {
//                 return GET_STRING(currList[0]);
//             }
//             /* else someone screwed up the alias table. */
//             /* *pErrorCode = U_INVALID_FORMAT_ERROR */
//         }
//     }

//     return NULL;
// }

// U_CAPI uint16_t U_EXPORT2
// ucnv_countAliases(const char *alias, UErrorCode *pErrorCode)
// {
//     return ucnv_io_countAliases(alias, pErrorCode);
// }


// U_CAPI const char* U_EXPORT2
// ucnv_getAlias(const char *alias, uint16_t n, UErrorCode *pErrorCode)
// {
//     return ucnv_io_getAlias(alias, n, pErrorCode);
// }

// U_CAPI void U_EXPORT2
// ucnv_getAliases(const char *alias, const char **aliases, UErrorCode *pErrorCode)
// {
//     ucnv_io_getAliases(alias, 0, aliases, pErrorCode);
// }

// U_CAPI uint16_t U_EXPORT2
// ucnv_countStandards(void)
// {
//     UErrorCode err = U_ZERO_ERROR;
//     return ucnv_io_countStandards(&err);
// }

// U_CAPI const char * U_EXPORT2
// ucnv_getCanonicalName(const char *alias, const char *standard, UErrorCode *pErrorCode) {
//     if (haveAliasData(pErrorCode) && isAlias(alias, pErrorCode)) {
//         uint32_t convNum = findTaggedConverterNum(alias, standard, pErrorCode);

//         if (convNum < gMainTable.converterListSize) {
//             return GET_STRING(gMainTable.converterList[convNum]);
//         }
//     }

//     return NULL;
// }

// U_CDECL_BEGIN


// static int32_t U_CALLCONV
// ucnv_io_countAllConverters(UEnumeration * /*enumerator*/, UErrorCode * /*pErrorCode*/) {
//     return gMainTable.converterListSize;
// }

// static const char * U_CALLCONV
// ucnv_io_nextAllConverters(UEnumeration *enumerator,
//                             int32_t* resultLength,
//                             UErrorCode * /*pErrorCode*/)
// {
//     uint16_t *myContext = (uint16_t *)(enumerator->context);

//     if (*myContext < gMainTable.converterListSize) {
//         const char *myStr = GET_STRING(gMainTable.converterList[(*myContext)++]);
//         if (resultLength) {
//             *resultLength = (int32_t)uprv_strlen(myStr);
//         }
//         return myStr;
//     }
//     /* Either we accessed a zero length list, or we enumerated too far. */
//     if (resultLength) {
//         *resultLength = 0;
//     }
//     return NULL;
// }

// static void U_CALLCONV
// ucnv_io_resetAllConverters(UEnumeration *enumerator, UErrorCode * /*pErrorCode*/) {
//     *((uint16_t *)(enumerator->context)) = 0;
// }
// U_CDECL_END
// static const UEnumeration gEnumAllConverters = {
//     NULL,
//     NULL,
//     ucnv_io_closeUEnumeration,
//     ucnv_io_countAllConverters,
//     uenum_unextDefault,
//     ucnv_io_nextAllConverters,
//     ucnv_io_resetAllConverters
// };

// U_CAPI UEnumeration * U_EXPORT2
// ucnv_openAllNames(UErrorCode *pErrorCode) {
//     UEnumeration *myEnum = NULL;
//     if (haveAliasData(pErrorCode)) {
//         uint16_t *myContext;

//         myEnum = static_cast<UEnumeration *>(uprv_malloc(sizeof(UEnumeration)));
//         if (myEnum == NULL) {
//             *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
//             return NULL;
//         }
//         uprv_memcpy(myEnum, &gEnumAllConverters, sizeof(UEnumeration));
//         myContext = static_cast<uint16_t *>(uprv_malloc(sizeof(uint16_t)));
//         if (myContext == NULL) {
//             *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
//             uprv_free(myEnum);
//             return NULL;
//         }
//         *myContext = 0;
//         myEnum->context = myContext;
//     }
//     return myEnum;
// }

// U_CAPI uint16_t
// ucnv_io_countKnownConverters(UErrorCode *pErrorCode) {
//     if (haveAliasData(pErrorCode)) {
//         return (uint16_t)gMainTable.converterListSize;
//     }
//     return 0;
// }

// /* alias table swapping ----------------------------------------------------- */

// U_CDECL_BEGIN

// typedef char * U_CALLCONV StripForCompareFn(char *dst, const char *name);
// U_CDECL_END


// /*
//  * row of a temporary array
//  *
//  * gets platform-endian charset string indexes and sorting indexes;
//  * after sorting this array by strings, the actual arrays are permutated
//  * according to the sorting indexes
//  */
// typedef struct TempRow {
//     uint16_t strIndex, sortIndex;
// } TempRow;

// typedef struct TempAliasTable {
//     const char *chars;
//     TempRow *rows;
//     uint16_t *resort;
//     StripForCompareFn *stripForCompare;
// } TempAliasTable;

// enum {
//     STACK_ROW_CAPACITY=500
// };

// static int32_t U_CALLCONV
// io_compareRows(const void *context, const void *left, const void *right) {
//     char strippedLeft[UCNV_MAX_CONVERTER_NAME_LENGTH],
//          strippedRight[UCNV_MAX_CONVERTER_NAME_LENGTH];

//     TempAliasTable *tempTable=(TempAliasTable *)context;
//     const char *chars=tempTable->chars;

//     return (int32_t)uprv_strcmp(tempTable->stripForCompare(strippedLeft, chars+2*((const TempRow *)left)->strIndex),
//                                 tempTable->stripForCompare(strippedRight, chars+2*((const TempRow *)right)->strIndex));
// }

// U_CAPI int32_t U_EXPORT2
// ucnv_swapAliases(const UDataSwapper *ds,
//                  const void *inData, int32_t length, void *outData,
//                  UErrorCode *pErrorCode) {
//     const UDataInfo *pInfo;
//     int32_t headerSize;

//     const uint16_t *inTable;
//     const uint32_t *inSectionSizes;
//     uint32_t toc[offsetsCount];
//     uint32_t offsets[offsetsCount]; /* 16-bit-addressed offsets from inTable/outTable */
//     uint32_t i, count, tocLength, topOffset;

//     TempRow rows[STACK_ROW_CAPACITY];
//     uint16_t resort[STACK_ROW_CAPACITY];
//     TempAliasTable tempTable;

//     /* udata_swapDataHeader checks the arguments */
//     headerSize=udata_swapDataHeader(ds, inData, length, outData, pErrorCode);
//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     /* check data format and format version */
//     pInfo=(const UDataInfo *)((const char *)inData+4);
//     if(!(
//         pInfo->dataFormat[0]==0x43 &&   /* dataFormat="CvAl" */
//         pInfo->dataFormat[1]==0x76 &&
//         pInfo->dataFormat[2]==0x41 &&
//         pInfo->dataFormat[3]==0x6c &&
//         pInfo->formatVersion[0]==3
//     )) {
//         udata_printError(ds, "ucnv_swapAliases(): data format %02x.%02x.%02x.%02x (format version %02x) is not an alias table\n",
//                          pInfo->dataFormat[0], pInfo->dataFormat[1],
//                          pInfo->dataFormat[2], pInfo->dataFormat[3],
//                          pInfo->formatVersion[0]);
//         *pErrorCode=U_UNSUPPORTED_ERROR;
//         return 0;
//     }

//     /* an alias table must contain at least the table of contents array */
//     if(length>=0 && (length-headerSize)<4*(1+minTocLength)) {
//         udata_printError(ds, "ucnv_swapAliases(): too few bytes (%d after header) for an alias table\n",
//                          length-headerSize);
//         *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//         return 0;
//     }

//     inSectionSizes=(const uint32_t *)((const char *)inData+headerSize);
//     inTable=(const uint16_t *)inSectionSizes;
//     uprv_memset(toc, 0, sizeof(toc));
//     toc[tocLengthIndex]=tocLength=ds->readUInt32(inSectionSizes[tocLengthIndex]);
//     if(tocLength<minTocLength || offsetsCount<=tocLength) {
//         udata_printError(ds, "ucnv_swapAliases(): table of contents contains unsupported number of sections (%u sections)\n", tocLength);
//         *pErrorCode=U_INVALID_FORMAT_ERROR;
//         return 0;
//     }

//     /* read the known part of the table of contents */
//     for(i=converterListIndex; i<=tocLength; ++i) {
//         toc[i]=ds->readUInt32(inSectionSizes[i]);
//     }

//     /* compute offsets */
//     uprv_memset(offsets, 0, sizeof(offsets));
//     offsets[converterListIndex]=2*(1+tocLength); /* count two 16-bit units per toc entry */
//     for(i=tagListIndex; i<=tocLength; ++i) {
//         offsets[i]=offsets[i-1]+toc[i-1];
//     }

//     /* compute the overall size of the after-header data, in numbers of 16-bit units */
//     topOffset=offsets[i-1]+toc[i-1];

//     if(length>=0) {
//         uint16_t *outTable;
//         const uint16_t *p, *p2;
//         uint16_t *q, *q2;
//         uint16_t oldIndex;

//         if((length-headerSize)<(2*(int32_t)topOffset)) {
//             udata_printError(ds, "ucnv_swapAliases(): too few bytes (%d after header) for an alias table\n",
//                              length-headerSize);
//             *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//             return 0;
//         }

//         outTable=(uint16_t *)((char *)outData+headerSize);

//         /* swap the entire table of contents */
//         ds->swapArray32(ds, inTable, 4*(1+tocLength), outTable, pErrorCode);

//         /* swap unormalized strings & normalized strings */
//         ds->swapInvChars(ds, inTable+offsets[stringTableIndex], 2*(int32_t)(toc[stringTableIndex]+toc[normalizedStringTableIndex]),
//                              outTable+offsets[stringTableIndex], pErrorCode);
//         if(U_FAILURE(*pErrorCode)) {
//             udata_printError(ds, "ucnv_swapAliases().swapInvChars(charset names) failed\n");
//             return 0;
//         }

//         if(ds->inCharset==ds->outCharset) {
//             /* no need to sort, just swap all 16-bit values together */
//             ds->swapArray16(ds,
//                             inTable+offsets[converterListIndex],
//                             2*(int32_t)(offsets[stringTableIndex]-offsets[converterListIndex]),
//                             outTable+offsets[converterListIndex],
//                             pErrorCode);
//         } else {
//             /* allocate the temporary table for sorting */
//             count=toc[aliasListIndex];

//             tempTable.chars=(const char *)(outTable+offsets[stringTableIndex]); /* sort by outCharset */

//             if(count<=STACK_ROW_CAPACITY) {
//                 tempTable.rows=rows;
//                 tempTable.resort=resort;
//             } else {
//                 tempTable.rows=(TempRow *)uprv_malloc(count*sizeof(TempRow)+count*2);
//                 if(tempTable.rows==NULL) {
//                     udata_printError(ds, "ucnv_swapAliases(): unable to allocate memory for sorting tables (max length: %u)\n",
//                                      count);
//                     *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
//                     return 0;
//                 }
//                 tempTable.resort=(uint16_t *)(tempTable.rows+count);
//             }

//             if(ds->outCharset==U_ASCII_FAMILY) {
//                 tempTable.stripForCompare=ucnv_io_stripASCIIForCompare;
//             } else /* U_EBCDIC_FAMILY */ {
//                 tempTable.stripForCompare=ucnv_io_stripEBCDICForCompare;
//             }

//             /*
//              * Sort unique aliases+mapped names.
//              *
//              * We need to sort the list again by outCharset strings because they
//              * sort differently for different charset families.
//              * First we set up a temporary table with the string indexes and
//              * sorting indexes and sort that.
//              * Then we permutate and copy/swap the actual values.
//              */
//             p=inTable+offsets[aliasListIndex];
//             q=outTable+offsets[aliasListIndex];

//             p2=inTable+offsets[untaggedConvArrayIndex];
//             q2=outTable+offsets[untaggedConvArrayIndex];

//             for(i=0; i<count; ++i) {
//                 tempTable.rows[i].strIndex=ds->readUInt16(p[i]);
//                 tempTable.rows[i].sortIndex=(uint16_t)i;
//             }

//             uprv_sortArray(tempTable.rows, (int32_t)count, sizeof(TempRow),
//                            io_compareRows, &tempTable,
//                            FALSE, pErrorCode);

//             if(U_SUCCESS(*pErrorCode)) {
//                 /* copy/swap/permutate items */
//                 if(p!=q) {
//                     for(i=0; i<count; ++i) {
//                         oldIndex=tempTable.rows[i].sortIndex;
//                         ds->swapArray16(ds, p+oldIndex, 2, q+i, pErrorCode);
//                         ds->swapArray16(ds, p2+oldIndex, 2, q2+i, pErrorCode);
//                     }
//                 } else {
//                     /*
//                      * If we swap in-place, then the permutation must use another
//                      * temporary array (tempTable.resort)
//                      * before the results are copied to the outBundle.
//                      */
//                     uint16_t *r=tempTable.resort;

//                     for(i=0; i<count; ++i) {
//                         oldIndex=tempTable.rows[i].sortIndex;
//                         ds->swapArray16(ds, p+oldIndex, 2, r+i, pErrorCode);
//                     }
//                     uprv_memcpy(q, r, 2*(size_t)count);

//                     for(i=0; i<count; ++i) {
//                         oldIndex=tempTable.rows[i].sortIndex;
//                         ds->swapArray16(ds, p2+oldIndex, 2, r+i, pErrorCode);
//                     }
//                     uprv_memcpy(q2, r, 2*(size_t)count);
//                 }
//             }

//             if(tempTable.rows!=rows) {
//                 uprv_free(tempTable.rows);
//             }

//             if(U_FAILURE(*pErrorCode)) {
//                 udata_printError(ds, "ucnv_swapAliases().uprv_sortArray(%u items) failed\n",
//                                  count);
//                 return 0;
//             }

//             /* swap remaining 16-bit values */
//             ds->swapArray16(ds,
//                             inTable+offsets[converterListIndex],
//                             2*(int32_t)(offsets[aliasListIndex]-offsets[converterListIndex]),
//                             outTable+offsets[converterListIndex],
//                             pErrorCode);
//             ds->swapArray16(ds,
//                             inTable+offsets[taggedAliasArrayIndex],
//                             2*(int32_t)(offsets[stringTableIndex]-offsets[taggedAliasArrayIndex]),
//                             outTable+offsets[taggedAliasArrayIndex],
//                             pErrorCode);
//         }
//     }

//     return headerSize+2*(int32_t)topOffset;
// }

// #endif


// /*
//  * Hey, Emacs, please set the following:
//  *
//  * Local Variables:
//  * indent-tabs-mode: nil
//  * End:
//  *
//  */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// *******************************************************************************
// *
// *   Copyright (C) 2003-2007, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *
// *******************************************************************************
// *   file name:  ucnv_set.c
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2004sep07
// *   created by: Markus W. Scherer
// *
// *   Conversion API functions using USet (ucnv_getUnicodeSet())
// *   moved here from ucnv.c for removing the dependency of other ucnv_
// *   implementation functions on the USet implementation.
// */

// #include "unicode/utypes.h"
// #include "unicode/uset.h"
// #include "unicode/ucnv.h"
// #include "ucnv_bld.h"
// #include "uset_imp.h"

// #if !UCONFIG_NO_CONVERSION

// U_CAPI void U_EXPORT2
// ucnv_getUnicodeSet(const UConverter *cnv,
//                    USet *setFillIn,
//                    UConverterUnicodeSet whichSet,
//                    UErrorCode *pErrorCode) {
//     /* argument checking */
//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return;
//     }
//     if(cnv==NULL || setFillIn==NULL || whichSet<UCNV_ROUNDTRIP_SET || UCNV_SET_COUNT<=whichSet) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

//     /* does this converter support this function? */
//     if(cnv->sharedData->impl->getUnicodeSet==NULL) {
//         *pErrorCode=U_UNSUPPORTED_ERROR;
//         return;
//     }

//     {
//         USetAdder sa={
//             NULL,
//             uset_add,
//             uset_addRange,
//             uset_addString,
//             uset_remove,
//             uset_removeRange
//         };
//         sa.set=setFillIn;

//         /* empty the set */
//         uset_clear(setFillIn);

//         /* call the converter to add the code points it supports */
//         cnv->sharedData->impl->getUnicodeSet(cnv, &sa, whichSet, pErrorCode);
//     }
// }

// #endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 2002-2015, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// **********************************************************************
// *   file name:  ucnv_u16.c
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2002jul01
// *   created by: Markus W. Scherer
// *
// *   UTF-16 converter implementation. Used to be in ucnv_utf.c.
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "unicode/ucnv.h"
// #include "unicode/uversion.h"
// #include "ucnv_bld.h"
// #include "ucnv_cnv.h"
// #include "cmemory.h"

// enum {
//     UCNV_NEED_TO_WRITE_BOM=1
// };

// U_CDECL_BEGIN
// /*
//  * The UTF-16 toUnicode implementation is also used for the Java-specific
//  * "with BOM" variants of UTF-16BE and UTF-16LE.
//  */
// static void  U_CALLCONV
// _UTF16ToUnicodeWithOffsets(UConverterToUnicodeArgs *pArgs,
//                            UErrorCode *pErrorCode);

// /* UTF-16BE ----------------------------------------------------------------- */

// #if U_IS_BIG_ENDIAN
// #   define _UTF16PEFromUnicodeWithOffsets   _UTF16BEFromUnicodeWithOffsets
// #else
// #   define _UTF16PEFromUnicodeWithOffsets   _UTF16LEFromUnicodeWithOffsets
// #endif


// static void  U_CALLCONV
// _UTF16BEFromUnicodeWithOffsets(UConverterFromUnicodeArgs *pArgs,
//                                UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const UChar *source;
//     char *target;
//     int32_t *offsets;

//     uint32_t targetCapacity, length, sourceIndex;
//     UChar c, trail;
//     char overflow[4];

//     source=pArgs->source;
//     length=(int32_t)(pArgs->sourceLimit-source);
//     if(length<=0) {
//         /* no input, nothing to do */
//         return;
//     }

//     cnv=pArgs->converter;

//     /* write the BOM if necessary */
//     if(cnv->fromUnicodeStatus==UCNV_NEED_TO_WRITE_BOM) {
//         static const char bom[]={ (char)0xfeu, (char)0xffu };
//         ucnv_fromUWriteBytes(cnv,
//                              bom, 2,
//                              &pArgs->target, pArgs->targetLimit,
//                              &pArgs->offsets, -1,
//                              pErrorCode);
//         cnv->fromUnicodeStatus=0;
//     }

//     target=pArgs->target;
//     if(target >= pArgs->targetLimit) {
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//         return;
//     }

//     targetCapacity=(uint32_t)(pArgs->targetLimit-target);
//     offsets=pArgs->offsets;
//     sourceIndex=0;

//     /* c!=0 indicates in several places outside the main loops that a surrogate was found */

//     if((c=(UChar)cnv->fromUChar32)!=0 && U16_IS_TRAIL(trail=*source) && targetCapacity>=4) {
//         /* the last buffer ended with a lead surrogate, output the surrogate pair */
//         ++source;
//         --length;
//         target[0]=(uint8_t)(c>>8);
//         target[1]=(uint8_t)c;
//         target[2]=(uint8_t)(trail>>8);
//         target[3]=(uint8_t)trail;
//         target+=4;
//         targetCapacity-=4;
//         if(offsets!=NULL) {
//             *offsets++=-1;
//             *offsets++=-1;
//             *offsets++=-1;
//             *offsets++=-1;
//         }
//         sourceIndex=1;
//         cnv->fromUChar32=c=0;
//     }

//     if(c==0) {
//         /* copy an even number of bytes for complete UChars */
//         uint32_t count=2*length;
//         if(count>targetCapacity) {
//             count=targetCapacity&~1;
//         }
//         /* count is even */
//         targetCapacity-=count;
//         count>>=1;
//         length-=count;

//         if(offsets==NULL) {
//             while(count>0) {
//                 c=*source++;
//                 if(U16_IS_SINGLE(c)) {
//                     target[0]=(uint8_t)(c>>8);
//                     target[1]=(uint8_t)c;
//                     target+=2;
//                 } else if(U16_IS_SURROGATE_LEAD(c) && count>=2 && U16_IS_TRAIL(trail=*source)) {
//                     ++source;
//                     --count;
//                     target[0]=(uint8_t)(c>>8);
//                     target[1]=(uint8_t)c;
//                     target[2]=(uint8_t)(trail>>8);
//                     target[3]=(uint8_t)trail;
//                     target+=4;
//                 } else {
//                     break;
//                 }
//                 --count;
//             }
//         } else {
//             while(count>0) {
//                 c=*source++;
//                 if(U16_IS_SINGLE(c)) {
//                     target[0]=(uint8_t)(c>>8);
//                     target[1]=(uint8_t)c;
//                     target+=2;
//                     *offsets++=sourceIndex;
//                     *offsets++=sourceIndex++;
//                 } else if(U16_IS_SURROGATE_LEAD(c) && count>=2 && U16_IS_TRAIL(trail=*source)) {
//                     ++source;
//                     --count;
//                     target[0]=(uint8_t)(c>>8);
//                     target[1]=(uint8_t)c;
//                     target[2]=(uint8_t)(trail>>8);
//                     target[3]=(uint8_t)trail;
//                     target+=4;
//                     *offsets++=sourceIndex;
//                     *offsets++=sourceIndex;
//                     *offsets++=sourceIndex;
//                     *offsets++=sourceIndex;
//                     sourceIndex+=2;
//                 } else {
//                     break;
//                 }
//                 --count;
//             }
//         }

//         if(count==0) {
//             /* done with the loop for complete UChars */
//             if(length>0 && targetCapacity>0) {
//                 /*
//                  * there is more input and some target capacity -
//                  * it must be targetCapacity==1 because otherwise
//                  * the above would have copied more;
//                  * prepare for overflow output
//                  */
//                 if(U16_IS_SINGLE(c=*source++)) {
//                     overflow[0]=(char)(c>>8);
//                     overflow[1]=(char)c;
//                     length=2; /* 2 bytes to output */
//                     c=0;
//                 /* } else { keep c for surrogate handling, length will be set there */
//                 }
//             } else {
//                 length=0;
//                 c=0;
//             }
//         } else {
//             /* keep c for surrogate handling, length will be set there */
//             targetCapacity+=2*count;
//         }
//     } else {
//         length=0; /* from here on, length counts the bytes in overflow[] */
//     }

//     if(c!=0) {
//         /*
//          * c is a surrogate, and
//          * - source or target too short
//          * - or the surrogate is unmatched
//          */
//         length=0;
//         if(U16_IS_SURROGATE_LEAD(c)) {
//             if(source<pArgs->sourceLimit) {
//                 if(U16_IS_TRAIL(trail=*source)) {
//                     /* output the surrogate pair, will overflow (see conditions comment above) */
//                     ++source;
//                     overflow[0]=(char)(c>>8);
//                     overflow[1]=(char)c;
//                     overflow[2]=(char)(trail>>8);
//                     overflow[3]=(char)trail;
//                     length=4; /* 4 bytes to output */
//                     c=0;
//                 } else {
//                     /* unmatched lead surrogate */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 }
//             } else {
//                 /* see if the trail surrogate is in the next buffer */
//             }
//         } else {
//             /* unmatched trail surrogate */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//         }
//         cnv->fromUChar32=c;
//     }

//     if(length>0) {
//         /* output length bytes with overflow (length>targetCapacity>0) */
//         ucnv_fromUWriteBytes(cnv,
//                              overflow, length,
//                              (char **)&target, pArgs->targetLimit,
//                              &offsets, sourceIndex,
//                              pErrorCode);
//         targetCapacity=(uint32_t)(pArgs->targetLimit-(char *)target);
//     }

//     if(U_SUCCESS(*pErrorCode) && source<pArgs->sourceLimit && targetCapacity==0) {
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//     }

//     /* write back the updated pointers */
//     pArgs->source=source;
//     pArgs->target=(char *)target;
//     pArgs->offsets=offsets;
// }

// static void  U_CALLCONV
// _UTF16BEToUnicodeWithOffsets(UConverterToUnicodeArgs *pArgs,
//                              UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const uint8_t *source;
//     UChar *target;
//     int32_t *offsets;

//     uint32_t targetCapacity, length, count, sourceIndex;
//     UChar c, trail;

//     if(pArgs->converter->mode<8) {
//         _UTF16ToUnicodeWithOffsets(pArgs, pErrorCode);
//         return;
//     }

//     cnv=pArgs->converter;
//     source=(const uint8_t *)pArgs->source;
//     length=(int32_t)((const uint8_t *)pArgs->sourceLimit-source);
//     if(length<=0 && cnv->toUnicodeStatus==0) {
//         /* no input, nothing to do */
//         return;
//     }

//     target=pArgs->target;
//     if(target >= pArgs->targetLimit) {
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//         return;
//     }

//     targetCapacity=(uint32_t)(pArgs->targetLimit-target);
//     offsets=pArgs->offsets;
//     sourceIndex=0;
//     c=0;

//     /* complete a partial UChar or pair from the last call */
//     if(cnv->toUnicodeStatus!=0) {
//         /*
//          * special case: single byte from a previous buffer,
//          * where the byte turned out not to belong to a trail surrogate
//          * and the preceding, unmatched lead surrogate was put into toUBytes[]
//          * for error handling
//          */
//         cnv->toUBytes[0]=(uint8_t)cnv->toUnicodeStatus;
//         cnv->toULength=1;
//         cnv->toUnicodeStatus=0;
//     }
//     if((count=cnv->toULength)!=0) {
//         uint8_t *p=cnv->toUBytes;
//         do {
//             p[count++]=*source++;
//             ++sourceIndex;
//             --length;
//             if(count==2) {
//                 c=((UChar)p[0]<<8)|p[1];
//                 if(U16_IS_SINGLE(c)) {
//                     /* output the BMP code point */
//                     *target++=c;
//                     if(offsets!=NULL) {
//                         *offsets++=-1;
//                     }
//                     --targetCapacity;
//                     count=0;
//                     c=0;
//                     break;
//                 } else if(U16_IS_SURROGATE_LEAD(c)) {
//                     /* continue collecting bytes for the trail surrogate */
//                     c=0; /* avoid unnecessary surrogate handling below */
//                 } else {
//                     /* fall through to error handling for an unmatched trail surrogate */
//                     break;
//                 }
//             } else if(count==4) {
//                 c=((UChar)p[0]<<8)|p[1];
//                 trail=((UChar)p[2]<<8)|p[3];
//                 if(U16_IS_TRAIL(trail)) {
//                     /* output the surrogate pair */
//                     *target++=c;
//                     if(targetCapacity>=2) {
//                         *target++=trail;
//                         if(offsets!=NULL) {
//                             *offsets++=-1;
//                             *offsets++=-1;
//                         }
//                         targetCapacity-=2;
//                     } else /* targetCapacity==1 */ {
//                         targetCapacity=0;
//                         cnv->UCharErrorBuffer[0]=trail;
//                         cnv->UCharErrorBufferLength=1;
//                         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                     }
//                     count=0;
//                     c=0;
//                     break;
//                 } else {
//                     /* unmatched lead surrogate, handle here for consistent toUBytes[] */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;

//                     /* back out reading the code unit after it */
//                     if(((const uint8_t *)pArgs->source-source)>=2) {
//                         source-=2;
//                     } else {
//                         /*
//                          * if the trail unit's first byte was in a previous buffer, then
//                          * we need to put it into a special place because toUBytes[] will be
//                          * used for the lead unit's bytes
//                          */
//                         cnv->toUnicodeStatus=0x100|p[2];
//                         --source;
//                     }
//                     cnv->toULength=2;

//                     /* write back the updated pointers */
//                     pArgs->source=(const char *)source;
//                     pArgs->target=target;
//                     pArgs->offsets=offsets;
//                     return;
//                 }
//             }
//         } while(length>0);
//         cnv->toULength=(int8_t)count;
//     }

//     /* copy an even number of bytes for complete UChars */
//     count=2*targetCapacity;
//     if(count>length) {
//         count=length&~1;
//     }
//     if(c==0 && count>0) {
//         length-=count;
//         count>>=1;
//         targetCapacity-=count;
//         if(offsets==NULL) {
//             do {
//                 c=((UChar)source[0]<<8)|source[1];
//                 source+=2;
//                 if(U16_IS_SINGLE(c)) {
//                     *target++=c;
//                 } else if(U16_IS_SURROGATE_LEAD(c) && count>=2 &&
//                           U16_IS_TRAIL(trail=((UChar)source[0]<<8)|source[1])
//                 ) {
//                     source+=2;
//                     --count;
//                     *target++=c;
//                     *target++=trail;
//                 } else {
//                     break;
//                 }
//             } while(--count>0);
//         } else {
//             do {
//                 c=((UChar)source[0]<<8)|source[1];
//                 source+=2;
//                 if(U16_IS_SINGLE(c)) {
//                     *target++=c;
//                     *offsets++=sourceIndex;
//                     sourceIndex+=2;
//                 } else if(U16_IS_SURROGATE_LEAD(c) && count>=2 &&
//                           U16_IS_TRAIL(trail=((UChar)source[0]<<8)|source[1])
//                 ) {
//                     source+=2;
//                     --count;
//                     *target++=c;
//                     *target++=trail;
//                     *offsets++=sourceIndex;
//                     *offsets++=sourceIndex;
//                     sourceIndex+=4;
//                 } else {
//                     break;
//                 }
//             } while(--count>0);
//         }

//         if(count==0) {
//             /* done with the loop for complete UChars */
//             c=0;
//         } else {
//             /* keep c for surrogate handling, trail will be set there */
//             length+=2*(count-1); /* one more byte pair was consumed than count decremented */
//             targetCapacity+=count;
//         }
//     }

//     if(c!=0) {
//         /*
//          * c is a surrogate, and
//          * - source or target too short
//          * - or the surrogate is unmatched
//          */
//         cnv->toUBytes[0]=(uint8_t)(c>>8);
//         cnv->toUBytes[1]=(uint8_t)c;
//         cnv->toULength=2;

//         if(U16_IS_SURROGATE_LEAD(c)) {
//             if(length>=2) {
//                 if(U16_IS_TRAIL(trail=((UChar)source[0]<<8)|source[1])) {
//                     /* output the surrogate pair, will overflow (see conditions comment above) */
//                     source+=2;
//                     length-=2;
//                     *target++=c;
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     cnv->UCharErrorBuffer[0]=trail;
//                     cnv->UCharErrorBufferLength=1;
//                     cnv->toULength=0;
//                     *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 } else {
//                     /* unmatched lead surrogate */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 }
//             } else {
//                 /* see if the trail surrogate is in the next buffer */
//             }
//         } else {
//             /* unmatched trail surrogate */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//         }
//     }

//     if(U_SUCCESS(*pErrorCode)) {
//         /* check for a remaining source byte */
//         if(length>0) {
//             if(targetCapacity==0) {
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             } else {
//                 /* it must be length==1 because otherwise the above would have copied more */
//                 cnv->toUBytes[cnv->toULength++]=*source++;
//             }
//         }
//     }

//     /* write back the updated pointers */
//     pArgs->source=(const char *)source;
//     pArgs->target=target;
//     pArgs->offsets=offsets;
// }

// static UChar32  U_CALLCONV
// _UTF16BEGetNextUChar(UConverterToUnicodeArgs *pArgs, UErrorCode *err) {
//     const uint8_t *s, *sourceLimit;
//     UChar32 c;

//     if(pArgs->converter->mode<8) {
//         return UCNV_GET_NEXT_UCHAR_USE_TO_U;
//     }

//     s=(const uint8_t *)pArgs->source;
//     sourceLimit=(const uint8_t *)pArgs->sourceLimit;

//     if(s>=sourceLimit) {
//         /* no input */
//         *err=U_INDEX_OUTOFBOUNDS_ERROR;
//         return 0xffff;
//     }

//     if(s+2>sourceLimit) {
//         /* only one byte: truncated UChar */
//         pArgs->converter->toUBytes[0]=*s++;
//         pArgs->converter->toULength=1;
//         pArgs->source=(const char *)s;
//         *err = U_TRUNCATED_CHAR_FOUND;
//         return 0xffff;
//     }

//     /* get one UChar */
//     c=((UChar32)*s<<8)|s[1];
//     s+=2;

//     /* check for a surrogate pair */
//     if(U_IS_SURROGATE(c)) {
//         if(U16_IS_SURROGATE_LEAD(c)) {
//             if(s+2<=sourceLimit) {
//                 UChar trail;

//                 /* get a second UChar and see if it is a trail surrogate */
//                 trail=((UChar)*s<<8)|s[1];
//                 if(U16_IS_TRAIL(trail)) {
//                     c=U16_GET_SUPPLEMENTARY(c, trail);
//                     s+=2;
//                 } else {
//                     /* unmatched lead surrogate */
//                     c=-2;
//                 }
//             } else {
//                 /* too few (2 or 3) bytes for a surrogate pair: truncated code point */
//                 uint8_t *bytes=pArgs->converter->toUBytes;
//                 s-=2;
//                 pArgs->converter->toULength=(int8_t)(sourceLimit-s);
//                 do {
//                     *bytes++=*s++;
//                 } while(s<sourceLimit);

//                 c=0xffff;
//                 *err=U_TRUNCATED_CHAR_FOUND;
//             }
//         } else {
//             /* unmatched trail surrogate */
//             c=-2;
//         }

//         if(c<0) {
//             /* write the unmatched surrogate */
//             uint8_t *bytes=pArgs->converter->toUBytes;
//             pArgs->converter->toULength=2;
//             *bytes=*(s-2);
//             bytes[1]=*(s-1);

//             c=0xffff;
//             *err=U_ILLEGAL_CHAR_FOUND;
//         }
//     }

//     pArgs->source=(const char *)s;
//     return c;
// }

// static void  U_CALLCONV
// _UTF16BEReset(UConverter *cnv, UConverterResetChoice choice) {
//     if(choice<=UCNV_RESET_TO_UNICODE) {
//         /* reset toUnicode state */
//         if(UCNV_GET_VERSION(cnv)==0) {
//             cnv->mode=8; /* no BOM handling */
//         } else {
//             cnv->mode=0; /* Java-specific "UnicodeBig" requires BE BOM or no BOM */
//         }
//     }
//     if(choice!=UCNV_RESET_TO_UNICODE && UCNV_GET_VERSION(cnv)==1) {
//         /* reset fromUnicode for "UnicodeBig": prepare to output the UTF-16BE BOM */
//         cnv->fromUnicodeStatus=UCNV_NEED_TO_WRITE_BOM;
//     }
// }

// static void  U_CALLCONV
// _UTF16BEOpen(UConverter *cnv,
//              UConverterLoadArgs *pArgs,
//              UErrorCode *pErrorCode) {
//     (void)pArgs;
//     if(UCNV_GET_VERSION(cnv)<=1) {
//         _UTF16BEReset(cnv, UCNV_RESET_BOTH);
//     } else {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//     }
// }

// static const char *  U_CALLCONV
// _UTF16BEGetName(const UConverter *cnv) {
//     if(UCNV_GET_VERSION(cnv)==0) {
//         return "UTF-16BE";
//     } else {
//         return "UTF-16BE,version=1";
//     }
// }
// U_CDECL_END

// static const UConverterImpl _UTF16BEImpl={
//     UCNV_UTF16_BigEndian,

//     NULL,
//     NULL,

//     _UTF16BEOpen,
//     NULL,
//     _UTF16BEReset,

//     _UTF16BEToUnicodeWithOffsets,
//     _UTF16BEToUnicodeWithOffsets,
//     _UTF16BEFromUnicodeWithOffsets,
//     _UTF16BEFromUnicodeWithOffsets,
//     _UTF16BEGetNextUChar,

//     NULL,
//     _UTF16BEGetName,
//     NULL,
//     NULL,
//     ucnv_getNonSurrogateUnicodeSet,

//     NULL,
//     NULL
// };

// static const UConverterStaticData _UTF16BEStaticData={
//     sizeof(UConverterStaticData),
//     "UTF-16BE",
//     1200, UCNV_IBM, UCNV_UTF16_BigEndian, 2, 2,
//     { 0xff, 0xfd, 0, 0 },2,FALSE,FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };


// const UConverterSharedData _UTF16BEData=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_UTF16BEStaticData, &_UTF16BEImpl);

// /* UTF-16LE ----------------------------------------------------------------- */
// U_CDECL_BEGIN
// static void  U_CALLCONV
// _UTF16LEFromUnicodeWithOffsets(UConverterFromUnicodeArgs *pArgs,
//                                UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const UChar *source;
//     char *target;
//     int32_t *offsets;

//     uint32_t targetCapacity, length, sourceIndex;
//     UChar c, trail;
//     char overflow[4];

//     source=pArgs->source;
//     length=(int32_t)(pArgs->sourceLimit-source);
//     if(length<=0) {
//         /* no input, nothing to do */
//         return;
//     }

//     cnv=pArgs->converter;

//     /* write the BOM if necessary */
//     if(cnv->fromUnicodeStatus==UCNV_NEED_TO_WRITE_BOM) {
//         static const char bom[]={ (char)0xffu, (char)0xfeu };
//         ucnv_fromUWriteBytes(cnv,
//                              bom, 2,
//                              &pArgs->target, pArgs->targetLimit,
//                              &pArgs->offsets, -1,
//                              pErrorCode);
//         cnv->fromUnicodeStatus=0;
//     }

//     target=pArgs->target;
//     if(target >= pArgs->targetLimit) {
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//         return;
//     }

//     targetCapacity=(uint32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;
//     sourceIndex=0;

//     /* c!=0 indicates in several places outside the main loops that a surrogate was found */

//     if((c=(UChar)cnv->fromUChar32)!=0 && U16_IS_TRAIL(trail=*source) && targetCapacity>=4) {
//         /* the last buffer ended with a lead surrogate, output the surrogate pair */
//         ++source;
//         --length;
//         target[0]=(uint8_t)c;
//         target[1]=(uint8_t)(c>>8);
//         target[2]=(uint8_t)trail;
//         target[3]=(uint8_t)(trail>>8);
//         target+=4;
//         targetCapacity-=4;
//         if(offsets!=NULL) {
//             *offsets++=-1;
//             *offsets++=-1;
//             *offsets++=-1;
//             *offsets++=-1;
//         }
//         sourceIndex=1;
//         cnv->fromUChar32=c=0;
//     }

//     if(c==0) {
//         /* copy an even number of bytes for complete UChars */
//         uint32_t count=2*length;
//         if(count>targetCapacity) {
//             count=targetCapacity&~1;
//         }
//         /* count is even */
//         targetCapacity-=count;
//         count>>=1;
//         length-=count;

//         if(offsets==NULL) {
//             while(count>0) {
//                 c=*source++;
//                 if(U16_IS_SINGLE(c)) {
//                     target[0]=(uint8_t)c;
//                     target[1]=(uint8_t)(c>>8);
//                     target+=2;
//                 } else if(U16_IS_SURROGATE_LEAD(c) && count>=2 && U16_IS_TRAIL(trail=*source)) {
//                     ++source;
//                     --count;
//                     target[0]=(uint8_t)c;
//                     target[1]=(uint8_t)(c>>8);
//                     target[2]=(uint8_t)trail;
//                     target[3]=(uint8_t)(trail>>8);
//                     target+=4;
//                 } else {
//                     break;
//                 }
//                 --count;
//             }
//         } else {
//             while(count>0) {
//                 c=*source++;
//                 if(U16_IS_SINGLE(c)) {
//                     target[0]=(uint8_t)c;
//                     target[1]=(uint8_t)(c>>8);
//                     target+=2;
//                     *offsets++=sourceIndex;
//                     *offsets++=sourceIndex++;
//                 } else if(U16_IS_SURROGATE_LEAD(c) && count>=2 && U16_IS_TRAIL(trail=*source)) {
//                     ++source;
//                     --count;
//                     target[0]=(uint8_t)c;
//                     target[1]=(uint8_t)(c>>8);
//                     target[2]=(uint8_t)trail;
//                     target[3]=(uint8_t)(trail>>8);
//                     target+=4;
//                     *offsets++=sourceIndex;
//                     *offsets++=sourceIndex;
//                     *offsets++=sourceIndex;
//                     *offsets++=sourceIndex;
//                     sourceIndex+=2;
//                 } else {
//                     break;
//                 }
//                 --count;
//             }
//         }

//         if(count==0) {
//             /* done with the loop for complete UChars */
//             if(length>0 && targetCapacity>0) {
//                 /*
//                  * there is more input and some target capacity -
//                  * it must be targetCapacity==1 because otherwise
//                  * the above would have copied more;
//                  * prepare for overflow output
//                  */
//                 if(U16_IS_SINGLE(c=*source++)) {
//                     overflow[0]=(char)c;
//                     overflow[1]=(char)(c>>8);
//                     length=2; /* 2 bytes to output */
//                     c=0;
//                 /* } else { keep c for surrogate handling, length will be set there */
//                 }
//             } else {
//                 length=0;
//                 c=0;
//             }
//         } else {
//             /* keep c for surrogate handling, length will be set there */
//             targetCapacity+=2*count;
//         }
//     } else {
//         length=0; /* from here on, length counts the bytes in overflow[] */
//     }

//     if(c!=0) {
//         /*
//          * c is a surrogate, and
//          * - source or target too short
//          * - or the surrogate is unmatched
//          */
//         length=0;
//         if(U16_IS_SURROGATE_LEAD(c)) {
//             if(source<pArgs->sourceLimit) {
//                 if(U16_IS_TRAIL(trail=*source)) {
//                     /* output the surrogate pair, will overflow (see conditions comment above) */
//                     ++source;
//                     overflow[0]=(char)c;
//                     overflow[1]=(char)(c>>8);
//                     overflow[2]=(char)trail;
//                     overflow[3]=(char)(trail>>8);
//                     length=4; /* 4 bytes to output */
//                     c=0;
//                 } else {
//                     /* unmatched lead surrogate */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 }
//             } else {
//                 /* see if the trail surrogate is in the next buffer */
//             }
//         } else {
//             /* unmatched trail surrogate */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//         }
//         cnv->fromUChar32=c;
//     }

//     if(length>0) {
//         /* output length bytes with overflow (length>targetCapacity>0) */
//         ucnv_fromUWriteBytes(cnv,
//                              overflow, length,
//                              &target, pArgs->targetLimit,
//                              &offsets, sourceIndex,
//                              pErrorCode);
//         targetCapacity=(uint32_t)(pArgs->targetLimit-(char *)target);
//     }

//     if(U_SUCCESS(*pErrorCode) && source<pArgs->sourceLimit && targetCapacity==0) {
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//     }

//     /* write back the updated pointers */
//     pArgs->source=source;
//     pArgs->target=target;
//     pArgs->offsets=offsets;
// }

// static void  U_CALLCONV
// _UTF16LEToUnicodeWithOffsets(UConverterToUnicodeArgs *pArgs,
//                              UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const uint8_t *source;
//     UChar *target;
//     int32_t *offsets;

//     uint32_t targetCapacity, length, count, sourceIndex;
//     UChar c, trail;

//     if(pArgs->converter->mode<8) {
//         _UTF16ToUnicodeWithOffsets(pArgs, pErrorCode);
//         return;
//     }

//     cnv=pArgs->converter;
//     source=(const uint8_t *)pArgs->source;
//     length=(int32_t)((const uint8_t *)pArgs->sourceLimit-source);
//     if(length<=0 && cnv->toUnicodeStatus==0) {
//         /* no input, nothing to do */
//         return;
//     }

//     target=pArgs->target;
//     if(target >= pArgs->targetLimit) {
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//         return;
//     }

//     targetCapacity=(uint32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;
//     sourceIndex=0;
//     c=0;

//     /* complete a partial UChar or pair from the last call */
//     if(cnv->toUnicodeStatus!=0) {
//         /*
//          * special case: single byte from a previous buffer,
//          * where the byte turned out not to belong to a trail surrogate
//          * and the preceding, unmatched lead surrogate was put into toUBytes[]
//          * for error handling
//          */
//         cnv->toUBytes[0]=(uint8_t)cnv->toUnicodeStatus;
//         cnv->toULength=1;
//         cnv->toUnicodeStatus=0;
//     }
//     if((count=cnv->toULength)!=0) {
//         uint8_t *p=cnv->toUBytes;
//         do {
//             p[count++]=*source++;
//             ++sourceIndex;
//             --length;
//             if(count==2) {
//                 c=((UChar)p[1]<<8)|p[0];
//                 if(U16_IS_SINGLE(c)) {
//                     /* output the BMP code point */
//                     *target++=c;
//                     if(offsets!=NULL) {
//                         *offsets++=-1;
//                     }
//                     --targetCapacity;
//                     count=0;
//                     c=0;
//                     break;
//                 } else if(U16_IS_SURROGATE_LEAD(c)) {
//                     /* continue collecting bytes for the trail surrogate */
//                     c=0; /* avoid unnecessary surrogate handling below */
//                 } else {
//                     /* fall through to error handling for an unmatched trail surrogate */
//                     break;
//                 }
//             } else if(count==4) {
//                 c=((UChar)p[1]<<8)|p[0];
//                 trail=((UChar)p[3]<<8)|p[2];
//                 if(U16_IS_TRAIL(trail)) {
//                     /* output the surrogate pair */
//                     *target++=c;
//                     if(targetCapacity>=2) {
//                         *target++=trail;
//                         if(offsets!=NULL) {
//                             *offsets++=-1;
//                             *offsets++=-1;
//                         }
//                         targetCapacity-=2;
//                     } else /* targetCapacity==1 */ {
//                         targetCapacity=0;
//                         cnv->UCharErrorBuffer[0]=trail;
//                         cnv->UCharErrorBufferLength=1;
//                         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                     }
//                     count=0;
//                     c=0;
//                     break;
//                 } else {
//                     /* unmatched lead surrogate, handle here for consistent toUBytes[] */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;

//                     /* back out reading the code unit after it */
//                     if(((const uint8_t *)pArgs->source-source)>=2) {
//                         source-=2;
//                     } else {
//                         /*
//                          * if the trail unit's first byte was in a previous buffer, then
//                          * we need to put it into a special place because toUBytes[] will be
//                          * used for the lead unit's bytes
//                          */
//                         cnv->toUnicodeStatus=0x100|p[2];
//                         --source;
//                     }
//                     cnv->toULength=2;

//                     /* write back the updated pointers */
//                     pArgs->source=(const char *)source;
//                     pArgs->target=target;
//                     pArgs->offsets=offsets;
//                     return;
//                 }
//             }
//         } while(length>0);
//         cnv->toULength=(int8_t)count;
//     }

//     /* copy an even number of bytes for complete UChars */
//     count=2*targetCapacity;
//     if(count>length) {
//         count=length&~1;
//     }
//     if(c==0 && count>0) {
//         length-=count;
//         count>>=1;
//         targetCapacity-=count;
//         if(offsets==NULL) {
//             do {
//                 c=((UChar)source[1]<<8)|source[0];
//                 source+=2;
//                 if(U16_IS_SINGLE(c)) {
//                     *target++=c;
//                 } else if(U16_IS_SURROGATE_LEAD(c) && count>=2 &&
//                           U16_IS_TRAIL(trail=((UChar)source[1]<<8)|source[0])
//                 ) {
//                     source+=2;
//                     --count;
//                     *target++=c;
//                     *target++=trail;
//                 } else {
//                     break;
//                 }
//             } while(--count>0);
//         } else {
//             do {
//                 c=((UChar)source[1]<<8)|source[0];
//                 source+=2;
//                 if(U16_IS_SINGLE(c)) {
//                     *target++=c;
//                     *offsets++=sourceIndex;
//                     sourceIndex+=2;
//                 } else if(U16_IS_SURROGATE_LEAD(c) && count>=2 &&
//                           U16_IS_TRAIL(trail=((UChar)source[1]<<8)|source[0])
//                 ) {
//                     source+=2;
//                     --count;
//                     *target++=c;
//                     *target++=trail;
//                     *offsets++=sourceIndex;
//                     *offsets++=sourceIndex;
//                     sourceIndex+=4;
//                 } else {
//                     break;
//                 }
//             } while(--count>0);
//         }

//         if(count==0) {
//             /* done with the loop for complete UChars */
//             c=0;
//         } else {
//             /* keep c for surrogate handling, trail will be set there */
//             length+=2*(count-1); /* one more byte pair was consumed than count decremented */
//             targetCapacity+=count;
//         }
//     }

//     if(c!=0) {
//         /*
//          * c is a surrogate, and
//          * - source or target too short
//          * - or the surrogate is unmatched
//          */
//         cnv->toUBytes[0]=(uint8_t)c;
//         cnv->toUBytes[1]=(uint8_t)(c>>8);
//         cnv->toULength=2;

//         if(U16_IS_SURROGATE_LEAD(c)) {
//             if(length>=2) {
//                 if(U16_IS_TRAIL(trail=((UChar)source[1]<<8)|source[0])) {
//                     /* output the surrogate pair, will overflow (see conditions comment above) */
//                     source+=2;
//                     length-=2;
//                     *target++=c;
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     cnv->UCharErrorBuffer[0]=trail;
//                     cnv->UCharErrorBufferLength=1;
//                     cnv->toULength=0;
//                     *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 } else {
//                     /* unmatched lead surrogate */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 }
//             } else {
//                 /* see if the trail surrogate is in the next buffer */
//             }
//         } else {
//             /* unmatched trail surrogate */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//         }
//     }

//     if(U_SUCCESS(*pErrorCode)) {
//         /* check for a remaining source byte */
//         if(length>0) {
//             if(targetCapacity==0) {
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             } else {
//                 /* it must be length==1 because otherwise the above would have copied more */
//                 cnv->toUBytes[cnv->toULength++]=*source++;
//             }
//         }
//     }

//     /* write back the updated pointers */
//     pArgs->source=(const char *)source;
//     pArgs->target=target;
//     pArgs->offsets=offsets;
// }

// static UChar32  U_CALLCONV
// _UTF16LEGetNextUChar(UConverterToUnicodeArgs *pArgs, UErrorCode *err) {
//     const uint8_t *s, *sourceLimit;
//     UChar32 c;

//     if(pArgs->converter->mode<8) {
//         return UCNV_GET_NEXT_UCHAR_USE_TO_U;
//     }

//     s=(const uint8_t *)pArgs->source;
//     sourceLimit=(const uint8_t *)pArgs->sourceLimit;

//     if(s>=sourceLimit) {
//         /* no input */
//         *err=U_INDEX_OUTOFBOUNDS_ERROR;
//         return 0xffff;
//     }

//     if(s+2>sourceLimit) {
//         /* only one byte: truncated UChar */
//         pArgs->converter->toUBytes[0]=*s++;
//         pArgs->converter->toULength=1;
//         pArgs->source=(const char *)s;
//         *err = U_TRUNCATED_CHAR_FOUND;
//         return 0xffff;
//     }

//     /* get one UChar */
//     c=((UChar32)s[1]<<8)|*s;
//     s+=2;

//     /* check for a surrogate pair */
//     if(U_IS_SURROGATE(c)) {
//         if(U16_IS_SURROGATE_LEAD(c)) {
//             if(s+2<=sourceLimit) {
//                 UChar trail;

//                 /* get a second UChar and see if it is a trail surrogate */
//                 trail=((UChar)s[1]<<8)|*s;
//                 if(U16_IS_TRAIL(trail)) {
//                     c=U16_GET_SUPPLEMENTARY(c, trail);
//                     s+=2;
//                 } else {
//                     /* unmatched lead surrogate */
//                     c=-2;
//                 }
//             } else {
//                 /* too few (2 or 3) bytes for a surrogate pair: truncated code point */
//                 uint8_t *bytes=pArgs->converter->toUBytes;
//                 s-=2;
//                 pArgs->converter->toULength=(int8_t)(sourceLimit-s);
//                 do {
//                     *bytes++=*s++;
//                 } while(s<sourceLimit);

//                 c=0xffff;
//                 *err=U_TRUNCATED_CHAR_FOUND;
//             }
//         } else {
//             /* unmatched trail surrogate */
//             c=-2;
//         }

//         if(c<0) {
//             /* write the unmatched surrogate */
//             uint8_t *bytes=pArgs->converter->toUBytes;
//             pArgs->converter->toULength=2;
//             *bytes=*(s-2);
//             bytes[1]=*(s-1);

//             c=0xffff;
//             *err=U_ILLEGAL_CHAR_FOUND;
//         }
//     }

//     pArgs->source=(const char *)s;
//     return c;
// }

// static void  U_CALLCONV
// _UTF16LEReset(UConverter *cnv, UConverterResetChoice choice) {
//     if(choice<=UCNV_RESET_TO_UNICODE) {
//         /* reset toUnicode state */
//         if(UCNV_GET_VERSION(cnv)==0) {
//             cnv->mode=8; /* no BOM handling */
//         } else {
//             cnv->mode=0; /* Java-specific "UnicodeLittle" requires LE BOM or no BOM */
//         }
//     }
//     if(choice!=UCNV_RESET_TO_UNICODE && UCNV_GET_VERSION(cnv)==1) {
//         /* reset fromUnicode for "UnicodeLittle": prepare to output the UTF-16LE BOM */
//         cnv->fromUnicodeStatus=UCNV_NEED_TO_WRITE_BOM;
//     }
// }

// static void  U_CALLCONV
// _UTF16LEOpen(UConverter *cnv,
//              UConverterLoadArgs *pArgs,
//              UErrorCode *pErrorCode) {
//     (void)pArgs;
//     if(UCNV_GET_VERSION(cnv)<=1) {
//         _UTF16LEReset(cnv, UCNV_RESET_BOTH);
//     } else {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//     }
// }

// static const char *  U_CALLCONV
// _UTF16LEGetName(const UConverter *cnv) {
//     if(UCNV_GET_VERSION(cnv)==0) {
//         return "UTF-16LE";
//     } else {
//         return "UTF-16LE,version=1";
//     }
// }
// U_CDECL_END

// static const UConverterImpl _UTF16LEImpl={
//     UCNV_UTF16_LittleEndian,

//     NULL,
//     NULL,

//     _UTF16LEOpen,
//     NULL,
//     _UTF16LEReset,

//     _UTF16LEToUnicodeWithOffsets,
//     _UTF16LEToUnicodeWithOffsets,
//     _UTF16LEFromUnicodeWithOffsets,
//     _UTF16LEFromUnicodeWithOffsets,
//     _UTF16LEGetNextUChar,

//     NULL,
//     _UTF16LEGetName,
//     NULL,
//     NULL,
//     ucnv_getNonSurrogateUnicodeSet,

//     NULL,
//     NULL
// };


// static const UConverterStaticData _UTF16LEStaticData={
//     sizeof(UConverterStaticData),
//     "UTF-16LE",
//     1202, UCNV_IBM, UCNV_UTF16_LittleEndian, 2, 2,
//     { 0xfd, 0xff, 0, 0 },2,FALSE,FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };


// const UConverterSharedData _UTF16LEData=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_UTF16LEStaticData, &_UTF16LEImpl);

// /* UTF-16 (Detect BOM) ------------------------------------------------------ */

// /*
//  * Detect a BOM at the beginning of the stream and select UTF-16BE or UTF-16LE
//  * accordingly.
//  * This is a simpler version of the UTF-32 converter, with
//  * fewer states for shorter BOMs.
//  *
//  * State values:
//  * 0    initial state
//  * 1    saw first byte
//  * 2..5 -
//  * 6..7 see _UTF16ToUnicodeWithOffsets() comments in state 1
//  * 8    UTF-16BE mode
//  * 9    UTF-16LE mode
//  *
//  * During detection: state==number of initial bytes seen so far.
//  *
//  * On output, emit U+FEFF as the first code point.
//  *
//  * Variants:
//  * - UTF-16,version=1 (Java "Unicode" encoding) treats a missing BOM as an error.
//  * - UTF-16BE,version=1 (Java "UnicodeBig" encoding) and
//  *   UTF-16LE,version=1 (Java "UnicodeLittle" encoding) treat a reverse BOM as an error.
//  */
// U_CDECL_BEGIN
// static void  U_CALLCONV
// _UTF16Reset(UConverter *cnv, UConverterResetChoice choice) {
//     if(choice<=UCNV_RESET_TO_UNICODE) {
//         /* reset toUnicode: state=0 */
//         cnv->mode=0;
//     }
//     if(choice!=UCNV_RESET_TO_UNICODE) {
//         /* reset fromUnicode: prepare to output the UTF-16PE BOM */
//         cnv->fromUnicodeStatus=UCNV_NEED_TO_WRITE_BOM;
//     }
// }
// U_CDECL_END
// extern const UConverterSharedData _UTF16v2Data;
// U_CDECL_BEGIN
// static void U_CALLCONV
// _UTF16Open(UConverter *cnv,
//            UConverterLoadArgs *pArgs,
//            UErrorCode *pErrorCode) {
//     if(UCNV_GET_VERSION(cnv)<=2) {
//         if(UCNV_GET_VERSION(cnv)==2 && !pArgs->onlyTestIsLoadable) {
//             /*
//              * Switch implementation, and switch the staticData that's different
//              * and was copied into the UConverter.
//              * (See ucnv_createConverterFromSharedData() in ucnv_bld.c.)
//              * UTF-16,version=2 fromUnicode() always writes a big-endian byte stream.
//              */
//             cnv->sharedData=(UConverterSharedData*)&_UTF16v2Data;
//             uprv_memcpy(cnv->subChars, _UTF16v2Data.staticData->subChar, UCNV_MAX_SUBCHAR_LEN);
//         }
//         _UTF16Reset(cnv, UCNV_RESET_BOTH);
//     } else {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//     }
// }

// static const char *  U_CALLCONV
// _UTF16GetName(const UConverter *cnv) {
//     if(UCNV_GET_VERSION(cnv)==0) {
//         return "UTF-16";
//     } else if(UCNV_GET_VERSION(cnv)==1) {
//         return "UTF-16,version=1";
//     } else {
//         return "UTF-16,version=2";
//     }
// }
// U_CDECL_END
// extern const UConverterSharedData _UTF16Data;

// static inline bool IS_UTF16BE(const UConverter *cnv) {
//     return ((cnv)->sharedData == &_UTF16BEData);
// }

// static inline bool IS_UTF16LE(const UConverter *cnv) {
//     return ((cnv)->sharedData == &_UTF16LEData);
// }

// static inline bool IS_UTF16(const UConverter *cnv) {
//     return ((cnv)->sharedData==&_UTF16Data) || ((cnv)->sharedData == &_UTF16v2Data);
// }

// U_CDECL_BEGIN
// static void U_CALLCONV
// _UTF16ToUnicodeWithOffsets(UConverterToUnicodeArgs *pArgs,
//                            UErrorCode *pErrorCode) {
//     UConverter *cnv=pArgs->converter;
//     const char *source=pArgs->source;
//     const char *sourceLimit=pArgs->sourceLimit;
//     int32_t *offsets=pArgs->offsets;

//     int32_t state, offsetDelta;
//     uint8_t b;

//     state=cnv->mode;

//     /*
//      * If we detect a BOM in this buffer, then we must add the BOM size to the
//      * offsets because the actual converter function will not see and count the BOM.
//      * offsetDelta will have the number of the BOM bytes that are in the current buffer.
//      */
//     offsetDelta=0;

//     while(source<sourceLimit && U_SUCCESS(*pErrorCode)) {
//         switch(state) {
//         case 0:
//             cnv->toUBytes[0]=(uint8_t)*source++;
//             cnv->toULength=1;
//             state=1;
//             break;
//         case 1:
//             /*
//              * Only inside this switch case can the state variable
//              * temporarily take two additional values:
//              * 6: BOM error, continue with BE
//              * 7: BOM error, continue with LE
//              */
//             b=*source;
//             if(cnv->toUBytes[0]==0xfe && b==0xff) {
//                 if(IS_UTF16LE(cnv)) {
//                     state=7; /* illegal reverse BOM for Java "UnicodeLittle" */
//                 } else {
//                     state=8; /* detect UTF-16BE */
//                 }
//             } else if(cnv->toUBytes[0]==0xff && b==0xfe) {
//                 if(IS_UTF16BE(cnv)) {
//                     state=6; /* illegal reverse BOM for Java "UnicodeBig" */
//                 } else {
//                     state=9; /* detect UTF-16LE */
//                 }
//             } else if((IS_UTF16(cnv) && UCNV_GET_VERSION(cnv)==1)) {
//                 state=6; /* illegal missing BOM for Java "Unicode" */
//             }
//             if(state>=8) {
//                 /* BOM detected, consume it */
//                 ++source;
//                 cnv->toULength=0;
//                 offsetDelta=(int32_t)(source-pArgs->source);
//             } else if(state<6) {
//                 /* ok: no BOM, and not a reverse BOM */
//                 if(source!=pArgs->source) {
//                     /* reset the source for a correct first offset */
//                     source=pArgs->source;
//                     cnv->toULength=0;
//                 }
//                 if(IS_UTF16LE(cnv)) {
//                     /* Make Java "UnicodeLittle" default to LE. */
//                     state=9;
//                 } else {
//                     /* Make standard UTF-16 and Java "UnicodeBig" default to BE. */
//                     state=8;
//                 }
//             } else {
//                 /*
//                  * error: missing BOM, or reverse BOM
//                  * UTF-16,version=1: Java-specific "Unicode" requires a BOM.
//                  * UTF-16BE,version=1: Java-specific "UnicodeBig" requires a BE BOM or no BOM.
//                  * UTF-16LE,version=1: Java-specific "UnicodeLittle" requires an LE BOM or no BOM.
//                  */
//                 /* report the non-BOM or reverse BOM as an illegal sequence */
//                 cnv->toUBytes[1]=b;
//                 cnv->toULength=2;
//                 pArgs->source=source+1;
//                 /* continue with conversion if the callback resets the error */
//                 /*
//                  * Make Java "Unicode" default to BE like standard UTF-16.
//                  * Make Java "UnicodeBig" and "UnicodeLittle" default
//                  * to their normal endiannesses.
//                  */
//                 cnv->mode=state+2;
//                 *pErrorCode=U_ILLEGAL_ESCAPE_SEQUENCE;
//                 return;
//             }
//             /* convert the rest of the stream */
//             cnv->mode=state;
//             continue;
//         case 8:
//             /* call UTF-16BE */
//             pArgs->source=source;
//             _UTF16BEToUnicodeWithOffsets(pArgs, pErrorCode);
//             source=pArgs->source;
//             break;
//         case 9:
//             /* call UTF-16LE */
//             pArgs->source=source;
//             _UTF16LEToUnicodeWithOffsets(pArgs, pErrorCode);
//             source=pArgs->source;
//             break;
//         default:
//             break; /* does not occur */
//         }
//     }

//     /* add BOM size to offsets - see comment at offsetDelta declaration */
//     if(offsets!=NULL && offsetDelta!=0) {
//         int32_t *offsetsLimit=pArgs->offsets;
//         while(offsets<offsetsLimit) {
//             *offsets++ += offsetDelta;
//         }
//     }

//     pArgs->source=source;

//     if(source==sourceLimit && pArgs->flush) {
//         /* handle truncated input */
//         switch(state) {
//         case 0:
//             break; /* no input at all, nothing to do */
//         case 8:
//             _UTF16BEToUnicodeWithOffsets(pArgs, pErrorCode);
//             break;
//         case 9:
//             _UTF16LEToUnicodeWithOffsets(pArgs, pErrorCode);
//             break;
//         default:
//             /* 0<state<8: framework will report truncation, nothing to do here */
//             break;
//         }
//     }

//     cnv->mode=state;
// }

// static UChar32 U_CALLCONV
// _UTF16GetNextUChar(UConverterToUnicodeArgs *pArgs,
//                    UErrorCode *pErrorCode) {
//     switch(pArgs->converter->mode) {
//     case 8:
//         return _UTF16BEGetNextUChar(pArgs, pErrorCode);
//     case 9:
//         return _UTF16LEGetNextUChar(pArgs, pErrorCode);
//     default:
//         return UCNV_GET_NEXT_UCHAR_USE_TO_U;
//     }
// }
// U_CDECL_END

// static const UConverterImpl _UTF16Impl = {
//     UCNV_UTF16,

//     NULL,
//     NULL,

//     _UTF16Open,
//     NULL,
//     _UTF16Reset,

//     _UTF16ToUnicodeWithOffsets,
//     _UTF16ToUnicodeWithOffsets,
//     _UTF16PEFromUnicodeWithOffsets,
//     _UTF16PEFromUnicodeWithOffsets,
//     _UTF16GetNextUChar,

//     NULL, /* ### TODO implement getStarters for all Unicode encodings?! */
//     _UTF16GetName,
//     NULL,
//     NULL,
//     ucnv_getNonSurrogateUnicodeSet,

//     NULL,
//     NULL
// };

// static const UConverterStaticData _UTF16StaticData = {
//     sizeof(UConverterStaticData),
//     "UTF-16",
//     1204, /* CCSID for BOM sensitive UTF-16 */
//     UCNV_IBM, UCNV_UTF16, 2, 2,
// #if U_IS_BIG_ENDIAN
//     { 0xff, 0xfd, 0, 0 }, 2,
// #else
//     { 0xfd, 0xff, 0, 0 }, 2,
// #endif
//     FALSE, FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };

// const UConverterSharedData _UTF16Data =
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_UTF16StaticData, &_UTF16Impl);

// static const UConverterImpl _UTF16v2Impl = {
//     UCNV_UTF16,

//     NULL,
//     NULL,

//     _UTF16Open,
//     NULL,
//     _UTF16Reset,

//     _UTF16ToUnicodeWithOffsets,
//     _UTF16ToUnicodeWithOffsets,
//     _UTF16BEFromUnicodeWithOffsets,
//     _UTF16BEFromUnicodeWithOffsets,
//     _UTF16GetNextUChar,

//     NULL, /* ### TODO implement getStarters for all Unicode encodings?! */
//     _UTF16GetName,
//     NULL,
//     NULL,
//     ucnv_getNonSurrogateUnicodeSet,

//     NULL,
//     NULL
// };

// static const UConverterStaticData _UTF16v2StaticData = {
//     sizeof(UConverterStaticData),
//     "UTF-16,version=2",
//     1204, /* CCSID for BOM sensitive UTF-16 */
//     UCNV_IBM, UCNV_UTF16, 2, 2,
//     { 0xff, 0xfd, 0, 0 }, 2,
//     FALSE, FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };

// const UConverterSharedData _UTF16v2Data =
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_UTF16v2StaticData, &_UTF16v2Impl);

// #endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 2002-2015, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// **********************************************************************
// *   file name:  ucnv_u32.c
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2002jul01
// *   created by: Markus W. Scherer
// *
// *   UTF-32 converter implementation. Used to be in ucnv_utf.c.
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION && !UCONFIG_ONLY_HTML_CONVERSION

// #include "unicode/ucnv.h"
// #include "unicode/utf.h"
// #include "ucnv_bld.h"
// #include "ucnv_cnv.h"
// #include "cmemory.h"

// #define MAXIMUM_UCS2            0x0000FFFF
// #define MAXIMUM_UTF             0x0010FFFF
// #define HALF_SHIFT              10
// #define HALF_BASE               0x0010000
// #define HALF_MASK               0x3FF
// #define SURROGATE_HIGH_START    0xD800
// #define SURROGATE_LOW_START     0xDC00

// /* -SURROGATE_LOW_START + HALF_BASE */
// #define SURROGATE_LOW_BASE      9216

// enum {
//     UCNV_NEED_TO_WRITE_BOM=1
// };

// /* UTF-32BE ----------------------------------------------------------------- */
// U_CDECL_BEGIN
// static void U_CALLCONV
// T_UConverter_toUnicode_UTF32_BE(UConverterToUnicodeArgs * args,
//                                 UErrorCode * err)
// {
//     const unsigned char *mySource = (unsigned char *) args->source;
//     UChar *myTarget = args->target;
//     const unsigned char *sourceLimit = (unsigned char *) args->sourceLimit;
//     const UChar *targetLimit = args->targetLimit;
//     unsigned char *toUBytes = args->converter->toUBytes;
//     uint32_t ch, i;

//     /* Restore state of current sequence */
//     if (args->converter->toULength > 0 && myTarget < targetLimit) {
//         i = args->converter->toULength;       /* restore # of bytes consumed */
//         args->converter->toULength = 0;

//         ch = args->converter->toUnicodeStatus - 1;/*Stores the previously calculated ch from a previous call*/
//         args->converter->toUnicodeStatus = 0;
//         goto morebytes;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit) {
//         i = 0;
//         ch = 0;
// morebytes:
//         while (i < sizeof(uint32_t)) {
//             if (mySource < sourceLimit) {
//                 ch = (ch << 8) | (uint8_t)(*mySource);
//                 toUBytes[i++] = (char) *(mySource++);
//             }
//             else {
//                 /* stores a partially calculated target*/
//                 /* + 1 to make 0 a valid character */
//                 args->converter->toUnicodeStatus = ch + 1;
//                 args->converter->toULength = (int8_t) i;
//                 goto donefornow;
//             }
//         }

//         if (ch <= MAXIMUM_UTF && !U_IS_SURROGATE(ch)) {
//             /* Normal valid byte when the loop has not prematurely terminated (i < inBytes) */
//             if (ch <= MAXIMUM_UCS2)
//             {
//                 /* fits in 16 bits */
//                 *(myTarget++) = (UChar) ch;
//             }
//             else {
//                 /* write out the surrogates */
//                 *(myTarget++) = U16_LEAD(ch);
//                 ch = U16_TRAIL(ch);
//                 if (myTarget < targetLimit) {
//                     *(myTarget++) = (UChar)ch;
//                 }
//                 else {
//                     /* Put in overflow buffer (not handled here) */
//                     args->converter->UCharErrorBuffer[0] = (UChar) ch;
//                     args->converter->UCharErrorBufferLength = 1;
//                     *err = U_BUFFER_OVERFLOW_ERROR;
//                     break;
//                 }
//             }
//         }
//         else {
//             args->converter->toULength = (int8_t)i;
//             *err = U_ILLEGAL_CHAR_FOUND;
//             break;
//         }
//     }

// donefornow:
//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err)) {
//         /* End of target buffer */
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = myTarget;
//     args->source = (const char *) mySource;
// }

// static void U_CALLCONV
// T_UConverter_toUnicode_UTF32_BE_OFFSET_LOGIC(UConverterToUnicodeArgs * args,
//                                              UErrorCode * err)
// {
//     const unsigned char *mySource = (unsigned char *) args->source;
//     UChar *myTarget = args->target;
//     int32_t *myOffsets = args->offsets;
//     const unsigned char *sourceLimit = (unsigned char *) args->sourceLimit;
//     const UChar *targetLimit = args->targetLimit;
//     unsigned char *toUBytes = args->converter->toUBytes;
//     uint32_t ch, i;
//     int32_t offsetNum = 0;

//     /* Restore state of current sequence */
//     if (args->converter->toULength > 0 && myTarget < targetLimit) {
//         i = args->converter->toULength;       /* restore # of bytes consumed */
//         args->converter->toULength = 0;

//         ch = args->converter->toUnicodeStatus - 1;/*Stores the previously calculated ch from a previous call*/
//         args->converter->toUnicodeStatus = 0;
//         goto morebytes;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit) {
//         i = 0;
//         ch = 0;
// morebytes:
//         while (i < sizeof(uint32_t)) {
//             if (mySource < sourceLimit) {
//                 ch = (ch << 8) | (uint8_t)(*mySource);
//                 toUBytes[i++] = (char) *(mySource++);
//             }
//             else {
//                 /* stores a partially calculated target*/
//                 /* + 1 to make 0 a valid character */
//                 args->converter->toUnicodeStatus = ch + 1;
//                 args->converter->toULength = (int8_t) i;
//                 goto donefornow;
//             }
//         }

//         if (ch <= MAXIMUM_UTF && !U_IS_SURROGATE(ch)) {
//             /* Normal valid byte when the loop has not prematurely terminated (i < inBytes) */
//             if (ch <= MAXIMUM_UCS2) {
//                 /* fits in 16 bits */
//                 *(myTarget++) = (UChar) ch;
//                 *(myOffsets++) = offsetNum;
//             }
//             else {
//                 /* write out the surrogates */
//                 *(myTarget++) = U16_LEAD(ch);
//                 *myOffsets++ = offsetNum;
//                 ch = U16_TRAIL(ch);
//                 if (myTarget < targetLimit)
//                 {
//                     *(myTarget++) = (UChar)ch;
//                     *(myOffsets++) = offsetNum;
//                 }
//                 else {
//                     /* Put in overflow buffer (not handled here) */
//                     args->converter->UCharErrorBuffer[0] = (UChar) ch;
//                     args->converter->UCharErrorBufferLength = 1;
//                     *err = U_BUFFER_OVERFLOW_ERROR;
//                     break;
//                 }
//             }
//         }
//         else {
//             args->converter->toULength = (int8_t)i;
//             *err = U_ILLEGAL_CHAR_FOUND;
//             break;
//         }
//         offsetNum += i;
//     }

// donefornow:
//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err))
//     {
//         /* End of target buffer */
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = myTarget;
//     args->source = (const char *) mySource;
//     args->offsets = myOffsets;
// }

// static void U_CALLCONV
// T_UConverter_fromUnicode_UTF32_BE(UConverterFromUnicodeArgs * args,
//                                   UErrorCode * err)
// {
//     const UChar *mySource = args->source;
//     unsigned char *myTarget;
//     const UChar *sourceLimit = args->sourceLimit;
//     const unsigned char *targetLimit = (unsigned char *) args->targetLimit;
//     UChar32 ch, ch2;
//     unsigned int indexToWrite;
//     unsigned char temp[sizeof(uint32_t)];

//     if(mySource >= sourceLimit) {
//         /* no input, nothing to do */
//         return;
//     }

//     /* write the BOM if necessary */
//     if(args->converter->fromUnicodeStatus==UCNV_NEED_TO_WRITE_BOM) {
//         static const char bom[]={ 0, 0, (char)0xfeu, (char)0xffu };
//         ucnv_fromUWriteBytes(args->converter,
//                              bom, 4,
//                              &args->target, args->targetLimit,
//                              &args->offsets, -1,
//                              err);
//         args->converter->fromUnicodeStatus=0;
//     }

//     myTarget = (unsigned char *) args->target;
//     temp[0] = 0;

//     if (args->converter->fromUChar32) {
//         ch = args->converter->fromUChar32;
//         args->converter->fromUChar32 = 0;
//         goto lowsurogate;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit) {
//         ch = *(mySource++);

//         if (U_IS_SURROGATE(ch)) {
//             if (U_IS_LEAD(ch)) {
// lowsurogate:
//                 if (mySource < sourceLimit) {
//                     ch2 = *mySource;
//                     if (U_IS_TRAIL(ch2)) {
//                         ch = ((ch - SURROGATE_HIGH_START) << HALF_SHIFT) + ch2 + SURROGATE_LOW_BASE;
//                         mySource++;
//                     }
//                     else {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         args->converter->fromUChar32 = ch;
//                         *err = U_ILLEGAL_CHAR_FOUND;
//                         break;
//                     }
//                 }
//                 else {
//                     /* ran out of source */
//                     args->converter->fromUChar32 = ch;
//                     if (args->flush) {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         *err = U_ILLEGAL_CHAR_FOUND;
//                     }
//                     break;
//                 }
//             }
//             else {
//                 /* this is an unmatched trail code unit (2nd surrogate) */
//                 /* callback(illegal) */
//                 args->converter->fromUChar32 = ch;
//                 *err = U_ILLEGAL_CHAR_FOUND;
//                 break;
//             }
//         }

//         /* We cannot get any larger than 10FFFF because we are coming from UTF-16 */
//         temp[1] = (uint8_t) (ch >> 16 & 0x1F);
//         temp[2] = (uint8_t) (ch >> 8);  /* unsigned cast implicitly does (ch & FF) */
//         temp[3] = (uint8_t) (ch);       /* unsigned cast implicitly does (ch & FF) */

//         for (indexToWrite = 0; indexToWrite <= sizeof(uint32_t) - 1; indexToWrite++) {
//             if (myTarget < targetLimit) {
//                 *(myTarget++) = temp[indexToWrite];
//             }
//             else {
//                 args->converter->charErrorBuffer[args->converter->charErrorBufferLength++] = temp[indexToWrite];
//                 *err = U_BUFFER_OVERFLOW_ERROR;
//             }
//         }
//     }

//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err)) {
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = (char *) myTarget;
//     args->source = mySource;
// }

// static void U_CALLCONV
// T_UConverter_fromUnicode_UTF32_BE_OFFSET_LOGIC(UConverterFromUnicodeArgs * args,
//                                                UErrorCode * err)
// {
//     const UChar *mySource = args->source;
//     unsigned char *myTarget;
//     int32_t *myOffsets;
//     const UChar *sourceLimit = args->sourceLimit;
//     const unsigned char *targetLimit = (unsigned char *) args->targetLimit;
//     UChar32 ch, ch2;
//     int32_t offsetNum = 0;
//     unsigned int indexToWrite;
//     unsigned char temp[sizeof(uint32_t)];

//     if(mySource >= sourceLimit) {
//         /* no input, nothing to do */
//         return;
//     }

//     /* write the BOM if necessary */
//     if(args->converter->fromUnicodeStatus==UCNV_NEED_TO_WRITE_BOM) {
//         static const char bom[]={ 0, 0, (char)0xfeu, (char)0xffu };
//         ucnv_fromUWriteBytes(args->converter,
//                              bom, 4,
//                              &args->target, args->targetLimit,
//                              &args->offsets, -1,
//                              err);
//         args->converter->fromUnicodeStatus=0;
//     }

//     myTarget = (unsigned char *) args->target;
//     myOffsets = args->offsets;
//     temp[0] = 0;

//     if (args->converter->fromUChar32) {
//         ch = args->converter->fromUChar32;
//         args->converter->fromUChar32 = 0;
//         goto lowsurogate;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit) {
//         ch = *(mySource++);

//         if (U_IS_SURROGATE(ch)) {
//             if (U_IS_LEAD(ch)) {
// lowsurogate:
//                 if (mySource < sourceLimit) {
//                     ch2 = *mySource;
//                     if (U_IS_TRAIL(ch2)) {
//                         ch = ((ch - SURROGATE_HIGH_START) << HALF_SHIFT) + ch2 + SURROGATE_LOW_BASE;
//                         mySource++;
//                     }
//                     else {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         args->converter->fromUChar32 = ch;
//                         *err = U_ILLEGAL_CHAR_FOUND;
//                         break;
//                     }
//                 }
//                 else {
//                     /* ran out of source */
//                     args->converter->fromUChar32 = ch;
//                     if (args->flush) {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         *err = U_ILLEGAL_CHAR_FOUND;
//                     }
//                     break;
//                 }
//             }
//             else {
//                 /* this is an unmatched trail code unit (2nd surrogate) */
//                 /* callback(illegal) */
//                 args->converter->fromUChar32 = ch;
//                 *err = U_ILLEGAL_CHAR_FOUND;
//                 break;
//             }
//         }

//         /* We cannot get any larger than 10FFFF because we are coming from UTF-16 */
//         temp[1] = (uint8_t) (ch >> 16 & 0x1F);
//         temp[2] = (uint8_t) (ch >> 8);  /* unsigned cast implicitly does (ch & FF) */
//         temp[3] = (uint8_t) (ch);       /* unsigned cast implicitly does (ch & FF) */

//         for (indexToWrite = 0; indexToWrite <= sizeof(uint32_t) - 1; indexToWrite++) {
//             if (myTarget < targetLimit) {
//                 *(myTarget++) = temp[indexToWrite];
//                 *(myOffsets++) = offsetNum;
//             }
//             else {
//                 args->converter->charErrorBuffer[args->converter->charErrorBufferLength++] = temp[indexToWrite];
//                 *err = U_BUFFER_OVERFLOW_ERROR;
//             }
//         }
//         offsetNum = offsetNum + 1 + (temp[1] != 0);
//     }

//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err)) {
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = (char *) myTarget;
//     args->source = mySource;
//     args->offsets = myOffsets;
// }

// static UChar32 U_CALLCONV
// T_UConverter_getNextUChar_UTF32_BE(UConverterToUnicodeArgs* args,
//                                    UErrorCode* err)
// {
//     const uint8_t *mySource;
//     UChar32 myUChar;
//     int32_t length;

//     mySource = (const uint8_t *)args->source;
//     if (mySource >= (const uint8_t *)args->sourceLimit)
//     {
//         /* no input */
//         *err = U_INDEX_OUTOFBOUNDS_ERROR;
//         return 0xffff;
//     }

//     length = (int32_t)((const uint8_t *)args->sourceLimit - mySource);
//     if (length < 4)
//     {
//         /* got a partial character */
//         uprv_memcpy(args->converter->toUBytes, mySource, length);
//         args->converter->toULength = (int8_t)length;
//         args->source = (const char *)(mySource + length);
//         *err = U_TRUNCATED_CHAR_FOUND;
//         return 0xffff;
//     }

//     /* Don't even try to do a direct cast because the value may be on an odd address. */
//     myUChar = ((UChar32)mySource[0] << 24)
//             | ((UChar32)mySource[1] << 16)
//             | ((UChar32)mySource[2] << 8)
//             | ((UChar32)mySource[3]);

//     args->source = (const char *)(mySource + 4);
//     if ((uint32_t)myUChar <= MAXIMUM_UTF && !U_IS_SURROGATE(myUChar)) {
//         return myUChar;
//     }

//     uprv_memcpy(args->converter->toUBytes, mySource, 4);
//     args->converter->toULength = 4;

//     *err = U_ILLEGAL_CHAR_FOUND;
//     return 0xffff;
// }
// U_CDECL_END
// static const UConverterImpl _UTF32BEImpl = {
//     UCNV_UTF32_BigEndian,

//     NULL,
//     NULL,

//     NULL,
//     NULL,
//     NULL,

//     T_UConverter_toUnicode_UTF32_BE,
//     T_UConverter_toUnicode_UTF32_BE_OFFSET_LOGIC,
//     T_UConverter_fromUnicode_UTF32_BE,
//     T_UConverter_fromUnicode_UTF32_BE_OFFSET_LOGIC,
//     T_UConverter_getNextUChar_UTF32_BE,

//     NULL,
//     NULL,
//     NULL,
//     NULL,
//     ucnv_getNonSurrogateUnicodeSet,

//     NULL,
//     NULL
// };

// /* The 1232 CCSID refers to any version of Unicode with any endianess of UTF-32 */
// static const UConverterStaticData _UTF32BEStaticData = {
//     sizeof(UConverterStaticData),
//     "UTF-32BE",
//     1232,
//     UCNV_IBM, UCNV_UTF32_BigEndian, 4, 4,
//     { 0, 0, 0xff, 0xfd }, 4, FALSE, FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };

// const UConverterSharedData _UTF32BEData =
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_UTF32BEStaticData, &_UTF32BEImpl);

// /* UTF-32LE ---------------------------------------------------------- */
// U_CDECL_BEGIN
// static void U_CALLCONV
// T_UConverter_toUnicode_UTF32_LE(UConverterToUnicodeArgs * args,
//                                 UErrorCode * err)
// {
//     const unsigned char *mySource = (unsigned char *) args->source;
//     UChar *myTarget = args->target;
//     const unsigned char *sourceLimit = (unsigned char *) args->sourceLimit;
//     const UChar *targetLimit = args->targetLimit;
//     unsigned char *toUBytes = args->converter->toUBytes;
//     uint32_t ch, i;

//     /* Restore state of current sequence */
//     if (args->converter->toULength > 0 && myTarget < targetLimit)
//     {
//         i = args->converter->toULength;       /* restore # of bytes consumed */
//         args->converter->toULength = 0;

//         /* Stores the previously calculated ch from a previous call*/
//         ch = args->converter->toUnicodeStatus - 1;
//         args->converter->toUnicodeStatus = 0;
//         goto morebytes;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit)
//     {
//         i = 0;
//         ch = 0;
// morebytes:
//         while (i < sizeof(uint32_t))
//         {
//             if (mySource < sourceLimit)
//             {
//                 ch |= ((uint8_t)(*mySource)) << (i * 8);
//                 toUBytes[i++] = (char) *(mySource++);
//             }
//             else
//             {
//                 /* stores a partially calculated target*/
//                 /* + 1 to make 0 a valid character */
//                 args->converter->toUnicodeStatus = ch + 1;
//                 args->converter->toULength = (int8_t) i;
//                 goto donefornow;
//             }
//         }

//         if (ch <= MAXIMUM_UTF && !U_IS_SURROGATE(ch)) {
//             /* Normal valid byte when the loop has not prematurely terminated (i < inBytes) */
//             if (ch <= MAXIMUM_UCS2) {
//                 /* fits in 16 bits */
//                 *(myTarget++) = (UChar) ch;
//             }
//             else {
//                 /* write out the surrogates */
//                 *(myTarget++) = U16_LEAD(ch);
//                 ch = U16_TRAIL(ch);
//                 if (myTarget < targetLimit) {
//                     *(myTarget++) = (UChar)ch;
//                 }
//                 else {
//                     /* Put in overflow buffer (not handled here) */
//                     args->converter->UCharErrorBuffer[0] = (UChar) ch;
//                     args->converter->UCharErrorBufferLength = 1;
//                     *err = U_BUFFER_OVERFLOW_ERROR;
//                     break;
//                 }
//             }
//         }
//         else {
//             args->converter->toULength = (int8_t)i;
//             *err = U_ILLEGAL_CHAR_FOUND;
//             break;
//         }
//     }

// donefornow:
//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err))
//     {
//         /* End of target buffer */
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = myTarget;
//     args->source = (const char *) mySource;
// }

// static void U_CALLCONV
// T_UConverter_toUnicode_UTF32_LE_OFFSET_LOGIC(UConverterToUnicodeArgs * args,
//                                              UErrorCode * err)
// {
//     const unsigned char *mySource = (unsigned char *) args->source;
//     UChar *myTarget = args->target;
//     int32_t *myOffsets = args->offsets;
//     const unsigned char *sourceLimit = (unsigned char *) args->sourceLimit;
//     const UChar *targetLimit = args->targetLimit;
//     unsigned char *toUBytes = args->converter->toUBytes;
//     uint32_t ch, i;
//     int32_t offsetNum = 0;

//     /* Restore state of current sequence */
//     if (args->converter->toULength > 0 && myTarget < targetLimit)
//     {
//         i = args->converter->toULength;       /* restore # of bytes consumed */
//         args->converter->toULength = 0;

//         /* Stores the previously calculated ch from a previous call*/
//         ch = args->converter->toUnicodeStatus - 1;
//         args->converter->toUnicodeStatus = 0;
//         goto morebytes;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit)
//     {
//         i = 0;
//         ch = 0;
// morebytes:
//         while (i < sizeof(uint32_t))
//         {
//             if (mySource < sourceLimit)
//             {
//                 ch |= ((uint8_t)(*mySource)) << (i * 8);
//                 toUBytes[i++] = (char) *(mySource++);
//             }
//             else
//             {
//                 /* stores a partially calculated target*/
//                 /* + 1 to make 0 a valid character */
//                 args->converter->toUnicodeStatus = ch + 1;
//                 args->converter->toULength = (int8_t) i;
//                 goto donefornow;
//             }
//         }

//         if (ch <= MAXIMUM_UTF && !U_IS_SURROGATE(ch))
//         {
//             /* Normal valid byte when the loop has not prematurely terminated (i < inBytes) */
//             if (ch <= MAXIMUM_UCS2)
//             {
//                 /* fits in 16 bits */
//                 *(myTarget++) = (UChar) ch;
//                 *(myOffsets++) = offsetNum;
//             }
//             else {
//                 /* write out the surrogates */
//                 *(myTarget++) = U16_LEAD(ch);
//                 *(myOffsets++) = offsetNum;
//                 ch = U16_TRAIL(ch);
//                 if (myTarget < targetLimit)
//                 {
//                     *(myTarget++) = (UChar)ch;
//                     *(myOffsets++) = offsetNum;
//                 }
//                 else
//                 {
//                     /* Put in overflow buffer (not handled here) */
//                     args->converter->UCharErrorBuffer[0] = (UChar) ch;
//                     args->converter->UCharErrorBufferLength = 1;
//                     *err = U_BUFFER_OVERFLOW_ERROR;
//                     break;
//                 }
//             }
//         }
//         else
//         {
//             args->converter->toULength = (int8_t)i;
//             *err = U_ILLEGAL_CHAR_FOUND;
//             break;
//         }
//         offsetNum += i;
//     }

// donefornow:
//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err))
//     {
//         /* End of target buffer */
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = myTarget;
//     args->source = (const char *) mySource;
//     args->offsets = myOffsets;
// }

// static void U_CALLCONV
// T_UConverter_fromUnicode_UTF32_LE(UConverterFromUnicodeArgs * args,
//                                   UErrorCode * err)
// {
//     const UChar *mySource = args->source;
//     unsigned char *myTarget;
//     const UChar *sourceLimit = args->sourceLimit;
//     const unsigned char *targetLimit = (unsigned char *) args->targetLimit;
//     UChar32 ch, ch2;
//     unsigned int indexToWrite;
//     unsigned char temp[sizeof(uint32_t)];

//     if(mySource >= sourceLimit) {
//         /* no input, nothing to do */
//         return;
//     }

//     /* write the BOM if necessary */
//     if(args->converter->fromUnicodeStatus==UCNV_NEED_TO_WRITE_BOM) {
//         static const char bom[]={ (char)0xffu, (char)0xfeu, 0, 0 };
//         ucnv_fromUWriteBytes(args->converter,
//                              bom, 4,
//                              &args->target, args->targetLimit,
//                              &args->offsets, -1,
//                              err);
//         args->converter->fromUnicodeStatus=0;
//     }

//     myTarget = (unsigned char *) args->target;
//     temp[3] = 0;

//     if (args->converter->fromUChar32)
//     {
//         ch = args->converter->fromUChar32;
//         args->converter->fromUChar32 = 0;
//         goto lowsurogate;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit)
//     {
//         ch = *(mySource++);

//         if (U16_IS_SURROGATE(ch)) {
//             if (U16_IS_LEAD(ch))
//             {
// lowsurogate:
//                 if (mySource < sourceLimit)
//                 {
//                     ch2 = *mySource;
//                     if (U16_IS_TRAIL(ch2)) {
//                         ch = ((ch - SURROGATE_HIGH_START) << HALF_SHIFT) + ch2 + SURROGATE_LOW_BASE;
//                         mySource++;
//                     }
//                     else {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         args->converter->fromUChar32 = ch;
//                         *err = U_ILLEGAL_CHAR_FOUND;
//                         break;
//                     }
//                 }
//                 else {
//                     /* ran out of source */
//                     args->converter->fromUChar32 = ch;
//                     if (args->flush) {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         *err = U_ILLEGAL_CHAR_FOUND;
//                     }
//                     break;
//                 }
//             }
//             else {
//                 /* this is an unmatched trail code unit (2nd surrogate) */
//                 /* callback(illegal) */
//                 args->converter->fromUChar32 = ch;
//                 *err = U_ILLEGAL_CHAR_FOUND;
//                 break;
//             }
//         }

//         /* We cannot get any larger than 10FFFF because we are coming from UTF-16 */
//         temp[2] = (uint8_t) (ch >> 16 & 0x1F);
//         temp[1] = (uint8_t) (ch >> 8);  /* unsigned cast implicitly does (ch & FF) */
//         temp[0] = (uint8_t) (ch);       /* unsigned cast implicitly does (ch & FF) */

//         for (indexToWrite = 0; indexToWrite <= sizeof(uint32_t) - 1; indexToWrite++)
//         {
//             if (myTarget < targetLimit)
//             {
//                 *(myTarget++) = temp[indexToWrite];
//             }
//             else
//             {
//                 args->converter->charErrorBuffer[args->converter->charErrorBufferLength++] = temp[indexToWrite];
//                 *err = U_BUFFER_OVERFLOW_ERROR;
//             }
//         }
//     }

//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err))
//     {
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = (char *) myTarget;
//     args->source = mySource;
// }

// static void U_CALLCONV
// T_UConverter_fromUnicode_UTF32_LE_OFFSET_LOGIC(UConverterFromUnicodeArgs * args,
//                                                UErrorCode * err)
// {
//     const UChar *mySource = args->source;
//     unsigned char *myTarget;
//     int32_t *myOffsets;
//     const UChar *sourceLimit = args->sourceLimit;
//     const unsigned char *targetLimit = (unsigned char *) args->targetLimit;
//     UChar32 ch, ch2;
//     unsigned int indexToWrite;
//     unsigned char temp[sizeof(uint32_t)];
//     int32_t offsetNum = 0;

//     if(mySource >= sourceLimit) {
//         /* no input, nothing to do */
//         return;
//     }

//     /* write the BOM if necessary */
//     if(args->converter->fromUnicodeStatus==UCNV_NEED_TO_WRITE_BOM) {
//         static const char bom[]={ (char)0xffu, (char)0xfeu, 0, 0 };
//         ucnv_fromUWriteBytes(args->converter,
//                              bom, 4,
//                              &args->target, args->targetLimit,
//                              &args->offsets, -1,
//                              err);
//         args->converter->fromUnicodeStatus=0;
//     }

//     myTarget = (unsigned char *) args->target;
//     myOffsets = args->offsets;
//     temp[3] = 0;

//     if (args->converter->fromUChar32)
//     {
//         ch = args->converter->fromUChar32;
//         args->converter->fromUChar32 = 0;
//         goto lowsurogate;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit)
//     {
//         ch = *(mySource++);

//         if (U16_IS_SURROGATE(ch)) {
//             if (U16_IS_LEAD(ch))
//             {
// lowsurogate:
//                 if (mySource < sourceLimit)
//                 {
//                     ch2 = *mySource;
//                     if (U16_IS_TRAIL(ch2))
//                     {
//                         ch = ((ch - SURROGATE_HIGH_START) << HALF_SHIFT) + ch2 + SURROGATE_LOW_BASE;
//                         mySource++;
//                     }
//                     else {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         args->converter->fromUChar32 = ch;
//                         *err = U_ILLEGAL_CHAR_FOUND;
//                         break;
//                     }
//                 }
//                 else {
//                     /* ran out of source */
//                     args->converter->fromUChar32 = ch;
//                     if (args->flush) {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         *err = U_ILLEGAL_CHAR_FOUND;
//                     }
//                     break;
//                 }
//             }
//             else {
//                 /* this is an unmatched trail code unit (2nd surrogate) */
//                 /* callback(illegal) */
//                 args->converter->fromUChar32 = ch;
//                 *err = U_ILLEGAL_CHAR_FOUND;
//                 break;
//             }
//         }

//         /* We cannot get any larger than 10FFFF because we are coming from UTF-16 */
//         temp[2] = (uint8_t) (ch >> 16 & 0x1F);
//         temp[1] = (uint8_t) (ch >> 8);  /* unsigned cast implicitly does (ch & FF) */
//         temp[0] = (uint8_t) (ch);       /* unsigned cast implicitly does (ch & FF) */

//         for (indexToWrite = 0; indexToWrite <= sizeof(uint32_t) - 1; indexToWrite++)
//         {
//             if (myTarget < targetLimit)
//             {
//                 *(myTarget++) = temp[indexToWrite];
//                 *(myOffsets++) = offsetNum;
//             }
//             else
//             {
//                 args->converter->charErrorBuffer[args->converter->charErrorBufferLength++] = temp[indexToWrite];
//                 *err = U_BUFFER_OVERFLOW_ERROR;
//             }
//         }
//         offsetNum = offsetNum + 1 + (temp[2] != 0);
//     }

//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err))
//     {
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = (char *) myTarget;
//     args->source = mySource;
//     args->offsets = myOffsets;
// }

// static UChar32 U_CALLCONV
// T_UConverter_getNextUChar_UTF32_LE(UConverterToUnicodeArgs* args,
//                                    UErrorCode* err)
// {
//     const uint8_t *mySource;
//     UChar32 myUChar;
//     int32_t length;

//     mySource = (const uint8_t *)args->source;
//     if (mySource >= (const uint8_t *)args->sourceLimit)
//     {
//         /* no input */
//         *err = U_INDEX_OUTOFBOUNDS_ERROR;
//         return 0xffff;
//     }

//     length = (int32_t)((const uint8_t *)args->sourceLimit - mySource);
//     if (length < 4)
//     {
//         /* got a partial character */
//         uprv_memcpy(args->converter->toUBytes, mySource, length);
//         args->converter->toULength = (int8_t)length;
//         args->source = (const char *)(mySource + length);
//         *err = U_TRUNCATED_CHAR_FOUND;
//         return 0xffff;
//     }

//     /* Don't even try to do a direct cast because the value may be on an odd address. */
//     myUChar = ((UChar32)mySource[3] << 24)
//             | ((UChar32)mySource[2] << 16)
//             | ((UChar32)mySource[1] << 8)
//             | ((UChar32)mySource[0]);

//     args->source = (const char *)(mySource + 4);
//     if ((uint32_t)myUChar <= MAXIMUM_UTF && !U_IS_SURROGATE(myUChar)) {
//         return myUChar;
//     }

//     uprv_memcpy(args->converter->toUBytes, mySource, 4);
//     args->converter->toULength = 4;

//     *err = U_ILLEGAL_CHAR_FOUND;
//     return 0xffff;
// }
// U_CDECL_END
// static const UConverterImpl _UTF32LEImpl = {
//     UCNV_UTF32_LittleEndian,

//     NULL,
//     NULL,

//     NULL,
//     NULL,
//     NULL,

//     T_UConverter_toUnicode_UTF32_LE,
//     T_UConverter_toUnicode_UTF32_LE_OFFSET_LOGIC,
//     T_UConverter_fromUnicode_UTF32_LE,
//     T_UConverter_fromUnicode_UTF32_LE_OFFSET_LOGIC,
//     T_UConverter_getNextUChar_UTF32_LE,

//     NULL,
//     NULL,
//     NULL,
//     NULL,
//     ucnv_getNonSurrogateUnicodeSet,

//     NULL,
//     NULL
// };

// /* The 1232 CCSID refers to any version of Unicode with any endianess of UTF-32 */
// static const UConverterStaticData _UTF32LEStaticData = {
//     sizeof(UConverterStaticData),
//     "UTF-32LE",
//     1234,
//     UCNV_IBM, UCNV_UTF32_LittleEndian, 4, 4,
//     { 0xfd, 0xff, 0, 0 }, 4, FALSE, FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };


// const UConverterSharedData _UTF32LEData =
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_UTF32LEStaticData, &_UTF32LEImpl);

// /* UTF-32 (Detect BOM) ------------------------------------------------------ */

// /*
//  * Detect a BOM at the beginning of the stream and select UTF-32BE or UTF-32LE
//  * accordingly.
//  *
//  * State values:
//  * 0    initial state
//  * 1    saw 00
//  * 2    saw 00 00
//  * 3    saw 00 00 FE
//  * 4    -
//  * 5    saw FF
//  * 6    saw FF FE
//  * 7    saw FF FE 00
//  * 8    UTF-32BE mode
//  * 9    UTF-32LE mode
//  *
//  * During detection: state&3==number of matching bytes so far.
//  *
//  * On output, emit U+FEFF as the first code point.
//  */
// U_CDECL_BEGIN
// static void U_CALLCONV
// _UTF32Reset(UConverter *cnv, UConverterResetChoice choice) {
//     if(choice<=UCNV_RESET_TO_UNICODE) {
//         /* reset toUnicode: state=0 */
//         cnv->mode=0;
//     }
//     if(choice!=UCNV_RESET_TO_UNICODE) {
//         /* reset fromUnicode: prepare to output the UTF-32PE BOM */
//         cnv->fromUnicodeStatus=UCNV_NEED_TO_WRITE_BOM;
//     }
// }

// static void U_CALLCONV
// _UTF32Open(UConverter *cnv,
//            UConverterLoadArgs *pArgs,
//            UErrorCode *pErrorCode) {
//     (void)pArgs;
//     (void)pErrorCode;
//     _UTF32Reset(cnv, UCNV_RESET_BOTH);
// }

// static const char utf32BOM[8]={ 0, 0, (char)0xfeu, (char)0xffu, (char)0xffu, (char)0xfeu, 0, 0 };

// static void U_CALLCONV
// _UTF32ToUnicodeWithOffsets(UConverterToUnicodeArgs *pArgs,
//                            UErrorCode *pErrorCode) {
//     UConverter *cnv=pArgs->converter;
//     const char *source=pArgs->source;
//     const char *sourceLimit=pArgs->sourceLimit;
//     int32_t *offsets=pArgs->offsets;

//     int32_t state, offsetDelta;
//     char b;

//     state=cnv->mode;

//     /*
//      * If we detect a BOM in this buffer, then we must add the BOM size to the
//      * offsets because the actual converter function will not see and count the BOM.
//      * offsetDelta will have the number of the BOM bytes that are in the current buffer.
//      */
//     offsetDelta=0;

//     while(source<sourceLimit && U_SUCCESS(*pErrorCode)) {
//         switch(state) {
//         case 0:
//             b=*source;
//             if(b==0) {
//                 state=1; /* could be 00 00 FE FF */
//             } else if(b==(char)0xffu) {
//                 state=5; /* could be FF FE 00 00 */
//             } else {
//                 state=8; /* default to UTF-32BE */
//                 continue;
//             }
//             ++source;
//             break;
//         case 1:
//         case 2:
//         case 3:
//         case 5:
//         case 6:
//         case 7:
//             if(*source==utf32BOM[state]) {
//                 ++state;
//                 ++source;
//                 if(state==4) {
//                     state=8; /* detect UTF-32BE */
//                     offsetDelta=(int32_t)(source-pArgs->source);
//                 } else if(state==8) {
//                     state=9; /* detect UTF-32LE */
//                     offsetDelta=(int32_t)(source-pArgs->source);
//                 }
//             } else {
//                 /* switch to UTF-32BE and pass the previous bytes */
//                 int32_t count=(int32_t)(source-pArgs->source); /* number of bytes from this buffer */

//                 /* reset the source */
//                 source=pArgs->source;

//                 if(count==(state&3)) {
//                     /* simple: all in the same buffer, just reset source */
//                 } else {
//                     UBool oldFlush=pArgs->flush;

//                     /* some of the bytes are from a previous buffer, replay those first */
//                     pArgs->source=utf32BOM+(state&4); /* select the correct BOM */
//                     pArgs->sourceLimit=pArgs->source+((state&3)-count); /* replay previous bytes */
//                     pArgs->flush=FALSE; /* this sourceLimit is not the real source stream limit */

//                     /* no offsets: bytes from previous buffer, and not enough for output */
//                     T_UConverter_toUnicode_UTF32_BE(pArgs, pErrorCode);

//                     /* restore real pointers; pArgs->source will be set in case 8/9 */
//                     pArgs->sourceLimit=sourceLimit;
//                     pArgs->flush=oldFlush;
//                 }
//                 state=8;
//                 continue;
//             }
//             break;
//         case 8:
//             /* call UTF-32BE */
//             pArgs->source=source;
//             if(offsets==NULL) {
//                 T_UConverter_toUnicode_UTF32_BE(pArgs, pErrorCode);
//             } else {
//                 T_UConverter_toUnicode_UTF32_BE_OFFSET_LOGIC(pArgs, pErrorCode);
//             }
//             source=pArgs->source;
//             break;
//         case 9:
//             /* call UTF-32LE */
//             pArgs->source=source;
//             if(offsets==NULL) {
//                 T_UConverter_toUnicode_UTF32_LE(pArgs, pErrorCode);
//             } else {
//                 T_UConverter_toUnicode_UTF32_LE_OFFSET_LOGIC(pArgs, pErrorCode);
//             }
//             source=pArgs->source;
//             break;
//         default:
//             break; /* does not occur */
//         }
//     }

//     /* add BOM size to offsets - see comment at offsetDelta declaration */
//     if(offsets!=NULL && offsetDelta!=0) {
//         int32_t *offsetsLimit=pArgs->offsets;
//         while(offsets<offsetsLimit) {
//             *offsets++ += offsetDelta;
//         }
//     }

//     pArgs->source=source;

//     if(source==sourceLimit && pArgs->flush) {
//         /* handle truncated input */
//         switch(state) {
//         case 0:
//             break; /* no input at all, nothing to do */
//         case 8:
//             T_UConverter_toUnicode_UTF32_BE(pArgs, pErrorCode);
//             break;
//         case 9:
//             T_UConverter_toUnicode_UTF32_LE(pArgs, pErrorCode);
//             break;
//         default:
//             /* handle 0<state<8: call UTF-32BE with too-short input */
//             pArgs->source=utf32BOM+(state&4); /* select the correct BOM */
//             pArgs->sourceLimit=pArgs->source+(state&3); /* replay bytes */

//             /* no offsets: not enough for output */
//             T_UConverter_toUnicode_UTF32_BE(pArgs, pErrorCode);
//             pArgs->source=source;
//             pArgs->sourceLimit=sourceLimit;
//             state=8;
//             break;
//         }
//     }

//     cnv->mode=state;
// }

// static UChar32 U_CALLCONV
// _UTF32GetNextUChar(UConverterToUnicodeArgs *pArgs,
//                    UErrorCode *pErrorCode) {
//     switch(pArgs->converter->mode) {
//     case 8:
//         return T_UConverter_getNextUChar_UTF32_BE(pArgs, pErrorCode);
//     case 9:
//         return T_UConverter_getNextUChar_UTF32_LE(pArgs, pErrorCode);
//     default:
//         return UCNV_GET_NEXT_UCHAR_USE_TO_U;
//     }
// }
// U_CDECL_END
// static const UConverterImpl _UTF32Impl = {
//     UCNV_UTF32,

//     NULL,
//     NULL,

//     _UTF32Open,
//     NULL,
//     _UTF32Reset,

//     _UTF32ToUnicodeWithOffsets,
//     _UTF32ToUnicodeWithOffsets,
// #if U_IS_BIG_ENDIAN
//     T_UConverter_fromUnicode_UTF32_BE,
//     T_UConverter_fromUnicode_UTF32_BE_OFFSET_LOGIC,
// #else
//     T_UConverter_fromUnicode_UTF32_LE,
//     T_UConverter_fromUnicode_UTF32_LE_OFFSET_LOGIC,
// #endif
//     _UTF32GetNextUChar,

//     NULL, /* ### TODO implement getStarters for all Unicode encodings?! */
//     NULL,
//     NULL,
//     NULL,
//     ucnv_getNonSurrogateUnicodeSet,

//     NULL,
//     NULL
// };

// /* The 1236 CCSID refers to any version of Unicode with a BOM sensitive endianess of UTF-32 */
// static const UConverterStaticData _UTF32StaticData = {
//     sizeof(UConverterStaticData),
//     "UTF-32",
//     1236,
//     UCNV_IBM, UCNV_UTF32, 4, 4,
// #if U_IS_BIG_ENDIAN
//     { 0, 0, 0xff, 0xfd }, 4,
// #else
//     { 0xfd, 0xff, 0, 0 }, 4,
// #endif
//     FALSE, FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };

// const UConverterSharedData _UTF32Data =
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_UTF32StaticData, &_UTF32Impl);

// #endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 2002-2016, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// **********************************************************************
// *   file name:  ucnv_u8.c
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2002jul01
// *   created by: Markus W. Scherer
// *
// *   UTF-8 converter implementation. Used to be in ucnv_utf.c.
// *
// *   Also, CESU-8 implementation, see UTR 26.
// *   The CESU-8 converter uses all the same functions as the
// *   UTF-8 converter, with a branch for converting supplementary code points.
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "unicode/ucnv.h"
// #include "unicode/utf.h"
// #include "unicode/utf8.h"
// #include "unicode/utf16.h"
// #include "uassert.h"
// #include "ucnv_bld.h"
// #include "ucnv_cnv.h"
// #include "cmemory.h"
// #include "ustr_imp.h"

// /* Prototypes --------------------------------------------------------------- */

// /* Keep these here to make finicky compilers happy */

// U_CFUNC void ucnv_fromUnicode_UTF8(UConverterFromUnicodeArgs *args,
//                                            UErrorCode *err);
// U_CFUNC void ucnv_fromUnicode_UTF8_OFFSETS_LOGIC(UConverterFromUnicodeArgs *args,
//                                                         UErrorCode *err);


// /* UTF-8 -------------------------------------------------------------------- */

// #define MAXIMUM_UCS2            0x0000FFFF

// static const uint32_t offsetsFromUTF8[5] = {0,
//   (uint32_t) 0x00000000, (uint32_t) 0x00003080, (uint32_t) 0x000E2080,
//   (uint32_t) 0x03C82080
// };

// static UBool hasCESU8Data(const UConverter *cnv)
// {
// #if UCONFIG_ONLY_HTML_CONVERSION
//     return FALSE;
// #else
//     return (UBool)(cnv->sharedData == &_CESU8Data);
// #endif
// }
// U_CDECL_BEGIN
// static void  U_CALLCONV ucnv_toUnicode_UTF8 (UConverterToUnicodeArgs * args,
//                                   UErrorCode * err)
// {
//     UConverter *cnv = args->converter;
//     const unsigned char *mySource = (unsigned char *) args->source;
//     UChar *myTarget = args->target;
//     const unsigned char *sourceLimit = (unsigned char *) args->sourceLimit;
//     const UChar *targetLimit = args->targetLimit;
//     unsigned char *toUBytes = cnv->toUBytes;
//     UBool isCESU8 = hasCESU8Data(cnv);
//     uint32_t ch, ch2 = 0;
//     int32_t i, inBytes;

//     /* Restore size of current sequence */
//     if (cnv->toULength > 0 && myTarget < targetLimit)
//     {
//         inBytes = cnv->mode;            /* restore # of bytes to consume */
//         i = cnv->toULength;             /* restore # of bytes consumed */
//         cnv->toULength = 0;

//         ch = cnv->toUnicodeStatus;/*Stores the previously calculated ch from a previous call*/
//         cnv->toUnicodeStatus = 0;
//         goto morebytes;
//     }


//     while (mySource < sourceLimit && myTarget < targetLimit)
//     {
//         ch = *(mySource++);
//         if (U8_IS_SINGLE(ch))        /* Simple case */
//         {
//             *(myTarget++) = (UChar) ch;
//         }
//         else
//         {
//             /* store the first char */
//             toUBytes[0] = (char)ch;
//             inBytes = U8_COUNT_BYTES_NON_ASCII(ch); /* lookup current sequence length */
//             i = 1;

// morebytes:
//             while (i < inBytes)
//             {
//                 if (mySource < sourceLimit)
//                 {
//                     toUBytes[i] = (char) (ch2 = *mySource);
//                     if (!icu::UTF8::isValidTrail(ch, static_cast<uint8_t>(ch2), i, inBytes) &&
//                             !(isCESU8 && i == 1 && ch == 0xed && U8_IS_TRAIL(ch2)))
//                     {
//                         break; /* i < inBytes */
//                     }
//                     ch = (ch << 6) + ch2;
//                     ++mySource;
//                     i++;
//                 }
//                 else
//                 {
//                     /* stores a partially calculated target*/
//                     cnv->toUnicodeStatus = ch;
//                     cnv->mode = inBytes;
//                     cnv->toULength = (int8_t) i;
//                     goto donefornow;
//                 }
//             }

//             // In CESU-8, only surrogates, not supplementary code points, are encoded directly.
//             if (i == inBytes && (!isCESU8 || i <= 3))
//             {
//                 /* Remove the accumulated high bits */
//                 ch -= offsetsFromUTF8[inBytes];

//                 /* Normal valid byte when the loop has not prematurely terminated (i < inBytes) */
//                 if (ch <= MAXIMUM_UCS2)
//                 {
//                     /* fits in 16 bits */
//                     *(myTarget++) = (UChar) ch;
//                 }
//                 else
//                 {
//                     /* write out the surrogates */
//                     *(myTarget++) = U16_LEAD(ch);
//                     ch = U16_TRAIL(ch);
//                     if (myTarget < targetLimit)
//                     {
//                         *(myTarget++) = (UChar)ch;
//                     }
//                     else
//                     {
//                         /* Put in overflow buffer (not handled here) */
//                         cnv->UCharErrorBuffer[0] = (UChar) ch;
//                         cnv->UCharErrorBufferLength = 1;
//                         *err = U_BUFFER_OVERFLOW_ERROR;
//                         break;
//                     }
//                 }
//             }
//             else
//             {
//                 cnv->toULength = (int8_t)i;
//                 *err = U_ILLEGAL_CHAR_FOUND;
//                 break;
//             }
//         }
//     }

// donefornow:
//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err))
//     {
//         /* End of target buffer */
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = myTarget;
//     args->source = (const char *) mySource;
// }

// static void  U_CALLCONV ucnv_toUnicode_UTF8_OFFSETS_LOGIC (UConverterToUnicodeArgs * args,
//                                                 UErrorCode * err)
// {
//     UConverter *cnv = args->converter;
//     const unsigned char *mySource = (unsigned char *) args->source;
//     UChar *myTarget = args->target;
//     int32_t *myOffsets = args->offsets;
//     int32_t offsetNum = 0;
//     const unsigned char *sourceLimit = (unsigned char *) args->sourceLimit;
//     const UChar *targetLimit = args->targetLimit;
//     unsigned char *toUBytes = cnv->toUBytes;
//     UBool isCESU8 = hasCESU8Data(cnv);
//     uint32_t ch, ch2 = 0;
//     int32_t i, inBytes;

//     /* Restore size of current sequence */
//     if (cnv->toULength > 0 && myTarget < targetLimit)
//     {
//         inBytes = cnv->mode;            /* restore # of bytes to consume */
//         i = cnv->toULength;             /* restore # of bytes consumed */
//         cnv->toULength = 0;

//         ch = cnv->toUnicodeStatus;/*Stores the previously calculated ch from a previous call*/
//         cnv->toUnicodeStatus = 0;
//         goto morebytes;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit)
//     {
//         ch = *(mySource++);
//         if (U8_IS_SINGLE(ch))        /* Simple case */
//         {
//             *(myTarget++) = (UChar) ch;
//             *(myOffsets++) = offsetNum++;
//         }
//         else
//         {
//             toUBytes[0] = (char)ch;
//             inBytes = U8_COUNT_BYTES_NON_ASCII(ch);
//             i = 1;

// morebytes:
//             while (i < inBytes)
//             {
//                 if (mySource < sourceLimit)
//                 {
//                     toUBytes[i] = (char) (ch2 = *mySource);
//                     if (!icu::UTF8::isValidTrail(ch, static_cast<uint8_t>(ch2), i, inBytes) &&
//                             !(isCESU8 && i == 1 && ch == 0xed && U8_IS_TRAIL(ch2)))
//                     {
//                         break; /* i < inBytes */
//                     }
//                     ch = (ch << 6) + ch2;
//                     ++mySource;
//                     i++;
//                 }
//                 else
//                 {
//                     cnv->toUnicodeStatus = ch;
//                     cnv->mode = inBytes;
//                     cnv->toULength = (int8_t)i;
//                     goto donefornow;
//                 }
//             }

//             // In CESU-8, only surrogates, not supplementary code points, are encoded directly.
//             if (i == inBytes && (!isCESU8 || i <= 3))
//             {
//                 /* Remove the accumulated high bits */
//                 ch -= offsetsFromUTF8[inBytes];

//                 /* Normal valid byte when the loop has not prematurely terminated (i < inBytes) */
//                 if (ch <= MAXIMUM_UCS2)
//                 {
//                     /* fits in 16 bits */
//                     *(myTarget++) = (UChar) ch;
//                     *(myOffsets++) = offsetNum;
//                 }
//                 else
//                 {
//                     /* write out the surrogates */
//                     *(myTarget++) = U16_LEAD(ch);
//                     *(myOffsets++) = offsetNum;
//                     ch = U16_TRAIL(ch);
//                     if (myTarget < targetLimit)
//                     {
//                         *(myTarget++) = (UChar)ch;
//                         *(myOffsets++) = offsetNum;
//                     }
//                     else
//                     {
//                         cnv->UCharErrorBuffer[0] = (UChar) ch;
//                         cnv->UCharErrorBufferLength = 1;
//                         *err = U_BUFFER_OVERFLOW_ERROR;
//                     }
//                 }
//                 offsetNum += i;
//             }
//             else
//             {
//                 cnv->toULength = (int8_t)i;
//                 *err = U_ILLEGAL_CHAR_FOUND;
//                 break;
//             }
//         }
//     }

// donefornow:
//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err))
//     {   /* End of target buffer */
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = myTarget;
//     args->source = (const char *) mySource;
//     args->offsets = myOffsets;
// }
// U_CDECL_END

// U_CFUNC void  U_CALLCONV ucnv_fromUnicode_UTF8 (UConverterFromUnicodeArgs * args,
//                                     UErrorCode * err)
// {
//     UConverter *cnv = args->converter;
//     const UChar *mySource = args->source;
//     const UChar *sourceLimit = args->sourceLimit;
//     uint8_t *myTarget = (uint8_t *) args->target;
//     const uint8_t *targetLimit = (uint8_t *) args->targetLimit;
//     uint8_t *tempPtr;
//     UChar32 ch;
//     uint8_t tempBuf[4];
//     int32_t indexToWrite;
//     UBool isNotCESU8 = !hasCESU8Data(cnv);

//     if (cnv->fromUChar32 && myTarget < targetLimit)
//     {
//         ch = cnv->fromUChar32;
//         cnv->fromUChar32 = 0;
//         goto lowsurrogate;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit)
//     {
//         ch = *(mySource++);

//         if (ch < 0x80)        /* Single byte */
//         {
//             *(myTarget++) = (uint8_t) ch;
//         }
//         else if (ch < 0x800)  /* Double byte */
//         {
//             *(myTarget++) = (uint8_t) ((ch >> 6) | 0xc0);
//             if (myTarget < targetLimit)
//             {
//                 *(myTarget++) = (uint8_t) ((ch & 0x3f) | 0x80);
//             }
//             else
//             {
//                 cnv->charErrorBuffer[0] = (uint8_t) ((ch & 0x3f) | 0x80);
//                 cnv->charErrorBufferLength = 1;
//                 *err = U_BUFFER_OVERFLOW_ERROR;
//             }
//         }
//         else {
//             /* Check for surrogates */
//             if(U16_IS_SURROGATE(ch) && isNotCESU8) {
// lowsurrogate:
//                 if (mySource < sourceLimit) {
//                     /* test both code units */
//                     if(U16_IS_SURROGATE_LEAD(ch) && U16_IS_TRAIL(*mySource)) {
//                         /* convert and consume this supplementary code point */
//                         ch=U16_GET_SUPPLEMENTARY(ch, *mySource);
//                         ++mySource;
//                         /* exit this condition tree */
//                     }
//                     else {
//                         /* this is an unpaired trail or lead code unit */
//                         /* callback(illegal) */
//                         cnv->fromUChar32 = ch;
//                         *err = U_ILLEGAL_CHAR_FOUND;
//                         break;
//                     }
//                 }
//                 else {
//                     /* no more input */
//                     cnv->fromUChar32 = ch;
//                     break;
//                 }
//             }

//             /* Do we write the buffer directly for speed,
//             or do we have to be careful about target buffer space? */
//             tempPtr = (((targetLimit - myTarget) >= 4) ? myTarget : tempBuf);

//             if (ch <= MAXIMUM_UCS2) {
//                 indexToWrite = 2;
//                 tempPtr[0] = (uint8_t) ((ch >> 12) | 0xe0);
//             }
//             else {
//                 indexToWrite = 3;
//                 tempPtr[0] = (uint8_t) ((ch >> 18) | 0xf0);
//                 tempPtr[1] = (uint8_t) (((ch >> 12) & 0x3f) | 0x80);
//             }
//             tempPtr[indexToWrite-1] = (uint8_t) (((ch >> 6) & 0x3f) | 0x80);
//             tempPtr[indexToWrite] = (uint8_t) ((ch & 0x3f) | 0x80);

//             if (tempPtr == myTarget) {
//                 /* There was enough space to write the codepoint directly. */
//                 myTarget += (indexToWrite + 1);
//             }
//             else {
//                 /* We might run out of room soon. Write it slowly. */
//                 for (; tempPtr <= (tempBuf + indexToWrite); tempPtr++) {
//                     if (myTarget < targetLimit) {
//                         *(myTarget++) = *tempPtr;
//                     }
//                     else {
//                         cnv->charErrorBuffer[cnv->charErrorBufferLength++] = *tempPtr;
//                         *err = U_BUFFER_OVERFLOW_ERROR;
//                     }
//                 }
//             }
//         }
//     }

//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err))
//     {
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = (char *) myTarget;
//     args->source = mySource;
// }

// U_CFUNC void  U_CALLCONV ucnv_fromUnicode_UTF8_OFFSETS_LOGIC (UConverterFromUnicodeArgs * args,
//                                                   UErrorCode * err)
// {
//     UConverter *cnv = args->converter;
//     const UChar *mySource = args->source;
//     int32_t *myOffsets = args->offsets;
//     const UChar *sourceLimit = args->sourceLimit;
//     uint8_t *myTarget = (uint8_t *) args->target;
//     const uint8_t *targetLimit = (uint8_t *) args->targetLimit;
//     uint8_t *tempPtr;
//     UChar32 ch;
//     int32_t offsetNum, nextSourceIndex;
//     int32_t indexToWrite;
//     uint8_t tempBuf[4];
//     UBool isNotCESU8 = !hasCESU8Data(cnv);

//     if (cnv->fromUChar32 && myTarget < targetLimit)
//     {
//         ch = cnv->fromUChar32;
//         cnv->fromUChar32 = 0;
//         offsetNum = -1;
//         nextSourceIndex = 0;
//         goto lowsurrogate;
//     } else {
//         offsetNum = 0;
//     }

//     while (mySource < sourceLimit && myTarget < targetLimit)
//     {
//         ch = *(mySource++);

//         if (ch < 0x80)        /* Single byte */
//         {
//             *(myOffsets++) = offsetNum++;
//             *(myTarget++) = (char) ch;
//         }
//         else if (ch < 0x800)  /* Double byte */
//         {
//             *(myOffsets++) = offsetNum;
//             *(myTarget++) = (uint8_t) ((ch >> 6) | 0xc0);
//             if (myTarget < targetLimit)
//             {
//                 *(myOffsets++) = offsetNum++;
//                 *(myTarget++) = (uint8_t) ((ch & 0x3f) | 0x80);
//             }
//             else
//             {
//                 cnv->charErrorBuffer[0] = (uint8_t) ((ch & 0x3f) | 0x80);
//                 cnv->charErrorBufferLength = 1;
//                 *err = U_BUFFER_OVERFLOW_ERROR;
//             }
//         }
//         else
//         /* Check for surrogates */
//         {
//             nextSourceIndex = offsetNum + 1;

//             if(U16_IS_SURROGATE(ch) && isNotCESU8) {
// lowsurrogate:
//                 if (mySource < sourceLimit) {
//                     /* test both code units */
//                     if(U16_IS_SURROGATE_LEAD(ch) && U16_IS_TRAIL(*mySource)) {
//                         /* convert and consume this supplementary code point */
//                         ch=U16_GET_SUPPLEMENTARY(ch, *mySource);
//                         ++mySource;
//                         ++nextSourceIndex;
//                         /* exit this condition tree */
//                     }
//                     else {
//                         /* this is an unpaired trail or lead code unit */
//                         /* callback(illegal) */
//                         cnv->fromUChar32 = ch;
//                         *err = U_ILLEGAL_CHAR_FOUND;
//                         break;
//                     }
//                 }
//                 else {
//                     /* no more input */
//                     cnv->fromUChar32 = ch;
//                     break;
//                 }
//             }

//             /* Do we write the buffer directly for speed,
//             or do we have to be careful about target buffer space? */
//             tempPtr = (((targetLimit - myTarget) >= 4) ? myTarget : tempBuf);

//             if (ch <= MAXIMUM_UCS2) {
//                 indexToWrite = 2;
//                 tempPtr[0] = (uint8_t) ((ch >> 12) | 0xe0);
//             }
//             else {
//                 indexToWrite = 3;
//                 tempPtr[0] = (uint8_t) ((ch >> 18) | 0xf0);
//                 tempPtr[1] = (uint8_t) (((ch >> 12) & 0x3f) | 0x80);
//             }
//             tempPtr[indexToWrite-1] = (uint8_t) (((ch >> 6) & 0x3f) | 0x80);
//             tempPtr[indexToWrite] = (uint8_t) ((ch & 0x3f) | 0x80);

//             if (tempPtr == myTarget) {
//                 /* There was enough space to write the codepoint directly. */
//                 myTarget += (indexToWrite + 1);
//                 myOffsets[0] = offsetNum;
//                 myOffsets[1] = offsetNum;
//                 myOffsets[2] = offsetNum;
//                 if (indexToWrite >= 3) {
//                     myOffsets[3] = offsetNum;
//                 }
//                 myOffsets += (indexToWrite + 1);
//             }
//             else {
//                 /* We might run out of room soon. Write it slowly. */
//                 for (; tempPtr <= (tempBuf + indexToWrite); tempPtr++) {
//                     if (myTarget < targetLimit)
//                     {
//                         *(myOffsets++) = offsetNum;
//                         *(myTarget++) = *tempPtr;
//                     }
//                     else
//                     {
//                         cnv->charErrorBuffer[cnv->charErrorBufferLength++] = *tempPtr;
//                         *err = U_BUFFER_OVERFLOW_ERROR;
//                     }
//                 }
//             }
//             offsetNum = nextSourceIndex;
//         }
//     }

//     if (mySource < sourceLimit && myTarget >= targetLimit && U_SUCCESS(*err))
//     {
//         *err = U_BUFFER_OVERFLOW_ERROR;
//     }

//     args->target = (char *) myTarget;
//     args->source = mySource;
//     args->offsets = myOffsets;
// }

// U_CDECL_BEGIN
// static UChar32 U_CALLCONV ucnv_getNextUChar_UTF8(UConverterToUnicodeArgs *args,
//                                                UErrorCode *err) {
//     UConverter *cnv;
//     const uint8_t *sourceInitial;
//     const uint8_t *source;
//     uint8_t myByte;
//     UChar32 ch;
//     int8_t i;

//     /* UTF-8 only here, the framework handles CESU-8 to combine surrogate pairs */

//     cnv = args->converter;
//     sourceInitial = source = (const uint8_t *)args->source;
//     if (source >= (const uint8_t *)args->sourceLimit)
//     {
//         /* no input */
//         *err = U_INDEX_OUTOFBOUNDS_ERROR;
//         return 0xffff;
//     }

//     myByte = (uint8_t)*(source++);
//     if (U8_IS_SINGLE(myByte))
//     {
//         args->source = (const char *)source;
//         return (UChar32)myByte;
//     }

//     uint16_t countTrailBytes = U8_COUNT_TRAIL_BYTES(myByte);
//     if (countTrailBytes == 0) {
//         cnv->toUBytes[0] = myByte;
//         cnv->toULength = 1;
//         *err = U_ILLEGAL_CHAR_FOUND;
//         args->source = (const char *)source;
//         return 0xffff;
//     }

//     /*The byte sequence is longer than the buffer area passed*/
//     if (((const char *)source + countTrailBytes) > args->sourceLimit)
//     {
//         /* check if all of the remaining bytes are trail bytes */
//         uint16_t extraBytesToWrite = countTrailBytes + 1;
//         cnv->toUBytes[0] = myByte;
//         i = 1;
//         *err = U_TRUNCATED_CHAR_FOUND;
//         while(source < (const uint8_t *)args->sourceLimit) {
//             uint8_t b = *source;
//             if(icu::UTF8::isValidTrail(myByte, b, i, extraBytesToWrite)) {
//                 cnv->toUBytes[i++] = b;
//                 ++source;
//             } else {
//                 /* error even before we run out of input */
//                 *err = U_ILLEGAL_CHAR_FOUND;
//                 break;
//             }
//         }
//         cnv->toULength = i;
//         args->source = (const char *)source;
//         return 0xffff;
//     }

//     ch = myByte << 6;
//     if(countTrailBytes == 2) {
//         uint8_t t1 = *source, t2;
//         if(U8_IS_VALID_LEAD3_AND_T1(myByte, t1) && U8_IS_TRAIL(t2 = *++source)) {
//             args->source = (const char *)(source + 1);
//             return (((ch + t1) << 6) + t2) - offsetsFromUTF8[3];
//         }
//     } else if(countTrailBytes == 1) {
//         uint8_t t1 = *source;
//         if(U8_IS_TRAIL(t1)) {
//             args->source = (const char *)(source + 1);
//             return (ch + t1) - offsetsFromUTF8[2];
//         }
//     } else {  // countTrailBytes == 3
//         uint8_t t1 = *source, t2, t3;
//         if(U8_IS_VALID_LEAD4_AND_T1(myByte, t1) && U8_IS_TRAIL(t2 = *++source) &&
//                 U8_IS_TRAIL(t3 = *++source)) {
//             args->source = (const char *)(source + 1);
//             return (((((ch + t1) << 6) + t2) << 6) + t3) - offsetsFromUTF8[4];
//         }
//     }
//     args->source = (const char *)source;

//     for(i = 0; sourceInitial < source; ++i) {
//         cnv->toUBytes[i] = *sourceInitial++;
//     }
//     cnv->toULength = i;
//     *err = U_ILLEGAL_CHAR_FOUND;
//     return 0xffff;
// }
// U_CDECL_END

// /* UTF-8-from-UTF-8 conversion functions ------------------------------------ */

// U_CDECL_BEGIN
// /* "Convert" UTF-8 to UTF-8: Validate and copy. Modified from ucnv_DBCSFromUTF8(). */
// static void U_CALLCONV
// ucnv_UTF8FromUTF8(UConverterFromUnicodeArgs *pFromUArgs,
//                   UConverterToUnicodeArgs *pToUArgs,
//                   UErrorCode *pErrorCode) {
//     UConverter *utf8;
//     const uint8_t *source, *sourceLimit;
//     uint8_t *target;
//     int32_t targetCapacity;
//     int32_t count;

//     int8_t oldToULength, toULength, toULimit;

//     UChar32 c;
//     uint8_t b, t1, t2;

//     /* set up the local pointers */
//     utf8=pToUArgs->converter;
//     source=(uint8_t *)pToUArgs->source;
//     sourceLimit=(uint8_t *)pToUArgs->sourceLimit;
//     target=(uint8_t *)pFromUArgs->target;
//     targetCapacity=(int32_t)(pFromUArgs->targetLimit-pFromUArgs->target);

//     /* get the converter state from the UTF-8 UConverter */
//     if(utf8->toULength > 0) {
//         toULength=oldToULength=utf8->toULength;
//         toULimit=(int8_t)utf8->mode;
//         c=(UChar32)utf8->toUnicodeStatus;
//     } else {
//         toULength=oldToULength=toULimit=0;
//         c = 0;
//     }

//     count=(int32_t)(sourceLimit-source)+oldToULength;
//     if(count<toULimit) {
//         /*
//          * Not enough input to complete the partial character.
//          * Jump to moreBytes below - it will not output to target.
//          */
//     } else if(targetCapacity<toULimit) {
//         /*
//          * Not enough target capacity to output the partial character.
//          * Let the standard converter handle this.
//          */
//         *pErrorCode=U_USING_DEFAULT_WARNING;
//         return;
//     } else {
//         // Use a single counter for source and target, counting the minimum of
//         // the source length and the target capacity.
//         // Let the standard converter handle edge cases.
//         if(count>targetCapacity) {
//             count=targetCapacity;
//         }

//         // The conversion loop checks count>0 only once per character.
//         // If the buffer ends with a truncated sequence,
//         // then we reduce the count to stop before that,
//         // and collect the remaining bytes after the conversion loop.

//         // Do not go back into the bytes that will be read for finishing a partial
//         // sequence from the previous buffer.
//         int32_t length=count-toULimit;
//         U8_TRUNCATE_IF_INCOMPLETE(source, 0, length);
//         count=toULimit+length;
//     }

//     if(c!=0) {
//         utf8->toUnicodeStatus=0;
//         utf8->toULength=0;
//         goto moreBytes;
//         /* See note in ucnv_SBCSFromUTF8() about this goto. */
//     }

//     /* conversion loop */
//     while(count>0) {
//         b=*source++;
//         if(U8_IS_SINGLE(b)) {
//             /* convert ASCII */
//             *target++=b;
//             --count;
//             continue;
//         } else {
//             if(b>=0xe0) {
//                 if( /* handle U+0800..U+FFFF inline */
//                     b<0xf0 &&
//                     U8_IS_VALID_LEAD3_AND_T1(b, t1=source[0]) &&
//                     U8_IS_TRAIL(t2=source[1])
//                 ) {
//                     source+=2;
//                     *target++=b;
//                     *target++=t1;
//                     *target++=t2;
//                     count-=3;
//                     continue;
//                 }
//             } else {
//                 if( /* handle U+0080..U+07FF inline */
//                     b>=0xc2 &&
//                     U8_IS_TRAIL(t1=*source)
//                 ) {
//                     ++source;
//                     *target++=b;
//                     *target++=t1;
//                     count-=2;
//                     continue;
//                 }
//             }

//             /* handle "complicated" and error cases, and continuing partial characters */
//             oldToULength=0;
//             toULength=1;
//             toULimit=U8_COUNT_BYTES_NON_ASCII(b);
//             c=b;
// moreBytes:
//             while(toULength<toULimit) {
//                 if(source<sourceLimit) {
//                     b=*source;
//                     if(icu::UTF8::isValidTrail(c, b, toULength, toULimit)) {
//                         ++source;
//                         ++toULength;
//                         c=(c<<6)+b;
//                     } else {
//                         break; /* sequence too short, stop with toULength<toULimit */
//                     }
//                 } else {
//                     /* store the partial UTF-8 character, compatible with the regular UTF-8 converter */
//                     source-=(toULength-oldToULength);
//                     while(oldToULength<toULength) {
//                         utf8->toUBytes[oldToULength++]=*source++;
//                     }
//                     utf8->toUnicodeStatus=c;
//                     utf8->toULength=toULength;
//                     utf8->mode=toULimit;
//                     pToUArgs->source=(char *)source;
//                     pFromUArgs->target=(char *)target;
//                     return;
//                 }
//             }

//             if(toULength!=toULimit) {
//                 /* error handling: illegal UTF-8 byte sequence */
//                 source-=(toULength-oldToULength);
//                 while(oldToULength<toULength) {
//                     utf8->toUBytes[oldToULength++]=*source++;
//                 }
//                 utf8->toULength=toULength;
//                 pToUArgs->source=(char *)source;
//                 pFromUArgs->target=(char *)target;
//                 *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 return;
//             }

//             /* copy the legal byte sequence to the target */
//             {
//                 int8_t i;

//                 for(i=0; i<oldToULength; ++i) {
//                     *target++=utf8->toUBytes[i];
//                 }
//                 source-=(toULength-oldToULength);
//                 for(; i<toULength; ++i) {
//                     *target++=*source++;
//                 }
//                 count-=toULength;
//             }
//         }
//     }
//     U_ASSERT(count>=0);

//     if(U_SUCCESS(*pErrorCode) && source<sourceLimit) {
//         if(target==(const uint8_t *)pFromUArgs->targetLimit) {
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//         } else {
//             b=*source;
//             toULimit=U8_COUNT_BYTES(b);
//             if(toULimit>(sourceLimit-source)) {
//                 /* collect a truncated byte sequence */
//                 toULength=0;
//                 c=b;
//                 for(;;) {
//                     utf8->toUBytes[toULength++]=b;
//                     if(++source==sourceLimit) {
//                         /* partial byte sequence at end of source */
//                         utf8->toUnicodeStatus=c;
//                         utf8->toULength=toULength;
//                         utf8->mode=toULimit;
//                         break;
//                     } else if(!icu::UTF8::isValidTrail(c, b=*source, toULength, toULimit)) {
//                         utf8->toULength=toULength;
//                         *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                         break;
//                     }
//                     c=(c<<6)+b;
//                 }
//             } else {
//                 /* partial-sequence target overflow: fall back to the pivoting implementation */
//                 *pErrorCode=U_USING_DEFAULT_WARNING;
//             }
//         }
//     }

//     /* write back the updated pointers */
//     pToUArgs->source=(char *)source;
//     pFromUArgs->target=(char *)target;
// }

// U_CDECL_END

// /* UTF-8 converter data ----------------------------------------------------- */

// static const UConverterImpl _UTF8Impl={
//     UCNV_UTF8,

//     NULL,
//     NULL,

//     NULL,
//     NULL,
//     NULL,

//     ucnv_toUnicode_UTF8,
//     ucnv_toUnicode_UTF8_OFFSETS_LOGIC,
//     ucnv_fromUnicode_UTF8,
//     ucnv_fromUnicode_UTF8_OFFSETS_LOGIC,
//     ucnv_getNextUChar_UTF8,

//     NULL,
//     NULL,
//     NULL,
//     NULL,
//     ucnv_getNonSurrogateUnicodeSet,

//     ucnv_UTF8FromUTF8,
//     ucnv_UTF8FromUTF8
// };

// /* The 1208 CCSID refers to any version of Unicode of UTF-8 */
// static const UConverterStaticData _UTF8StaticData={
//     sizeof(UConverterStaticData),
//     "UTF-8",
//     1208, UCNV_IBM, UCNV_UTF8,
//     1, 3, /* max 3 bytes per UChar from UTF-8 (4 bytes from surrogate _pair_) */
//     { 0xef, 0xbf, 0xbd, 0 },3,FALSE,FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };


// const UConverterSharedData _UTF8Data=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_UTF8StaticData, &_UTF8Impl);

// /* CESU-8 converter data ---------------------------------------------------- */

// static const UConverterImpl _CESU8Impl={
//     UCNV_CESU8,

//     NULL,
//     NULL,

//     NULL,
//     NULL,
//     NULL,

//     ucnv_toUnicode_UTF8,
//     ucnv_toUnicode_UTF8_OFFSETS_LOGIC,
//     ucnv_fromUnicode_UTF8,
//     ucnv_fromUnicode_UTF8_OFFSETS_LOGIC,
//     NULL,

//     NULL,
//     NULL,
//     NULL,
//     NULL,
//     ucnv_getCompleteUnicodeSet,

//     NULL,
//     NULL
// };

// static const UConverterStaticData _CESU8StaticData={
//     sizeof(UConverterStaticData),
//     "CESU-8",
//     9400, /* CCSID for CESU-8 */
//     UCNV_UNKNOWN, UCNV_CESU8, 1, 3,
//     { 0xef, 0xbf, 0xbd, 0 },3,FALSE,FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };


// const UConverterSharedData _CESU8Data=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_CESU8StaticData, &_CESU8Impl);

// #endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// ******************************************************************************
// *
// *   Copyright (C) 1998-2004, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *
// ******************************************************************************
// *
// *  ucnvdisp.c:
// *  Implements APIs for the ICU's codeset conversion library display names.
// *
// * Modification History:
// *
// *   Date        Name        Description
// *   04/04/99    helena      Fixed internal header inclusion.
// *   05/09/00    helena      Added implementation to handle fallback mappings.
// *   06/20/2000  helena      OS/400 port changes; mostly typecast.
// *   09/08/2004  grhoten     split from ucnv.c
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "unicode/ustring.h"
// #include "unicode/ures.h"
// #include "unicode/ucnv.h"
// #include "cstring.h"
// #include "ustr_imp.h"
// #include "ucnv_imp.h"
// #include "putilimp.h"

// U_CAPI int32_t U_EXPORT2
// ucnv_getDisplayName(const UConverter *cnv,
//                     const char *displayLocale,
//                     UChar *displayName, int32_t displayNameCapacity,
//                     UErrorCode *pErrorCode) {
//     UResourceBundle *rb;
//     const UChar *name;
//     int32_t length;
//     UErrorCode localStatus = U_ZERO_ERROR;

//     /* check arguments */
//     if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     if(cnv==NULL || displayNameCapacity<0 || (displayNameCapacity>0 && displayName==NULL)) {
//         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     /* open the resource bundle and get the display name string */
//     rb=ures_open(NULL, displayLocale, pErrorCode);
//     if(U_FAILURE(*pErrorCode)) {
//         return 0;
//     }

//     /* use the internal name as the key */
//     name=ures_getStringByKey(rb, cnv->sharedData->staticData->name, &length, &localStatus);
//     ures_close(rb);

//     if(U_SUCCESS(localStatus)) {
//         /* copy the string */
//         if (*pErrorCode == U_ZERO_ERROR) {
//             *pErrorCode = localStatus;
//         }
//         u_memcpy(displayName, name, uprv_min(length, displayNameCapacity)*U_SIZEOF_UCHAR);
//     } else {
//         /* convert the internal name into a Unicode string */
//         length=(int32_t)uprv_strlen(cnv->sharedData->staticData->name);
//         u_charsToUChars(cnv->sharedData->staticData->name, displayName, uprv_min(length, displayNameCapacity));
//     }
//     return u_terminateUChars(displayName, displayNameCapacity, length, pErrorCode);
// }

// #endif

// /*
//  * Hey, Emacs, please set the following:
//  *
//  * Local Variables:
//  * indent-tabs-mode: nil
//  * End:
//  *
//  */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 2000-2015, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// **********************************************************************
// *   file name:  ucnvlat1.cpp
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2000feb07
// *   created by: Markus W. Scherer
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "unicode/ucnv.h"
// #include "unicode/uset.h"
// #include "unicode/utf8.h"
// #include "ucnv_bld.h"
// #include "ucnv_cnv.h"
// #include "ustr_imp.h"

// /* control optimizations according to the platform */
// #define LATIN1_UNROLL_FROM_UNICODE 1

// /* ISO 8859-1 --------------------------------------------------------------- */

// /* This is a table-less and callback-less version of ucnv_MBCSSingleToBMPWithOffsets(). */
// U_CDECL_BEGIN
// static void U_CALLCONV
// _Latin1ToUnicodeWithOffsets(UConverterToUnicodeArgs *pArgs,
//                             UErrorCode *pErrorCode) {
//     const uint8_t *source;
//     UChar *target;
//     int32_t targetCapacity, length;
//     int32_t *offsets;

//     int32_t sourceIndex;

//     /* set up the local pointers */
//     source=(const uint8_t *)pArgs->source;
//     target=pArgs->target;
//     targetCapacity=(int32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;

//     sourceIndex=0;

//     /*
//      * since the conversion here is 1:1 UChar:uint8_t, we need only one counter
//      * for the minimum of the sourceLength and targetCapacity
//      */
//     length=(int32_t)((const uint8_t *)pArgs->sourceLimit-source);
//     if(length<=targetCapacity) {
//         targetCapacity=length;
//     } else {
//         /* target will be full */
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//         length=targetCapacity;
//     }

//     if(targetCapacity>=8) {
//         /* This loop is unrolled for speed and improved pipelining. */
//         int32_t count, loops;

//         loops=count=targetCapacity>>3;
//         length=targetCapacity&=0x7;
//         do {
//             target[0]=source[0];
//             target[1]=source[1];
//             target[2]=source[2];
//             target[3]=source[3];
//             target[4]=source[4];
//             target[5]=source[5];
//             target[6]=source[6];
//             target[7]=source[7];
//             target+=8;
//             source+=8;
//         } while(--count>0);

//         if(offsets!=NULL) {
//             do {
//                 offsets[0]=sourceIndex++;
//                 offsets[1]=sourceIndex++;
//                 offsets[2]=sourceIndex++;
//                 offsets[3]=sourceIndex++;
//                 offsets[4]=sourceIndex++;
//                 offsets[5]=sourceIndex++;
//                 offsets[6]=sourceIndex++;
//                 offsets[7]=sourceIndex++;
//                 offsets+=8;
//             } while(--loops>0);
//         }
//     }

//     /* conversion loop */
//     while(targetCapacity>0) {
//         *target++=*source++;
//         --targetCapacity;
//     }

//     /* write back the updated pointers */
//     pArgs->source=(const char *)source;
//     pArgs->target=target;

//     /* set offsets */
//     if(offsets!=NULL) {
//         while(length>0) {
//             *offsets++=sourceIndex++;
//             --length;
//         }
//         pArgs->offsets=offsets;
//     }
// }

// /* This is a table-less and callback-less version of ucnv_MBCSSingleGetNextUChar(). */
// static UChar32 U_CALLCONV
// _Latin1GetNextUChar(UConverterToUnicodeArgs *pArgs,
//                     UErrorCode *pErrorCode) {
//     const uint8_t *source=(const uint8_t *)pArgs->source;
//     if(source<(const uint8_t *)pArgs->sourceLimit) {
//         pArgs->source=(const char *)(source+1);
//         return *source;
//     }

//     /* no output because of empty input */
//     *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//     return 0xffff;
// }

// /* This is a table-less version of ucnv_MBCSSingleFromBMPWithOffsets(). */
// static void U_CALLCONV
// _Latin1FromUnicodeWithOffsets(UConverterFromUnicodeArgs *pArgs,
//                               UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const UChar *source, *sourceLimit;
//     uint8_t *target, *oldTarget;
//     int32_t targetCapacity, length;
//     int32_t *offsets;

//     UChar32 cp;
//     UChar c, max;

//     int32_t sourceIndex;

//     /* set up the local pointers */
//     cnv=pArgs->converter;
//     source=pArgs->source;
//     sourceLimit=pArgs->sourceLimit;
//     target=oldTarget=(uint8_t *)pArgs->target;
//     targetCapacity=(int32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;

//     if(cnv->sharedData==&_Latin1Data) {
//         max=0xff; /* Latin-1 */
//     } else {
//         max=0x7f; /* US-ASCII */
//     }

//     /* get the converter state from UConverter */
//     cp=cnv->fromUChar32;

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     sourceIndex= cp==0 ? 0 : -1;

//     /*
//      * since the conversion here is 1:1 UChar:uint8_t, we need only one counter
//      * for the minimum of the sourceLength and targetCapacity
//      */
//     length=(int32_t)(sourceLimit-source);
//     if(length<targetCapacity) {
//         targetCapacity=length;
//     }

//     /* conversion loop */
//     if(cp!=0 && targetCapacity>0) {
//         goto getTrail;
//     }

// #if LATIN1_UNROLL_FROM_UNICODE
//     /* unroll the loop with the most common case */
//     if(targetCapacity>=16) {
//         int32_t count, loops;
//         UChar u, oredChars;

//         loops=count=targetCapacity>>4;
//         do {
//             oredChars=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;
//             oredChars|=u=*source++;
//             *target++=(uint8_t)u;

//             /* were all 16 entries really valid? */
//             if(oredChars>max) {
//                 /* no, return to the first of these 16 */
//                 source-=16;
//                 target-=16;
//                 break;
//             }
//         } while(--count>0);
//         count=loops-count;
//         targetCapacity-=16*count;

//         if(offsets!=NULL) {
//             oldTarget+=16*count;
//             while(count>0) {
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 --count;
//             }
//         }
//     }
// #endif

//     /* conversion loop */
//     c=0;
//     while(targetCapacity>0 && (c=*source++)<=max) {
//         /* convert the Unicode code point */
//         *target++=(uint8_t)c;
//         --targetCapacity;
//     }

//     if(c>max) {
//         cp=c;
//         if(!U_IS_SURROGATE(cp)) {
//             /* callback(unassigned) */
//         } else if(U_IS_SURROGATE_LEAD(cp)) {
// getTrail:
//             if(source<sourceLimit) {
//                 /* test the following code unit */
//                 UChar trail=*source;
//                 if(U16_IS_TRAIL(trail)) {
//                     ++source;
//                     cp=U16_GET_SUPPLEMENTARY(cp, trail);
//                     /* this codepage does not map supplementary code points */
//                     /* callback(unassigned) */
//                 } else {
//                     /* this is an unmatched lead code unit (1st surrogate) */
//                     /* callback(illegal) */
//                 }
//             } else {
//                 /* no more input */
//                 cnv->fromUChar32=cp;
//                 goto noMoreInput;
//             }
//         } else {
//             /* this is an unmatched trail code unit (2nd surrogate) */
//             /* callback(illegal) */
//         }

//         *pErrorCode= U_IS_SURROGATE(cp) ? U_ILLEGAL_CHAR_FOUND : U_INVALID_CHAR_FOUND;
//         cnv->fromUChar32=cp;
//     }
// noMoreInput:

//     /* set offsets since the start */
//     if(offsets!=NULL) {
//         size_t count=target-oldTarget;
//         while(count>0) {
//             *offsets++=sourceIndex++;
//             --count;
//         }
//     }

//     if(U_SUCCESS(*pErrorCode) && source<sourceLimit && target>=(uint8_t *)pArgs->targetLimit) {
//         /* target is full */
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//     }

//     /* write back the updated pointers */
//     pArgs->source=source;
//     pArgs->target=(char *)target;
//     pArgs->offsets=offsets;
// }

// /* Convert UTF-8 to Latin-1. Adapted from ucnv_SBCSFromUTF8(). */
// static void U_CALLCONV
// ucnv_Latin1FromUTF8(UConverterFromUnicodeArgs *pFromUArgs,
//                     UConverterToUnicodeArgs *pToUArgs,
//                     UErrorCode *pErrorCode) {
//     UConverter *utf8;
//     const uint8_t *source, *sourceLimit;
//     uint8_t *target;
//     int32_t targetCapacity;

//     UChar32 c;
//     uint8_t b, t1;

//     /* set up the local pointers */
//     utf8=pToUArgs->converter;
//     source=(uint8_t *)pToUArgs->source;
//     sourceLimit=(uint8_t *)pToUArgs->sourceLimit;
//     target=(uint8_t *)pFromUArgs->target;
//     targetCapacity=(int32_t)(pFromUArgs->targetLimit-pFromUArgs->target);

//     /* get the converter state from the UTF-8 UConverter */
//     if (utf8->toULength > 0) {
//         c=(UChar32)utf8->toUnicodeStatus;
//     } else {
//         c = 0;
//     }
//     if(c!=0 && source<sourceLimit) {
//         if(targetCapacity==0) {
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             return;
//         } else if(c>=0xc2 && c<=0xc3 && (t1=(uint8_t)(*source-0x80)) <= 0x3f) {
//             ++source;
//             *target++=(uint8_t)(((c&3)<<6)|t1);
//             --targetCapacity;

//             utf8->toUnicodeStatus=0;
//             utf8->toULength=0;
//         } else {
//             /* complicated, illegal or unmappable input: fall back to the pivoting implementation */
//             *pErrorCode=U_USING_DEFAULT_WARNING;
//             return;
//         }
//     }

//     /*
//      * Make sure that the last byte sequence before sourceLimit is complete
//      * or runs into a lead byte.
//      * In the conversion loop compare source with sourceLimit only once
//      * per multi-byte character.
//      * For Latin-1, adjust sourceLimit only for 1 trail byte because
//      * the conversion loop handles at most 2-byte sequences.
//      */
//     if(source<sourceLimit && U8_IS_LEAD(*(sourceLimit-1))) {
//         --sourceLimit;
//     }

//     /* conversion loop */
//     while(source<sourceLimit) {
//         if(targetCapacity>0) {
//             b=*source++;
//             if(U8_IS_SINGLE(b)) {
//                 /* convert ASCII */
//                 *target++=(uint8_t)b;
//                 --targetCapacity;
//             } else if( /* handle U+0080..U+00FF inline */
//                        b>=0xc2 && b<=0xc3 &&
//                        (t1=(uint8_t)(*source-0x80)) <= 0x3f
//             ) {
//                 ++source;
//                 *target++=(uint8_t)(((b&3)<<6)|t1);
//                 --targetCapacity;
//             } else {
//                 /* complicated, illegal or unmappable input: fall back to the pivoting implementation */
//                 pToUArgs->source=(char *)(source-1);
//                 pFromUArgs->target=(char *)target;
//                 *pErrorCode=U_USING_DEFAULT_WARNING;
//                 return;
//             }
//         } else {
//             /* target is full */
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }
//     }

//     /*
//      * The sourceLimit may have been adjusted before the conversion loop
//      * to stop before a truncated sequence.
//      * If so, then collect the truncated sequence now.
//      * For Latin-1, there is at most exactly one lead byte because of the
//      * smaller sourceLimit adjustment logic.
//      */
//     if(U_SUCCESS(*pErrorCode) && source<(sourceLimit=(uint8_t *)pToUArgs->sourceLimit)) {
//         utf8->toUnicodeStatus=utf8->toUBytes[0]=b=*source++;
//         utf8->toULength=1;
//         utf8->mode=U8_COUNT_BYTES(b);
//     }

//     /* write back the updated pointers */
//     pToUArgs->source=(char *)source;
//     pFromUArgs->target=(char *)target;
// }

// static void U_CALLCONV
// _Latin1GetUnicodeSet(const UConverter *cnv,
//                      const USetAdder *sa,
//                      UConverterUnicodeSet which,
//                      UErrorCode *pErrorCode) {
//     (void)cnv;
//     (void)which;
//     (void)pErrorCode;
//     sa->addRange(sa->set, 0, 0xff);
// }
// U_CDECL_END


// static const UConverterImpl _Latin1Impl={
//     UCNV_LATIN_1,

//     NULL,
//     NULL,

//     NULL,
//     NULL,
//     NULL,

//     _Latin1ToUnicodeWithOffsets,
//     _Latin1ToUnicodeWithOffsets,
//     _Latin1FromUnicodeWithOffsets,
//     _Latin1FromUnicodeWithOffsets,
//     _Latin1GetNextUChar,

//     NULL,
//     NULL,
//     NULL,
//     NULL,
//     _Latin1GetUnicodeSet,

//     NULL,
//     ucnv_Latin1FromUTF8
// };

// static const UConverterStaticData _Latin1StaticData={
//     sizeof(UConverterStaticData),
//     "ISO-8859-1",
//     819, UCNV_IBM, UCNV_LATIN_1, 1, 1,
//     { 0x1a, 0, 0, 0 }, 1, FALSE, FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };

// const UConverterSharedData _Latin1Data=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_Latin1StaticData, &_Latin1Impl);

// /* US-ASCII ----------------------------------------------------------------- */

// U_CDECL_BEGIN
// /* This is a table-less version of ucnv_MBCSSingleToBMPWithOffsets(). */
// static void U_CALLCONV
// _ASCIIToUnicodeWithOffsets(UConverterToUnicodeArgs *pArgs,
//                            UErrorCode *pErrorCode) {
//     const uint8_t *source, *sourceLimit;
//     UChar *target, *oldTarget;
//     int32_t targetCapacity, length;
//     int32_t *offsets;

//     int32_t sourceIndex;

//     uint8_t c;

//     /* set up the local pointers */
//     source=(const uint8_t *)pArgs->source;
//     sourceLimit=(const uint8_t *)pArgs->sourceLimit;
//     target=oldTarget=pArgs->target;
//     targetCapacity=(int32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     sourceIndex=0;

//     /*
//      * since the conversion here is 1:1 UChar:uint8_t, we need only one counter
//      * for the minimum of the sourceLength and targetCapacity
//      */
//     length=(int32_t)(sourceLimit-source);
//     if(length<targetCapacity) {
//         targetCapacity=length;
//     }

//     if(targetCapacity>=8) {
//         /* This loop is unrolled for speed and improved pipelining. */
//         int32_t count, loops;
//         UChar oredChars;

//         loops=count=targetCapacity>>3;
//         do {
//             oredChars=target[0]=source[0];
//             oredChars|=target[1]=source[1];
//             oredChars|=target[2]=source[2];
//             oredChars|=target[3]=source[3];
//             oredChars|=target[4]=source[4];
//             oredChars|=target[5]=source[5];
//             oredChars|=target[6]=source[6];
//             oredChars|=target[7]=source[7];

//             /* were all 16 entries really valid? */
//             if(oredChars>0x7f) {
//                 /* no, return to the first of these 16 */
//                 break;
//             }
//             source+=8;
//             target+=8;
//         } while(--count>0);
//         count=loops-count;
//         targetCapacity-=count*8;

//         if(offsets!=NULL) {
//             oldTarget+=count*8;
//             while(count>0) {
//                 offsets[0]=sourceIndex++;
//                 offsets[1]=sourceIndex++;
//                 offsets[2]=sourceIndex++;
//                 offsets[3]=sourceIndex++;
//                 offsets[4]=sourceIndex++;
//                 offsets[5]=sourceIndex++;
//                 offsets[6]=sourceIndex++;
//                 offsets[7]=sourceIndex++;
//                 offsets+=8;
//                 --count;
//             }
//         }
//     }

//     /* conversion loop */
//     c=0;
//     while(targetCapacity>0 && (c=*source++)<=0x7f) {
//         *target++=c;
//         --targetCapacity;
//     }

//     if(c>0x7f) {
//         /* callback(illegal); copy the current bytes to toUBytes[] */
//         UConverter *cnv=pArgs->converter;
//         cnv->toUBytes[0]=c;
//         cnv->toULength=1;
//         *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//     } else if(source<sourceLimit && target>=pArgs->targetLimit) {
//         /* target is full */
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//     }

//     /* set offsets since the start */
//     if(offsets!=NULL) {
//         size_t count=target-oldTarget;
//         while(count>0) {
//             *offsets++=sourceIndex++;
//             --count;
//         }
//     }

//     /* write back the updated pointers */
//     pArgs->source=(const char *)source;
//     pArgs->target=target;
//     pArgs->offsets=offsets;
// }

// /* This is a table-less version of ucnv_MBCSSingleGetNextUChar(). */
// static UChar32 U_CALLCONV
// _ASCIIGetNextUChar(UConverterToUnicodeArgs *pArgs,
//                    UErrorCode *pErrorCode) {
//     const uint8_t *source;
//     uint8_t b;

//     source=(const uint8_t *)pArgs->source;
//     if(source<(const uint8_t *)pArgs->sourceLimit) {
//         b=*source++;
//         pArgs->source=(const char *)source;
//         if(b<=0x7f) {
//             return b;
//         } else {
//             UConverter *cnv=pArgs->converter;
//             cnv->toUBytes[0]=b;
//             cnv->toULength=1;
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//             return 0xffff;
//         }
//     }

//     /* no output because of empty input */
//     *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//     return 0xffff;
// }

// /* "Convert" UTF-8 to US-ASCII: Validate and copy. */
// static void U_CALLCONV
// ucnv_ASCIIFromUTF8(UConverterFromUnicodeArgs *pFromUArgs,
//                    UConverterToUnicodeArgs *pToUArgs,
//                    UErrorCode *pErrorCode) {
//     const uint8_t *source, *sourceLimit;
//     uint8_t *target;
//     int32_t targetCapacity, length;

//     uint8_t c;

//     if(pToUArgs->converter->toULength > 0) {
//         /* no handling of partial UTF-8 characters here, fall back to pivoting */
//         *pErrorCode=U_USING_DEFAULT_WARNING;
//         return;
//     }

//     /* set up the local pointers */
//     source=(const uint8_t *)pToUArgs->source;
//     sourceLimit=(const uint8_t *)pToUArgs->sourceLimit;
//     target=(uint8_t *)pFromUArgs->target;
//     targetCapacity=(int32_t)(pFromUArgs->targetLimit-pFromUArgs->target);

//     /*
//      * since the conversion here is 1:1 uint8_t:uint8_t, we need only one counter
//      * for the minimum of the sourceLength and targetCapacity
//      */
//     length=(int32_t)(sourceLimit-source);
//     if(length<targetCapacity) {
//         targetCapacity=length;
//     }

//     /* unroll the loop with the most common case */
//     if(targetCapacity>=16) {
//         int32_t count, loops;
//         uint8_t oredChars;

//         loops=count=targetCapacity>>4;
//         do {
//             oredChars=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;
//             oredChars|=*target++=*source++;

//             /* were all 16 entries really valid? */
//             if(oredChars>0x7f) {
//                 /* no, return to the first of these 16 */
//                 source-=16;
//                 target-=16;
//                 break;
//             }
//         } while(--count>0);
//         count=loops-count;
//         targetCapacity-=16*count;
//     }

//     /* conversion loop */
//     c=0;
//     while(targetCapacity>0 && (c=*source)<=0x7f) {
//         ++source;
//         *target++=c;
//         --targetCapacity;
//     }

//     if(c>0x7f) {
//         /* non-ASCII character, handle in standard converter */
//         *pErrorCode=U_USING_DEFAULT_WARNING;
//     } else if(source<sourceLimit && target>=(const uint8_t *)pFromUArgs->targetLimit) {
//         /* target is full */
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//     }

//     /* write back the updated pointers */
//     pToUArgs->source=(const char *)source;
//     pFromUArgs->target=(char *)target;
// }

// static void U_CALLCONV
// _ASCIIGetUnicodeSet(const UConverter *cnv,
//                     const USetAdder *sa,
//                     UConverterUnicodeSet which,
//                     UErrorCode *pErrorCode) {
//     (void)cnv;
//     (void)which;
//     (void)pErrorCode;
//     sa->addRange(sa->set, 0, 0x7f);
// }
// U_CDECL_END

// static const UConverterImpl _ASCIIImpl={
//     UCNV_US_ASCII,

//     NULL,
//     NULL,

//     NULL,
//     NULL,
//     NULL,

//     _ASCIIToUnicodeWithOffsets,
//     _ASCIIToUnicodeWithOffsets,
//     _Latin1FromUnicodeWithOffsets,
//     _Latin1FromUnicodeWithOffsets,
//     _ASCIIGetNextUChar,

//     NULL,
//     NULL,
//     NULL,
//     NULL,
//     _ASCIIGetUnicodeSet,

//     NULL,
//     ucnv_ASCIIFromUTF8
// };

// static const UConverterStaticData _ASCIIStaticData={
//     sizeof(UConverterStaticData),
//     "US-ASCII",
//     367, UCNV_IBM, UCNV_US_ASCII, 1, 1,
//     { 0x1a, 0, 0, 0 }, 1, FALSE, FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };

// const UConverterSharedData _ASCIIData=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_ASCIIStaticData, &_ASCIIImpl);

// #endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// ******************************************************************************
// *
// *   Copyright (C) 2000-2016, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *
// ******************************************************************************
// *   file name:  ucnvmbcs.cpp
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2000jul03
// *   created by: Markus W. Scherer
// *
// *   The current code in this file replaces the previous implementation
// *   of conversion code from multi-byte codepages to Unicode and back.
// *   This implementation supports the following:
// *   - legacy variable-length codepages with up to 4 bytes per character
// *   - all Unicode code points (up to 0x10ffff)
// *   - efficient distinction of unassigned vs. illegal byte sequences
// *   - it is possible in fromUnicode() to directly deal with simple
// *     stateful encodings (used for EBCDIC_STATEFUL)
// *   - it is possible to convert Unicode code points
// *     to a single zero byte (but not as a fallback except for SBCS)
// *
// *   Remaining limitations in fromUnicode:
// *   - byte sequences must not have leading zero bytes
// *   - except for SBCS codepages: no fallback mapping from Unicode to a zero byte
// *   - limitation to up to 4 bytes per character
// *
// *   ICU 2.8 (late 2003) adds a secondary data structure which lifts some of these
// *   limitations and adds m:n character mappings and other features.
// *   See ucnv_ext.h for details.
// *
// *   Change history:
// *
// *    5/6/2001       Ram       Moved  MBCS_SINGLE_RESULT_FROM_U,MBCS_STAGE_2_FROM_U,
// *                             MBCS_VALUE_2_FROM_STAGE_2, MBCS_VALUE_4_FROM_STAGE_2
// *                             macros to ucnvmbcs.h file
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION && !UCONFIG_NO_LEGACY_CONVERSION

// #include "unicode/ucnv.h"
// #include "unicode/ucnv_cb.h"
// #include "unicode/udata.h"
// #include "unicode/uset.h"
// #include "unicode/utf8.h"
// #include "unicode/utf16.h"
// #include "ucnv_bld.h"
// #include "ucnvmbcs.h"
// #include "ucnv_ext.h"
// #include "ucnv_cnv.h"
// #include "cmemory.h"
// #include "cstring.h"
// #include "umutex.h"
// #include "ustr_imp.h"

// /* control optimizations according to the platform */
// #define MBCS_UNROLL_SINGLE_TO_BMP 1
// #define MBCS_UNROLL_SINGLE_FROM_BMP 0

// /*
//  * _MBCSHeader versions 5.3 & 4.3
//  * (Note that the _MBCSHeader version is in addition to the converter formatVersion.)
//  *
//  * This version is optional. Version 5 is used for incompatible data format changes.
//  * makeconv will continue to generate version 4 files if possible.
//  *
//  * Changes from version 4:
//  *
//  * The main difference is an additional _MBCSHeader field with
//  * - the length (number of uint32_t) of the _MBCSHeader
//  * - flags for further incompatible data format changes
//  * - flags for further, backward compatible data format changes
//  *
//  * The MBCS_OPT_FROM_U flag indicates that most of the fromUnicode data is omitted from
//  * the file and needs to be reconstituted at load time.
//  * This requires a utf8Friendly format with an additional mbcsIndex table for fast
//  * (and UTF-8-friendly) fromUnicode conversion for Unicode code points up to maxFastUChar.
//  * (For details about these structures see below, and see ucnvmbcs.h.)
//  *
//  *   utf8Friendly also implies that the fromUnicode mappings are stored in ascending order
//  *   of the Unicode code points. (This requires that the .ucm file has the |0 etc.
//  *   precision markers for all mappings.)
//  *
//  *   All fallbacks have been moved to the extension table, leaving only roundtrips in the
//  *   omitted data that can be reconstituted from the toUnicode data.
//  *
//  *   Of the stage 2 table, the part corresponding to maxFastUChar and below is omitted.
//  *   With only roundtrip mappings in the base fromUnicode data, this part is fully
//  *   redundant with the mbcsIndex and will be reconstituted from that (also using the
//  *   stage 1 table which contains the information about how stage 2 was compacted).
//  *
//  *   The rest of the stage 2 table, the part for code points above maxFastUChar,
//  *   is stored in the file and will be appended to the reconstituted part.
//  *
//  *   The entire fromUBytes array is omitted from the file and will be reconstitued.
//  *   This is done by enumerating all toUnicode roundtrip mappings, performing
//  *   each mapping (using the stage 1 and reconstituted stage 2 tables) and
//  *   writing instead of reading the byte values.
//  *
//  * _MBCSHeader version 4.3
//  *
//  * Change from version 4.2:
//  * - Optional utf8Friendly data structures, with 64-entry stage 3 block
//  *   allocation for parts of the BMP, and an additional mbcsIndex in non-SBCS
//  *   files which can be used instead of stages 1 & 2.
//  *   Faster lookups for roundtrips from most commonly used characters,
//  *   and lookups from UTF-8 byte sequences with a natural bit distribution.
//  *   See ucnvmbcs.h for more details.
//  *
//  * Change from version 4.1:
//  * - Added an optional extension table structure at the end of the .cnv file.
//  *   It is present if the upper bits of the header flags field contains a non-zero
//  *   byte offset to it.
//  *   Files that contain only a conversion table and no base table
//  *   use the special outputType MBCS_OUTPUT_EXT_ONLY.
//  *   These contain the base table name between the MBCS header and the extension
//  *   data.
//  *
//  * Change from version 4.0:
//  * - Replace header.reserved with header.fromUBytesLength so that all
//  *   fields in the data have length.
//  *
//  * Changes from version 3 (for performance improvements):
//  * - new bit distribution for state table entries
//  * - reordered action codes
//  * - new data structure for single-byte fromUnicode
//  *   + stage 2 only contains indexes
//  *   + stage 3 stores 16 bits per character with classification bits 15..8
//  * - no multiplier for stage 1 entries
//  * - stage 2 for non-single-byte codepages contains the index and the flags in
//  *   one 32-bit value
//  * - 2-byte and 4-byte fromUnicode results are stored directly as 16/32-bit integers
//  *
//  * For more details about old versions of the MBCS data structure, see
//  * the corresponding versions of this file.
//  *
//  * Converting stateless codepage data ---------------------------------------***
//  * (or codepage data with simple states) to Unicode.
//  *
//  * Data structure and algorithm for converting from complex legacy codepages
//  * to Unicode. (Designed before 2000-may-22.)
//  *
//  * The basic idea is that the structure of legacy codepages can be described
//  * with state tables.
//  * When reading a byte stream, each input byte causes a state transition.
//  * Some transitions result in the output of a code point, some result in
//  * "unassigned" or "illegal" output.
//  * This is used here for character conversion.
//  *
//  * The data structure begins with a state table consisting of a row
//  * per state, with 256 entries (columns) per row for each possible input
//  * byte value.
//  * Each entry is 32 bits wide, with two formats distinguished by
//  * the sign bit (bit 31):
//  *
//  * One format for transitional entries (bit 31 not set) for non-final bytes, and
//  * one format for final entries (bit 31 set).
//  * Both formats contain the number of the next state in the same bit
//  * positions.
//  * State 0 is the initial state.
//  *
//  * Most of the time, the offset values of subsequent states are added
//  * up to a scalar value. This value will eventually be the index of
//  * the Unicode code point in a table that follows the state table.
//  * The effect is that the code points for final state table rows
//  * are contiguous. The code points of final state rows follow each other
//  * in the order of the references to those final states by previous
//  * states, etc.
//  *
//  * For some terminal states, the offset is itself the output Unicode
//  * code point (16 bits for a BMP code point or 20 bits for a supplementary
//  * code point (stored as code point minus 0x10000 so that 20 bits are enough).
//  * For others, the code point in the Unicode table is stored with either
//  * one or two code units: one for BMP code points, two for a pair of
//  * surrogates.
//  * All code points for a final state entry take up the same number of code
//  * units, regardless of whether they all actually _use_ the same number
//  * of code units. This is necessary for simple array access.
//  *
//  * An additional feature comes in with what in ICU is called "fallback"
//  * mappings:
//  *
//  * In addition to round-trippable, precise, 1:1 mappings, there are often
//  * mappings defined between similar, though not the same, characters.
//  * Typically, such mappings occur only in fromUnicode mapping tables because
//  * Unicode has a superset repertoire of most other codepages. However, it
//  * is possible to provide such mappings in the toUnicode tables, too.
//  * In this case, the fallback mappings are partly integrated into the
//  * general state tables because the structure of the encoding includes their
//  * byte sequences.
//  * For final entries in an initial state, fallback mappings are stored in
//  * the entry itself like with roundtrip mappings.
//  * For other final entries, they are stored in the code units table if
//  * the entry is for a pair of code units.
//  * For single-unit results in the code units table, there is no space to
//  * alternatively hold a fallback mapping; in this case, the code unit
//  * is stored as U+fffe (unassigned), and the fallback mapping needs to
//  * be looked up by the scalar offset value in a separate table.
//  *
//  * "Unassigned" state entries really mean "structurally unassigned",
//  * i.e., such a byte sequence will never have a mapping result.
//  *
//  * The interpretation of the bits in each entry is as follows:
//  *
//  * Bit 31 not set, not a terminal entry ("transitional"):
//  * 30..24 next state
//  * 23..0  offset delta, to be added up
//  *
//  * Bit 31 set, terminal ("final") entry:
//  * 30..24 next state (regardless of action code)
//  * 23..20 action code:
//  *        action codes 0 and 1 result in precise-mapping Unicode code points
//  *        0  valid byte sequence
//  *           19..16 not used, 0
//  *           15..0  16-bit Unicode BMP code point
//  *                  never U+fffe or U+ffff
//  *        1  valid byte sequence
//  *           19..0  20-bit Unicode supplementary code point
//  *                  never U+fffe or U+ffff
//  *
//  *        action codes 2 and 3 result in fallback (unidirectional-mapping) Unicode code points
//  *        2  valid byte sequence (fallback)
//  *           19..16 not used, 0
//  *           15..0  16-bit Unicode BMP code point as fallback result
//  *        3  valid byte sequence (fallback)
//  *           19..0  20-bit Unicode supplementary code point as fallback result
//  *
//  *        action codes 4 and 5 may result in roundtrip/fallback/unassigned/illegal results
//  *        depending on the code units they result in
//  *        4  valid byte sequence
//  *           19..9  not used, 0
//  *            8..0  final offset delta
//  *                  pointing to one 16-bit code unit which may be
//  *                  fffe  unassigned -- look for a fallback for this offset
//  *                  ffff  illegal
//  *        5  valid byte sequence
//  *           19..9  not used, 0
//  *            8..0  final offset delta
//  *                  pointing to two 16-bit code units
//  *                  (typically UTF-16 surrogates)
//  *                  the result depends on the first code unit as follows:
//  *                  0000..d7ff  roundtrip BMP code point (1st alone)
//  *                  d800..dbff  roundtrip surrogate pair (1st, 2nd)
//  *                  dc00..dfff  fallback surrogate pair (1st-400, 2nd)
//  *                  e000        roundtrip BMP code point (2nd alone)
//  *                  e001        fallback BMP code point (2nd alone)
//  *                  fffe        unassigned
//  *                  ffff        illegal
//  *           (the final offset deltas are at most 255 * 2,
//  *            times 2 because of storing code unit pairs)
//  *
//  *        6  unassigned byte sequence
//  *           19..16 not used, 0
//  *           15..0  16-bit Unicode BMP code point U+fffe (new with version 2)
//  *                  this does not contain a final offset delta because the main
//  *                  purpose of this action code is to save scalar offset values;
//  *                  therefore, fallback values cannot be assigned to byte
//  *                  sequences that result in this action code
//  *        7  illegal byte sequence
//  *           19..16 not used, 0
//  *           15..0  16-bit Unicode BMP code point U+ffff (new with version 2)
//  *        8  state change only
//  *           19..0  not used, 0
//  *           useful for state changes in simple stateful encodings,
//  *           at Shift-In/Shift-Out codes
//  *
//  *
//  *        9..15 reserved for future use
//  *           current implementations will only perform a state change
//  *           and ignore bits 19..0
//  *
//  * An encoding with contiguous ranges of unassigned byte sequences, like
//  * Shift-JIS and especially EUC-TW, can be stored efficiently by having
//  * at least two states for the trail bytes:
//  * One trail byte state that results in code points, and one that only
//  * has "unassigned" and "illegal" terminal states.
//  *
//  * Note: partly by accident, this data structure supports simple stateful
//  * encodings without any additional logic.
//  * Currently, only simple Shift-In/Shift-Out schemes are handled with
//  * appropriate state tables (especially EBCDIC_STATEFUL!).
//  *
//  * MBCS version 2 added:
//  * unassigned and illegal action codes have U+fffe and U+ffff
//  * instead of unused bits; this is useful for _MBCS_SINGLE_SIMPLE_GET_NEXT_BMP()
//  *
//  * Converting from Unicode to codepage bytes --------------------------------***
//  *
//  * The conversion data structure for fromUnicode is designed for the known
//  * structure of Unicode. It maps from 21-bit code points (0..0x10ffff) to
//  * a sequence of 1..4 bytes, in addition to a flag that indicates if there is
//  * a roundtrip mapping.
//  *
//  * The lookup is done with a 3-stage trie, using 11/6/4 bits for stage 1/2/3
//  * like in the character properties table.
//  * The beginning of the trie is at offsetFromUTable, the beginning of stage 3
//  * with the resulting bytes is at offsetFromUBytes.
//  *
//  * Beginning with version 4, single-byte codepages have a significantly different
//  * trie compared to other codepages.
//  * In all cases, the entry in stage 1 is directly the index of the block of
//  * 64 entries in stage 2.
//  *
//  * Single-byte lookup:
//  *
//  * Stage 2 only contains 16-bit indexes directly to the 16-blocks in stage 3.
//  * Stage 3 contains one 16-bit word per result:
//  * Bits 15..8 indicate the kind of result:
//  *    f  roundtrip result
//  *    c  fallback result from private-use code point
//  *    8  fallback result from other code points
//  *    0  unassigned
//  * Bits 7..0 contain the codepage byte. A zero byte is always possible.
//  *
//  * In version 4.3, the runtime code can build an sbcsIndex for a utf8Friendly
//  * file. For 2-byte UTF-8 byte sequences and some 3-byte sequences the lookup
//  * becomes a 2-stage (single-index) trie lookup with 6 bits for stage 3.
//  * ASCII code points can be looked up with a linear array access into stage 3.
//  * See maxFastUChar and other details in ucnvmbcs.h.
//  *
//  * Multi-byte lookup:
//  *
//  * Stage 2 contains a 32-bit word for each 16-block in stage 3:
//  * Bits 31..16 contain flags for which stage 3 entries contain roundtrip results
//  *             test: MBCS_FROM_U_IS_ROUNDTRIP(stage2Entry, c)
//  *             If this test is false, then a non-zero result will be interpreted as
//  *             a fallback mapping.
//  * Bits 15..0  contain the index to stage 3, which must be multiplied by 16*(bytes per char)
//  *
//  * Stage 3 contains 2, 3, or 4 bytes per result.
//  * 2 or 4 bytes are stored as uint16_t/uint32_t in platform endianness,
//  * while 3 bytes are stored as bytes in big-endian order.
//  * Leading zero bytes are ignored, and the number of bytes is counted.
//  * A zero byte mapping result is possible as a roundtrip result.
//  * For some output types, the actual result is processed from this;
//  * see ucnv_MBCSFromUnicodeWithOffsets().
//  *
//  * Note that stage 1 always contains 0x440=1088 entries (0x440==0x110000>>10),
//  * or (version 3 and up) for BMP-only codepages, it contains 64 entries.
//  *
//  * In version 4.3, a utf8Friendly file contains an mbcsIndex table.
//  * For 2-byte UTF-8 byte sequences and most 3-byte sequences the lookup
//  * becomes a 2-stage (single-index) trie lookup with 6 bits for stage 3.
//  * ASCII code points can be looked up with a linear array access into stage 3.
//  * See maxFastUChar, mbcsIndex and other details in ucnvmbcs.h.
//  *
//  * In version 3, stage 2 blocks may overlap by multiples of the multiplier
//  * for compaction.
//  * In version 4, stage 2 blocks (and for single-byte codepages, stage 3 blocks)
//  * may overlap by any number of entries.
//  *
//  * MBCS version 2 added:
//  * the converter checks for known output types, which allows
//  * adding new ones without crashing an unaware converter
//  */

// /**
//  * Callback from ucnv_MBCSEnumToUnicode(), takes 32 mappings from
//  * consecutive sequences of bytes, starting from the one encoded in value,
//  * to Unicode code points. (Multiple mappings to reduce per-function call overhead.)
//  * Does not currently support m:n mappings or reverse fallbacks.
//  * This function will not be called for sequences of bytes with leading zeros.
//  *
//  * @param context an opaque pointer, as passed into ucnv_MBCSEnumToUnicode()
//  * @param value contains 1..4 bytes of the first byte sequence, right-aligned
//  * @param codePoints resulting Unicode code points, or negative if a byte sequence does
//  *        not map to anything
//  * @return TRUE to continue enumeration, FALSE to stop
//  */
// typedef UBool U_CALLCONV
// UConverterEnumToUCallback(const void *context, uint32_t value, UChar32 codePoints[32]);

// static void U_CALLCONV
// ucnv_MBCSLoad(UConverterSharedData *sharedData,
//           UConverterLoadArgs *pArgs,
//           const uint8_t *raw,
//           UErrorCode *pErrorCode);

// static void U_CALLCONV
// ucnv_MBCSUnload(UConverterSharedData *sharedData);

// static void U_CALLCONV
// ucnv_MBCSOpen(UConverter *cnv,
//               UConverterLoadArgs *pArgs,
//               UErrorCode *pErrorCode);

// static UChar32 U_CALLCONV
// ucnv_MBCSGetNextUChar(UConverterToUnicodeArgs *pArgs,
//                   UErrorCode *pErrorCode);

// static void U_CALLCONV
// ucnv_MBCSGetStarters(const UConverter* cnv,
//                  UBool starters[256],
//                  UErrorCode *pErrorCode);

// U_CDECL_BEGIN
// static const char* U_CALLCONV
// ucnv_MBCSGetName(const UConverter *cnv);
// U_CDECL_END

// static void U_CALLCONV
// ucnv_MBCSWriteSub(UConverterFromUnicodeArgs *pArgs,
//               int32_t offsetIndex,
//               UErrorCode *pErrorCode);

// static UChar32 U_CALLCONV
// ucnv_MBCSGetNextUChar(UConverterToUnicodeArgs *pArgs,
//                   UErrorCode *pErrorCode);

// static void U_CALLCONV
// ucnv_SBCSFromUTF8(UConverterFromUnicodeArgs *pFromUArgs,
//                   UConverterToUnicodeArgs *pToUArgs,
//                   UErrorCode *pErrorCode);

// static void U_CALLCONV
// ucnv_MBCSGetUnicodeSet(const UConverter *cnv,
//                    const USetAdder *sa,
//                    UConverterUnicodeSet which,
//                    UErrorCode *pErrorCode);

// static void U_CALLCONV
// ucnv_DBCSFromUTF8(UConverterFromUnicodeArgs *pFromUArgs,
//                   UConverterToUnicodeArgs *pToUArgs,
//                   UErrorCode *pErrorCode);

// static const UConverterImpl _SBCSUTF8Impl={
//     UCNV_MBCS,

//     ucnv_MBCSLoad,
//     ucnv_MBCSUnload,

//     ucnv_MBCSOpen,
//     NULL,
//     NULL,

//     ucnv_MBCSToUnicodeWithOffsets,
//     ucnv_MBCSToUnicodeWithOffsets,
//     ucnv_MBCSFromUnicodeWithOffsets,
//     ucnv_MBCSFromUnicodeWithOffsets,
//     ucnv_MBCSGetNextUChar,

//     ucnv_MBCSGetStarters,
//     ucnv_MBCSGetName,
//     ucnv_MBCSWriteSub,
//     NULL,
//     ucnv_MBCSGetUnicodeSet,

//     NULL,
//     ucnv_SBCSFromUTF8
// };

// static const UConverterImpl _DBCSUTF8Impl={
//     UCNV_MBCS,

//     ucnv_MBCSLoad,
//     ucnv_MBCSUnload,

//     ucnv_MBCSOpen,
//     NULL,
//     NULL,

//     ucnv_MBCSToUnicodeWithOffsets,
//     ucnv_MBCSToUnicodeWithOffsets,
//     ucnv_MBCSFromUnicodeWithOffsets,
//     ucnv_MBCSFromUnicodeWithOffsets,
//     ucnv_MBCSGetNextUChar,

//     ucnv_MBCSGetStarters,
//     ucnv_MBCSGetName,
//     ucnv_MBCSWriteSub,
//     NULL,
//     ucnv_MBCSGetUnicodeSet,

//     NULL,
//     ucnv_DBCSFromUTF8
// };

// static const UConverterImpl _MBCSImpl={
//     UCNV_MBCS,

//     ucnv_MBCSLoad,
//     ucnv_MBCSUnload,

//     ucnv_MBCSOpen,
//     NULL,
//     NULL,

//     ucnv_MBCSToUnicodeWithOffsets,
//     ucnv_MBCSToUnicodeWithOffsets,
//     ucnv_MBCSFromUnicodeWithOffsets,
//     ucnv_MBCSFromUnicodeWithOffsets,
//     ucnv_MBCSGetNextUChar,

//     ucnv_MBCSGetStarters,
//     ucnv_MBCSGetName,
//     ucnv_MBCSWriteSub,
//     NULL,
//     ucnv_MBCSGetUnicodeSet,
//     NULL,
//     NULL
// };

// /* Static data is in tools/makeconv/ucnvstat.c for data-based
//  * converters. Be sure to update it as well.
//  */

// const UConverterSharedData _MBCSData={
//     sizeof(UConverterSharedData), 1,
//     NULL, NULL, FALSE, TRUE, &_MBCSImpl,
//     0, UCNV_MBCS_TABLE_INITIALIZER
// };


// /* GB 18030 data ------------------------------------------------------------ */

// /* helper macros for linear values for GB 18030 four-byte sequences */
// #define LINEAR_18030(a, b, c, d) ((((a)*10+(b))*126L+(c))*10L+(d))

// #define LINEAR_18030_BASE LINEAR_18030(0x81, 0x30, 0x81, 0x30)

// #define LINEAR(x) LINEAR_18030(x>>24, (x>>16)&0xff, (x>>8)&0xff, x&0xff)

// /*
//  * Some ranges of GB 18030 where both the Unicode code points and the
//  * GB four-byte sequences are contiguous and are handled algorithmically by
//  * the special callback functions below.
//  * The values are start & end of Unicode & GB codes.
//  *
//  * Note that single surrogates are not mapped by GB 18030
//  * as of the re-released mapping tables from 2000-nov-30.
//  */
// static const uint32_t
// gb18030Ranges[14][4]={
//     {0x10000, 0x10FFFF, LINEAR(0x90308130), LINEAR(0xE3329A35)},
//     {0x9FA6, 0xD7FF, LINEAR(0x82358F33), LINEAR(0x8336C738)},
//     {0x0452, 0x1E3E, LINEAR(0x8130D330), LINEAR(0x8135F436)},
//     {0x1E40, 0x200F, LINEAR(0x8135F438), LINEAR(0x8136A531)},
//     {0xE865, 0xF92B, LINEAR(0x8336D030), LINEAR(0x84308534)},
//     {0x2643, 0x2E80, LINEAR(0x8137A839), LINEAR(0x8138FD38)},
//     {0xFA2A, 0xFE2F, LINEAR(0x84309C38), LINEAR(0x84318537)},
//     {0x3CE1, 0x4055, LINEAR(0x8231D438), LINEAR(0x8232AF32)},
//     {0x361B, 0x3917, LINEAR(0x8230A633), LINEAR(0x8230F237)},
//     {0x49B8, 0x4C76, LINEAR(0x8234A131), LINEAR(0x8234E733)},
//     {0x4160, 0x4336, LINEAR(0x8232C937), LINEAR(0x8232F837)},
//     {0x478E, 0x4946, LINEAR(0x8233E838), LINEAR(0x82349638)},
//     {0x44D7, 0x464B, LINEAR(0x8233A339), LINEAR(0x8233C931)},
//     {0xFFE6, 0xFFFF, LINEAR(0x8431A234), LINEAR(0x8431A439)}
// };

// /* bit flag for UConverter.options indicating GB 18030 special handling */
// #define _MBCS_OPTION_GB18030 0x8000

// /* bit flag for UConverter.options indicating KEIS,JEF,JIF special handling */
// #define _MBCS_OPTION_KEIS 0x01000
// #define _MBCS_OPTION_JEF  0x02000
// #define _MBCS_OPTION_JIPS 0x04000

// #define KEIS_SO_CHAR_1 0x0A
// #define KEIS_SO_CHAR_2 0x42
// #define KEIS_SI_CHAR_1 0x0A
// #define KEIS_SI_CHAR_2 0x41

// #define JEF_SO_CHAR 0x28
// #define JEF_SI_CHAR 0x29

// #define JIPS_SO_CHAR_1 0x1A
// #define JIPS_SO_CHAR_2 0x70
// #define JIPS_SI_CHAR_1 0x1A
// #define JIPS_SI_CHAR_2 0x71

// enum SISO_Option {
//     SI,
//     SO
// };
// typedef enum SISO_Option SISO_Option;

// static int32_t getSISOBytes(SISO_Option option, uint32_t cnvOption, uint8_t *value) {
//     int32_t SISOLength = 0;

//     switch (option) {
//         case SI:
//             if ((cnvOption&_MBCS_OPTION_KEIS)!=0) {
//                 value[0] = KEIS_SI_CHAR_1;
//                 value[1] = KEIS_SI_CHAR_2;
//                 SISOLength = 2;
//             } else if ((cnvOption&_MBCS_OPTION_JEF)!=0) {
//                 value[0] = JEF_SI_CHAR;
//                 SISOLength = 1;
//             } else if ((cnvOption&_MBCS_OPTION_JIPS)!=0) {
//                 value[0] = JIPS_SI_CHAR_1;
//                 value[1] = JIPS_SI_CHAR_2;
//                 SISOLength = 2;
//             } else {
//                 value[0] = UCNV_SI;
//                 SISOLength = 1;
//             }
//             break;
//         case SO:
//             if ((cnvOption&_MBCS_OPTION_KEIS)!=0) {
//                 value[0] = KEIS_SO_CHAR_1;
//                 value[1] = KEIS_SO_CHAR_2;
//                 SISOLength = 2;
//             } else if ((cnvOption&_MBCS_OPTION_JEF)!=0) {
//                 value[0] = JEF_SO_CHAR;
//                 SISOLength = 1;
//             } else if ((cnvOption&_MBCS_OPTION_JIPS)!=0) {
//                 value[0] = JIPS_SO_CHAR_1;
//                 value[1] = JIPS_SO_CHAR_2;
//                 SISOLength = 2;
//             } else {
//                 value[0] = UCNV_SO;
//                 SISOLength = 1;
//             }
//             break;
//         default:
//             /* Should never happen. */
//             break;
//     }

//     return SISOLength;
// }

// /* Miscellaneous ------------------------------------------------------------ */

// /* similar to ucnv_MBCSGetNextUChar() but recursive */
// static UBool
// enumToU(UConverterMBCSTable *mbcsTable, int8_t stateProps[],
//         int32_t state, uint32_t offset,
//         uint32_t value,
//         UConverterEnumToUCallback *callback, const void *context,
//         UErrorCode *pErrorCode) {
//     UChar32 codePoints[32];
//     const int32_t *row;
//     const uint16_t *unicodeCodeUnits;
//     UChar32 anyCodePoints;
//     int32_t b, limit;

//     row=mbcsTable->stateTable[state];
//     unicodeCodeUnits=mbcsTable->unicodeCodeUnits;

//     value<<=8;
//     anyCodePoints=-1;  /* becomes non-negative if there is a mapping */

//     b=(stateProps[state]&0x38)<<2;
//     if(b==0 && stateProps[state]>=0x40) {
//         /* skip byte sequences with leading zeros because they are not stored in the fromUnicode table */
//         codePoints[0]=U_SENTINEL;
//         b=1;
//     }
//     limit=((stateProps[state]&7)+1)<<5;
//     while(b<limit) {
//         int32_t entry=row[b];
//         if(MBCS_ENTRY_IS_TRANSITION(entry)) {
//             int32_t nextState=MBCS_ENTRY_TRANSITION_STATE(entry);
//             if(stateProps[nextState]>=0) {
//                 /* recurse to a state with non-ignorable actions */
//                 if(!enumToU(
//                         mbcsTable, stateProps, nextState,
//                         offset+MBCS_ENTRY_TRANSITION_OFFSET(entry),
//                         value|(uint32_t)b,
//                         callback, context,
//                         pErrorCode)) {
//                     return FALSE;
//                 }
//             }
//             codePoints[b&0x1f]=U_SENTINEL;
//         } else {
//             UChar32 c;
//             int32_t action;

//             /*
//              * An if-else-if chain provides more reliable performance for
//              * the most common cases compared to a switch.
//              */
//             action=MBCS_ENTRY_FINAL_ACTION(entry);
//             if(action==MBCS_STATE_VALID_DIRECT_16) {
//                 /* output BMP code point */
//                 c=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             } else if(action==MBCS_STATE_VALID_16) {
//                 int32_t finalOffset=offset+MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 c=unicodeCodeUnits[finalOffset];
//                 if(c<0xfffe) {
//                     /* output BMP code point */
//                 } else {
//                     c=U_SENTINEL;
//                 }
//             } else if(action==MBCS_STATE_VALID_16_PAIR) {
//                 int32_t finalOffset=offset+MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 c=unicodeCodeUnits[finalOffset++];
//                 if(c<0xd800) {
//                     /* output BMP code point below 0xd800 */
//                 } else if(c<=0xdbff) {
//                     /* output roundtrip or fallback supplementary code point */
//                     c=((c&0x3ff)<<10)+unicodeCodeUnits[finalOffset]+(0x10000-0xdc00);
//                 } else if(c==0xe000) {
//                     /* output roundtrip BMP code point above 0xd800 or fallback BMP code point */
//                     c=unicodeCodeUnits[finalOffset];
//                 } else {
//                     c=U_SENTINEL;
//                 }
//             } else if(action==MBCS_STATE_VALID_DIRECT_20) {
//                 /* output supplementary code point */
//                 c=(UChar32)(MBCS_ENTRY_FINAL_VALUE(entry)+0x10000);
//             } else {
//                 c=U_SENTINEL;
//             }

//             codePoints[b&0x1f]=c;
//             anyCodePoints&=c;
//         }
//         if(((++b)&0x1f)==0) {
//             if(anyCodePoints>=0) {
//                 if(!callback(context, value|(uint32_t)(b-0x20), codePoints)) {
//                     return FALSE;
//                 }
//                 anyCodePoints=-1;
//             }
//         }
//     }
//     return TRUE;
// }

// /*
//  * Only called if stateProps[state]==-1.
//  * A recursive call may do stateProps[state]|=0x40 if this state is the target of an
//  * MBCS_STATE_CHANGE_ONLY.
//  */
// static int8_t
// getStateProp(const int32_t (*stateTable)[256], int8_t stateProps[], int state) {
//     const int32_t *row;
//     int32_t min, max, entry, nextState;

//     row=stateTable[state];
//     stateProps[state]=0;

//     /* find first non-ignorable state */
//     for(min=0;; ++min) {
//         entry=row[min];
//         nextState=MBCS_ENTRY_STATE(entry);
//         if(stateProps[nextState]==-1) {
//             getStateProp(stateTable, stateProps, nextState);
//         }
//         if(MBCS_ENTRY_IS_TRANSITION(entry)) {
//             if(stateProps[nextState]>=0) {
//                 break;
//             }
//         } else if(MBCS_ENTRY_FINAL_ACTION(entry)<MBCS_STATE_UNASSIGNED) {
//             break;
//         }
//         if(min==0xff) {
//             stateProps[state]=-0x40;  /* (int8_t)0xc0 */
//             return stateProps[state];
//         }
//     }
//     stateProps[state]|=(int8_t)((min>>5)<<3);

//     /* find last non-ignorable state */
//     for(max=0xff; min<max; --max) {
//         entry=row[max];
//         nextState=MBCS_ENTRY_STATE(entry);
//         if(stateProps[nextState]==-1) {
//             getStateProp(stateTable, stateProps, nextState);
//         }
//         if(MBCS_ENTRY_IS_TRANSITION(entry)) {
//             if(stateProps[nextState]>=0) {
//                 break;
//             }
//         } else if(MBCS_ENTRY_FINAL_ACTION(entry)<MBCS_STATE_UNASSIGNED) {
//             break;
//         }
//     }
//     stateProps[state]|=(int8_t)(max>>5);

//     /* recurse further and collect direct-state information */
//     while(min<=max) {
//         entry=row[min];
//         nextState=MBCS_ENTRY_STATE(entry);
//         if(stateProps[nextState]==-1) {
//             getStateProp(stateTable, stateProps, nextState);
//         }
//         if(MBCS_ENTRY_IS_FINAL(entry)) {
//             stateProps[nextState]|=0x40;
//             if(MBCS_ENTRY_FINAL_ACTION(entry)<=MBCS_STATE_FALLBACK_DIRECT_20) {
//                 stateProps[state]|=0x40;
//             }
//         }
//         ++min;
//     }
//     return stateProps[state];
// }

// /*
//  * Internal function enumerating the toUnicode data of an MBCS converter.
//  * Currently only used for reconstituting data for a MBCS_OPT_NO_FROM_U
//  * table, but could also be used for a future ucnv_getUnicodeSet() option
//  * that includes reverse fallbacks (after updating this function's implementation).
//  * Currently only handles roundtrip mappings.
//  * Does not currently handle extensions.
//  */
// static void
// ucnv_MBCSEnumToUnicode(UConverterMBCSTable *mbcsTable,
//                        UConverterEnumToUCallback *callback, const void *context,
//                        UErrorCode *pErrorCode) {
//     /*
//      * Properties for each state, to speed up the enumeration.
//      * Ignorable actions are unassigned/illegal/state-change-only:
//      * They do not lead to mappings.
//      *
//      * Bits 7..6:
//      * 1 direct/initial state (stateful converters have multiple)
//      * 0 non-initial state with transitions or with non-ignorable result actions
//      * -1 final state with only ignorable actions
//      *
//      * Bits 5..3:
//      * The lowest byte value with non-ignorable actions is
//      * value<<5 (rounded down).
//      *
//      * Bits 2..0:
//      * The highest byte value with non-ignorable actions is
//      * (value<<5)&0x1f (rounded up).
//      */
//     int8_t stateProps[MBCS_MAX_STATE_COUNT];
//     int32_t state;

//     uprv_memset(stateProps, -1, sizeof(stateProps));

//     /* recurse from state 0 and set all stateProps */
//     getStateProp(mbcsTable->stateTable, stateProps, 0);

//     for(state=0; state<mbcsTable->countStates; ++state) {
//         /*if(stateProps[state]==-1) {
//             printf("unused/unreachable <icu:state> %d\n", state);
//         }*/
//         if(stateProps[state]>=0x40) {
//             /* start from each direct state */
//             enumToU(
//                 mbcsTable, stateProps, state, 0, 0,
//                 callback, context,
//                 pErrorCode);
//         }
//     }
// }

// U_CFUNC void
// ucnv_MBCSGetFilteredUnicodeSetForUnicode(const UConverterSharedData *sharedData,
//                                          const USetAdder *sa,
//                                          UConverterUnicodeSet which,
//                                          UConverterSetFilter filter,
//                                          UErrorCode *pErrorCode) {
//     const UConverterMBCSTable *mbcsTable;
//     const uint16_t *table;

//     uint32_t st3;
//     uint16_t st1, maxStage1, st2;

//     UChar32 c;

//     /* enumerate the from-Unicode trie table */
//     mbcsTable=&sharedData->mbcs;
//     table=mbcsTable->fromUnicodeTable;
//     if(mbcsTable->unicodeMask&UCNV_HAS_SUPPLEMENTARY) {
//         maxStage1=0x440;
//     } else {
//         maxStage1=0x40;
//     }

//     c=0; /* keep track of the current code point while enumerating */

//     if(mbcsTable->outputType==MBCS_OUTPUT_1) {
//         const uint16_t *stage2, *stage3, *results;
//         uint16_t minValue;

//         results=(const uint16_t *)mbcsTable->fromUnicodeBytes;

//         /*
//          * Set a threshold variable for selecting which mappings to use.
//          * See ucnv_MBCSSingleFromBMPWithOffsets() and
//          * MBCS_SINGLE_RESULT_FROM_U() for details.
//          */
//         if(which==UCNV_ROUNDTRIP_SET) {
//             /* use only roundtrips */
//             minValue=0xf00;
//         } else /* UCNV_ROUNDTRIP_AND_FALLBACK_SET */ {
//             /* use all roundtrip and fallback results */
//             minValue=0x800;
//         }

//         for(st1=0; st1<maxStage1; ++st1) {
//             st2=table[st1];
//             if(st2>maxStage1) {
//                 stage2=table+st2;
//                 for(st2=0; st2<64; ++st2) {
//                     if((st3=stage2[st2])!=0) {
//                         /* read the stage 3 block */
//                         stage3=results+st3;

//                         do {
//                             if(*stage3++>=minValue) {
//                                 sa->add(sa->set, c);
//                             }
//                         } while((++c&0xf)!=0);
//                     } else {
//                         c+=16; /* empty stage 3 block */
//                     }
//                 }
//             } else {
//                 c+=1024; /* empty stage 2 block */
//             }
//         }
//     } else {
//         const uint32_t *stage2;
//         const uint8_t *stage3, *bytes;
//         uint32_t st3Multiplier;
//         uint32_t value;
//         UBool useFallback;

//         bytes=mbcsTable->fromUnicodeBytes;

//         useFallback=(UBool)(which==UCNV_ROUNDTRIP_AND_FALLBACK_SET);

//         switch(mbcsTable->outputType) {
//         case MBCS_OUTPUT_3:
//         case MBCS_OUTPUT_4_EUC:
//             st3Multiplier=3;
//             break;
//         case MBCS_OUTPUT_4:
//             st3Multiplier=4;
//             break;
//         default:
//             st3Multiplier=2;
//             break;
//         }

//         for(st1=0; st1<maxStage1; ++st1) {
//             st2=table[st1];
//             if(st2>(maxStage1>>1)) {
//                 stage2=(const uint32_t *)table+st2;
//                 for(st2=0; st2<64; ++st2) {
//                     if((st3=stage2[st2])!=0) {
//                         /* read the stage 3 block */
//                         stage3=bytes+st3Multiplier*16*(uint32_t)(uint16_t)st3;

//                         /* get the roundtrip flags for the stage 3 block */
//                         st3>>=16;

//                         /*
//                          * Add code points for which the roundtrip flag is set,
//                          * or which map to non-zero bytes if we use fallbacks.
//                          * See ucnv_MBCSFromUnicodeWithOffsets() for details.
//                          */
//                         switch(filter) {
//                         case UCNV_SET_FILTER_NONE:
//                             do {
//                                 if(st3&1) {
//                                     sa->add(sa->set, c);
//                                     stage3+=st3Multiplier;
//                                 } else if(useFallback) {
//                                     uint8_t b=0;
//                                     switch(st3Multiplier) {
//                                     case 4:
//                                         b|=*stage3++;
//                                         U_FALLTHROUGH;
//                                     case 3:
//                                         b|=*stage3++;
//                                         U_FALLTHROUGH;
//                                     case 2:
//                                         b|=stage3[0]|stage3[1];
//                                         stage3+=2;
//                                         U_FALLTHROUGH;
//                                     default:
//                                         break;
//                                     }
//                                     if(b!=0) {
//                                         sa->add(sa->set, c);
//                                     }
//                                 }
//                                 st3>>=1;
//                             } while((++c&0xf)!=0);
//                             break;
//                         case UCNV_SET_FILTER_DBCS_ONLY:
//                              /* Ignore single-byte results (<0x100). */
//                             do {
//                                 if(((st3&1)!=0 || useFallback) && *((const uint16_t *)stage3)>=0x100) {
//                                     sa->add(sa->set, c);
//                                 }
//                                 st3>>=1;
//                                 stage3+=2;  /* +=st3Multiplier */
//                             } while((++c&0xf)!=0);
//                             break;
//                         case UCNV_SET_FILTER_2022_CN:
//                              /* Only add code points that map to CNS 11643 planes 1 & 2 for non-EXT ISO-2022-CN. */
//                             do {
//                                 if(((st3&1)!=0 || useFallback) && ((value=*stage3)==0x81 || value==0x82)) {
//                                     sa->add(sa->set, c);
//                                 }
//                                 st3>>=1;
//                                 stage3+=3;  /* +=st3Multiplier */
//                             } while((++c&0xf)!=0);
//                             break;
//                         case UCNV_SET_FILTER_SJIS:
//                              /* Only add code points that map to Shift-JIS codes corresponding to JIS X 0208. */
//                             do {
//                                 if(((st3&1)!=0 || useFallback) && (value=*((const uint16_t *)stage3))>=0x8140 && value<=0xeffc) {
//                                     sa->add(sa->set, c);
//                                 }
//                                 st3>>=1;
//                                 stage3+=2;  /* +=st3Multiplier */
//                             } while((++c&0xf)!=0);
//                             break;
//                         case UCNV_SET_FILTER_GR94DBCS:
//                             /* Only add code points that map to ISO 2022 GR 94 DBCS codes (each byte A1..FE). */
//                             do {
//                                 if( ((st3&1)!=0 || useFallback) &&
//                                     (uint16_t)((value=*((const uint16_t *)stage3)) - 0xa1a1)<=(0xfefe - 0xa1a1) &&
//                                     (uint8_t)(value-0xa1)<=(0xfe - 0xa1)
//                                 ) {
//                                     sa->add(sa->set, c);
//                                 }
//                                 st3>>=1;
//                                 stage3+=2;  /* +=st3Multiplier */
//                             } while((++c&0xf)!=0);
//                             break;
//                         case UCNV_SET_FILTER_HZ:
//                             /* Only add code points that are suitable for HZ DBCS (lead byte A1..FD). */
//                             do {
//                                 if( ((st3&1)!=0 || useFallback) &&
//                                     (uint16_t)((value=*((const uint16_t *)stage3))-0xa1a1)<=(0xfdfe - 0xa1a1) &&
//                                     (uint8_t)(value-0xa1)<=(0xfe - 0xa1)
//                                 ) {
//                                     sa->add(sa->set, c);
//                                 }
//                                 st3>>=1;
//                                 stage3+=2;  /* +=st3Multiplier */
//                             } while((++c&0xf)!=0);
//                             break;
//                         default:
//                             *pErrorCode=U_INTERNAL_PROGRAM_ERROR;
//                             return;
//                         }
//                     } else {
//                         c+=16; /* empty stage 3 block */
//                     }
//                 }
//             } else {
//                 c+=1024; /* empty stage 2 block */
//             }
//         }
//     }

//     ucnv_extGetUnicodeSet(sharedData, sa, which, filter, pErrorCode);
// }

// U_CFUNC void
// ucnv_MBCSGetUnicodeSetForUnicode(const UConverterSharedData *sharedData,
//                                  const USetAdder *sa,
//                                  UConverterUnicodeSet which,
//                                  UErrorCode *pErrorCode) {
//     ucnv_MBCSGetFilteredUnicodeSetForUnicode(
//         sharedData, sa, which,
//         sharedData->mbcs.outputType==MBCS_OUTPUT_DBCS_ONLY ?
//             UCNV_SET_FILTER_DBCS_ONLY :
//             UCNV_SET_FILTER_NONE,
//         pErrorCode);
// }

// static void U_CALLCONV
// ucnv_MBCSGetUnicodeSet(const UConverter *cnv,
//                    const USetAdder *sa,
//                    UConverterUnicodeSet which,
//                    UErrorCode *pErrorCode) {
//     if(cnv->options&_MBCS_OPTION_GB18030) {
//         sa->addRange(sa->set, 0, 0xd7ff);
//         sa->addRange(sa->set, 0xe000, 0x10ffff);
//     } else {
//         ucnv_MBCSGetUnicodeSetForUnicode(cnv->sharedData, sa, which, pErrorCode);
//     }
// }

// /* conversion extensions for input not in the main table -------------------- */

// /*
//  * Hardcoded extension handling for GB 18030.
//  * Definition of LINEAR macros and gb18030Ranges see near the beginning of the file.
//  *
//  * In the future, conversion extensions may handle m:n mappings and delta tables,
//  * see http://source.icu-project.org/repos/icu/icuhtml/trunk/design/conversion/conversion_extensions.html
//  *
//  * If an input character cannot be mapped, then these functions set an error
//  * code. The framework will then call the callback function.
//  */

// /*
//  * @return if(U_FAILURE) return the code point for cnv->fromUChar32
//  *         else return 0 after output has been written to the target
//  */
// static UChar32
// _extFromU(UConverter *cnv, const UConverterSharedData *sharedData,
//           UChar32 cp,
//           const UChar **source, const UChar *sourceLimit,
//           uint8_t **target, const uint8_t *targetLimit,
//           int32_t **offsets, int32_t sourceIndex,
//           UBool flush,
//           UErrorCode *pErrorCode) {
//     const int32_t *cx;

//     cnv->useSubChar1=FALSE;

//     if( (cx=sharedData->mbcs.extIndexes)!=NULL &&
//         ucnv_extInitialMatchFromU(
//             cnv, cx,
//             cp, source, sourceLimit,
//             (char **)target, (char *)targetLimit,
//             offsets, sourceIndex,
//             flush,
//             pErrorCode)
//     ) {
//         return 0; /* an extension mapping handled the input */
//     }

//     /* GB 18030 */
//     if((cnv->options&_MBCS_OPTION_GB18030)!=0) {
//         const uint32_t *range;
//         int32_t i;

//         range=gb18030Ranges[0];
//         for(i=0; i<UPRV_LENGTHOF(gb18030Ranges); range+=4, ++i) {
//             if(range[0]<=(uint32_t)cp && (uint32_t)cp<=range[1]) {
//                 /* found the Unicode code point, output the four-byte sequence for it */
//                 uint32_t linear;
//                 char bytes[4];

//                 /* get the linear value of the first GB 18030 code in this range */
//                 linear=range[2]-LINEAR_18030_BASE;

//                 /* add the offset from the beginning of the range */
//                 linear+=((uint32_t)cp-range[0]);

//                 /* turn this into a four-byte sequence */
//                 bytes[3]=(char)(0x30+linear%10); linear/=10;
//                 bytes[2]=(char)(0x81+linear%126); linear/=126;
//                 bytes[1]=(char)(0x30+linear%10); linear/=10;
//                 bytes[0]=(char)(0x81+linear);

//                 /* output this sequence */
//                 ucnv_fromUWriteBytes(cnv,
//                                      bytes, 4, (char **)target, (char *)targetLimit,
//                                      offsets, sourceIndex, pErrorCode);
//                 return 0;
//             }
//         }
//     }

//     /* no mapping */
//     *pErrorCode=U_INVALID_CHAR_FOUND;
//     return cp;
// }

// /*
//  * Input sequence: cnv->toUBytes[0..length[
//  * @return if(U_FAILURE) return the length (toULength, byteIndex) for the input
//  *         else return 0 after output has been written to the target
//  */
// static int8_t
// _extToU(UConverter *cnv, const UConverterSharedData *sharedData,
//         int8_t length,
//         const uint8_t **source, const uint8_t *sourceLimit,
//         UChar **target, const UChar *targetLimit,
//         int32_t **offsets, int32_t sourceIndex,
//         UBool flush,
//         UErrorCode *pErrorCode) {
//     const int32_t *cx;

//     if( (cx=sharedData->mbcs.extIndexes)!=NULL &&
//         ucnv_extInitialMatchToU(
//             cnv, cx,
//             length, (const char **)source, (const char *)sourceLimit,
//             target, targetLimit,
//             offsets, sourceIndex,
//             flush,
//             pErrorCode)
//     ) {
//         return 0; /* an extension mapping handled the input */
//     }

//     /* GB 18030 */
//     if(length==4 && (cnv->options&_MBCS_OPTION_GB18030)!=0) {
//         const uint32_t *range;
//         uint32_t linear;
//         int32_t i;

//         linear=LINEAR_18030(cnv->toUBytes[0], cnv->toUBytes[1], cnv->toUBytes[2], cnv->toUBytes[3]);
//         range=gb18030Ranges[0];
//         for(i=0; i<UPRV_LENGTHOF(gb18030Ranges); range+=4, ++i) {
//             if(range[2]<=linear && linear<=range[3]) {
//                 /* found the sequence, output the Unicode code point for it */
//                 *pErrorCode=U_ZERO_ERROR;

//                 /* add the linear difference between the input and start sequences to the start code point */
//                 linear=range[0]+(linear-range[2]);

//                 /* output this code point */
//                 ucnv_toUWriteCodePoint(cnv, linear, target, targetLimit, offsets, sourceIndex, pErrorCode);

//                 return 0;
//             }
//         }
//     }

//     /* no mapping */
//     *pErrorCode=U_INVALID_CHAR_FOUND;
//     return length;
// }

// /* EBCDIC swap LF<->NL ------------------------------------------------------ */

// /*
//  * This code modifies a standard EBCDIC<->Unicode mapping table for
//  * OS/390 (z/OS) Unix System Services (Open Edition).
//  * The difference is in the mapping of Line Feed and New Line control codes:
//  * Standard EBCDIC maps
//  *
//  *   <U000A> \x25 |0
//  *   <U0085> \x15 |0
//  *
//  * but OS/390 USS EBCDIC swaps the control codes for LF and NL,
//  * mapping
//  *
//  *   <U000A> \x15 |0
//  *   <U0085> \x25 |0
//  *
//  * This code modifies a loaded standard EBCDIC<->Unicode mapping table
//  * by copying it into allocated memory and swapping the LF and NL values.
//  * It allows to support the same EBCDIC charset in both versions without
//  * duplicating the entire installed table.
//  */

// /* standard EBCDIC codes */
// #define EBCDIC_LF 0x25
// #define EBCDIC_NL 0x15

// /* standard EBCDIC codes with roundtrip flag as stored in Unicode-to-single-byte tables */
// #define EBCDIC_RT_LF 0xf25
// #define EBCDIC_RT_NL 0xf15

// /* Unicode code points */
// #define U_LF 0x0a
// #define U_NL 0x85

// static UBool
// _EBCDICSwapLFNL(UConverterSharedData *sharedData, UErrorCode *pErrorCode) {
//     UConverterMBCSTable *mbcsTable;

//     const uint16_t *table, *results;
//     const uint8_t *bytes;

//     int32_t (*newStateTable)[256];
//     uint16_t *newResults;
//     uint8_t *p;
//     char *name;

//     uint32_t stage2Entry;
//     uint32_t size, sizeofFromUBytes;

//     mbcsTable=&sharedData->mbcs;

//     table=mbcsTable->fromUnicodeTable;
//     bytes=mbcsTable->fromUnicodeBytes;
//     results=(const uint16_t *)bytes;

//     /*
//      * Check that this is an EBCDIC table with SBCS portion -
//      * SBCS or EBCDIC_STATEFUL with standard EBCDIC LF and NL mappings.
//      *
//      * If not, ignore the option. Options are always ignored if they do not apply.
//      */
//     if(!(
//          (mbcsTable->outputType==MBCS_OUTPUT_1 || mbcsTable->outputType==MBCS_OUTPUT_2_SISO) &&
//          mbcsTable->stateTable[0][EBCDIC_LF]==MBCS_ENTRY_FINAL(0, MBCS_STATE_VALID_DIRECT_16, U_LF) &&
//          mbcsTable->stateTable[0][EBCDIC_NL]==MBCS_ENTRY_FINAL(0, MBCS_STATE_VALID_DIRECT_16, U_NL)
//     )) {
//         return FALSE;
//     }

//     if(mbcsTable->outputType==MBCS_OUTPUT_1) {
//         if(!(
//              EBCDIC_RT_LF==MBCS_SINGLE_RESULT_FROM_U(table, results, U_LF) &&
//              EBCDIC_RT_NL==MBCS_SINGLE_RESULT_FROM_U(table, results, U_NL)
//         )) {
//             return FALSE;
//         }
//     } else /* MBCS_OUTPUT_2_SISO */ {
//         stage2Entry=MBCS_STAGE_2_FROM_U(table, U_LF);
//         if(!(
//              MBCS_FROM_U_IS_ROUNDTRIP(stage2Entry, U_LF)!=0 &&
//              EBCDIC_LF==MBCS_VALUE_2_FROM_STAGE_2(bytes, stage2Entry, U_LF)
//         )) {
//             return FALSE;
//         }

//         stage2Entry=MBCS_STAGE_2_FROM_U(table, U_NL);
//         if(!(
//              MBCS_FROM_U_IS_ROUNDTRIP(stage2Entry, U_NL)!=0 &&
//              EBCDIC_NL==MBCS_VALUE_2_FROM_STAGE_2(bytes, stage2Entry, U_NL)
//         )) {
//             return FALSE;
//         }
//     }

//     if(mbcsTable->fromUBytesLength>0) {
//         /*
//          * We _know_ the number of bytes in the fromUnicodeBytes array
//          * starting with header.version 4.1.
//          */
//         sizeofFromUBytes=mbcsTable->fromUBytesLength;
//     } else {
//         /*
//          * Otherwise:
//          * There used to be code to enumerate the fromUnicode
//          * trie and find the highest entry, but it was removed in ICU 3.2
//          * because it was not tested and caused a low code coverage number.
//          * See Jitterbug 3674.
//          * This affects only some .cnv file formats with a header.version
//          * below 4.1, and only when swaplfnl is requested.
//          *
//          * ucnvmbcs.c revision 1.99 is the last one with the
//          * ucnv_MBCSSizeofFromUBytes() function.
//          */
//         *pErrorCode=U_INVALID_FORMAT_ERROR;
//         return FALSE;
//     }

//     /*
//      * The table has an appropriate format.
//      * Allocate and build
//      * - a modified to-Unicode state table
//      * - a modified from-Unicode output array
//      * - a converter name string with the swap option appended
//      */
//     size=
//         mbcsTable->countStates*1024+
//         sizeofFromUBytes+
//         UCNV_MAX_CONVERTER_NAME_LENGTH+20;
//     p=(uint8_t *)uprv_malloc(size);
//     if(p==NULL) {
//         *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
//         return FALSE;
//     }

//     /* copy and modify the to-Unicode state table */
//     newStateTable=(int32_t (*)[256])p;
//     uprv_memcpy(newStateTable, mbcsTable->stateTable, mbcsTable->countStates*1024);

//     newStateTable[0][EBCDIC_LF]=MBCS_ENTRY_FINAL(0, MBCS_STATE_VALID_DIRECT_16, U_NL);
//     newStateTable[0][EBCDIC_NL]=MBCS_ENTRY_FINAL(0, MBCS_STATE_VALID_DIRECT_16, U_LF);

//     /* copy and modify the from-Unicode result table */
//     newResults=(uint16_t *)newStateTable[mbcsTable->countStates];
//     uprv_memcpy(newResults, bytes, sizeofFromUBytes);

//     /* conveniently, the table access macros work on the left side of expressions */
//     if(mbcsTable->outputType==MBCS_OUTPUT_1) {
//         MBCS_SINGLE_RESULT_FROM_U(table, newResults, U_LF)=EBCDIC_RT_NL;
//         MBCS_SINGLE_RESULT_FROM_U(table, newResults, U_NL)=EBCDIC_RT_LF;
//     } else /* MBCS_OUTPUT_2_SISO */ {
//         stage2Entry=MBCS_STAGE_2_FROM_U(table, U_LF);
//         MBCS_VALUE_2_FROM_STAGE_2(newResults, stage2Entry, U_LF)=EBCDIC_NL;

//         stage2Entry=MBCS_STAGE_2_FROM_U(table, U_NL);
//         MBCS_VALUE_2_FROM_STAGE_2(newResults, stage2Entry, U_NL)=EBCDIC_LF;
//     }

//     /* set the canonical converter name */
//     name=(char *)newResults+sizeofFromUBytes;
//     uprv_strcpy(name, sharedData->staticData->name);
//     uprv_strcat(name, UCNV_SWAP_LFNL_OPTION_STRING);

//     /* set the pointers */
//     icu::umtx_lock(NULL);
//     if(mbcsTable->swapLFNLStateTable==NULL) {
//         mbcsTable->swapLFNLStateTable=newStateTable;
//         mbcsTable->swapLFNLFromUnicodeBytes=(uint8_t *)newResults;
//         mbcsTable->swapLFNLName=name;

//         newStateTable=NULL;
//     }
//     icu::umtx_unlock(NULL);

//     /* release the allocated memory if another thread beat us to it */
//     if(newStateTable!=NULL) {
//         uprv_free(newStateTable);
//     }
//     return TRUE;
// }

// /* reconstitute omitted fromUnicode data ------------------------------------ */

// /* for details, compare with genmbcs.c MBCSAddFromUnicode() and transformEUC() */
// static UBool U_CALLCONV
// writeStage3Roundtrip(const void *context, uint32_t value, UChar32 codePoints[32]) {
//     UConverterMBCSTable *mbcsTable=(UConverterMBCSTable *)context;
//     const uint16_t *table;
//     uint32_t *stage2;
//     uint8_t *bytes, *p;
//     UChar32 c;
//     int32_t i, st3;

//     table=mbcsTable->fromUnicodeTable;
//     bytes=(uint8_t *)mbcsTable->fromUnicodeBytes;

//     /* for EUC outputTypes, modify the value like genmbcs.c's transformEUC() */
//     switch(mbcsTable->outputType) {
//     case MBCS_OUTPUT_3_EUC:
//         if(value<=0xffff) {
//             /* short sequences are stored directly */
//             /* code set 0 or 1 */
//         } else if(value<=0x8effff) {
//             /* code set 2 */
//             value&=0x7fff;
//         } else /* first byte is 0x8f */ {
//             /* code set 3 */
//             value&=0xff7f;
//         }
//         break;
//     case MBCS_OUTPUT_4_EUC:
//         if(value<=0xffffff) {
//             /* short sequences are stored directly */
//             /* code set 0 or 1 */
//         } else if(value<=0x8effffff) {
//             /* code set 2 */
//             value&=0x7fffff;
//         } else /* first byte is 0x8f */ {
//             /* code set 3 */
//             value&=0xff7fff;
//         }
//         break;
//     default:
//         break;
//     }

//     for(i=0; i<=0x1f; ++value, ++i) {
//         c=codePoints[i];
//         if(c<0) {
//             continue;
//         }

//         /* locate the stage 2 & 3 data */
//         stage2=((uint32_t *)table)+table[c>>10]+((c>>4)&0x3f);
//         p=bytes;
//         st3=(int32_t)(uint16_t)*stage2*16+(c&0xf);

//         /* write the codepage bytes into stage 3 */
//         switch(mbcsTable->outputType) {
//         case MBCS_OUTPUT_3:
//         case MBCS_OUTPUT_4_EUC:
//             p+=st3*3;
//             p[0]=(uint8_t)(value>>16);
//             p[1]=(uint8_t)(value>>8);
//             p[2]=(uint8_t)value;
//             break;
//         case MBCS_OUTPUT_4:
//             ((uint32_t *)p)[st3]=value;
//             break;
//         default:
//             /* 2 bytes per character */
//             ((uint16_t *)p)[st3]=(uint16_t)value;
//             break;
//         }

//         /* set the roundtrip flag */
//         *stage2|=(1UL<<(16+(c&0xf)));
//     }
//     return TRUE;
//  }

// static void
// reconstituteData(UConverterMBCSTable *mbcsTable,
//                  uint32_t stage1Length, uint32_t stage2Length,
//                  uint32_t fullStage2Length,  /* lengths are numbers of units, not bytes */
//                  UErrorCode *pErrorCode) {
//     uint16_t *stage1;
//     uint32_t *stage2;
//     uint32_t dataLength=stage1Length*2+fullStage2Length*4+mbcsTable->fromUBytesLength;
//     mbcsTable->reconstitutedData=(uint8_t *)uprv_malloc(dataLength);
//     if(mbcsTable->reconstitutedData==NULL) {
//         *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
//         return;
//     }
//     uprv_memset(mbcsTable->reconstitutedData, 0, dataLength);

//     /* copy existing data and reroute the pointers */
//     stage1=(uint16_t *)mbcsTable->reconstitutedData;
//     uprv_memcpy(stage1, mbcsTable->fromUnicodeTable, stage1Length*2);

//     stage2=(uint32_t *)(stage1+stage1Length);
//     uprv_memcpy(stage2+(fullStage2Length-stage2Length),
//                 mbcsTable->fromUnicodeTable+stage1Length,
//                 stage2Length*4);

//     mbcsTable->fromUnicodeTable=stage1;
//     mbcsTable->fromUnicodeBytes=(uint8_t *)(stage2+fullStage2Length);

//     /* indexes into stage 2 count from the bottom of the fromUnicodeTable */
//     stage2=(uint32_t *)stage1;

//     /* reconstitute the initial part of stage 2 from the mbcsIndex */
//     {
//         int32_t stageUTF8Length=((int32_t)mbcsTable->maxFastUChar+1)>>6;
//         int32_t stageUTF8Index=0;
//         int32_t st1, st2, st3, i;

//         for(st1=0; stageUTF8Index<stageUTF8Length; ++st1) {
//             st2=stage1[st1];
//             if(st2!=(int32_t)stage1Length/2) {
//                 /* each stage 2 block has 64 entries corresponding to 16 entries in the mbcsIndex */
//                 for(i=0; i<16; ++i) {
//                     st3=mbcsTable->mbcsIndex[stageUTF8Index++];
//                     if(st3!=0) {
//                         /* an stage 2 entry's index is per stage 3 16-block, not per stage 3 entry */
//                         st3>>=4;
//                         /*
//                          * 4 stage 2 entries point to 4 consecutive stage 3 16-blocks which are
//                          * allocated together as a single 64-block for access from the mbcsIndex
//                          */
//                         stage2[st2++]=st3++;
//                         stage2[st2++]=st3++;
//                         stage2[st2++]=st3++;
//                         stage2[st2++]=st3;
//                     } else {
//                         /* no stage 3 block, skip */
//                         st2+=4;
//                     }
//                 }
//             } else {
//                 /* no stage 2 block, skip */
//                 stageUTF8Index+=16;
//             }
//         }
//     }

//     /* reconstitute fromUnicodeBytes with roundtrips from toUnicode data */
//     ucnv_MBCSEnumToUnicode(mbcsTable, writeStage3Roundtrip, mbcsTable, pErrorCode);
// }

// /* MBCS setup functions ----------------------------------------------------- */

// static void U_CALLCONV
// ucnv_MBCSLoad(UConverterSharedData *sharedData,
//           UConverterLoadArgs *pArgs,
//           const uint8_t *raw,
//           UErrorCode *pErrorCode) {
//     UDataInfo info;
//     UConverterMBCSTable *mbcsTable=&sharedData->mbcs;
//     _MBCSHeader *header=(_MBCSHeader *)raw;
//     uint32_t offset;
//     uint32_t headerLength;
//     UBool noFromU=FALSE;

//     if(header->version[0]==4) {
//         headerLength=MBCS_HEADER_V4_LENGTH;
//     } else if(header->version[0]==5 && header->version[1]>=3 &&
//               (header->options&MBCS_OPT_UNKNOWN_INCOMPATIBLE_MASK)==0) {
//         headerLength=header->options&MBCS_OPT_LENGTH_MASK;
//         noFromU=(UBool)((header->options&MBCS_OPT_NO_FROM_U)!=0);
//     } else {
//         *pErrorCode=U_INVALID_TABLE_FORMAT;
//         return;
//     }

//     mbcsTable->outputType=(uint8_t)header->flags;
//     if(noFromU && mbcsTable->outputType==MBCS_OUTPUT_1) {
//         *pErrorCode=U_INVALID_TABLE_FORMAT;
//         return;
//     }

//     /* extension data, header version 4.2 and higher */
//     offset=header->flags>>8;
//     if(offset!=0) {
//         mbcsTable->extIndexes=(const int32_t *)(raw+offset);
//     }

//     if(mbcsTable->outputType==MBCS_OUTPUT_EXT_ONLY) {
//         UConverterLoadArgs args=UCNV_LOAD_ARGS_INITIALIZER;
//         UConverterSharedData *baseSharedData;
//         const int32_t *extIndexes;
//         const char *baseName;

//         /* extension-only file, load the base table and set values appropriately */
//         if((extIndexes=mbcsTable->extIndexes)==NULL) {
//             /* extension-only file without extension */
//             *pErrorCode=U_INVALID_TABLE_FORMAT;
//             return;
//         }

//         if(pArgs->nestedLoads!=1) {
//             /* an extension table must not be loaded as a base table */
//             *pErrorCode=U_INVALID_TABLE_FILE;
//             return;
//         }

//         /* load the base table */
//         baseName=(const char *)header+headerLength*4;
//         if(0==uprv_strcmp(baseName, sharedData->staticData->name)) {
//             /* forbid loading this same extension-only file */
//             *pErrorCode=U_INVALID_TABLE_FORMAT;
//             return;
//         }

//         /* TODO parse package name out of the prefix of the base name in the extension .cnv file? */
//         args.size=sizeof(UConverterLoadArgs);
//         args.nestedLoads=2;
//         args.onlyTestIsLoadable=pArgs->onlyTestIsLoadable;
//         args.reserved=pArgs->reserved;
//         args.options=pArgs->options;
//         args.pkg=pArgs->pkg;
//         args.name=baseName;
//         baseSharedData=ucnv_load(&args, pErrorCode);
//         if(U_FAILURE(*pErrorCode)) {
//             return;
//         }
//         if( baseSharedData->staticData->conversionType!=UCNV_MBCS ||
//             baseSharedData->mbcs.baseSharedData!=NULL
//         ) {
//             ucnv_unload(baseSharedData);
//             *pErrorCode=U_INVALID_TABLE_FORMAT;
//             return;
//         }
//         if(pArgs->onlyTestIsLoadable) {
//             /*
//              * Exit as soon as we know that we can load the converter
//              * and the format is valid and supported.
//              * The worst that can happen in the following code is a memory
//              * allocation error.
//              */
//             ucnv_unload(baseSharedData);
//             return;
//         }

//         /* copy the base table data */
//         uprv_memcpy(mbcsTable, &baseSharedData->mbcs, sizeof(UConverterMBCSTable));

//         /* overwrite values with relevant ones for the extension converter */
//         mbcsTable->baseSharedData=baseSharedData;
//         mbcsTable->extIndexes=extIndexes;

//         /*
//          * It would be possible to share the swapLFNL data with a base converter,
//          * but the generated name would have to be different, and the memory
//          * would have to be free'd only once.
//          * It is easier to just create the data for the extension converter
//          * separately when it is requested.
//          */
//         mbcsTable->swapLFNLStateTable=NULL;
//         mbcsTable->swapLFNLFromUnicodeBytes=NULL;
//         mbcsTable->swapLFNLName=NULL;

//         /*
//          * The reconstitutedData must be deleted only when the base converter
//          * is unloaded.
//          */
//         mbcsTable->reconstitutedData=NULL;

//         /*
//          * Set a special, runtime-only outputType if the extension converter
//          * is a DBCS version of a base converter that also maps single bytes.
//          */
//         if( sharedData->staticData->conversionType==UCNV_DBCS ||
//                 (sharedData->staticData->conversionType==UCNV_MBCS &&
//                  sharedData->staticData->minBytesPerChar>=2)
//         ) {
//             if(baseSharedData->mbcs.outputType==MBCS_OUTPUT_2_SISO) {
//                 /* the base converter is SI/SO-stateful */
//                 int32_t entry;

//                 /* get the dbcs state from the state table entry for SO=0x0e */
//                 entry=mbcsTable->stateTable[0][0xe];
//                 if( MBCS_ENTRY_IS_FINAL(entry) &&
//                     MBCS_ENTRY_FINAL_ACTION(entry)==MBCS_STATE_CHANGE_ONLY &&
//                     MBCS_ENTRY_FINAL_STATE(entry)!=0
//                 ) {
//                     mbcsTable->dbcsOnlyState=(uint8_t)MBCS_ENTRY_FINAL_STATE(entry);

//                     mbcsTable->outputType=MBCS_OUTPUT_DBCS_ONLY;
//                 }
//             } else if(
//                 baseSharedData->staticData->conversionType==UCNV_MBCS &&
//                 baseSharedData->staticData->minBytesPerChar==1 &&
//                 baseSharedData->staticData->maxBytesPerChar==2 &&
//                 mbcsTable->countStates<=127
//             ) {
//                 /* non-stateful base converter, need to modify the state table */
//                 int32_t (*newStateTable)[256];
//                 int32_t *state;
//                 int32_t i, count;

//                 /* allocate a new state table and copy the base state table contents */
//                 count=mbcsTable->countStates;
//                 newStateTable=(int32_t (*)[256])uprv_malloc((count+1)*1024);
//                 if(newStateTable==NULL) {
//                     ucnv_unload(baseSharedData);
//                     *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
//                     return;
//                 }

//                 uprv_memcpy(newStateTable, mbcsTable->stateTable, count*1024);

//                 /* change all final single-byte entries to go to a new all-illegal state */
//                 state=newStateTable[0];
//                 for(i=0; i<256; ++i) {
//                     if(MBCS_ENTRY_IS_FINAL(state[i])) {
//                         state[i]=MBCS_ENTRY_TRANSITION(count, 0);
//                     }
//                 }

//                 /* build the new all-illegal state */
//                 state=newStateTable[count];
//                 for(i=0; i<256; ++i) {
//                     state[i]=MBCS_ENTRY_FINAL(0, MBCS_STATE_ILLEGAL, 0);
//                 }
//                 mbcsTable->stateTable=(const int32_t (*)[256])newStateTable;
//                 mbcsTable->countStates=(uint8_t)(count+1);
//                 mbcsTable->stateTableOwned=TRUE;

//                 mbcsTable->outputType=MBCS_OUTPUT_DBCS_ONLY;
//             }
//         }

//         /*
//          * unlike below for files with base tables, do not get the unicodeMask
//          * from the sharedData; instead, use the base table's unicodeMask,
//          * which we copied in the memcpy above;
//          * this is necessary because the static data unicodeMask, especially
//          * the UCNV_HAS_SUPPLEMENTARY flag, is part of the base table data
//          */
//     } else {
//         /* conversion file with a base table; an additional extension table is optional */
//         /* make sure that the output type is known */
//         switch(mbcsTable->outputType) {
//         case MBCS_OUTPUT_1:
//         case MBCS_OUTPUT_2:
//         case MBCS_OUTPUT_3:
//         case MBCS_OUTPUT_4:
//         case MBCS_OUTPUT_3_EUC:
//         case MBCS_OUTPUT_4_EUC:
//         case MBCS_OUTPUT_2_SISO:
//             /* OK */
//             break;
//         default:
//             *pErrorCode=U_INVALID_TABLE_FORMAT;
//             return;
//         }
//         if(pArgs->onlyTestIsLoadable) {
//             /*
//              * Exit as soon as we know that we can load the converter
//              * and the format is valid and supported.
//              * The worst that can happen in the following code is a memory
//              * allocation error.
//              */
//             return;
//         }

//         mbcsTable->countStates=(uint8_t)header->countStates;
//         mbcsTable->countToUFallbacks=header->countToUFallbacks;
//         mbcsTable->stateTable=(const int32_t (*)[256])(raw+headerLength*4);
//         mbcsTable->toUFallbacks=(const _MBCSToUFallback *)(mbcsTable->stateTable+header->countStates);
//         mbcsTable->unicodeCodeUnits=(const uint16_t *)(raw+header->offsetToUCodeUnits);

//         mbcsTable->fromUnicodeTable=(const uint16_t *)(raw+header->offsetFromUTable);
//         mbcsTable->fromUnicodeBytes=(const uint8_t *)(raw+header->offsetFromUBytes);
//         mbcsTable->fromUBytesLength=header->fromUBytesLength;

//         /*
//          * converter versions 6.1 and up contain a unicodeMask that is
//          * used here to select the most efficient function implementations
//          */
//         info.size=sizeof(UDataInfo);
//         udata_getInfo((UDataMemory *)sharedData->dataMemory, &info);
//         if(info.formatVersion[0]>6 || (info.formatVersion[0]==6 && info.formatVersion[1]>=1)) {
//             /* mask off possible future extensions to be safe */
//             mbcsTable->unicodeMask=(uint8_t)(sharedData->staticData->unicodeMask&3);
//         } else {
//             /* for older versions, assume worst case: contains anything possible (prevent over-optimizations) */
//             mbcsTable->unicodeMask=UCNV_HAS_SUPPLEMENTARY|UCNV_HAS_SURROGATES;
//         }

//         /*
//          * _MBCSHeader.version 4.3 adds utf8Friendly data structures.
//          * Check for the header version, SBCS vs. MBCS, and for whether the
//          * data structures are optimized for code points as high as what the
//          * runtime code is designed for.
//          * The implementation does not handle mapping tables with entries for
//          * unpaired surrogates.
//          */
//         if( header->version[1]>=3 &&
//             (mbcsTable->unicodeMask&UCNV_HAS_SURROGATES)==0 &&
//             (mbcsTable->countStates==1 ?
//                 (header->version[2]>=(SBCS_FAST_MAX>>8)) :
//                 (header->version[2]>=(MBCS_FAST_MAX>>8))
//             )
//         ) {
//             mbcsTable->utf8Friendly=TRUE;

//             if(mbcsTable->countStates==1) {
//                 /*
//                  * SBCS: Stage 3 is allocated in 64-entry blocks for U+0000..SBCS_FAST_MAX or higher.
//                  * Build a table with indexes to each block, to be used instead of
//                  * the regular stage 1/2 table.
//                  */
//                 int32_t i;
//                 for(i=0; i<(SBCS_FAST_LIMIT>>6); ++i) {
//                     mbcsTable->sbcsIndex[i]=mbcsTable->fromUnicodeTable[mbcsTable->fromUnicodeTable[i>>4]+((i<<2)&0x3c)];
//                 }
//                 /* set SBCS_FAST_MAX to reflect the reach of sbcsIndex[] even if header->version[2]>(SBCS_FAST_MAX>>8) */
//                 mbcsTable->maxFastUChar=SBCS_FAST_MAX;
//             } else {
//                 /*
//                  * MBCS: Stage 3 is allocated in 64-entry blocks for U+0000..MBCS_FAST_MAX or higher.
//                  * The .cnv file is prebuilt with an additional stage table with indexes
//                  * to each block.
//                  */
//                 mbcsTable->mbcsIndex=(const uint16_t *)
//                     (mbcsTable->fromUnicodeBytes+
//                      (noFromU ? 0 : mbcsTable->fromUBytesLength));
//                 mbcsTable->maxFastUChar=(((UChar)header->version[2])<<8)|0xff;
//             }
//         }

//         /* calculate a bit set of 4 ASCII characters per bit that round-trip to ASCII bytes */
//         {
//             uint32_t asciiRoundtrips=0xffffffff;
//             int32_t i;

//             for(i=0; i<0x80; ++i) {
//                 if(mbcsTable->stateTable[0][i]!=MBCS_ENTRY_FINAL(0, MBCS_STATE_VALID_DIRECT_16, i)) {
//                     asciiRoundtrips&=~((uint32_t)1<<(i>>2));
//                 }
//             }
//             mbcsTable->asciiRoundtrips=asciiRoundtrips;
//         }

//         if(noFromU) {
//             uint32_t stage1Length=
//                 mbcsTable->unicodeMask&UCNV_HAS_SUPPLEMENTARY ?
//                     0x440 : 0x40;
//             uint32_t stage2Length=
//                 (header->offsetFromUBytes-header->offsetFromUTable)/4-
//                 stage1Length/2;
//             reconstituteData(mbcsTable, stage1Length, stage2Length, header->fullStage2Length, pErrorCode);
//         }
//     }

//     /* Set the impl pointer here so that it is set for both extension-only and base tables. */
//     if(mbcsTable->utf8Friendly) {
//         if(mbcsTable->countStates==1) {
//             sharedData->impl=&_SBCSUTF8Impl;
//         } else {
//             if(mbcsTable->outputType==MBCS_OUTPUT_2) {
//                 sharedData->impl=&_DBCSUTF8Impl;
//             }
//         }
//     }

//     if(mbcsTable->outputType==MBCS_OUTPUT_DBCS_ONLY || mbcsTable->outputType==MBCS_OUTPUT_2_SISO) {
//         /*
//          * MBCS_OUTPUT_DBCS_ONLY: No SBCS mappings, therefore ASCII does not roundtrip.
//          * MBCS_OUTPUT_2_SISO: Bypass the ASCII fastpath to handle prevLength correctly.
//          */
//         mbcsTable->asciiRoundtrips=0;
//     }
// }

// static void U_CALLCONV
// ucnv_MBCSUnload(UConverterSharedData *sharedData) {
//     UConverterMBCSTable *mbcsTable=&sharedData->mbcs;

//     if(mbcsTable->swapLFNLStateTable!=NULL) {
//         uprv_free(mbcsTable->swapLFNLStateTable);
//     }
//     if(mbcsTable->stateTableOwned) {
//         uprv_free((void *)mbcsTable->stateTable);
//     }
//     if(mbcsTable->baseSharedData!=NULL) {
//         ucnv_unload(mbcsTable->baseSharedData);
//     }
//     if(mbcsTable->reconstitutedData!=NULL) {
//         uprv_free(mbcsTable->reconstitutedData);
//     }
// }

// static void U_CALLCONV
// ucnv_MBCSOpen(UConverter *cnv,
//               UConverterLoadArgs *pArgs,
//               UErrorCode *pErrorCode) {
//     UConverterMBCSTable *mbcsTable;
//     const int32_t *extIndexes;
//     uint8_t outputType;
//     int8_t maxBytesPerUChar;

//     if(pArgs->onlyTestIsLoadable) {
//         return;
//     }

//     mbcsTable=&cnv->sharedData->mbcs;
//     outputType=mbcsTable->outputType;

//     if(outputType==MBCS_OUTPUT_DBCS_ONLY) {
//         /* the swaplfnl option does not apply, remove it */
//         cnv->options=pArgs->options&=~UCNV_OPTION_SWAP_LFNL;
//     }

//     if((pArgs->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         /* do this because double-checked locking is broken */
//         UBool isCached;

//         icu::umtx_lock(NULL);
//         isCached=mbcsTable->swapLFNLStateTable!=NULL;
//         icu::umtx_unlock(NULL);

//         if(!isCached) {
//             if(!_EBCDICSwapLFNL(cnv->sharedData, pErrorCode)) {
//                 if(U_FAILURE(*pErrorCode)) {
//                     return; /* something went wrong */
//                 }

//                 /* the option does not apply, remove it */
//                 cnv->options=pArgs->options&=~UCNV_OPTION_SWAP_LFNL;
//             }
//         }
//     }

//     if(uprv_strstr(pArgs->name, "18030")!=NULL) {
//         if(uprv_strstr(pArgs->name, "gb18030")!=NULL || uprv_strstr(pArgs->name, "GB18030")!=NULL) {
//             /* set a flag for GB 18030 mode, which changes the callback behavior */
//             cnv->options|=_MBCS_OPTION_GB18030;
//         }
//     } else if((uprv_strstr(pArgs->name, "KEIS")!=NULL) || (uprv_strstr(pArgs->name, "keis")!=NULL)) {
//         /* set a flag for KEIS converter, which changes the SI/SO character sequence */
//         cnv->options|=_MBCS_OPTION_KEIS;
//     } else if((uprv_strstr(pArgs->name, "JEF")!=NULL) || (uprv_strstr(pArgs->name, "jef")!=NULL)) {
//         /* set a flag for JEF converter, which changes the SI/SO character sequence */
//         cnv->options|=_MBCS_OPTION_JEF;
//     } else if((uprv_strstr(pArgs->name, "JIPS")!=NULL) || (uprv_strstr(pArgs->name, "jips")!=NULL)) {
//         /* set a flag for JIPS converter, which changes the SI/SO character sequence */
//         cnv->options|=_MBCS_OPTION_JIPS;
//     }

//     /* fix maxBytesPerUChar depending on outputType and options etc. */
//     if(outputType==MBCS_OUTPUT_2_SISO) {
//         cnv->maxBytesPerUChar=3; /* SO+DBCS */
//     }

//     extIndexes=mbcsTable->extIndexes;
//     if(extIndexes!=NULL) {
//         maxBytesPerUChar=(int8_t)UCNV_GET_MAX_BYTES_PER_UCHAR(extIndexes);
//         if(outputType==MBCS_OUTPUT_2_SISO) {
//             ++maxBytesPerUChar; /* SO + multiple DBCS */
//         }

//         if(maxBytesPerUChar>cnv->maxBytesPerUChar) {
//             cnv->maxBytesPerUChar=maxBytesPerUChar;
//         }
//     }

// #if 0
//     /*
//      * documentation of UConverter fields used for status
//      * all of these fields are (re)set to 0 by ucnv_bld.c and ucnv_reset()
//      */

//     /* toUnicode */
//     cnv->toUnicodeStatus=0;     /* offset */
//     cnv->mode=0;                /* state */
//     cnv->toULength=0;           /* byteIndex */

//     /* fromUnicode */
//     cnv->fromUChar32=0;
//     cnv->fromUnicodeStatus=1;   /* prevLength */
// #endif
// }

// U_CDECL_BEGIN

// static const char* U_CALLCONV
// ucnv_MBCSGetName(const UConverter *cnv) {
//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0 && cnv->sharedData->mbcs.swapLFNLName!=NULL) {
//         return cnv->sharedData->mbcs.swapLFNLName;
//     } else {
//         return cnv->sharedData->staticData->name;
//     }
// }
// U_CDECL_END


// /* MBCS-to-Unicode conversion functions ------------------------------------- */

// static UChar32 U_CALLCONV
// ucnv_MBCSGetFallback(UConverterMBCSTable *mbcsTable, uint32_t offset) {
//     const _MBCSToUFallback *toUFallbacks;
//     uint32_t i, start, limit;

//     limit=mbcsTable->countToUFallbacks;
//     if(limit>0) {
//         /* do a binary search for the fallback mapping */
//         toUFallbacks=mbcsTable->toUFallbacks;
//         start=0;
//         while(start<limit-1) {
//             i=(start+limit)/2;
//             if(offset<toUFallbacks[i].offset) {
//                 limit=i;
//             } else {
//                 start=i;
//             }
//         }

//         /* did we really find it? */
//         if(offset==toUFallbacks[start].offset) {
//             return toUFallbacks[start].codePoint;
//         }
//     }

//     return 0xfffe;
// }

// /* This version of ucnv_MBCSToUnicodeWithOffsets() is optimized for single-byte, single-state codepages. */
// static void
// ucnv_MBCSSingleToUnicodeWithOffsets(UConverterToUnicodeArgs *pArgs,
//                                 UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const uint8_t *source, *sourceLimit;
//     UChar *target;
//     const UChar *targetLimit;
//     int32_t *offsets;

//     const int32_t (*stateTable)[256];

//     int32_t sourceIndex;

//     int32_t entry;
//     UChar c;
//     uint8_t action;

//     /* set up the local pointers */
//     cnv=pArgs->converter;
//     source=(const uint8_t *)pArgs->source;
//     sourceLimit=(const uint8_t *)pArgs->sourceLimit;
//     target=pArgs->target;
//     targetLimit=pArgs->targetLimit;
//     offsets=pArgs->offsets;

//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         stateTable=(const int32_t (*)[256])cnv->sharedData->mbcs.swapLFNLStateTable;
//     } else {
//         stateTable=cnv->sharedData->mbcs.stateTable;
//     }

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     sourceIndex=0;

//     /* conversion loop */
//     while(source<sourceLimit) {
//         /*
//          * This following test is to see if available input would overflow the output.
//          * It does not catch output of more than one code unit that
//          * overflows as a result of a surrogate pair or callback output
//          * from the last source byte.
//          * Therefore, those situations also test for overflows and will
//          * then break the loop, too.
//          */
//         if(target>=targetLimit) {
//             /* target is full */
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }

//         entry=stateTable[0][*source++];
//         /* MBCS_ENTRY_IS_FINAL(entry) */

//         /* test the most common case first */
//         if(MBCS_ENTRY_FINAL_IS_VALID_DIRECT_16(entry)) {
//             /* output BMP code point */
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             if(offsets!=NULL) {
//                 *offsets++=sourceIndex;
//             }

//             /* normal end of action codes: prepare for a new character */
//             ++sourceIndex;
//             continue;
//         }

//         /*
//          * An if-else-if chain provides more reliable performance for
//          * the most common cases compared to a switch.
//          */
//         action=(uint8_t)(MBCS_ENTRY_FINAL_ACTION(entry));
//         if(action==MBCS_STATE_VALID_DIRECT_20 ||
//            (action==MBCS_STATE_FALLBACK_DIRECT_20 && UCNV_TO_U_USE_FALLBACK(cnv))
//         ) {
//             entry=MBCS_ENTRY_FINAL_VALUE(entry);
//             /* output surrogate pair */
//             *target++=(UChar)(0xd800|(UChar)(entry>>10));
//             if(offsets!=NULL) {
//                 *offsets++=sourceIndex;
//             }
//             c=(UChar)(0xdc00|(UChar)(entry&0x3ff));
//             if(target<targetLimit) {
//                 *target++=c;
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//             } else {
//                 /* target overflow */
//                 cnv->UCharErrorBuffer[0]=c;
//                 cnv->UCharErrorBufferLength=1;
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 break;
//             }

//             ++sourceIndex;
//             continue;
//         } else if(action==MBCS_STATE_FALLBACK_DIRECT_16) {
//             if(UCNV_TO_U_USE_FALLBACK(cnv)) {
//                 /* output BMP code point */
//                 *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }

//                 ++sourceIndex;
//                 continue;
//             }
//         } else if(action==MBCS_STATE_UNASSIGNED) {
//             /* just fall through */
//         } else if(action==MBCS_STATE_ILLEGAL) {
//             /* callback(illegal) */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//         } else {
//             /* reserved, must never occur */
//             ++sourceIndex;
//             continue;
//         }

//         if(U_FAILURE(*pErrorCode)) {
//             /* callback(illegal) */
//             break;
//         } else /* unassigned sequences indicated with byteIndex>0 */ {
//             /* try an extension mapping */
//             pArgs->source=(const char *)source;
//             cnv->toUBytes[0]=*(source-1);
//             cnv->toULength=_extToU(cnv, cnv->sharedData,
//                                     1, &source, sourceLimit,
//                                     &target, targetLimit,
//                                     &offsets, sourceIndex,
//                                     pArgs->flush,
//                                     pErrorCode);
//             sourceIndex+=1+(int32_t)(source-(const uint8_t *)pArgs->source);

//             if(U_FAILURE(*pErrorCode)) {
//                 /* not mappable or buffer overflow */
//                 break;
//             }
//         }
//     }

//     /* write back the updated pointers */
//     pArgs->source=(const char *)source;
//     pArgs->target=target;
//     pArgs->offsets=offsets;
// }

// /*
//  * This version of ucnv_MBCSSingleToUnicodeWithOffsets() is optimized for single-byte, single-state codepages
//  * that only map to and from the BMP.
//  * In addition to single-byte optimizations, the offset calculations
//  * become much easier.
//  */
// static void
// ucnv_MBCSSingleToBMPWithOffsets(UConverterToUnicodeArgs *pArgs,
//                             UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const uint8_t *source, *sourceLimit, *lastSource;
//     UChar *target;
//     int32_t targetCapacity, length;
//     int32_t *offsets;

//     const int32_t (*stateTable)[256];

//     int32_t sourceIndex;

//     int32_t entry;
//     uint8_t action;

//     /* set up the local pointers */
//     cnv=pArgs->converter;
//     source=(const uint8_t *)pArgs->source;
//     sourceLimit=(const uint8_t *)pArgs->sourceLimit;
//     target=pArgs->target;
//     targetCapacity=(int32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;

//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         stateTable=(const int32_t (*)[256])cnv->sharedData->mbcs.swapLFNLStateTable;
//     } else {
//         stateTable=cnv->sharedData->mbcs.stateTable;
//     }

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     sourceIndex=0;
//     lastSource=source;

//     /*
//      * since the conversion here is 1:1 UChar:uint8_t, we need only one counter
//      * for the minimum of the sourceLength and targetCapacity
//      */
//     length=(int32_t)(sourceLimit-source);
//     if(length<targetCapacity) {
//         targetCapacity=length;
//     }

// #if MBCS_UNROLL_SINGLE_TO_BMP
//     /* unrolling makes it faster on Pentium III/Windows 2000 */
//     /* unroll the loop with the most common case */
// unrolled:
//     if(targetCapacity>=16) {
//         int32_t count, loops, oredEntries;

//         loops=count=targetCapacity>>4;
//         do {
//             oredEntries=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             oredEntries|=entry=stateTable[0][*source++];
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);

//             /* were all 16 entries really valid? */
//             if(!MBCS_ENTRY_FINAL_IS_VALID_DIRECT_16(oredEntries)) {
//                 /* no, return to the first of these 16 */
//                 source-=16;
//                 target-=16;
//                 break;
//             }
//         } while(--count>0);
//         count=loops-count;
//         targetCapacity-=16*count;

//         if(offsets!=NULL) {
//             lastSource+=16*count;
//             while(count>0) {
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 --count;
//             }
//         }
//     }
// #endif

//     /* conversion loop */
//     while(targetCapacity > 0 && source < sourceLimit) {
//         entry=stateTable[0][*source++];
//         /* MBCS_ENTRY_IS_FINAL(entry) */

//         /* test the most common case first */
//         if(MBCS_ENTRY_FINAL_IS_VALID_DIRECT_16(entry)) {
//             /* output BMP code point */
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             --targetCapacity;
//             continue;
//         }

//         /*
//          * An if-else-if chain provides more reliable performance for
//          * the most common cases compared to a switch.
//          */
//         action=(uint8_t)(MBCS_ENTRY_FINAL_ACTION(entry));
//         if(action==MBCS_STATE_FALLBACK_DIRECT_16) {
//             if(UCNV_TO_U_USE_FALLBACK(cnv)) {
//                 /* output BMP code point */
//                 *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 --targetCapacity;
//                 continue;
//             }
//         } else if(action==MBCS_STATE_UNASSIGNED) {
//             /* just fall through */
//         } else if(action==MBCS_STATE_ILLEGAL) {
//             /* callback(illegal) */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//         } else {
//             /* reserved, must never occur */
//             continue;
//         }

//         /* set offsets since the start or the last extension */
//         if(offsets!=NULL) {
//             int32_t count=(int32_t)(source-lastSource);

//             /* predecrement: do not set the offset for the callback-causing character */
//             while(--count>0) {
//                 *offsets++=sourceIndex++;
//             }
//             /* offset and sourceIndex are now set for the current character */
//         }

//         if(U_FAILURE(*pErrorCode)) {
//             /* callback(illegal) */
//             break;
//         } else /* unassigned sequences indicated with byteIndex>0 */ {
//             /* try an extension mapping */
//             lastSource=source;
//             cnv->toUBytes[0]=*(source-1);
//             cnv->toULength=_extToU(cnv, cnv->sharedData,
//                                     1, &source, sourceLimit,
//                                     &target, pArgs->targetLimit,
//                                     &offsets, sourceIndex,
//                                     pArgs->flush,
//                                     pErrorCode);
//             sourceIndex+=1+(int32_t)(source-lastSource);

//             if(U_FAILURE(*pErrorCode)) {
//                 /* not mappable or buffer overflow */
//                 break;
//             }

//             /* recalculate the targetCapacity after an extension mapping */
//             targetCapacity=(int32_t)(pArgs->targetLimit-target);
//             length=(int32_t)(sourceLimit-source);
//             if(length<targetCapacity) {
//                 targetCapacity=length;
//             }
//         }

// #if MBCS_UNROLL_SINGLE_TO_BMP
//         /* unrolling makes it faster on Pentium III/Windows 2000 */
//         goto unrolled;
// #endif
//     }

//     if(U_SUCCESS(*pErrorCode) && source<sourceLimit && target>=pArgs->targetLimit) {
//         /* target is full */
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//     }

//     /* set offsets since the start or the last callback */
//     if(offsets!=NULL) {
//         size_t count=source-lastSource;
//         while(count>0) {
//             *offsets++=sourceIndex++;
//             --count;
//         }
//     }

//     /* write back the updated pointers */
//     pArgs->source=(const char *)source;
//     pArgs->target=target;
//     pArgs->offsets=offsets;
// }

// static UBool
// hasValidTrailBytes(const int32_t (*stateTable)[256], uint8_t state) {
//     const int32_t *row=stateTable[state];
//     int32_t b, entry;
//     /* First test for final entries in this state for some commonly valid byte values. */
//     entry=row[0xa1];
//     if( !MBCS_ENTRY_IS_TRANSITION(entry) &&
//         MBCS_ENTRY_FINAL_ACTION(entry)!=MBCS_STATE_ILLEGAL
//     ) {
//         return TRUE;
//     }
//     entry=row[0x41];
//     if( !MBCS_ENTRY_IS_TRANSITION(entry) &&
//         MBCS_ENTRY_FINAL_ACTION(entry)!=MBCS_STATE_ILLEGAL
//     ) {
//         return TRUE;
//     }
//     /* Then test for final entries in this state. */
//     for(b=0; b<=0xff; ++b) {
//         entry=row[b];
//         if( !MBCS_ENTRY_IS_TRANSITION(entry) &&
//             MBCS_ENTRY_FINAL_ACTION(entry)!=MBCS_STATE_ILLEGAL
//         ) {
//             return TRUE;
//         }
//     }
//     /* Then recurse for transition entries. */
//     for(b=0; b<=0xff; ++b) {
//         entry=row[b];
//         if( MBCS_ENTRY_IS_TRANSITION(entry) &&
//             hasValidTrailBytes(stateTable, (uint8_t)MBCS_ENTRY_TRANSITION_STATE(entry))
//         ) {
//             return TRUE;
//         }
//     }
//     return FALSE;
// }

// /*
//  * Is byte b a single/lead byte in this state?
//  * Recurse for transition states, because here we don't want to say that
//  * b is a lead byte if all byte sequences that start with b are illegal.
//  */
// static UBool
// isSingleOrLead(const int32_t (*stateTable)[256], uint8_t state, UBool isDBCSOnly, uint8_t b) {
//     const int32_t *row=stateTable[state];
//     int32_t entry=row[b];
//     if(MBCS_ENTRY_IS_TRANSITION(entry)) {   /* lead byte */
//         return hasValidTrailBytes(stateTable, (uint8_t)MBCS_ENTRY_TRANSITION_STATE(entry));
//     } else {
//         uint8_t action=(uint8_t)(MBCS_ENTRY_FINAL_ACTION(entry));
//         if(action==MBCS_STATE_CHANGE_ONLY && isDBCSOnly) {
//             return FALSE;   /* SI/SO are illegal for DBCS-only conversion */
//         } else {
//             return action!=MBCS_STATE_ILLEGAL;
//         }
//     }
// }

// U_CFUNC void
// ucnv_MBCSToUnicodeWithOffsets(UConverterToUnicodeArgs *pArgs,
//                           UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const uint8_t *source, *sourceLimit;
//     UChar *target;
//     const UChar *targetLimit;
//     int32_t *offsets;

//     const int32_t (*stateTable)[256];
//     const uint16_t *unicodeCodeUnits;

//     uint32_t offset;
//     uint8_t state;
//     int8_t byteIndex;
//     uint8_t *bytes;

//     int32_t sourceIndex, nextSourceIndex;

//     int32_t entry;
//     UChar c;
//     uint8_t action;

//     /* use optimized function if possible */
//     cnv=pArgs->converter;

//     if(cnv->preToULength>0) {
//         /*
//          * pass sourceIndex=-1 because we continue from an earlier buffer
//          * in the future, this may change with continuous offsets
//          */
//         ucnv_extContinueMatchToU(cnv, pArgs, -1, pErrorCode);

//         if(U_FAILURE(*pErrorCode) || cnv->preToULength<0) {
//             return;
//         }
//     }

//     if(cnv->sharedData->mbcs.countStates==1) {
//         if(!(cnv->sharedData->mbcs.unicodeMask&UCNV_HAS_SUPPLEMENTARY)) {
//             ucnv_MBCSSingleToBMPWithOffsets(pArgs, pErrorCode);
//         } else {
//             ucnv_MBCSSingleToUnicodeWithOffsets(pArgs, pErrorCode);
//         }
//         return;
//     }

//     /* set up the local pointers */
//     source=(const uint8_t *)pArgs->source;
//     sourceLimit=(const uint8_t *)pArgs->sourceLimit;
//     target=pArgs->target;
//     targetLimit=pArgs->targetLimit;
//     offsets=pArgs->offsets;

//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         stateTable=(const int32_t (*)[256])cnv->sharedData->mbcs.swapLFNLStateTable;
//     } else {
//         stateTable=cnv->sharedData->mbcs.stateTable;
//     }
//     unicodeCodeUnits=cnv->sharedData->mbcs.unicodeCodeUnits;

//     /* get the converter state from UConverter */
//     offset=cnv->toUnicodeStatus;
//     byteIndex=cnv->toULength;
//     bytes=cnv->toUBytes;

//     /*
//      * if we are in the SBCS state for a DBCS-only converter,
//      * then load the DBCS state from the MBCS data
//      * (dbcsOnlyState==0 if it is not a DBCS-only converter)
//      */
//     if((state=(uint8_t)(cnv->mode))==0) {
//         state=cnv->sharedData->mbcs.dbcsOnlyState;
//     }

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     sourceIndex=byteIndex==0 ? 0 : -1;
//     nextSourceIndex=0;

//     /* conversion loop */
//     while(source<sourceLimit) {
//         /*
//          * This following test is to see if available input would overflow the output.
//          * It does not catch output of more than one code unit that
//          * overflows as a result of a surrogate pair or callback output
//          * from the last source byte.
//          * Therefore, those situations also test for overflows and will
//          * then break the loop, too.
//          */
//         if(target>=targetLimit) {
//             /* target is full */
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }

//         if(byteIndex==0) {
//             /* optimized loop for 1/2-byte input and BMP output */
//             if(offsets==NULL) {
//                 do {
//                     entry=stateTable[state][*source];
//                     if(MBCS_ENTRY_IS_TRANSITION(entry)) {
//                         state=(uint8_t)MBCS_ENTRY_TRANSITION_STATE(entry);
//                         offset=MBCS_ENTRY_TRANSITION_OFFSET(entry);

//                         ++source;
//                         if( source<sourceLimit &&
//                             MBCS_ENTRY_IS_FINAL(entry=stateTable[state][*source]) &&
//                             MBCS_ENTRY_FINAL_ACTION(entry)==MBCS_STATE_VALID_16 &&
//                             (c=unicodeCodeUnits[offset+MBCS_ENTRY_FINAL_VALUE_16(entry)])<0xfffe
//                         ) {
//                             ++source;
//                             *target++=c;
//                             state=(uint8_t)MBCS_ENTRY_FINAL_STATE(entry); /* typically 0 */
//                             offset=0;
//                         } else {
//                             /* set the state and leave the optimized loop */
//                             bytes[0]=*(source-1);
//                             byteIndex=1;
//                             break;
//                         }
//                     } else {
//                         if(MBCS_ENTRY_FINAL_IS_VALID_DIRECT_16(entry)) {
//                             /* output BMP code point */
//                             ++source;
//                             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//                             state=(uint8_t)MBCS_ENTRY_FINAL_STATE(entry); /* typically 0 */
//                         } else {
//                             /* leave the optimized loop */
//                             break;
//                         }
//                     }
//                 } while(source<sourceLimit && target<targetLimit);
//             } else /* offsets!=NULL */ {
//                 do {
//                     entry=stateTable[state][*source];
//                     if(MBCS_ENTRY_IS_TRANSITION(entry)) {
//                         state=(uint8_t)MBCS_ENTRY_TRANSITION_STATE(entry);
//                         offset=MBCS_ENTRY_TRANSITION_OFFSET(entry);

//                         ++source;
//                         if( source<sourceLimit &&
//                             MBCS_ENTRY_IS_FINAL(entry=stateTable[state][*source]) &&
//                             MBCS_ENTRY_FINAL_ACTION(entry)==MBCS_STATE_VALID_16 &&
//                             (c=unicodeCodeUnits[offset+MBCS_ENTRY_FINAL_VALUE_16(entry)])<0xfffe
//                         ) {
//                             ++source;
//                             *target++=c;
//                             if(offsets!=NULL) {
//                                 *offsets++=sourceIndex;
//                                 sourceIndex=(nextSourceIndex+=2);
//                             }
//                             state=(uint8_t)MBCS_ENTRY_FINAL_STATE(entry); /* typically 0 */
//                             offset=0;
//                         } else {
//                             /* set the state and leave the optimized loop */
//                             ++nextSourceIndex;
//                             bytes[0]=*(source-1);
//                             byteIndex=1;
//                             break;
//                         }
//                     } else {
//                         if(MBCS_ENTRY_FINAL_IS_VALID_DIRECT_16(entry)) {
//                             /* output BMP code point */
//                             ++source;
//                             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//                             if(offsets!=NULL) {
//                                 *offsets++=sourceIndex;
//                                 sourceIndex=++nextSourceIndex;
//                             }
//                             state=(uint8_t)MBCS_ENTRY_FINAL_STATE(entry); /* typically 0 */
//                         } else {
//                             /* leave the optimized loop */
//                             break;
//                         }
//                     }
//                 } while(source<sourceLimit && target<targetLimit);
//             }

//             /*
//              * these tests and break statements could be put inside the loop
//              * if C had "break outerLoop" like Java
//              */
//             if(source>=sourceLimit) {
//                 break;
//             }
//             if(target>=targetLimit) {
//                 /* target is full */
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 break;
//             }

//             ++nextSourceIndex;
//             bytes[byteIndex++]=*source++;
//         } else /* byteIndex>0 */ {
//             ++nextSourceIndex;
//             entry=stateTable[state][bytes[byteIndex++]=*source++];
//         }

//         if(MBCS_ENTRY_IS_TRANSITION(entry)) {
//             state=(uint8_t)MBCS_ENTRY_TRANSITION_STATE(entry);
//             offset+=MBCS_ENTRY_TRANSITION_OFFSET(entry);
//             continue;
//         }

//         /* save the previous state for proper extension mapping with SI/SO-stateful converters */
//         cnv->mode=state;

//         /* set the next state early so that we can reuse the entry variable */
//         state=(uint8_t)MBCS_ENTRY_FINAL_STATE(entry); /* typically 0 */

//         /*
//          * An if-else-if chain provides more reliable performance for
//          * the most common cases compared to a switch.
//          */
//         action=(uint8_t)(MBCS_ENTRY_FINAL_ACTION(entry));
//         if(action==MBCS_STATE_VALID_16) {
//             offset+=MBCS_ENTRY_FINAL_VALUE_16(entry);
//             c=unicodeCodeUnits[offset];
//             if(c<0xfffe) {
//                 /* output BMP code point */
//                 *target++=c;
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 byteIndex=0;
//             } else if(c==0xfffe) {
//                 if(UCNV_TO_U_USE_FALLBACK(cnv) && (entry=(int32_t)ucnv_MBCSGetFallback(&cnv->sharedData->mbcs, offset))!=0xfffe) {
//                     /* output fallback BMP code point */
//                     *target++=(UChar)entry;
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     byteIndex=0;
//                 }
//             } else {
//                 /* callback(illegal) */
//                 *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//             }
//         } else if(action==MBCS_STATE_VALID_DIRECT_16) {
//             /* output BMP code point */
//             *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             if(offsets!=NULL) {
//                 *offsets++=sourceIndex;
//             }
//             byteIndex=0;
//         } else if(action==MBCS_STATE_VALID_16_PAIR) {
//             offset+=MBCS_ENTRY_FINAL_VALUE_16(entry);
//             c=unicodeCodeUnits[offset++];
//             if(c<0xd800) {
//                 /* output BMP code point below 0xd800 */
//                 *target++=c;
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 byteIndex=0;
//             } else if(UCNV_TO_U_USE_FALLBACK(cnv) ? c<=0xdfff : c<=0xdbff) {
//                 /* output roundtrip or fallback surrogate pair */
//                 *target++=(UChar)(c&0xdbff);
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 byteIndex=0;
//                 if(target<targetLimit) {
//                     *target++=unicodeCodeUnits[offset];
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                 } else {
//                     /* target overflow */
//                     cnv->UCharErrorBuffer[0]=unicodeCodeUnits[offset];
//                     cnv->UCharErrorBufferLength=1;
//                     *pErrorCode=U_BUFFER_OVERFLOW_ERROR;

//                     offset=0;
//                     break;
//                 }
//             } else if(UCNV_TO_U_USE_FALLBACK(cnv) ? (c&0xfffe)==0xe000 : c==0xe000) {
//                 /* output roundtrip BMP code point above 0xd800 or fallback BMP code point */
//                 *target++=unicodeCodeUnits[offset];
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 byteIndex=0;
//             } else if(c==0xffff) {
//                 /* callback(illegal) */
//                 *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//             }
//         } else if(action==MBCS_STATE_VALID_DIRECT_20 ||
//                   (action==MBCS_STATE_FALLBACK_DIRECT_20 && UCNV_TO_U_USE_FALLBACK(cnv))
//         ) {
//             entry=MBCS_ENTRY_FINAL_VALUE(entry);
//             /* output surrogate pair */
//             *target++=(UChar)(0xd800|(UChar)(entry>>10));
//             if(offsets!=NULL) {
//                 *offsets++=sourceIndex;
//             }
//             byteIndex=0;
//             c=(UChar)(0xdc00|(UChar)(entry&0x3ff));
//             if(target<targetLimit) {
//                 *target++=c;
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//             } else {
//                 /* target overflow */
//                 cnv->UCharErrorBuffer[0]=c;
//                 cnv->UCharErrorBufferLength=1;
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;

//                 offset=0;
//                 break;
//             }
//         } else if(action==MBCS_STATE_CHANGE_ONLY) {
//             /*
//              * This serves as a state change without any output.
//              * It is useful for reading simple stateful encodings,
//              * for example using just Shift-In/Shift-Out codes.
//              * The 21 unused bits may later be used for more sophisticated
//              * state transitions.
//              */
//             if(cnv->sharedData->mbcs.dbcsOnlyState==0) {
//                 byteIndex=0;
//             } else {
//                 /* SI/SO are illegal for DBCS-only conversion */
//                 state=(uint8_t)(cnv->mode); /* restore the previous state */

//                 /* callback(illegal) */
//                 *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//             }
//         } else if(action==MBCS_STATE_FALLBACK_DIRECT_16) {
//             if(UCNV_TO_U_USE_FALLBACK(cnv)) {
//                 /* output BMP code point */
//                 *target++=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 byteIndex=0;
//             }
//         } else if(action==MBCS_STATE_UNASSIGNED) {
//             /* just fall through */
//         } else if(action==MBCS_STATE_ILLEGAL) {
//             /* callback(illegal) */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//         } else {
//             /* reserved, must never occur */
//             byteIndex=0;
//         }

//         /* end of action codes: prepare for a new character */
//         offset=0;

//         if(byteIndex==0) {
//             sourceIndex=nextSourceIndex;
//         } else if(U_FAILURE(*pErrorCode)) {
//             /* callback(illegal) */
//             if(byteIndex>1) {
//                 /*
//                  * Ticket 5691: consistent illegal sequences:
//                  * - We include at least the first byte in the illegal sequence.
//                  * - If any of the non-initial bytes could be the start of a character,
//                  *   we stop the illegal sequence before the first one of those.
//                  */
//                 UBool isDBCSOnly=(UBool)(cnv->sharedData->mbcs.dbcsOnlyState!=0);
//                 int8_t i;
//                 for(i=1;
//                     i<byteIndex && !isSingleOrLead(stateTable, state, isDBCSOnly, bytes[i]);
//                     ++i) {}
//                 if(i<byteIndex) {
//                     /* Back out some bytes. */
//                     int8_t backOutDistance=byteIndex-i;
//                     int32_t bytesFromThisBuffer=(int32_t)(source-(const uint8_t *)pArgs->source);
//                     byteIndex=i;  /* length of reported illegal byte sequence */
//                     if(backOutDistance<=bytesFromThisBuffer) {
//                         source-=backOutDistance;
//                     } else {
//                         /* Back out bytes from the previous buffer: Need to replay them. */
//                         cnv->preToULength=(int8_t)(bytesFromThisBuffer-backOutDistance);
//                         /* preToULength is negative! */
//                         uprv_memcpy(cnv->preToU, bytes+i, -cnv->preToULength);
//                         source=(const uint8_t *)pArgs->source;
//                     }
//                 }
//             }
//             break;
//         } else /* unassigned sequences indicated with byteIndex>0 */ {
//             /* try an extension mapping */
//             pArgs->source=(const char *)source;
//             byteIndex=_extToU(cnv, cnv->sharedData,
//                               byteIndex, &source, sourceLimit,
//                               &target, targetLimit,
//                               &offsets, sourceIndex,
//                               pArgs->flush,
//                               pErrorCode);
//             sourceIndex=nextSourceIndex+=(int32_t)(source-(const uint8_t *)pArgs->source);

//             if(U_FAILURE(*pErrorCode)) {
//                 /* not mappable or buffer overflow */
//                 break;
//             }
//         }
//     }

//     /* set the converter state back into UConverter */
//     cnv->toUnicodeStatus=offset;
//     cnv->mode=state;
//     cnv->toULength=byteIndex;

//     /* write back the updated pointers */
//     pArgs->source=(const char *)source;
//     pArgs->target=target;
//     pArgs->offsets=offsets;
// }

// /*
//  * This version of ucnv_MBCSGetNextUChar() is optimized for single-byte, single-state codepages.
//  * We still need a conversion loop in case we find reserved action codes, which are to be ignored.
//  */
// static UChar32
// ucnv_MBCSSingleGetNextUChar(UConverterToUnicodeArgs *pArgs,
//                         UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const int32_t (*stateTable)[256];
//     const uint8_t *source, *sourceLimit;

//     int32_t entry;
//     uint8_t action;

//     /* set up the local pointers */
//     cnv=pArgs->converter;
//     source=(const uint8_t *)pArgs->source;
//     sourceLimit=(const uint8_t *)pArgs->sourceLimit;
//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         stateTable=(const int32_t (*)[256])cnv->sharedData->mbcs.swapLFNLStateTable;
//     } else {
//         stateTable=cnv->sharedData->mbcs.stateTable;
//     }

//     /* conversion loop */
//     while(source<sourceLimit) {
//         entry=stateTable[0][*source++];
//         /* MBCS_ENTRY_IS_FINAL(entry) */

//         /* write back the updated pointer early so that we can return directly */
//         pArgs->source=(const char *)source;

//         if(MBCS_ENTRY_FINAL_IS_VALID_DIRECT_16(entry)) {
//             /* output BMP code point */
//             return (UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//         }

//         /*
//          * An if-else-if chain provides more reliable performance for
//          * the most common cases compared to a switch.
//          */
//         action=(uint8_t)(MBCS_ENTRY_FINAL_ACTION(entry));
//         if( action==MBCS_STATE_VALID_DIRECT_20 ||
//             (action==MBCS_STATE_FALLBACK_DIRECT_20 && UCNV_TO_U_USE_FALLBACK(cnv))
//         ) {
//             /* output supplementary code point */
//             return (UChar32)(MBCS_ENTRY_FINAL_VALUE(entry)+0x10000);
//         } else if(action==MBCS_STATE_FALLBACK_DIRECT_16) {
//             if(UCNV_TO_U_USE_FALLBACK(cnv)) {
//                 /* output BMP code point */
//                 return (UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//             }
//         } else if(action==MBCS_STATE_UNASSIGNED) {
//             /* just fall through */
//         } else if(action==MBCS_STATE_ILLEGAL) {
//             /* callback(illegal) */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//         } else {
//             /* reserved, must never occur */
//             continue;
//         }

//         if(U_FAILURE(*pErrorCode)) {
//             /* callback(illegal) */
//             break;
//         } else /* unassigned sequence */ {
//             /* defer to the generic implementation */
//             pArgs->source=(const char *)source-1;
//             return UCNV_GET_NEXT_UCHAR_USE_TO_U;
//         }
//     }

//     /* no output because of empty input or only state changes */
//     *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//     return 0xffff;
// }

// /*
//  * Version of _MBCSToUnicodeWithOffsets() optimized for single-character
//  * conversion without offset handling.
//  *
//  * When a character does not have a mapping to Unicode, then we return to the
//  * generic ucnv_getNextUChar() code for extension/GB 18030 and error/callback
//  * handling.
//  * We also defer to the generic code in other complicated cases and have them
//  * ultimately handled by _MBCSToUnicodeWithOffsets() itself.
//  *
//  * All normal mappings and errors are handled here.
//  */
// static UChar32 U_CALLCONV
// ucnv_MBCSGetNextUChar(UConverterToUnicodeArgs *pArgs,
//                   UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const uint8_t *source, *sourceLimit, *lastSource;

//     const int32_t (*stateTable)[256];
//     const uint16_t *unicodeCodeUnits;

//     uint32_t offset;
//     uint8_t state;

//     int32_t entry;
//     UChar32 c;
//     uint8_t action;

//     /* use optimized function if possible */
//     cnv=pArgs->converter;

//     if(cnv->preToULength>0) {
//         /* use the generic code in ucnv_getNextUChar() to continue with a partial match */
//         return UCNV_GET_NEXT_UCHAR_USE_TO_U;
//     }

//     if(cnv->sharedData->mbcs.unicodeMask&UCNV_HAS_SURROGATES) {
//         /*
//          * Using the generic ucnv_getNextUChar() code lets us deal correctly
//          * with the rare case of a codepage that maps single surrogates
//          * without adding the complexity to this already complicated function here.
//          */
//         return UCNV_GET_NEXT_UCHAR_USE_TO_U;
//     } else if(cnv->sharedData->mbcs.countStates==1) {
//         return ucnv_MBCSSingleGetNextUChar(pArgs, pErrorCode);
//     }

//     /* set up the local pointers */
//     source=lastSource=(const uint8_t *)pArgs->source;
//     sourceLimit=(const uint8_t *)pArgs->sourceLimit;

//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         stateTable=(const int32_t (*)[256])cnv->sharedData->mbcs.swapLFNLStateTable;
//     } else {
//         stateTable=cnv->sharedData->mbcs.stateTable;
//     }
//     unicodeCodeUnits=cnv->sharedData->mbcs.unicodeCodeUnits;

//     /* get the converter state from UConverter */
//     offset=cnv->toUnicodeStatus;

//     /*
//      * if we are in the SBCS state for a DBCS-only converter,
//      * then load the DBCS state from the MBCS data
//      * (dbcsOnlyState==0 if it is not a DBCS-only converter)
//      */
//     if((state=(uint8_t)(cnv->mode))==0) {
//         state=cnv->sharedData->mbcs.dbcsOnlyState;
//     }

//     /* conversion loop */
//     c=U_SENTINEL;
//     while(source<sourceLimit) {
//         entry=stateTable[state][*source++];
//         if(MBCS_ENTRY_IS_TRANSITION(entry)) {
//             state=(uint8_t)MBCS_ENTRY_TRANSITION_STATE(entry);
//             offset+=MBCS_ENTRY_TRANSITION_OFFSET(entry);

//             /* optimization for 1/2-byte input and BMP output */
//             if( source<sourceLimit &&
//                 MBCS_ENTRY_IS_FINAL(entry=stateTable[state][*source]) &&
//                 MBCS_ENTRY_FINAL_ACTION(entry)==MBCS_STATE_VALID_16 &&
//                 (c=unicodeCodeUnits[offset+MBCS_ENTRY_FINAL_VALUE_16(entry)])<0xfffe
//             ) {
//                 ++source;
//                 state=(uint8_t)MBCS_ENTRY_FINAL_STATE(entry); /* typically 0 */
//                 /* output BMP code point */
//                 break;
//             }
//         } else {
//             /* save the previous state for proper extension mapping with SI/SO-stateful converters */
//             cnv->mode=state;

//             /* set the next state early so that we can reuse the entry variable */
//             state=(uint8_t)MBCS_ENTRY_FINAL_STATE(entry); /* typically 0 */

//             /*
//              * An if-else-if chain provides more reliable performance for
//              * the most common cases compared to a switch.
//              */
//             action=(uint8_t)(MBCS_ENTRY_FINAL_ACTION(entry));
//             if(action==MBCS_STATE_VALID_DIRECT_16) {
//                 /* output BMP code point */
//                 c=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 break;
//             } else if(action==MBCS_STATE_VALID_16) {
//                 offset+=MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 c=unicodeCodeUnits[offset];
//                 if(c<0xfffe) {
//                     /* output BMP code point */
//                     break;
//                 } else if(c==0xfffe) {
//                     if(UCNV_TO_U_USE_FALLBACK(cnv) && (c=ucnv_MBCSGetFallback(&cnv->sharedData->mbcs, offset))!=0xfffe) {
//                         break;
//                     }
//                 } else {
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 }
//             } else if(action==MBCS_STATE_VALID_16_PAIR) {
//                 offset+=MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 c=unicodeCodeUnits[offset++];
//                 if(c<0xd800) {
//                     /* output BMP code point below 0xd800 */
//                     break;
//                 } else if(UCNV_TO_U_USE_FALLBACK(cnv) ? c<=0xdfff : c<=0xdbff) {
//                     /* output roundtrip or fallback supplementary code point */
//                     c=((c&0x3ff)<<10)+unicodeCodeUnits[offset]+(0x10000-0xdc00);
//                     break;
//                 } else if(UCNV_TO_U_USE_FALLBACK(cnv) ? (c&0xfffe)==0xe000 : c==0xe000) {
//                     /* output roundtrip BMP code point above 0xd800 or fallback BMP code point */
//                     c=unicodeCodeUnits[offset];
//                     break;
//                 } else if(c==0xffff) {
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 }
//             } else if(action==MBCS_STATE_VALID_DIRECT_20 ||
//                       (action==MBCS_STATE_FALLBACK_DIRECT_20 && UCNV_TO_U_USE_FALLBACK(cnv))
//             ) {
//                 /* output supplementary code point */
//                 c=(UChar32)(MBCS_ENTRY_FINAL_VALUE(entry)+0x10000);
//                 break;
//             } else if(action==MBCS_STATE_CHANGE_ONLY) {
//                 /*
//                  * This serves as a state change without any output.
//                  * It is useful for reading simple stateful encodings,
//                  * for example using just Shift-In/Shift-Out codes.
//                  * The 21 unused bits may later be used for more sophisticated
//                  * state transitions.
//                  */
//                 if(cnv->sharedData->mbcs.dbcsOnlyState!=0) {
//                     /* SI/SO are illegal for DBCS-only conversion */
//                     state=(uint8_t)(cnv->mode); /* restore the previous state */

//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                 }
//             } else if(action==MBCS_STATE_FALLBACK_DIRECT_16) {
//                 if(UCNV_TO_U_USE_FALLBACK(cnv)) {
//                     /* output BMP code point */
//                     c=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//                     break;
//                 }
//             } else if(action==MBCS_STATE_UNASSIGNED) {
//                 /* just fall through */
//             } else if(action==MBCS_STATE_ILLEGAL) {
//                 /* callback(illegal) */
//                 *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//             } else {
//                 /* reserved (must never occur), or only state change */
//                 offset=0;
//                 lastSource=source;
//                 continue;
//             }

//             /* end of action codes: prepare for a new character */
//             offset=0;

//             if(U_FAILURE(*pErrorCode)) {
//                 /* callback(illegal) */
//                 break;
//             } else /* unassigned sequence */ {
//                 /* defer to the generic implementation */
//                 cnv->toUnicodeStatus=0;
//                 cnv->mode=state;
//                 pArgs->source=(const char *)lastSource;
//                 return UCNV_GET_NEXT_UCHAR_USE_TO_U;
//             }
//         }
//     }

//     if(c<0) {
//         if(U_SUCCESS(*pErrorCode) && source==sourceLimit && lastSource<source) {
//             /* incomplete character byte sequence */
//             uint8_t *bytes=cnv->toUBytes;
//             cnv->toULength=(int8_t)(source-lastSource);
//             do {
//                 *bytes++=*lastSource++;
//             } while(lastSource<source);
//             *pErrorCode=U_TRUNCATED_CHAR_FOUND;
//         } else if(U_FAILURE(*pErrorCode)) {
//             /* callback(illegal) */
//             /*
//              * Ticket 5691: consistent illegal sequences:
//              * - We include at least the first byte in the illegal sequence.
//              * - If any of the non-initial bytes could be the start of a character,
//              *   we stop the illegal sequence before the first one of those.
//              */
//             UBool isDBCSOnly=(UBool)(cnv->sharedData->mbcs.dbcsOnlyState!=0);
//             uint8_t *bytes=cnv->toUBytes;
//             *bytes++=*lastSource++;     /* first byte */
//             if(lastSource==source) {
//                 cnv->toULength=1;
//             } else /* lastSource<source: multi-byte character */ {
//                 int8_t i;
//                 for(i=1;
//                     lastSource<source && !isSingleOrLead(stateTable, state, isDBCSOnly, *lastSource);
//                     ++i
//                 ) {
//                     *bytes++=*lastSource++;
//                 }
//                 cnv->toULength=i;
//                 source=lastSource;
//             }
//         } else {
//             /* no output because of empty input or only state changes */
//             *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
//         }
//         c=0xffff;
//     }

//     /* set the converter state back into UConverter, ready for a new character */
//     cnv->toUnicodeStatus=0;
//     cnv->mode=state;

//     /* write back the updated pointer */
//     pArgs->source=(const char *)source;
//     return c;
// }

// #if 0
// /*
//  * Code disabled 2002dec09 (ICU 2.4) because it is not currently used in ICU. markus
//  * Removal improves code coverage.
//  */
// /**
//  * This version of ucnv_MBCSSimpleGetNextUChar() is optimized for single-byte, single-state codepages.
//  * It does not handle the EBCDIC swaplfnl option (set in UConverter).
//  * It does not handle conversion extensions (_extToU()).
//  */
// U_CFUNC UChar32
// ucnv_MBCSSingleSimpleGetNextUChar(UConverterSharedData *sharedData,
//                               uint8_t b, UBool useFallback) {
//     int32_t entry;
//     uint8_t action;

//     entry=sharedData->mbcs.stateTable[0][b];
//     /* MBCS_ENTRY_IS_FINAL(entry) */

//     if(MBCS_ENTRY_FINAL_IS_VALID_DIRECT_16(entry)) {
//         /* output BMP code point */
//         return (UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//     }

//     /*
//      * An if-else-if chain provides more reliable performance for
//      * the most common cases compared to a switch.
//      */
//     action=(uint8_t)(MBCS_ENTRY_FINAL_ACTION(entry));
//     if(action==MBCS_STATE_VALID_DIRECT_20) {
//         /* output supplementary code point */
//         return 0x10000+MBCS_ENTRY_FINAL_VALUE(entry);
//     } else if(action==MBCS_STATE_FALLBACK_DIRECT_16) {
//         if(!TO_U_USE_FALLBACK(useFallback)) {
//             return 0xfffe;
//         }
//         /* output BMP code point */
//         return (UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//     } else if(action==MBCS_STATE_FALLBACK_DIRECT_20) {
//         if(!TO_U_USE_FALLBACK(useFallback)) {
//             return 0xfffe;
//         }
//         /* output supplementary code point */
//         return 0x10000+MBCS_ENTRY_FINAL_VALUE(entry);
//     } else if(action==MBCS_STATE_UNASSIGNED) {
//         return 0xfffe;
//     } else if(action==MBCS_STATE_ILLEGAL) {
//         return 0xffff;
//     } else {
//         /* reserved, must never occur */
//         return 0xffff;
//     }
// }
// #endif

// /*
//  * This is a simple version of _MBCSGetNextUChar() that is used
//  * by other converter implementations.
//  * It only returns an "assigned" result if it consumes the entire input.
//  * It does not use state from the converter, nor error codes.
//  * It does not handle the EBCDIC swaplfnl option (set in UConverter).
//  * It handles conversion extensions but not GB 18030.
//  *
//  * Return value:
//  * U+fffe   unassigned
//  * U+ffff   illegal
//  * otherwise the Unicode code point
//  */
// U_CFUNC UChar32
// ucnv_MBCSSimpleGetNextUChar(UConverterSharedData *sharedData,
//                         const char *source, int32_t length,
//                         UBool useFallback) {
//     const int32_t (*stateTable)[256];
//     const uint16_t *unicodeCodeUnits;

//     uint32_t offset;
//     uint8_t state, action;

//     UChar32 c;
//     int32_t i, entry;

//     if(length<=0) {
//         /* no input at all: "illegal" */
//         return 0xffff;
//     }

// #if 0
// /*
//  * Code disabled 2002dec09 (ICU 2.4) because it is not currently used in ICU. markus
//  * TODO In future releases, verify that this function is never called for SBCS
//  * conversions, i.e., that sharedData->mbcs.countStates==1 is still true.
//  * Removal improves code coverage.
//  */
//     /* use optimized function if possible */
//     if(sharedData->mbcs.countStates==1) {
//         if(length==1) {
//             return ucnv_MBCSSingleSimpleGetNextUChar(sharedData, (uint8_t)*source, useFallback);
//         } else {
//             return 0xffff; /* illegal: more than a single byte for an SBCS converter */
//         }
//     }
// #endif

//     /* set up the local pointers */
//     stateTable=sharedData->mbcs.stateTable;
//     unicodeCodeUnits=sharedData->mbcs.unicodeCodeUnits;

//     /* converter state */
//     offset=0;
//     state=sharedData->mbcs.dbcsOnlyState;

//     /* conversion loop */
//     for(i=0;;) {
//         entry=stateTable[state][(uint8_t)source[i++]];
//         if(MBCS_ENTRY_IS_TRANSITION(entry)) {
//             state=(uint8_t)MBCS_ENTRY_TRANSITION_STATE(entry);
//             offset+=MBCS_ENTRY_TRANSITION_OFFSET(entry);

//             if(i==length) {
//                 return 0xffff; /* truncated character */
//             }
//         } else {
//             /*
//              * An if-else-if chain provides more reliable performance for
//              * the most common cases compared to a switch.
//              */
//             action=(uint8_t)(MBCS_ENTRY_FINAL_ACTION(entry));
//             if(action==MBCS_STATE_VALID_16) {
//                 offset+=MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 c=unicodeCodeUnits[offset];
//                 if(c!=0xfffe) {
//                     /* done */
//                 } else if(UCNV_TO_U_USE_FALLBACK(cnv)) {
//                     c=ucnv_MBCSGetFallback(&sharedData->mbcs, offset);
//                 /* else done with 0xfffe */
//                 }
//                 break;
//             } else if(action==MBCS_STATE_VALID_DIRECT_16) {
//                 /* output BMP code point */
//                 c=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 break;
//             } else if(action==MBCS_STATE_VALID_16_PAIR) {
//                 offset+=MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 c=unicodeCodeUnits[offset++];
//                 if(c<0xd800) {
//                     /* output BMP code point below 0xd800 */
//                 } else if(UCNV_TO_U_USE_FALLBACK(cnv) ? c<=0xdfff : c<=0xdbff) {
//                     /* output roundtrip or fallback supplementary code point */
//                     c=(UChar32)(((c&0x3ff)<<10)+unicodeCodeUnits[offset]+(0x10000-0xdc00));
//                 } else if(UCNV_TO_U_USE_FALLBACK(cnv) ? (c&0xfffe)==0xe000 : c==0xe000) {
//                     /* output roundtrip BMP code point above 0xd800 or fallback BMP code point */
//                     c=unicodeCodeUnits[offset];
//                 } else if(c==0xffff) {
//                     return 0xffff;
//                 } else {
//                     c=0xfffe;
//                 }
//                 break;
//             } else if(action==MBCS_STATE_VALID_DIRECT_20) {
//                 /* output supplementary code point */
//                 c=0x10000+MBCS_ENTRY_FINAL_VALUE(entry);
//                 break;
//             } else if(action==MBCS_STATE_FALLBACK_DIRECT_16) {
//                 if(!TO_U_USE_FALLBACK(useFallback)) {
//                     c=0xfffe;
//                     break;
//                 }
//                 /* output BMP code point */
//                 c=(UChar)MBCS_ENTRY_FINAL_VALUE_16(entry);
//                 break;
//             } else if(action==MBCS_STATE_FALLBACK_DIRECT_20) {
//                 if(!TO_U_USE_FALLBACK(useFallback)) {
//                     c=0xfffe;
//                     break;
//                 }
//                 /* output supplementary code point */
//                 c=0x10000+MBCS_ENTRY_FINAL_VALUE(entry);
//                 break;
//             } else if(action==MBCS_STATE_UNASSIGNED) {
//                 c=0xfffe;
//                 break;
//             }

//             /*
//              * forbid MBCS_STATE_CHANGE_ONLY for this function,
//              * and MBCS_STATE_ILLEGAL and reserved action codes
//              */
//             return 0xffff;
//         }
//     }

//     if(i!=length) {
//         /* illegal for this function: not all input consumed */
//         return 0xffff;
//     }

//     if(c==0xfffe) {
//         /* try an extension mapping */
//         const int32_t *cx=sharedData->mbcs.extIndexes;
//         if(cx!=NULL) {
//             return ucnv_extSimpleMatchToU(cx, source, length, useFallback);
//         }
//     }

//     return c;
// }

// /* MBCS-from-Unicode conversion functions ----------------------------------- */

// /* This version of ucnv_MBCSFromUnicodeWithOffsets() is optimized for double-byte codepages. */
// static void
// ucnv_MBCSDoubleFromUnicodeWithOffsets(UConverterFromUnicodeArgs *pArgs,
//                                   UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const UChar *source, *sourceLimit;
//     uint8_t *target;
//     int32_t targetCapacity;
//     int32_t *offsets;

//     const uint16_t *table;
//     const uint16_t *mbcsIndex;
//     const uint8_t *bytes;

//     UChar32 c;

//     int32_t sourceIndex, nextSourceIndex;

//     uint32_t stage2Entry;
//     uint32_t asciiRoundtrips;
//     uint32_t value;
//     uint8_t unicodeMask;

//     /* use optimized function if possible */
//     cnv=pArgs->converter;
//     unicodeMask=cnv->sharedData->mbcs.unicodeMask;

//     /* set up the local pointers */
//     source=pArgs->source;
//     sourceLimit=pArgs->sourceLimit;
//     target=(uint8_t *)pArgs->target;
//     targetCapacity=(int32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;

//     table=cnv->sharedData->mbcs.fromUnicodeTable;
//     mbcsIndex=cnv->sharedData->mbcs.mbcsIndex;
//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         bytes=cnv->sharedData->mbcs.swapLFNLFromUnicodeBytes;
//     } else {
//         bytes=cnv->sharedData->mbcs.fromUnicodeBytes;
//     }
//     asciiRoundtrips=cnv->sharedData->mbcs.asciiRoundtrips;

//     /* get the converter state from UConverter */
//     c=cnv->fromUChar32;

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     sourceIndex= c==0 ? 0 : -1;
//     nextSourceIndex=0;

//     /* conversion loop */
//     if(c!=0 && targetCapacity>0) {
//         goto getTrail;
//     }

//     while(source<sourceLimit) {
//         /*
//          * This following test is to see if available input would overflow the output.
//          * It does not catch output of more than one byte that
//          * overflows as a result of a multi-byte character or callback output
//          * from the last source character.
//          * Therefore, those situations also test for overflows and will
//          * then break the loop, too.
//          */
//         if(targetCapacity>0) {
//             /*
//              * Get a correct Unicode code point:
//              * a single UChar for a BMP code point or
//              * a matched surrogate pair for a "supplementary code point".
//              */
//             c=*source++;
//             ++nextSourceIndex;
//             if(c<=0x7f && IS_ASCII_ROUNDTRIP(c, asciiRoundtrips)) {
//                 *target++=(uint8_t)c;
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                     sourceIndex=nextSourceIndex;
//                 }
//                 --targetCapacity;
//                 c=0;
//                 continue;
//             }
//             /*
//              * utf8Friendly table: Test for <=0xd7ff rather than <=MBCS_FAST_MAX
//              * to avoid dealing with surrogates.
//              * MBCS_FAST_MAX must be >=0xd7ff.
//              */
//             if(c<=0xd7ff) {
//                 value=DBCS_RESULT_FROM_MOST_BMP(mbcsIndex, (const uint16_t *)bytes, c);
//                 /* There are only roundtrips (!=0) and no-mapping (==0) entries. */
//                 if(value==0) {
//                     goto unassigned;
//                 }
//                 /* output the value */
//             } else {
//                 /*
//                  * This also tests if the codepage maps single surrogates.
//                  * If it does, then surrogates are not paired but mapped separately.
//                  * Note that in this case unmatched surrogates are not detected.
//                  */
//                 if(U16_IS_SURROGATE(c) && !(unicodeMask&UCNV_HAS_SURROGATES)) {
//                     if(U16_IS_SURROGATE_LEAD(c)) {
// getTrail:
//                         if(source<sourceLimit) {
//                             /* test the following code unit */
//                             UChar trail=*source;
//                             if(U16_IS_TRAIL(trail)) {
//                                 ++source;
//                                 ++nextSourceIndex;
//                                 c=U16_GET_SUPPLEMENTARY(c, trail);
//                                 if(!(unicodeMask&UCNV_HAS_SUPPLEMENTARY)) {
//                                     /* BMP-only codepages are stored without stage 1 entries for supplementary code points */
//                                     /* callback(unassigned) */
//                                     goto unassigned;
//                                 }
//                                 /* convert this supplementary code point */
//                                 /* exit this condition tree */
//                             } else {
//                                 /* this is an unmatched lead code unit (1st surrogate) */
//                                 /* callback(illegal) */
//                                 *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                                 break;
//                             }
//                         } else {
//                             /* no more input */
//                             break;
//                         }
//                     } else {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                         break;
//                     }
//                 }

//                 /* convert the Unicode code point in c into codepage bytes */
//                 stage2Entry=MBCS_STAGE_2_FROM_U(table, c);

//                 /* get the bytes and the length for the output */
//                 /* MBCS_OUTPUT_2 */
//                 value=MBCS_VALUE_2_FROM_STAGE_2(bytes, stage2Entry, c);

//                 /* is this code point assigned, or do we use fallbacks? */
//                 if(!(MBCS_FROM_U_IS_ROUNDTRIP(stage2Entry, c) ||
//                      (UCNV_FROM_U_USE_FALLBACK(cnv, c) && value!=0))
//                 ) {
//                     /*
//                      * We allow a 0 byte output if the "assigned" bit is set for this entry.
//                      * There is no way with this data structure for fallback output
//                      * to be a zero byte.
//                      */

// unassigned:
//                     /* try an extension mapping */
//                     pArgs->source=source;
//                     c=_extFromU(cnv, cnv->sharedData,
//                                 c, &source, sourceLimit,
//                                 &target, target+targetCapacity,
//                                 &offsets, sourceIndex,
//                                 pArgs->flush,
//                                 pErrorCode);
//                     nextSourceIndex+=(int32_t)(source-pArgs->source);

//                     if(U_FAILURE(*pErrorCode)) {
//                         /* not mappable or buffer overflow */
//                         break;
//                     } else {
//                         /* a mapping was written to the target, continue */

//                         /* recalculate the targetCapacity after an extension mapping */
//                         targetCapacity=(int32_t)(pArgs->targetLimit-(char *)target);

//                         /* normal end of conversion: prepare for a new character */
//                         sourceIndex=nextSourceIndex;
//                         continue;
//                     }
//                 }
//             }

//             /* write the output character bytes from value and length */
//             /* from the first if in the loop we know that targetCapacity>0 */
//             if(value<=0xff) {
//                 /* this is easy because we know that there is enough space */
//                 *target++=(uint8_t)value;
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 --targetCapacity;
//             } else /* length==2 */ {
//                 *target++=(uint8_t)(value>>8);
//                 if(2<=targetCapacity) {
//                     *target++=(uint8_t)value;
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                         *offsets++=sourceIndex;
//                     }
//                     targetCapacity-=2;
//                 } else {
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     cnv->charErrorBuffer[0]=(char)value;
//                     cnv->charErrorBufferLength=1;

//                     /* target overflow */
//                     targetCapacity=0;
//                     *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                     c=0;
//                     break;
//                 }
//             }

//             /* normal end of conversion: prepare for a new character */
//             c=0;
//             sourceIndex=nextSourceIndex;
//             continue;
//         } else {
//             /* target is full */
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }
//     }

//     /* set the converter state back into UConverter */
//     cnv->fromUChar32=c;

//     /* write back the updated pointers */
//     pArgs->source=source;
//     pArgs->target=(char *)target;
//     pArgs->offsets=offsets;
// }

// /* This version of ucnv_MBCSFromUnicodeWithOffsets() is optimized for single-byte codepages. */
// static void
// ucnv_MBCSSingleFromUnicodeWithOffsets(UConverterFromUnicodeArgs *pArgs,
//                                   UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const UChar *source, *sourceLimit;
//     uint8_t *target;
//     int32_t targetCapacity;
//     int32_t *offsets;

//     const uint16_t *table;
//     const uint16_t *results;

//     UChar32 c;

//     int32_t sourceIndex, nextSourceIndex;

//     uint16_t value, minValue;
//     UBool hasSupplementary;

//     /* set up the local pointers */
//     cnv=pArgs->converter;
//     source=pArgs->source;
//     sourceLimit=pArgs->sourceLimit;
//     target=(uint8_t *)pArgs->target;
//     targetCapacity=(int32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;

//     table=cnv->sharedData->mbcs.fromUnicodeTable;
//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         results=(uint16_t *)cnv->sharedData->mbcs.swapLFNLFromUnicodeBytes;
//     } else {
//         results=(uint16_t *)cnv->sharedData->mbcs.fromUnicodeBytes;
//     }

//     if(cnv->useFallback) {
//         /* use all roundtrip and fallback results */
//         minValue=0x800;
//     } else {
//         /* use only roundtrips and fallbacks from private-use characters */
//         minValue=0xc00;
//     }
//     hasSupplementary=(UBool)(cnv->sharedData->mbcs.unicodeMask&UCNV_HAS_SUPPLEMENTARY);

//     /* get the converter state from UConverter */
//     c=cnv->fromUChar32;

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     sourceIndex= c==0 ? 0 : -1;
//     nextSourceIndex=0;

//     /* conversion loop */
//     if(c!=0 && targetCapacity>0) {
//         goto getTrail;
//     }

//     while(source<sourceLimit) {
//         /*
//          * This following test is to see if available input would overflow the output.
//          * It does not catch output of more than one byte that
//          * overflows as a result of a multi-byte character or callback output
//          * from the last source character.
//          * Therefore, those situations also test for overflows and will
//          * then break the loop, too.
//          */
//         if(targetCapacity>0) {
//             /*
//              * Get a correct Unicode code point:
//              * a single UChar for a BMP code point or
//              * a matched surrogate pair for a "supplementary code point".
//              */
//             c=*source++;
//             ++nextSourceIndex;
//             if(U16_IS_SURROGATE(c)) {
//                 if(U16_IS_SURROGATE_LEAD(c)) {
// getTrail:
//                     if(source<sourceLimit) {
//                         /* test the following code unit */
//                         UChar trail=*source;
//                         if(U16_IS_TRAIL(trail)) {
//                             ++source;
//                             ++nextSourceIndex;
//                             c=U16_GET_SUPPLEMENTARY(c, trail);
//                             if(!hasSupplementary) {
//                                 /* BMP-only codepages are stored without stage 1 entries for supplementary code points */
//                                 /* callback(unassigned) */
//                                 goto unassigned;
//                             }
//                             /* convert this supplementary code point */
//                             /* exit this condition tree */
//                         } else {
//                             /* this is an unmatched lead code unit (1st surrogate) */
//                             /* callback(illegal) */
//                             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                             break;
//                         }
//                     } else {
//                         /* no more input */
//                         break;
//                     }
//                 } else {
//                     /* this is an unmatched trail code unit (2nd surrogate) */
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                     break;
//                 }
//             }

//             /* convert the Unicode code point in c into codepage bytes */
//             value=MBCS_SINGLE_RESULT_FROM_U(table, results, c);

//             /* is this code point assigned, or do we use fallbacks? */
//             if(value>=minValue) {
//                 /* assigned, write the output character bytes from value and length */
//                 /* length==1 */
//                 /* this is easy because we know that there is enough space */
//                 *target++=(uint8_t)value;
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 --targetCapacity;

//                 /* normal end of conversion: prepare for a new character */
//                 c=0;
//                 sourceIndex=nextSourceIndex;
//             } else { /* unassigned */
// unassigned:
//                 /* try an extension mapping */
//                 pArgs->source=source;
//                 c=_extFromU(cnv, cnv->sharedData,
//                             c, &source, sourceLimit,
//                             &target, target+targetCapacity,
//                             &offsets, sourceIndex,
//                             pArgs->flush,
//                             pErrorCode);
//                 nextSourceIndex+=(int32_t)(source-pArgs->source);

//                 if(U_FAILURE(*pErrorCode)) {
//                     /* not mappable or buffer overflow */
//                     break;
//                 } else {
//                     /* a mapping was written to the target, continue */

//                     /* recalculate the targetCapacity after an extension mapping */
//                     targetCapacity=(int32_t)(pArgs->targetLimit-(char *)target);

//                     /* normal end of conversion: prepare for a new character */
//                     sourceIndex=nextSourceIndex;
//                 }
//             }
//         } else {
//             /* target is full */
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }
//     }

//     /* set the converter state back into UConverter */
//     cnv->fromUChar32=c;

//     /* write back the updated pointers */
//     pArgs->source=source;
//     pArgs->target=(char *)target;
//     pArgs->offsets=offsets;
// }

// /*
//  * This version of ucnv_MBCSFromUnicode() is optimized for single-byte codepages
//  * that map only to and from the BMP.
//  * In addition to single-byte/state optimizations, the offset calculations
//  * become much easier.
//  * It would be possible to use the sbcsIndex for UTF-8-friendly tables,
//  * but measurements have shown that this diminishes performance
//  * in more cases than it improves it.
//  * See SVN revision 21013 (2007-feb-06) for the last version with #if switches
//  * for various MBCS and SBCS optimizations.
//  */
// static void
// ucnv_MBCSSingleFromBMPWithOffsets(UConverterFromUnicodeArgs *pArgs,
//                               UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const UChar *source, *sourceLimit, *lastSource;
//     uint8_t *target;
//     int32_t targetCapacity, length;
//     int32_t *offsets;

//     const uint16_t *table;
//     const uint16_t *results;

//     UChar32 c;

//     int32_t sourceIndex;

//     uint32_t asciiRoundtrips;
//     uint16_t value, minValue;

//     /* set up the local pointers */
//     cnv=pArgs->converter;
//     source=pArgs->source;
//     sourceLimit=pArgs->sourceLimit;
//     target=(uint8_t *)pArgs->target;
//     targetCapacity=(int32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;

//     table=cnv->sharedData->mbcs.fromUnicodeTable;
//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         results=(uint16_t *)cnv->sharedData->mbcs.swapLFNLFromUnicodeBytes;
//     } else {
//         results=(uint16_t *)cnv->sharedData->mbcs.fromUnicodeBytes;
//     }
//     asciiRoundtrips=cnv->sharedData->mbcs.asciiRoundtrips;

//     if(cnv->useFallback) {
//         /* use all roundtrip and fallback results */
//         minValue=0x800;
//     } else {
//         /* use only roundtrips and fallbacks from private-use characters */
//         minValue=0xc00;
//     }

//     /* get the converter state from UConverter */
//     c=cnv->fromUChar32;

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     sourceIndex= c==0 ? 0 : -1;
//     lastSource=source;

//     /*
//      * since the conversion here is 1:1 UChar:uint8_t, we need only one counter
//      * for the minimum of the sourceLength and targetCapacity
//      */
//     length=(int32_t)(sourceLimit-source);
//     if(length<targetCapacity) {
//         targetCapacity=length;
//     }

//     /* conversion loop */
//     if(c!=0 && targetCapacity>0) {
//         goto getTrail;
//     }

// #if MBCS_UNROLL_SINGLE_FROM_BMP
//     /* unrolling makes it slower on Pentium III/Windows 2000?! */
//     /* unroll the loop with the most common case */
// unrolled:
//     if(targetCapacity>=4) {
//         int32_t count, loops;
//         uint16_t andedValues;

//         loops=count=targetCapacity>>2;
//         do {
//             c=*source++;
//             andedValues=value=MBCS_SINGLE_RESULT_FROM_U(table, results, c);
//             *target++=(uint8_t)value;
//             c=*source++;
//             andedValues&=value=MBCS_SINGLE_RESULT_FROM_U(table, results, c);
//             *target++=(uint8_t)value;
//             c=*source++;
//             andedValues&=value=MBCS_SINGLE_RESULT_FROM_U(table, results, c);
//             *target++=(uint8_t)value;
//             c=*source++;
//             andedValues&=value=MBCS_SINGLE_RESULT_FROM_U(table, results, c);
//             *target++=(uint8_t)value;

//             /* were all 4 entries really valid? */
//             if(andedValues<minValue) {
//                 /* no, return to the first of these 4 */
//                 source-=4;
//                 target-=4;
//                 break;
//             }
//         } while(--count>0);
//         count=loops-count;
//         targetCapacity-=4*count;

//         if(offsets!=NULL) {
//             lastSource+=4*count;
//             while(count>0) {
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 *offsets++=sourceIndex++;
//                 --count;
//             }
//         }

//         c=0;
//     }
// #endif

//     while(targetCapacity>0) {
//         /*
//          * Get a correct Unicode code point:
//          * a single UChar for a BMP code point or
//          * a matched surrogate pair for a "supplementary code point".
//          */
//         c=*source++;
//         /*
//          * Do not immediately check for single surrogates:
//          * Assume that they are unassigned and check for them in that case.
//          * This speeds up the conversion of assigned characters.
//          */
//         /* convert the Unicode code point in c into codepage bytes */
//         if(c<=0x7f && IS_ASCII_ROUNDTRIP(c, asciiRoundtrips)) {
//             *target++=(uint8_t)c;
//             --targetCapacity;
//             c=0;
//             continue;
//         }
//         value=MBCS_SINGLE_RESULT_FROM_U(table, results, c);
//         /* is this code point assigned, or do we use fallbacks? */
//         if(value>=minValue) {
//             /* assigned, write the output character bytes from value and length */
//             /* length==1 */
//             /* this is easy because we know that there is enough space */
//             *target++=(uint8_t)value;
//             --targetCapacity;

//             /* normal end of conversion: prepare for a new character */
//             c=0;
//             continue;
//         } else if(!U16_IS_SURROGATE(c)) {
//             /* normal, unassigned BMP character */
//         } else if(U16_IS_SURROGATE_LEAD(c)) {
// getTrail:
//             if(source<sourceLimit) {
//                 /* test the following code unit */
//                 UChar trail=*source;
//                 if(U16_IS_TRAIL(trail)) {
//                     ++source;
//                     c=U16_GET_SUPPLEMENTARY(c, trail);
//                     /* this codepage does not map supplementary code points */
//                     /* callback(unassigned) */
//                 } else {
//                     /* this is an unmatched lead code unit (1st surrogate) */
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                     break;
//                 }
//             } else {
//                 /* no more input */
//                 if (pArgs->flush) {
//                     *pErrorCode=U_TRUNCATED_CHAR_FOUND;
//                 }
//                 break;
//             }
//         } else {
//             /* this is an unmatched trail code unit (2nd surrogate) */
//             /* callback(illegal) */
//             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//             break;
//         }

//         /* c does not have a mapping */

//         /* get the number of code units for c to correctly advance sourceIndex */
//         length=U16_LENGTH(c);

//         /* set offsets since the start or the last extension */
//         if(offsets!=NULL) {
//             int32_t count=(int32_t)(source-lastSource);

//             /* do not set the offset for this character */
//             count-=length;

//             while(count>0) {
//                 *offsets++=sourceIndex++;
//                 --count;
//             }
//             /* offsets and sourceIndex are now set for the current character */
//         }

//         /* try an extension mapping */
//         lastSource=source;
//         c=_extFromU(cnv, cnv->sharedData,
//                     c, &source, sourceLimit,
//                     &target, (const uint8_t *)(pArgs->targetLimit),
//                     &offsets, sourceIndex,
//                     pArgs->flush,
//                     pErrorCode);
//         sourceIndex+=length+(int32_t)(source-lastSource);
//         lastSource=source;

//         if(U_FAILURE(*pErrorCode)) {
//             /* not mappable or buffer overflow */
//             break;
//         } else {
//             /* a mapping was written to the target, continue */

//             /* recalculate the targetCapacity after an extension mapping */
//             targetCapacity=(int32_t)(pArgs->targetLimit-(char *)target);
//             length=(int32_t)(sourceLimit-source);
//             if(length<targetCapacity) {
//                 targetCapacity=length;
//             }
//         }

// #if MBCS_UNROLL_SINGLE_FROM_BMP
//         /* unrolling makes it slower on Pentium III/Windows 2000?! */
//         goto unrolled;
// #endif
//     }

//     if(U_SUCCESS(*pErrorCode) && source<sourceLimit && target>=(uint8_t *)pArgs->targetLimit) {
//         /* target is full */
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//     }

//     /* set offsets since the start or the last callback */
//     if(offsets!=NULL) {
//         size_t count=source-lastSource;
//         if (count > 0 && *pErrorCode == U_TRUNCATED_CHAR_FOUND) {
//             /*
//             Caller gave us a partial supplementary character,
//             which this function couldn't convert in any case.
//             The callback will handle the offset.
//             */
//             count--;
//         }
//         while(count>0) {
//             *offsets++=sourceIndex++;
//             --count;
//         }
//     }

//     /* set the converter state back into UConverter */
//     cnv->fromUChar32=c;

//     /* write back the updated pointers */
//     pArgs->source=source;
//     pArgs->target=(char *)target;
//     pArgs->offsets=offsets;
// }

// U_CFUNC void
// ucnv_MBCSFromUnicodeWithOffsets(UConverterFromUnicodeArgs *pArgs,
//                             UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     const UChar *source, *sourceLimit;
//     uint8_t *target;
//     int32_t targetCapacity;
//     int32_t *offsets;

//     const uint16_t *table;
//     const uint16_t *mbcsIndex;
//     const uint8_t *p, *bytes;
//     uint8_t outputType;

//     UChar32 c;

//     int32_t prevSourceIndex, sourceIndex, nextSourceIndex;

//     uint32_t stage2Entry;
//     uint32_t asciiRoundtrips;
//     uint32_t value;
//     /* Shift-In and Shift-Out byte sequences differ by encoding scheme. */
//     uint8_t siBytes[2] = {0, 0};
//     uint8_t soBytes[2] = {0, 0};
//     uint8_t siLength, soLength;
//     int32_t length = 0, prevLength;
//     uint8_t unicodeMask;

//     cnv=pArgs->converter;

//     if(cnv->preFromUFirstCP>=0) {
//         /*
//          * pass sourceIndex=-1 because we continue from an earlier buffer
//          * in the future, this may change with continuous offsets
//          */
//         ucnv_extContinueMatchFromU(cnv, pArgs, -1, pErrorCode);

//         if(U_FAILURE(*pErrorCode) || cnv->preFromULength<0) {
//             return;
//         }
//     }

//     /* use optimized function if possible */
//     outputType=cnv->sharedData->mbcs.outputType;
//     unicodeMask=cnv->sharedData->mbcs.unicodeMask;
//     if(outputType==MBCS_OUTPUT_1 && !(unicodeMask&UCNV_HAS_SURROGATES)) {
//         if(!(unicodeMask&UCNV_HAS_SUPPLEMENTARY)) {
//             ucnv_MBCSSingleFromBMPWithOffsets(pArgs, pErrorCode);
//         } else {
//             ucnv_MBCSSingleFromUnicodeWithOffsets(pArgs, pErrorCode);
//         }
//         return;
//     } else if(outputType==MBCS_OUTPUT_2 && cnv->sharedData->mbcs.utf8Friendly) {
//         ucnv_MBCSDoubleFromUnicodeWithOffsets(pArgs, pErrorCode);
//         return;
//     }

//     /* set up the local pointers */
//     source=pArgs->source;
//     sourceLimit=pArgs->sourceLimit;
//     target=(uint8_t *)pArgs->target;
//     targetCapacity=(int32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;

//     table=cnv->sharedData->mbcs.fromUnicodeTable;
//     if(cnv->sharedData->mbcs.utf8Friendly) {
//         mbcsIndex=cnv->sharedData->mbcs.mbcsIndex;
//     } else {
//         mbcsIndex=NULL;
//     }
//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         bytes=cnv->sharedData->mbcs.swapLFNLFromUnicodeBytes;
//     } else {
//         bytes=cnv->sharedData->mbcs.fromUnicodeBytes;
//     }
//     asciiRoundtrips=cnv->sharedData->mbcs.asciiRoundtrips;

//     /* get the converter state from UConverter */
//     c=cnv->fromUChar32;

//     if(outputType==MBCS_OUTPUT_2_SISO) {
//         prevLength=cnv->fromUnicodeStatus;
//         if(prevLength==0) {
//             /* set the real value */
//             prevLength=1;
//         }
//     } else {
//         /* prevent fromUnicodeStatus from being set to something non-0 */
//         prevLength=0;
//     }

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     prevSourceIndex=-1;
//     sourceIndex= c==0 ? 0 : -1;
//     nextSourceIndex=0;

//     /* Get the SI/SO character for the converter */
//     siLength = static_cast<uint8_t>(getSISOBytes(SI, cnv->options, siBytes));
//     soLength = static_cast<uint8_t>(getSISOBytes(SO, cnv->options, soBytes));

//     /* conversion loop */
//     /*
//      * This is another piece of ugly code:
//      * A goto into the loop if the converter state contains a first surrogate
//      * from the previous function call.
//      * It saves me to check in each loop iteration a check of if(c==0)
//      * and duplicating the trail-surrogate-handling code in the else
//      * branch of that check.
//      * I could not find any other way to get around this other than
//      * using a function call for the conversion and callback, which would
//      * be even more inefficient.
//      *
//      * Markus Scherer 2000-jul-19
//      */
//     if(c!=0 && targetCapacity>0) {
//         goto getTrail;
//     }

//     while(source<sourceLimit) {
//         /*
//          * This following test is to see if available input would overflow the output.
//          * It does not catch output of more than one byte that
//          * overflows as a result of a multi-byte character or callback output
//          * from the last source character.
//          * Therefore, those situations also test for overflows and will
//          * then break the loop, too.
//          */
//         if(targetCapacity>0) {
//             /*
//              * Get a correct Unicode code point:
//              * a single UChar for a BMP code point or
//              * a matched surrogate pair for a "supplementary code point".
//              */
//             c=*source++;
//             ++nextSourceIndex;
//             if(c<=0x7f && IS_ASCII_ROUNDTRIP(c, asciiRoundtrips)) {
//                 *target++=(uint8_t)c;
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                     prevSourceIndex=sourceIndex;
//                     sourceIndex=nextSourceIndex;
//                 }
//                 --targetCapacity;
//                 c=0;
//                 continue;
//             }
//             /*
//              * utf8Friendly table: Test for <=0xd7ff rather than <=MBCS_FAST_MAX
//              * to avoid dealing with surrogates.
//              * MBCS_FAST_MAX must be >=0xd7ff.
//              */
//             if(c<=0xd7ff && mbcsIndex!=NULL) {
//                 value=mbcsIndex[c>>6];

//                 /* get the bytes and the length for the output (copied from below and adapted for utf8Friendly data) */
//                 /* There are only roundtrips (!=0) and no-mapping (==0) entries. */
//                 switch(outputType) {
//                 case MBCS_OUTPUT_2:
//                     value=((const uint16_t *)bytes)[value +(c&0x3f)];
//                     if(value<=0xff) {
//                         if(value==0) {
//                             goto unassigned;
//                         } else {
//                             length=1;
//                         }
//                     } else {
//                         length=2;
//                     }
//                     break;
//                 case MBCS_OUTPUT_2_SISO:
//                     /* 1/2-byte stateful with Shift-In/Shift-Out */
//                     /*
//                      * Save the old state in the converter object
//                      * right here, then change the local prevLength state variable if necessary.
//                      * Then, if this character turns out to be unassigned or a fallback that
//                      * is not taken, the callback code must not save the new state in the converter
//                      * because the new state is for a character that is not output.
//                      * However, the callback must still restore the state from the converter
//                      * in case the callback function changed it for its output.
//                      */
//                     cnv->fromUnicodeStatus=prevLength; /* save the old state */
//                     value=((const uint16_t *)bytes)[value +(c&0x3f)];
//                     if(value<=0xff) {
//                         if(value==0) {
//                             goto unassigned;
//                         } else if(prevLength<=1) {
//                             length=1;
//                         } else {
//                             /* change from double-byte mode to single-byte */
//                             if (siLength == 1) {
//                                 value|=(uint32_t)siBytes[0]<<8;
//                                 length = 2;
//                             } else if (siLength == 2) {
//                                 value|=(uint32_t)siBytes[1]<<8;
//                                 value|=(uint32_t)siBytes[0]<<16;
//                                 length = 3;
//                             }
//                             prevLength=1;
//                         }
//                     } else {
//                         if(prevLength==2) {
//                             length=2;
//                         } else {
//                             /* change from single-byte mode to double-byte */
//                             if (soLength == 1) {
//                                 value|=(uint32_t)soBytes[0]<<16;
//                                 length = 3;
//                             } else if (soLength == 2) {
//                                 value|=(uint32_t)soBytes[1]<<16;
//                                 value|=(uint32_t)soBytes[0]<<24;
//                                 length = 4;
//                             }
//                             prevLength=2;
//                         }
//                     }
//                     break;
//                 case MBCS_OUTPUT_DBCS_ONLY:
//                     /* table with single-byte results, but only DBCS mappings used */
//                     value=((const uint16_t *)bytes)[value +(c&0x3f)];
//                     if(value<=0xff) {
//                         /* no mapping or SBCS result, not taken for DBCS-only */
//                         goto unassigned;
//                     } else {
//                         length=2;
//                     }
//                     break;
//                 case MBCS_OUTPUT_3:
//                     p=bytes+(value+(c&0x3f))*3;
//                     value=((uint32_t)*p<<16)|((uint32_t)p[1]<<8)|p[2];
//                     if(value<=0xff) {
//                         if(value==0) {
//                             goto unassigned;
//                         } else {
//                             length=1;
//                         }
//                     } else if(value<=0xffff) {
//                         length=2;
//                     } else {
//                         length=3;
//                     }
//                     break;
//                 case MBCS_OUTPUT_4:
//                     value=((const uint32_t *)bytes)[value +(c&0x3f)];
//                     if(value<=0xff) {
//                         if(value==0) {
//                             goto unassigned;
//                         } else {
//                             length=1;
//                         }
//                     } else if(value<=0xffff) {
//                         length=2;
//                     } else if(value<=0xffffff) {
//                         length=3;
//                     } else {
//                         length=4;
//                     }
//                     break;
//                 case MBCS_OUTPUT_3_EUC:
//                     value=((const uint16_t *)bytes)[value +(c&0x3f)];
//                     /* EUC 16-bit fixed-length representation */
//                     if(value<=0xff) {
//                         if(value==0) {
//                             goto unassigned;
//                         } else {
//                             length=1;
//                         }
//                     } else if((value&0x8000)==0) {
//                         value|=0x8e8000;
//                         length=3;
//                     } else if((value&0x80)==0) {
//                         value|=0x8f0080;
//                         length=3;
//                     } else {
//                         length=2;
//                     }
//                     break;
//                 case MBCS_OUTPUT_4_EUC:
//                     p=bytes+(value+(c&0x3f))*3;
//                     value=((uint32_t)*p<<16)|((uint32_t)p[1]<<8)|p[2];
//                     /* EUC 16-bit fixed-length representation applied to the first two bytes */
//                     if(value<=0xff) {
//                         if(value==0) {
//                             goto unassigned;
//                         } else {
//                             length=1;
//                         }
//                     } else if(value<=0xffff) {
//                         length=2;
//                     } else if((value&0x800000)==0) {
//                         value|=0x8e800000;
//                         length=4;
//                     } else if((value&0x8000)==0) {
//                         value|=0x8f008000;
//                         length=4;
//                     } else {
//                         length=3;
//                     }
//                     break;
//                 default:
//                     /* must not occur */
//                     /*
//                      * To avoid compiler warnings that value & length may be
//                      * used without having been initialized, we set them here.
//                      * In reality, this is unreachable code.
//                      * Not having a default branch also causes warnings with
//                      * some compilers.
//                      */
//                     value=0;
//                     length=0;
//                     break;
//                 }
//                 /* output the value */
//             } else {
//                 /*
//                  * This also tests if the codepage maps single surrogates.
//                  * If it does, then surrogates are not paired but mapped separately.
//                  * Note that in this case unmatched surrogates are not detected.
//                  */
//                 if(U16_IS_SURROGATE(c) && !(unicodeMask&UCNV_HAS_SURROGATES)) {
//                     if(U16_IS_SURROGATE_LEAD(c)) {
// getTrail:
//                         if(source<sourceLimit) {
//                             /* test the following code unit */
//                             UChar trail=*source;
//                             if(U16_IS_TRAIL(trail)) {
//                                 ++source;
//                                 ++nextSourceIndex;
//                                 c=U16_GET_SUPPLEMENTARY(c, trail);
//                                 if(!(unicodeMask&UCNV_HAS_SUPPLEMENTARY)) {
//                                     /* BMP-only codepages are stored without stage 1 entries for supplementary code points */
//                                     cnv->fromUnicodeStatus=prevLength; /* save the old state */
//                                     /* callback(unassigned) */
//                                     goto unassigned;
//                                 }
//                                 /* convert this supplementary code point */
//                                 /* exit this condition tree */
//                             } else {
//                                 /* this is an unmatched lead code unit (1st surrogate) */
//                                 /* callback(illegal) */
//                                 *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                                 break;
//                             }
//                         } else {
//                             /* no more input */
//                             break;
//                         }
//                     } else {
//                         /* this is an unmatched trail code unit (2nd surrogate) */
//                         /* callback(illegal) */
//                         *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                         break;
//                     }
//                 }

//                 /* convert the Unicode code point in c into codepage bytes */

//                 /*
//                  * The basic lookup is a triple-stage compact array (trie) lookup.
//                  * For details see the beginning of this file.
//                  *
//                  * Single-byte codepages are handled with a different data structure
//                  * by _MBCSSingle... functions.
//                  *
//                  * The result consists of a 32-bit value from stage 2 and
//                  * a pointer to as many bytes as are stored per character.
//                  * The pointer points to the character's bytes in stage 3.
//                  * Bits 15..0 of the stage 2 entry contain the stage 3 index
//                  * for that pointer, while bits 31..16 are flags for which of
//                  * the 16 characters in the block are roundtrip-assigned.
//                  *
//                  * For 2-byte and 4-byte codepages, the bytes are stored as uint16_t
//                  * respectively as uint32_t, in the platform encoding.
//                  * For 3-byte codepages, the bytes are always stored in big-endian order.
//                  *
//                  * For EUC encodings that use only either 0x8e or 0x8f as the first
//                  * byte of their longest byte sequences, the first two bytes in
//                  * this third stage indicate with their 7th bits whether these bytes
//                  * are to be written directly or actually need to be preceeded by
//                  * one of the two Single-Shift codes. With this, the third stage
//                  * stores one byte fewer per character than the actual maximum length of
//                  * EUC byte sequences.
//                  *
//                  * Other than that, leading zero bytes are removed and the other
//                  * bytes output. A single zero byte may be output if the "assigned"
//                  * bit in stage 2 was on.
//                  * The data structure does not support zero byte output as a fallback,
//                  * and also does not allow output of leading zeros.
//                  */
//                 stage2Entry=MBCS_STAGE_2_FROM_U(table, c);

//                 /* get the bytes and the length for the output */
//                 switch(outputType) {
//                 case MBCS_OUTPUT_2:
//                     value=MBCS_VALUE_2_FROM_STAGE_2(bytes, stage2Entry, c);
//                     if(value<=0xff) {
//                         length=1;
//                     } else {
//                         length=2;
//                     }
//                     break;
//                 case MBCS_OUTPUT_2_SISO:
//                     /* 1/2-byte stateful with Shift-In/Shift-Out */
//                     /*
//                      * Save the old state in the converter object
//                      * right here, then change the local prevLength state variable if necessary.
//                      * Then, if this character turns out to be unassigned or a fallback that
//                      * is not taken, the callback code must not save the new state in the converter
//                      * because the new state is for a character that is not output.
//                      * However, the callback must still restore the state from the converter
//                      * in case the callback function changed it for its output.
//                      */
//                     cnv->fromUnicodeStatus=prevLength; /* save the old state */
//                     value=MBCS_VALUE_2_FROM_STAGE_2(bytes, stage2Entry, c);
//                     if(value<=0xff) {
//                         if(value==0 && MBCS_FROM_U_IS_ROUNDTRIP(stage2Entry, c)==0) {
//                             /* no mapping, leave value==0 */
//                             length=0;
//                         } else if(prevLength<=1) {
//                             length=1;
//                         } else {
//                             /* change from double-byte mode to single-byte */
//                             if (siLength == 1) {
//                                 value|=(uint32_t)siBytes[0]<<8;
//                                 length = 2;
//                             } else if (siLength == 2) {
//                                 value|=(uint32_t)siBytes[1]<<8;
//                                 value|=(uint32_t)siBytes[0]<<16;
//                                 length = 3;
//                             }
//                             prevLength=1;
//                         }
//                     } else {
//                         if(prevLength==2) {
//                             length=2;
//                         } else {
//                             /* change from single-byte mode to double-byte */
//                             if (soLength == 1) {
//                                 value|=(uint32_t)soBytes[0]<<16;
//                                 length = 3;
//                             } else if (soLength == 2) {
//                                 value|=(uint32_t)soBytes[1]<<16;
//                                 value|=(uint32_t)soBytes[0]<<24;
//                                 length = 4;
//                             }
//                             prevLength=2;
//                         }
//                     }
//                     break;
//                 case MBCS_OUTPUT_DBCS_ONLY:
//                     /* table with single-byte results, but only DBCS mappings used */
//                     value=MBCS_VALUE_2_FROM_STAGE_2(bytes, stage2Entry, c);
//                     if(value<=0xff) {
//                         /* no mapping or SBCS result, not taken for DBCS-only */
//                         value=stage2Entry=0; /* stage2Entry=0 to reset roundtrip flags */
//                         length=0;
//                     } else {
//                         length=2;
//                     }
//                     break;
//                 case MBCS_OUTPUT_3:
//                     p=MBCS_POINTER_3_FROM_STAGE_2(bytes, stage2Entry, c);
//                     value=((uint32_t)*p<<16)|((uint32_t)p[1]<<8)|p[2];
//                     if(value<=0xff) {
//                         length=1;
//                     } else if(value<=0xffff) {
//                         length=2;
//                     } else {
//                         length=3;
//                     }
//                     break;
//                 case MBCS_OUTPUT_4:
//                     value=MBCS_VALUE_4_FROM_STAGE_2(bytes, stage2Entry, c);
//                     if(value<=0xff) {
//                         length=1;
//                     } else if(value<=0xffff) {
//                         length=2;
//                     } else if(value<=0xffffff) {
//                         length=3;
//                     } else {
//                         length=4;
//                     }
//                     break;
//                 case MBCS_OUTPUT_3_EUC:
//                     value=MBCS_VALUE_2_FROM_STAGE_2(bytes, stage2Entry, c);
//                     /* EUC 16-bit fixed-length representation */
//                     if(value<=0xff) {
//                         length=1;
//                     } else if((value&0x8000)==0) {
//                         value|=0x8e8000;
//                         length=3;
//                     } else if((value&0x80)==0) {
//                         value|=0x8f0080;
//                         length=3;
//                     } else {
//                         length=2;
//                     }
//                     break;
//                 case MBCS_OUTPUT_4_EUC:
//                     p=MBCS_POINTER_3_FROM_STAGE_2(bytes, stage2Entry, c);
//                     value=((uint32_t)*p<<16)|((uint32_t)p[1]<<8)|p[2];
//                     /* EUC 16-bit fixed-length representation applied to the first two bytes */
//                     if(value<=0xff) {
//                         length=1;
//                     } else if(value<=0xffff) {
//                         length=2;
//                     } else if((value&0x800000)==0) {
//                         value|=0x8e800000;
//                         length=4;
//                     } else if((value&0x8000)==0) {
//                         value|=0x8f008000;
//                         length=4;
//                     } else {
//                         length=3;
//                     }
//                     break;
//                 default:
//                     /* must not occur */
//                     /*
//                      * To avoid compiler warnings that value & length may be
//                      * used without having been initialized, we set them here.
//                      * In reality, this is unreachable code.
//                      * Not having a default branch also causes warnings with
//                      * some compilers.
//                      */
//                     value=stage2Entry=0; /* stage2Entry=0 to reset roundtrip flags */
//                     length=0;
//                     break;
//                 }

//                 /* is this code point assigned, or do we use fallbacks? */
//                 if(!(MBCS_FROM_U_IS_ROUNDTRIP(stage2Entry, c)!=0 ||
//                      (UCNV_FROM_U_USE_FALLBACK(cnv, c) && value!=0))
//                 ) {
//                     /*
//                      * We allow a 0 byte output if the "assigned" bit is set for this entry.
//                      * There is no way with this data structure for fallback output
//                      * to be a zero byte.
//                      */

// unassigned:
//                     /* try an extension mapping */
//                     pArgs->source=source;
//                     c=_extFromU(cnv, cnv->sharedData,
//                                 c, &source, sourceLimit,
//                                 &target, target+targetCapacity,
//                                 &offsets, sourceIndex,
//                                 pArgs->flush,
//                                 pErrorCode);
//                     nextSourceIndex+=(int32_t)(source-pArgs->source);
//                     prevLength=cnv->fromUnicodeStatus; /* restore SISO state */

//                     if(U_FAILURE(*pErrorCode)) {
//                         /* not mappable or buffer overflow */
//                         break;
//                     } else {
//                         /* a mapping was written to the target, continue */

//                         /* recalculate the targetCapacity after an extension mapping */
//                         targetCapacity=(int32_t)(pArgs->targetLimit-(char *)target);

//                         /* normal end of conversion: prepare for a new character */
//                         if(offsets!=NULL) {
//                             prevSourceIndex=sourceIndex;
//                             sourceIndex=nextSourceIndex;
//                         }
//                         continue;
//                     }
//                 }
//             }

//             /* write the output character bytes from value and length */
//             /* from the first if in the loop we know that targetCapacity>0 */
//             if(length<=targetCapacity) {
//                 if(offsets==NULL) {
//                     switch(length) {
//                         /* each branch falls through to the next one */
//                     case 4:
//                         *target++=(uint8_t)(value>>24);
//                         U_FALLTHROUGH;
//                     case 3:
//                         *target++=(uint8_t)(value>>16);
//                         U_FALLTHROUGH;
//                     case 2:
//                         *target++=(uint8_t)(value>>8);
//                         U_FALLTHROUGH;
//                     case 1:
//                         *target++=(uint8_t)value;
//                         U_FALLTHROUGH;
//                     default:
//                         /* will never occur */
//                         break;
//                     }
//                 } else {
//                     switch(length) {
//                         /* each branch falls through to the next one */
//                     case 4:
//                         *target++=(uint8_t)(value>>24);
//                         *offsets++=sourceIndex;
//                         U_FALLTHROUGH;
//                     case 3:
//                         *target++=(uint8_t)(value>>16);
//                         *offsets++=sourceIndex;
//                         U_FALLTHROUGH;
//                     case 2:
//                         *target++=(uint8_t)(value>>8);
//                         *offsets++=sourceIndex;
//                         U_FALLTHROUGH;
//                     case 1:
//                         *target++=(uint8_t)value;
//                         *offsets++=sourceIndex;
//                         U_FALLTHROUGH;
//                     default:
//                         /* will never occur */
//                         break;
//                     }
//                 }
//                 targetCapacity-=length;
//             } else {
//                 uint8_t *charErrorBuffer;

//                 /*
//                  * We actually do this backwards here:
//                  * In order to save an intermediate variable, we output
//                  * first to the overflow buffer what does not fit into the
//                  * regular target.
//                  */
//                 /* we know that 1<=targetCapacity<length<=4 */
//                 length-=targetCapacity;
//                 charErrorBuffer=(uint8_t *)cnv->charErrorBuffer;
//                 switch(length) {
//                     /* each branch falls through to the next one */
//                 case 3:
//                     *charErrorBuffer++=(uint8_t)(value>>16);
//                     U_FALLTHROUGH;
//                 case 2:
//                     *charErrorBuffer++=(uint8_t)(value>>8);
//                     U_FALLTHROUGH;
//                 case 1:
//                     *charErrorBuffer=(uint8_t)value;
//                     U_FALLTHROUGH;
//                 default:
//                     /* will never occur */
//                     break;
//                 }
//                 cnv->charErrorBufferLength=(int8_t)length;

//                 /* now output what fits into the regular target */
//                 value>>=8*length; /* length was reduced by targetCapacity */
//                 switch(targetCapacity) {
//                     /* each branch falls through to the next one */
//                 case 3:
//                     *target++=(uint8_t)(value>>16);
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     U_FALLTHROUGH;
//                 case 2:
//                     *target++=(uint8_t)(value>>8);
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     U_FALLTHROUGH;
//                 case 1:
//                     *target++=(uint8_t)value;
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     U_FALLTHROUGH;
//                 default:
//                     /* will never occur */
//                     break;
//                 }

//                 /* target overflow */
//                 targetCapacity=0;
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 c=0;
//                 break;
//             }

//             /* normal end of conversion: prepare for a new character */
//             c=0;
//             if(offsets!=NULL) {
//                 prevSourceIndex=sourceIndex;
//                 sourceIndex=nextSourceIndex;
//             }
//             continue;
//         } else {
//             /* target is full */
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }
//     }

//     /*
//      * the end of the input stream and detection of truncated input
//      * are handled by the framework, but for EBCDIC_STATEFUL conversion
//      * we need to emit an SI at the very end
//      *
//      * conditions:
//      *   successful
//      *   EBCDIC_STATEFUL in DBCS mode
//      *   end of input and no truncated input
//      */
//     if( U_SUCCESS(*pErrorCode) &&
//         outputType==MBCS_OUTPUT_2_SISO && prevLength==2 &&
//         pArgs->flush && source>=sourceLimit && c==0
//     ) {
//         /* EBCDIC_STATEFUL ending with DBCS: emit an SI to return the output stream to SBCS */
//         if(targetCapacity>0) {
//             *target++=(uint8_t)siBytes[0];
//             if (siLength == 2) {
//                 if (targetCapacity<2) {
//                     cnv->charErrorBuffer[0]=(uint8_t)siBytes[1];
//                     cnv->charErrorBufferLength=1;
//                     *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 } else {
//                     *target++=(uint8_t)siBytes[1];
//                 }
//             }
//             if(offsets!=NULL) {
//                 /* set the last source character's index (sourceIndex points at sourceLimit now) */
//                 *offsets++=prevSourceIndex;
//             }
//         } else {
//             /* target is full */
//             cnv->charErrorBuffer[0]=(uint8_t)siBytes[0];
//             if (siLength == 2) {
//                 cnv->charErrorBuffer[1]=(uint8_t)siBytes[1];
//             }
//             cnv->charErrorBufferLength=siLength;
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//         }
//         prevLength=1; /* we switched into SBCS */
//     }

//     /* set the converter state back into UConverter */
//     cnv->fromUChar32=c;
//     cnv->fromUnicodeStatus=prevLength;

//     /* write back the updated pointers */
//     pArgs->source=source;
//     pArgs->target=(char *)target;
//     pArgs->offsets=offsets;
// }

// /*
//  * This is another simple conversion function for internal use by other
//  * conversion implementations.
//  * It does not use the converter state nor call callbacks.
//  * It does not handle the EBCDIC swaplfnl option (set in UConverter).
//  * It handles conversion extensions but not GB 18030.
//  *
//  * It converts one single Unicode code point into codepage bytes, encoded
//  * as one 32-bit value. The function returns the number of bytes in *pValue:
//  * 1..4 the number of bytes in *pValue
//  * 0    unassigned (*pValue undefined)
//  * -1   illegal (currently not used, *pValue undefined)
//  *
//  * *pValue will contain the resulting bytes with the last byte in bits 7..0,
//  * the second to last byte in bits 15..8, etc.
//  * Currently, the function assumes but does not check that 0<=c<=0x10ffff.
//  */
// U_CFUNC int32_t
// ucnv_MBCSFromUChar32(UConverterSharedData *sharedData,
//                  UChar32 c, uint32_t *pValue,
//                  UBool useFallback) {
//     const int32_t *cx;
//     const uint16_t *table;
// #if 0
// /* #if 0 because this is not currently used in ICU - reduce code, increase code coverage */
//     const uint8_t *p;
// #endif
//     uint32_t stage2Entry;
//     uint32_t value;
//     int32_t length;

//     /* BMP-only codepages are stored without stage 1 entries for supplementary code points */
//     if(c<=0xffff || (sharedData->mbcs.unicodeMask&UCNV_HAS_SUPPLEMENTARY)) {
//         table=sharedData->mbcs.fromUnicodeTable;

//         /* convert the Unicode code point in c into codepage bytes (same as in _MBCSFromUnicodeWithOffsets) */
//         if(sharedData->mbcs.outputType==MBCS_OUTPUT_1) {
//             value=MBCS_SINGLE_RESULT_FROM_U(table, (uint16_t *)sharedData->mbcs.fromUnicodeBytes, c);
//             /* is this code point assigned, or do we use fallbacks? */
//             if(useFallback ? value>=0x800 : value>=0xc00) {
//                 *pValue=value&0xff;
//                 return 1;
//             }
//         } else /* outputType!=MBCS_OUTPUT_1 */ {
//             stage2Entry=MBCS_STAGE_2_FROM_U(table, c);

//             /* get the bytes and the length for the output */
//             switch(sharedData->mbcs.outputType) {
//             case MBCS_OUTPUT_2:
//                 value=MBCS_VALUE_2_FROM_STAGE_2(sharedData->mbcs.fromUnicodeBytes, stage2Entry, c);
//                 if(value<=0xff) {
//                     length=1;
//                 } else {
//                     length=2;
//                 }
//                 break;
// #if 0
// /* #if 0 because this is not currently used in ICU - reduce code, increase code coverage */
//             case MBCS_OUTPUT_DBCS_ONLY:
//                 /* table with single-byte results, but only DBCS mappings used */
//                 value=MBCS_VALUE_2_FROM_STAGE_2(sharedData->mbcs.fromUnicodeBytes, stage2Entry, c);
//                 if(value<=0xff) {
//                     /* no mapping or SBCS result, not taken for DBCS-only */
//                     value=stage2Entry=0; /* stage2Entry=0 to reset roundtrip flags */
//                     length=0;
//                 } else {
//                     length=2;
//                 }
//                 break;
//             case MBCS_OUTPUT_3:
//                 p=MBCS_POINTER_3_FROM_STAGE_2(sharedData->mbcs.fromUnicodeBytes, stage2Entry, c);
//                 value=((uint32_t)*p<<16)|((uint32_t)p[1]<<8)|p[2];
//                 if(value<=0xff) {
//                     length=1;
//                 } else if(value<=0xffff) {
//                     length=2;
//                 } else {
//                     length=3;
//                 }
//                 break;
//             case MBCS_OUTPUT_4:
//                 value=MBCS_VALUE_4_FROM_STAGE_2(sharedData->mbcs.fromUnicodeBytes, stage2Entry, c);
//                 if(value<=0xff) {
//                     length=1;
//                 } else if(value<=0xffff) {
//                     length=2;
//                 } else if(value<=0xffffff) {
//                     length=3;
//                 } else {
//                     length=4;
//                 }
//                 break;
//             case MBCS_OUTPUT_3_EUC:
//                 value=MBCS_VALUE_2_FROM_STAGE_2(sharedData->mbcs.fromUnicodeBytes, stage2Entry, c);
//                 /* EUC 16-bit fixed-length representation */
//                 if(value<=0xff) {
//                     length=1;
//                 } else if((value&0x8000)==0) {
//                     value|=0x8e8000;
//                     length=3;
//                 } else if((value&0x80)==0) {
//                     value|=0x8f0080;
//                     length=3;
//                 } else {
//                     length=2;
//                 }
//                 break;
//             case MBCS_OUTPUT_4_EUC:
//                 p=MBCS_POINTER_3_FROM_STAGE_2(sharedData->mbcs.fromUnicodeBytes, stage2Entry, c);
//                 value=((uint32_t)*p<<16)|((uint32_t)p[1]<<8)|p[2];
//                 /* EUC 16-bit fixed-length representation applied to the first two bytes */
//                 if(value<=0xff) {
//                     length=1;
//                 } else if(value<=0xffff) {
//                     length=2;
//                 } else if((value&0x800000)==0) {
//                     value|=0x8e800000;
//                     length=4;
//                 } else if((value&0x8000)==0) {
//                     value|=0x8f008000;
//                     length=4;
//                 } else {
//                     length=3;
//                 }
//                 break;
// #endif
//             default:
//                 /* must not occur */
//                 return -1;
//             }

//             /* is this code point assigned, or do we use fallbacks? */
//             if( MBCS_FROM_U_IS_ROUNDTRIP(stage2Entry, c) ||
//                 (FROM_U_USE_FALLBACK(useFallback, c) && value!=0)
//             ) {
//                 /*
//                  * We allow a 0 byte output if the "assigned" bit is set for this entry.
//                  * There is no way with this data structure for fallback output
//                  * to be a zero byte.
//                  */
//                 /* assigned */
//                 *pValue=value;
//                 return length;
//             }
//         }
//     }

//     cx=sharedData->mbcs.extIndexes;
//     if(cx!=NULL) {
//         length=ucnv_extSimpleMatchFromU(cx, c, pValue, useFallback);
//         return length>=0 ? length : -length;  /* return abs(length); */
//     }

//     /* unassigned */
//     return 0;
// }


// #if 0
// /*
//  * This function has been moved to ucnv2022.c for inlining.
//  * This implementation is here only for documentation purposes
//  */

// /**
//  * This version of ucnv_MBCSFromUChar32() is optimized for single-byte codepages.
//  * It does not handle the EBCDIC swaplfnl option (set in UConverter).
//  * It does not handle conversion extensions (_extFromU()).
//  *
//  * It returns the codepage byte for the code point, or -1 if it is unassigned.
//  */
// U_CFUNC int32_t
// ucnv_MBCSSingleFromUChar32(UConverterSharedData *sharedData,
//                        UChar32 c,
//                        UBool useFallback) {
//     const uint16_t *table;
//     int32_t value;

//     /* BMP-only codepages are stored without stage 1 entries for supplementary code points */
//     if(c>=0x10000 && !(sharedData->mbcs.unicodeMask&UCNV_HAS_SUPPLEMENTARY)) {
//         return -1;
//     }

//     /* convert the Unicode code point in c into codepage bytes (same as in _MBCSFromUnicodeWithOffsets) */
//     table=sharedData->mbcs.fromUnicodeTable;

//     /* get the byte for the output */
//     value=MBCS_SINGLE_RESULT_FROM_U(table, (uint16_t *)sharedData->mbcs.fromUnicodeBytes, c);
//     /* is this code point assigned, or do we use fallbacks? */
//     if(useFallback ? value>=0x800 : value>=0xc00) {
//         return value&0xff;
//     } else {
//         return -1;
//     }
// }
// #endif

// /* MBCS-from-UTF-8 conversion functions ------------------------------------- */

// /* offsets for n-byte UTF-8 sequences that were calculated with ((lead<<6)+trail)<<6+trail... */
// static const UChar32
// utf8_offsets[5]={ 0, 0, 0x3080, 0xE2080, 0x3C82080 };

// static void U_CALLCONV
// ucnv_SBCSFromUTF8(UConverterFromUnicodeArgs *pFromUArgs,
//                   UConverterToUnicodeArgs *pToUArgs,
//                   UErrorCode *pErrorCode) {
//     UConverter *utf8, *cnv;
//     const uint8_t *source, *sourceLimit;
//     uint8_t *target;
//     int32_t targetCapacity;

//     const uint16_t *table, *sbcsIndex;
//     const uint16_t *results;

//     int8_t oldToULength, toULength, toULimit;

//     UChar32 c;
//     uint8_t b, t1, t2;

//     uint32_t asciiRoundtrips;
//     uint16_t value, minValue = 0;
//     UBool hasSupplementary;

//     /* set up the local pointers */
//     utf8=pToUArgs->converter;
//     cnv=pFromUArgs->converter;
//     source=(uint8_t *)pToUArgs->source;
//     sourceLimit=(uint8_t *)pToUArgs->sourceLimit;
//     target=(uint8_t *)pFromUArgs->target;
//     targetCapacity=(int32_t)(pFromUArgs->targetLimit-pFromUArgs->target);

//     table=cnv->sharedData->mbcs.fromUnicodeTable;
//     sbcsIndex=cnv->sharedData->mbcs.sbcsIndex;
//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         results=(uint16_t *)cnv->sharedData->mbcs.swapLFNLFromUnicodeBytes;
//     } else {
//         results=(uint16_t *)cnv->sharedData->mbcs.fromUnicodeBytes;
//     }
//     asciiRoundtrips=cnv->sharedData->mbcs.asciiRoundtrips;

//     if(cnv->useFallback) {
//         /* use all roundtrip and fallback results */
//         minValue=0x800;
//     } else {
//         /* use only roundtrips and fallbacks from private-use characters */
//         minValue=0xc00;
//     }
//     hasSupplementary=(UBool)(cnv->sharedData->mbcs.unicodeMask&UCNV_HAS_SUPPLEMENTARY);

//     /* get the converter state from the UTF-8 UConverter */
//     if(utf8->toULength > 0) {
//         toULength=oldToULength=utf8->toULength;
//         toULimit=(int8_t)utf8->mode;
//         c=(UChar32)utf8->toUnicodeStatus;
//     } else {
//         toULength=oldToULength=toULimit=0;
//         c = 0;
//     }

//     // The conversion loop checks source<sourceLimit only once per 1/2/3-byte character.
//     // If the buffer ends with a truncated 2- or 3-byte sequence,
//     // then we reduce the sourceLimit to before that,
//     // and collect the remaining bytes after the conversion loop.
//     {
//         // Do not go back into the bytes that will be read for finishing a partial
//         // sequence from the previous buffer.
//         int32_t length=(int32_t)(sourceLimit-source) - (toULimit-oldToULength);
//         if(length>0) {
//             uint8_t b1=*(sourceLimit-1);
//             if(U8_IS_SINGLE(b1)) {
//                 // common ASCII character
//             } else if(U8_IS_TRAIL(b1) && length>=2) {
//                 uint8_t b2=*(sourceLimit-2);
//                 if(0xe0<=b2 && b2<0xf0 && U8_IS_VALID_LEAD3_AND_T1(b2, b1)) {
//                     // truncated 3-byte sequence
//                     sourceLimit-=2;
//                 }
//             } else if(0xc2<=b1 && b1<0xf0) {
//                 // truncated 2- or 3-byte sequence
//                 --sourceLimit;
//             }
//         }
//     }

//     if(c!=0 && targetCapacity>0) {
//         utf8->toUnicodeStatus=0;
//         utf8->toULength=0;
//         goto moreBytes;
//         /*
//          * Note: We could avoid the goto by duplicating some of the moreBytes
//          * code, but only up to the point of collecting a complete UTF-8
//          * sequence; then recurse for the toUBytes[toULength]
//          * and then continue with normal conversion.
//          *
//          * If so, move this code to just after initializing the minimum
//          * set of local variables for reading the UTF-8 input
//          * (utf8, source, target, limits but not cnv, table, minValue, etc.).
//          *
//          * Potential advantages:
//          * - avoid the goto
//          * - oldToULength could become a local variable in just those code blocks
//          *   that deal with buffer boundaries
//          * - possibly faster if the goto prevents some compiler optimizations
//          *   (this would need measuring to confirm)
//          * Disadvantage:
//          * - code duplication
//          */
//     }

//     /* conversion loop */
//     while(source<sourceLimit) {
//         if(targetCapacity>0) {
//             b=*source++;
//             if(U8_IS_SINGLE(b)) {
//                 /* convert ASCII */
//                 if(IS_ASCII_ROUNDTRIP(b, asciiRoundtrips)) {
//                     *target++=(uint8_t)b;
//                     --targetCapacity;
//                     continue;
//                 } else {
//                     c=b;
//                     value=SBCS_RESULT_FROM_UTF8(sbcsIndex, results, 0, c);
//                 }
//             } else {
//                 if(b<0xe0) {
//                     if( /* handle U+0080..U+07FF inline */
//                         b>=0xc2 &&
//                         (t1=(uint8_t)(*source-0x80)) <= 0x3f
//                     ) {
//                         c=b&0x1f;
//                         ++source;
//                         value=SBCS_RESULT_FROM_UTF8(sbcsIndex, results, c, t1);
//                         if(value>=minValue) {
//                             *target++=(uint8_t)value;
//                             --targetCapacity;
//                             continue;
//                         } else {
//                             c=(c<<6)|t1;
//                         }
//                     } else {
//                         c=-1;
//                     }
//                 } else if(b==0xe0) {
//                     if( /* handle U+0800..U+0FFF inline */
//                         (t1=(uint8_t)(source[0]-0x80)) <= 0x3f && t1 >= 0x20 &&
//                         (t2=(uint8_t)(source[1]-0x80)) <= 0x3f
//                     ) {
//                         c=t1;
//                         source+=2;
//                         value=SBCS_RESULT_FROM_UTF8(sbcsIndex, results, c, t2);
//                         if(value>=minValue) {
//                             *target++=(uint8_t)value;
//                             --targetCapacity;
//                             continue;
//                         } else {
//                             c=(c<<6)|t2;
//                         }
//                     } else {
//                         c=-1;
//                     }
//                 } else {
//                     c=-1;
//                 }

//                 if(c<0) {
//                     /* handle "complicated" and error cases, and continuing partial characters */
//                     oldToULength=0;
//                     toULength=1;
//                     toULimit=U8_COUNT_BYTES_NON_ASCII(b);
//                     c=b;
// moreBytes:
//                     while(toULength<toULimit) {
//                         /*
//                          * The sourceLimit may have been adjusted before the conversion loop
//                          * to stop before a truncated sequence.
//                          * Here we need to use the real limit in case we have two truncated
//                          * sequences at the end.
//                          * See ticket #7492.
//                          */
//                         if(source<(uint8_t *)pToUArgs->sourceLimit) {
//                             b=*source;
//                             if(icu::UTF8::isValidTrail(c, b, toULength, toULimit)) {
//                                 ++source;
//                                 ++toULength;
//                                 c=(c<<6)+b;
//                             } else {
//                                 break; /* sequence too short, stop with toULength<toULimit */
//                             }
//                         } else {
//                             /* store the partial UTF-8 character, compatible with the regular UTF-8 converter */
//                             source-=(toULength-oldToULength);
//                             while(oldToULength<toULength) {
//                                 utf8->toUBytes[oldToULength++]=*source++;
//                             }
//                             utf8->toUnicodeStatus=c;
//                             utf8->toULength=toULength;
//                             utf8->mode=toULimit;
//                             pToUArgs->source=(char *)source;
//                             pFromUArgs->target=(char *)target;
//                             return;
//                         }
//                     }

//                     if(toULength==toULimit) {
//                         c-=utf8_offsets[toULength];
//                         if(toULength<=3) {  /* BMP */
//                             value=MBCS_SINGLE_RESULT_FROM_U(table, results, c);
//                         } else {
//                             /* supplementary code point */
//                             if(!hasSupplementary) {
//                                 /* BMP-only codepages are stored without stage 1 entries for supplementary code points */
//                                 value=0;
//                             } else {
//                                 value=MBCS_SINGLE_RESULT_FROM_U(table, results, c);
//                             }
//                         }
//                     } else {
//                         /* error handling: illegal UTF-8 byte sequence */
//                         source-=(toULength-oldToULength);
//                         while(oldToULength<toULength) {
//                             utf8->toUBytes[oldToULength++]=*source++;
//                         }
//                         utf8->toULength=toULength;
//                         pToUArgs->source=(char *)source;
//                         pFromUArgs->target=(char *)target;
//                         *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                         return;
//                     }
//                 }
//             }

//             if(value>=minValue) {
//                 /* output the mapping for c */
//                 *target++=(uint8_t)value;
//                 --targetCapacity;
//             } else {
//                 /* value<minValue means c is unassigned (unmappable) */
//                 /*
//                  * Try an extension mapping.
//                  * Pass in no source because we don't have UTF-16 input.
//                  * If we have a partial match on c, we will return and revert
//                  * to UTF-8->UTF-16->charset conversion.
//                  */
//                 static const UChar nul=0;
//                 const UChar *noSource=&nul;
//                 c=_extFromU(cnv, cnv->sharedData,
//                             c, &noSource, noSource,
//                             &target, target+targetCapacity,
//                             NULL, -1,
//                             pFromUArgs->flush,
//                             pErrorCode);

//                 if(U_FAILURE(*pErrorCode)) {
//                     /* not mappable or buffer overflow */
//                     cnv->fromUChar32=c;
//                     break;
//                 } else if(cnv->preFromUFirstCP>=0) {
//                     /*
//                      * Partial match, return and revert to pivoting.
//                      * In normal from-UTF-16 conversion, we would just continue
//                      * but then exit the loop because the extension match would
//                      * have consumed the source.
//                      */
//                     *pErrorCode=U_USING_DEFAULT_WARNING;
//                     break;
//                 } else {
//                     /* a mapping was written to the target, continue */

//                     /* recalculate the targetCapacity after an extension mapping */
//                     targetCapacity=(int32_t)(pFromUArgs->targetLimit-(char *)target);
//                 }
//             }
//         } else {
//             /* target is full */
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }
//     }

//     /*
//      * The sourceLimit may have been adjusted before the conversion loop
//      * to stop before a truncated sequence.
//      * If so, then collect the truncated sequence now.
//      */
//     if(U_SUCCESS(*pErrorCode) &&
//             cnv->preFromUFirstCP<0 &&
//             source<(sourceLimit=(uint8_t *)pToUArgs->sourceLimit)) {
//         c=utf8->toUBytes[0]=b=*source++;
//         toULength=1;
//         toULimit=U8_COUNT_BYTES(b);
//         while(source<sourceLimit) {
//             utf8->toUBytes[toULength++]=b=*source++;
//             c=(c<<6)+b;
//         }
//         utf8->toUnicodeStatus=c;
//         utf8->toULength=toULength;
//         utf8->mode=toULimit;
//     }

//     /* write back the updated pointers */
//     pToUArgs->source=(char *)source;
//     pFromUArgs->target=(char *)target;
// }

// static void U_CALLCONV
// ucnv_DBCSFromUTF8(UConverterFromUnicodeArgs *pFromUArgs,
//                   UConverterToUnicodeArgs *pToUArgs,
//                   UErrorCode *pErrorCode) {
//     UConverter *utf8, *cnv;
//     const uint8_t *source, *sourceLimit;
//     uint8_t *target;
//     int32_t targetCapacity;

//     const uint16_t *table, *mbcsIndex;
//     const uint16_t *results;

//     int8_t oldToULength, toULength, toULimit;

//     UChar32 c;
//     uint8_t b, t1, t2;

//     uint32_t stage2Entry;
//     uint32_t asciiRoundtrips;
//     uint16_t value = 0;
//     UBool hasSupplementary;

//     /* set up the local pointers */
//     utf8=pToUArgs->converter;
//     cnv=pFromUArgs->converter;
//     source=(uint8_t *)pToUArgs->source;
//     sourceLimit=(uint8_t *)pToUArgs->sourceLimit;
//     target=(uint8_t *)pFromUArgs->target;
//     targetCapacity=(int32_t)(pFromUArgs->targetLimit-pFromUArgs->target);

//     table=cnv->sharedData->mbcs.fromUnicodeTable;
//     mbcsIndex=cnv->sharedData->mbcs.mbcsIndex;
//     if((cnv->options&UCNV_OPTION_SWAP_LFNL)!=0) {
//         results=(uint16_t *)cnv->sharedData->mbcs.swapLFNLFromUnicodeBytes;
//     } else {
//         results=(uint16_t *)cnv->sharedData->mbcs.fromUnicodeBytes;
//     }
//     asciiRoundtrips=cnv->sharedData->mbcs.asciiRoundtrips;

//     hasSupplementary=(UBool)(cnv->sharedData->mbcs.unicodeMask&UCNV_HAS_SUPPLEMENTARY);

//     /* get the converter state from the UTF-8 UConverter */
//     if(utf8->toULength > 0) {
//         toULength=oldToULength=utf8->toULength;
//         toULimit=(int8_t)utf8->mode;
//         c=(UChar32)utf8->toUnicodeStatus;
//     } else {
//         toULength=oldToULength=toULimit=0;
//         c = 0;
//     }

//     // The conversion loop checks source<sourceLimit only once per 1/2/3-byte character.
//     // If the buffer ends with a truncated 2- or 3-byte sequence,
//     // then we reduce the sourceLimit to before that,
//     // and collect the remaining bytes after the conversion loop.
//     {
//         // Do not go back into the bytes that will be read for finishing a partial
//         // sequence from the previous buffer.
//         int32_t length=(int32_t)(sourceLimit-source) - (toULimit-oldToULength);
//         if(length>0) {
//             uint8_t b1=*(sourceLimit-1);
//             if(U8_IS_SINGLE(b1)) {
//                 // common ASCII character
//             } else if(U8_IS_TRAIL(b1) && length>=2) {
//                 uint8_t b2=*(sourceLimit-2);
//                 if(0xe0<=b2 && b2<0xf0 && U8_IS_VALID_LEAD3_AND_T1(b2, b1)) {
//                     // truncated 3-byte sequence
//                     sourceLimit-=2;
//                 }
//             } else if(0xc2<=b1 && b1<0xf0) {
//                 // truncated 2- or 3-byte sequence
//                 --sourceLimit;
//             }
//         }
//     }

//     if(c!=0 && targetCapacity>0) {
//         utf8->toUnicodeStatus=0;
//         utf8->toULength=0;
//         goto moreBytes;
//         /* See note in ucnv_SBCSFromUTF8() about this goto. */
//     }

//     /* conversion loop */
//     while(source<sourceLimit) {
//         if(targetCapacity>0) {
//             b=*source++;
//             if(U8_IS_SINGLE(b)) {
//                 /* convert ASCII */
//                 if(IS_ASCII_ROUNDTRIP(b, asciiRoundtrips)) {
//                     *target++=b;
//                     --targetCapacity;
//                     continue;
//                 } else {
//                     value=DBCS_RESULT_FROM_UTF8(mbcsIndex, results, 0, b);
//                     if(value==0) {
//                         c=b;
//                         goto unassigned;
//                     }
//                 }
//             } else {
//                 if(b>=0xe0) {
//                     if( /* handle U+0800..U+D7FF inline */
//                         b<=0xed &&  // do not assume maxFastUChar>0xd7ff
//                         U8_IS_VALID_LEAD3_AND_T1(b, t1=source[0]) &&
//                         (t2=(uint8_t)(source[1]-0x80)) <= 0x3f
//                     ) {
//                         c=((b&0xf)<<6)|(t1&0x3f);
//                         source+=2;
//                         value=DBCS_RESULT_FROM_UTF8(mbcsIndex, results, c, t2);
//                         if(value==0) {
//                             c=(c<<6)|t2;
//                             goto unassigned;
//                         }
//                     } else {
//                         c=-1;
//                     }
//                 } else {
//                     if( /* handle U+0080..U+07FF inline */
//                         b>=0xc2 &&
//                         (t1=(uint8_t)(*source-0x80)) <= 0x3f
//                     ) {
//                         c=b&0x1f;
//                         ++source;
//                         value=DBCS_RESULT_FROM_UTF8(mbcsIndex, results, c, t1);
//                         if(value==0) {
//                             c=(c<<6)|t1;
//                             goto unassigned;
//                         }
//                     } else {
//                         c=-1;
//                     }
//                 }

//                 if(c<0) {
//                     /* handle "complicated" and error cases, and continuing partial characters */
//                     oldToULength=0;
//                     toULength=1;
//                     toULimit=U8_COUNT_BYTES_NON_ASCII(b);
//                     c=b;
// moreBytes:
//                     while(toULength<toULimit) {
//                         /*
//                          * The sourceLimit may have been adjusted before the conversion loop
//                          * to stop before a truncated sequence.
//                          * Here we need to use the real limit in case we have two truncated
//                          * sequences at the end.
//                          * See ticket #7492.
//                          */
//                         if(source<(uint8_t *)pToUArgs->sourceLimit) {
//                             b=*source;
//                             if(icu::UTF8::isValidTrail(c, b, toULength, toULimit)) {
//                                 ++source;
//                                 ++toULength;
//                                 c=(c<<6)+b;
//                             } else {
//                                 break; /* sequence too short, stop with toULength<toULimit */
//                             }
//                         } else {
//                             /* store the partial UTF-8 character, compatible with the regular UTF-8 converter */
//                             source-=(toULength-oldToULength);
//                             while(oldToULength<toULength) {
//                                 utf8->toUBytes[oldToULength++]=*source++;
//                             }
//                             utf8->toUnicodeStatus=c;
//                             utf8->toULength=toULength;
//                             utf8->mode=toULimit;
//                             pToUArgs->source=(char *)source;
//                             pFromUArgs->target=(char *)target;
//                             return;
//                         }
//                     }

//                     if(toULength==toULimit) {
//                         c-=utf8_offsets[toULength];
//                         if(toULength<=3) {  /* BMP */
//                             stage2Entry=MBCS_STAGE_2_FROM_U(table, c);
//                         } else {
//                             /* supplementary code point */
//                             if(!hasSupplementary) {
//                                 /* BMP-only codepages are stored without stage 1 entries for supplementary code points */
//                                 stage2Entry=0;
//                             } else {
//                                 stage2Entry=MBCS_STAGE_2_FROM_U(table, c);
//                             }
//                         }
//                     } else {
//                         /* error handling: illegal UTF-8 byte sequence */
//                         source-=(toULength-oldToULength);
//                         while(oldToULength<toULength) {
//                             utf8->toUBytes[oldToULength++]=*source++;
//                         }
//                         utf8->toULength=toULength;
//                         pToUArgs->source=(char *)source;
//                         pFromUArgs->target=(char *)target;
//                         *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                         return;
//                     }

//                     /* get the bytes and the length for the output */
//                     /* MBCS_OUTPUT_2 */
//                     value=MBCS_VALUE_2_FROM_STAGE_2(results, stage2Entry, c);

//                     /* is this code point assigned, or do we use fallbacks? */
//                     if(!(MBCS_FROM_U_IS_ROUNDTRIP(stage2Entry, c) ||
//                          (UCNV_FROM_U_USE_FALLBACK(cnv, c) && value!=0))
//                     ) {
//                         goto unassigned;
//                     }
//                 }
//             }

//             /* write the output character bytes from value and length */
//             /* from the first if in the loop we know that targetCapacity>0 */
//             if(value<=0xff) {
//                 /* this is easy because we know that there is enough space */
//                 *target++=(uint8_t)value;
//                 --targetCapacity;
//             } else /* length==2 */ {
//                 *target++=(uint8_t)(value>>8);
//                 if(2<=targetCapacity) {
//                     *target++=(uint8_t)value;
//                     targetCapacity-=2;
//                 } else {
//                     cnv->charErrorBuffer[0]=(char)value;
//                     cnv->charErrorBufferLength=1;

//                     /* target overflow */
//                     *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                     break;
//                 }
//             }
//             continue;

// unassigned:
//             {
//                 /*
//                  * Try an extension mapping.
//                  * Pass in no source because we don't have UTF-16 input.
//                  * If we have a partial match on c, we will return and revert
//                  * to UTF-8->UTF-16->charset conversion.
//                  */
//                 static const UChar nul=0;
//                 const UChar *noSource=&nul;
//                 c=_extFromU(cnv, cnv->sharedData,
//                             c, &noSource, noSource,
//                             &target, target+targetCapacity,
//                             NULL, -1,
//                             pFromUArgs->flush,
//                             pErrorCode);

//                 if(U_FAILURE(*pErrorCode)) {
//                     /* not mappable or buffer overflow */
//                     cnv->fromUChar32=c;
//                     break;
//                 } else if(cnv->preFromUFirstCP>=0) {
//                     /*
//                      * Partial match, return and revert to pivoting.
//                      * In normal from-UTF-16 conversion, we would just continue
//                      * but then exit the loop because the extension match would
//                      * have consumed the source.
//                      */
//                     *pErrorCode=U_USING_DEFAULT_WARNING;
//                     break;
//                 } else {
//                     /* a mapping was written to the target, continue */

//                     /* recalculate the targetCapacity after an extension mapping */
//                     targetCapacity=(int32_t)(pFromUArgs->targetLimit-(char *)target);
//                     continue;
//                 }
//             }
//         } else {
//             /* target is full */
//             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//             break;
//         }
//     }

//     /*
//      * The sourceLimit may have been adjusted before the conversion loop
//      * to stop before a truncated sequence.
//      * If so, then collect the truncated sequence now.
//      */
//     if(U_SUCCESS(*pErrorCode) &&
//             cnv->preFromUFirstCP<0 &&
//             source<(sourceLimit=(uint8_t *)pToUArgs->sourceLimit)) {
//         c=utf8->toUBytes[0]=b=*source++;
//         toULength=1;
//         toULimit=U8_COUNT_BYTES(b);
//         while(source<sourceLimit) {
//             utf8->toUBytes[toULength++]=b=*source++;
//             c=(c<<6)+b;
//         }
//         utf8->toUnicodeStatus=c;
//         utf8->toULength=toULength;
//         utf8->mode=toULimit;
//     }

//     /* write back the updated pointers */
//     pToUArgs->source=(char *)source;
//     pFromUArgs->target=(char *)target;
// }

// /* miscellaneous ------------------------------------------------------------ */

// static void U_CALLCONV
// ucnv_MBCSGetStarters(const UConverter* cnv,
//                  UBool starters[256],
//                  UErrorCode *) {
//     const int32_t *state0;
//     int i;

//     state0=cnv->sharedData->mbcs.stateTable[cnv->sharedData->mbcs.dbcsOnlyState];
//     for(i=0; i<256; ++i) {
//         /* all bytes that cause a state transition from state 0 are lead bytes */
//         starters[i]= (UBool)MBCS_ENTRY_IS_TRANSITION(state0[i]);
//     }
// }

// /*
//  * This is an internal function that allows other converter implementations
//  * to check whether a byte is a lead byte.
//  */
// U_CFUNC UBool
// ucnv_MBCSIsLeadByte(UConverterSharedData *sharedData, char byte) {
//     return (UBool)MBCS_ENTRY_IS_TRANSITION(sharedData->mbcs.stateTable[0][(uint8_t)byte]);
// }

// static void U_CALLCONV
// ucnv_MBCSWriteSub(UConverterFromUnicodeArgs *pArgs,
//               int32_t offsetIndex,
//               UErrorCode *pErrorCode) {
//     UConverter *cnv=pArgs->converter;
//     char *p, *subchar;
//     char buffer[4];
//     int32_t length;

//     /* first, select between subChar and subChar1 */
//     if( cnv->subChar1!=0 &&
//         (cnv->sharedData->mbcs.extIndexes!=NULL ?
//             cnv->useSubChar1 :
//             (cnv->invalidUCharBuffer[0]<=0xff))
//     ) {
//         /* select subChar1 if it is set (not 0) and the unmappable Unicode code point is up to U+00ff (IBM MBCS behavior) */
//         subchar=(char *)&cnv->subChar1;
//         length=1;
//     } else {
//         /* select subChar in all other cases */
//         subchar=(char *)cnv->subChars;
//         length=cnv->subCharLen;
//     }

//     /* reset the selector for the next code point */
//     cnv->useSubChar1=FALSE;

//     if (cnv->sharedData->mbcs.outputType == MBCS_OUTPUT_2_SISO) {
//         p=buffer;

//         /* fromUnicodeStatus contains prevLength */
//         switch(length) {
//         case 1:
//             if(cnv->fromUnicodeStatus==2) {
//                 /* DBCS mode and SBCS sub char: change to SBCS */
//                 cnv->fromUnicodeStatus=1;
//                 *p++=UCNV_SI;
//             }
//             *p++=subchar[0];
//             break;
//         case 2:
//             if(cnv->fromUnicodeStatus<=1) {
//                 /* SBCS mode and DBCS sub char: change to DBCS */
//                 cnv->fromUnicodeStatus=2;
//                 *p++=UCNV_SO;
//             }
//             *p++=subchar[0];
//             *p++=subchar[1];
//             break;
//         default:
//             *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
//             return;
//         }
//         subchar=buffer;
//         length=(int32_t)(p-buffer);
//     }

//     ucnv_cbFromUWriteBytes(pArgs, subchar, length, offsetIndex, pErrorCode);
// }

// U_CFUNC UConverterType
// ucnv_MBCSGetType(const UConverter* converter) {
//     /* SBCS, DBCS, and EBCDIC_STATEFUL are replaced by MBCS, but here we cheat a little */
//     if(converter->sharedData->mbcs.countStates==1) {
//         return (UConverterType)UCNV_SBCS;
//     } else if((converter->sharedData->mbcs.outputType&0xff)==MBCS_OUTPUT_2_SISO) {
//         return (UConverterType)UCNV_EBCDIC_STATEFUL;
//     } else if(converter->sharedData->staticData->minBytesPerChar==2 && converter->sharedData->staticData->maxBytesPerChar==2) {
//         return (UConverterType)UCNV_DBCS;
//     }
//     return (UConverterType)UCNV_MBCS;
// }

// #endif /* #if !UCONFIG_NO_LEGACY_CONVERSION */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// ******************************************************************************
// *
// *   Copyright (C) 2000-2016, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *
// ******************************************************************************
// *   file name:  ucnvscsu.c
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2000nov18
// *   created by: Markus W. Scherer
// *
// *   This is an implementation of the Standard Compression Scheme for Unicode
// *   as defined in http://www.unicode.org/unicode/reports/tr6/ .
// *   Reserved commands and window settings are treated as illegal sequences and
// *   will result in callback calls.
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION && !UCONFIG_ONLY_HTML_CONVERSION

// #include "unicode/ucnv.h"
// #include "unicode/ucnv_cb.h"
// #include "unicode/utf16.h"
// #include "ucnv_bld.h"
// #include "ucnv_cnv.h"
// #include "cmemory.h"

// /* SCSU definitions --------------------------------------------------------- */

// /* SCSU command byte values */
// enum {
//     SQ0=0x01, /* Quote from window pair 0 */
//     SQ7=0x08, /* Quote from window pair 7 */
//     SDX=0x0B, /* Define a window as extended */
//     Srs=0x0C, /* reserved */
//     SQU=0x0E, /* Quote a single Unicode character */
//     SCU=0x0F, /* Change to Unicode mode */
//     SC0=0x10, /* Select window 0 */
//     SC7=0x17, /* Select window 7 */
//     SD0=0x18, /* Define and select window 0 */
//     SD7=0x1F, /* Define and select window 7 */

//     UC0=0xE0, /* Select window 0 */
//     UC7=0xE7, /* Select window 7 */
//     UD0=0xE8, /* Define and select window 0 */
//     UD7=0xEF, /* Define and select window 7 */
//     UQU=0xF0, /* Quote a single Unicode character */
//     UDX=0xF1, /* Define a Window as extended */
//     Urs=0xF2  /* reserved */
// };

// enum {
//     /*
//      * Unicode code points from 3400 to E000 are not adressible by
//      * dynamic window, since in these areas no short run alphabets are
//      * found. Therefore add gapOffset to all values from gapThreshold.
//      */
//     gapThreshold=0x68,
//     gapOffset=0xAC00,

//     /* values between reservedStart and fixedThreshold are reserved */
//     reservedStart=0xA8,

//     /* use table of predefined fixed offsets for values from fixedThreshold */
//     fixedThreshold=0xF9
// };

// /* constant offsets for the 8 static windows */
// static const uint32_t staticOffsets[8]={
//     0x0000, /* ASCII for quoted tags */
//     0x0080, /* Latin - 1 Supplement (for access to punctuation) */
//     0x0100, /* Latin Extended-A */
//     0x0300, /* Combining Diacritical Marks */
//     0x2000, /* General Punctuation */
//     0x2080, /* Currency Symbols */
//     0x2100, /* Letterlike Symbols and Number Forms */
//     0x3000  /* CJK Symbols and punctuation */
// };

// /* initial offsets for the 8 dynamic (sliding) windows */
// static const uint32_t initialDynamicOffsets[8]={
//     0x0080, /* Latin-1 */
//     0x00C0, /* Latin Extended A */
//     0x0400, /* Cyrillic */
//     0x0600, /* Arabic */
//     0x0900, /* Devanagari */
//     0x3040, /* Hiragana */
//     0x30A0, /* Katakana */
//     0xFF00  /* Fullwidth ASCII */
// };

// /* Table of fixed predefined Offsets */
// static const uint32_t fixedOffsets[]={
//     /* 0xF9 */ 0x00C0, /* Latin-1 Letters + half of Latin Extended A */
//     /* 0xFA */ 0x0250, /* IPA extensions */
//     /* 0xFB */ 0x0370, /* Greek */
//     /* 0xFC */ 0x0530, /* Armenian */
//     /* 0xFD */ 0x3040, /* Hiragana */
//     /* 0xFE */ 0x30A0, /* Katakana */
//     /* 0xFF */ 0xFF60  /* Halfwidth Katakana */
// };

// /* state values */
// enum {
//     readCommand,
//     quotePairOne,
//     quotePairTwo,
//     quoteOne,
//     definePairOne,
//     definePairTwo,
//     defineOne
// };

// typedef struct SCSUData {
//     /* dynamic window offsets, intitialize to default values from initialDynamicOffsets */
//     uint32_t toUDynamicOffsets[8];
//     uint32_t fromUDynamicOffsets[8];

//     /* state machine state - toUnicode */
//     UBool toUIsSingleByteMode;
//     uint8_t toUState;
//     int8_t toUQuoteWindow, toUDynamicWindow;
//     uint8_t toUByteOne;
//     uint8_t toUPadding[3];

//     /* state machine state - fromUnicode */
//     UBool fromUIsSingleByteMode;
//     int8_t fromUDynamicWindow;

//     /*
//      * windowUse[] keeps track of the use of the dynamic windows:
//      * At nextWindowUseIndex there is the least recently used window,
//      * and the following windows (in a wrapping manner) are more and more
//      * recently used.
//      * At nextWindowUseIndex-1 there is the most recently used window.
//      */
//     uint8_t locale;
//     int8_t nextWindowUseIndex;
//     int8_t windowUse[8];
// } SCSUData;

// static const int8_t initialWindowUse[8]={ 7, 0, 3, 2, 4, 5, 6, 1 };
// static const int8_t initialWindowUse_ja[8]={ 3, 2, 4, 1, 0, 7, 5, 6 };

// enum {
//     lGeneric, l_ja
// };

// /* SCSU setup functions ----------------------------------------------------- */
// U_CDECL_BEGIN
// static void U_CALLCONV
// _SCSUReset(UConverter *cnv, UConverterResetChoice choice) {
//     SCSUData *scsu=(SCSUData *)cnv->extraInfo;

//     if(choice<=UCNV_RESET_TO_UNICODE) {
//         /* reset toUnicode */
//         uprv_memcpy(scsu->toUDynamicOffsets, initialDynamicOffsets, 32);

//         scsu->toUIsSingleByteMode=TRUE;
//         scsu->toUState=readCommand;
//         scsu->toUQuoteWindow=scsu->toUDynamicWindow=0;
//         scsu->toUByteOne=0;

//         cnv->toULength=0;
//     }
//     if(choice!=UCNV_RESET_TO_UNICODE) {
//         /* reset fromUnicode */
//         uprv_memcpy(scsu->fromUDynamicOffsets, initialDynamicOffsets, 32);

//         scsu->fromUIsSingleByteMode=TRUE;
//         scsu->fromUDynamicWindow=0;

//         scsu->nextWindowUseIndex=0;
//         switch(scsu->locale) {
//         case l_ja:
//             uprv_memcpy(scsu->windowUse, initialWindowUse_ja, 8);
//             break;
//         default:
//             uprv_memcpy(scsu->windowUse, initialWindowUse, 8);
//             break;
//         }

//         cnv->fromUChar32=0;
//     }
// }

// static void U_CALLCONV
// _SCSUOpen(UConverter *cnv,
//           UConverterLoadArgs *pArgs,
//           UErrorCode *pErrorCode) {
//     const char *locale=pArgs->locale;
//     if(pArgs->onlyTestIsLoadable) {
//         return;
//     }
//     cnv->extraInfo=uprv_malloc(sizeof(SCSUData));
//     if(cnv->extraInfo!=NULL) {
//         if(locale!=NULL && locale[0]=='j' && locale[1]=='a' && (locale[2]==0 || locale[2]=='_')) {
//             ((SCSUData *)cnv->extraInfo)->locale=l_ja;
//         } else {
//             ((SCSUData *)cnv->extraInfo)->locale=lGeneric;
//         }
//         _SCSUReset(cnv, UCNV_RESET_BOTH);
//     } else {
//         *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
//     }

//     /* Set the substitution character U+fffd as a Unicode string. */
//     cnv->subUChars[0]=0xfffd;
//     cnv->subCharLen=-1;
// }

// static void U_CALLCONV
// _SCSUClose(UConverter *cnv) {
//     if(cnv->extraInfo!=NULL) {
//         if(!cnv->isExtraLocal) {
//             uprv_free(cnv->extraInfo);
//         }
//         cnv->extraInfo=NULL;
//     }
// }

// /* SCSU-to-Unicode conversion functions ------------------------------------- */

// static void U_CALLCONV
// _SCSUToUnicodeWithOffsets(UConverterToUnicodeArgs *pArgs,
//                           UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     SCSUData *scsu;
//     const uint8_t *source, *sourceLimit;
//     UChar *target;
//     const UChar *targetLimit;
//     int32_t *offsets;
//     UBool isSingleByteMode;
//     uint8_t state, byteOne;
//     int8_t quoteWindow, dynamicWindow;

//     int32_t sourceIndex, nextSourceIndex;

//     uint8_t b;

//     /* set up the local pointers */
//     cnv=pArgs->converter;
//     scsu=(SCSUData *)cnv->extraInfo;

//     source=(const uint8_t *)pArgs->source;
//     sourceLimit=(const uint8_t *)pArgs->sourceLimit;
//     target=pArgs->target;
//     targetLimit=pArgs->targetLimit;
//     offsets=pArgs->offsets;

//     /* get the state machine state */
//     isSingleByteMode=scsu->toUIsSingleByteMode;
//     state=scsu->toUState;
//     quoteWindow=scsu->toUQuoteWindow;
//     dynamicWindow=scsu->toUDynamicWindow;
//     byteOne=scsu->toUByteOne;

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     sourceIndex=state==readCommand ? 0 : -1;
//     nextSourceIndex=0;

//     /*
//      * conversion "loop"
//      *
//      * For performance, this is not a normal C loop.
//      * Instead, there are two code blocks for the two SCSU modes.
//      * The function branches to either one, and a change of the mode is done with a goto to
//      * the other branch.
//      *
//      * Each branch has two conventional loops:
//      * - a fast-path loop for the most common codes in the mode
//      * - a loop for all other codes in the mode
//      * When the fast-path runs into a code that it cannot handle, its loop ends and it
//      * runs into the following loop to handle the other codes.
//      * The end of the input or output buffer is also handled by the slower loop.
//      * The slow loop jumps (goto) to the fast-path loop again as soon as possible.
//      *
//      * The callback handling is done by returning with an error code.
//      * The conversion framework actually calls the callback function.
//      */
//     if(isSingleByteMode) {
//         /* fast path for single-byte mode */
//         if(state==readCommand) {
// fastSingle:
//             while(source<sourceLimit && target<targetLimit && (b=*source)>=0x20) {
//                 ++source;
//                 ++nextSourceIndex;
//                 if(b<=0x7f) {
//                     /* write US-ASCII graphic character or DEL */
//                     *target++=(UChar)b;
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                 } else {
//                     /* write from dynamic window */
//                     uint32_t c=scsu->toUDynamicOffsets[dynamicWindow]+(b&0x7f);
//                     if(c<=0xffff) {
//                         *target++=(UChar)c;
//                         if(offsets!=NULL) {
//                             *offsets++=sourceIndex;
//                         }
//                     } else {
//                         /* output surrogate pair */
//                         *target++=(UChar)(0xd7c0+(c>>10));
//                         if(target<targetLimit) {
//                             *target++=(UChar)(0xdc00|(c&0x3ff));
//                             if(offsets!=NULL) {
//                                 *offsets++=sourceIndex;
//                                 *offsets++=sourceIndex;
//                             }
//                         } else {
//                             /* target overflow */
//                             if(offsets!=NULL) {
//                                 *offsets++=sourceIndex;
//                             }
//                             cnv->UCharErrorBuffer[0]=(UChar)(0xdc00|(c&0x3ff));
//                             cnv->UCharErrorBufferLength=1;
//                             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                             goto endloop;
//                         }
//                     }
//                 }
//                 sourceIndex=nextSourceIndex;
//             }
//         }

//         /* normal state machine for single-byte mode, minus handling for what fastSingle covers */
// singleByteMode:
//         while(source<sourceLimit) {
//             if(target>=targetLimit) {
//                 /* target is full */
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 break;
//             }
//             b=*source++;
//             ++nextSourceIndex;
//             switch(state) {
//             case readCommand:
//                 /* redundant conditions are commented out */
//                 /* here: b<0x20 because otherwise we would be in fastSingle */
//                 if((1UL<<b)&0x2601 /* binary 0010 0110 0000 0001, check for b==0xd || b==0xa || b==9 || b==0 */) {
//                     /* CR/LF/TAB/NUL */
//                     *target++=(UChar)b;
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     sourceIndex=nextSourceIndex;
//                     goto fastSingle;
//                 } else if(SC0<=b) {
//                     if(b<=SC7) {
//                         dynamicWindow=(int8_t)(b-SC0);
//                         sourceIndex=nextSourceIndex;
//                         goto fastSingle;
//                     } else /* if(SD0<=b && b<=SD7) */ {
//                         dynamicWindow=(int8_t)(b-SD0);
//                         state=defineOne;
//                     }
//                 } else if(/* SQ0<=b && */ b<=SQ7) {
//                     quoteWindow=(int8_t)(b-SQ0);
//                     state=quoteOne;
//                 } else if(b==SDX) {
//                     state=definePairOne;
//                 } else if(b==SQU) {
//                     state=quotePairOne;
//                 } else if(b==SCU) {
//                     sourceIndex=nextSourceIndex;
//                     isSingleByteMode=FALSE;
//                     goto fastUnicode;
//                 } else /* Srs */ {
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     goto endloop;
//                 }

//                 /* store the first byte of a multibyte sequence in toUBytes[] */
//                 cnv->toUBytes[0]=b;
//                 cnv->toULength=1;
//                 break;
//             case quotePairOne:
//                 byteOne=b;
//                 cnv->toUBytes[1]=b;
//                 cnv->toULength=2;
//                 state=quotePairTwo;
//                 break;
//             case quotePairTwo:
//                 *target++=(UChar)((byteOne<<8)|b);
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 sourceIndex=nextSourceIndex;
//                 state=readCommand;
//                 goto fastSingle;
//             case quoteOne:
//                 if(b<0x80) {
//                     /* all static offsets are in the BMP */
//                     *target++=(UChar)(staticOffsets[quoteWindow]+b);
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                 } else {
//                     /* write from dynamic window */
//                     uint32_t c=scsu->toUDynamicOffsets[quoteWindow]+(b&0x7f);
//                     if(c<=0xffff) {
//                         *target++=(UChar)c;
//                         if(offsets!=NULL) {
//                             *offsets++=sourceIndex;
//                         }
//                     } else {
//                         /* output surrogate pair */
//                         *target++=(UChar)(0xd7c0+(c>>10));
//                         if(target<targetLimit) {
//                             *target++=(UChar)(0xdc00|(c&0x3ff));
//                             if(offsets!=NULL) {
//                                 *offsets++=sourceIndex;
//                                 *offsets++=sourceIndex;
//                             }
//                         } else {
//                             /* target overflow */
//                             if(offsets!=NULL) {
//                                 *offsets++=sourceIndex;
//                             }
//                             cnv->UCharErrorBuffer[0]=(UChar)(0xdc00|(c&0x3ff));
//                             cnv->UCharErrorBufferLength=1;
//                             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                             goto endloop;
//                         }
//                     }
//                 }
//                 sourceIndex=nextSourceIndex;
//                 state=readCommand;
//                 goto fastSingle;
//             case definePairOne:
//                 dynamicWindow=(int8_t)((b>>5)&7);
//                 byteOne=(uint8_t)(b&0x1f);
//                 cnv->toUBytes[1]=b;
//                 cnv->toULength=2;
//                 state=definePairTwo;
//                 break;
//             case definePairTwo:
//                 scsu->toUDynamicOffsets[dynamicWindow]=0x10000+(byteOne<<15UL | b<<7UL);
//                 sourceIndex=nextSourceIndex;
//                 state=readCommand;
//                 goto fastSingle;
//             case defineOne:
//                 if(b==0) {
//                     /* callback(illegal): Reserved window offset value 0 */
//                     cnv->toUBytes[1]=b;
//                     cnv->toULength=2;
//                     goto endloop;
//                 } else if(b<gapThreshold) {
//                     scsu->toUDynamicOffsets[dynamicWindow]=b<<7UL;
//                 } else if((uint8_t)(b-gapThreshold)<(reservedStart-gapThreshold)) {
//                     scsu->toUDynamicOffsets[dynamicWindow]=(b<<7UL)+gapOffset;
//                 } else if(b>=fixedThreshold) {
//                     scsu->toUDynamicOffsets[dynamicWindow]=fixedOffsets[b-fixedThreshold];
//                 } else {
//                     /* callback(illegal): Reserved window offset value 0xa8..0xf8 */
//                     cnv->toUBytes[1]=b;
//                     cnv->toULength=2;
//                     goto endloop;
//                 }
//                 sourceIndex=nextSourceIndex;
//                 state=readCommand;
//                 goto fastSingle;
//             }
//         }
//     } else {
//         /* fast path for Unicode mode */
//         if(state==readCommand) {
// fastUnicode:
//             while(source+1<sourceLimit && target<targetLimit && (uint8_t)((b=*source)-UC0)>(Urs-UC0)) {
//                 *target++=(UChar)((b<<8)|source[1]);
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 sourceIndex=nextSourceIndex;
//                 nextSourceIndex+=2;
//                 source+=2;
//             }
//         }

//         /* normal state machine for Unicode mode */
// /* unicodeByteMode: */
//         while(source<sourceLimit) {
//             if(target>=targetLimit) {
//                 /* target is full */
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 break;
//             }
//             b=*source++;
//             ++nextSourceIndex;
//             switch(state) {
//             case readCommand:
//                 if((uint8_t)(b-UC0)>(Urs-UC0)) {
//                     byteOne=b;
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     state=quotePairTwo;
//                 } else if(/* UC0<=b && */ b<=UC7) {
//                     dynamicWindow=(int8_t)(b-UC0);
//                     sourceIndex=nextSourceIndex;
//                     isSingleByteMode=TRUE;
//                     goto fastSingle;
//                 } else if(/* UD0<=b && */ b<=UD7) {
//                     dynamicWindow=(int8_t)(b-UD0);
//                     isSingleByteMode=TRUE;
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     state=defineOne;
//                     goto singleByteMode;
//                 } else if(b==UDX) {
//                     isSingleByteMode=TRUE;
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     state=definePairOne;
//                     goto singleByteMode;
//                 } else if(b==UQU) {
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     state=quotePairOne;
//                 } else /* Urs */ {
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     goto endloop;
//                 }
//                 break;
//             case quotePairOne:
//                 byteOne=b;
//                 cnv->toUBytes[1]=b;
//                 cnv->toULength=2;
//                 state=quotePairTwo;
//                 break;
//             case quotePairTwo:
//                 *target++=(UChar)((byteOne<<8)|b);
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 sourceIndex=nextSourceIndex;
//                 state=readCommand;
//                 goto fastUnicode;
//             }
//         }
//     }
// endloop:

//     /* set the converter state back into UConverter */
//     if(U_FAILURE(*pErrorCode) && *pErrorCode!=U_BUFFER_OVERFLOW_ERROR) {
//         /* reset to deal with the next character */
//         state=readCommand;
//     } else if(state==readCommand) {
//         /* not in a multi-byte sequence, reset toULength */
//         cnv->toULength=0;
//     }
//     scsu->toUIsSingleByteMode=isSingleByteMode;
//     scsu->toUState=state;
//     scsu->toUQuoteWindow=quoteWindow;
//     scsu->toUDynamicWindow=dynamicWindow;
//     scsu->toUByteOne=byteOne;

//     /* write back the updated pointers */
//     pArgs->source=(const char *)source;
//     pArgs->target=target;
//     pArgs->offsets=offsets;
//     return;
// }

// /*
//  * Identical to _SCSUToUnicodeWithOffsets but without offset handling.
//  * If a change is made in the original function, then either
//  * change this function the same way or
//  * re-copy the original function and remove the variables
//  * offsets, sourceIndex, and nextSourceIndex.
//  */
// static void U_CALLCONV
// _SCSUToUnicode(UConverterToUnicodeArgs *pArgs,
//                UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     SCSUData *scsu;
//     const uint8_t *source, *sourceLimit;
//     UChar *target;
//     const UChar *targetLimit;
//     UBool isSingleByteMode;
//     uint8_t state, byteOne;
//     int8_t quoteWindow, dynamicWindow;

//     uint8_t b;

//     /* set up the local pointers */
//     cnv=pArgs->converter;
//     scsu=(SCSUData *)cnv->extraInfo;

//     source=(const uint8_t *)pArgs->source;
//     sourceLimit=(const uint8_t *)pArgs->sourceLimit;
//     target=pArgs->target;
//     targetLimit=pArgs->targetLimit;

//     /* get the state machine state */
//     isSingleByteMode=scsu->toUIsSingleByteMode;
//     state=scsu->toUState;
//     quoteWindow=scsu->toUQuoteWindow;
//     dynamicWindow=scsu->toUDynamicWindow;
//     byteOne=scsu->toUByteOne;

//     /*
//      * conversion "loop"
//      *
//      * For performance, this is not a normal C loop.
//      * Instead, there are two code blocks for the two SCSU modes.
//      * The function branches to either one, and a change of the mode is done with a goto to
//      * the other branch.
//      *
//      * Each branch has two conventional loops:
//      * - a fast-path loop for the most common codes in the mode
//      * - a loop for all other codes in the mode
//      * When the fast-path runs into a code that it cannot handle, its loop ends and it
//      * runs into the following loop to handle the other codes.
//      * The end of the input or output buffer is also handled by the slower loop.
//      * The slow loop jumps (goto) to the fast-path loop again as soon as possible.
//      *
//      * The callback handling is done by returning with an error code.
//      * The conversion framework actually calls the callback function.
//      */
//     if(isSingleByteMode) {
//         /* fast path for single-byte mode */
//         if(state==readCommand) {
// fastSingle:
//             while(source<sourceLimit && target<targetLimit && (b=*source)>=0x20) {
//                 ++source;
//                 if(b<=0x7f) {
//                     /* write US-ASCII graphic character or DEL */
//                     *target++=(UChar)b;
//                 } else {
//                     /* write from dynamic window */
//                     uint32_t c=scsu->toUDynamicOffsets[dynamicWindow]+(b&0x7f);
//                     if(c<=0xffff) {
//                         *target++=(UChar)c;
//                     } else {
//                         /* output surrogate pair */
//                         *target++=(UChar)(0xd7c0+(c>>10));
//                         if(target<targetLimit) {
//                             *target++=(UChar)(0xdc00|(c&0x3ff));
//                         } else {
//                             /* target overflow */
//                             cnv->UCharErrorBuffer[0]=(UChar)(0xdc00|(c&0x3ff));
//                             cnv->UCharErrorBufferLength=1;
//                             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                             goto endloop;
//                         }
//                     }
//                 }
//             }
//         }

//         /* normal state machine for single-byte mode, minus handling for what fastSingle covers */
// singleByteMode:
//         while(source<sourceLimit) {
//             if(target>=targetLimit) {
//                 /* target is full */
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 break;
//             }
//             b=*source++;
//             switch(state) {
//             case readCommand:
//                 /* redundant conditions are commented out */
//                 /* here: b<0x20 because otherwise we would be in fastSingle */
//                 if((1UL<<b)&0x2601 /* binary 0010 0110 0000 0001, check for b==0xd || b==0xa || b==9 || b==0 */) {
//                     /* CR/LF/TAB/NUL */
//                     *target++=(UChar)b;
//                     goto fastSingle;
//                 } else if(SC0<=b) {
//                     if(b<=SC7) {
//                         dynamicWindow=(int8_t)(b-SC0);
//                         goto fastSingle;
//                     } else /* if(SD0<=b && b<=SD7) */ {
//                         dynamicWindow=(int8_t)(b-SD0);
//                         state=defineOne;
//                     }
//                 } else if(/* SQ0<=b && */ b<=SQ7) {
//                     quoteWindow=(int8_t)(b-SQ0);
//                     state=quoteOne;
//                 } else if(b==SDX) {
//                     state=definePairOne;
//                 } else if(b==SQU) {
//                     state=quotePairOne;
//                 } else if(b==SCU) {
//                     isSingleByteMode=FALSE;
//                     goto fastUnicode;
//                 } else /* Srs */ {
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     goto endloop;
//                 }

//                 /* store the first byte of a multibyte sequence in toUBytes[] */
//                 cnv->toUBytes[0]=b;
//                 cnv->toULength=1;
//                 break;
//             case quotePairOne:
//                 byteOne=b;
//                 cnv->toUBytes[1]=b;
//                 cnv->toULength=2;
//                 state=quotePairTwo;
//                 break;
//             case quotePairTwo:
//                 *target++=(UChar)((byteOne<<8)|b);
//                 state=readCommand;
//                 goto fastSingle;
//             case quoteOne:
//                 if(b<0x80) {
//                     /* all static offsets are in the BMP */
//                     *target++=(UChar)(staticOffsets[quoteWindow]+b);
//                 } else {
//                     /* write from dynamic window */
//                     uint32_t c=scsu->toUDynamicOffsets[quoteWindow]+(b&0x7f);
//                     if(c<=0xffff) {
//                         *target++=(UChar)c;
//                     } else {
//                         /* output surrogate pair */
//                         *target++=(UChar)(0xd7c0+(c>>10));
//                         if(target<targetLimit) {
//                             *target++=(UChar)(0xdc00|(c&0x3ff));
//                         } else {
//                             /* target overflow */
//                             cnv->UCharErrorBuffer[0]=(UChar)(0xdc00|(c&0x3ff));
//                             cnv->UCharErrorBufferLength=1;
//                             *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                             goto endloop;
//                         }
//                     }
//                 }
//                 state=readCommand;
//                 goto fastSingle;
//             case definePairOne:
//                 dynamicWindow=(int8_t)((b>>5)&7);
//                 byteOne=(uint8_t)(b&0x1f);
//                 cnv->toUBytes[1]=b;
//                 cnv->toULength=2;
//                 state=definePairTwo;
//                 break;
//             case definePairTwo:
//                 scsu->toUDynamicOffsets[dynamicWindow]=0x10000+(byteOne<<15UL | b<<7UL);
//                 state=readCommand;
//                 goto fastSingle;
//             case defineOne:
//                 if(b==0) {
//                     /* callback(illegal): Reserved window offset value 0 */
//                     cnv->toUBytes[1]=b;
//                     cnv->toULength=2;
//                     goto endloop;
//                 } else if(b<gapThreshold) {
//                     scsu->toUDynamicOffsets[dynamicWindow]=b<<7UL;
//                 } else if((uint8_t)(b-gapThreshold)<(reservedStart-gapThreshold)) {
//                     scsu->toUDynamicOffsets[dynamicWindow]=(b<<7UL)+gapOffset;
//                 } else if(b>=fixedThreshold) {
//                     scsu->toUDynamicOffsets[dynamicWindow]=fixedOffsets[b-fixedThreshold];
//                 } else {
//                     /* callback(illegal): Reserved window offset value 0xa8..0xf8 */
//                     cnv->toUBytes[1]=b;
//                     cnv->toULength=2;
//                     goto endloop;
//                 }
//                 state=readCommand;
//                 goto fastSingle;
//             }
//         }
//     } else {
//         /* fast path for Unicode mode */
//         if(state==readCommand) {
// fastUnicode:
//             while(source+1<sourceLimit && target<targetLimit && (uint8_t)((b=*source)-UC0)>(Urs-UC0)) {
//                 *target++=(UChar)((b<<8)|source[1]);
//                 source+=2;
//             }
//         }

//         /* normal state machine for Unicode mode */
// /* unicodeByteMode: */
//         while(source<sourceLimit) {
//             if(target>=targetLimit) {
//                 /* target is full */
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 break;
//             }
//             b=*source++;
//             switch(state) {
//             case readCommand:
//                 if((uint8_t)(b-UC0)>(Urs-UC0)) {
//                     byteOne=b;
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     state=quotePairTwo;
//                 } else if(/* UC0<=b && */ b<=UC7) {
//                     dynamicWindow=(int8_t)(b-UC0);
//                     isSingleByteMode=TRUE;
//                     goto fastSingle;
//                 } else if(/* UD0<=b && */ b<=UD7) {
//                     dynamicWindow=(int8_t)(b-UD0);
//                     isSingleByteMode=TRUE;
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     state=defineOne;
//                     goto singleByteMode;
//                 } else if(b==UDX) {
//                     isSingleByteMode=TRUE;
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     state=definePairOne;
//                     goto singleByteMode;
//                 } else if(b==UQU) {
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     state=quotePairOne;
//                 } else /* Urs */ {
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                     cnv->toUBytes[0]=b;
//                     cnv->toULength=1;
//                     goto endloop;
//                 }
//                 break;
//             case quotePairOne:
//                 byteOne=b;
//                 cnv->toUBytes[1]=b;
//                 cnv->toULength=2;
//                 state=quotePairTwo;
//                 break;
//             case quotePairTwo:
//                 *target++=(UChar)((byteOne<<8)|b);
//                 state=readCommand;
//                 goto fastUnicode;
//             }
//         }
//     }
// endloop:

//     /* set the converter state back into UConverter */
//     if(U_FAILURE(*pErrorCode) && *pErrorCode!=U_BUFFER_OVERFLOW_ERROR) {
//         /* reset to deal with the next character */
//         state=readCommand;
//     } else if(state==readCommand) {
//         /* not in a multi-byte sequence, reset toULength */
//         cnv->toULength=0;
//     }
//     scsu->toUIsSingleByteMode=isSingleByteMode;
//     scsu->toUState=state;
//     scsu->toUQuoteWindow=quoteWindow;
//     scsu->toUDynamicWindow=dynamicWindow;
//     scsu->toUByteOne=byteOne;

//     /* write back the updated pointers */
//     pArgs->source=(const char *)source;
//     pArgs->target=target;
//     return;
// }
// U_CDECL_END
// /* SCSU-from-Unicode conversion functions ----------------------------------- */

// /*
//  * This SCSU Encoder is fairly simple but uses all SCSU commands to achieve
//  * reasonable results. The lookahead is minimal.
//  * Many cases are simple:
//  * A character fits directly into the current mode, a dynamic or static window,
//  * or is not compressible. These cases are tested first.
//  * Real compression heuristics are applied to the rest, in code branches for
//  * single/Unicode mode and BMP/supplementary code points.
//  * The heuristics used here are extremely simple.
//  */

// /* get the number of the window that this character is in, or -1 */
// static int8_t
// getWindow(const uint32_t offsets[8], uint32_t c) {
//     int i;
//     for(i=0; i<8; ++i) {
//         if((uint32_t)(c-offsets[i])<=0x7f) {
//             return (int8_t)(i);
//         }
//     }
//     return -1;
// }

// /* is the character in the dynamic window starting at the offset, or in the direct-encoded range? */
// static UBool
// isInOffsetWindowOrDirect(uint32_t offset, uint32_t c) {
//     return (UBool)(c<=offset+0x7f &&
//           (c>=offset || (c<=0x7f &&
//                         (c>=0x20 || (1UL<<c)&0x2601))));
//                                 /* binary 0010 0110 0000 0001,
//                                    check for b==0xd || b==0xa || b==9 || b==0 */
// }

// /*
//  * getNextDynamicWindow returns the next dynamic window to be redefined
//  */
// static int8_t
// getNextDynamicWindow(SCSUData *scsu) {
//     int8_t window=scsu->windowUse[scsu->nextWindowUseIndex];
//     if(++scsu->nextWindowUseIndex==8) {
//         scsu->nextWindowUseIndex=0;
//     }
//     return window;
// }

// /*
//  * useDynamicWindow() adjusts
//  * windowUse[] and nextWindowUseIndex for the algorithm to choose
//  * the next dynamic window to be defined;
//  * a subclass may override it and provide its own algorithm.
//  */
// static void
// useDynamicWindow(SCSUData *scsu, int8_t window) {
//     /*
//      * move the existing window, which just became the most recently used one,
//      * up in windowUse[] to nextWindowUseIndex-1
//      */

//     /* first, find the index of the window - backwards to favor the more recently used windows */
//     int i, j;

//     i=scsu->nextWindowUseIndex;
//     do {
//         if(--i<0) {
//             i=7;
//         }
//     } while(scsu->windowUse[i]!=window);

//     /* now copy each windowUse[i+1] to [i] */
//     j=i+1;
//     if(j==8) {
//         j=0;
//     }
//     while(j!=scsu->nextWindowUseIndex) {
//         scsu->windowUse[i]=scsu->windowUse[j];
//         i=j;
//         if(++j==8) { j=0; }
//     }

//     /* finally, set the window into the most recently used index */
//     scsu->windowUse[i]=window;
// }

// /*
//  * calculate the offset and the code for a dynamic window that contains the character
//  * takes fixed offsets into account
//  * the offset of the window is stored in the offset variable,
//  * the code is returned
//  *
//  * return offset code: -1 none  <=0xff code for SDn/UDn  else code for SDX/UDX, subtract 0x200 to get the true code
//  */
// static int
// getDynamicOffset(uint32_t c, uint32_t *pOffset) {
//     int i;

//     for(i=0; i<7; ++i) {
//         if((uint32_t)(c-fixedOffsets[i])<=0x7f) {
//             *pOffset=fixedOffsets[i];
//             return 0xf9+i;
//         }
//     }

//     if(c<0x80) {
//         /* No dynamic window for US-ASCII. */
//         return -1;
//     } else if(c<0x3400 ||
//               (uint32_t)(c-0x10000)<(0x14000-0x10000) ||
//               (uint32_t)(c-0x1d000)<=(0x1ffff-0x1d000)
//     ) {
//         /* This character is in a code range for a "small", i.e., reasonably windowable, script. */
//         *pOffset=c&0x7fffff80;
//         return (int)(c>>7);
//     } else if(0xe000<=c && c!=0xfeff && c<0xfff0) {
//         /* For these characters we need to take the gapOffset into account. */
//         *pOffset=c&0x7fffff80;
//         return (int)((c-gapOffset)>>7);
//     } else {
//         return -1;
//     }
// }
// U_CDECL_BEGIN
// /*
//  * Idea for compression:
//  *  - save SCSUData and other state before really starting work
//  *  - at endloop, see if compression could be better with just unicode mode
//  *  - don't do this if a callback has been called
//  *  - if unicode mode would be smaller, then override the results with it - may need SCU at the beginning
//  *  - different buffer handling!
//  *
//  * Drawback or need for corrective handling:
//  * it is desirable to encode U+feff as SQU fe ff for the SCSU signature, and
//  * it is desirable to start a document in US-ASCII/Latin-1 for as long as possible
//  * not only for compression but also for HTML/XML documents with following charset/encoding announcers.
//  *
//  * How to achieve both?
//  *  - Only replace the result after an SDX or SCU?
//  */

// static void U_CALLCONV
// _SCSUFromUnicodeWithOffsets(UConverterFromUnicodeArgs *pArgs,
//                             UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     SCSUData *scsu;
//     const UChar *source, *sourceLimit;
//     uint8_t *target;
//     int32_t targetCapacity;
//     int32_t *offsets;

//     UBool isSingleByteMode;
//     uint8_t dynamicWindow;
//     uint32_t currentOffset;

//     uint32_t c, delta;

//     int32_t sourceIndex, nextSourceIndex;

//     int32_t length;

//     /* variables for compression heuristics */
//     uint32_t offset;
//     UChar lead, trail;
//     int code;
//     int8_t window;

//     /* set up the local pointers */
//     cnv=pArgs->converter;
//     scsu=(SCSUData *)cnv->extraInfo;

//     /* set up the local pointers */
//     source=pArgs->source;
//     sourceLimit=pArgs->sourceLimit;
//     target=(uint8_t *)pArgs->target;
//     targetCapacity=(int32_t)(pArgs->targetLimit-pArgs->target);
//     offsets=pArgs->offsets;

//     /* get the state machine state */
//     isSingleByteMode=scsu->fromUIsSingleByteMode;
//     dynamicWindow=scsu->fromUDynamicWindow;
//     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow];

//     c=cnv->fromUChar32;

//     /* sourceIndex=-1 if the current character began in the previous buffer */
//     sourceIndex= c==0 ? 0 : -1;
//     nextSourceIndex=0;

//     /* similar conversion "loop" as in toUnicode */
// loop:
//     if(isSingleByteMode) {
//         if(c!=0 && targetCapacity>0) {
//             goto getTrailSingle;
//         }

//         /* state machine for single-byte mode */
// /* singleByteMode: */
//         while(source<sourceLimit) {
//             if(targetCapacity<=0) {
//                 /* target is full */
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 break;
//             }
//             c=*source++;
//             ++nextSourceIndex;

//             if((c-0x20)<=0x5f) {
//                 /* pass US-ASCII graphic character through */
//                 *target++=(uint8_t)c;
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 --targetCapacity;
//             } else if(c<0x20) {
//                 if((1UL<<c)&0x2601 /* binary 0010 0110 0000 0001, check for b==0xd || b==0xa || b==9 || b==0 */) {
//                     /* CR/LF/TAB/NUL */
//                     *target++=(uint8_t)c;
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     --targetCapacity;
//                 } else {
//                     /* quote C0 control character */
//                     c|=SQ0<<8;
//                     length=2;
//                     goto outputBytes;
//                 }
//             } else if((delta=c-currentOffset)<=0x7f) {
//                 /* use the current dynamic window */
//                 *target++=(uint8_t)(delta|0x80);
//                 if(offsets!=NULL) {
//                     *offsets++=sourceIndex;
//                 }
//                 --targetCapacity;
//             } else if(U16_IS_SURROGATE(c)) {
//                 if(U16_IS_SURROGATE_LEAD(c)) {
// getTrailSingle:
//                     lead=(UChar)c;
//                     if(source<sourceLimit) {
//                         /* test the following code unit */
//                         trail=*source;
//                         if(U16_IS_TRAIL(trail)) {
//                             ++source;
//                             ++nextSourceIndex;
//                             c=U16_GET_SUPPLEMENTARY(c, trail);
//                             /* convert this surrogate code point */
//                             /* exit this condition tree */
//                         } else {
//                             /* this is an unmatched lead code unit (1st surrogate) */
//                             /* callback(illegal) */
//                             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                             goto endloop;
//                         }
//                     } else {
//                         /* no more input */
//                         break;
//                     }
//                 } else {
//                     /* this is an unmatched trail code unit (2nd surrogate) */
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                     goto endloop;
//                 }

//                 /* compress supplementary character U+10000..U+10ffff */
//                 if((delta=c-currentOffset)<=0x7f) {
//                     /* use the current dynamic window */
//                     *target++=(uint8_t)(delta|0x80);
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     --targetCapacity;
//                 } else if((window=getWindow(scsu->fromUDynamicOffsets, c))>=0) {
//                     /* there is a dynamic window that contains this character, change to it */
//                     dynamicWindow=window;
//                     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow];
//                     useDynamicWindow(scsu, dynamicWindow);
//                     c=((uint32_t)(SC0+dynamicWindow)<<8)|(c-currentOffset)|0x80;
//                     length=2;
//                     goto outputBytes;
//                 } else if((code=getDynamicOffset(c, &offset))>=0) {
//                     /* might check if there are more characters in this window to come */
//                     /* define an extended window with this character */
//                     code-=0x200;
//                     dynamicWindow=getNextDynamicWindow(scsu);
//                     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow]=offset;
//                     useDynamicWindow(scsu, dynamicWindow);
//                     c=((uint32_t)SDX<<24)|((uint32_t)dynamicWindow<<21)|((uint32_t)code<<8)|(c-currentOffset)|0x80;
//                     length=4;
//                     goto outputBytes;
//                 } else {
//                     /* change to Unicode mode and output this (lead, trail) pair */
//                     isSingleByteMode=FALSE;
//                     *target++=(uint8_t)SCU;
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                     }
//                     --targetCapacity;
//                     c=((uint32_t)lead<<16)|trail;
//                     length=4;
//                     goto outputBytes;
//                 }
//             } else if(c<0xa0) {
//                 /* quote C1 control character */
//                 c=(c&0x7f)|(SQ0+1)<<8; /* SQ0+1==SQ1 */
//                 length=2;
//                 goto outputBytes;
//             } else if(c==0xfeff || c>=0xfff0) {
//                 /* quote signature character=byte order mark and specials */
//                 c|=SQU<<16;
//                 length=3;
//                 goto outputBytes;
//             } else {
//                 /* compress all other BMP characters */
//                 if((window=getWindow(scsu->fromUDynamicOffsets, c))>=0) {
//                     /* there is a window defined that contains this character - switch to it or quote from it? */
//                     if(source>=sourceLimit || isInOffsetWindowOrDirect(scsu->fromUDynamicOffsets[window], *source)) {
//                         /* change to dynamic window */
//                         dynamicWindow=window;
//                         currentOffset=scsu->fromUDynamicOffsets[dynamicWindow];
//                         useDynamicWindow(scsu, dynamicWindow);
//                         c=((uint32_t)(SC0+dynamicWindow)<<8)|(c-currentOffset)|0x80;
//                         length=2;
//                         goto outputBytes;
//                     } else {
//                         /* quote from dynamic window */
//                         c=((uint32_t)(SQ0+window)<<8)|(c-scsu->fromUDynamicOffsets[window])|0x80;
//                         length=2;
//                         goto outputBytes;
//                     }
//                 } else if((window=getWindow(staticOffsets, c))>=0) {
//                     /* quote from static window */
//                     c=((uint32_t)(SQ0+window)<<8)|(c-staticOffsets[window]);
//                     length=2;
//                     goto outputBytes;
//                 } else if((code=getDynamicOffset(c, &offset))>=0) {
//                     /* define a dynamic window with this character */
//                     dynamicWindow=getNextDynamicWindow(scsu);
//                     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow]=offset;
//                     useDynamicWindow(scsu, dynamicWindow);
//                     c=((uint32_t)(SD0+dynamicWindow)<<16)|((uint32_t)code<<8)|(c-currentOffset)|0x80;
//                     length=3;
//                     goto outputBytes;
//                 } else if((uint32_t)(c-0x3400)<(0xd800-0x3400) &&
//                           (source>=sourceLimit || (uint32_t)(*source-0x3400)<(0xd800-0x3400))
//                 ) {
//                     /*
//                      * this character is not compressible (a BMP ideograph or similar);
//                      * switch to Unicode mode if this is the last character in the block
//                      * or there is at least one more ideograph following immediately
//                      */
//                     isSingleByteMode=FALSE;
//                     c|=SCU<<16;
//                     length=3;
//                     goto outputBytes;
//                 } else {
//                     /* quote Unicode */
//                     c|=SQU<<16;
//                     length=3;
//                     goto outputBytes;
//                 }
//             }

//             /* normal end of conversion: prepare for a new character */
//             c=0;
//             sourceIndex=nextSourceIndex;
//         }
//     } else {
//         if(c!=0 && targetCapacity>0) {
//             goto getTrailUnicode;
//         }

//         /* state machine for Unicode mode */
// /* unicodeByteMode: */
//         while(source<sourceLimit) {
//             if(targetCapacity<=0) {
//                 /* target is full */
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 break;
//             }
//             c=*source++;
//             ++nextSourceIndex;

//             if((uint32_t)(c-0x3400)<(0xd800-0x3400)) {
//                 /* not compressible, write character directly */
//                 if(targetCapacity>=2) {
//                     *target++=(uint8_t)(c>>8);
//                     *target++=(uint8_t)c;
//                     if(offsets!=NULL) {
//                         *offsets++=sourceIndex;
//                         *offsets++=sourceIndex;
//                     }
//                     targetCapacity-=2;
//                 } else {
//                     length=2;
//                     goto outputBytes;
//                 }
//             } else if((uint32_t)(c-0x3400)>=(0xf300-0x3400) /* c<0x3400 || c>=0xf300 */) {
//                 /* compress BMP character if the following one is not an uncompressible ideograph */
//                 if(!(source<sourceLimit && (uint32_t)(*source-0x3400)<(0xd800-0x3400))) {
//                     if(((uint32_t)(c-0x30)<10 || (uint32_t)(c-0x61)<26 || (uint32_t)(c-0x41)<26)) {
//                         /* ASCII digit or letter */
//                         isSingleByteMode=TRUE;
//                         c|=((uint32_t)(UC0+dynamicWindow)<<8)|c;
//                         length=2;
//                         goto outputBytes;
//                     } else if((window=getWindow(scsu->fromUDynamicOffsets, c))>=0) {
//                         /* there is a dynamic window that contains this character, change to it */
//                         isSingleByteMode=TRUE;
//                         dynamicWindow=window;
//                         currentOffset=scsu->fromUDynamicOffsets[dynamicWindow];
//                         useDynamicWindow(scsu, dynamicWindow);
//                         c=((uint32_t)(UC0+dynamicWindow)<<8)|(c-currentOffset)|0x80;
//                         length=2;
//                         goto outputBytes;
//                     } else if((code=getDynamicOffset(c, &offset))>=0) {
//                         /* define a dynamic window with this character */
//                         isSingleByteMode=TRUE;
//                         dynamicWindow=getNextDynamicWindow(scsu);
//                         currentOffset=scsu->fromUDynamicOffsets[dynamicWindow]=offset;
//                         useDynamicWindow(scsu, dynamicWindow);
//                         c=((uint32_t)(UD0+dynamicWindow)<<16)|((uint32_t)code<<8)|(c-currentOffset)|0x80;
//                         length=3;
//                         goto outputBytes;
//                     }
//                 }

//                 /* don't know how to compress this character, just write it directly */
//                 length=2;
//                 goto outputBytes;
//             } else if(c<0xe000) {
//                 /* c is a surrogate */
//                 if(U16_IS_SURROGATE_LEAD(c)) {
// getTrailUnicode:
//                     lead=(UChar)c;
//                     if(source<sourceLimit) {
//                         /* test the following code unit */
//                         trail=*source;
//                         if(U16_IS_TRAIL(trail)) {
//                             ++source;
//                             ++nextSourceIndex;
//                             c=U16_GET_SUPPLEMENTARY(c, trail);
//                             /* convert this surrogate code point */
//                             /* exit this condition tree */
//                         } else {
//                             /* this is an unmatched lead code unit (1st surrogate) */
//                             /* callback(illegal) */
//                             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                             goto endloop;
//                         }
//                     } else {
//                         /* no more input */
//                         break;
//                     }
//                 } else {
//                     /* this is an unmatched trail code unit (2nd surrogate) */
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                     goto endloop;
//                 }

//                 /* compress supplementary character */
//                 if( (window=getWindow(scsu->fromUDynamicOffsets, c))>=0 &&
//                     !(source<sourceLimit && (uint32_t)(*source-0x3400)<(0xd800-0x3400))
//                 ) {
//                     /*
//                      * there is a dynamic window that contains this character and
//                      * the following character is not uncompressible,
//                      * change to the window
//                      */
//                     isSingleByteMode=TRUE;
//                     dynamicWindow=window;
//                     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow];
//                     useDynamicWindow(scsu, dynamicWindow);
//                     c=((uint32_t)(UC0+dynamicWindow)<<8)|(c-currentOffset)|0x80;
//                     length=2;
//                     goto outputBytes;
//                 } else if(source<sourceLimit && lead==*source && /* too lazy to check trail in same window as source[1] */
//                           (code=getDynamicOffset(c, &offset))>=0
//                 ) {
//                     /* two supplementary characters in (probably) the same window - define an extended one */
//                     isSingleByteMode=TRUE;
//                     code-=0x200;
//                     dynamicWindow=getNextDynamicWindow(scsu);
//                     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow]=offset;
//                     useDynamicWindow(scsu, dynamicWindow);
//                     c=((uint32_t)UDX<<24)|((uint32_t)dynamicWindow<<21)|((uint32_t)code<<8)|(c-currentOffset)|0x80;
//                     length=4;
//                     goto outputBytes;
//                 } else {
//                     /* don't know how to compress this character, just write it directly */
//                     c=((uint32_t)lead<<16)|trail;
//                     length=4;
//                     goto outputBytes;
//                 }
//             } else /* 0xe000<=c<0xf300 */ {
//                 /* quote to avoid SCSU tags */
//                 c|=UQU<<16;
//                 length=3;
//                 goto outputBytes;
//             }

//             /* normal end of conversion: prepare for a new character */
//             c=0;
//             sourceIndex=nextSourceIndex;
//         }
//     }
// endloop:

//     /* set the converter state back into UConverter */
//     scsu->fromUIsSingleByteMode=isSingleByteMode;
//     scsu->fromUDynamicWindow=dynamicWindow;

//     cnv->fromUChar32=c;

//     /* write back the updated pointers */
//     pArgs->source=source;
//     pArgs->target=(char *)target;
//     pArgs->offsets=offsets;
//     return;

// outputBytes:
//     /* write the output character bytes from c and length [code copied from ucnvmbcs.c] */
//     /* from the first if in the loop we know that targetCapacity>0 */
//     if(length<=targetCapacity) {
//         if(offsets==NULL) {
//             switch(length) {
//                 /* each branch falls through to the next one */
//             case 4:
//                 *target++=(uint8_t)(c>>24);
//                 U_FALLTHROUGH;
//             case 3:
//                 *target++=(uint8_t)(c>>16);
//                 U_FALLTHROUGH;
//             case 2:
//                 *target++=(uint8_t)(c>>8);
//                 U_FALLTHROUGH;
//             case 1:
//                 *target++=(uint8_t)c;
//                 U_FALLTHROUGH;
//             default:
//                 /* will never occur */
//                 break;
//             }
//         } else {
//             switch(length) {
//                 /* each branch falls through to the next one */
//             case 4:
//                 *target++=(uint8_t)(c>>24);
//                 *offsets++=sourceIndex;
//                 U_FALLTHROUGH;
//             case 3:
//                 *target++=(uint8_t)(c>>16);
//                 *offsets++=sourceIndex;
//                 U_FALLTHROUGH;
//             case 2:
//                 *target++=(uint8_t)(c>>8);
//                 *offsets++=sourceIndex;
//                 U_FALLTHROUGH;
//             case 1:
//                 *target++=(uint8_t)c;
//                 *offsets++=sourceIndex;
//                 U_FALLTHROUGH;
//             default:
//                 /* will never occur */
//                 break;
//             }
//         }
//         targetCapacity-=length;

//         /* normal end of conversion: prepare for a new character */
//         c=0;
//         sourceIndex=nextSourceIndex;
//         goto loop;
//     } else {
//         uint8_t *p;

//         /*
//          * We actually do this backwards here:
//          * In order to save an intermediate variable, we output
//          * first to the overflow buffer what does not fit into the
//          * regular target.
//          */
//         /* we know that 0<=targetCapacity<length<=4 */
//         /* targetCapacity==0 when SCU+supplementary where SCU used up targetCapacity==1 */
//         length-=targetCapacity;
//         p=(uint8_t *)cnv->charErrorBuffer;
//         switch(length) {
//             /* each branch falls through to the next one */
//         case 4:
//             *p++=(uint8_t)(c>>24);
//             U_FALLTHROUGH;
//         case 3:
//             *p++=(uint8_t)(c>>16);
//             U_FALLTHROUGH;
//         case 2:
//             *p++=(uint8_t)(c>>8);
//             U_FALLTHROUGH;
//         case 1:
//             *p=(uint8_t)c;
//             U_FALLTHROUGH;
//         default:
//             /* will never occur */
//             break;
//         }
//         cnv->charErrorBufferLength=(int8_t)length;

//         /* now output what fits into the regular target */
//         c>>=8*length; /* length was reduced by targetCapacity */
//         switch(targetCapacity) {
//             /* each branch falls through to the next one */
//         case 3:
//             *target++=(uint8_t)(c>>16);
//             if(offsets!=NULL) {
//                 *offsets++=sourceIndex;
//             }
//             U_FALLTHROUGH;
//         case 2:
//             *target++=(uint8_t)(c>>8);
//             if(offsets!=NULL) {
//                 *offsets++=sourceIndex;
//             }
//             U_FALLTHROUGH;
//         case 1:
//             *target++=(uint8_t)c;
//             if(offsets!=NULL) {
//                 *offsets++=sourceIndex;
//             }
//             U_FALLTHROUGH;
//         default:
//             break;
//         }

//         /* target overflow */
//         targetCapacity=0;
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//         c=0;
//         goto endloop;
//     }
// }

// /*
//  * Identical to _SCSUFromUnicodeWithOffsets but without offset handling.
//  * If a change is made in the original function, then either
//  * change this function the same way or
//  * re-copy the original function and remove the variables
//  * offsets, sourceIndex, and nextSourceIndex.
//  */
// static void U_CALLCONV
// _SCSUFromUnicode(UConverterFromUnicodeArgs *pArgs,
//                  UErrorCode *pErrorCode) {
//     UConverter *cnv;
//     SCSUData *scsu;
//     const UChar *source, *sourceLimit;
//     uint8_t *target;
//     int32_t targetCapacity;

//     UBool isSingleByteMode;
//     uint8_t dynamicWindow;
//     uint32_t currentOffset;

//     uint32_t c, delta;

//     int32_t length;

//     /* variables for compression heuristics */
//     uint32_t offset;
//     UChar lead, trail;
//     int code;
//     int8_t window;

//     /* set up the local pointers */
//     cnv=pArgs->converter;
//     scsu=(SCSUData *)cnv->extraInfo;

//     /* set up the local pointers */
//     source=pArgs->source;
//     sourceLimit=pArgs->sourceLimit;
//     target=(uint8_t *)pArgs->target;
//     targetCapacity=(int32_t)(pArgs->targetLimit-pArgs->target);

//     /* get the state machine state */
//     isSingleByteMode=scsu->fromUIsSingleByteMode;
//     dynamicWindow=scsu->fromUDynamicWindow;
//     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow];

//     c=cnv->fromUChar32;

//     /* similar conversion "loop" as in toUnicode */
// loop:
//     if(isSingleByteMode) {
//         if(c!=0 && targetCapacity>0) {
//             goto getTrailSingle;
//         }

//         /* state machine for single-byte mode */
// /* singleByteMode: */
//         while(source<sourceLimit) {
//             if(targetCapacity<=0) {
//                 /* target is full */
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 break;
//             }
//             c=*source++;

//             if((c-0x20)<=0x5f) {
//                 /* pass US-ASCII graphic character through */
//                 *target++=(uint8_t)c;
//                 --targetCapacity;
//             } else if(c<0x20) {
//                 if((1UL<<c)&0x2601 /* binary 0010 0110 0000 0001, check for b==0xd || b==0xa || b==9 || b==0 */) {
//                     /* CR/LF/TAB/NUL */
//                     *target++=(uint8_t)c;
//                     --targetCapacity;
//                 } else {
//                     /* quote C0 control character */
//                     c|=SQ0<<8;
//                     length=2;
//                     goto outputBytes;
//                 }
//             } else if((delta=c-currentOffset)<=0x7f) {
//                 /* use the current dynamic window */
//                 *target++=(uint8_t)(delta|0x80);
//                 --targetCapacity;
//             } else if(U16_IS_SURROGATE(c)) {
//                 if(U16_IS_SURROGATE_LEAD(c)) {
// getTrailSingle:
//                     lead=(UChar)c;
//                     if(source<sourceLimit) {
//                         /* test the following code unit */
//                         trail=*source;
//                         if(U16_IS_TRAIL(trail)) {
//                             ++source;
//                             c=U16_GET_SUPPLEMENTARY(c, trail);
//                             /* convert this surrogate code point */
//                             /* exit this condition tree */
//                         } else {
//                             /* this is an unmatched lead code unit (1st surrogate) */
//                             /* callback(illegal) */
//                             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                             goto endloop;
//                         }
//                     } else {
//                         /* no more input */
//                         break;
//                     }
//                 } else {
//                     /* this is an unmatched trail code unit (2nd surrogate) */
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                     goto endloop;
//                 }

//                 /* compress supplementary character U+10000..U+10ffff */
//                 if((delta=c-currentOffset)<=0x7f) {
//                     /* use the current dynamic window */
//                     *target++=(uint8_t)(delta|0x80);
//                     --targetCapacity;
//                 } else if((window=getWindow(scsu->fromUDynamicOffsets, c))>=0) {
//                     /* there is a dynamic window that contains this character, change to it */
//                     dynamicWindow=window;
//                     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow];
//                     useDynamicWindow(scsu, dynamicWindow);
//                     c=((uint32_t)(SC0+dynamicWindow)<<8)|(c-currentOffset)|0x80;
//                     length=2;
//                     goto outputBytes;
//                 } else if((code=getDynamicOffset(c, &offset))>=0) {
//                     /* might check if there are more characters in this window to come */
//                     /* define an extended window with this character */
//                     code-=0x200;
//                     dynamicWindow=getNextDynamicWindow(scsu);
//                     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow]=offset;
//                     useDynamicWindow(scsu, dynamicWindow);
//                     c=((uint32_t)SDX<<24)|((uint32_t)dynamicWindow<<21)|((uint32_t)code<<8)|(c-currentOffset)|0x80;
//                     length=4;
//                     goto outputBytes;
//                 } else {
//                     /* change to Unicode mode and output this (lead, trail) pair */
//                     isSingleByteMode=FALSE;
//                     *target++=(uint8_t)SCU;
//                     --targetCapacity;
//                     c=((uint32_t)lead<<16)|trail;
//                     length=4;
//                     goto outputBytes;
//                 }
//             } else if(c<0xa0) {
//                 /* quote C1 control character */
//                 c=(c&0x7f)|(SQ0+1)<<8; /* SQ0+1==SQ1 */
//                 length=2;
//                 goto outputBytes;
//             } else if(c==0xfeff || c>=0xfff0) {
//                 /* quote signature character=byte order mark and specials */
//                 c|=SQU<<16;
//                 length=3;
//                 goto outputBytes;
//             } else {
//                 /* compress all other BMP characters */
//                 if((window=getWindow(scsu->fromUDynamicOffsets, c))>=0) {
//                     /* there is a window defined that contains this character - switch to it or quote from it? */
//                     if(source>=sourceLimit || isInOffsetWindowOrDirect(scsu->fromUDynamicOffsets[window], *source)) {
//                         /* change to dynamic window */
//                         dynamicWindow=window;
//                         currentOffset=scsu->fromUDynamicOffsets[dynamicWindow];
//                         useDynamicWindow(scsu, dynamicWindow);
//                         c=((uint32_t)(SC0+dynamicWindow)<<8)|(c-currentOffset)|0x80;
//                         length=2;
//                         goto outputBytes;
//                     } else {
//                         /* quote from dynamic window */
//                         c=((uint32_t)(SQ0+window)<<8)|(c-scsu->fromUDynamicOffsets[window])|0x80;
//                         length=2;
//                         goto outputBytes;
//                     }
//                 } else if((window=getWindow(staticOffsets, c))>=0) {
//                     /* quote from static window */
//                     c=((uint32_t)(SQ0+window)<<8)|(c-staticOffsets[window]);
//                     length=2;
//                     goto outputBytes;
//                 } else if((code=getDynamicOffset(c, &offset))>=0) {
//                     /* define a dynamic window with this character */
//                     dynamicWindow=getNextDynamicWindow(scsu);
//                     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow]=offset;
//                     useDynamicWindow(scsu, dynamicWindow);
//                     c=((uint32_t)(SD0+dynamicWindow)<<16)|((uint32_t)code<<8)|(c-currentOffset)|0x80;
//                     length=3;
//                     goto outputBytes;
//                 } else if((uint32_t)(c-0x3400)<(0xd800-0x3400) &&
//                           (source>=sourceLimit || (uint32_t)(*source-0x3400)<(0xd800-0x3400))
//                 ) {
//                     /*
//                      * this character is not compressible (a BMP ideograph or similar);
//                      * switch to Unicode mode if this is the last character in the block
//                      * or there is at least one more ideograph following immediately
//                      */
//                     isSingleByteMode=FALSE;
//                     c|=SCU<<16;
//                     length=3;
//                     goto outputBytes;
//                 } else {
//                     /* quote Unicode */
//                     c|=SQU<<16;
//                     length=3;
//                     goto outputBytes;
//                 }
//             }

//             /* normal end of conversion: prepare for a new character */
//             c=0;
//         }
//     } else {
//         if(c!=0 && targetCapacity>0) {
//             goto getTrailUnicode;
//         }

//         /* state machine for Unicode mode */
// /* unicodeByteMode: */
//         while(source<sourceLimit) {
//             if(targetCapacity<=0) {
//                 /* target is full */
//                 *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//                 break;
//             }
//             c=*source++;

//             if((uint32_t)(c-0x3400)<(0xd800-0x3400)) {
//                 /* not compressible, write character directly */
//                 if(targetCapacity>=2) {
//                     *target++=(uint8_t)(c>>8);
//                     *target++=(uint8_t)c;
//                     targetCapacity-=2;
//                 } else {
//                     length=2;
//                     goto outputBytes;
//                 }
//             } else if((uint32_t)(c-0x3400)>=(0xf300-0x3400) /* c<0x3400 || c>=0xf300 */) {
//                 /* compress BMP character if the following one is not an uncompressible ideograph */
//                 if(!(source<sourceLimit && (uint32_t)(*source-0x3400)<(0xd800-0x3400))) {
//                     if(((uint32_t)(c-0x30)<10 || (uint32_t)(c-0x61)<26 || (uint32_t)(c-0x41)<26)) {
//                         /* ASCII digit or letter */
//                         isSingleByteMode=TRUE;
//                         c|=((uint32_t)(UC0+dynamicWindow)<<8)|c;
//                         length=2;
//                         goto outputBytes;
//                     } else if((window=getWindow(scsu->fromUDynamicOffsets, c))>=0) {
//                         /* there is a dynamic window that contains this character, change to it */
//                         isSingleByteMode=TRUE;
//                         dynamicWindow=window;
//                         currentOffset=scsu->fromUDynamicOffsets[dynamicWindow];
//                         useDynamicWindow(scsu, dynamicWindow);
//                         c=((uint32_t)(UC0+dynamicWindow)<<8)|(c-currentOffset)|0x80;
//                         length=2;
//                         goto outputBytes;
//                     } else if((code=getDynamicOffset(c, &offset))>=0) {
//                         /* define a dynamic window with this character */
//                         isSingleByteMode=TRUE;
//                         dynamicWindow=getNextDynamicWindow(scsu);
//                         currentOffset=scsu->fromUDynamicOffsets[dynamicWindow]=offset;
//                         useDynamicWindow(scsu, dynamicWindow);
//                         c=((uint32_t)(UD0+dynamicWindow)<<16)|((uint32_t)code<<8)|(c-currentOffset)|0x80;
//                         length=3;
//                         goto outputBytes;
//                     }
//                 }

//                 /* don't know how to compress this character, just write it directly */
//                 length=2;
//                 goto outputBytes;
//             } else if(c<0xe000) {
//                 /* c is a surrogate */
//                 if(U16_IS_SURROGATE_LEAD(c)) {
// getTrailUnicode:
//                     lead=(UChar)c;
//                     if(source<sourceLimit) {
//                         /* test the following code unit */
//                         trail=*source;
//                         if(U16_IS_TRAIL(trail)) {
//                             ++source;
//                             c=U16_GET_SUPPLEMENTARY(c, trail);
//                             /* convert this surrogate code point */
//                             /* exit this condition tree */
//                         } else {
//                             /* this is an unmatched lead code unit (1st surrogate) */
//                             /* callback(illegal) */
//                             *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                             goto endloop;
//                         }
//                     } else {
//                         /* no more input */
//                         break;
//                     }
//                 } else {
//                     /* this is an unmatched trail code unit (2nd surrogate) */
//                     /* callback(illegal) */
//                     *pErrorCode=U_ILLEGAL_CHAR_FOUND;
//                     goto endloop;
//                 }

//                 /* compress supplementary character */
//                 if( (window=getWindow(scsu->fromUDynamicOffsets, c))>=0 &&
//                     !(source<sourceLimit && (uint32_t)(*source-0x3400)<(0xd800-0x3400))
//                 ) {
//                     /*
//                      * there is a dynamic window that contains this character and
//                      * the following character is not uncompressible,
//                      * change to the window
//                      */
//                     isSingleByteMode=TRUE;
//                     dynamicWindow=window;
//                     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow];
//                     useDynamicWindow(scsu, dynamicWindow);
//                     c=((uint32_t)(UC0+dynamicWindow)<<8)|(c-currentOffset)|0x80;
//                     length=2;
//                     goto outputBytes;
//                 } else if(source<sourceLimit && lead==*source && /* too lazy to check trail in same window as source[1] */
//                           (code=getDynamicOffset(c, &offset))>=0
//                 ) {
//                     /* two supplementary characters in (probably) the same window - define an extended one */
//                     isSingleByteMode=TRUE;
//                     code-=0x200;
//                     dynamicWindow=getNextDynamicWindow(scsu);
//                     currentOffset=scsu->fromUDynamicOffsets[dynamicWindow]=offset;
//                     useDynamicWindow(scsu, dynamicWindow);
//                     c=((uint32_t)UDX<<24)|((uint32_t)dynamicWindow<<21)|((uint32_t)code<<8)|(c-currentOffset)|0x80;
//                     length=4;
//                     goto outputBytes;
//                 } else {
//                     /* don't know how to compress this character, just write it directly */
//                     c=((uint32_t)lead<<16)|trail;
//                     length=4;
//                     goto outputBytes;
//                 }
//             } else /* 0xe000<=c<0xf300 */ {
//                 /* quote to avoid SCSU tags */
//                 c|=UQU<<16;
//                 length=3;
//                 goto outputBytes;
//             }

//             /* normal end of conversion: prepare for a new character */
//             c=0;
//         }
//     }
// endloop:

//     /* set the converter state back into UConverter */
//     scsu->fromUIsSingleByteMode=isSingleByteMode;
//     scsu->fromUDynamicWindow=dynamicWindow;

//     cnv->fromUChar32=c;

//     /* write back the updated pointers */
//     pArgs->source=source;
//     pArgs->target=(char *)target;
//     return;

// outputBytes:
//     /* write the output character bytes from c and length [code copied from ucnvmbcs.c] */
//     /* from the first if in the loop we know that targetCapacity>0 */
//     if(length<=targetCapacity) {
//         switch(length) {
//             /* each branch falls through to the next one */
//         case 4:
//             *target++=(uint8_t)(c>>24);
//             U_FALLTHROUGH;
//         case 3:
//             *target++=(uint8_t)(c>>16);
//             U_FALLTHROUGH;
//         case 2:
//             *target++=(uint8_t)(c>>8);
//             U_FALLTHROUGH;
//         case 1:
//             *target++=(uint8_t)c;
//             U_FALLTHROUGH;
//         default:
//             /* will never occur */
//             break;
//         }
//         targetCapacity-=length;

//         /* normal end of conversion: prepare for a new character */
//         c=0;
//         goto loop;
//     } else {
//         uint8_t *p;

//         /*
//          * We actually do this backwards here:
//          * In order to save an intermediate variable, we output
//          * first to the overflow buffer what does not fit into the
//          * regular target.
//          */
//         /* we know that 0<=targetCapacity<length<=4 */
//         /* targetCapacity==0 when SCU+supplementary where SCU used up targetCapacity==1 */
//         length-=targetCapacity;
//         p=(uint8_t *)cnv->charErrorBuffer;
//         switch(length) {
//             /* each branch falls through to the next one */
//         case 4:
//             *p++=(uint8_t)(c>>24);
//             U_FALLTHROUGH;
//         case 3:
//             *p++=(uint8_t)(c>>16);
//             U_FALLTHROUGH;
//         case 2:
//             *p++=(uint8_t)(c>>8);
//             U_FALLTHROUGH;
//         case 1:
//             *p=(uint8_t)c;
//             U_FALLTHROUGH;
//         default:
//             /* will never occur */
//             break;
//         }
//         cnv->charErrorBufferLength=(int8_t)length;

//         /* now output what fits into the regular target */
//         c>>=8*length; /* length was reduced by targetCapacity */
//         switch(targetCapacity) {
//             /* each branch falls through to the next one */
//         case 3:
//             *target++=(uint8_t)(c>>16);
//             U_FALLTHROUGH;
//         case 2:
//             *target++=(uint8_t)(c>>8);
//             U_FALLTHROUGH;
//         case 1:
//             *target++=(uint8_t)c;
//             U_FALLTHROUGH;
//         default:
//             break;
//         }

//         /* target overflow */
//         targetCapacity=0;
//         *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
//         c=0;
//         goto endloop;
//     }
// }

// /* miscellaneous ------------------------------------------------------------ */

// static const char *  U_CALLCONV
// _SCSUGetName(const UConverter *cnv) {
//     SCSUData *scsu=(SCSUData *)cnv->extraInfo;

//     switch(scsu->locale) {
//     case l_ja:
//         return "SCSU,locale=ja";
//     default:
//         return "SCSU";
//     }
// }

// /* structure for SafeClone calculations */
// struct cloneSCSUStruct
// {
//     UConverter cnv;
//     SCSUData mydata;
// };

// static UConverter *  U_CALLCONV
// _SCSUSafeClone(const UConverter *cnv,
//                void *stackBuffer,
//                int32_t *pBufferSize,
//                UErrorCode *status)
// {
//     struct cloneSCSUStruct * localClone;
//     int32_t bufferSizeNeeded = sizeof(struct cloneSCSUStruct);

//     if (U_FAILURE(*status)){
//         return 0;
//     }

//     if (*pBufferSize == 0){ /* 'preflighting' request - set needed size into *pBufferSize */
//         *pBufferSize = bufferSizeNeeded;
//         return 0;
//     }

//     localClone = (struct cloneSCSUStruct *)stackBuffer;
//     /* ucnv.c/ucnv_safeClone() copied the main UConverter already */

//     uprv_memcpy(&localClone->mydata, cnv->extraInfo, sizeof(SCSUData));
//     localClone->cnv.extraInfo = &localClone->mydata;
//     localClone->cnv.isExtraLocal = TRUE;

//     return &localClone->cnv;
// }
// U_CDECL_END

// static const UConverterImpl _SCSUImpl={
//     UCNV_SCSU,

//     NULL,
//     NULL,

//     _SCSUOpen,
//     _SCSUClose,
//     _SCSUReset,

//     _SCSUToUnicode,
//     _SCSUToUnicodeWithOffsets,
//     _SCSUFromUnicode,
//     _SCSUFromUnicodeWithOffsets,
//     NULL,

//     NULL,
//     _SCSUGetName,
//     NULL,
//     _SCSUSafeClone,
//     ucnv_getCompleteUnicodeSet,
//     NULL,
//     NULL
// };

// static const UConverterStaticData _SCSUStaticData={
//     sizeof(UConverterStaticData),
//     "SCSU",
//     1212, /* CCSID for SCSU */
//     UCNV_IBM, UCNV_SCSU,
//     1, 3, /* one UChar generates at least 1 byte and at most 3 bytes */
//     /*
//      * The subchar here is ignored because _SCSUOpen() sets U+fffd as a Unicode
//      * substitution string.
//      */
//     { 0x0e, 0xff, 0xfd, 0 }, 3,
//     FALSE, FALSE,
//     0,
//     0,
//     { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } /* reserved */
// };

// const UConverterSharedData _SCSUData=
//         UCNV_IMMUTABLE_SHARED_DATA_INITIALIZER(&_SCSUStaticData, &_SCSUImpl);

// #endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// *******************************************************************************
// *
// *   Copyright (C) 2008-2011, International Business Machines
// *   Corporation, Google and others.  All Rights Reserved.
// *
// *******************************************************************************
// */
// // Author : eldawy@google.com (Mohamed Eldawy)
// // ucnvsel.cpp
// //
// // Purpose: To generate a list of encodings capable of handling
// // a given Unicode text
// //
// // Started 09-April-2008

// /**
//  * \file
//  *
//  * This is an implementation of an encoding selector.
//  * The goal is, given a unicode string, find the encodings
//  * this string can be mapped to. To make processing faster
//  * a trie is built when you call ucnvsel_open() that
//  * stores all encodings a codepoint can map to
//  */

// #include "unicode/ucnvsel.h"

// #if !UCONFIG_NO_CONVERSION

// #include <string.h>

// #include "unicode/uchar.h"
// #include "unicode/uniset.h"
// #include "unicode/ucnv.h"
// #include "unicode/ustring.h"
// #include "unicode/uchriter.h"
// #include "utrie2.h"
// #include "propsvec.h"
// #include "uassert.h"
// #include "ucmndata.h"
// #include "udataswp.h"
// #include "uenumimp.h"
// #include "cmemory.h"
// #include "cstring.h"

// U_NAMESPACE_USE

// struct UConverterSelector {
//   UTrie2 *trie;              // 16 bit trie containing offsets into pv
//   uint32_t* pv;              // table of bits!
//   int32_t pvCount;
//   char** encodings;          // which encodings did user ask to use?
//   int32_t encodingsCount;
//   int32_t encodingStrLength;
//   uint8_t* swapped;
//   UBool ownPv, ownEncodingStrings;
// };

// static void generateSelectorData(UConverterSelector* result,
//                                  UPropsVectors *upvec,
//                                  const USet* excludedCodePoints,
//                                  const UConverterUnicodeSet whichSet,
//                                  UErrorCode* status) {
//   if (U_FAILURE(*status)) {
//     return;
//   }

//   int32_t columns = (result->encodingsCount+31)/32;

//   // set errorValue to all-ones
//   for (int32_t col = 0; col < columns; col++) {
//     upvec_setValue(upvec, UPVEC_ERROR_VALUE_CP, UPVEC_ERROR_VALUE_CP,
//                    col, static_cast<uint32_t>(~0), static_cast<uint32_t>(~0), status);
//   }

//   for (int32_t i = 0; i < result->encodingsCount; ++i) {
//     uint32_t mask;
//     uint32_t column;
//     int32_t item_count;
//     int32_t j;
//     UConverter* test_converter = ucnv_open(result->encodings[i], status);
//     if (U_FAILURE(*status)) {
//       return;
//     }
//     USet* unicode_point_set;
//     unicode_point_set = uset_open(1, 0);  // empty set

//     ucnv_getUnicodeSet(test_converter, unicode_point_set,
//                        whichSet, status);
//     if (U_FAILURE(*status)) {
//       ucnv_close(test_converter);
//       return;
//     }

//     column = i / 32;
//     mask = 1 << (i%32);
//     // now iterate over intervals on set i!
//     item_count = uset_getItemCount(unicode_point_set);

//     for (j = 0; j < item_count; ++j) {
//       UChar32 start_char;
//       UChar32 end_char;
//       UErrorCode smallStatus = U_ZERO_ERROR;
//       uset_getItem(unicode_point_set, j, &start_char, &end_char, NULL, 0,
//                    &smallStatus);
//       if (U_FAILURE(smallStatus)) {
//         // this will be reached for the converters that fill the set with
//         // strings. Those should be ignored by our system
//       } else {
//         upvec_setValue(upvec, start_char, end_char, column, static_cast<uint32_t>(~0), mask,
//                        status);
//       }
//     }
//     ucnv_close(test_converter);
//     uset_close(unicode_point_set);
//     if (U_FAILURE(*status)) {
//       return;
//     }
//   }

//   // handle excluded encodings! Simply set their values to all 1's in the upvec
//   if (excludedCodePoints) {
//     int32_t item_count = uset_getItemCount(excludedCodePoints);
//     for (int32_t j = 0; j < item_count; ++j) {
//       UChar32 start_char;
//       UChar32 end_char;

//       uset_getItem(excludedCodePoints, j, &start_char, &end_char, NULL, 0,
//                    status);
//       for (int32_t col = 0; col < columns; col++) {
//         upvec_setValue(upvec, start_char, end_char, col, static_cast<uint32_t>(~0), static_cast<uint32_t>(~0),
//                       status);
//       }
//     }
//   }

//   // alright. Now, let's put things in the same exact form you'd get when you
//   // unserialize things.
//   result->trie = upvec_compactToUTrie2WithRowIndexes(upvec, status);
//   result->pv = upvec_cloneArray(upvec, &result->pvCount, NULL, status);
//   result->pvCount *= columns;  // number of uint32_t = rows * columns
//   result->ownPv = TRUE;
// }

// /* open a selector. If converterListSize is 0, build for all converters.
//    If excludedCodePoints is NULL, don't exclude any codepoints */
// U_CAPI UConverterSelector* U_EXPORT2
// ucnvsel_open(const char* const*  converterList, int32_t converterListSize,
//              const USet* excludedCodePoints,
//              const UConverterUnicodeSet whichSet, UErrorCode* status) {
//   // check if already failed
//   if (U_FAILURE(*status)) {
//     return NULL;
//   }
//   // ensure args make sense!
//   if (converterListSize < 0 || (converterList == NULL && converterListSize != 0)) {
//     *status = U_ILLEGAL_ARGUMENT_ERROR;
//     return NULL;
//   }

//   // allocate a new converter
//   LocalUConverterSelectorPointer newSelector(
//     (UConverterSelector*)uprv_malloc(sizeof(UConverterSelector)));
//   if (newSelector.isNull()) {
//     *status = U_MEMORY_ALLOCATION_ERROR;
//     return NULL;
//   }
//   uprv_memset(newSelector.getAlias(), 0, sizeof(UConverterSelector));

//   if (converterListSize == 0) {
//     converterList = NULL;
//     converterListSize = ucnv_countAvailable();
//   }
//   newSelector->encodings =
//     (char**)uprv_malloc(converterListSize * sizeof(char*));
//   if (!newSelector->encodings) {
//     *status = U_MEMORY_ALLOCATION_ERROR;
//     return NULL;
//   }
//   newSelector->encodings[0] = NULL;  // now we can call ucnvsel_close()

//   // make a backup copy of the list of converters
//   int32_t totalSize = 0;
//   int32_t i;
//   for (i = 0; i < converterListSize; i++) {
//     totalSize +=
//       (int32_t)uprv_strlen(converterList != NULL ? converterList[i] : ucnv_getAvailableName(i)) + 1;
//   }
//   // 4-align the totalSize to 4-align the size of the serialized form
//   int32_t encodingStrPadding = totalSize & 3;
//   if (encodingStrPadding != 0) {
//     encodingStrPadding = 4 - encodingStrPadding;
//   }
//   newSelector->encodingStrLength = totalSize += encodingStrPadding;
//   char* allStrings = (char*) uprv_malloc(totalSize);
//   if (!allStrings) {
//     *status = U_MEMORY_ALLOCATION_ERROR;
//     return NULL;
//   }

//   for (i = 0; i < converterListSize; i++) {
//     newSelector->encodings[i] = allStrings;
//     uprv_strcpy(newSelector->encodings[i],
//                 converterList != NULL ? converterList[i] : ucnv_getAvailableName(i));
//     allStrings += uprv_strlen(newSelector->encodings[i]) + 1;
//   }
//   while (encodingStrPadding > 0) {
//     *allStrings++ = 0;
//     --encodingStrPadding;
//   }

//   newSelector->ownEncodingStrings = TRUE;
//   newSelector->encodingsCount = converterListSize;
//   UPropsVectors *upvec = upvec_open((converterListSize+31)/32, status);
//   generateSelectorData(newSelector.getAlias(), upvec, excludedCodePoints, whichSet, status);
//   upvec_close(upvec);

//   if (U_FAILURE(*status)) {
//     return NULL;
//   }

//   return newSelector.orphan();
// }

// /* close opened selector */
// U_CAPI void U_EXPORT2
// ucnvsel_close(UConverterSelector *sel) {
//   if (!sel) {
//     return;
//   }
//   if (sel->ownEncodingStrings) {
//     uprv_free(sel->encodings[0]);
//   }
//   uprv_free(sel->encodings);
//   if (sel->ownPv) {
//     uprv_free(sel->pv);
//   }
//   utrie2_close(sel->trie);
//   uprv_free(sel->swapped);
//   uprv_free(sel);
// }

// static const UDataInfo dataInfo = {
//   sizeof(UDataInfo),
//   0,

//   U_IS_BIG_ENDIAN,
//   U_CHARSET_FAMILY,
//   U_SIZEOF_UCHAR,
//   0,

//   { 0x43, 0x53, 0x65, 0x6c },   /* dataFormat="CSel" */
//   { 1, 0, 0, 0 },               /* formatVersion */
//   { 0, 0, 0, 0 }                /* dataVersion */
// };

// enum {
//   UCNVSEL_INDEX_TRIE_SIZE,      // trie size in bytes
//   UCNVSEL_INDEX_PV_COUNT,       // number of uint32_t in the bit vectors
//   UCNVSEL_INDEX_NAMES_COUNT,    // number of encoding names
//   UCNVSEL_INDEX_NAMES_LENGTH,   // number of encoding name bytes including padding
//   UCNVSEL_INDEX_SIZE = 15,      // bytes following the DataHeader
//   UCNVSEL_INDEX_COUNT = 16
// };

// /*
//  * Serialized form of a UConverterSelector, formatVersion 1:
//  *
//  * The serialized form begins with a standard ICU DataHeader with a UDataInfo
//  * as the template above.
//  * This is followed by:
//  *   int32_t indexes[UCNVSEL_INDEX_COUNT];          // see index entry constants above
//  *   serialized UTrie2;                             // indexes[UCNVSEL_INDEX_TRIE_SIZE] bytes
//  *   uint32_t pv[indexes[UCNVSEL_INDEX_PV_COUNT]];  // bit vectors
//  *   char* encodingNames[indexes[UCNVSEL_INDEX_NAMES_LENGTH]];  // NUL-terminated strings + padding
//  */

// /* serialize a selector */
// U_CAPI int32_t U_EXPORT2
// ucnvsel_serialize(const UConverterSelector* sel,
//                   void* buffer, int32_t bufferCapacity, UErrorCode* status) {
//   // check if already failed
//   if (U_FAILURE(*status)) {
//     return 0;
//   }
//   // ensure args make sense!
//   uint8_t *p = (uint8_t *)buffer;
//   if (bufferCapacity < 0 ||
//       (bufferCapacity > 0 && (p == NULL || (U_POINTER_MASK_LSB(p, 3) != 0)))
//   ) {
//     *status = U_ILLEGAL_ARGUMENT_ERROR;
//     return 0;
//   }
//   // add up the size of the serialized form
//   int32_t serializedTrieSize = utrie2_serialize(sel->trie, NULL, 0, status);
//   if (*status != U_BUFFER_OVERFLOW_ERROR && U_FAILURE(*status)) {
//     return 0;
//   }
//   *status = U_ZERO_ERROR;

//   DataHeader header;
//   uprv_memset(&header, 0, sizeof(header));
//   header.dataHeader.headerSize = (uint16_t)((sizeof(header) + 15) & ~15);
//   header.dataHeader.magic1 = 0xda;
//   header.dataHeader.magic2 = 0x27;
//   uprv_memcpy(&header.info, &dataInfo, sizeof(dataInfo));

//   int32_t indexes[UCNVSEL_INDEX_COUNT] = {
//     serializedTrieSize,
//     sel->pvCount,
//     sel->encodingsCount,
//     sel->encodingStrLength
//   };

//   int32_t totalSize =
//     header.dataHeader.headerSize +
//     (int32_t)sizeof(indexes) +
//     serializedTrieSize +
//     sel->pvCount * 4 +
//     sel->encodingStrLength;
//   indexes[UCNVSEL_INDEX_SIZE] = totalSize - header.dataHeader.headerSize;
//   if (totalSize > bufferCapacity) {
//     *status = U_BUFFER_OVERFLOW_ERROR;
//     return totalSize;
//   }
//   // ok, save!
//   int32_t length = header.dataHeader.headerSize;
//   uprv_memcpy(p, &header, sizeof(header));
//   uprv_memset(p + sizeof(header), 0, length - sizeof(header));
//   p += length;

//   length = (int32_t)sizeof(indexes);
//   uprv_memcpy(p, indexes, length);
//   p += length;

//   utrie2_serialize(sel->trie, p, serializedTrieSize, status);
//   p += serializedTrieSize;

//   length = sel->pvCount * 4;
//   uprv_memcpy(p, sel->pv, length);
//   p += length;

//   uprv_memcpy(p, sel->encodings[0], sel->encodingStrLength);
//   p += sel->encodingStrLength;

//   return totalSize;
// }

// /**
//  * swap a selector into the desired Endianness and Asciiness of
//  * the system. Just as FYI, selectors are always saved in the format
//  * of the system that created them. They are only converted if used
//  * on another system. In other words, selectors created on different
//  * system can be different even if the params are identical (endianness
//  * and Asciiness differences only)
//  *
//  * @param ds pointer to data swapper containing swapping info
//  * @param inData pointer to incoming data
//  * @param length length of inData in bytes
//  * @param outData pointer to output data. Capacity should
//  *                be at least equal to capacity of inData
//  * @param status an in/out ICU UErrorCode
//  * @return 0 on failure, number of bytes swapped on success
//  *         number of bytes swapped can be smaller than length
//  */
// static int32_t
// ucnvsel_swap(const UDataSwapper *ds,
//              const void *inData, int32_t length,
//              void *outData, UErrorCode *status) {
//   /* udata_swapDataHeader checks the arguments */
//   int32_t headerSize = udata_swapDataHeader(ds, inData, length, outData, status);
//   if(U_FAILURE(*status)) {
//     return 0;
//   }

//   /* check data format and format version */
//   const UDataInfo *pInfo = (const UDataInfo *)((const char *)inData + 4);
//   if(!(
//     pInfo->dataFormat[0] == 0x43 &&  /* dataFormat="CSel" */
//     pInfo->dataFormat[1] == 0x53 &&
//     pInfo->dataFormat[2] == 0x65 &&
//     pInfo->dataFormat[3] == 0x6c
//   )) {
//     udata_printError(ds, "ucnvsel_swap(): data format %02x.%02x.%02x.%02x is not recognized as UConverterSelector data\n",
//                      pInfo->dataFormat[0], pInfo->dataFormat[1],
//                      pInfo->dataFormat[2], pInfo->dataFormat[3]);
//     *status = U_INVALID_FORMAT_ERROR;
//     return 0;
//   }
//   if(pInfo->formatVersion[0] != 1) {
//     udata_printError(ds, "ucnvsel_swap(): format version %02x is not supported\n",
//                      pInfo->formatVersion[0]);
//     *status = U_UNSUPPORTED_ERROR;
//     return 0;
//   }

//   if(length >= 0) {
//     length -= headerSize;
//     if(length < 16*4) {
//       udata_printError(ds, "ucnvsel_swap(): too few bytes (%d after header) for UConverterSelector data\n",
//                        length);
//       *status = U_INDEX_OUTOFBOUNDS_ERROR;
//       return 0;
//     }
//   }

//   const uint8_t *inBytes = (const uint8_t *)inData + headerSize;
//   uint8_t *outBytes = (uint8_t *)outData + headerSize;

//   /* read the indexes */
//   const int32_t *inIndexes = (const int32_t *)inBytes;
//   int32_t indexes[16];
//   int32_t i;
//   for(i = 0; i < 16; ++i) {
//     indexes[i] = udata_readInt32(ds, inIndexes[i]);
//   }

//   /* get the total length of the data */
//   int32_t size = indexes[UCNVSEL_INDEX_SIZE];
//   if(length >= 0) {
//     if(length < size) {
//       udata_printError(ds, "ucnvsel_swap(): too few bytes (%d after header) for all of UConverterSelector data\n",
//                        length);
//       *status = U_INDEX_OUTOFBOUNDS_ERROR;
//       return 0;
//     }

//     /* copy the data for inaccessible bytes */
//     if(inBytes != outBytes) {
//       uprv_memcpy(outBytes, inBytes, size);
//     }

//     int32_t offset = 0, count;

//     /* swap the int32_t indexes[] */
//     count = UCNVSEL_INDEX_COUNT*4;
//     ds->swapArray32(ds, inBytes, count, outBytes, status);
//     offset += count;

//     /* swap the UTrie2 */
//     count = indexes[UCNVSEL_INDEX_TRIE_SIZE];
//     utrie2_swap(ds, inBytes + offset, count, outBytes + offset, status);
//     offset += count;

//     /* swap the uint32_t pv[] */
//     count = indexes[UCNVSEL_INDEX_PV_COUNT]*4;
//     ds->swapArray32(ds, inBytes + offset, count, outBytes + offset, status);
//     offset += count;

//     /* swap the encoding names */
//     count = indexes[UCNVSEL_INDEX_NAMES_LENGTH];
//     ds->swapInvChars(ds, inBytes + offset, count, outBytes + offset, status);
//     offset += count;

//     U_ASSERT(offset == size);
//   }

//   return headerSize + size;
// }

// /* unserialize a selector */
// U_CAPI UConverterSelector* U_EXPORT2
// ucnvsel_openFromSerialized(const void* buffer, int32_t length, UErrorCode* status) {
//   // check if already failed
//   if (U_FAILURE(*status)) {
//     return NULL;
//   }
//   // ensure args make sense!
//   const uint8_t *p = (const uint8_t *)buffer;
//   if (length <= 0 ||
//       (length > 0 && (p == NULL || (U_POINTER_MASK_LSB(p, 3) != 0)))
//   ) {
//     *status = U_ILLEGAL_ARGUMENT_ERROR;
//     return NULL;
//   }
//   // header
//   if (length < 32) {
//     // not even enough space for a minimal header
//     *status = U_INDEX_OUTOFBOUNDS_ERROR;
//     return NULL;
//   }
//   const DataHeader *pHeader = (const DataHeader *)p;
//   if (!(
//     pHeader->dataHeader.magic1==0xda &&
//     pHeader->dataHeader.magic2==0x27 &&
//     pHeader->info.dataFormat[0] == 0x43 &&
//     pHeader->info.dataFormat[1] == 0x53 &&
//     pHeader->info.dataFormat[2] == 0x65 &&
//     pHeader->info.dataFormat[3] == 0x6c
//   )) {
//     /* header not valid or dataFormat not recognized */
//     *status = U_INVALID_FORMAT_ERROR;
//     return NULL;
//   }
//   if (pHeader->info.formatVersion[0] != 1) {
//     *status = U_UNSUPPORTED_ERROR;
//     return NULL;
//   }
//   uint8_t* swapped = NULL;
//   if (pHeader->info.isBigEndian != U_IS_BIG_ENDIAN ||
//       pHeader->info.charsetFamily != U_CHARSET_FAMILY
//   ) {
//     // swap the data
//     UDataSwapper *ds =
//       udata_openSwapperForInputData(p, length, U_IS_BIG_ENDIAN, U_CHARSET_FAMILY, status);
//     int32_t totalSize = ucnvsel_swap(ds, p, -1, NULL, status);
//     if (U_FAILURE(*status)) {
//       udata_closeSwapper(ds);
//       return NULL;
//     }
//     if (length < totalSize) {
//       udata_closeSwapper(ds);
//       *status = U_INDEX_OUTOFBOUNDS_ERROR;
//       return NULL;
//     }
//     swapped = (uint8_t*)uprv_malloc(totalSize);
//     if (swapped == NULL) {
//       udata_closeSwapper(ds);
//       *status = U_MEMORY_ALLOCATION_ERROR;
//       return NULL;
//     }
//     ucnvsel_swap(ds, p, length, swapped, status);
//     udata_closeSwapper(ds);
//     if (U_FAILURE(*status)) {
//       uprv_free(swapped);
//       return NULL;
//     }
//     p = swapped;
//     pHeader = (const DataHeader *)p;
//   }
//   if (length < (pHeader->dataHeader.headerSize + 16 * 4)) {
//     // not even enough space for the header and the indexes
//     uprv_free(swapped);
//     *status = U_INDEX_OUTOFBOUNDS_ERROR;
//     return NULL;
//   }
//   p += pHeader->dataHeader.headerSize;
//   length -= pHeader->dataHeader.headerSize;
//   // indexes
//   const int32_t *indexes = (const int32_t *)p;
//   if (length < indexes[UCNVSEL_INDEX_SIZE]) {
//     uprv_free(swapped);
//     *status = U_INDEX_OUTOFBOUNDS_ERROR;
//     return NULL;
//   }
//   p += UCNVSEL_INDEX_COUNT * 4;
//   // create and populate the selector object
//   UConverterSelector* sel = (UConverterSelector*)uprv_malloc(sizeof(UConverterSelector));
//   char **encodings =
//     (char **)uprv_malloc(
//       indexes[UCNVSEL_INDEX_NAMES_COUNT] * sizeof(char *));
//   if (sel == NULL || encodings == NULL) {
//     uprv_free(swapped);
//     uprv_free(sel);
//     uprv_free(encodings);
//     *status = U_MEMORY_ALLOCATION_ERROR;
//     return NULL;
//   }
//   uprv_memset(sel, 0, sizeof(UConverterSelector));
//   sel->pvCount = indexes[UCNVSEL_INDEX_PV_COUNT];
//   sel->encodings = encodings;
//   sel->encodingsCount = indexes[UCNVSEL_INDEX_NAMES_COUNT];
//   sel->encodingStrLength = indexes[UCNVSEL_INDEX_NAMES_LENGTH];
//   sel->swapped = swapped;
//   // trie
//   sel->trie = utrie2_openFromSerialized(UTRIE2_16_VALUE_BITS,
//                                         p, indexes[UCNVSEL_INDEX_TRIE_SIZE], NULL,
//                                         status);
//   p += indexes[UCNVSEL_INDEX_TRIE_SIZE];
//   if (U_FAILURE(*status)) {
//     ucnvsel_close(sel);
//     return NULL;
//   }
//   // bit vectors
//   sel->pv = (uint32_t *)p;
//   p += sel->pvCount * 4;
//   // encoding names
//   char* s = (char*)p;
//   for (int32_t i = 0; i < sel->encodingsCount; ++i) {
//     sel->encodings[i] = s;
//     s += uprv_strlen(s) + 1;
//   }
//   p += sel->encodingStrLength;

//   return sel;
// }

// // a bunch of functions for the enumeration thingie! Nothing fancy here. Just
// // iterate over the selected encodings
// struct Enumerator {
//   int16_t* index;
//   int16_t length;
//   int16_t cur;
//   const UConverterSelector* sel;
// };

// U_CDECL_BEGIN

// static void U_CALLCONV
// ucnvsel_close_selector_iterator(UEnumeration *enumerator) {
//   uprv_free(((Enumerator*)(enumerator->context))->index);
//   uprv_free(enumerator->context);
//   uprv_free(enumerator);
// }


// static int32_t U_CALLCONV
// ucnvsel_count_encodings(UEnumeration *enumerator, UErrorCode *status) {
//   // check if already failed
//   if (U_FAILURE(*status)) {
//     return 0;
//   }
//   return ((Enumerator*)(enumerator->context))->length;
// }


// static const char* U_CALLCONV ucnvsel_next_encoding(UEnumeration* enumerator,
//                                                  int32_t* resultLength,
//                                                  UErrorCode* status) {
//   // check if already failed
//   if (U_FAILURE(*status)) {
//     return NULL;
//   }

//   int16_t cur = ((Enumerator*)(enumerator->context))->cur;
//   const UConverterSelector* sel;
//   const char* result;
//   if (cur >= ((Enumerator*)(enumerator->context))->length) {
//     return NULL;
//   }
//   sel = ((Enumerator*)(enumerator->context))->sel;
//   result = sel->encodings[((Enumerator*)(enumerator->context))->index[cur] ];
//   ((Enumerator*)(enumerator->context))->cur++;
//   if (resultLength) {
//     *resultLength = (int32_t)uprv_strlen(result);
//   }
//   return result;
// }

// static void U_CALLCONV ucnvsel_reset_iterator(UEnumeration* enumerator,
//                                            UErrorCode* status) {
//   // check if already failed
//   if (U_FAILURE(*status)) {
//     return ;
//   }
//   ((Enumerator*)(enumerator->context))->cur = 0;
// }

// U_CDECL_END


// static const UEnumeration defaultEncodings = {
//   NULL,
//     NULL,
//     ucnvsel_close_selector_iterator,
//     ucnvsel_count_encodings,
//     uenum_unextDefault,
//     ucnvsel_next_encoding,
//     ucnvsel_reset_iterator
// };


// // internal fn to intersect two sets of masks
// // returns whether the mask has reduced to all zeros
// static UBool intersectMasks(uint32_t* dest, const uint32_t* source1, int32_t len) {
//   int32_t i;
//   uint32_t oredDest = 0;
//   for (i = 0 ; i < len ; ++i) {
//     oredDest |= (dest[i] &= source1[i]);
//   }
//   return oredDest == 0;
// }

// // internal fn to count how many 1's are there in a mask
// // algorithm taken from  http://graphics.stanford.edu/~seander/bithacks.html
// static int16_t countOnes(uint32_t* mask, int32_t len) {
//   int32_t i, totalOnes = 0;
//   for (i = 0 ; i < len ; ++i) {
//     uint32_t ent = mask[i];
//     for (; ent; totalOnes++)
//     {
//       ent &= ent - 1; // clear the least significant bit set
//     }
//   }
//   return static_cast<int16_t>(totalOnes);
// }


// /* internal function! */
// static UEnumeration *selectForMask(const UConverterSelector* sel,
//                                    uint32_t *theMask, UErrorCode *status) {
//   LocalMemory<uint32_t> mask(theMask);
//   // this is the context we will use. Store a table of indices to which
//   // encodings are legit.
//   LocalMemory<Enumerator> result(static_cast<Enumerator *>(uprv_malloc(sizeof(Enumerator))));
//   if (result.isNull()) {
//     *status = U_MEMORY_ALLOCATION_ERROR;
//     return nullptr;
//   }
//   result->index = nullptr;  // this will be allocated later!
//   result->length = result->cur = 0;
//   result->sel = sel;

//   LocalMemory<UEnumeration> en(static_cast<UEnumeration *>(uprv_malloc(sizeof(UEnumeration))));
//   if (en.isNull()) {
//     // TODO(markus): Combine Enumerator and UEnumeration into one struct.
//     *status = U_MEMORY_ALLOCATION_ERROR;
//     return nullptr;
//   }
//   memcpy(en.getAlias(), &defaultEncodings, sizeof(UEnumeration));

//   int32_t columns = (sel->encodingsCount+31)/32;
//   int16_t numOnes = countOnes(mask.getAlias(), columns);
//   // now, we know the exact space we need for index
//   if (numOnes > 0) {
//     result->index = static_cast<int16_t*>(uprv_malloc(numOnes * sizeof(int16_t)));
//     if (result->index == nullptr) {
//       *status = U_MEMORY_ALLOCATION_ERROR;
//       return nullptr;
//     }
//     int32_t i, j;
//     int16_t k = 0;
//     for (j = 0 ; j < columns; j++) {
//       uint32_t v = mask[j];
//       for (i = 0 ; i < 32 && k < sel->encodingsCount; i++, k++) {
//         if ((v & 1) != 0) {
//           result->index[result->length++] = k;
//         }
//         v >>= 1;
//       }
//     }
//   } //otherwise, index will remain NULL (and will never be touched by
//     //the enumerator code anyway)
//   en->context = result.orphan();
//   return en.orphan();
// }

// /* check a string against the selector - UTF16 version */
// U_CAPI UEnumeration * U_EXPORT2
// ucnvsel_selectForString(const UConverterSelector* sel,
//                         const UChar *s, int32_t length, UErrorCode *status) {
//   // check if already failed
//   if (U_FAILURE(*status)) {
//     return NULL;
//   }
//   // ensure args make sense!
//   if (sel == NULL || (s == NULL && length != 0)) {
//     *status = U_ILLEGAL_ARGUMENT_ERROR;
//     return NULL;
//   }

//   int32_t columns = (sel->encodingsCount+31)/32;
//   uint32_t* mask = (uint32_t*) uprv_malloc(columns * 4);
//   if (mask == NULL) {
//     *status = U_MEMORY_ALLOCATION_ERROR;
//     return NULL;
//   }
//   uprv_memset(mask, ~0, columns *4);

//   if(s!=NULL) {
//     const UChar *limit;
//     if (length >= 0) {
//       limit = s + length;
//     } else {
//       limit = NULL;
//     }

//     while (limit == NULL ? *s != 0 : s != limit) {
//       UChar32 c;
//       uint16_t pvIndex;
//       UTRIE2_U16_NEXT16(sel->trie, s, limit, c, pvIndex);
//       if (intersectMasks(mask, sel->pv+pvIndex, columns)) {
//         break;
//       }
//     }
//   }
//   return selectForMask(sel, mask, status);
// }

// /* check a string against the selector - UTF8 version */
// U_CAPI UEnumeration * U_EXPORT2
// ucnvsel_selectForUTF8(const UConverterSelector* sel,
//                       const char *s, int32_t length, UErrorCode *status) {
//   // check if already failed
//   if (U_FAILURE(*status)) {
//     return NULL;
//   }
//   // ensure args make sense!
//   if (sel == NULL || (s == NULL && length != 0)) {
//     *status = U_ILLEGAL_ARGUMENT_ERROR;
//     return NULL;
//   }

//   int32_t columns = (sel->encodingsCount+31)/32;
//   uint32_t* mask = (uint32_t*) uprv_malloc(columns * 4);
//   if (mask == NULL) {
//     *status = U_MEMORY_ALLOCATION_ERROR;
//     return NULL;
//   }
//   uprv_memset(mask, ~0, columns *4);

//   if (length < 0) {
//     length = (int32_t)uprv_strlen(s);
//   }

//   if(s!=NULL) {
//     const char *limit = s + length;

//     while (s != limit) {
//       uint16_t pvIndex;
//       UTRIE2_U8_NEXT16(sel->trie, s, limit, pvIndex);
//       if (intersectMasks(mask, sel->pv+pvIndex, columns)) {
//         break;
//       }
//     }
//   }
//   return selectForMask(sel, mask, status);
// }

// #endif  // !UCONFIG_NO_CONVERSION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2003-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ucol_swp.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2003sep10
*   created by: Markus W. Scherer
*
*   Swap collation binaries.
*/

 /* UDataInfo */




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2000-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  ucol_data.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011jul02
*   created by: Markus Scherer
*
* Private implementation header for C/C++ collation.
* Some file data structure definitions were moved here from i18n/ucol_imp.h
* so that the common library (via ucol_swp.cpp) need not depend on the i18n library at all.
*
* We do not want to move the collation swapper to the i18n library because
* a) the resource bundle swapper depends on it and would have to move too, and
* b) we might want to eventually implement runtime data swapping,
*    which might (or might not) be easier if all swappers are in the common library.
*/




#if !UCONFIG_NO_COLLATION

/* let us know whether reserved fields are reset to zero or junked */
#define UCOL_HEADER_MAGIC 0x20030618

typedef struct {
      int32_t size;
      /* all the offsets are in bytes */
      /* to get the address add to the header address and cast properly */
      uint32_t options; /* these are the default options for the collator */
      uint32_t UCAConsts; /* structure which holds values for indirect positioning and implicit ranges */
      uint32_t contractionUCACombos;        /* this one is needed only for UCA, to copy the appropriate contractions */
      uint32_t magic;            /* magic number - lets us know whether reserved data is reset or junked */
      uint32_t mappingPosition;  /* const uint8_t *mappingPosition; */
      uint32_t expansion;        /* uint32_t *expansion;            */
      uint32_t contractionIndex; /* UChar *contractionIndex;        */
      uint32_t contractionCEs;   /* uint32_t *contractionCEs;       */
      uint32_t contractionSize;  /* needed for various closures */
      /*int32_t latinOneMapping;*/ /* this is now handled in the trie itself *//* fast track to latin1 chars      */

      uint32_t endExpansionCE;      /* array of last collation element in
                                       expansion */
      uint32_t expansionCESize;     /* array of maximum expansion size
                                       corresponding to the expansion
                                       collation elements with last element
                                       in endExpansionCE*/
      int32_t  endExpansionCECount; /* size of endExpansionCE */
      uint32_t unsafeCP;            /* hash table of unsafe code points */
      uint32_t contrEndCP;          /* hash table of final code points  */
                                    /*   in contractions.               */

      int32_t contractionUCACombosSize;     /* number of UCA contraction items. */
                                            /*Length is contractionUCACombosSize*contractionUCACombosWidth*sizeof(UChar) */
      UBool jamoSpecial;                    /* is jamoSpecial */
      UBool isBigEndian;                    /* is this data big endian? from the UDataInfo header*/
      uint8_t charSetFamily;                /* what is the charset family of this data from the UDataInfo header*/
      uint8_t contractionUCACombosWidth;    /* width of UCA combos field */
      UVersionInfo version;
      UVersionInfo UCAVersion;              /* version of the UCA, read from file */
      UVersionInfo UCDVersion;              /* UCD version, obtained by u_getUnicodeVersion */
      UVersionInfo formatVersion;           /* format version from the UDataInfo header */
      uint32_t scriptToLeadByte;            /* offset to script to lead collation byte mapping data */
      uint32_t leadByteToScript;            /* offset to lead collation byte to script mapping data */
      uint8_t reserved[76];                 /* for future use */
} UCATableHeader;

typedef struct {
  uint32_t byteSize;
  uint32_t tableSize;
  uint32_t contsSize;
  uint32_t table;
  uint32_t conts;
  UVersionInfo UCAVersion;              /* version of the UCA, read from file */
  uint8_t padding[8];
} InverseUCATableHeader;

#endif  /* !UCONFIG_NO_COLLATION */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2003-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ucol_swp.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2003sep10
*   created by: Markus W. Scherer
*
*   Swap collation binaries.
*/




#if !UCONFIG_NO_COLLATION



/*
 * Does the data look like a collation binary?
 * @internal
 */
U_INTERNAL UBool U_EXPORT2
ucol_looksLikeCollationBinary(const UDataSwapper *ds,
                              const void *inData, int32_t length);

/**
 * Swap ICU collation data like ucadata.icu. See udataswp.h.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
ucol_swap(const UDataSwapper *ds,
          const void *inData, int32_t length, void *outData,
          UErrorCode *pErrorCode);

/**
 * Swap inverse UCA collation data (invuca.icu). See udataswp.h.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
ucol_swapInverseUCA(const UDataSwapper *ds,
                    const void *inData, int32_t length, void *outData,
                    UErrorCode *pErrorCode);

#endif /* #if !UCONFIG_NO_COLLATION */



/* swapping ----------------------------------------------------------------- */

#if !UCONFIG_NO_COLLATION

U_CAPI UBool U_EXPORT2
ucol_looksLikeCollationBinary(const UDataSwapper *ds,
                              const void *inData, int32_t length) {
    if(ds==NULL || inData==NULL || length<-1) {
        return FALSE;
    }

    // First check for format version 4+ which has a standard data header.
    UErrorCode errorCode=U_ZERO_ERROR;
    (void)udata_swapDataHeader(ds, inData, -1, NULL, &errorCode);
    if(U_SUCCESS(errorCode)) {
        const UDataInfo &info=*(const UDataInfo *)((const char *)inData+4);
        if(info.dataFormat[0]==0x55 &&   // dataFormat="UCol"
                info.dataFormat[1]==0x43 &&
                info.dataFormat[2]==0x6f &&
                info.dataFormat[3]==0x6c) {
            return TRUE;
        }
    }

    // Else check for format version 3.
    const UCATableHeader *inHeader=(const UCATableHeader *)inData;

    /*
     * The collation binary must contain at least the UCATableHeader,
     * starting with its size field.
     * sizeof(UCATableHeader)==42*4 in ICU 2.8
     * check the length against the header size before reading the size field
     */
    UCATableHeader header;
    uprv_memset(&header, 0, sizeof(header));
    if(length<0) {
        header.size=udata_readInt32(ds, inHeader->size);
    } else if((length<(42*4) || length<(header.size=udata_readInt32(ds, inHeader->size)))) {
        return FALSE;
    }

    header.magic=ds->readUInt32(inHeader->magic);
    if(!(
        header.magic==UCOL_HEADER_MAGIC &&
        inHeader->formatVersion[0]==3 /*&&
        inHeader->formatVersion[1]>=0*/
    )) {
        return FALSE;
    }

    if(inHeader->isBigEndian!=ds->inIsBigEndian || inHeader->charSetFamily!=ds->inCharset) {
        return FALSE;
    }

    return TRUE;
}

namespace {

/* swap a header-less collation formatVersion=3 binary, inside a resource bundle or ucadata.icu */
int32_t
swapFormatVersion3(const UDataSwapper *ds,
                   const void *inData, int32_t length, void *outData,
                   UErrorCode *pErrorCode) {
    const uint8_t *inBytes;
    uint8_t *outBytes;

    const UCATableHeader *inHeader;
    UCATableHeader *outHeader;
    UCATableHeader header;

    uint32_t count;

    /* argument checking in case we were not called from ucol_swap() */
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<-1 || (length>0 && outData==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    inBytes=(const uint8_t *)inData;
    outBytes=(uint8_t *)outData;

    inHeader=(const UCATableHeader *)inData;
    outHeader=(UCATableHeader *)outData;

    /*
     * The collation binary must contain at least the UCATableHeader,
     * starting with its size field.
     * sizeof(UCATableHeader)==42*4 in ICU 2.8
     * check the length against the header size before reading the size field
     */
    uprv_memset(&header, 0, sizeof(header));
    if(length<0) {
        header.size=udata_readInt32(ds, inHeader->size);
    } else if((length<(42*4) || length<(header.size=udata_readInt32(ds, inHeader->size)))) {
        udata_printError(ds, "ucol_swap(formatVersion=3): too few bytes (%d after header) for collation data\n",
                         length);
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    header.magic=ds->readUInt32(inHeader->magic);
    if(!(
        header.magic==UCOL_HEADER_MAGIC &&
        inHeader->formatVersion[0]==3 /*&&
        inHeader->formatVersion[1]>=0*/
    )) {
        udata_printError(ds, "ucol_swap(formatVersion=3): magic 0x%08x or format version %02x.%02x is not a collation binary\n",
                         header.magic,
                         inHeader->formatVersion[0], inHeader->formatVersion[1]);
        *pErrorCode=U_UNSUPPORTED_ERROR;
        return 0;
    }

    if(inHeader->isBigEndian!=ds->inIsBigEndian || inHeader->charSetFamily!=ds->inCharset) {
        udata_printError(ds, "ucol_swap(formatVersion=3): endianness %d or charset %d does not match the swapper\n",
                         inHeader->isBigEndian, inHeader->charSetFamily);
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return 0;
    }

    if(length>=0) {
        /* copy everything, takes care of data that needs no swapping */
        if(inBytes!=outBytes) {
            uprv_memcpy(outBytes, inBytes, header.size);
        }

        /* swap the necessary pieces in the order of their occurrence in the data */

        /* read more of the UCATableHeader (the size field was read above) */
        header.options=                 ds->readUInt32(inHeader->options);
        header.UCAConsts=               ds->readUInt32(inHeader->UCAConsts);
        header.contractionUCACombos=    ds->readUInt32(inHeader->contractionUCACombos);
        header.mappingPosition=         ds->readUInt32(inHeader->mappingPosition);
        header.expansion=               ds->readUInt32(inHeader->expansion);
        header.contractionIndex=        ds->readUInt32(inHeader->contractionIndex);
        header.contractionCEs=          ds->readUInt32(inHeader->contractionCEs);
        header.contractionSize=         ds->readUInt32(inHeader->contractionSize);
        header.endExpansionCE=          ds->readUInt32(inHeader->endExpansionCE);
        header.expansionCESize=         ds->readUInt32(inHeader->expansionCESize);
        header.endExpansionCECount=     udata_readInt32(ds, inHeader->endExpansionCECount);
        header.contractionUCACombosSize=udata_readInt32(ds, inHeader->contractionUCACombosSize);
        header.scriptToLeadByte=        ds->readUInt32(inHeader->scriptToLeadByte);
        header.leadByteToScript=        ds->readUInt32(inHeader->leadByteToScript);
        
        /* swap the 32-bit integers in the header */
        ds->swapArray32(ds, inHeader, (int32_t)((const char *)&inHeader->jamoSpecial-(const char *)inHeader),
                           outHeader, pErrorCode);
        ds->swapArray32(ds, &(inHeader->scriptToLeadByte), sizeof(header.scriptToLeadByte) + sizeof(header.leadByteToScript),
                           &(outHeader->scriptToLeadByte), pErrorCode);
        /* set the output platform properties */
        outHeader->isBigEndian=ds->outIsBigEndian;
        outHeader->charSetFamily=ds->outCharset;

        /* swap the options */
        if(header.options!=0) {
            ds->swapArray32(ds, inBytes+header.options, header.expansion-header.options,
                               outBytes+header.options, pErrorCode);
        }

        /* swap the expansions */
        if(header.mappingPosition!=0 && header.expansion!=0) {
            if(header.contractionIndex!=0) {
                /* expansions bounded by contractions */
                count=header.contractionIndex-header.expansion;
            } else {
                /* no contractions: expansions bounded by the main trie */
                count=header.mappingPosition-header.expansion;
            }
            ds->swapArray32(ds, inBytes+header.expansion, (int32_t)count,
                               outBytes+header.expansion, pErrorCode);
        }

        /* swap the contractions */
        if(header.contractionSize!=0) {
            /* contractionIndex: UChar[] */
            ds->swapArray16(ds, inBytes+header.contractionIndex, header.contractionSize*2,
                               outBytes+header.contractionIndex, pErrorCode);

            /* contractionCEs: CEs[] */
            ds->swapArray32(ds, inBytes+header.contractionCEs, header.contractionSize*4,
                               outBytes+header.contractionCEs, pErrorCode);
        }

        /* swap the main trie */
        if(header.mappingPosition!=0) {
            count=header.endExpansionCE-header.mappingPosition;
            utrie_swap(ds, inBytes+header.mappingPosition, (int32_t)count,
                          outBytes+header.mappingPosition, pErrorCode);
        }

        /* swap the max expansion table */
        if(header.endExpansionCECount!=0) {
            ds->swapArray32(ds, inBytes+header.endExpansionCE, header.endExpansionCECount*4,
                               outBytes+header.endExpansionCE, pErrorCode);
        }

        /* expansionCESize, unsafeCP, contrEndCP: uint8_t[], no need to swap */

        /* swap UCA constants */
        if(header.UCAConsts!=0) {
            /*
             * if UCAConsts!=0 then contractionUCACombos because we are swapping
             * the UCA data file, and we know that the UCA contains contractions
             */
            ds->swapArray32(ds, inBytes+header.UCAConsts, header.contractionUCACombos-header.UCAConsts,
                               outBytes+header.UCAConsts, pErrorCode);
        }

        /* swap UCA contractions */
        if(header.contractionUCACombosSize!=0) {
            count=header.contractionUCACombosSize*inHeader->contractionUCACombosWidth*U_SIZEOF_UCHAR;
            ds->swapArray16(ds, inBytes+header.contractionUCACombos, (int32_t)count,
                               outBytes+header.contractionUCACombos, pErrorCode);
        }
        
        /* swap the script to lead bytes */
        if(header.scriptToLeadByte!=0) {
            int indexCount = ds->readUInt16(*((uint16_t*)(inBytes+header.scriptToLeadByte))); // each entry = 2 * uint16
            int dataCount = ds->readUInt16(*((uint16_t*)(inBytes+header.scriptToLeadByte + 2))); // each entry = uint16
            ds->swapArray16(ds, inBytes+header.scriptToLeadByte, 
                                4 + (4 * indexCount) + (2 * dataCount),
                                outBytes+header.scriptToLeadByte, pErrorCode);
        }
        
        /* swap the lead byte to scripts */
        if(header.leadByteToScript!=0) {
            int indexCount = ds->readUInt16(*((uint16_t*)(inBytes+header.leadByteToScript))); // each entry = uint16
            int dataCount = ds->readUInt16(*((uint16_t*)(inBytes+header.leadByteToScript + 2))); // each entry = uint16
            ds->swapArray16(ds, inBytes+header.leadByteToScript, 
                                4 + (2 * indexCount) + (2 * dataCount),
                                outBytes+header.leadByteToScript, pErrorCode);
        }
    }

    return header.size;
}

// swap formatVersion 4 or 5 ----------------------------------------------- ***

// The following are copied from CollationDataReader, trading an awkward copy of constants
// for an awkward relocation of the i18n collationdatareader.h file into the common library.
// Keep them in sync!

enum {
    IX_INDEXES_LENGTH,  // 0
    IX_OPTIONS,
    IX_RESERVED2,
    IX_RESERVED3,

    IX_JAMO_CE32S_START,  // 4
    IX_REORDER_CODES_OFFSET,
    IX_REORDER_TABLE_OFFSET,
    IX_TRIE_OFFSET,

    IX_RESERVED8_OFFSET,  // 8
    IX_CES_OFFSET,
    IX_RESERVED10_OFFSET,
    IX_CE32S_OFFSET,

    IX_ROOT_ELEMENTS_OFFSET,  // 12
    IX_CONTEXTS_OFFSET,
    IX_UNSAFE_BWD_OFFSET,
    IX_FAST_LATIN_TABLE_OFFSET,

    IX_SCRIPTS_OFFSET,  // 16
    IX_COMPRESSIBLE_BYTES_OFFSET,
    IX_RESERVED18_OFFSET,
    IX_TOTAL_SIZE
};

int32_t
swapFormatVersion4(const UDataSwapper *ds,
                   const void *inData, int32_t length, void *outData,
                   UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }

    const uint8_t *inBytes=(const uint8_t *)inData;
    uint8_t *outBytes=(uint8_t *)outData;

    const int32_t *inIndexes=(const int32_t *)inBytes;
    int32_t indexes[IX_TOTAL_SIZE+1];

    // Need at least IX_INDEXES_LENGTH and IX_OPTIONS.
    if(0<=length && length<8) {
        udata_printError(ds, "ucol_swap(formatVersion=4): too few bytes "
                         "(%d after header) for collation data\n",
                         length);
        errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    int32_t indexesLength=indexes[0]=udata_readInt32(ds, inIndexes[0]);
    if(0<=length && length<(indexesLength*4)) {
        udata_printError(ds, "ucol_swap(formatVersion=4): too few bytes "
                         "(%d after header) for collation data\n",
                         length);
        errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    for(int32_t i=1; i<=IX_TOTAL_SIZE && i<indexesLength; ++i) {
        indexes[i]=udata_readInt32(ds, inIndexes[i]);
    }
    for(int32_t i=indexesLength; i<=IX_TOTAL_SIZE; ++i) {
        indexes[i]=-1;
    }
    inIndexes=NULL;  // Make sure we do not accidentally use these instead of indexes[].

    // Get the total length of the data.
    int32_t size;
    if(indexesLength>IX_TOTAL_SIZE) {
        size=indexes[IX_TOTAL_SIZE];
    } else if(indexesLength>IX_REORDER_CODES_OFFSET) {
        size=indexes[indexesLength-1];
    } else {
        size=indexesLength*4;
    }
    if(length<0) { return size; }

    if(length<size) {
        udata_printError(ds, "ucol_swap(formatVersion=4): too few bytes "
                         "(%d after header) for collation data\n",
                         length);
        errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    // Copy the data for inaccessible bytes and arrays of bytes.
    if(inBytes!=outBytes) {
        uprv_memcpy(outBytes, inBytes, size);
    }

    // Swap the int32_t indexes[].
    ds->swapArray32(ds, inBytes, indexesLength * 4, outBytes, &errorCode);

    // The following is a modified version of CollationDataReader::read().
    // Here we use indexes[] not inIndexes[] because
    // the inIndexes[] may not be in this machine's endianness.
    int32_t index;  // one of the indexes[] slots
    int32_t offset;  // byte offset for the index part
    // int32_t length;  // number of bytes in the index part

    index = IX_REORDER_CODES_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        ds->swapArray32(ds, inBytes + offset, length, outBytes + offset, &errorCode);
    }

    // Skip the IX_REORDER_TABLE_OFFSET byte array.

    index = IX_TRIE_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        utrie2_swap(ds, inBytes + offset, length, outBytes + offset, &errorCode);
    }

    index = IX_RESERVED8_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        udata_printError(ds, "ucol_swap(formatVersion=4): unknown data at IX_RESERVED8_OFFSET\n", length);
        errorCode = U_UNSUPPORTED_ERROR;
        return 0;
    }

    index = IX_CES_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        ds->swapArray64(ds, inBytes + offset, length, outBytes + offset, &errorCode);
    }

    index = IX_RESERVED10_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        udata_printError(ds, "ucol_swap(formatVersion=4): unknown data at IX_RESERVED10_OFFSET\n", length);
        errorCode = U_UNSUPPORTED_ERROR;
        return 0;
    }

    index = IX_CE32S_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        ds->swapArray32(ds, inBytes + offset, length, outBytes + offset, &errorCode);
    }

    index = IX_ROOT_ELEMENTS_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        ds->swapArray32(ds, inBytes + offset, length, outBytes + offset, &errorCode);
    }

    index = IX_CONTEXTS_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        ds->swapArray16(ds, inBytes + offset, length, outBytes + offset, &errorCode);
    }

    index = IX_UNSAFE_BWD_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        ds->swapArray16(ds, inBytes + offset, length, outBytes + offset, &errorCode);
    }

    index = IX_FAST_LATIN_TABLE_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        ds->swapArray16(ds, inBytes + offset, length, outBytes + offset, &errorCode);
    }

    index = IX_SCRIPTS_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        ds->swapArray16(ds, inBytes + offset, length, outBytes + offset, &errorCode);
    }

    // Skip the  IX_COMPRESSIBLE_BYTES_OFFSET byte array.

    index = IX_RESERVED18_OFFSET;
    offset = indexes[index];
    length = indexes[index + 1] - offset;
    if(length > 0) {
        udata_printError(ds, "ucol_swap(formatVersion=4): unknown data at IX_RESERVED18_OFFSET\n", length);
        errorCode = U_UNSUPPORTED_ERROR;
        return 0;
    }

    return size;
}

}  // namespace

/* swap ICU collation data like ucadata.icu */
U_CAPI int32_t U_EXPORT2
ucol_swap(const UDataSwapper *ds,
          const void *inData, int32_t length, void *outData,
          UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) { return 0; }

    /* udata_swapDataHeader checks the arguments */
    int32_t headerSize=udata_swapDataHeader(ds, inData, length, outData, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        // Try to swap the old format version which did not have a standard data header.
        *pErrorCode=U_ZERO_ERROR;
        return swapFormatVersion3(ds, inData, length, outData, pErrorCode);
    }

    /* check data format and format version */
    const UDataInfo &info=*(const UDataInfo *)((const char *)inData+4);
    if(!(
        info.dataFormat[0]==0x55 &&   // dataFormat="UCol"
        info.dataFormat[1]==0x43 &&
        info.dataFormat[2]==0x6f &&
        info.dataFormat[3]==0x6c &&
        (3<=info.formatVersion[0] && info.formatVersion[0]<=5)
    )) {
        udata_printError(ds, "ucol_swap(): data format %02x.%02x.%02x.%02x "
                         "(format version %02x.%02x) is not recognized as collation data\n",
                         info.dataFormat[0], info.dataFormat[1],
                         info.dataFormat[2], info.dataFormat[3],
                         info.formatVersion[0], info.formatVersion[1]);
        *pErrorCode=U_UNSUPPORTED_ERROR;
        return 0;
    }

    inData=(const char *)inData+headerSize;
    if(length>=0) { length-=headerSize; }
    outData=(char *)outData+headerSize;
    int32_t collationSize;
    if(info.formatVersion[0]>=4) {
        collationSize=swapFormatVersion4(ds, inData, length, outData, *pErrorCode);
    } else {
        collationSize=swapFormatVersion3(ds, inData, length, outData, pErrorCode);
    }
    if(U_SUCCESS(*pErrorCode)) {
        return headerSize+collationSize;
    } else {
        return 0;
    }
}

/* swap inverse UCA collation data (invuca.icu) */
U_CAPI int32_t U_EXPORT2
ucol_swapInverseUCA(const UDataSwapper *ds,
                    const void *inData, int32_t length, void *outData,
                    UErrorCode *pErrorCode) {
    const UDataInfo *pInfo;
    int32_t headerSize;

    const uint8_t *inBytes;
    uint8_t *outBytes;

    const InverseUCATableHeader *inHeader;
    InverseUCATableHeader *outHeader;
    InverseUCATableHeader header={ 0,0,0,0,0,{0,0,0,0},{0,0,0,0,0,0,0,0} };

    /* udata_swapDataHeader checks the arguments */
    headerSize=udata_swapDataHeader(ds, inData, length, outData, pErrorCode);
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* check data format and format version */
    pInfo=(const UDataInfo *)((const char *)inData+4);
    if(!(
        pInfo->dataFormat[0]==0x49 &&   /* dataFormat="InvC" */
        pInfo->dataFormat[1]==0x6e &&
        pInfo->dataFormat[2]==0x76 &&
        pInfo->dataFormat[3]==0x43 &&
        pInfo->formatVersion[0]==2 &&
        pInfo->formatVersion[1]>=1
    )) {
        udata_printError(ds, "ucol_swapInverseUCA(): data format %02x.%02x.%02x.%02x (format version %02x.%02x) is not an inverse UCA collation file\n",
                         pInfo->dataFormat[0], pInfo->dataFormat[1],
                         pInfo->dataFormat[2], pInfo->dataFormat[3],
                         pInfo->formatVersion[0], pInfo->formatVersion[1]);
        *pErrorCode=U_UNSUPPORTED_ERROR;
        return 0;
    }

    inBytes=(const uint8_t *)inData+headerSize;
    outBytes=(uint8_t *)outData+headerSize;

    inHeader=(const InverseUCATableHeader *)inBytes;
    outHeader=(InverseUCATableHeader *)outBytes;

    /*
     * The inverse UCA collation binary must contain at least the InverseUCATableHeader,
     * starting with its size field.
     * sizeof(UCATableHeader)==8*4 in ICU 2.8
     * check the length against the header size before reading the size field
     */
    if(length<0) {
        header.byteSize=udata_readInt32(ds, inHeader->byteSize);
    } else if(
        ((length-headerSize)<(8*4) ||
         (uint32_t)(length-headerSize)<(header.byteSize=udata_readInt32(ds, inHeader->byteSize)))
    ) {
        udata_printError(ds, "ucol_swapInverseUCA(): too few bytes (%d after header) for inverse UCA collation data\n",
                         length);
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    if(length>=0) {
        /* copy everything, takes care of data that needs no swapping */
        if(inBytes!=outBytes) {
            uprv_memcpy(outBytes, inBytes, header.byteSize);
        }

        /* swap the necessary pieces in the order of their occurrence in the data */

        /* read more of the InverseUCATableHeader (the byteSize field was read above) */
        header.tableSize=   ds->readUInt32(inHeader->tableSize);
        header.contsSize=   ds->readUInt32(inHeader->contsSize);
        header.table=       ds->readUInt32(inHeader->table);
        header.conts=       ds->readUInt32(inHeader->conts);

        /* swap the 32-bit integers in the header */
        ds->swapArray32(ds, inHeader, 5*4, outHeader, pErrorCode);

        /* swap the inverse table; tableSize counts uint32_t[3] rows */
        ds->swapArray32(ds, inBytes+header.table, header.tableSize*3*4,
                           outBytes+header.table, pErrorCode);

        /* swap the continuation table; contsSize counts UChars */
        ds->swapArray16(ds, inBytes+header.conts, header.contsSize*U_SIZEOF_UCHAR,
                           outBytes+header.conts, pErrorCode);
    }

    return headerSize+header.byteSize;
}

#endif /* #if !UCONFIG_NO_COLLATION */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// ucptrie.cpp (modified from utrie2.cpp)
// created: 2017dec29 Markus W. Scherer

// #define UCPTRIE_DEBUG
#ifdef UCPTRIE_DEBUG
#   include <stdio.h>
#endif










U_CAPI UCPTrie * U_EXPORT2
ucptrie_openFromBinary(UCPTrieType type, UCPTrieValueWidth valueWidth,
                       const void *data, int32_t length, int32_t *pActualLength,
                       UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) {
        return nullptr;
    }

    if (length <= 0 || (U_POINTER_MASK_LSB(data, 3) != 0) ||
            type < UCPTRIE_TYPE_ANY || UCPTRIE_TYPE_SMALL < type ||
            valueWidth < UCPTRIE_VALUE_BITS_ANY || UCPTRIE_VALUE_BITS_8 < valueWidth) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }

    // Enough data for a trie header?
    if (length < (int32_t)sizeof(UCPTrieHeader)) {
        *pErrorCode = U_INVALID_FORMAT_ERROR;
        return nullptr;
    }

    // Check the signature.
    const UCPTrieHeader *header = (const UCPTrieHeader *)data;
    if (header->signature != UCPTRIE_SIG) {
        *pErrorCode = U_INVALID_FORMAT_ERROR;
        return nullptr;
    }

    int32_t options = header->options;
    int32_t typeInt = (options >> 6) & 3;
    int32_t valueWidthInt = options & UCPTRIE_OPTIONS_VALUE_BITS_MASK;
    if (typeInt > UCPTRIE_TYPE_SMALL || valueWidthInt > UCPTRIE_VALUE_BITS_8 ||
            (options & UCPTRIE_OPTIONS_RESERVED_MASK) != 0) {
        *pErrorCode = U_INVALID_FORMAT_ERROR;
        return nullptr;
    }
    UCPTrieType actualType = (UCPTrieType)typeInt;
    UCPTrieValueWidth actualValueWidth = (UCPTrieValueWidth)valueWidthInt;
    if (type < 0) {
        type = actualType;
    }
    if (valueWidth < 0) {
        valueWidth = actualValueWidth;
    }
    if (type != actualType || valueWidth != actualValueWidth) {
        *pErrorCode = U_INVALID_FORMAT_ERROR;
        return nullptr;
    }

    // Get the length values and offsets.
    UCPTrie tempTrie;
    uprv_memset(&tempTrie, 0, sizeof(tempTrie));
    tempTrie.indexLength = header->indexLength;
    tempTrie.dataLength =
        ((options & UCPTRIE_OPTIONS_DATA_LENGTH_MASK) << 4) | header->dataLength;
    tempTrie.index3NullOffset = header->index3NullOffset;
    tempTrie.dataNullOffset =
        ((options & UCPTRIE_OPTIONS_DATA_NULL_OFFSET_MASK) << 8) | header->dataNullOffset;

    tempTrie.highStart = header->shiftedHighStart << UCPTRIE_SHIFT_2;
    tempTrie.shifted12HighStart = (tempTrie.highStart + 0xfff) >> 12;
    tempTrie.type = type;
    tempTrie.valueWidth = valueWidth;

    // Calculate the actual length.
    int32_t actualLength = (int32_t)sizeof(UCPTrieHeader) + tempTrie.indexLength * 2;
    if (valueWidth == UCPTRIE_VALUE_BITS_16) {
        actualLength += tempTrie.dataLength * 2;
    } else if (valueWidth == UCPTRIE_VALUE_BITS_32) {
        actualLength += tempTrie.dataLength * 4;
    } else {
        actualLength += tempTrie.dataLength;
    }
    if (length < actualLength) {
        *pErrorCode = U_INVALID_FORMAT_ERROR;  // Not enough bytes.
        return nullptr;
    }

    // Allocate the trie.
    UCPTrie *trie = (UCPTrie *)uprv_malloc(sizeof(UCPTrie));
    if (trie == nullptr) {
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    uprv_memcpy(trie, &tempTrie, sizeof(tempTrie));
#ifdef UCPTRIE_DEBUG
    trie->name = "fromSerialized";
#endif

    // Set the pointers to its index and data arrays.
    const uint16_t *p16 = (const uint16_t *)(header + 1);
    trie->index = p16;
    p16 += trie->indexLength;

    // Get the data.
    int32_t nullValueOffset = trie->dataNullOffset;
    if (nullValueOffset >= trie->dataLength) {
        nullValueOffset = trie->dataLength - UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET;
    }
    switch (valueWidth) {
    case UCPTRIE_VALUE_BITS_16:
        trie->data.ptr16 = p16;
        trie->nullValue = trie->data.ptr16[nullValueOffset];
        break;
    case UCPTRIE_VALUE_BITS_32:
        trie->data.ptr32 = (const uint32_t *)p16;
        trie->nullValue = trie->data.ptr32[nullValueOffset];
        break;
    case UCPTRIE_VALUE_BITS_8:
        trie->data.ptr8 = (const uint8_t *)p16;
        trie->nullValue = trie->data.ptr8[nullValueOffset];
        break;
    default:
        // Unreachable because valueWidth was checked above.
        *pErrorCode = U_INVALID_FORMAT_ERROR;
        return nullptr;
    }

    if (pActualLength != nullptr) {
        *pActualLength = actualLength;
    }
    return trie;
}

U_CAPI void U_EXPORT2
ucptrie_close(UCPTrie *trie) {
    uprv_free(trie);
}

U_CAPI UCPTrieType U_EXPORT2
ucptrie_getType(const UCPTrie *trie) {
    return (UCPTrieType)trie->type;
}

U_CAPI UCPTrieValueWidth U_EXPORT2
ucptrie_getValueWidth(const UCPTrie *trie) {
    return (UCPTrieValueWidth)trie->valueWidth;
}

U_CAPI int32_t U_EXPORT2
ucptrie_internalSmallIndex(const UCPTrie *trie, UChar32 c) {
    int32_t i1 = c >> UCPTRIE_SHIFT_1;
    if (trie->type == UCPTRIE_TYPE_FAST) {
        U_ASSERT(0xffff < c && c < trie->highStart);
        i1 += UCPTRIE_BMP_INDEX_LENGTH - UCPTRIE_OMITTED_BMP_INDEX_1_LENGTH;
    } else {
        U_ASSERT((uint32_t)c < (uint32_t)trie->highStart && trie->highStart > UCPTRIE_SMALL_LIMIT);
        i1 += UCPTRIE_SMALL_INDEX_LENGTH;
    }
    int32_t i3Block = trie->index[
        (int32_t)trie->index[i1] + ((c >> UCPTRIE_SHIFT_2) & UCPTRIE_INDEX_2_MASK)];
    int32_t i3 = (c >> UCPTRIE_SHIFT_3) & UCPTRIE_INDEX_3_MASK;
    int32_t dataBlock;
    if ((i3Block & 0x8000) == 0) {
        // 16-bit indexes
        dataBlock = trie->index[i3Block + i3];
    } else {
        // 18-bit indexes stored in groups of 9 entries per 8 indexes.
        i3Block = (i3Block & 0x7fff) + (i3 & ~7) + (i3 >> 3);
        i3 &= 7;
        dataBlock = ((int32_t)trie->index[i3Block++] << (2 + (2 * i3))) & 0x30000;
        dataBlock |= trie->index[i3Block + i3];
    }
    return dataBlock + (c & UCPTRIE_SMALL_DATA_MASK);
}

U_CAPI int32_t U_EXPORT2
ucptrie_internalSmallU8Index(const UCPTrie *trie, int32_t lt1, uint8_t t2, uint8_t t3) {
    UChar32 c = (lt1 << 12) | (t2 << 6) | t3;
    if (c >= trie->highStart) {
        // Possible because the UTF-8 macro compares with shifted12HighStart which may be higher.
        return trie->dataLength - UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET;
    }
    return ucptrie_internalSmallIndex(trie, c);
}

U_CAPI int32_t U_EXPORT2
ucptrie_internalU8PrevIndex(const UCPTrie *trie, UChar32 c,
                            const uint8_t *start, const uint8_t *src) {
    int32_t i, length;
    // Support 64-bit pointers by avoiding cast of arbitrary difference.
    if ((src - start) <= 7) {
        i = length = (int32_t)(src - start);
    } else {
        i = length = 7;
        start = src - 7;
    }
    c = utf8_prevCharSafeBody(start, 0, &i, c, -1);
    i = length - i;  // Number of bytes read backward from src.
    int32_t idx = _UCPTRIE_CP_INDEX(trie, 0xffff, c);
    return (idx << 3) | i;
}

namespace {

inline uint32_t getValue(UCPTrieData data, UCPTrieValueWidth valueWidth, int32_t dataIndex) {
    switch (valueWidth) {
    case UCPTRIE_VALUE_BITS_16:
        return data.ptr16[dataIndex];
    case UCPTRIE_VALUE_BITS_32:
        return data.ptr32[dataIndex];
    case UCPTRIE_VALUE_BITS_8:
        return data.ptr8[dataIndex];
    default:
        // Unreachable if the trie is properly initialized.
        return 0xffffffff;
    }
}

}  // namespace

U_CAPI uint32_t U_EXPORT2
ucptrie_get(const UCPTrie *trie, UChar32 c) {
    int32_t dataIndex;
    if ((uint32_t)c <= 0x7f) {
        // linear ASCII
        dataIndex = c;
    } else {
        UChar32 fastMax = trie->type == UCPTRIE_TYPE_FAST ? 0xffff : UCPTRIE_SMALL_MAX;
        dataIndex = _UCPTRIE_CP_INDEX(trie, fastMax, c);
    }
    return getValue(trie->data, (UCPTrieValueWidth)trie->valueWidth, dataIndex);
}

namespace {

constexpr int32_t MAX_UNICODE = 0x10ffff;

inline uint32_t maybeFilterValue(uint32_t value, uint32_t trieNullValue, uint32_t nullValue,
                                 UCPMapValueFilter *filter, const void *context) {
    if (value == trieNullValue) {
        value = nullValue;
    } else if (filter != nullptr) {
        value = filter(context, value);
    }
    return value;
}

UChar32 ucptrie_internal_getRange(const void *t, UChar32 start,
                 UCPMapValueFilter *filter, const void *context, uint32_t *pValue) {
    if ((uint32_t)start > MAX_UNICODE) {
        return U_SENTINEL;
    }
    const UCPTrie *trie = reinterpret_cast<const UCPTrie *>(t);
    UCPTrieValueWidth valueWidth = (UCPTrieValueWidth)trie->valueWidth;
    if (start >= trie->highStart) {
        if (pValue != nullptr) {
            int32_t di = trie->dataLength - UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET;
            uint32_t value = getValue(trie->data, valueWidth, di);
            if (filter != nullptr) { value = filter(context, value); }
            *pValue = value;
        }
        return MAX_UNICODE;
    }

    uint32_t nullValue = trie->nullValue;
    if (filter != nullptr) { nullValue = filter(context, nullValue); }
    const uint16_t *index = trie->index;

    int32_t prevI3Block = -1;
    int32_t prevBlock = -1;
    UChar32 c = start;
    uint32_t trieValue, value = nullValue;
    bool haveValue = false;
    do {
        int32_t i3Block;
        int32_t i3;
        int32_t i3BlockLength;
        int32_t dataBlockLength;
        if (c <= 0xffff && (trie->type == UCPTRIE_TYPE_FAST || c <= UCPTRIE_SMALL_MAX)) {
            i3Block = 0;
            i3 = c >> UCPTRIE_FAST_SHIFT;
            i3BlockLength = trie->type == UCPTRIE_TYPE_FAST ?
                UCPTRIE_BMP_INDEX_LENGTH : UCPTRIE_SMALL_INDEX_LENGTH;
            dataBlockLength = UCPTRIE_FAST_DATA_BLOCK_LENGTH;
        } else {
            // Use the multi-stage index.
            int32_t i1 = c >> UCPTRIE_SHIFT_1;
            if (trie->type == UCPTRIE_TYPE_FAST) {
                U_ASSERT(0xffff < c && c < trie->highStart);
                i1 += UCPTRIE_BMP_INDEX_LENGTH - UCPTRIE_OMITTED_BMP_INDEX_1_LENGTH;
            } else {
                U_ASSERT(c < trie->highStart && trie->highStart > UCPTRIE_SMALL_LIMIT);
                i1 += UCPTRIE_SMALL_INDEX_LENGTH;
            }
            i3Block = trie->index[
                (int32_t)trie->index[i1] + ((c >> UCPTRIE_SHIFT_2) & UCPTRIE_INDEX_2_MASK)];
            if (i3Block == prevI3Block && (c - start) >= UCPTRIE_CP_PER_INDEX_2_ENTRY) {
                // The index-3 block is the same as the previous one, and filled with value.
                U_ASSERT((c & (UCPTRIE_CP_PER_INDEX_2_ENTRY - 1)) == 0);
                c += UCPTRIE_CP_PER_INDEX_2_ENTRY;
                continue;
            }
            prevI3Block = i3Block;
            if (i3Block == trie->index3NullOffset) {
                // This is the index-3 null block.
                if (haveValue) {
                    if (nullValue != value) {
                        return c - 1;
                    }
                } else {
                    trieValue = trie->nullValue;
                    value = nullValue;
                    if (pValue != nullptr) { *pValue = nullValue; }
                    haveValue = true;
                }
                prevBlock = trie->dataNullOffset;
                c = (c + UCPTRIE_CP_PER_INDEX_2_ENTRY) & ~(UCPTRIE_CP_PER_INDEX_2_ENTRY - 1);
                continue;
            }
            i3 = (c >> UCPTRIE_SHIFT_3) & UCPTRIE_INDEX_3_MASK;
            i3BlockLength = UCPTRIE_INDEX_3_BLOCK_LENGTH;
            dataBlockLength = UCPTRIE_SMALL_DATA_BLOCK_LENGTH;
        }
        // Enumerate data blocks for one index-3 block.
        do {
            int32_t block;
            if ((i3Block & 0x8000) == 0) {
                block = index[i3Block + i3];
            } else {
                // 18-bit indexes stored in groups of 9 entries per 8 indexes.
                int32_t group = (i3Block & 0x7fff) + (i3 & ~7) + (i3 >> 3);
                int32_t gi = i3 & 7;
                block = ((int32_t)index[group++] << (2 + (2 * gi))) & 0x30000;
                block |= index[group + gi];
            }
            if (block == prevBlock && (c - start) >= dataBlockLength) {
                // The block is the same as the previous one, and filled with value.
                U_ASSERT((c & (dataBlockLength - 1)) == 0);
                c += dataBlockLength;
            } else {
                int32_t dataMask = dataBlockLength - 1;
                prevBlock = block;
                if (block == trie->dataNullOffset) {
                    // This is the data null block.
                    if (haveValue) {
                        if (nullValue != value) {
                            return c - 1;
                        }
                    } else {
                        trieValue = trie->nullValue;
                        value = nullValue;
                        if (pValue != nullptr) { *pValue = nullValue; }
                        haveValue = true;
                    }
                    c = (c + dataBlockLength) & ~dataMask;
                } else {
                    int32_t di = block + (c & dataMask);
                    uint32_t trieValue2 = getValue(trie->data, valueWidth, di);
                    if (haveValue) {
                        if (trieValue2 != trieValue) {
                            if (filter == nullptr ||
                                    maybeFilterValue(trieValue2, trie->nullValue, nullValue,
                                                     filter, context) != value) {
                                return c - 1;
                            }
                            trieValue = trieValue2;  // may or may not help
                        }
                    } else {
                        trieValue = trieValue2;
                        value = maybeFilterValue(trieValue2, trie->nullValue, nullValue,
                                                 filter, context);
                        if (pValue != nullptr) { *pValue = value; }
                        haveValue = true;
                    }
                    while ((++c & dataMask) != 0) {
                        trieValue2 = getValue(trie->data, valueWidth, ++di);
                        if (trieValue2 != trieValue) {
                            if (filter == nullptr ||
                                    maybeFilterValue(trieValue2, trie->nullValue, nullValue,
                                                     filter, context) != value) {
                                return c - 1;
                            }
                            trieValue = trieValue2;  // may or may not help
                        }
                    }
                }
            }
        } while (++i3 < i3BlockLength);
    } while (c < trie->highStart);
    U_ASSERT(haveValue);
    int32_t di = trie->dataLength - UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET;
    uint32_t highValue = getValue(trie->data, valueWidth, di);
    if (maybeFilterValue(highValue, trie->nullValue, nullValue,
                         filter, context) != value) {
        return c - 1;
    } else {
        return MAX_UNICODE;
    }
}

}  // namespace

U_CFUNC UChar32
ucptrie_internalGetRange(UCPTrieGetRange *getRange,
                         const void *trie, UChar32 start,
                         UCPMapRangeOption option, uint32_t surrogateValue,
                         UCPMapValueFilter *filter, const void *context, uint32_t *pValue) {
    if (option == UCPMAP_RANGE_NORMAL) {
        return getRange(trie, start, filter, context, pValue);
    }
    uint32_t value;
    if (pValue == nullptr) {
        // We need to examine the range value even if the caller does not want it.
        pValue = &value;
    }
    UChar32 surrEnd = option == UCPMAP_RANGE_FIXED_ALL_SURROGATES ? 0xdfff : 0xdbff;
    UChar32 end = getRange(trie, start, filter, context, pValue);
    if (end < 0xd7ff || start > surrEnd) {
        return end;
    }
    // The range overlaps with surrogates, or ends just before the first one.
    if (*pValue == surrogateValue) {
        if (end >= surrEnd) {
            // Surrogates followed by a non-surrogateValue range,
            // or surrogates are part of a larger surrogateValue range.
            return end;
        }
    } else {
        if (start <= 0xd7ff) {
            return 0xd7ff;  // Non-surrogateValue range ends before surrogateValue surrogates.
        }
        // Start is a surrogate with a non-surrogateValue code *unit* value.
        // Return a surrogateValue code *point* range.
        *pValue = surrogateValue;
        if (end > surrEnd) {
            return surrEnd;  // Surrogate range ends before non-surrogateValue rest of range.
        }
    }
    // See if the surrogateValue surrogate range can be merged with
    // an immediately following range.
    uint32_t value2;
    UChar32 end2 = getRange(trie, surrEnd + 1, filter, context, &value2);
    if (value2 == surrogateValue) {
        return end2;
    }
    return surrEnd;
}

U_CAPI UChar32 U_EXPORT2
ucptrie_getRange(const UCPTrie *trie, UChar32 start,
                 UCPMapRangeOption option, uint32_t surrogateValue,
                 UCPMapValueFilter *filter, const void *context, uint32_t *pValue) {
    return ucptrie_internalGetRange(ucptrie_internal_getRange, trie, start,
                                    option, surrogateValue,
                                    filter, context, pValue);
}

U_CAPI int32_t U_EXPORT2
ucptrie_toBinary(const UCPTrie *trie,
                 void *data, int32_t capacity,
                 UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) {
        return 0;
    }

    UCPTrieType type = (UCPTrieType)trie->type;
    UCPTrieValueWidth valueWidth = (UCPTrieValueWidth)trie->valueWidth;
    if (type < UCPTRIE_TYPE_FAST || UCPTRIE_TYPE_SMALL < type ||
            valueWidth < UCPTRIE_VALUE_BITS_16 || UCPTRIE_VALUE_BITS_8 < valueWidth ||
            capacity < 0 ||
            (capacity > 0 && (data == nullptr || (U_POINTER_MASK_LSB(data, 3) != 0)))) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    int32_t length = (int32_t)sizeof(UCPTrieHeader) + trie->indexLength * 2;
    switch (valueWidth) {
    case UCPTRIE_VALUE_BITS_16:
        length += trie->dataLength * 2;
        break;
    case UCPTRIE_VALUE_BITS_32:
        length += trie->dataLength * 4;
        break;
    case UCPTRIE_VALUE_BITS_8:
        length += trie->dataLength;
        break;
    default:
        // unreachable
        break;
    }
    if (capacity < length) {
        *pErrorCode = U_BUFFER_OVERFLOW_ERROR;
        return length;
    }

    char *bytes = (char *)data;
    UCPTrieHeader *header = (UCPTrieHeader *)bytes;
    header->signature = UCPTRIE_SIG;  // "Tri3"
    header->options = (uint16_t)(
        ((trie->dataLength & 0xf0000) >> 4) |
        ((trie->dataNullOffset & 0xf0000) >> 8) |
        (trie->type << 6) |
        valueWidth);
    header->indexLength = (uint16_t)trie->indexLength;
    header->dataLength = (uint16_t)trie->dataLength;
    header->index3NullOffset = trie->index3NullOffset;
    header->dataNullOffset = (uint16_t)trie->dataNullOffset;
    header->shiftedHighStart = trie->highStart >> UCPTRIE_SHIFT_2;
    bytes += sizeof(UCPTrieHeader);

    uprv_memcpy(bytes, trie->index, trie->indexLength * 2);
    bytes += trie->indexLength * 2;

    switch (valueWidth) {
    case UCPTRIE_VALUE_BITS_16:
        uprv_memcpy(bytes, trie->data.ptr16, trie->dataLength * 2);
        break;
    case UCPTRIE_VALUE_BITS_32:
        uprv_memcpy(bytes, trie->data.ptr32, trie->dataLength * 4);
        break;
    case UCPTRIE_VALUE_BITS_8:
        uprv_memcpy(bytes, trie->data.ptr8, trie->dataLength);
        break;
    default:
        // unreachable
        break;
    }
    return length;
}

namespace {

#ifdef UCPTRIE_DEBUG
long countNull(const UCPTrie *trie) {
    uint32_t nullValue=trie->nullValue;
    int32_t length=trie->dataLength;
    long count=0;
    switch (trie->valueWidth) {
    case UCPTRIE_VALUE_BITS_16:
        for(int32_t i=0; i<length; ++i) {
            if(trie->data.ptr16[i]==nullValue) { ++count; }
        }
        break;
    case UCPTRIE_VALUE_BITS_32:
        for(int32_t i=0; i<length; ++i) {
            if(trie->data.ptr32[i]==nullValue) { ++count; }
        }
        break;
    case UCPTRIE_VALUE_BITS_8:
        for(int32_t i=0; i<length; ++i) {
            if(trie->data.ptr8[i]==nullValue) { ++count; }
        }
        break;
    default:
        // unreachable
        break;
    }
    return count;
}

U_CFUNC void
ucptrie_printLengths(const UCPTrie *trie, const char *which) {
    long indexLength=trie->indexLength;
    long dataLength=(long)trie->dataLength;
    long totalLength=(long)sizeof(UCPTrieHeader)+indexLength*2+
            dataLength*(trie->valueWidth==UCPTRIE_VALUE_BITS_16 ? 2 :
                        trie->valueWidth==UCPTRIE_VALUE_BITS_32 ? 4 : 1);
    printf("**UCPTrieLengths(%s %s)** index:%6ld  data:%6ld  countNull:%6ld  serialized:%6ld\n",
           which, trie->name, indexLength, dataLength, countNull(trie), totalLength);
}
#endif

}  // namespace

// UCPMap ----
// Initially, this is the same as UCPTrie. This may well change.

U_CAPI uint32_t U_EXPORT2
ucpmap_get(const UCPMap *map, UChar32 c) {
    return ucptrie_get(reinterpret_cast<const UCPTrie *>(map), c);
}

U_CAPI UChar32 U_EXPORT2
ucpmap_getRange(const UCPMap *map, UChar32 start,
                UCPMapRangeOption option, uint32_t surrogateValue,
                UCPMapValueFilter *filter, const void *context, uint32_t *pValue) {
    return ucptrie_getRange(reinterpret_cast<const UCPTrie *>(map), start,
                            option, surrogateValue,
                            filter, context, pValue);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2002-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
*/



#if !UCONFIG_NO_FORMATTING

















// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2002-2006, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uenumimp.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:2
*
*   created on: 2002jul08
*   created by: Vladimir Weinstein
*/




U_CDECL_BEGIN

/** 
 * following are the type declarations for 
 * implementations of APIs. If any of these
 * functions are NULL, U_UNSUPPORTED_ERROR
 * is returned. If close is NULL, the enumeration
 * object is going to be released.
 * Initial error checking is done in the body
 * of API function, so the implementations 
 * need not to check the initial error condition.
 */

/**
 * Function type declaration for uenum_close().
 *
 * This function should cleanup the enumerator object
 *
 * @param en enumeration to be closed
 */
typedef void U_CALLCONV
UEnumClose(UEnumeration *en);

/**
 * Function type declaration for uenum_count().
 *
 * This function should count the number of elements
 * in this enumeration
 *
 * @param en enumeration to be counted
 * @param status pointer to UErrorCode variable
 * @return number of elements in enumeration
 */
typedef int32_t U_CALLCONV
UEnumCount(UEnumeration *en, UErrorCode *status);

/**
 * Function type declaration for uenum_unext().
 *
 * This function returns the next element as a UChar *,
 * or NULL after all elements haven been enumerated.
 *
 * @param en enumeration 
 * @param resultLength pointer to result length
 * @param status pointer to UErrorCode variable
 * @return next element as UChar *,
 *         or NULL after all elements haven been enumerated
 */
typedef const UChar* U_CALLCONV 
UEnumUNext(UEnumeration* en,
            int32_t* resultLength,
            UErrorCode* status);

/**
 * Function type declaration for uenum_next().
 *
 * This function returns the next element as a char *,
 * or NULL after all elements haven been enumerated.
 *
 * @param en enumeration 
 * @param resultLength pointer to result length
 * @param status pointer to UErrorCode variable
 * @return next element as char *,
 *         or NULL after all elements haven been enumerated
 */
typedef const char* U_CALLCONV 
UEnumNext(UEnumeration* en,
           int32_t* resultLength,
           UErrorCode* status);

/**
 * Function type declaration for uenum_reset().
 *
 * This function should reset the enumeration 
 * object
 *
 * @param en enumeration 
 * @param status pointer to UErrorCode variable
 */
typedef void U_CALLCONV 
UEnumReset(UEnumeration* en, 
            UErrorCode* status);


struct UEnumeration {
    /* baseContext. For the base class only. Don't touch! */
    void *baseContext;

    /* context. Use it for what you need */
    void *context;

    /** 
     * these are functions that will 
     * be used for APIs
     */
    /* called from uenum_close */
    UEnumClose *close;
    /* called from uenum_count */
    UEnumCount *count;
    /* called from uenum_unext */
    UEnumUNext *uNext;
    /* called from uenum_next */
    UEnumNext  *next;
    /* called from uenum_reset */
    UEnumReset *reset;
};

U_CDECL_END

/* This is the default implementation for uenum_unext().
 * It automatically converts the char * string to UChar *.
 * Don't call this directly.  This is called internally by uenum_unext
 * when a UEnumeration is defined with 'uNext' pointing to this
 * function.
 */
U_CAPI const UChar* U_EXPORT2
uenum_unextDefault(UEnumeration* en,
            int32_t* resultLength,
            UErrorCode* status);

/* This is the default implementation for uenum_next().
 * It automatically converts the UChar * string to char *.
 * Don't call this directly.  This is called internally by uenum_next
 * when a UEnumeration is defined with 'next' pointing to this
 * function.
 */
U_CAPI const char* U_EXPORT2
uenum_nextDefault(UEnumeration* en,
            int32_t* resultLength,
            UErrorCode* status);






// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 2009-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*/





struct UList;
typedef struct UList UList;

U_CAPI UList * U_EXPORT2 ulist_createEmptyList(UErrorCode *status);

U_CAPI void U_EXPORT2 ulist_addItemEndList(UList *list, const void *data, UBool forceDelete, UErrorCode *status);

U_CAPI void U_EXPORT2 ulist_addItemBeginList(UList *list, const void *data, UBool forceDelete, UErrorCode *status);

U_CAPI UBool U_EXPORT2 ulist_containsString(const UList *list, const char *data, int32_t length);

U_CAPI UBool U_EXPORT2 ulist_removeString(UList *list, const char *data);

U_CAPI void *U_EXPORT2 ulist_getNext(UList *list);

U_CAPI int32_t U_EXPORT2 ulist_getListSize(const UList *list);

U_CAPI void U_EXPORT2 ulist_resetList(UList *list);

U_CAPI void U_EXPORT2 ulist_deleteList(UList *list);

/*
 * The following are for use when creating UEnumeration object backed by UList.
 */
U_CAPI void U_EXPORT2 ulist_close_keyword_values_iterator(UEnumeration *en);

U_CAPI int32_t U_EXPORT2 ulist_count_keyword_values(UEnumeration *en, UErrorCode *status);

U_CAPI const char * U_EXPORT2 ulist_next_keyword_value(UEnumeration* en, int32_t *resultLength, UErrorCode* status);

U_CAPI void U_EXPORT2 ulist_reset_keyword_values_iterator(UEnumeration* en, UErrorCode* status);

U_CAPI UList * U_EXPORT2 ulist_getListFromEnum(UEnumeration *en);






using namespace icu;

//#define UCURR_DEBUG_EQUIV 1
#ifdef UCURR_DEBUG_EQUIV
#include "stdio.h"
#endif
//#define UCURR_DEBUG 1
#ifdef UCURR_DEBUG
#include "stdio.h"
#endif

typedef struct IsoCodeEntry {
    const UChar *isoCode; /* const because it's a reference to a resource bundle string. */
    UDate from;
    UDate to;
} IsoCodeEntry;

//------------------------------------------------------------
// Constants

// Default currency meta data of last resort.  We try to use the
// defaults encoded in the meta data resource bundle.  If there is a
// configuration/build error and these are not available, we use these
// hard-coded defaults (which should be identical).
static const int32_t LAST_RESORT_DATA[] = { 2, 0, 2, 0 };

// POW10[i] = 10^i, i=0..MAX_POW10
static const int32_t POW10[] = { 1, 10, 100, 1000, 10000, 100000,
                                 1000000, 10000000, 100000000, 1000000000 };

static const int32_t MAX_POW10 = UPRV_LENGTHOF(POW10) - 1;

#define ISO_CURRENCY_CODE_LENGTH 3

//------------------------------------------------------------
// Resource tags
//

static const char CURRENCY_DATA[] = "supplementalData";
// Tag for meta-data, in root.
static const char CURRENCY_META[] = "CurrencyMeta";

// Tag for map from countries to currencies, in root.
static const char CURRENCY_MAP[] = "CurrencyMap";

// Tag for default meta-data, in CURRENCY_META
static const char DEFAULT_META[] = "DEFAULT";

// Variant delimiter
static const char VAR_DELIM = '_';

// Tag for localized display names (symbols) of currencies
static const char CURRENCIES[] = "Currencies";
static const char CURRENCIES_NARROW[] = "Currencies%narrow";
static const char CURRENCYPLURALS[] = "CurrencyPlurals";

// ISO codes mapping table
static const UHashtable* gIsoCodes = NULL;
static icu::UInitOnce gIsoCodesInitOnce = U_INITONCE_INITIALIZER;

// Currency symbol equivalances
static const icu::Hashtable* gCurrSymbolsEquiv = NULL;
static icu::UInitOnce gCurrSymbolsEquivInitOnce = U_INITONCE_INITIALIZER;

U_NAMESPACE_BEGIN

// EquivIterator iterates over all strings that are equivalent to a given
// string, s. Note that EquivIterator will never yield s itself.
class EquivIterator : public icu::UMemory {
public:
    // Constructor. hash stores the equivalence relationships; s is the string
    // for which we find equivalent strings.
    inline EquivIterator(const icu::Hashtable& hash, const icu::UnicodeString& s)
        : _hash(hash) {
        _start = _current = &s;
    }
    inline ~EquivIterator() { }

    // next returns the next equivalent string or NULL if there are no more.
    // If s has no equivalent strings, next returns NULL on the first call.
    const icu::UnicodeString *next();
private:
    const icu::Hashtable& _hash;
    const icu::UnicodeString* _start;
    const icu::UnicodeString* _current;
};

const icu::UnicodeString *
EquivIterator::next() {
    const icu::UnicodeString* _next = (const icu::UnicodeString*) _hash.get(*_current);
    if (_next == NULL) {
        U_ASSERT(_current == _start);
        return NULL;
    }
    if (*_next == *_start) {
        return NULL;
    }
    _current = _next;
    return _next;
}

U_NAMESPACE_END

// makeEquivalent makes lhs and rhs equivalent by updating the equivalence
// relations in hash accordingly.
static void makeEquivalent(
    const icu::UnicodeString &lhs,
    const icu::UnicodeString &rhs,
    icu::Hashtable* hash, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return;
    }
    if (lhs == rhs) {
        // already equivalent
        return;
    }
    icu::EquivIterator leftIter(*hash, lhs);
    icu::EquivIterator rightIter(*hash, rhs);
    const icu::UnicodeString *firstLeft = leftIter.next();
    const icu::UnicodeString *firstRight = rightIter.next();
    const icu::UnicodeString *nextLeft = firstLeft;
    const icu::UnicodeString *nextRight = firstRight;
    while (nextLeft != NULL && nextRight != NULL) {
        if (*nextLeft == rhs || *nextRight == lhs) {
            // Already equivalent
            return;
        }
        nextLeft = leftIter.next();
        nextRight = rightIter.next();
    }
    // Not equivalent. Must join.
    icu::UnicodeString *newFirstLeft;
    icu::UnicodeString *newFirstRight;
    if (firstRight == NULL && firstLeft == NULL) {
        // Neither lhs or rhs belong to an equivalence circle, so we form
        // a new equivalnce circle of just lhs and rhs.
        newFirstLeft = new icu::UnicodeString(rhs);
        newFirstRight = new icu::UnicodeString(lhs);
    } else if (firstRight == NULL) {
        // lhs belongs to an equivalence circle, but rhs does not, so we link
        // rhs into lhs' circle.
        newFirstLeft = new icu::UnicodeString(rhs);
        newFirstRight = new icu::UnicodeString(*firstLeft);
    } else if (firstLeft == NULL) {
        // rhs belongs to an equivlance circle, but lhs does not, so we link
        // lhs into rhs' circle.
        newFirstLeft = new icu::UnicodeString(*firstRight);
        newFirstRight = new icu::UnicodeString(lhs);
    } else {
        // Both lhs and rhs belong to different equivalnce circles. We link
        // them together to form one single, larger equivalnce circle.
        newFirstLeft = new icu::UnicodeString(*firstRight);
        newFirstRight = new icu::UnicodeString(*firstLeft);
    }
    if (newFirstLeft == NULL || newFirstRight == NULL) {
        delete newFirstLeft;
        delete newFirstRight;
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    hash->put(lhs, (void *) newFirstLeft, status);
    hash->put(rhs, (void *) newFirstRight, status);
}

// countEquivalent counts how many strings are equivalent to s.
// hash stores all the equivalnce relations.
// countEquivalent does not include s itself in the count.
static int32_t countEquivalent(const icu::Hashtable &hash, const icu::UnicodeString &s) {
    int32_t result = 0;
    icu::EquivIterator iter(hash, s);
    while (iter.next() != NULL) {
        ++result;
    }
#ifdef UCURR_DEBUG_EQUIV
 {
   char tmp[200];
   s.extract(0,s.length(),tmp, "UTF-8");
   printf("CountEquivalent('%s') = %d\n", tmp, result);
 }
#endif
    return result;
}

static const icu::Hashtable* getCurrSymbolsEquiv();

//------------------------------------------------------------
// Code

/**
 * Cleanup callback func
 */
static UBool U_CALLCONV
isoCodes_cleanup(void)
{
    if (gIsoCodes != NULL) {
        uhash_close(const_cast<UHashtable *>(gIsoCodes));
        gIsoCodes = NULL;
    }
    gIsoCodesInitOnce.reset();
    return TRUE;
}

/**
 * Cleanup callback func
 */
static UBool U_CALLCONV
currSymbolsEquiv_cleanup(void)
{
    delete const_cast<icu::Hashtable *>(gCurrSymbolsEquiv);
    gCurrSymbolsEquiv = NULL;
    gCurrSymbolsEquivInitOnce.reset();
    return TRUE;
}

/**
 * Deleter for OlsonToMetaMappingEntry
 */
static void U_CALLCONV
deleteIsoCodeEntry(void *obj) {
    IsoCodeEntry *entry = (IsoCodeEntry*)obj;
    uprv_free(entry);
}

/**
 * Deleter for gCurrSymbolsEquiv.
 */
static void U_CALLCONV
deleteUnicode(void *obj) {
    icu::UnicodeString *entry = (icu::UnicodeString*)obj;
    delete entry;
}

/**
 * Unfortunately, we have to convert the UChar* currency code to char*
 * to use it as a resource key.
 */
static inline char*
myUCharsToChars(char* resultOfLen4, const UChar* currency) {
    u_UCharsToChars(currency, resultOfLen4, ISO_CURRENCY_CODE_LENGTH);
    resultOfLen4[ISO_CURRENCY_CODE_LENGTH] = 0;
    return resultOfLen4;
}

/**
 * Internal function to look up currency data.  Result is an array of
 * four integers.  The first is the fraction digits.  The second is the
 * rounding increment, or 0 if none.  The rounding increment is in
 * units of 10^(-fraction_digits).  The third and fourth are the same
 * except that they are those used in cash transations ( cashDigits
 * and cashRounding ).
 */
static const int32_t*
_findMetaData(const UChar* currency, UErrorCode& ec) {

    if (currency == 0 || *currency == 0) {
        if (U_SUCCESS(ec)) {
            ec = U_ILLEGAL_ARGUMENT_ERROR;
        }
        return LAST_RESORT_DATA;
    }

    // Get CurrencyMeta resource out of root locale file.  [This may
    // move out of the root locale file later; if it does, update this
    // code.]
    UResourceBundle* currencyData = ures_openDirect(U_ICUDATA_CURR, CURRENCY_DATA, &ec);
    UResourceBundle* currencyMeta = ures_getByKey(currencyData, CURRENCY_META, currencyData, &ec);

    if (U_FAILURE(ec)) {
        ures_close(currencyMeta);
        // Config/build error; return hard-coded defaults
        return LAST_RESORT_DATA;
    }

    // Look up our currency, or if that's not available, then DEFAULT
    char buf[ISO_CURRENCY_CODE_LENGTH+1];
    UErrorCode ec2 = U_ZERO_ERROR; // local error code: soft failure
    UResourceBundle* rb = ures_getByKey(currencyMeta, myUCharsToChars(buf, currency), NULL, &ec2);
      if (U_FAILURE(ec2)) {
        ures_close(rb);
        rb = ures_getByKey(currencyMeta,DEFAULT_META, NULL, &ec);
        if (U_FAILURE(ec)) {
            ures_close(currencyMeta);
            ures_close(rb);
            // Config/build error; return hard-coded defaults
            return LAST_RESORT_DATA;
        }
    }

    int32_t len;
    const int32_t *data = ures_getIntVector(rb, &len, &ec);
    if (U_FAILURE(ec) || len != 4) {
        // Config/build error; return hard-coded defaults
        if (U_SUCCESS(ec)) {
            ec = U_INVALID_FORMAT_ERROR;
        }
        ures_close(currencyMeta);
        ures_close(rb);
        return LAST_RESORT_DATA;
    }

    ures_close(currencyMeta);
    ures_close(rb);
    return data;
}

// -------------------------------------

static void
idForLocale(const char* locale, char* countryAndVariant, int capacity, UErrorCode* ec)
{
    ulocimp_getRegionForSupplementalData(locale, FALSE, countryAndVariant, capacity, ec);
}

// ------------------------------------------
//
// Registration
//
//-------------------------------------------

// don't use ICUService since we don't need fallback

U_CDECL_BEGIN
static UBool U_CALLCONV currency_cleanup(void);
U_CDECL_END

#if !UCONFIG_NO_SERVICE
struct CReg;

static UMutex gCRegLock;
static CReg* gCRegHead = 0;

struct CReg : public icu::UMemory {
    CReg *next;
    UChar iso[ISO_CURRENCY_CODE_LENGTH+1];
    char  id[ULOC_FULLNAME_CAPACITY];

    CReg(const UChar* _iso, const char* _id)
        : next(0)
    {
        int32_t len = (int32_t)uprv_strlen(_id);
        if (len > (int32_t)(sizeof(id)-1)) {
            len = (sizeof(id)-1);
        }
        uprv_strncpy(id, _id, len);
        id[len] = 0;
        u_memcpy(iso, _iso, ISO_CURRENCY_CODE_LENGTH);
        iso[ISO_CURRENCY_CODE_LENGTH] = 0;
    }

    static UCurrRegistryKey reg(const UChar* _iso, const char* _id, UErrorCode* status)
    {
        if (status && U_SUCCESS(*status) && _iso && _id) {
            CReg* n = new CReg(_iso, _id);
            if (n) {
                umtx_lock(&gCRegLock);
                if (!gCRegHead) {
                    /* register for the first time */
                    ucln_common_registerCleanup(UCLN_COMMON_CURRENCY, currency_cleanup);
                }
                n->next = gCRegHead;
                gCRegHead = n;
                umtx_unlock(&gCRegLock);
                return n;
            }
            *status = U_MEMORY_ALLOCATION_ERROR;
        }
        return 0;
    }

    static UBool unreg(UCurrRegistryKey key) {
        UBool found = FALSE;
        umtx_lock(&gCRegLock);

        CReg** p = &gCRegHead;
        while (*p) {
            if (*p == key) {
                *p = ((CReg*)key)->next;
                delete (CReg*)key;
                found = TRUE;
                break;
            }
            p = &((*p)->next);
        }

        umtx_unlock(&gCRegLock);
        return found;
    }

    static const UChar* get(const char* id) {
        const UChar* result = NULL;
        umtx_lock(&gCRegLock);
        CReg* p = gCRegHead;

        /* register cleanup of the mutex */
        ucln_common_registerCleanup(UCLN_COMMON_CURRENCY, currency_cleanup);
        while (p) {
            if (uprv_strcmp(id, p->id) == 0) {
                result = p->iso;
                break;
            }
            p = p->next;
        }
        umtx_unlock(&gCRegLock);
        return result;
    }

    /* This doesn't need to be thread safe. It's for u_cleanup only. */
    static void cleanup(void) {
        while (gCRegHead) {
            CReg* n = gCRegHead;
            gCRegHead = gCRegHead->next;
            delete n;
        }
    }
};

// -------------------------------------

U_CAPI UCurrRegistryKey U_EXPORT2
ucurr_register(const UChar* isoCode, const char* locale, UErrorCode *status)
{
    if (status && U_SUCCESS(*status)) {
        char id[ULOC_FULLNAME_CAPACITY];
        idForLocale(locale, id, sizeof(id), status);
        return CReg::reg(isoCode, id, status);
    }
    return NULL;
}

// -------------------------------------

U_CAPI UBool U_EXPORT2
ucurr_unregister(UCurrRegistryKey key, UErrorCode* status)
{
    if (status && U_SUCCESS(*status)) {
        return CReg::unreg(key);
    }
    return FALSE;
}
#endif /* UCONFIG_NO_SERVICE */

// -------------------------------------

/**
 * Release all static memory held by currency.
 */
/*The declaration here is needed so currency_cleanup(void)
 * can call this function.
 */
static UBool U_CALLCONV
currency_cache_cleanup(void);

U_CDECL_BEGIN
static UBool U_CALLCONV currency_cleanup(void) {
#if !UCONFIG_NO_SERVICE
    CReg::cleanup();
#endif
    /*
     * There might be some cached currency data or isoCodes data.
     */
    currency_cache_cleanup();
    isoCodes_cleanup();
    currSymbolsEquiv_cleanup();

    return TRUE;
}
U_CDECL_END

// -------------------------------------

U_CAPI int32_t U_EXPORT2
ucurr_forLocale(const char* locale,
                UChar* buff,
                int32_t buffCapacity,
                UErrorCode* ec) {
    if (U_FAILURE(*ec)) { return 0; }
    if (buffCapacity < 0 || (buff == nullptr && buffCapacity > 0)) {
        *ec = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    char currency[4];  // ISO currency codes are alpha3 codes.
    UErrorCode localStatus = U_ZERO_ERROR;
    int32_t resLen = uloc_getKeywordValue(locale, "currency",
                                          currency, UPRV_LENGTHOF(currency), &localStatus);
    if (U_SUCCESS(localStatus) && resLen == 3 && uprv_isInvariantString(currency, resLen)) {
        if (resLen < buffCapacity) {
            T_CString_toUpperCase(currency);
            u_charsToUChars(currency, buff, resLen);
        }
        return u_terminateUChars(buff, buffCapacity, resLen, ec);
    }

    // get country or country_variant in `id'
    char id[ULOC_FULLNAME_CAPACITY];
    idForLocale(locale, id, UPRV_LENGTHOF(id), ec);
    if (U_FAILURE(*ec)) {
        return 0;
    }

#if !UCONFIG_NO_SERVICE
    const UChar* result = CReg::get(id);
    if (result) {
        if(buffCapacity > u_strlen(result)) {
            u_strcpy(buff, result);
        }
        resLen = u_strlen(result);
        return u_terminateUChars(buff, buffCapacity, resLen, ec);
    }
#endif
    // Remove variants, which is only needed for registration.
    char *idDelim = uprv_strchr(id, VAR_DELIM);
    if (idDelim) {
        idDelim[0] = 0;
    }

    const UChar* s = NULL;  // Currency code from data file.
    if (id[0] == 0) {
        // No point looking in the data for an empty string.
        // This is what we would get.
        localStatus = U_MISSING_RESOURCE_ERROR;
    } else {
        // Look up the CurrencyMap element in the root bundle.
        localStatus = U_ZERO_ERROR;
        UResourceBundle *rb = ures_openDirect(U_ICUDATA_CURR, CURRENCY_DATA, &localStatus);
        UResourceBundle *cm = ures_getByKey(rb, CURRENCY_MAP, rb, &localStatus);
        UResourceBundle *countryArray = ures_getByKey(rb, id, cm, &localStatus);
        UResourceBundle *currencyReq = ures_getByIndex(countryArray, 0, NULL, &localStatus);
        s = ures_getStringByKey(currencyReq, "id", &resLen, &localStatus);
        ures_close(currencyReq);
        ures_close(countryArray);
    }

    if ((U_FAILURE(localStatus)) && strchr(id, '_') != 0) {
        // We don't know about it.  Check to see if we support the variant.
        uloc_getParent(locale, id, UPRV_LENGTHOF(id), ec);
        *ec = U_USING_FALLBACK_WARNING;
        // TODO: Loop over the shortened id rather than recursing and
        // looking again for a currency keyword.
        return ucurr_forLocale(id, buff, buffCapacity, ec);
    }
    if (*ec == U_ZERO_ERROR || localStatus != U_ZERO_ERROR) {
        // There is nothing to fallback to. Report the failure/warning if possible.
        *ec = localStatus;
    }
    if (U_SUCCESS(*ec)) {
        if(buffCapacity > resLen) {
            u_strcpy(buff, s);
        }
    }
    return u_terminateUChars(buff, buffCapacity, resLen, ec);
}

// end registration

/**
 * Modify the given locale name by removing the rightmost _-delimited
 * element.  If there is none, empty the string ("" == root).
 * NOTE: The string "root" is not recognized; do not use it.
 * @return TRUE if the fallback happened; FALSE if locale is already
 * root ("").
 */
static UBool fallback(char *loc) {
    if (!*loc) {
        return FALSE;
    }
    UErrorCode status = U_ZERO_ERROR;
    if (uprv_strcmp(loc, "en_GB") == 0) {
        // HACK: See #13368.  We need "en_GB" to fall back to "en_001" instead of "en"
        // in order to consume the correct data strings.  This hack will be removed
        // when proper data sink loading is implemented here.
        // NOTE: "001" adds 1 char over "GB".  However, both call sites allocate
        // arrays with length ULOC_FULLNAME_CAPACITY (plenty of room for en_001).
        uprv_strcpy(loc + 3, "001");
    } else {
        uloc_getParent(loc, loc, (int32_t)uprv_strlen(loc), &status);
    }
 /*
    char *i = uprv_strrchr(loc, '_');
    if (i == NULL) {
        i = loc;
    }
    *i = 0;
 */
    return TRUE;
}


U_CAPI const UChar* U_EXPORT2
ucurr_getName(const UChar* currency,
              const char* locale,
              UCurrNameStyle nameStyle,
              UBool* isChoiceFormat, // fillin
              int32_t* len, // fillin
              UErrorCode* ec) {

    // Look up the Currencies resource for the given locale.  The
    // Currencies locale data looks like this:
    //|en {
    //|  Currencies {
    //|    USD { "US$", "US Dollar" }
    //|    CHF { "Sw F", "Swiss Franc" }
    //|    INR { "=0#Rs|1#Re|1<Rs", "=0#Rupees|1#Rupee|1<Rupees" }
    //|    //...
    //|  }
    //|}

    if (U_FAILURE(*ec)) {
        return 0;
    }

    int32_t choice = (int32_t) nameStyle;
    if (choice < 0 || choice > 2) {
        *ec = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    // In the future, resource bundles may implement multi-level
    // fallback.  That is, if a currency is not found in the en_US
    // Currencies data, then the en Currencies data will be searched.
    // Currently, if a Currencies datum exists in en_US and en, the
    // en_US entry hides that in en.

    // We want multi-level fallback for this resource, so we implement
    // it manually.

    // Use a separate UErrorCode here that does not propagate out of
    // this function.
    UErrorCode ec2 = U_ZERO_ERROR;

    char loc[ULOC_FULLNAME_CAPACITY];
    uloc_getName(locale, loc, sizeof(loc), &ec2);
    if (U_FAILURE(ec2) || ec2 == U_STRING_NOT_TERMINATED_WARNING) {
        *ec = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    char buf[ISO_CURRENCY_CODE_LENGTH+1];
    myUCharsToChars(buf, currency);

    /* Normalize the keyword value to uppercase */
    T_CString_toUpperCase(buf);

    const UChar* s = NULL;
    ec2 = U_ZERO_ERROR;
    LocalUResourceBundlePointer rb(ures_open(U_ICUDATA_CURR, loc, &ec2));

    if (nameStyle == UCURR_NARROW_SYMBOL_NAME) {
        CharString key;
        key.append(CURRENCIES_NARROW, ec2);
        key.append("/", ec2);
        key.append(buf, ec2);
        s = ures_getStringByKeyWithFallback(rb.getAlias(), key.data(), len, &ec2);
        if (ec2 == U_MISSING_RESOURCE_ERROR) {
            *ec = U_USING_FALLBACK_WARNING;
            ec2 = U_ZERO_ERROR;
            choice = UCURR_SYMBOL_NAME;
        }
    }
    if (s == NULL) {
        ures_getByKey(rb.getAlias(), CURRENCIES, rb.getAlias(), &ec2);
        ures_getByKeyWithFallback(rb.getAlias(), buf, rb.getAlias(), &ec2);
        s = ures_getStringByIndex(rb.getAlias(), choice, len, &ec2);
    }

    // If we've succeeded we're done.  Otherwise, try to fallback.
    // If that fails (because we are already at root) then exit.
    if (U_SUCCESS(ec2)) {
        if (ec2 == U_USING_DEFAULT_WARNING
            || (ec2 == U_USING_FALLBACK_WARNING && *ec != U_USING_DEFAULT_WARNING)) {
            *ec = ec2;
        }
    }

    // We no longer support choice format data in names.  Data should not contain
    // choice patterns.
    if (isChoiceFormat != NULL) {
        *isChoiceFormat = FALSE;
    }
    if (U_SUCCESS(ec2)) {
        U_ASSERT(s != NULL);
        return s;
    }

    // If we fail to find a match, use the ISO 4217 code
    *len = u_strlen(currency); // Should == ISO_CURRENCY_CODE_LENGTH, but maybe not...?
    *ec = U_USING_DEFAULT_WARNING;
    return currency;
}

U_CAPI const UChar* U_EXPORT2
ucurr_getPluralName(const UChar* currency,
                    const char* locale,
                    UBool* isChoiceFormat,
                    const char* pluralCount,
                    int32_t* len, // fillin
                    UErrorCode* ec) {
    // Look up the Currencies resource for the given locale.  The
    // Currencies locale data looks like this:
    //|en {
    //|  CurrencyPlurals {
    //|    USD{
    //|      one{"US dollar"}
    //|      other{"US dollars"}
    //|    }
    //|  }
    //|}

    if (U_FAILURE(*ec)) {
        return 0;
    }

    // Use a separate UErrorCode here that does not propagate out of
    // this function.
    UErrorCode ec2 = U_ZERO_ERROR;

    char loc[ULOC_FULLNAME_CAPACITY];
    uloc_getName(locale, loc, sizeof(loc), &ec2);
    if (U_FAILURE(ec2) || ec2 == U_STRING_NOT_TERMINATED_WARNING) {
        *ec = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    char buf[ISO_CURRENCY_CODE_LENGTH+1];
    myUCharsToChars(buf, currency);

    const UChar* s = NULL;
    ec2 = U_ZERO_ERROR;
    UResourceBundle* rb = ures_open(U_ICUDATA_CURR, loc, &ec2);

    rb = ures_getByKey(rb, CURRENCYPLURALS, rb, &ec2);

    // Fetch resource with multi-level resource inheritance fallback
    rb = ures_getByKeyWithFallback(rb, buf, rb, &ec2);

    s = ures_getStringByKeyWithFallback(rb, pluralCount, len, &ec2);
    if (U_FAILURE(ec2)) {
        //  fall back to "other"
        ec2 = U_ZERO_ERROR;
        s = ures_getStringByKeyWithFallback(rb, "other", len, &ec2);
        if (U_FAILURE(ec2)) {
            ures_close(rb);
            // fall back to long name in Currencies
            return ucurr_getName(currency, locale, UCURR_LONG_NAME,
                                 isChoiceFormat, len, ec);
        }
    }
    ures_close(rb);

    // If we've succeeded we're done.  Otherwise, try to fallback.
    // If that fails (because we are already at root) then exit.
    if (U_SUCCESS(ec2)) {
        if (ec2 == U_USING_DEFAULT_WARNING
            || (ec2 == U_USING_FALLBACK_WARNING && *ec != U_USING_DEFAULT_WARNING)) {
            *ec = ec2;
        }
        U_ASSERT(s != NULL);
        return s;
    }

    // If we fail to find a match, use the ISO 4217 code
    *len = u_strlen(currency); // Should == ISO_CURRENCY_CODE_LENGTH, but maybe not...?
    *ec = U_USING_DEFAULT_WARNING;
    return currency;
}


//========================================================================
// Following are structure and function for parsing currency names

#define NEED_TO_BE_DELETED 0x1

// TODO: a better way to define this?
#define MAX_CURRENCY_NAME_LEN 100

typedef struct {
    const char* IsoCode;  // key
    UChar* currencyName;  // value
    int32_t currencyNameLen;  // value length
    int32_t flag;  // flags
} CurrencyNameStruct;


#ifndef MIN
#define MIN(a,b) (((a)<(b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) (((a)<(b)) ? (b) : (a))
#endif


// Comparason function used in quick sort.
static int U_CALLCONV currencyNameComparator(const void* a, const void* b) {
    const CurrencyNameStruct* currName_1 = (const CurrencyNameStruct*)a;
    const CurrencyNameStruct* currName_2 = (const CurrencyNameStruct*)b;
    for (int32_t i = 0;
         i < MIN(currName_1->currencyNameLen, currName_2->currencyNameLen);
         ++i) {
        if (currName_1->currencyName[i] < currName_2->currencyName[i]) {
            return -1;
        }
        if (currName_1->currencyName[i] > currName_2->currencyName[i]) {
            return 1;
        }
    }
    if (currName_1->currencyNameLen < currName_2->currencyNameLen) {
        return -1;
    } else if (currName_1->currencyNameLen > currName_2->currencyNameLen) {
        return 1;
    }
    return 0;
}


// Give a locale, return the maximum number of currency names associated with
// this locale.
// It gets currency names from resource bundles using fallback.
// It is the maximum number because in the fallback chain, some of the
// currency names are duplicated.
// For example, given locale as "en_US", the currency names get from resource
// bundle in "en_US" and "en" are duplicated. The fallback mechanism will count
// all currency names in "en_US" and "en".
static void
getCurrencyNameCount(const char* loc, int32_t* total_currency_name_count, int32_t* total_currency_symbol_count) {
    U_NAMESPACE_USE
    *total_currency_name_count = 0;
    *total_currency_symbol_count = 0;
    const UChar* s = NULL;
    char locale[ULOC_FULLNAME_CAPACITY];
    uprv_strcpy(locale, loc);
    const icu::Hashtable *currencySymbolsEquiv = getCurrSymbolsEquiv();
    for (;;) {
        UErrorCode ec2 = U_ZERO_ERROR;
        // TODO: ures_openDirect?
        UResourceBundle* rb = ures_open(U_ICUDATA_CURR, locale, &ec2);
        UResourceBundle* curr = ures_getByKey(rb, CURRENCIES, NULL, &ec2);
        int32_t n = ures_getSize(curr);
        for (int32_t i=0; i<n; ++i) {
            UResourceBundle* names = ures_getByIndex(curr, i, NULL, &ec2);
            int32_t len;
            s = ures_getStringByIndex(names, UCURR_SYMBOL_NAME, &len, &ec2);
            ++(*total_currency_symbol_count);  // currency symbol
            if (currencySymbolsEquiv != NULL) {
                *total_currency_symbol_count += countEquivalent(*currencySymbolsEquiv, UnicodeString(TRUE, s, len));
            }
            ++(*total_currency_symbol_count); // iso code
            ++(*total_currency_name_count); // long name
            ures_close(names);
        }

        // currency plurals
        UErrorCode ec3 = U_ZERO_ERROR;
        UResourceBundle* curr_p = ures_getByKey(rb, CURRENCYPLURALS, NULL, &ec3);
        n = ures_getSize(curr_p);
        for (int32_t i=0; i<n; ++i) {
            UResourceBundle* names = ures_getByIndex(curr_p, i, NULL, &ec3);
            *total_currency_name_count += ures_getSize(names);
            ures_close(names);
        }
        ures_close(curr_p);
        ures_close(curr);
        ures_close(rb);

        if (!fallback(locale)) {
            break;
        }
    }
}

static UChar*
toUpperCase(const UChar* source, int32_t len, const char* locale) {
    UChar* dest = NULL;
    UErrorCode ec = U_ZERO_ERROR;
    int32_t destLen = u_strToUpper(dest, 0, source, len, locale, &ec);

    ec = U_ZERO_ERROR;
    dest = (UChar*)uprv_malloc(sizeof(UChar) * MAX(destLen, len));
    u_strToUpper(dest, destLen, source, len, locale, &ec);
    if (U_FAILURE(ec)) {
        u_memcpy(dest, source, len);
    }
    return dest;
}


// Collect all available currency names associated with the given locale
// (enable fallback chain).
// Read currenc names defined in resource bundle "Currencies" and
// "CurrencyPlural", enable fallback chain.
// return the malloc-ed currency name arrays and the total number of currency
// names in the array.
static void
collectCurrencyNames(const char* locale,
                     CurrencyNameStruct** currencyNames,
                     int32_t* total_currency_name_count,
                     CurrencyNameStruct** currencySymbols,
                     int32_t* total_currency_symbol_count,
                     UErrorCode& ec) {
    U_NAMESPACE_USE
    const icu::Hashtable *currencySymbolsEquiv = getCurrSymbolsEquiv();
    // Look up the Currencies resource for the given locale.
    UErrorCode ec2 = U_ZERO_ERROR;

    char loc[ULOC_FULLNAME_CAPACITY];
    uloc_getName(locale, loc, sizeof(loc), &ec2);
    if (U_FAILURE(ec2) || ec2 == U_STRING_NOT_TERMINATED_WARNING) {
        ec = U_ILLEGAL_ARGUMENT_ERROR;
    }

    // Get maximum currency name count first.
    getCurrencyNameCount(loc, total_currency_name_count, total_currency_symbol_count);

    *currencyNames = (CurrencyNameStruct*)uprv_malloc
        (sizeof(CurrencyNameStruct) * (*total_currency_name_count));
    *currencySymbols = (CurrencyNameStruct*)uprv_malloc
        (sizeof(CurrencyNameStruct) * (*total_currency_symbol_count));

    if(currencyNames == NULL || currencySymbols == NULL) {
      ec = U_MEMORY_ALLOCATION_ERROR;
    }

    if (U_FAILURE(ec)) return;

    const UChar* s = NULL;  // currency name
    char* iso = NULL;  // currency ISO code

    *total_currency_name_count = 0;
    *total_currency_symbol_count = 0;

    UErrorCode ec3 = U_ZERO_ERROR;
    UErrorCode ec4 = U_ZERO_ERROR;

    // Using hash to remove duplicates caused by locale fallback
    UHashtable* currencyIsoCodes = uhash_open(uhash_hashChars, uhash_compareChars, NULL, &ec3);
    UHashtable* currencyPluralIsoCodes = uhash_open(uhash_hashChars, uhash_compareChars, NULL, &ec4);
    for (int32_t localeLevel = 0; ; ++localeLevel) {
        ec2 = U_ZERO_ERROR;
        // TODO: ures_openDirect
        UResourceBundle* rb = ures_open(U_ICUDATA_CURR, loc, &ec2);
        UResourceBundle* curr = ures_getByKey(rb, CURRENCIES, NULL, &ec2);
        int32_t n = ures_getSize(curr);
        for (int32_t i=0; i<n; ++i) {
            UResourceBundle* names = ures_getByIndex(curr, i, NULL, &ec2);
            int32_t len;
            s = ures_getStringByIndex(names, UCURR_SYMBOL_NAME, &len, &ec2);
            // TODO: uhash_put wont change key/value?
            iso = (char*)ures_getKey(names);
            if (localeLevel == 0) {
                uhash_put(currencyIsoCodes, iso, iso, &ec3);
            } else {
                if (uhash_get(currencyIsoCodes, iso) != NULL) {
                    ures_close(names);
                    continue;
                } else {
                    uhash_put(currencyIsoCodes, iso, iso, &ec3);
                }
            }
            // Add currency symbol.
            (*currencySymbols)[*total_currency_symbol_count].IsoCode = iso;
            (*currencySymbols)[*total_currency_symbol_count].currencyName = (UChar*)s;
            (*currencySymbols)[*total_currency_symbol_count].flag = 0;
            (*currencySymbols)[(*total_currency_symbol_count)++].currencyNameLen = len;
            // Add equivalent symbols
            if (currencySymbolsEquiv != NULL) {
                UnicodeString str(TRUE, s, len);
                icu::EquivIterator iter(*currencySymbolsEquiv, str);
                const UnicodeString *symbol;
                while ((symbol = iter.next()) != NULL) {
                    (*currencySymbols)[*total_currency_symbol_count].IsoCode = iso;
                    (*currencySymbols)[*total_currency_symbol_count].currencyName =
                        const_cast<UChar*>(symbol->getBuffer());
                    (*currencySymbols)[*total_currency_symbol_count].flag = 0;
                    (*currencySymbols)[(*total_currency_symbol_count)++].currencyNameLen = symbol->length();
                }
            }

            // Add currency long name.
            s = ures_getStringByIndex(names, UCURR_LONG_NAME, &len, &ec2);
            (*currencyNames)[*total_currency_name_count].IsoCode = iso;
            UChar* upperName = toUpperCase(s, len, locale);
            (*currencyNames)[*total_currency_name_count].currencyName = upperName;
            (*currencyNames)[*total_currency_name_count].flag = NEED_TO_BE_DELETED;
            (*currencyNames)[(*total_currency_name_count)++].currencyNameLen = len;

            // put (iso, 3, and iso) in to array
            // Add currency ISO code.
            (*currencySymbols)[*total_currency_symbol_count].IsoCode = iso;
            (*currencySymbols)[*total_currency_symbol_count].currencyName = (UChar*)uprv_malloc(sizeof(UChar)*3);
            // Must convert iso[] into Unicode
            u_charsToUChars(iso, (*currencySymbols)[*total_currency_symbol_count].currencyName, 3);
            (*currencySymbols)[*total_currency_symbol_count].flag = NEED_TO_BE_DELETED;
            (*currencySymbols)[(*total_currency_symbol_count)++].currencyNameLen = 3;

            ures_close(names);
        }

        // currency plurals
        UErrorCode ec5 = U_ZERO_ERROR;
        UResourceBundle* curr_p = ures_getByKey(rb, CURRENCYPLURALS, NULL, &ec5);
        n = ures_getSize(curr_p);
        for (int32_t i=0; i<n; ++i) {
            UResourceBundle* names = ures_getByIndex(curr_p, i, NULL, &ec5);
            iso = (char*)ures_getKey(names);
            // Using hash to remove duplicated ISO codes in fallback chain.
            if (localeLevel == 0) {
                uhash_put(currencyPluralIsoCodes, iso, iso, &ec4);
            } else {
                if (uhash_get(currencyPluralIsoCodes, iso) != NULL) {
                    ures_close(names);
                    continue;
                } else {
                    uhash_put(currencyPluralIsoCodes, iso, iso, &ec4);
                }
            }
            int32_t num = ures_getSize(names);
            int32_t len;
            for (int32_t j = 0; j < num; ++j) {
                // TODO: remove duplicates between singular name and
                // currency long name?
                s = ures_getStringByIndex(names, j, &len, &ec5);
                (*currencyNames)[*total_currency_name_count].IsoCode = iso;
                UChar* upperName = toUpperCase(s, len, locale);
                (*currencyNames)[*total_currency_name_count].currencyName = upperName;
                (*currencyNames)[*total_currency_name_count].flag = NEED_TO_BE_DELETED;
                (*currencyNames)[(*total_currency_name_count)++].currencyNameLen = len;
            }
            ures_close(names);
        }
        ures_close(curr_p);
        ures_close(curr);
        ures_close(rb);

        if (!fallback(loc)) {
            break;
        }
    }

    uhash_close(currencyIsoCodes);
    uhash_close(currencyPluralIsoCodes);

    // quick sort the struct
    qsort(*currencyNames, *total_currency_name_count,
          sizeof(CurrencyNameStruct), currencyNameComparator);
    qsort(*currencySymbols, *total_currency_symbol_count,
          sizeof(CurrencyNameStruct), currencyNameComparator);

#ifdef UCURR_DEBUG
    printf("currency name count: %d\n", *total_currency_name_count);
    for (int32_t index = 0; index < *total_currency_name_count; ++index) {
        printf("index: %d\n", index);
        printf("iso: %s\n", (*currencyNames)[index].IsoCode);
        char curNameBuf[1024];
        memset(curNameBuf, 0, 1024);
        u_austrncpy(curNameBuf, (*currencyNames)[index].currencyName, (*currencyNames)[index].currencyNameLen);
        printf("currencyName: %s\n", curNameBuf);
        printf("len: %d\n", (*currencyNames)[index].currencyNameLen);
    }
    printf("currency symbol count: %d\n", *total_currency_symbol_count);
    for (int32_t index = 0; index < *total_currency_symbol_count; ++index) {
        printf("index: %d\n", index);
        printf("iso: %s\n", (*currencySymbols)[index].IsoCode);
        char curNameBuf[1024];
        memset(curNameBuf, 0, 1024);
        u_austrncpy(curNameBuf, (*currencySymbols)[index].currencyName, (*currencySymbols)[index].currencyNameLen);
        printf("currencySymbol: %s\n", curNameBuf);
        printf("len: %d\n", (*currencySymbols)[index].currencyNameLen);
    }
#endif
    // fail on hashtable errors
    if (U_FAILURE(ec3)) {
      ec = ec3;
      return;
    }
    if (U_FAILURE(ec4)) {
      ec = ec4;
      return;
    }
}

// @param  currencyNames: currency names array
// @param  indexInCurrencyNames: the index of the character in currency names
//         array against which the comparison is done
// @param  key: input text char to compare against
// @param  begin(IN/OUT): the begin index of matching range in currency names array
// @param  end(IN/OUT): the end index of matching range in currency names array.
static int32_t
binarySearch(const CurrencyNameStruct* currencyNames,
             int32_t indexInCurrencyNames,
             const UChar key,
             int32_t* begin, int32_t* end) {
#ifdef UCURR_DEBUG
    printf("key = %x\n", key);
#endif
   int32_t first = *begin;
   int32_t last = *end;
   while (first <= last) {
       int32_t mid = (first + last) / 2;  // compute mid point.
       if (indexInCurrencyNames >= currencyNames[mid].currencyNameLen) {
           first = mid + 1;
       } else {
           if (key > currencyNames[mid].currencyName[indexInCurrencyNames]) {
               first = mid + 1;
           }
           else if (key < currencyNames[mid].currencyName[indexInCurrencyNames]) {
               last = mid - 1;
           }
           else {
                // Find a match, and looking for ranges
                // Now do two more binary searches. First, on the left side for
                // the greatest L such that CurrencyNameStruct[L] < key.
                int32_t L = *begin;
                int32_t R = mid;

#ifdef UCURR_DEBUG
                printf("mid = %d\n", mid);
#endif
                while (L < R) {
                    int32_t M = (L + R) / 2;
#ifdef UCURR_DEBUG
                    printf("L = %d, R = %d, M = %d\n", L, R, M);
#endif
                    if (indexInCurrencyNames >= currencyNames[M].currencyNameLen) {
                        L = M + 1;
                    } else {
                        if (currencyNames[M].currencyName[indexInCurrencyNames] < key) {
                            L = M + 1;
                        } else {
#ifdef UCURR_DEBUG
                            U_ASSERT(currencyNames[M].currencyName[indexInCurrencyNames] == key);
#endif
                            R = M;
                        }
                    }
                }
#ifdef UCURR_DEBUG
                U_ASSERT(L == R);
#endif
                *begin = L;
#ifdef UCURR_DEBUG
                printf("begin = %d\n", *begin);
                U_ASSERT(currencyNames[*begin].currencyName[indexInCurrencyNames] == key);
#endif

                // Now for the second search, finding the least R such that
                // key < CurrencyNameStruct[R].
                L = mid;
                R = *end;
                while (L < R) {
                    int32_t M = (L + R) / 2;
#ifdef UCURR_DEBUG
                    printf("L = %d, R = %d, M = %d\n", L, R, M);
#endif
                    if (currencyNames[M].currencyNameLen < indexInCurrencyNames) {
                        L = M + 1;
                    } else {
                        if (currencyNames[M].currencyName[indexInCurrencyNames] > key) {
                            R = M;
                        } else {
#ifdef UCURR_DEBUG
                            U_ASSERT(currencyNames[M].currencyName[indexInCurrencyNames] == key);
#endif
                            L = M + 1;
                        }
                    }
                }
#ifdef UCURR_DEBUG
                U_ASSERT(L == R);
#endif
                if (currencyNames[R].currencyName[indexInCurrencyNames] > key) {
                    *end = R - 1;
                } else {
                    *end = R;
                }
#ifdef UCURR_DEBUG
                printf("end = %d\n", *end);
#endif

                // now, found the range. check whether there is exact match
                if (currencyNames[*begin].currencyNameLen == indexInCurrencyNames + 1) {
                    return *begin;  // find range and exact match.
                }
                return -1;  // find range, but no exact match.
           }
       }
   }
   *begin = -1;
   *end = -1;
   return -1;    // failed to find range.
}


// Linear search "text" in "currencyNames".
// @param  begin, end: the begin and end index in currencyNames, within which
//         range should the search be performed.
// @param  textLen: the length of the text to be compared
// @param  maxMatchLen(IN/OUT): passing in the computed max matching length
//                              pass out the new max  matching length
// @param  maxMatchIndex: the index in currencyName which has the longest
//                        match with input text.
static void
linearSearch(const CurrencyNameStruct* currencyNames,
             int32_t begin, int32_t end,
             const UChar* text, int32_t textLen,
             int32_t *partialMatchLen,
             int32_t *maxMatchLen, int32_t* maxMatchIndex) {
    int32_t initialPartialMatchLen = *partialMatchLen;
    for (int32_t index = begin; index <= end; ++index) {
        int32_t len = currencyNames[index].currencyNameLen;
        if (len > *maxMatchLen && len <= textLen &&
            uprv_memcmp(currencyNames[index].currencyName, text, len * sizeof(UChar)) == 0) {
            *partialMatchLen = MAX(*partialMatchLen, len);
            *maxMatchIndex = index;
            *maxMatchLen = len;
#ifdef UCURR_DEBUG
            printf("maxMatchIndex = %d, maxMatchLen = %d\n",
                   *maxMatchIndex, *maxMatchLen);
#endif
        } else {
            // Check for partial matches.
            for (int32_t i=initialPartialMatchLen; i<MIN(len, textLen); i++) {
                if (currencyNames[index].currencyName[i] != text[i]) {
                    break;
                }
                *partialMatchLen = MAX(*partialMatchLen, i + 1);
            }
        }
    }
}

#define LINEAR_SEARCH_THRESHOLD 10

// Find longest match between "text" and currency names in "currencyNames".
// @param  total_currency_count: total number of currency names in CurrencyNames.
// @param  textLen: the length of the text to be compared
// @param  maxMatchLen: passing in the computed max matching length
//                              pass out the new max  matching length
// @param  maxMatchIndex: the index in currencyName which has the longest
//                        match with input text.
static void
searchCurrencyName(const CurrencyNameStruct* currencyNames,
                   int32_t total_currency_count,
                   const UChar* text, int32_t textLen,
                   int32_t *partialMatchLen,
                   int32_t* maxMatchLen, int32_t* maxMatchIndex) {
    *maxMatchIndex = -1;
    *maxMatchLen = 0;
    int32_t matchIndex = -1;
    int32_t binarySearchBegin = 0;
    int32_t binarySearchEnd = total_currency_count - 1;
    // It is a variant of binary search.
    // For example, given the currency names in currencyNames array are:
    // A AB ABC AD AZ B BB BBEX BBEXYZ BS C D E....
    // and the input text is BBEXST
    // The first round binary search search "B" in the text against
    // the first char in currency names, and find the first char matching range
    // to be "B BB BBEX BBEXYZ BS" (and the maximum matching "B").
    // The 2nd round binary search search the second "B" in the text against
    // the 2nd char in currency names, and narrow the matching range to
    // "BB BBEX BBEXYZ" (and the maximum matching "BB").
    // The 3rd round returnes the range as "BBEX BBEXYZ" (without changing
    // maximum matching).
    // The 4th round returns the same range (the maximum matching is "BBEX").
    // The 5th round returns no matching range.
    for (int32_t index = 0; index < textLen; ++index) {
        // matchIndex saves the one with exact match till the current point.
        // [binarySearchBegin, binarySearchEnd] saves the matching range.
        matchIndex = binarySearch(currencyNames, index,
                                  text[index],
                                  &binarySearchBegin, &binarySearchEnd);
        if (binarySearchBegin == -1) { // did not find the range
            break;
        }
        *partialMatchLen = MAX(*partialMatchLen, index + 1);
        if (matchIndex != -1) {
            // find an exact match for text from text[0] to text[index]
            // in currencyNames array.
            *maxMatchLen = index + 1;
            *maxMatchIndex = matchIndex;
        }
        if (binarySearchEnd - binarySearchBegin < LINEAR_SEARCH_THRESHOLD) {
            // linear search if within threshold.
            linearSearch(currencyNames, binarySearchBegin, binarySearchEnd,
                         text, textLen,
                         partialMatchLen,
                         maxMatchLen, maxMatchIndex);
            break;
        }
    }
    return;
}

//========================= currency name cache =====================
typedef struct {
    char locale[ULOC_FULLNAME_CAPACITY];  //key
    // currency names, case insensitive
    CurrencyNameStruct* currencyNames;  // value
    int32_t totalCurrencyNameCount;  // currency name count
    // currency symbols and ISO code, case sensitive
    CurrencyNameStruct* currencySymbols; // value
    int32_t totalCurrencySymbolCount;  // count
    // reference count.
    // reference count is set to 1 when an entry is put to cache.
    // it increases by 1 before accessing, and decreased by 1 after accessing.
    // The entry is deleted when ref count is zero, which means
    // the entry is replaced out of cache and no process is accessing it.
    int32_t refCount;
} CurrencyNameCacheEntry;


#define CURRENCY_NAME_CACHE_NUM 10

// Reserve 10 cache entries.
static CurrencyNameCacheEntry* currCache[CURRENCY_NAME_CACHE_NUM] = {NULL};
// Using an index to indicate which entry to be replaced when cache is full.
// It is a simple round-robin replacement strategy.
static int8_t currentCacheEntryIndex = 0;

static UMutex gCurrencyCacheMutex;

// Cache deletion
static void
deleteCurrencyNames(CurrencyNameStruct* currencyNames, int32_t count) {
    for (int32_t index = 0; index < count; ++index) {
        if ( (currencyNames[index].flag & NEED_TO_BE_DELETED) ) {
            uprv_free(currencyNames[index].currencyName);
        }
    }
    uprv_free(currencyNames);
}


static void
deleteCacheEntry(CurrencyNameCacheEntry* entry) {
    deleteCurrencyNames(entry->currencyNames, entry->totalCurrencyNameCount);
    deleteCurrencyNames(entry->currencySymbols, entry->totalCurrencySymbolCount);
    uprv_free(entry);
}


// Cache clean up
static UBool U_CALLCONV
currency_cache_cleanup(void) {
    for (int32_t i = 0; i < CURRENCY_NAME_CACHE_NUM; ++i) {
        if (currCache[i]) {
            deleteCacheEntry(currCache[i]);
            currCache[i] = 0;
        }
    }
    return TRUE;
}


/**
 * Loads the currency name data from the cache, or from resource bundles if necessary.
 * The refCount is automatically incremented.  It is the caller's responsibility
 * to decrement it when done!
 */
static CurrencyNameCacheEntry*
getCacheEntry(const char* locale, UErrorCode& ec) {

    int32_t total_currency_name_count = 0;
    CurrencyNameStruct* currencyNames = NULL;
    int32_t total_currency_symbol_count = 0;
    CurrencyNameStruct* currencySymbols = NULL;
    CurrencyNameCacheEntry* cacheEntry = NULL;

    umtx_lock(&gCurrencyCacheMutex);
    // in order to handle racing correctly,
    // not putting 'search' in a separate function.
    int8_t found = -1;
    for (int8_t i = 0; i < CURRENCY_NAME_CACHE_NUM; ++i) {
        if (currCache[i]!= NULL &&
            uprv_strcmp(locale, currCache[i]->locale) == 0) {
            found = i;
            break;
        }
    }
    if (found != -1) {
        cacheEntry = currCache[found];
        ++(cacheEntry->refCount);
    }
    umtx_unlock(&gCurrencyCacheMutex);
    if (found == -1) {
        collectCurrencyNames(locale, &currencyNames, &total_currency_name_count, &currencySymbols, &total_currency_symbol_count, ec);
        if (U_FAILURE(ec)) {
            return NULL;
        }
        umtx_lock(&gCurrencyCacheMutex);
        // check again.
        for (int8_t i = 0; i < CURRENCY_NAME_CACHE_NUM; ++i) {
            if (currCache[i]!= NULL &&
                uprv_strcmp(locale, currCache[i]->locale) == 0) {
                found = i;
                break;
            }
        }
        if (found == -1) {
            // insert new entry to
            // currentCacheEntryIndex % CURRENCY_NAME_CACHE_NUM
            // and remove the existing entry
            // currentCacheEntryIndex % CURRENCY_NAME_CACHE_NUM
            // from cache.
            cacheEntry = currCache[currentCacheEntryIndex];
            if (cacheEntry) {
                --(cacheEntry->refCount);
                // delete if the ref count is zero
                if (cacheEntry->refCount == 0) {
                    deleteCacheEntry(cacheEntry);
                }
            }
            cacheEntry = (CurrencyNameCacheEntry*)uprv_malloc(sizeof(CurrencyNameCacheEntry));
            currCache[currentCacheEntryIndex] = cacheEntry;
            uprv_strcpy(cacheEntry->locale, locale);
            cacheEntry->currencyNames = currencyNames;
            cacheEntry->totalCurrencyNameCount = total_currency_name_count;
            cacheEntry->currencySymbols = currencySymbols;
            cacheEntry->totalCurrencySymbolCount = total_currency_symbol_count;
            cacheEntry->refCount = 2; // one for cache, one for reference
            currentCacheEntryIndex = (currentCacheEntryIndex + 1) % CURRENCY_NAME_CACHE_NUM;
            ucln_common_registerCleanup(UCLN_COMMON_CURRENCY, currency_cleanup);
        } else {
            deleteCurrencyNames(currencyNames, total_currency_name_count);
            deleteCurrencyNames(currencySymbols, total_currency_symbol_count);
            cacheEntry = currCache[found];
            ++(cacheEntry->refCount);
        }
        umtx_unlock(&gCurrencyCacheMutex);
    }

    return cacheEntry;
}

static void releaseCacheEntry(CurrencyNameCacheEntry* cacheEntry) {
    umtx_lock(&gCurrencyCacheMutex);
    --(cacheEntry->refCount);
    if (cacheEntry->refCount == 0) {  // remove
        deleteCacheEntry(cacheEntry);
    }
    umtx_unlock(&gCurrencyCacheMutex);
}

U_CAPI void
uprv_parseCurrency(const char* locale,
                   const icu::UnicodeString& text,
                   icu::ParsePosition& pos,
                   int8_t type,
                   int32_t* partialMatchLen,
                   UChar* result,
                   UErrorCode& ec) {
    U_NAMESPACE_USE
    if (U_FAILURE(ec)) {
        return;
    }
    CurrencyNameCacheEntry* cacheEntry = getCacheEntry(locale, ec);
    if (U_FAILURE(ec)) {
        return;
    }

    int32_t total_currency_name_count = cacheEntry->totalCurrencyNameCount;
    CurrencyNameStruct* currencyNames = cacheEntry->currencyNames;
    int32_t total_currency_symbol_count = cacheEntry->totalCurrencySymbolCount;
    CurrencyNameStruct* currencySymbols = cacheEntry->currencySymbols;

    int32_t start = pos.getIndex();

    UChar inputText[MAX_CURRENCY_NAME_LEN];
    UChar upperText[MAX_CURRENCY_NAME_LEN];
    int32_t textLen = MIN(MAX_CURRENCY_NAME_LEN, text.length() - start);
    text.extract(start, textLen, inputText);
    UErrorCode ec1 = U_ZERO_ERROR;
    textLen = u_strToUpper(upperText, MAX_CURRENCY_NAME_LEN, inputText, textLen, locale, &ec1);

    // Make sure partialMatchLen is initialized
    *partialMatchLen = 0;

    int32_t max = 0;
    int32_t matchIndex = -1;
    // case in-sensitive comparision against currency names
    searchCurrencyName(currencyNames, total_currency_name_count,
                       upperText, textLen, partialMatchLen, &max, &matchIndex);

#ifdef UCURR_DEBUG
    printf("search in names, max = %d, matchIndex = %d\n", max, matchIndex);
#endif

    int32_t maxInSymbol = 0;
    int32_t matchIndexInSymbol = -1;
    if (type != UCURR_LONG_NAME) {  // not name only
        // case sensitive comparison against currency symbols and ISO code.
        searchCurrencyName(currencySymbols, total_currency_symbol_count,
                           inputText, textLen,
                           partialMatchLen,
                           &maxInSymbol, &matchIndexInSymbol);
    }

#ifdef UCURR_DEBUG
    printf("search in symbols, maxInSymbol = %d, matchIndexInSymbol = %d\n", maxInSymbol, matchIndexInSymbol);
    if(matchIndexInSymbol != -1) {
      printf("== ISO=%s\n", currencySymbols[matchIndexInSymbol].IsoCode);
    }
#endif

    if (max >= maxInSymbol && matchIndex != -1) {
        u_charsToUChars(currencyNames[matchIndex].IsoCode, result, 4);
        pos.setIndex(start + max);
    } else if (maxInSymbol >= max && matchIndexInSymbol != -1) {
        u_charsToUChars(currencySymbols[matchIndexInSymbol].IsoCode, result, 4);
        pos.setIndex(start + maxInSymbol);
    }

    // decrease reference count
    releaseCacheEntry(cacheEntry);
}

void uprv_currencyLeads(const char* locale, icu::UnicodeSet& result, UErrorCode& ec) {
    U_NAMESPACE_USE
    if (U_FAILURE(ec)) {
        return;
    }
    CurrencyNameCacheEntry* cacheEntry = getCacheEntry(locale, ec);
    if (U_FAILURE(ec)) {
        return;
    }

    for (int32_t i=0; i<cacheEntry->totalCurrencySymbolCount; i++) {
        const CurrencyNameStruct& info = cacheEntry->currencySymbols[i];
        UChar32 cp;
        U16_GET(info.currencyName, 0, 0, info.currencyNameLen, cp);
        result.add(cp);
    }

    for (int32_t i=0; i<cacheEntry->totalCurrencyNameCount; i++) {
        const CurrencyNameStruct& info = cacheEntry->currencyNames[i];
        UChar32 cp;
        U16_GET(info.currencyName, 0, 0, info.currencyNameLen, cp);
        result.add(cp);
    }

    // decrease reference count
    releaseCacheEntry(cacheEntry);
}


/**
 * Internal method.  Given a currency ISO code and a locale, return
 * the "static" currency name.  This is usually the same as the
 * UCURR_SYMBOL_NAME, but if the latter is a choice format, then the
 * format is applied to the number 2.0 (to yield the more common
 * plural) to return a static name.
 *
 * This is used for backward compatibility with old currency logic in
 * DecimalFormat and DecimalFormatSymbols.
 */
U_CAPI void
uprv_getStaticCurrencyName(const UChar* iso, const char* loc,
                           icu::UnicodeString& result, UErrorCode& ec)
{
    U_NAMESPACE_USE

    int32_t len;
    const UChar* currname = ucurr_getName(iso, loc, UCURR_SYMBOL_NAME,
                                          nullptr /* isChoiceFormat */, &len, &ec);
    if (U_SUCCESS(ec)) {
        result.setTo(currname, len);
    }
}

U_CAPI int32_t U_EXPORT2
ucurr_getDefaultFractionDigits(const UChar* currency, UErrorCode* ec) {
    return ucurr_getDefaultFractionDigitsForUsage(currency,UCURR_USAGE_STANDARD,ec);
}

U_DRAFT int32_t U_EXPORT2
ucurr_getDefaultFractionDigitsForUsage(const UChar* currency, const UCurrencyUsage usage, UErrorCode* ec) {
    int32_t fracDigits = 0;
    if (U_SUCCESS(*ec)) {
        switch (usage) {
            case UCURR_USAGE_STANDARD:
                fracDigits = (_findMetaData(currency, *ec))[0];
                break;
            case UCURR_USAGE_CASH:
                fracDigits = (_findMetaData(currency, *ec))[2];
                break;
            default:
                *ec = U_UNSUPPORTED_ERROR;
        }
    }
    return fracDigits;
}

U_CAPI double U_EXPORT2
ucurr_getRoundingIncrement(const UChar* currency, UErrorCode* ec) {
    return ucurr_getRoundingIncrementForUsage(currency, UCURR_USAGE_STANDARD, ec);
}

U_DRAFT double U_EXPORT2
ucurr_getRoundingIncrementForUsage(const UChar* currency, const UCurrencyUsage usage, UErrorCode* ec) {
    double result = 0.0;

    const int32_t *data = _findMetaData(currency, *ec);
    if (U_SUCCESS(*ec)) {
        int32_t fracDigits;
        int32_t increment;
        switch (usage) {
            case UCURR_USAGE_STANDARD:
                fracDigits = data[0];
                increment = data[1];
                break;
            case UCURR_USAGE_CASH:
                fracDigits = data[2];
                increment = data[3];
                break;
            default:
                *ec = U_UNSUPPORTED_ERROR;
                return result;
        }

        // If the meta data is invalid, return 0.0
        if (fracDigits < 0 || fracDigits > MAX_POW10) {
            *ec = U_INVALID_FORMAT_ERROR;
        } else {
            // A rounding value of 0 or 1 indicates no rounding.
            if (increment >= 2) {
                // Return (increment) / 10^(fracDigits).  The only actual rounding data,
                // as of this writing, is CHF { 2, 5 }.
                result = double(increment) / POW10[fracDigits];
            }
        }
    }

    return result;
}

U_CDECL_BEGIN

typedef struct UCurrencyContext {
    uint32_t currType; /* UCurrCurrencyType */
    uint32_t listIdx;
} UCurrencyContext;

/*
Please keep this list in alphabetical order.
You can look at the CLDR supplemental data or ISO-4217 for the meaning of some
of these items.
ISO-4217: http://www.iso.org/iso/en/prods-services/popstds/currencycodeslist.html
*/
static const struct CurrencyList {
    const char *currency;
    uint32_t currType;
} gCurrencyList[] = {
    {"ADP", UCURR_COMMON|UCURR_DEPRECATED},
    {"AED", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"AFA", UCURR_COMMON|UCURR_DEPRECATED},
    {"AFN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"ALK", UCURR_COMMON|UCURR_DEPRECATED},
    {"ALL", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"AMD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"ANG", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"AOA", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"AOK", UCURR_COMMON|UCURR_DEPRECATED},
    {"AON", UCURR_COMMON|UCURR_DEPRECATED},
    {"AOR", UCURR_COMMON|UCURR_DEPRECATED},
    {"ARA", UCURR_COMMON|UCURR_DEPRECATED},
    {"ARL", UCURR_COMMON|UCURR_DEPRECATED},
    {"ARM", UCURR_COMMON|UCURR_DEPRECATED},
    {"ARP", UCURR_COMMON|UCURR_DEPRECATED},
    {"ARS", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"ATS", UCURR_COMMON|UCURR_DEPRECATED},
    {"AUD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"AWG", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"AZM", UCURR_COMMON|UCURR_DEPRECATED},
    {"AZN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BAD", UCURR_COMMON|UCURR_DEPRECATED},
    {"BAM", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BAN", UCURR_COMMON|UCURR_DEPRECATED},
    {"BBD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BDT", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BEC", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"BEF", UCURR_COMMON|UCURR_DEPRECATED},
    {"BEL", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"BGL", UCURR_COMMON|UCURR_DEPRECATED},
    {"BGM", UCURR_COMMON|UCURR_DEPRECATED},
    {"BGN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BGO", UCURR_COMMON|UCURR_DEPRECATED},
    {"BHD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BIF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BMD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BND", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BOB", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BOL", UCURR_COMMON|UCURR_DEPRECATED},
    {"BOP", UCURR_COMMON|UCURR_DEPRECATED},
    {"BOV", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"BRB", UCURR_COMMON|UCURR_DEPRECATED},
    {"BRC", UCURR_COMMON|UCURR_DEPRECATED},
    {"BRE", UCURR_COMMON|UCURR_DEPRECATED},
    {"BRL", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BRN", UCURR_COMMON|UCURR_DEPRECATED},
    {"BRR", UCURR_COMMON|UCURR_DEPRECATED},
    {"BRZ", UCURR_COMMON|UCURR_DEPRECATED},
    {"BSD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BTN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BUK", UCURR_COMMON|UCURR_DEPRECATED},
    {"BWP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BYB", UCURR_COMMON|UCURR_DEPRECATED},
    {"BYN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"BYR", UCURR_COMMON|UCURR_DEPRECATED},
    {"BZD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"CAD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"CDF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"CHE", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"CHF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"CHW", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"CLE", UCURR_COMMON|UCURR_DEPRECATED},
    {"CLF", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"CLP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"CNH", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"CNX", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"CNY", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"COP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"COU", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"CRC", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"CSD", UCURR_COMMON|UCURR_DEPRECATED},
    {"CSK", UCURR_COMMON|UCURR_DEPRECATED},
    {"CUC", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"CUP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"CVE", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"CYP", UCURR_COMMON|UCURR_DEPRECATED},
    {"CZK", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"DDM", UCURR_COMMON|UCURR_DEPRECATED},
    {"DEM", UCURR_COMMON|UCURR_DEPRECATED},
    {"DJF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"DKK", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"DOP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"DZD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"ECS", UCURR_COMMON|UCURR_DEPRECATED},
    {"ECV", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"EEK", UCURR_COMMON|UCURR_DEPRECATED},
    {"EGP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"EQE", UCURR_COMMON|UCURR_DEPRECATED}, // questionable, remove?
    {"ERN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"ESA", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"ESB", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"ESP", UCURR_COMMON|UCURR_DEPRECATED},
    {"ETB", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"EUR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"FIM", UCURR_COMMON|UCURR_DEPRECATED},
    {"FJD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"FKP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"FRF", UCURR_COMMON|UCURR_DEPRECATED},
    {"GBP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"GEK", UCURR_COMMON|UCURR_DEPRECATED},
    {"GEL", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"GHC", UCURR_COMMON|UCURR_DEPRECATED},
    {"GHS", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"GIP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"GMD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"GNF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"GNS", UCURR_COMMON|UCURR_DEPRECATED},
    {"GQE", UCURR_COMMON|UCURR_DEPRECATED},
    {"GRD", UCURR_COMMON|UCURR_DEPRECATED},
    {"GTQ", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"GWE", UCURR_COMMON|UCURR_DEPRECATED},
    {"GWP", UCURR_COMMON|UCURR_DEPRECATED},
    {"GYD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"HKD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"HNL", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"HRD", UCURR_COMMON|UCURR_DEPRECATED},
    {"HRK", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"HTG", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"HUF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"IDR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"IEP", UCURR_COMMON|UCURR_DEPRECATED},
    {"ILP", UCURR_COMMON|UCURR_DEPRECATED},
    {"ILR", UCURR_COMMON|UCURR_DEPRECATED},
    {"ILS", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"INR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"IQD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"IRR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"ISJ", UCURR_COMMON|UCURR_DEPRECATED},
    {"ISK", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"ITL", UCURR_COMMON|UCURR_DEPRECATED},
    {"JMD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"JOD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"JPY", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"KES", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"KGS", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"KHR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"KMF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"KPW", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"KRH", UCURR_COMMON|UCURR_DEPRECATED},
    {"KRO", UCURR_COMMON|UCURR_DEPRECATED},
    {"KRW", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"KWD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"KYD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"KZT", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"LAK", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"LBP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"LKR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"LRD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"LSL", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"LSM", UCURR_COMMON|UCURR_DEPRECATED}, // questionable, remove?
    {"LTL", UCURR_COMMON|UCURR_DEPRECATED},
    {"LTT", UCURR_COMMON|UCURR_DEPRECATED},
    {"LUC", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"LUF", UCURR_COMMON|UCURR_DEPRECATED},
    {"LUL", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"LVL", UCURR_COMMON|UCURR_DEPRECATED},
    {"LVR", UCURR_COMMON|UCURR_DEPRECATED},
    {"LYD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MAD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MAF", UCURR_COMMON|UCURR_DEPRECATED},
    {"MCF", UCURR_COMMON|UCURR_DEPRECATED},
    {"MDC", UCURR_COMMON|UCURR_DEPRECATED},
    {"MDL", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MGA", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MGF", UCURR_COMMON|UCURR_DEPRECATED},
    {"MKD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MKN", UCURR_COMMON|UCURR_DEPRECATED},
    {"MLF", UCURR_COMMON|UCURR_DEPRECATED},
    {"MMK", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MNT", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MOP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MRO", UCURR_COMMON|UCURR_DEPRECATED},
    {"MRU", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MTL", UCURR_COMMON|UCURR_DEPRECATED},
    {"MTP", UCURR_COMMON|UCURR_DEPRECATED},
    {"MUR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MVP", UCURR_COMMON|UCURR_DEPRECATED}, // questionable, remove?
    {"MVR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MWK", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MXN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MXP", UCURR_COMMON|UCURR_DEPRECATED},
    {"MXV", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"MYR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"MZE", UCURR_COMMON|UCURR_DEPRECATED},
    {"MZM", UCURR_COMMON|UCURR_DEPRECATED},
    {"MZN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"NAD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"NGN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"NIC", UCURR_COMMON|UCURR_DEPRECATED},
    {"NIO", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"NLG", UCURR_COMMON|UCURR_DEPRECATED},
    {"NOK", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"NPR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"NZD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"OMR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"PAB", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"PEI", UCURR_COMMON|UCURR_DEPRECATED},
    {"PEN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"PES", UCURR_COMMON|UCURR_DEPRECATED},
    {"PGK", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"PHP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"PKR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"PLN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"PLZ", UCURR_COMMON|UCURR_DEPRECATED},
    {"PTE", UCURR_COMMON|UCURR_DEPRECATED},
    {"PYG", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"QAR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"RHD", UCURR_COMMON|UCURR_DEPRECATED},
    {"ROL", UCURR_COMMON|UCURR_DEPRECATED},
    {"RON", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"RSD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"RUB", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"RUR", UCURR_COMMON|UCURR_DEPRECATED},
    {"RWF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SAR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SBD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SCR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SDD", UCURR_COMMON|UCURR_DEPRECATED},
    {"SDG", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SDP", UCURR_COMMON|UCURR_DEPRECATED},
    {"SEK", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SGD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SHP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SIT", UCURR_COMMON|UCURR_DEPRECATED},
    {"SKK", UCURR_COMMON|UCURR_DEPRECATED},
    {"SLL", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SOS", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SRD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SRG", UCURR_COMMON|UCURR_DEPRECATED},
    {"SSP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"STD", UCURR_COMMON|UCURR_DEPRECATED},
    {"STN", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SUR", UCURR_COMMON|UCURR_DEPRECATED},
    {"SVC", UCURR_COMMON|UCURR_DEPRECATED},
    {"SYP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"SZL", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"THB", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"TJR", UCURR_COMMON|UCURR_DEPRECATED},
    {"TJS", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"TMM", UCURR_COMMON|UCURR_DEPRECATED},
    {"TMT", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"TND", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"TOP", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"TPE", UCURR_COMMON|UCURR_DEPRECATED},
    {"TRL", UCURR_COMMON|UCURR_DEPRECATED},
    {"TRY", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"TTD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"TWD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"TZS", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"UAH", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"UAK", UCURR_COMMON|UCURR_DEPRECATED},
    {"UGS", UCURR_COMMON|UCURR_DEPRECATED},
    {"UGX", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"USD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"USN", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"USS", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"UYI", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"UYP", UCURR_COMMON|UCURR_DEPRECATED},
    {"UYU", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"UZS", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"VEB", UCURR_COMMON|UCURR_DEPRECATED},
    {"VEF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"VND", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"VNN", UCURR_COMMON|UCURR_DEPRECATED},
    {"VUV", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"WST", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"XAF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"XAG", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XAU", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XBA", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XBB", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XBC", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XBD", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XCD", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"XDR", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XEU", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"XFO", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XFU", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XOF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"XPD", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XPF", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"XPT", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XRE", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"XSU", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XTS", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XUA", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"XXX", UCURR_UNCOMMON|UCURR_NON_DEPRECATED},
    {"YDD", UCURR_COMMON|UCURR_DEPRECATED},
    {"YER", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"YUD", UCURR_COMMON|UCURR_DEPRECATED},
    {"YUM", UCURR_COMMON|UCURR_DEPRECATED},
    {"YUN", UCURR_COMMON|UCURR_DEPRECATED},
    {"YUR", UCURR_COMMON|UCURR_DEPRECATED},
    {"ZAL", UCURR_UNCOMMON|UCURR_DEPRECATED},
    {"ZAR", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"ZMK", UCURR_COMMON|UCURR_DEPRECATED},
    {"ZMW", UCURR_COMMON|UCURR_NON_DEPRECATED},
    {"ZRN", UCURR_COMMON|UCURR_DEPRECATED},
    {"ZRZ", UCURR_COMMON|UCURR_DEPRECATED},
    {"ZWD", UCURR_COMMON|UCURR_DEPRECATED},
    {"ZWL", UCURR_COMMON|UCURR_DEPRECATED},
    {"ZWR", UCURR_COMMON|UCURR_DEPRECATED},
    { NULL, 0 } // Leave here to denote the end of the list.
};

#define UCURR_MATCHES_BITMASK(variable, typeToMatch) \
    ((typeToMatch) == UCURR_ALL || ((variable) & (typeToMatch)) == (typeToMatch))

static int32_t U_CALLCONV
ucurr_countCurrencyList(UEnumeration *enumerator, UErrorCode * /*pErrorCode*/) {
    UCurrencyContext *myContext = (UCurrencyContext *)(enumerator->context);
    uint32_t currType = myContext->currType;
    int32_t count = 0;

    /* Count the number of items matching the type we are looking for. */
    for (int32_t idx = 0; gCurrencyList[idx].currency != NULL; idx++) {
        if (UCURR_MATCHES_BITMASK(gCurrencyList[idx].currType, currType)) {
            count++;
        }
    }
    return count;
}

static const char* U_CALLCONV
ucurr_nextCurrencyList(UEnumeration *enumerator,
                        int32_t* resultLength,
                        UErrorCode * /*pErrorCode*/)
{
    UCurrencyContext *myContext = (UCurrencyContext *)(enumerator->context);

    /* Find the next in the list that matches the type we are looking for. */
    while (myContext->listIdx < UPRV_LENGTHOF(gCurrencyList)-1) {
        const struct CurrencyList *currItem = &gCurrencyList[myContext->listIdx++];
        if (UCURR_MATCHES_BITMASK(currItem->currType, myContext->currType))
        {
            if (resultLength) {
                *resultLength = 3; /* Currency codes are only 3 chars long */
            }
            return currItem->currency;
        }
    }
    /* We enumerated too far. */
    if (resultLength) {
        *resultLength = 0;
    }
    return NULL;
}

static void U_CALLCONV
ucurr_resetCurrencyList(UEnumeration *enumerator, UErrorCode * /*pErrorCode*/) {
    ((UCurrencyContext *)(enumerator->context))->listIdx = 0;
}

static void U_CALLCONV
ucurr_closeCurrencyList(UEnumeration *enumerator) {
    uprv_free(enumerator->context);
    uprv_free(enumerator);
}

static void U_CALLCONV
ucurr_createCurrencyList(UHashtable *isoCodes, UErrorCode* status){
    UErrorCode localStatus = U_ZERO_ERROR;

    // Look up the CurrencyMap element in the root bundle.
    UResourceBundle *rb = ures_openDirect(U_ICUDATA_CURR, CURRENCY_DATA, &localStatus);
    UResourceBundle *currencyMapArray = ures_getByKey(rb, CURRENCY_MAP, rb, &localStatus);

    if (U_SUCCESS(localStatus)) {
        // process each entry in currency map
        for (int32_t i=0; i<ures_getSize(currencyMapArray); i++) {
            // get the currency resource
            UResourceBundle *currencyArray = ures_getByIndex(currencyMapArray, i, NULL, &localStatus);
            // process each currency
            if (U_SUCCESS(localStatus)) {
                for (int32_t j=0; j<ures_getSize(currencyArray); j++) {
                    // get the currency resource
                    UResourceBundle *currencyRes = ures_getByIndex(currencyArray, j, NULL, &localStatus);
                    IsoCodeEntry *entry = (IsoCodeEntry*)uprv_malloc(sizeof(IsoCodeEntry));
                    if (entry == NULL) {
                        *status = U_MEMORY_ALLOCATION_ERROR;
                        return;
                    }

                    // get the ISO code
                    int32_t isoLength = 0;
                    UResourceBundle *idRes = ures_getByKey(currencyRes, "id", NULL, &localStatus);
                    if (idRes == NULL) {
                        continue;
                    }
                    const UChar *isoCode = ures_getString(idRes, &isoLength, &localStatus);

                    // get from date
                    UDate fromDate = U_DATE_MIN;
                    UResourceBundle *fromRes = ures_getByKey(currencyRes, "from", NULL, &localStatus);

                    if (U_SUCCESS(localStatus)) {
                        int32_t fromLength = 0;
                        const int32_t *fromArray = ures_getIntVector(fromRes, &fromLength, &localStatus);
                        int64_t currDate64 = (int64_t)fromArray[0] << 32;
                        currDate64 |= ((int64_t)fromArray[1] & (int64_t)INT64_C(0x00000000FFFFFFFF));
                        fromDate = (UDate)currDate64;
                    }
                    ures_close(fromRes);

                    // get to date
                    UDate toDate = U_DATE_MAX;
                    localStatus = U_ZERO_ERROR;
                    UResourceBundle *toRes = ures_getByKey(currencyRes, "to", NULL, &localStatus);

                    if (U_SUCCESS(localStatus)) {
                        int32_t toLength = 0;
                        const int32_t *toArray = ures_getIntVector(toRes, &toLength, &localStatus);
                        int64_t currDate64 = (int64_t)toArray[0] << 32;
                        currDate64 |= ((int64_t)toArray[1] & (int64_t)INT64_C(0x00000000FFFFFFFF));
                        toDate = (UDate)currDate64;
                    }
                    ures_close(toRes);

                    ures_close(idRes);
                    ures_close(currencyRes);

                    entry->isoCode = isoCode;
                    entry->from = fromDate;
                    entry->to = toDate;

                    localStatus = U_ZERO_ERROR;
                    uhash_put(isoCodes, (UChar *)isoCode, entry, &localStatus);
                }
            } else {
                *status = localStatus;
            }
            ures_close(currencyArray);
        }
    } else {
        *status = localStatus;
    }

    ures_close(currencyMapArray);
}

static const UEnumeration gEnumCurrencyList = {
    NULL,
    NULL,
    ucurr_closeCurrencyList,
    ucurr_countCurrencyList,
    uenum_unextDefault,
    ucurr_nextCurrencyList,
    ucurr_resetCurrencyList
};
U_CDECL_END


static void U_CALLCONV initIsoCodes(UErrorCode &status) {
    U_ASSERT(gIsoCodes == NULL);
    ucln_common_registerCleanup(UCLN_COMMON_CURRENCY, currency_cleanup);

    UHashtable *isoCodes = uhash_open(uhash_hashUChars, uhash_compareUChars, NULL, &status);
    if (U_FAILURE(status)) {
        return;
    }
    uhash_setValueDeleter(isoCodes, deleteIsoCodeEntry);

    ucurr_createCurrencyList(isoCodes, &status);
    if (U_FAILURE(status)) {
        uhash_close(isoCodes);
        return;
    }
    gIsoCodes = isoCodes;  // Note: gIsoCodes is const. Once set up here it is never altered,
                           //       and read only access is safe without synchronization.
}

static void populateCurrSymbolsEquiv(icu::Hashtable *hash, UErrorCode &status) {
    if (U_FAILURE(status)) { return; }
    for (auto& entry : unisets::kCurrencyEntries) {
        UnicodeString exemplar(entry.exemplar);
        const UnicodeSet* set = unisets::get(entry.key);
        if (set == nullptr) { return; }
        UnicodeSetIterator it(*set);
        while (it.next()) {
            UnicodeString value = it.getString();
            if (value == exemplar) {
                // No need to mark the exemplar character as an equivalent
                continue;
            }
            makeEquivalent(exemplar, value, hash, status);
            if (U_FAILURE(status)) { return; }
        }
    }
}

static void U_CALLCONV initCurrSymbolsEquiv() {
    U_ASSERT(gCurrSymbolsEquiv == NULL);
    UErrorCode status = U_ZERO_ERROR;
    ucln_common_registerCleanup(UCLN_COMMON_CURRENCY, currency_cleanup);
    icu::Hashtable *temp = new icu::Hashtable(status);
    if (temp == NULL) {
        return;
    }
    if (U_FAILURE(status)) {
        delete temp;
        return;
    }
    temp->setValueDeleter(deleteUnicode);
    populateCurrSymbolsEquiv(temp, status);
    if (U_FAILURE(status)) {
        delete temp;
        return;
    }
    gCurrSymbolsEquiv = temp;
}

U_CAPI UBool U_EXPORT2
ucurr_isAvailable(const UChar* isoCode, UDate from, UDate to, UErrorCode* eErrorCode) {
    umtx_initOnce(gIsoCodesInitOnce, &initIsoCodes, *eErrorCode);
    if (U_FAILURE(*eErrorCode)) {
        return FALSE;
    }

    IsoCodeEntry* result = (IsoCodeEntry *) uhash_get(gIsoCodes, isoCode);
    if (result == NULL) {
        return FALSE;
    } else if (from > to) {
        *eErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    } else if  ((from > result->to) || (to < result->from)) {
        return FALSE;
    }
    return TRUE;
}

static const icu::Hashtable* getCurrSymbolsEquiv() {
    umtx_initOnce(gCurrSymbolsEquivInitOnce, &initCurrSymbolsEquiv);
    return gCurrSymbolsEquiv;
}

U_CAPI UEnumeration * U_EXPORT2
ucurr_openISOCurrencies(uint32_t currType, UErrorCode *pErrorCode) {
    UEnumeration *myEnum = NULL;
    UCurrencyContext *myContext;

    myEnum = (UEnumeration*)uprv_malloc(sizeof(UEnumeration));
    if (myEnum == NULL) {
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    uprv_memcpy(myEnum, &gEnumCurrencyList, sizeof(UEnumeration));
    myContext = (UCurrencyContext*)uprv_malloc(sizeof(UCurrencyContext));
    if (myContext == NULL) {
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
        uprv_free(myEnum);
        return NULL;
    }
    myContext->currType = currType;
    myContext->listIdx = 0;
    myEnum->context = myContext;
    return myEnum;
}

U_CAPI int32_t U_EXPORT2
ucurr_countCurrencies(const char* locale,
                 UDate date,
                 UErrorCode* ec)
{
    int32_t currCount = 0;

    if (ec != NULL && U_SUCCESS(*ec))
    {
        // local variables
        UErrorCode localStatus = U_ZERO_ERROR;
        char id[ULOC_FULLNAME_CAPACITY];
        uloc_getKeywordValue(locale, "currency", id, ULOC_FULLNAME_CAPACITY, &localStatus);

        // get country or country_variant in `id'
        idForLocale(locale, id, sizeof(id), ec);

        if (U_FAILURE(*ec))
        {
            return 0;
        }

        // Remove variants, which is only needed for registration.
        char *idDelim = strchr(id, VAR_DELIM);
        if (idDelim)
        {
            idDelim[0] = 0;
        }

        // Look up the CurrencyMap element in the root bundle.
        UResourceBundle *rb = ures_openDirect(U_ICUDATA_CURR, CURRENCY_DATA, &localStatus);
        UResourceBundle *cm = ures_getByKey(rb, CURRENCY_MAP, rb, &localStatus);

        // Using the id derived from the local, get the currency data
        UResourceBundle *countryArray = ures_getByKey(rb, id, cm, &localStatus);

        // process each currency to see which one is valid for the given date
        if (U_SUCCESS(localStatus))
        {
            for (int32_t i=0; i<ures_getSize(countryArray); i++)
            {
                // get the currency resource
                UResourceBundle *currencyRes = ures_getByIndex(countryArray, i, NULL, &localStatus);

                // get the from date
                int32_t fromLength = 0;
                UResourceBundle *fromRes = ures_getByKey(currencyRes, "from", NULL, &localStatus);
                const int32_t *fromArray = ures_getIntVector(fromRes, &fromLength, &localStatus);

                int64_t currDate64 = (int64_t)fromArray[0] << 32;
                currDate64 |= ((int64_t)fromArray[1] & (int64_t)INT64_C(0x00000000FFFFFFFF));
                UDate fromDate = (UDate)currDate64;

                if (ures_getSize(currencyRes)> 2)
                {
                    int32_t toLength = 0;
                    UResourceBundle *toRes = ures_getByKey(currencyRes, "to", NULL, &localStatus);
                    const int32_t *toArray = ures_getIntVector(toRes, &toLength, &localStatus);

                    currDate64 = (int64_t)toArray[0] << 32;
                    currDate64 |= ((int64_t)toArray[1] & (int64_t)INT64_C(0x00000000FFFFFFFF));
                    UDate toDate = (UDate)currDate64;

                    if ((fromDate <= date) && (date < toDate))
                    {
                        currCount++;
                    }

                    ures_close(toRes);
                }
                else
                {
                    if (fromDate <= date)
                    {
                        currCount++;
                    }
                }

                // close open resources
                ures_close(currencyRes);
                ures_close(fromRes);

            } // end For loop
        } // end if (U_SUCCESS(localStatus))

        ures_close(countryArray);

        // Check for errors
        if (*ec == U_ZERO_ERROR || localStatus != U_ZERO_ERROR)
        {
            // There is nothing to fallback to.
            // Report the failure/warning if possible.
            *ec = localStatus;
        }

        if (U_SUCCESS(*ec))
        {
            // no errors
            return currCount;
        }

    }

    // If we got here, either error code is invalid or
    // some argument passed is no good.
    return 0;
}

U_CAPI int32_t U_EXPORT2
ucurr_forLocaleAndDate(const char* locale,
                UDate date,
                int32_t index,
                UChar* buff,
                int32_t buffCapacity,
                UErrorCode* ec)
{
    int32_t resLen = 0;
	int32_t currIndex = 0;
    const UChar* s = NULL;

    if (ec != NULL && U_SUCCESS(*ec))
    {
        // check the arguments passed
        if ((buff && buffCapacity) || !buffCapacity )
        {
            // local variables
            UErrorCode localStatus = U_ZERO_ERROR;
            char id[ULOC_FULLNAME_CAPACITY];
            resLen = uloc_getKeywordValue(locale, "currency", id, ULOC_FULLNAME_CAPACITY, &localStatus);

            // get country or country_variant in `id'
            idForLocale(locale, id, sizeof(id), ec);
            if (U_FAILURE(*ec))
            {
                return 0;
            }

            // Remove variants, which is only needed for registration.
            char *idDelim = strchr(id, VAR_DELIM);
            if (idDelim)
            {
                idDelim[0] = 0;
            }

            // Look up the CurrencyMap element in the root bundle.
            UResourceBundle *rb = ures_openDirect(U_ICUDATA_CURR, CURRENCY_DATA, &localStatus);
            UResourceBundle *cm = ures_getByKey(rb, CURRENCY_MAP, rb, &localStatus);

            // Using the id derived from the local, get the currency data
            UResourceBundle *countryArray = ures_getByKey(rb, id, cm, &localStatus);

            // process each currency to see which one is valid for the given date
            bool matchFound = false;
            if (U_SUCCESS(localStatus))
            {
                if ((index <= 0) || (index> ures_getSize(countryArray)))
                {
                    // requested index is out of bounds
                    ures_close(countryArray);
                    return 0;
                }

                for (int32_t i=0; i<ures_getSize(countryArray); i++)
                {
                    // get the currency resource
                    UResourceBundle *currencyRes = ures_getByIndex(countryArray, i, NULL, &localStatus);
                    s = ures_getStringByKey(currencyRes, "id", &resLen, &localStatus);

                    // get the from date
                    int32_t fromLength = 0;
                    UResourceBundle *fromRes = ures_getByKey(currencyRes, "from", NULL, &localStatus);
                    const int32_t *fromArray = ures_getIntVector(fromRes, &fromLength, &localStatus);

                    int64_t currDate64 = (int64_t)fromArray[0] << 32;
                    currDate64 |= ((int64_t)fromArray[1] & (int64_t)INT64_C(0x00000000FFFFFFFF));
                    UDate fromDate = (UDate)currDate64;

                    if (ures_getSize(currencyRes)> 2)
                    {
                        int32_t toLength = 0;
                        UResourceBundle *toRes = ures_getByKey(currencyRes, "to", NULL, &localStatus);
                        const int32_t *toArray = ures_getIntVector(toRes, &toLength, &localStatus);

                        currDate64 = (int64_t)toArray[0] << 32;
                        currDate64 |= ((int64_t)toArray[1] & (int64_t)INT64_C(0x00000000FFFFFFFF));
                        UDate toDate = (UDate)currDate64;

                        if ((fromDate <= date) && (date < toDate))
                        {
                            currIndex++;
                            if (currIndex == index)
                            {
                                matchFound = true;
                            }
                        }

                        ures_close(toRes);
                    }
                    else
                    {
                        if (fromDate <= date)
                        {
                            currIndex++;
                            if (currIndex == index)
                            {
                                matchFound = true;
                            }
                        }
                    }

                    // close open resources
                    ures_close(currencyRes);
                    ures_close(fromRes);

                    // check for loop exit
                    if (matchFound)
                    {
                        break;
                    }

                } // end For loop
            }

            ures_close(countryArray);

            // Check for errors
            if (*ec == U_ZERO_ERROR || localStatus != U_ZERO_ERROR)
            {
                // There is nothing to fallback to.
                // Report the failure/warning if possible.
                *ec = localStatus;
            }

            if (U_SUCCESS(*ec))
            {
                // no errors
                if((buffCapacity> resLen) && matchFound)
                {
                    // write out the currency value
                    u_strcpy(buff, s);
                }
                else
                {
                    return 0;
                }
            }

            // return null terminated currency string
            return u_terminateUChars(buff, buffCapacity, resLen, ec);
        }
        else
        {
            // illegal argument encountered
            *ec = U_ILLEGAL_ARGUMENT_ERROR;
        }

    }

    // If we got here, either error code is invalid or
    // some argument passed is no good.
    return resLen;
}

static const UEnumeration ucurr_defaultKeywordValues = {
    NULL,
    NULL,
    ulist_close_keyword_values_iterator,
    ulist_count_keyword_values,
    uenum_unextDefault,
    ulist_next_keyword_value,
    ulist_reset_keyword_values_iterator
};

U_CAPI UEnumeration *U_EXPORT2 ucurr_getKeywordValuesForLocale(const char *key, const char *locale, UBool commonlyUsed, UErrorCode* status) {
    // Resolve region
    char prefRegion[ULOC_COUNTRY_CAPACITY];
    ulocimp_getRegionForSupplementalData(locale, TRUE, prefRegion, sizeof(prefRegion), status);

    // Read value from supplementalData
    UList *values = ulist_createEmptyList(status);
    UList *otherValues = ulist_createEmptyList(status);
    UEnumeration *en = (UEnumeration *)uprv_malloc(sizeof(UEnumeration));
    if (U_FAILURE(*status) || en == NULL) {
        if (en == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
        } else {
            uprv_free(en);
        }
        ulist_deleteList(values);
        ulist_deleteList(otherValues);
        return NULL;
    }
    memcpy(en, &ucurr_defaultKeywordValues, sizeof(UEnumeration));
    en->context = values;

    UResourceBundle *bundle = ures_openDirect(U_ICUDATA_CURR, "supplementalData", status);
    ures_getByKey(bundle, "CurrencyMap", bundle, status);
    UResourceBundle bundlekey, regbndl, curbndl, to;
    ures_initStackObject(&bundlekey);
    ures_initStackObject(&regbndl);
    ures_initStackObject(&curbndl);
    ures_initStackObject(&to);

    while (U_SUCCESS(*status) && ures_hasNext(bundle)) {
        ures_getNextResource(bundle, &bundlekey, status);
        if (U_FAILURE(*status)) {
            break;
        }
        const char *region = ures_getKey(&bundlekey);
        UBool isPrefRegion = uprv_strcmp(region, prefRegion) == 0 ? TRUE : FALSE;
        if (!isPrefRegion && commonlyUsed) {
            // With commonlyUsed=true, we do not put
            // currencies for other regions in the
            // result list.
            continue;
        }
        ures_getByKey(bundle, region, &regbndl, status);
        if (U_FAILURE(*status)) {
            break;
        }
        while (U_SUCCESS(*status) && ures_hasNext(&regbndl)) {
            ures_getNextResource(&regbndl, &curbndl, status);
            if (ures_getType(&curbndl) != URES_TABLE) {
                // Currently, an empty ARRAY is mixed in.
                continue;
            }
            char *curID = (char *)uprv_malloc(sizeof(char) * ULOC_KEYWORDS_CAPACITY);
            int32_t curIDLength = ULOC_KEYWORDS_CAPACITY;
            if (curID == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                break;
            }

#if U_CHARSET_FAMILY==U_ASCII_FAMILY
            ures_getUTF8StringByKey(&curbndl, "id", curID, &curIDLength, TRUE, status);
            /* optimize - use the utf-8 string */
#else
            {
                       const UChar* defString = ures_getStringByKey(&curbndl, "id", &curIDLength, status);
                       if(U_SUCCESS(*status)) {
			   if(curIDLength+1 > ULOC_KEYWORDS_CAPACITY) {
				*status = U_BUFFER_OVERFLOW_ERROR;
			   } else {
                           	u_UCharsToChars(defString, curID, curIDLength+1);
			   }
                       }
            }
#endif

            if (U_FAILURE(*status)) {
                break;
            }
            UBool hasTo = FALSE;
            ures_getByKey(&curbndl, "to", &to, status);
            if (U_FAILURE(*status)) {
                // Do nothing here...
                *status = U_ZERO_ERROR;
            } else {
                hasTo = TRUE;
            }
            if (isPrefRegion && !hasTo && !ulist_containsString(values, curID, (int32_t)uprv_strlen(curID))) {
                // Currently active currency for the target country
                ulist_addItemEndList(values, curID, TRUE, status);
            } else if (!ulist_containsString(otherValues, curID, (int32_t)uprv_strlen(curID)) && !commonlyUsed) {
                ulist_addItemEndList(otherValues, curID, TRUE, status);
            } else {
                uprv_free(curID);
            }
        }

    }
    if (U_SUCCESS(*status)) {
        if (commonlyUsed) {
            if (ulist_getListSize(values) == 0) {
                // This could happen if no valid region is supplied in the input
                // locale. In this case, we use the CLDR's default.
                uenum_close(en);
                en = ucurr_getKeywordValuesForLocale(key, "und", TRUE, status);
            }
        } else {
            // Consolidate the list
            char *value = NULL;
            ulist_resetList(otherValues);
            while ((value = (char *)ulist_getNext(otherValues)) != NULL) {
                if (!ulist_containsString(values, value, (int32_t)uprv_strlen(value))) {
                    char *tmpValue = (char *)uprv_malloc(sizeof(char) * ULOC_KEYWORDS_CAPACITY);
                    uprv_memcpy(tmpValue, value, uprv_strlen(value) + 1);
                    ulist_addItemEndList(values, tmpValue, TRUE, status);
                    if (U_FAILURE(*status)) {
                        break;
                    }
                }
            }
        }

        ulist_resetList((UList *)(en->context));
    } else {
        ulist_deleteList(values);
        uprv_free(en);
        values = NULL;
        en = NULL;
    }
    ures_close(&to);
    ures_close(&curbndl);
    ures_close(&regbndl);
    ures_close(&bundlekey);
    ures_close(bundle);

    ulist_deleteList(otherValues);

    return en;
}


U_CAPI int32_t U_EXPORT2
ucurr_getNumericCode(const UChar* currency) {
    int32_t code = 0;
    if (currency && u_strlen(currency) == ISO_CURRENCY_CODE_LENGTH) {
        UErrorCode status = U_ZERO_ERROR;

        UResourceBundle *bundle = ures_openDirect(0, "currencyNumericCodes", &status);
        ures_getByKey(bundle, "codeMap", bundle, &status);
        if (U_SUCCESS(status)) {
            char alphaCode[ISO_CURRENCY_CODE_LENGTH+1];
            myUCharsToChars(alphaCode, currency);
            T_CString_toUpperCase(alphaCode);
            ures_getByKey(bundle, alphaCode, bundle, &status);
            int tmpCode = ures_getInt(bundle, &status);
            if (U_SUCCESS(status)) {
                code = tmpCode;
            }
        }
        ures_close(bundle);
    }
    return code;
}
#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  udata.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999oct25
*   created by: Markus W. Scherer
*/

  /* U_PLATFORM etc. */

#ifdef __GNUC__
/* if gcc
#define ATTRIBUTE_WEAK __attribute__ ((weak))
might have to #include some other header
*/
#endif


















/***********************************************************************
*
*   Notes on the organization of the ICU data implementation
*
*      All of the public API is defined in udata.h
*
*      The implementation is split into several files...
*
*         - udata.c  (this file) contains higher level code that knows about
*                     the search paths for locating data, caching opened data, etc.
*
*         - umapfile.c  contains the low level platform-specific code for actually loading
*                     (memory mapping, file reading, whatever) data into memory.
*
*         - ucmndata.c  deals with the tables of contents of ICU data items within
*                     an ICU common format data file.  The implementation includes
*                     an abstract interface and support for multiple TOC formats.
*                     All knowledge of any specific TOC format is encapsulated here.
*
*         - udatamem.c has code for managing UDataMemory structs.  These are little
*                     descriptor objects for blocks of memory holding ICU data of
*                     various types.
*/

/* configuration ---------------------------------------------------------- */

/* If you are excruciatingly bored turn this on .. */
/* #define UDATA_DEBUG 1 */

#if defined(UDATA_DEBUG)
#   include <stdio.h>
#endif

U_NAMESPACE_USE

/*
 *  Forward declarations
 */
static UDataMemory *udata_findCachedData(const char *path, UErrorCode &err);

/***********************************************************************
*
*    static (Global) data
*
************************************************************************/

/*
 * Pointers to the common ICU data.
 *
 * We store multiple pointers to ICU data packages and iterate through them
 * when looking for a data item.
 *
 * It is possible to combine this with dependency inversion:
 * One or more data package libraries may export
 * functions that each return a pointer to their piece of the ICU data,
 * and this file would import them as weak functions, without a
 * strong linker dependency from the common library on the data library.
 *
 * Then we can have applications depend on only that part of ICU's data
 * that they really need, reducing the size of binaries that take advantage
 * of this.
 */
static UDataMemory *gCommonICUDataArray[10] = { NULL };   // Access protected by icu global mutex.

static u_atomic_int32_t gHaveTriedToLoadCommonData = ATOMIC_INT32_T_INITIALIZER(0);  //  See extendICUData().

static UHashtable  *gCommonDataCache = NULL;  /* Global hash table of opened ICU data files.  */
static icu::UInitOnce gCommonDataCacheInitOnce = U_INITONCE_INITIALIZER;

#if !defined(ICU_DATA_DIR_WINDOWS)
static UDataFileAccess  gDataFileAccess = UDATA_DEFAULT_ACCESS;  // Access not synchronized.
                                                                 // Modifying is documented as thread-unsafe.
#else
// If we are using the Windows data directory, then look in one spot only.
static UDataFileAccess  gDataFileAccess = UDATA_NO_FILES;
#endif

static UBool U_CALLCONV
udata_cleanup(void)
{
    int32_t i;

    if (gCommonDataCache) {             /* Delete the cache of user data mappings.  */
        uhash_close(gCommonDataCache);  /*   Table owns the contents, and will delete them. */
        gCommonDataCache = NULL;        /*   Cleanup is not thread safe.                */
    }
    gCommonDataCacheInitOnce.reset();

    for (i = 0; i < UPRV_LENGTHOF(gCommonICUDataArray) && gCommonICUDataArray[i] != NULL; ++i) {
        udata_close(gCommonICUDataArray[i]);
        gCommonICUDataArray[i] = NULL;
    }
    gHaveTriedToLoadCommonData = 0;

    return TRUE;                   /* Everything was cleaned up */
}

static UBool U_CALLCONV
findCommonICUDataByName(const char *inBasename, UErrorCode &err)
{
    UBool found = FALSE;
    int32_t i;

    UDataMemory  *pData = udata_findCachedData(inBasename, err);
    if (U_FAILURE(err) || pData == NULL)
        return FALSE;

    {
        Mutex lock;
        for (i = 0; i < UPRV_LENGTHOF(gCommonICUDataArray); ++i) {
            if ((gCommonICUDataArray[i] != NULL) && (gCommonICUDataArray[i]->pHeader == pData->pHeader)) {
                /* The data pointer is already in the array. */
                found = TRUE;
                break;
            }
        }
    }
    return found;
}


/*
 * setCommonICUData.   Set a UDataMemory to be the global ICU Data
 */
static UBool
setCommonICUData(UDataMemory *pData,     /*  The new common data.  Belongs to caller, we copy it. */
                 UBool       warn,       /*  If true, set USING_DEFAULT warning if ICUData was    */
                                         /*    changed by another thread before we got to it.     */
                 UErrorCode *pErr)
{
    UDataMemory  *newCommonData = UDataMemory_createNewInstance(pErr);
    int32_t i;
    UBool didUpdate = FALSE;
    if (U_FAILURE(*pErr)) {
        return FALSE;
    }

    /*  For the assignment, other threads must cleanly see either the old            */
    /*    or the new, not some partially initialized new.  The old can not be        */
    /*    deleted - someone may still have a pointer to it lying around in           */
    /*    their locals.                                                              */
    UDatamemory_assign(newCommonData, pData);
    umtx_lock(NULL);
    for (i = 0; i < UPRV_LENGTHOF(gCommonICUDataArray); ++i) {
        if (gCommonICUDataArray[i] == NULL) {
            gCommonICUDataArray[i] = newCommonData;
            didUpdate = TRUE;
            break;
        } else if (gCommonICUDataArray[i]->pHeader == pData->pHeader) {
            /* The same data pointer is already in the array. */
            break;
        }
    }
    umtx_unlock(NULL);

    if (i == UPRV_LENGTHOF(gCommonICUDataArray) && warn) {
        *pErr = U_USING_DEFAULT_WARNING;
    }
    if (didUpdate) {
        ucln_common_registerCleanup(UCLN_COMMON_UDATA, udata_cleanup);
    } else {
        uprv_free(newCommonData);
    }
    return didUpdate;
}

#if !defined(ICU_DATA_DIR_WINDOWS)

static UBool
setCommonICUDataPointer(const void *pData, UBool /*warn*/, UErrorCode *pErrorCode) {
    UDataMemory tData;
    UDataMemory_init(&tData);
    UDataMemory_setData(&tData, pData);
    udata_checkCommonData(&tData, pErrorCode);
    return setCommonICUData(&tData, FALSE, pErrorCode);
}

#endif

static const char *
findBasename(const char *path) {
    const char *basename=uprv_strrchr(path, U_FILE_SEP_CHAR);
    if(basename==NULL) {
        return path;
    } else {
        return basename+1;
    }
}

#ifdef UDATA_DEBUG
static const char *
packageNameFromPath(const char *path)
{
    if((path == NULL) || (*path == 0)) {
        return U_ICUDATA_NAME;
    }

    path = findBasename(path);

    if((path == NULL) || (*path == 0)) {
        return U_ICUDATA_NAME;
    }

    return path;
}
#endif

/*----------------------------------------------------------------------*
 *                                                                      *
 *   Cache for common data                                              *
 *      Functions for looking up or adding entries to a cache of        *
 *      data that has been previously opened.  Avoids a potentially     *
 *      expensive operation of re-opening the data for subsequent       *
 *      uses.                                                           *
 *                                                                      *
 *      Data remains cached for the duration of the process.            *
 *                                                                      *
 *----------------------------------------------------------------------*/

typedef struct DataCacheElement {
    char          *name;
    UDataMemory   *item;
} DataCacheElement;



/*
 * Deleter function for DataCacheElements.
 *         udata cleanup function closes the hash table; hash table in turn calls back to
 *         here for each entry.
 */
static void U_CALLCONV DataCacheElement_deleter(void *pDCEl) {
    DataCacheElement *p = (DataCacheElement *)pDCEl;
    udata_close(p->item);              /* unmaps storage */
    uprv_free(p->name);                /* delete the hash key string. */
    uprv_free(pDCEl);                  /* delete 'this'          */
}

static void U_CALLCONV udata_initHashTable(UErrorCode &err) {
    U_ASSERT(gCommonDataCache == NULL);
    gCommonDataCache = uhash_open(uhash_hashChars, uhash_compareChars, NULL, &err);
    if (U_FAILURE(err)) {
       return;
    }
    U_ASSERT(gCommonDataCache != NULL);
    uhash_setValueDeleter(gCommonDataCache, DataCacheElement_deleter);
    ucln_common_registerCleanup(UCLN_COMMON_UDATA, udata_cleanup);
}

 /*   udata_getCacheHashTable()
  *     Get the hash table used to store the data cache entries.
  *     Lazy create it if it doesn't yet exist.
  */
static UHashtable *udata_getHashTable(UErrorCode &err) {
    umtx_initOnce(gCommonDataCacheInitOnce, &udata_initHashTable, err);
    return gCommonDataCache;
}



static UDataMemory *udata_findCachedData(const char *path, UErrorCode &err)
{
    UHashtable        *htable;
    UDataMemory       *retVal = NULL;
    DataCacheElement  *el;
    const char        *baseName;

    htable = udata_getHashTable(err);
    if (U_FAILURE(err)) {
        return NULL;
    }

    baseName = findBasename(path);   /* Cache remembers only the base name, not the full path. */
    umtx_lock(NULL);
    el = (DataCacheElement *)uhash_get(htable, baseName);
    umtx_unlock(NULL);
    if (el != NULL) {
        retVal = el->item;
    }
#ifdef UDATA_DEBUG
    fprintf(stderr, "Cache: [%s] -> %p\n", baseName, (void*) retVal);
#endif
    return retVal;
}


static UDataMemory *udata_cacheDataItem(const char *path, UDataMemory *item, UErrorCode *pErr) {
    DataCacheElement *newElement;
    const char       *baseName;
    int32_t           nameLen;
    UHashtable       *htable;
    DataCacheElement *oldValue = NULL;
    UErrorCode        subErr = U_ZERO_ERROR;

    htable = udata_getHashTable(*pErr);
    if (U_FAILURE(*pErr)) {
        return NULL;
    }

    /* Create a new DataCacheElement - the thingy we store in the hash table -
     * and copy the supplied path and UDataMemoryItems into it.
     */
    newElement = (DataCacheElement *)uprv_malloc(sizeof(DataCacheElement));
    if (newElement == NULL) {
        *pErr = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    newElement->item = UDataMemory_createNewInstance(pErr);
    if (U_FAILURE(*pErr)) {
        uprv_free(newElement);
        return NULL;
    }
    UDatamemory_assign(newElement->item, item);

    baseName = findBasename(path);
    nameLen = (int32_t)uprv_strlen(baseName);
    newElement->name = (char *)uprv_malloc(nameLen+1);
    if (newElement->name == NULL) {
        *pErr = U_MEMORY_ALLOCATION_ERROR;
        uprv_free(newElement->item);
        uprv_free(newElement);
        return NULL;
    }
    uprv_strcpy(newElement->name, baseName);

    /* Stick the new DataCacheElement into the hash table.
    */
    umtx_lock(NULL);
    oldValue = (DataCacheElement *)uhash_get(htable, path);
    if (oldValue != NULL) {
        subErr = U_USING_DEFAULT_WARNING;
    }
    else {
        uhash_put(
            htable,
            newElement->name,               /* Key   */
            newElement,                     /* Value */
            &subErr);
    }
    umtx_unlock(NULL);

#ifdef UDATA_DEBUG
    fprintf(stderr, "Cache: [%s] <<< %p : %s. vFunc=%p\n", newElement->name,
    (void*) newElement->item, u_errorName(subErr), (void*) newElement->item->vFuncs);
#endif

    if (subErr == U_USING_DEFAULT_WARNING || U_FAILURE(subErr)) {
        *pErr = subErr; /* copy sub err unto fillin ONLY if something happens. */
        uprv_free(newElement->name);
        uprv_free(newElement->item);
        uprv_free(newElement);
        return oldValue ? oldValue->item : NULL;
    }

    return newElement->item;
}

/*----------------------------------------------------------------------*==============
 *                                                                      *
 *  Path management.  Could be shared with other tools/etc if need be   *
 * later on.                                                            *
 *                                                                      *
 *----------------------------------------------------------------------*/

U_NAMESPACE_BEGIN

class UDataPathIterator
{
public:
    UDataPathIterator(const char *path, const char *pkg,
                      const char *item, const char *suffix, UBool doCheckLastFour,
                      UErrorCode *pErrorCode);
    const char *next(UErrorCode *pErrorCode);

private:
    const char *path;                              /* working path (u_icudata_Dir) */
    const char *nextPath;                          /* path following this one */
    const char *basename;                          /* item's basename (icudt22e_mt.res)*/

    StringPiece suffix;                            /* item suffix (can be null) */

    uint32_t    basenameLen;                       /* length of basename */

    CharString  itemPath;                          /* path passed in with item name */
    CharString  pathBuffer;                        /* output path for this it'ion */
    CharString  packageStub;                       /* example:  "/icudt28b". Will ignore that leaf in set paths. */

    UBool       checkLastFour;                     /* if TRUE then allow paths such as '/foo/myapp.dat'
                                                    * to match, checks last 4 chars of suffix with
                                                    * last 4 of path, then previous chars. */
};

/**
 * @param iter    The iterator to be initialized. Its current state does not matter.
 * @param inPath  The full pathname to be iterated over.  If NULL, defaults to U_ICUDATA_NAME
 * @param pkg     Package which is being searched for, ex "icudt28l".  Will ignore leaf directories such as /icudt28l
 * @param item    Item to be searched for.  Can include full path, such as /a/b/foo.dat
 * @param inSuffix  Optional item suffix, if not-null (ex. ".dat") then 'path' can contain 'item' explicitly.
 *             Ex:   'stuff.dat' would be found in '/a/foo:/tmp/stuff.dat:/bar/baz' as item #2.
 *                   '/blarg/stuff.dat' would also be found.
 *  Note: inSuffix may also be the 'item' being searched for as well, (ex: "ibm-5348_P100-1997.cnv"), in which case
 *        the 'item' parameter is often the same as pkg. (Though sometimes might have a tree part as well, ex: "icudt62l-curr").
 */
UDataPathIterator::UDataPathIterator(const char *inPath, const char *pkg,
                                     const char *item, const char *inSuffix, UBool doCheckLastFour,
                                     UErrorCode *pErrorCode)
{
#ifdef UDATA_DEBUG
        fprintf(stderr, "SUFFIX1=%s PATH=%s\n", inSuffix, inPath);
#endif
    /** Path **/
    if(inPath == NULL) {
        path = u_getDataDirectory();
    } else {
        path = inPath;
    }

    /** Package **/
    if(pkg != NULL) {
      packageStub.append(U_FILE_SEP_CHAR, *pErrorCode).append(pkg, *pErrorCode);
#ifdef UDATA_DEBUG
      fprintf(stderr, "STUB=%s [%d]\n", packageStub.data(), packageStub.length());
#endif
    }

    /** Item **/
    basename = findBasename(item);
    basenameLen = (int32_t)uprv_strlen(basename);

    /** Item path **/
    if(basename == item) {
        nextPath = path;
    } else {
        itemPath.append(item, (int32_t)(basename-item), *pErrorCode);
        nextPath = itemPath.data();
    }
#ifdef UDATA_DEBUG
    fprintf(stderr, "SUFFIX=%s [%p]\n", inSuffix, (void*) inSuffix);
#endif

    /** Suffix  **/
    if(inSuffix != NULL) {
        suffix = inSuffix;
    } else {
        suffix = "";
    }

    checkLastFour = doCheckLastFour;

    /* pathBuffer will hold the output path strings returned by this iterator */

#ifdef UDATA_DEBUG
    fprintf(stderr, "0: init %s -> [path=%s], [base=%s], [suff=%s], [itempath=%s], [nextpath=%s], [checklast4=%s]\n",
            item,
            path,
            basename,
            suffix.data(),
            itemPath.data(),
            nextPath,
            checkLastFour?"TRUE":"false");
#endif
}

/**
 * Get the next path on the list.
 *
 * @param iter The Iter to be used
 * @param len  If set, pointer to the length of the returned path, for convenience.
 * @return Pointer to the next path segment, or NULL if there are no more.
 */
const char *UDataPathIterator::next(UErrorCode *pErrorCode)
{
    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }

    const char *currentPath = NULL;
    int32_t     pathLen = 0;
    const char *pathBasename;

    do
    {
        if( nextPath == NULL ) {
            break;
        }
        currentPath = nextPath;

        if(nextPath == itemPath.data()) { /* we were processing item's path. */
            nextPath = path; /* start with regular path next tm. */
            pathLen = (int32_t)uprv_strlen(currentPath);
        } else {
            /* fix up next for next time */
            nextPath = uprv_strchr(currentPath, U_PATH_SEP_CHAR);
            if(nextPath == NULL) {
                /* segment: entire path */
                pathLen = (int32_t)uprv_strlen(currentPath);
            } else {
                /* segment: until next segment */
                pathLen = (int32_t)(nextPath - currentPath);
                /* skip divider */
                nextPath ++;
            }
        }

        if(pathLen == 0) {
            continue;
        }

#ifdef UDATA_DEBUG
        fprintf(stderr, "rest of path (IDD) = %s\n", currentPath);
        fprintf(stderr, "                     ");
        {
            int32_t qqq;
            for(qqq=0;qqq<pathLen;qqq++)
            {
                fprintf(stderr, " ");
            }

            fprintf(stderr, "^\n");
        }
#endif
        pathBuffer.clear().append(currentPath, pathLen, *pErrorCode);

        /* check for .dat files */
        pathBasename = findBasename(pathBuffer.data());

        if(checkLastFour == TRUE &&
           (pathLen>=4) &&
           uprv_strncmp(pathBuffer.data() +(pathLen-4), suffix.data(), 4)==0 && /* suffix matches */
           uprv_strncmp(findBasename(pathBuffer.data()), basename, basenameLen)==0  && /* base matches */
           uprv_strlen(pathBasename)==(basenameLen+4)) { /* base+suffix = full len */

#ifdef UDATA_DEBUG
            fprintf(stderr, "Have %s file on the path: %s\n", suffix.data(), pathBuffer.data());
#endif
            /* do nothing */
        }
        else
        {       /* regular dir path */
            if(pathBuffer[pathLen-1] != U_FILE_SEP_CHAR) {
                if((pathLen>=4) &&
                   uprv_strncmp(pathBuffer.data()+(pathLen-4), ".dat", 4) == 0)
                {
#ifdef UDATA_DEBUG
                    fprintf(stderr, "skipping non-directory .dat file %s\n", pathBuffer.data());
#endif
                    continue;
                }

                /* Check if it is a directory with the same name as our package */
                if(!packageStub.isEmpty() &&
                   (pathLen > packageStub.length()) &&
                   !uprv_strcmp(pathBuffer.data() + pathLen - packageStub.length(), packageStub.data())) {
#ifdef UDATA_DEBUG
                  fprintf(stderr, "Found stub %s (will add package %s of len %d)\n", packageStub.data(), basename, basenameLen);
#endif
                  pathBuffer.truncate(pathLen - packageStub.length());
                }
                pathBuffer.append(U_FILE_SEP_CHAR, *pErrorCode);
            }

            /* + basename */
            pathBuffer.append(packageStub.data()+1, packageStub.length()-1, *pErrorCode);

            if (!suffix.empty())  /* tack on suffix */
            {
                if (suffix.length() > 4) {
                    // If the suffix is actually an item ("ibm-5348_P100-1997.cnv") and not an extension (".res")
                    // then we need to ensure that the path ends with a separator.
                    pathBuffer.ensureEndsWithFileSeparator(*pErrorCode);
                }
                pathBuffer.append(suffix, *pErrorCode);
            }
        }

#ifdef UDATA_DEBUG
        fprintf(stderr, " -->  %s\n", pathBuffer.data());
#endif

        return pathBuffer.data();

    } while(path);

    /* fell way off the end */
    return NULL;
}

U_NAMESPACE_END

/* ==================================================================================*/


/*----------------------------------------------------------------------*
 *                                                                      *
 *  Add a static reference to the common data library                   *
 *   Unless overridden by an explicit udata_setCommonData, this will be *
 *      our common data.                                                *
 *                                                                      *
 *----------------------------------------------------------------------*/
#if !defined(ICU_DATA_DIR_WINDOWS)
// When using the Windows system data, we expect only a single data file.
extern "C" const unsigned char U_DATA_API U_ICUDATA_ENTRY_POINT [];
#endif

/*
 * This would be a good place for weak-linkage declarations of
 * partial-data-library access functions where each returns a pointer
 * to its data package, if it is linked in.
 */
/*
extern const void *uprv_getICUData_collation(void) ATTRIBUTE_WEAK;
extern const void *uprv_getICUData_conversion(void) ATTRIBUTE_WEAK;
*/

/*----------------------------------------------------------------------*
 *                                                                      *
 *   openCommonData   Attempt to open a common format (.dat) file       *
 *                    Map it into memory (if it's not there already)    *
 *                    and return a UDataMemory object for it.           *
 *                                                                      *
 *                    If the requested data is already open and cached  *
 *                       just return the cached UDataMem object.        *
 *                                                                      *
 *----------------------------------------------------------------------*/
static UDataMemory *
openCommonData(const char *path,          /*  Path from OpenChoice?          */
               int32_t commonDataIndex,   /*  ICU Data (index >= 0) if path == NULL */
               UErrorCode *pErrorCode)
{
    UDataMemory tData;
    const char *pathBuffer;
    const char *inBasename;

    if (U_FAILURE(*pErrorCode)) {
        return NULL;
    }

    UDataMemory_init(&tData);

    /* ??????? TODO revisit this */
    if (commonDataIndex >= 0) {
        /* "mini-cache" for common ICU data */
        if(commonDataIndex >= UPRV_LENGTHOF(gCommonICUDataArray)) {
            return NULL;
        }
        {
            Mutex lock;
            if(gCommonICUDataArray[commonDataIndex] != NULL) {
                return gCommonICUDataArray[commonDataIndex];
            }
#if !defined(ICU_DATA_DIR_WINDOWS)
// When using the Windows system data, we expect only a single data file.
            int32_t i;
            for(i = 0; i < commonDataIndex; ++i) {
                if((void*) gCommonICUDataArray[i]->pHeader == (void*) &U_ICUDATA_ENTRY_POINT) {
                    /* The linked-in data is already in the list. */
                    return NULL;
                }
            }
#endif
        }

        /* Add the linked-in data to the list. */
        /*
         * This is where we would check and call weakly linked partial-data-library
         * access functions.
         */
        /*
        if (uprv_getICUData_collation) {
            setCommonICUDataPointer(uprv_getICUData_collation(), FALSE, pErrorCode);
        }
        if (uprv_getICUData_conversion) {
            setCommonICUDataPointer(uprv_getICUData_conversion(), FALSE, pErrorCode);
        }
        */
#if !defined(ICU_DATA_DIR_WINDOWS)
// When using the Windows system data, we expect only a single data file.
        setCommonICUDataPointer(&U_ICUDATA_ENTRY_POINT, FALSE, pErrorCode);
        {
            Mutex lock;
            return gCommonICUDataArray[commonDataIndex];
        }
#endif
    }


    /* request is NOT for ICU Data.  */

    /* Find the base name portion of the supplied path.   */
    /*   inBasename will be left pointing somewhere within the original path string.      */
    inBasename = findBasename(path);
#ifdef UDATA_DEBUG
    fprintf(stderr, "inBasename = %s\n", inBasename);
#endif

    if(*inBasename==0) {
        /* no basename.     This will happen if the original path was a directory name,   */
        /*    like  "a/b/c/".   (Fallback to separate files will still work.)             */
#ifdef UDATA_DEBUG
        fprintf(stderr, "ocd: no basename in %s, bailing.\n", path);
#endif
        if (U_SUCCESS(*pErrorCode)) {
            *pErrorCode=U_FILE_ACCESS_ERROR;
        }
        return NULL;
    }

   /* Is the requested common data file already open and cached?                     */
   /*   Note that the cache is keyed by the base name only.  The rest of the path,   */
   /*     if any, is not considered.                                                 */
    UDataMemory  *dataToReturn = udata_findCachedData(inBasename, *pErrorCode);
    if (dataToReturn != NULL || U_FAILURE(*pErrorCode)) {
        return dataToReturn;
    }

    /* Requested item is not in the cache.
     * Hunt it down, trying all the path locations
     */

    UDataPathIterator iter(u_getDataDirectory(), inBasename, path, ".dat", TRUE, pErrorCode);

    while ((UDataMemory_isLoaded(&tData)==FALSE) && (pathBuffer = iter.next(pErrorCode)) != NULL)
    {
#ifdef UDATA_DEBUG
        fprintf(stderr, "ocd: trying path %s - ", pathBuffer);
#endif
        uprv_mapFile(&tData, pathBuffer, pErrorCode);
#ifdef UDATA_DEBUG
        fprintf(stderr, "%s\n", UDataMemory_isLoaded(&tData)?"LOADED":"not loaded");
#endif
    }
    if (U_FAILURE(*pErrorCode)) {
        return NULL;
    }

#if defined(OS390_STUBDATA) && defined(OS390BATCH)
    if (!UDataMemory_isLoaded(&tData)) {
        char ourPathBuffer[1024];
        /* One more chance, for extendCommonData() */
        uprv_strncpy(ourPathBuffer, path, 1019);
        ourPathBuffer[1019]=0;
        uprv_strcat(ourPathBuffer, ".dat");
        uprv_mapFile(&tData, ourPathBuffer, pErrorCode);
    }
#endif

    if (U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if (!UDataMemory_isLoaded(&tData)) {
        /* no common data */
        *pErrorCode=U_FILE_ACCESS_ERROR;
        return NULL;
    }

    /* we have mapped a file, check its header */
    udata_checkCommonData(&tData, pErrorCode);


    /* Cache the UDataMemory struct for this .dat file,
     *   so we won't need to hunt it down and map it again next time
     *   something is needed from it.                */
    return udata_cacheDataItem(inBasename, &tData, pErrorCode);
}


/*----------------------------------------------------------------------*
 *                                                                      *
 *   extendICUData   If the full set of ICU data was not loaded at      *
 *                   program startup, load it now.  This function will  *
 *                   be called when the lookup of an ICU data item in   *
 *                   the common ICU data fails.                         *
 *                                                                      *
 *                   return true if new data is loaded, false otherwise.*
 *                                                                      *
 *----------------------------------------------------------------------*/
static UBool extendICUData(UErrorCode *pErr)
{
    UDataMemory   *pData;
    UDataMemory   copyPData;
    UBool         didUpdate = FALSE;

    /*
     * There is a chance for a race condition here.
     * Normally, ICU data is loaded from a DLL or via mmap() and
     * setCommonICUData() will detect if the same address is set twice.
     * If ICU is built with data loading via fread() then the address will
     * be different each time the common data is loaded and we may add
     * multiple copies of the data.
     * In this case, use a mutex to prevent the race.
     * Use a specific mutex to avoid nested locks of the global mutex.
     */
#if MAP_IMPLEMENTATION==MAP_STDIO
    static UMutex extendICUDataMutex;
    umtx_lock(&extendICUDataMutex);
#endif
    if(!umtx_loadAcquire(gHaveTriedToLoadCommonData)) {
        /* See if we can explicitly open a .dat file for the ICUData. */
        pData = openCommonData(
                   U_ICUDATA_NAME,            /*  "icudt20l" , for example.          */
                   -1,                        /*  Pretend we're not opening ICUData  */
                   pErr);

        /* How about if there is no pData, eh... */

       UDataMemory_init(&copyPData);
       if(pData != NULL) {
          UDatamemory_assign(&copyPData, pData);
          copyPData.map = 0;              /* The mapping for this data is owned by the hash table */
          copyPData.mapAddr = 0;          /*   which will unmap it when ICU is shut down.         */
                                          /* CommonICUData is also unmapped when ICU is shut down.*/
                                          /* To avoid unmapping the data twice, zero out the map  */
                                          /*   fields in the UDataMemory that we're assigning     */
                                          /*   to CommonICUData.                                  */

          didUpdate = /* no longer using this result */
              setCommonICUData(&copyPData,/*  The new common data.                                */
                       FALSE,             /*  No warnings if write didn't happen                  */
                       pErr);             /*  setCommonICUData honors errors; NOP if error set    */
        }

        umtx_storeRelease(gHaveTriedToLoadCommonData, 1);
    }

    didUpdate = findCommonICUDataByName(U_ICUDATA_NAME, *pErr);  /* Return 'true' when a racing writes out the extended                 */
                                                          /* data after another thread has failed to see it (in openCommonData), so     */
                                                          /* extended data can be examined.                                             */
                                                          /* Also handles a race through here before gHaveTriedToLoadCommonData is set. */

#if MAP_IMPLEMENTATION==MAP_STDIO
    umtx_unlock(&extendICUDataMutex);
#endif
    return didUpdate;               /* Return true if ICUData pointer was updated.   */
                                    /*   (Could potentially have been done by another thread racing */
                                    /*   us through here, but that's fine, we still return true    */
                                    /*   so that current thread will also examine extended data.   */
}

/*----------------------------------------------------------------------*
 *                                                                      *
 *   udata_setCommonData                                                *
 *                                                                      *
 *----------------------------------------------------------------------*/
U_CAPI void U_EXPORT2
udata_setCommonData(const void *data, UErrorCode *pErrorCode) {
    UDataMemory dataMemory;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return;
    }

    if(data==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    /* set the data pointer and test for validity */
    UDataMemory_init(&dataMemory);
    UDataMemory_setData(&dataMemory, data);
    udata_checkCommonData(&dataMemory, pErrorCode);
    if (U_FAILURE(*pErrorCode)) {return;}

    /* we have good data */
    /* Set it up as the ICU Common Data.  */
    setCommonICUData(&dataMemory, TRUE, pErrorCode);
}

/*---------------------------------------------------------------------------
 *
 *  udata_setAppData
 *
 *---------------------------------------------------------------------------- */
U_CAPI void U_EXPORT2
udata_setAppData(const char *path, const void *data, UErrorCode *err)
{
    UDataMemory     udm;

    if(err==NULL || U_FAILURE(*err)) {
        return;
    }
    if(data==NULL) {
        *err=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    UDataMemory_init(&udm);
    UDataMemory_setData(&udm, data);
    udata_checkCommonData(&udm, err);
    udata_cacheDataItem(path, &udm, err);
}

/*----------------------------------------------------------------------------*
 *                                                                            *
 *  checkDataItem     Given a freshly located/loaded data item, either        *
 *                    an entry in a common file or a separately loaded file,  *
 *                    sanity check its header, and see if the data is         *
 *                    acceptable to the app.                                  *
 *                    If the data is good, create and return a UDataMemory    *
 *                    object that can be returned to the application.         *
 *                    Return NULL on any sort of failure.                     *
 *                                                                            *
 *----------------------------------------------------------------------------*/
static UDataMemory *
checkDataItem
(
 const DataHeader         *pHeader,         /* The data item to be checked.                */
 UDataMemoryIsAcceptable  *isAcceptable,    /* App's call-back function                    */
 void                     *context,         /*   pass-thru param for above.                */
 const char               *type,            /*   pass-thru param for above.                */
 const char               *name,            /*   pass-thru param for above.                */
 UErrorCode               *nonFatalErr,     /* Error code if this data was not acceptable  */
                                            /*   but openChoice should continue with       */
                                            /*   trying to get data from fallback path.    */
 UErrorCode               *fatalErr         /* Bad error, caller should return immediately */
 )
{
    UDataMemory  *rDataMem = NULL;          /* the new UDataMemory, to be returned.        */

    if (U_FAILURE(*fatalErr)) {
        return NULL;
    }

    if(pHeader->dataHeader.magic1==0xda &&
        pHeader->dataHeader.magic2==0x27 &&
        (isAcceptable==NULL || isAcceptable(context, type, name, &pHeader->info))
    ) {
        rDataMem=UDataMemory_createNewInstance(fatalErr);
        if (U_FAILURE(*fatalErr)) {
            return NULL;
        }
        rDataMem->pHeader = pHeader;
    } else {
        /* the data is not acceptable, look further */
        /* If we eventually find something good, this errorcode will be */
        /*    cleared out.                                              */
        *nonFatalErr=U_INVALID_FORMAT_ERROR;
    }
    return rDataMem;
}

/**
 * @return 0 if not loaded, 1 if loaded or err
 */
static UDataMemory *doLoadFromIndividualFiles(const char *pkgName,
        const char *dataPath, const char *tocEntryPathSuffix,
            /* following arguments are the same as doOpenChoice itself */
            const char *path, const char *type, const char *name,
             UDataMemoryIsAcceptable *isAcceptable, void *context,
             UErrorCode *subErrorCode,
             UErrorCode *pErrorCode)
{
    const char         *pathBuffer;
    UDataMemory         dataMemory;
    UDataMemory *pEntryData;

    /* look in ind. files: package\nam.typ  ========================= */
    /* init path iterator for individual files */
    UDataPathIterator iter(dataPath, pkgName, path, tocEntryPathSuffix, FALSE, pErrorCode);

    while ((pathBuffer = iter.next(pErrorCode)) != NULL)
    {
#ifdef UDATA_DEBUG
        fprintf(stderr, "UDATA: trying individual file %s\n", pathBuffer);
#endif
        if (uprv_mapFile(&dataMemory, pathBuffer, pErrorCode))
        {
            pEntryData = checkDataItem(dataMemory.pHeader, isAcceptable, context, type, name, subErrorCode, pErrorCode);
            if (pEntryData != NULL) {
                /* Data is good.
                *  Hand off ownership of the backing memory to the user's UDataMemory.
                *  and return it.   */
                pEntryData->mapAddr = dataMemory.mapAddr;
                pEntryData->map     = dataMemory.map;

#ifdef UDATA_DEBUG
                fprintf(stderr, "** Mapped file: %s\n", pathBuffer);
#endif
                return pEntryData;
            }

            /* the data is not acceptable, or some error occurred.  Either way, unmap the memory */
            udata_close(&dataMemory);

            /* If we had a nasty error, bail out completely.  */
            if (U_FAILURE(*pErrorCode)) {
                return NULL;
            }

            /* Otherwise remember that we found data but didn't like it for some reason  */
            *subErrorCode=U_INVALID_FORMAT_ERROR;
        }
#ifdef UDATA_DEBUG
        fprintf(stderr, "%s\n", UDataMemory_isLoaded(&dataMemory)?"LOADED":"not loaded");
#endif
    }
    return NULL;
}

/**
 * @return 0 if not loaded, 1 if loaded or err
 */
static UDataMemory *doLoadFromCommonData(UBool isICUData, const char * /*pkgName*/,
        const char * /*dataPath*/, const char * /*tocEntryPathSuffix*/, const char *tocEntryName,
            /* following arguments are the same as doOpenChoice itself */
            const char *path, const char *type, const char *name,
             UDataMemoryIsAcceptable *isAcceptable, void *context,
             UErrorCode *subErrorCode,
             UErrorCode *pErrorCode)
{
    UDataMemory        *pEntryData;
    const DataHeader   *pHeader;
    UDataMemory        *pCommonData;
    int32_t            commonDataIndex;
    UBool              checkedExtendedICUData = FALSE;
    /* try to get common data.  The loop is for platforms such as the 390 that do
     *  not initially load the full set of ICU data.  If the lookup of an ICU data item
     *  fails, the full (but slower to load) set is loaded, the and the loop repeats,
     *  trying the lookup again.  Once the full set of ICU data is loaded, the loop wont
     *  repeat because the full set will be checked the first time through.
     *
     *  The loop also handles the fallback to a .dat file if the application linked
     *   to the stub data library rather than a real library.
     */
    for (commonDataIndex = isICUData ? 0 : -1;;) {
        pCommonData=openCommonData(path, commonDataIndex, subErrorCode); /** search for pkg **/

        if(U_SUCCESS(*subErrorCode) && pCommonData!=NULL) {
            int32_t length;

            /* look up the data piece in the common data */
            pHeader=pCommonData->vFuncs->Lookup(pCommonData, tocEntryName, &length, subErrorCode);
#ifdef UDATA_DEBUG
            fprintf(stderr, "%s: pHeader=%p - %s\n", tocEntryName, (void*) pHeader, u_errorName(*subErrorCode));
#endif

            if(pHeader!=NULL) {
                pEntryData = checkDataItem(pHeader, isAcceptable, context, type, name, subErrorCode, pErrorCode);
#ifdef UDATA_DEBUG
                fprintf(stderr, "pEntryData=%p\n", (void*) pEntryData);
#endif
                if (U_FAILURE(*pErrorCode)) {
                    return NULL;
                }
                if (pEntryData != NULL) {
                    pEntryData->length = length;
                    return pEntryData;
                }
            }
        }
        // If we failed due to being out-of-memory, then stop early and report the error.
        if (*subErrorCode == U_MEMORY_ALLOCATION_ERROR) {
            *pErrorCode = *subErrorCode;
            return NULL;
        }
        /* Data wasn't found.  If we were looking for an ICUData item and there is
         * more data available, load it and try again,
         * otherwise break out of this loop. */
        if (!isICUData) {
            return NULL;
        } else if (pCommonData != NULL) {
            ++commonDataIndex;  /* try the next data package */
        } else if ((!checkedExtendedICUData) && extendICUData(subErrorCode)) {
            checkedExtendedICUData = TRUE;
            /* try this data package slot again: it changed from NULL to non-NULL */
        } else {
            return NULL;
        }
    }
}

/*
 * Identify the Time Zone resources that are subject to special override data loading.
 */
static UBool isTimeZoneFile(const char *name, const char *type) {
    return ((uprv_strcmp(type, "res") == 0) &&
            (uprv_strcmp(name, "zoneinfo64") == 0 ||
             uprv_strcmp(name, "timezoneTypes") == 0 ||
             uprv_strcmp(name, "windowsZones") == 0 ||
             uprv_strcmp(name, "metaZones") == 0));
}

/*
 *  A note on the ownership of Mapped Memory
 *
 *  For common format files, ownership resides with the UDataMemory object
 *    that lives in the cache of opened common data.  These UDataMemorys are private
 *    to the udata implementation, and are never seen directly by users.
 *
 *    The UDataMemory objects returned to users will have the address of some desired
 *    data within the mapped region, but they wont have the mapping info itself, and thus
 *    won't cause anything to be removed from memory when they are closed.
 *
 *  For individual data files, the UDataMemory returned to the user holds the
 *  information necessary to unmap the data on close.  If the user independently
 *  opens the same data file twice, two completely independent mappings will be made.
 *  (There is no cache of opened data items from individual files, only a cache of
 *   opened Common Data files, that is, files containing a collection of data items.)
 *
 *  For common data passed in from the user via udata_setAppData() or
 *  udata_setCommonData(), ownership remains with the user.
 *
 *  UDataMemory objects themselves, as opposed to the memory they describe,
 *  can be anywhere - heap, stack/local or global.
 *  They have a flag to indicate when they're heap allocated and thus
 *  must be deleted when closed.
 */


/*----------------------------------------------------------------------------*
 *                                                                            *
 * main data loading functions                                                *
 *                                                                            *
 *----------------------------------------------------------------------------*/
static UDataMemory *
doOpenChoice(const char *path, const char *type, const char *name,
             UDataMemoryIsAcceptable *isAcceptable, void *context,
             UErrorCode *pErrorCode)
{
    UDataMemory         *retVal = NULL;

    const char         *dataPath;

    int32_t             tocEntrySuffixIndex;
    const char         *tocEntryPathSuffix;
    UErrorCode          subErrorCode=U_ZERO_ERROR;
    const char         *treeChar;

    UBool               isICUData = FALSE;


    FileTracer::traceOpen(path, type, name);


    /* Is this path ICU data? */
    if(path == NULL ||
       !strcmp(path, U_ICUDATA_ALIAS) ||  /* "ICUDATA" */
       !uprv_strncmp(path, U_ICUDATA_NAME U_TREE_SEPARATOR_STRING, /* "icudt26e-" */
                     uprv_strlen(U_ICUDATA_NAME U_TREE_SEPARATOR_STRING)) ||
       !uprv_strncmp(path, U_ICUDATA_ALIAS U_TREE_SEPARATOR_STRING, /* "ICUDATA-" */
                     uprv_strlen(U_ICUDATA_ALIAS U_TREE_SEPARATOR_STRING))) {
      isICUData = TRUE;
    }

#if (U_FILE_SEP_CHAR != U_FILE_ALT_SEP_CHAR)  /* Windows:  try "foo\bar" and "foo/bar" */
    /* remap from alternate path char to the main one */
    CharString altSepPath;
    if(path) {
        if(uprv_strchr(path,U_FILE_ALT_SEP_CHAR) != NULL) {
            altSepPath.append(path, *pErrorCode);
            char *p;
            while ((p = uprv_strchr(altSepPath.data(), U_FILE_ALT_SEP_CHAR)) != NULL) {
                *p = U_FILE_SEP_CHAR;
            }
#if defined (UDATA_DEBUG)
            fprintf(stderr, "Changed path from [%s] to [%s]\n", path, altSepPath.s);
#endif
            path = altSepPath.data();
        }
    }
#endif

    CharString tocEntryName; /* entry name in tree format. ex:  'icudt28b/coll/ar.res' */
    CharString tocEntryPath; /* entry name in path format. ex:  'icudt28b\\coll\\ar.res' */

    CharString pkgName;
    CharString treeName;

    /* ======= Set up strings */
    if(path==NULL) {
        pkgName.append(U_ICUDATA_NAME, *pErrorCode);
    } else {
        const char *pkg;
        const char *first;
        pkg = uprv_strrchr(path, U_FILE_SEP_CHAR);
        first = uprv_strchr(path, U_FILE_SEP_CHAR);
        if(uprv_pathIsAbsolute(path) || (pkg != first)) { /* more than one slash in the path- not a tree name */
            /* see if this is an /absolute/path/to/package  path */
            if(pkg) {
                pkgName.append(pkg+1, *pErrorCode);
            } else {
                pkgName.append(path, *pErrorCode);
            }
        } else {
            treeChar = uprv_strchr(path, U_TREE_SEPARATOR);
            if(treeChar) {
                treeName.append(treeChar+1, *pErrorCode); /* following '-' */
                if(isICUData) {
                    pkgName.append(U_ICUDATA_NAME, *pErrorCode);
                } else {
                    pkgName.append(path, (int32_t)(treeChar-path), *pErrorCode);
                    if (first == NULL) {
                        /*
                        This user data has no path, but there is a tree name.
                        Look up the correct path from the data cache later.
                        */
                        path = pkgName.data();
                    }
                }
            } else {
                if(isICUData) {
                    pkgName.append(U_ICUDATA_NAME, *pErrorCode);
                } else {
                    pkgName.append(path, *pErrorCode);
                }
            }
        }
    }

#ifdef UDATA_DEBUG
    fprintf(stderr, " P=%s T=%s\n", pkgName.data(), treeName.data());
#endif

    /* setting up the entry name and file name
     * Make up a full name by appending the type to the supplied
     *  name, assuming that a type was supplied.
     */

    /* prepend the package */
    tocEntryName.append(pkgName, *pErrorCode);
    tocEntryPath.append(pkgName, *pErrorCode);
    tocEntrySuffixIndex = tocEntryName.length();

    if(!treeName.isEmpty()) {
        tocEntryName.append(U_TREE_ENTRY_SEP_CHAR, *pErrorCode).append(treeName, *pErrorCode);
        tocEntryPath.append(U_FILE_SEP_CHAR, *pErrorCode).append(treeName, *pErrorCode);
    }

    tocEntryName.append(U_TREE_ENTRY_SEP_CHAR, *pErrorCode).append(name, *pErrorCode);
    tocEntryPath.append(U_FILE_SEP_CHAR, *pErrorCode).append(name, *pErrorCode);
    if(type!=NULL && *type!=0) {
        tocEntryName.append(".", *pErrorCode).append(type, *pErrorCode);
        tocEntryPath.append(".", *pErrorCode).append(type, *pErrorCode);
    }
    // The +1 is for the U_FILE_SEP_CHAR that is always appended above.
    tocEntryPathSuffix = tocEntryPath.data() + tocEntrySuffixIndex + 1; /* suffix starts here */

#ifdef UDATA_DEBUG
    fprintf(stderr, " tocEntryName = %s\n", tocEntryName.data());
    fprintf(stderr, " tocEntryPath = %s\n", tocEntryName.data());
#endif

#if !defined(ICU_DATA_DIR_WINDOWS)
    if(path == NULL) {
        path = COMMON_DATA_NAME; /* "icudt26e" */
    }
#else
    // When using the Windows system data, we expects only a single data file.
    path = COMMON_DATA_NAME; /* "icudt26e" */
#endif

    /************************ Begin loop looking for ind. files ***************/
#ifdef UDATA_DEBUG
    fprintf(stderr, "IND: inBasename = %s, pkg=%s\n", "(n/a)", packageNameFromPath(path));
#endif

    /* End of dealing with a null basename */
    dataPath = u_getDataDirectory();

    /****    Time zone individual files override  */
    if (isICUData && isTimeZoneFile(name, type)) {
        const char *tzFilesDir = u_getTimeZoneFilesDirectory(pErrorCode);
        if (tzFilesDir[0] != 0) {
#ifdef UDATA_DEBUG
            fprintf(stderr, "Trying Time Zone Files directory = %s\n", tzFilesDir);
#endif
            retVal = doLoadFromIndividualFiles(/* pkgName.data() */ "", tzFilesDir, tocEntryPathSuffix,
                            /* path */ "", type, name, isAcceptable, context, &subErrorCode, pErrorCode);
            if((retVal != NULL) || U_FAILURE(*pErrorCode)) {
                return retVal;
            }
        }
    }

    /****    COMMON PACKAGE  - only if packages are first. */
    if(gDataFileAccess == UDATA_PACKAGES_FIRST) {
#ifdef UDATA_DEBUG
        fprintf(stderr, "Trying packages (UDATA_PACKAGES_FIRST)\n");
#endif
        /* #2 */
        retVal = doLoadFromCommonData(isICUData,
                            pkgName.data(), dataPath, tocEntryPathSuffix, tocEntryName.data(),
                            path, type, name, isAcceptable, context, &subErrorCode, pErrorCode);
        if((retVal != NULL) || U_FAILURE(*pErrorCode)) {
            return retVal;
        }
    }

    /****    INDIVIDUAL FILES  */
    if((gDataFileAccess==UDATA_PACKAGES_FIRST) ||
       (gDataFileAccess==UDATA_FILES_FIRST)) {
#ifdef UDATA_DEBUG
        fprintf(stderr, "Trying individual files\n");
#endif
        /* Check to make sure that there is a dataPath to iterate over */
        if ((dataPath && *dataPath) || !isICUData) {
            retVal = doLoadFromIndividualFiles(pkgName.data(), dataPath, tocEntryPathSuffix,
                            path, type, name, isAcceptable, context, &subErrorCode, pErrorCode);
            if((retVal != NULL) || U_FAILURE(*pErrorCode)) {
                return retVal;
            }
        }
    }

    /****    COMMON PACKAGE  */
    if((gDataFileAccess==UDATA_ONLY_PACKAGES) ||
       (gDataFileAccess==UDATA_FILES_FIRST)) {
#ifdef UDATA_DEBUG
        fprintf(stderr, "Trying packages (UDATA_ONLY_PACKAGES || UDATA_FILES_FIRST)\n");
#endif
        retVal = doLoadFromCommonData(isICUData,
                            pkgName.data(), dataPath, tocEntryPathSuffix, tocEntryName.data(),
                            path, type, name, isAcceptable, context, &subErrorCode, pErrorCode);
        if((retVal != NULL) || U_FAILURE(*pErrorCode)) {
            return retVal;
        }
    }

    /* Load from DLL.  If we haven't attempted package load, we also haven't had any chance to
        try a DLL (static or setCommonData/etc)  load.
         If we ever have a "UDATA_ONLY_FILES", add it to the or list here.  */
    if(gDataFileAccess==UDATA_NO_FILES) {
#ifdef UDATA_DEBUG
        fprintf(stderr, "Trying common data (UDATA_NO_FILES)\n");
#endif
        retVal = doLoadFromCommonData(isICUData,
                            pkgName.data(), "", tocEntryPathSuffix, tocEntryName.data(),
                            path, type, name, isAcceptable, context, &subErrorCode, pErrorCode);
        if((retVal != NULL) || U_FAILURE(*pErrorCode)) {
            return retVal;
        }
    }

    /* data not found */
    if(U_SUCCESS(*pErrorCode)) {
        if(U_SUCCESS(subErrorCode)) {
            /* file not found */
            *pErrorCode=U_FILE_ACCESS_ERROR;
        } else {
            /* entry point not found or rejected */
            *pErrorCode=subErrorCode;
        }
    }
    return retVal;
}



/* API ---------------------------------------------------------------------- */

U_CAPI UDataMemory * U_EXPORT2
udata_open(const char *path, const char *type, const char *name,
           UErrorCode *pErrorCode) {
#ifdef UDATA_DEBUG
  fprintf(stderr, "udata_open(): Opening: %s : %s . %s\n", (path?path:"NULL"), name, type);
    fflush(stderr);
#endif

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return NULL;
    } else if(name==NULL || *name==0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    } else {
        return doOpenChoice(path, type, name, NULL, NULL, pErrorCode);
    }
}



U_CAPI UDataMemory * U_EXPORT2
udata_openChoice(const char *path, const char *type, const char *name,
                 UDataMemoryIsAcceptable *isAcceptable, void *context,
                 UErrorCode *pErrorCode) {
#ifdef UDATA_DEBUG
  fprintf(stderr, "udata_openChoice(): Opening: %s : %s . %s\n", (path?path:"NULL"), name, type);
#endif

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return NULL;
    } else if(name==NULL || *name==0 || isAcceptable==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    } else {
        return doOpenChoice(path, type, name, isAcceptable, context, pErrorCode);
    }
}



U_CAPI void U_EXPORT2
udata_getInfo(UDataMemory *pData, UDataInfo *pInfo) {
    if(pInfo!=NULL) {
        if(pData!=NULL && pData->pHeader!=NULL) {
            const UDataInfo *info=&pData->pHeader->info;
            uint16_t dataInfoSize=udata_getInfoSize(info);
            if(pInfo->size>dataInfoSize) {
                pInfo->size=dataInfoSize;
            }
            uprv_memcpy((uint16_t *)pInfo+1, (const uint16_t *)info+1, pInfo->size-2);
            if(info->isBigEndian!=U_IS_BIG_ENDIAN) {
                /* opposite endianness */
                uint16_t x=info->reservedWord;
                pInfo->reservedWord=(uint16_t)((x<<8)|(x>>8));
            }
        } else {
            pInfo->size=0;
        }
    }
}


U_CAPI void U_EXPORT2 udata_setFileAccess(UDataFileAccess access, UErrorCode * /*status*/)
{
    // Note: this function is documented as not thread safe.
    gDataFileAccess = access;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************/


/*----------------------------------------------------------------------------------
 *
 *  UDataMemory     A class-like struct that serves as a handle to a piece of memory
 *                  that contains some ICU data (resource, converters, whatever.)
 *
 *                  When an application opens ICU data (with udata_open, for example,
 *                  a UDataMemory * is returned.
 *
 *----------------------------------------------------------------------------------*/







U_CFUNC void UDataMemory_init(UDataMemory *This) {
    uprv_memset(This, 0, sizeof(UDataMemory));
    This->length=-1;
}


U_CFUNC void UDatamemory_assign(UDataMemory *dest, UDataMemory *source) {
    /* UDataMemory Assignment.  Destination UDataMemory must be initialized first.  */
    UBool mallocedFlag = dest->heapAllocated;
    uprv_memcpy(dest, source, sizeof(UDataMemory));
    dest->heapAllocated = mallocedFlag;
}

U_CFUNC UDataMemory *UDataMemory_createNewInstance(UErrorCode *pErr) {
    UDataMemory *This;

    if (U_FAILURE(*pErr)) {
        return NULL;
    }
    This = (UDataMemory *)uprv_malloc(sizeof(UDataMemory));
    if (This == NULL) {
        *pErr = U_MEMORY_ALLOCATION_ERROR; }
    else {
        UDataMemory_init(This);
        This->heapAllocated = TRUE;
    }
    return This;
}


U_CFUNC const DataHeader *
UDataMemory_normalizeDataPointer(const void *p) {
    /* allow the data to be optionally prepended with an alignment-forcing double value */
    const DataHeader *pdh = (const DataHeader *)p;
    if(pdh==NULL || (pdh->dataHeader.magic1==0xda && pdh->dataHeader.magic2==0x27)) {
        return pdh;
    } else {
#if U_PLATFORM == U_PF_OS400
        /*
        TODO: Fix this once the compiler implements this feature. Keep in sync with genccode.c

        This is here because this platform can't currently put
        const data into the read-only pages of an object or
        shared library (service program). Only strings are allowed in read-only
        pages, so we use char * strings to store the data.

        In order to prevent the beginning of the data from ever matching the
        magic numbers we must skip the initial double.
        [grhoten 4/24/2003]
        */
        return (const DataHeader *)*((const void **)p+1);
#else
        return (const DataHeader *)((const double *)p+1);
#endif
    }
}


U_CFUNC void UDataMemory_setData (UDataMemory *This, const void *dataAddr) {
    This->pHeader = UDataMemory_normalizeDataPointer(dataAddr);
}


U_CAPI void U_EXPORT2
udata_close(UDataMemory *pData) {
    if(pData!=NULL) {
        uprv_unmapFile(pData);
        if(pData->heapAllocated ) {
            uprv_free(pData);
        } else {
            UDataMemory_init(pData);
        }
    }
}

U_CAPI const void * U_EXPORT2
udata_getMemory(UDataMemory *pData) {
    if(pData!=NULL && pData->pHeader!=NULL) {
        return (char *)(pData->pHeader)+udata_getHeaderSize(pData->pHeader);
    } else {
        return NULL;
    }
}

/**
 * Get the length of the data item if possible.
 * The length may be up to 15 bytes larger than the actual data.
 *
 * TODO Consider making this function public.
 * It would have to return the actual length in more cases.
 * For example, the length of the last item in a .dat package could be
 * computed from the size of the whole .dat package minus the offset of the
 * last item.
 * The size of a file that was directly memory-mapped could be determined
 * using some system API.
 *
 * In order to get perfect values for all data items, we may have to add a
 * length field to UDataInfo, but that complicates data generation
 * and may be overkill.
 *
 * @param pData The data item.
 * @return the length of the data item, or -1 if not known
 * @internal Currently used only in cintltst/udatatst.c
 */
U_CAPI int32_t U_EXPORT2
udata_getLength(const UDataMemory *pData) {
    if(pData!=NULL && pData->pHeader!=NULL && pData->length>=0) {
        /*
         * subtract the header size,
         * return only the size of the actual data starting at udata_getMemory()
         */
        return pData->length-udata_getHeaderSize(pData->pHeader);
    } else {
        return -1;
    }
}

/**
 * Get the memory including the data header.
 * Used in cintltst/udatatst.c
 * @internal
 */
U_CAPI const void * U_EXPORT2
udata_getRawMemory(const UDataMemory *pData) {
    if(pData!=NULL && pData->pHeader!=NULL) {
        return pData->pHeader;
    } else {
        return NULL;
    }
}

U_CFUNC UBool UDataMemory_isLoaded(const UDataMemory *This) {
    return This->pHeader != NULL;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2003-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  udataswp.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2003jun05
*   created by: Markus W. Scherer
*
*   Definitions for ICU data transformations for different platforms,
*   changing between big- and little-endian data and/or between
*   charset families (ASCII<->EBCDIC).
*/

#include <stdarg.h>

 /* UDataInfo */
 /* DataHeader */



/* swapping primitives ------------------------------------------------------ */

static int32_t U_CALLCONV
uprv_swapArray16(const UDataSwapper *ds,
                 const void *inData, int32_t length, void *outData,
                 UErrorCode *pErrorCode) {
    const uint16_t *p;
    uint16_t *q;
    int32_t count;
    uint16_t x;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 || (length&1)!=0 || outData==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* setup and swapping */
    p=(const uint16_t *)inData;
    q=(uint16_t *)outData;
    count=length/2;
    while(count>0) {
        x=*p++;
        *q++=(uint16_t)((x<<8)|(x>>8));
        --count;
    }

    return length;
}

static int32_t U_CALLCONV
uprv_copyArray16(const UDataSwapper *ds,
                 const void *inData, int32_t length, void *outData,
                 UErrorCode *pErrorCode) {
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 || (length&1)!=0 || outData==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if(length>0 && inData!=outData) {
        uprv_memcpy(outData, inData, length);
    }
    return length;
}

static int32_t U_CALLCONV
uprv_swapArray32(const UDataSwapper *ds,
                 const void *inData, int32_t length, void *outData,
                 UErrorCode *pErrorCode) {
    const uint32_t *p;
    uint32_t *q;
    int32_t count;
    uint32_t x;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 || (length&3)!=0 || outData==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* setup and swapping */
    p=(const uint32_t *)inData;
    q=(uint32_t *)outData;
    count=length/4;
    while(count>0) {
        x=*p++;
        *q++=(uint32_t)((x<<24)|((x<<8)&0xff0000)|((x>>8)&0xff00)|(x>>24));
        --count;
    }

    return length;
}

static int32_t U_CALLCONV
uprv_copyArray32(const UDataSwapper *ds,
                 const void *inData, int32_t length, void *outData,
                 UErrorCode *pErrorCode) {
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 || (length&3)!=0 || outData==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if(length>0 && inData!=outData) {
        uprv_memcpy(outData, inData, length);
    }
    return length;
}

static int32_t U_CALLCONV
uprv_swapArray64(const UDataSwapper *ds,
                 const void *inData, int32_t length, void *outData,
                 UErrorCode *pErrorCode) {
    const uint64_t *p;
    uint64_t *q;
    int32_t count;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 || (length&7)!=0 || outData==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* setup and swapping */
    p=(const uint64_t *)inData;
    q=(uint64_t *)outData;
    count=length/8;
    while(count>0) {
        uint64_t x=*p++;
        x=(x<<56)|((x&0xff00)<<40)|((x&0xff0000)<<24)|((x&0xff000000)<<8)|
            ((x>>8)&0xff000000)|((x>>24)&0xff0000)|((x>>40)&0xff00)|(x>>56);
        *q++=x;
        --count;
    }

    return length;
}

static int32_t U_CALLCONV
uprv_copyArray64(const UDataSwapper *ds,
                 const void *inData, int32_t length, void *outData,
                 UErrorCode *pErrorCode) {
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 || (length&7)!=0 || outData==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if(length>0 && inData!=outData) {
        uprv_memcpy(outData, inData, length);
    }
    return length;
}

static uint16_t U_CALLCONV
uprv_readSwapUInt16(uint16_t x) {
    return (uint16_t)((x<<8)|(x>>8));
}

static uint16_t U_CALLCONV
uprv_readDirectUInt16(uint16_t x) {
    return x;
}

static uint32_t U_CALLCONV
uprv_readSwapUInt32(uint32_t x) {
    return (uint32_t)((x<<24)|((x<<8)&0xff0000)|((x>>8)&0xff00)|(x>>24));
}

static uint32_t U_CALLCONV
uprv_readDirectUInt32(uint32_t x) {
    return x;
}

static void U_CALLCONV
uprv_writeSwapUInt16(uint16_t *p, uint16_t x) {
    *p=(uint16_t)((x<<8)|(x>>8));
}

static void U_CALLCONV
uprv_writeDirectUInt16(uint16_t *p, uint16_t x) {
    *p=x;
}

static void U_CALLCONV
uprv_writeSwapUInt32(uint32_t *p, uint32_t x) {
    *p=(uint32_t)((x<<24)|((x<<8)&0xff0000)|((x>>8)&0xff00)|(x>>24));
}

static void U_CALLCONV
uprv_writeDirectUInt32(uint32_t *p, uint32_t x) {
    *p=x;
}

U_CAPI int16_t U_EXPORT2
udata_readInt16(const UDataSwapper *ds, int16_t x) {
    return (int16_t)ds->readUInt16((uint16_t)x);
}

U_CAPI int32_t U_EXPORT2
udata_readInt32(const UDataSwapper *ds, int32_t x) {
    return (int32_t)ds->readUInt32((uint32_t)x);
}

/**
 * Swap a block of invariant, NUL-terminated strings, but not padding
 * bytes after the last string.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
udata_swapInvStringBlock(const UDataSwapper *ds,
                         const void *inData, int32_t length, void *outData,
                         UErrorCode *pErrorCode) {
    const char *inChars;
    int32_t stringsLength;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 || (length>0 && outData==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* reduce the strings length to not include bytes after the last NUL */
    inChars=(const char *)inData;
    stringsLength=length;
    while(stringsLength>0 && inChars[stringsLength-1]!=0) {
        --stringsLength;
    }

    /* swap up to the last NUL */
    ds->swapInvChars(ds, inData, stringsLength, outData, pErrorCode);

    /* copy the bytes after the last NUL */
    if(inData!=outData && length>stringsLength) {
        uprv_memcpy((char *)outData+stringsLength, inChars+stringsLength, length-stringsLength);
    }

    /* return the length including padding bytes */
    if(U_SUCCESS(*pErrorCode)) {
        return length;
    } else {
        return 0;
    }
}

U_CAPI void U_EXPORT2
udata_printError(const UDataSwapper *ds,
                 const char *fmt,
                 ...) {
    va_list args;

    if(ds->printError!=NULL) {
        va_start(args, fmt);
        ds->printError(ds->printErrorContext, fmt, args);
        va_end(args);
    }
}

/* swap a data header ------------------------------------------------------- */

U_CAPI int32_t U_EXPORT2
udata_swapDataHeader(const UDataSwapper *ds,
                     const void *inData, int32_t length, void *outData,
                     UErrorCode *pErrorCode) {
    const DataHeader *pHeader;
    uint16_t headerSize, infoSize;

    /* argument checking */
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<-1 || (length>0 && outData==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* check minimum length and magic bytes */
    pHeader=(const DataHeader *)inData;
    if( (length>=0 && length<(int32_t)sizeof(DataHeader)) ||
        pHeader->dataHeader.magic1!=0xda ||
        pHeader->dataHeader.magic2!=0x27 ||
        pHeader->info.sizeofUChar!=2
    ) {
        udata_printError(ds, "udata_swapDataHeader(): initial bytes do not look like ICU data\n");
        *pErrorCode=U_UNSUPPORTED_ERROR;
        return 0;
    }

    headerSize=ds->readUInt16(pHeader->dataHeader.headerSize);
    infoSize=ds->readUInt16(pHeader->info.size);

    if( headerSize<sizeof(DataHeader) ||
        infoSize<sizeof(UDataInfo) ||
        headerSize<(sizeof(pHeader->dataHeader)+infoSize) ||
        (length>=0 && length<headerSize)
    ) {
        udata_printError(ds, "udata_swapDataHeader(): header size mismatch - headerSize %d infoSize %d length %d\n",
                         headerSize, infoSize, length);
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    if(length>0) {
        DataHeader *outHeader;
        const char *s;
        int32_t maxLength;

        /* Most of the fields are just bytes and need no swapping. */
        if(inData!=outData) {
            uprv_memcpy(outData, inData, headerSize);
        }
        outHeader=(DataHeader *)outData;

        outHeader->info.isBigEndian = ds->outIsBigEndian;
        outHeader->info.charsetFamily = ds->outCharset;

        /* swap headerSize */
        ds->swapArray16(ds, &pHeader->dataHeader.headerSize, 2, &outHeader->dataHeader.headerSize, pErrorCode);

        /* swap UDataInfo size and reservedWord */
        ds->swapArray16(ds, &pHeader->info.size, 4, &outHeader->info.size, pErrorCode);

        /* swap copyright statement after the UDataInfo */
        infoSize+=sizeof(pHeader->dataHeader);
        s=(const char *)inData+infoSize;
        maxLength=headerSize-infoSize;
        /* get the length of the string */
        for(length=0; length<maxLength && s[length]!=0; ++length) {}
        /* swap the string contents */
        ds->swapInvChars(ds, s, length, (char *)outData+infoSize, pErrorCode);
    }

    return headerSize;
}

/* API functions ------------------------------------------------------------ */

U_CAPI UDataSwapper * U_EXPORT2
udata_openSwapper(UBool inIsBigEndian, uint8_t inCharset,
                  UBool outIsBigEndian, uint8_t outCharset,
                  UErrorCode *pErrorCode) {
    UDataSwapper *swapper;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if(inCharset>U_EBCDIC_FAMILY || outCharset>U_EBCDIC_FAMILY) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    /* allocate the swapper */
    swapper=(UDataSwapper *)uprv_malloc(sizeof(UDataSwapper));
    if(swapper==NULL) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    uprv_memset(swapper, 0, sizeof(UDataSwapper));

    /* set values and functions pointers according to in/out parameters */
    swapper->inIsBigEndian=inIsBigEndian;
    swapper->inCharset=inCharset;
    swapper->outIsBigEndian=outIsBigEndian;
    swapper->outCharset=outCharset;

    swapper->readUInt16= inIsBigEndian==U_IS_BIG_ENDIAN ? uprv_readDirectUInt16 : uprv_readSwapUInt16;
    swapper->readUInt32= inIsBigEndian==U_IS_BIG_ENDIAN ? uprv_readDirectUInt32 : uprv_readSwapUInt32;

    swapper->writeUInt16= outIsBigEndian==U_IS_BIG_ENDIAN ? uprv_writeDirectUInt16 : uprv_writeSwapUInt16;
    swapper->writeUInt32= outIsBigEndian==U_IS_BIG_ENDIAN ? uprv_writeDirectUInt32 : uprv_writeSwapUInt32;

    swapper->compareInvChars= outCharset==U_ASCII_FAMILY ? uprv_compareInvAscii : uprv_compareInvEbcdic;

    if(inIsBigEndian==outIsBigEndian) {
        swapper->swapArray16=uprv_copyArray16;
        swapper->swapArray32=uprv_copyArray32;
        swapper->swapArray64=uprv_copyArray64;
    } else {
        swapper->swapArray16=uprv_swapArray16;
        swapper->swapArray32=uprv_swapArray32;
        swapper->swapArray64=uprv_swapArray64;
    }

    if(inCharset==U_ASCII_FAMILY) {
        swapper->swapInvChars= outCharset==U_ASCII_FAMILY ? uprv_copyAscii : uprv_ebcdicFromAscii;
    } else /* U_EBCDIC_FAMILY */ {
        swapper->swapInvChars= outCharset==U_EBCDIC_FAMILY ? uprv_copyEbcdic : uprv_asciiFromEbcdic;
    }

    return swapper;
}

U_CAPI UDataSwapper * U_EXPORT2
udata_openSwapperForInputData(const void *data, int32_t length,
                              UBool outIsBigEndian, uint8_t outCharset,
                              UErrorCode *pErrorCode) {
    const DataHeader *pHeader;
    uint16_t headerSize, infoSize;
    UBool inIsBigEndian;
    int8_t inCharset;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if( data==NULL ||
        (length>=0 && length<(int32_t)sizeof(DataHeader)) ||
        outCharset>U_EBCDIC_FAMILY
    ) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    pHeader=(const DataHeader *)data;
    if( (length>=0 && length<(int32_t)sizeof(DataHeader)) ||
        pHeader->dataHeader.magic1!=0xda ||
        pHeader->dataHeader.magic2!=0x27 ||
        pHeader->info.sizeofUChar!=2
    ) {
        *pErrorCode=U_UNSUPPORTED_ERROR;
        return 0;
    }

    inIsBigEndian=(UBool)pHeader->info.isBigEndian;
    inCharset=pHeader->info.charsetFamily;

    if(inIsBigEndian==U_IS_BIG_ENDIAN) {
        headerSize=pHeader->dataHeader.headerSize;
        infoSize=pHeader->info.size;
    } else {
        headerSize=uprv_readSwapUInt16(pHeader->dataHeader.headerSize);
        infoSize=uprv_readSwapUInt16(pHeader->info.size);
    }

    if( headerSize<sizeof(DataHeader) ||
        infoSize<sizeof(UDataInfo) ||
        headerSize<(sizeof(pHeader->dataHeader)+infoSize) ||
        (length>=0 && length<headerSize)
    ) {
        *pErrorCode=U_UNSUPPORTED_ERROR;
        return 0;
    }

    return udata_openSwapper(inIsBigEndian, inCharset, outIsBigEndian, outCharset, pErrorCode);
}

U_CAPI void U_EXPORT2
udata_closeSwapper(UDataSwapper *ds) {
    uprv_free(ds);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2002-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uenum.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:2
*
*   created on: 2002jul08
*   created by: Vladimir Weinstein
*/





/* Layout of the baseContext buffer. */
typedef struct {
    int32_t len;  /* number of bytes available starting at 'data' */
    char    data; /* actual data starts here */
} _UEnumBuffer;

/* Extra bytes to allocate in the baseContext buffer. */
static const int32_t PAD = 8;

/* Return a pointer to the baseContext buffer, possibly allocating
   or reallocating it if at least 'capacity' bytes are not available. */
static void* _getBuffer(UEnumeration* en, int32_t capacity) {

    if (en->baseContext != NULL) {
        if (((_UEnumBuffer*) en->baseContext)->len < capacity) {
            capacity += PAD;
            en->baseContext = uprv_realloc(en->baseContext,
                                           sizeof(int32_t) + capacity);
            if (en->baseContext == NULL) {
                return NULL;
            }
            ((_UEnumBuffer*) en->baseContext)->len = capacity;
        }
    } else {
        capacity += PAD;
        en->baseContext = uprv_malloc(sizeof(int32_t) + capacity);
        if (en->baseContext == NULL) {
            return NULL;
        }
        ((_UEnumBuffer*) en->baseContext)->len = capacity;
    }
    
    return (void*) & ((_UEnumBuffer*) en->baseContext)->data;
}

U_CAPI void U_EXPORT2
uenum_close(UEnumeration* en)
{
    if (en) {
        if (en->close != NULL) {
            if (en->baseContext) {
                uprv_free(en->baseContext);
            }
            en->close(en);
        } else { /* this seems dangerous, but we better kill the object */
            uprv_free(en);
        }
    }
}

U_CAPI int32_t U_EXPORT2
uenum_count(UEnumeration* en, UErrorCode* status)
{
    if (!en || U_FAILURE(*status)) {
        return -1;
    }
    if (en->count != NULL) {
        return en->count(en, status);
    } else {
        *status = U_UNSUPPORTED_ERROR;
        return -1;
    }
}

/* Don't call this directly. Only uenum_unext should be calling this. */
U_CAPI const UChar* U_EXPORT2
uenum_unextDefault(UEnumeration* en,
            int32_t* resultLength,
            UErrorCode* status)
{
    UChar *ustr = NULL;
    int32_t len = 0;
    if (en->next != NULL) {
        const char *cstr = en->next(en, &len, status);
        if (cstr != NULL) {
            ustr = (UChar*) _getBuffer(en, (len+1) * sizeof(UChar));
            if (ustr == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
            } else {
                u_charsToUChars(cstr, ustr, len+1);
            }
        }
    } else {
        *status = U_UNSUPPORTED_ERROR;
    }
    if (resultLength) {
        *resultLength = len;
    }
    return ustr;
}

/* Don't call this directly. Only uenum_next should be calling this. */
U_CAPI const char* U_EXPORT2
uenum_nextDefault(UEnumeration* en,
            int32_t* resultLength,
            UErrorCode* status)
{
    if (en->uNext != NULL) {
        char *tempCharVal;
        const UChar *tempUCharVal = en->uNext(en, resultLength, status);
        if (tempUCharVal == NULL) {
            return NULL;
        }
        tempCharVal = (char*)
            _getBuffer(en, (*resultLength+1) * sizeof(char));
        if (!tempCharVal) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        u_UCharsToChars(tempUCharVal, tempCharVal, *resultLength + 1);
        return tempCharVal;
    } else {
        *status = U_UNSUPPORTED_ERROR;
        return NULL;
    }
}

U_CAPI const UChar* U_EXPORT2
uenum_unext(UEnumeration* en,
            int32_t* resultLength,
            UErrorCode* status)
{
    if (!en || U_FAILURE(*status)) {
        return NULL;
    }
    if (en->uNext != NULL) {
        return en->uNext(en, resultLength, status);
    } else {
        *status = U_UNSUPPORTED_ERROR;
        return NULL;
    }
}

U_CAPI const char* U_EXPORT2
uenum_next(UEnumeration* en,
          int32_t* resultLength,
          UErrorCode* status)
{
    if (!en || U_FAILURE(*status)) {
        return NULL;
    }
    if (en->next != NULL) {
        if (resultLength != NULL) {
            return en->next(en, resultLength, status);
        }
        else {
            int32_t dummyLength=0;
            return en->next(en, &dummyLength, status);
        }
    } else {
        *status = U_UNSUPPORTED_ERROR;
        return NULL;
    }
}

U_CAPI void U_EXPORT2
uenum_reset(UEnumeration* en, UErrorCode* status)
{
    if (!en || U_FAILURE(*status)) {
        return;
    }
    if (en->reset != NULL) {
        en->reset(en, status);
    } else {
        *status = U_UNSUPPORTED_ERROR;
    }
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*   Date        Name        Description
*   03/22/00    aliu        Adapted from original C++ ICU Hashtable.
*   07/06/01    aliu        Modified to support int32_t keys on
*                           platforms with sizeof(void*) < 32.
******************************************************************************
*/








/* This hashtable is implemented as a double hash.  All elements are
 * stored in a single array with no secondary storage for collision
 * resolution (no linked list, etc.).  When there is a hash collision
 * (when two unequal keys have the same hashcode) we resolve this by
 * using a secondary hash.  The secondary hash is an increment
 * computed as a hash function (a different one) of the primary
 * hashcode.  This increment is added to the initial hash value to
 * obtain further slots assigned to the same hash code.  For this to
 * work, the length of the array and the increment must be relatively
 * prime.  The easiest way to achieve this is to have the length of
 * the array be prime, and the increment be any value from
 * 1..length-1.
 *
 * Hashcodes are 32-bit integers.  We make sure all hashcodes are
 * non-negative by masking off the top bit.  This has two effects: (1)
 * modulo arithmetic is simplified.  If we allowed negative hashcodes,
 * then when we computed hashcode % length, we could get a negative
 * result, which we would then have to adjust back into range.  It's
 * simpler to just make hashcodes non-negative. (2) It makes it easy
 * to check for empty vs. occupied slots in the table.  We just mark
 * empty or deleted slots with a negative hashcode.
 *
 * The central function is _uhash_find().  This function looks for a
 * slot matching the given key and hashcode.  If one is found, it
 * returns a pointer to that slot.  If the table is full, and no match
 * is found, it returns NULL -- in theory.  This would make the code
 * more complicated, since all callers of _uhash_find() would then
 * have to check for a NULL result.  To keep this from happening, we
 * don't allow the table to fill.  When there is only one
 * empty/deleted slot left, uhash_put() will refuse to increase the
 * count, and fail.  This simplifies the code.  In practice, one will
 * seldom encounter this using default UHashtables.  However, if a
 * hashtable is set to a U_FIXED resize policy, or if memory is
 * exhausted, then the table may fill.
 *
 * High and low water ratios control rehashing.  They establish levels
 * of fullness (from 0 to 1) outside of which the data array is
 * reallocated and repopulated.  Setting the low water ratio to zero
 * means the table will never shrink.  Setting the high water ratio to
 * one means the table will never grow.  The ratios should be
 * coordinated with the ratio between successive elements of the
 * PRIMES table, so that when the primeIndex is incremented or
 * decremented during rehashing, it brings the ratio of count / length
 * back into the desired range (between low and high water ratios).
 */

/********************************************************************
 * PRIVATE Constants, Macros
 ********************************************************************/

/* This is a list of non-consecutive primes chosen such that
 * PRIMES[i+1] ~ 2*PRIMES[i].  (Currently, the ratio ranges from 1.81
 * to 2.18; the inverse ratio ranges from 0.459 to 0.552.)  If this
 * ratio is changed, the low and high water ratios should also be
 * adjusted to suit.
 *
 * These prime numbers were also chosen so that they are the largest
 * prime number while being less than a power of two.
 */
static const int32_t PRIMES[] = {
    7, 13, 31, 61, 127, 251, 509, 1021, 2039, 4093, 8191, 16381, 32749,
    65521, 131071, 262139, 524287, 1048573, 2097143, 4194301, 8388593,
    16777213, 33554393, 67108859, 134217689, 268435399, 536870909,
    1073741789, 2147483647 /*, 4294967291 */
};

#define PRIMES_LENGTH UPRV_LENGTHOF(PRIMES)
#define DEFAULT_PRIME_INDEX 4

/* These ratios are tuned to the PRIMES array such that a resize
 * places the table back into the zone of non-resizing.  That is,
 * after a call to _uhash_rehash(), a subsequent call to
 * _uhash_rehash() should do nothing (should not churn).  This is only
 * a potential problem with U_GROW_AND_SHRINK.
 */
static const float RESIZE_POLICY_RATIO_TABLE[6] = {
    /* low, high water ratio */
    0.0F, 0.5F, /* U_GROW: Grow on demand, do not shrink */
    0.1F, 0.5F, /* U_GROW_AND_SHRINK: Grow and shrink on demand */
    0.0F, 1.0F  /* U_FIXED: Never change size */
};

/*
  Invariants for hashcode values:

  * DELETED < 0
  * EMPTY < 0
  * Real hashes >= 0

  Hashcodes may not start out this way, but internally they are
  adjusted so that they are always positive.  We assume 32-bit
  hashcodes; adjust these constants for other hashcode sizes.
*/
#define HASH_DELETED    ((int32_t) 0x80000000)
#define HASH_EMPTY      ((int32_t) HASH_DELETED + 1)

#define IS_EMPTY_OR_DELETED(x) ((x) < 0)

/* This macro expects a UHashTok.pointer as its keypointer and
   valuepointer parameters */
#define HASH_DELETE_KEY_VALUE(hash, keypointer, valuepointer) UPRV_BLOCK_MACRO_BEGIN { \
    if (hash->keyDeleter != NULL && keypointer != NULL) { \
        (*hash->keyDeleter)(keypointer); \
    } \
    if (hash->valueDeleter != NULL && valuepointer != NULL) { \
        (*hash->valueDeleter)(valuepointer); \
    } \
} UPRV_BLOCK_MACRO_END

/*
 * Constants for hinting whether a key or value is an integer
 * or a pointer.  If a hint bit is zero, then the associated
 * token is assumed to be an integer.
 */
#define HINT_KEY_POINTER   (1)
#define HINT_VALUE_POINTER (2)

/********************************************************************
 * PRIVATE Implementation
 ********************************************************************/

static UHashTok
_uhash_setElement(UHashtable *hash, UHashElement* e,
                  int32_t hashcode,
                  UHashTok key, UHashTok value, int8_t hint) {

    UHashTok oldValue = e->value;
    if (hash->keyDeleter != NULL && e->key.pointer != NULL &&
        e->key.pointer != key.pointer) { /* Avoid double deletion */
        (*hash->keyDeleter)(e->key.pointer);
    }
    if (hash->valueDeleter != NULL) {
        if (oldValue.pointer != NULL &&
            oldValue.pointer != value.pointer) { /* Avoid double deletion */
            (*hash->valueDeleter)(oldValue.pointer);
        }
        oldValue.pointer = NULL;
    }
    /* Compilers should copy the UHashTok union correctly, but even if
     * they do, memory heap tools (e.g. BoundsChecker) can get
     * confused when a pointer is cloaked in a union and then copied.
     * TO ALLEVIATE THIS, we use hints (based on what API the user is
     * calling) to copy pointers when we know the user thinks
     * something is a pointer. */
    if (hint & HINT_KEY_POINTER) {
        e->key.pointer = key.pointer;
    } else {
        e->key = key;
    }
    if (hint & HINT_VALUE_POINTER) {
        e->value.pointer = value.pointer;
    } else {
        e->value = value;
    }
    e->hashcode = hashcode;
    return oldValue;
}

/**
 * Assumes that the given element is not empty or deleted.
 */
static UHashTok
_uhash_internalRemoveElement(UHashtable *hash, UHashElement* e) {
    UHashTok empty;
    U_ASSERT(!IS_EMPTY_OR_DELETED(e->hashcode));
    --hash->count;
    empty.pointer = NULL; empty.integer = 0;
    return _uhash_setElement(hash, e, HASH_DELETED, empty, empty, 0);
}

static void
_uhash_internalSetResizePolicy(UHashtable *hash, enum UHashResizePolicy policy) {
    U_ASSERT(hash != NULL);
    U_ASSERT(((int32_t)policy) >= 0);
    U_ASSERT(((int32_t)policy) < 3);
    hash->lowWaterRatio  = RESIZE_POLICY_RATIO_TABLE[policy * 2];
    hash->highWaterRatio = RESIZE_POLICY_RATIO_TABLE[policy * 2 + 1];
}

/**
 * Allocate internal data array of a size determined by the given
 * prime index.  If the index is out of range it is pinned into range.
 * If the allocation fails the status is set to
 * U_MEMORY_ALLOCATION_ERROR and all array storage is freed.  In
 * either case the previous array pointer is overwritten.
 *
 * Caller must ensure primeIndex is in range 0..PRIME_LENGTH-1.
 */
static void
_uhash_allocate(UHashtable *hash,
                int32_t primeIndex,
                UErrorCode *status) {

    UHashElement *p, *limit;
    UHashTok emptytok;

    if (U_FAILURE(*status)) return;

    U_ASSERT(primeIndex >= 0 && primeIndex < PRIMES_LENGTH);

    hash->primeIndex = static_cast<int8_t>(primeIndex);
    hash->length = PRIMES[primeIndex];

    p = hash->elements = (UHashElement*)
        uprv_malloc(sizeof(UHashElement) * hash->length);

    if (hash->elements == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    emptytok.pointer = NULL; /* Only one of these two is needed */
    emptytok.integer = 0;    /* but we don't know which one. */

    limit = p + hash->length;
    while (p < limit) {
        p->key = emptytok;
        p->value = emptytok;
        p->hashcode = HASH_EMPTY;
        ++p;
    }

    hash->count = 0;
    hash->lowWaterMark = (int32_t)(hash->length * hash->lowWaterRatio);
    hash->highWaterMark = (int32_t)(hash->length * hash->highWaterRatio);
}

static UHashtable*
_uhash_init(UHashtable *result,
              UHashFunction *keyHash,
              UKeyComparator *keyComp,
              UValueComparator *valueComp,
              int32_t primeIndex,
              UErrorCode *status)
{
    if (U_FAILURE(*status)) return NULL;
    U_ASSERT(keyHash != NULL);
    U_ASSERT(keyComp != NULL);

    result->keyHasher       = keyHash;
    result->keyComparator   = keyComp;
    result->valueComparator = valueComp;
    result->keyDeleter      = NULL;
    result->valueDeleter    = NULL;
    result->allocated       = FALSE;
    _uhash_internalSetResizePolicy(result, U_GROW);

    _uhash_allocate(result, primeIndex, status);

    if (U_FAILURE(*status)) {
        return NULL;
    }

    return result;
}

static UHashtable*
_uhash_create(UHashFunction *keyHash,
              UKeyComparator *keyComp,
              UValueComparator *valueComp,
              int32_t primeIndex,
              UErrorCode *status) {
    UHashtable *result;

    if (U_FAILURE(*status)) return NULL;

    result = (UHashtable*) uprv_malloc(sizeof(UHashtable));
    if (result == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    _uhash_init(result, keyHash, keyComp, valueComp, primeIndex, status);
    result->allocated       = TRUE;

    if (U_FAILURE(*status)) {
        uprv_free(result);
        return NULL;
    }

    return result;
}

/**
 * Look for a key in the table, or if no such key exists, the first
 * empty slot matching the given hashcode.  Keys are compared using
 * the keyComparator function.
 *
 * First find the start position, which is the hashcode modulo
 * the length.  Test it to see if it is:
 *
 * a. identical:  First check the hash values for a quick check,
 *    then compare keys for equality using keyComparator.
 * b. deleted
 * c. empty
 *
 * Stop if it is identical or empty, otherwise continue by adding a
 * "jump" value (moduloing by the length again to keep it within
 * range) and retesting.  For efficiency, there need enough empty
 * values so that the searchs stop within a reasonable amount of time.
 * This can be changed by changing the high/low water marks.
 *
 * In theory, this function can return NULL, if it is full (no empty
 * or deleted slots) and if no matching key is found.  In practice, we
 * prevent this elsewhere (in uhash_put) by making sure the last slot
 * in the table is never filled.
 *
 * The size of the table should be prime for this algorithm to work;
 * otherwise we are not guaranteed that the jump value (the secondary
 * hash) is relatively prime to the table length.
 */
static UHashElement*
_uhash_find(const UHashtable *hash, UHashTok key,
            int32_t hashcode) {

    int32_t firstDeleted = -1;  /* assume invalid index */
    int32_t theIndex, startIndex;
    int32_t jump = 0; /* lazy evaluate */
    int32_t tableHash;
    UHashElement *elements = hash->elements;

    hashcode &= 0x7FFFFFFF; /* must be positive */
    startIndex = theIndex = (hashcode ^ 0x4000000) % hash->length;

    do {
        tableHash = elements[theIndex].hashcode;
        if (tableHash == hashcode) {          /* quick check */
            if ((*hash->keyComparator)(key, elements[theIndex].key)) {
                return &(elements[theIndex]);
            }
        } else if (!IS_EMPTY_OR_DELETED(tableHash)) {
            /* We have hit a slot which contains a key-value pair,
             * but for which the hash code does not match.  Keep
             * looking.
             */
        } else if (tableHash == HASH_EMPTY) { /* empty, end o' the line */
            break;
        } else if (firstDeleted < 0) { /* remember first deleted */
            firstDeleted = theIndex;
        }
        if (jump == 0) { /* lazy compute jump */
            /* The jump value must be relatively prime to the table
             * length.  As long as the length is prime, then any value
             * 1..length-1 will be relatively prime to it.
             */
            jump = (hashcode % (hash->length - 1)) + 1;
        }
        theIndex = (theIndex + jump) % hash->length;
    } while (theIndex != startIndex);

    if (firstDeleted >= 0) {
        theIndex = firstDeleted; /* reset if had deleted slot */
    } else if (tableHash != HASH_EMPTY) {
        /* We get to this point if the hashtable is full (no empty or
         * deleted slots), and we've failed to find a match.  THIS
         * WILL NEVER HAPPEN as long as uhash_put() makes sure that
         * count is always < length.
         */
        UPRV_UNREACHABLE;
    }
    return &(elements[theIndex]);
}

/**
 * Attempt to grow or shrink the data arrays in order to make the
 * count fit between the high and low water marks.  hash_put() and
 * hash_remove() call this method when the count exceeds the high or
 * low water marks.  This method may do nothing, if memory allocation
 * fails, or if the count is already in range, or if the length is
 * already at the low or high limit.  In any case, upon return the
 * arrays will be valid.
 */
static void
_uhash_rehash(UHashtable *hash, UErrorCode *status) {

    UHashElement *old = hash->elements;
    int32_t oldLength = hash->length;
    int32_t newPrimeIndex = hash->primeIndex;
    int32_t i;

    if (hash->count > hash->highWaterMark) {
        if (++newPrimeIndex >= PRIMES_LENGTH) {
            return;
        }
    } else if (hash->count < hash->lowWaterMark) {
        if (--newPrimeIndex < 0) {
            return;
        }
    } else {
        return;
    }

    _uhash_allocate(hash, newPrimeIndex, status);

    if (U_FAILURE(*status)) {
        hash->elements = old;
        hash->length = oldLength;
        return;
    }

    for (i = oldLength - 1; i >= 0; --i) {
        if (!IS_EMPTY_OR_DELETED(old[i].hashcode)) {
            UHashElement *e = _uhash_find(hash, old[i].key, old[i].hashcode);
            U_ASSERT(e != NULL);
            U_ASSERT(e->hashcode == HASH_EMPTY);
            e->key = old[i].key;
            e->value = old[i].value;
            e->hashcode = old[i].hashcode;
            ++hash->count;
        }
    }

    uprv_free(old);
}

static UHashTok
_uhash_remove(UHashtable *hash,
              UHashTok key) {
    /* First find the position of the key in the table.  If the object
     * has not been removed already, remove it.  If the user wanted
     * keys deleted, then delete it also.  We have to put a special
     * hashcode in that position that means that something has been
     * deleted, since when we do a find, we have to continue PAST any
     * deleted values.
     */
    UHashTok result;
    UHashElement* e = _uhash_find(hash, key, hash->keyHasher(key));
    U_ASSERT(e != NULL);
    result.pointer = NULL;
    result.integer = 0;
    if (!IS_EMPTY_OR_DELETED(e->hashcode)) {
        result = _uhash_internalRemoveElement(hash, e);
        if (hash->count < hash->lowWaterMark) {
            UErrorCode status = U_ZERO_ERROR;
            _uhash_rehash(hash, &status);
        }
    }
    return result;
}

static UHashTok
_uhash_put(UHashtable *hash,
           UHashTok key,
           UHashTok value,
           int8_t hint,
           UErrorCode *status) {

    /* Put finds the position in the table for the new value.  If the
     * key is already in the table, it is deleted, if there is a
     * non-NULL keyDeleter.  Then the key, the hash and the value are
     * all put at the position in their respective arrays.
     */
    int32_t hashcode;
    UHashElement* e;
    UHashTok emptytok;

    if (U_FAILURE(*status)) {
        goto err;
    }
    U_ASSERT(hash != NULL);
    /* Cannot always check pointer here or iSeries sees NULL every time. */
    if ((hint & HINT_VALUE_POINTER) && value.pointer == NULL) {
        /* Disallow storage of NULL values, since NULL is returned by
         * get() to indicate an absent key.  Storing NULL == removing.
         */
        return _uhash_remove(hash, key);
    }
    if (hash->count > hash->highWaterMark) {
        _uhash_rehash(hash, status);
        if (U_FAILURE(*status)) {
            goto err;
        }
    }

    hashcode = (*hash->keyHasher)(key);
    e = _uhash_find(hash, key, hashcode);
    U_ASSERT(e != NULL);

    if (IS_EMPTY_OR_DELETED(e->hashcode)) {
        /* Important: We must never actually fill the table up.  If we
         * do so, then _uhash_find() will return NULL, and we'll have
         * to check for NULL after every call to _uhash_find().  To
         * avoid this we make sure there is always at least one empty
         * or deleted slot in the table.  This only is a problem if we
         * are out of memory and rehash isn't working.
         */
        ++hash->count;
        if (hash->count == hash->length) {
            /* Don't allow count to reach length */
            --hash->count;
            *status = U_MEMORY_ALLOCATION_ERROR;
            goto err;
        }
    }

    /* We must in all cases handle storage properly.  If there was an
     * old key, then it must be deleted (if the deleter != NULL).
     * Make hashcodes stored in table positive.
     */
    return _uhash_setElement(hash, e, hashcode & 0x7FFFFFFF, key, value, hint);

 err:
    /* If the deleters are non-NULL, this method adopts its key and/or
     * value arguments, and we must be sure to delete the key and/or
     * value in all cases, even upon failure.
     */
    HASH_DELETE_KEY_VALUE(hash, key.pointer, value.pointer);
    emptytok.pointer = NULL; emptytok.integer = 0;
    return emptytok;
}


/********************************************************************
 * PUBLIC API
 ********************************************************************/

U_CAPI UHashtable* U_EXPORT2
uhash_open(UHashFunction *keyHash,
           UKeyComparator *keyComp,
           UValueComparator *valueComp,
           UErrorCode *status) {

    return _uhash_create(keyHash, keyComp, valueComp, DEFAULT_PRIME_INDEX, status);
}

U_CAPI UHashtable* U_EXPORT2
uhash_openSize(UHashFunction *keyHash,
               UKeyComparator *keyComp,
               UValueComparator *valueComp,
               int32_t size,
               UErrorCode *status) {

    /* Find the smallest index i for which PRIMES[i] >= size. */
    int32_t i = 0;
    while (i<(PRIMES_LENGTH-1) && PRIMES[i]<size) {
        ++i;
    }

    return _uhash_create(keyHash, keyComp, valueComp, i, status);
}

U_CAPI UHashtable* U_EXPORT2
uhash_init(UHashtable *fillinResult,
           UHashFunction *keyHash,
           UKeyComparator *keyComp,
           UValueComparator *valueComp,
           UErrorCode *status) {

    return _uhash_init(fillinResult, keyHash, keyComp, valueComp, DEFAULT_PRIME_INDEX, status);
}

U_CAPI UHashtable* U_EXPORT2
uhash_initSize(UHashtable *fillinResult,
               UHashFunction *keyHash,
               UKeyComparator *keyComp,
               UValueComparator *valueComp,
               int32_t size,
               UErrorCode *status) {

    // Find the smallest index i for which PRIMES[i] >= size.
    int32_t i = 0;
    while (i<(PRIMES_LENGTH-1) && PRIMES[i]<size) {
        ++i;
    }
    return _uhash_init(fillinResult, keyHash, keyComp, valueComp, i, status);
}

U_CAPI void U_EXPORT2
uhash_close(UHashtable *hash) {
    if (hash == NULL) {
        return;
    }
    if (hash->elements != NULL) {
        if (hash->keyDeleter != NULL || hash->valueDeleter != NULL) {
            int32_t pos=UHASH_FIRST;
            UHashElement *e;
            while ((e = (UHashElement*) uhash_nextElement(hash, &pos)) != NULL) {
                HASH_DELETE_KEY_VALUE(hash, e->key.pointer, e->value.pointer);
            }
        }
        uprv_free(hash->elements);
        hash->elements = NULL;
    }
    if (hash->allocated) {
        uprv_free(hash);
    }
}

U_CAPI UHashFunction *U_EXPORT2
uhash_setKeyHasher(UHashtable *hash, UHashFunction *fn) {
    UHashFunction *result = hash->keyHasher;
    hash->keyHasher = fn;
    return result;
}

U_CAPI UKeyComparator *U_EXPORT2
uhash_setKeyComparator(UHashtable *hash, UKeyComparator *fn) {
    UKeyComparator *result = hash->keyComparator;
    hash->keyComparator = fn;
    return result;
}
U_CAPI UValueComparator *U_EXPORT2
uhash_setValueComparator(UHashtable *hash, UValueComparator *fn){
    UValueComparator *result = hash->valueComparator;
    hash->valueComparator = fn;
    return result;
}

U_CAPI UObjectDeleter *U_EXPORT2
uhash_setKeyDeleter(UHashtable *hash, UObjectDeleter *fn) {
    UObjectDeleter *result = hash->keyDeleter;
    hash->keyDeleter = fn;
    return result;
}

U_CAPI UObjectDeleter *U_EXPORT2
uhash_setValueDeleter(UHashtable *hash, UObjectDeleter *fn) {
    UObjectDeleter *result = hash->valueDeleter;
    hash->valueDeleter = fn;
    return result;
}

U_CAPI void U_EXPORT2
uhash_setResizePolicy(UHashtable *hash, enum UHashResizePolicy policy) {
    UErrorCode status = U_ZERO_ERROR;
    _uhash_internalSetResizePolicy(hash, policy);
    hash->lowWaterMark  = (int32_t)(hash->length * hash->lowWaterRatio);
    hash->highWaterMark = (int32_t)(hash->length * hash->highWaterRatio);
    _uhash_rehash(hash, &status);
}

U_CAPI int32_t U_EXPORT2
uhash_count(const UHashtable *hash) {
    return hash->count;
}

U_CAPI void* U_EXPORT2
uhash_get(const UHashtable *hash,
          const void* key) {
    UHashTok keyholder;
    keyholder.pointer = (void*) key;
    return _uhash_find(hash, keyholder, hash->keyHasher(keyholder))->value.pointer;
}

U_CAPI void* U_EXPORT2
uhash_iget(const UHashtable *hash,
           int32_t key) {
    UHashTok keyholder;
    keyholder.integer = key;
    return _uhash_find(hash, keyholder, hash->keyHasher(keyholder))->value.pointer;
}

U_CAPI int32_t U_EXPORT2
uhash_geti(const UHashtable *hash,
           const void* key) {
    UHashTok keyholder;
    keyholder.pointer = (void*) key;
    return _uhash_find(hash, keyholder, hash->keyHasher(keyholder))->value.integer;
}

U_CAPI int32_t U_EXPORT2
uhash_igeti(const UHashtable *hash,
           int32_t key) {
    UHashTok keyholder;
    keyholder.integer = key;
    return _uhash_find(hash, keyholder, hash->keyHasher(keyholder))->value.integer;
}

U_CAPI void* U_EXPORT2
uhash_put(UHashtable *hash,
          void* key,
          void* value,
          UErrorCode *status) {
    UHashTok keyholder, valueholder;
    keyholder.pointer = key;
    valueholder.pointer = value;
    return _uhash_put(hash, keyholder, valueholder,
                      HINT_KEY_POINTER | HINT_VALUE_POINTER,
                      status).pointer;
}

U_CAPI void* U_EXPORT2
uhash_iput(UHashtable *hash,
           int32_t key,
           void* value,
           UErrorCode *status) {
    UHashTok keyholder, valueholder;
    keyholder.integer = key;
    valueholder.pointer = value;
    return _uhash_put(hash, keyholder, valueholder,
                      HINT_VALUE_POINTER,
                      status).pointer;
}

U_CAPI int32_t U_EXPORT2
uhash_puti(UHashtable *hash,
           void* key,
           int32_t value,
           UErrorCode *status) {
    UHashTok keyholder, valueholder;
    keyholder.pointer = key;
    valueholder.integer = value;
    return _uhash_put(hash, keyholder, valueholder,
                      HINT_KEY_POINTER,
                      status).integer;
}


U_CAPI int32_t U_EXPORT2
uhash_iputi(UHashtable *hash,
           int32_t key,
           int32_t value,
           UErrorCode *status) {
    UHashTok keyholder, valueholder;
    keyholder.integer = key;
    valueholder.integer = value;
    return _uhash_put(hash, keyholder, valueholder,
                      0, /* neither is a ptr */
                      status).integer;
}

U_CAPI void* U_EXPORT2
uhash_remove(UHashtable *hash,
             const void* key) {
    UHashTok keyholder;
    keyholder.pointer = (void*) key;
    return _uhash_remove(hash, keyholder).pointer;
}

U_CAPI void* U_EXPORT2
uhash_iremove(UHashtable *hash,
              int32_t key) {
    UHashTok keyholder;
    keyholder.integer = key;
    return _uhash_remove(hash, keyholder).pointer;
}

U_CAPI int32_t U_EXPORT2
uhash_removei(UHashtable *hash,
              const void* key) {
    UHashTok keyholder;
    keyholder.pointer = (void*) key;
    return _uhash_remove(hash, keyholder).integer;
}

U_CAPI int32_t U_EXPORT2
uhash_iremovei(UHashtable *hash,
               int32_t key) {
    UHashTok keyholder;
    keyholder.integer = key;
    return _uhash_remove(hash, keyholder).integer;
}

U_CAPI void U_EXPORT2
uhash_removeAll(UHashtable *hash) {
    int32_t pos = UHASH_FIRST;
    const UHashElement *e;
    U_ASSERT(hash != NULL);
    if (hash->count != 0) {
        while ((e = uhash_nextElement(hash, &pos)) != NULL) {
            uhash_removeElement(hash, e);
        }
    }
    U_ASSERT(hash->count == 0);
}

U_CAPI const UHashElement* U_EXPORT2
uhash_find(const UHashtable *hash, const void* key) {
    UHashTok keyholder;
    const UHashElement *e;
    keyholder.pointer = (void*) key;
    e = _uhash_find(hash, keyholder, hash->keyHasher(keyholder));
    return IS_EMPTY_OR_DELETED(e->hashcode) ? NULL : e;
}

U_CAPI const UHashElement* U_EXPORT2
uhash_nextElement(const UHashtable *hash, int32_t *pos) {
    /* Walk through the array until we find an element that is not
     * EMPTY and not DELETED.
     */
    int32_t i;
    U_ASSERT(hash != NULL);
    for (i = *pos + 1; i < hash->length; ++i) {
        if (!IS_EMPTY_OR_DELETED(hash->elements[i].hashcode)) {
            *pos = i;
            return &(hash->elements[i]);
        }
    }

    /* No more elements */
    return NULL;
}

U_CAPI void* U_EXPORT2
uhash_removeElement(UHashtable *hash, const UHashElement* e) {
    U_ASSERT(hash != NULL);
    U_ASSERT(e != NULL);
    if (!IS_EMPTY_OR_DELETED(e->hashcode)) {
        UHashElement *nce = (UHashElement *)e;
        return _uhash_internalRemoveElement(hash, nce).pointer;
    }
    return NULL;
}

/********************************************************************
 * UHashTok convenience
 ********************************************************************/

/**
 * Return a UHashTok for an integer.
 */
/*U_CAPI UHashTok U_EXPORT2
uhash_toki(int32_t i) {
    UHashTok tok;
    tok.integer = i;
    return tok;
}*/

/**
 * Return a UHashTok for a pointer.
 */
/*U_CAPI UHashTok U_EXPORT2
uhash_tokp(void* p) {
    UHashTok tok;
    tok.pointer = p;
    return tok;
}*/

/********************************************************************
 * PUBLIC Key Hash Functions
 ********************************************************************/

U_CAPI int32_t U_EXPORT2
uhash_hashUChars(const UHashTok key) {
    const UChar *s = (const UChar *)key.pointer;
    return s == NULL ? 0 : ustr_hashUCharsN(s, u_strlen(s));
}

U_CAPI int32_t U_EXPORT2
uhash_hashChars(const UHashTok key) {
    const char *s = (const char *)key.pointer;
    return s == NULL ? 0 : static_cast<int32_t>(ustr_hashCharsN(s, static_cast<int32_t>(uprv_strlen(s))));
}

U_CAPI int32_t U_EXPORT2
uhash_hashIChars(const UHashTok key) {
    const char *s = (const char *)key.pointer;
    return s == NULL ? 0 : ustr_hashICharsN(s, static_cast<int32_t>(uprv_strlen(s)));
}

U_CAPI UBool U_EXPORT2
uhash_equals(const UHashtable* hash1, const UHashtable* hash2){
    int32_t count1, count2, pos, i;

    if(hash1==hash2){
        return TRUE;
    }

    /*
     * Make sure that we are comparing 2 valid hashes of the same type
     * with valid comparison functions.
     * Without valid comparison functions, a binary comparison
     * of the hash values will yield random results on machines
     * with 64-bit pointers and 32-bit integer hashes.
     * A valueComparator is normally optional.
     */
    if (hash1==NULL || hash2==NULL ||
        hash1->keyComparator != hash2->keyComparator ||
        hash1->valueComparator != hash2->valueComparator ||
        hash1->valueComparator == NULL)
    {
        /*
        Normally we would return an error here about incompatible hash tables,
        but we return FALSE instead.
        */
        return FALSE;
    }

    count1 = uhash_count(hash1);
    count2 = uhash_count(hash2);
    if(count1!=count2){
        return FALSE;
    }

    pos=UHASH_FIRST;
    for(i=0; i<count1; i++){
        const UHashElement* elem1 = uhash_nextElement(hash1, &pos);
        const UHashTok key1 = elem1->key;
        const UHashTok val1 = elem1->value;
        /* here the keys are not compared, instead the key form hash1 is used to fetch
         * value from hash2. If the hashes are equal then then both hashes should
         * contain equal values for the same key!
         */
        const UHashElement* elem2 = _uhash_find(hash2, key1, hash2->keyHasher(key1));
        const UHashTok val2 = elem2->value;
        if(hash1->valueComparator(val1, val2)==FALSE){
            return FALSE;
        }
    }
    return TRUE;
}

/********************************************************************
 * PUBLIC Comparator Functions
 ********************************************************************/

U_CAPI UBool U_EXPORT2
uhash_compareUChars(const UHashTok key1, const UHashTok key2) {
    const UChar *p1 = (const UChar*) key1.pointer;
    const UChar *p2 = (const UChar*) key2.pointer;
    if (p1 == p2) {
        return TRUE;
    }
    if (p1 == NULL || p2 == NULL) {
        return FALSE;
    }
    while (*p1 != 0 && *p1 == *p2) {
        ++p1;
        ++p2;
    }
    return (UBool)(*p1 == *p2);
}

U_CAPI UBool U_EXPORT2
uhash_compareChars(const UHashTok key1, const UHashTok key2) {
    const char *p1 = (const char*) key1.pointer;
    const char *p2 = (const char*) key2.pointer;
    if (p1 == p2) {
        return TRUE;
    }
    if (p1 == NULL || p2 == NULL) {
        return FALSE;
    }
    while (*p1 != 0 && *p1 == *p2) {
        ++p1;
        ++p2;
    }
    return (UBool)(*p1 == *p2);
}

U_CAPI UBool U_EXPORT2
uhash_compareIChars(const UHashTok key1, const UHashTok key2) {
    const char *p1 = (const char*) key1.pointer;
    const char *p2 = (const char*) key2.pointer;
    if (p1 == p2) {
        return TRUE;
    }
    if (p1 == NULL || p2 == NULL) {
        return FALSE;
    }
    while (*p1 != 0 && uprv_tolower(*p1) == uprv_tolower(*p2)) {
        ++p1;
        ++p2;
    }
    return (UBool)(*p1 == *p2);
}

/********************************************************************
 * PUBLIC int32_t Support Functions
 ********************************************************************/

U_CAPI int32_t U_EXPORT2
uhash_hashLong(const UHashTok key) {
    return key.integer;
}

U_CAPI UBool U_EXPORT2
uhash_compareLong(const UHashTok key1, const UHashTok key2) {
    return (UBool)(key1.integer == key2.integer);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 1997-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*   Date        Name        Description
*   03/22/00    aliu        Creation.
*   07/06/01    aliu        Modified to support int32_t keys on
*                           platforms with sizeof(void*) < 32.
******************************************************************************
*/



/**
 * Deleter for Hashtable objects.
 */
U_CAPI void U_EXPORT2
uhash_deleteHashtable(void *obj) {
    U_NAMESPACE_USE
    delete (Hashtable*) obj;
}

//eof
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
//  *******************************************************************************
//  *
//  *   Copyright (C) 2003-2014, International Business Machines
//  *   Corporation and others.  All Rights Reserved.
//  *
//  *******************************************************************************
//  *   file name:  uidna.cpp
//  *   encoding:   UTF-8
//  *   tab size:   8 (not used)
//  *   indentation:4
//  *
//  *   created on: 2003feb1
//  *   created by: Ram Viswanadha
//  */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_IDNA

// #include "unicode/uidna.h"
// #include "unicode/ustring.h"
// #include "unicode/usprep.h"
// #include "punycode.h"
// #include "ustr_imp.h"
// #include "cmemory.h"
// #include "uassert.h"
// #include "sprpimpl.h"

// /* it is official IDNA ACE Prefix is "xn--" */
// static const UChar ACE_PREFIX[] ={ 0x0078,0x006E,0x002d,0x002d } ;
// #define ACE_PREFIX_LENGTH 4

// #define MAX_LABEL_LENGTH 63
// /* The Max length of the labels should not be more than MAX_LABEL_LENGTH */
// #define MAX_LABEL_BUFFER_SIZE 100

// #define MAX_DOMAIN_NAME_LENGTH 255
// /* The Max length of the domain names should not be more than MAX_DOMAIN_NAME_LENGTH */
// #define MAX_IDN_BUFFER_SIZE   MAX_DOMAIN_NAME_LENGTH+1

// #define LOWER_CASE_DELTA 0x0020
// #define HYPHEN           0x002D
// #define FULL_STOP        0x002E
// #define CAPITAL_A        0x0041
// #define CAPITAL_Z        0x005A

// inline static UChar
// toASCIILower(UChar ch){
//     if(CAPITAL_A <= ch && ch <= CAPITAL_Z){
//         return ch + LOWER_CASE_DELTA;
//     }
//     return ch;
// }

// inline static UBool
// startsWithPrefix(const UChar* src , int32_t srcLength){
//     if(srcLength < ACE_PREFIX_LENGTH){
//         return FALSE;
//     }

//     for(int8_t i=0; i< ACE_PREFIX_LENGTH; i++){
//         if(toASCIILower(src[i]) != ACE_PREFIX[i]){
//             return FALSE;
//         }
//     }
//     return TRUE;
// }


// inline static int32_t
// compareCaseInsensitiveASCII(const UChar* s1, int32_t s1Len,
//                             const UChar* s2, int32_t s2Len){

//     int32_t minLength;
//     int32_t lengthResult;

//     // are we comparing different lengths?
//     if(s1Len != s2Len) {
//         if(s1Len < s2Len) {
//             minLength = s1Len;
//             lengthResult = -1;
//         } else {
//             minLength = s2Len;
//             lengthResult = 1;
//         }
//     } else {
//         // ok the lengths are equal
//         minLength = s1Len;
//         lengthResult = 0;
//     }

//     UChar c1,c2;
//     int32_t rc;

//     for(int32_t i =0;/* no condition */;i++) {

//         /* If we reach the ends of both strings then they match */
//         if(i == minLength) {
//             return lengthResult;
//         }

//         c1 = s1[i];
//         c2 = s2[i];

//         /* Case-insensitive comparison */
//         if(c1!=c2) {
//             rc=(int32_t)toASCIILower(c1)-(int32_t)toASCIILower(c2);
//             if(rc!=0) {
//                 lengthResult=rc;
//                 break;
//             }
//         }
//     }
//     return lengthResult;
// }


// /**
//  * Ascertain if the given code point is a label separator as
//  * defined by the IDNA RFC
//  *
//  * @param ch The code point to be ascertained
//  * @return true if the char is a label separator
//  * @stable ICU 2.8
//  */
// static inline UBool isLabelSeparator(UChar ch){
//     switch(ch){
//         case 0x002e:
//         case 0x3002:
//         case 0xFF0E:
//         case 0xFF61:
//             return TRUE;
//         default:
//             return FALSE;
//     }
// }

// // returns the length of the label excluding the separator
// // if *limit == separator then the length returned does not include
// // the separtor.
// static inline int32_t
// getNextSeparator(UChar *src, int32_t srcLength,
//                  UChar **limit, UBool *done){
//     if(srcLength == -1){
//         int32_t i;
//         for(i=0 ; ;i++){
//             if(src[i] == 0){
//                 *limit = src + i; // point to null
//                 *done = TRUE;
//                 return i;
//             }
//             if(isLabelSeparator(src[i])){
//                 *limit = src + (i+1); // go past the delimiter
//                 return i;

//             }
//         }
//     }else{
//         int32_t i;
//         for(i=0;i<srcLength;i++){
//             if(isLabelSeparator(src[i])){
//                 *limit = src + (i+1); // go past the delimiter
//                 return i;
//             }
//         }
//         // we have not found the delimiter
//         // if(i==srcLength)
//         *limit = src+srcLength;
//         *done = TRUE;

//         return i;
//     }
// }
// static inline UBool isLDHChar(UChar ch){
//     // high runner case
//     if(ch>0x007A){
//         return FALSE;
//     }
//     //[\\u002D \\u0030-\\u0039 \\u0041-\\u005A \\u0061-\\u007A]
//     if( (ch==0x002D) ||
//         (0x0030 <= ch && ch <= 0x0039) ||
//         (0x0041 <= ch && ch <= 0x005A) ||
//         (0x0061 <= ch && ch <= 0x007A)
//       ){
//         return TRUE;
//     }
//     return FALSE;
// }

// static int32_t
// _internal_toASCII(const UChar* src, int32_t srcLength,
//                   UChar* dest, int32_t destCapacity,
//                   int32_t options,
//                   UStringPrepProfile* nameprep,
//                   UParseError* parseError,
//                   UErrorCode* status)
// {

//     // TODO Revisit buffer handling. The label should not be over 63 ASCII characters. ICU4J may need to be updated too.
//     UChar b1Stack[MAX_LABEL_BUFFER_SIZE], b2Stack[MAX_LABEL_BUFFER_SIZE];
//     //initialize pointers to stack buffers
//     UChar  *b1 = b1Stack, *b2 = b2Stack;
//     int32_t b1Len=0, b2Len,
//             b1Capacity = MAX_LABEL_BUFFER_SIZE,
//             b2Capacity = MAX_LABEL_BUFFER_SIZE ,
//             reqLength=0;

//     int32_t namePrepOptions = ((options & UIDNA_ALLOW_UNASSIGNED) != 0) ? USPREP_ALLOW_UNASSIGNED: 0;
//     UBool* caseFlags = NULL;

//     // the source contains all ascii codepoints
//     UBool srcIsASCII  = TRUE;
//     // assume the source contains all LDH codepoints
//     UBool srcIsLDH = TRUE;

//     int32_t j=0;

//     //get the options
//     UBool useSTD3ASCIIRules = (UBool)((options & UIDNA_USE_STD3_RULES) != 0);

//     int32_t failPos = -1;

//     if(srcLength == -1){
//         srcLength = u_strlen(src);
//     }

//     if(srcLength > b1Capacity){
//         b1 = (UChar*) uprv_malloc(srcLength * U_SIZEOF_UCHAR);
//         if(b1==NULL){
//             *status = U_MEMORY_ALLOCATION_ERROR;
//             goto CLEANUP;
//         }
//         b1Capacity = srcLength;
//     }

//     // step 1
//     for( j=0;j<srcLength;j++){
//         if(src[j] > 0x7F){
//             srcIsASCII = FALSE;
//         }
//         b1[b1Len++] = src[j];
//     }

//     // step 2 is performed only if the source contains non ASCII
//     if(srcIsASCII == FALSE){

//         // step 2
//         b1Len = usprep_prepare(nameprep, src, srcLength, b1, b1Capacity, namePrepOptions, parseError, status);

//         if(*status == U_BUFFER_OVERFLOW_ERROR){
//             // redo processing of string
//             // we do not have enough room so grow the buffer
//             if(b1 != b1Stack){
//                 uprv_free(b1);
//             }
//             b1 = (UChar*) uprv_malloc(b1Len * U_SIZEOF_UCHAR);
//             if(b1==NULL){
//                 *status = U_MEMORY_ALLOCATION_ERROR;
//                 goto CLEANUP;
//             }

//             *status = U_ZERO_ERROR; // reset error

//             b1Len = usprep_prepare(nameprep, src, srcLength, b1, b1Len, namePrepOptions, parseError, status);
//         }
//     }
//     // error bail out
//     if(U_FAILURE(*status)){
//         goto CLEANUP;
//     }
//     if(b1Len == 0){
//         *status = U_IDNA_ZERO_LENGTH_LABEL_ERROR;
//         goto CLEANUP;
//     }

//     // for step 3 & 4
//     srcIsASCII = TRUE;
//     for( j=0;j<b1Len;j++){
//         // check if output of usprep_prepare is all ASCII
//         if(b1[j] > 0x7F){
//             srcIsASCII = FALSE;
//         }else if(isLDHChar(b1[j])==FALSE){  // if the char is in ASCII range verify that it is an LDH character
//             srcIsLDH = FALSE;
//             failPos = j;
//         }
//     }
//     if(useSTD3ASCIIRules == TRUE){
//         // verify 3a and 3b
//         // 3(a) Verify the absence of non-LDH ASCII code points; that is, the
//         //  absence of 0..2C, 2E..2F, 3A..40, 5B..60, and 7B..7F.
//         // 3(b) Verify the absence of leading and trailing hyphen-minus; that
//         //  is, the absence of U+002D at the beginning and end of the
//         //  sequence.
//         if( srcIsLDH == FALSE /* source at this point should not contain anyLDH characters */
//             || b1[0] ==  HYPHEN || b1[b1Len-1] == HYPHEN){
//             *status = U_IDNA_STD3_ASCII_RULES_ERROR;

//             /* populate the parseError struct */
//             if(srcIsLDH==FALSE){
//                 // failPos is always set the index of failure
//                 uprv_syntaxError(b1,failPos, b1Len,parseError);
//             }else if(b1[0] == HYPHEN){
//                 // fail position is 0
//                 uprv_syntaxError(b1,0,b1Len,parseError);
//             }else{
//                 // the last index in the source is always length-1
//                 uprv_syntaxError(b1, (b1Len>0) ? b1Len-1 : b1Len, b1Len,parseError);
//             }

//             goto CLEANUP;
//         }
//     }
//     // Step 4: if the source is ASCII then proceed to step 8
//     if(srcIsASCII){
//         if(b1Len <= destCapacity){
//             u_memmove(dest, b1, b1Len);
//             reqLength = b1Len;
//         }else{
//             reqLength = b1Len;
//             goto CLEANUP;
//         }
//     }else{
//         // step 5 : verify the sequence does not begin with ACE prefix
//         if(!startsWithPrefix(b1,b1Len)){

//             //step 6: encode the sequence with punycode

//             // do not preserve the case flags for now!
//             // TODO: Preserve the case while implementing the RFE
//             // caseFlags = (UBool*) uprv_malloc(b1Len * sizeof(UBool));
//             // uprv_memset(caseFlags,TRUE,b1Len);

//             b2Len = u_strToPunycode(b1,b1Len,b2,b2Capacity,caseFlags, status);

//             if(*status == U_BUFFER_OVERFLOW_ERROR){
//                 // redo processing of string
//                 /* we do not have enough room so grow the buffer*/
//                 b2 = (UChar*) uprv_malloc(b2Len * U_SIZEOF_UCHAR);
//                 if(b2 == NULL){
//                     *status = U_MEMORY_ALLOCATION_ERROR;
//                     goto CLEANUP;
//                 }

//                 *status = U_ZERO_ERROR; // reset error

//                 b2Len = u_strToPunycode(b1,b1Len,b2,b2Len,caseFlags, status);
//             }
//             //error bail out
//             if(U_FAILURE(*status)){
//                 goto CLEANUP;
//             }
//             // TODO : Reconsider while implementing the case preserve RFE
//             // convert all codepoints to lower case ASCII
//             // toASCIILower(b2,b2Len);
//             reqLength = b2Len+ACE_PREFIX_LENGTH;

//             if(reqLength > destCapacity){
//                 *status = U_BUFFER_OVERFLOW_ERROR;
//                 goto CLEANUP;
//             }
//             //Step 7: prepend the ACE prefix
//             u_memcpy(dest, ACE_PREFIX, ACE_PREFIX_LENGTH);
//             //Step 6: copy the contents in b2 into dest
//             u_memcpy(dest+ACE_PREFIX_LENGTH, b2, b2Len);

//         }else{
//             *status = U_IDNA_ACE_PREFIX_ERROR;
//             //position of failure is 0
//             uprv_syntaxError(b1,0,b1Len,parseError);
//             goto CLEANUP;
//         }
//     }
//     // step 8: verify the length of label
//     if(reqLength > MAX_LABEL_LENGTH){
//         *status = U_IDNA_LABEL_TOO_LONG_ERROR;
//     }

// CLEANUP:
//     if(b1 != b1Stack){
//         uprv_free(b1);
//     }
//     if(b2 != b2Stack){
//         uprv_free(b2);
//     }
//     uprv_free(caseFlags);

//     return u_terminateUChars(dest, destCapacity, reqLength, status);
// }

// static int32_t
// _internal_toUnicode(const UChar* src, int32_t srcLength,
//                     UChar* dest, int32_t destCapacity,
//                     int32_t options,
//                     UStringPrepProfile* nameprep,
//                     UParseError* parseError,
//                     UErrorCode* status)
// {

//     //get the options
//     //UBool useSTD3ASCIIRules = (UBool)((options & UIDNA_USE_STD3_RULES) != 0);
//     int32_t namePrepOptions = ((options & UIDNA_ALLOW_UNASSIGNED) != 0) ? USPREP_ALLOW_UNASSIGNED: 0;

//     // TODO Revisit buffer handling. The label should not be over 63 ASCII characters. ICU4J may need to be updated too.
//     UChar b1Stack[MAX_LABEL_BUFFER_SIZE], b2Stack[MAX_LABEL_BUFFER_SIZE], b3Stack[MAX_LABEL_BUFFER_SIZE];

//     //initialize pointers to stack buffers
//     UChar  *b1 = b1Stack, *b2 = b2Stack, *b1Prime=NULL, *b3=b3Stack;
//     int32_t b1Len = 0, b2Len, b1PrimeLen, b3Len,
//             b1Capacity = MAX_LABEL_BUFFER_SIZE,
//             b2Capacity = MAX_LABEL_BUFFER_SIZE,
//             b3Capacity = MAX_LABEL_BUFFER_SIZE,
//             reqLength=0;

//     UBool* caseFlags = NULL;

//     UBool srcIsASCII = TRUE;
//     /*UBool srcIsLDH = TRUE;
//     int32_t failPos =0;*/

//     // step 1: find out if all the codepoints in src are ASCII
//     if(srcLength==-1){
//         srcLength = 0;
//         for(;src[srcLength]!=0;){
//             if(src[srcLength]> 0x7f){
//                 srcIsASCII = FALSE;
//             }/*else if(isLDHChar(src[srcLength])==FALSE){
//                 // here we do not assemble surrogates
//                 // since we know that LDH code points
//                 // are in the ASCII range only
//                 srcIsLDH = FALSE;
//                 failPos = srcLength;
//             }*/
//             srcLength++;
//         }
//     }else if(srcLength > 0){
//         for(int32_t j=0; j<srcLength; j++){
//             if(src[j]> 0x7f){
//                 srcIsASCII = FALSE;
//                 break;
//             }/*else if(isLDHChar(src[j])==FALSE){
//                 // here we do not assemble surrogates
//                 // since we know that LDH code points
//                 // are in the ASCII range only
//                 srcIsLDH = FALSE;
//                 failPos = j;
//             }*/
//         }
//     }else{
//         return 0;
//     }

//     if(srcIsASCII == FALSE){
//         // step 2: process the string
//         b1Len = usprep_prepare(nameprep, src, srcLength, b1, b1Capacity, namePrepOptions, parseError, status);
//         if(*status == U_BUFFER_OVERFLOW_ERROR){
//             // redo processing of string
//             /* we do not have enough room so grow the buffer*/
//             b1 = (UChar*) uprv_malloc(b1Len * U_SIZEOF_UCHAR);
//             if(b1==NULL){
//                 *status = U_MEMORY_ALLOCATION_ERROR;
//                 goto CLEANUP;
//             }

//             *status = U_ZERO_ERROR; // reset error

//             b1Len = usprep_prepare(nameprep, src, srcLength, b1, b1Len, namePrepOptions, parseError, status);
//         }
//         //bail out on error
//         if(U_FAILURE(*status)){
//             goto CLEANUP;
//         }
//     }else{

//         //just point src to b1
//         b1 = (UChar*) src;
//         b1Len = srcLength;
//     }

//     // The RFC states that
//     // <quote>
//     // ToUnicode never fails. If any step fails, then the original input
//     // is returned immediately in that step.
//     // </quote>

//     //step 3: verify ACE Prefix
//     if(startsWithPrefix(b1,b1Len)){

//         //step 4: Remove the ACE Prefix
//         b1Prime = b1 + ACE_PREFIX_LENGTH;
//         b1PrimeLen  = b1Len - ACE_PREFIX_LENGTH;

//         //step 5: Decode using punycode
//         b2Len = u_strFromPunycode(b1Prime, b1PrimeLen, b2, b2Capacity, caseFlags,status);

//         if(*status == U_BUFFER_OVERFLOW_ERROR){
//             // redo processing of string
//             /* we do not have enough room so grow the buffer*/
//             b2 = (UChar*) uprv_malloc(b2Len * U_SIZEOF_UCHAR);
//             if(b2==NULL){
//                 *status = U_MEMORY_ALLOCATION_ERROR;
//                 goto CLEANUP;
//             }

//             *status = U_ZERO_ERROR; // reset error

//             b2Len =  u_strFromPunycode(b1Prime, b1PrimeLen, b2, b2Len, caseFlags, status);
//         }


//         //step 6:Apply toASCII
//         b3Len = uidna_toASCII(b2, b2Len, b3, b3Capacity, options, parseError, status);

//         if(*status == U_BUFFER_OVERFLOW_ERROR){
//             // redo processing of string
//             /* we do not have enough room so grow the buffer*/
//             b3 = (UChar*) uprv_malloc(b3Len * U_SIZEOF_UCHAR);
//             if(b3==NULL){
//                 *status = U_MEMORY_ALLOCATION_ERROR;
//                 goto CLEANUP;
//             }

//             *status = U_ZERO_ERROR; // reset error

//             b3Len =  uidna_toASCII(b2,b2Len,b3,b3Len,options,parseError, status);

//         }
//         //bail out on error
//         if(U_FAILURE(*status)){
//             goto CLEANUP;
//         }

//         //step 7: verify
//         if(compareCaseInsensitiveASCII(b1, b1Len, b3, b3Len) !=0){
//             // Cause the original to be returned.
//             *status = U_IDNA_VERIFICATION_ERROR;
//             goto CLEANUP;
//         }

//         //step 8: return output of step 5
//         reqLength = b2Len;
//         if(b2Len <= destCapacity) {
//             u_memmove(dest, b2, b2Len);
//         }
//     }
//     else{
//         // See the start of this if statement for why this is commented out.
//         // verify that STD3 ASCII rules are satisfied
//         /*if(useSTD3ASCIIRules == TRUE){
//             if( srcIsLDH == FALSE // source contains some non-LDH characters
//                 || src[0] ==  HYPHEN || src[srcLength-1] == HYPHEN){
//                 *status = U_IDNA_STD3_ASCII_RULES_ERROR;

//                 // populate the parseError struct
//                 if(srcIsLDH==FALSE){
//                     // failPos is always set the index of failure
//                     uprv_syntaxError(src,failPos, srcLength,parseError);
//                 }else if(src[0] == HYPHEN){
//                     // fail position is 0
//                     uprv_syntaxError(src,0,srcLength,parseError);
//                 }else{
//                     // the last index in the source is always length-1
//                     uprv_syntaxError(src, (srcLength>0) ? srcLength-1 : srcLength, srcLength,parseError);
//                 }

//                 goto CLEANUP;
//             }
//         }*/
//         // just return the source
//         //copy the source to destination
//         if(srcLength <= destCapacity){
//             u_memmove(dest, src, srcLength);
//         }
//         reqLength = srcLength;
//     }


// CLEANUP:

//     if(b1 != b1Stack && b1!=src){
//         uprv_free(b1);
//     }
//     if(b2 != b2Stack){
//         uprv_free(b2);
//     }
//     uprv_free(caseFlags);

//     // The RFC states that
//     // <quote>
//     // ToUnicode never fails. If any step fails, then the original input
//     // is returned immediately in that step.
//     // </quote>
//     // So if any step fails lets copy source to destination
//     if(U_FAILURE(*status)){
//         //copy the source to destination
//         if(dest && srcLength <= destCapacity){
//             // srcLength should have already been set earlier.
//             U_ASSERT(srcLength >= 0);
//             u_memmove(dest, src, srcLength);
//         }
//         reqLength = srcLength;
//         *status = U_ZERO_ERROR;
//     }

//     return u_terminateUChars(dest, destCapacity, reqLength, status);
// }

// U_CAPI int32_t U_EXPORT2
// uidna_toASCII(const UChar* src, int32_t srcLength,
//               UChar* dest, int32_t destCapacity,
//               int32_t options,
//               UParseError* parseError,
//               UErrorCode* status){

//     if(status == NULL || U_FAILURE(*status)){
//         return 0;
//     }
//     if((src==NULL) || (srcLength < -1) || (destCapacity<0) || (!dest && destCapacity > 0)){
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     UStringPrepProfile* nameprep = usprep_openByType(USPREP_RFC3491_NAMEPREP, status);

//     if(U_FAILURE(*status)){
//         return -1;
//     }

//     int32_t retLen = _internal_toASCII(src, srcLength, dest, destCapacity, options, nameprep, parseError, status);

//     /* close the profile*/
//     usprep_close(nameprep);

//     return retLen;
// }

// U_CAPI int32_t U_EXPORT2
// uidna_toUnicode(const UChar* src, int32_t srcLength,
//                 UChar* dest, int32_t destCapacity,
//                 int32_t options,
//                 UParseError* parseError,
//                 UErrorCode* status){

//     if(status == NULL || U_FAILURE(*status)){
//         return 0;
//     }
//     if( (src==NULL) || (srcLength < -1) || (destCapacity<0) || (!dest && destCapacity > 0)){
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     UStringPrepProfile* nameprep = usprep_openByType(USPREP_RFC3491_NAMEPREP, status);

//     if(U_FAILURE(*status)){
//         return -1;
//     }

//     int32_t retLen = _internal_toUnicode(src, srcLength, dest, destCapacity, options, nameprep, parseError, status);

//     usprep_close(nameprep);

//     return retLen;
// }


// U_CAPI int32_t U_EXPORT2
// uidna_IDNToASCII(  const UChar *src, int32_t srcLength,
//                    UChar* dest, int32_t destCapacity,
//                    int32_t options,
//                    UParseError *parseError,
//                    UErrorCode *status){

//     if(status == NULL || U_FAILURE(*status)){
//         return 0;
//     }
//     if((src==NULL) || (srcLength < -1) || (destCapacity<0) || (!dest && destCapacity > 0)){
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     int32_t reqLength = 0;

//     UStringPrepProfile* nameprep = usprep_openByType(USPREP_RFC3491_NAMEPREP, status);

//     if(U_FAILURE(*status)){
//         return 0;
//     }

//     //initialize pointers
//     UChar *delimiter = (UChar*)src;
//     UChar *labelStart = (UChar*)src;
//     UChar *currentDest = (UChar*) dest;
//     int32_t remainingLen = srcLength;
//     int32_t remainingDestCapacity = destCapacity;
//     int32_t labelLen = 0, labelReqLength = 0;
//     UBool done = FALSE;


//     for(;;){

//         labelLen = getNextSeparator(labelStart,remainingLen, &delimiter,&done);
//         labelReqLength = 0;
//         if(!(labelLen==0 && done)){// make sure this is not a root label separator.

//             labelReqLength = _internal_toASCII( labelStart, labelLen,
//                                                 currentDest, remainingDestCapacity,
//                                                 options, nameprep,
//                                                 parseError, status);

//             if(*status == U_BUFFER_OVERFLOW_ERROR){

//                 *status = U_ZERO_ERROR; // reset error
//                 remainingDestCapacity = 0;
//             }
//         }


//         if(U_FAILURE(*status)){
//             break;
//         }

//         reqLength +=labelReqLength;
//         // adjust the destination pointer
//         if(labelReqLength < remainingDestCapacity){
//             currentDest = currentDest + labelReqLength;
//             remainingDestCapacity -= labelReqLength;
//         }else{
//             // should never occur
//             remainingDestCapacity = 0;
//         }

//         if(done == TRUE){
//             break;
//         }

//         // add the label separator
//         if(remainingDestCapacity > 0){
//             *currentDest++ = FULL_STOP;
//             remainingDestCapacity--;
//         }
//         reqLength++;

//         labelStart = delimiter;
//         if(remainingLen >0 ){
//             remainingLen = (int32_t)(srcLength - (delimiter - src));
//         }

//     }

//     if(reqLength > MAX_DOMAIN_NAME_LENGTH){
//         *status = U_IDNA_DOMAIN_NAME_TOO_LONG_ERROR;
//     }

//     usprep_close(nameprep);

//     return u_terminateUChars(dest, destCapacity, reqLength, status);
// }

// U_CAPI int32_t U_EXPORT2
// uidna_IDNToUnicode(  const UChar* src, int32_t srcLength,
//                      UChar* dest, int32_t destCapacity,
//                      int32_t options,
//                      UParseError* parseError,
//                      UErrorCode* status){

//     if(status == NULL || U_FAILURE(*status)){
//         return 0;
//     }
//     if((src==NULL) || (srcLength < -1) || (destCapacity<0) || (!dest && destCapacity > 0)){
//         *status = U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     int32_t reqLength = 0;

//     UStringPrepProfile* nameprep = usprep_openByType(USPREP_RFC3491_NAMEPREP, status);

//     if(U_FAILURE(*status)){
//         return 0;
//     }

//     //initialize pointers
//     UChar *delimiter = (UChar*)src;
//     UChar *labelStart = (UChar*)src;
//     UChar *currentDest = (UChar*) dest;
//     int32_t remainingLen = srcLength;
//     int32_t remainingDestCapacity = destCapacity;
//     int32_t labelLen = 0, labelReqLength = 0;
//     UBool done = FALSE;

//     for(;;){

//         labelLen = getNextSeparator(labelStart,remainingLen, &delimiter,&done);

//         // The RFC states that
//         // <quote>
//         // ToUnicode never fails. If any step fails, then the original input
//         // is returned immediately in that step.
//         // </quote>
//         // _internal_toUnicode will copy the label.
//         /*if(labelLen==0 && done==FALSE){
//             *status = U_IDNA_ZERO_LENGTH_LABEL_ERROR;
//             break;
//         }*/

//         labelReqLength = _internal_toUnicode(labelStart, labelLen,
//                                              currentDest, remainingDestCapacity,
//                                              options, nameprep,
//                                              parseError, status);

//         if(*status == U_BUFFER_OVERFLOW_ERROR){
//             *status = U_ZERO_ERROR; // reset error
//             remainingDestCapacity = 0;
//         }

//         if(U_FAILURE(*status)){
//             break;
//         }

//         reqLength +=labelReqLength;
//         // adjust the destination pointer
//         if(labelReqLength < remainingDestCapacity){
//             currentDest = currentDest + labelReqLength;
//             remainingDestCapacity -= labelReqLength;
//         }else{
//             // should never occur
//             remainingDestCapacity = 0;
//         }

//         if(done == TRUE){
//             break;
//         }

//         // add the label separator
//         // Unlike the ToASCII operation we don't normalize the label separators
//         if(remainingDestCapacity > 0){
//             *currentDest++ = *(labelStart + labelLen);
//             remainingDestCapacity--;
//         }
//         reqLength++;

//         labelStart = delimiter;
//         if(remainingLen >0 ){
//             remainingLen = (int32_t)(srcLength - (delimiter - src));
//         }

//     }

//     if(reqLength > MAX_DOMAIN_NAME_LENGTH){
//         *status = U_IDNA_DOMAIN_NAME_TOO_LONG_ERROR;
//     }

//     usprep_close(nameprep);

//     return u_terminateUChars(dest, destCapacity, reqLength, status);
// }

// U_CAPI int32_t U_EXPORT2
// uidna_compare(  const UChar *s1, int32_t length1,
//                 const UChar *s2, int32_t length2,
//                 int32_t options,
//                 UErrorCode* status){

//     if(status == NULL || U_FAILURE(*status)){
//         return -1;
//     }

//     UChar b1Stack[MAX_IDN_BUFFER_SIZE], b2Stack[MAX_IDN_BUFFER_SIZE];
//     UChar *b1 = b1Stack, *b2 = b2Stack;
//     int32_t b1Len, b2Len, b1Capacity = MAX_IDN_BUFFER_SIZE, b2Capacity = MAX_IDN_BUFFER_SIZE;
//     int32_t result=-1;

//     UParseError parseError;

//     b1Len = uidna_IDNToASCII(s1, length1, b1, b1Capacity, options, &parseError, status);
//     if(*status == U_BUFFER_OVERFLOW_ERROR){
//         // redo processing of string
//         b1 = (UChar*) uprv_malloc(b1Len * U_SIZEOF_UCHAR);
//         if(b1==NULL){
//             *status = U_MEMORY_ALLOCATION_ERROR;
//             goto CLEANUP;
//         }

//         *status = U_ZERO_ERROR; // reset error

//         b1Len = uidna_IDNToASCII(s1,length1,b1,b1Len, options, &parseError, status);

//     }

//     b2Len = uidna_IDNToASCII(s2,length2, b2,b2Capacity, options, &parseError, status);
//     if(*status == U_BUFFER_OVERFLOW_ERROR){
//         // redo processing of string
//         b2 = (UChar*) uprv_malloc(b2Len * U_SIZEOF_UCHAR);
//         if(b2==NULL){
//             *status = U_MEMORY_ALLOCATION_ERROR;
//             goto CLEANUP;
//         }

//         *status = U_ZERO_ERROR; // reset error

//         b2Len = uidna_IDNToASCII(s2, length2, b2, b2Len, options, &parseError, status);

//     }
//     // when toASCII is applied all label separators are replaced with FULL_STOP
//     result = compareCaseInsensitiveASCII(b1,b1Len,b2,b2Len);

// CLEANUP:
//     if(b1 != b1Stack){
//         uprv_free(b1);
//     }

//     if(b2 != b2Stack){
//         uprv_free(b2);
//     }

//     return result;
// }

// #endif /* #if !UCONFIG_NO_IDNA */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2001-2015, International Business Machines
*                Corporation and others. All Rights Reserved.
******************************************************************************
*   file name:  uinit.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001July05
*   created by: George Rhoten
*/







// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
//  **********************************************************************
//  *   Copyright (C) 1999-2006, International Business Machines
//  *   Corporation and others.  All Rights Reserved.
//  **********************************************************************
//  *
//  *
//  *  ucnv_io.h:
//  *  defines  variables and functions pertaining to converter name resolution
//  *  aspect of the conversion code
//  */

// #ifndef UCNV_IO_H
// #define UCNV_IO_H

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "udataswp.h"

// #define UCNV_AMBIGUOUS_ALIAS_MAP_BIT 0x8000
// #define UCNV_CONTAINS_OPTION_BIT 0x4000
// #define UCNV_CONVERTER_INDEX_MASK 0xFFF
// #define UCNV_NUM_RESERVED_TAGS 2
// #define UCNV_NUM_HIDDEN_TAGS 1

// enum {
//     UCNV_IO_UNNORMALIZED,
//     UCNV_IO_STD_NORMALIZED,
//     UCNV_IO_NORM_TYPE_COUNT
// };

// typedef struct {
//     uint16_t stringNormalizationType;
//     uint16_t containsCnvOptionInfo;
// } UConverterAliasOptions;

// typedef struct UConverterAlias {
//     const uint16_t *converterList;
//     const uint16_t *tagList;
//     const uint16_t *aliasList;
//     const uint16_t *untaggedConvArray;
//     const uint16_t *taggedAliasArray;
//     const uint16_t *taggedAliasLists;
//     const UConverterAliasOptions *optionTable;
//     const uint16_t *stringTable;
//     const uint16_t *normalizedStringTable;

//     uint32_t converterListSize;
//     uint32_t tagListSize;
//     uint32_t aliasListSize;
//     uint32_t untaggedConvArraySize;
//     uint32_t taggedAliasArraySize;
//     uint32_t taggedAliasListsSize;
//     uint32_t optionTableSize;
//     uint32_t stringTableSize;
//     uint32_t normalizedStringTableSize;
// } UConverterAlias;

// /**
//  * \var ucnv_io_stripForCompare
//  * Remove the underscores, dashes and spaces from the name, and convert
//  * the name to lower case.
//  * @param dst The destination buffer, which is <= the buffer of name.
//  * @param dst The destination buffer, which is <= the buffer of name.
//  * @see ucnv_compareNames
//  * @return the destination buffer.
//  */
// #if U_CHARSET_FAMILY==U_ASCII_FAMILY
// #   define ucnv_io_stripForCompare ucnv_io_stripASCIIForCompare
// #elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
// #   define ucnv_io_stripForCompare ucnv_io_stripEBCDICForCompare
// #else
// #   error U_CHARSET_FAMILY is not valid
// #endif

// U_CAPI char * U_CALLCONV
// ucnv_io_stripASCIIForCompare(char *dst, const char *name);

// U_CAPI char * U_CALLCONV
// ucnv_io_stripEBCDICForCompare(char *dst, const char *name);

// /**
//  * Map a converter alias name to a canonical converter name.
//  * The alias is searched for case-insensitively, the converter name
//  * is returned in mixed-case.
//  * Returns NULL if the alias is not found.
//  * @param alias The alias name to be searched.
//  * @param containsOption A return value stating whether the returned converter name contains an option (a comma)
//  * @param pErrorCode The error code
//  * @return the converter name in mixed-case, return NULL if the alias is not found.
//  */
// U_CAPI const char *
// ucnv_io_getConverterName(const char *alias, UBool *containsOption, UErrorCode *pErrorCode);

// /**
//  * Return the number of all known converter names (no aliases).
//  * @param pErrorCode The error code
//  * @return the number of all aliases
//  */
// U_CAPI uint16_t
// ucnv_io_countKnownConverters(UErrorCode *pErrorCode);

// /**
//  * Swap an ICU converter alias table. See implementation for details.
//  * @internal
//  */
// U_CAPI int32_t U_EXPORT2
// ucnv_swapAliases(const UDataSwapper *ds,
//                  const void *inData, int32_t length, void *outData,
//                  UErrorCode *pErrorCode);

// #endif

// #endif /* _UCNV_IO */

// /*
//  * Hey, Emacs, please set the following:
//  *
//  * Local Variables:
//  * indent-tabs-mode: nil
//  * End:
//  *
//  */




U_NAMESPACE_BEGIN

static UInitOnce gICUInitOnce = U_INITONCE_INITIALIZER;

static UBool U_CALLCONV uinit_cleanup() {
    gICUInitOnce.reset();
    return TRUE;
}

static void U_CALLCONV
initData(UErrorCode &status)
{
#if UCONFIG_ENABLE_PLUGINS
    /* initialize plugins */
    uplug_init(&status);
#endif

#if !UCONFIG_NO_CONVERSION
    /*
     * 2005-may-02
     *
     * ICU4C 3.4 (jitterbug 4497) hardcodes the data for Unicode character
     * properties for APIs that want to be fast.
     * Therefore, we need not load them here nor check for errors.
     * Instead, we load the converter alias table to see if any ICU data
     * is available.
     * Users should really open the service objects they need and check
     * for errors there, to make sure that the actual items they need are
     * available.
     */
    ucnv_io_countKnownConverters(&status);
#endif
    ucln_common_registerCleanup(UCLN_COMMON_UINIT, uinit_cleanup);
}

U_NAMESPACE_END

U_NAMESPACE_USE

/*
 * ICU Initialization Function. Need not be called.
 */
U_CAPI void U_EXPORT2
u_init(UErrorCode *status) {
    UTRACE_ENTRY_OC(UTRACE_U_INIT);
    umtx_initOnce(gICUInitOnce, &initData, *status);
    UTRACE_EXIT_STATUS(*status);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1999-2010, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uinvchar.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:2
*
*   created on: 2004sep14
*   created by: Markus W. Scherer
*
*   Functions for handling invariant characters, moved here from putil.c
*   for better modularization.
*/









/* invariant-character handling --------------------------------------------- */

/*
 * These maps for ASCII to/from EBCDIC map invariant characters (see utypes.h)
 * appropriately for most EBCDIC codepages.
 *
 * They currently also map most other ASCII graphic characters,
 * appropriately for codepages 37 and 1047.
 * Exceptions: The characters for []^ have different codes in 37 & 1047.
 * Both versions are mapped to ASCII.
 *
 *    ASCII 37 1047
 * [     5B BA   AD
 * ]     5D BB   BD
 * ^     5E B0   5F
 *
 * There are no mappings for variant characters from Unicode to EBCDIC.
 *
 * Currently, C0 control codes are also included in these maps.
 * Exceptions: S/390 Open Edition swaps LF and NEL codes compared with other
 * EBCDIC platforms; both codes (15 and 25) are mapped to ASCII LF (0A),
 * but there is no mapping for ASCII LF back to EBCDIC.
 *
 *    ASCII EBCDIC S/390-OE
 * LF    0A     25       15
 * NEL   85     15       25
 *
 * The maps below explicitly exclude the variant
 * control and graphical characters that are in ASCII-based
 * codepages at 0x80 and above.
 * "No mapping" is expressed by mapping to a 00 byte.
 *
 * These tables do not establish a converter or a codepage.
 */

static const uint8_t asciiFromEbcdic[256]={
    0x00, 0x01, 0x02, 0x03, 0x00, 0x09, 0x00, 0x7f, 0x00, 0x00, 0x00, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
    0x10, 0x11, 0x12, 0x13, 0x00, 0x0a, 0x08, 0x00, 0x18, 0x19, 0x00, 0x00, 0x1c, 0x1d, 0x1e, 0x1f,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x17, 0x1b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x06, 0x07,
    0x00, 0x00, 0x16, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x14, 0x15, 0x00, 0x1a,

    0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x2e, 0x3c, 0x28, 0x2b, 0x7c,
    0x26, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x21, 0x24, 0x2a, 0x29, 0x3b, 0x5e,
    0x2d, 0x2f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x2c, 0x25, 0x5f, 0x3e, 0x3f,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x60, 0x3a, 0x23, 0x40, 0x27, 0x3d, 0x22,

    0x00, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x7e, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x00, 0x00, 0x00, 0x5b, 0x00, 0x00,
    0x5e, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x5b, 0x5d, 0x00, 0x5d, 0x00, 0x00,

    0x7b, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x7d, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 0x52, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x5c, 0x00, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static const uint8_t ebcdicFromAscii[256]={
    0x00, 0x01, 0x02, 0x03, 0x37, 0x2d, 0x2e, 0x2f, 0x16, 0x05, 0x00, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
    0x10, 0x11, 0x12, 0x13, 0x3c, 0x3d, 0x32, 0x26, 0x18, 0x19, 0x3f, 0x27, 0x1c, 0x1d, 0x1e, 0x1f,
    0x40, 0x00, 0x7f, 0x00, 0x00, 0x6c, 0x50, 0x7d, 0x4d, 0x5d, 0x5c, 0x4e, 0x6b, 0x60, 0x4b, 0x61,
    0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0x7a, 0x5e, 0x4c, 0x7e, 0x6e, 0x6f,

    0x00, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6,
    0xd7, 0xd8, 0xd9, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0x00, 0x00, 0x00, 0x00, 0x6d,
    0x00, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96,
    0x97, 0x98, 0x99, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0x00, 0x00, 0x00, 0x00, 0x07,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

/* Same as asciiFromEbcdic[] except maps all letters to lowercase. */
static const uint8_t lowercaseAsciiFromEbcdic[256]={
    0x00, 0x01, 0x02, 0x03, 0x00, 0x09, 0x00, 0x7f, 0x00, 0x00, 0x00, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
    0x10, 0x11, 0x12, 0x13, 0x00, 0x0a, 0x08, 0x00, 0x18, 0x19, 0x00, 0x00, 0x1c, 0x1d, 0x1e, 0x1f,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x17, 0x1b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x05, 0x06, 0x07,
    0x00, 0x00, 0x16, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x14, 0x15, 0x00, 0x1a,

    0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x2e, 0x3c, 0x28, 0x2b, 0x7c,
    0x26, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x21, 0x24, 0x2a, 0x29, 0x3b, 0x5e,
    0x2d, 0x2f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x2c, 0x25, 0x5f, 0x3e, 0x3f,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x60, 0x3a, 0x23, 0x40, 0x27, 0x3d, 0x22,

    0x00, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x7e, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x00, 0x00, 0x00, 0x5b, 0x00, 0x00,
    0x5e, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x5b, 0x5d, 0x00, 0x5d, 0x00, 0x00,

    0x7b, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x7d, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x7c, 0x00, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

/*
 * Bit sets indicating which characters of the ASCII repertoire
 * (by ASCII/Unicode code) are "invariant".
 * See utypes.h for more details.
 *
 * As invariant are considered the characters of the ASCII repertoire except
 * for the following:
 * 21  '!' <exclamation mark>
 * 23  '#' <number sign>
 * 24  '$' <dollar sign>
 *
 * 40  '@' <commercial at>
 *
 * 5b  '[' <left bracket>
 * 5c  '\' <backslash>
 * 5d  ']' <right bracket>
 * 5e  '^' <circumflex>
 *
 * 60  '`' <grave accent>
 *
 * 7b  '{' <left brace>
 * 7c  '|' <vertical line>
 * 7d  '}' <right brace>
 * 7e  '~' <tilde>
 */
static const uint32_t invariantChars[4]={
    0xfffffbff, /* 00..1f but not 0a */
    0xffffffe5, /* 20..3f but not 21 23 24 */
    0x87fffffe, /* 40..5f but not 40 5b..5e */
    0x87fffffe  /* 60..7f but not 60 7b..7e */
};

/*
 * test unsigned types (or values known to be non-negative) for invariant characters,
 * tests ASCII-family character values
 */
#define UCHAR_IS_INVARIANT(c) (((c)<=0x7f) && (invariantChars[(c)>>5]&((uint32_t)1<<((c)&0x1f)))!=0)

/* test signed types for invariant characters, adds test for positive values */
#define SCHAR_IS_INVARIANT(c) ((0<=(c)) && UCHAR_IS_INVARIANT(c))

#if U_CHARSET_FAMILY==U_ASCII_FAMILY
#define CHAR_TO_UCHAR(c) c
#define UCHAR_TO_CHAR(c) c
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
#define CHAR_TO_UCHAR(u) asciiFromEbcdic[u]
#define UCHAR_TO_CHAR(u) ebcdicFromAscii[u]
#else
#   error U_CHARSET_FAMILY is not valid
#endif


U_CAPI void U_EXPORT2
u_charsToUChars(const char *cs, UChar *us, int32_t length) {
    UChar u;
    uint8_t c;

    /*
     * Allow the entire ASCII repertoire to be mapped _to_ Unicode.
     * For EBCDIC systems, this works for characters with codes from
     * codepages 37 and 1047 or compatible.
     */
    while(length>0) {
        c=(uint8_t)(*cs++);
        u=(UChar)CHAR_TO_UCHAR(c);
        U_ASSERT((u!=0 || c==0)); /* only invariant chars converted? */
        *us++=u;
        --length;
    }
}

U_CAPI void U_EXPORT2
u_UCharsToChars(const UChar *us, char *cs, int32_t length) {
    UChar u;

    while(length>0) {
        u=*us++;
        if(!UCHAR_IS_INVARIANT(u)) {
            U_ASSERT(FALSE); /* Variant characters were used. These are not portable in ICU. */
            u=0;
        }
        *cs++=(char)UCHAR_TO_CHAR(u);
        --length;
    }
}

U_CAPI UBool U_EXPORT2
uprv_isInvariantString(const char *s, int32_t length) {
    uint8_t c;

    for(;;) {
        if(length<0) {
            /* NUL-terminated */
            c=(uint8_t)*s++;
            if(c==0) {
                break;
            }
        } else {
            /* count length */
            if(length==0) {
                break;
            }
            --length;
            c=(uint8_t)*s++;
            if(c==0) {
                continue; /* NUL is invariant */
            }
        }
        /* c!=0 now, one branch below checks c==0 for variant characters */

        /*
         * no assertions here because these functions are legitimately called
         * for strings with variant characters
         */
#if U_CHARSET_FAMILY==U_ASCII_FAMILY
        if(!UCHAR_IS_INVARIANT(c)) {
            return FALSE; /* found a variant char */
        }
#elif U_CHARSET_FAMILY==U_EBCDIC_FAMILY
        c=CHAR_TO_UCHAR(c);
        if(c==0 || !UCHAR_IS_INVARIANT(c)) {
            return FALSE; /* found a variant char */
        }
#else
#   error U_CHARSET_FAMILY is not valid
#endif
    }
    return TRUE;
}

U_CAPI UBool U_EXPORT2
uprv_isInvariantUString(const UChar *s, int32_t length) {
    UChar c;

    for(;;) {
        if(length<0) {
            /* NUL-terminated */
            c=*s++;
            if(c==0) {
                break;
            }
        } else {
            /* count length */
            if(length==0) {
                break;
            }
            --length;
            c=*s++;
        }

        /*
         * no assertions here because these functions are legitimately called
         * for strings with variant characters
         */
        if(!UCHAR_IS_INVARIANT(c)) {
            return FALSE; /* found a variant char */
        }
    }
    return TRUE;
}

/* UDataSwapFn implementations used in udataswp.c ------- */

/* convert ASCII to EBCDIC and verify that all characters are invariant */
U_CAPI int32_t U_EXPORT2
uprv_ebcdicFromAscii(const UDataSwapper *ds,
                     const void *inData, int32_t length, void *outData,
                     UErrorCode *pErrorCode) {
    const uint8_t *s;
    uint8_t *t;
    uint8_t c;

    int32_t count;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 || (length>0 && outData==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* setup and swapping */
    s=(const uint8_t *)inData;
    t=(uint8_t *)outData;
    count=length;
    while(count>0) {
        c=*s++;
        if(!UCHAR_IS_INVARIANT(c)) {
            udata_printError(ds, "uprv_ebcdicFromAscii() string[%d] contains a variant character in position %d\n",
                             length, length-count);
            *pErrorCode=U_INVALID_CHAR_FOUND;
            return 0;
        }
        *t++=ebcdicFromAscii[c];
        --count;
    }

    return length;
}

/* this function only checks and copies ASCII strings without conversion */
U_CFUNC int32_t
uprv_copyAscii(const UDataSwapper *ds,
               const void *inData, int32_t length, void *outData,
               UErrorCode *pErrorCode) {
    const uint8_t *s;
    uint8_t c;

    int32_t count;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 || (length>0 && outData==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* setup and checking */
    s=(const uint8_t *)inData;
    count=length;
    while(count>0) {
        c=*s++;
        if(!UCHAR_IS_INVARIANT(c)) {
            udata_printError(ds, "uprv_copyFromAscii() string[%d] contains a variant character in position %d\n",
                             length, length-count);
            *pErrorCode=U_INVALID_CHAR_FOUND;
            return 0;
        }
        --count;
    }

    if(length>0 && inData!=outData) {
        uprv_memcpy(outData, inData, length);
    }

    return length;
}

/* convert EBCDIC to ASCII and verify that all characters are invariant */
U_CFUNC int32_t
uprv_asciiFromEbcdic(const UDataSwapper *ds,
                     const void *inData, int32_t length, void *outData,
                     UErrorCode *pErrorCode) {
    const uint8_t *s;
    uint8_t *t;
    uint8_t c;

    int32_t count;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 ||  (length>0 && outData==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* setup and swapping */
    s=(const uint8_t *)inData;
    t=(uint8_t *)outData;
    count=length;
    while(count>0) {
        c=*s++;
        if(c!=0 && ((c=asciiFromEbcdic[c])==0 || !UCHAR_IS_INVARIANT(c))) {
            udata_printError(ds, "uprv_asciiFromEbcdic() string[%d] contains a variant character in position %d\n",
                             length, length-count);
            *pErrorCode=U_INVALID_CHAR_FOUND;
            return 0;
        }
        *t++=c;
        --count;
    }

    return length;
}

/* this function only checks and copies EBCDIC strings without conversion */
U_CFUNC int32_t
uprv_copyEbcdic(const UDataSwapper *ds,
                const void *inData, int32_t length, void *outData,
                UErrorCode *pErrorCode) {
    const uint8_t *s;
    uint8_t c;

    int32_t count;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || length<0 || (length>0 && outData==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* setup and checking */
    s=(const uint8_t *)inData;
    count=length;
    while(count>0) {
        c=*s++;
        if(c!=0 && ((c=asciiFromEbcdic[c])==0 || !UCHAR_IS_INVARIANT(c))) {
            udata_printError(ds, "uprv_copyEbcdic() string[%] contains a variant character in position %d\n",
                             length, length-count);
            *pErrorCode=U_INVALID_CHAR_FOUND;
            return 0;
        }
        --count;
    }

    if(length>0 && inData!=outData) {
        uprv_memcpy(outData, inData, length);
    }

    return length;
}

U_CFUNC UBool
uprv_isEbcdicAtSign(char c) {
    static const uint8_t ebcdicAtSigns[] = {
        0x7C, 0x44, 0x66, 0x80, 0xAC, 0xAE, 0xAF, 0xB5, 0xEC, 0xEF, 0x00 };
    return c != 0 && uprv_strchr((const char *)ebcdicAtSigns, c) != nullptr;
}

/* compare invariant strings; variant characters compare less than others and unlike each other */
U_CFUNC int32_t
uprv_compareInvAscii(const UDataSwapper *ds,
                     const char *outString, int32_t outLength,
                     const UChar *localString, int32_t localLength) {
    (void)ds;
    int32_t minLength;
    UChar32 c1, c2;
    uint8_t c;

    if(outString==NULL || outLength<-1 || localString==NULL || localLength<-1) {
        return 0;
    }

    if(outLength<0) {
        outLength=(int32_t)uprv_strlen(outString);
    }
    if(localLength<0) {
        localLength=u_strlen(localString);
    }

    minLength= outLength<localLength ? outLength : localLength;

    while(minLength>0) {
        c=(uint8_t)*outString++;
        if(UCHAR_IS_INVARIANT(c)) {
            c1=c;
        } else {
            c1=-1;
        }

        c2=*localString++;
        if(!UCHAR_IS_INVARIANT(c2)) {
            c2=-2;
        }

        if((c1-=c2)!=0) {
            return c1;
        }

        --minLength;
    }

    /* strings start with same prefix, compare lengths */
    return outLength-localLength;
}

U_CFUNC int32_t
uprv_compareInvEbcdic(const UDataSwapper *ds,
                      const char *outString, int32_t outLength,
                      const UChar *localString, int32_t localLength) {
    (void)ds;
    int32_t minLength;
    UChar32 c1, c2;
    uint8_t c;

    if(outString==NULL || outLength<-1 || localString==NULL || localLength<-1) {
        return 0;
    }

    if(outLength<0) {
        outLength=(int32_t)uprv_strlen(outString);
    }
    if(localLength<0) {
        localLength=u_strlen(localString);
    }

    minLength= outLength<localLength ? outLength : localLength;

    while(minLength>0) {
        c=(uint8_t)*outString++;
        if(c==0) {
            c1=0;
        } else if((c1=asciiFromEbcdic[c])!=0 && UCHAR_IS_INVARIANT(c1)) {
            /* c1 is set */
        } else {
            c1=-1;
        }

        c2=*localString++;
        if(!UCHAR_IS_INVARIANT(c2)) {
            c2=-2;
        }

        if((c1-=c2)!=0) {
            return c1;
        }

        --minLength;
    }

    /* strings start with same prefix, compare lengths */
    return outLength-localLength;
}

U_CAPI int32_t U_EXPORT2
uprv_compareInvEbcdicAsAscii(const char *s1, const char *s2) {
    int32_t c1, c2;

    for(;; ++s1, ++s2) {
        c1=(uint8_t)*s1;
        c2=(uint8_t)*s2;
        if(c1!=c2) {
            if(c1!=0 && ((c1=asciiFromEbcdic[c1])==0 || !UCHAR_IS_INVARIANT(c1))) {
                c1=-(int32_t)(uint8_t)*s1;
            }
            if(c2!=0 && ((c2=asciiFromEbcdic[c2])==0 || !UCHAR_IS_INVARIANT(c2))) {
                c2=-(int32_t)(uint8_t)*s2;
            }
            return c1-c2;
        } else if(c1==0) {
            return 0;
        }
    }
}

U_CAPI char U_EXPORT2
uprv_ebcdicToAscii(char c) {
    return (char)asciiFromEbcdic[(uint8_t)c];
}

U_CAPI char U_EXPORT2
uprv_ebcdicToLowercaseAscii(char c) {
    return (char)lowercaseAsciiFromEbcdic[(uint8_t)c];
}

U_INTERNAL uint8_t* U_EXPORT2
uprv_aestrncpy(uint8_t *dst, const uint8_t *src, int32_t n)
{
  uint8_t *orig_dst = dst;

  if(n==-1) { 
    n = static_cast<int32_t>(uprv_strlen((const char*)src)+1); /* copy NUL */
  }
  /* copy non-null */
  while(*src && n>0) {
    *(dst++) = asciiFromEbcdic[*(src++)];
    n--;
  }
  /* pad */
  while(n>0) {
    *(dst++) = 0;
    n--;
  }
  return orig_dst;
}

U_INTERNAL uint8_t* U_EXPORT2
uprv_eastrncpy(uint8_t *dst, const uint8_t *src, int32_t n)
{
  uint8_t *orig_dst = dst;

  if(n==-1) { 
    n = static_cast<int32_t>(uprv_strlen((const char*)src)+1); /* copy NUL */
  }
  /* copy non-null */
  while(*src && n>0) {
    char ch = ebcdicFromAscii[*(src++)];
    if(ch == 0) {
      ch = ebcdicFromAscii[0x3f]; /* questionmark (subchar) */
    }
    *(dst++) = ch;
    n--;
  }
  /* pad */
  while(n>0) {
    *(dst++) = 0;
    n--;
  }
  return orig_dst;
}

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2002-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uiter.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2002jan18
*   created by: Markus W. Scherer
*/











U_NAMESPACE_USE

#define IS_EVEN(n) (((n)&1)==0)
#define IS_POINTER_EVEN(p) IS_EVEN((size_t)p)

U_CDECL_BEGIN

/* No-Op UCharIterator implementation for illegal input --------------------- */

static int32_t U_CALLCONV
noopGetIndex(UCharIterator * /*iter*/, UCharIteratorOrigin /*origin*/) {
    return 0;
}

static int32_t U_CALLCONV
noopMove(UCharIterator * /*iter*/, int32_t /*delta*/, UCharIteratorOrigin /*origin*/) {
    return 0;
}

static UBool U_CALLCONV
noopHasNext(UCharIterator * /*iter*/) {
    return FALSE;
}

static UChar32 U_CALLCONV
noopCurrent(UCharIterator * /*iter*/) {
    return U_SENTINEL;
}

static uint32_t U_CALLCONV
noopGetState(const UCharIterator * /*iter*/) {
    return UITER_NO_STATE;
}

static void U_CALLCONV
noopSetState(UCharIterator * /*iter*/, uint32_t /*state*/, UErrorCode *pErrorCode) {
    *pErrorCode=U_UNSUPPORTED_ERROR;
}

static const UCharIterator noopIterator={
    0, 0, 0, 0, 0, 0,
    noopGetIndex,
    noopMove,
    noopHasNext,
    noopHasNext,
    noopCurrent,
    noopCurrent,
    noopCurrent,
    NULL,
    noopGetState,
    noopSetState
};

/* UCharIterator implementation for simple strings -------------------------- */

/*
 * This is an implementation of a code unit (UChar) iterator
 * for UChar * strings.
 *
 * The UCharIterator.context field holds a pointer to the string.
 */

static int32_t U_CALLCONV
stringIteratorGetIndex(UCharIterator *iter, UCharIteratorOrigin origin) {
    switch(origin) {
    case UITER_ZERO:
        return 0;
    case UITER_START:
        return iter->start;
    case UITER_CURRENT:
        return iter->index;
    case UITER_LIMIT:
        return iter->limit;
    case UITER_LENGTH:
        return iter->length;
    default:
        /* not a valid origin */
        /* Should never get here! */
        return -1;
    }
}

static int32_t U_CALLCONV
stringIteratorMove(UCharIterator *iter, int32_t delta, UCharIteratorOrigin origin) {
    int32_t pos;

    switch(origin) {
    case UITER_ZERO:
        pos=delta;
        break;
    case UITER_START:
        pos=iter->start+delta;
        break;
    case UITER_CURRENT:
        pos=iter->index+delta;
        break;
    case UITER_LIMIT:
        pos=iter->limit+delta;
        break;
    case UITER_LENGTH:
        pos=iter->length+delta;
        break;
    default:
        return -1;  /* Error */
    }

    if(pos<iter->start) {
        pos=iter->start;
    } else if(pos>iter->limit) {
        pos=iter->limit;
    }

    return iter->index=pos;
}

static UBool U_CALLCONV
stringIteratorHasNext(UCharIterator *iter) {
    return iter->index<iter->limit;
}

static UBool U_CALLCONV
stringIteratorHasPrevious(UCharIterator *iter) {
    return iter->index>iter->start;
}

static UChar32 U_CALLCONV
stringIteratorCurrent(UCharIterator *iter) {
    if(iter->index<iter->limit) {
        return ((const UChar *)(iter->context))[iter->index];
    } else {
        return U_SENTINEL;
    }
}

static UChar32 U_CALLCONV
stringIteratorNext(UCharIterator *iter) {
    if(iter->index<iter->limit) {
        return ((const UChar *)(iter->context))[iter->index++];
    } else {
        return U_SENTINEL;
    }
}

static UChar32 U_CALLCONV
stringIteratorPrevious(UCharIterator *iter) {
    if(iter->index>iter->start) {
        return ((const UChar *)(iter->context))[--iter->index];
    } else {
        return U_SENTINEL;
    }
}

static uint32_t U_CALLCONV
stringIteratorGetState(const UCharIterator *iter) {
    return (uint32_t)iter->index;
}

static void U_CALLCONV
stringIteratorSetState(UCharIterator *iter, uint32_t state, UErrorCode *pErrorCode) {
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        /* do nothing */
    } else if(iter==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
    } else if((int32_t)state<iter->start || iter->limit<(int32_t)state) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
    } else {
        iter->index=(int32_t)state;
    }
}

static const UCharIterator stringIterator={
    0, 0, 0, 0, 0, 0,
    stringIteratorGetIndex,
    stringIteratorMove,
    stringIteratorHasNext,
    stringIteratorHasPrevious,
    stringIteratorCurrent,
    stringIteratorNext,
    stringIteratorPrevious,
    NULL,
    stringIteratorGetState,
    stringIteratorSetState
};

U_CAPI void U_EXPORT2
uiter_setString(UCharIterator *iter, const UChar *s, int32_t length) {
    if(iter!=0) {
        if(s!=0 && length>=-1) {
            *iter=stringIterator;
            iter->context=s;
            if(length>=0) {
                iter->length=length;
            } else {
                iter->length=u_strlen(s);
            }
            iter->limit=iter->length;
        } else {
            *iter=noopIterator;
        }
    }
}

/* UCharIterator implementation for UTF-16BE strings ------------------------ */

/*
 * This is an implementation of a code unit (UChar) iterator
 * for UTF-16BE strings, i.e., strings in byte-vectors where
 * each UChar is stored as a big-endian pair of bytes.
 *
 * The UCharIterator.context field holds a pointer to the string.
 * Everything works just like with a normal UChar iterator (uiter_setString),
 * except that UChars are assembled from byte pairs.
 */

/* internal helper function */
static inline UChar32
utf16BEIteratorGet(UCharIterator *iter, int32_t index) {
    const uint8_t *p=(const uint8_t *)iter->context;
    return ((UChar)p[2*index]<<8)|(UChar)p[2*index+1];
}

static UChar32 U_CALLCONV
utf16BEIteratorCurrent(UCharIterator *iter) {
    int32_t index;

    if((index=iter->index)<iter->limit) {
        return utf16BEIteratorGet(iter, index);
    } else {
        return U_SENTINEL;
    }
}

static UChar32 U_CALLCONV
utf16BEIteratorNext(UCharIterator *iter) {
    int32_t index;

    if((index=iter->index)<iter->limit) {
        iter->index=index+1;
        return utf16BEIteratorGet(iter, index);
    } else {
        return U_SENTINEL;
    }
}

static UChar32 U_CALLCONV
utf16BEIteratorPrevious(UCharIterator *iter) {
    int32_t index;

    if((index=iter->index)>iter->start) {
        iter->index=--index;
        return utf16BEIteratorGet(iter, index);
    } else {
        return U_SENTINEL;
    }
}

static const UCharIterator utf16BEIterator={
    0, 0, 0, 0, 0, 0,
    stringIteratorGetIndex,
    stringIteratorMove,
    stringIteratorHasNext,
    stringIteratorHasPrevious,
    utf16BEIteratorCurrent,
    utf16BEIteratorNext,
    utf16BEIteratorPrevious,
    NULL,
    stringIteratorGetState,
    stringIteratorSetState
};

/*
 * Count the number of UChars in a UTF-16BE string before a terminating UChar NUL,
 * i.e., before a pair of 0 bytes where the first 0 byte is at an even
 * offset from s.
 */
static int32_t
utf16BE_strlen(const char *s) {
    if(IS_POINTER_EVEN(s)) {
        /*
         * even-aligned, call u_strlen(s)
         * we are probably on a little-endian machine, but searching for UChar NUL
         * does not care about endianness
         */
        return u_strlen((const UChar *)s);
    } else {
        /* odd-aligned, search for pair of 0 bytes */
        const char *p=s;

        while(!(*p==0 && p[1]==0)) {
            p+=2;
        }
        return (int32_t)((p-s)/2);
    }
}

U_CAPI void U_EXPORT2
uiter_setUTF16BE(UCharIterator *iter, const char *s, int32_t length) {
    if(iter!=NULL) {
        /* allow only even-length strings (the input length counts bytes) */
        if(s!=NULL && (length==-1 || (length>=0 && IS_EVEN(length)))) {
            /* length/=2, except that >>=1 also works for -1 (-1/2==0, -1>>1==-1) */
            length>>=1;

            if(U_IS_BIG_ENDIAN && IS_POINTER_EVEN(s)) {
                /* big-endian machine and 2-aligned UTF-16BE string: use normal UChar iterator */
                uiter_setString(iter, (const UChar *)s, length);
                return;
            }

            *iter=utf16BEIterator;
            iter->context=s;
            if(length>=0) {
                iter->length=length;
            } else {
                iter->length=utf16BE_strlen(s);
            }
            iter->limit=iter->length;
        } else {
            *iter=noopIterator;
        }
    }
}

/* UCharIterator wrapper around CharacterIterator --------------------------- */

/*
 * This is wrapper code around a C++ CharacterIterator to
 * look like a C UCharIterator.
 *
 * The UCharIterator.context field holds a pointer to the CharacterIterator.
 */

static int32_t U_CALLCONV
characterIteratorGetIndex(UCharIterator *iter, UCharIteratorOrigin origin) {
    switch(origin) {
    case UITER_ZERO:
        return 0;
    case UITER_START:
        return ((CharacterIterator *)(iter->context))->startIndex();
    case UITER_CURRENT:
        return ((CharacterIterator *)(iter->context))->getIndex();
    case UITER_LIMIT:
        return ((CharacterIterator *)(iter->context))->endIndex();
    case UITER_LENGTH:
        return ((CharacterIterator *)(iter->context))->getLength();
    default:
        /* not a valid origin */
        /* Should never get here! */
        return -1;
    }
}

static int32_t U_CALLCONV
characterIteratorMove(UCharIterator *iter, int32_t delta, UCharIteratorOrigin origin) {
    switch(origin) {
    case UITER_ZERO:
        ((CharacterIterator *)(iter->context))->setIndex(delta);
        return ((CharacterIterator *)(iter->context))->getIndex();
    case UITER_START:
    case UITER_CURRENT:
    case UITER_LIMIT:
        return ((CharacterIterator *)(iter->context))->move(delta, (CharacterIterator::EOrigin)origin);
    case UITER_LENGTH:
        ((CharacterIterator *)(iter->context))->setIndex(((CharacterIterator *)(iter->context))->getLength()+delta);
        return ((CharacterIterator *)(iter->context))->getIndex();
    default:
        /* not a valid origin */
        /* Should never get here! */
        return -1;
    }
}

static UBool U_CALLCONV
characterIteratorHasNext(UCharIterator *iter) {
    return ((CharacterIterator *)(iter->context))->hasNext();
}

static UBool U_CALLCONV
characterIteratorHasPrevious(UCharIterator *iter) {
    return ((CharacterIterator *)(iter->context))->hasPrevious();
}

static UChar32 U_CALLCONV
characterIteratorCurrent(UCharIterator *iter) {
    UChar32 c;

    c=((CharacterIterator *)(iter->context))->current();
    if(c!=0xffff || ((CharacterIterator *)(iter->context))->hasNext()) {
        return c;
    } else {
        return U_SENTINEL;
    }
}

static UChar32 U_CALLCONV
characterIteratorNext(UCharIterator *iter) {
    if(((CharacterIterator *)(iter->context))->hasNext()) {
        return ((CharacterIterator *)(iter->context))->nextPostInc();
    } else {
        return U_SENTINEL;
    }
}

static UChar32 U_CALLCONV
characterIteratorPrevious(UCharIterator *iter) {
    if(((CharacterIterator *)(iter->context))->hasPrevious()) {
        return ((CharacterIterator *)(iter->context))->previous();
    } else {
        return U_SENTINEL;
    }
}

static uint32_t U_CALLCONV
characterIteratorGetState(const UCharIterator *iter) {
    return ((CharacterIterator *)(iter->context))->getIndex();
}

static void U_CALLCONV
characterIteratorSetState(UCharIterator *iter, uint32_t state, UErrorCode *pErrorCode) {
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        /* do nothing */
    } else if(iter==NULL || iter->context==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
    } else if((int32_t)state<((CharacterIterator *)(iter->context))->startIndex() || ((CharacterIterator *)(iter->context))->endIndex()<(int32_t)state) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
    } else {
        ((CharacterIterator *)(iter->context))->setIndex((int32_t)state);
    }
}

static const UCharIterator characterIteratorWrapper={
    0, 0, 0, 0, 0, 0,
    characterIteratorGetIndex,
    characterIteratorMove,
    characterIteratorHasNext,
    characterIteratorHasPrevious,
    characterIteratorCurrent,
    characterIteratorNext,
    characterIteratorPrevious,
    NULL,
    characterIteratorGetState,
    characterIteratorSetState
};

U_CAPI void U_EXPORT2
uiter_setCharacterIterator(UCharIterator *iter, CharacterIterator *charIter) {
    if(iter!=0) {
        if(charIter!=0) {
            *iter=characterIteratorWrapper;
            iter->context=charIter;
        } else {
            *iter=noopIterator;
        }
    }
}

/* UCharIterator wrapper around Replaceable --------------------------------- */

/*
 * This is an implementation of a code unit (UChar) iterator
 * based on a Replaceable object.
 *
 * The UCharIterator.context field holds a pointer to the Replaceable.
 * UCharIterator.length and UCharIterator.index hold Replaceable.length()
 * and the iteration index.
 */

static UChar32 U_CALLCONV
replaceableIteratorCurrent(UCharIterator *iter) {
    if(iter->index<iter->limit) {
        return ((Replaceable *)(iter->context))->charAt(iter->index);
    } else {
        return U_SENTINEL;
    }
}

static UChar32 U_CALLCONV
replaceableIteratorNext(UCharIterator *iter) {
    if(iter->index<iter->limit) {
        return ((Replaceable *)(iter->context))->charAt(iter->index++);
    } else {
        return U_SENTINEL;
    }
}

static UChar32 U_CALLCONV
replaceableIteratorPrevious(UCharIterator *iter) {
    if(iter->index>iter->start) {
        return ((Replaceable *)(iter->context))->charAt(--iter->index);
    } else {
        return U_SENTINEL;
    }
}

static const UCharIterator replaceableIterator={
    0, 0, 0, 0, 0, 0,
    stringIteratorGetIndex,
    stringIteratorMove,
    stringIteratorHasNext,
    stringIteratorHasPrevious,
    replaceableIteratorCurrent,
    replaceableIteratorNext,
    replaceableIteratorPrevious,
    NULL,
    stringIteratorGetState,
    stringIteratorSetState
};

U_CAPI void U_EXPORT2
uiter_setReplaceable(UCharIterator *iter, const Replaceable *rep) {
    if(iter!=0) {
        if(rep!=0) {
            *iter=replaceableIterator;
            iter->context=rep;
            iter->limit=iter->length=rep->length();
        } else {
            *iter=noopIterator;
        }
    }
}

/* UCharIterator implementation for UTF-8 strings --------------------------- */

/*
 * Possible, probably necessary only for an implementation for arbitrary
 * converters:
 * Maintain a buffer (ring buffer?) for a piece of converted 16-bit text.
 * This would require to turn reservedFn into a close function and
 * to introduce a uiter_close(iter).
 */

#define UITER_CNV_CAPACITY 16

/*
 * Minimal implementation:
 * Maintain a single-UChar buffer for an additional surrogate.
 * The caller must not modify start and limit because they are used internally.
 *
 * Use UCharIterator fields as follows:
 *   context        pointer to UTF-8 string
 *   length         UTF-16 length of the string; -1 until lazy evaluation
 *   start          current UTF-8 index
 *   index          current UTF-16 index; may be -1="unknown" after setState()
 *   limit          UTF-8 length of the string
 *   reservedField  supplementary code point
 *
 * Since UCharIterator delivers 16-bit code units, the iteration can be
 * currently in the middle of the byte sequence for a supplementary code point.
 * In this case, reservedField will contain that code point and start will
 * point to after the corresponding byte sequence. The UTF-16 index will be
 * one less than what it would otherwise be corresponding to the UTF-8 index.
 * Otherwise, reservedField will be 0.
 */

/*
 * Possible optimization for NUL-terminated UTF-8 and UTF-16 strings:
 * Add implementations that do not call strlen() for iteration but check for NUL.
 */

static int32_t U_CALLCONV
utf8IteratorGetIndex(UCharIterator *iter, UCharIteratorOrigin origin) {
    switch(origin) {
    case UITER_ZERO:
    case UITER_START:
        return 0;
    case UITER_CURRENT:
        if(iter->index<0) {
            /* the current UTF-16 index is unknown after setState(), count from the beginning */
            const uint8_t *s;
            UChar32 c;
            int32_t i, limit, index;

            s=(const uint8_t *)iter->context;
            i=index=0;
            limit=iter->start; /* count up to the UTF-8 index */
            while(i<limit) {
                U8_NEXT_OR_FFFD(s, i, limit, c);
                index+=U16_LENGTH(c);
            }

            iter->start=i; /* just in case setState() did not get us to a code point boundary */
            if(i==iter->limit) {
                iter->length=index; /* in case it was <0 or wrong */
            }
            if(iter->reservedField!=0) {
                --index; /* we are in the middle of a supplementary code point */
            }
            iter->index=index;
        }
        return iter->index;
    case UITER_LIMIT:
    case UITER_LENGTH:
        if(iter->length<0) {
            const uint8_t *s;
            UChar32 c;
            int32_t i, limit, length;

            s=(const uint8_t *)iter->context;
            if(iter->index<0) {
                /*
                 * the current UTF-16 index is unknown after setState(),
                 * we must first count from the beginning to here
                 */
                i=length=0;
                limit=iter->start;

                /* count from the beginning to the current index */
                while(i<limit) {
                    U8_NEXT_OR_FFFD(s, i, limit, c);
                    length+=U16_LENGTH(c);
                }

                /* assume i==limit==iter->start, set the UTF-16 index */
                iter->start=i; /* just in case setState() did not get us to a code point boundary */
                iter->index= iter->reservedField!=0 ? length-1 : length;
            } else {
                i=iter->start;
                length=iter->index;
                if(iter->reservedField!=0) {
                    ++length;
                }
            }

            /* count from the current index to the end */
            limit=iter->limit;
            while(i<limit) {
                U8_NEXT_OR_FFFD(s, i, limit, c);
                length+=U16_LENGTH(c);
            }
            iter->length=length;
        }
        return iter->length;
    default:
        /* not a valid origin */
        /* Should never get here! */
        return -1;
    }
}

static int32_t U_CALLCONV
utf8IteratorMove(UCharIterator *iter, int32_t delta, UCharIteratorOrigin origin) {
    const uint8_t *s;
    UChar32 c;
    int32_t pos; /* requested UTF-16 index */
    int32_t i; /* UTF-8 index */
    UBool havePos;

    /* calculate the requested UTF-16 index */
    switch(origin) {
    case UITER_ZERO:
    case UITER_START:
        pos=delta;
        havePos=TRUE;
        /* iter->index<0 (unknown) is possible */
        break;
    case UITER_CURRENT:
        if(iter->index>=0) {
            pos=iter->index+delta;
            havePos=TRUE;
        } else {
            /* the current UTF-16 index is unknown after setState(), use only delta */
            pos=0;
            havePos=FALSE;
        }
        break;
    case UITER_LIMIT:
    case UITER_LENGTH:
        if(iter->length>=0) {
            pos=iter->length+delta;
            havePos=TRUE;
        } else {
            /* pin to the end, avoid counting the length */
            iter->index=-1;
            iter->start=iter->limit;
            iter->reservedField=0;
            if(delta>=0) {
                return UITER_UNKNOWN_INDEX;
            } else {
                /* the current UTF-16 index is unknown, use only delta */
                pos=0;
                havePos=FALSE;
            }
        }
        break;
    default:
        return -1;  /* Error */
    }

    if(havePos) {
        /* shortcuts: pinning to the edges of the string */
        if(pos<=0) {
            iter->index=iter->start=iter->reservedField=0;
            return 0;
        } else if(iter->length>=0 && pos>=iter->length) {
            iter->index=iter->length;
            iter->start=iter->limit;
            iter->reservedField=0;
            return iter->index;
        }

        /* minimize the number of U8_NEXT/PREV operations */
        if(iter->index<0 || pos<iter->index/2) {
            /* go forward from the start instead of backward from the current index */
            iter->index=iter->start=iter->reservedField=0;
        } else if(iter->length>=0 && (iter->length-pos)<(pos-iter->index)) {
            /*
             * if we have the UTF-16 index and length and the new position is
             * closer to the end than the current index,
             * then go backward from the end instead of forward from the current index
             */
            iter->index=iter->length;
            iter->start=iter->limit;
            iter->reservedField=0;
        }

        delta=pos-iter->index;
        if(delta==0) {
            return iter->index; /* nothing to do */
        }
    } else {
        /* move relative to unknown UTF-16 index */
        if(delta==0) {
            return UITER_UNKNOWN_INDEX; /* nothing to do */
        } else if(-delta>=iter->start) {
            /* moving backwards by more UChars than there are UTF-8 bytes, pin to 0 */
            iter->index=iter->start=iter->reservedField=0;
            return 0;
        } else if(delta>=(iter->limit-iter->start)) {
            /* moving forward by more UChars than the remaining UTF-8 bytes, pin to the end */
            iter->index=iter->length; /* may or may not be <0 (unknown) */
            iter->start=iter->limit;
            iter->reservedField=0;
            return iter->index>=0 ? iter->index : (int32_t)UITER_UNKNOWN_INDEX;
        }
    }

    /* delta!=0 */

    /* move towards the requested position, pin to the edges of the string */
    s=(const uint8_t *)iter->context;
    pos=iter->index; /* could be <0 (unknown) */
    i=iter->start;
    if(delta>0) {
        /* go forward */
        int32_t limit=iter->limit;
        if(iter->reservedField!=0) {
            iter->reservedField=0;
            ++pos;
            --delta;
        }
        while(delta>0 && i<limit) {
            U8_NEXT_OR_FFFD(s, i, limit, c);
            if(c<=0xffff) {
                ++pos;
                --delta;
            } else if(delta>=2) {
                pos+=2;
                delta-=2;
            } else /* delta==1 */ {
                /* stop in the middle of a supplementary code point */
                iter->reservedField=c;
                ++pos;
                break; /* delta=0; */
            }
        }
        if(i==limit) {
            if(iter->length<0 && iter->index>=0) {
                iter->length= iter->reservedField==0 ? pos : pos+1;
            } else if(iter->index<0 && iter->length>=0) {
                iter->index= iter->reservedField==0 ? iter->length : iter->length-1;
            }
        }
    } else /* delta<0 */ {
        /* go backward */
        if(iter->reservedField!=0) {
            iter->reservedField=0;
            i-=4; /* we stayed behind the supplementary code point; go before it now */
            --pos;
            ++delta;
        }
        while(delta<0 && i>0) {
            U8_PREV_OR_FFFD(s, 0, i, c);
            if(c<=0xffff) {
                --pos;
                ++delta;
            } else if(delta<=-2) {
                pos-=2;
                delta+=2;
            } else /* delta==-1 */ {
                /* stop in the middle of a supplementary code point */
                i+=4; /* back to behind this supplementary code point for consistent state */
                iter->reservedField=c;
                --pos;
                break; /* delta=0; */
            }
        }
    }

    iter->start=i;
    if(iter->index>=0) {
        return iter->index=pos;
    } else {
        /* we started with index<0 (unknown) so pos is bogus */
        if(i<=1) {
            return iter->index=i; /* reached the beginning */
        } else {
            /* we still don't know the UTF-16 index */
            return UITER_UNKNOWN_INDEX;
        }
    }
}

static UBool U_CALLCONV
utf8IteratorHasNext(UCharIterator *iter) {
    return iter->start<iter->limit || iter->reservedField!=0;
}

static UBool U_CALLCONV
utf8IteratorHasPrevious(UCharIterator *iter) {
    return iter->start>0;
}

static UChar32 U_CALLCONV
utf8IteratorCurrent(UCharIterator *iter) {
    if(iter->reservedField!=0) {
        return U16_TRAIL(iter->reservedField);
    } else if(iter->start<iter->limit) {
        const uint8_t *s=(const uint8_t *)iter->context;
        UChar32 c;
        int32_t i=iter->start;

        U8_NEXT_OR_FFFD(s, i, iter->limit, c);
        if(c<=0xffff) {
            return c;
        } else {
            return U16_LEAD(c);
        }
    } else {
        return U_SENTINEL;
    }
}

static UChar32 U_CALLCONV
utf8IteratorNext(UCharIterator *iter) {
    int32_t index;

    if(iter->reservedField!=0) {
        UChar trail=U16_TRAIL(iter->reservedField);
        iter->reservedField=0;
        if((index=iter->index)>=0) {
            iter->index=index+1;
        }
        return trail;
    } else if(iter->start<iter->limit) {
        const uint8_t *s=(const uint8_t *)iter->context;
        UChar32 c;

        U8_NEXT_OR_FFFD(s, iter->start, iter->limit, c);
        if((index=iter->index)>=0) {
            iter->index=++index;
            if(iter->length<0 && iter->start==iter->limit) {
                iter->length= c<=0xffff ? index : index+1;
            }
        } else if(iter->start==iter->limit && iter->length>=0) {
            iter->index= c<=0xffff ? iter->length : iter->length-1;
        }
        if(c<=0xffff) {
            return c;
        } else {
            iter->reservedField=c;
            return U16_LEAD(c);
        }
    } else {
        return U_SENTINEL;
    }
}

static UChar32 U_CALLCONV
utf8IteratorPrevious(UCharIterator *iter) {
    int32_t index;

    if(iter->reservedField!=0) {
        UChar lead=U16_LEAD(iter->reservedField);
        iter->reservedField=0;
        iter->start-=4; /* we stayed behind the supplementary code point; go before it now */
        if((index=iter->index)>0) {
            iter->index=index-1;
        }
        return lead;
    } else if(iter->start>0) {
        const uint8_t *s=(const uint8_t *)iter->context;
        UChar32 c;

        U8_PREV_OR_FFFD(s, 0, iter->start, c);
        if((index=iter->index)>0) {
            iter->index=index-1;
        } else if(iter->start<=1) {
            iter->index= c<=0xffff ? iter->start : iter->start+1;
        }
        if(c<=0xffff) {
            return c;
        } else {
            iter->start+=4; /* back to behind this supplementary code point for consistent state */
            iter->reservedField=c;
            return U16_TRAIL(c);
        }
    } else {
        return U_SENTINEL;
    }
}

static uint32_t U_CALLCONV
utf8IteratorGetState(const UCharIterator *iter) {
    uint32_t state=(uint32_t)(iter->start<<1);
    if(iter->reservedField!=0) {
        state|=1;
    }
    return state;
}

static void U_CALLCONV
utf8IteratorSetState(UCharIterator *iter,
                     uint32_t state,
                     UErrorCode *pErrorCode)
{
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        /* do nothing */
    } else if(iter==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
    } else if(state==utf8IteratorGetState(iter)) {
        /* setting to the current state: no-op */
    } else {
        int32_t index=(int32_t)(state>>1); /* UTF-8 index */
        state&=1; /* 1 if in surrogate pair, must be index>=4 */

        if((state==0 ? index<0 : index<4) || iter->limit<index) {
            *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        } else {
            iter->start=index; /* restore UTF-8 byte index */
            if(index<=1) {
                iter->index=index;
            } else {
                iter->index=-1; /* unknown UTF-16 index */
            }
            if(state==0) {
                iter->reservedField=0;
            } else {
                /* verified index>=4 above */
                UChar32 c;
                U8_PREV_OR_FFFD((const uint8_t *)iter->context, 0, index, c);
                if(c<=0xffff) {
                    *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                } else {
                    iter->reservedField=c;
                }
            }
        }
    }
}

static const UCharIterator utf8Iterator={
    0, 0, 0, 0, 0, 0,
    utf8IteratorGetIndex,
    utf8IteratorMove,
    utf8IteratorHasNext,
    utf8IteratorHasPrevious,
    utf8IteratorCurrent,
    utf8IteratorNext,
    utf8IteratorPrevious,
    NULL,
    utf8IteratorGetState,
    utf8IteratorSetState
};

U_CAPI void U_EXPORT2
uiter_setUTF8(UCharIterator *iter, const char *s, int32_t length) {
    if(iter!=0) {
        if(s!=0 && length>=-1) {
            *iter=utf8Iterator;
            iter->context=s;
            if(length>=0) {
                iter->limit=length;
            } else {
                iter->limit=(int32_t)uprv_strlen(s);
            }
            iter->length= iter->limit<=1 ? iter->limit : -1;
        } else {
            *iter=noopIterator;
        }
    }
}

/* Helper functions --------------------------------------------------------- */

U_CAPI UChar32 U_EXPORT2
uiter_current32(UCharIterator *iter) {
    UChar32 c, c2;

    c=iter->current(iter);
    if(U16_IS_SURROGATE(c)) {
        if(U16_IS_SURROGATE_LEAD(c)) {
            /*
             * go to the next code unit
             * we know that we are not at the limit because c!=U_SENTINEL
             */
            iter->move(iter, 1, UITER_CURRENT);
            if(U16_IS_TRAIL(c2=iter->current(iter))) {
                c=U16_GET_SUPPLEMENTARY(c, c2);
            }

            /* undo index movement */
            iter->move(iter, -1, UITER_CURRENT);
        } else {
            if(U16_IS_LEAD(c2=iter->previous(iter))) {
                c=U16_GET_SUPPLEMENTARY(c2, c);
            }
            if(c2>=0) {
                /* undo index movement */
                iter->move(iter, 1, UITER_CURRENT);
            }
        }
    }
    return c;
}

U_CAPI UChar32 U_EXPORT2
uiter_next32(UCharIterator *iter) {
    UChar32 c, c2;

    c=iter->next(iter);
    if(U16_IS_LEAD(c)) {
        if(U16_IS_TRAIL(c2=iter->next(iter))) {
            c=U16_GET_SUPPLEMENTARY(c, c2);
        } else if(c2>=0) {
            /* unmatched first surrogate, undo index movement */
            iter->move(iter, -1, UITER_CURRENT);
        }
    }
    return c;
}

U_CAPI UChar32 U_EXPORT2
uiter_previous32(UCharIterator *iter) {
    UChar32 c, c2;

    c=iter->previous(iter);
    if(U16_IS_TRAIL(c)) {
        if(U16_IS_LEAD(c2=iter->previous(iter))) {
            c=U16_GET_SUPPLEMENTARY(c2, c);
        } else if(c2>=0) {
            /* unmatched second surrogate, undo index movement */
            iter->move(iter, 1, UITER_CURRENT);
        }
    }
    return c;
}

U_CAPI uint32_t U_EXPORT2
uiter_getState(const UCharIterator *iter) {
    if(iter==NULL || iter->getState==NULL) {
        return UITER_NO_STATE;
    } else {
        return iter->getState(iter);
    }
}

U_CAPI void U_EXPORT2
uiter_setState(UCharIterator *iter, uint32_t state, UErrorCode *pErrorCode) {
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        /* do nothing */
    } else if(iter==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
    } else if(iter->setState==NULL) {
        *pErrorCode=U_UNSUPPORTED_ERROR;
    } else {
        iter->setState(iter, state, pErrorCode);
    }
}

U_CDECL_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 2009-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*/






typedef struct UListNode UListNode;
struct UListNode {
    void *data;
    
    UListNode *next;
    UListNode *previous;
    
    /* When data is created with uprv_malloc, needs to be freed during deleteList function. */
    UBool forceDelete;
};

struct UList {
    UListNode *curr;
    UListNode *head;
    UListNode *tail;
    
    int32_t size;
};

static void ulist_addFirstItem(UList *list, UListNode *newItem);

U_CAPI UList *U_EXPORT2 ulist_createEmptyList(UErrorCode *status) {
    UList *newList = NULL;
    
    if (U_FAILURE(*status)) {
        return NULL;
    }
    
    newList = (UList *)uprv_malloc(sizeof(UList));
    if (newList == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    
    newList->curr = NULL;
    newList->head = NULL;
    newList->tail = NULL;
    newList->size = 0;
    
    return newList;
}

/*
 * Function called by addItemEndList or addItemBeginList when the first item is added to the list.
 * This function properly sets the pointers for the first item added.
 */
static void ulist_addFirstItem(UList *list, UListNode *newItem) {
    newItem->next = NULL;
    newItem->previous = NULL;
    list->head = newItem;
    list->tail = newItem;
}

static void ulist_removeItem(UList *list, UListNode *p) {
    if (p->previous == NULL) {
        // p is the list head.
        list->head = p->next;
    } else {
        p->previous->next = p->next;
    }
    if (p->next == NULL) {
        // p is the list tail.
        list->tail = p->previous;
    } else {
        p->next->previous = p->previous;
    }
    if (p == list->curr) {
        list->curr = p->next;
    }
    --list->size;
    if (p->forceDelete) {
        uprv_free(p->data);
    }
    uprv_free(p);
}

U_CAPI void U_EXPORT2 ulist_addItemEndList(UList *list, const void *data, UBool forceDelete, UErrorCode *status) {
    UListNode *newItem = NULL;
    
    if (U_FAILURE(*status) || list == NULL || data == NULL) {
        if (forceDelete) {
            uprv_free((void *)data);
        }
        return;
    }
    
    newItem = (UListNode *)uprv_malloc(sizeof(UListNode));
    if (newItem == NULL) {
        if (forceDelete) {
            uprv_free((void *)data);
        }
        *status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    newItem->data = (void *)(data);
    newItem->forceDelete = forceDelete;
    
    if (list->size == 0) {
        ulist_addFirstItem(list, newItem);
    } else {
        newItem->next = NULL;
        newItem->previous = list->tail;
        list->tail->next = newItem;
        list->tail = newItem;
    }
    
    list->size++;
}

U_CAPI void U_EXPORT2 ulist_addItemBeginList(UList *list, const void *data, UBool forceDelete, UErrorCode *status) {
    UListNode *newItem = NULL;
    
    if (U_FAILURE(*status) || list == NULL || data == NULL) {
        if (forceDelete) {
            uprv_free((void *)data);
        }
        return;
    }
    
    newItem = (UListNode *)uprv_malloc(sizeof(UListNode));
    if (newItem == NULL) {
        if (forceDelete) {
            uprv_free((void *)data);
        }
        *status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    newItem->data = (void *)(data);
    newItem->forceDelete = forceDelete;
    
    if (list->size == 0) {
        ulist_addFirstItem(list, newItem);
    } else {
        newItem->previous = NULL;
        newItem->next = list->head;
        list->head->previous = newItem;
        list->head = newItem;
    }
    
    list->size++;
}

U_CAPI UBool U_EXPORT2 ulist_containsString(const UList *list, const char *data, int32_t length) {
    if (list != NULL) {
        const UListNode *pointer;
        for (pointer = list->head; pointer != NULL; pointer = pointer->next) {
            if (length == (int32_t)uprv_strlen((const char *)pointer->data)) {
                if (uprv_memcmp(data, pointer->data, length) == 0) {
                    return TRUE;
                }
            }
        }
    }
    return FALSE;
}

U_CAPI UBool U_EXPORT2 ulist_removeString(UList *list, const char *data) {
    if (list != NULL) {
        UListNode *pointer;
        for (pointer = list->head; pointer != NULL; pointer = pointer->next) {
            if (uprv_strcmp(data, (const char *)pointer->data) == 0) {
                ulist_removeItem(list, pointer);
                // Remove only the first occurrence, like Java LinkedList.remove(Object).
                return TRUE;
            }
        }
    }
    return FALSE;
}

U_CAPI void *U_EXPORT2 ulist_getNext(UList *list) {
    UListNode *curr = NULL;
    
    if (list == NULL || list->curr == NULL) {
        return NULL;
    }
    
    curr = list->curr;
    list->curr = curr->next;
    
    return curr->data;
}

U_CAPI int32_t U_EXPORT2 ulist_getListSize(const UList *list) {
    if (list != NULL) {
        return list->size;
    }
    
    return -1;
}

U_CAPI void U_EXPORT2 ulist_resetList(UList *list) {
    if (list != NULL) {
        list->curr = list->head;
    }
}

U_CAPI void U_EXPORT2 ulist_deleteList(UList *list) {
    UListNode *listHead = NULL;

    if (list != NULL) {
        listHead = list->head;
        while (listHead != NULL) {
            UListNode *listPointer = listHead->next;

            if (listHead->forceDelete) {
                uprv_free(listHead->data);
            }

            uprv_free(listHead);
            listHead = listPointer;
        }
        uprv_free(list);
        list = NULL;
    }
}

U_CAPI void U_EXPORT2 ulist_close_keyword_values_iterator(UEnumeration *en) {
    if (en != NULL) {
        ulist_deleteList((UList *)(en->context));
        uprv_free(en);
    }
}

U_CAPI int32_t U_EXPORT2 ulist_count_keyword_values(UEnumeration *en, UErrorCode *status) {
    if (U_FAILURE(*status)) {
        return -1;
    }
    
    return ulist_getListSize((UList *)(en->context));
}

U_CAPI const char * U_EXPORT2 ulist_next_keyword_value(UEnumeration *en, int32_t *resultLength, UErrorCode *status) {
    const char *s;
    if (U_FAILURE(*status)) {
        return NULL;
    }

    s = (const char *)ulist_getNext((UList *)(en->context));
    if (s != NULL && resultLength != NULL) {
        *resultLength = static_cast<int32_t>(uprv_strlen(s));
    }
    return s;
}

U_CAPI void U_EXPORT2 ulist_reset_keyword_values_iterator(UEnumeration *en, UErrorCode *status) {
    if (U_FAILURE(*status)) {
        return ;
    }
    
    ulist_resetList((UList *)(en->context));
}

U_CAPI UList * U_EXPORT2 ulist_getListFromEnum(UEnumeration *en) {
    return (UList *)(en->context);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*
* File ULOC.CPP
*
* Modification History:
*
*   Date        Name        Description
*   04/01/97    aliu        Creation.
*   08/21/98    stephen     JDK 1.2 sync
*   12/08/98    rtg         New Locale implementation and C API
*   03/15/99    damiba      overhaul.
*   04/06/99    stephen     changed setDefault() to realloc and copy
*   06/14/99    stephen     Changed calls to ures_open for new params
*   07/21/99    stephen     Modified setDefault() to propagate to C++
*   05/14/04    alan        7 years later: refactored, cleaned up, fixed bugs,
*                           brought canonicalization code into line with spec
*****************************************************************************/

/*
   POSIX's locale format, from putil.c: [no spaces]

     ll [ _CC ] [ . MM ] [ @ VV]

     l = lang, C = ctry, M = charmap, V = variant
*/











// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1996-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File locmap.h      : Locale Mapping Classes
* 
*
* Created by: Helena Shih
*
* Modification History:
*
*  Date        Name        Description
*  3/11/97     aliu        Added setId().
*  4/20/99     Madhu       Added T_convertToPosix()
* 09/18/00     george      Removed the memory leaks.
* 08/23/01     george      Convert to C
*============================================================================
*/




#define LANGUAGE_LCID(hostID) (uint16_t)(0x03FF & hostID)

U_CAPI int32_t uprv_convertToPosix(uint32_t hostid, char* posixID, int32_t posixIDCapacity, UErrorCode* status);

/* Don't call these functions directly. Use uloc_getLCID instead. */
U_CAPI uint32_t uprv_convertToLCIDPlatform(const char* localeID, UErrorCode* status); // Leverage platform conversion if possible
U_CAPI uint32_t uprv_convertToLCID(const char* langID, const char* posixID, UErrorCode* status);







#include <stdio.h> /* for sprintf */

U_NAMESPACE_USE

/* ### Declarations **************************************************/

/* Locale stuff from locid.cpp */
U_CFUNC void locale_set_default(const char *id);
U_CFUNC const char *locale_get_default(void);
U_CFUNC int32_t
locale_getKeywords(const char *localeID,
            char prev,
            char *keywords, int32_t keywordCapacity,
            char *values, int32_t valuesCapacity, int32_t *valLen,
            UBool valuesToo,
            UErrorCode *status);

/* ### Data tables **************************************************/

/**
 * Table of language codes, both 2- and 3-letter, with preference
 * given to 2-letter codes where possible.  Includes 3-letter codes
 * that lack a 2-letter equivalent.
 *
 * This list must be in sorted order.  This list is returned directly
 * to the user by some API.
 *
 * This list must be kept in sync with LANGUAGES_3, with corresponding
 * entries matched.
 *
 * This table should be terminated with a NULL entry, followed by a
 * second list, and another NULL entry.  The first list is visible to
 * user code when this array is returned by API.  The second list
 * contains codes we support, but do not expose through user API.
 *
 * Notes
 *
 * Tables updated per http://lcweb.loc.gov/standards/iso639-2/ to
 * include the revisions up to 2001/7/27 *CWB*
 *
 * The 3 character codes are the terminology codes like RFC 3066.  This
 * is compatible with prior ICU codes
 *
 * "in" "iw" "ji" "jw" & "sh" have been withdrawn but are still in the
 * table but now at the end of the table because 3 character codes are
 * duplicates.  This avoids bad searches going from 3 to 2 character
 * codes.
 *
 * The range qaa-qtz is reserved for local use
 */
/* Generated using org.unicode.cldr.icu.GenerateISO639LanguageTables */
/* ISO639 table version is 20150505 */
/* Subsequent hand addition of selected languages */
static const char * const LANGUAGES[] = {
    "aa",  "ab",  "ace", "ach", "ada", "ady", "ae",  "aeb",
    "af",  "afh", "agq", "ain", "ak",  "akk", "akz", "ale",
    "aln", "alt", "am",  "an",  "ang", "anp", "ar",  "arc",
    "arn", "aro", "arp", "arq", "ars", "arw", "ary", "arz", "as",
    "asa", "ase", "ast", "av",  "avk", "awa", "ay",  "az",
    "ba",  "bal", "ban", "bar", "bas", "bax", "bbc", "bbj",
    "be",  "bej", "bem", "bew", "bez", "bfd", "bfq", "bg",
    "bgn", "bho", "bi",  "bik", "bin", "bjn", "bkm", "bla",
    "bm",  "bn",  "bo",  "bpy", "bqi", "br",  "bra", "brh",
    "brx", "bs",  "bss", "bua", "bug", "bum", "byn", "byv",
    "ca",  "cad", "car", "cay", "cch", "ccp", "ce",  "ceb", "cgg",
    "ch",  "chb", "chg", "chk", "chm", "chn", "cho", "chp",
    "chr", "chy", "ckb", "co",  "cop", "cps", "cr",  "crh",
    "cs",  "csb", "cu",  "cv",  "cy",
    "da",  "dak", "dar", "dav", "de",  "del", "den", "dgr",
    "din", "dje", "doi", "dsb", "dtp", "dua", "dum", "dv",
    "dyo", "dyu", "dz",  "dzg",
    "ebu", "ee",  "efi", "egl", "egy", "eka", "el",  "elx",
    "en",  "enm", "eo",  "es",  "esu", "et",  "eu",  "ewo",
    "ext",
    "fa",  "fan", "fat", "ff",  "fi",  "fil", "fit", "fj",
    "fo",  "fon", "fr",  "frc", "frm", "fro", "frp", "frr",
    "frs", "fur", "fy",
    "ga",  "gaa", "gag", "gan", "gay", "gba", "gbz", "gd",
    "gez", "gil", "gl",  "glk", "gmh", "gn",  "goh", "gom",
    "gon", "gor", "got", "grb", "grc", "gsw", "gu",  "guc",
    "gur", "guz", "gv",  "gwi",
    "ha",  "hai", "hak", "haw", "he",  "hi",  "hif", "hil",
    "hit", "hmn", "ho",  "hr",  "hsb", "hsn", "ht",  "hu",
    "hup", "hy",  "hz",
    "ia",  "iba", "ibb", "id",  "ie",  "ig",  "ii",  "ik",
    "ilo", "inh", "io",  "is",  "it",  "iu",  "izh",
    "ja",  "jam", "jbo", "jgo", "jmc", "jpr", "jrb", "jut",
    "jv",
    "ka",  "kaa", "kab", "kac", "kaj", "kam", "kaw", "kbd",
    "kbl", "kcg", "kde", "kea", "ken", "kfo", "kg",  "kgp",
    "kha", "kho", "khq", "khw", "ki",  "kiu", "kj",  "kk",
    "kkj", "kl",  "kln", "km",  "kmb", "kn",  "ko",  "koi",
    "kok", "kos", "kpe", "kr",  "krc", "kri", "krj", "krl",
    "kru", "ks",  "ksb", "ksf", "ksh", "ku",  "kum", "kut",
    "kv",  "kw",  "ky",
    "la",  "lad", "lag", "lah", "lam", "lb",  "lez", "lfn",
    "lg",  "li",  "lij", "liv", "lkt", "lmo", "ln",  "lo",
    "lol", "loz", "lrc", "lt",  "ltg", "lu",  "lua", "lui",
    "lun", "luo", "lus", "luy", "lv",  "lzh", "lzz",
    "mad", "maf", "mag", "mai", "mak", "man", "mas", "mde",
    "mdf", "mdh", "mdr", "men", "mer", "mfe", "mg",  "mga",
    "mgh", "mgo", "mh",  "mi",  "mic", "min", "mis", "mk",
    "ml",  "mn",  "mnc", "mni", "mo",
    "moh", "mos", "mr",  "mrj",
    "ms",  "mt",  "mua", "mul", "mus", "mwl", "mwr", "mwv",
    "my",  "mye", "myv", "mzn",
    "na",  "nan", "nap", "naq", "nb",  "nd",  "nds", "ne",
    "new", "ng",  "nia", "niu", "njo", "nl",  "nmg", "nn",
    "nnh", "no",  "nog", "non", "nov", "nqo", "nr",  "nso",
    "nus", "nv",  "nwc", "ny",  "nym", "nyn", "nyo", "nzi",
    "oc",  "oj",  "om",  "or",  "os",  "osa", "ota",
    "pa",  "pag", "pal", "pam", "pap", "pau", "pcd", "pdc",
    "pdt", "peo", "pfl", "phn", "pi",  "pl",  "pms", "pnt",
    "pon", "prg", "pro", "ps",  "pt",
    "qu",  "quc", "qug",
    "raj", "rap", "rar", "rgn", "rif", "rm",  "rn",  "ro",
    "rof", "rom", "rtm", "ru",  "rue", "rug", "rup",
    "rw",  "rwk",
    "sa",  "sad", "sah", "sam", "saq", "sas", "sat", "saz",
    "sba", "sbp", "sc",  "scn", "sco", "sd",  "sdc", "sdh",
    "se",  "see", "seh", "sei", "sel", "ses", "sg",  "sga",
    "sgs", "shi", "shn", "shu", "si",  "sid", "sk",
    "sl",  "sli", "sly", "sm",  "sma", "smj", "smn", "sms",
    "sn",  "snk", "so",  "sog", "sq",  "sr",  "srn", "srr",
    "ss",  "ssy", "st",  "stq", "su",  "suk", "sus", "sux",
    "sv",  "sw",  "swb", "swc", "syc", "syr", "szl",
    "ta",  "tcy", "te",  "tem", "teo", "ter", "tet", "tg",
    "th",  "ti",  "tig", "tiv", "tk",  "tkl", "tkr", "tl",
    "tlh", "tli", "tly", "tmh", "tn",  "to",  "tog", "tpi",
    "tr",  "tru", "trv", "ts",  "tsd", "tsi", "tt",  "ttt",
    "tum", "tvl", "tw",  "twq", "ty",  "tyv", "tzm",
    "udm", "ug",  "uga", "uk",  "umb", "und", "ur",  "uz",
    "vai", "ve",  "vec", "vep", "vi",  "vls", "vmf", "vo",
    "vot", "vro", "vun",
    "wa",  "wae", "wal", "war", "was", "wbp", "wo",  "wuu",
    "xal", "xh",  "xmf", "xog",
    "yao", "yap", "yav", "ybb", "yi",  "yo",  "yrl", "yue",
    "za",  "zap", "zbl", "zea", "zen", "zgh", "zh",  "zu",
    "zun", "zxx", "zza",
NULL,
    "in",  "iw",  "ji",  "jw",  "sh",    /* obsolete language codes */
NULL
};

static const char* const DEPRECATED_LANGUAGES[]={
    "in", "iw", "ji", "jw", NULL, NULL
};
static const char* const REPLACEMENT_LANGUAGES[]={
    "id", "he", "yi", "jv", NULL, NULL
};

/**
 * Table of 3-letter language codes.
 *
 * This is a lookup table used to convert 3-letter language codes to
 * their 2-letter equivalent, where possible.  It must be kept in sync
 * with LANGUAGES.  For all valid i, LANGUAGES[i] must refer to the
 * same language as LANGUAGES_3[i].  The commented-out lines are
 * copied from LANGUAGES to make eyeballing this baby easier.
 *
 * Where a 3-letter language code has no 2-letter equivalent, the
 * 3-letter code occupies both LANGUAGES[i] and LANGUAGES_3[i].
 *
 * This table should be terminated with a NULL entry, followed by a
 * second list, and another NULL entry.  The two lists correspond to
 * the two lists in LANGUAGES.
 */
/* Generated using org.unicode.cldr.icu.GenerateISO639LanguageTables */
/* ISO639 table version is 20150505 */
/* Subsequent hand addition of selected languages */
static const char * const LANGUAGES_3[] = {
    "aar", "abk", "ace", "ach", "ada", "ady", "ave", "aeb",
    "afr", "afh", "agq", "ain", "aka", "akk", "akz", "ale",
    "aln", "alt", "amh", "arg", "ang", "anp", "ara", "arc",
    "arn", "aro", "arp", "arq", "ars", "arw", "ary", "arz", "asm",
    "asa", "ase", "ast", "ava", "avk", "awa", "aym", "aze",
    "bak", "bal", "ban", "bar", "bas", "bax", "bbc", "bbj",
    "bel", "bej", "bem", "bew", "bez", "bfd", "bfq", "bul",
    "bgn", "bho", "bis", "bik", "bin", "bjn", "bkm", "bla",
    "bam", "ben", "bod", "bpy", "bqi", "bre", "bra", "brh",
    "brx", "bos", "bss", "bua", "bug", "bum", "byn", "byv",
    "cat", "cad", "car", "cay", "cch", "ccp", "che", "ceb", "cgg",
    "cha", "chb", "chg", "chk", "chm", "chn", "cho", "chp",
    "chr", "chy", "ckb", "cos", "cop", "cps", "cre", "crh",
    "ces", "csb", "chu", "chv", "cym",
    "dan", "dak", "dar", "dav", "deu", "del", "den", "dgr",
    "din", "dje", "doi", "dsb", "dtp", "dua", "dum", "div",
    "dyo", "dyu", "dzo", "dzg",
    "ebu", "ewe", "efi", "egl", "egy", "eka", "ell", "elx",
    "eng", "enm", "epo", "spa", "esu", "est", "eus", "ewo",
    "ext",
    "fas", "fan", "fat", "ful", "fin", "fil", "fit", "fij",
    "fao", "fon", "fra", "frc", "frm", "fro", "frp", "frr",
    "frs", "fur", "fry",
    "gle", "gaa", "gag", "gan", "gay", "gba", "gbz", "gla",
    "gez", "gil", "glg", "glk", "gmh", "grn", "goh", "gom",
    "gon", "gor", "got", "grb", "grc", "gsw", "guj", "guc",
    "gur", "guz", "glv", "gwi",
    "hau", "hai", "hak", "haw", "heb", "hin", "hif", "hil",
    "hit", "hmn", "hmo", "hrv", "hsb", "hsn", "hat", "hun",
    "hup", "hye", "her",
    "ina", "iba", "ibb", "ind", "ile", "ibo", "iii", "ipk",
    "ilo", "inh", "ido", "isl", "ita", "iku", "izh",
    "jpn", "jam", "jbo", "jgo", "jmc", "jpr", "jrb", "jut",
    "jav",
    "kat", "kaa", "kab", "kac", "kaj", "kam", "kaw", "kbd",
    "kbl", "kcg", "kde", "kea", "ken", "kfo", "kon", "kgp",
    "kha", "kho", "khq", "khw", "kik", "kiu", "kua", "kaz",
    "kkj", "kal", "kln", "khm", "kmb", "kan", "kor", "koi",
    "kok", "kos", "kpe", "kau", "krc", "kri", "krj", "krl",
    "kru", "kas", "ksb", "ksf", "ksh", "kur", "kum", "kut",
    "kom", "cor", "kir",
    "lat", "lad", "lag", "lah", "lam", "ltz", "lez", "lfn",
    "lug", "lim", "lij", "liv", "lkt", "lmo", "lin", "lao",
    "lol", "loz", "lrc", "lit", "ltg", "lub", "lua", "lui",
    "lun", "luo", "lus", "luy", "lav", "lzh", "lzz",
    "mad", "maf", "mag", "mai", "mak", "man", "mas", "mde",
    "mdf", "mdh", "mdr", "men", "mer", "mfe", "mlg", "mga",
    "mgh", "mgo", "mah", "mri", "mic", "min", "mis", "mkd",
    "mal", "mon", "mnc", "mni", "mol",
    "moh", "mos", "mar", "mrj",
    "msa", "mlt", "mua", "mul", "mus", "mwl", "mwr", "mwv",
    "mya", "mye", "myv", "mzn",
    "nau", "nan", "nap", "naq", "nob", "nde", "nds", "nep",
    "new", "ndo", "nia", "niu", "njo", "nld", "nmg", "nno",
    "nnh", "nor", "nog", "non", "nov", "nqo", "nbl", "nso",
    "nus", "nav", "nwc", "nya", "nym", "nyn", "nyo", "nzi",
    "oci", "oji", "orm", "ori", "oss", "osa", "ota",
    "pan", "pag", "pal", "pam", "pap", "pau", "pcd", "pdc",
    "pdt", "peo", "pfl", "phn", "pli", "pol", "pms", "pnt",
    "pon", "prg", "pro", "pus", "por",
    "que", "quc", "qug",
    "raj", "rap", "rar", "rgn", "rif", "roh", "run", "ron",
    "rof", "rom", "rtm", "rus", "rue", "rug", "rup",
    "kin", "rwk",
    "san", "sad", "sah", "sam", "saq", "sas", "sat", "saz",
    "sba", "sbp", "srd", "scn", "sco", "snd", "sdc", "sdh",
    "sme", "see", "seh", "sei", "sel", "ses", "sag", "sga",
    "sgs", "shi", "shn", "shu", "sin", "sid", "slk",
    "slv", "sli", "sly", "smo", "sma", "smj", "smn", "sms",
    "sna", "snk", "som", "sog", "sqi", "srp", "srn", "srr",
    "ssw", "ssy", "sot", "stq", "sun", "suk", "sus", "sux",
    "swe", "swa", "swb", "swc", "syc", "syr", "szl",
    "tam", "tcy", "tel", "tem", "teo", "ter", "tet", "tgk",
    "tha", "tir", "tig", "tiv", "tuk", "tkl", "tkr", "tgl",
    "tlh", "tli", "tly", "tmh", "tsn", "ton", "tog", "tpi",
    "tur", "tru", "trv", "tso", "tsd", "tsi", "tat", "ttt",
    "tum", "tvl", "twi", "twq", "tah", "tyv", "tzm",
    "udm", "uig", "uga", "ukr", "umb", "und", "urd", "uzb",
    "vai", "ven", "vec", "vep", "vie", "vls", "vmf", "vol",
    "vot", "vro", "vun",
    "wln", "wae", "wal", "war", "was", "wbp", "wol", "wuu",
    "xal", "xho", "xmf", "xog",
    "yao", "yap", "yav", "ybb", "yid", "yor", "yrl", "yue",
    "zha", "zap", "zbl", "zea", "zen", "zgh", "zho", "zul",
    "zun", "zxx", "zza",
NULL,
/*  "in",  "iw",  "ji",  "jw",  "sh",                          */
    "ind", "heb", "yid", "jaw", "srp",
NULL
};

/**
 * Table of 2-letter country codes.
 *
 * This list must be in sorted order.  This list is returned directly
 * to the user by some API.
 *
 * This list must be kept in sync with COUNTRIES_3, with corresponding
 * entries matched.
 *
 * This table should be terminated with a NULL entry, followed by a
 * second list, and another NULL entry.  The first list is visible to
 * user code when this array is returned by API.  The second list
 * contains codes we support, but do not expose through user API.
 *
 * Notes:
 *
 * ZR(ZAR) is now CD(COD) and FX(FXX) is PS(PSE) as per
 * http://www.evertype.com/standards/iso3166/iso3166-1-en.html added
 * new codes keeping the old ones for compatibility updated to include
 * 1999/12/03 revisions *CWB*
 *
 * RO(ROM) is now RO(ROU) according to
 * http://www.iso.org/iso/en/prods-services/iso3166ma/03updates-on-iso-3166/nlv3e-rou.html
 */
static const char * const COUNTRIES[] = {
    "AD",  "AE",  "AF",  "AG",  "AI",  "AL",  "AM",
    "AO",  "AQ",  "AR",  "AS",  "AT",  "AU",  "AW",  "AX",  "AZ",
    "BA",  "BB",  "BD",  "BE",  "BF",  "BG",  "BH",  "BI",
    "BJ",  "BL",  "BM",  "BN",  "BO",  "BQ",  "BR",  "BS",  "BT",  "BV",
    "BW",  "BY",  "BZ",  "CA",  "CC",  "CD",  "CF",  "CG",
    "CH",  "CI",  "CK",  "CL",  "CM",  "CN",  "CO",  "CR",
    "CU",  "CV",  "CW",  "CX",  "CY",  "CZ",  "DE",  "DJ",  "DK",
    "DM",  "DO",  "DZ",  "EC",  "EE",  "EG",  "EH",  "ER",
    "ES",  "ET",  "FI",  "FJ",  "FK",  "FM",  "FO",  "FR",
    "GA",  "GB",  "GD",  "GE",  "GF",  "GG",  "GH",  "GI",  "GL",
    "GM",  "GN",  "GP",  "GQ",  "GR",  "GS",  "GT",  "GU",
    "GW",  "GY",  "HK",  "HM",  "HN",  "HR",  "HT",  "HU",
    "ID",  "IE",  "IL",  "IM",  "IN",  "IO",  "IQ",  "IR",  "IS",
    "IT",  "JE",  "JM",  "JO",  "JP",  "KE",  "KG",  "KH",  "KI",
    "KM",  "KN",  "KP",  "KR",  "KW",  "KY",  "KZ",  "LA",
    "LB",  "LC",  "LI",  "LK",  "LR",  "LS",  "LT",  "LU",
    "LV",  "LY",  "MA",  "MC",  "MD",  "ME",  "MF",  "MG",  "MH",  "MK",
    "ML",  "MM",  "MN",  "MO",  "MP",  "MQ",  "MR",  "MS",
    "MT",  "MU",  "MV",  "MW",  "MX",  "MY",  "MZ",  "NA",
    "NC",  "NE",  "NF",  "NG",  "NI",  "NL",  "NO",  "NP",
    "NR",  "NU",  "NZ",  "OM",  "PA",  "PE",  "PF",  "PG",
    "PH",  "PK",  "PL",  "PM",  "PN",  "PR",  "PS",  "PT",
    "PW",  "PY",  "QA",  "RE",  "RO",  "RS",  "RU",  "RW",  "SA",
    "SB",  "SC",  "SD",  "SE",  "SG",  "SH",  "SI",  "SJ",
    "SK",  "SL",  "SM",  "SN",  "SO",  "SR",  "SS",  "ST",  "SV",
    "SX",  "SY",  "SZ",  "TC",  "TD",  "TF",  "TG",  "TH",  "TJ",
    "TK",  "TL",  "TM",  "TN",  "TO",  "TR",  "TT",  "TV",
    "TW",  "TZ",  "UA",  "UG",  "UM",  "US",  "UY",  "UZ",
    "VA",  "VC",  "VE",  "VG",  "VI",  "VN",  "VU",  "WF",
    "WS",  "YE",  "YT",  "ZA",  "ZM",  "ZW",
NULL,
    "AN",  "BU", "CS", "FX", "RO", "SU", "TP", "YD", "YU", "ZR",   /* obsolete country codes */
NULL
};

static const char* const DEPRECATED_COUNTRIES[] = {
    "AN", "BU", "CS", "DD", "DY", "FX", "HV", "NH", "RH", "SU", "TP", "UK", "VD", "YD", "YU", "ZR", NULL, NULL /* deprecated country list */
};
static const char* const REPLACEMENT_COUNTRIES[] = {
/*  "AN", "BU", "CS", "DD", "DY", "FX", "HV", "NH", "RH", "SU", "TP", "UK", "VD", "YD", "YU", "ZR" */
    "CW", "MM", "RS", "DE", "BJ", "FR", "BF", "VU", "ZW", "RU", "TL", "GB", "VN", "YE", "RS", "CD", NULL, NULL  /* replacement country codes */
};

/**
 * Table of 3-letter country codes.
 *
 * This is a lookup table used to convert 3-letter country codes to
 * their 2-letter equivalent.  It must be kept in sync with COUNTRIES.
 * For all valid i, COUNTRIES[i] must refer to the same country as
 * COUNTRIES_3[i].  The commented-out lines are copied from COUNTRIES
 * to make eyeballing this baby easier.
 *
 * This table should be terminated with a NULL entry, followed by a
 * second list, and another NULL entry.  The two lists correspond to
 * the two lists in COUNTRIES.
 */
static const char * const COUNTRIES_3[] = {
/*  "AD",  "AE",  "AF",  "AG",  "AI",  "AL",  "AM",      */
    "AND", "ARE", "AFG", "ATG", "AIA", "ALB", "ARM",
/*  "AO",  "AQ",  "AR",  "AS",  "AT",  "AU",  "AW",  "AX",  "AZ",     */
    "AGO", "ATA", "ARG", "ASM", "AUT", "AUS", "ABW", "ALA", "AZE",
/*  "BA",  "BB",  "BD",  "BE",  "BF",  "BG",  "BH",  "BI",     */
    "BIH", "BRB", "BGD", "BEL", "BFA", "BGR", "BHR", "BDI",
/*  "BJ",  "BL",  "BM",  "BN",  "BO",  "BQ",  "BR",  "BS",  "BT",  "BV",     */
    "BEN", "BLM", "BMU", "BRN", "BOL", "BES", "BRA", "BHS", "BTN", "BVT",
/*  "BW",  "BY",  "BZ",  "CA",  "CC",  "CD",  "CF",  "CG",     */
    "BWA", "BLR", "BLZ", "CAN", "CCK", "COD", "CAF", "COG",
/*  "CH",  "CI",  "CK",  "CL",  "CM",  "CN",  "CO",  "CR",     */
    "CHE", "CIV", "COK", "CHL", "CMR", "CHN", "COL", "CRI",
/*  "CU",  "CV",  "CW",  "CX",  "CY",  "CZ",  "DE",  "DJ",  "DK",     */
    "CUB", "CPV", "CUW", "CXR", "CYP", "CZE", "DEU", "DJI", "DNK",
/*  "DM",  "DO",  "DZ",  "EC",  "EE",  "EG",  "EH",  "ER",     */
    "DMA", "DOM", "DZA", "ECU", "EST", "EGY", "ESH", "ERI",
/*  "ES",  "ET",  "FI",  "FJ",  "FK",  "FM",  "FO",  "FR",     */
    "ESP", "ETH", "FIN", "FJI", "FLK", "FSM", "FRO", "FRA",
/*  "GA",  "GB",  "GD",  "GE",  "GF",  "GG",  "GH",  "GI",  "GL",     */
    "GAB", "GBR", "GRD", "GEO", "GUF", "GGY", "GHA", "GIB", "GRL",
/*  "GM",  "GN",  "GP",  "GQ",  "GR",  "GS",  "GT",  "GU",     */
    "GMB", "GIN", "GLP", "GNQ", "GRC", "SGS", "GTM", "GUM",
/*  "GW",  "GY",  "HK",  "HM",  "HN",  "HR",  "HT",  "HU",     */
    "GNB", "GUY", "HKG", "HMD", "HND", "HRV", "HTI", "HUN",
/*  "ID",  "IE",  "IL",  "IM",  "IN",  "IO",  "IQ",  "IR",  "IS" */
    "IDN", "IRL", "ISR", "IMN", "IND", "IOT", "IRQ", "IRN", "ISL",
/*  "IT",  "JE",  "JM",  "JO",  "JP",  "KE",  "KG",  "KH",  "KI",     */
    "ITA", "JEY", "JAM", "JOR", "JPN", "KEN", "KGZ", "KHM", "KIR",
/*  "KM",  "KN",  "KP",  "KR",  "KW",  "KY",  "KZ",  "LA",     */
    "COM", "KNA", "PRK", "KOR", "KWT", "CYM", "KAZ", "LAO",
/*  "LB",  "LC",  "LI",  "LK",  "LR",  "LS",  "LT",  "LU",     */
    "LBN", "LCA", "LIE", "LKA", "LBR", "LSO", "LTU", "LUX",
/*  "LV",  "LY",  "MA",  "MC",  "MD",  "ME",  "MF",  "MG",  "MH",  "MK",     */
    "LVA", "LBY", "MAR", "MCO", "MDA", "MNE", "MAF", "MDG", "MHL", "MKD",
/*  "ML",  "MM",  "MN",  "MO",  "MP",  "MQ",  "MR",  "MS",     */
    "MLI", "MMR", "MNG", "MAC", "MNP", "MTQ", "MRT", "MSR",
/*  "MT",  "MU",  "MV",  "MW",  "MX",  "MY",  "MZ",  "NA",     */
    "MLT", "MUS", "MDV", "MWI", "MEX", "MYS", "MOZ", "NAM",
/*  "NC",  "NE",  "NF",  "NG",  "NI",  "NL",  "NO",  "NP",     */
    "NCL", "NER", "NFK", "NGA", "NIC", "NLD", "NOR", "NPL",
/*  "NR",  "NU",  "NZ",  "OM",  "PA",  "PE",  "PF",  "PG",     */
    "NRU", "NIU", "NZL", "OMN", "PAN", "PER", "PYF", "PNG",
/*  "PH",  "PK",  "PL",  "PM",  "PN",  "PR",  "PS",  "PT",     */
    "PHL", "PAK", "POL", "SPM", "PCN", "PRI", "PSE", "PRT",
/*  "PW",  "PY",  "QA",  "RE",  "RO",  "RS",  "RU",  "RW",  "SA",     */
    "PLW", "PRY", "QAT", "REU", "ROU", "SRB", "RUS", "RWA", "SAU",
/*  "SB",  "SC",  "SD",  "SE",  "SG",  "SH",  "SI",  "SJ",     */
    "SLB", "SYC", "SDN", "SWE", "SGP", "SHN", "SVN", "SJM",
/*  "SK",  "SL",  "SM",  "SN",  "SO",  "SR",  "SS",  "ST",  "SV",     */
    "SVK", "SLE", "SMR", "SEN", "SOM", "SUR", "SSD", "STP", "SLV",
/*  "SX",  "SY",  "SZ",  "TC",  "TD",  "TF",  "TG",  "TH",  "TJ",     */
    "SXM", "SYR", "SWZ", "TCA", "TCD", "ATF", "TGO", "THA", "TJK",
/*  "TK",  "TL",  "TM",  "TN",  "TO",  "TR",  "TT",  "TV",     */
    "TKL", "TLS", "TKM", "TUN", "TON", "TUR", "TTO", "TUV",
/*  "TW",  "TZ",  "UA",  "UG",  "UM",  "US",  "UY",  "UZ",     */
    "TWN", "TZA", "UKR", "UGA", "UMI", "USA", "URY", "UZB",
/*  "VA",  "VC",  "VE",  "VG",  "VI",  "VN",  "VU",  "WF",     */
    "VAT", "VCT", "VEN", "VGB", "VIR", "VNM", "VUT", "WLF",
/*  "WS",  "YE",  "YT",  "ZA",  "ZM",  "ZW",          */
    "WSM", "YEM", "MYT", "ZAF", "ZMB", "ZWE",
NULL,
/*  "AN",  "BU",  "CS",  "FX",  "RO", "SU",  "TP",  "YD",  "YU",  "ZR" */
    "ANT", "BUR", "SCG", "FXX", "ROM", "SUN", "TMP", "YMD", "YUG", "ZAR",
NULL
};

typedef struct CanonicalizationMap {
    const char *id;          /* input ID */
    const char *canonicalID; /* canonicalized output ID */
} CanonicalizationMap;

/**
 * A map to canonicalize locale IDs.  This handles a variety of
 * different semantic kinds of transformations.
 */
static const CanonicalizationMap CANONICALIZE_MAP[] = {
    { "art_LOJBAN",     "jbo" }, /* registered name */
    { "hy__AREVELA",    "hy" }, /* Registered IANA variant */
    { "hy__AREVMDA",    "hyw" }, /* Registered IANA variant */
    { "zh_GAN",         "gan" }, /* registered name */
    { "zh_GUOYU",       "zh" }, /* registered name */
    { "zh_HAKKA",       "hak" }, /* registered name */
    { "zh_MIN_NAN",     "nan" }, /* registered name */
    { "zh_WUU",         "wuu" }, /* registered name */
    { "zh_XIANG",       "hsn" }, /* registered name */
    { "zh_YUE",         "yue" }, /* registered name */
};

/* ### BCP47 Conversion *******************************************/
/* Test if the locale id has BCP47 u extension and does not have '@' */
#define _hasBCP47Extension(id) (id && uprv_strstr(id, "@") == NULL && getShortestSubtagLength(localeID) == 1)
/* Converts the BCP47 id to Unicode id. Does nothing to id if conversion fails */
#define _ConvertBCP47(finalID, id, buffer, length,err) UPRV_BLOCK_MACRO_BEGIN { \
    if (uloc_forLanguageTag(id, buffer, length, NULL, err) <= 0 || \
            U_FAILURE(*err) || *err == U_STRING_NOT_TERMINATED_WARNING) { \
        finalID=id; \
        if (*err == U_STRING_NOT_TERMINATED_WARNING) { *err = U_BUFFER_OVERFLOW_ERROR; } \
    } else { \
        finalID=buffer; \
    } \
} UPRV_BLOCK_MACRO_END
/* Gets the size of the shortest subtag in the given localeID. */
static int32_t getShortestSubtagLength(const char *localeID) {
    int32_t localeIDLength = static_cast<int32_t>(uprv_strlen(localeID));
    int32_t length = localeIDLength;
    int32_t tmpLength = 0;
    int32_t i;
    UBool reset = TRUE;

    for (i = 0; i < localeIDLength; i++) {
        if (localeID[i] != '_' && localeID[i] != '-') {
            if (reset) {
                tmpLength = 0;
                reset = FALSE;
            }
            tmpLength++;
        } else {
            if (tmpLength != 0 && tmpLength < length) {
                length = tmpLength;
            }
            reset = TRUE;
        }
    }

    return length;
}

/* ### Keywords **************************************************/
#define UPRV_ISDIGIT(c) (((c) >= '0') && ((c) <= '9'))
#define UPRV_ISALPHANUM(c) (uprv_isASCIILetter(c) || UPRV_ISDIGIT(c) )
/* Punctuation/symbols allowed in legacy key values */
#define UPRV_OK_VALUE_PUNCTUATION(c) ((c) == '_' || (c) == '-' || (c) == '+' || (c) == '/')

#define ULOC_KEYWORD_BUFFER_LEN 25
#define ULOC_MAX_NO_KEYWORDS 25

U_CAPI const char * U_EXPORT2
locale_getKeywordsStart(const char *localeID) {
    const char *result = NULL;
    if((result = uprv_strchr(localeID, '@')) != NULL) {
        return result;
    }
#if (U_CHARSET_FAMILY == U_EBCDIC_FAMILY)
    else {
        /* We do this because the @ sign is variant, and the @ sign used on one
        EBCDIC machine won't be compiled the same way on other EBCDIC based
        machines. */
        static const uint8_t ebcdicSigns[] = { 0x7C, 0x44, 0x66, 0x80, 0xAC, 0xAE, 0xAF, 0xB5, 0xEC, 0xEF, 0x00 };
        const uint8_t *charToFind = ebcdicSigns;
        while(*charToFind) {
            if((result = uprv_strchr(localeID, *charToFind)) != NULL) {
                return result;
            }
            charToFind++;
        }
    }
#endif
    return NULL;
}

/**
 * @param buf buffer of size [ULOC_KEYWORD_BUFFER_LEN]
 * @param keywordName incoming name to be canonicalized
 * @param status return status (keyword too long)
 * @return length of the keyword name
 */
static int32_t locale_canonKeywordName(char *buf, const char *keywordName, UErrorCode *status)
{
  int32_t keywordNameLen = 0;

  for (; *keywordName != 0; keywordName++) {
    if (!UPRV_ISALPHANUM(*keywordName)) {
      *status = U_ILLEGAL_ARGUMENT_ERROR; /* malformed keyword name */
      return 0;
    }
    if (keywordNameLen < ULOC_KEYWORD_BUFFER_LEN - 1) {
      buf[keywordNameLen++] = uprv_tolower(*keywordName);
    } else {
      /* keyword name too long for internal buffer */
      *status = U_INTERNAL_PROGRAM_ERROR;
      return 0;
    }
  }
  if (keywordNameLen == 0) {
    *status = U_ILLEGAL_ARGUMENT_ERROR; /* empty keyword name */
    return 0;
  }
  buf[keywordNameLen] = 0; /* terminate */

  return keywordNameLen;
}

typedef struct {
    char keyword[ULOC_KEYWORD_BUFFER_LEN];
    int32_t keywordLen;
    const char *valueStart;
    int32_t valueLen;
} KeywordStruct;

static int32_t U_CALLCONV
compareKeywordStructs(const void * /*context*/, const void *left, const void *right) {
    const char* leftString = ((const KeywordStruct *)left)->keyword;
    const char* rightString = ((const KeywordStruct *)right)->keyword;
    return uprv_strcmp(leftString, rightString);
}

static int32_t
_getKeywords(const char *localeID,
             char prev,
             char *keywords, int32_t keywordCapacity,
             char *values, int32_t valuesCapacity, int32_t *valLen,
             UBool valuesToo,
             UErrorCode *status)
{
    KeywordStruct keywordList[ULOC_MAX_NO_KEYWORDS];

    int32_t maxKeywords = ULOC_MAX_NO_KEYWORDS;
    int32_t numKeywords = 0;
    const char* pos = localeID;
    const char* equalSign = NULL;
    const char* semicolon = NULL;
    int32_t i = 0, j, n;
    int32_t keywordsLen = 0;
    int32_t valuesLen = 0;

    if(prev == '@') { /* start of keyword definition */
        /* we will grab pairs, trim spaces, lowercase keywords, sort and return */
        do {
            UBool duplicate = FALSE;
            /* skip leading spaces */
            while(*pos == ' ') {
                pos++;
            }
            if (!*pos) { /* handle trailing "; " */
                break;
            }
            if(numKeywords == maxKeywords) {
                *status = U_INTERNAL_PROGRAM_ERROR;
                return 0;
            }
            equalSign = uprv_strchr(pos, '=');
            semicolon = uprv_strchr(pos, ';');
            /* lack of '=' [foo@currency] is illegal */
            /* ';' before '=' [foo@currency;collation=pinyin] is illegal */
            if(!equalSign || (semicolon && semicolon<equalSign)) {
                *status = U_INVALID_FORMAT_ERROR;
                return 0;
            }
            /* need to normalize both keyword and keyword name */
            if(equalSign - pos >= ULOC_KEYWORD_BUFFER_LEN) {
                /* keyword name too long for internal buffer */
                *status = U_INTERNAL_PROGRAM_ERROR;
                return 0;
            }
            for(i = 0, n = 0; i < equalSign - pos; ++i) {
                if (pos[i] != ' ') {
                    keywordList[numKeywords].keyword[n++] = uprv_tolower(pos[i]);
                }
            }

            /* zero-length keyword is an error. */
            if (n == 0) {
                *status = U_INVALID_FORMAT_ERROR;
                return 0;
            }

            keywordList[numKeywords].keyword[n] = 0;
            keywordList[numKeywords].keywordLen = n;
            /* now grab the value part. First we skip the '=' */
            equalSign++;
            /* then we leading spaces */
            while(*equalSign == ' ') {
                equalSign++;
            }

            /* Premature end or zero-length value */
            if (!*equalSign || equalSign == semicolon) {
                *status = U_INVALID_FORMAT_ERROR;
                return 0;
            }

            keywordList[numKeywords].valueStart = equalSign;

            pos = semicolon;
            i = 0;
            if(pos) {
                while(*(pos - i - 1) == ' ') {
                    i++;
                }
                keywordList[numKeywords].valueLen = (int32_t)(pos - equalSign - i);
                pos++;
            } else {
                i = (int32_t)uprv_strlen(equalSign);
                while(i && equalSign[i-1] == ' ') {
                    i--;
                }
                keywordList[numKeywords].valueLen = i;
            }
            /* If this is a duplicate keyword, then ignore it */
            for (j=0; j<numKeywords; ++j) {
                if (uprv_strcmp(keywordList[j].keyword, keywordList[numKeywords].keyword) == 0) {
                    duplicate = TRUE;
                    break;
                }
            }
            if (!duplicate) {
                ++numKeywords;
            }
        } while(pos);

        /* now we have a list of keywords */
        /* we need to sort it */
        uprv_sortArray(keywordList, numKeywords, sizeof(KeywordStruct), compareKeywordStructs, NULL, FALSE, status);

        /* Now construct the keyword part */
        for(i = 0; i < numKeywords; i++) {
            if(keywordsLen + keywordList[i].keywordLen + 1< keywordCapacity) {
                uprv_strcpy(keywords+keywordsLen, keywordList[i].keyword);
                if(valuesToo) {
                    keywords[keywordsLen + keywordList[i].keywordLen] = '=';
                } else {
                    keywords[keywordsLen + keywordList[i].keywordLen] = 0;
                }
            }
            keywordsLen += keywordList[i].keywordLen + 1;
            if(valuesToo) {
                if(keywordsLen + keywordList[i].valueLen <= keywordCapacity) {
                    uprv_strncpy(keywords+keywordsLen, keywordList[i].valueStart, keywordList[i].valueLen);
                }
                keywordsLen += keywordList[i].valueLen;

                if(i < numKeywords - 1) {
                    if(keywordsLen < keywordCapacity) {
                        keywords[keywordsLen] = ';';
                    }
                    keywordsLen++;
                }
            }
            if(values) {
                if(valuesLen + keywordList[i].valueLen + 1< valuesCapacity) {
                    uprv_strcpy(values+valuesLen, keywordList[i].valueStart);
                    values[valuesLen + keywordList[i].valueLen] = 0;
                }
                valuesLen += keywordList[i].valueLen + 1;
            }
        }
        if(values) {
            values[valuesLen] = 0;
            if(valLen) {
                *valLen = valuesLen;
            }
        }
        return u_terminateChars(keywords, keywordCapacity, keywordsLen, status);
    } else {
        return 0;
    }
}

U_CFUNC int32_t
locale_getKeywords(const char *localeID,
                   char prev,
                   char *keywords, int32_t keywordCapacity,
                   char *values, int32_t valuesCapacity, int32_t *valLen,
                   UBool valuesToo,
                   UErrorCode *status) {
    return _getKeywords(localeID, prev, keywords, keywordCapacity,
                        values, valuesCapacity, valLen, valuesToo,
                        status);
}

U_CAPI int32_t U_EXPORT2
uloc_getKeywordValue(const char* localeID,
                     const char* keywordName,
                     char* buffer, int32_t bufferCapacity,
                     UErrorCode* status)
{
    if (buffer != nullptr) {
        buffer[0] = '\0';
    }
    const char* startSearchHere = NULL;
    const char* nextSeparator = NULL;
    char keywordNameBuffer[ULOC_KEYWORD_BUFFER_LEN];
    char localeKeywordNameBuffer[ULOC_KEYWORD_BUFFER_LEN];
    int32_t result = 0;

    if(status && U_SUCCESS(*status) && localeID) {
      char tempBuffer[ULOC_FULLNAME_CAPACITY];
      const char* tmpLocaleID;

      if (keywordName == NULL || keywordName[0] == 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
      }

      locale_canonKeywordName(keywordNameBuffer, keywordName, status);
      if(U_FAILURE(*status)) {
        return 0;
      }

      if (_hasBCP47Extension(localeID)) {
          _ConvertBCP47(tmpLocaleID, localeID, tempBuffer, sizeof(tempBuffer), status);
      } else {
          tmpLocaleID=localeID;
      }

      startSearchHere = locale_getKeywordsStart(tmpLocaleID);
      if(startSearchHere == NULL) {
          /* no keywords, return at once */
          return 0;
      }

      /* find the first keyword */
      while(startSearchHere) {
          const char* keyValueTail;
          int32_t keyValueLen;

          startSearchHere++; /* skip @ or ; */
          nextSeparator = uprv_strchr(startSearchHere, '=');
          if(!nextSeparator) {
              *status = U_ILLEGAL_ARGUMENT_ERROR; /* key must have =value */
              return 0;
          }
          /* strip leading & trailing spaces (TC decided to tolerate these) */
          while(*startSearchHere == ' ') {
              startSearchHere++;
          }
          keyValueTail = nextSeparator;
          while (keyValueTail > startSearchHere && *(keyValueTail-1) == ' ') {
              keyValueTail--;
          }
          /* now keyValueTail points to first char after the keyName */
          /* copy & normalize keyName from locale */
          if (startSearchHere == keyValueTail) {
              *status = U_ILLEGAL_ARGUMENT_ERROR; /* empty keyword name in passed-in locale */
              return 0;
          }
          keyValueLen = 0;
          while (startSearchHere < keyValueTail) {
            if (!UPRV_ISALPHANUM(*startSearchHere)) {
              *status = U_ILLEGAL_ARGUMENT_ERROR; /* malformed keyword name */
              return 0;
            }
            if (keyValueLen < ULOC_KEYWORD_BUFFER_LEN - 1) {
              localeKeywordNameBuffer[keyValueLen++] = uprv_tolower(*startSearchHere++);
            } else {
              /* keyword name too long for internal buffer */
              *status = U_INTERNAL_PROGRAM_ERROR;
              return 0;
            }
          }
          localeKeywordNameBuffer[keyValueLen] = 0; /* terminate */

          startSearchHere = uprv_strchr(nextSeparator, ';');

          if(uprv_strcmp(keywordNameBuffer, localeKeywordNameBuffer) == 0) {
               /* current entry matches the keyword. */
             nextSeparator++; /* skip '=' */
              /* First strip leading & trailing spaces (TC decided to tolerate these) */
              while(*nextSeparator == ' ') {
                nextSeparator++;
              }
              keyValueTail = (startSearchHere)? startSearchHere: nextSeparator + uprv_strlen(nextSeparator);
              while(keyValueTail > nextSeparator && *(keyValueTail-1) == ' ') {
                keyValueTail--;
              }
              /* Now copy the value, but check well-formedness */
              if (nextSeparator == keyValueTail) {
                *status = U_ILLEGAL_ARGUMENT_ERROR; /* empty key value name in passed-in locale */
                return 0;
              }
              keyValueLen = 0;
              while (nextSeparator < keyValueTail) {
                if (!UPRV_ISALPHANUM(*nextSeparator) && !UPRV_OK_VALUE_PUNCTUATION(*nextSeparator)) {
                  *status = U_ILLEGAL_ARGUMENT_ERROR; /* malformed key value */
                  return 0;
                }
                if (keyValueLen < bufferCapacity) {
                  /* Should we lowercase value to return here? Tests expect as-is. */
                  buffer[keyValueLen++] = *nextSeparator++;
                } else { /* keep advancing so we return correct length in case of overflow */
                  keyValueLen++;
                  nextSeparator++;
                }
              }
              result = u_terminateChars(buffer, bufferCapacity, keyValueLen, status);
              return result;
          }
      }
    }
    return 0;
}

U_CAPI int32_t U_EXPORT2
uloc_setKeywordValue(const char* keywordName,
                     const char* keywordValue,
                     char* buffer, int32_t bufferCapacity,
                     UErrorCode* status)
{
    /* TODO: sorting. removal. */
    int32_t keywordNameLen;
    int32_t keywordValueLen;
    int32_t bufLen;
    int32_t needLen = 0;
    char keywordNameBuffer[ULOC_KEYWORD_BUFFER_LEN];
    char keywordValueBuffer[ULOC_KEYWORDS_CAPACITY+1];
    char localeKeywordNameBuffer[ULOC_KEYWORD_BUFFER_LEN];
    int32_t rc;
    char* nextSeparator = NULL;
    char* nextEqualsign = NULL;
    char* startSearchHere = NULL;
    char* keywordStart = NULL;
    CharString updatedKeysAndValues;
    int32_t updatedKeysAndValuesLen;
    UBool handledInputKeyAndValue = FALSE;
    char keyValuePrefix = '@';

    if(U_FAILURE(*status)) {
        return -1;
    }
    if (keywordName == NULL || keywordName[0] == 0 || bufferCapacity <= 1) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    bufLen = (int32_t)uprv_strlen(buffer);
    if(bufferCapacity<bufLen) {
        /* The capacity is less than the length?! Is this NULL terminated? */
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    keywordNameLen = locale_canonKeywordName(keywordNameBuffer, keywordName, status);
    if(U_FAILURE(*status)) {
        return 0;
    }

    keywordValueLen = 0;
    if(keywordValue) {
        while (*keywordValue != 0) {
            if (!UPRV_ISALPHANUM(*keywordValue) && !UPRV_OK_VALUE_PUNCTUATION(*keywordValue)) {
                *status = U_ILLEGAL_ARGUMENT_ERROR; /* malformed key value */
                return 0;
            }
            if (keywordValueLen < ULOC_KEYWORDS_CAPACITY) {
                /* Should we force lowercase in value to set? */
                keywordValueBuffer[keywordValueLen++] = *keywordValue++;
            } else {
                /* keywordValue too long for internal buffer */
                *status = U_INTERNAL_PROGRAM_ERROR;
                return 0;
            }
        }
    }
    keywordValueBuffer[keywordValueLen] = 0; /* terminate */

    startSearchHere = (char*)locale_getKeywordsStart(buffer);
    if(startSearchHere == NULL || (startSearchHere[1]==0)) {
        if(keywordValueLen == 0) { /* no keywords = nothing to remove */
            return bufLen;
        }

        needLen = bufLen+1+keywordNameLen+1+keywordValueLen;
        if(startSearchHere) { /* had a single @ */
            needLen--; /* already had the @ */
            /* startSearchHere points at the @ */
        } else {
            startSearchHere=buffer+bufLen;
        }
        if(needLen >= bufferCapacity) {
            *status = U_BUFFER_OVERFLOW_ERROR;
            return needLen; /* no change */
        }
        *startSearchHere++ = '@';
        uprv_strcpy(startSearchHere, keywordNameBuffer);
        startSearchHere += keywordNameLen;
        *startSearchHere++ = '=';
        uprv_strcpy(startSearchHere, keywordValueBuffer);
        return needLen;
    } /* end shortcut - no @ */

    keywordStart = startSearchHere;
    /* search for keyword */
    while(keywordStart) {
        const char* keyValueTail;
        int32_t keyValueLen;

        keywordStart++; /* skip @ or ; */
        nextEqualsign = uprv_strchr(keywordStart, '=');
        if (!nextEqualsign) {
            *status = U_ILLEGAL_ARGUMENT_ERROR; /* key must have =value */
            return 0;
        }
        /* strip leading & trailing spaces (TC decided to tolerate these) */
        while(*keywordStart == ' ') {
            keywordStart++;
        }
        keyValueTail = nextEqualsign;
        while (keyValueTail > keywordStart && *(keyValueTail-1) == ' ') {
            keyValueTail--;
        }
        /* now keyValueTail points to first char after the keyName */
        /* copy & normalize keyName from locale */
        if (keywordStart == keyValueTail) {
            *status = U_ILLEGAL_ARGUMENT_ERROR; /* empty keyword name in passed-in locale */
            return 0;
        }
        keyValueLen = 0;
        while (keywordStart < keyValueTail) {
            if (!UPRV_ISALPHANUM(*keywordStart)) {
                *status = U_ILLEGAL_ARGUMENT_ERROR; /* malformed keyword name */
                return 0;
            }
            if (keyValueLen < ULOC_KEYWORD_BUFFER_LEN - 1) {
                localeKeywordNameBuffer[keyValueLen++] = uprv_tolower(*keywordStart++);
            } else {
                /* keyword name too long for internal buffer */
                *status = U_INTERNAL_PROGRAM_ERROR;
                return 0;
            }
        }
        localeKeywordNameBuffer[keyValueLen] = 0; /* terminate */

        nextSeparator = uprv_strchr(nextEqualsign, ';');

        /* start processing the value part */
        nextEqualsign++; /* skip '=' */
        /* First strip leading & trailing spaces (TC decided to tolerate these) */
        while(*nextEqualsign == ' ') {
            nextEqualsign++;
        }
        keyValueTail = (nextSeparator)? nextSeparator: nextEqualsign + uprv_strlen(nextEqualsign);
        while(keyValueTail > nextEqualsign && *(keyValueTail-1) == ' ') {
            keyValueTail--;
        }
        if (nextEqualsign == keyValueTail) {
            *status = U_ILLEGAL_ARGUMENT_ERROR; /* empty key value in passed-in locale */
            return 0;
        }

        rc = uprv_strcmp(keywordNameBuffer, localeKeywordNameBuffer);
        if(rc == 0) {
            /* Current entry matches the input keyword. Update the entry */
            if(keywordValueLen > 0) { /* updating a value */
                updatedKeysAndValues.append(keyValuePrefix, *status);
                keyValuePrefix = ';'; /* for any subsequent key-value pair */
                updatedKeysAndValues.append(keywordNameBuffer, keywordNameLen, *status);
                updatedKeysAndValues.append('=', *status);
                updatedKeysAndValues.append(keywordValueBuffer, keywordValueLen, *status);
            } /* else removing this entry, don't emit anything */
            handledInputKeyAndValue = TRUE;
        } else {
           /* input keyword sorts earlier than current entry, add before current entry */
            if (rc < 0 && keywordValueLen > 0 && !handledInputKeyAndValue) {
                /* insert new entry at this location */
                updatedKeysAndValues.append(keyValuePrefix, *status);
                keyValuePrefix = ';'; /* for any subsequent key-value pair */
                updatedKeysAndValues.append(keywordNameBuffer, keywordNameLen, *status);
                updatedKeysAndValues.append('=', *status);
                updatedKeysAndValues.append(keywordValueBuffer, keywordValueLen, *status);
                handledInputKeyAndValue = TRUE;
            }
            /* copy the current entry */
            updatedKeysAndValues.append(keyValuePrefix, *status);
            keyValuePrefix = ';'; /* for any subsequent key-value pair */
            updatedKeysAndValues.append(localeKeywordNameBuffer, keyValueLen, *status);
            updatedKeysAndValues.append('=', *status);
            updatedKeysAndValues.append(nextEqualsign, static_cast<int32_t>(keyValueTail-nextEqualsign), *status);
        }
        if (!nextSeparator && keywordValueLen > 0 && !handledInputKeyAndValue) {
            /* append new entry at the end, it sorts later than existing entries */
            updatedKeysAndValues.append(keyValuePrefix, *status);
            /* skip keyValuePrefix update, no subsequent key-value pair */
            updatedKeysAndValues.append(keywordNameBuffer, keywordNameLen, *status);
            updatedKeysAndValues.append('=', *status);
            updatedKeysAndValues.append(keywordValueBuffer, keywordValueLen, *status);
            handledInputKeyAndValue = TRUE;
        }
        keywordStart = nextSeparator;
    } /* end loop searching */

    /* Any error from updatedKeysAndValues.append above would be internal and not due to
     * problems with the passed-in locale. So if we did encounter problems with the
     * passed-in locale above, those errors took precedence and overrode any error
     * status from updatedKeysAndValues.append, and also caused a return of 0. If there
     * are errors here they are from updatedKeysAndValues.append; they do cause an
     * error return but the passed-in locale is unmodified and the original bufLen is
     * returned.
     */
    if (!handledInputKeyAndValue || U_FAILURE(*status)) {
        /* if input key/value specified removal of a keyword not present in locale, or
         * there was an error in CharString.append, leave original locale alone. */
        return bufLen;
    }

    updatedKeysAndValuesLen = updatedKeysAndValues.length();
    /* needLen = length of the part before '@' + length of updated key-value part including '@' */
    needLen = (int32_t)(startSearchHere - buffer) + updatedKeysAndValuesLen;
    if(needLen >= bufferCapacity) {
        *status = U_BUFFER_OVERFLOW_ERROR;
        return needLen; /* no change */
    }
    if (updatedKeysAndValuesLen > 0) {
        uprv_strncpy(startSearchHere, updatedKeysAndValues.data(), updatedKeysAndValuesLen);
    }
    buffer[needLen]=0;
    return needLen;
}

/* ### ID parsing implementation **************************************************/

#define _isPrefixLetter(a) ((a=='x')||(a=='X')||(a=='i')||(a=='I'))

/*returns TRUE if one of the special prefixes is here (s=string)
  'x-' or 'i-' */
#define _isIDPrefix(s) (_isPrefixLetter(s[0])&&_isIDSeparator(s[1]))

/* Dot terminates it because of POSIX form  where dot precedes the codepage
 * except for variant
 */
#define _isTerminator(a)  ((a==0)||(a=='.')||(a=='@'))

/**
 * Lookup 'key' in the array 'list'.  The array 'list' should contain
 * a NULL entry, followed by more entries, and a second NULL entry.
 *
 * The 'list' param should be LANGUAGES, LANGUAGES_3, COUNTRIES, or
 * COUNTRIES_3.
 */
static int16_t _findIndex(const char* const* list, const char* key)
{
    const char* const* anchor = list;
    int32_t pass = 0;

    /* Make two passes through two NULL-terminated arrays at 'list' */
    while (pass++ < 2) {
        while (*list) {
            if (uprv_strcmp(key, *list) == 0) {
                return (int16_t)(list - anchor);
            }
            list++;
        }
        ++list;     /* skip final NULL *CWB*/
    }
    return -1;
}

/* count the length of src while copying it to dest; return strlen(src) */
static inline int32_t
_copyCount(char *dest, int32_t destCapacity, const char *src) {
    const char *anchor;
    char c;

    anchor=src;
    for(;;) {
        if((c=*src)==0) {
            return (int32_t)(src-anchor);
        }
        if(destCapacity<=0) {
            return (int32_t)((src-anchor)+uprv_strlen(src));
        }
        ++src;
        *dest++=c;
        --destCapacity;
    }
}

U_CFUNC const char*
uloc_getCurrentCountryID(const char* oldID){
    int32_t offset = _findIndex(DEPRECATED_COUNTRIES, oldID);
    if (offset >= 0) {
        return REPLACEMENT_COUNTRIES[offset];
    }
    return oldID;
}
U_CFUNC const char*
uloc_getCurrentLanguageID(const char* oldID){
    int32_t offset = _findIndex(DEPRECATED_LANGUAGES, oldID);
    if (offset >= 0) {
        return REPLACEMENT_LANGUAGES[offset];
    }
    return oldID;
}
/*
 * the internal functions _getLanguage(), _getCountry(), _getVariant()
 * avoid duplicating code to handle the earlier locale ID pieces
 * in the functions for the later ones by
 * setting the *pEnd pointer to where they stopped parsing
 *
 * TODO try to use this in Locale
 */
U_CFUNC int32_t
ulocimp_getLanguage(const char *localeID,
                    char *language, int32_t languageCapacity,
                    const char **pEnd) {
    int32_t i=0;
    int32_t offset;
    char lang[4]={ 0, 0, 0, 0 }; /* temporary buffer to hold language code for searching */

    if (uprv_stricmp(localeID, "root") == 0) {
        localeID += 4;
    } else if (uprv_strnicmp(localeID, "und", 3) == 0 &&
               (localeID[3] == '\0' ||
                localeID[3] == '-' ||
                localeID[3] == '_' ||
                localeID[3] == '@')) {
        localeID += 3;
    }

    /* if it starts with i- or x- then copy that prefix */
    if(_isIDPrefix(localeID)) {
        if(i<languageCapacity) {
            language[i]=(char)uprv_tolower(*localeID);
        }
        if(i<languageCapacity) {
            language[i+1]='-';
        }
        i+=2;
        localeID+=2;
    }

    /* copy the language as far as possible and count its length */
    while(!_isTerminator(*localeID) && !_isIDSeparator(*localeID)) {
        if(i<languageCapacity) {
            language[i]=(char)uprv_tolower(*localeID);
        }
        if(i<3) {
            U_ASSERT(i>=0);
            lang[i]=(char)uprv_tolower(*localeID);
        }
        i++;
        localeID++;
    }

    if(i==3) {
        /* convert 3 character code to 2 character code if possible *CWB*/
        offset=_findIndex(LANGUAGES_3, lang);
        if(offset>=0) {
            i=_copyCount(language, languageCapacity, LANGUAGES[offset]);
        }
    }

    if(pEnd!=NULL) {
        *pEnd=localeID;
    }
    return i;
}

U_CFUNC int32_t
ulocimp_getScript(const char *localeID,
                  char *script, int32_t scriptCapacity,
                  const char **pEnd)
{
    int32_t idLen = 0;

    if (pEnd != NULL) {
        *pEnd = localeID;
    }

    /* copy the second item as far as possible and count its length */
    while(!_isTerminator(localeID[idLen]) && !_isIDSeparator(localeID[idLen])
            && uprv_isASCIILetter(localeID[idLen])) {
        idLen++;
    }

    /* If it's exactly 4 characters long, then it's a script and not a country. */
    if (idLen == 4) {
        int32_t i;
        if (pEnd != NULL) {
            *pEnd = localeID+idLen;
        }
        if(idLen > scriptCapacity) {
            idLen = scriptCapacity;
        }
        if (idLen >= 1) {
            script[0]=(char)uprv_toupper(*(localeID++));
        }
        for (i = 1; i < idLen; i++) {
            script[i]=(char)uprv_tolower(*(localeID++));
        }
    }
    else {
        idLen = 0;
    }
    return idLen;
}

U_CFUNC int32_t
ulocimp_getCountry(const char *localeID,
                   char *country, int32_t countryCapacity,
                   const char **pEnd)
{
    int32_t idLen=0;
    char cnty[ULOC_COUNTRY_CAPACITY]={ 0, 0, 0, 0 };
    int32_t offset;

    /* copy the country as far as possible and count its length */
    while(!_isTerminator(localeID[idLen]) && !_isIDSeparator(localeID[idLen])) {
        if(idLen<(ULOC_COUNTRY_CAPACITY-1)) {   /*CWB*/
            cnty[idLen]=(char)uprv_toupper(localeID[idLen]);
        }
        idLen++;
    }

    /* the country should be either length 2 or 3 */
    if (idLen == 2 || idLen == 3) {
        UBool gotCountry = FALSE;
        /* convert 3 character code to 2 character code if possible *CWB*/
        if(idLen==3) {
            offset=_findIndex(COUNTRIES_3, cnty);
            if(offset>=0) {
                idLen=_copyCount(country, countryCapacity, COUNTRIES[offset]);
                gotCountry = TRUE;
            }
        }
        if (!gotCountry) {
            int32_t i = 0;
            for (i = 0; i < idLen; i++) {
                if (i < countryCapacity) {
                    country[i]=(char)uprv_toupper(localeID[i]);
                }
            }
        }
        localeID+=idLen;
    } else {
        idLen = 0;
    }

    if(pEnd!=NULL) {
        *pEnd=localeID;
    }

    return idLen;
}

/**
 * @param needSeparator if true, then add leading '_' if any variants
 * are added to 'variant'
 */
static int32_t
_getVariantEx(const char *localeID,
              char prev,
              char *variant, int32_t variantCapacity,
              UBool needSeparator) {
    int32_t i=0;

    /* get one or more variant tags and separate them with '_' */
    if(_isIDSeparator(prev)) {
        /* get a variant string after a '-' or '_' */
        while(!_isTerminator(*localeID)) {
            if (needSeparator) {
                if (i<variantCapacity) {
                    variant[i] = '_';
                }
                ++i;
                needSeparator = FALSE;
            }
            if(i<variantCapacity) {
                variant[i]=(char)uprv_toupper(*localeID);
                if(variant[i]=='-') {
                    variant[i]='_';
                }
            }
            i++;
            localeID++;
        }
    }

    /* if there is no variant tag after a '-' or '_' then look for '@' */
    if(i==0) {
        if(prev=='@') {
            /* keep localeID */
        } else if((localeID=locale_getKeywordsStart(localeID))!=NULL) {
            ++localeID; /* point after the '@' */
        } else {
            return 0;
        }
        while(!_isTerminator(*localeID)) {
            if (needSeparator) {
                if (i<variantCapacity) {
                    variant[i] = '_';
                }
                ++i;
                needSeparator = FALSE;
            }
            if(i<variantCapacity) {
                variant[i]=(char)uprv_toupper(*localeID);
                if(variant[i]=='-' || variant[i]==',') {
                    variant[i]='_';
                }
            }
            i++;
            localeID++;
        }
    }

    return i;
}

static int32_t
_getVariant(const char *localeID,
            char prev,
            char *variant, int32_t variantCapacity) {
    return _getVariantEx(localeID, prev, variant, variantCapacity, FALSE);
}

/* Keyword enumeration */

typedef struct UKeywordsContext {
    char* keywords;
    char* current;
} UKeywordsContext;

U_CDECL_BEGIN

static void U_CALLCONV
uloc_kw_closeKeywords(UEnumeration *enumerator) {
    uprv_free(((UKeywordsContext *)enumerator->context)->keywords);
    uprv_free(enumerator->context);
    uprv_free(enumerator);
}

static int32_t U_CALLCONV
uloc_kw_countKeywords(UEnumeration *en, UErrorCode * /*status*/) {
    char *kw = ((UKeywordsContext *)en->context)->keywords;
    int32_t result = 0;
    while(*kw) {
        result++;
        kw += uprv_strlen(kw)+1;
    }
    return result;
}

static const char * U_CALLCONV
uloc_kw_nextKeyword(UEnumeration* en,
                    int32_t* resultLength,
                    UErrorCode* /*status*/) {
    const char* result = ((UKeywordsContext *)en->context)->current;
    int32_t len = 0;
    if(*result) {
        len = (int32_t)uprv_strlen(((UKeywordsContext *)en->context)->current);
        ((UKeywordsContext *)en->context)->current += len+1;
    } else {
        result = NULL;
    }
    if (resultLength) {
        *resultLength = len;
    }
    return result;
}

static void U_CALLCONV
uloc_kw_resetKeywords(UEnumeration* en,
                      UErrorCode* /*status*/) {
    ((UKeywordsContext *)en->context)->current = ((UKeywordsContext *)en->context)->keywords;
}

U_CDECL_END


static const UEnumeration gKeywordsEnum = {
    NULL,
    NULL,
    uloc_kw_closeKeywords,
    uloc_kw_countKeywords,
    uenum_unextDefault,
    uloc_kw_nextKeyword,
    uloc_kw_resetKeywords
};

U_CAPI UEnumeration* U_EXPORT2
uloc_openKeywordList(const char *keywordList, int32_t keywordListSize, UErrorCode* status)
{
    LocalMemory<UKeywordsContext> myContext;
    LocalMemory<UEnumeration> result;

    if (U_FAILURE(*status)) {
        return nullptr;
    }
    myContext.adoptInstead(static_cast<UKeywordsContext *>(uprv_malloc(sizeof(UKeywordsContext))));
    result.adoptInstead(static_cast<UEnumeration *>(uprv_malloc(sizeof(UEnumeration))));
    if (myContext.isNull() || result.isNull()) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    uprv_memcpy(result.getAlias(), &gKeywordsEnum, sizeof(UEnumeration));
    myContext->keywords = static_cast<char *>(uprv_malloc(keywordListSize+1));
    if (myContext->keywords == nullptr) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    uprv_memcpy(myContext->keywords, keywordList, keywordListSize);
    myContext->keywords[keywordListSize] = 0;
    myContext->current = myContext->keywords;
    result->context = myContext.orphan();
    return result.orphan();
}

U_CAPI UEnumeration* U_EXPORT2
uloc_openKeywords(const char* localeID,
                        UErrorCode* status)
{
    int32_t i=0;
    char keywords[256];
    int32_t keywordsCapacity = 256;
    char tempBuffer[ULOC_FULLNAME_CAPACITY];
    const char* tmpLocaleID;

    if(status==NULL || U_FAILURE(*status)) {
        return 0;
    }

    if (_hasBCP47Extension(localeID)) {
        _ConvertBCP47(tmpLocaleID, localeID, tempBuffer, sizeof(tempBuffer), status);
    } else {
        if (localeID==NULL) {
           localeID=uloc_getDefault();
        }
        tmpLocaleID=localeID;
    }

    /* Skip the language */
    ulocimp_getLanguage(tmpLocaleID, NULL, 0, &tmpLocaleID);
    if(_isIDSeparator(*tmpLocaleID)) {
        const char *scriptID;
        /* Skip the script if available */
        ulocimp_getScript(tmpLocaleID+1, NULL, 0, &scriptID);
        if(scriptID != tmpLocaleID+1) {
            /* Found optional script */
            tmpLocaleID = scriptID;
        }
        /* Skip the Country */
        if (_isIDSeparator(*tmpLocaleID)) {
            ulocimp_getCountry(tmpLocaleID+1, NULL, 0, &tmpLocaleID);
            if(_isIDSeparator(*tmpLocaleID)) {
                _getVariant(tmpLocaleID+1, *tmpLocaleID, NULL, 0);
            }
        }
    }

    /* keywords are located after '@' */
    if((tmpLocaleID = locale_getKeywordsStart(tmpLocaleID)) != NULL) {
        i=locale_getKeywords(tmpLocaleID+1, '@', keywords, keywordsCapacity, NULL, 0, NULL, FALSE, status);
    }

    if(i) {
        return uloc_openKeywordList(keywords, i, status);
    } else {
        return NULL;
    }
}


/* bit-flags for 'options' parameter of _canonicalize */
#define _ULOC_STRIP_KEYWORDS 0x2
#define _ULOC_CANONICALIZE   0x1

#define OPTION_SET(options, mask) ((options & mask) != 0)

static const char i_default[] = {'i', '-', 'd', 'e', 'f', 'a', 'u', 'l', 't'};
#define I_DEFAULT_LENGTH UPRV_LENGTHOF(i_default)

/**
 * Canonicalize the given localeID, to level 1 or to level 2,
 * depending on the options.  To specify level 1, pass in options=0.
 * To specify level 2, pass in options=_ULOC_CANONICALIZE.
 *
 * This is the code underlying uloc_getName and uloc_canonicalize.
 */
static int32_t
_canonicalize(const char* localeID,
              char* result,
              int32_t resultCapacity,
              uint32_t options,
              UErrorCode* err) {
    int32_t j, len, fieldCount=0, scriptSize=0, variantSize=0, nameCapacity;
    char localeBuffer[ULOC_FULLNAME_CAPACITY];
    char tempBuffer[ULOC_FULLNAME_CAPACITY];
    const char* origLocaleID;
    const char* tmpLocaleID;
    const char* keywordAssign = NULL;
    const char* separatorIndicator = NULL;
    char* name;
    char* variant = NULL; /* pointer into name, or NULL */

    if (U_FAILURE(*err)) {
        return 0;
    }

    if (_hasBCP47Extension(localeID)) {
        _ConvertBCP47(tmpLocaleID, localeID, tempBuffer, sizeof(tempBuffer), err);
    } else {
        if (localeID==NULL) {
           localeID=uloc_getDefault();
        }
        tmpLocaleID=localeID;
    }

    origLocaleID=tmpLocaleID;

    /* if we are doing a full canonicalization, then put results in
       localeBuffer, if necessary; otherwise send them to result. */
    if (/*OPTION_SET(options, _ULOC_CANONICALIZE) &&*/
        (result == NULL || resultCapacity < (int32_t)sizeof(localeBuffer))) {
        name = localeBuffer;
        nameCapacity = (int32_t)sizeof(localeBuffer);
    } else {
        name = result;
        nameCapacity = resultCapacity;
    }

    /* get all pieces, one after another, and separate with '_' */
    len=ulocimp_getLanguage(tmpLocaleID, name, nameCapacity, &tmpLocaleID);

    if(len == I_DEFAULT_LENGTH && uprv_strncmp(origLocaleID, i_default, len) == 0) {
        const char *d = uloc_getDefault();

        len = (int32_t)uprv_strlen(d);

        if (name != NULL) {
            uprv_memcpy(name, d, len);
        }
    } else if(_isIDSeparator(*tmpLocaleID)) {
        const char *scriptID;

        ++fieldCount;
        if(len<nameCapacity) {
            name[len]='_';
        }
        ++len;

        scriptSize=ulocimp_getScript(tmpLocaleID+1,
            (len<nameCapacity ? name+len : NULL), nameCapacity-len, &scriptID);
        if(scriptSize > 0) {
            /* Found optional script */
            tmpLocaleID = scriptID;
            ++fieldCount;
            len+=scriptSize;
            if (_isIDSeparator(*tmpLocaleID)) {
                /* If there is something else, then we add the _ */
                if(len<nameCapacity) {
                    name[len]='_';
                }
                ++len;
            }
        }

        if (_isIDSeparator(*tmpLocaleID)) {
            const char *cntryID;
            int32_t cntrySize = ulocimp_getCountry(tmpLocaleID+1,
                (len<nameCapacity ? name+len : NULL), nameCapacity-len, &cntryID);
            if (cntrySize > 0) {
                /* Found optional country */
                tmpLocaleID = cntryID;
                len+=cntrySize;
            }
            if(_isIDSeparator(*tmpLocaleID)) {
                /* If there is something else, then we add the _  if we found country before. */
                if (cntrySize >= 0 && ! _isIDSeparator(*(tmpLocaleID+1)) ) {
                    ++fieldCount;
                    if(len<nameCapacity) {
                        name[len]='_';
                    }
                    ++len;
                }

                variantSize = _getVariant(tmpLocaleID+1, *tmpLocaleID,
                    (len<nameCapacity ? name+len : NULL), nameCapacity-len);
                if (variantSize > 0) {
                    variant = len<nameCapacity ? name+len : NULL;
                    len += variantSize;
                    tmpLocaleID += variantSize + 1; /* skip '_' and variant */
                }
            }
        }
    }

    /* Copy POSIX-style charset specifier, if any [mr.utf8] */
    if (!OPTION_SET(options, _ULOC_CANONICALIZE) && *tmpLocaleID == '.') {
        UBool done = FALSE;
        do {
            char c = *tmpLocaleID;
            switch (c) {
            case 0:
            case '@':
                done = TRUE;
                break;
            default:
                if (len<nameCapacity) {
                    name[len] = c;
                }
                ++len;
                ++tmpLocaleID;
                break;
            }
        } while (!done);
    }

    /* Scan ahead to next '@' and determine if it is followed by '=' and/or ';'
       After this, tmpLocaleID either points to '@' or is NULL */
    if ((tmpLocaleID=locale_getKeywordsStart(tmpLocaleID))!=NULL) {
        keywordAssign = uprv_strchr(tmpLocaleID, '=');
        separatorIndicator = uprv_strchr(tmpLocaleID, ';');
    }

    /* Copy POSIX-style variant, if any [mr@FOO] */
    if (!OPTION_SET(options, _ULOC_CANONICALIZE) &&
        tmpLocaleID != NULL && keywordAssign == NULL) {
        for (;;) {
            char c = *tmpLocaleID;
            if (c == 0) {
                break;
            }
            if (len<nameCapacity) {
                name[len] = c;
            }
            ++len;
            ++tmpLocaleID;
        }
    }

    if (OPTION_SET(options, _ULOC_CANONICALIZE)) {
        /* Handle @FOO variant if @ is present and not followed by = */
        if (tmpLocaleID!=NULL && keywordAssign==NULL) {
            int32_t posixVariantSize;
            /* Add missing '_' if needed */
            if (fieldCount < 2 || (fieldCount < 3 && scriptSize > 0)) {
                do {
                    if(len<nameCapacity) {
                        name[len]='_';
                    }
                    ++len;
                    ++fieldCount;
                } while(fieldCount<2);
            }
            posixVariantSize = _getVariantEx(tmpLocaleID+1, '@', name+len, nameCapacity-len,
                                             (UBool)(variantSize > 0));
            if (posixVariantSize > 0) {
                if (variant == NULL) {
                    variant = name+len;
                }
                len += posixVariantSize;
                variantSize += posixVariantSize;
            }
        }

        /* Look up the ID in the canonicalization map */
        for (j=0; j<UPRV_LENGTHOF(CANONICALIZE_MAP); j++) {
            const char* id = CANONICALIZE_MAP[j].id;
            int32_t n = (int32_t)uprv_strlen(id);
            if (len == n && uprv_strncmp(name, id, n) == 0) {
                if (n == 0 && tmpLocaleID != NULL) {
                    break; /* Don't remap "" if keywords present */
                }
                len = _copyCount(name, nameCapacity, CANONICALIZE_MAP[j].canonicalID);
                break;
            }
        }
    }

    if (!OPTION_SET(options, _ULOC_STRIP_KEYWORDS)) {
        if (tmpLocaleID!=NULL && keywordAssign!=NULL &&
            (!separatorIndicator || separatorIndicator > keywordAssign)) {
            if(len<nameCapacity) {
                name[len]='@';
            }
            ++len;
            ++fieldCount;
            len += _getKeywords(tmpLocaleID+1, '@', (len<nameCapacity ? name+len : NULL), nameCapacity-len,
                                NULL, 0, NULL, TRUE, err);
        }
    }

    if (U_SUCCESS(*err) && result != NULL && name == localeBuffer) {
        uprv_strncpy(result, localeBuffer, (len > resultCapacity) ? resultCapacity : len);
    }

    return u_terminateChars(result, resultCapacity, len, err);
}

/* ### ID parsing API **************************************************/

U_CAPI int32_t  U_EXPORT2
uloc_getParent(const char*    localeID,
               char* parent,
               int32_t parentCapacity,
               UErrorCode* err)
{
    const char *lastUnderscore;
    int32_t i;

    if (U_FAILURE(*err))
        return 0;

    if (localeID == NULL)
        localeID = uloc_getDefault();

    lastUnderscore=uprv_strrchr(localeID, '_');
    if(lastUnderscore!=NULL) {
        i=(int32_t)(lastUnderscore-localeID);
    } else {
        i=0;
    }

    if (i > 0) {
        if (uprv_strnicmp(localeID, "und_", 4) == 0) {
            localeID += 3;
            i -= 3;
            uprv_memmove(parent, localeID, uprv_min(i, parentCapacity));
        } else if (parent != localeID) {
            uprv_memcpy(parent, localeID, uprv_min(i, parentCapacity));
        }
    }

    return u_terminateChars(parent, parentCapacity, i, err);
}

U_CAPI int32_t U_EXPORT2
uloc_getLanguage(const char*    localeID,
         char* language,
         int32_t languageCapacity,
         UErrorCode* err)
{
    /* uloc_getLanguage will return a 2 character iso-639 code if one exists. *CWB*/
    int32_t i=0;

    if (err==NULL || U_FAILURE(*err)) {
        return 0;
    }

    if(localeID==NULL) {
        localeID=uloc_getDefault();
    }

    i=ulocimp_getLanguage(localeID, language, languageCapacity, NULL);
    return u_terminateChars(language, languageCapacity, i, err);
}

U_CAPI int32_t U_EXPORT2
uloc_getScript(const char*    localeID,
         char* script,
         int32_t scriptCapacity,
         UErrorCode* err)
{
    int32_t i=0;

    if(err==NULL || U_FAILURE(*err)) {
        return 0;
    }

    if(localeID==NULL) {
        localeID=uloc_getDefault();
    }

    /* skip the language */
    ulocimp_getLanguage(localeID, NULL, 0, &localeID);
    if(_isIDSeparator(*localeID)) {
        i=ulocimp_getScript(localeID+1, script, scriptCapacity, NULL);
    }
    return u_terminateChars(script, scriptCapacity, i, err);
}

U_CAPI int32_t  U_EXPORT2
uloc_getCountry(const char* localeID,
            char* country,
            int32_t countryCapacity,
            UErrorCode* err)
{
    int32_t i=0;

    if(err==NULL || U_FAILURE(*err)) {
        return 0;
    }

    if(localeID==NULL) {
        localeID=uloc_getDefault();
    }

    /* Skip the language */
    ulocimp_getLanguage(localeID, NULL, 0, &localeID);
    if(_isIDSeparator(*localeID)) {
        const char *scriptID;
        /* Skip the script if available */
        ulocimp_getScript(localeID+1, NULL, 0, &scriptID);
        if(scriptID != localeID+1) {
            /* Found optional script */
            localeID = scriptID;
        }
        if(_isIDSeparator(*localeID)) {
            i=ulocimp_getCountry(localeID+1, country, countryCapacity, NULL);
        }
    }
    return u_terminateChars(country, countryCapacity, i, err);
}

U_CAPI int32_t  U_EXPORT2
uloc_getVariant(const char* localeID,
                char* variant,
                int32_t variantCapacity,
                UErrorCode* err)
{
    char tempBuffer[ULOC_FULLNAME_CAPACITY];
    const char* tmpLocaleID;
    int32_t i=0;

    if(err==NULL || U_FAILURE(*err)) {
        return 0;
    }

    if (_hasBCP47Extension(localeID)) {
        _ConvertBCP47(tmpLocaleID, localeID, tempBuffer, sizeof(tempBuffer), err);
    } else {
        if (localeID==NULL) {
           localeID=uloc_getDefault();
        }
        tmpLocaleID=localeID;
    }

    /* Skip the language */
    ulocimp_getLanguage(tmpLocaleID, NULL, 0, &tmpLocaleID);
    if(_isIDSeparator(*tmpLocaleID)) {
        const char *scriptID;
        /* Skip the script if available */
        ulocimp_getScript(tmpLocaleID+1, NULL, 0, &scriptID);
        if(scriptID != tmpLocaleID+1) {
            /* Found optional script */
            tmpLocaleID = scriptID;
        }
        /* Skip the Country */
        if (_isIDSeparator(*tmpLocaleID)) {
            const char *cntryID;
            ulocimp_getCountry(tmpLocaleID+1, NULL, 0, &cntryID);
            if (cntryID != tmpLocaleID+1) {
                /* Found optional country */
                tmpLocaleID = cntryID;
            }
            if(_isIDSeparator(*tmpLocaleID)) {
                /* If there was no country ID, skip a possible extra IDSeparator */
                if (tmpLocaleID != cntryID && _isIDSeparator(tmpLocaleID[1])) {
                    tmpLocaleID++;
                }
                i=_getVariant(tmpLocaleID+1, *tmpLocaleID, variant, variantCapacity);
            }
        }
    }

    /* removed by weiv. We don't want to handle POSIX variants anymore. Use canonicalization function */
    /* if we do not have a variant tag yet then try a POSIX variant after '@' */
/*
    if(!haveVariant && (localeID=uprv_strrchr(localeID, '@'))!=NULL) {
        i=_getVariant(localeID+1, '@', variant, variantCapacity);
    }
*/
    return u_terminateChars(variant, variantCapacity, i, err);
}

U_CAPI int32_t  U_EXPORT2
uloc_getName(const char* localeID,
             char* name,
             int32_t nameCapacity,
             UErrorCode* err)
{
    return _canonicalize(localeID, name, nameCapacity, 0, err);
}

U_CAPI int32_t  U_EXPORT2
uloc_getBaseName(const char* localeID,
                 char* name,
                 int32_t nameCapacity,
                 UErrorCode* err)
{
    return _canonicalize(localeID, name, nameCapacity, _ULOC_STRIP_KEYWORDS, err);
}

U_CAPI int32_t  U_EXPORT2
uloc_canonicalize(const char* localeID,
                  char* name,
                  int32_t nameCapacity,
                  UErrorCode* err)
{
    return _canonicalize(localeID, name, nameCapacity, _ULOC_CANONICALIZE, err);
}

U_CAPI const char*  U_EXPORT2
uloc_getISO3Language(const char* localeID)
{
    int16_t offset;
    char lang[ULOC_LANG_CAPACITY];
    UErrorCode err = U_ZERO_ERROR;

    if (localeID == NULL)
    {
        localeID = uloc_getDefault();
    }
    uloc_getLanguage(localeID, lang, ULOC_LANG_CAPACITY, &err);
    if (U_FAILURE(err))
        return "";
    offset = _findIndex(LANGUAGES, lang);
    if (offset < 0)
        return "";
    return LANGUAGES_3[offset];
}

U_CAPI const char*  U_EXPORT2
uloc_getISO3Country(const char* localeID)
{
    int16_t offset;
    char cntry[ULOC_LANG_CAPACITY];
    UErrorCode err = U_ZERO_ERROR;

    if (localeID == NULL)
    {
        localeID = uloc_getDefault();
    }
    uloc_getCountry(localeID, cntry, ULOC_LANG_CAPACITY, &err);
    if (U_FAILURE(err))
        return "";
    offset = _findIndex(COUNTRIES, cntry);
    if (offset < 0)
        return "";

    return COUNTRIES_3[offset];
}

U_CAPI uint32_t  U_EXPORT2
uloc_getLCID(const char* localeID)
{
    UErrorCode status = U_ZERO_ERROR;
    char       langID[ULOC_FULLNAME_CAPACITY];
    uint32_t   lcid = 0;

    /* Check for incomplete id. */
    if (!localeID || uprv_strlen(localeID) < 2) {
        return 0;
    }

    // First, attempt Windows platform lookup if available, but fall
    // through to catch any special cases (ICU vs Windows name differences).
    lcid = uprv_convertToLCIDPlatform(localeID, &status);
    if (U_FAILURE(status)) {
        return 0;
    }
    if (lcid > 0) {
        // Windows found an LCID, return that
        return lcid;
    }

    uloc_getLanguage(localeID, langID, sizeof(langID), &status);
    if (U_FAILURE(status) || status == U_STRING_NOT_TERMINATED_WARNING) {
        return 0;
    }

    if (uprv_strchr(localeID, '@')) {
        // uprv_convertToLCID does not support keywords other than collation.
        // Remove all keywords except collation.
        int32_t len;
        char collVal[ULOC_KEYWORDS_CAPACITY];
        char tmpLocaleID[ULOC_FULLNAME_CAPACITY];

        len = uloc_getKeywordValue(localeID, "collation", collVal,
            UPRV_LENGTHOF(collVal) - 1, &status);

        if (U_SUCCESS(status) && len > 0) {
            collVal[len] = 0;

            len = uloc_getBaseName(localeID, tmpLocaleID,
                UPRV_LENGTHOF(tmpLocaleID) - 1, &status);

            if (U_SUCCESS(status) && len > 0) {
                tmpLocaleID[len] = 0;

                len = uloc_setKeywordValue("collation", collVal, tmpLocaleID,
                    UPRV_LENGTHOF(tmpLocaleID) - len - 1, &status);

                if (U_SUCCESS(status) && len > 0) {
                    tmpLocaleID[len] = 0;
                    return uprv_convertToLCID(langID, tmpLocaleID, &status);
                }
            }
        }

        // fall through - all keywords are simply ignored
        status = U_ZERO_ERROR;
    }

    return uprv_convertToLCID(langID, localeID, &status);
}

U_CAPI int32_t U_EXPORT2
uloc_getLocaleForLCID(uint32_t hostid, char *locale, int32_t localeCapacity,
                UErrorCode *status)
{
    return uprv_convertToPosix(hostid, locale, localeCapacity, status);
}

/* ### Default locale **************************************************/

U_CAPI const char*  U_EXPORT2
uloc_getDefault()
{
    return locale_get_default();
}

U_CAPI void  U_EXPORT2
uloc_setDefault(const char*   newDefaultLocale,
             UErrorCode* err)
{
    if (U_FAILURE(*err))
        return;
    /* the error code isn't currently used for anything by this function*/

    /* propagate change to C++ */
    locale_set_default(newDefaultLocale);
}

/**
 * Returns a list of all 2-letter language codes defined in ISO 639.  This is a pointer
 * to an array of pointers to arrays of char.  All of these pointers are owned
 * by ICU-- do not delete them, and do not write through them.  The array is
 * terminated with a null pointer.
 */
U_CAPI const char* const*  U_EXPORT2
uloc_getISOLanguages()
{
    return LANGUAGES;
}

/**
 * Returns a list of all 2-letter country codes defined in ISO 639.  This is a
 * pointer to an array of pointers to arrays of char.  All of these pointers are
 * owned by ICU-- do not delete them, and do not write through them.  The array is
 * terminated with a null pointer.
 */
U_CAPI const char* const*  U_EXPORT2
uloc_getISOCountries()
{
    return COUNTRIES;
}


/* this function to be moved into cstring.c later */
static char gDecimal = 0;

static /* U_CAPI */
double
/* U_EXPORT2 */
_uloc_strtod(const char *start, char **end) {
    char *decimal;
    char *myEnd;
    char buf[30];
    double rv;
    if (!gDecimal) {
        char rep[5];
        /* For machines that decide to change the decimal on you,
        and try to be too smart with localization.
        This normally should be just a '.'. */
        sprintf(rep, "%+1.1f", 1.0);
        gDecimal = rep[2];
    }

    if(gDecimal == '.') {
        return uprv_strtod(start, end); /* fall through to OS */
    } else {
        uprv_strncpy(buf, start, 29);
        buf[29]=0;
        decimal = uprv_strchr(buf, '.');
        if(decimal) {
            *decimal = gDecimal;
        } else {
            return uprv_strtod(start, end); /* no decimal point */
        }
        rv = uprv_strtod(buf, &myEnd);
        if(end) {
            *end = (char*)(start+(myEnd-buf)); /* cast away const (to follow uprv_strtod API.) */
        }
        return rv;
    }
}

typedef struct {
    float q;
    int32_t dummy;  /* to avoid uninitialized memory copy from qsort */
    char locale[ULOC_FULLNAME_CAPACITY+1];
} _acceptLangItem;

static int32_t U_CALLCONV
uloc_acceptLanguageCompare(const void * /*context*/, const void *a, const void *b)
{
    const _acceptLangItem *aa = (const _acceptLangItem*)a;
    const _acceptLangItem *bb = (const _acceptLangItem*)b;

    int32_t rc = 0;
    if(bb->q < aa->q) {
        rc = -1;  /* A > B */
    } else if(bb->q > aa->q) {
        rc = 1;   /* A < B */
    } else {
        rc = 0;   /* A = B */
    }

    if(rc==0) {
        rc = uprv_stricmp(aa->locale, bb->locale);
    }

#if defined(ULOC_DEBUG)
    /*  fprintf(stderr, "a:[%s:%g], b:[%s:%g] -> %d\n",
    aa->locale, aa->q,
    bb->locale, bb->q,
    rc);*/
#endif

    return rc;
}

/*
mt-mt, ja;q=0.76, en-us;q=0.95, en;q=0.92, en-gb;q=0.89, fr;q=0.87, iu-ca;q=0.84, iu;q=0.82, ja-jp;q=0.79, mt;q=0.97, de-de;q=0.74, de;q=0.71, es;q=0.68, it-it;q=0.66, it;q=0.63, vi-vn;q=0.61, vi;q=0.58, nl-nl;q=0.55, nl;q=0.53
*/

U_CAPI int32_t U_EXPORT2
uloc_acceptLanguageFromHTTP(char *result, int32_t resultAvailable, UAcceptResult *outResult,
                            const char *httpAcceptLanguage,
                            UEnumeration* availableLocales,
                            UErrorCode *status)
{
  MaybeStackArray<_acceptLangItem, 4> items; // Struct for collecting items.
    char tmp[ULOC_FULLNAME_CAPACITY +1];
    int32_t n = 0;
    const char *itemEnd;
    const char *paramEnd;
    const char *s;
    const char *t;
    int32_t res;
    int32_t i;
    int32_t l = (int32_t)uprv_strlen(httpAcceptLanguage);

    if(U_FAILURE(*status)) {
        return -1;
    }

    for(s=httpAcceptLanguage;s&&*s;) {
        while(isspace(*s)) /* eat space at the beginning */
            s++;
        itemEnd=uprv_strchr(s,',');
        paramEnd=uprv_strchr(s,';');
        if(!itemEnd) {
            itemEnd = httpAcceptLanguage+l; /* end of string */
        }
        if(paramEnd && paramEnd<itemEnd) {
            /* semicolon (;) is closer than end (,) */
            t = paramEnd+1;
            if(*t=='q') {
                t++;
            }
            while(isspace(*t)) {
                t++;
            }
            if(*t=='=') {
                t++;
            }
            while(isspace(*t)) {
                t++;
            }
            items[n].q = (float)_uloc_strtod(t,NULL);
        } else {
            /* no semicolon - it's 1.0 */
            items[n].q = 1.0f;
            paramEnd = itemEnd;
        }
        items[n].dummy=0;
        /* eat spaces prior to semi */
        for(t=(paramEnd-1);(paramEnd>s)&&isspace(*t);t--)
            ;
        int32_t slen = static_cast<int32_t>(((t+1)-s));
        if(slen > ULOC_FULLNAME_CAPACITY) {
          *status = U_BUFFER_OVERFLOW_ERROR;
          return -1; // too big
        }
        uprv_strncpy(items[n].locale, s, slen);
        items[n].locale[slen]=0; // terminate
        int32_t clen = uloc_canonicalize(items[n].locale, tmp, UPRV_LENGTHOF(tmp)-1, status);
        if(U_FAILURE(*status)) return -1;
        if((clen!=slen) || (uprv_strncmp(items[n].locale, tmp, slen))) {
            // canonicalization had an effect- copy back
            uprv_strncpy(items[n].locale, tmp, clen);
            items[n].locale[clen] = 0; // terminate
        }
#if defined(ULOC_DEBUG)
        /*fprintf(stderr,"%d: s <%s> q <%g>\n", n, j[n].locale, j[n].q);*/
#endif
        n++;
        s = itemEnd;
        while(*s==',') { /* eat duplicate commas */
            s++;
        }
        if(n>=items.getCapacity()) { // If we need more items
          if(NULL == items.resize(items.getCapacity()*2, items.getCapacity())) {
              *status = U_MEMORY_ALLOCATION_ERROR;
              return -1;
          }
#if defined(ULOC_DEBUG)
          fprintf(stderr,"malloced at size %d\n", items.getCapacity());
#endif
        }
    }
    uprv_sortArray(items.getAlias(), n, sizeof(items[0]), uloc_acceptLanguageCompare, NULL, TRUE, status);
    if (U_FAILURE(*status)) {
        return -1;
    }
    LocalMemory<const char*> strs(NULL);
    if (strs.allocateInsteadAndReset(n) == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return -1;
    }
    for(i=0;i<n;i++) {
#if defined(ULOC_DEBUG)
        /*fprintf(stderr,"%d: s <%s> q <%g>\n", i, j[i].locale, j[i].q);*/
#endif
        strs[i]=items[i].locale;
    }
    res =  uloc_acceptLanguage(result, resultAvailable, outResult,
                               strs.getAlias(), n, availableLocales, status);
    return res;
}


U_CAPI int32_t U_EXPORT2
uloc_acceptLanguage(char *result, int32_t resultAvailable,
                    UAcceptResult *outResult, const char **acceptList,
                    int32_t acceptListCount,
                    UEnumeration* availableLocales,
                    UErrorCode *status)
{
    int32_t i,j;
    int32_t len;
    int32_t maxLen=0;
    char tmp[ULOC_FULLNAME_CAPACITY+1];
    const char *l;
    char **fallbackList;
    if(U_FAILURE(*status)) {
        return -1;
    }
    fallbackList = static_cast<char **>(uprv_malloc((size_t)(sizeof(fallbackList[0])*acceptListCount)));
    if(fallbackList==NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return -1;
    }
    for(i=0;i<acceptListCount;i++) {
#if defined(ULOC_DEBUG)
        fprintf(stderr,"%02d: %s\n", i, acceptList[i]);
#endif
        while((l=uenum_next(availableLocales, NULL, status)) != NULL) {
#if defined(ULOC_DEBUG)
            fprintf(stderr,"  %s\n", l);
#endif
            len = (int32_t)uprv_strlen(l);
            if(!uprv_strcmp(acceptList[i], l)) {
                if(outResult) {
                    *outResult = ULOC_ACCEPT_VALID;
                }
#if defined(ULOC_DEBUG)
                fprintf(stderr, "MATCH! %s\n", l);
#endif
                if(len>0) {
                    uprv_strncpy(result, l, uprv_min(len, resultAvailable));
                }
                for(j=0;j<i;j++) {
                    uprv_free(fallbackList[j]);
                }
                uprv_free(fallbackList);
                return u_terminateChars(result, resultAvailable, len, status);
            }
            if(len>maxLen) {
                maxLen = len;
            }
        }
        uenum_reset(availableLocales, status);
        /* save off parent info */
        if(uloc_getParent(acceptList[i], tmp, UPRV_LENGTHOF(tmp), status)!=0) {
            fallbackList[i] = uprv_strdup(tmp);
        } else {
            fallbackList[i]=0;
        }
    }

    for(maxLen--;maxLen>0;maxLen--) {
        for(i=0;i<acceptListCount;i++) {
            if(fallbackList[i] && ((int32_t)uprv_strlen(fallbackList[i])==maxLen)) {
#if defined(ULOC_DEBUG)
                fprintf(stderr,"Try: [%s]", fallbackList[i]);
#endif
                while((l=uenum_next(availableLocales, NULL, status)) != NULL) {
#if defined(ULOC_DEBUG)
                    fprintf(stderr,"  %s\n", l);
#endif
                    len = (int32_t)uprv_strlen(l);
                    if(!uprv_strcmp(fallbackList[i], l)) {
                        if(outResult) {
                            *outResult = ULOC_ACCEPT_FALLBACK;
                        }
#if defined(ULOC_DEBUG)
                        fprintf(stderr, "fallback MATCH! %s\n", l);
#endif
                        if(len>0) {
                            uprv_strncpy(result, l, uprv_min(len, resultAvailable));
                        }
                        for(j=0;j<acceptListCount;j++) {
                            uprv_free(fallbackList[j]);
                        }
                        uprv_free(fallbackList);
                        return u_terminateChars(result, resultAvailable, len, status);
                    }
                }
                uenum_reset(availableLocales, status);

                if(uloc_getParent(fallbackList[i], tmp, UPRV_LENGTHOF(tmp), status)!=0) {
                    uprv_free(fallbackList[i]);
                    fallbackList[i] = uprv_strdup(tmp);
                } else {
                    uprv_free(fallbackList[i]);
                    fallbackList[i]=0;
                }
            }
        }
        if(outResult) {
            *outResult = ULOC_ACCEPT_FAILED;
        }
    }
    for(i=0;i<acceptListCount;i++) {
        uprv_free(fallbackList[i]);
    }
    uprv_free(fallbackList);
    return -1;
}

U_CAPI const char* U_EXPORT2
uloc_toUnicodeLocaleKey(const char* keyword)
{
    const char* bcpKey = ulocimp_toBcpKey(keyword);
    if (bcpKey == NULL && ultag_isUnicodeLocaleKey(keyword, -1)) {
        // unknown keyword, but syntax is fine..
        return keyword;
    }
    return bcpKey;
}

U_CAPI const char* U_EXPORT2
uloc_toUnicodeLocaleType(const char* keyword, const char* value)
{
    const char* bcpType = ulocimp_toBcpType(keyword, value, NULL, NULL);
    if (bcpType == NULL && ultag_isUnicodeLocaleType(value, -1)) {
        // unknown keyword, but syntax is fine..
        return value;
    }
    return bcpType;
}

static UBool
isWellFormedLegacyKey(const char* legacyKey)
{
    const char* p = legacyKey;
    while (*p) {
        if (!UPRV_ISALPHANUM(*p)) {
            return FALSE;
        }
        p++;
    }
    return TRUE;
}

static UBool
isWellFormedLegacyType(const char* legacyType)
{
    const char* p = legacyType;
    int32_t alphaNumLen = 0;
    while (*p) {
        if (*p == '_' || *p == '/' || *p == '-') {
            if (alphaNumLen == 0) {
                return FALSE;
            }
            alphaNumLen = 0;
        } else if (UPRV_ISALPHANUM(*p)) {
            alphaNumLen++;
        } else {
            return FALSE;
        }
        p++;
    }
    return (alphaNumLen != 0);
}

U_CAPI const char* U_EXPORT2
uloc_toLegacyKey(const char* keyword)
{
    const char* legacyKey = ulocimp_toLegacyKey(keyword);
    if (legacyKey == NULL) {
        // Checks if the specified locale key is well-formed with the legacy locale syntax.
        //
        // Note:
        //  LDML/CLDR provides some definition of keyword syntax in
        //  * http://www.unicode.org/reports/tr35/#Unicode_locale_identifier and
        //  * http://www.unicode.org/reports/tr35/#Old_Locale_Extension_Syntax
        //  Keys can only consist of [0-9a-zA-Z].
        if (isWellFormedLegacyKey(keyword)) {
            return keyword;
        }
    }
    return legacyKey;
}

U_CAPI const char* U_EXPORT2
uloc_toLegacyType(const char* keyword, const char* value)
{
    const char* legacyType = ulocimp_toLegacyType(keyword, value, NULL, NULL);
    if (legacyType == NULL) {
        // Checks if the specified locale type is well-formed with the legacy locale syntax.
        //
        // Note:
        //  LDML/CLDR provides some definition of keyword syntax in
        //  * http://www.unicode.org/reports/tr35/#Unicode_locale_identifier and
        //  * http://www.unicode.org/reports/tr35/#Old_Locale_Extension_Syntax
        //  Values (types) can only consist of [0-9a-zA-Z], plus for legacy values
        //  we allow [/_-+] in the middle (e.g. "Etc/GMT+1", "Asia/Tel_Aviv")
        if (isWellFormedLegacyType(value)) {
            return value;
        }
    }
    return legacyType;
}

/*eof*/
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2014-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/
#include <algorithm>














 /* for InvChar functions */

static UHashtable* gLocExtKeyMap = NULL;
static icu::UInitOnce gLocExtKeyMapInitOnce = U_INITONCE_INITIALIZER;

// bit flags for special types
typedef enum {
    SPECIALTYPE_NONE = 0,
    SPECIALTYPE_CODEPOINTS = 1,
    SPECIALTYPE_REORDER_CODE = 2,
    SPECIALTYPE_RG_KEY_VALUE = 4
} SpecialType;

struct LocExtKeyData : public icu::UMemory {
    const char*     legacyId;
    const char*     bcpId;
    icu::LocalUHashtablePointer typeMap;
    uint32_t        specialTypes;
};

struct LocExtType : public icu::UMemory {
    const char*     legacyId;
    const char*     bcpId;
};

static icu::MemoryPool<icu::CharString>* gKeyTypeStringPool = NULL;
static icu::MemoryPool<LocExtKeyData>* gLocExtKeyDataEntries = NULL;
static icu::MemoryPool<LocExtType>* gLocExtTypeEntries = NULL;

U_CDECL_BEGIN

static UBool U_CALLCONV
uloc_key_type_cleanup(void) {
    if (gLocExtKeyMap != NULL) {
        uhash_close(gLocExtKeyMap);
        gLocExtKeyMap = NULL;
    }

    delete gLocExtKeyDataEntries;
    gLocExtKeyDataEntries = NULL;

    delete gLocExtTypeEntries;
    gLocExtTypeEntries = NULL;

    delete gKeyTypeStringPool;
    gKeyTypeStringPool = NULL;

    gLocExtKeyMapInitOnce.reset();
    return TRUE;
}

U_CDECL_END


static void U_CALLCONV
initFromResourceBundle(UErrorCode& sts) {
    U_NAMESPACE_USE
    ucln_common_registerCleanup(UCLN_COMMON_LOCALE_KEY_TYPE, uloc_key_type_cleanup);

    gLocExtKeyMap = uhash_open(uhash_hashIChars, uhash_compareIChars, NULL, &sts);

    LocalUResourceBundlePointer keyTypeDataRes(ures_openDirect(NULL, "keyTypeData", &sts));
    LocalUResourceBundlePointer keyMapRes(ures_getByKey(keyTypeDataRes.getAlias(), "keyMap", NULL, &sts));
    LocalUResourceBundlePointer typeMapRes(ures_getByKey(keyTypeDataRes.getAlias(), "typeMap", NULL, &sts));

    if (U_FAILURE(sts)) {
        return;
    }

    UErrorCode tmpSts = U_ZERO_ERROR;
    LocalUResourceBundlePointer typeAliasRes(ures_getByKey(keyTypeDataRes.getAlias(), "typeAlias", NULL, &tmpSts));
    tmpSts = U_ZERO_ERROR;
    LocalUResourceBundlePointer bcpTypeAliasRes(ures_getByKey(keyTypeDataRes.getAlias(), "bcpTypeAlias", NULL, &tmpSts));

    // initialize pools storing dynamically allocated objects
    gKeyTypeStringPool = new icu::MemoryPool<icu::CharString>;
    if (gKeyTypeStringPool == NULL) {
        sts = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    gLocExtKeyDataEntries = new icu::MemoryPool<LocExtKeyData>;
    if (gLocExtKeyDataEntries == NULL) {
        sts = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    gLocExtTypeEntries = new icu::MemoryPool<LocExtType>;
    if (gLocExtTypeEntries == NULL) {
        sts = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    // iterate through keyMap resource
    LocalUResourceBundlePointer keyMapEntry;

    while (ures_hasNext(keyMapRes.getAlias())) {
        keyMapEntry.adoptInstead(ures_getNextResource(keyMapRes.getAlias(), keyMapEntry.orphan(), &sts));
        if (U_FAILURE(sts)) {
            break;
        }
        const char* legacyKeyId = ures_getKey(keyMapEntry.getAlias());
        UnicodeString uBcpKeyId = ures_getUnicodeString(keyMapEntry.getAlias(), &sts);
        if (U_FAILURE(sts)) {
            break;
        }

        // empty value indicates that BCP key is same with the legacy key.
        const char* bcpKeyId = legacyKeyId;
        if (!uBcpKeyId.isEmpty()) {
            icu::CharString* bcpKeyIdBuf = gKeyTypeStringPool->create();
            if (bcpKeyIdBuf == NULL) {
                sts = U_MEMORY_ALLOCATION_ERROR;
                break;
            }
            bcpKeyIdBuf->appendInvariantChars(uBcpKeyId, sts);
            if (U_FAILURE(sts)) {
                break;
            }
            bcpKeyId = bcpKeyIdBuf->data();
        }

        UBool isTZ = uprv_strcmp(legacyKeyId, "timezone") == 0;

        UHashtable* typeDataMap = uhash_open(uhash_hashIChars, uhash_compareIChars, NULL, &sts);
        if (U_FAILURE(sts)) {
            break;
        }
        uint32_t specialTypes = SPECIALTYPE_NONE;

        LocalUResourceBundlePointer typeAliasResByKey;
        LocalUResourceBundlePointer bcpTypeAliasResByKey;

        if (typeAliasRes.isValid()) {
            tmpSts = U_ZERO_ERROR;
            typeAliasResByKey.adoptInstead(ures_getByKey(typeAliasRes.getAlias(), legacyKeyId, NULL, &tmpSts));
            if (U_FAILURE(tmpSts)) {
                typeAliasResByKey.orphan();
            }
        }
        if (bcpTypeAliasRes.isValid()) {
            tmpSts = U_ZERO_ERROR;
            bcpTypeAliasResByKey.adoptInstead(ures_getByKey(bcpTypeAliasRes.getAlias(), bcpKeyId, NULL, &tmpSts));
            if (U_FAILURE(tmpSts)) {
                bcpTypeAliasResByKey.orphan();
            }
        }

        // look up type map for the key, and walk through the mapping data
        tmpSts = U_ZERO_ERROR;
        LocalUResourceBundlePointer typeMapResByKey(ures_getByKey(typeMapRes.getAlias(), legacyKeyId, NULL, &tmpSts));
        if (U_FAILURE(tmpSts)) {
            // type map for each key must exist
            UPRV_UNREACHABLE;
        } else {
            LocalUResourceBundlePointer typeMapEntry;

            while (ures_hasNext(typeMapResByKey.getAlias())) {
                typeMapEntry.adoptInstead(ures_getNextResource(typeMapResByKey.getAlias(), typeMapEntry.orphan(), &sts));
                if (U_FAILURE(sts)) {
                    break;
                }
                const char* legacyTypeId = ures_getKey(typeMapEntry.getAlias());

                // special types
                if (uprv_strcmp(legacyTypeId, "CODEPOINTS") == 0) {
                    specialTypes |= SPECIALTYPE_CODEPOINTS;
                    continue;
                }
                if (uprv_strcmp(legacyTypeId, "REORDER_CODE") == 0) {
                    specialTypes |= SPECIALTYPE_REORDER_CODE;
                    continue;
                }
                if (uprv_strcmp(legacyTypeId, "RG_KEY_VALUE") == 0) {
                    specialTypes |= SPECIALTYPE_RG_KEY_VALUE;
                    continue;
                }

                if (isTZ) {
                    // a timezone key uses a colon instead of a slash in the resource.
                    // e.g. America:Los_Angeles
                    if (uprv_strchr(legacyTypeId, ':') != NULL) {
                        icu::CharString* legacyTypeIdBuf =
                                gKeyTypeStringPool->create(legacyTypeId, sts);
                        if (legacyTypeIdBuf == NULL) {
                            sts = U_MEMORY_ALLOCATION_ERROR;
                            break;
                        }
                        if (U_FAILURE(sts)) {
                            break;
                        }
                        std::replace(
                                legacyTypeIdBuf->data(),
                                legacyTypeIdBuf->data() + legacyTypeIdBuf->length(),
                                ':', '/');
                        legacyTypeId = legacyTypeIdBuf->data();
                    }
                }

                UnicodeString uBcpTypeId = ures_getUnicodeString(typeMapEntry.getAlias(), &sts);
                if (U_FAILURE(sts)) {
                    break;
                }

                // empty value indicates that BCP type is same with the legacy type.
                const char* bcpTypeId = legacyTypeId;
                if (!uBcpTypeId.isEmpty()) {
                    icu::CharString* bcpTypeIdBuf = gKeyTypeStringPool->create();
                    if (bcpTypeIdBuf == NULL) {
                        sts = U_MEMORY_ALLOCATION_ERROR;
                        break;
                    }
                    bcpTypeIdBuf->appendInvariantChars(uBcpTypeId, sts);
                    if (U_FAILURE(sts)) {
                        break;
                    }
                    bcpTypeId = bcpTypeIdBuf->data();
                }

                // Note: legacy type value should never be
                // equivalent to bcp type value of a different
                // type under the same key. So we use a single
                // map for lookup.
                LocExtType* t = gLocExtTypeEntries->create();
                if (t == NULL) {
                    sts = U_MEMORY_ALLOCATION_ERROR;
                    break;
                }
                t->bcpId = bcpTypeId;
                t->legacyId = legacyTypeId;

                uhash_put(typeDataMap, (void*)legacyTypeId, t, &sts);
                if (bcpTypeId != legacyTypeId) {
                    // different type value
                    uhash_put(typeDataMap, (void*)bcpTypeId, t, &sts);
                }
                if (U_FAILURE(sts)) {
                    break;
                }

                // also put aliases in the map
                if (typeAliasResByKey.isValid()) {
                    LocalUResourceBundlePointer typeAliasDataEntry;

                    ures_resetIterator(typeAliasResByKey.getAlias());
                    while (ures_hasNext(typeAliasResByKey.getAlias()) && U_SUCCESS(sts)) {
                        int32_t toLen;
                        typeAliasDataEntry.adoptInstead(ures_getNextResource(typeAliasResByKey.getAlias(), typeAliasDataEntry.orphan(), &sts));
                        const UChar* to = ures_getString(typeAliasDataEntry.getAlias(), &toLen, &sts);
                        if (U_FAILURE(sts)) {
                            break;
                        }
                        // check if this is an alias of canoncal legacy type
                        if (uprv_compareInvWithUChar(NULL, legacyTypeId, -1, to, toLen) == 0) {
                            const char* from = ures_getKey(typeAliasDataEntry.getAlias());
                            if (isTZ) {
                                // replace colon with slash if necessary
                                if (uprv_strchr(from, ':') != NULL) {
                                    icu::CharString* fromBuf =
                                            gKeyTypeStringPool->create(from, sts);
                                    if (fromBuf == NULL) {
                                        sts = U_MEMORY_ALLOCATION_ERROR;
                                        break;
                                    }
                                    if (U_FAILURE(sts)) {
                                        break;
                                    }
                                    std::replace(
                                            fromBuf->data(),
                                            fromBuf->data() + fromBuf->length(),
                                            ':', '/');
                                    from = fromBuf->data();
                                }
                            }
                            uhash_put(typeDataMap, (void*)from, t, &sts);
                        }
                    }
                    if (U_FAILURE(sts)) {
                        break;
                    }
                }

                if (bcpTypeAliasResByKey.isValid()) {
                    LocalUResourceBundlePointer bcpTypeAliasDataEntry;

                    ures_resetIterator(bcpTypeAliasResByKey.getAlias());
                    while (ures_hasNext(bcpTypeAliasResByKey.getAlias()) && U_SUCCESS(sts)) {
                        int32_t toLen;
                        bcpTypeAliasDataEntry.adoptInstead(ures_getNextResource(bcpTypeAliasResByKey.getAlias(), bcpTypeAliasDataEntry.orphan(), &sts));
                        const UChar* to = ures_getString(bcpTypeAliasDataEntry.getAlias(), &toLen, &sts);
                        if (U_FAILURE(sts)) {
                            break;
                        }
                        // check if this is an alias of bcp type
                        if (uprv_compareInvWithUChar(NULL, bcpTypeId, -1, to, toLen) == 0) {
                            const char* from = ures_getKey(bcpTypeAliasDataEntry.getAlias());
                            uhash_put(typeDataMap, (void*)from, t, &sts);
                        }
                    }
                    if (U_FAILURE(sts)) {
                        break;
                    }
                }
            }
        }
        if (U_FAILURE(sts)) {
            break;
        }

        LocExtKeyData* keyData = gLocExtKeyDataEntries->create();
        if (keyData == NULL) {
            sts = U_MEMORY_ALLOCATION_ERROR;
            break;
        }
        keyData->bcpId = bcpKeyId;
        keyData->legacyId = legacyKeyId;
        keyData->specialTypes = specialTypes;
        keyData->typeMap.adoptInstead(typeDataMap);

        uhash_put(gLocExtKeyMap, (void*)legacyKeyId, keyData, &sts);
        if (legacyKeyId != bcpKeyId) {
            // different key value
            uhash_put(gLocExtKeyMap, (void*)bcpKeyId, keyData, &sts);
        }
        if (U_FAILURE(sts)) {
            break;
        }
    }
}

static UBool
init() {
    UErrorCode sts = U_ZERO_ERROR;
    umtx_initOnce(gLocExtKeyMapInitOnce, &initFromResourceBundle, sts);
    if (U_FAILURE(sts)) {
        return FALSE;
    }
    return TRUE;
}

static UBool
isSpecialTypeCodepoints(const char* val) {
    int32_t subtagLen = 0;
    const char* p = val;
    while (*p) {
        if (*p == '-') {
            if (subtagLen < 4 || subtagLen > 6) {
                return FALSE;
            }
            subtagLen = 0;
        } else if ((*p >= '0' && *p <= '9') ||
                    (*p >= 'A' && *p <= 'F') || // A-F/a-f are contiguous
                    (*p >= 'a' && *p <= 'f')) { // also in EBCDIC
            subtagLen++;
        } else {
            return FALSE;
        }
        p++;
    }
    return (subtagLen >= 4 && subtagLen <= 6);
}

static UBool
isSpecialTypeReorderCode(const char* val) {
    int32_t subtagLen = 0;
    const char* p = val;
    while (*p) {
        if (*p == '-') {
            if (subtagLen < 3 || subtagLen > 8) {
                return FALSE;
            }
            subtagLen = 0;
        } else if (uprv_isASCIILetter(*p)) {
            subtagLen++;
        } else {
            return FALSE;
        }
        p++;
    }
    return (subtagLen >=3 && subtagLen <=8);
}

static UBool
isSpecialTypeRgKeyValue(const char* val) {
    int32_t subtagLen = 0;
    const char* p = val;
    while (*p) {
        if ( (subtagLen < 2 && uprv_isASCIILetter(*p)) ||
                    (subtagLen >= 2 && (*p == 'Z' || *p == 'z')) ) {
            subtagLen++;
        } else {
            return FALSE;
        }
        p++;
    }
    return (subtagLen == 6);
}

U_CFUNC const char*
ulocimp_toBcpKey(const char* key) {
    if (!init()) {
        return NULL;
    }

    LocExtKeyData* keyData = (LocExtKeyData*)uhash_get(gLocExtKeyMap, key);
    if (keyData != NULL) {
        return keyData->bcpId;
    }
    return NULL;
}

U_CFUNC const char*
ulocimp_toLegacyKey(const char* key) {
    if (!init()) {
        return NULL;
    }

    LocExtKeyData* keyData = (LocExtKeyData*)uhash_get(gLocExtKeyMap, key);
    if (keyData != NULL) {
        return keyData->legacyId;
    }
    return NULL;
}

U_CFUNC const char*
ulocimp_toBcpType(const char* key, const char* type, UBool* isKnownKey, UBool* isSpecialType) {
    if (isKnownKey != NULL) {
        *isKnownKey = FALSE;
    }
    if (isSpecialType != NULL) {
        *isSpecialType = FALSE;
    }

    if (!init()) {
        return NULL;
    }

    LocExtKeyData* keyData = (LocExtKeyData*)uhash_get(gLocExtKeyMap, key);
    if (keyData != NULL) {
        if (isKnownKey != NULL) {
            *isKnownKey = TRUE;
        }
        LocExtType* t = (LocExtType*)uhash_get(keyData->typeMap.getAlias(), type);
        if (t != NULL) {
            return t->bcpId;
        }
        if (keyData->specialTypes != SPECIALTYPE_NONE) {
            UBool matched = FALSE;
            if (keyData->specialTypes & SPECIALTYPE_CODEPOINTS) {
                matched = isSpecialTypeCodepoints(type);
            }
            if (!matched && keyData->specialTypes & SPECIALTYPE_REORDER_CODE) {
                matched = isSpecialTypeReorderCode(type);
            }
            if (!matched && keyData->specialTypes & SPECIALTYPE_RG_KEY_VALUE) {
                matched = isSpecialTypeRgKeyValue(type);
            }
            if (matched) {
                if (isSpecialType != NULL) {
                    *isSpecialType = TRUE;
                }
                return type;
            }
        }
    }
    return NULL;
}


U_CFUNC const char*
ulocimp_toLegacyType(const char* key, const char* type, UBool* isKnownKey, UBool* isSpecialType) {
    if (isKnownKey != NULL) {
        *isKnownKey = FALSE;
    }
    if (isSpecialType != NULL) {
        *isSpecialType = FALSE;
    }

    if (!init()) {
        return NULL;
    }

    LocExtKeyData* keyData = (LocExtKeyData*)uhash_get(gLocExtKeyMap, key);
    if (keyData != NULL) {
        if (isKnownKey != NULL) {
            *isKnownKey = TRUE;
        }
        LocExtType* t = (LocExtType*)uhash_get(keyData->typeMap.getAlias(), type);
        if (t != NULL) {
            return t->legacyId;
        }
        if (keyData->specialTypes != SPECIALTYPE_NONE) {
            UBool matched = FALSE;
            if (keyData->specialTypes & SPECIALTYPE_CODEPOINTS) {
                matched = isSpecialTypeCodepoints(type);
            }
            if (!matched && keyData->specialTypes & SPECIALTYPE_REORDER_CODE) {
                matched = isSpecialTypeReorderCode(type);
            }
            if (!matched && keyData->specialTypes & SPECIALTYPE_RG_KEY_VALUE) {
                matched = isSpecialTypeRgKeyValue(type);
            }
            if (matched) {
                if (isSpecialType != NULL) {
                    *isSpecialType = TRUE;
                }
                return type;
            }
        }
    }
    return NULL;
}

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2009-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/


















/* struct holding a single variant */
typedef struct VariantListEntry {
    const char              *variant;
    struct VariantListEntry *next;
} VariantListEntry;

/* struct holding a single attribute value */
struct AttributeListEntry : public icu::UMemory {
    const char              *attribute;
    struct AttributeListEntry *next;
};

/* struct holding a single extension */
struct ExtensionListEntry : public icu::UMemory {
    const char                  *key;
    const char                  *value;
    struct ExtensionListEntry   *next;
};

#define MAXEXTLANG 3
typedef struct ULanguageTag {
    char                *buf;   /* holding parsed subtags */
    const char          *language;
    const char          *extlang[MAXEXTLANG];
    const char          *script;
    const char          *region;
    VariantListEntry    *variants;
    ExtensionListEntry  *extensions;
    const char          *privateuse;
    const char          *grandfathered;
} ULanguageTag;

#define MINLEN 2
#define SEP '-'
#define PRIVATEUSE 'x'
#define LDMLEXT 'u'

#define LOCALE_SEP '_'
#define LOCALE_EXT_SEP '@'
#define LOCALE_KEYWORD_SEP ';'
#define LOCALE_KEY_TYPE_SEP '='

#define ISALPHA(c) uprv_isASCIILetter(c)
#define ISNUMERIC(c) ((c)>='0' && (c)<='9')

static const char EMPTY[] = "";
static const char LANG_UND[] = "und";
static const char PRIVATEUSE_KEY[] = "x";
static const char _POSIX[] = "_POSIX";
static const char POSIX_KEY[] = "va";
static const char POSIX_VALUE[] = "posix";
static const char LOCALE_ATTRIBUTE_KEY[] = "attribute";
static const char PRIVUSE_VARIANT_PREFIX[] = "lvariant";
static const char LOCALE_TYPE_YES[] = "yes";

#define LANG_UND_LEN 3

/*
 Updated on 2018-09-12 from
 https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry .

 This table has 2 parts. The parts for Grandfathered tags is generated by the
 following scripts from the IANA language tag registry.

 curl  https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry |\
 egrep -A 7 'Type: grandfathered' | \
 egrep 'Tag|Prefe' | grep -B1 'Preferred' | grep -v '^--' | \
 awk -n '/Tag/ {printf("    \"%s\", ", $2);} /Preferred/ {printf("\"%s\",\n", $2);}' |\
 tr 'A-Z' 'a-z'


 The 2nd part is made of five ICU-specific entries. They're kept for
 the backward compatibility for now, even though there are no preferred
 values. They may have to be removed for the strict BCP 47 compliance.

*/
static const char* const GRANDFATHERED[] = {
/*  grandfathered   preferred */
    "art-lojban",   "jbo",
    "en-gb-oed",    "en-gb-oxendict",
    "i-ami",        "ami",
    "i-bnn",        "bnn",
    "i-hak",        "hak",
    "i-klingon",    "tlh",
    "i-lux",        "lb",
    "i-navajo",     "nv",
    "i-pwn",        "pwn",
    "i-tao",        "tao",
    "i-tay",        "tay",
    "i-tsu",        "tsu",
    "no-bok",       "nb",
    "no-nyn",       "nn",
    "sgn-be-fr",    "sfb",
    "sgn-be-nl",    "vgt",
    "sgn-ch-de",    "sgg",
    "zh-guoyu",     "cmn",
    "zh-hakka",     "hak",
    "zh-min-nan",   "nan",
    "zh-xiang",     "hsn",

    // Grandfathered tags with no preferred value in the IANA
    // registry. Kept for now for the backward compatibility
    // because ICU has mapped them this way.
    "cel-gaulish",  "xtg-x-cel-gaulish",
    "i-default",    "en-x-i-default",
    "i-enochian",   "und-x-i-enochian",
    "i-mingo",      "see-x-i-mingo",
    "zh-min",       "nan-x-zh-min",
};

/*
 Updated on 2018-09-12 from
 https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry .

 The table lists redundant tags with preferred value in the IANA languate tag registry.
 It's generated with the following command:

 curl  https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry |\
 grep 'Type: redundant' -A 5 | egrep '^(Tag:|Prefer)' | grep -B1 'Preferred' | \
 awk -n '/Tag/ {printf("    \"%s\",       ", $2);} /Preferred/ {printf("\"%s\",\n", $2);}' | \
 tr 'A-Z' 'a-z'

 In addition, ja-latn-hepburn-heploc is mapped to ja-latn-alalc97 because
 a variant tag 'hepburn-heploc' has the preferred subtag, 'alaic97'.
*/

static const char* const REDUNDANT[] = {
//  redundant       preferred
    "sgn-br",       "bzs",
    "sgn-co",       "csn",
    "sgn-de",       "gsg",
    "sgn-dk",       "dsl",
    "sgn-es",       "ssp",
    "sgn-fr",       "fsl",
    "sgn-gb",       "bfi",
    "sgn-gr",       "gss",
    "sgn-ie",       "isg",
    "sgn-it",       "ise",
    "sgn-jp",       "jsl",
    "sgn-mx",       "mfs",
    "sgn-ni",       "ncs",
    "sgn-nl",       "dse",
    "sgn-no",       "nsl",
    "sgn-pt",       "psr",
    "sgn-se",       "swl",
    "sgn-us",       "ase",
    "sgn-za",       "sfs",
    "zh-cmn",       "cmn",
    "zh-cmn-hans",  "cmn-hans",
    "zh-cmn-hant",  "cmn-hant",
    "zh-gan",       "gan",
    "zh-wuu",       "wuu",
    "zh-yue",       "yue",

    // variant tag with preferred value
    "ja-latn-hepburn-heploc", "ja-latn-alalc97",
};

/*
  Updated on 2018-09-12 from
  https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry .

  grep 'Type: language' -A 7 language-subtag-registry  | egrep 'Subtag|Prefe' | \
  grep -B1 'Preferred' | grep -v '^--' | \
  awk -n '/Subtag/ {printf("    \"%s\",       ", $2);} /Preferred/ {printf("\"%s\",\n", $2);}'

  Make sure that 2-letter language subtags come before 3-letter subtags.
*/
static const char DEPRECATEDLANGS[][4] = {
/*  deprecated  new */
    "in",       "id",
    "iw",       "he",
    "ji",       "yi",
    "jw",       "jv",
    "mo",       "ro",
    "aam",       "aas",
    "adp",       "dz",
    "aue",       "ktz",
    "ayx",       "nun",
    "bgm",       "bcg",
    "bjd",       "drl",
    "ccq",       "rki",
    "cjr",       "mom",
    "cka",       "cmr",
    "cmk",       "xch",
    "coy",       "pij",
    "cqu",       "quh",
    "drh",       "khk",
    "drw",       "prs",
    "gav",       "dev",
    "gfx",       "vaj",
    "ggn",       "gvr",
    "gti",       "nyc",
    "guv",       "duz",
    "hrr",       "jal",
    "ibi",       "opa",
    "ilw",       "gal",
    "jeg",       "oyb",
    "kgc",       "tdf",
    "kgh",       "kml",
    "koj",       "kwv",
    "krm",       "bmf",
    "ktr",       "dtp",
    "kvs",       "gdj",
    "kwq",       "yam",
    "kxe",       "tvd",
    "kzj",       "dtp",
    "kzt",       "dtp",
    "lii",       "raq",
    "lmm",       "rmx",
    "meg",       "cir",
    "mst",       "mry",
    "mwj",       "vaj",
    "myt",       "mry",
    "nad",       "xny",
    "ncp",       "kdz",
    "nnx",       "ngv",
    "nts",       "pij",
    "oun",       "vaj",
    "pcr",       "adx",
    "pmc",       "huw",
    "pmu",       "phr",
    "ppa",       "bfy",
    "ppr",       "lcq",
    "pry",       "prt",
    "puz",       "pub",
    "sca",       "hle",
    "skk",       "oyb",
    "tdu",       "dtp",
    "thc",       "tpo",
    "thx",       "oyb",
    "tie",       "ras",
    "tkk",       "twm",
    "tlw",       "weo",
    "tmp",       "tyj",
    "tne",       "kak",
    "tnf",       "prs",
    "tsf",       "taj",
    "uok",       "ema",
    "xba",       "cax",
    "xia",       "acn",
    "xkh",       "waw",
    "xsj",       "suj",
    "ybd",       "rki",
    "yma",       "lrr",
    "ymt",       "mtm",
    "yos",       "zom",
    "yuu",       "yug",
};

/*
  Updated on 2018-04-24 from

  curl  https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry | \
  grep 'Type: region' -A 7 | egrep 'Subtag|Prefe' | \
  grep -B1 'Preferred' | \
  awk -n '/Subtag/ {printf("    \"%s\",       ", $2);} /Preferred/ {printf("\"%s\",\n", $2);}'
*/
static const char DEPRECATEDREGIONS[][3] = {
/*  deprecated  new */
    "BU",       "MM",
    "DD",       "DE",
    "FX",       "FR",
    "TP",       "TL",
    "YD",       "YE",
    "ZR",       "CD",
};

/*
* -------------------------------------------------
*
* These ultag_ functions may be exposed as APIs later
*
* -------------------------------------------------
*/

static ULanguageTag*
ultag_parse(const char* tag, int32_t tagLen, int32_t* parsedLen, UErrorCode* status);

static void
ultag_close(ULanguageTag* langtag);

static const char*
ultag_getLanguage(const ULanguageTag* langtag);

#if 0
static const char*
ultag_getJDKLanguage(const ULanguageTag* langtag);
#endif

static const char*
ultag_getExtlang(const ULanguageTag* langtag, int32_t idx);

static int32_t
ultag_getExtlangSize(const ULanguageTag* langtag);

static const char*
ultag_getScript(const ULanguageTag* langtag);

static const char*
ultag_getRegion(const ULanguageTag* langtag);

static const char*
ultag_getVariant(const ULanguageTag* langtag, int32_t idx);

static int32_t
ultag_getVariantsSize(const ULanguageTag* langtag);

static const char*
ultag_getExtensionKey(const ULanguageTag* langtag, int32_t idx);

static const char*
ultag_getExtensionValue(const ULanguageTag* langtag, int32_t idx);

static int32_t
ultag_getExtensionsSize(const ULanguageTag* langtag);

static const char*
ultag_getPrivateUse(const ULanguageTag* langtag);

#if 0
static const char*
ultag_getGrandfathered(const ULanguageTag* langtag);
#endif

U_NAMESPACE_BEGIN

/**
 * \class LocalULanguageTagPointer
 * "Smart pointer" class, closes a ULanguageTag via ultag_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @internal
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalULanguageTagPointer, ULanguageTag, ultag_close);

U_NAMESPACE_END

/*
* -------------------------------------------------
*
* Language subtag syntax validation functions
*
* -------------------------------------------------
*/

static UBool
_isAlphaString(const char* s, int32_t len) {
    int32_t i;
    for (i = 0; i < len; i++) {
        if (!ISALPHA(*(s + i))) {
            return FALSE;
        }
    }
    return TRUE;
}

static UBool
_isNumericString(const char* s, int32_t len) {
    int32_t i;
    for (i = 0; i < len; i++) {
        if (!ISNUMERIC(*(s + i))) {
            return FALSE;
        }
    }
    return TRUE;
}

static UBool
_isAlphaNumericString(const char* s, int32_t len) {
    int32_t i;
    for (i = 0; i < len; i++) {
        if (!ISALPHA(*(s + i)) && !ISNUMERIC(*(s + i))) {
            return FALSE;
        }
    }
    return TRUE;
}

static UBool
_isAlphaNumericStringLimitedLength(const char* s, int32_t len, int32_t min, int32_t max) {
    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }
    if (len >= min && len <= max && _isAlphaNumericString(s, len)) {
        return TRUE;
    }
    return FALSE;
}

U_CFUNC UBool
ultag_isLanguageSubtag(const char* s, int32_t len) {
    /*
     * unicode_language_subtag = alpha{2,3} | alpha{5,8};
     * NOTE: Per ICUTC 2019/01/23- accepting alpha 4
     * See ICU-20372
     */
    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }
    if (len >= 2 && len <= 8 && _isAlphaString(s, len)) {
        return TRUE;
    }
    return FALSE;
}

static UBool
_isExtlangSubtag(const char* s, int32_t len) {
    /*
     * extlang       = 3ALPHA              ; selected ISO 639 codes
     *                 *2("-" 3ALPHA)      ; permanently reserved
     */
    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }
    if (len == 3 && _isAlphaString(s, len)) {
        return TRUE;
    }
    return FALSE;
}

U_CFUNC UBool
ultag_isScriptSubtag(const char* s, int32_t len) {
    /*
     * script        = 4ALPHA              ; ISO 15924 code
     */
    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }
    if (len == 4 && _isAlphaString(s, len)) {
        return TRUE;
    }
    return FALSE;
}

U_CFUNC UBool
ultag_isRegionSubtag(const char* s, int32_t len) {
    /*
     * region        = 2ALPHA              ; ISO 3166-1 code
     *               / 3DIGIT              ; UN M.49 code
     */
    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }
    if (len == 2 && _isAlphaString(s, len)) {
        return TRUE;
    }
    if (len == 3 && _isNumericString(s, len)) {
        return TRUE;
    }
    return FALSE;
}

static UBool
_isVariantSubtag(const char* s, int32_t len) {
    /*
     * variant       = 5*8alphanum         ; registered variants
     *               / (DIGIT 3alphanum)
     */
    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }
    if (_isAlphaNumericStringLimitedLength(s, len, 5, 8)) {
        return TRUE;
    }
    if (len == 4 && ISNUMERIC(*s) && _isAlphaNumericString(s + 1, 3)) {
        return TRUE;
    }
    return FALSE;
}

static UBool
_isSepListOf(UBool (*test)(const char*, int32_t), const char* s, int32_t len) {
    const char *p = s;
    const char *pSubtag = NULL;

    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }

    while ((p - s) < len) {
        if (*p == SEP) {
            if (pSubtag == NULL) {
                return FALSE;
            }
            if (!test(pSubtag, (int32_t)(p - pSubtag))) {
                return FALSE;
            }
            pSubtag = NULL;
        } else if (pSubtag == NULL) {
            pSubtag = p;
        }
        p++;
    }
    if (pSubtag == NULL) {
        return FALSE;
    }
    return test(pSubtag, (int32_t)(p - pSubtag));
}

U_CFUNC UBool
ultag_isVariantSubtags(const char* s, int32_t len) {
    return _isSepListOf(&_isVariantSubtag, s, len);
}

// This is for the ICU-specific "lvariant" handling.
static UBool
_isPrivateuseVariantSubtag(const char* s, int32_t len) {
    /*
     * variant       = 1*8alphanum         ; registered variants
     *               / (DIGIT 3alphanum)
     */
    return _isAlphaNumericStringLimitedLength(s, len , 1, 8);
}

static UBool
_isExtensionSingleton(const char* s, int32_t len) {
    /*
     * extension     = singleton 1*("-" (2*8alphanum))
     *
     * singleton     = DIGIT               ; 0 - 9
     *               / %x41-57             ; A - W
     *               / %x59-5A             ; Y - Z
     *               / %x61-77             ; a - w
     *               / %x79-7A             ; y - z
     */
    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }
    if (len == 1 && (ISALPHA(*s) || ISNUMERIC(*s)) && (uprv_tolower(*s) != PRIVATEUSE)) {
        return TRUE;
    }
    return FALSE;
}

static UBool
_isExtensionSubtag(const char* s, int32_t len) {
    /*
     * extension     = singleton 1*("-" (2*8alphanum))
     */
    return _isAlphaNumericStringLimitedLength(s, len, 2, 8);
}

U_CFUNC UBool
ultag_isExtensionSubtags(const char* s, int32_t len) {
    return _isSepListOf(&_isExtensionSubtag, s, len);
}

static UBool
_isPrivateuseValueSubtag(const char* s, int32_t len) {
    /*
     * privateuse    = "x" 1*("-" (1*8alphanum))
     */
    return _isAlphaNumericStringLimitedLength(s, len, 1, 8);
}

U_CFUNC UBool
ultag_isPrivateuseValueSubtags(const char* s, int32_t len) {
    return _isSepListOf(&_isPrivateuseValueSubtag, s, len);
}

U_CFUNC UBool
ultag_isUnicodeLocaleAttribute(const char* s, int32_t len) {
    /*
     * attribute = alphanum{3,8} ;
     */
    return _isAlphaNumericStringLimitedLength(s, len , 3, 8);
}

U_CFUNC UBool
ultag_isUnicodeLocaleAttributes(const char* s, int32_t len) {
    return _isSepListOf(&ultag_isUnicodeLocaleAttribute, s, len);
}

U_CFUNC UBool
ultag_isUnicodeLocaleKey(const char* s, int32_t len) {
    /*
     * key = alphanum alpha ;
     */
    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }
    if (len == 2 && (ISALPHA(*s) || ISNUMERIC(*s)) && ISALPHA(s[1])) {
        return TRUE;
    }
    return FALSE;
}

U_CFUNC UBool
_isUnicodeLocaleTypeSubtag(const char*s, int32_t len) {
    /*
     * alphanum{3,8}
     */
    return _isAlphaNumericStringLimitedLength(s, len , 3, 8);
}

U_CFUNC UBool
ultag_isUnicodeLocaleType(const char*s, int32_t len) {
    /*
     * type = alphanum{3,8} (sep alphanum{3,8})* ;
     */
    return _isSepListOf(&_isUnicodeLocaleTypeSubtag, s, len);
}

static UBool
_isTKey(const char* s, int32_t len)
{
    /*
     * tkey = alpha digit ;
     */
    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }
    if (len == 2 && ISALPHA(*s) && ISNUMERIC(*(s + 1))) {
        return TRUE;
    }
    return FALSE;
}

static UBool
_isTValue(const char* s, int32_t len)
{
    /*
     * tvalue = (sep alphanum{3,8})+ ;
     */
    return _isAlphaNumericStringLimitedLength(s, len , 3, 8);
}

static UBool
_isTransformedExtensionSubtag(int32_t& state, const char* s, int32_t len)
{
    const int32_t kStart = 0;       // Start, wait for unicode_language_subtag, tkey or end
    const int32_t kGotLanguage = 1; // Got unicode_language_subtag, wait for unicode_script_subtag,
                                    // unicode_region_subtag, unicode_variant_subtag, tkey or end
    const int32_t kGotScript = 2;   // Got unicode_script_subtag, wait for unicode_region_subtag,
                                    // unicode_variant_subtag, tkey, or end
    const int32_t kGotRegion = 3;   // Got unicode_region_subtag, wait for unicode_variant_subtag,
                                    // tkey, or end.
    const int32_t kGotVariant = 4;  // Got unicode_variant_subtag, wait for unicode_variant_subtag
                                    // tkey or end.
    const int32_t kGotTKey = -1;    // Got tkey, wait for tvalue. ERROR if stop here.
    const int32_t kGotTValue = 6;   // Got tvalue, wait for tkey, tvalue or end

    switch (state) {
        case kStart:
            if (ultag_isLanguageSubtag(s, len)) {
                state = kGotLanguage;
                return TRUE;
            }
            if (_isTKey(s, len)) {
                state = kGotTKey;
                return TRUE;
            }
            return FALSE;
        case kGotLanguage:
            if (ultag_isScriptSubtag(s, len)) {
                state = kGotScript;
                return TRUE;
            }
            U_FALLTHROUGH;
        case kGotScript:
            if (ultag_isRegionSubtag(s, len)) {
                state = kGotRegion;
                return TRUE;
            }
            U_FALLTHROUGH;
        case kGotRegion:
            U_FALLTHROUGH;
        case kGotVariant:
            if (_isVariantSubtag(s, len)) {
                state = kGotVariant;
                return TRUE;
            }
            if (_isTKey(s, len)) {
                state = kGotTKey;
                return TRUE;
            }
            return FALSE;
        case kGotTKey:
            if (_isTValue(s, len)) {
                state = kGotTValue;
                return TRUE;
            }
            return FALSE;
        case kGotTValue:
            if (_isTKey(s, len)) {
                state = kGotTKey;
                return TRUE;
            }
            if (_isTValue(s, len)) {
                return TRUE;
            }
            return FALSE;
    }
    return FALSE;
}

static UBool
_isUnicodeExtensionSubtag(int32_t& state, const char* s, int32_t len)
{
    const int32_t kStart = 0;         // Start, wait for a key or attribute or end
    const int32_t kGotKey = 1;        // Got a key, wait for type or key or end
    const int32_t kGotType = 2;       // Got a type, wait for key or end

    switch (state) {
        case kStart:
            if (ultag_isUnicodeLocaleKey(s, len)) {
                state = kGotKey;
                return TRUE;
            }
            if (ultag_isUnicodeLocaleAttribute(s, len)) {
                return TRUE;
            }
            return FALSE;
        case kGotKey:
            if (ultag_isUnicodeLocaleKey(s, len)) {
                return TRUE;
            }
            if (_isUnicodeLocaleTypeSubtag(s, len)) {
                state = kGotType;
                return TRUE;
            }
            return FALSE;
        case kGotType:
            if (ultag_isUnicodeLocaleKey(s, len)) {
                state = kGotKey;
                return TRUE;
            }
            if (_isUnicodeLocaleTypeSubtag(s, len)) {
                return TRUE;
            }
            return FALSE;
    }
    return FALSE;
}

static UBool
_isStatefulSepListOf(UBool (*test)(int32_t&, const char*, int32_t), const char* s, int32_t len)
{
    int32_t state = 0;
    const char* p;
    const char* start = s;
    int32_t subtagLen = 0;

    if (len < 0) {
        len = (int32_t)uprv_strlen(s);
    }

    for (p = s; len > 0; p++, len--) {
        if (*p == SEP) {
            if (!test(state, start, subtagLen)) {
                return FALSE;
            }
            subtagLen = 0;
            start = p + 1;
        } else {
            subtagLen++;
        }
    }

    if (test(state, start, subtagLen) && state >= 0) {
        return TRUE;
    }
    return FALSE;
}

U_CFUNC UBool
ultag_isTransformedExtensionSubtags(const char* s, int32_t len)
{
    return _isStatefulSepListOf(&_isTransformedExtensionSubtag, s, len);
}

U_CFUNC UBool
ultag_isUnicodeExtensionSubtags(const char* s, int32_t len) {
    return _isStatefulSepListOf(&_isUnicodeExtensionSubtag, s, len);
}


/*
* -------------------------------------------------
*
* Helper functions
*
* -------------------------------------------------
*/

static UBool
_addVariantToList(VariantListEntry **first, VariantListEntry *var) {
    UBool bAdded = TRUE;

    if (*first == NULL) {
        var->next = NULL;
        *first = var;
    } else {
        VariantListEntry *prev, *cur;
        int32_t cmp;

        /* variants order should be preserved */
        prev = NULL;
        cur = *first;
        while (TRUE) {
            if (cur == NULL) {
                prev->next = var;
                var->next = NULL;
                break;
            }

            /* Checking for duplicate variant */
            cmp = uprv_compareInvCharsAsAscii(var->variant, cur->variant);
            if (cmp == 0) {
                /* duplicated variant */
                bAdded = FALSE;
                break;
            }
            prev = cur;
            cur = cur->next;
        }
    }

    return bAdded;
}

static UBool
_addAttributeToList(AttributeListEntry **first, AttributeListEntry *attr) {
    UBool bAdded = TRUE;

    if (*first == NULL) {
        attr->next = NULL;
        *first = attr;
    } else {
        AttributeListEntry *prev, *cur;
        int32_t cmp;

        /* reorder variants in alphabetical order */
        prev = NULL;
        cur = *first;
        while (TRUE) {
            if (cur == NULL) {
                prev->next = attr;
                attr->next = NULL;
                break;
            }
            cmp = uprv_compareInvCharsAsAscii(attr->attribute, cur->attribute);
            if (cmp < 0) {
                if (prev == NULL) {
                    *first = attr;
                } else {
                    prev->next = attr;
                }
                attr->next = cur;
                break;
            }
            if (cmp == 0) {
                /* duplicated variant */
                bAdded = FALSE;
                break;
            }
            prev = cur;
            cur = cur->next;
        }
    }

    return bAdded;
}


static UBool
_addExtensionToList(ExtensionListEntry **first, ExtensionListEntry *ext, UBool localeToBCP) {
    UBool bAdded = TRUE;

    if (*first == NULL) {
        ext->next = NULL;
        *first = ext;
    } else {
        ExtensionListEntry *prev, *cur;
        int32_t cmp;

        /* reorder variants in alphabetical order */
        prev = NULL;
        cur = *first;
        while (TRUE) {
            if (cur == NULL) {
                prev->next = ext;
                ext->next = NULL;
                break;
            }
            if (localeToBCP) {
                /* special handling for locale to bcp conversion */
                int32_t len, curlen;

                len = (int32_t)uprv_strlen(ext->key);
                curlen = (int32_t)uprv_strlen(cur->key);

                if (len == 1 && curlen == 1) {
                    if (*(ext->key) == *(cur->key)) {
                        cmp = 0;
                    } else if (*(ext->key) == PRIVATEUSE) {
                        cmp = 1;
                    } else if (*(cur->key) == PRIVATEUSE) {
                        cmp = -1;
                    } else {
                        cmp = *(ext->key) - *(cur->key);
                    }
                } else if (len == 1) {
                    cmp = *(ext->key) - LDMLEXT;
                } else if (curlen == 1) {
                    cmp = LDMLEXT - *(cur->key);
                } else {
                    cmp = uprv_compareInvCharsAsAscii(ext->key, cur->key);
                    /* Both are u extension keys - we need special handling for 'attribute' */
                    if (cmp != 0) {
                        if (uprv_strcmp(cur->key, LOCALE_ATTRIBUTE_KEY) == 0) {
                            cmp = 1;
                        } else if (uprv_strcmp(ext->key, LOCALE_ATTRIBUTE_KEY) == 0) {
                            cmp = -1;
                        }
                    }
                }
            } else {
                cmp = uprv_compareInvCharsAsAscii(ext->key, cur->key);
            }
            if (cmp < 0) {
                if (prev == NULL) {
                    *first = ext;
                } else {
                    prev->next = ext;
                }
                ext->next = cur;
                break;
            }
            if (cmp == 0) {
                /* duplicated extension key */
                bAdded = FALSE;
                break;
            }
            prev = cur;
            cur = cur->next;
        }
    }

    return bAdded;
}

static void
_initializeULanguageTag(ULanguageTag* langtag) {
    int32_t i;

    langtag->buf = NULL;

    langtag->language = EMPTY;
    for (i = 0; i < MAXEXTLANG; i++) {
        langtag->extlang[i] = NULL;
    }

    langtag->script = EMPTY;
    langtag->region = EMPTY;

    langtag->variants = NULL;
    langtag->extensions = NULL;

    langtag->grandfathered = EMPTY;
    langtag->privateuse = EMPTY;
}

static void
_appendLanguageToLanguageTag(const char* localeID, icu::ByteSink& sink, UBool strict, UErrorCode* status) {
    char buf[ULOC_LANG_CAPACITY];
    UErrorCode tmpStatus = U_ZERO_ERROR;
    int32_t len, i;

    if (U_FAILURE(*status)) {
        return;
    }

    len = uloc_getLanguage(localeID, buf, sizeof(buf), &tmpStatus);
    if (U_FAILURE(tmpStatus) || tmpStatus == U_STRING_NOT_TERMINATED_WARNING) {
        if (strict) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        len = 0;
    }

    /* Note: returned language code is in lower case letters */

    if (len == 0) {
        sink.Append(LANG_UND, LANG_UND_LEN);
    } else if (!ultag_isLanguageSubtag(buf, len)) {
            /* invalid language code */
        if (strict) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        sink.Append(LANG_UND, LANG_UND_LEN);
    } else {
        /* resolve deprecated */
        for (i = 0; i < UPRV_LENGTHOF(DEPRECATEDLANGS); i += 2) {
            // 2-letter deprecated subtags are listede before 3-letter
            // ones in DEPRECATEDLANGS[]. Get out of loop on coming
            // across the 1st 3-letter subtag, if the input is a 2-letter code.
            // to avoid continuing to try when there's no match.
            if (uprv_strlen(buf) < uprv_strlen(DEPRECATEDLANGS[i])) break;
            if (uprv_compareInvCharsAsAscii(buf, DEPRECATEDLANGS[i]) == 0) {
                uprv_strcpy(buf, DEPRECATEDLANGS[i + 1]);
                len = (int32_t)uprv_strlen(buf);
                break;
            }
        }
        sink.Append(buf, len);
    }
}

static void
_appendScriptToLanguageTag(const char* localeID, icu::ByteSink& sink, UBool strict, UErrorCode* status) {
    char buf[ULOC_SCRIPT_CAPACITY];
    UErrorCode tmpStatus = U_ZERO_ERROR;
    int32_t len;

    if (U_FAILURE(*status)) {
        return;
    }

    len = uloc_getScript(localeID, buf, sizeof(buf), &tmpStatus);
    if (U_FAILURE(tmpStatus) || tmpStatus == U_STRING_NOT_TERMINATED_WARNING) {
        if (strict) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
        }
        return;
    }

    if (len > 0) {
        if (!ultag_isScriptSubtag(buf, len)) {
            /* invalid script code */
            if (strict) {
                *status = U_ILLEGAL_ARGUMENT_ERROR;
            }
            return;
        } else {
            sink.Append("-", 1);
            sink.Append(buf, len);
        }
    }
}

static void
_appendRegionToLanguageTag(const char* localeID, icu::ByteSink& sink, UBool strict, UErrorCode* status) {
    char buf[ULOC_COUNTRY_CAPACITY];
    UErrorCode tmpStatus = U_ZERO_ERROR;
    int32_t len;

    if (U_FAILURE(*status)) {
        return;
    }

    len = uloc_getCountry(localeID, buf, sizeof(buf), &tmpStatus);
    if (U_FAILURE(tmpStatus) || tmpStatus == U_STRING_NOT_TERMINATED_WARNING) {
        if (strict) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
        }
        return;
    }

    if (len > 0) {
        if (!ultag_isRegionSubtag(buf, len)) {
            /* invalid region code */
            if (strict) {
                *status = U_ILLEGAL_ARGUMENT_ERROR;
            }
            return;
        } else {
            sink.Append("-", 1);
            /* resolve deprecated */
            for (int i = 0; i < UPRV_LENGTHOF(DEPRECATEDREGIONS); i += 2) {
                if (uprv_compareInvCharsAsAscii(buf, DEPRECATEDREGIONS[i]) == 0) {
                    uprv_strcpy(buf, DEPRECATEDREGIONS[i + 1]);
                    len = (int32_t)uprv_strlen(buf);
                    break;
                }
            }
            sink.Append(buf, len);
        }
    }
}

static void
_appendVariantsToLanguageTag(const char* localeID, icu::ByteSink& sink, UBool strict, UBool *hadPosix, UErrorCode* status) {
    char buf[ULOC_FULLNAME_CAPACITY];
    UErrorCode tmpStatus = U_ZERO_ERROR;
    int32_t len, i;

    if (U_FAILURE(*status)) {
        return;
    }

    len = uloc_getVariant(localeID, buf, sizeof(buf), &tmpStatus);
    if (U_FAILURE(tmpStatus) || tmpStatus == U_STRING_NOT_TERMINATED_WARNING) {
        if (strict) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
        }
        return;
    }

    if (len > 0) {
        char *p, *pVar;
        UBool bNext = TRUE;
        VariantListEntry *var;
        VariantListEntry *varFirst = NULL;

        pVar = NULL;
        p = buf;
        while (bNext) {
            if (*p == SEP || *p == LOCALE_SEP || *p == 0) {
                if (*p == 0) {
                    bNext = FALSE;
                } else {
                    *p = 0; /* terminate */
                }
                if (pVar == NULL) {
                    if (strict) {
                        *status = U_ILLEGAL_ARGUMENT_ERROR;
                        break;
                    }
                    /* ignore empty variant */
                } else {
                    /* ICU uses upper case letters for variants, but
                       the canonical format is lowercase in BCP47 */
                    for (i = 0; *(pVar + i) != 0; i++) {
                        *(pVar + i) = uprv_tolower(*(pVar + i));
                    }

                    /* validate */
                    if (_isVariantSubtag(pVar, -1)) {
                        if (uprv_strcmp(pVar,POSIX_VALUE) || len != (int32_t)uprv_strlen(POSIX_VALUE)) {
                            /* emit the variant to the list */
                            var = (VariantListEntry*)uprv_malloc(sizeof(VariantListEntry));
                            if (var == NULL) {
                                *status = U_MEMORY_ALLOCATION_ERROR;
                                break;
                            }
                            var->variant = pVar;
                            if (!_addVariantToList(&varFirst, var)) {
                                /* duplicated variant */
                                uprv_free(var);
                                if (strict) {
                                    *status = U_ILLEGAL_ARGUMENT_ERROR;
                                    break;
                                }
                            }
                        } else {
                            /* Special handling for POSIX variant, need to remember that we had it and then */
                            /* treat it like an extension later. */
                            *hadPosix = TRUE;
                        }
                    } else if (strict) {
                        *status = U_ILLEGAL_ARGUMENT_ERROR;
                        break;
                    } else if (_isPrivateuseValueSubtag(pVar, -1)) {
                        /* Handle private use subtags separately */
                        break;
                    }
                }
                /* reset variant starting position */
                pVar = NULL;
            } else if (pVar == NULL) {
                pVar = p;
            }
            p++;
        }

        if (U_SUCCESS(*status)) {
            if (varFirst != NULL) {
                int32_t varLen;

                /* write out validated/normalized variants to the target */
                var = varFirst;
                while (var != NULL) {
                    sink.Append("-", 1);
                    varLen = (int32_t)uprv_strlen(var->variant);
                    sink.Append(var->variant, varLen);
                    var = var->next;
                }
            }
        }

        /* clean up */
        var = varFirst;
        while (var != NULL) {
            VariantListEntry *tmpVar = var->next;
            uprv_free(var);
            var = tmpVar;
        }

        if (U_FAILURE(*status)) {
            return;
        }
    }
}

static void
_appendKeywordsToLanguageTag(const char* localeID, icu::ByteSink& sink, UBool strict, UBool hadPosix, UErrorCode* status) {
    char attrBuf[ULOC_KEYWORD_AND_VALUES_CAPACITY] = { 0 };
    int32_t attrBufLength = 0;

    icu::MemoryPool<AttributeListEntry> attrPool;
    icu::MemoryPool<ExtensionListEntry> extPool;
    icu::MemoryPool<icu::CharString> strPool;

    icu::LocalUEnumerationPointer keywordEnum(uloc_openKeywords(localeID, status));
    if (U_FAILURE(*status) && !hadPosix) {
        return;
    }
    if (keywordEnum.isValid() || hadPosix) {
        /* reorder extensions */
        int32_t len;
        const char *key;
        ExtensionListEntry *firstExt = NULL;
        ExtensionListEntry *ext;
        AttributeListEntry *firstAttr = NULL;
        AttributeListEntry *attr;
        icu::MemoryPool<icu::CharString> extBufPool;
        const char *bcpKey=nullptr, *bcpValue=nullptr;
        UErrorCode tmpStatus = U_ZERO_ERROR;
        int32_t keylen;
        UBool isBcpUExt;

        while (TRUE) {
            icu::CharString buf;
            key = uenum_next(keywordEnum.getAlias(), NULL, status);
            if (key == NULL) {
                break;
            }
            char* buffer;
            int32_t resultCapacity = ULOC_KEYWORD_AND_VALUES_CAPACITY;

            for (;;) {
                buffer = buf.getAppendBuffer(
                        /*minCapacity=*/resultCapacity,
                        /*desiredCapacityHint=*/resultCapacity,
                        resultCapacity,
                        tmpStatus);

                if (U_FAILURE(tmpStatus)) {
                    break;
                }

                len = uloc_getKeywordValue(
                        localeID, key, buffer, resultCapacity, &tmpStatus);

                if (tmpStatus != U_BUFFER_OVERFLOW_ERROR) {
                    break;
                }

                resultCapacity = len;
                tmpStatus = U_ZERO_ERROR;
            }

            if (U_FAILURE(tmpStatus)) {
                if (tmpStatus == U_MEMORY_ALLOCATION_ERROR) {
                    *status = U_MEMORY_ALLOCATION_ERROR;
                    break;
                }
                if (strict) {
                    *status = U_ILLEGAL_ARGUMENT_ERROR;
                    break;
                }
                /* ignore this keyword */
                tmpStatus = U_ZERO_ERROR;
                continue;
            }

            buf.append(buffer, len, tmpStatus);
            if (tmpStatus == U_STRING_NOT_TERMINATED_WARNING) {
                tmpStatus = U_ZERO_ERROR;  // Terminators provided by CharString.
            }

            keylen = (int32_t)uprv_strlen(key);
            isBcpUExt = (keylen > 1);

            /* special keyword used for representing Unicode locale attributes */
            if (uprv_strcmp(key, LOCALE_ATTRIBUTE_KEY) == 0) {
                if (len > 0) {
                    int32_t i = 0;
                    while (TRUE) {
                        attrBufLength = 0;
                        for (; i < len; i++) {
                            if (buf[i] != '-') {
                                attrBuf[attrBufLength++] = buf[i];
                            } else {
                                i++;
                                break;
                            }
                        }
                        if (attrBufLength > 0) {
                            attrBuf[attrBufLength] = 0;

                        } else if (i >= len){
                            break;
                        }

                        /* create AttributeListEntry */
                        attr = attrPool.create();
                        if (attr == NULL) {
                            *status = U_MEMORY_ALLOCATION_ERROR;
                            break;
                        }
                        icu::CharString* attrValue =
                                strPool.create(attrBuf, attrBufLength, *status);
                        if (attrValue == NULL) {
                            *status = U_MEMORY_ALLOCATION_ERROR;
                            break;
                        }
                        if (U_FAILURE(*status)) {
                            break;
                        }
                        attr->attribute = attrValue->data();

                        if (!_addAttributeToList(&firstAttr, attr)) {
                            if (strict) {
                                *status = U_ILLEGAL_ARGUMENT_ERROR;
                                break;
                            }
                        }
                    }
                    /* for a place holder ExtensionListEntry */
                    bcpKey = LOCALE_ATTRIBUTE_KEY;
                    bcpValue = NULL;
                }
            } else if (isBcpUExt) {
                bcpKey = uloc_toUnicodeLocaleKey(key);
                if (bcpKey == NULL) {
                    if (strict) {
                        *status = U_ILLEGAL_ARGUMENT_ERROR;
                        break;
                    }
                    continue;
                }

                /* we've checked buf is null-terminated above */
                bcpValue = uloc_toUnicodeLocaleType(key, buf.data());
                if (bcpValue == NULL) {
                    if (strict) {
                        *status = U_ILLEGAL_ARGUMENT_ERROR;
                        break;
                    }
                    continue;
                }
                if (bcpValue == buf.data()) {
                    /*
                    When uloc_toUnicodeLocaleType(key, buf) returns the
                    input value as is, the value is well-formed, but has
                    no known mapping. This implementation normalizes the
                    value to lower case
                    */
                    icu::CharString* extBuf = extBufPool.create();
                    if (extBuf == nullptr) {
                        *status = U_MEMORY_ALLOCATION_ERROR;
                        break;
                    }
                    int32_t bcpValueLen = static_cast<int32_t>(uprv_strlen(bcpValue));
                    int32_t resultCapacity;
                    char* pExtBuf = extBuf->getAppendBuffer(
                            /*minCapacity=*/bcpValueLen,
                            /*desiredCapacityHint=*/bcpValueLen,
                            resultCapacity,
                            tmpStatus);
                    if (U_FAILURE(tmpStatus)) {
                        *status = tmpStatus;
                        break;
                    }

                    uprv_strcpy(pExtBuf, bcpValue);
                    T_CString_toLowerCase(pExtBuf);

                    extBuf->append(pExtBuf, bcpValueLen, tmpStatus);
                    if (U_FAILURE(tmpStatus)) {
                        *status = tmpStatus;
                        break;
                    }

                    bcpValue = extBuf->data();
                }
            } else {
                if (*key == PRIVATEUSE) {
                    if (!ultag_isPrivateuseValueSubtags(buf.data(), len)) {
                        if (strict) {
                            *status = U_ILLEGAL_ARGUMENT_ERROR;
                            break;
                        }
                        continue;
                    }
                } else {
                    if (!_isExtensionSingleton(key, keylen) || !ultag_isExtensionSubtags(buf.data(), len)) {
                        if (strict) {
                            *status = U_ILLEGAL_ARGUMENT_ERROR;
                            break;
                        }
                        continue;
                    }
                }
                bcpKey = key;
                icu::CharString* extBuf =
                    extBufPool.create(buf.data(), len, tmpStatus);
                if (extBuf == nullptr) {
                    *status = U_MEMORY_ALLOCATION_ERROR;
                    break;
                }
                if (U_FAILURE(tmpStatus)) {
                    *status = tmpStatus;
                    break;
                }
                bcpValue = extBuf->data();
            }

            /* create ExtensionListEntry */
            ext = extPool.create();
            if (ext == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                break;
            }
            ext->key = bcpKey;
            ext->value = bcpValue;

            if (!_addExtensionToList(&firstExt, ext, TRUE)) {
                if (strict) {
                    *status = U_ILLEGAL_ARGUMENT_ERROR;
                    break;
                }
            }
        }

        /* Special handling for POSIX variant - add the keywords for POSIX */
        if (hadPosix) {
            /* create ExtensionListEntry for POSIX */
            ext = extPool.create();
            if (ext == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            ext->key = POSIX_KEY;
            ext->value = POSIX_VALUE;

            if (!_addExtensionToList(&firstExt, ext, TRUE)) {
                // Silently ignore errors.
            }
        }

        if (U_SUCCESS(*status) && (firstExt != NULL || firstAttr != NULL)) {
            UBool startLDMLExtension = FALSE;
            for (ext = firstExt; ext; ext = ext->next) {
                if (!startLDMLExtension && uprv_strlen(ext->key) > 1) {
                    /* first LDML u singlton extension */
                   sink.Append("-u", 2);
                   startLDMLExtension = TRUE;
                }

                /* write out the sorted BCP47 attributes, extensions and private use */
                if (uprv_strcmp(ext->key, LOCALE_ATTRIBUTE_KEY) == 0) {
                    /* write the value for the attributes */
                    for (attr = firstAttr; attr; attr = attr->next) {
                        sink.Append("-", 1);
                        sink.Append(
                                attr->attribute, static_cast<int32_t>(uprv_strlen(attr->attribute)));
                    }
                } else {
                    sink.Append("-", 1);
                    sink.Append(ext->key, static_cast<int32_t>(uprv_strlen(ext->key)));
                    sink.Append("-", 1);
                    sink.Append(ext->value, static_cast<int32_t>(uprv_strlen(ext->value)));
                }
            }
        }
    }
}

/**
 * Append keywords parsed from LDML extension value
 * e.g. "u-ca-gregory-co-trad" -> {calendar = gregorian} {collation = traditional}
 * Note: char* buf is used for storing keywords
 */
static void
_appendLDMLExtensionAsKeywords(const char* ldmlext, ExtensionListEntry** appendTo, icu::MemoryPool<ExtensionListEntry>& extPool, icu::MemoryPool<icu::CharString>& kwdBuf, UBool *posixVariant, UErrorCode *status) {
    const char *pTag;   /* beginning of current subtag */
    const char *pKwds;  /* beginning of key-type pairs */
    UBool variantExists = *posixVariant;

    ExtensionListEntry *kwdFirst = NULL;    /* first LDML keyword */
    ExtensionListEntry *kwd, *nextKwd;

    int32_t len;

    /* Reset the posixVariant value */
    *posixVariant = FALSE;

    pTag = ldmlext;
    pKwds = NULL;

    {
        AttributeListEntry *attrFirst = NULL;   /* first attribute */
        AttributeListEntry *attr, *nextAttr;

        char attrBuf[ULOC_KEYWORD_AND_VALUES_CAPACITY];
        int32_t attrBufIdx = 0;

        icu::MemoryPool<AttributeListEntry> attrPool;

        /* Iterate through u extension attributes */
        while (*pTag) {
            /* locate next separator char */
            for (len = 0; *(pTag + len) && *(pTag + len) != SEP; len++);

            if (ultag_isUnicodeLocaleKey(pTag, len)) {
                pKwds = pTag;
                break;
            }

            /* add this attribute to the list */
            attr = attrPool.create();
            if (attr == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }

            if (len < (int32_t)sizeof(attrBuf) - attrBufIdx) {
                uprv_memcpy(&attrBuf[attrBufIdx], pTag, len);
                attrBuf[attrBufIdx + len] = 0;
                attr->attribute = &attrBuf[attrBufIdx];
                attrBufIdx += (len + 1);
            } else {
                *status = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }

            // duplicate attribute is ignored, causes no error.
            _addAttributeToList(&attrFirst, attr);

            /* next tag */
            pTag += len;
            if (*pTag) {
                /* next to the separator */
                pTag++;
            }
        }

        if (attrFirst) {
            /* emit attributes as an LDML keyword, e.g. attribute=attr1-attr2 */

            kwd = extPool.create();
            if (kwd == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }

            icu::CharString* value = kwdBuf.create();
            if (value == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }

            /* attribute subtags sorted in alphabetical order as type */
            attr = attrFirst;
            while (attr != NULL) {
                nextAttr = attr->next;
                if (attr != attrFirst) {
                    value->append('-', *status);
                }
                value->append(attr->attribute, *status);
                attr = nextAttr;
            }
            if (U_FAILURE(*status)) {
                return;
            }

            kwd->key = LOCALE_ATTRIBUTE_KEY;
            kwd->value = value->data();

            if (!_addExtensionToList(&kwdFirst, kwd, FALSE)) {
                *status = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
        }
    }

    if (pKwds) {
        const char *pBcpKey = NULL;     /* u extenstion key subtag */
        const char *pBcpType = NULL;    /* beginning of u extension type subtag(s) */
        int32_t bcpKeyLen = 0;
        int32_t bcpTypeLen = 0;
        UBool isDone = FALSE;

        pTag = pKwds;
        /* BCP47 representation of LDML key/type pairs */
        while (!isDone) {
            const char *pNextBcpKey = NULL;
            int32_t nextBcpKeyLen = 0;
            UBool emitKeyword = FALSE;

            if (*pTag) {
                /* locate next separator char */
                for (len = 0; *(pTag + len) && *(pTag + len) != SEP; len++);

                if (ultag_isUnicodeLocaleKey(pTag, len)) {
                    if (pBcpKey) {
                        emitKeyword = TRUE;
                        pNextBcpKey = pTag;
                        nextBcpKeyLen = len;
                    } else {
                        pBcpKey = pTag;
                        bcpKeyLen = len;
                    }
                } else {
                    U_ASSERT(pBcpKey != NULL);
                    /* within LDML type subtags */
                    if (pBcpType) {
                        bcpTypeLen += (len + 1);
                    } else {
                        pBcpType = pTag;
                        bcpTypeLen = len;
                    }
                }

                /* next tag */
                pTag += len;
                if (*pTag) {
                    /* next to the separator */
                    pTag++;
                }
            } else {
                /* processing last one */
                emitKeyword = TRUE;
                isDone = TRUE;
            }

            if (emitKeyword) {
                const char *pKey = NULL;    /* LDML key */
                const char *pType = NULL;   /* LDML type */

                char bcpKeyBuf[9];          /* BCP key length is always 2 for now */

                U_ASSERT(pBcpKey != NULL);

                if (bcpKeyLen >= (int32_t)sizeof(bcpKeyBuf)) {
                    /* the BCP key is invalid */
                    *status = U_ILLEGAL_ARGUMENT_ERROR;
                    return;
                }

                uprv_strncpy(bcpKeyBuf, pBcpKey, bcpKeyLen);
                bcpKeyBuf[bcpKeyLen] = 0;

                /* u extension key to LDML key */
                pKey = uloc_toLegacyKey(bcpKeyBuf);
                if (pKey == NULL) {
                    *status = U_ILLEGAL_ARGUMENT_ERROR;
                    return;
                }
                if (pKey == bcpKeyBuf) {
                    /*
                    The key returned by toLegacyKey points to the input buffer.
                    We normalize the result key to lower case.
                    */
                    T_CString_toLowerCase(bcpKeyBuf);
                    icu::CharString* key = kwdBuf.create(bcpKeyBuf, bcpKeyLen, *status);
                    if (key == NULL) {
                        *status = U_MEMORY_ALLOCATION_ERROR;
                        return;
                    }
                    if (U_FAILURE(*status)) {
                        return;
                    }
                    pKey = key->data();
                }

                if (pBcpType) {
                    char bcpTypeBuf[128];       /* practically long enough even considering multiple subtag type */
                    if (bcpTypeLen >= (int32_t)sizeof(bcpTypeBuf)) {
                        /* the BCP type is too long */
                        *status = U_ILLEGAL_ARGUMENT_ERROR;
                        return;
                    }

                    uprv_strncpy(bcpTypeBuf, pBcpType, bcpTypeLen);
                    bcpTypeBuf[bcpTypeLen] = 0;

                    /* BCP type to locale type */
                    pType = uloc_toLegacyType(pKey, bcpTypeBuf);
                    if (pType == NULL) {
                        *status = U_ILLEGAL_ARGUMENT_ERROR;
                        return;
                    }
                    if (pType == bcpTypeBuf) {
                        /*
                        The type returned by toLegacyType points to the input buffer.
                        We normalize the result type to lower case.
                        */
                        /* normalize to lower case */
                        T_CString_toLowerCase(bcpTypeBuf);
                        icu::CharString* type = kwdBuf.create(bcpTypeBuf, bcpTypeLen, *status);
                        if (type == NULL) {
                            *status = U_MEMORY_ALLOCATION_ERROR;
                            return;
                        }
                        if (U_FAILURE(*status)) {
                            return;
                        }
                        pType = type->data();
                    }
                } else {
                    /* typeless - default type value is "yes" */
                    pType = LOCALE_TYPE_YES;
                }

                /* Special handling for u-va-posix, since we want to treat this as a variant,
                   not as a keyword */
                if (!variantExists && !uprv_strcmp(pKey, POSIX_KEY) && !uprv_strcmp(pType, POSIX_VALUE) ) {
                    *posixVariant = TRUE;
                } else {
                    /* create an ExtensionListEntry for this keyword */
                    kwd = extPool.create();
                    if (kwd == NULL) {
                        *status = U_MEMORY_ALLOCATION_ERROR;
                        return;
                    }

                    kwd->key = pKey;
                    kwd->value = pType;

                    if (!_addExtensionToList(&kwdFirst, kwd, FALSE)) {
                        // duplicate keyword is allowed, Only the first
                        // is honored.
                    }
                }

                pBcpKey = pNextBcpKey;
                bcpKeyLen = pNextBcpKey != NULL ? nextBcpKeyLen : 0;
                pBcpType = NULL;
                bcpTypeLen = 0;
            }
        }
    }

    kwd = kwdFirst;
    while (kwd != NULL) {
        nextKwd = kwd->next;
        _addExtensionToList(appendTo, kwd, FALSE);
        kwd = nextKwd;
    }
}


static void
_appendKeywords(ULanguageTag* langtag, icu::ByteSink& sink, UErrorCode* status) {
    int32_t i, n;
    int32_t len;
    ExtensionListEntry *kwdFirst = NULL;
    ExtensionListEntry *kwd;
    const char *key, *type;
    icu::MemoryPool<ExtensionListEntry> extPool;
    icu::MemoryPool<icu::CharString> kwdBuf;
    UBool posixVariant = FALSE;

    if (U_FAILURE(*status)) {
        return;
    }

    /* Determine if variants already exists */
    if (ultag_getVariantsSize(langtag)) {
        posixVariant = TRUE;
    }

    n = ultag_getExtensionsSize(langtag);

    /* resolve locale keywords and reordering keys */
    for (i = 0; i < n; i++) {
        key = ultag_getExtensionKey(langtag, i);
        type = ultag_getExtensionValue(langtag, i);
        if (*key == LDMLEXT) {
            _appendLDMLExtensionAsKeywords(type, &kwdFirst, extPool, kwdBuf, &posixVariant, status);
            if (U_FAILURE(*status)) {
                break;
            }
        } else {
            kwd = extPool.create();
            if (kwd == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                break;
            }
            kwd->key = key;
            kwd->value = type;
            if (!_addExtensionToList(&kwdFirst, kwd, FALSE)) {
                *status = U_ILLEGAL_ARGUMENT_ERROR;
                break;
            }
        }
    }

    if (U_SUCCESS(*status)) {
        type = ultag_getPrivateUse(langtag);
        if ((int32_t)uprv_strlen(type) > 0) {
            /* add private use as a keyword */
            kwd = extPool.create();
            if (kwd == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
            } else {
                kwd->key = PRIVATEUSE_KEY;
                kwd->value = type;
                if (!_addExtensionToList(&kwdFirst, kwd, FALSE)) {
                    *status = U_ILLEGAL_ARGUMENT_ERROR;
                }
            }
        }
    }

    /* If a POSIX variant was in the extensions, write it out before writing the keywords. */

    if (U_SUCCESS(*status) && posixVariant) {
        len = (int32_t) uprv_strlen(_POSIX);
        sink.Append(_POSIX, len);
    }

    if (U_SUCCESS(*status) && kwdFirst != NULL) {
        /* write out the sorted keywords */
        UBool firstValue = TRUE;
        kwd = kwdFirst;
        do {
            if (firstValue) {
                sink.Append("@", 1);
                firstValue = FALSE;
            } else {
                sink.Append(";", 1);
            }

            /* key */
            len = (int32_t)uprv_strlen(kwd->key);
            sink.Append(kwd->key, len);
            sink.Append("=", 1);

            /* type */
            len = (int32_t)uprv_strlen(kwd->value);
            sink.Append(kwd->value, len);

            kwd = kwd->next;
        } while (kwd);
    }
}

static void
_appendPrivateuseToLanguageTag(const char* localeID, icu::ByteSink& sink, UBool strict, UBool hadPosix, UErrorCode* status) {
    (void)hadPosix;
    char buf[ULOC_FULLNAME_CAPACITY];
    char tmpAppend[ULOC_FULLNAME_CAPACITY];
    UErrorCode tmpStatus = U_ZERO_ERROR;
    int32_t len, i;
    int32_t reslen = 0;
    int32_t capacity = sizeof tmpAppend;

    if (U_FAILURE(*status)) {
        return;
    }

    len = uloc_getVariant(localeID, buf, sizeof(buf), &tmpStatus);
    if (U_FAILURE(tmpStatus) || tmpStatus == U_STRING_NOT_TERMINATED_WARNING) {
        if (strict) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
        }
        return;
    }

    if (len > 0) {
        char *p, *pPriv;
        UBool bNext = TRUE;
        UBool firstValue = TRUE;
        UBool writeValue;

        pPriv = NULL;
        p = buf;
        while (bNext) {
            writeValue = FALSE;
            if (*p == SEP || *p == LOCALE_SEP || *p == 0) {
                if (*p == 0) {
                    bNext = FALSE;
                } else {
                    *p = 0; /* terminate */
                }
                if (pPriv != NULL) {
                    /* Private use in the canonical format is lowercase in BCP47 */
                    for (i = 0; *(pPriv + i) != 0; i++) {
                        *(pPriv + i) = uprv_tolower(*(pPriv + i));
                    }

                    /* validate */
                    if (_isPrivateuseValueSubtag(pPriv, -1)) {
                        if (firstValue) {
                            if (!_isVariantSubtag(pPriv, -1)) {
                                writeValue = TRUE;
                            }
                        } else {
                            writeValue = TRUE;
                        }
                    } else if (strict) {
                        *status = U_ILLEGAL_ARGUMENT_ERROR;
                        break;
                    } else {
                        break;
                    }

                    if (writeValue) {
                        if (reslen < capacity) {
                            tmpAppend[reslen++] = SEP;
                        }

                        if (firstValue) {
                            if (reslen < capacity) {
                                tmpAppend[reslen++] = *PRIVATEUSE_KEY;
                            }

                            if (reslen < capacity) {
                                tmpAppend[reslen++] = SEP;
                            }

                            len = (int32_t)uprv_strlen(PRIVUSE_VARIANT_PREFIX);
                            if (reslen < capacity) {
                                uprv_memcpy(tmpAppend + reslen, PRIVUSE_VARIANT_PREFIX, uprv_min(len, capacity - reslen));
                            }
                            reslen += len;

                            if (reslen < capacity) {
                                tmpAppend[reslen++] = SEP;
                            }

                            firstValue = FALSE;
                        }

                        len = (int32_t)uprv_strlen(pPriv);
                        if (reslen < capacity) {
                            uprv_memcpy(tmpAppend + reslen, pPriv, uprv_min(len, capacity - reslen));
                        }
                        reslen += len;
                    }
                }
                /* reset private use starting position */
                pPriv = NULL;
            } else if (pPriv == NULL) {
                pPriv = p;
            }
            p++;
        }

        if (U_FAILURE(*status)) {
            return;
        }
    }

    if (U_SUCCESS(*status)) {
        len = reslen;
        sink.Append(tmpAppend, len);
    }
}

/*
* -------------------------------------------------
*
* ultag_ functions
*
* -------------------------------------------------
*/

/* Bit flags used by the parser */
#define LANG 0x0001
#define EXTL 0x0002
#define SCRT 0x0004
#define REGN 0x0008
#define VART 0x0010
#define EXTS 0x0020
#define EXTV 0x0040
#define PRIV 0x0080

/**
 * Ticket #12705 - Visual Studio 2015 Update 3 contains a new code optimizer which has problems optimizing
 * this function. (See https://blogs.msdn.microsoft.com/vcblog/2016/05/04/new-code-optimizer/ )
 * As a workaround, we will turn off optimization just for this function on VS2015 Update 3 and above.
 */
#if (defined(_MSC_VER) && (_MSC_VER >= 1900) && defined(_MSC_FULL_VER) && (_MSC_FULL_VER >= 190024210))
#pragma optimize( "", off )
#endif

static ULanguageTag*
ultag_parse(const char* tag, int32_t tagLen, int32_t* parsedLen, UErrorCode* status) {
    char *tagBuf;
    int16_t next;
    char *pSubtag, *pNext, *pLastGoodPosition;
    int32_t subtagLen;
    int32_t extlangIdx;
    ExtensionListEntry *pExtension;
    char *pExtValueSubtag, *pExtValueSubtagEnd;
    int32_t i;
    UBool privateuseVar = FALSE;
    int32_t grandfatheredLen = 0;

    if (parsedLen != NULL) {
        *parsedLen = 0;
    }

    if (U_FAILURE(*status)) {
        return NULL;
    }

    if (tagLen < 0) {
        tagLen = (int32_t)uprv_strlen(tag);
    }

    /* copy the entire string */
    tagBuf = (char*)uprv_malloc(tagLen + 1);
    if (tagBuf == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    uprv_memcpy(tagBuf, tag, tagLen);
    *(tagBuf + tagLen) = 0;

    /* create a ULanguageTag */
    icu::LocalULanguageTagPointer t(
            (ULanguageTag*)uprv_malloc(sizeof(ULanguageTag)));
    if (t.isNull()) {
        uprv_free(tagBuf);
        *status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    _initializeULanguageTag(t.getAlias());
    t->buf = tagBuf;

    if (tagLen < MINLEN) {
        /* the input tag is too short - return empty ULanguageTag */
        return t.orphan();
    }

    size_t parsedLenDelta = 0;
    // Grandfathered tag will be consider together. Grandfathered tag with intervening
    // script and region such as art-DE-lojban or art-Latn-lojban won't be
    // matched.
    /* check if the tag is grandfathered */
    for (i = 0; i < UPRV_LENGTHOF(GRANDFATHERED); i += 2) {
        int32_t checkGrandfatheredLen = static_cast<int32_t>(uprv_strlen(GRANDFATHERED[i]));
        if (tagLen < checkGrandfatheredLen) {
            continue;
        }
        if (tagLen > checkGrandfatheredLen && tagBuf[checkGrandfatheredLen] != '-') {
            // make sure next char is '-'.
            continue;
        }
        if (uprv_strnicmp(GRANDFATHERED[i], tagBuf, checkGrandfatheredLen) == 0) {
            int32_t newTagLength;

            grandfatheredLen = checkGrandfatheredLen;  /* back up for output parsedLen */
            int32_t replacementLen = static_cast<int32_t>(uprv_strlen(GRANDFATHERED[i+1]));
            newTagLength = replacementLen + tagLen - checkGrandfatheredLen;
            if (tagLen < newTagLength) {
                uprv_free(tagBuf);
                tagBuf = (char*)uprv_malloc(newTagLength + 1);
                if (tagBuf == NULL) {
                    *status = U_MEMORY_ALLOCATION_ERROR;
                    return NULL;
                }
                t->buf = tagBuf;
                tagLen = newTagLength;
            }
            parsedLenDelta = checkGrandfatheredLen - replacementLen;
            uprv_strcpy(t->buf, GRANDFATHERED[i + 1]);
            if (checkGrandfatheredLen != tagLen) {
                uprv_strcpy(t->buf + replacementLen, tag + checkGrandfatheredLen);
            }
            break;
        }
    }

    if (grandfatheredLen == 0) {
        for (i = 0; i < UPRV_LENGTHOF(REDUNDANT); i += 2) {
            const char* redundantTag = REDUNDANT[i];
            size_t redundantTagLen = uprv_strlen(redundantTag);
            // The preferred tag for a redundant tag is always shorter than redundant
            // tag. A redundant tag may or may not be followed by other subtags.
            // (i.e. "zh-yue" or "zh-yue-u-co-pinyin").
            if (uprv_strnicmp(redundantTag, tagBuf, static_cast<uint32_t>(redundantTagLen)) == 0) {
                const char* redundantTagEnd = tagBuf + redundantTagLen;
                if (*redundantTagEnd  == '\0' || *redundantTagEnd == SEP) {
                    const char* preferredTag = REDUNDANT[i + 1];
                    size_t preferredTagLen = uprv_strlen(preferredTag);
                    uprv_strncpy(t->buf, preferredTag, preferredTagLen);
                    if (*redundantTagEnd == SEP) {
                        uprv_memmove(tagBuf + preferredTagLen,
                                     redundantTagEnd,
                                     tagLen - redundantTagLen + 1);
                    } else {
                        tagBuf[preferredTagLen] = '\0';
                    }
                    // parsedLen should be the length of the input
                    // before redundantTag is replaced by preferredTag.
                    // Save the delta to add it back later.
                    parsedLenDelta = redundantTagLen - preferredTagLen;
                    break;
                }
            }
        }
    }

    /*
     * langtag      =   language
     *                  ["-" script]
     *                  ["-" region]
     *                  *("-" variant)
     *                  *("-" extension)
     *                  ["-" privateuse]
     */

    next = LANG | PRIV;
    pNext = pLastGoodPosition = tagBuf;
    extlangIdx = 0;
    pExtension = NULL;
    pExtValueSubtag = NULL;
    pExtValueSubtagEnd = NULL;

    while (pNext) {
        char *pSep;

        pSubtag = pNext;

        /* locate next separator char */
        pSep = pSubtag;
        while (*pSep) {
            if (*pSep == SEP) {
                break;
            }
            pSep++;
        }
        if (*pSep == 0) {
            /* last subtag */
            pNext = NULL;
        } else {
            pNext = pSep + 1;
        }
        subtagLen = (int32_t)(pSep - pSubtag);

        if (next & LANG) {
            if (ultag_isLanguageSubtag(pSubtag, subtagLen)) {
                *pSep = 0;  /* terminate */
                // TODO: move deprecated language code handling here.
                t->language = T_CString_toLowerCase(pSubtag);

                pLastGoodPosition = pSep;
                next = SCRT | REGN | VART | EXTS | PRIV;
                if (subtagLen <= 3)
                  next |= EXTL;
                continue;
            }
        }
        if (next & EXTL) {
            if (_isExtlangSubtag(pSubtag, subtagLen)) {
                *pSep = 0;
                t->extlang[extlangIdx++] = T_CString_toLowerCase(pSubtag);

                pLastGoodPosition = pSep;
                if (extlangIdx < 3) {
                    next = EXTL | SCRT | REGN | VART | EXTS | PRIV;
                } else {
                    next = SCRT | REGN | VART | EXTS | PRIV;
                }
                continue;
            }
        }
        if (next & SCRT) {
            if (ultag_isScriptSubtag(pSubtag, subtagLen)) {
                char *p = pSubtag;

                *pSep = 0;

                /* to title case */
                *p = uprv_toupper(*p);
                p++;
                for (; *p; p++) {
                    *p = uprv_tolower(*p);
                }

                t->script = pSubtag;

                pLastGoodPosition = pSep;
                next = REGN | VART | EXTS | PRIV;
                continue;
            }
        }
        if (next & REGN) {
            if (ultag_isRegionSubtag(pSubtag, subtagLen)) {
                *pSep = 0;
                // TODO: move deprecated region code handling here.
                t->region = T_CString_toUpperCase(pSubtag);

                pLastGoodPosition = pSep;
                next = VART | EXTS | PRIV;
                continue;
            }
        }
        if (next & VART) {
            if (_isVariantSubtag(pSubtag, subtagLen) ||
               (privateuseVar && _isPrivateuseVariantSubtag(pSubtag, subtagLen))) {
                VariantListEntry *var;
                UBool isAdded;

                var = (VariantListEntry*)uprv_malloc(sizeof(VariantListEntry));
                if (var == NULL) {
                    *status = U_MEMORY_ALLOCATION_ERROR;
                    return NULL;
                }
                *pSep = 0;
                var->variant = T_CString_toUpperCase(pSubtag);
                isAdded = _addVariantToList(&(t->variants), var);
                if (!isAdded) {
                    /* duplicated variant entry */
                    uprv_free(var);
                    break;
                }
                pLastGoodPosition = pSep;
                next = VART | EXTS | PRIV;
                continue;
            }
        }
        if (next & EXTS) {
            if (_isExtensionSingleton(pSubtag, subtagLen)) {
                if (pExtension != NULL) {
                    if (pExtValueSubtag == NULL || pExtValueSubtagEnd == NULL) {
                        /* the previous extension is incomplete */
                        uprv_free(pExtension);
                        pExtension = NULL;
                        break;
                    }

                    /* terminate the previous extension value */
                    *pExtValueSubtagEnd = 0;
                    pExtension->value = T_CString_toLowerCase(pExtValueSubtag);

                    /* insert the extension to the list */
                    if (_addExtensionToList(&(t->extensions), pExtension, FALSE)) {
                        pLastGoodPosition = pExtValueSubtagEnd;
                    } else {
                        /* stop parsing here */
                        uprv_free(pExtension);
                        pExtension = NULL;
                        break;
                    }
                }

                /* create a new extension */
                pExtension = (ExtensionListEntry*)uprv_malloc(sizeof(ExtensionListEntry));
                if (pExtension == NULL) {
                    *status = U_MEMORY_ALLOCATION_ERROR;
                    return NULL;
                }
                *pSep = 0;
                pExtension->key = T_CString_toLowerCase(pSubtag);
                pExtension->value = NULL;   /* will be set later */

                /*
                 * reset the start and the end location of extension value
                 * subtags for this extension
                 */
                pExtValueSubtag = NULL;
                pExtValueSubtagEnd = NULL;

                next = EXTV;
                continue;
            }
        }
        if (next & EXTV) {
            if (_isExtensionSubtag(pSubtag, subtagLen)) {
                if (pExtValueSubtag == NULL) {
                    /* if the start postion of this extension's value is not yet,
                        this one is the first value subtag */
                    pExtValueSubtag = pSubtag;
                }

                /* Mark the end of this subtag */
                pExtValueSubtagEnd = pSep;
                next = EXTS | EXTV | PRIV;

                continue;
            }
        }
        if (next & PRIV) {
            if (uprv_tolower(*pSubtag) == PRIVATEUSE && subtagLen == 1) {
                char *pPrivuseVal;

                if (pExtension != NULL) {
                    /* Process the last extension */
                    if (pExtValueSubtag == NULL || pExtValueSubtagEnd == NULL) {
                        /* the previous extension is incomplete */
                        uprv_free(pExtension);
                        pExtension = NULL;
                        break;
                    } else {
                        /* terminate the previous extension value */
                        *pExtValueSubtagEnd = 0;
                        pExtension->value = T_CString_toLowerCase(pExtValueSubtag);

                        /* insert the extension to the list */
                        if (_addExtensionToList(&(t->extensions), pExtension, FALSE)) {
                            pLastGoodPosition = pExtValueSubtagEnd;
                            pExtension = NULL;
                        } else {
                        /* stop parsing here */
                            uprv_free(pExtension);
                            pExtension = NULL;
                            break;
                        }
                    }
                }

                /* The rest of part will be private use value subtags */
                if (pNext == NULL) {
                    /* empty private use subtag */
                    break;
                }
                /* back up the private use value start position */
                pPrivuseVal = pNext;

                /* validate private use value subtags */
                while (pNext) {
                    pSubtag = pNext;
                    pSep = pSubtag;
                    while (*pSep) {
                        if (*pSep == SEP) {
                            break;
                        }
                        pSep++;
                    }
                    if (*pSep == 0) {
                        /* last subtag */
                        pNext = NULL;
                    } else {
                        pNext = pSep + 1;
                    }
                    subtagLen = (int32_t)(pSep - pSubtag);

                    if (uprv_strncmp(pSubtag, PRIVUSE_VARIANT_PREFIX, uprv_strlen(PRIVUSE_VARIANT_PREFIX)) == 0) {
                        *pSep = 0;
                        next = VART;
                        privateuseVar = TRUE;
                        break;
                    } else if (_isPrivateuseValueSubtag(pSubtag, subtagLen)) {
                        pLastGoodPosition = pSep;
                    } else {
                        break;
                    }
                }

                if (next == VART) {
                    continue;
                }

                if (pLastGoodPosition - pPrivuseVal > 0) {
                    *pLastGoodPosition = 0;
                    t->privateuse = T_CString_toLowerCase(pPrivuseVal);
                }
                /* No more subtags, exiting the parse loop */
                break;
            }
            break;
        }

        /* If we fell through here, it means this subtag is illegal - quit parsing */
        break;
    }

    if (pExtension != NULL) {
        /* Process the last extension */
        if (pExtValueSubtag == NULL || pExtValueSubtagEnd == NULL) {
            /* the previous extension is incomplete */
            uprv_free(pExtension);
        } else {
            /* terminate the previous extension value */
            *pExtValueSubtagEnd = 0;
            pExtension->value = T_CString_toLowerCase(pExtValueSubtag);
            /* insert the extension to the list */
            if (_addExtensionToList(&(t->extensions), pExtension, FALSE)) {
                pLastGoodPosition = pExtValueSubtagEnd;
            } else {
                uprv_free(pExtension);
            }
        }
    }

    if (parsedLen != NULL) {
        *parsedLen = (int32_t)(pLastGoodPosition - t->buf + parsedLenDelta);
    }

    return t.orphan();
}

/**
* Ticket #12705 - Turn optimization back on.
*/
#if (defined(_MSC_VER) && (_MSC_VER >= 1900) && defined(_MSC_FULL_VER) && (_MSC_FULL_VER >= 190024210))
#pragma optimize( "", on )
#endif

static void
ultag_close(ULanguageTag* langtag) {

    if (langtag == NULL) {
        return;
    }

    uprv_free(langtag->buf);

    if (langtag->variants) {
        VariantListEntry *curVar = langtag->variants;
        while (curVar) {
            VariantListEntry *nextVar = curVar->next;
            uprv_free(curVar);
            curVar = nextVar;
        }
    }

    if (langtag->extensions) {
        ExtensionListEntry *curExt = langtag->extensions;
        while (curExt) {
            ExtensionListEntry *nextExt = curExt->next;
            uprv_free(curExt);
            curExt = nextExt;
        }
    }

    uprv_free(langtag);
}

static const char*
ultag_getLanguage(const ULanguageTag* langtag) {
    return langtag->language;
}

#if 0
static const char*
ultag_getJDKLanguage(const ULanguageTag* langtag) {
    int32_t i;
    for (i = 0; DEPRECATEDLANGS[i] != NULL; i += 2) {
        if (uprv_compareInvCharsAsAscii(DEPRECATEDLANGS[i], langtag->language) == 0) {
            return DEPRECATEDLANGS[i + 1];
        }
    }
    return langtag->language;
}
#endif

static const char*
ultag_getExtlang(const ULanguageTag* langtag, int32_t idx) {
    if (idx >= 0 && idx < MAXEXTLANG) {
        return langtag->extlang[idx];
    }
    return NULL;
}

static int32_t
ultag_getExtlangSize(const ULanguageTag* langtag) {
    int32_t size = 0;
    int32_t i;
    for (i = 0; i < MAXEXTLANG; i++) {
        if (langtag->extlang[i]) {
            size++;
        }
    }
    return size;
}

static const char*
ultag_getScript(const ULanguageTag* langtag) {
    return langtag->script;
}

static const char*
ultag_getRegion(const ULanguageTag* langtag) {
    return langtag->region;
}

static const char*
ultag_getVariant(const ULanguageTag* langtag, int32_t idx) {
    const char *var = NULL;
    VariantListEntry *cur = langtag->variants;
    int32_t i = 0;
    while (cur) {
        if (i == idx) {
            var = cur->variant;
            break;
        }
        cur = cur->next;
        i++;
    }
    return var;
}

static int32_t
ultag_getVariantsSize(const ULanguageTag* langtag) {
    int32_t size = 0;
    VariantListEntry *cur = langtag->variants;
    while (TRUE) {
        if (cur == NULL) {
            break;
        }
        size++;
        cur = cur->next;
    }
    return size;
}

static const char*
ultag_getExtensionKey(const ULanguageTag* langtag, int32_t idx) {
    const char *key = NULL;
    ExtensionListEntry *cur = langtag->extensions;
    int32_t i = 0;
    while (cur) {
        if (i == idx) {
            key = cur->key;
            break;
        }
        cur = cur->next;
        i++;
    }
    return key;
}

static const char*
ultag_getExtensionValue(const ULanguageTag* langtag, int32_t idx) {
    const char *val = NULL;
    ExtensionListEntry *cur = langtag->extensions;
    int32_t i = 0;
    while (cur) {
        if (i == idx) {
            val = cur->value;
            break;
        }
        cur = cur->next;
        i++;
    }
    return val;
}

static int32_t
ultag_getExtensionsSize(const ULanguageTag* langtag) {
    int32_t size = 0;
    ExtensionListEntry *cur = langtag->extensions;
    while (TRUE) {
        if (cur == NULL) {
            break;
        }
        size++;
        cur = cur->next;
    }
    return size;
}

static const char*
ultag_getPrivateUse(const ULanguageTag* langtag) {
    return langtag->privateuse;
}

#if 0
static const char*
ultag_getGrandfathered(const ULanguageTag* langtag) {
    return langtag->grandfathered;
}
#endif


/*
* -------------------------------------------------
*
* Locale/BCP47 conversion APIs, exposed as uloc_*
*
* -------------------------------------------------
*/
U_CAPI int32_t U_EXPORT2
uloc_toLanguageTag(const char* localeID,
                   char* langtag,
                   int32_t langtagCapacity,
                   UBool strict,
                   UErrorCode* status) {
    if (U_FAILURE(*status)) {
        return 0;
    }

    icu::CheckedArrayByteSink sink(langtag, langtagCapacity);
    ulocimp_toLanguageTag(localeID, sink, strict, status);

    int32_t reslen = sink.NumberOfBytesAppended();

    if (U_FAILURE(*status)) {
        return reslen;
    }

    if (sink.Overflowed()) {
        *status = U_BUFFER_OVERFLOW_ERROR;
    } else {
        u_terminateChars(langtag, langtagCapacity, reslen, status);
    }

    return reslen;
}


U_CAPI void U_EXPORT2
ulocimp_toLanguageTag(const char* localeID,
                      icu::ByteSink& sink,
                      UBool strict,
                      UErrorCode* status) {
    icu::CharString canonical;
    int32_t reslen;
    UErrorCode tmpStatus = U_ZERO_ERROR;
    UBool hadPosix = FALSE;
    const char* pKeywordStart;

    /* Note: uloc_canonicalize returns "en_US_POSIX" for input locale ID "".  See #6835 */
    int32_t resultCapacity = static_cast<int32_t>(uprv_strlen(localeID));
    if (resultCapacity > 0) {
        char* buffer;

        for (;;) {
            buffer = canonical.getAppendBuffer(
                    /*minCapacity=*/resultCapacity,
                    /*desiredCapacityHint=*/resultCapacity,
                    resultCapacity,
                    tmpStatus);

            if (U_FAILURE(tmpStatus)) {
                *status = tmpStatus;
                return;
            }

            reslen =
                uloc_canonicalize(localeID, buffer, resultCapacity, &tmpStatus);

            if (tmpStatus != U_BUFFER_OVERFLOW_ERROR) {
                break;
            }

            resultCapacity = reslen;
            tmpStatus = U_ZERO_ERROR;
        }

        if (U_FAILURE(tmpStatus)) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }

        canonical.append(buffer, reslen, tmpStatus);
        if (tmpStatus == U_STRING_NOT_TERMINATED_WARNING) {
            tmpStatus = U_ZERO_ERROR;  // Terminators provided by CharString.
        }

        if (U_FAILURE(tmpStatus)) {
            *status = tmpStatus;
            return;
        }
    }

    /* For handling special case - private use only tag */
    pKeywordStart = locale_getKeywordsStart(canonical.data());
    if (pKeywordStart == canonical.data()) {
        int kwdCnt = 0;
        UBool done = FALSE;

        icu::LocalUEnumerationPointer kwdEnum(uloc_openKeywords(canonical.data(), &tmpStatus));
        if (U_SUCCESS(tmpStatus)) {
            kwdCnt = uenum_count(kwdEnum.getAlias(), &tmpStatus);
            if (kwdCnt == 1) {
                const char *key;
                int32_t len = 0;

                key = uenum_next(kwdEnum.getAlias(), &len, &tmpStatus);
                if (len == 1 && *key == PRIVATEUSE) {
                    char buf[ULOC_KEYWORD_AND_VALUES_CAPACITY];
                    buf[0] = PRIVATEUSE;
                    buf[1] = SEP;
                    len = uloc_getKeywordValue(localeID, key, &buf[2], sizeof(buf) - 2, &tmpStatus);
                    if (U_SUCCESS(tmpStatus)) {
                        if (ultag_isPrivateuseValueSubtags(&buf[2], len)) {
                            /* return private use only tag */
                            sink.Append(buf, len + 2);
                            done = TRUE;
                        } else if (strict) {
                            *status = U_ILLEGAL_ARGUMENT_ERROR;
                            done = TRUE;
                        }
                        /* if not strict mode, then "und" will be returned */
                    } else {
                        *status = U_ILLEGAL_ARGUMENT_ERROR;
                        done = TRUE;
                    }
                }
            }
            if (done) {
                return;
            }
        }
    }

    _appendLanguageToLanguageTag(canonical.data(), sink, strict, status);
    _appendScriptToLanguageTag(canonical.data(), sink, strict, status);
    _appendRegionToLanguageTag(canonical.data(), sink, strict, status);
    _appendVariantsToLanguageTag(canonical.data(), sink, strict, &hadPosix, status);
    _appendKeywordsToLanguageTag(canonical.data(), sink, strict, hadPosix, status);
    _appendPrivateuseToLanguageTag(canonical.data(), sink, strict, hadPosix, status);
}


U_CAPI int32_t U_EXPORT2
uloc_forLanguageTag(const char* langtag,
                    char* localeID,
                    int32_t localeIDCapacity,
                    int32_t* parsedLength,
                    UErrorCode* status) {
    if (U_FAILURE(*status)) {
        return 0;
    }

    icu::CheckedArrayByteSink sink(localeID, localeIDCapacity);
    ulocimp_forLanguageTag(langtag, -1, sink, parsedLength, status);

    int32_t reslen = sink.NumberOfBytesAppended();

    if (U_FAILURE(*status)) {
        return reslen;
    }

    if (sink.Overflowed()) {
        *status = U_BUFFER_OVERFLOW_ERROR;
    } else {
        u_terminateChars(localeID, localeIDCapacity, reslen, status);
    }

    return reslen;
}


U_CAPI void U_EXPORT2
ulocimp_forLanguageTag(const char* langtag,
                       int32_t tagLen,
                       icu::ByteSink& sink,
                       int32_t* parsedLength,
                       UErrorCode* status) {
    UBool isEmpty = TRUE;
    const char *subtag, *p;
    int32_t len;
    int32_t i, n;
    UBool noRegion = TRUE;

    icu::LocalULanguageTagPointer lt(ultag_parse(langtag, tagLen, parsedLength, status));
    if (U_FAILURE(*status)) {
        return;
    }

    /* language */
    subtag = ultag_getExtlangSize(lt.getAlias()) > 0 ? ultag_getExtlang(lt.getAlias(), 0) : ultag_getLanguage(lt.getAlias());
    if (uprv_compareInvCharsAsAscii(subtag, LANG_UND) != 0) {
        len = (int32_t)uprv_strlen(subtag);
        if (len > 0) {
            sink.Append(subtag, len);
            isEmpty = FALSE;
        }
    }

    /* script */
    subtag = ultag_getScript(lt.getAlias());
    len = (int32_t)uprv_strlen(subtag);
    if (len > 0) {
        sink.Append("_", 1);
        isEmpty = FALSE;

        /* write out the script in title case */
        char c = uprv_toupper(*subtag);
        sink.Append(&c, 1);
        sink.Append(subtag + 1, len - 1);
    }

    /* region */
    subtag = ultag_getRegion(lt.getAlias());
    len = (int32_t)uprv_strlen(subtag);
    if (len > 0) {
        sink.Append("_", 1);
        isEmpty = FALSE;

        /* write out the region in upper case */
        p = subtag;
        while (*p) {
            char c = uprv_toupper(*p);
            sink.Append(&c, 1);
            p++;
        }
        noRegion = FALSE;
    }

    /* variants */
    n = ultag_getVariantsSize(lt.getAlias());
    if (n > 0) {
        if (noRegion) {
            sink.Append("_", 1);
            isEmpty = FALSE;
        }

        for (i = 0; i < n; i++) {
            subtag = ultag_getVariant(lt.getAlias(), i);
            sink.Append("_", 1);

            /* write out the variant in upper case */
            p = subtag;
            while (*p) {
                char c = uprv_toupper(*p);
                sink.Append(&c, 1);
                p++;
            }
        }
    }

    /* keywords */
    n = ultag_getExtensionsSize(lt.getAlias());
    subtag = ultag_getPrivateUse(lt.getAlias());
    if (n > 0 || uprv_strlen(subtag) > 0) {
        if (isEmpty && n > 0) {
            /* need a language */
            sink.Append(LANG_UND, LANG_UND_LEN);
        }
        _appendKeywords(lt.getAlias(), sink, status);
    }
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1997-2006, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
* This file contains platform independent math.
*/



U_CAPI int32_t U_EXPORT2
uprv_max(int32_t x, int32_t y)
{
    return (x > y ? x : y);
}

U_CAPI int32_t U_EXPORT2
uprv_min(int32_t x, int32_t y)
{
    return (x > y ? y : x);
}

// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// umutablecptrie.cpp (inspired by utrie2_builder.cpp)
// created: 2017dec29 Markus W. Scherer

// #define UCPTRIE_DEBUG
#ifdef UCPTRIE_DEBUG
#   include <stdio.h>
#endif










// ICU-20235 In case Microsoft math.h has defined this, undefine it.
#ifdef OVERFLOW
#undef OVERFLOW
#endif

U_NAMESPACE_BEGIN

namespace {

constexpr int32_t MAX_UNICODE = 0x10ffff;

constexpr int32_t UNICODE_LIMIT = 0x110000;
constexpr int32_t BMP_LIMIT = 0x10000;
constexpr int32_t ASCII_LIMIT = 0x80;

constexpr int32_t I_LIMIT = UNICODE_LIMIT >> UCPTRIE_SHIFT_3;
constexpr int32_t BMP_I_LIMIT = BMP_LIMIT >> UCPTRIE_SHIFT_3;
constexpr int32_t ASCII_I_LIMIT = ASCII_LIMIT >> UCPTRIE_SHIFT_3;

constexpr int32_t SMALL_DATA_BLOCKS_PER_BMP_BLOCK = (1 << (UCPTRIE_FAST_SHIFT - UCPTRIE_SHIFT_3));

// Flag values for data blocks.
constexpr uint8_t ALL_SAME = 0;
constexpr uint8_t MIXED = 1;
constexpr uint8_t SAME_AS = 2;

/** Start with allocation of 16k data entries. */
constexpr int32_t INITIAL_DATA_LENGTH = ((int32_t)1 << 14);

/** Grow about 8x each time. */
constexpr int32_t MEDIUM_DATA_LENGTH = ((int32_t)1 << 17);

/**
 * Maximum length of the build-time data array.
 * One entry per 0x110000 code points.
 */
constexpr int32_t MAX_DATA_LENGTH = UNICODE_LIMIT;

// Flag values for index-3 blocks while compacting/building.
constexpr uint8_t I3_NULL = 0;
constexpr uint8_t I3_BMP = 1;
constexpr uint8_t I3_16 = 2;
constexpr uint8_t I3_18 = 3;

constexpr int32_t INDEX_3_18BIT_BLOCK_LENGTH = UCPTRIE_INDEX_3_BLOCK_LENGTH + UCPTRIE_INDEX_3_BLOCK_LENGTH / 8;

class AllSameBlocks;
class MixedBlocks;

class MutableCodePointTrie : public UMemory {
public:
    MutableCodePointTrie(uint32_t initialValue, uint32_t errorValue, UErrorCode &errorCode);
    MutableCodePointTrie(const MutableCodePointTrie &other, UErrorCode &errorCode);
    MutableCodePointTrie(const MutableCodePointTrie &other) = delete;
    ~MutableCodePointTrie();

    MutableCodePointTrie &operator=(const MutableCodePointTrie &other) = delete;

    static MutableCodePointTrie *fromUCPMap(const UCPMap *map, UErrorCode &errorCode);
    static MutableCodePointTrie *fromUCPTrie(const UCPTrie *trie, UErrorCode &errorCode);

    uint32_t get(UChar32 c) const;
    int32_t getRange(UChar32 start, UCPMapValueFilter *filter, const void *context,
                     uint32_t *pValue) const;

    void set(UChar32 c, uint32_t value, UErrorCode &errorCode);
    void setRange(UChar32 start, UChar32 end, uint32_t value, UErrorCode &errorCode);

    UCPTrie *build(UCPTrieType type, UCPTrieValueWidth valueWidth, UErrorCode &errorCode);

private:
    void clear();

    bool ensureHighStart(UChar32 c);
    int32_t allocDataBlock(int32_t blockLength);
    int32_t getDataBlock(int32_t i);

    void maskValues(uint32_t mask);
    UChar32 findHighStart() const;
    int32_t compactWholeDataBlocks(int32_t fastILimit, AllSameBlocks &allSameBlocks);
    int32_t compactData(
            int32_t fastILimit, uint32_t *newData, int32_t newDataCapacity,
            int32_t dataNullIndex, MixedBlocks &mixedBlocks, UErrorCode &errorCode);
    int32_t compactIndex(int32_t fastILimit, MixedBlocks &mixedBlocks, UErrorCode &errorCode);
    int32_t compactTrie(int32_t fastILimit, UErrorCode &errorCode);

    uint32_t *index = nullptr;
    int32_t indexCapacity = 0;
    int32_t index3NullOffset = -1;
    uint32_t *data = nullptr;
    int32_t dataCapacity = 0;
    int32_t dataLength = 0;
    int32_t dataNullOffset = -1;

    uint32_t origInitialValue;
    uint32_t initialValue;
    uint32_t errorValue;
    UChar32 highStart;
    uint32_t highValue;
#ifdef UCPTRIE_DEBUG
public:
    const char *name;
#endif
private:
    /** Temporary array while building the final data. */
    uint16_t *index16 = nullptr;
    uint8_t flags[UNICODE_LIMIT >> UCPTRIE_SHIFT_3];
};

MutableCodePointTrie::MutableCodePointTrie(uint32_t iniValue, uint32_t errValue, UErrorCode &errorCode) :
        origInitialValue(iniValue), initialValue(iniValue), errorValue(errValue),
        highStart(0), highValue(initialValue)
#ifdef UCPTRIE_DEBUG
        , name("open")
#endif
        {
    if (U_FAILURE(errorCode)) { return; }
    index = (uint32_t *)uprv_malloc(BMP_I_LIMIT * 4);
    data = (uint32_t *)uprv_malloc(INITIAL_DATA_LENGTH * 4);
    if (index == nullptr || data == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    indexCapacity = BMP_I_LIMIT;
    dataCapacity = INITIAL_DATA_LENGTH;
}

MutableCodePointTrie::MutableCodePointTrie(const MutableCodePointTrie &other, UErrorCode &errorCode) :
        index3NullOffset(other.index3NullOffset),
        dataNullOffset(other.dataNullOffset),
        origInitialValue(other.origInitialValue), initialValue(other.initialValue),
        errorValue(other.errorValue),
        highStart(other.highStart), highValue(other.highValue)
#ifdef UCPTRIE_DEBUG
        , name("mutable clone")
#endif
        {
    if (U_FAILURE(errorCode)) { return; }
    int32_t iCapacity = highStart <= BMP_LIMIT ? BMP_I_LIMIT : I_LIMIT;
    index = (uint32_t *)uprv_malloc(iCapacity * 4);
    data = (uint32_t *)uprv_malloc(other.dataCapacity * 4);
    if (index == nullptr || data == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    indexCapacity = iCapacity;
    dataCapacity = other.dataCapacity;

    int32_t iLimit = highStart >> UCPTRIE_SHIFT_3;
    uprv_memcpy(flags, other.flags, iLimit);
    uprv_memcpy(index, other.index, iLimit * 4);
    uprv_memcpy(data, other.data, (size_t)other.dataLength * 4);
    dataLength = other.dataLength;
    U_ASSERT(other.index16 == nullptr);
}

MutableCodePointTrie::~MutableCodePointTrie() {
    uprv_free(index);
    uprv_free(data);
    uprv_free(index16);
}

MutableCodePointTrie *MutableCodePointTrie::fromUCPMap(const UCPMap *map, UErrorCode &errorCode) {
    // Use the highValue as the initialValue to reduce the highStart.
    uint32_t errorValue = ucpmap_get(map, -1);
    uint32_t initialValue = ucpmap_get(map, 0x10ffff);
    LocalPointer<MutableCodePointTrie> mutableTrie(
        new MutableCodePointTrie(initialValue, errorValue, errorCode),
        errorCode);
    if (U_FAILURE(errorCode)) {
        return nullptr;
    }
    UChar32 start = 0, end;
    uint32_t value;
    while ((end = ucpmap_getRange(map, start, UCPMAP_RANGE_NORMAL, 0,
                                  nullptr, nullptr, &value)) >= 0) {
        if (value != initialValue) {
            if (start == end) {
                mutableTrie->set(start, value, errorCode);
            } else {
                mutableTrie->setRange(start, end, value, errorCode);
            }
        }
        start = end + 1;
    }
    if (U_SUCCESS(errorCode)) {
        return mutableTrie.orphan();
    } else {
        return nullptr;
    }
}

MutableCodePointTrie *MutableCodePointTrie::fromUCPTrie(const UCPTrie *trie, UErrorCode &errorCode) {
    // Use the highValue as the initialValue to reduce the highStart.
    uint32_t errorValue;
    uint32_t initialValue;
    switch (trie->valueWidth) {
    case UCPTRIE_VALUE_BITS_16:
        errorValue = trie->data.ptr16[trie->dataLength - UCPTRIE_ERROR_VALUE_NEG_DATA_OFFSET];
        initialValue = trie->data.ptr16[trie->dataLength - UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET];
        break;
    case UCPTRIE_VALUE_BITS_32:
        errorValue = trie->data.ptr32[trie->dataLength - UCPTRIE_ERROR_VALUE_NEG_DATA_OFFSET];
        initialValue = trie->data.ptr32[trie->dataLength - UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET];
        break;
    case UCPTRIE_VALUE_BITS_8:
        errorValue = trie->data.ptr8[trie->dataLength - UCPTRIE_ERROR_VALUE_NEG_DATA_OFFSET];
        initialValue = trie->data.ptr8[trie->dataLength - UCPTRIE_HIGH_VALUE_NEG_DATA_OFFSET];
        break;
    default:
        // Unreachable if the trie is properly initialized.
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    LocalPointer<MutableCodePointTrie> mutableTrie(
        new MutableCodePointTrie(initialValue, errorValue, errorCode),
        errorCode);
    if (U_FAILURE(errorCode)) {
        return nullptr;
    }
    UChar32 start = 0, end;
    uint32_t value;
    while ((end = ucptrie_getRange(trie, start, UCPMAP_RANGE_NORMAL, 0,
                                   nullptr, nullptr, &value)) >= 0) {
        if (value != initialValue) {
            if (start == end) {
                mutableTrie->set(start, value, errorCode);
            } else {
                mutableTrie->setRange(start, end, value, errorCode);
            }
        }
        start = end + 1;
    }
    if (U_SUCCESS(errorCode)) {
        return mutableTrie.orphan();
    } else {
        return nullptr;
    }
}

void MutableCodePointTrie::clear() {
    index3NullOffset = dataNullOffset = -1;
    dataLength = 0;
    highValue = initialValue = origInitialValue;
    highStart = 0;
    uprv_free(index16);
    index16 = nullptr;
}

uint32_t MutableCodePointTrie::get(UChar32 c) const {
    if ((uint32_t)c > MAX_UNICODE) {
        return errorValue;
    }
    if (c >= highStart) {
        return highValue;
    }
    int32_t i = c >> UCPTRIE_SHIFT_3;
    if (flags[i] == ALL_SAME) {
        return index[i];
    } else {
        return data[index[i] + (c & UCPTRIE_SMALL_DATA_MASK)];
    }
}

inline uint32_t maybeFilterValue(uint32_t value, uint32_t initialValue, uint32_t nullValue,
                                 UCPMapValueFilter *filter, const void *context) {
    if (value == initialValue) {
        value = nullValue;
    } else if (filter != nullptr) {
        value = filter(context, value);
    }
    return value;
}

UChar32 MutableCodePointTrie::getRange(
        UChar32 start, UCPMapValueFilter *filter, const void *context,
        uint32_t *pValue) const {
    if ((uint32_t)start > MAX_UNICODE) {
        return U_SENTINEL;
    }
    if (start >= highStart) {
        if (pValue != nullptr) {
            uint32_t value = highValue;
            if (filter != nullptr) { value = filter(context, value); }
            *pValue = value;
        }
        return MAX_UNICODE;
    }
    uint32_t nullValue = initialValue;
    if (filter != nullptr) { nullValue = filter(context, nullValue); }
    UChar32 c = start;
    uint32_t trieValue, value;
    bool haveValue = false;
    int32_t i = c >> UCPTRIE_SHIFT_3;
    do {
        if (flags[i] == ALL_SAME) {
            uint32_t trieValue2 = index[i];
            if (haveValue) {
                if (trieValue2 != trieValue) {
                    if (filter == nullptr ||
                            maybeFilterValue(trieValue2, initialValue, nullValue,
                                             filter, context) != value) {
                        return c - 1;
                    }
                    trieValue = trieValue2;  // may or may not help
                }
            } else {
                trieValue = trieValue2;
                value = maybeFilterValue(trieValue2, initialValue, nullValue, filter, context);
                if (pValue != nullptr) { *pValue = value; }
                haveValue = true;
            }
            c = (c + UCPTRIE_SMALL_DATA_BLOCK_LENGTH) & ~UCPTRIE_SMALL_DATA_MASK;
        } else /* MIXED */ {
            int32_t di = index[i] + (c & UCPTRIE_SMALL_DATA_MASK);
            uint32_t trieValue2 = data[di];
            if (haveValue) {
                if (trieValue2 != trieValue) {
                    if (filter == nullptr ||
                            maybeFilterValue(trieValue2, initialValue, nullValue,
                                             filter, context) != value) {
                        return c - 1;
                    }
                    trieValue = trieValue2;  // may or may not help
                }
            } else {
                trieValue = trieValue2;
                value = maybeFilterValue(trieValue2, initialValue, nullValue, filter, context);
                if (pValue != nullptr) { *pValue = value; }
                haveValue = true;
            }
            while ((++c & UCPTRIE_SMALL_DATA_MASK) != 0) {
                trieValue2 = data[++di];
                if (trieValue2 != trieValue) {
                    if (filter == nullptr ||
                            maybeFilterValue(trieValue2, initialValue, nullValue,
                                             filter, context) != value) {
                        return c - 1;
                    }
                }
                trieValue = trieValue2;  // may or may not help
            }
        }
        ++i;
    } while (c < highStart);
    U_ASSERT(haveValue);
    if (maybeFilterValue(highValue, initialValue, nullValue,
                         filter, context) != value) {
        return c - 1;
    } else {
        return MAX_UNICODE;
    }
}

void
umutablecptrie_writeBlock(uint32_t *block, uint32_t value) {
    uint32_t *limit = block + UCPTRIE_SMALL_DATA_BLOCK_LENGTH;
    while (block < limit) {
        *block++ = value;
    }
}

bool MutableCodePointTrie::ensureHighStart(UChar32 c) {
    if (c >= highStart) {
        // Round up to a UCPTRIE_CP_PER_INDEX_2_ENTRY boundary to simplify compaction.
        c = (c + UCPTRIE_CP_PER_INDEX_2_ENTRY) & ~(UCPTRIE_CP_PER_INDEX_2_ENTRY - 1);
        int32_t i = highStart >> UCPTRIE_SHIFT_3;
        int32_t iLimit = c >> UCPTRIE_SHIFT_3;
        if (iLimit > indexCapacity) {
            uint32_t *newIndex = (uint32_t *)uprv_malloc(I_LIMIT * 4);
            if (newIndex == nullptr) { return false; }
            uprv_memcpy(newIndex, index, i * 4);
            uprv_free(index);
            index = newIndex;
            indexCapacity = I_LIMIT;
        }
        do {
            flags[i] = ALL_SAME;
            index[i] = initialValue;
        } while(++i < iLimit);
        highStart = c;
    }
    return true;
}

int32_t MutableCodePointTrie::allocDataBlock(int32_t blockLength) {
    int32_t newBlock = dataLength;
    int32_t newTop = newBlock + blockLength;
    if (newTop > dataCapacity) {
        int32_t capacity;
        if (dataCapacity < MEDIUM_DATA_LENGTH) {
            capacity = MEDIUM_DATA_LENGTH;
        } else if (dataCapacity < MAX_DATA_LENGTH) {
            capacity = MAX_DATA_LENGTH;
        } else {
            // Should never occur.
            // Either MAX_DATA_LENGTH is incorrect,
            // or the code writes more values than should be possible.
            return -1;
        }
        uint32_t *newData = (uint32_t *)uprv_malloc(capacity * 4);
        if (newData == nullptr) {
            return -1;
        }
        uprv_memcpy(newData, data, (size_t)dataLength * 4);
        uprv_free(data);
        data = newData;
        dataCapacity = capacity;
    }
    dataLength = newTop;
    return newBlock;
}

/**
 * No error checking for illegal arguments.
 *
 * @return -1 if no new data block available (out of memory in data array)
 * @internal
 */
int32_t MutableCodePointTrie::getDataBlock(int32_t i) {
    if (flags[i] == MIXED) {
        return index[i];
    }
    if (i < BMP_I_LIMIT) {
        int32_t newBlock = allocDataBlock(UCPTRIE_FAST_DATA_BLOCK_LENGTH);
        if (newBlock < 0) { return newBlock; }
        int32_t iStart = i & ~(SMALL_DATA_BLOCKS_PER_BMP_BLOCK -1);
        int32_t iLimit = iStart + SMALL_DATA_BLOCKS_PER_BMP_BLOCK;
        do {
            U_ASSERT(flags[iStart] == ALL_SAME);
            umutablecptrie_writeBlock(data + newBlock, index[iStart]);
            flags[iStart] = MIXED;
            index[iStart++] = newBlock;
            newBlock += UCPTRIE_SMALL_DATA_BLOCK_LENGTH;
        } while (iStart < iLimit);
        return index[i];
    } else {
        int32_t newBlock = allocDataBlock(UCPTRIE_SMALL_DATA_BLOCK_LENGTH);
        if (newBlock < 0) { return newBlock; }
        umutablecptrie_writeBlock(data + newBlock, index[i]);
        flags[i] = MIXED;
        index[i] = newBlock;
        return newBlock;
    }
}

void MutableCodePointTrie::set(UChar32 c, uint32_t value, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) {
        return;
    }
    if ((uint32_t)c > MAX_UNICODE) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    int32_t block;
    if (!ensureHighStart(c) || (block = getDataBlock(c >> UCPTRIE_SHIFT_3)) < 0) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    data[block + (c & UCPTRIE_SMALL_DATA_MASK)] = value;
}

void
fillBlock(uint32_t *block, UChar32 start, UChar32 limit, uint32_t value) {
    uint32_t *pLimit = block + limit;
    block += start;
    while (block < pLimit) {
        *block++ = value;
    }
}

void MutableCodePointTrie::setRange(UChar32 start, UChar32 end, uint32_t value, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) {
        return;
    }
    if ((uint32_t)start > MAX_UNICODE || (uint32_t)end > MAX_UNICODE || start > end) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if (!ensureHighStart(end)) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    UChar32 limit = end + 1;
    if (start & UCPTRIE_SMALL_DATA_MASK) {
        // Set partial block at [start..following block boundary[.
        int32_t block = getDataBlock(start >> UCPTRIE_SHIFT_3);
        if (block < 0) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
        }

        UChar32 nextStart = (start + UCPTRIE_SMALL_DATA_MASK) & ~UCPTRIE_SMALL_DATA_MASK;
        if (nextStart <= limit) {
            fillBlock(data + block, start & UCPTRIE_SMALL_DATA_MASK, UCPTRIE_SMALL_DATA_BLOCK_LENGTH,
                      value);
            start = nextStart;
        } else {
            fillBlock(data + block, start & UCPTRIE_SMALL_DATA_MASK, limit & UCPTRIE_SMALL_DATA_MASK,
                      value);
            return;
        }
    }

    // Number of positions in the last, partial block.
    int32_t rest = limit & UCPTRIE_SMALL_DATA_MASK;

    // Round down limit to a block boundary.
    limit &= ~UCPTRIE_SMALL_DATA_MASK;

    // Iterate over all-value blocks.
    while (start < limit) {
        int32_t i = start >> UCPTRIE_SHIFT_3;
        if (flags[i] == ALL_SAME) {
            index[i] = value;
        } else /* MIXED */ {
            fillBlock(data + index[i], 0, UCPTRIE_SMALL_DATA_BLOCK_LENGTH, value);
        }
        start += UCPTRIE_SMALL_DATA_BLOCK_LENGTH;
    }

    if (rest > 0) {
        // Set partial block at [last block boundary..limit[.
        int32_t block = getDataBlock(start >> UCPTRIE_SHIFT_3);
        if (block < 0) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
        }

        fillBlock(data + block, 0, rest, value);
    }
}

/* compaction --------------------------------------------------------------- */

void MutableCodePointTrie::maskValues(uint32_t mask) {
    initialValue &= mask;
    errorValue &= mask;
    highValue &= mask;
    int32_t iLimit = highStart >> UCPTRIE_SHIFT_3;
    for (int32_t i = 0; i < iLimit; ++i) {
        if (flags[i] == ALL_SAME) {
            index[i] &= mask;
        }
    }
    for (int32_t i = 0; i < dataLength; ++i) {
        data[i] &= mask;
    }
}

template<typename UIntA, typename UIntB>
bool equalBlocks(const UIntA *s, const UIntB *t, int32_t length) {
    while (length > 0 && *s == *t) {
        ++s;
        ++t;
        --length;
    }
    return length == 0;
}

bool allValuesSameAs(const uint32_t *p, int32_t length, uint32_t value) {
    const uint32_t *pLimit = p + length;
    while (p < pLimit && *p == value) { ++p; }
    return p == pLimit;
}

/** Search for an identical block. */
int32_t findSameBlock(const uint16_t *p, int32_t pStart, int32_t length,
                      const uint16_t *q, int32_t qStart, int32_t blockLength) {
    // Ensure that we do not even partially get past length.
    length -= blockLength;

    q += qStart;
    while (pStart <= length) {
        if (equalBlocks(p + pStart, q, blockLength)) {
            return pStart;
        }
        ++pStart;
    }
    return -1;
}

int32_t findAllSameBlock(const uint32_t *p, int32_t start, int32_t limit,
                         uint32_t value, int32_t blockLength) {
    // Ensure that we do not even partially get past limit.
    limit -= blockLength;

    for (int32_t block = start; block <= limit; ++block) {
        if (p[block] == value) {
            for (int32_t i = 1;; ++i) {
                if (i == blockLength) {
                    return block;
                }
                if (p[block + i] != value) {
                    block += i;
                    break;
                }
            }
        }
    }
    return -1;
}

/**
 * Look for maximum overlap of the beginning of the other block
 * with the previous, adjacent block.
 */
template<typename UIntA, typename UIntB>
int32_t getOverlap(const UIntA *p, int32_t length,
                   const UIntB *q, int32_t qStart, int32_t blockLength) {
    int32_t overlap = blockLength - 1;
    U_ASSERT(overlap <= length);
    q += qStart;
    while (overlap > 0 && !equalBlocks(p + (length - overlap), q, overlap)) {
        --overlap;
    }
    return overlap;
}

int32_t getAllSameOverlap(const uint32_t *p, int32_t length, uint32_t value,
                          int32_t blockLength) {
    int32_t min = length - (blockLength - 1);
    int32_t i = length;
    while (min < i && p[i - 1] == value) { --i; }
    return length - i;
}

bool isStartOfSomeFastBlock(uint32_t dataOffset, const uint32_t index[], int32_t fastILimit) {
    for (int32_t i = 0; i < fastILimit; i += SMALL_DATA_BLOCKS_PER_BMP_BLOCK) {
        if (index[i] == dataOffset) {
            return true;
        }
    }
    return false;
}

/**
 * Finds the start of the last range in the trie by enumerating backward.
 * Indexes for code points higher than this will be omitted.
 */
UChar32 MutableCodePointTrie::findHighStart() const {
    int32_t i = highStart >> UCPTRIE_SHIFT_3;
    while (i > 0) {
        bool match;
        if (flags[--i] == ALL_SAME) {
            match = index[i] == highValue;
        } else /* MIXED */ {
            const uint32_t *p = data + index[i];
            for (int32_t j = 0;; ++j) {
                if (j == UCPTRIE_SMALL_DATA_BLOCK_LENGTH) {
                    match = true;
                    break;
                }
                if (p[j] != highValue) {
                    match = false;
                    break;
                }
            }
        }
        if (!match) {
            return (i + 1) << UCPTRIE_SHIFT_3;
        }
    }
    return 0;
}

class AllSameBlocks {
public:
    static constexpr int32_t NEW_UNIQUE = -1;
    static constexpr int32_t OVERFLOW = -2;

    AllSameBlocks() : length(0), mostRecent(-1) {}

    int32_t findOrAdd(int32_t index, int32_t count, uint32_t value) {
        if (mostRecent >= 0 && values[mostRecent] == value) {
            refCounts[mostRecent] += count;
            return indexes[mostRecent];
        }
        for (int32_t i = 0; i < length; ++i) {
            if (values[i] == value) {
                mostRecent = i;
                refCounts[i] += count;
                return indexes[i];
            }
        }
        if (length == CAPACITY) {
            return OVERFLOW;
        }
        mostRecent = length;
        indexes[length] = index;
        values[length] = value;
        refCounts[length++] = count;
        return NEW_UNIQUE;
    }

    /** Replaces the block which has the lowest reference count. */
    void add(int32_t index, int32_t count, uint32_t value) {
        U_ASSERT(length == CAPACITY);
        int32_t least = -1;
        int32_t leastCount = I_LIMIT;
        for (int32_t i = 0; i < length; ++i) {
            U_ASSERT(values[i] != value);
            if (refCounts[i] < leastCount) {
                least = i;
                leastCount = refCounts[i];
            }
        }
        U_ASSERT(least >= 0);
        mostRecent = least;
        indexes[least] = index;
        values[least] = value;
        refCounts[least] = count;
    }

    int32_t findMostUsed() const {
        if (length == 0) { return -1; }
        int32_t max = -1;
        int32_t maxCount = 0;
        for (int32_t i = 0; i < length; ++i) {
            if (refCounts[i] > maxCount) {
                max = i;
                maxCount = refCounts[i];
            }
        }
        return indexes[max];
    }

private:
    static constexpr int32_t CAPACITY = 32;

    int32_t length;
    int32_t mostRecent;

    int32_t indexes[CAPACITY];
    uint32_t values[CAPACITY];
    int32_t refCounts[CAPACITY];
};

// Custom hash table for mixed-value blocks to be found anywhere in the
// compacted data or index so far.
class MixedBlocks {
public:
    MixedBlocks() {}
    ~MixedBlocks() {
        uprv_free(table);
    }

    bool init(int32_t maxLength, int32_t newBlockLength) {
        // We store actual data indexes + 1 to reserve 0 for empty entries.
        int32_t maxDataIndex = maxLength - newBlockLength + 1;
        int32_t newLength;
        if (maxDataIndex <= 0xfff) {  // 4k
            newLength = 6007;
            shift = 12;
            mask = 0xfff;
        } else if (maxDataIndex <= 0x7fff) {  // 32k
            newLength = 50021;
            shift = 15;
            mask = 0x7fff;
        } else if (maxDataIndex <= 0x1ffff) {  // 128k
            newLength = 200003;
            shift = 17;
            mask = 0x1ffff;
        } else {
            // maxDataIndex up to around MAX_DATA_LENGTH, ca. 1.1M
            newLength = 1500007;
            shift = 21;
            mask = 0x1fffff;
        }
        if (newLength > capacity) {
            uprv_free(table);
            table = (uint32_t *)uprv_malloc(newLength * 4);
            if (table == nullptr) {
                return false;
            }
            capacity = newLength;
        }
        length = newLength;
        uprv_memset(table, 0, length * 4);

        blockLength = newBlockLength;
        return true;
    }

    template<typename UInt>
    void extend(const UInt *data, int32_t minStart, int32_t prevDataLength, int32_t newDataLength) {
        int32_t start = prevDataLength - blockLength;
        if (start >= minStart) {
            ++start;  // Skip the last block that we added last time.
        } else {
            start = minStart;  // Begin with the first full block.
        }
        for (int32_t end = newDataLength - blockLength; start <= end; ++start) {
            uint32_t hashCode = makeHashCode(data, start);
            addEntry(data, start, hashCode, start);
        }
    }

    template<typename UIntA, typename UIntB>
    int32_t findBlock(const UIntA *data, const UIntB *blockData, int32_t blockStart) const {
        uint32_t hashCode = makeHashCode(blockData, blockStart);
        int32_t entryIndex = findEntry(data, blockData, blockStart, hashCode);
        if (entryIndex >= 0) {
            return (table[entryIndex] & mask) - 1;
        } else {
            return -1;
        }
    }

    int32_t findAllSameBlock(const uint32_t *data, uint32_t blockValue) const {
        uint32_t hashCode = makeHashCode(blockValue);
        int32_t entryIndex = findEntry(data, blockValue, hashCode);
        if (entryIndex >= 0) {
            return (table[entryIndex] & mask) - 1;
        } else {
            return -1;
        }
    }

private:
    template<typename UInt>
    uint32_t makeHashCode(const UInt *blockData, int32_t blockStart) const {
        int32_t blockLimit = blockStart + blockLength;
        uint32_t hashCode = blockData[blockStart++];
        do {
            hashCode = 37 * hashCode + blockData[blockStart++];
        } while (blockStart < blockLimit);
        return hashCode;
    }

    uint32_t makeHashCode(uint32_t blockValue) const {
        uint32_t hashCode = blockValue;
        for (int32_t i = 1; i < blockLength; ++i) {
            hashCode = 37 * hashCode + blockValue;
        }
        return hashCode;
    }

    template<typename UInt>
    void addEntry(const UInt *data, int32_t blockStart, uint32_t hashCode, int32_t dataIndex) {
        U_ASSERT(0 <= dataIndex && dataIndex < (int32_t)mask);
        int32_t entryIndex = findEntry(data, data, blockStart, hashCode);
        if (entryIndex < 0) {
            table[~entryIndex] = (hashCode << shift) | (dataIndex + 1);
        }
    }

    template<typename UIntA, typename UIntB>
    int32_t findEntry(const UIntA *data, const UIntB *blockData, int32_t blockStart,
                      uint32_t hashCode) const {
        uint32_t shiftedHashCode = hashCode << shift;
        int32_t initialEntryIndex = (hashCode % (length - 1)) + 1;  // 1..length-1
        for (int32_t entryIndex = initialEntryIndex;;) {
            uint32_t entry = table[entryIndex];
            if (entry == 0) {
                return ~entryIndex;
            }
            if ((entry & ~mask) == shiftedHashCode) {
                int32_t dataIndex = (entry & mask) - 1;
                if (equalBlocks(data + dataIndex, blockData + blockStart, blockLength)) {
                    return entryIndex;
                }
            }
            entryIndex = nextIndex(initialEntryIndex, entryIndex);
        }
    }

    int32_t findEntry(const uint32_t *data, uint32_t blockValue, uint32_t hashCode) const {
        uint32_t shiftedHashCode = hashCode << shift;
        int32_t initialEntryIndex = (hashCode % (length - 1)) + 1;  // 1..length-1
        for (int32_t entryIndex = initialEntryIndex;;) {
            uint32_t entry = table[entryIndex];
            if (entry == 0) {
                return ~entryIndex;
            }
            if ((entry & ~mask) == shiftedHashCode) {
                int32_t dataIndex = (entry & mask) - 1;
                if (allValuesSameAs(data + dataIndex, blockLength, blockValue)) {
                    return entryIndex;
                }
            }
            entryIndex = nextIndex(initialEntryIndex, entryIndex);
        }
    }

    inline int32_t nextIndex(int32_t initialEntryIndex, int32_t entryIndex) const {
        // U_ASSERT(0 < initialEntryIndex && initialEntryIndex < length);
        return (entryIndex + initialEntryIndex) % length;
    }

    // Hash table.
    // The length is a prime number, larger than the maximum data length.
    // The "shift" lower bits store a data index + 1.
    // The remaining upper bits store a partial hashCode of the block data values.
    uint32_t *table = nullptr;
    int32_t capacity = 0;
    int32_t length = 0;
    int32_t shift = 0;
    uint32_t mask = 0;

    int32_t blockLength = 0;
};

int32_t MutableCodePointTrie::compactWholeDataBlocks(int32_t fastILimit, AllSameBlocks &allSameBlocks) {
#ifdef UCPTRIE_DEBUG
    bool overflow = false;
#endif

    // ASCII data will be stored as a linear table, even if the following code
    // does not yet count it that way.
    int32_t newDataCapacity = ASCII_LIMIT;
    // Add room for a small data null block in case it would match the start of
    // a fast data block where dataNullOffset must not be set in that case.
    newDataCapacity += UCPTRIE_SMALL_DATA_BLOCK_LENGTH;
    // Add room for special values (errorValue, highValue) and padding.
    newDataCapacity += 4;
    int32_t iLimit = highStart >> UCPTRIE_SHIFT_3;
    int32_t blockLength = UCPTRIE_FAST_DATA_BLOCK_LENGTH;
    int32_t inc = SMALL_DATA_BLOCKS_PER_BMP_BLOCK;
    for (int32_t i = 0; i < iLimit; i += inc) {
        if (i == fastILimit) {
            blockLength = UCPTRIE_SMALL_DATA_BLOCK_LENGTH;
            inc = 1;
        }
        uint32_t value = index[i];
        if (flags[i] == MIXED) {
            // Really mixed?
            const uint32_t *p = data + value;
            value = *p;
            if (allValuesSameAs(p + 1, blockLength - 1, value)) {
                flags[i] = ALL_SAME;
                index[i] = value;
                // Fall through to ALL_SAME handling.
            } else {
                newDataCapacity += blockLength;
                continue;
            }
        } else {
            U_ASSERT(flags[i] == ALL_SAME);
            if (inc > 1) {
                // Do all of the fast-range data block's ALL_SAME parts have the same value?
                bool allSame = true;
                int32_t next_i = i + inc;
                for (int32_t j = i + 1; j < next_i; ++j) {
                    U_ASSERT(flags[j] == ALL_SAME);
                    if (index[j] != value) {
                        allSame = false;
                        break;
                    }
                }
                if (!allSame) {
                    // Turn it into a MIXED block.
                    if (getDataBlock(i) < 0) {
                        return -1;
                    }
                    newDataCapacity += blockLength;
                    continue;
                }
            }
        }
        // Is there another ALL_SAME block with the same value?
        int32_t other = allSameBlocks.findOrAdd(i, inc, value);
        if (other == AllSameBlocks::OVERFLOW) {
            // The fixed-size array overflowed. Slow check for a duplicate block.
#ifdef UCPTRIE_DEBUG
            if (!overflow) {
                puts("UCPTrie AllSameBlocks overflow");
                overflow = true;
            }
#endif
            int32_t jInc = SMALL_DATA_BLOCKS_PER_BMP_BLOCK;
            for (int32_t j = 0;; j += jInc) {
                if (j == i) {
                    allSameBlocks.add(i, inc, value);
                    break;
                }
                if (j == fastILimit) {
                    jInc = 1;
                }
                if (flags[j] == ALL_SAME && index[j] == value) {
                    allSameBlocks.add(j, jInc + inc, value);
                    other = j;
                    break;
                    // We could keep counting blocks with the same value
                    // before we add the first one, which may improve compaction in rare cases,
                    // but it would make it slower.
                }
            }
        }
        if (other >= 0) {
            flags[i] = SAME_AS;
            index[i] = other;
        } else {
            // New unique same-value block.
            newDataCapacity += blockLength;
        }
    }
    return newDataCapacity;
}

#ifdef UCPTRIE_DEBUG
#   define DEBUG_DO(expr) expr
#else
#   define DEBUG_DO(expr)
#endif

#ifdef UCPTRIE_DEBUG
// Braille symbols: U+28xx = UTF-8 E2 A0 80..E2 A3 BF
int32_t appendValue(char s[], int32_t length, uint32_t value) {
    value ^= value >> 16;
    value ^= value >> 8;
    s[length] = 0xE2;
    s[length + 1] = (char)(0xA0 + ((value >> 6) & 3));
    s[length + 2] = (char)(0x80 + (value & 0x3F));
    return length + 3;
}

void printBlock(const uint32_t *block, int32_t blockLength, uint32_t value,
                UChar32 start, int32_t overlap, uint32_t initialValue) {
    char s[UCPTRIE_FAST_DATA_BLOCK_LENGTH * 3 + 3];
    int32_t length = 0;
    int32_t i;
    for (i = 0; i < overlap; ++i) {
        length = appendValue(s, length, 0);  // Braille blank
    }
    s[length++] = '|';
    for (; i < blockLength; ++i) {
        if (block != nullptr) {
            value = block[i];
        }
        if (value == initialValue) {
            value = 0x40;  // Braille lower left dot
        }
        length = appendValue(s, length, value);
    }
    s[length] = 0;
    start += overlap;
    if (start <= 0xffff) {
        printf("    %04lX  %s|\n", (long)start, s);
    } else if (start <= 0xfffff) {
        printf("   %5lX  %s|\n", (long)start, s);
    } else {
        printf("  %6lX  %s|\n", (long)start, s);
    }
}
#endif

/**
 * Compacts a build-time trie.
 *
 * The compaction
 * - removes blocks that are identical with earlier ones
 * - overlaps each new non-duplicate block as much as possible with the previously-written one
 * - works with fast-range data blocks whose length is a multiple of that of
 *   higher-code-point data blocks
 *
 * It does not try to find an optimal order of writing, deduplicating, and overlapping blocks.
 */
int32_t MutableCodePointTrie::compactData(
        int32_t fastILimit, uint32_t *newData, int32_t newDataCapacity,
        int32_t dataNullIndex, MixedBlocks &mixedBlocks, UErrorCode &errorCode) {
#ifdef UCPTRIE_DEBUG
    int32_t countSame=0, sumOverlaps=0;
    bool printData = dataLength == 29088 /* line.brk */ ||
        // dataLength == 30048 /* CanonIterData */ ||
        dataLength == 50400 /* zh.txt~stroke */;
#endif

    // The linear ASCII data has been copied into newData already.
    int32_t newDataLength = 0;
    for (int32_t i = 0; newDataLength < ASCII_LIMIT;
            newDataLength += UCPTRIE_FAST_DATA_BLOCK_LENGTH, i += SMALL_DATA_BLOCKS_PER_BMP_BLOCK) {
        index[i] = newDataLength;
#ifdef UCPTRIE_DEBUG
        if (printData) {
            printBlock(newData + newDataLength, UCPTRIE_FAST_DATA_BLOCK_LENGTH, 0, newDataLength, 0, initialValue);
        }
#endif
    }

    int32_t blockLength = UCPTRIE_FAST_DATA_BLOCK_LENGTH;
    if (!mixedBlocks.init(newDataCapacity, blockLength)) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    mixedBlocks.extend(newData, 0, 0, newDataLength);

    int32_t iLimit = highStart >> UCPTRIE_SHIFT_3;
    int32_t inc = SMALL_DATA_BLOCKS_PER_BMP_BLOCK;
    int32_t fastLength = 0;
    for (int32_t i = ASCII_I_LIMIT; i < iLimit; i += inc) {
        if (i == fastILimit) {
            blockLength = UCPTRIE_SMALL_DATA_BLOCK_LENGTH;
            inc = 1;
            fastLength = newDataLength;
            if (!mixedBlocks.init(newDataCapacity, blockLength)) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return 0;
            }
            mixedBlocks.extend(newData, 0, 0, newDataLength);
        }
        if (flags[i] == ALL_SAME) {
            uint32_t value = index[i];
            // Find an earlier part of the data array of length blockLength
            // that is filled with this value.
            int32_t n = mixedBlocks.findAllSameBlock(newData, value);
            // If we find a match, and the current block is the data null block,
            // and it is not a fast block but matches the start of a fast block,
            // then we need to continue looking.
            // This is because this small block is shorter than the fast block,
            // and not all of the rest of the fast block is filled with this value.
            // Otherwise trie.getRange() would detect that the fast block starts at
            // dataNullOffset and assume incorrectly that it is filled with the null value.
            while (n >= 0 && i == dataNullIndex && i >= fastILimit && n < fastLength &&
                    isStartOfSomeFastBlock(n, index, fastILimit)) {
                n = findAllSameBlock(newData, n + 1, newDataLength, value, blockLength);
            }
            if (n >= 0) {
                DEBUG_DO(++countSame);
                index[i] = n;
            } else {
                n = getAllSameOverlap(newData, newDataLength, value, blockLength);
                DEBUG_DO(sumOverlaps += n);
#ifdef UCPTRIE_DEBUG
                if (printData) {
                    printBlock(nullptr, blockLength, value, i << UCPTRIE_SHIFT_3, n, initialValue);
                }
#endif
                index[i] = newDataLength - n;
                int32_t prevDataLength = newDataLength;
                while (n < blockLength) {
                    newData[newDataLength++] = value;
                    ++n;
                }
                mixedBlocks.extend(newData, 0, prevDataLength, newDataLength);
            }
        } else if (flags[i] == MIXED) {
            const uint32_t *block = data + index[i];
            int32_t n = mixedBlocks.findBlock(newData, block, 0);
            if (n >= 0) {
                DEBUG_DO(++countSame);
                index[i] = n;
            } else {
                n = getOverlap(newData, newDataLength, block, 0, blockLength);
                DEBUG_DO(sumOverlaps += n);
#ifdef UCPTRIE_DEBUG
                if (printData) {
                    printBlock(block, blockLength, 0, i << UCPTRIE_SHIFT_3, n, initialValue);
                }
#endif
                index[i] = newDataLength - n;
                int32_t prevDataLength = newDataLength;
                while (n < blockLength) {
                    newData[newDataLength++] = block[n++];
                }
                mixedBlocks.extend(newData, 0, prevDataLength, newDataLength);
            }
        } else /* SAME_AS */ {
            uint32_t j = index[i];
            index[i] = index[j];
        }
    }

#ifdef UCPTRIE_DEBUG
    /* we saved some space */
    printf("compacting UCPTrie: count of 32-bit data words %lu->%lu  countSame=%ld  sumOverlaps=%ld\n",
            (long)dataLength, (long)newDataLength, (long)countSame, (long)sumOverlaps);
#endif
    return newDataLength;
}

int32_t MutableCodePointTrie::compactIndex(int32_t fastILimit, MixedBlocks &mixedBlocks,
                                           UErrorCode &errorCode) {
    int32_t fastIndexLength = fastILimit >> (UCPTRIE_FAST_SHIFT - UCPTRIE_SHIFT_3);
    if ((highStart >> UCPTRIE_FAST_SHIFT) <= fastIndexLength) {
        // Only the linear fast index, no multi-stage index tables.
        index3NullOffset = UCPTRIE_NO_INDEX3_NULL_OFFSET;
        return fastIndexLength;
    }

    // Condense the fast index table.
    // Also, does it contain an index-3 block with all dataNullOffset?
    uint16_t fastIndex[UCPTRIE_BMP_INDEX_LENGTH];  // fastIndexLength
    int32_t i3FirstNull = -1;
    for (int32_t i = 0, j = 0; i < fastILimit; ++j) {
        uint32_t i3 = index[i];
        fastIndex[j] = (uint16_t)i3;
        if (i3 == (uint32_t)dataNullOffset) {
            if (i3FirstNull < 0) {
                i3FirstNull = j;
            } else if (index3NullOffset < 0 &&
                    (j - i3FirstNull + 1) == UCPTRIE_INDEX_3_BLOCK_LENGTH) {
                index3NullOffset = i3FirstNull;
            }
        } else {
            i3FirstNull = -1;
        }
        // Set the index entries that compactData() skipped.
        // Needed when the multi-stage index covers the fast index range as well.
        int32_t iNext = i + SMALL_DATA_BLOCKS_PER_BMP_BLOCK;
        while (++i < iNext) {
            i3 += UCPTRIE_SMALL_DATA_BLOCK_LENGTH;
            index[i] = i3;
        }
    }

    if (!mixedBlocks.init(fastIndexLength, UCPTRIE_INDEX_3_BLOCK_LENGTH)) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    mixedBlocks.extend(fastIndex, 0, 0, fastIndexLength);

    // Examine index-3 blocks. For each determine one of:
    // - same as the index-3 null block
    // - same as a fast-index block
    // - 16-bit indexes
    // - 18-bit indexes
    // We store this in the first flags entry for the index-3 block.
    //
    // Also determine an upper limit for the index-3 table length.
    int32_t index3Capacity = 0;
    i3FirstNull = index3NullOffset;
    bool hasLongI3Blocks = false;
    // If the fast index covers the whole BMP, then
    // the multi-stage index is only for supplementary code points.
    // Otherwise, the multi-stage index covers all of Unicode.
    int32_t iStart = fastILimit < BMP_I_LIMIT ? 0 : BMP_I_LIMIT;
    int32_t iLimit = highStart >> UCPTRIE_SHIFT_3;
    for (int32_t i = iStart; i < iLimit;) {
        int32_t j = i;
        int32_t jLimit = i + UCPTRIE_INDEX_3_BLOCK_LENGTH;
        uint32_t oredI3 = 0;
        bool isNull = true;
        do {
            uint32_t i3 = index[j];
            oredI3 |= i3;
            if (i3 != (uint32_t)dataNullOffset) {
                isNull = false;
            }
        } while (++j < jLimit);
        if (isNull) {
            flags[i] = I3_NULL;
            if (i3FirstNull < 0) {
                if (oredI3 <= 0xffff) {
                    index3Capacity += UCPTRIE_INDEX_3_BLOCK_LENGTH;
                } else {
                    index3Capacity += INDEX_3_18BIT_BLOCK_LENGTH;
                    hasLongI3Blocks = true;
                }
                i3FirstNull = 0;
            }
        } else {
            if (oredI3 <= 0xffff) {
                int32_t n = mixedBlocks.findBlock(fastIndex, index, i);
                if (n >= 0) {
                    flags[i] = I3_BMP;
                    index[i] = n;
                } else {
                    flags[i] = I3_16;
                    index3Capacity += UCPTRIE_INDEX_3_BLOCK_LENGTH;
                }
            } else {
                flags[i] = I3_18;
                index3Capacity += INDEX_3_18BIT_BLOCK_LENGTH;
                hasLongI3Blocks = true;
            }
        }
        i = j;
    }

    int32_t index2Capacity = (iLimit - iStart) >> UCPTRIE_SHIFT_2_3;

    // Length of the index-1 table, rounded up.
    int32_t index1Length = (index2Capacity + UCPTRIE_INDEX_2_MASK) >> UCPTRIE_SHIFT_1_2;

    // Index table: Fast index, index-1, index-3, index-2.
    // +1 for possible index table padding.
    int32_t index16Capacity = fastIndexLength + index1Length + index3Capacity + index2Capacity + 1;
    index16 = (uint16_t *)uprv_malloc(index16Capacity * 2);
    if (index16 == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    uprv_memcpy(index16, fastIndex, fastIndexLength * 2);

    if (!mixedBlocks.init(index16Capacity, UCPTRIE_INDEX_3_BLOCK_LENGTH)) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    MixedBlocks longI3Blocks;
    if (hasLongI3Blocks) {
        if (!longI3Blocks.init(index16Capacity, INDEX_3_18BIT_BLOCK_LENGTH)) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return 0;
        }
    }

    // Compact the index-3 table and write an uncompacted version of the index-2 table.
    uint16_t index2[UNICODE_LIMIT >> UCPTRIE_SHIFT_2];  // index2Capacity
    int32_t i2Length = 0;
    i3FirstNull = index3NullOffset;
    int32_t index3Start = fastIndexLength + index1Length;
    int32_t indexLength = index3Start;
    for (int32_t i = iStart; i < iLimit; i += UCPTRIE_INDEX_3_BLOCK_LENGTH) {
        int32_t i3;
        uint8_t f = flags[i];
        if (f == I3_NULL && i3FirstNull < 0) {
            // First index-3 null block. Write & overlap it like a normal block, then remember it.
            f = dataNullOffset <= 0xffff ? I3_16 : I3_18;
            i3FirstNull = 0;
        }
        if (f == I3_NULL) {
            i3 = index3NullOffset;
        } else if (f == I3_BMP) {
            i3 = index[i];
        } else if (f == I3_16) {
            int32_t n = mixedBlocks.findBlock(index16, index, i);
            if (n >= 0) {
                i3 = n;
            } else {
                if (indexLength == index3Start) {
                    // No overlap at the boundary between the index-1 and index-3 tables.
                    n = 0;
                } else {
                    n = getOverlap(index16, indexLength,
                                   index, i, UCPTRIE_INDEX_3_BLOCK_LENGTH);
                }
                i3 = indexLength - n;
                int32_t prevIndexLength = indexLength;
                while (n < UCPTRIE_INDEX_3_BLOCK_LENGTH) {
                    index16[indexLength++] = index[i + n++];
                }
                mixedBlocks.extend(index16, index3Start, prevIndexLength, indexLength);
                if (hasLongI3Blocks) {
                    longI3Blocks.extend(index16, index3Start, prevIndexLength, indexLength);
                }
            }
        } else {
            U_ASSERT(f == I3_18);
            U_ASSERT(hasLongI3Blocks);
            // Encode an index-3 block that contains one or more data indexes exceeding 16 bits.
            int32_t j = i;
            int32_t jLimit = i + UCPTRIE_INDEX_3_BLOCK_LENGTH;
            int32_t k = indexLength;
            do {
                ++k;
                uint32_t v = index[j++];
                uint32_t upperBits = (v & 0x30000) >> 2;
                index16[k++] = v;
                v = index[j++];
                upperBits |= (v & 0x30000) >> 4;
                index16[k++] = v;
                v = index[j++];
                upperBits |= (v & 0x30000) >> 6;
                index16[k++] = v;
                v = index[j++];
                upperBits |= (v & 0x30000) >> 8;
                index16[k++] = v;
                v = index[j++];
                upperBits |= (v & 0x30000) >> 10;
                index16[k++] = v;
                v = index[j++];
                upperBits |= (v & 0x30000) >> 12;
                index16[k++] = v;
                v = index[j++];
                upperBits |= (v & 0x30000) >> 14;
                index16[k++] = v;
                v = index[j++];
                upperBits |= (v & 0x30000) >> 16;
                index16[k++] = v;
                index16[k - 9] = upperBits;
            } while (j < jLimit);
            int32_t n = longI3Blocks.findBlock(index16, index16, indexLength);
            if (n >= 0) {
                i3 = n | 0x8000;
            } else {
                if (indexLength == index3Start) {
                    // No overlap at the boundary between the index-1 and index-3 tables.
                    n = 0;
                } else {
                    n = getOverlap(index16, indexLength,
                                   index16, indexLength, INDEX_3_18BIT_BLOCK_LENGTH);
                }
                i3 = (indexLength - n) | 0x8000;
                int32_t prevIndexLength = indexLength;
                if (n > 0) {
                    int32_t start = indexLength;
                    while (n < INDEX_3_18BIT_BLOCK_LENGTH) {
                        index16[indexLength++] = index16[start + n++];
                    }
                } else {
                    indexLength += INDEX_3_18BIT_BLOCK_LENGTH;
                }
                mixedBlocks.extend(index16, index3Start, prevIndexLength, indexLength);
                if (hasLongI3Blocks) {
                    longI3Blocks.extend(index16, index3Start, prevIndexLength, indexLength);
                }
            }
        }
        if (index3NullOffset < 0 && i3FirstNull >= 0) {
            index3NullOffset = i3;
        }
        // Set the index-2 table entry.
        index2[i2Length++] = i3;
    }
    U_ASSERT(i2Length == index2Capacity);
    U_ASSERT(indexLength <= index3Start + index3Capacity);

    if (index3NullOffset < 0) {
        index3NullOffset = UCPTRIE_NO_INDEX3_NULL_OFFSET;
    }
    if (indexLength >= (UCPTRIE_NO_INDEX3_NULL_OFFSET + UCPTRIE_INDEX_3_BLOCK_LENGTH)) {
        // The index-3 offsets exceed 15 bits, or
        // the last one cannot be distinguished from the no-null-block value.
        errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    // Compact the index-2 table and write the index-1 table.
    static_assert(UCPTRIE_INDEX_2_BLOCK_LENGTH == UCPTRIE_INDEX_3_BLOCK_LENGTH,
                  "must re-init mixedBlocks");
    int32_t blockLength = UCPTRIE_INDEX_2_BLOCK_LENGTH;
    int32_t i1 = fastIndexLength;
    for (int32_t i = 0; i < i2Length; i += blockLength) {
        int32_t n;
        if ((i2Length - i) >= blockLength) {
            // normal block
            U_ASSERT(blockLength == UCPTRIE_INDEX_2_BLOCK_LENGTH);
            n = mixedBlocks.findBlock(index16, index2, i);
        } else {
            // highStart is inside the last index-2 block. Shorten it.
            blockLength = i2Length - i;
            n = findSameBlock(index16, index3Start, indexLength,
                              index2, i, blockLength);
        }
        int32_t i2;
        if (n >= 0) {
            i2 = n;
        } else {
            if (indexLength == index3Start) {
                // No overlap at the boundary between the index-1 and index-3/2 tables.
                n = 0;
            } else {
                n = getOverlap(index16, indexLength, index2, i, blockLength);
            }
            i2 = indexLength - n;
            int32_t prevIndexLength = indexLength;
            while (n < blockLength) {
                index16[indexLength++] = index2[i + n++];
            }
            mixedBlocks.extend(index16, index3Start, prevIndexLength, indexLength);
        }
        // Set the index-1 table entry.
        index16[i1++] = i2;
    }
    U_ASSERT(i1 == index3Start);
    U_ASSERT(indexLength <= index16Capacity);

#ifdef UCPTRIE_DEBUG
    /* we saved some space */
    printf("compacting UCPTrie: count of 16-bit index words %lu->%lu\n",
            (long)iLimit, (long)indexLength);
#endif

    return indexLength;
}

int32_t MutableCodePointTrie::compactTrie(int32_t fastILimit, UErrorCode &errorCode) {
    // Find the real highStart and round it up.
    U_ASSERT((highStart & (UCPTRIE_CP_PER_INDEX_2_ENTRY - 1)) == 0);
    highValue = get(MAX_UNICODE);
    int32_t realHighStart = findHighStart();
    realHighStart = (realHighStart + (UCPTRIE_CP_PER_INDEX_2_ENTRY - 1)) &
        ~(UCPTRIE_CP_PER_INDEX_2_ENTRY - 1);
    if (realHighStart == UNICODE_LIMIT) {
        highValue = initialValue;
    }

#ifdef UCPTRIE_DEBUG
    printf("UCPTrie: highStart U+%06lx  highValue 0x%lx  initialValue 0x%lx\n",
            (long)realHighStart, (long)highValue, (long)initialValue);
#endif

    // We always store indexes and data values for the fast range.
    // Pin highStart to the top of that range while building.
    UChar32 fastLimit = fastILimit << UCPTRIE_SHIFT_3;
    if (realHighStart < fastLimit) {
        for (int32_t i = (realHighStart >> UCPTRIE_SHIFT_3); i < fastILimit; ++i) {
            flags[i] = ALL_SAME;
            index[i] = highValue;
        }
        highStart = fastLimit;
    } else {
        highStart = realHighStart;
    }

    uint32_t asciiData[ASCII_LIMIT];
    for (int32_t i = 0; i < ASCII_LIMIT; ++i) {
        asciiData[i] = get(i);
    }

    // First we look for which data blocks have the same value repeated over the whole block,
    // deduplicate such blocks, find a good null data block (for faster enumeration),
    // and get an upper bound for the necessary data array length.
    AllSameBlocks allSameBlocks;
    int32_t newDataCapacity = compactWholeDataBlocks(fastILimit, allSameBlocks);
    if (newDataCapacity < 0) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    uint32_t *newData = (uint32_t *)uprv_malloc(newDataCapacity * 4);
    if (newData == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    uprv_memcpy(newData, asciiData, sizeof(asciiData));

    int32_t dataNullIndex = allSameBlocks.findMostUsed();

    MixedBlocks mixedBlocks;
    int32_t newDataLength = compactData(fastILimit, newData, newDataCapacity,
                                        dataNullIndex, mixedBlocks, errorCode);
    if (U_FAILURE(errorCode)) { return 0; }
    U_ASSERT(newDataLength <= newDataCapacity);
    uprv_free(data);
    data = newData;
    dataCapacity = newDataCapacity;
    dataLength = newDataLength;
    if (dataLength > (0x3ffff + UCPTRIE_SMALL_DATA_BLOCK_LENGTH)) {
        // The offset of the last data block is too high to be stored in the index table.
        errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    if (dataNullIndex >= 0) {
        dataNullOffset = index[dataNullIndex];
#ifdef UCPTRIE_DEBUG
        if (data[dataNullOffset] != initialValue) {
            printf("UCPTrie initialValue %lx -> more common nullValue %lx\n",
                   (long)initialValue, (long)data[dataNullOffset]);
        }
#endif
        initialValue = data[dataNullOffset];
    } else {
        dataNullOffset = UCPTRIE_NO_DATA_NULL_OFFSET;
    }

    int32_t indexLength = compactIndex(fastILimit, mixedBlocks, errorCode);
    highStart = realHighStart;
    return indexLength;
}

UCPTrie *MutableCodePointTrie::build(UCPTrieType type, UCPTrieValueWidth valueWidth, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) {
        return nullptr;
    }
    if (type < UCPTRIE_TYPE_FAST || UCPTRIE_TYPE_SMALL < type ||
            valueWidth < UCPTRIE_VALUE_BITS_16 || UCPTRIE_VALUE_BITS_8 < valueWidth) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }

    // The mutable trie always stores 32-bit values.
    // When we build a UCPTrie for a smaller value width, we first mask off unused bits
    // before compacting the data.
    switch (valueWidth) {
    case UCPTRIE_VALUE_BITS_32:
        break;
    case UCPTRIE_VALUE_BITS_16:
        maskValues(0xffff);
        break;
    case UCPTRIE_VALUE_BITS_8:
        maskValues(0xff);
        break;
    default:
        break;
    }

    UChar32 fastLimit = type == UCPTRIE_TYPE_FAST ? BMP_LIMIT : UCPTRIE_SMALL_LIMIT;
    int32_t indexLength = compactTrie(fastLimit >> UCPTRIE_SHIFT_3, errorCode);
    if (U_FAILURE(errorCode)) {
        clear();
        return nullptr;
    }

    // Ensure data table alignment: The index length must be even for uint32_t data.
    if (valueWidth == UCPTRIE_VALUE_BITS_32 && (indexLength & 1) != 0) {
        index16[indexLength++] = 0xffee;  // arbitrary value
    }

    // Make the total trie structure length a multiple of 4 bytes by padding the data table,
    // and store special values as the last two data values.
    int32_t length = indexLength * 2;
    if (valueWidth == UCPTRIE_VALUE_BITS_16) {
        if (((indexLength ^ dataLength) & 1) != 0) {
            // padding
            data[dataLength++] = errorValue;
        }
        if (data[dataLength - 1] != errorValue || data[dataLength - 2] != highValue) {
            data[dataLength++] = highValue;
            data[dataLength++] = errorValue;
        }
        length += dataLength * 2;
    } else if (valueWidth == UCPTRIE_VALUE_BITS_32) {
        // 32-bit data words never need padding to a multiple of 4 bytes.
        if (data[dataLength - 1] != errorValue || data[dataLength - 2] != highValue) {
            if (data[dataLength - 1] != highValue) {
                data[dataLength++] = highValue;
            }
            data[dataLength++] = errorValue;
        }
        length += dataLength * 4;
    } else {
        int32_t and3 = (length + dataLength) & 3;
        if (and3 == 0 && data[dataLength - 1] == errorValue && data[dataLength - 2] == highValue) {
            // all set
        } else if(and3 == 3 && data[dataLength - 1] == highValue) {
            data[dataLength++] = errorValue;
        } else {
            while (and3 != 2) {
                data[dataLength++] = highValue;
                and3 = (and3 + 1) & 3;
            }
            data[dataLength++] = highValue;
            data[dataLength++] = errorValue;
        }
        length += dataLength;
    }

    // Calculate the total length of the UCPTrie as a single memory block.
    length += sizeof(UCPTrie);
    U_ASSERT((length & 3) == 0);

    uint8_t *bytes = (uint8_t *)uprv_malloc(length);
    if (bytes == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        clear();
        return nullptr;
    }
    UCPTrie *trie = reinterpret_cast<UCPTrie *>(bytes);
    uprv_memset(trie, 0, sizeof(UCPTrie));
    trie->indexLength = indexLength;
    trie->dataLength = dataLength;

    trie->highStart = highStart;
    // Round up shifted12HighStart to a multiple of 0x1000 for easy testing from UTF-8 lead bytes.
    // Runtime code needs to then test for the real highStart as well.
    trie->shifted12HighStart = (highStart + 0xfff) >> 12;
    trie->type = type;
    trie->valueWidth = valueWidth;

    trie->index3NullOffset = index3NullOffset;
    trie->dataNullOffset = dataNullOffset;
    trie->nullValue = initialValue;

    bytes += sizeof(UCPTrie);

    // Fill the index and data arrays.
    uint16_t *dest16 = (uint16_t *)bytes;
    trie->index = dest16;

    if (highStart <= fastLimit) {
        // Condense only the fast index from the mutable-trie index.
        for (int32_t i = 0, j = 0; j < indexLength; i += SMALL_DATA_BLOCKS_PER_BMP_BLOCK, ++j) {
            *dest16++ = (uint16_t)index[i];  // dest16[j]
        }
    } else {
        uprv_memcpy(dest16, index16, indexLength * 2);
        dest16 += indexLength;
    }
    bytes += indexLength * 2;

    // Write the data array.
    const uint32_t *p = data;
    switch (valueWidth) {
    case UCPTRIE_VALUE_BITS_16:
        // Write 16-bit data values.
        trie->data.ptr16 = dest16;
        for (int32_t i = dataLength; i > 0; --i) {
            *dest16++ = (uint16_t)*p++;
        }
        break;
    case UCPTRIE_VALUE_BITS_32:
        // Write 32-bit data values.
        trie->data.ptr32 = (uint32_t *)bytes;
        uprv_memcpy(bytes, p, (size_t)dataLength * 4);
        break;
    case UCPTRIE_VALUE_BITS_8:
        // Write 8-bit data values.
        trie->data.ptr8 = bytes;
        for (int32_t i = dataLength; i > 0; --i) {
            *bytes++ = (uint8_t)*p++;
        }
        break;
    default:
        // Will not occur, valueWidth checked at the beginning.
        break;
    }

#ifdef UCPTRIE_DEBUG
    trie->name = name;

    ucptrie_printLengths(trie, "");
#endif

    clear();
    return trie;
}

}  // namespace

U_NAMESPACE_END

U_NAMESPACE_USE

U_CAPI UMutableCPTrie * U_EXPORT2
umutablecptrie_open(uint32_t initialValue, uint32_t errorValue, UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) {
        return nullptr;
    }
    LocalPointer<MutableCodePointTrie> trie(
        new MutableCodePointTrie(initialValue, errorValue, *pErrorCode), *pErrorCode);
    if (U_FAILURE(*pErrorCode)) {
        return nullptr;
    }
    return reinterpret_cast<UMutableCPTrie *>(trie.orphan());
}

U_CAPI UMutableCPTrie * U_EXPORT2
umutablecptrie_clone(const UMutableCPTrie *other, UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) {
        return nullptr;
    }
    if (other == nullptr) {
        return nullptr;
    }
    LocalPointer<MutableCodePointTrie> clone(
        new MutableCodePointTrie(*reinterpret_cast<const MutableCodePointTrie *>(other), *pErrorCode), *pErrorCode);
    if (U_FAILURE(*pErrorCode)) {
        return nullptr;
    }
    return reinterpret_cast<UMutableCPTrie *>(clone.orphan());
}

U_CAPI void U_EXPORT2
umutablecptrie_close(UMutableCPTrie *trie) {
    delete reinterpret_cast<MutableCodePointTrie *>(trie);
}

U_CAPI UMutableCPTrie * U_EXPORT2
umutablecptrie_fromUCPMap(const UCPMap *map, UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) {
        return nullptr;
    }
    if (map == nullptr) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    return reinterpret_cast<UMutableCPTrie *>(MutableCodePointTrie::fromUCPMap(map, *pErrorCode));
}

U_CAPI UMutableCPTrie * U_EXPORT2
umutablecptrie_fromUCPTrie(const UCPTrie *trie, UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) {
        return nullptr;
    }
    if (trie == nullptr) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    return reinterpret_cast<UMutableCPTrie *>(MutableCodePointTrie::fromUCPTrie(trie, *pErrorCode));
}

U_CAPI uint32_t U_EXPORT2
umutablecptrie_get(const UMutableCPTrie *trie, UChar32 c) {
    return reinterpret_cast<const MutableCodePointTrie *>(trie)->get(c);
}

namespace {

UChar32 umutablecptrie_internal_getRange(const void *trie, UChar32 start,
                 UCPMapValueFilter *filter, const void *context, uint32_t *pValue) {
    return reinterpret_cast<const MutableCodePointTrie *>(trie)->
        getRange(start, filter, context, pValue);
}

}  // namespace

U_CAPI UChar32 U_EXPORT2
umutablecptrie_getRange(const UMutableCPTrie *trie, UChar32 start,
                        UCPMapRangeOption option, uint32_t surrogateValue,
                        UCPMapValueFilter *filter, const void *context, uint32_t *pValue) {
    return ucptrie_internalGetRange(umutablecptrie_internal_getRange, trie, start,
                                    option, surrogateValue,
                                    filter, context, pValue);
}

U_CAPI void U_EXPORT2
umutablecptrie_set(UMutableCPTrie *trie, UChar32 c, uint32_t value, UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) {
        return;
    }
    reinterpret_cast<MutableCodePointTrie *>(trie)->set(c, value, *pErrorCode);
}

U_CAPI void U_EXPORT2
umutablecptrie_setRange(UMutableCPTrie *trie, UChar32 start, UChar32 end,
                   uint32_t value, UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) {
        return;
    }
    reinterpret_cast<MutableCodePointTrie *>(trie)->setRange(start, end, value, *pErrorCode);
}

/* Compact and internally serialize the trie. */
U_CAPI UCPTrie * U_EXPORT2
umutablecptrie_buildImmutable(UMutableCPTrie *trie, UCPTrieType type, UCPTrieValueWidth valueWidth,
                              UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode)) {
        return nullptr;
    }
    return reinterpret_cast<MutableCodePointTrie *>(trie)->build(type, valueWidth, *pErrorCode);
}

#ifdef UCPTRIE_DEBUG
U_CFUNC void umutablecptrie_setName(UMutableCPTrie *trie, const char *name) {
    reinterpret_cast<MutableCodePointTrie *>(trie)->name = name;
}
#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File umutex.cpp
*
* Modification History:
*
*   Date        Name        Description
*   04/02/97    aliu        Creation.
*   04/07/99    srl         updated
*   05/13/99    stephen     Changed to umutex (from cmutex).
*   11/22/99    aliu        Make non-global mutex autoinitialize [j151]
******************************************************************************
*/








U_NAMESPACE_BEGIN


#if defined(U_USER_MUTEX_CPP)
// Support for including an alternate implementation of mutexes has been withdrawn.
// See issue ICU-20185.
#error U_USER_MUTEX_CPP not supported
#endif


/*************************************************************************************************
 *
 *  ICU Mutex wrappers.
 *
 *************************************************************************************************/

namespace {
std::mutex *initMutex;
std::condition_variable *initCondition;

// The ICU global mutex.
// Used when ICU implementation code passes nullptr for the mutex pointer.
UMutex globalMutex;

std::once_flag initFlag;
std::once_flag *pInitFlag = &initFlag;

}  // Anonymous namespace

U_CDECL_BEGIN
static UBool U_CALLCONV umtx_cleanup() {
    initMutex->~mutex();
    initCondition->~condition_variable();
    UMutex::cleanup();

    // Reset the once_flag, by destructing it and creating a fresh one in its place.
    // Do not use this trick anywhere else in ICU; use umtx_initOnce, not std::call_once().
    pInitFlag->~once_flag();
    pInitFlag = new(&initFlag) std::once_flag();
    return true;
}

static void U_CALLCONV umtx_init() {
    initMutex = STATIC_NEW(std::mutex);
    initCondition = STATIC_NEW(std::condition_variable);
    ucln_common_registerCleanup(UCLN_COMMON_MUTEX, umtx_cleanup);
}
U_CDECL_END


std::mutex *UMutex::getMutex() {
    std::mutex *retPtr = fMutex.load(std::memory_order_acquire);
    if (retPtr == nullptr) {
        std::call_once(*pInitFlag, umtx_init);
        std::lock_guard<std::mutex> guard(*initMutex);
        retPtr = fMutex.load(std::memory_order_acquire);
        if (retPtr == nullptr) {
            fMutex = new(fStorage) std::mutex();
            retPtr = fMutex;
            fListLink = gListHead;
            gListHead = this;
        }
    }
    U_ASSERT(retPtr != nullptr);
    return retPtr;
}

UMutex *UMutex::gListHead = nullptr;

void UMutex::cleanup() {
    UMutex *next = nullptr;
    for (UMutex *m = gListHead; m != nullptr; m = next) {
        (*m->fMutex).~mutex();
        m->fMutex = nullptr;
        next = m->fListLink;
        m->fListLink = nullptr;
    }
    gListHead = nullptr;
}


U_CAPI void  U_EXPORT2
umtx_lock(UMutex *mutex) {
    if (mutex == nullptr) {
        mutex = &globalMutex;
    }
    mutex->lock();
}


U_CAPI void  U_EXPORT2
umtx_unlock(UMutex* mutex)
{
    if (mutex == nullptr) {
        mutex = &globalMutex;
    }
    mutex->unlock();
}


/*************************************************************************************************
 *
 *  UInitOnce Implementation
 *
 *************************************************************************************************/

// This function is called when a test of a UInitOnce::fState reveals that
//   initialization has not completed, that we either need to call the init
//   function on this thread, or wait for some other thread to complete.
//
// The actual call to the init function is made inline by template code
//   that knows the C++ types involved. This function returns true if
//   the caller needs to call the Init function.
//
U_COMMON_API UBool U_EXPORT2
umtx_initImplPreInit(UInitOnce &uio) {
    std::call_once(*pInitFlag, umtx_init);
    std::unique_lock<std::mutex> lock(*initMutex);
    if (umtx_loadAcquire(uio.fState) == 0) {
        umtx_storeRelease(uio.fState, 1);
        return true;      // Caller will next call the init function.
    } else {
        while (umtx_loadAcquire(uio.fState) == 1) {
            // Another thread is currently running the initialization.
            // Wait until it completes.
            initCondition->wait(lock);
        }
        U_ASSERT(uio.fState == 2);
        return false;
    }
}


// This function is called by the thread that ran an initialization function,
// just after completing the function.
//   Some threads may be waiting on the condition, requiring the broadcast wakeup.
//   Some threads may be racing to test the fState variable outside of the mutex,
//   requiring the use of store/release when changing its value.

U_COMMON_API void U_EXPORT2
umtx_initImplPostInit(UInitOnce &uio) {
    {
        std::unique_lock<std::mutex> lock(*initMutex);
        umtx_storeRelease(uio.fState, 2);
    }
    initCondition->notify_all();
}

U_NAMESPACE_END

/*************************************************************************************************
 *
 *  Deprecated functions for setting user mutexes.
 *
 *************************************************************************************************/

U_DEPRECATED void U_EXPORT2
u_setMutexFunctions(const void * /*context */, UMtxInitFn *, UMtxFn *,
                    UMtxFn *,  UMtxFn *, UErrorCode *status) {
    if (U_SUCCESS(*status)) {
        *status = U_UNSUPPORTED_ERROR;
    }
    return;
}



U_DEPRECATED void U_EXPORT2
u_setAtomicIncDecFunctions(const void * /*context */, UMtxAtomicFn *, UMtxAtomicFn *,
                           UErrorCode *status) {
    if (U_SUCCESS(*status)) {
        *status = U_UNSUPPORTED_ERROR;
    }
    return;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  unames.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999oct04
*   created by: Markus W. Scherer
*/
















U_NAMESPACE_BEGIN

/* prototypes ------------------------------------------------------------- */

static const char ICU_UNAMES_DATA_NAME[] = "unames";
static const char ICU_UNAMES_DATA_TYPE[] = "icu";

#define GROUP_SHIFT 5
#define LINES_PER_GROUP (1L<<GROUP_SHIFT)
#define GROUP_MASK (LINES_PER_GROUP-1)

/*
 * This struct was replaced by explicitly accessing equivalent
 * fields from triples of uint16_t.
 * The Group struct was padded to 8 bytes on compilers for early ARM CPUs,
 * which broke the assumption that sizeof(Group)==6 and that the ++ operator
 * would advance by 6 bytes (3 uint16_t).
 *
 * We can't just change the data structure because it's loaded from a data file,
 * and we don't want to make it less compact, so we changed the access code.
 *
 * For details see ICU tickets 6331 and 6008.
typedef struct {
    uint16_t groupMSB,
             offsetHigh, offsetLow; / * avoid padding * /
} Group;
 */
enum {
    GROUP_MSB,
    GROUP_OFFSET_HIGH,
    GROUP_OFFSET_LOW,
    GROUP_LENGTH
};

/*
 * Get the 32-bit group offset.
 * @param group (const uint16_t *) pointer to a Group triple of uint16_t
 * @return group offset (int32_t)
 */
#define GET_GROUP_OFFSET(group) ((int32_t)(group)[GROUP_OFFSET_HIGH]<<16|(group)[GROUP_OFFSET_LOW])

#define NEXT_GROUP(group) ((group)+GROUP_LENGTH)
#define PREV_GROUP(group) ((group)-GROUP_LENGTH)

typedef struct {
    uint32_t start, end;
    uint8_t type, variant;
    uint16_t size;
} AlgorithmicRange;

typedef struct {
    uint32_t tokenStringOffset, groupsOffset, groupStringOffset, algNamesOffset;
} UCharNames;

/*
 * Get the groups table from a UCharNames struct.
 * The groups table consists of one uint16_t groupCount followed by
 * groupCount groups. Each group is a triple of uint16_t, see GROUP_LENGTH
 * and the comment for the old struct Group above.
 *
 * @param names (const UCharNames *) pointer to the UCharNames indexes
 * @return (const uint16_t *) pointer to the groups table
 */
#define GET_GROUPS(names) (const uint16_t *)((const char *)names+names->groupsOffset)

typedef struct {
    const char *otherName;
    UChar32 code;
} FindName;

#define DO_FIND_NAME NULL

static UDataMemory *uCharNamesData=NULL;
static UCharNames *uCharNames=NULL;
static icu::UInitOnce gCharNamesInitOnce = U_INITONCE_INITIALIZER;

/*
 * Maximum length of character names (regular & 1.0).
 */
static int32_t gMaxNameLength=0;

/*
 * Set of chars used in character names (regular & 1.0).
 * Chars are platform-dependent (can be EBCDIC).
 */
static uint32_t gNameSet[8]={ 0 };

#define U_NONCHARACTER_CODE_POINT U_CHAR_CATEGORY_COUNT
#define U_LEAD_SURROGATE U_CHAR_CATEGORY_COUNT + 1
#define U_TRAIL_SURROGATE U_CHAR_CATEGORY_COUNT + 2

#define U_CHAR_EXTENDED_CATEGORY_COUNT (U_CHAR_CATEGORY_COUNT + 3)

static const char * const charCatNames[U_CHAR_EXTENDED_CATEGORY_COUNT] = {
    "unassigned",
    "uppercase letter",
    "lowercase letter",
    "titlecase letter",
    "modifier letter",
    "other letter",
    "non spacing mark",
    "enclosing mark",
    "combining spacing mark",
    "decimal digit number",
    "letter number",
    "other number",
    "space separator",
    "line separator",
    "paragraph separator",
    "control",
    "format",
    "private use area",
    "surrogate",
    "dash punctuation",
    "start punctuation",
    "end punctuation",
    "connector punctuation",
    "other punctuation",
    "math symbol",
    "currency symbol",
    "modifier symbol",
    "other symbol",
    "initial punctuation",
    "final punctuation",
    "noncharacter",
    "lead surrogate",
    "trail surrogate"
};

/* implementation ----------------------------------------------------------- */

static UBool U_CALLCONV unames_cleanup(void)
{
    if(uCharNamesData) {
        udata_close(uCharNamesData);
        uCharNamesData = NULL;
    }
    if(uCharNames) {
        uCharNames = NULL;
    }
    gCharNamesInitOnce.reset();
    gMaxNameLength=0;
    return TRUE;
}

static UBool U_CALLCONV
isAcceptable(void * /*context*/,
             const char * /*type*/, const char * /*name*/,
             const UDataInfo *pInfo) {
    return (UBool)(
        pInfo->size>=20 &&
        pInfo->isBigEndian==U_IS_BIG_ENDIAN &&
        pInfo->charsetFamily==U_CHARSET_FAMILY &&
        pInfo->dataFormat[0]==0x75 &&   /* dataFormat="unam" */
        pInfo->dataFormat[1]==0x6e &&
        pInfo->dataFormat[2]==0x61 &&
        pInfo->dataFormat[3]==0x6d &&
        pInfo->formatVersion[0]==1);
}

static void U_CALLCONV
loadCharNames(UErrorCode &status) {
    U_ASSERT(uCharNamesData == NULL);
    U_ASSERT(uCharNames == NULL);

    uCharNamesData = udata_openChoice(NULL, ICU_UNAMES_DATA_TYPE, ICU_UNAMES_DATA_NAME, isAcceptable, NULL, &status);
    if(U_FAILURE(status)) {
        uCharNamesData = NULL;
    } else {
        uCharNames = (UCharNames *)udata_getMemory(uCharNamesData);
    }
    ucln_common_registerCleanup(UCLN_COMMON_UNAMES, unames_cleanup);
}


static UBool
isDataLoaded(UErrorCode *pErrorCode) {
    umtx_initOnce(gCharNamesInitOnce, &loadCharNames, *pErrorCode);
    return U_SUCCESS(*pErrorCode);
}

#define WRITE_CHAR(buffer, bufferLength, bufferPos, c) UPRV_BLOCK_MACRO_BEGIN { \
    if((bufferLength)>0) { \
        *(buffer)++=c; \
        --(bufferLength); \
    } \
    ++(bufferPos); \
} UPRV_BLOCK_MACRO_END

#define U_ISO_COMMENT U_CHAR_NAME_CHOICE_COUNT

/*
 * Important: expandName() and compareName() are almost the same -
 * apply fixes to both.
 *
 * UnicodeData.txt uses ';' as a field separator, so no
 * field can contain ';' as part of its contents.
 * In unames.dat, it is marked as token[';']==-1 only if the
 * semicolon is used in the data file - which is iff we
 * have Unicode 1.0 names or ISO comments or aliases.
 * So, it will be token[';']==-1 if we store U1.0 names/ISO comments/aliases
 * although we know that it will never be part of a name.
 */
static uint16_t
expandName(UCharNames *names,
           const uint8_t *name, uint16_t nameLength, UCharNameChoice nameChoice,
           char *buffer, uint16_t bufferLength) {
    uint16_t *tokens=(uint16_t *)names+8;
    uint16_t token, tokenCount=*tokens++, bufferPos=0;
    uint8_t *tokenStrings=(uint8_t *)names+names->tokenStringOffset;
    uint8_t c;

    if(nameChoice!=U_UNICODE_CHAR_NAME && nameChoice!=U_EXTENDED_CHAR_NAME) {
        /*
         * skip the modern name if it is not requested _and_
         * if the semicolon byte value is a character, not a token number
         */
        if((uint8_t)';'>=tokenCount || tokens[(uint8_t)';']==(uint16_t)(-1)) {
            int fieldIndex= nameChoice==U_ISO_COMMENT ? 2 : nameChoice;
            do {
                while(nameLength>0) {
                    --nameLength;
                    if(*name++==';') {
                        break;
                    }
                }
            } while(--fieldIndex>0);
        } else {
            /*
             * the semicolon byte value is a token number, therefore
             * only modern names are stored in unames.dat and there is no
             * such requested alternate name here
             */
            nameLength=0;
        }
    }

    /* write each letter directly, and write a token word per token */
    while(nameLength>0) {
        --nameLength;
        c=*name++;

        if(c>=tokenCount) {
            if(c!=';') {
                /* implicit letter */
                WRITE_CHAR(buffer, bufferLength, bufferPos, c);
            } else {
                /* finished */
                break;
            }
        } else {
            token=tokens[c];
            if(token==(uint16_t)(-2)) {
                /* this is a lead byte for a double-byte token */
                token=tokens[c<<8|*name++];
                --nameLength;
            }
            if(token==(uint16_t)(-1)) {
                if(c!=';') {
                    /* explicit letter */
                    WRITE_CHAR(buffer, bufferLength, bufferPos, c);
                } else {
                    /* stop, but skip the semicolon if we are seeking
                       extended names and there was no 2.0 name but there
                       is a 1.0 name. */
                    if(!bufferPos && nameChoice == U_EXTENDED_CHAR_NAME) {
                        if ((uint8_t)';'>=tokenCount || tokens[(uint8_t)';']==(uint16_t)(-1)) {
                            continue;
                        }
                    }
                    /* finished */
                    break;
                }
            } else {
                /* write token word */
                uint8_t *tokenString=tokenStrings+token;
                while((c=*tokenString++)!=0) {
                    WRITE_CHAR(buffer, bufferLength, bufferPos, c);
                }
            }
        }
    }

    /* zero-terminate */
    if(bufferLength>0) {
        *buffer=0;
    }

    return bufferPos;
}

/*
 * compareName() is almost the same as expandName() except that it compares
 * the currently expanded name to an input name.
 * It returns the match/no match result as soon as possible.
 */
static UBool
compareName(UCharNames *names,
            const uint8_t *name, uint16_t nameLength, UCharNameChoice nameChoice,
            const char *otherName) {
    uint16_t *tokens=(uint16_t *)names+8;
    uint16_t token, tokenCount=*tokens++;
    uint8_t *tokenStrings=(uint8_t *)names+names->tokenStringOffset;
    uint8_t c;
    const char *origOtherName = otherName;

    if(nameChoice!=U_UNICODE_CHAR_NAME && nameChoice!=U_EXTENDED_CHAR_NAME) {
        /*
         * skip the modern name if it is not requested _and_
         * if the semicolon byte value is a character, not a token number
         */
        if((uint8_t)';'>=tokenCount || tokens[(uint8_t)';']==(uint16_t)(-1)) {
            int fieldIndex= nameChoice==U_ISO_COMMENT ? 2 : nameChoice;
            do {
                while(nameLength>0) {
                    --nameLength;
                    if(*name++==';') {
                        break;
                    }
                }
            } while(--fieldIndex>0);
        } else {
            /*
             * the semicolon byte value is a token number, therefore
             * only modern names are stored in unames.dat and there is no
             * such requested alternate name here
             */
            nameLength=0;
        }
    }

    /* compare each letter directly, and compare a token word per token */
    while(nameLength>0) {
        --nameLength;
        c=*name++;

        if(c>=tokenCount) {
            if(c!=';') {
                /* implicit letter */
                if((char)c!=*otherName++) {
                    return FALSE;
                }
            } else {
                /* finished */
                break;
            }
        } else {
            token=tokens[c];
            if(token==(uint16_t)(-2)) {
                /* this is a lead byte for a double-byte token */
                token=tokens[c<<8|*name++];
                --nameLength;
            }
            if(token==(uint16_t)(-1)) {
                if(c!=';') {
                    /* explicit letter */
                    if((char)c!=*otherName++) {
                        return FALSE;
                    }
                } else {
                    /* stop, but skip the semicolon if we are seeking
                       extended names and there was no 2.0 name but there
                       is a 1.0 name. */
                    if(otherName == origOtherName && nameChoice == U_EXTENDED_CHAR_NAME) {
                        if ((uint8_t)';'>=tokenCount || tokens[(uint8_t)';']==(uint16_t)(-1)) {
                            continue;
                        }
                    }
                    /* finished */
                    break;
                }
            } else {
                /* write token word */
                uint8_t *tokenString=tokenStrings+token;
                while((c=*tokenString++)!=0) {
                    if((char)c!=*otherName++) {
                        return FALSE;
                    }
                }
            }
        }
    }

    /* complete match? */
    return (UBool)(*otherName==0);
}

static uint8_t getCharCat(UChar32 cp) {
    uint8_t cat;

    if (U_IS_UNICODE_NONCHAR(cp)) {
        return U_NONCHARACTER_CODE_POINT;
    }

    if ((cat = u_charType(cp)) == U_SURROGATE) {
        cat = U_IS_LEAD(cp) ? U_LEAD_SURROGATE : U_TRAIL_SURROGATE;
    }

    return cat;
}

static const char *getCharCatName(UChar32 cp) {
    uint8_t cat = getCharCat(cp);

    /* Return unknown if the table of names above is not up to
       date. */

    if (cat >= UPRV_LENGTHOF(charCatNames)) {
        return "unknown";
    } else {
        return charCatNames[cat];
    }
}

static uint16_t getExtName(uint32_t code, char *buffer, uint16_t bufferLength) {
    const char *catname = getCharCatName(code);
    uint16_t length = 0;

    UChar32 cp;
    int ndigits, i;

    WRITE_CHAR(buffer, bufferLength, length, '<');
    while (catname[length - 1]) {
        WRITE_CHAR(buffer, bufferLength, length, catname[length - 1]);
    }
    WRITE_CHAR(buffer, bufferLength, length, '-');
    for (cp = code, ndigits = 0; cp; ++ndigits, cp >>= 4)
        ;
    if (ndigits < 4)
        ndigits = 4;
    for (cp = code, i = ndigits; (cp || i > 0) && bufferLength; cp >>= 4, bufferLength--) {
        uint8_t v = (uint8_t)(cp & 0xf);
        buffer[--i] = (v < 10 ? '0' + v : 'A' + v - 10);
    }
    buffer += ndigits;
    length += static_cast<uint16_t>(ndigits);
    WRITE_CHAR(buffer, bufferLength, length, '>');

    return length;
}

/*
 * getGroup() does a binary search for the group that contains the
 * Unicode code point "code".
 * The return value is always a valid Group* that may contain "code"
 * or else is the highest group before "code".
 * If the lowest group is after "code", then that one is returned.
 */
static const uint16_t *
getGroup(UCharNames *names, uint32_t code) {
    const uint16_t *groups=GET_GROUPS(names);
    uint16_t groupMSB=(uint16_t)(code>>GROUP_SHIFT),
             start=0,
             limit=*groups++,
             number;

    /* binary search for the group of names that contains the one for code */
    while(start<limit-1) {
        number=(uint16_t)((start+limit)/2);
        if(groupMSB<groups[number*GROUP_LENGTH+GROUP_MSB]) {
            limit=number;
        } else {
            start=number;
        }
    }

    /* return this regardless of whether it is an exact match */
    return groups+start*GROUP_LENGTH;
}

/*
 * expandGroupLengths() reads a block of compressed lengths of 32 strings and
 * expands them into offsets and lengths for each string.
 * Lengths are stored with a variable-width encoding in consecutive nibbles:
 * If a nibble<0xc, then it is the length itself (0=empty string).
 * If a nibble>=0xc, then it forms a length value with the following nibble.
 * Calculation see below.
 * The offsets and lengths arrays must be at least 33 (one more) long because
 * there is no check here at the end if the last nibble is still used.
 */
static const uint8_t *
expandGroupLengths(const uint8_t *s,
                   uint16_t offsets[LINES_PER_GROUP+1], uint16_t lengths[LINES_PER_GROUP+1]) {
    /* read the lengths of the 32 strings in this group and get each string's offset */
    uint16_t i=0, offset=0, length=0;
    uint8_t lengthByte;

    /* all 32 lengths must be read to get the offset of the first group string */
    while(i<LINES_PER_GROUP) {
        lengthByte=*s++;

        /* read even nibble - MSBs of lengthByte */
        if(length>=12) {
            /* double-nibble length spread across two bytes */
            length=(uint16_t)(((length&0x3)<<4|lengthByte>>4)+12);
            lengthByte&=0xf;
        } else if((lengthByte /* &0xf0 */)>=0xc0) {
            /* double-nibble length spread across this one byte */
            length=(uint16_t)((lengthByte&0x3f)+12);
        } else {
            /* single-nibble length in MSBs */
            length=(uint16_t)(lengthByte>>4);
            lengthByte&=0xf;
        }

        *offsets++=offset;
        *lengths++=length;

        offset+=length;
        ++i;

        /* read odd nibble - LSBs of lengthByte */
        if((lengthByte&0xf0)==0) {
            /* this nibble was not consumed for a double-nibble length above */
            length=lengthByte;
            if(length<12) {
                /* single-nibble length in LSBs */
                *offsets++=offset;
                *lengths++=length;

                offset+=length;
                ++i;
            }
        } else {
            length=0;   /* prevent double-nibble detection in the next iteration */
        }
    }

    /* now, s is at the first group string */
    return s;
}

static uint16_t
expandGroupName(UCharNames *names, const uint16_t *group,
                uint16_t lineNumber, UCharNameChoice nameChoice,
                char *buffer, uint16_t bufferLength) {
    uint16_t offsets[LINES_PER_GROUP+2], lengths[LINES_PER_GROUP+2];
    const uint8_t *s=(uint8_t *)names+names->groupStringOffset+GET_GROUP_OFFSET(group);
    s=expandGroupLengths(s, offsets, lengths);
    return expandName(names, s+offsets[lineNumber], lengths[lineNumber], nameChoice,
                      buffer, bufferLength);
}

static uint16_t
getName(UCharNames *names, uint32_t code, UCharNameChoice nameChoice,
        char *buffer, uint16_t bufferLength) {
    const uint16_t *group=getGroup(names, code);
    if((uint16_t)(code>>GROUP_SHIFT)==group[GROUP_MSB]) {
        return expandGroupName(names, group, (uint16_t)(code&GROUP_MASK), nameChoice,
                               buffer, bufferLength);
    } else {
        /* group not found */
        /* zero-terminate */
        if(bufferLength>0) {
            *buffer=0;
        }
        return 0;
    }
}

/*
 * enumGroupNames() enumerates all the names in a 32-group
 * and either calls the enumerator function or finds a given input name.
 */
static UBool
enumGroupNames(UCharNames *names, const uint16_t *group,
               UChar32 start, UChar32 end,
               UEnumCharNamesFn *fn, void *context,
               UCharNameChoice nameChoice) {
    uint16_t offsets[LINES_PER_GROUP+2], lengths[LINES_PER_GROUP+2];
    const uint8_t *s=(uint8_t *)names+names->groupStringOffset+GET_GROUP_OFFSET(group);

    s=expandGroupLengths(s, offsets, lengths);
    if(fn!=DO_FIND_NAME) {
        char buffer[200];
        uint16_t length;

        while(start<=end) {
            length=expandName(names, s+offsets[start&GROUP_MASK], lengths[start&GROUP_MASK], nameChoice, buffer, sizeof(buffer));
            if (!length && nameChoice == U_EXTENDED_CHAR_NAME) {
                buffer[length = getExtName(start, buffer, sizeof(buffer))] = 0;
            }
            /* here, we assume that the buffer is large enough */
            if(length>0) {
                if(!fn(context, start, nameChoice, buffer, length)) {
                    return FALSE;
                }
            }
            ++start;
        }
    } else {
        const char *otherName=((FindName *)context)->otherName;
        while(start<=end) {
            if(compareName(names, s+offsets[start&GROUP_MASK], lengths[start&GROUP_MASK], nameChoice, otherName)) {
                ((FindName *)context)->code=start;
                return FALSE;
            }
            ++start;
        }
    }
    return TRUE;
}

/*
 * enumExtNames enumerate extended names.
 * It only needs to do it if it is called with a real function and not
 * with the dummy DO_FIND_NAME, because u_charFromName() does a check
 * for extended names by itself.
 */
static UBool
enumExtNames(UChar32 start, UChar32 end,
             UEnumCharNamesFn *fn, void *context)
{
    if(fn!=DO_FIND_NAME) {
        char buffer[200];
        uint16_t length;

        while(start<=end) {
            buffer[length = getExtName(start, buffer, sizeof(buffer))] = 0;
            /* here, we assume that the buffer is large enough */
            if(length>0) {
                if(!fn(context, start, U_EXTENDED_CHAR_NAME, buffer, length)) {
                    return FALSE;
                }
            }
            ++start;
        }
    }

    return TRUE;
}

static UBool
enumNames(UCharNames *names,
          UChar32 start, UChar32 limit,
          UEnumCharNamesFn *fn, void *context,
          UCharNameChoice nameChoice) {
    uint16_t startGroupMSB, endGroupMSB, groupCount;
    const uint16_t *group, *groupLimit;

    startGroupMSB=(uint16_t)(start>>GROUP_SHIFT);
    endGroupMSB=(uint16_t)((limit-1)>>GROUP_SHIFT);

    /* find the group that contains start, or the highest before it */
    group=getGroup(names, start);

    if(startGroupMSB<group[GROUP_MSB] && nameChoice==U_EXTENDED_CHAR_NAME) {
        /* enumerate synthetic names between start and the group start */
        UChar32 extLimit=((UChar32)group[GROUP_MSB]<<GROUP_SHIFT);
        if(extLimit>limit) {
            extLimit=limit;
        }
        if(!enumExtNames(start, extLimit-1, fn, context)) {
            return FALSE;
        }
        start=extLimit;
    }

    if(startGroupMSB==endGroupMSB) {
        if(startGroupMSB==group[GROUP_MSB]) {
            /* if start and limit-1 are in the same group, then enumerate only in that one */
            return enumGroupNames(names, group, start, limit-1, fn, context, nameChoice);
        }
    } else {
        const uint16_t *groups=GET_GROUPS(names);
        groupCount=*groups++;
        groupLimit=groups+groupCount*GROUP_LENGTH;

        if(startGroupMSB==group[GROUP_MSB]) {
            /* enumerate characters in the partial start group */
            if((start&GROUP_MASK)!=0) {
                if(!enumGroupNames(names, group,
                                   start, ((UChar32)startGroupMSB<<GROUP_SHIFT)+LINES_PER_GROUP-1,
                                   fn, context, nameChoice)) {
                    return FALSE;
                }
                group=NEXT_GROUP(group); /* continue with the next group */
            }
        } else if(startGroupMSB>group[GROUP_MSB]) {
            /* make sure that we start enumerating with the first group after start */
            const uint16_t *nextGroup=NEXT_GROUP(group);
            if (nextGroup < groupLimit && nextGroup[GROUP_MSB] > startGroupMSB && nameChoice == U_EXTENDED_CHAR_NAME) {
                UChar32 end = nextGroup[GROUP_MSB] << GROUP_SHIFT;
                if (end > limit) {
                    end = limit;
                }
                if (!enumExtNames(start, end - 1, fn, context)) {
                    return FALSE;
                }
            }
            group=nextGroup;
        }

        /* enumerate entire groups between the start- and end-groups */
        while(group<groupLimit && group[GROUP_MSB]<endGroupMSB) {
            const uint16_t *nextGroup;
            start=(UChar32)group[GROUP_MSB]<<GROUP_SHIFT;
            if(!enumGroupNames(names, group, start, start+LINES_PER_GROUP-1, fn, context, nameChoice)) {
                return FALSE;
            }
            nextGroup=NEXT_GROUP(group);
            if (nextGroup < groupLimit && nextGroup[GROUP_MSB] > group[GROUP_MSB] + 1 && nameChoice == U_EXTENDED_CHAR_NAME) {
                UChar32 end = nextGroup[GROUP_MSB] << GROUP_SHIFT;
                if (end > limit) {
                    end = limit;
                }
                if (!enumExtNames((group[GROUP_MSB] + 1) << GROUP_SHIFT, end - 1, fn, context)) {
                    return FALSE;
                }
            }
            group=nextGroup;
        }

        /* enumerate within the end group (group[GROUP_MSB]==endGroupMSB) */
        if(group<groupLimit && group[GROUP_MSB]==endGroupMSB) {
            return enumGroupNames(names, group, (limit-1)&~GROUP_MASK, limit-1, fn, context, nameChoice);
        } else if (nameChoice == U_EXTENDED_CHAR_NAME && group == groupLimit) {
            UChar32 next = (PREV_GROUP(group)[GROUP_MSB] + 1) << GROUP_SHIFT;
            if (next > start) {
                start = next;
            }
        } else {
            return TRUE;
        }
    }

    /* we have not found a group, which means everything is made of
       extended names. */
    if (nameChoice == U_EXTENDED_CHAR_NAME) {
        if (limit > UCHAR_MAX_VALUE + 1) {
            limit = UCHAR_MAX_VALUE + 1;
        }
        return enumExtNames(start, limit - 1, fn, context);
    }

    return TRUE;
}

static uint16_t
writeFactorSuffix(const uint16_t *factors, uint16_t count,
                  const char *s, /* suffix elements */
                  uint32_t code,
                  uint16_t indexes[8], /* output fields from here */
                  const char *elementBases[8], const char *elements[8],
                  char *buffer, uint16_t bufferLength) {
    uint16_t i, factor, bufferPos=0;
    char c;

    /* write elements according to the factors */

    /*
     * the factorized elements are determined by modulo arithmetic
     * with the factors of this algorithm
     *
     * note that for fewer operations, count is decremented here
     */
    --count;
    for(i=count; i>0; --i) {
        factor=factors[i];
        indexes[i]=(uint16_t)(code%factor);
        code/=factor;
    }
    /*
     * we don't need to calculate the last modulus because start<=code<=end
     * guarantees here that code<=factors[0]
     */
    indexes[0]=(uint16_t)code;

    /* write each element */
    for(;;) {
        if(elementBases!=NULL) {
            *elementBases++=s;
        }

        /* skip indexes[i] strings */
        factor=indexes[i];
        while(factor>0) {
            while(*s++!=0) {}
            --factor;
        }
        if(elements!=NULL) {
            *elements++=s;
        }

        /* write element */
        while((c=*s++)!=0) {
            WRITE_CHAR(buffer, bufferLength, bufferPos, c);
        }

        /* we do not need to perform the rest of this loop for i==count - break here */
        if(i>=count) {
            break;
        }

        /* skip the rest of the strings for this factors[i] */
        factor=(uint16_t)(factors[i]-indexes[i]-1);
        while(factor>0) {
            while(*s++!=0) {}
            --factor;
        }

        ++i;
    }

    /* zero-terminate */
    if(bufferLength>0) {
        *buffer=0;
    }

    return bufferPos;
}

/*
 * Important:
 * Parts of findAlgName() are almost the same as some of getAlgName().
 * Fixes must be applied to both.
 */
static uint16_t
getAlgName(AlgorithmicRange *range, uint32_t code, UCharNameChoice nameChoice,
        char *buffer, uint16_t bufferLength) {
    uint16_t bufferPos=0;

    /* Only the normative character name can be algorithmic. */
    if(nameChoice!=U_UNICODE_CHAR_NAME && nameChoice!=U_EXTENDED_CHAR_NAME) {
        /* zero-terminate */
        if(bufferLength>0) {
            *buffer=0;
        }
        return 0;
    }

    switch(range->type) {
    case 0: {
        /* name = prefix hex-digits */
        const char *s=(const char *)(range+1);
        char c;

        uint16_t i, count;

        /* copy prefix */
        while((c=*s++)!=0) {
            WRITE_CHAR(buffer, bufferLength, bufferPos, c);
        }

        /* write hexadecimal code point value */
        count=range->variant;

        /* zero-terminate */
        if(count<bufferLength) {
            buffer[count]=0;
        }

        for(i=count; i>0;) {
            if(--i<bufferLength) {
                c=(char)(code&0xf);
                if(c<10) {
                    c+='0';
                } else {
                    c+='A'-10;
                }
                buffer[i]=c;
            }
            code>>=4;
        }

        bufferPos+=count;
        break;
    }
    case 1: {
        /* name = prefix factorized-elements */
        uint16_t indexes[8];
        const uint16_t *factors=(const uint16_t *)(range+1);
        uint16_t count=range->variant;
        const char *s=(const char *)(factors+count);
        char c;

        /* copy prefix */
        while((c=*s++)!=0) {
            WRITE_CHAR(buffer, bufferLength, bufferPos, c);
        }

        bufferPos+=writeFactorSuffix(factors, count,
                                     s, code-range->start, indexes, NULL, NULL, buffer, bufferLength);
        break;
    }
    default:
        /* undefined type */
        /* zero-terminate */
        if(bufferLength>0) {
            *buffer=0;
        }
        break;
    }

    return bufferPos;
}

/*
 * Important: enumAlgNames() and findAlgName() are almost the same.
 * Any fix must be applied to both.
 */
static UBool
enumAlgNames(AlgorithmicRange *range,
             UChar32 start, UChar32 limit,
             UEnumCharNamesFn *fn, void *context,
             UCharNameChoice nameChoice) {
    char buffer[200];
    uint16_t length;

    if(nameChoice!=U_UNICODE_CHAR_NAME && nameChoice!=U_EXTENDED_CHAR_NAME) {
        return TRUE;
    }

    switch(range->type) {
    case 0: {
        char *s, *end;
        char c;

        /* get the full name of the start character */
        length=getAlgName(range, (uint32_t)start, nameChoice, buffer, sizeof(buffer));
        if(length<=0) {
            return TRUE;
        }

        /* call the enumerator function with this first character */
        if(!fn(context, start, nameChoice, buffer, length)) {
            return FALSE;
        }

        /* go to the end of the name; all these names have the same length */
        end=buffer;
        while(*end!=0) {
            ++end;
        }

        /* enumerate the rest of the names */
        while(++start<limit) {
            /* increment the hexadecimal number on a character-basis */
            s=end;
            for (;;) {
                c=*--s;
                if(('0'<=c && c<'9') || ('A'<=c && c<'F')) {
                    *s=(char)(c+1);
                    break;
                } else if(c=='9') {
                    *s='A';
                    break;
                } else if(c=='F') {
                    *s='0';
                }
            }

            if(!fn(context, start, nameChoice, buffer, length)) {
                return FALSE;
            }
        }
        break;
    }
    case 1: {
        uint16_t indexes[8];
        const char *elementBases[8], *elements[8];
        const uint16_t *factors=(const uint16_t *)(range+1);
        uint16_t count=range->variant;
        const char *s=(const char *)(factors+count);
        char *suffix, *t;
        uint16_t prefixLength, i, idx;

        char c;

        /* name = prefix factorized-elements */

        /* copy prefix */
        suffix=buffer;
        prefixLength=0;
        while((c=*s++)!=0) {
            *suffix++=c;
            ++prefixLength;
        }

        /* append the suffix of the start character */
        length=(uint16_t)(prefixLength+writeFactorSuffix(factors, count,
                                              s, (uint32_t)start-range->start,
                                              indexes, elementBases, elements,
                                              suffix, (uint16_t)(sizeof(buffer)-prefixLength)));

        /* call the enumerator function with this first character */
        if(!fn(context, start, nameChoice, buffer, length)) {
            return FALSE;
        }

        /* enumerate the rest of the names */
        while(++start<limit) {
            /* increment the indexes in lexical order bound by the factors */
            i=count;
            for (;;) {
                idx=(uint16_t)(indexes[--i]+1);
                if(idx<factors[i]) {
                    /* skip one index and its element string */
                    indexes[i]=idx;
                    s=elements[i];
                    while(*s++!=0) {
                    }
                    elements[i]=s;
                    break;
                } else {
                    /* reset this index to 0 and its element string to the first one */
                    indexes[i]=0;
                    elements[i]=elementBases[i];
                }
            }

            /* to make matters a little easier, just append all elements to the suffix */
            t=suffix;
            length=prefixLength;
            for(i=0; i<count; ++i) {
                s=elements[i];
                while((c=*s++)!=0) {
                    *t++=c;
                    ++length;
                }
            }
            /* zero-terminate */
            *t=0;

            if(!fn(context, start, nameChoice, buffer, length)) {
                return FALSE;
            }
        }
        break;
    }
    default:
        /* undefined type */
        break;
    }

    return TRUE;
}

/*
 * findAlgName() is almost the same as enumAlgNames() except that it
 * returns the code point for a name if it fits into the range.
 * It returns 0xffff otherwise.
 */
static UChar32
findAlgName(AlgorithmicRange *range, UCharNameChoice nameChoice, const char *otherName) {
    UChar32 code;

    if(nameChoice!=U_UNICODE_CHAR_NAME && nameChoice!=U_EXTENDED_CHAR_NAME) {
        return 0xffff;
    }

    switch(range->type) {
    case 0: {
        /* name = prefix hex-digits */
        const char *s=(const char *)(range+1);
        char c;

        uint16_t i, count;

        /* compare prefix */
        while((c=*s++)!=0) {
            if((char)c!=*otherName++) {
                return 0xffff;
            }
        }

        /* read hexadecimal code point value */
        count=range->variant;
        code=0;
        for(i=0; i<count; ++i) {
            c=*otherName++;
            if('0'<=c && c<='9') {
                code=(code<<4)|(c-'0');
            } else if('A'<=c && c<='F') {
                code=(code<<4)|(c-'A'+10);
            } else {
                return 0xffff;
            }
        }

        /* does it fit into the range? */
        if(*otherName==0 && range->start<=(uint32_t)code && (uint32_t)code<=range->end) {
            return code;
        }
        break;
    }
    case 1: {
        char buffer[64];
        uint16_t indexes[8];
        const char *elementBases[8], *elements[8];
        const uint16_t *factors=(const uint16_t *)(range+1);
        uint16_t count=range->variant;
        const char *s=(const char *)(factors+count), *t;
        UChar32 start, limit;
        uint16_t i, idx;

        char c;

        /* name = prefix factorized-elements */

        /* compare prefix */
        while((c=*s++)!=0) {
            if((char)c!=*otherName++) {
                return 0xffff;
            }
        }

        start=(UChar32)range->start;
        limit=(UChar32)(range->end+1);

        /* initialize the suffix elements for enumeration; indexes should all be set to 0 */
        writeFactorSuffix(factors, count, s, 0,
                          indexes, elementBases, elements, buffer, sizeof(buffer));

        /* compare the first suffix */
        if(0==uprv_strcmp(otherName, buffer)) {
            return start;
        }

        /* enumerate and compare the rest of the suffixes */
        while(++start<limit) {
            /* increment the indexes in lexical order bound by the factors */
            i=count;
            for (;;) {
                idx=(uint16_t)(indexes[--i]+1);
                if(idx<factors[i]) {
                    /* skip one index and its element string */
                    indexes[i]=idx;
                    s=elements[i];
                    while(*s++!=0) {}
                    elements[i]=s;
                    break;
                } else {
                    /* reset this index to 0 and its element string to the first one */
                    indexes[i]=0;
                    elements[i]=elementBases[i];
                }
            }

            /* to make matters a little easier, just compare all elements of the suffix */
            t=otherName;
            for(i=0; i<count; ++i) {
                s=elements[i];
                while((c=*s++)!=0) {
                    if(c!=*t++) {
                        s=""; /* does not match */
                        i=99;
                    }
                }
            }
            if(i<99 && *t==0) {
                return start;
            }
        }
        break;
    }
    default:
        /* undefined type */
        break;
    }

    return 0xffff;
}

/* sets of name characters, maximum name lengths ---------------------------- */

#define SET_ADD(set, c) ((set)[(uint8_t)c>>5]|=((uint32_t)1<<((uint8_t)c&0x1f)))
#define SET_CONTAINS(set, c) (((set)[(uint8_t)c>>5]&((uint32_t)1<<((uint8_t)c&0x1f)))!=0)

static int32_t
calcStringSetLength(uint32_t set[8], const char *s) {
    int32_t length=0;
    char c;

    while((c=*s++)!=0) {
        SET_ADD(set, c);
        ++length;
    }
    return length;
}

static int32_t
calcAlgNameSetsLengths(int32_t maxNameLength) {
    AlgorithmicRange *range;
    uint32_t *p;
    uint32_t rangeCount;
    int32_t length;

    /* enumerate algorithmic ranges */
    p=(uint32_t *)((uint8_t *)uCharNames+uCharNames->algNamesOffset);
    rangeCount=*p;
    range=(AlgorithmicRange *)(p+1);
    while(rangeCount>0) {
        switch(range->type) {
        case 0:
            /* name = prefix + (range->variant times) hex-digits */
            /* prefix */
            length=calcStringSetLength(gNameSet, (const char *)(range+1))+range->variant;
            if(length>maxNameLength) {
                maxNameLength=length;
            }
            break;
        case 1: {
            /* name = prefix factorized-elements */
            const uint16_t *factors=(const uint16_t *)(range+1);
            const char *s;
            int32_t i, count=range->variant, factor, factorLength, maxFactorLength;

            /* prefix length */
            s=(const char *)(factors+count);
            length=calcStringSetLength(gNameSet, s);
            s+=length+1; /* start of factor suffixes */

            /* get the set and maximum factor suffix length for each factor */
            for(i=0; i<count; ++i) {
                maxFactorLength=0;
                for(factor=factors[i]; factor>0; --factor) {
                    factorLength=calcStringSetLength(gNameSet, s);
                    s+=factorLength+1;
                    if(factorLength>maxFactorLength) {
                        maxFactorLength=factorLength;
                    }
                }
                length+=maxFactorLength;
            }

            if(length>maxNameLength) {
                maxNameLength=length;
            }
            break;
        }
        default:
            /* unknown type */
            break;
        }

        range=(AlgorithmicRange *)((uint8_t *)range+range->size);
        --rangeCount;
    }
    return maxNameLength;
}

static int32_t
calcExtNameSetsLengths(int32_t maxNameLength) {
    int32_t i, length;

    for(i=0; i<UPRV_LENGTHOF(charCatNames); ++i) {
        /*
         * for each category, count the length of the category name
         * plus 9=
         * 2 for <>
         * 1 for -
         * 6 for most hex digits per code point
         */
        length=9+calcStringSetLength(gNameSet, charCatNames[i]);
        if(length>maxNameLength) {
            maxNameLength=length;
        }
    }
    return maxNameLength;
}

static int32_t
calcNameSetLength(const uint16_t *tokens, uint16_t tokenCount, const uint8_t *tokenStrings, int8_t *tokenLengths,
                  uint32_t set[8],
                  const uint8_t **pLine, const uint8_t *lineLimit) {
    const uint8_t *line=*pLine;
    int32_t length=0, tokenLength;
    uint16_t c, token;

    while(line!=lineLimit && (c=*line++)!=(uint8_t)';') {
        if(c>=tokenCount) {
            /* implicit letter */
            SET_ADD(set, c);
            ++length;
        } else {
            token=tokens[c];
            if(token==(uint16_t)(-2)) {
                /* this is a lead byte for a double-byte token */
                c=c<<8|*line++;
                token=tokens[c];
            }
            if(token==(uint16_t)(-1)) {
                /* explicit letter */
                SET_ADD(set, c);
                ++length;
            } else {
                /* count token word */
                if(tokenLengths!=NULL) {
                    /* use cached token length */
                    tokenLength=tokenLengths[c];
                    if(tokenLength==0) {
                        tokenLength=calcStringSetLength(set, (const char *)tokenStrings+token);
                        tokenLengths[c]=(int8_t)tokenLength;
                    }
                } else {
                    tokenLength=calcStringSetLength(set, (const char *)tokenStrings+token);
                }
                length+=tokenLength;
            }
        }
    }

    *pLine=line;
    return length;
}

static void
calcGroupNameSetsLengths(int32_t maxNameLength) {
    uint16_t offsets[LINES_PER_GROUP+2], lengths[LINES_PER_GROUP+2];

    uint16_t *tokens=(uint16_t *)uCharNames+8;
    uint16_t tokenCount=*tokens++;
    uint8_t *tokenStrings=(uint8_t *)uCharNames+uCharNames->tokenStringOffset;

    int8_t *tokenLengths;

    const uint16_t *group;
    const uint8_t *s, *line, *lineLimit;

    int32_t groupCount, lineNumber, length;

    tokenLengths=(int8_t *)uprv_malloc(tokenCount);
    if(tokenLengths!=NULL) {
        uprv_memset(tokenLengths, 0, tokenCount);
    }

    group=GET_GROUPS(uCharNames);
    groupCount=*group++;

    /* enumerate all groups */
    while(groupCount>0) {
        s=(uint8_t *)uCharNames+uCharNames->groupStringOffset+GET_GROUP_OFFSET(group);
        s=expandGroupLengths(s, offsets, lengths);

        /* enumerate all lines in each group */
        for(lineNumber=0; lineNumber<LINES_PER_GROUP; ++lineNumber) {
            line=s+offsets[lineNumber];
            length=lengths[lineNumber];
            if(length==0) {
                continue;
            }

            lineLimit=line+length;

            /* read regular name */
            length=calcNameSetLength(tokens, tokenCount, tokenStrings, tokenLengths, gNameSet, &line, lineLimit);
            if(length>maxNameLength) {
                maxNameLength=length;
            }
            if(line==lineLimit) {
                continue;
            }

            /* read Unicode 1.0 name */
            length=calcNameSetLength(tokens, tokenCount, tokenStrings, tokenLengths, gNameSet, &line, lineLimit);
            if(length>maxNameLength) {
                maxNameLength=length;
            }
            if(line==lineLimit) {
                continue;
            }

            /* read ISO comment */
            /*length=calcNameSetLength(tokens, tokenCount, tokenStrings, tokenLengths, gISOCommentSet, &line, lineLimit);*/
        }

        group=NEXT_GROUP(group);
        --groupCount;
    }

    if(tokenLengths!=NULL) {
        uprv_free(tokenLengths);
    }

    /* set gMax... - name length last for threading */
    gMaxNameLength=maxNameLength;
}

static UBool
calcNameSetsLengths(UErrorCode *pErrorCode) {
    static const char extChars[]="0123456789ABCDEF<>-";
    int32_t i, maxNameLength;

    if(gMaxNameLength!=0) {
        return TRUE;
    }

    if(!isDataLoaded(pErrorCode)) {
        return FALSE;
    }

    /* set hex digits, used in various names, and <>-, used in extended names */
    for(i=0; i<(int32_t)sizeof(extChars)-1; ++i) {
        SET_ADD(gNameSet, extChars[i]);
    }

    /* set sets and lengths from algorithmic names */
    maxNameLength=calcAlgNameSetsLengths(0);

    /* set sets and lengths from extended names */
    maxNameLength=calcExtNameSetsLengths(maxNameLength);

    /* set sets and lengths from group names, set global maximum values */
    calcGroupNameSetsLengths(maxNameLength);

    return TRUE;
}

U_NAMESPACE_END

/* public API --------------------------------------------------------------- */

U_NAMESPACE_USE

U_CAPI int32_t U_EXPORT2
u_charName(UChar32 code, UCharNameChoice nameChoice,
           char *buffer, int32_t bufferLength,
           UErrorCode *pErrorCode) {
     AlgorithmicRange *algRange;
    uint32_t *p;
    uint32_t i;
    int32_t length;

    /* check the argument values */
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    } else if(nameChoice>=U_CHAR_NAME_CHOICE_COUNT ||
              bufferLength<0 || (bufferLength>0 && buffer==NULL)
    ) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if((uint32_t)code>UCHAR_MAX_VALUE || !isDataLoaded(pErrorCode)) {
        return u_terminateChars(buffer, bufferLength, 0, pErrorCode);
    }

    length=0;

    /* try algorithmic names first */
    p=(uint32_t *)((uint8_t *)uCharNames+uCharNames->algNamesOffset);
    i=*p;
    algRange=(AlgorithmicRange *)(p+1);
    while(i>0) {
        if(algRange->start<=(uint32_t)code && (uint32_t)code<=algRange->end) {
            length=getAlgName(algRange, (uint32_t)code, nameChoice, buffer, (uint16_t)bufferLength);
            break;
        }
        algRange=(AlgorithmicRange *)((uint8_t *)algRange+algRange->size);
        --i;
    }

    if(i==0) {
        if (nameChoice == U_EXTENDED_CHAR_NAME) {
            length = getName(uCharNames, (uint32_t )code, U_EXTENDED_CHAR_NAME, buffer, (uint16_t) bufferLength);
            if (!length) {
                /* extended character name */
                length = getExtName((uint32_t) code, buffer, (uint16_t) bufferLength);
            }
        } else {
            /* normal character name */
            length=getName(uCharNames, (uint32_t)code, nameChoice, buffer, (uint16_t)bufferLength);
        }
    }

    return u_terminateChars(buffer, bufferLength, length, pErrorCode);
}

U_CAPI int32_t U_EXPORT2
u_getISOComment(UChar32 /*c*/,
                char *dest, int32_t destCapacity,
                UErrorCode *pErrorCode) {
    /* check the argument values */
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    } else if(destCapacity<0 || (destCapacity>0 && dest==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    return u_terminateChars(dest, destCapacity, 0, pErrorCode);
}

U_CAPI UChar32 U_EXPORT2
u_charFromName(UCharNameChoice nameChoice,
               const char *name,
               UErrorCode *pErrorCode) {
    char upper[120], lower[120];
    FindName findName;
    AlgorithmicRange *algRange;
    uint32_t *p;
    uint32_t i;
    UChar32 cp = 0;
    char c0;
    static constexpr UChar32 error = 0xffff;     /* Undefined, but use this for backwards compatibility. */

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return error;
    }

    if(nameChoice>=U_CHAR_NAME_CHOICE_COUNT || name==NULL || *name==0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return error;
    }

    if(!isDataLoaded(pErrorCode)) {
        return error;
    }

    /* construct the uppercase and lowercase of the name first */
    for(i=0; i<sizeof(upper); ++i) {
        if((c0=*name++)!=0) {
            upper[i]=uprv_toupper(c0);
            lower[i]=uprv_tolower(c0);
        } else {
            upper[i]=lower[i]=0;
            break;
        }
    }
    if(i==sizeof(upper)) {
        /* name too long, there is no such character */
        *pErrorCode = U_ILLEGAL_CHAR_FOUND;
        return error;
    }
    // i==strlen(name)==strlen(lower)==strlen(upper)

    /* try extended names first */
    if (lower[0] == '<') {
        if (nameChoice == U_EXTENDED_CHAR_NAME && lower[--i] == '>') {
            // Parse a string like "<category-HHHH>" where HHHH is a hex code point.
            uint32_t limit = i;
            while (i >= 3 && lower[--i] != '-') {}

            // There should be 1 to 8 hex digits.
            int32_t hexLength = limit - (i + 1);
            if (i >= 2 && lower[i] == '-' && 1 <= hexLength && hexLength <= 8) {
                uint32_t cIdx;

                lower[i] = 0;

                for (++i; i < limit; ++i) {
                    if (lower[i] >= '0' && lower[i] <= '9') {
                        cp = (cp << 4) + lower[i] - '0';
                    } else if (lower[i] >= 'a' && lower[i] <= 'f') {
                        cp = (cp << 4) + lower[i] - 'a' + 10;
                    } else {
                        *pErrorCode = U_ILLEGAL_CHAR_FOUND;
                        return error;
                    }
                    // Prevent signed-integer overflow and out-of-range code points.
                    if (cp > UCHAR_MAX_VALUE) {
                        *pErrorCode = U_ILLEGAL_CHAR_FOUND;
                        return error;
                    }
                }

                /* Now validate the category name.
                   We could use a binary search, or a trie, if
                   we really wanted to. */
                uint8_t cat = getCharCat(cp);
                for (lower[i] = 0, cIdx = 0; cIdx < UPRV_LENGTHOF(charCatNames); ++cIdx) {

                    if (!uprv_strcmp(lower + 1, charCatNames[cIdx])) {
                        if (cat == cIdx) {
                            return cp;
                        }
                        break;
                    }
                }
            }
        }

        *pErrorCode = U_ILLEGAL_CHAR_FOUND;
        return error;
    }

    /* try algorithmic names now */
    p=(uint32_t *)((uint8_t *)uCharNames+uCharNames->algNamesOffset);
    i=*p;
    algRange=(AlgorithmicRange *)(p+1);
    while(i>0) {
        if((cp=findAlgName(algRange, nameChoice, upper))!=0xffff) {
            return cp;
        }
        algRange=(AlgorithmicRange *)((uint8_t *)algRange+algRange->size);
        --i;
    }

    /* normal character name */
    findName.otherName=upper;
    findName.code=error;
    enumNames(uCharNames, 0, UCHAR_MAX_VALUE + 1, DO_FIND_NAME, &findName, nameChoice);
    if (findName.code == error) {
         *pErrorCode = U_ILLEGAL_CHAR_FOUND;
    }
    return findName.code;
}

U_CAPI void U_EXPORT2
u_enumCharNames(UChar32 start, UChar32 limit,
                UEnumCharNamesFn *fn,
                void *context,
                UCharNameChoice nameChoice,
                UErrorCode *pErrorCode) {
    AlgorithmicRange *algRange;
    uint32_t *p;
    uint32_t i;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return;
    }

    if(nameChoice>=U_CHAR_NAME_CHOICE_COUNT || fn==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    if((uint32_t) limit > UCHAR_MAX_VALUE + 1) {
        limit = UCHAR_MAX_VALUE + 1;
    }
    if((uint32_t)start>=(uint32_t)limit) {
        return;
    }

    if(!isDataLoaded(pErrorCode)) {
        return;
    }

    /* interleave the data-driven ones with the algorithmic ones */
    /* iterate over all algorithmic ranges; assume that they are in ascending order */
    p=(uint32_t *)((uint8_t *)uCharNames+uCharNames->algNamesOffset);
    i=*p;
    algRange=(AlgorithmicRange *)(p+1);
    while(i>0) {
        /* enumerate the character names before the current algorithmic range */
        /* here: start<limit */
        if((uint32_t)start<algRange->start) {
            if((uint32_t)limit<=algRange->start) {
                enumNames(uCharNames, start, limit, fn, context, nameChoice);
                return;
            }
            if(!enumNames(uCharNames, start, (UChar32)algRange->start, fn, context, nameChoice)) {
                return;
            }
            start=(UChar32)algRange->start;
        }
        /* enumerate the character names in the current algorithmic range */
        /* here: algRange->start<=start<limit */
        if((uint32_t)start<=algRange->end) {
            if((uint32_t)limit<=(algRange->end+1)) {
                enumAlgNames(algRange, start, limit, fn, context, nameChoice);
                return;
            }
            if(!enumAlgNames(algRange, start, (UChar32)algRange->end+1, fn, context, nameChoice)) {
                return;
            }
            start=(UChar32)algRange->end+1;
        }
        /* continue to the next algorithmic range (here: start<limit) */
        algRange=(AlgorithmicRange *)((uint8_t *)algRange+algRange->size);
        --i;
    }
    /* enumerate the character names after the last algorithmic range */
    enumNames(uCharNames, start, limit, fn, context, nameChoice);
}

U_CAPI int32_t U_EXPORT2
uprv_getMaxCharNameLength() {
    UErrorCode errorCode=U_ZERO_ERROR;
    if(calcNameSetsLengths(&errorCode)) {
        return gMaxNameLength;
    } else {
        return 0;
    }
}

/**
 * Converts the char set cset into a Unicode set uset.
 * @param cset Set of 256 bit flags corresponding to a set of chars.
 * @param uset USet to receive characters. Existing contents are deleted.
 */
static void
charSetToUSet(uint32_t cset[8], const USetAdder *sa) {
    UChar us[256];
    char cs[256];

    int32_t i, length;
    UErrorCode errorCode;

    errorCode=U_ZERO_ERROR;

    if(!calcNameSetsLengths(&errorCode)) {
        return;
    }

    /* build a char string with all chars that are used in character names */
    length=0;
    for(i=0; i<256; ++i) {
        if(SET_CONTAINS(cset, i)) {
            cs[length++]=(char)i;
        }
    }

    /* convert the char string to a UChar string */
    u_charsToUChars(cs, us, length);

    /* add each UChar to the USet */
    for(i=0; i<length; ++i) {
        if(us[i]!=0 || cs[i]==0) { /* non-invariant chars become (UChar)0 */
            sa->add(sa->set, us[i]);
        }
    }
}

/**
 * Fills set with characters that are used in Unicode character names.
 * @param set USet to receive characters.
 */
U_CAPI void U_EXPORT2
uprv_getCharNameCharacters(const USetAdder *sa) {
    charSetToUSet(gNameSet, sa);
}

/* data swapping ------------------------------------------------------------ */

/*
 * The token table contains non-negative entries for token bytes,
 * and -1 for bytes that represent themselves in the data file's charset.
 * -2 entries are used for lead bytes.
 *
 * Direct bytes (-1 entries) must be translated from the input charset family
 * to the output charset family.
 * makeTokenMap() writes a permutation mapping for this.
 * Use it once for single-/lead-byte tokens and once more for all trail byte
 * tokens. (';' is an unused trail byte marked with -1.)
 */
static void
makeTokenMap(const UDataSwapper *ds,
             int16_t tokens[], uint16_t tokenCount,
             uint8_t map[256],
             UErrorCode *pErrorCode) {
    UBool usedOutChar[256];
    uint16_t i, j;
    uint8_t c1, c2;

    if(U_FAILURE(*pErrorCode)) {
        return;
    }

    if(ds->inCharset==ds->outCharset) {
        /* Same charset family: identity permutation */
        for(i=0; i<256; ++i) {
            map[i]=(uint8_t)i;
        }
    } else {
        uprv_memset(map, 0, 256);
        uprv_memset(usedOutChar, 0, 256);

        if(tokenCount>256) {
            tokenCount=256;
        }

        /* set the direct bytes (byte 0 always maps to itself) */
        for(i=1; i<tokenCount; ++i) {
            if(tokens[i]==-1) {
                /* convert the direct byte character */
                c1=(uint8_t)i;
                ds->swapInvChars(ds, &c1, 1, &c2, pErrorCode);
                if(U_FAILURE(*pErrorCode)) {
                    udata_printError(ds, "unames/makeTokenMap() finds variant character 0x%02x used (input charset family %d)\n",
                                     i, ds->inCharset);
                    return;
                }

                /* enter the converted character into the map and mark it used */
                map[c1]=c2;
                usedOutChar[c2]=TRUE;
            }
        }

        /* set the mappings for the rest of the permutation */
        for(i=j=1; i<tokenCount; ++i) {
            /* set mappings that were not set for direct bytes */
            if(map[i]==0) {
                /* set an output byte value that was not used as an output byte above */
                while(usedOutChar[j]) {
                    ++j;
                }
                map[i]=(uint8_t)j++;
            }
        }

        /*
         * leave mappings at tokenCount and above unset if tokenCount<256
         * because they won't be used
         */
    }
}

U_CAPI int32_t U_EXPORT2
uchar_swapNames(const UDataSwapper *ds,
                const void *inData, int32_t length, void *outData,
                UErrorCode *pErrorCode) {
    const UDataInfo *pInfo;
    int32_t headerSize;

    const uint8_t *inBytes;
    uint8_t *outBytes;

    uint32_t tokenStringOffset, groupsOffset, groupStringOffset, algNamesOffset,
             offset, i, count, stringsCount;

    const AlgorithmicRange *inRange;
    AlgorithmicRange *outRange;

    /* udata_swapDataHeader checks the arguments */
    headerSize=udata_swapDataHeader(ds, inData, length, outData, pErrorCode);
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* check data format and format version */
    pInfo=(const UDataInfo *)((const char *)inData+4);
    if(!(
        pInfo->dataFormat[0]==0x75 &&   /* dataFormat="unam" */
        pInfo->dataFormat[1]==0x6e &&
        pInfo->dataFormat[2]==0x61 &&
        pInfo->dataFormat[3]==0x6d &&
        pInfo->formatVersion[0]==1
    )) {
        udata_printError(ds, "uchar_swapNames(): data format %02x.%02x.%02x.%02x (format version %02x) is not recognized as unames.icu\n",
                         pInfo->dataFormat[0], pInfo->dataFormat[1],
                         pInfo->dataFormat[2], pInfo->dataFormat[3],
                         pInfo->formatVersion[0]);
        *pErrorCode=U_UNSUPPORTED_ERROR;
        return 0;
    }

    inBytes=(const uint8_t *)inData+headerSize;
    outBytes=(uint8_t *)outData+headerSize;
    if(length<0) {
        algNamesOffset=ds->readUInt32(((const uint32_t *)inBytes)[3]);
    } else {
        length-=headerSize;
        if( length<20 ||
            (uint32_t)length<(algNamesOffset=ds->readUInt32(((const uint32_t *)inBytes)[3]))
        ) {
            udata_printError(ds, "uchar_swapNames(): too few bytes (%d after header) for unames.icu\n",
                             length);
            *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }
    }

    if(length<0) {
        /* preflighting: iterate through algorithmic ranges */
        offset=algNamesOffset;
        count=ds->readUInt32(*((const uint32_t *)(inBytes+offset)));
        offset+=4;

        for(i=0; i<count; ++i) {
            inRange=(const AlgorithmicRange *)(inBytes+offset);
            offset+=ds->readUInt16(inRange->size);
        }
    } else {
        /* swap data */
        const uint16_t *p;
        uint16_t *q, *temp;

        int16_t tokens[512];
        uint16_t tokenCount;

        uint8_t map[256], trailMap[256];

        /* copy the data for inaccessible bytes */
        if(inBytes!=outBytes) {
            uprv_memcpy(outBytes, inBytes, length);
        }

        /* the initial 4 offsets first */
        tokenStringOffset=ds->readUInt32(((const uint32_t *)inBytes)[0]);
        groupsOffset=ds->readUInt32(((const uint32_t *)inBytes)[1]);
        groupStringOffset=ds->readUInt32(((const uint32_t *)inBytes)[2]);
        ds->swapArray32(ds, inBytes, 16, outBytes, pErrorCode);

        /*
         * now the tokens table
         * it needs to be permutated along with the compressed name strings
         */
        p=(const uint16_t *)(inBytes+16);
        q=(uint16_t *)(outBytes+16);

        /* read and swap the tokenCount */
        tokenCount=ds->readUInt16(*p);
        ds->swapArray16(ds, p, 2, q, pErrorCode);
        ++p;
        ++q;

        /* read the first 512 tokens and make the token maps */
        if(tokenCount<=512) {
            count=tokenCount;
        } else {
            count=512;
        }
        for(i=0; i<count; ++i) {
            tokens[i]=udata_readInt16(ds, p[i]);
        }
        for(; i<512; ++i) {
            tokens[i]=0; /* fill the rest of the tokens array if tokenCount<512 */
        }
        makeTokenMap(ds, tokens, tokenCount, map, pErrorCode);
        makeTokenMap(ds, tokens+256, (uint16_t)(tokenCount>256 ? tokenCount-256 : 0), trailMap, pErrorCode);
        if(U_FAILURE(*pErrorCode)) {
            return 0;
        }

        /*
         * swap and permutate the tokens
         * go through a temporary array to support in-place swapping
         */
        temp=(uint16_t *)uprv_malloc(tokenCount*2);
        if(temp==NULL) {
            udata_printError(ds, "out of memory swapping %u unames.icu tokens\n",
                             tokenCount);
            *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
            return 0;
        }

        /* swap and permutate single-/lead-byte tokens */
        for(i=0; i<tokenCount && i<256; ++i) {
            ds->swapArray16(ds, p+i, 2, temp+map[i], pErrorCode);
        }

        /* swap and permutate trail-byte tokens */
        for(; i<tokenCount; ++i) {
            ds->swapArray16(ds, p+i, 2, temp+(i&0xffffff00)+trailMap[i&0xff], pErrorCode);
        }

        /* copy the result into the output and free the temporary array */
        uprv_memcpy(q, temp, tokenCount*2);
        uprv_free(temp);

        /*
         * swap the token strings but not a possible padding byte after
         * the terminating NUL of the last string
         */
        udata_swapInvStringBlock(ds, inBytes+tokenStringOffset, (int32_t)(groupsOffset-tokenStringOffset),
                                    outBytes+tokenStringOffset, pErrorCode);
        if(U_FAILURE(*pErrorCode)) {
            udata_printError(ds, "uchar_swapNames(token strings) failed\n");
            return 0;
        }

        /* swap the group table */
        count=ds->readUInt16(*((const uint16_t *)(inBytes+groupsOffset)));
        ds->swapArray16(ds, inBytes+groupsOffset, (int32_t)((1+count*3)*2),
                           outBytes+groupsOffset, pErrorCode);

        /*
         * swap the group strings
         * swap the string bytes but not the nibble-encoded string lengths
         */
        if(ds->inCharset!=ds->outCharset) {
            uint16_t offsets[LINES_PER_GROUP+1], lengths[LINES_PER_GROUP+1];

            const uint8_t *inStrings, *nextInStrings;
            uint8_t *outStrings;

            uint8_t c;

            inStrings=inBytes+groupStringOffset;
            outStrings=outBytes+groupStringOffset;

            stringsCount=algNamesOffset-groupStringOffset;

            /* iterate through string groups until only a few padding bytes are left */
            while(stringsCount>32) {
                nextInStrings=expandGroupLengths(inStrings, offsets, lengths);

                /* move past the length bytes */
                stringsCount-=(uint32_t)(nextInStrings-inStrings);
                outStrings+=nextInStrings-inStrings;
                inStrings=nextInStrings;

                count=offsets[31]+lengths[31]; /* total number of string bytes in this group */
                stringsCount-=count;

                /* swap the string bytes using map[] and trailMap[] */
                while(count>0) {
                    c=*inStrings++;
                    *outStrings++=map[c];
                    if(tokens[c]!=-2) {
                        --count;
                    } else {
                        /* token lead byte: swap the trail byte, too */
                        *outStrings++=trailMap[*inStrings++];
                        count-=2;
                    }
                }
            }
        }

        /* swap the algorithmic ranges */
        offset=algNamesOffset;
        count=ds->readUInt32(*((const uint32_t *)(inBytes+offset)));
        ds->swapArray32(ds, inBytes+offset, 4, outBytes+offset, pErrorCode);
        offset+=4;

        for(i=0; i<count; ++i) {
            if(offset>(uint32_t)length) {
                udata_printError(ds, "uchar_swapNames(): too few bytes (%d after header) for unames.icu algorithmic range %u\n",
                                 length, i);
                *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                return 0;
            }

            inRange=(const AlgorithmicRange *)(inBytes+offset);
            outRange=(AlgorithmicRange *)(outBytes+offset);
            offset+=ds->readUInt16(inRange->size);

            ds->swapArray32(ds, inRange, 8, outRange, pErrorCode);
            ds->swapArray16(ds, &inRange->size, 2, &outRange->size, pErrorCode);
            switch(inRange->type) {
            case 0:
                /* swap prefix string */
                ds->swapInvChars(ds, inRange+1, (int32_t)uprv_strlen((const char *)(inRange+1)),
                                    outRange+1, pErrorCode);
                if(U_FAILURE(*pErrorCode)) {
                    udata_printError(ds, "uchar_swapNames(prefix string of algorithmic range %u) failed\n",
                                     i);
                    return 0;
                }
                break;
            case 1:
                {
                    /* swap factors and the prefix and factor strings */
                    uint32_t factorsCount;

                    factorsCount=inRange->variant;
                    p=(const uint16_t *)(inRange+1);
                    q=(uint16_t *)(outRange+1);
                    ds->swapArray16(ds, p, (int32_t)(factorsCount*2), q, pErrorCode);

                    /* swap the strings, up to the last terminating NUL */
                    p+=factorsCount;
                    q+=factorsCount;
                    stringsCount=(uint32_t)((inBytes+offset)-(const uint8_t *)p);
                    while(stringsCount>0 && ((const uint8_t *)p)[stringsCount-1]!=0) {
                        --stringsCount;
                    }
                    ds->swapInvChars(ds, p, (int32_t)stringsCount, q, pErrorCode);
                }
                break;
            default:
                udata_printError(ds, "uchar_swapNames(): unknown type %u of algorithmic range %u\n",
                                 inRange->type, i);
                *pErrorCode=U_UNSUPPORTED_ERROR;
                return 0;
            }
        }
    }

    return headerSize+(int32_t)offset;
}

/*
 * Hey, Emacs, please set the following:
 *
 * Local Variables:
 * indent-tabs-mode: nil
 * End:
 *
 */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2015, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*
* File unifiedcache.cpp
******************************************************************************
*/



#include <algorithm>      // For std::max()
#include <mutex>





static icu::UnifiedCache *unifiedcache_gCache = NULL;
static std::mutex *unifiedcache_gCacheMutex = nullptr;
static std::condition_variable *gInProgressValueAddedCond;
static icu::UInitOnce unifiedcache_gCacheInitOnce = U_INITONCE_INITIALIZER;

static const int32_t MAX_EVICT_ITERATIONS = 10;
static const int32_t DEFAULT_MAX_UNUSED = 1000;
static const int32_t DEFAULT_PERCENTAGE_OF_IN_USE = 100;


U_CDECL_BEGIN
static UBool U_CALLCONV unifiedcache_cleanup() {
    unifiedcache_gCacheInitOnce.reset();
    delete unifiedcache_gCache;
    unifiedcache_gCache = nullptr;
    unifiedcache_gCacheMutex->~mutex();
    unifiedcache_gCacheMutex = nullptr;
    gInProgressValueAddedCond->~condition_variable();
    gInProgressValueAddedCond = nullptr;
    return TRUE;
}
U_CDECL_END


U_NAMESPACE_BEGIN

U_CAPI int32_t U_EXPORT2
ucache_hashKeys(const UHashTok key) {
    const CacheKeyBase *ckey = (const CacheKeyBase *) key.pointer;
    return ckey->hashCode();
}

U_CAPI UBool U_EXPORT2
ucache_compareKeys(const UHashTok key1, const UHashTok key2) {
    const CacheKeyBase *p1 = (const CacheKeyBase *) key1.pointer;
    const CacheKeyBase *p2 = (const CacheKeyBase *) key2.pointer;
    return *p1 == *p2;
}

U_CAPI void U_EXPORT2
ucache_deleteKey(void *obj) {
    CacheKeyBase *p = (CacheKeyBase *) obj;
    delete p;
}

CacheKeyBase::~CacheKeyBase() {
}

static void U_CALLCONV unified_cache_cacheInit(UErrorCode &status) {
    U_ASSERT(unifiedcache_gCache == NULL);
    ucln_common_registerCleanup(
            UCLN_COMMON_UNIFIED_CACHE, unifiedcache_cleanup);

    unifiedcache_gCacheMutex = STATIC_NEW(std::mutex);
    gInProgressValueAddedCond = STATIC_NEW(std::condition_variable);
    unifiedcache_gCache = new UnifiedCache(status);
    if (unifiedcache_gCache == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    if (U_FAILURE(status)) {
        delete unifiedcache_gCache;
        unifiedcache_gCache = NULL;
        return;
    }
}

UnifiedCache *UnifiedCache::getInstance(UErrorCode &status) {
    umtx_initOnce(unifiedcache_gCacheInitOnce, &unified_cache_cacheInit, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    U_ASSERT(unifiedcache_gCache != NULL);
    return unifiedcache_gCache;
}

UnifiedCache::UnifiedCache(UErrorCode &status) :
        fHashtable(NULL),
        fEvictPos(UHASH_FIRST),
        fNumValuesTotal(0),
        fNumValuesInUse(0),
        fMaxUnused(DEFAULT_MAX_UNUSED),
        fMaxPercentageOfInUse(DEFAULT_PERCENTAGE_OF_IN_USE),
        fAutoEvictedCount(0),
        fNoValue(nullptr) {
    if (U_FAILURE(status)) {
        return;
    }
    fNoValue = new SharedObject();
    if (fNoValue == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    fNoValue->softRefCount = 1;  // Add fake references to prevent fNoValue from being deleted
    fNoValue->hardRefCount = 1;  // when other references to it are removed.
    fNoValue->cachePtr = this;

    fHashtable = uhash_open(
            &ucache_hashKeys,
            &ucache_compareKeys,
            NULL,
            &status);
    if (U_FAILURE(status)) {
        return;
    }
    uhash_setKeyDeleter(fHashtable, &ucache_deleteKey);
}

void UnifiedCache::setEvictionPolicy(
        int32_t count, int32_t percentageOfInUseItems, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return;
    }
    if (count < 0 || percentageOfInUseItems < 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    std::lock_guard<std::mutex> lock(*unifiedcache_gCacheMutex);
    fMaxUnused = count;
    fMaxPercentageOfInUse = percentageOfInUseItems;
}

int32_t UnifiedCache::unusedCount() const {
    std::lock_guard<std::mutex> lock(*unifiedcache_gCacheMutex);
    return uhash_count(fHashtable) - fNumValuesInUse;
}

int64_t UnifiedCache::autoEvictedCount() const {
    std::lock_guard<std::mutex> lock(*unifiedcache_gCacheMutex);
    return fAutoEvictedCount;
}

int32_t UnifiedCache::keyCount() const {
    std::lock_guard<std::mutex> lock(*unifiedcache_gCacheMutex);
    return uhash_count(fHashtable);
}

void UnifiedCache::flush() const {
    std::lock_guard<std::mutex> lock(*unifiedcache_gCacheMutex);

    // Use a loop in case cache items that are flushed held hard references to
    // other cache items making those additional cache items eligible for
    // flushing.
    while (_flush(FALSE));
}

void UnifiedCache::handleUnreferencedObject() const {
    std::lock_guard<std::mutex> lock(*unifiedcache_gCacheMutex);
    --fNumValuesInUse;
    _runEvictionSlice();
}

#ifdef UNIFIED_CACHE_DEBUG
#include <stdio.h>

void UnifiedCache::dump() {
    UErrorCode status = U_ZERO_ERROR;
    const UnifiedCache *cache = getInstance(status);
    if (U_FAILURE(status)) {
        fprintf(stderr, "Unified Cache: Error fetching cache.\n");
        return;
    }
    cache->dumpContents();
}

void UnifiedCache::dumpContents() const {
    std::lock_guard<std::mutex> lock(*unifiedcache_gCacheMutex);
    _dumpContents();
}

// Dumps content of cache.
// On entry, unifiedcache_gCacheMutex must be held.
// On exit, cache contents dumped to stderr.
void UnifiedCache::_dumpContents() const {
    int32_t pos = UHASH_FIRST;
    const UHashElement *element = uhash_nextElement(fHashtable, &pos);
    char buffer[256];
    int32_t cnt = 0;
    for (; element != NULL; element = uhash_nextElement(fHashtable, &pos)) {
        const SharedObject *sharedObject =
                (const SharedObject *) element->value.pointer;
        const CacheKeyBase *key =
                (const CacheKeyBase *) element->key.pointer;
        if (sharedObject->hasHardReferences()) {
            ++cnt;
            fprintf(
                    stderr,
                    "Unified Cache: Key '%s', error %d, value %p, total refcount %d, soft refcount %d\n",
                    key->writeDescription(buffer, 256),
                    key->creationStatus,
                    sharedObject == fNoValue ? NULL :sharedObject,
                    sharedObject->getRefCount(),
                    sharedObject->getSoftRefCount());
        }
    }
    fprintf(stderr, "Unified Cache: %d out of a total of %d still have hard references\n", cnt, uhash_count(fHashtable));
}
#endif

UnifiedCache::~UnifiedCache() {
    // Try our best to clean up first.
    flush();
    {
        // Now all that should be left in the cache are entries that refer to
        // each other and entries with hard references from outside the cache.
        // Nothing we can do about these so proceed to wipe out the cache.
        std::lock_guard<std::mutex> lock(*unifiedcache_gCacheMutex);
        _flush(TRUE);
    }
    uhash_close(fHashtable);
    fHashtable = nullptr;
    delete fNoValue;
    fNoValue = nullptr;
}

const UHashElement *
UnifiedCache::_nextElement() const {
    const UHashElement *element = uhash_nextElement(fHashtable, &fEvictPos);
    if (element == NULL) {
        fEvictPos = UHASH_FIRST;
        return uhash_nextElement(fHashtable, &fEvictPos);
    }
    return element;
}

UBool UnifiedCache::_flush(UBool all) const {
    UBool result = FALSE;
    int32_t origSize = uhash_count(fHashtable);
    for (int32_t i = 0; i < origSize; ++i) {
        const UHashElement *element = _nextElement();
        if (element == nullptr) {
            break;
        }
        if (all || _isEvictable(element)) {
            const SharedObject *sharedObject =
                    (const SharedObject *) element->value.pointer;
            U_ASSERT(sharedObject->cachePtr == this);
            uhash_removeElement(fHashtable, element);
            removeSoftRef(sharedObject);    // Deletes the sharedObject when softRefCount goes to zero.
            result = TRUE;
        }
    }
    return result;
}

int32_t UnifiedCache::_computeCountOfItemsToEvict() const {
    int32_t totalItems = uhash_count(fHashtable);
    int32_t evictableItems = totalItems - fNumValuesInUse;

    int32_t unusedLimitByPercentage = fNumValuesInUse * fMaxPercentageOfInUse / 100;
    int32_t unusedLimit = unusedLimitByPercentage > fMaxUnused ? unusedLimitByPercentage : fMaxUnused;
    int32_t countOfItemsToEvict = evictableItems > unusedLimit ? evictableItems - unusedLimit : 0;
    return countOfItemsToEvict;
}

void UnifiedCache::_runEvictionSlice() const {
    int32_t maxItemsToEvict = _computeCountOfItemsToEvict();
    if (maxItemsToEvict <= 0) {
        return;
    }
    for (int32_t i = 0; i < MAX_EVICT_ITERATIONS; ++i) {
        const UHashElement *element = _nextElement();
        if (element == nullptr) {
            break;
        }
        if (_isEvictable(element)) {
            const SharedObject *sharedObject =
                    (const SharedObject *) element->value.pointer;
            uhash_removeElement(fHashtable, element);
            removeSoftRef(sharedObject);   // Deletes sharedObject when SoftRefCount goes to zero.
            ++fAutoEvictedCount;
            if (--maxItemsToEvict == 0) {
                break;
            }
        }
    }
}

void UnifiedCache::_putNew(
        const CacheKeyBase &key,
        const SharedObject *value,
        const UErrorCode creationStatus,
        UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return;
    }
    CacheKeyBase *keyToAdopt = key.clone();
    if (keyToAdopt == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    keyToAdopt->fCreationStatus = creationStatus;
    if (value->softRefCount == 0) {
        _registerMaster(keyToAdopt, value);
    }
    void *oldValue = uhash_put(fHashtable, keyToAdopt, (void *) value, &status);
    U_ASSERT(oldValue == nullptr);
    (void)oldValue;
    if (U_SUCCESS(status)) {
        value->softRefCount++;
    }
}

void UnifiedCache::_putIfAbsentAndGet(
        const CacheKeyBase &key,
        const SharedObject *&value,
        UErrorCode &status) const {
    std::lock_guard<std::mutex> lock(*unifiedcache_gCacheMutex);
    const UHashElement *element = uhash_find(fHashtable, &key);
    if (element != NULL && !_inProgress(element)) {
        _fetch(element, value, status);
        return;
    }
    if (element == NULL) {
        UErrorCode putError = U_ZERO_ERROR;
        // best-effort basis only.
        _putNew(key, value, status, putError);
    } else {
        _put(element, value, status);
    }
    // Run an eviction slice. This will run even if we added a master entry
    // which doesn't increase the unused count, but that is still o.k
    _runEvictionSlice();
}


UBool UnifiedCache::_poll(
        const CacheKeyBase &key,
        const SharedObject *&value,
        UErrorCode &status) const {
    U_ASSERT(value == NULL);
    U_ASSERT(status == U_ZERO_ERROR);
    std::unique_lock<std::mutex> lock(*unifiedcache_gCacheMutex);
    const UHashElement *element = uhash_find(fHashtable, &key);

    // If the hash table contains an inProgress placeholder entry for this key,
    // this means that another thread is currently constructing the value object.
    // Loop, waiting for that construction to complete.
     while (element != NULL && _inProgress(element)) {
         gInProgressValueAddedCond->wait(lock);
         element = uhash_find(fHashtable, &key);
    }

    // If the hash table contains an entry for the key,
    // fetch out the contents and return them.
    if (element != NULL) {
         _fetch(element, value, status);
        return TRUE;
    }

    // The hash table contained nothing for this key.
    // Insert an inProgress place holder value.
    // Our caller will create the final value and update the hash table.
    _putNew(key, fNoValue, U_ZERO_ERROR, status);
    return FALSE;
}

void UnifiedCache::_get(
        const CacheKeyBase &key,
        const SharedObject *&value,
        const void *creationContext,
        UErrorCode &status) const {
    U_ASSERT(value == NULL);
    U_ASSERT(status == U_ZERO_ERROR);
    if (_poll(key, value, status)) {
        if (value == fNoValue) {
            SharedObject::clearPtr(value);
        }
        return;
    }
    if (U_FAILURE(status)) {
        return;
    }
    value = key.createObject(creationContext, status);
    U_ASSERT(value == NULL || value->hasHardReferences());
    U_ASSERT(value != NULL || status != U_ZERO_ERROR);
    if (value == NULL) {
        SharedObject::copyPtr(fNoValue, value);
    }
    _putIfAbsentAndGet(key, value, status);
    if (value == fNoValue) {
        SharedObject::clearPtr(value);
    }
}

void UnifiedCache::_registerMaster(
            const CacheKeyBase *theKey, const SharedObject *value) const {
    theKey->fIsMaster = true;
    value->cachePtr = this;
    ++fNumValuesTotal;
    ++fNumValuesInUse;
}

void UnifiedCache::_put(
        const UHashElement *element,
        const SharedObject *value,
        const UErrorCode status) const {
    U_ASSERT(_inProgress(element));
    const CacheKeyBase *theKey = (const CacheKeyBase *) element->key.pointer;
    const SharedObject *oldValue = (const SharedObject *) element->value.pointer;
    theKey->fCreationStatus = status;
    if (value->softRefCount == 0) {
        _registerMaster(theKey, value);
    }
    value->softRefCount++;
    UHashElement *ptr = const_cast<UHashElement *>(element);
    ptr->value.pointer = (void *) value;
    U_ASSERT(oldValue == fNoValue);
    removeSoftRef(oldValue);

    // Tell waiting threads that we replace in-progress status with
    // an error.
    gInProgressValueAddedCond->notify_all();
}

void UnifiedCache::_fetch(
        const UHashElement *element,
        const SharedObject *&value,
        UErrorCode &status) const {
    const CacheKeyBase *theKey = (const CacheKeyBase *) element->key.pointer;
    status = theKey->fCreationStatus;

    // Since we have the cache lock, calling regular SharedObject add/removeRef
    // could cause us to deadlock on ourselves since they may need to lock
    // the cache mutex.
    removeHardRef(value);
    value = static_cast<const SharedObject *>(element->value.pointer);
    addHardRef(value);
}


UBool UnifiedCache::_inProgress(const UHashElement* element) const {
    UErrorCode status = U_ZERO_ERROR;
    const SharedObject * value = NULL;
    _fetch(element, value, status);
    UBool result = _inProgress(value, status);
    removeHardRef(value);
    return result;
}

UBool UnifiedCache::_inProgress(
        const SharedObject* theValue, UErrorCode creationStatus) const {
    return (theValue == fNoValue && creationStatus == U_ZERO_ERROR);
}

UBool UnifiedCache::_isEvictable(const UHashElement *element) const
{
    const CacheKeyBase *theKey = (const CacheKeyBase *) element->key.pointer;
    const SharedObject *theValue =
            (const SharedObject *) element->value.pointer;

    // Entries that are under construction are never evictable
    if (_inProgress(theValue, theKey->fCreationStatus)) {
        return FALSE;
    }

    // We can evict entries that are either not a master or have just
    // one reference (The one reference being from the cache itself).
    return (!theKey->fIsMaster || (theValue->softRefCount == 1 && theValue->noHardReferences()));
}

void UnifiedCache::removeSoftRef(const SharedObject *value) const {
    U_ASSERT(value->cachePtr == this);
    U_ASSERT(value->softRefCount > 0);
    if (--value->softRefCount == 0) {
        --fNumValuesTotal;
        if (value->noHardReferences()) {
            delete value;
        } else {
            // This path only happens from flush(all). Which only happens from the
            // UnifiedCache destructor.  Nulling out value.cacheptr changes the behavior
            // of value.removeRef(), causing the deletion to be done there.
            value->cachePtr = nullptr;
        }
    }
}

int32_t UnifiedCache::removeHardRef(const SharedObject *value) const {
    int refCount = 0;
    if (value) {
        refCount = umtx_atomic_dec(&value->hardRefCount);
        U_ASSERT(refCount >= 0);
        if (refCount == 0) {
            --fNumValuesInUse;
        }
    }
    return refCount;
}

int32_t UnifiedCache::addHardRef(const SharedObject *value) const {
    int refCount = 0;
    if (value) {
        refCount = umtx_atomic_inc(&value->hardRefCount);
        U_ASSERT(refCount >= 1);
        if (refCount == 1) {
            fNumValuesInUse++;
        }
    }
    return refCount;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2001-2012, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
*   Date        Name        Description
*   07/18/01    aliu        Creation.
**********************************************************************
*/





U_NAMESPACE_BEGIN
UOBJECT_DEFINE_ABSTRACT_RTTI_IMPLEMENTATION(UnicodeFilter)


/* Define this here due to the lack of another file.
   It can't be defined in the header */
UnicodeMatcher::~UnicodeMatcher() {}

UnicodeFilter::~UnicodeFilter() {}

/**
 * UnicodeFunctor API.
 *   Note that UnicodeMatcher is a base class of UnicodeFilter.
 */
UnicodeMatcher* UnicodeFilter::toMatcher() const {
  return const_cast<UnicodeFilter *>(this);
}

void UnicodeFilter::setData(const TransliterationRuleData*) {}

/**
 * Default implementation of UnicodeMatcher::matches() for Unicode
 * filters.  Matches a single code point at offset (either one or
 * two 16-bit code units).
 */
UMatchDegree UnicodeFilter::matches(const Replaceable& text,
                                    int32_t& offset,
                                    int32_t limit,
                                    UBool incremental) {
    UChar32 c;
    if (offset < limit &&
        contains(c = text.char32At(offset))) {
        offset += U16_LENGTH(c);
        return U_MATCH;
    }
    if (offset > limit &&
        contains(c = text.char32At(offset))) {
        // Backup offset by 1, unless the preceding character is a
        // surrogate pair -- then backup by 2 (keep offset pointing at
        // the lead surrogate).
        --offset;
        if (offset >= 0) {
            offset -= U16_LENGTH(text.char32At(offset)) - 1;
        }
        return U_MATCH;
    }
    if (incremental && offset == limit) {
        return U_PARTIAL_MATCH;
    }
    return U_MISMATCH;
}

U_NAMESPACE_END

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2002-2004, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
*/



U_NAMESPACE_BEGIN

UOBJECT_DEFINE_ABSTRACT_RTTI_IMPLEMENTATION(UnicodeFunctor)

UnicodeFunctor::~UnicodeFunctor() {}

UnicodeMatcher* UnicodeFunctor::toMatcher() const {
    return 0;
}

UnicodeReplacer* UnicodeFunctor::toReplacer() const {
    return 0;
}

U_NAMESPACE_END

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1999-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*   Date        Name        Description
*   10/20/99    alan        Creation.
**********************************************************************
*/



















// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2007, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  unisetspan.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2007mar01
*   created by: Markus W. Scherer
*/





U_NAMESPACE_BEGIN

/*
 * Implement span() etc. for a set with strings.
 * Avoid recursion because of its exponential complexity.
 * Instead, try multiple paths at once and track them with an IndexList.
 */
class UnicodeSetStringSpan : public UMemory {
public:
    /*
     * Which span() variant will be used?
     * The object is either built for one variant and used once,
     * or built for all and may be used many times.
     */
    enum {
        FWD             = 0x20,
        BACK            = 0x10,
        UTF16           = 8,
        UTF8            = 4,
        CONTAINED       = 2,
        NOT_CONTAINED   = 1,

        ALL             = 0x3f,

        FWD_UTF16_CONTAINED     = FWD  | UTF16 |     CONTAINED,
        FWD_UTF16_NOT_CONTAINED = FWD  | UTF16 | NOT_CONTAINED,
        FWD_UTF8_CONTAINED      = FWD  | UTF8  |     CONTAINED,
        FWD_UTF8_NOT_CONTAINED  = FWD  | UTF8  | NOT_CONTAINED,
        BACK_UTF16_CONTAINED    = BACK | UTF16 |     CONTAINED,
        BACK_UTF16_NOT_CONTAINED= BACK | UTF16 | NOT_CONTAINED,
        BACK_UTF8_CONTAINED     = BACK | UTF8  |     CONTAINED,
        BACK_UTF8_NOT_CONTAINED = BACK | UTF8  | NOT_CONTAINED
    };

    UnicodeSetStringSpan(const UnicodeSet &set, const UVector &setStrings, uint32_t which);

    // Copy constructor. Assumes which==ALL for a frozen set.
    UnicodeSetStringSpan(const UnicodeSetStringSpan &otherStringSpan, const UVector &newParentSetStrings);

    ~UnicodeSetStringSpan();

    /*
     * Do the strings need to be checked in span() etc.?
     * @return TRUE if strings need to be checked (call span() here),
     *         FALSE if not (use a BMPSet for best performance).
     */
    inline UBool needsStringSpanUTF16();
    inline UBool needsStringSpanUTF8();

    // For fast UnicodeSet::contains(c).
    inline UBool contains(UChar32 c) const;

    int32_t span(const UChar *s, int32_t length, USetSpanCondition spanCondition) const;

    int32_t spanBack(const UChar *s, int32_t length, USetSpanCondition spanCondition) const;

    int32_t spanUTF8(const uint8_t *s, int32_t length, USetSpanCondition spanCondition) const;

    int32_t spanBackUTF8(const uint8_t *s, int32_t length, USetSpanCondition spanCondition) const;

private:
    // Special spanLength byte values.
    enum {
        // The spanLength is >=0xfe.
        LONG_SPAN=0xfe,
        // All code points in the string are contained in the parent set.
        ALL_CP_CONTAINED=0xff
    };

    // Add a starting or ending string character to the spanNotSet
    // so that a character span ends before any string.
    void addToSpanNotSet(UChar32 c);

    int32_t spanNot(const UChar *s, int32_t length) const;
    int32_t spanNotBack(const UChar *s, int32_t length) const;
    int32_t spanNotUTF8(const uint8_t *s, int32_t length) const;
    int32_t spanNotBackUTF8(const uint8_t *s, int32_t length) const;

    // Set for span(). Same as parent but without strings.
    UnicodeSet spanSet;

    // Set for span(not contained).
    // Same as spanSet, plus characters that start or end strings.
    UnicodeSet *pSpanNotSet;

    // The strings of the parent set.
    const UVector &strings;

    // Pointer to the UTF-8 string lengths.
    // Also pointer to further allocated storage for meta data and
    // UTF-8 string contents as necessary.
    int32_t *utf8Lengths;

    // Pointer to the part of the (utf8Lengths) memory block that stores
    // the lengths of span(), spanBack() etc. for each string.
    uint8_t *spanLengths;

    // Pointer to the part of the (utf8Lengths) memory block that stores
    // the UTF-8 versions of the parent set's strings.
    uint8_t *utf8;

    // Number of bytes for all UTF-8 versions of strings together.
    int32_t utf8Length;

    // Maximum lengths of relevant strings.
    int32_t maxLength16;
    int32_t maxLength8;

    // Set up for all variants of span()?
    UBool all;

    // Memory for small numbers and lengths of strings.
    // For example, for 8 strings:
    // 8 UTF-8 lengths, 8*4 bytes span lengths, 8*2 3-byte UTF-8 characters
    // = 112 bytes = int32_t[28].
    int32_t staticLengths[32];
};

UBool UnicodeSetStringSpan::needsStringSpanUTF16() {
    return (UBool)(maxLength16!=0);
}

UBool UnicodeSetStringSpan::needsStringSpanUTF8() {
    return (UBool)(maxLength8!=0);
}

UBool UnicodeSetStringSpan::contains(UChar32 c) const {
    return spanSet.contains(c);
}

U_NAMESPACE_END



// Define UChar constants using hex for EBCDIC compatibility
// Used #define to reduce private static exports and memory access time.
#define uniset_SET_OPEN        ((UChar)0x005B) /*[*/
#define uniset_SET_CLOSE       ((UChar)0x005D) /*]*/
#define uniset_HYPHEN          ((UChar)0x002D) /*-*/
#define uniset_COMPLEMENT      ((UChar)0x005E) /*^*/
#define uniset_COLON           ((UChar)0x003A) /*:*/
#define uniset_BACKSLASH       ((UChar)0x005C) /*\*/
#define uniset_INTERSECTION    ((UChar)0x0026) /*&*/
#define uniset_UPPER_U         ((UChar)0x0055) /*U*/
#define uniset_LOWER_U         ((UChar)0x0075) /*u*/
#define uniset_OPEN_BRACE      ((UChar)123)    /*{*/
#define uniset_CLOSE_BRACE     ((UChar)125)    /*}*/
#define uniset_UPPER_P         ((UChar)0x0050) /*P*/
#define uniset_LOWER_P         ((UChar)0x0070) /*p*/
#define uniset_UPPER_N         ((UChar)78)     /*N*/
#define uniset_EQUALS          ((UChar)0x003D) /*=*/

// HIGH_VALUE > all valid values. 110000 for codepoints
#define UNICODESET_HIGH 0x0110000

// LOW <= all valid values. ZERO for codepoints
#define UNICODESET_LOW 0x000000

/** Max list [0, 1, 2, ..., max code point, HIGH] */
constexpr int32_t uniset_MAX_LENGTH = UNICODESET_HIGH + 1;

U_NAMESPACE_BEGIN

SymbolTable::~SymbolTable() {}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(UnicodeSet)

/**
 * Modify the given UChar32 variable so that it is in range, by
 * pinning values < UNICODESET_LOW to UNICODESET_LOW, and
 * pinning values > UNICODESET_HIGH-1 to UNICODESET_HIGH-1.
 * It modifies its argument in-place and also returns it.
 */
static inline UChar32 pinCodePoint(UChar32& c) {
    if (c < UNICODESET_LOW) {
        c = UNICODESET_LOW;
    } else if (c > (UNICODESET_HIGH-1)) {
        c = (UNICODESET_HIGH-1);
    }
    return c;
}

//----------------------------------------------------------------
// Debugging
//----------------------------------------------------------------

// DO NOT DELETE THIS CODE.  This code is used to debug memory leaks.
// To enable the debugging, define the symbol DEBUG_MEM in the line
// below.  This will result in text being sent to stdout that looks
// like this:
//   DEBUG UnicodeSet: ct 0x00A39B20; 397 [\u0A81-\u0A83\u0A85-
//   DEBUG UnicodeSet: dt 0x00A39B20; 396 [\u0A81-\u0A83\u0A85-
// Each line lists a construction (ct) or destruction (dt) event, the
// object address, the number of outstanding objects after the event,
// and the pattern of the object in question.

// #define DEBUG_MEM

#ifdef DEBUG_MEM
#include <stdio.h>
static int32_t _dbgCount = 0;

static inline void _dbgct(UnicodeSet* set) {
    UnicodeString str;
    set->toPattern(str, TRUE);
    char buf[40];
    str.extract(0, 39, buf, "");
    printf("DEBUG UnicodeSet: ct 0x%08X; %d %s\n", set, ++_dbgCount, buf);
}

static inline void _dbgdt(UnicodeSet* set) {
    UnicodeString str;
    set->toPattern(str, TRUE);
    char buf[40];
    str.extract(0, 39, buf, "");
    printf("DEBUG UnicodeSet: dt 0x%08X; %d %s\n", set, --_dbgCount, buf);
}

#else

#ifndef _dbgct
#define _dbgct(me)
#endif
#ifndef _dbgdt
#define _dbgdt(me)
#endif

#endif

//----------------------------------------------------------------
// UnicodeString in UVector support
//----------------------------------------------------------------

static void U_CALLCONV cloneUnicodeString(UElement *dst, UElement *src) {
    dst->pointer = new UnicodeString(*(UnicodeString*)src->pointer);
}

static int8_t U_CALLCONV compareUnicodeString(UElement t1, UElement t2) {
    const UnicodeString &a = *(const UnicodeString*)t1.pointer;
    const UnicodeString &b = *(const UnicodeString*)t2.pointer;
    return a.compare(b);
}

UBool UnicodeSet::hasStrings() const {
    return strings != nullptr && !strings->isEmpty();
}

int32_t UnicodeSet::stringsSize() const {
    return strings == nullptr ? 0 : strings->size();
}

UBool UnicodeSet::stringsContains(const UnicodeString &s) const {
    return strings != nullptr && strings->contains((void*) &s);
}

//----------------------------------------------------------------
// Constructors &c
//----------------------------------------------------------------

/**
 * Constructs an empty set.
 */
UnicodeSet::UnicodeSet() {
    list[0] = UNICODESET_HIGH;
    _dbgct(this);
}

/**
 * Constructs a set containing the given range. If <code>end >
 * start</code> then an empty set is created.
 *
 * @param start first character, inclusive, of range
 * @param end last character, inclusive, of range
 */
UnicodeSet::UnicodeSet(UChar32 start, UChar32 end) {
    list[0] = UNICODESET_HIGH;
    add(start, end);
    _dbgct(this);
}

/**
 * Constructs a set that is identical to the given UnicodeSet.
 */
UnicodeSet::UnicodeSet(const UnicodeSet& o) : UnicodeFilter(o) {
    *this = o;
    _dbgct(this);
}

// Copy-construct as thawed.
UnicodeSet::UnicodeSet(const UnicodeSet& o, UBool /* asThawed */) : UnicodeFilter(o) {
    if (ensureCapacity(o.len)) {
        // *this = o except for bmpSet and stringSpan
        len = o.len;
        uprv_memcpy(list, o.list, (size_t)len*sizeof(UChar32));
        if (o.hasStrings()) {
            UErrorCode status = U_ZERO_ERROR;
            if (!allocateStrings(status) ||
                    (strings->assign(*o.strings, cloneUnicodeString, status), U_FAILURE(status))) {
                setToBogus();
                return;
            }
        }
        if (o.pat) {
            setPattern(o.pat, o.patLen);
        }
        _dbgct(this);
    }
}

/**
 * Destructs the set.
 */
UnicodeSet::~UnicodeSet() {
    _dbgdt(this); // first!
    if (list != stackList) {
        uprv_free(list);
    }
    delete bmpSet;
    if (buffer != stackList) {
        uprv_free(buffer);
    }
    delete strings;
    delete stringSpan;
    releasePattern();
}

/**
 * Assigns this object to be a copy of another.
 */
UnicodeSet& UnicodeSet::operator=(const UnicodeSet& o) {
    return copyFrom(o, FALSE);
}

UnicodeSet& UnicodeSet::copyFrom(const UnicodeSet& o, UBool asThawed) {
    if (this == &o) {
        return *this;
    }
    if (isFrozen()) {
        return *this;
    }
    if (o.isBogus()) {
        setToBogus();
        return *this;
    }
    if (!ensureCapacity(o.len)) {
        // ensureCapacity will mark the UnicodeSet as Bogus if OOM failure happens.
        return *this;
    }
    len = o.len;
    uprv_memcpy(list, o.list, (size_t)len*sizeof(UChar32));
    if (o.bmpSet != nullptr && !asThawed) {
        bmpSet = new BMPSet(*o.bmpSet, list, len);
        if (bmpSet == NULL) { // Check for memory allocation error.
            setToBogus();
            return *this;
        }
    }
    if (o.hasStrings()) {
        UErrorCode status = U_ZERO_ERROR;
        if ((strings == nullptr && !allocateStrings(status)) ||
                (strings->assign(*o.strings, cloneUnicodeString, status), U_FAILURE(status))) {
            setToBogus();
            return *this;
        }
    } else if (hasStrings()) {
        strings->removeAllElements();
    }
    if (o.stringSpan != nullptr && !asThawed) {
        stringSpan = new UnicodeSetStringSpan(*o.stringSpan, *strings);
        if (stringSpan == NULL) { // Check for memory allocation error.
            setToBogus();
            return *this;
        }
    }
    releasePattern();
    if (o.pat) {
        setPattern(o.pat, o.patLen);
    }
    return *this;
}

/**
 * Returns a copy of this object.  All UnicodeMatcher objects have
 * to support cloning in order to allow classes using
 * UnicodeMatchers, such as Transliterator, to implement cloning.
 */
UnicodeSet* UnicodeSet::clone() const {
    return new UnicodeSet(*this);
}

UnicodeSet *UnicodeSet::cloneAsThawed() const {
    return new UnicodeSet(*this, TRUE);
}

/**
 * Compares the specified object with this set for equality.  Returns
 * <tt>true</tt> if the two sets
 * have the same size, and every member of the specified set is
 * contained in this set (or equivalently, every member of this set is
 * contained in the specified set).
 *
 * @param o set to be compared for equality with this set.
 * @return <tt>true</tt> if the specified set is equal to this set.
 */
UBool UnicodeSet::operator==(const UnicodeSet& o) const {
    if (len != o.len) return FALSE;
    for (int32_t i = 0; i < len; ++i) {
        if (list[i] != o.list[i]) return FALSE;
    }
    if (hasStrings() != o.hasStrings()) { return FALSE; }
    if (hasStrings() && *strings != *o.strings) return FALSE;
    return TRUE;
}

/**
 * Returns the hash code value for this set.
 *
 * @return the hash code value for this set.
 * @see Object#hashCode()
 */
int32_t UnicodeSet::hashCode(void) const {
    uint32_t result = static_cast<uint32_t>(len);
    for (int32_t i = 0; i < len; ++i) {
        result *= 1000003u;
        result += list[i];
    }
    return static_cast<int32_t>(result);
}

//----------------------------------------------------------------
// Public API
//----------------------------------------------------------------

/**
 * Returns the number of elements in this set (its cardinality),
 * Note than the elements of a set may include both individual
 * codepoints and strings.
 *
 * @return the number of elements in this set (its cardinality).
 */
int32_t UnicodeSet::size(void) const {
    int32_t n = 0;
    int32_t count = getRangeCount();
    for (int32_t i = 0; i < count; ++i) {
        n += getRangeEnd(i) - getRangeStart(i) + 1;
    }
    return n + stringsSize();
}

/**
 * Returns <tt>true</tt> if this set contains no elements.
 *
 * @return <tt>true</tt> if this set contains no elements.
 */
UBool UnicodeSet::isEmpty(void) const {
    return len == 1 && !hasStrings();
}

/**
 * Returns true if this set contains the given character.
 * @param c character to be checked for containment
 * @return true if the test condition is met
 */
UBool UnicodeSet::contains(UChar32 c) const {
    // Set i to the index of the start item greater than ch
    // We know we will terminate without length test!
    // LATER: for large sets, add binary search
    //int32_t i = -1;
    //for (;;) {
    //    if (c < list[++i]) break;
    //}
    if (bmpSet != NULL) {
        return bmpSet->contains(c);
    }
    if (stringSpan != NULL) {
        return stringSpan->contains(c);
    }
    if (c >= UNICODESET_HIGH) { // Don't need to check LOW bound
        return FALSE;
    }
    int32_t i = findCodePoint(c);
    return (UBool)(i & 1); // return true if odd
}

/**
 * Returns the smallest value i such that c < list[i].  Caller
 * must ensure that c is a legal value or this method will enter
 * an infinite loop.  This method performs a binary search.
 * @param c a character in the range MIN_VALUE..MAX_VALUE
 * inclusive
 * @return the smallest integer i in the range 0..len-1,
 * inclusive, such that c < list[i]
 */
int32_t UnicodeSet::findCodePoint(UChar32 c) const {
    /* Examples:
                                       findCodePoint(c)
       set              list[]         c=0 1 3 4 7 8
       ===              ==============   ===========
       []               [110000]         0 0 0 0 0 0
       [\u0000-\u0003]  [0, 4, 110000]   1 1 1 2 2 2
       [\u0004-\u0007]  [4, 8, 110000]   0 0 0 1 1 2
       [:Any:]          [0, 110000]      1 1 1 1 1 1
     */

    // Return the smallest i such that c < list[i].  Assume
    // list[len - 1] == HIGH and that c is legal (0..HIGH-1).
    if (c < list[0])
        return 0;
    // High runner test.  c is often after the last range, so an
    // initial check for this condition pays off.
    int32_t lo = 0;
    int32_t hi = len - 1;
    if (lo >= hi || c >= list[hi-1])
        return hi;
    // invariant: c >= list[lo]
    // invariant: c < list[hi]
    for (;;) {
        int32_t i = (lo + hi) >> 1;
        if (i == lo) {
            break; // Found!
        } else if (c < list[i]) {
            hi = i;
        } else {
            lo = i;
        }
    }
    return hi;
}

/**
 * Returns true if this set contains every character
 * of the given range.
 * @param start first character, inclusive, of the range
 * @param end last character, inclusive, of the range
 * @return true if the test condition is met
 */
UBool UnicodeSet::contains(UChar32 start, UChar32 end) const {
    //int32_t i = -1;
    //for (;;) {
    //    if (start < list[++i]) break;
    //}
    int32_t i = findCodePoint(start);
    return ((i & 1) != 0 && end < list[i]);
}

/**
 * Returns <tt>true</tt> if this set contains the given
 * multicharacter string.
 * @param s string to be checked for containment
 * @return <tt>true</tt> if this set contains the specified string
 */
UBool UnicodeSet::contains(const UnicodeString& s) const {
    if (s.length() == 0) return FALSE;
    int32_t cp = getSingleCP(s);
    if (cp < 0) {
        return stringsContains(s);
    } else {
        return contains((UChar32) cp);
    }
}

/**
 * Returns true if this set contains all the characters and strings
 * of the given set.
 * @param c set to be checked for containment
 * @return true if the test condition is met
 */
UBool UnicodeSet::containsAll(const UnicodeSet& c) const {
    // The specified set is a subset if all of its pairs are contained in
    // this set.  It's possible to code this more efficiently in terms of
    // direct manipulation of the inversion lists if the need arises.
    int32_t n = c.getRangeCount();
    for (int i=0; i<n; ++i) {
        if (!contains(c.getRangeStart(i), c.getRangeEnd(i))) {
            return FALSE;
        }
    }
    return !c.hasStrings() || (strings != nullptr && strings->containsAll(*c.strings));
}

/**
 * Returns true if this set contains all the characters
 * of the given string.
 * @param s string containing characters to be checked for containment
 * @return true if the test condition is met
 */
UBool UnicodeSet::containsAll(const UnicodeString& s) const {
    return (UBool)(span(s.getBuffer(), s.length(), USET_SPAN_CONTAINED) ==
                   s.length());
}

/**
 * Returns true if this set contains none of the characters
 * of the given range.
 * @param start first character, inclusive, of the range
 * @param end last character, inclusive, of the range
 * @return true if the test condition is met
 */
UBool UnicodeSet::containsNone(UChar32 start, UChar32 end) const {
    //int32_t i = -1;
    //for (;;) {
    //    if (start < list[++i]) break;
    //}
    int32_t i = findCodePoint(start);
    return ((i & 1) == 0 && end < list[i]);
}

/**
 * Returns true if this set contains none of the characters and strings
 * of the given set.
 * @param c set to be checked for containment
 * @return true if the test condition is met
 */
UBool UnicodeSet::containsNone(const UnicodeSet& c) const {
    // The specified set is a subset if all of its pairs are contained in
    // this set.  It's possible to code this more efficiently in terms of
    // direct manipulation of the inversion lists if the need arises.
    int32_t n = c.getRangeCount();
    for (int32_t i=0; i<n; ++i) {
        if (!containsNone(c.getRangeStart(i), c.getRangeEnd(i))) {
            return FALSE;
        }
    }
    return strings == nullptr || !c.hasStrings() || strings->containsNone(*c.strings);
}

/**
 * Returns true if this set contains none of the characters
 * of the given string.
 * @param s string containing characters to be checked for containment
 * @return true if the test condition is met
 */
UBool UnicodeSet::containsNone(const UnicodeString& s) const {
    return (UBool)(span(s.getBuffer(), s.length(), USET_SPAN_NOT_CONTAINED) ==
                   s.length());
}

/**
 * Returns <tt>true</tt> if this set contains any character whose low byte
 * is the given value.  This is used by <tt>RuleBasedTransliterator</tt> for
 * indexing.
 */
UBool UnicodeSet::matchesIndexValue(uint8_t v) const {
    /* The index value v, in the range [0,255], is contained in this set if
     * it is contained in any pair of this set.  Pairs either have the high
     * bytes equal, or unequal.  If the high bytes are equal, then we have
     * aaxx..aayy, where aa is the high byte.  Then v is contained if xx <=
     * v <= yy.  If the high bytes are unequal we have aaxx..bbyy, bb>aa.
     * Then v is contained if xx <= v || v <= yy.  (This is identical to the
     * time zone month containment logic.)
     */
    int32_t i;
    int32_t rangeCount=getRangeCount();
    for (i=0; i<rangeCount; ++i) {
        UChar32 low = getRangeStart(i);
        UChar32 high = getRangeEnd(i);
        if ((low & ~0xFF) == (high & ~0xFF)) {
            if ((low & 0xFF) <= v && v <= (high & 0xFF)) {
                return TRUE;
            }
        } else if ((low & 0xFF) <= v || v <= (high & 0xFF)) {
            return TRUE;
        }
    }
    if (hasStrings()) {
        for (i=0; i<strings->size(); ++i) {
            const UnicodeString& s = *(const UnicodeString*)strings->elementAt(i);
            //if (s.length() == 0) {
            //    // Empty strings match everything
            //    return TRUE;
            //}
            // assert(s.length() != 0); // We enforce this elsewhere
            UChar32 c = s.char32At(0);
            if ((c & 0xFF) == v) {
                return TRUE;
            }
        }
    }
    return FALSE;
}

/**
 * Implementation of UnicodeMatcher::matches().  Always matches the
 * longest possible multichar string.
 */
UMatchDegree UnicodeSet::matches(const Replaceable& text,
                                 int32_t& offset,
                                 int32_t limit,
                                 UBool incremental) {
    if (offset == limit) {
        // Strings, if any, have length != 0, so we don't worry
        // about them here.  If we ever allow zero-length strings
        // we much check for them here.
        if (contains(U_ETHER)) {
            return incremental ? U_PARTIAL_MATCH : U_MATCH;
        } else {
            return U_MISMATCH;
        }
    } else {
        if (hasStrings()) { // try strings first

            // might separate forward and backward loops later
            // for now they are combined

            // TODO Improve efficiency of this, at least in the forward
            // direction, if not in both.  In the forward direction we
            // can assume the strings are sorted.

            int32_t i;
            UBool forward = offset < limit;

            // firstChar is the leftmost char to match in the
            // forward direction or the rightmost char to match in
            // the reverse direction.
            UChar firstChar = text.charAt(offset);

            // If there are multiple strings that can match we
            // return the longest match.
            int32_t highWaterLength = 0;

            for (i=0; i<strings->size(); ++i) {
                const UnicodeString& trial = *(const UnicodeString*)strings->elementAt(i);

                //if (trial.length() == 0) {
                //    return U_MATCH; // null-string always matches
                //}
                // assert(trial.length() != 0); // We ensure this elsewhere

                UChar c = trial.charAt(forward ? 0 : trial.length() - 1);

                // Strings are sorted, so we can optimize in the
                // forward direction.
                if (forward && c > firstChar) break;
                if (c != firstChar) continue;

                int32_t matchLen = matchRest(text, offset, limit, trial);

                if (incremental) {
                    int32_t maxLen = forward ? limit-offset : offset-limit;
                    if (matchLen == maxLen) {
                        // We have successfully matched but only up to limit.
                        return U_PARTIAL_MATCH;
                    }
                }

                if (matchLen == trial.length()) {
                    // We have successfully matched the whole string.
                    if (matchLen > highWaterLength) {
                        highWaterLength = matchLen;
                    }
                    // In the forward direction we know strings
                    // are sorted so we can bail early.
                    if (forward && matchLen < highWaterLength) {
                        break;
                    }
                    continue;
                }
            }

            // We've checked all strings without a partial match.
            // If we have full matches, return the longest one.
            if (highWaterLength != 0) {
                offset += forward ? highWaterLength : -highWaterLength;
                return U_MATCH;
            }
        }
        return UnicodeFilter::matches(text, offset, limit, incremental);
    }
}

/**
 * Returns the longest match for s in text at the given position.
 * If limit > start then match forward from start+1 to limit
 * matching all characters except s.charAt(0).  If limit < start,
 * go backward starting from start-1 matching all characters
 * except s.charAt(s.length()-1).  This method assumes that the
 * first character, text.charAt(start), matches s, so it does not
 * check it.
 * @param text the text to match
 * @param start the first character to match.  In the forward
 * direction, text.charAt(start) is matched against s.charAt(0).
 * In the reverse direction, it is matched against
 * s.charAt(s.length()-1).
 * @param limit the limit offset for matching, either last+1 in
 * the forward direction, or last-1 in the reverse direction,
 * where last is the index of the last character to match.
 * @return If part of s matches up to the limit, return |limit -
 * start|.  If all of s matches before reaching the limit, return
 * s.length().  If there is a mismatch between s and text, return
 * 0
 */
int32_t UnicodeSet::matchRest(const Replaceable& text,
                              int32_t start, int32_t limit,
                              const UnicodeString& s) {
    int32_t i;
    int32_t maxLen;
    int32_t slen = s.length();
    if (start < limit) {
        maxLen = limit - start;
        if (maxLen > slen) maxLen = slen;
        for (i = 1; i < maxLen; ++i) {
            if (text.charAt(start + i) != s.charAt(i)) return 0;
        }
    } else {
        maxLen = start - limit;
        if (maxLen > slen) maxLen = slen;
        --slen; // <=> slen = s.length() - 1;
        for (i = 1; i < maxLen; ++i) {
            if (text.charAt(start - i) != s.charAt(slen - i)) return 0;
        }
    }
    return maxLen;
}

/**
 * Implement of UnicodeMatcher
 */
void UnicodeSet::addMatchSetTo(UnicodeSet& toUnionTo) const {
    toUnionTo.addAll(*this);
}

/**
 * Returns the index of the given character within this set, where
 * the set is ordered by ascending code point.  If the character
 * is not in this set, return -1.  The inverse of this method is
 * <code>charAt()</code>.
 * @return an index from 0..size()-1, or -1
 */
int32_t UnicodeSet::indexOf(UChar32 c) const {
    if (c < MIN_VALUE || c > MAX_VALUE) {
        return -1;
    }
    int32_t i = 0;
    int32_t n = 0;
    for (;;) {
        UChar32 start = list[i++];
        if (c < start) {
            return -1;
        }
        UChar32 limit = list[i++];
        if (c < limit) {
            return n + c - start;
        }
        n += limit - start;
    }
}

/**
 * Returns the character at the given index within this set, where
 * the set is ordered by ascending code point.  If the index is
 * out of range, return (UChar32)-1.  The inverse of this method is
 * <code>indexOf()</code>.
 * @param index an index from 0..size()-1
 * @return the character at the given index, or (UChar32)-1.
 */
UChar32 UnicodeSet::charAt(int32_t index) const {
    if (index >= 0) {
        // len2 is the largest even integer <= len, that is, it is len
        // for even values and len-1 for odd values.  With odd values
        // the last entry is UNICODESET_HIGH.
        int32_t len2 = len & ~1;
        for (int32_t i=0; i < len2;) {
            UChar32 start = list[i++];
            int32_t count = list[i++] - start;
            if (index < count) {
                return (UChar32)(start + index);
            }
            index -= count;
        }
    }
    return (UChar32)-1;
}

/**
 * Make this object represent the range <code>start - end</code>.
 * If <code>end > start</code> then this object is set to an
 * an empty range.
 *
 * @param start first character in the set, inclusive
 * @rparam end last character in the set, inclusive
 */
UnicodeSet& UnicodeSet::set(UChar32 start, UChar32 end) {
    clear();
    complement(start, end);
    return *this;
}

/**
 * Adds the specified range to this set if it is not already
 * present.  If this set already contains the specified range,
 * the call leaves this set unchanged.  If <code>end > start</code>
 * then an empty range is added, leaving the set unchanged.
 *
 * @param start first character, inclusive, of range to be added
 * to this set.
 * @param end last character, inclusive, of range to be added
 * to this set.
 */
UnicodeSet& UnicodeSet::add(UChar32 start, UChar32 end) {
    if (pinCodePoint(start) < pinCodePoint(end)) {
        UChar32 limit = end + 1;
        // Fast path for adding a new range after the last one.
        // Odd list length: [..., lastStart, lastLimit, HIGH]
        if ((len & 1) != 0) {
            // If the list is empty, set lastLimit low enough to not be adjacent to 0.
            UChar32 lastLimit = len == 1 ? -2 : list[len - 2];
            if (lastLimit <= start && !isFrozen() && !isBogus()) {
                if (lastLimit == start) {
                    // Extend the last range.
                    list[len - 2] = limit;
                    if (limit == UNICODESET_HIGH) {
                        --len;
                    }
                } else {
                    list[len - 1] = start;
                    if (limit < UNICODESET_HIGH) {
                        if (ensureCapacity(len + 2)) {
                            list[len++] = limit;
                            list[len++] = UNICODESET_HIGH;
                        }
                    } else {  // limit == UNICODESET_HIGH
                        if (ensureCapacity(len + 1)) {
                            list[len++] = UNICODESET_HIGH;
                        }
                    }
                }
                releasePattern();
                return *this;
            }
        }
        // This is slow. Could be much faster using findCodePoint(start)
        // and modifying the list, dealing with adjacent & overlapping ranges.
        UChar32 range[3] = { start, limit, UNICODESET_HIGH };
        add(range, 2, 0);
    } else if (start == end) {
        add(start);
    }
    return *this;
}

// #define DEBUG_US_ADD

#ifdef DEBUG_US_ADD
#include <stdio.h>
void dump(UChar32 c) {
    if (c <= 0xFF) {
        printf("%c", (char)c);
    } else {
        printf("U+%04X", c);
    }
}
void dump(const UChar32* list, int32_t len) {
    printf("[");
    for (int32_t i=0; i<len; ++i) {
        if (i != 0) printf(", ");
        dump(list[i]);
    }
    printf("]");
}
#endif

/**
 * Adds the specified character to this set if it is not already
 * present.  If this set already contains the specified character,
 * the call leaves this set unchanged.
 */
UnicodeSet& UnicodeSet::add(UChar32 c) {
    // find smallest i such that c < list[i]
    // if odd, then it is IN the set
    // if even, then it is OUT of the set
    int32_t i = findCodePoint(pinCodePoint(c));

    // already in set?
    if ((i & 1) != 0  || isFrozen() || isBogus()) return *this;

    // HIGH is 0x110000
    // assert(list[len-1] == HIGH);

    // empty = [HIGH]
    // [start_0, limit_0, start_1, limit_1, HIGH]

    // [..., start_k-1, limit_k-1, start_k, limit_k, ..., HIGH]
    //                             ^
    //                             list[i]

    // i == 0 means c is before the first range

#ifdef DEBUG_US_ADD
    printf("Add of ");
    dump(c);
    printf(" found at %d", i);
    printf(": ");
    dump(list, len);
    printf(" => ");
#endif

    if (c == list[i]-1) {
        // c is before start of next range
        list[i] = c;
        // if we touched the HIGH mark, then add a new one
        if (c == (UNICODESET_HIGH - 1)) {
            if (!ensureCapacity(len+1)) {
                // ensureCapacity will mark the object as Bogus if OOM failure happens.
                return *this;
            }
            list[len++] = UNICODESET_HIGH;
        }
        if (i > 0 && c == list[i-1]) {
            // collapse adjacent ranges

            // [..., start_k-1, c, c, limit_k, ..., HIGH]
            //                     ^
            //                     list[i]

            //for (int32_t k=i-1; k<len-2; ++k) {
            //    list[k] = list[k+2];
            //}
            UChar32* dst = list + i - 1;
            UChar32* src = dst + 2;
            UChar32* srclimit = list + len;
            while (src < srclimit) *(dst++) = *(src++);

            len -= 2;
        }
    }

    else if (i > 0 && c == list[i-1]) {
        // c is after end of prior range
        list[i-1]++;
        // no need to check for collapse here
    }

    else {
        // At this point we know the new char is not adjacent to
        // any existing ranges, and it is not 10FFFF.


        // [..., start_k-1, limit_k-1, start_k, limit_k, ..., HIGH]
        //                             ^
        //                             list[i]

        // [..., start_k-1, limit_k-1, c, c+1, start_k, limit_k, ..., HIGH]
        //                             ^
        //                             list[i]

        if (!ensureCapacity(len+2)) {
            // ensureCapacity will mark the object as Bogus if OOM failure happens.
            return *this;
        }

        UChar32 *p = list + i;
        uprv_memmove(p + 2, p, (len - i) * sizeof(*p));
        list[i] = c;
        list[i+1] = c+1;
        len += 2;
    }

#ifdef DEBUG_US_ADD
    dump(list, len);
    printf("\n");

    for (i=1; i<len; ++i) {
        if (list[i] <= list[i-1]) {
            // Corrupt array!
            printf("ERROR: list has been corrupted\n");
            exit(1);
        }
    }
#endif

    releasePattern();
    return *this;
}

/**
 * Adds the specified multicharacter to this set if it is not already
 * present.  If this set already contains the multicharacter,
 * the call leaves this set unchanged.
 * Thus "ch" => {"ch"}
 * <br><b>Warning: you cannot add an empty string ("") to a UnicodeSet.</b>
 * @param s the source string
 * @return the modified set, for chaining
 */
UnicodeSet& UnicodeSet::add(const UnicodeString& s) {
    if (s.length() == 0 || isFrozen() || isBogus()) return *this;
    int32_t cp = getSingleCP(s);
    if (cp < 0) {
        if (!stringsContains(s)) {
            _add(s);
            releasePattern();
        }
    } else {
        add((UChar32)cp);
    }
    return *this;
}

/**
 * Adds the given string, in order, to 'strings'.  The given string
 * must have been checked by the caller to not be empty and to not
 * already be in 'strings'.
 */
void UnicodeSet::_add(const UnicodeString& s) {
    if (isFrozen() || isBogus()) {
        return;
    }
    UErrorCode ec = U_ZERO_ERROR;
    if (strings == nullptr && !allocateStrings(ec)) {
        setToBogus();
        return;
    }
    UnicodeString* t = new UnicodeString(s);
    if (t == NULL) { // Check for memory allocation error.
        setToBogus();
        return;
    }
    strings->sortedInsert(t, compareUnicodeString, ec);
    if (U_FAILURE(ec)) {
        setToBogus();
        delete t;
    }
}

/**
 * @return a code point IF the string consists of a single one.
 * otherwise returns -1.
 * @param string to test
 */
int32_t UnicodeSet::getSingleCP(const UnicodeString& s) {
    //if (s.length() < 1) {
    //    throw new IllegalArgumentException("Can't use zero-length strings in UnicodeSet");
    //}
    if (s.length() > 2) return -1;
    if (s.length() == 1) return s.charAt(0);

    // at this point, len = 2
    UChar32 cp = s.char32At(0);
    if (cp > 0xFFFF) { // is surrogate pair
        return cp;
    }
    return -1;
}

/**
 * Adds each of the characters in this string to the set. Thus "ch" => {"c", "h"}
 * If this set already any particular character, it has no effect on that character.
 * @param the source string
 * @return the modified set, for chaining
 */
UnicodeSet& UnicodeSet::addAll(const UnicodeString& s) {
    UChar32 cp;
    for (int32_t i = 0; i < s.length(); i += U16_LENGTH(cp)) {
        cp = s.char32At(i);
        add(cp);
    }
    return *this;
}

/**
 * Retains EACH of the characters in this string. Note: "ch" == {"c", "h"}
 * If this set already any particular character, it has no effect on that character.
 * @param the source string
 * @return the modified set, for chaining
 */
UnicodeSet& UnicodeSet::retainAll(const UnicodeString& s) {
    UnicodeSet set;
    set.addAll(s);
    retainAll(set);
    return *this;
}

/**
 * Complement EACH of the characters in this string. Note: "ch" == {"c", "h"}
 * If this set already any particular character, it has no effect on that character.
 * @param the source string
 * @return the modified set, for chaining
 */
UnicodeSet& UnicodeSet::complementAll(const UnicodeString& s) {
    UnicodeSet set;
    set.addAll(s);
    complementAll(set);
    return *this;
}

/**
 * Remove EACH of the characters in this string. Note: "ch" == {"c", "h"}
 * If this set already any particular character, it has no effect on that character.
 * @param the source string
 * @return the modified set, for chaining
 */
UnicodeSet& UnicodeSet::removeAll(const UnicodeString& s) {
    UnicodeSet set;
    set.addAll(s);
    removeAll(set);
    return *this;
}

UnicodeSet& UnicodeSet::removeAllStrings() {
    if (!isFrozen() && hasStrings()) {
        strings->removeAllElements();
        releasePattern();
    }
    return *this;
}


/**
 * Makes a set from a multicharacter string. Thus "ch" => {"ch"}
 * <br><b>Warning: you cannot add an empty string ("") to a UnicodeSet.</b>
 * @param the source string
 * @return a newly created set containing the given string
 */
UnicodeSet* U_EXPORT2 UnicodeSet::createFrom(const UnicodeString& s) {
    UnicodeSet *set = new UnicodeSet();
    if (set != NULL) { // Check for memory allocation error.
        set->add(s);
    }
    return set;
}


/**
 * Makes a set from each of the characters in the string. Thus "ch" => {"c", "h"}
 * @param the source string
 * @return a newly created set containing the given characters
 */
UnicodeSet* U_EXPORT2 UnicodeSet::createFromAll(const UnicodeString& s) {
    UnicodeSet *set = new UnicodeSet();
    if (set != NULL) { // Check for memory allocation error.
        set->addAll(s);
    }
    return set;
}

/**
 * Retain only the elements in this set that are contained in the
 * specified range.  If <code>end > start</code> then an empty range is
 * retained, leaving the set empty.
 *
 * @param start first character, inclusive, of range to be retained
 * to this set.
 * @param end last character, inclusive, of range to be retained
 * to this set.
 */
UnicodeSet& UnicodeSet::retain(UChar32 start, UChar32 end) {
    if (pinCodePoint(start) <= pinCodePoint(end)) {
        UChar32 range[3] = { start, end+1, UNICODESET_HIGH };
        retain(range, 2, 0);
    } else {
        clear();
    }
    return *this;
}

UnicodeSet& UnicodeSet::retain(UChar32 c) {
    return retain(c, c);
}

/**
 * Removes the specified range from this set if it is present.
 * The set will not contain the specified range once the call
 * returns.  If <code>end > start</code> then an empty range is
 * removed, leaving the set unchanged.
 *
 * @param start first character, inclusive, of range to be removed
 * from this set.
 * @param end last character, inclusive, of range to be removed
 * from this set.
 */
UnicodeSet& UnicodeSet::remove(UChar32 start, UChar32 end) {
    if (pinCodePoint(start) <= pinCodePoint(end)) {
        UChar32 range[3] = { start, end+1, UNICODESET_HIGH };
        retain(range, 2, 2);
    }
    return *this;
}

/**
 * Removes the specified character from this set if it is present.
 * The set will not contain the specified range once the call
 * returns.
 */
UnicodeSet& UnicodeSet::remove(UChar32 c) {
    return remove(c, c);
}

/**
 * Removes the specified string from this set if it is present.
 * The set will not contain the specified character once the call
 * returns.
 * @param the source string
 * @return the modified set, for chaining
 */
UnicodeSet& UnicodeSet::remove(const UnicodeString& s) {
    if (s.length() == 0 || isFrozen() || isBogus()) return *this;
    int32_t cp = getSingleCP(s);
    if (cp < 0) {
        if (strings != nullptr && strings->removeElement((void*) &s)) {
            releasePattern();
        }
    } else {
        remove((UChar32)cp, (UChar32)cp);
    }
    return *this;
}

/**
 * Complements the specified range in this set.  Any character in
 * the range will be removed if it is in this set, or will be
 * added if it is not in this set.  If <code>end > start</code>
 * then an empty range is xor'ed, leaving the set unchanged.
 *
 * @param start first character, inclusive, of range to be removed
 * from this set.
 * @param end last character, inclusive, of range to be removed
 * from this set.
 */
UnicodeSet& UnicodeSet::complement(UChar32 start, UChar32 end) {
    if (isFrozen() || isBogus()) {
        return *this;
    }
    if (pinCodePoint(start) <= pinCodePoint(end)) {
        UChar32 range[3] = { start, end+1, UNICODESET_HIGH };
        exclusiveOr(range, 2, 0);
    }
    releasePattern();
    return *this;
}

UnicodeSet& UnicodeSet::complement(UChar32 c) {
    return complement(c, c);
}

/**
 * This is equivalent to
 * <code>complement(MIN_VALUE, MAX_VALUE)</code>.
 */
UnicodeSet& UnicodeSet::complement(void) {
    if (isFrozen() || isBogus()) {
        return *this;
    }
    if (list[0] == UNICODESET_LOW) {
        uprv_memmove(list, list + 1, (size_t)(len-1)*sizeof(UChar32));
        --len;
    } else {
        if (!ensureCapacity(len+1)) {
            return *this;
        }
        uprv_memmove(list + 1, list, (size_t)len*sizeof(UChar32));
        list[0] = UNICODESET_LOW;
        ++len;
    }
    releasePattern();
    return *this;
}

/**
 * Complement the specified string in this set.
 * The set will not contain the specified string once the call
 * returns.
 * <br><b>Warning: you cannot add an empty string ("") to a UnicodeSet.</b>
 * @param s the string to complement
 * @return this object, for chaining
 */
UnicodeSet& UnicodeSet::complement(const UnicodeString& s) {
    if (s.length() == 0 || isFrozen() || isBogus()) return *this;
    int32_t cp = getSingleCP(s);
    if (cp < 0) {
        if (stringsContains(s)) {
            strings->removeElement((void*) &s);
        } else {
            _add(s);
        }
        releasePattern();
    } else {
        complement((UChar32)cp, (UChar32)cp);
    }
    return *this;
}

/**
 * Adds all of the elements in the specified set to this set if
 * they're not already present.  This operation effectively
 * modifies this set so that its value is the <i>union</i> of the two
 * sets.  The behavior of this operation is unspecified if the specified
 * collection is modified while the operation is in progress.
 *
 * @param c set whose elements are to be added to this set.
 * @see #add(char, char)
 */
UnicodeSet& UnicodeSet::addAll(const UnicodeSet& c) {
    if ( c.len>0 && c.list!=NULL ) {
        add(c.list, c.len, 0);
    }

    // Add strings in order
    if ( c.strings!=NULL ) {
        for (int32_t i=0; i<c.strings->size(); ++i) {
            const UnicodeString* s = (const UnicodeString*)c.strings->elementAt(i);
            if (!stringsContains(*s)) {
                _add(*s);
            }
        }
    }
    return *this;
}

/**
 * Retains only the elements in this set that are contained in the
 * specified set.  In other words, removes from this set all of
 * its elements that are not contained in the specified set.  This
 * operation effectively modifies this set so that its value is
 * the <i>intersection</i> of the two sets.
 *
 * @param c set that defines which elements this set will retain.
 */
UnicodeSet& UnicodeSet::retainAll(const UnicodeSet& c) {
    if (isFrozen() || isBogus()) {
        return *this;
    }
    retain(c.list, c.len, 0);
    if (hasStrings()) {
        if (!c.hasStrings()) {
            strings->removeAllElements();
        } else {
            strings->retainAll(*c.strings);
        }
    }
    return *this;
}

/**
 * Removes from this set all of its elements that are contained in the
 * specified set.  This operation effectively modifies this
 * set so that its value is the <i>asymmetric set difference</i> of
 * the two sets.
 *
 * @param c set that defines which elements will be removed from
 *          this set.
 */
UnicodeSet& UnicodeSet::removeAll(const UnicodeSet& c) {
    if (isFrozen() || isBogus()) {
        return *this;
    }
    retain(c.list, c.len, 2);
    if (hasStrings() && c.hasStrings()) {
        strings->removeAll(*c.strings);
    }
    return *this;
}

/**
 * Complements in this set all elements contained in the specified
 * set.  Any character in the other set will be removed if it is
 * in this set, or will be added if it is not in this set.
 *
 * @param c set that defines which elements will be xor'ed from
 *          this set.
 */
UnicodeSet& UnicodeSet::complementAll(const UnicodeSet& c) {
    if (isFrozen() || isBogus()) {
        return *this;
    }
    exclusiveOr(c.list, c.len, 0);

    if (c.strings != nullptr) {
        for (int32_t i=0; i<c.strings->size(); ++i) {
            void* e = c.strings->elementAt(i);
            if (strings == nullptr || !strings->removeElement(e)) {
                _add(*(const UnicodeString*)e);
            }
        }
    }
    return *this;
}

/**
 * Removes all of the elements from this set.  This set will be
 * empty after this call returns.
 */
UnicodeSet& UnicodeSet::clear(void) {
    if (isFrozen()) {
        return *this;
    }
    list[0] = UNICODESET_HIGH;
    len = 1;
    releasePattern();
    if (strings != NULL) {
        strings->removeAllElements();
    }
    // Remove bogus
    fFlags = 0;
    return *this;
}

/**
 * Iteration method that returns the number of ranges contained in
 * this set.
 * @see #getRangeStart
 * @see #getRangeEnd
 */
int32_t UnicodeSet::getRangeCount() const {
    return len/2;
}

/**
 * Iteration method that returns the first character in the
 * specified range of this set.
 * @see #getRangeCount
 * @see #getRangeEnd
 */
UChar32 UnicodeSet::getRangeStart(int32_t index) const {
    return list[index*2];
}

/**
 * Iteration method that returns the last character in the
 * specified range of this set.
 * @see #getRangeStart
 * @see #getRangeEnd
 */
UChar32 UnicodeSet::getRangeEnd(int32_t index) const {
    return list[index*2 + 1] - 1;
}

const UnicodeString* UnicodeSet::getString(int32_t index) const {
    return (const UnicodeString*) strings->elementAt(index);
}

/**
 * Reallocate this objects internal structures to take up the least
 * possible space, without changing this object's value.
 */
UnicodeSet& UnicodeSet::compact() {
    if (isFrozen() || isBogus()) {
        return *this;
    }
    // Delete buffer first to defragment memory less.
    if (buffer != stackList) {
        uprv_free(buffer);
        buffer = NULL;
        bufferCapacity = 0;
    }
    if (list == stackList) {
        // pass
    } else if (len <= INITIAL_CAPACITY) {
        uprv_memcpy(stackList, list, len * sizeof(UChar32));
        uprv_free(list);
        list = stackList;
        capacity = INITIAL_CAPACITY;
    } else if ((len + 7) < capacity) {
        // If we have more than a little unused capacity, shrink it to len.
        UChar32* temp = (UChar32*) uprv_realloc(list, sizeof(UChar32) * len);
        if (temp) {
            list = temp;
            capacity = len;
        }
        // else what the heck happened?! We allocated less memory!
        // Oh well. We'll keep our original array.
    }
    if (strings != nullptr && strings->isEmpty()) {
        delete strings;
        strings = nullptr;
    }
    return *this;
}

#ifdef DEBUG_SERIALIZE
#include <stdio.h>
#endif

/**
 * Deserialize constructor.
 */
UnicodeSet::UnicodeSet(const uint16_t data[], int32_t dataLen, ESerialization serialization,
                       UErrorCode &ec) {

  if(U_FAILURE(ec)) {
    setToBogus();
    return;
  }

  if( (serialization != kSerialized)
      || (data==NULL)
      || (dataLen < 1)) {
    ec = U_ILLEGAL_ARGUMENT_ERROR;
    setToBogus();
    return;
  }

  // bmp?
  int32_t headerSize = ((data[0]&0x8000)) ?2:1;
  int32_t bmpLength = (headerSize==1)?data[0]:data[1];

  int32_t newLength = (((data[0]&0x7FFF)-bmpLength)/2)+bmpLength;
#ifdef DEBUG_SERIALIZE
  printf("dataLen %d headerSize %d bmpLen %d len %d. data[0]=%X/%X/%X/%X\n", dataLen,headerSize,bmpLength,newLength, data[0],data[1],data[2],data[3]);
#endif
  if(!ensureCapacity(newLength + 1)) {  // +1 for HIGH
    return;
  }
  // copy bmp
  int32_t i;
  for(i = 0; i< bmpLength;i++) {
    list[i] = data[i+headerSize];
#ifdef DEBUG_SERIALIZE
    printf("<<16@%d[%d] %X\n", i+headerSize, i, list[i]);
#endif
  }
  // copy smp
  for(i=bmpLength;i<newLength;i++) {
    list[i] = ((UChar32)data[headerSize+bmpLength+(i-bmpLength)*2+0] << 16) +
              ((UChar32)data[headerSize+bmpLength+(i-bmpLength)*2+1]);
#ifdef DEBUG_SERIALIZE
    printf("<<32@%d+[%d] %lX\n", headerSize+bmpLength+i, i, list[i]);
#endif
  }
  U_ASSERT(i == newLength);
  if (i == 0 || list[i - 1] != UNICODESET_HIGH) {
    list[i++] = UNICODESET_HIGH;
  }
  len = i;
}


int32_t UnicodeSet::serialize(uint16_t *dest, int32_t destCapacity, UErrorCode& ec) const {
    int32_t bmpLength, length, destLength;

    if (U_FAILURE(ec)) {
        return 0;
    }

    if (destCapacity<0 || (destCapacity>0 && dest==NULL)) {
        ec=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* count necessary 16-bit units */
    length=this->len-1; // Subtract 1 to ignore final UNICODESET_HIGH
    // assert(length>=0);
    if (length==0) {
        /* empty set */
        if (destCapacity>0) {
            *dest=0;
        } else {
            ec=U_BUFFER_OVERFLOW_ERROR;
        }
        return 1;
    }
    /* now length>0 */

    if (this->list[length-1]<=0xffff) {
        /* all BMP */
        bmpLength=length;
    } else if (this->list[0]>=0x10000) {
        /* all supplementary */
        bmpLength=0;
        length*=2;
    } else {
        /* some BMP, some supplementary */
        for (bmpLength=0; bmpLength<length && this->list[bmpLength]<=0xffff; ++bmpLength) {}
        length=bmpLength+2*(length-bmpLength);
    }
#ifdef DEBUG_SERIALIZE
    printf(">> bmpLength%d length%d len%d\n", bmpLength, length, len);
#endif
    /* length: number of 16-bit array units */
    if (length>0x7fff) {
        /* there are only 15 bits for the length in the first serialized word */
        ec=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    /*
     * total serialized length:
     * number of 16-bit array units (length) +
     * 1 length unit (always) +
     * 1 bmpLength unit (if there are supplementary values)
     */
    destLength=length+((length>bmpLength)?2:1);
    if (destLength<=destCapacity) {
        const UChar32 *p;
        int32_t i;

#ifdef DEBUG_SERIALIZE
        printf("writeHdr\n");
#endif
        *dest=(uint16_t)length;
        if (length>bmpLength) {
            *dest|=0x8000;
            *++dest=(uint16_t)bmpLength;
        }
        ++dest;

        /* write the BMP part of the array */
        p=this->list;
        for (i=0; i<bmpLength; ++i) {
#ifdef DEBUG_SERIALIZE
          printf("writebmp: %x\n", (int)*p);
#endif
            *dest++=(uint16_t)*p++;
        }

        /* write the supplementary part of the array */
        for (; i<length; i+=2) {
#ifdef DEBUG_SERIALIZE
          printf("write32: %x\n", (int)*p);
#endif
            *dest++=(uint16_t)(*p>>16);
            *dest++=(uint16_t)*p++;
        }
    } else {
        ec=U_BUFFER_OVERFLOW_ERROR;
    }
    return destLength;
}

//----------------------------------------------------------------
// Implementation: Utility methods
//----------------------------------------------------------------

/**
 * Allocate our strings vector and return TRUE if successful.
 */
UBool UnicodeSet::allocateStrings(UErrorCode &status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    strings = new UVector(uprv_deleteUObject,
                          uhash_compareUnicodeString, 1, status);
    if (strings == NULL) { // Check for memory allocation error.
        status = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    if (U_FAILURE(status)) {
        delete strings;
        strings = NULL;
        return FALSE;
    }
    return TRUE;
}

int32_t UnicodeSet::nextCapacity(int32_t minCapacity) {
    // Grow exponentially to reduce the frequency of allocations.
    if (minCapacity < INITIAL_CAPACITY) {
        return minCapacity + INITIAL_CAPACITY;
    } else if (minCapacity <= 2500) {
        return 5 * minCapacity;
    } else {
        int32_t newCapacity = 2 * minCapacity;
        if (newCapacity > uniset_MAX_LENGTH) {
            newCapacity = uniset_MAX_LENGTH;
        }
        return newCapacity;
    }
}

bool UnicodeSet::ensureCapacity(int32_t newLen) {
    if (newLen > uniset_MAX_LENGTH) {
        newLen = uniset_MAX_LENGTH;
    }
    if (newLen <= capacity) {
        return true;
    }
    int32_t newCapacity = nextCapacity(newLen);
    UChar32* temp = (UChar32*) uprv_malloc(newCapacity * sizeof(UChar32));
    if (temp == NULL) {
        setToBogus(); // set the object to bogus state if an OOM failure occurred.
        return false;
    }
    // Copy only the actual contents.
    uprv_memcpy(temp, list, len * sizeof(UChar32));
    if (list != stackList) {
        uprv_free(list);
    }
    list = temp;
    capacity = newCapacity;
    return true;
}

bool UnicodeSet::ensureBufferCapacity(int32_t newLen) {
    if (newLen > uniset_MAX_LENGTH) {
        newLen = uniset_MAX_LENGTH;
    }
    if (newLen <= bufferCapacity) {
        return true;
    }
    int32_t newCapacity = nextCapacity(newLen);
    UChar32* temp = (UChar32*) uprv_malloc(newCapacity * sizeof(UChar32));
    if (temp == NULL) {
        setToBogus();
        return false;
    }
    // The buffer has no contents to be copied.
    // It is always filled from scratch after this call.
    if (buffer != stackList) {
        uprv_free(buffer);
    }
    buffer = temp;
    bufferCapacity = newCapacity;
    return true;
}

/**
 * Swap list and buffer.
 */
void UnicodeSet::swapBuffers(void) {
    // swap list and buffer
    UChar32* temp = list;
    list = buffer;
    buffer = temp;

    int32_t c = capacity;
    capacity = bufferCapacity;
    bufferCapacity = c;
}

void UnicodeSet::setToBogus() {
    clear(); // Remove everything in the set.
    fFlags = kIsBogus;
}

//----------------------------------------------------------------
// Implementation: Fundamental operators
//----------------------------------------------------------------

static inline UChar32 unichar_max(UChar32 a, UChar32 b) {
    return (a > b) ? a : b;
}

// polarity = 0, 3 is normal: x xor y
// polarity = 1, 2: x xor ~y == x === y

void UnicodeSet::exclusiveOr(const UChar32* other, int32_t otherLen, int8_t polarity) {
    if (isFrozen() || isBogus()) {
        return;
    }
    if (!ensureBufferCapacity(len + otherLen)) {
        return;
    }

    int32_t i = 0, j = 0, k = 0;
    UChar32 a = list[i++];
    UChar32 b;
    if (polarity == 1 || polarity == 2) {
        b = UNICODESET_LOW;
        if (other[j] == UNICODESET_LOW) { // skip base if already LOW
            ++j;
            b = other[j];
        }
    } else {
        b = other[j++];
    }
    // simplest of all the routines
    // sort the values, discarding identicals!
    for (;;) {
        if (a < b) {
            buffer[k++] = a;
            a = list[i++];
        } else if (b < a) {
            buffer[k++] = b;
            b = other[j++];
        } else if (a != UNICODESET_HIGH) { // at this point, a == b
            // discard both values!
            a = list[i++];
            b = other[j++];
        } else { // DONE!
            buffer[k++] = UNICODESET_HIGH;
            len = k;
            break;
        }
    }
    swapBuffers();
    releasePattern();
}

// polarity = 0 is normal: x union y
// polarity = 2: x union ~y
// polarity = 1: ~x union y
// polarity = 3: ~x union ~y

void UnicodeSet::add(const UChar32* other, int32_t otherLen, int8_t polarity) {
    if (isFrozen() || isBogus() || other==NULL) {
        return;
    }
    if (!ensureBufferCapacity(len + otherLen)) {
        return;
    }

    int32_t i = 0, j = 0, k = 0;
    UChar32 a = list[i++];
    UChar32 b = other[j++];
    // change from xor is that we have to check overlapping pairs
    // polarity bit 1 means a is second, bit 2 means b is.
    for (;;) {
        switch (polarity) {
          case 0: // both first; take lower if unequal
            if (a < b) { // take a
                // Back up over overlapping ranges in buffer[]
                if (k > 0 && a <= buffer[k-1]) {
                    // Pick latter end value in buffer[] vs. list[]
                    a = unichar_max(list[i], buffer[--k]);
                } else {
                    // No overlap
                    buffer[k++] = a;
                    a = list[i];
                }
                i++; // Common if/else code factored out
                polarity ^= 1;
            } else if (b < a) { // take b
                if (k > 0 && b <= buffer[k-1]) {
                    b = unichar_max(other[j], buffer[--k]);
                } else {
                    buffer[k++] = b;
                    b = other[j];
                }
                j++;
                polarity ^= 2;
            } else { // a == b, take a, drop b
                if (a == UNICODESET_HIGH) goto loop_end;
                // This is symmetrical; it doesn't matter if
                // we backtrack with a or b. - liu
                if (k > 0 && a <= buffer[k-1]) {
                    a = unichar_max(list[i], buffer[--k]);
                } else {
                    // No overlap
                    buffer[k++] = a;
                    a = list[i];
                }
                i++;
                polarity ^= 1;
                b = other[j++];
                polarity ^= 2;
            }
            break;
          case 3: // both second; take higher if unequal, and drop other
            if (b <= a) { // take a
                if (a == UNICODESET_HIGH) goto loop_end;
                buffer[k++] = a;
            } else { // take b
                if (b == UNICODESET_HIGH) goto loop_end;
                buffer[k++] = b;
            }
            a = list[i++];
            polarity ^= 1;   // factored common code
            b = other[j++];
            polarity ^= 2;
            break;
          case 1: // a second, b first; if b < a, overlap
            if (a < b) { // no overlap, take a
                buffer[k++] = a; a = list[i++]; polarity ^= 1;
            } else if (b < a) { // OVERLAP, drop b
                b = other[j++];
                polarity ^= 2;
            } else { // a == b, drop both!
                if (a == UNICODESET_HIGH) goto loop_end;
                a = list[i++];
                polarity ^= 1;
                b = other[j++];
                polarity ^= 2;
            }
            break;
          case 2: // a first, b second; if a < b, overlap
            if (b < a) { // no overlap, take b
                buffer[k++] = b;
                b = other[j++];
                polarity ^= 2;
            } else  if (a < b) { // OVERLAP, drop a
                a = list[i++];
                polarity ^= 1;
            } else { // a == b, drop both!
                if (a == UNICODESET_HIGH) goto loop_end;
                a = list[i++];
                polarity ^= 1;
                b = other[j++];
                polarity ^= 2;
            }
            break;
        }
    }
 loop_end:
    buffer[k++] = UNICODESET_HIGH;    // terminate
    len = k;
    swapBuffers();
    releasePattern();
}

// polarity = 0 is normal: x intersect y
// polarity = 2: x intersect ~y == set-minus
// polarity = 1: ~x intersect y
// polarity = 3: ~x intersect ~y

void UnicodeSet::retain(const UChar32* other, int32_t otherLen, int8_t polarity) {
    if (isFrozen() || isBogus()) {
        return;
    }
    if (!ensureBufferCapacity(len + otherLen)) {
        return;
    }

    int32_t i = 0, j = 0, k = 0;
    UChar32 a = list[i++];
    UChar32 b = other[j++];
    // change from xor is that we have to check overlapping pairs
    // polarity bit 1 means a is second, bit 2 means b is.
    for (;;) {
        switch (polarity) {
          case 0: // both first; drop the smaller
            if (a < b) { // drop a
                a = list[i++];
                polarity ^= 1;
            } else if (b < a) { // drop b
                b = other[j++];
                polarity ^= 2;
            } else { // a == b, take one, drop other
                if (a == UNICODESET_HIGH) goto loop_end;
                buffer[k++] = a;
                a = list[i++];
                polarity ^= 1;
                b = other[j++];
                polarity ^= 2;
            }
            break;
          case 3: // both second; take lower if unequal
            if (a < b) { // take a
                buffer[k++] = a;
                a = list[i++];
                polarity ^= 1;
            } else if (b < a) { // take b
                buffer[k++] = b;
                b = other[j++];
                polarity ^= 2;
            } else { // a == b, take one, drop other
                if (a == UNICODESET_HIGH) goto loop_end;
                buffer[k++] = a;
                a = list[i++];
                polarity ^= 1;
                b = other[j++];
                polarity ^= 2;
            }
            break;
          case 1: // a second, b first;
            if (a < b) { // NO OVERLAP, drop a
                a = list[i++];
                polarity ^= 1;
            } else if (b < a) { // OVERLAP, take b
                buffer[k++] = b;
                b = other[j++];
                polarity ^= 2;
            } else { // a == b, drop both!
                if (a == UNICODESET_HIGH) goto loop_end;
                a = list[i++];
                polarity ^= 1;
                b = other[j++];
                polarity ^= 2;
            }
            break;
          case 2: // a first, b second; if a < b, overlap
            if (b < a) { // no overlap, drop b
                b = other[j++];
                polarity ^= 2;
            } else  if (a < b) { // OVERLAP, take a
                buffer[k++] = a;
                a = list[i++];
                polarity ^= 1;
            } else { // a == b, drop both!
                if (a == UNICODESET_HIGH) goto loop_end;
                a = list[i++];
                polarity ^= 1;
                b = other[j++];
                polarity ^= 2;
            }
            break;
        }
    }
 loop_end:
    buffer[k++] = UNICODESET_HIGH;    // terminate
    len = k;
    swapBuffers();
    releasePattern();
}

/**
 * Append the <code>toPattern()</code> representation of a
 * string to the given <code>StringBuffer</code>.
 */
void UnicodeSet::_appendToPat(UnicodeString& buf, const UnicodeString& s, UBool
escapeUnprintable) {
    UChar32 cp;
    for (int32_t i = 0; i < s.length(); i += U16_LENGTH(cp)) {
        _appendToPat(buf, cp = s.char32At(i), escapeUnprintable);
    }
}

/**
 * Append the <code>toPattern()</code> representation of a
 * character to the given <code>StringBuffer</code>.
 */
void UnicodeSet::_appendToPat(UnicodeString& buf, UChar32 c, UBool
escapeUnprintable) {
    if (escapeUnprintable && ICU_Utility::isUnprintable(c)) {
        // Use hex escape notation (\uxxxx or \Uxxxxxxxx) for anything
        // unprintable
        if (ICU_Utility::escapeUnprintable(buf, c)) {
            return;
        }
    }
    // Okay to let ':' pass through
    switch (c) {
    case uniset_SET_OPEN:
    case uniset_SET_CLOSE:
    case uniset_HYPHEN:
    case uniset_COMPLEMENT:
    case uniset_INTERSECTION:
    case uniset_BACKSLASH:
    case uniset_OPEN_BRACE:
    case uniset_CLOSE_BRACE:
    case uniset_COLON:
    case SymbolTable::SYMBOL_REF:
        buf.append(uniset_BACKSLASH);
        break;
    default:
        // Escape whitespace
        if (PatternProps::isWhiteSpace(c)) {
            buf.append(uniset_BACKSLASH);
        }
        break;
    }
    buf.append(c);
}

/**
 * Append a string representation of this set to result.  This will be
 * a cleaned version of the string passed to applyPattern(), if there
 * is one.  Otherwise it will be generated.
 */
UnicodeString& UnicodeSet::_toPattern(UnicodeString& result,
                                      UBool escapeUnprintable) const
{
    if (pat != NULL) {
        int32_t i;
        int32_t backslashCount = 0;
        for (i=0; i<patLen; ) {
            UChar32 c;
            U16_NEXT(pat, i, patLen, c);
            if (escapeUnprintable && ICU_Utility::isUnprintable(c)) {
                // If the unprintable character is preceded by an odd
                // number of backslashes, then it has been escaped.
                // Before unescaping it, we delete the final
                // backslash.
                if ((backslashCount % 2) == 1) {
                    result.truncate(result.length() - 1);
                }
                ICU_Utility::escapeUnprintable(result, c);
                backslashCount = 0;
            } else {
                result.append(c);
                if (c == uniset_BACKSLASH) {
                    ++backslashCount;
                } else {
                    backslashCount = 0;
                }
            }
        }
        return result;
    }

    return _generatePattern(result, escapeUnprintable);
}

/**
 * Returns a string representation of this set.  If the result of
 * calling this function is passed to a UnicodeSet constructor, it
 * will produce another set that is equal to this one.
 */
UnicodeString& UnicodeSet::toPattern(UnicodeString& result,
                                     UBool escapeUnprintable) const
{
    result.truncate(0);
    return _toPattern(result, escapeUnprintable);
}

/**
 * Generate and append a string representation of this set to result.
 * This does not use this.pat, the cleaned up copy of the string
 * passed to applyPattern().
 */
UnicodeString& UnicodeSet::_generatePattern(UnicodeString& result,
                                            UBool escapeUnprintable) const
{
    result.append(uniset_SET_OPEN);

//  // Check against the predefined categories.  We implicitly build
//  // up ALL category sets the first time toPattern() is called.
//  for (int8_t cat=0; cat<Unicode::GENERAL_TYPES_COUNT; ++cat) {
//      if (*this == getCategorySet(cat)) {
//          result.append(uniset_COLON);
//          result.append(CATEGORY_NAMES, cat*2, 2);
//          return result.append(CATEGORY_CLOSE);
//      }
//  }

    int32_t count = getRangeCount();

    // If the set contains at least 2 intervals and includes both
    // MIN_VALUE and MAX_VALUE, then the inverse representation will
    // be more economical.
    if (count > 1 &&
        getRangeStart(0) == MIN_VALUE &&
        getRangeEnd(count-1) == MAX_VALUE) {

        // Emit the inverse
        result.append(uniset_COMPLEMENT);

        for (int32_t i = 1; i < count; ++i) {
            UChar32 start = getRangeEnd(i-1)+1;
            UChar32 end = getRangeStart(i)-1;
            _appendToPat(result, start, escapeUnprintable);
            if (start != end) {
                if ((start+1) != end) {
                    result.append(uniset_HYPHEN);
                }
                _appendToPat(result, end, escapeUnprintable);
            }
        }
    }

    // Default; emit the ranges as pairs
    else {
        for (int32_t i = 0; i < count; ++i) {
            UChar32 start = getRangeStart(i);
            UChar32 end = getRangeEnd(i);
            _appendToPat(result, start, escapeUnprintable);
            if (start != end) {
                if ((start+1) != end) {
                    result.append(uniset_HYPHEN);
                }
                _appendToPat(result, end, escapeUnprintable);
            }
        }
    }

    if (strings != nullptr) {
        for (int32_t i = 0; i<strings->size(); ++i) {
            result.append(uniset_OPEN_BRACE);
            _appendToPat(result,
                         *(const UnicodeString*) strings->elementAt(i),
                         escapeUnprintable);
            result.append(uniset_CLOSE_BRACE);
        }
    }
    return result.append(uniset_SET_CLOSE);
}

/**
* Release existing cached pattern
*/
void UnicodeSet::releasePattern() {
    if (pat) {
        uprv_free(pat);
        pat = NULL;
        patLen = 0;
    }
}

/**
* Set the new pattern to cache.
*/
void UnicodeSet::setPattern(const char16_t *newPat, int32_t newPatLen) {
    releasePattern();
    pat = (UChar *)uprv_malloc((newPatLen + 1) * sizeof(UChar));
    if (pat) {
        patLen = newPatLen;
        u_memcpy(pat, newPat, patLen);
        pat[patLen] = 0;
    }
    // else we don't care if malloc failed. This was just a nice cache.
    // We can regenerate an equivalent pattern later when requested.
}

UnicodeSet *UnicodeSet::freeze() {
    if(!isFrozen() && !isBogus()) {
        compact();

        // Optimize contains() and span() and similar functions.
        if (hasStrings()) {
            stringSpan = new UnicodeSetStringSpan(*this, *strings, UnicodeSetStringSpan::ALL);
            if (stringSpan == nullptr) {
                setToBogus();
                return this;
            } else if (!stringSpan->needsStringSpanUTF16()) {
                // All strings are irrelevant for span() etc. because
                // all of each string's code points are contained in this set.
                // Do not check needsStringSpanUTF8() because UTF-8 has at most as
                // many relevant strings as UTF-16.
                // (Thus needsStringSpanUTF8() implies needsStringSpanUTF16().)
                delete stringSpan;
                stringSpan = NULL;
            }
        }
        if (stringSpan == NULL) {
            // No span-relevant strings: Optimize for code point spans.
            bmpSet=new BMPSet(list, len);
            if (bmpSet == NULL) { // Check for memory allocation error.
                setToBogus();
            }
        }
    }
    return this;
}

int32_t UnicodeSet::span(const UChar *s, int32_t length, USetSpanCondition spanCondition) const {
    if(length>0 && bmpSet!=NULL) {
        return (int32_t)(bmpSet->span(s, s+length, spanCondition)-s);
    }
    if(length<0) {
        length=u_strlen(s);
    }
    if(length==0) {
        return 0;
    }
    if(stringSpan!=NULL) {
        return stringSpan->span(s, length, spanCondition);
    } else if(hasStrings()) {
        uint32_t which= spanCondition==USET_SPAN_NOT_CONTAINED ?
                            UnicodeSetStringSpan::FWD_UTF16_NOT_CONTAINED :
                            UnicodeSetStringSpan::FWD_UTF16_CONTAINED;
        UnicodeSetStringSpan strSpan(*this, *strings, which);
        if(strSpan.needsStringSpanUTF16()) {
            return strSpan.span(s, length, spanCondition);
        }
    }

    if(spanCondition!=USET_SPAN_NOT_CONTAINED) {
        spanCondition=USET_SPAN_CONTAINED;  // Pin to 0/1 values.
    }

    UChar32 c;
    int32_t start=0, prev=0;
    do {
        U16_NEXT(s, start, length, c);
        if(spanCondition!=contains(c)) {
            break;
        }
    } while((prev=start)<length);
    return prev;
}

int32_t UnicodeSet::spanBack(const UChar *s, int32_t length, USetSpanCondition spanCondition) const {
    if(length>0 && bmpSet!=NULL) {
        return (int32_t)(bmpSet->spanBack(s, s+length, spanCondition)-s);
    }
    if(length<0) {
        length=u_strlen(s);
    }
    if(length==0) {
        return 0;
    }
    if(stringSpan!=NULL) {
        return stringSpan->spanBack(s, length, spanCondition);
    } else if(hasStrings()) {
        uint32_t which= spanCondition==USET_SPAN_NOT_CONTAINED ?
                            UnicodeSetStringSpan::BACK_UTF16_NOT_CONTAINED :
                            UnicodeSetStringSpan::BACK_UTF16_CONTAINED;
        UnicodeSetStringSpan strSpan(*this, *strings, which);
        if(strSpan.needsStringSpanUTF16()) {
            return strSpan.spanBack(s, length, spanCondition);
        }
    }

    if(spanCondition!=USET_SPAN_NOT_CONTAINED) {
        spanCondition=USET_SPAN_CONTAINED;  // Pin to 0/1 values.
    }

    UChar32 c;
    int32_t prev=length;
    do {
        U16_PREV(s, 0, length, c);
        if(spanCondition!=contains(c)) {
            break;
        }
    } while((prev=length)>0);
    return prev;
}

int32_t UnicodeSet::spanUTF8(const char *s, int32_t length, USetSpanCondition spanCondition) const {
    if(length>0 && bmpSet!=NULL) {
        const uint8_t *s0=(const uint8_t *)s;
        return (int32_t)(bmpSet->spanUTF8(s0, length, spanCondition)-s0);
    }
    if(length<0) {
        length=(int32_t)uprv_strlen(s);
    }
    if(length==0) {
        return 0;
    }
    if(stringSpan!=NULL) {
        return stringSpan->spanUTF8((const uint8_t *)s, length, spanCondition);
    } else if(hasStrings()) {
        uint32_t which= spanCondition==USET_SPAN_NOT_CONTAINED ?
                            UnicodeSetStringSpan::FWD_UTF8_NOT_CONTAINED :
                            UnicodeSetStringSpan::FWD_UTF8_CONTAINED;
        UnicodeSetStringSpan strSpan(*this, *strings, which);
        if(strSpan.needsStringSpanUTF8()) {
            return strSpan.spanUTF8((const uint8_t *)s, length, spanCondition);
        }
    }

    if(spanCondition!=USET_SPAN_NOT_CONTAINED) {
        spanCondition=USET_SPAN_CONTAINED;  // Pin to 0/1 values.
    }

    UChar32 c;
    int32_t start=0, prev=0;
    do {
        U8_NEXT_OR_FFFD(s, start, length, c);
        if(spanCondition!=contains(c)) {
            break;
        }
    } while((prev=start)<length);
    return prev;
}

int32_t UnicodeSet::spanBackUTF8(const char *s, int32_t length, USetSpanCondition spanCondition) const {
    if(length>0 && bmpSet!=NULL) {
        const uint8_t *s0=(const uint8_t *)s;
        return bmpSet->spanBackUTF8(s0, length, spanCondition);
    }
    if(length<0) {
        length=(int32_t)uprv_strlen(s);
    }
    if(length==0) {
        return 0;
    }
    if(stringSpan!=NULL) {
        return stringSpan->spanBackUTF8((const uint8_t *)s, length, spanCondition);
    } else if(hasStrings()) {
        uint32_t which= spanCondition==USET_SPAN_NOT_CONTAINED ?
                            UnicodeSetStringSpan::BACK_UTF8_NOT_CONTAINED :
                            UnicodeSetStringSpan::BACK_UTF8_CONTAINED;
        UnicodeSetStringSpan strSpan(*this, *strings, which);
        if(strSpan.needsStringSpanUTF8()) {
            return strSpan.spanBackUTF8((const uint8_t *)s, length, spanCondition);
        }
    }

    if(spanCondition!=USET_SPAN_NOT_CONTAINED) {
        spanCondition=USET_SPAN_CONTAINED;  // Pin to 0/1 values.
    }

    UChar32 c;
    int32_t prev=length;
    do {
        U8_PREV_OR_FFFD(s, 0, length, c);
        if(spanCondition!=contains(c)) {
            break;
        }
    } while((prev=length)>0);
    return prev;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uniset_closure.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011may30
*   created by: Markus W. Scherer
*
*   UnicodeSet::closeOver() and related methods moved here from uniset_props.cpp
*   to simplify dependencies.
*   In particular, this depends on the BreakIterator, but the BreakIterator
*   code also builds UnicodeSets from patterns and needs uniset_props.
*/











U_NAMESPACE_BEGIN

// TODO memory debugging provided inside uniset.cpp
// could be made available here but probably obsolete with use of modern
// memory leak checker tools
#ifndef _dbgct
#define _dbgct(me)
#endif

//----------------------------------------------------------------
// Constructors &c
//----------------------------------------------------------------

UnicodeSet::UnicodeSet(const UnicodeString& pattern,
                       uint32_t options,
                       const SymbolTable* symbols,
                       UErrorCode& status) {
    applyPattern(pattern, options, symbols, status);
    _dbgct(this);
}

UnicodeSet::UnicodeSet(const UnicodeString& pattern, ParsePosition& pos,
                       uint32_t options,
                       const SymbolTable* symbols,
                       UErrorCode& status) {
    applyPattern(pattern, pos, options, symbols, status);
    _dbgct(this);
}

//----------------------------------------------------------------
// Public API
//----------------------------------------------------------------

UnicodeSet& UnicodeSet::applyPattern(const UnicodeString& pattern,
                                     uint32_t options,
                                     const SymbolTable* symbols,
                                     UErrorCode& status) {
    ParsePosition pos(0);
    applyPattern(pattern, pos, options, symbols, status);
    if (U_FAILURE(status)) return *this;

    int32_t i = pos.getIndex();

    if (options & USET_IGNORE_SPACE) {
        // Skip over trailing whitespace
        ICU_Utility::skipWhitespace(pattern, i, TRUE);
    }

    if (i != pattern.length()) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return *this;
}

UnicodeSet& UnicodeSet::applyPattern(const UnicodeString& pattern,
                              ParsePosition& pos,
                              uint32_t options,
                              const SymbolTable* symbols,
                              UErrorCode& status) {
    if (U_FAILURE(status)) {
        return *this;
    }
    if (isFrozen()) {
        status = U_NO_WRITE_PERMISSION;
        return *this;
    }
    // Need to build the pattern in a temporary string because
    // _applyPattern calls add() etc., which set pat to empty.
    UnicodeString rebuiltPat;
    RuleCharacterIterator chars(pattern, symbols, pos);
    applyPattern(chars, symbols, rebuiltPat, options, &UnicodeSet::closeOver, 0, status);
    if (U_FAILURE(status)) return *this;
    if (chars.inVariable()) {
        // syntaxError(chars, "Extra chars in variable value");
        status = U_MALFORMED_SET;
        return *this;
    }
    setPattern(rebuiltPat);
    return *this;
}

// USetAdder implementation
// Does not use uset.h to reduce code dependencies
static void U_CALLCONV
_set_add(USet *set, UChar32 c) {
    ((UnicodeSet *)set)->add(c);
}

static void U_CALLCONV
_set_addRange(USet *set, UChar32 start, UChar32 end) {
    ((UnicodeSet *)set)->add(start, end);
}

static void U_CALLCONV
_set_addString(USet *set, const UChar *str, int32_t length) {
    ((UnicodeSet *)set)->add(UnicodeString((UBool)(length<0), str, length));
}

//----------------------------------------------------------------
// Case folding API
//----------------------------------------------------------------

// add the result of a full case mapping to the set
// use str as a temporary string to avoid constructing one
static inline void
addCaseMapping(UnicodeSet &set, int32_t result, const UChar *full, UnicodeString &str) {
    if(result >= 0) {
        if(result > UCASE_MAX_STRING_LENGTH) {
            // add a single-code point case mapping
            set.add(result);
        } else {
            // add a string case mapping from full with length result
            str.setTo((UBool)FALSE, full, result);
            set.add(str);
        }
    }
    // result < 0: the code point mapped to itself, no need to add it
    // see ucase.h
}

UnicodeSet& UnicodeSet::closeOver(int32_t attribute) {
    if (isFrozen() || isBogus()) {
        return *this;
    }
    if (attribute & (USET_CASE_INSENSITIVE | USET_ADD_CASE_MAPPINGS)) {
        {
            UnicodeSet foldSet(*this);
            UnicodeString str;
            USetAdder sa = {
                foldSet.toUSet(),
                _set_add,
                _set_addRange,
                _set_addString,
                NULL, // don't need remove()
                NULL // don't need removeRange()
            };

            // start with input set to guarantee inclusion
            // USET_CASE: remove strings because the strings will actually be reduced (folded);
            //            therefore, start with no strings and add only those needed
            if ((attribute & USET_CASE_INSENSITIVE) && foldSet.hasStrings()) {
                foldSet.strings->removeAllElements();
            }

            int32_t n = getRangeCount();
            UChar32 result;
            const UChar *full;

            for (int32_t i=0; i<n; ++i) {
                UChar32 start = getRangeStart(i);
                UChar32 end   = getRangeEnd(i);

                if (attribute & USET_CASE_INSENSITIVE) {
                    // full case closure
                    for (UChar32 cp=start; cp<=end; ++cp) {
                        ucase_addCaseClosure(cp, &sa);
                    }
                } else {
                    // add case mappings
                    // (does not add long s for regular s, or Kelvin for k, for example)
                    for (UChar32 cp=start; cp<=end; ++cp) {
                        result = ucase_toFullLower(cp, NULL, NULL, &full, UCASE_LOC_ROOT);
                        addCaseMapping(foldSet, result, full, str);

                        result = ucase_toFullTitle(cp, NULL, NULL, &full, UCASE_LOC_ROOT);
                        addCaseMapping(foldSet, result, full, str);

                        result = ucase_toFullUpper(cp, NULL, NULL, &full, UCASE_LOC_ROOT);
                        addCaseMapping(foldSet, result, full, str);

                        result = ucase_toFullFolding(cp, &full, 0);
                        addCaseMapping(foldSet, result, full, str);
                    }
                }
            }
            if (hasStrings()) {
                if (attribute & USET_CASE_INSENSITIVE) {
                    for (int32_t j=0; j<strings->size(); ++j) {
                        str = *(const UnicodeString *) strings->elementAt(j);
                        str.foldCase();
                        if(!ucase_addStringCaseClosure(str.getBuffer(), str.length(), &sa)) {
                            foldSet.add(str); // does not map to code points: add the folded string itself
                        }
                    }
                } else {
                    Locale root("");
#if !UCONFIG_NO_BREAK_ITERATION
                    UErrorCode status = U_ZERO_ERROR;
                    BreakIterator *bi = BreakIterator::createWordInstance(root, status);
                    if (U_SUCCESS(status)) {
#endif
                        const UnicodeString *pStr;

                        for (int32_t j=0; j<strings->size(); ++j) {
                            pStr = (const UnicodeString *) strings->elementAt(j);
                            (str = *pStr).toLower(root);
                            foldSet.add(str);
#if !UCONFIG_NO_BREAK_ITERATION
                            (str = *pStr).toTitle(bi, root);
                            foldSet.add(str);
#endif
                            (str = *pStr).toUpper(root);
                            foldSet.add(str);
                            (str = *pStr).foldCase();
                            foldSet.add(str);
                        }
#if !UCONFIG_NO_BREAK_ITERATION
                    }
                    delete bi;
#endif
                }
            }
            *this = foldSet;
        }
    }
    return *this;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1999-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uniset_props.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004aug25
*   created by: Markus W. Scherer
*
*   Character property dependent functions moved here from uniset.cpp
*/




























U_NAMESPACE_USE

// Define UChar constants using hex for EBCDIC compatibility
// Used #define to reduce private static exports and memory access time.
#define uniset_props_SET_OPEN        ((UChar)0x005B) /*[*/
#define uniset_props_SET_CLOSE       ((UChar)0x005D) /*]*/
#define uniset_props_HYPHEN          ((UChar)0x002D) /*-*/
#define uniset_props_COMPLEMENT      ((UChar)0x005E) /*^*/
#define uniset_props_COLON           ((UChar)0x003A) /*:*/
#define uniset_props_BACKSLASH       ((UChar)0x005C) /*\*/
#define uniset_props_INTERSECTION    ((UChar)0x0026) /*&*/
#define uniset_props_UPPER_U         ((UChar)0x0055) /*U*/
#define uniset_props_LOWER_U         ((UChar)0x0075) /*u*/
#define uniset_props_OPEN_BRACE      ((UChar)123)    /*{*/
#define uniset_props_CLOSE_BRACE     ((UChar)125)    /*}*/
#define uniset_props_UPPER_P         ((UChar)0x0050) /*P*/
#define uniset_props_LOWER_P         ((UChar)0x0070) /*p*/
#define uniset_props_UPPER_N         ((UChar)78)     /*N*/
#define uniset_props_EQUALS          ((UChar)0x003D) /*=*/

//static const UChar POSIX_OPEN[]  = { uniset_props_SET_OPEN,uniset_props_COLON,0 };  // "[:"
static const UChar uniset_props_POSIX_CLOSE[] = { uniset_props_COLON,uniset_props_SET_CLOSE,0 };  // ":]"
//static const UChar PERL_OPEN[]   = { uniset_props_BACKSLASH,uniset_props_LOWER_P,0 }; // "\\p"
//static const UChar PERL_CLOSE[]  = { uniset_props_CLOSE_BRACE,0 };    // "}"
//static const UChar NAME_OPEN[]   = { uniset_props_BACKSLASH,uniset_props_UPPER_N,0 };  // "\\N"
static const UChar uniset_props_HYPHEN_RIGHT_BRACE[] = {uniset_props_HYPHEN,uniset_props_SET_CLOSE,0}; /*-]*/

// Special property set IDs
static const char uniset_props_ANY[]   = "uniset_props_ANY";   // [\u0000-\U0010FFFF]
static const char uniset_props_ASCII[] = "uniset_props_ASCII"; // [\u0000-\u007F]
static const char uniset_props_ASSIGNED[] = "Assigned"; // [:^Cn:]

// Unicode name property alias
#define uniset_props_NAME_PROP "na"
#define uniset_props_NAME_PROP_LENGTH 2

/**
 * Delimiter string used in patterns to close a category reference:
 * ":]".  Example: "[:Lu:]".
 */
//static const UChar CATEGORY_CLOSE[] = {uniset_props_COLON, uniset_props_SET_CLOSE, 0x0000}; /* ":]" */

// Cached sets ------------------------------------------------------------- ***

U_CDECL_BEGIN
static UBool U_CALLCONV uset_cleanup();

static UnicodeSet *uni32Singleton;
static icu::UInitOnce uni32InitOnce = U_INITONCE_INITIALIZER;

/**
 * Cleanup function for UnicodeSet
 */
static UBool U_CALLCONV uset_cleanup(void) {
    delete uni32Singleton;
    uni32Singleton = NULL;
    uni32InitOnce.reset();
    return TRUE;
}

U_CDECL_END

U_NAMESPACE_BEGIN

namespace {

// Cache some sets for other services -------------------------------------- ***
void U_CALLCONV createUni32Set(UErrorCode &errorCode) {
    U_ASSERT(uni32Singleton == NULL);
    uni32Singleton = new UnicodeSet(UNICODE_STRING_SIMPLE("[:age=3.2:]"), errorCode);
    if(uni32Singleton==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
    } else {
        uni32Singleton->freeze();
    }
    ucln_common_registerCleanup(UCLN_COMMON_USET, uset_cleanup);
}

// helper functions for matching of pattern syntax pieces ------------------ ***
// these functions are parallel to the PERL_OPEN etc. strings above

// using these functions is not only faster than UnicodeString::compare() and
// caseCompare(), but they also make UnicodeSet work for simple patterns when
// no Unicode properties data is available - when caseCompare() fails

static inline UBool
isPerlOpen(const UnicodeString &pattern, int32_t pos) {
    UChar c;
    return pattern.charAt(pos)==uniset_props_BACKSLASH && ((c=pattern.charAt(pos+1))==uniset_props_LOWER_P || c==uniset_props_UPPER_P);
}

/*static inline UBool
isPerlClose(const UnicodeString &pattern, int32_t pos) {
    return pattern.charAt(pos)==uniset_props_CLOSE_BRACE;
}*/

static inline UBool
isNameOpen(const UnicodeString &pattern, int32_t pos) {
    return pattern.charAt(pos)==uniset_props_BACKSLASH && pattern.charAt(pos+1)==uniset_props_UPPER_N;
}

static inline UBool
isPOSIXOpen(const UnicodeString &pattern, int32_t pos) {
    return pattern.charAt(pos)==uniset_props_SET_OPEN && pattern.charAt(pos+1)==uniset_props_COLON;
}

/*static inline UBool
isPOSIXClose(const UnicodeString &pattern, int32_t pos) {
    return pattern.charAt(pos)==uniset_props_COLON && pattern.charAt(pos+1)==uniset_props_SET_CLOSE;
}*/

// TODO memory debugging provided inside uniset.cpp
// could be made available here but probably obsolete with use of modern
// memory leak checker tools
#ifndef _dbgct
#define _dbgct(me)
#endif

}  // namespace

U_CFUNC UnicodeSet *
uniset_getUnicode32Instance(UErrorCode &errorCode) {
    umtx_initOnce(uni32InitOnce, &createUni32Set, errorCode);
    return uni32Singleton;
}

//----------------------------------------------------------------
// Constructors &c
//----------------------------------------------------------------

/**
 * Constructs a set from the given pattern, optionally ignoring
 * white space.  See the class description for the syntax of the
 * pattern language.
 * @param pattern a string specifying what characters are in the set
 */
UnicodeSet::UnicodeSet(const UnicodeString& pattern,
                       UErrorCode& status) {
    applyPattern(pattern, status);
    _dbgct(this);
}

//----------------------------------------------------------------
// Public API
//----------------------------------------------------------------

UnicodeSet& UnicodeSet::applyPattern(const UnicodeString& pattern,
                                     UErrorCode& status) {
    // Equivalent to
    //   return applyPattern(pattern, USET_IGNORE_SPACE, NULL, status);
    // but without dependency on closeOver().
    ParsePosition pos(0);
    applyPatternIgnoreSpace(pattern, pos, NULL, status);
    if (U_FAILURE(status)) return *this;

    int32_t i = pos.getIndex();
    // Skip over trailing whitespace
    ICU_Utility::skipWhitespace(pattern, i, TRUE);
    if (i != pattern.length()) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return *this;
}

void
UnicodeSet::applyPatternIgnoreSpace(const UnicodeString& pattern,
                                    ParsePosition& pos,
                                    const SymbolTable* symbols,
                                    UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    if (isFrozen()) {
        status = U_NO_WRITE_PERMISSION;
        return;
    }
    // Need to build the pattern in a temporary string because
    // _applyPattern calls add() etc., which set pat to empty.
    UnicodeString rebuiltPat;
    RuleCharacterIterator chars(pattern, symbols, pos);
    applyPattern(chars, symbols, rebuiltPat, USET_IGNORE_SPACE, NULL, 0, status);
    if (U_FAILURE(status)) return;
    if (chars.inVariable()) {
        // syntaxError(chars, "Extra chars in variable value");
        status = U_MALFORMED_SET;
        return;
    }
    setPattern(rebuiltPat);
}

/**
 * Return true if the given position, in the given pattern, appears
 * to be the start of a UnicodeSet pattern.
 */
UBool UnicodeSet::resemblesPattern(const UnicodeString& pattern, int32_t pos) {
    return ((pos+1) < pattern.length() &&
            pattern.charAt(pos) == (UChar)91/*[*/) ||
        resemblesPropertyPattern(pattern, pos);
}

//----------------------------------------------------------------
// Implementation: Pattern parsing
//----------------------------------------------------------------

namespace {

/**
 * A small all-inline class to manage a UnicodeSet pointer.  Add
 * operator->() etc. as needed.
 */
class UnicodeSetPointer {
    UnicodeSet* p;
public:
    inline UnicodeSetPointer() : p(0) {}
    inline ~UnicodeSetPointer() { delete p; }
    inline UnicodeSet* pointer() { return p; }
    inline UBool allocate() {
        if (p == 0) {
            p = new UnicodeSet();
        }
        return p != 0;
    }
};

constexpr int32_t MAX_DEPTH = 100;

}  // namespace

/**
 * Parse the pattern from the given RuleCharacterIterator.  The
 * iterator is advanced over the parsed pattern.
 * @param chars iterator over the pattern characters.  Upon return
 * it will be advanced to the first character after the parsed
 * pattern, or the end of the iteration if all characters are
 * parsed.
 * @param symbols symbol table to use to parse and dereference
 * variables, or null if none.
 * @param rebuiltPat the pattern that was parsed, rebuilt or
 * copied from the input pattern, as appropriate.
 * @param options a bit mask of zero or more of the following:
 * IGNORE_SPACE, CASE.
 */
void UnicodeSet::applyPattern(RuleCharacterIterator& chars,
                              const SymbolTable* symbols,
                              UnicodeString& rebuiltPat,
                              uint32_t options,
                              UnicodeSet& (UnicodeSet::*caseClosure)(int32_t attribute),
                              int32_t depth,
                              UErrorCode& ec) {
    if (U_FAILURE(ec)) return;
    if (depth > MAX_DEPTH) {
        ec = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    // Syntax characters: [ ] ^ - & { }

    // Recognized special forms for chars, sets: c-c s-s s&s

    int32_t opts = RuleCharacterIterator::PARSE_VARIABLES |
                   RuleCharacterIterator::PARSE_ESCAPES;
    if ((options & USET_IGNORE_SPACE) != 0) {
        opts |= RuleCharacterIterator::SKIP_WHITESPACE;
    }

    UnicodeString patLocal, buf;
    UBool usePat = FALSE;
    UnicodeSetPointer scratch;
    RuleCharacterIterator::Pos backup;

    // mode: 0=before [, 1=between [...], 2=after ]
    // lastItem: 0=none, 1=char, 2=set
    int8_t lastItem = 0, mode = 0;
    UChar32 lastChar = 0;
    UChar op = 0;

    UBool invert = FALSE;

    clear();

    while (mode != 2 && !chars.atEnd()) {
        U_ASSERT((lastItem == 0 && op == 0) ||
                 (lastItem == 1 && (op == 0 || op == uniset_props_HYPHEN /*'-'*/)) ||
                 (lastItem == 2 && (op == 0 || op == uniset_props_HYPHEN /*'-'*/ ||
                                    op == uniset_props_INTERSECTION /*'&'*/)));

        UChar32 c = 0;
        UBool literal = FALSE;
        UnicodeSet* nested = 0; // alias - do not delete

        // -------- Check for property pattern

        // setMode: 0=none, 1=unicodeset, 2=propertypat, 3=preparsed
        int8_t setMode = 0;
        if (resemblesPropertyPattern(chars, opts)) {
            setMode = 2;
        }

        // -------- Parse '[' of opening delimiter OR nested set.
        // If there is a nested set, use `setMode' to define how
        // the set should be parsed.  If the '[' is part of the
        // opening delimiter for this pattern, parse special
        // strings "[", "[^", "[-", and "[^-".  Check for stand-in
        // characters representing a nested set in the symbol
        // table.

        else {
            // Prepare to backup if necessary
            chars.getPos(backup);
            c = chars.next(opts, literal, ec);
            if (U_FAILURE(ec)) return;

            if (c == 0x5B /*'['*/ && !literal) {
                if (mode == 1) {
                    chars.setPos(backup); // backup
                    setMode = 1;
                } else {
                    // Handle opening '[' delimiter
                    mode = 1;
                    patLocal.append((UChar) 0x5B /*'['*/);
                    chars.getPos(backup); // prepare to backup
                    c = chars.next(opts, literal, ec);
                    if (U_FAILURE(ec)) return;
                    if (c == 0x5E /*'^'*/ && !literal) {
                        invert = TRUE;
                        patLocal.append((UChar) 0x5E /*'^'*/);
                        chars.getPos(backup); // prepare to backup
                        c = chars.next(opts, literal, ec);
                        if (U_FAILURE(ec)) return;
                    }
                    // Fall through to handle special leading '-';
                    // otherwise restart loop for nested [], \p{}, etc.
                    if (c == uniset_props_HYPHEN /*'-'*/) {
                        literal = TRUE;
                        // Fall through to handle literal '-' below
                    } else {
                        chars.setPos(backup); // backup
                        continue;
                    }
                }
            } else if (symbols != 0) {
                const UnicodeFunctor *m = symbols->lookupMatcher(c);
                if (m != 0) {
                    const UnicodeSet *ms = dynamic_cast<const UnicodeSet *>(m);
                    if (ms == NULL) {
                        ec = U_MALFORMED_SET;
                        return;
                    }
                    // casting away const, but `nested' won't be modified
                    // (important not to modify stored set)
                    nested = const_cast<UnicodeSet*>(ms);
                    setMode = 3;
                }
            }
        }

        // -------- Handle a nested set.  This either is inline in
        // the pattern or represented by a stand-in that has
        // previously been parsed and was looked up in the symbol
        // table.

        if (setMode != 0) {
            if (lastItem == 1) {
                if (op != 0) {
                    // syntaxError(chars, "Char expected after operator");
                    ec = U_MALFORMED_SET;
                    return;
                }
                add(lastChar, lastChar);
                _appendToPat(patLocal, lastChar, FALSE);
                lastItem = 0;
                op = 0;
            }

            if (op == uniset_props_HYPHEN /*'-'*/ || op == uniset_props_INTERSECTION /*'&'*/) {
                patLocal.append(op);
            }

            if (nested == 0) {
                // lazy allocation
                if (!scratch.allocate()) {
                    ec = U_MEMORY_ALLOCATION_ERROR;
                    return;
                }
                nested = scratch.pointer();
            }
            switch (setMode) {
            case 1:
                nested->applyPattern(chars, symbols, patLocal, options, caseClosure, depth + 1, ec);
                break;
            case 2:
                chars.skipIgnored(opts);
                nested->applyPropertyPattern(chars, patLocal, ec);
                if (U_FAILURE(ec)) return;
                break;
            case 3: // `nested' already parsed
                nested->_toPattern(patLocal, FALSE);
                break;
            }

            usePat = TRUE;

            if (mode == 0) {
                // Entire pattern is a category; leave parse loop
                *this = *nested;
                mode = 2;
                break;
            }

            switch (op) {
            case uniset_props_HYPHEN: /*'-'*/
                removeAll(*nested);
                break;
            case uniset_props_INTERSECTION: /*'&'*/
                retainAll(*nested);
                break;
            case 0:
                addAll(*nested);
                break;
            }

            op = 0;
            lastItem = 2;

            continue;
        }

        if (mode == 0) {
            // syntaxError(chars, "Missing '['");
            ec = U_MALFORMED_SET;
            return;
        }

        // -------- Parse special (syntax) characters.  If the
        // current character is not special, or if it is escaped,
        // then fall through and handle it below.

        if (!literal) {
            switch (c) {
            case 0x5D /*']'*/:
                if (lastItem == 1) {
                    add(lastChar, lastChar);
                    _appendToPat(patLocal, lastChar, FALSE);
                }
                // Treat final trailing '-' as a literal
                if (op == uniset_props_HYPHEN /*'-'*/) {
                    add(op, op);
                    patLocal.append(op);
                } else if (op == uniset_props_INTERSECTION /*'&'*/) {
                    // syntaxError(chars, "Trailing '&'");
                    ec = U_MALFORMED_SET;
                    return;
                }
                patLocal.append((UChar) 0x5D /*']'*/);
                mode = 2;
                continue;
            case uniset_props_HYPHEN /*'-'*/:
                if (op == 0) {
                    if (lastItem != 0) {
                        op = (UChar) c;
                        continue;
                    } else {
                        // Treat final trailing '-' as a literal
                        add(c, c);
                        c = chars.next(opts, literal, ec);
                        if (U_FAILURE(ec)) return;
                        if (c == 0x5D /*']'*/ && !literal) {
                            patLocal.append(uniset_props_HYPHEN_RIGHT_BRACE, 2);
                            mode = 2;
                            continue;
                        }
                    }
                }
                // syntaxError(chars, "'-' not after char or set");
                ec = U_MALFORMED_SET;
                return;
            case uniset_props_INTERSECTION /*'&'*/:
                if (lastItem == 2 && op == 0) {
                    op = (UChar) c;
                    continue;
                }
                // syntaxError(chars, "'&' not after set");
                ec = U_MALFORMED_SET;
                return;
            case 0x5E /*'^'*/:
                // syntaxError(chars, "'^' not after '['");
                ec = U_MALFORMED_SET;
                return;
            case 0x7B /*'{'*/:
                if (op != 0) {
                    // syntaxError(chars, "Missing operand after operator");
                    ec = U_MALFORMED_SET;
                    return;
                }
                if (lastItem == 1) {
                    add(lastChar, lastChar);
                    _appendToPat(patLocal, lastChar, FALSE);
                }
                lastItem = 0;
                buf.truncate(0);
                {
                    UBool ok = FALSE;
                    while (!chars.atEnd()) {
                        c = chars.next(opts, literal, ec);
                        if (U_FAILURE(ec)) return;
                        if (c == 0x7D /*'}'*/ && !literal) {
                            ok = TRUE;
                            break;
                        }
                        buf.append(c);
                    }
                    if (buf.length() < 1 || !ok) {
                        // syntaxError(chars, "Invalid multicharacter string");
                        ec = U_MALFORMED_SET;
                        return;
                    }
                }
                // We have new string. Add it to set and continue;
                // we don't need to drop through to the further
                // processing
                add(buf);
                patLocal.append((UChar) 0x7B /*'{'*/);
                _appendToPat(patLocal, buf, FALSE);
                patLocal.append((UChar) 0x7D /*'}'*/);
                continue;
            case SymbolTable::SYMBOL_REF:
                //         symbols  nosymbols
                // [a-$]   error    error (ambiguous)
                // [a$]    anchor   anchor
                // [a-$x]  var "x"* literal '$'
                // [a-$.]  error    literal '$'
                // *We won't get here in the case of var "x"
                {
                    chars.getPos(backup);
                    c = chars.next(opts, literal, ec);
                    if (U_FAILURE(ec)) return;
                    UBool anchor = (c == 0x5D /*']'*/ && !literal);
                    if (symbols == 0 && !anchor) {
                        c = SymbolTable::SYMBOL_REF;
                        chars.setPos(backup);
                        break; // literal '$'
                    }
                    if (anchor && op == 0) {
                        if (lastItem == 1) {
                            add(lastChar, lastChar);
                            _appendToPat(patLocal, lastChar, FALSE);
                        }
                        add(U_ETHER);
                        usePat = TRUE;
                        patLocal.append((UChar) SymbolTable::SYMBOL_REF);
                        patLocal.append((UChar) 0x5D /*']'*/);
                        mode = 2;
                        continue;
                    }
                    // syntaxError(chars, "Unquoted '$'");
                    ec = U_MALFORMED_SET;
                    return;
                }
            default:
                break;
            }
        }

        // -------- Parse literal characters.  This includes both
        // escaped chars ("\u4E01") and non-syntax characters
        // ("a").

        switch (lastItem) {
        case 0:
            lastItem = 1;
            lastChar = c;
            break;
        case 1:
            if (op == uniset_props_HYPHEN /*'-'*/) {
                if (lastChar >= c) {
                    // Don't allow redundant (a-a) or empty (b-a) ranges;
                    // these are most likely typos.
                    // syntaxError(chars, "Invalid range");
                    ec = U_MALFORMED_SET;
                    return;
                }
                add(lastChar, c);
                _appendToPat(patLocal, lastChar, FALSE);
                patLocal.append(op);
                _appendToPat(patLocal, c, FALSE);
                lastItem = 0;
                op = 0;
            } else {
                add(lastChar, lastChar);
                _appendToPat(patLocal, lastChar, FALSE);
                lastChar = c;
            }
            break;
        case 2:
            if (op != 0) {
                // syntaxError(chars, "Set expected after operator");
                ec = U_MALFORMED_SET;
                return;
            }
            lastChar = c;
            lastItem = 1;
            break;
        }
    }

    if (mode != 2) {
        // syntaxError(chars, "Missing ']'");
        ec = U_MALFORMED_SET;
        return;
    }

    chars.skipIgnored(opts);

    /**
     * Handle global flags (invert, case insensitivity).  If this
     * pattern should be compiled case-insensitive, then we need
     * to close over case BEFORE uniset_props_COMPLEMENTING.  This makes
     * patterns like /[^abc]/i work.
     */
    if ((options & USET_CASE_INSENSITIVE) != 0) {
        (this->*caseClosure)(USET_CASE_INSENSITIVE);
    }
    else if ((options & USET_ADD_CASE_MAPPINGS) != 0) {
        (this->*caseClosure)(USET_ADD_CASE_MAPPINGS);
    }
    if (invert) {
        complement();
    }

    // Use the rebuilt pattern (patLocal) only if necessary.  Prefer the
    // generated pattern.
    if (usePat) {
        rebuiltPat.append(patLocal);
    } else {
        _generatePattern(rebuiltPat, FALSE);
    }
    if (isBogus() && U_SUCCESS(ec)) {
        // We likely ran out of memory. AHHH!
        ec = U_MEMORY_ALLOCATION_ERROR;
    }
}

//----------------------------------------------------------------
// Property set implementation
//----------------------------------------------------------------

namespace {

static UBool numericValueFilter(UChar32 ch, void* context) {
    return u_getNumericValue(ch) == *(double*)context;
}

static UBool generalCategoryMaskFilter(UChar32 ch, void* context) {
    int32_t value = *(int32_t*)context;
    return (U_GET_GC_MASK((UChar32) ch) & value) != 0;
}

static UBool versionFilter(UChar32 ch, void* context) {
    static const UVersionInfo none = { 0, 0, 0, 0 };
    UVersionInfo v;
    u_charAge(ch, v);
    UVersionInfo* version = (UVersionInfo*)context;
    return uprv_memcmp(&v, &none, sizeof(v)) > 0 && uprv_memcmp(&v, version, sizeof(v)) <= 0;
}

typedef struct {
    UProperty prop;
    int32_t value;
} IntPropertyContext;

static UBool intPropertyFilter(UChar32 ch, void* context) {
    IntPropertyContext* c = (IntPropertyContext*)context;
    return u_getIntPropertyValue((UChar32) ch, c->prop) == c->value;
}

static UBool scriptExtensionsFilter(UChar32 ch, void* context) {
    return uscript_hasScript(ch, *(UScriptCode*)context);
}

}  // namespace

/**
 * Generic filter-based scanning code for UCD property UnicodeSets.
 */
void UnicodeSet::applyFilter(UnicodeSet::Filter filter,
                             void* context,
                             const UnicodeSet* inclusions,
                             UErrorCode &status) {
    if (U_FAILURE(status)) return;

    // Logically, walk through all Unicode characters, noting the start
    // and end of each range for which filter.contain(c) is
    // true.  Add each range to a set.
    //
    // To improve performance, use an inclusions set which
    // encodes information about character ranges that are known
    // to have identical properties.
    // inclusions contains the first characters of
    // same-value ranges for the given property.

    clear();

    UChar32 startHasProperty = -1;
    int32_t limitRange = inclusions->getRangeCount();

    for (int j=0; j<limitRange; ++j) {
        // get current range
        UChar32 start = inclusions->getRangeStart(j);
        UChar32 end = inclusions->getRangeEnd(j);

        // for all the code points in the range, process
        for (UChar32 ch = start; ch <= end; ++ch) {
            // only add to this UnicodeSet on inflection points --
            // where the hasProperty value changes to false
            if ((*filter)(ch, context)) {
                if (startHasProperty < 0) {
                    startHasProperty = ch;
                }
            } else if (startHasProperty >= 0) {
                add(startHasProperty, ch-1);
                startHasProperty = -1;
            }
        }
    }
    if (startHasProperty >= 0) {
        add((UChar32)startHasProperty, (UChar32)0x10FFFF);
    }
    if (isBogus() && U_SUCCESS(status)) {
        // We likely ran out of memory. AHHH!
        status = U_MEMORY_ALLOCATION_ERROR;
    }
}

namespace {

static UBool mungeCharName(char* dst, const char* src, int32_t dstCapacity) {
    /* Note: we use ' ' in compiler code page */
    int32_t j = 0;
    char ch;
    --dstCapacity; /* make room for term. zero */
    while ((ch = *src++) != 0) {
        if (ch == ' ' && (j==0 || (j>0 && dst[j-1]==' '))) {
            continue;
        }
        if (j >= dstCapacity) return FALSE;
        dst[j++] = ch;
    }
    if (j > 0 && dst[j-1] == ' ') --j;
    dst[j] = 0;
    return TRUE;
}

}  // namespace

//----------------------------------------------------------------
// Property set API
//----------------------------------------------------------------

#define FAIL(ec) UPRV_BLOCK_MACRO_BEGIN { \
    ec=U_ILLEGAL_ARGUMENT_ERROR; \
    return *this; \
} UPRV_BLOCK_MACRO_END

UnicodeSet&
UnicodeSet::applyIntPropertyValue(UProperty prop, int32_t value, UErrorCode& ec) {
    if (U_FAILURE(ec) || isFrozen()) { return *this; }
    if (prop == UCHAR_GENERAL_CATEGORY_MASK) {
        const UnicodeSet* inclusions = CharacterProperties::getInclusionsForProperty(prop, ec);
        applyFilter(generalCategoryMaskFilter, &value, inclusions, ec);
    } else if (prop == UCHAR_SCRIPT_EXTENSIONS) {
        const UnicodeSet* inclusions = CharacterProperties::getInclusionsForProperty(prop, ec);
        UScriptCode script = (UScriptCode)value;
        applyFilter(scriptExtensionsFilter, &script, inclusions, ec);
    } else if (0 <= prop && prop < UCHAR_BINARY_LIMIT) {
        if (value == 0 || value == 1) {
            const USet *set = u_getBinaryPropertySet(prop, &ec);
            if (U_FAILURE(ec)) { return *this; }
            copyFrom(*UnicodeSet::fromUSet(set), TRUE);
            if (value == 0) {
                complement();
            }
        } else {
            clear();
        }
    } else if (UCHAR_INT_START <= prop && prop < UCHAR_INT_LIMIT) {
        const UnicodeSet* inclusions = CharacterProperties::getInclusionsForProperty(prop, ec);
        IntPropertyContext c = {prop, value};
        applyFilter(intPropertyFilter, &c, inclusions, ec);
    } else {
        ec = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return *this;
}

UnicodeSet&
UnicodeSet::applyPropertyAlias(const UnicodeString& prop,
                               const UnicodeString& value,
                               UErrorCode& ec) {
    if (U_FAILURE(ec) || isFrozen()) return *this;

    // prop and value used to be converted to char * using the default
    // converter instead of the invariant conversion.
    // This should not be necessary because all Unicode property and value
    // names use only invariant characters.
    // If there are any variant characters, then we won't find them anyway.
    // Checking first avoids assertion failures in the conversion.
    if( !uprv_isInvariantUString(prop.getBuffer(), prop.length()) ||
        !uprv_isInvariantUString(value.getBuffer(), value.length())
    ) {
        FAIL(ec);
    }
    CharString pname, vname;
    pname.appendInvariantChars(prop, ec);
    vname.appendInvariantChars(value, ec);
    if (U_FAILURE(ec)) return *this;

    UProperty p;
    int32_t v;
    UBool invert = FALSE;

    if (value.length() > 0) {
        p = u_getPropertyEnum(pname.data());
        if (p == UCHAR_INVALID_CODE) FAIL(ec);

        // Treat gc as gcm
        if (p == UCHAR_GENERAL_CATEGORY) {
            p = UCHAR_GENERAL_CATEGORY_MASK;
        }

        if ((p >= UCHAR_BINARY_START && p < UCHAR_BINARY_LIMIT) ||
            (p >= UCHAR_INT_START && p < UCHAR_INT_LIMIT) ||
            (p >= UCHAR_MASK_START && p < UCHAR_MASK_LIMIT)) {
            v = u_getPropertyValueEnum(p, vname.data());
            if (v == UCHAR_INVALID_CODE) {
                // Handle numeric CCC
                if (p == UCHAR_CANONICAL_COMBINING_CLASS ||
                    p == UCHAR_TRAIL_CANONICAL_COMBINING_CLASS ||
                    p == UCHAR_LEAD_CANONICAL_COMBINING_CLASS) {
                    char* end;
                    double val = uprv_strtod(vname.data(), &end);
                    // Anything between 0 and 255 is valid even if unused.
                    // Cast double->int only after range check.
                    // We catch NaN here because comparing it with both 0 and 255 will be false
                    // (as are all comparisons with NaN).
                    if (*end != 0 || !(0 <= val && val <= 255) ||
                            (v = (int32_t)val) != val) {
                        // non-integral value or outside 0..255, or trailing junk
                        FAIL(ec);
                    }
                } else {
                    FAIL(ec);
                }
            }
        }

        else {

            switch (p) {
            case UCHAR_NUMERIC_VALUE:
                {
                    char* end;
                    double val = uprv_strtod(vname.data(), &end);
                    if (*end != 0) {
                        FAIL(ec);
                    }
                    applyFilter(numericValueFilter, &val,
                                CharacterProperties::getInclusionsForProperty(p, ec), ec);
                    return *this;
                }
            case UCHAR_NAME:
                {
                    // Must munge name, since u_charFromName() does not do
                    // 'loose' matching.
                    char buf[128]; // it suffices that this be > uprv_getMaxCharNameLength
                    if (!mungeCharName(buf, vname.data(), sizeof(buf))) FAIL(ec);
                    UChar32 ch = u_charFromName(U_EXTENDED_CHAR_NAME, buf, &ec);
                    if (U_SUCCESS(ec)) {
                        clear();
                        add(ch);
                        return *this;
                    } else {
                        FAIL(ec);
                    }
                }
            case UCHAR_UNICODE_1_NAME:
                // ICU 49 deprecates the Unicode_1_Name property APIs.
                FAIL(ec);
            case UCHAR_AGE:
                {
                    // Must munge name, since u_versionFromString() does not do
                    // 'loose' matching.
                    char buf[128];
                    if (!mungeCharName(buf, vname.data(), sizeof(buf))) FAIL(ec);
                    UVersionInfo version;
                    u_versionFromString(version, buf);
                    applyFilter(versionFilter, &version,
                                CharacterProperties::getInclusionsForProperty(p, ec), ec);
                    return *this;
                }
            case UCHAR_SCRIPT_EXTENSIONS:
                v = u_getPropertyValueEnum(UCHAR_SCRIPT, vname.data());
                if (v == UCHAR_INVALID_CODE) {
                    FAIL(ec);
                }
                // fall through to calling applyIntPropertyValue()
                break;
            default:
                // p is a non-binary, non-enumerated property that we
                // don't support (yet).
                FAIL(ec);
            }
        }
    }

    else {
        // value is empty.  Interpret as General Category, Script, or
        // Binary property.
        p = UCHAR_GENERAL_CATEGORY_MASK;
        v = u_getPropertyValueEnum(p, pname.data());
        if (v == UCHAR_INVALID_CODE) {
            p = UCHAR_SCRIPT;
            v = u_getPropertyValueEnum(p, pname.data());
            if (v == UCHAR_INVALID_CODE) {
                p = u_getPropertyEnum(pname.data());
                if (p >= UCHAR_BINARY_START && p < UCHAR_BINARY_LIMIT) {
                    v = 1;
                } else if (0 == uprv_comparePropertyNames(uniset_props_ANY, pname.data())) {
                    set(MIN_VALUE, MAX_VALUE);
                    return *this;
                } else if (0 == uprv_comparePropertyNames(uniset_props_ASCII, pname.data())) {
                    set(0, 0x7F);
                    return *this;
                } else if (0 == uprv_comparePropertyNames(uniset_props_ASSIGNED, pname.data())) {
                    // [:Assigned:]=[:^Cn:]
                    p = UCHAR_GENERAL_CATEGORY_MASK;
                    v = U_GC_CN_MASK;
                    invert = TRUE;
                } else {
                    FAIL(ec);
                }
            }
        }
    }

    applyIntPropertyValue(p, v, ec);
    if(invert) {
        complement();
    }

    if (isBogus() && U_SUCCESS(ec)) {
        // We likely ran out of memory. AHHH!
        ec = U_MEMORY_ALLOCATION_ERROR;
    }
    return *this;
}

//----------------------------------------------------------------
// Property set patterns
//----------------------------------------------------------------

/**
 * Return true if the given position, in the given pattern, appears
 * to be the start of a property set pattern.
 */
UBool UnicodeSet::resemblesPropertyPattern(const UnicodeString& pattern,
                                           int32_t pos) {
    // Patterns are at least 5 characters long
    if ((pos+5) > pattern.length()) {
        return FALSE;
    }

    // Look for an opening [:, [:^, \p, or \P
    return isPOSIXOpen(pattern, pos) || isPerlOpen(pattern, pos) || isNameOpen(pattern, pos);
}

/**
 * Return true if the given iterator appears to point at a
 * property pattern.  Regardless of the result, return with the
 * iterator unchanged.
 * @param chars iterator over the pattern characters.  Upon return
 * it will be unchanged.
 * @param iterOpts RuleCharacterIterator options
 */
UBool UnicodeSet::resemblesPropertyPattern(RuleCharacterIterator& chars,
                                           int32_t iterOpts) {
    // NOTE: literal will always be FALSE, because we don't parse escapes.
    UBool result = FALSE, literal;
    UErrorCode ec = U_ZERO_ERROR;
    iterOpts &= ~RuleCharacterIterator::PARSE_ESCAPES;
    RuleCharacterIterator::Pos pos;
    chars.getPos(pos);
    UChar32 c = chars.next(iterOpts, literal, ec);
    if (c == 0x5B /*'['*/ || c == 0x5C /*'\\'*/) {
        UChar32 d = chars.next(iterOpts & ~RuleCharacterIterator::SKIP_WHITESPACE,
                               literal, ec);
        result = (c == 0x5B /*'['*/) ? (d == 0x3A /*':'*/) :
                 (d == 0x4E /*'N'*/ || d == 0x70 /*'p'*/ || d == 0x50 /*'P'*/);
    }
    chars.setPos(pos);
    return result && U_SUCCESS(ec);
}

/**
 * Parse the given property pattern at the given parse position.
 */
UnicodeSet& UnicodeSet::applyPropertyPattern(const UnicodeString& pattern,
                                             ParsePosition& ppos,
                                             UErrorCode &ec) {
    int32_t pos = ppos.getIndex();

    UBool posix = FALSE; // true for [:pat:], false for \p{pat} \P{pat} \N{pat}
    UBool isName = FALSE; // true for \N{pat}, o/w false
    UBool invert = FALSE;

    if (U_FAILURE(ec)) return *this;

    // Minimum length is 5 characters, e.g. \p{L}
    if ((pos+5) > pattern.length()) {
        FAIL(ec);
    }

    // On entry, ppos should point to one of the following locations:
    // Look for an opening [:, [:^, \p, or \P
    if (isPOSIXOpen(pattern, pos)) {
        posix = TRUE;
        pos += 2;
        pos = ICU_Utility::skipWhitespace(pattern, pos);
        if (pos < pattern.length() && pattern.charAt(pos) == uniset_props_COMPLEMENT) {
            ++pos;
            invert = TRUE;
        }
    } else if (isPerlOpen(pattern, pos) || isNameOpen(pattern, pos)) {
        UChar c = pattern.charAt(pos+1);
        invert = (c == uniset_props_UPPER_P);
        isName = (c == uniset_props_UPPER_N);
        pos += 2;
        pos = ICU_Utility::skipWhitespace(pattern, pos);
        if (pos == pattern.length() || pattern.charAt(pos++) != uniset_props_OPEN_BRACE) {
            // Syntax error; "\p" or "\P" not followed by "{"
            FAIL(ec);
        }
    } else {
        // Open delimiter not seen
        FAIL(ec);
    }

    // Look for the matching close delimiter, either :] or }
    int32_t close;
    if (posix) {
      close = pattern.indexOf(uniset_props_POSIX_CLOSE, 2, pos);
    } else {
      close = pattern.indexOf(uniset_props_CLOSE_BRACE, pos);
    }
    if (close < 0) {
        // Syntax error; close delimiter missing
        FAIL(ec);
    }

    // Look for an '=' sign.  If this is present, we will parse a
    // medium \p{gc=Cf} or long \p{GeneralCategory=Format}
    // pattern.
    int32_t equals = pattern.indexOf(uniset_props_EQUALS, pos);
    UnicodeString propName, valueName;
    if (equals >= 0 && equals < close && !isName) {
        // Equals seen; parse medium/long pattern
        pattern.extractBetween(pos, equals, propName);
        pattern.extractBetween(equals+1, close, valueName);
    }

    else {
        // Handle case where no '=' is seen, and \N{}
        pattern.extractBetween(pos, close, propName);

        // Handle \N{name}
        if (isName) {
            // This is a little inefficient since it means we have to
            // parse uniset_props_NAME_PROP back to UCHAR_NAME even though we already
            // know it's UCHAR_NAME.  If we refactor the API to
            // support args of (UProperty, char*) then we can remove
            // uniset_props_NAME_PROP and make this a little more efficient.
            valueName = propName;
            propName = UnicodeString(uniset_props_NAME_PROP, uniset_props_NAME_PROP_LENGTH, US_INV);
        }
    }

    applyPropertyAlias(propName, valueName, ec);

    if (U_SUCCESS(ec)) {
        if (invert) {
            complement();
        }

        // Move to the limit position after the close delimiter if the
        // parse succeeded.
        ppos.setIndex(close + (posix ? 2 : 1));
    }

    return *this;
}

/**
 * Parse a property pattern.
 * @param chars iterator over the pattern characters.  Upon return
 * it will be advanced to the first character after the parsed
 * pattern, or the end of the iteration if all characters are
 * parsed.
 * @param rebuiltPat the pattern that was parsed, rebuilt or
 * copied from the input pattern, as appropriate.
 */
void UnicodeSet::applyPropertyPattern(RuleCharacterIterator& chars,
                                      UnicodeString& rebuiltPat,
                                      UErrorCode& ec) {
    if (U_FAILURE(ec)) return;
    UnicodeString pattern;
    chars.lookahead(pattern);
    ParsePosition pos(0);
    applyPropertyPattern(pattern, pos, ec);
    if (U_FAILURE(ec)) return;
    if (pos.getIndex() == 0) {
        // syntaxError(chars, "Invalid property pattern");
        ec = U_MALFORMED_SET;
        return;
    }
    chars.jumpahead(pos.getIndex());
    rebuiltPat.append(pattern, 0, pos.getIndex());
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2007-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  unisetspan.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2007mar01
*   created by: Markus W. Scherer
*/










U_NAMESPACE_BEGIN

/*
 * List of offsets from the current position from where to try matching
 * a code point or a string.
 * Store offsets rather than indexes to simplify the code and use the same list
 * for both increments (in span()) and decrements (in spanBack()).
 *
 * Assumption: The maximum offset is limited, and the offsets that are stored
 * at any one time are relatively dense, that is, there are normally no gaps of
 * hundreds or thousands of offset values.
 *
 * The implementation uses a circular buffer of byte flags,
 * each indicating whether the corresponding offset is in the list.
 * This avoids inserting into a sorted list of offsets (or absolute indexes) and
 * physically moving part of the list.
 *
 * Note: In principle, the caller should setMaxLength() to the maximum of the
 * max string length and U16_LENGTH/U8_LENGTH to account for
 * "long" single code points.
 * However, this implementation uses at least a staticList with more than
 * U8_LENGTH entries anyway.
 *
 * Note: If maxLength were guaranteed to be no more than 32 or 64,
 * the list could be stored as bit flags in a single integer.
 * Rather than handling a circular buffer with a start list index,
 * the integer would simply be shifted when lower offsets are removed.
 * UnicodeSet does not have a limit on the lengths of strings.
 */
class OffsetList {  // Only ever stack-allocated, does not need to inherit UMemory.
public:
    OffsetList() : list(staticList), capacity(0), length(0), start(0) {}

    ~OffsetList() {
        if(list!=staticList) {
            uprv_free(list);
        }
    }

    // Call exactly once if the list is to be used.
    void setMaxLength(int32_t maxLength) {
        if(maxLength<=(int32_t)sizeof(staticList)) {
            capacity=(int32_t)sizeof(staticList);
        } else {
            UBool *l=(UBool *)uprv_malloc(maxLength);
            if(l!=NULL) {
                list=l;
                capacity=maxLength;
            }
        }
        uprv_memset(list, 0, capacity);
    }

    void clear() {
        uprv_memset(list, 0, capacity);
        start=length=0;
    }

    UBool isEmpty() const {
        return (UBool)(length==0);
    }

    // Reduce all stored offsets by delta, used when the current position
    // moves by delta.
    // There must not be any offsets lower than delta.
    // If there is an offset equal to delta, it is removed.
    // delta=[1..maxLength]
    void shift(int32_t delta) {
        int32_t i=start+delta;
        if(i>=capacity) {
            i-=capacity;
        }
        if(list[i]) {
            list[i]=FALSE;
            --length;
        }
        start=i;
    }

    // Add an offset. The list must not contain it yet.
    // offset=[1..maxLength]
    void addOffset(int32_t offset) {
        int32_t i=start+offset;
        if(i>=capacity) {
            i-=capacity;
        }
        list[i]=TRUE;
        ++length;
    }

    // offset=[1..maxLength]
    UBool containsOffset(int32_t offset) const {
        int32_t i=start+offset;
        if(i>=capacity) {
            i-=capacity;
        }
        return list[i];
    }

    // Find the lowest stored offset from a non-empty list, remove it,
    // and reduce all other offsets by this minimum.
    // Returns [1..maxLength].
    int32_t popMinimum() {
        // Look for the next offset in list[start+1..capacity-1].
        int32_t i=start, result;
        while(++i<capacity) {
            if(list[i]) {
                list[i]=FALSE;
                --length;
                result=i-start;
                start=i;
                return result;
            }
        }
        // i==capacity

        // Wrap around and look for the next offset in list[0..start].
        // Since the list is not empty, there will be one.
        result=capacity-start;
        i=0;
        while(!list[i]) {
            ++i;
        }
        list[i]=FALSE;
        --length;
        start=i;
        return result+=i;
    }

private:
    UBool *list;
    int32_t capacity;
    int32_t length;
    int32_t start;

    UBool staticList[16];
};

// Get the number of UTF-8 bytes for a UTF-16 (sub)string.
static int32_t
getUTF8Length(const UChar *s, int32_t length) {
    UErrorCode errorCode=U_ZERO_ERROR;
    int32_t length8=0;
    u_strToUTF8(NULL, 0, &length8, s, length, &errorCode);
    if(U_SUCCESS(errorCode) || errorCode==U_BUFFER_OVERFLOW_ERROR) {
        return length8;
    } else {
        // The string contains an unpaired surrogate.
        // Ignore this string.
        return 0;
    }
}

// Append the UTF-8 version of the string to t and return the appended UTF-8 length.
static int32_t
appendUTF8(const UChar *s, int32_t length, uint8_t *t, int32_t capacity) {
    UErrorCode errorCode=U_ZERO_ERROR;
    int32_t length8=0;
    u_strToUTF8((char *)t, capacity, &length8, s, length, &errorCode);
    if(U_SUCCESS(errorCode)) {
        return length8;
    } else {
        // The string contains an unpaired surrogate.
        // Ignore this string.
        return 0;
    }
}

static inline uint8_t
makeSpanLengthByte(int32_t spanLength) {
    // 0xfe==UnicodeSetStringSpan::LONG_SPAN
    return spanLength<0xfe ? (uint8_t)spanLength : (uint8_t)0xfe;
}

// Construct for all variants of span(), or only for any one variant.
// Initialize as little as possible, for single use.
UnicodeSetStringSpan::UnicodeSetStringSpan(const UnicodeSet &set,
                                           const UVector &setStrings,
                                           uint32_t which)
        : spanSet(0, 0x10ffff), pSpanNotSet(NULL), strings(setStrings),
          utf8Lengths(NULL), spanLengths(NULL), utf8(NULL),
          utf8Length(0),
          maxLength16(0), maxLength8(0),
          all((UBool)(which==ALL)) {
    spanSet.retainAll(set);
    if(which&NOT_CONTAINED) {
        // Default to the same sets.
        // addToSpanNotSet() will create a separate set if necessary.
        pSpanNotSet=&spanSet;
    }

    // Determine if the strings even need to be taken into account at all for span() etc.
    // If any string is relevant, then all strings need to be used for
    // span(longest match) but only the relevant ones for span(while contained).
    // TODO: Possible optimization: Distinguish CONTAINED vs. LONGEST_MATCH
    //   and do not store UTF-8 strings if !thisRelevant and CONTAINED.
    //   (Only store irrelevant UTF-8 strings for LONGEST_MATCH where they are relevant after all.)
    // Also count the lengths of the UTF-8 versions of the strings for memory allocation.
    int32_t stringsLength=strings.size();

    int32_t i, spanLength;
    UBool someRelevant=FALSE;
    for(i=0; i<stringsLength; ++i) {
        const UnicodeString &string=*(const UnicodeString *)strings.elementAt(i);
        const UChar *s16=string.getBuffer();
        int32_t length16=string.length();
        UBool thisRelevant;
        spanLength=spanSet.span(s16, length16, USET_SPAN_CONTAINED);
        if(spanLength<length16) {  // Relevant string.
            someRelevant=thisRelevant=TRUE;
        } else {
            thisRelevant=FALSE;
        }
        if((which&UTF16) && length16>maxLength16) {
            maxLength16=length16;
        }
        if((which&UTF8) && (thisRelevant || (which&CONTAINED))) {
            int32_t length8=getUTF8Length(s16, length16);
            utf8Length+=length8;
            if(length8>maxLength8) {
                maxLength8=length8;
            }
        }
    }
    if(!someRelevant) {
        maxLength16=maxLength8=0;
        return;
    }

    // Freeze after checking for the need to use strings at all because freezing
    // a set takes some time and memory which are wasted if there are no relevant strings.
    if(all) {
        spanSet.freeze();
    }

    uint8_t *spanBackLengths;
    uint8_t *spanUTF8Lengths;
    uint8_t *spanBackUTF8Lengths;

    // Allocate a block of meta data.
    int32_t allocSize;
    if(all) {
        // UTF-8 lengths, 4 sets of span lengths, UTF-8 strings.
        allocSize=stringsLength*(4+1+1+1+1)+utf8Length;
    } else {
        allocSize=stringsLength;  // One set of span lengths.
        if(which&UTF8) {
            // UTF-8 lengths and UTF-8 strings.
            allocSize+=stringsLength*4+utf8Length;
        }
    }
    if(allocSize<=(int32_t)sizeof(staticLengths)) {
        utf8Lengths=staticLengths;
    } else {
        utf8Lengths=(int32_t *)uprv_malloc(allocSize);
        if(utf8Lengths==NULL) {
            maxLength16=maxLength8=0;  // Prevent usage by making needsStringSpanUTF16/8() return FALSE.
            return;  // Out of memory.
        }
    }

    if(all) {
        // Store span lengths for all span() variants.
        spanLengths=(uint8_t *)(utf8Lengths+stringsLength);
        spanBackLengths=spanLengths+stringsLength;
        spanUTF8Lengths=spanBackLengths+stringsLength;
        spanBackUTF8Lengths=spanUTF8Lengths+stringsLength;
        utf8=spanBackUTF8Lengths+stringsLength;
    } else {
        // Store span lengths for only one span() variant.
        if(which&UTF8) {
            spanLengths=(uint8_t *)(utf8Lengths+stringsLength);
            utf8=spanLengths+stringsLength;
        } else {
            spanLengths=(uint8_t *)utf8Lengths;
        }
        spanBackLengths=spanUTF8Lengths=spanBackUTF8Lengths=spanLengths;
    }

    // Set the meta data and pSpanNotSet and write the UTF-8 strings.
    int32_t utf8Count=0;  // Count UTF-8 bytes written so far.

    for(i=0; i<stringsLength; ++i) {
        const UnicodeString &string=*(const UnicodeString *)strings.elementAt(i);
        const UChar *s16=string.getBuffer();
        int32_t length16=string.length();
        spanLength=spanSet.span(s16, length16, USET_SPAN_CONTAINED);
        if(spanLength<length16) {  // Relevant string.
            if(which&UTF16) {
                if(which&CONTAINED) {
                    if(which&FWD) {
                        spanLengths[i]=makeSpanLengthByte(spanLength);
                    }
                    if(which&BACK) {
                        spanLength=length16-spanSet.spanBack(s16, length16, USET_SPAN_CONTAINED);
                        spanBackLengths[i]=makeSpanLengthByte(spanLength);
                    }
                } else /* not CONTAINED, not all, but NOT_CONTAINED */ {
                    spanLengths[i]=spanBackLengths[i]=0;  // Only store a relevant/irrelevant flag.
                }
            }
            if(which&UTF8) {
                uint8_t *s8=utf8+utf8Count;
                int32_t length8=appendUTF8(s16, length16, s8, utf8Length-utf8Count);
                utf8Count+=utf8Lengths[i]=length8;
                if(length8==0) {  // Irrelevant for UTF-8 because not representable in UTF-8.
                    spanUTF8Lengths[i]=spanBackUTF8Lengths[i]=(uint8_t)ALL_CP_CONTAINED;
                } else {  // Relevant for UTF-8.
                    if(which&CONTAINED) {
                        if(which&FWD) {
                            spanLength=spanSet.spanUTF8((const char *)s8, length8, USET_SPAN_CONTAINED);
                            spanUTF8Lengths[i]=makeSpanLengthByte(spanLength);
                        }
                        if(which&BACK) {
                            spanLength=length8-spanSet.spanBackUTF8((const char *)s8, length8, USET_SPAN_CONTAINED);
                            spanBackUTF8Lengths[i]=makeSpanLengthByte(spanLength);
                        }
                    } else /* not CONTAINED, not all, but NOT_CONTAINED */ {
                        spanUTF8Lengths[i]=spanBackUTF8Lengths[i]=0;  // Only store a relevant/irrelevant flag.
                    }
                }
            }
            if(which&NOT_CONTAINED) {
                // Add string start and end code points to the spanNotSet so that
                // a span(while not contained) stops before any string.
                UChar32 c;
                if(which&FWD) {
                    int32_t len=0;
                    U16_NEXT(s16, len, length16, c);
                    addToSpanNotSet(c);
                }
                if(which&BACK) {
                    int32_t len=length16;
                    U16_PREV(s16, 0, len, c);
                    addToSpanNotSet(c);
                }
            }
        } else {  // Irrelevant string.
            if(which&UTF8) {
                if(which&CONTAINED) {  // Only necessary for LONGEST_MATCH.
                    uint8_t *s8=utf8+utf8Count;
                    int32_t length8=appendUTF8(s16, length16, s8, utf8Length-utf8Count);
                    utf8Count+=utf8Lengths[i]=length8;
                } else {
                    utf8Lengths[i]=0;
                }
            }
            if(all) {
                spanLengths[i]=spanBackLengths[i]=
                    spanUTF8Lengths[i]=spanBackUTF8Lengths[i]=
                        (uint8_t)ALL_CP_CONTAINED;
            } else {
                // All spanXYZLengths pointers contain the same address.
                spanLengths[i]=(uint8_t)ALL_CP_CONTAINED;
            }
        }
    }

    // Finish.
    if(all) {
        pSpanNotSet->freeze();
    }
}

// Copy constructor. Assumes which==ALL for a frozen set.
UnicodeSetStringSpan::UnicodeSetStringSpan(const UnicodeSetStringSpan &otherStringSpan,
                                           const UVector &newParentSetStrings)
        : spanSet(otherStringSpan.spanSet), pSpanNotSet(NULL), strings(newParentSetStrings),
          utf8Lengths(NULL), spanLengths(NULL), utf8(NULL),
          utf8Length(otherStringSpan.utf8Length),
          maxLength16(otherStringSpan.maxLength16), maxLength8(otherStringSpan.maxLength8),
          all(TRUE) {
    if(otherStringSpan.pSpanNotSet==&otherStringSpan.spanSet) {
        pSpanNotSet=&spanSet;
    } else {
        pSpanNotSet=otherStringSpan.pSpanNotSet->clone();
    }

    // Allocate a block of meta data.
    // UTF-8 lengths, 4 sets of span lengths, UTF-8 strings.
    int32_t stringsLength=strings.size();
    int32_t allocSize=stringsLength*(4+1+1+1+1)+utf8Length;
    if(allocSize<=(int32_t)sizeof(staticLengths)) {
        utf8Lengths=staticLengths;
    } else {
        utf8Lengths=(int32_t *)uprv_malloc(allocSize);
        if(utf8Lengths==NULL) {
            maxLength16=maxLength8=0;  // Prevent usage by making needsStringSpanUTF16/8() return FALSE.
            return;  // Out of memory.
        }
    }

    spanLengths=(uint8_t *)(utf8Lengths+stringsLength);
    utf8=spanLengths+stringsLength*4;
    uprv_memcpy(utf8Lengths, otherStringSpan.utf8Lengths, allocSize);
}

UnicodeSetStringSpan::~UnicodeSetStringSpan() {
    if(pSpanNotSet!=NULL && pSpanNotSet!=&spanSet) {
        delete pSpanNotSet;
    }
    if(utf8Lengths!=NULL && utf8Lengths!=staticLengths) {
        uprv_free(utf8Lengths);
    }
}

void UnicodeSetStringSpan::addToSpanNotSet(UChar32 c) {
    if(pSpanNotSet==NULL || pSpanNotSet==&spanSet) {
        if(spanSet.contains(c)) {
            return;  // Nothing to do.
        }
        UnicodeSet *newSet=spanSet.cloneAsThawed();
        if(newSet==NULL) {
            return;  // Out of memory.
        } else {
            pSpanNotSet=newSet;
        }
    }
    pSpanNotSet->add(c);
}

// Compare strings without any argument checks. Requires length>0.
static inline UBool
matches16(const UChar *s, const UChar *t, int32_t length) {
    do {
        if(*s++!=*t++) {
            return FALSE;
        }
    } while(--length>0);
    return TRUE;
}

static inline UBool
matches8(const uint8_t *s, const uint8_t *t, int32_t length) {
    do {
        if(*s++!=*t++) {
            return FALSE;
        }
    } while(--length>0);
    return TRUE;
}

// Compare 16-bit Unicode strings (which may be malformed UTF-16)
// at code point boundaries.
// That is, each edge of a match must not be in the middle of a surrogate pair.
static inline UBool
matches16CPB(const UChar *s, int32_t start, int32_t limit, const UChar *t, int32_t length) {
    s+=start;
    limit-=start;
    return matches16(s, t, length) &&
           !(0<start && U16_IS_LEAD(s[-1]) && U16_IS_TRAIL(s[0])) &&
           !(length<limit && U16_IS_LEAD(s[length-1]) && U16_IS_TRAIL(s[length]));
}

// Does the set contain the next code point?
// If so, return its length; otherwise return its negative length.
static inline int32_t
spanOne(const UnicodeSet &set, const UChar *s, int32_t length) {
    UChar c=*s, c2;
    if(c>=0xd800 && c<=0xdbff && length>=2 && U16_IS_TRAIL(c2=s[1])) {
        return set.contains(U16_GET_SUPPLEMENTARY(c, c2)) ? 2 : -2;
    }
    return set.contains(c) ? 1 : -1;
}

static inline int32_t
spanOneBack(const UnicodeSet &set, const UChar *s, int32_t length) {
    UChar c=s[length-1], c2;
    if(c>=0xdc00 && c<=0xdfff && length>=2 && U16_IS_LEAD(c2=s[length-2])) {
        return set.contains(U16_GET_SUPPLEMENTARY(c2, c)) ? 2 : -2;
    }
    return set.contains(c) ? 1 : -1;
}

static inline int32_t
spanOneUTF8(const UnicodeSet &set, const uint8_t *s, int32_t length) {
    UChar32 c=*s;
    if(U8_IS_SINGLE(c)) {
        return set.contains(c) ? 1 : -1;
    }
    // Take advantage of non-ASCII fastpaths in U8_NEXT_OR_FFFD().
    int32_t i=0;
    U8_NEXT_OR_FFFD(s, i, length, c);
    return set.contains(c) ? i : -i;
}

static inline int32_t
spanOneBackUTF8(const UnicodeSet &set, const uint8_t *s, int32_t length) {
    UChar32 c=s[length-1];
    if(U8_IS_SINGLE(c)) {
        return set.contains(c) ? 1 : -1;
    }
    int32_t i=length-1;
    c=utf8_prevCharSafeBody(s, 0, &i, c, -3);
    length-=i;
    return set.contains(c) ? length : -length;
}

/*
 * Note: In span() when spanLength==0 (after a string match, or at the beginning
 * after an empty code point span) and in spanNot() and spanNotUTF8(),
 * string matching could use a binary search
 * because all string matches are done from the same start index.
 *
 * For UTF-8, this would require a comparison function that returns UTF-16 order.
 *
 * This optimization should not be necessary for normal UnicodeSets because
 * most sets have no strings, and most sets with strings have
 * very few very short strings.
 * For cases with many strings, it might be better to use a different API
 * and implementation with a DFA (state machine).
 */

/*
 * Algorithm for span(USET_SPAN_CONTAINED)
 *
 * Theoretical algorithm:
 * - Iterate through the string, and at each code point boundary:
 *   + If the code point there is in the set, then remember to continue after it.
 *   + If a set string matches at the current position, then remember to continue after it.
 *   + Either recursively span for each code point or string match,
 *     or recursively span for all but the shortest one and
 *     iteratively continue the span with the shortest local match.
 *   + Remember the longest recursive span (the farthest end point).
 *   + If there is no match at the current position, neither for the code point there
 *     nor for any set string, then stop and return the longest recursive span length.
 *
 * Optimized implementation:
 *
 * (We assume that most sets will have very few very short strings.
 * A span using a string-less set is extremely fast.)
 *
 * Create and cache a spanSet which contains all of the single code points
 * of the original set but none of its strings.
 *
 * - Start with spanLength=spanSet.span(USET_SPAN_CONTAINED).
 * - Loop:
 *   + Try to match each set string at the end of the spanLength.
 *     ~ Set strings that start with set-contained code points must be matched
 *       with a partial overlap because the recursive algorithm would have tried
 *       to match them at every position.
 *     ~ Set strings that entirely consist of set-contained code points
 *       are irrelevant for span(USET_SPAN_CONTAINED) because the
 *       recursive algorithm would continue after them anyway
 *       and find the longest recursive match from their end.
 *     ~ Rather than recursing, note each end point of a set string match.
 *   + If no set string matched after spanSet.span(), then return
 *     with where the spanSet.span() ended.
 *   + If at least one set string matched after spanSet.span(), then
 *     pop the shortest string match end point and continue
 *     the loop, trying to match all set strings from there.
 *   + If at least one more set string matched after a previous string match,
 *     then test if the code point after the previous string match is also
 *     contained in the set.
 *     Continue the loop with the shortest end point of either this code point
 *     or a matching set string.
 *   + If no more set string matched after a previous string match,
 *     then try another spanLength=spanSet.span(USET_SPAN_CONTAINED).
 *     Stop if spanLength==0, otherwise continue the loop.
 *
 * By noting each end point of a set string match,
 * the function visits each string position at most once and finishes
 * in linear time.
 *
 * The recursive algorithm may visit the same string position many times
 * if multiple paths lead to it and finishes in exponential time.
 */

/*
 * Algorithm for span(USET_SPAN_SIMPLE)
 *
 * Theoretical algorithm:
 * - Iterate through the string, and at each code point boundary:
 *   + If the code point there is in the set, then remember to continue after it.
 *   + If a set string matches at the current position, then remember to continue after it.
 *   + Continue from the farthest match position and ignore all others.
 *   + If there is no match at the current position,
 *     then stop and return the current position.
 *
 * Optimized implementation:
 *
 * (Same assumption and spanSet as above.)
 *
 * - Start with spanLength=spanSet.span(USET_SPAN_CONTAINED).
 * - Loop:
 *   + Try to match each set string at the end of the spanLength.
 *     ~ Set strings that start with set-contained code points must be matched
 *       with a partial overlap because the standard algorithm would have tried
 *       to match them earlier.
 *     ~ Set strings that entirely consist of set-contained code points
 *       must be matched with a full overlap because the longest-match algorithm
 *       would hide set string matches that end earlier.
 *       Such set strings need not be matched earlier inside the code point span
 *       because the standard algorithm would then have continued after
 *       the set string match anyway.
 *     ~ Remember the longest set string match (farthest end point) from the earliest
 *       starting point.
 *   + If no set string matched after spanSet.span(), then return
 *     with where the spanSet.span() ended.
 *   + If at least one set string matched, then continue the loop after the
 *     longest match from the earliest position.
 *   + If no more set string matched after a previous string match,
 *     then try another spanLength=spanSet.span(USET_SPAN_CONTAINED).
 *     Stop if spanLength==0, otherwise continue the loop.
 */

int32_t UnicodeSetStringSpan::span(const UChar *s, int32_t length, USetSpanCondition spanCondition) const {
    if(spanCondition==USET_SPAN_NOT_CONTAINED) {
        return spanNot(s, length);
    }
    int32_t spanLength=spanSet.span(s, length, USET_SPAN_CONTAINED);
    if(spanLength==length) {
        return length;
    }

    // Consider strings; they may overlap with the span.
    OffsetList offsets;
    if(spanCondition==USET_SPAN_CONTAINED) {
        // Use offset list to try all possibilities.
        offsets.setMaxLength(maxLength16);
    }
    int32_t pos=spanLength, rest=length-pos;
    int32_t i, stringsLength=strings.size();
    for(;;) {
        if(spanCondition==USET_SPAN_CONTAINED) {
            for(i=0; i<stringsLength; ++i) {
                int32_t overlap=spanLengths[i];
                if(overlap==ALL_CP_CONTAINED) {
                    continue;  // Irrelevant string.
                }
                const UnicodeString &string=*(const UnicodeString *)strings.elementAt(i);
                const UChar *s16=string.getBuffer();
                int32_t length16=string.length();

                // Try to match this string at pos-overlap..pos.
                if(overlap>=LONG_SPAN) {
                    overlap=length16;
                    // While contained: No point matching fully inside the code point span.
                    U16_BACK_1(s16, 0, overlap);  // Length of the string minus the last code point.
                }
                if(overlap>spanLength) {
                    overlap=spanLength;
                }
                int32_t inc=length16-overlap;  // Keep overlap+inc==length16.
                for(;;) {
                    if(inc>rest) {
                        break;
                    }
                    // Try to match if the increment is not listed already.
                    if(!offsets.containsOffset(inc) && matches16CPB(s, pos-overlap, length, s16, length16)) {
                        if(inc==rest) {
                            return length;  // Reached the end of the string.
                        }
                        offsets.addOffset(inc);
                    }
                    if(overlap==0) {
                        break;
                    }
                    --overlap;
                    ++inc;
                }
            }
        } else /* USET_SPAN_SIMPLE */ {
            int32_t maxInc=0, maxOverlap=0;
            for(i=0; i<stringsLength; ++i) {
                int32_t overlap=spanLengths[i];
                // For longest match, we do need to try to match even an all-contained string
                // to find the match from the earliest start.

                const UnicodeString &string=*(const UnicodeString *)strings.elementAt(i);
                const UChar *s16=string.getBuffer();
                int32_t length16=string.length();

                // Try to match this string at pos-overlap..pos.
                if(overlap>=LONG_SPAN) {
                    overlap=length16;
                    // Longest match: Need to match fully inside the code point span
                    // to find the match from the earliest start.
                }
                if(overlap>spanLength) {
                    overlap=spanLength;
                }
                int32_t inc=length16-overlap;  // Keep overlap+inc==length16.
                for(;;) {
                    if(inc>rest || overlap<maxOverlap) {
                        break;
                    }
                    // Try to match if the string is longer or starts earlier.
                    if( (overlap>maxOverlap || /* redundant overlap==maxOverlap && */ inc>maxInc) &&
                        matches16CPB(s, pos-overlap, length, s16, length16)
                    ) {
                        maxInc=inc;  // Longest match from earliest start.
                        maxOverlap=overlap;
                        break;
                    }
                    --overlap;
                    ++inc;
                }
            }

            if(maxInc!=0 || maxOverlap!=0) {
                // Longest-match algorithm, and there was a string match.
                // Simply continue after it.
                pos+=maxInc;
                rest-=maxInc;
                if(rest==0) {
                    return length;  // Reached the end of the string.
                }
                spanLength=0;  // Match strings from after a string match.
                continue;
            }
        }
        // Finished trying to match all strings at pos.

        if(spanLength!=0 || pos==0) {
            // The position is after an unlimited code point span (spanLength!=0),
            // not after a string match.
            // The only position where spanLength==0 after a span is pos==0.
            // Otherwise, an unlimited code point span is only tried again when no
            // strings match, and if such a non-initial span fails we stop.
            if(offsets.isEmpty()) {
                return pos;  // No strings matched after a span.
            }
            // Match strings from after the next string match.
        } else {
            // The position is after a string match (or a single code point).
            if(offsets.isEmpty()) {
                // No more strings matched after a previous string match.
                // Try another code point span from after the last string match.
                spanLength=spanSet.span(s+pos, rest, USET_SPAN_CONTAINED);
                if( spanLength==rest || // Reached the end of the string, or
                    spanLength==0       // neither strings nor span progressed.
                ) {
                    return pos+spanLength;
                }
                pos+=spanLength;
                rest-=spanLength;
                continue;  // spanLength>0: Match strings from after a span.
            } else {
                // Try to match only one code point from after a string match if some
                // string matched beyond it, so that we try all possible positions
                // and don't overshoot.
                spanLength=spanOne(spanSet, s+pos, rest);
                if(spanLength>0) {
                    if(spanLength==rest) {
                        return length;  // Reached the end of the string.
                    }
                    // Match strings after this code point.
                    // There cannot be any increments below it because UnicodeSet strings
                    // contain multiple code points.
                    pos+=spanLength;
                    rest-=spanLength;
                    offsets.shift(spanLength);
                    spanLength=0;
                    continue;  // Match strings from after a single code point.
                }
                // Match strings from after the next string match.
            }
        }
        int32_t minOffset=offsets.popMinimum();
        pos+=minOffset;
        rest-=minOffset;
        spanLength=0;  // Match strings from after a string match.
    }
}

int32_t UnicodeSetStringSpan::spanBack(const UChar *s, int32_t length, USetSpanCondition spanCondition) const {
    if(spanCondition==USET_SPAN_NOT_CONTAINED) {
        return spanNotBack(s, length);
    }
    int32_t pos=spanSet.spanBack(s, length, USET_SPAN_CONTAINED);
    if(pos==0) {
        return 0;
    }
    int32_t spanLength=length-pos;

    // Consider strings; they may overlap with the span.
    OffsetList offsets;
    if(spanCondition==USET_SPAN_CONTAINED) {
        // Use offset list to try all possibilities.
        offsets.setMaxLength(maxLength16);
    }
    int32_t i, stringsLength=strings.size();
    uint8_t *spanBackLengths=spanLengths;
    if(all) {
        spanBackLengths+=stringsLength;
    }
    for(;;) {
        if(spanCondition==USET_SPAN_CONTAINED) {
            for(i=0; i<stringsLength; ++i) {
                int32_t overlap=spanBackLengths[i];
                if(overlap==ALL_CP_CONTAINED) {
                    continue;  // Irrelevant string.
                }
                const UnicodeString &string=*(const UnicodeString *)strings.elementAt(i);
                const UChar *s16=string.getBuffer();
                int32_t length16=string.length();

                // Try to match this string at pos-(length16-overlap)..pos-length16.
                if(overlap>=LONG_SPAN) {
                    overlap=length16;
                    // While contained: No point matching fully inside the code point span.
                    int32_t len1=0;
                    U16_FWD_1(s16, len1, overlap);
                    overlap-=len1;  // Length of the string minus the first code point.
                }
                if(overlap>spanLength) {
                    overlap=spanLength;
                }
                int32_t dec=length16-overlap;  // Keep dec+overlap==length16.
                for(;;) {
                    if(dec>pos) {
                        break;
                    }
                    // Try to match if the decrement is not listed already.
                    if(!offsets.containsOffset(dec) && matches16CPB(s, pos-dec, length, s16, length16)) {
                        if(dec==pos) {
                            return 0;  // Reached the start of the string.
                        }
                        offsets.addOffset(dec);
                    }
                    if(overlap==0) {
                        break;
                    }
                    --overlap;
                    ++dec;
                }
            }
        } else /* USET_SPAN_SIMPLE */ {
            int32_t maxDec=0, maxOverlap=0;
            for(i=0; i<stringsLength; ++i) {
                int32_t overlap=spanBackLengths[i];
                // For longest match, we do need to try to match even an all-contained string
                // to find the match from the latest end.

                const UnicodeString &string=*(const UnicodeString *)strings.elementAt(i);
                const UChar *s16=string.getBuffer();
                int32_t length16=string.length();

                // Try to match this string at pos-(length16-overlap)..pos-length16.
                if(overlap>=LONG_SPAN) {
                    overlap=length16;
                    // Longest match: Need to match fully inside the code point span
                    // to find the match from the latest end.
                }
                if(overlap>spanLength) {
                    overlap=spanLength;
                }
                int32_t dec=length16-overlap;  // Keep dec+overlap==length16.
                for(;;) {
                    if(dec>pos || overlap<maxOverlap) {
                        break;
                    }
                    // Try to match if the string is longer or ends later.
                    if( (overlap>maxOverlap || /* redundant overlap==maxOverlap && */ dec>maxDec) &&
                        matches16CPB(s, pos-dec, length, s16, length16)
                    ) {
                        maxDec=dec;  // Longest match from latest end.
                        maxOverlap=overlap;
                        break;
                    }
                    --overlap;
                    ++dec;
                }
            }

            if(maxDec!=0 || maxOverlap!=0) {
                // Longest-match algorithm, and there was a string match.
                // Simply continue before it.
                pos-=maxDec;
                if(pos==0) {
                    return 0;  // Reached the start of the string.
                }
                spanLength=0;  // Match strings from before a string match.
                continue;
            }
        }
        // Finished trying to match all strings at pos.

        if(spanLength!=0 || pos==length) {
            // The position is before an unlimited code point span (spanLength!=0),
            // not before a string match.
            // The only position where spanLength==0 before a span is pos==length.
            // Otherwise, an unlimited code point span is only tried again when no
            // strings match, and if such a non-initial span fails we stop.
            if(offsets.isEmpty()) {
                return pos;  // No strings matched before a span.
            }
            // Match strings from before the next string match.
        } else {
            // The position is before a string match (or a single code point).
            if(offsets.isEmpty()) {
                // No more strings matched before a previous string match.
                // Try another code point span from before the last string match.
                int32_t oldPos=pos;
                pos=spanSet.spanBack(s, oldPos, USET_SPAN_CONTAINED);
                spanLength=oldPos-pos;
                if( pos==0 ||           // Reached the start of the string, or
                    spanLength==0       // neither strings nor span progressed.
                ) {
                    return pos;
                }
                continue;  // spanLength>0: Match strings from before a span.
            } else {
                // Try to match only one code point from before a string match if some
                // string matched beyond it, so that we try all possible positions
                // and don't overshoot.
                spanLength=spanOneBack(spanSet, s, pos);
                if(spanLength>0) {
                    if(spanLength==pos) {
                        return 0;  // Reached the start of the string.
                    }
                    // Match strings before this code point.
                    // There cannot be any decrements below it because UnicodeSet strings
                    // contain multiple code points.
                    pos-=spanLength;
                    offsets.shift(spanLength);
                    spanLength=0;
                    continue;  // Match strings from before a single code point.
                }
                // Match strings from before the next string match.
            }
        }
        pos-=offsets.popMinimum();
        spanLength=0;  // Match strings from before a string match.
    }
}

int32_t UnicodeSetStringSpan::spanUTF8(const uint8_t *s, int32_t length, USetSpanCondition spanCondition) const {
    if(spanCondition==USET_SPAN_NOT_CONTAINED) {
        return spanNotUTF8(s, length);
    }
    int32_t spanLength=spanSet.spanUTF8((const char *)s, length, USET_SPAN_CONTAINED);
    if(spanLength==length) {
        return length;
    }

    // Consider strings; they may overlap with the span.
    OffsetList offsets;
    if(spanCondition==USET_SPAN_CONTAINED) {
        // Use offset list to try all possibilities.
        offsets.setMaxLength(maxLength8);
    }
    int32_t pos=spanLength, rest=length-pos;
    int32_t i, stringsLength=strings.size();
    uint8_t *spanUTF8Lengths=spanLengths;
    if(all) {
        spanUTF8Lengths+=2*stringsLength;
    }
    for(;;) {
        const uint8_t *s8=utf8;
        int32_t length8;
        if(spanCondition==USET_SPAN_CONTAINED) {
            for(i=0; i<stringsLength; ++i) {
                length8=utf8Lengths[i];
                if(length8==0) {
                    continue;  // String not representable in UTF-8.
                }
                int32_t overlap=spanUTF8Lengths[i];
                if(overlap==ALL_CP_CONTAINED) {
                    s8+=length8;
                    continue;  // Irrelevant string.
                }

                // Try to match this string at pos-overlap..pos.
                if(overlap>=LONG_SPAN) {
                    overlap=length8;
                    // While contained: No point matching fully inside the code point span.
                    U8_BACK_1(s8, 0, overlap);  // Length of the string minus the last code point.
                }
                if(overlap>spanLength) {
                    overlap=spanLength;
                }
                int32_t inc=length8-overlap;  // Keep overlap+inc==length8.
                for(;;) {
                    if(inc>rest) {
                        break;
                    }
                    // Try to match if the increment is not listed already.
                    // Match at code point boundaries. (The UTF-8 strings were converted
                    // from UTF-16 and are guaranteed to be well-formed.)
                    if(!U8_IS_TRAIL(s[pos-overlap]) &&
                            !offsets.containsOffset(inc) &&
                            matches8(s+pos-overlap, s8, length8)) {
                        if(inc==rest) {
                            return length;  // Reached the end of the string.
                        }
                        offsets.addOffset(inc);
                    }
                    if(overlap==0) {
                        break;
                    }
                    --overlap;
                    ++inc;
                }
                s8+=length8;
            }
        } else /* USET_SPAN_SIMPLE */ {
            int32_t maxInc=0, maxOverlap=0;
            for(i=0; i<stringsLength; ++i) {
                length8=utf8Lengths[i];
                if(length8==0) {
                    continue;  // String not representable in UTF-8.
                }
                int32_t overlap=spanUTF8Lengths[i];
                // For longest match, we do need to try to match even an all-contained string
                // to find the match from the earliest start.

                // Try to match this string at pos-overlap..pos.
                if(overlap>=LONG_SPAN) {
                    overlap=length8;
                    // Longest match: Need to match fully inside the code point span
                    // to find the match from the earliest start.
                }
                if(overlap>spanLength) {
                    overlap=spanLength;
                }
                int32_t inc=length8-overlap;  // Keep overlap+inc==length8.
                for(;;) {
                    if(inc>rest || overlap<maxOverlap) {
                        break;
                    }
                    // Try to match if the string is longer or starts earlier.
                    // Match at code point boundaries. (The UTF-8 strings were converted
                    // from UTF-16 and are guaranteed to be well-formed.)
                    if(!U8_IS_TRAIL(s[pos-overlap]) &&
                            (overlap>maxOverlap ||
                                /* redundant overlap==maxOverlap && */ inc>maxInc) &&
                            matches8(s+pos-overlap, s8, length8)) {
                        maxInc=inc;  // Longest match from earliest start.
                        maxOverlap=overlap;
                        break;
                    }
                    --overlap;
                    ++inc;
                }
                s8+=length8;
            }

            if(maxInc!=0 || maxOverlap!=0) {
                // Longest-match algorithm, and there was a string match.
                // Simply continue after it.
                pos+=maxInc;
                rest-=maxInc;
                if(rest==0) {
                    return length;  // Reached the end of the string.
                }
                spanLength=0;  // Match strings from after a string match.
                continue;
            }
        }
        // Finished trying to match all strings at pos.

        if(spanLength!=0 || pos==0) {
            // The position is after an unlimited code point span (spanLength!=0),
            // not after a string match.
            // The only position where spanLength==0 after a span is pos==0.
            // Otherwise, an unlimited code point span is only tried again when no
            // strings match, and if such a non-initial span fails we stop.
            if(offsets.isEmpty()) {
                return pos;  // No strings matched after a span.
            }
            // Match strings from after the next string match.
        } else {
            // The position is after a string match (or a single code point).
            if(offsets.isEmpty()) {
                // No more strings matched after a previous string match.
                // Try another code point span from after the last string match.
                spanLength=spanSet.spanUTF8((const char *)s+pos, rest, USET_SPAN_CONTAINED);
                if( spanLength==rest || // Reached the end of the string, or
                    spanLength==0       // neither strings nor span progressed.
                ) {
                    return pos+spanLength;
                }
                pos+=spanLength;
                rest-=spanLength;
                continue;  // spanLength>0: Match strings from after a span.
            } else {
                // Try to match only one code point from after a string match if some
                // string matched beyond it, so that we try all possible positions
                // and don't overshoot.
                spanLength=spanOneUTF8(spanSet, s+pos, rest);
                if(spanLength>0) {
                    if(spanLength==rest) {
                        return length;  // Reached the end of the string.
                    }
                    // Match strings after this code point.
                    // There cannot be any increments below it because UnicodeSet strings
                    // contain multiple code points.
                    pos+=spanLength;
                    rest-=spanLength;
                    offsets.shift(spanLength);
                    spanLength=0;
                    continue;  // Match strings from after a single code point.
                }
                // Match strings from after the next string match.
            }
        }
        int32_t minOffset=offsets.popMinimum();
        pos+=minOffset;
        rest-=minOffset;
        spanLength=0;  // Match strings from after a string match.
    }
}

int32_t UnicodeSetStringSpan::spanBackUTF8(const uint8_t *s, int32_t length, USetSpanCondition spanCondition) const {
    if(spanCondition==USET_SPAN_NOT_CONTAINED) {
        return spanNotBackUTF8(s, length);
    }
    int32_t pos=spanSet.spanBackUTF8((const char *)s, length, USET_SPAN_CONTAINED);
    if(pos==0) {
        return 0;
    }
    int32_t spanLength=length-pos;

    // Consider strings; they may overlap with the span.
    OffsetList offsets;
    if(spanCondition==USET_SPAN_CONTAINED) {
        // Use offset list to try all possibilities.
        offsets.setMaxLength(maxLength8);
    }
    int32_t i, stringsLength=strings.size();
    uint8_t *spanBackUTF8Lengths=spanLengths;
    if(all) {
        spanBackUTF8Lengths+=3*stringsLength;
    }
    for(;;) {
        const uint8_t *s8=utf8;
        int32_t length8;
        if(spanCondition==USET_SPAN_CONTAINED) {
            for(i=0; i<stringsLength; ++i) {
                length8=utf8Lengths[i];
                if(length8==0) {
                    continue;  // String not representable in UTF-8.
                }
                int32_t overlap=spanBackUTF8Lengths[i];
                if(overlap==ALL_CP_CONTAINED) {
                    s8+=length8;
                    continue;  // Irrelevant string.
                }

                // Try to match this string at pos-(length8-overlap)..pos-length8.
                if(overlap>=LONG_SPAN) {
                    overlap=length8;
                    // While contained: No point matching fully inside the code point span.
                    int32_t len1=0;
                    U8_FWD_1(s8, len1, overlap);
                    overlap-=len1;  // Length of the string minus the first code point.
                }
                if(overlap>spanLength) {
                    overlap=spanLength;
                }
                int32_t dec=length8-overlap;  // Keep dec+overlap==length8.
                for(;;) {
                    if(dec>pos) {
                        break;
                    }
                    // Try to match if the decrement is not listed already.
                    // Match at code point boundaries. (The UTF-8 strings were converted
                    // from UTF-16 and are guaranteed to be well-formed.)
                    if( !U8_IS_TRAIL(s[pos-dec]) &&
                        !offsets.containsOffset(dec) &&
                        matches8(s+pos-dec, s8, length8)
                    ) {
                        if(dec==pos) {
                            return 0;  // Reached the start of the string.
                        }
                        offsets.addOffset(dec);
                    }
                    if(overlap==0) {
                        break;
                    }
                    --overlap;
                    ++dec;
                }
                s8+=length8;
            }
        } else /* USET_SPAN_SIMPLE */ {
            int32_t maxDec=0, maxOverlap=0;
            for(i=0; i<stringsLength; ++i) {
                length8=utf8Lengths[i];
                if(length8==0) {
                    continue;  // String not representable in UTF-8.
                }
                int32_t overlap=spanBackUTF8Lengths[i];
                // For longest match, we do need to try to match even an all-contained string
                // to find the match from the latest end.

                // Try to match this string at pos-(length8-overlap)..pos-length8.
                if(overlap>=LONG_SPAN) {
                    overlap=length8;
                    // Longest match: Need to match fully inside the code point span
                    // to find the match from the latest end.
                }
                if(overlap>spanLength) {
                    overlap=spanLength;
                }
                int32_t dec=length8-overlap;  // Keep dec+overlap==length8.
                for(;;) {
                    if(dec>pos || overlap<maxOverlap) {
                        break;
                    }
                    // Try to match if the string is longer or ends later.
                    // Match at code point boundaries. (The UTF-8 strings were converted
                    // from UTF-16 and are guaranteed to be well-formed.)
                    if( !U8_IS_TRAIL(s[pos-dec]) &&
                        (overlap>maxOverlap || /* redundant overlap==maxOverlap && */ dec>maxDec) &&
                        matches8(s+pos-dec, s8, length8)
                    ) {
                        maxDec=dec;  // Longest match from latest end.
                        maxOverlap=overlap;
                        break;
                    }
                    --overlap;
                    ++dec;
                }
                s8+=length8;
            }

            if(maxDec!=0 || maxOverlap!=0) {
                // Longest-match algorithm, and there was a string match.
                // Simply continue before it.
                pos-=maxDec;
                if(pos==0) {
                    return 0;  // Reached the start of the string.
                }
                spanLength=0;  // Match strings from before a string match.
                continue;
            }
        }
        // Finished trying to match all strings at pos.

        if(spanLength!=0 || pos==length) {
            // The position is before an unlimited code point span (spanLength!=0),
            // not before a string match.
            // The only position where spanLength==0 before a span is pos==length.
            // Otherwise, an unlimited code point span is only tried again when no
            // strings match, and if such a non-initial span fails we stop.
            if(offsets.isEmpty()) {
                return pos;  // No strings matched before a span.
            }
            // Match strings from before the next string match.
        } else {
            // The position is before a string match (or a single code point).
            if(offsets.isEmpty()) {
                // No more strings matched before a previous string match.
                // Try another code point span from before the last string match.
                int32_t oldPos=pos;
                pos=spanSet.spanBackUTF8((const char *)s, oldPos, USET_SPAN_CONTAINED);
                spanLength=oldPos-pos;
                if( pos==0 ||           // Reached the start of the string, or
                    spanLength==0       // neither strings nor span progressed.
                ) {
                    return pos;
                }
                continue;  // spanLength>0: Match strings from before a span.
            } else {
                // Try to match only one code point from before a string match if some
                // string matched beyond it, so that we try all possible positions
                // and don't overshoot.
                spanLength=spanOneBackUTF8(spanSet, s, pos);
                if(spanLength>0) {
                    if(spanLength==pos) {
                        return 0;  // Reached the start of the string.
                    }
                    // Match strings before this code point.
                    // There cannot be any decrements below it because UnicodeSet strings
                    // contain multiple code points.
                    pos-=spanLength;
                    offsets.shift(spanLength);
                    spanLength=0;
                    continue;  // Match strings from before a single code point.
                }
                // Match strings from before the next string match.
            }
        }
        pos-=offsets.popMinimum();
        spanLength=0;  // Match strings from before a string match.
    }
}

/*
 * Algorithm for spanNot()==span(USET_SPAN_NOT_CONTAINED)
 *
 * Theoretical algorithm:
 * - Iterate through the string, and at each code point boundary:
 *   + If the code point there is in the set, then return with the current position.
 *   + If a set string matches at the current position, then return with the current position.
 *
 * Optimized implementation:
 *
 * (Same assumption as for span() above.)
 *
 * Create and cache a spanNotSet which contains all of the single code points
 * of the original set but none of its strings.
 * For each set string add its initial code point to the spanNotSet.
 * (Also add its final code point for spanNotBack().)
 *
 * - Loop:
 *   + Do spanLength=spanNotSet.span(USET_SPAN_NOT_CONTAINED).
 *   + If the current code point is in the original set, then
 *     return the current position.
 *   + If any set string matches at the current position, then
 *     return the current position.
 *   + If there is no match at the current position, neither for the code point there
 *     nor for any set string, then skip this code point and continue the loop.
 *     This happens for set-string-initial code points that were added to spanNotSet
 *     when there is not actually a match for such a set string.
 */

int32_t UnicodeSetStringSpan::spanNot(const UChar *s, int32_t length) const {
    int32_t pos=0, rest=length;
    int32_t i, stringsLength=strings.size();
    do {
        // Span until we find a code point from the set,
        // or a code point that starts or ends some string.
        i=pSpanNotSet->span(s+pos, rest, USET_SPAN_NOT_CONTAINED);
        if(i==rest) {
            return length;  // Reached the end of the string.
        }
        pos+=i;
        rest-=i;

        // Check whether the current code point is in the original set,
        // without the string starts and ends.
        int32_t cpLength=spanOne(spanSet, s+pos, rest);
        if(cpLength>0) {
            return pos;  // There is a set element at pos.
        }

        // Try to match the strings at pos.
        for(i=0; i<stringsLength; ++i) {
            if(spanLengths[i]==ALL_CP_CONTAINED) {
                continue;  // Irrelevant string.
            }
            const UnicodeString &string=*(const UnicodeString *)strings.elementAt(i);
            const UChar *s16=string.getBuffer();
            int32_t length16=string.length();
            if(length16<=rest && matches16CPB(s, pos, length, s16, length16)) {
                return pos;  // There is a set element at pos.
            }
        }

        // The span(while not contained) ended on a string start/end which is
        // not in the original set. Skip this code point and continue.
        // cpLength<0
        pos-=cpLength;
        rest+=cpLength;
    } while(rest!=0);
    return length;  // Reached the end of the string.
}

int32_t UnicodeSetStringSpan::spanNotBack(const UChar *s, int32_t length) const {
    int32_t pos=length;
    int32_t i, stringsLength=strings.size();
    do {
        // Span until we find a code point from the set,
        // or a code point that starts or ends some string.
        pos=pSpanNotSet->spanBack(s, pos, USET_SPAN_NOT_CONTAINED);
        if(pos==0) {
            return 0;  // Reached the start of the string.
        }

        // Check whether the current code point is in the original set,
        // without the string starts and ends.
        int32_t cpLength=spanOneBack(spanSet, s, pos);
        if(cpLength>0) {
            return pos;  // There is a set element at pos.
        }

        // Try to match the strings at pos.
        for(i=0; i<stringsLength; ++i) {
            // Use spanLengths rather than a spanBackLengths pointer because
            // it is easier and we only need to know whether the string is irrelevant
            // which is the same in either array.
            if(spanLengths[i]==ALL_CP_CONTAINED) {
                continue;  // Irrelevant string.
            }
            const UnicodeString &string=*(const UnicodeString *)strings.elementAt(i);
            const UChar *s16=string.getBuffer();
            int32_t length16=string.length();
            if(length16<=pos && matches16CPB(s, pos-length16, length, s16, length16)) {
                return pos;  // There is a set element at pos.
            }
        }

        // The span(while not contained) ended on a string start/end which is
        // not in the original set. Skip this code point and continue.
        // cpLength<0
        pos+=cpLength;
    } while(pos!=0);
    return 0;  // Reached the start of the string.
}

int32_t UnicodeSetStringSpan::spanNotUTF8(const uint8_t *s, int32_t length) const {
    int32_t pos=0, rest=length;
    int32_t i, stringsLength=strings.size();
    uint8_t *spanUTF8Lengths=spanLengths;
    if(all) {
        spanUTF8Lengths+=2*stringsLength;
    }
    do {
        // Span until we find a code point from the set,
        // or a code point that starts or ends some string.
        i=pSpanNotSet->spanUTF8((const char *)s+pos, rest, USET_SPAN_NOT_CONTAINED);
        if(i==rest) {
            return length;  // Reached the end of the string.
        }
        pos+=i;
        rest-=i;

        // Check whether the current code point is in the original set,
        // without the string starts and ends.
        int32_t cpLength=spanOneUTF8(spanSet, s+pos, rest);
        if(cpLength>0) {
            return pos;  // There is a set element at pos.
        }

        // Try to match the strings at pos.
        const uint8_t *s8=utf8;
        int32_t length8;
        for(i=0; i<stringsLength; ++i) {
            length8=utf8Lengths[i];
            // ALL_CP_CONTAINED: Irrelevant string.
            if(length8!=0 && spanUTF8Lengths[i]!=ALL_CP_CONTAINED && length8<=rest && matches8(s+pos, s8, length8)) {
                return pos;  // There is a set element at pos.
            }
            s8+=length8;
        }

        // The span(while not contained) ended on a string start/end which is
        // not in the original set. Skip this code point and continue.
        // cpLength<0
        pos-=cpLength;
        rest+=cpLength;
    } while(rest!=0);
    return length;  // Reached the end of the string.
}

int32_t UnicodeSetStringSpan::spanNotBackUTF8(const uint8_t *s, int32_t length) const {
    int32_t pos=length;
    int32_t i, stringsLength=strings.size();
    uint8_t *spanBackUTF8Lengths=spanLengths;
    if(all) {
        spanBackUTF8Lengths+=3*stringsLength;
    }
    do {
        // Span until we find a code point from the set,
        // or a code point that starts or ends some string.
        pos=pSpanNotSet->spanBackUTF8((const char *)s, pos, USET_SPAN_NOT_CONTAINED);
        if(pos==0) {
            return 0;  // Reached the start of the string.
        }

        // Check whether the current code point is in the original set,
        // without the string starts and ends.
        int32_t cpLength=spanOneBackUTF8(spanSet, s, pos);
        if(cpLength>0) {
            return pos;  // There is a set element at pos.
        }

        // Try to match the strings at pos.
        const uint8_t *s8=utf8;
        int32_t length8;
        for(i=0; i<stringsLength; ++i) {
            length8=utf8Lengths[i];
            // ALL_CP_CONTAINED: Irrelevant string.
            if(length8!=0 && spanBackUTF8Lengths[i]!=ALL_CP_CONTAINED && length8<=pos && matches8(s+pos-length8, s8, length8)) {
                return pos;  // There is a set element at pos.
            }
            s8+=length8;
        }

        // The span(while not contained) ended on a string start/end which is
        // not in the original set. Skip this code point and continue.
        // cpLength<0
        pos+=cpLength;
    } while(pos!=0);
    return 0;  // Reached the start of the string.
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 1999-2016, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*
* File unistr.cpp
*
* Modification History:
*
*   Date        Name        Description
*   09/25/98    stephen     Creation.
*   04/20/99    stephen     Overhauled per 4/16 code review.
*   07/09/99    stephen     Renamed {hi,lo},{byte,word} to icu_X for HP/UX
*   11/18/99    aliu        Added handleReplaceBetween() to make inherit from
*                           Replaceable.
*   06/25/01    grhoten     Removed the dependency on iostream
******************************************************************************
*/















#if 0

#include <iostream>
using namespace std;

//DEBUGGING
void
print(const UnicodeString& s,
      const char *name)
{
  UChar c;
  cout << name << ":|";
  for(int i = 0; i < s.length(); ++i) {
    c = s[i];
    if(c>= 0x007E || c < 0x0020)
      cout << "[0x" << hex << s[i] << "]";
    else
      cout << (char) s[i];
  }
  cout << '|' << endl;
}

void
print(const UChar *s,
      int32_t len,
      const char *name)
{
  UChar c;
  cout << name << ":|";
  for(int i = 0; i < len; ++i) {
    c = s[i];
    if(c>= 0x007E || c < 0x0020)
      cout << "[0x" << hex << s[i] << "]";
    else
      cout << (char) s[i];
  }
  cout << '|' << endl;
}
// END DEBUGGING
#endif

// Local function definitions for now

// need to copy areas that may overlap
static
inline void
us_arrayCopy(const UChar *src, int32_t srcStart,
         UChar *dst, int32_t dstStart, int32_t count)
{
  if(count>0) {
    uprv_memmove(dst+dstStart, src+srcStart, (size_t)count*sizeof(*src));
  }
}

// u_unescapeAt() callback to get a UChar from a UnicodeString
U_CDECL_BEGIN
static UChar U_CALLCONV
UnicodeString_charAt(int32_t offset, void *context) {
    return ((icu::UnicodeString*) context)->charAt(offset);
}
U_CDECL_END

U_NAMESPACE_BEGIN

/* The Replaceable virtual destructor can't be defined in the header
   due to how AIX works with multiple definitions of virtual functions.
*/
Replaceable::~Replaceable() {}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(UnicodeString)

UnicodeString U_EXPORT2
operator+ (const UnicodeString &s1, const UnicodeString &s2) {
    return
        UnicodeString(s1.length()+s2.length()+1, (UChar32)0, 0).
            append(s1).
                append(s2);
}

//========================================
// Reference Counting functions, put at top of file so that optimizing compilers
//                               have a chance to automatically inline.
//========================================

void
UnicodeString::addRef() {
  umtx_atomic_inc((u_atomic_int32_t *)fUnion.fFields.fArray - 1);
}

int32_t
UnicodeString::removeRef() {
  return umtx_atomic_dec((u_atomic_int32_t *)fUnion.fFields.fArray - 1);
}

int32_t
UnicodeString::refCount() const {
  return umtx_loadAcquire(*((u_atomic_int32_t *)fUnion.fFields.fArray - 1));
}

void
UnicodeString::releaseArray() {
  if((fUnion.fFields.fLengthAndFlags & kRefCounted) && removeRef() == 0) {
    uprv_free((int32_t *)fUnion.fFields.fArray - 1);
  }
}



//========================================
// Constructors
//========================================

// The default constructor is inline in unistr.h.

UnicodeString::UnicodeString(int32_t capacity, UChar32 c, int32_t count) {
  fUnion.fFields.fLengthAndFlags = 0;
  if(count <= 0 || (uint32_t)c > 0x10ffff) {
    // just allocate and do not do anything else
    allocate(capacity);
  } else if(c <= 0xffff) {
    int32_t length = count;
    if(capacity < length) {
      capacity = length;
    }
    if(allocate(capacity)) {
      UChar *array = getArrayStart();
      UChar unit = (UChar)c;
      for(int32_t i = 0; i < length; ++i) {
        array[i] = unit;
      }
      setLength(length);
    }
  } else {  // supplementary code point, write surrogate pairs
    if(count > (INT32_MAX / 2)) {
      // We would get more than 2G UChars.
      allocate(capacity);
      return;
    }
    int32_t length = count * 2;
    if(capacity < length) {
      capacity = length;
    }
    if(allocate(capacity)) {
      UChar *array = getArrayStart();
      UChar lead = U16_LEAD(c);
      UChar trail = U16_TRAIL(c);
      for(int32_t i = 0; i < length; i += 2) {
        array[i] = lead;
        array[i + 1] = trail;
      }
      setLength(length);
    }
  }
}

UnicodeString::UnicodeString(UChar ch) {
  fUnion.fFields.fLengthAndFlags = kLength1 | kShortString;
  fUnion.fStackFields.fBuffer[0] = ch;
}

UnicodeString::UnicodeString(UChar32 ch) {
  fUnion.fFields.fLengthAndFlags = kShortString;
  int32_t i = 0;
  UBool isError = FALSE;
  U16_APPEND(fUnion.fStackFields.fBuffer, i, US_STACKBUF_SIZE, ch, isError);
  // We test isError so that the compiler does not complain that we don't.
  // If isError then i==0 which is what we want anyway.
  if(!isError) {
    setShortLength(i);
  }
}

UnicodeString::UnicodeString(const UChar *text) {
  fUnion.fFields.fLengthAndFlags = kShortString;
  doAppend(text, 0, -1);
}

UnicodeString::UnicodeString(const UChar *text,
                             int32_t textLength) {
  fUnion.fFields.fLengthAndFlags = kShortString;
  doAppend(text, 0, textLength);
}

UnicodeString::UnicodeString(UBool isTerminated,
                             ConstChar16Ptr textPtr,
                             int32_t textLength) {
  fUnion.fFields.fLengthAndFlags = kReadonlyAlias;
  const UChar *text = textPtr;
  if(text == NULL) {
    // treat as an empty string, do not alias
    setToEmpty();
  } else if(textLength < -1 ||
            (textLength == -1 && !isTerminated) ||
            (textLength >= 0 && isTerminated && text[textLength] != 0)
  ) {
    setToBogus();
  } else {
    if(textLength == -1) {
      // text is terminated, or else it would have failed the above test
      textLength = u_strlen(text);
    }
    setArray(const_cast<UChar *>(text), textLength,
             isTerminated ? textLength + 1 : textLength);
  }
}

UnicodeString::UnicodeString(UChar *buff,
                             int32_t buffLength,
                             int32_t buffCapacity) {
  fUnion.fFields.fLengthAndFlags = kWritableAlias;
  if(buff == NULL) {
    // treat as an empty string, do not alias
    setToEmpty();
  } else if(buffLength < -1 || buffCapacity < 0 || buffLength > buffCapacity) {
    setToBogus();
  } else {
    if(buffLength == -1) {
      // fLength = u_strlen(buff); but do not look beyond buffCapacity
      const UChar *p = buff, *limit = buff + buffCapacity;
      while(p != limit && *p != 0) {
        ++p;
      }
      buffLength = (int32_t)(p - buff);
    }
    setArray(buff, buffLength, buffCapacity);
  }
}

UnicodeString::UnicodeString(const char *src, int32_t length, EInvariant) {
  fUnion.fFields.fLengthAndFlags = kShortString;
  if(src==NULL) {
    // treat as an empty string
  } else {
    if(length<0) {
      length=(int32_t)uprv_strlen(src);
    }
    if(cloneArrayIfNeeded(length, length, FALSE)) {
      u_charsToUChars(src, getArrayStart(), length);
      setLength(length);
    } else {
      setToBogus();
    }
  }
}

#if U_CHARSET_IS_UTF8

UnicodeString::UnicodeString(const char *codepageData) {
  fUnion.fFields.fLengthAndFlags = kShortString;
  if(codepageData != 0) {
    setToUTF8(codepageData);
  }
}

UnicodeString::UnicodeString(const char *codepageData, int32_t dataLength) {
  fUnion.fFields.fLengthAndFlags = kShortString;
  // if there's nothing to convert, do nothing
  if(codepageData == 0 || dataLength == 0 || dataLength < -1) {
    return;
  }
  if(dataLength == -1) {
    dataLength = (int32_t)uprv_strlen(codepageData);
  }
  setToUTF8(StringPiece(codepageData, dataLength));
}

// else see unistr_cnv.cpp
#endif

UnicodeString::UnicodeString(const UnicodeString& that) {
  fUnion.fFields.fLengthAndFlags = kShortString;
  copyFrom(that);
}

UnicodeString::UnicodeString(UnicodeString &&src) U_NOEXCEPT {
  copyFieldsFrom(src, TRUE);
}

UnicodeString::UnicodeString(const UnicodeString& that,
                             int32_t srcStart) {
  fUnion.fFields.fLengthAndFlags = kShortString;
  setTo(that, srcStart);
}

UnicodeString::UnicodeString(const UnicodeString& that,
                             int32_t srcStart,
                             int32_t srcLength) {
  fUnion.fFields.fLengthAndFlags = kShortString;
  setTo(that, srcStart, srcLength);
}

// Replaceable base class clone() default implementation, does not clone
Replaceable *
Replaceable::clone() const {
  return NULL;
}

// UnicodeString overrides clone() with a real implementation
UnicodeString *
UnicodeString::clone() const {
  return new UnicodeString(*this);
}

//========================================
// array allocation
//========================================

namespace {

const int32_t kGrowSize = 128;

// The number of bytes for one int32_t reference counter and capacity UChars
// must fit into a 32-bit size_t (at least when on a 32-bit platform).
// We also add one for the NUL terminator, to avoid reallocation in getTerminatedBuffer(),
// and round up to a multiple of 16 bytes.
// This means that capacity must be at most (0xfffffff0 - 4) / 2 - 1 = 0x7ffffff5.
// (With more complicated checks we could go up to 0x7ffffffd without rounding up,
// but that does not seem worth it.)
const int32_t kMaxCapacity = 0x7ffffff5;

int32_t getGrowCapacity(int32_t newLength) {
  int32_t growSize = (newLength >> 2) + kGrowSize;
  if(growSize <= (kMaxCapacity - newLength)) {
    return newLength + growSize;
  } else {
    return kMaxCapacity;
  }
}

}  // namespace

UBool
UnicodeString::allocate(int32_t capacity) {
  if(capacity <= US_STACKBUF_SIZE) {
    fUnion.fFields.fLengthAndFlags = kShortString;
    return TRUE;
  }
  if(capacity <= kMaxCapacity) {
    ++capacity;  // for the NUL
    // Switch to size_t which is unsigned so that we can allocate up to 4GB.
    // Reference counter + UChars.
    size_t numBytes = sizeof(int32_t) + (size_t)capacity * U_SIZEOF_UCHAR;
    // Round up to a multiple of 16.
    numBytes = (numBytes + 15) & ~15;
    int32_t *array = (int32_t *) uprv_malloc(numBytes);
    if(array != NULL) {
      // set initial refCount and point behind the refCount
      *array++ = 1;
      numBytes -= sizeof(int32_t);

      // have fArray point to the first UChar
      fUnion.fFields.fArray = (UChar *)array;
      fUnion.fFields.fCapacity = (int32_t)(numBytes / U_SIZEOF_UCHAR);
      fUnion.fFields.fLengthAndFlags = kLongString;
      return TRUE;
    }
  }
  fUnion.fFields.fLengthAndFlags = kIsBogus;
  fUnion.fFields.fArray = 0;
  fUnion.fFields.fCapacity = 0;
  return FALSE;
}

//========================================
// Destructor
//========================================

#ifdef UNISTR_COUNT_FINAL_STRING_LENGTHS
static u_atomic_int32_t finalLengthCounts[0x400];  // UnicodeString::kMaxShortLength+1
static u_atomic_int32_t beyondCount(0);

U_CAPI void unistr_printLengths() {
  int32_t i;
  for(i = 0; i <= 59; ++i) {
    printf("%2d,  %9d\n", i, (int32_t)finalLengthCounts[i]);
  }
  int32_t beyond = beyondCount;
  for(; i < UPRV_LENGTHOF(finalLengthCounts); ++i) {
    beyond += finalLengthCounts[i];
  }
  printf(">59, %9d\n", beyond);
}
#endif

UnicodeString::~UnicodeString()
{
#ifdef UNISTR_COUNT_FINAL_STRING_LENGTHS
  // Count lengths of strings at the end of their lifetime.
  // Useful for discussion of a desirable stack buffer size.
  // Count the contents length, not the optional NUL terminator nor further capacity.
  // Ignore open-buffer strings and strings which alias external storage.
  if((fUnion.fFields.fLengthAndFlags&(kOpenGetBuffer|kReadonlyAlias|kWritableAlias)) == 0) {
    if(hasShortLength()) {
      umtx_atomic_inc(finalLengthCounts + getShortLength());
    } else {
      umtx_atomic_inc(&beyondCount);
    }
  }
#endif

  releaseArray();
}

//========================================
// Factory methods
//========================================

UnicodeString UnicodeString::fromUTF8(StringPiece utf8) {
  UnicodeString result;
  result.setToUTF8(utf8);
  return result;
}

UnicodeString UnicodeString::fromUTF32(const UChar32 *utf32, int32_t length) {
  UnicodeString result;
  int32_t capacity;
  // Most UTF-32 strings will be BMP-only and result in a same-length
  // UTF-16 string. We overestimate the capacity just slightly,
  // just in case there are a few supplementary characters.
  if(length <= US_STACKBUF_SIZE) {
    capacity = US_STACKBUF_SIZE;
  } else {
    capacity = length + (length >> 4) + 4;
  }
  do {
    UChar *utf16 = result.getBuffer(capacity);
    int32_t length16;
    UErrorCode errorCode = U_ZERO_ERROR;
    u_strFromUTF32WithSub(utf16, result.getCapacity(), &length16,
        utf32, length,
        0xfffd,  // Substitution character.
        NULL,    // Don't care about number of substitutions.
        &errorCode);
    result.releaseBuffer(length16);
    if(errorCode == U_BUFFER_OVERFLOW_ERROR) {
      capacity = length16 + 1;  // +1 for the terminating NUL.
      continue;
    } else if(U_FAILURE(errorCode)) {
      result.setToBogus();
    }
    break;
  } while(TRUE);
  return result;
}

//========================================
// Assignment
//========================================

UnicodeString &
UnicodeString::operator=(const UnicodeString &src) {
  return copyFrom(src);
}

UnicodeString &
UnicodeString::fastCopyFrom(const UnicodeString &src) {
  return copyFrom(src, TRUE);
}

UnicodeString &
UnicodeString::copyFrom(const UnicodeString &src, UBool fastCopy) {
  // if assigning to ourselves, do nothing
  if(this == &src) {
    return *this;
  }

  // is the right side bogus?
  if(src.isBogus()) {
    setToBogus();
    return *this;
  }

  // delete the current contents
  releaseArray();

  if(src.isEmpty()) {
    // empty string - use the stack buffer
    setToEmpty();
    return *this;
  }

  // fLength>0 and not an "open" src.getBuffer(minCapacity)
  fUnion.fFields.fLengthAndFlags = src.fUnion.fFields.fLengthAndFlags;
  switch(src.fUnion.fFields.fLengthAndFlags & kAllStorageFlags) {
  case kShortString:
    // short string using the stack buffer, do the same
    uprv_memcpy(fUnion.fStackFields.fBuffer, src.fUnion.fStackFields.fBuffer,
                getShortLength() * U_SIZEOF_UCHAR);
    break;
  case kLongString:
    // src uses a refCounted string buffer, use that buffer with refCount
    // src is const, use a cast - we don't actually change it
    ((UnicodeString &)src).addRef();
    // copy all fields, share the reference-counted buffer
    fUnion.fFields.fArray = src.fUnion.fFields.fArray;
    fUnion.fFields.fCapacity = src.fUnion.fFields.fCapacity;
    if(!hasShortLength()) {
      fUnion.fFields.fLength = src.fUnion.fFields.fLength;
    }
    break;
  case kReadonlyAlias:
    if(fastCopy) {
      // src is a readonly alias, do the same
      // -> maintain the readonly alias as such
      fUnion.fFields.fArray = src.fUnion.fFields.fArray;
      fUnion.fFields.fCapacity = src.fUnion.fFields.fCapacity;
      if(!hasShortLength()) {
        fUnion.fFields.fLength = src.fUnion.fFields.fLength;
      }
      break;
    }
    // else if(!fastCopy) fall through to case kWritableAlias
    // -> allocate a new buffer and copy the contents
    U_FALLTHROUGH;
  case kWritableAlias: {
    // src is a writable alias; we make a copy of that instead
    int32_t srcLength = src.length();
    if(allocate(srcLength)) {
      u_memcpy(getArrayStart(), src.getArrayStart(), srcLength);
      setLength(srcLength);
      break;
    }
    // if there is not enough memory, then fall through to setting to bogus
    U_FALLTHROUGH;
  }
  default:
    // if src is bogus, set ourselves to bogus
    // do not call setToBogus() here because fArray and flags are not consistent here
    fUnion.fFields.fLengthAndFlags = kIsBogus;
    fUnion.fFields.fArray = 0;
    fUnion.fFields.fCapacity = 0;
    break;
  }

  return *this;
}

UnicodeString &UnicodeString::operator=(UnicodeString &&src) U_NOEXCEPT {
  // No explicit check for self move assignment, consistent with standard library.
  // Self move assignment causes no crash nor leak but might make the object bogus.
  releaseArray();
  copyFieldsFrom(src, TRUE);
  return *this;
}

// Same as move assignment except without memory management.
void UnicodeString::copyFieldsFrom(UnicodeString &src, UBool setSrcToBogus) U_NOEXCEPT {
  int16_t lengthAndFlags = fUnion.fFields.fLengthAndFlags = src.fUnion.fFields.fLengthAndFlags;
  if(lengthAndFlags & kUsingStackBuffer) {
    // Short string using the stack buffer, copy the contents.
    // Check for self assignment to prevent "overlap in memcpy" warnings,
    // although it should be harmless to copy a buffer to itself exactly.
    if(this != &src) {
      uprv_memcpy(fUnion.fStackFields.fBuffer, src.fUnion.fStackFields.fBuffer,
                  getShortLength() * U_SIZEOF_UCHAR);
    }
  } else {
    // In all other cases, copy all fields.
    fUnion.fFields.fArray = src.fUnion.fFields.fArray;
    fUnion.fFields.fCapacity = src.fUnion.fFields.fCapacity;
    if(!hasShortLength()) {
      fUnion.fFields.fLength = src.fUnion.fFields.fLength;
    }
    if(setSrcToBogus) {
      // Set src to bogus without releasing any memory.
      src.fUnion.fFields.fLengthAndFlags = kIsBogus;
      src.fUnion.fFields.fArray = NULL;
      src.fUnion.fFields.fCapacity = 0;
    }
  }
}

void UnicodeString::swap(UnicodeString &other) U_NOEXCEPT {
  UnicodeString temp;  // Empty short string: Known not to need releaseArray().
  // Copy fields without resetting source values in between.
  temp.copyFieldsFrom(*this, FALSE);
  this->copyFieldsFrom(other, FALSE);
  other.copyFieldsFrom(temp, FALSE);
  // Set temp to an empty string so that other's memory is not released twice.
  temp.fUnion.fFields.fLengthAndFlags = kShortString;
}

//========================================
// Miscellaneous operations
//========================================

UnicodeString UnicodeString::unescape() const {
    UnicodeString result(length(), (UChar32)0, (int32_t)0); // construct with capacity
    if (result.isBogus()) {
        return result;
    }
    const UChar *array = getBuffer();
    int32_t len = length();
    int32_t prev = 0;
    for (int32_t i=0;;) {
        if (i == len) {
            result.append(array, prev, len - prev);
            break;
        }
        if (array[i++] == 0x5C /*'\\'*/) {
            result.append(array, prev, (i - 1) - prev);
            UChar32 c = unescapeAt(i); // advances i
            if (c < 0) {
                result.remove(); // return empty string
                break; // invalid escape sequence
            }
            result.append(c);
            prev = i;
        }
    }
    return result;
}

UChar32 UnicodeString::unescapeAt(int32_t &offset) const {
    return u_unescapeAt(UnicodeString_charAt, &offset, length(), (void*)this);
}

//========================================
// Read-only implementation
//========================================
UBool
UnicodeString::doEquals(const UnicodeString &text, int32_t len) const {
  // Requires: this & text not bogus and have same lengths.
  // Byte-wise comparison works for equality regardless of endianness.
  return uprv_memcmp(getArrayStart(), text.getArrayStart(), len * U_SIZEOF_UCHAR) == 0;
}

int8_t
UnicodeString::doCompare( int32_t start,
              int32_t length,
              const UChar *srcChars,
              int32_t srcStart,
              int32_t srcLength) const
{
  // compare illegal string values
  if(isBogus()) {
    return -1;
  }

  // pin indices to legal values
  pinIndices(start, length);

  if(srcChars == NULL) {
    // treat const UChar *srcChars==NULL as an empty string
    return length == 0 ? 0 : 1;
  }

  // get the correct pointer
  const UChar *chars = getArrayStart();

  chars += start;
  srcChars += srcStart;

  int32_t minLength;
  int8_t lengthResult;

  // get the srcLength if necessary
  if(srcLength < 0) {
    srcLength = u_strlen(srcChars + srcStart);
  }

  // are we comparing different lengths?
  if(length != srcLength) {
    if(length < srcLength) {
      minLength = length;
      lengthResult = -1;
    } else {
      minLength = srcLength;
      lengthResult = 1;
    }
  } else {
    minLength = length;
    lengthResult = 0;
  }

  /*
   * note that uprv_memcmp() returns an int but we return an int8_t;
   * we need to take care not to truncate the result -
   * one way to do this is to right-shift the value to
   * move the sign bit into the lower 8 bits and making sure that this
   * does not become 0 itself
   */

  if(minLength > 0 && chars != srcChars) {
    int32_t result;

#   if U_IS_BIG_ENDIAN
      // big-endian: byte comparison works
      result = uprv_memcmp(chars, srcChars, minLength * sizeof(UChar));
      if(result != 0) {
        return (int8_t)(result >> 15 | 1);
      }
#   else
      // little-endian: compare UChar units
      do {
        result = ((int32_t)*(chars++) - (int32_t)*(srcChars++));
        if(result != 0) {
          return (int8_t)(result >> 15 | 1);
        }
      } while(--minLength > 0);
#   endif
  }
  return lengthResult;
}

/* String compare in code point order - doCompare() compares in code unit order. */
int8_t
UnicodeString::doCompareCodePointOrder(int32_t start,
                                       int32_t length,
                                       const UChar *srcChars,
                                       int32_t srcStart,
                                       int32_t srcLength) const
{
  // compare illegal string values
  // treat const UChar *srcChars==NULL as an empty string
  if(isBogus()) {
    return -1;
  }

  // pin indices to legal values
  pinIndices(start, length);

  if(srcChars == NULL) {
    srcStart = srcLength = 0;
  }

  int32_t diff = uprv_strCompare(getArrayStart() + start, length, (srcChars!=NULL)?(srcChars + srcStart):NULL, srcLength, FALSE, TRUE);
  /* translate the 32-bit result into an 8-bit one */
  if(diff!=0) {
    return (int8_t)(diff >> 15 | 1);
  } else {
    return 0;
  }
}

int32_t
UnicodeString::getLength() const {
    return length();
}

UChar
UnicodeString::getCharAt(int32_t offset) const {
  return charAt(offset);
}

UChar32
UnicodeString::getChar32At(int32_t offset) const {
  return char32At(offset);
}

UChar32
UnicodeString::char32At(int32_t offset) const
{
  int32_t len = length();
  if((uint32_t)offset < (uint32_t)len) {
    const UChar *array = getArrayStart();
    UChar32 c;
    U16_GET(array, 0, offset, len, c);
    return c;
  } else {
    return kInvalidUChar;
  }
}

int32_t
UnicodeString::getChar32Start(int32_t offset) const {
  if((uint32_t)offset < (uint32_t)length()) {
    const UChar *array = getArrayStart();
    U16_SET_CP_START(array, 0, offset);
    return offset;
  } else {
    return 0;
  }
}

int32_t
UnicodeString::getChar32Limit(int32_t offset) const {
  int32_t len = length();
  if((uint32_t)offset < (uint32_t)len) {
    const UChar *array = getArrayStart();
    U16_SET_CP_LIMIT(array, 0, offset, len);
    return offset;
  } else {
    return len;
  }
}

int32_t
UnicodeString::countChar32(int32_t start, int32_t length) const {
  pinIndices(start, length);
  // if(isBogus()) then fArray==0 and start==0 - u_countChar32() checks for NULL
  return u_countChar32(getArrayStart()+start, length);
}

UBool
UnicodeString::hasMoreChar32Than(int32_t start, int32_t length, int32_t number) const {
  pinIndices(start, length);
  // if(isBogus()) then fArray==0 and start==0 - u_strHasMoreChar32Than() checks for NULL
  return u_strHasMoreChar32Than(getArrayStart()+start, length, number);
}

int32_t
UnicodeString::moveIndex32(int32_t index, int32_t delta) const {
  // pin index
  int32_t len = length();
  if(index<0) {
    index=0;
  } else if(index>len) {
    index=len;
  }

  const UChar *array = getArrayStart();
  if(delta>0) {
    U16_FWD_N(array, index, len, delta);
  } else {
    U16_BACK_N(array, 0, index, -delta);
  }

  return index;
}

void
UnicodeString::doExtract(int32_t start,
             int32_t length,
             UChar *dst,
             int32_t dstStart) const
{
  // pin indices to legal values
  pinIndices(start, length);

  // do not copy anything if we alias dst itself
  const UChar *array = getArrayStart();
  if(array + start != dst + dstStart) {
    us_arrayCopy(array, start, dst, dstStart, length);
  }
}

int32_t
UnicodeString::extract(Char16Ptr dest, int32_t destCapacity,
                       UErrorCode &errorCode) const {
  int32_t len = length();
  if(U_SUCCESS(errorCode)) {
    if(isBogus() || destCapacity<0 || (destCapacity>0 && dest==0)) {
      errorCode=U_ILLEGAL_ARGUMENT_ERROR;
    } else {
      const UChar *array = getArrayStart();
      if(len>0 && len<=destCapacity && array!=dest) {
        u_memcpy(dest, array, len);
      }
      return u_terminateUChars(dest, destCapacity, len, &errorCode);
    }
  }

  return len;
}

int32_t
UnicodeString::extract(int32_t start,
                       int32_t length,
                       char *target,
                       int32_t targetCapacity,
                       enum EInvariant) const
{
  // if the arguments are illegal, then do nothing
  if(targetCapacity < 0 || (targetCapacity > 0 && target == NULL)) {
    return 0;
  }

  // pin the indices to legal values
  pinIndices(start, length);

  if(length <= targetCapacity) {
    u_UCharsToChars(getArrayStart() + start, target, length);
  }
  UErrorCode status = U_ZERO_ERROR;
  return u_terminateChars(target, targetCapacity, length, &status);
}

UnicodeString
UnicodeString::tempSubString(int32_t start, int32_t len) const {
  pinIndices(start, len);
  const UChar *array = getBuffer();  // not getArrayStart() to check kIsBogus & kOpenGetBuffer
  if(array==NULL) {
    array=fUnion.fStackFields.fBuffer;  // anything not NULL because that would make an empty string
    len=-2;  // bogus result string
  }
  return UnicodeString(FALSE, array + start, len);
}

int32_t
UnicodeString::toUTF8(int32_t start, int32_t len,
                      char *target, int32_t capacity) const {
  pinIndices(start, len);
  int32_t length8;
  UErrorCode errorCode = U_ZERO_ERROR;
  u_strToUTF8WithSub(target, capacity, &length8,
                     getBuffer() + start, len,
                     0xFFFD,  // Standard substitution character.
                     NULL,    // Don't care about number of substitutions.
                     &errorCode);
  return length8;
}

#if U_CHARSET_IS_UTF8

int32_t
UnicodeString::extract(int32_t start, int32_t len,
                       char *target, uint32_t dstSize) const {
  // if the arguments are illegal, then do nothing
  if(/*dstSize < 0 || */(dstSize > 0 && target == 0)) {
    return 0;
  }
  return toUTF8(start, len, target, dstSize <= 0x7fffffff ? (int32_t)dstSize : 0x7fffffff);
}

// else see unistr_cnv.cpp
#endif

void
UnicodeString::extractBetween(int32_t start,
                  int32_t limit,
                  UnicodeString& target) const {
  pinIndex(start);
  pinIndex(limit);
  doExtract(start, limit - start, target);
}

// When converting from UTF-16 to UTF-8, the result will have at most 3 times
// as many bytes as the source has UChars.
// The "worst cases" are writing systems like Indic, Thai and CJK with
// 3:1 bytes:UChars.
void
UnicodeString::toUTF8(ByteSink &sink) const {
  int32_t length16 = length();
  if(length16 != 0) {
    char stackBuffer[1024];
    int32_t capacity = (int32_t)sizeof(stackBuffer);
    UBool utf8IsOwned = FALSE;
    char *utf8 = sink.GetAppendBuffer(length16 < capacity ? length16 : capacity,
                                      3*length16,
                                      stackBuffer, capacity,
                                      &capacity);
    int32_t length8 = 0;
    UErrorCode errorCode = U_ZERO_ERROR;
    u_strToUTF8WithSub(utf8, capacity, &length8,
                       getBuffer(), length16,
                       0xFFFD,  // Standard substitution character.
                       NULL,    // Don't care about number of substitutions.
                       &errorCode);
    if(errorCode == U_BUFFER_OVERFLOW_ERROR) {
      utf8 = (char *)uprv_malloc(length8);
      if(utf8 != NULL) {
        utf8IsOwned = TRUE;
        errorCode = U_ZERO_ERROR;
        u_strToUTF8WithSub(utf8, length8, &length8,
                           getBuffer(), length16,
                           0xFFFD,  // Standard substitution character.
                           NULL,    // Don't care about number of substitutions.
                           &errorCode);
      } else {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
      }
    }
    if(U_SUCCESS(errorCode)) {
      sink.Append(utf8, length8);
      sink.Flush();
    }
    if(utf8IsOwned) {
      uprv_free(utf8);
    }
  }
}

int32_t
UnicodeString::toUTF32(UChar32 *utf32, int32_t capacity, UErrorCode &errorCode) const {
  int32_t length32=0;
  if(U_SUCCESS(errorCode)) {
    // getBuffer() and u_strToUTF32WithSub() check for illegal arguments.
    u_strToUTF32WithSub(utf32, capacity, &length32,
        getBuffer(), length(),
        0xfffd,  // Substitution character.
        NULL,    // Don't care about number of substitutions.
        &errorCode);
  }
  return length32;
}

int32_t
UnicodeString::indexOf(const UChar *srcChars,
               int32_t srcStart,
               int32_t srcLength,
               int32_t start,
               int32_t length) const
{
  if(isBogus() || srcChars == 0 || srcStart < 0 || srcLength == 0) {
    return -1;
  }

  // UnicodeString does not find empty substrings
  if(srcLength < 0 && srcChars[srcStart] == 0) {
    return -1;
  }

  // get the indices within bounds
  pinIndices(start, length);

  // find the first occurrence of the substring
  const UChar *array = getArrayStart();
  const UChar *match = u_strFindFirst(array + start, length, srcChars + srcStart, srcLength);
  if(match == NULL) {
    return -1;
  } else {
    return (int32_t)(match - array);
  }
}

int32_t
UnicodeString::doIndexOf(UChar c,
             int32_t start,
             int32_t length) const
{
  // pin indices
  pinIndices(start, length);

  // find the first occurrence of c
  const UChar *array = getArrayStart();
  const UChar *match = u_memchr(array + start, c, length);
  if(match == NULL) {
    return -1;
  } else {
    return (int32_t)(match - array);
  }
}

int32_t
UnicodeString::doIndexOf(UChar32 c,
                         int32_t start,
                         int32_t length) const {
  // pin indices
  pinIndices(start, length);

  // find the first occurrence of c
  const UChar *array = getArrayStart();
  const UChar *match = u_memchr32(array + start, c, length);
  if(match == NULL) {
    return -1;
  } else {
    return (int32_t)(match - array);
  }
}

int32_t
UnicodeString::lastIndexOf(const UChar *srcChars,
               int32_t srcStart,
               int32_t srcLength,
               int32_t start,
               int32_t length) const
{
  if(isBogus() || srcChars == 0 || srcStart < 0 || srcLength == 0) {
    return -1;
  }

  // UnicodeString does not find empty substrings
  if(srcLength < 0 && srcChars[srcStart] == 0) {
    return -1;
  }

  // get the indices within bounds
  pinIndices(start, length);

  // find the last occurrence of the substring
  const UChar *array = getArrayStart();
  const UChar *match = u_strFindLast(array + start, length, srcChars + srcStart, srcLength);
  if(match == NULL) {
    return -1;
  } else {
    return (int32_t)(match - array);
  }
}

int32_t
UnicodeString::doLastIndexOf(UChar c,
                 int32_t start,
                 int32_t length) const
{
  if(isBogus()) {
    return -1;
  }

  // pin indices
  pinIndices(start, length);

  // find the last occurrence of c
  const UChar *array = getArrayStart();
  const UChar *match = u_memrchr(array + start, c, length);
  if(match == NULL) {
    return -1;
  } else {
    return (int32_t)(match - array);
  }
}

int32_t
UnicodeString::doLastIndexOf(UChar32 c,
                             int32_t start,
                             int32_t length) const {
  // pin indices
  pinIndices(start, length);

  // find the last occurrence of c
  const UChar *array = getArrayStart();
  const UChar *match = u_memrchr32(array + start, c, length);
  if(match == NULL) {
    return -1;
  } else {
    return (int32_t)(match - array);
  }
}

//========================================
// Write implementation
//========================================

UnicodeString&
UnicodeString::findAndReplace(int32_t start,
                  int32_t length,
                  const UnicodeString& oldText,
                  int32_t oldStart,
                  int32_t oldLength,
                  const UnicodeString& newText,
                  int32_t newStart,
                  int32_t newLength)
{
  if(isBogus() || oldText.isBogus() || newText.isBogus()) {
    return *this;
  }

  pinIndices(start, length);
  oldText.pinIndices(oldStart, oldLength);
  newText.pinIndices(newStart, newLength);

  if(oldLength == 0) {
    return *this;
  }

  while(length > 0 && length >= oldLength) {
    int32_t pos = indexOf(oldText, oldStart, oldLength, start, length);
    if(pos < 0) {
      // no more oldText's here: done
      break;
    } else {
      // we found oldText, replace it by newText and go beyond it
      replace(pos, oldLength, newText, newStart, newLength);
      length -= pos + oldLength - start;
      start = pos + newLength;
    }
  }

  return *this;
}


void
UnicodeString::setToBogus()
{
  releaseArray();

  fUnion.fFields.fLengthAndFlags = kIsBogus;
  fUnion.fFields.fArray = 0;
  fUnion.fFields.fCapacity = 0;
}

// turn a bogus string into an empty one
void
UnicodeString::unBogus() {
  if(fUnion.fFields.fLengthAndFlags & kIsBogus) {
    setToEmpty();
  }
}

const char16_t *
UnicodeString::getTerminatedBuffer() {
  if(!isWritable()) {
    return nullptr;
  }
  UChar *array = getArrayStart();
  int32_t len = length();
  if(len < getCapacity()) {
    if(fUnion.fFields.fLengthAndFlags & kBufferIsReadonly) {
      // If len<capacity on a read-only alias, then array[len] is
      // either the original NUL (if constructed with (TRUE, s, length))
      // or one of the original string contents characters (if later truncated),
      // therefore we can assume that array[len] is initialized memory.
      if(array[len] == 0) {
        return array;
      }
    } else if(((fUnion.fFields.fLengthAndFlags & kRefCounted) == 0 || refCount() == 1)) {
      // kRefCounted: Do not write the NUL if the buffer is shared.
      // That is mostly safe, except when the length of one copy was modified
      // without copy-on-write, e.g., via truncate(newLength) or remove(void).
      // Then the NUL would be written into the middle of another copy's string.

      // Otherwise, the buffer is fully writable and it is anyway safe to write the NUL.
      // Do not test if there is a NUL already because it might be uninitialized memory.
      // (That would be safe, but tools like valgrind & Purify would complain.)
      array[len] = 0;
      return array;
    }
  }
  if(len<INT32_MAX && cloneArrayIfNeeded(len+1)) {
    array = getArrayStart();
    array[len] = 0;
    return array;
  } else {
    return nullptr;
  }
}

// setTo() analogous to the readonly-aliasing constructor with the same signature
UnicodeString &
UnicodeString::setTo(UBool isTerminated,
                     ConstChar16Ptr textPtr,
                     int32_t textLength)
{
  if(fUnion.fFields.fLengthAndFlags & kOpenGetBuffer) {
    // do not modify a string that has an "open" getBuffer(minCapacity)
    return *this;
  }

  const UChar *text = textPtr;
  if(text == NULL) {
    // treat as an empty string, do not alias
    releaseArray();
    setToEmpty();
    return *this;
  }

  if( textLength < -1 ||
      (textLength == -1 && !isTerminated) ||
      (textLength >= 0 && isTerminated && text[textLength] != 0)
  ) {
    setToBogus();
    return *this;
  }

  releaseArray();

  if(textLength == -1) {
    // text is terminated, or else it would have failed the above test
    textLength = u_strlen(text);
  }
  fUnion.fFields.fLengthAndFlags = kReadonlyAlias;
  setArray((UChar *)text, textLength, isTerminated ? textLength + 1 : textLength);
  return *this;
}

// setTo() analogous to the writable-aliasing constructor with the same signature
UnicodeString &
UnicodeString::setTo(UChar *buffer,
                     int32_t buffLength,
                     int32_t buffCapacity) {
  if(fUnion.fFields.fLengthAndFlags & kOpenGetBuffer) {
    // do not modify a string that has an "open" getBuffer(minCapacity)
    return *this;
  }

  if(buffer == NULL) {
    // treat as an empty string, do not alias
    releaseArray();
    setToEmpty();
    return *this;
  }

  if(buffLength < -1 || buffCapacity < 0 || buffLength > buffCapacity) {
    setToBogus();
    return *this;
  } else if(buffLength == -1) {
    // buffLength = u_strlen(buff); but do not look beyond buffCapacity
    const UChar *p = buffer, *limit = buffer + buffCapacity;
    while(p != limit && *p != 0) {
      ++p;
    }
    buffLength = (int32_t)(p - buffer);
  }

  releaseArray();

  fUnion.fFields.fLengthAndFlags = kWritableAlias;
  setArray(buffer, buffLength, buffCapacity);
  return *this;
}

UnicodeString &UnicodeString::setToUTF8(StringPiece utf8) {
  unBogus();
  int32_t length = utf8.length();
  int32_t capacity;
  // The UTF-16 string will be at most as long as the UTF-8 string.
  if(length <= US_STACKBUF_SIZE) {
    capacity = US_STACKBUF_SIZE;
  } else {
    capacity = length + 1;  // +1 for the terminating NUL.
  }
  UChar *utf16 = getBuffer(capacity);
  int32_t length16;
  UErrorCode errorCode = U_ZERO_ERROR;
  u_strFromUTF8WithSub(utf16, getCapacity(), &length16,
      utf8.data(), length,
      0xfffd,  // Substitution character.
      NULL,    // Don't care about number of substitutions.
      &errorCode);
  releaseBuffer(length16);
  if(U_FAILURE(errorCode)) {
    setToBogus();
  }
  return *this;
}

UnicodeString&
UnicodeString::setCharAt(int32_t offset,
             UChar c)
{
  int32_t len = length();
  if(cloneArrayIfNeeded() && len > 0) {
    if(offset < 0) {
      offset = 0;
    } else if(offset >= len) {
      offset = len - 1;
    }

    getArrayStart()[offset] = c;
  }
  return *this;
}

UnicodeString&
UnicodeString::replace(int32_t start,
               int32_t _length,
               UChar32 srcChar) {
  UChar buffer[U16_MAX_LENGTH];
  int32_t count = 0;
  UBool isError = FALSE;
  U16_APPEND(buffer, count, U16_MAX_LENGTH, srcChar, isError);
  // We test isError so that the compiler does not complain that we don't.
  // If isError (srcChar is not a valid code point) then count==0 which means
  // we remove the source segment rather than replacing it with srcChar.
  return doReplace(start, _length, buffer, 0, isError ? 0 : count);
}

UnicodeString&
UnicodeString::append(UChar32 srcChar) {
  UChar buffer[U16_MAX_LENGTH];
  int32_t _length = 0;
  UBool isError = FALSE;
  U16_APPEND(buffer, _length, U16_MAX_LENGTH, srcChar, isError);
  // We test isError so that the compiler does not complain that we don't.
  // If isError then _length==0 which turns the doAppend() into a no-op anyway.
  return isError ? *this : doAppend(buffer, 0, _length);
}

UnicodeString&
UnicodeString::doReplace( int32_t start,
              int32_t length,
              const UnicodeString& src,
              int32_t srcStart,
              int32_t srcLength)
{
  // pin the indices to legal values
  src.pinIndices(srcStart, srcLength);

  // get the characters from src
  // and replace the range in ourselves with them
  return doReplace(start, length, src.getArrayStart(), srcStart, srcLength);
}

UnicodeString&
UnicodeString::doReplace(int32_t start,
             int32_t length,
             const UChar *srcChars,
             int32_t srcStart,
             int32_t srcLength)
{
  if(!isWritable()) {
    return *this;
  }

  int32_t oldLength = this->length();

  // optimize (read-only alias).remove(0, start) and .remove(start, end)
  if((fUnion.fFields.fLengthAndFlags&kBufferIsReadonly) && srcLength == 0) {
    if(start == 0) {
      // remove prefix by adjusting the array pointer
      pinIndex(length);
      fUnion.fFields.fArray += length;
      fUnion.fFields.fCapacity -= length;
      setLength(oldLength - length);
      return *this;
    } else {
      pinIndex(start);
      if(length >= (oldLength - start)) {
        // remove suffix by reducing the length (like truncate())
        setLength(start);
        fUnion.fFields.fCapacity = start;  // not NUL-terminated any more
        return *this;
      }
    }
  }

  if(start == oldLength) {
    return doAppend(srcChars, srcStart, srcLength);
  }

  if(srcChars == 0) {
    srcLength = 0;
  } else {
    // Perform all remaining operations relative to srcChars + srcStart.
    // From this point forward, do not use srcStart.
    srcChars += srcStart;
    if (srcLength < 0) {
      // get the srcLength if necessary
      srcLength = u_strlen(srcChars);
    }
  }

  // pin the indices to legal values
  pinIndices(start, length);

  // Calculate the size of the string after the replace.
  // Avoid int32_t overflow.
  int32_t newLength = oldLength - length;
  if(srcLength > (INT32_MAX - newLength)) {
    setToBogus();
    return *this;
  }
  newLength += srcLength;

  // Check for insertion into ourself
  const UChar *oldArray = getArrayStart();
  if (isBufferWritable() &&
      oldArray < srcChars + srcLength &&
      srcChars < oldArray + oldLength) {
    // Copy into a new UnicodeString and start over
    UnicodeString copy(srcChars, srcLength);
    if (copy.isBogus()) {
      setToBogus();
      return *this;
    }
    return doReplace(start, length, copy.getArrayStart(), 0, srcLength);
  }

  // cloneArrayIfNeeded(doCopyArray=FALSE) may change fArray but will not copy the current contents;
  // therefore we need to keep the current fArray
  UChar oldStackBuffer[US_STACKBUF_SIZE];
  if((fUnion.fFields.fLengthAndFlags&kUsingStackBuffer) && (newLength > US_STACKBUF_SIZE)) {
    // copy the stack buffer contents because it will be overwritten with
    // fUnion.fFields values
    u_memcpy(oldStackBuffer, oldArray, oldLength);
    oldArray = oldStackBuffer;
  }

  // clone our array and allocate a bigger array if needed
  int32_t *bufferToDelete = 0;
  if(!cloneArrayIfNeeded(newLength, getGrowCapacity(newLength),
                         FALSE, &bufferToDelete)
  ) {
    return *this;
  }

  // now do the replace

  UChar *newArray = getArrayStart();
  if(newArray != oldArray) {
    // if fArray changed, then we need to copy everything except what will change
    us_arrayCopy(oldArray, 0, newArray, 0, start);
    us_arrayCopy(oldArray, start + length,
                 newArray, start + srcLength,
                 oldLength - (start + length));
  } else if(length != srcLength) {
    // fArray did not change; copy only the portion that isn't changing, leaving a hole
    us_arrayCopy(oldArray, start + length,
                 newArray, start + srcLength,
                 oldLength - (start + length));
  }

  // now fill in the hole with the new string
  us_arrayCopy(srcChars, 0, newArray, start, srcLength);

  setLength(newLength);

  // delayed delete in case srcChars == fArray when we started, and
  // to keep oldArray alive for the above operations
  if (bufferToDelete) {
    uprv_free(bufferToDelete);
  }

  return *this;
}

// Versions of doReplace() only for append() variants.
// doReplace() and doAppend() optimize for different cases.

UnicodeString&
UnicodeString::doAppend(const UnicodeString& src, int32_t srcStart, int32_t srcLength) {
  if(srcLength == 0) {
    return *this;
  }

  // pin the indices to legal values
  src.pinIndices(srcStart, srcLength);
  return doAppend(src.getArrayStart(), srcStart, srcLength);
}

UnicodeString&
UnicodeString::doAppend(const UChar *srcChars, int32_t srcStart, int32_t srcLength) {
  if(!isWritable() || srcLength == 0 || srcChars == NULL) {
    return *this;
  }

  // Perform all remaining operations relative to srcChars + srcStart.
  // From this point forward, do not use srcStart.
  srcChars += srcStart;

  if(srcLength < 0) {
    // get the srcLength if necessary
    if((srcLength = u_strlen(srcChars)) == 0) {
      return *this;
    }
  }

  int32_t oldLength = length();
  int32_t newLength = oldLength + srcLength;

  // Check for append onto ourself
  const UChar* oldArray = getArrayStart();
  if (isBufferWritable() &&
      oldArray < srcChars + srcLength &&
      srcChars < oldArray + oldLength) {
    // Copy into a new UnicodeString and start over
    UnicodeString copy(srcChars, srcLength);
    if (copy.isBogus()) {
      setToBogus();
      return *this;
    }
    return doAppend(copy.getArrayStart(), 0, srcLength);
  }

  // optimize append() onto a large-enough, owned string
  if((newLength <= getCapacity() && isBufferWritable()) ||
      cloneArrayIfNeeded(newLength, getGrowCapacity(newLength))) {
    UChar *newArray = getArrayStart();
    // Do not copy characters when
    //   UChar *buffer=str.getAppendBuffer(...);
    // is followed by
    //   str.append(buffer, length);
    // or
    //   str.appendString(buffer, length)
    // or similar.
    if(srcChars != newArray + oldLength) {
      us_arrayCopy(srcChars, 0, newArray, oldLength, srcLength);
    }
    setLength(newLength);
  }
  return *this;
}

/**
 * Replaceable API
 */
void
UnicodeString::handleReplaceBetween(int32_t start,
                                    int32_t limit,
                                    const UnicodeString& text) {
    replaceBetween(start, limit, text);
}

/**
 * Replaceable API
 */
void
UnicodeString::copy(int32_t start, int32_t limit, int32_t dest) {
    if (limit <= start) {
        return; // Nothing to do; avoid bogus malloc call
    }
    UChar* text = (UChar*) uprv_malloc( sizeof(UChar) * (limit - start) );
    // Check to make sure text is not null.
    if (text != NULL) {
	    extractBetween(start, limit, text, 0);
	    insert(dest, text, 0, limit - start);
	    uprv_free(text);
    }
}

/**
 * Replaceable API
 *
 * NOTE: This is for the Replaceable class.  There is no rep.cpp,
 * so we implement this function here.
 */
UBool Replaceable::hasMetaData() const {
    return TRUE;
}

/**
 * Replaceable API
 */
UBool UnicodeString::hasMetaData() const {
    return FALSE;
}

UnicodeString&
UnicodeString::doReverse(int32_t start, int32_t length) {
  if(length <= 1 || !cloneArrayIfNeeded()) {
    return *this;
  }

  // pin the indices to legal values
  pinIndices(start, length);
  if(length <= 1) {  // pinIndices() might have shrunk the length
    return *this;
  }

  UChar *left = getArrayStart() + start;
  UChar *right = left + length - 1;  // -1 for inclusive boundary (length>=2)
  UChar swap;
  UBool hasSupplementary = FALSE;

  // Before the loop we know left<right because length>=2.
  do {
    hasSupplementary |= (UBool)U16_IS_LEAD(swap = *left);
    hasSupplementary |= (UBool)U16_IS_LEAD(*left++ = *right);
    *right-- = swap;
  } while(left < right);
  // Make sure to test the middle code unit of an odd-length string.
  // Redundant if the length is even.
  hasSupplementary |= (UBool)U16_IS_LEAD(*left);

  /* if there are supplementary code points in the reversed range, then re-swap their surrogates */
  if(hasSupplementary) {
    UChar swap2;

    left = getArrayStart() + start;
    right = left + length - 1; // -1 so that we can look at *(left+1) if left<right
    while(left < right) {
      if(U16_IS_TRAIL(swap = *left) && U16_IS_LEAD(swap2 = *(left + 1))) {
        *left++ = swap2;
        *left++ = swap;
      } else {
        ++left;
      }
    }
  }

  return *this;
}

UBool
UnicodeString::padLeading(int32_t targetLength,
                          UChar padChar)
{
  int32_t oldLength = length();
  if(oldLength >= targetLength || !cloneArrayIfNeeded(targetLength)) {
    return FALSE;
  } else {
    // move contents up by padding width
    UChar *array = getArrayStart();
    int32_t start = targetLength - oldLength;
    us_arrayCopy(array, 0, array, start, oldLength);

    // fill in padding character
    while(--start >= 0) {
      array[start] = padChar;
    }
    setLength(targetLength);
    return TRUE;
  }
}

UBool
UnicodeString::padTrailing(int32_t targetLength,
                           UChar padChar)
{
  int32_t oldLength = length();
  if(oldLength >= targetLength || !cloneArrayIfNeeded(targetLength)) {
    return FALSE;
  } else {
    // fill in padding character
    UChar *array = getArrayStart();
    int32_t length = targetLength;
    while(--length >= oldLength) {
      array[length] = padChar;
    }
    setLength(targetLength);
    return TRUE;
  }
}

//========================================
// Hashing
//========================================
int32_t
UnicodeString::doHashCode() const
{
    /* Delegate hash computation to uhash.  This makes UnicodeString
     * hashing consistent with UChar* hashing.  */
    int32_t hashCode = ustr_hashUCharsN(getArrayStart(), length());
    if (hashCode == kInvalidHashCode) {
        hashCode = kEmptyHashCode;
    }
    return hashCode;
}

//========================================
// External Buffer
//========================================

char16_t *
UnicodeString::getBuffer(int32_t minCapacity) {
  if(minCapacity>=-1 && cloneArrayIfNeeded(minCapacity)) {
    fUnion.fFields.fLengthAndFlags|=kOpenGetBuffer;
    setZeroLength();
    return getArrayStart();
  } else {
    return nullptr;
  }
}

void
UnicodeString::releaseBuffer(int32_t newLength) {
  if(fUnion.fFields.fLengthAndFlags&kOpenGetBuffer && newLength>=-1) {
    // set the new fLength
    int32_t capacity=getCapacity();
    if(newLength==-1) {
      // the new length is the string length, capped by fCapacity
      const UChar *array=getArrayStart(), *p=array, *limit=array+capacity;
      while(p<limit && *p!=0) {
        ++p;
      }
      newLength=(int32_t)(p-array);
    } else if(newLength>capacity) {
      newLength=capacity;
    }
    setLength(newLength);
    fUnion.fFields.fLengthAndFlags&=~kOpenGetBuffer;
  }
}

//========================================
// Miscellaneous
//========================================
UBool
UnicodeString::cloneArrayIfNeeded(int32_t newCapacity,
                                  int32_t growCapacity,
                                  UBool doCopyArray,
                                  int32_t **pBufferToDelete,
                                  UBool forceClone) {
  // default parameters need to be static, therefore
  // the defaults are -1 to have convenience defaults
  if(newCapacity == -1) {
    newCapacity = getCapacity();
  }

  // while a getBuffer(minCapacity) is "open",
  // prevent any modifications of the string by returning FALSE here
  // if the string is bogus, then only an assignment or similar can revive it
  if(!isWritable()) {
    return FALSE;
  }

  /*
   * We need to make a copy of the array if
   * the buffer is read-only, or
   * the buffer is refCounted (shared), and refCount>1, or
   * the buffer is too small.
   * Return FALSE if memory could not be allocated.
   */
  if(forceClone ||
     fUnion.fFields.fLengthAndFlags & kBufferIsReadonly ||
     (fUnion.fFields.fLengthAndFlags & kRefCounted && refCount() > 1) ||
     newCapacity > getCapacity()
  ) {
    // check growCapacity for default value and use of the stack buffer
    if(growCapacity < 0) {
      growCapacity = newCapacity;
    } else if(newCapacity <= US_STACKBUF_SIZE && growCapacity > US_STACKBUF_SIZE) {
      growCapacity = US_STACKBUF_SIZE;
    }

    // save old values
    UChar oldStackBuffer[US_STACKBUF_SIZE];
    UChar *oldArray;
    int32_t oldLength = length();
    int16_t flags = fUnion.fFields.fLengthAndFlags;

    if(flags&kUsingStackBuffer) {
      U_ASSERT(!(flags&kRefCounted)); /* kRefCounted and kUsingStackBuffer are mutally exclusive */
      if(doCopyArray && growCapacity > US_STACKBUF_SIZE) {
        // copy the stack buffer contents because it will be overwritten with
        // fUnion.fFields values
        us_arrayCopy(fUnion.fStackFields.fBuffer, 0, oldStackBuffer, 0, oldLength);
        oldArray = oldStackBuffer;
      } else {
        oldArray = NULL; // no need to copy from the stack buffer to itself
      }
    } else {
      oldArray = fUnion.fFields.fArray;
      U_ASSERT(oldArray!=NULL); /* when stack buffer is not used, oldArray must have a non-NULL reference */
    }

    // allocate a new array
    if(allocate(growCapacity) ||
       (newCapacity < growCapacity && allocate(newCapacity))
    ) {
      if(doCopyArray) {
        // copy the contents
        // do not copy more than what fits - it may be smaller than before
        int32_t minLength = oldLength;
        newCapacity = getCapacity();
        if(newCapacity < minLength) {
          minLength = newCapacity;
        }
        if(oldArray != NULL) {
          us_arrayCopy(oldArray, 0, getArrayStart(), 0, minLength);
        }
        setLength(minLength);
      } else {
        setZeroLength();
      }

      // release the old array
      if(flags & kRefCounted) {
        // the array is refCounted; decrement and release if 0
        u_atomic_int32_t *pRefCount = ((u_atomic_int32_t *)oldArray - 1);
        if(umtx_atomic_dec(pRefCount) == 0) {
          if(pBufferToDelete == 0) {
              // Note: cast to (void *) is needed with MSVC, where u_atomic_int32_t
              // is defined as volatile. (Volatile has useful non-standard behavior
              //   with this compiler.)
            uprv_free((void *)pRefCount);
          } else {
            // the caller requested to delete it himself
            *pBufferToDelete = (int32_t *)pRefCount;
          }
        }
      }
    } else {
      // not enough memory for growCapacity and not even for the smaller newCapacity
      // reset the old values for setToBogus() to release the array
      if(!(flags&kUsingStackBuffer)) {
        fUnion.fFields.fArray = oldArray;
      }
      fUnion.fFields.fLengthAndFlags = flags;
      setToBogus();
      return FALSE;
    }
  }
  return TRUE;
}

// UnicodeStringAppendable ------------------------------------------------- ***

UnicodeStringAppendable::~UnicodeStringAppendable() {}

UBool
UnicodeStringAppendable::appendCodeUnit(UChar c) {
  return str.doAppend(&c, 0, 1).isWritable();
}

UBool
UnicodeStringAppendable::appendCodePoint(UChar32 c) {
  UChar buffer[U16_MAX_LENGTH];
  int32_t cLength = 0;
  UBool isError = FALSE;
  U16_APPEND(buffer, cLength, U16_MAX_LENGTH, c, isError);
  return !isError && str.doAppend(buffer, 0, cLength).isWritable();
}

UBool
UnicodeStringAppendable::appendString(const UChar *s, int32_t length) {
  return str.doAppend(s, 0, length).isWritable();
}

UBool
UnicodeStringAppendable::reserveAppendCapacity(int32_t appendCapacity) {
  return str.cloneArrayIfNeeded(str.length() + appendCapacity);
}

UChar *
UnicodeStringAppendable::getAppendBuffer(int32_t minCapacity,
                                         int32_t desiredCapacityHint,
                                         UChar *scratch, int32_t scratchCapacity,
                                         int32_t *resultCapacity) {
  if(minCapacity < 1 || scratchCapacity < minCapacity) {
    *resultCapacity = 0;
    return NULL;
  }
  int32_t oldLength = str.length();
  if(minCapacity <= (kMaxCapacity - oldLength) &&
      desiredCapacityHint <= (kMaxCapacity - oldLength) &&
      str.cloneArrayIfNeeded(oldLength + minCapacity, oldLength + desiredCapacityHint)) {
    *resultCapacity = str.getCapacity() - oldLength;
    return str.getArrayStart() + oldLength;
  }
  *resultCapacity = scratchCapacity;
  return scratch;
}

U_NAMESPACE_END

U_NAMESPACE_USE

U_CAPI int32_t U_EXPORT2
uhash_hashUnicodeString(const UElement key) {
    const UnicodeString *str = (const UnicodeString*) key.pointer;
    return (str == NULL) ? 0 : str->hashCode();
}

// Moved here from uhash_us.cpp so that using a UVector of UnicodeString*
// does not depend on hashtable code.
U_CAPI UBool U_EXPORT2
uhash_compareUnicodeString(const UElement key1, const UElement key2) {
    const UnicodeString *str1 = (const UnicodeString*) key1.pointer;
    const UnicodeString *str2 = (const UnicodeString*) key2.pointer;
    if (str1 == str2) {
        return TRUE;
    }
    if (str1 == NULL || str2 == NULL) {
        return FALSE;
    }
    return *str1 == *str2;
}

// #ifdef U_STATIC_IMPLEMENTATION
// /*
// This should never be called. It is defined here to make sure that the
// virtual vector deleting destructor is defined within unistr.cpp.
// The vector deleting destructor is already a part of UObject,
// but defining it here makes sure that it is included with this object file.
// This makes sure that static library dependencies are kept to a minimum.
// */
// static void uprv_UnicodeStringDummy(void) {
//     delete [] (new UnicodeString[2]);
// }
// #endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1999-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  unistr_case.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:2
*
*   created on: 2004aug19
*   created by: Markus W. Scherer
*
*   Case-mapping functions moved here from unistr.cpp
*/















U_NAMESPACE_BEGIN

//========================================
// Read-only implementation
//========================================

int8_t
UnicodeString::doCaseCompare(int32_t start,
                             int32_t length,
                             const UChar *srcChars,
                             int32_t srcStart,
                             int32_t srcLength,
                             uint32_t options) const
{
  // compare illegal string values
  // treat const UChar *srcChars==NULL as an empty string
  if(isBogus()) {
    return -1;
  }

  // pin indices to legal values
  pinIndices(start, length);

  if(srcChars == NULL) {
    srcStart = srcLength = 0;
  }

  // get the correct pointer
  const UChar *chars = getArrayStart();

  chars += start;
  if(srcStart!=0) {
    srcChars += srcStart;
  }

  if(chars != srcChars) {
    UErrorCode errorCode=U_ZERO_ERROR;
    int32_t result=u_strcmpFold(chars, length, srcChars, srcLength,
                                options|U_COMPARE_IGNORE_CASE, &errorCode);
    if(result!=0) {
      return (int8_t)(result >> 24 | 1);
    }
  } else {
    // get the srcLength if necessary
    if(srcLength < 0) {
      srcLength = u_strlen(srcChars + srcStart);
    }
    if(length != srcLength) {
      return (int8_t)((length - srcLength) >> 24 | 1);
    }
  }
  return 0;
}

//========================================
// Write implementation
//========================================

UnicodeString &
UnicodeString::caseMap(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
                       UStringCaseMapper *stringCaseMapper) {
  if(isEmpty() || !isWritable()) {
    // nothing to do
    return *this;
  }

  UChar oldBuffer[2 * US_STACKBUF_SIZE];
  UChar *oldArray;
  int32_t oldLength = length();
  int32_t newLength;
  UBool writable = isBufferWritable();
  UErrorCode errorCode = U_ZERO_ERROR;

#if !UCONFIG_NO_BREAK_ITERATION
  // Read-only alias to the original string contents for the titlecasing BreakIterator.
  // We cannot set the iterator simply to *this because *this is being modified.
  UnicodeString oldString;
#endif

  // Try to avoid heap-allocating a new character array for this string.
  if (writable ? oldLength <= UPRV_LENGTHOF(oldBuffer) : oldLength < US_STACKBUF_SIZE) {
    // Short string: Copy the contents into a temporary buffer and
    // case-map back into the current array, or into the stack buffer.
    UChar *buffer = getArrayStart();
    int32_t capacity;
    oldArray = oldBuffer;
    u_memcpy(oldBuffer, buffer, oldLength);
    if (writable) {
      capacity = getCapacity();
    } else {
      // Switch from the read-only alias or shared heap buffer to the stack buffer.
      if (!cloneArrayIfNeeded(US_STACKBUF_SIZE, US_STACKBUF_SIZE, /* doCopyArray= */ FALSE)) {
        return *this;
      }
      U_ASSERT(fUnion.fFields.fLengthAndFlags & kUsingStackBuffer);
      buffer = fUnion.fStackFields.fBuffer;
      capacity = US_STACKBUF_SIZE;
    }
#if !UCONFIG_NO_BREAK_ITERATION
    if (iter != nullptr) {
      oldString.setTo(FALSE, oldArray, oldLength);
      iter->setText(oldString);
    }
#endif
    newLength = stringCaseMapper(caseLocale, options, UCASEMAP_BREAK_ITERATOR
                                 buffer, capacity,
                                 oldArray, oldLength, NULL, errorCode);
    if (U_SUCCESS(errorCode)) {
      setLength(newLength);
      return *this;
    } else if (errorCode == U_BUFFER_OVERFLOW_ERROR) {
      // common overflow handling below
    } else {
      setToBogus();
      return *this;
    }
  } else {
    // Longer string or read-only buffer:
    // Collect only changes and then apply them to this string.
    // Case mapping often changes only small parts of a string,
    // and often does not change its length.
    oldArray = getArrayStart();
    Edits edits;
    UChar replacementChars[200];
#if !UCONFIG_NO_BREAK_ITERATION
    if (iter != nullptr) {
      oldString.setTo(FALSE, oldArray, oldLength);
      iter->setText(oldString);
    }
#endif
    stringCaseMapper(caseLocale, options | U_OMIT_UNCHANGED_TEXT, UCASEMAP_BREAK_ITERATOR
                     replacementChars, UPRV_LENGTHOF(replacementChars),
                     oldArray, oldLength, &edits, errorCode);
    if (U_SUCCESS(errorCode)) {
      // Grow the buffer at most once, not for multiple doReplace() calls.
      newLength = oldLength + edits.lengthDelta();
      if (newLength > oldLength && !cloneArrayIfNeeded(newLength, newLength)) {
        return *this;
      }
      for (Edits::Iterator ei = edits.getCoarseChangesIterator(); ei.next(errorCode);) {
        doReplace(ei.destinationIndex(), ei.oldLength(),
                  replacementChars, ei.replacementIndex(), ei.newLength());
      }
      if (U_FAILURE(errorCode)) {
        setToBogus();
      }
      return *this;
    } else if (errorCode == U_BUFFER_OVERFLOW_ERROR) {
      // common overflow handling below
      newLength = oldLength + edits.lengthDelta();
    } else {
      setToBogus();
      return *this;
    }
  }

  // Handle buffer overflow, newLength is known.
  // We need to allocate a new buffer for the internal string case mapping function.
  // This is very similar to how doReplace() keeps the old array pointer
  // and deletes the old array itself after it is done.
  // In addition, we are forcing cloneArrayIfNeeded() to always allocate a new array.
  int32_t *bufferToDelete = 0;
  if (!cloneArrayIfNeeded(newLength, newLength, FALSE, &bufferToDelete, TRUE)) {
    return *this;
  }
  errorCode = U_ZERO_ERROR;
  // No need to iter->setText() again: The case mapper restarts via iter->first().
  newLength = stringCaseMapper(caseLocale, options, UCASEMAP_BREAK_ITERATOR
                               getArrayStart(), getCapacity(),
                               oldArray, oldLength, NULL, errorCode);
  if (bufferToDelete) {
    uprv_free(bufferToDelete);
  }
  if (U_SUCCESS(errorCode)) {
    setLength(newLength);
  } else {
    setToBogus();
  }
  return *this;
}

UnicodeString &
UnicodeString::foldCase(uint32_t options) {
  return caseMap(UCASE_LOC_ROOT, options, UCASEMAP_BREAK_ITERATOR_NULL ustrcase_internalFold);
}

U_NAMESPACE_END

// Defined here to reduce dependencies on break iterator
U_CAPI int32_t U_EXPORT2
uhash_hashCaselessUnicodeString(const UElement key) {
    U_NAMESPACE_USE
    const UnicodeString *str = (const UnicodeString*) key.pointer;
    if (str == NULL) {
        return 0;
    }
    // Inefficient; a better way would be to have a hash function in
    // UnicodeString that does case folding on the fly.
    UnicodeString copy(*str);
    return copy.foldCase().hashCode();
}

// Defined here to reduce dependencies on break iterator
U_CAPI UBool U_EXPORT2
uhash_compareCaselessUnicodeString(const UElement key1, const UElement key2) {
    U_NAMESPACE_USE
    const UnicodeString *str1 = (const UnicodeString*) key1.pointer;
    const UnicodeString *str2 = (const UnicodeString*) key2.pointer;
    if (str1 == str2) {
        return TRUE;
    }
    if (str1 == NULL || str2 == NULL) {
        return FALSE;
    }
    return str1->caseCompare(*str2, U_FOLD_CASE_DEFAULT) == 0;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  unistr_case_locale.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011may31
*   created by: Markus W. Scherer
*
*   Locale-sensitive case mapping functions (ones that call uloc_getDefault())
*   were moved here to break dependency cycles among parts of the common library.
*/







U_NAMESPACE_BEGIN

//========================================
// Write implementation
//========================================

UnicodeString &
UnicodeString::toLower() {
  return caseMap(ustrcase_getCaseLocale(NULL), 0,
                 UCASEMAP_BREAK_ITERATOR_NULL ustrcase_internalToLower);
}

UnicodeString &
UnicodeString::toLower(const Locale &locale) {
  return caseMap(ustrcase_getCaseLocale(locale.getBaseName()), 0,
                 UCASEMAP_BREAK_ITERATOR_NULL ustrcase_internalToLower);
}

UnicodeString &
UnicodeString::toUpper() {
  return caseMap(ustrcase_getCaseLocale(NULL), 0,
                 UCASEMAP_BREAK_ITERATOR_NULL ustrcase_internalToUpper);
}

UnicodeString &
UnicodeString::toUpper(const Locale &locale) {
  return caseMap(ustrcase_getCaseLocale(locale.getBaseName()), 0,
                 UCASEMAP_BREAK_ITERATOR_NULL ustrcase_internalToUpper);
}

U_NAMESPACE_END
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// *******************************************************************************
// *
// *   Copyright (C) 1999-2014, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *
// *******************************************************************************
// *   file name:  unistr_cnv.cpp
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:2
// *
// *   created on: 2004aug19
// *   created by: Markus W. Scherer
// *
// *   Character conversion functions moved here from unistr.cpp
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "unicode/putil.h"
// #include "cstring.h"
// #include "cmemory.h"
// #include "unicode/ustring.h"
// #include "unicode/unistr.h"
// #include "unicode/ucnv.h"
// #include "ucnv_imp.h"
// #include "putilimp.h"
// #include "ustr_cnv.h"
// #include "ustr_imp.h"

// U_NAMESPACE_BEGIN

// //========================================
// // Constructors
// //========================================

// #if !U_CHARSET_IS_UTF8

// UnicodeString::UnicodeString(const char *codepageData) {
//     fUnion.fFields.fLengthAndFlags = kShortString;
//     if(codepageData != 0) {
//         doCodepageCreate(codepageData, (int32_t)uprv_strlen(codepageData), 0);
//     }
// }

// UnicodeString::UnicodeString(const char *codepageData,
//                              int32_t dataLength) {
//     fUnion.fFields.fLengthAndFlags = kShortString;
//     if(codepageData != 0) {
//         doCodepageCreate(codepageData, dataLength, 0);
//     }
// }

// // else see unistr.cpp
// #endif

// UnicodeString::UnicodeString(const char *codepageData,
//                              const char *codepage) {
//     fUnion.fFields.fLengthAndFlags = kShortString;
//     if(codepageData != 0) {
//         doCodepageCreate(codepageData, (int32_t)uprv_strlen(codepageData), codepage);
//     }
// }

// UnicodeString::UnicodeString(const char *codepageData,
//                              int32_t dataLength,
//                              const char *codepage) {
//     fUnion.fFields.fLengthAndFlags = kShortString;
//     if(codepageData != 0) {
//         doCodepageCreate(codepageData, dataLength, codepage);
//     }
// }

// UnicodeString::UnicodeString(const char *src, int32_t srcLength,
//                              UConverter *cnv,
//                              UErrorCode &errorCode) {
//     fUnion.fFields.fLengthAndFlags = kShortString;
//     if(U_SUCCESS(errorCode)) {
//         // check arguments
//         if(src==NULL) {
//             // treat as an empty string, do nothing more
//         } else if(srcLength<-1) {
//             errorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         } else {
//             // get input length
//             if(srcLength==-1) {
//                 srcLength=(int32_t)uprv_strlen(src);
//             }
//             if(srcLength>0) {
//                 if(cnv!=0) {
//                     // use the provided converter
//                     ucnv_resetToUnicode(cnv);
//                     doCodepageCreate(src, srcLength, cnv, errorCode);
//                 } else {
//                     // use the default converter
//                     cnv=u_getDefaultConverter(&errorCode);
//                     doCodepageCreate(src, srcLength, cnv, errorCode);
//                     u_releaseDefaultConverter(cnv);
//                 }
//             }
//         }

//         if(U_FAILURE(errorCode)) {
//             setToBogus();
//         }
//     }
// }

// //========================================
// // Codeset conversion
// //========================================

// #if !U_CHARSET_IS_UTF8

// int32_t
// UnicodeString::extract(int32_t start,
//                        int32_t length,
//                        char *target,
//                        uint32_t dstSize) const {
//     return extract(start, length, target, dstSize, 0);
// }

// // else see unistr.cpp
// #endif

// int32_t
// UnicodeString::extract(int32_t start,
//                        int32_t length,
//                        char *target,
//                        uint32_t dstSize,
//                        const char *codepage) const
// {
//     // if the arguments are illegal, then do nothing
//     if(/*dstSize < 0 || */(dstSize > 0 && target == 0)) {
//         return 0;
//     }

//     // pin the indices to legal values
//     pinIndices(start, length);

//     // We need to cast dstSize to int32_t for all subsequent code.
//     // I don't know why the API was defined with uint32_t but we are stuck with it.
//     // Also, dstSize==0xffffffff means "unlimited" but if we use target+dstSize
//     // as a limit in some functions, it may wrap around and yield a pointer
//     // that compares less-than target.
//     int32_t capacity;
//     if(dstSize < 0x7fffffff) {
//         // Assume that the capacity is real and a limit pointer won't wrap around.
//         capacity = (int32_t)dstSize;
//     } else {
//         // Pin the capacity so that a limit pointer does not wrap around.
//         char *targetLimit = (char *)U_MAX_PTR(target);
//         // U_MAX_PTR(target) returns a targetLimit that is at most 0x7fffffff
//         // greater than target and does not wrap around the top of the address space.
//         capacity = (int32_t)(targetLimit - target);
//     }

//     // create the converter
//     UConverter *converter;
//     UErrorCode status = U_ZERO_ERROR;

//     // just write the NUL if the string length is 0
//     if(length == 0) {
//         return u_terminateChars(target, capacity, 0, &status);
//     }

//     // if the codepage is the default, use our cache
//     // if it is an empty string, then use the "invariant character" conversion
//     if (codepage == 0) {
//         const char *defaultName = ucnv_getDefaultName();
//         if(UCNV_FAST_IS_UTF8(defaultName)) {
//             return toUTF8(start, length, target, capacity);
//         }
//         converter = u_getDefaultConverter(&status);
//     } else if (*codepage == 0) {
//         // use the "invariant characters" conversion
//         int32_t destLength;
//         if(length <= capacity) {
//             destLength = length;
//         } else {
//             destLength = capacity;
//         }
//         u_UCharsToChars(getArrayStart() + start, target, destLength);
//         return u_terminateChars(target, capacity, length, &status);
//     } else {
//         converter = ucnv_open(codepage, &status);
//     }

//     length = doExtract(start, length, target, capacity, converter, status);

//     // close the converter
//     if (codepage == 0) {
//         u_releaseDefaultConverter(converter);
//     } else {
//         ucnv_close(converter);
//     }

//     return length;
// }

// int32_t
// UnicodeString::extract(char *dest, int32_t destCapacity,
//                        UConverter *cnv,
//                        UErrorCode &errorCode) const
// {
//     if(U_FAILURE(errorCode)) {
//         return 0;
//     }

//     if(isBogus() || destCapacity<0 || (destCapacity>0 && dest==0)) {
//         errorCode=U_ILLEGAL_ARGUMENT_ERROR;
//         return 0;
//     }

//     // nothing to do?
//     if(isEmpty()) {
//         return u_terminateChars(dest, destCapacity, 0, &errorCode);
//     }

//     // get the converter
//     UBool isDefaultConverter;
//     if(cnv==0) {
//         isDefaultConverter=TRUE;
//         cnv=u_getDefaultConverter(&errorCode);
//         if(U_FAILURE(errorCode)) {
//             return 0;
//         }
//     } else {
//         isDefaultConverter=FALSE;
//         ucnv_resetFromUnicode(cnv);
//     }

//     // convert
//     int32_t len=doExtract(0, length(), dest, destCapacity, cnv, errorCode);

//     // release the converter
//     if(isDefaultConverter) {
//         u_releaseDefaultConverter(cnv);
//     }

//     return len;
// }

// int32_t
// UnicodeString::doExtract(int32_t start, int32_t length,
//                          char *dest, int32_t destCapacity,
//                          UConverter *cnv,
//                          UErrorCode &errorCode) const
// {
//     if(U_FAILURE(errorCode)) {
//         if(destCapacity!=0) {
//             *dest=0;
//         }
//         return 0;
//     }

//     const UChar *src=getArrayStart()+start, *srcLimit=src+length;
//     char *originalDest=dest;
//     const char *destLimit;

//     if(destCapacity==0) {
//         destLimit=dest=0;
//     } else if(destCapacity==-1) {
//         // Pin the limit to U_MAX_PTR if the "magic" destCapacity is used.
//         destLimit=(char*)U_MAX_PTR(dest);
//         // for NUL-termination, translate into highest int32_t
//         destCapacity=0x7fffffff;
//     } else {
//         destLimit=dest+destCapacity;
//     }

//     // perform the conversion
//     ucnv_fromUnicode(cnv, &dest, destLimit, &src, srcLimit, 0, TRUE, &errorCode);
//     length=(int32_t)(dest-originalDest);

//     // if an overflow occurs, then get the preflighting length
//     if(errorCode==U_BUFFER_OVERFLOW_ERROR) {
//         char buffer[1024];

//         destLimit=buffer+sizeof(buffer);
//         do {
//             dest=buffer;
//             errorCode=U_ZERO_ERROR;
//             ucnv_fromUnicode(cnv, &dest, destLimit, &src, srcLimit, 0, TRUE, &errorCode);
//             length+=(int32_t)(dest-buffer);
//         } while(errorCode==U_BUFFER_OVERFLOW_ERROR);
//     }

//     return u_terminateChars(originalDest, destCapacity, length, &errorCode);
// }

// void
// UnicodeString::doCodepageCreate(const char *codepageData,
//                                 int32_t dataLength,
//                                 const char *codepage)
// {
//     // if there's nothing to convert, do nothing
//     if(codepageData == 0 || dataLength == 0 || dataLength < -1) {
//         return;
//     }
//     if(dataLength == -1) {
//         dataLength = (int32_t)uprv_strlen(codepageData);
//     }

//     UErrorCode status = U_ZERO_ERROR;

//     // create the converter
//     // if the codepage is the default, use our cache
//     // if it is an empty string, then use the "invariant character" conversion
//     UConverter *converter;
//     if (codepage == 0) {
//         const char *defaultName = ucnv_getDefaultName();
//         if(UCNV_FAST_IS_UTF8(defaultName)) {
//             setToUTF8(StringPiece(codepageData, dataLength));
//             return;
//         }
//         converter = u_getDefaultConverter(&status);
//     } else if(*codepage == 0) {
//         // use the "invariant characters" conversion
//         if(cloneArrayIfNeeded(dataLength, dataLength, FALSE)) {
//             u_charsToUChars(codepageData, getArrayStart(), dataLength);
//             setLength(dataLength);
//         } else {
//             setToBogus();
//         }
//         return;
//     } else {
//         converter = ucnv_open(codepage, &status);
//     }

//     // if we failed, set the appropriate flags and return
//     if(U_FAILURE(status)) {
//         setToBogus();
//         return;
//     }

//     // perform the conversion
//     doCodepageCreate(codepageData, dataLength, converter, status);
//     if(U_FAILURE(status)) {
//         setToBogus();
//     }

//     // close the converter
//     if(codepage == 0) {
//         u_releaseDefaultConverter(converter);
//     } else {
//         ucnv_close(converter);
//     }
// }

// void
// UnicodeString::doCodepageCreate(const char *codepageData,
//                                 int32_t dataLength,
//                                 UConverter *converter,
//                                 UErrorCode &status)
// {
//     if(U_FAILURE(status)) {
//         return;
//     }

//     // set up the conversion parameters
//     const char *mySource     = codepageData;
//     const char *mySourceEnd  = mySource + dataLength;
//     UChar *array, *myTarget;

//     // estimate the size needed:
//     int32_t arraySize;
//     if(dataLength <= US_STACKBUF_SIZE) {
//         // try to use the stack buffer
//         arraySize = US_STACKBUF_SIZE;
//     } else {
//         // 1.25 UChar's per source byte should cover most cases
//         arraySize = dataLength + (dataLength >> 2);
//     }

//     // we do not care about the current contents
//     UBool doCopyArray = FALSE;
//     for(;;) {
//         if(!cloneArrayIfNeeded(arraySize, arraySize, doCopyArray)) {
//             setToBogus();
//             break;
//         }

//         // perform the conversion
//         array = getArrayStart();
//         myTarget = array + length();
//         ucnv_toUnicode(converter, &myTarget,  array + getCapacity(),
//             &mySource, mySourceEnd, 0, TRUE, &status);

//         // update the conversion parameters
//         setLength((int32_t)(myTarget - array));

//         // allocate more space and copy data, if needed
//         if(status == U_BUFFER_OVERFLOW_ERROR) {
//             // reset the error code
//             status = U_ZERO_ERROR;

//             // keep the previous conversion results
//             doCopyArray = TRUE;

//             // estimate the new size needed, larger than before
//             // try 2 UChar's per remaining source byte
//             arraySize = (int32_t)(length() + 2 * (mySourceEnd - mySource));
//         } else {
//             break;
//         }
//     }
// }

// U_NAMESPACE_END

// #endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1999-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  unistr_props.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:2
*
*   created on: 2004aug25
*   created by: Markus W. Scherer
*
*   Character property dependent functions moved here from unistr.cpp
*/






U_NAMESPACE_BEGIN

UnicodeString& 
UnicodeString::trim()
{
  if(isBogus()) {
    return *this;
  }

  UChar *array = getArrayStart();
  UChar32 c;
  int32_t oldLength = this->length();
  int32_t i = oldLength, length;

  // first cut off trailing white space
  for(;;) {
    length = i;
    if(i <= 0) {
      break;
    }
    U16_PREV(array, 0, i, c);
    if(!(c == 0x20 || u_isWhitespace(c))) {
      break;
    }
  }
  if(length < oldLength) {
    setLength(length);
  }

  // find leading white space
  int32_t start;
  i = 0;
  for(;;) {
    start = i;
    if(i >= length) {
      break;
    }
    U16_NEXT(array, i, length, c);
    if(!(c == 0x20 || u_isWhitespace(c))) {
      break;
    }
  }

  // move string forward over leading white space
  if(start > 0) {
    doReplace(0, start, 0, 0, 0);
  }

  return *this;
}

U_NAMESPACE_END
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// *******************************************************************************
// *   Copyright (C) 2011, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *******************************************************************************
// *   file name:  unistr_titlecase_brkiter.cpp
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:2
// *
// *   created on: 2011may30
// *   created by: Markus W. Scherer
// *
// *   Titlecasing functions that are based on BreakIterator
// *   were moved here to break dependency cycles among parts of the common library.
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/brkiter.h"
// #include "unicode/locid.h"
// #include "unicode/ucasemap.h"
// #include "unicode/unistr.h"
// #include "ucasemap_imp.h"

// U_NAMESPACE_BEGIN

// UnicodeString &
// UnicodeString::toTitle(BreakIterator *iter) {
//     return toTitle(iter, Locale::getDefault(), 0);
// }

// UnicodeString &
// UnicodeString::toTitle(BreakIterator *iter, const Locale &locale) {
//     return toTitle(iter, locale, 0);
// }

// UnicodeString &
// UnicodeString::toTitle(BreakIterator *iter, const Locale &locale, uint32_t options) {
//     LocalPointer<BreakIterator> ownedIter;
//     UErrorCode errorCode = U_ZERO_ERROR;
//     iter = ustrcase_getTitleBreakIterator(&locale, "", options, iter, ownedIter, errorCode);
//     if (iter == nullptr) {
//         setToBogus();
//         return *this;
//     }
//     caseMap(ustrcase_getCaseLocale(locale.getBaseName()), options, iter, ustrcase_internalToTitle);
//     return *this;
// }

// U_NAMESPACE_END

// #endif  // !UCONFIG_NO_BREAK_ITERATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (c) 1996-2014, International Business Machines
* Corporation and others. All Rights Reserved.
******************************************************************************
* File unorm.cpp
*
* Created by: Vladimir Weinstein 12052000
*
* Modification history :
*
* Date        Name        Description
* 02/01/01    synwee      Added normalization quickcheck enum and method.
* 02/12/01    synwee      Commented out quickcheck util api has been approved
*                         Added private method for doing FCD checks
* 02/23/01    synwee      Modified quickcheck and checkFCE to run through 
*                         string for codepoints < 0x300 for the normalization 
*                         mode NFC.
* 05/25/01+   Markus Scherer total rewrite, implement all normalization here
*                         instead of just wrappers around normlzr.cpp,
*                         load unorm.dat, support Unicode 3.1 with
*                         supplementary code points, etc.
* 2009-nov..2010-jan  Markus Scherer  total rewrite, new Normalizer2 API & code
*/



#if !UCONFIG_NO_NORMALIZATION







// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2001-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  unormimp.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001may25
*   created by: Markus W. Scherer
*/




#if !UCONFIG_NO_NORMALIZATION



/*
 * The 2001-2010 implementation of the normalization code loads its data from
 * unorm.icu, which is generated with the gennorm tool.
 * The format of that file is described at the end of this file.
 */

/* norm32 value constants */
enum {
    /* quick check flags 0..3 set mean "no" for their forms */
    _NORM_QC_NFC=0x11,          /* no|maybe */
    _NORM_QC_NFKC=0x22,         /* no|maybe */
    _NORM_QC_NFD=4,             /* no */
    _NORM_QC_NFKD=8,            /* no */

    _NORM_QC_ANY_NO=0xf,

    /* quick check flags 4..5 mean "maybe" for their forms; test flags>=_NORM_QC_MAYBE */
    _NORM_QC_MAYBE=0x10,
    _NORM_QC_ANY_MAYBE=0x30,

    _NORM_QC_MASK=0x3f,

    _NORM_COMBINES_FWD=0x40,
    _NORM_COMBINES_BACK=0x80,
    _NORM_COMBINES_ANY=0xc0,

    _NORM_CC_SHIFT=8,           /* UnicodeData.txt combining class in bits 15..8 */
    _NORM_CC_MASK=0xff00,

    _NORM_EXTRA_SHIFT=16,               /* 16 bits for the index to UChars and other extra data */
    _NORM_EXTRA_INDEX_TOP=0xfc00,       /* start of surrogate specials after shift */

    _NORM_EXTRA_SURROGATE_MASK=0x3ff,
    _NORM_EXTRA_SURROGATE_TOP=0x3f0,    /* hangul etc. */

    _NORM_EXTRA_HANGUL=_NORM_EXTRA_SURROGATE_TOP,
    _NORM_EXTRA_JAMO_L,
    _NORM_EXTRA_JAMO_V,
    _NORM_EXTRA_JAMO_T
};

/* norm32 value constants using >16 bits */
#define _NORM_MIN_SPECIAL       0xfc000000
#define _NORM_SURROGATES_TOP    0xfff00000
#define _NORM_MIN_HANGUL        0xfff00000
#define _NORM_MIN_JAMO_V        0xfff20000
#define _NORM_JAMO_V_TOP        0xfff30000

/* value constants for auxTrie */
enum {
    _NORM_AUX_COMP_EX_SHIFT=10,
    _NORM_AUX_UNSAFE_SHIFT=11,
    _NORM_AUX_NFC_SKIPPABLE_F_SHIFT=12
};

#define _NORM_AUX_MAX_FNC           ((int32_t)1<<_NORM_AUX_COMP_EX_SHIFT)

#define _NORM_AUX_FNC_MASK          (uint32_t)(_NORM_AUX_MAX_FNC-1)
#define _NORM_AUX_COMP_EX_MASK      ((uint32_t)1<<_NORM_AUX_COMP_EX_SHIFT)
#define _NORM_AUX_UNSAFE_MASK       ((uint32_t)1<<_NORM_AUX_UNSAFE_SHIFT)
#define _NORM_AUX_NFC_SKIP_F_MASK   ((uint32_t)1<<_NORM_AUX_NFC_SKIPPABLE_F_SHIFT)

/* canonStartSets[0..31] contains indexes for what is in the array */
enum {
    _NORM_SET_INDEX_CANON_SETS_LENGTH,      /* number of uint16_t in canonical starter sets */
    _NORM_SET_INDEX_CANON_BMP_TABLE_LENGTH, /* number of uint16_t in the BMP search table (contains pairs) */
    _NORM_SET_INDEX_CANON_SUPP_TABLE_LENGTH,/* number of uint16_t in the supplementary search table (contains triplets) */

    /* from formatVersion 2.3: */
    _NORM_SET_INDEX_NX_CJK_COMPAT_OFFSET,   /* uint16_t offset from canonStartSets[0] to the
                                               exclusion set for CJK compatibility characters */
    _NORM_SET_INDEX_NX_UNICODE32_OFFSET,    /* uint16_t offset from canonStartSets[0] to the
                                               exclusion set for Unicode 3.2 characters */
    _NORM_SET_INDEX_NX_RESERVED_OFFSET,     /* uint16_t offset from canonStartSets[0] to the
                                               end of the previous exclusion set */

    _NORM_SET_INDEX_TOP=32                  /* changing this requires a new formatVersion */
};

/* more constants for canonical starter sets */

/* 14 bit indexes to canonical USerializedSets */
#define _NORM_MAX_CANON_SETS            0x4000

/* single-code point BMP sets are encoded directly in the search table except if result=0x4000..0x7fff */
#define _NORM_CANON_SET_BMP_MASK        0xc000
#define _NORM_CANON_SET_BMP_IS_INDEX    0x4000

/* indexes[] value names */
enum {
    _NORM_INDEX_TRIE_SIZE,              /* number of bytes in normalization trie */
    _NORM_INDEX_UCHAR_COUNT,            /* number of UChars in extra data */

    _NORM_INDEX_COMBINE_DATA_COUNT,     /* number of uint16_t words for combining data */
    _NORM_INDEX_COMBINE_FWD_COUNT,      /* number of code points that combine forward */
    _NORM_INDEX_COMBINE_BOTH_COUNT,     /* number of code points that combine forward and backward */
    _NORM_INDEX_COMBINE_BACK_COUNT,     /* number of code points that combine backward */

    _NORM_INDEX_MIN_NFC_NO_MAYBE,       /* first code point with quick check NFC NO/MAYBE */
    _NORM_INDEX_MIN_NFKC_NO_MAYBE,      /* first code point with quick check NFKC NO/MAYBE */
    _NORM_INDEX_MIN_NFD_NO_MAYBE,       /* first code point with quick check NFD NO/MAYBE */
    _NORM_INDEX_MIN_NFKD_NO_MAYBE,      /* first code point with quick check NFKD NO/MAYBE */

    _NORM_INDEX_FCD_TRIE_SIZE,          /* number of bytes in FCD trie */

    _NORM_INDEX_AUX_TRIE_SIZE,          /* number of bytes in the auxiliary trie */
    _NORM_INDEX_CANON_SET_COUNT,        /* number of uint16_t in the array of serialized USet */

    _NORM_INDEX_TOP=32                  /* changing this requires a new formatVersion */
};

enum {
    /* FCD check: everything below this code point is known to have a 0 lead combining class */
    _NORM_MIN_WITH_LEAD_CC=0x300
};

enum {
    /**
     * Bit 7 of the length byte for a decomposition string in extra data is
     * a flag indicating whether the decomposition string is
     * preceded by a 16-bit word with the leading and trailing cc
     * of the decomposition (like for A-umlaut);
     * if not, then both cc's are zero (like for compatibility ideographs).
     */
    _NORM_DECOMP_FLAG_LENGTH_HAS_CC=0x80,
    /**
     * Bits 6..0 of the length byte contain the actual length.
     */
    _NORM_DECOMP_LENGTH_MASK=0x7f
};

/** Constants for options flags for normalization. */
enum {
    /** Options bit 0, do not decompose Hangul syllables. */
    UNORM_NX_HANGUL=1,
    /** Options bit 1, do not decompose CJK compatibility characters. */
    UNORM_NX_CJK_COMPAT=2
};

/**
 * Description of the format of unorm.icu version 2.3.
 *
 * Main change from version 1 to version 2:
 * Use of new, common UTrie instead of normalization-specific tries.
 * Change to version 2.1: add third/auxiliary trie with associated data.
 * Change to version 2.2: add skippable (f) flag data (_NORM_AUX_NFC_SKIP_F_MASK).
 * Change to version 2.3: add serialized sets for normalization exclusions
 *                        stored inside canonStartSets[]
 *
 * For more details of how to use the data structures see the code
 * in unorm.cpp (runtime normalization code) and
 * in gennorm.c and gennorm/store.c (build-time data generation).
 *
 * For the serialized format of UTrie see utrie.c/UTrieHeader.
 *
 * - Overall partition
 *
 * unorm.dat customarily begins with a UDataInfo structure, see udata.h and .c.
 * After that there are the following structures:
 *
 * int32_t indexes[_NORM_INDEX_TOP];            -- _NORM_INDEX_TOP=32, see enum in this file
 *
 * UTrie normTrie;                              -- size in bytes=indexes[_NORM_INDEX_TRIE_SIZE]
 * 
 * uint16_t extraData[extraDataTop];            -- extraDataTop=indexes[_NORM_INDEX_UCHAR_COUNT]
 *                                                 extraData[0] contains the number of units for
 *                                                 FC_NFKC_Closure (formatVersion>=2.1)
 *
 * uint16_t combiningTable[combiningTableTop];  -- combiningTableTop=indexes[_NORM_INDEX_COMBINE_DATA_COUNT]
 *                                                 combiningTableTop may include one 16-bit padding unit
 *                                                 to make sure that fcdTrie is 32-bit-aligned
 *
 * UTrie fcdTrie;                               -- size in bytes=indexes[_NORM_INDEX_FCD_TRIE_SIZE]
 *
 * UTrie auxTrie;                               -- size in bytes=indexes[_NORM_INDEX_AUX_TRIE_SIZE]
 *
 * uint16_t canonStartSets[canonStartSetsTop]   -- canonStartSetsTop=indexes[_NORM_INDEX_CANON_SET_COUNT]
 *                                                 serialized USets and binary search tables, see below
 *
 *
 * The indexes array contains lengths and sizes of the following arrays and structures
 * as well as the following values:
 *  indexes[_NORM_INDEX_COMBINE_FWD_COUNT]=combineFwdTop
 *      -- one more than the highest combining index computed for forward-only-combining characters
 *  indexes[_NORM_INDEX_COMBINE_BOTH_COUNT]=combineBothTop-combineFwdTop
 *      -- number of combining indexes computed for both-ways-combining characters
 *  indexes[_NORM_INDEX_COMBINE_BACK_COUNT]=combineBackTop-combineBothTop
 *      -- number of combining indexes computed for backward-only-combining characters
 *
 *  indexes[_NORM_INDEX_MIN_NF*_NO_MAYBE] (where *={ C, D, KC, KD })
 *      -- first code point with a quick check NF* value of NO/MAYBE
 *
 *
 * - Tries
 *
 * The main structures are two UTrie tables ("compact arrays"),
 * each with one index array and one data array.
 * See utrie.h and utrie.c.
 *
 *
 * - Tries in unorm.dat
 *
 * The first trie (normTrie above)
 * provides data for the NF* quick checks and normalization.
 * The second trie (fcdTrie above) provides data just for FCD checks.
 *
 *
 * - norm32 data words from the first trie
 *
 * The norm32Table contains one 32-bit word "norm32" per code point.
 * It contains the following bit fields:
 * 31..16   extra data index, _NORM_EXTRA_SHIFT is used to shift this field down
 *          if this index is <_NORM_EXTRA_INDEX_TOP then it is an index into
 *              extraData[] where variable-length normalization data for this
 *              code point is found
 *          if this index is <_NORM_EXTRA_INDEX_TOP+_NORM_EXTRA_SURROGATE_TOP
 *              then this is a norm32 for a leading surrogate, and the index
 *              value is used together with the following trailing surrogate
 *              code unit in the second trie access
 *          if this index is >=_NORM_EXTRA_INDEX_TOP+_NORM_EXTRA_SURROGATE_TOP
 *              then this is a norm32 for a "special" character,
 *              i.e., the character is a Hangul syllable or a Jamo
 *              see _NORM_EXTRA_HANGUL etc.
 *          generally, instead of extracting this index from the norm32 and
 *              comparing it with the above constants,
 *              the normalization code compares the entire norm32 value
 *              with _NORM_MIN_SPECIAL, _NORM_SURROGATES_TOP, _NORM_MIN_HANGUL etc.
 *
 * 15..8    combining class (cc) according to UnicodeData.txt
 *
 *  7..6    _NORM_COMBINES_ANY flags, used in composition to see if a character
 *              combines with any following or preceding character(s)
 *              at all
 *     7    _NORM_COMBINES_BACK
 *     6    _NORM_COMBINES_FWD
 *
 *  5..0    quick check flags, set for "no" or "maybe", with separate flags for
 *              each normalization form
 *              the higher bits are "maybe" flags; for NF*D there are no such flags
 *              the lower bits are "no" flags for all forms, in the same order
 *              as the "maybe" flags,
 *              which is (MSB to LSB): NFKD NFD NFKC NFC
 *  5..4    _NORM_QC_ANY_MAYBE
 *  3..0    _NORM_QC_ANY_NO
 *              see further related constants
 *
 *
 * - Extra data per code point
 *
 * "Extra data" is referenced by the index in norm32.
 * It is variable-length data. It is only present, and only those parts
 * of it are, as needed for a given character.
 * The norm32 extra data index is added to the beginning of extraData[]
 * to get to a vector of 16-bit words with data at the following offsets:
 *
 * [-1]     Combining index for composition.
 *              Stored only if norm32&_NORM_COMBINES_ANY .
 * [0]      Lengths of the canonical and compatibility decomposition strings.
 *              Stored only if there are decompositions, i.e.,
 *              if norm32&(_NORM_QC_NFD|_NORM_QC_NFKD)
 *          High byte: length of NFKD, or 0 if none
 *          Low byte: length of NFD, or 0 if none
 *          Each length byte also has another flag:
 *              Bit 7 of a length byte is set if there are non-zero
 *              combining classes (cc's) associated with the respective
 *              decomposition. If this flag is set, then the decomposition
 *              is preceded by a 16-bit word that contains the
 *              leading and trailing cc's.
 *              Bits 6..0 of a length byte are the length of the
 *              decomposition string, not counting the cc word.
 * [1..n]   NFD
 * [n+1..]  NFKD
 *
 * Each of the two decompositions consists of up to two parts:
 * - The 16-bit words with the leading and trailing cc's.
 *   This is only stored if bit 7 of the corresponding length byte
 *   is set. In this case, at least one of the cc's is not zero.
 *   High byte: leading cc==cc of the first code point in the decomposition string
 *   Low byte: trailing cc==cc of the last code point in the decomposition string
 * - The decomposition string in UTF-16, with length code units.
 *
 *
 * - Combining indexes and combiningTable[]
 *
 * Combining indexes are stored at the [-1] offset of the extra data
 * if the character combines forward or backward with any other characters.
 * They are used for (re)composition in NF*C.
 * Values of combining indexes are arranged according to whether a character
 * combines forward, backward, or both ways:
 *    forward-only < both ways < backward-only
 *
 * The index values for forward-only and both-ways combining characters
 * are indexes into the combiningTable[].
 * The index values for backward-only combining characters are simply
 * incremented from the preceding index values to be unique.
 *
 * In the combiningTable[], a variable-length list
 * of variable-length (back-index, code point) pair entries is stored
 * for each forward-combining character.
 *
 * These back-indexes are the combining indexes of both-ways or backward-only
 * combining characters that the forward-combining character combines with.
 *
 * Each list is sorted in ascending order of back-indexes.
 * Each list is terminated with the last back-index having bit 15 set.
 *
 * Each pair (back-index, code point) takes up either 2 or 3
 * 16-bit words.
 * The first word of a list entry is the back-index, with its bit 15 set if
 * this is the last pair in the list.
 *
 * The second word contains flags in bits 15..13 that determine
 * if there is a third word and how the combined character is encoded:
 * 15   set if there is a third word in this list entry
 * 14   set if the result is a supplementary character
 * 13   set if the result itself combines forward
 *
 * According to these bits 15..14 of the second word,
 * the result character is encoded as follows:
 * 00 or 01 The result is <=0x1fff and stored in bits 12..0 of
 *          the second word.
 * 10       The result is 0x2000..0xffff and stored in the third word.
 *          Bits 12..0 of the second word are not used.
 * 11       The result is a supplementary character.
 *          Bits 9..0 of the leading surrogate are in bits 9..0 of
 *          the second word.
 *          Add 0xd800 to these bits to get the complete surrogate.
 *          Bits 12..10 of the second word are not used.
 *          The trailing surrogate is stored in the third word.
 *
 *
 * - FCD trie
 *
 * The FCD trie is very simple.
 * It is a folded trie with 16-bit data words.
 * In each word, the high byte contains the leading cc of the character,
 * and the low byte contains the trailing cc of the character.
 * These cc's are the cc's of the first and last code points in the
 * canonical decomposition of the character.
 *
 * Since all 16 bits are used for cc's, lead surrogates must be tested
 * by checking the code unit instead of the trie data.
 * This is done only if the 16-bit data word is not zero.
 * If the code unit is a leading surrogate and the data word is not zero,
 * then instead of cc's it contains the offset for the second trie lookup.
 *
 *
 * - Auxiliary trie and data
 *
 * The auxiliary 16-bit trie contains data for additional properties.
 * Bits
 * 15..13   reserved
 *     12   not NFC_Skippable (f) (formatVersion>=2.2)
 *     11   flag: not a safe starter for canonical closure
 *     10   composition exclusion
 *  9.. 0   index into extraData[] to FC_NFKC_Closure string
 *          (not for lead surrogate),
 *          or lead surrogate offset (for lead surrogate, if 9..0 not zero)
 *
 * - FC_NFKC_Closure strings in extraData[]
 *
 * Strings are either stored as a single code unit or as the length
 * followed by that many units.
 *   const UChar *s=extraData+(index from auxTrie data bits 9..0);
 *   int32_t length;
 *   if(*s<0xff00) {
 *     // s points to the single-unit string
 *     length=1;
 *   } else {
 *     length=*s&0xff;
 *     ++s;
 *   }
 *
 * Conditions for "NF* Skippable" from Mark Davis' com.ibm.text.UCD.NFSkippable:
 * (used in NormalizerTransliterator)
 *
 * A skippable character is
 * a) unassigned, or ALL of the following:
 * b) of combining class 0.
 * c) not decomposed by this normalization form.
 * AND if NFC or NFKC,
 * d) can never compose with a previous character.
 * e) can never compose with a following character.
 * f) can never change if another character is added.
 *    Example: a-breve might satisfy all but f, but if you
 *    add an ogonek it changes to a-ogonek + breve
 *
 * a)..e) must be tested from norm32.
 * Since f) is more complicated, the (not-)NFC_Skippable flag (f) is built
 * into the auxiliary trie.
 * The same bit is used for NFC and NFKC; (c) differs for them.
 * As usual, we build the "not skippable" flags so that unassigned
 * code points get a 0 bit.
 * This bit is only valid after (a)..(e) test FALSE; test NFD_NO before (f) as well.
 * Test Hangul LV syllables entirely in code.
 *
 *
 * - structure inside canonStartSets[]
 *
 * This array maps from code points c to sets of code points (USerializedSet).
 * The result sets are the code points whose canonical decompositions start
 * with c.
 *
 * canonStartSets[] contains the following sub-arrays:
 *
 * indexes[_NORM_SET_INDEX_TOP]
 *   - contains lengths of sub-arrays etc.
 *
 * startSets[indexes[_NORM_SET_INDEX_CANON_SETS_LENGTH]-_NORM_SET_INDEX_TOP]
 *   - contains serialized sets (USerializedSet) of canonical starters for
 *     enumerating canonically equivalent strings
 *     indexes[_NORM_SET_INDEX_CANON_SETS_LENGTH] includes _NORM_SET_INDEX_TOP
 *     for details about the structure see uset.c
 *
 * bmpTable[indexes[_NORM_SET_INDEX_CANON_BMP_TABLE_LENGTH]]
 *   - a sorted search table for BMP code points whose results are
 *     either indexes to USerializedSets or single code points for
 *     single-code point sets;
 *     each entry is a pair of { code point, result } with result=(binary) yy xxxxxx xxxxxxxx
 *     if yy==01 then there is a USerializedSet at canonStartSets+x
 *     else build a USerializedSet with result as the single code point
 *
 * suppTable[indexes[_NORM_SET_INDEX_CANON_SUPP_TABLE_LENGTH]]
 *   - a sorted search table for supplementary code points whose results are
 *     either indexes to USerializedSets or single code points for
 *     single-code point sets;
 *     each entry is a triplet of { high16(cp), low16(cp), result }
 *     each code point's high-word may contain extra data in bits 15..5:
 *     if the high word has bit 15 set, then build a set with a single code point
 *     which is (((high16(cp)&0x1f00)<<8)|result;
 *     else there is a USerializedSet at canonStartSets+result
 *
 * FormatVersion 2.3 adds 2 serialized sets for normalization exclusions.
 * They are stored in the data file so that the runtime normalization code need
 * not depend on other properties and their data and implementation files.
 * The _NORM_SET_INDEX_NX_..._OFFSET offsets in the canonStartSets index table
 * give the location for each set.
 * There is no set stored for UNORM_NX_HANGUL because it's trivial to create
 * without using properties.
 *
 * Set contents:
 *
 * _NORM_SET_INDEX_NX_CJK_COMPAT_OFFSET (for UNORM_NX_CJK_COMPAT)
 *     [[:Ideographic:]&[:NFD_QC=No:]]
 *     =[CJK Ideographs]&[has canonical decomposition]
 *
 * _NORM_SET_INDEX_NX_UNICODE32_OFFSET (for UNORM_UNICODE_3_2)
 *     [:^Age=3.2:]
 *     =set with all code points that were not designated by the specified Unicode version
 *
 * _NORM_SET_INDEX_NX_RESERVED_OFFSET
 *     This is an offset that points to where the next, future set would start.
 *     Currently it indicates where the previous set ends, and thus its length.
 *     The name for this enum constant may in the future be applied to different
 *     index slots. In order to get the limit of a set, use its index slot and
 *     the immediately following one regardless of that one's enum name.
 */

#endif /* #if !UCONFIG_NO_NORMALIZATION */





U_NAMESPACE_USE

/* quick check functions ---------------------------------------------------- */

U_CAPI UNormalizationCheckResult U_EXPORT2
unorm_quickCheck(const UChar *src,
                 int32_t srcLength, 
                 UNormalizationMode mode,
                 UErrorCode *pErrorCode) {
    const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, *pErrorCode);
    return unorm2_quickCheck((const UNormalizer2 *)n2, src, srcLength, pErrorCode);
}

U_CAPI UNormalizationCheckResult U_EXPORT2
unorm_quickCheckWithOptions(const UChar *src, int32_t srcLength, 
                            UNormalizationMode mode, int32_t options,
                            UErrorCode *pErrorCode) {
    const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, *pErrorCode);
    if(options&UNORM_UNICODE_3_2) {
        FilteredNormalizer2 fn2(*n2, *uniset_getUnicode32Instance(*pErrorCode));
        return unorm2_quickCheck(
            reinterpret_cast<const UNormalizer2 *>(static_cast<Normalizer2 *>(&fn2)),
            src, srcLength, pErrorCode);
    } else {
        return unorm2_quickCheck((const UNormalizer2 *)n2, src, srcLength, pErrorCode);
    }
}

U_CAPI UBool U_EXPORT2
unorm_isNormalized(const UChar *src, int32_t srcLength,
                   UNormalizationMode mode,
                   UErrorCode *pErrorCode) {
    const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, *pErrorCode);
    return unorm2_isNormalized((const UNormalizer2 *)n2, src, srcLength, pErrorCode);
}

U_CAPI UBool U_EXPORT2
unorm_isNormalizedWithOptions(const UChar *src, int32_t srcLength,
                              UNormalizationMode mode, int32_t options,
                              UErrorCode *pErrorCode) {
    const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, *pErrorCode);
    if(options&UNORM_UNICODE_3_2) {
        FilteredNormalizer2 fn2(*n2, *uniset_getUnicode32Instance(*pErrorCode));
        return unorm2_isNormalized(
            reinterpret_cast<const UNormalizer2 *>(static_cast<Normalizer2 *>(&fn2)),
            src, srcLength, pErrorCode);
    } else {
        return unorm2_isNormalized((const UNormalizer2 *)n2, src, srcLength, pErrorCode);
    }
}

/* normalize() API ---------------------------------------------------------- */

/** Public API for normalizing. */
U_CAPI int32_t U_EXPORT2
unorm_normalize(const UChar *src, int32_t srcLength,
                UNormalizationMode mode, int32_t options,
                UChar *dest, int32_t destCapacity,
                UErrorCode *pErrorCode) {
    const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, *pErrorCode);
    if(options&UNORM_UNICODE_3_2) {
        FilteredNormalizer2 fn2(*n2, *uniset_getUnicode32Instance(*pErrorCode));
        return unorm2_normalize(
            reinterpret_cast<const UNormalizer2 *>(static_cast<Normalizer2 *>(&fn2)),
            src, srcLength, dest, destCapacity, pErrorCode);
    } else {
        return unorm2_normalize((const UNormalizer2 *)n2,
            src, srcLength, dest, destCapacity, pErrorCode);
    }
}


/* iteration functions ------------------------------------------------------ */

static int32_t
_iterate(UCharIterator *src, UBool forward,
              UChar *dest, int32_t destCapacity,
              const Normalizer2 *n2,
              UBool doNormalize, UBool *pNeededToNormalize,
              UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(destCapacity<0 || (dest==NULL && destCapacity>0) || src==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if(pNeededToNormalize!=NULL) {
        *pNeededToNormalize=FALSE;
    }
    if(!(forward ? src->hasNext(src) : src->hasPrevious(src))) {
        return u_terminateUChars(dest, destCapacity, 0, pErrorCode);
    }

    UnicodeString buffer;
    UChar32 c;
    if(forward) {
        /* get one character and ignore its properties */
        buffer.append(uiter_next32(src));
        /* get all following characters until we see a boundary */
        while((c=uiter_next32(src))>=0) {
            if(n2->hasBoundaryBefore(c)) {
                /* back out the latest movement to stop at the boundary */
                src->move(src, -U16_LENGTH(c), UITER_CURRENT);
                break;
            } else {
                buffer.append(c);
            }
        }
    } else {
        while((c=uiter_previous32(src))>=0) {
            /* always write this character to the front of the buffer */
            buffer.insert(0, c);
            /* stop if this just-copied character is a boundary */
            if(n2->hasBoundaryBefore(c)) {
                break;
            }
        }
    }

    UnicodeString destString(dest, 0, destCapacity);
    if(buffer.length()>0 && doNormalize) {
        n2->normalize(buffer, destString, *pErrorCode).extract(dest, destCapacity, *pErrorCode);
        if(pNeededToNormalize!=NULL && U_SUCCESS(*pErrorCode)) {
            *pNeededToNormalize= destString!=buffer;
        }
        return destString.length();
    } else {
        /* just copy the source characters */
        return buffer.extract(dest, destCapacity, *pErrorCode);
    }
}

static int32_t
unorm_iterate(UCharIterator *src, UBool forward,
              UChar *dest, int32_t destCapacity,
              UNormalizationMode mode, int32_t options,
              UBool doNormalize, UBool *pNeededToNormalize,
              UErrorCode *pErrorCode) {
    const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, *pErrorCode);
    if(options&UNORM_UNICODE_3_2) {
        const UnicodeSet *uni32 = uniset_getUnicode32Instance(*pErrorCode);
        if(U_FAILURE(*pErrorCode)) {
            return 0;
        }
        FilteredNormalizer2 fn2(*n2, *uni32);
        return _iterate(src, forward, dest, destCapacity,
            &fn2, doNormalize, pNeededToNormalize, pErrorCode);
    }
    return _iterate(src, forward, dest, destCapacity,
            n2, doNormalize, pNeededToNormalize, pErrorCode);
}

U_CAPI int32_t U_EXPORT2
unorm_previous(UCharIterator *src,
               UChar *dest, int32_t destCapacity,
               UNormalizationMode mode, int32_t options,
               UBool doNormalize, UBool *pNeededToNormalize,
               UErrorCode *pErrorCode) {
    return unorm_iterate(src, FALSE,
                         dest, destCapacity,
                         mode, options,
                         doNormalize, pNeededToNormalize,
                         pErrorCode);
}

U_CAPI int32_t U_EXPORT2
unorm_next(UCharIterator *src,
           UChar *dest, int32_t destCapacity,
           UNormalizationMode mode, int32_t options,
           UBool doNormalize, UBool *pNeededToNormalize,
           UErrorCode *pErrorCode) {
    return unorm_iterate(src, TRUE,
                         dest, destCapacity,
                         mode, options,
                         doNormalize, pNeededToNormalize,
                         pErrorCode);
}

/* Concatenation of normalized strings -------------------------------------- */

static int32_t
_concatenate(const UChar *left, int32_t leftLength,
                  const UChar *right, int32_t rightLength,
                  UChar *dest, int32_t destCapacity,
                  const Normalizer2 *n2,
                  UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(destCapacity<0 || (dest==NULL && destCapacity>0) ||
        left==NULL || leftLength<-1 || right==NULL || rightLength<-1) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* check for overlapping right and destination */
    if( dest!=NULL &&
        ((right>=dest && right<(dest+destCapacity)) ||
         (rightLength>0 && dest>=right && dest<(right+rightLength)))
    ) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* allow left==dest */
    UnicodeString destString;
    if(left==dest) {
        destString.setTo(dest, leftLength, destCapacity);
    } else {
        destString.setTo(dest, 0, destCapacity);
        destString.append(left, leftLength);
    }
    return n2->append(destString, UnicodeString(rightLength<0, right, rightLength), *pErrorCode).
           extract(dest, destCapacity, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
unorm_concatenate(const UChar *left, int32_t leftLength,
                  const UChar *right, int32_t rightLength,
                  UChar *dest, int32_t destCapacity,
                  UNormalizationMode mode, int32_t options,
                  UErrorCode *pErrorCode) {
    const Normalizer2 *n2=Normalizer2Factory::getInstance(mode, *pErrorCode);
    if(options&UNORM_UNICODE_3_2) {
        const UnicodeSet *uni32 = uniset_getUnicode32Instance(*pErrorCode);
        if(U_FAILURE(*pErrorCode)) {
            return 0;
        }
        FilteredNormalizer2 fn2(*n2, *uni32);
        return _concatenate(left, leftLength, right, rightLength,
            dest, destCapacity, &fn2, pErrorCode);
    }
    return _concatenate(left, leftLength, right, rightLength,
        dest, destCapacity, n2, pErrorCode);
}

#endif /* #if !UCONFIG_NO_NORMALIZATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2001-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  unormcmp.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004sep13
*   created by: Markus W. Scherer
*
*   unorm_compare() function moved here from unorm.cpp for better modularization.
*   Depends on both normalization and case folding.
*   Allows unorm.cpp to not depend on any character properties code.
*/



#if !UCONFIG_NO_NORMALIZATION









U_NAMESPACE_USE

/* compare canonically equivalent ------------------------------------------- */

/*
 * Compare two strings for canonical equivalence.
 * Further options include case-insensitive comparison and
 * code point order (as opposed to code unit order).
 *
 * In this function, canonical equivalence is optional as well.
 * If canonical equivalence is tested, then both strings must fulfill
 * the FCD check.
 *
 * Semantically, this is equivalent to
 *   strcmp[CodePointOrder](NFD(foldCase(s1)), NFD(foldCase(s2)))
 * where code point order, NFD and foldCase are all optional.
 *
 * String comparisons almost always yield results before processing both strings
 * completely.
 * They are generally more efficient working incrementally instead of
 * performing the sub-processing (strlen, normalization, case-folding)
 * on the entire strings first.
 *
 * It is also unnecessary to not normalize identical characters.
 *
 * This function works in principle as follows:
 *
 * loop {
 *   get one code unit c1 from s1 (-1 if end of source)
 *   get one code unit c2 from s2 (-1 if end of source)
 *
 *   if(either string finished) {
 *     return result;
 *   }
 *   if(c1==c2) {
 *     continue;
 *   }
 *
 *   // c1!=c2
 *   try to decompose/case-fold c1/c2, and continue if one does;
 *
 *   // still c1!=c2 and neither decomposes/case-folds, return result
 *   return c1-c2;
 * }
 *
 * When a character decomposes, then the pointer for that source changes to
 * the decomposition, pushing the previous pointer onto a stack.
 * When the end of the decomposition is reached, then the code unit reader
 * pops the previous source from the stack.
 * (Same for case-folding.)
 *
 * This is complicated further by operating on variable-width UTF-16.
 * The top part of the loop works on code units, while lookups for decomposition
 * and case-folding need code points.
 * Code points are assembled after the equality/end-of-source part.
 * The source pointer is only advanced beyond all code units when the code point
 * actually decomposes/case-folds.
 *
 * If we were on a trail surrogate unit when assembling a code point,
 * and the code point decomposes/case-folds, then the decomposition/folding
 * result must be compared with the part of the other string that corresponds to
 * this string's lead surrogate.
 * Since we only assemble a code point when hitting a trail unit when the
 * preceding lead units were identical, we back up the other string by one unit
 * in such a case.
 *
 * The optional code point order comparison at the end works with
 * the same fix-up as the other code point order comparison functions.
 * See ustring.c and the comment near the end of this function.
 *
 * Assumption: A decomposition or case-folding result string never contains
 * a single surrogate. This is a safe assumption in the Unicode Standard.
 * Therefore, we do not need to check for surrogate pairs across
 * decomposition/case-folding boundaries.
 *
 * Further assumptions (see verifications tstnorm.cpp):
 * The API function checks for FCD first, while the core function
 * first case-folds and then decomposes. This requires that case-folding does not
 * un-FCD any strings.
 *
 * The API function may also NFD the input and turn off decomposition.
 * This requires that case-folding does not un-NFD strings either.
 *
 * TODO If any of the above two assumptions is violated,
 * then this entire code must be re-thought.
 * If this happens, then a simple solution is to case-fold both strings up front
 * and to turn off UNORM_INPUT_IS_FCD.
 * We already do this when not both strings are in FCD because makeFCD
 * would be a partial NFD before the case folding, which does not work.
 * Note that all of this is only a problem when case-folding _and_
 * canonical equivalence come together.
 * (Comments in unorm_compare() are more up to date than this TODO.)
 */

/* stack element for previous-level source/decomposition pointers */
struct unormcmp_CmpEquivLevel {
    const UChar *start, *s, *limit;
};
typedef struct unormcmp_CmpEquivLevel unormcmp_CmpEquivLevel;

/**
 * Internal option for unorm_cmpEquivFold() for decomposing.
 * If not set, just do strcasecmp().
 */
#define _COMPARE_EQUIV 0x80000

/* internal function */
static int32_t
unorm_cmpEquivFold(const UChar *s1, int32_t length1,
                   const UChar *s2, int32_t length2,
                   uint32_t options,
                   UErrorCode *pErrorCode) {
    const Normalizer2Impl *nfcImpl;

    /* current-level start/limit - s1/s2 as current */
    const UChar *start1, *start2, *limit1, *limit2;

    /* decomposition and case folding variables */
    const UChar *p;
    int32_t length;

    /* stacks of previous-level start/current/limit */
    unormcmp_CmpEquivLevel stack1[2], stack2[2];

    /* buffers for algorithmic decompositions */
    UChar decomp1[4], decomp2[4];

    /* case folding buffers, only use current-level start/limit */
    UChar fold1[UCASE_MAX_STRING_LENGTH+1], fold2[UCASE_MAX_STRING_LENGTH+1];

    /* track which is the current level per string */
    int32_t level1, level2;

    /* current code units, and code points for lookups */
    UChar32 c1, c2, cp1, cp2;

    /* no argument error checking because this itself is not an API */

    /*
     * assume that at least one of the options _COMPARE_EQUIV and U_COMPARE_IGNORE_CASE is set
     * otherwise this function must behave exactly as uprv_strCompare()
     * not checking for that here makes testing this function easier
     */

    /* normalization/properties data loaded? */
    if((options&_COMPARE_EQUIV)!=0) {
        nfcImpl=Normalizer2Factory::getNFCImpl(*pErrorCode);
    } else {
        nfcImpl=NULL;
    }
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* initialize */
    start1=s1;
    if(length1==-1) {
        limit1=NULL;
    } else {
        limit1=s1+length1;
    }

    start2=s2;
    if(length2==-1) {
        limit2=NULL;
    } else {
        limit2=s2+length2;
    }

    level1=level2=0;
    c1=c2=-1;

    /* comparison loop */
    for(;;) {
        /*
         * here a code unit value of -1 means "get another code unit"
         * below it will mean "this source is finished"
         */

        if(c1<0) {
            /* get next code unit from string 1, post-increment */
            for(;;) {
                if(s1==limit1 || ((c1=*s1)==0 && (limit1==NULL || (options&_STRNCMP_STYLE)))) {
                    if(level1==0) {
                        c1=-1;
                        break;
                    }
                } else {
                    ++s1;
                    break;
                }

                /* reached end of level buffer, pop one level */
                do {
                    --level1;
                    start1=stack1[level1].start;    /*Not uninitialized*/
                } while(start1==NULL);
                s1=stack1[level1].s;                /*Not uninitialized*/
                limit1=stack1[level1].limit;        /*Not uninitialized*/
            }
        }

        if(c2<0) {
            /* get next code unit from string 2, post-increment */
            for(;;) {
                if(s2==limit2 || ((c2=*s2)==0 && (limit2==NULL || (options&_STRNCMP_STYLE)))) {
                    if(level2==0) {
                        c2=-1;
                        break;
                    }
                } else {
                    ++s2;
                    break;
                }

                /* reached end of level buffer, pop one level */
                do {
                    --level2;
                    start2=stack2[level2].start;    /*Not uninitialized*/
                } while(start2==NULL);
                s2=stack2[level2].s;                /*Not uninitialized*/
                limit2=stack2[level2].limit;        /*Not uninitialized*/
            }
        }

        /*
         * compare c1 and c2
         * either variable c1, c2 is -1 only if the corresponding string is finished
         */
        if(c1==c2) {
            if(c1<0) {
                return 0;   /* c1==c2==-1 indicating end of strings */
            }
            c1=c2=-1;       /* make us fetch new code units */
            continue;
        } else if(c1<0) {
            return -1;      /* string 1 ends before string 2 */
        } else if(c2<0) {
            return 1;       /* string 2 ends before string 1 */
        }
        /* c1!=c2 && c1>=0 && c2>=0 */

        /* get complete code points for c1, c2 for lookups if either is a surrogate */
        cp1=c1;
        if(U_IS_SURROGATE(c1)) {
            UChar c;

            if(U_IS_SURROGATE_LEAD(c1)) {
                if(s1!=limit1 && U16_IS_TRAIL(c=*s1)) {
                    /* advance ++s1; only below if cp1 decomposes/case-folds */
                    cp1=U16_GET_SUPPLEMENTARY(c1, c);
                }
            } else /* isTrail(c1) */ {
                if(start1<=(s1-2) && U16_IS_LEAD(c=*(s1-2))) {
                    cp1=U16_GET_SUPPLEMENTARY(c, c1);
                }
            }
        }

        cp2=c2;
        if(U_IS_SURROGATE(c2)) {
            UChar c;

            if(U_IS_SURROGATE_LEAD(c2)) {
                if(s2!=limit2 && U16_IS_TRAIL(c=*s2)) {
                    /* advance ++s2; only below if cp2 decomposes/case-folds */
                    cp2=U16_GET_SUPPLEMENTARY(c2, c);
                }
            } else /* isTrail(c2) */ {
                if(start2<=(s2-2) && U16_IS_LEAD(c=*(s2-2))) {
                    cp2=U16_GET_SUPPLEMENTARY(c, c2);
                }
            }
        }

        /*
         * go down one level for each string
         * continue with the main loop as soon as there is a real change
         */

        if( level1==0 && (options&U_COMPARE_IGNORE_CASE) &&
            (length=ucase_toFullFolding((UChar32)cp1, &p, options))>=0
        ) {
            /* cp1 case-folds to the code point "length" or to p[length] */
            if(U_IS_SURROGATE(c1)) {
                if(U_IS_SURROGATE_LEAD(c1)) {
                    /* advance beyond source surrogate pair if it case-folds */
                    ++s1;
                } else /* isTrail(c1) */ {
                    /*
                     * we got a supplementary code point when hitting its trail surrogate,
                     * therefore the lead surrogate must have been the same as in the other string;
                     * compare this decomposition with the lead surrogate in the other string
                     * remember that this simulates bulk text replacement:
                     * the decomposition would replace the entire code point
                     */
                    --s2;
                    c2=*(s2-1);
                }
            }

            /* push current level pointers */
            stack1[0].start=start1;
            stack1[0].s=s1;
            stack1[0].limit=limit1;
            ++level1;

            /* copy the folding result to fold1[] */
            if(length<=UCASE_MAX_STRING_LENGTH) {
                u_memcpy(fold1, p, length);
            } else {
                int32_t i=0;
                U16_APPEND_UNSAFE(fold1, i, length);
                length=i;
            }

            /* set next level pointers to case folding */
            start1=s1=fold1;
            limit1=fold1+length;

            /* get ready to read from decomposition, continue with loop */
            c1=-1;
            continue;
        }

        if( level2==0 && (options&U_COMPARE_IGNORE_CASE) &&
            (length=ucase_toFullFolding((UChar32)cp2, &p, options))>=0
        ) {
            /* cp2 case-folds to the code point "length" or to p[length] */
            if(U_IS_SURROGATE(c2)) {
                if(U_IS_SURROGATE_LEAD(c2)) {
                    /* advance beyond source surrogate pair if it case-folds */
                    ++s2;
                } else /* isTrail(c2) */ {
                    /*
                     * we got a supplementary code point when hitting its trail surrogate,
                     * therefore the lead surrogate must have been the same as in the other string;
                     * compare this decomposition with the lead surrogate in the other string
                     * remember that this simulates bulk text replacement:
                     * the decomposition would replace the entire code point
                     */
                    --s1;
                    c1=*(s1-1);
                }
            }

            /* push current level pointers */
            stack2[0].start=start2;
            stack2[0].s=s2;
            stack2[0].limit=limit2;
            ++level2;

            /* copy the folding result to fold2[] */
            if(length<=UCASE_MAX_STRING_LENGTH) {
                u_memcpy(fold2, p, length);
            } else {
                int32_t i=0;
                U16_APPEND_UNSAFE(fold2, i, length);
                length=i;
            }

            /* set next level pointers to case folding */
            start2=s2=fold2;
            limit2=fold2+length;

            /* get ready to read from decomposition, continue with loop */
            c2=-1;
            continue;
        }

        if( level1<2 && (options&_COMPARE_EQUIV) &&
            0!=(p=nfcImpl->getDecomposition((UChar32)cp1, decomp1, length))
        ) {
            /* cp1 decomposes into p[length] */
            if(U_IS_SURROGATE(c1)) {
                if(U_IS_SURROGATE_LEAD(c1)) {
                    /* advance beyond source surrogate pair if it decomposes */
                    ++s1;
                } else /* isTrail(c1) */ {
                    /*
                     * we got a supplementary code point when hitting its trail surrogate,
                     * therefore the lead surrogate must have been the same as in the other string;
                     * compare this decomposition with the lead surrogate in the other string
                     * remember that this simulates bulk text replacement:
                     * the decomposition would replace the entire code point
                     */
                    --s2;
                    c2=*(s2-1);
                }
            }

            /* push current level pointers */
            stack1[level1].start=start1;
            stack1[level1].s=s1;
            stack1[level1].limit=limit1;
            ++level1;

            /* set empty intermediate level if skipped */
            if(level1<2) {
                stack1[level1++].start=NULL;
            }

            /* set next level pointers to decomposition */
            start1=s1=p;
            limit1=p+length;

            /* get ready to read from decomposition, continue with loop */
            c1=-1;
            continue;
        }

        if( level2<2 && (options&_COMPARE_EQUIV) &&
            0!=(p=nfcImpl->getDecomposition((UChar32)cp2, decomp2, length))
        ) {
            /* cp2 decomposes into p[length] */
            if(U_IS_SURROGATE(c2)) {
                if(U_IS_SURROGATE_LEAD(c2)) {
                    /* advance beyond source surrogate pair if it decomposes */
                    ++s2;
                } else /* isTrail(c2) */ {
                    /*
                     * we got a supplementary code point when hitting its trail surrogate,
                     * therefore the lead surrogate must have been the same as in the other string;
                     * compare this decomposition with the lead surrogate in the other string
                     * remember that this simulates bulk text replacement:
                     * the decomposition would replace the entire code point
                     */
                    --s1;
                    c1=*(s1-1);
                }
            }

            /* push current level pointers */
            stack2[level2].start=start2;
            stack2[level2].s=s2;
            stack2[level2].limit=limit2;
            ++level2;

            /* set empty intermediate level if skipped */
            if(level2<2) {
                stack2[level2++].start=NULL;
            }

            /* set next level pointers to decomposition */
            start2=s2=p;
            limit2=p+length;

            /* get ready to read from decomposition, continue with loop */
            c2=-1;
            continue;
        }

        /*
         * no decomposition/case folding, max level for both sides:
         * return difference result
         *
         * code point order comparison must not just return cp1-cp2
         * because when single surrogates are present then the surrogate pairs
         * that formed cp1 and cp2 may be from different string indexes
         *
         * example: { d800 d800 dc01 } vs. { d800 dc00 }, compare at second code units
         * c1=d800 cp1=10001 c2=dc00 cp2=10000
         * cp1-cp2>0 but c1-c2<0 and in fact in UTF-32 it is { d800 10001 } < { 10000 }
         *
         * therefore, use same fix-up as in ustring.c/uprv_strCompare()
         * except: uprv_strCompare() fetches c=*s while this functions fetches c=*s++
         * so we have slightly different pointer/start/limit comparisons here
         */

        if(c1>=0xd800 && c2>=0xd800 && (options&U_COMPARE_CODE_POINT_ORDER)) {
            /* subtract 0x2800 from BMP code points to make them smaller than supplementary ones */
            if(
                (c1<=0xdbff && s1!=limit1 && U16_IS_TRAIL(*s1)) ||
                (U16_IS_TRAIL(c1) && start1!=(s1-1) && U16_IS_LEAD(*(s1-2)))
            ) {
                /* part of a surrogate pair, leave >=d800 */
            } else {
                /* BMP code point - may be surrogate code point - make <d800 */
                c1-=0x2800;
            }

            if(
                (c2<=0xdbff && s2!=limit2 && U16_IS_TRAIL(*s2)) ||
                (U16_IS_TRAIL(c2) && start2!=(s2-1) && U16_IS_LEAD(*(s2-2)))
            ) {
                /* part of a surrogate pair, leave >=d800 */
            } else {
                /* BMP code point - may be surrogate code point - make <d800 */
                c2-=0x2800;
            }
        }

        return c1-c2;
    }
}

static
UBool _normalize(const Normalizer2 *n2, const UChar *s, int32_t length,
                UnicodeString &normalized, UErrorCode *pErrorCode) {
    UnicodeString str(length<0, s, length);

    // check if s fulfill the conditions
    int32_t spanQCYes=n2->spanQuickCheckYes(str, *pErrorCode);
    if (U_FAILURE(*pErrorCode)) {
        return FALSE;
    }
    /*
     * ICU 2.4 had a further optimization:
     * If both strings were not in FCD, then they were both NFD'ed,
     * and the _COMPARE_EQUIV option was turned off.
     * It is not entirely clear that this is valid with the current
     * definition of the canonical caseless match.
     * Therefore, ICU 2.6 removes that optimization.
     */
    if(spanQCYes<str.length()) {
        UnicodeString unnormalized=str.tempSubString(spanQCYes);
        normalized.setTo(FALSE, str.getBuffer(), spanQCYes);
        n2->normalizeSecondAndAppend(normalized, unnormalized, *pErrorCode);
        if (U_SUCCESS(*pErrorCode)) {
            return TRUE;
        }
    }
    return FALSE;
}

U_CAPI int32_t U_EXPORT2
unorm_compare(const UChar *s1, int32_t length1,
              const UChar *s2, int32_t length2,
              uint32_t options,
              UErrorCode *pErrorCode) {
    /* argument checking */
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(s1==0 || length1<-1 || s2==0 || length2<-1) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    UnicodeString fcd1, fcd2;
    int32_t normOptions=(int32_t)(options>>UNORM_COMPARE_NORM_OPTIONS_SHIFT);
    options|=_COMPARE_EQUIV;

    /*
     * UAX #21 Case Mappings, as fixed for Unicode version 4
     * (see Jitterbug 2021), defines a canonical caseless match as
     *
     * A string X is a canonical caseless match
     * for a string Y if and only if
     * NFD(toCasefold(NFD(X))) = NFD(toCasefold(NFD(Y)))
     *
     * For better performance, we check for FCD (or let the caller tell us that
     * both strings are in FCD) for the inner normalization.
     * BasicNormalizerTest::FindFoldFCDExceptions() makes sure that
     * case-folding preserves the FCD-ness of a string.
     * The outer normalization is then only performed by unorm_cmpEquivFold()
     * when there is a difference.
     *
     * Exception: When using the Turkic case-folding option, we do perform
     * full NFD first. This is because in the Turkic case precomposed characters
     * with 0049 capital I or 0069 small i fold differently whether they
     * are first decomposed or not, so an FCD check - a check only for
     * canonical order - is not sufficient.
     */
    if(!(options&UNORM_INPUT_IS_FCD) || (options&U_FOLD_CASE_EXCLUDE_SPECIAL_I)) {
        const Normalizer2 *n2;
        if(options&U_FOLD_CASE_EXCLUDE_SPECIAL_I) {
            n2=Normalizer2::getNFDInstance(*pErrorCode);
        } else {
            n2=Normalizer2Factory::getFCDInstance(*pErrorCode);
        }
        if (U_FAILURE(*pErrorCode)) {
            return 0;
        }

        if(normOptions&UNORM_UNICODE_3_2) {
            const UnicodeSet *uni32=uniset_getUnicode32Instance(*pErrorCode);
            FilteredNormalizer2 fn2(*n2, *uni32);
            if(_normalize(&fn2, s1, length1, fcd1, pErrorCode)) {
                s1=fcd1.getBuffer();
                length1=fcd1.length();
            }
            if(_normalize(&fn2, s2, length2, fcd2, pErrorCode)) {
                s2=fcd2.getBuffer();
                length2=fcd2.length();
            }
        } else {
            if(_normalize(n2, s1, length1, fcd1, pErrorCode)) {
                s1=fcd1.getBuffer();
                length1=fcd1.length();
            }
            if(_normalize(n2, s2, length2, fcd2, pErrorCode)) {
                s2=fcd2.getBuffer();
                length2=fcd2.length();
            }
        }
    }

    if(U_SUCCESS(*pErrorCode)) {
        return unorm_cmpEquivFold(s1, length1, s2, length2, options, pErrorCode);
    } else {
        return 0;
    }
}

#endif /* #if !UCONFIG_NO_NORMALIZATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2002-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  uobject.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2002jun26
*   created by: Markus W. Scherer
*/




U_NAMESPACE_BEGIN

#if U_OVERRIDE_CXX_ALLOCATION

/*
 * Default implementation of UMemory::new/delete
 * using uprv_malloc() and uprv_free().
 *
 * For testing, this is used together with a list of imported symbols to verify
 * that ICU is not using the global ::new and ::delete operators.
 *
 * These operators can be implemented like this or any other appropriate way
 * when customizing ICU for certain environments.
 * Whenever ICU is customized in binary incompatible ways please be sure
 * to use library name suffixes to distinguish such libraries from
 * the standard build.
 *
 * Instead of just modifying these C++ new/delete operators, it is usually best
 * to modify the uprv_malloc()/uprv_free()/uprv_realloc() functions in cmemory.c.
 *
 * Memory test on Windows/MSVC 6:
 * The global operators new and delete look as follows:
 *   04F 00000000 UNDEF  notype ()    External     | ??2@YAPAXI@Z (void * __cdecl operator new(unsigned int))
 *   03F 00000000 UNDEF  notype ()    External     | ??3@YAXPAX@Z (void __cdecl operator delete(void *))
 *
 * These lines are from output generated by the MSVC 6 tool dumpbin with
 * dumpbin /symbols *.obj
 *
 * ??2@YAPAXI@Z and ??3@YAXPAX@Z are the linker symbols in the .obj
 * files and are imported from msvcrtd.dll (in a debug build).
 *
 * Make sure that with the UMemory operators new and delete defined these two symbols
 * do not appear in the dumpbin /symbols output for the ICU libraries!
 *
 * If such a symbol appears in the output then look in the preceding lines in the output
 * for which file and function calls the global new or delete operator,
 * and replace with uprv_malloc/uprv_free.
 */

void * U_EXPORT2 UMemory::operator new(size_t size) U_NOEXCEPT {
    return uprv_malloc(size);
}

void U_EXPORT2 UMemory::operator delete(void *p) U_NOEXCEPT {
    if(p!=NULL) {
        uprv_free(p);
    }
}

void * U_EXPORT2 UMemory::operator new[](size_t size) U_NOEXCEPT {
    return uprv_malloc(size);
}

void U_EXPORT2 UMemory::operator delete[](void *p) U_NOEXCEPT {
    if(p!=NULL) {
        uprv_free(p);
    }
}

#if U_HAVE_DEBUG_LOCATION_NEW
void * U_EXPORT2 UMemory::operator new(size_t size, const char* /*file*/, int /*line*/) U_NOEXCEPT {
    return UMemory::operator new(size);
}

void U_EXPORT2 UMemory::operator delete(void* p, const char* /*file*/, int /*line*/) U_NOEXCEPT {
    UMemory::operator delete(p);
}
#endif /* U_HAVE_DEBUG_LOCATION_NEW */


#endif

UObject::~UObject() {}

UClassID UObject::getDynamicClassID() const { return NULL; }

U_NAMESPACE_END

U_NAMESPACE_USE

U_CAPI void U_EXPORT2
uprv_deleteUObject(void *obj) {
    delete static_cast<UObject *>(obj);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2002-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uprops.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2002feb24
*   created by: Markus W. Scherer
*
*   Implementations for mostly non-core Unicode character properties
*   stored in uprops.icu.
*
*   With the APIs implemented here, almost all properties files and
*   their associated implementation files are used from this file,
*   including those for normalization and case mappings.
*/
















// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// ulayout_props.h
// created: 2019feb12 Markus W. Scherer




// file definitions ------------------------------------------------------------

#define ULAYOUT_DATA_NAME "ulayout"
#define ULAYOUT_DATA_TYPE "icu"

// data format "Layo"
#define ULAYOUT_FMT_0 0x4c
#define ULAYOUT_FMT_1 0x61
#define ULAYOUT_FMT_2 0x79
#define ULAYOUT_FMT_3 0x6f

// indexes into indexes[]
enum {
    // Element 0 stores the length of the indexes[] array.
    ULAYOUT_IX_INDEXES_LENGTH,
    // Elements 1..7 store the tops of consecutive code point tries.
    // No trie is stored if the difference between two of these is less than 16.
    ULAYOUT_IX_INPC_TRIE_TOP,
    ULAYOUT_IX_INSC_TRIE_TOP,
    ULAYOUT_IX_VO_TRIE_TOP,
    ULAYOUT_IX_RESERVED_TOP,

    ULAYOUT_IX_TRIES_TOP = 7,

    ULAYOUT_IX_MAX_VALUES = 9,

    // Length of indexes[]. Multiple of 4 to 16-align the tries.
    ULAYOUT_IX_COUNT = 12
};

constexpr int32_t ULAYOUT_MAX_INPC_SHIFT = 24;
constexpr int32_t ULAYOUT_MAX_INSC_SHIFT = 16;
constexpr int32_t ULAYOUT_MAX_VO_SHIFT = 8;




U_NAMESPACE_USE

// Unicode text layout properties data -----------------------------------------

namespace {

icu::UInitOnce gLayoutInitOnce = U_INITONCE_INITIALIZER;
UDataMemory *gLayoutMemory = nullptr;

UCPTrie *gInpcTrie = nullptr;  // Indic_Positional_Category
UCPTrie *gInscTrie = nullptr;  // Indic_Syllabic_Category
UCPTrie *gVoTrie = nullptr;  // Vertical_Orientation

int32_t gMaxInpcValue = 0;
int32_t gMaxInscValue = 0;
int32_t gMaxVoValue = 0;

UBool U_CALLCONV uprops_cleanup() {
    udata_close(gLayoutMemory);
    gLayoutMemory = nullptr;

    ucptrie_close(gInpcTrie);
    gInpcTrie = nullptr;
    ucptrie_close(gInscTrie);
    gInscTrie = nullptr;
    ucptrie_close(gVoTrie);
    gVoTrie = nullptr;

    gMaxInpcValue = 0;
    gMaxInscValue = 0;
    gMaxVoValue = 0;

    gLayoutInitOnce.reset();
    return TRUE;
}

UBool U_CALLCONV
ulayout_isAcceptable(void * /*context*/,
                     const char * /* type */, const char * /*name*/,
                     const UDataInfo *pInfo) {
    return pInfo->size >= 20 &&
        pInfo->isBigEndian == U_IS_BIG_ENDIAN &&
        pInfo->charsetFamily == U_CHARSET_FAMILY &&
        pInfo->dataFormat[0] == ULAYOUT_FMT_0 &&
        pInfo->dataFormat[1] == ULAYOUT_FMT_1 &&
        pInfo->dataFormat[2] == ULAYOUT_FMT_2 &&
        pInfo->dataFormat[3] == ULAYOUT_FMT_3 &&
        pInfo->formatVersion[0] == 1;
}

// UInitOnce singleton initialization function
void U_CALLCONV ulayout_load(UErrorCode &errorCode) {
    gLayoutMemory = udata_openChoice(
        nullptr, ULAYOUT_DATA_TYPE, ULAYOUT_DATA_NAME,
        ulayout_isAcceptable, nullptr, &errorCode);
    if (U_FAILURE(errorCode)) { return; }

    const uint8_t *inBytes = (const uint8_t *)udata_getMemory(gLayoutMemory);
    const int32_t *inIndexes = (const int32_t *)inBytes;
    int32_t indexesLength = inIndexes[ULAYOUT_IX_INDEXES_LENGTH];
    if (indexesLength < 12) {
        errorCode = U_INVALID_FORMAT_ERROR;  // Not enough indexes.
        return;
    }
    int32_t offset = indexesLength * 4;
    int32_t top = inIndexes[ULAYOUT_IX_INPC_TRIE_TOP];
    int32_t trieSize = top - offset;
    if (trieSize >= 16) {
        gInpcTrie = ucptrie_openFromBinary(
            UCPTRIE_TYPE_ANY, UCPTRIE_VALUE_BITS_ANY,
            inBytes + offset, trieSize, nullptr, &errorCode);
    }
    offset = top;
    top = inIndexes[ULAYOUT_IX_INSC_TRIE_TOP];
    trieSize = top - offset;
    if (trieSize >= 16) {
        gInscTrie = ucptrie_openFromBinary(
            UCPTRIE_TYPE_ANY, UCPTRIE_VALUE_BITS_ANY,
            inBytes + offset, trieSize, nullptr, &errorCode);
    }
    offset = top;
    top = inIndexes[ULAYOUT_IX_VO_TRIE_TOP];
    trieSize = top - offset;
    if (trieSize >= 16) {
        gVoTrie = ucptrie_openFromBinary(
            UCPTRIE_TYPE_ANY, UCPTRIE_VALUE_BITS_ANY,
            inBytes + offset, trieSize, nullptr, &errorCode);
    }

    uint32_t maxValues = inIndexes[ULAYOUT_IX_MAX_VALUES];
    gMaxInpcValue = maxValues >> ULAYOUT_MAX_INPC_SHIFT;
    gMaxInscValue = (maxValues >> ULAYOUT_MAX_INSC_SHIFT) & 0xff;
    gMaxVoValue = (maxValues >> ULAYOUT_MAX_VO_SHIFT) & 0xff;

    ucln_common_registerCleanup(UCLN_COMMON_UPROPS, uprops_cleanup);
}

UBool ulayout_ensureData(UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return FALSE; }
    umtx_initOnce(gLayoutInitOnce, &ulayout_load, errorCode);
    return U_SUCCESS(errorCode);
}

UBool ulayout_ensureData() {
    UErrorCode errorCode = U_ZERO_ERROR;
    return ulayout_ensureData(errorCode);
}

}  // namespace

/* general properties API functions ----------------------------------------- */

struct BinaryProperty;

typedef UBool BinaryPropertyContains(const BinaryProperty &prop, UChar32 c, UProperty which);

struct BinaryProperty {
    int32_t column;  // SRC_PROPSVEC column, or "source" if mask==0
    uint32_t mask;
    BinaryPropertyContains *contains;
};

static UBool defaultContains(const BinaryProperty &prop, UChar32 c, UProperty /*which*/) {
    /* systematic, directly stored properties */
    return (u_getUnicodeProperties(c, prop.column)&prop.mask)!=0;
}

static UBool caseBinaryPropertyContains(const BinaryProperty &/*prop*/, UChar32 c, UProperty which) {
    return static_cast<UBool>(ucase_hasBinaryProperty(c, which));
}

static UBool isBidiControl(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return ubidi_isBidiControl(c);
}

static UBool isMirrored(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return ubidi_isMirrored(c);
}

static UBool isJoinControl(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return ubidi_isJoinControl(c);
}

#if UCONFIG_NO_NORMALIZATION
static UBool hasFullCompositionExclusion(const BinaryProperty &, UChar32, UProperty) {
    return FALSE;
}
#else
static UBool hasFullCompositionExclusion(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    // By definition, Full_Composition_Exclusion is the same as NFC_QC=No.
    UErrorCode errorCode=U_ZERO_ERROR;
    const Normalizer2Impl *impl=Normalizer2Factory::getNFCImpl(errorCode);
    return U_SUCCESS(errorCode) && impl->isCompNo(impl->getNorm16(c));
}
#endif

// UCHAR_NF*_INERT properties
#if UCONFIG_NO_NORMALIZATION
static UBool isNormInert(const BinaryProperty &, UChar32, UProperty) {
    return FALSE;
}
#else
static UBool isNormInert(const BinaryProperty &/*prop*/, UChar32 c, UProperty which) {
    UErrorCode errorCode=U_ZERO_ERROR;
    const Normalizer2 *norm2=Normalizer2Factory::getInstance(
        (UNormalizationMode)(which-UCHAR_NFD_INERT+UNORM_NFD), errorCode);
    return U_SUCCESS(errorCode) && norm2->isInert(c);
}
#endif

#if UCONFIG_NO_NORMALIZATION
static UBool changesWhenCasefolded(const BinaryProperty &, UChar32, UProperty) {
    return FALSE;
}
#else
static UBool changesWhenCasefolded(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    UnicodeString nfd;
    UErrorCode errorCode=U_ZERO_ERROR;
    const Normalizer2 *nfcNorm2=Normalizer2::getNFCInstance(errorCode);
    if(U_FAILURE(errorCode)) {
        return FALSE;
    }
    if(nfcNorm2->getDecomposition(c, nfd)) {
        /* c has a decomposition */
        if(nfd.length()==1) {
            c=nfd[0];  /* single BMP code point */
        } else if(nfd.length()<=U16_MAX_LENGTH &&
                  nfd.length()==U16_LENGTH(c=nfd.char32At(0))
        ) {
            /* single supplementary code point */
        } else {
            c=U_SENTINEL;
        }
    } else if(c<0) {
        return FALSE;  /* protect against bad input */
    }
    if(c>=0) {
        /* single code point */
        const UChar *resultString;
        return (UBool)(ucase_toFullFolding(c, &resultString, U_FOLD_CASE_DEFAULT)>=0);
    } else {
        /* guess some large but stack-friendly capacity */
        UChar dest[2*UCASE_MAX_STRING_LENGTH];
        int32_t destLength;
        destLength=u_strFoldCase(dest, UPRV_LENGTHOF(dest),
                                  nfd.getBuffer(), nfd.length(),
                                  U_FOLD_CASE_DEFAULT, &errorCode);
        return (UBool)(U_SUCCESS(errorCode) &&
                       0!=u_strCompare(nfd.getBuffer(), nfd.length(),
                                       dest, destLength, FALSE));
    }
}
#endif

#if UCONFIG_NO_NORMALIZATION
static UBool changesWhenNFKC_Casefolded(const BinaryProperty &, UChar32, UProperty) {
    return FALSE;
}
#else
static UBool changesWhenNFKC_Casefolded(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    UErrorCode errorCode=U_ZERO_ERROR;
    const Normalizer2Impl *kcf=Normalizer2Factory::getNFKC_CFImpl(errorCode);
    if(U_FAILURE(errorCode)) {
        return FALSE;
    }
    UnicodeString src(c);
    UnicodeString dest;
    {
        // The ReorderingBuffer must be in a block because its destructor
        // needs to release dest's buffer before we look at its contents.
        ReorderingBuffer buffer(*kcf, dest);
        // Small destCapacity for NFKC_CF(c).
        if(buffer.init(5, errorCode)) {
            const UChar *srcArray=src.getBuffer();
            kcf->compose(srcArray, srcArray+src.length(), FALSE,
                          TRUE, buffer, errorCode);
        }
    }
    return U_SUCCESS(errorCode) && dest!=src;
}
#endif

#if UCONFIG_NO_NORMALIZATION
static UBool isCanonSegmentStarter(const BinaryProperty &, UChar32, UProperty) {
    return FALSE;
}
#else
static UBool isCanonSegmentStarter(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    UErrorCode errorCode=U_ZERO_ERROR;
    const Normalizer2Impl *impl=Normalizer2Factory::getNFCImpl(errorCode);
    return
        U_SUCCESS(errorCode) && impl->ensureCanonIterData(errorCode) &&
        impl->isCanonSegmentStarter(c);
}
#endif

static UBool isPOSIX_alnum(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return u_isalnumPOSIX(c);
}

static UBool isPOSIX_blank(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return u_isblank(c);
}

static UBool isPOSIX_graph(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return u_isgraphPOSIX(c);
}

static UBool isPOSIX_print(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return u_isprintPOSIX(c);
}

static UBool isPOSIX_xdigit(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return u_isxdigit(c);
}

static UBool isRegionalIndicator(const BinaryProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    // Property starts are a subset of lb=RI etc.
    return 0x1F1E6<=c && c<=0x1F1FF;
}

static const BinaryProperty binProps[UCHAR_BINARY_LIMIT]={
    /*
     * column and mask values for binary properties from u_getUnicodeProperties().
     * Must be in order of corresponding UProperty,
     * and there must be exactly one entry per binary UProperty.
     *
     * Properties with mask==0 are handled in code.
     * For them, column is the UPropertySource value.
     */
    { 1,                U_MASK(UPROPS_ALPHABETIC), defaultContains },
    { 1,                U_MASK(UPROPS_ASCII_HEX_DIGIT), defaultContains },
    { UPROPS_SRC_BIDI,  0, isBidiControl },
    { UPROPS_SRC_BIDI,  0, isMirrored },
    { 1,                U_MASK(UPROPS_DASH), defaultContains },
    { 1,                U_MASK(UPROPS_DEFAULT_IGNORABLE_CODE_POINT), defaultContains },
    { 1,                U_MASK(UPROPS_DEPRECATED), defaultContains },
    { 1,                U_MASK(UPROPS_DIACRITIC), defaultContains },
    { 1,                U_MASK(UPROPS_EXTENDER), defaultContains },
    { UPROPS_SRC_NFC,   0, hasFullCompositionExclusion },
    { 1,                U_MASK(UPROPS_GRAPHEME_BASE), defaultContains },
    { 1,                U_MASK(UPROPS_GRAPHEME_EXTEND), defaultContains },
    { 1,                U_MASK(UPROPS_GRAPHEME_LINK), defaultContains },
    { 1,                U_MASK(UPROPS_HEX_DIGIT), defaultContains },
    { 1,                U_MASK(UPROPS_HYPHEN), defaultContains },
    { 1,                U_MASK(UPROPS_ID_CONTINUE), defaultContains },
    { 1,                U_MASK(UPROPS_ID_START), defaultContains },
    { 1,                U_MASK(UPROPS_IDEOGRAPHIC), defaultContains },
    { 1,                U_MASK(UPROPS_IDS_BINARY_OPERATOR), defaultContains },
    { 1,                U_MASK(UPROPS_IDS_TRINARY_OPERATOR), defaultContains },
    { UPROPS_SRC_BIDI,  0, isJoinControl },
    { 1,                U_MASK(UPROPS_LOGICAL_ORDER_EXCEPTION), defaultContains },
    { UPROPS_SRC_CASE,  0, caseBinaryPropertyContains },  // UCHAR_LOWERCASE
    { 1,                U_MASK(UPROPS_MATH), defaultContains },
    { 1,                U_MASK(UPROPS_NONCHARACTER_CODE_POINT), defaultContains },
    { 1,                U_MASK(UPROPS_QUOTATION_MARK), defaultContains },
    { 1,                U_MASK(UPROPS_RADICAL), defaultContains },
    { UPROPS_SRC_CASE,  0, caseBinaryPropertyContains },  // UCHAR_SOFT_DOTTED
    { 1,                U_MASK(UPROPS_TERMINAL_PUNCTUATION), defaultContains },
    { 1,                U_MASK(UPROPS_UNIFIED_IDEOGRAPH), defaultContains },
    { UPROPS_SRC_CASE,  0, caseBinaryPropertyContains },  // UCHAR_UPPERCASE
    { 1,                U_MASK(UPROPS_WHITE_SPACE), defaultContains },
    { 1,                U_MASK(UPROPS_XID_CONTINUE), defaultContains },
    { 1,                U_MASK(UPROPS_XID_START), defaultContains },
    { UPROPS_SRC_CASE,  0, caseBinaryPropertyContains },  // UCHAR_CASE_SENSITIVE
    { 1,                U_MASK(UPROPS_S_TERM), defaultContains },
    { 1,                U_MASK(UPROPS_VARIATION_SELECTOR), defaultContains },
    { UPROPS_SRC_NFC,   0, isNormInert },  // UCHAR_NFD_INERT
    { UPROPS_SRC_NFKC,  0, isNormInert },  // UCHAR_NFKD_INERT
    { UPROPS_SRC_NFC,   0, isNormInert },  // UCHAR_NFC_INERT
    { UPROPS_SRC_NFKC,  0, isNormInert },  // UCHAR_NFKC_INERT
    { UPROPS_SRC_NFC_CANON_ITER, 0, isCanonSegmentStarter },
    { 1,                U_MASK(UPROPS_PATTERN_SYNTAX), defaultContains },
    { 1,                U_MASK(UPROPS_PATTERN_WHITE_SPACE), defaultContains },
    { UPROPS_SRC_CHAR_AND_PROPSVEC,  0, isPOSIX_alnum },
    { UPROPS_SRC_CHAR,  0, isPOSIX_blank },
    { UPROPS_SRC_CHAR,  0, isPOSIX_graph },
    { UPROPS_SRC_CHAR,  0, isPOSIX_print },
    { UPROPS_SRC_CHAR,  0, isPOSIX_xdigit },
    { UPROPS_SRC_CASE,  0, caseBinaryPropertyContains },  // UCHAR_CASED
    { UPROPS_SRC_CASE,  0, caseBinaryPropertyContains },  // UCHAR_CASE_IGNORABLE
    { UPROPS_SRC_CASE,  0, caseBinaryPropertyContains },  // UCHAR_CHANGES_WHEN_LOWERCASED
    { UPROPS_SRC_CASE,  0, caseBinaryPropertyContains },  // UCHAR_CHANGES_WHEN_UPPERCASED
    { UPROPS_SRC_CASE,  0, caseBinaryPropertyContains },  // UCHAR_CHANGES_WHEN_TITLECASED
    { UPROPS_SRC_CASE_AND_NORM,  0, changesWhenCasefolded },
    { UPROPS_SRC_CASE,  0, caseBinaryPropertyContains },  // UCHAR_CHANGES_WHEN_CASEMAPPED
    { UPROPS_SRC_NFKC_CF, 0, changesWhenNFKC_Casefolded },
    { 2,                U_MASK(UPROPS_2_EMOJI), defaultContains },
    { 2,                U_MASK(UPROPS_2_EMOJI_PRESENTATION), defaultContains },
    { 2,                U_MASK(UPROPS_2_EMOJI_MODIFIER), defaultContains },
    { 2,                U_MASK(UPROPS_2_EMOJI_MODIFIER_BASE), defaultContains },
    { 2,                U_MASK(UPROPS_2_EMOJI_COMPONENT), defaultContains },
    { 2,                0, isRegionalIndicator },
    { 1,                U_MASK(UPROPS_PREPENDED_CONCATENATION_MARK), defaultContains },
    { 2,                U_MASK(UPROPS_2_EXTENDED_PICTOGRAPHIC), defaultContains },
};

U_CAPI UBool U_EXPORT2
u_hasBinaryProperty(UChar32 c, UProperty which) {
    /* c is range-checked in the functions that are called from here */
    if(which<UCHAR_BINARY_START || UCHAR_BINARY_LIMIT<=which) {
        /* not a known binary property */
        return FALSE;
    } else {
        const BinaryProperty &prop=binProps[which];
        return prop.contains(prop, c, which);
    }
}

struct IntProperty;

typedef int32_t IntPropertyGetValue(const IntProperty &prop, UChar32 c, UProperty which);
typedef int32_t IntPropertyGetMaxValue(const IntProperty &prop, UProperty which);

struct IntProperty {
    int32_t column;  // SRC_PROPSVEC column, or "source" if mask==0
    uint32_t mask;
    int32_t shift;  // =maxValue if getMaxValueFromShift() is used
    IntPropertyGetValue *getValue;
    IntPropertyGetMaxValue *getMaxValue;
};

static int32_t defaultGetValue(const IntProperty &prop, UChar32 c, UProperty /*which*/) {
    /* systematic, directly stored properties */
    return (int32_t)(u_getUnicodeProperties(c, prop.column)&prop.mask)>>prop.shift;
}

static int32_t defaultGetMaxValue(const IntProperty &prop, UProperty /*which*/) {
    return (uprv_getMaxValues(prop.column)&prop.mask)>>prop.shift;
}

static int32_t getMaxValueFromShift(const IntProperty &prop, UProperty /*which*/) {
    return prop.shift;
}

static int32_t getBiDiClass(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return (int32_t)u_charDirection(c);
}

static int32_t getBiDiPairedBracketType(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return (int32_t)ubidi_getPairedBracketType(c);
}

static int32_t biDiGetMaxValue(const IntProperty &/*prop*/, UProperty which) {
    return ubidi_getMaxValue(which);
}

#if UCONFIG_NO_NORMALIZATION
static int32_t getCombiningClass(const IntProperty &, UChar32, UProperty) {
    return 0;
}
#else
static int32_t getCombiningClass(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return u_getCombiningClass(c);
}
#endif

static int32_t getGeneralCategory(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return (int32_t)u_charType(c);
}

static int32_t getJoiningGroup(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return ubidi_getJoiningGroup(c);
}

static int32_t getJoiningType(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return ubidi_getJoiningType(c);
}

static int32_t getNumericType(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    int32_t ntv=(int32_t)GET_NUMERIC_TYPE_VALUE(u_getMainProperties(c));
    return UPROPS_NTV_GET_TYPE(ntv);
}

static int32_t getScript(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    UErrorCode errorCode=U_ZERO_ERROR;
    return (int32_t)uscript_getScript(c, &errorCode);
}

static int32_t scriptGetMaxValue(const IntProperty &/*prop*/, UProperty /*which*/) {
    uint32_t scriptX=uprv_getMaxValues(0)&UPROPS_SCRIPT_X_MASK;
    return uprops_mergeScriptCodeOrIndex(scriptX);
}

/*
 * Map some of the Grapheme Cluster Break values to Hangul Syllable Types.
 * Hangul_Syllable_Type is fully redundant with a subset of Grapheme_Cluster_Break.
 */
static const UHangulSyllableType gcbToHst[]={
    U_HST_NOT_APPLICABLE,   /* U_GCB_OTHER */
    U_HST_NOT_APPLICABLE,   /* U_GCB_CONTROL */
    U_HST_NOT_APPLICABLE,   /* U_GCB_CR */
    U_HST_NOT_APPLICABLE,   /* U_GCB_EXTEND */
    U_HST_LEADING_JAMO,     /* U_GCB_L */
    U_HST_NOT_APPLICABLE,   /* U_GCB_LF */
    U_HST_LV_SYLLABLE,      /* U_GCB_LV */
    U_HST_LVT_SYLLABLE,     /* U_GCB_LVT */
    U_HST_TRAILING_JAMO,    /* U_GCB_T */
    U_HST_VOWEL_JAMO        /* U_GCB_V */
    /*
     * Omit GCB values beyond what we need for hst.
     * The code below checks for the array length.
     */
};

static int32_t getHangulSyllableType(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    /* see comments on gcbToHst[] above */
    int32_t gcb=(int32_t)(u_getUnicodeProperties(c, 2)&UPROPS_GCB_MASK)>>UPROPS_GCB_SHIFT;
    if(gcb<UPRV_LENGTHOF(gcbToHst)) {
        return gcbToHst[gcb];
    } else {
        return U_HST_NOT_APPLICABLE;
    }
}

#if UCONFIG_NO_NORMALIZATION
static int32_t getNormQuickCheck(const IntProperty &, UChar32, UProperty) {
    return 0;
}
#else
static int32_t getNormQuickCheck(const IntProperty &/*prop*/, UChar32 c, UProperty which) {
    return (int32_t)unorm_getQuickCheck(c, (UNormalizationMode)(which-UCHAR_NFD_QUICK_CHECK+UNORM_NFD));
}
#endif

#if UCONFIG_NO_NORMALIZATION
static int32_t getLeadCombiningClass(const IntProperty &, UChar32, UProperty) {
    return 0;
}
#else
static int32_t getLeadCombiningClass(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return unorm_getFCD16(c)>>8;
}
#endif

#if UCONFIG_NO_NORMALIZATION
static int32_t getTrailCombiningClass(const IntProperty &, UChar32, UProperty) {
    return 0;
}
#else
static int32_t getTrailCombiningClass(const IntProperty &/*prop*/, UChar32 c, UProperty /*which*/) {
    return unorm_getFCD16(c)&0xff;
}
#endif

static int32_t getInPC(const IntProperty &, UChar32 c, UProperty) {
    return ulayout_ensureData() && gInpcTrie != nullptr ? ucptrie_get(gInpcTrie, c) : 0;
}

static int32_t getInSC(const IntProperty &, UChar32 c, UProperty) {
    return ulayout_ensureData() && gInscTrie != nullptr ? ucptrie_get(gInscTrie, c) : 0;
}

static int32_t getVo(const IntProperty &, UChar32 c, UProperty) {
    return ulayout_ensureData() && gVoTrie != nullptr ? ucptrie_get(gVoTrie, c) : 0;
}

static int32_t layoutGetMaxValue(const IntProperty &/*prop*/, UProperty which) {
    if (!ulayout_ensureData()) { return 0; }
    switch (which) {
    case UCHAR_INDIC_POSITIONAL_CATEGORY:
        return gMaxInpcValue;
    case UCHAR_INDIC_SYLLABIC_CATEGORY:
        return gMaxInscValue;
    case UCHAR_VERTICAL_ORIENTATION:
        return gMaxVoValue;
    default:
        return 0;
    }
}

static const IntProperty intProps[UCHAR_INT_LIMIT-UCHAR_INT_START]={
    /*
     * column, mask and shift values for int-value properties from u_getUnicodeProperties().
     * Must be in order of corresponding UProperty,
     * and there must be exactly one entry per int UProperty.
     *
     * Properties with mask==0 are handled in code.
     * For them, column is the UPropertySource value.
     */
    { UPROPS_SRC_BIDI,  0, 0,                               getBiDiClass, biDiGetMaxValue },
    { 0,                UPROPS_BLOCK_MASK, UPROPS_BLOCK_SHIFT, defaultGetValue, defaultGetMaxValue },
    { UPROPS_SRC_NFC,   0, 0xff,                            getCombiningClass, getMaxValueFromShift },
    { 2,                UPROPS_DT_MASK, 0,                  defaultGetValue, defaultGetMaxValue },
    { 0,                UPROPS_EA_MASK, UPROPS_EA_SHIFT,    defaultGetValue, defaultGetMaxValue },
    { UPROPS_SRC_CHAR,  0, (int32_t)U_CHAR_CATEGORY_COUNT-1,getGeneralCategory, getMaxValueFromShift },
    { UPROPS_SRC_BIDI,  0, 0,                               getJoiningGroup, biDiGetMaxValue },
    { UPROPS_SRC_BIDI,  0, 0,                               getJoiningType, biDiGetMaxValue },
    { 2,                UPROPS_LB_MASK, UPROPS_LB_SHIFT,    defaultGetValue, defaultGetMaxValue },
    { UPROPS_SRC_CHAR,  0, (int32_t)U_NT_COUNT-1,           getNumericType, getMaxValueFromShift },
    { UPROPS_SRC_PROPSVEC, 0, 0,                            getScript, scriptGetMaxValue },
    { UPROPS_SRC_PROPSVEC, 0, (int32_t)U_HST_COUNT-1,       getHangulSyllableType, getMaxValueFromShift },
    // UCHAR_NFD_QUICK_CHECK: max=1=YES -- never "maybe", only "no" or "yes"
    { UPROPS_SRC_NFC,   0, (int32_t)UNORM_YES,              getNormQuickCheck, getMaxValueFromShift },
    // UCHAR_NFKD_QUICK_CHECK: max=1=YES -- never "maybe", only "no" or "yes"
    { UPROPS_SRC_NFKC,  0, (int32_t)UNORM_YES,              getNormQuickCheck, getMaxValueFromShift },
    // UCHAR_NFC_QUICK_CHECK: max=2=MAYBE
    { UPROPS_SRC_NFC,   0, (int32_t)UNORM_MAYBE,            getNormQuickCheck, getMaxValueFromShift },
    // UCHAR_NFKC_QUICK_CHECK: max=2=MAYBE
    { UPROPS_SRC_NFKC,  0, (int32_t)UNORM_MAYBE,            getNormQuickCheck, getMaxValueFromShift },
    { UPROPS_SRC_NFC,   0, 0xff,                            getLeadCombiningClass, getMaxValueFromShift },
    { UPROPS_SRC_NFC,   0, 0xff,                            getTrailCombiningClass, getMaxValueFromShift },
    { 2,                UPROPS_GCB_MASK, UPROPS_GCB_SHIFT,  defaultGetValue, defaultGetMaxValue },
    { 2,                UPROPS_SB_MASK, UPROPS_SB_SHIFT,    defaultGetValue, defaultGetMaxValue },
    { 2,                UPROPS_WB_MASK, UPROPS_WB_SHIFT,    defaultGetValue, defaultGetMaxValue },
    { UPROPS_SRC_BIDI,  0, 0,                               getBiDiPairedBracketType, biDiGetMaxValue },
    { UPROPS_SRC_INPC,  0, 0,                               getInPC, layoutGetMaxValue },
    { UPROPS_SRC_INSC,  0, 0,                               getInSC, layoutGetMaxValue },
    { UPROPS_SRC_VO,    0, 0,                               getVo, layoutGetMaxValue },
};

U_CAPI int32_t U_EXPORT2
u_getIntPropertyValue(UChar32 c, UProperty which) {
    if(which<UCHAR_INT_START) {
        if(UCHAR_BINARY_START<=which && which<UCHAR_BINARY_LIMIT) {
            const BinaryProperty &prop=binProps[which];
            return prop.contains(prop, c, which);
        }
    } else if(which<UCHAR_INT_LIMIT) {
        const IntProperty &prop=intProps[which-UCHAR_INT_START];
        return prop.getValue(prop, c, which);
    } else if(which==UCHAR_GENERAL_CATEGORY_MASK) {
        return U_MASK(u_charType(c));
    }
    return 0;  // undefined
}

U_CAPI int32_t U_EXPORT2
u_getIntPropertyMinValue(UProperty /*which*/) {
    return 0; /* all binary/enum/int properties have a minimum value of 0 */
}

U_CAPI int32_t U_EXPORT2
u_getIntPropertyMaxValue(UProperty which) {
    if(which<UCHAR_INT_START) {
        if(UCHAR_BINARY_START<=which && which<UCHAR_BINARY_LIMIT) {
            return 1;  // maximum TRUE for all binary properties
        }
    } else if(which<UCHAR_INT_LIMIT) {
        const IntProperty &prop=intProps[which-UCHAR_INT_START];
        return prop.getMaxValue(prop, which);
    }
    return -1;  // undefined
}

U_CFUNC UPropertySource U_EXPORT2
uprops_getSource(UProperty which) {
    if(which<UCHAR_BINARY_START) {
        return UPROPS_SRC_NONE; /* undefined */
    } else if(which<UCHAR_BINARY_LIMIT) {
        const BinaryProperty &prop=binProps[which];
        if(prop.mask!=0) {
            return UPROPS_SRC_PROPSVEC;
        } else {
            return (UPropertySource)prop.column;
        }
    } else if(which<UCHAR_INT_START) {
        return UPROPS_SRC_NONE; /* undefined */
    } else if(which<UCHAR_INT_LIMIT) {
        const IntProperty &prop=intProps[which-UCHAR_INT_START];
        if(prop.mask!=0) {
            return UPROPS_SRC_PROPSVEC;
        } else {
            return (UPropertySource)prop.column;
        }
    } else if(which<UCHAR_STRING_START) {
        switch(which) {
        case UCHAR_GENERAL_CATEGORY_MASK:
        case UCHAR_NUMERIC_VALUE:
            return UPROPS_SRC_CHAR;

        default:
            return UPROPS_SRC_NONE;
        }
    } else if(which<UCHAR_STRING_LIMIT) {
        switch(which) {
        case UCHAR_AGE:
            return UPROPS_SRC_PROPSVEC;

        case UCHAR_BIDI_MIRRORING_GLYPH:
            return UPROPS_SRC_BIDI;

        case UCHAR_CASE_FOLDING:
        case UCHAR_LOWERCASE_MAPPING:
        case UCHAR_SIMPLE_CASE_FOLDING:
        case UCHAR_SIMPLE_LOWERCASE_MAPPING:
        case UCHAR_SIMPLE_TITLECASE_MAPPING:
        case UCHAR_SIMPLE_UPPERCASE_MAPPING:
        case UCHAR_TITLECASE_MAPPING:
        case UCHAR_UPPERCASE_MAPPING:
            return UPROPS_SRC_CASE;

        case UCHAR_ISO_COMMENT:
        case UCHAR_NAME:
        case UCHAR_UNICODE_1_NAME:
            return UPROPS_SRC_NAMES;

        default:
            return UPROPS_SRC_NONE;
        }
    } else {
        switch(which) {
        case UCHAR_SCRIPT_EXTENSIONS:
            return UPROPS_SRC_PROPSVEC;
        default:
            return UPROPS_SRC_NONE; /* undefined */
        }
    }
}

U_CFUNC void U_EXPORT2
uprops_addPropertyStarts(UPropertySource src, const USetAdder *sa, UErrorCode *pErrorCode) {
    if (!ulayout_ensureData(*pErrorCode)) { return; }
    const UCPTrie *trie;
    switch (src) {
    case UPROPS_SRC_INPC:
        trie = gInpcTrie;
        break;
    case UPROPS_SRC_INSC:
        trie = gInscTrie;
        break;
    case UPROPS_SRC_VO:
        trie = gVoTrie;
        break;
    default:
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    if (trie == nullptr) {
        *pErrorCode = U_MISSING_RESOURCE_ERROR;
        return;
    }

    // Add the start code point of each same-value range of the trie.
    UChar32 start = 0, end;
    while ((end = ucptrie_getRange(trie, start, UCPMAP_RANGE_NORMAL, 0,
                                   nullptr, nullptr, nullptr)) >= 0) {
        sa->add(sa->set, start);
        start = end + 1;
    }
}

#if !UCONFIG_NO_NORMALIZATION

U_CAPI int32_t U_EXPORT2
u_getFC_NFKC_Closure(UChar32 c, UChar *dest, int32_t destCapacity, UErrorCode *pErrorCode) {
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(destCapacity<0 || (dest==NULL && destCapacity>0)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    // Compute the FC_NFKC_Closure on the fly:
    // We have the API for complete coverage of Unicode properties, although
    // this value by itself is not useful via API.
    // (What could be useful is a custom normalization table that combines
    // case folding and NFKC.)
    // For the derivation, see Unicode's DerivedNormalizationProps.txt.
    const Normalizer2 *nfkc=Normalizer2::getNFKCInstance(*pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    // first: b = NFKC(Fold(a))
    UnicodeString folded1String;
    const UChar *folded1;
    int32_t folded1Length=ucase_toFullFolding(c, &folded1, U_FOLD_CASE_DEFAULT);
    if(folded1Length<0) {
        const Normalizer2Impl *nfkcImpl=Normalizer2Factory::getImpl(nfkc);
        if(nfkcImpl->getCompQuickCheck(nfkcImpl->getNorm16(c))!=UNORM_NO) {
            return u_terminateUChars(dest, destCapacity, 0, pErrorCode);  // c does not change at all under CaseFolding+NFKC
        }
        folded1String.setTo(c);
    } else {
        if(folded1Length>UCASE_MAX_STRING_LENGTH) {
            folded1String.setTo(folded1Length);
        } else {
            folded1String.setTo(FALSE, folded1, folded1Length);
        }
    }
    UnicodeString kc1=nfkc->normalize(folded1String, *pErrorCode);
    // second: c = NFKC(Fold(b))
    UnicodeString folded2String(kc1);
    UnicodeString kc2=nfkc->normalize(folded2String.foldCase(), *pErrorCode);
    // if (c != b) add the mapping from a to c
    if(U_FAILURE(*pErrorCode) || kc1==kc2) {
        return u_terminateUChars(dest, destCapacity, 0, pErrorCode);
    } else {
        return kc2.extract(dest, destCapacity, *pErrorCode);
    }
}

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1997-2006, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ures_cnv.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004aug25
*   created by: Markus W. Scherer
*
*   Character conversion functions moved here from uresbund.c
*/







// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 1999-2010, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// **********************************************************************
// *   file name:  ustr_cnv.h
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2004Aug27
// *   created by: George Rhoten
// */

// #ifndef USTR_CNV_IMP_H
// #define USTR_CNV_IMP_H

// #include "unicode/utypes.h"
// #include "unicode/ucnv.h"

// #if !UCONFIG_NO_CONVERSION

// /**
//  * Get the default converter. This is a commonly used converter
//  * that is used for the ustring and UnicodeString API.
//  * Remember to use the u_releaseDefaultConverter when you are done.
//  * @internal
//  */
// U_CAPI UConverter* U_EXPORT2
// u_getDefaultConverter(UErrorCode *status);


// /**
//  * Release the default converter to the converter cache.
//  * @internal
//  */
// U_CAPI void U_EXPORT2
// u_releaseDefaultConverter(UConverter *converter);

// /**
//  * Flush the default converter, if cached.
//  * @internal
//  */
// U_CAPI void U_EXPORT2
// u_flushDefaultConverter(void);

// #endif

// #endif


U_CAPI UResourceBundle * U_EXPORT2
ures_openU(const UChar *myPath, 
           const char *localeID, 
           UErrorCode *status)
{
    char pathBuffer[1024];
    int32_t length;
    char *path = pathBuffer;

    if(status==NULL || U_FAILURE(*status)) {
        return NULL;
    }
    if(myPath==NULL) {
        path = NULL;
    }
    else {
        length=u_strlen(myPath);
        if(length>=(int32_t)sizeof(pathBuffer)) {
            *status=U_ILLEGAL_ARGUMENT_ERROR;
            return NULL;
        } else if(uprv_isInvariantUString(myPath, length)) {
            /*
             * the invariant converter is sufficient for package and tree names
             * and is more efficient
             */
            u_UCharsToChars(myPath, path, length+1); /* length+1 to include the NUL */
        } else {
#if !UCONFIG_NO_CONVERSION
            /* use the default converter to support variant-character paths */
            UConverter *cnv=u_getDefaultConverter(status);
            length=ucnv_fromUChars(cnv, path, (int32_t)sizeof(pathBuffer), myPath, length, status);
            u_releaseDefaultConverter(cnv);
            if(U_FAILURE(*status)) {
                return NULL;
            }
            if(length>=(int32_t)sizeof(pathBuffer)) {
                /* not NUL-terminated - path too long */
                *status=U_ILLEGAL_ARGUMENT_ERROR;
                return NULL;
            }
#else
            /* the default converter is not available */
            *status=U_UNSUPPORTED_ERROR;
            return NULL;
#endif
        }
    }

    return ures_open(path, localeID, status);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 1997-2016, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*
* File uresbund.cpp
*
* Modification History:
*
*   Date        Name        Description
*   04/01/97    aliu        Creation.
*   06/14/99    stephen     Removed functions taking a filename suffix.
*   07/20/99    stephen     Changed for UResourceBundle typedef'd to void*
*   11/09/99    weiv            Added ures_getLocale()
*   March 2000  weiv        Total overhaul - using data in DLLs
*   06/20/2000  helena      OS/400 port changes; mostly typecast.
*   06/24/02    weiv        Added support for resource sharing
******************************************************************************
*/





















using namespace icu;

/*
Static uresbund_cache for already opened resource bundles - mostly for keeping fallback info
TODO: This uresbund_cache should probably be removed when the deprecated code is
      completely removed.
*/
static UHashtable *uresbund_cache = NULL;
static icu::UInitOnce uresbund_gCacheInitOnce = U_INITONCE_INITIALIZER;

static UMutex resbMutex;

/* INTERNAL: hashes an entry  */
static int32_t U_CALLCONV hashEntry(const UHashTok parm) {
    UResourceDataEntry *b = (UResourceDataEntry *)parm.pointer;
    UHashTok namekey, pathkey;
    namekey.pointer = b->fName;
    pathkey.pointer = b->fPath;
    return uhash_hashChars(namekey)+37u*uhash_hashChars(pathkey);
}

/* INTERNAL: compares two entries */
static UBool U_CALLCONV compareEntries(const UHashTok p1, const UHashTok p2) {
    UResourceDataEntry *b1 = (UResourceDataEntry *)p1.pointer;
    UResourceDataEntry *b2 = (UResourceDataEntry *)p2.pointer;
    UHashTok name1, name2, path1, path2;
    name1.pointer = b1->fName;
    name2.pointer = b2->fName;
    path1.pointer = b1->fPath;
    path2.pointer = b2->fPath;
    return (UBool)(uhash_compareChars(name1, name2) &&
        uhash_compareChars(path1, path2));
}


/**
 *  Internal function, gets parts of locale name according
 *  to the position of '_' character
 */
static UBool chopLocale(char *name) {
    char *i = uprv_strrchr(name, '_');

    if(i != NULL) {
        *i = '\0';
        return TRUE;
    }

    return FALSE;
}

/**
 *  Internal function
 */
static void entryIncrease(UResourceDataEntry *entry) {
    Mutex lock(&resbMutex);
    entry->fCountExisting++;
    while(entry->fParent != NULL) {
      entry = entry->fParent;
      entry->fCountExisting++;
    }
}

/**
 *  Internal function. Tries to find a resource in given Resource
 *  Bundle, as well as in its parents
 */
static const ResourceData *getFallbackData(const UResourceBundle* resBundle, const char* * resTag, UResourceDataEntry* *realData, Resource *res, UErrorCode *status) {
    UResourceDataEntry *resB = resBundle->fData;
    int32_t indexR = -1;
    int32_t i = 0;
    *res = RES_BOGUS;
    if(resB != NULL) {
        if(resB->fBogus == U_ZERO_ERROR) { /* if this resource is real, */
            *res = res_getTableItemByKey(&(resB->fData), resB->fData.rootRes, &indexR, resTag); /* try to get data from there */
            i++;
        }
        if(resBundle->fHasFallback == TRUE) {
            while(*res == RES_BOGUS && resB->fParent != NULL) { /* Otherwise, we'll look in parents */
                resB = resB->fParent;
                if(resB->fBogus == U_ZERO_ERROR) {
                    i++;
                    *res = res_getTableItemByKey(&(resB->fData), resB->fData.rootRes, &indexR, resTag);
                }
            }
        }

        if(*res != RES_BOGUS) { /* If the resource is found in parents, we need to adjust the error */
            if(i>1) {
                if(uprv_strcmp(resB->fName, uloc_getDefault())==0 || uprv_strcmp(resB->fName, kRootLocaleName)==0) {
                    *status = U_USING_DEFAULT_WARNING;
                } else {
                    *status = U_USING_FALLBACK_WARNING;
                }
            }
            *realData = resB;
            return (&(resB->fData));
        } else { /* If resource is not found, we need to give an error */
            *status = U_MISSING_RESOURCE_ERROR;
            return NULL;
        }
    } else {
            *status = U_MISSING_RESOURCE_ERROR;
            return NULL;
    }
}

static void
free_entry(UResourceDataEntry *entry) {
    UResourceDataEntry *alias;
    res_unload(&(entry->fData));
    if(entry->fName != NULL && entry->fName != entry->fNameBuffer) {
        uprv_free(entry->fName);
    }
    if(entry->fPath != NULL) {
        uprv_free(entry->fPath);
    }
    if(entry->fPool != NULL) {
        --entry->fPool->fCountExisting;
    }
    alias = entry->fAlias;
    if(alias != NULL) {
        while(alias->fAlias != NULL) {
            alias = alias->fAlias;
        }
        --alias->fCountExisting;
    }
    uprv_free(entry);
}

/* Works just like ucnv_flushCache() */
static int32_t ures_flushCache()
{
    UResourceDataEntry *resB;
    int32_t pos;
    int32_t rbDeletedNum = 0;
    const UHashElement *e;
    UBool deletedMore;

    /*if shared data hasn't even been lazy evaluated yet
    * return 0
    */
    Mutex lock(&resbMutex);
    if (uresbund_cache == NULL) {
        return 0;
    }

    do {
        deletedMore = FALSE;
        /*creates an enumeration to iterate through every element in the table */
        pos = UHASH_FIRST;
        while ((e = uhash_nextElement(uresbund_cache, &pos)) != NULL)
        {
            resB = (UResourceDataEntry *) e->value.pointer;
            /* Deletes only if reference counter == 0
             * Don't worry about the children of this node.
             * Those will eventually get deleted too, if not already.
             * Don't worry about the parents of this node.
             * Those will eventually get deleted too, if not already.
             */
            /* 04/05/2002 [weiv] fCountExisting should now be accurate. If it's not zero, that means that    */
            /* some resource bundles are still open somewhere. */

            if (resB->fCountExisting == 0) {
                rbDeletedNum++;
                deletedMore = TRUE;
                uhash_removeElement(uresbund_cache, e);
                free_entry(resB);
            }
        }
        /*
         * Do it again to catch bundles (aliases, pool bundle) whose fCountExisting
         * got decremented by free_entry().
         */
    } while(deletedMore);

    return rbDeletedNum;
}

#ifdef URES_DEBUG
#include <stdio.h>

U_CAPI UBool U_EXPORT2 ures_dumpCacheContents(void) {
  UBool uresbund_cacheNotEmpty = FALSE;
  int32_t pos = UHASH_FIRST;
  const UHashElement *e;
  UResourceDataEntry *resB;

    Mutex lock(&resbMutex);
    if (uresbund_cache == NULL) {
      fprintf(stderr,"%s:%d: RB Cache is NULL.\n", __FILE__, __LINE__);
      return FALSE;
    }

    while ((e = uhash_nextElement(uresbund_cache, &pos)) != NULL) {
      uresbund_cacheNotEmpty=TRUE;
      resB = (UResourceDataEntry *) e->value.pointer;
      fprintf(stderr,"%s:%d: RB Cache: Entry @0x%p, refcount %d, name %s:%s.  Pool 0x%p, alias 0x%p, parent 0x%p\n",
              __FILE__, __LINE__,
              (void*)resB, resB->fCountExisting,
              resB->fName?resB->fName:"NULL",
              resB->fPath?resB->fPath:"NULL",
              (void*)resB->fPool,
              (void*)resB->fAlias,
              (void*)resB->fParent);
    }

    fprintf(stderr,"%s:%d: RB Cache still contains %d items.\n", __FILE__, __LINE__, uhash_count(uresbund_cache));
    return uresbund_cacheNotEmpty;
}

#endif

static UBool U_CALLCONV ures_cleanup(void)
{
    if (uresbund_cache != NULL) {
        ures_flushCache();
        uhash_close(uresbund_cache);
        uresbund_cache = NULL;
    }
    uresbund_gCacheInitOnce.reset();
    return TRUE;
}

/** INTERNAL: Initializes the uresbund_cache for resources */
static void U_CALLCONV createCache(UErrorCode &status) {
    U_ASSERT(uresbund_cache == NULL);
    uresbund_cache = uhash_open(hashEntry, compareEntries, NULL, &status);
    ucln_common_registerCleanup(UCLN_COMMON_URES, ures_cleanup);
}

static void initCache(UErrorCode *status) {
    umtx_initOnce(uresbund_gCacheInitOnce, &createCache, *status);
}

/** INTERNAL: sets the name (locale) of the resource bundle to given name */

static void setEntryName(UResourceDataEntry *res, const char *name, UErrorCode *status) {
    int32_t len = (int32_t)uprv_strlen(name);
    if(res->fName != NULL && res->fName != res->fNameBuffer) {
        uprv_free(res->fName);
    }
    if (len < (int32_t)sizeof(res->fNameBuffer)) {
        res->fName = res->fNameBuffer;
    }
    else {
        res->fName = (char *)uprv_malloc(len+1);
    }
    if(res->fName == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
    } else {
        uprv_strcpy(res->fName, name);
    }
}

static UResourceDataEntry *
getPoolEntry(const char *path, UErrorCode *status);

/**
 *  INTERNAL: Inits and opens an entry from a data DLL.
 *    CAUTION:  resbMutex must be locked when calling this function.
 */
static UResourceDataEntry *init_entry(const char *localeID, const char *path, UErrorCode *status) {
    UResourceDataEntry *r = NULL;
    UResourceDataEntry find;
    /*int32_t hashValue;*/
    const char *name;
    char aliasName[100] = { 0 };
    int32_t aliasLen = 0;
    /*UBool isAlias = FALSE;*/
    /*UHashTok hashkey; */

    if(U_FAILURE(*status)) {
        return NULL;
    }

    /* here we try to deduce the right locale name */
    if(localeID == NULL) { /* if localeID is NULL, we're trying to open default locale */
        name = uloc_getDefault();
    } else if(*localeID == 0) { /* if localeID is "" then we try to open root locale */
        name = kRootLocaleName;
    } else { /* otherwise, we'll open what we're given */
        name = localeID;
    }

    find.fName = (char *)name;
    find.fPath = (char *)path;

    /* calculate the hash value of the entry */
    /*hashkey.pointer = (void *)&find;*/
    /*hashValue = hashEntry(hashkey);*/

    /* check to see if we already have this entry */
    r = (UResourceDataEntry *)uhash_get(uresbund_cache, &find);
    if(r == NULL) {
        /* if the entry is not yet in the hash table, we'll try to construct a new one */
        r = (UResourceDataEntry *) uprv_malloc(sizeof(UResourceDataEntry));
        if(r == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }

        uprv_memset(r, 0, sizeof(UResourceDataEntry));
        /*r->fHashKey = hashValue;*/

        setEntryName(r, name, status);
        if (U_FAILURE(*status)) {
            uprv_free(r);
            return NULL;
        }

        if(path != NULL) {
            r->fPath = (char *)uprv_strdup(path);
            if(r->fPath == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                uprv_free(r);
                return NULL;
            }
        }

        /* this is the actual loading */
        res_load(&(r->fData), r->fPath, r->fName, status);

        if (U_FAILURE(*status)) {
            /* if we failed to load due to an out-of-memory error, exit early. */
            if (*status == U_MEMORY_ALLOCATION_ERROR) {
                uprv_free(r);
                return NULL;
            }
            /* we have no such entry in dll, so it will always use fallback */
            *status = U_USING_FALLBACK_WARNING;
            r->fBogus = U_USING_FALLBACK_WARNING;
        } else { /* if we have a regular entry */
            Resource aliasres;
            if (r->fData.usesPoolBundle) {
                r->fPool = getPoolEntry(r->fPath, status);
                if (U_SUCCESS(*status)) {
                    const int32_t *poolIndexes = r->fPool->fData.pRoot + 1;
                    if(r->fData.pRoot[1 + URES_INDEX_POOL_CHECKSUM] == poolIndexes[URES_INDEX_POOL_CHECKSUM]) {
                        r->fData.poolBundleKeys = (const char *)(poolIndexes + (poolIndexes[URES_INDEX_LENGTH] & 0xff));
                        r->fData.poolBundleStrings = r->fPool->fData.p16BitUnits;
                    } else {
                        r->fBogus = *status = U_INVALID_FORMAT_ERROR;
                    }
                } else {
                    r->fBogus = *status;
                }
            }
            if (U_SUCCESS(*status)) {
                /* handle the alias by trying to get out the %%Alias tag.*/
                /* We'll try to get alias string from the bundle */
                aliasres = res_getResource(&(r->fData), "%%ALIAS");
                if (aliasres != RES_BOGUS) {
                    // No tracing: called during initial data loading
                    const UChar *alias = res_getStringNoTrace(&(r->fData), aliasres, &aliasLen);
                    if(alias != NULL && aliasLen > 0) { /* if there is actual alias - unload and load new data */
                        u_UCharsToChars(alias, aliasName, aliasLen+1);
                        r->fAlias = init_entry(aliasName, path, status);
                    }
                }
            }
        }

        {
            UResourceDataEntry *oldR = NULL;
            if((oldR = (UResourceDataEntry *)uhash_get(uresbund_cache, r)) == NULL) { /* if the data is not uresbund_cached */
                /* just insert it in the uresbund_cache */
                UErrorCode uresbund_cacheStatus = U_ZERO_ERROR;
                uhash_put(uresbund_cache, (void *)r, r, &uresbund_cacheStatus);
                if (U_FAILURE(uresbund_cacheStatus)) {
                    *status = uresbund_cacheStatus;
                    free_entry(r);
                    r = NULL;
                }
            } else {
                /* somebody have already inserted it while we were working, discard newly opened data */
                /* Also, we could get here IF we opened an alias */
                free_entry(r);
                r = oldR;
            }
        }

    }
    if(r != NULL) {
        /* return the real bundle */
        while(r->fAlias != NULL) {
            r = r->fAlias;
        }
        r->fCountExisting++; /* we increase its reference count */
        /* if the resource has a warning */
        /* we don't want to overwrite a status with no error */
        if(r->fBogus != U_ZERO_ERROR && U_SUCCESS(*status)) {
             *status = r->fBogus; /* set the returning status */
        }
    }
    return r;
}

static UResourceDataEntry *
getPoolEntry(const char *path, UErrorCode *status) {
    UResourceDataEntry *poolBundle = init_entry(kPoolBundleName, path, status);
    if( U_SUCCESS(*status) &&
        (poolBundle == NULL || poolBundle->fBogus != U_ZERO_ERROR || !poolBundle->fData.isPoolBundle)
    ) {
        *status = U_INVALID_FORMAT_ERROR;
    }
    return poolBundle;
}

/* INTERNAL: */
/*   CAUTION:  resbMutex must be locked when calling this function! */
static UResourceDataEntry *
findFirstExisting(const char* path, char* name,
                  UBool *isRoot, UBool *hasChopped, UBool *isDefault, UErrorCode* status) {
    UResourceDataEntry *r = NULL;
    UBool hasRealData = FALSE;
    const char *defaultLoc = uloc_getDefault();
    *hasChopped = TRUE; /* we're starting with a fresh name */

    while(*hasChopped && !hasRealData) {
        r = init_entry(name, path, status);
        /* Null pointer test */
        if (U_FAILURE(*status)) {
            return NULL;
        }
        *isDefault = (UBool)(uprv_strncmp(name, defaultLoc, uprv_strlen(name)) == 0);
        hasRealData = (UBool)(r->fBogus == U_ZERO_ERROR);
        if(!hasRealData) {
            /* this entry is not real. We will discard it. */
            /* However, the parent line for this entry is  */
            /* not to be used - as there might be parent   */
            /* lines in uresbund_cache from previous openings that  */
            /* are not updated yet. */
            r->fCountExisting--;
            /*entryCloseInt(r);*/
            r = NULL;
            *status = U_USING_FALLBACK_WARNING;
        } else {
            uprv_strcpy(name, r->fName); /* this is needed for supporting aliases */
        }

        *isRoot = (UBool)(uprv_strcmp(name, kRootLocaleName) == 0);

        /*Fallback data stuff*/
        *hasChopped = chopLocale(name);
        if (*hasChopped && *name == '\0') {
            uprv_strcpy(name, "und");
        }
    }
    return r;
}

static void ures_setIsStackObject( UResourceBundle* resB, UBool state) {
    if(state) {
        resB->fMagic1 = 0;
        resB->fMagic2 = 0;
    } else {
        resB->fMagic1 = MAGIC1;
        resB->fMagic2 = MAGIC2;
    }
}

static UBool ures_isStackObject(const UResourceBundle* resB) {
  return((resB->fMagic1 == MAGIC1 && resB->fMagic2 == MAGIC2)?FALSE:TRUE);
}


U_CFUNC void ures_initStackObject(UResourceBundle* resB) {
  uprv_memset(resB, 0, sizeof(UResourceBundle));
  ures_setIsStackObject(resB, TRUE);
}

U_NAMESPACE_BEGIN

StackUResourceBundle::StackUResourceBundle() {
    ures_initStackObject(&bundle);
}

StackUResourceBundle::~StackUResourceBundle() {
    ures_close(&bundle);
}

U_NAMESPACE_END

static UBool  // returns U_SUCCESS(*status)
loadParentsExceptRoot(UResourceDataEntry *&t1,
                      char name[], int32_t nameCapacity,
                      UBool usingUSRData, char usrDataPath[], UErrorCode *status) {
    if (U_FAILURE(*status)) { return FALSE; }
    UBool hasChopped = TRUE;
    while (hasChopped && t1->fParent == NULL && !t1->fData.noFallback &&
            res_getResource(&t1->fData,"%%ParentIsRoot") == RES_BOGUS) {
        Resource parentRes = res_getResource(&t1->fData, "%%Parent");
        if (parentRes != RES_BOGUS) {  // An explicit parent was found.
            int32_t parentLocaleLen = 0;
            // No tracing: called during initial data loading
            const UChar *parentLocaleName = res_getStringNoTrace(&(t1->fData), parentRes, &parentLocaleLen);
            if(parentLocaleName != NULL && 0 < parentLocaleLen && parentLocaleLen < nameCapacity) {
                u_UCharsToChars(parentLocaleName, name, parentLocaleLen + 1);
                if (uprv_strcmp(name, kRootLocaleName) == 0) {
                    return TRUE;
                }
            }
        }
        // Insert regular parents.
        UErrorCode parentStatus = U_ZERO_ERROR;
        UResourceDataEntry *t2 = init_entry(name, t1->fPath, &parentStatus);
        if (U_FAILURE(parentStatus)) {
            *status = parentStatus;
            return FALSE;
        }
        UResourceDataEntry *u2 = NULL;
        UErrorCode usrStatus = U_ZERO_ERROR;
        if (usingUSRData) {  // This code inserts user override data into the inheritance chain.
            u2 = init_entry(name, usrDataPath, &usrStatus);
            // If we failed due to out-of-memory, report that to the caller and exit early.
            if (usrStatus == U_MEMORY_ALLOCATION_ERROR) {
                *status = usrStatus;
                return FALSE;
            }
        }

        if (usingUSRData && U_SUCCESS(usrStatus) && u2->fBogus == U_ZERO_ERROR) {
            t1->fParent = u2;
            u2->fParent = t2;
        } else {
            t1->fParent = t2;
            if (usingUSRData) {
                // The USR override data wasn't found, set it to be deleted.
                u2->fCountExisting = 0;
            }
        }
        t1 = t2;
        hasChopped = chopLocale(name);
    }
    return TRUE;
}

static UBool  // returns U_SUCCESS(*status)
insertRootBundle(UResourceDataEntry *&t1, UErrorCode *status) {
    if (U_FAILURE(*status)) { return FALSE; }
    UErrorCode parentStatus = U_ZERO_ERROR;
    UResourceDataEntry *t2 = init_entry(kRootLocaleName, t1->fPath, &parentStatus);
    if (U_FAILURE(parentStatus)) {
        *status = parentStatus;
        return FALSE;
    }
    t1->fParent = t2;
    t1 = t2;
    return TRUE;
}

enum UResOpenType {
    /**
     * Open a resource bundle for the locale;
     * if there is not even a base language bundle, then fall back to the default locale;
     * if there is no bundle for that either, then load the root bundle.
     *
     * This is the default bundle loading behavior.
     */
    URES_OPEN_LOCALE_DEFAULT_ROOT,
    // TODO: ICU ticket #11271 "consistent default locale across locale trees"
    // Add an option to look at the main locale tree for whether to
    // fall back to root directly (if the locale has main data) or
    // fall back to the default locale first (if the locale does not even have main data).
    /**
     * Open a resource bundle for the locale;
     * if there is not even a base language bundle, then load the root bundle;
     * never fall back to the default locale.
     *
     * This is used for algorithms that have good pan-Unicode default behavior,
     * such as case mappings, collation, and segmentation (BreakIterator).
     */
    URES_OPEN_LOCALE_ROOT,
    /**
     * Open a resource bundle for the exact bundle name as requested;
     * no fallbacks, do not load parent bundles.
     *
     * This is used for supplemental (non-locale) data.
     */
    URES_OPEN_DIRECT
};
typedef enum UResOpenType UResOpenType;

static UResourceDataEntry *entryOpen(const char* path, const char* localeID,
                                     UResOpenType openType, UErrorCode* status) {
    U_ASSERT(openType != URES_OPEN_DIRECT);
    UErrorCode intStatus = U_ZERO_ERROR;
    UResourceDataEntry *r = NULL;
    UResourceDataEntry *t1 = NULL;
    UBool isDefault = FALSE;
    UBool isRoot = FALSE;
    UBool hasRealData = FALSE;
    UBool hasChopped = TRUE;
    UBool usingUSRData = U_USE_USRDATA && ( path == NULL || uprv_strncmp(path,U_ICUDATA_NAME,8) == 0);

    char name[ULOC_FULLNAME_CAPACITY];
    char usrDataPath[96];

    initCache(status);

    if(U_FAILURE(*status)) {
        return NULL;
    }

    uprv_strncpy(name, localeID, sizeof(name) - 1);
    name[sizeof(name) - 1] = 0;

    if ( usingUSRData ) {
        if ( path == NULL ) {
            uprv_strcpy(usrDataPath, U_USRDATA_NAME);
        } else {
            uprv_strncpy(usrDataPath, path, sizeof(usrDataPath) - 1);
            usrDataPath[0] = 'u';
            usrDataPath[1] = 's';
            usrDataPath[2] = 'r';
            usrDataPath[sizeof(usrDataPath) - 1] = 0;
        }
    }

    Mutex lock(&resbMutex);    // Lock resbMutex until the end of this function.

    /* We're going to skip all the locales that do not have any data */
    r = findFirstExisting(path, name, &isRoot, &hasChopped, &isDefault, &intStatus);

    // If we failed due to out-of-memory, report the failure and exit early.
    if (intStatus == U_MEMORY_ALLOCATION_ERROR) {
        *status = intStatus;
        goto finish;
    }

    if(r != NULL) { /* if there is one real locale, we can look for parents. */
        t1 = r;
        hasRealData = TRUE;
        if ( usingUSRData ) {  /* This code inserts user override data into the inheritance chain */
            UErrorCode usrStatus = U_ZERO_ERROR;
            UResourceDataEntry *u1 = init_entry(t1->fName, usrDataPath, &usrStatus);
            // If we failed due to out-of-memory, report the failure and exit early.
            if (intStatus == U_MEMORY_ALLOCATION_ERROR) {
                *status = intStatus;
                goto finish;
            }
            if ( u1 != NULL ) {
                if(u1->fBogus == U_ZERO_ERROR) {
                    u1->fParent = t1;
                    r = u1;
                } else {
                    /* the USR override data wasn't found, set it to be deleted */
                    u1->fCountExisting = 0;
                }
            }
        }
        if (hasChopped && !isRoot) {
            if (!loadParentsExceptRoot(t1, name, UPRV_LENGTHOF(name), usingUSRData, usrDataPath, status)) {
                goto finish;
            }
        }
    }

    /* we could have reached this point without having any real data */
    /* if that is the case, we need to chain in the default locale   */
    if(r==NULL && openType == URES_OPEN_LOCALE_DEFAULT_ROOT && !isDefault && !isRoot) {
        /* insert default locale */
        uprv_strcpy(name, uloc_getDefault());
        r = findFirstExisting(path, name, &isRoot, &hasChopped, &isDefault, &intStatus);
        // If we failed due to out-of-memory, report the failure and exit early.
        if (intStatus == U_MEMORY_ALLOCATION_ERROR) {
            *status = intStatus;
            goto finish;
        }
        intStatus = U_USING_DEFAULT_WARNING;
        if(r != NULL) { /* the default locale exists */
            t1 = r;
            hasRealData = TRUE;
            isDefault = TRUE;
            // TODO: Why not if (usingUSRData) { ... } like in the non-default-locale code path?
            if (hasChopped && !isRoot) {
                if (!loadParentsExceptRoot(t1, name, UPRV_LENGTHOF(name), usingUSRData, usrDataPath, status)) {
                    goto finish;
                }
            }
        }
    }

    /* we could still have r == NULL at this point - maybe even default locale is not */
    /* present */
    if(r == NULL) {
        uprv_strcpy(name, kRootLocaleName);
        r = findFirstExisting(path, name, &isRoot, &hasChopped, &isDefault, &intStatus);
        // If we failed due to out-of-memory, report the failure and exit early.
        if (intStatus == U_MEMORY_ALLOCATION_ERROR) {
            *status = intStatus;
            goto finish;
        }
        if(r != NULL) {
            t1 = r;
            intStatus = U_USING_DEFAULT_WARNING;
            hasRealData = TRUE;
        } else { /* we don't even have the root locale */
            *status = U_MISSING_RESOURCE_ERROR;
            goto finish;
        }
    } else if(!isRoot && uprv_strcmp(t1->fName, kRootLocaleName) != 0 &&
            t1->fParent == NULL && !r->fData.noFallback) {
        if (!insertRootBundle(t1, status)) {
            goto finish;
        }
        if(!hasRealData) {
            r->fBogus = U_USING_DEFAULT_WARNING;
        }
    }

    // TODO: Does this ever loop?
    while(r != NULL && !isRoot && t1->fParent != NULL) {
        t1->fParent->fCountExisting++;
        t1 = t1->fParent;
    }

finish:
    if(U_SUCCESS(*status)) {
        if(intStatus != U_ZERO_ERROR) {
            *status = intStatus;
        }
        return r;
    } else {
        return NULL;
    }
}

/**
 * Version of entryOpen() and findFirstExisting() for ures_openDirect(),
 * with no fallbacks.
 * Parent and root locale bundles are loaded if
 * the requested bundle does not have the "nofallback" flag.
 */
static UResourceDataEntry *
entryOpenDirect(const char* path, const char* localeID, UErrorCode* status) {
    initCache(status);
    if(U_FAILURE(*status)) {
        return NULL;
    }

    Mutex lock(&resbMutex);
    // findFirstExisting() without fallbacks.
    UResourceDataEntry *r = init_entry(localeID, path, status);
    if(U_SUCCESS(*status)) {
        if(r->fBogus != U_ZERO_ERROR) {
            r->fCountExisting--;
            r = NULL;
        }
    } else {
        r = NULL;
    }

    // Some code depends on the ures_openDirect() bundle to have a parent bundle chain,
    // unless it is marked with "nofallback".
    UResourceDataEntry *t1 = r;
    if(r != NULL && uprv_strcmp(localeID, kRootLocaleName) != 0 &&  // not root
            r->fParent == NULL && !r->fData.noFallback &&
            uprv_strlen(localeID) < ULOC_FULLNAME_CAPACITY) {
        char name[ULOC_FULLNAME_CAPACITY];
        uprv_strcpy(name, localeID);
        if(!chopLocale(name) || uprv_strcmp(name, kRootLocaleName) == 0 ||
                loadParentsExceptRoot(t1, name, UPRV_LENGTHOF(name), FALSE, NULL, status)) {
            if(uprv_strcmp(t1->fName, kRootLocaleName) != 0 && t1->fParent == NULL) {
                insertRootBundle(t1, status);
            }
        }
        if(U_FAILURE(*status)) {
            r = NULL;
        }
    }

    if(r != NULL) {
        // TODO: Does this ever loop?
        while(t1->fParent != NULL) {
            t1->fParent->fCountExisting++;
            t1 = t1->fParent;
        }
    }
    return r;
}

/**
 * Functions to create and destroy resource bundles.
 *     CAUTION:  resbMutex must be locked when calling this function.
 */
/* INTERNAL: */
static void entryCloseInt(UResourceDataEntry *resB) {
    UResourceDataEntry *p = resB;

    while(resB != NULL) {
        p = resB->fParent;
        resB->fCountExisting--;

        /* Entries are left in the uresbund_cache. TODO: add ures_flushCache() to force a flush
         of the uresbund_cache. */
/*
        if(resB->fCountExisting <= 0) {
            uhash_remove(uresbund_cache, resB);
            if(resB->fBogus == U_ZERO_ERROR) {
                res_unload(&(resB->fData));
            }
            if(resB->fName != NULL) {
                uprv_free(resB->fName);
            }
            if(resB->fPath != NULL) {
                uprv_free(resB->fPath);
            }
            uprv_free(resB);
        }
*/

        resB = p;
    }
}

/**
 *  API: closes a resource bundle and cleans up.
 */

static void entryClose(UResourceDataEntry *resB) {
  Mutex lock(&resbMutex);
  entryCloseInt(resB);
}

/*
U_CFUNC void ures_setResPath(UResourceBundle *resB, const char* toAdd) {
  if(resB->fResPath == NULL) {
    resB->fResPath = resB->fResBuf;
    *(resB->fResPath) = 0;
  }
  resB->fResPathLen = uprv_strlen(toAdd);
  if(RES_BUFSIZE <= resB->fResPathLen+1) {
    if(resB->fResPath == resB->fResBuf) {
      resB->fResPath = (char *)uprv_malloc((resB->fResPathLen+1)*sizeof(char));
    } else {
      resB->fResPath = (char *)uprv_realloc(resB->fResPath, (resB->fResPathLen+1)*sizeof(char));
    }
  }
  uprv_strcpy(resB->fResPath, toAdd);
}
*/
static void ures_appendResPath(UResourceBundle *resB, const char* toAdd, int32_t lenToAdd, UErrorCode *status) {
    int32_t resPathLenOrig = resB->fResPathLen;
    if(resB->fResPath == NULL) {
        resB->fResPath = resB->fResBuf;
        *(resB->fResPath) = 0;
        resB->fResPathLen = 0;
    }
    resB->fResPathLen += lenToAdd;
    if(RES_BUFSIZE <= resB->fResPathLen+1) {
        if(resB->fResPath == resB->fResBuf) {
            resB->fResPath = (char *)uprv_malloc((resB->fResPathLen+1)*sizeof(char));
            /* Check that memory was allocated correctly. */
            if (resB->fResPath == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            uprv_strcpy(resB->fResPath, resB->fResBuf);
        } else {
            char *temp = (char *)uprv_realloc(resB->fResPath, (resB->fResPathLen+1)*sizeof(char));
            /* Check that memory was reallocated correctly. */
            if (temp == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            resB->fResPath = temp;
        }
    }
    uprv_strcpy(resB->fResPath + resPathLenOrig, toAdd);
}

static void ures_freeResPath(UResourceBundle *resB) {
    if (resB->fResPath && resB->fResPath != resB->fResBuf) {
        uprv_free(resB->fResPath);
    }
    resB->fResPath = NULL;
    resB->fResPathLen = 0;
}

static void
ures_closeBundle(UResourceBundle* resB, UBool freeBundleObj)
{
    if(resB != NULL) {
        if(resB->fData != NULL) {
            entryClose(resB->fData);
        }
        if(resB->fVersion != NULL) {
            uprv_free(resB->fVersion);
        }
        ures_freeResPath(resB);

        if(ures_isStackObject(resB) == FALSE && freeBundleObj) {
            uprv_free(resB);
        }
#if 0 /*U_DEBUG*/
        else {
            /* poison the data */
            uprv_memset(resB, -1, sizeof(UResourceBundle));
        }
#endif
    }
}

U_CAPI void  U_EXPORT2
ures_close(UResourceBundle* resB)
{
    ures_closeBundle(resB, TRUE);
}

static UResourceBundle *init_resb_result(const ResourceData *rdata, Resource r,
                                         const char *key, int32_t idx, UResourceDataEntry *realData,
                                         const UResourceBundle *parent, int32_t noAlias,
                                         UResourceBundle *resB, UErrorCode *status)
{
    if(status == NULL || U_FAILURE(*status)) {
        return resB;
    }
    if (parent == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    if(RES_GET_TYPE(r) == URES_ALIAS) { /* This is an alias, need to exchange with real data */
        if(noAlias < URES_MAX_ALIAS_LEVEL) {
            int32_t len = 0;
            const UChar *alias = res_getAlias(rdata, r, &len);
            if(len > 0) {
                /* we have an alias, now let's cut it up */
                char stackAlias[200];
                char *chAlias = NULL, *path = NULL, *locale = NULL, *keyPath = NULL;
                int32_t capacity;

                /*
                * Allocate enough space for both the char * version
                * of the alias and parent->fResPath.
                *
                * We do this so that res_findResource() can modify the path,
                * which allows us to remove redundant _res_findResource() variants
                * in uresdata.c.
                * res_findResource() now NUL-terminates each segment so that table keys
                * can always be compared with strcmp() instead of strncmp().
                * Saves code there and simplifies testing and code coverage.
                *
                * markus 2003oct17
                */
                ++len; /* count the terminating NUL */
                if(parent->fResPath != NULL) {
                    capacity = (int32_t)uprv_strlen(parent->fResPath) + 1;
                } else {
                    capacity = 0;
                }
                if(capacity < len) {
                    capacity = len;
                }
                if(capacity <= (int32_t)sizeof(stackAlias)) {
                    capacity = (int32_t)sizeof(stackAlias);
                    chAlias = stackAlias;
                } else {
                    chAlias = (char *)uprv_malloc(capacity);
                    /* test for NULL */
                    if(chAlias == NULL) {
                        *status = U_MEMORY_ALLOCATION_ERROR;
                        return NULL;
                    }
                }
                u_UCharsToChars(alias, chAlias, len);

                if(*chAlias == RES_PATH_SEPARATOR) {
                    /* there is a path included */
                    locale = uprv_strchr(chAlias+1, RES_PATH_SEPARATOR);
                    if(locale == NULL) {
                        locale = uprv_strchr(chAlias, 0); /* avoid locale == NULL to make code below work */
                    } else {
                        *locale = 0;
                        locale++;
                    }
                    path = chAlias+1;
                    if(uprv_strcmp(path, "LOCALE") == 0) {
                        /* this is an XPath alias, starting with "/LOCALE/" */
                        /* it contains the path to a resource which should be looked up */
                        /* starting in the requested locale */
                        keyPath = locale;
                        locale = parent->fTopLevelData->fName; /* this is the requested locale's name */
                        path = realData->fPath; /* we will be looking in the same package */
                    } else {
                        if(uprv_strcmp(path, "ICUDATA") == 0) { /* want ICU data */
                            path = NULL;
                        }
                        keyPath = uprv_strchr(locale, RES_PATH_SEPARATOR);
                        if(keyPath) {
                            *keyPath = 0;
                            keyPath++;
                        }
                    }
                } else {
                    /* no path, start with a locale */
                    locale = chAlias;
                    keyPath = uprv_strchr(locale, RES_PATH_SEPARATOR);
                    if(keyPath) {
                        *keyPath = 0;
                        keyPath++;
                    }
                    path = realData->fPath;
                }


                {
                    /* got almost everything, let's try to open */
                    /* first, open the bundle with real data */
                    UResourceBundle *result = resB;
                    const char* temp = NULL;
                    UErrorCode intStatus = U_ZERO_ERROR;
                    UResourceBundle *mainRes = ures_openDirect(path, locale, &intStatus);
                    if(U_SUCCESS(intStatus)) {
                        if(keyPath == NULL) {
                            /* no key path. This means that we are going to
                            * to use the corresponding resource from
                            * another bundle
                            */
                            /* first, we are going to get a corresponding parent
                            * resource to the one we are searching.
                            */
                            char *aKey = parent->fResPath;
                            if(aKey) {
                                uprv_strcpy(chAlias, aKey); /* allocated large enough above */
                                aKey = chAlias;
                                r = res_findResource(&(mainRes->fResData), mainRes->fRes, &aKey, &temp);
                            } else {
                                r = mainRes->fRes;
                            }
                            if(key) {
                                /* we need to make keyPath from parent's fResPath and
                                * current key, if there is a key associated
                                */
                                len = (int32_t)(uprv_strlen(key) + 1);
                                if(len > capacity) {
                                    capacity = len;
                                    if(chAlias == stackAlias) {
                                        chAlias = (char *)uprv_malloc(capacity);
                                    } else {
                                        chAlias = (char *)uprv_realloc(chAlias, capacity);
                                    }
                                    if(chAlias == NULL) {
                                        ures_close(mainRes);
                                        *status = U_MEMORY_ALLOCATION_ERROR;
                                        return NULL;
                                    }
                                }
                                uprv_memcpy(chAlias, key, len);
                                aKey = chAlias;
                                r = res_findResource(&(mainRes->fResData), r, &aKey, &temp);
                            } else if(idx != -1) {
                                /* if there is no key, but there is an index, try to get by the index */
                                /* here we have either a table or an array, so get the element */
                                int32_t type = RES_GET_TYPE(r);
                                if(URES_IS_TABLE(type)) {
                                    r = res_getTableItemByIndex(&(mainRes->fResData), r, idx, (const char **)&aKey);
                                } else { /* array */
                                    r = res_getArrayItem(&(mainRes->fResData), r, idx);
                                }
                            }
                            if(r != RES_BOGUS) {
                                result = init_resb_result(&(mainRes->fResData), r, temp, -1, mainRes->fData, mainRes, noAlias+1, resB, status);
                            } else {
                                *status = U_MISSING_RESOURCE_ERROR;
                                result = resB;
                            }
                        } else {
                            /* this one is a bit trickier.
                            * we start finding keys, but after we resolve one alias, the path might continue.
                            * Consider:
                            *     aliastest:alias { "testtypes/anotheralias/Sequence" }
                            *     anotheralias:alias { "/ICUDATA/sh/CollationElements" }
                            * aliastest resource should finally have the sequence, not collation elements.
                            */
                            UResourceDataEntry *dataEntry = mainRes->fData;
                            char stackPath[URES_MAX_BUFFER_SIZE];
                            char *pathBuf = stackPath, *myPath = pathBuf;
                            if(uprv_strlen(keyPath) >= UPRV_LENGTHOF(stackPath)) {
                                pathBuf = (char *)uprv_malloc((uprv_strlen(keyPath)+1)*sizeof(char));
                                if(pathBuf == NULL) {
                                    *status = U_MEMORY_ALLOCATION_ERROR;
                                    ures_close(mainRes);
                                    return NULL;
                                }
                            }
                            uprv_strcpy(pathBuf, keyPath);
                            result = mainRes;
                            /* now we have fallback following here */
                            do {
                                r = dataEntry->fData.rootRes;
                                /* this loop handles 'found' resources over several levels */
                                while(*myPath && U_SUCCESS(*status)) {
                                    r = res_findResource(&(dataEntry->fData), r, &myPath, &temp);
                                    if(r != RES_BOGUS) { /* found a resource, but it might be an indirection */
                                        resB = init_resb_result(&(dataEntry->fData), r, temp, -1, dataEntry, result, noAlias+1, resB, status);
                                        result = resB;
                                        if(result) {
                                            r = result->fRes; /* switch to a new resource, possibly a new tree */
                                            dataEntry = result->fData;
                                        }
                                    } else { /* no resource found, we don't really want to look anymore on this level */
                                        break;
                                    }
                                }
                                dataEntry = dataEntry->fParent;
                                uprv_strcpy(pathBuf, keyPath);
                                myPath = pathBuf;
                            } while(r == RES_BOGUS && dataEntry != NULL);
                            if(r == RES_BOGUS) {
                                *status = U_MISSING_RESOURCE_ERROR;
                                result = resB;
                            }
                            if(pathBuf != stackPath) {
                                uprv_free(pathBuf);
                            }
                        }
                    } else { /* we failed to open the resource we're aliasing to */
                        *status = intStatus;
                    }
                    if(chAlias != stackAlias) {
                        uprv_free(chAlias);
                    }
                    if(mainRes != result) {
                        ures_close(mainRes);
                    }
                    ResourceTracer(resB).maybeTrace("getalias");
                    return result;
                }
            } else {
                /* bad alias, should be an error */
                *status = U_ILLEGAL_ARGUMENT_ERROR;
                return resB;
            }
        } else {
            *status = U_TOO_MANY_ALIASES_ERROR;
            return resB;
        }
    }
    if(resB == NULL) {
        resB = (UResourceBundle *)uprv_malloc(sizeof(UResourceBundle));
        /* test for NULL */
        if (resB == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        ures_setIsStackObject(resB, FALSE);
        resB->fResPath = NULL;
        resB->fResPathLen = 0;
    } else {
        if(resB->fData != NULL) {
            entryClose(resB->fData);
        }
        if(resB->fVersion != NULL) {
            uprv_free(resB->fVersion);
        }
        /*
        weiv: if stack object was passed in, it doesn't really need to be reinited,
        since the purpose of initing is to remove stack junk. However, at this point
        we would not do anything to an allocated object, so stack object should be
        treated the same
        */
        /*
        if(ures_isStackObject(resB) != FALSE) {
        ures_initStackObject(resB);
        }
        */
        if(parent != resB) {
            ures_freeResPath(resB);
        }
    }
    resB->fData = realData;
    entryIncrease(resB->fData);
    resB->fHasFallback = FALSE;
    resB->fIsTopLevel = FALSE;
    resB->fIndex = -1;
    resB->fKey = key;
    /*resB->fParentRes = parent;*/
    resB->fTopLevelData = parent->fTopLevelData;
    if(parent->fResPath && parent != resB) {
        ures_appendResPath(resB, parent->fResPath, parent->fResPathLen, status);
    }
    if(key != NULL) {
        ures_appendResPath(resB, key, (int32_t)uprv_strlen(key), status);
        if(resB->fResPath[resB->fResPathLen-1] != RES_PATH_SEPARATOR) {
            ures_appendResPath(resB, RES_PATH_SEPARATOR_S, 1, status);
        }
    } else if(idx >= 0) {
        char buf[256];
        int32_t len = T_CString_integerToString(buf, idx, 10);
        ures_appendResPath(resB, buf, len, status);
        if(resB->fResPath[resB->fResPathLen-1] != RES_PATH_SEPARATOR) {
            ures_appendResPath(resB, RES_PATH_SEPARATOR_S, 1, status);
        }
    }
    /* Make sure that Purify doesn't complain about uninitialized memory copies. */
    {
        int32_t usedLen = ((resB->fResBuf == resB->fResPath) ? resB->fResPathLen : 0);
        uprv_memset(resB->fResBuf + usedLen, 0, sizeof(resB->fResBuf) - usedLen);
    }

    resB->fVersion = NULL;
    resB->fRes = r;
    /*resB->fParent = parent->fRes;*/
    uprv_memmove(&resB->fResData, rdata, sizeof(ResourceData));
    resB->fSize = res_countArrayItems(&(resB->fResData), resB->fRes);
    ResourceTracer(resB).trace("get");
    return resB;
}

UResourceBundle *ures_copyResb(UResourceBundle *r, const UResourceBundle *original, UErrorCode *status) {
    UBool isStackObject;
    if(U_FAILURE(*status) || r == original) {
        return r;
    }
    if(original != NULL) {
        if(r == NULL) {
            isStackObject = FALSE;
            r = (UResourceBundle *)uprv_malloc(sizeof(UResourceBundle));
            /* test for NULL */
            if (r == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
                return NULL;
            }
        } else {
            isStackObject = ures_isStackObject(r);
            ures_closeBundle(r, FALSE);
        }
        uprv_memcpy(r, original, sizeof(UResourceBundle));
        r->fResPath = NULL;
        r->fResPathLen = 0;
        if(original->fResPath) {
            ures_appendResPath(r, original->fResPath, original->fResPathLen, status);
        }
        ures_setIsStackObject(r, isStackObject);
        if(r->fData != NULL) {
            entryIncrease(r->fData);
        }
    }
    return r;
}

/**
 * Functions to retrieve data from resource bundles.
 */

U_CAPI const UChar* U_EXPORT2 ures_getString(const UResourceBundle* resB, int32_t* len, UErrorCode* status) {
    const UChar *s;
    if (status==NULL || U_FAILURE(*status)) {
        return NULL;
    }
    if(resB == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    s = res_getString({resB}, &(resB->fResData), resB->fRes, len);
    if (s == NULL) {
        *status = U_RESOURCE_TYPE_MISMATCH;
    }
    return s;
}

static const char *
ures_toUTF8String(const UChar *s16, int32_t length16,
                  char *dest, int32_t *pLength,
                  UBool forceCopy,
                  UErrorCode *status) {
    int32_t capacity;

    if (U_FAILURE(*status)) {
        return NULL;
    }
    if (pLength != NULL) {
        capacity = *pLength;
    } else {
        capacity = 0;
    }
    if (capacity < 0 || (capacity > 0 && dest == NULL)) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if (length16 == 0) {
        /* empty string, return as read-only pointer */
        if (pLength != NULL) {
            *pLength = 0;
        }
        if (forceCopy) {
            u_terminateChars(dest, capacity, 0, status);
            return dest;
        } else {
            return "";
        }
    } else {
        /* We need to transform the string to the destination buffer. */
        if (capacity < length16) {
            /* No chance for the string to fit. Pure preflighting. */
            return u_strToUTF8(NULL, 0, pLength, s16, length16, status);
        }
        if (!forceCopy && (length16 <= 0x2aaaaaaa)) {
            /*
             * We know the string will fit into dest because each UChar turns
             * into at most three UTF-8 bytes. Fill the latter part of dest
             * so that callers do not expect to use dest as a string pointer,
             * hopefully leading to more robust code for when resource bundles
             * may store UTF-8 natively.
             * (In which case dest would not be used at all.)
             *
             * We do not do this if forceCopy=TRUE because then the caller
             * expects the string to start exactly at dest.
             *
             * The test above for <= 0x2aaaaaaa prevents overflows.
             * The +1 is for the NUL terminator.
             */
            int32_t maxLength = 3 * length16 + 1;
            if (capacity > maxLength) {
                dest += capacity - maxLength;
                capacity = maxLength;
            }
        }
        return u_strToUTF8(dest, capacity, pLength, s16, length16, status);
    }
}

U_CAPI const char * U_EXPORT2
ures_getUTF8String(const UResourceBundle *resB,
                   char *dest, int32_t *pLength,
                   UBool forceCopy,
                   UErrorCode *status) {
    int32_t length16;
    const UChar *s16 = ures_getString(resB, &length16, status);
    return ures_toUTF8String(s16, length16, dest, pLength, forceCopy, status);
}

U_CAPI const uint8_t* U_EXPORT2 ures_getBinary(const UResourceBundle* resB, int32_t* len,
                                               UErrorCode*               status) {
  const uint8_t *p;
  if (status==NULL || U_FAILURE(*status)) {
    return NULL;
  }
  if(resB == NULL) {
    *status = U_ILLEGAL_ARGUMENT_ERROR;
    return NULL;
  }
  p = res_getBinary({resB}, &(resB->fResData), resB->fRes, len);
  if (p == NULL) {
    *status = U_RESOURCE_TYPE_MISMATCH;
  }
  return p;
}

U_CAPI const int32_t* U_EXPORT2 ures_getIntVector(const UResourceBundle* resB, int32_t* len,
                                                   UErrorCode*               status) {
  const int32_t *p;
  if (status==NULL || U_FAILURE(*status)) {
    return NULL;
  }
  if(resB == NULL) {
    *status = U_ILLEGAL_ARGUMENT_ERROR;
    return NULL;
  }
  p = res_getIntVector({resB}, &(resB->fResData), resB->fRes, len);
  if (p == NULL) {
    *status = U_RESOURCE_TYPE_MISMATCH;
  }
  return p;
}

/* this function returns a signed integer */
/* it performs sign extension */
U_CAPI int32_t U_EXPORT2 ures_getInt(const UResourceBundle* resB, UErrorCode *status) {
  if (status==NULL || U_FAILURE(*status)) {
    return 0xffffffff;
  }
  if(resB == NULL) {
    *status = U_ILLEGAL_ARGUMENT_ERROR;
    return 0xffffffff;
  }
  if(RES_GET_TYPE(resB->fRes) != URES_INT) {
    *status = U_RESOURCE_TYPE_MISMATCH;
    return 0xffffffff;
  }
  return res_getInt({resB}, resB->fRes);
}

U_CAPI uint32_t U_EXPORT2 ures_getUInt(const UResourceBundle* resB, UErrorCode *status) {
  if (status==NULL || U_FAILURE(*status)) {
    return 0xffffffff;
  }
  if(resB == NULL) {
    *status = U_ILLEGAL_ARGUMENT_ERROR;
    return 0xffffffff;
  }
  if(RES_GET_TYPE(resB->fRes) != URES_INT) {
    *status = U_RESOURCE_TYPE_MISMATCH;
    return 0xffffffff;
  }
  return res_getUInt({resB}, resB->fRes);
}

U_CAPI UResType U_EXPORT2 ures_getType(const UResourceBundle *resB) {
  if(resB == NULL) {
    return URES_NONE;
  }
  return res_getPublicType(resB->fRes);
}

U_CAPI const char * U_EXPORT2 ures_getKey(const UResourceBundle *resB) {
  //
  // TODO: Trace ures_getKey? I guess not usually.
  //
  // We usually get the key string to decide whether we want the value, or to
  // make a key-value pair. Tracing the value should suffice.
  //
  // However, I believe we have some data (e.g., in res_index) where the key
  // strings are the data. Tracing the enclosing table should suffice.
  //
  if(resB == NULL) {
    return NULL;
  }
  return(resB->fKey);
}

U_CAPI int32_t U_EXPORT2 ures_getSize(const UResourceBundle *resB) {
  if(resB == NULL) {
    return 0;
  }

  return resB->fSize;
}

static const UChar* ures_getStringWithAlias(const UResourceBundle *resB, Resource r, int32_t sIndex, int32_t *len, UErrorCode *status) {
  if(RES_GET_TYPE(r) == URES_ALIAS) {
    const UChar* result = 0;
    UResourceBundle *tempRes = ures_getByIndex(resB, sIndex, NULL, status);
    result = ures_getString(tempRes, len, status);
    ures_close(tempRes);
    return result;
  } else {
    return res_getString({resB, sIndex}, &(resB->fResData), r, len);
  }
}

U_CAPI void U_EXPORT2 ures_resetIterator(UResourceBundle *resB){
  if(resB == NULL) {
    return;
  }
  resB->fIndex = -1;
}

U_CAPI UBool U_EXPORT2 ures_hasNext(const UResourceBundle *resB) {
  if(resB == NULL) {
    return FALSE;
  }
  return (UBool)(resB->fIndex < resB->fSize-1);
}

U_CAPI const UChar* U_EXPORT2 ures_getNextString(UResourceBundle *resB, int32_t* len, const char ** key, UErrorCode *status) {
  Resource r = RES_BOGUS;

  if (status==NULL || U_FAILURE(*status)) {
    return NULL;
  }
  if(resB == NULL) {
    *status = U_ILLEGAL_ARGUMENT_ERROR;
    return NULL;
  }

  if(resB->fIndex == resB->fSize-1) {
    *status = U_INDEX_OUTOFBOUNDS_ERROR;
  } else {
    resB->fIndex++;
    switch(RES_GET_TYPE(resB->fRes)) {
    case URES_STRING:
    case URES_STRING_V2:
      return res_getString({resB}, &(resB->fResData), resB->fRes, len);
    case URES_TABLE:
    case URES_TABLE16:
    case URES_TABLE32:
      r = res_getTableItemByIndex(&(resB->fResData), resB->fRes, resB->fIndex, key);
      if(r == RES_BOGUS && resB->fHasFallback) {
        /* TODO: do the fallback */
      }
      return ures_getStringWithAlias(resB, r, resB->fIndex, len, status);
    case URES_ARRAY:
    case URES_ARRAY16:
      r = res_getArrayItem(&(resB->fResData), resB->fRes, resB->fIndex);
      if(r == RES_BOGUS && resB->fHasFallback) {
        /* TODO: do the fallback */
      }
      return ures_getStringWithAlias(resB, r, resB->fIndex, len, status);
    case URES_ALIAS:
      return ures_getStringWithAlias(resB, resB->fRes, resB->fIndex, len, status);
    case URES_INT:
    case URES_BINARY:
    case URES_INT_VECTOR:
        *status = U_RESOURCE_TYPE_MISMATCH;
        U_FALLTHROUGH;
    default:
      return NULL;
    }
  }

  return NULL;
}

U_CAPI UResourceBundle* U_EXPORT2 ures_getNextResource(UResourceBundle *resB, UResourceBundle *fillIn, UErrorCode *status) {
    const char *key = NULL;
    Resource r = RES_BOGUS;

    if (status==NULL || U_FAILURE(*status)) {
            /*return NULL;*/
            return fillIn;
    }
    if(resB == NULL) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
            /*return NULL;*/
            return fillIn;
    }

    if(resB->fIndex == resB->fSize-1) {
      *status = U_INDEX_OUTOFBOUNDS_ERROR;
      /*return NULL;*/
    } else {
        resB->fIndex++;
        switch(RES_GET_TYPE(resB->fRes)) {
        case URES_INT:
        case URES_BINARY:
        case URES_STRING:
        case URES_STRING_V2:
        case URES_INT_VECTOR:
            return ures_copyResb(fillIn, resB, status);
        case URES_TABLE:
        case URES_TABLE16:
        case URES_TABLE32:
            r = res_getTableItemByIndex(&(resB->fResData), resB->fRes, resB->fIndex, &key);
            if(r == RES_BOGUS && resB->fHasFallback) {
                /* TODO: do the fallback */
            }
            return init_resb_result(&(resB->fResData), r, key, resB->fIndex, resB->fData, resB, 0, fillIn, status);
        case URES_ARRAY:
        case URES_ARRAY16:
            r = res_getArrayItem(&(resB->fResData), resB->fRes, resB->fIndex);
            if(r == RES_BOGUS && resB->fHasFallback) {
                /* TODO: do the fallback */
            }
            return init_resb_result(&(resB->fResData), r, key, resB->fIndex, resB->fData, resB, 0, fillIn, status);
        default:
            /*return NULL;*/
            return fillIn;
        }
    }
    /*return NULL;*/
    return fillIn;
}

U_CAPI UResourceBundle* U_EXPORT2 ures_getByIndex(const UResourceBundle *resB, int32_t indexR, UResourceBundle *fillIn, UErrorCode *status) {
    const char* key = NULL;
    Resource r = RES_BOGUS;

    if (status==NULL || U_FAILURE(*status)) {
        /*return NULL;*/
        return fillIn;
    }
    if(resB == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        /*return NULL;*/
        return fillIn;
    }

    if(indexR >= 0 && resB->fSize > indexR) {
        switch(RES_GET_TYPE(resB->fRes)) {
        case URES_INT:
        case URES_BINARY:
        case URES_STRING:
        case URES_STRING_V2:
        case URES_INT_VECTOR:
            return ures_copyResb(fillIn, resB, status);
        case URES_TABLE:
        case URES_TABLE16:
        case URES_TABLE32:
            r = res_getTableItemByIndex(&(resB->fResData), resB->fRes, indexR, &key);
            if(r == RES_BOGUS && resB->fHasFallback) {
                /* TODO: do the fallback */
            }
            return init_resb_result(&(resB->fResData), r, key, indexR, resB->fData, resB, 0, fillIn, status);
        case URES_ARRAY:
        case URES_ARRAY16:
            r = res_getArrayItem(&(resB->fResData), resB->fRes, indexR);
            if(r == RES_BOGUS && resB->fHasFallback) {
                /* TODO: do the fallback */
            }
            return init_resb_result(&(resB->fResData), r, key, indexR, resB->fData, resB, 0, fillIn, status);
        default:
            /*return NULL;*/
            return fillIn;
        }
    } else {
        *status = U_MISSING_RESOURCE_ERROR;
    }
    /*return NULL;*/
    return fillIn;
}

U_CAPI const UChar* U_EXPORT2 ures_getStringByIndex(const UResourceBundle *resB, int32_t indexS, int32_t* len, UErrorCode *status) {
    const char* key = NULL;
    Resource r = RES_BOGUS;

    if (status==NULL || U_FAILURE(*status)) {
        return NULL;
    }
    if(resB == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if(indexS >= 0 && resB->fSize > indexS) {
        switch(RES_GET_TYPE(resB->fRes)) {
        case URES_STRING:
        case URES_STRING_V2:
            return res_getString({resB}, &(resB->fResData), resB->fRes, len);
        case URES_TABLE:
        case URES_TABLE16:
        case URES_TABLE32:
            r = res_getTableItemByIndex(&(resB->fResData), resB->fRes, indexS, &key);
            if(r == RES_BOGUS && resB->fHasFallback) {
                /* TODO: do the fallback */
            }
            return ures_getStringWithAlias(resB, r, indexS, len, status);
        case URES_ARRAY:
        case URES_ARRAY16:
            r = res_getArrayItem(&(resB->fResData), resB->fRes, indexS);
            if(r == RES_BOGUS && resB->fHasFallback) {
                /* TODO: do the fallback */
            }
            return ures_getStringWithAlias(resB, r, indexS, len, status);
        case URES_ALIAS:
            return ures_getStringWithAlias(resB, resB->fRes, indexS, len, status);
        case URES_INT:
        case URES_BINARY:
        case URES_INT_VECTOR:
            *status = U_RESOURCE_TYPE_MISMATCH;
            break;
        default:
          /* must not occur */
          *status = U_INTERNAL_PROGRAM_ERROR;
          break;
        }
    } else {
        *status = U_MISSING_RESOURCE_ERROR;
    }
    return NULL;
}

U_CAPI const char * U_EXPORT2
ures_getUTF8StringByIndex(const UResourceBundle *resB,
                          int32_t idx,
                          char *dest, int32_t *pLength,
                          UBool forceCopy,
                          UErrorCode *status) {
    int32_t length16;
    const UChar *s16 = ures_getStringByIndex(resB, idx, &length16, status);
    return ures_toUTF8String(s16, length16, dest, pLength, forceCopy, status);
}

/*U_CAPI const char *ures_getResPath(UResourceBundle *resB) {
  return resB->fResPath;
}*/

U_CAPI UResourceBundle* U_EXPORT2
ures_findResource(const char* path, UResourceBundle *fillIn, UErrorCode *status)
{
  UResourceBundle *first = NULL;
  UResourceBundle *result = fillIn;
  char *packageName = NULL;
  char *pathToResource = NULL, *save = NULL;
  char *locale = NULL, *localeEnd = NULL;
  int32_t length;

  if(status == NULL || U_FAILURE(*status)) {
    return result;
  }

  length = (int32_t)(uprv_strlen(path)+1);
  save = pathToResource = (char *)uprv_malloc(length*sizeof(char));
  /* test for NULL */
  if(pathToResource == NULL) {
    *status = U_MEMORY_ALLOCATION_ERROR;
    return result;
  }
  uprv_memcpy(pathToResource, path, length);

  locale = pathToResource;
  if(*pathToResource == RES_PATH_SEPARATOR) { /* there is a path specification */
    pathToResource++;
    packageName = pathToResource;
    pathToResource = uprv_strchr(pathToResource, RES_PATH_SEPARATOR);
    if(pathToResource == NULL) {
      *status = U_ILLEGAL_ARGUMENT_ERROR;
    } else {
      *pathToResource = 0;
      locale = pathToResource+1;
    }
  }

  localeEnd = uprv_strchr(locale, RES_PATH_SEPARATOR);
  if(localeEnd != NULL) {
    *localeEnd = 0;
  }

  first = ures_open(packageName, locale, status);

  if(U_SUCCESS(*status)) {
    if(localeEnd) {
      result = ures_findSubResource(first, localeEnd+1, fillIn, status);
    } else {
      result = ures_copyResb(fillIn, first, status);
    }
    ures_close(first);
  }
  uprv_free(save);
  return result;
}

U_CAPI UResourceBundle* U_EXPORT2
ures_findSubResource(const UResourceBundle *resB, char* path, UResourceBundle *fillIn, UErrorCode *status)
{
  Resource res = RES_BOGUS;
  UResourceBundle *result = fillIn;
  const char *key;

  if(status == NULL || U_FAILURE(*status)) {
    return result;
  }

  /* here we do looping and circular alias checking */
  /* this loop is here because aliasing is resolved on this level, not on res level */
  /* so, when we encounter an alias, it is not an aggregate resource, so we return */
  do {
    res = res_findResource(&(resB->fResData), resB->fRes, &path, &key);
    if(res != RES_BOGUS) {
        result = init_resb_result(&(resB->fResData), res, key, -1, resB->fData, resB, 0, fillIn, status);
        resB = result;
    } else {
        *status = U_MISSING_RESOURCE_ERROR;
        break;
    }
  } while(*path); /* there is more stuff in the path */

  return result;
}
U_INTERNAL const UChar* U_EXPORT2
ures_getStringByKeyWithFallback(const UResourceBundle *resB,
                                const char* inKey,
                                int32_t* len,
                                UErrorCode *status) {

    UResourceBundle stack;
    const UChar* retVal = NULL;
    ures_initStackObject(&stack);
    ures_getByKeyWithFallback(resB, inKey, &stack, status);
    int32_t length;
    retVal = ures_getString(&stack, &length, status);
    ures_close(&stack);
    if (U_FAILURE(*status)) {
        return NULL;
    }
    if (length == 3 && retVal[0] == EMPTY_SET && retVal[1] == EMPTY_SET && retVal[2] == EMPTY_SET ) {
        retVal = NULL;
        length = 0;
        *status = U_MISSING_RESOURCE_ERROR;
    }
    if (len != NULL) {
        *len = length;
    }
    return retVal;
}

/*
  Like res_getTableItemByKey but accepts full paths like "NumberElements/latn/patternsShort".
*/
static Resource getTableItemByKeyPath(const ResourceData *pResData, Resource table, const char *key) {
  Resource resource = table;  /* The current resource */
  icu::CharString path;
  UErrorCode errorCode = U_ZERO_ERROR;
  path.append(key, errorCode);
  if (U_FAILURE(errorCode)) { return RES_BOGUS; }
  char *pathPart = path.data();  /* Path from current resource to desired resource */
  UResType type = (UResType)RES_GET_TYPE(resource);  /* the current resource type */
  while (*pathPart && resource != RES_BOGUS && URES_IS_CONTAINER(type)) {
    char *nextPathPart = uprv_strchr(pathPart, RES_PATH_SEPARATOR);
    if (nextPathPart != NULL) {
      *nextPathPart = 0;  /* Terminating null for this part of path. */
      nextPathPart++;
    } else {
      nextPathPart = uprv_strchr(pathPart, 0);
    }
    int32_t t;
    const char *pathP = pathPart;
    resource = res_getTableItemByKey(pResData, resource, &t, &pathP);
    type = (UResType)RES_GET_TYPE(resource);
    pathPart = nextPathPart;
  }
  if (*pathPart) {
    return RES_BOGUS;
  }
  return resource;
}

U_CAPI UResourceBundle* U_EXPORT2
ures_getByKeyWithFallback(const UResourceBundle *resB,
                          const char* inKey,
                          UResourceBundle *fillIn,
                          UErrorCode *status) {
    Resource res = RES_BOGUS, rootRes = RES_BOGUS;
    /*UResourceDataEntry *realData = NULL;*/
    UResourceBundle *helper = NULL;

    if (status==NULL || U_FAILURE(*status)) {
        return fillIn;
    }
    if(resB == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return fillIn;
    }

    int32_t type = RES_GET_TYPE(resB->fRes);
    if(URES_IS_TABLE(type)) {
        res = getTableItemByKeyPath(&(resB->fResData), resB->fRes, inKey);
        const char* key = inKey;
        if(res == RES_BOGUS) {
            UResourceDataEntry *dataEntry = resB->fData;
            CharString path;
            char *myPath = NULL;
            const char* resPath = resB->fResPath;
            int32_t len = resB->fResPathLen;
            while(res == RES_BOGUS && dataEntry->fParent != NULL) { /* Otherwise, we'll look in parents */
                dataEntry = dataEntry->fParent;
                rootRes = dataEntry->fData.rootRes;

                if(dataEntry->fBogus == U_ZERO_ERROR) {
                    path.clear();
                    if (len > 0) {
                        path.append(resPath, len, *status);
                    }
                    path.append(inKey, *status);
                    if (U_FAILURE(*status)) {
                        ures_close(helper);
                        return fillIn;
                    }
                    myPath = path.data();
                    key = inKey;
                    do {
                        res = res_findResource(&(dataEntry->fData), rootRes, &myPath, &key);
                        if (RES_GET_TYPE(res) == URES_ALIAS && *myPath) {
                            /* We hit an alias, but we didn't finish following the path. */
                            helper = init_resb_result(&(dataEntry->fData), res, NULL, -1, dataEntry, resB, 0, helper, status);
                            /*helper = init_resb_result(&(dataEntry->fData), res, inKey, -1, dataEntry, resB, 0, helper, status);*/
                            if(helper) {
                              dataEntry = helper->fData;
                              rootRes = helper->fRes;
                              resPath = helper->fResPath;
                              len = helper->fResPathLen;

                            } else {
                              break;
                            }
                        }
                    } while(*myPath); /* Continue until the whole path is consumed */
                }
            }
            /*const ResourceData *rd = getFallbackData(resB, &key, &realData, &res, status);*/
            if(res != RES_BOGUS) {
              /* check if resB->fResPath gives the right name here */
                if(uprv_strcmp(dataEntry->fName, uloc_getDefault())==0 || uprv_strcmp(dataEntry->fName, kRootLocaleName)==0) {
                    *status = U_USING_DEFAULT_WARNING;
                } else {
                    *status = U_USING_FALLBACK_WARNING;
                }

                fillIn = init_resb_result(&(dataEntry->fData), res, inKey, -1, dataEntry, resB, 0, fillIn, status);
            } else {
                *status = U_MISSING_RESOURCE_ERROR;
            }
        } else {
            fillIn = init_resb_result(&(resB->fResData), res, key, -1, resB->fData, resB, 0, fillIn, status);
        }
    }
    else {
        *status = U_RESOURCE_TYPE_MISMATCH;
    }
    ures_close(helper);
    return fillIn;
}

namespace {

void getAllItemsWithFallback(
        const UResourceBundle *bundle, ResourceDataValue &value,
        ResourceSink &sink,
        UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return; }
    // We recursively enumerate child-first,
    // only storing parent items in the absence of child items.
    // The sink needs to store a placeholder value for the no-fallback/no-inheritance marker
    // to prevent a parent item from being stored.
    //
    // It would be possible to recursively enumerate parent-first,
    // overriding parent items with child items.
    // When the sink sees the no-fallback/no-inheritance marker,
    // then it would remove the parent's item.
    // We would deserialize parent values even though they are overridden in a child bundle.
    value.setData(&bundle->fResData);
    UResourceDataEntry *parentEntry = bundle->fData->fParent;
    UBool hasParent = parentEntry != NULL && U_SUCCESS(parentEntry->fBogus);
    value.setResource(bundle->fRes, ResourceTracer(bundle));
    sink.put(bundle->fKey, value, !hasParent, errorCode);
    if (hasParent) {
        // We might try to query the sink whether
        // any fallback from the parent bundle is still possible.

        // Turn the parent UResourceDataEntry into a UResourceBundle,
        // much like in ures_openWithType().
        // TODO: See if we can refactor ures_getByKeyWithFallback()
        // and pull out an inner function that takes and returns a UResourceDataEntry
        // so that we need not create UResourceBundle objects.
        UResourceBundle parentBundle;
        ures_initStackObject(&parentBundle);
        parentBundle.fTopLevelData = parentBundle.fData = parentEntry;
        // TODO: What is the difference between bundle fData and fTopLevelData?
        uprv_memcpy(&parentBundle.fResData, &parentEntry->fData, sizeof(ResourceData));
        // TODO: Try to replace bundle.fResData with just using bundle.fData->fData.
        parentBundle.fHasFallback = !parentBundle.fResData.noFallback;
        parentBundle.fIsTopLevel = TRUE;
        parentBundle.fRes = parentBundle.fResData.rootRes;
        parentBundle.fSize = res_countArrayItems(&(parentBundle.fResData), parentBundle.fRes);
        parentBundle.fIndex = -1;
        entryIncrease(parentEntry);

        // Look up the container item in the parent bundle.
        UResourceBundle containerBundle;
        ures_initStackObject(&containerBundle);
        const UResourceBundle *rb;
        UErrorCode pathErrorCode = U_ZERO_ERROR;  // Ignore if parents up to root do not have this path.
        if (bundle->fResPath == NULL || *bundle->fResPath == 0) {
            rb = &parentBundle;
        } else {
            rb = ures_getByKeyWithFallback(&parentBundle, bundle->fResPath,
                                           &containerBundle, &pathErrorCode);
        }
        if (U_SUCCESS(pathErrorCode)) {
            getAllItemsWithFallback(rb, value, sink, errorCode);
        }
        ures_close(&containerBundle);
        ures_close(&parentBundle);
    }
}

}  // namespace

// Requires a ResourceDataValue fill-in, so that we need not cast from a ResourceValue.
// Unfortunately, the caller must know which subclass to make and pass in.
// Alternatively, we could make it as polymorphic as in Java by
// returning a ResourceValue pointer (possibly wrapped into a LocalPointer)
// that the caller then owns.
//
// Also requires a UResourceBundle fill-in, so that the value's ResourceTracer
// can point to a non-local bundle.
// Without tracing, the child bundle could be a function-local object.
U_CAPI void U_EXPORT2
ures_getValueWithFallback(const UResourceBundle *bundle, const char *path,
                          UResourceBundle *tempFillIn,
                          ResourceDataValue &value, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return; }
    if (path == nullptr) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    const UResourceBundle *rb;
    if (*path == 0) {
        // empty path
        rb = bundle;
    } else {
        rb = ures_getByKeyWithFallback(bundle, path, tempFillIn, &errorCode);
        if (U_FAILURE(errorCode)) {
            return;
        }
    }
    value.setData(&rb->fResData);
    value.setResource(rb->fRes, ResourceTracer(rb));
}

U_CAPI void U_EXPORT2
ures_getAllItemsWithFallback(const UResourceBundle *bundle, const char *path,
                             icu::ResourceSink &sink, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return; }
    if (path == nullptr) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    StackUResourceBundle stackBundle;
    const UResourceBundle *rb;
    if (*path == 0) {
        // empty path
        rb = bundle;
    } else {
        rb = ures_getByKeyWithFallback(bundle, path, stackBundle.getAlias(), &errorCode);
        if (U_FAILURE(errorCode)) {
            return;
        }
    }
    // Get all table items with fallback.
    ResourceDataValue value;
    getAllItemsWithFallback(rb, value, sink, errorCode);
}

U_CAPI UResourceBundle* U_EXPORT2 ures_getByKey(const UResourceBundle *resB, const char* inKey, UResourceBundle *fillIn, UErrorCode *status) {
    Resource res = RES_BOGUS;
    UResourceDataEntry *realData = NULL;
    const char *key = inKey;

    if (status==NULL || U_FAILURE(*status)) {
        return fillIn;
    }
    if(resB == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return fillIn;
    }

    int32_t type = RES_GET_TYPE(resB->fRes);
    if(URES_IS_TABLE(type)) {
        int32_t t;
        res = res_getTableItemByKey(&(resB->fResData), resB->fRes, &t, &key);
        if(res == RES_BOGUS) {
            key = inKey;
            if(resB->fHasFallback == TRUE) {
                const ResourceData *rd = getFallbackData(resB, &key, &realData, &res, status);
                if(U_SUCCESS(*status)) {
                  /* check if resB->fResPath gives the right name here */
                    return init_resb_result(rd, res, key, -1, realData, resB, 0, fillIn, status);
                } else {
                    *status = U_MISSING_RESOURCE_ERROR;
                }
            } else {
                *status = U_MISSING_RESOURCE_ERROR;
            }
        } else {
            return init_resb_result(&(resB->fResData), res, key, -1, resB->fData, resB, 0, fillIn, status);
        }
    }
#if 0
    /* this is a kind of TODO item. If we have an array with an index table, we could do this. */
    /* not currently */
    else if(RES_GET_TYPE(resB->fRes) == URES_ARRAY && resB->fHasFallback == TRUE) {
        /* here should go a first attempt to locate the key using index table */
        const ResourceData *rd = getFallbackData(resB, &key, &realData, &res, status);
        if(U_SUCCESS(*status)) {
            return init_resb_result(rd, res, key, realData, resB, fillIn, status);
        } else {
            *status = U_MISSING_RESOURCE_ERROR;
        }
    }
#endif
    else {
        *status = U_RESOURCE_TYPE_MISMATCH;
    }
    return fillIn;
}

U_CAPI const UChar* U_EXPORT2 ures_getStringByKey(const UResourceBundle *resB, const char* inKey, int32_t* len, UErrorCode *status) {
    Resource res = RES_BOGUS;
    UResourceDataEntry *realData = NULL;
    const char* key = inKey;

    if (status==NULL || U_FAILURE(*status)) {
        return NULL;
    }
    if(resB == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    int32_t type = RES_GET_TYPE(resB->fRes);
    if(URES_IS_TABLE(type)) {
        int32_t t=0;

        res = res_getTableItemByKey(&(resB->fResData), resB->fRes, &t, &key);

        if(res == RES_BOGUS) {
            key = inKey;
            if(resB->fHasFallback == TRUE) {
                const ResourceData *rd = getFallbackData(resB, &key, &realData, &res, status);
                if(U_SUCCESS(*status)) {
                    switch (RES_GET_TYPE(res)) {
                    case URES_STRING:
                    case URES_STRING_V2:
                        return res_getString({resB, key}, rd, res, len);
                    case URES_ALIAS:
                      {
                        const UChar* result = 0;
                        UResourceBundle *tempRes = ures_getByKey(resB, inKey, NULL, status);
                        result = ures_getString(tempRes, len, status);
                        ures_close(tempRes);
                        return result;
                      }
                    default:
                        *status = U_RESOURCE_TYPE_MISMATCH;
                    }
                } else {
                    *status = U_MISSING_RESOURCE_ERROR;
                }
            } else {
                *status = U_MISSING_RESOURCE_ERROR;
            }
        } else {
            switch (RES_GET_TYPE(res)) {
            case URES_STRING:
            case URES_STRING_V2:
                return res_getString({resB, key}, &(resB->fResData), res, len);
            case URES_ALIAS:
              {
                const UChar* result = 0;
                UResourceBundle *tempRes = ures_getByKey(resB, inKey, NULL, status);
                result = ures_getString(tempRes, len, status);
                ures_close(tempRes);
                return result;
              }
            default:
                *status = U_RESOURCE_TYPE_MISMATCH;
            }
        }
    }
#if 0
    /* this is a kind of TODO item. If we have an array with an index table, we could do this. */
    /* not currently */
    else if(RES_GET_TYPE(resB->fRes) == URES_ARRAY && resB->fHasFallback == TRUE) {
        /* here should go a first attempt to locate the key using index table */
        const ResourceData *rd = getFallbackData(resB, &key, &realData, &res, status);
        if(U_SUCCESS(*status)) {
            // TODO: Tracing
            return res_getString(rd, res, len);
        } else {
            *status = U_MISSING_RESOURCE_ERROR;
        }
    }
#endif
    else {
        *status = U_RESOURCE_TYPE_MISMATCH;
    }
    return NULL;
}

U_CAPI const char * U_EXPORT2
ures_getUTF8StringByKey(const UResourceBundle *resB,
                        const char *key,
                        char *dest, int32_t *pLength,
                        UBool forceCopy,
                        UErrorCode *status) {
    int32_t length16;
    const UChar *s16 = ures_getStringByKey(resB, key, &length16, status);
    return ures_toUTF8String(s16, length16, dest, pLength, forceCopy, status);
}

/* TODO: clean from here down */

/**
 *  INTERNAL: Get the name of the first real locale (not placeholder)
 *  that has resource bundle data.
 */
U_INTERNAL const char*  U_EXPORT2
ures_getLocaleInternal(const UResourceBundle* resourceBundle, UErrorCode* status)
{
    if (status==NULL || U_FAILURE(*status)) {
        return NULL;
    }
    if (!resourceBundle) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    } else {
      return resourceBundle->fData->fName;
    }
}

U_CAPI const char* U_EXPORT2
ures_getLocale(const UResourceBundle* resourceBundle,
               UErrorCode* status)
{
  return ures_getLocaleInternal(resourceBundle, status);
}


U_CAPI const char* U_EXPORT2
ures_getLocaleByType(const UResourceBundle* resourceBundle,
                     ULocDataLocaleType type,
                     UErrorCode* status) {
    if (status==NULL || U_FAILURE(*status)) {
        return NULL;
    }
    if (!resourceBundle) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    } else {
        switch(type) {
        case ULOC_ACTUAL_LOCALE:
            return resourceBundle->fData->fName;
        case ULOC_VALID_LOCALE:
            return resourceBundle->fTopLevelData->fName;
        case ULOC_REQUESTED_LOCALE:
        default:
            *status = U_ILLEGAL_ARGUMENT_ERROR;
            return NULL;
        }
    }
}

U_CFUNC const char* ures_getName(const UResourceBundle* resB) {
  if(resB == NULL) {
    return NULL;
  }

  return resB->fData->fName;
}

#ifdef URES_DEBUG
U_CFUNC const char* ures_getPath(const UResourceBundle* resB) {
  if(resB == NULL) {
    return NULL;
  }

  return resB->fData->fPath;
}
#endif

static UResourceBundle*
ures_openWithType(UResourceBundle *r, const char* path, const char* localeID,
                  UResOpenType openType, UErrorCode* status) {
    if(U_FAILURE(*status)) {
        return NULL;
    }

    UResourceDataEntry *entry;
    if(openType != URES_OPEN_DIRECT) {
        /* first "canonicalize" the locale ID */
        char canonLocaleID[ULOC_FULLNAME_CAPACITY];
        uloc_getBaseName(localeID, canonLocaleID, UPRV_LENGTHOF(canonLocaleID), status);
        if(U_FAILURE(*status) || *status == U_STRING_NOT_TERMINATED_WARNING) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
            return NULL;
        }
        entry = entryOpen(path, canonLocaleID, openType, status);
    } else {
        entry = entryOpenDirect(path, localeID, status);
    }
    if(U_FAILURE(*status)) {
        return NULL;
    }
    if(entry == NULL) {
        *status = U_MISSING_RESOURCE_ERROR;
        return NULL;
    }

    UBool isStackObject;
    if(r == NULL) {
        r = (UResourceBundle *)uprv_malloc(sizeof(UResourceBundle));
        if(r == NULL) {
            entryClose(entry);
            *status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        isStackObject = FALSE;
    } else {  // fill-in
        isStackObject = ures_isStackObject(r);
        ures_closeBundle(r, FALSE);
    }
    uprv_memset(r, 0, sizeof(UResourceBundle));
    ures_setIsStackObject(r, isStackObject);

    r->fTopLevelData = r->fData = entry;
    uprv_memcpy(&r->fResData, &entry->fData, sizeof(ResourceData));
    r->fHasFallback = openType != URES_OPEN_DIRECT && !r->fResData.noFallback;
    r->fIsTopLevel = TRUE;
    r->fRes = r->fResData.rootRes;
    r->fSize = res_countArrayItems(&(r->fResData), r->fRes);
    r->fIndex = -1;

    ResourceTracer(r).traceOpen();

    return r;
}

U_CAPI UResourceBundle* U_EXPORT2
ures_open(const char* path, const char* localeID, UErrorCode* status) {
    return ures_openWithType(NULL, path, localeID, URES_OPEN_LOCALE_DEFAULT_ROOT, status);
}

U_CAPI UResourceBundle* U_EXPORT2
ures_openNoDefault(const char* path, const char* localeID, UErrorCode* status) {
    return ures_openWithType(NULL, path, localeID, URES_OPEN_LOCALE_ROOT, status);
}

/**
 *  Opens a resource bundle without "canonicalizing" the locale name. No fallback will be performed
 *  or sought. However, alias substitution will happen!
 */
U_CAPI UResourceBundle*  U_EXPORT2
ures_openDirect(const char* path, const char* localeID, UErrorCode* status) {
    return ures_openWithType(NULL, path, localeID, URES_OPEN_DIRECT, status);
}

/**
 *  Internal API: This function is used to open a resource bundle
 *  proper fallback chaining is executed while initialization.
 *  The result is stored in uresbund_cache for later fallback search.
 *
 * Same as ures_open(), but uses the fill-in parameter and does not allocate a new bundle.
 */
U_INTERNAL void U_EXPORT2
ures_openFillIn(UResourceBundle *r, const char* path,
                const char* localeID, UErrorCode* status) {
    if(U_SUCCESS(*status) && r == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    ures_openWithType(r, path, localeID, URES_OPEN_LOCALE_DEFAULT_ROOT, status);
}

/**
 * Same as ures_openDirect(), but uses the fill-in parameter and does not allocate a new bundle.
 */
U_INTERNAL void U_EXPORT2
ures_openDirectFillIn(UResourceBundle *r, const char* path, const char* localeID, UErrorCode* status) {
    if(U_SUCCESS(*status) && r == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    ures_openWithType(r, path, localeID, URES_OPEN_DIRECT, status);
}

/**
 *  API: Counts members. For arrays and tables, returns number of resources.
 *  For strings, returns 1.
 */
U_CAPI int32_t  U_EXPORT2
ures_countArrayItems(const UResourceBundle* resourceBundle,
                  const char* resourceKey,
                  UErrorCode* status)
{
    UResourceBundle resData;
    ures_initStackObject(&resData);
    if (status==NULL || U_FAILURE(*status)) {
        return 0;
    }
    if(resourceBundle == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    ures_getByKey(resourceBundle, resourceKey, &resData, status);

    if(resData.fResData.data != NULL) {
        int32_t result = res_countArrayItems(&resData.fResData, resData.fRes);
        ures_close(&resData);
        return result;
    } else {
        *status = U_MISSING_RESOURCE_ERROR;
        ures_close(&resData);
        return 0;
    }
}

/**
 * Internal function.
 * Return the version number associated with this ResourceBundle as a string.
 *
 * @param resourceBundle The resource bundle for which the version is checked.
 * @return  A version number string as specified in the resource bundle or its parent.
 *          The caller does not own this string.
 * @see ures_getVersion
 * @internal
 */
U_INTERNAL const char* U_EXPORT2
ures_getVersionNumberInternal(const UResourceBundle *resourceBundle)
{
    if (!resourceBundle) return NULL;

    if(resourceBundle->fVersion == NULL) {

        /* If the version ID has not been built yet, then do so.  Retrieve */
        /* the minor version from the file. */
        UErrorCode status = U_ZERO_ERROR;
        int32_t minor_len = 0;
        int32_t len;

        const UChar* minor_version = ures_getStringByKey(resourceBundle, kVersionTag, &minor_len, &status);

        /* Determine the length of of the final version string.  This is */
        /* the length of the major part + the length of the separator */
        /* (==1) + the length of the minor part (+ 1 for the zero byte at */
        /* the end). */

        len = (minor_len > 0) ? minor_len : 1;

        /* Allocate the string, and build it up. */
        /* + 1 for zero byte */


        ((UResourceBundle *)resourceBundle)->fVersion = (char *)uprv_malloc(1 + len);
        /* Check for null pointer. */
        if (((UResourceBundle *)resourceBundle)->fVersion == NULL) {
            return NULL;
        }

        if(minor_len > 0) {
            u_UCharsToChars(minor_version, resourceBundle->fVersion , minor_len);
            resourceBundle->fVersion[len] =  '\0';
        }
        else {
            uprv_strcpy(resourceBundle->fVersion, kDefaultMinorVersion);
        }
    }

    return resourceBundle->fVersion;
}

U_CAPI const char*  U_EXPORT2
ures_getVersionNumber(const UResourceBundle*   resourceBundle)
{
    return ures_getVersionNumberInternal(resourceBundle);
}

U_CAPI void U_EXPORT2 ures_getVersion(const UResourceBundle* resB, UVersionInfo versionInfo) {
    if (!resB) return;

    u_versionFromString(versionInfo, ures_getVersionNumberInternal(resB));
}

/** Tree support functions *******************************/
#define INDEX_LOCALE_NAME "res_index"
#define INDEX_TAG         "InstalledLocales"
#define DEFAULT_TAG       "default"

#if defined(URES_TREE_DEBUG)
#include <stdio.h>
#endif

typedef struct ULocalesContext {
    UResourceBundle installed;
    UResourceBundle curr;
} ULocalesContext;

static void U_CALLCONV
ures_loc_closeLocales(UEnumeration *enumerator) {
    ULocalesContext *ctx = (ULocalesContext *)enumerator->context;
    ures_close(&ctx->curr);
    ures_close(&ctx->installed);
    uprv_free(ctx);
    uprv_free(enumerator);
}

static int32_t U_CALLCONV
ures_loc_countLocales(UEnumeration *en, UErrorCode * /*status*/) {
    ULocalesContext *ctx = (ULocalesContext *)en->context;
    return ures_getSize(&ctx->installed);
}

U_CDECL_BEGIN


static const char * U_CALLCONV
ures_loc_nextLocale(UEnumeration* en,
                    int32_t* resultLength,
                    UErrorCode* status) {
    ULocalesContext *ctx = (ULocalesContext *)en->context;
    UResourceBundle *res = &(ctx->installed);
    UResourceBundle *k = NULL;
    const char *result = NULL;
    int32_t len = 0;
    if(ures_hasNext(res) && (k = ures_getNextResource(res, &ctx->curr, status)) != 0) {
        result = ures_getKey(k);
        len = (int32_t)uprv_strlen(result);
    }
    if (resultLength) {
        *resultLength = len;
    }
    return result;
}

static void U_CALLCONV
ures_loc_resetLocales(UEnumeration* en,
                      UErrorCode* /*status*/) {
    UResourceBundle *res = &((ULocalesContext *)en->context)->installed;
    ures_resetIterator(res);
}

U_CDECL_END

static const UEnumeration gLocalesEnum = {
    NULL,
        NULL,
        ures_loc_closeLocales,
        ures_loc_countLocales,
        uenum_unextDefault,
        ures_loc_nextLocale,
        ures_loc_resetLocales
};


U_CAPI UEnumeration* U_EXPORT2
ures_openAvailableLocales(const char *path, UErrorCode *status)
{
    UResourceBundle *idx = NULL;
    UEnumeration *en = NULL;
    ULocalesContext *myContext = NULL;

    if(U_FAILURE(*status)) {
        return NULL;
    }
    myContext = static_cast<ULocalesContext *>(uprv_malloc(sizeof(ULocalesContext)));
    en =  (UEnumeration *)uprv_malloc(sizeof(UEnumeration));
    if(!en || !myContext) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        uprv_free(en);
        uprv_free(myContext);
        return NULL;
    }
    uprv_memcpy(en, &gLocalesEnum, sizeof(UEnumeration));

    ures_initStackObject(&myContext->installed);
    ures_initStackObject(&myContext->curr);
    idx = ures_openDirect(path, INDEX_LOCALE_NAME, status);
    ures_getByKey(idx, INDEX_TAG, &myContext->installed, status);
    if(U_SUCCESS(*status)) {
#if defined(URES_TREE_DEBUG)
        fprintf(stderr, "Got %s::%s::[%s] : %s\n",
            path, INDEX_LOCALE_NAME, INDEX_TAG, ures_getKey(&myContext->installed));
#endif
        en->context = myContext;
    } else {
#if defined(URES_TREE_DEBUG)
        fprintf(stderr, "%s open failed - %s\n", path, u_errorName(*status));
#endif
        ures_close(&myContext->installed);
        uprv_free(myContext);
        uprv_free(en);
        en = NULL;
    }

    ures_close(idx);

    return en;
}

static UBool isLocaleInList(UEnumeration *locEnum, const char *locToSearch, UErrorCode *status) {
    const char *loc;
    while ((loc = uenum_next(locEnum, NULL, status)) != NULL) {
        if (uprv_strcmp(loc, locToSearch) == 0) {
            return TRUE;
        }
    }
    return FALSE;
}

U_CAPI int32_t U_EXPORT2
ures_getFunctionalEquivalent(char *result, int32_t resultCapacity,
                             const char *path, const char *resName, const char *keyword, const char *locid,
                             UBool *isAvailable, UBool omitDefault, UErrorCode *status)
{
    char kwVal[1024] = ""; /* value of keyword 'keyword' */
    char defVal[1024] = ""; /* default value for given locale */
    char defLoc[1024] = ""; /* default value for given locale */
    char base[1024] = ""; /* base locale */
    char found[1024];
    char parent[1024];
    char full[1024] = "";
    UResourceBundle bund1, bund2;
    UResourceBundle *res = NULL;
    UErrorCode subStatus = U_ZERO_ERROR;
    int32_t length = 0;
    if(U_FAILURE(*status)) return 0;
    uloc_getKeywordValue(locid, keyword, kwVal, 1024-1,&subStatus);
    if(!uprv_strcmp(kwVal, DEFAULT_TAG)) {
        kwVal[0]=0;
    }
    uloc_getBaseName(locid, base, 1024-1,&subStatus);
#if defined(URES_TREE_DEBUG)
    fprintf(stderr, "getFunctionalEquivalent: \"%s\" [%s=%s] in %s - %s\n",
            locid, keyword, kwVal, base, u_errorName(subStatus));
#endif
    ures_initStackObject(&bund1);
    ures_initStackObject(&bund2);


    uprv_strcpy(parent, base);
    uprv_strcpy(found, base);

    if(isAvailable) {
        UEnumeration *locEnum = ures_openAvailableLocales(path, &subStatus);
        *isAvailable = TRUE;
        if (U_SUCCESS(subStatus)) {
            *isAvailable = isLocaleInList(locEnum, parent, &subStatus);
        }
        uenum_close(locEnum);
    }

    if(U_FAILURE(subStatus)) {
        *status = subStatus;
        return 0;
    }

    do {
        subStatus = U_ZERO_ERROR;
        res = ures_open(path, parent, &subStatus);
        if(((subStatus == U_USING_FALLBACK_WARNING) ||
            (subStatus == U_USING_DEFAULT_WARNING)) && isAvailable)
        {
            *isAvailable = FALSE;
        }
        isAvailable = NULL; /* only want to set this the first time around */

#if defined(URES_TREE_DEBUG)
        fprintf(stderr, "%s;%s -> %s [%s]\n", path?path:"ICUDATA", parent, u_errorName(subStatus), ures_getLocale(res, &subStatus));
#endif
        if(U_FAILURE(subStatus)) {
            *status = subStatus;
        } else if(subStatus == U_ZERO_ERROR) {
            ures_getByKey(res,resName,&bund1, &subStatus);
            if(subStatus == U_ZERO_ERROR) {
                const UChar *defUstr;
                int32_t defLen;
                /* look for default item */
#if defined(URES_TREE_DEBUG)
                fprintf(stderr, "%s;%s : loaded default -> %s\n",
                    path?path:"ICUDATA", parent, u_errorName(subStatus));
#endif
                defUstr = ures_getStringByKey(&bund1, DEFAULT_TAG, &defLen, &subStatus);
                if(U_SUCCESS(subStatus) && defLen) {
                    u_UCharsToChars(defUstr, defVal, u_strlen(defUstr));
#if defined(URES_TREE_DEBUG)
                    fprintf(stderr, "%s;%s -> default %s=%s,  %s\n",
                        path?path:"ICUDATA", parent, keyword, defVal, u_errorName(subStatus));
#endif
                    uprv_strcpy(defLoc, parent);
                    if(kwVal[0]==0) {
                        uprv_strcpy(kwVal, defVal);
#if defined(URES_TREE_DEBUG)
                        fprintf(stderr, "%s;%s -> kwVal =  %s\n",
                            path?path:"ICUDATA", parent, keyword, kwVal);
#endif
                    }
                }
            }
        }

        subStatus = U_ZERO_ERROR;

        if (res != NULL) {
            uprv_strcpy(found, ures_getLocaleByType(res, ULOC_VALID_LOCALE, &subStatus));
        }

        uloc_getParent(found,parent,sizeof(parent),&subStatus);
        ures_close(res);
    } while(!defVal[0] && *found && uprv_strcmp(found, "root") != 0 && U_SUCCESS(*status));

    /* Now, see if we can find the kwVal collator.. start the search over.. */
    uprv_strcpy(parent, base);
    uprv_strcpy(found, base);

    do {
        subStatus = U_ZERO_ERROR;
        res = ures_open(path, parent, &subStatus);
        if((subStatus == U_USING_FALLBACK_WARNING) && isAvailable) {
            *isAvailable = FALSE;
        }
        isAvailable = NULL; /* only want to set this the first time around */

#if defined(URES_TREE_DEBUG)
        fprintf(stderr, "%s;%s -> %s (looking for %s)\n",
            path?path:"ICUDATA", parent, u_errorName(subStatus), kwVal);
#endif
        if(U_FAILURE(subStatus)) {
            *status = subStatus;
        } else if(subStatus == U_ZERO_ERROR) {
            ures_getByKey(res,resName,&bund1, &subStatus);
#if defined(URES_TREE_DEBUG)
/**/ fprintf(stderr,"@%d [%s] %s\n", __LINE__, resName, u_errorName(subStatus));
#endif
            if(subStatus == U_ZERO_ERROR) {
                ures_getByKey(&bund1, kwVal, &bund2, &subStatus);
#if defined(URES_TREE_DEBUG)
/**/ fprintf(stderr,"@%d [%s] %s\n", __LINE__, kwVal, u_errorName(subStatus));
#endif
                if(subStatus == U_ZERO_ERROR) {
#if defined(URES_TREE_DEBUG)
                    fprintf(stderr, "%s;%s -> full0 %s=%s,  %s\n",
                        path?path:"ICUDATA", parent, keyword, kwVal, u_errorName(subStatus));
#endif
                    uprv_strcpy(full, parent);
                    if(*full == 0) {
                        uprv_strcpy(full, "root");
                    }
                        /* now, recalculate default kw if need be */
                        if(uprv_strlen(defLoc) > uprv_strlen(full)) {
                          const UChar *defUstr;
                          int32_t defLen;
                          /* look for default item */
#if defined(URES_TREE_DEBUG)
                            fprintf(stderr, "%s;%s -> recalculating Default0\n",
                                    path?path:"ICUDATA", full);
#endif
                          defUstr = ures_getStringByKey(&bund1, DEFAULT_TAG, &defLen, &subStatus);
                          if(U_SUCCESS(subStatus) && defLen) {
                            u_UCharsToChars(defUstr, defVal, u_strlen(defUstr));
#if defined(URES_TREE_DEBUG)
                            fprintf(stderr, "%s;%s -> default0 %s=%s,  %s\n",
                                    path?path:"ICUDATA", full, keyword, defVal, u_errorName(subStatus));
#endif
                            uprv_strcpy(defLoc, full);
                          }
                        } /* end of recalculate default KW */
#if defined(URES_TREE_DEBUG)
                        else {
                          fprintf(stderr, "No trim0,  %s <= %s\n", defLoc, full);
                        }
#endif
                } else {
#if defined(URES_TREE_DEBUG)
                    fprintf(stderr, "err=%s in %s looking for %s\n",
                        u_errorName(subStatus), parent, kwVal);
#endif
                }
            }
        }

        subStatus = U_ZERO_ERROR;

        uprv_strcpy(found, parent);
        uloc_getParent(found,parent,1023,&subStatus);
        ures_close(res);
    } while(!full[0] && *found && U_SUCCESS(*status));

    if((full[0]==0) && uprv_strcmp(kwVal, defVal)) {
#if defined(URES_TREE_DEBUG)
        fprintf(stderr, "Failed to locate kw %s - try default %s\n", kwVal, defVal);
#endif
        uprv_strcpy(kwVal, defVal);
        uprv_strcpy(parent, base);
        uprv_strcpy(found, base);

        do { /* search for 'default' named item */
            subStatus = U_ZERO_ERROR;
            res = ures_open(path, parent, &subStatus);
            if((subStatus == U_USING_FALLBACK_WARNING) && isAvailable) {
                *isAvailable = FALSE;
            }
            isAvailable = NULL; /* only want to set this the first time around */

#if defined(URES_TREE_DEBUG)
            fprintf(stderr, "%s;%s -> %s (looking for default %s)\n",
                path?path:"ICUDATA", parent, u_errorName(subStatus), kwVal);
#endif
            if(U_FAILURE(subStatus)) {
                *status = subStatus;
            } else if(subStatus == U_ZERO_ERROR) {
                ures_getByKey(res,resName,&bund1, &subStatus);
                if(subStatus == U_ZERO_ERROR) {
                    ures_getByKey(&bund1, kwVal, &bund2, &subStatus);
                    if(subStatus == U_ZERO_ERROR) {
#if defined(URES_TREE_DEBUG)
                        fprintf(stderr, "%s;%s -> full1 %s=%s,  %s\n", path?path:"ICUDATA",
                            parent, keyword, kwVal, u_errorName(subStatus));
#endif
                        uprv_strcpy(full, parent);
                        if(*full == 0) {
                            uprv_strcpy(full, "root");
                        }

                        /* now, recalculate default kw if need be */
                        if(uprv_strlen(defLoc) > uprv_strlen(full)) {
                          const UChar *defUstr;
                          int32_t defLen;
                          /* look for default item */
#if defined(URES_TREE_DEBUG)
                            fprintf(stderr, "%s;%s -> recalculating Default1\n",
                                    path?path:"ICUDATA", full);
#endif
                          defUstr = ures_getStringByKey(&bund1, DEFAULT_TAG, &defLen, &subStatus);
                          if(U_SUCCESS(subStatus) && defLen) {
                            u_UCharsToChars(defUstr, defVal, u_strlen(defUstr));
#if defined(URES_TREE_DEBUG)
                            fprintf(stderr, "%s;%s -> default %s=%s,  %s\n",
                                    path?path:"ICUDATA", full, keyword, defVal, u_errorName(subStatus));
#endif
                            uprv_strcpy(defLoc, full);
                          }
                        } /* end of recalculate default KW */
#if defined(URES_TREE_DEBUG)
                        else {
                          fprintf(stderr, "No trim1,  %s <= %s\n", defLoc, full);
                        }
#endif
                    }
                }
            }
            subStatus = U_ZERO_ERROR;

            uprv_strcpy(found, parent);
            uloc_getParent(found,parent,1023,&subStatus);
            ures_close(res);
        } while(!full[0] && *found && U_SUCCESS(*status));
    }

    if(U_SUCCESS(*status)) {
        if(!full[0]) {
#if defined(URES_TREE_DEBUG)
          fprintf(stderr, "Still could not load keyword %s=%s\n", keyword, kwVal);
#endif
          *status = U_MISSING_RESOURCE_ERROR;
        } else if(omitDefault) {
#if defined(URES_TREE_DEBUG)
          fprintf(stderr,"Trim? full=%s, defLoc=%s, found=%s\n", full, defLoc, found);
#endif
          if(uprv_strlen(defLoc) <= uprv_strlen(full)) {
            /* found the keyword in a *child* of where the default tag was present. */
            if(!uprv_strcmp(kwVal, defVal)) { /* if the requested kw is default, */
              /* and the default is in or in an ancestor of the current locale */
#if defined(URES_TREE_DEBUG)
              fprintf(stderr, "Removing unneeded var %s=%s\n", keyword, kwVal);
#endif
              kwVal[0]=0;
            }
          }
        }
        uprv_strcpy(found, full);
        if(kwVal[0]) {
            uprv_strcat(found, "@");
            uprv_strcat(found, keyword);
            uprv_strcat(found, "=");
            uprv_strcat(found, kwVal);
        } else if(!omitDefault) {
            uprv_strcat(found, "@");
            uprv_strcat(found, keyword);
            uprv_strcat(found, "=");
            uprv_strcat(found, defVal);
        }
    }
    /* we found the default locale - no need to repeat it.*/

    ures_close(&bund1);
    ures_close(&bund2);

    length = (int32_t)uprv_strlen(found);

    if(U_SUCCESS(*status)) {
        int32_t copyLength = uprv_min(length, resultCapacity);
        if(copyLength>0) {
            uprv_strncpy(result, found, copyLength);
        }
        if(length == 0) {
          *status = U_MISSING_RESOURCE_ERROR;
        }
    } else {
        length = 0;
        result[0]=0;
    }
    return u_terminateChars(result, resultCapacity, length, status);
}

U_CAPI UEnumeration* U_EXPORT2
ures_getKeywordValues(const char *path, const char *keyword, UErrorCode *status)
{
#define VALUES_BUF_SIZE 2048
#define VALUES_LIST_SIZE 512

    char       valuesBuf[VALUES_BUF_SIZE];
    int32_t    valuesIndex = 0;
    const char *valuesList[VALUES_LIST_SIZE];
    int32_t    valuesCount = 0;

    const char *locale;
    int32_t     locLen;

    UEnumeration *locs = NULL;

    UResourceBundle    item;
    UResourceBundle    subItem;

    ures_initStackObject(&item);
    ures_initStackObject(&subItem);
    locs = ures_openAvailableLocales(path, status);

    if(U_FAILURE(*status)) {
        ures_close(&item);
        ures_close(&subItem);
        return NULL;
    }

    valuesBuf[0]=0;
    valuesBuf[1]=0;

    while((locale = uenum_next(locs, &locLen, status)) != 0) {
        UResourceBundle   *bund = NULL;
        UResourceBundle   *subPtr = NULL;
        UErrorCode subStatus = U_ZERO_ERROR; /* don't fail if a bundle is unopenable */
        bund = ures_openDirect(path, locale, &subStatus);

#if defined(URES_TREE_DEBUG)
        if(!bund || U_FAILURE(subStatus)) {
            fprintf(stderr, "%s-%s values: Can't open %s locale - skipping. (%s)\n",
                path?path:"<ICUDATA>", keyword, locale, u_errorName(subStatus));
        }
#endif

        ures_getByKey(bund, keyword, &item, &subStatus);

        if(!bund || U_FAILURE(subStatus)) {
#if defined(URES_TREE_DEBUG)
            fprintf(stderr, "%s-%s values: Can't find in %s - skipping. (%s)\n",
                path?path:"<ICUDATA>", keyword, locale, u_errorName(subStatus));
#endif
            ures_close(bund);
            bund = NULL;
            continue;
        }

        while((subPtr = ures_getNextResource(&item,&subItem,&subStatus)) != 0
            && U_SUCCESS(subStatus)) {
            const char *k;
            int32_t i;
            k = ures_getKey(subPtr);

#if defined(URES_TREE_DEBUG)
            /* fprintf(stderr, "%s | %s | %s | %s\n", path?path:"<ICUDATA>", keyword, locale, k); */
#endif
            if(k == NULL || *k == 0 ||
                    uprv_strcmp(k, DEFAULT_TAG) == 0 || uprv_strncmp(k, "private-", 8) == 0) {
                // empty or "default" or unlisted type
                continue;
            }
            for(i=0; i<valuesCount; i++) {
                if(!uprv_strcmp(valuesList[i],k)) {
                    k = NULL; /* found duplicate */
                    break;
                }
            }
            if(k != NULL) {
                int32_t kLen = (int32_t)uprv_strlen(k);
                if((valuesCount >= (VALUES_LIST_SIZE-1)) ||       /* no more space in list .. */
                    ((valuesIndex+kLen+1+1) >= VALUES_BUF_SIZE)) { /* no more space in buffer (string + 2 nulls) */
                    *status = U_ILLEGAL_ARGUMENT_ERROR; /* out of space.. */
                } else {
                    uprv_strcpy(valuesBuf+valuesIndex, k);
                    valuesList[valuesCount++] = valuesBuf+valuesIndex;
                    valuesIndex += kLen;
#if defined(URES_TREE_DEBUG)
                    fprintf(stderr, "%s | %s | %s | [%s]   (UNIQUE)\n",
                        path?path:"<ICUDATA>", keyword, locale, k);
#endif
                    valuesBuf[valuesIndex++] = 0; /* terminate */
                }
            }
        }
        ures_close(bund);
    }
    valuesBuf[valuesIndex++] = 0; /* terminate */

    ures_close(&item);
    ures_close(&subItem);
    uenum_close(locs);
#if defined(URES_TREE_DEBUG)
    fprintf(stderr, "%s:  size %d, #%d\n", u_errorName(*status),
        valuesIndex, valuesCount);
#endif
    return uloc_openKeywordList(valuesBuf, valuesIndex, status);
}
#if 0
/* This code isn't needed, and given the documentation warnings the implementation is suspect */
U_INTERNAL UBool U_EXPORT2
ures_equal(const UResourceBundle* res1, const UResourceBundle* res2){
    if(res1==NULL || res2==NULL){
        return res1==res2; /* pointer comparision */
    }
    if(res1->fKey==NULL||  res2->fKey==NULL){
        return (res1->fKey==res2->fKey);
    }else{
        if(uprv_strcmp(res1->fKey, res2->fKey)!=0){
            return FALSE;
        }
    }
    if(uprv_strcmp(res1->fData->fName, res2->fData->fName)!=0){
        return FALSE;
    }
    if(res1->fData->fPath == NULL||  res2->fData->fPath==NULL){
        return (res1->fData->fPath == res2->fData->fPath);
    }else{
        if(uprv_strcmp(res1->fData->fPath, res2->fData->fPath)!=0){
            return FALSE;
        }
    }
    if(uprv_strcmp(res1->fData->fParent->fName, res2->fData->fParent->fName)!=0){
        return FALSE;
    }
    if(uprv_strcmp(res1->fData->fParent->fPath, res2->fData->fParent->fPath)!=0){
        return FALSE;
    }
    if(uprv_strncmp(res1->fResPath, res2->fResPath, res1->fResPathLen)!=0){
        return FALSE;
    }
    if(res1->fRes != res2->fRes){
        return FALSE;
    }
    return TRUE;
}
U_INTERNAL UResourceBundle* U_EXPORT2
ures_clone(const UResourceBundle* res, UErrorCode* status){
    UResourceBundle* bundle = NULL;
    UResourceBundle* ret = NULL;
    if(U_FAILURE(*status) || res == NULL){
        return NULL;
    }
    bundle = ures_open(res->fData->fPath, res->fData->fName, status);
    if(res->fResPath!=NULL){
        ret = ures_findSubResource(bundle, res->fResPath, NULL, status);
        ures_close(bundle);
    }else{
        ret = bundle;
    }
    return ret;
}
U_INTERNAL const UResourceBundle* U_EXPORT2
ures_getParentBundle(const UResourceBundle* res){
    if(res==NULL){
        return NULL;
    }
    return res->fParentRes;
}
#endif

U_INTERNAL void U_EXPORT2
ures_getVersionByKey(const UResourceBundle* res, const char *key, UVersionInfo ver, UErrorCode *status) {
  const UChar *str;
  int32_t len;
  str = ures_getStringByKey(res, key, &len, status);
  if(U_SUCCESS(*status)) {
    u_versionFromUString(ver, str);
  }
}

/* eof */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1999-2016, International Business Machines Corporation
*               and others. All Rights Reserved.
*******************************************************************************
*   file name:  uresdata.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999dec08
*   created by: Markus W. Scherer
* Modification History:
*
*   Date        Name        Description
*   06/20/2000  helena      OS/400 port changes; mostly typecast.
*   06/24/02    weiv        Added support for resource sharing
*/

















/*
 * Resource access helpers
 */

/* get a const char* pointer to the key with the keyOffset byte offset from pRoot */
#define RES_GET_KEY16(pResData, keyOffset) \
    ((keyOffset)<(pResData)->localKeyLimit ? \
        (const char *)(pResData)->pRoot+(keyOffset) : \
        (pResData)->poolBundleKeys+(keyOffset)-(pResData)->localKeyLimit)

#define RES_GET_KEY32(pResData, keyOffset) \
    ((keyOffset)>=0 ? \
        (const char *)(pResData)->pRoot+(keyOffset) : \
        (pResData)->poolBundleKeys+((keyOffset)&0x7fffffff))

#define URESDATA_ITEM_NOT_FOUND -1

/* empty resources, returned when the resource offset is 0 */
static const uint16_t gEmpty16=0;

static const struct {
    int32_t length;
    int32_t res;
} gEmpty32={ 0, 0 };

static const struct {
    int32_t length;
    UChar nul;
    UChar pad;
} uresdata_gEmptyString={ 0, 0, 0 };

/*
 * All the type-access functions assume that
 * the resource is of the expected type.
 */

static int32_t
_res_findTableItem(const ResourceData *pResData, const uint16_t *keyOffsets, int32_t length,
                   const char *key, const char **realKey) {
    const char *tableKey;
    int32_t mid, start, limit;
    int result;

    /* do a binary search for the key */
    start=0;
    limit=length;
    while(start<limit) {
        mid = (start + limit) / 2;
        tableKey = RES_GET_KEY16(pResData, keyOffsets[mid]);
        if (pResData->useNativeStrcmp) {
            result = uprv_strcmp(key, tableKey);
        } else {
            result = uprv_compareInvCharsAsAscii(key, tableKey);
        }
        if (result < 0) {
            limit = mid;
        } else if (result > 0) {
            start = mid + 1;
        } else {
            /* We found it! */
            *realKey=tableKey;
            return mid;
        }
    }
    return URESDATA_ITEM_NOT_FOUND;  /* not found or table is empty. */
}

static int32_t
_res_findTable32Item(const ResourceData *pResData, const int32_t *keyOffsets, int32_t length,
                     const char *key, const char **realKey) {
    const char *tableKey;
    int32_t mid, start, limit;
    int result;

    /* do a binary search for the key */
    start=0;
    limit=length;
    while(start<limit) {
        mid = (start + limit) / 2;
        tableKey = RES_GET_KEY32(pResData, keyOffsets[mid]);
        if (pResData->useNativeStrcmp) {
            result = uprv_strcmp(key, tableKey);
        } else {
            result = uprv_compareInvCharsAsAscii(key, tableKey);
        }
        if (result < 0) {
            limit = mid;
        } else if (result > 0) {
            start = mid + 1;
        } else {
            /* We found it! */
            *realKey=tableKey;
            return mid;
        }
    }
    return URESDATA_ITEM_NOT_FOUND;  /* not found or table is empty. */
}

/* helper for res_load() ---------------------------------------------------- */

static UBool U_CALLCONV
uresdata_isAcceptable(void *context,
             const char * /*type*/, const char * /*name*/,
             const UDataInfo *pInfo) {
    uprv_memcpy(context, pInfo->formatVersion, 4);
    return (UBool)(
        pInfo->size>=20 &&
        pInfo->isBigEndian==U_IS_BIG_ENDIAN &&
        pInfo->charsetFamily==U_CHARSET_FAMILY &&
        pInfo->sizeofUChar==U_SIZEOF_UCHAR &&
        pInfo->dataFormat[0]==0x52 &&   /* dataFormat="ResB" */
        pInfo->dataFormat[1]==0x65 &&
        pInfo->dataFormat[2]==0x73 &&
        pInfo->dataFormat[3]==0x42 &&
        (1<=pInfo->formatVersion[0] && pInfo->formatVersion[0]<=3));
}

/* semi-public functions ---------------------------------------------------- */

static void
res_init(ResourceData *pResData,
         UVersionInfo formatVersion, const void *inBytes, int32_t length,
         UErrorCode *errorCode) {
    UResType rootType;

    /* get the root resource */
    pResData->pRoot=(const int32_t *)inBytes;
    pResData->rootRes=(Resource)*pResData->pRoot;
    pResData->p16BitUnits=&gEmpty16;

    /* formatVersion 1.1 must have a root item and at least 5 indexes */
    if(length>=0 && (length/4)<((formatVersion[0]==1 && formatVersion[1]==0) ? 1 : 1+5)) {
        *errorCode=U_INVALID_FORMAT_ERROR;
        res_unload(pResData);
        return;
    }

    /* currently, we accept only resources that have a Table as their roots */
    rootType=(UResType)RES_GET_TYPE(pResData->rootRes);
    if(!URES_IS_TABLE(rootType)) {
        *errorCode=U_INVALID_FORMAT_ERROR;
        res_unload(pResData);
        return;
    }

    if(formatVersion[0]==1 && formatVersion[1]==0) {
        pResData->localKeyLimit=0x10000;  /* greater than any 16-bit key string offset */
    } else {
        /* bundles with formatVersion 1.1 and later contain an indexes[] array */
        const int32_t *indexes=pResData->pRoot+1;
        int32_t indexLength=indexes[URES_INDEX_LENGTH]&0xff;
        if(indexLength<=URES_INDEX_MAX_TABLE_LENGTH) {
            *errorCode=U_INVALID_FORMAT_ERROR;
            res_unload(pResData);
            return;
        }
        if( length>=0 &&
            (length<((1+indexLength)<<2) ||
             length<(indexes[URES_INDEX_BUNDLE_TOP]<<2))
        ) {
            *errorCode=U_INVALID_FORMAT_ERROR;
            res_unload(pResData);
            return;
        }
        if(indexes[URES_INDEX_KEYS_TOP]>(1+indexLength)) {
            pResData->localKeyLimit=indexes[URES_INDEX_KEYS_TOP]<<2;
        }
        if(formatVersion[0]>=3) {
            // In formatVersion 1, the indexLength took up this whole int.
            // In version 2, bits 31..8 were reserved and always 0.
            // In version 3, they contain bits 23..0 of the poolStringIndexLimit.
            // Bits 27..24 are in indexes[URES_INDEX_ATTRIBUTES] bits 15..12.
            pResData->poolStringIndexLimit=(int32_t)((uint32_t)indexes[URES_INDEX_LENGTH]>>8);
        }
        if(indexLength>URES_INDEX_ATTRIBUTES) {
            int32_t att=indexes[URES_INDEX_ATTRIBUTES];
            pResData->noFallback=(UBool)(att&URES_ATT_NO_FALLBACK);
            pResData->isPoolBundle=(UBool)((att&URES_ATT_IS_POOL_BUNDLE)!=0);
            pResData->usesPoolBundle=(UBool)((att&URES_ATT_USES_POOL_BUNDLE)!=0);
            pResData->poolStringIndexLimit|=(att&0xf000)<<12;  // bits 15..12 -> 27..24
            pResData->poolStringIndex16Limit=(int32_t)((uint32_t)att>>16);
        }
        if((pResData->isPoolBundle || pResData->usesPoolBundle) && indexLength<=URES_INDEX_POOL_CHECKSUM) {
            *errorCode=U_INVALID_FORMAT_ERROR;
            res_unload(pResData);
            return;
        }
        if( indexLength>URES_INDEX_16BIT_TOP &&
            indexes[URES_INDEX_16BIT_TOP]>indexes[URES_INDEX_KEYS_TOP]
        ) {
            pResData->p16BitUnits=(const uint16_t *)(pResData->pRoot+indexes[URES_INDEX_KEYS_TOP]);
        }
    }

    if(formatVersion[0]==1 || U_CHARSET_FAMILY==U_ASCII_FAMILY) {
        /*
         * formatVersion 1: compare key strings in native-charset order
         * formatVersion 2 and up: compare key strings in ASCII order
         */
        pResData->useNativeStrcmp=TRUE;
    }
}

U_CAPI void U_EXPORT2
res_read(ResourceData *pResData,
         const UDataInfo *pInfo, const void *inBytes, int32_t length,
         UErrorCode *errorCode) {
    UVersionInfo formatVersion;

    uprv_memset(pResData, 0, sizeof(ResourceData));
    if(U_FAILURE(*errorCode)) {
        return;
    }
    if(!uresdata_isAcceptable(formatVersion, NULL, NULL, pInfo)) {
        *errorCode=U_INVALID_FORMAT_ERROR;
        return;
    }
    res_init(pResData, formatVersion, inBytes, length, errorCode);
}

U_CFUNC void
res_load(ResourceData *pResData,
         const char *path, const char *name, UErrorCode *errorCode) {
    UVersionInfo formatVersion;

    uprv_memset(pResData, 0, sizeof(ResourceData));

    /* load the ResourceBundle file */
    pResData->data=udata_openChoice(path, "res", name, uresdata_isAcceptable, formatVersion, errorCode);
    if(U_FAILURE(*errorCode)) {
        return;
    }

    /* get its memory and initialize *pResData */
    res_init(pResData, formatVersion, udata_getMemory(pResData->data), -1, errorCode);
}

U_CFUNC void
res_unload(ResourceData *pResData) {
    if(pResData->data!=NULL) {
        udata_close(pResData->data);
        pResData->data=NULL;
    }
}

static const int8_t gPublicTypes[URES_LIMIT] = {
    URES_STRING,
    URES_BINARY,
    URES_TABLE,
    URES_ALIAS,

    URES_TABLE,     /* URES_TABLE32 */
    URES_TABLE,     /* URES_TABLE16 */
    URES_STRING,    /* URES_STRING_V2 */
    URES_INT,

    URES_ARRAY,
    URES_ARRAY,     /* URES_ARRAY16 */
    URES_NONE,
    URES_NONE,

    URES_NONE,
    URES_NONE,
    URES_INT_VECTOR,
    URES_NONE
};

U_CAPI UResType U_EXPORT2
res_getPublicType(Resource res) {
    return (UResType)gPublicTypes[RES_GET_TYPE(res)];
}

U_CAPI const UChar * U_EXPORT2
res_getStringNoTrace(const ResourceData *pResData, Resource res, int32_t *pLength) {
    const UChar *p;
    uint32_t offset=RES_GET_OFFSET(res);
    int32_t length;
    if(RES_GET_TYPE(res)==URES_STRING_V2) {
        int32_t first;
        if((int32_t)offset<pResData->poolStringIndexLimit) {
            p=(const UChar *)pResData->poolBundleStrings+offset;
        } else {
            p=(const UChar *)pResData->p16BitUnits+(offset-pResData->poolStringIndexLimit);
        }
        first=*p;
        if(!U16_IS_TRAIL(first)) {
            length=u_strlen(p);
        } else if(first<0xdfef) {
            length=first&0x3ff;
            ++p;
        } else if(first<0xdfff) {
            length=((first-0xdfef)<<16)|p[1];
            p+=2;
        } else {
            length=((int32_t)p[1]<<16)|p[2];
            p+=3;
        }
    } else if(res==offset) /* RES_GET_TYPE(res)==URES_STRING */ {
        const int32_t *p32= res==0 ? &uresdata_gEmptyString.length : pResData->pRoot+res;
        length=*p32++;
        p=(const UChar *)p32;
    } else {
        p=NULL;
        length=0;
    }
    if(pLength) {
        *pLength=length;
    }
    return p;
}

namespace {

/**
 * CLDR string value (three empty-set symbols)=={2205, 2205, 2205}
 * prevents fallback to the parent bundle.
 * TODO: combine with other code that handles this marker, use EMPTY_SET constant.
 * TODO: maybe move to uresbund.cpp?
 */
UBool isNoInheritanceMarker(const ResourceData *pResData, Resource res) {
    uint32_t offset=RES_GET_OFFSET(res);
    if (offset == 0) {
        // empty string
    } else if (res == offset) {
        const int32_t *p32=pResData->pRoot+res;
        int32_t length=*p32;
        const UChar *p=(const UChar *)p32;
        return length == 3 && p[2] == 0x2205 && p[3] == 0x2205 && p[4] == 0x2205;
    } else if (RES_GET_TYPE(res) == URES_STRING_V2) {
        const UChar *p;
        if((int32_t)offset<pResData->poolStringIndexLimit) {
            p=(const UChar *)pResData->poolBundleStrings+offset;
        } else {
            p=(const UChar *)pResData->p16BitUnits+(offset-pResData->poolStringIndexLimit);
        }
        int32_t first=*p;
        if (first == 0x2205) {  // implicit length
            return p[1] == 0x2205 && p[2] == 0x2205 && p[3] == 0;
        } else if (first == 0xdc03) {  // explicit length 3 (should not occur)
            return p[1] == 0x2205 && p[2] == 0x2205 && p[3] == 0x2205;
        } else {
            // Assume that the string has not been stored with more length units than necessary.
            return FALSE;
        }
    }
    return FALSE;
}

int32_t getStringArray(const ResourceData *pResData, const icu::ResourceArray &array,
                       icu::UnicodeString *dest, int32_t capacity,
                       UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    if(dest == NULL ? capacity != 0 : capacity < 0) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    int32_t length = array.getSize();
    if(length == 0) {
        return 0;
    }
    if(length > capacity) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return length;
    }
    for(int32_t i = 0; i < length; ++i) {
        int32_t sLength;
        // No tracing: handled by the caller
        const UChar *s = res_getStringNoTrace(pResData, array.internalGetResource(pResData, i), &sLength);
        if(s == NULL) {
            errorCode = U_RESOURCE_TYPE_MISMATCH;
            return 0;
        }
        dest[i].setTo(TRUE, s, sLength);
    }
    return length;
}

}  // namespace

U_CAPI const UChar * U_EXPORT2
res_getAlias(const ResourceData *pResData, Resource res, int32_t *pLength) {
    const UChar *p;
    uint32_t offset=RES_GET_OFFSET(res);
    int32_t length;
    if(RES_GET_TYPE(res)==URES_ALIAS) {
        const int32_t *p32= offset==0 ? &uresdata_gEmptyString.length : pResData->pRoot+offset;
        length=*p32++;
        p=(const UChar *)p32;
    } else {
        p=NULL;
        length=0;
    }
    if(pLength) {
        *pLength=length;
    }
    return p;
}

U_CAPI const uint8_t * U_EXPORT2
res_getBinaryNoTrace(const ResourceData *pResData, Resource res, int32_t *pLength) {
    const uint8_t *p;
    uint32_t offset=RES_GET_OFFSET(res);
    int32_t length;
    if(RES_GET_TYPE(res)==URES_BINARY) {
        const int32_t *p32= offset==0 ? (const int32_t*)&gEmpty32 : pResData->pRoot+offset;
        length=*p32++;
        p=(const uint8_t *)p32;
    } else {
        p=NULL;
        length=0;
    }
    if(pLength) {
        *pLength=length;
    }
    return p;
}


U_CAPI const int32_t * U_EXPORT2
res_getIntVectorNoTrace(const ResourceData *pResData, Resource res, int32_t *pLength) {
    const int32_t *p;
    uint32_t offset=RES_GET_OFFSET(res);
    int32_t length;
    if(RES_GET_TYPE(res)==URES_INT_VECTOR) {
        p= offset==0 ? (const int32_t *)&gEmpty32 : pResData->pRoot+offset;
        length=*p++;
    } else {
        p=NULL;
        length=0;
    }
    if(pLength) {
        *pLength=length;
    }
    return p;
}

U_CAPI int32_t U_EXPORT2
res_countArrayItems(const ResourceData *pResData, Resource res) {
    uint32_t offset=RES_GET_OFFSET(res);
    switch(RES_GET_TYPE(res)) {
    case URES_STRING:
    case URES_STRING_V2:
    case URES_BINARY:
    case URES_ALIAS:
    case URES_INT:
    case URES_INT_VECTOR:
        return 1;
    case URES_ARRAY:
    case URES_TABLE32:
        return offset==0 ? 0 : *(pResData->pRoot+offset);
    case URES_TABLE:
        return offset==0 ? 0 : *((const uint16_t *)(pResData->pRoot+offset));
    case URES_ARRAY16:
    case URES_TABLE16:
        return pResData->p16BitUnits[offset];
    default:
        return 0;
    }
}

U_NAMESPACE_BEGIN

ResourceDataValue::~ResourceDataValue() {}

UResType ResourceDataValue::getType() const {
    return res_getPublicType(res);
}

const UChar *ResourceDataValue::getString(int32_t &length, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    const UChar *s = res_getString(fTraceInfo, &getData(), res, &length);
    if(s == NULL) {
        errorCode = U_RESOURCE_TYPE_MISMATCH;
    }
    return s;
}

const UChar *ResourceDataValue::getAliasString(int32_t &length, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    const UChar *s = res_getAlias(&getData(), res, &length);
    if(s == NULL) {
        errorCode = U_RESOURCE_TYPE_MISMATCH;
    }
    return s;
}

int32_t ResourceDataValue::getInt(UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    if(RES_GET_TYPE(res) != URES_INT) {
        errorCode = U_RESOURCE_TYPE_MISMATCH;
    }
    return res_getInt(fTraceInfo, res);
}

uint32_t ResourceDataValue::getUInt(UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    if(RES_GET_TYPE(res) != URES_INT) {
        errorCode = U_RESOURCE_TYPE_MISMATCH;
    }
    return res_getUInt(fTraceInfo, res);
}

const int32_t *ResourceDataValue::getIntVector(int32_t &length, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    const int32_t *iv = res_getIntVector(fTraceInfo, &getData(), res, &length);
    if(iv == NULL) {
        errorCode = U_RESOURCE_TYPE_MISMATCH;
    }
    return iv;
}

const uint8_t *ResourceDataValue::getBinary(int32_t &length, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    const uint8_t *b = res_getBinary(fTraceInfo, &getData(), res, &length);
    if(b == NULL) {
        errorCode = U_RESOURCE_TYPE_MISMATCH;
    }
    return b;
}

ResourceArray ResourceDataValue::getArray(UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return ResourceArray();
    }
    const uint16_t *items16 = NULL;
    const Resource *items32 = NULL;
    uint32_t offset=RES_GET_OFFSET(res);
    int32_t length = 0;
    switch(RES_GET_TYPE(res)) {
    case URES_ARRAY:
        if (offset!=0) {  // empty if offset==0
            items32 = (const Resource *)getData().pRoot+offset;
            length = *items32++;
        }
        break;
    case URES_ARRAY16:
        items16 = getData().p16BitUnits+offset;
        length = *items16++;
        break;
    default:
        errorCode = U_RESOURCE_TYPE_MISMATCH;
        return ResourceArray();
    }
    return ResourceArray(items16, items32, length, fTraceInfo);
}

ResourceTable ResourceDataValue::getTable(UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return ResourceTable();
    }
    const uint16_t *keys16 = NULL;
    const int32_t *keys32 = NULL;
    const uint16_t *items16 = NULL;
    const Resource *items32 = NULL;
    uint32_t offset = RES_GET_OFFSET(res);
    int32_t length = 0;
    switch(RES_GET_TYPE(res)) {
    case URES_TABLE:
        if (offset != 0) {  // empty if offset==0
            keys16 = (const uint16_t *)(getData().pRoot+offset);
            length = *keys16++;
            items32 = (const Resource *)(keys16+length+(~length&1));
        }
        break;
    case URES_TABLE16:
        keys16 = getData().p16BitUnits+offset;
        length = *keys16++;
        items16 = keys16 + length;
        break;
    case URES_TABLE32:
        if (offset != 0) {  // empty if offset==0
            keys32 = getData().pRoot+offset;
            length = *keys32++;
            items32 = (const Resource *)keys32 + length;
        }
        break;
    default:
        errorCode = U_RESOURCE_TYPE_MISMATCH;
        return ResourceTable();
    }
    return ResourceTable(keys16, keys32, items16, items32, length, fTraceInfo);
}

UBool ResourceDataValue::isNoInheritanceMarker() const {
    return ::isNoInheritanceMarker(&getData(), res);
}

int32_t ResourceDataValue::getStringArray(UnicodeString *dest, int32_t capacity,
                                          UErrorCode &errorCode) const {
    return ::getStringArray(&getData(), getArray(errorCode), dest, capacity, errorCode);
}

int32_t ResourceDataValue::getStringArrayOrStringAsArray(UnicodeString *dest, int32_t capacity,
                                                         UErrorCode &errorCode) const {
    if(URES_IS_ARRAY(res)) {
        return ::getStringArray(&getData(), getArray(errorCode), dest, capacity, errorCode);
    }
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    if(dest == NULL ? capacity != 0 : capacity < 0) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    if(capacity < 1) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return 1;
    }
    int32_t sLength;
    const UChar *s = res_getString(fTraceInfo, &getData(), res, &sLength);
    if(s != NULL) {
        dest[0].setTo(TRUE, s, sLength);
        return 1;
    }
    errorCode = U_RESOURCE_TYPE_MISMATCH;
    return 0;
}

UnicodeString ResourceDataValue::getStringOrFirstOfArray(UErrorCode &errorCode) const {
    UnicodeString us;
    if(U_FAILURE(errorCode)) {
        return us;
    }
    int32_t sLength;
    const UChar *s = res_getString(fTraceInfo, &getData(), res, &sLength);
    if(s != NULL) {
        us.setTo(TRUE, s, sLength);
        return us;
    }
    ResourceArray array = getArray(errorCode);
    if(U_FAILURE(errorCode)) {
        return us;
    }
    if(array.getSize() > 0) {
        // Tracing is already performed above (unimportant for trace that this is an array)
        s = res_getStringNoTrace(&getData(), array.internalGetResource(&getData(), 0), &sLength);
        if(s != NULL) {
            us.setTo(TRUE, s, sLength);
            return us;
        }
    }
    errorCode = U_RESOURCE_TYPE_MISMATCH;
    return us;
}

U_NAMESPACE_END

static Resource
makeResourceFrom16(const ResourceData *pResData, int32_t res16) {
    if(res16<pResData->poolStringIndex16Limit) {
        // Pool string, nothing to do.
    } else {
        // Local string, adjust the 16-bit offset to a regular one,
        // with a larger pool string index limit.
        res16=res16-pResData->poolStringIndex16Limit+pResData->poolStringIndexLimit;
    }
    return URES_MAKE_RESOURCE(URES_STRING_V2, res16);
}

U_CAPI Resource U_EXPORT2
res_getTableItemByKey(const ResourceData *pResData, Resource table,
                      int32_t *indexR, const char **key) {
    uint32_t offset=RES_GET_OFFSET(table);
    int32_t length;
    int32_t idx;
    if(key == NULL || *key == NULL) {
        return RES_BOGUS;
    }
    switch(RES_GET_TYPE(table)) {
    case URES_TABLE: {
        if (offset!=0) { /* empty if offset==0 */
            const uint16_t *p= (const uint16_t *)(pResData->pRoot+offset);
            length=*p++;
            *indexR=idx=_res_findTableItem(pResData, p, length, *key, key);
            if(idx>=0) {
                const Resource *p32=(const Resource *)(p+length+(~length&1));
                return p32[idx];
            }
        }
        break;
    }
    case URES_TABLE16: {
        const uint16_t *p=pResData->p16BitUnits+offset;
        length=*p++;
        *indexR=idx=_res_findTableItem(pResData, p, length, *key, key);
        if(idx>=0) {
            return makeResourceFrom16(pResData, p[length+idx]);
        }
        break;
    }
    case URES_TABLE32: {
        if (offset!=0) { /* empty if offset==0 */
            const int32_t *p= pResData->pRoot+offset;
            length=*p++;
            *indexR=idx=_res_findTable32Item(pResData, p, length, *key, key);
            if(idx>=0) {
                return (Resource)p[length+idx];
            }
        }
        break;
    }
    default:
        break;
    }
    return RES_BOGUS;
}

U_CAPI Resource U_EXPORT2
res_getTableItemByIndex(const ResourceData *pResData, Resource table,
                        int32_t indexR, const char **key) {
    uint32_t offset=RES_GET_OFFSET(table);
    int32_t length;
    if (indexR < 0) {
        return RES_BOGUS;
    }
    switch(RES_GET_TYPE(table)) {
    case URES_TABLE: {
        if (offset != 0) { /* empty if offset==0 */
            const uint16_t *p= (const uint16_t *)(pResData->pRoot+offset);
            length=*p++;
            if(indexR<length) {
                const Resource *p32=(const Resource *)(p+length+(~length&1));
                if(key!=NULL) {
                    *key=RES_GET_KEY16(pResData, p[indexR]);
                }
                return p32[indexR];
            }
        }
        break;
    }
    case URES_TABLE16: {
        const uint16_t *p=pResData->p16BitUnits+offset;
        length=*p++;
        if(indexR<length) {
            if(key!=NULL) {
                *key=RES_GET_KEY16(pResData, p[indexR]);
            }
            return makeResourceFrom16(pResData, p[length+indexR]);
        }
        break;
    }
    case URES_TABLE32: {
        if (offset != 0) { /* empty if offset==0 */
            const int32_t *p= pResData->pRoot+offset;
            length=*p++;
            if(indexR<length) {
                if(key!=NULL) {
                    *key=RES_GET_KEY32(pResData, p[indexR]);
                }
                return (Resource)p[length+indexR];
            }
        }
        break;
    }
    default:
        break;
    }
    return RES_BOGUS;
}

U_CAPI Resource U_EXPORT2
res_getResource(const ResourceData *pResData, const char *key) {
    const char *realKey=key;
    int32_t idx;
    return res_getTableItemByKey(pResData, pResData->rootRes, &idx, &realKey);
}


UBool icu::ResourceTable::getKeyAndValue(int32_t i,
                                         const char *&key, icu::ResourceValue &value) const {
    if(0 <= i && i < length) {
        icu::ResourceDataValue &rdValue = static_cast<icu::ResourceDataValue &>(value);
        if (keys16 != nullptr) {
            key = RES_GET_KEY16(&rdValue.getData(), keys16[i]);
        } else {
            key = RES_GET_KEY32(&rdValue.getData(), keys32[i]);
        }
        Resource res;
        if (items16 != nullptr) {
            res = makeResourceFrom16(&rdValue.getData(), items16[i]);
        } else {
            res = items32[i];
        }
        // Note: the ResourceTracer keeps a reference to the field of this
        // ResourceTable. This is OK because the ResourceTable should remain
        // alive for the duration that fields are being read from it
        // (including nested fields).
        rdValue.setResource(res, ResourceTracer(fTraceInfo, key));
        return TRUE;
    }
    return FALSE;
}

UBool icu::ResourceTable::findValue(const char *key, ResourceValue &value) const {
    icu::ResourceDataValue &rdValue = static_cast<icu::ResourceDataValue &>(value);
    const char *realKey = nullptr;
    int32_t i;
    if (keys16 != nullptr) {
        i = _res_findTableItem(&rdValue.getData(), keys16, length, key, &realKey);
    } else {
        i = _res_findTable32Item(&rdValue.getData(), keys32, length, key, &realKey);
    }
    if (i >= 0) {
        Resource res;
        if (items16 != nullptr) {
            res = makeResourceFrom16(&rdValue.getData(), items16[i]);
        } else {
            res = items32[i];
        }
        // Same note about lifetime as in getKeyAndValue().
        rdValue.setResource(res, ResourceTracer(fTraceInfo, key));
        return TRUE;
    }
    return FALSE;
}

U_CAPI Resource U_EXPORT2
res_getArrayItem(const ResourceData *pResData, Resource array, int32_t indexR) {
    uint32_t offset=RES_GET_OFFSET(array);
    if (indexR < 0) {
        return RES_BOGUS;
    }
    switch(RES_GET_TYPE(array)) {
    case URES_ARRAY: {
        if (offset!=0) { /* empty if offset==0 */
            const int32_t *p= pResData->pRoot+offset;
            if(indexR<*p) {
                return (Resource)p[1+indexR];
            }
        }
        break;
    }
    case URES_ARRAY16: {
        const uint16_t *p=pResData->p16BitUnits+offset;
        if(indexR<*p) {
            return makeResourceFrom16(pResData, p[1+indexR]);
        }
        break;
    }
    default:
        break;
    }
    return RES_BOGUS;
}

uint32_t icu::ResourceArray::internalGetResource(const ResourceData *pResData, int32_t i) const {
    if (items16 != NULL) {
        return makeResourceFrom16(pResData, items16[i]);
    } else {
        return items32[i];
    }
}

UBool icu::ResourceArray::getValue(int32_t i, icu::ResourceValue &value) const {
    if(0 <= i && i < length) {
        icu::ResourceDataValue &rdValue = static_cast<icu::ResourceDataValue &>(value);
        // Note: the ResourceTracer keeps a reference to the field of this
        // ResourceArray. This is OK because the ResourceArray should remain
        // alive for the duration that fields are being read from it
        // (including nested fields).
        rdValue.setResource(
            internalGetResource(&rdValue.getData(), i),
            ResourceTracer(fTraceInfo, i));
        return TRUE;
    }
    return FALSE;
}

U_CFUNC Resource
res_findResource(const ResourceData *pResData, Resource r, char** path, const char** key) {
  char *pathP = *path, *nextSepP = *path;
  char *closeIndex = NULL;
  Resource t1 = r;
  Resource t2;
  int32_t indexR = 0;
  UResType type = (UResType)RES_GET_TYPE(t1);

  /* if you come in with an empty path, you'll be getting back the same resource */
  if(!uprv_strlen(pathP)) {
      return r;
  }

  /* one needs to have an aggregate resource in order to search in it */
  if(!URES_IS_CONTAINER(type)) {
      return RES_BOGUS;
  }

  while(nextSepP && *pathP && t1 != RES_BOGUS && URES_IS_CONTAINER(type)) {
    /* Iteration stops if: the path has been consumed, we found a non-existing
     * resource (t1 == RES_BOGUS) or we found a scalar resource (including alias)
     */
    nextSepP = uprv_strchr(pathP, RES_PATH_SEPARATOR);
    /* if there are more separators, terminate string
     * and set path to the remaining part of the string
     */
    if(nextSepP != NULL) {
      if(nextSepP == pathP) {
        // Empty key string.
        return RES_BOGUS;
      }
      *nextSepP = 0; /* overwrite the separator with a NUL to terminate the key */
      *path = nextSepP+1;
    } else {
      *path = uprv_strchr(pathP, 0);
    }

    /* if the resource is a table */
    /* try the key based access */
    if(URES_IS_TABLE(type)) {
      *key = pathP;
      t2 = res_getTableItemByKey(pResData, t1, &indexR, key);
      if(t2 == RES_BOGUS) {
        /* if we fail to get the resource by key, maybe we got an index */
        indexR = uprv_strtol(pathP, &closeIndex, 10);
        if(indexR >= 0 && *closeIndex == 0) {
          /* if we indeed have an index, try to get the item by index */
          t2 = res_getTableItemByIndex(pResData, t1, indexR, key);
        } // else t2 is already RES_BOGUS
      }
    } else if(URES_IS_ARRAY(type)) {
      indexR = uprv_strtol(pathP, &closeIndex, 10);
      if(indexR >= 0 && *closeIndex == 0) {
        t2 = res_getArrayItem(pResData, t1, indexR);
      } else {
        t2 = RES_BOGUS; /* have an array, but don't have a valid index */
      }
      *key = NULL;
    } else { /* can't do much here, except setting t2 to bogus */
      t2 = RES_BOGUS;
    }
    t1 = t2;
    type = (UResType)RES_GET_TYPE(t1);
    /* position pathP to next resource key/index */
    pathP = *path;
  }

  return t1;
}

/* resource bundle swapping ------------------------------------------------- */

/*
 * Need to always enumerate the entire item tree,
 * track the lowest address of any item to use as the limit for char keys[],
 * track the highest address of any item to return the size of the data.
 *
 * We should have thought of storing those in the data...
 * It is possible to extend the data structure by putting additional values
 * in places that are inaccessible by ordinary enumeration of the item tree.
 * For example, additional integers could be stored at the beginning or
 * end of the key strings; this could be indicated by a minor version number,
 * and the data swapping would have to know about these values.
 *
 * The data structure does not forbid keys to be shared, so we must swap
 * all keys once instead of each key when it is referenced.
 *
 * These swapping functions assume that a resource bundle always has a length
 * that is a multiple of 4 bytes.
 * Currently, this is trivially true because genrb writes bundle tree leaves
 * physically first, before their branches, so that the root table with its
 * array of resource items (uint32_t values) is always last.
 */

/* definitions for table sorting ------------------------ */

/*
 * row of a temporary array
 *
 * gets platform-endian key string indexes and sorting indexes;
 * after sorting this array by keys, the actual key/value arrays are permutated
 * according to the sorting indexes
 */
typedef struct Row {
    int32_t keyIndex, sortIndex;
} Row;

static int32_t U_CALLCONV
ures_compareRows(const void *context, const void *left, const void *right) {
    const char *keyChars=(const char *)context;
    return (int32_t)uprv_strcmp(keyChars+((const Row *)left)->keyIndex,
                                keyChars+((const Row *)right)->keyIndex);
}

typedef struct TempTable {
    const char *keyChars;
    Row *rows;
    int32_t *resort;
    uint32_t *resFlags;
    int32_t localKeyLimit;
    uint8_t majorFormatVersion;
} TempTable;

enum {
    STACK_ROW_CAPACITY=200
};

/* The table item key string is not locally available. */
static const char *const gUnknownKey="";

/* resource table key for collation binaries: "%%CollationBin" */
static const UChar gCollationBinKey[]={
    0x25, 0x25,
    0x43, 0x6f, 0x6c, 0x6c, 0x61, 0x74, 0x69, 0x6f, 0x6e,
    0x42, 0x69, 0x6e,
    0
};

/*
 * swap one resource item
 */
static void
ures_swapResource(const UDataSwapper *ds,
                  const Resource *inBundle, Resource *outBundle,
                  Resource res, /* caller swaps res itself */
                  const char *key,
                  TempTable *pTempTable,
                  UErrorCode *pErrorCode) {
    const Resource *p;
    Resource *q;
    int32_t offset, count;

    switch(RES_GET_TYPE(res)) {
    case URES_TABLE16:
    case URES_STRING_V2:
    case URES_INT:
    case URES_ARRAY16:
        /* integer, or points to 16-bit units, nothing to do here */
        return;
    default:
        break;
    }

    /* all other types use an offset to point to their data */
    offset=(int32_t)RES_GET_OFFSET(res);
    if(offset==0) {
        /* special offset indicating an empty item */
        return;
    }
    if(pTempTable->resFlags[offset>>5]&((uint32_t)1<<(offset&0x1f))) {
        /* we already swapped this resource item */
        return;
    } else {
        /* mark it as swapped now */
        pTempTable->resFlags[offset>>5]|=((uint32_t)1<<(offset&0x1f));
    }

    p=inBundle+offset;
    q=outBundle+offset;

    switch(RES_GET_TYPE(res)) {
    case URES_ALIAS:
        /* physically same value layout as string, fall through */
        U_FALLTHROUGH;
    case URES_STRING:
        count=udata_readInt32(ds, (int32_t)*p);
        /* swap length */
        ds->swapArray32(ds, p, 4, q, pErrorCode);
        /* swap each UChar (the terminating NUL would not change) */
        ds->swapArray16(ds, p+1, 2*count, q+1, pErrorCode);
        break;
    case URES_BINARY:
        count=udata_readInt32(ds, (int32_t)*p);
        /* swap length */
        ds->swapArray32(ds, p, 4, q, pErrorCode);
        /* no need to swap or copy bytes - ures_swap() copied them all */

        /* swap known formats */
#if !UCONFIG_NO_COLLATION
        if( key!=NULL &&  /* the binary is in a table */
            (key!=gUnknownKey ?
                /* its table key string is "%%CollationBin" */
                0==ds->compareInvChars(ds, key, -1,
                                       gCollationBinKey, UPRV_LENGTHOF(gCollationBinKey)-1) :
                /* its table key string is unknown but it looks like a collation binary */
                ucol_looksLikeCollationBinary(ds, p+1, count))
        ) {
            ucol_swap(ds, p+1, count, q+1, pErrorCode);
        }
#endif
        break;
    case URES_TABLE:
    case URES_TABLE32:
        {
            const uint16_t *pKey16;
            uint16_t *qKey16;

            const int32_t *pKey32;
            int32_t *qKey32;

            Resource item;
            int32_t i, oldIndex;

            if(RES_GET_TYPE(res)==URES_TABLE) {
                /* get table item count */
                pKey16=(const uint16_t *)p;
                qKey16=(uint16_t *)q;
                count=ds->readUInt16(*pKey16);

                pKey32=qKey32=NULL;

                /* swap count */
                ds->swapArray16(ds, pKey16++, 2, qKey16++, pErrorCode);

                offset+=((1+count)+1)/2;
            } else {
                /* get table item count */
                pKey32=(const int32_t *)p;
                qKey32=(int32_t *)q;
                count=udata_readInt32(ds, *pKey32);

                pKey16=qKey16=NULL;

                /* swap count */
                ds->swapArray32(ds, pKey32++, 4, qKey32++, pErrorCode);

                offset+=1+count;
            }

            if(count==0) {
                break;
            }

            p=inBundle+offset; /* pointer to table resources */
            q=outBundle+offset;

            /* recurse */
            for(i=0; i<count; ++i) {
                const char *itemKey=gUnknownKey;
                if(pKey16!=NULL) {
                    int32_t keyOffset=ds->readUInt16(pKey16[i]);
                    if(keyOffset<pTempTable->localKeyLimit) {
                        itemKey=(const char *)outBundle+keyOffset;
                    }
                } else {
                    int32_t keyOffset=udata_readInt32(ds, pKey32[i]);
                    if(keyOffset>=0) {
                        itemKey=(const char *)outBundle+keyOffset;
                    }
                }
                item=ds->readUInt32(p[i]);
                ures_swapResource(ds, inBundle, outBundle, item, itemKey, pTempTable, pErrorCode);
                if(U_FAILURE(*pErrorCode)) {
                    udata_printError(ds, "ures_swapResource(table res=%08x)[%d].recurse(%08x) failed\n",
                                     res, i, item);
                    return;
                }
            }

            if(pTempTable->majorFormatVersion>1 || ds->inCharset==ds->outCharset) {
                /* no need to sort, just swap the offset/value arrays */
                if(pKey16!=NULL) {
                    ds->swapArray16(ds, pKey16, count*2, qKey16, pErrorCode);
                    ds->swapArray32(ds, p, count*4, q, pErrorCode);
                } else {
                    /* swap key offsets and items as one array */
                    ds->swapArray32(ds, pKey32, count*2*4, qKey32, pErrorCode);
                }
                break;
            }

            /*
             * We need to sort tables by outCharset key strings because they
             * sort differently for different charset families.
             * ures_swap() already set pTempTable->keyChars appropriately.
             * First we set up a temporary table with the key indexes and
             * sorting indexes and sort that.
             * Then we permutate and copy/swap the actual values.
             */
            if(pKey16!=NULL) {
                for(i=0; i<count; ++i) {
                    pTempTable->rows[i].keyIndex=ds->readUInt16(pKey16[i]);
                    pTempTable->rows[i].sortIndex=i;
                }
            } else {
                for(i=0; i<count; ++i) {
                    pTempTable->rows[i].keyIndex=udata_readInt32(ds, pKey32[i]);
                    pTempTable->rows[i].sortIndex=i;
                }
            }
            uprv_sortArray(pTempTable->rows, count, sizeof(Row),
                           ures_compareRows, pTempTable->keyChars,
                           FALSE, pErrorCode);
            if(U_FAILURE(*pErrorCode)) {
                udata_printError(ds, "ures_swapResource(table res=%08x).uprv_sortArray(%d items) failed\n",
                                 res, count);
                return;
            }

            /*
             * copy/swap/permutate items
             *
             * If we swap in-place, then the permutation must use another
             * temporary array (pTempTable->resort)
             * before the results are copied to the outBundle.
             */
            /* keys */
            if(pKey16!=NULL) {
                uint16_t *rKey16;

                if(pKey16!=qKey16) {
                    rKey16=qKey16;
                } else {
                    rKey16=(uint16_t *)pTempTable->resort;
                }
                for(i=0; i<count; ++i) {
                    oldIndex=pTempTable->rows[i].sortIndex;
                    ds->swapArray16(ds, pKey16+oldIndex, 2, rKey16+i, pErrorCode);
                }
                if(qKey16!=rKey16) {
                    uprv_memcpy(qKey16, rKey16, 2*count);
                }
            } else {
                int32_t *rKey32;

                if(pKey32!=qKey32) {
                    rKey32=qKey32;
                } else {
                    rKey32=pTempTable->resort;
                }
                for(i=0; i<count; ++i) {
                    oldIndex=pTempTable->rows[i].sortIndex;
                    ds->swapArray32(ds, pKey32+oldIndex, 4, rKey32+i, pErrorCode);
                }
                if(qKey32!=rKey32) {
                    uprv_memcpy(qKey32, rKey32, 4*count);
                }
            }

            /* resources */
            {
                Resource *r;


                if(p!=q) {
                    r=q;
                } else {
                    r=(Resource *)pTempTable->resort;
                }
                for(i=0; i<count; ++i) {
                    oldIndex=pTempTable->rows[i].sortIndex;
                    ds->swapArray32(ds, p+oldIndex, 4, r+i, pErrorCode);
                }
                if(q!=r) {
                    uprv_memcpy(q, r, 4*count);
                }
            }
        }
        break;
    case URES_ARRAY:
        {
            Resource item;
            int32_t i;

            count=udata_readInt32(ds, (int32_t)*p);
            /* swap length */
            ds->swapArray32(ds, p++, 4, q++, pErrorCode);

            /* recurse */
            for(i=0; i<count; ++i) {
                item=ds->readUInt32(p[i]);
                ures_swapResource(ds, inBundle, outBundle, item, NULL, pTempTable, pErrorCode);
                if(U_FAILURE(*pErrorCode)) {
                    udata_printError(ds, "ures_swapResource(array res=%08x)[%d].recurse(%08x) failed\n",
                                     res, i, item);
                    return;
                }
            }

            /* swap items */
            ds->swapArray32(ds, p, 4*count, q, pErrorCode);
        }
        break;
    case URES_INT_VECTOR:
        count=udata_readInt32(ds, (int32_t)*p);
        /* swap length and each integer */
        ds->swapArray32(ds, p, 4*(1+count), q, pErrorCode);
        break;
    default:
        /* also catches RES_BOGUS */
        *pErrorCode=U_UNSUPPORTED_ERROR;
        break;
    }
}

U_CAPI int32_t U_EXPORT2
ures_swap(const UDataSwapper *ds,
          const void *inData, int32_t length, void *outData,
          UErrorCode *pErrorCode) {
    const UDataInfo *pInfo;
    const Resource *inBundle;
    Resource rootRes;
    int32_t headerSize, maxTableLength;

    Row rows[STACK_ROW_CAPACITY];
    int32_t resort[STACK_ROW_CAPACITY];
    TempTable tempTable;

    const int32_t *inIndexes;

    /* the following integers count Resource item offsets (4 bytes each), not bytes */
    int32_t bundleLength, indexLength, keysBottom, keysTop, resBottom, top;

    /* udata_swapDataHeader checks the arguments */
    headerSize=udata_swapDataHeader(ds, inData, length, outData, pErrorCode);
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* check data format and format version */
    pInfo=(const UDataInfo *)((const char *)inData+4);
    if(!(
        pInfo->dataFormat[0]==0x52 &&   /* dataFormat="ResB" */
        pInfo->dataFormat[1]==0x65 &&
        pInfo->dataFormat[2]==0x73 &&
        pInfo->dataFormat[3]==0x42 &&
        /* formatVersion 1.1+ or 2.x or 3.x */
        ((pInfo->formatVersion[0]==1 && pInfo->formatVersion[1]>=1) ||
            pInfo->formatVersion[0]==2 || pInfo->formatVersion[0]==3)
    )) {
        udata_printError(ds, "ures_swap(): data format %02x.%02x.%02x.%02x (format version %02x.%02x) is not a resource bundle\n",
                         pInfo->dataFormat[0], pInfo->dataFormat[1],
                         pInfo->dataFormat[2], pInfo->dataFormat[3],
                         pInfo->formatVersion[0], pInfo->formatVersion[1]);
        *pErrorCode=U_UNSUPPORTED_ERROR;
        return 0;
    }
    tempTable.majorFormatVersion=pInfo->formatVersion[0];

    /* a resource bundle must contain at least one resource item */
    if(length<0) {
        bundleLength=-1;
    } else {
        bundleLength=(length-headerSize)/4;

        /* formatVersion 1.1 must have a root item and at least 5 indexes */
        if(bundleLength<(1+5)) {
            udata_printError(ds, "ures_swap(): too few bytes (%d after header) for a resource bundle\n",
                             length-headerSize);
            *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }
    }

    inBundle=(const Resource *)((const char *)inData+headerSize);
    rootRes=ds->readUInt32(*inBundle);

    /* formatVersion 1.1 adds the indexes[] array */
    inIndexes=(const int32_t *)(inBundle+1);

    indexLength=udata_readInt32(ds, inIndexes[URES_INDEX_LENGTH])&0xff;
    if(indexLength<=URES_INDEX_MAX_TABLE_LENGTH) {
        udata_printError(ds, "ures_swap(): too few indexes for a 1.1+ resource bundle\n");
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }
    keysBottom=1+indexLength;
    keysTop=udata_readInt32(ds, inIndexes[URES_INDEX_KEYS_TOP]);
    if(indexLength>URES_INDEX_16BIT_TOP) {
        resBottom=udata_readInt32(ds, inIndexes[URES_INDEX_16BIT_TOP]);
    } else {
        resBottom=keysTop;
    }
    top=udata_readInt32(ds, inIndexes[URES_INDEX_BUNDLE_TOP]);
    maxTableLength=udata_readInt32(ds, inIndexes[URES_INDEX_MAX_TABLE_LENGTH]);

    if(0<=bundleLength && bundleLength<top) {
        udata_printError(ds, "ures_swap(): resource top %d exceeds bundle length %d\n",
                         top, bundleLength);
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }
    if(keysTop>(1+indexLength)) {
        tempTable.localKeyLimit=keysTop<<2;
    } else {
        tempTable.localKeyLimit=0;
    }

    if(length>=0) {
        Resource *outBundle=(Resource *)((char *)outData+headerSize);

        /* track which resources we have already swapped */
        uint32_t stackResFlags[STACK_ROW_CAPACITY];
        int32_t resFlagsLength;

        /*
         * We need one bit per 4 resource bundle bytes so that we can track
         * every possible Resource for whether we have swapped it already.
         * Multiple Resource words can refer to the same bundle offsets
         * for sharing identical values.
         * We could optimize this by allocating only for locations above
         * where Resource values are stored (above keys & strings).
         */
        resFlagsLength=(length+31)>>5;          /* number of bytes needed */
        resFlagsLength=(resFlagsLength+3)&~3;   /* multiple of 4 bytes for uint32_t */
        if(resFlagsLength<=(int32_t)sizeof(stackResFlags)) {
            tempTable.resFlags=stackResFlags;
        } else {
            tempTable.resFlags=(uint32_t *)uprv_malloc(resFlagsLength);
            if(tempTable.resFlags==NULL) {
                udata_printError(ds, "ures_swap(): unable to allocate memory for tracking resources\n");
                *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
                return 0;
            }
        }
        uprv_memset(tempTable.resFlags, 0, resFlagsLength);

        /* copy the bundle for binary and inaccessible data */
        if(inData!=outData) {
            uprv_memcpy(outBundle, inBundle, 4*top);
        }

        /* swap the key strings, but not the padding bytes (0xaa) after the last string and its NUL */
        udata_swapInvStringBlock(ds, inBundle+keysBottom, 4*(keysTop-keysBottom),
                                    outBundle+keysBottom, pErrorCode);
        if(U_FAILURE(*pErrorCode)) {
            udata_printError(ds, "ures_swap().udata_swapInvStringBlock(keys[%d]) failed\n", 4*(keysTop-keysBottom));
            return 0;
        }

        /* swap the 16-bit units (strings, table16, array16) */
        if(keysTop<resBottom) {
            ds->swapArray16(ds, inBundle+keysTop, (resBottom-keysTop)*4, outBundle+keysTop, pErrorCode);
            if(U_FAILURE(*pErrorCode)) {
                udata_printError(ds, "ures_swap().swapArray16(16-bit units[%d]) failed\n", 2*(resBottom-keysTop));
                return 0;
            }
        }

        /* allocate the temporary table for sorting resource tables */
        tempTable.keyChars=(const char *)outBundle; /* sort by outCharset */
        if(tempTable.majorFormatVersion>1 || maxTableLength<=STACK_ROW_CAPACITY) {
            tempTable.rows=rows;
            tempTable.resort=resort;
        } else {
            tempTable.rows=(Row *)uprv_malloc(maxTableLength*sizeof(Row)+maxTableLength*4);
            if(tempTable.rows==NULL) {
                udata_printError(ds, "ures_swap(): unable to allocate memory for sorting tables (max length: %d)\n",
                                 maxTableLength);
                *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
                if(tempTable.resFlags!=stackResFlags) {
                    uprv_free(tempTable.resFlags);
                }
                return 0;
            }
            tempTable.resort=(int32_t *)(tempTable.rows+maxTableLength);
        }

        /* swap the resources */
        ures_swapResource(ds, inBundle, outBundle, rootRes, NULL, &tempTable, pErrorCode);
        if(U_FAILURE(*pErrorCode)) {
            udata_printError(ds, "ures_swapResource(root res=%08x) failed\n",
                             rootRes);
        }

        if(tempTable.rows!=rows) {
            uprv_free(tempTable.rows);
        }
        if(tempTable.resFlags!=stackResFlags) {
            uprv_free(tempTable.resFlags);
        }

        /* swap the root resource and indexes */
        ds->swapArray32(ds, inBundle, keysBottom*4, outBundle, pErrorCode);
    }

    return headerSize+4*top;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1999-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*
* File USC_IMPL.C
*
* Modification History:
*
*   Date        Name        Description
*   07/08/2002  Eric Mader  Creation.
******************************************************************************
*/


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1999-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*
* File USC_IMPL.H
*
* Modification History:
*
*   Date        Name        Description
*   07/08/2002  Eric Mader  Creation.
******************************************************************************
*/




/**
 * <code>UScriptRun</code> is used to find runs of characters in
 * the same script. It implements a simple iterator over an array
 * of characters. The iterator will resolve script-neutral characters
 * like punctuation into the script of the surrounding characters.
 *
 * The iterator will try to match paired punctuation. If it sees an
 * opening punctuation character, it will remember the script that
 * was assigned to that character, and assign the same script to the
 * matching closing punctuation.
 *
 * Scripts are chosen based on the <code>UScriptCode</code> enumeration.
 * No attempt is made to combine related scripts into a single run. In
 * particular, Hiragana, Katakana, and Han characters will appear in seperate
 * runs.

 * Here is an example of how to iterate over script runs:
 * <pre>
 * \code
 * void printScriptRuns(const UChar *text, int32_t length)
 * {
 *     UErrorCode error = U_ZERO_ERROR;
 *     UScriptRun *scriptRun = uscript_openRun(text, testLength, &error);
 *     int32_t start = 0, limit = 0;
 *     UScriptCode code = USCRIPT_INVALID_CODE;
 *
 *     while (uscript_nextRun(&start, &limit, &code)) {
 *         printf("Script '%s' from %d to %d.\n", uscript_getName(code), start, limit);
 *     }
 *
 *     uscript_closeRun(scriptRun);
 *  }
 * </pre>
 */
struct UScriptRun;

typedef struct UScriptRun UScriptRun;

/**
 * Create a <code>UScriptRun</code> object for iterating over the given text. This object must
 * be freed using <code>uscript_closeRun()</code>. Note that this object does not copy the source text,
 * only the pointer to it. You must make sure that the pointer remains valid until you call
 * <code>uscript_closeRun()</code> or <code>uscript_setRunText()</code>.
 *
 * @param src is the address of the array of characters over which to iterate.
 *        if <code>src == NULL</code> and <code>length == 0</code>,
 *        an empty <code>UScriptRun</code> object will be returned.
 *
 * @param length is the number of characters over which to iterate.
 *
 * @param pErrorCode is a pointer to a valid <code>UErrorCode</code> value. If this value
 *        indicates a failure on entry, the function will immediately return.
 *        On exit the value will indicate the success of the operation.
 *
 * @return the address of <code>UScriptRun</code> object which will iterate over the text,
 *         or <code>NULL</code> if the operation failed.
 */
U_CAPI UScriptRun * U_EXPORT2
uscript_openRun(const UChar *src, int32_t length, UErrorCode *pErrorCode);

/**
 * Frees the given <code>UScriptRun</code> object and any storage associated with it.
 * On return, scriptRun no longer points to a valid <code>UScriptRun</code> object.
 *
 * @param scriptRun is the <code>UScriptRun</code> object which will be freed.
 */
U_CAPI void U_EXPORT2
uscript_closeRun(UScriptRun *scriptRun);

/**
 * Reset the <code>UScriptRun</code> object so that it will start iterating from
 * the beginning.
 *
 * @param scriptRun is the address of the <code>UScriptRun</code> object to be reset.
 */
U_CAPI void U_EXPORT2
uscript_resetRun(UScriptRun *scriptRun);

/**
 * Change the text over which the given <code>UScriptRun</code> object iterates.
 *
 * @param scriptRun is the <code>UScriptRun</code> object which will be changed.
 *
 * @param src is the address of the new array of characters over which to iterate.
 *        If <code>src == NULL</code> and <code>length == 0</code>,
 *        the <code>UScriptRun</code> object will become empty.
 *
 * @param length is the new number of characters over which to iterate
 *
 * @param pErrorCode is a pointer to a valid <code>UErrorCode</code> value. If this value
 *        indicates a failure on entry, the function will immediately return.
 *        On exit the value will indicate the success of the operation.
 */
U_CAPI void U_EXPORT2
uscript_setRunText(UScriptRun *scriptRun, const UChar *src, int32_t length, UErrorCode *pErrorCode);

/**
 * Advance the <code>UScriptRun</code> object to the next script run, return the start and limit
 * offsets, and the script of the run.
 *
 * @param scriptRun is the address of the <code>UScriptRun</code> object.
 *
 * @param pRunStart is a pointer to the variable to receive the starting offset of the next run.
 *        This pointer can be <code>NULL</code> if the value is not needed.
 *
 * @param pRunLimit is a pointer to the variable to receive the limit offset of the next run.
 *        This pointer can be <code>NULL</code> if the value is not needed.
 *
 * @param pRunScript is a pointer to the variable to receive the UScriptCode for the
 *        script of the current run. This pointer can be <code>NULL</code> if the value is not needed.
 *
 * @return true if there was another script run.
 */
U_CAPI UBool U_EXPORT2
uscript_nextRun(UScriptRun *scriptRun, int32_t *pRunStart, int32_t *pRunLimit, UScriptCode *pRunScript);




#define PAREN_STACK_DEPTH 32

#define MOD(sp) ((sp) % PAREN_STACK_DEPTH)
#define LIMIT_INC(sp) (((sp) < PAREN_STACK_DEPTH)? (sp) + 1 : PAREN_STACK_DEPTH)
#define INC(sp,count) (MOD((sp) + (count)))
#define INC1(sp) (INC(sp, 1))
#define DEC(sp,count) (MOD((sp) + PAREN_STACK_DEPTH - (count)))
#define DEC1(sp) (DEC(sp, 1))
#define STACK_IS_EMPTY(scriptRun) ((scriptRun)->pushCount <= 0)
#define STACK_IS_NOT_EMPTY(scriptRun) (! STACK_IS_EMPTY(scriptRun))
#define TOP(scriptRun) ((scriptRun)->parenStack[(scriptRun)->parenSP])
#define SYNC_FIXUP(scriptRun) ((scriptRun)->fixupCount = 0)

struct ParenStackEntry
{
    int32_t pairIndex;
    UScriptCode scriptCode;
};

struct UScriptRun
{
    int32_t textLength;
    const UChar *textArray;

    int32_t scriptStart;
    int32_t scriptLimit;
    UScriptCode scriptCode;

    struct ParenStackEntry parenStack[PAREN_STACK_DEPTH];
    int32_t parenSP;
    int32_t pushCount;
    int32_t fixupCount;
};

static int8_t highBit(int32_t value);

static const UChar32 pairedChars[] = {
    0x0028, 0x0029, /* ascii paired punctuation */
    0x003c, 0x003e,
    0x005b, 0x005d,
    0x007b, 0x007d,
    0x00ab, 0x00bb, /* guillemets */
    0x2018, 0x2019, /* general punctuation */
    0x201c, 0x201d,
    0x2039, 0x203a,
    0x3008, 0x3009, /* chinese paired punctuation */
    0x300a, 0x300b,
    0x300c, 0x300d,
    0x300e, 0x300f,
    0x3010, 0x3011,
    0x3014, 0x3015,
    0x3016, 0x3017,
    0x3018, 0x3019,
    0x301a, 0x301b
};

static void push(UScriptRun *scriptRun, int32_t pairIndex, UScriptCode scriptCode)
{
    scriptRun->pushCount  = LIMIT_INC(scriptRun->pushCount);
    scriptRun->fixupCount = LIMIT_INC(scriptRun->fixupCount);
    
    scriptRun->parenSP = INC1(scriptRun->parenSP);
    scriptRun->parenStack[scriptRun->parenSP].pairIndex  = pairIndex;
    scriptRun->parenStack[scriptRun->parenSP].scriptCode = scriptCode;
}

static void pop(UScriptRun *scriptRun)
{
    if (STACK_IS_EMPTY(scriptRun)) {
        return;
    }
    
    if (scriptRun->fixupCount > 0) {
        scriptRun->fixupCount -= 1;
    }
    
    scriptRun->pushCount -= 1;
    scriptRun->parenSP = DEC1(scriptRun->parenSP);
    
    /* If the stack is now empty, reset the stack
       pointers to their initial values.
     */
    if (STACK_IS_EMPTY(scriptRun)) {
        scriptRun->parenSP = -1;
    }
}

static void fixup(UScriptRun *scriptRun, UScriptCode scriptCode)
{
    int32_t fixupSP = DEC(scriptRun->parenSP, scriptRun->fixupCount);
    
    while (scriptRun->fixupCount-- > 0) {
        fixupSP = INC1(fixupSP);
        scriptRun->parenStack[fixupSP].scriptCode = scriptCode;
    }
}

static int8_t
highBit(int32_t value)
{
    int8_t bit = 0;

    if (value <= 0) {
        return -32;
    }

    if (value >= 1 << 16) {
        value >>= 16;
        bit += 16;
    }

    if (value >= 1 << 8) {
        value >>= 8;
        bit += 8;
    }

    if (value >= 1 << 4) {
        value >>= 4;
        bit += 4;
    }

    if (value >= 1 << 2) {
        value >>= 2;
        bit += 2;
    }

    if (value >= 1 << 1) {
        //value >>= 1;
        bit += 1;
    }

    return bit;
}

static int32_t
getPairIndex(UChar32 ch)
{
    int32_t pairedCharCount = UPRV_LENGTHOF(pairedChars);
    int32_t pairedCharPower = 1 << highBit(pairedCharCount);
    int32_t pairedCharExtra = pairedCharCount - pairedCharPower;

    int32_t probe = pairedCharPower;
    int32_t pairIndex = 0;

    if (ch >= pairedChars[pairedCharExtra]) {
        pairIndex = pairedCharExtra;
    }

    while (probe > (1 << 0)) {
        probe >>= 1;

        if (ch >= pairedChars[pairIndex + probe]) {
            pairIndex += probe;
        }
    }

    if (pairedChars[pairIndex] != ch) {
        pairIndex = -1;
    }

    return pairIndex;
}

static UBool
sameScript(UScriptCode scriptOne, UScriptCode scriptTwo)
{
    return scriptOne <= USCRIPT_INHERITED || scriptTwo <= USCRIPT_INHERITED || scriptOne == scriptTwo;
}

U_CAPI UScriptRun * U_EXPORT2
uscript_openRun(const UChar *src, int32_t length, UErrorCode *pErrorCode)
{
    UScriptRun *result = NULL;

    if (pErrorCode == NULL || U_FAILURE(*pErrorCode)) {
        return NULL;
    }

    result = (UScriptRun *)uprv_malloc(sizeof (UScriptRun));

    if (result == NULL) {
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    uscript_setRunText(result, src, length, pErrorCode);

    /* Release the UScriptRun if uscript_setRunText() returns an error */
    if (U_FAILURE(*pErrorCode)) {
        uprv_free(result);
        result = NULL;
    }

    return result;
}

U_CAPI void U_EXPORT2
uscript_closeRun(UScriptRun *scriptRun)
{
    if (scriptRun != NULL) {
        uprv_free(scriptRun);
    }
}

U_CAPI void U_EXPORT2
uscript_resetRun(UScriptRun *scriptRun)
{
    if (scriptRun != NULL) {
        scriptRun->scriptStart = 0;
        scriptRun->scriptLimit = 0;
        scriptRun->scriptCode  = USCRIPT_INVALID_CODE;
        scriptRun->parenSP     = -1;
        scriptRun->pushCount   =  0;
        scriptRun->fixupCount  =  0;
    }
}

U_CAPI void U_EXPORT2
uscript_setRunText(UScriptRun *scriptRun, const UChar *src, int32_t length, UErrorCode *pErrorCode)
{
    if (pErrorCode == NULL || U_FAILURE(*pErrorCode)) {
        return;
    }

    if (scriptRun == NULL || length < 0 || ((src == NULL) != (length == 0))) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    scriptRun->textArray  = src;
    scriptRun->textLength = length;

    uscript_resetRun(scriptRun);
}

U_CAPI UBool U_EXPORT2
uscript_nextRun(UScriptRun *scriptRun, int32_t *pRunStart, int32_t *pRunLimit, UScriptCode *pRunScript)
{
    UErrorCode error = U_ZERO_ERROR;

    /* if we've fallen off the end of the text, we're done */
    if (scriptRun == NULL || scriptRun->scriptLimit >= scriptRun->textLength) {
        return FALSE;
    }
    
    SYNC_FIXUP(scriptRun);
    scriptRun->scriptCode = USCRIPT_COMMON;

    for (scriptRun->scriptStart = scriptRun->scriptLimit; scriptRun->scriptLimit < scriptRun->textLength; scriptRun->scriptLimit += 1) {
        UChar   high = scriptRun->textArray[scriptRun->scriptLimit];
        UChar32 ch   = high;
        UScriptCode sc;
        int32_t pairIndex;

        /*
         * if the character is a high surrogate and it's not the last one
         * in the text, see if it's followed by a low surrogate
         */
        if (high >= 0xD800 && high <= 0xDBFF && scriptRun->scriptLimit < scriptRun->textLength - 1) {
            UChar low = scriptRun->textArray[scriptRun->scriptLimit + 1];

            /*
             * if it is followed by a low surrogate,
             * consume it and form the full character
             */
            if (low >= 0xDC00 && low <= 0xDFFF) {
                ch = (high - 0xD800) * 0x0400 + low - 0xDC00 + 0x10000;
                scriptRun->scriptLimit += 1;
            }
        }

        sc = uscript_getScript(ch, &error);
        pairIndex = getPairIndex(ch);

        /*
         * Paired character handling:
         *
         * if it's an open character, push it onto the stack.
         * if it's a close character, find the matching open on the
         * stack, and use that script code. Any non-matching open
         * characters above it on the stack will be poped.
         */
        if (pairIndex >= 0) {
            if ((pairIndex & 1) == 0) {
                push(scriptRun, pairIndex, scriptRun->scriptCode);
            } else {
                int32_t pi = pairIndex & ~1;

                while (STACK_IS_NOT_EMPTY(scriptRun) && TOP(scriptRun).pairIndex != pi) {
                    pop(scriptRun);
                }

                if (STACK_IS_NOT_EMPTY(scriptRun)) {
                    sc = TOP(scriptRun).scriptCode;
                }
            }
        }

        if (sameScript(scriptRun->scriptCode, sc)) {
            if (scriptRun->scriptCode <= USCRIPT_INHERITED && sc > USCRIPT_INHERITED) {
                scriptRun->scriptCode = sc;

                fixup(scriptRun, scriptRun->scriptCode);
            }

            /*
             * if this character is a close paired character,
             * pop the matching open character from the stack
             */
            if (pairIndex >= 0 && (pairIndex & 1) != 0) {
                pop(scriptRun);
            }
        } else {
            /*
             * if the run broke on a surrogate pair,
             * end it before the high surrogate
             */
            if (ch >= 0x10000) {
                scriptRun->scriptLimit -= 1;
            }

            break;
        }
    }


    if (pRunStart != NULL) {
        *pRunStart = scriptRun->scriptStart;
    }

    if (pRunLimit != NULL) {
        *pRunLimit = scriptRun->scriptLimit;
    }

    if (pRunScript != NULL) {
        *pRunScript = scriptRun->scriptCode;
    }

    return TRUE;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1997-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*
* File USCRIPT.C
*
* Modification History:
*
*   Date        Name        Description
*   07/06/2001    Ram         Creation.
******************************************************************************
*/










static const UScriptCode JAPANESE[3] = { USCRIPT_KATAKANA, USCRIPT_HIRAGANA, USCRIPT_HAN };
static const UScriptCode KOREAN[2] = { USCRIPT_HANGUL, USCRIPT_HAN };
static const UScriptCode HAN_BOPO[2] = { USCRIPT_HAN, USCRIPT_BOPOMOFO };

static int32_t
setCodes(const UScriptCode *src, int32_t length,
         UScriptCode *dest, int32_t capacity, UErrorCode *err) {
    int32_t i;
    if(U_FAILURE(*err)) { return 0; }
    if(length > capacity) {
        *err = U_BUFFER_OVERFLOW_ERROR;
        return length;
    }
    for(i = 0; i < length; ++i) {
        dest[i] = src[i];
    }
    return length;
}

static int32_t
setOneCode(UScriptCode script, UScriptCode *scripts, int32_t capacity, UErrorCode *err) {
    if(U_FAILURE(*err)) { return 0; }
    if(1 > capacity) {
        *err = U_BUFFER_OVERFLOW_ERROR;
        return 1;
    }
    scripts[0] = script;
    return 1;
}

static int32_t
getCodesFromLocale(const char *locale,
                   UScriptCode *scripts, int32_t capacity, UErrorCode *err) {
    UErrorCode internalErrorCode = U_ZERO_ERROR;
    char lang[8];
    char script[8];
    int32_t scriptLength;
    if(U_FAILURE(*err)) { return 0; }
    // Multi-script languages, equivalent to the LocaleScript data
    // that we used to load from locale resource bundles.
    /*length = */ uloc_getLanguage(locale, lang, UPRV_LENGTHOF(lang), &internalErrorCode);
    if(U_FAILURE(internalErrorCode) || internalErrorCode == U_STRING_NOT_TERMINATED_WARNING) {
        return 0;
    }
    if(0 == uprv_strcmp(lang, "ja")) {
        return setCodes(JAPANESE, UPRV_LENGTHOF(JAPANESE), scripts, capacity, err);
    }
    if(0 == uprv_strcmp(lang, "ko")) {
        return setCodes(KOREAN, UPRV_LENGTHOF(KOREAN), scripts, capacity, err);
    }
    scriptLength = uloc_getScript(locale, script, UPRV_LENGTHOF(script), &internalErrorCode);
    if(U_FAILURE(internalErrorCode) || internalErrorCode == U_STRING_NOT_TERMINATED_WARNING) {
        return 0;
    }
    if(0 == uprv_strcmp(lang, "zh") && 0 == uprv_strcmp(script, "Hant")) {
        return setCodes(HAN_BOPO, UPRV_LENGTHOF(HAN_BOPO), scripts, capacity, err);
    }
    // Explicit script code.
    if(scriptLength != 0) {
        UScriptCode scriptCode = (UScriptCode)u_getPropertyValueEnum(UCHAR_SCRIPT, script);
        if(scriptCode != USCRIPT_INVALID_CODE) {
            if(scriptCode == USCRIPT_SIMPLIFIED_HAN || scriptCode == USCRIPT_TRADITIONAL_HAN) {
                scriptCode = USCRIPT_HAN;
            }
            return setOneCode(scriptCode, scripts, capacity, err);
        }
    }
    return 0;
}

/* TODO: this is a bad API and should be deprecated, ticket #11141 */
U_CAPI int32_t  U_EXPORT2
uscript_getCode(const char* nameOrAbbrOrLocale,
                UScriptCode* fillIn,
                int32_t capacity,
                UErrorCode* err){
    UBool triedCode;
    UErrorCode internalErrorCode;
    int32_t length;

    if(U_FAILURE(*err)) {
        return 0;
    }
    if(nameOrAbbrOrLocale==NULL ||
            (fillIn == NULL ? capacity != 0 : capacity < 0)) {
        *err = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    triedCode = FALSE;
    if(uprv_strchr(nameOrAbbrOrLocale, '-')==NULL && uprv_strchr(nameOrAbbrOrLocale, '_')==NULL ){
        /* try long and abbreviated script names first */
        UScriptCode code = (UScriptCode) u_getPropertyValueEnum(UCHAR_SCRIPT, nameOrAbbrOrLocale);
        if(code!=USCRIPT_INVALID_CODE) {
            return setOneCode(code, fillIn, capacity, err);
        }
        triedCode = TRUE;
    }
    internalErrorCode = U_ZERO_ERROR;
    length = getCodesFromLocale(nameOrAbbrOrLocale, fillIn, capacity, err);
    if(U_FAILURE(*err) || length != 0) {
        return length;
    }
    icu::CharString likely;
    {
        icu::CharStringByteSink sink(&likely);
        ulocimp_addLikelySubtags(nameOrAbbrOrLocale, sink, &internalErrorCode);
    }
    if(U_SUCCESS(internalErrorCode) && internalErrorCode != U_STRING_NOT_TERMINATED_WARNING) {
        length = getCodesFromLocale(likely.data(), fillIn, capacity, err);
        if(U_FAILURE(*err) || length != 0) {
            return length;
        }
    }
    if(!triedCode) {
        /* still not found .. try long and abbreviated script names again */
        UScriptCode code = (UScriptCode) u_getPropertyValueEnum(UCHAR_SCRIPT, nameOrAbbrOrLocale);
        if(code!=USCRIPT_INVALID_CODE) {
            return setOneCode(code, fillIn, capacity, err);
        }
    }
    return 0;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2013-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  uscript_props.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2013feb16
*   created by: Markus W. Scherer
*/








namespace {

// Script metadata (script properties).
// See http://unicode.org/cldr/trac/browser/trunk/common/properties/scriptMetadata.txt

// 0 = NOT_ENCODED, no sample character, default false script properties.
// Bits 20.. 0: sample character

// Bits 23..21: usage
const int32_t up_UNKNOWN = 1 << 21;
const int32_t up_EXCLUSION = 2 << 21;
const int32_t up_LIMITED_USE = 3 << 21;
// st int32_t ASPIRATIONAL = 4 << 21; -- not used any more since Unicode 10
const int32_t up_RECOMMENDED = 5 << 21;

// Bits 31..24: Single-bit flags
const int32_t up_RTL = 1 << 24;
const int32_t up_LB_LETTERS = 1 << 25;
const int32_t up_CASED = 1 << 26;

const int32_t up_SCRIPT_PROPS[] = {
    // Begin copy-paste output from
    // tools/trunk/unicode/py/parsescriptmetadata.py
    0x0040 | up_RECOMMENDED,  // Zyyy
    0x0308 | up_RECOMMENDED,  // Zinh
    0x0628 | up_RECOMMENDED | up_RTL,  // Arab
    0x0531 | up_RECOMMENDED | up_CASED,  // Armn
    0x0995 | up_RECOMMENDED,  // Beng
    0x3105 | up_RECOMMENDED | up_LB_LETTERS,  // Bopo
    0x13C4 | up_LIMITED_USE | up_CASED,  // Cher
    0x03E2 | up_EXCLUSION | up_CASED,  // Copt
    0x042F | up_RECOMMENDED | up_CASED,  // Cyrl
    0x10414 | up_EXCLUSION | up_CASED,  // Dsrt
    0x0905 | up_RECOMMENDED,  // Deva
    0x12A0 | up_RECOMMENDED,  // Ethi
    0x10D3 | up_RECOMMENDED,  // Geor
    0x10330 | up_EXCLUSION,  // Goth
    0x03A9 | up_RECOMMENDED | up_CASED,  // Grek
    0x0A95 | up_RECOMMENDED,  // Gujr
    0x0A15 | up_RECOMMENDED,  // Guru
    0x5B57 | up_RECOMMENDED | up_LB_LETTERS,  // Hani
    0xAC00 | up_RECOMMENDED,  // Hang
    0x05D0 | up_RECOMMENDED | up_RTL,  // Hebr
    0x304B | up_RECOMMENDED | up_LB_LETTERS,  // Hira
    0x0C95 | up_RECOMMENDED,  // Knda
    0x30AB | up_RECOMMENDED | up_LB_LETTERS,  // Kana
    0x1780 | up_RECOMMENDED | up_LB_LETTERS,  // Khmr
    0x0EA5 | up_RECOMMENDED | up_LB_LETTERS,  // Laoo
    0x004C | up_RECOMMENDED | up_CASED,  // Latn
    0x0D15 | up_RECOMMENDED,  // Mlym
    0x1826 | up_EXCLUSION,  // Mong
    0x1000 | up_RECOMMENDED | up_LB_LETTERS,  // Mymr
    0x168F | up_EXCLUSION,  // Ogam
    0x10300 | up_EXCLUSION,  // Ital
    0x0B15 | up_RECOMMENDED,  // Orya
    0x16A0 | up_EXCLUSION,  // Runr
    0x0D85 | up_RECOMMENDED,  // Sinh
    0x0710 | up_LIMITED_USE | up_RTL,  // Syrc
    0x0B95 | up_RECOMMENDED,  // Taml
    0x0C15 | up_RECOMMENDED,  // Telu
    0x078C | up_RECOMMENDED | up_RTL,  // Thaa
    0x0E17 | up_RECOMMENDED | up_LB_LETTERS,  // Thai
    0x0F40 | up_RECOMMENDED,  // Tibt
    0x14C0 | up_LIMITED_USE,  // Cans
    0xA288 | up_LIMITED_USE | up_LB_LETTERS,  // Yiii
    0x1703 | up_EXCLUSION,  // Tglg
    0x1723 | up_EXCLUSION,  // Hano
    0x1743 | up_EXCLUSION,  // Buhd
    0x1763 | up_EXCLUSION,  // Tagb
    0x280E | up_UNKNOWN,  // Brai
    0x10800 | up_EXCLUSION | up_RTL,  // Cprt
    0x1900 | up_LIMITED_USE,  // Limb
    0x10000 | up_EXCLUSION,  // Linb
    0x10480 | up_EXCLUSION,  // Osma
    0x10450 | up_EXCLUSION,  // Shaw
    0x1950 | up_LIMITED_USE | up_LB_LETTERS,  // Tale
    0x10380 | up_EXCLUSION,  // Ugar
    0,
    0x1A00 | up_EXCLUSION,  // Bugi
    0x2C00 | up_EXCLUSION | up_CASED,  // Glag
    0x10A00 | up_EXCLUSION | up_RTL,  // Khar
    0xA800 | up_LIMITED_USE,  // Sylo
    0x1980 | up_LIMITED_USE | up_LB_LETTERS,  // Talu
    0x2D30 | up_LIMITED_USE,  // Tfng
    0x103A0 | up_EXCLUSION,  // Xpeo
    0x1B05 | up_LIMITED_USE,  // Bali
    0x1BC0 | up_LIMITED_USE,  // Batk
    0,
    0x11005 | up_EXCLUSION,  // Brah
    0xAA00 | up_LIMITED_USE,  // Cham
    0,
    0,
    0,
    0,
    0x13153 | up_EXCLUSION,  // Egyp
    0,
    0x5B57 | up_RECOMMENDED | up_LB_LETTERS,  // Hans
    0x5B57 | up_RECOMMENDED | up_LB_LETTERS,  // Hant
    0x16B1C | up_EXCLUSION,  // Hmng
    0x10CA1 | up_EXCLUSION | up_RTL | up_CASED,  // Hung
    0,
    0xA984 | up_LIMITED_USE,  // Java
    0xA90A | up_LIMITED_USE,  // Kali
    0,
    0,
    0x1C00 | up_LIMITED_USE,  // Lepc
    0x10647 | up_EXCLUSION,  // Lina
    0x0840 | up_LIMITED_USE | up_RTL,  // Mand
    0,
    0x10980 | up_EXCLUSION | up_RTL,  // Mero
    0x07CA | up_LIMITED_USE | up_RTL,  // Nkoo
    0x10C00 | up_EXCLUSION | up_RTL,  // Orkh
    0x1036B | up_EXCLUSION,  // Perm
    0xA840 | up_EXCLUSION,  // Phag
    0x10900 | up_EXCLUSION | up_RTL,  // Phnx
    0x16F00 | up_LIMITED_USE,  // Plrd
    0,
    0,
    0,
    0,
    0,
    0,
    0xA549 | up_LIMITED_USE,  // Vaii
    0,
    0x12000 | up_EXCLUSION,  // Xsux
    0,
    0xFDD0 | up_UNKNOWN,  // Zzzz
    0x102A0 | up_EXCLUSION,  // Cari
    0x304B | up_RECOMMENDED | up_LB_LETTERS,  // Jpan
    0x1A20 | up_LIMITED_USE | up_LB_LETTERS,  // Lana
    0x10280 | up_EXCLUSION,  // Lyci
    0x10920 | up_EXCLUSION | up_RTL,  // Lydi
    0x1C5A | up_LIMITED_USE,  // Olck
    0xA930 | up_EXCLUSION,  // Rjng
    0xA882 | up_LIMITED_USE,  // Saur
    0x1D850 | up_EXCLUSION,  // Sgnw
    0x1B83 | up_LIMITED_USE,  // Sund
    0,
    0xABC0 | up_LIMITED_USE,  // Mtei
    0x10840 | up_EXCLUSION | up_RTL,  // Armi
    0x10B00 | up_EXCLUSION | up_RTL,  // Avst
    0x11103 | up_LIMITED_USE,  // Cakm
    0xAC00 | up_RECOMMENDED,  // Kore
    0x11083 | up_EXCLUSION,  // Kthi
    0x10AD8 | up_EXCLUSION | up_RTL,  // Mani
    0x10B60 | up_EXCLUSION | up_RTL,  // Phli
    0x10B8F | up_EXCLUSION | up_RTL,  // Phlp
    0,
    0x10B40 | up_EXCLUSION | up_RTL,  // Prti
    0x0800 | up_EXCLUSION | up_RTL,  // Samr
    0xAA80 | up_LIMITED_USE | up_LB_LETTERS,  // Tavt
    0,
    0,
    0xA6A0 | up_LIMITED_USE,  // Bamu
    0xA4D0 | up_LIMITED_USE,  // Lisu
    0,
    0x10A60 | up_EXCLUSION | up_RTL,  // Sarb
    0x16AE6 | up_EXCLUSION,  // Bass
    0x1BC20 | up_EXCLUSION,  // Dupl
    0x10500 | up_EXCLUSION,  // Elba
    0x11315 | up_EXCLUSION,  // Gran
    0,
    0,
    0x1E802 | up_EXCLUSION | up_RTL,  // Mend
    0x109A0 | up_EXCLUSION | up_RTL,  // Merc
    0x10A95 | up_EXCLUSION | up_RTL,  // Narb
    0x10896 | up_EXCLUSION | up_RTL,  // Nbat
    0x10873 | up_EXCLUSION | up_RTL,  // Palm
    0x112BE | up_EXCLUSION,  // Sind
    0x118B4 | up_EXCLUSION | up_CASED,  // Wara
    0,
    0,
    0x16A4F | up_EXCLUSION,  // Mroo
    0x1B1C4 | up_EXCLUSION | up_LB_LETTERS,  // Nshu
    0x11183 | up_EXCLUSION,  // Shrd
    0x110D0 | up_EXCLUSION,  // Sora
    0x11680 | up_EXCLUSION,  // Takr
    0x18229 | up_EXCLUSION | up_LB_LETTERS,  // Tang
    0,
    0x14400 | up_EXCLUSION,  // Hluw
    0x11208 | up_EXCLUSION,  // Khoj
    0x11484 | up_EXCLUSION,  // Tirh
    0x10537 | up_EXCLUSION,  // Aghb
    0x11152 | up_EXCLUSION,  // Mahj
    0x11717 | up_EXCLUSION | up_LB_LETTERS,  // Ahom
    0x108F4 | up_EXCLUSION | up_RTL,  // Hatr
    0x1160E | up_EXCLUSION,  // Modi
    0x1128F | up_EXCLUSION,  // Mult
    0x11AC0 | up_EXCLUSION,  // Pauc
    0x1158E | up_EXCLUSION,  // Sidd
    0x1E909 | up_LIMITED_USE | up_RTL | up_CASED,  // Adlm
    0x11C0E | up_EXCLUSION,  // Bhks
    0x11C72 | up_EXCLUSION,  // Marc
    0x11412 | up_LIMITED_USE,  // Newa
    0x104B5 | up_LIMITED_USE | up_CASED,  // Osge
    0x5B57 | up_RECOMMENDED | up_LB_LETTERS,  // Hanb
    0x1112 | up_RECOMMENDED,  // Jamo
    0,
    0x11D10 | up_EXCLUSION,  // Gonm
    0x11A5C | up_EXCLUSION,  // Soyo
    0x11A0B | up_EXCLUSION,  // Zanb
    0x1180B | up_EXCLUSION,  // Dogr
    0x11D71 | up_LIMITED_USE,  // Gong
    0x11EE5 | up_EXCLUSION,  // Maka
    0x16E40 | up_EXCLUSION | up_CASED,  // Medf
    0x10D12 | up_LIMITED_USE | up_RTL,  // Rohg
    0x10F42 | up_EXCLUSION | up_RTL,  // Sogd
    0x10F19 | up_EXCLUSION | up_RTL,  // Sogo
    0x10FF1 | up_EXCLUSION | up_RTL,  // Elym
    0x1E108 | up_LIMITED_USE,  // Hmnp
    0x119CE | up_EXCLUSION,  // Nand
    0x1E2E1 | up_LIMITED_USE,  // Wcho
    0x10FBF | up_EXCLUSION | up_RTL,  // Chrs
    0x1190C | up_EXCLUSION,  // Diak
    0x18C65 | up_EXCLUSION | up_LB_LETTERS,  // Kits
    0x10E88 | up_EXCLUSION | up_RTL,  // Yezi
    // End copy-paste from parsescriptmetadata.py
};

int32_t getScriptProps(UScriptCode script) {
    if (0 <= script && script < UPRV_LENGTHOF(up_SCRIPT_PROPS)) {
        return up_SCRIPT_PROPS[script];
    } else {
        return 0;
    }
}

}  // namespace

U_CAPI int32_t U_EXPORT2
uscript_getSampleString(UScriptCode script, UChar *dest, int32_t capacity, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) { return 0; }
    if(capacity < 0 || (capacity > 0 && dest == NULL)) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    int32_t sampleChar = getScriptProps(script) & 0x1fffff;
    int32_t length;
    if(sampleChar == 0) {
        length = 0;
    } else {
        length = U16_LENGTH(sampleChar);
        if(length <= capacity) {
            int32_t i = 0;
            U16_APPEND_UNSAFE(dest, i, sampleChar);
        }
    }
    return u_terminateUChars(dest, capacity, length, pErrorCode);
}

U_COMMON_API icu::UnicodeString U_EXPORT2
uscript_getSampleUnicodeString(UScriptCode script) {
    icu::UnicodeString sample;
    int32_t sampleChar = getScriptProps(script) & 0x1fffff;
    if(sampleChar != 0) {
        sample.append(sampleChar);
    }
    return sample;
}

U_CAPI UScriptUsage U_EXPORT2
uscript_getUsage(UScriptCode script) {
    return (UScriptUsage)((getScriptProps(script) >> 21) & 7);
}

U_CAPI UBool U_EXPORT2
uscript_isRightToLeft(UScriptCode script) {
    return (getScriptProps(script) & up_RTL) != 0;
}

U_CAPI UBool U_EXPORT2
uscript_breaksBetweenLetters(UScriptCode script) {
    return (getScriptProps(script) & up_LB_LETTERS) != 0;
}

U_CAPI UBool U_EXPORT2
uscript_isCased(UScriptCode script) {
    return (getScriptProps(script) & up_CASED) != 0;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2002-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uset.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2002mar07
*   created by: Markus W. Scherer
*
*   There are functions to efficiently serialize a USet into an array of uint16_t
*   and functions to use such a serialized form efficiently without
*   instantiating a new USet.
*/









U_NAMESPACE_USE

U_CAPI USet* U_EXPORT2
uset_openEmpty() {
    return (USet*) new UnicodeSet();
}

U_CAPI USet* U_EXPORT2
uset_open(UChar32 start, UChar32 end) {
    return (USet*) new UnicodeSet(start, end);
}

U_CAPI void U_EXPORT2
uset_close(USet* set) {
    delete (UnicodeSet*) set;
}

U_CAPI USet * U_EXPORT2
uset_clone(const USet *set) {
    return (USet*) (((UnicodeSet*) set)->UnicodeSet::clone());
}

U_CAPI UBool U_EXPORT2
uset_isFrozen(const USet *set) {
    return ((UnicodeSet*) set)->UnicodeSet::isFrozen();
}

U_CAPI void U_EXPORT2
uset_freeze(USet *set) {
    ((UnicodeSet*) set)->UnicodeSet::freeze();
}

U_CAPI USet * U_EXPORT2
uset_cloneAsThawed(const USet *set) {
    return (USet*) (((UnicodeSet*) set)->UnicodeSet::cloneAsThawed());
}

U_CAPI void U_EXPORT2
uset_set(USet* set,
     UChar32 start, UChar32 end) {
    ((UnicodeSet*) set)->UnicodeSet::set(start, end);
}

U_CAPI void U_EXPORT2
uset_addAll(USet* set, const USet *additionalSet) {
    ((UnicodeSet*) set)->UnicodeSet::addAll(*((const UnicodeSet*)additionalSet));
}

U_CAPI void U_EXPORT2
uset_add(USet* set, UChar32 c) {
    ((UnicodeSet*) set)->UnicodeSet::add(c);
}

U_CAPI void U_EXPORT2
uset_addRange(USet* set, UChar32 start, UChar32 end) {
    ((UnicodeSet*) set)->UnicodeSet::add(start, end);    
}

U_CAPI void U_EXPORT2
uset_addString(USet* set, const UChar* str, int32_t strLen) {
    // UnicodeString handles -1 for strLen
    UnicodeString s(strLen<0, str, strLen);
    ((UnicodeSet*) set)->UnicodeSet::add(s);
}

U_CAPI void U_EXPORT2
uset_addAllCodePoints(USet* set, const UChar *str, int32_t strLen) {
    // UnicodeString handles -1 for strLen
    UnicodeString s(str, strLen);
    ((UnicodeSet*) set)->UnicodeSet::addAll(s);
}

U_CAPI void U_EXPORT2
uset_remove(USet* set, UChar32 c) {
    ((UnicodeSet*) set)->UnicodeSet::remove(c);
}

U_CAPI void U_EXPORT2
uset_removeRange(USet* set, UChar32 start, UChar32 end) {
    ((UnicodeSet*) set)->UnicodeSet::remove(start, end);
}

U_CAPI void U_EXPORT2
uset_removeString(USet* set, const UChar* str, int32_t strLen) {
    UnicodeString s(strLen==-1, str, strLen);
    ((UnicodeSet*) set)->UnicodeSet::remove(s);
}

U_CAPI void U_EXPORT2
uset_removeAll(USet* set, const USet* remove) {
    ((UnicodeSet*) set)->UnicodeSet::removeAll(*(const UnicodeSet*)remove);
}

U_CAPI void U_EXPORT2
uset_retain(USet* set, UChar32 start, UChar32 end) {
    ((UnicodeSet*) set)->UnicodeSet::retain(start, end);
}

U_CAPI void U_EXPORT2
uset_retainAll(USet* set, const USet* retain) {
    ((UnicodeSet*) set)->UnicodeSet::retainAll(*(const UnicodeSet*)retain);
}

U_CAPI void U_EXPORT2
uset_compact(USet* set) {
    ((UnicodeSet*) set)->UnicodeSet::compact();
}

U_CAPI void U_EXPORT2
uset_complement(USet* set) {
    ((UnicodeSet*) set)->UnicodeSet::complement();
}

U_CAPI void U_EXPORT2
uset_complementAll(USet* set, const USet* complement) {
    ((UnicodeSet*) set)->UnicodeSet::complementAll(*(const UnicodeSet*)complement);
}

U_CAPI void U_EXPORT2
uset_clear(USet* set) {
    ((UnicodeSet*) set)->UnicodeSet::clear();
}

U_CAPI void U_EXPORT2
uset_removeAllStrings(USet* set) {
    ((UnicodeSet*) set)->UnicodeSet::removeAllStrings();
}

U_CAPI UBool U_EXPORT2
uset_isEmpty(const USet* set) {
    return ((const UnicodeSet*) set)->UnicodeSet::isEmpty();
}

U_CAPI UBool U_EXPORT2
uset_contains(const USet* set, UChar32 c) {
    return ((const UnicodeSet*) set)->UnicodeSet::contains(c);
}

U_CAPI UBool U_EXPORT2
uset_containsRange(const USet* set, UChar32 start, UChar32 end) {
    return ((const UnicodeSet*) set)->UnicodeSet::contains(start, end);
}

U_CAPI UBool U_EXPORT2
uset_containsString(const USet* set, const UChar* str, int32_t strLen) {
    UnicodeString s(strLen==-1, str, strLen);
    return ((const UnicodeSet*) set)->UnicodeSet::contains(s);
}

U_CAPI UBool U_EXPORT2
uset_containsAll(const USet* set1, const USet* set2) {
    return ((const UnicodeSet*) set1)->UnicodeSet::containsAll(* (const UnicodeSet*) set2);
}

U_CAPI UBool U_EXPORT2
uset_containsAllCodePoints(const USet* set, const UChar *str, int32_t strLen) {
    // Create a string alias, since nothing is being added to the set.
    UnicodeString s(strLen==-1, str, strLen);
    return ((const UnicodeSet*) set)->UnicodeSet::containsAll(s);
}

U_CAPI UBool U_EXPORT2
uset_containsNone(const USet* set1, const USet* set2) {
    return ((const UnicodeSet*) set1)->UnicodeSet::containsNone(* (const UnicodeSet*) set2);
}

U_CAPI UBool U_EXPORT2
uset_containsSome(const USet* set1, const USet* set2) {
    return ((const UnicodeSet*) set1)->UnicodeSet::containsSome(* (const UnicodeSet*) set2);
}

U_CAPI int32_t U_EXPORT2
uset_span(const USet *set, const UChar *s, int32_t length, USetSpanCondition spanCondition) {
    return ((UnicodeSet*) set)->UnicodeSet::span(s, length, spanCondition);
}

U_CAPI int32_t U_EXPORT2
uset_spanBack(const USet *set, const UChar *s, int32_t length, USetSpanCondition spanCondition) {
    return ((UnicodeSet*) set)->UnicodeSet::spanBack(s, length, spanCondition);
}

U_CAPI int32_t U_EXPORT2
uset_spanUTF8(const USet *set, const char *s, int32_t length, USetSpanCondition spanCondition) {
    return ((UnicodeSet*) set)->UnicodeSet::spanUTF8(s, length, spanCondition);
}

U_CAPI int32_t U_EXPORT2
uset_spanBackUTF8(const USet *set, const char *s, int32_t length, USetSpanCondition spanCondition) {
    return ((UnicodeSet*) set)->UnicodeSet::spanBackUTF8(s, length, spanCondition);
}

U_CAPI UBool U_EXPORT2
uset_equals(const USet* set1, const USet* set2) {
    return *(const UnicodeSet*)set1 == *(const UnicodeSet*)set2;
}

U_CAPI int32_t U_EXPORT2
uset_indexOf(const USet* set, UChar32 c) {
    return ((UnicodeSet*) set)->UnicodeSet::indexOf(c);
}

U_CAPI UChar32 U_EXPORT2
uset_charAt(const USet* set, int32_t index) {
    return ((UnicodeSet*) set)->UnicodeSet::charAt(index);
}

U_CAPI int32_t U_EXPORT2
uset_size(const USet* set) {
    return ((const UnicodeSet*) set)->UnicodeSet::size();
}

U_NAMESPACE_BEGIN
/**
 * This class only exists to provide access to the UnicodeSet private
 * USet support API.  Declaring a class a friend is more portable than
 * trying to declare extern "C" functions as friends.
 */
class USetAccess /* not : public UObject because all methods are static */ {
public:
    /* Try to have the compiler inline these*/
    inline static int32_t getStringCount(const UnicodeSet& set) {
        return set.stringsSize();
    }
    inline static const UnicodeString* getString(const UnicodeSet& set,
                                                 int32_t i) {
        return set.getString(i);
    }
private:
    /* do not instantiate*/
    USetAccess();
};
U_NAMESPACE_END

U_CAPI int32_t U_EXPORT2
uset_getItemCount(const USet* uset) {
    const UnicodeSet& set = *(const UnicodeSet*)uset;
    return set.getRangeCount() + USetAccess::getStringCount(set);
}

U_CAPI int32_t U_EXPORT2
uset_getItem(const USet* uset, int32_t itemIndex,
             UChar32* start, UChar32* end,
             UChar* str, int32_t strCapacity,
             UErrorCode* ec) {
    if (U_FAILURE(*ec)) return 0;
    const UnicodeSet& set = *(const UnicodeSet*)uset;
    int32_t rangeCount;

    if (itemIndex < 0) {
        *ec = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    } else if (itemIndex < (rangeCount = set.getRangeCount())) {
        *start = set.getRangeStart(itemIndex);
        *end = set.getRangeEnd(itemIndex);
        return 0;
    } else {
        itemIndex -= rangeCount;
        if (itemIndex < USetAccess::getStringCount(set)) {
            const UnicodeString* s = USetAccess::getString(set, itemIndex);
            return s->extract(str, strCapacity, *ec);
        } else {
            *ec = U_INDEX_OUTOFBOUNDS_ERROR;
            return -1;
        }
    }
}

//U_CAPI int32_t U_EXPORT2
//uset_getRangeCount(const USet* set) {
//    return ((const UnicodeSet*) set)->getRangeCount();
//}
//
//U_CAPI UBool U_EXPORT2
//uset_getRange(const USet* set, int32_t rangeIndex,
//              UChar32* pStart, UChar32* pEnd) {
//    if ((uint32_t) rangeIndex >= (uint32_t) uset_getRangeCount(set)) {
//        return FALSE;
//    }
//    const UnicodeSet* us = (const UnicodeSet*) set;
//    *pStart = us->getRangeStart(rangeIndex);
//    *pEnd = us->getRangeEnd(rangeIndex);
//    return TRUE;
//}

/*
 * Serialize a USet into 16-bit units.
 * Store BMP code points as themselves with one 16-bit unit each.
 *
 * Important: the code points in the array are in ascending order,
 * therefore all BMP code points precede all supplementary code points.
 *
 * Store each supplementary code point in 2 16-bit units,
 * simply with higher-then-lower 16-bit halfs.
 *
 * Precede the entire list with the length.
 * If there are supplementary code points, then set bit 15 in the length
 * and add the bmpLength between it and the array.
 *
 * In other words:
 * - all BMP:            (length=bmpLength) BMP, .., BMP
 * - some supplementary: (length|0x8000) (bmpLength<length) BMP, .., BMP, supp-high, supp-low, ..
 */
U_CAPI int32_t U_EXPORT2
uset_serialize(const USet* set, uint16_t* dest, int32_t destCapacity, UErrorCode* ec) {
    if (ec==NULL || U_FAILURE(*ec)) {
        return 0;
    }

    return ((const UnicodeSet*) set)->UnicodeSet::serialize(dest, destCapacity,* ec);
}

U_CAPI UBool U_EXPORT2
uset_getSerializedSet(USerializedSet* fillSet, const uint16_t* src, int32_t srcLength) {
    int32_t length;

    if(fillSet==NULL) {
        return FALSE;
    }
    if(src==NULL || srcLength<=0) {
        fillSet->length=fillSet->bmpLength=0;
        return FALSE;
    }

    length=*src++;
    if(length&0x8000) {
        /* there are supplementary values */
        length&=0x7fff;
        if(srcLength<(2+length)) {
            fillSet->length=fillSet->bmpLength=0;
            return FALSE;
        }
        fillSet->bmpLength=*src++;
    } else {
        /* only BMP values */
        if(srcLength<(1+length)) {
            fillSet->length=fillSet->bmpLength=0;
            return FALSE;
        }
        fillSet->bmpLength=length;
    }
    fillSet->array=src;
    fillSet->length=length;
    return TRUE;
}

U_CAPI void U_EXPORT2
uset_setSerializedToOne(USerializedSet* fillSet, UChar32 c) {
    if(fillSet==NULL || (uint32_t)c>0x10ffff) {
        return;
    }

    fillSet->array=fillSet->staticArray;
    if(c<0xffff) {
        fillSet->bmpLength=fillSet->length=2;
        fillSet->staticArray[0]=(uint16_t)c;
        fillSet->staticArray[1]=(uint16_t)c+1;
    } else if(c==0xffff) {
        fillSet->bmpLength=1;
        fillSet->length=3;
        fillSet->staticArray[0]=0xffff;
        fillSet->staticArray[1]=1;
        fillSet->staticArray[2]=0;
    } else if(c<0x10ffff) {
        fillSet->bmpLength=0;
        fillSet->length=4;
        fillSet->staticArray[0]=(uint16_t)(c>>16);
        fillSet->staticArray[1]=(uint16_t)c;
        ++c;
        fillSet->staticArray[2]=(uint16_t)(c>>16);
        fillSet->staticArray[3]=(uint16_t)c;
    } else /* c==0x10ffff */ {
        fillSet->bmpLength=0;
        fillSet->length=2;
        fillSet->staticArray[0]=0x10;
        fillSet->staticArray[1]=0xffff;
    }
}

U_CAPI UBool U_EXPORT2
uset_serializedContains(const USerializedSet* set, UChar32 c) {
    const uint16_t* array;

    if(set==NULL || (uint32_t)c>0x10ffff) {
        return FALSE;
    }

    array=set->array;
    if(c<=0xffff) {
        /* find c in the BMP part */
        int32_t lo = 0;
        int32_t hi = set->bmpLength-1;
        if (c < array[0]) {
            hi = 0;
        } else if (c < array[hi]) {
            for(;;) {
                int32_t i = (lo + hi) >> 1;
                if (i == lo) {
                    break;  // Done!
                } else if (c < array[i]) {
                    hi = i;
                } else {
                    lo = i;
                }
            }
        } else {
            hi += 1;
        }
        return (UBool)(hi&1);
    } else {
        /* find c in the supplementary part */
        uint16_t high=(uint16_t)(c>>16), low=(uint16_t)c;
        int32_t base = set->bmpLength;
        int32_t lo = 0;
        int32_t hi = set->length - 2 - base;
        if (high < array[base] || (high==array[base] && low<array[base+1])) {
            hi = 0;
        } else if (high < array[base+hi] || (high==array[base+hi] && low<array[base+hi+1])) {
            for (;;) {
                int32_t i = ((lo + hi) >> 1) & ~1;  // Guarantee even result
                int32_t iabs = i + base;
                if (i == lo) {
                    break;  // Done!
                } else if (high < array[iabs] || (high==array[iabs] && low<array[iabs+1])) {
                    hi = i;
                } else {
                    lo = i;
                }
            }
        } else {
            hi += 2;
        }
        /* count pairs of 16-bit units even per BMP and check if the number of pairs is odd */
        return (UBool)(((hi+(base<<1))&2)!=0);
    }
}

U_CAPI int32_t U_EXPORT2
uset_getSerializedRangeCount(const USerializedSet* set) {
    if(set==NULL) {
        return 0;
    }

    return (set->bmpLength+(set->length-set->bmpLength)/2+1)/2;
}

U_CAPI UBool U_EXPORT2
uset_getSerializedRange(const USerializedSet* set, int32_t rangeIndex,
                        UChar32* pStart, UChar32* pEnd) {
    const uint16_t* array;
    int32_t bmpLength, length;

    if(set==NULL || rangeIndex<0 || pStart==NULL || pEnd==NULL) {
        return FALSE;
    }

    array=set->array;
    length=set->length;
    bmpLength=set->bmpLength;

    rangeIndex*=2; /* address start/limit pairs */
    if(rangeIndex<bmpLength) {
        *pStart=array[rangeIndex++];
        if(rangeIndex<bmpLength) {
            *pEnd=array[rangeIndex]-1;
        } else if(rangeIndex<length) {
            *pEnd=((((int32_t)array[rangeIndex])<<16)|array[rangeIndex+1])-1;
        } else {
            *pEnd=0x10ffff;
        }
        return TRUE;
    } else {
        rangeIndex-=bmpLength;
        rangeIndex*=2; /* address pairs of pairs of units */
        length-=bmpLength;
        if(rangeIndex<length) {
            array+=bmpLength;
            *pStart=(((int32_t)array[rangeIndex])<<16)|array[rangeIndex+1];
            rangeIndex+=2;
            if(rangeIndex<length) {
                *pEnd=((((int32_t)array[rangeIndex])<<16)|array[rangeIndex+1])-1;
            } else {
                *pEnd=0x10ffff;
            }
            return TRUE;
        } else {
            return FALSE;
        }
    }
}

// TODO The old, internal uset.c had an efficient uset_containsOne function.
// Returned the one and only code point, or else -1 or something.
// Consider adding such a function to both C and C++ UnicodeSet/uset.
// See tools/gennorm/store.c for usage, now usetContainsOne there.

// TODO Investigate incorporating this code into UnicodeSet to improve
// efficiency.
// ---
// #define USET_GROW_DELTA 20
// 
// static int32_t
// findChar(const UChar32* array, int32_t length, UChar32 c) {
//     int32_t i;
// 
//     /* check the last range limit first for more efficient appending */
//     if(length>0) {
//         if(c>=array[length-1]) {
//             return length;
//         }
// 
//         /* do not check the last range limit again in the loop below */
//         --length;
//     }
// 
//     for(i=0; i<length && c>=array[i]; ++i) {}
//     return i;
// }
// 
// static UBool
// addRemove(USet* set, UChar32 c, int32_t doRemove) {
//     int32_t i, length, more;
// 
//     if(set==NULL || (uint32_t)c>0x10ffff) {
//         return FALSE;
//     }
// 
//     length=set->length;
//     i=findChar(set->array, length, c);
//     if((i&1)^doRemove) {
//         /* c is already in the set */
//         return TRUE;
//     }
// 
//     /* how many more array items do we need? */
//     if(i<length && (c+1)==set->array[i]) {
//         /* c is just before the following range, extend that in-place by one */
//         set->array[i]=c;
//         if(i>0) {
//             --i;
//             if(c==set->array[i]) {
//                 /* the previous range collapsed, remove it */
//                 set->length=length-=2;
//                 if(i<length) {
//                     uprv_memmove(set->array+i, set->array+i+2, (length-i)*4);
//                 }
//             }
//         }
//         return TRUE;
//     } else if(i>0 && c==set->array[i-1]) {
//         /* c is just after the previous range, extend that in-place by one */
//         if(++c<=0x10ffff) {
//             set->array[i-1]=c;
//             if(i<length && c==set->array[i]) {
//                 /* the following range collapsed, remove it */
//                 --i;
//                 set->length=length-=2;
//                 if(i<length) {
//                     uprv_memmove(set->array+i, set->array+i+2, (length-i)*4);
//                 }
//             }
//         } else {
//             /* extend the previous range (had limit 0x10ffff) to the end of Unicode */
//             set->length=i-1;
//         }
//         return TRUE;
//     } else if(i==length && c==0x10ffff) {
//         /* insert one range limit c */
//         more=1;
//     } else {
//         /* insert two range limits c, c+1 */
//         more=2;
//     }
// 
//     /* insert <more> range limits */
//     if(length+more>set->capacity) {
//         /* reallocate */
//         int32_t newCapacity=set->capacity+set->capacity/2+USET_GROW_DELTA;
//         UChar32* newArray=(UChar32* )uprv_malloc(newCapacity*4);
//         if(newArray==NULL) {
//             return FALSE;
//         }
//         set->capacity=newCapacity;
//         uprv_memcpy(newArray, set->array, length*4);
// 
//         if(set->array!=set->staticBuffer) {
//             uprv_free(set->array);
//         }
//         set->array=newArray;
//     }
// 
//     if(i<length) {
//         uprv_memmove(set->array+i+more, set->array+i, (length-i)*4);
//     }
//     set->array[i]=c;
//     if(more==2) {
//         set->array[i+1]=c+1;
//     }
//     set->length+=more;
// 
//     return TRUE;
// }
// 
// U_CAPI UBool U_EXPORT2
// uset_add(USet* set, UChar32 c) {
//     return addRemove(set, c, 0);
// }
// 
// U_CAPI void U_EXPORT2
// uset_remove(USet* set, UChar32 c) {
//     addRemove(set, c, 1);
// }
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2002-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  uset_props.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004aug30
*   created by: Markus W. Scherer
*
*   C wrappers around UnicodeSet functions that are implemented in
*   uniset_props.cpp, split off for modularization.
*/









U_NAMESPACE_USE

U_CAPI USet* U_EXPORT2
uset_openPattern(const UChar* pattern, int32_t patternLength,
                 UErrorCode* ec)
{
    UnicodeString pat(patternLength==-1, pattern, patternLength);
    UnicodeSet* set = new UnicodeSet(pat, *ec);
    /* test for NULL */
    if(set == 0) {
        *ec = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }

    if (U_FAILURE(*ec)) {
        delete set;
        set = NULL;
    }
    return (USet*) set;
}

U_CAPI USet* U_EXPORT2
uset_openPatternOptions(const UChar* pattern, int32_t patternLength,
                 uint32_t options,
                 UErrorCode* ec)
{
    UnicodeString pat(patternLength==-1, pattern, patternLength);
    UnicodeSet* set = new UnicodeSet(pat, options, NULL, *ec);
    /* test for NULL */
    if(set == 0) {
        *ec = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }

    if (U_FAILURE(*ec)) {
        delete set;
        set = NULL;
    }
    return (USet*) set;
}


U_CAPI int32_t U_EXPORT2 
uset_applyPattern(USet *set,
                  const UChar *pattern, int32_t patternLength,
                  uint32_t options,
                  UErrorCode *status){

    // status code needs to be checked since we 
    // dereference it
    if(status == NULL || U_FAILURE(*status)){
        return 0;
    }

    // check only the set paramenter
    // if pattern is NULL or null terminate
    // UnicodeString constructor takes care of it
    if(set == NULL){
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    UnicodeString pat(pattern, patternLength);

    ParsePosition pos;
   
    ((UnicodeSet*) set)->applyPattern(pat, pos, options, NULL, *status);
    
    return pos.getIndex();
}

U_CAPI void U_EXPORT2
uset_applyIntPropertyValue(USet* set,
               UProperty prop, int32_t value, UErrorCode* ec) {
    ((UnicodeSet*) set)->applyIntPropertyValue(prop, value, *ec);
}

U_CAPI void U_EXPORT2
uset_applyPropertyAlias(USet* set,
                        const UChar *prop, int32_t propLength,
                        const UChar *value, int32_t valueLength,
            UErrorCode* ec) {

    UnicodeString p(prop, propLength);
    UnicodeString v(value, valueLength);

    ((UnicodeSet*) set)->applyPropertyAlias(p, v, *ec);
}

U_CAPI UBool U_EXPORT2
uset_resemblesPattern(const UChar *pattern, int32_t patternLength,
                      int32_t pos) {

    UnicodeString pat(pattern, patternLength);

    return ((pos+1) < pat.length() &&
            pat.charAt(pos) == (UChar)91/*[*/) ||
            UnicodeSet::resemblesPattern(pat, pos);
}

U_CAPI int32_t U_EXPORT2
uset_toPattern(const USet* set,
               UChar* result, int32_t resultCapacity,
               UBool escapeUnprintable,
               UErrorCode* ec) {
    UnicodeString pat;
    ((const UnicodeSet*) set)->toPattern(pat, escapeUnprintable);
    return pat.extract(result, resultCapacity, *ec);
}

U_CAPI void U_EXPORT2
uset_closeOver(USet* set, int32_t attributes) {
    ((UnicodeSet*) set)->UnicodeSet::closeOver(attributes);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2002-2006, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
*/





U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(UnicodeSetIterator)

/**
 * Create an iterator
 * @param set set to iterate over
 */
UnicodeSetIterator::UnicodeSetIterator(const UnicodeSet& uSet) {
    cpString  = NULL;
    reset(uSet);
}

/**
 * Create an iterator. Convenience for when the contents are to be set later.
 */
UnicodeSetIterator::UnicodeSetIterator() {
    this->set = NULL;
    cpString  = NULL;
    reset();
}

UnicodeSetIterator::~UnicodeSetIterator() {
    delete cpString;
}

/**
 * Returns the next element in the set.
 * @return true if there was another element in the set.
 * if so, if codepoint == IS_STRING, the value is a string in the string field
 * else the value is a single code point in the codepoint field.
 * <br>You are guaranteed that the codepoints are in sorted order, and the strings are in sorted order,
 * and that all code points are returned before any strings are returned.
 * <br>Note also that the codepointEnd is undefined after calling this method.
 */
UBool UnicodeSetIterator::next() {
    if (nextElement <= endElement) {
        codepoint = codepointEnd = nextElement++;
        string = NULL;
        return TRUE;
    }
    if (range < endRange) {
        loadRange(++range);
        codepoint = codepointEnd = nextElement++;
        string = NULL;
        return TRUE;
    }

    if (nextString >= stringCount) return FALSE;
    codepoint = (UChar32)IS_STRING; // signal that value is actually a string
    string = (const UnicodeString*) set->strings->elementAt(nextString++);
    return TRUE;
}

/**
 * @return true if there was another element in the set.
 * if so, if codepoint == IS_STRING, the value is a string in the string field
 * else the value is a range of codepoints in the <codepoint, codepointEnd> fields.
 * <br>Note that the codepoints are in sorted order, and the strings are in sorted order,
 * and that all code points are returned before any strings are returned.
 * <br>You are guaranteed that the ranges are in sorted order, and the strings are in sorted order,
 * and that all ranges are returned before any strings are returned.
 * <br>You are also guaranteed that ranges are disjoint and non-contiguous.
 * <br>Note also that the codepointEnd is undefined after calling this method.
 */
UBool UnicodeSetIterator::nextRange() {
    string = NULL;
    if (nextElement <= endElement) {
        codepointEnd = endElement;
        codepoint = nextElement;
        nextElement = endElement+1;
        return TRUE;
    }
    if (range < endRange) {
        loadRange(++range);
        codepointEnd = endElement;
        codepoint = nextElement;
        nextElement = endElement+1;
        return TRUE;
    }

    if (nextString >= stringCount) return FALSE;
    codepoint = (UChar32)IS_STRING; // signal that value is actually a string
    string = (const UnicodeString*) set->strings->elementAt(nextString++);
    return TRUE;
}

/**
 *@param set the set to iterate over. This allows reuse of the iterator.
 */
void UnicodeSetIterator::reset(const UnicodeSet& uSet) {
    this->set = &uSet;
    reset();
}

/**
 * Resets to the start, to allow the iteration to start over again.
 */
void UnicodeSetIterator::reset() {
    if (set == NULL) {
        // Set up indices to empty iteration
        endRange = -1;
        stringCount = 0;
    } else {
        endRange = set->getRangeCount() - 1;
        stringCount = set->stringsSize();
    }
    range = 0;
    endElement = -1;
    nextElement = 0;            
    if (endRange >= 0) {
        loadRange(range);
    }
    nextString = 0;
    string = NULL;
}

void UnicodeSetIterator::loadRange(int32_t iRange) {
    nextElement = set->getRangeStart(iRange);
    endElement = set->getRangeEnd(iRange);
}


const UnicodeString& UnicodeSetIterator::getString()  {
    if (string==NULL && codepoint!=(UChar32)IS_STRING) {
       if (cpString == NULL) {
          cpString = new UnicodeString();
       }
       if (cpString != NULL) {
          cpString->setTo((UChar32)codepoint);
       }
       string = cpString;
    }
    return *string;
}

U_NAMESPACE_END

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ******************************************************************************
 *
 *   Copyright (C) 2000-2016, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 *
 ******************************************************************************
 *   file name:  ushape.cpp
 *   encoding:   UTF-8
 *   tab size:   8 (not used)
 *   indentation:4
 *
 *   created on: 2000jun29
 *   created by: Markus W. Scherer
 *
 *   Arabic letter shaping implemented by Ayman Roshdy
 */











/*
 * This implementation is designed for 16-bit Unicode strings.
 * The main assumption is that the Arabic characters and their
 * presentation forms each fit into a single UChar.
 * With UTF-8, they occupy 2 or 3 bytes, and more than the ASCII
 * characters.
 */

/*
 * ### TODO in general for letter shaping:
 * - the letter shaping code is UTF-16-unaware; needs update
 *   + especially invertBuffer()?!
 * - needs to handle the "Arabic Tail" that is used in some legacy codepages
 *   as a glyph fragment of wide-glyph letters
 *   + IBM Unicode conversion tables map it to U+200B (ZWSP)
 *   + IBM Egypt has proposed to encode the tail in Unicode among Arabic Presentation Forms
 *   + Unicode 3.2 added U+FE73 ARABIC TAIL FRAGMENT
 */

/* definitions for Arabic letter shaping ------------------------------------ */

#define USHAPE_IRRELEVANT 4
#define USHAPE_LAMTYPE    16
#define USHAPE_ALEFTYPE   32
#define USHAPE_LINKR      1
#define USHAPE_LINKL      2
#define USHAPE_APRESENT   8
#define USHAPE_SHADDA     64
#define USHAPE_CSHADDA    128
#define USHAPE_COMBINE    (USHAPE_SHADDA+USHAPE_CSHADDA)

#define USHAPE_HAMZAFE_CHAR       0xfe80
#define USHAPE_HAMZA06_CHAR       0x0621
#define USHAPE_YEH_HAMZA_CHAR     0x0626
#define USHAPE_YEH_HAMZAFE_CHAR   0xFE89
#define USHAPE_LAMALEF_SPACE_SUB  0xFFFF
#define USHAPE_TASHKEEL_SPACE_SUB 0xFFFE
#define USHAPE_NEW_TAIL_CHAR      0xFE73
#define USHAPE_OLD_TAIL_CHAR      0x200B
#define USHAPE_LAM_CHAR           0x0644
#define USHAPE_SPACE_CHAR         0x0020
#define USHAPE_SHADDA_CHAR        0xFE7C
#define USHAPE_TATWEEL_CHAR       0x0640
#define USHAPE_SHADDA_USHAPE_TATWEEL_CHAR  0xFE7D
#define USHAPE_SHADDA06_CHAR      0x0651

#define USHAPE_SHAPE_MODE   0
#define USHAPE_DESHAPE_MODE 1

struct uShapeVariables {
     UChar tailChar;
     uint32_t uShapeLamalefBegin;
     uint32_t uShapeLamalefEnd;
     uint32_t uShapeTashkeelBegin;
     uint32_t uShapeTashkeelEnd;
     int spacesRelativeToTextBeginEnd;
};

static const uint8_t tailFamilyIsolatedFinal[] = {
    /* FEB1 */ 1,
    /* FEB2 */ 1,
    /* FEB3 */ 0,
    /* FEB4 */ 0,
    /* FEB5 */ 1,
    /* FEB6 */ 1,
    /* FEB7 */ 0,
    /* FEB8 */ 0,
    /* FEB9 */ 1,
    /* FEBA */ 1,
    /* FEBB */ 0,
    /* FEBC */ 0,
    /* FEBD */ 1,
    /* FEBE */ 1
};

static const uint8_t tashkeelMedial[] = {
    /* FE70 */ 0,
    /* FE71 */ 1,
    /* FE72 */ 0,
    /* FE73 */ 0,
    /* FE74 */ 0,
    /* FE75 */ 0,
    /* FE76 */ 0,
    /* FE77 */ 1,
    /* FE78 */ 0,
    /* FE79 */ 1,
    /* FE7A */ 0,
    /* FE7B */ 1,
    /* FE7C */ 0,
    /* FE7D */ 1,
    /* FE7E */ 0,
    /* FE7F */ 1
};

static const UChar yehHamzaToYeh[] =
{
/* isolated*/ 0xFEEF,
/* final   */ 0xFEF0
};

static const uint8_t IrrelevantPos[] = {
    0x0, 0x2, 0x4, 0x6,
    0x8, 0xA, 0xC, 0xE
};


static const UChar convertLamAlef[] =
{
/*FEF5*/    0x0622,
/*FEF6*/    0x0622,
/*FEF7*/    0x0623,
/*FEF8*/    0x0623,
/*FEF9*/    0x0625,
/*FEFA*/    0x0625,
/*FEFB*/    0x0627,
/*FEFC*/    0x0627
};

static const UChar araLink[178]=
{
  1           + 32 + 256 * 0x11,/*0x0622*/
  1           + 32 + 256 * 0x13,/*0x0623*/
  1                + 256 * 0x15,/*0x0624*/
  1           + 32 + 256 * 0x17,/*0x0625*/
  1 + 2            + 256 * 0x19,/*0x0626*/
  1           + 32 + 256 * 0x1D,/*0x0627*/
  1 + 2            + 256 * 0x1F,/*0x0628*/
  1                + 256 * 0x23,/*0x0629*/
  1 + 2            + 256 * 0x25,/*0x062A*/
  1 + 2            + 256 * 0x29,/*0x062B*/
  1 + 2            + 256 * 0x2D,/*0x062C*/
  1 + 2            + 256 * 0x31,/*0x062D*/
  1 + 2            + 256 * 0x35,/*0x062E*/
  1                + 256 * 0x39,/*0x062F*/
  1                + 256 * 0x3B,/*0x0630*/
  1                + 256 * 0x3D,/*0x0631*/
  1                + 256 * 0x3F,/*0x0632*/
  1 + 2            + 256 * 0x41,/*0x0633*/
  1 + 2            + 256 * 0x45,/*0x0634*/
  1 + 2            + 256 * 0x49,/*0x0635*/
  1 + 2            + 256 * 0x4D,/*0x0636*/
  1 + 2            + 256 * 0x51,/*0x0637*/
  1 + 2            + 256 * 0x55,/*0x0638*/
  1 + 2            + 256 * 0x59,/*0x0639*/
  1 + 2            + 256 * 0x5D,/*0x063A*/
  0, 0, 0, 0, 0,                /*0x063B-0x063F*/
  1 + 2,                        /*0x0640*/
  1 + 2            + 256 * 0x61,/*0x0641*/
  1 + 2            + 256 * 0x65,/*0x0642*/
  1 + 2            + 256 * 0x69,/*0x0643*/
  1 + 2       + 16 + 256 * 0x6D,/*0x0644*/
  1 + 2            + 256 * 0x71,/*0x0645*/
  1 + 2            + 256 * 0x75,/*0x0646*/
  1 + 2            + 256 * 0x79,/*0x0647*/
  1                + 256 * 0x7D,/*0x0648*/
  1                + 256 * 0x7F,/*0x0649*/
  1 + 2            + 256 * 0x81,/*0x064A*/
         4         + 256 * 1,   /*0x064B*/
         4 + 128   + 256 * 1,   /*0x064C*/
         4 + 128   + 256 * 1,   /*0x064D*/
         4 + 128   + 256 * 1,   /*0x064E*/
         4 + 128   + 256 * 1,   /*0x064F*/
         4 + 128   + 256 * 1,   /*0x0650*/
         4 + 64    + 256 * 3,   /*0x0651*/
         4         + 256 * 1,   /*0x0652*/
         4         + 256 * 7,   /*0x0653*/
         4         + 256 * 8,   /*0x0654*/
         4         + 256 * 8,   /*0x0655*/
         4         + 256 * 1,   /*0x0656*/
  0, 0, 0, 0, 0,                /*0x0657-0x065B*/
  1                + 256 * 0x85,/*0x065C*/
  1                + 256 * 0x87,/*0x065D*/
  1                + 256 * 0x89,/*0x065E*/
  1                + 256 * 0x8B,/*0x065F*/
  0, 0, 0, 0, 0,                /*0x0660-0x0664*/
  0, 0, 0, 0, 0,                /*0x0665-0x0669*/
  0, 0, 0, 0, 0, 0,             /*0x066A-0x066F*/
         4         + 256 * 6,   /*0x0670*/
  1        + 8     + 256 * 0x00,/*0x0671*/
  1            + 32,            /*0x0672*/
  1            + 32,            /*0x0673*/
  0,                            /*0x0674*/
  1            + 32,            /*0x0675*/
  1, 1,                         /*0x0676-0x0677*/
  1 + 2,                        /*0x0678*/
  1 + 2 + 8        + 256 * 0x16,/*0x0679*/
  1 + 2 + 8        + 256 * 0x0E,/*0x067A*/
  1 + 2 + 8        + 256 * 0x02,/*0x067B*/
  1+2, 1+2,                     /*0x67C-0x067D*/
  1+2+8+256 * 0x06, 1+2, 1+2, 1+2, 1+2, 1+2, /*0x067E-0x0683*/
  1+2, 1+2, 1+2+8+256 * 0x2A, 1+2,           /*0x0684-0x0687*/
  1     + 8        + 256 * 0x38,/*0x0688*/
  1, 1, 1,                      /*0x0689-0x068B*/
  1     + 8        + 256 * 0x34,/*0x068C*/
  1     + 8        + 256 * 0x32,/*0x068D*/
  1     + 8        + 256 * 0x36,/*0x068E*/
  1, 1,                         /*0x068F-0x0690*/
  1     + 8        + 256 * 0x3C,/*0x0691*/
  1, 1, 1, 1, 1, 1, 1+8+256 * 0x3A, 1,       /*0x0692-0x0699*/
  1+2, 1+2, 1+2, 1+2, 1+2, 1+2, /*0x069A-0x06A3*/
  1+2, 1+2, 1+2, 1+2,           /*0x069A-0x06A3*/
  1+2, 1+2, 1+2, 1+2, 1+2, 1+2+8+256 * 0x3E, /*0x06A4-0x06AD*/
  1+2, 1+2, 1+2, 1+2,           /*0x06A4-0x06AD*/
  1+2, 1+2+8+256 * 0x42, 1+2, 1+2, 1+2, 1+2, /*0x06AE-0x06B7*/
  1+2, 1+2, 1+2, 1+2,           /*0x06AE-0x06B7*/
  1+2, 1+2,                     /*0x06B8-0x06B9*/
  1     + 8        + 256 * 0x4E,/*0x06BA*/
  1 + 2 + 8        + 256 * 0x50,/*0x06BB*/
  1+2, 1+2,                     /*0x06BC-0x06BD*/
  1 + 2 + 8        + 256 * 0x5A,/*0x06BE*/
  1+2,                          /*0x06BF*/
  1     + 8        + 256 * 0x54,/*0x06C0*/
  1 + 2 + 8        + 256 * 0x56,/*0x06C1*/
  1, 1, 1,                      /*0x06C2-0x06C4*/
  1     + 8        + 256 * 0x90,/*0x06C5*/
  1     + 8        + 256 * 0x89,/*0x06C6*/
  1     + 8        + 256 * 0x87,/*0x06C7*/
  1     + 8        + 256 * 0x8B,/*0x06C8*/
  1     + 8        + 256 * 0x92,/*0x06C9*/
  1,                            /*0x06CA*/
  1     + 8        + 256 * 0x8E,/*0x06CB*/
  1 + 2 + 8        + 256 * 0xAC,/*0x06CC*/
  1,                            /*0x06CD*/
  1+2, 1+2,                     /*0x06CE-0x06CF*/
  1 + 2 + 8        + 256 * 0x94,/*0x06D0*/
  1+2,                          /*0x06D1*/
  1     + 8        + 256 * 0x5E,/*0x06D2*/
  1     + 8        + 256 * 0x60 /*0x06D3*/
};

static const uint8_t presALink[] = {
/***********0*****1*****2*****3*****4*****5*****6*****7*****8*****9*****A*****B*****C*****D*****E*****F*/
/*FB5*/    0,    1,    0,    0,    0,    0,    0,    1,    2,1 + 2,    0,    0,    0,    0,    0,    0,
/*FB6*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FB7*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    1,    2,1 + 2,    0,    0,
/*FB8*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    1,    0,    0,    0,    1,
/*FB9*/    2,1 + 2,    0,    1,    2,1 + 2,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FBA*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FBB*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FBC*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FBD*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FBE*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FBF*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    1,    2,1 + 2,
/*FC0*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FC1*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FC2*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FC3*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FC4*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
/*FC5*/    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    4,    4,
/*FC6*/    4,    4,    4
};

static const uint8_t presBLink[]=
{
/***********0*****1*****2*****3*****4*****5*****6*****7*****8*****9*****A*****B*****C*****D*****E*****F*/
/*FE7*/1 + 2,1 + 2,1 + 2,    0,1 + 2,    0,1 + 2,1 + 2,1 + 2,1 + 2,1 + 2,1 + 2,1 + 2,1 + 2,1 + 2,1 + 2,
/*FE8*/    0,    0,    1,    0,    1,    0,    1,    0,    1,    0,    1,    2,1 + 2,    0,    1,    0,
/*FE9*/    1,    2,1 + 2,    0,    1,    0,    1,    2,1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,
/*FEA*/1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,1 + 2,    0,    1,    0,    1,    0,    1,    0,
/*FEB*/    1,    0,    1,    2,1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,
/*FEC*/1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,
/*FED*/1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,
/*FEE*/1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,1 + 2,    0,    1,    2,1 + 2,    0,    1,    0,
/*FEF*/    1,    0,    1,    2,1 + 2,    0,    1,    0,    1,    0,    1,    0,    1,    0,    0,    0
};

static const UChar convertFBto06[] =
{
/***********0******1******2******3******4******5******6******7******8******9******A******B******C******D******E******F***/
/*FB5*/   0x671, 0x671, 0x67B, 0x67B, 0x67B, 0x67B, 0x67E, 0x67E, 0x67E, 0x67E,     0,     0,     0,     0, 0x67A, 0x67A,
/*FB6*/   0x67A, 0x67A,     0,     0,     0,     0, 0x679, 0x679, 0x679, 0x679,     0,     0,     0,     0,     0,     0,
/*FB7*/       0,     0,     0,     0,     0,     0,     0,     0,     0,     0, 0x686, 0x686, 0x686, 0x686,     0,     0,
/*FB8*/       0,     0, 0x68D, 0x68D, 0x68C, 0x68C, 0x68E, 0x68E, 0x688, 0x688, 0x698, 0x698, 0x691, 0x691, 0x6A9, 0x6A9,
/*FB9*/   0x6A9, 0x6A9, 0x6AF, 0x6AF, 0x6AF, 0x6AF,     0,     0,     0,     0,     0,     0,     0,     0, 0x6BA, 0x6BA,
/*FBA*/   0x6BB, 0x6BB, 0x6BB, 0x6BB, 0x6C0, 0x6C0, 0x6C1, 0x6C1, 0x6C1, 0x6C1, 0x6BE, 0x6BE, 0x6BE, 0x6BE, 0x6d2, 0x6D2,
/*FBB*/   0x6D3, 0x6D3,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
/*FBC*/       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
/*FBD*/       0,     0,     0,     0,     0,     0,     0, 0x6C7, 0x6C7, 0x6C6, 0x6C6, 0x6C8, 0x6C8,     0, 0x6CB, 0x6CB,
/*FBE*/   0x6C5, 0x6C5, 0x6C9, 0x6C9, 0x6D0, 0x6D0, 0x6D0, 0x6D0,     0,     0,     0,     0,     0,     0,     0,     0,
/*FBF*/       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0, 0x6CC, 0x6CC, 0x6CC, 0x6CC
};

static const UChar convertFEto06[] =
{
/***********0******1******2******3******4******5******6******7******8******9******A******B******C******D******E******F***/
/*FE7*/   0x64B, 0x64B, 0x64C, 0x64C, 0x64D, 0x64D, 0x64E, 0x64E, 0x64F, 0x64F, 0x650, 0x650, 0x651, 0x651, 0x652, 0x652,
/*FE8*/   0x621, 0x622, 0x622, 0x623, 0x623, 0x624, 0x624, 0x625, 0x625, 0x626, 0x626, 0x626, 0x626, 0x627, 0x627, 0x628,
/*FE9*/   0x628, 0x628, 0x628, 0x629, 0x629, 0x62A, 0x62A, 0x62A, 0x62A, 0x62B, 0x62B, 0x62B, 0x62B, 0x62C, 0x62C, 0x62C,
/*FEA*/   0x62C, 0x62D, 0x62D, 0x62D, 0x62D, 0x62E, 0x62E, 0x62E, 0x62E, 0x62F, 0x62F, 0x630, 0x630, 0x631, 0x631, 0x632,
/*FEB*/   0x632, 0x633, 0x633, 0x633, 0x633, 0x634, 0x634, 0x634, 0x634, 0x635, 0x635, 0x635, 0x635, 0x636, 0x636, 0x636,
/*FEC*/   0x636, 0x637, 0x637, 0x637, 0x637, 0x638, 0x638, 0x638, 0x638, 0x639, 0x639, 0x639, 0x639, 0x63A, 0x63A, 0x63A,
/*FED*/   0x63A, 0x641, 0x641, 0x641, 0x641, 0x642, 0x642, 0x642, 0x642, 0x643, 0x643, 0x643, 0x643, 0x644, 0x644, 0x644,
/*FEE*/   0x644, 0x645, 0x645, 0x645, 0x645, 0x646, 0x646, 0x646, 0x646, 0x647, 0x647, 0x647, 0x647, 0x648, 0x648, 0x649,
/*FEF*/   0x649, 0x64A, 0x64A, 0x64A, 0x64A, 0x65C, 0x65C, 0x65D, 0x65D, 0x65E, 0x65E, 0x65F, 0x65F
};

static const uint8_t shapeTable[4][4][4]=
{
  { {0,0,0,0}, {0,0,0,0}, {0,1,0,3}, {0,1,0,1} },
  { {0,0,2,2}, {0,0,1,2}, {0,1,1,2}, {0,1,1,3} },
  { {0,0,0,0}, {0,0,0,0}, {0,1,0,3}, {0,1,0,3} },
  { {0,0,1,2}, {0,0,1,2}, {0,1,1,2}, {0,1,1,3} }
};

/*
 * This function shapes European digits to Arabic-Indic digits
 * in-place, writing over the input characters.
 * Since we know that we are only looking for BMP code points,
 * we can safely just work with code units (again, at least UTF-16).
 */
static void
_shapeToArabicDigitsWithContext(UChar *s, int32_t length,
                                UChar digitBase,
                                UBool isLogical, UBool lastStrongWasAL) {
    int32_t i;
    UChar c;

    digitBase-=0x30;

    /* the iteration direction depends on the type of input */
    if(isLogical) {
        for(i=0; i<length; ++i) {
            c=s[i];
            switch(ubidi_getClass(c)) {
            case U_LEFT_TO_RIGHT: /* L */
            case U_RIGHT_TO_LEFT: /* R */
                lastStrongWasAL=FALSE;
                break;
            case U_RIGHT_TO_LEFT_ARABIC: /* AL */
                lastStrongWasAL=TRUE;
                break;
            case U_EUROPEAN_NUMBER: /* EN */
                if(lastStrongWasAL && (uint32_t)(c-0x30)<10) {
                    s[i]=(UChar)(digitBase+c); /* digitBase+(c-0x30) - digitBase was modified above */
                }
                break;
            default :
                break;
            }
        }
    } else {
        for(i=length; i>0; /* pre-decrement in the body */) {
            c=s[--i];
            switch(ubidi_getClass(c)) {
            case U_LEFT_TO_RIGHT: /* L */
            case U_RIGHT_TO_LEFT: /* R */
                lastStrongWasAL=FALSE;
                break;
            case U_RIGHT_TO_LEFT_ARABIC: /* AL */
                lastStrongWasAL=TRUE;
                break;
            case U_EUROPEAN_NUMBER: /* EN */
                if(lastStrongWasAL && (uint32_t)(c-0x30)<10) {
                    s[i]=(UChar)(digitBase+c); /* digitBase+(c-0x30) - digitBase was modified above */
                }
                break;
            default :
                break;
            }
        }
    }
}

/*
 *Name     : invertBuffer
 *Function : This function inverts the buffer, it's used
 *           in case the user specifies the buffer to be
 *           U_SHAPE_TEXT_DIRECTION_LOGICAL
 */
static void
invertBuffer(UChar *buffer, int32_t size, uint32_t /*options*/, int32_t lowlimit, int32_t highlimit) {
    UChar temp;
    int32_t i=0,j=0;
    for(i=lowlimit,j=size-highlimit-1;i<j;i++,j--) {
        temp = buffer[i];
        buffer[i] = buffer[j];
        buffer[j] = temp;
    }
}

/*
 *Name     : changeLamAlef
 *Function : Converts the Alef characters into an equivalent
 *           LamAlef location in the 0x06xx Range, this is an
 *           intermediate stage in the operation of the program
 *           later it'll be converted into the 0xFExx LamAlefs
 *           in the shaping function.
 */
static inline UChar
changeLamAlef(UChar ch) {
    switch(ch) {
    case 0x0622 :
        return 0x065C;
    case 0x0623 :
        return 0x065D;
    case 0x0625 :
        return 0x065E;
    case 0x0627 :
        return 0x065F;
    }
    return 0;
}

/*
 *Name     : getLink
 *Function : Resolves the link between the characters as
 *           Arabic characters have four forms :
 *           Isolated, Initial, Middle and Final Form
 */
static UChar
getLink(UChar ch) {
    if(ch >= 0x0622 && ch <= 0x06D3) {
        return(araLink[ch-0x0622]);
    } else if(ch == 0x200D) {
        return(3);
    } else if(ch >= 0x206D && ch <= 0x206F) {
        return(4);
    }else if(ch >= 0xFB50 && ch <= 0xFC62) {
        return(presALink[ch-0xFB50]);
    } else if(ch >= 0xFE70 && ch <= 0xFEFC) {
        return(presBLink[ch-0xFE70]);
    }else {
        return(0);
    }
}

/*
 *Name     : countSpaces
 *Function : Counts the number of spaces
 *           at each end of the logical buffer
 */
static void
countSpaces(UChar *dest, int32_t size, uint32_t /*options*/, int32_t *spacesCountl, int32_t *spacesCountr) {
    int32_t i = 0;
    int32_t countl = 0,countr = 0;
    while((dest[i] == USHAPE_SPACE_CHAR) && (countl < size)) {
       countl++;
       i++;
    }
    if (countl < size) {  /* the entire buffer is not all space */
        while(dest[size-1] == USHAPE_SPACE_CHAR) {
            countr++;
            size--;
        }
    }
    *spacesCountl = countl;
    *spacesCountr = countr;
}

/*
 *Name     : isTashkeelChar
 *Function : Returns 1 for Tashkeel characters in 06 range else return 0
 */
static inline int32_t
isTashkeelChar(UChar ch) {
    return (int32_t)( ch>=0x064B && ch<= 0x0652 );
}

/*
 *Name     : isTashkeelCharFE
 *Function : Returns 1 for Tashkeel characters in FE range else return 0
 */
static inline int32_t
isTashkeelCharFE(UChar ch) {
    return (int32_t)( ch>=0xFE70 && ch<= 0xFE7F );
}

/*
 *Name     : isAlefChar
 *Function : Returns 1 for Alef characters else return 0
 */
static inline int32_t
isAlefChar(UChar ch) {
    return (int32_t)( (ch==0x0622)||(ch==0x0623)||(ch==0x0625)||(ch==0x0627) );
}

/*
 *Name     : isLamAlefChar
 *Function : Returns 1 for LamAlef characters else return 0
 */
static inline int32_t
isLamAlefChar(UChar ch) {
    return (int32_t)((ch>=0xFEF5)&&(ch<=0xFEFC) );
}

/*BIDI
 *Name     : isTailChar
 *Function : returns 1 if the character matches one of the tail characters (0xfe73 or 0x200b) otherwise returns 0
 */

static inline int32_t
isTailChar(UChar ch) {
    if(ch == USHAPE_OLD_TAIL_CHAR || ch == USHAPE_NEW_TAIL_CHAR){
            return 1;
    }else{
            return 0;
    }
}

/*BIDI
 *Name     : isSeenTailFamilyChar
 *Function : returns 1 if the character is a seen family isolated character
 *           in the FE range otherwise returns 0
 */

static inline int32_t
isSeenTailFamilyChar(UChar ch) {
    if(ch >= 0xfeb1 && ch < 0xfebf){
            return tailFamilyIsolatedFinal [ch - 0xFEB1];
    }else{
            return 0;
    }
}

 /* Name     : isSeenFamilyChar
  * Function : returns 1 if the character is a seen family character in the Unicode
  *            06 range otherwise returns 0
 */

static inline int32_t
isSeenFamilyChar(UChar  ch){
    if(ch >= 0x633 && ch <= 0x636){
        return 1;
    }else {
        return 0;
    }
}

/*Start of BIDI*/
/*
 *Name     : isAlefMaksouraChar
 *Function : returns 1 if the character is a Alef Maksoura Final or isolated
 *           otherwise returns 0
 */
static inline int32_t
isAlefMaksouraChar(UChar ch) {
    return (int32_t)( (ch == 0xFEEF) || ( ch == 0xFEF0) || (ch == 0x0649));
}

/*
 * Name     : isYehHamzaChar
 * Function : returns 1 if the character is a yehHamza isolated or yehhamza
 *            final is found otherwise returns 0
 */
static inline int32_t
isYehHamzaChar(UChar ch) {
    if((ch==0xFE89)||(ch==0xFE8A)){
        return 1;
    }else{
        return 0;
    }
}

 /*
 * Name: isTashkeelOnTatweelChar
 * Function: Checks if the Tashkeel Character is on Tatweel or not,if the
 *           Tashkeel on tatweel (FE range), it returns 1 else if the
 *           Tashkeel with shadda on tatweel (FC range)return 2 otherwise
 *           returns 0
 */
static inline int32_t
isTashkeelOnTatweelChar(UChar ch){
    if(ch >= 0xfe70 && ch <= 0xfe7f && ch != USHAPE_NEW_TAIL_CHAR && ch != 0xFE75 && ch != USHAPE_SHADDA_USHAPE_TATWEEL_CHAR)
    {
        return tashkeelMedial [ch - 0xFE70];
    }else if( (ch >= 0xfcf2 && ch <= 0xfcf4) || (ch == USHAPE_SHADDA_USHAPE_TATWEEL_CHAR)) {
        return 2;
    }else{
        return 0;
    }
}

/*
 * Name: isIsolatedTashkeelChar
 * Function: Checks if the Tashkeel Character is in the isolated form
 *           (i.e. Unicode FE range) returns 1 else if the Tashkeel
 *           with shadda is in the isolated form (i.e. Unicode FC range)
 *           returns 2 otherwise returns 0
 */
static inline int32_t
isIsolatedTashkeelChar(UChar ch){
    if(ch >= 0xfe70 && ch <= 0xfe7f && ch != USHAPE_NEW_TAIL_CHAR && ch != 0xFE75){
        return (1 - tashkeelMedial [ch - 0xFE70]);
    }else if(ch >= 0xfc5e && ch <= 0xfc63){
        return 1;
    }else{
        return 0;
    }
}




/*
 *Name     : calculateSize
 *Function : This function calculates the destSize to be used in preflighting
 *           when the destSize is equal to 0
 *           It is used also to calculate the new destsize in case the
 *           destination buffer will be resized.
 */

static int32_t
calculateSize(const UChar *source, int32_t sourceLength,
int32_t destSize,uint32_t options) {
    int32_t i = 0;

    int lamAlefOption = 0;
    int tashkeelOption = 0;

    destSize = sourceLength;

    if (((options&U_SHAPE_LETTERS_MASK) == U_SHAPE_LETTERS_SHAPE ||
        ((options&U_SHAPE_LETTERS_MASK) == U_SHAPE_LETTERS_SHAPE_TASHKEEL_ISOLATED )) &&
        ((options&U_SHAPE_LAMALEF_MASK) == U_SHAPE_LAMALEF_RESIZE )){
            lamAlefOption = 1;
    }
    if((options&U_SHAPE_LETTERS_MASK) == U_SHAPE_LETTERS_SHAPE &&
       ((options&U_SHAPE_TASHKEEL_MASK) == U_SHAPE_TASHKEEL_RESIZE ) ){
            tashkeelOption = 1;
        }

    if(lamAlefOption || tashkeelOption){
        if((options&U_SHAPE_TEXT_DIRECTION_MASK)==U_SHAPE_TEXT_DIRECTION_VISUAL_LTR) {
            for(i=0;i<sourceLength;i++) {
                if( ((isAlefChar(source[i]))&& (i<(sourceLength-1)) &&(source[i+1] == USHAPE_LAM_CHAR)) || (isTashkeelCharFE(source[i])) ) {
                        destSize--;
                    }
                }
            }else if((options&U_SHAPE_TEXT_DIRECTION_MASK)==U_SHAPE_TEXT_DIRECTION_LOGICAL) {
                for(i=0;i<sourceLength;i++) {
                    if( ( (source[i] == USHAPE_LAM_CHAR) && (i<(sourceLength-1)) && (isAlefChar(source[i+1]))) || (isTashkeelCharFE(source[i])) ) {
                        destSize--;
                    }
                }
            }
        }

    if ((options&U_SHAPE_LETTERS_MASK) == U_SHAPE_LETTERS_UNSHAPE){
        if ( (options&U_SHAPE_LAMALEF_MASK) == U_SHAPE_LAMALEF_RESIZE){
            for(i=0;i<sourceLength;i++) {
                if(isLamAlefChar(source[i]))
                destSize++;
            }
        }
    }

    return destSize;
}

/*
 *Name     : handleTashkeelWithTatweel
 *Function : Replaces Tashkeel as following:
 *            Case 1 :if the Tashkeel on tatweel, replace it with Tatweel.
 *            Case 2 :if the Tashkeel aggregated with Shadda on Tatweel, replace
 *                   it with Shadda on Tatweel.
 *            Case 3: if the Tashkeel is isolated replace it with Space.
 *
 */
static int32_t
handleTashkeelWithTatweel(UChar *dest, int32_t sourceLength,
             int32_t /*destSize*/, uint32_t /*options*/,
             UErrorCode * /*pErrorCode*/) {
                 int i;
                 for(i = 0; i < sourceLength; i++){
                     if((isTashkeelOnTatweelChar(dest[i]) == 1)){
                         dest[i] = USHAPE_TATWEEL_CHAR;
                    }else if((isTashkeelOnTatweelChar(dest[i]) == 2)){
                         dest[i] = USHAPE_SHADDA_USHAPE_TATWEEL_CHAR;
                    }else if(isIsolatedTashkeelChar(dest[i]) && dest[i] != USHAPE_SHADDA_CHAR){
                         dest[i] = USHAPE_SPACE_CHAR;
                    }
                 }
                 return sourceLength;
}



/*
 *Name     : handleGeneratedSpaces
 *Function : The shapeUnicode function converts Lam + Alef into LamAlef + space,
 *           and Tashkeel to space.
 *           handleGeneratedSpaces function puts these generated spaces
 *           according to the options the user specifies. LamAlef and Tashkeel
 *           spaces can be replaced at begin, at end, at near or decrease the
 *           buffer size.
 *
 *           There is also Auto option for LamAlef and tashkeel, which will put
 *           the spaces at end of the buffer (or end of text if the user used
 *           the option U_SHAPE_SPACES_RELATIVE_TO_TEXT_BEGIN_END).
 *
 *           If the text type was visual_LTR and the option
 *           U_SHAPE_SPACES_RELATIVE_TO_TEXT_BEGIN_END was selected the END
 *           option will place the space at the beginning of the buffer and
 *           BEGIN will place the space at the end of the buffer.
 */

static int32_t
handleGeneratedSpaces(UChar *dest, int32_t sourceLength,
                    int32_t destSize,
                    uint32_t options,
                    UErrorCode *pErrorCode,struct uShapeVariables shapeVars ) {

    int32_t i = 0, j = 0;
    int32_t count = 0;
    UChar *tempbuffer=NULL;

    int lamAlefOption = 0;
    int tashkeelOption = 0;
    int shapingMode = USHAPE_SHAPE_MODE;

    if (shapingMode == 0){
        if ( (options&U_SHAPE_LAMALEF_MASK) == U_SHAPE_LAMALEF_RESIZE ){
            lamAlefOption = 1;
        }
        if ( (options&U_SHAPE_TASHKEEL_MASK) == U_SHAPE_TASHKEEL_RESIZE ){
            tashkeelOption = 1;
        }
    }

    tempbuffer = (UChar *)uprv_malloc((sourceLength+1)*U_SIZEOF_UCHAR);
    /* Test for NULL */
    if(tempbuffer == NULL) {
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }


    if (lamAlefOption || tashkeelOption){
        uprv_memset(tempbuffer, 0, (sourceLength+1)*U_SIZEOF_UCHAR);

        i = j = 0; count = 0;
        while(i < sourceLength) {
            if ( (lamAlefOption && dest[i] == USHAPE_LAMALEF_SPACE_SUB) ||
               (tashkeelOption && dest[i] == USHAPE_TASHKEEL_SPACE_SUB) ){
                j--;
                count++;
            } else {
                tempbuffer[j] = dest[i];
            }
            i++;
            j++;
        }

        while(count >= 0) {
            tempbuffer[i] = 0x0000;
            i--;
            count--;
        }

        u_memcpy(dest, tempbuffer, sourceLength);
        destSize = u_strlen(dest);
    }

      lamAlefOption = 0;

    if (shapingMode == 0){
        if ( (options&U_SHAPE_LAMALEF_MASK) == U_SHAPE_LAMALEF_NEAR ){
            lamAlefOption = 1;
        }
    }

    if (lamAlefOption){
        /* Lam+Alef is already shaped into LamAlef + FFFF */
        i = 0;
        while(i < sourceLength) {
            if(lamAlefOption&&dest[i] == USHAPE_LAMALEF_SPACE_SUB){
                dest[i] = USHAPE_SPACE_CHAR;
            }
            i++;
        }
        destSize = sourceLength;
    }
    lamAlefOption = 0;
    tashkeelOption = 0;

    if (shapingMode == 0) {
        if ( ((options&U_SHAPE_LAMALEF_MASK) == shapeVars.uShapeLamalefBegin) ||
              (((options&U_SHAPE_LAMALEF_MASK) == U_SHAPE_LAMALEF_AUTO )
              && (shapeVars.spacesRelativeToTextBeginEnd==1)) ) {
            lamAlefOption = 1;
        }
        if ( (options&U_SHAPE_TASHKEEL_MASK) == shapeVars.uShapeTashkeelBegin ) {
            tashkeelOption = 1;
        }
    }

    if(lamAlefOption || tashkeelOption){
        uprv_memset(tempbuffer, 0, (sourceLength+1)*U_SIZEOF_UCHAR);

        i = j = sourceLength; count = 0;

        while(i >= 0) {
            if ( (lamAlefOption && dest[i] == USHAPE_LAMALEF_SPACE_SUB) ||
                 (tashkeelOption && dest[i] == USHAPE_TASHKEEL_SPACE_SUB) ){
                j++;
                count++;
            }else {
                tempbuffer[j] = dest[i];
            }
            i--;
            j--;
        }

        for(i=0 ;i < count; i++){
                tempbuffer[i] = USHAPE_SPACE_CHAR;
        }

        u_memcpy(dest, tempbuffer, sourceLength);
        destSize = sourceLength;
    }



    lamAlefOption = 0;
    tashkeelOption = 0;

    if (shapingMode == 0) {
        if ( ((options&U_SHAPE_LAMALEF_MASK) == shapeVars.uShapeLamalefEnd) ||
              (((options&U_SHAPE_LAMALEF_MASK) == U_SHAPE_LAMALEF_AUTO )
              && (shapeVars.spacesRelativeToTextBeginEnd==0)) ) {
            lamAlefOption = 1;
        }
        if ( (options&U_SHAPE_TASHKEEL_MASK) == shapeVars.uShapeTashkeelEnd ){
            tashkeelOption = 1;
        }
    }

    if(lamAlefOption || tashkeelOption){
        uprv_memset(tempbuffer, 0, (sourceLength+1)*U_SIZEOF_UCHAR);

        i = j = 0; count = 0;
        while(i < sourceLength) {
            if ( (lamAlefOption && dest[i] == USHAPE_LAMALEF_SPACE_SUB) ||
                 (tashkeelOption && dest[i] == USHAPE_TASHKEEL_SPACE_SUB) ){
                j--;
                count++;
            }else {
                tempbuffer[j] = dest[i];
            }
            i++;
            j++;
        }

        while(count >= 0) {
            tempbuffer[i] = USHAPE_SPACE_CHAR;
            i--;
            count--;
        }

        u_memcpy(dest, tempbuffer, sourceLength);
        destSize = sourceLength;
    }


    if(tempbuffer){
        uprv_free(tempbuffer);
    }

    return destSize;
}

/*
 *Name     :expandCompositCharAtBegin
 *Function :Expands the LamAlef character to Lam and Alef consuming the required
 *         space from beginning of the buffer. If the text type was visual_LTR
 *         and the option U_SHAPE_SPACES_RELATIVE_TO_TEXT_BEGIN_END was selected
 *         the spaces will be located at end of buffer.
 *         If there are no spaces to expand the LamAlef, an error
 *         will be set to U_NO_SPACE_AVAILABLE as defined in utypes.h
 */

static int32_t
expandCompositCharAtBegin(UChar *dest, int32_t sourceLength, int32_t destSize,UErrorCode *pErrorCode) {
    int32_t      i = 0,j = 0;
    int32_t      countl = 0;
    UChar    *tempbuffer=NULL;

    tempbuffer = (UChar *)uprv_malloc((sourceLength+1)*U_SIZEOF_UCHAR);

    /* Test for NULL */
    if(tempbuffer == NULL) {
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }

        uprv_memset(tempbuffer, 0, (sourceLength+1)*U_SIZEOF_UCHAR);

        i = 0;
        while(dest[i] == USHAPE_SPACE_CHAR) {
            countl++;
            i++;
        }

        i = j = sourceLength-1;

        while(i >= 0 && j >= 0) {
            if( countl>0 && isLamAlefChar(dest[i])) {
                tempbuffer[j] = USHAPE_LAM_CHAR;
                /* to ensure the array index is within the range */
                U_ASSERT(dest[i] >= 0xFEF5u
                    && dest[i]-0xFEF5u < UPRV_LENGTHOF(convertLamAlef));
                tempbuffer[j-1] = convertLamAlef[ dest[i] - 0xFEF5 ];
                j--;
                countl--;
            }else {
                 if( countl == 0 && isLamAlefChar(dest[i]) ) {
                     *pErrorCode=U_NO_SPACE_AVAILABLE;
                     }
                 tempbuffer[j] = dest[i];
            }
            i--;
            j--;
        }
        u_memcpy(dest, tempbuffer, sourceLength);

        uprv_free(tempbuffer);

        destSize = sourceLength;
        return destSize;
}

/*
 *Name     : expandCompositCharAtEnd
 *Function : Expands the LamAlef character to Lam and Alef consuming the
 *           required space from end of the buffer. If the text type was
 *           Visual LTR and the option U_SHAPE_SPACES_RELATIVE_TO_TEXT_BEGIN_END
 *           was used, the spaces will be consumed from begin of buffer. If
 *           there are no spaces to expand the LamAlef, an error
 *           will be set to U_NO_SPACE_AVAILABLE as defined in utypes.h
 */

static int32_t
expandCompositCharAtEnd(UChar *dest, int32_t sourceLength, int32_t destSize,UErrorCode *pErrorCode) {
    int32_t      i = 0,j = 0;

    int32_t      countr = 0;
    int32_t  inpsize = sourceLength;

    UChar    *tempbuffer=NULL;
    tempbuffer = (UChar *)uprv_malloc((sourceLength+1)*U_SIZEOF_UCHAR);

    /* Test for NULL */
    if(tempbuffer == NULL) {
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
         return 0;
    }

    uprv_memset(tempbuffer, 0, (sourceLength+1)*U_SIZEOF_UCHAR);

    while(dest[inpsize-1] == USHAPE_SPACE_CHAR) {
        countr++;
        inpsize--;
    }

    i = sourceLength - countr - 1;
    j = sourceLength - 1;

    while(i >= 0 && j >= 0) {
        if( countr>0 && isLamAlefChar(dest[i]) ) {
            tempbuffer[j] = USHAPE_LAM_CHAR;
            tempbuffer[j-1] = convertLamAlef[ dest[i] - 0xFEF5 ];
            j--;
            countr--;
        }else {
            if ((countr == 0) && isLamAlefChar(dest[i]) ) {
                *pErrorCode=U_NO_SPACE_AVAILABLE;
            }
            tempbuffer[j] = dest[i];
        }
        i--;
        j--;
    }

    if(countr > 0) {
        u_memmove(tempbuffer, tempbuffer+countr, sourceLength);
        if(u_strlen(tempbuffer) < sourceLength) {
            for(i=sourceLength-1;i>=sourceLength-countr;i--) {
                tempbuffer[i] = USHAPE_SPACE_CHAR;
            }
        }
    }
    u_memcpy(dest, tempbuffer, sourceLength);

    uprv_free(tempbuffer);

    destSize = sourceLength;
    return destSize;
}

/*
 *Name     : expandCompositCharAtNear
 *Function : Expands the LamAlef character into Lam + Alef, YehHamza character
 *           into Yeh + Hamza, SeenFamily character into SeenFamily character
 *           + Tail, while consuming the space next to the character.
 *           If there are no spaces next to the character, an error
 *           will be set to U_NO_SPACE_AVAILABLE as defined in utypes.h
 */

static int32_t
expandCompositCharAtNear(UChar *dest, int32_t sourceLength, int32_t destSize,UErrorCode *pErrorCode,
                         int yehHamzaOption, int seenTailOption, int lamAlefOption, struct uShapeVariables shapeVars) {
    int32_t      i = 0;


    UChar    lamalefChar, yehhamzaChar;

    for(i = 0 ;i<=sourceLength-1;i++) {
            if (seenTailOption && isSeenTailFamilyChar(dest[i])) {
                if ((i>0) && (dest[i-1] == USHAPE_SPACE_CHAR) ) {
                    dest[i-1] = shapeVars.tailChar;
                }else {
                    *pErrorCode=U_NO_SPACE_AVAILABLE;
                }
            }else if(yehHamzaOption && (isYehHamzaChar(dest[i])) ) {
                if ((i>0) && (dest[i-1] == USHAPE_SPACE_CHAR) ) {
                    yehhamzaChar = dest[i];
                    dest[i] = yehHamzaToYeh[yehhamzaChar - USHAPE_YEH_HAMZAFE_CHAR];
                    dest[i-1] = USHAPE_HAMZAFE_CHAR;
                }else {

                    *pErrorCode=U_NO_SPACE_AVAILABLE;
                }
            }else if(lamAlefOption && isLamAlefChar(dest[i+1])) {
                if(dest[i] == USHAPE_SPACE_CHAR){
                    lamalefChar = dest[i+1];
                    dest[i+1] = USHAPE_LAM_CHAR;
                    dest[i] = convertLamAlef[ lamalefChar - 0xFEF5 ];
                }else {
                    *pErrorCode=U_NO_SPACE_AVAILABLE;
                }
            }
       }
       destSize = sourceLength;
       return destSize;
}
 /*
 * Name     : expandCompositChar
 * Function : LamAlef, need special handling, since it expands from one
 *            character into two characters while shaping or deshaping.
 *            In order to expand it, near or far spaces according to the
 *            options user specifies. Also buffer size can be increased.
 *
 *            For SeenFamily characters and YehHamza only the near option is
 *            supported, while for LamAlef we can take spaces from begin, end,
 *            near or even increase the buffer size.
 *            There is also the Auto option for LamAlef only, which will first
 *            search for a space at end, begin then near, respectively.
 *            If there are no spaces to expand these characters, an error will be set to
 *            U_NO_SPACE_AVAILABLE as defined in utypes.h
 */

static int32_t
expandCompositChar(UChar *dest, int32_t sourceLength,
              int32_t destSize,uint32_t options,
              UErrorCode *pErrorCode, int shapingMode,struct uShapeVariables shapeVars) {

    int32_t      i = 0,j = 0;

    UChar    *tempbuffer=NULL;
    int yehHamzaOption = 0;
    int seenTailOption = 0;
    int lamAlefOption = 0;

    if (shapingMode == 1){
        if ( (options&U_SHAPE_LAMALEF_MASK) == U_SHAPE_LAMALEF_AUTO){

            if(shapeVars.spacesRelativeToTextBeginEnd == 0) {
                destSize = expandCompositCharAtEnd(dest, sourceLength, destSize, pErrorCode);

                if(*pErrorCode == U_NO_SPACE_AVAILABLE) {
                    *pErrorCode = U_ZERO_ERROR;
                    destSize = expandCompositCharAtBegin(dest, sourceLength, destSize, pErrorCode);
                }
            }else {
                destSize = expandCompositCharAtBegin(dest, sourceLength, destSize, pErrorCode);

                if(*pErrorCode == U_NO_SPACE_AVAILABLE) {
                    *pErrorCode = U_ZERO_ERROR;
                    destSize = expandCompositCharAtEnd(dest, sourceLength, destSize, pErrorCode);
                }
            }

            if(*pErrorCode == U_NO_SPACE_AVAILABLE) {
                *pErrorCode = U_ZERO_ERROR;
                destSize = expandCompositCharAtNear(dest, sourceLength, destSize, pErrorCode, yehHamzaOption,
                                                seenTailOption, 1,shapeVars);
            }
        }
    }

    if (shapingMode == 1){
        if ( (options&U_SHAPE_LAMALEF_MASK) == shapeVars.uShapeLamalefEnd){
            destSize = expandCompositCharAtEnd(dest, sourceLength, destSize, pErrorCode);
        }
    }

    if (shapingMode == 1){
        if ( (options&U_SHAPE_LAMALEF_MASK) == shapeVars.uShapeLamalefBegin){
            destSize = expandCompositCharAtBegin(dest, sourceLength, destSize, pErrorCode);
        }
    }

    if (shapingMode == 0){
         if ((options&U_SHAPE_YEHHAMZA_MASK) == U_SHAPE_YEHHAMZA_TWOCELL_NEAR){
             yehHamzaOption = 1;
         }
         if ((options&U_SHAPE_SEEN_MASK) == U_SHAPE_SEEN_TWOCELL_NEAR){
            seenTailOption = 1;
         }
    }
    if (shapingMode == 1) {
        if ( (options&U_SHAPE_LAMALEF_MASK) == U_SHAPE_LAMALEF_NEAR) {
            lamAlefOption = 1;
        }
    }


    if (yehHamzaOption || seenTailOption || lamAlefOption){
        destSize = expandCompositCharAtNear(dest, sourceLength, destSize, pErrorCode, yehHamzaOption,
                                            seenTailOption,lamAlefOption,shapeVars);
    }


    if (shapingMode == 1){
        if ( (options&U_SHAPE_LAMALEF_MASK) == U_SHAPE_LAMALEF_RESIZE){
            destSize = calculateSize(dest,sourceLength,destSize,options);
            tempbuffer = (UChar *)uprv_malloc((destSize+1)*U_SIZEOF_UCHAR);

            /* Test for NULL */
            if(tempbuffer == NULL) {
                *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
                return 0;
            }

            uprv_memset(tempbuffer, 0, (destSize+1)*U_SIZEOF_UCHAR);

            i = j = 0;
            while(i < destSize && j < destSize) {
                if(isLamAlefChar(dest[i]) ) {
                    tempbuffer[j] = convertLamAlef[ dest[i] - 0xFEF5 ];
                    tempbuffer[j+1] = USHAPE_LAM_CHAR;
                    j++;
                }else {
                    tempbuffer[j] = dest[i];
                }
                i++;
                j++;
            }

            u_memcpy(dest, tempbuffer, destSize);
        }
    }

    if(tempbuffer) {
        uprv_free(tempbuffer);
    }
    return destSize;
}

/*
 *Name     : shapeUnicode
 *Function : Converts an Arabic Unicode buffer in 06xx Range into a shaped
 *           arabic Unicode buffer in FExx Range
 */
static int32_t
shapeUnicode(UChar *dest, int32_t sourceLength,
             int32_t destSize,uint32_t options,
             UErrorCode *pErrorCode,
             int tashkeelFlag, struct uShapeVariables shapeVars) {

    int32_t          i, iend;
    int32_t          step;
    int32_t          lastPos,Nx, Nw;
    unsigned int     Shape;
    int32_t          lamalef_found = 0;
    int32_t seenfamFound = 0, yehhamzaFound =0, tashkeelFound  = 0;
    UChar            prevLink = 0, lastLink = 0, currLink, nextLink = 0;
    UChar            wLamalef;

    /*
     * Converts the input buffer from FExx Range into 06xx Range
     * to make sure that all characters are in the 06xx range
     * even the lamalef is converted to the special region in
     * the 06xx range
     */
    if ((options & U_SHAPE_PRESERVE_PRESENTATION_MASK)  == U_SHAPE_PRESERVE_PRESENTATION_NOOP) {
        for (i = 0; i < sourceLength; i++) {
            UChar inputChar  = dest[i];
            if ( (inputChar >= 0xFB50) && (inputChar <= 0xFBFF)) {
                UChar c = convertFBto06 [ (inputChar - 0xFB50) ];
                if (c != 0)
                    dest[i] = c;
            } else if ( (inputChar >= 0xFE70) && (inputChar <= 0xFEFC)) {
                dest[i] = convertFEto06 [ (inputChar - 0xFE70) ] ;
            } else {
                dest[i] = inputChar ;
            }
        }
    }


    /* sets the index to the end of the buffer, together with the step point to -1 */
    i = sourceLength - 1;
    iend = -1;
    step = -1;

    /*
     * This function resolves the link between the characters .
     * Arabic characters have four forms :
     * Isolated Form, Initial Form, Middle Form and Final Form
     */
    currLink = getLink(dest[i]);

    lastPos = i;
    Nx = -2, Nw = 0;

    while (i != iend) {
        /* If high byte of currLink > 0 then more than one shape */
        if ((currLink & 0xFF00) > 0 || (getLink(dest[i]) & USHAPE_IRRELEVANT) != 0) {
            Nw = i + step;
            while (Nx < 0) {         /* we need to know about next char */
                if(Nw == iend) {
                    nextLink = 0;
                    Nx = 3000;
                } else {
                    nextLink = getLink(dest[Nw]);
                    if((nextLink & USHAPE_IRRELEVANT) == 0) {
                        Nx = Nw;
                    } else {
                        Nw = Nw + step;
                    }
                }
            }

            if ( ((currLink & USHAPE_ALEFTYPE) > 0)  &&  ((lastLink & USHAPE_LAMTYPE) > 0) ) {
                lamalef_found = 1;
                wLamalef = changeLamAlef(dest[i]); /*get from 0x065C-0x065f */
                if ( wLamalef != 0) {
                    dest[i] = USHAPE_LAMALEF_SPACE_SUB;            /* The default case is to drop the Alef and replace */
                    dest[lastPos] =wLamalef;     /* it by USHAPE_LAMALEF_SPACE_SUB which is the last character in the  */
                    i=lastPos;                   /* unicode private use area, this is done to make   */
                }                                /* sure that removeLamAlefSpaces() handles only the */
                lastLink = prevLink;             /* spaces generated during lamalef generation.      */
                currLink = getLink(wLamalef);    /* USHAPE_LAMALEF_SPACE_SUB is added here and is replaced by spaces   */
            }                                    /* in removeLamAlefSpaces()                         */

            if ((i > 0) && (dest[i-1] == USHAPE_SPACE_CHAR)){
                if ( isSeenFamilyChar(dest[i])) {
                    seenfamFound = 1;
                } else if (dest[i] == USHAPE_YEH_HAMZA_CHAR) {
                    yehhamzaFound = 1;
                }
            }
            else if(i==0){
                if ( isSeenFamilyChar(dest[i])){
                    seenfamFound = 1;
                } else if (dest[i] == USHAPE_YEH_HAMZA_CHAR) {
                    yehhamzaFound = 1;
                }
            }

            /*
             * get the proper shape according to link ability of neighbors
             * and of character; depends on the order of the shapes
             * (isolated, initial, middle, final) in the compatibility area
             */
            Shape = shapeTable[nextLink & (USHAPE_LINKR + USHAPE_LINKL)]
                              [lastLink & (USHAPE_LINKR + USHAPE_LINKL)]
                              [currLink & (USHAPE_LINKR + USHAPE_LINKL)];

            if ((currLink & (USHAPE_LINKR+USHAPE_LINKL)) == 1) {
                Shape &= 1;
            } else if(isTashkeelChar(dest[i])) {
                if( (lastLink & USHAPE_LINKL) && (nextLink & USHAPE_LINKR) && (tashkeelFlag == 1) &&
                     dest[i] != 0x064C && dest[i] != 0x064D )
                {
                    Shape = 1;
                    if( (nextLink&USHAPE_ALEFTYPE) == USHAPE_ALEFTYPE && (lastLink&USHAPE_LAMTYPE) == USHAPE_LAMTYPE ) {
                        Shape = 0;
                    }
                } else if(tashkeelFlag == 2 && dest[i] == USHAPE_SHADDA06_CHAR){
                    Shape = 1;
                } else {
                    Shape = 0;
                }
            }
            if ((dest[i] ^ 0x0600) < 0x100) {
                if ( isTashkeelChar(dest[i]) ){
                    if (tashkeelFlag == 2  && dest[i] != USHAPE_SHADDA06_CHAR){
                        dest[i] = USHAPE_TASHKEEL_SPACE_SUB;
                        tashkeelFound  = 1;
                    } else {
                        /* to ensure the array index is within the range */
                        U_ASSERT(dest[i] >= 0x064Bu
                            && dest[i]-0x064Bu < UPRV_LENGTHOF(IrrelevantPos));
                        dest[i] =  0xFE70 + IrrelevantPos[(dest[i] - 0x064B)] + static_cast<UChar>(Shape);
                    }
                }else if ((currLink & USHAPE_APRESENT) > 0) {
                    dest[i] = (UChar)(0xFB50 + (currLink >> 8) + Shape);
                }else if ((currLink >> 8) > 0 && (currLink & USHAPE_IRRELEVANT) == 0) {
                    dest[i] = (UChar)(0xFE70 + (currLink >> 8) + Shape);
                }
            }
        }

        /* move one notch forward */
        if ((currLink & USHAPE_IRRELEVANT) == 0) {
            prevLink = lastLink;
            lastLink = currLink;
            lastPos = i;
        }

        i = i + step;
        if (i == Nx) {
            currLink = nextLink;
            Nx = -2;
        } else if(i != iend) {
            currLink = getLink(dest[i]);
        }
    }
    destSize = sourceLength;
    if ( (lamalef_found != 0 ) || (tashkeelFound  != 0) ){
        destSize = handleGeneratedSpaces(dest,sourceLength,destSize,options,pErrorCode, shapeVars);
    }

    if ( (seenfamFound != 0) || (yehhamzaFound != 0) ) {
        destSize = expandCompositChar(dest, sourceLength,destSize,options,pErrorCode, USHAPE_SHAPE_MODE,shapeVars);
    }
    return destSize;
}

/*
 *Name     : deShapeUnicode
 *Function : Converts an Arabic Unicode buffer in FExx Range into unshaped
 *           arabic Unicode buffer in 06xx Range
 */
static int32_t
deShapeUnicode(UChar *dest, int32_t sourceLength,
               int32_t destSize,uint32_t options,
               UErrorCode *pErrorCode, struct uShapeVariables shapeVars) {
    int32_t i = 0;
    int32_t lamalef_found = 0;
    int32_t yehHamzaComposeEnabled = 0;
    int32_t seenComposeEnabled = 0;

    yehHamzaComposeEnabled = ((options&U_SHAPE_YEHHAMZA_MASK) == U_SHAPE_YEHHAMZA_TWOCELL_NEAR) ? 1 : 0;
    seenComposeEnabled = ((options&U_SHAPE_SEEN_MASK) == U_SHAPE_SEEN_TWOCELL_NEAR)? 1 : 0;

    /*
     *This for loop changes the buffer from the Unicode FE range to
     *the Unicode 06 range
     */

    for(i = 0; i < sourceLength; i++) {
        UChar  inputChar = dest[i];
        if ( (inputChar >= 0xFB50) && (inputChar <= 0xFBFF)) { /* FBxx Arabic range */
            UChar c = convertFBto06 [ (inputChar - 0xFB50) ];
            if (c != 0)
                dest[i] = c;
        } else if( (yehHamzaComposeEnabled == 1) && ((inputChar == USHAPE_HAMZA06_CHAR) || (inputChar == USHAPE_HAMZAFE_CHAR))
                && (i < (sourceLength - 1)) && isAlefMaksouraChar(dest[i+1] )) {
                 dest[i] = USHAPE_SPACE_CHAR;
                 dest[i+1] = USHAPE_YEH_HAMZA_CHAR;
        } else if ( (seenComposeEnabled == 1) && (isTailChar(inputChar)) && (i< (sourceLength - 1))
                        && (isSeenTailFamilyChar(dest[i+1])) ) {
                dest[i] = USHAPE_SPACE_CHAR;
        } else if (( inputChar >= 0xFE70) && (inputChar <= 0xFEF4 )) { /* FExx Arabic range */
                dest[i] = convertFEto06 [ (inputChar - 0xFE70) ];
        } else {
            dest[i] = inputChar ;
        }

        if( isLamAlefChar(dest[i]) )
            lamalef_found = 1;
    }

   destSize = sourceLength;
   if (lamalef_found != 0){
          destSize = expandCompositChar(dest,sourceLength,destSize,options,pErrorCode,USHAPE_DESHAPE_MODE, shapeVars);
   }
   return destSize;
}

/*
 ****************************************
 * u_shapeArabic
 ****************************************
 */

U_CAPI int32_t U_EXPORT2
u_shapeArabic(const UChar *source, int32_t sourceLength,
              UChar *dest, int32_t destCapacity,
              uint32_t options,
              UErrorCode *pErrorCode) {

    int32_t destLength;
    struct  uShapeVariables shapeVars = { USHAPE_OLD_TAIL_CHAR,U_SHAPE_LAMALEF_BEGIN,U_SHAPE_LAMALEF_END,U_SHAPE_TASHKEEL_BEGIN,U_SHAPE_TASHKEEL_END,0};

    /* usual error checking */
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* make sure that no reserved options values are used; allow dest==NULL only for preflighting */
    if( source==NULL || sourceLength<-1 || (dest==NULL && destCapacity!=0) || destCapacity<0 ||
                (((options&U_SHAPE_TASHKEEL_MASK) > 0) &&
                 ((options&U_SHAPE_LETTERS_SHAPE_TASHKEEL_ISOLATED) == U_SHAPE_LETTERS_SHAPE_TASHKEEL_ISOLATED) ) ||
                (((options&U_SHAPE_TASHKEEL_MASK) > 0) &&
                 ((options&U_SHAPE_LETTERS_MASK) == U_SHAPE_LETTERS_UNSHAPE)) ||
                (options&U_SHAPE_DIGIT_TYPE_RESERVED)==U_SHAPE_DIGIT_TYPE_RESERVED ||
                (options&U_SHAPE_DIGITS_MASK)==U_SHAPE_DIGITS_RESERVED ||
                ((options&U_SHAPE_LAMALEF_MASK) != U_SHAPE_LAMALEF_RESIZE  &&
                (options&U_SHAPE_AGGREGATE_TASHKEEL_MASK) != 0) ||
                ((options&U_SHAPE_AGGREGATE_TASHKEEL_MASK) == U_SHAPE_AGGREGATE_TASHKEEL &&
                (options&U_SHAPE_LETTERS_SHAPE_TASHKEEL_ISOLATED) != U_SHAPE_LETTERS_SHAPE_TASHKEEL_ISOLATED)
    )
    {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    /* Validate  lamalef options */
    if(((options&U_SHAPE_LAMALEF_MASK) > 0)&&
              !(((options & U_SHAPE_LAMALEF_MASK)==U_SHAPE_LAMALEF_BEGIN) ||
                ((options & U_SHAPE_LAMALEF_MASK)==U_SHAPE_LAMALEF_END ) ||
                ((options & U_SHAPE_LAMALEF_MASK)==U_SHAPE_LAMALEF_RESIZE )||
                 ((options & U_SHAPE_LAMALEF_MASK)==U_SHAPE_LAMALEF_AUTO) ||
                 ((options & U_SHAPE_LAMALEF_MASK)==U_SHAPE_LAMALEF_NEAR)))
    {
         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    /* Validate  Tashkeel options */
    if(((options&U_SHAPE_TASHKEEL_MASK) > 0)&&
                   !(((options & U_SHAPE_TASHKEEL_MASK)==U_SHAPE_TASHKEEL_BEGIN) ||
                     ((options & U_SHAPE_TASHKEEL_MASK)==U_SHAPE_TASHKEEL_END )
                    ||((options & U_SHAPE_TASHKEEL_MASK)==U_SHAPE_TASHKEEL_RESIZE )||
                    ((options & U_SHAPE_TASHKEEL_MASK)==U_SHAPE_TASHKEEL_REPLACE_BY_TATWEEL)))
    {
         *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    /* determine the source length */
    if(sourceLength==-1) {
        sourceLength=u_strlen(source);
    }
    if(sourceLength<=0) {
        return u_terminateUChars(dest, destCapacity, 0, pErrorCode);
    }

    /* check that source and destination do not overlap */
    if( dest!=NULL &&
        ((source<=dest && dest<source+sourceLength) ||
         (dest<=source && source<dest+destCapacity))) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* Does Options contain the new Seen Tail Unicode code point option */
    if ( (options&U_SHAPE_TAIL_TYPE_MASK) == U_SHAPE_TAIL_NEW_UNICODE){
        shapeVars.tailChar = USHAPE_NEW_TAIL_CHAR;
    }else {
        shapeVars.tailChar = USHAPE_OLD_TAIL_CHAR;
    }

    if((options&U_SHAPE_LETTERS_MASK)!=U_SHAPE_LETTERS_NOOP) {
        UChar buffer[300];
        UChar *tempbuffer, *tempsource = NULL;
        int32_t outputSize, spacesCountl=0, spacesCountr=0;

        if((options&U_SHAPE_AGGREGATE_TASHKEEL_MASK)>0) {
            int32_t logical_order = (options&U_SHAPE_TEXT_DIRECTION_MASK) == U_SHAPE_TEXT_DIRECTION_LOGICAL;
            int32_t aggregate_tashkeel =
                        (options&(U_SHAPE_AGGREGATE_TASHKEEL_MASK+U_SHAPE_LETTERS_SHAPE_TASHKEEL_ISOLATED)) ==
                        (U_SHAPE_AGGREGATE_TASHKEEL+U_SHAPE_LETTERS_SHAPE_TASHKEEL_ISOLATED);
            int step=logical_order?1:-1;
            int j=logical_order?-1:2*sourceLength;
            int i=logical_order?-1:sourceLength;
            int end=logical_order?sourceLength:-1;
            int aggregation_possible = 1;
            UChar prev = 0;
            UChar prevLink, currLink = 0;
            int newSourceLength = 0;
            tempsource = (UChar *)uprv_malloc(2*sourceLength*U_SIZEOF_UCHAR);
            if(tempsource == NULL) {
                *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
                return 0;
            }

            while ((i+=step) != end) {
                prevLink = currLink;
                currLink = getLink(source[i]);
                if (aggregate_tashkeel && ((prevLink|currLink)&USHAPE_COMBINE) == USHAPE_COMBINE && aggregation_possible) {
                    aggregation_possible = 0;
                    tempsource[j] = (prev<source[i]?prev:source[i])-0x064C+0xFC5E;
                    currLink = getLink(tempsource[j]);
                } else {
                    aggregation_possible = 1;
                    tempsource[j+=step] = source[i];
                    prev = source[i];
                    newSourceLength++;
                }
            }
            source = tempsource+(logical_order?0:j);
            sourceLength = newSourceLength;
        }

        /* calculate destination size */
        /* TODO: do we ever need to do this pure preflighting? */
        if(((options&U_SHAPE_LAMALEF_MASK)==U_SHAPE_LAMALEF_RESIZE) ||
           ((options&U_SHAPE_TASHKEEL_MASK)==U_SHAPE_TASHKEEL_RESIZE)) {
            outputSize=calculateSize(source,sourceLength,destCapacity,options);
        } else {
            outputSize=sourceLength;
        }

        if(outputSize>destCapacity) {
            *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
                if (tempsource != NULL) uprv_free(tempsource);
            return outputSize;
        }

        /*
         * need a temporary buffer of size max(outputSize, sourceLength)
         * because at first we copy source->temp
         */
        if(sourceLength>outputSize) {
            outputSize=sourceLength;
        }

        /* Start of Arabic letter shaping part */
        if(outputSize<=UPRV_LENGTHOF(buffer)) {
            outputSize=UPRV_LENGTHOF(buffer);
            tempbuffer=buffer;
        } else {
            tempbuffer = (UChar *)uprv_malloc(outputSize*U_SIZEOF_UCHAR);

            /*Test for NULL*/
            if(tempbuffer == NULL) {
                *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
                if (tempsource != NULL) uprv_free(tempsource);
                return 0;
            }
        }
        u_memcpy(tempbuffer, source, sourceLength);
        if (tempsource != NULL){
            uprv_free(tempsource);
        }

        if(sourceLength<outputSize) {
            uprv_memset(tempbuffer+sourceLength, 0, (outputSize-sourceLength)*U_SIZEOF_UCHAR);
        }

        if((options&U_SHAPE_TEXT_DIRECTION_MASK) == U_SHAPE_TEXT_DIRECTION_LOGICAL) {
            countSpaces(tempbuffer,sourceLength,options,&spacesCountl,&spacesCountr);
            invertBuffer(tempbuffer,sourceLength,options,spacesCountl,spacesCountr);
        }

        if((options&U_SHAPE_TEXT_DIRECTION_MASK) == U_SHAPE_TEXT_DIRECTION_VISUAL_LTR) {
            if((options&U_SHAPE_SPACES_RELATIVE_TO_TEXT_MASK) == U_SHAPE_SPACES_RELATIVE_TO_TEXT_BEGIN_END) {
                shapeVars.spacesRelativeToTextBeginEnd = 1;
                shapeVars.uShapeLamalefBegin = U_SHAPE_LAMALEF_END;
                shapeVars.uShapeLamalefEnd = U_SHAPE_LAMALEF_BEGIN;
                shapeVars.uShapeTashkeelBegin = U_SHAPE_TASHKEEL_END;
                shapeVars.uShapeTashkeelEnd = U_SHAPE_TASHKEEL_BEGIN;
            }
        }

        switch(options&U_SHAPE_LETTERS_MASK) {
        case U_SHAPE_LETTERS_SHAPE :
             if( (options&U_SHAPE_TASHKEEL_MASK)> 0
                 && ((options&U_SHAPE_TASHKEEL_MASK) !=U_SHAPE_TASHKEEL_REPLACE_BY_TATWEEL)) {
                /* Call the shaping function with tashkeel flag == 2 for removal of tashkeel */
                destLength = shapeUnicode(tempbuffer,sourceLength,destCapacity,options,pErrorCode,2,shapeVars);
             }else {
                /* default Call the shaping function with tashkeel flag == 1 */
                destLength = shapeUnicode(tempbuffer,sourceLength,destCapacity,options,pErrorCode,1,shapeVars);

                /*After shaping text check if user wants to remove tashkeel and replace it with tatweel*/
                if( (options&U_SHAPE_TASHKEEL_MASK) == U_SHAPE_TASHKEEL_REPLACE_BY_TATWEEL){
                  destLength = handleTashkeelWithTatweel(tempbuffer,destLength,destCapacity,options,pErrorCode);
                }
            }
            break;
        case U_SHAPE_LETTERS_SHAPE_TASHKEEL_ISOLATED :
            /* Call the shaping function with tashkeel flag == 0 */
            destLength = shapeUnicode(tempbuffer,sourceLength,destCapacity,options,pErrorCode,0,shapeVars);
            break;

        case U_SHAPE_LETTERS_UNSHAPE :
            /* Call the deshaping function */
            destLength = deShapeUnicode(tempbuffer,sourceLength,destCapacity,options,pErrorCode,shapeVars);
            break;
        default :
            /* will never occur because of validity checks above */
            destLength = 0;
            break;
        }

        /*
         * TODO: (markus 2002aug01)
         * For as long as we always preflight the outputSize above
         * we should U_ASSERT(outputSize==destLength)
         * except for the adjustment above before the tempbuffer allocation
         */

        if((options&U_SHAPE_TEXT_DIRECTION_MASK) == U_SHAPE_TEXT_DIRECTION_LOGICAL) {
            countSpaces(tempbuffer,destLength,options,&spacesCountl,&spacesCountr);
            invertBuffer(tempbuffer,destLength,options,spacesCountl,spacesCountr);
        }
        u_memcpy(dest, tempbuffer, uprv_min(destLength, destCapacity));

        if(tempbuffer!=buffer) {
            uprv_free(tempbuffer);
        }

        if(destLength>destCapacity) {
            *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
            return destLength;
        }

        /* End of Arabic letter shaping part */
    } else {
        /*
         * No letter shaping:
         * just make sure the destination is large enough and copy the string.
         */
        if(destCapacity<sourceLength) {
            /* this catches preflighting, too */
            *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
            return sourceLength;
        }
        u_memcpy(dest, source, sourceLength);
        destLength=sourceLength;
    }

    /*
     * Perform number shaping.
     * With UTF-16 or UTF-32, the length of the string is constant.
     * The easiest way to do this is to operate on the destination and
     * "shape" the digits in-place.
     */
    if((options&U_SHAPE_DIGITS_MASK)!=U_SHAPE_DIGITS_NOOP) {
        UChar digitBase;
        int32_t i;

        /* select the requested digit group */
        switch(options&U_SHAPE_DIGIT_TYPE_MASK) {
        case U_SHAPE_DIGIT_TYPE_AN:
            digitBase=0x660; /* Unicode: "Arabic-Indic digits" */
            break;
        case U_SHAPE_DIGIT_TYPE_AN_EXTENDED:
            digitBase=0x6f0; /* Unicode: "Eastern Arabic-Indic digits (Persian and Urdu)" */
            break;
        default:
            /* will never occur because of validity checks above */
            digitBase=0;
            break;
        }

        /* perform the requested operation */
        switch(options&U_SHAPE_DIGITS_MASK) {
        case U_SHAPE_DIGITS_EN2AN:
            /* add (digitBase-'0') to each European (ASCII) digit code point */
            digitBase-=0x30;
            for(i=0; i<destLength; ++i) {
                if(((uint32_t)dest[i]-0x30)<10) {
                    dest[i]+=digitBase;
                }
            }
            break;
        case U_SHAPE_DIGITS_AN2EN:
            /* subtract (digitBase-'0') from each Arabic digit code point */
            for(i=0; i<destLength; ++i) {
                if(((uint32_t)dest[i]-(uint32_t)digitBase)<10) {
                    dest[i]-=digitBase-0x30;
                }
            }
            break;
        case U_SHAPE_DIGITS_ALEN2AN_INIT_LR:
            _shapeToArabicDigitsWithContext(dest, destLength,
                                            digitBase,
                                            (UBool)((options&U_SHAPE_TEXT_DIRECTION_MASK)==U_SHAPE_TEXT_DIRECTION_LOGICAL),
                                            FALSE);
            break;
        case U_SHAPE_DIGITS_ALEN2AN_INIT_AL:
            _shapeToArabicDigitsWithContext(dest, destLength,
                                            digitBase,
                                            (UBool)((options&U_SHAPE_TEXT_DIRECTION_MASK)==U_SHAPE_TEXT_DIRECTION_LOGICAL),
                                            TRUE);
            break;
        default:
            /* will never occur because of validity checks above */
            break;
        }
    }

    return u_terminateUChars(dest, destCapacity, destLength, pErrorCode);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 *
 *   Copyright (C) 2003-2016, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 *
 *******************************************************************************
 *   file name:  usprep.cpp
 *   encoding:   UTF-8
 *   tab size:   8 (not used)
 *   indentation:4
 *
 *   created on: 2003jul2
 *   created by: Ram Viswanadha
 */



#if !UCONFIG_NO_IDNA

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 *
 *   Copyright (C) 2003-2014, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 *
 *******************************************************************************
 *   file name:  usprep.h
 *   encoding:   UTF-8
 *   tab size:   8 (not used)
 *   indentation:4
 *
 *   created on: 2003jul2
 *   created by: Ram Viswanadha
 */


/**
 * \file 
 * \brief C API: Implements the StringPrep algorithm.
 */




/**
 *
 * StringPrep API implements the StingPrep framework as described by RFC 3454.
 * StringPrep prepares Unicode strings for use in network protocols.
 * Profiles of StingPrep are set of rules and data according to with the
 * Unicode Strings are prepared. Each profiles contains tables which describe
 * how a code point should be treated. The tables are broadly classified into
 * <ul>
 *     <li> Unassigned Table: Contains code points that are unassigned 
 *          in the Unicode Version supported by StringPrep. Currently 
 *          RFC 3454 supports Unicode 3.2. </li>
 *     <li> Prohibited Table: Contains code points that are prohibited from
 *          the output of the StringPrep processing function. </li>
 *     <li> Mapping Table: Contains code points that are deleted from the output or case mapped. </li>
 * </ul>
 * 
 * The procedure for preparing Unicode strings:
 * <ol>
 *      <li> Map: For each character in the input, check if it has a mapping
 *           and, if so, replace it with its mapping. </li>
 *      <li> Normalize: Possibly normalize the result of step 1 using Unicode
 *           normalization. </li>
 *      <li> Prohibit: Check for any characters that are not allowed in the
 *        output.  If any are found, return an error.</li>
 *      <li> Check bidi: Possibly check for right-to-left characters, and if
 *           any are found, make sure that the whole string satisfies the
 *           requirements for bidirectional strings.  If the string does not
 *           satisfy the requirements for bidirectional strings, return an
 *           error.  </li>
 * </ol>
 * @author Ram Viswanadha
 */
#if !UCONFIG_NO_IDNA



/**
 * The StringPrep profile
 * @stable ICU 2.8
 */
typedef struct UStringPrepProfile UStringPrepProfile;


/** 
 * Option to prohibit processing of unassigned code points in the input
 * 
 * @see  usprep_prepare
 * @stable ICU 2.8
 */
#define USPREP_DEFAULT 0x0000

/** 
 * Option to allow processing of unassigned code points in the input
 * 
 * @see  usprep_prepare
 * @stable ICU 2.8
 */
#define USPREP_ALLOW_UNASSIGNED 0x0001

/**
 * enums for the standard stringprep profile types
 * supported by usprep_openByType.
 * @see usprep_openByType
 * @stable ICU 4.2
 */
typedef enum UStringPrepProfileType {
    /**
     * RFC3491 Nameprep
     * @stable ICU 4.2
     */
    USPREP_RFC3491_NAMEPREP,
    /**
     * RFC3530 nfs4_cs_prep
     * @stable ICU 4.2
     */
	USPREP_RFC3530_NFS4_CS_PREP,
    /**
     * RFC3530 nfs4_cs_prep with case insensitive option
     * @stable ICU 4.2
     */
	USPREP_RFC3530_NFS4_CS_PREP_CI,
    /**
     * RFC3530 nfs4_cis_prep
     * @stable ICU 4.2
     */
	USPREP_RFC3530_NFS4_CIS_PREP,
    /**
     * RFC3530 nfs4_mixed_prep for prefix
     * @stable ICU 4.2
     */
	USPREP_RFC3530_NFS4_MIXED_PREP_PREFIX,
    /**
     * RFC3530 nfs4_mixed_prep for suffix
     * @stable ICU 4.2
     */
	USPREP_RFC3530_NFS4_MIXED_PREP_SUFFIX,
    /**
     * RFC3722 iSCSI
     * @stable ICU 4.2
     */
	USPREP_RFC3722_ISCSI,
    /**
     * RFC3920 XMPP Nodeprep
     * @stable ICU 4.2
     */
	USPREP_RFC3920_NODEPREP,
    /**
     * RFC3920 XMPP Resourceprep
     * @stable ICU 4.2
     */
	USPREP_RFC3920_RESOURCEPREP,
    /**
     * RFC4011 Policy MIB Stringprep
     * @stable ICU 4.2
     */
	USPREP_RFC4011_MIB,
    /**
     * RFC4013 SASLprep
     * @stable ICU 4.2
     */
    USPREP_RFC4013_SASLPREP,
    /**
     * RFC4505 trace
     * @stable ICU 4.2
     */
	USPREP_RFC4505_TRACE,
    /**
     * RFC4518 LDAP
     * @stable ICU 4.2
     */
	USPREP_RFC4518_LDAP,
    /**
     * RFC4518 LDAP for case ignore, numeric and stored prefix
     * matching rules
     * @stable ICU 4.2
     */
	USPREP_RFC4518_LDAP_CI
} UStringPrepProfileType;

/**
 * Creates a StringPrep profile from the data file.
 *
 * @param path      string containing the full path pointing to the directory
 *                  where the profile reside followed by the package name
 *                  e.g. "/usr/resource/my_app/profiles/mydata" on a Unix system.
 *                  if NULL, ICU default data files will be used.
 * @param fileName  name of the profile file to be opened
 * @param status    ICU error code in/out parameter. Must not be NULL.
 *                  Must fulfill U_SUCCESS before the function call.
 * @return Pointer to UStringPrepProfile that is opened. Should be closed by
 * calling usprep_close()
 * @see usprep_close()
 * @stable ICU 2.8
 */
U_STABLE UStringPrepProfile* U_EXPORT2
usprep_open(const char* path, 
            const char* fileName,
            UErrorCode* status);

/**
 * Creates a StringPrep profile for the specified profile type.
 *
 * @param type		The profile type
 * @param status    ICU error code in/out parameter. Must not be NULL.
 *                  Must fulfill U_SUCCESS before the function call.
 * @return          Pointer to UStringPrepProfile that is opened. Should be closed by
 *                  calling usprep_close()
 * @see usprep_close()
 * @stable ICU 4.2
 */
U_STABLE UStringPrepProfile* U_EXPORT2
usprep_openByType(UStringPrepProfileType type,
				  UErrorCode* status);

/**
 * Closes the profile
 * @param profile The profile to close
 * @stable ICU 2.8
 */
U_STABLE void U_EXPORT2
usprep_close(UStringPrepProfile* profile);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUStringPrepProfilePointer
 * "Smart pointer" class, closes a UStringPrepProfile via usprep_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUStringPrepProfilePointer, UStringPrepProfile, usprep_close);

U_NAMESPACE_END

#endif

/**
 * Prepare the input buffer for use in applications with the given profile. This operation maps, normalizes(NFKC),
 * checks for prohibited and BiDi characters in the order defined by RFC 3454
 * depending on the options specified in the profile.
 *
 * @param prep          The profile to use 
 * @param src           Pointer to UChar buffer containing the string to prepare
 * @param srcLength     Number of characters in the source string
 * @param dest          Pointer to the destination buffer to receive the output
 * @param destCapacity  The capacity of destination array
 * @param options       A bit set of options:
 *
 *  - USPREP_DEFAULT            Prohibit processing of unassigned code points in the input
 *
 *  - USPREP_ALLOW_UNASSIGNED   Treat the unassigned code points are in the input 
 *                              as normal Unicode code points.
 *
 * @param parseError        Pointer to UParseError struct to receive information on position 
 *                          of error if an error is encountered. Can be NULL.
 * @param status            ICU in/out error code parameter.
 *                          U_INVALID_CHAR_FOUND if src contains
 *                          unmatched single surrogates.
 *                          U_INDEX_OUTOFBOUNDS_ERROR if src contains
 *                          too many code points.
 *                          U_BUFFER_OVERFLOW_ERROR if destCapacity is not enough
 * @return The number of UChars in the destination buffer
 * @stable ICU 2.8
 */

U_STABLE int32_t U_EXPORT2
usprep_prepare(   const UStringPrepProfile* prep,
                  const UChar* src, int32_t srcLength, 
                  UChar* dest, int32_t destCapacity,
                  int32_t options,
                  UParseError* parseError,
                  UErrorCode* status );


#endif /* #if !UCONFIG_NO_IDNA */









// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 *
 *   Copyright (C) 2003-2006, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 *
 *******************************************************************************
 *   file name:  sprpimpl.h
 *   encoding:   UTF-8
 *   tab size:   8 (not used)
 *   indentation:4
 *
 *   created on: 2003feb1
 *   created by: Ram Viswanadha
 */




#if !UCONFIG_NO_IDNA









#define _SPREP_DATA_TYPE "spp"

enum UStringPrepType{
    USPREP_UNASSIGNED           = 0x0000 ,
    USPREP_MAP                  = 0x0001 ,
    USPREP_PROHIBITED           = 0x0002 , 
    USPREP_DELETE               = 0x0003 ,
    USPREP_TYPE_LIMIT           = 0x0004  
};

typedef enum UStringPrepType UStringPrepType;

#ifdef USPREP_TYPE_NAMES_ARRAY
static const char* usprepTypeNames[] ={
    "UNASSIGNED" ,          
    "MAP" , 
    "PROHIBITED" ,        
    "DELETE",
    "TYPE_LIMIT" 
};
#endif

enum{
    _SPREP_NORMALIZATION_ON = 0x0001,
    _SPREP_CHECK_BIDI_ON    = 0x0002
};

enum{
    _SPREP_TYPE_THRESHOLD       = 0xFFF0,
    _SPREP_MAX_INDEX_VALUE      = 0x3FBF,   /*16139*/ 
    _SPREP_MAX_INDEX_TOP_LENGTH = 0x0003
};

/* indexes[] value names */
enum {
    _SPREP_INDEX_TRIE_SIZE                  = 0, /* number of bytes in StringPrep trie */
    _SPREP_INDEX_MAPPING_DATA_SIZE          = 1, /* The array that contains the mapping   */
    _SPREP_NORM_CORRECTNS_LAST_UNI_VERSION  = 2, /* The index of Unicode version of last entry in NormalizationCorrections.txt */ 
    _SPREP_ONE_UCHAR_MAPPING_INDEX_START    = 3, /* The starting index of 1 UChar mapping index in the mapping data array */
    _SPREP_TWO_UCHARS_MAPPING_INDEX_START   = 4, /* The starting index of 2 UChars mapping index in the mapping data array */
    _SPREP_THREE_UCHARS_MAPPING_INDEX_START = 5, /* The starting index of 3 UChars mapping index in the mapping data array */
    _SPREP_FOUR_UCHARS_MAPPING_INDEX_START  = 6, /* The starting index of 4 UChars mapping index in the mapping data array */
    _SPREP_OPTIONS                          = 7, /* Bit set of options to turn on in the profile */
    _SPREP_INDEX_TOP=16                          /* changing this requires a new formatVersion */
};

typedef struct UStringPrepKey UStringPrepKey;


struct UStringPrepKey{
    char* name;
    char* path;
};

struct UStringPrepProfile{
    int32_t indexes[_SPREP_INDEX_TOP];
    UTrie sprepTrie;
    const uint16_t* mappingData;
    UDataMemory* sprepData;
    int32_t refCount;
    UBool isDataLoaded;
    UBool doNFKC;
    UBool checkBiDi;
};

/**
 * Helper function for populating the UParseError struct
 * @internal
 */
U_CAPI void U_EXPORT2
uprv_syntaxError(const UChar* rules, 
                 int32_t pos,
                 int32_t rulesLen,
                 UParseError* parseError);


/**
 * Swap StringPrep .spp profile data. See udataswp.h.
 * @internal
 */
U_CAPI int32_t U_EXPORT2
usprep_swap(const UDataSwapper *ds,
            const void *inData, int32_t length, void *outData,
            UErrorCode *pErrorCode);

#endif /* #if !UCONFIG_NO_IDNA */










U_NAMESPACE_USE

U_CDECL_BEGIN

/*
Static cache for already opened StringPrep profiles
*/
static UHashtable *SHARED_DATA_HASHTABLE = NULL;
static icu::UInitOnce gSharedDataInitOnce = U_INITONCE_INITIALIZER;

static UMutex usprepMutex;
/* format version of spp file */
//static uint8_t formatVersion[4]={ 0, 0, 0, 0 };

/* the Unicode version of the sprep data */
static UVersionInfo dataVersion={ 0, 0, 0, 0 };

/* Profile names must be aligned to UStringPrepProfileType */
static const char * const PROFILE_NAMES[] = {
    "rfc3491",      /* USPREP_RFC3491_NAMEPREP */
    "rfc3530cs",    /* USPREP_RFC3530_NFS4_CS_PREP */
    "rfc3530csci",  /* USPREP_RFC3530_NFS4_CS_PREP_CI */
    "rfc3491",      /* USPREP_RFC3530_NSF4_CIS_PREP */
    "rfc3530mixp",  /* USPREP_RFC3530_NSF4_MIXED_PREP_PREFIX */
    "rfc3491",      /* USPREP_RFC3530_NSF4_MIXED_PREP_SUFFIX */
    "rfc3722",      /* USPREP_RFC3722_ISCSI */
    "rfc3920node",  /* USPREP_RFC3920_NODEPREP */
    "rfc3920res",   /* USPREP_RFC3920_RESOURCEPREP */
    "rfc4011",      /* USPREP_RFC4011_MIB */
    "rfc4013",      /* USPREP_RFC4013_SASLPREP */
    "rfc4505",      /* USPREP_RFC4505_TRACE */
    "rfc4518",      /* USPREP_RFC4518_LDAP */
    "rfc4518ci",    /* USPREP_RFC4518_LDAP_CI */
};

static UBool U_CALLCONV
isSPrepAcceptable(void * /* context */,
             const char * /* type */, 
             const char * /* name */,
             const UDataInfo *pInfo) {
    if(
        pInfo->size>=20 &&
        pInfo->isBigEndian==U_IS_BIG_ENDIAN &&
        pInfo->charsetFamily==U_CHARSET_FAMILY &&
        pInfo->dataFormat[0]==0x53 &&   /* dataFormat="SPRP" */
        pInfo->dataFormat[1]==0x50 &&
        pInfo->dataFormat[2]==0x52 &&
        pInfo->dataFormat[3]==0x50 &&
        pInfo->formatVersion[0]==3 &&
        pInfo->formatVersion[2]==UTRIE_SHIFT &&
        pInfo->formatVersion[3]==UTRIE_INDEX_SHIFT
    ) {
        //uprv_memcpy(formatVersion, pInfo->formatVersion, 4);
        uprv_memcpy(dataVersion, pInfo->dataVersion, 4);
        return TRUE;
    } else {
        return FALSE;
    }
}

static int32_t U_CALLCONV
getSPrepFoldingOffset(uint32_t data) {

    return (int32_t)data;

}

/* hashes an entry  */
static int32_t U_CALLCONV 
hashEntry(const UHashTok parm) {
    UStringPrepKey *b = (UStringPrepKey *)parm.pointer;
    UHashTok namekey, pathkey;
    namekey.pointer = b->name;
    pathkey.pointer = b->path;
    uint32_t unsignedHash = static_cast<uint32_t>(uhash_hashChars(namekey)) +
            37u * static_cast<uint32_t>(uhash_hashChars(pathkey));
    return static_cast<int32_t>(unsignedHash);
}

/* compares two entries */
static UBool U_CALLCONV 
compareEntries(const UHashTok p1, const UHashTok p2) {
    UStringPrepKey *b1 = (UStringPrepKey *)p1.pointer;
    UStringPrepKey *b2 = (UStringPrepKey *)p2.pointer;
    UHashTok name1, name2, path1, path2;
    name1.pointer = b1->name;
    name2.pointer = b2->name;
    path1.pointer = b1->path;
    path2.pointer = b2->path;
    return ((UBool)(uhash_compareChars(name1, name2) & 
        uhash_compareChars(path1, path2)));
}

static void 
usprep_unload(UStringPrepProfile* data){
    udata_close(data->sprepData);
}

static int32_t 
usprep_internal_flushCache(UBool noRefCount){
    UStringPrepProfile *profile = NULL;
    UStringPrepKey  *key  = NULL;
    int32_t pos = UHASH_FIRST;
    int32_t deletedNum = 0;
    const UHashElement *e;

    /*
     * if shared data hasn't even been lazy evaluated yet
     * return 0
     */
    umtx_lock(&usprepMutex);
    if (SHARED_DATA_HASHTABLE == NULL) {
        umtx_unlock(&usprepMutex);
        return 0;
    }

    /*creates an enumeration to iterate through every element in the table */
    while ((e = uhash_nextElement(SHARED_DATA_HASHTABLE, &pos)) != NULL)
    {
        profile = (UStringPrepProfile *) e->value.pointer;
        key  = (UStringPrepKey *) e->key.pointer;

        if ((noRefCount== FALSE && profile->refCount == 0) || 
             noRefCount== TRUE) {
            deletedNum++;
            uhash_removeElement(SHARED_DATA_HASHTABLE, e);

            /* unload the data */
            usprep_unload(profile);

            if(key->name != NULL) {
                uprv_free(key->name);
                key->name=NULL;
            }
            if(key->path != NULL) {
                uprv_free(key->path);
                key->path=NULL;
            }
            uprv_free(profile);
            uprv_free(key);
        }
       
    }
    umtx_unlock(&usprepMutex);

    return deletedNum;
}

/* Works just like ucnv_flushCache() 
static int32_t 
usprep_flushCache(){
    return usprep_internal_flushCache(FALSE);
}
*/

static UBool U_CALLCONV usprep_cleanup(void){
    if (SHARED_DATA_HASHTABLE != NULL) {
        usprep_internal_flushCache(TRUE);
        if (SHARED_DATA_HASHTABLE != NULL && uhash_count(SHARED_DATA_HASHTABLE) == 0) {
            uhash_close(SHARED_DATA_HASHTABLE);
            SHARED_DATA_HASHTABLE = NULL;
        }
    }
    gSharedDataInitOnce.reset();
    return (SHARED_DATA_HASHTABLE == NULL);
}
U_CDECL_END


/** Initializes the cache for resources */
static void U_CALLCONV
createCache(UErrorCode &status) {
    SHARED_DATA_HASHTABLE = uhash_open(hashEntry, compareEntries, NULL, &status);
    if (U_FAILURE(status)) {
        SHARED_DATA_HASHTABLE = NULL;
    }
    ucln_common_registerCleanup(UCLN_COMMON_USPREP, usprep_cleanup);
}

static void 
initCache(UErrorCode *status) {
    umtx_initOnce(gSharedDataInitOnce, &createCache, *status);
}

static UBool U_CALLCONV
loadData(UStringPrepProfile* profile, 
         const char* path, 
         const char* name, 
         const char* type, 
         UErrorCode* errorCode) {
    /* load Unicode SPREP data from file */    
    UTrie _sprepTrie={ 0,0,0,0,0,0,0 };
    UDataMemory *dataMemory;
    const int32_t *p=NULL;
    const uint8_t *pb;
    UVersionInfo normUnicodeVersion;
    int32_t normUniVer, sprepUniVer, normCorrVer;

    if(errorCode==NULL || U_FAILURE(*errorCode)) {
        return 0;
    }

    /* open the data outside the mutex block */
    //TODO: change the path
    dataMemory=udata_openChoice(path, type, name, isSPrepAcceptable, NULL, errorCode);
    if(U_FAILURE(*errorCode)) {
        return FALSE;
    }

    p=(const int32_t *)udata_getMemory(dataMemory);
    pb=(const uint8_t *)(p+_SPREP_INDEX_TOP);
    utrie_unserialize(&_sprepTrie, pb, p[_SPREP_INDEX_TRIE_SIZE], errorCode);
    _sprepTrie.getFoldingOffset=getSPrepFoldingOffset;


    if(U_FAILURE(*errorCode)) {
        udata_close(dataMemory);
        return FALSE;
    }

    /* in the mutex block, set the data for this process */
    umtx_lock(&usprepMutex);
    if(profile->sprepData==NULL) {
        profile->sprepData=dataMemory;
        dataMemory=NULL;
        uprv_memcpy(&profile->indexes, p, sizeof(profile->indexes));
        uprv_memcpy(&profile->sprepTrie, &_sprepTrie, sizeof(UTrie));
    } else {
        p=(const int32_t *)udata_getMemory(profile->sprepData);
    }
    umtx_unlock(&usprepMutex);
    /* initialize some variables */
    profile->mappingData=(uint16_t *)((uint8_t *)(p+_SPREP_INDEX_TOP)+profile->indexes[_SPREP_INDEX_TRIE_SIZE]);
    
    u_getUnicodeVersion(normUnicodeVersion);
    normUniVer = (normUnicodeVersion[0] << 24) + (normUnicodeVersion[1] << 16) + 
                 (normUnicodeVersion[2] << 8 ) + (normUnicodeVersion[3]);
    sprepUniVer = (dataVersion[0] << 24) + (dataVersion[1] << 16) + 
                  (dataVersion[2] << 8 ) + (dataVersion[3]);
    normCorrVer = profile->indexes[_SPREP_NORM_CORRECTNS_LAST_UNI_VERSION];
    
    if(U_FAILURE(*errorCode)){
        udata_close(dataMemory);
        return FALSE;
    }
    if( normUniVer < sprepUniVer && /* the Unicode version of SPREP file must be less than the Unicode Vesion of the normalization data */
        normUniVer < normCorrVer && /* the Unicode version of the NormalizationCorrections.txt file should be less than the Unicode Vesion of the normalization data */
        ((profile->indexes[_SPREP_OPTIONS] & _SPREP_NORMALIZATION_ON) > 0) /* normalization turned on*/
      ){
        *errorCode = U_INVALID_FORMAT_ERROR;
        udata_close(dataMemory);
        return FALSE;
    }
    profile->isDataLoaded = TRUE;

    /* if a different thread set it first, then close the extra data */
    if(dataMemory!=NULL) {
        udata_close(dataMemory); /* NULL if it was set correctly */
    }


    return profile->isDataLoaded;
}

static UStringPrepProfile* 
usprep_getProfile(const char* path, 
                  const char* name,
                  UErrorCode *status){

    UStringPrepProfile* profile = NULL;

    initCache(status);

    if(U_FAILURE(*status)){
        return NULL;
    }

    UStringPrepKey stackKey;
    /* 
     * const is cast way to save malloc, strcpy and free calls 
     * we use the passed in pointers for fetching the data from the 
     * hash table which is safe
     */
    stackKey.name = (char*) name;
    stackKey.path = (char*) path;

    /* fetch the data from the cache */
    umtx_lock(&usprepMutex);
    profile = (UStringPrepProfile*) (uhash_get(SHARED_DATA_HASHTABLE,&stackKey));
    if(profile != NULL) {
        profile->refCount++;
    }
    umtx_unlock(&usprepMutex);
    
    if(profile == NULL) {
        /* else load the data and put the data in the cache */
        LocalMemory<UStringPrepProfile> newProfile;
        if(newProfile.allocateInsteadAndReset() == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }

        /* load the data */
        if(!loadData(newProfile.getAlias(), path, name, _SPREP_DATA_TYPE, status) || U_FAILURE(*status) ){
            return NULL;
        }

        /* get the options */
        newProfile->doNFKC = (UBool)((newProfile->indexes[_SPREP_OPTIONS] & _SPREP_NORMALIZATION_ON) > 0);
        newProfile->checkBiDi = (UBool)((newProfile->indexes[_SPREP_OPTIONS] & _SPREP_CHECK_BIDI_ON) > 0);

        LocalMemory<UStringPrepKey> key;
        LocalMemory<char> keyName;
        LocalMemory<char> keyPath;
        if( key.allocateInsteadAndReset() == NULL ||
            keyName.allocateInsteadAndCopy(static_cast<int32_t>(uprv_strlen(name)+1)) == NULL ||
            (path != NULL &&
             keyPath.allocateInsteadAndCopy(static_cast<int32_t>(uprv_strlen(path)+1)) == NULL)
         ) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            usprep_unload(newProfile.getAlias());
            return NULL;
        }

        umtx_lock(&usprepMutex);
        // If another thread already inserted the same key/value, refcount and cleanup our thread data
        profile = (UStringPrepProfile*) (uhash_get(SHARED_DATA_HASHTABLE,&stackKey));
        if(profile != NULL) {
            profile->refCount++;
            usprep_unload(newProfile.getAlias());
        }
        else {
            /* initialize the key members */
            key->name = keyName.orphan();
            uprv_strcpy(key->name, name);
            if(path != NULL){
                key->path = keyPath.orphan();
                uprv_strcpy(key->path, path);
            }        
            profile = newProfile.orphan();
    
            /* add the data object to the cache */
            profile->refCount = 1;
            uhash_put(SHARED_DATA_HASHTABLE, key.orphan(), profile, status);
        }
        umtx_unlock(&usprepMutex);
    }

    return profile;
}

U_CAPI UStringPrepProfile* U_EXPORT2
usprep_open(const char* path, 
            const char* name,
            UErrorCode* status){

    if(status == NULL || U_FAILURE(*status)){
        return NULL;
    }
       
    /* initialize the profile struct members */
    return usprep_getProfile(path,name,status);
}

U_CAPI UStringPrepProfile* U_EXPORT2
usprep_openByType(UStringPrepProfileType type,
				  UErrorCode* status) {
    if(status == NULL || U_FAILURE(*status)){
        return NULL;
    }
    int32_t index = (int32_t)type;
    if (index < 0 || index >= UPRV_LENGTHOF(PROFILE_NAMES)) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    return usprep_open(NULL, PROFILE_NAMES[index], status);
}

U_CAPI void U_EXPORT2
usprep_close(UStringPrepProfile* profile){
    if(profile==NULL){
        return;
    }

    umtx_lock(&usprepMutex);
    /* decrement the ref count*/
    if(profile->refCount > 0){
        profile->refCount--;
    }
    umtx_unlock(&usprepMutex);
    
}

U_CFUNC void 
uprv_syntaxError(const UChar* rules, 
                 int32_t pos,
                 int32_t rulesLen,
                 UParseError* parseError){
    if(parseError == NULL){
        return;
    }
    parseError->offset = pos;
    parseError->line = 0 ; // we are not using line numbers 
    
    // for pre-context
    int32_t start = (pos < U_PARSE_CONTEXT_LEN)? 0 : (pos - (U_PARSE_CONTEXT_LEN-1));
    int32_t limit = pos;
    
    u_memcpy(parseError->preContext,rules+start,limit-start);
    //null terminate the buffer
    parseError->preContext[limit-start] = 0;
    
    // for post-context; include error rules[pos]
    start = pos;
    limit = start + (U_PARSE_CONTEXT_LEN-1);
    if (limit > rulesLen) {
        limit = rulesLen;
    }
    if (start < rulesLen) {
        u_memcpy(parseError->postContext,rules+start,limit-start);
    }
    //null terminate the buffer
    parseError->postContext[limit-start]= 0;
}


static inline UStringPrepType
getValues(uint16_t trieWord, int16_t& value, UBool& isIndex){

    UStringPrepType type;
    if(trieWord == 0){
        /* 
         * Initial value stored in the mapping table 
         * just return USPREP_TYPE_LIMIT .. so that
         * the source codepoint is copied to the destination
         */
        type = USPREP_TYPE_LIMIT;
        isIndex =FALSE;
        value = 0;
    }else if(trieWord >= _SPREP_TYPE_THRESHOLD){
        type = (UStringPrepType) (trieWord - _SPREP_TYPE_THRESHOLD);
        isIndex =FALSE;
        value = 0;
    }else{
        /* get the type */
        type = USPREP_MAP;
        /* ascertain if the value is index or delta */
        if(trieWord & 0x02){
            isIndex = TRUE;
            value = trieWord  >> 2; //mask off the lower 2 bits and shift
        }else{
            isIndex = FALSE;
            value = (int16_t)trieWord;
            value =  (value >> 2);
        }
 
        if((trieWord>>2) == _SPREP_MAX_INDEX_VALUE){
            type = USPREP_DELETE;
            isIndex =FALSE;
            value = 0;
        }
    }
    return type;
}

// TODO: change to writing to UnicodeString not UChar *
static int32_t 
usprep_map(  const UStringPrepProfile* profile, 
             const UChar* src, int32_t srcLength, 
             UChar* dest, int32_t destCapacity,
             int32_t options,
             UParseError* parseError,
             UErrorCode* status ){
    
    uint16_t result;
    int32_t destIndex=0;
    int32_t srcIndex;
    UBool allowUnassigned = (UBool) ((options & USPREP_ALLOW_UNASSIGNED)>0);
    UStringPrepType type;
    int16_t value;
    UBool isIndex;
    const int32_t* indexes = profile->indexes;

    // no error checking the caller check for error and arguments
    // no string length check the caller finds out the string length

    for(srcIndex=0;srcIndex<srcLength;){
        UChar32 ch;

        U16_NEXT(src,srcIndex,srcLength,ch);
        
        result=0;

        UTRIE_GET16(&profile->sprepTrie,ch,result);
        
        type = getValues(result, value, isIndex);

        // check if the source codepoint is unassigned
        if(type == USPREP_UNASSIGNED && allowUnassigned == FALSE){

            uprv_syntaxError(src,srcIndex-U16_LENGTH(ch), srcLength,parseError);
            *status = U_STRINGPREP_UNASSIGNED_ERROR;
            return 0;
            
        }else if(type == USPREP_MAP){
            
            int32_t index, length;

            if(isIndex){
                index = value;
                if(index >= indexes[_SPREP_ONE_UCHAR_MAPPING_INDEX_START] &&
                         index < indexes[_SPREP_TWO_UCHARS_MAPPING_INDEX_START]){
                    length = 1;
                }else if(index >= indexes[_SPREP_TWO_UCHARS_MAPPING_INDEX_START] &&
                         index < indexes[_SPREP_THREE_UCHARS_MAPPING_INDEX_START]){
                    length = 2;
                }else if(index >= indexes[_SPREP_THREE_UCHARS_MAPPING_INDEX_START] &&
                         index < indexes[_SPREP_FOUR_UCHARS_MAPPING_INDEX_START]){
                    length = 3;
                }else{
                    length = profile->mappingData[index++];
         
                }

                /* copy mapping to destination */
                for(int32_t i=0; i< length; i++){
                    if(destIndex < destCapacity  ){
                        dest[destIndex] = profile->mappingData[index+i];
                    }
                    destIndex++; /* for pre-flighting */
                }  
                continue;
            }else{
                // subtract the delta to arrive at the code point
                ch -= value;
            }

        }else if(type==USPREP_DELETE){
             // just consume the codepoint and contine
            continue;
        }
        //copy the code point into destination
        if(ch <= 0xFFFF){
            if(destIndex < destCapacity ){
                dest[destIndex] = (UChar)ch;
            }
            destIndex++;
        }else{
            if(destIndex+1 < destCapacity ){
                dest[destIndex]   = U16_LEAD(ch);
                dest[destIndex+1] = U16_TRAIL(ch);
            }
            destIndex +=2;
        }
       
    }
        
    return u_terminateUChars(dest, destCapacity, destIndex, status);
}

/*
   1) Map -- For each character in the input, check if it has a mapping
      and, if so, replace it with its mapping.  

   2) Normalize -- Possibly normalize the result of step 1 using Unicode
      normalization. 

   3) Prohibit -- Check for any characters that are not allowed in the
      output.  If any are found, return an error.  

   4) Check bidi -- Possibly check for right-to-left characters, and if
      any are found, make sure that the whole string satisfies the
      requirements for bidirectional strings.  If the string does not
      satisfy the requirements for bidirectional strings, return an
      error.  
      [Unicode3.2] defines several bidirectional categories; each character
       has one bidirectional category assigned to it.  For the purposes of
       the requirements below, an "RandALCat character" is a character that
       has Unicode bidirectional categories "R" or "AL"; an "LCat character"
       is a character that has Unicode bidirectional category "L".  Note


       that there are many characters which fall in neither of the above
       definitions; Latin digits (<U+0030> through <U+0039>) are examples of
       this because they have bidirectional category "EN".

       In any profile that specifies bidirectional character handling, all
       three of the following requirements MUST be met:

       1) The characters in section 5.8 MUST be prohibited.

       2) If a string contains any RandALCat character, the string MUST NOT
          contain any LCat character.

       3) If a string contains any RandALCat character, a RandALCat
          character MUST be the first character of the string, and a
          RandALCat character MUST be the last character of the string.
*/
U_CAPI int32_t U_EXPORT2
usprep_prepare(   const UStringPrepProfile* profile,
                  const UChar* src, int32_t srcLength, 
                  UChar* dest, int32_t destCapacity,
                  int32_t options,
                  UParseError* parseError,
                  UErrorCode* status ){

    // check error status
    if(U_FAILURE(*status)){
        return 0;
    }

    //check arguments
    if(profile==NULL ||
            (src==NULL ? srcLength!=0 : srcLength<-1) ||
            (dest==NULL ? destCapacity!=0 : destCapacity<0)) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    //get the string length
    if(srcLength < 0){
        srcLength = u_strlen(src);
    }
    // map
    UnicodeString s1;
    UChar *b1 = s1.getBuffer(srcLength);
    if(b1==NULL){
        *status = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    int32_t b1Len = usprep_map(profile, src, srcLength,
                               b1, s1.getCapacity(), options, parseError, status);
    s1.releaseBuffer(U_SUCCESS(*status) ? b1Len : 0);

    if(*status == U_BUFFER_OVERFLOW_ERROR){
        // redo processing of string
        /* we do not have enough room so grow the buffer*/
        b1 = s1.getBuffer(b1Len);
        if(b1==NULL){
            *status = U_MEMORY_ALLOCATION_ERROR;
            return 0;
        }

        *status = U_ZERO_ERROR; // reset error
        b1Len = usprep_map(profile, src, srcLength,
                           b1, s1.getCapacity(), options, parseError, status);
        s1.releaseBuffer(U_SUCCESS(*status) ? b1Len : 0);
    }
    if(U_FAILURE(*status)){
        return 0;
    }

    // normalize
    UnicodeString s2;
    if(profile->doNFKC){
        const Normalizer2 *n2 = Normalizer2::getNFKCInstance(*status);
        FilteredNormalizer2 fn2(*n2, *uniset_getUnicode32Instance(*status));
        if(U_FAILURE(*status)){
            return 0;
        }
        fn2.normalize(s1, s2, *status);
    }else{
        s2.fastCopyFrom(s1);
    }
    if(U_FAILURE(*status)){
        return 0;
    }

    // Prohibit and checkBiDi in one pass
    const UChar *b2 = s2.getBuffer();
    int32_t b2Len = s2.length();
    UCharDirection direction=U_CHAR_DIRECTION_COUNT, firstCharDir=U_CHAR_DIRECTION_COUNT;
    UBool leftToRight=FALSE, rightToLeft=FALSE;
    int32_t rtlPos =-1, ltrPos =-1;

    for(int32_t b2Index=0; b2Index<b2Len;){
        UChar32 ch = 0;
        U16_NEXT(b2, b2Index, b2Len, ch);

        uint16_t result;
        UTRIE_GET16(&profile->sprepTrie,ch,result);

        int16_t value;
        UBool isIndex;
        UStringPrepType type = getValues(result, value, isIndex);

        if( type == USPREP_PROHIBITED || 
            ((result < _SPREP_TYPE_THRESHOLD) && (result & 0x01) /* first bit says it the code point is prohibited*/)
           ){
            *status = U_STRINGPREP_PROHIBITED_ERROR;
            uprv_syntaxError(b2, b2Index-U16_LENGTH(ch), b2Len, parseError);
            return 0;
        }

        if(profile->checkBiDi) {
            direction = ubidi_getClass(ch);
            if(firstCharDir == U_CHAR_DIRECTION_COUNT){
                firstCharDir = direction;
            }
            if(direction == U_LEFT_TO_RIGHT){
                leftToRight = TRUE;
                ltrPos = b2Index-1;
            }
            if(direction == U_RIGHT_TO_LEFT || direction == U_RIGHT_TO_LEFT_ARABIC){
                rightToLeft = TRUE;
                rtlPos = b2Index-1;
            }
        }
    }
    if(profile->checkBiDi == TRUE){
        // satisfy 2
        if( leftToRight == TRUE && rightToLeft == TRUE){
            *status = U_STRINGPREP_CHECK_BIDI_ERROR;
            uprv_syntaxError(b2,(rtlPos>ltrPos) ? rtlPos : ltrPos, b2Len, parseError);
            return 0;
        }

        //satisfy 3
        if( rightToLeft == TRUE && 
            !((firstCharDir == U_RIGHT_TO_LEFT || firstCharDir == U_RIGHT_TO_LEFT_ARABIC) &&
              (direction == U_RIGHT_TO_LEFT || direction == U_RIGHT_TO_LEFT_ARABIC))
           ){
            *status = U_STRINGPREP_CHECK_BIDI_ERROR;
            uprv_syntaxError(b2, rtlPos, b2Len, parseError);
            return FALSE;
        }
    }
    return s2.extract(dest, destCapacity, *status);
}


/* data swapping ------------------------------------------------------------ */

U_CAPI int32_t U_EXPORT2
usprep_swap(const UDataSwapper *ds,
            const void *inData, int32_t length, void *outData,
            UErrorCode *pErrorCode) {
    const UDataInfo *pInfo;
    int32_t headerSize;

    const uint8_t *inBytes;
    uint8_t *outBytes;

    const int32_t *inIndexes;
    int32_t indexes[16];

    int32_t i, offset, count, size;

    /* udata_swapDataHeader checks the arguments */
    headerSize=udata_swapDataHeader(ds, inData, length, outData, pErrorCode);
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* check data format and format version */
    pInfo=(const UDataInfo *)((const char *)inData+4);
    if(!(
        pInfo->dataFormat[0]==0x53 &&   /* dataFormat="SPRP" */
        pInfo->dataFormat[1]==0x50 &&
        pInfo->dataFormat[2]==0x52 &&
        pInfo->dataFormat[3]==0x50 &&
        pInfo->formatVersion[0]==3
    )) {
        udata_printError(ds, "usprep_swap(): data format %02x.%02x.%02x.%02x (format version %02x) is not recognized as StringPrep .spp data\n",
                         pInfo->dataFormat[0], pInfo->dataFormat[1],
                         pInfo->dataFormat[2], pInfo->dataFormat[3],
                         pInfo->formatVersion[0]);
        *pErrorCode=U_UNSUPPORTED_ERROR;
        return 0;
    }

    inBytes=(const uint8_t *)inData+headerSize;
    outBytes=(uint8_t *)outData+headerSize;

    inIndexes=(const int32_t *)inBytes;

    if(length>=0) {
        length-=headerSize;
        if(length<16*4) {
            udata_printError(ds, "usprep_swap(): too few bytes (%d after header) for StringPrep .spp data\n",
                             length);
            *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }
    }

    /* read the first 16 indexes (ICU 2.8/format version 3: _SPREP_INDEX_TOP==16, might grow) */
    for(i=0; i<16; ++i) {
        indexes[i]=udata_readInt32(ds, inIndexes[i]);
    }

    /* calculate the total length of the data */
    size=
        16*4+ /* size of indexes[] */
        indexes[_SPREP_INDEX_TRIE_SIZE]+
        indexes[_SPREP_INDEX_MAPPING_DATA_SIZE];

    if(length>=0) {
        if(length<size) {
            udata_printError(ds, "usprep_swap(): too few bytes (%d after header) for all of StringPrep .spp data\n",
                             length);
            *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }

        /* copy the data for inaccessible bytes */
        if(inBytes!=outBytes) {
            uprv_memcpy(outBytes, inBytes, size);
        }

        offset=0;

        /* swap the int32_t indexes[] */
        count=16*4;
        ds->swapArray32(ds, inBytes, count, outBytes, pErrorCode);
        offset+=count;

        /* swap the UTrie */
        count=indexes[_SPREP_INDEX_TRIE_SIZE];
        utrie_swap(ds, inBytes+offset, count, outBytes+offset, pErrorCode);
        offset+=count;

        /* swap the uint16_t mappingTable[] */
        count=indexes[_SPREP_INDEX_MAPPING_DATA_SIZE];
        ds->swapArray16(ds, inBytes+offset, count, outBytes+offset, pErrorCode);
        //offset+=count;
    }

    return headerSize+size;
}

#endif /* #if !UCONFIG_NO_IDNA */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2003-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/



U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(UStack)

UStack::UStack(UErrorCode &status) :
    UVector(status)
{
}

UStack::UStack(int32_t initialCapacity, UErrorCode &status) :
    UVector(initialCapacity, status)
{
}

UStack::UStack(UObjectDeleter *d, UElementsAreEqual *c, UErrorCode &status) :
    UVector(d, c, status)
{
}

UStack::UStack(UObjectDeleter *d, UElementsAreEqual *c, int32_t initialCapacity, UErrorCode &status) :
    UVector(d, c, initialCapacity, status)
{
}

UStack::~UStack() {}

void* UStack::pop(void) {
    int32_t n = size() - 1;
    void* result = 0;
    if (n >= 0) {
        result = elementAt(n);
        removeElementAt(n);
    }
    return result;
}

int32_t UStack::popi(void) {
    int32_t n = size() - 1;
    int32_t result = 0;
    if (n >= 0) {
        result = elementAti(n);
        removeElementAt(n);
    }
    return result;
}

int32_t UStack::search(void* obj) const {
    int32_t i = indexOf(obj);
    return (i >= 0) ? size() - i : i;
}

U_NAMESPACE_END
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// *******************************************************************************
// *
// *   Copyright (C) 1998-2014, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *
// *******************************************************************************
// *   file name:  ustr_cnv.cpp
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2004aug24
// *   created by: Markus W. Scherer
// *
// *   Character conversion functions moved here from ustring.c
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_CONVERSION

// #include "unicode/ustring.h"
// #include "unicode/ucnv.h"
// #include "cstring.h"
// #include "cmemory.h"
// #include "umutex.h"
// #include "ustr_cnv.h"
// #include "ucnv_bld.h"

// /* mutexed access to a shared default converter ----------------------------- */

// static UConverter *gDefaultConverter = NULL;

// U_CAPI UConverter* U_EXPORT2
// u_getDefaultConverter(UErrorCode *status)
// {
//     UConverter *converter = NULL;

//     if (gDefaultConverter != NULL) {
//         icu::umtx_lock(NULL);

//         /* need to check to make sure it wasn't taken out from under us */
//         if (gDefaultConverter != NULL) {
//             converter = gDefaultConverter;
//             gDefaultConverter = NULL;
//         }
//         icu::umtx_unlock(NULL);
//     }

//     /* if the cache was empty, create a converter */
//     if(converter == NULL) {
//         converter = ucnv_open(NULL, status);
//         if(U_FAILURE(*status)) {
//             ucnv_close(converter);
//             converter = NULL;
//         }
//     }

//     return converter;
// }

// U_CAPI void U_EXPORT2
// u_releaseDefaultConverter(UConverter *converter)
// {
//     if(gDefaultConverter == NULL) {
//         if (converter != NULL) {
//             ucnv_reset(converter);
//         }
//         ucnv_enableCleanup();
//         icu::umtx_lock(NULL);
//         if(gDefaultConverter == NULL) {
//             gDefaultConverter = converter;
//             converter = NULL;
//         }
//         icu::umtx_unlock(NULL);
//     }

//     if(converter != NULL) {
//         ucnv_close(converter);
//     }
// }

// U_CAPI void U_EXPORT2
// u_flushDefaultConverter()
// {
//     UConverter *converter = NULL;

//     if (gDefaultConverter != NULL) {
//         icu::umtx_lock(NULL);

//         /* need to check to make sure it wasn't taken out from under us */
//         if (gDefaultConverter != NULL) {
//             converter = gDefaultConverter;
//             gDefaultConverter = NULL;
//         }
//         icu::umtx_unlock(NULL);
//     }

//     /* if the cache was populated, flush it */
//     if(converter != NULL) {
//          ucnv_close(converter);
//     }
// }


// /* conversions between char* and UChar* ------------------------------------- */

// /* maximum string length for u_uastrcpy() and u_austrcpy() implementations */
// #define MAX_STRLEN 0x0FFFFFFF

// /*
//  returns the minimum of (the length of the null-terminated string) and n.
// */
// static int32_t u_astrnlen(const char *s1, int32_t n)
// {
//     int32_t len = 0;

//     if (s1)
//     {
//         while (n-- && *(s1++))
//         {
//             len++;
//         }
//     }
//     return len;
// }

// U_CAPI UChar*  U_EXPORT2
// u_uastrncpy(UChar *ucs1,
//            const char *s2,
//            int32_t n)
// {
//   UChar *target = ucs1;
//   UErrorCode err = U_ZERO_ERROR;
//   UConverter *cnv = u_getDefaultConverter(&err);
//   if(U_SUCCESS(err) && cnv != NULL) {
//     ucnv_reset(cnv);
//     ucnv_toUnicode(cnv,
//                    &target,
//                    ucs1+n,
//                    &s2,
//                    s2+u_astrnlen(s2, n),
//                    NULL,
//                    TRUE,
//                    &err);
//     ucnv_reset(cnv); /* be good citizens */
//     u_releaseDefaultConverter(cnv);
//     if(U_FAILURE(err) && (err != U_BUFFER_OVERFLOW_ERROR) ) {
//       *ucs1 = 0; /* failure */
//     }
//     if(target < (ucs1+n)) { /* U_BUFFER_OVERFLOW_ERROR isn't an err, just means no termination will happen. */
//       *target = 0;  /* terminate */
//     }
//   } else {
//     *ucs1 = 0;
//   }
//   return ucs1;
// }

// U_CAPI UChar*  U_EXPORT2
// u_uastrcpy(UChar *ucs1,
//           const char *s2 )
// {
//   UErrorCode err = U_ZERO_ERROR;
//   UConverter *cnv = u_getDefaultConverter(&err);
//   if(U_SUCCESS(err) && cnv != NULL) {
//     ucnv_toUChars(cnv,
//                     ucs1,
//                     MAX_STRLEN,
//                     s2,
//                     (int32_t)uprv_strlen(s2),
//                     &err);
//     u_releaseDefaultConverter(cnv);
//     if(U_FAILURE(err)) {
//       *ucs1 = 0;
//     }
//   } else {
//     *ucs1 = 0;
//   }
//   return ucs1;
// }

// /*
//  returns the minimum of (the length of the null-terminated string) and n.
// */
// static int32_t u_ustrnlen(const UChar *ucs1, int32_t n)
// {
//     int32_t len = 0;

//     if (ucs1)
//     {
//         while (n-- && *(ucs1++))
//         {
//             len++;
//         }
//     }
//     return len;
// }

// U_CAPI char*  U_EXPORT2
// u_austrncpy(char *s1,
//         const UChar *ucs2,
//         int32_t n)
// {
//   char *target = s1;
//   UErrorCode err = U_ZERO_ERROR;
//   UConverter *cnv = u_getDefaultConverter(&err);
//   if(U_SUCCESS(err) && cnv != NULL) {
//     ucnv_reset(cnv);
//     ucnv_fromUnicode(cnv,
//                   &target,
//                   s1+n,
//                   &ucs2,
//                   ucs2+u_ustrnlen(ucs2, n),
//                   NULL,
//                   TRUE,
//                   &err);
//     ucnv_reset(cnv); /* be good citizens */
//     u_releaseDefaultConverter(cnv);
//     if(U_FAILURE(err) && (err != U_BUFFER_OVERFLOW_ERROR) ) {
//       *s1 = 0; /* failure */
//     }
//     if(target < (s1+n)) { /* U_BUFFER_OVERFLOW_ERROR isn't an err, just means no termination will happen. */
//       *target = 0;  /* terminate */
//     }
//   } else {
//     *s1 = 0;
//   }
//   return s1;
// }

// U_CAPI char*  U_EXPORT2
// u_austrcpy(char *s1,
//          const UChar *ucs2 )
// {
//   UErrorCode err = U_ZERO_ERROR;
//   UConverter *cnv = u_getDefaultConverter(&err);
//   if(U_SUCCESS(err) && cnv != NULL) {
//     int32_t len = ucnv_fromUChars(cnv,
//                   s1,
//                   MAX_STRLEN,
//                   ucs2,
//                   -1,
//                   &err);
//     u_releaseDefaultConverter(cnv);
//     s1[len] = 0;
//   } else {
//     *s1 = 0;
//   }
//   return s1;
// }

// #endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// *******************************************************************************
// *   Copyright (C) 2011, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// *******************************************************************************
// *   file name:  ustr_titlecase_brkiter.cpp
// *   encoding:   UTF-8
// *   tab size:   8 (not used)
// *   indentation:4
// *
// *   created on: 2011may30
// *   created by: Markus W. Scherer
// *
// *   Titlecasing functions that are based on BreakIterator
// *   were moved here to break dependency cycles among parts of the common library.
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/brkiter.h"
// #include "unicode/casemap.h"
// #include "unicode/chariter.h"
// #include "unicode/localpointer.h"
// #include "unicode/ubrk.h"
// #include "unicode/ucasemap.h"
// #include "unicode/utext.h"
// #include "cmemory.h"
// #include "uassert.h"
// #include "ucase.h"
// #include "ucasemap_imp.h"

// U_NAMESPACE_BEGIN

// /**
//  * Whole-string BreakIterator.
//  * Titlecasing only calls setText(), first(), and next().
//  * We implement the rest only to satisfy the abstract interface.
//  */
// class WholeStringBreakIterator : public BreakIterator {
// public:
//     WholeStringBreakIterator() : BreakIterator(), length(0) {}
//     ~WholeStringBreakIterator() U_OVERRIDE;
//     UBool operator==(const BreakIterator&) const U_OVERRIDE;
//     WholeStringBreakIterator *clone() const U_OVERRIDE;
//     static UClassID U_EXPORT2 getStaticClassID();
//     UClassID getDynamicClassID() const U_OVERRIDE;
//     CharacterIterator &getText() const U_OVERRIDE;
//     UText *getUText(UText *fillIn, UErrorCode &errorCode) const U_OVERRIDE;
//     void  setText(const UnicodeString &text) U_OVERRIDE;
//     void  setText(UText *text, UErrorCode &errorCode) U_OVERRIDE;
//     void  adoptText(CharacterIterator* it) U_OVERRIDE;
//     int32_t first() U_OVERRIDE;
//     int32_t last() U_OVERRIDE;
//     int32_t previous() U_OVERRIDE;
//     int32_t next() U_OVERRIDE;
//     int32_t current() const U_OVERRIDE;
//     int32_t following(int32_t offset) U_OVERRIDE;
//     int32_t preceding(int32_t offset) U_OVERRIDE;
//     UBool isBoundary(int32_t offset) U_OVERRIDE;
//     int32_t next(int32_t n) U_OVERRIDE;
//     WholeStringBreakIterator *createBufferClone(void *stackBuffer, int32_t &BufferSize,
//                                                 UErrorCode &errorCode) U_OVERRIDE;
//     WholeStringBreakIterator &refreshInputText(UText *input, UErrorCode &errorCode) U_OVERRIDE;

// private:
//     int32_t length;
// };

// UOBJECT_DEFINE_RTTI_IMPLEMENTATION(WholeStringBreakIterator)

// WholeStringBreakIterator::~WholeStringBreakIterator() {}
// UBool WholeStringBreakIterator::operator==(const BreakIterator&) const { return FALSE; }
// WholeStringBreakIterator *WholeStringBreakIterator::clone() const { return nullptr; }

// CharacterIterator &WholeStringBreakIterator::getText() const {
//     UPRV_UNREACHABLE;  // really should not be called
// }
// UText *WholeStringBreakIterator::getUText(UText * /*fillIn*/, UErrorCode &errorCode) const {
//     if (U_SUCCESS(errorCode)) {
//         errorCode = U_UNSUPPORTED_ERROR;
//     }
//     return nullptr;
// }

// void  WholeStringBreakIterator::setText(const UnicodeString &text) {
//     length = text.length();
// }
// void  WholeStringBreakIterator::setText(UText *text, UErrorCode &errorCode) {
//     if (U_SUCCESS(errorCode)) {
//         int64_t length64 = utext_nativeLength(text);
//         if (length64 <= INT32_MAX) {
//             length = (int32_t)length64;
//         } else {
//             errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
//         }
//     }
// }
// void  WholeStringBreakIterator::adoptText(CharacterIterator*) {
//     UPRV_UNREACHABLE;  // should not be called
// }

// int32_t WholeStringBreakIterator::first() { return 0; }
// int32_t WholeStringBreakIterator::last() { return length; }
// int32_t WholeStringBreakIterator::previous() { return 0; }
// int32_t WholeStringBreakIterator::next() { return length; }
// int32_t WholeStringBreakIterator::current() const { return 0; }
// int32_t WholeStringBreakIterator::following(int32_t /*offset*/) { return length; }
// int32_t WholeStringBreakIterator::preceding(int32_t /*offset*/) { return 0; }
// UBool WholeStringBreakIterator::isBoundary(int32_t /*offset*/) { return FALSE; }
// int32_t WholeStringBreakIterator::next(int32_t /*n*/) { return length; }

// WholeStringBreakIterator *WholeStringBreakIterator::createBufferClone(
//         void * /*stackBuffer*/, int32_t & /*BufferSize*/, UErrorCode &errorCode) {
//     if (U_SUCCESS(errorCode)) {
//         errorCode = U_UNSUPPORTED_ERROR;
//     }
//     return nullptr;
// }
// WholeStringBreakIterator &WholeStringBreakIterator::refreshInputText(
//         UText * /*input*/, UErrorCode &errorCode) {
//     if (U_SUCCESS(errorCode)) {
//         errorCode = U_UNSUPPORTED_ERROR;
//     }
//     return *this;
// }

// U_CFUNC
// BreakIterator *ustrcase_getTitleBreakIterator(
//         const Locale *locale, const char *locID, uint32_t options, BreakIterator *iter,
//         LocalPointer<BreakIterator> &ownedIter, UErrorCode &errorCode) {
//     if (U_FAILURE(errorCode)) { return nullptr; }
//     options &= U_TITLECASE_ITERATOR_MASK;
//     if (options != 0 && iter != nullptr) {
//         errorCode = U_ILLEGAL_ARGUMENT_ERROR;
//         return nullptr;
//     }
//     if (iter == nullptr) {
//         switch (options) {
//         case 0:
//             iter = BreakIterator::createWordInstance(
//                 locale != nullptr ? *locale : Locale(locID), errorCode);
//             break;
//         case U_TITLECASE_WHOLE_STRING:
//             iter = new WholeStringBreakIterator();
//             if (iter == nullptr) {
//                 errorCode = U_MEMORY_ALLOCATION_ERROR;
//             }
//             break;
//         case U_TITLECASE_SENTENCES:
//             iter = BreakIterator::createSentenceInstance(
//                 locale != nullptr ? *locale : Locale(locID), errorCode);
//             break;
//         default:
//             errorCode = U_ILLEGAL_ARGUMENT_ERROR;
//             break;
//         }
//         ownedIter.adoptInstead(iter);
//     }
//     return iter;
// }

// int32_t CaseMap::toTitle(
//         const char *locale, uint32_t options, BreakIterator *iter,
//         const UChar *src, int32_t srcLength,
//         UChar *dest, int32_t destCapacity, Edits *edits,
//         UErrorCode &errorCode) {
//     LocalPointer<BreakIterator> ownedIter;
//     iter = ustrcase_getTitleBreakIterator(nullptr, locale, options, iter, ownedIter, errorCode);
//     if(iter==NULL) {
//         return 0;
//     }
//     UnicodeString s(srcLength<0, src, srcLength);
//     iter->setText(s);
//     return ustrcase_map(
//         ustrcase_getCaseLocale(locale), options, iter,
//         dest, destCapacity,
//         src, srcLength,
//         ustrcase_internalToTitle, edits, errorCode);
// }

// U_NAMESPACE_END

// U_NAMESPACE_USE

// U_CAPI int32_t U_EXPORT2
// u_strToTitle(UChar *dest, int32_t destCapacity,
//              const UChar *src, int32_t srcLength,
//              UBreakIterator *titleIter,
//              const char *locale,
//              UErrorCode *pErrorCode) {
//     LocalPointer<BreakIterator> ownedIter;
//     BreakIterator *iter = ustrcase_getTitleBreakIterator(
//         nullptr, locale, 0, reinterpret_cast<BreakIterator *>(titleIter),
//         ownedIter, *pErrorCode);
//     if (iter == nullptr) {
//         return 0;
//     }
//     UnicodeString s(srcLength<0, src, srcLength);
//     iter->setText(s);
//     return ustrcase_mapWithOverlap(
//         ustrcase_getCaseLocale(locale), 0, iter,
//         dest, destCapacity,
//         src, srcLength,
//         ustrcase_internalToTitle, *pErrorCode);
// }

// U_CAPI int32_t U_EXPORT2
// ucasemap_toTitle(UCaseMap *csm,
//                  UChar *dest, int32_t destCapacity,
//                  const UChar *src, int32_t srcLength,
//                  UErrorCode *pErrorCode) {
//     if (U_FAILURE(*pErrorCode)) {
//         return 0;
//     }
//     if (csm->iter == NULL) {
//         LocalPointer<BreakIterator> ownedIter;
//         BreakIterator *iter = ustrcase_getTitleBreakIterator(
//             nullptr, csm->locale, csm->options, nullptr, ownedIter, *pErrorCode);
//         if (iter == nullptr) {
//             return 0;
//         }
//         csm->iter = ownedIter.orphan();
//     }
//     UnicodeString s(srcLength<0, src, srcLength);
//     csm->iter->setText(s);
//     return ustrcase_map(
//         csm->caseLocale, csm->options, csm->iter,
//         dest, destCapacity,
//         src, srcLength,
//         ustrcase_internalToTitle, NULL, *pErrorCode);
// }

// #endif  // !UCONFIG_NO_BREAK_ITERATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2001-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ustr_wcs.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2004sep07
*   created by: Markus W. Scherer
*
*   u_strToWCS() and u_strFromWCS() functions
*   moved here from ustrtrns.c for better modularization.
*/









#if defined(U_WCHAR_IS_UTF16) || defined(U_WCHAR_IS_UTF32) || !UCONFIG_NO_CONVERSION

#define _STACK_BUFFER_CAPACITY 1000
#define _BUFFER_CAPACITY_MULTIPLIER 2

#if !defined(U_WCHAR_IS_UTF16) && !defined(U_WCHAR_IS_UTF32)
// TODO: We should use CharString for char buffers and UnicodeString for UChar buffers.
// Then we could change this to work only with wchar_t buffers.
static inline UBool 
u_growAnyBufferFromStatic(void *context,
                       void **pBuffer, int32_t *pCapacity, int32_t reqCapacity,
                       int32_t length, int32_t size) {
    // Use char* not void* to avoid the compiler's strict-aliasing assumptions
    // and related warnings.
    char *newBuffer=(char *)uprv_malloc(reqCapacity*size);
    if(newBuffer!=NULL) {
        if(length>0) {
            uprv_memcpy(newBuffer, *pBuffer, (size_t)length*size);
        }
        *pCapacity=reqCapacity;
    } else {
        *pCapacity=0;
    }

    /* release the old pBuffer if it was not statically allocated */
    if(*pBuffer!=(char *)context) {
        uprv_free(*pBuffer);
    }

    *pBuffer=newBuffer;
    return (UBool)(newBuffer!=NULL);
}

/* helper function */
static wchar_t* 
_strToWCS(wchar_t *dest, 
           int32_t destCapacity,
           int32_t *pDestLength,
           const UChar *src, 
           int32_t srcLength,
           UErrorCode *pErrorCode){

    char stackBuffer [_STACK_BUFFER_CAPACITY];
    char* tempBuf = stackBuffer;
    int32_t tempBufCapacity = _STACK_BUFFER_CAPACITY;
    char* tempBufLimit = stackBuffer + tempBufCapacity;
    UConverter* conv = NULL;
    char* saveBuf = tempBuf;
    wchar_t* intTarget=NULL;
    int32_t intTargetCapacity=0;
    int count=0,retVal=0;
    
    const UChar *pSrcLimit =NULL;
    const UChar *pSrc = src;

    conv = u_getDefaultConverter(pErrorCode);
    
    if(U_FAILURE(*pErrorCode)){
        return NULL;
    }
    
    if(srcLength == -1){
        srcLength = u_strlen(pSrc);
    }
    
    pSrcLimit = pSrc + srcLength;

    for(;;) {
        /* reset the error state */
        *pErrorCode = U_ZERO_ERROR;

        /* convert to chars using default converter */
        ucnv_fromUnicode(conv,&tempBuf,tempBufLimit,&pSrc,pSrcLimit,NULL,(UBool)(pSrc==pSrcLimit),pErrorCode);
        count =(tempBuf - saveBuf);
        
        /* This should rarely occur */
        if(*pErrorCode==U_BUFFER_OVERFLOW_ERROR){
            tempBuf = saveBuf;
            
            /* we dont have enough room on the stack grow the buffer */
            int32_t newCapacity = 2 * srcLength;
            if(newCapacity <= tempBufCapacity) {
                newCapacity = _BUFFER_CAPACITY_MULTIPLIER * tempBufCapacity;
            }
            if(!u_growAnyBufferFromStatic(stackBuffer,(void**) &tempBuf, &tempBufCapacity,
                    newCapacity, count, 1)) {
                goto cleanup;
            }
          
           saveBuf = tempBuf;
           tempBufLimit = tempBuf + tempBufCapacity;
           tempBuf = tempBuf + count;

        } else {
            break;
        }
    }

    if(U_FAILURE(*pErrorCode)){
        goto cleanup;
    }

    /* done with conversion null terminate the char buffer */
    if(count>=tempBufCapacity){
        tempBuf = saveBuf;
        /* we dont have enough room on the stack grow the buffer */
        if(!u_growAnyBufferFromStatic(stackBuffer,(void**) &tempBuf, &tempBufCapacity, 
                count+1, count, 1)) {
            goto cleanup;
        }              
       saveBuf = tempBuf;
    }
    
    saveBuf[count]=0;
      

    /* allocate more space than required 
     * here we assume that every char requires 
     * no more than 2 wchar_ts
     */
    intTargetCapacity =  (count * _BUFFER_CAPACITY_MULTIPLIER + 1) /*for null termination */;
    intTarget = (wchar_t*)uprv_malloc( intTargetCapacity * sizeof(wchar_t) );

    if(intTarget){

        int32_t nulLen = 0;
        int32_t remaining = intTargetCapacity;
        wchar_t* pIntTarget=intTarget;
        tempBuf = saveBuf;
        
        /* now convert the mbs to wcs */
        for(;;){
            
            /* we can call the system API since we are sure that
             * there is atleast 1 null in the input
             */
            retVal = uprv_mbstowcs(pIntTarget,(tempBuf+nulLen),remaining);
            
            if(retVal==-1){
                *pErrorCode = U_INVALID_CHAR_FOUND;
                break;
            }else if(retVal== remaining){/* should never occur */
                int numWritten = (pIntTarget-intTarget);
                u_growAnyBufferFromStatic(NULL,(void**) &intTarget,
                                          &intTargetCapacity,
                                          intTargetCapacity * _BUFFER_CAPACITY_MULTIPLIER,
                                          numWritten,
                                          sizeof(wchar_t));
                pIntTarget = intTarget;
                remaining=intTargetCapacity;

                if(nulLen!=count){ /*there are embedded nulls*/
                    pIntTarget+=numWritten;
                    remaining-=numWritten;
                }

            }else{
                int32_t nulVal;
                /*scan for nulls */
                /* we donot check for limit since tempBuf is null terminated */
                while(tempBuf[nulLen++] != 0){
                }
                nulVal = (nulLen < srcLength) ? 1 : 0; 
                pIntTarget = pIntTarget + retVal+nulVal;
                remaining -=(retVal+nulVal);
            
                /* check if we have reached the source limit*/
                if(nulLen>=(count)){
                    break;
                }
            }
        }
        count = (int32_t)(pIntTarget-intTarget);
       
        if(0 < count && count <= destCapacity){
            uprv_memcpy(dest, intTarget, (size_t)count*sizeof(wchar_t));
        }  

        if(pDestLength){
            *pDestLength = count;
        }

        /* free the allocated memory */
        uprv_free(intTarget);

    }else{
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
    }
cleanup:
    /* are we still using stack buffer */
    if(stackBuffer != saveBuf){
        uprv_free(saveBuf);
    }
    u_terminateWChars(dest,destCapacity,count,pErrorCode);

    u_releaseDefaultConverter(conv);

    return dest;
}
#endif

U_CAPI wchar_t* U_EXPORT2
u_strToWCS(wchar_t *dest, 
           int32_t destCapacity,
           int32_t *pDestLength,
           const UChar *src, 
           int32_t srcLength,
           UErrorCode *pErrorCode){

    /* args check */
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)){
        return NULL;
    }
        
    if( (src==NULL && srcLength!=0) || srcLength < -1 ||
        (destCapacity<0) || (dest == NULL && destCapacity > 0)
    ) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    
#ifdef U_WCHAR_IS_UTF16
    /* wchar_t is UTF-16 just do a memcpy */
    if(srcLength == -1){
        srcLength = u_strlen(src);
    }
    if(0 < srcLength && srcLength <= destCapacity){
        u_memcpy((UChar *)dest, src, srcLength);
    }
    if(pDestLength){
       *pDestLength = srcLength;
    }

    u_terminateUChars((UChar *)dest,destCapacity,srcLength,pErrorCode);

    return dest;

#elif defined U_WCHAR_IS_UTF32
    
    return (wchar_t*)u_strToUTF32((UChar32*)dest, destCapacity, pDestLength,
                                  src, srcLength, pErrorCode);

#else
    
    return _strToWCS(dest,destCapacity,pDestLength,src,srcLength, pErrorCode);
    
#endif

}

#if !defined(U_WCHAR_IS_UTF16) && !defined(U_WCHAR_IS_UTF32)
/* helper function */
static UChar* 
_strFromWCS( UChar   *dest,
             int32_t destCapacity, 
             int32_t *pDestLength,
             const wchar_t *src,
             int32_t srcLength,
             UErrorCode *pErrorCode)
{
    int32_t retVal =0, count =0 ;
    UConverter* conv = NULL;
    UChar* pTarget = NULL;
    UChar* pTargetLimit = NULL;
    UChar* target = NULL;
    
    UChar uStack [_STACK_BUFFER_CAPACITY];

    wchar_t wStack[_STACK_BUFFER_CAPACITY];
    wchar_t* pWStack = wStack;


    char cStack[_STACK_BUFFER_CAPACITY];
    int32_t cStackCap = _STACK_BUFFER_CAPACITY;
    char* pCSrc=cStack;
    char* pCSave=pCSrc;
    char* pCSrcLimit=NULL;

    const wchar_t* pSrc = src;
    const wchar_t* pSrcLimit = NULL;

    if(srcLength ==-1){
        /* if the wchar_t source is null terminated we can safely
         * assume that there are no embedded nulls, this is a fast
         * path for null terminated strings.
         */
        for(;;){
            /* convert wchars  to chars */
            retVal = uprv_wcstombs(pCSrc,src, cStackCap);
    
            if(retVal == -1){
                *pErrorCode = U_ILLEGAL_CHAR_FOUND;
                goto cleanup;
            }else if(retVal >= (cStackCap-1)){
                /* Should rarely occur */
                u_growAnyBufferFromStatic(cStack,(void**)&pCSrc,&cStackCap,
                    cStackCap * _BUFFER_CAPACITY_MULTIPLIER, 0, sizeof(char));
                pCSave = pCSrc;
            }else{
                /* converted every thing */
                pCSrc = pCSrc+retVal;
                break;
            }
        }
        
    }else{
        /* here the source is not null terminated 
         * so it may have nulls embeded and we need to
         * do some extra processing 
         */
        int32_t remaining =cStackCap;
        
        pSrcLimit = src + srcLength;

        for(;;){
            int32_t nulLen = 0;

            /* find nulls in the string */
            while(nulLen<srcLength && pSrc[nulLen++]!=0){
            }

            if((pSrc+nulLen) < pSrcLimit){
                /* check if we have enough room in pCSrc */
                if(remaining < (nulLen * MB_CUR_MAX)){
                    /* should rarely occur */
                    int32_t len = (pCSrc-pCSave);
                    pCSrc = pCSave;
                    /* we do not have enough room so grow the buffer*/
                    u_growAnyBufferFromStatic(cStack,(void**)&pCSrc,&cStackCap,
                           _BUFFER_CAPACITY_MULTIPLIER*cStackCap+(nulLen*MB_CUR_MAX),len,sizeof(char));

                    pCSave = pCSrc;
                    pCSrc = pCSave+len;
                    remaining = cStackCap-(pCSrc - pCSave);
                }

                /* we have found a null  so convert the 
                 * chunk from begining of non-null char to null
                 */
                retVal = uprv_wcstombs(pCSrc,pSrc,remaining);

                if(retVal==-1){
                    /* an error occurred bail out */
                    *pErrorCode = U_ILLEGAL_CHAR_FOUND;
                    goto cleanup;
                }

                pCSrc += retVal+1 /* already null terminated */;

                pSrc += nulLen; /* skip past the null */
                srcLength-=nulLen; /* decrement the srcLength */
                remaining -= (pCSrc-pCSave);


            }else{
                /* the source is not null terminated and we are 
                 * end of source so we copy the source to a temp buffer
                 * null terminate it and convert wchar_ts to chars
                 */
                if(nulLen >= _STACK_BUFFER_CAPACITY){
                    /* Should rarely occcur */
                    /* allocate new buffer buffer */
                    pWStack =(wchar_t*) uprv_malloc(sizeof(wchar_t) * (nulLen + 1));
                    if(pWStack==NULL){
                        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
                        goto cleanup;
                    }
                }
                if(nulLen>0){
                    /* copy the contents to tempStack */
                    uprv_memcpy(pWStack, pSrc, (size_t)nulLen*sizeof(wchar_t));
                }
            
                /* null terminate the tempBuffer */
                pWStack[nulLen] =0 ;
            
                if(remaining < (nulLen * MB_CUR_MAX)){
                    /* Should rarely occur */
                    int32_t len = (pCSrc-pCSave);
                    pCSrc = pCSave;
                    /* we do not have enough room so grow the buffer*/
                    u_growAnyBufferFromStatic(cStack,(void**)&pCSrc,&cStackCap,
                           cStackCap+(nulLen*MB_CUR_MAX),len,sizeof(char));

                    pCSave = pCSrc;
                    pCSrc = pCSave+len;
                    remaining = cStackCap-(pCSrc - pCSave);
                }
                /* convert to chars */
                retVal = uprv_wcstombs(pCSrc,pWStack,remaining);
            
                pCSrc += retVal;
                pSrc  += nulLen;
                srcLength-=nulLen; /* decrement the srcLength */
                break;
            }
        }
    }

    /* OK..now we have converted from wchar_ts to chars now 
     * convert chars to UChars 
     */
    pCSrcLimit = pCSrc;
    pCSrc = pCSave;
    pTarget = target= dest;
    pTargetLimit = dest + destCapacity;    
    
    conv= u_getDefaultConverter(pErrorCode);
    
    if(U_FAILURE(*pErrorCode)|| conv==NULL){
        goto cleanup;
    }
    
    for(;;) {
        
        *pErrorCode = U_ZERO_ERROR;
        
        /* convert to stack buffer*/
        ucnv_toUnicode(conv,&pTarget,pTargetLimit,(const char**)&pCSrc,pCSrcLimit,NULL,(UBool)(pCSrc==pCSrcLimit),pErrorCode);
        
        /* increment count to number written to stack */
        count+= pTarget - target;
        
        if(*pErrorCode==U_BUFFER_OVERFLOW_ERROR){
            target = uStack;
            pTarget = uStack;
            pTargetLimit = uStack + _STACK_BUFFER_CAPACITY;
        } else {
            break;
        }
        
    }
    
    if(pDestLength){
        *pDestLength =count;
    }

    u_terminateUChars(dest,destCapacity,count,pErrorCode);
    
cleanup:
 
    if(cStack != pCSave){
        uprv_free(pCSave);
    }

    if(wStack != pWStack){
        uprv_free(pWStack);
    }
    
    u_releaseDefaultConverter(conv);

    return dest;
}
#endif

U_CAPI UChar* U_EXPORT2
u_strFromWCS(UChar   *dest,
             int32_t destCapacity, 
             int32_t *pDestLength,
             const wchar_t *src,
             int32_t srcLength,
             UErrorCode *pErrorCode)
{

    /* args check */
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)){
        return NULL;
    }

    if( (src==NULL && srcLength!=0) || srcLength < -1 ||
        (destCapacity<0) || (dest == NULL && destCapacity > 0)
    ) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

#ifdef U_WCHAR_IS_UTF16
    /* wchar_t is UTF-16 just do a memcpy */
    if(srcLength == -1){
        srcLength = u_strlen((const UChar *)src);
    }
    if(0 < srcLength && srcLength <= destCapacity){
        u_memcpy(dest, (const UChar *)src, srcLength);
    }
    if(pDestLength){
       *pDestLength = srcLength;
    }

    u_terminateUChars(dest,destCapacity,srcLength,pErrorCode);

    return dest;

#elif defined U_WCHAR_IS_UTF32
    
    return u_strFromUTF32(dest, destCapacity, pDestLength,
                          (UChar32*)src, srcLength, pErrorCode);

#else

    return _strFromWCS(dest,destCapacity,pDestLength,src,srcLength,pErrorCode);  

#endif

}

#endif /* #if !defined(U_WCHAR_IS_UTF16) && !defined(U_WCHAR_IS_UTF32) && !UCONFIG_NO_CONVERSION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2001-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  ustrcase.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2002feb20
*   created by: Markus W. Scherer
*
*   Implementation file for string casing C API functions.
*   Uses functions from uchar.c for basic functionality that requires access
*   to the Unicode Character Database (uprops.dat).
*/

















U_NAMESPACE_BEGIN

namespace {

int32_t checkOverflowAndEditsError(int32_t destIndex, int32_t destCapacity,
                                   Edits *edits, UErrorCode &errorCode) {
    if (U_SUCCESS(errorCode)) {
        if (destIndex > destCapacity) {
            errorCode = U_BUFFER_OVERFLOW_ERROR;
        } else if (edits != NULL) {
            edits->copyErrorTo(errorCode);
        }
    }
    return destIndex;
}

/* Appends a full case mapping result, see UCASE_MAX_STRING_LENGTH. */
inline int32_t
appendResult(UChar *dest, int32_t destIndex, int32_t destCapacity,
             int32_t result, const UChar *s,
             int32_t cpLength, uint32_t options, icu::Edits *edits) {
    UChar32 c;
    int32_t length;

    /* decode the result */
    if(result<0) {
        /* (not) original code point */
        if(edits!=NULL) {
            edits->addUnchanged(cpLength);
        }
        if(options & U_OMIT_UNCHANGED_TEXT) {
            return destIndex;
        }
        c=~result;
        if(destIndex<destCapacity && c<=0xffff) {  // BMP slightly-fastpath
            dest[destIndex++]=(UChar)c;
            return destIndex;
        }
        length=cpLength;
    } else {
        if(result<=UCASE_MAX_STRING_LENGTH) {
            c=U_SENTINEL;
            length=result;
        } else if(destIndex<destCapacity && result<=0xffff) {  // BMP slightly-fastpath
            dest[destIndex++]=(UChar)result;
            if(edits!=NULL) {
                edits->addReplace(cpLength, 1);
            }
            return destIndex;
        } else {
            c=result;
            length=U16_LENGTH(c);
        }
        if(edits!=NULL) {
            edits->addReplace(cpLength, length);
        }
    }
    if(length>(INT32_MAX-destIndex)) {
        return -1;  // integer overflow
    }

    if(destIndex<destCapacity) {
        /* append the result */
        if(c>=0) {
            /* code point */
            UBool isError=FALSE;
            U16_APPEND(dest, destIndex, destCapacity, c, isError);
            if(isError) {
                /* overflow, nothing written */
                destIndex+=length;
            }
        } else {
            /* string */
            if((destIndex+length)<=destCapacity) {
                while(length>0) {
                    dest[destIndex++]=*s++;
                    --length;
                }
            } else {
                /* overflow */
                destIndex+=length;
            }
        }
    } else {
        /* preflight */
        destIndex+=length;
    }
    return destIndex;
}

inline int32_t
appendUChar(UChar *dest, int32_t destIndex, int32_t destCapacity, UChar c) {
    if(destIndex<destCapacity) {
        dest[destIndex]=c;
    } else if(destIndex==INT32_MAX) {
        return -1;  // integer overflow
    }
    return destIndex+1;
}

int32_t
appendNonEmptyUnchanged(UChar *dest, int32_t destIndex, int32_t destCapacity,
                        const UChar *s, int32_t length, uint32_t options, icu::Edits *edits) {
    if(edits!=NULL) {
        edits->addUnchanged(length);
    }
    if(options & U_OMIT_UNCHANGED_TEXT) {
        return destIndex;
    }
    if(length>(INT32_MAX-destIndex)) {
        return -1;  // integer overflow
    }
    if((destIndex+length)<=destCapacity) {
        u_memcpy(dest+destIndex, s, length);
    }
    return destIndex + length;
}

inline int32_t
appendUnchanged(UChar *dest, int32_t destIndex, int32_t destCapacity,
                const UChar *s, int32_t length, uint32_t options, icu::Edits *edits) {
    if (length <= 0) {
        return destIndex;
    }
    return appendNonEmptyUnchanged(dest, destIndex, destCapacity, s, length, options, edits);
}

UChar32 U_CALLCONV
utf16_caseContextIterator(void *context, int8_t dir) {
    UCaseContext *csc=(UCaseContext *)context;
    UChar32 c;

    if(dir<0) {
        /* reset for backward iteration */
        csc->index=csc->cpStart;
        csc->dir=dir;
    } else if(dir>0) {
        /* reset for forward iteration */
        csc->index=csc->cpLimit;
        csc->dir=dir;
    } else {
        /* continue current iteration direction */
        dir=csc->dir;
    }

    if(dir<0) {
        if(csc->start<csc->index) {
            U16_PREV((const UChar *)csc->p, csc->start, csc->index, c);
            return c;
        }
    } else {
        if(csc->index<csc->limit) {
            U16_NEXT((const UChar *)csc->p, csc->index, csc->limit, c);
            return c;
        }
    }
    return U_SENTINEL;
}

/**
 * caseLocale >= 0: Lowercases [srcStart..srcLimit[ but takes context [0..srcLength[ into account.
 * caseLocale < 0: Case-folds [srcStart..srcLimit[.
 */
int32_t toLower(int32_t caseLocale, uint32_t options,
                UChar *dest, int32_t destCapacity,
                const UChar *src, UCaseContext *csc, int32_t srcStart, int32_t srcLimit,
                icu::Edits *edits, UErrorCode &errorCode) {
    const int8_t *latinToLower;
    if (caseLocale == UCASE_LOC_ROOT ||
            (caseLocale >= 0 ?
                !(caseLocale == UCASE_LOC_TURKISH || caseLocale == UCASE_LOC_LITHUANIAN) :
                (options & _FOLD_CASE_OPTIONS_MASK) == U_FOLD_CASE_DEFAULT)) {
        latinToLower = LatinCase::TO_LOWER_NORMAL;
    } else {
        latinToLower = LatinCase::TO_LOWER_TR_LT;
    }
    const UTrie2 *trie = ucase_getTrie();
    int32_t destIndex = 0;
    int32_t prev = srcStart;
    int32_t srcIndex = srcStart;
    for (;;) {
        // fast path for simple cases
        UChar lead = 0;
        while (srcIndex < srcLimit) {
            lead = src[srcIndex];
            int32_t delta;
            if (lead < LatinCase::LONG_S) {
                int8_t d = latinToLower[lead];
                if (d == LatinCase::EXC) { break; }
                ++srcIndex;
                if (d == 0) { continue; }
                delta = d;
            } else if (lead >= 0xd800) {
                break;  // surrogate or higher
            } else {
                uint16_t props = UTRIE2_GET16_FROM_U16_SINGLE_LEAD(trie, lead);
                if (UCASE_HAS_EXCEPTION(props)) { break; }
                ++srcIndex;
                if (!UCASE_IS_UPPER_OR_TITLE(props) || (delta = UCASE_GET_DELTA(props)) == 0) {
                    continue;
                }
            }
            lead += static_cast<UChar>(delta);
            destIndex = appendUnchanged(dest, destIndex, destCapacity,
                                        src + prev, srcIndex - 1 - prev, options, edits);
            if (destIndex >= 0) {
                destIndex = appendUChar(dest, destIndex, destCapacity, lead);
                if (edits != nullptr) {
                    edits->addReplace(1, 1);
                }
            }
            if (destIndex < 0) {
                errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
                return 0;
            }
            prev = srcIndex;
        }
        if (srcIndex >= srcLimit) {
            break;
        }
        // slow path
        int32_t cpStart = srcIndex++;
        UChar trail;
        UChar32 c;
        if (U16_IS_LEAD(lead) && srcIndex < srcLimit && U16_IS_TRAIL(trail = src[srcIndex])) {
            c = U16_GET_SUPPLEMENTARY(lead, trail);
            ++srcIndex;
        } else {
            c = lead;
        }
        const UChar *s;
        if (caseLocale >= 0) {
            csc->cpStart = cpStart;
            csc->cpLimit = srcIndex;
            c = ucase_toFullLower(c, utf16_caseContextIterator, csc, &s, caseLocale);
        } else {
            c = ucase_toFullFolding(c, &s, options);
        }
        if (c >= 0) {
            destIndex = appendUnchanged(dest, destIndex, destCapacity,
                                        src + prev, cpStart - prev, options, edits);
            if (destIndex >= 0) {
                destIndex = appendResult(dest, destIndex, destCapacity, c, s,
                                         srcIndex - cpStart, options, edits);
            }
            if (destIndex < 0) {
                errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
                return 0;
            }
            prev = srcIndex;
        }
    }
    destIndex = appendUnchanged(dest, destIndex, destCapacity,
                                src + prev, srcIndex - prev, options, edits);
    if (destIndex < 0) {
        errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }
    return destIndex;
}

int32_t toUpper(int32_t caseLocale, uint32_t options,
                UChar *dest, int32_t destCapacity,
                const UChar *src, UCaseContext *csc, int32_t srcLength,
                icu::Edits *edits, UErrorCode &errorCode) {
    const int8_t *latinToUpper;
    if (caseLocale == UCASE_LOC_TURKISH) {
        latinToUpper = LatinCase::TO_UPPER_TR;
    } else {
        latinToUpper = LatinCase::TO_UPPER_NORMAL;
    }
    const UTrie2 *trie = ucase_getTrie();
    int32_t destIndex = 0;
    int32_t prev = 0;
    int32_t srcIndex = 0;
    for (;;) {
        // fast path for simple cases
        UChar lead = 0;
        while (srcIndex < srcLength) {
            lead = src[srcIndex];
            int32_t delta;
            if (lead < LatinCase::LONG_S) {
                int8_t d = latinToUpper[lead];
                if (d == LatinCase::EXC) { break; }
                ++srcIndex;
                if (d == 0) { continue; }
                delta = d;
            } else if (lead >= 0xd800) {
                break;  // surrogate or higher
            } else {
                uint16_t props = UTRIE2_GET16_FROM_U16_SINGLE_LEAD(trie, lead);
                if (UCASE_HAS_EXCEPTION(props)) { break; }
                ++srcIndex;
                if (UCASE_GET_TYPE(props) != UCASE_LOWER || (delta = UCASE_GET_DELTA(props)) == 0) {
                    continue;
                }
            }
            lead += static_cast<UChar>(delta);
            destIndex = appendUnchanged(dest, destIndex, destCapacity,
                                        src + prev, srcIndex - 1 - prev, options, edits);
            if (destIndex >= 0) {
                destIndex = appendUChar(dest, destIndex, destCapacity, lead);
                if (edits != nullptr) {
                    edits->addReplace(1, 1);
                }
            }
            if (destIndex < 0) {
                errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
                return 0;
            }
            prev = srcIndex;
        }
        if (srcIndex >= srcLength) {
            break;
        }
        // slow path
        int32_t cpStart;
        csc->cpStart = cpStart = srcIndex++;
        UChar trail;
        UChar32 c;
        if (U16_IS_LEAD(lead) && srcIndex < srcLength && U16_IS_TRAIL(trail = src[srcIndex])) {
            c = U16_GET_SUPPLEMENTARY(lead, trail);
            ++srcIndex;
        } else {
            c = lead;
        }
        csc->cpLimit = srcIndex;
        const UChar *s;
        c = ucase_toFullUpper(c, utf16_caseContextIterator, csc, &s, caseLocale);
        if (c >= 0) {
            destIndex = appendUnchanged(dest, destIndex, destCapacity,
                                        src + prev, cpStart - prev, options, edits);
            if (destIndex >= 0) {
                destIndex = appendResult(dest, destIndex, destCapacity, c, s,
                                         srcIndex - cpStart, options, edits);
            }
            if (destIndex < 0) {
                errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
                return 0;
            }
            prev = srcIndex;
        }
    }
    destIndex = appendUnchanged(dest, destIndex, destCapacity,
                                src + prev, srcIndex - prev, options, edits);
    if (destIndex < 0) {
        errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }
    return destIndex;
}

}  // namespace

U_NAMESPACE_END

U_NAMESPACE_USE

#if !UCONFIG_NO_BREAK_ITERATION

U_CFUNC int32_t U_CALLCONV
ustrcase_internalToTitle(int32_t caseLocale, uint32_t options, BreakIterator *iter,
                         UChar *dest, int32_t destCapacity,
                         const UChar *src, int32_t srcLength,
                         icu::Edits *edits,
                         UErrorCode &errorCode) {
    if (!ustrcase_checkTitleAdjustmentOptions(options, errorCode)) {
        return 0;
    }

    /* set up local variables */
    UCaseContext csc=UCASECONTEXT_INITIALIZER;
    csc.p=(void *)src;
    csc.limit=srcLength;
    int32_t destIndex=0;
    int32_t prev=0;
    UBool isFirstIndex=TRUE;

    /* titlecasing loop */
    while(prev<srcLength) {
        /* find next index where to titlecase */
        int32_t index;
        if(isFirstIndex) {
            isFirstIndex=FALSE;
            index=iter->first();
        } else {
            index=iter->next();
        }
        if(index==UBRK_DONE || index>srcLength) {
            index=srcLength;
        }

        /*
         * Segment [prev..index[ into 3 parts:
         * a) skipped characters (copy as-is) [prev..titleStart[
         * b) first letter (titlecase)              [titleStart..titleLimit[
         * c) subsequent characters (lowercase)                 [titleLimit..index[
         */
        if(prev<index) {
            // Find and copy skipped characters [prev..titleStart[
            int32_t titleStart=prev;
            int32_t titleLimit=prev;
            UChar32 c;
            U16_NEXT(src, titleLimit, index, c);
            if ((options&U_TITLECASE_NO_BREAK_ADJUSTMENT)==0) {
                // Adjust the titlecasing index to the next cased character,
                // or to the next letter/number/symbol/private use.
                // Stop with titleStart<titleLimit<=index
                // if there is a character to be titlecased,
                // or else stop with titleStart==titleLimit==index.
                UBool toCased = (options&U_TITLECASE_ADJUST_TO_CASED) != 0;
                while (toCased ? UCASE_NONE==ucase_getType(c) : !ustrcase_isLNS(c)) {
                    titleStart=titleLimit;
                    if(titleLimit==index) {
                        break;
                    }
                    U16_NEXT(src, titleLimit, index, c);
                }
                if (prev < titleStart) {
                    destIndex=appendUnchanged(dest, destIndex, destCapacity,
                                              src+prev, titleStart-prev, options, edits);
                    if(destIndex<0) {
                        errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                        return 0;
                    }
                }
            }

            if(titleStart<titleLimit) {
                /* titlecase c which is from [titleStart..titleLimit[ */
                csc.cpStart=titleStart;
                csc.cpLimit=titleLimit;
                const UChar *s;
                c=ucase_toFullTitle(c, utf16_caseContextIterator, &csc, &s, caseLocale);
                destIndex=appendResult(dest, destIndex, destCapacity, c, s,
                                       titleLimit-titleStart, options, edits);
                if(destIndex<0) {
                    errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                    return 0;
                }

                /* Special case Dutch IJ titlecasing */
                if (titleStart+1 < index &&
                        caseLocale == UCASE_LOC_DUTCH &&
                        (src[titleStart] == 0x0049 || src[titleStart] == 0x0069)) {
                    if (src[titleStart+1] == 0x006A) {
                        destIndex=appendUChar(dest, destIndex, destCapacity, 0x004A);
                        if(destIndex<0) {
                            errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                            return 0;
                        }
                        if(edits!=NULL) {
                            edits->addReplace(1, 1);
                        }
                        titleLimit++;
                    } else if (src[titleStart+1] == 0x004A) {
                        // Keep the capital J from getting lowercased.
                        destIndex=appendUnchanged(dest, destIndex, destCapacity,
                                                  src+titleStart+1, 1, options, edits);
                        if(destIndex<0) {
                            errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                            return 0;
                        }
                        titleLimit++;
                    }
                }

                /* lowercase [titleLimit..index[ */
                if(titleLimit<index) {
                    if((options&U_TITLECASE_NO_LOWERCASE)==0) {
                        /* Normal operation: Lowercase the rest of the word. */
                        destIndex+=
                            toLower(
                                caseLocale, options,
                                dest+destIndex, destCapacity-destIndex,
                                src, &csc, titleLimit, index,
                                edits, errorCode);
                        if(errorCode==U_BUFFER_OVERFLOW_ERROR) {
                            errorCode=U_ZERO_ERROR;
                        }
                        if(U_FAILURE(errorCode)) {
                            return destIndex;
                        }
                    } else {
                        /* Optionally just copy the rest of the word unchanged. */
                        destIndex=appendUnchanged(dest, destIndex, destCapacity,
                                                  src+titleLimit, index-titleLimit, options, edits);
                        if(destIndex<0) {
                            errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                            return 0;
                        }
                    }
                }
            }
        }

        prev=index;
    }

    return checkOverflowAndEditsError(destIndex, destCapacity, edits, errorCode);
}

#endif  // !UCONFIG_NO_BREAK_ITERATION

U_NAMESPACE_BEGIN
namespace GreekUpper {

// Data generated by prototype code, see
// http://site.icu-project.org/design/case/greek-upper
// TODO: Move this data into ucase.icu.
static const uint16_t data0370[] = {
    // U+0370..03FF
    0x0370,
    0x0370,
    0x0372,
    0x0372,
    0,
    0,
    0x0376,
    0x0376,
    0,
    0,
    0x037A,
    0x03FD,
    0x03FE,
    0x03FF,
    0,
    0x037F,
    0,
    0,
    0,
    0,
    0,
    0,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT | HAS_DIALYTIKA,
    0x0391 | HAS_VOWEL,
    0x0392,
    0x0393,
    0x0394,
    0x0395 | HAS_VOWEL,
    0x0396,
    0x0397 | HAS_VOWEL,
    0x0398,
    0x0399 | HAS_VOWEL,
    0x039A,
    0x039B,
    0x039C,
    0x039D,
    0x039E,
    0x039F | HAS_VOWEL,
    0x03A0,
    0x03A1,
    0,
    0x03A3,
    0x03A4,
    0x03A5 | HAS_VOWEL,
    0x03A6,
    0x03A7,
    0x03A8,
    0x03A9 | HAS_VOWEL,
    0x0399 | HAS_VOWEL | HAS_DIALYTIKA,
    0x03A5 | HAS_VOWEL | HAS_DIALYTIKA,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT | HAS_DIALYTIKA,
    0x0391 | HAS_VOWEL,
    0x0392,
    0x0393,
    0x0394,
    0x0395 | HAS_VOWEL,
    0x0396,
    0x0397 | HAS_VOWEL,
    0x0398,
    0x0399 | HAS_VOWEL,
    0x039A,
    0x039B,
    0x039C,
    0x039D,
    0x039E,
    0x039F | HAS_VOWEL,
    0x03A0,
    0x03A1,
    0x03A3,
    0x03A3,
    0x03A4,
    0x03A5 | HAS_VOWEL,
    0x03A6,
    0x03A7,
    0x03A8,
    0x03A9 | HAS_VOWEL,
    0x0399 | HAS_VOWEL | HAS_DIALYTIKA,
    0x03A5 | HAS_VOWEL | HAS_DIALYTIKA,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03CF,
    0x0392,
    0x0398,
    0x03D2,
    0x03D2 | HAS_ACCENT,
    0x03D2 | HAS_DIALYTIKA,
    0x03A6,
    0x03A0,
    0x03CF,
    0x03D8,
    0x03D8,
    0x03DA,
    0x03DA,
    0x03DC,
    0x03DC,
    0x03DE,
    0x03DE,
    0x03E0,
    0x03E0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0x039A,
    0x03A1,
    0x03F9,
    0x037F,
    0x03F4,
    0x0395 | HAS_VOWEL,
    0,
    0x03F7,
    0x03F7,
    0x03F9,
    0x03FA,
    0x03FA,
    0x03FC,
    0x03FD,
    0x03FE,
    0x03FF,
};

static const uint16_t data1F00[] = {
    // U+1F00..1FFF
    0x0391 | HAS_VOWEL,
    0x0391 | HAS_VOWEL,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL,
    0x0391 | HAS_VOWEL,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL,
    0x0395 | HAS_VOWEL,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0,
    0,
    0x0395 | HAS_VOWEL,
    0x0395 | HAS_VOWEL,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0,
    0,
    0x0397 | HAS_VOWEL,
    0x0397 | HAS_VOWEL,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL,
    0x0397 | HAS_VOWEL,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL,
    0x0399 | HAS_VOWEL,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL,
    0x0399 | HAS_VOWEL,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x039F | HAS_VOWEL,
    0x039F | HAS_VOWEL,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0,
    0,
    0x039F | HAS_VOWEL,
    0x039F | HAS_VOWEL,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0,
    0,
    0x03A5 | HAS_VOWEL,
    0x03A5 | HAS_VOWEL,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0,
    0x03A5 | HAS_VOWEL,
    0,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL,
    0x03A9 | HAS_VOWEL,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL,
    0x03A9 | HAS_VOWEL,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0,
    0,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL,
    0x0391 | HAS_VOWEL,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0391 | HAS_VOWEL,
    0x0391 | HAS_VOWEL,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_ACCENT,
    0x0391 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0,
    0x0399 | HAS_VOWEL,
    0,
    0,
    0,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0395 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_ACCENT,
    0x0397 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0,
    0,
    0,
    0x0399 | HAS_VOWEL,
    0x0399 | HAS_VOWEL,
    0x0399 | HAS_VOWEL | HAS_ACCENT | HAS_DIALYTIKA,
    0x0399 | HAS_VOWEL | HAS_ACCENT | HAS_DIALYTIKA,
    0,
    0,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT | HAS_DIALYTIKA,
    0x0399 | HAS_VOWEL,
    0x0399 | HAS_VOWEL,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0x0399 | HAS_VOWEL | HAS_ACCENT,
    0,
    0,
    0,
    0,
    0x03A5 | HAS_VOWEL,
    0x03A5 | HAS_VOWEL,
    0x03A5 | HAS_VOWEL | HAS_ACCENT | HAS_DIALYTIKA,
    0x03A5 | HAS_VOWEL | HAS_ACCENT | HAS_DIALYTIKA,
    0x03A1,
    0x03A1,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT | HAS_DIALYTIKA,
    0x03A5 | HAS_VOWEL,
    0x03A5 | HAS_VOWEL,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A5 | HAS_VOWEL | HAS_ACCENT,
    0x03A1,
    0,
    0,
    0,
    0,
    0,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI | HAS_ACCENT,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x039F | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_ACCENT,
    0x03A9 | HAS_VOWEL | HAS_YPOGEGRAMMENI,
    0,
    0,
    0,
};

// U+2126 Ohm sign
static const uint16_t data2126 = 0x03A9 | HAS_VOWEL;

uint32_t getLetterData(UChar32 c) {
    if (c < 0x370 || 0x2126 < c || (0x3ff < c && c < 0x1f00)) {
        return 0;
    } else if (c <= 0x3ff) {
        return data0370[c - 0x370];
    } else if (c <= 0x1fff) {
        return data1F00[c - 0x1f00];
    } else if (c == 0x2126) {
        return data2126;
    } else {
        return 0;
    }
}

uint32_t getDiacriticData(UChar32 c) {
    switch (c) {
    case 0x0300:  // varia
    case 0x0301:  // tonos = oxia
    case 0x0342:  // perispomeni
    case 0x0302:  // circumflex can look like perispomeni
    case 0x0303:  // tilde can look like perispomeni
    case 0x0311:  // inverted breve can look like perispomeni
        return HAS_ACCENT;
    case 0x0308:  // dialytika = diaeresis
        return HAS_COMBINING_DIALYTIKA;
    case 0x0344:  // dialytika tonos
        return HAS_COMBINING_DIALYTIKA | HAS_ACCENT;
    case 0x0345:  // ypogegrammeni = iota subscript
        return HAS_YPOGEGRAMMENI;
    case 0x0304:  // macron
    case 0x0306:  // breve
    case 0x0313:  // comma above
    case 0x0314:  // reversed comma above
    case 0x0343:  // koronis
        return HAS_OTHER_GREEK_DIACRITIC;
    default:
        return 0;
    }
}

UBool isFollowedByCasedLetter(const UChar *s, int32_t i, int32_t length) {
    while (i < length) {
        UChar32 c;
        U16_NEXT(s, i, length, c);
        int32_t type = ucase_getTypeOrIgnorable(c);
        if ((type & UCASE_IGNORABLE) != 0) {
            // Case-ignorable, continue with the loop.
        } else if (type != UCASE_NONE) {
            return TRUE;  // Followed by cased letter.
        } else {
            return FALSE;  // Uncased and not case-ignorable.
        }
    }
    return FALSE;  // Not followed by cased letter.
}

/**
 * Greek string uppercasing with a state machine.
 * Probably simpler than a stateless function that has to figure out complex context-before
 * for each character.
 * TODO: Try to re-consolidate one way or another with the non-Greek function.
 */
int32_t toUpper(uint32_t options,
                UChar *dest, int32_t destCapacity,
                const UChar *src, int32_t srcLength,
                Edits *edits,
                UErrorCode &errorCode) {
    int32_t destIndex=0;
    uint32_t state = 0;
    for (int32_t i = 0; i < srcLength;) {
        int32_t nextIndex = i;
        UChar32 c;
        U16_NEXT(src, nextIndex, srcLength, c);
        uint32_t nextState = 0;
        int32_t type = ucase_getTypeOrIgnorable(c);
        if ((type & UCASE_IGNORABLE) != 0) {
            // c is case-ignorable
            nextState |= (state & AFTER_CASED);
        } else if (type != UCASE_NONE) {
            // c is cased
            nextState |= AFTER_CASED;
        }
        uint32_t data = getLetterData(c);
        if (data > 0) {
            uint32_t upper = data & UPPER_MASK;
            // Add a dialytika to this iota or ypsilon vowel
            // if we removed a tonos from the previous vowel,
            // and that previous vowel did not also have (or gain) a dialytika.
            // Adding one only to the final vowel in a longer sequence
            // (which does not occur in normal writing) would require lookahead.
            // Set the same flag as for preserving an existing dialytika.
            if ((data & HAS_VOWEL) != 0 && (state & AFTER_VOWEL_WITH_ACCENT) != 0 &&
                    (upper == 0x399 || upper == 0x3A5)) {
                data |= HAS_DIALYTIKA;
            }
            int32_t numYpogegrammeni = 0;  // Map each one to a trailing, spacing, capital iota.
            if ((data & HAS_YPOGEGRAMMENI) != 0) {
                numYpogegrammeni = 1;
            }
            // Skip combining diacritics after this Greek letter.
            while (nextIndex < srcLength) {
                uint32_t diacriticData = getDiacriticData(src[nextIndex]);
                if (diacriticData != 0) {
                    data |= diacriticData;
                    if ((diacriticData & HAS_YPOGEGRAMMENI) != 0) {
                        ++numYpogegrammeni;
                    }
                    ++nextIndex;
                } else {
                    break;  // not a Greek diacritic
                }
            }
            if ((data & HAS_VOWEL_AND_ACCENT_AND_DIALYTIKA) == HAS_VOWEL_AND_ACCENT) {
                nextState |= AFTER_VOWEL_WITH_ACCENT;
            }
            // Map according to Greek rules.
            UBool addTonos = FALSE;
            if (upper == 0x397 &&
                    (data & HAS_ACCENT) != 0 &&
                    numYpogegrammeni == 0 &&
                    (state & AFTER_CASED) == 0 &&
                    !isFollowedByCasedLetter(src, nextIndex, srcLength)) {
                // Keep disjunctive "or" with (only) a tonos.
                // We use the same "word boundary" conditions as for the Final_Sigma test.
                if (i == nextIndex) {
                    upper = 0x389;  // Preserve the precomposed form.
                } else {
                    addTonos = TRUE;
                }
            } else if ((data & HAS_DIALYTIKA) != 0) {
                // Preserve a vowel with dialytika in precomposed form if it exists.
                if (upper == 0x399) {
                    upper = 0x3AA;
                    data &= ~HAS_EITHER_DIALYTIKA;
                } else if (upper == 0x3A5) {
                    upper = 0x3AB;
                    data &= ~HAS_EITHER_DIALYTIKA;
                }
            }

            UBool change;
            if (edits == nullptr && (options & U_OMIT_UNCHANGED_TEXT) == 0) {
                change = TRUE;  // common, simple usage
            } else {
                // Find out first whether we are changing the text.
                change = src[i] != upper || numYpogegrammeni > 0;
                int32_t i2 = i + 1;
                if ((data & HAS_EITHER_DIALYTIKA) != 0) {
                    change |= i2 >= nextIndex || src[i2] != 0x308;
                    ++i2;
                }
                if (addTonos) {
                    change |= i2 >= nextIndex || src[i2] != 0x301;
                    ++i2;
                }
                int32_t oldLength = nextIndex - i;
                int32_t newLength = (i2 - i) + numYpogegrammeni;
                change |= oldLength != newLength;
                if (change) {
                    if (edits != NULL) {
                        edits->addReplace(oldLength, newLength);
                    }
                } else {
                    if (edits != NULL) {
                        edits->addUnchanged(oldLength);
                    }
                    // Write unchanged text?
                    change = (options & U_OMIT_UNCHANGED_TEXT) == 0;
                }
            }

            if (change) {
                destIndex=appendUChar(dest, destIndex, destCapacity, (UChar)upper);
                if (destIndex >= 0 && (data & HAS_EITHER_DIALYTIKA) != 0) {
                    destIndex=appendUChar(dest, destIndex, destCapacity, 0x308);  // restore or add a dialytika
                }
                if (destIndex >= 0 && addTonos) {
                    destIndex=appendUChar(dest, destIndex, destCapacity, 0x301);
                }
                while (destIndex >= 0 && numYpogegrammeni > 0) {
                    destIndex=appendUChar(dest, destIndex, destCapacity, 0x399);
                    --numYpogegrammeni;
                }
                if(destIndex<0) {
                    errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
                    return 0;
                }
            }
        } else {
            const UChar *s;
            c=ucase_toFullUpper(c, NULL, NULL, &s, UCASE_LOC_GREEK);
            destIndex = appendResult(dest, destIndex, destCapacity, c, s,
                                     nextIndex - i, options, edits);
            if (destIndex < 0) {
                errorCode = U_INDEX_OUTOFBOUNDS_ERROR;
                return 0;
            }
        }
        i = nextIndex;
        state = nextState;
    }

    return destIndex;
}

}  // namespace GreekUpper
U_NAMESPACE_END

/* functions available in the common library (for unistr_case.cpp) */

U_CFUNC int32_t U_CALLCONV
ustrcase_internalToLower(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_UNUSED
                         UChar *dest, int32_t destCapacity,
                         const UChar *src, int32_t srcLength,
                         icu::Edits *edits,
                         UErrorCode &errorCode) {
    UCaseContext csc=UCASECONTEXT_INITIALIZER;
    csc.p=(void *)src;
    csc.limit=srcLength;
    int32_t destIndex = toLower(
        caseLocale, options,
        dest, destCapacity,
        src, &csc, 0, srcLength,
        edits, errorCode);
    return checkOverflowAndEditsError(destIndex, destCapacity, edits, errorCode);
}

U_CFUNC int32_t U_CALLCONV
ustrcase_internalToUpper(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_UNUSED
                         UChar *dest, int32_t destCapacity,
                         const UChar *src, int32_t srcLength,
                         icu::Edits *edits,
                         UErrorCode &errorCode) {
    int32_t destIndex;
    if (caseLocale == UCASE_LOC_GREEK) {
        destIndex = GreekUpper::toUpper(options, dest, destCapacity,
                                        src, srcLength, edits, errorCode);
    } else {
        UCaseContext csc=UCASECONTEXT_INITIALIZER;
        csc.p=(void *)src;
        csc.limit=srcLength;
        destIndex = toUpper(
            caseLocale, options,
            dest, destCapacity,
            src, &csc, srcLength,
            edits, errorCode);
    }
    return checkOverflowAndEditsError(destIndex, destCapacity, edits, errorCode);
}

U_CFUNC int32_t U_CALLCONV
ustrcase_internalFold(int32_t /* caseLocale */, uint32_t options, UCASEMAP_BREAK_ITERATOR_UNUSED
                      UChar *dest, int32_t destCapacity,
                      const UChar *src, int32_t srcLength,
                      icu::Edits *edits,
                      UErrorCode &errorCode) {
    int32_t destIndex = toLower(
        -1, options,
        dest, destCapacity,
        src, nullptr, 0, srcLength,
        edits, errorCode);
    return checkOverflowAndEditsError(destIndex, destCapacity, edits, errorCode);
}

U_CFUNC int32_t
ustrcase_map(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
             UChar *dest, int32_t destCapacity,
             const UChar *src, int32_t srcLength,
             UStringCaseMapper *stringCaseMapper,
             icu::Edits *edits,
             UErrorCode &errorCode) {
    int32_t destLength;

    /* check argument values */
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    if( destCapacity<0 ||
        (dest==NULL && destCapacity>0) ||
        src==NULL ||
        srcLength<-1
    ) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* get the string length */
    if(srcLength==-1) {
        srcLength=u_strlen(src);
    }

    /* check for overlapping source and destination */
    if( dest!=NULL &&
        ((src>=dest && src<(dest+destCapacity)) ||
         (dest>=src && dest<(src+srcLength)))
    ) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if (edits != nullptr && (options & U_EDITS_NO_RESET) == 0) {
        edits->reset();
    }
    destLength=stringCaseMapper(caseLocale, options, UCASEMAP_BREAK_ITERATOR
                                dest, destCapacity, src, srcLength, edits, errorCode);
    return u_terminateUChars(dest, destCapacity, destLength, &errorCode);
}

U_CFUNC int32_t
ustrcase_mapWithOverlap(int32_t caseLocale, uint32_t options, UCASEMAP_BREAK_ITERATOR_PARAM
                        UChar *dest, int32_t destCapacity,
                        const UChar *src, int32_t srcLength,
                        UStringCaseMapper *stringCaseMapper,
                        UErrorCode &errorCode) {
    UChar buffer[300];
    UChar *temp;

    int32_t destLength;

    /* check argument values */
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    if( destCapacity<0 ||
        (dest==NULL && destCapacity>0) ||
        src==NULL ||
        srcLength<-1
    ) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* get the string length */
    if(srcLength==-1) {
        srcLength=u_strlen(src);
    }

    /* check for overlapping source and destination */
    if( dest!=NULL &&
        ((src>=dest && src<(dest+destCapacity)) ||
         (dest>=src && dest<(src+srcLength)))
    ) {
        /* overlap: provide a temporary destination buffer and later copy the result */
        if(destCapacity<=UPRV_LENGTHOF(buffer)) {
            /* the stack buffer is large enough */
            temp=buffer;
        } else {
            /* allocate a buffer */
            temp=(UChar *)uprv_malloc(destCapacity*U_SIZEOF_UCHAR);
            if(temp==NULL) {
                errorCode=U_MEMORY_ALLOCATION_ERROR;
                return 0;
            }
        }
    } else {
        temp=dest;
    }

    destLength=stringCaseMapper(caseLocale, options, UCASEMAP_BREAK_ITERATOR
                                temp, destCapacity, src, srcLength, NULL, errorCode);
    if(temp!=dest) {
        /* copy the result string to the destination buffer */
        if (U_SUCCESS(errorCode) && 0 < destLength && destLength <= destCapacity) {
            u_memmove(dest, temp, destLength);
        }
        if(temp!=buffer) {
            uprv_free(temp);
        }
    }

    return u_terminateUChars(dest, destCapacity, destLength, &errorCode);
}

/* public API functions */

U_CAPI int32_t U_EXPORT2
u_strFoldCase(UChar *dest, int32_t destCapacity,
              const UChar *src, int32_t srcLength,
              uint32_t options,
              UErrorCode *pErrorCode) {
    return ustrcase_mapWithOverlap(
        UCASE_LOC_ROOT, options, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ustrcase_internalFold, *pErrorCode);
}

U_NAMESPACE_BEGIN

int32_t CaseMap::fold(
        uint32_t options,
        const UChar *src, int32_t srcLength,
        UChar *dest, int32_t destCapacity, Edits *edits,
        UErrorCode &errorCode) {
    return ustrcase_map(
        UCASE_LOC_ROOT, options, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ustrcase_internalFold, edits, errorCode);
}

U_NAMESPACE_END

/* case-insensitive string comparisons -------------------------------------- */

/*
 * This function is a copy of unorm_cmpEquivFold() minus the parts for
 * canonical equivalence.
 * Keep the functions in sync, and see there for how this works.
 * The duplication is for modularization:
 * It makes caseless (but not canonical caseless) matches independent of
 * the normalization code.
 */

/* stack element for previous-level source/decomposition pointers */
struct ustrcase_CmpEquivLevel {
    const UChar *start, *s, *limit;
};
typedef struct ustrcase_CmpEquivLevel ustrcase_CmpEquivLevel;

/**
 * Internal implementation code comparing string with case fold.
 * This function is called from u_strcmpFold() and u_caseInsensitivePrefixMatch().
 *
 * @param s1            input string 1
 * @param length1       length of string 1, or -1 (NULL terminated)
 * @param s2            input string 2
 * @param length2       length of string 2, or -1 (NULL terminated)
 * @param options       compare options
 * @param matchLen1     (output) length of partial prefix match in s1
 * @param matchLen2     (output) length of partial prefix match in s2
 * @param pErrorCode    receives error status
 * @return The result of comparison
 */
static int32_t _cmpFold(
            const UChar *s1, int32_t length1,
            const UChar *s2, int32_t length2,
            uint32_t options,
            int32_t *matchLen1, int32_t *matchLen2,
            UErrorCode *pErrorCode) {
    int32_t cmpRes = 0;

    /* current-level start/limit - s1/s2 as current */
    const UChar *start1, *start2, *limit1, *limit2;

    /* points to the original start address */
    const UChar *org1, *org2;

    /* points to the end of match + 1 */
    const UChar *m1, *m2;

    /* case folding variables */
    const UChar *p;
    int32_t length;

    /* stacks of previous-level start/current/limit */
    ustrcase_CmpEquivLevel stack1[2], stack2[2];

    /* case folding buffers, only use current-level start/limit */
    UChar fold1[UCASE_MAX_STRING_LENGTH+1], fold2[UCASE_MAX_STRING_LENGTH+1];

    /* track which is the current level per string */
    int32_t level1, level2;

    /* current code units, and code points for lookups */
    UChar32 c1, c2, cp1, cp2;

    /* no argument error checking because this itself is not an API */

    /*
     * assume that at least the option U_COMPARE_IGNORE_CASE is set
     * otherwise this function would have to behave exactly as uprv_strCompare()
     */
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }

    /* initialize */
    if(matchLen1) {
        U_ASSERT(matchLen2 !=NULL);
        *matchLen1=0;
        *matchLen2=0;
    }

    start1=m1=org1=s1;
    if(length1==-1) {
        limit1=NULL;
    } else {
        limit1=s1+length1;
    }

    start2=m2=org2=s2;
    if(length2==-1) {
        limit2=NULL;
    } else {
        limit2=s2+length2;
    }

    level1=level2=0;
    c1=c2=-1;

    /* comparison loop */
    for(;;) {
        /*
         * here a code unit value of -1 means "get another code unit"
         * below it will mean "this source is finished"
         */

        if(c1<0) {
            /* get next code unit from string 1, post-increment */
            for(;;) {
                if(s1==limit1 || ((c1=*s1)==0 && (limit1==NULL || (options&_STRNCMP_STYLE)))) {
                    if(level1==0) {
                        c1=-1;
                        break;
                    }
                } else {
                    ++s1;
                    break;
                }

                /* reached end of level buffer, pop one level */
                do {
                    --level1;
                    start1=stack1[level1].start;    /*Not uninitialized*/
                } while(start1==NULL);
                s1=stack1[level1].s;                /*Not uninitialized*/
                limit1=stack1[level1].limit;        /*Not uninitialized*/
            }
        }

        if(c2<0) {
            /* get next code unit from string 2, post-increment */
            for(;;) {
                if(s2==limit2 || ((c2=*s2)==0 && (limit2==NULL || (options&_STRNCMP_STYLE)))) {
                    if(level2==0) {
                        c2=-1;
                        break;
                    }
                } else {
                    ++s2;
                    break;
                }

                /* reached end of level buffer, pop one level */
                do {
                    --level2;
                    start2=stack2[level2].start;    /*Not uninitialized*/
                } while(start2==NULL);
                s2=stack2[level2].s;                /*Not uninitialized*/
                limit2=stack2[level2].limit;        /*Not uninitialized*/
            }
        }

        /*
         * compare c1 and c2
         * either variable c1, c2 is -1 only if the corresponding string is finished
         */
        if(c1==c2) {
            const UChar *next1, *next2;

            if(c1<0) {
                cmpRes=0;   /* c1==c2==-1 indicating end of strings */
                break;
            }

            /*
             * Note: Move the match positions in both strings at the same time
             *      only when corresponding code point(s) in the original strings
             *      are fully consumed. For example, when comparing s1="Fust" and
             *      s2="Fu\u00dfball", s2[2] is folded into "ss", and s1[2] matches
             *      the first code point in the case-folded data. But the second "s"
             *      has no matching code point in s1, so this implementation returns
             *      2 as the prefix match length ("Fu").
             */
            next1=next2=NULL;
            if(level1==0) {
                next1=s1;
            } else if(s1==limit1) {
                /* Note: This implementation only use a single level of stack.
                 *      If this code needs to be changed to use multiple levels
                 *      of stacks, the code above should check if the current
                 *      code is at the end of all stacks.
                 */
                U_ASSERT(level1==1);

                /* is s1 at the end of the current stack? */
                next1=stack1[0].s;
            }

            if (next1!=NULL) {
                if(level2==0) {
                    next2=s2;
                } else if(s2==limit2) {
                    U_ASSERT(level2==1);

                    /* is s2 at the end of the current stack? */
                    next2=stack2[0].s;
                }
                if(next2!=NULL) {
                    m1=next1;
                    m2=next2;
                }
            }
            c1=c2=-1;       /* make us fetch new code units */
            continue;
        } else if(c1<0) {
            cmpRes=-1;      /* string 1 ends before string 2 */
            break;
        } else if(c2<0) {
            cmpRes=1;       /* string 2 ends before string 1 */
            break;
        }
        /* c1!=c2 && c1>=0 && c2>=0 */

        /* get complete code points for c1, c2 for lookups if either is a surrogate */
        cp1=c1;
        if(U_IS_SURROGATE(c1)) {
            UChar c;

            if(U_IS_SURROGATE_LEAD(c1)) {
                if(s1!=limit1 && U16_IS_TRAIL(c=*s1)) {
                    /* advance ++s1; only below if cp1 decomposes/case-folds */
                    cp1=U16_GET_SUPPLEMENTARY(c1, c);
                }
            } else /* isTrail(c1) */ {
                if(start1<=(s1-2) && U16_IS_LEAD(c=*(s1-2))) {
                    cp1=U16_GET_SUPPLEMENTARY(c, c1);
                }
            }
        }

        cp2=c2;
        if(U_IS_SURROGATE(c2)) {
            UChar c;

            if(U_IS_SURROGATE_LEAD(c2)) {
                if(s2!=limit2 && U16_IS_TRAIL(c=*s2)) {
                    /* advance ++s2; only below if cp2 decomposes/case-folds */
                    cp2=U16_GET_SUPPLEMENTARY(c2, c);
                }
            } else /* isTrail(c2) */ {
                if(start2<=(s2-2) && U16_IS_LEAD(c=*(s2-2))) {
                    cp2=U16_GET_SUPPLEMENTARY(c, c2);
                }
            }
        }

        /*
         * go down one level for each string
         * continue with the main loop as soon as there is a real change
         */

        if( level1==0 &&
            (length=ucase_toFullFolding((UChar32)cp1, &p, options))>=0
        ) {
            /* cp1 case-folds to the code point "length" or to p[length] */
            if(U_IS_SURROGATE(c1)) {
                if(U_IS_SURROGATE_LEAD(c1)) {
                    /* advance beyond source surrogate pair if it case-folds */
                    ++s1;
                } else /* isTrail(c1) */ {
                    /*
                     * we got a supplementary code point when hitting its trail surrogate,
                     * therefore the lead surrogate must have been the same as in the other string;
                     * compare this decomposition with the lead surrogate in the other string
                     * remember that this simulates bulk text replacement:
                     * the decomposition would replace the entire code point
                     */
                    --s2;
                    --m2;
                    c2=*(s2-1);
                }
            }

            /* push current level pointers */
            stack1[0].start=start1;
            stack1[0].s=s1;
            stack1[0].limit=limit1;
            ++level1;

            /* copy the folding result to fold1[] */
            if(length<=UCASE_MAX_STRING_LENGTH) {
                u_memcpy(fold1, p, length);
            } else {
                int32_t i=0;
                U16_APPEND_UNSAFE(fold1, i, length);
                length=i;
            }

            /* set next level pointers to case folding */
            start1=s1=fold1;
            limit1=fold1+length;

            /* get ready to read from decomposition, continue with loop */
            c1=-1;
            continue;
        }

        if( level2==0 &&
            (length=ucase_toFullFolding((UChar32)cp2, &p, options))>=0
        ) {
            /* cp2 case-folds to the code point "length" or to p[length] */
            if(U_IS_SURROGATE(c2)) {
                if(U_IS_SURROGATE_LEAD(c2)) {
                    /* advance beyond source surrogate pair if it case-folds */
                    ++s2;
                } else /* isTrail(c2) */ {
                    /*
                     * we got a supplementary code point when hitting its trail surrogate,
                     * therefore the lead surrogate must have been the same as in the other string;
                     * compare this decomposition with the lead surrogate in the other string
                     * remember that this simulates bulk text replacement:
                     * the decomposition would replace the entire code point
                     */
                    --s1;
                    --m2;
                    c1=*(s1-1);
                }
            }

            /* push current level pointers */
            stack2[0].start=start2;
            stack2[0].s=s2;
            stack2[0].limit=limit2;
            ++level2;

            /* copy the folding result to fold2[] */
            if(length<=UCASE_MAX_STRING_LENGTH) {
                u_memcpy(fold2, p, length);
            } else {
                int32_t i=0;
                U16_APPEND_UNSAFE(fold2, i, length);
                length=i;
            }

            /* set next level pointers to case folding */
            start2=s2=fold2;
            limit2=fold2+length;

            /* get ready to read from decomposition, continue with loop */
            c2=-1;
            continue;
        }

        /*
         * no decomposition/case folding, max level for both sides:
         * return difference result
         *
         * code point order comparison must not just return cp1-cp2
         * because when single surrogates are present then the surrogate pairs
         * that formed cp1 and cp2 may be from different string indexes
         *
         * example: { d800 d800 dc01 } vs. { d800 dc00 }, compare at second code units
         * c1=d800 cp1=10001 c2=dc00 cp2=10000
         * cp1-cp2>0 but c1-c2<0 and in fact in UTF-32 it is { d800 10001 } < { 10000 }
         *
         * therefore, use same fix-up as in ustring.c/uprv_strCompare()
         * except: uprv_strCompare() fetches c=*s while this functions fetches c=*s++
         * so we have slightly different pointer/start/limit comparisons here
         */

        if(c1>=0xd800 && c2>=0xd800 && (options&U_COMPARE_CODE_POINT_ORDER)) {
            /* subtract 0x2800 from BMP code points to make them smaller than supplementary ones */
            if(
                (c1<=0xdbff && s1!=limit1 && U16_IS_TRAIL(*s1)) ||
                (U16_IS_TRAIL(c1) && start1!=(s1-1) && U16_IS_LEAD(*(s1-2)))
            ) {
                /* part of a surrogate pair, leave >=d800 */
            } else {
                /* BMP code point - may be surrogate code point - make <d800 */
                c1-=0x2800;
            }

            if(
                (c2<=0xdbff && s2!=limit2 && U16_IS_TRAIL(*s2)) ||
                (U16_IS_TRAIL(c2) && start2!=(s2-1) && U16_IS_LEAD(*(s2-2)))
            ) {
                /* part of a surrogate pair, leave >=d800 */
            } else {
                /* BMP code point - may be surrogate code point - make <d800 */
                c2-=0x2800;
            }
        }

        cmpRes=c1-c2;
        break;
    }

    if(matchLen1) {
        *matchLen1=static_cast<int32_t>(m1-org1);
        *matchLen2=static_cast<int32_t>(m2-org2);
    }
    return cmpRes;
}

/* internal function */
U_CFUNC int32_t
u_strcmpFold(const UChar *s1, int32_t length1,
             const UChar *s2, int32_t length2,
             uint32_t options,
             UErrorCode *pErrorCode) {
    return _cmpFold(s1, length1, s2, length2, options, NULL, NULL, pErrorCode);
}

/* public API functions */

U_CAPI int32_t U_EXPORT2
u_strCaseCompare(const UChar *s1, int32_t length1,
                 const UChar *s2, int32_t length2,
                 uint32_t options,
                 UErrorCode *pErrorCode) {
    /* argument checking */
    if(pErrorCode==0 || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(s1==NULL || length1<-1 || s2==NULL || length2<-1) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    return u_strcmpFold(s1, length1, s2, length2,
                        options|U_COMPARE_IGNORE_CASE,
                        pErrorCode);
}

U_CAPI int32_t U_EXPORT2
u_strcasecmp(const UChar *s1, const UChar *s2, uint32_t options) {
    UErrorCode errorCode=U_ZERO_ERROR;
    return u_strcmpFold(s1, -1, s2, -1,
                        options|U_COMPARE_IGNORE_CASE,
                        &errorCode);
}

U_CAPI int32_t U_EXPORT2
u_memcasecmp(const UChar *s1, const UChar *s2, int32_t length, uint32_t options) {
    UErrorCode errorCode=U_ZERO_ERROR;
    return u_strcmpFold(s1, length, s2, length,
                        options|U_COMPARE_IGNORE_CASE,
                        &errorCode);
}

U_CAPI int32_t U_EXPORT2
u_strncasecmp(const UChar *s1, const UChar *s2, int32_t n, uint32_t options) {
    UErrorCode errorCode=U_ZERO_ERROR;
    return u_strcmpFold(s1, n, s2, n,
                        options|(U_COMPARE_IGNORE_CASE|_STRNCMP_STYLE),
                        &errorCode);
}

/* internal API - detect length of shared prefix */
U_CAPI void
u_caseInsensitivePrefixMatch(const UChar *s1, int32_t length1,
                             const UChar *s2, int32_t length2,
                             uint32_t options,
                             int32_t *matchLen1, int32_t *matchLen2,
                             UErrorCode *pErrorCode) {
    _cmpFold(s1, length1, s2, length2, options,
        matchLen1, matchLen2, pErrorCode);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  ustrcase_locale.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011may31
*   created by: Markus W. Scherer
*
*   Locale-sensitive case mapping functions (ones that call uloc_getDefault())
*   were moved here to break dependency cycles among parts of the common library.
*/











U_CFUNC int32_t
ustrcase_getCaseLocale(const char *locale) {
    if (locale == NULL) {
        locale = uloc_getDefault();
    }
    if (*locale == 0) {
        return UCASE_LOC_ROOT;
    } else {
        return ucase_getCaseLocale(locale);
    }
}

/* public API functions */

U_CAPI int32_t U_EXPORT2
u_strToLower(UChar *dest, int32_t destCapacity,
             const UChar *src, int32_t srcLength,
             const char *locale,
             UErrorCode *pErrorCode) {
    return ustrcase_mapWithOverlap(
        ustrcase_getCaseLocale(locale), 0, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ustrcase_internalToLower, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
u_strToUpper(UChar *dest, int32_t destCapacity,
             const UChar *src, int32_t srcLength,
             const char *locale,
             UErrorCode *pErrorCode) {
    return ustrcase_mapWithOverlap(
        ustrcase_getCaseLocale(locale), 0, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ustrcase_internalToUpper, *pErrorCode);
}

U_NAMESPACE_BEGIN

int32_t CaseMap::toLower(
        const char *locale, uint32_t options,
        const UChar *src, int32_t srcLength,
        UChar *dest, int32_t destCapacity, Edits *edits,
        UErrorCode &errorCode) {
    return ustrcase_map(
        ustrcase_getCaseLocale(locale), options, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ustrcase_internalToLower, edits, errorCode);
}

int32_t CaseMap::toUpper(
        const char *locale, uint32_t options,
        const UChar *src, int32_t srcLength,
        UChar *dest, int32_t destCapacity, Edits *edits,
        UErrorCode &errorCode) {
    return ustrcase_map(
        ustrcase_getCaseLocale(locale), options, UCASEMAP_BREAK_ITERATOR_NULL
        dest, destCapacity,
        src, srcLength,
        ustrcase_internalToUpper, edits, errorCode);
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2002-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: November 11 2002
* Since: ICU 2.4
**********************************************************************
*/
  // for 'typeid' to work





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2002-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: November 11 2002
* Since: ICU 2.4
**********************************************************************
*/




//----------------------------------------------------------------------
U_NAMESPACE_BEGIN

/**
 * A wrapper to make a UEnumeration into a StringEnumeration.  The
 * wrapper adopts the UEnumeration is wraps.
 */
class U_COMMON_API UStringEnumeration : public StringEnumeration {

public:
    /**
     * Constructor.  This constructor adopts its UEnumeration
     * argument.
     * @param uenum a UEnumeration object.  This object takes
     * ownership of 'uenum' and will close it in its destructor.  The
     * caller must not call uenum_close on 'uenum' after calling this
     * constructor.
     */
    UStringEnumeration(UEnumeration* uenum);

    /**
     * Destructor.  This closes the UEnumeration passed in to the
     * constructor.
     */
    virtual ~UStringEnumeration();

    /**
     * Return the number of elements that the iterator traverses.
     * @param status the error code.
     * @return number of elements in the iterator.
     */
    virtual int32_t count(UErrorCode& status) const;

    virtual const char* next(int32_t *resultLength, UErrorCode& status);

    /**
     * Returns the next element a UnicodeString*.  If there are no
     * more elements, returns NULL.
     * @param status the error code.
     * @return a pointer to the string, or NULL.
     */
    virtual const UnicodeString* snext(UErrorCode& status);

    /**
     * Resets the iterator.
     * @param status the error code.
     */
    virtual void reset(UErrorCode& status);

    /**
     * ICU4C "poor man's RTTI", returns a UClassID for the actual ICU class.
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * ICU4C "poor man's RTTI", returns a UClassID for this ICU class.
     */
    static UClassID U_EXPORT2 getStaticClassID();

    static UStringEnumeration * U_EXPORT2 fromUEnumeration(
            UEnumeration *enumToAdopt, UErrorCode &status);
private:
    UEnumeration *uenum; // owned
};

U_NAMESPACE_END






U_NAMESPACE_BEGIN
// StringEnumeration implementation ---------------------------------------- ***

StringEnumeration::StringEnumeration()
    : chars(charsBuffer), charsCapacity(sizeof(charsBuffer)) {
}

StringEnumeration::~StringEnumeration() {
    if (chars != NULL && chars != charsBuffer) {
        uprv_free(chars);
    }
}

// StringEnumeration base class clone() default implementation, does not clone
StringEnumeration *
StringEnumeration::clone() const {
  return NULL;
}

const char *
StringEnumeration::next(int32_t *resultLength, UErrorCode &status) {
    const UnicodeString *s=snext(status);
    if(U_SUCCESS(status) && s!=NULL) {
        unistr=*s;
        ensureCharsCapacity(unistr.length()+1, status);
        if(U_SUCCESS(status)) {
            if(resultLength!=NULL) {
                *resultLength=unistr.length();
            }
            unistr.extract(0, INT32_MAX, chars, charsCapacity, US_INV);
            return chars;
        }
    }

    return NULL;
}

const UChar *
StringEnumeration::unext(int32_t *resultLength, UErrorCode &status) {
    const UnicodeString *s=snext(status);
    if(U_SUCCESS(status) && s!=NULL) {
        unistr=*s;
        if(resultLength!=NULL) {
            *resultLength=unistr.length();
        }
        return unistr.getTerminatedBuffer();
    }

    return NULL;
}

const UnicodeString *
StringEnumeration::snext(UErrorCode &status) {
    int32_t length;
    const char *s=next(&length, status);
    return setChars(s, length, status);
}

void
StringEnumeration::ensureCharsCapacity(int32_t capacity, UErrorCode &status) {
    if(U_SUCCESS(status) && capacity>charsCapacity) {
        if(capacity<(charsCapacity+charsCapacity/2)) {
            // avoid allocation thrashing
            capacity=charsCapacity+charsCapacity/2;
        }
        if(chars!=charsBuffer) {
            uprv_free(chars);
        }
        chars=(char *)uprv_malloc(capacity);
        if(chars==NULL) {
            chars=charsBuffer;
            charsCapacity=sizeof(charsBuffer);
            status=U_MEMORY_ALLOCATION_ERROR;
        } else {
            charsCapacity=capacity;
        }
    }
}

UnicodeString *
StringEnumeration::setChars(const char *s, int32_t length, UErrorCode &status) {
    if(U_SUCCESS(status) && s!=NULL) {
        if(length<0) {
            length=(int32_t)uprv_strlen(s);
        }

        UChar *buffer=unistr.getBuffer(length+1);
        if(buffer!=NULL) {
            u_charsToUChars(s, buffer, length);
            buffer[length]=0;
            unistr.releaseBuffer(length);
            return &unistr;
        } else {
            status=U_MEMORY_ALLOCATION_ERROR;
        }
    }

    return NULL;
}
UBool
StringEnumeration::operator==(const StringEnumeration& that)const {
    return typeid(*this) == typeid(that);
}

UBool
StringEnumeration::operator!=(const StringEnumeration& that)const {
    return !operator==(that);
}

// UStringEnumeration implementation --------------------------------------- ***

UStringEnumeration * U_EXPORT2
UStringEnumeration::fromUEnumeration(
        UEnumeration *uenumToAdopt, UErrorCode &status) {
    if (U_FAILURE(status)) {
        uenum_close(uenumToAdopt);
        return NULL;
    }
    UStringEnumeration *result = new UStringEnumeration(uenumToAdopt);
    if (result == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        uenum_close(uenumToAdopt);
        return NULL;
    }
    return result;
}

UStringEnumeration::UStringEnumeration(UEnumeration* _uenum) :
    uenum(_uenum) {
    U_ASSERT(_uenum != 0);
}

UStringEnumeration::~UStringEnumeration() {
    uenum_close(uenum);
}

int32_t UStringEnumeration::count(UErrorCode& status) const {
    return uenum_count(uenum, &status);
}

const char *UStringEnumeration::next(int32_t *resultLength, UErrorCode &status) {
    return uenum_next(uenum, resultLength, &status);
}

const UnicodeString* UStringEnumeration::snext(UErrorCode& status) {
    int32_t length;
    const UChar* str = uenum_unext(uenum, &length, &status);
    if (str == 0 || U_FAILURE(status)) {
        return 0;
    }
    return &unistr.setTo(str, length);
}

void UStringEnumeration::reset(UErrorCode& status) {
    uenum_reset(uenum, &status);
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(UStringEnumeration)
U_NAMESPACE_END

// C wrapper --------------------------------------------------------------- ***

#define ustrenum_THIS(en) ((icu::StringEnumeration*)(en->context))

U_CDECL_BEGIN

/**
 * Wrapper API to make StringEnumeration look like UEnumeration.
 */
static void U_CALLCONV
ustrenum_close(UEnumeration* en) {
    delete ustrenum_THIS(en);
    uprv_free(en);
}

/**
 * Wrapper API to make StringEnumeration look like UEnumeration.
 */
static int32_t U_CALLCONV
ustrenum_count(UEnumeration* en,
               UErrorCode* ec)
{
    return ustrenum_THIS(en)->count(*ec);
}

/**
 * Wrapper API to make StringEnumeration look like UEnumeration.
 */
static const UChar* U_CALLCONV
ustrenum_unext(UEnumeration* en,
               int32_t* resultLength,
               UErrorCode* ec)
{
    return ustrenum_THIS(en)->unext(resultLength, *ec);
}

/**
 * Wrapper API to make StringEnumeration look like UEnumeration.
 */
static const char* U_CALLCONV
ustrenum_next(UEnumeration* en,
              int32_t* resultLength,
              UErrorCode* ec)
{
    return ustrenum_THIS(en)->next(resultLength, *ec);
}

/**
 * Wrapper API to make StringEnumeration look like UEnumeration.
 */
static void U_CALLCONV
ustrenum_reset(UEnumeration* en,
               UErrorCode* ec)
{
    ustrenum_THIS(en)->reset(*ec);
}

/**
 * Pseudo-vtable for UEnumeration wrapper around StringEnumeration.
 * The StringEnumeration pointer will be stored in 'context'.
 */
static const UEnumeration USTRENUM_VT = {
    NULL,
    NULL, // store StringEnumeration pointer here
    ustrenum_close,
    ustrenum_count,
    ustrenum_unext,
    ustrenum_next,
    ustrenum_reset
};

U_CDECL_END

/**
 * Given a StringEnumeration, wrap it in a UEnumeration.  The
 * StringEnumeration is adopted; after this call, the caller must not
 * delete it (regardless of error status).
 */
U_CAPI UEnumeration* U_EXPORT2
uenum_openFromStringEnumeration(icu::StringEnumeration* adopted, UErrorCode* ec) {
    UEnumeration* result = NULL;
    if (U_SUCCESS(*ec) && adopted != NULL) {
        result = (UEnumeration*) uprv_malloc(sizeof(UEnumeration));
        if (result == NULL) {
            *ec = U_MEMORY_ALLOCATION_ERROR;
        } else {
            uprv_memcpy(result, &USTRENUM_VT, sizeof(USTRENUM_VT));
            result->context = adopted;
        }
    }
    if (result == NULL) {
        delete adopted;
    }
    return result;
}

// C wrapper --------------------------------------------------------------- ***

U_CDECL_BEGIN

typedef struct UCharStringEnumeration {
    UEnumeration uenum;
    int32_t index, count;
} UCharStringEnumeration;

static void U_CALLCONV
ucharstrenum_close(UEnumeration* en) {
    uprv_free(en);
}

static int32_t U_CALLCONV
ucharstrenum_count(UEnumeration* en,
                   UErrorCode* /*ec*/) {
    return ((UCharStringEnumeration*)en)->count;
}

static const UChar* U_CALLCONV
ucharstrenum_unext(UEnumeration* en,
                  int32_t* resultLength,
                  UErrorCode* /*ec*/) {
    UCharStringEnumeration *e = (UCharStringEnumeration*) en;
    if (e->index >= e->count) {
        return NULL;
    }
    const UChar* result = ((const UChar**)e->uenum.context)[e->index++];
    if (resultLength) {
        *resultLength = (int32_t)u_strlen(result);
    }
    return result;
}


static const char* U_CALLCONV
ucharstrenum_next(UEnumeration* en,
                  int32_t* resultLength,
                  UErrorCode* /*ec*/) {
    UCharStringEnumeration *e = (UCharStringEnumeration*) en;
    if (e->index >= e->count) {
        return NULL;
    }
    const char* result = ((const char**)e->uenum.context)[e->index++];
    if (resultLength) {
        *resultLength = (int32_t)uprv_strlen(result);
    }
    return result;
}

static void U_CALLCONV
ucharstrenum_reset(UEnumeration* en,
                   UErrorCode* /*ec*/) {
    ((UCharStringEnumeration*)en)->index = 0;
}

static const UEnumeration UCHARSTRENUM_VT = {
    NULL,
    NULL, // store StringEnumeration pointer here
    ucharstrenum_close,
    ucharstrenum_count,
    uenum_unextDefault,
    ucharstrenum_next,
    ucharstrenum_reset
};

static const UEnumeration UCHARSTRENUM_U_VT = {
    NULL,
    NULL, // store StringEnumeration pointer here
    ucharstrenum_close,
    ucharstrenum_count,
    ucharstrenum_unext,
    uenum_nextDefault,
    ucharstrenum_reset
};

U_CDECL_END

U_CAPI UEnumeration* U_EXPORT2
uenum_openCharStringsEnumeration(const char* const strings[], int32_t count,
                                 UErrorCode* ec) {
    UCharStringEnumeration* result = NULL;
    if (U_SUCCESS(*ec) && count >= 0 && (count == 0 || strings != 0)) {
        result = (UCharStringEnumeration*) uprv_malloc(sizeof(UCharStringEnumeration));
        if (result == NULL) {
            *ec = U_MEMORY_ALLOCATION_ERROR;
        } else {
            U_ASSERT((char*)result==(char*)(&result->uenum));
            uprv_memcpy(result, &UCHARSTRENUM_VT, sizeof(UCHARSTRENUM_VT));
            result->uenum.context = (void*)strings;
            result->index = 0;
            result->count = count;
        }
    }
    return (UEnumeration*) result;
}

U_CAPI UEnumeration* U_EXPORT2
uenum_openUCharStringsEnumeration(const UChar* const strings[], int32_t count,
                                 UErrorCode* ec) {
    UCharStringEnumeration* result = NULL;
    if (U_SUCCESS(*ec) && count >= 0 && (count == 0 || strings != 0)) {
        result = (UCharStringEnumeration*) uprv_malloc(sizeof(UCharStringEnumeration));
        if (result == NULL) {
            *ec = U_MEMORY_ALLOCATION_ERROR;
        } else {
            U_ASSERT((char*)result==(char*)(&result->uenum));
            uprv_memcpy(result, &UCHARSTRENUM_U_VT, sizeof(UCHARSTRENUM_U_VT));
            result->uenum.context = (void*)strings;
            result->index = 0;
            result->count = count;
        }
    }
    return (UEnumeration*) result;
}


// end C Wrapper
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2001-2006, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/





/***
 * Fills in a UChar* string with the radix-based representation of a
 * uint32_t number padded with zeroes to minwidth.  The result
 * will be null terminated if there is room.
 *
 * @param buffer UChar buffer to receive result
 * @param capacity capacity of buffer
 * @param i the unsigned number to be formatted
 * @param radix the radix from 2..36
 * @param minwidth the minimum width.  If the result is narrower than
 *        this, '0's will be added on the left.  Must be <=
 *        capacity.
 * @return the length of the result, not including any terminating
 *        null
 */
U_CAPI int32_t U_EXPORT2
uprv_itou (UChar * buffer, int32_t capacity,
           uint32_t i, uint32_t radix, int32_t minwidth)
{
    int32_t length = 0;
    int digit;
    int32_t j;
    UChar temp;

    do{
        digit = (int)(i % radix);
        buffer[length++]=(UChar)(digit<=9?(0x0030+digit):(0x0030+digit+7));
        i=i/radix;
    } while(i && length<capacity);

    while (length < minwidth){
        buffer[length++] = (UChar) 0x0030;/*zero padding */
    }
    /* null terminate the buffer */
    if(length<capacity){
        buffer[length] = (UChar) 0x0000;
    }

    /* Reverses the string */
    for (j = 0; j < (length / 2); j++){
        temp = buffer[(length-1) - j];
        buffer[(length-1) - j] = buffer[j];
        buffer[j] = temp;
    }
    return length;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1998-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File ustring.cpp
*
* Modification History:
*
*   Date        Name        Description
*   12/07/98    bertrand    Creation.
******************************************************************************
*/











/* ANSI string.h - style functions ------------------------------------------ */

/* U+ffff is the highest BMP code point, the highest one that fits into a 16-bit UChar */
#define U_BMP_MAX 0xffff

/* Forward binary string search functions ----------------------------------- */

/*
 * Test if a substring match inside a string is at code point boundaries.
 * All pointers refer to the same buffer.
 * The limit pointer may be NULL, all others must be real pointers.
 */
static inline UBool
isMatchAtCPBoundary(const UChar *start, const UChar *match, const UChar *matchLimit, const UChar *limit) {
    if(U16_IS_TRAIL(*match) && start!=match && U16_IS_LEAD(*(match-1))) {
        /* the leading edge of the match is in the middle of a surrogate pair */
        return FALSE;
    }
    if(U16_IS_LEAD(*(matchLimit-1)) && match!=limit && U16_IS_TRAIL(*matchLimit)) {
        /* the trailing edge of the match is in the middle of a surrogate pair */
        return FALSE;
    }
    return TRUE;
}

U_CAPI UChar * U_EXPORT2
u_strFindFirst(const UChar *s, int32_t length,
               const UChar *sub, int32_t subLength) {
    const UChar *start, *p, *q, *subLimit;
    UChar c, cs, cq;

    if(sub==NULL || subLength<-1) {
        return (UChar *)s;
    }
    if(s==NULL || length<-1) {
        return NULL;
    }

    start=s;

    if(length<0 && subLength<0) {
        /* both strings are NUL-terminated */
        if((cs=*sub++)==0) {
            return (UChar *)s;
        }
        if(*sub==0 && !U16_IS_SURROGATE(cs)) {
            /* the substring consists of a single, non-surrogate BMP code point */
            return u_strchr(s, cs);
        }

        while((c=*s++)!=0) {
            if(c==cs) {
                /* found first substring UChar, compare rest */
                p=s;
                q=sub;
                for(;;) {
                    if((cq=*q)==0) {
                        if(isMatchAtCPBoundary(start, s-1, p, NULL)) {
                            return (UChar *)(s-1); /* well-formed match */
                        } else {
                            break; /* no match because surrogate pair is split */
                        }
                    }
                    if((c=*p)==0) {
                        return NULL; /* no match, and none possible after s */
                    }
                    if(c!=cq) {
                        break; /* no match */
                    }
                    ++p;
                    ++q;
                }
            }
        }

        /* not found */
        return NULL;
    }

    if(subLength<0) {
        subLength=u_strlen(sub);
    }
    if(subLength==0) {
        return (UChar *)s;
    }

    /* get sub[0] to search for it fast */
    cs=*sub++;
    --subLength;
    subLimit=sub+subLength;

    if(subLength==0 && !U16_IS_SURROGATE(cs)) {
        /* the substring consists of a single, non-surrogate BMP code point */
        return length<0 ? u_strchr(s, cs) : u_memchr(s, cs, length);
    }

    if(length<0) {
        /* s is NUL-terminated */
        while((c=*s++)!=0) {
            if(c==cs) {
                /* found first substring UChar, compare rest */
                p=s;
                q=sub;
                for(;;) {
                    if(q==subLimit) {
                        if(isMatchAtCPBoundary(start, s-1, p, NULL)) {
                            return (UChar *)(s-1); /* well-formed match */
                        } else {
                            break; /* no match because surrogate pair is split */
                        }
                    }
                    if((c=*p)==0) {
                        return NULL; /* no match, and none possible after s */
                    }
                    if(c!=*q) {
                        break; /* no match */
                    }
                    ++p;
                    ++q;
                }
            }
        }
    } else {
        const UChar *limit, *preLimit;

        /* subLength was decremented above */
        if(length<=subLength) {
            return NULL; /* s is shorter than sub */
        }

        limit=s+length;

        /* the substring must start before preLimit */
        preLimit=limit-subLength;

        while(s!=preLimit) {
            c=*s++;
            if(c==cs) {
                /* found first substring UChar, compare rest */
                p=s;
                q=sub;
                for(;;) {
                    if(q==subLimit) {
                        if(isMatchAtCPBoundary(start, s-1, p, limit)) {
                            return (UChar *)(s-1); /* well-formed match */
                        } else {
                            break; /* no match because surrogate pair is split */
                        }
                    }
                    if(*p!=*q) {
                        break; /* no match */
                    }
                    ++p;
                    ++q;
                }
            }
        }
    }

    /* not found */
    return NULL;
}

U_CAPI UChar * U_EXPORT2
u_strstr(const UChar *s, const UChar *substring) {
    return u_strFindFirst(s, -1, substring, -1);
}

U_CAPI UChar * U_EXPORT2
u_strchr(const UChar *s, UChar c) {
    if(U16_IS_SURROGATE(c)) {
        /* make sure to not find half of a surrogate pair */
        return u_strFindFirst(s, -1, &c, 1);
    } else {
        UChar cs;

        /* trivial search for a BMP code point */
        for(;;) {
            if((cs=*s)==c) {
                return (UChar *)s;
            }
            if(cs==0) {
                return NULL;
            }
            ++s;
        }
    }
}

U_CAPI UChar * U_EXPORT2
u_strchr32(const UChar *s, UChar32 c) {
    if((uint32_t)c<=U_BMP_MAX) {
        /* find BMP code point */
        return u_strchr(s, (UChar)c);
    } else if((uint32_t)c<=UCHAR_MAX_VALUE) {
        /* find supplementary code point as surrogate pair */
        UChar cs, lead=U16_LEAD(c), trail=U16_TRAIL(c);

        while((cs=*s++)!=0) {
            if(cs==lead && *s==trail) {
                return (UChar *)(s-1);
            }
        }
        return NULL;
    } else {
        /* not a Unicode code point, not findable */
        return NULL;
    }
}

U_CAPI UChar * U_EXPORT2
u_memchr(const UChar *s, UChar c, int32_t count) {
    if(count<=0) {
        return NULL; /* no string */
    } else if(U16_IS_SURROGATE(c)) {
        /* make sure to not find half of a surrogate pair */
        return u_strFindFirst(s, count, &c, 1);
    } else {
        /* trivial search for a BMP code point */
        const UChar *limit=s+count;
        do {
            if(*s==c) {
                return (UChar *)s;
            }
        } while(++s!=limit);
        return NULL;
    }
}

U_CAPI UChar * U_EXPORT2
u_memchr32(const UChar *s, UChar32 c, int32_t count) {
    if((uint32_t)c<=U_BMP_MAX) {
        /* find BMP code point */
        return u_memchr(s, (UChar)c, count);
    } else if(count<2) {
        /* too short for a surrogate pair */
        return NULL;
    } else if((uint32_t)c<=UCHAR_MAX_VALUE) {
        /* find supplementary code point as surrogate pair */
        const UChar *limit=s+count-1; /* -1 so that we do not need a separate check for the trail unit */
        UChar lead=U16_LEAD(c), trail=U16_TRAIL(c);

        do {
            if(*s==lead && *(s+1)==trail) {
                return (UChar *)s;
            }
        } while(++s!=limit);
        return NULL;
    } else {
        /* not a Unicode code point, not findable */
        return NULL;
    }
}

/* Backward binary string search functions ---------------------------------- */

U_CAPI UChar * U_EXPORT2
u_strFindLast(const UChar *s, int32_t length,
              const UChar *sub, int32_t subLength) {
    const UChar *start, *limit, *p, *q, *subLimit;
    UChar c, cs;

    if(sub==NULL || subLength<-1) {
        return (UChar *)s;
    }
    if(s==NULL || length<-1) {
        return NULL;
    }

    /*
     * This implementation is more lazy than the one for u_strFindFirst():
     * There is no special search code for NUL-terminated strings.
     * It does not seem to be worth it for searching substrings to
     * search forward and find all matches like in u_strrchr() and similar.
     * Therefore, we simply get both string lengths and search backward.
     *
     * markus 2002oct23
     */

    if(subLength<0) {
        subLength=u_strlen(sub);
    }
    if(subLength==0) {
        return (UChar *)s;
    }

    /* get sub[subLength-1] to search for it fast */
    subLimit=sub+subLength;
    cs=*(--subLimit);
    --subLength;

    if(subLength==0 && !U16_IS_SURROGATE(cs)) {
        /* the substring consists of a single, non-surrogate BMP code point */
        return length<0 ? u_strrchr(s, cs) : u_memrchr(s, cs, length);
    }

    if(length<0) {
        length=u_strlen(s);
    }

    /* subLength was decremented above */
    if(length<=subLength) {
        return NULL; /* s is shorter than sub */
    }

    start=s;
    limit=s+length;

    /* the substring must start no later than s+subLength */
    s+=subLength;

    while(s!=limit) {
        c=*(--limit);
        if(c==cs) {
            /* found last substring UChar, compare rest */
            p=limit;
            q=subLimit;
            for(;;) {
                if(q==sub) {
                    if(isMatchAtCPBoundary(start, p, limit+1, start+length)) {
                        return (UChar *)p; /* well-formed match */
                    } else {
                        break; /* no match because surrogate pair is split */
                    }
                }
                if(*(--p)!=*(--q)) {
                    break; /* no match */
                }
            }
        }
    }

    /* not found */
    return NULL;
}

U_CAPI UChar * U_EXPORT2
u_strrstr(const UChar *s, const UChar *substring) {
    return u_strFindLast(s, -1, substring, -1);
}

U_CAPI UChar * U_EXPORT2
u_strrchr(const UChar *s, UChar c) {
    if(U16_IS_SURROGATE(c)) {
        /* make sure to not find half of a surrogate pair */
        return u_strFindLast(s, -1, &c, 1);
    } else {
        const UChar *result=NULL;
        UChar cs;

        /* trivial search for a BMP code point */
        for(;;) {
            if((cs=*s)==c) {
                result=s;
            }
            if(cs==0) {
                return (UChar *)result;
            }
            ++s;
        }
    }
}

U_CAPI UChar * U_EXPORT2
u_strrchr32(const UChar *s, UChar32 c) {
    if((uint32_t)c<=U_BMP_MAX) {
        /* find BMP code point */
        return u_strrchr(s, (UChar)c);
    } else if((uint32_t)c<=UCHAR_MAX_VALUE) {
        /* find supplementary code point as surrogate pair */
        const UChar *result=NULL;
        UChar cs, lead=U16_LEAD(c), trail=U16_TRAIL(c);

        while((cs=*s++)!=0) {
            if(cs==lead && *s==trail) {
                result=s-1;
            }
        }
        return (UChar *)result;
    } else {
        /* not a Unicode code point, not findable */
        return NULL;
    }
}

U_CAPI UChar * U_EXPORT2
u_memrchr(const UChar *s, UChar c, int32_t count) {
    if(count<=0) {
        return NULL; /* no string */
    } else if(U16_IS_SURROGATE(c)) {
        /* make sure to not find half of a surrogate pair */
        return u_strFindLast(s, count, &c, 1);
    } else {
        /* trivial search for a BMP code point */
        const UChar *limit=s+count;
        do {
            if(*(--limit)==c) {
                return (UChar *)limit;
            }
        } while(s!=limit);
        return NULL;
    }
}

U_CAPI UChar * U_EXPORT2
u_memrchr32(const UChar *s, UChar32 c, int32_t count) {
    if((uint32_t)c<=U_BMP_MAX) {
        /* find BMP code point */
        return u_memrchr(s, (UChar)c, count);
    } else if(count<2) {
        /* too short for a surrogate pair */
        return NULL;
    } else if((uint32_t)c<=UCHAR_MAX_VALUE) {
        /* find supplementary code point as surrogate pair */
        const UChar *limit=s+count-1;
        UChar lead=U16_LEAD(c), trail=U16_TRAIL(c);

        do {
            if(*limit==trail && *(limit-1)==lead) {
                return (UChar *)(limit-1);
            }
        } while(s!=--limit);
        return NULL;
    } else {
        /* not a Unicode code point, not findable */
        return NULL;
    }
}

/* Tokenization functions --------------------------------------------------- */

/*
 * Match each code point in a string against each code point in the matchSet.
 * Return the index of the first string code point that
 * is (polarity==TRUE) or is not (FALSE) contained in the matchSet.
 * Return -(string length)-1 if there is no such code point.
 */
static int32_t
_matchFromSet(const UChar *string, const UChar *matchSet, UBool polarity) {
    int32_t matchLen, matchBMPLen, strItr, matchItr;
    UChar32 stringCh, matchCh;
    UChar c, c2;

    /* first part of matchSet contains only BMP code points */
    matchBMPLen = 0;
    while((c = matchSet[matchBMPLen]) != 0 && U16_IS_SINGLE(c)) {
        ++matchBMPLen;
    }

    /* second part of matchSet contains BMP and supplementary code points */
    matchLen = matchBMPLen;
    while(matchSet[matchLen] != 0) {
        ++matchLen;
    }

    for(strItr = 0; (c = string[strItr]) != 0;) {
        ++strItr;
        if(U16_IS_SINGLE(c)) {
            if(polarity) {
                for(matchItr = 0; matchItr < matchLen; ++matchItr) {
                    if(c == matchSet[matchItr]) {
                        return strItr - 1; /* one matches */
                    }
                }
            } else {
                for(matchItr = 0; matchItr < matchLen; ++matchItr) {
                    if(c == matchSet[matchItr]) {
                        goto endloop;
                    }
                }
                return strItr - 1; /* none matches */
            }
        } else {
            /*
             * No need to check for string length before U16_IS_TRAIL
             * because c2 could at worst be the terminating NUL.
             */
            if(U16_IS_SURROGATE_LEAD(c) && U16_IS_TRAIL(c2 = string[strItr])) {
                ++strItr;
                stringCh = U16_GET_SUPPLEMENTARY(c, c2);
            } else {
                stringCh = c; /* unpaired trail surrogate */
            }

            if(polarity) {
                for(matchItr = matchBMPLen; matchItr < matchLen;) {
                    U16_NEXT(matchSet, matchItr, matchLen, matchCh);
                    if(stringCh == matchCh) {
                        return strItr - U16_LENGTH(stringCh); /* one matches */
                    }
                }
            } else {
                for(matchItr = matchBMPLen; matchItr < matchLen;) {
                    U16_NEXT(matchSet, matchItr, matchLen, matchCh);
                    if(stringCh == matchCh) {
                        goto endloop;
                    }
                }
                return strItr - U16_LENGTH(stringCh); /* none matches */
            }
        }
endloop:
        /* wish C had continue with labels like Java... */;
    }

    /* Didn't find it. */
    return -strItr-1;
}

/* Search for a codepoint in a string that matches one of the matchSet codepoints. */
U_CAPI UChar * U_EXPORT2
u_strpbrk(const UChar *string, const UChar *matchSet)
{
    int32_t idx = _matchFromSet(string, matchSet, TRUE);
    if(idx >= 0) {
        return (UChar *)string + idx;
    } else {
        return NULL;
    }
}

/* Search for a codepoint in a string that matches one of the matchSet codepoints. */
U_CAPI int32_t U_EXPORT2
u_strcspn(const UChar *string, const UChar *matchSet)
{
    int32_t idx = _matchFromSet(string, matchSet, TRUE);
    if(idx >= 0) {
        return idx;
    } else {
        return -idx - 1; /* == u_strlen(string) */
    }
}

/* Search for a codepoint in a string that does not match one of the matchSet codepoints. */
U_CAPI int32_t U_EXPORT2
u_strspn(const UChar *string, const UChar *matchSet)
{
    int32_t idx = _matchFromSet(string, matchSet, FALSE);
    if(idx >= 0) {
        return idx;
    } else {
        return -idx - 1; /* == u_strlen(string) */
    }
}

/* ----- Text manipulation functions --- */

U_CAPI UChar* U_EXPORT2
u_strtok_r(UChar    *src,
     const UChar    *delim,
           UChar   **saveState)
{
    UChar *tokSource;
    UChar *nextToken;
    uint32_t nonDelimIdx;

    /* If saveState is NULL, the user messed up. */
    if (src != NULL) {
        tokSource = src;
        *saveState = src; /* Set to "src" in case there are no delimiters */
    }
    else if (*saveState) {
        tokSource = *saveState;
    }
    else {
        /* src == NULL && *saveState == NULL */
        /* This shouldn't happen. We already finished tokenizing. */
        return NULL;
    }

    /* Skip initial delimiters */
    nonDelimIdx = u_strspn(tokSource, delim);
    tokSource = &tokSource[nonDelimIdx];

    if (*tokSource) {
        nextToken = u_strpbrk(tokSource, delim);
        if (nextToken != NULL) {
            /* Create a token */
            *(nextToken++) = 0;
            *saveState = nextToken;
            return tokSource;
        }
        else if (*saveState) {
            /* Return the last token */
            *saveState = NULL;
            return tokSource;
        }
    }
    else {
        /* No tokens were found. Only delimiters were left. */
        *saveState = NULL;
    }
    return NULL;
}

/* Miscellaneous functions -------------------------------------------------- */

U_CAPI UChar* U_EXPORT2
u_strcat(UChar     *dst,
    const UChar     *src)
{
    UChar *anchor = dst;            /* save a pointer to start of dst */

    while(*dst != 0) {              /* To end of first string          */
        ++dst;
    }
    while((*(dst++) = *(src++)) != 0) {     /* copy string 2 over              */
    }

    return anchor;
}

U_CAPI UChar*  U_EXPORT2
u_strncat(UChar     *dst,
     const UChar     *src,
     int32_t     n )
{
    if(n > 0) {
        UChar *anchor = dst;            /* save a pointer to start of dst */

        while(*dst != 0) {              /* To end of first string          */
            ++dst;
        }
        while((*dst = *src) != 0) {     /* copy string 2 over              */
            ++dst;
            if(--n == 0) {
                *dst = 0;
                break;
            }
            ++src;
        }

        return anchor;
    } else {
        return dst;
    }
}

/* ----- Text property functions --- */

U_CAPI int32_t   U_EXPORT2
u_strcmp(const UChar *s1,
    const UChar *s2)
{
    UChar  c1, c2;

    for(;;) {
        c1=*s1++;
        c2=*s2++;
        if (c1 != c2 || c1 == 0) {
            break;
        }
    }
    return (int32_t)c1 - (int32_t)c2;
}

U_CFUNC int32_t U_EXPORT2
uprv_strCompare(const UChar *s1, int32_t length1,
                const UChar *s2, int32_t length2,
                UBool strncmpStyle, UBool codePointOrder) {
    const UChar *start1, *start2, *limit1, *limit2;
    UChar c1, c2;

    /* setup for fix-up */
    start1=s1;
    start2=s2;

    /* compare identical prefixes - they do not need to be fixed up */
    if(length1<0 && length2<0) {
        /* strcmp style, both NUL-terminated */
        if(s1==s2) {
            return 0;
        }

        for(;;) {
            c1=*s1;
            c2=*s2;
            if(c1!=c2) {
                break;
            }
            if(c1==0) {
                return 0;
            }
            ++s1;
            ++s2;
        }

        /* setup for fix-up */
        limit1=limit2=NULL;
    } else if(strncmpStyle) {
        /* special handling for strncmp, assume length1==length2>=0 but also check for NUL */
        if(s1==s2) {
            return 0;
        }

        limit1=start1+length1;

        for(;;) {
            /* both lengths are same, check only one limit */
            if(s1==limit1) {
                return 0;
            }

            c1=*s1;
            c2=*s2;
            if(c1!=c2) {
                break;
            }
            if(c1==0) {
                return 0;
            }
            ++s1;
            ++s2;
        }

        /* setup for fix-up */
        limit2=start2+length1; /* use length1 here, too, to enforce assumption */
    } else {
        /* memcmp/UnicodeString style, both length-specified */
        int32_t lengthResult;

        if(length1<0) {
            length1=u_strlen(s1);
        }
        if(length2<0) {
            length2=u_strlen(s2);
        }

        /* limit1=start1+min(lenght1, length2) */
        if(length1<length2) {
            lengthResult=-1;
            limit1=start1+length1;
        } else if(length1==length2) {
            lengthResult=0;
            limit1=start1+length1;
        } else /* length1>length2 */ {
            lengthResult=1;
            limit1=start1+length2;
        }

        if(s1==s2) {
            return lengthResult;
        }

        for(;;) {
            /* check pseudo-limit */
            if(s1==limit1) {
                return lengthResult;
            }

            c1=*s1;
            c2=*s2;
            if(c1!=c2) {
                break;
            }
            ++s1;
            ++s2;
        }

        /* setup for fix-up */
        limit1=start1+length1;
        limit2=start2+length2;
    }

    /* if both values are in or above the surrogate range, fix them up */
    if(c1>=0xd800 && c2>=0xd800 && codePointOrder) {
        /* subtract 0x2800 from BMP code points to make them smaller than supplementary ones */
        if(
            (c1<=0xdbff && (s1+1)!=limit1 && U16_IS_TRAIL(*(s1+1))) ||
            (U16_IS_TRAIL(c1) && start1!=s1 && U16_IS_LEAD(*(s1-1)))
        ) {
            /* part of a surrogate pair, leave >=d800 */
        } else {
            /* BMP code point - may be surrogate code point - make <d800 */
            c1-=0x2800;
        }

        if(
            (c2<=0xdbff && (s2+1)!=limit2 && U16_IS_TRAIL(*(s2+1))) ||
            (U16_IS_TRAIL(c2) && start2!=s2 && U16_IS_LEAD(*(s2-1)))
        ) {
            /* part of a surrogate pair, leave >=d800 */
        } else {
            /* BMP code point - may be surrogate code point - make <d800 */
            c2-=0x2800;
        }
    }

    /* now c1 and c2 are in the requested (code unit or code point) order */
    return (int32_t)c1-(int32_t)c2;
}

/*
 * Compare two strings as presented by UCharIterators.
 * Use code unit or code point order.
 * When the function returns, it is undefined where the iterators
 * have stopped.
 */
U_CAPI int32_t U_EXPORT2
u_strCompareIter(UCharIterator *iter1, UCharIterator *iter2, UBool codePointOrder) {
    UChar32 c1, c2;

    /* argument checking */
    if(iter1==NULL || iter2==NULL) {
        return 0; /* bad arguments */
    }
    if(iter1==iter2) {
        return 0; /* identical iterators */
    }

    /* reset iterators to start? */
    iter1->move(iter1, 0, UITER_START);
    iter2->move(iter2, 0, UITER_START);

    /* compare identical prefixes - they do not need to be fixed up */
    for(;;) {
        c1=iter1->next(iter1);
        c2=iter2->next(iter2);
        if(c1!=c2) {
            break;
        }
        if(c1==-1) {
            return 0;
        }
    }

    /* if both values are in or above the surrogate range, fix them up */
    if(c1>=0xd800 && c2>=0xd800 && codePointOrder) {
        /* subtract 0x2800 from BMP code points to make them smaller than supplementary ones */
        if(
            (c1<=0xdbff && U16_IS_TRAIL(iter1->current(iter1))) ||
            (U16_IS_TRAIL(c1) && (iter1->previous(iter1), U16_IS_LEAD(iter1->previous(iter1))))
        ) {
            /* part of a surrogate pair, leave >=d800 */
        } else {
            /* BMP code point - may be surrogate code point - make <d800 */
            c1-=0x2800;
        }

        if(
            (c2<=0xdbff && U16_IS_TRAIL(iter2->current(iter2))) ||
            (U16_IS_TRAIL(c2) && (iter2->previous(iter2), U16_IS_LEAD(iter2->previous(iter2))))
        ) {
            /* part of a surrogate pair, leave >=d800 */
        } else {
            /* BMP code point - may be surrogate code point - make <d800 */
            c2-=0x2800;
        }
    }

    /* now c1 and c2 are in the requested (code unit or code point) order */
    return (int32_t)c1-(int32_t)c2;
}

#if 0
/*
 * u_strCompareIter() does not leave the iterators _on_ the different units.
 * This is possible but would cost a few extra indirect function calls to back
 * up if the last unit (c1 or c2 respectively) was >=0.
 *
 * Consistently leaving them _behind_ the different units is not an option
 * because the current "unit" is the end of the string if that is reached,
 * and in such a case the iterator does not move.
 * For example, when comparing "ab" with "abc", both iterators rest _on_ the end
 * of their strings. Calling previous() on each does not move them to where
 * the comparison fails.
 *
 * So the simplest semantics is to not define where the iterators end up.
 *
 * The following fragment is part of what would need to be done for backing up.
 */
void fragment {
        /* iff a surrogate is part of a surrogate pair, leave >=d800 */
        if(c1<=0xdbff) {
            if(!U16_IS_TRAIL(iter1->current(iter1))) {
                /* lead surrogate code point - make <d800 */
                c1-=0x2800;
            }
        } else if(c1<=0xdfff) {
            int32_t idx=iter1->getIndex(iter1, UITER_CURRENT);
            iter1->previous(iter1); /* ==c1 */
            if(!U16_IS_LEAD(iter1->previous(iter1))) {
                /* trail surrogate code point - make <d800 */
                c1-=0x2800;
            }
            /* go back to behind where the difference is */
            iter1->move(iter1, idx, UITER_ZERO);
        } else /* 0xe000<=c1<=0xffff */ {
            /* BMP code point - make <d800 */
            c1-=0x2800;
        }
}
#endif

U_CAPI int32_t U_EXPORT2
u_strCompare(const UChar *s1, int32_t length1,
             const UChar *s2, int32_t length2,
             UBool codePointOrder) {
    /* argument checking */
    if(s1==NULL || length1<-1 || s2==NULL || length2<-1) {
        return 0;
    }
    return uprv_strCompare(s1, length1, s2, length2, FALSE, codePointOrder);
}

/* String compare in code point order - u_strcmp() compares in code unit order. */
U_CAPI int32_t U_EXPORT2
u_strcmpCodePointOrder(const UChar *s1, const UChar *s2) {
    return uprv_strCompare(s1, -1, s2, -1, FALSE, TRUE);
}

U_CAPI int32_t   U_EXPORT2
u_strncmp(const UChar     *s1,
     const UChar     *s2,
     int32_t     n)
{
    if(n > 0) {
        int32_t rc;
        for(;;) {
            rc = (int32_t)*s1 - (int32_t)*s2;
            if(rc != 0 || *s1 == 0 || --n == 0) {
                return rc;
            }
            ++s1;
            ++s2;
        }
    } else {
        return 0;
    }
}

U_CAPI int32_t U_EXPORT2
u_strncmpCodePointOrder(const UChar *s1, const UChar *s2, int32_t n) {
    return uprv_strCompare(s1, n, s2, n, TRUE, TRUE);
}

U_CAPI UChar* U_EXPORT2
u_strcpy(UChar     *dst,
    const UChar     *src)
{
    UChar *anchor = dst;            /* save a pointer to start of dst */

    while((*(dst++) = *(src++)) != 0) {     /* copy string 2 over              */
    }

    return anchor;
}

U_CAPI UChar*  U_EXPORT2
u_strncpy(UChar     *dst,
     const UChar     *src,
     int32_t     n)
{
    UChar *anchor = dst;            /* save a pointer to start of dst */

    /* copy string 2 over */
    while(n > 0 && (*(dst++) = *(src++)) != 0) {
        --n;
    }

    return anchor;
}

U_CAPI int32_t   U_EXPORT2
u_strlen(const UChar *s)
{
// #if U_SIZEOF_WCHAR_T == U_SIZEOF_UCHAR
//     return (int32_t)uprv_wcslen((const wchar_t *)s);
// #else
    const UChar *t = s;
    while(*t != 0) {
      ++t;
    }
    return t - s;
// #endif
}

U_CAPI int32_t U_EXPORT2
u_countChar32(const UChar *s, int32_t length) {
    int32_t count;

    if(s==NULL || length<-1) {
        return 0;
    }

    count=0;
    if(length>=0) {
        while(length>0) {
            ++count;
            if(U16_IS_LEAD(*s) && length>=2 && U16_IS_TRAIL(*(s+1))) {
                s+=2;
                length-=2;
            } else {
                ++s;
                --length;
            }
        }
    } else /* length==-1 */ {
        UChar c;

        for(;;) {
            if((c=*s++)==0) {
                break;
            }
            ++count;

            /*
             * sufficient to look ahead one because of UTF-16;
             * safe to look ahead one because at worst that would be the terminating NUL
             */
            if(U16_IS_LEAD(c) && U16_IS_TRAIL(*s)) {
                ++s;
            }
        }
    }
    return count;
}

U_CAPI UBool U_EXPORT2
u_strHasMoreChar32Than(const UChar *s, int32_t length, int32_t number) {

    if(number<0) {
        return TRUE;
    }
    if(s==NULL || length<-1) {
        return FALSE;
    }

    if(length==-1) {
        /* s is NUL-terminated */
        UChar c;

        /* count code points until they exceed */
        for(;;) {
            if((c=*s++)==0) {
                return FALSE;
            }
            if(number==0) {
                return TRUE;
            }
            if(U16_IS_LEAD(c) && U16_IS_TRAIL(*s)) {
                ++s;
            }
            --number;
        }
    } else {
        /* length>=0 known */
        const UChar *limit;
        int32_t maxSupplementary;

        /* s contains at least (length+1)/2 code points: <=2 UChars per cp */
        if(((length+1)/2)>number) {
            return TRUE;
        }

        /* check if s does not even contain enough UChars */
        maxSupplementary=length-number;
        if(maxSupplementary<=0) {
            return FALSE;
        }
        /* there are maxSupplementary=length-number more UChars than asked-for code points */

        /*
         * count code points until they exceed and also check that there are
         * no more than maxSupplementary supplementary code points (UChar pairs)
         */
        limit=s+length;
        for(;;) {
            if(s==limit) {
                return FALSE;
            }
            if(number==0) {
                return TRUE;
            }
            if(U16_IS_LEAD(*s++) && s!=limit && U16_IS_TRAIL(*s)) {
                ++s;
                if(--maxSupplementary<=0) {
                    /* too many pairs - too few code points */
                    return FALSE;
                }
            }
            --number;
        }
    }
}

U_CAPI UChar * U_EXPORT2
u_memcpy(UChar *dest, const UChar *src, int32_t count) {
    if(count > 0) {
        uprv_memcpy(dest, src, (size_t)count*U_SIZEOF_UCHAR);
    }
    return dest;
}

U_CAPI UChar * U_EXPORT2
u_memmove(UChar *dest, const UChar *src, int32_t count) {
    if(count > 0) {
        uprv_memmove(dest, src, (size_t)count*U_SIZEOF_UCHAR);
    }
    return dest;
}

U_CAPI UChar * U_EXPORT2
u_memset(UChar *dest, UChar c, int32_t count) {
    if(count > 0) {
        UChar *ptr = dest;
        UChar *limit = dest + count;

        while (ptr < limit) {
            *(ptr++) = c;
        }
    }
    return dest;
}

U_CAPI int32_t U_EXPORT2
u_memcmp(const UChar *buf1, const UChar *buf2, int32_t count) {
    if(count > 0) {
        const UChar *limit = buf1 + count;
        int32_t result;

        while (buf1 < limit) {
            result = (int32_t)(uint16_t)*buf1 - (int32_t)(uint16_t)*buf2;
            if (result != 0) {
                return result;
            }
            buf1++;
            buf2++;
        }
    }
    return 0;
}

U_CAPI int32_t U_EXPORT2
u_memcmpCodePointOrder(const UChar *s1, const UChar *s2, int32_t count) {
    return uprv_strCompare(s1, count, s2, count, FALSE, TRUE);
}

/* u_unescape & support fns ------------------------------------------------- */

/* This map must be in ASCENDING ORDER OF THE ESCAPE CODE */
static const UChar UNESCAPE_MAP[] = {
    /*"   0x22, 0x22 */
    /*'   0x27, 0x27 */
    /*?   0x3F, 0x3F */
    /*\   0x5C, 0x5C */
    /*a*/ 0x61, 0x07,
    /*b*/ 0x62, 0x08,
    /*e*/ 0x65, 0x1b,
    /*f*/ 0x66, 0x0c,
    /*n*/ 0x6E, 0x0a,
    /*r*/ 0x72, 0x0d,
    /*t*/ 0x74, 0x09,
    /*v*/ 0x76, 0x0b
};
enum { UNESCAPE_MAP_LENGTH = UPRV_LENGTHOF(UNESCAPE_MAP) };

/* Convert one octal digit to a numeric value 0..7, or -1 on failure */
static int8_t _digit8(UChar c) {
    if (c >= 0x0030 && c <= 0x0037) {
        return (int8_t)(c - 0x0030);
    }
    return -1;
}

/* Convert one hex digit to a numeric value 0..F, or -1 on failure */
static int8_t _digit16(UChar c) {
    if (c >= 0x0030 && c <= 0x0039) {
        return (int8_t)(c - 0x0030);
    }
    if (c >= 0x0041 && c <= 0x0046) {
        return (int8_t)(c - (0x0041 - 10));
    }
    if (c >= 0x0061 && c <= 0x0066) {
        return (int8_t)(c - (0x0061 - 10));
    }
    return -1;
}

/* Parse a single escape sequence.  Although this method deals in
 * UChars, it does not use C++ or UnicodeString.  This allows it to
 * be used from C contexts. */
U_CAPI UChar32 U_EXPORT2
u_unescapeAt(UNESCAPE_CHAR_AT charAt,
             int32_t *offset,
             int32_t length,
             void *context) {

    int32_t start = *offset;
    UChar c;
    UChar32 result = 0;
    int8_t n = 0;
    int8_t minDig = 0;
    int8_t maxDig = 0;
    int8_t bitsPerDigit = 4;
    int8_t dig;
    int32_t i;
    UBool braces = FALSE;

    /* Check that offset is in range */
    if (*offset < 0 || *offset >= length) {
        goto err;
    }

    /* Fetch first UChar after '\\' */
    c = charAt((*offset)++, context);

    /* Convert hexadecimal and octal escapes */
    switch (c) {
    case 0x0075 /*'u'*/:
        minDig = maxDig = 4;
        break;
    case 0x0055 /*'U'*/:
        minDig = maxDig = 8;
        break;
    case 0x0078 /*'x'*/:
        minDig = 1;
        if (*offset < length && charAt(*offset, context) == 0x7B /*{*/) {
            ++(*offset);
            braces = TRUE;
            maxDig = 8;
        } else {
            maxDig = 2;
        }
        break;
    default:
        dig = _digit8(c);
        if (dig >= 0) {
            minDig = 1;
            maxDig = 3;
            n = 1; /* Already have first octal digit */
            bitsPerDigit = 3;
            result = dig;
        }
        break;
    }
    if (minDig != 0) {
        while (*offset < length && n < maxDig) {
            c = charAt(*offset, context);
            dig = (int8_t)((bitsPerDigit == 3) ? _digit8(c) : _digit16(c));
            if (dig < 0) {
                break;
            }
            result = (result << bitsPerDigit) | dig;
            ++(*offset);
            ++n;
        }
        if (n < minDig) {
            goto err;
        }
        if (braces) {
            if (c != 0x7D /*}*/) {
                goto err;
            }
            ++(*offset);
        }
        if (result < 0 || result >= 0x110000) {
            goto err;
        }
        /* If an escape sequence specifies a lead surrogate, see if
         * there is a trail surrogate after it, either as an escape or
         * as a literal.  If so, join them up into a supplementary.
         */
        if (*offset < length && U16_IS_LEAD(result)) {
            int32_t ahead = *offset + 1;
            c = charAt(*offset, context);
            if (c == 0x5C /*'\\'*/ && ahead < length) {
                c = (UChar) u_unescapeAt(charAt, &ahead, length, context);
            }
            if (U16_IS_TRAIL(c)) {
                *offset = ahead;
                result = U16_GET_SUPPLEMENTARY(result, c);
            }
        }
        return result;
    }

    /* Convert C-style escapes in table */
    for (i=0; i<UNESCAPE_MAP_LENGTH; i+=2) {
        if (c == UNESCAPE_MAP[i]) {
            return UNESCAPE_MAP[i+1];
        } else if (c < UNESCAPE_MAP[i]) {
            break;
        }
    }

    /* Map \cX to control-X: X & 0x1F */
    if (c == 0x0063 /*'c'*/ && *offset < length) {
        c = charAt((*offset)++, context);
        if (U16_IS_LEAD(c) && *offset < length) {
            UChar c2 = charAt(*offset, context);
            if (U16_IS_TRAIL(c2)) {
                ++(*offset);
                c = (UChar) U16_GET_SUPPLEMENTARY(c, c2); /* [sic] */
            }
        }
        return 0x1F & c;
    }

    /* If no special forms are recognized, then consider
     * the backslash to generically escape the next character.
     * Deal with surrogate pairs. */
    if (U16_IS_LEAD(c) && *offset < length) {
        UChar c2 = charAt(*offset, context);
        if (U16_IS_TRAIL(c2)) {
            ++(*offset);
            return U16_GET_SUPPLEMENTARY(c, c2);
        }
    }
    return c;

 err:
    /* Invalid escape sequence */
    *offset = start; /* Reset to initial value */
    return (UChar32)0xFFFFFFFF;
}

/* u_unescapeAt() callback to return a UChar from a char* */
static UChar U_CALLCONV
_charPtr_charAt(int32_t offset, void *context) {
    UChar c16;
    /* It would be more efficient to access the invariant tables
     * directly but there is no API for that. */
    u_charsToUChars(((char*) context) + offset, &c16, 1);
    return c16;
}

/* Append an escape-free segment of the text; used by u_unescape() */
static void _appendUChars(UChar *dest, int32_t destCapacity,
                          const char *src, int32_t srcLen) {
    if (destCapacity < 0) {
        destCapacity = 0;
    }
    if (srcLen > destCapacity) {
        srcLen = destCapacity;
    }
    u_charsToUChars(src, dest, srcLen);
}

/* Do an invariant conversion of char* -> UChar*, with escape parsing */
U_CAPI int32_t U_EXPORT2
u_unescape(const char *src, UChar *dest, int32_t destCapacity) {
    const char *segment = src;
    int32_t i = 0;
    char c;

    while ((c=*src) != 0) {
        /* '\\' intentionally written as compiler-specific
         * character constant to correspond to compiler-specific
         * char* constants. */
        if (c == '\\') {
            int32_t lenParsed = 0;
            UChar32 c32;
            if (src != segment) {
                if (dest != NULL) {
                    _appendUChars(dest + i, destCapacity - i,
                                  segment, (int32_t)(src - segment));
                }
                i += (int32_t)(src - segment);
            }
            ++src; /* advance past '\\' */
            c32 = (UChar32)u_unescapeAt(_charPtr_charAt, &lenParsed, (int32_t)uprv_strlen(src), (void*)src);
            if (lenParsed == 0) {
                goto err;
            }
            src += lenParsed; /* advance past escape seq. */
            if (dest != NULL && U16_LENGTH(c32) <= (destCapacity - i)) {
                U16_APPEND_UNSAFE(dest, i, c32);
            } else {
                i += U16_LENGTH(c32);
            }
            segment = src;
        } else {
            ++src;
        }
    }
    if (src != segment) {
        if (dest != NULL) {
            _appendUChars(dest + i, destCapacity - i,
                          segment, (int32_t)(src - segment));
        }
        i += (int32_t)(src - segment);
    }
    if (dest != NULL && i < destCapacity) {
        dest[i] = 0;
    }
    return i;

 err:
    if (dest != NULL && destCapacity > 0) {
        *dest = 0;
    }
    return 0;
}

/* NUL-termination of strings ----------------------------------------------- */

/**
 * NUL-terminate a string no matter what its type.
 * Set warning and error codes accordingly.
 */
#define __TERMINATE_STRING(dest, destCapacity, length, pErrorCode) UPRV_BLOCK_MACRO_BEGIN { \
    if(pErrorCode!=NULL && U_SUCCESS(*pErrorCode)) {                    \
        /* not a public function, so no complete argument checking */   \
                                                                        \
        if(length<0) {                                                  \
            /* assume that the caller handles this */                   \
        } else if(length<destCapacity) {                                \
            /* NUL-terminate the string, the NUL fits */                \
            dest[length]=0;                                             \
            /* unset the not-terminated warning but leave all others */ \
            if(*pErrorCode==U_STRING_NOT_TERMINATED_WARNING) {          \
                *pErrorCode=U_ZERO_ERROR;                               \
            }                                                           \
        } else if(length==destCapacity) {                               \
            /* unable to NUL-terminate, but the string itself fit - set a warning code */ \
            *pErrorCode=U_STRING_NOT_TERMINATED_WARNING;                \
        } else /* length>destCapacity */ {                              \
            /* even the string itself did not fit - set an error code */ \
            *pErrorCode=U_BUFFER_OVERFLOW_ERROR;                        \
        }                                                               \
    } \
} UPRV_BLOCK_MACRO_END

U_CAPI int32_t U_EXPORT2
u_terminateUChars(UChar *dest, int32_t destCapacity, int32_t length, UErrorCode *pErrorCode) {
    __TERMINATE_STRING(dest, destCapacity, length, pErrorCode);
    return length;
}

U_CAPI int32_t U_EXPORT2
u_terminateChars(char *dest, int32_t destCapacity, int32_t length, UErrorCode *pErrorCode) {
    __TERMINATE_STRING(dest, destCapacity, length, pErrorCode);
    return length;
}

U_CAPI int32_t U_EXPORT2
u_terminateUChar32s(UChar32 *dest, int32_t destCapacity, int32_t length, UErrorCode *pErrorCode) {
    __TERMINATE_STRING(dest, destCapacity, length, pErrorCode);
    return length;
}

U_CAPI int32_t U_EXPORT2
u_terminateWChars(wchar_t *dest, int32_t destCapacity, int32_t length, UErrorCode *pErrorCode) {
    __TERMINATE_STRING(dest, destCapacity, length, pErrorCode);
    return length;
}

// Compute the hash code for a string -------------------------------------- ***

// Moved here from uhash.c so that UnicodeString::hashCode() does not depend
// on UHashtable code.

/*
  Compute the hash by iterating sparsely over about 32 (up to 63)
  characters spaced evenly through the string.  For each character,
  multiply the previous hash value by a prime number and add the new
  character in, like a linear congruential random number generator,
  producing a pseudorandom deterministic value well distributed over
  the output range. [LIU]
*/

#define STRING_HASH(TYPE, STR, STRLEN, DEREF) UPRV_BLOCK_MACRO_BEGIN { \
    uint32_t hash = 0;                        \
    const TYPE *p = (const TYPE*) STR;        \
    if (p != NULL) {                          \
        int32_t len = (int32_t)(STRLEN);      \
        int32_t inc = ((len - 32) / 32) + 1;  \
        const TYPE *limit = p + len;          \
        while (p<limit) {                     \
            hash = (hash * 37) + DEREF;       \
            p += inc;                         \
        }                                     \
    }                                         \
    return static_cast<int32_t>(hash);        \
} UPRV_BLOCK_MACRO_END

/* Used by UnicodeString to compute its hashcode - Not public API. */
U_CAPI int32_t U_EXPORT2
ustr_hashUCharsN(const UChar *str, int32_t length) {
    STRING_HASH(UChar, str, length, *p);
}

U_CAPI int32_t U_EXPORT2
ustr_hashCharsN(const char *str, int32_t length) {
    STRING_HASH(uint8_t, str, length, *p);
}

U_CAPI int32_t U_EXPORT2
ustr_hashICharsN(const char *str, int32_t length) {
    STRING_HASH(char, str, length, (uint8_t)uprv_tolower(*p));
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2001-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
* File ustrtrns.cpp
*
* Modification History:
*
*   Date        Name        Description
*   9/10/2001    Ram    Creation.
******************************************************************************
*/

/*******************************************************************************
 *
 * u_strTo* and u_strFrom* APIs
 * WCS functions moved to ustr_wcs.c for better modularization
 *
 *******************************************************************************
 */












U_CAPI UChar* U_EXPORT2 
u_strFromUTF32WithSub(UChar *dest,
               int32_t destCapacity,
               int32_t *pDestLength,
               const UChar32 *src,
               int32_t srcLength,
               UChar32 subchar, int32_t *pNumSubstitutions,
               UErrorCode *pErrorCode) {
    const UChar32 *srcLimit;
    UChar32 ch;
    UChar *destLimit;
    UChar *pDest;
    int32_t reqLength;
    int32_t numSubstitutions;

    /* args check */
    if(U_FAILURE(*pErrorCode)){
        return NULL;
    }
    if( (src==NULL && srcLength!=0) || srcLength < -1 ||
        (destCapacity<0) || (dest == NULL && destCapacity > 0) ||
        subchar > 0x10ffff || U_IS_SURROGATE(subchar)
    ) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if(pNumSubstitutions != NULL) {
        *pNumSubstitutions = 0;
    }

    pDest = dest;
    destLimit = (dest!=NULL)?(dest + destCapacity):NULL;
    reqLength = 0;
    numSubstitutions = 0;

    if(srcLength < 0) {
        /* simple loop for conversion of a NUL-terminated BMP string */
        while((ch=*src) != 0 &&
              ((uint32_t)ch < 0xd800 || (0xe000 <= ch && ch <= 0xffff))) {
            ++src;
            if(pDest < destLimit) {
                *pDest++ = (UChar)ch;
            } else {
                ++reqLength;
            }
        }
        srcLimit = src;
        if(ch != 0) {
            /* "complicated" case, find the end of the remaining string */
            while(*++srcLimit != 0) {}
        }
    } else {
      srcLimit = (src!=NULL)?(src + srcLength):NULL;
    }

    /* convert with length */
    while(src < srcLimit) {
        ch = *src++;
        do {
            /* usually "loops" once; twice only for writing subchar */
            if((uint32_t)ch < 0xd800 || (0xe000 <= ch && ch <= 0xffff)) {
                if(pDest < destLimit) {
                    *pDest++ = (UChar)ch;
                } else {
                    ++reqLength;
                }
                break;
            } else if(0x10000 <= ch && ch <= 0x10ffff) {
                if(pDest!=NULL && ((pDest + 2) <= destLimit)) {
                    *pDest++ = U16_LEAD(ch);
                    *pDest++ = U16_TRAIL(ch);
                } else {
                    reqLength += 2;
                }
                break;
            } else if((ch = subchar) < 0) {
                /* surrogate code point, or not a Unicode code point at all */
                *pErrorCode = U_INVALID_CHAR_FOUND;
                return NULL;
            } else {
                ++numSubstitutions;
            }
        } while(TRUE);
    }

    reqLength += (int32_t)(pDest - dest);
    if(pDestLength) {
        *pDestLength = reqLength;
    }
    if(pNumSubstitutions != NULL) {
        *pNumSubstitutions = numSubstitutions;
    }

    /* Terminate the buffer */
    u_terminateUChars(dest, destCapacity, reqLength, pErrorCode);
    
    return dest;
}

U_CAPI UChar* U_EXPORT2 
u_strFromUTF32(UChar *dest,
               int32_t destCapacity, 
               int32_t *pDestLength,
               const UChar32 *src,
               int32_t srcLength,
               UErrorCode *pErrorCode) {
    return u_strFromUTF32WithSub(
            dest, destCapacity, pDestLength,
            src, srcLength,
            U_SENTINEL, NULL,
            pErrorCode);
}

U_CAPI UChar32* U_EXPORT2 
u_strToUTF32WithSub(UChar32 *dest,
             int32_t destCapacity,
             int32_t *pDestLength,
             const UChar *src,
             int32_t srcLength,
             UChar32 subchar, int32_t *pNumSubstitutions,
             UErrorCode *pErrorCode) {
    const UChar *srcLimit;
    UChar32 ch;
    UChar ch2;
    UChar32 *destLimit;
    UChar32 *pDest;
    int32_t reqLength;
    int32_t numSubstitutions;

    /* args check */
    if(U_FAILURE(*pErrorCode)){
        return NULL;
    }
    if( (src==NULL && srcLength!=0) || srcLength < -1 ||
        (destCapacity<0) || (dest == NULL && destCapacity > 0) ||
        subchar > 0x10ffff || U_IS_SURROGATE(subchar)
    ) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if(pNumSubstitutions != NULL) {
        *pNumSubstitutions = 0;
    }

    pDest = dest;
    destLimit = (dest!=NULL)?(dest + destCapacity):NULL;
    reqLength = 0;
    numSubstitutions = 0;

    if(srcLength < 0) {
        /* simple loop for conversion of a NUL-terminated BMP string */
        while((ch=*src) != 0 && !U16_IS_SURROGATE(ch)) {
            ++src;
            if(pDest < destLimit) {
                *pDest++ = ch;
            } else {
                ++reqLength;
            }
        }
        srcLimit = src;
        if(ch != 0) {
            /* "complicated" case, find the end of the remaining string */
            while(*++srcLimit != 0) {}
        }
    } else {
        srcLimit = (src!=NULL)?(src + srcLength):NULL;
    }

    /* convert with length */
    while(src < srcLimit) {
        ch = *src++;
        if(!U16_IS_SURROGATE(ch)) {
            /* write or count ch below */
        } else if(U16_IS_SURROGATE_LEAD(ch) && src < srcLimit && U16_IS_TRAIL(ch2 = *src)) {
            ++src;
            ch = U16_GET_SUPPLEMENTARY(ch, ch2);
        } else if((ch = subchar) < 0) {
            /* unpaired surrogate */
            *pErrorCode = U_INVALID_CHAR_FOUND;
            return NULL;
        } else {
            ++numSubstitutions;
        }
        if(pDest < destLimit) {
            *pDest++ = ch;
        } else {
            ++reqLength;
        }
    }

    reqLength += (int32_t)(pDest - dest);
    if(pDestLength) {
        *pDestLength = reqLength;
    }
    if(pNumSubstitutions != NULL) {
        *pNumSubstitutions = numSubstitutions;
    }

    /* Terminate the buffer */
    u_terminateUChar32s(dest, destCapacity, reqLength, pErrorCode);

    return dest;
}

U_CAPI UChar32* U_EXPORT2 
u_strToUTF32(UChar32 *dest, 
             int32_t destCapacity,
             int32_t *pDestLength,
             const UChar *src, 
             int32_t srcLength,
             UErrorCode *pErrorCode) {
    return u_strToUTF32WithSub(
            dest, destCapacity, pDestLength,
            src, srcLength,
            U_SENTINEL, NULL,
            pErrorCode);
}

U_CAPI UChar* U_EXPORT2
u_strFromUTF8WithSub(UChar *dest,
              int32_t destCapacity,
              int32_t *pDestLength,
              const char* src,
              int32_t srcLength,
              UChar32 subchar, int32_t *pNumSubstitutions,
              UErrorCode *pErrorCode){
    /* args check */
    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if( (src==NULL && srcLength!=0) || srcLength < -1 ||
        (destCapacity<0) || (dest == NULL && destCapacity > 0) ||
        subchar > 0x10ffff || U_IS_SURROGATE(subchar)
    ) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if(pNumSubstitutions!=NULL) {
        *pNumSubstitutions=0;
    }
    UChar *pDest = dest;
    UChar *pDestLimit = dest+destCapacity;
    int32_t reqLength = 0;
    int32_t numSubstitutions=0;

    /*
     * Inline processing of UTF-8 byte sequences:
     *
     * Byte sequences for the most common characters are handled inline in
     * the conversion loops. In order to reduce the path lengths for those
     * characters, the tests are arranged in a kind of binary search.
     * ASCII (<=0x7f) is checked first, followed by the dividing point
     * between 2- and 3-byte sequences (0xe0).
     * The 3-byte branch is tested first to speed up CJK text.
     * The compiler should combine the subtractions for the two tests for 0xe0.
     * Each branch then tests for the other end of its range.
     */

    if(srcLength < 0){
        /*
         * Transform a NUL-terminated string.
         * The code explicitly checks for NULs only in the lead byte position.
         * A NUL byte in the trail byte position fails the trail byte range check anyway.
         */
        int32_t i;
        UChar32 c;
        for(i = 0; (c = (uint8_t)src[i]) != 0 && (pDest < pDestLimit);) {
            // modified copy of U8_NEXT()
            ++i;
            if(U8_IS_SINGLE(c)) {
                *pDest++=(UChar)c;
            } else {
                uint8_t __t1, __t2;
                if( /* handle U+0800..U+FFFF inline */
                        (0xe0<=(c) && (c)<0xf0) &&
                        U8_IS_VALID_LEAD3_AND_T1((c), src[i]) &&
                        (__t2=src[(i)+1]-0x80)<=0x3f) {
                    *pDest++ = (((c)&0xf)<<12)|((src[i]&0x3f)<<6)|__t2;
                    i+=2;
                } else if( /* handle U+0080..U+07FF inline */
                        ((c)<0xe0 && (c)>=0xc2) &&
                        (__t1=src[i]-0x80)<=0x3f) {
                    *pDest++ = (((c)&0x1f)<<6)|__t1;
                    ++(i);
                } else {
                    /* function call for "complicated" and error cases */
                    (c)=utf8_nextCharSafeBody((const uint8_t *)src, &(i), -1, c, -1);
                    if(c<0 && (++numSubstitutions, c = subchar) < 0) {
                        *pErrorCode = U_INVALID_CHAR_FOUND;
                        return NULL;
                    } else if(c<=0xFFFF) {
                        *(pDest++)=(UChar)c;
                    } else {
                        *(pDest++)=U16_LEAD(c);
                        if(pDest<pDestLimit) {
                            *(pDest++)=U16_TRAIL(c);
                        } else {
                            reqLength++;
                            break;
                        }
                    }
                }
            }
        }

        /* Pre-flight the rest of the string. */
        while((c = (uint8_t)src[i]) != 0) {
            // modified copy of U8_NEXT()
            ++i;
            if(U8_IS_SINGLE(c)) {
                ++reqLength;
            } else {
                uint8_t __t1, __t2;
                if( /* handle U+0800..U+FFFF inline */
                        (0xe0<=(c) && (c)<0xf0) &&
                        U8_IS_VALID_LEAD3_AND_T1((c), src[i]) &&
                        (__t2=src[(i)+1]-0x80)<=0x3f) {
                    ++reqLength;
                    i+=2;
                } else if( /* handle U+0080..U+07FF inline */
                        ((c)<0xe0 && (c)>=0xc2) &&
                        (__t1=src[i]-0x80)<=0x3f) {
                    ++reqLength;
                    ++(i);
                } else {
                    /* function call for "complicated" and error cases */
                    (c)=utf8_nextCharSafeBody((const uint8_t *)src, &(i), -1, c, -1);
                    if(c<0 && (++numSubstitutions, c = subchar) < 0) {
                        *pErrorCode = U_INVALID_CHAR_FOUND;
                        return NULL;
                    }
                    reqLength += U16_LENGTH(c);
                }
            }
        }
    } else /* srcLength >= 0 */ {
        /* Faster loop without ongoing checking for srcLength and pDestLimit. */
        int32_t i = 0;
        UChar32 c;
        for(;;) {
            /*
             * Each iteration of the inner loop progresses by at most 3 UTF-8
             * bytes and one UChar, for most characters.
             * For supplementary code points (4 & 2), which are rare,
             * there is an additional adjustment.
             */
            int32_t count = (int32_t)(pDestLimit - pDest);
            int32_t count2 = (srcLength - i) / 3;
            if(count > count2) {
                count = count2; /* min(remaining dest, remaining src/3) */
            }
            if(count < 3) {
                /*
                 * Too much overhead if we get near the end of the string,
                 * continue with the next loop.
                 */
                break;
            }

            do {
                // modified copy of U8_NEXT()
                c = (uint8_t)src[i++];
                if(U8_IS_SINGLE(c)) {
                    *pDest++=(UChar)c;
                } else {
                    uint8_t __t1, __t2;
                    if( /* handle U+0800..U+FFFF inline */
                            (0xe0<=(c) && (c)<0xf0) &&
                            ((i)+1)<srcLength &&
                            U8_IS_VALID_LEAD3_AND_T1((c), src[i]) &&
                            (__t2=src[(i)+1]-0x80)<=0x3f) {
                        *pDest++ = (((c)&0xf)<<12)|((src[i]&0x3f)<<6)|__t2;
                        i+=2;
                    } else if( /* handle U+0080..U+07FF inline */
                            ((c)<0xe0 && (c)>=0xc2) &&
                            ((i)!=srcLength) &&
                            (__t1=src[i]-0x80)<=0x3f) {
                        *pDest++ = (((c)&0x1f)<<6)|__t1;
                        ++(i);
                    } else {
                        if(c >= 0xf0 || subchar > 0xffff) {
                            // We may read up to four bytes and write up to two UChars,
                            // which we didn't account for with computing count,
                            // so we adjust it here.
                            if(--count == 0) {
                                --i;  // back out byte c
                                break;
                            }
                        }

                        /* function call for "complicated" and error cases */
                        (c)=utf8_nextCharSafeBody((const uint8_t *)src, &(i), srcLength, c, -1);
                        if(c<0 && (++numSubstitutions, c = subchar) < 0) {
                            *pErrorCode = U_INVALID_CHAR_FOUND;
                            return NULL;
                        } else if(c<=0xFFFF) {
                            *(pDest++)=(UChar)c;
                        } else {
                            *(pDest++)=U16_LEAD(c);
                            *(pDest++)=U16_TRAIL(c);
                        }
                    }
                }
            } while(--count > 0);
        }

        while(i < srcLength && (pDest < pDestLimit)) {
            // modified copy of U8_NEXT()
            c = (uint8_t)src[i++];
            if(U8_IS_SINGLE(c)) {
                *pDest++=(UChar)c;
            } else {
                uint8_t __t1, __t2;
                if( /* handle U+0800..U+FFFF inline */
                        (0xe0<=(c) && (c)<0xf0) &&
                        ((i)+1)<srcLength &&
                        U8_IS_VALID_LEAD3_AND_T1((c), src[i]) &&
                        (__t2=src[(i)+1]-0x80)<=0x3f) {
                    *pDest++ = (((c)&0xf)<<12)|((src[i]&0x3f)<<6)|__t2;
                    i+=2;
                } else if( /* handle U+0080..U+07FF inline */
                        ((c)<0xe0 && (c)>=0xc2) &&
                        ((i)!=srcLength) &&
                        (__t1=src[i]-0x80)<=0x3f) {
                    *pDest++ = (((c)&0x1f)<<6)|__t1;
                    ++(i);
                } else {
                    /* function call for "complicated" and error cases */
                    (c)=utf8_nextCharSafeBody((const uint8_t *)src, &(i), srcLength, c, -1);
                    if(c<0 && (++numSubstitutions, c = subchar) < 0) {
                        *pErrorCode = U_INVALID_CHAR_FOUND;
                        return NULL;
                    } else if(c<=0xFFFF) {
                        *(pDest++)=(UChar)c;
                    } else {
                        *(pDest++)=U16_LEAD(c);
                        if(pDest<pDestLimit) {
                            *(pDest++)=U16_TRAIL(c);
                        } else {
                            reqLength++;
                            break;
                        }
                    }
                }
            }
        }

        /* Pre-flight the rest of the string. */
        while(i < srcLength) {
            // modified copy of U8_NEXT()
            c = (uint8_t)src[i++];
            if(U8_IS_SINGLE(c)) {
                ++reqLength;
            } else {
                uint8_t __t1, __t2;
                if( /* handle U+0800..U+FFFF inline */
                        (0xe0<=(c) && (c)<0xf0) &&
                        ((i)+1)<srcLength &&
                        U8_IS_VALID_LEAD3_AND_T1((c), src[i]) &&
                        (__t2=src[(i)+1]-0x80)<=0x3f) {
                    ++reqLength;
                    i+=2;
                } else if( /* handle U+0080..U+07FF inline */
                        ((c)<0xe0 && (c)>=0xc2) &&
                        ((i)!=srcLength) &&
                        (__t1=src[i]-0x80)<=0x3f) {
                    ++reqLength;
                    ++(i);
                } else {
                    /* function call for "complicated" and error cases */
                    (c)=utf8_nextCharSafeBody((const uint8_t *)src, &(i), srcLength, c, -1);
                    if(c<0 && (++numSubstitutions, c = subchar) < 0) {
                        *pErrorCode = U_INVALID_CHAR_FOUND;
                        return NULL;
                    }
                    reqLength += U16_LENGTH(c);
                }
            }
        }
    }

    reqLength+=(int32_t)(pDest - dest);

    if(pNumSubstitutions!=NULL) {
        *pNumSubstitutions=numSubstitutions;
    }

    if(pDestLength){
        *pDestLength = reqLength;
    }

    /* Terminate the buffer */
    u_terminateUChars(dest,destCapacity,reqLength,pErrorCode);

    return dest;
}

U_CAPI UChar* U_EXPORT2
u_strFromUTF8(UChar *dest,
              int32_t destCapacity,
              int32_t *pDestLength,
              const char* src,
              int32_t srcLength,
              UErrorCode *pErrorCode){
    return u_strFromUTF8WithSub(
            dest, destCapacity, pDestLength,
            src, srcLength,
            U_SENTINEL, NULL,
            pErrorCode);
}

U_CAPI UChar * U_EXPORT2
u_strFromUTF8Lenient(UChar *dest,
                     int32_t destCapacity,
                     int32_t *pDestLength,
                     const char *src,
                     int32_t srcLength,
                     UErrorCode *pErrorCode) {
    UChar *pDest = dest;
    UChar32 ch;
    int32_t reqLength = 0;
    uint8_t* pSrc = (uint8_t*) src;

    /* args check */
    if(U_FAILURE(*pErrorCode)){
        return NULL;
    }
        
    if( (src==NULL && srcLength!=0) || srcLength < -1 ||
        (destCapacity<0) || (dest == NULL && destCapacity > 0)
    ) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if(srcLength < 0) {
        /* Transform a NUL-terminated string. */
        UChar *pDestLimit = (dest!=NULL)?(dest+destCapacity):NULL;
        uint8_t t1, t2, t3; /* trail bytes */

        while(((ch = *pSrc) != 0) && (pDest < pDestLimit)) {
            if(ch < 0xc0) {
                /*
                 * ASCII, or a trail byte in lead position which is treated like
                 * a single-byte sequence for better character boundary
                 * resynchronization after illegal sequences.
                 */
                *pDest++=(UChar)ch;
                ++pSrc;
                continue;
            } else if(ch < 0xe0) { /* U+0080..U+07FF */
                if((t1 = pSrc[1]) != 0) {
                    /* 0x3080 = (0xc0 << 6) + 0x80 */
                    *pDest++ = (UChar)((ch << 6) + t1 - 0x3080);
                    pSrc += 2;
                    continue;
                }
            } else if(ch < 0xf0) { /* U+0800..U+FFFF */
                if((t1 = pSrc[1]) != 0 && (t2 = pSrc[2]) != 0) {
                    /* no need for (ch & 0xf) because the upper bits are truncated after <<12 in the cast to (UChar) */
                    /* 0x2080 = (0x80 << 6) + 0x80 */
                    *pDest++ = (UChar)((ch << 12) + (t1 << 6) + t2 - 0x2080);
                    pSrc += 3;
                    continue;
                }
            } else /* f0..f4 */ { /* U+10000..U+10FFFF */
                if((t1 = pSrc[1]) != 0 && (t2 = pSrc[2]) != 0 && (t3 = pSrc[3]) != 0) {
                    pSrc += 4;
                    /* 0x3c82080 = (0xf0 << 18) + (0x80 << 12) + (0x80 << 6) + 0x80 */
                    ch = (ch << 18) + (t1 << 12) + (t2 << 6) + t3 - 0x3c82080;
                    *(pDest++) = U16_LEAD(ch);
                    if(pDest < pDestLimit) {
                        *(pDest++) = U16_TRAIL(ch);
                    } else {
                        reqLength = 1;
                        break;
                    }
                    continue;
                }
            }

            /* truncated character at the end */
            *pDest++ = 0xfffd;
            while(*++pSrc != 0) {}
            break;
        }

        /* Pre-flight the rest of the string. */
        while((ch = *pSrc) != 0) {
            if(ch < 0xc0) {
                /*
                 * ASCII, or a trail byte in lead position which is treated like
                 * a single-byte sequence for better character boundary
                 * resynchronization after illegal sequences.
                 */
                ++reqLength;
                ++pSrc;
                continue;
            } else if(ch < 0xe0) { /* U+0080..U+07FF */
                if(pSrc[1] != 0) {
                    ++reqLength;
                    pSrc += 2;
                    continue;
                }
            } else if(ch < 0xf0) { /* U+0800..U+FFFF */
                if(pSrc[1] != 0 && pSrc[2] != 0) {
                    ++reqLength;
                    pSrc += 3;
                    continue;
                }
            } else /* f0..f4 */ { /* U+10000..U+10FFFF */
                if(pSrc[1] != 0 && pSrc[2] != 0 && pSrc[3] != 0) {
                    reqLength += 2;
                    pSrc += 4;
                    continue;
                }
            }

            /* truncated character at the end */
            ++reqLength;
            break;
        }
    } else /* srcLength >= 0 */ {
      const uint8_t *pSrcLimit = (pSrc!=NULL)?(pSrc + srcLength):NULL;

        /*
         * This function requires that if srcLength is given, then it must be
         * destCapatity >= srcLength so that we need not check for
         * destination buffer overflow in the loop.
         */
        if(destCapacity < srcLength) {
            if(pDestLength != NULL) {
                *pDestLength = srcLength; /* this likely overestimates the true destLength! */
            }
            *pErrorCode = U_BUFFER_OVERFLOW_ERROR;
            return NULL;
        }

        if((pSrcLimit - pSrc) >= 4) {
            pSrcLimit -= 3; /* temporarily reduce pSrcLimit */

            /* in this loop, we can always access at least 4 bytes, up to pSrc+3 */
            do {
                ch = *pSrc++;
                if(ch < 0xc0) {
                    /*
                     * ASCII, or a trail byte in lead position which is treated like
                     * a single-byte sequence for better character boundary
                     * resynchronization after illegal sequences.
                     */
                    *pDest++=(UChar)ch;
                } else if(ch < 0xe0) { /* U+0080..U+07FF */
                    /* 0x3080 = (0xc0 << 6) + 0x80 */
                    *pDest++ = (UChar)((ch << 6) + *pSrc++ - 0x3080);
                } else if(ch < 0xf0) { /* U+0800..U+FFFF */
                    /* no need for (ch & 0xf) because the upper bits are truncated after <<12 in the cast to (UChar) */
                    /* 0x2080 = (0x80 << 6) + 0x80 */
                    ch = (ch << 12) + (*pSrc++ << 6);
                    *pDest++ = (UChar)(ch + *pSrc++ - 0x2080);
                } else /* f0..f4 */ { /* U+10000..U+10FFFF */
                    /* 0x3c82080 = (0xf0 << 18) + (0x80 << 12) + (0x80 << 6) + 0x80 */
                    ch = (ch << 18) + (*pSrc++ << 12);
                    ch += *pSrc++ << 6;
                    ch += *pSrc++ - 0x3c82080;
                    *(pDest++) = U16_LEAD(ch);
                    *(pDest++) = U16_TRAIL(ch);
                }
            } while(pSrc < pSrcLimit);

            pSrcLimit += 3; /* restore original pSrcLimit */
        }

        while(pSrc < pSrcLimit) {
            ch = *pSrc++;
            if(ch < 0xc0) {
                /*
                 * ASCII, or a trail byte in lead position which is treated like
                 * a single-byte sequence for better character boundary
                 * resynchronization after illegal sequences.
                 */
                *pDest++=(UChar)ch;
                continue;
            } else if(ch < 0xe0) { /* U+0080..U+07FF */
                if(pSrc < pSrcLimit) {
                    /* 0x3080 = (0xc0 << 6) + 0x80 */
                    *pDest++ = (UChar)((ch << 6) + *pSrc++ - 0x3080);
                    continue;
                }
            } else if(ch < 0xf0) { /* U+0800..U+FFFF */
                if((pSrcLimit - pSrc) >= 2) {
                    /* no need for (ch & 0xf) because the upper bits are truncated after <<12 in the cast to (UChar) */
                    /* 0x2080 = (0x80 << 6) + 0x80 */
                    ch = (ch << 12) + (*pSrc++ << 6);
                    *pDest++ = (UChar)(ch + *pSrc++ - 0x2080);
                    pSrc += 3;
                    continue;
                }
            } else /* f0..f4 */ { /* U+10000..U+10FFFF */
                if((pSrcLimit - pSrc) >= 3) {
                    /* 0x3c82080 = (0xf0 << 18) + (0x80 << 12) + (0x80 << 6) + 0x80 */
                    ch = (ch << 18) + (*pSrc++ << 12);
                    ch += *pSrc++ << 6;
                    ch += *pSrc++ - 0x3c82080;
                    *(pDest++) = U16_LEAD(ch);
                    *(pDest++) = U16_TRAIL(ch);
                    pSrc += 4;
                    continue;
                }
            }

            /* truncated character at the end */
            *pDest++ = 0xfffd;
            break;
        }
    }

    reqLength+=(int32_t)(pDest - dest);

    if(pDestLength){
        *pDestLength = reqLength;
    }

    /* Terminate the buffer */
    u_terminateUChars(dest,destCapacity,reqLength,pErrorCode);

    return dest;
}

static inline uint8_t *
_appendUTF8(uint8_t *pDest, UChar32 c) {
    /* it is 0<=c<=0x10ffff and not a surrogate if called by a validating function */
    if((c)<=0x7f) {
        *pDest++=(uint8_t)c;
    } else if(c<=0x7ff) {
        *pDest++=(uint8_t)((c>>6)|0xc0);
        *pDest++=(uint8_t)((c&0x3f)|0x80);
    } else if(c<=0xffff) {
        *pDest++=(uint8_t)((c>>12)|0xe0);
        *pDest++=(uint8_t)(((c>>6)&0x3f)|0x80);
        *pDest++=(uint8_t)(((c)&0x3f)|0x80);
    } else /* if((uint32_t)(c)<=0x10ffff) */ {
        *pDest++=(uint8_t)(((c)>>18)|0xf0);
        *pDest++=(uint8_t)((((c)>>12)&0x3f)|0x80);
        *pDest++=(uint8_t)((((c)>>6)&0x3f)|0x80);
        *pDest++=(uint8_t)(((c)&0x3f)|0x80);
    }
    return pDest;
}

   
U_CAPI char* U_EXPORT2 
u_strToUTF8WithSub(char *dest,
            int32_t destCapacity,
            int32_t *pDestLength,
            const UChar *pSrc,
            int32_t srcLength,
            UChar32 subchar, int32_t *pNumSubstitutions,
            UErrorCode *pErrorCode){
    int32_t reqLength=0;
    uint32_t ch=0,ch2=0;
    uint8_t *pDest = (uint8_t *)dest;
    uint8_t *pDestLimit = (pDest!=NULL)?(pDest + destCapacity):NULL;
    int32_t numSubstitutions;

    /* args check */
    if(U_FAILURE(*pErrorCode)){
        return NULL;
    }
        
    if( (pSrc==NULL && srcLength!=0) || srcLength < -1 ||
        (destCapacity<0) || (dest == NULL && destCapacity > 0) ||
        subchar > 0x10ffff || U_IS_SURROGATE(subchar)
    ) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if(pNumSubstitutions!=NULL) {
        *pNumSubstitutions=0;
    }
    numSubstitutions=0;

    if(srcLength==-1) {
        while((ch=*pSrc)!=0) {
            ++pSrc;
            if(ch <= 0x7f) {
                if(pDest<pDestLimit) {
                    *pDest++ = (uint8_t)ch;
                } else {
                    reqLength = 1;
                    break;
                }
            } else if(ch <= 0x7ff) {
                if((pDestLimit - pDest) >= 2) {
                    *pDest++=(uint8_t)((ch>>6)|0xc0);
                    *pDest++=(uint8_t)((ch&0x3f)|0x80);
                } else {
                    reqLength = 2;
                    break;
                }
            } else if(ch <= 0xd7ff || ch >= 0xe000) {
                if((pDestLimit - pDest) >= 3) {
                    *pDest++=(uint8_t)((ch>>12)|0xe0);
                    *pDest++=(uint8_t)(((ch>>6)&0x3f)|0x80);
                    *pDest++=(uint8_t)((ch&0x3f)|0x80);
                } else {
                    reqLength = 3;
                    break;
                }
            } else /* ch is a surrogate */ {
                int32_t length;

                /*need not check for NUL because NUL fails U16_IS_TRAIL() anyway*/
                if(U16_IS_SURROGATE_LEAD(ch) && U16_IS_TRAIL(ch2=*pSrc)) { 
                    ++pSrc;
                    ch=U16_GET_SUPPLEMENTARY(ch, ch2);
                } else if(subchar>=0) {
                    ch=subchar;
                    ++numSubstitutions;
                } else {
                    /* Unicode 3.2 forbids surrogate code points in UTF-8 */
                    *pErrorCode = U_INVALID_CHAR_FOUND;
                    return NULL;
                }

                length = U8_LENGTH(ch);
                if((pDestLimit - pDest) >= length) {
                    /* convert and append*/
                    pDest=_appendUTF8(pDest, ch);
                } else {
                    reqLength = length;
                    break;
                }
            }
        }
        while((ch=*pSrc++)!=0) {
            if(ch<=0x7f) {
                ++reqLength;
            } else if(ch<=0x7ff) {
                reqLength+=2;
            } else if(!U16_IS_SURROGATE(ch)) {
                reqLength+=3;
            } else if(U16_IS_SURROGATE_LEAD(ch) && U16_IS_TRAIL(ch2=*pSrc)) {
                ++pSrc;
                reqLength+=4;
            } else if(subchar>=0) {
                reqLength+=U8_LENGTH(subchar);
                ++numSubstitutions;
            } else {
                /* Unicode 3.2 forbids surrogate code points in UTF-8 */
                *pErrorCode = U_INVALID_CHAR_FOUND;
                return NULL;
            }
        }
    } else {
        const UChar *pSrcLimit = (pSrc!=NULL)?(pSrc+srcLength):NULL;
        int32_t count;

        /* Faster loop without ongoing checking for pSrcLimit and pDestLimit. */
        for(;;) {
            /*
             * Each iteration of the inner loop progresses by at most 3 UTF-8
             * bytes and one UChar, for most characters.
             * For supplementary code points (4 & 2), which are rare,
             * there is an additional adjustment.
             */
            count = (int32_t)((pDestLimit - pDest) / 3);
            srcLength = (int32_t)(pSrcLimit - pSrc);
            if(count > srcLength) {
                count = srcLength; /* min(remaining dest/3, remaining src) */
            }
            if(count < 3) {
                /*
                 * Too much overhead if we get near the end of the string,
                 * continue with the next loop.
                 */
                break;
            }
            do {
                ch=*pSrc++;
                if(ch <= 0x7f) {
                    *pDest++ = (uint8_t)ch;
                } else if(ch <= 0x7ff) {
                    *pDest++=(uint8_t)((ch>>6)|0xc0);
                    *pDest++=(uint8_t)((ch&0x3f)|0x80);
                } else if(ch <= 0xd7ff || ch >= 0xe000) {
                    *pDest++=(uint8_t)((ch>>12)|0xe0);
                    *pDest++=(uint8_t)(((ch>>6)&0x3f)|0x80);
                    *pDest++=(uint8_t)((ch&0x3f)|0x80);
                } else /* ch is a surrogate */ {
                    /*
                     * We will read two UChars and probably output four bytes,
                     * which we didn't account for with computing count,
                     * so we adjust it here.
                     */
                    if(--count == 0) {
                        --pSrc; /* undo ch=*pSrc++ for the lead surrogate */
                        break;  /* recompute count */
                    }

                    if(U16_IS_SURROGATE_LEAD(ch) && U16_IS_TRAIL(ch2=*pSrc)) { 
                        ++pSrc;
                        ch=U16_GET_SUPPLEMENTARY(ch, ch2);

                        /* writing 4 bytes per 2 UChars is ok */
                        *pDest++=(uint8_t)((ch>>18)|0xf0);
                        *pDest++=(uint8_t)(((ch>>12)&0x3f)|0x80);
                        *pDest++=(uint8_t)(((ch>>6)&0x3f)|0x80);
                        *pDest++=(uint8_t)((ch&0x3f)|0x80);
                    } else  {
                        /* Unicode 3.2 forbids surrogate code points in UTF-8 */
                        if(subchar>=0) {
                            ch=subchar;
                            ++numSubstitutions;
                        } else {
                            *pErrorCode = U_INVALID_CHAR_FOUND;
                            return NULL;
                        }

                        /* convert and append*/
                        pDest=_appendUTF8(pDest, ch);
                    }
                }
            } while(--count > 0);
        }

        while(pSrc<pSrcLimit) {
            ch=*pSrc++;
            if(ch <= 0x7f) {
                if(pDest<pDestLimit) {
                    *pDest++ = (uint8_t)ch;
                } else {
                    reqLength = 1;
                    break;
                }
            } else if(ch <= 0x7ff) {
                if((pDestLimit - pDest) >= 2) {
                    *pDest++=(uint8_t)((ch>>6)|0xc0);
                    *pDest++=(uint8_t)((ch&0x3f)|0x80);
                } else {
                    reqLength = 2;
                    break;
                }
            } else if(ch <= 0xd7ff || ch >= 0xe000) {
                if((pDestLimit - pDest) >= 3) {
                    *pDest++=(uint8_t)((ch>>12)|0xe0);
                    *pDest++=(uint8_t)(((ch>>6)&0x3f)|0x80);
                    *pDest++=(uint8_t)((ch&0x3f)|0x80);
                } else {
                    reqLength = 3;
                    break;
                }
            } else /* ch is a surrogate */ {
                int32_t length;

                if(U16_IS_SURROGATE_LEAD(ch) && pSrc<pSrcLimit && U16_IS_TRAIL(ch2=*pSrc)) { 
                    ++pSrc;
                    ch=U16_GET_SUPPLEMENTARY(ch, ch2);
                } else if(subchar>=0) {
                    ch=subchar;
                    ++numSubstitutions;
                } else {
                    /* Unicode 3.2 forbids surrogate code points in UTF-8 */
                    *pErrorCode = U_INVALID_CHAR_FOUND;
                    return NULL;
                }

                length = U8_LENGTH(ch);
                if((pDestLimit - pDest) >= length) {
                    /* convert and append*/
                    pDest=_appendUTF8(pDest, ch);
                } else {
                    reqLength = length;
                    break;
                }
            }
        }
        while(pSrc<pSrcLimit) {
            ch=*pSrc++;
            if(ch<=0x7f) {
                ++reqLength;
            } else if(ch<=0x7ff) {
                reqLength+=2;
            } else if(!U16_IS_SURROGATE(ch)) {
                reqLength+=3;
            } else if(U16_IS_SURROGATE_LEAD(ch) && pSrc<pSrcLimit && U16_IS_TRAIL(ch2=*pSrc)) {
                ++pSrc;
                reqLength+=4;
            } else if(subchar>=0) {
                reqLength+=U8_LENGTH(subchar);
                ++numSubstitutions;
            } else {
                /* Unicode 3.2 forbids surrogate code points in UTF-8 */
                *pErrorCode = U_INVALID_CHAR_FOUND;
                return NULL;
            }
        }
    }

    reqLength+=(int32_t)(pDest - (uint8_t *)dest);

    if(pNumSubstitutions!=NULL) {
        *pNumSubstitutions=numSubstitutions;
    }

    if(pDestLength){
        *pDestLength = reqLength;
    }

    /* Terminate the buffer */
    u_terminateChars(dest, destCapacity, reqLength, pErrorCode);
    return dest;
}

U_CAPI char* U_EXPORT2 
u_strToUTF8(char *dest,
            int32_t destCapacity,
            int32_t *pDestLength,
            const UChar *pSrc,
            int32_t srcLength,
            UErrorCode *pErrorCode){
    return u_strToUTF8WithSub(
            dest, destCapacity, pDestLength,
            pSrc, srcLength,
            U_SENTINEL, NULL,
            pErrorCode);
}

U_CAPI UChar* U_EXPORT2
u_strFromJavaModifiedUTF8WithSub(
        UChar *dest,
        int32_t destCapacity,
        int32_t *pDestLength,
        const char *src,
        int32_t srcLength,
        UChar32 subchar, int32_t *pNumSubstitutions,
        UErrorCode *pErrorCode) {
    /* args check */
    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if( (src==NULL && srcLength!=0) || srcLength < -1 ||
        (dest==NULL && destCapacity!=0) || destCapacity<0 ||
        subchar > 0x10ffff || U_IS_SURROGATE(subchar)
    ) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if(pNumSubstitutions!=NULL) {
        *pNumSubstitutions=0;
    }
    UChar *pDest = dest;
    UChar *pDestLimit = dest+destCapacity;
    int32_t reqLength = 0;
    int32_t numSubstitutions=0;

    if(srcLength < 0) {
        /*
         * Transform a NUL-terminated ASCII string.
         * Handle non-ASCII strings with slower code.
         */
        UChar32 c;
        while(((c = (uint8_t)*src) != 0) && c <= 0x7f && (pDest < pDestLimit)) {
            *pDest++=(UChar)c;
            ++src;
        }
        if(c == 0) {
            reqLength=(int32_t)(pDest - dest);
            if(pDestLength) {
                *pDestLength = reqLength;
            }

            /* Terminate the buffer */
            u_terminateUChars(dest, destCapacity, reqLength, pErrorCode);
            return dest;
        }
        srcLength = static_cast<int32_t>(uprv_strlen(src));
    }

    /* Faster loop without ongoing checking for srcLength and pDestLimit. */
    UChar32 ch;
    uint8_t t1, t2;
    int32_t i = 0;
    for(;;) {
        int32_t count = (int32_t)(pDestLimit - pDest);
        int32_t count2 = srcLength - i;
        if(count >= count2 && srcLength > 0 && U8_IS_SINGLE(*src)) {
            /* fast ASCII loop */
            int32_t start = i;
            uint8_t b;
            while(i < srcLength && U8_IS_SINGLE(b = src[i])) {
                *pDest++=b;
                ++i;
            }
            int32_t delta = i - start;
            count -= delta;
            count2 -= delta;
        }
        /*
         * Each iteration of the inner loop progresses by at most 3 UTF-8
         * bytes and one UChar.
         */
        if(subchar > 0xFFFF) {
            break;
        }
        count2 /= 3;
        if(count > count2) {
            count = count2; /* min(remaining dest, remaining src/3) */
        }
        if(count < 3) {
            /*
             * Too much overhead if we get near the end of the string,
             * continue with the next loop.
             */
            break;
        }
        do {
            ch = (uint8_t)src[i++];
            if(U8_IS_SINGLE(ch)) {
                *pDest++=(UChar)ch;
            } else {
                if(ch >= 0xe0) {
                    if( /* handle U+0000..U+FFFF inline */
                        ch <= 0xef &&
                        (t1 = (uint8_t)(src[i] - 0x80)) <= 0x3f &&
                        (t2 = (uint8_t)(src[i+1] - 0x80)) <= 0x3f
                    ) {
                        /* no need for (ch & 0xf) because the upper bits are truncated after <<12 in the cast to (UChar) */
                        *pDest++ = (UChar)((ch << 12) | (t1 << 6) | t2);
                        i += 2;
                        continue;
                    }
                } else {
                    if( /* handle U+0000..U+07FF inline */
                        ch >= 0xc0 &&
                        (t1 = (uint8_t)(src[i] - 0x80)) <= 0x3f
                    ) {
                        *pDest++ = (UChar)(((ch & 0x1f) << 6) | t1);
                        ++i;
                        continue;
                    }
                }

                if(subchar < 0) {
                    *pErrorCode = U_INVALID_CHAR_FOUND;
                    return NULL;
                } else if(subchar > 0xffff && --count == 0) {
                    /*
                     * We need to write two UChars, adjusted count for that,
                     * and ran out of space.
                     */
                    --i;  // back out byte ch
                    break;
                } else {
                    /* function call for error cases */
                    utf8_nextCharSafeBody((const uint8_t *)src, &(i), srcLength, ch, -1);
                    ++numSubstitutions;
                    *(pDest++)=(UChar)subchar;
                }
            }
        } while(--count > 0);
    }

    while(i < srcLength && (pDest < pDestLimit)) {
        ch = (uint8_t)src[i++];
        if(U8_IS_SINGLE(ch)){
            *pDest++=(UChar)ch;
        } else {
            if(ch >= 0xe0) {
                if( /* handle U+0000..U+FFFF inline */
                    ch <= 0xef &&
                    (i+1) < srcLength &&
                    (t1 = (uint8_t)(src[i] - 0x80)) <= 0x3f &&
                    (t2 = (uint8_t)(src[i+1] - 0x80)) <= 0x3f
                ) {
                    /* no need for (ch & 0xf) because the upper bits are truncated after <<12 in the cast to (UChar) */
                    *pDest++ = (UChar)((ch << 12) | (t1 << 6) | t2);
                    i += 2;
                    continue;
                }
            } else {
                if( /* handle U+0000..U+07FF inline */
                    ch >= 0xc0 &&
                    i < srcLength &&
                    (t1 = (uint8_t)(src[i] - 0x80)) <= 0x3f
                ) {
                    *pDest++ = (UChar)(((ch & 0x1f) << 6) | t1);
                    ++i;
                    continue;
                }
            }

            if(subchar < 0) {
                *pErrorCode = U_INVALID_CHAR_FOUND;
                return NULL;
            } else {
                /* function call for error cases */
                utf8_nextCharSafeBody((const uint8_t *)src, &(i), srcLength, ch, -1);
                ++numSubstitutions;
                if(subchar<=0xFFFF) {
                    *(pDest++)=(UChar)subchar;
                } else {
                    *(pDest++)=U16_LEAD(subchar);
                    if(pDest<pDestLimit) {
                        *(pDest++)=U16_TRAIL(subchar);
                    } else {
                        reqLength++;
                        break;
                    }
                }
            }
        }
    }

    /* Pre-flight the rest of the string. */
    while(i < srcLength) {
        ch = (uint8_t)src[i++];
        if(U8_IS_SINGLE(ch)) {
            reqLength++;
        } else {
            if(ch >= 0xe0) {
                if( /* handle U+0000..U+FFFF inline */
                    ch <= 0xef &&
                    (i+1) < srcLength &&
                    (uint8_t)(src[i] - 0x80) <= 0x3f &&
                    (uint8_t)(src[i+1] - 0x80) <= 0x3f
                ) {
                    reqLength++;
                    i += 2;
                    continue;
                }
            } else {
                if( /* handle U+0000..U+07FF inline */
                    ch >= 0xc0 &&
                    i < srcLength &&
                    (uint8_t)(src[i] - 0x80) <= 0x3f
                ) {
                    reqLength++;
                    ++i;
                    continue;
                }
            }

            if(subchar < 0) {
                *pErrorCode = U_INVALID_CHAR_FOUND;
                return NULL;
            } else {
                /* function call for error cases */
                utf8_nextCharSafeBody((const uint8_t *)src, &(i), srcLength, ch, -1);
                ++numSubstitutions;
                reqLength+=U16_LENGTH(ch);
            }
        }
    }

    if(pNumSubstitutions!=NULL) {
        *pNumSubstitutions=numSubstitutions;
    }

    reqLength+=(int32_t)(pDest - dest);
    if(pDestLength) {
        *pDestLength = reqLength;
    }

    /* Terminate the buffer */
    u_terminateUChars(dest, destCapacity, reqLength, pErrorCode);
    return dest;
}

U_CAPI char* U_EXPORT2 
u_strToJavaModifiedUTF8(
        char *dest,
        int32_t destCapacity,
        int32_t *pDestLength,
        const UChar *src, 
        int32_t srcLength,
        UErrorCode *pErrorCode) {
    int32_t reqLength=0;
    uint32_t ch=0;
    uint8_t *pDest = (uint8_t *)dest;
    uint8_t *pDestLimit = pDest + destCapacity;
    const UChar *pSrcLimit;
    int32_t count;

    /* args check */
    if(U_FAILURE(*pErrorCode)){
        return NULL;
    }
    if( (src==NULL && srcLength!=0) || srcLength < -1 ||
        (dest==NULL && destCapacity!=0) || destCapacity<0
    ) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if(srcLength==-1) {
        /* Convert NUL-terminated ASCII, then find the string length. */
        while((ch=*src)<=0x7f && ch != 0 && pDest<pDestLimit) {
            *pDest++ = (uint8_t)ch;
            ++src;
        }
        if(ch == 0) {
            reqLength=(int32_t)(pDest - (uint8_t *)dest);
            if(pDestLength) {
                *pDestLength = reqLength;
            }

            /* Terminate the buffer */
            u_terminateChars(dest, destCapacity, reqLength, pErrorCode);
            return dest;
        }
        srcLength = u_strlen(src);
    }

    /* Faster loop without ongoing checking for pSrcLimit and pDestLimit. */
    pSrcLimit = (src!=NULL)?(src+srcLength):NULL;
    for(;;) {
        count = (int32_t)(pDestLimit - pDest);
        srcLength = (int32_t)(pSrcLimit - src);
        if(count >= srcLength && srcLength > 0 && *src <= 0x7f) {
            /* fast ASCII loop */
            const UChar *prevSrc = src;
            int32_t delta;
            while(src < pSrcLimit && (ch = *src) <= 0x7f && ch != 0) {
                *pDest++=(uint8_t)ch;
                ++src;
            }
            delta = (int32_t)(src - prevSrc);
            count -= delta;
            srcLength -= delta;
        }
        /*
         * Each iteration of the inner loop progresses by at most 3 UTF-8
         * bytes and one UChar.
         */
        count /= 3;
        if(count > srcLength) {
            count = srcLength; /* min(remaining dest/3, remaining src) */
        }
        if(count < 3) {
            /*
             * Too much overhead if we get near the end of the string,
             * continue with the next loop.
             */
            break;
        }
        do {
            ch=*src++;
            if(ch <= 0x7f && ch != 0) {
                *pDest++ = (uint8_t)ch;
            } else if(ch <= 0x7ff) {
                *pDest++=(uint8_t)((ch>>6)|0xc0);
                *pDest++=(uint8_t)((ch&0x3f)|0x80);
            } else {
                *pDest++=(uint8_t)((ch>>12)|0xe0);
                *pDest++=(uint8_t)(((ch>>6)&0x3f)|0x80);
                *pDest++=(uint8_t)((ch&0x3f)|0x80);
            }
        } while(--count > 0);
    }

    while(src<pSrcLimit) {
        ch=*src++;
        if(ch <= 0x7f && ch != 0) {
            if(pDest<pDestLimit) {
                *pDest++ = (uint8_t)ch;
            } else {
                reqLength = 1;
                break;
            }
        } else if(ch <= 0x7ff) {
            if((pDestLimit - pDest) >= 2) {
                *pDest++=(uint8_t)((ch>>6)|0xc0);
                *pDest++=(uint8_t)((ch&0x3f)|0x80);
            } else {
                reqLength = 2;
                break;
            }
        } else {
            if((pDestLimit - pDest) >= 3) {
                *pDest++=(uint8_t)((ch>>12)|0xe0);
                *pDest++=(uint8_t)(((ch>>6)&0x3f)|0x80);
                *pDest++=(uint8_t)((ch&0x3f)|0x80);
            } else {
                reqLength = 3;
                break;
            }
        }
    }
    while(src<pSrcLimit) {
        ch=*src++;
        if(ch <= 0x7f && ch != 0) {
            ++reqLength;
        } else if(ch<=0x7ff) {
            reqLength+=2;
        } else {
            reqLength+=3;
        }
    }

    reqLength+=(int32_t)(pDest - (uint8_t *)dest);
    if(pDestLength){
        *pDestLength = reqLength;
    }

    /* Terminate the buffer */
    u_terminateChars(dest, destCapacity, reqLength, pErrorCode);
    return dest;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2005-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  utext.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2005apr12
*   created by: Markus W. Scherer
*/















U_NAMESPACE_USE

#define I32_FLAG(bitIndex) ((int32_t)1<<(bitIndex))


static UBool
utext_access(UText *ut, int64_t index, UBool forward) {
    return ut->pFuncs->access(ut, index, forward);
}



U_CAPI UBool U_EXPORT2
utext_moveIndex32(UText *ut, int32_t delta) {
    UChar32  c;
    if (delta > 0) {
        do {
            if(ut->chunkOffset>=ut->chunkLength && !utext_access(ut, ut->chunkNativeLimit, TRUE)) {
                return FALSE;
            }
            c = ut->chunkContents[ut->chunkOffset];
            if (U16_IS_SURROGATE(c)) {
                c = utext_next32(ut);
                if (c == U_SENTINEL) {
                    return FALSE;
                }
            } else {
                ut->chunkOffset++;
            }
        } while(--delta>0);

    } else if (delta<0) {
        do {
            if(ut->chunkOffset<=0 && !utext_access(ut, ut->chunkNativeStart, FALSE)) {
                return FALSE;
            }
            c = ut->chunkContents[ut->chunkOffset-1];
            if (U16_IS_SURROGATE(c)) {
                c = utext_previous32(ut);
                if (c == U_SENTINEL) {
                    return FALSE;
                }
            } else {
                ut->chunkOffset--;
            }
        } while(++delta<0);
    }

    return TRUE;
}


U_CAPI int64_t U_EXPORT2
utext_nativeLength(UText *ut) {
    return ut->pFuncs->nativeLength(ut);
}


U_CAPI UBool U_EXPORT2
utext_isLengthExpensive(const UText *ut) {
    UBool r = (ut->providerProperties & I32_FLAG(UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE)) != 0;
    return r;
}


U_CAPI int64_t U_EXPORT2
utext_getNativeIndex(const UText *ut) {
    if(ut->chunkOffset <= ut->nativeIndexingLimit) {
        return ut->chunkNativeStart+ut->chunkOffset;
    } else {
        return ut->pFuncs->mapOffsetToNative(ut);
    }
}


U_CAPI void U_EXPORT2
utext_setNativeIndex(UText *ut, int64_t index) {
    if(index<ut->chunkNativeStart || index>=ut->chunkNativeLimit) {
        // The desired position is outside of the current chunk.
        // Access the new position.  Assume a forward iteration from here,
        // which will also be optimimum for a single random access.
        // Reverse iterations may suffer slightly.
        ut->pFuncs->access(ut, index, TRUE);
    } else if((int32_t)(index - ut->chunkNativeStart) <= ut->nativeIndexingLimit) {
        // utf-16 indexing.
        ut->chunkOffset=(int32_t)(index-ut->chunkNativeStart);
    } else {
         ut->chunkOffset=ut->pFuncs->mapNativeIndexToUTF16(ut, index);
    }
    // The convention is that the index must always be on a code point boundary.
    // Adjust the index position if it is in the middle of a surrogate pair.
    if (ut->chunkOffset<ut->chunkLength) {
        UChar c= ut->chunkContents[ut->chunkOffset];
        if (U16_IS_TRAIL(c)) {
            if (ut->chunkOffset==0) {
                ut->pFuncs->access(ut, ut->chunkNativeStart, FALSE);
            }
            if (ut->chunkOffset>0) {
                UChar lead = ut->chunkContents[ut->chunkOffset-1];
                if (U16_IS_LEAD(lead)) {
                    ut->chunkOffset--;
                }
            }
        }
    }
}



U_CAPI int64_t U_EXPORT2
utext_getPreviousNativeIndex(UText *ut) {
    //
    //  Fast-path the common case.
    //     Common means current position is not at the beginning of a chunk
    //     and the preceding character is not supplementary.
    //
    int32_t i = ut->chunkOffset - 1;
    int64_t result;
    if (i >= 0) {
        UChar c = ut->chunkContents[i];
        if (U16_IS_TRAIL(c) == FALSE) {
            if (i <= ut->nativeIndexingLimit) {
                result = ut->chunkNativeStart + i;
            } else {
                ut->chunkOffset = i;
                result = ut->pFuncs->mapOffsetToNative(ut);
                ut->chunkOffset++;
            }
            return result;
        }
    }

    // If at the start of text, simply return 0.
    if (ut->chunkOffset==0 && ut->chunkNativeStart==0) {
        return 0;
    }

    // Harder, less common cases.  We are at a chunk boundary, or on a surrogate.
    //    Keep it simple, use other functions to handle the edges.
    //
    utext_previous32(ut);
    result = UTEXT_GETNATIVEINDEX(ut);
    utext_next32(ut);
    return result;
}


//
//  utext_current32.  Get the UChar32 at the current position.
//                    UText iteration position is always on a code point boundary,
//                    never on the trail half of a surrogate pair.
//
U_CAPI UChar32 U_EXPORT2
utext_current32(UText *ut) {
    UChar32  c;
    if (ut->chunkOffset==ut->chunkLength) {
        // Current position is just off the end of the chunk.
        if (ut->pFuncs->access(ut, ut->chunkNativeLimit, TRUE) == FALSE) {
            // Off the end of the text.
            return U_SENTINEL;
        }
    }

    c = ut->chunkContents[ut->chunkOffset];
    if (U16_IS_LEAD(c) == FALSE) {
        // Normal, non-supplementary case.
        return c;
    }

    //
    //  Possible supplementary char.
    //
    UChar32   trail = 0;
    UChar32   supplementaryC = c;
    if ((ut->chunkOffset+1) < ut->chunkLength) {
        // The trail surrogate is in the same chunk.
        trail = ut->chunkContents[ut->chunkOffset+1];
    } else {
        //  The trail surrogate is in a different chunk.
        //     Because we must maintain the iteration position, we need to switch forward
        //     into the new chunk, get the trail surrogate, then revert the chunk back to the
        //     original one.
        //     An edge case to be careful of:  the entire text may end with an unpaired
        //        leading surrogate.  The attempt to access the trail will fail, but
        //        the original position before the unpaired lead still needs to be restored.
        int64_t  nativePosition = ut->chunkNativeLimit;
        int32_t  originalOffset = ut->chunkOffset;
        if (ut->pFuncs->access(ut, nativePosition, TRUE)) {
            trail = ut->chunkContents[ut->chunkOffset];
        }
        UBool r = ut->pFuncs->access(ut, nativePosition, FALSE);  // reverse iteration flag loads preceding chunk
        U_ASSERT(r==TRUE);
        ut->chunkOffset = originalOffset;
        if(!r) {
            return U_SENTINEL;
        }
    }

    if (U16_IS_TRAIL(trail)) {
        supplementaryC = U16_GET_SUPPLEMENTARY(c, trail);
    }
    return supplementaryC;

}


U_CAPI UChar32 U_EXPORT2
utext_char32At(UText *ut, int64_t nativeIndex) {
    UChar32 c = U_SENTINEL;

    // Fast path the common case.
    if (nativeIndex>=ut->chunkNativeStart && nativeIndex < ut->chunkNativeStart + ut->nativeIndexingLimit) {
        ut->chunkOffset = (int32_t)(nativeIndex - ut->chunkNativeStart);
        c = ut->chunkContents[ut->chunkOffset];
        if (U16_IS_SURROGATE(c) == FALSE) {
            return c;
        }
    }


    utext_setNativeIndex(ut, nativeIndex);
    if (nativeIndex>=ut->chunkNativeStart && ut->chunkOffset<ut->chunkLength) {
        c = ut->chunkContents[ut->chunkOffset];
        if (U16_IS_SURROGATE(c)) {
            // For surrogates, let current32() deal with the complications
            //    of supplementaries that may span chunk boundaries.
            c = utext_current32(ut);
        }
    }
    return c;
}


U_CAPI UChar32 U_EXPORT2
utext_next32(UText *ut) {
    UChar32       c;

    if (ut->chunkOffset >= ut->chunkLength) {
        if (ut->pFuncs->access(ut, ut->chunkNativeLimit, TRUE) == FALSE) {
            return U_SENTINEL;
        }
    }

    c = ut->chunkContents[ut->chunkOffset++];
    if (U16_IS_LEAD(c) == FALSE) {
        // Normal case, not supplementary.
        //   (A trail surrogate seen here is just returned as is, as a surrogate value.
        //    It cannot be part of a pair.)
        return c;
    }

    if (ut->chunkOffset >= ut->chunkLength) {
        if (ut->pFuncs->access(ut, ut->chunkNativeLimit, TRUE) == FALSE) {
            // c is an unpaired lead surrogate at the end of the text.
            // return it as it is.
            return c;
        }
    }
    UChar32 trail = ut->chunkContents[ut->chunkOffset];
    if (U16_IS_TRAIL(trail) == FALSE) {
        // c was an unpaired lead surrogate, not at the end of the text.
        // return it as it is (unpaired).  Iteration position is on the
        // following character, possibly in the next chunk, where the
        //  trail surrogate would have been if it had existed.
        return c;
    }

    UChar32 supplementary = U16_GET_SUPPLEMENTARY(c, trail);
    ut->chunkOffset++;   // move iteration position over the trail surrogate.
    return supplementary;
    }


U_CAPI UChar32 U_EXPORT2
utext_previous32(UText *ut) {
    UChar32       c;

    if (ut->chunkOffset <= 0) {
        if (ut->pFuncs->access(ut, ut->chunkNativeStart, FALSE) == FALSE) {
            return U_SENTINEL;
        }
    }
    ut->chunkOffset--;
    c = ut->chunkContents[ut->chunkOffset];
    if (U16_IS_TRAIL(c) == FALSE) {
        // Normal case, not supplementary.
        //   (A lead surrogate seen here is just returned as is, as a surrogate value.
        //    It cannot be part of a pair.)
        return c;
    }

    if (ut->chunkOffset <= 0) {
        if (ut->pFuncs->access(ut, ut->chunkNativeStart, FALSE) == FALSE) {
            // c is an unpaired trail surrogate at the start of the text.
            // return it as it is.
            return c;
        }
    }

    UChar32 lead = ut->chunkContents[ut->chunkOffset-1];
    if (U16_IS_LEAD(lead) == FALSE) {
        // c was an unpaired trail surrogate, not at the end of the text.
        // return it as it is (unpaired).  Iteration position is at c
        return c;
    }

    UChar32 supplementary = U16_GET_SUPPLEMENTARY(lead, c);
    ut->chunkOffset--;   // move iteration position over the lead surrogate.
    return supplementary;
}



U_CAPI UChar32 U_EXPORT2
utext_next32From(UText *ut, int64_t index) {
    UChar32       c      = U_SENTINEL;

    if(index<ut->chunkNativeStart || index>=ut->chunkNativeLimit) {
        // Desired position is outside of the current chunk.
        if(!ut->pFuncs->access(ut, index, TRUE)) {
            // no chunk available here
            return U_SENTINEL;
        }
    } else if (index - ut->chunkNativeStart  <= (int64_t)ut->nativeIndexingLimit) {
        // Desired position is in chunk, with direct 1:1 native to UTF16 indexing
        ut->chunkOffset = (int32_t)(index - ut->chunkNativeStart);
    } else {
        // Desired position is in chunk, with non-UTF16 indexing.
        ut->chunkOffset = ut->pFuncs->mapNativeIndexToUTF16(ut, index);
    }

    c = ut->chunkContents[ut->chunkOffset++];
    if (U16_IS_SURROGATE(c)) {
        // Surrogates.  Many edge cases.  Use other functions that already
        //              deal with the problems.
        utext_setNativeIndex(ut, index);
        c = utext_next32(ut);
    }
    return c;
}


U_CAPI UChar32 U_EXPORT2
utext_previous32From(UText *ut, int64_t index) {
    //
    //  Return the character preceding the specified index.
    //  Leave the iteration position at the start of the character that was returned.
    //
    UChar32     cPrev;    // The character preceding cCurr, which is what we will return.

    // Address the chunk containg the position preceding the incoming index
    // A tricky edge case:
    //   We try to test the requested native index against the chunkNativeStart to determine
    //    whether the character preceding the one at the index is in the current chunk.
    //    BUT, this test can fail with UTF-8 (or any other multibyte encoding), when the
    //    requested index is on something other than the first position of the first char.
    //
    if(index<=ut->chunkNativeStart || index>ut->chunkNativeLimit) {
        // Requested native index is outside of the current chunk.
        if(!ut->pFuncs->access(ut, index, FALSE)) {
            // no chunk available here
            return U_SENTINEL;
        }
    } else if(index - ut->chunkNativeStart <= (int64_t)ut->nativeIndexingLimit) {
        // Direct UTF-16 indexing.
        ut->chunkOffset = (int32_t)(index - ut->chunkNativeStart);
    } else {
        ut->chunkOffset=ut->pFuncs->mapNativeIndexToUTF16(ut, index);
        if (ut->chunkOffset==0 && !ut->pFuncs->access(ut, index, FALSE)) {
            // no chunk available here
            return U_SENTINEL;
        }
    }

    //
    // Simple case with no surrogates.
    //
    ut->chunkOffset--;
    cPrev = ut->chunkContents[ut->chunkOffset];

    if (U16_IS_SURROGATE(cPrev)) {
        // Possible supplementary.  Many edge cases.
        // Let other functions do the heavy lifting.
        utext_setNativeIndex(ut, index);
        cPrev = utext_previous32(ut);
    }
    return cPrev;
}


U_CAPI int32_t U_EXPORT2
utext_extract(UText *ut,
             int64_t start, int64_t limit,
             UChar *dest, int32_t destCapacity,
             UErrorCode *status) {
                 return ut->pFuncs->extract(ut, start, limit, dest, destCapacity, status);
             }



U_CAPI UBool U_EXPORT2
utext_equals(const UText *a, const UText *b) {
    if (a==NULL || b==NULL ||
        a->magic != UTEXT_MAGIC ||
        b->magic != UTEXT_MAGIC) {
            // Null or invalid arguments don't compare equal to anything.
            return FALSE;
    }

    if (a->pFuncs != b->pFuncs) {
        // Different types of text providers.
        return FALSE;
    }

    if (a->context != b->context) {
        // Different sources (different strings)
        return FALSE;
    }
    if (utext_getNativeIndex(a) != utext_getNativeIndex(b)) {
        // Different current position in the string.
        return FALSE;
    }

    return TRUE;
}

U_CAPI UBool U_EXPORT2
utext_isWritable(const UText *ut)
{
    UBool b = (ut->providerProperties & I32_FLAG(UTEXT_PROVIDER_WRITABLE)) != 0;
    return b;
}


U_CAPI void U_EXPORT2
utext_freeze(UText *ut) {
    // Zero out the WRITABLE flag.
    ut->providerProperties &= ~(I32_FLAG(UTEXT_PROVIDER_WRITABLE));
}


U_CAPI UBool U_EXPORT2
utext_hasMetaData(const UText *ut)
{
    UBool b = (ut->providerProperties & I32_FLAG(UTEXT_PROVIDER_HAS_META_DATA)) != 0;
    return b;
}



U_CAPI int32_t U_EXPORT2
utext_replace(UText *ut,
             int64_t nativeStart, int64_t nativeLimit,
             const UChar *replacementText, int32_t replacementLength,
             UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return 0;
    }
    if ((ut->providerProperties & I32_FLAG(UTEXT_PROVIDER_WRITABLE)) == 0) {
        *status = U_NO_WRITE_PERMISSION;
        return 0;
    }
    int32_t i = ut->pFuncs->replace(ut, nativeStart, nativeLimit, replacementText, replacementLength, status);
    return i;
}

U_CAPI void U_EXPORT2
utext_copy(UText *ut,
          int64_t nativeStart, int64_t nativeLimit,
          int64_t destIndex,
          UBool move,
          UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return;
    }
    if ((ut->providerProperties & I32_FLAG(UTEXT_PROVIDER_WRITABLE)) == 0) {
        *status = U_NO_WRITE_PERMISSION;
        return;
    }
    ut->pFuncs->copy(ut, nativeStart, nativeLimit, destIndex, move, status);
}



U_CAPI UText * U_EXPORT2
utext_clone(UText *dest, const UText *src, UBool deep, UBool readOnly, UErrorCode *status) {
    if (U_FAILURE(*status)) {
        return dest;
    }
    UText *result = src->pFuncs->clone(dest, src, deep, status);
    if (U_FAILURE(*status)) {
        return result;
    }
    if (result == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return result;
    }
    if (readOnly) {
        utext_freeze(result);
    }
    return result;
}



//------------------------------------------------------------------------------
//
//   UText common functions implementation
//
//------------------------------------------------------------------------------

//
//  UText.flags bit definitions
//
enum {
    UTEXT_HEAP_ALLOCATED  = 1,      //  1 if ICU has allocated this UText struct on the heap.
                                    //  0 if caller provided storage for the UText.

    UTEXT_EXTRA_HEAP_ALLOCATED = 2, //  1 if ICU has allocated extra storage as a separate
                                    //     heap block.
                                    //  0 if there is no separate allocation.  Either no extra
                                    //     storage was requested, or it is appended to the end
                                    //     of the main UText storage.

    UTEXT_OPEN = 4                  //  1 if this UText is currently open
                                    //  0 if this UText is not open.
};


//
//  Extended form of a UText.  The purpose is to aid in computing the total size required
//    when a provider asks for a UText to be allocated with extra storage.

struct ExtendedUText {
    UText          ut;
    max_align_t    extension;
};

static const UText emptyText = UTEXT_INITIALIZER;

U_CAPI UText * U_EXPORT2
utext_setup(UText *ut, int32_t extraSpace, UErrorCode *status) {
    if (U_FAILURE(*status)) {
        return ut;
    }

    if (ut == NULL) {
        // We need to heap-allocate storage for the new UText
        int32_t spaceRequired = sizeof(UText);
        if (extraSpace > 0) {
            spaceRequired = sizeof(ExtendedUText) + extraSpace - sizeof(max_align_t);
        }
        ut = (UText *)uprv_malloc(spaceRequired);
        if (ut == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        } else {
            *ut = emptyText;
            ut->flags |= UTEXT_HEAP_ALLOCATED;
            if (spaceRequired>0) {
                ut->extraSize = extraSpace;
                ut->pExtra    = &((ExtendedUText *)ut)->extension;
            }
        }
    } else {
        // We have been supplied with an already existing UText.
        // Verify that it really appears to be a UText.
        if (ut->magic != UTEXT_MAGIC) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
            return ut;
        }
        // If the ut is already open and there's a provider supplied close
        //   function, call it.
        if ((ut->flags & UTEXT_OPEN) && ut->pFuncs->close != NULL)  {
            ut->pFuncs->close(ut);
        }
        ut->flags &= ~UTEXT_OPEN;

        // If extra space was requested by our caller, check whether
        //   sufficient already exists, and allocate new if needed.
        if (extraSpace > ut->extraSize) {
            // Need more space.  If there is existing separately allocated space,
            //   delete it first, then allocate new space.
            if (ut->flags & UTEXT_EXTRA_HEAP_ALLOCATED) {
                uprv_free(ut->pExtra);
                ut->extraSize = 0;
            }
            ut->pExtra = uprv_malloc(extraSpace);
            if (ut->pExtra == NULL) {
                *status = U_MEMORY_ALLOCATION_ERROR;
            } else {
                ut->extraSize = extraSpace;
                ut->flags |= UTEXT_EXTRA_HEAP_ALLOCATED;
            }
        }
    }
    if (U_SUCCESS(*status)) {
        ut->flags |= UTEXT_OPEN;

        // Initialize all remaining fields of the UText.
        //
        ut->context             = NULL;
        ut->chunkContents       = NULL;
        ut->p                   = NULL;
        ut->q                   = NULL;
        ut->r                   = NULL;
        ut->a                   = 0;
        ut->b                   = 0;
        ut->c                   = 0;
        ut->chunkOffset         = 0;
        ut->chunkLength         = 0;
        ut->chunkNativeStart    = 0;
        ut->chunkNativeLimit    = 0;
        ut->nativeIndexingLimit = 0;
        ut->providerProperties  = 0;
        ut->privA               = 0;
        ut->privB               = 0;
        ut->privC               = 0;
        ut->privP               = NULL;
        if (ut->pExtra!=NULL && ut->extraSize>0)
            uprv_memset(ut->pExtra, 0, ut->extraSize);

    }
    return ut;
}


U_CAPI UText * U_EXPORT2
utext_close(UText *ut) {
    if (ut==NULL ||
        ut->magic != UTEXT_MAGIC ||
        (ut->flags & UTEXT_OPEN) == 0)
    {
        // The supplied ut is not an open UText.
        // Do nothing.
        return ut;
    }

    // If the provider gave us a close function, call it now.
    // This will clean up anything allocated specifically by the provider.
    if (ut->pFuncs->close != NULL) {
        ut->pFuncs->close(ut);
    }
    ut->flags &= ~UTEXT_OPEN;

    // If we (the framework) allocated the UText or subsidiary storage,
    //   delete it.
    if (ut->flags & UTEXT_EXTRA_HEAP_ALLOCATED) {
        uprv_free(ut->pExtra);
        ut->pExtra = NULL;
        ut->flags &= ~UTEXT_EXTRA_HEAP_ALLOCATED;
        ut->extraSize = 0;
    }

    // Zero out function table of the closed UText.  This is a defensive move,
    //   inteded to cause applications that inadvertantly use a closed
    //   utext to crash with null pointer errors.
    ut->pFuncs        = NULL;

    if (ut->flags & UTEXT_HEAP_ALLOCATED) {
        // This UText was allocated by UText setup.  We need to free it.
        // Clear magic, so we can detect if the user messes up and immediately
        //  tries to reopen another UText using the deleted storage.
        ut->magic = 0;
        uprv_free(ut);
        ut = NULL;
    }
    return ut;
}




//
// invalidateChunk   Reset a chunk to have no contents, so that the next call
//                   to access will cause new data to load.
//                   This is needed when copy/move/replace operate directly on the
//                   backing text, potentially putting it out of sync with the
//                   contents in the chunk.
//
static void
invalidateChunk(UText *ut) {
    ut->chunkLength = 0;
    ut->chunkNativeLimit = 0;
    ut->chunkNativeStart = 0;
    ut->chunkOffset = 0;
    ut->nativeIndexingLimit = 0;
}

//
// pinIndex        Do range pinning on a native index parameter.
//                 64 bit pinning is done in place.
//                 32 bit truncated result is returned as a convenience for
//                        use in providers that don't need 64 bits.
static int32_t
pinIndex(int64_t &index, int64_t limit) {
    if (index<0) {
        index = 0;
    } else if (index > limit) {
        index = limit;
    }
    return (int32_t)index;
}


U_CDECL_BEGIN

//
// Pointer relocation function,
//   a utility used by shallow clone.
//   Adjust a pointer that refers to something within one UText (the source)
//   to refer to the same relative offset within a another UText (the target)
//
static void adjustPointer(UText *dest, const void **destPtr, const UText *src) {
    // convert all pointers to (char *) so that byte address arithmetic will work.
    char  *dptr = (char *)*destPtr;
    char  *dUText = (char *)dest;
    char  *sUText = (char *)src;

    if (dptr >= (char *)src->pExtra && dptr < ((char*)src->pExtra)+src->extraSize) {
        // target ptr was to something within the src UText's pExtra storage.
        //   relocate it into the target UText's pExtra region.
        *destPtr = ((char *)dest->pExtra) + (dptr - (char *)src->pExtra);
    } else if (dptr>=sUText && dptr < sUText+src->sizeOfStruct) {
        // target ptr was pointing to somewhere within the source UText itself.
        //   Move it to the same offset within the target UText.
        *destPtr = dUText + (dptr-sUText);
    }
}


//
//  Clone.  This is a generic copy-the-utext-by-value clone function that can be
//          used as-is with some utext types, and as a helper by other clones.
//
static UText * U_CALLCONV
shallowTextClone(UText * dest, const UText * src, UErrorCode * status) {
    if (U_FAILURE(*status)) {
        return NULL;
    }
    int32_t  srcExtraSize = src->extraSize;

    //
    // Use the generic text_setup to allocate storage if required.
    //
    dest = utext_setup(dest, srcExtraSize, status);
    if (U_FAILURE(*status)) {
        return dest;
    }

    //
    //  flags (how the UText was allocated) and the pointer to the
    //   extra storage must retain the values in the cloned utext that
    //   were set up by utext_setup.  Save them separately before
    //   copying the whole struct.
    //
    void *destExtra = dest->pExtra;
    int32_t flags   = dest->flags;


    //
    //  Copy the whole UText struct by value.
    //  Any "Extra" storage is copied also.
    //
    int sizeToCopy = src->sizeOfStruct;
    if (sizeToCopy > dest->sizeOfStruct) {
        sizeToCopy = dest->sizeOfStruct;
    }
    uprv_memcpy(dest, src, sizeToCopy);
    dest->pExtra = destExtra;
    dest->flags  = flags;
    if (srcExtraSize > 0) {
        uprv_memcpy(dest->pExtra, src->pExtra, srcExtraSize);
    }

    //
    // Relocate any pointers in the target that refer to the UText itself
    //   to point to the cloned copy rather than the original source.
    //
    adjustPointer(dest, &dest->context, src);
    adjustPointer(dest, &dest->p, src);
    adjustPointer(dest, &dest->q, src);
    adjustPointer(dest, &dest->r, src);
    adjustPointer(dest, (const void **)&dest->chunkContents, src);

    // The newly shallow-cloned UText does _not_ own the underlying storage for the text.
    // (The source for the clone may or may not have owned the text.)

    dest->providerProperties &= ~I32_FLAG(UTEXT_PROVIDER_OWNS_TEXT);

    return dest;
}


U_CDECL_END



//------------------------------------------------------------------------------
//
//     UText implementation for UTF-8 char * strings (read-only)
//     Limitation:  string length must be <= 0x7fffffff in length.
//                  (length must for in an int32_t variable)
//
//         Use of UText data members:
//              context    pointer to UTF-8 string
//              utext.b    is the input string length (bytes).
//              utext.c    Length scanned so far in string
//                           (for optimizing finding length of zero terminated strings.)
//              utext.p    pointer to the current buffer
//              utext.q    pointer to the other buffer.
//
//------------------------------------------------------------------------------

// Chunk size.
//     Must be less than 85 (256/3), because of byte mapping from UChar indexes to native indexes.
//     Worst case is three native bytes to one UChar.  (Supplemenaries are 4 native bytes
//     to two UChars.)
//     The longest illegal byte sequence treated as a single error (and converted to U+FFFD)
//     is a three-byte sequence (truncated four-byte sequence).
//
enum { UTF8_TEXT_CHUNK_SIZE=32 };

//
// UTF8Buf  Two of these structs will be set up in the UText's extra allocated space.
//          Each contains the UChar chunk buffer, the to and from native maps, and
//          header info.
//
//     because backwards iteration fills the buffers starting at the end and
//     working towards the front, the filled part of the buffers may not begin
//     at the start of the available storage for the buffers.
//
//     Buffer size is one bigger than the specified UTF8_TEXT_CHUNK_SIZE to allow for
//     the last character added being a supplementary, and thus requiring a surrogate
//     pair.  Doing this is simpler than checking for the edge case.
//

struct UTF8Buf {
    int32_t   bufNativeStart;                        // Native index of first char in UChar buf
    int32_t   bufNativeLimit;                        // Native index following last char in buf.
    int32_t   bufStartIdx;                           // First filled position in buf.
    int32_t   bufLimitIdx;                           // Limit of filled range in buf.
    int32_t   bufNILimit;                            // Limit of native indexing part of buf
    int32_t   toUCharsMapStart;                      // Native index corresponding to
                                                     //   mapToUChars[0].
                                                     //   Set to bufNativeStart when filling forwards.
                                                     //   Set to computed value when filling backwards.

    UChar     buf[UTF8_TEXT_CHUNK_SIZE+4];           // The UChar buffer.  Requires one extra position beyond the
                                                     //   the chunk size, to allow for surrogate at the end.
                                                     //   Length must be identical to mapToNative array, below,
                                                     //   because of the way indexing works when the array is
                                                     //   filled backwards during a reverse iteration.  Thus,
                                                     //   the additional extra size.
    uint8_t   mapToNative[UTF8_TEXT_CHUNK_SIZE+4];   // map UChar index in buf to
                                                     //  native offset from bufNativeStart.
                                                     //  Requires two extra slots,
                                                     //    one for a supplementary starting in the last normal position,
                                                     //    and one for an entry for the buffer limit position.
    uint8_t   mapToUChars[UTF8_TEXT_CHUNK_SIZE*3+6]; // Map native offset from bufNativeStart to
                                                     //   correspoding offset in filled part of buf.
    int32_t   align;
};

U_CDECL_BEGIN

//
//   utf8TextLength
//
//        Get the length of the string.  If we don't already know it,
//              we'll need to scan for the trailing  nul.
//
static int64_t U_CALLCONV
utf8TextLength(UText *ut) {
    if (ut->b < 0) {
        // Zero terminated string, and we haven't scanned to the end yet.
        // Scan it now.
        const char *r = (const char *)ut->context + ut->c;
        while (*r != 0) {
            r++;
        }
        if ((r - (const char *)ut->context) < 0x7fffffff) {
            ut->b = (int32_t)(r - (const char *)ut->context);
        } else {
            // Actual string was bigger (more than 2 gig) than we
            //   can handle.  Clip it to 2 GB.
            ut->b = 0x7fffffff;
        }
        ut->providerProperties &= ~I32_FLAG(UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE);
    }
    return ut->b;
}






static UBool U_CALLCONV
utf8TextAccess(UText *ut, int64_t index, UBool forward) {
    //
    //  Apologies to those who are allergic to goto statements.
    //    Consider each goto to a labelled block to be the equivalent of
    //         call the named block as if it were a function();
    //         return;
    //
    const uint8_t *s8=(const uint8_t *)ut->context;
    UTF8Buf *u8b = NULL;
    int32_t  length = ut->b;         // Length of original utf-8
    int32_t  ix= (int32_t)index;     // Requested index, trimmed to 32 bits.
    int32_t  mapIndex = 0;
    if (index<0) {
        ix=0;
    } else if (index > 0x7fffffff) {
        // Strings with 64 bit lengths not supported by this UTF-8 provider.
        ix = 0x7fffffff;
    }

    // Pin requested index to the string length.
    if (ix>length) {
        if (length>=0) {
            ix=length;
        } else if (ix>=ut->c) {
            // Zero terminated string, and requested index is beyond
            //   the region that has already been scanned.
            //   Scan up to either the end of the string or to the
            //   requested position, whichever comes first.
            while (ut->c<ix && s8[ut->c]!=0) {
                ut->c++;
            }
            //  TODO:  support for null terminated string length > 32 bits.
            if (s8[ut->c] == 0) {
                // We just found the actual length of the string.
                //  Trim the requested index back to that.
                ix     = ut->c;
                ut->b  = ut->c;
                length = ut->c;
                ut->providerProperties &= ~I32_FLAG(UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE);
            }
        }
    }

    //
    // Dispatch to the appropriate action for a forward iteration request.
    //
    if (forward) {
        if (ix==ut->chunkNativeLimit) {
            // Check for normal sequential iteration cases first.
            if (ix==length) {
                // Just reached end of string
                // Don't swap buffers, but do set the
                //   current buffer position.
                ut->chunkOffset = ut->chunkLength;
                return FALSE;
            } else {
                // End of current buffer.
                //   check whether other buffer already has what we need.
                UTF8Buf *altB = (UTF8Buf *)ut->q;
                if (ix>=altB->bufNativeStart && ix<altB->bufNativeLimit) {
                    goto swapBuffers;
                }
            }
        }

        // A random access.  Desired index could be in either or niether buf.
        // For optimizing the order of testing, first check for the index
        //    being in the other buffer.  This will be the case for uses that
        //    move back and forth over a fairly limited range
        {
            u8b = (UTF8Buf *)ut->q;   // the alternate buffer
            if (ix>=u8b->bufNativeStart && ix<u8b->bufNativeLimit) {
                // Requested index is in the other buffer.
                goto swapBuffers;
            }
            if (ix == length) {
                // Requested index is end-of-string.
                //   (this is the case of randomly seeking to the end.
                //    The case of iterating off the end is handled earlier.)
                if (ix == ut->chunkNativeLimit) {
                    // Current buffer extends up to the end of the string.
                    //   Leave it as the current buffer.
                    ut->chunkOffset = ut->chunkLength;
                    return FALSE;
                }
                if (ix == u8b->bufNativeLimit) {
                    // Alternate buffer extends to the end of string.
                    //   Swap it in as the current buffer.
                    goto swapBuffersAndFail;
                }

                // Neither existing buffer extends to the end of the string.
                goto makeStubBuffer;
            }

            if (ix<ut->chunkNativeStart || ix>=ut->chunkNativeLimit) {
                // Requested index is in neither buffer.
                goto fillForward;
            }

            // Requested index is in this buffer.
            u8b = (UTF8Buf *)ut->p;   // the current buffer
            mapIndex = ix - u8b->toUCharsMapStart;
            U_ASSERT(mapIndex < (int32_t)sizeof(UTF8Buf::mapToUChars));
            ut->chunkOffset = u8b->mapToUChars[mapIndex] - u8b->bufStartIdx;
            return TRUE;

        }
    }


    //
    // Dispatch to the appropriate action for a
    //   Backwards Diretion iteration request.
    //
    if (ix==ut->chunkNativeStart) {
        // Check for normal sequential iteration cases first.
        if (ix==0) {
            // Just reached the start of string
            // Don't swap buffers, but do set the
            //   current buffer position.
            ut->chunkOffset = 0;
            return FALSE;
        } else {
            // Start of current buffer.
            //   check whether other buffer already has what we need.
            UTF8Buf *altB = (UTF8Buf *)ut->q;
            if (ix>altB->bufNativeStart && ix<=altB->bufNativeLimit) {
                goto swapBuffers;
            }
        }
    }

    // A random access.  Desired index could be in either or niether buf.
    // For optimizing the order of testing,
    //    Most likely case:  in the other buffer.
    //    Second most likely: in neither buffer.
    //    Unlikely, but must work:  in the current buffer.
    u8b = (UTF8Buf *)ut->q;   // the alternate buffer
    if (ix>u8b->bufNativeStart && ix<=u8b->bufNativeLimit) {
        // Requested index is in the other buffer.
        goto swapBuffers;
    }
    // Requested index is start-of-string.
    //   (this is the case of randomly seeking to the start.
    //    The case of iterating off the start is handled earlier.)
    if (ix==0) {
        if (u8b->bufNativeStart==0) {
            // Alternate buffer contains the data for the start string.
            // Make it be the current buffer.
            goto swapBuffersAndFail;
        } else {
            // Request for data before the start of string,
            //   neither buffer is usable.
            //   set up a zero-length buffer.
            goto makeStubBuffer;
        }
    }

    if (ix<=ut->chunkNativeStart || ix>ut->chunkNativeLimit) {
        // Requested index is in neither buffer.
        goto fillReverse;
    }

    // Requested index is in this buffer.
    //   Set the utf16 buffer index.
    u8b = (UTF8Buf *)ut->p;
    mapIndex = ix - u8b->toUCharsMapStart;
    ut->chunkOffset = u8b->mapToUChars[mapIndex] - u8b->bufStartIdx;
    if (ut->chunkOffset==0) {
        // This occurs when the first character in the text is
        //   a multi-byte UTF-8 char, and the requested index is to
        //   one of the trailing bytes.  Because there is no preceding ,
        //   character, this access fails.  We can't pick up on the
        //   situation sooner because the requested index is not zero.
        return FALSE;
    } else {
        return TRUE;
    }



swapBuffers:
    //  The alternate buffer (ut->q) has the string data that was requested.
    //  Swap the primary and alternate buffers, and set the
    //   chunk index into the new primary buffer.
    {
        u8b   = (UTF8Buf *)ut->q;
        ut->q = ut->p;
        ut->p = u8b;
        ut->chunkContents       = &u8b->buf[u8b->bufStartIdx];
        ut->chunkLength         = u8b->bufLimitIdx - u8b->bufStartIdx;
        ut->chunkNativeStart    = u8b->bufNativeStart;
        ut->chunkNativeLimit    = u8b->bufNativeLimit;
        ut->nativeIndexingLimit = u8b->bufNILimit;

        // Index into the (now current) chunk
        // Use the map to set the chunk index.  It's more trouble than it's worth
        //    to check whether native indexing can be used.
        U_ASSERT(ix>=u8b->bufNativeStart);
        U_ASSERT(ix<=u8b->bufNativeLimit);
        mapIndex = ix - u8b->toUCharsMapStart;
        U_ASSERT(mapIndex>=0);
        U_ASSERT(mapIndex<(int32_t)sizeof(u8b->mapToUChars));
        ut->chunkOffset = u8b->mapToUChars[mapIndex] - u8b->bufStartIdx;

        return TRUE;
    }


 swapBuffersAndFail:
    // We got a request for either the start or end of the string,
    //  with iteration continuing in the out-of-bounds direction.
    // The alternate buffer already contains the data up to the
    //  start/end.
    // Swap the buffers, then return failure, indicating that we couldn't
    //  make things correct for continuing the iteration in the requested
    //  direction.  The position & buffer are correct should the
    //  user decide to iterate in the opposite direction.
    u8b   = (UTF8Buf *)ut->q;
    ut->q = ut->p;
    ut->p = u8b;
    ut->chunkContents       = &u8b->buf[u8b->bufStartIdx];
    ut->chunkLength         = u8b->bufLimitIdx - u8b->bufStartIdx;
    ut->chunkNativeStart    = u8b->bufNativeStart;
    ut->chunkNativeLimit    = u8b->bufNativeLimit;
    ut->nativeIndexingLimit = u8b->bufNILimit;

    // Index into the (now current) chunk
    //  For this function  (swapBuffersAndFail), the requested index
    //    will always be at either the start or end of the chunk.
    if (ix==u8b->bufNativeLimit) {
        ut->chunkOffset = ut->chunkLength;
    } else  {
        ut->chunkOffset = 0;
        U_ASSERT(ix == u8b->bufNativeStart);
    }
    return FALSE;

makeStubBuffer:
    //   The user has done a seek/access past the start or end
    //   of the string.  Rather than loading data that is likely
    //   to never be used, just set up a zero-length buffer at
    //   the position.
    u8b = (UTF8Buf *)ut->q;
    u8b->bufNativeStart   = ix;
    u8b->bufNativeLimit   = ix;
    u8b->bufStartIdx      = 0;
    u8b->bufLimitIdx      = 0;
    u8b->bufNILimit       = 0;
    u8b->toUCharsMapStart = ix;
    u8b->mapToNative[0]   = 0;
    u8b->mapToUChars[0]   = 0;
    goto swapBuffersAndFail;



fillForward:
    {
        // Move the incoming index to a code point boundary.
        U8_SET_CP_START(s8, 0, ix);

        // Swap the UText buffers.
        //  We want to fill what was previously the alternate buffer,
        //  and make what was the current buffer be the new alternate.
        UTF8Buf *u8b_swap = (UTF8Buf *)ut->q;
        ut->q = ut->p;
        ut->p = u8b_swap;

        int32_t strLen = ut->b;
        UBool   nulTerminated = FALSE;
        if (strLen < 0) {
            strLen = 0x7fffffff;
            nulTerminated = TRUE;
        }

        UChar   *buf = u8b_swap->buf;
        uint8_t *mapToNative  = u8b_swap->mapToNative;
        uint8_t *mapToUChars  = u8b_swap->mapToUChars;
        int32_t  destIx       = 0;
        int32_t  srcIx        = ix;
        UBool    seenNonAscii = FALSE;
        UChar32  c = 0;

        // Fill the chunk buffer and mapping arrays.
        while (destIx<UTF8_TEXT_CHUNK_SIZE) {
            c = s8[srcIx];
            if (c>0 && c<0x80) {
                // Special case ASCII range for speed.
                //   zero is excluded to simplify bounds checking.
                buf[destIx] = (UChar)c;
                mapToNative[destIx]    = (uint8_t)(srcIx - ix);
                mapToUChars[srcIx-ix]  = (uint8_t)destIx;
                srcIx++;
                destIx++;
            } else {
                // General case, handle everything.
                if (seenNonAscii == FALSE) {
                    seenNonAscii = TRUE;
                    u8b_swap->bufNILimit = destIx;
                }

                int32_t  cIx      = srcIx;
                int32_t  dIx      = destIx;
                int32_t  dIxSaved = destIx;
                U8_NEXT_OR_FFFD(s8, srcIx, strLen, c);
                if (c==0 && nulTerminated) {
                    srcIx--;
                    break;
                }

                U16_APPEND_UNSAFE(buf, destIx, c);
                do {
                    mapToNative[dIx++] = (uint8_t)(cIx - ix);
                } while (dIx < destIx);

                do {
                    mapToUChars[cIx++ - ix] = (uint8_t)dIxSaved;
                } while (cIx < srcIx);
            }
            if (srcIx>=strLen) {
                break;
            }

        }

        //  store Native <--> Chunk Map entries for the end of the buffer.
        //    There is no actual character here, but the index position is valid.
        mapToNative[destIx]     = (uint8_t)(srcIx - ix);
        mapToUChars[srcIx - ix] = (uint8_t)destIx;

        //  fill in Buffer descriptor
        u8b_swap->bufNativeStart     = ix;
        u8b_swap->bufNativeLimit     = srcIx;
        u8b_swap->bufStartIdx        = 0;
        u8b_swap->bufLimitIdx        = destIx;
        if (seenNonAscii == FALSE) {
            u8b_swap->bufNILimit     = destIx;
        }
        u8b_swap->toUCharsMapStart   = u8b_swap->bufNativeStart;

        // Set UText chunk to refer to this buffer.
        ut->chunkContents       = buf;
        ut->chunkOffset         = 0;
        ut->chunkLength         = u8b_swap->bufLimitIdx;
        ut->chunkNativeStart    = u8b_swap->bufNativeStart;
        ut->chunkNativeLimit    = u8b_swap->bufNativeLimit;
        ut->nativeIndexingLimit = u8b_swap->bufNILimit;

        // For zero terminated strings, keep track of the maximum point
        //   scanned so far.
        if (nulTerminated && srcIx>ut->c) {
            ut->c = srcIx;
            if (c==0) {
                // We scanned to the end.
                //   Remember the actual length.
                ut->b = srcIx;
                ut->providerProperties &= ~I32_FLAG(UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE);
            }
        }
        return TRUE;
    }


fillReverse:
    {
        // Move the incoming index to a code point boundary.
        // Can only do this if the incoming index is somewhere in the interior of the string.
        //   If index is at the end, there is no character there to look at.
        if (ix != ut->b) {
            // Note: this function will only move the index back if it is on a trail byte
            //       and there is a preceding lead byte and the sequence from the lead
            //       through this trail could be part of a valid UTF-8 sequence
            //       Otherwise the index remains unchanged.
            U8_SET_CP_START(s8, 0, ix);
        }

        // Swap the UText buffers.
        //  We want to fill what was previously the alternate buffer,
        //  and make what was the current buffer be the new alternate.
        UTF8Buf *u8b_swap = (UTF8Buf *)ut->q;
        ut->q = ut->p;
        ut->p = u8b_swap;

        UChar   *buf = u8b_swap->buf;
        uint8_t *mapToNative = u8b_swap->mapToNative;
        uint8_t *mapToUChars = u8b_swap->mapToUChars;
        int32_t  toUCharsMapStart = ix - sizeof(UTF8Buf::mapToUChars) + 1;
        // Note that toUCharsMapStart can be negative. Happens when the remaining
        // text from current position to the beginning is less than the buffer size.
        // + 1 because mapToUChars must have a slot at the end for the bufNativeLimit entry.
        int32_t  destIx = UTF8_TEXT_CHUNK_SIZE+2;   // Start in the overflow region
                                                    //   at end of buffer to leave room
                                                    //   for a surrogate pair at the
                                                    //   buffer start.
        int32_t  srcIx  = ix;
        int32_t  bufNILimit = destIx;
        UChar32   c;

        // Map to/from Native Indexes, fill in for the position at the end of
        //   the buffer.
        //
        mapToNative[destIx] = (uint8_t)(srcIx - toUCharsMapStart);
        mapToUChars[srcIx - toUCharsMapStart] = (uint8_t)destIx;

        // Fill the chunk buffer
        // Work backwards, filling from the end of the buffer towards the front.
        //
        while (destIx>2 && (srcIx - toUCharsMapStart > 5) && (srcIx > 0)) {
            srcIx--;
            destIx--;

            // Get last byte of the UTF-8 character
            c = s8[srcIx];
            if (c<0x80) {
                // Special case ASCII range for speed.
                buf[destIx] = (UChar)c;
                U_ASSERT(toUCharsMapStart <= srcIx);
                mapToUChars[srcIx - toUCharsMapStart] = (uint8_t)destIx;
                mapToNative[destIx] = (uint8_t)(srcIx - toUCharsMapStart);
            } else {
                // General case, handle everything non-ASCII.

                int32_t  sIx      = srcIx;  // ix of last byte of multi-byte u8 char

                // Get the full character from the UTF8 string.
                //   use code derived from tbe macros in utf8.h
                //   Leaves srcIx pointing at the first byte of the UTF-8 char.
                //
                c=utf8_prevCharSafeBody(s8, 0, &srcIx, c, -3);
                // leaves srcIx at first byte of the multi-byte char.

                // Store the character in UTF-16 buffer.
                if (c<0x10000) {
                    buf[destIx] = (UChar)c;
                    mapToNative[destIx] = (uint8_t)(srcIx - toUCharsMapStart);
                } else {
                    buf[destIx]         = U16_TRAIL(c);
                    mapToNative[destIx] = (uint8_t)(srcIx - toUCharsMapStart);
                    buf[--destIx]       = U16_LEAD(c);
                    mapToNative[destIx] = (uint8_t)(srcIx - toUCharsMapStart);
                }

                // Fill in the map from native indexes to UChars buf index.
                do {
                    mapToUChars[sIx-- - toUCharsMapStart] = (uint8_t)destIx;
                } while (sIx >= srcIx);
                U_ASSERT(toUCharsMapStart <= (srcIx+1));

                // Set native indexing limit to be the current position.
                //   We are processing a non-ascii, non-native-indexing char now;
                //     the limit will be here if the rest of the chars to be
                //     added to this buffer are ascii.
                bufNILimit = destIx;
            }
        }
        u8b_swap->bufNativeStart     = srcIx;
        u8b_swap->bufNativeLimit     = ix;
        u8b_swap->bufStartIdx        = destIx;
        u8b_swap->bufLimitIdx        = UTF8_TEXT_CHUNK_SIZE+2;
        u8b_swap->bufNILimit         = bufNILimit - u8b_swap->bufStartIdx;
        u8b_swap->toUCharsMapStart   = toUCharsMapStart;

        ut->chunkContents       = &buf[u8b_swap->bufStartIdx];
        ut->chunkLength         = u8b_swap->bufLimitIdx - u8b_swap->bufStartIdx;
        ut->chunkOffset         = ut->chunkLength;
        ut->chunkNativeStart    = u8b_swap->bufNativeStart;
        ut->chunkNativeLimit    = u8b_swap->bufNativeLimit;
        ut->nativeIndexingLimit = u8b_swap->bufNILimit;
        return TRUE;
    }

}



//
//  This is a slightly modified copy of u_strFromUTF8,
//     Inserts a Replacement Char rather than failing on invalid UTF-8
//     Removes unnecessary features.
//
static UChar*
utext_strFromUTF8(UChar *dest,
              int32_t destCapacity,
              int32_t *pDestLength,
              const char* src,
              int32_t srcLength,        // required.  NUL terminated not supported.
              UErrorCode *pErrorCode
              )
{

    UChar *pDest = dest;
    UChar *pDestLimit = (dest!=NULL)?(dest+destCapacity):NULL;
    UChar32 ch=0;
    int32_t index = 0;
    int32_t reqLength = 0;
    uint8_t* pSrc = (uint8_t*) src;


    while((index < srcLength)&&(pDest<pDestLimit)){
        ch = pSrc[index++];
        if(ch <=0x7f){
            *pDest++=(UChar)ch;
        }else{
            ch=utf8_nextCharSafeBody(pSrc, &index, srcLength, ch, -3);
            if(U_IS_BMP(ch)){
                *(pDest++)=(UChar)ch;
            }else{
                *(pDest++)=U16_LEAD(ch);
                if(pDest<pDestLimit){
                    *(pDest++)=U16_TRAIL(ch);
                }else{
                    reqLength++;
                    break;
                }
            }
        }
    }
    /* donot fill the dest buffer just count the UChars needed */
    while(index < srcLength){
        ch = pSrc[index++];
        if(ch <= 0x7f){
            reqLength++;
        }else{
            ch=utf8_nextCharSafeBody(pSrc, &index, srcLength, ch, -3);
            reqLength+=U16_LENGTH(ch);
        }
    }

    reqLength+=(int32_t)(pDest - dest);

    if(pDestLength){
        *pDestLength = reqLength;
    }

    /* Terminate the buffer */
    u_terminateUChars(dest,destCapacity,reqLength,pErrorCode);

    return dest;
}



static int32_t U_CALLCONV
utf8TextExtract(UText *ut,
                int64_t start, int64_t limit,
                UChar *dest, int32_t destCapacity,
                UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(destCapacity<0 || (dest==NULL && destCapacity>0)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    int32_t  length  = ut->b;
    int32_t  start32 = pinIndex(start, length);
    int32_t  limit32 = pinIndex(limit, length);

    if(start32>limit32) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }


    // adjust the incoming indexes to land on code point boundaries if needed.
    //    adjust by no more than three, because that is the largest number of trail bytes
    //    in a well formed UTF8 character.
    const uint8_t *buf = (const uint8_t *)ut->context;
    int i;
    if (start32 < ut->chunkNativeLimit) {
        for (i=0; i<3; i++) {
            if (U8_IS_SINGLE(buf[start32]) || U8_IS_LEAD(buf[start32]) || start32==0) {
                break;
            }
            start32--;
        }
    }

    if (limit32 < ut->chunkNativeLimit) {
        for (i=0; i<3; i++) {
            if (U8_IS_SINGLE(buf[limit32]) || U8_IS_LEAD(buf[limit32]) || limit32==0) {
                break;
            }
            limit32--;
        }
    }

    // Do the actual extract.
    int32_t destLength=0;
    utext_strFromUTF8(dest, destCapacity, &destLength,
                    (const char *)ut->context+start32, limit32-start32,
                    pErrorCode);
    utf8TextAccess(ut, limit32, TRUE);
    return destLength;
}

//
// utf8TextMapOffsetToNative
//
// Map a chunk (UTF-16) offset to a native index.
static int64_t U_CALLCONV
utf8TextMapOffsetToNative(const UText *ut) {
    //
    UTF8Buf *u8b = (UTF8Buf *)ut->p;
    U_ASSERT(ut->chunkOffset>ut->nativeIndexingLimit && ut->chunkOffset<=ut->chunkLength);
    int32_t nativeOffset = u8b->mapToNative[ut->chunkOffset + u8b->bufStartIdx] + u8b->toUCharsMapStart;
    U_ASSERT(nativeOffset >= ut->chunkNativeStart && nativeOffset <= ut->chunkNativeLimit);
    return nativeOffset;
}

//
// Map a native index to the corrsponding chunk offset
//
static int32_t U_CALLCONV
utf8TextMapIndexToUTF16(const UText *ut, int64_t index64) {
    U_ASSERT(index64 <= 0x7fffffff);
    int32_t index = (int32_t)index64;
    UTF8Buf *u8b = (UTF8Buf *)ut->p;
    U_ASSERT(index>=ut->chunkNativeStart+ut->nativeIndexingLimit);
    U_ASSERT(index<=ut->chunkNativeLimit);
    int32_t mapIndex = index - u8b->toUCharsMapStart;
    U_ASSERT(mapIndex < (int32_t)sizeof(UTF8Buf::mapToUChars));
    int32_t offset = u8b->mapToUChars[mapIndex] - u8b->bufStartIdx;
    U_ASSERT(offset>=0 && offset<=ut->chunkLength);
    return offset;
}

static UText * U_CALLCONV
utf8TextClone(UText *dest, const UText *src, UBool deep, UErrorCode *status)
{
    // First do a generic shallow clone.  Does everything needed for the UText struct itself.
    dest = shallowTextClone(dest, src, status);

    // For deep clones, make a copy of the string.
    //  The copied storage is owned by the newly created clone.
    //
    // TODO:  There is an isssue with using utext_nativeLength().
    //        That function is non-const in cases where the input was NUL terminated
    //          and the length has not yet been determined.
    //        This function (clone()) is const.
    //        There potentially a thread safety issue lurking here.
    //
    if (deep && U_SUCCESS(*status)) {
        int32_t  len = (int32_t)utext_nativeLength((UText *)src);
        char *copyStr = (char *)uprv_malloc(len+1);
        if (copyStr == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
        } else {
            uprv_memcpy(copyStr, src->context, len+1);
            dest->context = copyStr;
            dest->providerProperties |= I32_FLAG(UTEXT_PROVIDER_OWNS_TEXT);
        }
    }
    return dest;
}


static void U_CALLCONV
utf8TextClose(UText *ut) {
    // Most of the work of close is done by the generic UText framework close.
    // All that needs to be done here is to delete the UTF8 string if the UText
    //  owns it.  This occurs if the UText was created by cloning.
    if (ut->providerProperties & I32_FLAG(UTEXT_PROVIDER_OWNS_TEXT)) {
        char *s = (char *)ut->context;
        uprv_free(s);
        ut->context = NULL;
    }
}

U_CDECL_END


static const struct UTextFuncs utf8Funcs =
{
    sizeof(UTextFuncs),
    0, 0, 0,             // Reserved alignment padding
    utf8TextClone,
    utf8TextLength,
    utf8TextAccess,
    utf8TextExtract,
    NULL,                /* replace*/
    NULL,                /* copy   */
    utf8TextMapOffsetToNative,
    utf8TextMapIndexToUTF16,
    utf8TextClose,
    NULL,                // spare 1
    NULL,                // spare 2
    NULL                 // spare 3
};


static const char utext_gEmptyString[] = {0};

U_CAPI UText * U_EXPORT2
utext_openUTF8(UText *ut, const char *s, int64_t length, UErrorCode *status) {
    if(U_FAILURE(*status)) {
        return NULL;
    }
    if(s==NULL && length==0) {
        s = utext_gEmptyString;
    }

    if(s==NULL || length<-1 || length>INT32_MAX) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    ut = utext_setup(ut, sizeof(UTF8Buf) * 2, status);
    if (U_FAILURE(*status)) {
        return ut;
    }

    ut->pFuncs  = &utf8Funcs;
    ut->context = s;
    ut->b       = (int32_t)length;
    ut->c       = (int32_t)length;
    if (ut->c < 0) {
        ut->c = 0;
        ut->providerProperties |= I32_FLAG(UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE);
    }
    ut->p = ut->pExtra;
    ut->q = (char *)ut->pExtra + sizeof(UTF8Buf);
    return ut;

}








//------------------------------------------------------------------------------
//
//     UText implementation wrapper for Replaceable (read/write)
//
//         Use of UText data members:
//            context    pointer to Replaceable.
//            p          pointer to Replaceable if it is owned by the UText.
//
//------------------------------------------------------------------------------



// minimum chunk size for this implementation: 3
// to allow for possible trimming for code point boundaries
enum { REP_TEXT_CHUNK_SIZE=10 };

struct ReplExtra {
    /*
     * Chunk UChars.
     * +1 to simplify filling with surrogate pair at the end.
     */
    UChar s[REP_TEXT_CHUNK_SIZE+1];
};


U_CDECL_BEGIN

static UText * U_CALLCONV
repTextClone(UText *dest, const UText *src, UBool deep, UErrorCode *status) {
    // First do a generic shallow clone.  Does everything needed for the UText struct itself.
    dest = shallowTextClone(dest, src, status);

    // For deep clones, make a copy of the Replaceable.
    //  The copied Replaceable storage is owned by the newly created UText clone.
    //  A non-NULL pointer in UText.p is the signal to the close() function to delete
    //    it.
    //
    if (deep && U_SUCCESS(*status)) {
        const Replaceable *replSrc = (const Replaceable *)src->context;
        dest->context = replSrc->clone();
        dest->providerProperties |= I32_FLAG(UTEXT_PROVIDER_OWNS_TEXT);

        // with deep clone, the copy is writable, even when the source is not.
        dest->providerProperties |= I32_FLAG(UTEXT_PROVIDER_WRITABLE);
    }
    return dest;
}


static void U_CALLCONV
repTextClose(UText *ut) {
    // Most of the work of close is done by the generic UText framework close.
    // All that needs to be done here is delete the Replaceable if the UText
    //  owns it.  This occurs if the UText was created by cloning.
    if (ut->providerProperties & I32_FLAG(UTEXT_PROVIDER_OWNS_TEXT)) {
        Replaceable *rep = (Replaceable *)ut->context;
        delete rep;
        ut->context = NULL;
    }
}


static int64_t U_CALLCONV
repTextLength(UText *ut) {
    const Replaceable *replSrc = (const Replaceable *)ut->context;
    int32_t  len = replSrc->length();
    return len;
}


static UBool U_CALLCONV
repTextAccess(UText *ut, int64_t index, UBool forward) {
    const Replaceable *rep=(const Replaceable *)ut->context;
    int32_t length=rep->length();   // Full length of the input text (bigger than a chunk)

    // clip the requested index to the limits of the text.
    int32_t index32 = pinIndex(index, length);
    U_ASSERT(index<=INT32_MAX);


    /*
     * Compute start/limit boundaries around index, for a segment of text
     * to be extracted.
     * To allow for the possibility that our user gave an index to the trailing
     * half of a surrogate pair, we must request one extra preceding UChar when
     * going in the forward direction.  This will ensure that the buffer has the
     * entire code point at the specified index.
     */
    if(forward) {

        if (index32>=ut->chunkNativeStart && index32<ut->chunkNativeLimit) {
            // Buffer already contains the requested position.
            ut->chunkOffset = (int32_t)(index - ut->chunkNativeStart);
            return TRUE;
        }
        if (index32>=length && ut->chunkNativeLimit==length) {
            // Request for end of string, and buffer already extends up to it.
            // Can't get the data, but don't change the buffer.
            ut->chunkOffset = length - (int32_t)ut->chunkNativeStart;
            return FALSE;
        }

        ut->chunkNativeLimit = index + REP_TEXT_CHUNK_SIZE - 1;
        // Going forward, so we want to have the buffer with stuff at and beyond
        //   the requested index.  The -1 gets us one code point before the
        //   requested index also, to handle the case of the index being on
        //   a trail surrogate of a surrogate pair.
        if(ut->chunkNativeLimit > length) {
            ut->chunkNativeLimit = length;
        }
        // unless buffer ran off end, start is index-1.
        ut->chunkNativeStart = ut->chunkNativeLimit - REP_TEXT_CHUNK_SIZE;
        if(ut->chunkNativeStart < 0) {
            ut->chunkNativeStart = 0;
        }
    } else {
        // Reverse iteration.  Fill buffer with data preceding the requested index.
        if (index32>ut->chunkNativeStart && index32<=ut->chunkNativeLimit) {
            // Requested position already in buffer.
            ut->chunkOffset = index32 - (int32_t)ut->chunkNativeStart;
            return TRUE;
        }
        if (index32==0 && ut->chunkNativeStart==0) {
            // Request for start, buffer already begins at start.
            //  No data, but keep the buffer as is.
            ut->chunkOffset = 0;
            return FALSE;
        }

        // Figure out the bounds of the chunk to extract for reverse iteration.
        // Need to worry about chunk not splitting surrogate pairs, and while still
        // containing the data we need.
        // Fix by requesting a chunk that includes an extra UChar at the end.
        // If this turns out to be a lead surrogate, we can lop it off and still have
        //   the data we wanted.
        ut->chunkNativeStart = index32 + 1 - REP_TEXT_CHUNK_SIZE;
        if (ut->chunkNativeStart < 0) {
            ut->chunkNativeStart = 0;
        }

        ut->chunkNativeLimit = index32 + 1;
        if (ut->chunkNativeLimit > length) {
            ut->chunkNativeLimit = length;
        }
    }

    // Extract the new chunk of text from the Replaceable source.
    ReplExtra *ex = (ReplExtra *)ut->pExtra;
    // UnicodeString with its buffer a writable alias to the chunk buffer
    UnicodeString buffer(ex->s, 0 /*buffer length*/, REP_TEXT_CHUNK_SIZE /*buffer capacity*/);
    rep->extractBetween((int32_t)ut->chunkNativeStart, (int32_t)ut->chunkNativeLimit, buffer);

    ut->chunkContents  = ex->s;
    ut->chunkLength    = (int32_t)(ut->chunkNativeLimit - ut->chunkNativeStart);
    ut->chunkOffset    = (int32_t)(index32 - ut->chunkNativeStart);

    // Surrogate pairs from the input text must not span chunk boundaries.
    // If end of chunk could be the start of a surrogate, trim it off.
    if (ut->chunkNativeLimit < length &&
        U16_IS_LEAD(ex->s[ut->chunkLength-1])) {
            ut->chunkLength--;
            ut->chunkNativeLimit--;
            if (ut->chunkOffset > ut->chunkLength) {
                ut->chunkOffset = ut->chunkLength;
            }
        }

    // if the first UChar in the chunk could be the trailing half of a surrogate pair,
    // trim it off.
    if(ut->chunkNativeStart>0 && U16_IS_TRAIL(ex->s[0])) {
        ++(ut->chunkContents);
        ++(ut->chunkNativeStart);
        --(ut->chunkLength);
        --(ut->chunkOffset);
    }

    // adjust the index/chunkOffset to a code point boundary
    U16_SET_CP_START(ut->chunkContents, 0, ut->chunkOffset);

    // Use fast indexing for get/setNativeIndex()
    ut->nativeIndexingLimit = ut->chunkLength;

    return TRUE;
}



static int32_t U_CALLCONV
repTextExtract(UText *ut,
               int64_t start, int64_t limit,
               UChar *dest, int32_t destCapacity,
               UErrorCode *status) {
    const Replaceable *rep=(const Replaceable *)ut->context;
    int32_t  length=rep->length();

    if(U_FAILURE(*status)) {
        return 0;
    }
    if(destCapacity<0 || (dest==NULL && destCapacity>0)) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
    }
    if(start>limit) {
        *status=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    int32_t  start32 = pinIndex(start, length);
    int32_t  limit32 = pinIndex(limit, length);

    // adjust start, limit if they point to trail half of surrogates
    if (start32<length && U16_IS_TRAIL(rep->charAt(start32)) &&
        U_IS_SUPPLEMENTARY(rep->char32At(start32))){
            start32--;
    }
    if (limit32<length && U16_IS_TRAIL(rep->charAt(limit32)) &&
        U_IS_SUPPLEMENTARY(rep->char32At(limit32))){
            limit32--;
    }

    length=limit32-start32;
    if(length>destCapacity) {
        limit32 = start32 + destCapacity;
    }
    UnicodeString buffer(dest, 0, destCapacity); // writable alias
    rep->extractBetween(start32, limit32, buffer);
    repTextAccess(ut, limit32, TRUE);

    return u_terminateUChars(dest, destCapacity, length, status);
}

static int32_t U_CALLCONV
repTextReplace(UText *ut,
               int64_t start, int64_t limit,
               const UChar *src, int32_t length,
               UErrorCode *status) {
    Replaceable *rep=(Replaceable *)ut->context;
    int32_t oldLength;

    if(U_FAILURE(*status)) {
        return 0;
    }
    if(src==NULL && length!=0) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    oldLength=rep->length(); // will subtract from new length
    if(start>limit ) {
        *status=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    int32_t start32 = pinIndex(start, oldLength);
    int32_t limit32 = pinIndex(limit, oldLength);

    // Snap start & limit to code point boundaries.
    if (start32<oldLength && U16_IS_TRAIL(rep->charAt(start32)) &&
        start32>0 && U16_IS_LEAD(rep->charAt(start32-1)))
    {
            start32--;
    }
    if (limit32<oldLength && U16_IS_LEAD(rep->charAt(limit32-1)) &&
        U16_IS_TRAIL(rep->charAt(limit32)))
    {
            limit32++;
    }

    // Do the actual replace operation using methods of the Replaceable class
    UnicodeString replStr((UBool)(length<0), src, length); // read-only alias
    rep->handleReplaceBetween(start32, limit32, replStr);
    int32_t newLength = rep->length();
    int32_t lengthDelta = newLength - oldLength;

    // Is the UText chunk buffer OK?
    if (ut->chunkNativeLimit > start32) {
        // this replace operation may have impacted the current chunk.
        // invalidate it, which will force a reload on the next access.
        invalidateChunk(ut);
    }

    // set the iteration position to the end of the newly inserted replacement text.
    int32_t newIndexPos = limit32 + lengthDelta;
    repTextAccess(ut, newIndexPos, TRUE);

    return lengthDelta;
}


static void U_CALLCONV
repTextCopy(UText *ut,
                int64_t start, int64_t limit,
                int64_t destIndex,
                UBool move,
                UErrorCode *status)
{
    Replaceable *rep=(Replaceable *)ut->context;
    int32_t length=rep->length();

    if(U_FAILURE(*status)) {
        return;
    }
    if (start>limit || (start<destIndex && destIndex<limit))
    {
        *status=U_INDEX_OUTOFBOUNDS_ERROR;
        return;
    }

    int32_t start32     = pinIndex(start, length);
    int32_t limit32     = pinIndex(limit, length);
    int32_t destIndex32 = pinIndex(destIndex, length);

    // TODO:  snap input parameters to code point boundaries.

    if(move) {
        // move: copy to destIndex, then replace original with nothing
        int32_t segLength=limit32-start32;
        rep->copy(start32, limit32, destIndex32);
        if(destIndex32<start32) {
            start32+=segLength;
            limit32+=segLength;
        }
        rep->handleReplaceBetween(start32, limit32, UnicodeString());
    } else {
        // copy
        rep->copy(start32, limit32, destIndex32);
    }

    // If the change to the text touched the region in the chunk buffer,
    //  invalidate the buffer.
    int32_t firstAffectedIndex = destIndex32;
    if (move && start32<firstAffectedIndex) {
        firstAffectedIndex = start32;
    }
    if (firstAffectedIndex < ut->chunkNativeLimit) {
        // changes may have affected range covered by the chunk
        invalidateChunk(ut);
    }

    // Put iteration position at the newly inserted (moved) block,
    int32_t  nativeIterIndex = destIndex32 + limit32 - start32;
    if (move && destIndex32>start32) {
        // moved a block of text towards the end of the string.
        nativeIterIndex = destIndex32;
    }

    // Set position, reload chunk if needed.
    repTextAccess(ut, nativeIterIndex, TRUE);
}

static const struct UTextFuncs repFuncs =
{
    sizeof(UTextFuncs),
    0, 0, 0,           // Reserved alignment padding
    repTextClone,
    repTextLength,
    repTextAccess,
    repTextExtract,
    repTextReplace,
    repTextCopy,
    NULL,              // MapOffsetToNative,
    NULL,              // MapIndexToUTF16,
    repTextClose,
    NULL,              // spare 1
    NULL,              // spare 2
    NULL               // spare 3
};


U_CAPI UText * U_EXPORT2
utext_openReplaceable(UText *ut, Replaceable *rep, UErrorCode *status)
{
    if(U_FAILURE(*status)) {
        return NULL;
    }
    if(rep==NULL) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    ut = utext_setup(ut, sizeof(ReplExtra), status);
    if(U_FAILURE(*status)) {
        return ut;
    }

    ut->providerProperties = I32_FLAG(UTEXT_PROVIDER_WRITABLE);
    if(rep->hasMetaData()) {
        ut->providerProperties |=I32_FLAG(UTEXT_PROVIDER_HAS_META_DATA);
    }

    ut->pFuncs  = &repFuncs;
    ut->context =  rep;
    return ut;
}

U_CDECL_END








//------------------------------------------------------------------------------
//
//     UText implementation for UnicodeString (read/write)  and
//                    for const UnicodeString (read only)
//             (same implementation, only the flags are different)
//
//         Use of UText data members:
//            context    pointer to UnicodeString
//            p          pointer to UnicodeString IF this UText owns the string
//                       and it must be deleted on close().  NULL otherwise.
//
//------------------------------------------------------------------------------

U_CDECL_BEGIN


static UText * U_CALLCONV
unistrTextClone(UText *dest, const UText *src, UBool deep, UErrorCode *status) {
    // First do a generic shallow clone.  Does everything needed for the UText struct itself.
    dest = shallowTextClone(dest, src, status);

    // For deep clones, make a copy of the UnicodeSring.
    //  The copied UnicodeString storage is owned by the newly created UText clone.
    //  A non-NULL pointer in UText.p is the signal to the close() function to delete
    //    the UText.
    //
    if (deep && U_SUCCESS(*status)) {
        const UnicodeString *srcString = (const UnicodeString *)src->context;
        dest->context = new UnicodeString(*srcString);
        dest->providerProperties |= I32_FLAG(UTEXT_PROVIDER_OWNS_TEXT);

        // with deep clone, the copy is writable, even when the source is not.
        dest->providerProperties |= I32_FLAG(UTEXT_PROVIDER_WRITABLE);
    }
    return dest;
}

static void U_CALLCONV
unistrTextClose(UText *ut) {
    // Most of the work of close is done by the generic UText framework close.
    // All that needs to be done here is delete the UnicodeString if the UText
    //  owns it.  This occurs if the UText was created by cloning.
    if (ut->providerProperties & I32_FLAG(UTEXT_PROVIDER_OWNS_TEXT)) {
        UnicodeString *str = (UnicodeString *)ut->context;
        delete str;
        ut->context = NULL;
    }
}


static int64_t U_CALLCONV
unistrTextLength(UText *t) {
    return ((const UnicodeString *)t->context)->length();
}


static UBool U_CALLCONV
unistrTextAccess(UText *ut, int64_t index, UBool  forward) {
    int32_t length  = ut->chunkLength;
    ut->chunkOffset = pinIndex(index, length);

    // Check whether request is at the start or end
    UBool retVal = (forward && index<length) || (!forward && index>0);
    return retVal;
}



static int32_t U_CALLCONV
unistrTextExtract(UText *t,
                  int64_t start, int64_t limit,
                  UChar *dest, int32_t destCapacity,
                  UErrorCode *pErrorCode) {
    const UnicodeString *us=(const UnicodeString *)t->context;
    int32_t length=us->length();

    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(destCapacity<0 || (dest==NULL && destCapacity>0)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
    }
    if(start<0 || start>limit) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    int32_t start32 = start<length ? us->getChar32Start((int32_t)start) : length;
    int32_t limit32 = limit<length ? us->getChar32Start((int32_t)limit) : length;

    length=limit32-start32;
    if (destCapacity>0 && dest!=NULL) {
        int32_t trimmedLength = length;
        if(trimmedLength>destCapacity) {
            trimmedLength=destCapacity;
        }
        us->extract(start32, trimmedLength, dest);
        t->chunkOffset = start32+trimmedLength;
    } else {
        t->chunkOffset = start32;
    }
    u_terminateUChars(dest, destCapacity, length, pErrorCode);
    return length;
}

static int32_t U_CALLCONV
unistrTextReplace(UText *ut,
                  int64_t start, int64_t limit,
                  const UChar *src, int32_t length,
                  UErrorCode *pErrorCode) {
    UnicodeString *us=(UnicodeString *)ut->context;
    int32_t oldLength;

    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(src==NULL && length!=0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
    }
    if(start>limit) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }
    oldLength=us->length();
    int32_t start32 = pinIndex(start, oldLength);
    int32_t limit32 = pinIndex(limit, oldLength);
    if (start32 < oldLength) {
        start32 = us->getChar32Start(start32);
    }
    if (limit32 < oldLength) {
        limit32 = us->getChar32Start(limit32);
    }

    // replace
    us->replace(start32, limit32-start32, src, length);
    int32_t newLength = us->length();

    // Update the chunk description.
    ut->chunkContents    = us->getBuffer();
    ut->chunkLength      = newLength;
    ut->chunkNativeLimit = newLength;
    ut->nativeIndexingLimit = newLength;

    // Set iteration position to the point just following the newly inserted text.
    int32_t lengthDelta = newLength - oldLength;
    ut->chunkOffset = limit32 + lengthDelta;

    return lengthDelta;
}

static void U_CALLCONV
unistrTextCopy(UText *ut,
               int64_t start, int64_t limit,
               int64_t destIndex,
               UBool move,
               UErrorCode *pErrorCode) {
    UnicodeString *us=(UnicodeString *)ut->context;
    int32_t length=us->length();

    if(U_FAILURE(*pErrorCode)) {
        return;
    }
    int32_t start32 = pinIndex(start, length);
    int32_t limit32 = pinIndex(limit, length);
    int32_t destIndex32 = pinIndex(destIndex, length);

    if( start32>limit32 || (start32<destIndex32 && destIndex32<limit32)) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return;
    }

    if(move) {
        // move: copy to destIndex, then remove original
        int32_t segLength=limit32-start32;
        us->copy(start32, limit32, destIndex32);
        if(destIndex32<start32) {
            start32+=segLength;
        }
        us->remove(start32, segLength);
    } else {
        // copy
        us->copy(start32, limit32, destIndex32);
    }

    // update chunk description, set iteration position.
    ut->chunkContents = us->getBuffer();
    if (move==FALSE) {
        // copy operation, string length grows
        ut->chunkLength += limit32-start32;
        ut->chunkNativeLimit = ut->chunkLength;
        ut->nativeIndexingLimit = ut->chunkLength;
    }

    // Iteration position to end of the newly inserted text.
    ut->chunkOffset = destIndex32+limit32-start32;
    if (move && destIndex32>start32) {
        ut->chunkOffset = destIndex32;
    }

}

static const struct UTextFuncs unistrFuncs =
{
    sizeof(UTextFuncs),
    0, 0, 0,             // Reserved alignment padding
    unistrTextClone,
    unistrTextLength,
    unistrTextAccess,
    unistrTextExtract,
    unistrTextReplace,
    unistrTextCopy,
    NULL,                // MapOffsetToNative,
    NULL,                // MapIndexToUTF16,
    unistrTextClose,
    NULL,                // spare 1
    NULL,                // spare 2
    NULL                 // spare 3
};



U_CDECL_END


U_CAPI UText * U_EXPORT2
utext_openUnicodeString(UText *ut, UnicodeString *s, UErrorCode *status) {
    ut = utext_openConstUnicodeString(ut, s, status);
    if (U_SUCCESS(*status)) {
        ut->providerProperties |= I32_FLAG(UTEXT_PROVIDER_WRITABLE);
    }
    return ut;
}



U_CAPI UText * U_EXPORT2
utext_openConstUnicodeString(UText *ut, const UnicodeString *s, UErrorCode *status) {
    if (U_SUCCESS(*status) && s->isBogus()) {
        // The UnicodeString is bogus, but we still need to detach the UText
        //   from whatever it was hooked to before, if anything.
        utext_openUChars(ut, NULL, 0, status);
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return ut;
    }
    ut = utext_setup(ut, 0, status);
    //    note:  use the standard (writable) function table for UnicodeString.
    //           The flag settings disable writing, so having the functions in
    //           the table is harmless.
    if (U_SUCCESS(*status)) {
        ut->pFuncs              = &unistrFuncs;
        ut->context             = s;
        ut->providerProperties  = I32_FLAG(UTEXT_PROVIDER_STABLE_CHUNKS);
        ut->chunkContents       = s->getBuffer();
        ut->chunkLength         = s->length();
        ut->chunkNativeStart    = 0;
        ut->chunkNativeLimit    = ut->chunkLength;
        ut->nativeIndexingLimit = ut->chunkLength;
    }
    return ut;
}

//------------------------------------------------------------------------------
//
//     UText implementation for const UChar * strings
//
//         Use of UText data members:
//            context    pointer to UnicodeString
//            a          length.  -1 if not yet known.
//
//         TODO:  support 64 bit lengths.
//
//------------------------------------------------------------------------------

U_CDECL_BEGIN


static UText * U_CALLCONV
ucstrTextClone(UText *dest, const UText * src, UBool deep, UErrorCode * status) {
    // First do a generic shallow clone.
    dest = shallowTextClone(dest, src, status);

    // For deep clones, make a copy of the string.
    //  The copied storage is owned by the newly created clone.
    //  A non-NULL pointer in UText.p is the signal to the close() function to delete
    //    it.
    //
    if (deep && U_SUCCESS(*status)) {
        U_ASSERT(utext_nativeLength(dest) < INT32_MAX);
        int32_t  len = (int32_t)utext_nativeLength(dest);

        // The cloned string IS going to be NUL terminated, whether or not the original was.
        const UChar *srcStr = (const UChar *)src->context;
        UChar *copyStr = (UChar *)uprv_malloc((len+1) * sizeof(UChar));
        if (copyStr == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
        } else {
            int64_t i;
            for (i=0; i<len; i++) {
                copyStr[i] = srcStr[i];
            }
            copyStr[len] = 0;
            dest->context = copyStr;
            dest->providerProperties |= I32_FLAG(UTEXT_PROVIDER_OWNS_TEXT);
        }
    }
    return dest;
}


static void U_CALLCONV
ucstrTextClose(UText *ut) {
    // Most of the work of close is done by the generic UText framework close.
    // All that needs to be done here is delete the string if the UText
    //  owns it.  This occurs if the UText was created by cloning.
    if (ut->providerProperties & I32_FLAG(UTEXT_PROVIDER_OWNS_TEXT)) {
        UChar *s = (UChar *)ut->context;
        uprv_free(s);
        ut->context = NULL;
    }
}



static int64_t U_CALLCONV
ucstrTextLength(UText *ut) {
    if (ut->a < 0) {
        // null terminated, we don't yet know the length.  Scan for it.
        //    Access is not convenient for doing this
        //    because the current interation postion can't be changed.
        const UChar  *str = (const UChar *)ut->context;
        for (;;) {
            if (str[ut->chunkNativeLimit] == 0) {
                break;
            }
            ut->chunkNativeLimit++;
        }
        ut->a = ut->chunkNativeLimit;
        ut->chunkLength = (int32_t)ut->chunkNativeLimit;
        ut->nativeIndexingLimit = ut->chunkLength;
        ut->providerProperties &= ~I32_FLAG(UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE);
    }
    return ut->a;
}


static UBool U_CALLCONV
ucstrTextAccess(UText *ut, int64_t index, UBool  forward) {
    const UChar *str   = (const UChar *)ut->context;

    // pin the requested index to the bounds of the string,
    //  and set current iteration position.
    if (index<0) {
        index = 0;
    } else if (index < ut->chunkNativeLimit) {
        // The request data is within the chunk as it is known so far.
        // Put index on a code point boundary.
        U16_SET_CP_START(str, 0, index);
    } else if (ut->a >= 0) {
        // We know the length of this string, and the user is requesting something
        // at or beyond the length.  Pin the requested index to the length.
        index = ut->a;
    } else {
        // Null terminated string, length not yet known, and the requested index
        //  is beyond where we have scanned so far.
        //  Scan to 32 UChars beyond the requested index.  The strategy here is
        //  to avoid fully scanning a long string when the caller only wants to
        //  see a few characters at its beginning.
        int32_t scanLimit = (int32_t)index + 32;
        if ((index + 32)>INT32_MAX || (index + 32)<0 ) {   // note: int64 expression
            scanLimit = INT32_MAX;
        }

        int32_t chunkLimit = (int32_t)ut->chunkNativeLimit;
        for (; chunkLimit<scanLimit; chunkLimit++) {
            if (str[chunkLimit] == 0) {
                // We found the end of the string.  Remember it, pin the requested index to it,
                //  and bail out of here.
                ut->a = chunkLimit;
                ut->chunkLength = chunkLimit;
                ut->nativeIndexingLimit = chunkLimit;
                if (index >= chunkLimit) {
                    index = chunkLimit;
                } else {
                    U16_SET_CP_START(str, 0, index);
                }

                ut->chunkNativeLimit = chunkLimit;
                ut->providerProperties &= ~I32_FLAG(UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE);
                goto breakout;
            }
        }
        // We scanned through the next batch of UChars without finding the end.
        U16_SET_CP_START(str, 0, index);
        if (chunkLimit == INT32_MAX) {
            // Scanned to the limit of a 32 bit length.
            // Forceably trim the overlength string back so length fits in int32
            //  TODO:  add support for 64 bit strings.
            ut->a = chunkLimit;
            ut->chunkLength = chunkLimit;
            ut->nativeIndexingLimit = chunkLimit;
            if (index > chunkLimit) {
                index = chunkLimit;
            }
            ut->chunkNativeLimit = chunkLimit;
            ut->providerProperties &= ~I32_FLAG(UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE);
        } else {
            // The endpoint of a chunk must not be left in the middle of a surrogate pair.
            // If the current end is on a lead surrogate, back the end up by one.
            // It doesn't matter if the end char happens to be an unpaired surrogate,
            //    and it's simpler not to worry about it.
            if (U16_IS_LEAD(str[chunkLimit-1])) {
                --chunkLimit;
            }
            // Null-terminated chunk with end still unknown.
            // Update the chunk length to reflect what has been scanned thus far.
            // That the full length is still unknown is (still) flagged by
            //    ut->a being < 0.
            ut->chunkNativeLimit = chunkLimit;
            ut->nativeIndexingLimit = chunkLimit;
            ut->chunkLength = chunkLimit;
        }

    }
breakout:
    U_ASSERT(index<=INT32_MAX);
    ut->chunkOffset = (int32_t)index;

    // Check whether request is at the start or end
    UBool retVal = (forward && index<ut->chunkNativeLimit) || (!forward && index>0);
    return retVal;
}



static int32_t U_CALLCONV
ucstrTextExtract(UText *ut,
                  int64_t start, int64_t limit,
                  UChar *dest, int32_t destCapacity,
                  UErrorCode *pErrorCode)
{
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(destCapacity<0 || (dest==NULL && destCapacity>0) || start>limit) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    //const UChar *s=(const UChar *)ut->context;
    int32_t si, di;

    int32_t start32;
    int32_t limit32;

    // Access the start.  Does two things we need:
    //   Pins 'start' to the length of the string, if it came in out-of-bounds.
    //   Snaps 'start' to the beginning of a code point.
    ucstrTextAccess(ut, start, TRUE);
    const UChar *s=ut->chunkContents;
    start32 = ut->chunkOffset;

    int32_t strLength=(int32_t)ut->a;
    if (strLength >= 0) {
        limit32 = pinIndex(limit, strLength);
    } else {
        limit32 = pinIndex(limit, INT32_MAX);
    }
    di = 0;
    for (si=start32; si<limit32; si++) {
        if (strLength<0 && s[si]==0) {
            // Just hit the end of a null-terminated string.
            ut->a = si;               // set string length for this UText
            ut->chunkNativeLimit    = si;
            ut->chunkLength         = si;
            ut->nativeIndexingLimit = si;
            strLength               = si;
            limit32                 = si;
            break;
        }
        U_ASSERT(di>=0); /* to ensure di never exceeds INT32_MAX, which must not happen logically */
        if (di<destCapacity) {
            // only store if there is space.
            dest[di] = s[si];
        } else {
            if (strLength>=0) {
                // We have filled the destination buffer, and the string length is known.
                //  Cut the loop short.  There is no need to scan string termination.
                di = limit32 - start32;
                si = limit32;
                break;
            }
        }
        di++;
    }

    // If the limit index points to a lead surrogate of a pair,
    //   add the corresponding trail surrogate to the destination.
    if (si>0 && U16_IS_LEAD(s[si-1]) &&
            ((si<strLength || strLength<0)  && U16_IS_TRAIL(s[si])))
    {
        if (di<destCapacity) {
            // store only if there is space in the output buffer.
            dest[di++] = s[si];
        }
        si++;
    }

    // Put iteration position at the point just following the extracted text
    if (si <= ut->chunkNativeLimit) {
        ut->chunkOffset = si;
    } else {
        ucstrTextAccess(ut, si, TRUE);
    }

    // Add a terminating NUL if space in the buffer permits,
    // and set the error status as required.
    u_terminateUChars(dest, destCapacity, di, pErrorCode);
    return di;
}

static const struct UTextFuncs ucstrFuncs =
{
    sizeof(UTextFuncs),
    0, 0, 0,           // Reserved alignment padding
    ucstrTextClone,
    ucstrTextLength,
    ucstrTextAccess,
    ucstrTextExtract,
    NULL,              // Replace
    NULL,              // Copy
    NULL,              // MapOffsetToNative,
    NULL,              // MapIndexToUTF16,
    ucstrTextClose,
    NULL,              // spare 1
    NULL,              // spare 2
    NULL,              // spare 3
};

U_CDECL_END

static const UChar gEmptyUString[] = {0};

U_CAPI UText * U_EXPORT2
utext_openUChars(UText *ut, const UChar *s, int64_t length, UErrorCode *status) {
    if (U_FAILURE(*status)) {
        return NULL;
    }
    if(s==NULL && length==0) {
        s = gEmptyUString;
    }
    if (s==NULL || length < -1 || length>INT32_MAX) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    ut = utext_setup(ut, 0, status);
    if (U_SUCCESS(*status)) {
        ut->pFuncs               = &ucstrFuncs;
        ut->context              = s;
        ut->providerProperties   = I32_FLAG(UTEXT_PROVIDER_STABLE_CHUNKS);
        if (length==-1) {
            ut->providerProperties |= I32_FLAG(UTEXT_PROVIDER_LENGTH_IS_EXPENSIVE);
        }
        ut->a                    = length;
        ut->chunkContents        = s;
        ut->chunkNativeStart     = 0;
        ut->chunkNativeLimit     = length>=0? length : 0;
        ut->chunkLength          = (int32_t)ut->chunkNativeLimit;
        ut->chunkOffset          = 0;
        ut->nativeIndexingLimit  = ut->chunkLength;
    }
    return ut;
}


//------------------------------------------------------------------------------
//
//     UText implementation for text from ICU CharacterIterators
//
//         Use of UText data members:
//            context    pointer to the CharacterIterator
//            a          length of the full text.
//            p          pointer to  buffer 1
//            b          start index of local buffer 1 contents
//            q          pointer to buffer 2
//            c          start index of local buffer 2 contents
//            r          pointer to the character iterator if the UText owns it.
//                       Null otherwise.
//
//------------------------------------------------------------------------------
#define CIBufSize 16

U_CDECL_BEGIN
static void U_CALLCONV
charIterTextClose(UText *ut) {
    // Most of the work of close is done by the generic UText framework close.
    // All that needs to be done here is delete the CharacterIterator if the UText
    //  owns it.  This occurs if the UText was created by cloning.
    CharacterIterator *ci = (CharacterIterator *)ut->r;
    delete ci;
    ut->r = NULL;
}

static int64_t U_CALLCONV
charIterTextLength(UText *ut) {
    return (int32_t)ut->a;
}

static UBool U_CALLCONV
charIterTextAccess(UText *ut, int64_t index, UBool  forward) {
    CharacterIterator *ci   = (CharacterIterator *)ut->context;

    int32_t clippedIndex = (int32_t)index;
    if (clippedIndex<0) {
        clippedIndex=0;
    } else if (clippedIndex>=ut->a) {
        clippedIndex=(int32_t)ut->a;
    }
    int32_t neededIndex = clippedIndex;
    if (!forward && neededIndex>0) {
        // reverse iteration, want the position just before what was asked for.
        neededIndex--;
    } else if (forward && neededIndex==ut->a && neededIndex>0) {
        // Forward iteration, don't ask for something past the end of the text.
        neededIndex--;
    }

    // Find the native index of the start of the buffer containing what we want.
    neededIndex -= neededIndex % CIBufSize;

    UChar *buf = NULL;
    UBool  needChunkSetup = TRUE;
    int    i;
    if (ut->chunkNativeStart == neededIndex) {
        // The buffer we want is already the current chunk.
        needChunkSetup = FALSE;
    } else if (ut->b == neededIndex) {
        // The first buffer (buffer p) has what we need.
        buf = (UChar *)ut->p;
    } else if (ut->c == neededIndex) {
        // The second buffer (buffer q) has what we need.
        buf = (UChar *)ut->q;
    } else {
        // Neither buffer already has what we need.
        // Load new data from the character iterator.
        // Use the buf that is not the current buffer.
        buf = (UChar *)ut->p;
        if (ut->p == ut->chunkContents) {
            buf = (UChar *)ut->q;
        }
        ci->setIndex(neededIndex);
        for (i=0; i<CIBufSize; i++) {
            buf[i] = ci->nextPostInc();
            if (i+neededIndex > ut->a) {
                break;
            }
        }
    }

    // We have a buffer with the data we need.
    // Set it up as the current chunk, if it wasn't already.
    if (needChunkSetup) {
        ut->chunkContents = buf;
        ut->chunkLength   = CIBufSize;
        ut->chunkNativeStart = neededIndex;
        ut->chunkNativeLimit = neededIndex + CIBufSize;
        if (ut->chunkNativeLimit > ut->a) {
            ut->chunkNativeLimit = ut->a;
            ut->chunkLength  = (int32_t)(ut->chunkNativeLimit)-(int32_t)(ut->chunkNativeStart);
        }
        ut->nativeIndexingLimit = ut->chunkLength;
        U_ASSERT(ut->chunkOffset>=0 && ut->chunkOffset<=CIBufSize);
    }
    ut->chunkOffset = clippedIndex - (int32_t)ut->chunkNativeStart;
    UBool success = (forward? ut->chunkOffset<ut->chunkLength : ut->chunkOffset>0);
    return success;
}

static UText * U_CALLCONV
charIterTextClone(UText *dest, const UText *src, UBool deep, UErrorCode * status) {
    if (U_FAILURE(*status)) {
        return NULL;
    }

    if (deep) {
        // There is no CharacterIterator API for cloning the underlying text storage.
        *status = U_UNSUPPORTED_ERROR;
        return NULL;
    } else {
        CharacterIterator *srcCI =(CharacterIterator *)src->context;
        srcCI = srcCI->clone();
        dest = utext_openCharacterIterator(dest, srcCI, status);
        if (U_FAILURE(*status)) {
            return dest;
        }
        // cast off const on getNativeIndex.
        //   For CharacterIterator based UTexts, this is safe, the operation is const.
        int64_t  ix = utext_getNativeIndex((UText *)src);
        utext_setNativeIndex(dest, ix);
        dest->r = srcCI;    // flags that this UText owns the CharacterIterator
    }
    return dest;
}

static int32_t U_CALLCONV
charIterTextExtract(UText *ut,
                  int64_t start, int64_t limit,
                  UChar *dest, int32_t destCapacity,
                  UErrorCode *status)
{
    if(U_FAILURE(*status)) {
        return 0;
    }
    if(destCapacity<0 || (dest==NULL && destCapacity>0) || start>limit) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    int32_t  length  = (int32_t)ut->a;
    int32_t  start32 = pinIndex(start, length);
    int32_t  limit32 = pinIndex(limit, length);
    int32_t  desti   = 0;
    int32_t  srci;
    int32_t  copyLimit;

    CharacterIterator *ci = (CharacterIterator *)ut->context;
    ci->setIndex32(start32);   // Moves ix to lead of surrogate pair, if needed.
    srci = ci->getIndex();
    copyLimit = srci;
    while (srci<limit32) {
        UChar32 c = ci->next32PostInc();
        int32_t  len = U16_LENGTH(c);
        U_ASSERT(desti+len>0); /* to ensure desti+len never exceeds MAX_INT32, which must not happen logically */
        if (desti+len <= destCapacity) {
            U16_APPEND_UNSAFE(dest, desti, c);
            copyLimit = srci+len;
        } else {
            desti += len;
            *status = U_BUFFER_OVERFLOW_ERROR;
        }
        srci += len;
    }

    charIterTextAccess(ut, copyLimit, TRUE);

    u_terminateUChars(dest, destCapacity, desti, status);
    return desti;
}

static const struct UTextFuncs charIterFuncs =
{
    sizeof(UTextFuncs),
    0, 0, 0,             // Reserved alignment padding
    charIterTextClone,
    charIterTextLength,
    charIterTextAccess,
    charIterTextExtract,
    NULL,                // Replace
    NULL,                // Copy
    NULL,                // MapOffsetToNative,
    NULL,                // MapIndexToUTF16,
    charIterTextClose,
    NULL,                // spare 1
    NULL,                // spare 2
    NULL                 // spare 3
};
U_CDECL_END


U_CAPI UText * U_EXPORT2
utext_openCharacterIterator(UText *ut, CharacterIterator *ci, UErrorCode *status) {
    if (U_FAILURE(*status)) {
        return NULL;
    }

    if (ci->startIndex() > 0) {
        // No support for CharacterIterators that do not start indexing from zero.
        *status = U_UNSUPPORTED_ERROR;
        return NULL;
    }

    // Extra space in UText for 2 buffers of CIBufSize UChars each.
    int32_t  extraSpace = 2 * CIBufSize * sizeof(UChar);
    ut = utext_setup(ut, extraSpace, status);
    if (U_SUCCESS(*status)) {
        ut->pFuncs                = &charIterFuncs;
        ut->context              = ci;
        ut->providerProperties   = 0;
        ut->a                    = ci->endIndex();        // Length of text
        ut->p                    = ut->pExtra;            // First buffer
        ut->b                    = -1;                    // Native index of first buffer contents
        ut->q                    = (UChar*)ut->pExtra+CIBufSize;  // Second buffer
        ut->c                    = -1;                    // Native index of second buffer contents

        // Initialize current chunk contents to be empty.
        //   First access will fault something in.
        //   Note:  The initial nativeStart and chunkOffset must sum to zero
        //          so that getNativeIndex() will correctly compute to zero
        //          if no call to Access() has ever been made.  They can't be both
        //          zero without Access() thinking that the chunk is valid.
        ut->chunkContents        = (UChar *)ut->p;
        ut->chunkNativeStart     = -1;
        ut->chunkOffset          = 1;
        ut->chunkNativeLimit     = 0;
        ut->chunkLength          = 0;
        ut->nativeIndexingLimit  = ut->chunkOffset;  // enables native indexing
    }
    return ut;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  utf_impl.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 1999sep13
*   created by: Markus W. Scherer
*
*   This file provides implementation functions for macros in the utfXX.h
*   that would otherwise be too long as macros.
*/

/* set import/export definitions */
#ifndef U_UTF8_IMPL
#   define U_UTF8_IMPL
#endif






/*
 * Table of the number of utf8 trail bytes, indexed by the lead byte.
 * Used by the deprecated macro UTF8_COUNT_TRAIL_BYTES, defined in utf_old.h
 *
 * The current macro, U8_COUNT_TRAIL_BYTES, does _not_ use this table.
 *
 * Note that this table cannot be removed, even if UTF8_COUNT_TRAIL_BYTES were
 * changed to no longer use it. References to the table from expansions of UTF8_COUNT_TRAIL_BYTES
 * may exist in old client code that must continue to run with newer icu library versions.
 *
 * This table could be replaced on many machines by
 * a few lines of assembler code using an
 * "index of first 0-bit from msb" instruction and
 * one or two more integer instructions.
 *
 * For example, on an i386, do something like
 * - MOV AL, leadByte
 * - NOT AL         (8-bit, leave b15..b8==0..0, reverse only b7..b0)
 * - MOV AH, 0
 * - BSR BX, AX     (16-bit)
 * - MOV AX, 6      (result)
 * - JZ finish      (ZF==1 if leadByte==0xff)
 * - SUB AX, BX (result)
 * -finish:
 * (BSR: Bit Scan Reverse, scans for a 1-bit, starting from the MSB)
 */
extern "C" U_EXPORT const uint8_t
utf8_countTrailBytes[256]={
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    // illegal C0 & C1
    // 2-byte lead bytes C2..DF
    0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,

    // 3-byte lead bytes E0..EF
    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    // 4-byte lead bytes F0..F4
    // illegal F5..FF
    3, 3, 3, 3, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

static const UChar32
utf8_errorValue[6]={
    // Same values as UTF8_ERROR_VALUE_1, UTF8_ERROR_VALUE_2, UTF_ERROR_VALUE,
    // but without relying on the obsolete unicode/utf_old.h.
    0x15, 0x9f, 0xffff,
    0x10ffff
};

static UChar32
errorValue(int32_t count, int8_t strict) {
    if(strict>=0) {
        return utf8_errorValue[count];
    } else if(strict==-3) {
        return 0xfffd;
    } else {
        return U_SENTINEL;
    }
}

/*
 * Handle the non-inline part of the U8_NEXT() and U8_NEXT_FFFD() macros
 * and their obsolete sibling UTF8_NEXT_CHAR_SAFE().
 *
 * U8_NEXT() supports NUL-terminated strings indicated via length<0.
 *
 * The "strict" parameter controls the error behavior:
 * <0  "Safe" behavior of U8_NEXT():
 *     -1: All illegal byte sequences yield U_SENTINEL=-1.
 *     -2: Same as -1, except for lenient treatment of surrogate code points as legal.
 *         Some implementations use this for roundtripping of
 *         Unicode 16-bit strings that are not well-formed UTF-16, that is, they
 *         contain unpaired surrogates.
 *     -3: All illegal byte sequences yield U+FFFD.
 *  0  Obsolete "safe" behavior of UTF8_NEXT_CHAR_SAFE(..., FALSE):
 *     All illegal byte sequences yield a positive code point such that this
 *     result code point would be encoded with the same number of bytes as
 *     the illegal sequence.
 * >0  Obsolete "strict" behavior of UTF8_NEXT_CHAR_SAFE(..., TRUE):
 *     Same as the obsolete "safe" behavior, but non-characters are also treated
 *     like illegal sequences.
 *
 * Note that a UBool is the same as an int8_t.
 */
U_CAPI UChar32 U_EXPORT2
utf8_nextCharSafeBody(const uint8_t *s, int32_t *pi, int32_t length, UChar32 c, UBool strict) {
    // *pi is one after byte c.
    int32_t i=*pi;
    // length can be negative for NUL-terminated strings: Read and validate one byte at a time.
    if(i==length || c>0xf4) {
        // end of string, or not a lead byte
    } else if(c>=0xf0) {
        // Test for 4-byte sequences first because
        // U8_NEXT() handles shorter valid sequences inline.
        uint8_t t1=s[i], t2, t3;
        c&=7;
        if(U8_IS_VALID_LEAD4_AND_T1(c, t1) &&
                ++i!=length && (t2=s[i]-0x80)<=0x3f &&
                ++i!=length && (t3=s[i]-0x80)<=0x3f) {
            ++i;
            c=(c<<18)|((t1&0x3f)<<12)|(t2<<6)|t3;
            // strict: forbid non-characters like U+fffe
            if(strict<=0 || !U_IS_UNICODE_NONCHAR(c)) {
                *pi=i;
                return c;
            }
        }
    } else if(c>=0xe0) {
        c&=0xf;
        if(strict!=-2) {
            uint8_t t1=s[i], t2;
            if(U8_IS_VALID_LEAD3_AND_T1(c, t1) &&
                    ++i!=length && (t2=s[i]-0x80)<=0x3f) {
                ++i;
                c=(c<<12)|((t1&0x3f)<<6)|t2;
                // strict: forbid non-characters like U+fffe
                if(strict<=0 || !U_IS_UNICODE_NONCHAR(c)) {
                    *pi=i;
                    return c;
                }
            }
        } else {
            // strict=-2 -> lenient: allow surrogates
            uint8_t t1=s[i]-0x80, t2;
            if(t1<=0x3f && (c>0 || t1>=0x20) &&
                    ++i!=length && (t2=s[i]-0x80)<=0x3f) {
                *pi=i+1;
                return (c<<12)|(t1<<6)|t2;
            }
        }
    } else if(c>=0xc2) {
        uint8_t t1=s[i]-0x80;
        if(t1<=0x3f) {
            *pi=i+1;
            return ((c-0xc0)<<6)|t1;
        }
    }  // else 0x80<=c<0xc2 is not a lead byte

    /* error handling */
    c=errorValue(i-*pi, strict);
    *pi=i;
    return c;
}

U_CAPI int32_t U_EXPORT2
utf8_appendCharSafeBody(uint8_t *s, int32_t i, int32_t length, UChar32 c, UBool *pIsError) {
    if((uint32_t)(c)<=0x7ff) {
        if((i)+1<(length)) {
            (s)[(i)++]=(uint8_t)(((c)>>6)|0xc0);
            (s)[(i)++]=(uint8_t)(((c)&0x3f)|0x80);
            return i;
        }
    } else if((uint32_t)(c)<=0xffff) {
        /* Starting with Unicode 3.2, surrogate code points must not be encoded in UTF-8. */
        if((i)+2<(length) && !U_IS_SURROGATE(c)) {
            (s)[(i)++]=(uint8_t)(((c)>>12)|0xe0);
            (s)[(i)++]=(uint8_t)((((c)>>6)&0x3f)|0x80);
            (s)[(i)++]=(uint8_t)(((c)&0x3f)|0x80);
            return i;
        }
    } else if((uint32_t)(c)<=0x10ffff) {
        if((i)+3<(length)) {
            (s)[(i)++]=(uint8_t)(((c)>>18)|0xf0);
            (s)[(i)++]=(uint8_t)((((c)>>12)&0x3f)|0x80);
            (s)[(i)++]=(uint8_t)((((c)>>6)&0x3f)|0x80);
            (s)[(i)++]=(uint8_t)(((c)&0x3f)|0x80);
            return i;
        }
    }
    /* c>0x10ffff or not enough space, write an error value */
    if(pIsError!=NULL) {
        *pIsError=TRUE;
    } else {
        length-=i;
        if(length>0) {
            int32_t offset;
            if(length>3) {
                length=3;
            }
            s+=i;
            offset=0;
            c=utf8_errorValue[length-1];
            U8_APPEND_UNSAFE(s, offset, c);
            i=i+offset;
        }
    }
    return i;
}

U_CAPI UChar32 U_EXPORT2
utf8_prevCharSafeBody(const uint8_t *s, int32_t start, int32_t *pi, UChar32 c, UBool strict) {
    // *pi is the index of byte c.
    int32_t i=*pi;
    if(U8_IS_TRAIL(c) && i>start) {
        uint8_t b1=s[--i];
        if(U8_IS_LEAD(b1)) {
            if(b1<0xe0) {
                *pi=i;
                return ((b1-0xc0)<<6)|(c&0x3f);
            } else if(b1<0xf0 ? U8_IS_VALID_LEAD3_AND_T1(b1, c) : U8_IS_VALID_LEAD4_AND_T1(b1, c)) {
                // Truncated 3- or 4-byte sequence.
                *pi=i;
                return errorValue(1, strict);
            }
        } else if(U8_IS_TRAIL(b1) && i>start) {
            // Extract the value bits from the last trail byte.
            c&=0x3f;
            uint8_t b2=s[--i];
            if(0xe0<=b2 && b2<=0xf4) {
                if(b2<0xf0) {
                    b2&=0xf;
                    if(strict!=-2) {
                        if(U8_IS_VALID_LEAD3_AND_T1(b2, b1)) {
                            *pi=i;
                            c=(b2<<12)|((b1&0x3f)<<6)|c;
                            if(strict<=0 || !U_IS_UNICODE_NONCHAR(c)) {
                                return c;
                            } else {
                                // strict: forbid non-characters like U+fffe
                                return errorValue(2, strict);
                            }
                        }
                    } else {
                        // strict=-2 -> lenient: allow surrogates
                        b1-=0x80;
                        if((b2>0 || b1>=0x20)) {
                            *pi=i;
                            return (b2<<12)|(b1<<6)|c;
                        }
                    }
                } else if(U8_IS_VALID_LEAD4_AND_T1(b2, b1)) {
                    // Truncated 4-byte sequence.
                    *pi=i;
                    return errorValue(2, strict);
                }
            } else if(U8_IS_TRAIL(b2) && i>start) {
                uint8_t b3=s[--i];
                if(0xf0<=b3 && b3<=0xf4) {
                    b3&=7;
                    if(U8_IS_VALID_LEAD4_AND_T1(b3, b2)) {
                        *pi=i;
                        c=(b3<<18)|((b2&0x3f)<<12)|((b1&0x3f)<<6)|c;
                        if(strict<=0 || !U_IS_UNICODE_NONCHAR(c)) {
                            return c;
                        } else {
                            // strict: forbid non-characters like U+fffe
                            return errorValue(3, strict);
                        }
                    }
                }
            }
        }
    }
    return errorValue(0, strict);
}

U_CAPI int32_t U_EXPORT2
utf8_back1SafeBody(const uint8_t *s, int32_t start, int32_t i) {
    // Same as utf8_prevCharSafeBody(..., strict=-1) minus assembling code points.
    int32_t orig_i=i;
    uint8_t c=s[i];
    if(U8_IS_TRAIL(c) && i>start) {
        uint8_t b1=s[--i];
        if(U8_IS_LEAD(b1)) {
            if(b1<0xe0 ||
                    (b1<0xf0 ? U8_IS_VALID_LEAD3_AND_T1(b1, c) : U8_IS_VALID_LEAD4_AND_T1(b1, c))) {
                return i;
            }
        } else if(U8_IS_TRAIL(b1) && i>start) {
            uint8_t b2=s[--i];
            if(0xe0<=b2 && b2<=0xf4) {
                if(b2<0xf0 ? U8_IS_VALID_LEAD3_AND_T1(b2, b1) : U8_IS_VALID_LEAD4_AND_T1(b2, b1)) {
                    return i;
                }
            } else if(U8_IS_TRAIL(b2) && i>start) {
                uint8_t b3=s[--i];
                if(0xf0<=b3 && b3<=0xf4 && U8_IS_VALID_LEAD4_AND_T1(b3, b2)) {
                    return i;
                }
            }
        }
    }
    return orig_i;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (c) 2001-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*   Date        Name        Description
*   11/19/2001  aliu        Creation.
**********************************************************************
*/






// Define UChar constants using hex for EBCDIC compatibility

static const UChar u_util_BACKSLASH  = 0x005C; /*\*/
static const UChar u_util_UPPER_U    = 0x0055; /*U*/
static const UChar u_util_LOWER_U    = 0x0075; /*u*/
static const UChar u_util_APOSTROPHE = 0x0027; // '\''
static const UChar u_util_SPACE      = 0x0020; // ' '

// "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
static const UChar u_util_DIGITS[] = {
    48,49,50,51,52,53,54,55,56,57,
    65,66,67,68,69,70,71,72,73,74,
    75,76,77,78,79,80,81,82,83,84,
    85,86,87,88,89,90
};

U_NAMESPACE_BEGIN

UnicodeString& ICU_Utility::appendNumber(UnicodeString& result, int32_t n,
                                     int32_t radix, int32_t minDigits) {
    if (radix < 2 || radix > 36) {
        // Bogus radix
        return result.append((UChar)63/*?*/);
    }
    // Handle negatives
    if (n < 0) {
        n = -n;
        result.append((UChar)45/*-*/);
    }
    // First determine the number of digits
    int32_t nn = n;
    int32_t r = 1;
    while (nn >= radix) {
        nn /= radix;
        r *= radix;
        --minDigits;
    }
    // Now generate the digits
    while (--minDigits > 0) {
        result.append(u_util_DIGITS[0]);
    }
    while (r > 0) {
        int32_t digit = n / r;
        result.append(u_util_DIGITS[digit]);
        n -= digit * r;
        r /= radix;
    }
    return result;
}

/**
 * Return true if the character is NOT printable ASCII.
 */
UBool ICU_Utility::isUnprintable(UChar32 c) {
    return !(c >= 0x20 && c <= 0x7E);
}

/**
 * Escape unprintable characters using \uxxxx notation for U+0000 to
 * U+FFFF and \Uxxxxxxxx for U+10000 and above.  If the character is
 * printable ASCII, then do nothing and return FALSE.  Otherwise,
 * append the escaped notation and return TRUE.
 */
UBool ICU_Utility::escapeUnprintable(UnicodeString& result, UChar32 c) {
    if (isUnprintable(c)) {
        result.append(u_util_BACKSLASH);
        if (c & ~0xFFFF) {
            result.append(u_util_UPPER_U);
            result.append(u_util_DIGITS[0xF&(c>>28)]);
            result.append(u_util_DIGITS[0xF&(c>>24)]);
            result.append(u_util_DIGITS[0xF&(c>>20)]);
            result.append(u_util_DIGITS[0xF&(c>>16)]);
        } else {
            result.append(u_util_LOWER_U);
        }
        result.append(u_util_DIGITS[0xF&(c>>12)]);
        result.append(u_util_DIGITS[0xF&(c>>8)]);
        result.append(u_util_DIGITS[0xF&(c>>4)]);
        result.append(u_util_DIGITS[0xF&c]);
        return TRUE;
    }
    return FALSE;
}

/**
 * Returns the index of a character, ignoring quoted text.
 * For example, in the string "abc'hide'h", the 'h' in "hide" will not be
 * found by a search for 'h'.
 */
// FOR FUTURE USE.  DISABLE FOR NOW for coverage reasons.
/*
int32_t ICU_Utility::quotedIndexOf(const UnicodeString& text,
                               int32_t start, int32_t limit,
                               UChar charToFind) {
    for (int32_t i=start; i<limit; ++i) {
        UChar c = text.charAt(i);
        if (c == u_util_BACKSLASH) {
            ++i;
        } else if (c == u_util_APOSTROPHE) {
            while (++i < limit
                   && text.charAt(i) != u_util_APOSTROPHE) {}
        } else if (c == charToFind) {
            return i;
        }
    }
    return -1;
}
*/

/**
 * Skip over a sequence of zero or more white space characters at pos.
 * @param advance if true, advance pos to the first non-white-space
 * character at or after pos, or str.length(), if there is none.
 * Otherwise leave pos unchanged.
 * @return the index of the first non-white-space character at or
 * after pos, or str.length(), if there is none.
 */
int32_t ICU_Utility::skipWhitespace(const UnicodeString& str, int32_t& pos,
                                    UBool advance) {
    int32_t p = pos;
    const UChar* s = str.getBuffer();
    p = (int32_t)(PatternProps::skipWhiteSpace(s + p, str.length() - p) - s);
    if (advance) {
        pos = p;
    }
    return p;
}

/**
 * Skip over Pattern_White_Space in a Replaceable.
 * Skipping may be done in the forward or
 * reverse direction.  In either case, the leftmost index will be
 * inclusive, and the rightmost index will be exclusive.  That is,
 * given a range defined as [start, limit), the call
 * skipWhitespace(text, start, limit) will advance start past leading
 * whitespace, whereas the call skipWhitespace(text, limit, start),
 * will back up limit past trailing whitespace.
 * @param text the text to be analyzed
 * @param pos either the start or limit of a range of 'text', to skip
 * leading or trailing whitespace, respectively
 * @param stop either the limit or start of a range of 'text', to skip
 * leading or trailing whitespace, respectively
 * @return the new start or limit, depending on what was passed in to
 * 'pos'
 */
//?FOR FUTURE USE.  DISABLE FOR NOW for coverage reasons.
//?int32_t ICU_Utility::skipWhitespace(const Replaceable& text,
//?                                    int32_t pos, int32_t stop) {
//?    UChar32 c;
//?    UBool isForward = (stop >= pos);
//?
//?    if (!isForward) {
//?        --pos; // pos is a limit, so back up by one
//?    }
//?
//?    while (pos != stop &&
//?           PatternProps::isWhiteSpace(c = text.char32At(pos))) {
//?        if (isForward) {
//?            pos += U16_LENGTH(c);
//?        } else {
//?            pos -= U16_LENGTH(c);
//?        }
//?    }
//?
//?    if (!isForward) {
//?        ++pos; // make pos back into a limit
//?    }
//?
//?    return pos;
//?}

/**
 * Parse a single non-whitespace character 'ch', optionally
 * preceded by whitespace.
 * @param id the string to be parsed
 * @param pos INPUT-OUTPUT parameter.  On input, pos[0] is the
 * offset of the first character to be parsed.  On output, pos[0]
 * is the index after the last parsed character.  If the parse
 * fails, pos[0] will be unchanged.
 * @param ch the non-whitespace character to be parsed.
 * @return true if 'ch' is seen preceded by zero or more
 * whitespace characters.
 */
UBool ICU_Utility::parseChar(const UnicodeString& id, int32_t& pos, UChar ch) {
    int32_t start = pos;
    skipWhitespace(id, pos, TRUE);
    if (pos == id.length() ||
        id.charAt(pos) != ch) {
        pos = start;
        return FALSE;
    }
    ++pos;
    return TRUE;
}

/**
 * Parse a pattern string within the given Replaceable and a parsing
 * pattern.  Characters are matched literally and case-sensitively
 * except for the following special characters:
 *
 * ~  zero or more Pattern_White_Space chars
 *
 * If end of pattern is reached with all matches along the way,
 * pos is advanced to the first unparsed index and returned.
 * Otherwise -1 is returned.
 * @param pat pattern that controls parsing
 * @param text text to be parsed, starting at index
 * @param index offset to first character to parse
 * @param limit offset after last character to parse
 * @return index after last parsed character, or -1 on parse failure.
 */
int32_t ICU_Utility::parsePattern(const UnicodeString& pat,
                                  const Replaceable& text,
                                  int32_t index,
                                  int32_t limit) {
    int32_t ipat = 0;

    // empty pattern matches immediately
    if (ipat == pat.length()) {
        return index;
    }

    UChar32 cpat = pat.char32At(ipat);

    while (index < limit) {
        UChar32 c = text.char32At(index);

        // parse \s*
        if (cpat == 126 /*~*/) {
            if (PatternProps::isWhiteSpace(c)) {
                index += U16_LENGTH(c);
                continue;
            } else {
                if (++ipat == pat.length()) {
                    return index; // success; c unparsed
                }
                // fall thru; process c again with next cpat
            }
        }

        // parse literal
        else if (c == cpat) {
            index += U16_LENGTH(c);
            ipat += U16_LENGTH(cpat);
            if (ipat == pat.length()) {
                return index; // success; c parsed
            }
            // fall thru; get next cpat
        }

        // match failure of literal
        else {
            return -1;
        }

        cpat = pat.char32At(ipat);
    }

    return -1; // text ended before end of pat
}

int32_t ICU_Utility::parseAsciiInteger(const UnicodeString& str, int32_t& pos) {
    int32_t result = 0;
    UChar c;
    while (pos < str.length() && (c = str.charAt(pos)) >= u'0' && c <= u'9') {
        result = result * 10 + (c - u'0');
        pos++;
    }
    return result;
}

/**
 * Append a character to a rule that is being built up.  To flush
 * the quoteBuf to rule, make one final call with isLiteral == TRUE.
 * If there is no final character, pass in (UChar32)-1 as c.
 * @param rule the string to append the character to
 * @param c the character to append, or (UChar32)-1 if none.
 * @param isLiteral if true, then the given character should not be
 * quoted or escaped.  Usually this means it is a syntactic element
 * such as > or $
 * @param escapeUnprintable if true, then unprintable characters
 * should be escaped using \uxxxx or \Uxxxxxxxx.  These escapes will
 * appear outside of quotes.
 * @param quoteBuf a buffer which is used to build up quoted
 * substrings.  The caller should initially supply an empty buffer,
 * and thereafter should not modify the buffer.  The buffer should be
 * cleared out by, at the end, calling this method with a literal
 * character.
 */
void ICU_Utility::appendToRule(UnicodeString& rule,
                               UChar32 c,
                               UBool isLiteral,
                               UBool escapeUnprintable,
                               UnicodeString& quoteBuf) {
    // If we are escaping unprintables, then escape them outside
    // quotes.  \u and \U are not recognized within quotes.  The same
    // logic applies to literals, but literals are never escaped.
    if (isLiteral ||
        (escapeUnprintable && ICU_Utility::isUnprintable(c))) {
        if (quoteBuf.length() > 0) {
            // We prefer backslash u_util_APOSTROPHE to double u_util_APOSTROPHE
            // (more readable, less similar to ") so if there are
            // double u_util_APOSTROPHEs at the ends, we pull them outside
            // of the quote.

            // If the first thing in the quoteBuf is u_util_APOSTROPHE
            // (doubled) then pull it out.
            while (quoteBuf.length() >= 2 &&
                   quoteBuf.charAt(0) == u_util_APOSTROPHE &&
                   quoteBuf.charAt(1) == u_util_APOSTROPHE) {
                rule.append(u_util_BACKSLASH).append(u_util_APOSTROPHE);
                quoteBuf.remove(0, 2);
            }
            // If the last thing in the quoteBuf is u_util_APOSTROPHE
            // (doubled) then remove and count it and add it after.
            int32_t trailingCount = 0;
            while (quoteBuf.length() >= 2 &&
                   quoteBuf.charAt(quoteBuf.length()-2) == u_util_APOSTROPHE &&
                   quoteBuf.charAt(quoteBuf.length()-1) == u_util_APOSTROPHE) {
                quoteBuf.truncate(quoteBuf.length()-2);
                ++trailingCount;
            }
            if (quoteBuf.length() > 0) {
                rule.append(u_util_APOSTROPHE);
                rule.append(quoteBuf);
                rule.append(u_util_APOSTROPHE);
                quoteBuf.truncate(0);
            }
            while (trailingCount-- > 0) {
                rule.append(u_util_BACKSLASH).append(u_util_APOSTROPHE);
            }
        }
        if (c != (UChar32)-1) {
            /* Since spaces are ignored during parsing, they are
             * emitted only for readability.  We emit one here
             * only if there isn't already one at the end of the
             * rule.
             */
            if (c == u_util_SPACE) {
                int32_t len = rule.length();
                if (len > 0 && rule.charAt(len-1) != c) {
                    rule.append(c);
                }
            } else if (!escapeUnprintable || !ICU_Utility::escapeUnprintable(rule, c)) {
                rule.append(c);
            }
        }
    }

    // Escape ' and '\' and don't begin a quote just for them
    else if (quoteBuf.length() == 0 &&
             (c == u_util_APOSTROPHE || c == u_util_BACKSLASH)) {
        rule.append(u_util_BACKSLASH);
        rule.append(c);
    }

    // Specials (printable ascii that isn't [0-9a-zA-Z]) and
    // whitespace need quoting.  Also append stuff to quotes if we are
    // building up a quoted substring already.
    else if (quoteBuf.length() > 0 ||
             (c >= 0x0021 && c <= 0x007E &&
              !((c >= 0x0030/*'0'*/ && c <= 0x0039/*'9'*/) ||
                (c >= 0x0041/*'A'*/ && c <= 0x005A/*'Z'*/) ||
                (c >= 0x0061/*'a'*/ && c <= 0x007A/*'z'*/))) ||
             PatternProps::isWhiteSpace(c)) {
        quoteBuf.append(c);
        // Double ' within a quote
        if (c == u_util_APOSTROPHE) {
            quoteBuf.append(c);
        }
    }

    // Otherwise just append
    else {
        rule.append(c);
    }
}

void ICU_Utility::appendToRule(UnicodeString& rule,
                               const UnicodeString& text,
                               UBool isLiteral,
                               UBool escapeUnprintable,
                               UnicodeString& quoteBuf) {
    for (int32_t i=0; i<text.length(); ++i) {
        appendToRule(rule, text[i], isLiteral, escapeUnprintable, quoteBuf);
    }
}

/**
 * Given a matcher reference, which may be null, append its
 * pattern as a literal to the given rule.
 */
void ICU_Utility::appendToRule(UnicodeString& rule,
                               const UnicodeMatcher* matcher,
                               UBool escapeUnprintable,
                               UnicodeString& quoteBuf) {
    if (matcher != NULL) {
        UnicodeString pat;
        appendToRule(rule, matcher->toPattern(pat, escapeUnprintable),
                     TRUE, escapeUnprintable, quoteBuf);
    }
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (c) 2001-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*   Date        Name        Description
*   11/19/2001  aliu        Creation.
**********************************************************************
*/






U_NAMESPACE_BEGIN

/**
 * Parse an integer at pos, either of the form \d+ or of the form
 * 0x[0-9A-Fa-f]+ or 0[0-7]+, that is, in standard decimal, hex,
 * or octal format.
 * @param pos INPUT-OUTPUT parameter.  On input, the first
 * character to parse.  On output, the character after the last
 * parsed character.
 */
int32_t ICU_Utility::parseInteger(const UnicodeString& rule, int32_t& pos, int32_t limit) {
    int32_t count = 0;
    int32_t value = 0;
    int32_t p = pos;
    int8_t radix = 10;

    if (p < limit && rule.charAt(p) == 48 /*0*/) {
        if (p+1 < limit && (rule.charAt(p+1) == 0x78 /*x*/ || rule.charAt(p+1) == 0x58 /*X*/)) {
            p += 2;
            radix = 16;
        }
        else {
            p++;
            count = 1;
            radix = 8;
        }
    }

    while (p < limit) {
        int32_t d = u_digit(rule.charAt(p++), radix);
        if (d < 0) {
            --p;
            break;
        }
        ++count;
        int32_t v = (value * radix) + d;
        if (v <= value) {
            // If there are too many input digits, at some point
            // the value will go negative, e.g., if we have seen
            // "0x8000000" already and there is another '0', when
            // we parse the next 0 the value will go negative.
            return 0;
        }
        value = v;
    }
    if (count > 0) {
        pos = p;
    }
    return value;
}

/**
 * Parse a pattern string starting at offset pos.  Keywords are
 * matched case-insensitively.  Spaces may be skipped and may be
 * optional or required.  Integer values may be parsed, and if
 * they are, they will be returned in the given array.  If
 * successful, the offset of the next non-space character is
 * returned.  On failure, -1 is returned.
 * @param pattern must only contain lowercase characters, which
 * will match their uppercase equivalents as well.  A space
 * character matches one or more required spaces.  A '~' character
 * matches zero or more optional spaces.  A '#' character matches
 * an integer and stores it in parsedInts, which the caller must
 * ensure has enough capacity.
 * @param parsedInts array to receive parsed integers.  Caller
 * must ensure that parsedInts.length is >= the number of '#'
 * signs in 'pattern'.
 * @return the position after the last character parsed, or -1 if
 * the parse failed
 */
int32_t ICU_Utility::parsePattern(const UnicodeString& rule, int32_t pos, int32_t limit,
                              const UnicodeString& pattern, int32_t* parsedInts) {
    // TODO Update this to handle surrogates
    int32_t p;
    int32_t intCount = 0; // number of integers parsed
    for (int32_t i=0; i<pattern.length(); ++i) {
        UChar cpat = pattern.charAt(i);
        UChar c;
        switch (cpat) {
        case 32 /*' '*/:
            if (pos >= limit) {
                return -1;
            }
            c = rule.charAt(pos++);
            if (!PatternProps::isWhiteSpace(c)) {
                return -1;
            }
            // FALL THROUGH to skipWhitespace
            U_FALLTHROUGH;
        case 126 /*'~'*/:
            pos = skipWhitespace(rule, pos);
            break;
        case 35 /*'#'*/:
            p = pos;
            parsedInts[intCount++] = parseInteger(rule, p, limit);
            if (p == pos) {
                // Syntax error; failed to parse integer
                return -1;
            }
            pos = p;
            break;
        default:
            if (pos >= limit) {
                return -1;
            }
            c = (UChar) u_tolower(rule.charAt(pos++));
            if (c != cpat) {
                return -1;
            }
            break;
        }
    }
    return pos;
}

/**
 * Parse a Unicode identifier from the given string at the given
 * position.  Return the identifier, or an empty string if there
 * is no identifier.
 * @param str the string to parse
 * @param pos INPUT-OUPUT parameter.  On INPUT, pos is the
 * first character to examine.  It must be less than str.length(),
 * and it must not point to a whitespace character.  That is, must
 * have pos < str.length().  On
 * OUTPUT, the position after the last parsed character.
 * @return the Unicode identifier, or an empty string if there is
 * no valid identifier at pos.
 */
UnicodeString ICU_Utility::parseUnicodeIdentifier(const UnicodeString& str, int32_t& pos) {
    // assert(pos < str.length());
    UnicodeString buf;
    int p = pos;
    while (p < str.length()) {
        UChar32 ch = str.char32At(p);
        if (buf.length() == 0) {
            if (u_isIDStart(ch)) {
                buf.append(ch);
            } else {
                buf.truncate(0);
                return buf;
            }
        } else {
            if (u_isIDPart(ch)) {
                buf.append(ch);
            } else {
                break;
            }
        }
        p += U16_LENGTH(ch);
    }
    pos = p;
    return buf;
}

/**
 * Parse an unsigned 31-bit integer at the given offset.  Use
 * UCharacter.digit() to parse individual characters into digits.
 * @param text the text to be parsed
 * @param pos INPUT-OUTPUT parameter.  On entry, pos[0] is the
 * offset within text at which to start parsing; it should point
 * to a valid digit.  On exit, pos[0] is the offset after the last
 * parsed character.  If the parse failed, it will be unchanged on
 * exit.  Must be >= 0 on entry.
 * @param radix the radix in which to parse; must be >= 2 and <=
 * 36.
 * @return a non-negative parsed number, or -1 upon parse failure.
 * Parse fails if there are no digits, that is, if pos[0] does not
 * point to a valid digit on entry, or if the number to be parsed
 * does not fit into a 31-bit unsigned integer.
 */
int32_t ICU_Utility::parseNumber(const UnicodeString& text,
                                 int32_t& pos, int8_t radix) {
    // assert(pos[0] >= 0);
    // assert(radix >= 2);
    // assert(radix <= 36);
    int32_t n = 0;
    int32_t p = pos;
    while (p < text.length()) {
        UChar32 ch = text.char32At(p);
        int32_t d = u_digit(ch, radix);
        if (d < 0) {
            break;
        }
        n = radix*n + d;
        // ASSUME that when a 32-bit integer overflows it becomes
        // negative.  E.g., 214748364 * 10 + 8 => negative value.
        if (n < 0) {
            return -1;
        }
        ++p;
    }
    if (p == pos) {
        return -1;
    }
    pos = p;
    return n;
}

U_NAMESPACE_END

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2003-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  utrace.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*/








static UTraceEntry     *pTraceEntryFunc = NULL;
static UTraceExit      *pTraceExitFunc  = NULL;
static UTraceData      *pTraceDataFunc  = NULL;
static const void      *gTraceContext   = NULL;

/**
 * \var utrace_level
 * Trace level variable. Negative for "off".
 */
static int32_t
utrace_level = UTRACE_ERROR;

U_CAPI void U_EXPORT2
utrace_entry(int32_t fnNumber) {
    if (pTraceEntryFunc != NULL) {
        (*pTraceEntryFunc)(gTraceContext, fnNumber);
    }
}


static const char gExitFmt[]             = "Returns.";
static const char gExitFmtValue[]        = "Returns %d.";
static const char gExitFmtStatus[]       = "Returns.  Status = %d.";
static const char gExitFmtValueStatus[]  = "Returns %d.  Status = %d.";
static const char gExitFmtPtrStatus[]    = "Returns %d.  Status = %p.";

U_CAPI void U_EXPORT2
utrace_exit(int32_t fnNumber, int32_t returnType, ...) {
    if (pTraceExitFunc != NULL) {
        va_list     args;
        const char *fmt;

        switch (returnType) {
        case 0:
            fmt = gExitFmt;
            break;
        case UTRACE_EXITV_I32:
            fmt = gExitFmtValue;
            break;
        case UTRACE_EXITV_STATUS:
            fmt = gExitFmtStatus;
            break;
        case UTRACE_EXITV_I32 | UTRACE_EXITV_STATUS:
            fmt = gExitFmtValueStatus;
            break;
        case UTRACE_EXITV_PTR | UTRACE_EXITV_STATUS:
            fmt = gExitFmtPtrStatus;
            break;
        default:
            UPRV_UNREACHABLE;
        }

        va_start(args, returnType);
        (*pTraceExitFunc)(gTraceContext, fnNumber, fmt, args);
        va_end(args);
    }
}
 

 
U_CAPI void U_EXPORT2 
utrace_data(int32_t fnNumber, int32_t level, const char *fmt, ...) {
    if (pTraceDataFunc != NULL) {
           va_list args;
           va_start(args, fmt ); 
           (*pTraceDataFunc)(gTraceContext, fnNumber, level, fmt, args);
           va_end(args);
    }
}


static void outputChar(char c, char *outBuf, int32_t *outIx, int32_t capacity, int32_t indent) {
    int32_t i;
    /* Check whether a start of line indenting is needed.  Three cases:
     *   1.  At the start of the first line  (output index == 0).
     *   2.  At the start of subsequent lines  (preceeding char in buffer == '\n')
     *   3.  When preflighting buffer len (buffer capacity is exceeded), when
     *       a \n is output.  Ideally we wouldn't do the indent until the following char
     *       is received, but that won't work because there's no place to remember that
     *       the preceding char was \n.  Meaning that we may overstimate the
     *       buffer size needed.  No harm done.
     */
    if (*outIx==0 ||   /* case 1. */
        (c!='\n' && c!=0 && *outIx < capacity && outBuf[(*outIx)-1]=='\n') ||  /* case 2. */
        (c=='\n' && *outIx>=capacity))    /* case 3 */
    {
        /* At the start of a line.  Indent. */
        for(i=0; i<indent; i++) {
            if (*outIx < capacity) {
                outBuf[*outIx] = ' ';
            }
            (*outIx)++;
        }
    }

    if (*outIx < capacity) {
        outBuf[*outIx] = c;
    }
    if (c != 0) {
        /* Nulls only appear as end-of-string terminators.  Move them to the output
         *  buffer, but do not update the length of the buffer, so that any
         *  following output will overwrite the null. */
        (*outIx)++;
    }
}

static void outputHexBytes(int64_t val, int32_t charsToOutput,
                           char *outBuf, int32_t *outIx, int32_t capacity) {
    static const char gHexChars[] = "0123456789abcdef";
    int32_t shiftCount;
    for  (shiftCount=(charsToOutput-1)*4; shiftCount >= 0; shiftCount-=4) {
        char c = gHexChars[(val >> shiftCount) & 0xf];
        outputChar(c, outBuf, outIx, capacity, 0);
    }
}

/* Output a pointer value in hex.  Work with any size of pointer   */
static void outputPtrBytes(void *val, char *outBuf, int32_t *outIx, int32_t capacity) {
    uint32_t  i;
    int32_t  incVal = 1;              /* +1 for big endian, -1 for little endian          */
    char     *p     = (char *)&val;   /* point to current byte to output in the ptr val  */

#if !U_IS_BIG_ENDIAN
    /* Little Endian.  Move p to most significant end of the value      */
    incVal = -1;
    p += sizeof(void *) - 1;
#endif

    /* Loop through the bytes of the ptr as it sits in memory, from 
     * most significant to least significant end                    */
    for (i=0; i<sizeof(void *); i++) {
        outputHexBytes(*p, 2, outBuf, outIx, capacity);
        p += incVal;
    }
}

static void outputString(const char *s, char *outBuf, int32_t *outIx, int32_t capacity, int32_t indent) {
    int32_t i = 0;
    char    c;
    if (s==NULL) {
        s = "*NULL*";
    }
    do {
        c = s[i++];
        outputChar(c, outBuf, outIx, capacity, indent);
    } while (c != 0);
}
        


static void outputUString(const UChar *s, int32_t len, 
                          char *outBuf, int32_t *outIx, int32_t capacity, int32_t indent) {
    int32_t i = 0;
    UChar   c;
    if (s==NULL) {
        outputString(NULL, outBuf, outIx, capacity, indent);
        return;
    }

    for (i=0; i<len || len==-1; i++) {
        c = s[i];
        outputHexBytes(c, 4, outBuf, outIx, capacity);
        outputChar(' ', outBuf, outIx, capacity, indent);
        if (len == -1 && c==0) {
            break;
        }
    }
}
        
U_CAPI int32_t U_EXPORT2
utrace_vformat(char *outBuf, int32_t capacity, int32_t indent, const char *fmt, va_list args) {
    int32_t   outIx  = 0;
    int32_t   fmtIx  = 0;
    char      fmtC;
    char      c;
    int32_t   intArg;
    int64_t   longArg = 0;
    char      *ptrArg;

    /*   Loop runs once for each character in the format string.
     */
    for (;;) {
        fmtC = fmt[fmtIx++];
        if (fmtC != '%') {
            /* Literal character, not part of a %sequence.  Just copy it to the output. */
            outputChar(fmtC, outBuf, &outIx, capacity, indent);
            if (fmtC == 0) {
                /* We hit the null that terminates the format string.
                 * This is the normal (and only) exit from the loop that
                 * interprets the format
                 */
                break;
            }
            continue;
        }

        /* We encountered a '%'.  Pick up the following format char */
        fmtC = fmt[fmtIx++];

        switch (fmtC) {
        case 'c':
            /* single 8 bit char   */
            c = (char)va_arg(args, int32_t);
            outputChar(c, outBuf, &outIx, capacity, indent);
            break;

        case 's':
            /* char * string, null terminated.  */
            ptrArg = va_arg(args, char *);
            outputString((const char *)ptrArg, outBuf, &outIx, capacity, indent);
            break;

        case 'S':
            /* UChar * string, with length, len==-1 for null terminated. */
            ptrArg = va_arg(args, char *);             /* Ptr    */
            intArg =(int32_t)va_arg(args, int32_t);    /* Length */
            outputUString((const UChar *)ptrArg, intArg, outBuf, &outIx, capacity, indent);
            break;

        case 'b':
            /*  8 bit int  */
            intArg = va_arg(args, int);
            outputHexBytes(intArg, 2, outBuf, &outIx, capacity);
            break;

        case 'h':
            /*  16 bit int  */
            intArg = va_arg(args, int);
            outputHexBytes(intArg, 4, outBuf, &outIx, capacity);
            break;

        case 'd':
            /*  32 bit int  */
            intArg = va_arg(args, int);
            outputHexBytes(intArg, 8, outBuf, &outIx, capacity);
            break;

        case 'l':
            /*  64 bit long  */
            longArg = va_arg(args, int64_t);
            outputHexBytes(longArg, 16, outBuf, &outIx, capacity);
            break;
            
        case 'p':
            /*  Pointers.   */
            ptrArg = va_arg(args, char *);
            outputPtrBytes(ptrArg, outBuf, &outIx, capacity);
            break;

        case 0:
            /* Single '%' at end of fmt string.  Output as literal '%'.   
             * Back up index into format string so that the terminating null will be
             * re-fetched in the outer loop, causing it to terminate.
             */
            outputChar('%', outBuf, &outIx, capacity, indent);
            fmtIx--;
            break;

        case 'v':
            {
                /* Vector of values, e.g. %vh */
                char     vectorType;
                int32_t  vectorLen;
                const char   *i8Ptr;
                int16_t  *i16Ptr;
                int32_t  *i32Ptr;
                int64_t  *i64Ptr;
                void     **ptrPtr;
                int32_t   charsToOutput = 0;
                int32_t   i;
                
                vectorType = fmt[fmtIx];    /* b, h, d, l, p, etc. */
                if (vectorType != 0) {
                    fmtIx++;
                }
                i8Ptr = (const char *)va_arg(args, void*);
                i16Ptr = (int16_t *)i8Ptr;
                i32Ptr = (int32_t *)i8Ptr;
                i64Ptr = (int64_t *)i8Ptr;
                ptrPtr = (void **)i8Ptr;
                vectorLen =(int32_t)va_arg(args, int32_t);
                if (ptrPtr == NULL) {
                    outputString("*NULL* ", outBuf, &outIx, capacity, indent);
                } else {
                    for (i=0; i<vectorLen || vectorLen==-1; i++) { 
                        switch (vectorType) {
                        case 'b':
                            charsToOutput = 2;
                            longArg = *i8Ptr++;
                            break;
                        case 'h':
                            charsToOutput = 4;
                            longArg = *i16Ptr++;
                            break;
                        case 'd':
                            charsToOutput = 8;
                            longArg = *i32Ptr++;
                            break;
                        case 'l':
                            charsToOutput = 16;
                            longArg = *i64Ptr++;
                            break;
                        case 'p':
                            charsToOutput = 0;
                            outputPtrBytes(*ptrPtr, outBuf, &outIx, capacity);
                            longArg = *ptrPtr==NULL? 0: 1;    /* test for null terminated array. */
                            ptrPtr++;
                            break;
                        case 'c':
                            charsToOutput = 0;
                            outputChar(*i8Ptr, outBuf, &outIx, capacity, indent);
                            longArg = *i8Ptr;    /* for test for null terminated array. */
                            i8Ptr++;
                            break;
                        case 's':
                            charsToOutput = 0;
                            outputString((const char *)*ptrPtr, outBuf, &outIx, capacity, indent);
                            outputChar('\n', outBuf, &outIx, capacity, indent);
                            longArg = *ptrPtr==NULL? 0: 1;   /* for test for null term. array. */
                            ptrPtr++;
                            break;

                        case 'S':
                            charsToOutput = 0;
                            outputUString((const UChar *)*ptrPtr, -1, outBuf, &outIx, capacity, indent);
                            outputChar('\n', outBuf, &outIx, capacity, indent);
                            longArg = *ptrPtr==NULL? 0: 1;   /* for test for null term. array. */
                            ptrPtr++;
                            break;

                            
                        }
                        if (charsToOutput > 0) {
                            outputHexBytes(longArg, charsToOutput, outBuf, &outIx, capacity);
                            outputChar(' ', outBuf, &outIx, capacity, indent);
                        }
                        if (vectorLen == -1 && longArg == 0) {
                            break;
                        }
                    }
                }
                outputChar('[', outBuf, &outIx, capacity, indent);
                outputHexBytes(vectorLen, 8, outBuf, &outIx, capacity);
                outputChar(']', outBuf, &outIx, capacity, indent);
            }
            break;


        default:
            /* %. in format string, where . is some character not in the set
             *    of recognized format chars.  Just output it as if % wasn't there.
             *    (Covers "%%" outputing a single '%')
             */
             outputChar(fmtC, outBuf, &outIx, capacity, indent);
        }
    }
    outputChar(0, outBuf, &outIx, capacity, indent);  /* Make sure that output is null terminated  */
    return outIx + 1;     /* outIx + 1 because outIx does not increment when outputing final null. */
}




U_CAPI int32_t U_EXPORT2
utrace_format(char *outBuf, int32_t capacity,
                int32_t indent, const char *fmt,  ...) {
    int32_t retVal;
    va_list args;
    va_start(args, fmt ); 
    retVal = utrace_vformat(outBuf, capacity, indent, fmt, args);
    va_end(args);
    return retVal;
}


U_CAPI void U_EXPORT2
utrace_setFunctions(const void *context,
                    UTraceEntry *e, UTraceExit *x, UTraceData *d) {
    pTraceEntryFunc = e;
    pTraceExitFunc  = x;
    pTraceDataFunc  = d;
    gTraceContext   = context;
}


U_CAPI void U_EXPORT2
utrace_getFunctions(const void **context,
                    UTraceEntry **e, UTraceExit **x, UTraceData **d) {
    *e = pTraceEntryFunc;
    *x = pTraceExitFunc;
    *d = pTraceDataFunc;
    *context = gTraceContext;
}

U_CAPI void U_EXPORT2
utrace_setLevel(int32_t level) {
    if (level < UTRACE_OFF) {
        level = UTRACE_OFF;
    }
    if (level > UTRACE_VERBOSE) {
        level = UTRACE_VERBOSE;
    }
    utrace_level = level;
}

U_CAPI int32_t U_EXPORT2
utrace_getLevel() {
    return utrace_level;
}


U_CFUNC UBool 
utrace_cleanup() {
    pTraceEntryFunc = NULL;
    pTraceExitFunc  = NULL;
    pTraceDataFunc  = NULL;
    utrace_level    = UTRACE_OFF;
    gTraceContext   = NULL;
    return TRUE;
}


static const char * const
trFnName[] = {
    "u_init",
    "u_cleanup",
    NULL
};


static const char * const
trConvNames[] = {
    "ucnv_open",
    "ucnv_openPackage",
    "ucnv_openAlgorithmic",
    "ucnv_clone",
    "ucnv_close",
    "ucnv_flushCache",
    "ucnv_load",
    "ucnv_unload",
    NULL
};

    
static const char * const
trCollNames[] = {
    "ucol_open",
    "ucol_close",
    "ucol_strcoll",
    "ucol_getSortKey",
    "ucol_getLocale",
    "ucol_nextSortKeyPart",
    "ucol_strcollIter",
    "ucol_openFromShortString",
    "ucol_strcollUTF8",
    NULL
};


static const char* const
trResDataNames[] = {
    "resc",
    "bundle-open",
    "file-open",
    "res-open",
    NULL
};

                
U_CAPI const char * U_EXPORT2
utrace_functionName(int32_t fnNumber) {
    if(UTRACE_FUNCTION_START <= fnNumber && fnNumber < UTRACE_FUNCTION_LIMIT) {
        return trFnName[fnNumber];
    } else if(UTRACE_CONVERSION_START <= fnNumber && fnNumber < UTRACE_CONVERSION_LIMIT) {
        return trConvNames[fnNumber - UTRACE_CONVERSION_START];
    } else if(UTRACE_COLLATION_START <= fnNumber && fnNumber < UTRACE_COLLATION_LIMIT){
        return trCollNames[fnNumber - UTRACE_COLLATION_START];
    } else if(UTRACE_UDATA_START <= fnNumber && fnNumber < UTRACE_RES_DATA_LIMIT){
        return trResDataNames[fnNumber - UTRACE_UDATA_START];
    } else {
        return "[BOGUS Trace Function Number]";
    }
}

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2001-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  utrie.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001oct20
*   created by: Markus W. Scherer
*
*   This is a common implementation of a "folded" trie.
*   It is a kind of compressed, serializable table of 16- or 32-bit values associated with
*   Unicode code points (0..0x10ffff).
*/

#ifdef UTRIE_DEBUG
#   include <stdio.h>
#endif





/* miscellaneous ------------------------------------------------------------ */

#undef ABS
#define ABS(x) ((x)>=0 ? (x) : -(x))

static inline UBool
utrie_equal_uint32(const uint32_t *s, const uint32_t *t, int32_t length) {
    while(length>0 && *s==*t) {
        ++s;
        ++t;
        --length;
    }
    return (UBool)(length==0);
}

/* Building a trie ----------------------------------------------------------*/

U_CAPI UNewTrie * U_EXPORT2
utrie_open(UNewTrie *fillIn,
           uint32_t *aliasData, int32_t maxDataLength,
           uint32_t initialValue, uint32_t leadUnitValue,
           UBool latin1Linear) {
    UNewTrie *trie;
    int32_t i, j;

    if( maxDataLength<UTRIE_DATA_BLOCK_LENGTH ||
        (latin1Linear && maxDataLength<1024)
    ) {
        return NULL;
    }

    if(fillIn!=NULL) {
        trie=fillIn;
    } else {
        trie=(UNewTrie *)uprv_malloc(sizeof(UNewTrie));
        if(trie==NULL) {
            return NULL;
        }
    }
    uprv_memset(trie, 0, sizeof(UNewTrie));
    trie->isAllocated= (UBool)(fillIn==NULL);

    if(aliasData!=NULL) {
        trie->data=aliasData;
        trie->isDataAllocated=FALSE;
    } else {
        trie->data=(uint32_t *)uprv_malloc(maxDataLength*4);
        if(trie->data==NULL) {
            uprv_free(trie);
            return NULL;
        }
        trie->isDataAllocated=TRUE;
    }

    /* preallocate and reset the first data block (block index 0) */
    j=UTRIE_DATA_BLOCK_LENGTH;

    if(latin1Linear) {
        /* preallocate and reset the first block (number 0) and Latin-1 (U+0000..U+00ff) after that */
        /* made sure above that maxDataLength>=1024 */

        /* set indexes to point to consecutive data blocks */
        i=0;
        do {
            /* do this at least for trie->index[0] even if that block is only partly used for Latin-1 */
            trie->index[i++]=j;
            j+=UTRIE_DATA_BLOCK_LENGTH;
        } while(i<(256>>UTRIE_SHIFT));
    }

    /* reset the initially allocated blocks to the initial value */
    trie->dataLength=j;
    while(j>0) {
        trie->data[--j]=initialValue;
    }

    trie->leadUnitValue=leadUnitValue;
    trie->indexLength=UTRIE_MAX_INDEX_LENGTH;
    trie->dataCapacity=maxDataLength;
    trie->isLatin1Linear=latin1Linear;
    trie->isCompacted=FALSE;
    return trie;
}

U_CAPI UNewTrie * U_EXPORT2
utrie_clone(UNewTrie *fillIn, const UNewTrie *other, uint32_t *aliasData, int32_t aliasDataCapacity) {
    UNewTrie *trie;
    UBool isDataAllocated;

    /* do not clone if other is not valid or already compacted */
    if(other==NULL || other->data==NULL || other->isCompacted) {
        return NULL;
    }

    /* clone data */
    if(aliasData!=NULL && aliasDataCapacity>=other->dataCapacity) {
        isDataAllocated=FALSE;
    } else {
        aliasDataCapacity=other->dataCapacity;
        aliasData=(uint32_t *)uprv_malloc(other->dataCapacity*4);
        if(aliasData==NULL) {
            return NULL;
        }
        isDataAllocated=TRUE;
    }

    trie=utrie_open(fillIn, aliasData, aliasDataCapacity,
                    other->data[0], other->leadUnitValue,
                    other->isLatin1Linear);
    if(trie==NULL) {
        uprv_free(aliasData);
    } else {
        uprv_memcpy(trie->index, other->index, sizeof(trie->index));
        uprv_memcpy(trie->data, other->data, (size_t)other->dataLength*4);
        trie->dataLength=other->dataLength;
        trie->isDataAllocated=isDataAllocated;
    }

    return trie;
}

U_CAPI void U_EXPORT2
utrie_close(UNewTrie *trie) {
    if(trie!=NULL) {
        if(trie->isDataAllocated) {
            uprv_free(trie->data);
            trie->data=NULL;
        }
        if(trie->isAllocated) {
            uprv_free(trie);
        }
    }
}

U_CAPI uint32_t * U_EXPORT2
utrie_getData(UNewTrie *trie, int32_t *pLength) {
    if(trie==NULL || pLength==NULL) {
        return NULL;
    }

    *pLength=trie->dataLength;
    return trie->data;
}

static int32_t
utrie_allocDataBlock(UNewTrie *trie) {
    int32_t newBlock, newTop;

    newBlock=trie->dataLength;
    newTop=newBlock+UTRIE_DATA_BLOCK_LENGTH;
    if(newTop>trie->dataCapacity) {
        /* out of memory in the data array */
        return -1;
    }
    trie->dataLength=newTop;
    return newBlock;
}

/**
 * No error checking for illegal arguments.
 *
 * @return -1 if no new data block available (out of memory in data array)
 * @internal
 */
static int32_t
utrie_getDataBlock(UNewTrie *trie, UChar32 c) {
    int32_t indexValue, newBlock;

    c>>=UTRIE_SHIFT;
    indexValue=trie->index[c];
    if(indexValue>0) {
        return indexValue;
    }

    /* allocate a new data block */
    newBlock=utrie_allocDataBlock(trie);
    if(newBlock<0) {
        /* out of memory in the data array */
        return -1;
    }
    trie->index[c]=newBlock;

    /* copy-on-write for a block from a setRange() */
    uprv_memcpy(trie->data+newBlock, trie->data-indexValue, 4*UTRIE_DATA_BLOCK_LENGTH);
    return newBlock;
}

/**
 * @return TRUE if the value was successfully set
 */
U_CAPI UBool U_EXPORT2
utrie_set32(UNewTrie *trie, UChar32 c, uint32_t value) {
    int32_t block;

    /* valid, uncompacted trie and valid c? */
    if(trie==NULL || trie->isCompacted || (uint32_t)c>0x10ffff) {
        return FALSE;
    }

    block=utrie_getDataBlock(trie, c);
    if(block<0) {
        return FALSE;
    }

    trie->data[block+(c&UTRIE_MASK)]=value;
    return TRUE;
}

U_CAPI uint32_t U_EXPORT2
utrie_get32(UNewTrie *trie, UChar32 c, UBool *pInBlockZero) {
    int32_t block;

    /* valid, uncompacted trie and valid c? */
    if(trie==NULL || trie->isCompacted || (uint32_t)c>0x10ffff) {
        if(pInBlockZero!=NULL) {
            *pInBlockZero=TRUE;
        }
        return 0;
    }

    block=trie->index[c>>UTRIE_SHIFT];
    if(pInBlockZero!=NULL) {
        *pInBlockZero= (UBool)(block==0);
    }

    return trie->data[ABS(block)+(c&UTRIE_MASK)];
}

/**
 * @internal
 */
static void
utrie_fillBlock(uint32_t *block, UChar32 start, UChar32 limit,
                uint32_t value, uint32_t initialValue, UBool overwrite) {
    uint32_t *pLimit;

    pLimit=block+limit;
    block+=start;
    if(overwrite) {
        while(block<pLimit) {
            *block++=value;
        }
    } else {
        while(block<pLimit) {
            if(*block==initialValue) {
                *block=value;
            }
            ++block;
        }
    }
}

U_CAPI UBool U_EXPORT2
utrie_setRange32(UNewTrie *trie, UChar32 start, UChar32 limit, uint32_t value, UBool overwrite) {
    /*
     * repeat value in [start..limit[
     * mark index values for repeat-data blocks by setting bit 31 of the index values
     * fill around existing values if any, if(overwrite)
     */
    uint32_t initialValue;
    int32_t block, rest, repeatBlock;

    /* valid, uncompacted trie and valid indexes? */
    if( trie==NULL || trie->isCompacted ||
        (uint32_t)start>0x10ffff || (uint32_t)limit>0x110000 || start>limit
    ) {
        return FALSE;
    }
    if(start==limit) {
        return TRUE; /* nothing to do */
    }

    initialValue=trie->data[0];
    if(start&UTRIE_MASK) {
        UChar32 nextStart;

        /* set partial block at [start..following block boundary[ */
        block=utrie_getDataBlock(trie, start);
        if(block<0) {
            return FALSE;
        }

        nextStart=(start+UTRIE_DATA_BLOCK_LENGTH)&~UTRIE_MASK;
        if(nextStart<=limit) {
            utrie_fillBlock(trie->data+block, start&UTRIE_MASK, UTRIE_DATA_BLOCK_LENGTH,
                            value, initialValue, overwrite);
            start=nextStart;
        } else {
            utrie_fillBlock(trie->data+block, start&UTRIE_MASK, limit&UTRIE_MASK,
                            value, initialValue, overwrite);
            return TRUE;
        }
    }

    /* number of positions in the last, partial block */
    rest=limit&UTRIE_MASK;

    /* round down limit to a block boundary */
    limit&=~UTRIE_MASK;

    /* iterate over all-value blocks */
    if(value==initialValue) {
        repeatBlock=0;
    } else {
        repeatBlock=-1;
    }
    while(start<limit) {
        /* get index value */
        block=trie->index[start>>UTRIE_SHIFT];
        if(block>0) {
            /* already allocated, fill in value */
            utrie_fillBlock(trie->data+block, 0, UTRIE_DATA_BLOCK_LENGTH, value, initialValue, overwrite);
        } else if(trie->data[-block]!=value && (block==0 || overwrite)) {
            /* set the repeatBlock instead of the current block 0 or range block */
            if(repeatBlock>=0) {
                trie->index[start>>UTRIE_SHIFT]=-repeatBlock;
            } else {
                /* create and set and fill the repeatBlock */
                repeatBlock=utrie_getDataBlock(trie, start);
                if(repeatBlock<0) {
                    return FALSE;
                }

                /* set the negative block number to indicate that it is a repeat block */
                trie->index[start>>UTRIE_SHIFT]=-repeatBlock;
                utrie_fillBlock(trie->data+repeatBlock, 0, UTRIE_DATA_BLOCK_LENGTH, value, initialValue, TRUE);
            }
        }

        start+=UTRIE_DATA_BLOCK_LENGTH;
    }

    if(rest>0) {
        /* set partial block at [last block boundary..limit[ */
        block=utrie_getDataBlock(trie, start);
        if(block<0) {
            return FALSE;
        }

        utrie_fillBlock(trie->data+block, 0, rest, value, initialValue, overwrite);
    }

    return TRUE;
}

static int32_t
_findSameIndexBlock(const int32_t *idx, int32_t indexLength,
                    int32_t otherBlock) {
    int32_t block, i;

    for(block=UTRIE_BMP_INDEX_LENGTH; block<indexLength; block+=UTRIE_SURROGATE_BLOCK_COUNT) {
        for(i=0; i<UTRIE_SURROGATE_BLOCK_COUNT; ++i) {
            if(idx[block+i]!=idx[otherBlock+i]) {
                break;
            }
        }
        if(i==UTRIE_SURROGATE_BLOCK_COUNT) {
            return block;
        }
    }
    return indexLength;
}

/*
 * Fold the normalization data for supplementary code points into
 * a compact area on top of the BMP-part of the trie index,
 * with the lead surrogates indexing this compact area.
 *
 * Duplicate the index values for lead surrogates:
 * From inside the BMP area, where some may be overridden with folded values,
 * to just after the BMP area, where they can be retrieved for
 * code point lookups.
 */
static void
utrie_fold(UNewTrie *trie, UNewTrieGetFoldedValue *getFoldedValue, UErrorCode *pErrorCode) {
    int32_t leadIndexes[UTRIE_SURROGATE_BLOCK_COUNT];
    int32_t *idx;
    uint32_t value;
    UChar32 c;
    int32_t indexLength, block;
#ifdef UTRIE_DEBUG
    int countLeadCUWithData=0;
#endif

    idx=trie->index;

    /* copy the lead surrogate indexes into a temporary array */
    uprv_memcpy(leadIndexes, idx+(0xd800>>UTRIE_SHIFT), 4*UTRIE_SURROGATE_BLOCK_COUNT);

    /*
     * set all values for lead surrogate code *units* to leadUnitValue
     * so that, by default, runtime lookups will find no data for associated
     * supplementary code points, unless there is data for such code points
     * which will result in a non-zero folding value below that is set for
     * the respective lead units
     *
     * the above saved the indexes for surrogate code *points*
     * fill the indexes with simplified code from utrie_setRange32()
     */
    if(trie->leadUnitValue==trie->data[0]) {
        block=0; /* leadUnitValue==initialValue, use all-initial-value block */
    } else {
        /* create and fill the repeatBlock */
        block=utrie_allocDataBlock(trie);
        if(block<0) {
            /* data table overflow */
            *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        utrie_fillBlock(trie->data+block, 0, UTRIE_DATA_BLOCK_LENGTH, trie->leadUnitValue, trie->data[0], TRUE);
        block=-block; /* negative block number to indicate that it is a repeat block */
    }
    for(c=(0xd800>>UTRIE_SHIFT); c<(0xdc00>>UTRIE_SHIFT); ++c) {
        trie->index[c]=block;
    }

    /*
     * Fold significant index values into the area just after the BMP indexes.
     * In case the first lead surrogate has significant data,
     * its index block must be used first (in which case the folding is a no-op).
     * Later all folded index blocks are moved up one to insert the copied
     * lead surrogate indexes.
     */
    indexLength=UTRIE_BMP_INDEX_LENGTH;

    /* search for any index (stage 1) entries for supplementary code points */
    for(c=0x10000; c<0x110000;) {
        if(idx[c>>UTRIE_SHIFT]!=0) {
            /* there is data, treat the full block for a lead surrogate */
            c&=~0x3ff;

#ifdef UTRIE_DEBUG
            ++countLeadCUWithData;
            /* printf("supplementary data for lead surrogate U+%04lx\n", (long)(0xd7c0+(c>>10))); */
#endif

            /* is there an identical index block? */
            block=_findSameIndexBlock(idx, indexLength, c>>UTRIE_SHIFT);

            /*
             * get a folded value for [c..c+0x400[ and,
             * if different from the value for the lead surrogate code point,
             * set it for the lead surrogate code unit
             */
            value=getFoldedValue(trie, c, block+UTRIE_SURROGATE_BLOCK_COUNT);
            if(value!=utrie_get32(trie, U16_LEAD(c), NULL)) {
                if(!utrie_set32(trie, U16_LEAD(c), value)) {
                    /* data table overflow */
                    *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
                    return;
                }

                /* if we did not find an identical index block... */
                if(block==indexLength) {
                    /* move the actual index (stage 1) entries from the supplementary position to the new one */
                    uprv_memmove(idx+indexLength,
                                 idx+(c>>UTRIE_SHIFT),
                                 4*UTRIE_SURROGATE_BLOCK_COUNT);
                    indexLength+=UTRIE_SURROGATE_BLOCK_COUNT;
                }
            }
            c+=0x400;
        } else {
            c+=UTRIE_DATA_BLOCK_LENGTH;
        }
    }
#ifdef UTRIE_DEBUG
    if(countLeadCUWithData>0) {
        printf("supplementary data for %d lead surrogates\n", countLeadCUWithData);
    }
#endif

    /*
     * index array overflow?
     * This is to guarantee that a folding offset is of the form
     * UTRIE_BMP_INDEX_LENGTH+n*UTRIE_SURROGATE_BLOCK_COUNT with n=0..1023.
     * If the index is too large, then n>=1024 and more than 10 bits are necessary.
     *
     * In fact, it can only ever become n==1024 with completely unfoldable data and
     * the additional block of duplicated values for lead surrogates.
     */
    if(indexLength>=UTRIE_MAX_INDEX_LENGTH) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return;
    }

    /*
     * make space for the lead surrogate index block and
     * insert it between the BMP indexes and the folded ones
     */
    uprv_memmove(idx+UTRIE_BMP_INDEX_LENGTH+UTRIE_SURROGATE_BLOCK_COUNT,
                 idx+UTRIE_BMP_INDEX_LENGTH,
                 4*(indexLength-UTRIE_BMP_INDEX_LENGTH));
    uprv_memcpy(idx+UTRIE_BMP_INDEX_LENGTH,
                leadIndexes,
                4*UTRIE_SURROGATE_BLOCK_COUNT);
    indexLength+=UTRIE_SURROGATE_BLOCK_COUNT;

#ifdef UTRIE_DEBUG
    printf("trie index count: BMP %ld  all Unicode %ld  folded %ld\n",
           UTRIE_BMP_INDEX_LENGTH, (long)UTRIE_MAX_INDEX_LENGTH, indexLength);
#endif

    trie->indexLength=indexLength;
}

/*
 * Set a value in the trie index map to indicate which data block
 * is referenced and which one is not.
 * utrie_compact() will remove data blocks that are not used at all.
 * Set
 * - 0 if it is used
 * - -1 if it is not used
 */
static void
_findUnusedBlocks(UNewTrie *trie) {
    int32_t i;

    /* fill the entire map with "not used" */
    uprv_memset(trie->map, 0xff, (UTRIE_MAX_BUILD_TIME_DATA_LENGTH>>UTRIE_SHIFT)*4);

    /* mark each block that _is_ used with 0 */
    for(i=0; i<trie->indexLength; ++i) {
        trie->map[ABS(trie->index[i])>>UTRIE_SHIFT]=0;
    }

    /* never move the all-initial-value block 0 */
    trie->map[0]=0;
}

static int32_t
_findSameDataBlock(const uint32_t *data, int32_t dataLength,
                   int32_t otherBlock, int32_t step) {
    int32_t block;

    /* ensure that we do not even partially get past dataLength */
    dataLength-=UTRIE_DATA_BLOCK_LENGTH;

    for(block=0; block<=dataLength; block+=step) {
        if(utrie_equal_uint32(data+block, data+otherBlock, UTRIE_DATA_BLOCK_LENGTH)) {
            return block;
        }
    }
    return -1;
}

/*
 * Compact a folded build-time trie.
 *
 * The compaction
 * - removes blocks that are identical with earlier ones
 * - overlaps adjacent blocks as much as possible (if overlap==TRUE)
 * - moves blocks in steps of the data granularity
 * - moves and overlaps blocks that overlap with multiple values in the overlap region
 *
 * It does not
 * - try to move and overlap blocks that are not already adjacent
 */
static void
utrie_compact(UNewTrie *trie, UBool overlap, UErrorCode *pErrorCode) {
    int32_t i, start, newStart, overlapStart;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return;
    }

    /* valid, uncompacted trie? */
    if(trie==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if(trie->isCompacted) {
        return; /* nothing left to do */
    }

    /* compaction */

    /* initialize the index map with "block is used/unused" flags */
    _findUnusedBlocks(trie);

    /* if Latin-1 is preallocated and linear, then do not compact Latin-1 data */
    if(trie->isLatin1Linear && UTRIE_SHIFT<=8) {
        overlapStart=UTRIE_DATA_BLOCK_LENGTH+256;
    } else {
        overlapStart=UTRIE_DATA_BLOCK_LENGTH;
    }

    newStart=UTRIE_DATA_BLOCK_LENGTH;
    for(start=newStart; start<trie->dataLength;) {
        /*
         * start: index of first entry of current block
         * newStart: index where the current block is to be moved
         *           (right after current end of already-compacted data)
         */

        /* skip blocks that are not used */
        if(trie->map[start>>UTRIE_SHIFT]<0) {
            /* advance start to the next block */
            start+=UTRIE_DATA_BLOCK_LENGTH;

            /* leave newStart with the previous block! */
            continue;
        }

        /* search for an identical block */
        if( start>=overlapStart &&
            (i=_findSameDataBlock(trie->data, newStart, start,
                            overlap ? UTRIE_DATA_GRANULARITY : UTRIE_DATA_BLOCK_LENGTH))
             >=0
        ) {
            /* found an identical block, set the other block's index value for the current block */
            trie->map[start>>UTRIE_SHIFT]=i;

            /* advance start to the next block */
            start+=UTRIE_DATA_BLOCK_LENGTH;

            /* leave newStart with the previous block! */
            continue;
        }

        /* see if the beginning of this block can be overlapped with the end of the previous block */
        if(overlap && start>=overlapStart) {
            /* look for maximum overlap (modulo granularity) with the previous, adjacent block */
            for(i=UTRIE_DATA_BLOCK_LENGTH-UTRIE_DATA_GRANULARITY;
                i>0 && !utrie_equal_uint32(trie->data+(newStart-i), trie->data+start, i);
                i-=UTRIE_DATA_GRANULARITY) {}
        } else {
            i=0;
        }

        if(i>0) {
            /* some overlap */
            trie->map[start>>UTRIE_SHIFT]=newStart-i;

            /* move the non-overlapping indexes to their new positions */
            start+=i;
            for(i=UTRIE_DATA_BLOCK_LENGTH-i; i>0; --i) {
                trie->data[newStart++]=trie->data[start++];
            }
        } else if(newStart<start) {
            /* no overlap, just move the indexes to their new positions */
            trie->map[start>>UTRIE_SHIFT]=newStart;
            for(i=UTRIE_DATA_BLOCK_LENGTH; i>0; --i) {
                trie->data[newStart++]=trie->data[start++];
            }
        } else /* no overlap && newStart==start */ {
            trie->map[start>>UTRIE_SHIFT]=start;
            newStart+=UTRIE_DATA_BLOCK_LENGTH;
            start=newStart;
        }
    }

    /* now adjust the index (stage 1) table */
    for(i=0; i<trie->indexLength; ++i) {
        trie->index[i]=trie->map[ABS(trie->index[i])>>UTRIE_SHIFT];
    }

#ifdef UTRIE_DEBUG
    /* we saved some space */
    printf("compacting trie: count of 32-bit words %lu->%lu\n",
            (long)trie->dataLength, (long)newStart);
#endif

    trie->dataLength=newStart;
}

/* serialization ------------------------------------------------------------ */

/*
 * Default function for the folding value:
 * Just store the offset (16 bits) if there is any non-initial-value entry.
 *
 * The offset parameter is never 0.
 * Returning the offset itself is safe for UTRIE_SHIFT>=5 because
 * for UTRIE_SHIFT==5 the maximum index length is UTRIE_MAX_INDEX_LENGTH==0x8800
 * which fits into 16-bit trie values;
 * for higher UTRIE_SHIFT, UTRIE_MAX_INDEX_LENGTH decreases.
 *
 * Theoretically, it would be safer for all possible UTRIE_SHIFT including
 * those of 4 and lower to return offset>>UTRIE_SURROGATE_BLOCK_BITS
 * which would always result in a value of 0x40..0x43f
 * (start/end 1k blocks of supplementary Unicode code points).
 * However, this would be uglier, and would not work for some existing
 * binary data file formats.
 *
 * Also, we do not plan to change UTRIE_SHIFT because it would change binary
 * data file formats, and we would probably not make it smaller because of
 * the then even larger BMP index length even for empty tries.
 */
static uint32_t U_CALLCONV
defaultGetFoldedValue(UNewTrie *trie, UChar32 start, int32_t offset) {
    uint32_t value, initialValue;
    UChar32 limit;
    UBool inBlockZero;

    initialValue=trie->data[0];
    limit=start+0x400;
    while(start<limit) {
        value=utrie_get32(trie, start, &inBlockZero);
        if(inBlockZero) {
            start+=UTRIE_DATA_BLOCK_LENGTH;
        } else if(value!=initialValue) {
            return (uint32_t)offset;
        } else {
            ++start;
        }
    }
    return 0;
}

U_CAPI int32_t U_EXPORT2
utrie_serialize(UNewTrie *trie, void *dt, int32_t capacity,
                UNewTrieGetFoldedValue *getFoldedValue,
                UBool reduceTo16Bits,
                UErrorCode *pErrorCode) {
    UTrieHeader *header;
    uint32_t *p;
    uint16_t *dest16;
    int32_t i, length;
    uint8_t* data = NULL;

    /* argument check */
    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }

    if(trie==NULL || capacity<0 || (capacity>0 && dt==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    if(getFoldedValue==NULL) {
        getFoldedValue=defaultGetFoldedValue;
    }

    data = (uint8_t*)dt;
    /* fold and compact if necessary, also checks that indexLength is within limits */
    if(!trie->isCompacted) {
        /* compact once without overlap to improve folding */
        utrie_compact(trie, FALSE, pErrorCode);

        /* fold the supplementary part of the index array */
        utrie_fold(trie, getFoldedValue, pErrorCode);

        /* compact again with overlap for minimum data array length */
        utrie_compact(trie, TRUE, pErrorCode);

        trie->isCompacted=TRUE;
        if(U_FAILURE(*pErrorCode)) {
            return 0;
        }
    }

    /* is dataLength within limits? */
    if( (reduceTo16Bits ? (trie->dataLength+trie->indexLength) : trie->dataLength) >= UTRIE_MAX_DATA_LENGTH) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
    }

    length=sizeof(UTrieHeader)+2*trie->indexLength;
    if(reduceTo16Bits) {
        length+=2*trie->dataLength;
    } else {
        length+=4*trie->dataLength;
    }

    if(length>capacity) {
        return length; /* preflighting */
    }

#ifdef UTRIE_DEBUG
    printf("**UTrieLengths(serialize)** index:%6ld  data:%6ld  serialized:%6ld\n",
           (long)trie->indexLength, (long)trie->dataLength, (long)length);
#endif

    /* set the header fields */
    header=(UTrieHeader *)data;
    data+=sizeof(UTrieHeader);

    header->signature=0x54726965; /* "Trie" */
    header->options=UTRIE_SHIFT | (UTRIE_INDEX_SHIFT<<UTRIE_OPTIONS_INDEX_SHIFT);

    if(!reduceTo16Bits) {
        header->options|=UTRIE_OPTIONS_DATA_IS_32_BIT;
    }
    if(trie->isLatin1Linear) {
        header->options|=UTRIE_OPTIONS_LATIN1_IS_LINEAR;
    }

    header->indexLength=trie->indexLength;
    header->dataLength=trie->dataLength;

    /* write the index (stage 1) array and the 16/32-bit data (stage 2) array */
    if(reduceTo16Bits) {
        /* write 16-bit index values shifted right by UTRIE_INDEX_SHIFT, after adding indexLength */
        p=(uint32_t *)trie->index;
        dest16=(uint16_t *)data;
        for(i=trie->indexLength; i>0; --i) {
            *dest16++=(uint16_t)((*p++ + trie->indexLength)>>UTRIE_INDEX_SHIFT);
        }

        /* write 16-bit data values */
        p=trie->data;
        for(i=trie->dataLength; i>0; --i) {
            *dest16++=(uint16_t)*p++;
        }
    } else {
        /* write 16-bit index values shifted right by UTRIE_INDEX_SHIFT */
        p=(uint32_t *)trie->index;
        dest16=(uint16_t *)data;
        for(i=trie->indexLength; i>0; --i) {
            *dest16++=(uint16_t)(*p++ >> UTRIE_INDEX_SHIFT);
        }

        /* write 32-bit data values */
        uprv_memcpy(dest16, trie->data, 4*(size_t)trie->dataLength);
    }

    return length;
}

/* inverse to defaultGetFoldedValue() */
U_CAPI int32_t U_EXPORT2
utrie_defaultGetFoldingOffset(uint32_t data) {
    return (int32_t)data;
}

U_CAPI int32_t U_EXPORT2
utrie_unserialize(UTrie *trie, const void *data, int32_t length, UErrorCode *pErrorCode) {
    const UTrieHeader *header;
    const uint16_t *p16;
    uint32_t options;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return -1;
    }

    /* enough data for a trie header? */
    if(length<(int32_t)sizeof(UTrieHeader)) {
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return -1;
    }

    /* check the signature */
    header=(const UTrieHeader *)data;
    if(header->signature!=0x54726965) {
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return -1;
    }

    /* get the options and check the shift values */
    options=header->options;
    if( (options&UTRIE_OPTIONS_SHIFT_MASK)!=UTRIE_SHIFT ||
        ((options>>UTRIE_OPTIONS_INDEX_SHIFT)&UTRIE_OPTIONS_SHIFT_MASK)!=UTRIE_INDEX_SHIFT
    ) {
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return -1;
    }
    trie->isLatin1Linear= (UBool)((options&UTRIE_OPTIONS_LATIN1_IS_LINEAR)!=0);

    /* get the length values */
    trie->indexLength=header->indexLength;
    trie->dataLength=header->dataLength;

    length-=(int32_t)sizeof(UTrieHeader);

    /* enough data for the index? */
    if(length<2*trie->indexLength) {
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return -1;
    }
    p16=(const uint16_t *)(header+1);
    trie->index=p16;
    p16+=trie->indexLength;
    length-=2*trie->indexLength;

    /* get the data */
    if(options&UTRIE_OPTIONS_DATA_IS_32_BIT) {
        if(length<4*trie->dataLength) {
            *pErrorCode=U_INVALID_FORMAT_ERROR;
            return -1;
        }
        trie->data32=(const uint32_t *)p16;
        trie->initialValue=trie->data32[0];
        length=(int32_t)sizeof(UTrieHeader)+2*trie->indexLength+4*trie->dataLength;
    } else {
        if(length<2*trie->dataLength) {
            *pErrorCode=U_INVALID_FORMAT_ERROR;
            return -1;
        }

        /* the "data16" data is used via the index pointer */
        trie->data32=NULL;
        trie->initialValue=trie->index[trie->indexLength];
        length=(int32_t)sizeof(UTrieHeader)+2*trie->indexLength+2*trie->dataLength;
    }

    trie->getFoldingOffset=utrie_defaultGetFoldingOffset;

    return length;
}

U_CAPI int32_t U_EXPORT2
utrie_unserializeDummy(UTrie *trie,
                       void *data, int32_t length,
                       uint32_t initialValue, uint32_t leadUnitValue,
                       UBool make16BitTrie,
                       UErrorCode *pErrorCode) {
    uint16_t *p16;
    int32_t actualLength, latin1Length, i, limit;
    uint16_t block;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return -1;
    }

    /* calculate the actual size of the dummy trie data */

    /* max(Latin-1, block 0) */
    latin1Length= 256; /*UTRIE_SHIFT<=8 ? 256 : UTRIE_DATA_BLOCK_LENGTH;*/

    trie->indexLength=UTRIE_BMP_INDEX_LENGTH+UTRIE_SURROGATE_BLOCK_COUNT;
    trie->dataLength=latin1Length;
    if(leadUnitValue!=initialValue) {
        trie->dataLength+=UTRIE_DATA_BLOCK_LENGTH;
    }

    actualLength=trie->indexLength*2;
    if(make16BitTrie) {
        actualLength+=trie->dataLength*2;
    } else {
        actualLength+=trie->dataLength*4;
    }

    /* enough space for the dummy trie? */
    if(length<actualLength) {
        *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
        return actualLength;
    }

    trie->isLatin1Linear=TRUE;
    trie->initialValue=initialValue;

    /* fill the index and data arrays */
    p16=(uint16_t *)data;
    trie->index=p16;

    if(make16BitTrie) {
        /* indexes to block 0 */
        block=(uint16_t)(trie->indexLength>>UTRIE_INDEX_SHIFT);
        limit=trie->indexLength;
        for(i=0; i<limit; ++i) {
            p16[i]=block;
        }

        if(leadUnitValue!=initialValue) {
            /* indexes for lead surrogate code units to the block after Latin-1 */
            block+=(uint16_t)(latin1Length>>UTRIE_INDEX_SHIFT);
            i=0xd800>>UTRIE_SHIFT;
            limit=0xdc00>>UTRIE_SHIFT;
            for(; i<limit; ++i) {
                p16[i]=block;
            }
        }

        trie->data32=NULL;

        /* Latin-1 data */
        p16+=trie->indexLength;
        for(i=0; i<latin1Length; ++i) {
            p16[i]=(uint16_t)initialValue;
        }

        /* data for lead surrogate code units */
        if(leadUnitValue!=initialValue) {
            limit=latin1Length+UTRIE_DATA_BLOCK_LENGTH;
            for(/* i=latin1Length */; i<limit; ++i) {
                p16[i]=(uint16_t)leadUnitValue;
            }
        }
    } else {
        uint32_t *p32;

        /* indexes to block 0 */
        uprv_memset(p16, 0, trie->indexLength*2);

        if(leadUnitValue!=initialValue) {
            /* indexes for lead surrogate code units to the block after Latin-1 */
            block=(uint16_t)(latin1Length>>UTRIE_INDEX_SHIFT);
            i=0xd800>>UTRIE_SHIFT;
            limit=0xdc00>>UTRIE_SHIFT;
            for(; i<limit; ++i) {
                p16[i]=block;
            }
        }

        trie->data32=p32=(uint32_t *)(p16+trie->indexLength);

        /* Latin-1 data */
        for(i=0; i<latin1Length; ++i) {
            p32[i]=initialValue;
        }

        /* data for lead surrogate code units */
        if(leadUnitValue!=initialValue) {
            limit=latin1Length+UTRIE_DATA_BLOCK_LENGTH;
            for(/* i=latin1Length */; i<limit; ++i) {
                p32[i]=leadUnitValue;
            }
        }
    }

    trie->getFoldingOffset=utrie_defaultGetFoldingOffset;

    return actualLength;
}

/* enumeration -------------------------------------------------------------- */

/* default UTrieEnumValue() returns the input value itself */
static uint32_t U_CALLCONV
utrie_enumSameValue(const void * /*context*/, uint32_t value) {
    return value;
}

/**
 * Enumerate all ranges of code points with the same relevant values.
 * The values are transformed from the raw trie entries by the enumValue function.
 */
U_CAPI void U_EXPORT2
utrie_enum(const UTrie *trie,
           UTrieEnumValue *enumValue, UTrieEnumRange *enumRange, const void *context) {
    const uint32_t *data32;
    const uint16_t *idx;

    uint32_t value, prevValue, initialValue;
    UChar32 c, prev;
    int32_t l, i, j, block, prevBlock, nullBlock, offset;

    /* check arguments */
    if(trie==NULL || trie->index==NULL || enumRange==NULL) {
        return;
    }
    if(enumValue==NULL) {
        enumValue=utrie_enumSameValue;
    }

    idx=trie->index;
    data32=trie->data32;

    /* get the enumeration value that corresponds to an initial-value trie data entry */
    initialValue=enumValue(context, trie->initialValue);

    if(data32==NULL) {
        nullBlock=trie->indexLength;
    } else {
        nullBlock=0;
    }

    /* set variables for previous range */
    prevBlock=nullBlock;
    prev=0;
    prevValue=initialValue;

    /* enumerate BMP - the main loop enumerates data blocks */
    for(i=0, c=0; c<=0xffff; ++i) {
        if(c==0xd800) {
            /* skip lead surrogate code _units_, go to lead surr. code _points_ */
            i=UTRIE_BMP_INDEX_LENGTH;
        } else if(c==0xdc00) {
            /* go back to regular BMP code points */
            i=c>>UTRIE_SHIFT;
        }

        block=idx[i]<<UTRIE_INDEX_SHIFT;
        if(block==prevBlock) {
            /* the block is the same as the previous one, and filled with value */
            c+=UTRIE_DATA_BLOCK_LENGTH;
        } else if(block==nullBlock) {
            /* this is the all-initial-value block */
            if(prevValue!=initialValue) {
                if(prev<c) {
                    if(!enumRange(context, prev, c, prevValue)) {
                        return;
                    }
                }
                prevBlock=nullBlock;
                prev=c;
                prevValue=initialValue;
            }
            c+=UTRIE_DATA_BLOCK_LENGTH;
        } else {
            prevBlock=block;
            for(j=0; j<UTRIE_DATA_BLOCK_LENGTH; ++j) {
                value=enumValue(context, data32!=NULL ? data32[block+j] : idx[block+j]);
                if(value!=prevValue) {
                    if(prev<c) {
                        if(!enumRange(context, prev, c, prevValue)) {
                            return;
                        }
                    }
                    if(j>0) {
                        /* the block is not filled with all the same value */
                        prevBlock=-1;
                    }
                    prev=c;
                    prevValue=value;
                }
                ++c;
            }
        }
    }

    /* enumerate supplementary code points */
    for(l=0xd800; l<0xdc00;) {
        /* lead surrogate access */
        offset=idx[l>>UTRIE_SHIFT]<<UTRIE_INDEX_SHIFT;
        if(offset==nullBlock) {
            /* no entries for a whole block of lead surrogates */
            if(prevValue!=initialValue) {
                if(prev<c) {
                    if(!enumRange(context, prev, c, prevValue)) {
                        return;
                    }
                }
                prevBlock=nullBlock;
                prev=c;
                prevValue=initialValue;
            }

            l+=UTRIE_DATA_BLOCK_LENGTH;
            c+=UTRIE_DATA_BLOCK_LENGTH<<10;
            continue;
        }

        value= data32!=NULL ? data32[offset+(l&UTRIE_MASK)] : idx[offset+(l&UTRIE_MASK)];

        /* enumerate trail surrogates for this lead surrogate */
        offset=trie->getFoldingOffset(value);
        if(offset<=0) {
            /* no data for this lead surrogate */
            if(prevValue!=initialValue) {
                if(prev<c) {
                    if(!enumRange(context, prev, c, prevValue)) {
                        return;
                    }
                }
                prevBlock=nullBlock;
                prev=c;
                prevValue=initialValue;
            }

            /* nothing else to do for the supplementary code points for this lead surrogate */
            c+=0x400;
        } else {
            /* enumerate code points for this lead surrogate */
            i=offset;
            offset+=UTRIE_SURROGATE_BLOCK_COUNT;
            do {
                /* copy of most of the body of the BMP loop */
                block=idx[i]<<UTRIE_INDEX_SHIFT;
                if(block==prevBlock) {
                    /* the block is the same as the previous one, and filled with value */
                    c+=UTRIE_DATA_BLOCK_LENGTH;
                } else if(block==nullBlock) {
                    /* this is the all-initial-value block */
                    if(prevValue!=initialValue) {
                        if(prev<c) {
                            if(!enumRange(context, prev, c, prevValue)) {
                                return;
                            }
                        }
                        prevBlock=nullBlock;
                        prev=c;
                        prevValue=initialValue;
                    }
                    c+=UTRIE_DATA_BLOCK_LENGTH;
                } else {
                    prevBlock=block;
                    for(j=0; j<UTRIE_DATA_BLOCK_LENGTH; ++j) {
                        value=enumValue(context, data32!=NULL ? data32[block+j] : idx[block+j]);
                        if(value!=prevValue) {
                            if(prev<c) {
                                if(!enumRange(context, prev, c, prevValue)) {
                                    return;
                                }
                            }
                            if(j>0) {
                                /* the block is not filled with all the same value */
                                prevBlock=-1;
                            }
                            prev=c;
                            prevValue=value;
                        }
                        ++c;
                    }
                }
            } while(++i<offset);
        }

        ++l;
    }

    /* deliver last range */
    enumRange(context, prev, c, prevValue);
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2001-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  utrie2.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2008aug16 (starting from a copy of utrie.c)
*   created by: Markus W. Scherer
*
*   This is a common implementation of a Unicode trie.
*   It is a kind of compressed, serializable table of 16- or 32-bit values associated with
*   Unicode code points (0..0x10ffff).
*   This is the second common version of a Unicode trie (hence the name UTrie2).
*   See utrie2.h for a comparison.
*
*   This file contains only the runtime and enumeration code, for read-only access.
*   See utrie2_builder.c for the builder code.
*/

#ifdef UCPTRIE_DEBUG

#endif





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2001-2008, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  utrie2_impl.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2008sep26 (split off from utrie2.c)
*   created by: Markus W. Scherer
*
*   Definitions needed for both runtime and builder code for UTrie2,
*   used by utrie2.c and utrie2_builder.c.
*/


#ifdef UCPTRIE_DEBUG

#endif


/* Public UTrie2 API implementation ----------------------------------------- */

/*
 * These definitions are mostly needed by utrie2.cpp,
 * but also by utrie2_serialize() and utrie2_swap().
 */

// UTrie2 signature values, in platform endianness and opposite endianness.
// The UTrie2 signature ASCII byte values spell "Tri2".
#define UTRIE2_SIG      0x54726932
#define UTRIE2_OE_SIG   0x32697254

/**
 * Trie data structure in serialized form:
 *
 * UTrie2Header header;
 * uint16_t index[header.index2Length];
 * uint16_t data[header.shiftedDataLength<<2];  -- or uint32_t data[...]
 * @internal
 */
typedef struct UTrie2Header {
    /** "Tri2" in big-endian US-ASCII (0x54726932) */
    uint32_t signature;

    /**
     * options bit field:
     * 15.. 4   reserved (0)
     *  3.. 0   UTrie2ValueBits valueBits
     */
    uint16_t options;

    /** UTRIE2_INDEX_1_OFFSET..UTRIE2_MAX_INDEX_LENGTH */
    uint16_t indexLength;

    /** (UTRIE2_DATA_START_OFFSET..UTRIE2_MAX_DATA_LENGTH)>>UTRIE2_INDEX_SHIFT */
    uint16_t shiftedDataLength;

    /** Null index and data blocks, not shifted. */
    uint16_t index2NullOffset, dataNullOffset;

    /**
     * First code point of the single-value range ending with U+10ffff,
     * rounded up and then shifted right by UTRIE2_SHIFT_1.
     */
    uint16_t shiftedHighStart;
} UTrie2Header;

/**
 * Constants for use with UTrie2Header.options.
 * @internal
 */
enum {
    /** Mask to get the UTrie2ValueBits valueBits from options. */
    UTRIE2_OPTIONS_VALUE_BITS_MASK=0xf
};

/* Building a trie ---------------------------------------------------------- */

/*
 * These definitions are mostly needed by utrie2_builder.c, but also by
 * utrie2_get32() and utrie2_enum().
 */

enum {
    /**
     * At build time, leave a gap in the index-2 table,
     * at least as long as the maximum lengths of the 2-byte UTF-8 index-2 table
     * and the supplementary index-1 table.
     * Round up to UTRIE2_INDEX_2_BLOCK_LENGTH for proper compacting.
     */
    UNEWTRIE2_INDEX_GAP_OFFSET=UTRIE2_INDEX_2_BMP_LENGTH,
    UNEWTRIE2_INDEX_GAP_LENGTH=
        ((UTRIE2_UTF8_2B_INDEX_2_LENGTH+UTRIE2_MAX_INDEX_1_LENGTH)+UTRIE2_INDEX_2_MASK)&
        ~UTRIE2_INDEX_2_MASK,

    /**
     * Maximum length of the build-time index-2 array.
     * Maximum number of Unicode code points (0x110000) shifted right by UTRIE2_SHIFT_2,
     * plus the part of the index-2 table for lead surrogate code points,
     * plus the build-time index gap,
     * plus the null index-2 block.
     */
    UNEWTRIE2_MAX_INDEX_2_LENGTH=
        (0x110000>>UTRIE2_SHIFT_2)+
        UTRIE2_LSCP_INDEX_2_LENGTH+
        UNEWTRIE2_INDEX_GAP_LENGTH+
        UTRIE2_INDEX_2_BLOCK_LENGTH,

    UNEWTRIE2_INDEX_1_LENGTH=0x110000>>UTRIE2_SHIFT_1
};

/**
 * Maximum length of the build-time data array.
 * One entry per 0x110000 code points, plus the illegal-UTF-8 block and the null block,
 * plus values for the 0x400 surrogate code units.
 */
#define UNEWTRIE2_MAX_DATA_LENGTH (0x110000+0x40+0x40+0x400)

/*
 * Build-time trie structure.
 *
 * Just using a boolean flag for "repeat use" could lead to data array overflow
 * because we would not be able to detect when a data block becomes unused.
 * It also leads to orphan data blocks that are kept through serialization.
 *
 * Need to use reference counting for data blocks,
 * and allocDataBlock() needs to look for a free block before increasing dataLength.
 *
 * This scheme seems like overkill for index-2 blocks since the whole index array is
 * preallocated anyway (unlike the growable data array).
 * Just allocating multiple index-2 blocks as needed.
 */
struct UNewTrie2 {
    int32_t index1[UNEWTRIE2_INDEX_1_LENGTH];
    int32_t index2[UNEWTRIE2_MAX_INDEX_2_LENGTH];
    uint32_t *data;
#ifdef UCPTRIE_DEBUG
    UMutableCPTrie *t3;
#endif

    uint32_t initialValue, errorValue;
    int32_t index2Length, dataCapacity, dataLength;
    int32_t firstFreeBlock;
    int32_t index2NullOffset, dataNullOffset;
    UChar32 highStart;
    UBool isCompacted;

    /**
     * Multi-purpose per-data-block table.
     *
     * Before compacting:
     *
     * Per-data-block reference counters/free-block list.
     *  0: unused
     * >0: reference counter (number of index-2 entries pointing here)
     * <0: next free data block in free-block list
     *
     * While compacting:
     *
     * Map of adjusted indexes, used in compactData() and compactIndex2().
     * Maps from original indexes to new ones.
     */
    int32_t map[UNEWTRIE2_MAX_DATA_LENGTH>>UTRIE2_SHIFT_2];
};




/* Public UTrie2 API implementation ----------------------------------------- */

static uint32_t
get32(const UNewTrie2 *trie, UChar32 c, UBool fromLSCP) {
    int32_t i2, block;

    if(c>=trie->highStart && (!U_IS_LEAD(c) || fromLSCP)) {
        return trie->data[trie->dataLength-UTRIE2_DATA_GRANULARITY];
    }

    if(U_IS_LEAD(c) && fromLSCP) {
        i2=(UTRIE2_LSCP_INDEX_2_OFFSET-(0xd800>>UTRIE2_SHIFT_2))+
            (c>>UTRIE2_SHIFT_2);
    } else {
        i2=trie->index1[c>>UTRIE2_SHIFT_1]+
            ((c>>UTRIE2_SHIFT_2)&UTRIE2_INDEX_2_MASK);
    }
    block=trie->index2[i2];
    return trie->data[block+(c&UTRIE2_DATA_MASK)];
}

U_CAPI uint32_t U_EXPORT2
utrie2_get32(const UTrie2 *trie, UChar32 c) {
    if(trie->data16!=NULL) {
        return UTRIE2_GET16(trie, c);
    } else if(trie->data32!=NULL) {
        return UTRIE2_GET32(trie, c);
    } else if((uint32_t)c>0x10ffff) {
        return trie->errorValue;
    } else {
        return get32(trie->newTrie, c, TRUE);
    }
}

U_CAPI uint32_t U_EXPORT2
utrie2_get32FromLeadSurrogateCodeUnit(const UTrie2 *trie, UChar32 c) {
    if(!U_IS_LEAD(c)) {
        return trie->errorValue;
    }
    if(trie->data16!=NULL) {
        return UTRIE2_GET16_FROM_U16_SINGLE_LEAD(trie, c);
    } else if(trie->data32!=NULL) {
        return UTRIE2_GET32_FROM_U16_SINGLE_LEAD(trie, c);
    } else {
        return get32(trie->newTrie, c, FALSE);
    }
}

static inline int32_t
u8Index(const UTrie2 *trie, UChar32 c, int32_t i) {
    int32_t idx=
        _UTRIE2_INDEX_FROM_CP(
            trie,
            trie->data32==NULL ? trie->indexLength : 0,
            c);
    return (idx<<3)|i;
}

U_CAPI int32_t U_EXPORT2
utrie2_internalU8NextIndex(const UTrie2 *trie, UChar32 c,
                           const uint8_t *src, const uint8_t *limit) {
    int32_t i, length;
    i=0;
    /* support 64-bit pointers by avoiding cast of arbitrary difference */
    if((limit-src)<=7) {
        length=(int32_t)(limit-src);
    } else {
        length=7;
    }
    c=utf8_nextCharSafeBody(src, &i, length, c, -1);
    return u8Index(trie, c, i);
}

U_CAPI int32_t U_EXPORT2
utrie2_internalU8PrevIndex(const UTrie2 *trie, UChar32 c,
                           const uint8_t *start, const uint8_t *src) {
    int32_t i, length;
    /* support 64-bit pointers by avoiding cast of arbitrary difference */
    if((src-start)<=7) {
        i=length=(int32_t)(src-start);
    } else {
        i=length=7;
        start=src-7;
    }
    c=utf8_prevCharSafeBody(start, 0, &i, c, -1);
    i=length-i;  /* number of bytes read backward from src */
    return u8Index(trie, c, i);
}

U_CAPI UTrie2 * U_EXPORT2
utrie2_openFromSerialized(UTrie2ValueBits valueBits,
                          const void *data, int32_t length, int32_t *pActualLength,
                          UErrorCode *pErrorCode) {
    const UTrie2Header *header;
    const uint16_t *p16;
    int32_t actualLength;

    UTrie2 tempTrie;
    UTrie2 *trie;

    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }

    // if( length<=0 || (U_POINTER_MASK_LSB(data, 3)!=0) ||
    //     valueBits<0 || UTRIE2_COUNT_VALUE_BITS<=valueBits
    // ) {
    //     *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
    //     return 0;
    // }

    /* enough data for a trie header? */
    if(length<(int32_t)sizeof(UTrie2Header)) {
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return 0;
    }

    /* check the signature */
    header=(const UTrie2Header *)data;
    if(header->signature!=UTRIE2_SIG) {
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return 0;
    }

    /* get the options */
    if(valueBits!=(UTrie2ValueBits)(header->options&UTRIE2_OPTIONS_VALUE_BITS_MASK)) {
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return 0;
    }

    /* get the length values and offsets */
    uprv_memset(&tempTrie, 0, sizeof(tempTrie));
    tempTrie.indexLength=header->indexLength;
    tempTrie.dataLength=header->shiftedDataLength<<UTRIE2_INDEX_SHIFT;
    tempTrie.index2NullOffset=header->index2NullOffset;
    tempTrie.dataNullOffset=header->dataNullOffset;

    tempTrie.highStart=header->shiftedHighStart<<UTRIE2_SHIFT_1;
    tempTrie.highValueIndex=tempTrie.dataLength-UTRIE2_DATA_GRANULARITY;
    if(valueBits==UTRIE2_16_VALUE_BITS) {
        tempTrie.highValueIndex+=tempTrie.indexLength;
    }

    /* calculate the actual length */
    actualLength=(int32_t)sizeof(UTrie2Header)+tempTrie.indexLength*2;
    if(valueBits==UTRIE2_16_VALUE_BITS) {
        actualLength+=tempTrie.dataLength*2;
    } else {
        actualLength+=tempTrie.dataLength*4;
    }
    if(length<actualLength) {
        *pErrorCode=U_INVALID_FORMAT_ERROR;  /* not enough bytes */
        return 0;
    }

    /* allocate the trie */
    trie=(UTrie2 *)uprv_malloc(sizeof(UTrie2));
    if(trie==NULL) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    uprv_memcpy(trie, &tempTrie, sizeof(tempTrie));
    trie->memory=(uint32_t *)data;
    trie->length=actualLength;
    trie->isMemoryOwned=FALSE;
#ifdef UTRIE2_DEBUG
    trie->name="fromSerialized";
#endif

    /* set the pointers to its index and data arrays */
    p16=(const uint16_t *)(header+1);
    trie->index=p16;
    p16+=trie->indexLength;

    /* get the data */
    switch(valueBits) {
    case UTRIE2_16_VALUE_BITS:
        trie->data16=p16;
        trie->data32=NULL;
        trie->initialValue=trie->index[trie->dataNullOffset];
        trie->errorValue=trie->data16[UTRIE2_BAD_UTF8_DATA_OFFSET];
        break;
    case UTRIE2_32_VALUE_BITS:
        trie->data16=NULL;
        trie->data32=(const uint32_t *)p16;
        trie->initialValue=trie->data32[trie->dataNullOffset];
        trie->errorValue=trie->data32[UTRIE2_BAD_UTF8_DATA_OFFSET];
        break;
    default:
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return 0;
    }

    if(pActualLength!=NULL) {
        *pActualLength=actualLength;
    }
    return trie;
}

U_CAPI UTrie2 * U_EXPORT2
utrie2_openDummy(UTrie2ValueBits valueBits,
                 uint32_t initialValue, uint32_t errorValue,
                 UErrorCode *pErrorCode) {
    UTrie2 *trie;
    UTrie2Header *header;
    uint32_t *p;
    uint16_t *dest16;
    int32_t indexLength, dataLength, length, i;
    int32_t dataMove;  /* >0 if the data is moved to the end of the index array */

    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }

    if(valueBits<0 || UTRIE2_COUNT_VALUE_BITS<=valueBits) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* calculate the total length of the dummy trie data */
    indexLength=UTRIE2_INDEX_1_OFFSET;
    dataLength=UTRIE2_DATA_START_OFFSET+UTRIE2_DATA_GRANULARITY;
    length=(int32_t)sizeof(UTrie2Header)+indexLength*2;
    if(valueBits==UTRIE2_16_VALUE_BITS) {
        length+=dataLength*2;
    } else {
        length+=dataLength*4;
    }

    /* allocate the trie */
    trie=(UTrie2 *)uprv_malloc(sizeof(UTrie2));
    if(trie==NULL) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    uprv_memset(trie, 0, sizeof(UTrie2));
    trie->memory=uprv_malloc(length);
    if(trie->memory==NULL) {
        uprv_free(trie);
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    trie->length=length;
    trie->isMemoryOwned=TRUE;

    /* set the UTrie2 fields */
    if(valueBits==UTRIE2_16_VALUE_BITS) {
        dataMove=indexLength;
    } else {
        dataMove=0;
    }

    trie->indexLength=indexLength;
    trie->dataLength=dataLength;
    trie->index2NullOffset=UTRIE2_INDEX_2_OFFSET;
    trie->dataNullOffset=(uint16_t)dataMove;
    trie->initialValue=initialValue;
    trie->errorValue=errorValue;
    trie->highStart=0;
    trie->highValueIndex=dataMove+UTRIE2_DATA_START_OFFSET;
#ifdef UTRIE2_DEBUG
    trie->name="dummy";
#endif

    /* set the header fields */
    header=(UTrie2Header *)trie->memory;

    header->signature=UTRIE2_SIG; /* "Tri2" */
    header->options=(uint16_t)valueBits;

    header->indexLength=(uint16_t)indexLength;
    header->shiftedDataLength=(uint16_t)(dataLength>>UTRIE2_INDEX_SHIFT);
    header->index2NullOffset=(uint16_t)UTRIE2_INDEX_2_OFFSET;
    header->dataNullOffset=(uint16_t)dataMove;
    header->shiftedHighStart=0;

    /* fill the index and data arrays */
    dest16=(uint16_t *)(header+1);
    trie->index=dest16;

    /* write the index-2 array values shifted right by UTRIE2_INDEX_SHIFT */
    for(i=0; i<UTRIE2_INDEX_2_BMP_LENGTH; ++i) {
        *dest16++=(uint16_t)(dataMove>>UTRIE2_INDEX_SHIFT);  /* null data block */
    }

    /* write UTF-8 2-byte index-2 values, not right-shifted */
    for(i=0; i<(0xc2-0xc0); ++i) {                                  /* C0..C1 */
        *dest16++=(uint16_t)(dataMove+UTRIE2_BAD_UTF8_DATA_OFFSET);
    }
    for(; i<(0xe0-0xc0); ++i) {                                     /* C2..DF */
        *dest16++=(uint16_t)dataMove;
    }

    /* write the 16/32-bit data array */
    switch(valueBits) {
    case UTRIE2_16_VALUE_BITS:
        /* write 16-bit data values */
        trie->data16=dest16;
        trie->data32=NULL;
        for(i=0; i<0x80; ++i) {
            *dest16++=(uint16_t)initialValue;
        }
        for(; i<0xc0; ++i) {
            *dest16++=(uint16_t)errorValue;
        }
        /* highValue and reserved values */
        for(i=0; i<UTRIE2_DATA_GRANULARITY; ++i) {
            *dest16++=(uint16_t)initialValue;
        }
        break;
    case UTRIE2_32_VALUE_BITS:
        /* write 32-bit data values */
        p=(uint32_t *)dest16;
        trie->data16=NULL;
        trie->data32=p;
        for(i=0; i<0x80; ++i) {
            *p++=initialValue;
        }
        for(; i<0xc0; ++i) {
            *p++=errorValue;
        }
        /* highValue and reserved values */
        for(i=0; i<UTRIE2_DATA_GRANULARITY; ++i) {
            *p++=initialValue;
        }
        break;
    default:
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    return trie;
}

U_CAPI void U_EXPORT2
utrie2_close(UTrie2 *trie) {
    if(trie!=NULL) {
        if(trie->isMemoryOwned) {
            uprv_free(trie->memory);
        }
        if(trie->newTrie!=NULL) {
            uprv_free(trie->newTrie->data);
#ifdef UCPTRIE_DEBUG
            umutablecptrie_close(trie->newTrie->t3);
#endif
            uprv_free(trie->newTrie);
        }
        uprv_free(trie);
    }
}

U_CAPI UBool U_EXPORT2
utrie2_isFrozen(const UTrie2 *trie) {
    return (UBool)(trie->newTrie==NULL);
}

U_CAPI int32_t U_EXPORT2
utrie2_serialize(const UTrie2 *trie,
                 void *data, int32_t capacity,
                 UErrorCode *pErrorCode) {
    /* argument check */
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }

    if( trie==NULL || trie->memory==NULL || trie->newTrie!=NULL ||
        capacity<0 || (capacity>0 && (data==NULL || (U_POINTER_MASK_LSB(data, 3)!=0)))
    ) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if(capacity>=trie->length) {
        uprv_memcpy(data, trie->memory, trie->length);
    } else {
        *pErrorCode=U_BUFFER_OVERFLOW_ERROR;
    }
    return trie->length;
}

/* enumeration -------------------------------------------------------------- */

#define MIN_VALUE(a, b) ((a)<(b) ? (a) : (b))

/* default UTrie2EnumValue() returns the input value itself */
static uint32_t U_CALLCONV
utrie2_enumSameValue(const void * /*context*/, uint32_t value) {
    return value;
}

/**
 * Enumerate all ranges of code points with the same relevant values.
 * The values are transformed from the raw trie entries by the enumValue function.
 *
 * Currently requires start<limit and both start and limit must be multiples
 * of UTRIE2_DATA_BLOCK_LENGTH.
 *
 * Optimizations:
 * - Skip a whole block if we know that it is filled with a single value,
 *   and it is the same as we visited just before.
 * - Handle the null block specially because we know a priori that it is filled
 *   with a single value.
 */
static void
enumEitherTrie(const UTrie2 *trie,
               UChar32 start, UChar32 limit,
               UTrie2EnumValue *enumValue, UTrie2EnumRange *enumRange, const void *context) {
    const uint32_t *data32;
    const uint16_t *idx;

    uint32_t value, prevValue, initialValue;
    UChar32 c, prev, highStart;
    int32_t j, i2Block, prevI2Block, index2NullOffset, block, prevBlock, nullBlock;

    if(enumRange==NULL) {
        return;
    }
    if(enumValue==NULL) {
        enumValue=utrie2_enumSameValue;
    }

    if(trie->newTrie==NULL) {
        /* frozen trie */
        idx=trie->index;
        U_ASSERT(idx!=NULL); /* the following code assumes trie->newTrie is not NULL when idx is NULL */
        data32=trie->data32;

        index2NullOffset=trie->index2NullOffset;
        nullBlock=trie->dataNullOffset;
    } else {
        /* unfrozen, mutable trie */
        idx=NULL;
        data32=trie->newTrie->data;
        U_ASSERT(data32!=NULL); /* the following code assumes idx is not NULL when data32 is NULL */

        index2NullOffset=trie->newTrie->index2NullOffset;
        nullBlock=trie->newTrie->dataNullOffset;
    }

    highStart=trie->highStart;

    /* get the enumeration value that corresponds to an initial-value trie data entry */
    initialValue=enumValue(context, trie->initialValue);

    /* set variables for previous range */
    prevI2Block=-1;
    prevBlock=-1;
    prev=start;
    prevValue=0;

    /* enumerate index-2 blocks */
    for(c=start; c<limit && c<highStart;) {
        /* Code point limit for iterating inside this i2Block. */
        UChar32 tempLimit=c+UTRIE2_CP_PER_INDEX_1_ENTRY;
        if(limit<tempLimit) {
            tempLimit=limit;
        }
        if(c<=0xffff) {
            if(!U_IS_SURROGATE(c)) {
                i2Block=c>>UTRIE2_SHIFT_2;
            } else if(U_IS_SURROGATE_LEAD(c)) {
                /*
                 * Enumerate values for lead surrogate code points, not code units:
                 * This special block has half the normal length.
                 */
                i2Block=UTRIE2_LSCP_INDEX_2_OFFSET;
                tempLimit=MIN_VALUE(0xdc00, limit);
            } else {
                /*
                 * Switch back to the normal part of the index-2 table.
                 * Enumerate the second half of the surrogates block.
                 */
                i2Block=0xd800>>UTRIE2_SHIFT_2;
                tempLimit=MIN_VALUE(0xe000, limit);
            }
        } else {
            /* supplementary code points */
            if(idx!=NULL) {
                i2Block=idx[(UTRIE2_INDEX_1_OFFSET-UTRIE2_OMITTED_BMP_INDEX_1_LENGTH)+
                              (c>>UTRIE2_SHIFT_1)];
            } else {
                i2Block=trie->newTrie->index1[c>>UTRIE2_SHIFT_1];
            }
            if(i2Block==prevI2Block && (c-prev)>=UTRIE2_CP_PER_INDEX_1_ENTRY) {
                /*
                 * The index-2 block is the same as the previous one, and filled with prevValue.
                 * Only possible for supplementary code points because the linear-BMP index-2
                 * table creates unique i2Block values.
                 */
                c+=UTRIE2_CP_PER_INDEX_1_ENTRY;
                continue;
            }
        }
        prevI2Block=i2Block;
        if(i2Block==index2NullOffset) {
            /* this is the null index-2 block */
            if(prevValue!=initialValue) {
                if(prev<c && !enumRange(context, prev, c-1, prevValue)) {
                    return;
                }
                prevBlock=nullBlock;
                prev=c;
                prevValue=initialValue;
            }
            c+=UTRIE2_CP_PER_INDEX_1_ENTRY;
        } else {
            /* enumerate data blocks for one index-2 block */
            int32_t i2, i2Limit;
            i2=(c>>UTRIE2_SHIFT_2)&UTRIE2_INDEX_2_MASK;
            if((c>>UTRIE2_SHIFT_1)==(tempLimit>>UTRIE2_SHIFT_1)) {
                i2Limit=(tempLimit>>UTRIE2_SHIFT_2)&UTRIE2_INDEX_2_MASK;
            } else {
                i2Limit=UTRIE2_INDEX_2_BLOCK_LENGTH;
            }
            for(; i2<i2Limit; ++i2) {
                if(idx!=NULL) {
                    block=(int32_t)idx[i2Block+i2]<<UTRIE2_INDEX_SHIFT;
                } else {
                    block=trie->newTrie->index2[i2Block+i2];
                }
                if(block==prevBlock && (c-prev)>=UTRIE2_DATA_BLOCK_LENGTH) {
                    /* the block is the same as the previous one, and filled with prevValue */
                    c+=UTRIE2_DATA_BLOCK_LENGTH;
                    continue;
                }
                prevBlock=block;
                if(block==nullBlock) {
                    /* this is the null data block */
                    if(prevValue!=initialValue) {
                        if(prev<c && !enumRange(context, prev, c-1, prevValue)) {
                            return;
                        }
                        prev=c;
                        prevValue=initialValue;
                    }
                    c+=UTRIE2_DATA_BLOCK_LENGTH;
                } else {
                    for(j=0; j<UTRIE2_DATA_BLOCK_LENGTH; ++j) {
                        value=enumValue(context, data32!=NULL ? data32[block+j] : idx[block+j]);
                        if(value!=prevValue) {
                            if(prev<c && !enumRange(context, prev, c-1, prevValue)) {
                                return;
                            }
                            prev=c;
                            prevValue=value;
                        }
                        ++c;
                    }
                }
            }
        }
    }

    if(c>limit) {
        c=limit;  /* could be higher if in the index2NullOffset */
    } else if(c<limit) {
        /* c==highStart<limit */
        uint32_t highValue;
        if(idx!=NULL) {
            highValue=
                data32!=NULL ?
                    data32[trie->highValueIndex] :
                    idx[trie->highValueIndex];
        } else {
            highValue=trie->newTrie->data[trie->newTrie->dataLength-UTRIE2_DATA_GRANULARITY];
        }
        value=enumValue(context, highValue);
        if(value!=prevValue) {
            if(prev<c && !enumRange(context, prev, c-1, prevValue)) {
                return;
            }
            prev=c;
            prevValue=value;
        }
        c=limit;
    }

    /* deliver last range */
    enumRange(context, prev, c-1, prevValue);
}

U_CAPI void U_EXPORT2
utrie2_enum(const UTrie2 *trie,
            UTrie2EnumValue *enumValue, UTrie2EnumRange *enumRange, const void *context) {
    enumEitherTrie(trie, 0, 0x110000, enumValue, enumRange, context);
}

U_CAPI void U_EXPORT2
utrie2_enumForLeadSurrogate(const UTrie2 *trie, UChar32 lead,
                            UTrie2EnumValue *enumValue, UTrie2EnumRange *enumRange,
                            const void *context) {
    if(!U16_IS_LEAD(lead)) {
        return;
    }
    lead=(lead-0xd7c0)<<10;   /* start code point */
    enumEitherTrie(trie, lead, lead+0x400, enumValue, enumRange, context);
}

/* C++ convenience wrappers ------------------------------------------------- */

U_NAMESPACE_BEGIN

uint16_t BackwardUTrie2StringIterator::previous16() {
    codePointLimit=codePointStart;
    if(start>=codePointStart) {
        codePoint=U_SENTINEL;
        return static_cast<uint16_t>(trie->errorValue);
    }
    uint16_t result;
    UTRIE2_U16_PREV16(trie, start, codePointStart, codePoint, result);
    return result;
}

uint16_t ForwardUTrie2StringIterator::next16() {
    codePointStart=codePointLimit;
    if(codePointLimit==limit) {
        codePoint=U_SENTINEL;
        return static_cast<uint16_t>(trie->errorValue);
    }
    uint16_t result;
    UTRIE2_U16_NEXT16(trie, codePointLimit, limit, codePoint, result);
    return result;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2001-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*   file name:  utrie2_builder.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2008sep26 (split off from utrie2.c)
*   created by: Markus W. Scherer
*
*   This is a common implementation of a Unicode trie.
*   It is a kind of compressed, serializable table of 16- or 32-bit values associated with
*   Unicode code points (0..0x10ffff).
*   This is the second common version of a Unicode trie (hence the name UTrie2).
*   See utrie2.h for a comparison.
*
*   This file contains only the builder code.
*   See utrie2.c for the runtime and enumeration code.
*/
// #define UTRIE2_DEBUG
#ifdef UTRIE2_DEBUG
#   include <stdio.h>
#endif
// #define UCPTRIE_DEBUG


#ifdef UCPTRIE_DEBUG



#endif




  // for utrie2_fromUTrie()

/* Implementation notes ----------------------------------------------------- */

/*
 * The UTRIE2_SHIFT_1, UTRIE2_SHIFT_2, UTRIE2_INDEX_SHIFT and other values
 * have been chosen to minimize trie sizes overall.
 * Most of the code is flexible enough to work with a range of values,
 * within certain limits.
 *
 * Exception: Support for separate values for lead surrogate code _units_
 * vs. code _points_ was added after the constants were fixed,
 * and has not been tested nor particularly designed for different constant values.
 * (Especially the utrie2_enum() code that jumps to the special LSCP index-2
 * part and back.)
 *
 * Requires UTRIE2_SHIFT_2<=6. Otherwise 0xc0 which is the top of the ASCII-linear data
 * including the bad-UTF-8-data block is not a multiple of UTRIE2_DATA_BLOCK_LENGTH
 * and map[block>>UTRIE2_SHIFT_2] (used in reference counting and compaction
 * remapping) stops working.
 *
 * Requires UTRIE2_SHIFT_1>=10 because utrie2_enumForLeadSurrogate()
 * assumes that a single index-2 block is used for 0x400 code points
 * corresponding to one lead surrogate.
 *
 * Requires UTRIE2_SHIFT_1<=16. Otherwise one single index-2 block contains
 * more than one Unicode plane, and the split of the index-2 table into a BMP
 * part and a supplementary part, with a gap in between, would not work.
 *
 * Requires UTRIE2_INDEX_SHIFT>=1 not because of the code but because
 * there is data with more than 64k distinct values,
 * for example for Unihan collation with a separate collation weight per
 * Han character.
 */

/* Building a trie ----------------------------------------------------------*/

enum {
    /** The null index-2 block, following the gap in the index-2 table. */
    UNEWTRIE2_INDEX_2_NULL_OFFSET=UNEWTRIE2_INDEX_GAP_OFFSET+UNEWTRIE2_INDEX_GAP_LENGTH,

    /** The start of allocated index-2 blocks. */
    UNEWTRIE2_INDEX_2_START_OFFSET=UNEWTRIE2_INDEX_2_NULL_OFFSET+UTRIE2_INDEX_2_BLOCK_LENGTH,

    /**
     * The null data block.
     * Length 64=0x40 even if UTRIE2_DATA_BLOCK_LENGTH is smaller,
     * to work with 6-bit trail bytes from 2-byte UTF-8.
     */
    UNEWTRIE2_DATA_NULL_OFFSET=UTRIE2_DATA_START_OFFSET,

    /** The start of allocated data blocks. */
    UNEWTRIE2_DATA_START_OFFSET=UNEWTRIE2_DATA_NULL_OFFSET+0x40,

    /**
     * The start of data blocks for U+0800 and above.
     * Below, compaction uses a block length of 64 for 2-byte UTF-8.
     * From here on, compaction uses UTRIE2_DATA_BLOCK_LENGTH.
     * Data values for 0x780 code points beyond ASCII.
     */
    UNEWTRIE2_DATA_0800_OFFSET=UNEWTRIE2_DATA_START_OFFSET+0x780
};

/* Start with allocation of 16k data entries. */
#define UNEWTRIE2_INITIAL_DATA_LENGTH ((int32_t)1<<14)

/* Grow about 8x each time. */
#define UNEWTRIE2_MEDIUM_DATA_LENGTH ((int32_t)1<<17)

static int32_t
allocIndex2Block(UNewTrie2 *trie);

U_CAPI UTrie2 * U_EXPORT2
utrie2_open(uint32_t initialValue, uint32_t errorValue, UErrorCode *pErrorCode) {
    UTrie2 *trie;
    UNewTrie2 *newTrie;
    uint32_t *data;
    int32_t i, j;

    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }

    trie=(UTrie2 *)uprv_malloc(sizeof(UTrie2));
    newTrie=(UNewTrie2 *)uprv_malloc(sizeof(UNewTrie2));
    data=(uint32_t *)uprv_malloc(UNEWTRIE2_INITIAL_DATA_LENGTH*4);
    if(trie==NULL || newTrie==NULL || data==NULL) {
        uprv_free(trie);
        uprv_free(newTrie);
        uprv_free(data);
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }

    uprv_memset(trie, 0, sizeof(UTrie2));
    trie->initialValue=initialValue;
    trie->errorValue=errorValue;
    trie->highStart=0x110000;
    trie->newTrie=newTrie;
#ifdef UTRIE2_DEBUG
    trie->name="open";
#endif

    newTrie->data=data;
#ifdef UCPTRIE_DEBUG
    newTrie->t3=umutablecptrie_open(initialValue, errorValue, pErrorCode);
#endif
    newTrie->dataCapacity=UNEWTRIE2_INITIAL_DATA_LENGTH;
    newTrie->initialValue=initialValue;
    newTrie->errorValue=errorValue;
    newTrie->highStart=0x110000;
    newTrie->firstFreeBlock=0;  /* no free block in the list */
    newTrie->isCompacted=FALSE;

    /*
     * preallocate and reset
     * - ASCII
     * - the bad-UTF-8-data block
     * - the null data block
     */
    for(i=0; i<0x80; ++i) {
        newTrie->data[i]=initialValue;
    }
    for(; i<0xc0; ++i) {
        newTrie->data[i]=errorValue;
    }
    for(i=UNEWTRIE2_DATA_NULL_OFFSET; i<UNEWTRIE2_DATA_START_OFFSET; ++i) {
        newTrie->data[i]=initialValue;
    }
    newTrie->dataNullOffset=UNEWTRIE2_DATA_NULL_OFFSET;
    newTrie->dataLength=UNEWTRIE2_DATA_START_OFFSET;

    /* set the index-2 indexes for the 2=0x80>>UTRIE2_SHIFT_2 ASCII data blocks */
    for(i=0, j=0; j<0x80; ++i, j+=UTRIE2_DATA_BLOCK_LENGTH) {
        newTrie->index2[i]=j;
        newTrie->map[i]=1;
    }
    /* reference counts for the bad-UTF-8-data block */
    for(; j<0xc0; ++i, j+=UTRIE2_DATA_BLOCK_LENGTH) {
        newTrie->map[i]=0;
    }
    /*
     * Reference counts for the null data block: all blocks except for the ASCII blocks.
     * Plus 1 so that we don't drop this block during compaction.
     * Plus as many as needed for lead surrogate code points.
     */
    /* i==newTrie->dataNullOffset */
    newTrie->map[i++]=
        (0x110000>>UTRIE2_SHIFT_2)-
        (0x80>>UTRIE2_SHIFT_2)+
        1+
        UTRIE2_LSCP_INDEX_2_LENGTH;
    j+=UTRIE2_DATA_BLOCK_LENGTH;
    for(; j<UNEWTRIE2_DATA_START_OFFSET; ++i, j+=UTRIE2_DATA_BLOCK_LENGTH) {
        newTrie->map[i]=0;
    }

    /*
     * set the remaining indexes in the BMP index-2 block
     * to the null data block
     */
    for(i=0x80>>UTRIE2_SHIFT_2; i<UTRIE2_INDEX_2_BMP_LENGTH; ++i) {
        newTrie->index2[i]=UNEWTRIE2_DATA_NULL_OFFSET;
    }

    /*
     * Fill the index gap with impossible values so that compaction
     * does not overlap other index-2 blocks with the gap.
     */
    for(i=0; i<UNEWTRIE2_INDEX_GAP_LENGTH; ++i) {
        newTrie->index2[UNEWTRIE2_INDEX_GAP_OFFSET+i]=-1;
    }

    /* set the indexes in the null index-2 block */
    for(i=0; i<UTRIE2_INDEX_2_BLOCK_LENGTH; ++i) {
        newTrie->index2[UNEWTRIE2_INDEX_2_NULL_OFFSET+i]=UNEWTRIE2_DATA_NULL_OFFSET;
    }
    newTrie->index2NullOffset=UNEWTRIE2_INDEX_2_NULL_OFFSET;
    newTrie->index2Length=UNEWTRIE2_INDEX_2_START_OFFSET;

    /* set the index-1 indexes for the linear index-2 block */
    for(i=0, j=0;
        i<UTRIE2_OMITTED_BMP_INDEX_1_LENGTH;
        ++i, j+=UTRIE2_INDEX_2_BLOCK_LENGTH
    ) {
        newTrie->index1[i]=j;
    }

    /* set the remaining index-1 indexes to the null index-2 block */
    for(; i<UNEWTRIE2_INDEX_1_LENGTH; ++i) {
        newTrie->index1[i]=UNEWTRIE2_INDEX_2_NULL_OFFSET;
    }

    /*
     * Preallocate and reset data for U+0080..U+07ff,
     * for 2-byte UTF-8 which will be compacted in 64-blocks
     * even if UTRIE2_DATA_BLOCK_LENGTH is smaller.
     */
    for(i=0x80; i<0x800; i+=UTRIE2_DATA_BLOCK_LENGTH) {
        utrie2_set32(trie, i, initialValue, pErrorCode);
    }

    return trie;
}

static UNewTrie2 *
cloneBuilder(const UNewTrie2 *other) {
    UNewTrie2 *trie;

    trie=(UNewTrie2 *)uprv_malloc(sizeof(UNewTrie2));
    if(trie==NULL) {
        return NULL;
    }

    trie->data=(uint32_t *)uprv_malloc(other->dataCapacity*4);
    if(trie->data==NULL) {
        uprv_free(trie);
        return NULL;
    }
#ifdef UCPTRIE_DEBUG
    if(other->t3==nullptr) {
        trie->t3=nullptr;
    } else {
        UErrorCode errorCode=U_ZERO_ERROR;
        trie->t3=umutablecptrie_clone(other->t3, &errorCode);
    }
#endif
    trie->dataCapacity=other->dataCapacity;

    /* clone data */
    uprv_memcpy(trie->index1, other->index1, sizeof(trie->index1));
    uprv_memcpy(trie->index2, other->index2, (size_t)other->index2Length*4);
    trie->index2NullOffset=other->index2NullOffset;
    trie->index2Length=other->index2Length;

    uprv_memcpy(trie->data, other->data, (size_t)other->dataLength*4);
    trie->dataNullOffset=other->dataNullOffset;
    trie->dataLength=other->dataLength;

    /* reference counters */
    if(other->isCompacted) {
        trie->firstFreeBlock=0;
    } else {
        uprv_memcpy(trie->map, other->map, ((size_t)other->dataLength>>UTRIE2_SHIFT_2)*4);
        trie->firstFreeBlock=other->firstFreeBlock;
    }

    trie->initialValue=other->initialValue;
    trie->errorValue=other->errorValue;
    trie->highStart=other->highStart;
    trie->isCompacted=other->isCompacted;

    return trie;
}

U_CAPI UTrie2 * U_EXPORT2
utrie2_clone(const UTrie2 *other, UErrorCode *pErrorCode) {
    UTrie2 *trie;

    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if(other==NULL || (other->memory==NULL && other->newTrie==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    trie=(UTrie2 *)uprv_malloc(sizeof(UTrie2));
    if(trie==NULL) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    uprv_memcpy(trie, other, sizeof(UTrie2));

    if(other->memory!=NULL) {
        trie->memory=uprv_malloc(other->length);
        if(trie->memory!=NULL) {
            trie->isMemoryOwned=TRUE;
            uprv_memcpy(trie->memory, other->memory, other->length);

            /* make the clone's pointers point to its own memory */
            trie->index=(uint16_t *)trie->memory+(other->index-(uint16_t *)other->memory);
            if(other->data16!=NULL) {
                trie->data16=(uint16_t *)trie->memory+(other->data16-(uint16_t *)other->memory);
            }
            if(other->data32!=NULL) {
                trie->data32=(uint32_t *)trie->memory+(other->data32-(uint32_t *)other->memory);
            }
        }
    } else /* other->newTrie!=NULL */ {
        trie->newTrie=cloneBuilder(other->newTrie);
    }

    if(trie->memory==NULL && trie->newTrie==NULL) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        uprv_free(trie);
        trie=NULL;
    }
    return trie;
}

typedef struct NewTrieAndStatus {
    UTrie2 *trie;
    UErrorCode errorCode;
    UBool exclusiveLimit;  /* rather than inclusive range end */
} NewTrieAndStatus;

static UBool U_CALLCONV
copyEnumRange(const void *context, UChar32 start, UChar32 end, uint32_t value) {
    NewTrieAndStatus *nt=(NewTrieAndStatus *)context;
    if(value!=nt->trie->initialValue) {
        if(nt->exclusiveLimit) {
            --end;
        }
        if(start==end) {
            utrie2_set32(nt->trie, start, value, &nt->errorCode);
        } else {
            utrie2_setRange32(nt->trie, start, end, value, TRUE, &nt->errorCode);
        }
        return U_SUCCESS(nt->errorCode);
    } else {
        return TRUE;
    }
}

#ifdef UTRIE2_DEBUG
static long countInitial(const UTrie2 *trie) {
    uint32_t initialValue=trie->initialValue;
    int32_t length=trie->dataLength;
    long count=0;
    if(trie->data16!=nullptr) {
        for(int32_t i=0; i<length; ++i) {
            if(trie->data16[i]==initialValue) { ++count; }
        }
    } else {
        for(int32_t i=0; i<length; ++i) {
            if(trie->data32[i]==initialValue) { ++count; }
        }
    }
    return count;
}

static void
utrie_printLengths(const UTrie *trie) {
    long indexLength=trie->indexLength;
    long dataLength=(long)trie->dataLength;
    long totalLength=(long)sizeof(UTrieHeader)+indexLength*2+dataLength*(trie->data32!=NULL ? 4 : 2);
    printf("**UTrieLengths** index:%6ld  data:%6ld  serialized:%6ld\n",
           indexLength, dataLength, totalLength);
}

static void
utrie2_printLengths(const UTrie2 *trie, const char *which) {
    long indexLength=trie->indexLength;
    long dataLength=(long)trie->dataLength;
    long totalLength=(long)sizeof(UTrie2Header)+indexLength*2+dataLength*(trie->data32!=NULL ? 4 : 2);
    printf("**UTrie2Lengths(%s %s)** index:%6ld  data:%6ld  countInitial:%6ld  serialized:%6ld\n",
           which, trie->name, indexLength, dataLength, countInitial(trie), totalLength);
}
#endif

U_CAPI UTrie2 * U_EXPORT2
utrie2_cloneAsThawed(const UTrie2 *other, UErrorCode *pErrorCode) {
    NewTrieAndStatus context;
    UChar lead;

    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if(other==NULL || (other->memory==NULL && other->newTrie==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    if(other->newTrie!=NULL && !other->newTrie->isCompacted) {
        return utrie2_clone(other, pErrorCode);  /* clone an unfrozen trie */
    }

    /* Clone the frozen trie by enumerating it and building a new one. */
    context.trie=utrie2_open(other->initialValue, other->errorValue, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    context.exclusiveLimit=FALSE;
    context.errorCode=*pErrorCode;
    utrie2_enum(other, NULL, copyEnumRange, &context);
    *pErrorCode=context.errorCode;
    for(lead=0xd800; lead<0xdc00; ++lead) {
        uint32_t value;
        if(other->data32==NULL) {
            value=UTRIE2_GET16_FROM_U16_SINGLE_LEAD(other, lead);
        } else {
            value=UTRIE2_GET32_FROM_U16_SINGLE_LEAD(other, lead);
        }
        if(value!=other->initialValue) {
            utrie2_set32ForLeadSurrogateCodeUnit(context.trie, lead, value, pErrorCode);
        }
    }
    if(U_FAILURE(*pErrorCode)) {
        utrie2_close(context.trie);
        context.trie=NULL;
    }
    return context.trie;
}

/* Almost the same as utrie2_cloneAsThawed() but copies a UTrie and freezes the clone. */
U_CAPI UTrie2 * U_EXPORT2
utrie2_fromUTrie(const UTrie *trie1, uint32_t errorValue, UErrorCode *pErrorCode) {
    NewTrieAndStatus context;
    UChar lead;

    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    if(trie1==NULL) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    context.trie=utrie2_open(trie1->initialValue, errorValue, pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    context.exclusiveLimit=TRUE;
    context.errorCode=*pErrorCode;
    utrie_enum(trie1, NULL, copyEnumRange, &context);
    *pErrorCode=context.errorCode;
    for(lead=0xd800; lead<0xdc00; ++lead) {
        uint32_t value;
        if(trie1->data32==NULL) {
            value=UTRIE_GET16_FROM_LEAD(trie1, lead);
        } else {
            value=UTRIE_GET32_FROM_LEAD(trie1, lead);
        }
        if(value!=trie1->initialValue) {
            utrie2_set32ForLeadSurrogateCodeUnit(context.trie, lead, value, pErrorCode);
        }
    }
    if(U_SUCCESS(*pErrorCode)) {
        utrie2_freeze(context.trie,
                      trie1->data32!=NULL ? UTRIE2_32_VALUE_BITS : UTRIE2_16_VALUE_BITS,
                      pErrorCode);
    }
#ifdef UTRIE2_DEBUG
    if(U_SUCCESS(*pErrorCode)) {
        utrie_printLengths(trie1);
        utrie2_printLengths(context.trie, "fromUTrie");
    }
#endif
    if(U_FAILURE(*pErrorCode)) {
        utrie2_close(context.trie);
        context.trie=NULL;
    }
    return context.trie;
}

static inline UBool
isInNullBlock(UNewTrie2 *trie, UChar32 c, UBool forLSCP) {
    int32_t i2, block;

    if(U_IS_LEAD(c) && forLSCP) {
        i2=(UTRIE2_LSCP_INDEX_2_OFFSET-(0xd800>>UTRIE2_SHIFT_2))+
            (c>>UTRIE2_SHIFT_2);
    } else {
        i2=trie->index1[c>>UTRIE2_SHIFT_1]+
            ((c>>UTRIE2_SHIFT_2)&UTRIE2_INDEX_2_MASK);
    }
    block=trie->index2[i2];
    return (UBool)(block==trie->dataNullOffset);
}

static int32_t
allocIndex2Block(UNewTrie2 *trie) {
    int32_t newBlock, newTop;

    newBlock=trie->index2Length;
    newTop=newBlock+UTRIE2_INDEX_2_BLOCK_LENGTH;
    if(newTop>UPRV_LENGTHOF(trie->index2)) {
        /*
         * Should never occur.
         * Either UTRIE2_MAX_BUILD_TIME_INDEX_LENGTH is incorrect,
         * or the code writes more values than should be possible.
         */
        return -1;
    }
    trie->index2Length=newTop;
    uprv_memcpy(trie->index2+newBlock, trie->index2+trie->index2NullOffset, UTRIE2_INDEX_2_BLOCK_LENGTH*4);
    return newBlock;
}

static int32_t
getIndex2Block(UNewTrie2 *trie, UChar32 c, UBool forLSCP) {
    int32_t i1, i2;

    if(U_IS_LEAD(c) && forLSCP) {
        return UTRIE2_LSCP_INDEX_2_OFFSET;
    }

    i1=c>>UTRIE2_SHIFT_1;
    i2=trie->index1[i1];
    if(i2==trie->index2NullOffset) {
        i2=allocIndex2Block(trie);
        if(i2<0) {
            return -1;  /* program error */
        }
        trie->index1[i1]=i2;
    }
    return i2;
}

static int32_t
allocDataBlock(UNewTrie2 *trie, int32_t copyBlock) {
    int32_t newBlock, newTop;

    if(trie->firstFreeBlock!=0) {
        /* get the first free block */
        newBlock=trie->firstFreeBlock;
        trie->firstFreeBlock=-trie->map[newBlock>>UTRIE2_SHIFT_2];
    } else {
        /* get a new block from the high end */
        newBlock=trie->dataLength;
        newTop=newBlock+UTRIE2_DATA_BLOCK_LENGTH;
        if(newTop>trie->dataCapacity) {
            /* out of memory in the data array */
            int32_t capacity;
            uint32_t *data;

            if(trie->dataCapacity<UNEWTRIE2_MEDIUM_DATA_LENGTH) {
                capacity=UNEWTRIE2_MEDIUM_DATA_LENGTH;
            } else if(trie->dataCapacity<UNEWTRIE2_MAX_DATA_LENGTH) {
                capacity=UNEWTRIE2_MAX_DATA_LENGTH;
            } else {
                /*
                 * Should never occur.
                 * Either UNEWTRIE2_MAX_DATA_LENGTH is incorrect,
                 * or the code writes more values than should be possible.
                 */
                return -1;
            }
            data=(uint32_t *)uprv_malloc(capacity*4);
            if(data==NULL) {
                return -1;
            }
            uprv_memcpy(data, trie->data, (size_t)trie->dataLength*4);
            uprv_free(trie->data);
            trie->data=data;
            trie->dataCapacity=capacity;
        }
        trie->dataLength=newTop;
    }
    uprv_memcpy(trie->data+newBlock, trie->data+copyBlock, UTRIE2_DATA_BLOCK_LENGTH*4);
    trie->map[newBlock>>UTRIE2_SHIFT_2]=0;
    return newBlock;
}

/* call when the block's reference counter reaches 0 */
static void
releaseDataBlock(UNewTrie2 *trie, int32_t block) {
    /* put this block at the front of the free-block chain */
    trie->map[block>>UTRIE2_SHIFT_2]=-trie->firstFreeBlock;
    trie->firstFreeBlock=block;
}

static inline UBool
isWritableBlock(UNewTrie2 *trie, int32_t block) {
    return (UBool)(block!=trie->dataNullOffset && 1==trie->map[block>>UTRIE2_SHIFT_2]);
}

static inline void
setIndex2Entry(UNewTrie2 *trie, int32_t i2, int32_t block) {
    int32_t oldBlock;
    ++trie->map[block>>UTRIE2_SHIFT_2];  /* increment first, in case block==oldBlock! */
    oldBlock=trie->index2[i2];
    if(0 == --trie->map[oldBlock>>UTRIE2_SHIFT_2]) {
        releaseDataBlock(trie, oldBlock);
    }
    trie->index2[i2]=block;
}

/**
 * No error checking for illegal arguments.
 *
 * @return -1 if no new data block available (out of memory in data array)
 * @internal
 */
static int32_t
getDataBlock(UNewTrie2 *trie, UChar32 c, UBool forLSCP) {
    int32_t i2, oldBlock, newBlock;

    i2=getIndex2Block(trie, c, forLSCP);
    if(i2<0) {
        return -1;  /* program error */
    }

    i2+=(c>>UTRIE2_SHIFT_2)&UTRIE2_INDEX_2_MASK;
    oldBlock=trie->index2[i2];
    if(isWritableBlock(trie, oldBlock)) {
        return oldBlock;
    }

    /* allocate a new data block */
    newBlock=allocDataBlock(trie, oldBlock);
    if(newBlock<0) {
        /* out of memory in the data array */
        return -1;
    }
    setIndex2Entry(trie, i2, newBlock);
    return newBlock;
}

/**
 * @return TRUE if the value was successfully set
 */
static void
set32(UNewTrie2 *trie,
      UChar32 c, UBool forLSCP, uint32_t value,
      UErrorCode *pErrorCode) {
    int32_t block;

    if(trie==NULL || trie->isCompacted) {
        *pErrorCode=U_NO_WRITE_PERMISSION;
        return;
    }
#ifdef UCPTRIE_DEBUG
    umutablecptrie_set(trie->t3, c, value, pErrorCode);
#endif

    block=getDataBlock(trie, c, forLSCP);
    if(block<0) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    trie->data[block+(c&UTRIE2_DATA_MASK)]=value;
}

U_CAPI void U_EXPORT2
utrie2_set32(UTrie2 *trie, UChar32 c, uint32_t value, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return;
    }
    if((uint32_t)c>0x10ffff) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    set32(trie->newTrie, c, TRUE, value, pErrorCode);
}

U_CAPI void U_EXPORT2
utrie2_set32ForLeadSurrogateCodeUnit(UTrie2 *trie,
                                     UChar32 c, uint32_t value,
                                     UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return;
    }
    if(!U_IS_LEAD(c)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    set32(trie->newTrie, c, FALSE, value, pErrorCode);
}

static void
utrie2builder_writeBlock(uint32_t *block, uint32_t value) {
    uint32_t *limit=block+UTRIE2_DATA_BLOCK_LENGTH;
    while(block<limit) {
        *block++=value;
    }
}

/**
 * initialValue is ignored if overwrite=TRUE
 * @internal
 */
static void
fillBlock(uint32_t *block, UChar32 start, UChar32 limit,
          uint32_t value, uint32_t initialValue, UBool overwrite) {
    uint32_t *pLimit;

    pLimit=block+limit;
    block+=start;
    if(overwrite) {
        while(block<pLimit) {
            *block++=value;
        }
    } else {
        while(block<pLimit) {
            if(*block==initialValue) {
                *block=value;
            }
            ++block;
        }
    }
}

U_CAPI void U_EXPORT2
utrie2_setRange32(UTrie2 *trie,
                  UChar32 start, UChar32 end,
                  uint32_t value, UBool overwrite,
                  UErrorCode *pErrorCode) {
    /*
     * repeat value in [start..end]
     * mark index values for repeat-data blocks by setting bit 31 of the index values
     * fill around existing values if any, if(overwrite)
     */
    UNewTrie2 *newTrie;
    int32_t block, rest, repeatBlock;
    UChar32 limit;

    if(U_FAILURE(*pErrorCode)) {
        return;
    }
    if((uint32_t)start>0x10ffff || (uint32_t)end>0x10ffff || start>end) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    newTrie=trie->newTrie;
    if(newTrie==NULL || newTrie->isCompacted) {
        *pErrorCode=U_NO_WRITE_PERMISSION;
        return;
    }
#ifdef UCPTRIE_DEBUG
    umutablecptrie_setRange(newTrie->t3, start, end, value, pErrorCode);
#endif
    if(!overwrite && value==newTrie->initialValue) {
        return; /* nothing to do */
    }

    limit=end+1;
    if(start&UTRIE2_DATA_MASK) {
        UChar32 nextStart;

        /* set partial block at [start..following block boundary[ */
        block=getDataBlock(newTrie, start, TRUE);
        if(block<0) {
            *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
            return;
        }

        nextStart=(start+UTRIE2_DATA_MASK)&~UTRIE2_DATA_MASK;
        if(nextStart<=limit) {
            fillBlock(newTrie->data+block, start&UTRIE2_DATA_MASK, UTRIE2_DATA_BLOCK_LENGTH,
                      value, newTrie->initialValue, overwrite);
            start=nextStart;
        } else {
            fillBlock(newTrie->data+block, start&UTRIE2_DATA_MASK, limit&UTRIE2_DATA_MASK,
                      value, newTrie->initialValue, overwrite);
            return;
        }
    }

    /* number of positions in the last, partial block */
    rest=limit&UTRIE2_DATA_MASK;

    /* round down limit to a block boundary */
    limit&=~UTRIE2_DATA_MASK;

    /* iterate over all-value blocks */
    if(value==newTrie->initialValue) {
        repeatBlock=newTrie->dataNullOffset;
    } else {
        repeatBlock=-1;
    }

    while(start<limit) {
        int32_t i2;
        UBool setRepeatBlock=FALSE;

        if(value==newTrie->initialValue && isInNullBlock(newTrie, start, TRUE)) {
            start+=UTRIE2_DATA_BLOCK_LENGTH; /* nothing to do */
            continue;
        }

        /* get index value */
        i2=getIndex2Block(newTrie, start, TRUE);
        if(i2<0) {
            *pErrorCode=U_INTERNAL_PROGRAM_ERROR;
            return;
        }
        i2+=(start>>UTRIE2_SHIFT_2)&UTRIE2_INDEX_2_MASK;
        block=newTrie->index2[i2];
        if(isWritableBlock(newTrie, block)) {
            /* already allocated */
            if(overwrite && block>=UNEWTRIE2_DATA_0800_OFFSET) {
                /*
                 * We overwrite all values, and it's not a
                 * protected (ASCII-linear or 2-byte UTF-8) block:
                 * replace with the repeatBlock.
                 */
                setRepeatBlock=TRUE;
            } else {
                /* !overwrite, or protected block: just write the values into this block */
                fillBlock(newTrie->data+block,
                          0, UTRIE2_DATA_BLOCK_LENGTH,
                          value, newTrie->initialValue, overwrite);
            }
        } else if(newTrie->data[block]!=value && (overwrite || block==newTrie->dataNullOffset)) {
            /*
             * Set the repeatBlock instead of the null block or previous repeat block:
             *
             * If !isWritableBlock() then all entries in the block have the same value
             * because it's the null block or a range block (the repeatBlock from a previous
             * call to utrie2_setRange32()).
             * No other blocks are used multiple times before compacting.
             *
             * The null block is the only non-writable block with the initialValue because
             * of the repeatBlock initialization above. (If value==initialValue, then
             * the repeatBlock will be the null data block.)
             *
             * We set our repeatBlock if the desired value differs from the block's value,
             * and if we overwrite any data or if the data is all initial values
             * (which is the same as the block being the null block, see above).
             */
            setRepeatBlock=TRUE;
        }
        if(setRepeatBlock) {
            if(repeatBlock>=0) {
                setIndex2Entry(newTrie, i2, repeatBlock);
            } else {
                /* create and set and fill the repeatBlock */
                repeatBlock=getDataBlock(newTrie, start, TRUE);
                if(repeatBlock<0) {
                    *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
                    return;
                }
                utrie2builder_writeBlock(newTrie->data+repeatBlock, value);
            }
        }

        start+=UTRIE2_DATA_BLOCK_LENGTH;
    }

    if(rest>0) {
        /* set partial block at [last block boundary..limit[ */
        block=getDataBlock(newTrie, start, TRUE);
        if(block<0) {
            *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
            return;
        }

        fillBlock(newTrie->data+block, 0, rest, value, newTrie->initialValue, overwrite);
    }

    return;
}

/* compaction --------------------------------------------------------------- */

static inline UBool
utrie2_builder_equal_int32(const int32_t *s, const int32_t *t, int32_t length) {
    while(length>0 && *s==*t) {
        ++s;
        ++t;
        --length;
    }
    return (UBool)(length==0);
}

static inline UBool
utrie2_builder_equal_uint32(const uint32_t *s, const uint32_t *t, int32_t length) {
    while(length>0 && *s==*t) {
        ++s;
        ++t;
        --length;
    }
    return (UBool)(length==0);
}

static int32_t
findSameIndex2Block(const int32_t *idx, int32_t index2Length, int32_t otherBlock) {
    int32_t block;

    /* ensure that we do not even partially get past index2Length */
    index2Length-=UTRIE2_INDEX_2_BLOCK_LENGTH;

    for(block=0; block<=index2Length; ++block) {
        if(utrie2_builder_equal_int32(idx+block, idx+otherBlock, UTRIE2_INDEX_2_BLOCK_LENGTH)) {
            return block;
        }
    }
    return -1;
}

static int32_t
findSameDataBlock(const uint32_t *data, int32_t dataLength, int32_t otherBlock, int32_t blockLength) {
    int32_t block;

    /* ensure that we do not even partially get past dataLength */
    dataLength-=blockLength;

    for(block=0; block<=dataLength; block+=UTRIE2_DATA_GRANULARITY) {
        if(utrie2_builder_equal_uint32(data+block, data+otherBlock, blockLength)) {
            return block;
        }
    }
    return -1;
}

/*
 * Find the start of the last range in the trie by enumerating backward.
 * Indexes for supplementary code points higher than this will be omitted.
 */
static UChar32
findHighStart(UNewTrie2 *trie, uint32_t highValue) {
    const uint32_t *data32;

    uint32_t value, initialValue;
    UChar32 c, prev;
    int32_t i1, i2, j, i2Block, prevI2Block, index2NullOffset, block, prevBlock, nullBlock;

    data32=trie->data;
    initialValue=trie->initialValue;

    index2NullOffset=trie->index2NullOffset;
    nullBlock=trie->dataNullOffset;

    /* set variables for previous range */
    if(highValue==initialValue) {
        prevI2Block=index2NullOffset;
        prevBlock=nullBlock;
    } else {
        prevI2Block=-1;
        prevBlock=-1;
    }
    prev=0x110000;

    /* enumerate index-2 blocks */
    i1=UNEWTRIE2_INDEX_1_LENGTH;
    c=prev;
    while(c>0) {
        i2Block=trie->index1[--i1];
        if(i2Block==prevI2Block) {
            /* the index-2 block is the same as the previous one, and filled with highValue */
            c-=UTRIE2_CP_PER_INDEX_1_ENTRY;
            continue;
        }
        prevI2Block=i2Block;
        if(i2Block==index2NullOffset) {
            /* this is the null index-2 block */
            if(highValue!=initialValue) {
                return c;
            }
            c-=UTRIE2_CP_PER_INDEX_1_ENTRY;
        } else {
            /* enumerate data blocks for one index-2 block */
            for(i2=UTRIE2_INDEX_2_BLOCK_LENGTH; i2>0;) {
                block=trie->index2[i2Block+ --i2];
                if(block==prevBlock) {
                    /* the block is the same as the previous one, and filled with highValue */
                    c-=UTRIE2_DATA_BLOCK_LENGTH;
                    continue;
                }
                prevBlock=block;
                if(block==nullBlock) {
                    /* this is the null data block */
                    if(highValue!=initialValue) {
                        return c;
                    }
                    c-=UTRIE2_DATA_BLOCK_LENGTH;
                } else {
                    for(j=UTRIE2_DATA_BLOCK_LENGTH; j>0;) {
                        value=data32[block+ --j];
                        if(value!=highValue) {
                            return c;
                        }
                        --c;
                    }
                }
            }
        }
    }

    /* deliver last range */
    return 0;
}

/*
 * Compact a build-time trie.
 *
 * The compaction
 * - removes blocks that are identical with earlier ones
 * - overlaps adjacent blocks as much as possible (if overlap==TRUE)
 * - moves blocks in steps of the data granularity
 * - moves and overlaps blocks that overlap with multiple values in the overlap region
 *
 * It does not
 * - try to move and overlap blocks that are not already adjacent
 */
static void
compactData(UNewTrie2 *trie) {
#ifdef UTRIE2_DEBUG
    int32_t countSame=0, sumOverlaps=0;
#endif

    int32_t start, newStart, movedStart;
    int32_t blockLength, overlap;
    int32_t i, mapIndex, blockCount;

    /* do not compact linear-ASCII data */
    newStart=UTRIE2_DATA_START_OFFSET;
    for(start=0, i=0; start<newStart; start+=UTRIE2_DATA_BLOCK_LENGTH, ++i) {
        trie->map[i]=start;
    }

    /*
     * Start with a block length of 64 for 2-byte UTF-8,
     * then switch to UTRIE2_DATA_BLOCK_LENGTH.
     */
    blockLength=64;
    blockCount=blockLength>>UTRIE2_SHIFT_2;
    for(start=newStart; start<trie->dataLength;) {
        /*
         * start: index of first entry of current block
         * newStart: index where the current block is to be moved
         *           (right after current end of already-compacted data)
         */
        if(start==UNEWTRIE2_DATA_0800_OFFSET) {
            blockLength=UTRIE2_DATA_BLOCK_LENGTH;
            blockCount=1;
        }

        /* skip blocks that are not used */
        if(trie->map[start>>UTRIE2_SHIFT_2]<=0) {
            /* advance start to the next block */
            start+=blockLength;

            /* leave newStart with the previous block! */
            continue;
        }

        /* search for an identical block */
        if( (movedStart=findSameDataBlock(trie->data, newStart, start, blockLength))
             >=0
        ) {
#ifdef UTRIE2_DEBUG
            ++countSame;
#endif
            /* found an identical block, set the other block's index value for the current block */
            for(i=blockCount, mapIndex=start>>UTRIE2_SHIFT_2; i>0; --i) {
                trie->map[mapIndex++]=movedStart;
                movedStart+=UTRIE2_DATA_BLOCK_LENGTH;
            }

            /* advance start to the next block */
            start+=blockLength;

            /* leave newStart with the previous block! */
            continue;
        }

        /* see if the beginning of this block can be overlapped with the end of the previous block */
        /* look for maximum overlap (modulo granularity) with the previous, adjacent block */
        for(overlap=blockLength-UTRIE2_DATA_GRANULARITY;
            overlap>0 && !utrie2_builder_equal_uint32(trie->data+(newStart-overlap), trie->data+start, overlap);
            overlap-=UTRIE2_DATA_GRANULARITY) {}

#ifdef UTRIE2_DEBUG
            sumOverlaps+=overlap;
#endif
        if(overlap>0 || newStart<start) {
            /* some overlap, or just move the whole block */
            movedStart=newStart-overlap;
            for(i=blockCount, mapIndex=start>>UTRIE2_SHIFT_2; i>0; --i) {
                trie->map[mapIndex++]=movedStart;
                movedStart+=UTRIE2_DATA_BLOCK_LENGTH;
            }

            /* move the non-overlapping indexes to their new positions */
            start+=overlap;
            for(i=blockLength-overlap; i>0; --i) {
                trie->data[newStart++]=trie->data[start++];
            }
        } else /* no overlap && newStart==start */ {
            for(i=blockCount, mapIndex=start>>UTRIE2_SHIFT_2; i>0; --i) {
                trie->map[mapIndex++]=start;
                start+=UTRIE2_DATA_BLOCK_LENGTH;
            }
            newStart=start;
        }
    }

    /* now adjust the index-2 table */
    for(i=0; i<trie->index2Length; ++i) {
        if(i==UNEWTRIE2_INDEX_GAP_OFFSET) {
            /* Gap indexes are invalid (-1). Skip over the gap. */
            i+=UNEWTRIE2_INDEX_GAP_LENGTH;
        }
        trie->index2[i]=trie->map[trie->index2[i]>>UTRIE2_SHIFT_2];
    }
    trie->dataNullOffset=trie->map[trie->dataNullOffset>>UTRIE2_SHIFT_2];

    /* ensure dataLength alignment */
    while((newStart&(UTRIE2_DATA_GRANULARITY-1))!=0) {
        trie->data[newStart++]=trie->initialValue;
    }

#ifdef UTRIE2_DEBUG
    /* we saved some space */
    printf("compacting UTrie2: count of 32-bit data words %lu->%lu  countSame=%ld  sumOverlaps=%ld\n",
            (long)trie->dataLength, (long)newStart, (long)countSame, (long)sumOverlaps);
#endif

    trie->dataLength=newStart;
}

static void
compactIndex2(UNewTrie2 *trie) {
    int32_t i, start, newStart, movedStart, overlap;

    /* do not compact linear-BMP index-2 blocks */
    newStart=UTRIE2_INDEX_2_BMP_LENGTH;
    for(start=0, i=0; start<newStart; start+=UTRIE2_INDEX_2_BLOCK_LENGTH, ++i) {
        trie->map[i]=start;
    }

    /* Reduce the index table gap to what will be needed at runtime. */
    newStart+=UTRIE2_UTF8_2B_INDEX_2_LENGTH+((trie->highStart-0x10000)>>UTRIE2_SHIFT_1);

    for(start=UNEWTRIE2_INDEX_2_NULL_OFFSET; start<trie->index2Length;) {
        /*
         * start: index of first entry of current block
         * newStart: index where the current block is to be moved
         *           (right after current end of already-compacted data)
         */

        /* search for an identical block */
        if( (movedStart=findSameIndex2Block(trie->index2, newStart, start))
             >=0
        ) {
            /* found an identical block, set the other block's index value for the current block */
            trie->map[start>>UTRIE2_SHIFT_1_2]=movedStart;

            /* advance start to the next block */
            start+=UTRIE2_INDEX_2_BLOCK_LENGTH;

            /* leave newStart with the previous block! */
            continue;
        }

        /* see if the beginning of this block can be overlapped with the end of the previous block */
        /* look for maximum overlap with the previous, adjacent block */
        for(overlap=UTRIE2_INDEX_2_BLOCK_LENGTH-1;
            overlap>0 && !utrie2_builder_equal_int32(trie->index2+(newStart-overlap), trie->index2+start, overlap);
            --overlap) {}

        if(overlap>0 || newStart<start) {
            /* some overlap, or just move the whole block */
            trie->map[start>>UTRIE2_SHIFT_1_2]=newStart-overlap;

            /* move the non-overlapping indexes to their new positions */
            start+=overlap;
            for(i=UTRIE2_INDEX_2_BLOCK_LENGTH-overlap; i>0; --i) {
                trie->index2[newStart++]=trie->index2[start++];
            }
        } else /* no overlap && newStart==start */ {
            trie->map[start>>UTRIE2_SHIFT_1_2]=start;
            start+=UTRIE2_INDEX_2_BLOCK_LENGTH;
            newStart=start;
        }
    }

    /* now adjust the index-1 table */
    for(i=0; i<UNEWTRIE2_INDEX_1_LENGTH; ++i) {
        trie->index1[i]=trie->map[trie->index1[i]>>UTRIE2_SHIFT_1_2];
    }
    trie->index2NullOffset=trie->map[trie->index2NullOffset>>UTRIE2_SHIFT_1_2];

    /*
     * Ensure data table alignment:
     * Needs to be granularity-aligned for 16-bit trie
     * (so that dataMove will be down-shiftable),
     * and 2-aligned for uint32_t data.
     */
    while((newStart&((UTRIE2_DATA_GRANULARITY-1)|1))!=0) {
        /* Arbitrary value: 0x3fffc not possible for real data. */
        trie->index2[newStart++]=(int32_t)0xffff<<UTRIE2_INDEX_SHIFT;
    }

#ifdef UTRIE2_DEBUG
    /* we saved some space */
    printf("compacting UTrie2: count of 16-bit index words %lu->%lu\n",
            (long)trie->index2Length, (long)newStart);
#endif

    trie->index2Length=newStart;
}

static void
compactTrie(UTrie2 *trie, UErrorCode *pErrorCode) {
    UNewTrie2 *newTrie;
    UChar32 highStart, suppHighStart;
    uint32_t highValue;

    newTrie=trie->newTrie;

    /* find highStart and round it up */
    highValue=utrie2_get32(trie, 0x10ffff);
    highStart=findHighStart(newTrie, highValue);
    highStart=(highStart+(UTRIE2_CP_PER_INDEX_1_ENTRY-1))&~(UTRIE2_CP_PER_INDEX_1_ENTRY-1);
    if(highStart==0x110000) {
        highValue=trie->errorValue;
    }

    /*
     * Set trie->highStart only after utrie2_get32(trie, highStart).
     * Otherwise utrie2_get32(trie, highStart) would try to read the highValue.
     */
    trie->highStart=newTrie->highStart=highStart;

#ifdef UTRIE2_DEBUG
    printf("UTrie2: highStart U+%06lx  highValue 0x%lx  initialValue 0x%lx\n",
            (long)highStart, (long)highValue, (long)trie->initialValue);
#endif

    if(highStart<0x110000) {
        /* Blank out [highStart..10ffff] to release associated data blocks. */
        suppHighStart= highStart<=0x10000 ? 0x10000 : highStart;
        utrie2_setRange32(trie, suppHighStart, 0x10ffff, trie->initialValue, TRUE, pErrorCode);
        if(U_FAILURE(*pErrorCode)) {
            return;
        }
    }

    compactData(newTrie);
    if(highStart>0x10000) {
        compactIndex2(newTrie);
#ifdef UTRIE2_DEBUG
    } else {
        printf("UTrie2: highStart U+%04lx  count of 16-bit index words %lu->%lu\n",
                (long)highStart, (long)trie->newTrie->index2Length, (long)UTRIE2_INDEX_1_OFFSET);
#endif
    }

    /*
     * Store the highValue in the data array and round up the dataLength.
     * Must be done after compactData() because that assumes that dataLength
     * is a multiple of UTRIE2_DATA_BLOCK_LENGTH.
     */
    newTrie->data[newTrie->dataLength++]=highValue;
    while((newTrie->dataLength&(UTRIE2_DATA_GRANULARITY-1))!=0) {
        newTrie->data[newTrie->dataLength++]=trie->initialValue;
    }

    newTrie->isCompacted=TRUE;
}

/* serialization ------------------------------------------------------------ */

/**
 * Maximum length of the runtime index array.
 * Limited by its own 16-bit index values, and by uint16_t UTrie2Header.indexLength.
 * (The actual maximum length is lower,
 * (0x110000>>UTRIE2_SHIFT_2)+UTRIE2_UTF8_2B_INDEX_2_LENGTH+UTRIE2_MAX_INDEX_1_LENGTH.)
 */
#define UTRIE2_MAX_INDEX_LENGTH 0xffff

/**
 * Maximum length of the runtime data array.
 * Limited by 16-bit index values that are left-shifted by UTRIE2_INDEX_SHIFT,
 * and by uint16_t UTrie2Header.shiftedDataLength.
 */
#define UTRIE2_MAX_DATA_LENGTH (0xffff<<UTRIE2_INDEX_SHIFT)

/* Compact and internally serialize the trie. */
U_CAPI void U_EXPORT2
utrie2_freeze(UTrie2 *trie, UTrie2ValueBits valueBits, UErrorCode *pErrorCode) {
    UNewTrie2 *newTrie;
    UTrie2Header *header;
    uint32_t *p;
    uint16_t *dest16;
    int32_t i, length;
    int32_t allIndexesLength;
    int32_t dataMove;  /* >0 if the data is moved to the end of the index array */
    UChar32 highStart;

    /* argument check */
    if(U_FAILURE(*pErrorCode)) {
        return;
    }
    if( trie==NULL ||
        valueBits<0 || UTRIE2_COUNT_VALUE_BITS<=valueBits
    ) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    newTrie=trie->newTrie;
    if(newTrie==NULL) {
        /* already frozen */
        UTrie2ValueBits frozenValueBits=
            trie->data16!=NULL ? UTRIE2_16_VALUE_BITS : UTRIE2_32_VALUE_BITS;
        if(valueBits!=frozenValueBits) {
            *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        }
        return;
    }

    /* compact if necessary */
    if(!newTrie->isCompacted) {
        compactTrie(trie, pErrorCode);
        if(U_FAILURE(*pErrorCode)) {
            return;
        }
    }
    highStart=trie->highStart;

    if(highStart<=0x10000) {
        allIndexesLength=UTRIE2_INDEX_1_OFFSET;
    } else {
        allIndexesLength=newTrie->index2Length;
    }
    if(valueBits==UTRIE2_16_VALUE_BITS) {
        dataMove=allIndexesLength;
    } else {
        dataMove=0;
    }

    /* are indexLength and dataLength within limits? */
    if( /* for unshifted indexLength */
        allIndexesLength>UTRIE2_MAX_INDEX_LENGTH ||
        /* for unshifted dataNullOffset */
        (dataMove+newTrie->dataNullOffset)>0xffff ||
        /* for unshifted 2-byte UTF-8 index-2 values */
        (dataMove+UNEWTRIE2_DATA_0800_OFFSET)>0xffff ||
        /* for shiftedDataLength */
        (dataMove+newTrie->dataLength)>UTRIE2_MAX_DATA_LENGTH
    ) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return;
    }

    /* calculate the total serialized length */
    length=sizeof(UTrie2Header)+allIndexesLength*2;
    if(valueBits==UTRIE2_16_VALUE_BITS) {
        length+=newTrie->dataLength*2;
    } else {
        length+=newTrie->dataLength*4;
    }

    trie->memory=uprv_malloc(length);
    if(trie->memory==NULL) {
        *pErrorCode=U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    trie->length=length;
    trie->isMemoryOwned=TRUE;

    trie->indexLength=allIndexesLength;
    trie->dataLength=newTrie->dataLength;
    if(highStart<=0x10000) {
        trie->index2NullOffset=0xffff;
    } else {
        trie->index2NullOffset=static_cast<uint16_t>(UTRIE2_INDEX_2_OFFSET+newTrie->index2NullOffset);
    }
    trie->dataNullOffset=(uint16_t)(dataMove+newTrie->dataNullOffset);
    trie->highValueIndex=dataMove+trie->dataLength-UTRIE2_DATA_GRANULARITY;

    /* set the header fields */
    header=(UTrie2Header *)trie->memory;

    header->signature=UTRIE2_SIG; /* "Tri2" */
    header->options=(uint16_t)valueBits;

    header->indexLength=(uint16_t)trie->indexLength;
    header->shiftedDataLength=(uint16_t)(trie->dataLength>>UTRIE2_INDEX_SHIFT);
    header->index2NullOffset=trie->index2NullOffset;
    header->dataNullOffset=trie->dataNullOffset;
    header->shiftedHighStart=(uint16_t)(highStart>>UTRIE2_SHIFT_1);

    /* fill the index and data arrays */
    dest16=(uint16_t *)(header+1);
    trie->index=dest16;

    /* write the index-2 array values shifted right by UTRIE2_INDEX_SHIFT, after adding dataMove */
    p=(uint32_t *)newTrie->index2;
    for(i=UTRIE2_INDEX_2_BMP_LENGTH; i>0; --i) {
        *dest16++=(uint16_t)((dataMove + *p++)>>UTRIE2_INDEX_SHIFT);
    }

    /* write UTF-8 2-byte index-2 values, not right-shifted */
    for(i=0; i<(0xc2-0xc0); ++i) {                                  /* C0..C1 */
        *dest16++=(uint16_t)(dataMove+UTRIE2_BAD_UTF8_DATA_OFFSET);
    }
    for(; i<(0xe0-0xc0); ++i) {                                     /* C2..DF */
        *dest16++=(uint16_t)(dataMove+newTrie->index2[i<<(6-UTRIE2_SHIFT_2)]);
    }

    if(highStart>0x10000) {
        int32_t index1Length=(highStart-0x10000)>>UTRIE2_SHIFT_1;
        int32_t index2Offset=UTRIE2_INDEX_2_BMP_LENGTH+UTRIE2_UTF8_2B_INDEX_2_LENGTH+index1Length;

        /* write 16-bit index-1 values for supplementary code points */
        p=(uint32_t *)newTrie->index1+UTRIE2_OMITTED_BMP_INDEX_1_LENGTH;
        for(i=index1Length; i>0; --i) {
            *dest16++=(uint16_t)(UTRIE2_INDEX_2_OFFSET + *p++);
        }

        /*
         * write the index-2 array values for supplementary code points,
         * shifted right by UTRIE2_INDEX_SHIFT, after adding dataMove
         */
        p=(uint32_t *)newTrie->index2+index2Offset;
        for(i=newTrie->index2Length-index2Offset; i>0; --i) {
            *dest16++=(uint16_t)((dataMove + *p++)>>UTRIE2_INDEX_SHIFT);
        }
    }

    /* write the 16/32-bit data array */
    switch(valueBits) {
    case UTRIE2_16_VALUE_BITS:
        /* write 16-bit data values */
        trie->data16=dest16;
        trie->data32=NULL;
        p=newTrie->data;
        for(i=newTrie->dataLength; i>0; --i) {
            *dest16++=(uint16_t)*p++;
        }
        break;
    case UTRIE2_32_VALUE_BITS:
        /* write 32-bit data values */
        trie->data16=NULL;
        trie->data32=(uint32_t *)dest16;
        uprv_memcpy(dest16, newTrie->data, (size_t)newTrie->dataLength*4);
        break;
    default:
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

#ifdef UTRIE2_DEBUG
    utrie2_printLengths(trie, "");
#endif

#ifdef UCPTRIE_DEBUG
    umutablecptrie_setName(newTrie->t3, trie->name);
    ucptrie_close(
        umutablecptrie_buildImmutable(
            newTrie->t3, UCPTRIE_TYPE_FAST, (UCPTrieValueWidth)valueBits, pErrorCode));
#endif
    /* Delete the UNewTrie2. */
    uprv_free(newTrie->data);
    uprv_free(newTrie);
    trie->newTrie=NULL;
}
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

// utrie_swap.cpp
// created: 2018aug08 Markus W. Scherer








// These functions for swapping different generations of ICU code point tries are here
// so that their implementation files need not depend on swapper code,
// need not depend on each other, and so that other swapper code
// need not depend on other trie code.

namespace {

constexpr int32_t UTRIE_ASCII_LIMIT = 0x80;

}  // namespace

U_CAPI int32_t U_EXPORT2
utrie_swap(const UDataSwapper *ds,
           const void *inData, int32_t length, void *outData,
           UErrorCode *pErrorCode) {
    const UTrieHeader *inTrie;
    UTrieHeader trie;
    int32_t size;
    UBool dataIs32;

    if(pErrorCode==NULL || U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || (length>=0 && outData==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* setup and swapping */
    if(length>=0 && (uint32_t)length<sizeof(UTrieHeader)) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    inTrie=(const UTrieHeader *)inData;
    trie.signature=ds->readUInt32(inTrie->signature);
    trie.options=ds->readUInt32(inTrie->options);
    trie.indexLength=udata_readInt32(ds, inTrie->indexLength);
    trie.dataLength=udata_readInt32(ds, inTrie->dataLength);

    if( trie.signature!=0x54726965 ||
        (trie.options&UTRIE_OPTIONS_SHIFT_MASK)!=UTRIE_SHIFT ||
        ((trie.options>>UTRIE_OPTIONS_INDEX_SHIFT)&UTRIE_OPTIONS_SHIFT_MASK)!=UTRIE_INDEX_SHIFT ||
        trie.indexLength<UTRIE_BMP_INDEX_LENGTH ||
        (trie.indexLength&(UTRIE_SURROGATE_BLOCK_COUNT-1))!=0 ||
        trie.dataLength<UTRIE_DATA_BLOCK_LENGTH ||
        (trie.dataLength&(UTRIE_DATA_GRANULARITY-1))!=0 ||
        ((trie.options&UTRIE_OPTIONS_LATIN1_IS_LINEAR)!=0 && trie.dataLength<(UTRIE_DATA_BLOCK_LENGTH+0x100))
    ) {
        *pErrorCode=U_INVALID_FORMAT_ERROR; /* not a UTrie */
        return 0;
    }

    dataIs32=(UBool)((trie.options&UTRIE_OPTIONS_DATA_IS_32_BIT)!=0);
    size=sizeof(UTrieHeader)+trie.indexLength*2+trie.dataLength*(dataIs32?4:2);

    if(length>=0) {
        UTrieHeader *outTrie;

        if(length<size) {
            *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }

        outTrie=(UTrieHeader *)outData;

        /* swap the header */
        ds->swapArray32(ds, inTrie, sizeof(UTrieHeader), outTrie, pErrorCode);

        /* swap the index and the data */
        if(dataIs32) {
            ds->swapArray16(ds, inTrie+1, trie.indexLength*2, outTrie+1, pErrorCode);
            ds->swapArray32(ds, (const uint16_t *)(inTrie+1)+trie.indexLength, trie.dataLength*4,
                                     (uint16_t *)(outTrie+1)+trie.indexLength, pErrorCode);
        } else {
            ds->swapArray16(ds, inTrie+1, (trie.indexLength+trie.dataLength)*2, outTrie+1, pErrorCode);
        }
    }

    return size;
}

U_CAPI int32_t U_EXPORT2
utrie2_swap(const UDataSwapper *ds,
            const void *inData, int32_t length, void *outData,
            UErrorCode *pErrorCode) {
    const UTrie2Header *inTrie;
    UTrie2Header trie;
    int32_t dataLength, size;
    UTrie2ValueBits valueBits;

    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==NULL || inData==NULL || (length>=0 && outData==NULL)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* setup and swapping */
    if(length>=0 && length<(int32_t)sizeof(UTrie2Header)) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    inTrie=(const UTrie2Header *)inData;
    trie.signature=ds->readUInt32(inTrie->signature);
    trie.options=ds->readUInt16(inTrie->options);
    trie.indexLength=ds->readUInt16(inTrie->indexLength);
    trie.shiftedDataLength=ds->readUInt16(inTrie->shiftedDataLength);

    valueBits=(UTrie2ValueBits)(trie.options&UTRIE2_OPTIONS_VALUE_BITS_MASK);
    dataLength=(int32_t)trie.shiftedDataLength<<UTRIE2_INDEX_SHIFT;

    if( trie.signature!=UTRIE2_SIG ||
        valueBits<0 || UTRIE2_COUNT_VALUE_BITS<=valueBits ||
        trie.indexLength<UTRIE2_INDEX_1_OFFSET ||
        dataLength<UTRIE2_DATA_START_OFFSET
    ) {
        *pErrorCode=U_INVALID_FORMAT_ERROR; /* not a UTrie */
        return 0;
    }

    size=sizeof(UTrie2Header)+trie.indexLength*2;
    switch(valueBits) {
    case UTRIE2_16_VALUE_BITS:
        size+=dataLength*2;
        break;
    case UTRIE2_32_VALUE_BITS:
        size+=dataLength*4;
        break;
    default:
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return 0;
    }

    if(length>=0) {
        UTrie2Header *outTrie;

        if(length<size) {
            *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }

        outTrie=(UTrie2Header *)outData;

        /* swap the header */
        ds->swapArray32(ds, &inTrie->signature, 4, &outTrie->signature, pErrorCode);
        ds->swapArray16(ds, &inTrie->options, 12, &outTrie->options, pErrorCode);

        /* swap the index and the data */
        switch(valueBits) {
        case UTRIE2_16_VALUE_BITS:
            ds->swapArray16(ds, inTrie+1, (trie.indexLength+dataLength)*2, outTrie+1, pErrorCode);
            break;
        case UTRIE2_32_VALUE_BITS:
            ds->swapArray16(ds, inTrie+1, trie.indexLength*2, outTrie+1, pErrorCode);
            ds->swapArray32(ds, (const uint16_t *)(inTrie+1)+trie.indexLength, dataLength*4,
                                     (uint16_t *)(outTrie+1)+trie.indexLength, pErrorCode);
            break;
        default:
            *pErrorCode=U_INVALID_FORMAT_ERROR;
            return 0;
        }
    }

    return size;
}

U_CAPI int32_t U_EXPORT2
ucptrie_swap(const UDataSwapper *ds,
             const void *inData, int32_t length, void *outData,
             UErrorCode *pErrorCode) {
    const UCPTrieHeader *inTrie;
    UCPTrieHeader trie;
    int32_t dataLength, size;
    UCPTrieValueWidth valueWidth;

    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(ds==nullptr || inData==nullptr || (length>=0 && outData==nullptr)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    /* setup and swapping */
    if(length>=0 && length<(int32_t)sizeof(UCPTrieHeader)) {
        *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
        return 0;
    }

    inTrie=(const UCPTrieHeader *)inData;
    trie.signature=ds->readUInt32(inTrie->signature);
    trie.options=ds->readUInt16(inTrie->options);
    trie.indexLength=ds->readUInt16(inTrie->indexLength);
    trie.dataLength = ds->readUInt16(inTrie->dataLength);

    UCPTrieType type = (UCPTrieType)((trie.options >> 6) & 3);
    valueWidth = (UCPTrieValueWidth)(trie.options & UCPTRIE_OPTIONS_VALUE_BITS_MASK);
    dataLength = ((int32_t)(trie.options & UCPTRIE_OPTIONS_DATA_LENGTH_MASK) << 4) | trie.dataLength;

    int32_t minIndexLength = type == UCPTRIE_TYPE_FAST ?
        UCPTRIE_BMP_INDEX_LENGTH : UCPTRIE_SMALL_INDEX_LENGTH;
    if( trie.signature!=UCPTRIE_SIG ||
        type > UCPTRIE_TYPE_SMALL ||
        (trie.options & UCPTRIE_OPTIONS_RESERVED_MASK) != 0 ||
        valueWidth > UCPTRIE_VALUE_BITS_8 ||
        trie.indexLength < minIndexLength ||
        dataLength < UTRIE_ASCII_LIMIT
    ) {
        *pErrorCode=U_INVALID_FORMAT_ERROR; /* not a UCPTrie */
        return 0;
    }

    size=sizeof(UCPTrieHeader)+trie.indexLength*2;
    switch(valueWidth) {
    case UCPTRIE_VALUE_BITS_16:
        size+=dataLength*2;
        break;
    case UCPTRIE_VALUE_BITS_32:
        size+=dataLength*4;
        break;
    case UCPTRIE_VALUE_BITS_8:
        size+=dataLength;
        break;
    default:
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return 0;
    }

    if(length>=0) {
        UCPTrieHeader *outTrie;

        if(length<size) {
            *pErrorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            return 0;
        }

        outTrie=(UCPTrieHeader *)outData;

        /* swap the header */
        ds->swapArray32(ds, &inTrie->signature, 4, &outTrie->signature, pErrorCode);
        ds->swapArray16(ds, &inTrie->options, 12, &outTrie->options, pErrorCode);

        /* swap the index and the data */
        switch(valueWidth) {
        case UCPTRIE_VALUE_BITS_16:
            ds->swapArray16(ds, inTrie+1, (trie.indexLength+dataLength)*2, outTrie+1, pErrorCode);
            break;
        case UCPTRIE_VALUE_BITS_32:
            ds->swapArray16(ds, inTrie+1, trie.indexLength*2, outTrie+1, pErrorCode);
            ds->swapArray32(ds, (const uint16_t *)(inTrie+1)+trie.indexLength, dataLength*4,
                                     (uint16_t *)(outTrie+1)+trie.indexLength, pErrorCode);
            break;
        case UCPTRIE_VALUE_BITS_8:
            ds->swapArray16(ds, inTrie+1, trie.indexLength*2, outTrie+1, pErrorCode);
            if(inTrie!=outTrie) {
                uprv_memmove((outTrie+1)+trie.indexLength, (inTrie+1)+trie.indexLength, dataLength);
            }
            break;
        default:
            *pErrorCode=U_INVALID_FORMAT_ERROR;
            return 0;
        }
    }

    return size;
}

namespace {

/**
 * Gets the trie version from 32-bit-aligned memory containing the serialized form
 * of a UTrie (version 1), a UTrie2 (version 2), or a UCPTrie (version 3).
 *
 * @param data a pointer to 32-bit-aligned memory containing the serialized form of a trie
 * @param length the number of bytes available at data;
 *               can be more than necessary (see return value)
 * @param anyEndianOk If FALSE, only platform-endian serialized forms are recognized.
 *                    If TRUE, opposite-endian serialized forms are recognized as well.
 * @return the trie version of the serialized form, or 0 if it is not
 *         recognized as a serialized trie
 */
int32_t
getVersion(const void *data, int32_t length, UBool anyEndianOk) {
    uint32_t signature;
    if(length<16 || data==nullptr || (U_POINTER_MASK_LSB(data, 3)!=0)) {
        return 0;
    }
    signature=*(const uint32_t *)data;
    if(signature==UCPTRIE_SIG) {
        return 3;
    }
    if(anyEndianOk && signature==UCPTRIE_OE_SIG) {
        return 3;
    }
    if(signature==UTRIE2_SIG) {
        return 2;
    }
    if(anyEndianOk && signature==UTRIE2_OE_SIG) {
        return 2;
    }
    if(signature==UTRIE_SIG) {
        return 1;
    }
    if(anyEndianOk && signature==UTRIE_OE_SIG) {
        return 1;
    }
    return 0;
}

}  // namespace

U_CAPI int32_t U_EXPORT2
utrie_swapAnyVersion(const UDataSwapper *ds,
                     const void *inData, int32_t length, void *outData,
                     UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) { return 0; }
    switch(getVersion(inData, length, TRUE)) {
    case 1:
        return utrie_swap(ds, inData, length, outData, pErrorCode);
    case 2:
        return utrie2_swap(ds, inData, length, outData, pErrorCode);
    case 3:
        return ucptrie_swap(ds, inData, length, outData, pErrorCode);
    default:
        *pErrorCode=U_INVALID_FORMAT_ERROR;
        return 0;
    }
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  uts46.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010mar09
*   created by: Markus W. Scherer
*/



#if !UCONFIG_NO_IDNA

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2010-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  idna.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2010mar05
*   created by: Markus W. Scherer
*/


/**
 * \file
 * \brief C++ API: Internationalizing Domain Names in Applications (IDNA)
 */



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_IDNA



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 *
 *   Copyright (C) 2003-2014, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 *
 *******************************************************************************
 *   file name:  uidna.h
 *   encoding:   UTF-8
 *   tab size:   8 (not used)
 *   indentation:4
 *
 *   created on: 2003feb1
 *   created by: Ram Viswanadha
 */




#if !UCONFIG_NO_IDNA




/**
 * \file
 * \brief C API: Internationalizing Domain Names in Applications (IDNA)
 *
 * IDNA2008 is implemented according to UTS #46, see the IDNA C++ class in idna.h.
 *
 * The C API functions which do take a UIDNA * service object pointer
 * implement UTS #46 and IDNA2008.
 *
 * IDNA2003 is obsolete.
 * The C API functions which do not take a service object pointer
 * implement IDNA2003. They are all deprecated.
 */

/*
 * IDNA option bit set values.
 */
enum {
    /**
     * Default options value: None of the other options are set.
     * For use in static worker and factory methods.
     * @stable ICU 2.6
     */
    UIDNA_DEFAULT=0,
#ifndef U_HIDE_DEPRECATED_API
    /**
     * Option to allow unassigned code points in domain names and labels.
     * For use in static worker and factory methods.
     * <p>This option is ignored by the UTS46 implementation.
     * (UTS #46 disallows unassigned code points.)
     * @deprecated ICU 55 Use UTS #46 instead via uidna_openUTS46() or class IDNA.
     */
    UIDNA_ALLOW_UNASSIGNED=1,
#endif  /* U_HIDE_DEPRECATED_API */
    /**
     * Option to check whether the input conforms to the STD3 ASCII rules,
     * for example the restriction of labels to LDH characters
     * (ASCII Letters, Digits and Hyphen-Minus).
     * For use in static worker and factory methods.
     * @stable ICU 2.6
     */
    UIDNA_USE_STD3_RULES=2,
    /**
     * IDNA option to check for whether the input conforms to the BiDi rules.
     * For use in static worker and factory methods.
     * <p>This option is ignored by the IDNA2003 implementation.
     * (IDNA2003 always performs a BiDi check.)
     * @stable ICU 4.6
     */
    UIDNA_CHECK_BIDI=4,
    /**
     * IDNA option to check for whether the input conforms to the CONTEXTJ rules.
     * For use in static worker and factory methods.
     * <p>This option is ignored by the IDNA2003 implementation.
     * (The CONTEXTJ check is new in IDNA2008.)
     * @stable ICU 4.6
     */
    UIDNA_CHECK_CONTEXTJ=8,
    /**
     * IDNA option for nontransitional processing in ToASCII().
     * For use in static worker and factory methods.
     * <p>By default, ToASCII() uses transitional processing.
     * <p>This option is ignored by the IDNA2003 implementation.
     * (This is only relevant for compatibility of newer IDNA implementations with IDNA2003.)
     * @stable ICU 4.6
     */
    UIDNA_NONTRANSITIONAL_TO_ASCII=0x10,
    /**
     * IDNA option for nontransitional processing in ToUnicode().
     * For use in static worker and factory methods.
     * <p>By default, ToUnicode() uses transitional processing.
     * <p>This option is ignored by the IDNA2003 implementation.
     * (This is only relevant for compatibility of newer IDNA implementations with IDNA2003.)
     * @stable ICU 4.6
     */
    UIDNA_NONTRANSITIONAL_TO_UNICODE=0x20,
    /**
     * IDNA option to check for whether the input conforms to the CONTEXTO rules.
     * For use in static worker and factory methods.
     * <p>This option is ignored by the IDNA2003 implementation.
     * (The CONTEXTO check is new in IDNA2008.)
     * <p>This is for use by registries for IDNA2008 conformance.
     * UTS #46 does not require the CONTEXTO check.
     * @stable ICU 49
     */
    UIDNA_CHECK_CONTEXTO=0x40
};

/**
 * Opaque C service object type for the new IDNA API.
 * @stable ICU 4.6
 */
struct UIDNA;
typedef struct UIDNA UIDNA;  /**< C typedef for struct UIDNA. @stable ICU 4.6 */

/**
 * Returns a UIDNA instance which implements UTS #46.
 * Returns an unmodifiable instance, owned by the caller.
 * Cache it for multiple operations, and uidna_close() it when done.
 * The instance is thread-safe, that is, it can be used concurrently.
 *
 * For details about the UTS #46 implementation see the IDNA C++ class in idna.h.
 *
 * @param options Bit set to modify the processing and error checking.
 *                See option bit set values in uidna.h.
 * @param pErrorCode Standard ICU error code. Its input value must
 *                  pass the U_SUCCESS() test, or else the function returns
 *                  immediately. Check for U_FAILURE() on output or use with
 *                  function chaining. (See User Guide for details.)
 * @return the UTS #46 UIDNA instance, if successful
 * @stable ICU 4.6
 */
U_STABLE UIDNA * U_EXPORT2
uidna_openUTS46(uint32_t options, UErrorCode *pErrorCode);

/**
 * Closes a UIDNA instance.
 * @param idna UIDNA instance to be closed
 * @stable ICU 4.6
 */
U_STABLE void U_EXPORT2
uidna_close(UIDNA *idna);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUIDNAPointer
 * "Smart pointer" class, closes a UIDNA via uidna_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.6
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUIDNAPointer, UIDNA, uidna_close);

U_NAMESPACE_END

#endif

/**
 * Output container for IDNA processing errors.
 * Initialize with UIDNA_INFO_INITIALIZER:
 * \code
 * UIDNAInfo info = UIDNA_INFO_INITIALIZER;
 * int32_t length = uidna_nameToASCII(..., &info, &errorCode);
 * if(U_SUCCESS(errorCode) && info.errors!=0) { ... }
 * \endcode
 * @stable ICU 4.6
 */
typedef struct UIDNAInfo {
    /** sizeof(UIDNAInfo) @stable ICU 4.6 */
    int16_t size;
    /**
     * Set to TRUE if transitional and nontransitional processing produce different results.
     * For details see C++ IDNAInfo::isTransitionalDifferent().
     * @stable ICU 4.6
     */
    UBool isTransitionalDifferent;
    UBool reservedB3;  /**< Reserved field, do not use. @internal */
    /**
     * Bit set indicating IDNA processing errors. 0 if no errors.
     * See UIDNA_ERROR_... constants.
     * @stable ICU 4.6
     */
    uint32_t errors;
    int32_t reservedI2;  /**< Reserved field, do not use. @internal */
    int32_t reservedI3;  /**< Reserved field, do not use. @internal */
} UIDNAInfo;

/**
 * Static initializer for a UIDNAInfo struct.
 * @stable ICU 4.6
 */
#define UIDNA_INFO_INITIALIZER { \
    (int16_t)sizeof(UIDNAInfo), \
    FALSE, FALSE, \
    0, 0, 0 }

/**
 * Converts a single domain name label into its ASCII form for DNS lookup.
 * If any processing step fails, then pInfo->errors will be non-zero and
 * the result might not be an ASCII string.
 * The label might be modified according to the types of errors.
 * Labels with severe errors will be left in (or turned into) their Unicode form.
 *
 * The UErrorCode indicates an error only in exceptional cases,
 * such as a U_MEMORY_ALLOCATION_ERROR.
 *
 * @param idna UIDNA instance
 * @param label Input domain name label
 * @param length Label length, or -1 if NUL-terminated
 * @param dest Destination string buffer
 * @param capacity Destination buffer capacity
 * @param pInfo Output container of IDNA processing details.
 * @param pErrorCode Standard ICU error code. Its input value must
 *                  pass the U_SUCCESS() test, or else the function returns
 *                  immediately. Check for U_FAILURE() on output or use with
 *                  function chaining. (See User Guide for details.)
 * @return destination string length
 * @stable ICU 4.6
 */
U_STABLE int32_t U_EXPORT2
uidna_labelToASCII(const UIDNA *idna,
                   const UChar *label, int32_t length,
                   UChar *dest, int32_t capacity,
                   UIDNAInfo *pInfo, UErrorCode *pErrorCode);

/**
 * Converts a single domain name label into its Unicode form for human-readable display.
 * If any processing step fails, then pInfo->errors will be non-zero.
 * The label might be modified according to the types of errors.
 *
 * The UErrorCode indicates an error only in exceptional cases,
 * such as a U_MEMORY_ALLOCATION_ERROR.
 *
 * @param idna UIDNA instance
 * @param label Input domain name label
 * @param length Label length, or -1 if NUL-terminated
 * @param dest Destination string buffer
 * @param capacity Destination buffer capacity
 * @param pInfo Output container of IDNA processing details.
 * @param pErrorCode Standard ICU error code. Its input value must
 *                  pass the U_SUCCESS() test, or else the function returns
 *                  immediately. Check for U_FAILURE() on output or use with
 *                  function chaining. (See User Guide for details.)
 * @return destination string length
 * @stable ICU 4.6
 */
U_STABLE int32_t U_EXPORT2
uidna_labelToUnicode(const UIDNA *idna,
                     const UChar *label, int32_t length,
                     UChar *dest, int32_t capacity,
                     UIDNAInfo *pInfo, UErrorCode *pErrorCode);

/**
 * Converts a whole domain name into its ASCII form for DNS lookup.
 * If any processing step fails, then pInfo->errors will be non-zero and
 * the result might not be an ASCII string.
 * The domain name might be modified according to the types of errors.
 * Labels with severe errors will be left in (or turned into) their Unicode form.
 *
 * The UErrorCode indicates an error only in exceptional cases,
 * such as a U_MEMORY_ALLOCATION_ERROR.
 *
 * @param idna UIDNA instance
 * @param name Input domain name
 * @param length Domain name length, or -1 if NUL-terminated
 * @param dest Destination string buffer
 * @param capacity Destination buffer capacity
 * @param pInfo Output container of IDNA processing details.
 * @param pErrorCode Standard ICU error code. Its input value must
 *                  pass the U_SUCCESS() test, or else the function returns
 *                  immediately. Check for U_FAILURE() on output or use with
 *                  function chaining. (See User Guide for details.)
 * @return destination string length
 * @stable ICU 4.6
 */
U_STABLE int32_t U_EXPORT2
uidna_nameToASCII(const UIDNA *idna,
                  const UChar *name, int32_t length,
                  UChar *dest, int32_t capacity,
                  UIDNAInfo *pInfo, UErrorCode *pErrorCode);

/**
 * Converts a whole domain name into its Unicode form for human-readable display.
 * If any processing step fails, then pInfo->errors will be non-zero.
 * The domain name might be modified according to the types of errors.
 *
 * The UErrorCode indicates an error only in exceptional cases,
 * such as a U_MEMORY_ALLOCATION_ERROR.
 *
 * @param idna UIDNA instance
 * @param name Input domain name
 * @param length Domain name length, or -1 if NUL-terminated
 * @param dest Destination string buffer
 * @param capacity Destination buffer capacity
 * @param pInfo Output container of IDNA processing details.
 * @param pErrorCode Standard ICU error code. Its input value must
 *                  pass the U_SUCCESS() test, or else the function returns
 *                  immediately. Check for U_FAILURE() on output or use with
 *                  function chaining. (See User Guide for details.)
 * @return destination string length
 * @stable ICU 4.6
 */
U_STABLE int32_t U_EXPORT2
uidna_nameToUnicode(const UIDNA *idna,
                    const UChar *name, int32_t length,
                    UChar *dest, int32_t capacity,
                    UIDNAInfo *pInfo, UErrorCode *pErrorCode);

/* UTF-8 versions of the processing methods --------------------------------- */

/**
 * Converts a single domain name label into its ASCII form for DNS lookup.
 * UTF-8 version of uidna_labelToASCII(), same behavior.
 *
 * @param idna UIDNA instance
 * @param label Input domain name label
 * @param length Label length, or -1 if NUL-terminated
 * @param dest Destination string buffer
 * @param capacity Destination buffer capacity
 * @param pInfo Output container of IDNA processing details.
 * @param pErrorCode Standard ICU error code. Its input value must
 *                  pass the U_SUCCESS() test, or else the function returns
 *                  immediately. Check for U_FAILURE() on output or use with
 *                  function chaining. (See User Guide for details.)
 * @return destination string length
 * @stable ICU 4.6
 */
U_STABLE int32_t U_EXPORT2
uidna_labelToASCII_UTF8(const UIDNA *idna,
                        const char *label, int32_t length,
                        char *dest, int32_t capacity,
                        UIDNAInfo *pInfo, UErrorCode *pErrorCode);

/**
 * Converts a single domain name label into its Unicode form for human-readable display.
 * UTF-8 version of uidna_labelToUnicode(), same behavior.
 *
 * @param idna UIDNA instance
 * @param label Input domain name label
 * @param length Label length, or -1 if NUL-terminated
 * @param dest Destination string buffer
 * @param capacity Destination buffer capacity
 * @param pInfo Output container of IDNA processing details.
 * @param pErrorCode Standard ICU error code. Its input value must
 *                  pass the U_SUCCESS() test, or else the function returns
 *                  immediately. Check for U_FAILURE() on output or use with
 *                  function chaining. (See User Guide for details.)
 * @return destination string length
 * @stable ICU 4.6
 */
U_STABLE int32_t U_EXPORT2
uidna_labelToUnicodeUTF8(const UIDNA *idna,
                         const char *label, int32_t length,
                         char *dest, int32_t capacity,
                         UIDNAInfo *pInfo, UErrorCode *pErrorCode);

/**
 * Converts a whole domain name into its ASCII form for DNS lookup.
 * UTF-8 version of uidna_nameToASCII(), same behavior.
 *
 * @param idna UIDNA instance
 * @param name Input domain name
 * @param length Domain name length, or -1 if NUL-terminated
 * @param dest Destination string buffer
 * @param capacity Destination buffer capacity
 * @param pInfo Output container of IDNA processing details.
 * @param pErrorCode Standard ICU error code. Its input value must
 *                  pass the U_SUCCESS() test, or else the function returns
 *                  immediately. Check for U_FAILURE() on output or use with
 *                  function chaining. (See User Guide for details.)
 * @return destination string length
 * @stable ICU 4.6
 */
U_STABLE int32_t U_EXPORT2
uidna_nameToASCII_UTF8(const UIDNA *idna,
                       const char *name, int32_t length,
                       char *dest, int32_t capacity,
                       UIDNAInfo *pInfo, UErrorCode *pErrorCode);

/**
 * Converts a whole domain name into its Unicode form for human-readable display.
 * UTF-8 version of uidna_nameToUnicode(), same behavior.
 *
 * @param idna UIDNA instance
 * @param name Input domain name
 * @param length Domain name length, or -1 if NUL-terminated
 * @param dest Destination string buffer
 * @param capacity Destination buffer capacity
 * @param pInfo Output container of IDNA processing details.
 * @param pErrorCode Standard ICU error code. Its input value must
 *                  pass the U_SUCCESS() test, or else the function returns
 *                  immediately. Check for U_FAILURE() on output or use with
 *                  function chaining. (See User Guide for details.)
 * @return destination string length
 * @stable ICU 4.6
 */
U_STABLE int32_t U_EXPORT2
uidna_nameToUnicodeUTF8(const UIDNA *idna,
                        const char *name, int32_t length,
                        char *dest, int32_t capacity,
                        UIDNAInfo *pInfo, UErrorCode *pErrorCode);

/*
 * IDNA error bit set values.
 * When a domain name or label fails a processing step or does not meet the
 * validity criteria, then one or more of these error bits are set.
 */
enum {
    /**
     * A non-final domain name label (or the whole domain name) is empty.
     * @stable ICU 4.6
     */
    UIDNA_ERROR_EMPTY_LABEL=1,
    /**
     * A domain name label is longer than 63 bytes.
     * (See STD13/RFC1034 3.1. Name space specifications and terminology.)
     * This is only checked in ToASCII operations, and only if the output label is all-ASCII.
     * @stable ICU 4.6
     */
    UIDNA_ERROR_LABEL_TOO_LONG=2,
    /**
     * A domain name is longer than 255 bytes in its storage form.
     * (See STD13/RFC1034 3.1. Name space specifications and terminology.)
     * This is only checked in ToASCII operations, and only if the output domain name is all-ASCII.
     * @stable ICU 4.6
     */
    UIDNA_ERROR_DOMAIN_NAME_TOO_LONG=4,
    /**
     * A label starts with a hyphen-minus ('-').
     * @stable ICU 4.6
     */
    UIDNA_ERROR_LEADING_HYPHEN=8,
    /**
     * A label ends with a hyphen-minus ('-').
     * @stable ICU 4.6
     */
    UIDNA_ERROR_TRAILING_HYPHEN=0x10,
    /**
     * A label contains hyphen-minus ('-') in the third and fourth positions.
     * @stable ICU 4.6
     */
    UIDNA_ERROR_HYPHEN_3_4=0x20,
    /**
     * A label starts with a combining mark.
     * @stable ICU 4.6
     */
    UIDNA_ERROR_LEADING_COMBINING_MARK=0x40,
    /**
     * A label or domain name contains disallowed characters.
     * @stable ICU 4.6
     */
    UIDNA_ERROR_DISALLOWED=0x80,
    /**
     * A label starts with "xn--" but does not contain valid Punycode.
     * That is, an xn-- label failed Punycode decoding.
     * @stable ICU 4.6
     */
    UIDNA_ERROR_PUNYCODE=0x100,
    /**
     * A label contains a dot=full stop.
     * This can occur in an input string for a single-label function.
     * @stable ICU 4.6
     */
    UIDNA_ERROR_LABEL_HAS_DOT=0x200,
    /**
     * An ACE label does not contain a valid label string.
     * The label was successfully ACE (Punycode) decoded but the resulting
     * string had severe validation errors. For example,
     * it might contain characters that are not allowed in ACE labels,
     * or it might not be normalized.
     * @stable ICU 4.6
     */
    UIDNA_ERROR_INVALID_ACE_LABEL=0x400,
    /**
     * A label does not meet the IDNA BiDi requirements (for right-to-left characters).
     * @stable ICU 4.6
     */
    UIDNA_ERROR_BIDI=0x800,
    /**
     * A label does not meet the IDNA CONTEXTJ requirements.
     * @stable ICU 4.6
     */
    UIDNA_ERROR_CONTEXTJ=0x1000,
    /**
     * A label does not meet the IDNA CONTEXTO requirements for punctuation characters.
     * Some punctuation characters "Would otherwise have been DISALLOWED"
     * but are allowed in certain contexts. (RFC 5892)
     * @stable ICU 49
     */
    UIDNA_ERROR_CONTEXTO_PUNCTUATION=0x2000,
    /**
     * A label does not meet the IDNA CONTEXTO requirements for digits.
     * Arabic-Indic Digits (U+066x) must not be mixed with Extended Arabic-Indic Digits (U+06Fx).
     * @stable ICU 49
     */
    UIDNA_ERROR_CONTEXTO_DIGITS=0x4000
};

#ifndef U_HIDE_DEPRECATED_API

/* IDNA2003 API ------------------------------------------------------------- */

/**
 * IDNA2003: This function implements the ToASCII operation as defined in the IDNA RFC.
 * This operation is done on <b>single labels</b> before sending it to something that expects
 * ASCII names. A label is an individual part of a domain name. Labels are usually
 * separated by dots; e.g. "www.example.com" is composed of 3 labels "www","example", and "com".
 *
 * IDNA2003 API Overview:
 *
 * The uidna_ API implements the IDNA protocol as defined in the IDNA RFC
 * (http://www.ietf.org/rfc/rfc3490.txt).
 * The RFC defines 2 operations: ToASCII and ToUnicode. Domain name labels
 * containing non-ASCII code points are processed by the
 * ToASCII operation before passing it to resolver libraries. Domain names
 * that are obtained from resolver libraries are processed by the
 * ToUnicode operation before displaying the domain name to the user.
 * IDNA requires that implementations process input strings with Nameprep
 * (http://www.ietf.org/rfc/rfc3491.txt),
 * which is a profile of Stringprep (http://www.ietf.org/rfc/rfc3454.txt),
 * and then with Punycode (http://www.ietf.org/rfc/rfc3492.txt).
 * Implementations of IDNA MUST fully implement Nameprep and Punycode;
 * neither Nameprep nor Punycode are optional.
 * The input and output of ToASCII and ToUnicode operations are Unicode
 * and are designed to be chainable, i.e., applying ToASCII or ToUnicode operations
 * multiple times to an input string will yield the same result as applying the operation
 * once.
 * ToUnicode(ToUnicode(ToUnicode...(ToUnicode(string)))) == ToUnicode(string) 
 * ToASCII(ToASCII(ToASCII...(ToASCII(string))) == ToASCII(string).
 *
 * @param src               Input UChar array containing label in Unicode.
 * @param srcLength         Number of UChars in src, or -1 if NUL-terminated.
 * @param dest              Output UChar array with ASCII (ACE encoded) label.
 * @param destCapacity      Size of dest.
 * @param options           A bit set of options:
 *
 *  - UIDNA_DEFAULT             Use default options, i.e., do not process unassigned code points
 *                              and do not use STD3 ASCII rules
 *                              If unassigned code points are found the operation fails with 
 *                              U_UNASSIGNED_ERROR error code.
 *
 *  - UIDNA_ALLOW_UNASSIGNED    Unassigned values can be converted to ASCII for query operations
 *                              If this option is set, the unassigned code points are in the input 
 *                              are treated as normal Unicode code points.
 *
 *  - UIDNA_USE_STD3_RULES      Use STD3 ASCII rules for host name syntax restrictions
 *                              If this option is set and the input does not satisfy STD3 rules,  
 *                              the operation will fail with U_IDNA_STD3_ASCII_RULES_ERROR
 *
 * @param parseError        Pointer to UParseError struct to receive information on position 
 *                          of error if an error is encountered. Can be NULL.
 * @param status            ICU in/out error code parameter.
 *                          U_INVALID_CHAR_FOUND if src contains
 *                          unmatched single surrogates.
 *                          U_INDEX_OUTOFBOUNDS_ERROR if src contains
 *                          too many code points.
 *                          U_BUFFER_OVERFLOW_ERROR if destCapacity is not enough
 * @return The length of the result string, if successful - or in case of a buffer overflow,
 *         in which case it will be greater than destCapacity.
 * @deprecated ICU 55 Use UTS #46 instead via uidna_openUTS46() or class IDNA.
 */
U_DEPRECATED int32_t U_EXPORT2
uidna_toASCII(const UChar* src, int32_t srcLength, 
              UChar* dest, int32_t destCapacity,
              int32_t options,
              UParseError* parseError,
              UErrorCode* status);


/**
 * IDNA2003: This function implements the ToUnicode operation as defined in the IDNA RFC.
 * This operation is done on <b>single labels</b> before sending it to something that expects
 * Unicode names. A label is an individual part of a domain name. Labels are usually
 * separated by dots; for e.g. "www.example.com" is composed of 3 labels "www","example", and "com".
 *
 * @param src               Input UChar array containing ASCII (ACE encoded) label.
 * @param srcLength         Number of UChars in src, or -1 if NUL-terminated.
 * @param dest Output       Converted UChar array containing Unicode equivalent of label.
 * @param destCapacity      Size of dest.
 * @param options           A bit set of options:
 *
 *  - UIDNA_DEFAULT             Use default options, i.e., do not process unassigned code points
 *                              and do not use STD3 ASCII rules
 *                              If unassigned code points are found the operation fails with 
 *                              U_UNASSIGNED_ERROR error code.
 *
 *  - UIDNA_ALLOW_UNASSIGNED      Unassigned values can be converted to ASCII for query operations
 *                              If this option is set, the unassigned code points are in the input 
 *                              are treated as normal Unicode code points. <b> Note: </b> This option is 
 *                              required on toUnicode operation because the RFC mandates 
 *                              verification of decoded ACE input by applying toASCII and comparing
 *                              its output with source
 *
 *  - UIDNA_USE_STD3_RULES      Use STD3 ASCII rules for host name syntax restrictions
 *                              If this option is set and the input does not satisfy STD3 rules,  
 *                              the operation will fail with U_IDNA_STD3_ASCII_RULES_ERROR
 *
 * @param parseError        Pointer to UParseError struct to receive information on position 
 *                          of error if an error is encountered. Can be NULL.
 * @param status            ICU in/out error code parameter.
 *                          U_INVALID_CHAR_FOUND if src contains
 *                          unmatched single surrogates.
 *                          U_INDEX_OUTOFBOUNDS_ERROR if src contains
 *                          too many code points.
 *                          U_BUFFER_OVERFLOW_ERROR if destCapacity is not enough
 * @return The length of the result string, if successful - or in case of a buffer overflow,
 *         in which case it will be greater than destCapacity.
 * @deprecated ICU 55 Use UTS #46 instead via uidna_openUTS46() or class IDNA.
 */
U_DEPRECATED int32_t U_EXPORT2
uidna_toUnicode(const UChar* src, int32_t srcLength,
                UChar* dest, int32_t destCapacity,
                int32_t options,
                UParseError* parseError,
                UErrorCode* status);


/**
 * IDNA2003: Convenience function that implements the IDNToASCII operation as defined in the IDNA RFC.
 * This operation is done on complete domain names, e.g: "www.example.com". 
 * It is important to note that this operation can fail. If it fails, then the input 
 * domain name cannot be used as an Internationalized Domain Name and the application
 * should have methods defined to deal with the failure.
 *
 * <b>Note:</b> IDNA RFC specifies that a conformant application should divide a domain name
 * into separate labels, decide whether to apply allowUnassigned and useSTD3ASCIIRules on each, 
 * and then convert. This function does not offer that level of granularity. The options once  
 * set will apply to all labels in the domain name
 *
 * @param src               Input UChar array containing IDN in Unicode.
 * @param srcLength         Number of UChars in src, or -1 if NUL-terminated.
 * @param dest              Output UChar array with ASCII (ACE encoded) IDN.
 * @param destCapacity      Size of dest.
 * @param options           A bit set of options:
 *
 *  - UIDNA_DEFAULT             Use default options, i.e., do not process unassigned code points
 *                              and do not use STD3 ASCII rules
 *                              If unassigned code points are found the operation fails with 
 *                              U_UNASSIGNED_CODE_POINT_FOUND error code.
 *
 *  - UIDNA_ALLOW_UNASSIGNED    Unassigned values can be converted to ASCII for query operations
 *                              If this option is set, the unassigned code points are in the input 
 *                              are treated as normal Unicode code points.
 *
 *  - UIDNA_USE_STD3_RULES      Use STD3 ASCII rules for host name syntax restrictions
 *                              If this option is set and the input does not satisfy STD3 rules,  
 *                              the operation will fail with U_IDNA_STD3_ASCII_RULES_ERROR
 *
 * @param parseError        Pointer to UParseError struct to receive information on position 
 *                          of error if an error is encountered. Can be NULL.
 * @param status            ICU in/out error code parameter.
 *                          U_INVALID_CHAR_FOUND if src contains
 *                          unmatched single surrogates.
 *                          U_INDEX_OUTOFBOUNDS_ERROR if src contains
 *                          too many code points.
 *                          U_BUFFER_OVERFLOW_ERROR if destCapacity is not enough
 * @return The length of the result string, if successful - or in case of a buffer overflow,
 *         in which case it will be greater than destCapacity.
 * @deprecated ICU 55 Use UTS #46 instead via uidna_openUTS46() or class IDNA.
 */
U_DEPRECATED int32_t U_EXPORT2
uidna_IDNToASCII(  const UChar* src, int32_t srcLength,
                   UChar* dest, int32_t destCapacity,
                   int32_t options,
                   UParseError* parseError,
                   UErrorCode* status);

/**
 * IDNA2003: Convenience function that implements the IDNToUnicode operation as defined in the IDNA RFC.
 * This operation is done on complete domain names, e.g: "www.example.com". 
 *
 * <b>Note:</b> IDNA RFC specifies that a conformant application should divide a domain name
 * into separate labels, decide whether to apply allowUnassigned and useSTD3ASCIIRules on each, 
 * and then convert. This function does not offer that level of granularity. The options once  
 * set will apply to all labels in the domain name
 *
 * @param src               Input UChar array containing IDN in ASCII (ACE encoded) form.
 * @param srcLength         Number of UChars in src, or -1 if NUL-terminated.
 * @param dest Output       UChar array containing Unicode equivalent of source IDN.
 * @param destCapacity      Size of dest.
 * @param options           A bit set of options:
 *
 *  - UIDNA_DEFAULT             Use default options, i.e., do not process unassigned code points
 *                              and do not use STD3 ASCII rules
 *                              If unassigned code points are found the operation fails with 
 *                              U_UNASSIGNED_CODE_POINT_FOUND error code.
 *
 *  - UIDNA_ALLOW_UNASSIGNED    Unassigned values can be converted to ASCII for query operations
 *                              If this option is set, the unassigned code points are in the input 
 *                              are treated as normal Unicode code points.
 *
 *  - UIDNA_USE_STD3_RULES      Use STD3 ASCII rules for host name syntax restrictions
 *                              If this option is set and the input does not satisfy STD3 rules,  
 *                              the operation will fail with U_IDNA_STD3_ASCII_RULES_ERROR
 *
 * @param parseError        Pointer to UParseError struct to receive information on position 
 *                          of error if an error is encountered. Can be NULL.
 * @param status            ICU in/out error code parameter.
 *                          U_INVALID_CHAR_FOUND if src contains
 *                          unmatched single surrogates.
 *                          U_INDEX_OUTOFBOUNDS_ERROR if src contains
 *                          too many code points.
 *                          U_BUFFER_OVERFLOW_ERROR if destCapacity is not enough
 * @return The length of the result string, if successful - or in case of a buffer overflow,
 *         in which case it will be greater than destCapacity.
 * @deprecated ICU 55 Use UTS #46 instead via uidna_openUTS46() or class IDNA.
 */
U_DEPRECATED int32_t U_EXPORT2
uidna_IDNToUnicode(  const UChar* src, int32_t srcLength,
                     UChar* dest, int32_t destCapacity,
                     int32_t options,
                     UParseError* parseError,
                     UErrorCode* status);

/**
 * IDNA2003: Compare two IDN strings for equivalence.
 * This function splits the domain names into labels and compares them.
 * According to IDN RFC, whenever two labels are compared, they are 
 * considered equal if and only if their ASCII forms (obtained by 
 * applying toASCII) match using an case-insensitive ASCII comparison.
 * Two domain names are considered a match if and only if all labels 
 * match regardless of whether label separators match.
 *
 * @param s1                First source string.
 * @param length1           Length of first source string, or -1 if NUL-terminated.
 *
 * @param s2                Second source string.
 * @param length2           Length of second source string, or -1 if NUL-terminated.
 * @param options           A bit set of options:
 *
 *  - UIDNA_DEFAULT             Use default options, i.e., do not process unassigned code points
 *                              and do not use STD3 ASCII rules
 *                              If unassigned code points are found the operation fails with 
 *                              U_UNASSIGNED_CODE_POINT_FOUND error code.
 *
 *  - UIDNA_ALLOW_UNASSIGNED    Unassigned values can be converted to ASCII for query operations
 *                              If this option is set, the unassigned code points are in the input 
 *                              are treated as normal Unicode code points.
 *
 *  - UIDNA_USE_STD3_RULES      Use STD3 ASCII rules for host name syntax restrictions
 *                              If this option is set and the input does not satisfy STD3 rules,  
 *                              the operation will fail with U_IDNA_STD3_ASCII_RULES_ERROR
 *
 * @param status            ICU error code in/out parameter.
 *                          Must fulfill U_SUCCESS before the function call.
 * @return <0 or 0 or >0 as usual for string comparisons
 * @deprecated ICU 55 Use UTS #46 instead via uidna_openUTS46() or class IDNA.
 */
U_DEPRECATED int32_t U_EXPORT2
uidna_compare(  const UChar *s1, int32_t length1,
                const UChar *s2, int32_t length2,
                int32_t options,
                UErrorCode* status);

#endif  /* U_HIDE_DEPRECATED_API */

#endif /* #if !UCONFIG_NO_IDNA */




U_NAMESPACE_BEGIN

class IDNAInfo;

/**
 * Abstract base class for IDNA processing.
 * See http://www.unicode.org/reports/tr46/
 * and http://www.ietf.org/rfc/rfc3490.txt
 *
 * The IDNA class is not intended for public subclassing.
 *
 * This C++ API currently only implements UTS #46.
 * The uidna.h C API implements both UTS #46 (functions using UIDNA service object)
 * and IDNA2003 (functions that do not use a service object).
 * @stable ICU 4.6
 */
class U_COMMON_API IDNA : public UObject {
public:
    /**
     * Destructor.
     * @stable ICU 4.6
     */
    ~IDNA();

    /**
     * Returns an IDNA instance which implements UTS #46.
     * Returns an unmodifiable instance, owned by the caller.
     * Cache it for multiple operations, and delete it when done.
     * The instance is thread-safe, that is, it can be used concurrently.
     *
     * UTS #46 defines Unicode IDNA Compatibility Processing,
     * updated to the latest version of Unicode and compatible with both
     * IDNA2003 and IDNA2008.
     *
     * The worker functions use transitional processing, including deviation mappings,
     * unless UIDNA_NONTRANSITIONAL_TO_ASCII or UIDNA_NONTRANSITIONAL_TO_UNICODE
     * is used in which case the deviation characters are passed through without change.
     *
     * Disallowed characters are mapped to U+FFFD.
     *
     * For available options see the uidna.h header.
     * Operations with the UTS #46 instance do not support the
     * UIDNA_ALLOW_UNASSIGNED option.
     *
     * By default, the UTS #46 implementation allows all ASCII characters (as valid or mapped).
     * When the UIDNA_USE_STD3_RULES option is used, ASCII characters other than
     * letters, digits, hyphen (LDH) and dot/full stop are disallowed and mapped to U+FFFD.
     *
     * @param options Bit set to modify the processing and error checking.
     *                See option bit set values in uidna.h.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return the UTS #46 IDNA instance, if successful
     * @stable ICU 4.6
     */
    static IDNA *
    createUTS46Instance(uint32_t options, UErrorCode &errorCode);

    /**
     * Converts a single domain name label into its ASCII form for DNS lookup.
     * If any processing step fails, then info.hasErrors() will be TRUE and
     * the result might not be an ASCII string.
     * The label might be modified according to the types of errors.
     * Labels with severe errors will be left in (or turned into) their Unicode form.
     *
     * The UErrorCode indicates an error only in exceptional cases,
     * such as a U_MEMORY_ALLOCATION_ERROR.
     *
     * @param label Input domain name label
     * @param dest Destination string object
     * @param info Output container of IDNA processing details.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return dest
     * @stable ICU 4.6
     */
    virtual UnicodeString &
    labelToASCII(const UnicodeString &label, UnicodeString &dest,
                 IDNAInfo &info, UErrorCode &errorCode) const = 0;

    /**
     * Converts a single domain name label into its Unicode form for human-readable display.
     * If any processing step fails, then info.hasErrors() will be TRUE.
     * The label might be modified according to the types of errors.
     *
     * The UErrorCode indicates an error only in exceptional cases,
     * such as a U_MEMORY_ALLOCATION_ERROR.
     *
     * @param label Input domain name label
     * @param dest Destination string object
     * @param info Output container of IDNA processing details.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return dest
     * @stable ICU 4.6
     */
    virtual UnicodeString &
    labelToUnicode(const UnicodeString &label, UnicodeString &dest,
                   IDNAInfo &info, UErrorCode &errorCode) const = 0;

    /**
     * Converts a whole domain name into its ASCII form for DNS lookup.
     * If any processing step fails, then info.hasErrors() will be TRUE and
     * the result might not be an ASCII string.
     * The domain name might be modified according to the types of errors.
     * Labels with severe errors will be left in (or turned into) their Unicode form.
     *
     * The UErrorCode indicates an error only in exceptional cases,
     * such as a U_MEMORY_ALLOCATION_ERROR.
     *
     * @param name Input domain name
     * @param dest Destination string object
     * @param info Output container of IDNA processing details.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return dest
     * @stable ICU 4.6
     */
    virtual UnicodeString &
    nameToASCII(const UnicodeString &name, UnicodeString &dest,
                IDNAInfo &info, UErrorCode &errorCode) const = 0;

    /**
     * Converts a whole domain name into its Unicode form for human-readable display.
     * If any processing step fails, then info.hasErrors() will be TRUE.
     * The domain name might be modified according to the types of errors.
     *
     * The UErrorCode indicates an error only in exceptional cases,
     * such as a U_MEMORY_ALLOCATION_ERROR.
     *
     * @param name Input domain name
     * @param dest Destination string object
     * @param info Output container of IDNA processing details.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return dest
     * @stable ICU 4.6
     */
    virtual UnicodeString &
    nameToUnicode(const UnicodeString &name, UnicodeString &dest,
                  IDNAInfo &info, UErrorCode &errorCode) const = 0;

    // UTF-8 versions of the processing methods ---------------------------- ***

    /**
     * Converts a single domain name label into its ASCII form for DNS lookup.
     * UTF-8 version of labelToASCII(), same behavior.
     *
     * @param label Input domain name label
     * @param dest Destination byte sink; Flush()ed if successful
     * @param info Output container of IDNA processing details.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return dest
     * @stable ICU 4.6
     */
    virtual void
    labelToASCII_UTF8(StringPiece label, ByteSink &dest,
                      IDNAInfo &info, UErrorCode &errorCode) const;

    /**
     * Converts a single domain name label into its Unicode form for human-readable display.
     * UTF-8 version of labelToUnicode(), same behavior.
     *
     * @param label Input domain name label
     * @param dest Destination byte sink; Flush()ed if successful
     * @param info Output container of IDNA processing details.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return dest
     * @stable ICU 4.6
     */
    virtual void
    labelToUnicodeUTF8(StringPiece label, ByteSink &dest,
                       IDNAInfo &info, UErrorCode &errorCode) const;

    /**
     * Converts a whole domain name into its ASCII form for DNS lookup.
     * UTF-8 version of nameToASCII(), same behavior.
     *
     * @param name Input domain name
     * @param dest Destination byte sink; Flush()ed if successful
     * @param info Output container of IDNA processing details.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return dest
     * @stable ICU 4.6
     */
    virtual void
    nameToASCII_UTF8(StringPiece name, ByteSink &dest,
                     IDNAInfo &info, UErrorCode &errorCode) const;

    /**
     * Converts a whole domain name into its Unicode form for human-readable display.
     * UTF-8 version of nameToUnicode(), same behavior.
     *
     * @param name Input domain name
     * @param dest Destination byte sink; Flush()ed if successful
     * @param info Output container of IDNA processing details.
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return dest
     * @stable ICU 4.6
     */
    virtual void
    nameToUnicodeUTF8(StringPiece name, ByteSink &dest,
                      IDNAInfo &info, UErrorCode &errorCode) const;
};

class UTS46;

/**
 * Output container for IDNA processing errors.
 * The IDNAInfo class is not suitable for subclassing.
 * @stable ICU 4.6
 */
class U_COMMON_API IDNAInfo : public UMemory {
public:
    /**
     * Constructor for stack allocation.
     * @stable ICU 4.6
     */
    IDNAInfo() : errors(0), labelErrors(0), isTransDiff(FALSE), isBiDi(FALSE), isOkBiDi(TRUE) {}
    /**
     * Were there IDNA processing errors?
     * @return TRUE if there were processing errors
     * @stable ICU 4.6
     */
    UBool hasErrors() const { return errors!=0; }
    /**
     * Returns a bit set indicating IDNA processing errors.
     * See UIDNA_ERROR_... constants in uidna.h.
     * @return bit set of processing errors
     * @stable ICU 4.6
     */
    uint32_t getErrors() const { return errors; }
    /**
     * Returns TRUE if transitional and nontransitional processing produce different results.
     * This is the case when the input label or domain name contains
     * one or more deviation characters outside a Punycode label (see UTS #46).
     * <ul>
     * <li>With nontransitional processing, such characters are
     * copied to the destination string.
     * <li>With transitional processing, such characters are
     * mapped (sharp s/sigma) or removed (joiner/nonjoiner).
     * </ul>
     * @return TRUE if transitional and nontransitional processing produce different results
     * @stable ICU 4.6
     */
    UBool isTransitionalDifferent() const { return isTransDiff; }

private:
    friend class UTS46;

    IDNAInfo(const IDNAInfo &other);  // no copying
    IDNAInfo &operator=(const IDNAInfo &other);  // no copying

    void reset() {
        errors=labelErrors=0;
        isTransDiff=FALSE;
        isBiDi=FALSE;
        isOkBiDi=TRUE;
    }

    uint32_t errors, labelErrors;
    UBool isTransDiff;
    UBool isBiDi;
    UBool isOkBiDi;
};

U_NAMESPACE_END

#endif  // UCONFIG_NO_IDNA

#endif /* U_SHOW_CPLUSPLUS_API */








// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2002-2003, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  punycode.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2002jan31
*   created by: Markus W. Scherer
*/

/* This ICU code derived from: */
/*
punycode.c 0.4.0 (2001-Nov-17-Sat)
http://www.cs.berkeley.edu/~amc/idn/
Adam M. Costello
http://www.nicemice.net/amc/
*/




#if !UCONFIG_NO_IDNA

/**
 * u_strToPunycode() converts Unicode to Punycode.
 *
 * The input string must not contain single, unpaired surrogates.
 * The output will be represented as an array of ASCII code points.
 *
 * The output string is NUL-terminated according to normal ICU
 * string output rules.
 *
 * @param src Input Unicode string.
 *            This function handles a limited amount of code points
 *            (the limit is >=64).
 *            U_INDEX_OUTOFBOUNDS_ERROR is set if the limit is exceeded.
 * @param srcLength Number of UChars in src, or -1 if NUL-terminated.
 * @param dest Output Punycode array.
 * @param destCapacity Size of dest.
 * @param caseFlags Vector of boolean values, one per input UChar,
 *                  indicating that the corresponding character is to be
 *                  marked for the decoder optionally
 *                  uppercasing (TRUE) or lowercasing (FALSE)
 *                  the character.
 *                  ASCII characters are output directly in the case as marked.
 *                  Flags corresponding to trail surrogates are ignored.
 *                  If caseFlags==NULL then input characters are not
 *                  case-mapped.
 * @param pErrorCode ICU in/out error code parameter.
 *                   U_INVALID_CHAR_FOUND if src contains
 *                   unmatched single surrogates.
 *                   U_INDEX_OUTOFBOUNDS_ERROR if src contains
 *                   too many code points.
 * @return Number of ASCII characters in puny.
 *
 * @see u_strFromPunycode
 */
U_CFUNC int32_t
u_strToPunycode(const UChar *src, int32_t srcLength,
                UChar *dest, int32_t destCapacity,
                const UBool *caseFlags,
                UErrorCode *pErrorCode);

/**
 * u_strFromPunycode() converts Punycode to Unicode.
 * The Unicode string will be at most as long (in UChars)
 * than the Punycode string (in chars).
 *
 * @param src Input Punycode string.
 * @param srcLength Length of puny, or -1 if NUL-terminated
 * @param dest Output Unicode string buffer.
 * @param destCapacity Size of dest in number of UChars,
 *                     and of caseFlags in numbers of UBools.
 * @param caseFlags Output array for case flags as
 *                  defined by the Punycode string.
 *                  The caller should uppercase (TRUE) or lowercase (FASLE)
 *                  the corresponding character in dest.
 *                  For supplementary characters, only the lead surrogate
 *                  is marked, and FALSE is stored for the trail surrogate.
 *                  This is redundant and not necessary for ASCII characters
 *                  because they are already in the case indicated.
 *                  Can be NULL if the case flags are not needed.
 * @param pErrorCode ICU in/out error code parameter.
 *                   U_INVALID_CHAR_FOUND if a non-ASCII character
 *                   precedes the last delimiter ('-'),
 *                   or if an invalid character (not a-zA-Z0-9) is found
 *                   after the last delimiter.
 *                   U_ILLEGAL_CHAR_FOUND if the delta sequence is ill-formed.
 * @return Number of UChars written to dest.
 *
 * @see u_strToPunycode
 */
U_CFUNC int32_t
u_strFromPunycode(const UChar *src, int32_t srcLength,
                  UChar *dest, int32_t destCapacity,
                  UBool *caseFlags,
                  UErrorCode *pErrorCode);

#endif /* #if !UCONFIG_NO_IDNA */





// Note about tests for UIDNA_ERROR_DOMAIN_NAME_TOO_LONG:
//
// The domain name length limit is 255 octets in an internal DNS representation
// where the last ("root") label is the empty label
// represented by length byte 0 alone.
// In a conventional string, this translates to 253 characters, or 254
// if there is a trailing dot for the root label.

U_NAMESPACE_BEGIN

// Severe errors which usually result in a U+FFFD replacement character in the result string.
const uint32_t severeErrors=
    UIDNA_ERROR_LEADING_COMBINING_MARK|
    UIDNA_ERROR_DISALLOWED|
    UIDNA_ERROR_PUNYCODE|
    UIDNA_ERROR_LABEL_HAS_DOT|
    UIDNA_ERROR_INVALID_ACE_LABEL;

static inline UBool
isASCIIString(const UnicodeString &dest) {
    const UChar *s=dest.getBuffer();
    const UChar *limit=s+dest.length();
    while(s<limit) {
        if(*s++>0x7f) {
            return FALSE;
        }
    }
    return TRUE;
}

static UBool
isASCIIOkBiDi(const UChar *s, int32_t length);

static UBool
isASCIIOkBiDi(const char *s, int32_t length);

// IDNA class default implementations -------------------------------------- ***

IDNA::~IDNA() {}

void
IDNA::labelToASCII_UTF8(StringPiece label, ByteSink &dest,
                        IDNAInfo &info, UErrorCode &errorCode) const {
    if(U_SUCCESS(errorCode)) {
        UnicodeString destString;
        labelToASCII(UnicodeString::fromUTF8(label), destString,
                     info, errorCode).toUTF8(dest);
    }
}

void
IDNA::labelToUnicodeUTF8(StringPiece label, ByteSink &dest,
                         IDNAInfo &info, UErrorCode &errorCode) const {
    if(U_SUCCESS(errorCode)) {
        UnicodeString destString;
        labelToUnicode(UnicodeString::fromUTF8(label), destString,
                       info, errorCode).toUTF8(dest);
    }
}

void
IDNA::nameToASCII_UTF8(StringPiece name, ByteSink &dest,
                       IDNAInfo &info, UErrorCode &errorCode) const {
    if(U_SUCCESS(errorCode)) {
        UnicodeString destString;
        nameToASCII(UnicodeString::fromUTF8(name), destString,
                    info, errorCode).toUTF8(dest);
    }
}

void
IDNA::nameToUnicodeUTF8(StringPiece name, ByteSink &dest,
                        IDNAInfo &info, UErrorCode &errorCode) const {
    if(U_SUCCESS(errorCode)) {
        UnicodeString destString;
        nameToUnicode(UnicodeString::fromUTF8(name), destString,
                      info, errorCode).toUTF8(dest);
    }
}

// UTS46 class declaration ------------------------------------------------- ***

class UTS46 : public IDNA {
public:
    UTS46(uint32_t options, UErrorCode &errorCode);
    virtual ~UTS46();

    virtual UnicodeString &
    labelToASCII(const UnicodeString &label, UnicodeString &dest,
                 IDNAInfo &info, UErrorCode &errorCode) const;

    virtual UnicodeString &
    labelToUnicode(const UnicodeString &label, UnicodeString &dest,
                   IDNAInfo &info, UErrorCode &errorCode) const;

    virtual UnicodeString &
    nameToASCII(const UnicodeString &name, UnicodeString &dest,
                IDNAInfo &info, UErrorCode &errorCode) const;

    virtual UnicodeString &
    nameToUnicode(const UnicodeString &name, UnicodeString &dest,
                  IDNAInfo &info, UErrorCode &errorCode) const;

    virtual void
    labelToASCII_UTF8(StringPiece label, ByteSink &dest,
                      IDNAInfo &info, UErrorCode &errorCode) const;

    virtual void
    labelToUnicodeUTF8(StringPiece label, ByteSink &dest,
                       IDNAInfo &info, UErrorCode &errorCode) const;

    virtual void
    nameToASCII_UTF8(StringPiece name, ByteSink &dest,
                     IDNAInfo &info, UErrorCode &errorCode) const;

    virtual void
    nameToUnicodeUTF8(StringPiece name, ByteSink &dest,
                      IDNAInfo &info, UErrorCode &errorCode) const;

private:
    UnicodeString &
    process(const UnicodeString &src,
            UBool isLabel, UBool toASCII,
            UnicodeString &dest,
            IDNAInfo &info, UErrorCode &errorCode) const;

    void
    processUTF8(StringPiece src,
                UBool isLabel, UBool toASCII,
                ByteSink &dest,
                IDNAInfo &info, UErrorCode &errorCode) const;

    UnicodeString &
    processUnicode(const UnicodeString &src,
                   int32_t labelStart, int32_t mappingStart,
                   UBool isLabel, UBool toASCII,
                   UnicodeString &dest,
                   IDNAInfo &info, UErrorCode &errorCode) const;

    // returns the new dest.length()
    int32_t
    mapDevChars(UnicodeString &dest, int32_t labelStart, int32_t mappingStart,
                UErrorCode &errorCode) const;

    // returns the new label length
    int32_t
    processLabel(UnicodeString &dest,
                 int32_t labelStart, int32_t labelLength,
                 UBool toASCII,
                 IDNAInfo &info, UErrorCode &errorCode) const;
    int32_t
    markBadACELabel(UnicodeString &dest,
                    int32_t labelStart, int32_t labelLength,
                    UBool toASCII, IDNAInfo &info, UErrorCode &errorCode) const;

    void
    checkLabelBiDi(const UChar *label, int32_t labelLength, IDNAInfo &info) const;

    UBool
    isLabelOkContextJ(const UChar *label, int32_t labelLength) const;

    void
    checkLabelContextO(const UChar *label, int32_t labelLength, IDNAInfo &info) const;

    const Normalizer2 &uts46Norm2;  // uts46.nrm
    uint32_t options;
};

IDNA *
IDNA::createUTS46Instance(uint32_t options, UErrorCode &errorCode) {
    if(U_SUCCESS(errorCode)) {
        IDNA *idna=new UTS46(options, errorCode);
        if(idna==NULL) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
        } else if(U_FAILURE(errorCode)) {
            delete idna;
            idna=NULL;
        }
        return idna;
    } else {
        return NULL;
    }
}

// UTS46 implementation ---------------------------------------------------- ***

UTS46::UTS46(uint32_t opt, UErrorCode &errorCode)
        : uts46Norm2(*Normalizer2::getInstance(NULL, "uts46", UNORM2_COMPOSE, errorCode)),
          options(opt) {}

UTS46::~UTS46() {}

UnicodeString &
UTS46::labelToASCII(const UnicodeString &label, UnicodeString &dest,
                    IDNAInfo &info, UErrorCode &errorCode) const {
    return process(label, TRUE, TRUE, dest, info, errorCode);
}

UnicodeString &
UTS46::labelToUnicode(const UnicodeString &label, UnicodeString &dest,
                      IDNAInfo &info, UErrorCode &errorCode) const {
    return process(label, TRUE, FALSE, dest, info, errorCode);
}

UnicodeString &
UTS46::nameToASCII(const UnicodeString &name, UnicodeString &dest,
                   IDNAInfo &info, UErrorCode &errorCode) const {
    process(name, FALSE, TRUE, dest, info, errorCode);
    if( dest.length()>=254 && (info.errors&UIDNA_ERROR_DOMAIN_NAME_TOO_LONG)==0 &&
        isASCIIString(dest) &&
        (dest.length()>254 || dest[253]!=0x2e)
    ) {
        info.errors|=UIDNA_ERROR_DOMAIN_NAME_TOO_LONG;
    }
    return dest;
}

UnicodeString &
UTS46::nameToUnicode(const UnicodeString &name, UnicodeString &dest,
                     IDNAInfo &info, UErrorCode &errorCode) const {
    return process(name, FALSE, FALSE, dest, info, errorCode);
}

void
UTS46::labelToASCII_UTF8(StringPiece label, ByteSink &dest,
                         IDNAInfo &info, UErrorCode &errorCode) const {
    processUTF8(label, TRUE, TRUE, dest, info, errorCode);
}

void
UTS46::labelToUnicodeUTF8(StringPiece label, ByteSink &dest,
                          IDNAInfo &info, UErrorCode &errorCode) const {
    processUTF8(label, TRUE, FALSE, dest, info, errorCode);
}

void
UTS46::nameToASCII_UTF8(StringPiece name, ByteSink &dest,
                        IDNAInfo &info, UErrorCode &errorCode) const {
    processUTF8(name, FALSE, TRUE, dest, info, errorCode);
}

void
UTS46::nameToUnicodeUTF8(StringPiece name, ByteSink &dest,
                         IDNAInfo &info, UErrorCode &errorCode) const {
    processUTF8(name, FALSE, FALSE, dest, info, errorCode);
}

// UTS #46 data for ASCII characters.
// The normalizer (using uts46.nrm) maps uppercase ASCII letters to lowercase
// and passes through all other ASCII characters.
// If UIDNA_USE_STD3_RULES is set, then non-LDH characters are disallowed
// using this data.
// The ASCII fastpath also uses this data.
// Values: -1=disallowed  0==valid  1==mapped (lowercase)
static const int8_t asciiData[128]={
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    // 002D..002E; valid  #  HYPHEN-MINUS..FULL STOP
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  0,  0, -1,
    // 0030..0039; valid  #  DIGIT ZERO..DIGIT NINE
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1, -1, -1, -1, -1, -1,
    // 0041..005A; mapped  #  LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    -1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
     1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, -1, -1, -1, -1, -1,
    // 0061..007A; valid  #  LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1, -1, -1, -1, -1
};

UnicodeString &
UTS46::process(const UnicodeString &src,
               UBool isLabel, UBool toASCII,
               UnicodeString &dest,
               IDNAInfo &info, UErrorCode &errorCode) const {
    // uts46Norm2.normalize() would do all of this error checking and setup,
    // but with the ASCII fastpath we do not always call it, and do not
    // call it first.
    if(U_FAILURE(errorCode)) {
        dest.setToBogus();
        return dest;
    }
    const UChar *srcArray=src.getBuffer();
    if(&dest==&src || srcArray==NULL) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        dest.setToBogus();
        return dest;
    }
    // Arguments are fine, reset output values.
    dest.remove();
    info.reset();
    int32_t srcLength=src.length();
    if(srcLength==0) {
        info.errors|=UIDNA_ERROR_EMPTY_LABEL;
        return dest;
    }
    UChar *destArray=dest.getBuffer(srcLength);
    if(destArray==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return dest;
    }
    // ASCII fastpath
    UBool disallowNonLDHDot=(options&UIDNA_USE_STD3_RULES)!=0;
    int32_t labelStart=0;
    int32_t i;
    for(i=0;; ++i) {
        if(i==srcLength) {
            if(toASCII) {
                if((i-labelStart)>63) {
                    info.labelErrors|=UIDNA_ERROR_LABEL_TOO_LONG;
                }
                // There is a trailing dot if labelStart==i.
                if(!isLabel && i>=254 && (i>254 || labelStart<i)) {
                    info.errors|=UIDNA_ERROR_DOMAIN_NAME_TOO_LONG;
                }
            }
            info.errors|=info.labelErrors;
            dest.releaseBuffer(i);
            return dest;
        }
        UChar c=srcArray[i];
        if(c>0x7f) {
            break;
        }
        int cData=asciiData[c];
        if(cData>0) {
            destArray[i]=c+0x20;  // Lowercase an uppercase ASCII letter.
        } else if(cData<0 && disallowNonLDHDot) {
            break;  // Replacing with U+FFFD can be complicated for toASCII.
        } else {
            destArray[i]=c;
            if(c==0x2d) {  // hyphen
                if(i==(labelStart+3) && srcArray[i-1]==0x2d) {
                    // "??--..." is Punycode or forbidden.
                    ++i;  // '-' was copied to dest already
                    break;
                }
                if(i==labelStart) {
                    // label starts with "-"
                    info.labelErrors|=UIDNA_ERROR_LEADING_HYPHEN;
                }
                if((i+1)==srcLength || srcArray[i+1]==0x2e) {
                    // label ends with "-"
                    info.labelErrors|=UIDNA_ERROR_TRAILING_HYPHEN;
                }
            } else if(c==0x2e) {  // dot
                if(isLabel) {
                    // Replacing with U+FFFD can be complicated for toASCII.
                    ++i;  // '.' was copied to dest already
                    break;
                }
                if(i==labelStart) {
                    info.labelErrors|=UIDNA_ERROR_EMPTY_LABEL;
                }
                if(toASCII && (i-labelStart)>63) {
                    info.labelErrors|=UIDNA_ERROR_LABEL_TOO_LONG;
                }
                info.errors|=info.labelErrors;
                info.labelErrors=0;
                labelStart=i+1;
            }
        }
    }
    info.errors|=info.labelErrors;
    dest.releaseBuffer(i);
    processUnicode(src, labelStart, i, isLabel, toASCII, dest, info, errorCode);
    if( info.isBiDi && U_SUCCESS(errorCode) && (info.errors&severeErrors)==0 &&
        (!info.isOkBiDi || (labelStart>0 && !isASCIIOkBiDi(dest.getBuffer(), labelStart)))
    ) {
        info.errors|=UIDNA_ERROR_BIDI;
    }
    return dest;
}

void
UTS46::processUTF8(StringPiece src,
                   UBool isLabel, UBool toASCII,
                   ByteSink &dest,
                   IDNAInfo &info, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return;
    }
    const char *srcArray=src.data();
    int32_t srcLength=src.length();
    if(srcArray==NULL && srcLength!=0) {
        errorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    // Arguments are fine, reset output values.
    info.reset();
    if(srcLength==0) {
        info.errors|=UIDNA_ERROR_EMPTY_LABEL;
        dest.Flush();
        return;
    }
    UnicodeString destString;
    int32_t labelStart=0;
    if(srcLength<=256) {  // length of stackArray[]
        // ASCII fastpath
        char stackArray[256];
        int32_t destCapacity;
        char *destArray=dest.GetAppendBuffer(srcLength, srcLength+20,
                                             stackArray, UPRV_LENGTHOF(stackArray), &destCapacity);
        UBool disallowNonLDHDot=(options&UIDNA_USE_STD3_RULES)!=0;
        int32_t i;
        for(i=0;; ++i) {
            if(i==srcLength) {
                if(toASCII) {
                    if((i-labelStart)>63) {
                        info.labelErrors|=UIDNA_ERROR_LABEL_TOO_LONG;
                    }
                    // There is a trailing dot if labelStart==i.
                    if(!isLabel && i>=254 && (i>254 || labelStart<i)) {
                        info.errors|=UIDNA_ERROR_DOMAIN_NAME_TOO_LONG;
                    }
                }
                info.errors|=info.labelErrors;
                dest.Append(destArray, i);
                dest.Flush();
                return;
            }
            char c=srcArray[i];
            if((int8_t)c<0) {  // (uint8_t)c>0x7f
                break;
            }
            int cData=asciiData[(int)c];  // Cast: gcc warns about indexing with a char.
            if(cData>0) {
                destArray[i]=c+0x20;  // Lowercase an uppercase ASCII letter.
            } else if(cData<0 && disallowNonLDHDot) {
                break;  // Replacing with U+FFFD can be complicated for toASCII.
            } else {
                destArray[i]=c;
                if(c==0x2d) {  // hyphen
                    if(i==(labelStart+3) && srcArray[i-1]==0x2d) {
                        // "??--..." is Punycode or forbidden.
                        break;
                    }
                    if(i==labelStart) {
                        // label starts with "-"
                        info.labelErrors|=UIDNA_ERROR_LEADING_HYPHEN;
                    }
                    if((i+1)==srcLength || srcArray[i+1]==0x2e) {
                        // label ends with "-"
                        info.labelErrors|=UIDNA_ERROR_TRAILING_HYPHEN;
                    }
                } else if(c==0x2e) {  // dot
                    if(isLabel) {
                        break;  // Replacing with U+FFFD can be complicated for toASCII.
                    }
                    if(i==labelStart) {
                        info.labelErrors|=UIDNA_ERROR_EMPTY_LABEL;
                    }
                    if(toASCII && (i-labelStart)>63) {
                        info.labelErrors|=UIDNA_ERROR_LABEL_TOO_LONG;
                    }
                    info.errors|=info.labelErrors;
                    info.labelErrors=0;
                    labelStart=i+1;
                }
            }
        }
        info.errors|=info.labelErrors;
        // Convert the processed ASCII prefix of the current label to UTF-16.
        int32_t mappingStart=i-labelStart;
        destString=UnicodeString::fromUTF8(StringPiece(destArray+labelStart, mappingStart));
        // Output the previous ASCII labels and process the rest of src in UTF-16.
        dest.Append(destArray, labelStart);
        processUnicode(UnicodeString::fromUTF8(StringPiece(src, labelStart)), 0, mappingStart,
                       isLabel, toASCII,
                       destString, info, errorCode);
    } else {
        // src is too long for the ASCII fastpath implementation.
        processUnicode(UnicodeString::fromUTF8(src), 0, 0,
                       isLabel, toASCII,
                       destString, info, errorCode);
    }
    destString.toUTF8(dest);  // calls dest.Flush()
    if(toASCII && !isLabel) {
        // length==labelStart==254 means that there is a trailing dot (ok) and
        // destString is empty (do not index at 253-labelStart).
        int32_t length=labelStart+destString.length();
        if( length>=254 && isASCIIString(destString) &&
            (length>254 ||
             (labelStart<254 && destString[253-labelStart]!=0x2e))
        ) {
            info.errors|=UIDNA_ERROR_DOMAIN_NAME_TOO_LONG;
        }
    }
    if( info.isBiDi && U_SUCCESS(errorCode) && (info.errors&severeErrors)==0 &&
        (!info.isOkBiDi || (labelStart>0 && !isASCIIOkBiDi(srcArray, labelStart)))
    ) {
        info.errors|=UIDNA_ERROR_BIDI;
    }
}

UnicodeString &
UTS46::processUnicode(const UnicodeString &src,
                      int32_t labelStart, int32_t mappingStart,
                      UBool isLabel, UBool toASCII,
                      UnicodeString &dest,
                      IDNAInfo &info, UErrorCode &errorCode) const {
    if(mappingStart==0) {
        uts46Norm2.normalize(src, dest, errorCode);
    } else {
        uts46Norm2.normalizeSecondAndAppend(dest, src.tempSubString(mappingStart), errorCode);
    }
    if(U_FAILURE(errorCode)) {
        return dest;
    }
    UBool doMapDevChars=
        toASCII ? (options&UIDNA_NONTRANSITIONAL_TO_ASCII)==0 :
                  (options&UIDNA_NONTRANSITIONAL_TO_UNICODE)==0;
    const UChar *destArray=dest.getBuffer();
    int32_t destLength=dest.length();
    int32_t labelLimit=labelStart;
    while(labelLimit<destLength) {
        UChar c=destArray[labelLimit];
        if(c==0x2e && !isLabel) {
            int32_t labelLength=labelLimit-labelStart;
            int32_t newLength=processLabel(dest, labelStart, labelLength,
                                            toASCII, info, errorCode);
            info.errors|=info.labelErrors;
            info.labelErrors=0;
            if(U_FAILURE(errorCode)) {
                return dest;
            }
            destArray=dest.getBuffer();
            destLength+=newLength-labelLength;
            labelLimit=labelStart+=newLength+1;
            continue;
        } else if(c<0xdf) {
            // pass
        } else if(c<=0x200d && (c==0xdf || c==0x3c2 || c>=0x200c)) {
            info.isTransDiff=TRUE;
            if(doMapDevChars) {
                destLength=mapDevChars(dest, labelStart, labelLimit, errorCode);
                if(U_FAILURE(errorCode)) {
                    return dest;
                }
                destArray=dest.getBuffer();
                // All deviation characters have been mapped, no need to check for them again.
                doMapDevChars=FALSE;
                // Do not increment labelLimit in case c was removed.
                continue;
            }
        } else if(U16_IS_SURROGATE(c)) {
            if(U16_IS_SURROGATE_LEAD(c) ?
                    (labelLimit+1)==destLength || !U16_IS_TRAIL(destArray[labelLimit+1]) :
                    labelLimit==labelStart || !U16_IS_LEAD(destArray[labelLimit-1])) {
                // Map an unpaired surrogate to U+FFFD before normalization so that when
                // that removes characters we do not turn two unpaired ones into a pair.
                info.labelErrors|=UIDNA_ERROR_DISALLOWED;
                dest.setCharAt(labelLimit, 0xfffd);
                destArray=dest.getBuffer();
            }
        }
        ++labelLimit;
    }
    // Permit an empty label at the end (0<labelStart==labelLimit==destLength is ok)
    // but not an empty label elsewhere nor a completely empty domain name.
    // processLabel() sets UIDNA_ERROR_EMPTY_LABEL when labelLength==0.
    if(0==labelStart || labelStart<labelLimit) {
        processLabel(dest, labelStart, labelLimit-labelStart,
                      toASCII, info, errorCode);
        info.errors|=info.labelErrors;
    }
    return dest;
}

int32_t
UTS46::mapDevChars(UnicodeString &dest, int32_t labelStart, int32_t mappingStart,
                   UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    int32_t length=dest.length();
    UChar *s=dest.getBuffer(dest[mappingStart]==0xdf ? length+1 : length);
    if(s==NULL) {
        errorCode=U_MEMORY_ALLOCATION_ERROR;
        return length;
    }
    int32_t capacity=dest.getCapacity();
    UBool didMapDevChars=FALSE;
    int32_t readIndex=mappingStart, writeIndex=mappingStart;
    do {
        UChar c=s[readIndex++];
        switch(c) {
        case 0xdf:
            // Map sharp s to ss.
            didMapDevChars=TRUE;
            s[writeIndex++]=0x73;  // Replace sharp s with first s.
            // Insert second s and account for possible buffer reallocation.
            if(writeIndex==readIndex) {
                if(length==capacity) {
                    dest.releaseBuffer(length);
                    s=dest.getBuffer(length+1);
                    if(s==NULL) {
                        errorCode=U_MEMORY_ALLOCATION_ERROR;
                        return length;
                    }
                    capacity=dest.getCapacity();
                }
                u_memmove(s+writeIndex+1, s+writeIndex, length-writeIndex);
                ++readIndex;
            }
            s[writeIndex++]=0x73;
            ++length;
            break;
        case 0x3c2:  // Map final sigma to nonfinal sigma.
            didMapDevChars=TRUE;
            s[writeIndex++]=0x3c3;
            break;
        case 0x200c:  // Ignore/remove ZWNJ.
        case 0x200d:  // Ignore/remove ZWJ.
            didMapDevChars=TRUE;
            --length;
            break;
        default:
            // Only really necessary if writeIndex was different from readIndex.
            s[writeIndex++]=c;
            break;
        }
    } while(writeIndex<length);
    dest.releaseBuffer(length);
    if(didMapDevChars) {
        // Mapping deviation characters might have resulted in an un-NFC string.
        // We could use either the NFC or the UTS #46 normalizer.
        // By using the UTS #46 normalizer again, we avoid having to load a second .nrm data file.
        UnicodeString normalized;
        uts46Norm2.normalize(dest.tempSubString(labelStart), normalized, errorCode);
        if(U_SUCCESS(errorCode)) {
            dest.replace(labelStart, 0x7fffffff, normalized);
            if(dest.isBogus()) {
                errorCode=U_MEMORY_ALLOCATION_ERROR;
            }
            return dest.length();
        }
    }
    return length;
}

// Some non-ASCII characters are equivalent to sequences with
// non-LDH ASCII characters. To find them:
// grep disallowed_STD3_valid IdnaMappingTable.txt (or uts46.txt)
static inline UBool
isNonASCIIDisallowedSTD3Valid(UChar32 c) {
    return c==0x2260 || c==0x226E || c==0x226F;
}

// Replace the label in dest with the label string, if the label was modified.
// If &label==&dest then the label was modified in-place and labelLength
// is the new label length, different from label.length().
// If &label!=&dest then labelLength==label.length().
// Returns labelLength (= the new label length).
static int32_t
replaceLabel(UnicodeString &dest, int32_t destLabelStart, int32_t destLabelLength,
             const UnicodeString &label, int32_t labelLength, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    if(&label!=&dest) {
        dest.replace(destLabelStart, destLabelLength, label);
        if(dest.isBogus()) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            return 0;
        }
    }
    return labelLength;
}

int32_t
UTS46::processLabel(UnicodeString &dest,
                    int32_t labelStart, int32_t labelLength,
                    UBool toASCII,
                    IDNAInfo &info, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    UnicodeString fromPunycode;
    UnicodeString *labelString;
    const UChar *label=dest.getBuffer()+labelStart;
    int32_t destLabelStart=labelStart;
    int32_t destLabelLength=labelLength;
    UBool wasPunycode;
    if(labelLength>=4 && label[0]==0x78 && label[1]==0x6e && label[2]==0x2d && label[3]==0x2d) {
        // Label starts with "xn--", try to un-Punycode it.
        wasPunycode=TRUE;
        UChar *unicodeBuffer=fromPunycode.getBuffer(-1);  // capacity==-1: most labels should fit
        if(unicodeBuffer==NULL) {
            // Should never occur if we used capacity==-1 which uses the internal buffer.
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            return labelLength;
        }
        UErrorCode punycodeErrorCode=U_ZERO_ERROR;
        int32_t unicodeLength=u_strFromPunycode(label+4, labelLength-4,
                                                unicodeBuffer, fromPunycode.getCapacity(),
                                                NULL, &punycodeErrorCode);
        if(punycodeErrorCode==U_BUFFER_OVERFLOW_ERROR) {
            fromPunycode.releaseBuffer(0);
            unicodeBuffer=fromPunycode.getBuffer(unicodeLength);
            if(unicodeBuffer==NULL) {
                errorCode=U_MEMORY_ALLOCATION_ERROR;
                return labelLength;
            }
            punycodeErrorCode=U_ZERO_ERROR;
            unicodeLength=u_strFromPunycode(label+4, labelLength-4,
                                            unicodeBuffer, fromPunycode.getCapacity(),
                                            NULL, &punycodeErrorCode);
        }
        fromPunycode.releaseBuffer(unicodeLength);
        if(U_FAILURE(punycodeErrorCode)) {
            info.labelErrors|=UIDNA_ERROR_PUNYCODE;
            return markBadACELabel(dest, labelStart, labelLength, toASCII, info, errorCode);
        }
        // Check for NFC, and for characters that are not
        // valid or deviation characters according to the normalizer.
        // If there is something wrong, then the string will change.
        // Note that the normalizer passes through non-LDH ASCII and deviation characters.
        // Deviation characters are ok in Punycode even in transitional processing.
        // In the code further below, if we find non-LDH ASCII and we have UIDNA_USE_STD3_RULES
        // then we will set UIDNA_ERROR_INVALID_ACE_LABEL there too.
        UBool isValid=uts46Norm2.isNormalized(fromPunycode, errorCode);
        if(U_FAILURE(errorCode)) {
            return labelLength;
        }
        if(!isValid) {
            info.labelErrors|=UIDNA_ERROR_INVALID_ACE_LABEL;
            return markBadACELabel(dest, labelStart, labelLength, toASCII, info, errorCode);
        }
        labelString=&fromPunycode;
        label=fromPunycode.getBuffer();
        labelStart=0;
        labelLength=fromPunycode.length();
    } else {
        wasPunycode=FALSE;
        labelString=&dest;
    }
    // Validity check
    if(labelLength==0) {
        info.labelErrors|=UIDNA_ERROR_EMPTY_LABEL;
        return replaceLabel(dest, destLabelStart, destLabelLength,
                            *labelString, labelLength, errorCode);
    }
    // labelLength>0
    if(labelLength>=4 && label[2]==0x2d && label[3]==0x2d) {
        // label starts with "??--"
        info.labelErrors|=UIDNA_ERROR_HYPHEN_3_4;
    }
    if(label[0]==0x2d) {
        // label starts with "-"
        info.labelErrors|=UIDNA_ERROR_LEADING_HYPHEN;
    }
    if(label[labelLength-1]==0x2d) {
        // label ends with "-"
        info.labelErrors|=UIDNA_ERROR_TRAILING_HYPHEN;
    }
    // If the label was not a Punycode label, then it was the result of
    // mapping, normalization and label segmentation.
    // If the label was in Punycode, then we mapped it again above
    // and checked its validity.
    // Now we handle the STD3 restriction to LDH characters (if set)
    // and we look for U+FFFD which indicates disallowed characters
    // in a non-Punycode label or U+FFFD itself in a Punycode label.
    // We also check for dots which can come from the input to a single-label function.
    // Ok to cast away const because we own the UnicodeString.
    UChar *s=(UChar *)label;
    const UChar *limit=label+labelLength;
    UChar oredChars=0;
    // If we enforce STD3 rules, then ASCII characters other than LDH and dot are disallowed.
    UBool disallowNonLDHDot=(options&UIDNA_USE_STD3_RULES)!=0;
    do {
        UChar c=*s;
        if(c<=0x7f) {
            if(c==0x2e) {
                info.labelErrors|=UIDNA_ERROR_LABEL_HAS_DOT;
                *s=0xfffd;
            } else if(disallowNonLDHDot && asciiData[c]<0) {
                info.labelErrors|=UIDNA_ERROR_DISALLOWED;
                *s=0xfffd;
            }
        } else {
            oredChars|=c;
            if(disallowNonLDHDot && isNonASCIIDisallowedSTD3Valid(c)) {
                info.labelErrors|=UIDNA_ERROR_DISALLOWED;
                *s=0xfffd;
            } else if(c==0xfffd) {
                info.labelErrors|=UIDNA_ERROR_DISALLOWED;
            }
        }
        ++s;
    } while(s<limit);
    // Check for a leading combining mark after other validity checks
    // so that we don't report UIDNA_ERROR_DISALLOWED for the U+FFFD from here.
    UChar32 c;
    int32_t cpLength=0;
    // "Unsafe" is ok because unpaired surrogates were mapped to U+FFFD.
    U16_NEXT_UNSAFE(label, cpLength, c);
    if((U_GET_GC_MASK(c)&U_GC_M_MASK)!=0) {
        info.labelErrors|=UIDNA_ERROR_LEADING_COMBINING_MARK;
        labelString->replace(labelStart, cpLength, (UChar)0xfffd);
        label=labelString->getBuffer()+labelStart;
        labelLength+=1-cpLength;
        if(labelString==&dest) {
            destLabelLength=labelLength;
        }
    }
    if((info.labelErrors&severeErrors)==0) {
        // Do contextual checks only if we do not have U+FFFD from a severe error
        // because U+FFFD can make these checks fail.
        if((options&UIDNA_CHECK_BIDI)!=0 && (!info.isBiDi || info.isOkBiDi)) {
            checkLabelBiDi(label, labelLength, info);
        }
        if( (options&UIDNA_CHECK_CONTEXTJ)!=0 && (oredChars&0x200c)==0x200c &&
            !isLabelOkContextJ(label, labelLength)
        ) {
            info.labelErrors|=UIDNA_ERROR_CONTEXTJ;
        }
        if((options&UIDNA_CHECK_CONTEXTO)!=0 && oredChars>=0xb7) {
            checkLabelContextO(label, labelLength, info);
        }
        if(toASCII) {
            if(wasPunycode) {
                // Leave a Punycode label unchanged if it has no severe errors.
                if(destLabelLength>63) {
                    info.labelErrors|=UIDNA_ERROR_LABEL_TOO_LONG;
                }
                return destLabelLength;
            } else if(oredChars>=0x80) {
                // Contains non-ASCII characters.
                UnicodeString punycode;
                UChar *buffer=punycode.getBuffer(63);  // 63==maximum DNS label length
                if(buffer==NULL) {
                    errorCode=U_MEMORY_ALLOCATION_ERROR;
                    return destLabelLength;
                }
                buffer[0]=0x78;  // Write "xn--".
                buffer[1]=0x6e;
                buffer[2]=0x2d;
                buffer[3]=0x2d;
                int32_t punycodeLength=u_strToPunycode(label, labelLength,
                                                      buffer+4, punycode.getCapacity()-4,
                                                      NULL, &errorCode);
                if(errorCode==U_BUFFER_OVERFLOW_ERROR) {
                    errorCode=U_ZERO_ERROR;
                    punycode.releaseBuffer(4);
                    buffer=punycode.getBuffer(4+punycodeLength);
                    if(buffer==NULL) {
                        errorCode=U_MEMORY_ALLOCATION_ERROR;
                        return destLabelLength;
                    }
                    punycodeLength=u_strToPunycode(label, labelLength,
                                                  buffer+4, punycode.getCapacity()-4,
                                                  NULL, &errorCode);
                }
                punycodeLength+=4;
                punycode.releaseBuffer(punycodeLength);
                if(U_FAILURE(errorCode)) {
                    return destLabelLength;
                }
                if(punycodeLength>63) {
                    info.labelErrors|=UIDNA_ERROR_LABEL_TOO_LONG;
                }
                return replaceLabel(dest, destLabelStart, destLabelLength,
                                    punycode, punycodeLength, errorCode);
            } else {
                // all-ASCII label
                if(labelLength>63) {
                    info.labelErrors|=UIDNA_ERROR_LABEL_TOO_LONG;
                }
            }
        }
    } else {
        // If a Punycode label has severe errors,
        // then leave it but make sure it does not look valid.
        if(wasPunycode) {
            info.labelErrors|=UIDNA_ERROR_INVALID_ACE_LABEL;
            return markBadACELabel(dest, destLabelStart, destLabelLength, toASCII, info, errorCode);
        }
    }
    return replaceLabel(dest, destLabelStart, destLabelLength,
                        *labelString, labelLength, errorCode);
}

// Make sure an ACE label does not look valid.
// Append U+FFFD if the label has only LDH characters.
// If UIDNA_USE_STD3_RULES, also replace disallowed ASCII characters with U+FFFD.
int32_t
UTS46::markBadACELabel(UnicodeString &dest,
                       int32_t labelStart, int32_t labelLength,
                       UBool toASCII, IDNAInfo &info, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return 0;
    }
    UBool disallowNonLDHDot=(options&UIDNA_USE_STD3_RULES)!=0;
    UBool isASCII=TRUE;
    UBool onlyLDH=TRUE;
    const UChar *label=dest.getBuffer()+labelStart;
    // Ok to cast away const because we own the UnicodeString.
    UChar *s=(UChar *)label+4;  // After the initial "xn--".
    const UChar *limit=label+labelLength;
    do {
        UChar c=*s;
        if(c<=0x7f) {
            if(c==0x2e) {
                info.labelErrors|=UIDNA_ERROR_LABEL_HAS_DOT;
                *s=0xfffd;
                isASCII=onlyLDH=FALSE;
            } else if(asciiData[c]<0) {
                onlyLDH=FALSE;
                if(disallowNonLDHDot) {
                    *s=0xfffd;
                    isASCII=FALSE;
                }
            }
        } else {
            isASCII=onlyLDH=FALSE;
        }
    } while(++s<limit);
    if(onlyLDH) {
        dest.insert(labelStart+labelLength, (UChar)0xfffd);
        if(dest.isBogus()) {
            errorCode=U_MEMORY_ALLOCATION_ERROR;
            return 0;
        }
        ++labelLength;
    } else {
        if(toASCII && isASCII && labelLength>63) {
            info.labelErrors|=UIDNA_ERROR_LABEL_TOO_LONG;
        }
    }
    return labelLength;
}

const uint32_t L_MASK=U_MASK(U_LEFT_TO_RIGHT);
const uint32_t R_AL_MASK=U_MASK(U_RIGHT_TO_LEFT)|U_MASK(U_RIGHT_TO_LEFT_ARABIC);
const uint32_t L_R_AL_MASK=L_MASK|R_AL_MASK;

const uint32_t R_AL_AN_MASK=R_AL_MASK|U_MASK(U_ARABIC_NUMBER);

const uint32_t EN_AN_MASK=U_MASK(U_EUROPEAN_NUMBER)|U_MASK(U_ARABIC_NUMBER);
const uint32_t R_AL_EN_AN_MASK=R_AL_MASK|EN_AN_MASK;
const uint32_t L_EN_MASK=L_MASK|U_MASK(U_EUROPEAN_NUMBER);

const uint32_t ES_CS_ET_ON_BN_NSM_MASK=
    U_MASK(U_EUROPEAN_NUMBER_SEPARATOR)|
    U_MASK(U_COMMON_NUMBER_SEPARATOR)|
    U_MASK(U_EUROPEAN_NUMBER_TERMINATOR)|
    U_MASK(U_OTHER_NEUTRAL)|
    U_MASK(U_BOUNDARY_NEUTRAL)|
    U_MASK(U_DIR_NON_SPACING_MARK);
const uint32_t L_EN_ES_CS_ET_ON_BN_NSM_MASK=L_EN_MASK|ES_CS_ET_ON_BN_NSM_MASK;
const uint32_t R_AL_AN_EN_ES_CS_ET_ON_BN_NSM_MASK=R_AL_MASK|EN_AN_MASK|ES_CS_ET_ON_BN_NSM_MASK;

// We scan the whole label and check both for whether it contains RTL characters
// and whether it passes the BiDi Rule.
// In a BiDi domain name, all labels must pass the BiDi Rule, but we might find
// that a domain name is a BiDi domain name (has an RTL label) only after
// processing several earlier labels.
void
UTS46::checkLabelBiDi(const UChar *label, int32_t labelLength, IDNAInfo &info) const {
    // IDNA2008 BiDi rule
    // Get the directionality of the first character.
    UChar32 c;
    int32_t i=0;
    U16_NEXT_UNSAFE(label, i, c);
    uint32_t firstMask=U_MASK(u_charDirection(c));
    // 1. The first character must be a character with BIDI property L, R
    // or AL.  If it has the R or AL property, it is an RTL label; if it
    // has the L property, it is an LTR label.
    if((firstMask&~L_R_AL_MASK)!=0) {
        info.isOkBiDi=FALSE;
    }
    // Get the directionality of the last non-NSM character.
    uint32_t lastMask;
    for(;;) {
        if(i>=labelLength) {
            lastMask=firstMask;
            break;
        }
        U16_PREV_UNSAFE(label, labelLength, c);
        UCharDirection dir=u_charDirection(c);
        if(dir!=U_DIR_NON_SPACING_MARK) {
            lastMask=U_MASK(dir);
            break;
        }
    }
    // 3. In an RTL label, the end of the label must be a character with
    // BIDI property R, AL, EN or AN, followed by zero or more
    // characters with BIDI property NSM.
    // 6. In an LTR label, the end of the label must be a character with
    // BIDI property L or EN, followed by zero or more characters with
    // BIDI property NSM.
    if( (firstMask&L_MASK)!=0 ?
            (lastMask&~L_EN_MASK)!=0 :
            (lastMask&~R_AL_EN_AN_MASK)!=0
    ) {
        info.isOkBiDi=FALSE;
    }
    // Add the directionalities of the intervening characters.
    uint32_t mask=firstMask|lastMask;
    while(i<labelLength) {
        U16_NEXT_UNSAFE(label, i, c);
        mask|=U_MASK(u_charDirection(c));
    }
    if(firstMask&L_MASK) {
        // 5. In an LTR label, only characters with the BIDI properties L, EN,
        // ES, CS, ET, ON, BN and NSM are allowed.
        if((mask&~L_EN_ES_CS_ET_ON_BN_NSM_MASK)!=0) {
            info.isOkBiDi=FALSE;
        }
    } else {
        // 2. In an RTL label, only characters with the BIDI properties R, AL,
        // AN, EN, ES, CS, ET, ON, BN and NSM are allowed.
        if((mask&~R_AL_AN_EN_ES_CS_ET_ON_BN_NSM_MASK)!=0) {
            info.isOkBiDi=FALSE;
        }
        // 4. In an RTL label, if an EN is present, no AN may be present, and
        // vice versa.
        if((mask&EN_AN_MASK)==EN_AN_MASK) {
            info.isOkBiDi=FALSE;
        }
    }
    // An RTL label is a label that contains at least one character of type
    // R, AL or AN. [...]
    // A "BIDI domain name" is a domain name that contains at least one RTL
    // label. [...]
    // The following rule, consisting of six conditions, applies to labels
    // in BIDI domain names.
    if((mask&R_AL_AN_MASK)!=0) {
        info.isBiDi=TRUE;
    }
}

// Special code for the ASCII prefix of a BiDi domain name.
// The ASCII prefix is all-LTR.

// IDNA2008 BiDi rule, parts relevant to ASCII labels:
// 1. The first character must be a character with BIDI property L [...]
// 5. In an LTR label, only characters with the BIDI properties L, EN,
// ES, CS, ET, ON, BN and NSM are allowed.
// 6. In an LTR label, the end of the label must be a character with
// BIDI property L or EN [...]

// UTF-16 version, called for mapped ASCII prefix.
// Cannot contain uppercase A-Z.
// s[length-1] must be the trailing dot.
static UBool
isASCIIOkBiDi(const UChar *s, int32_t length) {
    int32_t labelStart=0;
    for(int32_t i=0; i<length; ++i) {
        UChar c=s[i];
        if(c==0x2e) {  // dot
            if(i>labelStart) {
                c=s[i-1];
                if(!(0x61<=c && c<=0x7a) && !(0x30<=c && c<=0x39)) {
                    // Last character in the label is not an L or EN.
                    return FALSE;
                }
            }
            labelStart=i+1;
        } else if(i==labelStart) {
            if(!(0x61<=c && c<=0x7a)) {
                // First character in the label is not an L.
                return FALSE;
            }
        } else {
            if(c<=0x20 && (c>=0x1c || (9<=c && c<=0xd))) {
                // Intermediate character in the label is a B, S or WS.
                return FALSE;
            }
        }
    }
    return TRUE;
}

// UTF-8 version, called for source ASCII prefix.
// Can contain uppercase A-Z.
// s[length-1] must be the trailing dot.
static UBool
isASCIIOkBiDi(const char *s, int32_t length) {
    int32_t labelStart=0;
    for(int32_t i=0; i<length; ++i) {
        char c=s[i];
        if(c==0x2e) {  // dot
            if(i>labelStart) {
                c=s[i-1];
                if(!(0x61<=c && c<=0x7a) && !(0x41<=c && c<=0x5a) && !(0x30<=c && c<=0x39)) {
                    // Last character in the label is not an L or EN.
                    return FALSE;
                }
            }
            labelStart=i+1;
        } else if(i==labelStart) {
            if(!(0x61<=c && c<=0x7a) && !(0x41<=c && c<=0x5a)) {
                // First character in the label is not an L.
                return FALSE;
            }
        } else {
            if(c<=0x20 && (c>=0x1c || (9<=c && c<=0xd))) {
                // Intermediate character in the label is a B, S or WS.
                return FALSE;
            }
        }
    }
    return TRUE;
}

UBool
UTS46::isLabelOkContextJ(const UChar *label, int32_t labelLength) const {
    // [IDNA2008-Tables]
    // 200C..200D  ; CONTEXTJ    # ZERO WIDTH NON-JOINER..ZERO WIDTH JOINER
    for(int32_t i=0; i<labelLength; ++i) {
        if(label[i]==0x200c) {
            // Appendix A.1. ZERO WIDTH NON-JOINER
            // Rule Set:
            //  False;
            //  If Canonical_Combining_Class(Before(cp)) .eq.  Virama Then True;
            //  If RegExpMatch((Joining_Type:{L,D})(Joining_Type:T)*\u200C
            //     (Joining_Type:T)*(Joining_Type:{R,D})) Then True;
            if(i==0) {
                return FALSE;
            }
            UChar32 c;
            int32_t j=i;
            U16_PREV_UNSAFE(label, j, c);
            if(uts46Norm2.getCombiningClass(c)==9) {
                continue;
            }
            // check precontext (Joining_Type:{L,D})(Joining_Type:T)*
            for(;;) {
                UJoiningType type=ubidi_getJoiningType(c);
                if(type==U_JT_TRANSPARENT) {
                    if(j==0) {
                        return FALSE;
                    }
                    U16_PREV_UNSAFE(label, j, c);
                } else if(type==U_JT_LEFT_JOINING || type==U_JT_DUAL_JOINING) {
                    break;  // precontext fulfilled
                } else {
                    return FALSE;
                }
            }
            // check postcontext (Joining_Type:T)*(Joining_Type:{R,D})
            for(j=i+1;;) {
                if(j==labelLength) {
                    return FALSE;
                }
                U16_NEXT_UNSAFE(label, j, c);
                UJoiningType type=ubidi_getJoiningType(c);
                if(type==U_JT_TRANSPARENT) {
                    // just skip this character
                } else if(type==U_JT_RIGHT_JOINING || type==U_JT_DUAL_JOINING) {
                    break;  // postcontext fulfilled
                } else {
                    return FALSE;
                }
            }
        } else if(label[i]==0x200d) {
            // Appendix A.2. ZERO WIDTH JOINER (U+200D)
            // Rule Set:
            //  False;
            //  If Canonical_Combining_Class(Before(cp)) .eq.  Virama Then True;
            if(i==0) {
                return FALSE;
            }
            UChar32 c;
            int32_t j=i;
            U16_PREV_UNSAFE(label, j, c);
            if(uts46Norm2.getCombiningClass(c)!=9) {
                return FALSE;
            }
        }
    }
    return TRUE;
}

void
UTS46::checkLabelContextO(const UChar *label, int32_t labelLength, IDNAInfo &info) const {
    int32_t labelEnd=labelLength-1;  // inclusive
    int32_t arabicDigits=0;  // -1 for 066x, +1 for 06Fx
    for(int32_t i=0; i<=labelEnd; ++i) {
        UChar32 c=label[i];
        if(c<0xb7) {
            // ASCII fastpath
        } else if(c<=0x6f9) {
            if(c==0xb7) {
                // Appendix A.3. MIDDLE DOT (U+00B7)
                // Rule Set:
                //  False;
                //  If Before(cp) .eq.  U+006C And
                //     After(cp) .eq.  U+006C Then True;
                if(!(0<i && label[i-1]==0x6c &&
                     i<labelEnd && label[i+1]==0x6c)) {
                    info.labelErrors|=UIDNA_ERROR_CONTEXTO_PUNCTUATION;
                }
            } else if(c==0x375) {
                // Appendix A.4. GREEK LOWER NUMERAL SIGN (KERAIA) (U+0375)
                // Rule Set:
                //  False;
                //  If Script(After(cp)) .eq.  Greek Then True;
                UScriptCode script=USCRIPT_INVALID_CODE;
                if(i<labelEnd) {
                    UErrorCode errorCode=U_ZERO_ERROR;
                    int32_t j=i+1;
                    U16_NEXT(label, j, labelLength, c);
                    script=uscript_getScript(c, &errorCode);
                }
                if(script!=USCRIPT_GREEK) {
                    info.labelErrors|=UIDNA_ERROR_CONTEXTO_PUNCTUATION;
                }
            } else if(c==0x5f3 || c==0x5f4) {
                // Appendix A.5. HEBREW PUNCTUATION GERESH (U+05F3)
                // Rule Set:
                //  False;
                //  If Script(Before(cp)) .eq.  Hebrew Then True;
                //
                // Appendix A.6. HEBREW PUNCTUATION GERSHAYIM (U+05F4)
                // Rule Set:
                //  False;
                //  If Script(Before(cp)) .eq.  Hebrew Then True;
                UScriptCode script=USCRIPT_INVALID_CODE;
                if(0<i) {
                    UErrorCode errorCode=U_ZERO_ERROR;
                    int32_t j=i;
                    U16_PREV(label, 0, j, c);
                    script=uscript_getScript(c, &errorCode);
                }
                if(script!=USCRIPT_HEBREW) {
                    info.labelErrors|=UIDNA_ERROR_CONTEXTO_PUNCTUATION;
                }
            } else if(0x660<=c /* && c<=0x6f9 */) {
                // Appendix A.8. ARABIC-INDIC DIGITS (0660..0669)
                // Rule Set:
                //  True;
                //  For All Characters:
                //    If cp .in. 06F0..06F9 Then False;
                //  End For;
                //
                // Appendix A.9. EXTENDED ARABIC-INDIC DIGITS (06F0..06F9)
                // Rule Set:
                //  True;
                //  For All Characters:
                //    If cp .in. 0660..0669 Then False;
                //  End For;
                if(c<=0x669) {
                    if(arabicDigits>0) {
                        info.labelErrors|=UIDNA_ERROR_CONTEXTO_DIGITS;
                    }
                    arabicDigits=-1;
                } else if(0x6f0<=c) {
                    if(arabicDigits<0) {
                        info.labelErrors|=UIDNA_ERROR_CONTEXTO_DIGITS;
                    }
                    arabicDigits=1;
                }
            }
        } else if(c==0x30fb) {
            // Appendix A.7. KATAKANA MIDDLE DOT (U+30FB)
            // Rule Set:
            //  False;
            //  For All Characters:
            //    If Script(cp) .in. {Hiragana, Katakana, Han} Then True;
            //  End For;
            UErrorCode errorCode=U_ZERO_ERROR;
            for(int j=0;;) {
                if(j>labelEnd) {
                    info.labelErrors|=UIDNA_ERROR_CONTEXTO_PUNCTUATION;
                    break;
                }
                U16_NEXT(label, j, labelLength, c);
                UScriptCode script=uscript_getScript(c, &errorCode);
                if(script==USCRIPT_HIRAGANA || script==USCRIPT_KATAKANA || script==USCRIPT_HAN) {
                    break;
                }
            }
        }
    }
}

U_NAMESPACE_END

// C API ------------------------------------------------------------------- ***

U_NAMESPACE_USE

U_CAPI UIDNA * U_EXPORT2
uidna_openUTS46(uint32_t options, UErrorCode *pErrorCode) {
    return reinterpret_cast<UIDNA *>(IDNA::createUTS46Instance(options, *pErrorCode));
}

U_CAPI void U_EXPORT2
uidna_close(UIDNA *idna) {
    delete reinterpret_cast<IDNA *>(idna);
}

static UBool
checkArgs(const void *label, int32_t length,
          void *dest, int32_t capacity,
          UIDNAInfo *pInfo, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return FALSE;
    }
    // sizeof(UIDNAInfo)=16 in the first API version.
    if(pInfo==NULL || pInfo->size<16) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
    if( (label==NULL ? length!=0 : length<-1) ||
        (dest==NULL ? capacity!=0 : capacity<0) ||
        (dest==label && label!=NULL)
    ) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
    // Set all *pInfo bytes to 0 except for the size field itself.
    uprv_memset(&pInfo->size+1, 0, pInfo->size-sizeof(pInfo->size));
    return TRUE;
}

static void
idnaInfoToStruct(IDNAInfo &info, UIDNAInfo *pInfo) {
    pInfo->isTransitionalDifferent=info.isTransitionalDifferent();
    pInfo->errors=info.getErrors();
}

U_CAPI int32_t U_EXPORT2
uidna_labelToASCII(const UIDNA *idna,
                   const UChar *label, int32_t length,
                   UChar *dest, int32_t capacity,
                   UIDNAInfo *pInfo, UErrorCode *pErrorCode) {
    if(!checkArgs(label, length, dest, capacity, pInfo, pErrorCode)) {
        return 0;
    }
    UnicodeString src((UBool)(length<0), label, length);
    UnicodeString destString(dest, 0, capacity);
    IDNAInfo info;
    reinterpret_cast<const IDNA *>(idna)->labelToASCII(src, destString, info, *pErrorCode);
    idnaInfoToStruct(info, pInfo);
    return destString.extract(dest, capacity, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uidna_labelToUnicode(const UIDNA *idna,
                     const UChar *label, int32_t length,
                     UChar *dest, int32_t capacity,
                     UIDNAInfo *pInfo, UErrorCode *pErrorCode) {
    if(!checkArgs(label, length, dest, capacity, pInfo, pErrorCode)) {
        return 0;
    }
    UnicodeString src((UBool)(length<0), label, length);
    UnicodeString destString(dest, 0, capacity);
    IDNAInfo info;
    reinterpret_cast<const IDNA *>(idna)->labelToUnicode(src, destString, info, *pErrorCode);
    idnaInfoToStruct(info, pInfo);
    return destString.extract(dest, capacity, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uidna_nameToASCII(const UIDNA *idna,
                  const UChar *name, int32_t length,
                  UChar *dest, int32_t capacity,
                  UIDNAInfo *pInfo, UErrorCode *pErrorCode) {
    if(!checkArgs(name, length, dest, capacity, pInfo, pErrorCode)) {
        return 0;
    }
    UnicodeString src((UBool)(length<0), name, length);
    UnicodeString destString(dest, 0, capacity);
    IDNAInfo info;
    reinterpret_cast<const IDNA *>(idna)->nameToASCII(src, destString, info, *pErrorCode);
    idnaInfoToStruct(info, pInfo);
    return destString.extract(dest, capacity, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uidna_nameToUnicode(const UIDNA *idna,
                    const UChar *name, int32_t length,
                    UChar *dest, int32_t capacity,
                    UIDNAInfo *pInfo, UErrorCode *pErrorCode) {
    if(!checkArgs(name, length, dest, capacity, pInfo, pErrorCode)) {
        return 0;
    }
    UnicodeString src((UBool)(length<0), name, length);
    UnicodeString destString(dest, 0, capacity);
    IDNAInfo info;
    reinterpret_cast<const IDNA *>(idna)->nameToUnicode(src, destString, info, *pErrorCode);
    idnaInfoToStruct(info, pInfo);
    return destString.extract(dest, capacity, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uidna_labelToASCII_UTF8(const UIDNA *idna,
                        const char *label, int32_t length,
                        char *dest, int32_t capacity,
                        UIDNAInfo *pInfo, UErrorCode *pErrorCode) {
    if(!checkArgs(label, length, dest, capacity, pInfo, pErrorCode)) {
        return 0;
    }
    StringPiece src(label, length<0 ? static_cast<int32_t>(uprv_strlen(label)) : length);
    CheckedArrayByteSink sink(dest, capacity);
    IDNAInfo info;
    reinterpret_cast<const IDNA *>(idna)->labelToASCII_UTF8(src, sink, info, *pErrorCode);
    idnaInfoToStruct(info, pInfo);
    return u_terminateChars(dest, capacity, sink.NumberOfBytesAppended(), pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uidna_labelToUnicodeUTF8(const UIDNA *idna,
                         const char *label, int32_t length,
                         char *dest, int32_t capacity,
                         UIDNAInfo *pInfo, UErrorCode *pErrorCode) {
    if(!checkArgs(label, length, dest, capacity, pInfo, pErrorCode)) {
        return 0;
    }
    StringPiece src(label, length<0 ? static_cast<int32_t>(uprv_strlen(label)) : length);
    CheckedArrayByteSink sink(dest, capacity);
    IDNAInfo info;
    reinterpret_cast<const IDNA *>(idna)->labelToUnicodeUTF8(src, sink, info, *pErrorCode);
    idnaInfoToStruct(info, pInfo);
    return u_terminateChars(dest, capacity, sink.NumberOfBytesAppended(), pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uidna_nameToASCII_UTF8(const UIDNA *idna,
                       const char *name, int32_t length,
                       char *dest, int32_t capacity,
                       UIDNAInfo *pInfo, UErrorCode *pErrorCode) {
    if(!checkArgs(name, length, dest, capacity, pInfo, pErrorCode)) {
        return 0;
    }
    StringPiece src(name, length<0 ? static_cast<int32_t>(uprv_strlen(name)) : length);
    CheckedArrayByteSink sink(dest, capacity);
    IDNAInfo info;
    reinterpret_cast<const IDNA *>(idna)->nameToASCII_UTF8(src, sink, info, *pErrorCode);
    idnaInfoToStruct(info, pInfo);
    return u_terminateChars(dest, capacity, sink.NumberOfBytesAppended(), pErrorCode);
}

U_CAPI int32_t U_EXPORT2
uidna_nameToUnicodeUTF8(const UIDNA *idna,
                        const char *name, int32_t length,
                        char *dest, int32_t capacity,
                        UIDNAInfo *pInfo, UErrorCode *pErrorCode) {
    if(!checkArgs(name, length, dest, capacity, pInfo, pErrorCode)) {
        return 0;
    }
    StringPiece src(name, length<0 ? static_cast<int32_t>(uprv_strlen(name)) : length);
    CheckedArrayByteSink sink(dest, capacity);
    IDNAInfo info;
    reinterpret_cast<const IDNA *>(idna)->nameToUnicodeUTF8(src, sink, info, *pErrorCode);
    idnaInfoToStruct(info, pInfo);
    return u_terminateChars(dest, capacity, sink.NumberOfBytesAppended(), pErrorCode);
}

#endif  // UCONFIG_NO_IDNA
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1997-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
*  FILE NAME : utypes.c (previously putil.c)
*
*   Date        Name        Description
*   10/07/2004  grhoten     split from putil.c
******************************************************************************
*/



/* u_errorName() ------------------------------------------------------------ */

static const char * const
_uErrorInfoName[U_ERROR_WARNING_LIMIT-U_ERROR_WARNING_START]={
    "U_USING_FALLBACK_WARNING",
    "U_USING_DEFAULT_WARNING",
    "U_SAFECLONE_ALLOCATED_WARNING",
    "U_STATE_OLD_WARNING",
    "U_STRING_NOT_TERMINATED_WARNING",
    "U_SORT_KEY_TOO_SHORT_WARNING",
    "U_AMBIGUOUS_ALIAS_WARNING",
    "U_DIFFERENT_UCA_VERSION",
    "U_PLUGIN_CHANGED_LEVEL_WARNING",
};

static const char * const
_uTransErrorName[U_PARSE_ERROR_LIMIT - U_PARSE_ERROR_START]={
    "U_BAD_VARIABLE_DEFINITION",
    "U_MALFORMED_RULE",
    "U_MALFORMED_SET",
    "U_MALFORMED_SYMBOL_REFERENCE",
    "U_MALFORMED_UNICODE_ESCAPE",
    "U_MALFORMED_VARIABLE_DEFINITION",
    "U_MALFORMED_VARIABLE_REFERENCE",
    "U_MISMATCHED_SEGMENT_DELIMITERS",
    "U_MISPLACED_ANCHOR_START",
    "U_MISPLACED_CURSOR_OFFSET",
    "U_MISPLACED_QUANTIFIER",
    "U_MISSING_OPERATOR",
    "U_MISSING_SEGMENT_CLOSE",
    "U_MULTIPLE_ANTE_CONTEXTS",
    "U_MULTIPLE_CURSORS",
    "U_MULTIPLE_POST_CONTEXTS",
    "U_TRAILING_BACKSLASH",
    "U_UNDEFINED_SEGMENT_REFERENCE",
    "U_UNDEFINED_VARIABLE",
    "U_UNQUOTED_SPECIAL",
    "U_UNTERMINATED_QUOTE",
    "U_RULE_MASK_ERROR",
    "U_MISPLACED_COMPOUND_FILTER",
    "U_MULTIPLE_COMPOUND_FILTERS",
    "U_INVALID_RBT_SYNTAX",
    "U_INVALID_PROPERTY_PATTERN",
    "U_MALFORMED_PRAGMA",
    "U_UNCLOSED_SEGMENT",
    "U_ILLEGAL_CHAR_IN_SEGMENT",
    "U_VARIABLE_RANGE_EXHAUSTED",
    "U_VARIABLE_RANGE_OVERLAP",
    "U_ILLEGAL_CHARACTER",
    "U_INTERNAL_TRANSLITERATOR_ERROR",
    "U_INVALID_ID",
    "U_INVALID_FUNCTION"
};

static const char * const
_uErrorName[U_STANDARD_ERROR_LIMIT]={
    "U_ZERO_ERROR",

    "U_ILLEGAL_ARGUMENT_ERROR",
    "U_MISSING_RESOURCE_ERROR",
    "U_INVALID_FORMAT_ERROR",
    "U_FILE_ACCESS_ERROR",
    "U_INTERNAL_PROGRAM_ERROR",
    "U_MESSAGE_PARSE_ERROR",
    "U_MEMORY_ALLOCATION_ERROR",
    "U_INDEX_OUTOFBOUNDS_ERROR",
    "U_PARSE_ERROR",
    "U_INVALID_CHAR_FOUND",
    "U_TRUNCATED_CHAR_FOUND",
    "U_ILLEGAL_CHAR_FOUND",
    "U_INVALID_TABLE_FORMAT",
    "U_INVALID_TABLE_FILE",
    "U_BUFFER_OVERFLOW_ERROR",
    "U_UNSUPPORTED_ERROR",
    "U_RESOURCE_TYPE_MISMATCH",
    "U_ILLEGAL_ESCAPE_SEQUENCE",
    "U_UNSUPPORTED_ESCAPE_SEQUENCE",
    "U_NO_SPACE_AVAILABLE",
    "U_CE_NOT_FOUND_ERROR",
    "U_PRIMARY_TOO_LONG_ERROR",
    "U_STATE_TOO_OLD_ERROR",
    "U_TOO_MANY_ALIASES_ERROR",
    "U_ENUM_OUT_OF_SYNC_ERROR",
    "U_INVARIANT_CONVERSION_ERROR",
    "U_INVALID_STATE_ERROR",
    "U_COLLATOR_VERSION_MISMATCH",
    "U_USELESS_COLLATOR_ERROR",
    "U_NO_WRITE_PERMISSION"
};
static const char * const
_uFmtErrorName[U_FMT_PARSE_ERROR_LIMIT - U_FMT_PARSE_ERROR_START] = {
    "U_UNEXPECTED_TOKEN",
    "U_MULTIPLE_DECIMAL_SEPARATORS",
    "U_MULTIPLE_EXPONENTIAL_SYMBOLS",
    "U_MALFORMED_EXPONENTIAL_PATTERN",
    "U_MULTIPLE_PERCENT_SYMBOLS",
    "U_MULTIPLE_PERMILL_SYMBOLS",
    "U_MULTIPLE_PAD_SPECIFIERS",
    "U_PATTERN_SYNTAX_ERROR",
    "U_ILLEGAL_PAD_POSITION",
    "U_UNMATCHED_BRACES",
    "U_UNSUPPORTED_PROPERTY",
    "U_UNSUPPORTED_ATTRIBUTE",
    "U_ARGUMENT_TYPE_MISMATCH",
    "U_DUPLICATE_KEYWORD",
    "U_UNDEFINED_KEYWORD",
    "U_DEFAULT_KEYWORD_MISSING",
    "U_DECIMAL_NUMBER_SYNTAX_ERROR",
    "U_FORMAT_INEXACT_ERROR",
    "U_NUMBER_ARG_OUTOFBOUNDS_ERROR",
    "U_NUMBER_SKELETON_SYNTAX_ERROR",
};

static const char * const
_uBrkErrorName[U_BRK_ERROR_LIMIT - U_BRK_ERROR_START] = {
    "U_BRK_INTERNAL_ERROR",
    "U_BRK_HEX_DIGITS_EXPECTED",
    "U_BRK_SEMICOLON_EXPECTED",
    "U_BRK_RULE_SYNTAX",
    "U_BRK_UNCLOSED_SET",
    "U_BRK_ASSIGN_ERROR",
    "U_BRK_VARIABLE_REDFINITION",
    "U_BRK_MISMATCHED_PAREN",
    "U_BRK_NEW_LINE_IN_QUOTED_STRING",
    "U_BRK_UNDEFINED_VARIABLE",
    "U_BRK_INIT_ERROR",
    "U_BRK_RULE_EMPTY_SET",
    "U_BRK_UNRECOGNIZED_OPTION",
    "U_BRK_MALFORMED_RULE_TAG"
};

static const char * const
_uRegexErrorName[U_REGEX_ERROR_LIMIT - U_REGEX_ERROR_START] = {
    "U_REGEX_INTERNAL_ERROR",
    "U_REGEX_RULE_SYNTAX",
    "U_REGEX_INVALID_STATE",
    "U_REGEX_BAD_ESCAPE_SEQUENCE",
    "U_REGEX_PROPERTY_SYNTAX",
    "U_REGEX_UNIMPLEMENTED",
    "U_REGEX_MISMATCHED_PAREN",
    "U_REGEX_NUMBER_TOO_BIG",
    "U_REGEX_BAD_INTERVAL",
    "U_REGEX_MAX_LT_MIN",
    "U_REGEX_INVALID_BACK_REF",
    "U_REGEX_INVALID_FLAG",
    "U_REGEX_LOOK_BEHIND_LIMIT",
    "U_REGEX_SET_CONTAINS_STRING",
    "U_REGEX_OCTAL_TOO_BIG",
    "U_REGEX_MISSING_CLOSE_BRACKET",
    "U_REGEX_INVALID_RANGE",
    "U_REGEX_STACK_OVERFLOW",
    "U_REGEX_TIME_OUT",
    "U_REGEX_STOPPED_BY_CALLER",
    "U_REGEX_PATTERN_TOO_BIG",
    "U_REGEX_INVALID_CAPTURE_GROUP_NAME"
};

static const char * const
_uIDNAErrorName[U_IDNA_ERROR_LIMIT - U_IDNA_ERROR_START] = {
     "U_STRINGPREP_PROHIBITED_ERROR",
     "U_STRINGPREP_UNASSIGNED_ERROR",
     "U_STRINGPREP_CHECK_BIDI_ERROR",
     "U_IDNA_STD3_ASCII_RULES_ERROR",
     "U_IDNA_ACE_PREFIX_ERROR",
     "U_IDNA_VERIFICATION_ERROR",
     "U_IDNA_LABEL_TOO_LONG_ERROR",
     "U_IDNA_ZERO_LENGTH_LABEL_ERROR",
     "U_IDNA_DOMAIN_NAME_TOO_LONG_ERROR"
};

static const char * const
_uPluginErrorName[U_PLUGIN_ERROR_LIMIT - U_PLUGIN_ERROR_START] = {
     "U_PLUGIN_TOO_HIGH",
     "U_PLUGIN_DIDNT_SET_LEVEL",
};

U_CAPI const char * U_EXPORT2
u_errorName(UErrorCode code) {
    if(U_ZERO_ERROR <= code && code < U_STANDARD_ERROR_LIMIT) {
        return _uErrorName[code];
    } else if(U_ERROR_WARNING_START <= code && code < U_ERROR_WARNING_LIMIT) {
        return _uErrorInfoName[code - U_ERROR_WARNING_START];
    } else if(U_PARSE_ERROR_START <= code && code < U_PARSE_ERROR_LIMIT){
        return _uTransErrorName[code - U_PARSE_ERROR_START];
    } else if(U_FMT_PARSE_ERROR_START <= code && code < U_FMT_PARSE_ERROR_LIMIT){
        return _uFmtErrorName[code - U_FMT_PARSE_ERROR_START];
    } else if (U_BRK_ERROR_START <= code  && code < U_BRK_ERROR_LIMIT){
        return _uBrkErrorName[code - U_BRK_ERROR_START];
    } else if (U_REGEX_ERROR_START <= code && code < U_REGEX_ERROR_LIMIT) {
        return _uRegexErrorName[code - U_REGEX_ERROR_START];
    } else if(U_IDNA_ERROR_START <= code && code < U_IDNA_ERROR_LIMIT) {
        return _uIDNAErrorName[code - U_IDNA_ERROR_START];
    } else if(U_PLUGIN_ERROR_START <= code && code < U_PLUGIN_ERROR_LIMIT) {
        return _uPluginErrorName[code - U_PLUGIN_ERROR_START];
    } else {
        return "[BOGUS UErrorCode]";
    }
}

/*
 * Hey, Emacs, please set the following:
 *
 * Local Variables:
 * indent-tabs-mode: nil
 * End:
 *
 */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 1999-2013, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*   Date        Name        Description
*   10/22/99    alan        Creation.
**********************************************************************
*/






U_NAMESPACE_BEGIN

#define UVECTOR_DEFAULT_CAPACITY 8

/*
 * Constants for hinting whether a key is an integer
 * or a pointer.  If a hint bit is zero, then the associated
 * token is assumed to be an integer. This is needed for iSeries
 */
#define HINT_KEY_POINTER   (1)
#define HINT_KEY_INTEGER   (0)

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(UVector)

UVector::UVector(UErrorCode &status) :
    count(0),
    capacity(0),
    elements(0),
    deleter(0),
    comparer(0)
{
    _init(UVECTOR_DEFAULT_CAPACITY, status);
}

UVector::UVector(int32_t initialCapacity, UErrorCode &status) :
    count(0),
    capacity(0),
    elements(0),
    deleter(0),
    comparer(0)
{
    _init(initialCapacity, status);
}

UVector::UVector(UObjectDeleter *d, UElementsAreEqual *c, UErrorCode &status) :
    count(0),
    capacity(0),
    elements(0),
    deleter(d),
    comparer(c)
{
    _init(UVECTOR_DEFAULT_CAPACITY, status);
}

UVector::UVector(UObjectDeleter *d, UElementsAreEqual *c, int32_t initialCapacity, UErrorCode &status) :
    count(0),
    capacity(0),
    elements(0),
    deleter(d),
    comparer(c)
{
    _init(initialCapacity, status);
}

void UVector::_init(int32_t initialCapacity, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return;
    }
    // Fix bogus initialCapacity values; avoid malloc(0) and integer overflow
    if ((initialCapacity < 1) || (initialCapacity > (int32_t)(INT32_MAX / sizeof(UElement)))) {
        initialCapacity = UVECTOR_DEFAULT_CAPACITY;
    }
    elements = (UElement *)uprv_malloc(sizeof(UElement)*initialCapacity);
    if (elements == 0) {
        status = U_MEMORY_ALLOCATION_ERROR;
    } else {
        capacity = initialCapacity;
    }
}

UVector::~UVector() {
    removeAllElements();
    uprv_free(elements);
    elements = 0;
}

/**
 * Assign this object to another (make this a copy of 'other').
 * Use the 'assign' function to assign each element.
 */
void UVector::assign(const UVector& other, UElementAssigner *assign, UErrorCode &ec) {
    if (ensureCapacity(other.count, ec)) {
        setSize(other.count, ec);
        if (U_SUCCESS(ec)) {
            for (int32_t i=0; i<other.count; ++i) {
                if (elements[i].pointer != 0 && deleter != 0) {
                    (*deleter)(elements[i].pointer);
                }
                (*assign)(&elements[i], &other.elements[i]);
            }
        }
    }
}

// This only does something sensible if this object has a non-null comparer
UBool UVector::operator==(const UVector& other) {
    int32_t i;
    if (count != other.count) return FALSE;
    if (comparer != NULL) {
        // Compare using this object's comparer
        for (i=0; i<count; ++i) {
            if (!(*comparer)(elements[i], other.elements[i])) {
                return FALSE;
            }
        }
    }
    return TRUE;
}

void UVector::addElement(void* obj, UErrorCode &status) {
    if (ensureCapacity(count + 1, status)) {
        elements[count++].pointer = obj;
    }
}

void UVector::addElement(int32_t elem, UErrorCode &status) {
    if (ensureCapacity(count + 1, status)) {
        elements[count].pointer = NULL;     // Pointers may be bigger than ints.
        elements[count].integer = elem;
        count++;
    }
}

void UVector::setElementAt(void* obj, int32_t index) {
    if (0 <= index && index < count) {
        if (elements[index].pointer != 0 && deleter != 0) {
            (*deleter)(elements[index].pointer);
        }
        elements[index].pointer = obj;
    }
    /* else index out of range */
}

void UVector::setElementAt(int32_t elem, int32_t index) {
    if (0 <= index && index < count) {
        if (elements[index].pointer != 0 && deleter != 0) {
            // TODO:  this should be an error.  mixing up ints and pointers.
            (*deleter)(elements[index].pointer);
        }
        elements[index].pointer = NULL;
        elements[index].integer = elem;
    }
    /* else index out of range */
}

void UVector::insertElementAt(void* obj, int32_t index, UErrorCode &status) {
    // must have 0 <= index <= count
    if (0 <= index && index <= count && ensureCapacity(count + 1, status)) {
        for (int32_t i=count; i>index; --i) {
            elements[i] = elements[i-1];
        }
        elements[index].pointer = obj;
        ++count;
    }
    /* else index out of range */
}

void UVector::insertElementAt(int32_t elem, int32_t index, UErrorCode &status) {
    // must have 0 <= index <= count
    if (0 <= index && index <= count && ensureCapacity(count + 1, status)) {
        for (int32_t i=count; i>index; --i) {
            elements[i] = elements[i-1];
        }
        elements[index].pointer = NULL;
        elements[index].integer = elem;
        ++count;
    }
    /* else index out of range */
}

void* UVector::elementAt(int32_t index) const {
    return (0 <= index && index < count) ? elements[index].pointer : 0;
}

int32_t UVector::elementAti(int32_t index) const {
    return (0 <= index && index < count) ? elements[index].integer : 0;
}

UBool UVector::containsAll(const UVector& other) const {
    for (int32_t i=0; i<other.size(); ++i) {
        if (indexOf(other.elements[i]) < 0) {
            return FALSE;
        }
    }
    return TRUE;
}

UBool UVector::containsNone(const UVector& other) const {
    for (int32_t i=0; i<other.size(); ++i) {
        if (indexOf(other.elements[i]) >= 0) {
            return FALSE;
        }
    }
    return TRUE;
}

UBool UVector::removeAll(const UVector& other) {
    UBool changed = FALSE;
    for (int32_t i=0; i<other.size(); ++i) {
        int32_t j = indexOf(other.elements[i]);
        if (j >= 0) {
            removeElementAt(j);
            changed = TRUE;
        }
    }
    return changed;
}

UBool UVector::retainAll(const UVector& other) {
    UBool changed = FALSE;
    for (int32_t j=size()-1; j>=0; --j) {
        int32_t i = other.indexOf(elements[j]);
        if (i < 0) {
            removeElementAt(j);
            changed = TRUE;
        }
    }
    return changed;
}

void UVector::removeElementAt(int32_t index) {
    void* e = orphanElementAt(index);
    if (e != 0 && deleter != 0) {
        (*deleter)(e);
    }
}

UBool UVector::removeElement(void* obj) {
    int32_t i = indexOf(obj);
    if (i >= 0) {
        removeElementAt(i);
        return TRUE;
    }
    return FALSE;
}

void UVector::removeAllElements(void) {
    if (deleter != 0) {
        for (int32_t i=0; i<count; ++i) {
            if (elements[i].pointer != 0) {
                (*deleter)(elements[i].pointer);
            }
        }
    }
    count = 0;
}

UBool   UVector::equals(const UVector &other) const {
    int      i;

    if (this->count != other.count) {
        return FALSE;
    }
    if (comparer == 0) {
        for (i=0; i<count; i++) {
            if (elements[i].pointer != other.elements[i].pointer) {
                return FALSE;
            }
        }
    } else {
        UElement key;
        for (i=0; i<count; i++) {
            key.pointer = &other.elements[i];
            if (!(*comparer)(key, elements[i])) {
                return FALSE;
            }
        }
    }
    return TRUE;
}



int32_t UVector::indexOf(void* obj, int32_t startIndex) const {
    UElement key;
    key.pointer = obj;
    return indexOf(key, startIndex, HINT_KEY_POINTER);
}

int32_t UVector::indexOf(int32_t obj, int32_t startIndex) const {
    UElement key;
    key.integer = obj;
    return indexOf(key, startIndex, HINT_KEY_INTEGER);
}

// This only works if this object has a non-null comparer
int32_t UVector::indexOf(UElement key, int32_t startIndex, int8_t hint) const {
    int32_t i;
    if (comparer != 0) {
        for (i=startIndex; i<count; ++i) {
            if ((*comparer)(key, elements[i])) {
                return i;
            }
        }
    } else {
        for (i=startIndex; i<count; ++i) {
            /* Pointers are not always the same size as ints so to perform
             * a valid comparision we need to know whether we are being
             * provided an int or a pointer. */
            if (hint & HINT_KEY_POINTER) {
                if (key.pointer == elements[i].pointer) {
                    return i;
                }
            } else {
                if (key.integer == elements[i].integer) {
                    return i;
                }
            }
        }
    }
    return -1;
}

UBool UVector::ensureCapacity(int32_t minimumCapacity, UErrorCode &status) {
	if (minimumCapacity < 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
	}
    if (capacity < minimumCapacity) {
        if (capacity > (INT32_MAX - 1) / 2) {        	// integer overflow check
        	status = U_ILLEGAL_ARGUMENT_ERROR;
        	return FALSE;
        }
        int32_t newCap = capacity * 2;
        if (newCap < minimumCapacity) {
            newCap = minimumCapacity;
        }
        if (newCap > (int32_t)(INT32_MAX / sizeof(UElement))) {	// integer overflow check
        	// We keep the original memory contents on bad minimumCapacity.
        	status = U_ILLEGAL_ARGUMENT_ERROR;
        	return FALSE;
        }
        UElement* newElems = (UElement *)uprv_realloc(elements, sizeof(UElement)*newCap);
        if (newElems == NULL) {
            // We keep the original contents on the memory failure on realloc or bad minimumCapacity.
            status = U_MEMORY_ALLOCATION_ERROR;
            return FALSE;
        }
        elements = newElems;
        capacity = newCap;
    }
    return TRUE;
}

/**
 * Change the size of this vector as follows: If newSize is smaller,
 * then truncate the array, possibly deleting held elements for i >=
 * newSize.  If newSize is larger, grow the array, filling in new
 * slots with NULL.
 */
void UVector::setSize(int32_t newSize, UErrorCode &status) {
    int32_t i;
    if (newSize < 0) {
        return;
    }
    if (newSize > count) {
        if (!ensureCapacity(newSize, status)) {
            return;
        }
        UElement empty;
        empty.pointer = NULL;
        empty.integer = 0;
        for (i=count; i<newSize; ++i) {
            elements[i] = empty;
        }
    } else {
        /* Most efficient to count down */
        for (i=count-1; i>=newSize; --i) {
            removeElementAt(i);
        }
    }
    count = newSize;
}

/**
 * Fill in the given array with all elements of this vector.
 */
void** UVector::toArray(void** result) const {
    void** a = result;
    for (int i=0; i<count; ++i) {
        *a++ = elements[i].pointer;
    }
    return result;
}

UObjectDeleter *UVector::setDeleter(UObjectDeleter *d) {
    UObjectDeleter *old = deleter;
    deleter = d;
    return old;
}

UElementsAreEqual *UVector::setComparer(UElementsAreEqual *d) {
    UElementsAreEqual *old = comparer;
    comparer = d;
    return old;
}

/**
 * Removes the element at the given index from this vector and
 * transfer ownership of it to the caller.  After this call, the
 * caller owns the result and must delete it and the vector entry
 * at 'index' is removed, shifting all subsequent entries back by
 * one index and shortening the size of the vector by one.  If the
 * index is out of range or if there is no item at the given index
 * then 0 is returned and the vector is unchanged.
 */
void* UVector::orphanElementAt(int32_t index) {
    void* e = 0;
    if (0 <= index && index < count) {
        e = elements[index].pointer;
        for (int32_t i=index; i<count-1; ++i) {
            elements[i] = elements[i+1];
        }
        --count;
    }
    /* else index out of range */
    return e;
}

/**
 * Insert the given object into this vector at its sorted position
 * as defined by 'compare'.  The current elements are assumed to
 * be sorted already.
 */
void UVector::sortedInsert(void* obj, UElementComparator *compare, UErrorCode& ec) {
    UElement e;
    e.pointer = obj;
    sortedInsert(e, compare, ec);
}

/**
 * Insert the given integer into this vector at its sorted position
 * as defined by 'compare'.  The current elements are assumed to
 * be sorted already.
 */
void UVector::sortedInsert(int32_t obj, UElementComparator *compare, UErrorCode& ec) {
    UElement e;
    e.integer = obj;
    sortedInsert(e, compare, ec);
}

// ASSUME elements[] IS CURRENTLY SORTED
void UVector::sortedInsert(UElement e, UElementComparator *compare, UErrorCode& ec) {
    // Perform a binary search for the location to insert tok at.  Tok
    // will be inserted between two elements a and b such that a <=
    // tok && tok < b, where there is a 'virtual' elements[-1] always
    // less than tok and a 'virtual' elements[count] always greater
    // than tok.
    int32_t min = 0, max = count;
    while (min != max) {
        int32_t probe = (min + max) / 2;
        int8_t c = (*compare)(elements[probe], e);
        if (c > 0) {
            max = probe;
        } else {
            // assert(c <= 0);
            min = probe + 1;
        }
    }
    if (ensureCapacity(count + 1, ec)) {
        for (int32_t i=count; i>min; --i) {
            elements[i] = elements[i-1];
        }
        elements[min] = e;
        ++count;
    }
}

/**
  *  Array sort comparator function.
  *  Used from UVector::sort()
  *  Conforms to function signature required for uprv_sortArray().
  *  This function is essentially just a wrapper, to make a
  *  UVector style comparator function usable with uprv_sortArray().
  *
  *  The context pointer to this function is a pointer back
  *  (with some extra indirection) to the user supplied comparator.
  *
  */
static int32_t U_CALLCONV
sortComparator(const void *context, const void *left, const void *right) {
    UElementComparator *compare = *static_cast<UElementComparator * const *>(context);
    UElement e1 = *static_cast<const UElement *>(left);
    UElement e2 = *static_cast<const UElement *>(right);
    int32_t result = (*compare)(e1, e2);
    return result;
}


/**
  *  Array sort comparison function for use from UVector::sorti()
  *  Compares int32_t vector elements.
  */
static int32_t U_CALLCONV
sortiComparator(const void * /*context */, const void *left, const void *right) {
    const UElement *e1 = static_cast<const UElement *>(left);
    const UElement *e2 = static_cast<const UElement *>(right);
    int32_t result = e1->integer < e2->integer? -1 :
                     e1->integer == e2->integer? 0 : 1;
    return result;
}

/**
  * Sort the vector, assuming it constains ints.
  *     (A more general sort would take a comparison function, but it's
  *     not clear whether UVector's UElementComparator or
  *     UComparator from uprv_sortAray would be more appropriate.)
  */
void UVector::sorti(UErrorCode &ec) {
    if (U_SUCCESS(ec)) {
        uprv_sortArray(elements, count, sizeof(UElement),
                       sortiComparator, NULL,  FALSE, &ec);
    }
}


/**
 *  Sort with a user supplied comparator.
 *
 *    The comparator function handling is confusing because the function type
 *    for UVector  (as defined for sortedInsert()) is different from the signature
 *    required by uprv_sortArray().  This is handled by passing the
 *    the UVector sort function pointer via the context pointer to a
 *    sortArray() comparator function, which can then call back to
 *    the original user functtion.
 *
 *    An additional twist is that it's not safe to pass a pointer-to-function
 *    as  a (void *) data pointer, so instead we pass a (data) pointer to a
 *    pointer-to-function variable.
 */
void UVector::sort(UElementComparator *compare, UErrorCode &ec) {
    if (U_SUCCESS(ec)) {
        uprv_sortArray(elements, count, sizeof(UElement),
                       sortComparator, &compare, FALSE, &ec);
    }
}


/**
 *  Stable sort with a user supplied comparator of type UComparator.
 */
void UVector::sortWithUComparator(UComparator *compare, const void *context, UErrorCode &ec) {
    if (U_SUCCESS(ec)) {
        uprv_sortArray(elements, count, sizeof(UElement),
                       compare, context, TRUE, &ec);
    }
}

U_NAMESPACE_END

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 1999-2015, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*   Date        Name        Description
*   10/22/99    alan        Creation.
**********************************************************************
*/





U_NAMESPACE_BEGIN

#define UVECTOR32_DEFAULT_CAPACITY 8

/*
 * Constants for hinting whether a key is an integer
 * or a pointer.  If a hint bit is zero, then the associated
 * token is assumed to be an integer. This is needed for iSeries
 */

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(UVector32)

UVector32::UVector32(UErrorCode &status) :
    count(0),
    capacity(0),
    maxCapacity(0),
    elements(NULL)
{
    _init(UVECTOR32_DEFAULT_CAPACITY, status);
}

UVector32::UVector32(int32_t initialCapacity, UErrorCode &status) :
    count(0),
    capacity(0),
    maxCapacity(0),
    elements(0)
{
    _init(initialCapacity, status);
}



void UVector32::_init(int32_t initialCapacity, UErrorCode &status) {
    // Fix bogus initialCapacity values; avoid malloc(0)
    if (initialCapacity < 1) {
        initialCapacity = UVECTOR32_DEFAULT_CAPACITY;
    }
    if (maxCapacity>0 && maxCapacity<initialCapacity) {
        initialCapacity = maxCapacity;
    }
    if (initialCapacity > (int32_t)(INT32_MAX / sizeof(int32_t))) {
        initialCapacity = uprv_min(UVECTOR32_DEFAULT_CAPACITY, maxCapacity);
    }
    elements = (int32_t *)uprv_malloc(sizeof(int32_t)*initialCapacity);
    if (elements == 0) {
        status = U_MEMORY_ALLOCATION_ERROR;
    } else {
        capacity = initialCapacity;
    }
}

UVector32::~UVector32() {
    uprv_free(elements);
    elements = 0;
}

/**
 * Assign this object to another (make this a copy of 'other').
 */
void UVector32::assign(const UVector32& other, UErrorCode &ec) {
    if (ensureCapacity(other.count, ec)) {
        setSize(other.count);
        for (int32_t i=0; i<other.count; ++i) {
            elements[i] = other.elements[i];
        }
    }
}


UBool UVector32::operator==(const UVector32& other) {
    int32_t i;
    if (count != other.count) return FALSE;
    for (i=0; i<count; ++i) {
        if (elements[i] != other.elements[i]) {
            return FALSE;
        }
    }
    return TRUE;
}


void UVector32::setElementAt(int32_t elem, int32_t index) {
    if (0 <= index && index < count) {
        elements[index] = elem;
    }
    /* else index out of range */
}

void UVector32::insertElementAt(int32_t elem, int32_t index, UErrorCode &status) {
    // must have 0 <= index <= count
    if (0 <= index && index <= count && ensureCapacity(count + 1, status)) {
        for (int32_t i=count; i>index; --i) {
            elements[i] = elements[i-1];
        }
        elements[index] = elem;
        ++count;
    }
    /* else index out of range */
}

UBool UVector32::containsAll(const UVector32& other) const {
    for (int32_t i=0; i<other.size(); ++i) {
        if (indexOf(other.elements[i]) < 0) {
            return FALSE;
        }
    }
    return TRUE;
}

UBool UVector32::containsNone(const UVector32& other) const {
    for (int32_t i=0; i<other.size(); ++i) {
        if (indexOf(other.elements[i]) >= 0) {
            return FALSE;
        }
    }
    return TRUE;
}

UBool UVector32::removeAll(const UVector32& other) {
    UBool changed = FALSE;
    for (int32_t i=0; i<other.size(); ++i) {
        int32_t j = indexOf(other.elements[i]);
        if (j >= 0) {
            removeElementAt(j);
            changed = TRUE;
        }
    }
    return changed;
}

UBool UVector32::retainAll(const UVector32& other) {
    UBool changed = FALSE;
    for (int32_t j=size()-1; j>=0; --j) {
        int32_t i = other.indexOf(elements[j]);
        if (i < 0) {
            removeElementAt(j);
            changed = TRUE;
        }
    }
    return changed;
}

void UVector32::removeElementAt(int32_t index) {
    if (index >= 0) {
        for (int32_t i=index; i<count-1; ++i) {
            elements[i] = elements[i+1];
        }
        --count;
    }
}

void UVector32::removeAllElements(void) {
    count = 0;
}

UBool   UVector32::equals(const UVector32 &other) const {
    int      i;

    if (this->count != other.count) {
        return FALSE;
    }
    for (i=0; i<count; i++) {
        if (elements[i] != other.elements[i]) {
            return FALSE;
        }
    }
    return TRUE;
}




int32_t UVector32::indexOf(int32_t key, int32_t startIndex) const {
    int32_t i;
    for (i=startIndex; i<count; ++i) {
        if (key == elements[i]) {
            return i;
        }
    }
    return -1;
}


UBool UVector32::expandCapacity(int32_t minimumCapacity, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (minimumCapacity < 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
    if (capacity >= minimumCapacity) {
        return TRUE;
    }
    if (maxCapacity>0 && minimumCapacity>maxCapacity) {
        status = U_BUFFER_OVERFLOW_ERROR;
        return FALSE;
    }
    if (capacity > (INT32_MAX - 1) / 2) {  // integer overflow check
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
    int32_t newCap = capacity * 2;
    if (newCap < minimumCapacity) {
        newCap = minimumCapacity;
    }
    if (maxCapacity > 0 && newCap > maxCapacity) {
        newCap = maxCapacity;
    }
    if (newCap > (int32_t)(INT32_MAX / sizeof(int32_t))) {  // integer overflow check
        // We keep the original memory contents on bad minimumCapacity/maxCapacity.
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
    int32_t* newElems = (int32_t *)uprv_realloc(elements, sizeof(int32_t)*newCap);
    if (newElems == NULL) {
        // We keep the original contents on the memory failure on realloc.
        status = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    elements = newElems;
    capacity = newCap;
    return TRUE;
}

void UVector32::setMaxCapacity(int32_t limit) {
    U_ASSERT(limit >= 0);
    if (limit < 0) {
        limit = 0;
    }
    if (limit > (int32_t)(INT32_MAX / sizeof(int32_t))) {  // integer overflow check for realloc
        //  Something is very wrong, don't realloc, leave capacity and maxCapacity unchanged
        return;
    }
    maxCapacity = limit;
    if (capacity <= maxCapacity || maxCapacity == 0) {
        // Current capacity is within the new limit.
        return;
    }

    // New maximum capacity is smaller than the current size.
    // Realloc the storage to the new, smaller size.
    int32_t* newElems = (int32_t *)uprv_realloc(elements, sizeof(int32_t)*maxCapacity);
    if (newElems == NULL) {
        // Realloc to smaller failed.
        //   Just keep what we had.  No need to call it a failure.
        return;
    }
    elements = newElems;
    capacity = maxCapacity;
    if (count > capacity) {
        count = capacity;
    }
}

/**
 * Change the size of this vector as follows: If newSize is smaller,
 * then truncate the array, possibly deleting held elements for i >=
 * newSize.  If newSize is larger, grow the array, filling in new
 * slots with NULL.
 */
void UVector32::setSize(int32_t newSize) {
    int32_t i;
    if (newSize < 0) {
        return;
    }
    if (newSize > count) {
        UErrorCode ec = U_ZERO_ERROR;
        if (!ensureCapacity(newSize, ec)) {
            return;
        }
        for (i=count; i<newSize; ++i) {
            elements[i] = 0;
        }
    }
    count = newSize;
}




/**
 * Insert the given integer into this vector at its sorted position
 * as defined by 'compare'.  The current elements are assumed to
 * be sorted already.
 */
void UVector32::sortedInsert(int32_t tok, UErrorCode& ec) {
    // Perform a binary search for the location to insert tok at.  Tok
    // will be inserted between two elements a and b such that a <=
    // tok && tok < b, where there is a 'virtual' elements[-1] always
    // less than tok and a 'virtual' elements[count] always greater
    // than tok.
    int32_t min = 0, max = count;
    while (min != max) {
        int32_t probe = (min + max) / 2;
        //int8_t c = (*compare)(elements[probe], tok);
        //if (c > 0) {
        if (elements[probe] > tok) {
            max = probe;
        } else {
            // assert(c <= 0);
            min = probe + 1;
        }
    }
    if (ensureCapacity(count + 1, ec)) {
        for (int32_t i=count; i>min; --i) {
            elements[i] = elements[i-1];
        }
        elements[min] = tok;
        ++count;
    }
}





U_NAMESPACE_END

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 1999-2015, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*/

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 1999-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
**********************************************************************
*/

//
//  UVector64 is a class implementing a vector of 64 bit integers.
//            It is similar to UVector32, but holds int64_t values rather than int32_t.
//            Most of the code is unchanged from UVector.
//







U_NAMESPACE_BEGIN



/**
 * <p>Ultralightweight C++ implementation of an <tt>int64_t</tt> vector
 * that has a subset of methods from UVector32
 *
 * <p>This is a very simple implementation, written to satisfy an
 * immediate porting need.  As such, it is not completely fleshed out,
 * and it aims for simplicity and conformity.  Nonetheless, it serves
 * its purpose (porting code from java that uses java.util.Vector)
 * well, and it could be easily made into a more robust vector class.
 *
 * <p><b>Design notes</b>
 *
 * <p>There is index bounds checking, but little is done about it.  If
 * indices are out of bounds, either nothing happens, or zero is
 * returned.  We <em>do</em> avoid indexing off into the weeds.
 *
 * <p>There is detection of out of memory, but the handling is very
 * coarse-grained -- similar to UnicodeString's protocol, but even
 * coarser.  The class contains <em>one static flag</em> that is set
 * when any call to <tt>new</tt> returns zero.  This allows the caller
 * to use several vectors and make just one check at the end to see if
 * a memory failure occurred.  This is more efficient than making a
 * check after each call on each vector when doing many operations on
 * multiple vectors.  The single static flag works best when memory
 * failures are infrequent, and when recovery options are limited or
 * nonexistent.
 *
 * <p><b>To do</b>
 *
 * <p>Improve the handling of index out of bounds errors.
 *
 */
class U_COMMON_API UVector64 : public UObject {
private:
    int32_t   count;

    int32_t   capacity;
    
    int32_t   maxCapacity;   // Limit beyond which capacity is not permitted to grow.

    int64_t*  elements;

public:
    UVector64(UErrorCode &status);

    UVector64(int32_t initialCapacity, UErrorCode &status);

    virtual ~UVector64();

    /**
     * Assign this object to another (make this a copy of 'other').
     * Use the 'assign' function to assign each element.
     */
    void assign(const UVector64& other, UErrorCode &ec);

    /**
     * Compare this vector with another.  They will be considered
     * equal if they are of the same size and all elements are equal,
     * as compared using this object's comparer.
     */
    UBool operator==(const UVector64& other);

    /**
     * Equivalent to !operator==()
     */
    inline UBool operator!=(const UVector64& other);

    //------------------------------------------------------------
    // subset of java.util.Vector API
    //------------------------------------------------------------

    inline void addElement(int64_t elem, UErrorCode &status);

    void setElementAt(int64_t elem, int32_t index);

    void insertElementAt(int64_t elem, int32_t index, UErrorCode &status);
    
    inline int64_t elementAti(int32_t index) const;

    //UBool equals(const UVector64 &other) const;

    inline int64_t lastElementi(void) const;

    //int32_t indexOf(int64_t elem, int32_t startIndex = 0) const;

    //UBool contains(int64_t elem) const;

    //UBool containsAll(const UVector64& other) const;

    //UBool removeAll(const UVector64& other);

    //UBool retainAll(const UVector64& other);

    //void removeElementAt(int32_t index);

    void removeAllElements();

    inline int32_t size(void) const;

    inline UBool isEmpty(void) const { return count == 0; }

    // Inline.  Use this one for speedy size check.
    inline UBool ensureCapacity(int32_t minimumCapacity, UErrorCode &status);

    // Out-of-line, handles actual growth.  Called by ensureCapacity() when necessary.
    UBool expandCapacity(int32_t minimumCapacity, UErrorCode &status);

    /**
     * Change the size of this vector as follows: If newSize is
     * smaller, then truncate the array, possibly deleting held
     * elements for i >= newSize.  If newSize is larger, grow the
     * array, filling in new slows with zero.
     */
    void setSize(int32_t newSize);

    //------------------------------------------------------------
    // New API
    //------------------------------------------------------------

    //UBool containsNone(const UVector64& other) const;


    //void sortedInsert(int64_t elem, UErrorCode& ec);

    /**
     * Returns a pointer to the internal array holding the vector.
     */
    inline int64_t *getBuffer() const;

    /**
     * Set the maximum allowed buffer capacity for this vector/stack.
     * Default with no limit set is unlimited, go until malloc() fails.
     * A Limit of zero means unlimited capacity.
     * Units are vector elements (64 bits each), not bytes.
     */
    void setMaxCapacity(int32_t limit);

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     */
    virtual UClassID getDynamicClassID() const;

private:
    void _init(int32_t initialCapacity, UErrorCode &status);

    // Disallow
    UVector64(const UVector64&);

    // Disallow
    UVector64& operator=(const UVector64&);


    //  API Functions for Stack operations.
    //  In the original UVector, these were in a separate derived class, UStack.
    //  Here in UVector64, they are all together.
public:
    //UBool empty(void) const;   // TODO:  redundant, same as empty().  Remove it?

    //int64_t peeki(void) const;
    
    inline int64_t popi(void);
    
    inline int64_t push(int64_t i, UErrorCode &status);

    inline int64_t *reserveBlock(int32_t size, UErrorCode &status);
    inline int64_t *popFrame(int32_t size);
};


// UVector64 inlines

inline UBool UVector64::ensureCapacity(int32_t minimumCapacity, UErrorCode &status) {
    if ((minimumCapacity >= 0) && (capacity >= minimumCapacity)) {
        return TRUE;
    } else {
        return expandCapacity(minimumCapacity, status);
    }
}

inline int64_t UVector64::elementAti(int32_t index) const {
    return (0 <= index && index < count) ? elements[index] : 0;
}


inline void UVector64::addElement(int64_t elem, UErrorCode &status) {
    if (ensureCapacity(count + 1, status)) {
        elements[count] = elem;
        count++;
    }
}

inline int64_t *UVector64::reserveBlock(int32_t size, UErrorCode &status) {
    if (ensureCapacity(count+size, status) == FALSE) {
        return NULL;
    }
    int64_t  *rp = elements+count;
    count += size;
    return rp;
}

inline int64_t *UVector64::popFrame(int32_t size) {
    U_ASSERT(count >= size);
    count -= size;
    if (count < 0) {
        count = 0;
    }
    return elements+count-size;
}



inline int32_t UVector64::size(void) const {
    return count;
}

inline int64_t UVector64::lastElementi(void) const {
    return elementAti(count-1);
}

inline UBool UVector64::operator!=(const UVector64& other) {
    return !operator==(other);
}

inline int64_t *UVector64::getBuffer() const {
    return elements;
}


// UStack inlines

inline int64_t UVector64::push(int64_t i, UErrorCode &status) {
    addElement(i, status);
    return i;
}

inline int64_t UVector64::popi(void) {
    int64_t result = 0;
    if (count > 0) {
        count--;
        result = elements[count];
    }
    return result;
}

U_NAMESPACE_END





U_NAMESPACE_BEGIN

#define UVECTOR64_DEFAULT_CAPACITY 8

/*
 * Constants for hinting whether a key is an integer
 * or a pointer.  If a hint bit is zero, then the associated
 * token is assumed to be an integer. This is needed for iSeries
 */

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(UVector64)

UVector64::UVector64(UErrorCode &status) :
    count(0),
    capacity(0),
    maxCapacity(0),
    elements(NULL)
{
    _init(UVECTOR64_DEFAULT_CAPACITY, status);
}

UVector64::UVector64(int32_t initialCapacity, UErrorCode &status) :
    count(0),
    capacity(0),
    maxCapacity(0),
    elements(0)
{
    _init(initialCapacity, status);
}



void UVector64::_init(int32_t initialCapacity, UErrorCode &status) {
    // Fix bogus initialCapacity values; avoid malloc(0)
    if (initialCapacity < 1) {
        initialCapacity = UVECTOR64_DEFAULT_CAPACITY;
    }
    if (maxCapacity>0 && maxCapacity<initialCapacity) {
        initialCapacity = maxCapacity;
    }
    if (initialCapacity > (int32_t)(INT32_MAX / sizeof(int64_t))) {
        initialCapacity = uprv_min(UVECTOR64_DEFAULT_CAPACITY, maxCapacity);
    }
    elements = (int64_t *)uprv_malloc(sizeof(int64_t)*initialCapacity);
    if (elements == 0) {
        status = U_MEMORY_ALLOCATION_ERROR;
    } else {
        capacity = initialCapacity;
    }
}

UVector64::~UVector64() {
    uprv_free(elements);
    elements = 0;
}

/**
 * Assign this object to another (make this a copy of 'other').
 */
void UVector64::assign(const UVector64& other, UErrorCode &ec) {
    if (ensureCapacity(other.count, ec)) {
        setSize(other.count);
        for (int32_t i=0; i<other.count; ++i) {
            elements[i] = other.elements[i];
        }
    }
}


UBool UVector64::operator==(const UVector64& other) {
    int32_t i;
    if (count != other.count) return FALSE;
    for (i=0; i<count; ++i) {
        if (elements[i] != other.elements[i]) {
            return FALSE;
        }
    }
    return TRUE;
}


void UVector64::setElementAt(int64_t elem, int32_t index) {
    if (0 <= index && index < count) {
        elements[index] = elem;
    }
    /* else index out of range */
}

void UVector64::insertElementAt(int64_t elem, int32_t index, UErrorCode &status) {
    // must have 0 <= index <= count
    if (0 <= index && index <= count && ensureCapacity(count + 1, status)) {
        for (int32_t i=count; i>index; --i) {
            elements[i] = elements[i-1];
        }
        elements[index] = elem;
        ++count;
    }
    /* else index out of range */
}

void UVector64::removeAllElements(void) {
    count = 0;
}

UBool UVector64::expandCapacity(int32_t minimumCapacity, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (minimumCapacity < 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
    if (capacity >= minimumCapacity) {
        return TRUE;
    }
    if (maxCapacity>0 && minimumCapacity>maxCapacity) {
        status = U_BUFFER_OVERFLOW_ERROR;
        return FALSE;
    }
    if (capacity > (INT32_MAX - 1) / 2) {  // integer overflow check
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
    int32_t newCap = capacity * 2;
    if (newCap < minimumCapacity) {
        newCap = minimumCapacity;
    }
    if (maxCapacity > 0 && newCap > maxCapacity) {
        newCap = maxCapacity;
    }
    if (newCap > (int32_t)(INT32_MAX / sizeof(int64_t))) {  // integer overflow check
        // We keep the original memory contents on bad minimumCapacity/maxCapacity.
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
    int64_t* newElems = (int64_t *)uprv_realloc(elements, sizeof(int64_t)*newCap);
    if (newElems == NULL) {
        // We keep the original contents on the memory failure on realloc.
        status = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    elements = newElems;
    capacity = newCap;
    return TRUE;
}

void UVector64::setMaxCapacity(int32_t limit) {
    U_ASSERT(limit >= 0);
    if (limit < 0) {
        limit = 0;
    }
    if (limit > (int32_t)(INT32_MAX / sizeof(int64_t))) {  // integer overflow check for realloc
        //  Something is very wrong, don't realloc, leave capacity and maxCapacity unchanged
        return;
    }
    maxCapacity = limit;
    if (capacity <= maxCapacity || maxCapacity == 0) {
        // Current capacity is within the new limit.
        return;
    }

    // New maximum capacity is smaller than the current size.
    // Realloc the storage to the new, smaller size.
    int64_t* newElems = (int64_t *)uprv_realloc(elements, sizeof(int64_t)*maxCapacity);
    if (newElems == NULL) {
        // Realloc to smaller failed.
        //   Just keep what we had.  No need to call it a failure.
        return;
    }
    elements = newElems;
    capacity = maxCapacity;
    if (count > capacity) {
        count = capacity;
    }
}

/**
 * Change the size of this vector as follows: If newSize is smaller,
 * then truncate the array, possibly deleting held elements for i >=
 * newSize.  If newSize is larger, grow the array, filling in new
 * slots with NULL.
 */
void UVector64::setSize(int32_t newSize) {
    int32_t i;
    if (newSize < 0) {
        return;
    }
    if (newSize > count) {
        UErrorCode ec = U_ZERO_ERROR;
        if (!ensureCapacity(newSize, ec)) {
            return;
        }
        for (i=count; i<newSize; ++i) {
            elements[i] = 0;
        }
    }
    count = newSize;
}

U_NAMESPACE_END

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2014, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_COLLATION

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2011-2014 International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API





#if !UCONFIG_NO_COLLATION

/**
 * \file
 * \brief C++ API: Index Characters
 */

U_CDECL_BEGIN

/**
 * Constants for Alphabetic Index Label Types.
 * The form of these enum constants anticipates having a plain C API
 * for Alphabetic Indexes that will also use them.
 * @stable ICU 4.8
 */
typedef enum UAlphabeticIndexLabelType {
    /**
     *  Normal Label, typically the starting letter of the names
     *  in the bucket with this label.
     * @stable ICU 4.8
     */
    U_ALPHAINDEX_NORMAL    = 0,

    /**
     * Undeflow Label.  The bucket with this label contains names
     * in scripts that sort before any of the bucket labels in this index.
     * @stable ICU 4.8
     */
    U_ALPHAINDEX_UNDERFLOW = 1,

    /**
     * Inflow Label.  The bucket with this label contains names
     * in scripts that sort between two of the bucket labels in this index.
     * Inflow labels are created when an index contains normal labels for
     * multiple scripts, and skips other scripts that sort between some of the
     * included scripts.
     * @stable ICU 4.8
     */
    U_ALPHAINDEX_INFLOW    = 2,

    /**
     * Overflow Label. Te bucket with this label contains names in scripts
     * that sort after all of the bucket labels in this index.
     * @stable ICU 4.8
     */
    U_ALPHAINDEX_OVERFLOW  = 3
} UAlphabeticIndexLabelType;


struct UHashtable;
U_CDECL_END

U_NAMESPACE_BEGIN

// Forward Declarations

class BucketList;
class Collator;
class RuleBasedCollator;
class StringEnumeration;
class UnicodeSet;
class UVector;

/**
 * AlphabeticIndex supports the creation of a UI index appropriate for a given language.
 * It can support either direct use, or use with a client that doesn't support localized collation.
 * The following is an example of what an index might look like in a UI:
 * 
 * <pre>
 *  <b>... A B C D E F G H I J K L M N O P Q R S T U V W X Y Z  ...</b>
 *
 *  <b>A</b>
 *     Addison
 *     Albertson
 *     Azensky
 *  <b>B</b>
 *     Baker
 *  ...
 * </pre>
 *
 * The class can generate a list of labels for use as a UI "index", that is, a list of
 * clickable characters (or character sequences) that allow the user to see a segment
 * (bucket) of a larger "target" list. That is, each label corresponds to a bucket in
 * the target list, where everything in the bucket is greater than or equal to the character
 * (according to the locale's collation). Strings can be added to the index;
 * they will be in sorted order in the right bucket.
 * <p>
 * The class also supports having buckets for strings before the first (underflow),
 * after the last (overflow), and between scripts (inflow). For example, if the index
 * is constructed with labels for Russian and English, Greek characters would fall
 * into an inflow bucket between the other two scripts.
 * <p>
 * The AlphabeticIndex class is not intended for public subclassing.
 *
 * <p><em>Note:</em> If you expect to have a lot of ASCII or Latin characters
 * as well as characters from the user's language,
 * then it is a good idea to call addLabels(Locale::getEnglish(), status).</p>
 *
 * <h2>Direct Use</h2>
 * <p>The following shows an example of building an index directly.
 *  The "show..." methods below are just to illustrate usage.
 *
 * <pre>
 * // Create a simple index.  "Item" is assumed to be an application
 * // defined type that the application's UI and other processing knows about,
 * //  and that has a name.
 *
 * UErrorCode status = U_ZERO_ERROR;
 * AlphabeticIndex index = new AlphabeticIndex(desiredLocale, status);
 * index->addLabels(additionalLocale, status);
 * for (Item *item in some source of Items ) {
 *     index->addRecord(item->name(), item, status);
 * }
 * ...
 * // Show index at top. We could skip or gray out empty buckets
 *
 * while (index->nextBucket(status)) {
 *     if (showAll || index->getBucketRecordCount() != 0) {
 *         showLabelAtTop(UI, index->getBucketLabel());
 *     }
 * }
 *  ...
 * // Show the buckets with their contents, skipping empty buckets
 *
 * index->resetBucketIterator(status);
 * while (index->nextBucket(status)) {
 *    if (index->getBucketRecordCount() != 0) {
 *        showLabelInList(UI, index->getBucketLabel());
 *        while (index->nextRecord(status)) {
 *            showIndexedItem(UI, static_cast<Item *>(index->getRecordData()))
 * </pre>
 *
 * The caller can build different UIs using this class.
 * For example, an index character could be omitted or grayed-out
 * if its bucket is empty. Small buckets could also be combined based on size, such as:
 *
 * <pre>
 * <b>... A-F G-N O-Z ...</b>
 * </pre>
 *
 * <h2>Client Support</h2>
 * <p>Callers can also use the AlphabeticIndex::ImmutableIndex, or the AlphabeticIndex itself,
 * to support sorting on a client that doesn't support AlphabeticIndex functionality.
 *
 * <p>The ImmutableIndex is both immutable and thread-safe.
 * The corresponding AlphabeticIndex methods are not thread-safe because
 * they "lazily" build the index buckets.
 * <ul>
 * <li>ImmutableIndex.getBucket(index) provides random access to all
 *     buckets and their labels and label types.
 * <li>The AlphabeticIndex bucket iterator or ImmutableIndex.getBucket(0..getBucketCount-1)
 *     can be used to get a list of the labels,
 *     such as "...", "A", "B",..., and send that list to the client.
 * <li>When the client has a new name, it sends that name to the server.
 * The server needs to call the following methods,
 * and communicate the bucketIndex and collationKey back to the client.
 *
 * <pre>
 * int32_t bucketIndex = index.getBucketIndex(name, status);
 * const UnicodeString &label = immutableIndex.getBucket(bucketIndex)->getLabel();  // optional
 * int32_t skLength = collator.getSortKey(name, sk, skCapacity);
 * </pre>
 *
 * <li>The client would put the name (and associated information) into its bucket for bucketIndex. The sort key sk is a
 * sequence of bytes that can be compared with a binary compare, and produce the right localized result.</li>
 * </ul>
 *
 * @stable ICU 4.8
 */
class U_I18N_API AlphabeticIndex: public UObject {
public:
     /**
      * An index "bucket" with a label string and type.
      * It is referenced by getBucketIndex(),
      * and returned by ImmutableIndex.getBucket().
      *
      * The Bucket class is not intended for public subclassing.
      * @stable ICU 51
      */
     class U_I18N_API Bucket : public UObject {
     public:
        /**
         * Destructor.
         * @stable ICU 51
         */
        virtual ~Bucket();

        /**
         * Returns the label string.
         *
         * @return the label string for the bucket
         * @stable ICU 51
         */
        const UnicodeString &getLabel() const { return label_; }
        /**
         * Returns whether this bucket is a normal, underflow, overflow, or inflow bucket.
         *
         * @return the bucket label type
         * @stable ICU 51
         */
        UAlphabeticIndexLabelType getLabelType() const { return labelType_; }

     private:
        friend class AlphabeticIndex;
        friend class BucketList;

        UnicodeString label_;
        UnicodeString lowerBoundary_;
        UAlphabeticIndexLabelType labelType_;
        Bucket *displayBucket_;
        int32_t displayIndex_;
        UVector *records_;  // Records are owned by the inputList_ vector.

        Bucket(const UnicodeString &label,   // Parameter strings are copied.
               const UnicodeString &lowerBoundary,
               UAlphabeticIndexLabelType type);
     };

    /**
     * Immutable, thread-safe version of AlphabeticIndex.
     * This class provides thread-safe methods for bucketing,
     * and random access to buckets and their properties,
     * but does not offer adding records to the index.
     *
     * The ImmutableIndex class is not intended for public subclassing.
     *
     * @stable ICU 51
     */
    class U_I18N_API ImmutableIndex : public UObject {
    public:
        /**
         * Destructor.
         * @stable ICU 51
         */
        virtual ~ImmutableIndex();

        /**
         * Returns the number of index buckets and labels, including underflow/inflow/overflow.
         *
         * @return the number of index buckets
         * @stable ICU 51
         */
        int32_t getBucketCount() const;

        /**
         * Finds the index bucket for the given name and returns the number of that bucket.
         * Use getBucket() to get the bucket's properties.
         *
         * @param name the string to be sorted into an index bucket
         * @param errorCode Error code, will be set with the reason if the
         *                  operation fails.
         * @return the bucket number for the name
         * @stable ICU 51
         */
        int32_t getBucketIndex(const UnicodeString &name, UErrorCode &errorCode) const;

        /**
         * Returns the index-th bucket. Returns NULL if the index is out of range.
         *
         * @param index bucket number
         * @return the index-th bucket
         * @stable ICU 51
         */
        const Bucket *getBucket(int32_t index) const;

    private:
        friend class AlphabeticIndex;

        ImmutableIndex(BucketList *bucketList, Collator *collatorPrimaryOnly)
                : buckets_(bucketList), collatorPrimaryOnly_(collatorPrimaryOnly) {}

        BucketList *buckets_;
        Collator *collatorPrimaryOnly_;
    };

    /**
     * Construct an AlphabeticIndex object for the specified locale.  If the locale's
     * data does not include index characters, a set of them will be
     * synthesized based on the locale's exemplar characters.  The locale
     * determines the sorting order for both the index characters and the
     * user item names appearing under each Index character.
     *
     * @param locale the desired locale.
     * @param status Error code, will be set with the reason if the construction
     *               of the AlphabeticIndex object fails.
     * @stable ICU 4.8
     */
     AlphabeticIndex(const Locale &locale, UErrorCode &status);

   /** 
     * Construct an AlphabeticIndex that uses a specific collator.
     * 
     * The index will be created with no labels; the addLabels() function must be called
     * after creation to add the desired labels to the index.
     * 
     * The index adopts the collator, and is responsible for deleting it. 
     * The caller should make no further use of the collator after creating the index.
     * 
     * @param collator The collator to use to order the contents of this index.
     * @param status Error code, will be set with the reason if the 
     *               operation fails.
     * @stable ICU 51
     */
    AlphabeticIndex(RuleBasedCollator *collator, UErrorCode &status);

    /**
     * Add Labels to this Index.  The labels are additions to those
     * that are already in the index; they do not replace the existing
     * ones.
     * @param additions The additional characters to add to the index, such as A-Z.
     * @param status Error code, will be set with the reason if the 
     *               operation fails.
     * @return this, for chaining
     * @stable ICU 4.8
     */
    virtual AlphabeticIndex &addLabels(const UnicodeSet &additions, UErrorCode &status);

    /**
     * Add the index characters from a Locale to the index.  The labels
     * are added to those that are already in the index; they do not replace the
     * existing index characters.  The collation order for this index is not
     * changed; it remains that of the locale that was originally specified
     * when creating this Index.
     *
     * @param locale The locale whose index characters are to be added.
     * @param status Error code, will be set with the reason if the 
     *               operation fails.
     * @return this, for chaining
     * @stable ICU 4.8
     */
    virtual AlphabeticIndex &addLabels(const Locale &locale, UErrorCode &status);

     /**
      * Destructor
      * @stable ICU 4.8
      */
    virtual ~AlphabeticIndex();

    /**
     * Builds an immutable, thread-safe version of this instance, without data records.
     *
     * @return an immutable index instance
     * @stable ICU 51
     */
    ImmutableIndex *buildImmutableIndex(UErrorCode &errorCode);

    /**
     * Get the Collator that establishes the ordering of the items in this index.
     * Ownership of the collator remains with the AlphabeticIndex instance.
     *
     * The returned collator is a reference to the internal collator used by this
     * index.  It may be safely used to compare the names of items or to get
     * sort keys for names.  However if any settings need to be changed,
     * or other non-const methods called, a cloned copy must be made first.
     *
     * @return The collator
     * @stable ICU 4.8
     */
    virtual const RuleBasedCollator &getCollator() const;


   /**
     * Get the default label used for abbreviated buckets *between* other index characters.
     * For example, consider the labels when Latin (X Y Z) and Greek (Α Β Γ) are used:
     *
     *     X Y Z ... Α Β Γ.
     *
     * @return inflow label
     * @stable ICU 4.8
     */
    virtual const UnicodeString &getInflowLabel() const;

   /**
     * Set the default label used for abbreviated buckets <i>between</i> other index characters.
     * An inflow label will be automatically inserted if two otherwise-adjacent label characters
     * are from different scripts, e.g. Latin and Cyrillic, and a third script, e.g. Greek,
     * sorts between the two.  The default inflow character is an ellipsis (...)
     *
     * @param inflowLabel the new Inflow label.
     * @param status Error code, will be set with the reason if the operation fails.
     * @return this
     * @stable ICU 4.8
     */
    virtual AlphabeticIndex &setInflowLabel(const UnicodeString &inflowLabel, UErrorCode &status);


   /**
     * Get the special label used for items that sort after the last normal label,
     * and that would not otherwise have an appropriate label.
     *
     * @return the overflow label
     * @stable ICU 4.8
     */
    virtual const UnicodeString &getOverflowLabel() const;


   /**
     * Set the label used for items that sort after the last normal label,
     * and that would not otherwise have an appropriate label.
     *
     * @param overflowLabel the new overflow label.
     * @param status Error code, will be set with the reason if the operation fails.
     * @return this
     * @stable ICU 4.8
     */
    virtual AlphabeticIndex &setOverflowLabel(const UnicodeString &overflowLabel, UErrorCode &status);

   /**
     * Get the special label used for items that sort before the first normal label,
     * and that would not otherwise have an appropriate label.
     *
     * @return underflow label
     * @stable ICU 4.8
     */
    virtual const UnicodeString &getUnderflowLabel() const;

   /**
     * Set the label used for items that sort before the first normal label,
     * and that would not otherwise have an appropriate label.
     *
     * @param underflowLabel the new underflow label.
     * @param status Error code, will be set with the reason if the operation fails.
     * @return this
     * @stable ICU 4.8
     */
    virtual AlphabeticIndex &setUnderflowLabel(const UnicodeString &underflowLabel, UErrorCode &status);


    /**
     * Get the limit on the number of labels permitted in the index.
     * The number does not include over, under and inflow labels.
     *
     * @return maxLabelCount maximum number of labels.
     * @stable ICU 4.8
     */
    virtual int32_t getMaxLabelCount() const;

    /**
     * Set a limit on the number of labels permitted in the index.
     * The number does not include over, under and inflow labels.
     * Currently, if the number is exceeded, then every
     * nth item is removed to bring the count down.
     * A more sophisticated mechanism may be available in the future.
     *
     * @param maxLabelCount the maximum number of labels.
     * @param status error code
     * @return This, for chaining
     * @stable ICU 4.8
     */
    virtual AlphabeticIndex &setMaxLabelCount(int32_t maxLabelCount, UErrorCode &status);


    /**
     * Add a record to the index.  Each record will be associated with an index Bucket
     *  based on the record's name.  The list of records for each bucket will be sorted
     *  based on the collation ordering of the names in the index's locale.
     *  Records with duplicate names are permitted; they will be kept in the order
     *  that they were added.
     *
     * @param name The display name for the Record.  The Record will be placed in
     *             a bucket based on this name.
     * @param data An optional pointer to user data associated with this
     *             item.  When iterating the contents of a bucket, both the
     *             data pointer the name will be available for each Record.
     * @param status  Error code, will be set with the reason if the operation fails.
     * @return        This, for chaining.
     * @stable ICU 4.8
     */
    virtual AlphabeticIndex &addRecord(const UnicodeString &name, const void *data, UErrorCode &status);

    /**
     * Remove all Records from the Index.  The set of Buckets, which define the headings under
     * which records are classified, is not altered.
     *
     * @param status  Error code, will be set with the reason if the operation fails.
     * @return        This, for chaining.
     * @stable ICU 4.8
     */
    virtual AlphabeticIndex &clearRecords(UErrorCode &status);


    /**  Get the number of labels in this index.
     *      Note: may trigger lazy index construction.
     *
     * @param status  Error code, will be set with the reason if the operation fails.
     * @return        The number of labels in this index, including any under, over or
     *                in-flow labels.
     * @stable ICU 4.8
     */
    virtual int32_t  getBucketCount(UErrorCode &status);


    /**  Get the total number of Records in this index, that is, the number
     *   of <name, data> pairs added.
     *
     * @param status  Error code, will be set with the reason if the operation fails.
     * @return        The number of records in this index, that is, the total number
     *                of (name, data) items added with addRecord().
     * @stable ICU 4.8
     */
    virtual int32_t  getRecordCount(UErrorCode &status);



    /**
     *   Given the name of a record, return the zero-based index of the Bucket
     *   in which the item should appear.  The name need not be in the index.
     *   A Record will not be added to the index by this function.
     *   Bucket numbers are zero-based, in Bucket iteration order.
     *
     * @param itemName  The name whose bucket position in the index is to be determined.
     * @param status  Error code, will be set with the reason if the operation fails.
     * @return The bucket number for this name.
     * @stable ICU 4.8
     *
     */
    virtual int32_t  getBucketIndex(const UnicodeString &itemName, UErrorCode &status);


    /**
     *   Get the zero based index of the current Bucket from an iteration
     *   over the Buckets of this index.  Return -1 if no iteration is in process.
     *   @return  the index of the current Bucket
     *   @stable ICU 4.8
     */
    virtual int32_t  getBucketIndex() const;


    /**
     *   Advance the iteration over the Buckets of this index.  Return FALSE if
     *   there are no more Buckets.
     *
     *   @param status  Error code, will be set with the reason if the operation fails.
     *   U_ENUM_OUT_OF_SYNC_ERROR will be reported if the index is modified while
     *   an enumeration of its contents are in process.
     *
     *   @return TRUE if success, FALSE if at end of iteration
     *   @stable ICU 4.8
     */
    virtual UBool nextBucket(UErrorCode &status);

    /**
     *   Return the name of the Label of the current bucket from an iteration over the buckets.
     *   If the iteration is before the first Bucket (nextBucket() has not been called),
     *   or after the last, return an empty string.
     *
     *   @return the bucket label.
     *   @stable ICU 4.8
     */
    virtual const UnicodeString &getBucketLabel() const;

    /**
     *  Return the type of the label for the current Bucket (selected by the
     *  iteration over Buckets.)
     *
     * @return the label type.
     * @stable ICU 4.8
     */
    virtual UAlphabeticIndexLabelType getBucketLabelType() const;

    /**
      * Get the number of <name, data> Records in the current Bucket.
      * If the current bucket iteration position is before the first label or after the
      * last, return 0.
      *
      *  @return the number of Records.
      *  @stable ICU 4.8
      */
    virtual int32_t getBucketRecordCount() const;


    /**
     *  Reset the Bucket iteration for this index.  The next call to nextBucket()
     *  will restart the iteration at the first label.
     *
     * @param status  Error code, will be set with the reason if the operation fails.
     * @return        this, for chaining.
     * @stable ICU 4.8
     */
    virtual AlphabeticIndex &resetBucketIterator(UErrorCode &status);

    /**
     * Advance to the next record in the current Bucket.
     * When nextBucket() is called, Record iteration is reset to just before the
     * first Record in the new Bucket.
     *
     *   @param status  Error code, will be set with the reason if the operation fails.
     *   U_ENUM_OUT_OF_SYNC_ERROR will be reported if the index is modified while
     *   an enumeration of its contents are in process.
     *   @return TRUE if successful, FALSE when the iteration advances past the last item.
     *   @stable ICU 4.8
     */
    virtual UBool nextRecord(UErrorCode &status);

    /**
     * Get the name of the current Record.
     * Return an empty string if the Record iteration position is before first
     * or after the last.
     *
     *  @return The name of the current index item.
     *  @stable ICU 4.8
     */
    virtual const UnicodeString &getRecordName() const;


    /**
     * Return the data pointer of the Record currently being iterated over.
     * Return NULL if the current iteration position before the first item in this Bucket,
     * or after the last.
     *
     *  @return The current Record's data pointer.
     *  @stable ICU 4.8
     */
    virtual const void *getRecordData() const;


    /**
     * Reset the Record iterator position to before the first Record in the current Bucket.
     *
     *  @return This, for chaining.
     *  @stable ICU 4.8
     */
    virtual AlphabeticIndex &resetRecordIterator();

private:
     /**
      * No Copy constructor.
      * @internal
      */
     AlphabeticIndex(const AlphabeticIndex &other);

     /**
      *   No assignment.
      */
     AlphabeticIndex &operator =(const AlphabeticIndex & /*other*/) { return *this;}

    /**
     * No Equality operators.
     * @internal
     */
     virtual UBool operator==(const AlphabeticIndex& other) const;

    /**
     * Inequality operator.
     * @internal
     */
     virtual UBool operator!=(const AlphabeticIndex& other) const;

     // Common initialization, for use from all constructors.
     void init(const Locale *locale, UErrorCode &status);

    /**
     * This method is called to get the index exemplars. Normally these come from the locale directly,
     * but if they aren't available, we have to synthesize them.
     */
    void addIndexExemplars(const Locale &locale, UErrorCode &status);
    /**
     * Add Chinese index characters from the tailoring.
     */
    UBool addChineseIndexCharacters(UErrorCode &errorCode);

    UVector *firstStringsInScript(UErrorCode &status);

    static UnicodeString separated(const UnicodeString &item);

    /**
     * Determine the best labels to use.
     * This is based on the exemplars, but we also process to make sure that they are unique,
     * and sort differently, and that the overall list is small enough.
     */
    void initLabels(UVector &indexCharacters, UErrorCode &errorCode) const;
    BucketList *createBucketList(UErrorCode &errorCode) const;
    void initBuckets(UErrorCode &errorCode);
    void clearBuckets();
    void internalResetBucketIterator();

public:

    //  The Record is declared public only to allow access from
    //  implementation code written in plain C.
    //  It is not intended for public use.

#ifndef U_HIDE_INTERNAL_API
    /**
     * A (name, data) pair, to be sorted by name into one of the index buckets.
     * The user data is not used by the index implementation.
     * \cond
     * @internal
     */
    struct Record: public UMemory {
        const UnicodeString  name_;
        const void           *data_;
        Record(const UnicodeString &name, const void *data);
        ~Record();
    };
    /** \endcond */
#endif  /* U_HIDE_INTERNAL_API */

private:

    /**
     * Holds all user records before they are distributed into buckets.
     * Type of contents is (Record *)
     * @internal
     */
    UVector  *inputList_;

    int32_t  labelsIterIndex_;        // Index of next item to return.
    int32_t  itemsIterIndex_;
    Bucket   *currentBucket_;         // While an iteration of the index in underway,
                                      //   point to the bucket for the current label.
                                      // NULL when no iteration underway.

    int32_t    maxLabelCount_;        // Limit on # of labels permitted in the index.

    UnicodeSet *initialLabels_;       // Initial (unprocessed) set of Labels.  Union
                                      //   of those explicitly set by the user plus
                                      //   those from locales.  Raw values, before
                                      //   crunching into bucket labels.

    UVector *firstCharsInScripts_;    // The first character from each script,
                                      //   in collation order.

    RuleBasedCollator *collator_;
    RuleBasedCollator *collatorPrimaryOnly_;

    // Lazy evaluated: null means that we have not built yet.
    BucketList *buckets_;

    UnicodeString  inflowLabel_;
    UnicodeString  overflowLabel_;
    UnicodeString  underflowLabel_;
    UnicodeString  overflowComparisonString_;

    UnicodeString emptyString_;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION

#endif /* U_SHOW_CPLUSPLUS_API */





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 1996-2016, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*/

/**
 * \file
 * \brief C++ API: The RuleBasedCollator class implements the Collator abstract base class.
 */

/**
* File tblcoll.h
*
* Created by: Helena Shih
*
* Modification History:
*
*  Date        Name        Description
*  2/5/97      aliu        Added streamIn and streamOut methods.  Added
*                          constructor which reads RuleBasedCollator object from
*                          a binary file.  Added writeToFile method which streams
*                          RuleBasedCollator out to a binary file.  The streamIn
*                          and streamOut methods use istream and ostream objects
*                          in binary mode.
*  2/12/97     aliu        Modified to use TableCollationData sub-object to
*                          hold invariant data.
*  2/13/97     aliu        Moved several methods into this class from Collation.
*                          Added a private RuleBasedCollator(Locale&) constructor,
*                          to be used by Collator::createDefault().  General
*                          clean up.
*  2/20/97     helena      Added clone, operator==, operator!=, operator=, and copy
*                          constructor and getDynamicClassID.
*  3/5/97      aliu        Modified constructFromFile() to add parameter
*                          specifying whether or not binary loading is to be
*                          attempted.  This is required for dynamic rule loading.
* 05/07/97     helena      Added memory allocation error detection.
*  6/17/97     helena      Added IDENTICAL strength for compare, changed getRules to
*                          use MergeCollation::getPattern.
*  6/20/97     helena      Java class name change.
*  8/18/97     helena      Added internal API documentation.
* 09/03/97     helena      Added createCollationKeyValues().
* 02/10/98     damiba      Added compare with "length" parameter
* 08/05/98     erm         Synched with 1.2 version of RuleBasedCollator.java
* 04/23/99     stephen     Removed EDecompositionMode, merged with
*                          Normalizer::EMode
* 06/14/99     stephen     Removed kResourceBundleSuffix
* 11/02/99     helena      Collator performance enhancements.  Eliminates the
*                          UnicodeString construction and special case for NO_OP.
* 11/23/99     srl         More performance enhancements. Updates to NormalizerIterator
*                          internal state management.
* 12/15/99     aliu        Update to support Thai collation.  Move NormalizerIterator
*                          to implementation file.
* 01/29/01     synwee      Modified into a C++ wrapper which calls C API
*                          (ucol.h)
* 2012-2014    markus      Rewritten in C++ again.
*/




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_COLLATION






U_NAMESPACE_BEGIN

struct CollationCacheEntry;
struct CollationData;
struct CollationSettings;
struct CollationTailoring;
/**
* @stable ICU 2.0
*/
class StringSearch;
/**
* @stable ICU 2.0
*/
class CollationElementIterator;
class CollationKey;
class SortKeyByteSink;
class UnicodeSet;
class UnicodeString;
class UVector64;

/**
 * The RuleBasedCollator class provides the implementation of
 * Collator, using data-driven tables. The user can create a customized
 * table-based collation.
 * <p>
 * For more information about the collation service see
 * <a href="http://userguide.icu-project.org/collation">the User Guide</a>.
 * <p>
 * Collation service provides correct sorting orders for most locales supported in ICU.
 * If specific data for a locale is not available, the orders eventually falls back
 * to the <a href="http://www.unicode.org/reports/tr35/tr35-collation.html#Root_Collation">CLDR root sort order</a>.
 * <p>
 * Sort ordering may be customized by providing your own set of rules. For more on
 * this subject see the <a href="http://userguide.icu-project.org/collation/customization">
 * Collation Customization</a> section of the User Guide.
 * <p>
 * Note, RuleBasedCollator is not to be subclassed.
 * @see        Collator
 */
class U_I18N_API RuleBasedCollator : public Collator {
public:
    /**
     * RuleBasedCollator constructor. This takes the table rules and builds a
     * collation table out of them. Please see RuleBasedCollator class
     * description for more details on the collation rule syntax.
     * @param rules the collation rules to build the collation table from.
     * @param status reporting a success or an error.
     * @stable ICU 2.0
     */
    RuleBasedCollator(const UnicodeString& rules, UErrorCode& status);

    /**
     * RuleBasedCollator constructor. This takes the table rules and builds a
     * collation table out of them. Please see RuleBasedCollator class
     * description for more details on the collation rule syntax.
     * @param rules the collation rules to build the collation table from.
     * @param collationStrength strength for comparison
     * @param status reporting a success or an error.
     * @stable ICU 2.0
     */
    RuleBasedCollator(const UnicodeString& rules,
                       ECollationStrength collationStrength,
                       UErrorCode& status);

    /**
     * RuleBasedCollator constructor. This takes the table rules and builds a
     * collation table out of them. Please see RuleBasedCollator class
     * description for more details on the collation rule syntax.
     * @param rules the collation rules to build the collation table from.
     * @param decompositionMode the normalisation mode
     * @param status reporting a success or an error.
     * @stable ICU 2.0
     */
    RuleBasedCollator(const UnicodeString& rules,
                    UColAttributeValue decompositionMode,
                    UErrorCode& status);

    /**
     * RuleBasedCollator constructor. This takes the table rules and builds a
     * collation table out of them. Please see RuleBasedCollator class
     * description for more details on the collation rule syntax.
     * @param rules the collation rules to build the collation table from.
     * @param collationStrength strength for comparison
     * @param decompositionMode the normalisation mode
     * @param status reporting a success or an error.
     * @stable ICU 2.0
     */
    RuleBasedCollator(const UnicodeString& rules,
                    ECollationStrength collationStrength,
                    UColAttributeValue decompositionMode,
                    UErrorCode& status);

#ifndef U_HIDE_INTERNAL_API
    /**
     * TODO: document & propose as public API
     * @internal
     */
    RuleBasedCollator(const UnicodeString &rules,
                      UParseError &parseError, UnicodeString &reason,
                      UErrorCode &errorCode);
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Copy constructor.
     * @param other the RuleBasedCollator object to be copied
     * @stable ICU 2.0
     */
    RuleBasedCollator(const RuleBasedCollator& other);


    /** Opens a collator from a collator binary image created using
    *  cloneBinary. Binary image used in instantiation of the
    *  collator remains owned by the user and should stay around for
    *  the lifetime of the collator. The API also takes a base collator
    *  which must be the root collator.
    *  @param bin binary image owned by the user and required through the
    *             lifetime of the collator
    *  @param length size of the image. If negative, the API will try to
    *                figure out the length of the image
    *  @param base Base collator, for lookup of untailored characters.
    *              Must be the root collator, must not be NULL.
    *              The base is required to be present through the lifetime of the collator.
    *  @param status for catching errors
    *  @return newly created collator
    *  @see cloneBinary
    *  @stable ICU 3.4
    */
    RuleBasedCollator(const uint8_t *bin, int32_t length,
                    const RuleBasedCollator *base,
                    UErrorCode &status);

    /**
     * Destructor.
     * @stable ICU 2.0
     */
    virtual ~RuleBasedCollator();

    /**
     * Assignment operator.
     * @param other other RuleBasedCollator object to copy from.
     * @stable ICU 2.0
     */
    RuleBasedCollator& operator=(const RuleBasedCollator& other);

    /**
     * Returns true if argument is the same as this object.
     * @param other Collator object to be compared.
     * @return true if arguments is the same as this object.
     * @stable ICU 2.0
     */
    virtual UBool operator==(const Collator& other) const;

    /**
     * Makes a copy of this object.
     * @return a copy of this object, owned by the caller
     * @stable ICU 2.0
     */
    virtual RuleBasedCollator* clone() const;

    /**
     * Creates a collation element iterator for the source string. The caller of
     * this method is responsible for the memory management of the return
     * pointer.
     * @param source the string over which the CollationElementIterator will
     *        iterate.
     * @return the collation element iterator of the source string using this as
     *         the based Collator.
     * @stable ICU 2.2
     */
    virtual CollationElementIterator* createCollationElementIterator(
                                           const UnicodeString& source) const;

    /**
     * Creates a collation element iterator for the source. The caller of this
     * method is responsible for the memory management of the returned pointer.
     * @param source the CharacterIterator which produces the characters over
     *        which the CollationElementItgerator will iterate.
     * @return the collation element iterator of the source using this as the
     *         based Collator.
     * @stable ICU 2.2
     */
    virtual CollationElementIterator* createCollationElementIterator(
                                         const CharacterIterator& source) const;

    // Make deprecated versions of Collator::compare() visible.
    using Collator::compare;

    /**
    * The comparison function compares the character data stored in two
    * different strings. Returns information about whether a string is less
    * than, greater than or equal to another string.
    * @param source the source string to be compared with.
    * @param target the string that is to be compared with the source string.
    * @param status possible error code
    * @return Returns an enum value. UCOL_GREATER if source is greater
    * than target; UCOL_EQUAL if source is equal to target; UCOL_LESS if source is less
    * than target
    * @stable ICU 2.6
    **/
    virtual UCollationResult compare(const UnicodeString& source,
                                     const UnicodeString& target,
                                     UErrorCode &status) const;

    /**
    * Does the same thing as compare but limits the comparison to a specified
    * length
    * @param source the source string to be compared with.
    * @param target the string that is to be compared with the source string.
    * @param length the length the comparison is limited to
    * @param status possible error code
    * @return Returns an enum value. UCOL_GREATER if source (up to the specified
    *         length) is greater than target; UCOL_EQUAL if source (up to specified
    *         length) is equal to target; UCOL_LESS if source (up to the specified
    *         length) is less  than target.
    * @stable ICU 2.6
    */
    virtual UCollationResult compare(const UnicodeString& source,
                                     const UnicodeString& target,
                                     int32_t length,
                                     UErrorCode &status) const;

    /**
    * The comparison function compares the character data stored in two
    * different string arrays. Returns information about whether a string array
    * is less than, greater than or equal to another string array.
    * @param source the source string array to be compared with.
    * @param sourceLength the length of the source string array.  If this value
    *        is equal to -1, the string array is null-terminated.
    * @param target the string that is to be compared with the source string.
    * @param targetLength the length of the target string array.  If this value
    *        is equal to -1, the string array is null-terminated.
    * @param status possible error code
    * @return Returns an enum value. UCOL_GREATER if source is greater
    * than target; UCOL_EQUAL if source is equal to target; UCOL_LESS if source is less
    * than target
    * @stable ICU 2.6
    */
    virtual UCollationResult compare(const char16_t* source, int32_t sourceLength,
                                     const char16_t* target, int32_t targetLength,
                                     UErrorCode &status) const;

    /**
     * Compares two strings using the Collator.
     * Returns whether the first one compares less than/equal to/greater than
     * the second one.
     * This version takes UCharIterator input.
     * @param sIter the first ("source") string iterator
     * @param tIter the second ("target") string iterator
     * @param status ICU status
     * @return UCOL_LESS, UCOL_EQUAL or UCOL_GREATER
     * @stable ICU 4.2
     */
    virtual UCollationResult compare(UCharIterator &sIter,
                                     UCharIterator &tIter,
                                     UErrorCode &status) const;

    /**
     * Compares two UTF-8 strings using the Collator.
     * Returns whether the first one compares less than/equal to/greater than
     * the second one.
     * This version takes UTF-8 input.
     * Note that a StringPiece can be implicitly constructed
     * from a std::string or a NUL-terminated const char * string.
     * @param source the first UTF-8 string
     * @param target the second UTF-8 string
     * @param status ICU status
     * @return UCOL_LESS, UCOL_EQUAL or UCOL_GREATER
     * @stable ICU 51
     */
    virtual UCollationResult compareUTF8(const StringPiece &source,
                                         const StringPiece &target,
                                         UErrorCode &status) const;

    /**
     * Transforms the string into a series of characters
     * that can be compared with CollationKey.compare().
     *
     * Note that sort keys are often less efficient than simply doing comparison.
     * For more details, see the ICU User Guide.
     *
     * @param source the source string.
     * @param key the transformed key of the source string.
     * @param status the error code status.
     * @return the transformed key.
     * @see CollationKey
     * @stable ICU 2.0
     */
    virtual CollationKey& getCollationKey(const UnicodeString& source,
                                          CollationKey& key,
                                          UErrorCode& status) const;

    /**
     * Transforms a specified region of the string into a series of characters
     * that can be compared with CollationKey.compare.
     *
     * Note that sort keys are often less efficient than simply doing comparison.
     * For more details, see the ICU User Guide.
     *
     * @param source the source string.
     * @param sourceLength the length of the source string.
     * @param key the transformed key of the source string.
     * @param status the error code status.
     * @return the transformed key.
     * @see CollationKey
     * @stable ICU 2.0
     */
    virtual CollationKey& getCollationKey(const char16_t *source,
                                          int32_t sourceLength,
                                          CollationKey& key,
                                          UErrorCode& status) const;

    /**
     * Generates the hash code for the rule-based collation object.
     * @return the hash code.
     * @stable ICU 2.0
     */
    virtual int32_t hashCode() const;

#ifndef U_FORCE_HIDE_DEPRECATED_API
    /**
    * Gets the locale of the Collator
    * @param type can be either requested, valid or actual locale. For more
    *             information see the definition of ULocDataLocaleType in
    *             uloc.h
    * @param status the error code status.
    * @return locale where the collation data lives. If the collator
    *         was instantiated from rules, locale is empty.
    * @deprecated ICU 2.8 likely to change in ICU 3.0, based on feedback
    */
    virtual Locale getLocale(ULocDataLocaleType type, UErrorCode& status) const;
#endif  // U_FORCE_HIDE_DEPRECATED_API

    /**
     * Gets the tailoring rules for this collator.
     * @return the collation tailoring from which this collator was created
     * @stable ICU 2.0
     */
    const UnicodeString& getRules() const;

    /**
     * Gets the version information for a Collator.
     * @param info the version # information, the result will be filled in
     * @stable ICU 2.0
     */
    virtual void getVersion(UVersionInfo info) const;

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Returns the maximum length of any expansion sequences that end with the
     * specified comparison order.
     *
     * This is specific to the kind of collation element values and sequences
     * returned by the CollationElementIterator.
     * Call CollationElementIterator::getMaxExpansion() instead.
     *
     * @param order a collation order returned by CollationElementIterator::previous
     *              or CollationElementIterator::next.
     * @return maximum size of the expansion sequences ending with the collation
     *         element, or 1 if the collation element does not occur at the end of
     *         any expansion sequence
     * @see CollationElementIterator#getMaxExpansion
     * @deprecated ICU 51 Use CollationElementIterator::getMaxExpansion() instead.
     */
    int32_t getMaxExpansion(int32_t order) const;
#endif  /* U_HIDE_DEPRECATED_API */

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     * @return The class ID for this object. All objects of a given class have
     *         the same class ID. Objects of other classes have different class
     *         IDs.
     * @stable ICU 2.0
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Returns the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * Base* polymorphic_pointer = createPolymorphicObject();
     * if (polymorphic_pointer->getDynamicClassID() ==
     *                                          Derived::getStaticClassID()) ...
     * </pre>
     * @return The class ID for all objects of this class.
     * @stable ICU 2.0
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Do not use this method: The caller and the ICU library might use different heaps.
     * Use cloneBinary() instead which writes to caller-provided memory.
     *
     * Returns a binary format of this collator.
     * @param length Returns the length of the data, in bytes
     * @param status the error code status.
     * @return memory, owned by the caller, of size 'length' bytes.
     * @deprecated ICU 52. Use cloneBinary() instead.
     */
    uint8_t *cloneRuleData(int32_t &length, UErrorCode &status) const;
#endif  /* U_HIDE_DEPRECATED_API */

    /** Creates a binary image of a collator. This binary image can be stored and
    *  later used to instantiate a collator using ucol_openBinary.
    *  This API supports preflighting.
    *  @param buffer a fill-in buffer to receive the binary image
    *  @param capacity capacity of the destination buffer
    *  @param status for catching errors
    *  @return size of the image
    *  @see ucol_openBinary
    *  @stable ICU 3.4
    */
    int32_t cloneBinary(uint8_t *buffer, int32_t capacity, UErrorCode &status) const;

    /**
     * Returns current rules. Delta defines whether full rules are returned or
     * just the tailoring.
     *
     * getRules(void) should normally be used instead.
     * See http://userguide.icu-project.org/collation/customization#TOC-Building-on-Existing-Locales
     * @param delta one of UCOL_TAILORING_ONLY, UCOL_FULL_RULES.
     * @param buffer UnicodeString to store the result rules
     * @stable ICU 2.2
     * @see UCOL_FULL_RULES
     */
    void getRules(UColRuleOption delta, UnicodeString &buffer) const;

    /**
     * Universal attribute setter
     * @param attr attribute type
     * @param value attribute value
     * @param status to indicate whether the operation went on smoothly or there were errors
     * @stable ICU 2.2
     */
    virtual void setAttribute(UColAttribute attr, UColAttributeValue value,
                              UErrorCode &status);

    /**
     * Universal attribute getter.
     * @param attr attribute type
     * @param status to indicate whether the operation went on smoothly or there were errors
     * @return attribute value
     * @stable ICU 2.2
     */
    virtual UColAttributeValue getAttribute(UColAttribute attr,
                                            UErrorCode &status) const;

    /**
     * Sets the variable top to the top of the specified reordering group.
     * The variable top determines the highest-sorting character
     * which is affected by UCOL_ALTERNATE_HANDLING.
     * If that attribute is set to UCOL_NON_IGNORABLE, then the variable top has no effect.
     * @param group one of UCOL_REORDER_CODE_SPACE, UCOL_REORDER_CODE_PUNCTUATION,
     *              UCOL_REORDER_CODE_SYMBOL, UCOL_REORDER_CODE_CURRENCY;
     *              or UCOL_REORDER_CODE_DEFAULT to restore the default max variable group
     * @param errorCode Standard ICU error code. Its input value must
     *                  pass the U_SUCCESS() test, or else the function returns
     *                  immediately. Check for U_FAILURE() on output or use with
     *                  function chaining. (See User Guide for details.)
     * @return *this
     * @see getMaxVariable
     * @stable ICU 53
     */
    virtual Collator &setMaxVariable(UColReorderCode group, UErrorCode &errorCode);

    /**
     * Returns the maximum reordering group whose characters are affected by UCOL_ALTERNATE_HANDLING.
     * @return the maximum variable reordering group.
     * @see setMaxVariable
     * @stable ICU 53
     */
    virtual UColReorderCode getMaxVariable() const;

#ifndef U_FORCE_HIDE_DEPRECATED_API
    /**
     * Sets the variable top to the primary weight of the specified string.
     *
     * Beginning with ICU 53, the variable top is pinned to
     * the top of one of the supported reordering groups,
     * and it must not be beyond the last of those groups.
     * See setMaxVariable().
     * @param varTop one or more (if contraction) char16_ts to which the variable top should be set
     * @param len length of variable top string. If -1 it is considered to be zero terminated.
     * @param status error code. If error code is set, the return value is undefined. Errors set by this function are: <br>
     *    U_CE_NOT_FOUND_ERROR if more than one character was passed and there is no such contraction<br>
     *    U_ILLEGAL_ARGUMENT_ERROR if the variable top is beyond
     *    the last reordering group supported by setMaxVariable()
     * @return variable top primary weight
     * @deprecated ICU 53 Call setMaxVariable() instead.
     */
    virtual uint32_t setVariableTop(const char16_t *varTop, int32_t len, UErrorCode &status);

    /**
     * Sets the variable top to the primary weight of the specified string.
     *
     * Beginning with ICU 53, the variable top is pinned to
     * the top of one of the supported reordering groups,
     * and it must not be beyond the last of those groups.
     * See setMaxVariable().
     * @param varTop a UnicodeString size 1 or more (if contraction) of char16_ts to which the variable top should be set
     * @param status error code. If error code is set, the return value is undefined. Errors set by this function are: <br>
     *    U_CE_NOT_FOUND_ERROR if more than one character was passed and there is no such contraction<br>
     *    U_ILLEGAL_ARGUMENT_ERROR if the variable top is beyond
     *    the last reordering group supported by setMaxVariable()
     * @return variable top primary weight
     * @deprecated ICU 53 Call setMaxVariable() instead.
     */
    virtual uint32_t setVariableTop(const UnicodeString &varTop, UErrorCode &status);

    /**
     * Sets the variable top to the specified primary weight.
     *
     * Beginning with ICU 53, the variable top is pinned to
     * the top of one of the supported reordering groups,
     * and it must not be beyond the last of those groups.
     * See setMaxVariable().
     * @param varTop primary weight, as returned by setVariableTop or ucol_getVariableTop
     * @param status error code
     * @deprecated ICU 53 Call setMaxVariable() instead.
     */
    virtual void setVariableTop(uint32_t varTop, UErrorCode &status);
#endif  // U_FORCE_HIDE_DEPRECATED_API

    /**
     * Gets the variable top value of a Collator.
     * @param status error code (not changed by function). If error code is set, the return value is undefined.
     * @return the variable top primary weight
     * @see getMaxVariable
     * @stable ICU 2.0
     */
    virtual uint32_t getVariableTop(UErrorCode &status) const;

    /**
     * Get a UnicodeSet that contains all the characters and sequences tailored in
     * this collator.
     * @param status      error code of the operation
     * @return a pointer to a UnicodeSet object containing all the
     *         code points and sequences that may sort differently than
     *         in the root collator. The object must be disposed of by using delete
     * @stable ICU 2.4
     */
    virtual UnicodeSet *getTailoredSet(UErrorCode &status) const;

    /**
     * Get the sort key as an array of bytes from a UnicodeString.
     *
     * Note that sort keys are often less efficient than simply doing comparison.
     * For more details, see the ICU User Guide.
     *
     * @param source string to be processed.
     * @param result buffer to store result in. If NULL, number of bytes needed
     *        will be returned.
     * @param resultLength length of the result buffer. If if not enough the
     *        buffer will be filled to capacity.
     * @return Number of bytes needed for storing the sort key
     * @stable ICU 2.0
     */
    virtual int32_t getSortKey(const UnicodeString& source, uint8_t *result,
                               int32_t resultLength) const;

    /**
     * Get the sort key as an array of bytes from a char16_t buffer.
     *
     * Note that sort keys are often less efficient than simply doing comparison.
     * For more details, see the ICU User Guide.
     *
     * @param source string to be processed.
     * @param sourceLength length of string to be processed. If -1, the string
     *        is 0 terminated and length will be decided by the function.
     * @param result buffer to store result in. If NULL, number of bytes needed
     *        will be returned.
     * @param resultLength length of the result buffer. If if not enough the
     *        buffer will be filled to capacity.
     * @return Number of bytes needed for storing the sort key
     * @stable ICU 2.2
     */
    virtual int32_t getSortKey(const char16_t *source, int32_t sourceLength,
                               uint8_t *result, int32_t resultLength) const;

    /**
     * Retrieves the reordering codes for this collator.
     * @param dest The array to fill with the script ordering.
     * @param destCapacity The length of dest. If it is 0, then dest may be NULL and the function
     *  will only return the length of the result without writing any codes (pre-flighting).
     * @param status A reference to an error code value, which must not indicate
     * a failure before the function call.
     * @return The length of the script ordering array.
     * @see ucol_setReorderCodes
     * @see Collator#getEquivalentReorderCodes
     * @see Collator#setReorderCodes
     * @stable ICU 4.8
     */
     virtual int32_t getReorderCodes(int32_t *dest,
                                     int32_t destCapacity,
                                     UErrorCode& status) const;

    /**
     * Sets the ordering of scripts for this collator.
     * @param reorderCodes An array of script codes in the new order. This can be NULL if the
     * length is also set to 0. An empty array will clear any reordering codes on the collator.
     * @param reorderCodesLength The length of reorderCodes.
     * @param status error code
     * @see ucol_setReorderCodes
     * @see Collator#getReorderCodes
     * @see Collator#getEquivalentReorderCodes
     * @stable ICU 4.8
     */
     virtual void setReorderCodes(const int32_t* reorderCodes,
                                  int32_t reorderCodesLength,
                                  UErrorCode& status) ;

    /**
     * Implements ucol_strcollUTF8().
     * @internal
     */
    virtual UCollationResult internalCompareUTF8(
            const char *left, int32_t leftLength,
            const char *right, int32_t rightLength,
            UErrorCode &errorCode) const;

    /** Get the short definition string for a collator. This internal API harvests the collator's
     *  locale and the attribute set and produces a string that can be used for opening
     *  a collator with the same attributes using the ucol_openFromShortString API.
     *  This string will be normalized.
     *  The structure and the syntax of the string is defined in the "Naming collators"
     *  section of the users guide:
     *  http://userguide.icu-project.org/collation/concepts#TOC-Collator-naming-scheme
     *  This function supports preflighting.
     *
     *  This is internal, and intended to be used with delegate converters.
     *
     *  @param locale a locale that will appear as a collators locale in the resulting
     *                short string definition. If NULL, the locale will be harvested
     *                from the collator.
     *  @param buffer space to hold the resulting string
     *  @param capacity capacity of the buffer
     *  @param status for returning errors. All the preflighting errors are featured
     *  @return length of the resulting string
     *  @see ucol_openFromShortString
     *  @see ucol_normalizeShortDefinitionString
     *  @see ucol_getShortDefinitionString
     *  @internal
     */
    virtual int32_t internalGetShortDefinitionString(const char *locale,
                                                     char *buffer,
                                                     int32_t capacity,
                                                     UErrorCode &status) const;

    /**
     * Implements ucol_nextSortKeyPart().
     * @internal
     */
    virtual int32_t internalNextSortKeyPart(
            UCharIterator *iter, uint32_t state[2],
            uint8_t *dest, int32_t count, UErrorCode &errorCode) const;

    // Do not enclose the default constructor with #ifndef U_HIDE_INTERNAL_API
    /**
     * Only for use in ucol_openRules().
     * @internal
     */
    RuleBasedCollator();

#ifndef U_HIDE_INTERNAL_API
    /**
     * Implements ucol_getLocaleByType().
     * Needed because the lifetime of the locale ID string must match that of the collator.
     * getLocale() returns a copy of a Locale, with minimal lifetime in a C wrapper.
     * @internal
     */
    const char *internalGetLocaleID(ULocDataLocaleType type, UErrorCode &errorCode) const;

    /**
     * Implements ucol_getContractionsAndExpansions().
     * Gets this collator's sets of contraction strings and/or
     * characters and strings that map to multiple collation elements (expansions).
     * If addPrefixes is TRUE, then contractions that are expressed as
     * prefix/pre-context rules are included.
     * @param contractions if not NULL, the set to hold the contractions
     * @param expansions if not NULL, the set to hold the expansions
     * @param addPrefixes include prefix contextual mappings
     * @param errorCode in/out ICU error code
     * @internal
     */
    void internalGetContractionsAndExpansions(
            UnicodeSet *contractions, UnicodeSet *expansions,
            UBool addPrefixes, UErrorCode &errorCode) const;

    /**
     * Adds the contractions that start with character c to the set.
     * Ignores prefixes. Used by AlphabeticIndex.
     * @internal
     */
    void internalAddContractions(UChar32 c, UnicodeSet &set, UErrorCode &errorCode) const;

    /**
     * Implements from-rule constructors, and ucol_openRules().
     * @internal
     */
    void internalBuildTailoring(
            const UnicodeString &rules,
            int32_t strength,
            UColAttributeValue decompositionMode,
            UParseError *outParseError, UnicodeString *outReason,
            UErrorCode &errorCode);

    /** @internal */
    static inline RuleBasedCollator *rbcFromUCollator(UCollator *uc) {
        return dynamic_cast<RuleBasedCollator *>(fromUCollator(uc));
    }
    /** @internal */
    static inline const RuleBasedCollator *rbcFromUCollator(const UCollator *uc) {
        return dynamic_cast<const RuleBasedCollator *>(fromUCollator(uc));
    }

    /**
     * Appends the CEs for the string to the vector.
     * @internal for tests & tools
     */
    void internalGetCEs(const UnicodeString &str, UVector64 &ces, UErrorCode &errorCode) const;
#endif  // U_HIDE_INTERNAL_API

protected:
   /**
    * Used internally by registration to define the requested and valid locales.
    * @param requestedLocale the requested locale
    * @param validLocale the valid locale
    * @param actualLocale the actual locale
    * @internal
    */
    virtual void setLocales(const Locale& requestedLocale, const Locale& validLocale, const Locale& actualLocale);

private:
    friend class CollationElementIterator;
    friend class Collator;

    RuleBasedCollator(const CollationCacheEntry *entry);

    /**
     * Enumeration of attributes that are relevant for short definition strings
     * (e.g., ucol_getShortDefinitionString()).
     * Effectively extends UColAttribute.
     */
    enum Attributes {
        ATTR_VARIABLE_TOP = UCOL_ATTRIBUTE_COUNT,
        ATTR_LIMIT
    };

    void adoptTailoring(CollationTailoring *t, UErrorCode &errorCode);

    // Both lengths must be <0 or else both must be >=0.
    UCollationResult doCompare(const char16_t *left, int32_t leftLength,
                               const char16_t *right, int32_t rightLength,
                               UErrorCode &errorCode) const;
    UCollationResult doCompare(const uint8_t *left, int32_t leftLength,
                               const uint8_t *right, int32_t rightLength,
                               UErrorCode &errorCode) const;

    void writeSortKey(const char16_t *s, int32_t length,
                      SortKeyByteSink &sink, UErrorCode &errorCode) const;

    void writeIdenticalLevel(const char16_t *s, const char16_t *limit,
                             SortKeyByteSink &sink, UErrorCode &errorCode) const;

    const CollationSettings &getDefaultSettings() const;

    void setAttributeDefault(int32_t attribute) {
        explicitlySetAttributes &= ~((uint32_t)1 << attribute);
    }
    void setAttributeExplicitly(int32_t attribute) {
        explicitlySetAttributes |= (uint32_t)1 << attribute;
    }
    UBool attributeHasBeenSetExplicitly(int32_t attribute) const {
        // assert(0 <= attribute < ATTR_LIMIT);
        return (UBool)((explicitlySetAttributes & ((uint32_t)1 << attribute)) != 0);
    }

    /**
     * Tests whether a character is "unsafe" for use as a collation starting point.
     *
     * @param c code point or code unit
     * @return TRUE if c is unsafe
     * @see CollationElementIterator#setOffset(int)
     */
    UBool isUnsafe(UChar32 c) const;

    static void U_CALLCONV computeMaxExpansions(const CollationTailoring *t, UErrorCode &errorCode);
    UBool initMaxExpansions(UErrorCode &errorCode) const;

    void setFastLatinOptions(CollationSettings &ownedSettings) const;

    const CollationData *data;
    const CollationSettings *settings;  // reference-counted
    const CollationTailoring *tailoring;  // alias of cacheEntry->tailoring
    const CollationCacheEntry *cacheEntry;  // reference-counted
    Locale validLocale;
    uint32_t explicitlySetAttributes;

    UBool actualLocaleIsSameAsValid;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION

#endif /* U_SHOW_CPLUSPLUS_API */



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*                                                                            *
* Copyright (C) 2003-2015, International Business Machines                   *
*                Corporation and others. All Rights Reserved.                *
*                                                                            *
******************************************************************************
*   file name:  ulocdata.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2003Oct21
*   created by: Ram Viswanadha
*/







/**
 * \file
 * \brief C API: Provides access to locale data.
 */

/** Forward declaration of the ULocaleData structure. @stable ICU 3.6 */
struct ULocaleData;

/** A locale data object. @stable ICU 3.6 */
typedef struct ULocaleData ULocaleData;



/** The possible types of exemplar character sets.
  * @stable ICU 3.4
  */
typedef enum ULocaleDataExemplarSetType  {
    /** Basic set @stable ICU 3.4 */
    ULOCDATA_ES_STANDARD=0,
    /** Auxiliary set @stable ICU 3.4 */
    ULOCDATA_ES_AUXILIARY=1,
    /** Index Character set @stable ICU 4.8 */
    ULOCDATA_ES_INDEX=2,
    /** Punctuation set @stable ICU 51 */
    ULOCDATA_ES_PUNCTUATION=3,
#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal ULocaleDataExemplarSetType value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    ULOCDATA_ES_COUNT=4
#endif  /* U_HIDE_DEPRECATED_API */
} ULocaleDataExemplarSetType;

/** The possible types of delimiters.
  * @stable ICU 3.4
  */
typedef enum ULocaleDataDelimiterType {
    /** Quotation start @stable ICU 3.4 */
    ULOCDATA_QUOTATION_START = 0,
    /** Quotation end @stable ICU 3.4 */
    ULOCDATA_QUOTATION_END = 1,
    /** Alternate quotation start @stable ICU 3.4 */
    ULOCDATA_ALT_QUOTATION_START = 2,
    /** Alternate quotation end @stable ICU 3.4 */
    ULOCDATA_ALT_QUOTATION_END = 3,
#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal ULocaleDataDelimiterType value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    ULOCDATA_DELIMITER_COUNT = 4
#endif  /* U_HIDE_DEPRECATED_API */
} ULocaleDataDelimiterType;

/**
 * Opens a locale data object for the given locale
 *
 * @param localeID  Specifies the locale associated with this locale
 *                  data object.
 * @param status    Pointer to error status code.
 * @stable ICU 3.4
 */
U_STABLE ULocaleData* U_EXPORT2
ulocdata_open(const char *localeID, UErrorCode *status);

/**
 * Closes a locale data object.
 *
 * @param uld       The locale data object to close
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
ulocdata_close(ULocaleData *uld);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalULocaleDataPointer
 * "Smart pointer" class, closes a ULocaleData via ulocdata_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalULocaleDataPointer, ULocaleData, ulocdata_close);

U_NAMESPACE_END

#endif

/**
 * Sets the "no Substitute" attribute of the locale data
 * object.  If true, then any methods associated with the
 * locale data object will return null when there is no
 * data available for that method, given the locale ID
 * supplied to ulocdata_open().
 *
 * @param uld       The locale data object to set.
 * @param setting   Value of the "no substitute" attribute.
 * @stable ICU 3.4
 */
U_STABLE void U_EXPORT2
ulocdata_setNoSubstitute(ULocaleData *uld, UBool setting);

/**
 * Retrieves the current "no Substitute" value of the locale data
 * object.  If true, then any methods associated with the
 * locale data object will return null when there is no
 * data available for that method, given the locale ID
 * supplied to ulocdata_open().
 *
 * @param uld       Pointer to the The locale data object to set.
 * @return UBool    Value of the "no substitute" attribute.
 * @stable ICU 3.4
 */
U_STABLE UBool U_EXPORT2
ulocdata_getNoSubstitute(ULocaleData *uld);

/**
 * Returns the set of exemplar characters for a locale.
 *
 * @param uld       Pointer to the locale data object from which the
 *                  exemplar character set is to be retrieved.
 * @param fillIn    Pointer to a USet object to receive the
 *                  exemplar character set for the given locale.  Previous
 *                  contents of fillIn are lost.  <em>If fillIn is NULL,
 *                  then a new USet is created and returned.  The caller
 *                  owns the result and must dispose of it by calling
 *                  uset_close.</em>
 * @param options   Bitmask for options to apply to the exemplar pattern.
 *                  Specify zero to retrieve the exemplar set as it is
 *                  defined in the locale data.  Specify
 *                  USET_CASE_INSENSITIVE to retrieve a case-folded
 *                  exemplar set.  See uset_applyPattern for a complete
 *                  list of valid options.  The USET_IGNORE_SPACE bit is
 *                  always set, regardless of the value of 'options'.
 * @param extype    Specifies the type of exemplar set to be retrieved.
 * @param status    Pointer to an input-output error code value;
 *                  must not be NULL.  Will be set to U_MISSING_RESOURCE_ERROR
 *                  if the requested data is not available.
 * @return USet*    Either fillIn, or if fillIn is NULL, a pointer to
 *                  a newly-allocated USet that the user must close.
 *                  In case of error, NULL is returned.
 * @stable ICU 3.4
 */
U_STABLE USet* U_EXPORT2
ulocdata_getExemplarSet(ULocaleData *uld, USet *fillIn,
                        uint32_t options, ULocaleDataExemplarSetType extype, UErrorCode *status);

/**
 * Returns one of the delimiter strings associated with a locale.
 *
 * @param uld           Pointer to the locale data object from which the
 *                      delimiter string is to be retrieved.
 * @param type          the type of delimiter to be retrieved.
 * @param result        A pointer to a buffer to receive the result.
 * @param resultLength  The maximum size of result.
 * @param status        Pointer to an error code value
 * @return int32_t      The total buffer size needed; if greater than resultLength,
 *                      the output was truncated.
 * @stable ICU 3.4
 */
U_STABLE int32_t U_EXPORT2
ulocdata_getDelimiter(ULocaleData *uld, ULocaleDataDelimiterType type, UChar *result, int32_t resultLength, UErrorCode *status);

/**
 * Enumeration for representing the measurement systems.
 * @stable ICU 2.8
 */
typedef enum UMeasurementSystem {
    UMS_SI,     /**< Measurement system specified by SI otherwise known as Metric system. @stable ICU 2.8 */
    UMS_US,     /**< Measurement system followed in the United States of America. @stable ICU 2.8 */
    UMS_UK,     /**< Mix of metric and imperial units used in Great Britain. @stable ICU 55 */
#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UMeasurementSystem value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UMS_LIMIT
#endif  /* U_HIDE_DEPRECATED_API */
} UMeasurementSystem;

/**
 * Returns the measurement system used in the locale specified by the localeID.
 * Please note that this API will change in ICU 3.6 and will use an ulocdata object.
 *
 * @param localeID      The id of the locale for which the measurement system to be retrieved.
 * @param status        Must be a valid pointer to an error code value,
 *                      which must not indicate a failure before the function call.
 * @return UMeasurementSystem the measurement system used in the locale.
 * @stable ICU 2.8
 */
U_STABLE UMeasurementSystem U_EXPORT2
ulocdata_getMeasurementSystem(const char *localeID, UErrorCode *status);

/**
 * Returns the element gives the normal business letter size, and customary units.
 * The units for the numbers are always in <em>milli-meters</em>.
 * For US since 8.5 and 11 do not yeild an integral value when converted to milli-meters,
 * the values are rounded off.
 * So for A4 size paper the height and width are 297 mm and 210 mm repectively,
 * and for US letter size the height and width are 279 mm and 216 mm respectively.
 * Please note that this API will change in ICU 3.6 and will use an ulocdata object.
 *
 * @param localeID      The id of the locale for which the paper size information to be retrieved.
 * @param height        A pointer to int to recieve the height information.
 * @param width         A pointer to int to recieve the width information.
 * @param status        Must be a valid pointer to an error code value,
 *                      which must not indicate a failure before the function call.
 * @stable ICU 2.8
 */
U_STABLE void U_EXPORT2
ulocdata_getPaperSize(const char *localeID, int32_t *height, int32_t *width, UErrorCode *status);

/**
 * Return the current CLDR version used by the library.
 * @param versionArray fillin that will recieve the version number
 * @param status error code - could be U_MISSING_RESOURCE_ERROR if the version was not found.
 * @stable ICU 4.2
 */
U_STABLE void U_EXPORT2
ulocdata_getCLDRVersion(UVersionInfo versionArray, UErrorCode *status);

/**
 * Returns locale display pattern associated with a locale.
 *
 * @param uld       Pointer to the locale data object from which the
 *                  exemplar character set is to be retrieved.
 * @param pattern   locale display pattern for locale.
 * @param patternCapacity the size of the buffer to store the locale display
 *                  pattern with.
 * @param status    Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return the actual buffer size needed for localeDisplayPattern.  If it's greater
 * than patternCapacity, the returned pattern will be truncated.
 *
 * @stable ICU 4.2
 */
U_STABLE int32_t U_EXPORT2
ulocdata_getLocaleDisplayPattern(ULocaleData *uld,
                                 UChar *pattern,
                                 int32_t patternCapacity,
                                 UErrorCode *status);


/**
 * Returns locale separator associated with a locale.
 *
 * @param uld       Pointer to the locale data object from which the
 *                  exemplar character set is to be retrieved.
 * @param separator locale separator for locale.
 * @param separatorCapacity the size of the buffer to store the locale
 *                  separator with.
 * @param status    Must be a valid pointer to an error code value,
 *                  which must not indicate a failure before the function call.
 * @return the actual buffer size needed for localeSeparator.  If it's greater
 * than separatorCapacity, the returned separator will be truncated.
 *
 * @stable ICU 4.2
 */
U_STABLE int32_t U_EXPORT2
ulocdata_getLocaleSeparator(ULocaleData *uld,
                            UChar *separator,
                            int32_t separatorCapacity,
                            UErrorCode *status);












//#include <string>
//#include <iostream>

U_NAMESPACE_BEGIN

namespace {

/**
 * Prefix string for Chinese index buckets.
 * See http://unicode.org/repos/cldr/trunk/specs/ldml/tr35-collation.html#Collation_Indexes
 */
const UChar BASE[1] = { 0xFDD0 };
const int32_t BASE_LENGTH = 1;

UBool isOneLabelBetterThanOther(const Normalizer2 &nfkdNormalizer,
                                const UnicodeString &one, const UnicodeString &other);

}  // namespace

static int32_t U_CALLCONV
collatorComparator(const void *context, const void *left, const void *right);

static int32_t U_CALLCONV
recordCompareFn(const void *context, const void *left, const void *right);

//  UVector<Record *> support function, delete a Record.
static void U_CALLCONV
alphaIndex_deleteRecord(void *obj) {
    delete static_cast<AlphabeticIndex::Record *>(obj);
}

namespace {

UnicodeString *ownedString(const UnicodeString &s, LocalPointer<UnicodeString> &owned,
                           UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return NULL; }
    if (owned.isValid()) {
        return owned.orphan();
    }
    UnicodeString *p = new UnicodeString(s);
    if (p == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
    }
    return p;
}

inline UnicodeString *getString(const UVector &list, int32_t i) {
    return static_cast<UnicodeString *>(list[i]);
}

inline AlphabeticIndex::Bucket *getBucket(const UVector &list, int32_t i) {
    return static_cast<AlphabeticIndex::Bucket *>(list[i]);
}

inline AlphabeticIndex::Record *getRecord(const UVector &list, int32_t i) {
    return static_cast<AlphabeticIndex::Record *>(list[i]);
}

/**
 * Like Java Collections.binarySearch(List, String, Comparator).
 *
 * @return the index>=0 where the item was found,
 *         or the index<0 for inserting the string at ~index in sorted order
 */
int32_t binarySearch(const UVector &list, const UnicodeString &s, const Collator &coll) {
    if (list.size() == 0) { return ~0; }
    int32_t start = 0;
    int32_t limit = list.size();
    for (;;) {
        int32_t i = (start + limit) / 2;
        const UnicodeString *si = static_cast<UnicodeString *>(list.elementAt(i));
        UErrorCode errorCode = U_ZERO_ERROR;
        UCollationResult cmp = coll.compare(s, *si, errorCode);
        if (cmp == UCOL_EQUAL) {
            return i;
        } else if (cmp < 0) {
            if (i == start) {
                return ~start;  // insert s before *si
            }
            limit = i;
        } else {
            if (i == start) {
                return ~(start + 1);  // insert s after *si
            }
            start = i;
        }
    }
}

}  // namespace

// The BucketList is not in the anonymous namespace because only Clang
// seems to support its use in other classes from there.
// However, we also don't need U_I18N_API because it is not used from outside the i18n library.
class BucketList : public UObject {
public:
    BucketList(UVector *bucketList, UVector *publicBucketList)
            : bucketList_(bucketList), immutableVisibleList_(publicBucketList) {
        int32_t displayIndex = 0;
        for (int32_t i = 0; i < publicBucketList->size(); ++i) {
            getBucket(*publicBucketList, i)->displayIndex_ = displayIndex++;
        }
    }

    // The virtual destructor must not be inline.
    // See ticket #8454 for details.
    virtual ~BucketList();

    int32_t getBucketCount() const {
        return immutableVisibleList_->size();
    }

    int32_t getBucketIndex(const UnicodeString &name, const Collator &collatorPrimaryOnly,
                           UErrorCode &errorCode) {
        // binary search
        int32_t start = 0;
        int32_t limit = bucketList_->size();
        while ((start + 1) < limit) {
            int32_t i = (start + limit) / 2;
            const AlphabeticIndex::Bucket *bucket = getBucket(*bucketList_, i);
            UCollationResult nameVsBucket =
                collatorPrimaryOnly.compare(name, bucket->lowerBoundary_, errorCode);
            if (nameVsBucket < 0) {
                limit = i;
            } else {
                start = i;
            }
        }
        const AlphabeticIndex::Bucket *bucket = getBucket(*bucketList_, start);
        if (bucket->displayBucket_ != NULL) {
            bucket = bucket->displayBucket_;
        }
        return bucket->displayIndex_;
    }

    /** All of the buckets, visible and invisible. */
    UVector *bucketList_;
    /** Just the visible buckets. */
    UVector *immutableVisibleList_;
};

BucketList::~BucketList() {
    delete bucketList_;
    if (immutableVisibleList_ != bucketList_) {
        delete immutableVisibleList_;
    }
}

AlphabeticIndex::ImmutableIndex::~ImmutableIndex() {
    delete buckets_;
    delete collatorPrimaryOnly_;
}

int32_t
AlphabeticIndex::ImmutableIndex::getBucketCount() const {
    return buckets_->getBucketCount();
}

int32_t
AlphabeticIndex::ImmutableIndex::getBucketIndex(
        const UnicodeString &name, UErrorCode &errorCode) const {
    return buckets_->getBucketIndex(name, *collatorPrimaryOnly_, errorCode);
}

const AlphabeticIndex::Bucket *
AlphabeticIndex::ImmutableIndex::getBucket(int32_t index) const {
    if (0 <= index && index < buckets_->getBucketCount()) {
        return icu::getBucket(*buckets_->immutableVisibleList_, index);
    } else {
        return NULL;
    }
}

AlphabeticIndex::AlphabeticIndex(const Locale &locale, UErrorCode &status)
        : inputList_(NULL),
          labelsIterIndex_(-1), itemsIterIndex_(0), currentBucket_(NULL),
          maxLabelCount_(99),
          initialLabels_(NULL), firstCharsInScripts_(NULL),
          collator_(NULL), collatorPrimaryOnly_(NULL),
          buckets_(NULL) {
    init(&locale, status);
}


AlphabeticIndex::AlphabeticIndex(RuleBasedCollator *collator, UErrorCode &status)
        : inputList_(NULL),
          labelsIterIndex_(-1), itemsIterIndex_(0), currentBucket_(NULL),
          maxLabelCount_(99),
          initialLabels_(NULL), firstCharsInScripts_(NULL),
          collator_(collator), collatorPrimaryOnly_(NULL),
          buckets_(NULL) {
    init(NULL, status);
}



AlphabeticIndex::~AlphabeticIndex() {
    delete collator_;
    delete collatorPrimaryOnly_;
    delete firstCharsInScripts_;
    delete buckets_;
    delete inputList_;
    delete initialLabels_;
}


AlphabeticIndex &AlphabeticIndex::addLabels(const UnicodeSet &additions, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return *this;
    }
    initialLabels_->addAll(additions);
    clearBuckets();
    return *this;
}


AlphabeticIndex &AlphabeticIndex::addLabels(const Locale &locale, UErrorCode &status) {
    addIndexExemplars(locale, status);
    clearBuckets();
    return *this;
}


AlphabeticIndex::ImmutableIndex *AlphabeticIndex::buildImmutableIndex(UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return NULL; }
    // In C++, the ImmutableIndex must own its copy of the BucketList,
    // even if it contains no records, for proper memory management.
    // We could clone the buckets_ if they are not NULL,
    // but that would be worth it only if this method is called multiple times,
    // or called after using the old-style bucket iterator API.
    LocalPointer<BucketList> immutableBucketList(createBucketList(errorCode));
    LocalPointer<RuleBasedCollator> coll(collatorPrimaryOnly_->clone());
    if (immutableBucketList.isNull() || coll.isNull()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    ImmutableIndex *immIndex = new ImmutableIndex(immutableBucketList.getAlias(), coll.getAlias());
    if (immIndex == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    // The ImmutableIndex adopted its parameter objects.
    immutableBucketList.orphan();
    coll.orphan();
    return immIndex;
}

int32_t AlphabeticIndex::getBucketCount(UErrorCode &status) {
    initBuckets(status);
    if (U_FAILURE(status)) {
        return 0;
    }
    return buckets_->getBucketCount();
}


int32_t AlphabeticIndex::getRecordCount(UErrorCode &status) {
    if (U_FAILURE(status) || inputList_ == NULL) {
        return 0;
    }
    return inputList_->size();
}

void AlphabeticIndex::initLabels(UVector &indexCharacters, UErrorCode &errorCode) const {
    const Normalizer2 *nfkdNormalizer = Normalizer2::getNFKDInstance(errorCode);
    if (U_FAILURE(errorCode)) { return; }

    const UnicodeString &firstScriptBoundary = *getString(*firstCharsInScripts_, 0);
    const UnicodeString &overflowBoundary =
        *getString(*firstCharsInScripts_, firstCharsInScripts_->size() - 1);

    // We make a sorted array of elements.
    // Some of the input may be redundant.
    // That is, we might have c, ch, d, where "ch" sorts just like "c", "h".
    // We filter out those cases.
    UnicodeSetIterator iter(*initialLabels_);
    while (iter.next()) {
        const UnicodeString *item = &iter.getString();
        LocalPointer<UnicodeString> ownedItem;
        UBool checkDistinct;
        int32_t itemLength = item->length();
        if (!item->hasMoreChar32Than(0, itemLength, 1)) {
            checkDistinct = FALSE;
        } else if(item->charAt(itemLength - 1) == 0x2a &&  // '*'
                item->charAt(itemLength - 2) != 0x2a) {
            // Use a label if it is marked with one trailing star,
            // even if the label string sorts the same when all contractions are suppressed.
            ownedItem.adoptInstead(new UnicodeString(*item, 0, itemLength - 1));
            item = ownedItem.getAlias();
            if (item == NULL) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            checkDistinct = FALSE;
        } else {
            checkDistinct = TRUE;
        }
        if (collatorPrimaryOnly_->compare(*item, firstScriptBoundary, errorCode) < 0) {
            // Ignore a primary-ignorable or non-alphabetic index character.
        } else if (collatorPrimaryOnly_->compare(*item, overflowBoundary, errorCode) >= 0) {
            // Ignore an index character that will land in the overflow bucket.
        } else if (checkDistinct &&
                collatorPrimaryOnly_->compare(*item, separated(*item), errorCode) == 0) {
            // Ignore a multi-code point index character that does not sort distinctly
            // from the sequence of its separate characters.
        } else {
            int32_t insertionPoint = binarySearch(indexCharacters, *item, *collatorPrimaryOnly_);
            if (insertionPoint < 0) {
                indexCharacters.insertElementAt(
                    ownedString(*item, ownedItem, errorCode), ~insertionPoint, errorCode);
            } else {
                const UnicodeString &itemAlreadyIn = *getString(indexCharacters, insertionPoint);
                if (isOneLabelBetterThanOther(*nfkdNormalizer, *item, itemAlreadyIn)) {
                    indexCharacters.setElementAt(
                        ownedString(*item, ownedItem, errorCode), insertionPoint);
                }
            }
        }
    }
    if (U_FAILURE(errorCode)) { return; }

    // if the result is still too large, cut down to maxLabelCount_ elements, by removing every nth element

    int32_t size = indexCharacters.size() - 1;
    if (size > maxLabelCount_) {
        int32_t count = 0;
        int32_t old = -1;
        for (int32_t i = 0; i < indexCharacters.size();) {
            ++count;
            int32_t bump = count * maxLabelCount_ / size;
            if (bump == old) {
                indexCharacters.removeElementAt(i);
            } else {
                old = bump;
                ++i;
            }
        }
    }
}

namespace {

const UnicodeString &fixLabel(const UnicodeString &current, UnicodeString &temp) {
    if (!current.startsWith(BASE, BASE_LENGTH)) {
        return current;
    }
    UChar rest = current.charAt(BASE_LENGTH);
    if (0x2800 < rest && rest <= 0x28FF) { // stroke count
        int32_t count = rest-0x2800;
        temp.setTo((UChar)(0x30 + count % 10));
        if (count >= 10) {
            count /= 10;
            temp.insert(0, (UChar)(0x30 + count % 10));
            if (count >= 10) {
                count /= 10;
                temp.insert(0, (UChar)(0x30 + count));
            }
        }
        return temp.append((UChar)0x5283);
    }
    return temp.setTo(current, BASE_LENGTH);
}

UBool hasMultiplePrimaryWeights(
        const RuleBasedCollator &coll, uint32_t variableTop,
        const UnicodeString &s, UVector64 &ces, UErrorCode &errorCode) {
    ces.removeAllElements();
    coll.internalGetCEs(s, ces, errorCode);
    if (U_FAILURE(errorCode)) { return FALSE; }
    UBool seenPrimary = FALSE;
    for (int32_t i = 0; i < ces.size(); ++i) {
        int64_t ce = ces.elementAti(i);
        uint32_t p = (uint32_t)(ce >> 32);
        if (p > variableTop) {
            // not primary ignorable
            if (seenPrimary) {
                return TRUE;
            }
            seenPrimary = TRUE;
        }
    }
    return FALSE;
}

}  // namespace

BucketList *AlphabeticIndex::createBucketList(UErrorCode &errorCode) const {
    // Initialize indexCharacters.
    UVector indexCharacters(errorCode);
    indexCharacters.setDeleter(uprv_deleteUObject);
    initLabels(indexCharacters, errorCode);
    if (U_FAILURE(errorCode)) { return NULL; }

    // Variables for hasMultiplePrimaryWeights().
    UVector64 ces(errorCode);
    uint32_t variableTop;
    if (collatorPrimaryOnly_->getAttribute(UCOL_ALTERNATE_HANDLING, errorCode) == UCOL_SHIFTED) {
        variableTop = collatorPrimaryOnly_->getVariableTop(errorCode);
    } else {
        variableTop = 0;
    }
    UBool hasInvisibleBuckets = FALSE;

    // Helper arrays for Chinese Pinyin collation.
    Bucket *asciiBuckets[26] = {
        NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
    };
    Bucket *pinyinBuckets[26] = {
        NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
    };
    UBool hasPinyin = FALSE;

    LocalPointer<UVector> bucketList(new UVector(errorCode), errorCode);
    if (U_FAILURE(errorCode)) {
        return NULL;
    }
    bucketList->setDeleter(uprv_deleteUObject);

    // underflow bucket
    Bucket *bucket = new Bucket(getUnderflowLabel(), emptyString_, U_ALPHAINDEX_UNDERFLOW);
    if (bucket == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    bucketList->addElement(bucket, errorCode);
    if (U_FAILURE(errorCode)) { return NULL; }

    UnicodeString temp;

    // fix up the list, adding underflow, additions, overflow
    // Insert inflow labels as needed.
    int32_t scriptIndex = -1;
    const UnicodeString *scriptUpperBoundary = &emptyString_;
    for (int32_t i = 0; i < indexCharacters.size(); ++i) {
        UnicodeString &current = *getString(indexCharacters, i);
        if (collatorPrimaryOnly_->compare(current, *scriptUpperBoundary, errorCode) >= 0) {
            // We crossed the script boundary into a new script.
            const UnicodeString &inflowBoundary = *scriptUpperBoundary;
            UBool skippedScript = FALSE;
            for (;;) {
                scriptUpperBoundary = getString(*firstCharsInScripts_, ++scriptIndex);
                if (collatorPrimaryOnly_->compare(current, *scriptUpperBoundary, errorCode) < 0) {
                    break;
                }
                skippedScript = TRUE;
            }
            if (skippedScript && bucketList->size() > 1) {
                // We are skipping one or more scripts,
                // and we are not just getting out of the underflow label.
                bucket = new Bucket(getInflowLabel(), inflowBoundary, U_ALPHAINDEX_INFLOW);
                if (bucket == NULL) {
                    errorCode = U_MEMORY_ALLOCATION_ERROR;
                    return NULL;
                }
                bucketList->addElement(bucket, errorCode);
            }
        }
        // Add a bucket with the current label.
        bucket = new Bucket(fixLabel(current, temp), current, U_ALPHAINDEX_NORMAL);
        if (bucket == NULL) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        bucketList->addElement(bucket, errorCode);
        // Remember ASCII and Pinyin buckets for Pinyin redirects.
        UChar c;
        if (current.length() == 1 && 0x41 <= (c = current.charAt(0)) && c <= 0x5A) {  // A-Z
            asciiBuckets[c - 0x41] = bucket;
        } else if (current.length() == BASE_LENGTH + 1 && current.startsWith(BASE, BASE_LENGTH) &&
                0x41 <= (c = current.charAt(BASE_LENGTH)) && c <= 0x5A) {
            pinyinBuckets[c - 0x41] = bucket;
            hasPinyin = TRUE;
        }
        // Check for multiple primary weights.
        if (!current.startsWith(BASE, BASE_LENGTH) &&
                hasMultiplePrimaryWeights(*collatorPrimaryOnly_, variableTop, current,
                                          ces, errorCode) &&
                current.charAt(current.length() - 1) != 0xFFFF /* !current.endsWith("\uffff") */) {
            // "AE-ligature" or "Sch" etc.
            for (int32_t j = bucketList->size() - 2;; --j) {
                Bucket *singleBucket = getBucket(*bucketList, j);
                if (singleBucket->labelType_ != U_ALPHAINDEX_NORMAL) {
                    // There is no single-character bucket since the last
                    // underflow or inflow label.
                    break;
                }
                if (singleBucket->displayBucket_ == NULL &&
                        !hasMultiplePrimaryWeights(*collatorPrimaryOnly_, variableTop,
                                                   singleBucket->lowerBoundary_,
                                                   ces, errorCode)) {
                    // Add an invisible bucket that redirects strings greater than the expansion
                    // to the previous single-character bucket.
                    // For example, after ... Q R S Sch we add Sch\uFFFF->S
                    // and after ... Q R S Sch Sch\uFFFF St we add St\uFFFF->S.
                    bucket = new Bucket(emptyString_,
                        UnicodeString(current).append((UChar)0xFFFF),
                        U_ALPHAINDEX_NORMAL);
                    if (bucket == NULL) {
                        errorCode = U_MEMORY_ALLOCATION_ERROR;
                        return NULL;
                    }
                    bucket->displayBucket_ = singleBucket;
                    bucketList->addElement(bucket, errorCode);
                    hasInvisibleBuckets = TRUE;
                    break;
                }
            }
        }
    }
    if (U_FAILURE(errorCode)) { return NULL; }
    if (bucketList->size() == 1) {
        // No real labels, show only the underflow label.
        BucketList *bl = new BucketList(bucketList.getAlias(), bucketList.getAlias());
        if (bl == NULL) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        bucketList.orphan();
        return bl;
    }
    // overflow bucket
    bucket = new Bucket(getOverflowLabel(), *scriptUpperBoundary, U_ALPHAINDEX_OVERFLOW);
    if (bucket == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    bucketList->addElement(bucket, errorCode); // final

    if (hasPinyin) {
        // Redirect Pinyin buckets.
        Bucket *asciiBucket = NULL;
        for (int32_t i = 0; i < 26; ++i) {
            if (asciiBuckets[i] != NULL) {
                asciiBucket = asciiBuckets[i];
            }
            if (pinyinBuckets[i] != NULL && asciiBucket != NULL) {
                pinyinBuckets[i]->displayBucket_ = asciiBucket;
                hasInvisibleBuckets = TRUE;
            }
        }
    }

    if (U_FAILURE(errorCode)) { return NULL; }
    if (!hasInvisibleBuckets) {
        BucketList *bl = new BucketList(bucketList.getAlias(), bucketList.getAlias());
        if (bl == NULL) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        bucketList.orphan();
        return bl;
    }
    // Merge inflow buckets that are visually adjacent.
    // Iterate backwards: Merge inflow into overflow rather than the other way around.
    int32_t i = bucketList->size() - 1;
    Bucket *nextBucket = getBucket(*bucketList, i);
    while (--i > 0) {
        bucket = getBucket(*bucketList, i);
        if (bucket->displayBucket_ != NULL) {
            continue;  // skip invisible buckets
        }
        if (bucket->labelType_ == U_ALPHAINDEX_INFLOW) {
            if (nextBucket->labelType_ != U_ALPHAINDEX_NORMAL) {
                bucket->displayBucket_ = nextBucket;
                continue;
            }
        }
        nextBucket = bucket;
    }

    LocalPointer<UVector> publicBucketList(new UVector(errorCode), errorCode);
    if (U_FAILURE(errorCode)) {
        return NULL;
    }
    // Do not call publicBucketList->setDeleter():
    // This vector shares its objects with the bucketList.
    for (int32_t j = 0; j < bucketList->size(); ++j) {
        bucket = getBucket(*bucketList, j);
        if (bucket->displayBucket_ == NULL) {
            publicBucketList->addElement(bucket, errorCode);
        }
    }
    if (U_FAILURE(errorCode)) { return NULL; }
    BucketList *bl = new BucketList(bucketList.getAlias(), publicBucketList.getAlias());
    if (bl == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    bucketList.orphan();
    publicBucketList.orphan();
    return bl;
}

/**
 * Creates an index, and buckets and sorts the list of records into the index.
 */
void AlphabeticIndex::initBuckets(UErrorCode &errorCode) {
    if (U_FAILURE(errorCode) || buckets_ != NULL) {
        return;
    }
    buckets_ = createBucketList(errorCode);
    if (U_FAILURE(errorCode) || inputList_ == NULL || inputList_->isEmpty()) {
        return;
    }

    // Sort the records by name.
    // Stable sort preserves input order of collation duplicates.
    inputList_->sortWithUComparator(recordCompareFn, collator_, errorCode);

    // Now, we traverse all of the input, which is now sorted.
    // If the item doesn't go in the current bucket, we find the next bucket that contains it.
    // This makes the process order n*log(n), since we just sort the list and then do a linear process.
    // However, if the user adds an item at a time and then gets the buckets, this isn't efficient, so
    // we need to improve it for that case.

    Bucket *currentBucket = getBucket(*buckets_->bucketList_, 0);
    int32_t bucketIndex = 1;
    Bucket *nextBucket;
    const UnicodeString *upperBoundary;
    if (bucketIndex < buckets_->bucketList_->size()) {
        nextBucket = getBucket(*buckets_->bucketList_, bucketIndex++);
        upperBoundary = &nextBucket->lowerBoundary_;
    } else {
        nextBucket = NULL;
        upperBoundary = NULL;
    }
    for (int32_t i = 0; i < inputList_->size(); ++i) {
        Record *r = getRecord(*inputList_, i);
        // if the current bucket isn't the right one, find the one that is
        // We have a special flag for the last bucket so that we don't look any further
        while (upperBoundary != NULL &&
                collatorPrimaryOnly_->compare(r->name_, *upperBoundary, errorCode) >= 0) {
            currentBucket = nextBucket;
            // now reset the boundary that we compare against
            if (bucketIndex < buckets_->bucketList_->size()) {
                nextBucket = getBucket(*buckets_->bucketList_, bucketIndex++);
                upperBoundary = &nextBucket->lowerBoundary_;
            } else {
                upperBoundary = NULL;
            }
        }
        // now put the record into the bucket.
        Bucket *bucket = currentBucket;
        if (bucket->displayBucket_ != NULL) {
            bucket = bucket->displayBucket_;
        }
        if (bucket->records_ == NULL) {
            bucket->records_ = new UVector(errorCode);
            if (bucket->records_ == NULL) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
        }
        bucket->records_->addElement(r, errorCode);
    }
}

void AlphabeticIndex::clearBuckets() {
    if (buckets_ != NULL) {
        delete buckets_;
        buckets_ = NULL;
        internalResetBucketIterator();
    }
}

void AlphabeticIndex::internalResetBucketIterator() {
    labelsIterIndex_ = -1;
    currentBucket_ = NULL;
}


void AlphabeticIndex::addIndexExemplars(const Locale &locale, UErrorCode &status) {
    LocalULocaleDataPointer uld(ulocdata_open(locale.getName(), &status));
    if (U_FAILURE(status)) {
        return;
    }

    UnicodeSet exemplars;
    ulocdata_getExemplarSet(uld.getAlias(), exemplars.toUSet(), 0, ULOCDATA_ES_INDEX, &status);
    if (U_SUCCESS(status)) {
        initialLabels_->addAll(exemplars);
        return;
    }
    status = U_ZERO_ERROR;  // Clear out U_MISSING_RESOURCE_ERROR

    // The locale data did not include explicit Index characters.
    // Synthesize a set of them from the locale's standard exemplar characters.
    ulocdata_getExemplarSet(uld.getAlias(), exemplars.toUSet(), 0, ULOCDATA_ES_STANDARD, &status);
    if (U_FAILURE(status)) {
        return;
    }

    // question: should we add auxiliary exemplars?
    if (exemplars.containsSome(0x61, 0x7A) /* a-z */ || exemplars.isEmpty()) {
        exemplars.add(0x61, 0x7A);
    }
    if (exemplars.containsSome(0xAC00, 0xD7A3)) {  // Hangul syllables
        // cut down to small list
        exemplars.remove(0xAC00, 0xD7A3).
            add(0xAC00).add(0xB098).add(0xB2E4).add(0xB77C).
            add(0xB9C8).add(0xBC14).add(0xC0AC).add(0xC544).
            add(0xC790).add(0xCC28).add(0xCE74).add(0xD0C0).
            add(0xD30C).add(0xD558);
    }
    if (exemplars.containsSome(0x1200, 0x137F)) {  // Ethiopic block
        // cut down to small list
        // make use of the fact that Ethiopic is allocated in 8's, where
        // the base is 0 mod 8.
        UnicodeSet ethiopic(UnicodeString(u"[ሀለሐመሠረሰሸቀቈቐቘበቨተቸኀኈነኘአከኰኸዀወዐዘዠየደዸጀገጐጘጠጨጰጸፀፈፐፘ]"), status);
        ethiopic.retainAll(exemplars);
        exemplars.remove(u'\x1200', 0x137F).addAll(ethiopic);
    }

    // Upper-case any that aren't already so.
    //   (We only do this for synthesized index characters.)
    UnicodeSetIterator it(exemplars);
    UnicodeString upperC;
    while (it.next()) {
        const UnicodeString &exemplarC = it.getString();
        upperC = exemplarC;
        upperC.toUpper(locale);
        initialLabels_->add(upperC);
    }
}

UBool AlphabeticIndex::addChineseIndexCharacters(UErrorCode &errorCode) {
    UnicodeSet contractions;
    collatorPrimaryOnly_->internalAddContractions(BASE[0], contractions, errorCode);
    if (U_FAILURE(errorCode) || contractions.isEmpty()) { return FALSE; }
    initialLabels_->addAll(contractions);
    UnicodeSetIterator iter(contractions);
    while (iter.next()) {
        const UnicodeString &s = iter.getString();
        U_ASSERT (s.startsWith(BASE, BASE_LENGTH));
        UChar c = s.charAt(s.length() - 1);
        if (0x41 <= c && c <= 0x5A) {  // A-Z
            // There are Pinyin labels, add ASCII A-Z labels as well.
            initialLabels_->add(0x41, 0x5A);  // A-Z
            break;
        }
    }
    return TRUE;
}


/*
 * Return the string with interspersed CGJs. Input must have more than 2 codepoints.
 */
static const UChar alphaindex_CGJ = 0x034F;
UnicodeString AlphabeticIndex::separated(const UnicodeString &item) {
    UnicodeString result;
    if (item.length() == 0) {
        return result;
    }
    int32_t i = 0;
    for (;;) {
        UChar32  cp = item.char32At(i);
        result.append(cp);
        i = item.moveIndex32(i, 1);
        if (i >= item.length()) {
            break;
        }
        result.append(alphaindex_CGJ);
    }
    return result;
}


UBool AlphabeticIndex::operator==(const AlphabeticIndex& /* other */) const {
    return FALSE;
}


UBool AlphabeticIndex::operator!=(const AlphabeticIndex& /* other */) const {
    return FALSE;
}


const RuleBasedCollator &AlphabeticIndex::getCollator() const {
    return *collator_;
}


const UnicodeString &AlphabeticIndex::getInflowLabel() const {
    return inflowLabel_;
}

const UnicodeString &AlphabeticIndex::getOverflowLabel() const {
    return overflowLabel_;
}


const UnicodeString &AlphabeticIndex::getUnderflowLabel() const {
    return underflowLabel_;
}


AlphabeticIndex &AlphabeticIndex::setInflowLabel(const UnicodeString &label, UErrorCode &/*status*/) {
    inflowLabel_ = label;
    clearBuckets();
    return *this;
}


AlphabeticIndex &AlphabeticIndex::setOverflowLabel(const UnicodeString &label, UErrorCode &/*status*/) {
    overflowLabel_ = label;
    clearBuckets();
    return *this;
}


AlphabeticIndex &AlphabeticIndex::setUnderflowLabel(const UnicodeString &label, UErrorCode &/*status*/) {
    underflowLabel_ = label;
    clearBuckets();
    return *this;
}


int32_t AlphabeticIndex::getMaxLabelCount() const {
    return maxLabelCount_;
}


AlphabeticIndex &AlphabeticIndex::setMaxLabelCount(int32_t maxLabelCount, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return *this;
    }
    if (maxLabelCount <= 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return *this;
    }
    maxLabelCount_ = maxLabelCount;
    clearBuckets();
    return *this;
}


//
//  init() - Common code for constructors.
//

void AlphabeticIndex::init(const Locale *locale, UErrorCode &status) {
    if (U_FAILURE(status)) { return; }
    if (locale == NULL && collator_ == NULL) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    initialLabels_         = new UnicodeSet();
    if (initialLabels_ == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    inflowLabel_.setTo((UChar)0x2026);    // Ellipsis
    overflowLabel_ = inflowLabel_;
    underflowLabel_ = inflowLabel_;

    if (collator_ == NULL) {
        Collator *coll = Collator::createInstance(*locale, status);
        if (U_FAILURE(status)) {
            delete coll;
            return;
        }
        if (coll == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        collator_ = dynamic_cast<RuleBasedCollator *>(coll);
        if (collator_ == NULL) {
            delete coll;
            status = U_UNSUPPORTED_ERROR;
            return;
        }
    }
    collatorPrimaryOnly_ = collator_->clone();
    if (collatorPrimaryOnly_ == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    collatorPrimaryOnly_->setAttribute(UCOL_STRENGTH, UCOL_PRIMARY, status);
    firstCharsInScripts_ = firstStringsInScript(status);
    if (U_FAILURE(status)) { return; }
    firstCharsInScripts_->sortWithUComparator(collatorComparator, collatorPrimaryOnly_, status);
    // Guard against a degenerate collator where
    // some script boundary strings are primary ignorable.
    for (;;) {
        if (U_FAILURE(status)) { return; }
        if (firstCharsInScripts_->isEmpty()) {
            // AlphabeticIndex requires some non-ignorable script boundary strings.
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        if (collatorPrimaryOnly_->compare(
                *static_cast<UnicodeString *>(firstCharsInScripts_->elementAt(0)),
                emptyString_, status) == UCOL_EQUAL) {
            firstCharsInScripts_->removeElementAt(0);
        } else {
            break;
        }
    }

    // Chinese index characters, which are specific to each of the several Chinese tailorings,
    // take precedence over the single locale data exemplar set per language.
    if (!addChineseIndexCharacters(status) && locale != NULL) {
        addIndexExemplars(*locale, status);
    }
}


//
//  Comparison function for UVector<UnicodeString *> sorting with a collator.
//
static int32_t U_CALLCONV
collatorComparator(const void *context, const void *left, const void *right) {
    const UElement *leftElement = static_cast<const UElement *>(left);
    const UElement *rightElement = static_cast<const UElement *>(right);
    const UnicodeString *leftString  = static_cast<const UnicodeString *>(leftElement->pointer);
    const UnicodeString *rightString = static_cast<const UnicodeString *>(rightElement->pointer);

    if (leftString == rightString) {
        // Catches case where both are NULL
        return 0;
    }
    if (leftString == NULL) {
        return 1;
    }
    if (rightString == NULL) {
        return -1;
    }
    const Collator *col = static_cast<const Collator *>(context);
    UErrorCode errorCode = U_ZERO_ERROR;
    return col->compare(*leftString, *rightString, errorCode);
}

//
//  Comparison function for UVector<Record *> sorting with a collator.
//
static int32_t U_CALLCONV
recordCompareFn(const void *context, const void *left, const void *right) {
    const UElement *leftElement = static_cast<const UElement *>(left);
    const UElement *rightElement = static_cast<const UElement *>(right);
    const AlphabeticIndex::Record *leftRec  = static_cast<const AlphabeticIndex::Record *>(leftElement->pointer);
    const AlphabeticIndex::Record *rightRec = static_cast<const AlphabeticIndex::Record *>(rightElement->pointer);
    const Collator *col = static_cast<const Collator *>(context);
    UErrorCode errorCode = U_ZERO_ERROR;
    return col->compare(leftRec->name_, rightRec->name_, errorCode);
}

UVector *AlphabeticIndex::firstStringsInScript(UErrorCode &status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    LocalPointer<UVector> dest(new UVector(status), status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    dest->setDeleter(uprv_deleteUObject);
    // Fetch the script-first-primary contractions which are defined in the root collator.
    // They all start with U+FDD1.
    UnicodeSet set;
    collatorPrimaryOnly_->internalAddContractions(0xFDD1, set, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    if (set.isEmpty()) {
        status = U_UNSUPPORTED_ERROR;
        return NULL;
    }
    UnicodeSetIterator iter(set);
    while (iter.next()) {
        const UnicodeString &boundary = iter.getString();
        uint32_t gcMask = U_GET_GC_MASK(boundary.char32At(1));
        if ((gcMask & (U_GC_L_MASK | U_GC_CN_MASK)) == 0) {
            // Ignore boundaries for the special reordering groups.
            // Take only those for "real scripts" (where the sample character is a Letter,
            // and the one for unassigned implicit weights (Cn).
            continue;
        }
        UnicodeString *s = new UnicodeString(boundary);
        if (s == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        dest->addElement(s, status);
    }
    return dest.orphan();
}


namespace {

/**
 * Returns true if one index character string is "better" than the other.
 * Shorter NFKD is better, and otherwise NFKD-binary-less-than is
 * better, and otherwise binary-less-than is better.
 */
UBool isOneLabelBetterThanOther(const Normalizer2 &nfkdNormalizer,
                                const UnicodeString &one, const UnicodeString &other) {
    // This is called with primary-equal strings, but never with one.equals(other).
    UErrorCode status = U_ZERO_ERROR;
    UnicodeString n1 = nfkdNormalizer.normalize(one, status);
    UnicodeString n2 = nfkdNormalizer.normalize(other, status);
    if (U_FAILURE(status)) { return FALSE; }
    int32_t result = n1.countChar32() - n2.countChar32();
    if (result != 0) {
        return result < 0;
    }
    result = n1.compareCodePointOrder(n2);
    if (result != 0) {
        return result < 0;
    }
    return one.compareCodePointOrder(other) < 0;
}

}  // namespace

//
//  Constructor & Destructor for AlphabeticIndex::Record
//
//     Records are internal only, instances are not directly surfaced in the public API.
//     This class is mostly struct-like, with all public fields.

AlphabeticIndex::Record::Record(const UnicodeString &name, const void *data)
        : name_(name), data_(data) {}

AlphabeticIndex::Record::~Record() {
}


AlphabeticIndex & AlphabeticIndex::addRecord(const UnicodeString &name, const void *data, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return *this;
    }
    if (inputList_ == NULL) {
        inputList_ = new UVector(status);
        if (inputList_ == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return *this;
        }
        inputList_->setDeleter(alphaIndex_deleteRecord);
    }
    Record *r = new Record(name, data);
    if (r == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return *this;
    }
    inputList_->addElement(r, status);
    clearBuckets();
    //std::string ss;
    //std::string ss2;
    //std::cout << "added record: name = \"" << r->name_.toUTF8String(ss) << "\"" <<
    //             "   sortingName = \"" << r->sortingName_.toUTF8String(ss2) << "\"" << std::endl;
    return *this;
}


AlphabeticIndex &AlphabeticIndex::clearRecords(UErrorCode &status) {
    if (U_SUCCESS(status) && inputList_ != NULL && !inputList_->isEmpty()) {
        inputList_->removeAllElements();
        clearBuckets();
    }
    return *this;
}

int32_t AlphabeticIndex::getBucketIndex(const UnicodeString &name, UErrorCode &status) {
    initBuckets(status);
    if (U_FAILURE(status)) {
        return 0;
    }
    return buckets_->getBucketIndex(name, *collatorPrimaryOnly_, status);
}


int32_t AlphabeticIndex::getBucketIndex() const {
    return labelsIterIndex_;
}


UBool AlphabeticIndex::nextBucket(UErrorCode &status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (buckets_ == NULL && currentBucket_ != NULL) {
        status = U_ENUM_OUT_OF_SYNC_ERROR;
        return FALSE;
    }
    initBuckets(status);
    if (U_FAILURE(status)) {
        return FALSE;
    }
    ++labelsIterIndex_;
    if (labelsIterIndex_ >= buckets_->getBucketCount()) {
        labelsIterIndex_ = buckets_->getBucketCount();
        return FALSE;
    }
    currentBucket_ = getBucket(*buckets_->immutableVisibleList_, labelsIterIndex_);
    resetRecordIterator();
    return TRUE;
}

const UnicodeString &AlphabeticIndex::getBucketLabel() const {
    if (currentBucket_ != NULL) {
        return currentBucket_->label_;
    } else {
        return emptyString_;
    }
}


UAlphabeticIndexLabelType AlphabeticIndex::getBucketLabelType() const {
    if (currentBucket_ != NULL) {
        return currentBucket_->labelType_;
    } else {
        return U_ALPHAINDEX_NORMAL;
    }
}


int32_t AlphabeticIndex::getBucketRecordCount() const {
    if (currentBucket_ != NULL && currentBucket_->records_ != NULL) {
        return currentBucket_->records_->size();
    } else {
        return 0;
    }
}


AlphabeticIndex &AlphabeticIndex::resetBucketIterator(UErrorCode &status) {
    if (U_FAILURE(status)) {
        return *this;
    }
    internalResetBucketIterator();
    return *this;
}


UBool AlphabeticIndex::nextRecord(UErrorCode &status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (currentBucket_ == NULL) {
        // We are trying to iterate over the items in a bucket, but there is no
        // current bucket from the enumeration of buckets.
        status = U_INVALID_STATE_ERROR;
        return FALSE;
    }
    if (buckets_ == NULL) {
        status = U_ENUM_OUT_OF_SYNC_ERROR;
        return FALSE;
    }
    if (currentBucket_->records_ == NULL) {
        return FALSE;
    }
    ++itemsIterIndex_;
    if (itemsIterIndex_ >= currentBucket_->records_->size()) {
        itemsIterIndex_  = currentBucket_->records_->size();
        return FALSE;
    }
    return TRUE;
}


const UnicodeString &AlphabeticIndex::getRecordName() const {
    const UnicodeString *retStr = &emptyString_;
    if (currentBucket_ != NULL && currentBucket_->records_ != NULL &&
        itemsIterIndex_ >= 0 &&
        itemsIterIndex_ < currentBucket_->records_->size()) {
            Record *item = static_cast<Record *>(currentBucket_->records_->elementAt(itemsIterIndex_));
            retStr = &item->name_;
    }
    return *retStr;
}

const void *AlphabeticIndex::getRecordData() const {
    const void *retPtr = NULL;
    if (currentBucket_ != NULL && currentBucket_->records_ != NULL &&
        itemsIterIndex_ >= 0 &&
        itemsIterIndex_ < currentBucket_->records_->size()) {
            Record *item = static_cast<Record *>(currentBucket_->records_->elementAt(itemsIterIndex_));
            retPtr = item->data_;
    }
    return retPtr;
}


AlphabeticIndex & AlphabeticIndex::resetRecordIterator() {
    itemsIterIndex_ = -1;
    return *this;
}



AlphabeticIndex::Bucket::Bucket(const UnicodeString &label,
                                const UnicodeString &lowerBoundary,
                                UAlphabeticIndexLabelType type)
        : label_(label), lowerBoundary_(lowerBoundary), labelType_(type),
          displayBucket_(NULL), displayIndex_(-1),
          records_(NULL) {
}


AlphabeticIndex::Bucket::~Bucket() {
    delete records_;
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/************************************************************************
 * Copyright (C) 1996-2012, International Business Machines Corporation
 * and others. All Rights Reserved.
 ************************************************************************
 *  2003-nov-07   srl       Port from Java
 */

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/************************************************************************
 * Copyright (C) 1996-2008, International Business Machines Corporation *
 * and others. All Rights Reserved.                                     *
 ************************************************************************
 *  2003-nov-07   srl       Port from Java
 */




#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 **********************************************************************
 * Copyright (c) 2003-2008, International Business Machines
 * Corporation and others.  All Rights Reserved.
 **********************************************************************
 * Author: Alan Liu
 * Created: September 2 2003
 * Since: ICU 2.8
 **********************************************************************
 */


#if !UCONFIG_NO_FORMATTING





U_NAMESPACE_BEGIN

/**
 * A utility class providing mathematical functions used by time zone
 * and calendar code.  Do not instantiate.  Formerly just named 'Math'.
 * @internal
 */
class ClockMath {
 public:
    /**
     * Divide two integers, returning the floor of the quotient.
     * Unlike the built-in division, this is mathematically
     * well-behaved.  E.g., <code>-1/4</code> => 0 but
     * <code>floorDivide(-1,4)</code> => -1.
     * @param numerator the numerator
     * @param denominator a divisor which must be != 0
     * @return the floor of the quotient
     */
    static int32_t floorDivide(int32_t numerator, int32_t denominator);

    /**
     * Divide two integers, returning the floor of the quotient.
     * Unlike the built-in division, this is mathematically
     * well-behaved.  E.g., <code>-1/4</code> => 0 but
     * <code>floorDivide(-1,4)</code> => -1.
     * @param numerator the numerator
     * @param denominator a divisor which must be != 0
     * @return the floor of the quotient
     */
    static int64_t floorDivide(int64_t numerator, int64_t denominator);

    /**
     * Divide two numbers, returning the floor of the quotient.
     * Unlike the built-in division, this is mathematically
     * well-behaved.  E.g., <code>-1/4</code> => 0 but
     * <code>floorDivide(-1,4)</code> => -1.
     * @param numerator the numerator
     * @param denominator a divisor which must be != 0
     * @return the floor of the quotient
     */
    static inline double floorDivide(double numerator, double denominator);

    /**
     * Divide two numbers, returning the floor of the quotient and
     * the modulus remainder.  Unlike the built-in division, this is
     * mathematically well-behaved.  E.g., <code>-1/4</code> => 0 and
     * <code>-1%4</code> => -1, but <code>floorDivide(-1,4)</code> =>
     * -1 with <code>remainder</code> => 3.  NOTE: If numerator is
     * too large, the returned quotient may overflow.
     * @param numerator the numerator
     * @param denominator a divisor which must be != 0
     * @param remainder output parameter to receive the
     * remainder. Unlike <code>numerator % denominator</code>, this
     * will always be non-negative, in the half-open range <code>[0,
     * |denominator|)</code>.
     * @return the floor of the quotient
     */
    static int32_t floorDivide(double numerator, int32_t denominator,
                               int32_t& remainder);

    /**
     * For a positive divisor, return the quotient and remainder
     * such that dividend = quotient*divisor + remainder and
     * 0 <= remainder < divisor.
     *
     * Works around edge-case bugs.  Handles pathological input
     * (divident >> divisor) reasonably.
     *
     * Calling with a divisor <= 0 is disallowed.
     */
    static double floorDivide(double dividend, double divisor,
                              double& remainder);
};

// Useful millisecond constants
#define kOneDay    (1.0 * U_MILLIS_PER_DAY)       //  86,400,000
#define kOneHour   (60*60*1000)
#define kOneMinute 60000
#define kOneSecond 1000
#define kOneMillisecond  1
#define kOneWeek   (7.0 * kOneDay) // 604,800,000

// Epoch constants
#define kJan1_1JulianDay  1721426 // January 1, year 1 (Gregorian)

#define kEpochStartAsJulianDay  2440588 // January 1, 1970 (Gregorian)

#define kEpochYear              1970


#define kEarliestViableMillis  -185331720384000000.0  // minimum representable by julian day  -1e17

#define kLatestViableMillis     185753453990400000.0  // max representable by julian day      +1e17

/**
 * The minimum supported Julian day.  This value is equivalent to
 * MIN_MILLIS.
 */
#define MIN_JULIAN (-0x7F000000)

/**
 * The minimum supported epoch milliseconds.  This value is equivalent
 * to MIN_JULIAN.
 */
#define MIN_MILLIS ((MIN_JULIAN - kEpochStartAsJulianDay) * kOneDay)

/**
 * The maximum supported Julian day.  This value is equivalent to
 * MAX_MILLIS.
 */
#define MAX_JULIAN (+0x7F000000)

/**
 * The maximum supported epoch milliseconds.  This value is equivalent
 * to MAX_JULIAN.
 */
#define MAX_MILLIS ((MAX_JULIAN - kEpochStartAsJulianDay) * kOneDay)

/**
 * A utility class providing proleptic Gregorian calendar functions
 * used by time zone and calendar code.  Do not instantiate.
 *
 * Note:  Unlike GregorianCalendar, all computations performed by this
 * class occur in the pure proleptic GregorianCalendar.
 */
class Grego {
 public:
    /**
     * Return TRUE if the given year is a leap year.
     * @param year Gregorian year, with 0 == 1 BCE, -1 == 2 BCE, etc.
     * @return TRUE if the year is a leap year
     */
    static inline UBool isLeapYear(int32_t year);

    /**
     * Return the number of days in the given month.
     * @param year Gregorian year, with 0 == 1 BCE, -1 == 2 BCE, etc.
     * @param month 0-based month, with 0==Jan
     * @return the number of days in the given month
     */
    static inline int8_t monthLength(int32_t year, int32_t month);

    /**
     * Return the length of a previous month of the Gregorian calendar.
     * @param y the extended year
     * @param m the 0-based month number
     * @return the number of days in the month previous to the given month
     */
    static inline int8_t previousMonthLength(int y, int m);

    /**
     * Convert a year, month, and day-of-month, given in the proleptic
     * Gregorian calendar, to 1970 epoch days.
     * @param year Gregorian year, with 0 == 1 BCE, -1 == 2 BCE, etc.
     * @param month 0-based month, with 0==Jan
     * @param dom 1-based day of month
     * @return the day number, with day 0 == Jan 1 1970
     */
    static double fieldsToDay(int32_t year, int32_t month, int32_t dom);
    
    /**
     * Convert a 1970-epoch day number to proleptic Gregorian year,
     * month, day-of-month, and day-of-week.
     * @param day 1970-epoch day (integral value)
     * @param year output parameter to receive year
     * @param month output parameter to receive month (0-based, 0==Jan)
     * @param dom output parameter to receive day-of-month (1-based)
     * @param dow output parameter to receive day-of-week (1-based, 1==Sun)
     * @param doy output parameter to receive day-of-year (1-based)
     */
    static void dayToFields(double day, int32_t& year, int32_t& month,
                            int32_t& dom, int32_t& dow, int32_t& doy);

    /**
     * Convert a 1970-epoch day number to proleptic Gregorian year,
     * month, day-of-month, and day-of-week.
     * @param day 1970-epoch day (integral value)
     * @param year output parameter to receive year
     * @param month output parameter to receive month (0-based, 0==Jan)
     * @param dom output parameter to receive day-of-month (1-based)
     * @param dow output parameter to receive day-of-week (1-based, 1==Sun)
     */
    static inline void dayToFields(double day, int32_t& year, int32_t& month,
                                   int32_t& dom, int32_t& dow);

    /**
     * Convert a 1970-epoch milliseconds to proleptic Gregorian year,
     * month, day-of-month, and day-of-week, day of year and millis-in-day.
     * @param time 1970-epoch milliseconds
     * @param year output parameter to receive year
     * @param month output parameter to receive month (0-based, 0==Jan)
     * @param dom output parameter to receive day-of-month (1-based)
     * @param dow output parameter to receive day-of-week (1-based, 1==Sun)
     * @param doy output parameter to receive day-of-year (1-based)
     * @param mid output parameter to recieve millis-in-day
     */
    static void timeToFields(UDate time, int32_t& year, int32_t& month,
                            int32_t& dom, int32_t& dow, int32_t& doy, int32_t& mid);

    /**
     * Return the day of week on the 1970-epoch day
     * @param day the 1970-epoch day (integral value)
     * @return the day of week
     */
    static int32_t dayOfWeek(double day);

    /**
     * Returns the ordinal number for the specified day of week within the month.
     * The valid return value is 1, 2, 3, 4 or -1.
     * @param year Gregorian year, with 0 == 1 BCE, -1 == 2 BCE, etc.
     * @param month 0-based month, with 0==Jan
     * @param dom 1-based day of month
     * @return The ordinal number for the specified day of week within the month
     */
    static int32_t dayOfWeekInMonth(int32_t year, int32_t month, int32_t dom);

    /**
     * Converts Julian day to time as milliseconds.
     * @param julian the given Julian day number.
     * @return time as milliseconds.
     * @internal
     */
    static inline double julianDayToMillis(int32_t julian);

    /**
     * Converts time as milliseconds to Julian day.
     * @param millis the given milliseconds.
     * @return the Julian day number.
     * @internal
     */
    static inline int32_t millisToJulianDay(double millis);

    /** 
     * Calculates the Gregorian day shift value for an extended year.
     * @param eyear Extended year 
     * @returns number of days to ADD to Julian in order to convert from J->G
     */
    static inline int32_t gregorianShift(int32_t eyear);

 private:
    static const int16_t DAYS_BEFORE[24];
    static const int8_t MONTH_LENGTH[24];
};

inline double ClockMath::floorDivide(double numerator, double denominator) {
    return uprv_floor(numerator / denominator);
}

inline UBool Grego::isLeapYear(int32_t year) {
    // year&0x3 == year%4
    return ((year&0x3) == 0) && ((year%100 != 0) || (year%400 == 0));
}

inline int8_t
Grego::monthLength(int32_t year, int32_t month) {
    return MONTH_LENGTH[month + (isLeapYear(year) ? 12 : 0)];
}

inline int8_t
Grego::previousMonthLength(int y, int m) {
  return (m > 0) ? monthLength(y, m-1) : 31;
}

inline void Grego::dayToFields(double day, int32_t& year, int32_t& month,
                               int32_t& dom, int32_t& dow) {
  int32_t doy_unused;
  dayToFields(day,year,month,dom,dow,doy_unused);
}

inline double Grego::julianDayToMillis(int32_t julian)
{
  return (julian - kEpochStartAsJulianDay) * kOneDay;
}

inline int32_t Grego::millisToJulianDay(double millis) {
  return (int32_t) (kEpochStartAsJulianDay + ClockMath::floorDivide(millis, (double)kOneDay));
}

inline int32_t Grego::gregorianShift(int32_t eyear) {
  int64_t y = (int64_t)eyear-1;
  int32_t gregShift = static_cast<int32_t>(ClockMath::floorDivide(y, (int64_t)400) - ClockMath::floorDivide(y, (int64_t)100) + 2);
  return gregShift;
}

U_NAMESPACE_END

#endif // !UCONFIG_NO_FORMATTING
  // for Math


U_NAMESPACE_BEGIN

/**
 * <code>CalendarAstronomer</code> is a class that can perform the calculations to
 * determine the positions of the sun and moon, the time of sunrise and
 * sunset, and other astronomy-related data.  The calculations it performs
 * are in some cases quite complicated, and this utility class saves you
 * the trouble of worrying about them.
 * <p>
 * The measurement of time is a very important part of astronomy.  Because
 * astronomical bodies are constantly in motion, observations are only valid
 * at a given moment in time.  Accordingly, each <code>CalendarAstronomer</code>
 * object has a <code>time</code> property that determines the date
 * and time for which its calculations are performed.  You can set and
 * retrieve this property with {@link #setDate setDate}, {@link #getDate getDate}
 * and related methods.
 * <p>
 * Almost all of the calculations performed by this class, or by any
 * astronomer, are approximations to various degrees of accuracy.  The
 * calculations in this class are mostly modelled after those described
 * in the book
 * <a href="http://www.amazon.com/exec/obidos/ISBN=0521356997" target="_top">
 * Practical Astronomy With Your Calculator</a>, by Peter J.
 * Duffett-Smith, Cambridge University Press, 1990.  This is an excellent
 * book, and if you want a greater understanding of how these calculations
 * are performed it a very good, readable starting point.
 * <p>
 * <strong>WARNING:</strong> This class is very early in its development, and
 * it is highly likely that its API will change to some degree in the future.
 * At the moment, it basically does just enough to support {@link IslamicCalendar}
 * and {@link ChineseCalendar}.
 *
 * @author Laura Werner
 * @author Alan Liu
 * @internal
 */
class U_I18N_API CalendarAstronomer : public UMemory {
public:
  // some classes

public:
  /**
   * Represents the position of an object in the sky relative to the ecliptic,
   * the plane of the earth's orbit around the Sun.
   * This is a spherical coordinate system in which the latitude
   * specifies the position north or south of the plane of the ecliptic.
   * The longitude specifies the position along the ecliptic plane
   * relative to the "First Point of Aries", which is the Sun's position in the sky
   * at the Vernal Equinox.
   * <p>
   * Note that Ecliptic objects are immutable and cannot be modified
   * once they are constructed.  This allows them to be passed and returned by
   * value without worrying about whether other code will modify them.
   *
   * @see CalendarAstronomer.Equatorial
   * @see CalendarAstronomer.Horizon
   * @internal
   */
  class U_I18N_API Ecliptic : public UMemory {
  public:
    /**
     * Constructs an Ecliptic coordinate object.
     * <p>
     * @param lat The ecliptic latitude, measured in radians.
     * @param lon The ecliptic longitude, measured in radians.
     * @internal
     */
    Ecliptic(double lat = 0, double lon = 0) {
      latitude = lat;
      longitude = lon;
    }

    /**
     * Setter for Ecliptic Coordinate object
     * @param lat The ecliptic latitude, measured in radians.
     * @param lon The ecliptic longitude, measured in radians.
     * @internal
     */
    void set(double lat, double lon) {
      latitude = lat;
      longitude = lon;
    }

    /**
     * Return a string representation of this object
     * @internal
     */
    UnicodeString toString() const;

    /**
     * The ecliptic latitude, in radians.  This specifies an object's
     * position north or south of the plane of the ecliptic,
     * with positive angles representing north.
     * @internal
     */
    double latitude;

    /**
     * The ecliptic longitude, in radians.
     * This specifies an object's position along the ecliptic plane
     * relative to the "First Point of Aries", which is the Sun's position
     * in the sky at the Vernal Equinox,
     * with positive angles representing east.
     * <p>
     * A bit of trivia: the first point of Aries is currently in the
     * constellation Pisces, due to the precession of the earth's axis.
     * @internal
     */
    double longitude;
  };

  /**
   * Represents the position of an
   * object in the sky relative to the plane of the earth's equator.
   * The <i>Right Ascension</i> specifies the position east or west
   * along the equator, relative to the sun's position at the vernal
   * equinox.  The <i>Declination</i> is the position north or south
   * of the equatorial plane.
   * <p>
   * Note that Equatorial objects are immutable and cannot be modified
   * once they are constructed.  This allows them to be passed and returned by
   * value without worrying about whether other code will modify them.
   *
   * @see CalendarAstronomer.Ecliptic
   * @see CalendarAstronomer.Horizon
   * @internal
   */
  class U_I18N_API Equatorial : public UMemory {
  public:
    /**
     * Constructs an Equatorial coordinate object.
     * <p>
     * @param asc The right ascension, measured in radians.
     * @param dec The declination, measured in radians.
     * @internal
     */
    Equatorial(double asc = 0, double dec = 0)
      : ascension(asc), declination(dec) { }

    /**
     * Setter
     * @param asc The right ascension, measured in radians.
     * @param dec The declination, measured in radians.
     * @internal
     */
    void set(double asc, double dec) {
      ascension = asc;
      declination = dec;
    }

    /**
     * Return a string representation of this object, with the
     * angles measured in degrees.
     * @internal
     */
    UnicodeString toString() const;

    /**
     * Return a string representation of this object with the right ascension
     * measured in hours, minutes, and seconds.
     * @internal
     */
    //String toHmsString() {
    //return radToHms(ascension) + "," + radToDms(declination);
    //}

    /**
     * The right ascension, in radians.
     * This is the position east or west along the equator
     * relative to the sun's position at the vernal equinox,
     * with positive angles representing East.
     * @internal
     */
    double ascension;

    /**
     * The declination, in radians.
     * This is the position north or south of the equatorial plane,
     * with positive angles representing north.
     * @internal
     */
    double declination;
  };

  /**
   * Represents the position of an  object in the sky relative to
   * the local horizon.
   * The <i>Altitude</i> represents the object's elevation above the horizon,
   * with objects below the horizon having a negative altitude.
   * The <i>Azimuth</i> is the geographic direction of the object from the
   * observer's position, with 0 representing north.  The azimuth increases
   * clockwise from north.
   * <p>
   * Note that Horizon objects are immutable and cannot be modified
   * once they are constructed.  This allows them to be passed and returned by
   * value without worrying about whether other code will modify them.
   *
   * @see CalendarAstronomer.Ecliptic
   * @see CalendarAstronomer.Equatorial
   * @internal
   */
  class U_I18N_API Horizon : public UMemory {
  public:
    /**
     * Constructs a Horizon coordinate object.
     * <p>
     * @param alt  The altitude, measured in radians above the horizon.
     * @param azim The azimuth, measured in radians clockwise from north.
     * @internal
     */
    Horizon(double alt=0, double azim=0)
      : altitude(alt), azimuth(azim) { }

    /**
     * Setter for Ecliptic Coordinate object
     * @param alt  The altitude, measured in radians above the horizon.
     * @param azim The azimuth, measured in radians clockwise from north.
     * @internal
     */
    void set(double alt, double azim) {
      altitude = alt;
      azimuth = azim;
    }

    /**
     * Return a string representation of this object, with the
     * angles measured in degrees.
     * @internal
     */
    UnicodeString toString() const;

    /**
     * The object's altitude above the horizon, in radians.
     * @internal
     */
    double altitude;

    /**
     * The object's direction, in radians clockwise from north.
     * @internal
     */
    double azimuth;
  };

public:
  //-------------------------------------------------------------------------
  // Assorted private data used for conversions
  //-------------------------------------------------------------------------

  // My own copies of these so compilers are more likely to optimize them away
  static const double PI;

  /**
   * The average number of solar days from one new moon to the next.  This is the time
   * it takes for the moon to return the same ecliptic longitude as the sun.
   * It is longer than the sidereal month because the sun's longitude increases
   * during the year due to the revolution of the earth around the sun.
   * Approximately 29.53.
   *
   * @see #SIDEREAL_MONTH
   * @internal
   * @deprecated ICU 2.4. This class may be removed or modified.
   */
  static const double SYNODIC_MONTH;

  //-------------------------------------------------------------------------
  // Constructors
  //-------------------------------------------------------------------------

  /**
   * Construct a new <code>CalendarAstronomer</code> object that is initialized to
   * the current date and time.
   * @internal
   */
  CalendarAstronomer();

  /**
   * Construct a new <code>CalendarAstronomer</code> object that is initialized to
   * the specified date and time.
   * @internal
   */
  CalendarAstronomer(UDate d);

  /**
   * Construct a new <code>CalendarAstronomer</code> object with the given
   * latitude and longitude.  The object's time is set to the current
   * date and time.
   * <p>
   * @param longitude The desired longitude, in <em>degrees</em> east of
   *                  the Greenwich meridian.
   *
   * @param latitude  The desired latitude, in <em>degrees</em>.  Positive
   *                  values signify North, negative South.
   *
   * @see java.util.Date#getTime()
   * @internal
   */
  CalendarAstronomer(double longitude, double latitude);

  /**
   * Destructor
   * @internal
   */
  ~CalendarAstronomer();

  //-------------------------------------------------------------------------
  // Time and date getters and setters
  //-------------------------------------------------------------------------

  /**
   * Set the current date and time of this <code>CalendarAstronomer</code> object.  All
   * astronomical calculations are performed based on this time setting.
   *
   * @param aTime the date and time, expressed as the number of milliseconds since
   *              1/1/1970 0:00 GMT (Gregorian).
   *
   * @see #setDate
   * @see #getTime
   * @internal
   */
  void setTime(UDate aTime);


  /**
   * Set the current date and time of this <code>CalendarAstronomer</code> object.  All
   * astronomical calculations are performed based on this time setting.
   *
   * @param aTime the date and time, expressed as the number of milliseconds since
   *              1/1/1970 0:00 GMT (Gregorian).
   *
   * @see #getTime
   * @internal
   */
  void setDate(UDate aDate) { setTime(aDate); }

  /**
   * Set the current date and time of this <code>CalendarAstronomer</code> object.  All
   * astronomical calculations are performed based on this time setting.
   *
   * @param jdn   the desired time, expressed as a "julian day number",
   *              which is the number of elapsed days since
   *              1/1/4713 BC (Julian), 12:00 GMT.  Note that julian day
   *              numbers start at <em>noon</em>.  To get the jdn for
   *              the corresponding midnight, subtract 0.5.
   *
   * @see #getJulianDay
   * @see #JULIAN_EPOCH_MS
   * @internal
   */
  void setJulianDay(double jdn);

  /**
   * Get the current time of this <code>CalendarAstronomer</code> object,
   * represented as the number of milliseconds since
   * 1/1/1970 AD 0:00 GMT (Gregorian).
   *
   * @see #setTime
   * @see #getDate
   * @internal
   */
  UDate getTime();

  /**
   * Get the current time of this <code>CalendarAstronomer</code> object,
   * expressed as a "julian day number", which is the number of elapsed
   * days since 1/1/4713 BC (Julian), 12:00 GMT.
   *
   * @see #setJulianDay
   * @see #JULIAN_EPOCH_MS
   * @internal
   */
  double getJulianDay();

  /**
   * Return this object's time expressed in julian centuries:
   * the number of centuries after 1/1/1900 AD, 12:00 GMT
   *
   * @see #getJulianDay
   * @internal
   */
  double getJulianCentury();

  /**
   * Returns the current Greenwich sidereal time, measured in hours
   * @internal
   */
  double getGreenwichSidereal();

private:
  double getSiderealOffset();
public:
  /**
   * Returns the current local sidereal time, measured in hours
   * @internal
   */
  double getLocalSidereal();

  /**
   * Converts local sidereal time to Universal Time.
   *
   * @param lst   The Local Sidereal Time, in hours since sidereal midnight
   *              on this object's current date.
   *
   * @return      The corresponding Universal Time, in milliseconds since
   *              1 Jan 1970, GMT.
   */
  //private:
  double lstToUT(double lst);

  /**
   *
   * Convert from ecliptic to equatorial coordinates.
   *
   * @param ecliptic     The ecliptic
   * @param result       Fillin result
   * @return reference to result
   */
  Equatorial& eclipticToEquatorial(Equatorial& result, const Ecliptic& ecliptic);

  /**
   * Convert from ecliptic to equatorial coordinates.
   *
   * @param eclipLong     The ecliptic longitude
   * @param eclipLat      The ecliptic latitude
   *
   * @return              The corresponding point in equatorial coordinates.
   * @internal
   */
  Equatorial& eclipticToEquatorial(Equatorial& result, double eclipLong, double eclipLat);

  /**
   * Convert from ecliptic longitude to equatorial coordinates.
   *
   * @param eclipLong     The ecliptic longitude
   *
   * @return              The corresponding point in equatorial coordinates.
   * @internal
   */
  Equatorial& eclipticToEquatorial(Equatorial& result, double eclipLong) ;

  /**
   * @internal
   */
  Horizon& eclipticToHorizon(Horizon& result, double eclipLong) ;

  //-------------------------------------------------------------------------
  // The Sun
  //-------------------------------------------------------------------------

  /**
   * The longitude of the sun at the time specified by this object.
   * The longitude is measured in radians along the ecliptic
   * from the "first point of Aries," the point at which the ecliptic
   * crosses the earth's equatorial plane at the vernal equinox.
   * <p>
   * Currently, this method uses an approximation of the two-body Kepler's
   * equation for the earth and the sun.  It does not take into account the
   * perturbations caused by the other planets, the moon, etc.
   * @internal
   */
  double getSunLongitude();

  /**
   * TODO Make this public when the entire class is package-private.
   */
  /*public*/ void getSunLongitude(double julianDay, double &longitude, double &meanAnomaly);

  /**
   * The position of the sun at this object's current date and time,
   * in equatorial coordinates.
   * @param result fillin for the result
   * @internal
   */
  Equatorial& getSunPosition(Equatorial& result);

public:
  /**
   * Constant representing the vernal equinox.
   * For use with {@link #getSunTime getSunTime}.
   * Note: In this case, "vernal" refers to the northern hemisphere's seasons.
   * @internal
   */
//  static double VERNAL_EQUINOX();

  /**
   * Constant representing the summer solstice.
   * For use with {@link #getSunTime getSunTime}.
   * Note: In this case, "summer" refers to the northern hemisphere's seasons.
   * @internal
   */
  static double SUMMER_SOLSTICE();

  /**
   * Constant representing the autumnal equinox.
   * For use with {@link #getSunTime getSunTime}.
   * Note: In this case, "autumn" refers to the northern hemisphere's seasons.
   * @internal
   */
//  static double AUTUMN_EQUINOX();

  /**
   * Constant representing the winter solstice.
   * For use with {@link #getSunTime getSunTime}.
   * Note: In this case, "winter" refers to the northern hemisphere's seasons.
   * @internal
   */
  static double WINTER_SOLSTICE();

  /**
   * Find the next time at which the sun's ecliptic longitude will have
   * the desired value.
   * @internal
   */
  UDate getSunTime(double desired, UBool next);

  /**
   * Returns the time (GMT) of sunrise or sunset on the local date to which
   * this calendar is currently set.
   *
   * NOTE: This method only works well if this object is set to a
   * time near local noon.  Because of variations between the local
   * official time zone and the geographic longitude, the
   * computation can flop over into an adjacent day if this object
   * is set to a time near local midnight.
   *
   * @internal
   */
  UDate getSunRiseSet(UBool rise);

  //-------------------------------------------------------------------------
  // The Moon
  //-------------------------------------------------------------------------

  /**
   * The position of the moon at the time set on this
   * object, in equatorial coordinates.
   * @internal
   * @return const reference to internal field of calendar astronomer. Do not use outside of the lifetime of this astronomer.
   */
  const Equatorial& getMoonPosition();

  /**
   * The "age" of the moon at the time specified in this object.
   * This is really the angle between the
   * current ecliptic longitudes of the sun and the moon,
   * measured in radians.
   *
   * @see #getMoonPhase
   * @internal
   */
  double getMoonAge();

  /**
   * Calculate the phase of the moon at the time set in this object.
   * The returned phase is a <code>double</code> in the range
   * <code>0 <= phase < 1</code>, interpreted as follows:
   * <ul>
   * <li>0.00: New moon
   * <li>0.25: First quarter
   * <li>0.50: Full moon
   * <li>0.75: Last quarter
   * </ul>
   *
   * @see #getMoonAge
   * @internal
   */
  double getMoonPhase();

  class U_I18N_API MoonAge : public UMemory {
  public:
    MoonAge(double l)
      :  value(l) { }
    void set(double l) { value = l; }
    double value;
  };

  /**
   * Constant representing a new moon.
   * For use with {@link #getMoonTime getMoonTime}
   * @internal
   */
  static const MoonAge NEW_MOON();

  /**
   * Constant representing the moon's first quarter.
   * For use with {@link #getMoonTime getMoonTime}
   * @internal
   */
//  static const MoonAge FIRST_QUARTER();

  /**
   * Constant representing a full moon.
   * For use with {@link #getMoonTime getMoonTime}
   * @internal
   */
  static const MoonAge FULL_MOON();

  /**
   * Constant representing the moon's last quarter.
   * For use with {@link #getMoonTime getMoonTime}
   * @internal
   */
//  static const MoonAge LAST_QUARTER();

  /**
   * Find the next or previous time at which the Moon's ecliptic
   * longitude will have the desired value.
   * <p>
   * @param desired   The desired longitude.
   * @param next      <tt>true</tt> if the next occurrance of the phase
   *                  is desired, <tt>false</tt> for the previous occurrance.
   * @internal
   */
  UDate getMoonTime(double desired, UBool next);
  UDate getMoonTime(const MoonAge& desired, UBool next);

  /**
   * Returns the time (GMT) of sunrise or sunset on the local date to which
   * this calendar is currently set.
   * @internal
   */
  UDate getMoonRiseSet(UBool rise);

  //-------------------------------------------------------------------------
  // Interpolation methods for finding the time at which a given event occurs
  //-------------------------------------------------------------------------

  // private
  class AngleFunc : public UMemory {
  public:
    virtual double eval(CalendarAstronomer&) = 0;
    virtual ~AngleFunc();
  };
  friend class AngleFunc;

  UDate timeOfAngle(AngleFunc& func, double desired,
                    double periodDays, double epsilon, UBool next);

  class CoordFunc : public UMemory {
  public:
    virtual void eval(Equatorial& result, CalendarAstronomer&) = 0;
    virtual ~CoordFunc();
  };
  friend class CoordFunc;

  double riseOrSet(CoordFunc& func, UBool rise,
                   double diameter, double refraction,
                   double epsilon);

  //-------------------------------------------------------------------------
  // Other utility methods
  //-------------------------------------------------------------------------
private:

  /**
   * Return the obliquity of the ecliptic (the angle between the ecliptic
   * and the earth's equator) at the current time.  This varies due to
   * the precession of the earth's axis.
   *
   * @return  the obliquity of the ecliptic relative to the equator,
   *          measured in radians.
   */
  double eclipticObliquity();

  //-------------------------------------------------------------------------
  // Private data
  //-------------------------------------------------------------------------
private:
  /**
   * Current time in milliseconds since 1/1/1970 AD
   * @see java.util.Date#getTime
   */
  UDate fTime;

  /* These aren't used yet, but they'll be needed for sunset calculations
   * and equatorial to horizon coordinate conversions
   */
  double fLongitude;
  double fLatitude;
  double fGmtOffset;

  //
  // The following fields are used to cache calculated results for improved
  // performance.  These values all depend on the current time setting
  // of this object, so the clearCache method is provided.
  //

  double    julianDay;
  double    julianCentury;
  double    sunLongitude;
  double    meanAnomalySun;
  double    moonLongitude;
  double    moonEclipLong;
  double    meanAnomalyMoon;
  double    eclipObliquity;
  double    siderealT0;
  double    siderealTime;

  void clearCache();

  Equatorial  moonPosition;
  UBool       moonPositionSet;

  /**
   * @internal
   */
//  UDate local(UDate localMillis);
};

U_NAMESPACE_END

struct UHashtable;

U_NAMESPACE_BEGIN

/**
 * Cache of month -> julian day
 * @internal
 */
class CalendarCache : public UMemory {
public:
  static int32_t get(CalendarCache** cache, int32_t key, UErrorCode &status);
  static void put(CalendarCache** cache, int32_t key, int32_t value, UErrorCode &status);
  virtual ~CalendarCache();
private:
  CalendarCache(int32_t size, UErrorCode& status);
  static void createCache(CalendarCache** cache, UErrorCode& status);
  /**
   * not implemented
   */
  CalendarCache();
  UHashtable *fTable;
};

U_NAMESPACE_END

#endif


#if !UCONFIG_NO_FORMATTING


#include <math.h>
#include <float.h>



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2001-2016, International Business Machines
*                Corporation and others. All Rights Reserved.
******************************************************************************
*   file name:  ucln_in.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001July05
*   created by: George Rhoten
*/





/*
Please keep the order of enums declared in same order
as the functions are suppose to be called.
It's usually best to have child dependencies called first. */
typedef enum ECleanupI18NType {
    UCLN_I18N_START = -1,
    UCLN_I18N_NUMBER_SKELETONS,
    UCLN_I18N_CURRENCY_SPACING,
    UCLN_I18N_SPOOF,
    UCLN_I18N_SPOOFDATA,
    UCLN_I18N_TRANSLITERATOR,
    UCLN_I18N_REGEX,
    UCLN_I18N_JAPANESE_CALENDAR,
    UCLN_I18N_ISLAMIC_CALENDAR,
    UCLN_I18N_CHINESE_CALENDAR,
    UCLN_I18N_HEBREW_CALENDAR,
    UCLN_I18N_ASTRO_CALENDAR,
    UCLN_I18N_DANGI_CALENDAR,
    UCLN_I18N_CALENDAR,
    UCLN_I18N_TIMEZONEFORMAT,
    UCLN_I18N_TZDBTIMEZONENAMES,
    UCLN_I18N_TIMEZONEGENERICNAMES,
    UCLN_I18N_TIMEZONENAMES,
    UCLN_I18N_ZONEMETA,
    UCLN_I18N_TIMEZONE,
    UCLN_I18N_DIGITLIST,
    UCLN_I18N_DECFMT,
    UCLN_I18N_NUMFMT,
    UCLN_I18N_ALLOWED_HOUR_FORMATS,
    UCLN_I18N_DAYPERIODRULES,
    UCLN_I18N_SMPDTFMT,
    UCLN_I18N_USEARCH,
    UCLN_I18N_COLLATOR,
    UCLN_I18N_UCOL_RES,
    UCLN_I18N_CSDET,
    UCLN_I18N_COLLATION_ROOT,
    UCLN_I18N_GENDERINFO,
    UCLN_I18N_CDFINFO,
    UCLN_I18N_REGION,
    UCLN_I18N_LIST_FORMATTER,
    UCLN_I18N_NUMSYS,
    UCLN_I18N_COUNT /* This must be last */
} ECleanupI18NType;

/* Main library cleanup registration function. */
/* See common/ucln.h for details on adding a cleanup function. */
/* Note: the global mutex must not be held when calling this function. */
U_CFUNC void U_EXPORT2 ucln_i18n_registerCleanup(ECleanupI18NType type,
                                                 cleanupFunc *func);



#include <stdio.h>  // for toString()

#if defined (PI)
#undef PI
#endif

#ifdef U_DEBUG_ASTRO
// # include "uresimp.h" // for debugging

// static void debug_astro_loc(const char *f, int32_t l)
// {
//   fprintf(stderr, "%s:%d: ", f, l);
// }

// static void debug_astro_msg(const char *pat, ...)
// {
//   va_list ap;
//   va_start(ap, pat);
//   vfprintf(stderr, pat, ap);
//   fflush(stderr);
// }
// #include "unicode/datefmt.h"
// #include "unicode/ustring.h"
// static const char * debug_astro_date(UDate d) {
//   static char gStrBuf[1024];
//   static DateFormat *df = NULL;
//   if(df == NULL) {
//     df = DateFormat::createDateTimeInstance(DateFormat::MEDIUM, DateFormat::MEDIUM, Locale::getUS());
//     df->adoptTimeZone(TimeZone::getGMT()->clone());
//   }
//   UnicodeString str;
//   df->format(d,str);
//   u_austrncpy(gStrBuf,str.getTerminatedBuffer(),sizeof(gStrBuf)-1);
//   return gStrBuf;
// }

// // must use double parens, i.e.:  U_DEBUG_ASTRO_MSG(("four is: %d",4));
// #define U_DEBUG_ASTRO_MSG(x) {debug_astro_loc(__FILE__,__LINE__);debug_astro_msg x;}
#else
#define U_DEBUG_ASTRO_MSG(x)
#endif

static inline UBool isINVALID(double d) {
  return(uprv_isNaN(d));
}

static icu::UMutex ccLock;

U_CDECL_BEGIN
static UBool calendar_astro_cleanup(void) {
  return TRUE;
}
U_CDECL_END

U_NAMESPACE_BEGIN

/**
 * The number of standard hours in one sidereal day.
 * Approximately 24.93.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
#define SIDEREAL_DAY (23.93446960027)

/**
 * The number of sidereal hours in one mean solar day.
 * Approximately 24.07.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
#define SOLAR_DAY  (24.065709816)

/**
 * The average number of solar days from one new moon to the next.  This is the time
 * it takes for the moon to return the same ecliptic longitude as the sun.
 * It is longer than the sidereal month because the sun's longitude increases
 * during the year due to the revolution of the earth around the sun.
 * Approximately 29.53.
 *
 * @see #SIDEREAL_MONTH
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
const double CalendarAstronomer::SYNODIC_MONTH  = 29.530588853;

/**
 * The average number of days it takes
 * for the moon to return to the same ecliptic longitude relative to the
 * stellar background.  This is referred to as the sidereal month.
 * It is shorter than the synodic month due to
 * the revolution of the earth around the sun.
 * Approximately 27.32.
 *
 * @see #SYNODIC_MONTH
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
#define SIDEREAL_MONTH  27.32166

/**
 * The average number number of days between successive vernal equinoxes.
 * Due to the precession of the earth's
 * axis, this is not precisely the same as the sidereal year.
 * Approximately 365.24
 *
 * @see #SIDEREAL_YEAR
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
#define TROPICAL_YEAR  365.242191

/**
 * The average number of days it takes
 * for the sun to return to the same position against the fixed stellar
 * background.  This is the duration of one orbit of the earth about the sun
 * as it would appear to an outside observer.
 * Due to the precession of the earth's
 * axis, this is not precisely the same as the tropical year.
 * Approximately 365.25.
 *
 * @see #TROPICAL_YEAR
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
#define SIDEREAL_YEAR  365.25636

//-------------------------------------------------------------------------
// Time-related constants
//-------------------------------------------------------------------------

/**
 * The number of milliseconds in one second.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
#define SECOND_MS  U_MILLIS_PER_SECOND

/**
 * The number of milliseconds in one minute.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
#define MINUTE_MS  U_MILLIS_PER_MINUTE

/**
 * The number of milliseconds in one hour.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
#define HOUR_MS   U_MILLIS_PER_HOUR

/**
 * The number of milliseconds in one day.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
#define DAY_MS U_MILLIS_PER_DAY

/**
 * The start of the julian day numbering scheme used by astronomers, which
 * is 1/1/4713 BC (Julian), 12:00 GMT.  This is given as the number of milliseconds
 * since 1/1/1970 AD (Gregorian), a negative number.
 * Note that julian day numbers and
 * the Julian calendar are <em>not</em> the same thing.  Also note that
 * julian days start at <em>noon</em>, not midnight.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
#define JULIAN_EPOCH_MS  -210866760000000.0


/**
 * Milliseconds value for 0.0 January 2000 AD.
 */
#define EPOCH_2000_MS  946598400000.0

//-------------------------------------------------------------------------
// Assorted private data used for conversions
//-------------------------------------------------------------------------

// My own copies of these so compilers are more likely to optimize them away
const double CalendarAstronomer::PI = 3.14159265358979323846;

#define CalendarAstronomer_PI2  (CalendarAstronomer::PI*2.0)
#define RAD_HOUR  ( 12 / CalendarAstronomer::PI )     // radians -> hours
#define DEG_RAD ( CalendarAstronomer::PI / 180 )      // degrees -> radians
#define RAD_DEG  ( 180 / CalendarAstronomer::PI )     // radians -> degrees

/***
 * Given 'value', add or subtract 'range' until 0 <= 'value' < range.
 * The modulus operator.
 */
inline static double normalize(double value, double range)  {
    return value - range * ClockMath::floorDivide(value, range);
}

/**
 * Normalize an angle so that it's in the range 0 - 2pi.
 * For positive angles this is just (angle % 2pi), but the Java
 * mod operator doesn't work that way for negative numbers....
 */
inline static double norm2PI(double angle)  {
    return normalize(angle, CalendarAstronomer::PI * 2.0);
}

/**
 * Normalize an angle into the range -PI - PI
 */
inline static  double normPI(double angle)  {
    return normalize(angle + CalendarAstronomer::PI, CalendarAstronomer::PI * 2.0) - CalendarAstronomer::PI;
}

//-------------------------------------------------------------------------
// Constructors
//-------------------------------------------------------------------------

/**
 * Construct a new <code>CalendarAstronomer</code> object that is initialized to
 * the current date and time.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
CalendarAstronomer::CalendarAstronomer():
  fTime(Calendar::getNow()), fLongitude(0.0), fLatitude(0.0), fGmtOffset(0.0), moonPosition(0,0), moonPositionSet(FALSE) {
  clearCache();
}

/**
 * Construct a new <code>CalendarAstronomer</code> object that is initialized to
 * the specified date and time.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
CalendarAstronomer::CalendarAstronomer(UDate d): fTime(d), fLongitude(0.0), fLatitude(0.0), fGmtOffset(0.0), moonPosition(0,0), moonPositionSet(FALSE) {
  clearCache();
}

/**
 * Construct a new <code>CalendarAstronomer</code> object with the given
 * latitude and longitude.  The object's time is set to the current
 * date and time.
 * <p>
 * @param longitude The desired longitude, in <em>degrees</em> east of
 *                  the Greenwich meridian.
 *
 * @param latitude  The desired latitude, in <em>degrees</em>.  Positive
 *                  values signify North, negative South.
 *
 * @see java.util.Date#getTime()
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
CalendarAstronomer::CalendarAstronomer(double longitude, double latitude) :
  fTime(Calendar::getNow()), moonPosition(0,0), moonPositionSet(FALSE) {
  fLongitude = normPI(longitude * (double)DEG_RAD);
  fLatitude  = normPI(latitude  * (double)DEG_RAD);
  fGmtOffset = (double)(fLongitude * 24. * (double)HOUR_MS / (double)CalendarAstronomer_PI2);
  clearCache();
}

CalendarAstronomer::~CalendarAstronomer()
{
}

//-------------------------------------------------------------------------
// Time and date getters and setters
//-------------------------------------------------------------------------

/**
 * Set the current date and time of this <code>CalendarAstronomer</code> object.  All
 * astronomical calculations are performed based on this time setting.
 *
 * @param aTime the date and time, expressed as the number of milliseconds since
 *              1/1/1970 0:00 GMT (Gregorian).
 *
 * @see #setDate
 * @see #getTime
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
void CalendarAstronomer::setTime(UDate aTime) {
    fTime = aTime;
    U_DEBUG_ASTRO_MSG(("setTime(%.1lf, %sL)\n", aTime, debug_astro_date(aTime+fGmtOffset)));
    clearCache();
}

/**
 * Set the current date and time of this <code>CalendarAstronomer</code> object.  All
 * astronomical calculations are performed based on this time setting.
 *
 * @param jdn   the desired time, expressed as a "julian day number",
 *              which is the number of elapsed days since
 *              1/1/4713 BC (Julian), 12:00 GMT.  Note that julian day
 *              numbers start at <em>noon</em>.  To get the jdn for
 *              the corresponding midnight, subtract 0.5.
 *
 * @see #getJulianDay
 * @see #JULIAN_EPOCH_MS
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
void CalendarAstronomer::setJulianDay(double jdn) {
    fTime = (double)(jdn * DAY_MS) + JULIAN_EPOCH_MS;
    clearCache();
    julianDay = jdn;
}

/**
 * Get the current time of this <code>CalendarAstronomer</code> object,
 * represented as the number of milliseconds since
 * 1/1/1970 AD 0:00 GMT (Gregorian).
 *
 * @see #setTime
 * @see #getDate
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
UDate CalendarAstronomer::getTime() {
    return fTime;
}

/**
 * Get the current time of this <code>CalendarAstronomer</code> object,
 * expressed as a "julian day number", which is the number of elapsed
 * days since 1/1/4713 BC (Julian), 12:00 GMT.
 *
 * @see #setJulianDay
 * @see #JULIAN_EPOCH_MS
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
double CalendarAstronomer::getJulianDay() {
    if (isINVALID(julianDay)) {
        julianDay = (fTime - (double)JULIAN_EPOCH_MS) / (double)DAY_MS;
    }
    return julianDay;
}

/**
 * Return this object's time expressed in julian centuries:
 * the number of centuries after 1/1/1900 AD, 12:00 GMT
 *
 * @see #getJulianDay
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
double CalendarAstronomer::getJulianCentury() {
    if (isINVALID(julianCentury)) {
        julianCentury = (getJulianDay() - 2415020.0) / 36525.0;
    }
    return julianCentury;
}

/**
 * Returns the current Greenwich sidereal time, measured in hours
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
double CalendarAstronomer::getGreenwichSidereal() {
    if (isINVALID(siderealTime)) {
        // See page 86 of "Practial Astronomy with your Calculator",
        // by Peter Duffet-Smith, for details on the algorithm.

        double UT = normalize(fTime/(double)HOUR_MS, 24.);

        siderealTime = normalize(getSiderealOffset() + UT*1.002737909, 24.);
    }
    return siderealTime;
}

double CalendarAstronomer::getSiderealOffset() {
    if (isINVALID(siderealT0)) {
        double JD  = uprv_floor(getJulianDay() - 0.5) + 0.5;
        double S   = JD - 2451545.0;
        double T   = S / 36525.0;
        siderealT0 = normalize(6.697374558 + 2400.051336*T + 0.000025862*T*T, 24);
    }
    return siderealT0;
}

/**
 * Returns the current local sidereal time, measured in hours
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
double CalendarAstronomer::getLocalSidereal() {
    return normalize(getGreenwichSidereal() + (fGmtOffset/(double)HOUR_MS), 24.);
}

/**
 * Converts local sidereal time to Universal Time.
 *
 * @param lst   The Local Sidereal Time, in hours since sidereal midnight
 *              on this object's current date.
 *
 * @return      The corresponding Universal Time, in milliseconds since
 *              1 Jan 1970, GMT.
 */
double CalendarAstronomer::lstToUT(double lst) {
    // Convert to local mean time
    double lt = normalize((lst - getSiderealOffset()) * 0.9972695663, 24);

    // Then find local midnight on this day
    double base = (DAY_MS * ClockMath::floorDivide(fTime + fGmtOffset,(double)DAY_MS)) - fGmtOffset;

    //out("    lt  =" + lt + " hours");
    //out("    base=" + new Date(base));

    return base + (long)(lt * HOUR_MS);
}


//-------------------------------------------------------------------------
// Coordinate transformations, all based on the current time of this object
//-------------------------------------------------------------------------

/**
 * Convert from ecliptic to equatorial coordinates.
 *
 * @param ecliptic  A point in the sky in ecliptic coordinates.
 * @return          The corresponding point in equatorial coordinates.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
CalendarAstronomer::Equatorial& CalendarAstronomer::eclipticToEquatorial(CalendarAstronomer::Equatorial& result, const CalendarAstronomer::Ecliptic& ecliptic)
{
    return eclipticToEquatorial(result, ecliptic.longitude, ecliptic.latitude);
}

/**
 * Convert from ecliptic to equatorial coordinates.
 *
 * @param eclipLong     The ecliptic longitude
 * @param eclipLat      The ecliptic latitude
 *
 * @return              The corresponding point in equatorial coordinates.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
CalendarAstronomer::Equatorial& CalendarAstronomer::eclipticToEquatorial(CalendarAstronomer::Equatorial& result, double eclipLong, double eclipLat)
{
    // See page 42 of "Practial Astronomy with your Calculator",
    // by Peter Duffet-Smith, for details on the algorithm.

    double obliq = eclipticObliquity();
    double sinE = ::sin(obliq);
    double cosE = cos(obliq);

    double sinL = ::sin(eclipLong);
    double cosL = cos(eclipLong);

    double sinB = ::sin(eclipLat);
    double cosB = cos(eclipLat);
    double tanB = tan(eclipLat);

    result.set(atan2(sinL*cosE - tanB*sinE, cosL),
        asin(sinB*cosE + cosB*sinE*sinL) );
    return result;
}

/**
 * Convert from ecliptic longitude to equatorial coordinates.
 *
 * @param eclipLong     The ecliptic longitude
 *
 * @return              The corresponding point in equatorial coordinates.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
CalendarAstronomer::Equatorial& CalendarAstronomer::eclipticToEquatorial(CalendarAstronomer::Equatorial& result, double eclipLong)
{
    return eclipticToEquatorial(result, eclipLong, 0);  // TODO: optimize
}

/**
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
CalendarAstronomer::Horizon& CalendarAstronomer::eclipticToHorizon(CalendarAstronomer::Horizon& result, double eclipLong)
{
    Equatorial equatorial;
    eclipticToEquatorial(equatorial, eclipLong);

    double H = getLocalSidereal()*CalendarAstronomer::PI/12 - equatorial.ascension;     // Hour-angle

    double sinH = ::sin(H);
    double cosH = cos(H);
    double sinD = ::sin(equatorial.declination);
    double cosD = cos(equatorial.declination);
    double sinL = ::sin(fLatitude);
    double cosL = cos(fLatitude);

    double altitude = asin(sinD*sinL + cosD*cosL*cosH);
    double azimuth  = atan2(-cosD*cosL*sinH, sinD - sinL * ::sin(altitude));

    result.set(azimuth, altitude);
    return result;
}


//-------------------------------------------------------------------------
// The Sun
//-------------------------------------------------------------------------

//
// Parameters of the Sun's orbit as of the epoch Jan 0.0 1990
// Angles are in radians (after multiplying by CalendarAstronomer::PI/180)
//
#define JD_EPOCH  2447891.5 // Julian day of epoch

#define SUN_ETA_G    (279.403303 * CalendarAstronomer::PI/180) // Ecliptic longitude at epoch
#define SUN_OMEGA_G  (282.768422 * CalendarAstronomer::PI/180) // Ecliptic longitude of perigee
#define SUN_E         0.016713          // Eccentricity of orbit
//double sunR0        1.495585e8        // Semi-major axis in KM
//double sunTheta0    (0.533128 * CalendarAstronomer::PI/180) // Angular diameter at R0

// The following three methods, which compute the sun parameters
// given above for an arbitrary epoch (whatever time the object is
// set to), make only a small difference as compared to using the
// above constants.  E.g., Sunset times might differ by ~12
// seconds.  Furthermore, the eta-g computation is befuddled by
// Duffet-Smith's incorrect coefficients (p.86).  I've corrected
// the first-order coefficient but the others may be off too - no
// way of knowing without consulting another source.

//  /**
//   * Return the sun's ecliptic longitude at perigee for the current time.
//   * See Duffett-Smith, p. 86.
//   * @return radians
//   */
//  private double getSunOmegaG() {
//      double T = getJulianCentury();
//      return (281.2208444 + (1.719175 + 0.000452778*T)*T) * DEG_RAD;
//  }

//  /**
//   * Return the sun's ecliptic longitude for the current time.
//   * See Duffett-Smith, p. 86.
//   * @return radians
//   */
//  private double getSunEtaG() {
//      double T = getJulianCentury();
//      //return (279.6966778 + (36000.76892 + 0.0003025*T)*T) * DEG_RAD;
//      //
//      // The above line is from Duffett-Smith, and yields manifestly wrong
//      // results.  The below constant is derived empirically to match the
//      // constant he gives for the 1990 EPOCH.
//      //
//      return (279.6966778 + (-0.3262541582718024 + 0.0003025*T)*T) * DEG_RAD;
//  }

//  /**
//   * Return the sun's eccentricity of orbit for the current time.
//   * See Duffett-Smith, p. 86.
//   * @return double
//   */
//  private double getSunE() {
//      double T = getJulianCentury();
//      return 0.01675104 - (0.0000418 + 0.000000126*T)*T;
//  }

/**
 * Find the "true anomaly" (longitude) of an object from
 * its mean anomaly and the eccentricity of its orbit.  This uses
 * an iterative solution to Kepler's equation.
 *
 * @param meanAnomaly   The object's longitude calculated as if it were in
 *                      a regular, circular orbit, measured in radians
 *                      from the point of perigee.
 *
 * @param eccentricity  The eccentricity of the orbit
 *
 * @return The true anomaly (longitude) measured in radians
 */
static double trueAnomaly(double meanAnomaly, double eccentricity)
{
    // First, solve Kepler's equation iteratively
    // Duffett-Smith, p.90
    double delta;
    double E = meanAnomaly;
    do {
        delta = E - eccentricity * ::sin(E) - meanAnomaly;
        E = E - delta / (1 - eccentricity * ::cos(E));
    }
    while (uprv_fabs(delta) > 1e-5); // epsilon = 1e-5 rad

    return 2.0 * ::atan( ::tan(E/2) * ::sqrt( (1+eccentricity)
                                             /(1-eccentricity) ) );
}

/**
 * The longitude of the sun at the time specified by this object.
 * The longitude is measured in radians along the ecliptic
 * from the "first point of Aries," the point at which the ecliptic
 * crosses the earth's equatorial plane at the vernal equinox.
 * <p>
 * Currently, this method uses an approximation of the two-body Kepler's
 * equation for the earth and the sun.  It does not take into account the
 * perturbations caused by the other planets, the moon, etc.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
double CalendarAstronomer::getSunLongitude()
{
    // See page 86 of "Practial Astronomy with your Calculator",
    // by Peter Duffet-Smith, for details on the algorithm.

    if (isINVALID(sunLongitude)) {
        getSunLongitude(getJulianDay(), sunLongitude, meanAnomalySun);
    }
    return sunLongitude;
}

/**
 * TODO Make this public when the entire class is package-private.
 */
/*public*/ void CalendarAstronomer::getSunLongitude(double jDay, double &longitude, double &meanAnomaly)
{
    // See page 86 of "Practial Astronomy with your Calculator",
    // by Peter Duffet-Smith, for details on the algorithm.

    double day = jDay - JD_EPOCH;       // Days since epoch

    // Find the angular distance the sun in a fictitious
    // circular orbit has travelled since the epoch.
    double epochAngle = norm2PI(CalendarAstronomer_PI2/TROPICAL_YEAR*day);

    // The epoch wasn't at the sun's perigee; find the angular distance
    // since perigee, which is called the "mean anomaly"
    meanAnomaly = norm2PI(epochAngle + SUN_ETA_G - SUN_OMEGA_G);

    // Now find the "true anomaly", e.g. the real solar longitude
    // by solving Kepler's equation for an elliptical orbit
    // NOTE: The 3rd ed. of the book lists omega_g and eta_g in different
    // equations; omega_g is to be correct.
    longitude =  norm2PI(trueAnomaly(meanAnomaly, SUN_E) + SUN_OMEGA_G);
}

/**
 * The position of the sun at this object's current date and time,
 * in equatorial coordinates.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
CalendarAstronomer::Equatorial& CalendarAstronomer::getSunPosition(CalendarAstronomer::Equatorial& result) {
    return eclipticToEquatorial(result, getSunLongitude(), 0);
}


/**
 * Constant representing the vernal equinox.
 * For use with {@link #getSunTime getSunTime}.
 * Note: In this case, "vernal" refers to the northern hemisphere's seasons.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
/*double CalendarAstronomer::VERNAL_EQUINOX() {
  return 0;
}*/

/**
 * Constant representing the summer solstice.
 * For use with {@link #getSunTime getSunTime}.
 * Note: In this case, "summer" refers to the northern hemisphere's seasons.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
double CalendarAstronomer::SUMMER_SOLSTICE() {
    return  (CalendarAstronomer::PI/2);
}

/**
 * Constant representing the autumnal equinox.
 * For use with {@link #getSunTime getSunTime}.
 * Note: In this case, "autumn" refers to the northern hemisphere's seasons.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
/*double CalendarAstronomer::AUTUMN_EQUINOX() {
  return  (CalendarAstronomer::PI);
}*/

/**
 * Constant representing the winter solstice.
 * For use with {@link #getSunTime getSunTime}.
 * Note: In this case, "winter" refers to the northern hemisphere's seasons.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
double CalendarAstronomer::WINTER_SOLSTICE() {
    return  ((CalendarAstronomer::PI*3)/2);
}

CalendarAstronomer::AngleFunc::~AngleFunc() {}

/**
 * Find the next time at which the sun's ecliptic longitude will have
 * the desired value.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
class SunTimeAngleFunc : public CalendarAstronomer::AngleFunc {
public:
    virtual ~SunTimeAngleFunc();
    virtual double eval(CalendarAstronomer& a) { return a.getSunLongitude(); }
};

SunTimeAngleFunc::~SunTimeAngleFunc() {}

UDate CalendarAstronomer::getSunTime(double desired, UBool next)
{
    SunTimeAngleFunc func;
    return timeOfAngle( func,
                        desired,
                        TROPICAL_YEAR,
                        MINUTE_MS,
                        next);
}

CalendarAstronomer::CoordFunc::~CoordFunc() {}

class RiseSetCoordFunc : public CalendarAstronomer::CoordFunc {
public:
    virtual ~RiseSetCoordFunc();
    virtual void eval(CalendarAstronomer::Equatorial& result, CalendarAstronomer&a) {  a.getSunPosition(result); }
};

RiseSetCoordFunc::~RiseSetCoordFunc() {}

UDate CalendarAstronomer::getSunRiseSet(UBool rise)
{
    UDate t0 = fTime;

    // Make a rough guess: 6am or 6pm local time on the current day
    double noon = ClockMath::floorDivide(fTime + fGmtOffset, (double)DAY_MS)*DAY_MS - fGmtOffset + (12*HOUR_MS);

    U_DEBUG_ASTRO_MSG(("Noon=%.2lf, %sL, gmtoff %.2lf\n", noon, debug_astro_date(noon+fGmtOffset), fGmtOffset));
    setTime(noon +  ((rise ? -6 : 6) * HOUR_MS));
    U_DEBUG_ASTRO_MSG(("added %.2lf ms as a guess,\n", ((rise ? -6. : 6.) * HOUR_MS)));

    RiseSetCoordFunc func;
    double t = riseOrSet(func,
                         rise,
                         .533 * DEG_RAD,        // Angular Diameter
                         34. /60.0 * DEG_RAD,    // Refraction correction
                         MINUTE_MS / 12.);       // Desired accuracy

    setTime(t0);
    return t;
}

// Commented out - currently unused. ICU 2.6, Alan
//    //-------------------------------------------------------------------------
//    // Alternate Sun Rise/Set
//    // See Duffett-Smith p.93
//    //-------------------------------------------------------------------------
//
//    // This yields worse results (as compared to USNO data) than getSunRiseSet().
//    /**
//     * TODO Make this when the entire class is package-private.
//     */
//    /*public*/ long getSunRiseSet2(boolean rise) {
//        // 1. Calculate coordinates of the sun's center for midnight
//        double jd = uprv_floor(getJulianDay() - 0.5) + 0.5;
//        double[] sl = getSunLongitude(jd);//        double lambda1 = sl[0];
//        Equatorial pos1 = eclipticToEquatorial(lambda1, 0);
//
//        // 2. Add ... to lambda to get position 24 hours later
//        double lambda2 = lambda1 + 0.985647*DEG_RAD;
//        Equatorial pos2 = eclipticToEquatorial(lambda2, 0);
//
//        // 3. Calculate LSTs of rising and setting for these two positions
//        double tanL = ::tan(fLatitude);
//        double H = ::acos(-tanL * ::tan(pos1.declination));
//        double lst1r = (CalendarAstronomer_PI2 + pos1.ascension - H) * 24 / CalendarAstronomer_PI2;
//        double lst1s = (pos1.ascension + H) * 24 / CalendarAstronomer_PI2;
//               H = ::acos(-tanL * ::tan(pos2.declination));
//        double lst2r = (CalendarAstronomer_PI2-H + pos2.ascension ) * 24 / CalendarAstronomer_PI2;
//        double lst2s = (H + pos2.ascension ) * 24 / CalendarAstronomer_PI2;
//        if (lst1r > 24) lst1r -= 24;
//        if (lst1s > 24) lst1s -= 24;
//        if (lst2r > 24) lst2r -= 24;
//        if (lst2s > 24) lst2s -= 24;
//
//        // 4. Convert LSTs to GSTs.  If GST1 > GST2, add 24 to GST2.
//        double gst1r = lstToGst(lst1r);
//        double gst1s = lstToGst(lst1s);
//        double gst2r = lstToGst(lst2r);
//        double gst2s = lstToGst(lst2s);
//        if (gst1r > gst2r) gst2r += 24;
//        if (gst1s > gst2s) gst2s += 24;
//
//        // 5. Calculate GST at 0h UT of this date
//        double t00 = utToGst(0);
//
//        // 6. Calculate GST at 0h on the observer's longitude
//        double offset = ::round(fLongitude*12/PI); // p.95 step 6; he _rounds_ to nearest 15 deg.
//        double t00p = t00 - offset*1.002737909;
//        if (t00p < 0) t00p += 24; // do NOT normalize
//
//        // 7. Adjust
//        if (gst1r < t00p) {
//            gst1r += 24;
//            gst2r += 24;
//        }
//        if (gst1s < t00p) {
//            gst1s += 24;
//            gst2s += 24;
//        }
//
//        // 8.
//        double gstr = (24.07*gst1r-t00*(gst2r-gst1r))/(24.07+gst1r-gst2r);
//        double gsts = (24.07*gst1s-t00*(gst2s-gst1s))/(24.07+gst1s-gst2s);
//
//        // 9. Correct for parallax, refraction, and sun's diameter
//        double dec = (pos1.declination + pos2.declination) / 2;
//        double psi = ::acos(sin(fLatitude) / cos(dec));
//        double x = 0.830725 * DEG_RAD; // parallax+refraction+diameter
//        double y = ::asin(sin(x) / ::sin(psi)) * RAD_DEG;
//        double delta_t = 240 * y / cos(dec) / 3600; // hours
//
//        // 10. Add correction to GSTs, subtract from GSTr
//        gstr -= delta_t;
//        gsts += delta_t;
//
//        // 11. Convert GST to UT and then to local civil time
//        double ut = gstToUt(rise ? gstr : gsts);
//        //System.out.println((rise?"rise=":"set=") + ut + ", delta_t=" + delta_t);
//        long midnight = DAY_MS * (time / DAY_MS); // Find UT midnight on this day
//        return midnight + (long) (ut * 3600000);
//    }

// Commented out - currently unused. ICU 2.6, Alan
//    /**
//     * Convert local sidereal time to Greenwich sidereal time.
//     * Section 15.  Duffett-Smith p.21
//     * @param lst in hours (0..24)
//     * @return GST in hours (0..24)
//     */
//    double lstToGst(double lst) {
//        double delta = fLongitude * 24 / CalendarAstronomer_PI2;
//        return normalize(lst - delta, 24);
//    }

// Commented out - currently unused. ICU 2.6, Alan
//    /**
//     * Convert UT to GST on this date.
//     * Section 12.  Duffett-Smith p.17
//     * @param ut in hours
//     * @return GST in hours
//     */
//    double utToGst(double ut) {
//        return normalize(getT0() + ut*1.002737909, 24);
//    }

// Commented out - currently unused. ICU 2.6, Alan
//    /**
//     * Convert GST to UT on this date.
//     * Section 13.  Duffett-Smith p.18
//     * @param gst in hours
//     * @return UT in hours
//     */
//    double gstToUt(double gst) {
//        return normalize(gst - getT0(), 24) * 0.9972695663;
//    }

// Commented out - currently unused. ICU 2.6, Alan
//    double getT0() {
//        // Common computation for UT <=> GST
//
//        // Find JD for 0h UT
//        double jd = uprv_floor(getJulianDay() - 0.5) + 0.5;
//
//        double s = jd - 2451545.0;
//        double t = s / 36525.0;
//        double t0 = 6.697374558 + (2400.051336 + 0.000025862*t)*t;
//        return t0;
//    }

// Commented out - currently unused. ICU 2.6, Alan
//    //-------------------------------------------------------------------------
//    // Alternate Sun Rise/Set
//    // See sci.astro FAQ
//    // http://www.faqs.org/faqs/astronomy/faq/part3/section-5.html
//    //-------------------------------------------------------------------------
//
//    // Note: This method appears to produce inferior accuracy as
//    // compared to getSunRiseSet().
//
//    /**
//     * TODO Make this when the entire class is package-private.
//     */
//    /*public*/ long getSunRiseSet3(boolean rise) {
//
//        // Compute day number for 0.0 Jan 2000 epoch
//        double d = (double)(time - EPOCH_2000_MS) / DAY_MS;
//
//        // Now compute the Local Sidereal Time, LST:
//        //
//        double LST  =  98.9818  +  0.985647352 * d  +  /*UT*15  +  long*/
//            fLongitude*RAD_DEG;
//        //
//        // (east long. positive).  Note that LST is here expressed in degrees,
//        // where 15 degrees corresponds to one hour.  Since LST really is an angle,
//        // it's convenient to use one unit---degrees---throughout.
//
//        //    COMPUTING THE SUN'S POSITION
//        //    ----------------------------
//        //
//        // To be able to compute the Sun's rise/set times, you need to be able to
//        // compute the Sun's position at any time.  First compute the "day
//        // number" d as outlined above, for the desired moment.  Next compute:
//        //
//        double oblecl = 23.4393 - 3.563E-7 * d;
//        //
//        double w  =  282.9404  +  4.70935E-5   * d;
//        double M  =  356.0470  +  0.9856002585 * d;
//        double e  =  0.016709  -  1.151E-9     * d;
//        //
//        // This is the obliquity of the ecliptic, plus some of the elements of
//        // the Sun's apparent orbit (i.e., really the Earth's orbit): w =
//        // argument of perihelion, M = mean anomaly, e = eccentricity.
//        // Semi-major axis is here assumed to be exactly 1.0 (while not strictly
//        // true, this is still an accurate approximation).  Next compute E, the
//        // eccentric anomaly:
//        //
//        double E = M + e*(180/PI) * ::sin(M*DEG_RAD) * ( 1.0 + e*cos(M*DEG_RAD) );
//        //
//        // where E and M are in degrees.  This is it---no further iterations are
//        // needed because we know e has a sufficiently small value.  Next compute
//        // the true anomaly, v, and the distance, r:
//        //
//        /*      r * cos(v)  =  */ double A  =  cos(E*DEG_RAD) - e;
//        /*      r * ::sin(v)  =  */ double B  =  ::sqrt(1 - e*e) * ::sin(E*DEG_RAD);
//        //
//        // and
//        //
//        //      r  =  sqrt( A*A + B*B )
//        double v  =  ::atan2( B, A )*RAD_DEG;
//        //
//        // The Sun's true longitude, slon, can now be computed:
//        //
//        double slon  =  v + w;
//        //
//        // Since the Sun is always at the ecliptic (or at least very very close to
//        // it), we can use simplified formulae to convert slon (the Sun's ecliptic
//        // longitude) to sRA and sDec (the Sun's RA and Dec):
//        //
//        //                   ::sin(slon) * cos(oblecl)
//        //     tan(sRA)  =  -------------------------
//        //            cos(slon)
//        //
//        //     ::sin(sDec) =  ::sin(oblecl) * ::sin(slon)
//        //
//        // As was the case when computing az, the Azimuth, if possible use an
//        // atan2() function to compute sRA.
//
//        double sRA = ::atan2(sin(slon*DEG_RAD) * cos(oblecl*DEG_RAD), cos(slon*DEG_RAD))*RAD_DEG;
//
//        double sin_sDec = ::sin(oblecl*DEG_RAD) * ::sin(slon*DEG_RAD);
//        double sDec = ::asin(sin_sDec)*RAD_DEG;
//
//        //    COMPUTING RISE AND SET TIMES
//        //    ----------------------------
//        //
//        // To compute when an object rises or sets, you must compute when it
//        // passes the meridian and the HA of rise/set.  Then the rise time is
//        // the meridian time minus HA for rise/set, and the set time is the
//        // meridian time plus the HA for rise/set.
//        //
//        // To find the meridian time, compute the Local Sidereal Time at 0h local
//        // time (or 0h UT if you prefer to work in UT) as outlined above---name
//        // that quantity LST0.  The Meridian Time, MT, will now be:
//        //
//        //     MT  =  RA - LST0
//        double MT = normalize(sRA - LST, 360);
//        //
//        // where "RA" is the object's Right Ascension (in degrees!).  If negative,
//        // add 360 deg to MT.  If the object is the Sun, leave the time as it is,
//        // but if it's stellar, multiply MT by 365.2422/366.2422, to convert from
//        // sidereal to solar time.  Now, compute HA for rise/set, name that
//        // quantity HA0:
//        //
//        //                 ::sin(h0)  -  ::sin(lat) * ::sin(Dec)
//        // cos(HA0)  =  ---------------------------------
//        //                      cos(lat) * cos(Dec)
//        //
//        // where h0 is the altitude selected to represent rise/set.  For a purely
//        // mathematical horizon, set h0 = 0 and simplify to:
//        //
//        //    cos(HA0)  =  - tan(lat) * tan(Dec)
//        //
//        // If you want to account for refraction on the atmosphere, set h0 = -35/60
//        // degrees (-35 arc minutes), and if you want to compute the rise/set times
//        // for the Sun's upper limb, set h0 = -50/60 (-50 arc minutes).
//        //
//        double h0 = -50/60 * DEG_RAD;
//
//        double HA0 = ::acos(
//          (sin(h0) - ::sin(fLatitude) * sin_sDec) /
//          (cos(fLatitude) * cos(sDec*DEG_RAD)))*RAD_DEG;
//
//        // When HA0 has been computed, leave it as it is for the Sun but multiply
//        // by 365.2422/366.2422 for stellar objects, to convert from sidereal to
//        // solar time.  Finally compute:
//        //
//        //    Rise time  =  MT - HA0
//        //    Set  time  =  MT + HA0
//        //
//        // convert the times from degrees to hours by dividing by 15.
//        //
//        // If you'd like to check that your calculations are accurate or just
//        // need a quick result, check the USNO's Sun or Moon Rise/Set Table,
//        // <URL:http://aa.usno.navy.mil/AA/data/docs/RS_OneYear.html>.
//
//        double result = MT + (rise ? -HA0 : HA0); // in degrees
//
//        // Find UT midnight on this day
//        long midnight = DAY_MS * (time / DAY_MS);
//
//        return midnight + (long) (result * 3600000 / 15);
//    }

//-------------------------------------------------------------------------
// The Moon
//-------------------------------------------------------------------------

#define moonL0  (318.351648 * CalendarAstronomer::PI/180 )   // Mean long. at epoch
#define moonP0 ( 36.340410 * CalendarAstronomer::PI/180 )   // Mean long. of perigee
#define moonN0 ( 318.510107 * CalendarAstronomer::PI/180 )   // Mean long. of node
#define moonI  (   5.145366 * CalendarAstronomer::PI/180 )   // Inclination of orbit
#define moonE  (   0.054900 )            // Eccentricity of orbit

// These aren't used right now
#define moonA  (   3.84401e5 )           // semi-major axis (km)
#define moonT0 (   0.5181 * CalendarAstronomer::PI/180 )     // Angular size at distance A
#define moonPi (   0.9507 * CalendarAstronomer::PI/180 )     // Parallax at distance A

/**
 * The position of the moon at the time set on this
 * object, in equatorial coordinates.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
const CalendarAstronomer::Equatorial& CalendarAstronomer::getMoonPosition()
{
    //
    // See page 142 of "Practial Astronomy with your Calculator",
    // by Peter Duffet-Smith, for details on the algorithm.
    //
    if (moonPositionSet == FALSE) {
        // Calculate the solar longitude.  Has the side effect of
        // filling in "meanAnomalySun" as well.
        getSunLongitude();

        //
        // Find the # of days since the epoch of our orbital parameters.
        // TODO: Convert the time of day portion into ephemeris time
        //
        double day = getJulianDay() - JD_EPOCH;       // Days since epoch

        // Calculate the mean longitude and anomaly of the moon, based on
        // a circular orbit.  Similar to the corresponding solar calculation.
        double meanLongitude = norm2PI(13.1763966*PI/180*day + moonL0);
        meanAnomalyMoon = norm2PI(meanLongitude - 0.1114041*PI/180 * day - moonP0);

        //
        // Calculate the following corrections:
        //  Evection:   the sun's gravity affects the moon's eccentricity
        //  Annual Eqn: variation in the effect due to earth-sun distance
        //  A3:         correction factor (for ???)
        //
        double evection = 1.2739*PI/180 * ::sin(2 * (meanLongitude - sunLongitude)
            - meanAnomalyMoon);
        double annual   = 0.1858*PI/180 * ::sin(meanAnomalySun);
        double a3       = 0.3700*PI/180 * ::sin(meanAnomalySun);

        meanAnomalyMoon += evection - annual - a3;

        //
        // More correction factors:
        //  center  equation of the center correction
        //  a4      yet another error correction (???)
        //
        // TODO: Skip the equation of the center correction and solve Kepler's eqn?
        //
        double center = 6.2886*PI/180 * ::sin(meanAnomalyMoon);
        double a4 =     0.2140*PI/180 * ::sin(2 * meanAnomalyMoon);

        // Now find the moon's corrected longitude
        moonLongitude = meanLongitude + evection + center - annual + a4;

        //
        // And finally, find the variation, caused by the fact that the sun's
        // gravitational pull on the moon varies depending on which side of
        // the earth the moon is on
        //
        double variation = 0.6583*CalendarAstronomer::PI/180 * ::sin(2*(moonLongitude - sunLongitude));

        moonLongitude += variation;

        //
        // What we've calculated so far is the moon's longitude in the plane
        // of its own orbit.  Now map to the ecliptic to get the latitude
        // and longitude.  First we need to find the longitude of the ascending
        // node, the position on the ecliptic where it is crossed by the moon's
        // orbit as it crosses from the southern to the northern hemisphere.
        //
        double nodeLongitude = norm2PI(moonN0 - 0.0529539*PI/180 * day);

        nodeLongitude -= 0.16*PI/180 * ::sin(meanAnomalySun);

        double y = ::sin(moonLongitude - nodeLongitude);
        double x = cos(moonLongitude - nodeLongitude);

        moonEclipLong = ::atan2(y*cos(moonI), x) + nodeLongitude;
        double moonEclipLat = ::asin(y * ::sin(moonI));

        eclipticToEquatorial(moonPosition, moonEclipLong, moonEclipLat);
        moonPositionSet = TRUE;
    }
    return moonPosition;
}

/**
 * The "age" of the moon at the time specified in this object.
 * This is really the angle between the
 * current ecliptic longitudes of the sun and the moon,
 * measured in radians.
 *
 * @see #getMoonPhase
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
double CalendarAstronomer::getMoonAge() {
    // See page 147 of "Practial Astronomy with your Calculator",
    // by Peter Duffet-Smith, for details on the algorithm.
    //
    // Force the moon's position to be calculated.  We're going to use
    // some the intermediate results cached during that calculation.
    //
    getMoonPosition();

    return norm2PI(moonEclipLong - sunLongitude);
}

/**
 * Calculate the phase of the moon at the time set in this object.
 * The returned phase is a <code>double</code> in the range
 * <code>0 <= phase < 1</code>, interpreted as follows:
 * <ul>
 * <li>0.00: New moon
 * <li>0.25: First quarter
 * <li>0.50: Full moon
 * <li>0.75: Last quarter
 * </ul>
 *
 * @see #getMoonAge
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
double CalendarAstronomer::getMoonPhase() {
    // See page 147 of "Practial Astronomy with your Calculator",
    // by Peter Duffet-Smith, for details on the algorithm.
    return 0.5 * (1 - cos(getMoonAge()));
}

/**
 * Constant representing a new moon.
 * For use with {@link #getMoonTime getMoonTime}
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
const CalendarAstronomer::MoonAge CalendarAstronomer::NEW_MOON() {
    return  CalendarAstronomer::MoonAge(0);
}

/**
 * Constant representing the moon's first quarter.
 * For use with {@link #getMoonTime getMoonTime}
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
/*const CalendarAstronomer::MoonAge CalendarAstronomer::FIRST_QUARTER() {
  return   CalendarAstronomer::MoonAge(CalendarAstronomer::PI/2);
}*/

/**
 * Constant representing a full moon.
 * For use with {@link #getMoonTime getMoonTime}
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
const CalendarAstronomer::MoonAge CalendarAstronomer::FULL_MOON() {
    return   CalendarAstronomer::MoonAge(CalendarAstronomer::PI);
}
/**
 * Constant representing the moon's last quarter.
 * For use with {@link #getMoonTime getMoonTime}
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */

class MoonTimeAngleFunc : public CalendarAstronomer::AngleFunc {
public:
    virtual ~MoonTimeAngleFunc();
    virtual double eval(CalendarAstronomer&a) { return a.getMoonAge(); }
};

MoonTimeAngleFunc::~MoonTimeAngleFunc() {}

/*const CalendarAstronomer::MoonAge CalendarAstronomer::LAST_QUARTER() {
  return  CalendarAstronomer::MoonAge((CalendarAstronomer::PI*3)/2);
}*/

/**
 * Find the next or previous time at which the Moon's ecliptic
 * longitude will have the desired value.
 * <p>
 * @param desired   The desired longitude.
 * @param next      <tt>true</tt> if the next occurrance of the phase
 *                  is desired, <tt>false</tt> for the previous occurrance.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
UDate CalendarAstronomer::getMoonTime(double desired, UBool next)
{
    MoonTimeAngleFunc func;
    return timeOfAngle( func,
                        desired,
                        SYNODIC_MONTH,
                        MINUTE_MS,
                        next);
}

/**
 * Find the next or previous time at which the moon will be in the
 * desired phase.
 * <p>
 * @param desired   The desired phase of the moon.
 * @param next      <tt>true</tt> if the next occurrance of the phase
 *                  is desired, <tt>false</tt> for the previous occurrance.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
UDate CalendarAstronomer::getMoonTime(const CalendarAstronomer::MoonAge& desired, UBool next) {
    return getMoonTime(desired.value, next);
}

class MoonRiseSetCoordFunc : public CalendarAstronomer::CoordFunc {
public:
    virtual ~MoonRiseSetCoordFunc();
    virtual void eval(CalendarAstronomer::Equatorial& result, CalendarAstronomer&a) { result = a.getMoonPosition(); }
};

MoonRiseSetCoordFunc::~MoonRiseSetCoordFunc() {}

/**
 * Returns the time (GMT) of sunrise or sunset on the local date to which
 * this calendar is currently set.
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
UDate CalendarAstronomer::getMoonRiseSet(UBool rise)
{
    MoonRiseSetCoordFunc func;
    return riseOrSet(func,
                     rise,
                     .533 * DEG_RAD,        // Angular Diameter
                     34 /60.0 * DEG_RAD,    // Refraction correction
                     MINUTE_MS);            // Desired accuracy
}

//-------------------------------------------------------------------------
// Interpolation methods for finding the time at which a given event occurs
//-------------------------------------------------------------------------

UDate CalendarAstronomer::timeOfAngle(AngleFunc& func, double desired,
                                      double periodDays, double epsilon, UBool next)
{
    // Find the value of the function at the current time
    double lastAngle = func.eval(*this);

    // Find out how far we are from the desired angle
    double deltaAngle = norm2PI(desired - lastAngle) ;

    // Using the average period, estimate the next (or previous) time at
    // which the desired angle occurs.
    double deltaT =  (deltaAngle + (next ? 0.0 : - CalendarAstronomer_PI2 )) * (periodDays*DAY_MS) / CalendarAstronomer_PI2;

    double lastDeltaT = deltaT; // Liu
    UDate startTime = fTime; // Liu

    setTime(fTime + uprv_ceil(deltaT));

    // Now iterate until we get the error below epsilon.  Throughout
    // this loop we use normPI to get values in the range -Pi to Pi,
    // since we're using them as correction factors rather than absolute angles.
    do {
        // Evaluate the function at the time we've estimated
        double angle = func.eval(*this);

        // Find the # of milliseconds per radian at this point on the curve
        double factor = uprv_fabs(deltaT / normPI(angle-lastAngle));

        // Correct the time estimate based on how far off the angle is
        deltaT = normPI(desired - angle) * factor;

        // HACK:
        //
        // If abs(deltaT) begins to diverge we need to quit this loop.
        // This only appears to happen when attempting to locate, for
        // example, a new moon on the day of the new moon.  E.g.:
        //
        // This result is correct:
        // newMoon(7508(Mon Jul 23 00:00:00 CST 1990,false))=
        //   Sun Jul 22 10:57:41 CST 1990
        //
        // But attempting to make the same call a day earlier causes deltaT
        // to diverge:
        // CalendarAstronomer.timeOfAngle() diverging: 1.348508727575625E9 ->
        //   1.3649828540224032E9
        // newMoon(7507(Sun Jul 22 00:00:00 CST 1990,false))=
        //   Sun Jul 08 13:56:15 CST 1990
        //
        // As a temporary solution, we catch this specific condition and
        // adjust our start time by one eighth period days (either forward
        // or backward) and try again.
        // Liu 11/9/00
        if (uprv_fabs(deltaT) > uprv_fabs(lastDeltaT)) {
            double delta = uprv_ceil (periodDays * DAY_MS / 8.0);
            setTime(startTime + (next ? delta : -delta));
            return timeOfAngle(func, desired, periodDays, epsilon, next);
        }

        lastDeltaT = deltaT;
        lastAngle = angle;

        setTime(fTime + uprv_ceil(deltaT));
    }
    while (uprv_fabs(deltaT) > epsilon);

    return fTime;
}

UDate CalendarAstronomer::riseOrSet(CoordFunc& func, UBool rise,
                                    double diameter, double refraction,
                                    double epsilon)
{
    Equatorial pos;
    double      tanL   = ::tan(fLatitude);
    double     deltaT = 0;
    int32_t         count = 0;

    //
    // Calculate the object's position at the current time, then use that
    // position to calculate the time of rising or setting.  The position
    // will be different at that time, so iterate until the error is allowable.
    //
    U_DEBUG_ASTRO_MSG(("setup rise=%s, dia=%.3lf, ref=%.3lf, eps=%.3lf\n",
        rise?"T":"F", diameter, refraction, epsilon));
    do {
        // See "Practical Astronomy With Your Calculator, section 33.
        func.eval(pos, *this);
        double angle = ::acos(-tanL * ::tan(pos.declination));
        double lst = ((rise ? CalendarAstronomer_PI2-angle : angle) + pos.ascension ) * 24 / CalendarAstronomer_PI2;

        // Convert from LST to Universal Time.
        UDate newTime = lstToUT( lst );

        deltaT = newTime - fTime;
        setTime(newTime);
        U_DEBUG_ASTRO_MSG(("%d] dT=%.3lf, angle=%.3lf, lst=%.3lf,   A=%.3lf/D=%.3lf\n",
            count, deltaT, angle, lst, pos.ascension, pos.declination));
    }
    while (++ count < 5 && uprv_fabs(deltaT) > epsilon);

    // Calculate the correction due to refraction and the object's angular diameter
    double cosD  = ::cos(pos.declination);
    double psi   = ::acos(sin(fLatitude) / cosD);
    double x     = diameter / 2 + refraction;
    double y     = ::asin(sin(x) / ::sin(psi));
    long  delta  = (long)((240 * y * RAD_DEG / cosD)*SECOND_MS);

    return fTime + (rise ? -delta : delta);
}
											   /**
 * Return the obliquity of the ecliptic (the angle between the ecliptic
 * and the earth's equator) at the current time.  This varies due to
 * the precession of the earth's axis.
 *
 * @return  the obliquity of the ecliptic relative to the equator,
 *          measured in radians.
 */
double CalendarAstronomer::eclipticObliquity() {
    if (isINVALID(eclipObliquity)) {
        const double epoch = 2451545.0;     // 2000 AD, January 1.5

        double T = (getJulianDay() - epoch) / 36525;

        eclipObliquity = 23.439292
            - 46.815/3600 * T
            - 0.0006/3600 * T*T
            + 0.00181/3600 * T*T*T;

        eclipObliquity *= DEG_RAD;
    }
    return eclipObliquity;
}


//-------------------------------------------------------------------------
// Private data
//-------------------------------------------------------------------------
void CalendarAstronomer::clearCache() {
    const double INVALID = uprv_getNaN();

    julianDay       = INVALID;
    julianCentury   = INVALID;
    sunLongitude    = INVALID;
    meanAnomalySun  = INVALID;
    moonLongitude   = INVALID;
    moonEclipLong   = INVALID;
    meanAnomalyMoon = INVALID;
    eclipObliquity  = INVALID;
    siderealTime    = INVALID;
    siderealT0      = INVALID;
    moonPositionSet = FALSE;
}

//private static void out(String s) {
//    System.out.println(s);
//}

//private static String deg(double rad) {
//    return Double.toString(rad * RAD_DEG);
//}

//private static String hours(long ms) {
//    return Double.toString((double)ms / HOUR_MS) + " hours";
//}

/**
 * @internal
 * @deprecated ICU 2.4. This class may be removed or modified.
 */
/*UDate CalendarAstronomer::local(UDate localMillis) {
  // TODO - srl ?
  TimeZone *tz = TimeZone::createDefault();
  int32_t rawOffset;
  int32_t dstOffset;
  UErrorCode status = U_ZERO_ERROR;
  tz->getOffset(localMillis, TRUE, rawOffset, dstOffset, status);
  delete tz;
  return localMillis - rawOffset;
}*/

// Debugging functions
UnicodeString CalendarAstronomer::Ecliptic::toString() const
{
#ifdef U_DEBUG_ASTRO
    char tmp[800];
    sprintf(tmp, "[%.5f,%.5f]", longitude*RAD_DEG, latitude*RAD_DEG);
    return UnicodeString(tmp, "");
#else
    return UnicodeString();
#endif
}

UnicodeString CalendarAstronomer::Equatorial::toString() const
{
#ifdef U_DEBUG_ASTRO
    char tmp[400];
    sprintf(tmp, "%f,%f",
        (ascension*RAD_DEG), (declination*RAD_DEG));
    return UnicodeString(tmp, "");
#else
    return UnicodeString();
#endif
}

UnicodeString CalendarAstronomer::Horizon::toString() const
{
#ifdef U_DEBUG_ASTRO
    char tmp[800];
    sprintf(tmp, "[%.5f,%.5f]", altitude*RAD_DEG, azimuth*RAD_DEG);
    return UnicodeString(tmp, "");
#else
    return UnicodeString();
#endif
}


//  static private String radToHms(double angle) {
//    int hrs = (int) (angle*RAD_HOUR);
//    int min = (int)((angle*RAD_HOUR - hrs) * 60);
//    int sec = (int)((angle*RAD_HOUR - hrs - min/60.0) * 3600);

//    return Integer.toString(hrs) + "h" + min + "m" + sec + "s";
//  }

//  static private String radToDms(double angle) {
//    int deg = (int) (angle*RAD_DEG);
//    int min = (int)((angle*RAD_DEG - deg) * 60);
//    int sec = (int)((angle*RAD_DEG - deg - min/60.0) * 3600);

//    return Integer.toString(deg) + "\u00b0" + min + "'" + sec + "\"";
//  }

// =============== Calendar Cache ================

void CalendarCache::createCache(CalendarCache** cache, UErrorCode& status) {
    ucln_i18n_registerCleanup(UCLN_I18N_ASTRO_CALENDAR, calendar_astro_cleanup);
    if(cache == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    } else {
        *cache = new CalendarCache(32, status);
        if(U_FAILURE(status)) {
            delete *cache;
            *cache = NULL;
        }
    }
}

int32_t CalendarCache::get(CalendarCache** cache, int32_t key, UErrorCode &status) {
    int32_t res;

    if(U_FAILURE(status)) {
        return 0;
    }
    umtx_lock(&ccLock);

    if(*cache == NULL) {
        createCache(cache, status);
        if(U_FAILURE(status)) {
            umtx_unlock(&ccLock);
            return 0;
        }
    }

    res = uhash_igeti((*cache)->fTable, key);
    U_DEBUG_ASTRO_MSG(("%p: GET: [%d] == %d\n", (*cache)->fTable, key, res));

    umtx_unlock(&ccLock);
    return res;
}

void CalendarCache::put(CalendarCache** cache, int32_t key, int32_t value, UErrorCode &status) {
    if(U_FAILURE(status)) {
        return;
    }
    umtx_lock(&ccLock);

    if(*cache == NULL) {
        createCache(cache, status);
        if(U_FAILURE(status)) {
            umtx_unlock(&ccLock);
            return;
        }
    }

    uhash_iputi((*cache)->fTable, key, value, &status);
    U_DEBUG_ASTRO_MSG(("%p: PUT: [%d] := %d\n", (*cache)->fTable, key, value));

    umtx_unlock(&ccLock);
}

CalendarCache::CalendarCache(int32_t size, UErrorCode &status) {
    fTable = uhash_openSize(uhash_hashLong, uhash_compareLong, NULL, size, &status);
    U_DEBUG_ASTRO_MSG(("%p: Opening.\n", fTable));
}

CalendarCache::~CalendarCache() {
    if(fTable != NULL) {
        U_DEBUG_ASTRO_MSG(("%p: Closing.\n", fTable));
        uhash_close(fTable);
    }
}

U_NAMESPACE_END

#endif //  !UCONFIG_NO_FORMATTING
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2013, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2013, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

/**
 * \file 
 * \brief C++ API: ICU TimeZone base class
 */



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2008, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/

/**
 * \file 
 * \brief C++ API: Time zone rule classes
 */



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2008, International Business Machines Corporation and         *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/



#if U_SHOW_CPLUSPLUS_API

/**
 * \file 
 * \brief C++ API: Rule for specifying date and time in an year
 */

#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN
/**
 * <code>DateTimeRule</code> is a class representing a time in a year by
 * a rule specified by month, day of month, day of week and
 * time in the day.
 * 
 * @stable ICU 3.8
 */
class U_I18N_API DateTimeRule : public UObject {
public:

    /**
     * Date rule type constants.
     * @stable ICU 3.8
     */
    enum DateRuleType {
        DOM = 0,        /**< The exact day of month,
                             for example, March 11. */
        DOW,            /**< The Nth occurence of the day of week,
                             for example, 2nd Sunday in March. */
        DOW_GEQ_DOM,    /**< The first occurence of the day of week on or after the day of monnth,
                             for example, first Sunday on or after March 8. */
        DOW_LEQ_DOM     /**< The last occurence of the day of week on or before the day of month,
                             for example, first Sunday on or before March 14. */
    };

    /**
     * Time rule type constants.
     * @stable ICU 3.8
     */
    enum TimeRuleType {
        WALL_TIME = 0,  /**< The local wall clock time */
        STANDARD_TIME,  /**< The local standard time */
        UTC_TIME        /**< The UTC time */
    };

    /**
     * Constructs a <code>DateTimeRule</code> by the day of month and
     * the time rule.  The date rule type for an instance created by
     * this constructor is <code>DOM</code>.
     * 
     * @param month         The rule month, for example, <code>Calendar::JANUARY</code>
     * @param dayOfMonth    The day of month, 1-based.
     * @param millisInDay   The milliseconds in the rule date.
     * @param timeType      The time type, <code>WALL_TIME</code> or <code>STANDARD_TIME</code>
     *                      or <code>UTC_TIME</code>.
     * @stable ICU 3.8
     */
    DateTimeRule(int32_t month, int32_t dayOfMonth,
        int32_t millisInDay, TimeRuleType timeType);

    /**
     * Constructs a <code>DateTimeRule</code> by the day of week and its oridinal
     * number and the time rule.  The date rule type for an instance created
     * by this constructor is <code>DOW</code>.
     * 
     * @param month         The rule month, for example, <code>Calendar::JANUARY</code>.
     * @param weekInMonth   The ordinal number of the day of week.  Negative number
     *                      may be used for specifying a rule date counted from the
     *                      end of the rule month.
     * @param dayOfWeek     The day of week, for example, <code>Calendar::SUNDAY</code>.
     * @param millisInDay   The milliseconds in the rule date.
     * @param timeType      The time type, <code>WALL_TIME</code> or <code>STANDARD_TIME</code>
     *                      or <code>UTC_TIME</code>.
     * @stable ICU 3.8
     */
    DateTimeRule(int32_t month, int32_t weekInMonth, int32_t dayOfWeek,
        int32_t millisInDay, TimeRuleType timeType);

    /**
     * Constructs a <code>DateTimeRule</code> by the first/last day of week
     * on or after/before the day of month and the time rule.  The date rule
     * type for an instance created by this constructor is either
     * <code>DOM_GEQ_DOM</code> or <code>DOM_LEQ_DOM</code>.
     * 
     * @param month         The rule month, for example, <code>Calendar::JANUARY</code>
     * @param dayOfMonth    The day of month, 1-based.
     * @param dayOfWeek     The day of week, for example, <code>Calendar::SUNDAY</code>.
     * @param after         true if the rule date is on or after the day of month.
     * @param millisInDay   The milliseconds in the rule date.
     * @param timeType      The time type, <code>WALL_TIME</code> or <code>STANDARD_TIME</code>
     *                      or <code>UTC_TIME</code>.
     * @stable ICU 3.8
     */
    DateTimeRule(int32_t month, int32_t dayOfMonth, int32_t dayOfWeek, UBool after,
        int32_t millisInDay, TimeRuleType timeType);

    /**
     * Copy constructor.
     * @param source    The DateTimeRule object to be copied.
     * @stable ICU 3.8
     */
    DateTimeRule(const DateTimeRule& source);

    /**
     * Destructor.
     * @stable ICU 3.8
     */
    ~DateTimeRule();

    /**
     * Clone this DateTimeRule object polymorphically. The caller owns the result and
     * should delete it when done.
     * @return    A copy of the object.
     * @stable ICU 3.8
     */
    DateTimeRule* clone() const;

    /**
     * Assignment operator.
     * @param right The object to be copied.
     * @stable ICU 3.8
     */
    DateTimeRule& operator=(const DateTimeRule& right);

    /**
     * Return true if the given DateTimeRule objects are semantically equal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given DateTimeRule objects are semantically equal.
     * @stable ICU 3.8
     */
    UBool operator==(const DateTimeRule& that) const;

    /**
     * Return true if the given DateTimeRule objects are semantically unequal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given DateTimeRule objects are semantically unequal.
     * @stable ICU 3.8
     */
    UBool operator!=(const DateTimeRule& that) const;

    /**
     * Gets the date rule type, such as <code>DOM</code>
     * @return The date rule type.
     * @stable ICU 3.8
     */
    DateRuleType getDateRuleType(void) const;

    /**
     * Gets the time rule type
     * @return The time rule type, either <code>WALL_TIME</code> or <code>STANDARD_TIME</code>
     *         or <code>UTC_TIME</code>.
     * @stable ICU 3.8
     */
    TimeRuleType getTimeRuleType(void) const;

    /**
     * Gets the rule month.
     * @return The rule month.
     * @stable ICU 3.8
     */
    int32_t getRuleMonth(void) const;

    /**
     * Gets the rule day of month.  When the date rule type
     * is <code>DOW</code>, the value is always 0.
     * @return The rule day of month
     * @stable ICU 3.8
     */
    int32_t getRuleDayOfMonth(void) const;

    /**
     * Gets the rule day of week.  When the date rule type
     * is <code>DOM</code>, the value is always 0.
     * @return The rule day of week.
     * @stable ICU 3.8
     */
    int32_t getRuleDayOfWeek(void) const;

    /**
     * Gets the ordinal number of the occurence of the day of week
     * in the month.  When the date rule type is not <code>DOW</code>,
     * the value is always 0.
     * @return The rule day of week ordinal number in the month.
     * @stable ICU 3.8
     */
    int32_t getRuleWeekInMonth(void) const;

    /**
     * Gets the rule time in the rule day.
     * @return The time in the rule day in milliseconds.
     * @stable ICU 3.8
     */
    int32_t getRuleMillisInDay(void) const;

private:
    int32_t fMonth;
    int32_t fDayOfMonth;
    int32_t fDayOfWeek;
    int32_t fWeekInMonth;
    int32_t fMillisInDay;
    DateRuleType fDateRuleType;
    TimeRuleType fTimeRuleType;

public:
    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 3.8
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 3.8
     */
    virtual UClassID getDynamicClassID(void) const;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



U_NAMESPACE_BEGIN

/**
 * <code>TimeZoneRule</code> is a class representing a rule for time zone.
 * <code>TimeZoneRule</code> has a set of time zone attributes, such as zone name,
 * raw offset (UTC offset for standard time) and daylight saving time offset.
 * 
 * @stable ICU 3.8
 */
class U_I18N_API TimeZoneRule : public UObject {
public:
    /**
     * Destructor.
     * @stable ICU 3.8
     */
    virtual ~TimeZoneRule();

    /**
     * Clone this TimeZoneRule object polymorphically. The caller owns the result and
     * should delete it when done.
     * @return  A copy of the object.
     * @stable ICU 3.8
     */
    virtual TimeZoneRule* clone() const = 0;

    /**
     * Return true if the given <code>TimeZoneRule</code> objects are semantically equal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZoneRule</code> objects are semantically equal.
     * @stable ICU 3.8
     */
    virtual UBool operator==(const TimeZoneRule& that) const;

    /**
     * Return true if the given <code>TimeZoneRule</code> objects are semantically unequal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZoneRule</code> objects are semantically unequal.
     * @stable ICU 3.8
     */
    virtual UBool operator!=(const TimeZoneRule& that) const;

    /**
     * Fills in "name" with the name of this time zone.
     * @param name  Receives the name of this time zone.
     * @return  A reference to "name"
     * @stable ICU 3.8
     */
    UnicodeString& getName(UnicodeString& name) const;

    /**
     * Gets the standard time offset.
     * @return  The standard time offset from UTC in milliseconds.
     * @stable ICU 3.8
     */
    int32_t getRawOffset(void) const;

    /**
     * Gets the amount of daylight saving delta time from the standard time.
     * @return  The amount of daylight saving offset used by this rule
     *          in milliseconds.
     * @stable ICU 3.8
     */
    int32_t getDSTSavings(void) const;

    /**
     * Returns if this rule represents the same rule and offsets as another.
     * When two <code>TimeZoneRule</code> objects differ only its names, this method
     * returns true.
     * @param other The <code>TimeZoneRule</code> object to be compared with.
     * @return  true if the other <code>TimeZoneRule</code> is the same as this one.
     * @stable ICU 3.8
     */
    virtual UBool isEquivalentTo(const TimeZoneRule& other) const;

    /**
     * Gets the very first time when this rule takes effect.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param result            Receives the very first time when this rule takes effect.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getFirstStart(int32_t prevRawOffset, int32_t prevDSTSavings, UDate& result) const = 0;

    /**
     * Gets the final time when this rule takes effect.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param result            Receives the final time when this rule takes effect.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getFinalStart(int32_t prevRawOffset, int32_t prevDSTSavings, UDate& result) const = 0;

    /**
     * Gets the first time when this rule takes effect after the specified time.
     * @param base              The first start time after this base time will be returned.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param inclusive         Whether the base time is inclusive or not.
     * @param result            Receives The first time when this rule takes effect after
     *                          the specified base time.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getNextStart(UDate base, int32_t prevRawOffset, int32_t prevDSTSavings,
        UBool inclusive, UDate& result) const = 0;

    /**
     * Gets the most recent time when this rule takes effect before the specified time.
     * @param base              The most recent time before this base time will be returned.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param inclusive         Whether the base time is inclusive or not.
     * @param result            Receives The most recent time when this rule takes effect before
     *                          the specified base time.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getPreviousStart(UDate base, int32_t prevRawOffset, int32_t prevDSTSavings,
        UBool inclusive, UDate& result) const = 0;

protected:

    /**
     * Constructs a <code>TimeZoneRule</code> with the name, the GMT offset of its
     * standard time and the amount of daylight saving offset adjustment.
     * @param name          The time zone name.
     * @param rawOffset     The UTC offset of its standard time in milliseconds.
     * @param dstSavings    The amount of daylight saving offset adjustment in milliseconds.
     *                      If this ia a rule for standard time, the value of this argument is 0.
     * @stable ICU 3.8
     */
    TimeZoneRule(const UnicodeString& name, int32_t rawOffset, int32_t dstSavings);

    /**
     * Copy constructor.
     * @param source    The TimeZoneRule object to be copied.
     * @stable ICU 3.8
     */
    TimeZoneRule(const TimeZoneRule& source);

    /**
     * Assignment operator.
     * @param right The object to be copied.
     * @stable ICU 3.8
     */
    TimeZoneRule& operator=(const TimeZoneRule& right);

private:
    UnicodeString fName; // time name
    int32_t fRawOffset;  // UTC offset of the standard time in milliseconds
    int32_t fDSTSavings; // DST saving amount in milliseconds
};

/**
 * <code>InitialTimeZoneRule</code> represents a time zone rule
 * representing a time zone effective from the beginning and
 * has no actual start times.
 * @stable ICU 3.8
 */
class U_I18N_API InitialTimeZoneRule : public TimeZoneRule {
public:
    /**
     * Constructs an <code>InitialTimeZoneRule</code> with the name, the GMT offset of its
     * standard time and the amount of daylight saving offset adjustment.
     * @param name          The time zone name.
     * @param rawOffset     The UTC offset of its standard time in milliseconds.
     * @param dstSavings    The amount of daylight saving offset adjustment in milliseconds.
     *                      If this ia a rule for standard time, the value of this argument is 0.
     * @stable ICU 3.8
     */
    InitialTimeZoneRule(const UnicodeString& name, int32_t rawOffset, int32_t dstSavings);

    /**
     * Copy constructor.
     * @param source    The InitialTimeZoneRule object to be copied.
     * @stable ICU 3.8
     */
    InitialTimeZoneRule(const InitialTimeZoneRule& source);

    /**
     * Destructor.
     * @stable ICU 3.8
     */
    virtual ~InitialTimeZoneRule();

    /**
     * Clone this InitialTimeZoneRule object polymorphically. The caller owns the result and
     * should delete it when done.
     * @return    A copy of the object.
     * @stable ICU 3.8
     */
    virtual InitialTimeZoneRule* clone() const;

    /**
     * Assignment operator.
     * @param right The object to be copied.
     * @stable ICU 3.8
     */
    InitialTimeZoneRule& operator=(const InitialTimeZoneRule& right);

    /**
     * Return true if the given <code>TimeZoneRule</code> objects are semantically equal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZoneRule</code> objects are semantically equal.
     * @stable ICU 3.8
     */
    virtual UBool operator==(const TimeZoneRule& that) const;

    /**
     * Return true if the given <code>TimeZoneRule</code> objects are semantically unequal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZoneRule</code> objects are semantically unequal.
     * @stable ICU 3.8
     */
    virtual UBool operator!=(const TimeZoneRule& that) const;

    /**
     * Gets the time when this rule takes effect in the given year.
     * @param year              The Gregorian year, with 0 == 1 BCE, -1 == 2 BCE, etc.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param result            Receives the start time in the year.
     * @return  true if this rule takes effect in the year and the result is set to
     *          "result".
     * @stable ICU 3.8
     */
    UBool getStartInYear(int32_t year, int32_t prevRawOffset, int32_t prevDSTSavings, UDate& result) const;

    /**
     * Returns if this rule represents the same rule and offsets as another.
     * When two <code>TimeZoneRule</code> objects differ only its names, this method
     * returns true.
     * @param that  The <code>TimeZoneRule</code> object to be compared with.
     * @return  true if the other <code>TimeZoneRule</code> is equivalent to this one.
     * @stable ICU 3.8
     */
    virtual UBool isEquivalentTo(const TimeZoneRule& that) const;

    /**
     * Gets the very first time when this rule takes effect.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param result            Receives the very first time when this rule takes effect.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getFirstStart(int32_t prevRawOffset, int32_t prevDSTSavings, UDate& result) const;

    /**
     * Gets the final time when this rule takes effect.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param result            Receives the final time when this rule takes effect.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getFinalStart(int32_t prevRawOffset, int32_t prevDSTSavings, UDate& result) const;

    /**
     * Gets the first time when this rule takes effect after the specified time.
     * @param base              The first start time after this base time will be returned.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param inclusive         Whether the base time is inclusive or not.
     * @param result            Receives The first time when this rule takes effect after
     *                          the specified base time.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getNextStart(UDate base, int32_t prevRawOffset, int32_t prevDSTSavings,
        UBool inclusive, UDate& result) const;

    /**
     * Gets the most recent time when this rule takes effect before the specified time.
     * @param base              The most recent time before this base time will be returned.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param inclusive         Whether the base time is inclusive or not.
     * @param result            Receives The most recent time when this rule takes effect before
     *                          the specified base time.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getPreviousStart(UDate base, int32_t prevRawOffset, int32_t prevDSTSavings,
        UBool inclusive, UDate& result) const;

public:
    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 3.8
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 3.8
     */
    virtual UClassID getDynamicClassID(void) const;
};

/**
 * <code>AnnualTimeZoneRule</code> is a class used for representing a time zone
 * rule which takes effect annually.  The calenday system used for the rule is
 * is based on Gregorian calendar
 * 
 * @stable ICU 3.8
 */
class U_I18N_API AnnualTimeZoneRule : public TimeZoneRule {
public:
    /**
     * The constant representing the maximum year used for designating
     * a rule is permanent.
     */
    static const int32_t MAX_YEAR;

    /**
     * Constructs a <code>AnnualTimeZoneRule</code> with the name, the GMT offset of its
     * standard time, the amount of daylight saving offset adjustment, the annual start
     * time rule and the start/until years.  The input DateTimeRule is copied by this
     * constructor, so the caller remains responsible for deleting the object.
     * @param name          The time zone name.
     * @param rawOffset     The GMT offset of its standard time in milliseconds.
     * @param dstSavings    The amount of daylight saving offset adjustment in
     *                      milliseconds.  If this ia a rule for standard time,
     *                      the value of this argument is 0.
     * @param dateTimeRule  The start date/time rule repeated annually.
     * @param startYear     The first year when this rule takes effect.
     * @param endYear       The last year when this rule takes effect.  If this
     *                      rule is effective forever in future, specify MAX_YEAR.
     * @stable ICU 3.8
     */
    AnnualTimeZoneRule(const UnicodeString& name, int32_t rawOffset, int32_t dstSavings,
            const DateTimeRule& dateTimeRule, int32_t startYear, int32_t endYear);

    /**
     * Constructs a <code>AnnualTimeZoneRule</code> with the name, the GMT offset of its
     * standard time, the amount of daylight saving offset adjustment, the annual start
     * time rule and the start/until years.  The input DateTimeRule object is adopted
     * by this object, therefore, the caller must not delete the object.
     * @param name          The time zone name.
     * @param rawOffset     The GMT offset of its standard time in milliseconds.
     * @param dstSavings    The amount of daylight saving offset adjustment in
     *                      milliseconds.  If this ia a rule for standard time,
     *                      the value of this argument is 0.
     * @param dateTimeRule  The start date/time rule repeated annually.
     * @param startYear     The first year when this rule takes effect.
     * @param endYear       The last year when this rule takes effect.  If this
     *                      rule is effective forever in future, specify MAX_YEAR.
     * @stable ICU 3.8
     */
    AnnualTimeZoneRule(const UnicodeString& name, int32_t rawOffset, int32_t dstSavings,
            DateTimeRule* dateTimeRule, int32_t startYear, int32_t endYear);

    /**
     * Copy constructor.
     * @param source    The AnnualTimeZoneRule object to be copied.
     * @stable ICU 3.8
     */
    AnnualTimeZoneRule(const AnnualTimeZoneRule& source);

    /**
     * Destructor.
     * @stable ICU 3.8
     */
    virtual ~AnnualTimeZoneRule();

    /**
     * Clone this AnnualTimeZoneRule object polymorphically. The caller owns the result and
     * should delete it when done.
     * @return    A copy of the object.
     * @stable ICU 3.8
     */
    virtual AnnualTimeZoneRule* clone() const;

    /**
     * Assignment operator.
     * @param right The object to be copied.
     * @stable ICU 3.8
     */
    AnnualTimeZoneRule& operator=(const AnnualTimeZoneRule& right);

    /**
     * Return true if the given <code>TimeZoneRule</code> objects are semantically equal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZoneRule</code> objects are semantically equal.
     * @stable ICU 3.8
     */
    virtual UBool operator==(const TimeZoneRule& that) const;

    /**
     * Return true if the given <code>TimeZoneRule</code> objects are semantically unequal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZoneRule</code> objects are semantically unequal.
     * @stable ICU 3.8
     */
    virtual UBool operator!=(const TimeZoneRule& that) const;

    /**
     * Gets the start date/time rule used by this rule.
     * @return  The <code>AnnualDateTimeRule</code> which represents the start date/time
     *          rule used by this time zone rule.
     * @stable ICU 3.8
     */
    const DateTimeRule* getRule(void) const;

    /**
     * Gets the first year when this rule takes effect.
     * @return  The start year of this rule.  The year is in Gregorian calendar
     *          with 0 == 1 BCE, -1 == 2 BCE, etc.
     * @stable ICU 3.8
     */
    int32_t getStartYear(void) const;

    /**
     * Gets the end year when this rule takes effect.
     * @return  The end year of this rule (inclusive). The year is in Gregorian calendar
     *          with 0 == 1 BCE, -1 == 2 BCE, etc.
     * @stable ICU 3.8
     */
    int32_t getEndYear(void) const;

    /**
     * Gets the time when this rule takes effect in the given year.
     * @param year              The Gregorian year, with 0 == 1 BCE, -1 == 2 BCE, etc.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param result            Receives the start time in the year.
     * @return  true if this rule takes effect in the year and the result is set to
     *          "result".
     * @stable ICU 3.8
     */
    UBool getStartInYear(int32_t year, int32_t prevRawOffset, int32_t prevDSTSavings, UDate& result) const;

    /**
     * Returns if this rule represents the same rule and offsets as another.
     * When two <code>TimeZoneRule</code> objects differ only its names, this method
     * returns true.
     * @param that  The <code>TimeZoneRule</code> object to be compared with.
     * @return  true if the other <code>TimeZoneRule</code> is equivalent to this one.
     * @stable ICU 3.8
     */
    virtual UBool isEquivalentTo(const TimeZoneRule& that) const;

    /**
     * Gets the very first time when this rule takes effect.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param result            Receives the very first time when this rule takes effect.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getFirstStart(int32_t prevRawOffset, int32_t prevDSTSavings, UDate& result) const;

    /**
     * Gets the final time when this rule takes effect.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param result            Receives the final time when this rule takes effect.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getFinalStart(int32_t prevRawOffset, int32_t prevDSTSavings, UDate& result) const;

    /**
     * Gets the first time when this rule takes effect after the specified time.
     * @param base              The first start time after this base time will be returned.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param inclusive         Whether the base time is inclusive or not.
     * @param result            Receives The first time when this rule takes effect after
     *                          the specified base time.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getNextStart(UDate base, int32_t prevRawOffset, int32_t prevDSTSavings,
        UBool inclusive, UDate& result) const;

    /**
     * Gets the most recent time when this rule takes effect before the specified time.
     * @param base              The most recent time before this base time will be returned.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param inclusive         Whether the base time is inclusive or not.
     * @param result            Receives The most recent time when this rule takes effect before
     *                          the specified base time.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getPreviousStart(UDate base, int32_t prevRawOffset, int32_t prevDSTSavings,
        UBool inclusive, UDate& result) const;


private:
    DateTimeRule* fDateTimeRule;
    int32_t fStartYear;
    int32_t fEndYear;

public:
    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 3.8
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 3.8
     */
    virtual UClassID getDynamicClassID(void) const;
};

/**
 * <code>TimeArrayTimeZoneRule</code> represents a time zone rule whose start times are
 * defined by an array of milliseconds since the standard base time.
 * 
 * @stable ICU 3.8
 */
class U_I18N_API TimeArrayTimeZoneRule : public TimeZoneRule {
public:
    /**
     * Constructs a <code>TimeArrayTimeZoneRule</code> with the name, the GMT offset of its
     * standard time, the amount of daylight saving offset adjustment and
     * the array of times when this rule takes effect.
     * @param name          The time zone name.
     * @param rawOffset     The UTC offset of its standard time in milliseconds.
     * @param dstSavings    The amount of daylight saving offset adjustment in
     *                      milliseconds.  If this ia a rule for standard time,
     *                      the value of this argument is 0.
     * @param startTimes    The array start times in milliseconds since the base time
     *                      (January 1, 1970, 00:00:00).
     * @param numStartTimes The number of elements in the parameter "startTimes"
     * @param timeRuleType  The time type of the start times, which is one of
     *                      <code>DataTimeRule::WALL_TIME</code>, <code>STANDARD_TIME</code>
     *                      and <code>UTC_TIME</code>.
     * @stable ICU 3.8
     */
    TimeArrayTimeZoneRule(const UnicodeString& name, int32_t rawOffset, int32_t dstSavings,
        const UDate* startTimes, int32_t numStartTimes, DateTimeRule::TimeRuleType timeRuleType);

    /**
     * Copy constructor.
     * @param source    The TimeArrayTimeZoneRule object to be copied.
     * @stable ICU 3.8
     */
    TimeArrayTimeZoneRule(const TimeArrayTimeZoneRule& source);

    /**
     * Destructor.
     * @stable ICU 3.8
     */
    virtual ~TimeArrayTimeZoneRule();

    /**
     * Clone this TimeArrayTimeZoneRule object polymorphically. The caller owns the result and
     * should delete it when done.
     * @return    A copy of the object.
     * @stable ICU 3.8
     */
    virtual TimeArrayTimeZoneRule* clone() const;

    /**
     * Assignment operator.
     * @param right The object to be copied.
     * @stable ICU 3.8
     */
    TimeArrayTimeZoneRule& operator=(const TimeArrayTimeZoneRule& right);

    /**
     * Return true if the given <code>TimeZoneRule</code> objects are semantically equal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZoneRule</code> objects are semantically equal.
     * @stable ICU 3.8
     */
    virtual UBool operator==(const TimeZoneRule& that) const;

    /**
     * Return true if the given <code>TimeZoneRule</code> objects are semantically unequal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZoneRule</code> objects are semantically unequal.
     * @stable ICU 3.8
     */
    virtual UBool operator!=(const TimeZoneRule& that) const;

    /**
     * Gets the time type of the start times used by this rule.  The return value
     * is either <code>DateTimeRule::WALL_TIME</code> or <code>STANDARD_TIME</code>
     * or <code>UTC_TIME</code>.
     * 
     * @return The time type used of the start times used by this rule.
     * @stable ICU 3.8
     */
    DateTimeRule::TimeRuleType getTimeType(void) const;

    /**
     * Gets a start time at the index stored in this rule.
     * @param index     The index of start times
     * @param result    Receives the start time at the index
     * @return  true if the index is within the valid range and
     *          and the result is set.  When false, the output
     *          parameger "result" is unchanged.
     * @stable ICU 3.8
     */
    UBool getStartTimeAt(int32_t index, UDate& result) const;

    /**
     * Returns the number of start times stored in this rule
     * @return The number of start times.
     * @stable ICU 3.8
     */
    int32_t countStartTimes(void) const;

    /**
     * Returns if this rule represents the same rule and offsets as another.
     * When two <code>TimeZoneRule</code> objects differ only its names, this method
     * returns true.
     * @param that  The <code>TimeZoneRule</code> object to be compared with.
     * @return  true if the other <code>TimeZoneRule</code> is equivalent to this one.
     * @stable ICU 3.8
     */
    virtual UBool isEquivalentTo(const TimeZoneRule& that) const;

    /**
     * Gets the very first time when this rule takes effect.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param result            Receives the very first time when this rule takes effect.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getFirstStart(int32_t prevRawOffset, int32_t prevDSTSavings, UDate& result) const;

    /**
     * Gets the final time when this rule takes effect.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param result            Receives the final time when this rule takes effect.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getFinalStart(int32_t prevRawOffset, int32_t prevDSTSavings, UDate& result) const;

    /**
     * Gets the first time when this rule takes effect after the specified time.
     * @param base              The first start time after this base time will be returned.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param inclusive         Whether the base time is inclusive or not.
     * @param result            Receives The first time when this rule takes effect after
     *                          the specified base time.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getNextStart(UDate base, int32_t prevRawOffset, int32_t prevDSTSavings,
        UBool inclusive, UDate& result) const;

    /**
     * Gets the most recent time when this rule takes effect before the specified time.
     * @param base              The most recent time before this base time will be returned.
     * @param prevRawOffset     The standard time offset from UTC before this rule
     *                          takes effect in milliseconds.
     * @param prevDSTSavings    The amount of daylight saving offset from the
     *                          standard time.
     * @param inclusive         Whether the base time is inclusive or not.
     * @param result            Receives The most recent time when this rule takes effect before
     *                          the specified base time.
     * @return  true if the start time is available.  When false is returned, output parameter
     *          "result" is unchanged.
     * @stable ICU 3.8
     */
    virtual UBool getPreviousStart(UDate base, int32_t prevRawOffset, int32_t prevDSTSavings,
        UBool inclusive, UDate& result) const;


private:
    enum { TIMEARRAY_STACK_BUFFER_SIZE = 32 };
    UBool initStartTimes(const UDate source[], int32_t size, UErrorCode& ec);
    UDate getUTC(UDate time, int32_t raw, int32_t dst) const;

    DateTimeRule::TimeRuleType  fTimeRuleType;
    int32_t fNumStartTimes;
    UDate*  fStartTimes;
    UDate   fLocalStartTimes[TIMEARRAY_STACK_BUFFER_SIZE];

public:
    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 3.8
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 3.8
     */
    virtual UClassID getDynamicClassID(void) const;
};


U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2008, International Business Machines Corporation and         *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/

/**
 * \file 
 * \brief C++ API: Time zone transition
 */



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

// Forward declaration
class TimeZoneRule;

/**
 * <code>TimeZoneTransition</code> is a class representing a time zone transition.
 * An instance has a time of transition and rules for both before and after the transition.
 * @stable ICU 3.8
 */
class U_I18N_API TimeZoneTransition : public UObject {
public:
    /**
     * Constructs a <code>TimeZoneTransition</code> with the time and the rules before/after
     * the transition.
     * 
     * @param time  The time of transition in milliseconds since the base time.
     * @param from  The time zone rule used before the transition.
     * @param to    The time zone rule used after the transition.
     * @stable ICU 3.8
     */
    TimeZoneTransition(UDate time, const TimeZoneRule& from, const TimeZoneRule& to);

    /**
     * Constructs an empty <code>TimeZoneTransition</code>
     * @stable ICU 3.8
     */
    TimeZoneTransition();

    /**
     * Copy constructor.
     * @param source    The TimeZoneTransition object to be copied.
     * @stable ICU 3.8
     */
    TimeZoneTransition(const TimeZoneTransition& source);

    /**
     * Destructor.
     * @stable ICU 3.8
     */
    ~TimeZoneTransition();

    /**
     * Clone this TimeZoneTransition object polymorphically. The caller owns the result and
     * should delete it when done.
     * @return  A copy of the object.
     * @stable ICU 3.8
     */
    TimeZoneTransition* clone() const;

    /**
     * Assignment operator.
     * @param right The object to be copied.
     * @stable ICU 3.8
     */
    TimeZoneTransition& operator=(const TimeZoneTransition& right);

    /**
     * Return true if the given TimeZoneTransition objects are semantically equal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given TimeZoneTransition objects are semantically equal.
     * @stable ICU 3.8
     */
    UBool operator==(const TimeZoneTransition& that) const;

    /**
     * Return true if the given TimeZoneTransition objects are semantically unequal. Objects
     * of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given TimeZoneTransition objects are semantically unequal.
     * @stable ICU 3.8
     */
    UBool operator!=(const TimeZoneTransition& that) const;

    /**
     * Returns the time of transition in milliseconds.
     * @return The time of the transition in milliseconds since the 1970 Jan 1 epoch time.
     * @stable ICU 3.8
     */
    UDate getTime(void) const;

    /**
     * Sets the time of transition in milliseconds.
     * @param time The time of the transition in milliseconds since the 1970 Jan 1 epoch time.
     * @stable ICU 3.8
     */
    void setTime(UDate time);

    /**
     * Returns the rule used before the transition.
     * @return The time zone rule used after the transition.
     * @stable ICU 3.8
     */
    const TimeZoneRule* getFrom(void) const;

    /**
     * Sets the rule used before the transition.  The caller remains
     * responsible for deleting the <code>TimeZoneRule</code> object.
     * @param from The time zone rule used before the transition.
     * @stable ICU 3.8
     */
    void setFrom(const TimeZoneRule& from);

    /**
     * Adopts the rule used before the transition.  The caller must
     * not delete the <code>TimeZoneRule</code> object passed in.
     * @param from The time zone rule used before the transition.
     * @stable ICU 3.8
     */
    void adoptFrom(TimeZoneRule* from);

    /**
     * Sets the rule used after the transition.  The caller remains
     * responsible for deleting the <code>TimeZoneRule</code> object.
     * @param to The time zone rule used after the transition.
     * @stable ICU 3.8
     */
    void setTo(const TimeZoneRule& to);

    /**
     * Adopts the rule used after the transition.  The caller must
     * not delete the <code>TimeZoneRule</code> object passed in.
     * @param to The time zone rule used after the transition.
     * @stable ICU 3.8
     */
    void adoptTo(TimeZoneRule* to);

    /**
     * Returns the rule used after the transition.
     * @return The time zone rule used after the transition.
     * @stable ICU 3.8
     */
    const TimeZoneRule* getTo(void) const;

private:
    UDate   fTime;
    TimeZoneRule*   fFrom;
    TimeZoneRule*   fTo;

public:
    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 3.8
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 3.8
     */
    virtual UClassID getDynamicClassID(void) const;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



U_NAMESPACE_BEGIN

// forward declarations
class UVector;

/**
 * <code>BasicTimeZone</code> is an abstract class extending <code>TimeZone</code>.
 * This class provides some additional methods to access time zone transitions and rules.
 * All ICU <code>TimeZone</code> concrete subclasses extend this class.
 * @stable ICU 3.8
 */
class U_I18N_API BasicTimeZone: public TimeZone {
public:
    /**
     * Destructor.
     * @stable ICU 3.8
     */
    virtual ~BasicTimeZone();

    /**
     * Clones this object polymorphically.
     * The caller owns the result and should delete it when done.
     * @return clone, or nullptr if an error occurred
     * @stable ICU 3.8
     */
    virtual BasicTimeZone* clone() const = 0;

    /**
     * Gets the first time zone transition after the base time.
     * @param base      The base time.
     * @param inclusive Whether the base time is inclusive or not.
     * @param result    Receives the first transition after the base time.
     * @return  TRUE if the transition is found.
     * @stable ICU 3.8
     */
    virtual UBool getNextTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const = 0;

    /**
     * Gets the most recent time zone transition before the base time.
     * @param base      The base time.
     * @param inclusive Whether the base time is inclusive or not.
     * @param result    Receives the most recent transition before the base time.
     * @return  TRUE if the transition is found.
     * @stable ICU 3.8
     */
    virtual UBool getPreviousTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const = 0;

    /**
     * Checks if the time zone has equivalent transitions in the time range.
     * This method returns true when all of transition times, from/to standard
     * offsets and DST savings used by this time zone match the other in the
     * time range.
     * @param tz    The <code>BasicTimeZone</code> object to be compared with.
     * @param start The start time of the evaluated time range (inclusive)
     * @param end   The end time of the evaluated time range (inclusive)
     * @param ignoreDstAmount
     *              When true, any transitions with only daylight saving amount
     *              changes will be ignored, except either of them is zero.
     *              For example, a transition from rawoffset 3:00/dstsavings 1:00
     *              to rawoffset 2:00/dstsavings 2:00 is excluded from the comparison,
     *              but a transtion from rawoffset 2:00/dstsavings 1:00 to
     *              rawoffset 3:00/dstsavings 0:00 is included.
     * @param ec    Output param to filled in with a success or an error.
     * @return      true if the other time zone has the equivalent transitions in the
     *              time range.
     * @stable ICU 3.8
     */
    virtual UBool hasEquivalentTransitions(const BasicTimeZone& tz, UDate start, UDate end,
        UBool ignoreDstAmount, UErrorCode& ec) const;

    /**
     * Returns the number of <code>TimeZoneRule</code>s which represents time transitions,
     * for this time zone, that is, all <code>TimeZoneRule</code>s for this time zone except
     * <code>InitialTimeZoneRule</code>.  The return value range is 0 or any positive value.
     * @param status    Receives error status code.
     * @return The number of <code>TimeZoneRule</code>s representing time transitions.
     * @stable ICU 3.8
     */
    virtual int32_t countTransitionRules(UErrorCode& status) const = 0;

    /**
     * Gets the <code>InitialTimeZoneRule</code> and the set of <code>TimeZoneRule</code>
     * which represent time transitions for this time zone.  On successful return,
     * the argument initial points to non-NULL <code>InitialTimeZoneRule</code> and
     * the array trsrules is filled with 0 or multiple <code>TimeZoneRule</code>
     * instances up to the size specified by trscount.  The results are referencing the
     * rule instance held by this time zone instance.  Therefore, after this time zone
     * is destructed, they are no longer available.
     * @param initial       Receives the initial timezone rule
     * @param trsrules      Receives the timezone transition rules
     * @param trscount      On input, specify the size of the array 'transitions' receiving
     *                      the timezone transition rules.  On output, actual number of
     *                      rules filled in the array will be set.
     * @param status        Receives error status code.
     * @stable ICU 3.8
     */
    virtual void getTimeZoneRules(const InitialTimeZoneRule*& initial,
        const TimeZoneRule* trsrules[], int32_t& trscount, UErrorCode& status) const = 0;

    /**
     * Gets the set of time zone rules valid at the specified time.  Some known external time zone
     * implementations are not capable to handle historic time zone rule changes.  Also some
     * implementations can only handle certain type of rule definitions.
     * If this time zone does not use any daylight saving time within about 1 year from the specified
     * time, only the <code>InitialTimeZone</code> is returned.  Otherwise, the rule for standard
     * time and daylight saving time transitions are returned in addition to the
     * <code>InitialTimeZoneRule</code>.  The standard and daylight saving time transition rules are
     * represented by <code>AnnualTimeZoneRule</code> with <code>DateTimeRule::DOW</code> for its date
     * rule and <code>DateTimeRule::WALL_TIME</code> for its time rule.  Because daylight saving time
     * rule is changing time to time in many time zones and also mapping a transition time rule to
     * different type is lossy transformation, the set of rules returned by this method may be valid
     * for short period of time.
     * The time zone rule objects returned by this method is owned by the caller, so the caller is
     * responsible for deleting them after use.
     * @param date      The date used for extracting time zone rules.
     * @param initial   Receives the <code>InitialTimeZone</code>, always not NULL.
     * @param std       Receives the <code>AnnualTimeZoneRule</code> for standard time transitions.
     *                  When this time time zone does not observe daylight saving times around the
     *                  specified date, NULL is set.
     * @param dst       Receives the <code>AnnualTimeZoneRule</code> for daylight saving time
     *                  transitions.  When this time zone does not observer daylight saving times
     *                  around the specified date, NULL is set.
     * @param status    Receives error status code.
     * @stable ICU 3.8
     */
    virtual void getSimpleRulesNear(UDate date, InitialTimeZoneRule*& initial,
        AnnualTimeZoneRule*& std, AnnualTimeZoneRule*& dst, UErrorCode& status) const;


#ifndef U_HIDE_INTERNAL_API
    /**
     * The time type option bit flags used by getOffsetFromLocal
     * @internal
     */
    enum {
        kStandard = 0x01,
        kDaylight = 0x03,
        kFormer = 0x04,
        kLatter = 0x0C
    };
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Get time zone offsets from local wall time.
     * @internal
     */
    virtual void getOffsetFromLocal(UDate date, int32_t nonExistingTimeOpt, int32_t duplicatedTimeOpt,
        int32_t& rawOffset, int32_t& dstOffset, UErrorCode& status) const;

protected:

#ifndef U_HIDE_INTERNAL_API
    /**
     * The time type option bit masks used by getOffsetFromLocal
     * @internal
     */
    enum {
        kStdDstMask = kDaylight,
        kFormerLatterMask = kLatter
    };
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Default constructor.
     * @stable ICU 3.8
     */
    BasicTimeZone();

    /**
     * Construct a timezone with a given ID.
     * @param id a system time zone ID
     * @stable ICU 3.8
     */
    BasicTimeZone(const UnicodeString &id);

    /**
     * Copy constructor.
     * @param source the object to be copied.
     * @stable ICU 3.8
     */
    BasicTimeZone(const BasicTimeZone& source);

    /**
     * Gets the set of TimeZoneRule instances applicable to the specified time and after.
     * @param start     The start date used for extracting time zone rules
     * @param initial   Receives the InitialTimeZone, always not NULL
     * @param transitionRules   Receives the transition rules, could be NULL
     * @param status    Receives error status code
     */
    void getTimeZoneRulesAfter(UDate start, InitialTimeZoneRule*& initial, UVector*& transitionRules,
        UErrorCode& status) const;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */






U_NAMESPACE_BEGIN

#define MILLIS_PER_YEAR (365*24*60*60*1000.0)

BasicTimeZone::BasicTimeZone()
: TimeZone() {
}

BasicTimeZone::BasicTimeZone(const UnicodeString &id)
: TimeZone(id) {
}

BasicTimeZone::BasicTimeZone(const BasicTimeZone& source)
: TimeZone(source) {
}

BasicTimeZone::~BasicTimeZone() {
}

UBool
BasicTimeZone::hasEquivalentTransitions(const BasicTimeZone& tz, UDate start, UDate end,
                                        UBool ignoreDstAmount, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (hasSameRules(tz)) {
        return TRUE;
    }
    // Check the offsets at the start time
    int32_t raw1, raw2, dst1, dst2;
    getOffset(start, FALSE, raw1, dst1, status);
    if (U_FAILURE(status)) {
        return FALSE;
    }
    tz.getOffset(start, FALSE, raw2, dst2, status);
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (ignoreDstAmount) {
        if ((raw1 + dst1 != raw2 + dst2)
            || (dst1 != 0 && dst2 == 0)
            || (dst1 == 0 && dst2 != 0)) {
            return FALSE;
        }
    } else {
        if (raw1 != raw2 || dst1 != dst2) {
            return FALSE;
        }            
    }
    // Check transitions in the range
    UDate time = start;
    TimeZoneTransition tr1, tr2;
    while (TRUE) {
        UBool avail1 = getNextTransition(time, FALSE, tr1);
        UBool avail2 = tz.getNextTransition(time, FALSE, tr2);

        if (ignoreDstAmount) {
            // Skip a transition which only differ the amount of DST savings
            while (TRUE) {
                if (avail1
                        && tr1.getTime() <= end
                        && (tr1.getFrom()->getRawOffset() + tr1.getFrom()->getDSTSavings()
                                == tr1.getTo()->getRawOffset() + tr1.getTo()->getDSTSavings())
                        && (tr1.getFrom()->getDSTSavings() != 0 && tr1.getTo()->getDSTSavings() != 0)) {
                    getNextTransition(tr1.getTime(), FALSE, tr1);
                } else {
                    break;
                }
            }
            while (TRUE) {
                if (avail2
                        && tr2.getTime() <= end
                        && (tr2.getFrom()->getRawOffset() + tr2.getFrom()->getDSTSavings()
                                == tr2.getTo()->getRawOffset() + tr2.getTo()->getDSTSavings())
                        && (tr2.getFrom()->getDSTSavings() != 0 && tr2.getTo()->getDSTSavings() != 0)) {
                    tz.getNextTransition(tr2.getTime(), FALSE, tr2);
                } else {
                    break;
                }
            }
        }

        UBool inRange1 = (avail1 && tr1.getTime() <= end);
        UBool inRange2 = (avail2 && tr2.getTime() <= end);
        if (!inRange1 && !inRange2) {
            // No more transition in the range
            break;
        }
        if (!inRange1 || !inRange2) {
            return FALSE;
        }
        if (tr1.getTime() != tr2.getTime()) {
            return FALSE;
        }
        if (ignoreDstAmount) {
            if (tr1.getTo()->getRawOffset() + tr1.getTo()->getDSTSavings()
                        != tr2.getTo()->getRawOffset() + tr2.getTo()->getDSTSavings()
                    || (tr1.getTo()->getDSTSavings() != 0 &&  tr2.getTo()->getDSTSavings() == 0)
                    || (tr1.getTo()->getDSTSavings() == 0 &&  tr2.getTo()->getDSTSavings() != 0)) {
                return FALSE;
            }
        } else {
            if (tr1.getTo()->getRawOffset() != tr2.getTo()->getRawOffset() ||
                tr1.getTo()->getDSTSavings() != tr2.getTo()->getDSTSavings()) {
                return FALSE;
            }
        }
        time = tr1.getTime();
    }
    return TRUE;
}

void
BasicTimeZone::getSimpleRulesNear(UDate date, InitialTimeZoneRule*& initial,
        AnnualTimeZoneRule*& std, AnnualTimeZoneRule*& dst, UErrorCode& status) const {
    initial = NULL;
    std = NULL;
    dst = NULL;
    if (U_FAILURE(status)) {
        return;
    }
    int32_t initialRaw, initialDst;
    UnicodeString initialName;

    AnnualTimeZoneRule *ar1 = NULL;
    AnnualTimeZoneRule *ar2 = NULL;
    UnicodeString name;

    UBool avail;
    TimeZoneTransition tr;
    // Get the next transition
    avail = getNextTransition(date, FALSE, tr);
    if (avail) {
        tr.getFrom()->getName(initialName);
        initialRaw = tr.getFrom()->getRawOffset();
        initialDst = tr.getFrom()->getDSTSavings();

        // Check if the next transition is either DST->STD or STD->DST and
        // within roughly 1 year from the specified date
        UDate nextTransitionTime = tr.getTime();
        if (((tr.getFrom()->getDSTSavings() == 0 && tr.getTo()->getDSTSavings() != 0)
              || (tr.getFrom()->getDSTSavings() != 0 && tr.getTo()->getDSTSavings() == 0))
            && (date + MILLIS_PER_YEAR > nextTransitionTime)) {
 
            int32_t year, month, dom, dow, doy, mid;
            UDate d;

            // Get local wall time for the next transition time
            Grego::timeToFields(nextTransitionTime + initialRaw + initialDst,
                year, month, dom, dow, doy, mid);
            int32_t weekInMonth = Grego::dayOfWeekInMonth(year, month, dom);
            // Create DOW rule
            DateTimeRule *dtr = new DateTimeRule(month, weekInMonth, dow, mid, DateTimeRule::WALL_TIME);
            tr.getTo()->getName(name);

            // Note:  SimpleTimeZone does not support raw offset change.
            // So we always use raw offset of the given time for the rule,
            // even raw offset is changed.  This will result that the result
            // zone to return wrong offset after the transition.
            // When we encounter such case, we do not inspect next next
            // transition for another rule.
            ar1 = new AnnualTimeZoneRule(name, initialRaw, tr.getTo()->getDSTSavings(),
                dtr, year, AnnualTimeZoneRule::MAX_YEAR);

            if (tr.getTo()->getRawOffset() == initialRaw) {
                // Get the next next transition
                avail = getNextTransition(nextTransitionTime, FALSE, tr);
                if (avail) {
                    // Check if the next next transition is either DST->STD or STD->DST
                    // and within roughly 1 year from the next transition
                    if (((tr.getFrom()->getDSTSavings() == 0 && tr.getTo()->getDSTSavings() != 0)
                          || (tr.getFrom()->getDSTSavings() != 0 && tr.getTo()->getDSTSavings() == 0))
                         && nextTransitionTime + MILLIS_PER_YEAR > tr.getTime()) {

                        // Get local wall time for the next transition time
                        Grego::timeToFields(tr.getTime() + tr.getFrom()->getRawOffset() + tr.getFrom()->getDSTSavings(),
                            year, month, dom, dow, doy, mid);
                        weekInMonth = Grego::dayOfWeekInMonth(year, month, dom);
                        // Generate another DOW rule
                        dtr = new DateTimeRule(month, weekInMonth, dow, mid, DateTimeRule::WALL_TIME);
                        tr.getTo()->getName(name);
                        ar2 = new AnnualTimeZoneRule(name, tr.getTo()->getRawOffset(), tr.getTo()->getDSTSavings(),
                            dtr, year - 1, AnnualTimeZoneRule::MAX_YEAR);

                        // Make sure this rule can be applied to the specified date
                        avail = ar2->getPreviousStart(date, tr.getFrom()->getRawOffset(), tr.getFrom()->getDSTSavings(), TRUE, d);
                        if (!avail || d > date
                                || initialRaw != tr.getTo()->getRawOffset()
                                || initialDst != tr.getTo()->getDSTSavings()) {
                            // We cannot use this rule as the second transition rule
                            delete ar2;
                            ar2 = NULL;
                        }
                    }
                }
            }
            if (ar2 == NULL) {
                // Try previous transition
                avail = getPreviousTransition(date, TRUE, tr);
                if (avail) {
                    // Check if the previous transition is either DST->STD or STD->DST.
                    // The actual transition time does not matter here.
                    if ((tr.getFrom()->getDSTSavings() == 0 && tr.getTo()->getDSTSavings() != 0)
                        || (tr.getFrom()->getDSTSavings() != 0 && tr.getTo()->getDSTSavings() == 0)) {

                        // Generate another DOW rule
                        Grego::timeToFields(tr.getTime() + tr.getFrom()->getRawOffset() + tr.getFrom()->getDSTSavings(),
                            year, month, dom, dow, doy, mid);
                        weekInMonth = Grego::dayOfWeekInMonth(year, month, dom);
                        dtr = new DateTimeRule(month, weekInMonth, dow, mid, DateTimeRule::WALL_TIME);
                        tr.getTo()->getName(name);

                        // second rule raw/dst offsets should match raw/dst offsets
                        // at the given time
                        ar2 = new AnnualTimeZoneRule(name, initialRaw, initialDst,
                            dtr, ar1->getStartYear() - 1, AnnualTimeZoneRule::MAX_YEAR);

                        // Check if this rule start after the first rule after the specified date
                        avail = ar2->getNextStart(date, tr.getFrom()->getRawOffset(), tr.getFrom()->getDSTSavings(), FALSE, d);
                        if (!avail || d <= nextTransitionTime) {
                            // We cannot use this rule as the second transition rule
                            delete ar2;
                            ar2 = NULL;
                        }
                    }
                }
            }
            if (ar2 == NULL) {
                // Cannot find a good pair of AnnualTimeZoneRule
                delete ar1;
                ar1 = NULL;
            } else {
                // The initial rule should represent the rule before the previous transition
                ar1->getName(initialName);
                initialRaw = ar1->getRawOffset();
                initialDst = ar1->getDSTSavings();
            }
        }
    }
    else {
        // Try the previous one
        avail = getPreviousTransition(date, TRUE, tr);
        if (avail) {
            tr.getTo()->getName(initialName);
            initialRaw = tr.getTo()->getRawOffset();
            initialDst = tr.getTo()->getDSTSavings();
        } else {
            // No transitions in the past.  Just use the current offsets
            getOffset(date, FALSE, initialRaw, initialDst, status);
            if (U_FAILURE(status)) {
                return;
            }
        }
    }
    // Set the initial rule
    initial = new InitialTimeZoneRule(initialName, initialRaw, initialDst);

    // Set the standard and daylight saving rules
    if (ar1 != NULL && ar2 != NULL) {
        if (ar1->getDSTSavings() != 0) {
            dst = ar1;
            std = ar2;
        } else {
            std = ar1;
            dst = ar2;
        }
    }
}

void
BasicTimeZone::getTimeZoneRulesAfter(UDate start, InitialTimeZoneRule*& initial,
                                     UVector*& transitionRules, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }

    const InitialTimeZoneRule *orgini;
    const TimeZoneRule **orgtrs = NULL;
    TimeZoneTransition tzt;
    UBool avail;
    UVector *orgRules = NULL;
    int32_t ruleCount;
    TimeZoneRule *r = NULL;
    UBool *done = NULL;
    InitialTimeZoneRule *res_initial = NULL;
    UVector *filteredRules = NULL;
    UnicodeString name;
    int32_t i;
    UDate time, t;
    UDate *newTimes = NULL;
    UDate firstStart;
    UBool bFinalStd = FALSE, bFinalDst = FALSE;

    // Original transition rules
    ruleCount = countTransitionRules(status);
    if (U_FAILURE(status)) {
        return;
    }
    orgRules = new UVector(ruleCount, status);
    if (U_FAILURE(status)) {
        return;
    }
    orgtrs = (const TimeZoneRule**)uprv_malloc(sizeof(TimeZoneRule*)*ruleCount);
    if (orgtrs == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        goto error;
    }
    getTimeZoneRules(orgini, orgtrs, ruleCount, status);
    if (U_FAILURE(status)) {
        goto error;
    }
    for (i = 0; i < ruleCount; i++) {
        orgRules->addElement(orgtrs[i]->clone(), status);
        if (U_FAILURE(status)) {
            goto error;
        }
    }
    uprv_free(orgtrs);
    orgtrs = NULL;

    avail = getPreviousTransition(start, TRUE, tzt);
    if (!avail) {
        // No need to filter out rules only applicable to time before the start
        initial = orgini->clone();
        transitionRules = orgRules;
        return;
    }

    done = (UBool*)uprv_malloc(sizeof(UBool)*ruleCount);
    if (done == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        goto error;
    }
    filteredRules = new UVector(status);
    if (U_FAILURE(status)) {
        goto error;
    }

    // Create initial rule
    tzt.getTo()->getName(name);
    res_initial = new InitialTimeZoneRule(name, tzt.getTo()->getRawOffset(),
        tzt.getTo()->getDSTSavings());

    // Mark rules which does not need to be processed
    for (i = 0; i < ruleCount; i++) {
        r = (TimeZoneRule*)orgRules->elementAt(i);
        avail = r->getNextStart(start, res_initial->getRawOffset(), res_initial->getDSTSavings(), FALSE, time);
        done[i] = !avail;
    }

    time = start;
    while (!bFinalStd || !bFinalDst) {
        avail = getNextTransition(time, FALSE, tzt);
        if (!avail) {
            break;
        }
        UDate updatedTime = tzt.getTime();
        if (updatedTime == time) {
            // Can get here if rules for start & end of daylight time have exactly
            // the same time.  
            // TODO:  fix getNextTransition() to prevent it?
            status = U_INVALID_STATE_ERROR;
            goto error;
        }
        time = updatedTime;
 
        const TimeZoneRule *toRule = tzt.getTo();
        for (i = 0; i < ruleCount; i++) {
            r = (TimeZoneRule*)orgRules->elementAt(i);
            if (*r == *toRule) {
                break;
            }
        }
        if (i >= ruleCount) {
            // This case should never happen
            status = U_INVALID_STATE_ERROR;
            goto error;
        }
        if (done[i]) {
            continue;
        }
        const TimeArrayTimeZoneRule *tar = dynamic_cast<const TimeArrayTimeZoneRule *>(toRule);
        const AnnualTimeZoneRule *ar;
        if (tar != NULL) {
            // Get the previous raw offset and DST savings before the very first start time
            TimeZoneTransition tzt0;
            t = start;
            while (TRUE) {
                avail = getNextTransition(t, FALSE, tzt0);
                if (!avail) {
                    break;
                }
                if (*(tzt0.getTo()) == *tar) {
                    break;
                }
                t = tzt0.getTime();
            }
            if (avail) {
                // Check if the entire start times to be added
                tar->getFirstStart(tzt.getFrom()->getRawOffset(), tzt.getFrom()->getDSTSavings(), firstStart);
                if (firstStart > start) {
                    // Just add the rule as is
                    filteredRules->addElement(tar->clone(), status);
                    if (U_FAILURE(status)) {
                        goto error;
                    }
                } else {
                    // Colllect transitions after the start time
                    int32_t startTimes;
                    DateTimeRule::TimeRuleType timeType;
                    int32_t idx;

                    startTimes = tar->countStartTimes();
                    timeType = tar->getTimeType();
                    for (idx = 0; idx < startTimes; idx++) {
                        tar->getStartTimeAt(idx, t);
                        if (timeType == DateTimeRule::STANDARD_TIME) {
                            t -= tzt.getFrom()->getRawOffset();
                        }
                        if (timeType == DateTimeRule::WALL_TIME) {
                            t -= tzt.getFrom()->getDSTSavings();
                        }
                        if (t > start) {
                            break;
                        }
                    }
                    int32_t asize = startTimes - idx;
                    if (asize > 0) {
                        newTimes = (UDate*)uprv_malloc(sizeof(UDate) * asize);
                        if (newTimes == NULL) {
                            status = U_MEMORY_ALLOCATION_ERROR;
                            goto error;
                        }
                        for (int32_t newidx = 0; newidx < asize; newidx++) {
                            tar->getStartTimeAt(idx + newidx, newTimes[newidx]);
                            if (U_FAILURE(status)) {
                                uprv_free(newTimes);
                                newTimes = NULL;
                                goto error;
                            }
                        }
                        tar->getName(name);
                        TimeArrayTimeZoneRule *newTar = new TimeArrayTimeZoneRule(name,
                            tar->getRawOffset(), tar->getDSTSavings(), newTimes, asize, timeType);
                        uprv_free(newTimes);
                        filteredRules->addElement(newTar, status);
                        if (U_FAILURE(status)) {
                            goto error;
                        }
                    }
                }
            }
        } else if ((ar = dynamic_cast<const AnnualTimeZoneRule *>(toRule)) != NULL) {
            ar->getFirstStart(tzt.getFrom()->getRawOffset(), tzt.getFrom()->getDSTSavings(), firstStart);
            if (firstStart == tzt.getTime()) {
                // Just add the rule as is
                filteredRules->addElement(ar->clone(), status);
                if (U_FAILURE(status)) {
                    goto error;
                }
            } else {
                // Calculate the transition year
                int32_t year, month, dom, dow, doy, mid;
                Grego::timeToFields(tzt.getTime(), year, month, dom, dow, doy, mid);
                // Re-create the rule
                ar->getName(name);
                AnnualTimeZoneRule *newAr = new AnnualTimeZoneRule(name, ar->getRawOffset(), ar->getDSTSavings(),
                    *(ar->getRule()), year, ar->getEndYear());
                filteredRules->addElement(newAr, status);
                if (U_FAILURE(status)) {
                    goto error;
                }
            }
            // check if this is a final rule
            if (ar->getEndYear() == AnnualTimeZoneRule::MAX_YEAR) {
                // After bot final standard and dst rules are processed,
                // exit this while loop.
                if (ar->getDSTSavings() == 0) {
                    bFinalStd = TRUE;
                } else {
                    bFinalDst = TRUE;
                }
            }
        }
        done[i] = TRUE;
    }

    // Set the results
    if (orgRules != NULL) {
        while (!orgRules->isEmpty()) {
            r = (TimeZoneRule*)orgRules->orphanElementAt(0);
            delete r;
        }
        delete orgRules;
    }
    if (done != NULL) {
        uprv_free(done);
    }

    initial = res_initial;
    transitionRules = filteredRules;
    return;

error:
    if (orgtrs != NULL) {
        uprv_free(orgtrs);
    }
    if (orgRules != NULL) {
        while (!orgRules->isEmpty()) {
            r = (TimeZoneRule*)orgRules->orphanElementAt(0);
            delete r;
        }
        delete orgRules;
    }
    if (done != NULL) {
        if (filteredRules != NULL) {
            while (!filteredRules->isEmpty()) {
                r = (TimeZoneRule*)filteredRules->orphanElementAt(0);
                delete r;
            }
            delete filteredRules;
        }
        delete res_initial;
        uprv_free(done);
    }

    initial = NULL;
    transitionRules = NULL;
}

void
BasicTimeZone::getOffsetFromLocal(UDate /*date*/, int32_t /*nonExistingTimeOpt*/, int32_t /*duplicatedTimeOpt*/,
                            int32_t& /*rawOffset*/, int32_t& /*dstOffset*/, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    status = U_UNSUPPORTED_ERROR;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2001-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  bocsu.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   Author: Markus W. Scherer
*
*   Modification history:
*   05/18/2001  weiv    Made into separate module
*/




#if !UCONFIG_NO_COLLATION



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2001-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  bocsu.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   Author: Markus W. Scherer
*
*   Modification history:
*   05/18/2001  weiv    Made into separate module
*/




#if !UCONFIG_NO_COLLATION

U_NAMESPACE_BEGIN

class ByteSink;

U_NAMESPACE_END

/*
 * "BOCSU"
 * Binary Ordered Compression Scheme for Unicode
 *
 * Specific application:
 *
 * Encode a Unicode string for the identical level of a sort key.
 * Restrictions:
 * - byte stream (unsigned 8-bit bytes)
 * - lexical order of the identical-level run must be
 *   the same as code point order for the string
 * - avoid byte values 0, 1, 2
 *
 * Method: Slope Detection
 * Remember the previous code point (initial 0).
 * For each cp in the string, encode the difference to the previous one.
 *
 * With a compact encoding of differences, this yields good results for
 * small scripts and UTF-like results otherwise.
 *
 * Encoding of differences:
 * - Similar to a UTF, encoding the length of the byte sequence in the lead bytes.
 * - Does not need to be friendly for decoding or random access
 *   (trail byte values may overlap with lead/single byte values).
 * - The signedness must be encoded as the most significant part.
 *
 * We encode differences with few bytes if their absolute values are small.
 * For correct ordering, we must treat the entire value range -10ffff..+10ffff
 * in ascending order, which forbids encoding the sign and the absolute value separately.
 * Instead, we split the lead byte range in the middle and encode non-negative values
 * going up and negative values going down.
 *
 * For very small absolute values, the difference is added to a middle byte value
 * for single-byte encoded differences.
 * For somewhat larger absolute values, the difference is divided by the number
 * of byte values available, the modulo is used for one trail byte, and the remainder
 * is added to a lead byte avoiding the single-byte range.
 * For large absolute values, the difference is similarly encoded in three bytes.
 *
 * This encoding does not use byte values 0, 1, 2, but uses all other byte values
 * for lead/single bytes so that the middle range of single bytes is as large
 * as possible.
 * Note that the lead byte ranges overlap some, but that the sequences as a whole
 * are well ordered. I.e., even if the lead byte is the same for sequences of different
 * lengths, the trail bytes establish correct order.
 * It would be possible to encode slightly larger ranges for each length (>1) by
 * subtracting the lower bound of the range. However, that would also slow down the
 * calculation.
 *
 * For the actual string encoding, an optimization moves the previous code point value
 * to the middle of its Unicode script block to minimize the differences in
 * same-script text runs.
 */

/* Do not use byte values 0, 1, 2 because they are separators in sort keys. */
#define SLOPE_MIN           3
#define SLOPE_MAX           0xff
#define SLOPE_MIDDLE        0x81

#define SLOPE_TAIL_COUNT    (SLOPE_MAX-SLOPE_MIN+1)

#define SLOPE_MAX_BYTES     4

/*
 * Number of lead bytes:
 * 1        middle byte for 0
 * 2*80=160 single bytes for !=0
 * 2*42=84  for double-byte values
 * 2*3=6    for 3-byte values
 * 2*1=2    for 4-byte values
 *
 * The sum must be <=SLOPE_TAIL_COUNT.
 *
 * Why these numbers?
 * - There should be >=128 single-byte values to cover 128-blocks
 *   with small scripts.
 * - There should be >=20902 single/double-byte values to cover Unihan.
 * - It helps CJK Extension B some if there are 3-byte values that cover
 *   the distance between them and Unihan.
 *   This also helps to jump among distant places in the BMP.
 * - Four-byte values are necessary to cover the rest of Unicode.
 *
 * Symmetrical lead byte counts are for convenience.
 * With an equal distribution of even and odd differences there is also
 * no advantage to asymmetrical lead byte counts.
 */
#define SLOPE_SINGLE        80
#define SLOPE_LEAD_2        42
#define SLOPE_LEAD_3        3
#define SLOPE_LEAD_4        1

/* The difference value range for single-byters. */
#define SLOPE_REACH_POS_1   SLOPE_SINGLE
#define SLOPE_REACH_NEG_1   (-SLOPE_SINGLE)

/* The difference value range for double-byters. */
#define SLOPE_REACH_POS_2   (SLOPE_LEAD_2*SLOPE_TAIL_COUNT+(SLOPE_LEAD_2-1))
#define SLOPE_REACH_NEG_2   (-SLOPE_REACH_POS_2-1)

/* The difference value range for 3-byters. */
#define SLOPE_REACH_POS_3   (SLOPE_LEAD_3*SLOPE_TAIL_COUNT*SLOPE_TAIL_COUNT+(SLOPE_LEAD_3-1)*SLOPE_TAIL_COUNT+(SLOPE_TAIL_COUNT-1))
#define SLOPE_REACH_NEG_3   (-SLOPE_REACH_POS_3-1)

/* The lead byte start values. */
#define SLOPE_START_POS_2   (SLOPE_MIDDLE+SLOPE_SINGLE+1)
#define SLOPE_START_POS_3   (SLOPE_START_POS_2+SLOPE_LEAD_2)

#define SLOPE_START_NEG_2   (SLOPE_MIDDLE+SLOPE_REACH_NEG_1)
#define SLOPE_START_NEG_3   (SLOPE_START_NEG_2-SLOPE_LEAD_2)

/*
 * Integer division and modulo with negative numerators
 * yields negative modulo results and quotients that are one more than
 * what we need here.
 */
#define NEGDIVMOD(n, d, m) UPRV_BLOCK_MACRO_BEGIN { \
    (m)=(n)%(d); \
    (n)/=(d); \
    if((m)<0) { \
        --(n); \
        (m)+=(d); \
    } \
} UPRV_BLOCK_MACRO_END

U_CFUNC UChar32
u_writeIdenticalLevelRun(UChar32 prev, const UChar *s, int32_t length, icu::ByteSink &sink);

#endif /* #if !UCONFIG_NO_COLLATION */



/*
 * encode one difference value -0x10ffff..+0x10ffff in 1..4 bytes,
 * preserving lexical order
 */
static uint8_t *
u_writeDiff(int32_t diff, uint8_t *p) {
    if(diff>=SLOPE_REACH_NEG_1) {
        if(diff<=SLOPE_REACH_POS_1) {
            *p++=(uint8_t)(SLOPE_MIDDLE+diff);
        } else if(diff<=SLOPE_REACH_POS_2) {
            *p++=(uint8_t)(SLOPE_START_POS_2+(diff/SLOPE_TAIL_COUNT));
            *p++=(uint8_t)(SLOPE_MIN+diff%SLOPE_TAIL_COUNT);
        } else if(diff<=SLOPE_REACH_POS_3) {
            p[2]=(uint8_t)(SLOPE_MIN+diff%SLOPE_TAIL_COUNT);
            diff/=SLOPE_TAIL_COUNT;
            p[1]=(uint8_t)(SLOPE_MIN+diff%SLOPE_TAIL_COUNT);
            *p=(uint8_t)(SLOPE_START_POS_3+(diff/SLOPE_TAIL_COUNT));
            p+=3;
        } else {
            p[3]=(uint8_t)(SLOPE_MIN+diff%SLOPE_TAIL_COUNT);
            diff/=SLOPE_TAIL_COUNT;
            p[2]=(uint8_t)(SLOPE_MIN+diff%SLOPE_TAIL_COUNT);
            diff/=SLOPE_TAIL_COUNT;
            p[1]=(uint8_t)(SLOPE_MIN+diff%SLOPE_TAIL_COUNT);
            *p=SLOPE_MAX;
            p+=4;
        }
    } else {
        int32_t m;

        if(diff>=SLOPE_REACH_NEG_2) {
            NEGDIVMOD(diff, SLOPE_TAIL_COUNT, m);
            *p++=(uint8_t)(SLOPE_START_NEG_2+diff);
            *p++=(uint8_t)(SLOPE_MIN+m);
        } else if(diff>=SLOPE_REACH_NEG_3) {
            NEGDIVMOD(diff, SLOPE_TAIL_COUNT, m);
            p[2]=(uint8_t)(SLOPE_MIN+m);
            NEGDIVMOD(diff, SLOPE_TAIL_COUNT, m);
            p[1]=(uint8_t)(SLOPE_MIN+m);
            *p=(uint8_t)(SLOPE_START_NEG_3+diff);
            p+=3;
        } else {
            NEGDIVMOD(diff, SLOPE_TAIL_COUNT, m);
            p[3]=(uint8_t)(SLOPE_MIN+m);
            NEGDIVMOD(diff, SLOPE_TAIL_COUNT, m);
            p[2]=(uint8_t)(SLOPE_MIN+m);
            NEGDIVMOD(diff, SLOPE_TAIL_COUNT, m);
            p[1]=(uint8_t)(SLOPE_MIN+m);
            *p=SLOPE_MIN;
            p+=4;
        }
    }
    return p;
}

/*
 * Encode the code points of a string as
 * a sequence of byte-encoded differences (slope detection),
 * preserving lexical order.
 *
 * Optimize the difference-taking for runs of Unicode text within
 * small scripts:
 *
 * Most small scripts are allocated within aligned 128-blocks of Unicode
 * code points. Lexical order is preserved if "prev" is always moved
 * into the middle of such a block.
 *
 * Additionally, "prev" is moved from anywhere in the Unihan
 * area into the middle of that area.
 * Note that the identical-level run in a sort key is generated from
 * NFD text - there are never Hangul characters included.
 */
U_CFUNC UChar32
u_writeIdenticalLevelRun(UChar32 prev, const UChar *s, int32_t length, icu::ByteSink &sink) {
    char scratch[64];
    int32_t capacity;

    int32_t i=0;
    while(i<length) {
        char *buffer=sink.GetAppendBuffer(1, length*2, scratch, (int32_t)sizeof(scratch), &capacity);
        uint8_t *p;
        // We must have capacity>=SLOPE_MAX_BYTES in case u_writeDiff() writes that much,
        // but we do not want to force the sink.GetAppendBuffer() to allocate
        // for a large min_capacity because we might actually only write one byte.
        if(capacity<16) {
            buffer=scratch;
            capacity=(int32_t)sizeof(scratch);
        }
        p=reinterpret_cast<uint8_t *>(buffer);
        uint8_t *lastSafe=p+capacity-SLOPE_MAX_BYTES;
        while(i<length && p<=lastSafe) {
            if(prev<0x4e00 || prev>=0xa000) {
                prev=(prev&~0x7f)-SLOPE_REACH_NEG_1;
            } else {
                /*
                 * Unihan U+4e00..U+9fa5:
                 * double-bytes down from the upper end
                 */
                prev=0x9fff-SLOPE_REACH_POS_2;
            }

            UChar32 c;
            U16_NEXT(s, i, length, c);
            if(c==0xfffe) {
                *p++=2;  // merge separator
                prev=0;
            } else {
                p=u_writeDiff(c-prev, p);
                prev=c;
            }
        }
        sink.Append(buffer, (int32_t)(p-reinterpret_cast<uint8_t *>(buffer)));
    }
    return prev;
}

#endif /* #if !UCONFIG_NO_COLLATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2003-2013, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* File BUDDHCAL.CPP
*
* Modification History:
*  05/13/2003    srl     copied from gregocal.cpp
*
*/



#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ********************************************************************************
 * Copyright (C) 2003-2013, International Business Machines Corporation
 * and others. All Rights Reserved.
 ********************************************************************************
 *
 * File BUDDHCAL.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   05/13/2003  srl          copied from gregocal.h
 ********************************************************************************
 */




#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
* Copyright (C) 1997-2013, International Business Machines Corporation and others.
* All Rights Reserved.
********************************************************************************
*
* File GREGOCAL.H
*
* Modification History:
*
*   Date        Name        Description
*   04/22/97    aliu        Overhauled header.
*    07/28/98    stephen        Sync with JDK 1.2
*    09/04/98    stephen        Re-sync with JDK 8/31 putback
*    09/14/98    stephen        Changed type of kOneDay, kOneWeek to double.
*                            Fixed bug in roll()
*   10/15/99    aliu        Fixed j31, incorrect WEEK_OF_YEAR computation.
*                           Added documentation of WEEK_OF_YEAR computation.
*   10/15/99    aliu        Fixed j32, cannot set date to Feb 29 2000 AD.
*                           {JDK bug 4210209 4209272}
*   11/07/2003  srl         Update, clean up documentation.
********************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING



/**
 * \file 
 * \brief C++ API: Concrete class which provides the standard calendar.
 */

U_NAMESPACE_BEGIN

/** 
 * Concrete class which provides the standard calendar used by most of the world.
 * <P>
 * The standard (Gregorian) calendar has 2 eras, BC and AD.
 * <P>
 * This implementation handles a single discontinuity, which corresponds by default to
 * the date the Gregorian calendar was originally instituted (October 15, 1582). Not all
 * countries adopted the Gregorian calendar then, so this cutover date may be changed by
 * the caller.
 * <P>
 * Prior to the institution of the Gregorian Calendar, New Year's Day was March 25. To
 * avoid confusion, this Calendar always uses January 1. A manual adjustment may be made
 * if desired for dates that are prior to the Gregorian changeover and which fall
 * between January 1 and March 24.
 *
 * <p>Values calculated for the <code>WEEK_OF_YEAR</code> field range from 1 to
 * 53.  Week 1 for a year is the first week that contains at least
 * <code>getMinimalDaysInFirstWeek()</code> days from that year.  It thus
 * depends on the values of <code>getMinimalDaysInFirstWeek()</code>,
 * <code>getFirstDayOfWeek()</code>, and the day of the week of January 1.
 * Weeks between week 1 of one year and week 1 of the following year are
 * numbered sequentially from 2 to 52 or 53 (as needed).
 *
 * <p>For example, January 1, 1998 was a Thursday.  If
 * <code>getFirstDayOfWeek()</code> is <code>MONDAY</code> and
 * <code>getMinimalDaysInFirstWeek()</code> is 4 (these are the values
 * reflecting ISO 8601 and many national standards), then week 1 of 1998 starts
 * on December 29, 1997, and ends on January 4, 1998.  If, however,
 * <code>getFirstDayOfWeek()</code> is <code>SUNDAY</code>, then week 1 of 1998
 * starts on January 4, 1998, and ends on January 10, 1998; the first three days
 * of 1998 then are part of week 53 of 1997.
 *
 * <p>Example for using GregorianCalendar:
 * <pre>
 * \code
 *     // get the supported ids for GMT-08:00 (Pacific Standard Time)
 *     UErrorCode success = U_ZERO_ERROR;
 *     const StringEnumeration *ids = TimeZone::createEnumeration(-8 * 60 * 60 * 1000);
 *     // if no ids were returned, something is wrong. get out.
 *     if (ids == 0 || ids->count(success) == 0) {
 *         return;
 *     }
 *
 *     // begin output
 *     cout << "Current Time" << endl;
 *
 *     // create a Pacific Standard Time time zone
 *     SimpleTimeZone* pdt = new SimpleTimeZone(-8 * 60 * 60 * 1000, ids->unext(NULL, success)));
 *
 *     // set up rules for daylight savings time
 *     pdt->setStartRule(UCAL_MARCH, 1, UCAL_SUNDAY, 2 * 60 * 60 * 1000);
 *     pdt->setEndRule(UCAL_NOVEMBER, 2, UCAL_SUNDAY, 2 * 60 * 60 * 1000);
 *
 *     // create a GregorianCalendar with the Pacific Daylight time zone
 *     // and the current date and time
 *     Calendar* calendar = new GregorianCalendar( pdt, success );
 *
 *     // print out a bunch of interesting things
 *     cout << "ERA: " << calendar->get( UCAL_ERA, success ) << endl;
 *     cout << "YEAR: " << calendar->get( UCAL_YEAR, success ) << endl;
 *     cout << "MONTH: " << calendar->get( UCAL_MONTH, success ) << endl;
 *     cout << "WEEK_OF_YEAR: " << calendar->get( UCAL_WEEK_OF_YEAR, success ) << endl;
 *     cout << "WEEK_OF_MONTH: " << calendar->get( UCAL_WEEK_OF_MONTH, success ) << endl;
 *     cout << "DATE: " << calendar->get( UCAL_DATE, success ) << endl;
 *     cout << "DAY_OF_MONTH: " << calendar->get( UCAL_DAY_OF_MONTH, success ) << endl;
 *     cout << "DAY_OF_YEAR: " << calendar->get( UCAL_DAY_OF_YEAR, success ) << endl;
 *     cout << "DAY_OF_WEEK: " << calendar->get( UCAL_DAY_OF_WEEK, success ) << endl;
 *     cout << "DAY_OF_WEEK_IN_MONTH: " << calendar->get( UCAL_DAY_OF_WEEK_IN_MONTH, success ) << endl;
 *     cout << "AM_PM: " << calendar->get( UCAL_AM_PM, success ) << endl;
 *     cout << "HOUR: " << calendar->get( UCAL_HOUR, success ) << endl;
 *     cout << "HOUR_OF_DAY: " << calendar->get( UCAL_HOUR_OF_DAY, success ) << endl;
 *     cout << "MINUTE: " << calendar->get( UCAL_MINUTE, success ) << endl;
 *     cout << "SECOND: " << calendar->get( UCAL_SECOND, success ) << endl;
 *     cout << "MILLISECOND: " << calendar->get( UCAL_MILLISECOND, success ) << endl;
 *     cout << "ZONE_OFFSET: " << (calendar->get( UCAL_ZONE_OFFSET, success )/(60*60*1000)) << endl;
 *     cout << "DST_OFFSET: " << (calendar->get( UCAL_DST_OFFSET, success )/(60*60*1000)) << endl;
 *
 *     cout << "Current Time, with hour reset to 3" << endl;
 *     calendar->clear(UCAL_HOUR_OF_DAY); // so doesn't override
 *     calendar->set(UCAL_HOUR, 3);
 *     cout << "ERA: " << calendar->get( UCAL_ERA, success ) << endl;
 *     cout << "YEAR: " << calendar->get( UCAL_YEAR, success ) << endl;
 *     cout << "MONTH: " << calendar->get( UCAL_MONTH, success ) << endl;
 *     cout << "WEEK_OF_YEAR: " << calendar->get( UCAL_WEEK_OF_YEAR, success ) << endl;
 *     cout << "WEEK_OF_MONTH: " << calendar->get( UCAL_WEEK_OF_MONTH, success ) << endl;
 *     cout << "DATE: " << calendar->get( UCAL_DATE, success ) << endl;
 *     cout << "DAY_OF_MONTH: " << calendar->get( UCAL_DAY_OF_MONTH, success ) << endl;
 *     cout << "DAY_OF_YEAR: " << calendar->get( UCAL_DAY_OF_YEAR, success ) << endl;
 *     cout << "DAY_OF_WEEK: " << calendar->get( UCAL_DAY_OF_WEEK, success ) << endl;
 *     cout << "DAY_OF_WEEK_IN_MONTH: " << calendar->get( UCAL_DAY_OF_WEEK_IN_MONTH, success ) << endl;
 *     cout << "AM_PM: " << calendar->get( UCAL_AM_PM, success ) << endl;
 *     cout << "HOUR: " << calendar->get( UCAL_HOUR, success ) << endl;
 *     cout << "HOUR_OF_DAY: " << calendar->get( UCAL_HOUR_OF_DAY, success ) << endl;
 *     cout << "MINUTE: " << calendar->get( UCAL_MINUTE, success ) << endl;
 *     cout << "SECOND: " << calendar->get( UCAL_SECOND, success ) << endl;
 *     cout << "MILLISECOND: " << calendar->get( UCAL_MILLISECOND, success ) << endl;
 *     cout << "ZONE_OFFSET: " << (calendar->get( UCAL_ZONE_OFFSET, success )/(60*60*1000)) << endl; // in hours
 *     cout << "DST_OFFSET: " << (calendar->get( UCAL_DST_OFFSET, success )/(60*60*1000)) << endl; // in hours
 *
 *     if (U_FAILURE(success)) {
 *         cout << "An error occured. success=" << u_errorName(success) << endl;
 *     }
 *
 *     delete ids;
 *     delete calendar; // also deletes pdt
 * \endcode
 * </pre>
 * @stable ICU 2.0
 */
class U_I18N_API GregorianCalendar: public Calendar {
public:

    /**
     * Useful constants for GregorianCalendar and TimeZone.
     * @stable ICU 2.0
     */
    enum EEras {
        BC,
        AD
    };

    /**
     * Constructs a default GregorianCalendar using the current time in the default time
     * zone with the default locale.
     *
     * @param success  Indicates the status of GregorianCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @stable ICU 2.0
     */
    GregorianCalendar(UErrorCode& success);

    /**
     * Constructs a GregorianCalendar based on the current time in the given time zone
     * with the default locale. Clients are no longer responsible for deleting the given
     * time zone object after it's adopted.
     *
     * @param zoneToAdopt     The given timezone.
     * @param success  Indicates the status of GregorianCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @stable ICU 2.0
     */
    GregorianCalendar(TimeZone* zoneToAdopt, UErrorCode& success);

    /**
     * Constructs a GregorianCalendar based on the current time in the given time zone
     * with the default locale.
     *
     * @param zone     The given timezone.
     * @param success  Indicates the status of GregorianCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @stable ICU 2.0
     */
    GregorianCalendar(const TimeZone& zone, UErrorCode& success);

    /**
     * Constructs a GregorianCalendar based on the current time in the default time zone
     * with the given locale.
     *
     * @param aLocale  The given locale.
     * @param success  Indicates the status of GregorianCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @stable ICU 2.0
     */
    GregorianCalendar(const Locale& aLocale, UErrorCode& success);

    /**
     * Constructs a GregorianCalendar based on the current time in the given time zone
     * with the given locale. Clients are no longer responsible for deleting the given
     * time zone object after it's adopted.
     *
     * @param zoneToAdopt     The given timezone.
     * @param aLocale  The given locale.
     * @param success  Indicates the status of GregorianCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @stable ICU 2.0
     */
    GregorianCalendar(TimeZone* zoneToAdopt, const Locale& aLocale, UErrorCode& success);

    /**
     * Constructs a GregorianCalendar based on the current time in the given time zone
     * with the given locale.
     *
     * @param zone     The given timezone.
     * @param aLocale  The given locale.
     * @param success  Indicates the status of GregorianCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @stable ICU 2.0
     */
    GregorianCalendar(const TimeZone& zone, const Locale& aLocale, UErrorCode& success);

    /**
     * Constructs a GregorianCalendar with the given AD date set in the default time
     * zone with the default locale.
     *
     * @param year     The value used to set the YEAR time field in the calendar.
     * @param month    The value used to set the MONTH time field in the calendar. Month
     *                 value is 0-based. e.g., 0 for January.
     * @param date     The value used to set the DATE time field in the calendar.
     * @param success  Indicates the status of GregorianCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @stable ICU 2.0
     */
    GregorianCalendar(int32_t year, int32_t month, int32_t date, UErrorCode& success);

    /**
     * Constructs a GregorianCalendar with the given AD date and time set for the
     * default time zone with the default locale.
     *
     * @param year     The value used to set the YEAR time field in the calendar.
     * @param month    The value used to set the MONTH time field in the calendar. Month
     *                 value is 0-based. e.g., 0 for January.
     * @param date     The value used to set the DATE time field in the calendar.
     * @param hour     The value used to set the HOUR_OF_DAY time field in the calendar.
     * @param minute   The value used to set the MINUTE time field in the calendar.
     * @param success  Indicates the status of GregorianCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @stable ICU 2.0
     */
    GregorianCalendar(int32_t year, int32_t month, int32_t date, int32_t hour, int32_t minute, UErrorCode& success);

    /**
     * Constructs a GregorianCalendar with the given AD date and time set for the
     * default time zone with the default locale.
     *
     * @param year     The value used to set the YEAR time field in the calendar.
     * @param month    The value used to set the MONTH time field in the calendar. Month
     *                 value is 0-based. e.g., 0 for January.
     * @param date     The value used to set the DATE time field in the calendar.
     * @param hour     The value used to set the HOUR_OF_DAY time field in the calendar.
     * @param minute   The value used to set the MINUTE time field in the calendar.
     * @param second   The value used to set the SECOND time field in the calendar.
     * @param success  Indicates the status of GregorianCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @stable ICU 2.0
     */
    GregorianCalendar(int32_t year, int32_t month, int32_t date, int32_t hour, int32_t minute, int32_t second, UErrorCode& success);

    /**
     * Destructor
     * @stable ICU 2.0
     */
    virtual ~GregorianCalendar();

    /**
     * Copy constructor
     * @param source    the object to be copied.
     * @stable ICU 2.0
     */
    GregorianCalendar(const GregorianCalendar& source);

    /**
     * Default assignment operator
     * @param right    the object to be copied.
     * @stable ICU 2.0
     */
    GregorianCalendar& operator=(const GregorianCalendar& right);

    /**
     * Create and return a polymorphic copy of this calendar.
     * @return    return a polymorphic copy of this calendar.
     * @stable ICU 2.0
     */
    virtual GregorianCalendar* clone() const;

    /**
     * Sets the GregorianCalendar change date. This is the point when the switch from
     * Julian dates to Gregorian dates occurred. Default is 00:00:00 local time, October
     * 15, 1582. Previous to this time and date will be Julian dates.
     *
     * @param date     The given Gregorian cutover date.
     * @param success  Output param set to success/failure code on exit.
     * @stable ICU 2.0
     */
    void setGregorianChange(UDate date, UErrorCode& success);

    /**
     * Gets the Gregorian Calendar change date. This is the point when the switch from
     * Julian dates to Gregorian dates occurred. Default is 00:00:00 local time, October
     * 15, 1582. Previous to this time and date will be Julian dates.
     *
     * @return   The Gregorian cutover time for this calendar.
     * @stable ICU 2.0
     */
    UDate getGregorianChange(void) const;

    /**
     * Return true if the given year is a leap year. Determination of whether a year is
     * a leap year is actually very complicated. We do something crude and mostly
     * correct here, but for a real determination you need a lot of contextual
     * information. For example, in Sweden, the change from Julian to Gregorian happened
     * in a complex way resulting in missed leap years and double leap years between
     * 1700 and 1753. Another example is that after the start of the Julian calendar in
     * 45 B.C., the leap years did not regularize until 8 A.D. This method ignores these
     * quirks, and pays attention only to the Julian onset date and the Gregorian
     * cutover (which can be changed).
     *
     * @param year  The given year.
     * @return      True if the given year is a leap year; false otherwise.
     * @stable ICU 2.0
     */
    UBool isLeapYear(int32_t year) const;

    /**
     * Returns TRUE if the given Calendar object is equivalent to this
     * one.  Calendar override.
     *
     * @param other the Calendar to be compared with this Calendar   
     * @stable ICU 2.4
     */
    virtual UBool isEquivalentTo(const Calendar& other) const;

#ifndef U_FORCE_HIDE_DEPRECATED_API
    /**
     * (Overrides Calendar) Rolls up or down by the given amount in the specified field.
     * For more information, see the documentation for Calendar::roll().
     *
     * @param field   The time field.
     * @param amount  Indicates amount to roll.
     * @param status  Output param set to success/failure code on exit. If any value
     *                previously set in the time field is invalid, this will be set to
     *                an error status.
     * @deprecated ICU 2.6. Use roll(UCalendarDateFields field, int32_t amount, UErrorCode& status) instead.
     */
    virtual void roll(EDateFields field, int32_t amount, UErrorCode& status);
#endif  // U_FORCE_HIDE_DEPRECATED_API

    /**
     * (Overrides Calendar) Rolls up or down by the given amount in the specified field.
     * For more information, see the documentation for Calendar::roll().
     *
     * @param field   The time field.
     * @param amount  Indicates amount to roll.
     * @param status  Output param set to success/failure code on exit. If any value
     *                previously set in the time field is invalid, this will be set to
     *                an error status.
     * @stable ICU 2.6.
     */
    virtual void roll(UCalendarDateFields field, int32_t amount, UErrorCode& status);

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Return the minimum value that this field could have, given the current date.
     * For the Gregorian calendar, this is the same as getMinimum() and getGreatestMinimum().
     * @param field    the time field.
     * @return         the minimum value that this field could have, given the current date.
     * @deprecated ICU 2.6. Use getActualMinimum(UCalendarDateFields field) instead.
     */
    int32_t getActualMinimum(EDateFields field) const;

    /**
     * Return the minimum value that this field could have, given the current date.
     * For the Gregorian calendar, this is the same as getMinimum() and getGreatestMinimum().
     * @param field    the time field.
     * @param status
     * @return         the minimum value that this field could have, given the current date.
     * @deprecated ICU 2.6. Use getActualMinimum(UCalendarDateFields field) instead. (Added to ICU 3.0 for signature consistency)
     */
    int32_t getActualMinimum(EDateFields field, UErrorCode& status) const;
#endif  /* U_HIDE_DEPRECATED_API */

    /**
     * Return the minimum value that this field could have, given the current date.
     * For the Gregorian calendar, this is the same as getMinimum() and getGreatestMinimum().
     * @param field    the time field.
     * @param status   error result.
     * @return         the minimum value that this field could have, given the current date.
     * @stable ICU 3.0
     */
    int32_t getActualMinimum(UCalendarDateFields field, UErrorCode &status) const;

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Return the maximum value that this field could have, given the current date.
     * For example, with the date "Feb 3, 1997" and the DAY_OF_MONTH field, the actual
     * maximum would be 28; for "Feb 3, 1996" it s 29.  Similarly for a Hebrew calendar,
     * for some years the actual maximum for MONTH is 12, and for others 13.
     * @param field    the time field.
     * @return         the maximum value that this field could have, given the current date.
     * @deprecated ICU 2.6. Use getActualMaximum(UCalendarDateFields field) instead.
     */
    int32_t getActualMaximum(EDateFields field) const;
#endif  /* U_HIDE_DEPRECATED_API */

    /**
     * Return the maximum value that this field could have, given the current date.
     * For example, with the date "Feb 3, 1997" and the DAY_OF_MONTH field, the actual
     * maximum would be 28; for "Feb 3, 1996" it s 29.  Similarly for a Hebrew calendar,
     * for some years the actual maximum for MONTH is 12, and for others 13.
     * @param field    the time field.
     * @param status   returns any errors that may result from this function call.
     * @return         the maximum value that this field could have, given the current date.
     * @stable ICU 2.6
     */
    virtual int32_t getActualMaximum(UCalendarDateFields field, UErrorCode& status) const;

    /**
     * (Overrides Calendar) Return true if the current date for this Calendar is in
     * Daylight Savings Time. Recognizes DST_OFFSET, if it is set.
     *
     * @param status Fill-in parameter which receives the status of this operation.
     * @return   True if the current date for this Calendar is in Daylight Savings Time,
     *           false, otherwise.
     * @stable ICU 2.0
     */
    virtual UBool inDaylightTime(UErrorCode& status) const;

public:

    /**
     * Override Calendar Returns a unique class ID POLYMORPHICALLY. Pure virtual
     * override. This method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone() methods call
     * this method.
     *
     * @return   The class ID for this object. All objects of a given class have the
     *           same class ID. Objects of other classes have different class IDs.
     * @stable ICU 2.0
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Return the class ID for this class. This is useful only for comparing to a return
     * value from getDynamicClassID(). For example:
     *
     *      Base* polymorphic_pointer = createPolymorphicObject();
     *      if (polymorphic_pointer->getDynamicClassID() ==
     *          Derived::getStaticClassID()) ...
     *
     * @return   The class ID for all objects of this class.
     * @stable ICU 2.0
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns the calendar type name string for this Calendar object.
     * The returned string is the legacy ICU calendar attribute value,
     * for example, "gregorian" or "japanese".
     *
     * For more details see the Calendar::getType() documentation.
     *
     * @return legacy calendar type name string
     * @stable ICU 49
     */
    virtual const char * getType() const;

 private:
    GregorianCalendar(); // default constructor not implemented

 protected:
    /**
     * Return the ERA.  We need a special method for this because the
     * default ERA is AD, but a zero (unset) ERA is BC.
     * @return    the ERA.
     * @internal
     */
    virtual int32_t internalGetEra() const;

    /**
     * Return the Julian day number of day before the first day of the
     * given month in the given extended year.  Subclasses should override
     * this method to implement their calendar system.
     * @param eyear the extended year
     * @param month the zero-based month, or 0 if useMonth is false
     * @param useMonth if false, compute the day before the first day of
     * the given year, otherwise, compute the day before the first day of
     * the given month
     * @return the Julian day number of the day before the first
     * day of the given month and year
     * @internal
     */
    virtual int32_t handleComputeMonthStart(int32_t eyear, int32_t month,
                                                   UBool useMonth) const;

    /**
     * Subclasses may override this.  This method calls
     * handleGetMonthLength() to obtain the calendar-specific month
     * length.
     * @param bestField which field to use to calculate the date 
     * @return julian day specified by calendar fields.
     * @internal
     */
    virtual int32_t handleComputeJulianDay(UCalendarDateFields bestField)  ;

    /**
     * Return the number of days in the given month of the given extended
     * year of this calendar system.  Subclasses should override this
     * method if they can provide a more correct or more efficient
     * implementation than the default implementation in Calendar.
     * @internal
     */
    virtual int32_t handleGetMonthLength(int32_t extendedYear, int32_t month) const;

    /**
     * Return the number of days in the given extended year of this
     * calendar system.  Subclasses should override this method if they can
     * provide a more correct or more efficient implementation than the
     * default implementation in Calendar.
     * @stable ICU 2.0
     */
    virtual int32_t handleGetYearLength(int32_t eyear) const;

    /**
     * return the length of the given month.
     * @param month    the given month.
     * @return    the length of the given month.
     * @internal
     */
    virtual int32_t monthLength(int32_t month) const;

    /**
     * return the length of the month according to the given year.
     * @param month    the given month.
     * @param year     the given year.
     * @return         the length of the month
     * @internal
     */
    virtual int32_t monthLength(int32_t month, int32_t year) const;

#ifndef U_HIDE_INTERNAL_API
    /**
     * return the length of the given year.
     * @param year    the given year.
     * @return        the length of the given year.
     * @internal
     */
    int32_t yearLength(int32_t year) const;
    
    /**
     * return the length of the year field.
     * @return    the length of the year field
     * @internal
     */
    int32_t yearLength(void) const;

    /**
     * After adjustments such as add(MONTH), add(YEAR), we don't want the
     * month to jump around.  E.g., we don't want Jan 31 + 1 month to go to Mar
     * 3, we want it to go to Feb 28.  Adjustments which might run into this
     * problem call this method to retain the proper month.
     * @internal
     */
    void pinDayOfMonth(void);
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Return the day number with respect to the epoch.  January 1, 1970 (Gregorian)
     * is day zero.
     * @param status Fill-in parameter which receives the status of this operation.
     * @return       the day number with respect to the epoch.  
     * @internal
     */
    virtual UDate getEpochDay(UErrorCode& status);

    /**
     * Subclass API for defining limits of different types.
     * Subclasses must implement this method to return limits for the
     * following fields:
     *
     * <pre>UCAL_ERA
     * UCAL_YEAR
     * UCAL_MONTH
     * UCAL_WEEK_OF_YEAR
     * UCAL_WEEK_OF_MONTH
     * UCAL_DATE (DAY_OF_MONTH on Java)
     * UCAL_DAY_OF_YEAR
     * UCAL_DAY_OF_WEEK_IN_MONTH
     * UCAL_YEAR_WOY
     * UCAL_EXTENDED_YEAR</pre>
     *
     * @param field one of the above field numbers
     * @param limitType one of <code>MINIMUM</code>, <code>GREATEST_MINIMUM</code>,
     * <code>LEAST_MAXIMUM</code>, or <code>MAXIMUM</code>
     * @internal
     */
    virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;

    /**
     * Return the extended year defined by the current fields.  This will
     * use the UCAL_EXTENDED_YEAR field or the UCAL_YEAR and supra-year fields (such
     * as UCAL_ERA) specific to the calendar system, depending on which set of
     * fields is newer.
     * @return the extended year
     * @internal
     */
    virtual int32_t handleGetExtendedYear();

    /** 
     * Subclasses may override this to convert from week fields 
     * (YEAR_WOY and WEEK_OF_YEAR) to an extended year in the case
     * where YEAR, EXTENDED_YEAR are not set.
     * The Gregorian implementation assumes a yearWoy in gregorian format, according to the current era.
     * @return the extended year, UCAL_EXTENDED_YEAR
     * @internal
     */
    virtual int32_t handleGetExtendedYearFromWeekFields(int32_t yearWoy, int32_t woy);


    /**
     * Subclasses may override this method to compute several fields
     * specific to each calendar system.  These are:
     *
     * <ul><li>ERA
     * <li>YEAR
     * <li>MONTH
     * <li>DAY_OF_MONTH
     * <li>DAY_OF_YEAR
     * <li>EXTENDED_YEAR</ul>
     *
     * <p>The GregorianCalendar implementation implements
     * a calendar with the specified Julian/Gregorian cutover date.
     * @internal
     */
    virtual void handleComputeFields(int32_t julianDay, UErrorCode &status);

 private:
    /**
     * Compute the julian day number of the given year.
     * @param isGregorian    if true, using Gregorian calendar, otherwise using Julian calendar
     * @param year           the given year.
     * @param isLeap         true if the year is a leap year.       
     * @return 
     */
    static double computeJulianDayOfYear(UBool isGregorian, int32_t year,
                                         UBool& isLeap);
    
    /**
     * Validates the values of the set time fields.  True if they're all valid.
     * @return    True if the set time fields are all valid.
     */
    UBool validateFields(void) const;

    /**
     * Validates the value of the given time field.  True if it's valid.
     */
    UBool boundsCheck(int32_t value, UCalendarDateFields field) const;

    /**
     * Return the pseudo-time-stamp for two fields, given their
     * individual pseudo-time-stamps.  If either of the fields
     * is unset, then the aggregate is unset.  Otherwise, the
     * aggregate is the later of the two stamps.
     * @param stamp_a    One given field.
     * @param stamp_b    Another given field.
     * @return the pseudo-time-stamp for two fields
     */
    int32_t aggregateStamp(int32_t stamp_a, int32_t stamp_b);

    /**
     * The point at which the Gregorian calendar rules are used, measured in
     * milliseconds from the standard epoch.  Default is October 15, 1582
     * (Gregorian) 00:00:00 UTC, that is, October 4, 1582 (Julian) is followed
     * by October 15, 1582 (Gregorian).  This corresponds to Julian day number
     * 2299161. This is measured from the standard epoch, not in Julian Days.
     */
    UDate                fGregorianCutover;

    /**
     * Julian day number of the Gregorian cutover
     */
    int32_t             fCutoverJulianDay;

    /**
     * Midnight, local time (using this Calendar's TimeZone) at or before the
     * gregorianCutover. This is a pure date value with no time of day or
     * timezone component.
     */
    UDate                 fNormalizedGregorianCutover;// = gregorianCutover;

    /**
     * The year of the gregorianCutover, with 0 representing
     * 1 BC, -1 representing 2 BC, etc.
     */
    int32_t fGregorianCutoverYear;// = 1582;

    /**
     * The year of the gregorianCutover, with 0 representing
     * 1 BC, -1 representing 2 BC, etc.
     */
    int32_t fGregorianCutoverJulianDay;// = 2299161;

    /**
     * Converts time as milliseconds to Julian date. The Julian date used here is not a
     * true Julian date, since it is measured from midnight, not noon.
     *
     * @param millis  The given milliseconds.
     * @return        The Julian date number.
     */
    static double millisToJulianDay(UDate millis);

    /**
     * Converts Julian date to time as milliseconds. The Julian date used here is not a
     * true Julian date, since it is measured from midnight, not noon.
     *
     * @param julian  The given Julian date number.
     * @return        Time as milliseconds.
     */
    static UDate julianDayToMillis(double julian);

    /**
     * Used by handleComputeJulianDay() and handleComputeMonthStart().
     * Temporary field indicating whether the calendar is currently Gregorian as opposed to Julian.
     */
    UBool fIsGregorian;

    /**
     * Used by handleComputeJulianDay() and handleComputeMonthStart().
     * Temporary field indicating that the sense of the gregorian cutover should be inverted
     * to handle certain calculations on and around the cutover date.
     */
    UBool fInvertGregorian;


 public: // internal implementation

    /**
     * @return TRUE if this calendar has the notion of a default century
     * @internal 
     */
    virtual UBool haveDefaultCentury() const;

    /**
     * @return the start of the default century
     * @internal
     */
    virtual UDate defaultCenturyStart() const;

    /**
     * @return the beginning year of the default century
     * @internal 
     */
    virtual int32_t defaultCenturyStartYear() const;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



U_NAMESPACE_BEGIN

/**
 * Concrete class which provides the Buddhist calendar.
 * <P>
 * <code>BuddhistCalendar</code> is a subclass of <code>GregorianCalendar</code>
 * that numbers years since the birth of the Buddha.  This is the civil calendar
 * in some predominantly Buddhist countries such as Thailand, and it is used for
 * religious purposes elsewhere.
 * <p>
 * The Buddhist calendar is identical to the Gregorian calendar in all respects
 * except for the year and era.  Years are numbered since the birth of the
 * Buddha in 543 BC (Gregorian), so that 1 AD (Gregorian) is equivalent to 544
 * BE (Buddhist Era) and 1998 AD is 2541 BE.
 * <p>
 * The Buddhist Calendar has only one allowable era: <code>BE</code>.  If the
 * calendar is not in lenient mode (see <code>setLenient</code>), dates before
 * 1/1/1 BE are rejected as an illegal argument.
 * <p>
 * @internal
 */
class BuddhistCalendar : public GregorianCalendar {
public:

    /**
     * Useful constants for BuddhistCalendar.  Only one Era.
     * @internal
     */
    enum EEras {
       BE
    };

    /**
     * Constructs a BuddhistCalendar based on the current time in the default time zone
     * with the given locale.
     *
     * @param aLocale  The given locale.
     * @param success  Indicates the status of BuddhistCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @internal
     */
    BuddhistCalendar(const Locale& aLocale, UErrorCode& success);


    /**
     * Destructor
     * @internal
     */
    virtual ~BuddhistCalendar();

    /**
     * Copy constructor
     * @param source    the object to be copied.
     * @internal
     */
    BuddhistCalendar(const BuddhistCalendar& source);

    /**
     * Default assignment operator
     * @param right    the object to be copied.
     * @internal
     */
    BuddhistCalendar& operator=(const BuddhistCalendar& right);

    /**
     * Create and return a polymorphic copy of this calendar.
     * @return    return a polymorphic copy of this calendar.
     * @internal
     */
    virtual BuddhistCalendar* clone() const;

public:
    /**
     * Override Calendar Returns a unique class ID POLYMORPHICALLY. Pure virtual
     * override. This method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone() methods call
     * this method.
     *
     * @return   The class ID for this object. All objects of a given class have the
     *           same class ID. Objects of other classes have different class IDs.
     * @internal
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Return the class ID for this class. This is useful only for comparing to a return
     * value from getDynamicClassID(). For example:
     *
     *      Base* polymorphic_pointer = createPolymorphicObject();
     *      if (polymorphic_pointer->getDynamicClassID() ==
     *          Derived::getStaticClassID()) ...
     *
     * @return   The class ID for all objects of this class.
     * @internal
     */
    U_I18N_API static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * return the calendar type, "buddhist".
     *
     * @return calendar type
     * @internal
     */
    virtual const char * getType() const;

private:
    BuddhistCalendar(); // default constructor not implemented

 protected:
    /**
     * Return the extended year defined by the current fields.  This will
     * use the UCAL_EXTENDED_YEAR field or the UCAL_YEAR and supra-year fields (such
     * as UCAL_ERA) specific to the calendar system, depending on which set of
     * fields is newer.
     * @return the extended year
     * @internal
     */
    virtual int32_t handleGetExtendedYear();
    /**
     * Subclasses may override this method to compute several fields
     * specific to each calendar system.  
     * @internal
     */
    virtual void handleComputeFields(int32_t julianDay, UErrorCode& status);
    /**
     * Subclass API for defining limits of different types.
     * @param field one of the field numbers
     * @param limitType one of <code>MINIMUM</code>, <code>GREATEST_MINIMUM</code>,
     * <code>LEAST_MAXIMUM</code>, or <code>MAXIMUM</code>
     * @internal
     */
    virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;
        /**
     * Return the Julian day number of day before the first day of the
     * given month in the given extended year.  Subclasses should override
     * this method to implement their calendar system.
     * @param eyear the extended year
     * @param month the zero-based month, or 0 if useMonth is false
     * @param useMonth if false, compute the day before the first day of
     * the given year, otherwise, compute the day before the first day of
     * the given month
     * @param return the Julian day number of the day before the first
     * day of the given month and year
     * @internal
     */
    virtual int32_t handleComputeMonthStart(int32_t eyear, int32_t month,
                                            UBool useMonth) const;

    /**
     * Returns TRUE because the Buddhist Calendar does have a default century
     * @internal
     */
    virtual UBool haveDefaultCentury() const;

    /**
     * Returns the date of the start of the default century
     * @return start of century - in milliseconds since epoch, 1970
     * @internal
     */
    virtual UDate defaultCenturyStart() const;

    /**
     * Returns the year in which the default century begins
     * @internal
     */
    virtual int32_t defaultCenturyStartYear() const;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */




#include <float.h>

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(BuddhistCalendar)

//static const int32_t kMaxEra = 0; // only 1 era

static const int32_t kBuddhistEraStart = -543;  // 544 BC (Gregorian)

static const int32_t buddhcal_kGregorianEpoch = 1970;    // used as the default value of EXTENDED_YEAR

BuddhistCalendar::BuddhistCalendar(const Locale& aLocale, UErrorCode& success)
:   GregorianCalendar(aLocale, success)
{
    setTimeInMillis(getNow(), success); // Call this again now that the vtable is set up properly.
}

BuddhistCalendar::~BuddhistCalendar()
{
}

BuddhistCalendar::BuddhistCalendar(const BuddhistCalendar& source)
: GregorianCalendar(source)
{
}

BuddhistCalendar& BuddhistCalendar::operator= ( const BuddhistCalendar& right)
{
    GregorianCalendar::operator=(right);
    return *this;
}

BuddhistCalendar* BuddhistCalendar::clone() const
{
    return new BuddhistCalendar(*this);
}

const char *BuddhistCalendar::getType() const
{
    return "buddhist";
}

int32_t BuddhistCalendar::handleGetExtendedYear()
{
    // EXTENDED_YEAR in BuddhistCalendar is a Gregorian year.
    // The default value of EXTENDED_YEAR is 1970 (Buddhist 2513)
    int32_t year;
    if (newerField(UCAL_EXTENDED_YEAR, UCAL_YEAR) == UCAL_EXTENDED_YEAR) {
        year = internalGet(UCAL_EXTENDED_YEAR, buddhcal_kGregorianEpoch);
    } else {
        // extended year is a gregorian year, where 1 = 1AD,  0 = 1BC, -1 = 2BC, etc
        year = internalGet(UCAL_YEAR, buddhcal_kGregorianEpoch - kBuddhistEraStart)
                + kBuddhistEraStart;
    }
    return year;
}

int32_t BuddhistCalendar::handleComputeMonthStart(int32_t eyear, int32_t month,

                                                  UBool useMonth) const
{
    return GregorianCalendar::handleComputeMonthStart(eyear, month, useMonth);
}

void BuddhistCalendar::handleComputeFields(int32_t julianDay, UErrorCode& status)
{
    GregorianCalendar::handleComputeFields(julianDay, status);
    int32_t y = internalGet(UCAL_EXTENDED_YEAR) - kBuddhistEraStart;
    internalSet(UCAL_ERA, 0);
    internalSet(UCAL_YEAR, y);
}

int32_t BuddhistCalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const
{
    if(field == UCAL_ERA) {
        return BE;
    } else {
        return GregorianCalendar::handleGetLimit(field,limitType);
    }
}

#if 0
void BuddhistCalendar::timeToFields(UDate theTime, UBool quick, UErrorCode& status)
{
    //Calendar::timeToFields(theTime, quick, status);

    int32_t era = internalGet(UCAL_ERA);
    int32_t year = internalGet(UCAL_YEAR);

    if(era == GregorianCalendar::BC) {
        year = 1-year;
        era = BuddhistCalendar::BE;
    } else if(era == GregorianCalendar::AD) {
        era = BuddhistCalendar::BE;
    } else {
        status = U_INTERNAL_PROGRAM_ERROR;
    }

    year = year - kBuddhistEraStart;

    internalSet(UCAL_ERA, era);
    internalSet(UCAL_YEAR, year);
}
#endif

/**
 * The system maintains a static default century start date.  This is initialized
 * the first time it is used. Once the system default century date and year
 * are set, they do not change.
 */
static UDate     buddhcal_gSystemDefaultCenturyStart       = DBL_MIN;
static int32_t   buddhcal_gSystemDefaultCenturyStartYear   = -1;
static icu::UInitOnce gBCInitOnce = U_INITONCE_INITIALIZER;


UBool BuddhistCalendar::haveDefaultCentury() const
{
    return TRUE;
}

static void U_CALLCONV
initializeBuddhCalSystemDefaultCentury()
{
    // initialize systemDefaultCentury and systemDefaultCenturyYear based
    // on the current time.  They'll be set to 80 years before
    // the current time.
    UErrorCode status = U_ZERO_ERROR;
    BuddhistCalendar calendar(Locale("@calendar=buddhist"),status);
    if (U_SUCCESS(status)) {
        calendar.setTime(Calendar::getNow(), status);
        calendar.add(UCAL_YEAR, -80, status);
        UDate    newStart =  calendar.getTime(status);
        int32_t  newYear  =  calendar.get(UCAL_YEAR, status);
        buddhcal_gSystemDefaultCenturyStartYear = newYear;
        buddhcal_gSystemDefaultCenturyStart = newStart;
    }
    // We have no recourse upon failure unless we want to propagate the failure
    // out.
}

UDate BuddhistCalendar::defaultCenturyStart() const
{
    // lazy-evaluate systemDefaultCenturyStart and systemDefaultCenturyStartYear
    umtx_initOnce(gBCInitOnce, &initializeBuddhCalSystemDefaultCentury);
    return buddhcal_gSystemDefaultCenturyStart;
}

int32_t BuddhistCalendar::defaultCenturyStartYear() const
{
    // lazy-evaluate systemDefaultCenturyStartYear and systemDefaultCenturyStart
    umtx_initOnce(gBCInitOnce, &initializeBuddhCalSystemDefaultCentury);
    return buddhcal_gSystemDefaultCenturyStartYear;
}


U_NAMESPACE_END

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2016, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* File CALENDAR.CPP
*
* Modification History:
*
*   Date        Name        Description
*   02/03/97    clhuang     Creation.
*   04/22/97    aliu        Cleaned up, fixed memory leak, made
*                           setWeekCountData() more robust.
*                           Moved platform code to TPlatformUtilities.
*   05/01/97    aliu        Made equals(), before(), after() arguments const.
*   05/20/97    aliu        Changed logic of when to compute fields and time
*                           to fix bugs.
*   08/12/97    aliu        Added equivalentTo.  Misc other fixes.
*   07/28/98    stephen     Sync up with JDK 1.2
*   09/02/98    stephen     Sync with JDK 1.2 8/31 build (getActualMin/Max)
*   03/17/99    stephen     Changed adoptTimeZone() - now fAreFieldsSet is
*                           set to FALSE to force update of time.
*******************************************************************************
*/

  // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ********************************************************************************
 * Copyright (C) 1997-2013, International Business Machines                     *
 * Corporation and others. All Rights Reserved.                                 *
 ********************************************************************************
 *
 * File SIMPLETZ.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   04/21/97    aliu        Overhauled header.
 *   08/10/98    stephen     JDK 1.2 sync
 *                           Added setStartRule() / setEndRule() overloads
 *                           Added hasSameRules()
 *   09/02/98    stephen     Added getOffset(monthLen)
 *                           Changed getOffset() to take UErrorCode
 *   07/09/99    stephen     Removed millisPerHour (unused, for HP compiler)
 *   12/02/99    aliu        Added TimeMode and constructor and setStart/EndRule
 *                           methods that take TimeMode. Added to docs.
 ********************************************************************************
 */




#if U_SHOW_CPLUSPLUS_API

/**
 * \file 
 * \brief C++ API: SimpleTimeZone is a concrete subclass of TimeZone.
 */
 
#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

// forward declaration
class InitialTimeZoneRule;
class TimeZoneTransition;
class AnnualTimeZoneRule;

/**
 * <code>SimpleTimeZone</code> is a concrete subclass of <code>TimeZone</code>
 * that represents a time zone for use with a Gregorian calendar. This
 * class does not handle historical changes.
 * <P>
 * When specifying daylight-savings-time begin and end dates, use a negative value for
 * <code>dayOfWeekInMonth</code> to indicate that <code>SimpleTimeZone</code> should
 * count from the end of the month backwards. For example, if Daylight Savings
 * Time starts or ends at the last Sunday a month, use <code>dayOfWeekInMonth = -1</code>
 * along with <code>dayOfWeek = UCAL_SUNDAY</code> to specify the rule.
 *
 * @see      Calendar
 * @see      GregorianCalendar
 * @see      TimeZone
 * @author   D. Goldsmith, Mark Davis, Chen-Lieh Huang, Alan Liu
 */
class U_I18N_API SimpleTimeZone: public BasicTimeZone {
public:

    /**
     * TimeMode is used, together with a millisecond offset after
     * midnight, to specify a rule transition time.  Most rules
     * transition at a local wall time, that is, according to the
     * current time in effect, either standard, or DST.  However, some
     * rules transition at local standard time, and some at a specific
     * UTC time.  Although it might seem that all times could be
     * converted to wall time, thus eliminating the need for this
     * parameter, this is not the case.
     * @stable ICU 2.0
     */
    enum TimeMode {
        WALL_TIME = 0,
        STANDARD_TIME,
        UTC_TIME
    };

    /**
     * Copy constructor
     * @param source the object to be copied.
     * @stable ICU 2.0
     */
    SimpleTimeZone(const SimpleTimeZone& source);

    /**
     * Default assignment operator
     * @param right    the object to be copied.
     * @stable ICU 2.0
     */
    SimpleTimeZone& operator=(const SimpleTimeZone& right);

    /**
     * Destructor
     * @stable ICU 2.0
     */
    virtual ~SimpleTimeZone();

    /**
     * Returns true if the two TimeZone objects are equal; that is, they have
     * the same ID, raw GMT offset, and DST rules.
     *
     * @param that  The SimpleTimeZone object to be compared with.
     * @return      True if the given time zone is equal to this time zone; false
     *              otherwise.
     * @stable ICU 2.0
     */
    virtual UBool operator==(const TimeZone& that) const;

    /**
     * Constructs a SimpleTimeZone with the given raw GMT offset and time zone ID,
     * and which doesn't observe daylight savings time.  Normally you should use
     * TimeZone::createInstance() to create a TimeZone instead of creating a
     * SimpleTimeZone directly with this constructor.
     *
     * @param rawOffsetGMT  The given base time zone offset to GMT.
     * @param ID         The timezone ID which is obtained from
     *                   TimeZone.getAvailableIDs.
     * @stable ICU 2.0
     */
    SimpleTimeZone(int32_t rawOffsetGMT, const UnicodeString& ID);

    /**
     * Construct a SimpleTimeZone with the given raw GMT offset, time zone ID,
     * and times to start and end daylight savings time. To create a TimeZone that
     * doesn't observe daylight savings time, don't use this constructor; use
     * SimpleTimeZone(rawOffset, ID) instead. Normally, you should use
     * TimeZone.createInstance() to create a TimeZone instead of creating a
     * SimpleTimeZone directly with this constructor.
     * <P>
     * Various types of daylight-savings time rules can be specfied by using different
     * values for startDay and startDayOfWeek and endDay and endDayOfWeek.  For a
     * complete explanation of how these parameters work, see the documentation for
     * setStartRule().
     *
     * @param rawOffsetGMT      The new SimpleTimeZone's raw GMT offset
     * @param ID                The new SimpleTimeZone's time zone ID.
     * @param savingsStartMonth The daylight savings starting month. Month is
     *                          0-based. eg, 0 for January.
     * @param savingsStartDayOfWeekInMonth   The daylight savings starting
     *                          day-of-week-in-month. See setStartRule() for a
     *                          complete explanation.
     * @param savingsStartDayOfWeek The daylight savings starting day-of-week.
     *                          See setStartRule() for a complete explanation.
     * @param savingsStartTime  The daylight savings starting time, expressed as the
     *                          number of milliseconds after midnight.
     * @param savingsEndMonth   The daylight savings ending month. Month is
     *                          0-based. eg, 0 for January.
     * @param savingsEndDayOfWeekInMonth     The daylight savings ending day-of-week-in-month.
     *                          See setStartRule() for a complete explanation.
     * @param savingsEndDayOfWeek The daylight savings ending day-of-week.
     *                          See setStartRule() for a complete explanation.
     * @param savingsEndTime    The daylight savings ending time, expressed as the
     *                          number of milliseconds after midnight.
     * @param status            An UErrorCode to receive the status.
     * @stable ICU 2.0
     */
    SimpleTimeZone(int32_t rawOffsetGMT, const UnicodeString& ID,
        int8_t savingsStartMonth, int8_t savingsStartDayOfWeekInMonth,
        int8_t savingsStartDayOfWeek, int32_t savingsStartTime,
        int8_t savingsEndMonth, int8_t savingsEndDayOfWeekInMonth,
        int8_t savingsEndDayOfWeek, int32_t savingsEndTime,
        UErrorCode& status);
    /**
     * Construct a SimpleTimeZone with the given raw GMT offset, time zone ID,
     * and times to start and end daylight savings time. To create a TimeZone that
     * doesn't observe daylight savings time, don't use this constructor; use
     * SimpleTimeZone(rawOffset, ID) instead. Normally, you should use
     * TimeZone.createInstance() to create a TimeZone instead of creating a
     * SimpleTimeZone directly with this constructor.
     * <P>
     * Various types of daylight-savings time rules can be specfied by using different
     * values for startDay and startDayOfWeek and endDay and endDayOfWeek.  For a
     * complete explanation of how these parameters work, see the documentation for
     * setStartRule().
     *
     * @param rawOffsetGMT      The new SimpleTimeZone's raw GMT offset
     * @param ID                The new SimpleTimeZone's time zone ID.
     * @param savingsStartMonth The daylight savings starting month. Month is
     *                          0-based. eg, 0 for January.
     * @param savingsStartDayOfWeekInMonth   The daylight savings starting
     *                          day-of-week-in-month. See setStartRule() for a
     *                          complete explanation.
     * @param savingsStartDayOfWeek The daylight savings starting day-of-week.
     *                          See setStartRule() for a complete explanation.
     * @param savingsStartTime  The daylight savings starting time, expressed as the
     *                          number of milliseconds after midnight.
     * @param savingsEndMonth   The daylight savings ending month. Month is
     *                          0-based. eg, 0 for January.
     * @param savingsEndDayOfWeekInMonth     The daylight savings ending day-of-week-in-month.
     *                          See setStartRule() for a complete explanation.
     * @param savingsEndDayOfWeek The daylight savings ending day-of-week.
     *                          See setStartRule() for a complete explanation.
     * @param savingsEndTime    The daylight savings ending time, expressed as the
     *                          number of milliseconds after midnight.
     * @param savingsDST        The number of milliseconds added to standard time
     *                          to get DST time. Default is one hour.
     * @param status            An UErrorCode to receive the status.
     * @stable ICU 2.0
     */
    SimpleTimeZone(int32_t rawOffsetGMT, const UnicodeString& ID,
        int8_t savingsStartMonth, int8_t savingsStartDayOfWeekInMonth,
        int8_t savingsStartDayOfWeek, int32_t savingsStartTime,
        int8_t savingsEndMonth, int8_t savingsEndDayOfWeekInMonth,
        int8_t savingsEndDayOfWeek, int32_t savingsEndTime,
        int32_t savingsDST, UErrorCode& status);

    /**
     * Construct a SimpleTimeZone with the given raw GMT offset, time zone ID,
     * and times to start and end daylight savings time. To create a TimeZone that
     * doesn't observe daylight savings time, don't use this constructor; use
     * SimpleTimeZone(rawOffset, ID) instead. Normally, you should use
     * TimeZone.createInstance() to create a TimeZone instead of creating a
     * SimpleTimeZone directly with this constructor.
     * <P>
     * Various types of daylight-savings time rules can be specfied by using different
     * values for startDay and startDayOfWeek and endDay and endDayOfWeek.  For a
     * complete explanation of how these parameters work, see the documentation for
     * setStartRule().
     *
     * @param rawOffsetGMT      The new SimpleTimeZone's raw GMT offset
     * @param ID                The new SimpleTimeZone's time zone ID.
     * @param savingsStartMonth The daylight savings starting month. Month is
     *                          0-based. eg, 0 for January.
     * @param savingsStartDayOfWeekInMonth   The daylight savings starting
     *                          day-of-week-in-month. See setStartRule() for a
     *                          complete explanation.
     * @param savingsStartDayOfWeek The daylight savings starting day-of-week.
     *                          See setStartRule() for a complete explanation.
     * @param savingsStartTime  The daylight savings starting time, expressed as the
     *                          number of milliseconds after midnight.
     * @param savingsStartTimeMode Whether the start time is local wall time, local
     *                          standard time, or UTC time. Default is local wall time.
     * @param savingsEndMonth   The daylight savings ending month. Month is
     *                          0-based. eg, 0 for January.
     * @param savingsEndDayOfWeekInMonth     The daylight savings ending day-of-week-in-month.
     *                          See setStartRule() for a complete explanation.
     * @param savingsEndDayOfWeek The daylight savings ending day-of-week.
     *                          See setStartRule() for a complete explanation.
     * @param savingsEndTime    The daylight savings ending time, expressed as the
     *                          number of milliseconds after midnight.
     * @param savingsEndTimeMode Whether the end time is local wall time, local
     *                          standard time, or UTC time. Default is local wall time.
     * @param savingsDST        The number of milliseconds added to standard time
     *                          to get DST time. Default is one hour.
     * @param status            An UErrorCode to receive the status.
     * @stable ICU 2.0
     */
    SimpleTimeZone(int32_t rawOffsetGMT, const UnicodeString& ID,
        int8_t savingsStartMonth, int8_t savingsStartDayOfWeekInMonth,
        int8_t savingsStartDayOfWeek, int32_t savingsStartTime,
        TimeMode savingsStartTimeMode,
        int8_t savingsEndMonth, int8_t savingsEndDayOfWeekInMonth,
        int8_t savingsEndDayOfWeek, int32_t savingsEndTime, TimeMode savingsEndTimeMode,
        int32_t savingsDST, UErrorCode& status);

    /**
     * Sets the daylight savings starting year, that is, the year this time zone began
     * observing its specified daylight savings time rules.  The time zone is considered
     * not to observe daylight savings time prior to that year; SimpleTimeZone doesn't
     * support historical daylight-savings-time rules.
     * @param year the daylight savings starting year.
     * @stable ICU 2.0
     */
    void setStartYear(int32_t year);

    /**
     * Sets the daylight savings starting rule. For example, in the U.S., Daylight Savings
     * Time starts at the second Sunday in March, at 2 AM in standard time.
     * Therefore, you can set the start rule by calling:
     * setStartRule(UCAL_MARCH, 2, UCAL_SUNDAY, 2*60*60*1000);
     * The dayOfWeekInMonth and dayOfWeek parameters together specify how to calculate
     * the exact starting date.  Their exact meaning depend on their respective signs,
     * allowing various types of rules to be constructed, as follows:
     * <ul>
     *   <li>If both dayOfWeekInMonth and dayOfWeek are positive, they specify the
     *       day of week in the month (e.g., (2, WEDNESDAY) is the second Wednesday
     *       of the month).</li>
     *   <li>If dayOfWeek is positive and dayOfWeekInMonth is negative, they specify
     *       the day of week in the month counting backward from the end of the month.
     *       (e.g., (-1, MONDAY) is the last Monday in the month)</li>
     *   <li>If dayOfWeek is zero and dayOfWeekInMonth is positive, dayOfWeekInMonth
     *       specifies the day of the month, regardless of what day of the week it is.
     *       (e.g., (10, 0) is the tenth day of the month)</li>
     *   <li>If dayOfWeek is zero and dayOfWeekInMonth is negative, dayOfWeekInMonth
     *       specifies the day of the month counting backward from the end of the
     *       month, regardless of what day of the week it is (e.g., (-2, 0) is the
     *       next-to-last day of the month).</li>
     *   <li>If dayOfWeek is negative and dayOfWeekInMonth is positive, they specify the
     *       first specified day of the week on or after the specfied day of the month.
     *       (e.g., (15, -SUNDAY) is the first Sunday after the 15th of the month
     *       [or the 15th itself if the 15th is a Sunday].)</li>
     *   <li>If dayOfWeek and DayOfWeekInMonth are both negative, they specify the
     *       last specified day of the week on or before the specified day of the month.
     *       (e.g., (-20, -TUESDAY) is the last Tuesday before the 20th of the month
     *       [or the 20th itself if the 20th is a Tuesday].)</li>
     * </ul>
     * @param month the daylight savings starting month. Month is 0-based.
     * eg, 0 for January.
     * @param dayOfWeekInMonth the daylight savings starting
     * day-of-week-in-month. Please see the member description for an example.
     * @param dayOfWeek the daylight savings starting day-of-week. Please see
     * the member description for an example.
     * @param time the daylight savings starting time. Please see the member
     * description for an example.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setStartRule(int32_t month, int32_t dayOfWeekInMonth, int32_t dayOfWeek,
                      int32_t time, UErrorCode& status);
    /**
     * Sets the daylight savings starting rule. For example, in the U.S., Daylight Savings
     * Time starts at the second Sunday in March, at 2 AM in standard time.
     * Therefore, you can set the start rule by calling:
     * setStartRule(UCAL_MARCH, 2, UCAL_SUNDAY, 2*60*60*1000);
     * The dayOfWeekInMonth and dayOfWeek parameters together specify how to calculate
     * the exact starting date.  Their exact meaning depend on their respective signs,
     * allowing various types of rules to be constructed, as follows:
     * <ul>
     *   <li>If both dayOfWeekInMonth and dayOfWeek are positive, they specify the
     *       day of week in the month (e.g., (2, WEDNESDAY) is the second Wednesday
     *       of the month).</li>
     *   <li>If dayOfWeek is positive and dayOfWeekInMonth is negative, they specify
     *       the day of week in the month counting backward from the end of the month.
     *       (e.g., (-1, MONDAY) is the last Monday in the month)</li>
     *   <li>If dayOfWeek is zero and dayOfWeekInMonth is positive, dayOfWeekInMonth
     *       specifies the day of the month, regardless of what day of the week it is.
     *       (e.g., (10, 0) is the tenth day of the month)</li>
     *   <li>If dayOfWeek is zero and dayOfWeekInMonth is negative, dayOfWeekInMonth
     *       specifies the day of the month counting backward from the end of the
     *       month, regardless of what day of the week it is (e.g., (-2, 0) is the
     *       next-to-last day of the month).</li>
     *   <li>If dayOfWeek is negative and dayOfWeekInMonth is positive, they specify the
     *       first specified day of the week on or after the specfied day of the month.
     *       (e.g., (15, -SUNDAY) is the first Sunday after the 15th of the month
     *       [or the 15th itself if the 15th is a Sunday].)</li>
     *   <li>If dayOfWeek and DayOfWeekInMonth are both negative, they specify the
     *       last specified day of the week on or before the specified day of the month.
     *       (e.g., (-20, -TUESDAY) is the last Tuesday before the 20th of the month
     *       [or the 20th itself if the 20th is a Tuesday].)</li>
     * </ul>
     * @param month the daylight savings starting month. Month is 0-based.
     * eg, 0 for January.
     * @param dayOfWeekInMonth the daylight savings starting
     * day-of-week-in-month. Please see the member description for an example.
     * @param dayOfWeek the daylight savings starting day-of-week. Please see
     * the member description for an example.
     * @param time the daylight savings starting time. Please see the member
     * description for an example.
     * @param mode whether the time is local wall time, local standard time,
     * or UTC time. Default is local wall time.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setStartRule(int32_t month, int32_t dayOfWeekInMonth, int32_t dayOfWeek,
                      int32_t time, TimeMode mode, UErrorCode& status);

    /**
     * Sets the DST start rule to a fixed date within a month.
     *
     * @param month         The month in which this rule occurs (0-based).
     * @param dayOfMonth    The date in that month (1-based).
     * @param time          The time of that day (number of millis after midnight)
     *                      when DST takes effect in local wall time, which is
     *                      standard time in this case.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setStartRule(int32_t month, int32_t dayOfMonth, int32_t time,
                      UErrorCode& status);
    /**
     * Sets the DST start rule to a fixed date within a month.
     *
     * @param month         The month in which this rule occurs (0-based).
     * @param dayOfMonth    The date in that month (1-based).
     * @param time          The time of that day (number of millis after midnight)
     *                      when DST takes effect in local wall time, which is
     *                      standard time in this case.
     * @param mode whether the time is local wall time, local standard time,
     * or UTC time. Default is local wall time.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setStartRule(int32_t month, int32_t dayOfMonth, int32_t time,
                      TimeMode mode, UErrorCode& status);

    /**
     * Sets the DST start rule to a weekday before or after a give date within
     * a month, e.g., the first Monday on or after the 8th.
     *
     * @param month         The month in which this rule occurs (0-based).
     * @param dayOfMonth    A date within that month (1-based).
     * @param dayOfWeek     The day of the week on which this rule occurs.
     * @param time          The time of that day (number of millis after midnight)
     *                      when DST takes effect in local wall time, which is
     *                      standard time in this case.
     * @param after         If true, this rule selects the first dayOfWeek on
     *                      or after dayOfMonth.  If false, this rule selects
     *                      the last dayOfWeek on or before dayOfMonth.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setStartRule(int32_t month, int32_t dayOfMonth, int32_t dayOfWeek,
                      int32_t time, UBool after, UErrorCode& status);
    /**
     * Sets the DST start rule to a weekday before or after a give date within
     * a month, e.g., the first Monday on or after the 8th.
     *
     * @param month         The month in which this rule occurs (0-based).
     * @param dayOfMonth    A date within that month (1-based).
     * @param dayOfWeek     The day of the week on which this rule occurs.
     * @param time          The time of that day (number of millis after midnight)
     *                      when DST takes effect in local wall time, which is
     *                      standard time in this case.
     * @param mode whether the time is local wall time, local standard time,
     * or UTC time. Default is local wall time.
     * @param after         If true, this rule selects the first dayOfWeek on
     *                      or after dayOfMonth.  If false, this rule selects
     *                      the last dayOfWeek on or before dayOfMonth.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setStartRule(int32_t month, int32_t dayOfMonth, int32_t dayOfWeek,
                      int32_t time, TimeMode mode, UBool after, UErrorCode& status);

    /**
     * Sets the daylight savings ending rule. For example, if Daylight
     * Savings Time ends at the last (-1) Sunday in October, at 2 AM in standard time.
     * Therefore, you can set the end rule by calling:
     * <pre>
     *    setEndRule(UCAL_OCTOBER, -1, UCAL_SUNDAY, 2*60*60*1000);
     * </pre>
     * Various other types of rules can be specified by manipulating the dayOfWeek
     * and dayOfWeekInMonth parameters.  For complete details, see the documentation
     * for setStartRule().
     *
     * @param month the daylight savings ending month. Month is 0-based.
     * eg, 0 for January.
     * @param dayOfWeekInMonth the daylight savings ending
     * day-of-week-in-month. See setStartRule() for a complete explanation.
     * @param dayOfWeek the daylight savings ending day-of-week. See setStartRule()
     * for a complete explanation.
     * @param time the daylight savings ending time. Please see the member
     * description for an example.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setEndRule(int32_t month, int32_t dayOfWeekInMonth, int32_t dayOfWeek,
                    int32_t time, UErrorCode& status);

    /**
     * Sets the daylight savings ending rule. For example, if Daylight
     * Savings Time ends at the last (-1) Sunday in October, at 2 AM in standard time.
     * Therefore, you can set the end rule by calling:
     * <pre>
     *    setEndRule(UCAL_OCTOBER, -1, UCAL_SUNDAY, 2*60*60*1000);
     * </pre>
     * Various other types of rules can be specified by manipulating the dayOfWeek
     * and dayOfWeekInMonth parameters.  For complete details, see the documentation
     * for setStartRule().
     *
     * @param month the daylight savings ending month. Month is 0-based.
     * eg, 0 for January.
     * @param dayOfWeekInMonth the daylight savings ending
     * day-of-week-in-month. See setStartRule() for a complete explanation.
     * @param dayOfWeek the daylight savings ending day-of-week. See setStartRule()
     * for a complete explanation.
     * @param time the daylight savings ending time. Please see the member
     * description for an example.
     * @param mode whether the time is local wall time, local standard time,
     * or UTC time. Default is local wall time.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setEndRule(int32_t month, int32_t dayOfWeekInMonth, int32_t dayOfWeek,
                    int32_t time, TimeMode mode, UErrorCode& status);

    /**
     * Sets the DST end rule to a fixed date within a month.
     *
     * @param month         The month in which this rule occurs (0-based).
     * @param dayOfMonth    The date in that month (1-based).
     * @param time          The time of that day (number of millis after midnight)
     *                      when DST ends in local wall time, which is daylight
     *                      time in this case.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setEndRule(int32_t month, int32_t dayOfMonth, int32_t time, UErrorCode& status);

    /**
     * Sets the DST end rule to a fixed date within a month.
     *
     * @param month         The month in which this rule occurs (0-based).
     * @param dayOfMonth    The date in that month (1-based).
     * @param time          The time of that day (number of millis after midnight)
     *                      when DST ends in local wall time, which is daylight
     *                      time in this case.
     * @param mode whether the time is local wall time, local standard time,
     * or UTC time. Default is local wall time.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setEndRule(int32_t month, int32_t dayOfMonth, int32_t time,
                    TimeMode mode, UErrorCode& status);

    /**
     * Sets the DST end rule to a weekday before or after a give date within
     * a month, e.g., the first Monday on or after the 8th.
     *
     * @param month         The month in which this rule occurs (0-based).
     * @param dayOfMonth    A date within that month (1-based).
     * @param dayOfWeek     The day of the week on which this rule occurs.
     * @param time          The time of that day (number of millis after midnight)
     *                      when DST ends in local wall time, which is daylight
     *                      time in this case.
     * @param after         If true, this rule selects the first dayOfWeek on
     *                      or after dayOfMonth.  If false, this rule selects
     *                      the last dayOfWeek on or before dayOfMonth.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setEndRule(int32_t month, int32_t dayOfMonth, int32_t dayOfWeek,
                    int32_t time, UBool after, UErrorCode& status);

    /**
     * Sets the DST end rule to a weekday before or after a give date within
     * a month, e.g., the first Monday on or after the 8th.
     *
     * @param month         The month in which this rule occurs (0-based).
     * @param dayOfMonth    A date within that month (1-based).
     * @param dayOfWeek     The day of the week on which this rule occurs.
     * @param time          The time of that day (number of millis after midnight)
     *                      when DST ends in local wall time, which is daylight
     *                      time in this case.
     * @param mode whether the time is local wall time, local standard time,
     * or UTC time. Default is local wall time.
     * @param after         If true, this rule selects the first dayOfWeek on
     *                      or after dayOfMonth.  If false, this rule selects
     *                      the last dayOfWeek on or before dayOfMonth.
     * @param status An UErrorCode
     * @stable ICU 2.0
     */
    void setEndRule(int32_t month, int32_t dayOfMonth, int32_t dayOfWeek,
                    int32_t time, TimeMode mode, UBool after, UErrorCode& status);

    /**
     * Returns the TimeZone's adjusted GMT offset (i.e., the number of milliseconds to add
     * to GMT to get local time in this time zone, taking daylight savings time into
     * account) as of a particular reference date.  The reference date is used to determine
     * whether daylight savings time is in effect and needs to be figured into the offset
     * that is returned (in other words, what is the adjusted GMT offset in this time zone
     * at this particular date and time?).  For the time zones produced by createTimeZone(),
     * the reference data is specified according to the Gregorian calendar, and the date
     * and time fields are in GMT, NOT local time.
     *
     * @param era        The reference date's era
     * @param year       The reference date's year
     * @param month      The reference date's month (0-based; 0 is January)
     * @param day        The reference date's day-in-month (1-based)
     * @param dayOfWeek  The reference date's day-of-week (1-based; 1 is Sunday)
     * @param millis     The reference date's milliseconds in day, UTT (NOT local time).
     * @param status     An UErrorCode to receive the status.
     * @return           The offset in milliseconds to add to GMT to get local time.
     * @stable ICU 2.0
     */
    virtual int32_t getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                              uint8_t dayOfWeek, int32_t millis, UErrorCode& status) const;

    /**
     * Gets the time zone offset, for current date, modified in case of
     * daylight savings. This is the offset to add *to* UTC to get local time.
     * @param era the era of the given date.
     * @param year the year in the given date.
     * @param month the month in the given date.
     * Month is 0-based. e.g., 0 for January.
     * @param day the day-in-month of the given date.
     * @param dayOfWeek the day-of-week of the given date.
     * @param milliseconds the millis in day in <em>standard</em> local time.
     * @param monthLength the length of the given month in days.
     * @param status     An UErrorCode to receive the status.
     * @return the offset to add *to* GMT to get local time.
     * @stable ICU 2.0
     */
    virtual int32_t getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                           uint8_t dayOfWeek, int32_t milliseconds,
                           int32_t monthLength, UErrorCode& status) const;
    /**
     * Gets the time zone offset, for current date, modified in case of
     * daylight savings. This is the offset to add *to* UTC to get local time.
     * @param era the era of the given date.
     * @param year the year in the given date.
     * @param month the month in the given date.
     * Month is 0-based. e.g., 0 for January.
     * @param day the day-in-month of the given date.
     * @param dayOfWeek the day-of-week of the given date.
     * @param milliseconds the millis in day in <em>standard</em> local time.
     * @param monthLength the length of the given month in days.
     * @param prevMonthLength length of the previous month in days.
     * @param status     An UErrorCode to receive the status.
     * @return the offset to add *to* GMT to get local time.
     * @stable ICU 2.0
     */
    virtual int32_t getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                              uint8_t dayOfWeek, int32_t milliseconds,
                              int32_t monthLength, int32_t prevMonthLength,
                              UErrorCode& status) const;

    /**
     * Redeclared TimeZone method.  This implementation simply calls
     * the base class method, which otherwise would be hidden.
     * @stable ICU 2.8
     */
    virtual void getOffset(UDate date, UBool local, int32_t& rawOffset,
                           int32_t& dstOffset, UErrorCode& ec) const;

    /**
     * Get time zone offsets from local wall time.
     * @internal
     */
    virtual void getOffsetFromLocal(UDate date, int32_t nonExistingTimeOpt, int32_t duplicatedTimeOpt,
        int32_t& rawOffset, int32_t& dstOffset, UErrorCode& status) const;

    /**
     * Returns the TimeZone's raw GMT offset (i.e., the number of milliseconds to add
     * to GMT to get local time, before taking daylight savings time into account).
     *
     * @return   The TimeZone's raw GMT offset.
     * @stable ICU 2.0
     */
    virtual int32_t getRawOffset(void) const;

    /**
     * Sets the TimeZone's raw GMT offset (i.e., the number of milliseconds to add
     * to GMT to get local time, before taking daylight savings time into account).
     *
     * @param offsetMillis  The new raw GMT offset for this time zone.
     * @stable ICU 2.0
     */
    virtual void setRawOffset(int32_t offsetMillis);

    /**
     * Sets the amount of time in ms that the clock is advanced during DST.
     * @param millisSavedDuringDST the number of milliseconds the time is
     * advanced with respect to standard time when the daylight savings rules
     * are in effect. Typically one hour (+3600000). The amount could be negative,
     * but not 0.
     * @param status  An UErrorCode to receive the status.
     * @stable ICU 2.0
     */
    void setDSTSavings(int32_t millisSavedDuringDST, UErrorCode& status);

    /**
     * Returns the amount of time in ms that the clock is advanced during DST.
     * @return the number of milliseconds the time is
     * advanced with respect to standard time when the daylight savings rules
     * are in effect. Typically one hour (+3600000). The amount could be negative,
     * but not 0.
     * @stable ICU 2.0
     */
    virtual int32_t getDSTSavings(void) const;

    /**
     * Queries if this TimeZone uses Daylight Savings Time.
     *
     * @return   True if this TimeZone uses Daylight Savings Time; false otherwise.
     * @stable ICU 2.0
     */
    virtual UBool useDaylightTime(void) const;

#ifndef U_FORCE_HIDE_DEPRECATED_API
    /**
     * Returns true if the given date is within the period when daylight savings time
     * is in effect; false otherwise.  If the TimeZone doesn't observe daylight savings
     * time, this functions always returns false.
     * This method is wasteful since it creates a new GregorianCalendar and
     * deletes it each time it is called. This is a deprecated method
     * and provided only for Java compatibility.
     *
     * @param date The date to test.
     * @param status  An UErrorCode to receive the status.
     * @return true if the given date is in Daylight Savings Time;
     * false otherwise.
     * @deprecated ICU 2.4. Use Calendar::inDaylightTime() instead.
     */
    virtual UBool inDaylightTime(UDate date, UErrorCode& status) const;
#endif  // U_FORCE_HIDE_DEPRECATED_API

    /**
     * Return true if this zone has the same rules and offset as another zone.
     * @param other the TimeZone object to be compared with
     * @return true if the given zone has the same rules and offset as this one
     * @stable ICU 2.0
     */
    UBool hasSameRules(const TimeZone& other) const;

    /**
     * Clones TimeZone objects polymorphically. Clients are responsible for deleting
     * the TimeZone object cloned.
     *
     * @return   A new copy of this TimeZone object.
     * @stable ICU 2.0
     */
    virtual SimpleTimeZone* clone() const;

    /**
     * Gets the first time zone transition after the base time.
     * @param base      The base time.
     * @param inclusive Whether the base time is inclusive or not.
     * @param result    Receives the first transition after the base time.
     * @return  TRUE if the transition is found.
     * @stable ICU 3.8
     */
    virtual UBool getNextTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const;

    /**
     * Gets the most recent time zone transition before the base time.
     * @param base      The base time.
     * @param inclusive Whether the base time is inclusive or not.
     * @param result    Receives the most recent transition before the base time.
     * @return  TRUE if the transition is found.
     * @stable ICU 3.8
     */
    virtual UBool getPreviousTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const;

    /**
     * Returns the number of <code>TimeZoneRule</code>s which represents time transitions,
     * for this time zone, that is, all <code>TimeZoneRule</code>s for this time zone except
     * <code>InitialTimeZoneRule</code>.  The return value range is 0 or any positive value.
     * @param status    Receives error status code.
     * @return The number of <code>TimeZoneRule</code>s representing time transitions.
     * @stable ICU 3.8
     */
    virtual int32_t countTransitionRules(UErrorCode& status) const;

    /**
     * Gets the <code>InitialTimeZoneRule</code> and the set of <code>TimeZoneRule</code>
     * which represent time transitions for this time zone.  On successful return,
     * the argument initial points to non-NULL <code>InitialTimeZoneRule</code> and
     * the array trsrules is filled with 0 or multiple <code>TimeZoneRule</code>
     * instances up to the size specified by trscount.  The results are referencing the
     * rule instance held by this time zone instance.  Therefore, after this time zone
     * is destructed, they are no longer available.
     * @param initial       Receives the initial timezone rule
     * @param trsrules      Receives the timezone transition rules
     * @param trscount      On input, specify the size of the array 'transitions' receiving
     *                      the timezone transition rules.  On output, actual number of
     *                      rules filled in the array will be set.
     * @param status        Receives error status code.
     * @stable ICU 3.8
     */
    virtual void getTimeZoneRules(const InitialTimeZoneRule*& initial,
        const TimeZoneRule* trsrules[], int32_t& trscount, UErrorCode& status) const;


public:

    /**
     * Override TimeZone Returns a unique class ID POLYMORPHICALLY. Pure virtual
     * override. This method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone() methods call
     * this method.
     *
     * @return   The class ID for this object. All objects of a given class have the
     *           same class ID. Objects of other classes have different class IDs.
     * @stable ICU 2.0
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Return the class ID for this class. This is useful only for comparing to a return
     * value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       Derived::getStaticClassID()) ...
     * </pre>
     * @return   The class ID for all objects of this class.
     * @stable ICU 2.0
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

private:
    /**
     * Constants specifying values of startMode and endMode.
     */
    enum EMode
    {
        DOM_MODE = 1,
        DOW_IN_MONTH_MODE,
        DOW_GE_DOM_MODE,
        DOW_LE_DOM_MODE
    };

    SimpleTimeZone(); // default constructor not implemented

    /**
     * Internal construction method.
     * @param rawOffsetGMT    The new SimpleTimeZone's raw GMT offset
     * @param startMonth      the month DST starts
     * @param startDay        the day DST starts
     * @param startDayOfWeek  the DOW DST starts
     * @param startTime       the time DST starts
     * @param startTimeMode   Whether the start time is local wall time, local
     *                        standard time, or UTC time. Default is local wall time.
     * @param endMonth        the month DST ends
     * @param endDay          the day DST ends
     * @param endDayOfWeek    the DOW DST ends
     * @param endTime         the time DST ends
     * @param endTimeMode     Whether the end time is local wall time, local
     *                        standard time, or UTC time. Default is local wall time.
     * @param dstSavings      The number of milliseconds added to standard time
     *                        to get DST time. Default is one hour.
     * @param status          An UErrorCode to receive the status.
     */
    void construct(int32_t rawOffsetGMT,
                   int8_t startMonth, int8_t startDay, int8_t startDayOfWeek,
                   int32_t startTime, TimeMode startTimeMode,
                   int8_t endMonth, int8_t endDay, int8_t endDayOfWeek,
                   int32_t endTime, TimeMode endTimeMode,
                   int32_t dstSavings, UErrorCode& status);

    /**
     * Compare a given date in the year to a rule. Return 1, 0, or -1, depending
     * on whether the date is after, equal to, or before the rule date. The
     * millis are compared directly against the ruleMillis, so any
     * standard-daylight adjustments must be handled by the caller.
     *
     * @return  1 if the date is after the rule date, -1 if the date is before
     *          the rule date, or 0 if the date is equal to the rule date.
     */
    static int32_t compareToRule(int8_t month, int8_t monthLen, int8_t prevMonthLen,
                                 int8_t dayOfMonth,
                                 int8_t dayOfWeek, int32_t millis, int32_t millisDelta,
                                 EMode ruleMode, int8_t ruleMonth, int8_t ruleDayOfWeek,
                                 int8_t ruleDay, int32_t ruleMillis);

    /**
     * Given a set of encoded rules in startDay and startDayOfMonth, decode
     * them and set the startMode appropriately.  Do the same for endDay and
     * endDayOfMonth.
     * <P>
     * Upon entry, the day of week variables may be zero or
     * negative, in order to indicate special modes.  The day of month
     * variables may also be negative.
     * <P>
     * Upon exit, the mode variables will be
     * set, and the day of week and day of month variables will be positive.
     * <P>
     * This method also recognizes a startDay or endDay of zero as indicating
     * no DST.
     */
    void decodeRules(UErrorCode& status);
    void decodeStartRule(UErrorCode& status);
    void decodeEndRule(UErrorCode& status);

    int8_t startMonth, startDay, startDayOfWeek;   // the month, day, DOW, and time DST starts
    int32_t startTime;
    TimeMode startTimeMode, endTimeMode; // Mode for startTime, endTime; see TimeMode
    int8_t endMonth, endDay, endDayOfWeek; // the month, day, DOW, and time DST ends
    int32_t endTime;
    int32_t startYear;  // the year these DST rules took effect
    int32_t rawOffset;  // the TimeZone's raw GMT offset
    UBool useDaylight; // flag indicating whether this TimeZone uses DST
    static const int8_t STATICMONTHLENGTH[12]; // lengths of the months
    EMode startMode, endMode;   // flags indicating what kind of rules the DST rules are

    /**
     * A positive value indicating the amount of time saved during DST in ms.
     * Typically one hour; sometimes 30 minutes.
     */
    int32_t dstSavings;

    /* Private for BasicTimeZone implementation */
    void checkTransitionRules(UErrorCode& status) const;
    void initTransitionRules(UErrorCode& status);
    void clearTransitionRules(void);
    void deleteTransitionRules(void);
    UBool   transitionRulesInitialized;
    InitialTimeZoneRule*    initialRule;
    TimeZoneTransition*     firstTransition;
    AnnualTimeZoneRule*     stdRule;
    AnnualTimeZoneRule*     dstRule;
};

inline void SimpleTimeZone::setStartRule(int32_t month, int32_t dayOfWeekInMonth,
                                         int32_t dayOfWeek,
                                         int32_t time, UErrorCode& status) {
    setStartRule(month, dayOfWeekInMonth, dayOfWeek, time, WALL_TIME, status);
}

inline void SimpleTimeZone::setStartRule(int32_t month, int32_t dayOfMonth,
                                         int32_t time,
                                         UErrorCode& status) {
    setStartRule(month, dayOfMonth, time, WALL_TIME, status);
}

inline void SimpleTimeZone::setStartRule(int32_t month, int32_t dayOfMonth,
                                         int32_t dayOfWeek,
                                         int32_t time, UBool after, UErrorCode& status) {
    setStartRule(month, dayOfMonth, dayOfWeek, time, WALL_TIME, after, status);
}

inline void SimpleTimeZone::setEndRule(int32_t month, int32_t dayOfWeekInMonth,
                                       int32_t dayOfWeek,
                                       int32_t time, UErrorCode& status) {
    setEndRule(month, dayOfWeekInMonth, dayOfWeek, time, WALL_TIME, status);
}

inline void SimpleTimeZone::setEndRule(int32_t month, int32_t dayOfMonth,
                                       int32_t time, UErrorCode& status) {
    setEndRule(month, dayOfMonth, time, WALL_TIME, status);
}

inline void SimpleTimeZone::setEndRule(int32_t month, int32_t dayOfMonth, int32_t dayOfWeek,
                                       int32_t time, UBool after, UErrorCode& status) {
    setEndRule(month, dayOfMonth, dayOfWeek, time, WALL_TIME, after, status);
}

inline void
SimpleTimeZone::getOffset(UDate date, UBool local, int32_t& rawOffsetRef,
                          int32_t& dstOffsetRef, UErrorCode& ec) const {
    TimeZone::getOffset(date, local, rawOffsetRef, dstOffsetRef, ec);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2013, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/



#if U_SHOW_CPLUSPLUS_API

/**
 * \file 
 * \brief C++ API: Rule based customizable time zone
 */

#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

// forward declaration
class UVector;
struct Transition;

/**
 * a BasicTimeZone subclass implemented in terms of InitialTimeZoneRule and TimeZoneRule instances
 * @see BasicTimeZone
 * @see InitialTimeZoneRule
 * @see TimeZoneRule
 */
class U_I18N_API RuleBasedTimeZone : public BasicTimeZone {
public:
    /**
     * Constructs a <code>RuleBasedTimeZone</code> object with the ID and the
     * <code>InitialTimeZoneRule</code>.  The input <code>InitialTimeZoneRule</code>
     * is adopted by this <code>RuleBasedTimeZone</code>, thus the caller must not
     * delete it.
     * @param id                The time zone ID.
     * @param initialRule       The initial time zone rule.
     * @stable ICU 3.8
     */
    RuleBasedTimeZone(const UnicodeString& id, InitialTimeZoneRule* initialRule);

    /**
     * Copy constructor.
     * @param source    The RuleBasedTimeZone object to be copied.
     * @stable ICU 3.8
     */
    RuleBasedTimeZone(const RuleBasedTimeZone& source);

    /**
     * Destructor.
     * @stable ICU 3.8
     */
    virtual ~RuleBasedTimeZone();

    /**
     * Assignment operator.
     * @param right The object to be copied.
     * @stable ICU 3.8
     */
    RuleBasedTimeZone& operator=(const RuleBasedTimeZone& right);

    /**
     * Return true if the given <code>TimeZone</code> objects are
     * semantically equal. Objects of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZone</code> objects are
      *semantically equal.
     * @stable ICU 3.8
     */
    virtual UBool operator==(const TimeZone& that) const;

    /**
     * Return true if the given <code>TimeZone</code> objects are
     * semantically unequal. Objects of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZone</code> objects are
     * semantically unequal.
     * @stable ICU 3.8
     */
    virtual UBool operator!=(const TimeZone& that) const;

    /**
     * Adds the <code>TimeZoneRule</code> which represents time transitions.
     * The <code>TimeZoneRule</code> must have start times, that is, the result
     * of isTransitionRule() must be true. Otherwise, U_ILLEGAL_ARGUMENT_ERROR
     * is set to the error code.
     * The input <code>TimeZoneRule</code> is adopted by this
     * <code>RuleBasedTimeZone</code> on successful completion of this method,
     * thus, the caller must not delete it when no error is returned.
     * After all rules are added, the caller must call complete() method to
     * make this <code>RuleBasedTimeZone</code> ready to handle common time
     * zone functions.
     * @param rule The <code>TimeZoneRule</code>.
     * @param status Output param to filled in with a success or an error.
     * @stable ICU 3.8
     */
    void addTransitionRule(TimeZoneRule* rule, UErrorCode& status);

    /**
     * Makes the <code>TimeZoneRule</code> ready to handle actual timezone
     * calcuation APIs.  This method collects time zone rules specified
     * by the caller via the constructor and addTransitionRule() and
     * builds internal structure for making the object ready to support
     * time zone APIs such as getOffset(), getNextTransition() and others.
     * @param status Output param to filled in with a success or an error.
     * @stable ICU 3.8
     */
    void complete(UErrorCode& status);

    /**
     * Clones TimeZone objects polymorphically. Clients are responsible for deleting
     * the TimeZone object cloned.
     *
     * @return   A new copy of this TimeZone object.
     * @stable ICU 3.8
     */
    virtual RuleBasedTimeZone* clone() const;

    /**
     * Returns the TimeZone's adjusted GMT offset (i.e., the number of milliseconds to add
     * to GMT to get local time in this time zone, taking daylight savings time into
     * account) as of a particular reference date.  The reference date is used to determine
     * whether daylight savings time is in effect and needs to be figured into the offset
     * that is returned (in other words, what is the adjusted GMT offset in this time zone
     * at this particular date and time?).  For the time zones produced by createTimeZone(),
     * the reference data is specified according to the Gregorian calendar, and the date
     * and time fields are local standard time.
     *
     * <p>Note: Don't call this method. Instead, call the getOffset(UDate...) overload,
     * which returns both the raw and the DST offset for a given time. This method
     * is retained only for backward compatibility.
     *
     * @param era        The reference date's era
     * @param year       The reference date's year
     * @param month      The reference date's month (0-based; 0 is January)
     * @param day        The reference date's day-in-month (1-based)
     * @param dayOfWeek  The reference date's day-of-week (1-based; 1 is Sunday)
     * @param millis     The reference date's milliseconds in day, local standard time
     * @param status     Output param to filled in with a success or an error.
     * @return           The offset in milliseconds to add to GMT to get local time.
     * @stable ICU 3.8
     */
    virtual int32_t getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                              uint8_t dayOfWeek, int32_t millis, UErrorCode& status) const;

    /**
     * Gets the time zone offset, for current date, modified in case of
     * daylight savings. This is the offset to add *to* UTC to get local time.
     *
     * <p>Note: Don't call this method. Instead, call the getOffset(UDate...) overload,
     * which returns both the raw and the DST offset for a given time. This method
     * is retained only for backward compatibility.
     *
     * @param era        The reference date's era
     * @param year       The reference date's year
     * @param month      The reference date's month (0-based; 0 is January)
     * @param day        The reference date's day-in-month (1-based)
     * @param dayOfWeek  The reference date's day-of-week (1-based; 1 is Sunday)
     * @param millis     The reference date's milliseconds in day, local standard time
     * @param monthLength The length of the given month in days.
     * @param status     Output param to filled in with a success or an error.
     * @return           The offset in milliseconds to add to GMT to get local time.
     * @stable ICU 3.8
     */
    virtual int32_t getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                           uint8_t dayOfWeek, int32_t millis,
                           int32_t monthLength, UErrorCode& status) const;

    /**
     * Returns the time zone raw and GMT offset for the given moment
     * in time.  Upon return, local-millis = GMT-millis + rawOffset +
     * dstOffset.  All computations are performed in the proleptic
     * Gregorian calendar.  The default implementation in the TimeZone
     * class delegates to the 8-argument getOffset().
     *
     * @param date moment in time for which to return offsets, in
     * units of milliseconds from January 1, 1970 0:00 GMT, either GMT
     * time or local wall time, depending on `local'.
     * @param local if true, `date' is local wall time; otherwise it
     * is in GMT time.
     * @param rawOffset output parameter to receive the raw offset, that
     * is, the offset not including DST adjustments
     * @param dstOffset output parameter to receive the DST offset,
     * that is, the offset to be added to `rawOffset' to obtain the
     * total offset between local and GMT time. If DST is not in
     * effect, this value is zero; otherwise it is a positive value,
     * typically one hour.
     * @param ec input-output error code
     * @stable ICU 3.8
     */
    virtual void getOffset(UDate date, UBool local, int32_t& rawOffset,
                           int32_t& dstOffset, UErrorCode& ec) const;

    /**
     * Sets the TimeZone's raw GMT offset (i.e., the number of milliseconds to add
     * to GMT to get local time, before taking daylight savings time into account).
     *
     * @param offsetMillis  The new raw GMT offset for this time zone.
     * @stable ICU 3.8
     */
    virtual void setRawOffset(int32_t offsetMillis);

    /**
     * Returns the TimeZone's raw GMT offset (i.e., the number of milliseconds to add
     * to GMT to get local time, before taking daylight savings time into account).
     *
     * @return   The TimeZone's raw GMT offset.
     * @stable ICU 3.8
     */
    virtual int32_t getRawOffset(void) const;

    /**
     * Queries if this time zone uses daylight savings time.
     * @return true if this time zone uses daylight savings time,
     * false, otherwise.
     * @stable ICU 3.8
     */
    virtual UBool useDaylightTime(void) const;

#ifndef U_FORCE_HIDE_DEPRECATED_API
    /**
     * Queries if the given date is in daylight savings time in
     * this time zone.
     * This method is wasteful since it creates a new GregorianCalendar and
     * deletes it each time it is called. This is a deprecated method
     * and provided only for Java compatibility.
     *
     * @param date the given UDate.
     * @param status Output param filled in with success/error code.
     * @return true if the given date is in daylight savings time,
     * false, otherwise.
     * @deprecated ICU 2.4. Use Calendar::inDaylightTime() instead.
     */
    virtual UBool inDaylightTime(UDate date, UErrorCode& status) const;
#endif  // U_FORCE_HIDE_DEPRECATED_API

    /**
     * Returns true if this zone has the same rule and offset as another zone.
     * That is, if this zone differs only in ID, if at all.
     * @param other the <code>TimeZone</code> object to be compared with
     * @return true if the given zone is the same as this one,
     * with the possible exception of the ID
     * @stable ICU 3.8
     */
    virtual UBool hasSameRules(const TimeZone& other) const;

    /**
     * Gets the first time zone transition after the base time.
     * @param base      The base time.
     * @param inclusive Whether the base time is inclusive or not.
     * @param result    Receives the first transition after the base time.
     * @return  TRUE if the transition is found.
     * @stable ICU 3.8
     */
    virtual UBool getNextTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const;

    /**
     * Gets the most recent time zone transition before the base time.
     * @param base      The base time.
     * @param inclusive Whether the base time is inclusive or not.
     * @param result    Receives the most recent transition before the base time.
     * @return  TRUE if the transition is found.
     * @stable ICU 3.8
     */
    virtual UBool getPreviousTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const;

    /**
     * Returns the number of <code>TimeZoneRule</code>s which represents time transitions,
     * for this time zone, that is, all <code>TimeZoneRule</code>s for this time zone except
     * <code>InitialTimeZoneRule</code>.  The return value range is 0 or any positive value.
     * @param status    Receives error status code.
     * @return The number of <code>TimeZoneRule</code>s representing time transitions.
     * @stable ICU 3.8
     */
    virtual int32_t countTransitionRules(UErrorCode& status) const;

    /**
     * Gets the <code>InitialTimeZoneRule</code> and the set of <code>TimeZoneRule</code>
     * which represent time transitions for this time zone.  On successful return,
     * the argument initial points to non-NULL <code>InitialTimeZoneRule</code> and
     * the array trsrules is filled with 0 or multiple <code>TimeZoneRule</code>
     * instances up to the size specified by trscount.  The results are referencing the
     * rule instance held by this time zone instance.  Therefore, after this time zone
     * is destructed, they are no longer available.
     * @param initial       Receives the initial timezone rule
     * @param trsrules      Receives the timezone transition rules
     * @param trscount      On input, specify the size of the array 'transitions' receiving
     *                      the timezone transition rules.  On output, actual number of
     *                      rules filled in the array will be set.
     * @param status        Receives error status code.
     * @stable ICU 3.8
     */
    virtual void getTimeZoneRules(const InitialTimeZoneRule*& initial,
        const TimeZoneRule* trsrules[], int32_t& trscount, UErrorCode& status) const;

    /**
     * Get time zone offsets from local wall time.
     * @internal
     */
    virtual void getOffsetFromLocal(UDate date, int32_t nonExistingTimeOpt, int32_t duplicatedTimeOpt,
        int32_t& rawOffset, int32_t& dstOffset, UErrorCode& status) const;

private:
    void deleteRules(void);
    void deleteTransitions(void);
    UVector* copyRules(UVector* source);
    TimeZoneRule* findRuleInFinal(UDate date, UBool local,
        int32_t NonExistingTimeOpt, int32_t DuplicatedTimeOpt) const;
    UBool findNext(UDate base, UBool inclusive, UDate& time, TimeZoneRule*& from, TimeZoneRule*& to) const;
    UBool findPrev(UDate base, UBool inclusive, UDate& time, TimeZoneRule*& from, TimeZoneRule*& to) const;
    int32_t getLocalDelta(int32_t rawBefore, int32_t dstBefore, int32_t rawAfter, int32_t dstAfter,
        int32_t NonExistingTimeOpt, int32_t DuplicatedTimeOpt) const;
    UDate getTransitionTime(Transition* transition, UBool local,
        int32_t NonExistingTimeOpt, int32_t DuplicatedTimeOpt) const;
    void getOffsetInternal(UDate date, UBool local, int32_t NonExistingTimeOpt, int32_t DuplicatedTimeOpt,
        int32_t& rawOffset, int32_t& dstOffset, UErrorCode& ec) const;
    void completeConst(UErrorCode &status) const;

    InitialTimeZoneRule *fInitialRule;
    UVector             *fHistoricRules;
    UVector             *fFinalRules;
    UVector             *fHistoricTransitions;
    UBool               fUpToDate;

public:
    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 3.8
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 3.8
     */
    virtual UClassID getDynamicClassID(void) const;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2013, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if U_SHOW_CPLUSPLUS_API

/**
 * \file 
 * \brief C++ API: RFC2445 VTIMEZONE support
 */

#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

class VTZWriter;
class VTZReader;
class UVector;

/**
 * <code>VTimeZone</code> is a class implementing RFC2445 VTIMEZONE.  You can create a
 * <code>VTimeZone</code> instance from a time zone ID supported by <code>TimeZone</code>.
 * With the <code>VTimeZone</code> instance created from the ID, you can write out the rule
 * in RFC2445 VTIMEZONE format.  Also, you can create a <code>VTimeZone</code> instance
 * from RFC2445 VTIMEZONE data stream, which allows you to calculate time
 * zone offset by the rules defined by the data. Or, you can create a
 * <code>VTimeZone</code> from any other ICU <code>BasicTimeZone</code>.
 * <br><br>
 * Note: The consumer of this class reading or writing VTIMEZONE data is responsible to
 * decode or encode Non-ASCII text.  Methods reading/writing VTIMEZONE data in this class
 * do nothing with MIME encoding.
 * @stable ICU 3.8
 */
class U_I18N_API VTimeZone : public BasicTimeZone {
public:
    /**
     * Copy constructor.
     * @param source    The <code>VTimeZone</code> object to be copied.
     * @stable ICU 3.8
     */
    VTimeZone(const VTimeZone& source);

    /**
     * Destructor.
     * @stable ICU 3.8
     */
    virtual ~VTimeZone();

    /**
     * Assignment operator.
     * @param right The object to be copied.
     * @stable ICU 3.8
     */
    VTimeZone& operator=(const VTimeZone& right);

    /**
     * Return true if the given <code>TimeZone</code> objects are
     * semantically equal. Objects of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZone</code> objects are
      *semantically equal.
     * @stable ICU 3.8
     */
    virtual UBool operator==(const TimeZone& that) const;

    /**
     * Return true if the given <code>TimeZone</code> objects are
     * semantically unequal. Objects of different subclasses are considered unequal.
     * @param that  The object to be compared with.
     * @return  true if the given <code>TimeZone</code> objects are
     * semantically unequal.
     * @stable ICU 3.8
     */
    virtual UBool operator!=(const TimeZone& that) const;

    /**
     * Create a <code>VTimeZone</code> instance by the time zone ID.
     * @param ID The time zone ID, such as America/New_York
     * @return A <code>VTimeZone</code> object initialized by the time zone ID,
     * or NULL when the ID is unknown.
     * @stable ICU 3.8
     */
    static VTimeZone* createVTimeZoneByID(const UnicodeString& ID);

    /**
     * Create a <code>VTimeZone</code> instance using a basic time zone.
     * @param basicTZ The basic time zone instance
     * @param status Output param to filled in with a success or an error.
     * @return A <code>VTimeZone</code> object initialized by the basic time zone.
     * @stable ICU 4.6
     */
    static VTimeZone* createVTimeZoneFromBasicTimeZone(const BasicTimeZone& basicTZ,
                                                       UErrorCode &status);

    /**
     * Create a <code>VTimeZone</code> instance by RFC2445 VTIMEZONE data
     * 
     * @param vtzdata The string including VTIMEZONE data block
     * @param status Output param to filled in with a success or an error.
     * @return A <code>VTimeZone</code> initialized by the VTIMEZONE data or
     * NULL if failed to load the rule from the VTIMEZONE data.
     * @stable ICU 3.8
     */
    static VTimeZone* createVTimeZone(const UnicodeString& vtzdata, UErrorCode& status);

    /**
     * Gets the RFC2445 TZURL property value.  When a <code>VTimeZone</code> instance was
     * created from VTIMEZONE data, the initial value is set by the TZURL property value
     * in the data.  Otherwise, the initial value is not set.
     * @param url Receives the RFC2445 TZURL property value.
     * @return TRUE if TZURL attribute is available and value is set.
     * @stable ICU 3.8
     */
    UBool getTZURL(UnicodeString& url) const;

    /**
     * Sets the RFC2445 TZURL property value.
     * @param url The TZURL property value.
     * @stable ICU 3.8
     */
    void setTZURL(const UnicodeString& url);

    /**
     * Gets the RFC2445 LAST-MODIFIED property value.  When a <code>VTimeZone</code> instance
     * was created from VTIMEZONE data, the initial value is set by the LAST-MODIFIED property
     * value in the data.  Otherwise, the initial value is not set.
     * @param lastModified Receives the last modified date.
     * @return TRUE if lastModified attribute is available and value is set.
     * @stable ICU 3.8
     */
    UBool getLastModified(UDate& lastModified) const;

    /**
     * Sets the RFC2445 LAST-MODIFIED property value.
     * @param lastModified The LAST-MODIFIED date.
     * @stable ICU 3.8
     */
    void setLastModified(UDate lastModified);

    /**
     * Writes RFC2445 VTIMEZONE data for this time zone
     * @param result Output param to filled in with the VTIMEZONE data.
     * @param status Output param to filled in with a success or an error.
     * @stable ICU 3.8
     */
    void write(UnicodeString& result, UErrorCode& status) const;

    /**
     * Writes RFC2445 VTIMEZONE data for this time zone applicalbe
     * for dates after the specified start time.
     * @param start The start date.
     * @param result Output param to filled in with the VTIMEZONE data.
     * @param status Output param to filled in with a success or an error.
     * @stable ICU 3.8
     */
    void write(UDate start, UnicodeString& result, UErrorCode& status) const;

    /**
     * Writes RFC2445 VTIMEZONE data applicalbe for the specified date.
     * Some common iCalendar implementations can only handle a single time
     * zone property or a pair of standard and daylight time properties using
     * BYDAY rule with day of week (such as BYDAY=1SUN).  This method produce
     * the VTIMEZONE data which can be handled these implementations.  The rules
     * produced by this method can be used only for calculating time zone offset
     * around the specified date.
     * @param time The date used for rule extraction.
     * @param result Output param to filled in with the VTIMEZONE data.
     * @param status Output param to filled in with a success or an error.
     * @stable ICU 3.8
     */
    void writeSimple(UDate time, UnicodeString& result, UErrorCode& status) const;

    /**
     * Clones TimeZone objects polymorphically. Clients are responsible for deleting
     * the TimeZone object cloned.
     * @return   A new copy of this TimeZone object.
     * @stable ICU 3.8
     */
    virtual VTimeZone* clone() const;

    /**
     * Returns the TimeZone's adjusted GMT offset (i.e., the number of milliseconds to add
     * to GMT to get local time in this time zone, taking daylight savings time into
     * account) as of a particular reference date.  The reference date is used to determine
     * whether daylight savings time is in effect and needs to be figured into the offset
     * that is returned (in other words, what is the adjusted GMT offset in this time zone
     * at this particular date and time?).  For the time zones produced by createTimeZone(),
     * the reference data is specified according to the Gregorian calendar, and the date
     * and time fields are local standard time.
     *
     * <p>Note: Don't call this method. Instead, call the getOffset(UDate...) overload,
     * which returns both the raw and the DST offset for a given time. This method
     * is retained only for backward compatibility.
     *
     * @param era        The reference date's era
     * @param year       The reference date's year
     * @param month      The reference date's month (0-based; 0 is January)
     * @param day        The reference date's day-in-month (1-based)
     * @param dayOfWeek  The reference date's day-of-week (1-based; 1 is Sunday)
     * @param millis     The reference date's milliseconds in day, local standard time
     * @param status     Output param to filled in with a success or an error.
     * @return           The offset in milliseconds to add to GMT to get local time.
     * @stable ICU 3.8
     */
    virtual int32_t getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                              uint8_t dayOfWeek, int32_t millis, UErrorCode& status) const;

    /**
     * Gets the time zone offset, for current date, modified in case of
     * daylight savings. This is the offset to add *to* UTC to get local time.
     *
     * <p>Note: Don't call this method. Instead, call the getOffset(UDate...) overload,
     * which returns both the raw and the DST offset for a given time. This method
     * is retained only for backward compatibility.
     *
     * @param era        The reference date's era
     * @param year       The reference date's year
     * @param month      The reference date's month (0-based; 0 is January)
     * @param day        The reference date's day-in-month (1-based)
     * @param dayOfWeek  The reference date's day-of-week (1-based; 1 is Sunday)
     * @param millis     The reference date's milliseconds in day, local standard time
     * @param monthLength The length of the given month in days.
     * @param status     Output param to filled in with a success or an error.
     * @return           The offset in milliseconds to add to GMT to get local time.
     * @stable ICU 3.8
     */
    virtual int32_t getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                           uint8_t dayOfWeek, int32_t millis,
                           int32_t monthLength, UErrorCode& status) const;

    /**
     * Returns the time zone raw and GMT offset for the given moment
     * in time.  Upon return, local-millis = GMT-millis + rawOffset +
     * dstOffset.  All computations are performed in the proleptic
     * Gregorian calendar.  The default implementation in the TimeZone
     * class delegates to the 8-argument getOffset().
     *
     * @param date moment in time for which to return offsets, in
     * units of milliseconds from January 1, 1970 0:00 GMT, either GMT
     * time or local wall time, depending on `local'.
     * @param local if true, `date' is local wall time; otherwise it
     * is in GMT time.
     * @param rawOffset output parameter to receive the raw offset, that
     * is, the offset not including DST adjustments
     * @param dstOffset output parameter to receive the DST offset,
     * that is, the offset to be added to `rawOffset' to obtain the
     * total offset between local and GMT time. If DST is not in
     * effect, this value is zero; otherwise it is a positive value,
     * typically one hour.
     * @param ec input-output error code
     * @stable ICU 3.8
     */
    virtual void getOffset(UDate date, UBool local, int32_t& rawOffset,
                           int32_t& dstOffset, UErrorCode& ec) const;

    /**
     * Sets the TimeZone's raw GMT offset (i.e., the number of milliseconds to add
     * to GMT to get local time, before taking daylight savings time into account).
     *
     * @param offsetMillis  The new raw GMT offset for this time zone.
     * @stable ICU 3.8
     */
    virtual void setRawOffset(int32_t offsetMillis);

    /**
     * Returns the TimeZone's raw GMT offset (i.e., the number of milliseconds to add
     * to GMT to get local time, before taking daylight savings time into account).
     *
     * @return   The TimeZone's raw GMT offset.
     * @stable ICU 3.8
     */
    virtual int32_t getRawOffset(void) const;

    /**
     * Queries if this time zone uses daylight savings time.
     * @return true if this time zone uses daylight savings time,
     * false, otherwise.
     * @stable ICU 3.8
     */
    virtual UBool useDaylightTime(void) const;

#ifndef U_FORCE_HIDE_DEPRECATED_API
    /**
     * Queries if the given date is in daylight savings time in
     * this time zone.
     * This method is wasteful since it creates a new GregorianCalendar and
     * deletes it each time it is called. This is a deprecated method
     * and provided only for Java compatibility.
     *
     * @param date the given UDate.
     * @param status Output param filled in with success/error code.
     * @return true if the given date is in daylight savings time,
     * false, otherwise.
     * @deprecated ICU 2.4. Use Calendar::inDaylightTime() instead.
     */
    virtual UBool inDaylightTime(UDate date, UErrorCode& status) const;
#endif  // U_FORCE_HIDE_DEPRECATED_API

    /**
     * Returns true if this zone has the same rule and offset as another zone.
     * That is, if this zone differs only in ID, if at all.
     * @param other the <code>TimeZone</code> object to be compared with
     * @return true if the given zone is the same as this one,
     * with the possible exception of the ID
     * @stable ICU 3.8
     */
    virtual UBool hasSameRules(const TimeZone& other) const;

    /**
     * Gets the first time zone transition after the base time.
     * @param base      The base time.
     * @param inclusive Whether the base time is inclusive or not.
     * @param result    Receives the first transition after the base time.
     * @return  TRUE if the transition is found.
     * @stable ICU 3.8
     */
    virtual UBool getNextTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const;

    /**
     * Gets the most recent time zone transition before the base time.
     * @param base      The base time.
     * @param inclusive Whether the base time is inclusive or not.
     * @param result    Receives the most recent transition before the base time.
     * @return  TRUE if the transition is found.
     * @stable ICU 3.8
     */
    virtual UBool getPreviousTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const;

    /**
     * Returns the number of <code>TimeZoneRule</code>s which represents time transitions,
     * for this time zone, that is, all <code>TimeZoneRule</code>s for this time zone except
     * <code>InitialTimeZoneRule</code>.  The return value range is 0 or any positive value.
     * @param status    Receives error status code.
     * @return The number of <code>TimeZoneRule</code>s representing time transitions.
     * @stable ICU 3.8
     */
    virtual int32_t countTransitionRules(UErrorCode& status) const;

    /**
     * Gets the <code>InitialTimeZoneRule</code> and the set of <code>TimeZoneRule</code>
     * which represent time transitions for this time zone.  On successful return,
     * the argument initial points to non-NULL <code>InitialTimeZoneRule</code> and
     * the array trsrules is filled with 0 or multiple <code>TimeZoneRule</code>
     * instances up to the size specified by trscount.  The results are referencing the
     * rule instance held by this time zone instance.  Therefore, after this time zone
     * is destructed, they are no longer available.
     * @param initial       Receives the initial timezone rule
     * @param trsrules      Receives the timezone transition rules
     * @param trscount      On input, specify the size of the array 'transitions' receiving
     *                      the timezone transition rules.  On output, actual number of
     *                      rules filled in the array will be set.
     * @param status        Receives error status code.
     * @stable ICU 3.8
     */
    virtual void getTimeZoneRules(const InitialTimeZoneRule*& initial,
        const TimeZoneRule* trsrules[], int32_t& trscount, UErrorCode& status) const;

private:
    enum { DEFAULT_VTIMEZONE_LINES = 100 };

    /**
     * Default constructor.
     */
    VTimeZone();
    static VTimeZone* createVTimeZone(VTZReader* reader);
    void write(VTZWriter& writer, UErrorCode& status) const;
    void write(UDate start, VTZWriter& writer, UErrorCode& status) const;
    void writeSimple(UDate time, VTZWriter& writer, UErrorCode& status) const;
    void load(VTZReader& reader, UErrorCode& status);
    void parse(UErrorCode& status);

    void writeZone(VTZWriter& w, BasicTimeZone& basictz, UVector* customProps,
        UErrorCode& status) const;

    void writeHeaders(VTZWriter& w, UErrorCode& status) const;
    void writeFooter(VTZWriter& writer, UErrorCode& status) const;

    void writeZonePropsByTime(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                              int32_t fromOffset, int32_t toOffset, UDate time, UBool withRDATE,
                              UErrorCode& status) const;
    void writeZonePropsByDOM(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                             int32_t fromOffset, int32_t toOffset,
                             int32_t month, int32_t dayOfMonth, UDate startTime, UDate untilTime,
                             UErrorCode& status) const;
    void writeZonePropsByDOW(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                             int32_t fromOffset, int32_t toOffset,
                             int32_t month, int32_t weekInMonth, int32_t dayOfWeek,
                             UDate startTime, UDate untilTime, UErrorCode& status) const;
    void writeZonePropsByDOW_GEQ_DOM(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                                     int32_t fromOffset, int32_t toOffset,
                                     int32_t month, int32_t dayOfMonth, int32_t dayOfWeek,
                                     UDate startTime, UDate untilTime, UErrorCode& status) const;
    void writeZonePropsByDOW_GEQ_DOM_sub(VTZWriter& writer, int32_t month, int32_t dayOfMonth,
                                         int32_t dayOfWeek, int32_t numDays,
                                         UDate untilTime, int32_t fromOffset, UErrorCode& status) const;
    void writeZonePropsByDOW_LEQ_DOM(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                                     int32_t fromOffset, int32_t toOffset,
                                     int32_t month, int32_t dayOfMonth, int32_t dayOfWeek,
                                     UDate startTime, UDate untilTime, UErrorCode& status) const;
    void writeFinalRule(VTZWriter& writer, UBool isDst, const AnnualTimeZoneRule* rule,
                        int32_t fromRawOffset, int32_t fromDSTSavings,
                        UDate startTime, UErrorCode& status) const;

    void beginZoneProps(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                        int32_t fromOffset, int32_t toOffset, UDate startTime, UErrorCode& status) const;
    void endZoneProps(VTZWriter& writer, UBool isDst, UErrorCode& status) const;
    void beginRRULE(VTZWriter& writer, int32_t month, UErrorCode& status) const;
    void appendUNTIL(VTZWriter& writer, const UnicodeString& until, UErrorCode& status) const;

    BasicTimeZone   *tz;
    UVector         *vtzlines;
    UnicodeString   tzurl;
    UDate           lastmod;
    UnicodeString   olsonzid;
    UnicodeString   icutzver;

public:
    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 3.8
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 3.8
     */
    virtual UClassID getDynamicClassID(void) const;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ********************************************************************************
 * Copyright (C) 2003-2013, International Business Machines Corporation
 * and others. All Rights Reserved.
 ********************************************************************************
 *
 * File BUDDHCAL.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   05/13/2003  srl          copied from gregocal.h
 *   06/29/2007  srl          copied from buddhcal.h
 ********************************************************************************
 */




#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

/**
 * Concrete class which provides the Taiwan calendar.
 * <P>
 * <code>TaiwanCalendar</code> is a subclass of <code>GregorianCalendar</code>
 * that numbers years since 1912
 * <p>
 * The Taiwan calendar is identical to the Gregorian calendar in all respects
 * except for the year and era.  Years are numbered since 1912 AD (Gregorian),
 * so that 1912 AD (Gregorian) is equivalent to 1 MINGUO (Minguo Era) and 1998 AD is 87 MINGUO.
 * <p>
 * The Taiwan Calendar has two eras: <code>BEFORE_MINGUO</code> and <code>MINGUO</code>.
 * <p>
 * @internal
 */
class TaiwanCalendar : public GregorianCalendar {
public:

    /**
     * Useful constants for TaiwanCalendar.  Only one Era.
     * @internal
     */
    enum EEras {
       BEFORE_MINGUO = 0,
       MINGUO  = 1
    };

    /**
     * Constructs a TaiwanCalendar based on the current time in the default time zone
     * with the given locale.
     *
     * @param aLocale  The given locale.
     * @param success  Indicates the status of TaiwanCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @internal
     */
    TaiwanCalendar(const Locale& aLocale, UErrorCode& success);


    /**
     * Destructor
     * @internal
     */
    virtual ~TaiwanCalendar();

    /**
     * Copy constructor
     * @param source    the object to be copied.
     * @internal
     */
    TaiwanCalendar(const TaiwanCalendar& source);

    /**
     * Default assignment operator
     * @param right    the object to be copied.
     * @internal
     */
    TaiwanCalendar& operator=(const TaiwanCalendar& right);

    /**
     * Create and return a polymorphic copy of this calendar.
     * @return    return a polymorphic copy of this calendar.
     * @internal
     */
    virtual TaiwanCalendar* clone() const;

public:
    /**
     * Override Calendar Returns a unique class ID POLYMORPHICALLY. Pure virtual
     * override. This method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone() methods call
     * this method.
     *
     * @return   The class ID for this object. All objects of a given class have the
     *           same class ID. Objects of other classes have different class IDs.
     * @internal
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Return the class ID for this class. This is useful only for comparing to a return
     * value from getDynamicClassID(). For example:
     *
     *      Base* polymorphic_pointer = createPolymorphicObject();
     *      if (polymorphic_pointer->getDynamicClassID() ==
     *          Derived::getStaticClassID()) ...
     *
     * @return   The class ID for all objects of this class.
     * @internal
     */
    U_I18N_API static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * return the calendar type, "Taiwan".
     *
     * @return calendar type
     * @internal
     */
    virtual const char * getType() const;

private:
    TaiwanCalendar(); // default constructor not implemented

 protected:
     /**
     * Return the extended year defined by the current fields.  This will
     * use the UCAL_EXTENDED_YEAR field or the UCAL_YEAR and supra-year fields (such
     * as UCAL_ERA) specific to the calendar system, depending on which set of
     * fields is newer.
     * @return the extended year
     * @internal
     */
    virtual int32_t handleGetExtendedYear();
    /**
     * Subclasses may override this method to compute several fields
     * specific to each calendar system.  
     * @internal
     */
    virtual void handleComputeFields(int32_t julianDay, UErrorCode& status);
    /**
     * Subclass API for defining limits of different types.
     * @param field one of the field numbers
     * @param limitType one of <code>MINIMUM</code>, <code>GREATEST_MINIMUM</code>,
     * <code>LEAST_MAXIMUM</code>, or <code>MAXIMUM</code>
     * @internal
     */
    virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;

    /**
     * Returns TRUE because the Taiwan Calendar does have a default century
     * @internal
     */
    virtual UBool haveDefaultCentury() const;

    /**
     * Returns the date of the start of the default century
     * @return start of century - in milliseconds since epoch, 1970
     * @internal
     */
    virtual UDate defaultCenturyStart() const;

    /**
     * Returns the year in which the default century begins
     * @internal
     */
    virtual int32_t defaultCenturyStartYear() const;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ********************************************************************************
 * Copyright (C) 2003-2008, International Business Machines Corporation
 * and others. All Rights Reserved.
 ********************************************************************************
 *
 * File JAPANCAL.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   05/13/2003  srl         copied from gregocal.h
 ********************************************************************************
 */




#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

/**
 * Concrete class which provides the Japanese calendar.
 * <P>
 * <code>JapaneseCalendar</code> is a subclass of <code>GregorianCalendar</code>
 * that numbers years and eras based on the reigns of the Japanese emperors.
 * The Japanese calendar is identical to the Gregorian calendar in all respects
 * except for the year and era.  The ascension of each  emperor to the throne
 * begins a new era, and the years of that era are numbered starting with the
 * year of ascension as year 1.
 * <p>
 * Note that in the year of an imperial ascension, there are two possible sets
 * of year and era values: that for the old era and for the new.  For example, a
 * new era began on January 7, 1989 AD.  Strictly speaking, the first six days
 * of that year were in the Showa era, e.g. "January 6, 64 Showa", while the rest
 * of the year was in the Heisei era, e.g. "January 7, 1 Heisei".  This class
 * handles this distinction correctly when computing dates.  However, in lenient
 * mode either form of date is acceptable as input. 
 * <p>
 * In modern times, eras have started on January 8, 1868 AD, Gregorian (Meiji),
 * July 30, 1912 (Taisho), December 25, 1926 (Showa), and January 7, 1989 (Heisei).  Constants
 * for these eras, suitable for use in the <code>UCAL_ERA</code> field, are provided
 * in this class.  Note that the <em>number</em> used for each era is more or
 * less arbitrary.  Currently, the era starting in 645 AD is era #0; however this
 * may change in the future.  Use the predefined constants rather than using actual,
 * absolute numbers.
 * <p>
 * Since ICU4C 63, start date of each era is imported from CLDR. CLDR era data
 * may contain tentative era in near future with placeholder names. By default,
 * such era data is not enabled. ICU4C users who want to test the behavior of
 * the future era can enable this one of following settings (in the priority
 * order):
 * <ol>
 * <li>Environment variable <code>ICU_ENABLE_TENTATIVE_ERA=true</code>.</li>
 * </nl>
 * @internal
 */
class JapaneseCalendar : public GregorianCalendar {
public:

    /**
     * Check environment variable. 
     * @internal
     */
    U_I18N_API static UBool U_EXPORT2 enableTentativeEra(void);

    /**
     * Useful constants for JapaneseCalendar.
     * Exported for use by test code.
     * @internal
     */
    U_I18N_API static uint32_t U_EXPORT2 getCurrentEra(void); // the current era

    /**
     * Constructs a JapaneseCalendar based on the current time in the default time zone
     * with the given locale.
     *
     * @param aLocale  The given locale.
     * @param success  Indicates the status of JapaneseCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @stable ICU 2.0
     */
    JapaneseCalendar(const Locale& aLocale, UErrorCode& success);


    /**
     * Destructor
     * @internal
     */
    virtual ~JapaneseCalendar();

    /**
     * Copy constructor
     * @param source    the object to be copied.
     * @internal
     */
    JapaneseCalendar(const JapaneseCalendar& source);

    /**
     * Default assignment operator
     * @param right    the object to be copied.
     * @internal
     */
    JapaneseCalendar& operator=(const JapaneseCalendar& right);

    /**
     * Create and return a polymorphic copy of this calendar.
     * @return    return a polymorphic copy of this calendar.
     * @internal
     */
    virtual JapaneseCalendar* clone() const;

    /**
     * Return the extended year defined by the current fields.  In the 
     * Japanese calendar case, this is equal to the equivalent extended Gregorian year.
     * @internal
     */
    virtual int32_t handleGetExtendedYear();

    /**
     * Return the maximum value that this field could have, given the current date.
     * @internal
     */
    virtual int32_t getActualMaximum(UCalendarDateFields field, UErrorCode& status) const;


public:
    /**
     * Override Calendar Returns a unique class ID POLYMORPHICALLY. Pure virtual
     * override. This method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone() methods call
     * this method.
     *
     * @return   The class ID for this object. All objects of a given class have the
     *           same class ID. Objects of other classes have different class IDs.
     * @internal
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Return the class ID for this class. This is useful only for comparing to a return
     * value from getDynamicClassID(). For example:
     *
     *      Base* polymorphic_pointer = createPolymorphicObject();
     *      if (polymorphic_pointer->getDynamicClassID() ==
     *          Derived::getStaticClassID()) ...
     *
     * @return   The class ID for all objects of this class.
     * @internal
     */
    U_I18N_API static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * return the calendar type, "japanese".
     *
     * @return calendar type
     * @internal
     */
    virtual const char * getType() const;

    /**
     * @return FALSE - no default century in Japanese
     * @internal 
     */
    virtual UBool haveDefaultCentury() const;

    /**
     * Not used - no default century.
     * @internal
     */
    virtual UDate defaultCenturyStart() const;
    /**
     * Not used - no default century.
     * @internal
     */
    virtual int32_t defaultCenturyStartYear() const;

private:
    JapaneseCalendar(); // default constructor not implemented

protected:
    /** 
     * Calculate the era for internal computation
     * @internal
     */
    virtual int32_t internalGetEra() const;

    /**
     * Compute fields from the JD
     * @internal
     */
    virtual void handleComputeFields(int32_t julianDay, UErrorCode& status);

    /**
     * Calculate the limit for a specified type of limit and field
     * @internal
     */
    virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;

    /***
     * Called by computeJulianDay.  Returns the default month (0-based) for the year,
     * taking year and era into account.  Will return the first month of the given era, if 
     * the current year is an ascension year.
     * @param eyear the extended year
     * @internal
     */
    virtual int32_t getDefaultMonthInYear(int32_t eyear);

    /***
     * Called by computeJulianDay.  Returns the default day (1-based) for the month,
     * taking currently-set year and era into account.  Will return the first day of the given
     * era, if the current month is an ascension year and month.
     * @param eyear the extended year
     * @param mon the month in the year
     * @internal
     */
    virtual int32_t getDefaultDayInMonth(int32_t eyear, int32_t month);
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ********************************************************************************
 * Copyright (C) 2003-2013, International Business Machines Corporation
 * and others. All Rights Reserved.
 ******************************************************************************
 *
 * File ISLAMCAL.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   10/14/2003  srl         ported from java IslamicCalendar
 *****************************************************************************
 */




#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

/**
 * <code>IslamicCalendar</code> is a subclass of <code>Calendar</code>
 * that implements the Islamic civil and religious calendars.  It
 * is used as the civil calendar in most of the Arab world and the
 * liturgical calendar of the Islamic faith worldwide.  This calendar
 * is also known as the "Hijri" calendar, since it starts at the time
 * of Mohammed's emigration (or "hijra") to Medinah on Thursday, 
 * July 15, 622 AD (Julian).
 * <p>
 * The Islamic calendar is strictly lunar, and thus an Islamic year of twelve
 * lunar months does not correspond to the solar year used by most other
 * calendar systems, including the Gregorian.  An Islamic year is, on average,
 * about 354 days long, so each successive Islamic year starts about 11 days
 * earlier in the corresponding Gregorian year.
 * <p>
 * Each month of the calendar starts when the new moon's crescent is visible
 * at sunset.  However, in order to keep the time fields in this class
 * synchronized with those of the other calendars and with local clock time,
 * we treat days and months as beginning at midnight,
 * roughly 6 hours after the corresponding sunset.
 * <p>
 * There are two main variants of the Islamic calendar in existence.  The first
 * is the <em>civil</em> calendar, which uses a fixed cycle of alternating 29-
 * and 30-day months, with a leap day added to the last month of 11 out of
 * every 30 years.  This calendar is easily calculated and thus predictable in
 * advance, so it is used as the civil calendar in a number of Arab countries.
 * This is the default behavior of a newly-created <code>IslamicCalendar</code>
 * object.
 * <p>
 * The Islamic <em>religious</em> calendar, however, is based on the <em>observation</em>
 * of the crescent moon.  It is thus affected by the position at which the
 * observations are made, seasonal variations in the time of sunset, the
 * eccentricities of the moon's orbit, and even the weather at the observation
 * site.  This makes it impossible to calculate in advance, and it causes the
 * start of a month in the religious calendar to differ from the civil calendar
 * by up to three days.
 * <p>
 * Using astronomical calculations for the position of the sun and moon, the
 * moon's illumination, and other factors, it is possible to determine the start
 * of a lunar month with a fairly high degree of certainty.  However, these
 * calculations are extremely complicated and thus slow, so most algorithms,
 * including the one used here, are only approximations of the true astronical
 * calculations.  At present, the approximations used in this class are fairly
 * simplistic; they will be improved in later versions of the code.
 * <p>
 * The {@link #setCivil setCivil} method determines
 * which approach is used to determine the start of a month.  By default, the
 * fixed-cycle civil calendar is used.  However, if <code>setCivil(false)</code>
 * is called, an approximation of the true lunar calendar will be used.
 *
 * @see GregorianCalendar
 *
 * @author Laura Werner
 * @author Alan Liu
 * @author Steven R. Loomis
 * @internal
 */
class U_I18N_API IslamicCalendar : public Calendar {
 public:
  //-------------------------------------------------------------------------
  // Constants...
  //-------------------------------------------------------------------------
  
  /**
   * Calendar type - civil or religious or um alqura
   * @internal 
   */
  enum ECalculationType {
    ASTRONOMICAL,
    CIVIL,
    UMALQURA,
    TBLA
  };
  
  /**
   * Constants for the months
   * @internal
   */
  enum EMonths {
    /**
     * Constant for Muharram, the 1st month of the Islamic year. 
     * @internal
     */
    MUHARRAM = 0,

    /**
     * Constant for Safar, the 2nd month of the Islamic year. 
     * @internal
     */
    SAFAR = 1,

    /**
     * Constant for Rabi' al-awwal (or Rabi' I), the 3rd month of the Islamic year. 
     * @internal 
     */
    RABI_1 = 2,

    /**
     * Constant for Rabi' al-thani or (Rabi' II), the 4th month of the Islamic year. 
     * @internal 
     */
    RABI_2 = 3,

    /**
     * Constant for Jumada al-awwal or (Jumada I), the 5th month of the Islamic year. 
     * @internal 
     */
    JUMADA_1 = 4,

    /**
     * Constant for Jumada al-thani or (Jumada II), the 6th month of the Islamic year. 
     * @internal 
     */
    JUMADA_2 = 5,

    /**
     * Constant for Rajab, the 7th month of the Islamic year. 
     * @internal 
     */
    RAJAB = 6,

    /**
     * Constant for Sha'ban, the 8th month of the Islamic year. 
     * @internal 
     */
    SHABAN = 7,

    /**
     * Constant for Ramadan, the 9th month of the Islamic year. 
     * @internal 
     */
    RAMADAN = 8,

    /**
     * Constant for Shawwal, the 10th month of the Islamic year. 
     * @internal 
     */
    SHAWWAL = 9,

    /**
     * Constant for Dhu al-Qi'dah, the 11th month of the Islamic year. 
     * @internal 
     */
    DHU_AL_QIDAH = 10,

    /**
     * Constant for Dhu al-Hijjah, the 12th month of the Islamic year. 
     * @internal 
     */
    DHU_AL_HIJJAH = 11,
    
    ISLAMIC_MONTH_MAX
  }; 


  //-------------------------------------------------------------------------
  // Constructors...
  //-------------------------------------------------------------------------

  /**
   * Constructs an IslamicCalendar based on the current time in the default time zone
   * with the given locale.
   *
   * @param aLocale  The given locale.
   * @param success  Indicates the status of IslamicCalendar object construction.
   *                 Returns U_ZERO_ERROR if constructed successfully.
   * @param type     The Islamic calendar calculation type. The default value is CIVIL.
   * @internal
   */
  IslamicCalendar(const Locale& aLocale, UErrorCode &success, ECalculationType type = CIVIL);

  /**
   * Copy Constructor
   * @internal
   */
  IslamicCalendar(const IslamicCalendar& other);

  /**
   * Destructor.
   * @internal
   */
  virtual ~IslamicCalendar();

  /**
   * Sets Islamic calendar calculation type used by this instance.
   *
   * @param type    The calendar calculation type, <code>CIVIL</code> to use the civil
   *                calendar, <code>ASTRONOMICAL</code> to use the astronomical calendar.
   * @internal
   */
  void setCalculationType(ECalculationType type, UErrorCode &status);
    
  /**
   * Returns <code>true</code> if this object is using the fixed-cycle civil
   * calendar, or <code>false</code> if using the religious, astronomical
   * calendar.
   * @internal
   */
  UBool isCivil();


  // TODO: copy c'tor, etc

  // clone
  virtual IslamicCalendar* clone() const;

 private:
  /**
   * Determine whether a year is a leap year in the Islamic civil calendar
   */
  static UBool civilLeapYear(int32_t year);
    
  /**
   * Return the day # on which the given year starts.  Days are counted
   * from the Hijri epoch, origin 0.
   */
  int32_t yearStart(int32_t year) const;

  /**
   * Return the day # on which the given month starts.  Days are counted
   * from the Hijri epoch, origin 0.
   *
   * @param year  The hijri year
   * @param year  The hijri month, 0-based
   */
  int32_t monthStart(int32_t year, int32_t month) const;
    
  /**
   * Find the day number on which a particular month of the true/lunar
   * Islamic calendar starts.
   *
   * @param month The month in question, origin 0 from the Hijri epoch
   *
   * @return The day number on which the given month starts.
   */
  int32_t trueMonthStart(int32_t month) const;

  /**
   * Return the "age" of the moon at the given time; this is the difference
   * in ecliptic latitude between the moon and the sun.  This method simply
   * calls CalendarAstronomer.moonAge, converts to degrees, 
   * and adjusts the resultto be in the range [-180, 180].
   *
   * @param time  The time at which the moon's age is desired,
   *              in millis since 1/1/1970.
   */
  static double moonAge(UDate time, UErrorCode &status);

  //-------------------------------------------------------------------------
  // Internal data....
  //
    
  /**
   * <code>CIVIL</code> if this object uses the fixed-cycle Islamic civil calendar,
   * and <code>ASTRONOMICAL</code> if it approximates the true religious calendar using
   * astronomical calculations for the time of the new moon.
   */
  ECalculationType cType;

  //----------------------------------------------------------------------
  // Calendar framework
  //----------------------------------------------------------------------
 protected:
  /**
   * @internal
   */
  virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;
  
  /**
   * Return the length (in days) of the given month.
   *
   * @param year  The hijri year
   * @param year  The hijri month, 0-based
   * @internal
   */
  virtual int32_t handleGetMonthLength(int32_t extendedYear, int32_t month) const;
  
  /**
   * Return the number of days in the given Islamic year
   * @internal
   */
  virtual int32_t handleGetYearLength(int32_t extendedYear) const;
    
  //-------------------------------------------------------------------------
  // Functions for converting from field values to milliseconds....
  //-------------------------------------------------------------------------

  // Return JD of start of given month/year
  /**
   * @internal
   */
  virtual int32_t handleComputeMonthStart(int32_t eyear, int32_t month, UBool useMonth) const;

  //-------------------------------------------------------------------------
  // Functions for converting from milliseconds to field values
  //-------------------------------------------------------------------------

  /**
   * @internal
   */
  virtual int32_t handleGetExtendedYear();

  /**
   * Override Calendar to compute several fields specific to the Islamic
   * calendar system.  These are:
   *
   * <ul><li>ERA
   * <li>YEAR
   * <li>MONTH
   * <li>DAY_OF_MONTH
   * <li>DAY_OF_YEAR
   * <li>EXTENDED_YEAR</ul>
   * 
   * The DAY_OF_WEEK and DOW_LOCAL fields are already set when this
   * method is called. The getGregorianXxx() methods return Gregorian
   * calendar equivalents for the given Julian day.
   * @internal
   */
  virtual void handleComputeFields(int32_t julianDay, UErrorCode &status);

  // UObject stuff
 public: 
  /**
   * @return   The class ID for this object. All objects of a given class have the
   *           same class ID. Objects of other classes have different class IDs.
   * @internal
   */
  virtual UClassID getDynamicClassID(void) const;

  /**
   * Return the class ID for this class. This is useful only for comparing to a return
   * value from getDynamicClassID(). For example:
   *
   *      Base* polymorphic_pointer = createPolymorphicObject();
   *      if (polymorphic_pointer->getDynamicClassID() ==
   *          Derived::getStaticClassID()) ...
   *
   * @return   The class ID for all objects of this class.
   * @internal
   */
  /*U_I18N_API*/ static UClassID U_EXPORT2 getStaticClassID(void);

  /**
   * return the calendar type, "buddhist".
   *
   * @return calendar type
   * @internal
   */
  virtual const char * getType() const;

 private:
  IslamicCalendar(); // default constructor not implemented

  // Default century.
 protected:

  /**
   * (Overrides Calendar) Return true if the current date for this Calendar is in
   * Daylight Savings Time. Recognizes DST_OFFSET, if it is set.
   *
   * @param status Fill-in parameter which receives the status of this operation.
   * @return   True if the current date for this Calendar is in Daylight Savings Time,
   *           false, otherwise.
   * @internal
   */
  virtual UBool inDaylightTime(UErrorCode& status) const;


  /**
   * Returns TRUE because the Islamic Calendar does have a default century
   * @internal
   */
  virtual UBool haveDefaultCentury() const;

  /**
   * Returns the date of the start of the default century
   * @return start of century - in milliseconds since epoch, 1970
   * @internal
   */
  virtual UDate defaultCenturyStart() const;

  /**
   * Returns the year in which the default century begins
   * @internal
   */
  virtual int32_t defaultCenturyStartYear() const;

 private:
  /**
   * Initializes the 100-year window that dates with 2-digit years
   * are considered to fall within so that its start date is 80 years
   * before the current time.
   */
  static void U_CALLCONV initializeSystemDefaultCentury(void);
};

U_NAMESPACE_END

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2003-2013, International Business Machines Corporation
* and others. All Rights Reserved.
******************************************************************************
*
* File HEBRWCAL.H
*
* Modification History:
*
*   Date        Name        Description
*   05/13/2003  srl          copied from gregocal.h
*   11/26/2003  srl          copied from buddhcal.h
******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

/**
 * <code>HebrewCalendar</code> is a subclass of <code>Calendar</code>
 * that that implements the traditional Hebrew calendar.
 * This is the civil calendar in Israel and the liturgical calendar
 * of the Jewish faith worldwide.
 * <p>
 * The Hebrew calendar is lunisolar and thus has a number of interesting
 * properties that distinguish it from the Gregorian.  Months start
 * on the day of (an arithmetic approximation of) each new moon.  Since the
 * solar year (approximately 365.24 days) is not an even multiple of
 * the lunar month (approximately 29.53 days) an extra "leap month" is
 * inserted in 7 out of every 19 years.  To make matters even more
 * interesting, the start of a year can be delayed by up to three days
 * in order to prevent certain holidays from falling on the Sabbath and
 * to prevent certain illegal year lengths.  Finally, the lengths of certain
 * months can vary depending on the number of days in the year.
 * <p>
 * The leap month is known as "Adar 1" and is inserted between the
 * months of Shevat and Adar in leap years.  Since the leap month does
 * not come at the end of the year, calculations involving
 * month numbers are particularly complex.  Users of this class should
 * make sure to use the {@link #roll roll} and {@link #add add} methods
 * rather than attempting to perform date arithmetic by manipulating
 * the fields directly.
 * <p>
 * <b>Note:</b> In the traditional Hebrew calendar, days start at sunset.
 * However, in order to keep the time fields in this class
 * synchronized with those of the other calendars and with local clock time,
 * we treat days and months as beginning at midnight,
 * roughly 6 hours after the corresponding sunset.
 * <p>
 * If you are interested in more information on the rules behind the Hebrew
 * calendar, see one of the following references:
 * <ul>
 * <li>"<a href="http://www.amazon.com/exec/obidos/ASIN/0521564743">Calendrical Calculations</a>",
 *      by Nachum Dershowitz & Edward Reingold, Cambridge University Press, 1997, pages 85-91.
 *
 * <li>Hebrew Calendar Science and Myths,
 *      <a href="http://www.geocities.com/Athens/1584/">
 *      http://www.geocities.com/Athens/1584/</a>
 *
 * <li>The Calendar FAQ,
 *      <a href="http://www.faqs.org/faqs/calendars/faq/">
 *      http://www.faqs.org/faqs/calendars/faq/</a>
 * </ul>
 * <p>
 * @see com.ibm.icu.util.GregorianCalendar
 *
 * @author Laura Werner
 * @author Alan Liu
 * @author Steven R. Loomis
 * <p>
 * @internal
 */
class U_I18N_API HebrewCalendar : public Calendar {
public:
  /**
   * Useful constants for HebrewCalendar.
   * @internal
   */
  enum EEras {
    /** 
     * Constant for Tishri, the 1st month of the Hebrew year. 
     */
      TISHRI,
      /**
     * Constant for Heshvan, the 2nd month of the Hebrew year. 
     */
      HESHVAN,
      /**
     * Constant for Kislev, the 3rd month of the Hebrew year. 
     */
      KISLEV,

    /**
     * Constant for Tevet, the 4th month of the Hebrew year. 
     */
      TEVET,

    /**
     * Constant for Shevat, the 5th month of the Hebrew year. 
     */
      SHEVAT,

    /**
     * Constant for Adar I, the 6th month of the Hebrew year
     * (present in leap years only). In non-leap years, the calendar
     * jumps from Shevat (5th month) to Adar (7th month).
     */
      ADAR_1,

    /** 
     * Constant for the Adar, the 7th month of the Hebrew year. 
     */
      ADAR,

    /**
     * Constant for Nisan, the 8th month of the Hebrew year. 
     */
      NISAN,

    /**
     * Constant for Iyar, the 9th month of the Hebrew year. 
     */
      IYAR,

    /**
     * Constant for Sivan, the 10th month of the Hebrew year. 
     */
      SIVAN,

    /**
     * Constant for Tammuz, the 11th month of the Hebrew year. 
     */
      TAMUZ,

    /**
     * Constant for Av, the 12th month of the Hebrew year. 
     */
      AV,

    /**
     * Constant for Elul, the 13th month of the Hebrew year. 
     */
      ELUL
    };

    /**
     * Constructs a HebrewCalendar based on the current time in the default time zone
     * with the given locale.
     *
     * @param aLocale  The given locale.
     * @param success  Indicates the status of HebrewCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @internal
     */
    HebrewCalendar(const Locale& aLocale, UErrorCode& success);


    /**
     * Destructor
     * @internal
     */
    virtual ~HebrewCalendar();

    /**
     * Copy constructor
     * @param source    the object to be copied.
     * @internal
     */
    HebrewCalendar(const HebrewCalendar& source);

    /**
     * Default assignment operator
     * @param right    the object to be copied.
     * @internal
     */
    HebrewCalendar& operator=(const HebrewCalendar& right);

    /**
     * Create and return a polymorphic copy of this calendar.
     * @return    return a polymorphic copy of this calendar.
     * @internal
     */
    virtual HebrewCalendar* clone() const;
    
public:
    /**
     * Override Calendar Returns a unique class ID POLYMORPHICALLY. Pure virtual
     * override. This method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone() methods call
     * this method.
     *
     * @return   The class ID for this object. All objects of a given class have the
     *           same class ID. Objects of other classes have different class IDs.
     * @internal
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Return the class ID for this class. This is useful only for comparing to a return
     * value from getDynamicClassID(). For example:
     *
     *      Base* polymorphic_pointer = createPolymorphicObject();
     *      if (polymorphic_pointer->getDynamicClassID() ==
     *          Derived::getStaticClassID()) ...
     *
     * @return   The class ID for all objects of this class.
     * @internal
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * return the calendar type, "hebrew".
     *
     * @return calendar type
     * @internal
     */
    virtual const char * getType() const;


    // Calendar API
 public:
    /**
     * (Overrides Calendar) UDate Arithmetic function. Adds the specified (signed) amount
     * of time to the given time field, based on the calendar's rules.  For more
     * information, see the documentation for Calendar::add().
     *
     * @param field   The time field.
     * @param amount  The amount of date or time to be added to the field.
     * @param status  Output param set to success/failure code on exit. If any value
     *                previously set in the time field is invalid, this will be set to
     *                an error status.
     */
    virtual void add(UCalendarDateFields field, int32_t amount, UErrorCode& status);
    /**
     * @deprecated ICU 2.6 use UCalendarDateFields instead of EDateFields
     */
    virtual void add(EDateFields field, int32_t amount, UErrorCode& status);


    /**
     * (Overrides Calendar) Rolls up or down by the given amount in the specified field.
     * For more information, see the documentation for Calendar::roll().
     *
     * @param field   The time field.
     * @param amount  Indicates amount to roll.
     * @param status  Output param set to success/failure code on exit. If any value
     *                previously set in the time field is invalid, this will be set to
     *                an error status.
     * @internal
     */
    virtual void roll(UCalendarDateFields field, int32_t amount, UErrorCode& status);

    /**
     * (Overrides Calendar) Rolls up or down by the given amount in the specified field.
     * For more information, see the documentation for Calendar::roll().
     *
     * @param field   The time field.
     * @param amount  Indicates amount to roll.
     * @param status  Output param set to success/failure code on exit. If any value
     *                previously set in the time field is invalid, this will be set to
     *                an error status.
     * @deprecated ICU 2.6. Use roll(UCalendarDateFields field, int32_t amount, UErrorCode& status) instead.
`     */
    virtual void roll(EDateFields field, int32_t amount, UErrorCode& status);

    /**
     * @internal 
     */
    static UBool isLeapYear(int32_t year) ;

 protected:

    /**
     * Subclass API for defining limits of different types.
     * Subclasses must implement this method to return limits for the
     * following fields:
     *
     * <pre>UCAL_ERA
     * UCAL_YEAR
     * UCAL_MONTH
     * UCAL_WEEK_OF_YEAR
     * UCAL_WEEK_OF_MONTH
     * UCAL_DATE (DAY_OF_MONTH on Java)
     * UCAL_DAY_OF_YEAR
     * UCAL_DAY_OF_WEEK_IN_MONTH
     * UCAL_YEAR_WOY
     * UCAL_EXTENDED_YEAR</pre>
     *
     * @param field one of the above field numbers
     * @param limitType one of <code>MINIMUM</code>, <code>GREATEST_MINIMUM</code>,
     * <code>LEAST_MAXIMUM</code>, or <code>MAXIMUM</code>
     * @internal
     */
    virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;

    /**
     * Return the number of days in the given month of the given extended
     * year of this calendar system.  Subclasses should override this
     * method if they can provide a more correct or more efficient
     * implementation than the default implementation in Calendar.
     * @internal
     */
    virtual int32_t handleGetMonthLength(int32_t extendedYear, int32_t month) const;

    /**
     * Return the number of days in the given extended year of this
     * calendar system.  Subclasses should override this method if they can
     * provide a more correct or more efficient implementation than the
     * default implementation in Calendar.
     * @stable ICU 2.0
     */
    virtual int32_t handleGetYearLength(int32_t eyear) const;
    /**
     * Subclasses may override this method to compute several fields
     * specific to each calendar system.  These are:
     *
     * <ul><li>ERA
     * <li>YEAR
     * <li>MONTH
     * <li>DAY_OF_MONTH
     * <li>DAY_OF_YEAR
     * <li>EXTENDED_YEAR</ul>
     *
     * <p>The GregorianCalendar implementation implements
     * a calendar with the specified Julian/Gregorian cutover date.
     * @internal
     */
    virtual void handleComputeFields(int32_t julianDay, UErrorCode &status);
    /**
     * Return the extended year defined by the current fields.  This will
     * use the UCAL_EXTENDED_YEAR field or the UCAL_YEAR and supra-year fields (such
     * as UCAL_ERA) specific to the calendar system, depending on which set of
     * fields is newer.
     * @return the extended year
     * @internal
     */
    virtual int32_t handleGetExtendedYear();
    /**
     * Return the Julian day number of day before the first day of the
     * given month in the given extended year.  Subclasses should override
     * this method to implement their calendar system.
     * @param eyear the extended year
     * @param month the zero-based month, or 0 if useMonth is false
     * @param useMonth if false, compute the day before the first day of
     * the given year, otherwise, compute the day before the first day of
     * the given month
     * @param return the Julian day number of the day before the first
     * day of the given month and year
     * @internal
     */
    virtual int32_t handleComputeMonthStart(int32_t eyear, int32_t month,
                                                   UBool useMonth) const;


    /**
     * Validate a single field of this calendar.
     * Overrides Calendar::validateField(int) to provide
     * special handling for month validation for Hebrew calendar.
     * @internal
     */
    virtual void validateField(UCalendarDateFields field, UErrorCode &status);

 protected:

  /**
   * (Overrides Calendar) Return true if the current date for this Calendar is in
   * Daylight Savings Time. Recognizes DST_OFFSET, if it is set.
   *
   * @param status Fill-in parameter which receives the status of this operation.
   * @return   True if the current date for this Calendar is in Daylight Savings Time,
   *           false, otherwise.
   * @internal
   */
  virtual UBool inDaylightTime(UErrorCode& status) const;

    /**
     * Returns TRUE because the Hebrew Calendar does have a default century
     * @internal
     */
    virtual UBool haveDefaultCentury() const;

    /**
     * Returns the date of the start of the default century
     * @return start of century - in milliseconds since epoch, 1970
     * @internal
     */
    virtual UDate defaultCenturyStart() const;

    /**
     * Returns the year in which the default century begins
     * @internal
     */
    virtual int32_t defaultCenturyStartYear() const;

 private: // Calendar-specific implementation
    /**
     * Finds the day # of the first day in the given Hebrew year.
     * To do this, we want to calculate the time of the Tishri 1 new moon
     * in that year.
     * <p>
     * The algorithm here is similar to ones described in a number of
     * references, including:
     * <ul>
     * <li>"Calendrical Calculations", by Nachum Dershowitz & Edward Reingold,
     *     Cambridge University Press, 1997, pages 85-91.
     *
     * <li>Hebrew Calendar Science and Myths,
     *     <a href="http://www.geocities.com/Athens/1584/">
     *     http://www.geocities.com/Athens/1584/</a>
     *
     * <li>The Calendar FAQ,
     *      <a href="http://www.faqs.org/faqs/calendars/faq/">
     *      http://www.faqs.org/faqs/calendars/faq/</a>
     * </ul>
     * @param year extended year
     * @return day number (JD)
     * @internal
     */
    static int32_t startOfYear(int32_t year, UErrorCode& status);

    static int32_t absoluteDayToDayOfWeek(int32_t day) ;
    
    /**
     * @internal 
     */
    int32_t yearType(int32_t year) const;

    /**
     * @internal 
     */
    static int32_t monthsInYear(int32_t year) ;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ******************************************************************************
 * Copyright (C) 2003-2013, International Business Machines Corporation
 * and others. All Rights Reserved.
 ******************************************************************************
 *
 * File PERSNCAL.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   9/23/2003 mehran        posted to icu-design
 *****************************************************************************
 */




#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

/**
 * <code>PersianCalendar</code> is a subclass of <code>Calendar</code>
 * that implements the Persian calendar.  It is used as the official
 * calendar in Iran.  This calendar is also known as the "Hijri Shamsi"
 * calendar, since it starts at the time of Mohammed's emigration (or
 * "hijra") to Medinah on Thursday, July 15, 622 AD (Julian) and is a
 * solar calendar system (or "shamsi").
 * <p>
 * The Persian calendar is strictly solar, and thus a Persian year has twelve
 * solar months. A Persian year is about 365 days long, except in leap years
 * which is 366 days long.
 * <p>
 * The six first months of Persian Calendar are 31 days long. The next five
 * months are 30 days long. The last month is 29 days long in normal years,
 * and 30 days long in leap years.
 *
 * @see GregorianCalendar
 *
 * @author Mehran Mehr
 * @internal
 */
class PersianCalendar : public Calendar {
 public:
  //-------------------------------------------------------------------------
  // Constants...
  //-------------------------------------------------------------------------
  /**
   * Constants for the months
   * @internal
   */
  enum EMonths {
    /**
     * Constant for Farvardin, the 1st month of the Persian year. 
     * @internal
     */
    FARVARDIN = 0,

    /**
     * Constant for Ordibehesht, the 2nd month of the Persian year. 
     * @internal
     */
    ORDIBEHESHT = 1,

    /**
     * Constant for Khordad, the 3rd month of the Persian year. 
     * @internal 
     */
    KHORDAD = 2,

    /**
     * Constant for Tir, the 4th month of the Persian year. 
     * @internal 
     */
    TIR = 3,

    /**
     * Constant for Mordad, the 5th month of the Persian year. 
     * @internal 
     */
    MORDAD = 4,

    /**
     * Constant for Shahrivar, the 6th month of the Persian year. 
     * @internal 
     */
    SHAHRIVAR = 5,

    /**
     * Constant for Mehr, the 7th month of the Persian year. 
     * @internal 
     */
    MEHR = 6,

    /**
     * Constant for Aban, the 8th month of the Persian year. 
     * @internal 
     */
    ABAN = 7,

    /**
     * Constant for Azar, the 9th month of the Persian year. 
     * @internal 
     */
    AZAR = 8,

    /**
     * Constant for Dei, the 10th month of the Persian year. 
     * @internal 
     */
    DEI = 9,

    /**
     * Constant for Bahman, the 11th month of the Persian year. 
     * @internal 
     */
    BAHMAN = 10,

    /**
     * Constant for Esfand, the 12th month of the Persian year. 
     * @internal 
     */
    ESFAND = 11,
    
    PERSIAN_MONTH_MAX
  }; 



  //-------------------------------------------------------------------------
  // Constructors...
  //-------------------------------------------------------------------------

  /**
   * Constructs a PersianCalendar based on the current time in the default time zone
   * with the given locale.
   *
   * @param aLocale  The given locale.
   * @param success  Indicates the status of PersianCalendar object construction.
   *                 Returns U_ZERO_ERROR if constructed successfully.
   * @internal
   */
  PersianCalendar(const Locale& aLocale, UErrorCode &success);

  /**
   * Copy Constructor
   * @internal
   */
  PersianCalendar(const PersianCalendar& other);

  /**
   * Destructor.
   * @internal
   */
  virtual ~PersianCalendar();

  // TODO: copy c'tor, etc

  // clone
  virtual PersianCalendar* clone() const;

 private:
  /**
   * Determine whether a year is a leap year in the Persian calendar
   */
  static UBool isLeapYear(int32_t year);
    
  /**
   * Return the day # on which the given year starts.  Days are counted
   * from the Hijri epoch, origin 0.
   */
  int32_t yearStart(int32_t year);

  /**
   * Return the day # on which the given month starts.  Days are counted
   * from the Hijri epoch, origin 0.
   *
   * @param year  The hijri shamsi year
   * @param year  The hijri shamsi month, 0-based
   */
  int32_t monthStart(int32_t year, int32_t month) const;
    
  //----------------------------------------------------------------------
  // Calendar framework
  //----------------------------------------------------------------------
 protected:
  /**
   * @internal
   */
  virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;
  
  /**
   * Return the length (in days) of the given month.
   *
   * @param year  The hijri shamsi year
   * @param year  The hijri shamsi month, 0-based
   * @internal
   */
  virtual int32_t handleGetMonthLength(int32_t extendedYear, int32_t month) const;
  
  /**
   * Return the number of days in the given Persian year
   * @internal
   */
  virtual int32_t handleGetYearLength(int32_t extendedYear) const;
    
  //-------------------------------------------------------------------------
  // Functions for converting from field values to milliseconds....
  //-------------------------------------------------------------------------

  // Return JD of start of given month/year
  /**
   * @internal
   */
  virtual int32_t handleComputeMonthStart(int32_t eyear, int32_t month, UBool useMonth) const;

  //-------------------------------------------------------------------------
  // Functions for converting from milliseconds to field values
  //-------------------------------------------------------------------------

  /**
   * @internal
   */
  virtual int32_t handleGetExtendedYear();

  /**
   * Override Calendar to compute several fields specific to the Persian
   * calendar system.  These are:
   *
   * <ul><li>ERA
   * <li>YEAR
   * <li>MONTH
   * <li>DAY_OF_MONTH
   * <li>DAY_OF_YEAR
   * <li>EXTENDED_YEAR</ul>
   * 
   * The DAY_OF_WEEK and DOW_LOCAL fields are already set when this
   * method is called. The getGregorianXxx() methods return Gregorian
   * calendar equivalents for the given Julian day.
   * @internal
   */
  virtual void handleComputeFields(int32_t julianDay, UErrorCode &status);

  // UObject stuff
 public: 
  /**
   * @return   The class ID for this object. All objects of a given class have the
   *           same class ID. Objects of other classes have different class IDs.
   * @internal
   */
  virtual UClassID getDynamicClassID(void) const;

  /**
   * Return the class ID for this class. This is useful only for comparing to a return
   * value from getDynamicClassID(). For example:
   *
   *      Base* polymorphic_pointer = createPolymorphicObject();
   *      if (polymorphic_pointer->getDynamicClassID() ==
   *          Derived::getStaticClassID()) ...
   *
   * @return   The class ID for all objects of this class.
   * @internal
   */
  U_I18N_API static UClassID U_EXPORT2 getStaticClassID(void);

  /**
   * return the calendar type, "persian".
   *
   * @return calendar type
   * @internal
   */
  virtual const char * getType() const;

 private:
  PersianCalendar(); // default constructor not implemented

 protected:

  /**
   * (Overrides Calendar) Return true if the current date for this Calendar is in
   * Daylight Savings Time. Recognizes DST_OFFSET, if it is set.
   *
   * @param status Fill-in parameter which receives the status of this operation.
   * @return   True if the current date for this Calendar is in Daylight Savings Time,
   *           false, otherwise.
   * @internal
   */
  virtual UBool inDaylightTime(UErrorCode& status) const;

  /**
   * Returns TRUE because the Persian Calendar does have a default century
   * @internal
   */
  virtual UBool haveDefaultCentury() const;

  /**
   * Returns the date of the start of the default century
   * @return start of century - in milliseconds since epoch, 1970
   * @internal
   */
  virtual UDate defaultCenturyStart() const;

  /**
   * Returns the year in which the default century begins
   * @internal
   */
  virtual int32_t defaultCenturyStartYear() const;
};

U_NAMESPACE_END

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *****************************************************************************
 * Copyright (C) 2003-2008, International Business Machines Corporation
 * and others. All Rights Reserved.
 *****************************************************************************
 *
 * File INDIANCAL.H
 *****************************************************************************
 */




#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

/**
 * Concrete class which provides the Indian calendar.
 * <P>
 * <code>IndianCalendar</code> is a subclass of <code>Calendar</code>
 * that numbers years since the begining of SAKA ERA.  This is the civil calendar
 * which is accepted by government of India as Indian National Calendar.
 * The two calendars most widely used in India today are the Vikrama calendar
 * followed in North India and the Shalivahana or Saka calendar which is followed
 * in South India and Maharashtra.

 * A variant of the Shalivahana Calendar was reformed and standardized as the
 * Indian National calendar in 1957.
 * <p>
 * Some details of Indian National Calendar (to be implemented) :
 * The Months
 * Month          Length      Start date (Gregorian)
 * =================================================
 * 1 Chaitra      30/31          March 22*
 * 2 Vaisakha     31             April 21
 * 3 Jyaistha     31             May 22
 * 4 Asadha       31             June 22
 * 5 Sravana      31             July 23
 * 6 Bhadra       31             August 23
 * 7 Asvina       30             September 23
 * 8 Kartika      30             October 23
 * 9 Agrahayana   30             November 22
 * 10 Pausa       30             December 22
 * 11 Magha       30             January 21
 * 12 Phalguna    30             February 20

 * In leap years, Chaitra has 31 days and starts on March 21 instead.
 * The leap years of Gregorian calendar and Indian National Calendar are in synchornization.
 * So When its a leap year in Gregorian calendar then Chaitra has 31 days.
 *
 * The Years
 * Years are counted in the Saka Era, which starts its year 0 in 78AD (by gregorian calendar).
 * So for eg. 9th June 2006 by Gregorian Calendar, is same as 19th of Jyaistha in 1928 of Saka
 * era by Indian National Calendar.
 * <p>
 * The Indian Calendar has only one allowable era: <code>Saka Era</code>.  If the
 * calendar is not in lenient mode (see <code>setLenient</code>), dates before
 * 1/1/1 Saka Era are rejected with an <code>IllegalArgumentException</code>.
 * <p>
 * @internal
 */


class U_I18N_API IndianCalendar : public Calendar {
public:
  /**
   * Useful constants for IndianCalendar.
   * @internal
   */
  enum EEras {
    /** 
     * Constant for Chaitra, the 1st month of the Indian year. 
     */
      CHAITRA,

      /**
     * Constant for Vaisakha, the 2nd month of the Indian year. 
     */
      VAISAKHA,

      /**
     * Constant for Jyaistha, the 3rd month of the Indian year. 
     */
      JYAISTHA,

    /**
     * Constant for Asadha, the 4th month of the Indian year. 
     */
      ASADHA,

    /**
     * Constant for Sravana, the 5th month of the Indian year. 
     */
      SRAVANA,

    /**
     * Constant for Bhadra the 6th month of the Indian year
     */
      BHADRA,

    /** 
     * Constant for the Asvina, the 7th month of the Indian year. 
     */
      ASVINA,

    /**
     * Constant for Kartika, the 8th month of the Indian year. 
     */
      KARTIKA,

    /**
     * Constant for Agrahayana, the 9th month of the Indian year. 
     */
      AGRAHAYANA,

    /**
     * Constant for Pausa, the 10th month of the Indian year. 
     */
      PAUSA,

    /**
     * Constant for Magha, the 11th month of the Indian year. 
     */
      MAGHA,

    /**
     * Constant for Phalguna, the 12th month of the Indian year. 
     */
      PHALGUNA
    };

  //-------------------------------------------------------------------------
  // Constructors...
  //-------------------------------------------------------------------------

  /**
   * Constructs an IndianCalendar based on the current time in the default time zone
   * with the given locale.
   *
   * @param aLocale  The given locale.
   * @param success  Indicates the status of IndianCalendar object construction.
   *                 Returns U_ZERO_ERROR if constructed successfully.
   * @param beCivil  Whether the calendar should be civil (default-TRUE) or religious (FALSE)
   * @internal
   */
  IndianCalendar(const Locale& aLocale, UErrorCode &success);

  /**
   * Copy Constructor
   * @internal
   */
  IndianCalendar(const IndianCalendar& other);

  /**
   * Destructor.
   * @internal
   */
  virtual ~IndianCalendar();

  /**
   * Determines whether this object uses the fixed-cycle Indian civil calendar
   * or an approximation of the religious, astronomical calendar.
   *
   * @param beCivil   <code>CIVIL</code> to use the civil calendar,
   *                  <code>ASTRONOMICAL</code> to use the astronomical calendar.
   * @internal
   */
  //void setCivil(ECivil beCivil, UErrorCode &status);
    
  /**
   * Returns <code>true</code> if this object is using the fixed-cycle civil
   * calendar, or <code>false</code> if using the religious, astronomical
   * calendar.
   * @internal
   */
  //UBool isCivil();


  // TODO: copy c'tor, etc

  // clone
  virtual IndianCalendar* clone() const;

 private:
  /**
   * Determine whether a year is the gregorian year a leap year 
   */
  //static UBool isGregorianLeap(int32_t year);
  //----------------------------------------------------------------------
  // Calendar framework
  //----------------------------------------------------------------------
 protected:
  /**
   * @internal
   */
  virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;
  
  /**
   * Return the length (in days) of the given month.
   *
   * @param year  The year in Saka era
   * @param year  The month(0-based) in Indian year
   * @internal
   */
  virtual int32_t handleGetMonthLength(int32_t extendedYear, int32_t month) const;
  
  /**
   * Return the number of days in the given Indian year
   * @internal
   */
  virtual int32_t handleGetYearLength(int32_t extendedYear) const;

  //-------------------------------------------------------------------------
  // Functions for converting from field values to milliseconds....
  //-------------------------------------------------------------------------

  // Return JD of start of given month/year
  /**
   * @internal
   */
  virtual int32_t handleComputeMonthStart(int32_t eyear, int32_t month, UBool useMonth) const;

  //-------------------------------------------------------------------------
  // Functions for converting from milliseconds to field values
  //-------------------------------------------------------------------------

  /**
   * @internal
   */
  virtual int32_t handleGetExtendedYear();

  /**
   * Override Calendar to compute several fields specific to the Indian
   * calendar system.  These are:
   *
   * <ul><li>ERA
   * <li>YEAR
   * <li>MONTH
   * <li>DAY_OF_MONTH
   * <li>DAY_OF_YEAR
   * <li>EXTENDED_YEAR</ul>
   * 
   * The DAY_OF_WEEK and DOW_LOCAL fields are already set when this
   * method is called. The getGregorianXxx() methods return Gregorian
   * calendar equivalents for the given Julian day.
   * @internal
   */
  virtual void handleComputeFields(int32_t julianDay, UErrorCode &status);

  // UObject stuff
 public: 
  /**
   * @return   The class ID for this object. All objects of a given class have the
   *           same class ID. Objects of other classes have different class IDs.
   * @internal
   */
  virtual UClassID getDynamicClassID(void) const;

  /**
   * Return the class ID for this class. This is useful only for comparing to a return
   * value from getDynamicClassID(). For example:
   *
   *      Base* polymorphic_pointer = createPolymorphicObject();
   *      if (polymorphic_pointer->getDynamicClassID() ==
   *          Derived::getStaticClassID()) ...
   *
   * @return   The class ID for all objects of this class.
   * @internal
   */
  static UClassID U_EXPORT2 getStaticClassID(void);

  /**
   * return the calendar type, "indian".
   *
   * @return calendar type
   * @internal
   */
  virtual const char * getType() const;

private:
  IndianCalendar(); // default constructor not implemented

  // Default century.
protected:

  /**
   * (Overrides Calendar) Return true if the current date for this Calendar is in
   * Daylight Savings Time. Recognizes DST_OFFSET, if it is set.
   *
   * @param status Fill-in parameter which receives the status of this operation.
   * @return   True if the current date for this Calendar is in Daylight Savings Time,
   *           false, otherwise.
   * @internal
   */
  virtual UBool inDaylightTime(UErrorCode& status) const;


  /**
   * Returns TRUE because the Indian Calendar does have a default century
   * @internal
   */
  virtual UBool haveDefaultCentury() const;

  /**
   * Returns the date of the start of the default century
   * @return start of century - in milliseconds since epoch, 1970
   * @internal
   */
  virtual UDate defaultCenturyStart() const;

  /**
   * Returns the year in which the default century begins
   * @internal
   */
  virtual int32_t defaultCenturyStartYear() const;
};

U_NAMESPACE_END

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *****************************************************************************
 * Copyright (C) 2007-2013, International Business Machines Corporation
 * and others. All Rights Reserved.
 *****************************************************************************
 *
 * File CHNSECAL.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   9/18/2007  ajmacher         ported from java ChineseCalendar
 *****************************************************************************
 */




#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

/**
 * <code>ChineseCalendar</code> is a concrete subclass of {@link Calendar}
 * that implements a traditional Chinese calendar.  The traditional Chinese
 * calendar is a lunisolar calendar: Each month starts on a new moon, and
 * the months are numbered according to solar events, specifically, to
 * guarantee that month 11 always contains the winter solstice.  In order
 * to accomplish this, leap months are inserted in certain years.  Leap
 * months are numbered the same as the month they follow.  The decision of
 * which month is a leap month depends on the relative movements of the sun
 * and moon.
 *
 * <p>This class defines one addition field beyond those defined by
 * <code>Calendar</code>: The <code>IS_LEAP_MONTH</code> field takes the
 * value of 0 for normal months, or 1 for leap months.
 *
 * <p>All astronomical computations are performed with respect to a time
 * zone of GMT+8:00 and a longitude of 120 degrees east.  Although some
 * calendars implement a historically more accurate convention of using
 * Beijing's local longitude (116 degrees 25 minutes east) and time zone
 * (GMT+7:45:40) for dates before 1929, we do not implement this here.
 *
 * <p>Years are counted in two different ways in the Chinese calendar.  The
 * first method is by sequential numbering from the 61st year of the reign
 * of Huang Di, 2637 BCE, which is designated year 1 on the Chinese
 * calendar.  The second method uses 60-year cycles from the same starting
 * point, which is designated year 1 of cycle 1.  In this class, the
 * <code>EXTENDED_YEAR</code> field contains the sequential year count.
 * The <code>ERA</code> field contains the cycle number, and the
 * <code>YEAR</code> field contains the year of the cycle, a value between
 * 1 and 60.
 *
 * <p>There is some variation in what is considered the starting point of
 * the calendar, with some sources starting in the first year of the reign
 * of Huang Di, rather than the 61st.  This gives continuous year numbers
 * 60 years greater and cycle numbers one greater than what this class
 * implements.
 *
 * <p>Because <code>ChineseCalendar</code> defines an additional field and
 * redefines the way the <code>ERA</code> field is used, it requires a new
 * format class, <code>ChineseDateFormat</code>.  As always, use the
 * methods <code>DateFormat.getXxxInstance(Calendar cal,...)</code> to
 * obtain a formatter for this calendar.
 *
 * <p>References:<ul>
 * 
 * <li>Dershowitz and Reingold, <i>Calendrical Calculations</i>,
 * Cambridge University Press, 1997</li>
 * 
 * <li>Helmer Aslaksen's
 * <a href="http://www.math.nus.edu.sg/aslaksen/calendar/chinese.shtml">
 * Chinese Calendar page</a></li>
 *
 * <li>The <a href="http://www.tondering.dk/claus/calendar.html">
 * Calendar FAQ</a></li>
 *
 * </ul>
 *
 * <p>
 * This class should only be subclassed to implement variants of the Chinese lunar calendar.</p>
 * <p>
 * ChineseCalendar usually should be instantiated using 
 * {@link com.ibm.icu.util.Calendar#getInstance(ULocale)} passing in a <code>ULocale</code>
 * with the tag <code>"@calendar=chinese"</code>.</p>
 *
 * @see com.ibm.icu.text.ChineseDateFormat
 * @see com.ibm.icu.util.Calendar
 * @author Alan Liu
 * @internal
 */
class U_I18N_API ChineseCalendar : public Calendar {
 public:
  //-------------------------------------------------------------------------
  // Constructors...
  //-------------------------------------------------------------------------

  /**
   * Constructs a ChineseCalendar based on the current time in the default time zone
   * with the given locale.
   *
   * @param aLocale  The given locale.
   * @param success  Indicates the status of ChineseCalendar object construction.
   *                 Returns U_ZERO_ERROR if constructed successfully.
   * @internal
   */
  ChineseCalendar(const Locale& aLocale, UErrorCode &success);

 protected:
 
   /**
   * Constructs a ChineseCalendar based on the current time in the default time zone
   * with the given locale, using the specified epoch year and time zone for
   * astronomical calculations.
   *
   * @param aLocale         The given locale.
   * @param epochYear       The epoch year to use for calculation.
   * @param zoneAstroCalc   The TimeZone to use for astronomical calculations. If null,
   *                        will be set appropriately for Chinese calendar (UTC + 8:00).
   * @param success         Indicates the status of ChineseCalendar object construction;
   *                        if successful, will not be changed to an error value.
   * @internal
   */
  ChineseCalendar(const Locale& aLocale, int32_t epochYear, const TimeZone* zoneAstroCalc, UErrorCode &success);

 public:
  /**
   * Copy Constructor
   * @internal
   */
  ChineseCalendar(const ChineseCalendar& other);

  /**
   * Destructor.
   * @internal
   */
  virtual ~ChineseCalendar();

  // clone
  virtual ChineseCalendar* clone() const;

 private:

  //-------------------------------------------------------------------------
  // Internal data....
  //-------------------------------------------------------------------------
    
  UBool isLeapYear;
  int32_t fEpochYear;   // Start year of this Chinese calendar instance.
  const TimeZone* fZoneAstroCalc;   // Zone used for the astronomical calculation
                                    // of this Chinese calendar instance.

  //----------------------------------------------------------------------
  // Calendar framework
  //----------------------------------------------------------------------

 protected:
  virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;
  virtual int32_t handleGetMonthLength(int32_t extendedYear, int32_t month) const;
  virtual int32_t handleComputeMonthStart(int32_t eyear, int32_t month, UBool useMonth) const;
  virtual int32_t handleGetExtendedYear();
  virtual void handleComputeFields(int32_t julianDay, UErrorCode &status);
  virtual const UFieldResolutionTable* getFieldResolutionTable() const;

 public:
  virtual void add(UCalendarDateFields field, int32_t amount, UErrorCode &status);
  virtual void add(EDateFields field, int32_t amount, UErrorCode &status);
  virtual void roll(UCalendarDateFields field, int32_t amount, UErrorCode &status);
  virtual void roll(EDateFields field, int32_t amount, UErrorCode &status);

  //----------------------------------------------------------------------
  // Internal methods & astronomical calculations
  //----------------------------------------------------------------------

 private:

  static const UFieldResolutionTable CHINESE_DATE_PRECEDENCE[];

  double daysToMillis(double days) const;
  double millisToDays(double millis) const;
  virtual int32_t winterSolstice(int32_t gyear) const;
  virtual int32_t newMoonNear(double days, UBool after) const;
  virtual int32_t synodicMonthsBetween(int32_t day1, int32_t day2) const;
  virtual int32_t majorSolarTerm(int32_t days) const;
  virtual UBool hasNoMajorSolarTerm(int32_t newMoon) const;
  virtual UBool isLeapMonthBetween(int32_t newMoon1, int32_t newMoon2) const;
  virtual void computeChineseFields(int32_t days, int32_t gyear,
                 int32_t gmonth, UBool setAllFields);
  virtual int32_t newYear(int32_t gyear) const;
  virtual void offsetMonth(int32_t newMoon, int32_t dom, int32_t delta);
  const TimeZone* getChineseCalZoneAstroCalc(void) const;

  // UObject stuff
 public: 
  /**
   * @return   The class ID for this object. All objects of a given class have the
   *           same class ID. Objects of other classes have different class IDs.
   * @internal
   */
  virtual UClassID getDynamicClassID(void) const;

  /**
   * Return the class ID for this class. This is useful only for comparing to a return
   * value from getDynamicClassID(). For example:
   *
   *      Base* polymorphic_pointer = createPolymorphicObject();
   *      if (polymorphic_pointer->getDynamicClassID() ==
   *          Derived::getStaticClassID()) ...
   *
   * @return   The class ID for all objects of this class.
   * @internal
   */
  static UClassID U_EXPORT2 getStaticClassID(void);

  /**
   * return the calendar type, "chinese".
   *
   * @return calendar type
   * @internal
   */
  virtual const char * getType() const;


 protected:
  /**
   * (Overrides Calendar) Return true if the current date for this Calendar is in
   * Daylight Savings Time. Recognizes DST_OFFSET, if it is set.
   *
   * @param status Fill-in parameter which receives the status of this operation.
   * @return   True if the current date for this Calendar is in Daylight Savings Time,
   *           false, otherwise.
   * @internal
   */
  virtual UBool inDaylightTime(UErrorCode& status) const;


  /**
   * Returns TRUE because the Islamic Calendar does have a default century
   * @internal
   */
  virtual UBool haveDefaultCentury() const;

  /**
   * Returns the date of the start of the default century
   * @return start of century - in milliseconds since epoch, 1970
   * @internal
   */
  virtual UDate defaultCenturyStart() const;

  /**
   * Returns the year in which the default century begins
   * @internal
   */
  virtual int32_t defaultCenturyStartYear() const;

 private: // default century stuff.

  /**
   * Returns the beginning date of the 100-year window that dates 
   * with 2-digit years are considered to fall within.
   */
  UDate         internalGetDefaultCenturyStart(void) const;

  /**
   * Returns the first year of the 100-year window that dates with 
   * 2-digit years are considered to fall within.
   */
  int32_t          internalGetDefaultCenturyStartYear(void) const;

  ChineseCalendar(); // default constructor not implemented
};

U_NAMESPACE_END

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2003 - 2013, International Business Machines Corporation and  *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2003 - 2008, International Business Machines Corporation and  *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

/**
 * Base class for EthiopicCalendar and CopticCalendar.
 * @internal
 */
class U_I18N_API CECalendar : public Calendar {

protected:
    //-------------------------------------------------------------------------
    // Constructors...
    //-------------------------------------------------------------------------

    /**
     * Constructs a CECalendar based on the current time in the default time zone
     * with the given locale with the Julian epoch offiset
     *
     * @param aLocale  The given locale.
     * @param success  Indicates the status of CECalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @internal
     */
    CECalendar(const Locale& aLocale, UErrorCode& success);

    /**
     * Copy Constructor
     * @internal
     */
    CECalendar (const CECalendar& other);

    /**
     * Destructor.
     * @internal
     */
    virtual ~CECalendar();

    /**
     * Default assignment operator
     * @param right    Calendar object to be copied
     * @internal
     */
    CECalendar& operator=(const CECalendar& right);

protected:
    //-------------------------------------------------------------------------
    // Calendar framework
    //-------------------------------------------------------------------------

    /**
     * Return JD of start of given month/extended year
     * @internal
     */
    virtual int32_t handleComputeMonthStart(int32_t eyear, int32_t month, UBool useMonth) const;

    /**
     * Calculate the limit for a specified type of limit and field
     * @internal
     */
    virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;

    /**
     * (Overrides Calendar) Return true if the current date for this Calendar is in
     * Daylight Savings Time. Recognizes DST_OFFSET, if it is set.
     *
     * @param status Fill-in parameter which receives the status of this operation.
     * @return   True if the current date for this Calendar is in Daylight Savings Time,
     *           false, otherwise.
     * @internal
     */
    virtual UBool inDaylightTime(UErrorCode&) const;

    /**
     * Returns TRUE because Coptic/Ethiopic Calendar does have a default century
     * @internal
     */
    virtual UBool haveDefaultCentury() const;

protected:
    /**
     * The Coptic and Ethiopic calendars differ only in their epochs.
     * This method must be implemented by CECalendar subclasses to
     * return the date offset from Julian
     * @internal
     */
    virtual int32_t getJDEpochOffset() const = 0;

    /**
     * Convert an Coptic/Ethiopic year, month, and day to a Julian day.
     *
     * @param year the extended year
     * @param month the month
     * @param day the day
     * @param jdEpochOffset the epoch offset from Julian epoch
     * @return Julian day
     * @internal
     */
    static int32_t ceToJD(int32_t year, int32_t month, int32_t date,
        int32_t jdEpochOffset);

    /**
     * Convert a Julian day to an Coptic/Ethiopic year, month and day
     *
     * @param julianDay the Julian day
     * @param jdEpochOffset the epoch offset from Julian epoch
     * @param year receives the extended year
     * @param month receives the month
     * @param date receives the day
     * @internal
     */
    static void jdToCE(int32_t julianDay, int32_t jdEpochOffset,
        int32_t& year, int32_t& month, int32_t& day);
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */


U_NAMESPACE_BEGIN

/**
 * Implement the Coptic calendar system.
 * @internal
 */
class CopticCalendar : public CECalendar {
  
public:
    /**
     * Useful constants for CopticCalendar.
     * @internal
     */
    enum EMonths {
        /** 
         * Constant for &#x03c9;&#x03bf;&#x03b3;&#x03c4;/&#x062a;&#xfeee;&#xfe97;,
         * the 1st month of the Coptic year. 
         */
        TOUT,
        
        /** 
         * Constant for &#x03a0;&#x03b1;&#x03bf;&#x03c0;&#x03b9;/&#xfeea;&#xfe91;&#xfe8e;&#xfe91;,
         * the 2nd month of the Coptic year. 
         */
        BABA,

        /** 
         * Constant for &#x0391;&#x03b8;&#x03bf;&#x03c1;/&#x0631;&#xfeee;&#xfe97;&#xfe8e;&#xfeeb;,
         * the 3rd month of the Coptic year. 
         */
        HATOR,

        /** 
         * Constant for &#x03a7;&#x03bf;&#x03b9;&#x03b1;&#x03ba;/&#xfeda;&#xfeec;&#xfef4;&#xfedb;,
         * the 4th month of the Coptic year. 
         */
        KIAHK,

        /** 
         * Constant for &#x03a4;&#x03c9;&#x03b2;&#x03b9;/&#x0637;&#xfeee;&#xfe92;&#xfeeb;,
         * the 5th month of the Coptic year. 
         */
        TOBA,

        /** 
         * Constant for &#x039c;&#x03b5;&#x03e3;&#x03b9;&#x03c1;/&#xfeae;&#xfef4;&#xfeb8;&#xfee3;&#x0623;,
         * the 6th month of the Coptic year. 
         */
        AMSHIR,

        /** 
         * Constant for &#x03a0;&#x03b1;&#x03c1;&#x03b5;&#x03bc;&#x03e9;&#x03b1;&#x03c4;/&#x062a;&#xfe8e;&#xfeec;&#xfee3;&#xfeae;&#xfe91;,
         * the 7th month of the Coptic year. 
         */
        BARAMHAT,

        /** 
         * Constant for &#x03a6;&#x03b1;&#x03c1;&#x03bc;&#x03bf;&#x03b8;&#x03b9;/&#x0647;&#x062f;&#xfeee;&#xfee3;&#xfeae;&#xfe91;, 
         * the 8th month of the Coptic year. 
         */
        BARAMOUDA,

        /** 
         * Constant for &#x03a0;&#x03b1;&#x03e3;&#x03b1;&#x03bd;/&#xfeb2;&#xfee8;&#xfeb8;&#xfe91;,
         * the 9th month of the Coptic year. 
         */
        BASHANS,

        /** 
         * Constant for &#x03a0;&#x03b1;&#x03c9;&#x03bd;&#x03b9;/&#xfeea;&#xfee7;&#x0624;&#xfeee;&#xfe91;,
         * the 10th month of the Coptic year. 
         */
        PAONA,

        /** 
         * Constant for &#x0395;&#x03c0;&#x03b7;&#x03c0;/&#xfe90;&#xfef4;&#xfe91;&#x0623;,
         * the 11th month of the Coptic year. 
         */
        EPEP,

        /** 
         * Constant for &#x039c;&#x03b5;&#x03f2;&#x03c9;&#x03c1;&#x03b7;/&#x0649;&#xfeae;&#xfeb4;&#xfee3;,
         * the 12th month of the Coptic year. 
         */
        MESRA,

        /** 
         * Constant for &#x03a0;&#x03b9;&#x03ba;&#x03bf;&#x03b3;&#x03eb;&#x03b9;
         * &#x03bc;&#x03b1;&#x03b2;&#x03bf;&#x03c4;/&#xfeae;&#xfef4;&#xfed0;&#xfebc;&#xfedf;&#x0627;
         * &#xfeae;&#xfeec;&#xfeb8;&#xfedf;&#x0627;,
         * the 13th month of the Coptic year. 
         */
        NASIE
    };

    enum EEras {
        BCE,    // Before the epoch
        CE      // After the epoch
    };

    /**
     * Constructs a CopticCalendar based on the current time in the default time zone
     * with the given locale.
     *
     * @param aLocale  The given locale.
     * @param success  Indicates the status of CopticCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @internal
     */
    CopticCalendar(const Locale& aLocale, UErrorCode& success);

    /**
     * Copy Constructor
     * @internal
     */
    CopticCalendar (const CopticCalendar& other);

    /**
     * Destructor.
     * @internal
     */
    virtual ~CopticCalendar();

    /**
     * Create and return a polymorphic copy of this calendar.
     * @return    return a polymorphic copy of this calendar.
     * @internal
     */
    virtual CopticCalendar* clone() const;

    /**
     * return the calendar type, "coptic"
     * @return calendar type
     * @internal
     */
    const char * getType() const;

protected:
    //-------------------------------------------------------------------------
    // Calendar framework
    //-------------------------------------------------------------------------

    /**
     * Return the extended year defined by the current fields.
     * @internal
     */
    virtual int32_t handleGetExtendedYear();

    /**
     * Compute fields from the JD
     * @internal
     */
    virtual void handleComputeFields(int32_t julianDay, UErrorCode &status);

    /**
     * Returns the date of the start of the default century
     * @return start of century - in milliseconds since epoch, 1970
     * @internal
     */
    virtual UDate defaultCenturyStart() const;

    /**
     * Returns the year in which the default century begins
     * @internal
     */
    virtual int32_t defaultCenturyStartYear() const;

    /**
     * Return the date offset from Julian
     * @internal
     */
    virtual int32_t getJDEpochOffset() const;


public:
    /**
     * Override Calendar Returns a unique class ID POLYMORPHICALLY. Pure virtual
     * override. This method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone() methods call
     * this method.
     *
     * @return   The class ID for this object. All objects of a given class have the
     *           same class ID. Objects of other classes have different class IDs.
     * @internal
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Return the class ID for this class. This is useful only for comparing to a return
     * value from getDynamicClassID(). For example:
     *
     *      Base* polymorphic_pointer = createPolymorphicObject();
     *      if (polymorphic_pointer->getDynamicClassID() ==
     *          Derived::getStaticClassID()) ...
     *
     * @return   The class ID for all objects of this class.
     * @internal
     */
    U_I18N_API static UClassID U_EXPORT2 getStaticClassID(void);  

#if 0
    // We do not want to introduce this API in ICU4C.
    // It was accidentally introduced in ICU4J as a public API.
public:
    //-------------------------------------------------------------------------
    // Calendar system Conversion methods...
    //-------------------------------------------------------------------------
    /**
     * Convert an Coptic year, month, and day to a Julian day.
     *
     * @param year the extended year
     * @param month the month
     * @param day the day
     * @return Julian day
     * @internal
     */
    static int32_t copticToJD(int32_t year, int32_t month, int32_t day);
#endif
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *****************************************************************************
 * Copyright (C) 2013, International Business Machines Corporation
 * and others. All Rights Reserved.
 *****************************************************************************
 *
 * File DANGICAL.H
 *****************************************************************************
 */




#if !UCONFIG_NO_FORMATTING





U_NAMESPACE_BEGIN

/**
 * <p><code>DangiCalendar</code> is a concrete subclass of {@link Calendar}
 * that implements a traditional Korean lunisolar calendar.</p>
 *
 * <p>DangiCalendar usually should be instantiated using 
 * {@link com.ibm.icu.util.Calendar#getInstance(ULocale)} passing in a <code>ULocale</code>
 * with the tag <code>"@calendar=dangi"</code>.</p>
 *
 * @internal
 */
class DangiCalendar : public ChineseCalendar {
 public:
  //-------------------------------------------------------------------------
  // Constructors...
  //-------------------------------------------------------------------------

  /**
   * Constructs a DangiCalendar based on the current time in the default time zone
   * with the given locale.
   *
   * @param aLocale  The given locale.
   * @param success  Indicates the status of DangiCalendar object construction.
   *                 Returns U_ZERO_ERROR if constructed successfully.
   * @internal
   */
  DangiCalendar(const Locale& aLocale, UErrorCode &success);

  /**
   * Copy Constructor
   * @internal
   */
  DangiCalendar(const DangiCalendar& other);

  /**
   * Destructor.
   * @internal
   */
  virtual ~DangiCalendar();

  /**
   * Clone.
   * @internal
   */
  virtual DangiCalendar* clone() const;

  //----------------------------------------------------------------------
  // Internal methods & astronomical calculations
  //----------------------------------------------------------------------

 private:

  const TimeZone* getDangiCalZoneAstroCalc(void) const;

  // UObject stuff
 public: 
  /**
   * @return   The class ID for this object. All objects of a given class have the
   *           same class ID. Objects of other classes have different class IDs.
   * @internal
   */
  virtual UClassID getDynamicClassID(void) const;

  /**
   * Return the class ID for this class. This is useful only for comparing to a return
   * value from getDynamicClassID(). For example:
   *
   *      Base* polymorphic_pointer = createPolymorphicObject();
   *      if (polymorphic_pointer->getDynamicClassID() ==
   *          Derived::getStaticClassID()) ...
   *
   * @return   The class ID for all objects of this class.
   * @internal
   */
  U_I18N_API static UClassID U_EXPORT2 getStaticClassID(void);

  /**
   * return the calendar type, "dangi".
   *
   * @return calendar type
   * @internal
   */
  const char * getType() const;


 private:
 
  DangiCalendar(); // default constructor not implemented
};

U_NAMESPACE_END

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2003 - 2013, International Business Machines Corporation and  *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

/**
 * Implement the Ethiopic calendar system.
 * @internal
 */
class EthiopicCalendar : public CECalendar {

public:
    /**
     * Calendar type - use Amete Alem era for all the time or not
     * @internal 
     */
    enum EEraType {
        AMETE_MIHRET_ERA,
        AMETE_ALEM_ERA
    };

    /**
     * Useful constants for EthiopicCalendar.
     * @internal
     */
    enum EMonths {
        /** 
         * Constant for &#x1218;&#x1235;&#x12a8;&#x1228;&#x121d;, the 1st month of the Ethiopic year.
         */
        MESKEREM,

        /** 
         * Constant for &#x1325;&#x1245;&#x121d;&#x1275;, the 2nd month of the Ethiopic year.
         */
        TEKEMT,

        /** 
         * Constant for &#x1285;&#x12f3;&#x122d;, the 3rd month of the Ethiopic year. 
         */
        HEDAR,

        /** 
         * Constant for &#x1273;&#x1285;&#x1223;&#x1225;, the 4th month of the Ethiopic year. 
         */
        TAHSAS,

        /** 
         * Constant for &#x1325;&#x122d;, the 5th month of the Ethiopic year. 
         */
        TER,

        /** 
         * Constant for &#x12e8;&#x12ab;&#x1272;&#x1275;, the 6th month of the Ethiopic year. 
         */
        YEKATIT,

        /** 
         * Constant for &#x1218;&#x130b;&#x1262;&#x1275;, the 7th month of the Ethiopic year. 
         */
        MEGABIT,

        /** 
         * Constant for &#x121a;&#x12eb;&#x12dd;&#x12eb;, the 8th month of the Ethiopic year. 
         */
        MIAZIA,

        /** 
         * Constant for &#x130d;&#x1295;&#x1266;&#x1275;, the 9th month of the Ethiopic year. 
         */
        GENBOT,

        /** 
         * Constant for &#x1230;&#x1294;, the 10th month of the Ethiopic year. 
         */
        SENE,

        /** 
         * Constant for &#x1210;&#x121d;&#x120c;, the 11th month of the Ethiopic year. 
         */
        HAMLE,

        /** 
         * Constant for &#x1290;&#x1210;&#x1234;, the 12th month of the Ethiopic year. 
         */
        NEHASSA,

        /** 
         * Constant for &#x1333;&#x1309;&#x121c;&#x1295;, the 13th month of the Ethiopic year. 
         */
        PAGUMEN
    };

    enum EEras {
        AMETE_ALEM,     // Before the epoch
        AMETE_MIHRET    // After the epoch
    };

    /**
     * Constructs a EthiopicCalendar based on the current time in the default time zone
     * with the given locale.
     *
     * @param aLocale  The given locale.
     * @param success  Indicates the status of EthiopicCalendar object construction.
     *                 Returns U_ZERO_ERROR if constructed successfully.
     * @param type     Whether this Ethiopic calendar use Amete Mihrret (default) or
     *                 only use Amete Alem for all the time.
     * @internal
     */
    EthiopicCalendar(const Locale& aLocale, UErrorCode& success, EEraType type = AMETE_MIHRET_ERA);

    /**
     * Copy Constructor
     * @internal
     */
    EthiopicCalendar(const EthiopicCalendar& other);

    /**
     * Destructor.
     * @internal
     */
    virtual ~EthiopicCalendar();

    /**
     * Create and return a polymorphic copy of this calendar.
     * @return    return a polymorphic copy of this calendar.
     * @internal
     */
    virtual EthiopicCalendar* clone() const;

    /**
     * return the calendar type, "ethiopic"
     * @return calendar type
     * @internal
     */
    virtual const char * getType() const;

    /**
     * Set Alem or Mihret era.
     * @param onOff Set Amete Alem era if true, otherwise set Amete Mihret era.
     * @internal
     */
    void setAmeteAlemEra (UBool onOff);

    /**
     * Return true if this calendar is set to the Amete Alem era.
     * @return true if set to the Amete Alem era.
     * @internal
     */
    UBool isAmeteAlemEra() const;

protected:
    //-------------------------------------------------------------------------
    // Calendar framework
    //-------------------------------------------------------------------------

    /**
     * Return the extended year defined by the current fields.
     * @internal
     */
    virtual int32_t handleGetExtendedYear();

    /**
     * Compute fields from the JD
     * @internal
     */
    virtual void handleComputeFields(int32_t julianDay, UErrorCode &status);

    /**
     * Calculate the limit for a specified type of limit and field
     * @internal
     */
    virtual int32_t handleGetLimit(UCalendarDateFields field, ELimitType limitType) const;

    /**
     * Returns the date of the start of the default century
     * @return start of century - in milliseconds since epoch, 1970
     * @internal
     */
    virtual UDate defaultCenturyStart() const;

    /**
     * Returns the year in which the default century begins
     * @internal
     */
    virtual int32_t defaultCenturyStartYear() const;

    /**
     * Return the date offset from Julian
     * @internal
     */
    virtual int32_t getJDEpochOffset() const;

private:
    /**
     * When eraType is AMETE_ALEM_ERA, then this calendar use only AMETE_ALEM
     * for the era. Otherwise (default), this calendar uses both AMETE_ALEM
     * and AMETE_MIHRET.
     *
     * EXTENDED_YEAR        AMETE_ALEM_ERA     AMETE_MIHRET_ERA
     *             0       Amete Alem 5500      Amete Alem 5500
     *             1        Amete Mihret 1      Amete Alem 5501
     */
    EEraType eraType;

public:
    /**
     * Override Calendar Returns a unique class ID POLYMORPHICALLY. Pure virtual
     * override. This method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone() methods call
     * this method.
     *
     * @return   The class ID for this object. All objects of a given class have the
     *           same class ID. Objects of other classes have different class IDs.
     * @internal
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Return the class ID for this class. This is useful only for comparing to a return
     * value from getDynamicClassID(). For example:
     *
     *      Base* polymorphic_pointer = createPolymorphicObject();
     *      if (polymorphic_pointer->getDynamicClassID() ==
     *          Derived::getStaticClassID()) ...
     *
     * @return   The class ID for all objects of this class.
     * @internal
     */
    U_I18N_API static UClassID U_EXPORT2 getStaticClassID(void);  

#if 0
// We do not want to introduce this API in ICU4C.
// It was accidentally introduced in ICU4J as a public API.

public:
    //-------------------------------------------------------------------------
    // Calendar system Conversion methods...
    //-------------------------------------------------------------------------

    /**
     * Convert an Ethiopic year, month, and day to a Julian day.
     *
     * @param year the extended year
     * @param month the month
     * @param day the day
     * @return Julian day
     * @internal
     */
    int32_t ethiopicToJD(int32_t year, int32_t month, int32_t day);
#endif
};

U_NAMESPACE_END
#endif /* #if !UCONFIG_NO_FORMATTING */










// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2003-2013, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: July 21 2003
* Since: ICU 2.8
**********************************************************************
*/



#if !UCONFIG_NO_FORMATTING




struct UResourceBundle;

U_NAMESPACE_BEGIN

class SimpleTimeZone;

/**
 * A time zone based on the Olson tz database.  Olson time zones change
 * behavior over time.  The raw offset, rules, presence or absence of
 * daylight savings time, and even the daylight savings amount can all
 * vary.
 *
 * This class uses a resource bundle named "zoneinfo".  Zoneinfo is a
 * table containing different kinds of resources.  In several places,
 * zones are referred to using integers.  A zone's integer is a number
 * from 0..n-1, where n is the number of zones, with the zones sorted
 * in lexicographic order.
 *
 * 1. Zones.  These have keys corresponding to the Olson IDs, e.g.,
 * "Asia/Shanghai".  Each resource describes the behavior of the given
 * zone.  Zones come in two different formats.
 *
 *   a. Zone (table).  A zone is a table resource contains several
 *   type of resources below:
 *  
 *   - typeOffsets:intvector (Required)
 *  
 *   Sets of UTC raw/dst offset pairs in seconds.  Entries at
 *   2n represents raw offset and 2n+1 represents dst offset
 *   paired with the raw offset at 2n.  The very first pair represents
 *   the initial zone offset (before the first transition) always.
 *
 *   - trans:intvector (Optional) 
 *  
 *   List of transition times represented by 32bit seconds from the
 *   epoch (1970-01-01T00:00Z) in ascending order.
 *  
 *   - transPre32/transPost32:intvector (Optional)
 *  
 *   List of transition times before/after 32bit minimum seconds.
 *   Each time is represented by a pair of 32bit integer.
 * 
 *   - typeMap:bin (Optional)
 *  
 *   Array of bytes representing the mapping between each transition
 *   time (transPre32/trans/transPost32) and its corresponding offset
 *   data (typeOffsets).
 *  
 *   - finalRule:string (Optional)
 *  
 *   If a recurrent transition rule is applicable to a zone forever
 *   after the final transition time, finalRule represents the rule
 *   in Rules data.
 *  
 *   - finalRaw:int (Optional)
 *   
 *   When finalRule is available, finalRaw is required and specifies
 *   the raw (base) offset of the rule.
 *   
 *   - finalYear:int (Optional)
 *   
 *   When finalRule is available, finalYear is required and specifies
 *   the start year of the rule.
 *   
 *   - links:intvector (Optional)
 *   
 *   When this zone data is shared with other zones, links specifies
 *   all zones including the zone itself.  Each zone is referenced by
 *   integer index.
 * 
 *  b. Link (int, length 1).  A link zone is an int resource.  The
 *  integer is the zone number of the target zone.  The key of this
 *  resource is an alternate name for the target zone.  This data
 *  is corresponding to Link data in the tz database.
 *
 *
 * 2. Rules.  These have keys corresponding to the Olson rule IDs,
 * with an underscore prepended, e.g., "_EU".  Each resource describes
 * the behavior of the given rule using an intvector, containing the
 * onset list, the cessation list, and the DST savings.  The onset and
 * cessation lists consist of the month, dowim, dow, time, and time
 * mode.  The end result is that the 11 integers describing the rule
 * can be passed directly into the SimpleTimeZone 13-argument
 * constructor (the other two arguments will be the raw offset, taken
 * from the complex zone element 5, and the ID string, which is not
 * used), with the times and the DST savings multiplied by 1000 to
 * scale from seconds to milliseconds.
 *
 * 3. Regions.  An array specifies mapping between zones and regions.
 * Each item is either a 2-letter ISO country code or "001"
 * (UN M.49 - World).  This data is generated from "zone.tab"
 * in the tz database.
 */
class U_I18N_API OlsonTimeZone: public BasicTimeZone {
 public:
    /**
     * Construct from a resource bundle.
     * @param top the top-level zoneinfo resource bundle.  This is used
     * to lookup the rule that `res' may refer to, if there is one.
     * @param res the resource bundle of the zone to be constructed
     * @param tzid the time zone ID
     * @param ec input-output error code
     */
    OlsonTimeZone(const UResourceBundle* top,
                  const UResourceBundle* res,
                  const UnicodeString& tzid,
                  UErrorCode& ec);

    /**
     * Copy constructor
     */
    OlsonTimeZone(const OlsonTimeZone& other);

    /**
     * Destructor
     */
    virtual ~OlsonTimeZone();

    /**
     * Assignment operator
     */
    OlsonTimeZone& operator=(const OlsonTimeZone& other);

    /**
     * Returns true if the two TimeZone objects are equal.
     */
    virtual UBool operator==(const TimeZone& other) const;

    /**
     * TimeZone API.
     */
    virtual OlsonTimeZone* clone() const;

    /**
     * TimeZone API.
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * TimeZone API.
     */
    virtual UClassID getDynamicClassID() const;
    
    /**
     * TimeZone API.  Do not call this; prefer getOffset(UDate,...).
     */
    virtual int32_t getOffset(uint8_t era, int32_t year, int32_t month,
                              int32_t day, uint8_t dayOfWeek,
                              int32_t millis, UErrorCode& ec) const;

    /**
     * TimeZone API.  Do not call this; prefer getOffset(UDate,...).
     */
    virtual int32_t getOffset(uint8_t era, int32_t year, int32_t month,
                              int32_t day, uint8_t dayOfWeek,
                              int32_t millis, int32_t monthLength,
                              UErrorCode& ec) const;

    /**
     * TimeZone API.
     */
    virtual void getOffset(UDate date, UBool local, int32_t& rawOffset,
                   int32_t& dstOffset, UErrorCode& ec) const;

    /**
     * BasicTimeZone API.
     */
    virtual void getOffsetFromLocal(UDate date, int32_t nonExistingTimeOpt, int32_t duplicatedTimeOpt,
        int32_t& rawoff, int32_t& dstoff, UErrorCode& ec) const;

    /**
     * TimeZone API.  This method has no effect since objects of this
     * class are quasi-immutable (the base class allows the ID to be
     * changed).
     */
    virtual void setRawOffset(int32_t offsetMillis);

    /**
     * TimeZone API.  For a historical zone, the raw offset can change
     * over time, so this API is not useful.  In order to approximate
     * expected behavior, this method returns the raw offset for the
     * current moment in time.
     */
    virtual int32_t getRawOffset() const;

    /**
     * TimeZone API.  For a historical zone, whether DST is used or
     * not varies over time.  In order to approximate expected
     * behavior, this method returns TRUE if DST is observed at any
     * point in the current year.
     */
    virtual UBool useDaylightTime() const;

    /**
     * TimeZone API.
     */
    virtual UBool inDaylightTime(UDate date, UErrorCode& ec) const;

    /**
     * TimeZone API.
     */
    virtual int32_t getDSTSavings() const;

    /**
     * TimeZone API.  Also comare historic transitions.
     */
    virtual UBool hasSameRules(const TimeZone& other) const;

    /**
     * BasicTimeZone API.
     * Gets the first time zone transition after the base time.
     * @param base      The base time.
     * @param inclusive Whether the base time is inclusive or not.
     * @param result    Receives the first transition after the base time.
     * @return  TRUE if the transition is found.
     */
    virtual UBool getNextTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const;

    /**
     * BasicTimeZone API.
     * Gets the most recent time zone transition before the base time.
     * @param base      The base time.
     * @param inclusive Whether the base time is inclusive or not.
     * @param result    Receives the most recent transition before the base time.
     * @return  TRUE if the transition is found.
     */
    virtual UBool getPreviousTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const;

    /**
     * BasicTimeZone API.
     * Returns the number of <code>TimeZoneRule</code>s which represents time transitions,
     * for this time zone, that is, all <code>TimeZoneRule</code>s for this time zone except
     * <code>InitialTimeZoneRule</code>.  The return value range is 0 or any positive value.
     * @param status    Receives error status code.
     * @return The number of <code>TimeZoneRule</code>s representing time transitions.
     */
    virtual int32_t countTransitionRules(UErrorCode& status) const;

    /**
     * Gets the <code>InitialTimeZoneRule</code> and the set of <code>TimeZoneRule</code>
     * which represent time transitions for this time zone.  On successful return,
     * the argument initial points to non-NULL <code>InitialTimeZoneRule</code> and
     * the array trsrules is filled with 0 or multiple <code>TimeZoneRule</code>
     * instances up to the size specified by trscount.  The results are referencing the
     * rule instance held by this time zone instance.  Therefore, after this time zone
     * is destructed, they are no longer available.
     * @param initial       Receives the initial timezone rule
     * @param trsrules      Receives the timezone transition rules
     * @param trscount      On input, specify the size of the array 'transitions' receiving
     *                      the timezone transition rules.  On output, actual number of
     *                      rules filled in the array will be set.
     * @param status        Receives error status code.
     */
    virtual void getTimeZoneRules(const InitialTimeZoneRule*& initial,
        const TimeZoneRule* trsrules[], int32_t& trscount, UErrorCode& status) const;

    /**
     * Internal API returning the canonical ID of this zone.
     * This ID won't be affected by setID().
     */
    const UChar *getCanonicalID() const;

private:
    /**
     * Default constructor.  Creates a time zone with an empty ID and
     * a fixed GMT offset of zero.
     */
    OlsonTimeZone();

private:

    void constructEmpty();

    void getHistoricalOffset(UDate date, UBool local,
        int32_t NonExistingTimeOpt, int32_t DuplicatedTimeOpt,
        int32_t& rawoff, int32_t& dstoff) const;

    int16_t transitionCount() const;

    int64_t transitionTimeInSeconds(int16_t transIdx) const;
    double transitionTime(int16_t transIdx) const;

    /*
     * Following 3 methods return an offset at the given transition time index.
     * When the index is negative, return the initial offset.
     */
    int32_t zoneOffsetAt(int16_t transIdx) const;
    int32_t rawOffsetAt(int16_t transIdx) const;
    int32_t dstOffsetAt(int16_t transIdx) const;

    /*
     * Following methods return the initial offset.
     */
    int32_t initialRawOffset() const;
    int32_t initialDstOffset() const;

    /**
     * Number of transitions in each time range
     */
    int16_t transitionCountPre32;
    int16_t transitionCount32;
    int16_t transitionCountPost32;

    /**
     * Time of each transition in seconds from 1970 epoch before 32bit second range (<= 1900).
     * Each transition in this range is represented by a pair of int32_t.
     * Length is transitionCount int32_t's.  NULL if no transitions in this range.
     */
    const int32_t *transitionTimesPre32; // alias into res; do not delete

    /**
     * Time of each transition in seconds from 1970 epoch in 32bit second range.
     * Length is transitionCount int32_t's.  NULL if no transitions in this range.
     */
    const int32_t *transitionTimes32; // alias into res; do not delete

    /**
     * Time of each transition in seconds from 1970 epoch after 32bit second range (>= 2038).
     * Each transition in this range is represented by a pair of int32_t.
     * Length is transitionCount int32_t's.  NULL if no transitions in this range.
     */
    const int32_t *transitionTimesPost32; // alias into res; do not delete

    /**
     * Number of types, 1..255
     */
    int16_t typeCount;

    /**
     * Offset from GMT in seconds for each type.
     * Length is typeCount int32_t's.  At least one type (a pair of int32_t)
     * is required.
     */
    const int32_t *typeOffsets; // alias into res; do not delete

    /**
     * Type description data, consisting of transitionCount uint8_t
     * type indices (from 0..typeCount-1).
     * Length is transitionCount int16_t's.  NULL if no transitions.
     */
    const uint8_t *typeMapData; // alias into res; do not delete

    /**
     * A SimpleTimeZone that governs the behavior for date >= finalMillis.
     */
    SimpleTimeZone *finalZone; // owned, may be NULL

    /**
     * For date >= finalMillis, the finalZone will be used.
     */
    double finalStartMillis;

    /**
     * For year >= finalYear, the finalZone will be used.
     */
    int32_t finalStartYear;

    /*
     * Canonical (CLDR) ID of this zone
     */
    const UChar *canonicalID;

    /* BasicTimeZone support */
    void clearTransitionRules(void);
    void deleteTransitionRules(void);
    void checkTransitionRules(UErrorCode& status) const;

  public:    // Internal, for access from plain C code
    void initTransitionRules(UErrorCode& status);
  private:

    InitialTimeZoneRule *initialRule;
    TimeZoneTransition  *firstTZTransition;
    int16_t             firstTZTransitionIdx;
    TimeZoneTransition  *firstFinalTZTransition;
    TimeArrayTimeZoneRule   **historicRules;
    int16_t             historicRuleCount;
    SimpleTimeZone      *finalZoneWithStartYear; // hack
    UInitOnce           transitionRulesInitOnce = U_INITONCE_INITIALIZER;
};

inline int16_t
OlsonTimeZone::transitionCount() const {
    return transitionCountPre32 + transitionCount32 + transitionCountPost32;
}

inline double
OlsonTimeZone::transitionTime(int16_t transIdx) const {
    return (double)transitionTimeInSeconds(transIdx) * U_MILLIS_PER_SECOND;
}

inline int32_t
OlsonTimeZone::zoneOffsetAt(int16_t transIdx) const {
    int16_t typeIdx = (transIdx >= 0 ? typeMapData[transIdx] : 0) << 1;
    return typeOffsets[typeIdx] + typeOffsets[typeIdx + 1];
}

inline int32_t
OlsonTimeZone::rawOffsetAt(int16_t transIdx) const {
    int16_t typeIdx = (transIdx >= 0 ? typeMapData[transIdx] : 0) << 1;
    return typeOffsets[typeIdx];
}

inline int32_t
OlsonTimeZone::dstOffsetAt(int16_t transIdx) const {
    int16_t typeIdx = (transIdx >= 0 ? typeMapData[transIdx] : 0) << 1;
    return typeOffsets[typeIdx + 1];
}

inline int32_t
OlsonTimeZone::initialRawOffset() const {
    return typeOffsets[0];
}

inline int32_t
OlsonTimeZone::initialDstOffset() const {
    return typeOffsets[1];
}

inline const UChar*
OlsonTimeZone::getCanonicalID() const {
    return canonicalID;
}


U_NAMESPACE_END

#endif // !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2014, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* sharedcalendar.h
*/





U_NAMESPACE_BEGIN

class Calendar;

class U_I18N_API SharedCalendar : public SharedObject {
public:
    SharedCalendar(Calendar *calToAdopt) : ptr(calToAdopt) { }
    virtual ~SharedCalendar();
    const Calendar *get() const { return ptr; }
    const Calendar *operator->() const { return ptr; }
    const Calendar &operator*() const { return *ptr; }
private:
    Calendar *ptr;
    SharedCalendar(const SharedCalendar &);
    SharedCalendar &operator=(const SharedCalendar &);
};

U_NAMESPACE_END





#if !UCONFIG_NO_SERVICE
static icu::ICULocaleService* gService = NULL;
static icu::UInitOnce gServiceInitOnce = U_INITONCE_INITIALIZER;

// INTERNAL - for cleanup
U_CDECL_BEGIN
static UBool calendar_cleanup(void) {
#if !UCONFIG_NO_SERVICE
    if (gService) {
        delete gService;
        gService = NULL;
    }
    gServiceInitOnce.reset();
#endif
    return TRUE;
}
U_CDECL_END
#endif

// ------------------------------------------
//
// Registration
//
//-------------------------------------------
//#define U_DEBUG_CALSVC 1
//

#if defined( U_DEBUG_CALSVC ) || defined (U_DEBUG_CAL)

/**
 * fldName was removed as a duplicate implementation.
 * use  udbg_ services instead,
 * which depend on include files and library from ../tools/toolutil, the following circular link:
 *   CPPFLAGS+=-I$(top_srcdir)/tools/toolutil
 *   LIBS+=$(LIBICUTOOLUTIL)
 */
// #include "udbgutil.h"
#include <stdio.h>

/**
* convert a UCalendarDateFields into a string - for debugging
* @param f field enum
* @return static string to the field name
* @internal
*/

const char* fldName(UCalendarDateFields f) {
    return udbg_enumName(UDBG_UCalendarDateFields, (int32_t)f);
}

#if UCAL_DEBUG_DUMP
// from CalendarTest::calToStr - but doesn't modify contents.
void ucal_dump(const Calendar &cal) {
    cal.dump();
}

void Calendar::dump() const {
    int i;
    fprintf(stderr, "@calendar=%s, timeset=%c, fieldset=%c, allfields=%c, virtualset=%c, t=%.2f",
        getType(), fIsTimeSet?'y':'n',  fAreFieldsSet?'y':'n',  fAreAllFieldsSet?'y':'n',
        fAreFieldsVirtuallySet?'y':'n',
        fTime);

    // can add more things here: DST, zone, etc.
    fprintf(stderr, "\n");
    for(i = 0;i<UCAL_FIELD_COUNT;i++) {
        int n;
        const char *f = fldName((UCalendarDateFields)i);
        fprintf(stderr, "  %25s: %-11ld", f, fFields[i]);
        if(fStamp[i] == kUnset) {
            fprintf(stderr, " (unset) ");
        } else if(fStamp[i] == kInternallySet) {
            fprintf(stderr, " (internally set) ");
            //} else if(fStamp[i] == kInternalDefault) {
            //    fprintf(stderr, " (internal default) ");
        } else {
            fprintf(stderr, " %%%d ", fStamp[i]);
        }
        fprintf(stderr, "\n");

    }
}

U_CFUNC void ucal_dump(UCalendar* cal) {
    ucal_dump( *((Calendar*)cal)  );
}
#endif

#endif

/* Max value for stamp allowable before recalculation */
#define STAMP_MAX 10000

static const char * const gCalTypes[] = {
    "gregorian",
    "japanese",
    "buddhist",
    "roc",
    "persian",
    "islamic-civil",
    "islamic",
    "hebrew",
    "chinese",
    "indian",
    "coptic",
    "ethiopic",
    "ethiopic-amete-alem",
    "iso8601",
    "dangi",
    "islamic-umalqura",
    "islamic-tbla",
    "islamic-rgsa",
    NULL
};

// Must be in the order of gCalTypes above
typedef enum ECalType {
    CALTYPE_UNKNOWN = -1,
    CALTYPE_GREGORIAN = 0,
    CALTYPE_JAPANESE,
    CALTYPE_BUDDHIST,
    CALTYPE_ROC,
    CALTYPE_PERSIAN,
    CALTYPE_ISLAMIC_CIVIL,
    CALTYPE_ISLAMIC,
    CALTYPE_HEBREW,
    CALTYPE_CHINESE,
    CALTYPE_INDIAN,
    CALTYPE_COPTIC,
    CALTYPE_ETHIOPIC,
    CALTYPE_ETHIOPIC_AMETE_ALEM,
    CALTYPE_ISO8601,
    CALTYPE_DANGI,
    CALTYPE_ISLAMIC_UMALQURA,
    CALTYPE_ISLAMIC_TBLA,
    CALTYPE_ISLAMIC_RGSA
} ECalType;

U_NAMESPACE_BEGIN

SharedCalendar::~SharedCalendar() {
    delete ptr;
}

template<> U_I18N_API
const SharedCalendar *LocaleCacheKey<SharedCalendar>::createObject(
        const void * /*unusedCreationContext*/, UErrorCode &status) const {
    Calendar *calendar = Calendar::makeInstance(fLoc, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    SharedCalendar *shared = new SharedCalendar(calendar);
    if (shared == NULL) {
        delete calendar;
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    shared->addRef();
    return shared;
}

static ECalType getCalendarType(const char *s) {
    for (int i = 0; gCalTypes[i] != NULL; i++) {
        if (uprv_stricmp(s, gCalTypes[i]) == 0) {
            return (ECalType)i;
        }
    }
    return CALTYPE_UNKNOWN;
}

#if !UCONFIG_NO_SERVICE
// Only used with service registration.
static UBool isStandardSupportedKeyword(const char *keyword, UErrorCode& status) {
    if(U_FAILURE(status)) {
        return FALSE;
    }
    ECalType calType = getCalendarType(keyword);
    return (calType != CALTYPE_UNKNOWN);
}

// only used with service registration.
static void getCalendarKeyword(const UnicodeString &id, char *targetBuffer, int32_t targetBufferSize) {
    UnicodeString calendarKeyword = UNICODE_STRING_SIMPLE("calendar=");
    int32_t calKeyLen = calendarKeyword.length();
    int32_t keyLen = 0;

    int32_t keywordIdx = id.indexOf((UChar)0x003D); /* '=' */
    if (id[0] == 0x40/*'@'*/
        && id.compareBetween(1, keywordIdx+1, calendarKeyword, 0, calKeyLen) == 0)
    {
        keyLen = id.extract(keywordIdx+1, id.length(), targetBuffer, targetBufferSize, US_INV);
    }
    targetBuffer[keyLen] = 0;
}
#endif

static ECalType getCalendarTypeForLocale(const char *locid) {
    UErrorCode status = U_ZERO_ERROR;
    ECalType calType = CALTYPE_UNKNOWN;

    //TODO: ULOC_FULL_NAME is out of date and too small..
    char canonicalName[256];

    // canonicalize, so grandfathered variant will be transformed to keywords
    // e.g ja_JP_TRADITIONAL -> ja_JP@calendar=japanese
    // NOTE: Since ICU-20187, ja_JP_TRADITIONAL no longer canonicalizes, and
    // the Gregorian calendar is returned instead.
    int32_t canonicalLen = uloc_canonicalize(locid, canonicalName, sizeof(canonicalName) - 1, &status);
    if (U_FAILURE(status)) {
        return CALTYPE_GREGORIAN;
    }
    canonicalName[canonicalLen] = 0;    // terminate

    char calTypeBuf[32];
    int32_t calTypeBufLen;

    calTypeBufLen = uloc_getKeywordValue(canonicalName, "calendar", calTypeBuf, sizeof(calTypeBuf) - 1, &status);
    if (U_SUCCESS(status)) {
        calTypeBuf[calTypeBufLen] = 0;
        calType = getCalendarType(calTypeBuf);
        if (calType != CALTYPE_UNKNOWN) {
            return calType;
        }
    }
    status = U_ZERO_ERROR;

    // when calendar keyword is not available or not supported, read supplementalData
    // to get the default calendar type for the locale's region
    char region[ULOC_COUNTRY_CAPACITY];
    (void)ulocimp_getRegionForSupplementalData(canonicalName, TRUE, region, sizeof(region), &status);
    if (U_FAILURE(status)) {
        return CALTYPE_GREGORIAN;
    }

    // Read preferred calendar values from supplementalData calendarPreference
    UResourceBundle *rb = ures_openDirect(NULL, "supplementalData", &status);
    ures_getByKey(rb, "calendarPreferenceData", rb, &status);
    UResourceBundle *order = ures_getByKey(rb, region, NULL, &status);
    if (status == U_MISSING_RESOURCE_ERROR && rb != NULL) {
        status = U_ZERO_ERROR;
        order = ures_getByKey(rb, "001", NULL, &status);
    }

    calTypeBuf[0] = 0;
    if (U_SUCCESS(status) && order != NULL) {
        // the first calender type is the default for the region
        int32_t len = 0;
        const UChar *uCalType = ures_getStringByIndex(order, 0, &len, &status);
        if (len < (int32_t)sizeof(calTypeBuf)) {
            u_UCharsToChars(uCalType, calTypeBuf, len);
            *(calTypeBuf + len) = 0; // terminate;
            calType = getCalendarType(calTypeBuf);
        }
    }

    ures_close(order);
    ures_close(rb);

    if (calType == CALTYPE_UNKNOWN) {
        // final fallback
        calType = CALTYPE_GREGORIAN;
    }
    return calType;
}

static Calendar *createStandardCalendar(ECalType calType, const Locale &loc, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    LocalPointer<Calendar> cal;

    switch (calType) {
        case CALTYPE_GREGORIAN:
            cal.adoptInsteadAndCheckErrorCode(new GregorianCalendar(loc, status), status);
            break;
        case CALTYPE_JAPANESE:
            cal.adoptInsteadAndCheckErrorCode(new JapaneseCalendar(loc, status), status);
            break;
        case CALTYPE_BUDDHIST:
            cal.adoptInsteadAndCheckErrorCode(new BuddhistCalendar(loc, status), status);
            break;
        case CALTYPE_ROC:
            cal.adoptInsteadAndCheckErrorCode(new TaiwanCalendar(loc, status), status);
            break;
        case CALTYPE_PERSIAN:
            cal.adoptInsteadAndCheckErrorCode(new PersianCalendar(loc, status), status);
            break;
        case CALTYPE_ISLAMIC_TBLA:
            cal.adoptInsteadAndCheckErrorCode(new IslamicCalendar(loc, status, IslamicCalendar::TBLA), status);
            break;
        case CALTYPE_ISLAMIC_CIVIL:
            cal.adoptInsteadAndCheckErrorCode(new IslamicCalendar(loc, status, IslamicCalendar::CIVIL), status);
            break;
        case CALTYPE_ISLAMIC_RGSA:
            // default any region specific not handled individually to islamic
        case CALTYPE_ISLAMIC:
            cal.adoptInsteadAndCheckErrorCode(new IslamicCalendar(loc, status, IslamicCalendar::ASTRONOMICAL), status);
            break;
        case CALTYPE_ISLAMIC_UMALQURA:
            cal.adoptInsteadAndCheckErrorCode(new IslamicCalendar(loc, status, IslamicCalendar::UMALQURA), status);
            break;
        case CALTYPE_HEBREW:
            cal.adoptInsteadAndCheckErrorCode(new HebrewCalendar(loc, status), status);
            break;
        case CALTYPE_CHINESE:
            cal.adoptInsteadAndCheckErrorCode(new ChineseCalendar(loc, status), status);
            break;
        case CALTYPE_INDIAN:
            cal.adoptInsteadAndCheckErrorCode(new IndianCalendar(loc, status), status);
            break;
        case CALTYPE_COPTIC:
            cal.adoptInsteadAndCheckErrorCode(new CopticCalendar(loc, status), status);
            break;
        case CALTYPE_ETHIOPIC:
            cal.adoptInsteadAndCheckErrorCode(new EthiopicCalendar(loc, status, EthiopicCalendar::AMETE_MIHRET_ERA), status);
            break;
        case CALTYPE_ETHIOPIC_AMETE_ALEM:
            cal.adoptInsteadAndCheckErrorCode(new EthiopicCalendar(loc, status, EthiopicCalendar::AMETE_ALEM_ERA), status);
            break;
        case CALTYPE_ISO8601:
            cal.adoptInsteadAndCheckErrorCode(new GregorianCalendar(loc, status), status);
            if (cal.isValid()) {
                cal->setFirstDayOfWeek(UCAL_MONDAY);
                cal->setMinimalDaysInFirstWeek(4);
            }
            break;
        case CALTYPE_DANGI:
            cal.adoptInsteadAndCheckErrorCode(new DangiCalendar(loc, status), status);
            break;
        default:
            status = U_UNSUPPORTED_ERROR;
    }
    return cal.orphan();
}


#if !UCONFIG_NO_SERVICE

// -------------------------------------

/**
* a Calendar Factory which creates the "basic" calendar types, that is, those
* shipped with ICU.
*/
class BasicCalendarFactory : public LocaleKeyFactory {
public:
    /**
    * @param calendarType static const string (caller owns storage - will be aliased) to calendar type
    */
    BasicCalendarFactory()
        : LocaleKeyFactory(LocaleKeyFactory::INVISIBLE) { }

    virtual ~BasicCalendarFactory();

protected:
    //virtual UBool isSupportedID( const UnicodeString& id, UErrorCode& status) const {
    //  if(U_FAILURE(status)) {
    //    return FALSE;
    //  }
    //  char keyword[ULOC_FULLNAME_CAPACITY];
    //  getCalendarKeyword(id, keyword, (int32_t)sizeof(keyword));
    //  return isStandardSupportedKeyword(keyword, status);
    //}

    virtual void updateVisibleIDs(Hashtable& result, UErrorCode& status) const
    {
        if (U_SUCCESS(status)) {
            for(int32_t i=0;gCalTypes[i] != NULL;i++) {
                UnicodeString id((UChar)0x40); /* '@' a variant character */
                id.append(UNICODE_STRING_SIMPLE("calendar="));
                id.append(UnicodeString(gCalTypes[i], -1, US_INV));
                result.put(id, (void*)this, status);
            }
        }
    }

    virtual UObject* create(const ICUServiceKey& key, const ICUService* /*service*/, UErrorCode& status) const {
#ifdef U_DEBUG_CALSVC
        if(dynamic_cast<const LocaleKey*>(&key) == NULL) {
            fprintf(stderr, "::create - not a LocaleKey!\n");
        }
#endif
        const LocaleKey& lkey = (LocaleKey&)key;
        Locale curLoc;  // current locale
        Locale canLoc;  // Canonical locale

        lkey.currentLocale(curLoc);
        lkey.canonicalLocale(canLoc);

        char keyword[ULOC_FULLNAME_CAPACITY];
        UnicodeString str;

        key.currentID(str);
        getCalendarKeyword(str, keyword, (int32_t) sizeof(keyword));

#ifdef U_DEBUG_CALSVC
        fprintf(stderr, "BasicCalendarFactory::create() - cur %s, can %s\n", (const char*)curLoc.getName(), (const char*)canLoc.getName());
#endif

        if(!isStandardSupportedKeyword(keyword,status)) {  // Do we handle this type?
#ifdef U_DEBUG_CALSVC

            fprintf(stderr, "BasicCalendarFactory - not handling %s.[%s]\n", (const char*) curLoc.getName(), tmp );
#endif
            return NULL;
        }

        return createStandardCalendar(getCalendarType(keyword), canLoc, status);
    }
};

BasicCalendarFactory::~BasicCalendarFactory() {}

/**
* A factory which looks up the DefaultCalendar resource to determine which class of calendar to use
*/

class DefaultCalendarFactory : public ICUResourceBundleFactory {
public:
    DefaultCalendarFactory() : ICUResourceBundleFactory() { }
    virtual ~DefaultCalendarFactory();
protected:
    virtual UObject* create(const ICUServiceKey& key, const ICUService* /*service*/, UErrorCode& status) const  {

        LocaleKey &lkey = (LocaleKey&)key;
        Locale loc;
        lkey.currentLocale(loc);

        UnicodeString *ret = new UnicodeString();
        if (ret == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
        } else {
            ret->append((UChar)0x40); // '@' is a variant character
            ret->append(UNICODE_STRING("calendar=", 9));
            ret->append(UnicodeString(gCalTypes[getCalendarTypeForLocale(loc.getName())], -1, US_INV));
        }
        return ret;
    }
};

DefaultCalendarFactory::~DefaultCalendarFactory() {}

// -------------------------------------
class CalendarService : public ICULocaleService {
public:
    CalendarService()
        : ICULocaleService(UNICODE_STRING_SIMPLE("Calendar"))
    {
        UErrorCode status = U_ZERO_ERROR;
        registerFactory(new DefaultCalendarFactory(), status);
    }

    virtual ~CalendarService();

    virtual UObject* cloneInstance(UObject* instance) const {
        UnicodeString *s = dynamic_cast<UnicodeString *>(instance);
        if(s != NULL) {
            return s->clone();
        } else {
#ifdef U_DEBUG_CALSVC_F
            UErrorCode status2 = U_ZERO_ERROR;
            fprintf(stderr, "Cloning a %s calendar with tz=%ld\n", ((Calendar*)instance)->getType(), ((Calendar*)instance)->get(UCAL_ZONE_OFFSET, status2));
#endif
            return ((Calendar*)instance)->clone();
        }
    }

    virtual UObject* handleDefault(const ICUServiceKey& key, UnicodeString* /*actualID*/, UErrorCode& status) const {
        LocaleKey& lkey = (LocaleKey&)key;
        //int32_t kind = lkey.kind();

        Locale loc;
        lkey.canonicalLocale(loc);

#ifdef U_DEBUG_CALSVC
        Locale loc2;
        lkey.currentLocale(loc2);
        fprintf(stderr, "CalSvc:handleDefault for currentLoc %s, canloc %s\n", (const char*)loc.getName(),  (const char*)loc2.getName());
#endif
        Calendar *nc =  new GregorianCalendar(loc, status);
        if (nc == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return nc;
        }

#ifdef U_DEBUG_CALSVC
        UErrorCode status2 = U_ZERO_ERROR;
        fprintf(stderr, "New default calendar has tz=%d\n", ((Calendar*)nc)->get(UCAL_ZONE_OFFSET, status2));
#endif
        return nc;
    }

    virtual UBool isDefault() const {
        return countFactories() == 1;
    }
};

CalendarService::~CalendarService() {}

// -------------------------------------

static inline UBool
isCalendarServiceUsed() {
    return !gServiceInitOnce.isReset();
}

// -------------------------------------

static void U_CALLCONV
initCalendarService(UErrorCode &status)
{
#ifdef U_DEBUG_CALSVC
        fprintf(stderr, "Spinning up Calendar Service\n");
#endif
    ucln_i18n_registerCleanup(UCLN_I18N_CALENDAR, calendar_cleanup);
    gService = new CalendarService();
    if (gService == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
        return;
        }
#ifdef U_DEBUG_CALSVC
        fprintf(stderr, "Registering classes..\n");
#endif

        // Register all basic instances.
    gService->registerFactory(new BasicCalendarFactory(),status);

#ifdef U_DEBUG_CALSVC
        fprintf(stderr, "Done..\n");
#endif

        if(U_FAILURE(status)) {
#ifdef U_DEBUG_CALSVC
            fprintf(stderr, "err (%s) registering classes, deleting service.....\n", u_errorName(status));
#endif
        delete gService;
        gService = NULL;
    }
        }

static ICULocaleService*
getCalendarService(UErrorCode &status)
{
    umtx_initOnce(gServiceInitOnce, &initCalendarService, status);
    return gService;
}

URegistryKey Calendar::registerFactory(ICUServiceFactory* toAdopt, UErrorCode& status)
{
    return getCalendarService(status)->registerFactory(toAdopt, status);
}

UBool Calendar::unregister(URegistryKey key, UErrorCode& status) {
    return getCalendarService(status)->unregister(key, status);
}
#endif /* UCONFIG_NO_SERVICE */

// -------------------------------------

static const int32_t kCalendarLimits[UCAL_FIELD_COUNT][4] = {
    //    Minimum  Greatest min      Least max   Greatest max
    {/*N/A*/-1,       /*N/A*/-1,     /*N/A*/-1,       /*N/A*/-1}, // ERA
    {/*N/A*/-1,       /*N/A*/-1,     /*N/A*/-1,       /*N/A*/-1}, // YEAR
    {/*N/A*/-1,       /*N/A*/-1,     /*N/A*/-1,       /*N/A*/-1}, // MONTH
    {/*N/A*/-1,       /*N/A*/-1,     /*N/A*/-1,       /*N/A*/-1}, // WEEK_OF_YEAR
    {/*N/A*/-1,       /*N/A*/-1,     /*N/A*/-1,       /*N/A*/-1}, // WEEK_OF_MONTH
    {/*N/A*/-1,       /*N/A*/-1,     /*N/A*/-1,       /*N/A*/-1}, // DAY_OF_MONTH
    {/*N/A*/-1,       /*N/A*/-1,     /*N/A*/-1,       /*N/A*/-1}, // DAY_OF_YEAR
    {           1,            1,             7,             7  }, // DAY_OF_WEEK
    {/*N/A*/-1,       /*N/A*/-1,     /*N/A*/-1,       /*N/A*/-1}, // DAY_OF_WEEK_IN_MONTH
    {           0,            0,             1,             1  }, // AM_PM
    {           0,            0,            11,            11  }, // HOUR
    {           0,            0,            23,            23  }, // HOUR_OF_DAY
    {           0,            0,            59,            59  }, // MINUTE
    {           0,            0,            59,            59  }, // SECOND
    {           0,            0,           999,           999  }, // MILLISECOND
    {-12*kOneHour, -12*kOneHour,   12*kOneHour,   15*kOneHour  }, // ZONE_OFFSET
    {           0,            0,    1*kOneHour,    1*kOneHour  }, // DST_OFFSET
    {/*N/A*/-1,       /*N/A*/-1,     /*N/A*/-1,       /*N/A*/-1}, // YEAR_WOY
    {           1,            1,             7,             7  }, // DOW_LOCAL
    {/*N/A*/-1,       /*N/A*/-1,     /*N/A*/-1,       /*N/A*/-1}, // EXTENDED_YEAR
    { -0x7F000000,  -0x7F000000,    0x7F000000,    0x7F000000  }, // JULIAN_DAY
    {           0,            0, 24*kOneHour-1, 24*kOneHour-1  }, // MILLISECONDS_IN_DAY
    {           0,            0,             1,             1  }, // IS_LEAP_MONTH
};

// Resource bundle tags read by this class
static const char gCalendar[] = "calendar";
static const char gMonthNames[] = "monthNames";
static const char gGregorian[] = "gregorian";

// Data flow in Calendar
// ---------------------

// The current time is represented in two ways by Calendar: as UTC
// milliseconds from the epoch start (1 January 1970 0:00 UTC), and as local
// fields such as MONTH, HOUR, AM_PM, etc.  It is possible to compute the
// millis from the fields, and vice versa.  The data needed to do this
// conversion is encapsulated by a TimeZone object owned by the Calendar.
// The data provided by the TimeZone object may also be overridden if the
// user sets the ZONE_OFFSET and/or DST_OFFSET fields directly. The class
// keeps track of what information was most recently set by the caller, and
// uses that to compute any other information as needed.

// If the user sets the fields using set(), the data flow is as follows.
// This is implemented by the Calendar subclass's computeTime() method.
// During this process, certain fields may be ignored.  The disambiguation
// algorithm for resolving which fields to pay attention to is described
// above.

//   local fields (YEAR, MONTH, DATE, HOUR, MINUTE, etc.)
//           |
//           | Using Calendar-specific algorithm
//           V
//   local standard millis
//           |
//           | Using TimeZone or user-set ZONE_OFFSET / DST_OFFSET
//           V
//   UTC millis (in time data member)

// If the user sets the UTC millis using setTime(), the data flow is as
// follows.  This is implemented by the Calendar subclass's computeFields()
// method.

//   UTC millis (in time data member)
//           |
//           | Using TimeZone getOffset()
//           V
//   local standard millis
//           |
//           | Using Calendar-specific algorithm
//           V
//   local fields (YEAR, MONTH, DATE, HOUR, MINUTE, etc.)

// In general, a round trip from fields, through local and UTC millis, and
// back out to fields is made when necessary.  This is implemented by the
// complete() method.  Resolving a partial set of fields into a UTC millis
// value allows all remaining fields to be generated from that value.  If
// the Calendar is lenient, the fields are also renormalized to standard
// ranges when they are regenerated.

// -------------------------------------

Calendar::Calendar(UErrorCode& success)
:   UObject(),
fIsTimeSet(FALSE),
fAreFieldsSet(FALSE),
fAreAllFieldsSet(FALSE),
fAreFieldsVirtuallySet(FALSE),
fNextStamp((int32_t)kMinimumUserStamp),
fTime(0),
fLenient(TRUE),
fZone(NULL),
fRepeatedWallTime(UCAL_WALLTIME_LAST),
fSkippedWallTime(UCAL_WALLTIME_LAST)
{
    validLocale[0] = 0;
    actualLocale[0] = 0;
    clear();
    if (U_FAILURE(success)) {
        return;
    }
    fZone = TimeZone::createDefault();
    if (fZone == NULL) {
        success = U_MEMORY_ALLOCATION_ERROR;
    }
    setWeekData(Locale::getDefault(), NULL, success);
}

// -------------------------------------

Calendar::Calendar(TimeZone* zone, const Locale& aLocale, UErrorCode& success)
:   UObject(),
fIsTimeSet(FALSE),
fAreFieldsSet(FALSE),
fAreAllFieldsSet(FALSE),
fAreFieldsVirtuallySet(FALSE),
fNextStamp((int32_t)kMinimumUserStamp),
fTime(0),
fLenient(TRUE),
fZone(NULL),
fRepeatedWallTime(UCAL_WALLTIME_LAST),
fSkippedWallTime(UCAL_WALLTIME_LAST)
{
    validLocale[0] = 0;
    actualLocale[0] = 0;
    if (U_FAILURE(success)) {
        delete zone;
        return;
    }
    if(zone == 0) {
#if defined (U_DEBUG_CAL)
        fprintf(stderr, "%s:%d: ILLEGAL ARG because timezone cannot be 0\n",
            __FILE__, __LINE__);
#endif
        success = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    clear();
    fZone = zone;
    setWeekData(aLocale, NULL, success);
}

// -------------------------------------

Calendar::Calendar(const TimeZone& zone, const Locale& aLocale, UErrorCode& success)
:   UObject(),
fIsTimeSet(FALSE),
fAreFieldsSet(FALSE),
fAreAllFieldsSet(FALSE),
fAreFieldsVirtuallySet(FALSE),
fNextStamp((int32_t)kMinimumUserStamp),
fTime(0),
fLenient(TRUE),
fZone(NULL),
fRepeatedWallTime(UCAL_WALLTIME_LAST),
fSkippedWallTime(UCAL_WALLTIME_LAST)
{
    validLocale[0] = 0;
    actualLocale[0] = 0;
    if (U_FAILURE(success)) {
        return;
    }
    clear();
    fZone = zone.clone();
    if (fZone == NULL) {
        success = U_MEMORY_ALLOCATION_ERROR;
    }
    setWeekData(aLocale, NULL, success);
}

// -------------------------------------

Calendar::~Calendar()
{
    delete fZone;
}

// -------------------------------------

Calendar::Calendar(const Calendar &source)
:   UObject(source)
{
    fZone = NULL;
    *this = source;
}

// -------------------------------------

Calendar &
Calendar::operator=(const Calendar &right)
{
    if (this != &right) {
        uprv_arrayCopy(right.fFields, fFields, UCAL_FIELD_COUNT);
        uprv_arrayCopy(right.fIsSet, fIsSet, UCAL_FIELD_COUNT);
        uprv_arrayCopy(right.fStamp, fStamp, UCAL_FIELD_COUNT);
        fTime                    = right.fTime;
        fIsTimeSet               = right.fIsTimeSet;
        fAreAllFieldsSet         = right.fAreAllFieldsSet;
        fAreFieldsSet            = right.fAreFieldsSet;
        fAreFieldsVirtuallySet   = right.fAreFieldsVirtuallySet;
        fLenient                 = right.fLenient;
        fRepeatedWallTime        = right.fRepeatedWallTime;
        fSkippedWallTime         = right.fSkippedWallTime;
        delete fZone;
        fZone = NULL;
        if (right.fZone != NULL) {
            fZone                = right.fZone->clone();
        }
        fFirstDayOfWeek          = right.fFirstDayOfWeek;
        fMinimalDaysInFirstWeek  = right.fMinimalDaysInFirstWeek;
        fWeekendOnset            = right.fWeekendOnset;
        fWeekendOnsetMillis      = right.fWeekendOnsetMillis;
        fWeekendCease            = right.fWeekendCease;
        fWeekendCeaseMillis      = right.fWeekendCeaseMillis;
        fNextStamp               = right.fNextStamp;
        uprv_strncpy(validLocale, right.validLocale, sizeof(validLocale));
        uprv_strncpy(actualLocale, right.actualLocale, sizeof(actualLocale));
        validLocale[sizeof(validLocale)-1] = 0;
        actualLocale[sizeof(validLocale)-1] = 0;
    }

    return *this;
}

// -------------------------------------

Calendar* U_EXPORT2
Calendar::createInstance(UErrorCode& success)
{
    return createInstance(TimeZone::createDefault(), Locale::getDefault(), success);
}

// -------------------------------------

Calendar* U_EXPORT2
Calendar::createInstance(const TimeZone& zone, UErrorCode& success)
{
    return createInstance(zone, Locale::getDefault(), success);
}

// -------------------------------------

Calendar* U_EXPORT2
Calendar::createInstance(const Locale& aLocale, UErrorCode& success)
{
    return createInstance(TimeZone::createDefault(), aLocale, success);
}

// ------------------------------------- Adopting

// Note: this is the bottleneck that actually calls the service routines.

Calendar * U_EXPORT2
Calendar::makeInstance(const Locale& aLocale, UErrorCode& success) {
    if (U_FAILURE(success)) {
        return NULL;
    }

    Locale actualLoc;
    UObject* u = NULL;

#if !UCONFIG_NO_SERVICE
    if (isCalendarServiceUsed()) {
        u = getCalendarService(success)->get(aLocale, LocaleKey::KIND_ANY, &actualLoc, success);
    }
    else
#endif
    {
        u = createStandardCalendar(getCalendarTypeForLocale(aLocale.getName()), aLocale, success);
    }
    Calendar* c = NULL;

    if(U_FAILURE(success) || !u) {
        if(U_SUCCESS(success)) { // Propagate some kind of err
            success = U_INTERNAL_PROGRAM_ERROR;
        }
        return NULL;
    }

#if !UCONFIG_NO_SERVICE
    const UnicodeString* str = dynamic_cast<const UnicodeString*>(u);
    if(str != NULL) {
        // It's a unicode string telling us what type of calendar to load ("gregorian", etc)
        // Create a Locale over this string
        Locale l("");
        LocaleUtility::initLocaleFromName(*str, l);

#ifdef U_DEBUG_CALSVC
        fprintf(stderr, "Calendar::createInstance(%s), looking up [%s]\n", aLocale.getName(), l.getName());
#endif

        Locale actualLoc2;
        delete u;
        u = NULL;

        // Don't overwrite actualLoc, since the actual loc from this call
        // may be something like "@calendar=gregorian" -- TODO investigate
        // further...
        c = (Calendar*)getCalendarService(success)->get(l, LocaleKey::KIND_ANY, &actualLoc2, success);

        if(U_FAILURE(success) || !c) {
            if(U_SUCCESS(success)) {
                success = U_INTERNAL_PROGRAM_ERROR; // Propagate some err
            }
            return NULL;
        }

        str = dynamic_cast<const UnicodeString*>(c);
        if(str != NULL) {
            // recursed! Second lookup returned a UnicodeString.
            // Perhaps DefaultCalendar{} was set to another locale.
#ifdef U_DEBUG_CALSVC
            char tmp[200];
            // Extract a char* out of it..
            int32_t len = str->length();
            int32_t actLen = sizeof(tmp)-1;
            if(len > actLen) {
                len = actLen;
            }
            str->extract(0,len,tmp);
            tmp[len]=0;

            fprintf(stderr, "err - recursed, 2nd lookup was unistring %s\n", tmp);
#endif
            success = U_MISSING_RESOURCE_ERROR;  // requested a calendar type which could NOT be found.
            delete c;
            return NULL;
        }
#ifdef U_DEBUG_CALSVC
        fprintf(stderr, "%p: setting week count data to locale %s, actual locale %s\n", c, (const char*)aLocale.getName(), (const char *)actualLoc.getName());
#endif
        c->setWeekData(aLocale, c->getType(), success);  // set the correct locale (this was an indirected calendar)

        char keyword[ULOC_FULLNAME_CAPACITY];
        UErrorCode tmpStatus = U_ZERO_ERROR;
        l.getKeywordValue("calendar", keyword, ULOC_FULLNAME_CAPACITY, tmpStatus);
        if (U_SUCCESS(tmpStatus) && uprv_strcmp(keyword, "iso8601") == 0) {
            c->setFirstDayOfWeek(UCAL_MONDAY);
            c->setMinimalDaysInFirstWeek(4);
        }
    }
    else
#endif /* UCONFIG_NO_SERVICE */
    {
        // a calendar was returned - we assume the factory did the right thing.
        c = (Calendar*)u;
    }

    return c;
}

Calendar* U_EXPORT2
Calendar::createInstance(TimeZone* zone, const Locale& aLocale, UErrorCode& success)
{
    LocalPointer<TimeZone> zonePtr(zone);
    const SharedCalendar *shared = NULL;
    UnifiedCache::getByLocale(aLocale, shared, success);
    if (U_FAILURE(success)) {
        return NULL;
    }
    Calendar *c = (*shared)->clone();
    shared->removeRef();
    if (c == NULL) {
        success = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    // Now, reset calendar to default state:
    c->adoptTimeZone(zonePtr.orphan()); //  Set the correct time zone
    c->setTimeInMillis(getNow(), success); // let the new calendar have the current time.

    return c;
}

// -------------------------------------

Calendar* U_EXPORT2
Calendar::createInstance(const TimeZone& zone, const Locale& aLocale, UErrorCode& success)
{
    Calendar* c = createInstance(aLocale, success);
    if(U_SUCCESS(success) && c) {
        c->setTimeZone(zone);
    }
    return c;
}

// -------------------------------------

void U_EXPORT2
Calendar::getCalendarTypeFromLocale(
        const Locale &aLocale,
        char *typeBuffer,
        int32_t typeBufferSize,
        UErrorCode &success) {
    const SharedCalendar *shared = NULL;
    UnifiedCache::getByLocale(aLocale, shared, success);
    if (U_FAILURE(success)) {
        return;
    }
    uprv_strncpy(typeBuffer, (*shared)->getType(), typeBufferSize);
    shared->removeRef();
    if (typeBuffer[typeBufferSize - 1]) {
        success = U_BUFFER_OVERFLOW_ERROR;
    }
}

UBool
Calendar::operator==(const Calendar& that) const
{
    UErrorCode status = U_ZERO_ERROR;
    return isEquivalentTo(that) &&
        getTimeInMillis(status) == that.getTimeInMillis(status) &&
        U_SUCCESS(status);
}

UBool
Calendar::isEquivalentTo(const Calendar& other) const
{
    return typeid(*this) == typeid(other) &&
        fLenient                == other.fLenient &&
        fRepeatedWallTime       == other.fRepeatedWallTime &&
        fSkippedWallTime        == other.fSkippedWallTime &&
        fFirstDayOfWeek         == other.fFirstDayOfWeek &&
        fMinimalDaysInFirstWeek == other.fMinimalDaysInFirstWeek &&
        fWeekendOnset           == other.fWeekendOnset &&
        fWeekendOnsetMillis     == other.fWeekendOnsetMillis &&
        fWeekendCease           == other.fWeekendCease &&
        fWeekendCeaseMillis     == other.fWeekendCeaseMillis &&
        *fZone                  == *other.fZone;
}

// -------------------------------------

UBool
Calendar::equals(const Calendar& when, UErrorCode& status) const
{
    return (this == &when ||
        getTime(status) == when.getTime(status));
}

// -------------------------------------

UBool
Calendar::before(const Calendar& when, UErrorCode& status) const
{
    return (this != &when &&
        getTimeInMillis(status) < when.getTimeInMillis(status));
}

// -------------------------------------

UBool
Calendar::after(const Calendar& when, UErrorCode& status) const
{
    return (this != &when &&
        getTimeInMillis(status) > when.getTimeInMillis(status));
}

// -------------------------------------


const Locale* U_EXPORT2
Calendar::getAvailableLocales(int32_t& count)
{
    return Locale::getAvailableLocales(count);
}

// -------------------------------------

StringEnumeration* U_EXPORT2
Calendar::getKeywordValuesForLocale(const char* key,
                    const Locale& locale, UBool commonlyUsed, UErrorCode& status)
{
    // This is a wrapper over ucal_getKeywordValuesForLocale
    UEnumeration *uenum = ucal_getKeywordValuesForLocale(key, locale.getName(),
                                                        commonlyUsed, &status);
    if (U_FAILURE(status)) {
        uenum_close(uenum);
        return NULL;
    }
    UStringEnumeration* ustringenum = new UStringEnumeration(uenum);
    if (ustringenum == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return ustringenum;
}

// -------------------------------------

UDate U_EXPORT2
Calendar::getNow()
{
    return uprv_getUTCtime(); // return as milliseconds
}

// -------------------------------------

/**
* Gets this Calendar's current time as a long.
* @return the current time as UTC milliseconds from the epoch.
*/
double
Calendar::getTimeInMillis(UErrorCode& status) const
{
    if(U_FAILURE(status))
        return 0.0;

    if ( ! fIsTimeSet)
        ((Calendar*)this)->updateTime(status);

    /* Test for buffer overflows */
    if(U_FAILURE(status)) {
        return 0.0;
    }
    return fTime;
}

// -------------------------------------

/**
* Sets this Calendar's current time from the given long value.
* A status of U_ILLEGAL_ARGUMENT_ERROR is set when millis is
* outside the range permitted by a Calendar object when not in lenient mode.
* when in lenient mode the out of range values are pinned to their respective min/max.
* @param date the new time in UTC milliseconds from the epoch.
*/
void
Calendar::setTimeInMillis( double millis, UErrorCode& status ) {
    if(U_FAILURE(status))
        return;

    if (millis > MAX_MILLIS) {
        if(isLenient()) {
            millis = MAX_MILLIS;
        } else {
		    status = U_ILLEGAL_ARGUMENT_ERROR;
		    return;
        }
    } else if (millis < MIN_MILLIS) {
        if(isLenient()) {
            millis = MIN_MILLIS;
        } else {
    		status = U_ILLEGAL_ARGUMENT_ERROR;
	    	return;
        }
    }

    fTime = millis;
    fAreFieldsSet = fAreAllFieldsSet = FALSE;
    fIsTimeSet = fAreFieldsVirtuallySet = TRUE;

    for (int32_t i=0; i<UCAL_FIELD_COUNT; ++i) {
        fFields[i]     = 0;
        fStamp[i]     = kUnset;
        fIsSet[i]     = FALSE;
    }


}

// -------------------------------------

int32_t
Calendar::get(UCalendarDateFields field, UErrorCode& status) const
{
    // field values are only computed when actually requested; for more on when computation
    // of various things happens, see the "data flow in Calendar" description at the top
    // of this file
    if (U_SUCCESS(status)) ((Calendar*)this)->complete(status); // Cast away const
    return U_SUCCESS(status) ? fFields[field] : 0;
}

// -------------------------------------

void
Calendar::set(UCalendarDateFields field, int32_t value)
{
    if (fAreFieldsVirtuallySet) {
        UErrorCode ec = U_ZERO_ERROR;
        computeFields(ec);
    }
    fFields[field]     = value;
    /* Ensure that the fNextStamp value doesn't go pass max value for int32_t */
    if (fNextStamp == STAMP_MAX) {
        recalculateStamp();
    }
    fStamp[field]     = fNextStamp++;
    fIsSet[field]     = TRUE; // Remove later
    fIsTimeSet = fAreFieldsSet = fAreFieldsVirtuallySet = FALSE;
}

// -------------------------------------

void
Calendar::set(int32_t year, int32_t month, int32_t date)
{
    set(UCAL_YEAR, year);
    set(UCAL_MONTH, month);
    set(UCAL_DATE, date);
}

// -------------------------------------

void
Calendar::set(int32_t year, int32_t month, int32_t date, int32_t hour, int32_t minute)
{
    set(UCAL_YEAR, year);
    set(UCAL_MONTH, month);
    set(UCAL_DATE, date);
    set(UCAL_HOUR_OF_DAY, hour);
    set(UCAL_MINUTE, minute);
}

// -------------------------------------

void
Calendar::set(int32_t year, int32_t month, int32_t date, int32_t hour, int32_t minute, int32_t second)
{
    set(UCAL_YEAR, year);
    set(UCAL_MONTH, month);
    set(UCAL_DATE, date);
    set(UCAL_HOUR_OF_DAY, hour);
    set(UCAL_MINUTE, minute);
    set(UCAL_SECOND, second);
}

// -------------------------------------
// For now the full getRelatedYear implementation is here;
// per #10752 move the non-default implementation to subclasses
// (default implementation will do no year adjustment)

static int32_t gregoYearFromIslamicStart(int32_t year) {
    // ad hoc conversion, improve under #10752
    // rough est for now, ok for grego 1846-2138,
    // otherwise occasionally wrong (for 3% of years)
    int cycle, offset, shift = 0;
    if (year >= 1397) {
        cycle = (year - 1397) / 67;
        offset = (year - 1397) % 67;
        shift = 2*cycle + ((offset >= 33)? 1: 0);
    } else {
        cycle = (year - 1396) / 67 - 1;
        offset = -(year - 1396) % 67;
        shift = 2*cycle + ((offset <= 33)? 1: 0);
    }
    return year + 579 - shift;
}

int32_t Calendar::getRelatedYear(UErrorCode &status) const
{
    if (U_FAILURE(status)) {
        return 0;
    }
    int32_t year = get(UCAL_EXTENDED_YEAR, status);
    if (U_FAILURE(status)) {
        return 0;
    }
    // modify for calendar type
    ECalType type = getCalendarType(getType());
    switch (type) {
        case CALTYPE_PERSIAN:
            year += 622; break;
        case CALTYPE_HEBREW:
            year -= 3760; break;
        case CALTYPE_CHINESE:
            year -= 2637; break;
        case CALTYPE_INDIAN:
            year += 79; break;
        case CALTYPE_COPTIC:
            year += 284; break;
        case CALTYPE_ETHIOPIC:
            year += 8; break;
        case CALTYPE_ETHIOPIC_AMETE_ALEM:
            year -=5492; break;
        case CALTYPE_DANGI:
            year -= 2333; break;
        case CALTYPE_ISLAMIC_CIVIL:
        case CALTYPE_ISLAMIC:
        case CALTYPE_ISLAMIC_UMALQURA:
        case CALTYPE_ISLAMIC_TBLA:
        case CALTYPE_ISLAMIC_RGSA:
            year = gregoYearFromIslamicStart(year); break;
        default:
            // CALTYPE_GREGORIAN
            // CALTYPE_JAPANESE
            // CALTYPE_BUDDHIST
            // CALTYPE_ROC
            // CALTYPE_ISO8601
            // do nothing, EXTENDED_YEAR same as Gregorian
            break;
    }
    return year;
}

// -------------------------------------
// For now the full setRelatedYear implementation is here;
// per #10752 move the non-default implementation to subclasses
// (default implementation will do no year adjustment)

static int32_t firstIslamicStartYearFromGrego(int32_t year) {
    // ad hoc conversion, improve under #10752
    // rough est for now, ok for grego 1846-2138,
    // otherwise occasionally wrong (for 3% of years)
    int cycle, offset, shift = 0;
    if (year >= 1977) {
        cycle = (year - 1977) / 65;
        offset = (year - 1977) % 65;
        shift = 2*cycle + ((offset >= 32)? 1: 0);
    } else {
        cycle = (year - 1976) / 65 - 1;
        offset = -(year - 1976) % 65;
        shift = 2*cycle + ((offset <= 32)? 1: 0);
    }
    return year - 579 + shift;
}
void Calendar::setRelatedYear(int32_t year)
{
    // modify for calendar type
    ECalType type = getCalendarType(getType());
    switch (type) {
        case CALTYPE_PERSIAN:
            year -= 622; break;
        case CALTYPE_HEBREW:
            year += 3760; break;
        case CALTYPE_CHINESE:
            year += 2637; break;
        case CALTYPE_INDIAN:
            year -= 79; break;
        case CALTYPE_COPTIC:
            year -= 284; break;
        case CALTYPE_ETHIOPIC:
            year -= 8; break;
        case CALTYPE_ETHIOPIC_AMETE_ALEM:
            year +=5492; break;
        case CALTYPE_DANGI:
            year += 2333; break;
        case CALTYPE_ISLAMIC_CIVIL:
        case CALTYPE_ISLAMIC:
        case CALTYPE_ISLAMIC_UMALQURA:
        case CALTYPE_ISLAMIC_TBLA:
        case CALTYPE_ISLAMIC_RGSA:
            year = firstIslamicStartYearFromGrego(year); break;
        default:
            // CALTYPE_GREGORIAN
            // CALTYPE_JAPANESE
            // CALTYPE_BUDDHIST
            // CALTYPE_ROC
            // CALTYPE_ISO8601
            // do nothing, EXTENDED_YEAR same as Gregorian
            break;
    }
    // set extended year
    set(UCAL_EXTENDED_YEAR, year);
}

// -------------------------------------

void
Calendar::clear()
{
    for (int32_t i=0; i<UCAL_FIELD_COUNT; ++i) {
        fFields[i]     = 0; // Must do this; other code depends on it
        fStamp[i]     = kUnset;
        fIsSet[i]     = FALSE; // Remove later
    }
    fIsTimeSet = fAreFieldsSet = fAreAllFieldsSet = fAreFieldsVirtuallySet = FALSE;
    // fTime is not 'cleared' - may be used if no fields are set.
}

// -------------------------------------

void
Calendar::clear(UCalendarDateFields field)
{
    if (fAreFieldsVirtuallySet) {
        UErrorCode ec = U_ZERO_ERROR;
        computeFields(ec);
    }
    fFields[field]         = 0;
    fStamp[field]         = kUnset;
    fIsSet[field]         = FALSE; // Remove later
    fIsTimeSet = fAreFieldsSet = fAreAllFieldsSet = fAreFieldsVirtuallySet = FALSE;
}

// -------------------------------------

UBool
Calendar::isSet(UCalendarDateFields field) const
{
    return fAreFieldsVirtuallySet || (fStamp[field] != kUnset);
}


int32_t Calendar::newestStamp(UCalendarDateFields first, UCalendarDateFields last, int32_t bestStampSoFar) const
{
    int32_t bestStamp = bestStampSoFar;
    for (int32_t i=(int32_t)first; i<=(int32_t)last; ++i) {
        if (fStamp[i] > bestStamp) {
            bestStamp = fStamp[i];
        }
    }
    return bestStamp;
}


// -------------------------------------

void
Calendar::complete(UErrorCode& status)
{
    if (!fIsTimeSet) {
        updateTime(status);
        /* Test for buffer overflows */
        if(U_FAILURE(status)) {
            return;
        }
    }
    if (!fAreFieldsSet) {
        computeFields(status); // fills in unset fields
        /* Test for buffer overflows */
        if(U_FAILURE(status)) {
            return;
        }
        fAreFieldsSet         = TRUE;
        fAreAllFieldsSet     = TRUE;
    }
}

//-------------------------------------------------------------------------
// Protected utility methods for use by subclasses.  These are very handy
// for implementing add, roll, and computeFields.
//-------------------------------------------------------------------------

/**
* Adjust the specified field so that it is within
* the allowable range for the date to which this calendar is set.
* For example, in a Gregorian calendar pinning the {@link #DAY_OF_MONTH DAY_OF_MONTH}
* field for a calendar set to April 31 would cause it to be set
* to April 30.
* <p>
* <b>Subclassing:</b>
* <br>
* This utility method is intended for use by subclasses that need to implement
* their own overrides of {@link #roll roll} and {@link #add add}.
* <p>
* <b>Note:</b>
* <code>pinField</code> is implemented in terms of
* {@link #getActualMinimum getActualMinimum}
* and {@link #getActualMaximum getActualMaximum}.  If either of those methods uses
* a slow, iterative algorithm for a particular field, it would be
* unwise to attempt to call <code>pinField</code> for that field.  If you
* really do need to do so, you should override this method to do
* something more efficient for that field.
* <p>
* @param field The calendar field whose value should be pinned.
*
* @see #getActualMinimum
* @see #getActualMaximum
* @stable ICU 2.0
*/
void Calendar::pinField(UCalendarDateFields field, UErrorCode& status) {
    int32_t max = getActualMaximum(field, status);
    int32_t min = getActualMinimum(field, status);

    if (fFields[field] > max) {
        set(field, max);
    } else if (fFields[field] < min) {
        set(field, min);
    }
}


void Calendar::computeFields(UErrorCode &ec)
{
  if (U_FAILURE(ec)) {
        return;
    }
    // Compute local wall millis
    double localMillis = internalGetTime();
    int32_t rawOffset, dstOffset;
    getTimeZone().getOffset(localMillis, FALSE, rawOffset, dstOffset, ec);
    localMillis += (rawOffset + dstOffset);

    // Mark fields as set.  Do this before calling handleComputeFields().
    uint32_t mask =   //fInternalSetMask;
        (1 << UCAL_ERA) |
        (1 << UCAL_YEAR) |
        (1 << UCAL_MONTH) |
        (1 << UCAL_DAY_OF_MONTH) | // = UCAL_DATE
        (1 << UCAL_DAY_OF_YEAR) |
        (1 << UCAL_EXTENDED_YEAR);

    for (int32_t i=0; i<UCAL_FIELD_COUNT; ++i) {
        if ((mask & 1) == 0) {
            fStamp[i] = kInternallySet;
            fIsSet[i] = TRUE; // Remove later
        } else {
            fStamp[i] = kUnset;
            fIsSet[i] = FALSE; // Remove later
        }
        mask >>= 1;
    }

    // We used to check for and correct extreme millis values (near
    // Long.MIN_VALUE or Long.MAX_VALUE) here.  Such values would cause
    // overflows from positive to negative (or vice versa) and had to
    // be manually tweaked.  We no longer need to do this because we
    // have limited the range of supported dates to those that have a
    // Julian day that fits into an int.  This allows us to implement a
    // JULIAN_DAY field and also removes some inelegant code. - Liu
    // 11/6/00

    int32_t days =  (int32_t)ClockMath::floorDivide(localMillis, (double)kOneDay);

    internalSet(UCAL_JULIAN_DAY,days + kEpochStartAsJulianDay);

#if defined (U_DEBUG_CAL)
    //fprintf(stderr, "%s:%d- Hmm! Jules @ %d, as per %.0lf millis\n",
    //__FILE__, __LINE__, fFields[UCAL_JULIAN_DAY], localMillis);
#endif

    computeGregorianAndDOWFields(fFields[UCAL_JULIAN_DAY], ec);

    // Call framework method to have subclass compute its fields.
    // These must include, at a minimum, MONTH, DAY_OF_MONTH,
    // EXTENDED_YEAR, YEAR, DAY_OF_YEAR.  This method will call internalSet(),
    // which will update stamp[].
    handleComputeFields(fFields[UCAL_JULIAN_DAY], ec);

    // Compute week-related fields, based on the subclass-computed
    // fields computed by handleComputeFields().
    computeWeekFields(ec);

    // Compute time-related fields.  These are indepent of the date and
    // of the subclass algorithm.  They depend only on the local zone
    // wall milliseconds in day.
    int32_t millisInDay =  (int32_t) (localMillis - (days * kOneDay));
    fFields[UCAL_MILLISECONDS_IN_DAY] = millisInDay;
    fFields[UCAL_MILLISECOND] = millisInDay % 1000;
    millisInDay /= 1000;
    fFields[UCAL_SECOND] = millisInDay % 60;
    millisInDay /= 60;
    fFields[UCAL_MINUTE] = millisInDay % 60;
    millisInDay /= 60;
    fFields[UCAL_HOUR_OF_DAY] = millisInDay;
    fFields[UCAL_AM_PM] = millisInDay / 12; // Assume AM == 0
    fFields[UCAL_HOUR] = millisInDay % 12;
    fFields[UCAL_ZONE_OFFSET] = rawOffset;
    fFields[UCAL_DST_OFFSET] = dstOffset;
}

uint8_t Calendar::julianDayToDayOfWeek(double julian)
{
    // If julian is negative, then julian%7 will be negative, so we adjust
    // accordingly.  We add 1 because Julian day 0 is Monday.
    int8_t dayOfWeek = (int8_t) uprv_fmod(julian + 1, 7);

    uint8_t result = (uint8_t)(dayOfWeek + ((dayOfWeek < 0) ? (7+UCAL_SUNDAY ) : UCAL_SUNDAY));
    return result;
}

/**
* Compute the Gregorian calendar year, month, and day of month from
* the given Julian day.  These values are not stored in fields, but in
* member variables gregorianXxx.  Also compute the DAY_OF_WEEK and
* DOW_LOCAL fields.
*/
void Calendar::computeGregorianAndDOWFields(int32_t julianDay, UErrorCode &ec)
{
    computeGregorianFields(julianDay, ec);

    // Compute day of week: JD 0 = Monday
    int32_t dow = julianDayToDayOfWeek(julianDay);
    internalSet(UCAL_DAY_OF_WEEK,dow);

    // Calculate 1-based localized day of week
    int32_t dowLocal = dow - getFirstDayOfWeek() + 1;
    if (dowLocal < 1) {
        dowLocal += 7;
    }
    internalSet(UCAL_DOW_LOCAL,dowLocal);
    fFields[UCAL_DOW_LOCAL] = dowLocal;
}

/**
* Compute the Gregorian calendar year, month, and day of month from the
* Julian day.  These values are not stored in fields, but in member
* variables gregorianXxx.  They are used for time zone computations and by
* subclasses that are Gregorian derivatives.  Subclasses may call this
* method to perform a Gregorian calendar millis->fields computation.
*/
void Calendar::computeGregorianFields(int32_t julianDay, UErrorCode & /* ec */) {
    int32_t gregorianDayOfWeekUnused;
    Grego::dayToFields(julianDay - kEpochStartAsJulianDay, fGregorianYear, fGregorianMonth, fGregorianDayOfMonth, gregorianDayOfWeekUnused, fGregorianDayOfYear);
}

/**
* Compute the fields WEEK_OF_YEAR, YEAR_WOY, WEEK_OF_MONTH,
* DAY_OF_WEEK_IN_MONTH, and DOW_LOCAL from EXTENDED_YEAR, YEAR,
* DAY_OF_WEEK, and DAY_OF_YEAR.  The latter fields are computed by the
* subclass based on the calendar system.
*
* <p>The YEAR_WOY field is computed simplistically.  It is equal to YEAR
* most of the time, but at the year boundary it may be adjusted to YEAR-1
* or YEAR+1 to reflect the overlap of a week into an adjacent year.  In
* this case, a simple increment or decrement is performed on YEAR, even
* though this may yield an invalid YEAR value.  For instance, if the YEAR
* is part of a calendar system with an N-year cycle field CYCLE, then
* incrementing the YEAR may involve incrementing CYCLE and setting YEAR
* back to 0 or 1.  This is not handled by this code, and in fact cannot be
* simply handled without having subclasses define an entire parallel set of
* fields for fields larger than or equal to a year.  This additional
* complexity is not warranted, since the intention of the YEAR_WOY field is
* to support ISO 8601 notation, so it will typically be used with a
* proleptic Gregorian calendar, which has no field larger than a year.
*/
void Calendar::computeWeekFields(UErrorCode &ec) {
    if(U_FAILURE(ec)) {
        return;
    }
    int32_t eyear = fFields[UCAL_EXTENDED_YEAR];
    int32_t dayOfWeek = fFields[UCAL_DAY_OF_WEEK];
    int32_t dayOfYear = fFields[UCAL_DAY_OF_YEAR];

    // WEEK_OF_YEAR start
    // Compute the week of the year.  For the Gregorian calendar, valid week
    // numbers run from 1 to 52 or 53, depending on the year, the first day
    // of the week, and the minimal days in the first week.  For other
    // calendars, the valid range may be different -- it depends on the year
    // length.  Days at the start of the year may fall into the last week of
    // the previous year; days at the end of the year may fall into the
    // first week of the next year.  ASSUME that the year length is less than
    // 7000 days.
    int32_t yearOfWeekOfYear = eyear;
    int32_t relDow = (dayOfWeek + 7 - getFirstDayOfWeek()) % 7; // 0..6
    int32_t relDowJan1 = (dayOfWeek - dayOfYear + 7001 - getFirstDayOfWeek()) % 7; // 0..6
    int32_t woy = (dayOfYear - 1 + relDowJan1) / 7; // 0..53
    if ((7 - relDowJan1) >= getMinimalDaysInFirstWeek()) {
        ++woy;
    }

    // Adjust for weeks at the year end that overlap into the previous or
    // next calendar year.
    if (woy == 0) {
        // We are the last week of the previous year.
        // Check to see if we are in the last week; if so, we need
        // to handle the case in which we are the first week of the
        // next year.

        int32_t prevDoy = dayOfYear + handleGetYearLength(eyear - 1);
        woy = weekNumber(prevDoy, dayOfWeek);
        yearOfWeekOfYear--;
    } else {
        int32_t lastDoy = handleGetYearLength(eyear);
        // Fast check: For it to be week 1 of the next year, the DOY
        // must be on or after L-5, where L is yearLength(), then it
        // cannot possibly be week 1 of the next year:
        //          L-5                  L
        // doy: 359 360 361 362 363 364 365 001
        // dow:      1   2   3   4   5   6   7
        if (dayOfYear >= (lastDoy - 5)) {
            int32_t lastRelDow = (relDow + lastDoy - dayOfYear) % 7;
            if (lastRelDow < 0) {
                lastRelDow += 7;
            }
            if (((6 - lastRelDow) >= getMinimalDaysInFirstWeek()) &&
                ((dayOfYear + 7 - relDow) > lastDoy)) {
                    woy = 1;
                    yearOfWeekOfYear++;
                }
        }
    }
    fFields[UCAL_WEEK_OF_YEAR] = woy;
    fFields[UCAL_YEAR_WOY] = yearOfWeekOfYear;
    // WEEK_OF_YEAR end

    int32_t dayOfMonth = fFields[UCAL_DAY_OF_MONTH];
    fFields[UCAL_WEEK_OF_MONTH] = weekNumber(dayOfMonth, dayOfWeek);
    fFields[UCAL_DAY_OF_WEEK_IN_MONTH] = (dayOfMonth-1) / 7 + 1;
#if defined (U_DEBUG_CAL)
    if(fFields[UCAL_DAY_OF_WEEK_IN_MONTH]==0) fprintf(stderr, "%s:%d: DOWIM %d on %g\n",
        __FILE__, __LINE__,fFields[UCAL_DAY_OF_WEEK_IN_MONTH], fTime);
#endif
}


int32_t Calendar::weekNumber(int32_t desiredDay, int32_t dayOfPeriod, int32_t dayOfWeek)
{
    // Determine the day of the week of the first day of the period
    // in question (either a year or a month).  Zero represents the
    // first day of the week on this calendar.
    int32_t periodStartDayOfWeek = (dayOfWeek - getFirstDayOfWeek() - dayOfPeriod + 1) % 7;
    if (periodStartDayOfWeek < 0) periodStartDayOfWeek += 7;

    // Compute the week number.  Initially, ignore the first week, which
    // may be fractional (or may not be).  We add periodStartDayOfWeek in
    // order to fill out the first week, if it is fractional.
    int32_t weekNo = (desiredDay + periodStartDayOfWeek - 1)/7;

    // If the first week is long enough, then count it.  If
    // the minimal days in the first week is one, or if the period start
    // is zero, we always increment weekNo.
    if ((7 - periodStartDayOfWeek) >= getMinimalDaysInFirstWeek()) ++weekNo;

    return weekNo;
}

void Calendar::handleComputeFields(int32_t /* julianDay */, UErrorCode &/* status */)
{
    internalSet(UCAL_MONTH, getGregorianMonth());
    internalSet(UCAL_DAY_OF_MONTH, getGregorianDayOfMonth());
    internalSet(UCAL_DAY_OF_YEAR, getGregorianDayOfYear());
    int32_t eyear = getGregorianYear();
    internalSet(UCAL_EXTENDED_YEAR, eyear);
    int32_t era = GregorianCalendar::AD;
    if (eyear < 1) {
        era = GregorianCalendar::BC;
        eyear = 1 - eyear;
    }
    internalSet(UCAL_ERA, era);
    internalSet(UCAL_YEAR, eyear);
}
// -------------------------------------


void Calendar::roll(EDateFields field, int32_t amount, UErrorCode& status)
{
    roll((UCalendarDateFields)field, amount, status);
}

void Calendar::roll(UCalendarDateFields field, int32_t amount, UErrorCode& status)
{
    if (amount == 0) {
        return; // Nothing to do
    }

    complete(status);

    if(U_FAILURE(status)) {
        return;
    }
    switch (field) {
    case UCAL_DAY_OF_MONTH:
    case UCAL_AM_PM:
    case UCAL_MINUTE:
    case UCAL_SECOND:
    case UCAL_MILLISECOND:
    case UCAL_MILLISECONDS_IN_DAY:
    case UCAL_ERA:
        // These are the standard roll instructions.  These work for all
        // simple cases, that is, cases in which the limits are fixed, such
        // as the hour, the day of the month, and the era.
        {
            int32_t min = getActualMinimum(field,status);
            int32_t max = getActualMaximum(field,status);
            int32_t gap = max - min + 1;

            int32_t value = internalGet(field) + amount;
            value = (value - min) % gap;
            if (value < 0) {
                value += gap;
            }
            value += min;

            set(field, value);
            return;
        }

    case UCAL_HOUR:
    case UCAL_HOUR_OF_DAY:
        // Rolling the hour is difficult on the ONSET and CEASE days of
        // daylight savings.  For example, if the change occurs at
        // 2 AM, we have the following progression:
        // ONSET: 12 Std -> 1 Std -> 3 Dst -> 4 Dst
        // CEASE: 12 Dst -> 1 Dst -> 1 Std -> 2 Std
        // To get around this problem we don't use fields; we manipulate
        // the time in millis directly.
        {
            // Assume min == 0 in calculations below
            double start = getTimeInMillis(status);
            int32_t oldHour = internalGet(field);
            int32_t max = getMaximum(field);
            int32_t newHour = (oldHour + amount) % (max + 1);
            if (newHour < 0) {
                newHour += max + 1;
            }
            setTimeInMillis(start + kOneHour * (newHour - oldHour),status);
            return;
        }

    case UCAL_MONTH:
        // Rolling the month involves both pinning the final value
        // and adjusting the DAY_OF_MONTH if necessary.  We only adjust the
        // DAY_OF_MONTH if, after updating the MONTH field, it is illegal.
        // E.g., <jan31>.roll(MONTH, 1) -> <feb28> or <feb29>.
        {
            int32_t max = getActualMaximum(UCAL_MONTH, status);
            int32_t mon = (internalGet(UCAL_MONTH) + amount) % (max+1);

            if (mon < 0) {
                mon += (max + 1);
            }
            set(UCAL_MONTH, mon);

            // Keep the day of month in range.  We don't want to spill over
            // into the next month; e.g., we don't want jan31 + 1 mo -> feb31 ->
            // mar3.
            pinField(UCAL_DAY_OF_MONTH,status);
            return;
        }

    case UCAL_YEAR:
    case UCAL_YEAR_WOY:
        {
            // * If era==0 and years go backwards in time, change sign of amount.
            // * Until we have new API per #9393, we temporarily hardcode knowledge of
            //   which calendars have era 0 years that go backwards.
            UBool era0WithYearsThatGoBackwards = FALSE;
            int32_t era = get(UCAL_ERA, status);
            if (era == 0) {
                const char * calType = getType();
                if ( uprv_strcmp(calType,"gregorian")==0 || uprv_strcmp(calType,"roc")==0 || uprv_strcmp(calType,"coptic")==0 ) {
                    amount = -amount;
                    era0WithYearsThatGoBackwards = TRUE;
                }
            }
            int32_t newYear = internalGet(field) + amount;
            if (era > 0 || newYear >= 1) {
                int32_t maxYear = getActualMaximum(field, status);
                if (maxYear < 32768) {
                    // this era has real bounds, roll should wrap years
                    if (newYear < 1) {
                        newYear = maxYear - ((-newYear) % maxYear);
                    } else if (newYear > maxYear) {
                        newYear = ((newYear - 1) % maxYear) + 1;
                    }
                // else era is unbounded, just pin low year instead of wrapping
                } else if (newYear < 1) {
                    newYear = 1;
                }
            // else we are in era 0 with newYear < 1;
            // calendars with years that go backwards must pin the year value at 0,
            // other calendars can have years < 0 in era 0
            } else if (era0WithYearsThatGoBackwards) {
                newYear = 1;
            }
            set(field, newYear);
            pinField(UCAL_MONTH,status);
            pinField(UCAL_DAY_OF_MONTH,status);
            return;
        }

    case UCAL_EXTENDED_YEAR:
        // Rolling the year can involve pinning the DAY_OF_MONTH.
        set(field, internalGet(field) + amount);
        pinField(UCAL_MONTH,status);
        pinField(UCAL_DAY_OF_MONTH,status);
        return;

    case UCAL_WEEK_OF_MONTH:
        {
            // This is tricky, because during the roll we may have to shift
            // to a different day of the week.  For example:

            //    s  m  t  w  r  f  s
            //          1  2  3  4  5
            //    6  7  8  9 10 11 12

            // When rolling from the 6th or 7th back one week, we go to the
            // 1st (assuming that the first partial week counts).  The same
            // thing happens at the end of the month.

            // The other tricky thing is that we have to figure out whether
            // the first partial week actually counts or not, based on the
            // minimal first days in the week.  And we have to use the
            // correct first day of the week to delineate the week
            // boundaries.

            // Here's our algorithm.  First, we find the real boundaries of
            // the month.  Then we discard the first partial week if it
            // doesn't count in this locale.  Then we fill in the ends with
            // phantom days, so that the first partial week and the last
            // partial week are full weeks.  We then have a nice square
            // block of weeks.  We do the usual rolling within this block,
            // as is done elsewhere in this method.  If we wind up on one of
            // the phantom days that we added, we recognize this and pin to
            // the first or the last day of the month.  Easy, eh?

            // Normalize the DAY_OF_WEEK so that 0 is the first day of the week
            // in this locale.  We have dow in 0..6.
            int32_t dow = internalGet(UCAL_DAY_OF_WEEK) - getFirstDayOfWeek();
            if (dow < 0) dow += 7;

            // Find the day of the week (normalized for locale) for the first
            // of the month.
            int32_t fdm = (dow - internalGet(UCAL_DAY_OF_MONTH) + 1) % 7;
            if (fdm < 0) fdm += 7;

            // Get the first day of the first full week of the month,
            // including phantom days, if any.  Figure out if the first week
            // counts or not; if it counts, then fill in phantom days.  If
            // not, advance to the first real full week (skip the partial week).
            int32_t start;
            if ((7 - fdm) < getMinimalDaysInFirstWeek())
                start = 8 - fdm; // Skip the first partial week
            else
                start = 1 - fdm; // This may be zero or negative

            // Get the day of the week (normalized for locale) for the last
            // day of the month.
            int32_t monthLen = getActualMaximum(UCAL_DAY_OF_MONTH, status);
            int32_t ldm = (monthLen - internalGet(UCAL_DAY_OF_MONTH) + dow) % 7;
            // We know monthLen >= DAY_OF_MONTH so we skip the += 7 step here.

            // Get the limit day for the blocked-off rectangular month; that
            // is, the day which is one past the last day of the month,
            // after the month has already been filled in with phantom days
            // to fill out the last week.  This day has a normalized DOW of 0.
            int32_t limit = monthLen + 7 - ldm;

            // Now roll between start and (limit - 1).
            int32_t gap = limit - start;
            int32_t day_of_month = (internalGet(UCAL_DAY_OF_MONTH) + amount*7 -
                start) % gap;
            if (day_of_month < 0) day_of_month += gap;
            day_of_month += start;

            // Finally, pin to the real start and end of the month.
            if (day_of_month < 1) day_of_month = 1;
            if (day_of_month > monthLen) day_of_month = monthLen;

            // Set the DAY_OF_MONTH.  We rely on the fact that this field
            // takes precedence over everything else (since all other fields
            // are also set at this point).  If this fact changes (if the
            // disambiguation algorithm changes) then we will have to unset
            // the appropriate fields here so that DAY_OF_MONTH is attended
            // to.
            set(UCAL_DAY_OF_MONTH, day_of_month);
            return;
        }
    case UCAL_WEEK_OF_YEAR:
        {
            // This follows the outline of WEEK_OF_MONTH, except it applies
            // to the whole year.  Please see the comment for WEEK_OF_MONTH
            // for general notes.

            // Normalize the DAY_OF_WEEK so that 0 is the first day of the week
            // in this locale.  We have dow in 0..6.
            int32_t dow = internalGet(UCAL_DAY_OF_WEEK) - getFirstDayOfWeek();
            if (dow < 0) dow += 7;

            // Find the day of the week (normalized for locale) for the first
            // of the year.
            int32_t fdy = (dow - internalGet(UCAL_DAY_OF_YEAR) + 1) % 7;
            if (fdy < 0) fdy += 7;

            // Get the first day of the first full week of the year,
            // including phantom days, if any.  Figure out if the first week
            // counts or not; if it counts, then fill in phantom days.  If
            // not, advance to the first real full week (skip the partial week).
            int32_t start;
            if ((7 - fdy) < getMinimalDaysInFirstWeek())
                start = 8 - fdy; // Skip the first partial week
            else
                start = 1 - fdy; // This may be zero or negative

            // Get the day of the week (normalized for locale) for the last
            // day of the year.
            int32_t yearLen = getActualMaximum(UCAL_DAY_OF_YEAR,status);
            int32_t ldy = (yearLen - internalGet(UCAL_DAY_OF_YEAR) + dow) % 7;
            // We know yearLen >= DAY_OF_YEAR so we skip the += 7 step here.

            // Get the limit day for the blocked-off rectangular year; that
            // is, the day which is one past the last day of the year,
            // after the year has already been filled in with phantom days
            // to fill out the last week.  This day has a normalized DOW of 0.
            int32_t limit = yearLen + 7 - ldy;

            // Now roll between start and (limit - 1).
            int32_t gap = limit - start;
            int32_t day_of_year = (internalGet(UCAL_DAY_OF_YEAR) + amount*7 -
                start) % gap;
            if (day_of_year < 0) day_of_year += gap;
            day_of_year += start;

            // Finally, pin to the real start and end of the month.
            if (day_of_year < 1) day_of_year = 1;
            if (day_of_year > yearLen) day_of_year = yearLen;

            // Make sure that the year and day of year are attended to by
            // clearing other fields which would normally take precedence.
            // If the disambiguation algorithm is changed, this section will
            // have to be updated as well.
            set(UCAL_DAY_OF_YEAR, day_of_year);
            clear(UCAL_MONTH);
            return;
        }
    case UCAL_DAY_OF_YEAR:
        {
            // Roll the day of year using millis.  Compute the millis for
            // the start of the year, and get the length of the year.
            double delta = amount * kOneDay; // Scale up from days to millis
            double min2 = internalGet(UCAL_DAY_OF_YEAR)-1;
            min2 *= kOneDay;
            min2 = internalGetTime() - min2;

            //      double min2 = internalGetTime() - (internalGet(UCAL_DAY_OF_YEAR) - 1.0) * kOneDay;
            double newtime;

            double yearLength = getActualMaximum(UCAL_DAY_OF_YEAR,status);
            double oneYear = yearLength;
            oneYear *= kOneDay;
            newtime = uprv_fmod((internalGetTime() + delta - min2), oneYear);
            if (newtime < 0) newtime += oneYear;
            setTimeInMillis(newtime + min2, status);
            return;
        }
    case UCAL_DAY_OF_WEEK:
    case UCAL_DOW_LOCAL:
        {
            // Roll the day of week using millis.  Compute the millis for
            // the start of the week, using the first day of week setting.
            // Restrict the millis to [start, start+7days).
            double delta = amount * kOneDay; // Scale up from days to millis
            // Compute the number of days before the current day in this
            // week.  This will be a value 0..6.
            int32_t leadDays = internalGet(field);
            leadDays -= (field == UCAL_DAY_OF_WEEK) ? getFirstDayOfWeek() : 1;
            if (leadDays < 0) leadDays += 7;
            double min2 = internalGetTime() - leadDays * kOneDay;
            double newtime = uprv_fmod((internalGetTime() + delta - min2), kOneWeek);
            if (newtime < 0) newtime += kOneWeek;
            setTimeInMillis(newtime + min2, status);
            return;
        }
    case UCAL_DAY_OF_WEEK_IN_MONTH:
        {
            // Roll the day of week in the month using millis.  Determine
            // the first day of the week in the month, and then the last,
            // and then roll within that range.
            double delta = amount * kOneWeek; // Scale up from weeks to millis
            // Find the number of same days of the week before this one
            // in this month.
            int32_t preWeeks = (internalGet(UCAL_DAY_OF_MONTH) - 1) / 7;
            // Find the number of same days of the week after this one
            // in this month.
            int32_t postWeeks = (getActualMaximum(UCAL_DAY_OF_MONTH,status) -
                internalGet(UCAL_DAY_OF_MONTH)) / 7;
            // From these compute the min and gap millis for rolling.
            double min2 = internalGetTime() - preWeeks * kOneWeek;
            double gap2 = kOneWeek * (preWeeks + postWeeks + 1); // Must add 1!
            // Roll within this range
            double newtime = uprv_fmod((internalGetTime() + delta - min2), gap2);
            if (newtime < 0) newtime += gap2;
            setTimeInMillis(newtime + min2, status);
            return;
        }
    case UCAL_JULIAN_DAY:
        set(field, internalGet(field) + amount);
        return;
    default:
        // Other fields cannot be rolled by this method
#if defined (U_DEBUG_CAL)
        fprintf(stderr, "%s:%d: ILLEGAL ARG because of roll on non-rollable field %s\n",
            __FILE__, __LINE__,fldName(field));
#endif
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

void Calendar::add(EDateFields field, int32_t amount, UErrorCode& status)
{
    Calendar::add((UCalendarDateFields)field, amount, status);
}

// -------------------------------------
void Calendar::add(UCalendarDateFields field, int32_t amount, UErrorCode& status)
{
    if (amount == 0) {
        return;   // Do nothing!
    }

    // We handle most fields in the same way.  The algorithm is to add
    // a computed amount of millis to the current millis.  The only
    // wrinkle is with DST (and/or a change to the zone's UTC offset, which
    // we'll include with DST) -- for some fields, like the DAY_OF_MONTH,
    // we don't want the wall time to shift due to changes in DST.  If the
    // result of the add operation is to move from DST to Standard, or
    // vice versa, we need to adjust by an hour forward or back,
    // respectively.  For such fields we set keepWallTimeInvariant to TRUE.

    // We only adjust the DST for fields larger than an hour.  For
    // fields smaller than an hour, we cannot adjust for DST without
    // causing problems.  for instance, if you add one hour to April 5,
    // 1998, 1:00 AM, in PST, the time becomes "2:00 AM PDT" (an
    // illegal value), but then the adjustment sees the change and
    // compensates by subtracting an hour.  As a result the time
    // doesn't advance at all.

    // For some fields larger than a day, such as a UCAL_MONTH, we pin the
    // UCAL_DAY_OF_MONTH.  This allows <March 31>.add(UCAL_MONTH, 1) to be
    // <April 30>, rather than <April 31> => <May 1>.

    double delta = amount; // delta in ms
    UBool keepWallTimeInvariant = TRUE;

    switch (field) {
    case UCAL_ERA:
        set(field, get(field, status) + amount);
        pinField(UCAL_ERA, status);
        return;

    case UCAL_YEAR:
    case UCAL_YEAR_WOY:
      {
        // * If era=0 and years go backwards in time, change sign of amount.
        // * Until we have new API per #9393, we temporarily hardcode knowledge of
        //   which calendars have era 0 years that go backwards.
        // * Note that for UCAL_YEAR (but not UCAL_YEAR_WOY) we could instead handle
        //   this by applying the amount to the UCAL_EXTENDED_YEAR field; but since
        //   we would still need to handle UCAL_YEAR_WOY as below, might as well
        //   also handle UCAL_YEAR the same way.
        int32_t era = get(UCAL_ERA, status);
        if (era == 0) {
          const char * calType = getType();
          if ( uprv_strcmp(calType,"gregorian")==0 || uprv_strcmp(calType,"roc")==0 || uprv_strcmp(calType,"coptic")==0 ) {
            amount = -amount;
          }
        }
      }
      // Fall through into normal handling
      U_FALLTHROUGH;
    case UCAL_EXTENDED_YEAR:
    case UCAL_MONTH:
      {
        UBool oldLenient = isLenient();
        setLenient(TRUE);
        set(field, get(field, status) + amount);
        pinField(UCAL_DAY_OF_MONTH, status);
        if(oldLenient==FALSE) {
          complete(status); /* force recalculate */
          setLenient(oldLenient);
        }
      }
      return;

    case UCAL_WEEK_OF_YEAR:
    case UCAL_WEEK_OF_MONTH:
    case UCAL_DAY_OF_WEEK_IN_MONTH:
        delta *= kOneWeek;
        break;

    case UCAL_AM_PM:
        delta *= 12 * kOneHour;
        break;

    case UCAL_DAY_OF_MONTH:
    case UCAL_DAY_OF_YEAR:
    case UCAL_DAY_OF_WEEK:
    case UCAL_DOW_LOCAL:
    case UCAL_JULIAN_DAY:
        delta *= kOneDay;
        break;

    case UCAL_HOUR_OF_DAY:
    case UCAL_HOUR:
        delta *= kOneHour;
        keepWallTimeInvariant = FALSE;
        break;

    case UCAL_MINUTE:
        delta *= kOneMinute;
        keepWallTimeInvariant = FALSE;
        break;

    case UCAL_SECOND:
        delta *= kOneSecond;
        keepWallTimeInvariant = FALSE;
        break;

    case UCAL_MILLISECOND:
    case UCAL_MILLISECONDS_IN_DAY:
        keepWallTimeInvariant = FALSE;
        break;

    default:
#if defined (U_DEBUG_CAL)
        fprintf(stderr, "%s:%d: ILLEGAL ARG because field %s not addable",
            __FILE__, __LINE__, fldName(field));
#endif
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
        //  throw new IllegalArgumentException("Calendar.add(" + fieldName(field) +
        //                                     ") not supported");
    }

    // In order to keep the wall time invariant (for fields where this is
    // appropriate), check the combined DST & ZONE offset before and
    // after the add() operation. If it changes, then adjust the millis
    // to compensate.
    int32_t prevOffset = 0;
    int32_t prevWallTime = 0;
    if (keepWallTimeInvariant) {
        prevOffset = get(UCAL_DST_OFFSET, status) + get(UCAL_ZONE_OFFSET, status);
        prevWallTime = get(UCAL_MILLISECONDS_IN_DAY, status);
    }

    setTimeInMillis(getTimeInMillis(status) + delta, status);

    if (keepWallTimeInvariant) {
        int32_t newWallTime = get(UCAL_MILLISECONDS_IN_DAY, status);
        if (newWallTime != prevWallTime) {
            // There is at least one zone transition between the base
            // time and the result time. As the result, wall time has
            // changed.
            UDate t = internalGetTime();
            int32_t newOffset = get(UCAL_DST_OFFSET, status) + get(UCAL_ZONE_OFFSET, status);
            if (newOffset != prevOffset) {
                // When the difference of the previous UTC offset and
                // the new UTC offset exceeds 1 full day, we do not want
                // to roll over/back the date. For now, this only happens
                // in Samoa (Pacific/Apia) on Dec 30, 2011. See ticket:9452.
                int32_t adjAmount = prevOffset - newOffset;
                adjAmount = adjAmount >= 0 ? adjAmount % (int32_t)kOneDay : -(-adjAmount % (int32_t)kOneDay);
                if (adjAmount != 0) {
                    setTimeInMillis(t + adjAmount, status);
                    newWallTime = get(UCAL_MILLISECONDS_IN_DAY, status);
                }
                if (newWallTime != prevWallTime) {
                    // The result wall time or adjusted wall time was shifted because
                    // the target wall time does not exist on the result date.
                    switch (fSkippedWallTime) {
                    case UCAL_WALLTIME_FIRST:
                        if (adjAmount > 0) {
                            setTimeInMillis(t, status);
                        }
                        break;
                    case UCAL_WALLTIME_LAST:
                        if (adjAmount < 0) {
                            setTimeInMillis(t, status);
                        }
                        break;
                    case UCAL_WALLTIME_NEXT_VALID:
                        UDate tmpT = adjAmount > 0 ? internalGetTime() : t;
                        UDate immediatePrevTrans;
                        UBool hasTransition = getImmediatePreviousZoneTransition(tmpT, &immediatePrevTrans, status);
                        if (U_SUCCESS(status) && hasTransition) {
                            setTimeInMillis(immediatePrevTrans, status);
                        }
                        break;
                    }
                }
            }
        }
    }
}

// -------------------------------------
int32_t Calendar::fieldDifference(UDate when, EDateFields field, UErrorCode& status) {
    return fieldDifference(when, (UCalendarDateFields) field, status);
}

int32_t Calendar::fieldDifference(UDate targetMs, UCalendarDateFields field, UErrorCode& ec) {
    if (U_FAILURE(ec)) return 0;
    int32_t min = 0;
    double startMs = getTimeInMillis(ec);
    // Always add from the start millis.  This accomodates
    // operations like adding years from February 29, 2000 up to
    // February 29, 2004.  If 1, 1, 1, 1 is added to the year
    // field, the DOM gets pinned to 28 and stays there, giving an
    // incorrect DOM difference of 1.  We have to add 1, reset, 2,
    // reset, 3, reset, 4.
    if (startMs < targetMs) {
        int32_t max = 1;
        // Find a value that is too large
        while (U_SUCCESS(ec)) {
            setTimeInMillis(startMs, ec);
            add(field, max, ec);
            double ms = getTimeInMillis(ec);
            if (ms == targetMs) {
                return max;
            } else if (ms > targetMs) {
                break;
            } else if (max < INT32_MAX) {
                min = max;
                max <<= 1;
                if (max < 0) {
                    max = INT32_MAX;
                }
            } else {
                // Field difference too large to fit into int32_t
#if defined (U_DEBUG_CAL)
                fprintf(stderr, "%s:%d: ILLEGAL ARG because field %s's max too large for int32_t\n",
                    __FILE__, __LINE__, fldName(field));
#endif
                ec = U_ILLEGAL_ARGUMENT_ERROR;
            }
        }
        // Do a binary search
        while ((max - min) > 1 && U_SUCCESS(ec)) {
            int32_t t = min + (max - min)/2; // make sure intermediate values don't exceed INT32_MAX
            setTimeInMillis(startMs, ec);
            add(field, t, ec);
            double ms = getTimeInMillis(ec);
            if (ms == targetMs) {
                return t;
            } else if (ms > targetMs) {
                max = t;
            } else {
                min = t;
            }
        }
    } else if (startMs > targetMs) {
        int32_t max = -1;
        // Find a value that is too small
        while (U_SUCCESS(ec)) {
            setTimeInMillis(startMs, ec);
            add(field, max, ec);
            double ms = getTimeInMillis(ec);
            if (ms == targetMs) {
                return max;
            } else if (ms < targetMs) {
                break;
            } else {
                min = max;
                max <<= 1;
                if (max == 0) {
                    // Field difference too large to fit into int32_t
#if defined (U_DEBUG_CAL)
                    fprintf(stderr, "%s:%d: ILLEGAL ARG because field %s's max too large for int32_t\n",
                        __FILE__, __LINE__, fldName(field));
#endif
                    ec = U_ILLEGAL_ARGUMENT_ERROR;
                }
            }
        }
        // Do a binary search
        while ((min - max) > 1 && U_SUCCESS(ec)) {
            int32_t t = min + (max - min)/2; // make sure intermediate values don't exceed INT32_MAX
            setTimeInMillis(startMs, ec);
            add(field, t, ec);
            double ms = getTimeInMillis(ec);
            if (ms == targetMs) {
                return t;
            } else if (ms < targetMs) {
                max = t;
            } else {
                min = t;
            }
        }
    }
    // Set calendar to end point
    setTimeInMillis(startMs, ec);
    add(field, min, ec);

    /* Test for buffer overflows */
    if(U_FAILURE(ec)) {
        return 0;
    }
    return min;
}

// -------------------------------------

void
Calendar::adoptTimeZone(TimeZone* zone)
{
    // Do nothing if passed-in zone is NULL
    if (zone == NULL) return;

    // fZone should always be non-null
    delete fZone;
    fZone = zone;

    // if the zone changes, we need to recompute the time fields
    fAreFieldsSet = FALSE;
}

// -------------------------------------
void
Calendar::setTimeZone(const TimeZone& zone)
{
    adoptTimeZone(zone.clone());
}

// -------------------------------------

const TimeZone&
Calendar::getTimeZone() const
{
    U_ASSERT(fZone != NULL);
    return *fZone;
}

// -------------------------------------

TimeZone*
Calendar::orphanTimeZone()
{
    // we let go of the time zone; the new time zone is the system default time zone
    TimeZone *defaultZone = TimeZone::createDefault();
    if (defaultZone == NULL) {
        // No error handling available. Must keep fZone non-NULL, there are many unchecked uses.
        return NULL;
    }
    TimeZone *z = fZone;
    fZone = defaultZone;
    return z;
}

// -------------------------------------

void
Calendar::setLenient(UBool lenient)
{
    fLenient = lenient;
}

// -------------------------------------

UBool
Calendar::isLenient() const
{
    return fLenient;
}

// -------------------------------------

void
Calendar::setRepeatedWallTimeOption(UCalendarWallTimeOption option)
{
    if (option == UCAL_WALLTIME_LAST || option == UCAL_WALLTIME_FIRST) {
        fRepeatedWallTime = option;
    }
}

// -------------------------------------

UCalendarWallTimeOption
Calendar::getRepeatedWallTimeOption(void) const
{
    return fRepeatedWallTime;
}

// -------------------------------------

void
Calendar::setSkippedWallTimeOption(UCalendarWallTimeOption option)
{
    fSkippedWallTime = option;
}

// -------------------------------------

UCalendarWallTimeOption
Calendar::getSkippedWallTimeOption(void) const
{
    return fSkippedWallTime;
}

// -------------------------------------

void
Calendar::setFirstDayOfWeek(UCalendarDaysOfWeek value)
{
    if (fFirstDayOfWeek != value &&
        value >= UCAL_SUNDAY && value <= UCAL_SATURDAY) {
            fFirstDayOfWeek = value;
            fAreFieldsSet = FALSE;
        }
}

// -------------------------------------

Calendar::EDaysOfWeek
Calendar::getFirstDayOfWeek() const
{
    return (Calendar::EDaysOfWeek)fFirstDayOfWeek;
}

UCalendarDaysOfWeek
Calendar::getFirstDayOfWeek(UErrorCode & /*status*/) const
{
    return fFirstDayOfWeek;
}
// -------------------------------------

void
Calendar::setMinimalDaysInFirstWeek(uint8_t value)
{
    // Values less than 1 have the same effect as 1; values greater
    // than 7 have the same effect as 7. However, we normalize values
    // so operator== and so forth work.
    if (value < 1) {
        value = 1;
    } else if (value > 7) {
        value = 7;
    }
    if (fMinimalDaysInFirstWeek != value) {
        fMinimalDaysInFirstWeek = value;
        fAreFieldsSet = FALSE;
    }
}

// -------------------------------------

uint8_t
Calendar::getMinimalDaysInFirstWeek() const
{
    return fMinimalDaysInFirstWeek;
}

// -------------------------------------
// weekend functions, just dummy implementations for now (for API freeze)

UCalendarWeekdayType
Calendar::getDayOfWeekType(UCalendarDaysOfWeek dayOfWeek, UErrorCode &status) const
{
    if (U_FAILURE(status)) {
        return UCAL_WEEKDAY;
    }
    if (dayOfWeek < UCAL_SUNDAY || dayOfWeek > UCAL_SATURDAY) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return UCAL_WEEKDAY;
    }
    if (fWeekendOnset == fWeekendCease) {
        if (dayOfWeek != fWeekendOnset)
            return UCAL_WEEKDAY;
        return (fWeekendOnsetMillis == 0) ? UCAL_WEEKEND : UCAL_WEEKEND_ONSET;
    }
    if (fWeekendOnset < fWeekendCease) {
        if (dayOfWeek < fWeekendOnset || dayOfWeek > fWeekendCease) {
            return UCAL_WEEKDAY;
        }
    } else {
        if (dayOfWeek > fWeekendCease && dayOfWeek < fWeekendOnset) {
            return UCAL_WEEKDAY;
        }
    }
    if (dayOfWeek == fWeekendOnset) {
        return (fWeekendOnsetMillis == 0) ? UCAL_WEEKEND : UCAL_WEEKEND_ONSET;
    }
    if (dayOfWeek == fWeekendCease) {
        return (fWeekendCeaseMillis >= 86400000) ? UCAL_WEEKEND : UCAL_WEEKEND_CEASE;
    }
    return UCAL_WEEKEND;
}

int32_t
Calendar::getWeekendTransition(UCalendarDaysOfWeek dayOfWeek, UErrorCode &status) const
{
    if (U_FAILURE(status)) {
        return 0;
    }
    if (dayOfWeek == fWeekendOnset) {
        return fWeekendOnsetMillis;
    } else if (dayOfWeek == fWeekendCease) {
        return fWeekendCeaseMillis;
    }
    status = U_ILLEGAL_ARGUMENT_ERROR;
    return 0;
}

UBool
Calendar::isWeekend(UDate date, UErrorCode &status) const
{
    if (U_FAILURE(status)) {
        return FALSE;
    }
    // clone the calendar so we don't mess with the real one.
    Calendar *work = this->clone();
    if (work == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    UBool result = FALSE;
    work->setTime(date, status);
    if (U_SUCCESS(status)) {
        result = work->isWeekend();
    }
    delete work;
    return result;
}

UBool
Calendar::isWeekend(void) const
{
    UErrorCode status = U_ZERO_ERROR;
    UCalendarDaysOfWeek dayOfWeek = (UCalendarDaysOfWeek)get(UCAL_DAY_OF_WEEK, status);
    UCalendarWeekdayType dayType = getDayOfWeekType(dayOfWeek, status);
    if (U_SUCCESS(status)) {
        switch (dayType) {
            case UCAL_WEEKDAY:
                return FALSE;
            case UCAL_WEEKEND:
                return TRUE;
            case UCAL_WEEKEND_ONSET:
            case UCAL_WEEKEND_CEASE:
                // Use internalGet() because the above call to get() populated all fields.
                {
                    int32_t millisInDay = internalGet(UCAL_MILLISECONDS_IN_DAY);
                    int32_t transitionMillis = getWeekendTransition(dayOfWeek, status);
                    if (U_SUCCESS(status)) {
                        return (dayType == UCAL_WEEKEND_ONSET)?
                            (millisInDay >= transitionMillis):
                            (millisInDay <  transitionMillis);
                    }
                    // else fall through, return FALSE
                    U_FALLTHROUGH;
                }
            default:
                break;
        }
    }
    return FALSE;
}

// ------------------------------------- limits

int32_t
Calendar::getMinimum(EDateFields field) const {
    return getLimit((UCalendarDateFields) field,UCAL_LIMIT_MINIMUM);
}

int32_t
Calendar::getMinimum(UCalendarDateFields field) const
{
    return getLimit(field,UCAL_LIMIT_MINIMUM);
}

// -------------------------------------
int32_t
Calendar::getMaximum(EDateFields field) const
{
    return getLimit((UCalendarDateFields) field,UCAL_LIMIT_MAXIMUM);
}

int32_t
Calendar::getMaximum(UCalendarDateFields field) const
{
    return getLimit(field,UCAL_LIMIT_MAXIMUM);
}

// -------------------------------------
int32_t
Calendar::getGreatestMinimum(EDateFields field) const
{
    return getLimit((UCalendarDateFields)field,UCAL_LIMIT_GREATEST_MINIMUM);
}

int32_t
Calendar::getGreatestMinimum(UCalendarDateFields field) const
{
    return getLimit(field,UCAL_LIMIT_GREATEST_MINIMUM);
}

// -------------------------------------
int32_t
Calendar::getLeastMaximum(EDateFields field) const
{
    return getLimit((UCalendarDateFields) field,UCAL_LIMIT_LEAST_MAXIMUM);
}

int32_t
Calendar::getLeastMaximum(UCalendarDateFields field) const
{
    return getLimit( field,UCAL_LIMIT_LEAST_MAXIMUM);
}

// -------------------------------------
int32_t
Calendar::getActualMinimum(EDateFields field, UErrorCode& status) const
{
    return getActualMinimum((UCalendarDateFields) field, status);
}

int32_t Calendar::getLimit(UCalendarDateFields field, ELimitType limitType) const {
    switch (field) {
    case UCAL_DAY_OF_WEEK:
    case UCAL_AM_PM:
    case UCAL_HOUR:
    case UCAL_HOUR_OF_DAY:
    case UCAL_MINUTE:
    case UCAL_SECOND:
    case UCAL_MILLISECOND:
    case UCAL_ZONE_OFFSET:
    case UCAL_DST_OFFSET:
    case UCAL_DOW_LOCAL:
    case UCAL_JULIAN_DAY:
    case UCAL_MILLISECONDS_IN_DAY:
    case UCAL_IS_LEAP_MONTH:
        return kCalendarLimits[field][limitType];

    case UCAL_WEEK_OF_MONTH:
        {
            int32_t limit;
            if (limitType == UCAL_LIMIT_MINIMUM) {
                limit = getMinimalDaysInFirstWeek() == 1 ? 1 : 0;
            } else if (limitType == UCAL_LIMIT_GREATEST_MINIMUM) {
                limit = 1;
            } else {
                int32_t minDaysInFirst = getMinimalDaysInFirstWeek();
                int32_t daysInMonth = handleGetLimit(UCAL_DAY_OF_MONTH, limitType);
                if (limitType == UCAL_LIMIT_LEAST_MAXIMUM) {
                    limit = (daysInMonth + (7 - minDaysInFirst)) / 7;
                } else { // limitType == UCAL_LIMIT_MAXIMUM
                    limit = (daysInMonth + 6 + (7 - minDaysInFirst)) / 7;
                }
            }
            return limit;
        }
    default:
        return handleGetLimit(field, limitType);
    }
}


int32_t
Calendar::getActualMinimum(UCalendarDateFields field, UErrorCode& status) const
{
    int32_t fieldValue = getGreatestMinimum(field);
    int32_t endValue = getMinimum(field);

    // if we know that the minimum value is always the same, just return it
    if (fieldValue == endValue) {
        return fieldValue;
    }

    // clone the calendar so we don't mess with the real one, and set it to
    // accept anything for the field values
    Calendar *work = this->clone();
    if (work == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    work->setLenient(TRUE);

    // now try each value from getLeastMaximum() to getMaximum() one by one until
    // we get a value that normalizes to another value.  The last value that
    // normalizes to itself is the actual minimum for the current date
    int32_t result = fieldValue;

    do {
        work->set(field, fieldValue);
        if (work->get(field, status) != fieldValue) {
            break;
        }
        else {
            result = fieldValue;
            fieldValue--;
        }
    } while (fieldValue >= endValue);

    delete work;

    /* Test for buffer overflows */
    if(U_FAILURE(status)) {
        return 0;
    }
    return result;
}

// -------------------------------------



/**
* Ensure that each field is within its valid range by calling {@link
* #validateField(int)} on each field that has been set.  This method
* should only be called if this calendar is not lenient.
* @see #isLenient
* @see #validateField(int)
*/
void Calendar::validateFields(UErrorCode &status) {
    for (int32_t field = 0; U_SUCCESS(status) && (field < UCAL_FIELD_COUNT); field++) {
        if (fStamp[field] >= kMinimumUserStamp) {
            validateField((UCalendarDateFields)field, status);
        }
    }
}

/**
* Validate a single field of this calendar.  Subclasses should
* override this method to validate any calendar-specific fields.
* Generic fields can be handled by
* <code>Calendar.validateField()</code>.
* @see #validateField(int, int, int)
*/
void Calendar::validateField(UCalendarDateFields field, UErrorCode &status) {
    int32_t y;
    switch (field) {
    case UCAL_DAY_OF_MONTH:
        y = handleGetExtendedYear();
        validateField(field, 1, handleGetMonthLength(y, internalGet(UCAL_MONTH)), status);
        break;
    case UCAL_DAY_OF_YEAR:
        y = handleGetExtendedYear();
        validateField(field, 1, handleGetYearLength(y), status);
        break;
    case UCAL_DAY_OF_WEEK_IN_MONTH:
        if (internalGet(field) == 0) {
#if defined (U_DEBUG_CAL)
            fprintf(stderr, "%s:%d: ILLEGAL ARG because DOW in month cannot be 0\n",
                __FILE__, __LINE__);
#endif
            status = U_ILLEGAL_ARGUMENT_ERROR; // "DAY_OF_WEEK_IN_MONTH cannot be zero"
            return;
        }
        validateField(field, getMinimum(field), getMaximum(field), status);
        break;
    default:
        validateField(field, getMinimum(field), getMaximum(field), status);
        break;
    }
}

/**
* Validate a single field of this calendar given its minimum and
* maximum allowed value.  If the field is out of range, throw a
* descriptive <code>IllegalArgumentException</code>.  Subclasses may
* use this method in their implementation of {@link
* #validateField(int)}.
*/
void Calendar::validateField(UCalendarDateFields field, int32_t min, int32_t max, UErrorCode& status)
{
    int32_t value = fFields[field];
    if (value < min || value > max) {
#if defined (U_DEBUG_CAL)
        fprintf(stderr, "%s:%d: ILLEGAL ARG because of field %s out of range %d..%d  at %d\n",
            __FILE__, __LINE__,fldName(field),min,max,value);
#endif
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
}

// -------------------------

const UFieldResolutionTable* Calendar::getFieldResolutionTable() const {
    return kDatePrecedence;
}


UCalendarDateFields Calendar::newerField(UCalendarDateFields defaultField, UCalendarDateFields alternateField) const
{
    if (fStamp[alternateField] > fStamp[defaultField]) {
        return alternateField;
    }
    return defaultField;
}

UCalendarDateFields Calendar::resolveFields(const UFieldResolutionTable* precedenceTable) {
    int32_t bestField = UCAL_FIELD_COUNT;
    int32_t tempBestField;
    for (int32_t g=0; precedenceTable[g][0][0] != -1 && (bestField == UCAL_FIELD_COUNT); ++g) {
        int32_t bestStamp = kUnset;
        for (int32_t l=0; precedenceTable[g][l][0] != -1; ++l) {
            int32_t lineStamp = kUnset;
            // Skip over first entry if it is negative
            for (int32_t i=((precedenceTable[g][l][0]>=kResolveRemap)?1:0); precedenceTable[g][l][i]!=-1; ++i) {
                U_ASSERT(precedenceTable[g][l][i] < UCAL_FIELD_COUNT);
                int32_t s = fStamp[precedenceTable[g][l][i]];
                // If any field is unset then don't use this line
                if (s == kUnset) {
                    goto linesInGroup;
                } else if(s > lineStamp) {
                    lineStamp = s;
                }
            }
            // Record new maximum stamp & field no.
            if (lineStamp > bestStamp) {
                tempBestField = precedenceTable[g][l][0]; // First field refers to entire line
                if (tempBestField >= kResolveRemap) {
                    tempBestField &= (kResolveRemap-1);
                    // This check is needed to resolve some issues with UCAL_YEAR precedence mapping
                    if (tempBestField != UCAL_DATE || (fStamp[UCAL_WEEK_OF_MONTH] < fStamp[tempBestField])) {
                        bestField = tempBestField;
                    }
                } else {
                    bestField = tempBestField;
                }

                if (bestField == tempBestField) {
                    bestStamp = lineStamp;
                }
            }
linesInGroup:
            ;
        }
    }
    return (UCalendarDateFields)bestField;
}

const UFieldResolutionTable Calendar::kDatePrecedence[] =
{
    {
        { UCAL_DAY_OF_MONTH, kResolveSTOP },
        { UCAL_WEEK_OF_YEAR, UCAL_DAY_OF_WEEK, kResolveSTOP },
        { UCAL_WEEK_OF_MONTH, UCAL_DAY_OF_WEEK, kResolveSTOP },
        { UCAL_DAY_OF_WEEK_IN_MONTH, UCAL_DAY_OF_WEEK, kResolveSTOP },
        { UCAL_WEEK_OF_YEAR, UCAL_DOW_LOCAL, kResolveSTOP },
        { UCAL_WEEK_OF_MONTH, UCAL_DOW_LOCAL, kResolveSTOP },
        { UCAL_DAY_OF_WEEK_IN_MONTH, UCAL_DOW_LOCAL, kResolveSTOP },
        { UCAL_DAY_OF_YEAR, kResolveSTOP },
        { kResolveRemap | UCAL_DAY_OF_MONTH, UCAL_YEAR, kResolveSTOP },  // if YEAR is set over YEAR_WOY use DAY_OF_MONTH
        { kResolveRemap | UCAL_WEEK_OF_YEAR, UCAL_YEAR_WOY, kResolveSTOP },  // if YEAR_WOY is set,  calc based on WEEK_OF_YEAR
        { kResolveSTOP }
    },
    {
        { UCAL_WEEK_OF_YEAR, kResolveSTOP },
        { UCAL_WEEK_OF_MONTH, kResolveSTOP },
        { UCAL_DAY_OF_WEEK_IN_MONTH, kResolveSTOP },
        { kResolveRemap | UCAL_DAY_OF_WEEK_IN_MONTH, UCAL_DAY_OF_WEEK, kResolveSTOP },
        { kResolveRemap | UCAL_DAY_OF_WEEK_IN_MONTH, UCAL_DOW_LOCAL, kResolveSTOP },
        { kResolveSTOP }
    },
    {{kResolveSTOP}}
};


const UFieldResolutionTable Calendar::kDOWPrecedence[] =
{
    {
        { UCAL_DAY_OF_WEEK,kResolveSTOP, kResolveSTOP },
        { UCAL_DOW_LOCAL,kResolveSTOP, kResolveSTOP },
        {kResolveSTOP}
    },
    {{kResolveSTOP}}
};

// precedence for calculating a year
const UFieldResolutionTable Calendar::kYearPrecedence[] =
{
    {
        { UCAL_YEAR, kResolveSTOP },
        { UCAL_EXTENDED_YEAR, kResolveSTOP },
        { UCAL_YEAR_WOY, UCAL_WEEK_OF_YEAR, kResolveSTOP },  // YEAR_WOY is useless without WEEK_OF_YEAR
        { kResolveSTOP }
    },
    {{kResolveSTOP}}
};


// -------------------------


void Calendar::computeTime(UErrorCode& status) {
    if (!isLenient()) {
        validateFields(status);
        if (U_FAILURE(status)) {
            return;
        }
    }

    // Compute the Julian day
    int32_t julianDay = computeJulianDay();

    double millis = Grego::julianDayToMillis(julianDay);

#if defined (U_DEBUG_CAL)
    //  int32_t julianInsanityCheck =  (int32_t)ClockMath::floorDivide(millis, kOneDay);
    //  julianInsanityCheck += kEpochStartAsJulianDay;
    //  if(1 || julianInsanityCheck != julianDay) {
    //    fprintf(stderr, "%s:%d- D'oh- computed jules %d, to mills (%s)%.lf, recomputed %d\n",
    //            __FILE__, __LINE__, julianDay, millis<0.0?"NEG":"", millis, julianInsanityCheck);
    //  }
#endif

    double millisInDay;

    // We only use MILLISECONDS_IN_DAY if it has been set by the user.
    // This makes it possible for the caller to set the calendar to a
    // time and call clear(MONTH) to reset the MONTH to January.  This
    // is legacy behavior.  Without this, clear(MONTH) has no effect,
    // since the internally set JULIAN_DAY is used.
    if (fStamp[UCAL_MILLISECONDS_IN_DAY] >= ((int32_t)kMinimumUserStamp) &&
            newestStamp(UCAL_AM_PM, UCAL_MILLISECOND, kUnset) <= fStamp[UCAL_MILLISECONDS_IN_DAY]) {
        millisInDay = internalGet(UCAL_MILLISECONDS_IN_DAY);
    } else {
        millisInDay = computeMillisInDay();
    }

    UDate t = 0;
    if (fStamp[UCAL_ZONE_OFFSET] >= ((int32_t)kMinimumUserStamp) || fStamp[UCAL_DST_OFFSET] >= ((int32_t)kMinimumUserStamp)) {
        t = millis + millisInDay - (internalGet(UCAL_ZONE_OFFSET) + internalGet(UCAL_DST_OFFSET));
    } else {
        // Compute the time zone offset and DST offset.  There are two potential
        // ambiguities here.  We'll assume a 2:00 am (wall time) switchover time
        // for discussion purposes here.
        //
        // 1. The positive offset change such as transition into DST.
        //    Here, a designated time of 2:00 am - 2:59 am does not actually exist.
        //    For this case, skippedWallTime option specifies the behavior.
        //    For example, 2:30 am is interpreted as;
        //      - WALLTIME_LAST(default): 3:30 am (DST) (interpreting 2:30 am as 31 minutes after 1:59 am (STD))
        //      - WALLTIME_FIRST: 1:30 am (STD) (interpreting 2:30 am as 30 minutes before 3:00 am (DST))
        //      - WALLTIME_NEXT_VALID: 3:00 am (DST) (next valid time after 2:30 am on a wall clock)
        // 2. The negative offset change such as transition out of DST.
        //    Here, a designated time of 1:00 am - 1:59 am can be in standard or DST.  Both are valid
        //    representations (the rep jumps from 1:59:59 DST to 1:00:00 Std).
        //    For this case, repeatedWallTime option specifies the behavior.
        //    For example, 1:30 am is interpreted as;
        //      - WALLTIME_LAST(default): 1:30 am (STD) - latter occurrence
        //      - WALLTIME_FIRST: 1:30 am (DST) - former occurrence
        //
        // In addition to above, when calendar is strict (not default), wall time falls into
        // the skipped time range will be processed as an error case.
        //
        // These special cases are mostly handled in #computeZoneOffset(long), except WALLTIME_NEXT_VALID
        // at positive offset change. The protected method computeZoneOffset(long) is exposed to Calendar
        // subclass implementations and marked as @stable. Strictly speaking, WALLTIME_NEXT_VALID
        // should be also handled in the same place, but we cannot change the code flow without deprecating
        // the protected method.
        //
        // We use the TimeZone object, unless the user has explicitly set the ZONE_OFFSET
        // or DST_OFFSET fields; then we use those fields.

        if (!isLenient() || fSkippedWallTime == UCAL_WALLTIME_NEXT_VALID) {
            // When strict, invalidate a wall time falls into a skipped wall time range.
            // When lenient and skipped wall time option is WALLTIME_NEXT_VALID,
            // the result time will be adjusted to the next valid time (on wall clock).
            int32_t zoneOffset = computeZoneOffset(millis, millisInDay, status);
            UDate tmpTime = millis + millisInDay - zoneOffset;

            int32_t raw, dst;
            fZone->getOffset(tmpTime, FALSE, raw, dst, status);

            if (U_SUCCESS(status)) {
                // zoneOffset != (raw + dst) only when the given wall time fall into
                // a skipped wall time range caused by positive zone offset transition.
                if (zoneOffset != (raw + dst)) {
                    if (!isLenient()) {
                        status = U_ILLEGAL_ARGUMENT_ERROR;
                    } else {
                        U_ASSERT(fSkippedWallTime == UCAL_WALLTIME_NEXT_VALID);
                        // Adjust time to the next valid wall clock time.
                        // At this point, tmpTime is on or after the zone offset transition causing
                        // the skipped time range.
                        UDate immediatePrevTransition;
                        UBool hasTransition = getImmediatePreviousZoneTransition(tmpTime, &immediatePrevTransition, status);
                        if (U_SUCCESS(status) && hasTransition) {
                            t = immediatePrevTransition;
                        }
                    }
                } else {
                    t = tmpTime;
                }
            }
        } else {
            t = millis + millisInDay - computeZoneOffset(millis, millisInDay, status);
        }
    }
    if (U_SUCCESS(status)) {
        internalSetTime(t);
    }
}

/**
 * Find the previous zone transtion near the given time.
 */
UBool Calendar::getImmediatePreviousZoneTransition(UDate base, UDate *transitionTime, UErrorCode& status) const {
    BasicTimeZone *btz = getBasicTimeZone();
    if (btz) {
        TimeZoneTransition trans;
        UBool hasTransition = btz->getPreviousTransition(base, TRUE, trans);
        if (hasTransition) {
            *transitionTime = trans.getTime();
            return TRUE;
        } else {
            // Could not find any transitions.
            // Note: This should never happen.
            status = U_INTERNAL_PROGRAM_ERROR;
        }
    } else {
        // If not BasicTimeZone, return unsupported error for now.
        // TODO: We may support non-BasicTimeZone in future.
        status = U_UNSUPPORTED_ERROR;
    }
    return FALSE;
}

/**
* Compute the milliseconds in the day from the fields.  This is a
* value from 0 to 23:59:59.999 inclusive, unless fields are out of
* range, in which case it can be an arbitrary value.  This value
* reflects local zone wall time.
* @stable ICU 2.0
*/
double Calendar::computeMillisInDay() {
  // Do the time portion of the conversion.

    double millisInDay = 0;

    // Find the best set of fields specifying the time of day.  There
    // are only two possibilities here; the HOUR_OF_DAY or the
    // AM_PM and the HOUR.
    int32_t hourOfDayStamp = fStamp[UCAL_HOUR_OF_DAY];
    int32_t hourStamp = (fStamp[UCAL_HOUR] > fStamp[UCAL_AM_PM])?fStamp[UCAL_HOUR]:fStamp[UCAL_AM_PM];
    int32_t bestStamp = (hourStamp > hourOfDayStamp) ? hourStamp : hourOfDayStamp;

    // Hours
    if (bestStamp != kUnset) {
        if (bestStamp == hourOfDayStamp) {
            // Don't normalize here; let overflow bump into the next period.
            // This is consistent with how we handle other fields.
            millisInDay += internalGet(UCAL_HOUR_OF_DAY);
        } else {
            // Don't normalize here; let overflow bump into the next period.
            // This is consistent with how we handle other fields.
            millisInDay += internalGet(UCAL_HOUR);
            millisInDay += 12 * internalGet(UCAL_AM_PM); // Default works for unset AM_PM
        }
    }

    // We use the fact that unset == 0; we start with millisInDay
    // == HOUR_OF_DAY.
    millisInDay *= 60;
    millisInDay += internalGet(UCAL_MINUTE); // now have minutes
    millisInDay *= 60;
    millisInDay += internalGet(UCAL_SECOND); // now have seconds
    millisInDay *= 1000;
    millisInDay += internalGet(UCAL_MILLISECOND); // now have millis

    return millisInDay;
}

/**
* This method can assume EXTENDED_YEAR has been set.
* @param millis milliseconds of the date fields
* @param millisInDay milliseconds of the time fields; may be out
* or range.
* @stable ICU 2.0
*/
int32_t Calendar::computeZoneOffset(double millis, double millisInDay, UErrorCode &ec) {
    int32_t rawOffset, dstOffset;
    UDate wall = millis + millisInDay;
    BasicTimeZone* btz = getBasicTimeZone();
    if (btz) {
        int duplicatedTimeOpt = (fRepeatedWallTime == UCAL_WALLTIME_FIRST) ? BasicTimeZone::kFormer : BasicTimeZone::kLatter;
        int nonExistingTimeOpt = (fSkippedWallTime == UCAL_WALLTIME_FIRST) ? BasicTimeZone::kLatter : BasicTimeZone::kFormer;
        btz->getOffsetFromLocal(wall, nonExistingTimeOpt, duplicatedTimeOpt, rawOffset, dstOffset, ec);
    } else {
        const TimeZone& tz = getTimeZone();
        // By default, TimeZone::getOffset behaves UCAL_WALLTIME_LAST for both.
        tz.getOffset(wall, TRUE, rawOffset, dstOffset, ec);

        UBool sawRecentNegativeShift = FALSE;
        if (fRepeatedWallTime == UCAL_WALLTIME_FIRST) {
            // Check if the given wall time falls into repeated time range
            UDate tgmt = wall - (rawOffset + dstOffset);

            // Any negative zone transition within last 6 hours?
            // Note: The maximum historic negative zone transition is -3 hours in the tz database.
            // 6 hour window would be sufficient for this purpose.
            int32_t tmpRaw, tmpDst;
            tz.getOffset(tgmt - 6*60*60*1000, FALSE, tmpRaw, tmpDst, ec);
            int32_t offsetDelta = (rawOffset + dstOffset) - (tmpRaw + tmpDst);

            U_ASSERT(offsetDelta < -6*60*60*1000);
            if (offsetDelta < 0) {
                sawRecentNegativeShift = TRUE;
                // Negative shift within last 6 hours. When UCAL_WALLTIME_FIRST is used and the given wall time falls
                // into the repeated time range, use offsets before the transition.
                // Note: If it does not fall into the repeated time range, offsets remain unchanged below.
                tz.getOffset(wall + offsetDelta, TRUE, rawOffset, dstOffset, ec);
            }
        }
        if (!sawRecentNegativeShift && fSkippedWallTime == UCAL_WALLTIME_FIRST) {
            // When skipped wall time option is WALLTIME_FIRST,
            // recalculate offsets from the resolved time (non-wall).
            // When the given wall time falls into skipped wall time,
            // the offsets will be based on the zone offsets AFTER
            // the transition (which means, earliest possibe interpretation).
            UDate tgmt = wall - (rawOffset + dstOffset);
            tz.getOffset(tgmt, FALSE, rawOffset, dstOffset, ec);
        }
    }
    return rawOffset + dstOffset;
}

int32_t Calendar::computeJulianDay()
{
    // We want to see if any of the date fields is newer than the
    // JULIAN_DAY.  If not, then we use JULIAN_DAY.  If so, then we do
    // the normal resolution.  We only use JULIAN_DAY if it has been
    // set by the user.  This makes it possible for the caller to set
    // the calendar to a time and call clear(MONTH) to reset the MONTH
    // to January.  This is legacy behavior.  Without this,
    // clear(MONTH) has no effect, since the internally set JULIAN_DAY
    // is used.
    if (fStamp[UCAL_JULIAN_DAY] >= (int32_t)kMinimumUserStamp) {
        int32_t bestStamp = newestStamp(UCAL_ERA, UCAL_DAY_OF_WEEK_IN_MONTH, kUnset);
        bestStamp = newestStamp(UCAL_YEAR_WOY, UCAL_EXTENDED_YEAR, bestStamp);
        if (bestStamp <= fStamp[UCAL_JULIAN_DAY]) {
            return internalGet(UCAL_JULIAN_DAY);
        }
    }

    UCalendarDateFields bestField = resolveFields(getFieldResolutionTable());
    if (bestField == UCAL_FIELD_COUNT) {
        bestField = UCAL_DAY_OF_MONTH;
    }

    return handleComputeJulianDay(bestField);
}

// -------------------------------------------

int32_t Calendar::handleComputeJulianDay(UCalendarDateFields bestField)  {
    UBool useMonth = (bestField == UCAL_DAY_OF_MONTH ||
        bestField == UCAL_WEEK_OF_MONTH ||
        bestField == UCAL_DAY_OF_WEEK_IN_MONTH);
    int32_t year;

    if (bestField == UCAL_WEEK_OF_YEAR && newerField(UCAL_YEAR_WOY, UCAL_YEAR) == UCAL_YEAR_WOY) {
        year = internalGet(UCAL_YEAR_WOY);
    } else {
        year = handleGetExtendedYear();
    }

    internalSet(UCAL_EXTENDED_YEAR, year);

#if defined (U_DEBUG_CAL)
    fprintf(stderr, "%s:%d: bestField= %s - y=%d\n", __FILE__, __LINE__, fldName(bestField), year);
#endif

    // Get the Julian day of the day BEFORE the start of this year.
    // If useMonth is true, get the day before the start of the month.

    // give calendar subclass a chance to have a default 'first' month
    int32_t month;

    if(isSet(UCAL_MONTH)) {
        month = internalGet(UCAL_MONTH);
    } else {
        month = getDefaultMonthInYear(year);
    }

    int32_t julianDay = handleComputeMonthStart(year, useMonth ? month : 0, useMonth);

    if (bestField == UCAL_DAY_OF_MONTH) {

        // give calendar subclass a chance to have a default 'first' dom
        int32_t dayOfMonth;
        if(isSet(UCAL_DAY_OF_MONTH)) {
            dayOfMonth = internalGet(UCAL_DAY_OF_MONTH,1);
        } else {
            dayOfMonth = getDefaultDayInMonth(year, month);
        }
        return julianDay + dayOfMonth;
    }

    if (bestField == UCAL_DAY_OF_YEAR) {
        return julianDay + internalGet(UCAL_DAY_OF_YEAR);
    }

    int32_t firstDayOfWeek = getFirstDayOfWeek(); // Localized fdw

    // At this point julianDay is the 0-based day BEFORE the first day of
    // January 1, year 1 of the given calendar.  If julianDay == 0, it
    // specifies (Jan. 1, 1) - 1, in whatever calendar we are using (Julian
    // or Gregorian). (or it is before the month we are in, if useMonth is True)

    // At this point we need to process the WEEK_OF_MONTH or
    // WEEK_OF_YEAR, which are similar, or the DAY_OF_WEEK_IN_MONTH.
    // First, perform initial shared computations.  These locate the
    // first week of the period.

    // Get the 0-based localized DOW of day one of the month or year.
    // Valid range 0..6.
    int32_t first = julianDayToDayOfWeek(julianDay + 1) - firstDayOfWeek;
    if (first < 0) {
        first += 7;
    }

    int32_t dowLocal = getLocalDOW();

    // Find the first target DOW (dowLocal) in the month or year.
    // Actually, it may be just before the first of the month or year.
    // It will be an integer from -5..7.
    int32_t date = 1 - first + dowLocal;

    if (bestField == UCAL_DAY_OF_WEEK_IN_MONTH) {
        // Adjust the target DOW to be in the month or year.
        if (date < 1) {
            date += 7;
        }

        // The only trickiness occurs if the day-of-week-in-month is
        // negative.
        int32_t dim = internalGet(UCAL_DAY_OF_WEEK_IN_MONTH, 1);
        if (dim >= 0) {
            date += 7*(dim - 1);

        } else {
            // Move date to the last of this day-of-week in this month,
            // then back up as needed.  If dim==-1, we don't back up at
            // all.  If dim==-2, we back up once, etc.  Don't back up
            // past the first of the given day-of-week in this month.
            // Note that we handle -2, -3, etc. correctly, even though
            // values < -1 are technically disallowed.
            int32_t m = internalGet(UCAL_MONTH, UCAL_JANUARY);
            int32_t monthLength = handleGetMonthLength(year, m);
            date += ((monthLength - date) / 7 + dim + 1) * 7;
        }
    } else {
#if defined (U_DEBUG_CAL)
        fprintf(stderr, "%s:%d - bf= %s\n", __FILE__, __LINE__, fldName(bestField));
#endif

        if(bestField == UCAL_WEEK_OF_YEAR) {  // ------------------------------------- WOY -------------
            if(!isSet(UCAL_YEAR_WOY) ||  // YWOY not set at all or
                ( (resolveFields(kYearPrecedence) != UCAL_YEAR_WOY) // YWOY doesn't have precedence
                && (fStamp[UCAL_YEAR_WOY]!=kInternallySet) ) ) // (excluding where all fields are internally set - then YWOY is used)
            {
                // need to be sure to stay in 'real' year.
                int32_t woy = internalGet(bestField);

                int32_t nextJulianDay = handleComputeMonthStart(year+1, 0, FALSE); // jd of day before jan 1
                int32_t nextFirst = julianDayToDayOfWeek(nextJulianDay + 1) - firstDayOfWeek;

                if (nextFirst < 0) { // 0..6 ldow of Jan 1
                    nextFirst += 7;
                }

                if(woy==1) {  // FIRST WEEK ---------------------------------
#if defined (U_DEBUG_CAL)
                    fprintf(stderr, "%s:%d - woy=%d, yp=%d, nj(%d)=%d, nf=%d", __FILE__, __LINE__,
                        internalGet(bestField), resolveFields(kYearPrecedence), year+1,
                        nextJulianDay, nextFirst);

                    fprintf(stderr, " next: %d DFW,  min=%d   \n", (7-nextFirst), getMinimalDaysInFirstWeek() );
#endif

                    // nextFirst is now the localized DOW of Jan 1  of y-woy+1
                    if((nextFirst > 0) &&   // Jan 1 starts on FDOW
                        (7-nextFirst) >= getMinimalDaysInFirstWeek()) // or enough days in the week
                    {
                        // Jan 1 of (yearWoy+1) is in yearWoy+1 - recalculate JD to next year
#if defined (U_DEBUG_CAL)
                        fprintf(stderr, "%s:%d - was going to move JD from %d to %d [d%d]\n", __FILE__, __LINE__,
                            julianDay, nextJulianDay, (nextJulianDay-julianDay));
#endif
                        julianDay = nextJulianDay;

                        // recalculate 'first' [0-based local dow of jan 1]
                        first = julianDayToDayOfWeek(julianDay + 1) - firstDayOfWeek;
                        if (first < 0) {
                            first += 7;
                        }
                        // recalculate date.
                        date = 1 - first + dowLocal;
                    }
                } else if(woy>=getLeastMaximum(bestField)) {
                    // could be in the last week- find out if this JD would overstep
                    int32_t testDate = date;
                    if ((7 - first) < getMinimalDaysInFirstWeek()) {
                        testDate += 7;
                    }

                    // Now adjust for the week number.
                    testDate += 7 * (woy - 1);

#if defined (U_DEBUG_CAL)
                    fprintf(stderr, "%s:%d - y=%d, y-1=%d doy%d, njd%d (C.F. %d)\n",
                        __FILE__, __LINE__, year, year-1, testDate, julianDay+testDate, nextJulianDay);
#endif
                    if(julianDay+testDate > nextJulianDay) { // is it past Dec 31?  (nextJulianDay is day BEFORE year+1's  Jan 1)
                        // Fire up the calculating engines.. retry YWOY = (year-1)
                        julianDay = handleComputeMonthStart(year-1, 0, FALSE); // jd before Jan 1 of previous year
                        first = julianDayToDayOfWeek(julianDay + 1) - firstDayOfWeek; // 0 based local dow   of first week

                        if(first < 0) { // 0..6
                            first += 7;
                        }
                        date = 1 - first + dowLocal;

#if defined (U_DEBUG_CAL)
                        fprintf(stderr, "%s:%d - date now %d, jd%d, ywoy%d\n",
                            __FILE__, __LINE__, date, julianDay, year-1);
#endif


                    } /* correction needed */
                } /* leastmaximum */
            } /* resolvefields(year) != year_woy */
        } /* bestfield != week_of_year */

        // assert(bestField == WEEK_OF_MONTH || bestField == WEEK_OF_YEAR)
        // Adjust for minimal days in first week
        if ((7 - first) < getMinimalDaysInFirstWeek()) {
            date += 7;
        }

        // Now adjust for the week number.
        date += 7 * (internalGet(bestField) - 1);
    }

    return julianDay + date;
}

int32_t
Calendar::getDefaultMonthInYear(int32_t /*eyear*/)
{
    return 0;
}

int32_t
Calendar::getDefaultDayInMonth(int32_t /*eyear*/, int32_t /*month*/)
{
    return 1;
}


int32_t Calendar::getLocalDOW()
{
  // Get zero-based localized DOW, valid range 0..6.  This is the DOW
    // we are looking for.
    int32_t dowLocal = 0;
    switch (resolveFields(kDOWPrecedence)) {
    case UCAL_DAY_OF_WEEK:
        dowLocal = internalGet(UCAL_DAY_OF_WEEK) - fFirstDayOfWeek;
        break;
    case UCAL_DOW_LOCAL:
        dowLocal = internalGet(UCAL_DOW_LOCAL) - 1;
        break;
    default:
        break;
    }
    dowLocal = dowLocal % 7;
    if (dowLocal < 0) {
        dowLocal += 7;
    }
    return dowLocal;
}

int32_t Calendar::handleGetExtendedYearFromWeekFields(int32_t yearWoy, int32_t woy)
{
    // We have UCAL_YEAR_WOY and UCAL_WEEK_OF_YEAR - from those, determine
    // what year we fall in, so that other code can set it properly.
    // (code borrowed from computeWeekFields and handleComputeJulianDay)
    //return yearWoy;

    // First, we need a reliable DOW.
    UCalendarDateFields bestField = resolveFields(kDatePrecedence); // !! Note: if subclasses have a different table, they should override handleGetExtendedYearFromWeekFields

    // Now, a local DOW
    int32_t dowLocal = getLocalDOW(); // 0..6
    int32_t firstDayOfWeek = getFirstDayOfWeek(); // Localized fdw
    int32_t jan1Start = handleComputeMonthStart(yearWoy, 0, FALSE);
    int32_t nextJan1Start = handleComputeMonthStart(yearWoy+1, 0, FALSE); // next year's Jan1 start

    // At this point julianDay is the 0-based day BEFORE the first day of
    // January 1, year 1 of the given calendar.  If julianDay == 0, it
    // specifies (Jan. 1, 1) - 1, in whatever calendar we are using (Julian
    // or Gregorian). (or it is before the month we are in, if useMonth is True)

    // At this point we need to process the WEEK_OF_MONTH or
    // WEEK_OF_YEAR, which are similar, or the DAY_OF_WEEK_IN_MONTH.
    // First, perform initial shared computations.  These locate the
    // first week of the period.

    // Get the 0-based localized DOW of day one of the month or year.
    // Valid range 0..6.
    int32_t first = julianDayToDayOfWeek(jan1Start + 1) - firstDayOfWeek;
    if (first < 0) {
        first += 7;
    }

    //// (nextFirst was not used below)
    // int32_t nextFirst = julianDayToDayOfWeek(nextJan1Start + 1) - firstDayOfWeek;
    // if (nextFirst < 0) {
    //     nextFirst += 7;
    //}

    int32_t minDays = getMinimalDaysInFirstWeek();
    UBool jan1InPrevYear = FALSE;  // January 1st in the year of WOY is the 1st week?  (i.e. first week is < minimal )
    //UBool nextJan1InPrevYear = FALSE; // January 1st of Year of WOY + 1 is in the first week?

    if((7 - first) < minDays) {
        jan1InPrevYear = TRUE;
    }

    //   if((7 - nextFirst) < minDays) {
    //     nextJan1InPrevYear = TRUE;
    //   }

    switch(bestField) {
    case UCAL_WEEK_OF_YEAR:
        if(woy == 1) {
            if(jan1InPrevYear == TRUE) {
                // the first week of January is in the previous year
                // therefore WOY1 is always solidly within yearWoy
                return yearWoy;
            } else {
                // First WOY is split between two years
                if( dowLocal < first) { // we are prior to Jan 1
                    return yearWoy-1; // previous year
                } else {
                    return yearWoy; // in this year
                }
            }
        } else if(woy >= getLeastMaximum(bestField)) {
            // we _might_ be in the last week..
            int32_t jd =  // Calculate JD of our target day:
                jan1Start +  // JD of Jan 1
                (7-first) + //  days in the first week (Jan 1.. )
                (woy-1)*7 + // add the weeks of the year
                dowLocal;   // the local dow (0..6) of last week
            if(jan1InPrevYear==FALSE) {
                jd -= 7; // woy already includes Jan 1's week.
            }

            if( (jd+1) >= nextJan1Start ) {
                // we are in week 52 or 53 etc. - actual year is yearWoy+1
                return yearWoy+1;
            } else {
                // still in yearWoy;
                return yearWoy;
            }
        } else {
            // we're not possibly in the last week -must be ywoy
            return yearWoy;
        }

    case UCAL_DATE:
        if((internalGet(UCAL_MONTH)==0) &&
            (woy >= getLeastMaximum(UCAL_WEEK_OF_YEAR))) {
                return yearWoy+1; // month 0, late woy = in the next year
            } else if(woy==1) {
                //if(nextJan1InPrevYear) {
                if(internalGet(UCAL_MONTH)==0) {
                    return yearWoy;
                } else {
                    return yearWoy-1;
                }
                //}
            }

            //(internalGet(UCAL_DATE) <= (7-first)) /* && in minDow  */ ) {
            //within 1st week and in this month..
            //return yearWoy+1;
            return yearWoy;

    default: // assume the year is appropriate
        return yearWoy;
    }
}

int32_t Calendar::handleGetMonthLength(int32_t extendedYear, int32_t month) const
{
    return handleComputeMonthStart(extendedYear, month+1, TRUE) -
        handleComputeMonthStart(extendedYear, month, TRUE);
}

int32_t Calendar::handleGetYearLength(int32_t eyear) const  {
    return handleComputeMonthStart(eyear+1, 0, FALSE) -
        handleComputeMonthStart(eyear, 0, FALSE);
}

int32_t
Calendar::getActualMaximum(UCalendarDateFields field, UErrorCode& status) const
{
    int32_t result;
    switch (field) {
    case UCAL_DATE:
        {
            if(U_FAILURE(status)) return 0;
            Calendar *cal = clone();
            if(!cal) { status = U_MEMORY_ALLOCATION_ERROR; return 0; }
            cal->setLenient(TRUE);
            cal->prepareGetActual(field,FALSE,status);
            result = handleGetMonthLength(cal->get(UCAL_EXTENDED_YEAR, status), cal->get(UCAL_MONTH, status));
            delete cal;
        }
        break;

    case UCAL_DAY_OF_YEAR:
        {
            if(U_FAILURE(status)) return 0;
            Calendar *cal = clone();
            if(!cal) { status = U_MEMORY_ALLOCATION_ERROR; return 0; }
            cal->setLenient(TRUE);
            cal->prepareGetActual(field,FALSE,status);
            result = handleGetYearLength(cal->get(UCAL_EXTENDED_YEAR, status));
            delete cal;
        }
        break;

    case UCAL_DAY_OF_WEEK:
    case UCAL_AM_PM:
    case UCAL_HOUR:
    case UCAL_HOUR_OF_DAY:
    case UCAL_MINUTE:
    case UCAL_SECOND:
    case UCAL_MILLISECOND:
    case UCAL_ZONE_OFFSET:
    case UCAL_DST_OFFSET:
    case UCAL_DOW_LOCAL:
    case UCAL_JULIAN_DAY:
    case UCAL_MILLISECONDS_IN_DAY:
        // These fields all have fixed minima/maxima
        result = getMaximum(field);
        break;

    default:
        // For all other fields, do it the hard way....
        result = getActualHelper(field, getLeastMaximum(field), getMaximum(field),status);
        break;
    }
    return result;
}


/**
* Prepare this calendar for computing the actual minimum or maximum.
* This method modifies this calendar's fields; it is called on a
* temporary calendar.
*
* <p>Rationale: The semantics of getActualXxx() is to return the
* maximum or minimum value that the given field can take, taking into
* account other relevant fields.  In general these other fields are
* larger fields.  For example, when computing the actual maximum
* DATE, the current value of DATE itself is ignored,
* as is the value of any field smaller.
*
* <p>The time fields all have fixed minima and maxima, so we don't
* need to worry about them.  This also lets us set the
* MILLISECONDS_IN_DAY to zero to erase any effects the time fields
* might have when computing date fields.
*
* <p>DAY_OF_WEEK is adjusted specially for the WEEK_OF_MONTH and
* WEEK_OF_YEAR fields to ensure that they are computed correctly.
* @internal
*/
void Calendar::prepareGetActual(UCalendarDateFields field, UBool isMinimum, UErrorCode &status)
{
    set(UCAL_MILLISECONDS_IN_DAY, 0);

    switch (field) {
    case UCAL_YEAR:
    case UCAL_EXTENDED_YEAR:
        set(UCAL_DAY_OF_YEAR, getGreatestMinimum(UCAL_DAY_OF_YEAR));
        break;

    case UCAL_YEAR_WOY:
        set(UCAL_WEEK_OF_YEAR, getGreatestMinimum(UCAL_WEEK_OF_YEAR));
        U_FALLTHROUGH;
    case UCAL_MONTH:
        set(UCAL_DATE, getGreatestMinimum(UCAL_DATE));
        break;

    case UCAL_DAY_OF_WEEK_IN_MONTH:
        // For dowim, the maximum occurs for the DOW of the first of the
        // month.
        set(UCAL_DATE, 1);
        set(UCAL_DAY_OF_WEEK, get(UCAL_DAY_OF_WEEK, status)); // Make this user set
        break;

    case UCAL_WEEK_OF_MONTH:
    case UCAL_WEEK_OF_YEAR:
        // If we're counting weeks, set the day of the week to either the
        // first or last localized DOW.  We know the last week of a month
        // or year will contain the first day of the week, and that the
        // first week will contain the last DOW.
        {
            int32_t dow = fFirstDayOfWeek;
            if (isMinimum) {
                dow = (dow + 6) % 7; // set to last DOW
                if (dow < UCAL_SUNDAY) {
                    dow += 7;
                }
            }
#if defined (U_DEBUG_CAL)
            fprintf(stderr, "prepareGetActualHelper(WOM/WOY) - dow=%d\n", dow);
#endif
            set(UCAL_DAY_OF_WEEK, dow);
        }
        break;
    default:
        break;
    }

    // Do this last to give it the newest time stamp
    set(field, getGreatestMinimum(field));
}

int32_t Calendar::getActualHelper(UCalendarDateFields field, int32_t startValue, int32_t endValue, UErrorCode &status) const
{
#if defined (U_DEBUG_CAL)
    fprintf(stderr, "getActualHelper(%d,%d .. %d, %s)\n", field, startValue, endValue, u_errorName(status));
#endif
    if (startValue == endValue) {
        // if we know that the maximum value is always the same, just return it
        return startValue;
    }

    int32_t delta = (endValue > startValue) ? 1 : -1;

    // clone the calendar so we don't mess with the real one, and set it to
    // accept anything for the field values
    if(U_FAILURE(status)) return startValue;
    Calendar *work = clone();
    if(!work) { status = U_MEMORY_ALLOCATION_ERROR; return startValue; }

    // need to resolve time here, otherwise, fields set for actual limit
    // may cause conflict with fields previously set (but not yet resolved).
    work->complete(status);

    work->setLenient(TRUE);
    work->prepareGetActual(field, delta < 0, status);

    // now try each value from the start to the end one by one until
    // we get a value that normalizes to another value.  The last value that
    // normalizes to itself is the actual maximum for the current date
    work->set(field, startValue);

    // prepareGetActual sets the first day of week in the same week with
    // the first day of a month.  Unlike WEEK_OF_YEAR, week number for the
    // week which contains days from both previous and current month is
    // not unique.  For example, last several days in the previous month
    // is week 5, and the rest of week is week 1.
    int32_t result = startValue;
    if ((work->get(field, status) != startValue
         && field != UCAL_WEEK_OF_MONTH && delta > 0 ) || U_FAILURE(status)) {
#if defined (U_DEBUG_CAL)
        fprintf(stderr, "getActualHelper(fld %d) - got  %d (not %d) - %s\n", field, work->get(field,status), startValue, u_errorName(status));
#endif
    } else {
        do {
            startValue += delta;
            work->add(field, delta, status);
            if (work->get(field, status) != startValue || U_FAILURE(status)) {
#if defined (U_DEBUG_CAL)
                fprintf(stderr, "getActualHelper(fld %d) - got  %d (not %d), BREAK - %s\n", field, work->get(field,status), startValue, u_errorName(status));
#endif
                break;
            }
            result = startValue;
        } while (startValue != endValue);
    }
    delete work;
#if defined (U_DEBUG_CAL)
    fprintf(stderr, "getActualHelper(%d) = %d\n", field, result);
#endif
    return result;
}




// -------------------------------------

void
Calendar::setWeekData(const Locale& desiredLocale, const char *type, UErrorCode& status)
{

    if (U_FAILURE(status)) return;

    fFirstDayOfWeek = UCAL_SUNDAY;
    fMinimalDaysInFirstWeek = 1;
    fWeekendOnset = UCAL_SATURDAY;
    fWeekendOnsetMillis = 0;
    fWeekendCease = UCAL_SUNDAY;
    fWeekendCeaseMillis = 86400000; // 24*60*60*1000

    // Since week and weekend data is territory based instead of language based,
    // we may need to tweak the locale that we are using to try to get the appropriate
    // values, using the following logic:
    // 1). If the locale has a language but no territory, use the territory as defined by
    //     the likely subtags.
    // 2). If the locale has a script designation then we ignore it,
    //     then remove it ( i.e. "en_Latn_US" becomes "en_US" )

    UErrorCode myStatus = U_ZERO_ERROR;

    Locale min(desiredLocale);
    min.minimizeSubtags(myStatus);
    Locale useLocale;
    if ( uprv_strlen(desiredLocale.getCountry()) == 0 ||
         (uprv_strlen(desiredLocale.getScript()) > 0 && uprv_strlen(min.getScript()) == 0) ) {
        myStatus = U_ZERO_ERROR;
        Locale max(desiredLocale);
        max.addLikelySubtags(myStatus);
        useLocale = Locale(max.getLanguage(),max.getCountry());
    } else {
        useLocale = desiredLocale;
    }

    /* The code here is somewhat of a hack, since week data and weekend data aren't really tied to
       a specific calendar, they aren't truly locale data.  But this is the only place where valid and
       actual locale can be set, so we take a shot at it here by loading a representative resource
       from the calendar data.  The code used to use the dateTimeElements resource to get first day
       of week data, but this was moved to supplemental data under ticket 7755. (JCE) */

    // Get the monthNames resource bundle for the calendar 'type'. Fallback to gregorian if the resource is not
    // found.
    LocalUResourceBundlePointer calData(ures_open(NULL, useLocale.getBaseName(), &status));
    ures_getByKey(calData.getAlias(), gCalendar, calData.getAlias(), &status);

    LocalUResourceBundlePointer monthNames;
    if (type != NULL && *type != '\0' && uprv_strcmp(type, gGregorian) != 0) {
        monthNames.adoptInstead(ures_getByKeyWithFallback(calData.getAlias(), type, NULL, &status));
        ures_getByKeyWithFallback(monthNames.getAlias(), gMonthNames,
                                  monthNames.getAlias(), &status);
    }

    if (monthNames.isNull() || status == U_MISSING_RESOURCE_ERROR) {
        status = U_ZERO_ERROR;
        monthNames.adoptInstead(ures_getByKeyWithFallback(calData.getAlias(), gGregorian,
                                                          monthNames.orphan(), &status));
        ures_getByKeyWithFallback(monthNames.getAlias(), gMonthNames,
                                  monthNames.getAlias(), &status);
    }

    if (U_SUCCESS(status)) {
        U_LOCALE_BASED(locBased,*this);
        locBased.setLocaleIDs(ures_getLocaleByType(monthNames.getAlias(), ULOC_VALID_LOCALE, &status),
                              ures_getLocaleByType(monthNames.getAlias(), ULOC_ACTUAL_LOCALE, &status));
    } else {
        status = U_USING_FALLBACK_WARNING;
        return;
    }

    char region[ULOC_COUNTRY_CAPACITY];
    (void)ulocimp_getRegionForSupplementalData(desiredLocale.getName(), TRUE, region, sizeof(region), &status);

    // Read week data values from supplementalData week data
    UResourceBundle *rb = ures_openDirect(NULL, "supplementalData", &status);
    ures_getByKey(rb, "weekData", rb, &status);
    UResourceBundle *weekData = ures_getByKey(rb, region, NULL, &status);
    if (status == U_MISSING_RESOURCE_ERROR && rb != NULL) {
        status = U_ZERO_ERROR;
        weekData = ures_getByKey(rb, "001", NULL, &status);
    }

    if (U_FAILURE(status)) {
        status = U_USING_FALLBACK_WARNING;
    } else {
        int32_t arrLen;
        const int32_t *weekDataArr = ures_getIntVector(weekData,&arrLen,&status);
        if( U_SUCCESS(status) && arrLen == 6
                && 1 <= weekDataArr[0] && weekDataArr[0] <= 7
                && 1 <= weekDataArr[1] && weekDataArr[1] <= 7
                && 1 <= weekDataArr[2] && weekDataArr[2] <= 7
                && 1 <= weekDataArr[4] && weekDataArr[4] <= 7) {
            fFirstDayOfWeek = (UCalendarDaysOfWeek)weekDataArr[0];
            fMinimalDaysInFirstWeek = (uint8_t)weekDataArr[1];
            fWeekendOnset = (UCalendarDaysOfWeek)weekDataArr[2];
            fWeekendOnsetMillis = weekDataArr[3];
            fWeekendCease = (UCalendarDaysOfWeek)weekDataArr[4];
            fWeekendCeaseMillis = weekDataArr[5];
        } else {
            status = U_INVALID_FORMAT_ERROR;
        }
    }
    ures_close(weekData);
    ures_close(rb);
}

/**
* Recompute the time and update the status fields isTimeSet
* and areFieldsSet.  Callers should check isTimeSet and only
* call this method if isTimeSet is false.
*/
void
Calendar::updateTime(UErrorCode& status)
{
    computeTime(status);
    if(U_FAILURE(status))
        return;

    // If we are lenient, we need to recompute the fields to normalize
    // the values.  Also, if we haven't set all the fields yet (i.e.,
    // in a newly-created object), we need to fill in the fields. [LIU]
    if (isLenient() || ! fAreAllFieldsSet)
        fAreFieldsSet = FALSE;

    fIsTimeSet = TRUE;
    fAreFieldsVirtuallySet = FALSE;
}

Locale
Calendar::getLocale(ULocDataLocaleType type, UErrorCode& status) const {
    U_LOCALE_BASED(locBased, *this);
    return locBased.getLocale(type, status);
}

const char *
Calendar::getLocaleID(ULocDataLocaleType type, UErrorCode& status) const {
    U_LOCALE_BASED(locBased, *this);
    return locBased.getLocaleID(type, status);
}

void
Calendar::recalculateStamp() {
    int32_t index;
    int32_t currentValue;
    int32_t j, i;

    fNextStamp = 1;

    for (j = 0; j < UCAL_FIELD_COUNT; j++) {
        currentValue = STAMP_MAX;
        index = -1;
        for (i = 0; i < UCAL_FIELD_COUNT; i++) {
            if (fStamp[i] > fNextStamp && fStamp[i] < currentValue) {
                currentValue = fStamp[i];
                index = i;
            }
        }

        if (index >= 0) {
            fStamp[index] = ++fNextStamp;
        } else {
            break;
        }
    }
    fNextStamp++;
}

// Deprecated function. This doesn't need to be inline.
void
Calendar::internalSet(EDateFields field, int32_t value)
{
    internalSet((UCalendarDateFields) field, value);
}

BasicTimeZone*
Calendar::getBasicTimeZone(void) const {
    if (dynamic_cast<const OlsonTimeZone *>(fZone) != NULL
        || dynamic_cast<const SimpleTimeZone *>(fZone) != NULL
        || dynamic_cast<const RuleBasedTimeZone *>(fZone) != NULL
        || dynamic_cast<const VTimeZone *>(fZone) != NULL) {
        return (BasicTimeZone*)fZone;
    }
    return NULL;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */


//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2003 - 2009, International Business Machines Corporation and  *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING


   //Math

U_NAMESPACE_BEGIN

static const int32_t CECAL_LIMITS[UCAL_FIELD_COUNT][4] = {
    // Minimum  Greatest    Least  Maximum
    //           Minimum  Maximum
    {        0,        0,        1,        1}, // ERA
    {        1,        1,  5000000,  5000000}, // YEAR
    {        0,        0,       12,       12}, // MONTH
    {        1,        1,       52,       53}, // WEEK_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // WEEK_OF_MONTH
    {        1,        1,        5,       30}, // DAY_OF_MONTH
    {        1,        1,      365,      366}, // DAY_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DAY_OF_WEEK
    {       -1,       -1,        1,        5}, // DAY_OF_WEEK_IN_MONTH
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // AM_PM
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR_OF_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MINUTE
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // SECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // ZONE_OFFSET
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DST_OFFSET
    { -5000000, -5000000,  5000000,  5000000}, // YEAR_WOY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DOW_LOCAL
    { -5000000, -5000000,  5000000,  5000000}, // EXTENDED_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // JULIAN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECONDS_IN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // IS_LEAP_MONTH
};

//-------------------------------------------------------------------------
// Constructors...
//-------------------------------------------------------------------------

CECalendar::CECalendar(const Locale& aLocale, UErrorCode& success)
:   Calendar(TimeZone::createDefault(), aLocale, success)
{
    setTimeInMillis(getNow(), success);
}

CECalendar::CECalendar (const CECalendar& other)
:   Calendar(other)
{
}

CECalendar::~CECalendar()
{
}

CECalendar&
CECalendar::operator=(const CECalendar& right)
{
    Calendar::operator=(right);
    return *this;
}

//-------------------------------------------------------------------------
// Calendar framework
//-------------------------------------------------------------------------

int32_t
CECalendar::handleComputeMonthStart(int32_t eyear,int32_t emonth, UBool /*useMonth*/) const
{
    return ceToJD(eyear, emonth, 0, getJDEpochOffset());
}

int32_t
CECalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const
{
    return CECAL_LIMITS[field][limitType];
}

UBool
CECalendar::inDaylightTime(UErrorCode& status) const
{
    if (U_FAILURE(status) || !getTimeZone().useDaylightTime()) {
        return FALSE;
    }

    // Force an update of the state of the Calendar.
    ((CECalendar*)this)->complete(status); // cast away const

    return (UBool)(U_SUCCESS(status) ? (internalGet(UCAL_DST_OFFSET) != 0) : FALSE);
}

UBool
CECalendar::haveDefaultCentury() const
{
    return TRUE;
}

//-------------------------------------------------------------------------
// Calendar system Conversion methods...
//-------------------------------------------------------------------------
int32_t
CECalendar::ceToJD(int32_t year, int32_t month, int32_t date, int32_t jdEpochOffset)
{
    // handle month > 12, < 0 (e.g. from add/set)
    if ( month >= 0 ) {
        year += month/13;
        month %= 13;
    } else {
        ++month;
        year += month/13 - 1;
        month = month%13 + 12;
    }
    return (int32_t) (
        jdEpochOffset                   // difference from Julian epoch to 1,1,1
        + 365 * year                    // number of days from years
        + ClockMath::floorDivide(year, 4)    // extra day of leap year
        + 30 * month                    // number of days from months (months are 0-based)
        + date - 1                      // number of days for present month (1 based)
        );
}

void
CECalendar::jdToCE(int32_t julianDay, int32_t jdEpochOffset, int32_t& year, int32_t& month, int32_t& day)
{
    int32_t c4; // number of 4 year cycle (1461 days)
    int32_t r4; // remainder of 4 year cycle, always positive

    c4 = ClockMath::floorDivide(julianDay - jdEpochOffset, 1461, r4);

    year = 4 * c4 + (r4/365 - r4/1460); // 4 * <number of 4year cycle> + <years within the last cycle>

    int32_t doy = (r4 == 1460) ? 365 : (r4 % 365); // days in present year

    month = doy / 30;       // 30 -> Coptic/Ethiopic month length up to 12th month
    day = (doy % 30) + 1;   // 1-based days in a month
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ******************************************************************************
 * Copyright (C) 2007-2014, International Business Machines Corporation
 * and others. All Rights Reserved.
 ******************************************************************************
 *
 * File CHNSECAL.CPP
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   9/18/2007  ajmacher         ported from java ChineseCalendar
 *****************************************************************************
 */



#if !UCONFIG_NO_FORMATTING


#include <float.h>
 // Math
 // CalendarAstronomer




// Debugging
#ifdef U_DEBUG_CHNSECAL
# include <stdio.h>
# include <stdarg.h>
static void debug_chnsecal_loc(const char *f, int32_t l)
{
    fprintf(stderr, "%s:%d: ", f, l);
}

static void debug_chnsecal_msg(const char *pat, ...)
{
    va_list ap;
    va_start(ap, pat);
    vfprintf(stderr, pat, ap);
    fflush(stderr);
}
// must use double parens, i.e.:  U_DEBUG_CHNSECAL_MSG(("four is: %d",4));
#define U_DEBUG_CHNSECAL_MSG(x) {debug_chnsecal_loc(__FILE__,__LINE__);debug_chnsecal_msg x;}
#else
#define U_DEBUG_CHNSECAL_MSG(x)
#endif


// --- The cache --
static icu::UMutex astroLock;
static icu::CalendarAstronomer *gChineseCalendarAstro = NULL;

// Lazy Creation & Access synchronized by class CalendarCache with a mutex.
static icu::CalendarCache *gChineseCalendarWinterSolsticeCache = NULL;
static icu::CalendarCache *gChineseCalendarNewYearCache = NULL;

static icu::TimeZone *gChineseCalendarZoneAstroCalc = NULL;
static icu::UInitOnce gChineseCalendarZoneAstroCalcInitOnce = U_INITONCE_INITIALIZER;

/**
 * The start year of the Chinese calendar, the 61st year of the reign
 * of Huang Di.  Some sources use the first year of his reign,
 * resulting in EXTENDED_YEAR values 60 years greater and ERA (cycle)
 * values one greater.
 */
static const int32_t CHINESE_EPOCH_YEAR = -2636; // Gregorian year

/**
 * The offset from GMT in milliseconds at which we perform astronomical
 * computations.  Some sources use a different historically accurate
 * offset of GMT+7:45:40 for years before 1929; we do not do this.
 */
static const int32_t CHINA_OFFSET = 8 * kOneHour;

/**
 * Value to be added or subtracted from the local days of a new moon to
 * get close to the next or prior new moon, but not cross it.  Must be
 * >= 1 and < CalendarAstronomer.SYNODIC_MONTH.
 */
static const int32_t SYNODIC_GAP = 25;


U_CDECL_BEGIN
static UBool calendar_chinese_cleanup(void) {
    if (gChineseCalendarAstro) {
        delete gChineseCalendarAstro;
        gChineseCalendarAstro = NULL;
    }
    if (gChineseCalendarWinterSolsticeCache) {
        delete gChineseCalendarWinterSolsticeCache;
        gChineseCalendarWinterSolsticeCache = NULL;
    }
    if (gChineseCalendarNewYearCache) {
        delete gChineseCalendarNewYearCache;
        gChineseCalendarNewYearCache = NULL;
    }
    if (gChineseCalendarZoneAstroCalc) {
        delete gChineseCalendarZoneAstroCalc;
        gChineseCalendarZoneAstroCalc = NULL;
    }
    gChineseCalendarZoneAstroCalcInitOnce.reset();
    return TRUE;
}
U_CDECL_END

U_NAMESPACE_BEGIN


// Implementation of the ChineseCalendar class


//-------------------------------------------------------------------------
// Constructors...
//-------------------------------------------------------------------------


ChineseCalendar* ChineseCalendar::clone() const {
    return new ChineseCalendar(*this);
}

ChineseCalendar::ChineseCalendar(const Locale& aLocale, UErrorCode& success)
:   Calendar(TimeZone::createDefault(), aLocale, success),
    isLeapYear(FALSE),
    fEpochYear(CHINESE_EPOCH_YEAR),
    fZoneAstroCalc(getChineseCalZoneAstroCalc())
{
    setTimeInMillis(getNow(), success); // Call this again now that the vtable is set up properly.
}

ChineseCalendar::ChineseCalendar(const Locale& aLocale, int32_t epochYear,
                                const TimeZone* zoneAstroCalc, UErrorCode &success)
:   Calendar(TimeZone::createDefault(), aLocale, success),
    isLeapYear(FALSE),
    fEpochYear(epochYear),
    fZoneAstroCalc(zoneAstroCalc)
{
    setTimeInMillis(getNow(), success); // Call this again now that the vtable is set up properly.
}

ChineseCalendar::ChineseCalendar(const ChineseCalendar& other) : Calendar(other) {
    isLeapYear = other.isLeapYear;
    fEpochYear = other.fEpochYear;
    fZoneAstroCalc = other.fZoneAstroCalc;
}

ChineseCalendar::~ChineseCalendar()
{
}

const char *ChineseCalendar::getType() const {
    return "chinese";
}

static void U_CALLCONV initChineseCalZoneAstroCalc() {
    gChineseCalendarZoneAstroCalc = new SimpleTimeZone(CHINA_OFFSET, UNICODE_STRING_SIMPLE("CHINA_ZONE") );
    ucln_i18n_registerCleanup(UCLN_I18N_CHINESE_CALENDAR, calendar_chinese_cleanup);
}

const TimeZone* ChineseCalendar::getChineseCalZoneAstroCalc(void) const {
    umtx_initOnce(gChineseCalendarZoneAstroCalcInitOnce, &initChineseCalZoneAstroCalc);
    return gChineseCalendarZoneAstroCalc;
}

//-------------------------------------------------------------------------
// Minimum / Maximum access functions
//-------------------------------------------------------------------------


static const int32_t CHNSECAL_LIMITS[UCAL_FIELD_COUNT][4] = {
    // Minimum  Greatest     Least    Maximum
    //           Minimum   Maximum
    {        1,        1,    83333,    83333}, // ERA
    {        1,        1,       60,       60}, // YEAR
    {        0,        0,       11,       11}, // MONTH
    {        1,        1,       50,       55}, // WEEK_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // WEEK_OF_MONTH
    {        1,        1,       29,       30}, // DAY_OF_MONTH
    {        1,        1,      353,      385}, // DAY_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DAY_OF_WEEK
    {       -1,       -1,        5,        5}, // DAY_OF_WEEK_IN_MONTH
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // AM_PM
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR_OF_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MINUTE
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // SECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // ZONE_OFFSET
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DST_OFFSET
    { -5000000, -5000000,  5000000,  5000000}, // YEAR_WOY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DOW_LOCAL
    { -5000000, -5000000,  5000000,  5000000}, // EXTENDED_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // JULIAN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECONDS_IN_DAY
    {        0,        0,        1,        1}, // IS_LEAP_MONTH
};


/**
* @draft ICU 2.4
*/
int32_t ChineseCalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const {
    return CHNSECAL_LIMITS[field][limitType];
}


//----------------------------------------------------------------------
// Calendar framework
//----------------------------------------------------------------------

/**
 * Implement abstract Calendar method to return the extended year
 * defined by the current fields.  This will use either the ERA and
 * YEAR field as the cycle and year-of-cycle, or the EXTENDED_YEAR
 * field as the continuous year count, depending on which is newer.
 * @stable ICU 2.8
 */
int32_t ChineseCalendar::handleGetExtendedYear() {
    int32_t year;
    if (newestStamp(UCAL_ERA, UCAL_YEAR, kUnset) <= fStamp[UCAL_EXTENDED_YEAR]) {
        year = internalGet(UCAL_EXTENDED_YEAR, 1); // Default to year 1
    } else {
        int32_t cycle = internalGet(UCAL_ERA, 1) - 1; // 0-based cycle
        // adjust to the instance specific epoch
        year = cycle * 60 + internalGet(UCAL_YEAR, 1) - (fEpochYear - CHINESE_EPOCH_YEAR);
    }
    return year;
}

/**
 * Override Calendar method to return the number of days in the given
 * extended year and month.
 *
 * <p>Note: This method also reads the IS_LEAP_MONTH field to determine
 * whether or not the given month is a leap month.
 * @stable ICU 2.8
 */
int32_t ChineseCalendar::handleGetMonthLength(int32_t extendedYear, int32_t month) const {
    int32_t thisStart = handleComputeMonthStart(extendedYear, month, TRUE) -
        kEpochStartAsJulianDay + 1; // Julian day -> local days
    int32_t nextStart = newMoonNear(thisStart + SYNODIC_GAP, TRUE);
    return nextStart - thisStart;
}

/**
 * Override Calendar to compute several fields specific to the Chinese
 * calendar system.  These are:
 *
 * <ul><li>ERA
 * <li>YEAR
 * <li>MONTH
 * <li>DAY_OF_MONTH
 * <li>DAY_OF_YEAR
 * <li>EXTENDED_YEAR</ul>
 *
 * The DAY_OF_WEEK and DOW_LOCAL fields are already set when this
 * method is called.  The getGregorianXxx() methods return Gregorian
 * calendar equivalents for the given Julian day.
 *
 * <p>Compute the ChineseCalendar-specific field IS_LEAP_MONTH.
 * @stable ICU 2.8
 */
void ChineseCalendar::handleComputeFields(int32_t julianDay, UErrorCode &/*status*/) {

    computeChineseFields(julianDay - kEpochStartAsJulianDay, // local days
                         getGregorianYear(), getGregorianMonth(),
                         TRUE); // set all fields
}

/**
 * Field resolution table that incorporates IS_LEAP_MONTH.
 */
const UFieldResolutionTable ChineseCalendar::CHINESE_DATE_PRECEDENCE[] =
{
    {
        { UCAL_DAY_OF_MONTH, kResolveSTOP },
        { UCAL_WEEK_OF_YEAR, UCAL_DAY_OF_WEEK, kResolveSTOP },
        { UCAL_WEEK_OF_MONTH, UCAL_DAY_OF_WEEK, kResolveSTOP },
        { UCAL_DAY_OF_WEEK_IN_MONTH, UCAL_DAY_OF_WEEK, kResolveSTOP },
        { UCAL_WEEK_OF_YEAR, UCAL_DOW_LOCAL, kResolveSTOP },
        { UCAL_WEEK_OF_MONTH, UCAL_DOW_LOCAL, kResolveSTOP },
        { UCAL_DAY_OF_WEEK_IN_MONTH, UCAL_DOW_LOCAL, kResolveSTOP },
        { UCAL_DAY_OF_YEAR, kResolveSTOP },
        { kResolveRemap | UCAL_DAY_OF_MONTH, UCAL_IS_LEAP_MONTH, kResolveSTOP },
        { kResolveSTOP }
    },
    {
        { UCAL_WEEK_OF_YEAR, kResolveSTOP },
        { UCAL_WEEK_OF_MONTH, kResolveSTOP },
        { UCAL_DAY_OF_WEEK_IN_MONTH, kResolveSTOP },
        { kResolveRemap | UCAL_DAY_OF_WEEK_IN_MONTH, UCAL_DAY_OF_WEEK, kResolveSTOP },
        { kResolveRemap | UCAL_DAY_OF_WEEK_IN_MONTH, UCAL_DOW_LOCAL, kResolveSTOP },
        { kResolveSTOP }
    },
    {{kResolveSTOP}}
};

/**
 * Override Calendar to add IS_LEAP_MONTH to the field resolution
 * table.
 * @stable ICU 2.8
 */
const UFieldResolutionTable* ChineseCalendar::getFieldResolutionTable() const {
    return CHINESE_DATE_PRECEDENCE;
}

/**
 * Return the Julian day number of day before the first day of the
 * given month in the given extended year.
 *
 * <p>Note: This method reads the IS_LEAP_MONTH field to determine
 * whether the given month is a leap month.
 * @param eyear the extended year
 * @param month the zero-based month.  The month is also determined
 * by reading the IS_LEAP_MONTH field.
 * @return the Julian day number of the day before the first
 * day of the given month and year
 * @stable ICU 2.8
 */
int32_t ChineseCalendar::handleComputeMonthStart(int32_t eyear, int32_t month, UBool useMonth) const {

    ChineseCalendar *nonConstThis = (ChineseCalendar*)this; // cast away const

    // If the month is out of range, adjust it into range, and
    // modify the extended year value accordingly.
    if (month < 0 || month > 11) {
        double m = month;
        eyear += (int32_t)ClockMath::floorDivide(m, 12.0, m);
        month = (int32_t)m;
    }

    int32_t gyear = eyear + fEpochYear - 1; // Gregorian year
    int32_t theNewYear = newYear(gyear);
    int32_t newMoon = newMoonNear(theNewYear + month * 29, TRUE);

    int32_t julianDay = newMoon + kEpochStartAsJulianDay;

    // Save fields for later restoration
    int32_t saveMonth = internalGet(UCAL_MONTH);
    int32_t saveIsLeapMonth = internalGet(UCAL_IS_LEAP_MONTH);

    // Ignore IS_LEAP_MONTH field if useMonth is false
    int32_t isLeapMonth = useMonth ? saveIsLeapMonth : 0;

    UErrorCode status = U_ZERO_ERROR;
    nonConstThis->computeGregorianFields(julianDay, status);
    if (U_FAILURE(status))
        return 0;

    // This will modify the MONTH and IS_LEAP_MONTH fields (only)
    nonConstThis->computeChineseFields(newMoon, getGregorianYear(),
                         getGregorianMonth(), FALSE);

    if (month != internalGet(UCAL_MONTH) ||
        isLeapMonth != internalGet(UCAL_IS_LEAP_MONTH)) {
        newMoon = newMoonNear(newMoon + SYNODIC_GAP, TRUE);
        julianDay = newMoon + kEpochStartAsJulianDay;
    }

    nonConstThis->internalSet(UCAL_MONTH, saveMonth);
    nonConstThis->internalSet(UCAL_IS_LEAP_MONTH, saveIsLeapMonth);

    return julianDay - 1;
}


/**
 * Override Calendar to handle leap months properly.
 * @stable ICU 2.8
 */
void ChineseCalendar::add(UCalendarDateFields field, int32_t amount, UErrorCode& status) {
    switch (field) {
    case UCAL_MONTH:
        if (amount != 0) {
            int32_t dom = get(UCAL_DAY_OF_MONTH, status);
            if (U_FAILURE(status)) break;
            int32_t day = get(UCAL_JULIAN_DAY, status) - kEpochStartAsJulianDay; // Get local day
            if (U_FAILURE(status)) break;
            int32_t moon = day - dom + 1; // New moon
            offsetMonth(moon, dom, amount);
        }
        break;
    default:
        Calendar::add(field, amount, status);
        break;
    }
}

/**
 * Override Calendar to handle leap months properly.
 * @stable ICU 2.8
 */
void ChineseCalendar::add(EDateFields field, int32_t amount, UErrorCode& status) {
    add((UCalendarDateFields)field, amount, status);
}

/**
 * Override Calendar to handle leap months properly.
 * @stable ICU 2.8
 */
void ChineseCalendar::roll(UCalendarDateFields field, int32_t amount, UErrorCode& status) {
    switch (field) {
    case UCAL_MONTH:
        if (amount != 0) {
            int32_t dom = get(UCAL_DAY_OF_MONTH, status);
            if (U_FAILURE(status)) break;
            int32_t day = get(UCAL_JULIAN_DAY, status) - kEpochStartAsJulianDay; // Get local day
            if (U_FAILURE(status)) break;
            int32_t moon = day - dom + 1; // New moon (start of this month)

            // Note throughout the following:  Months 12 and 1 are never
            // followed by a leap month (D&R p. 185).

            // Compute the adjusted month number m.  This is zero-based
            // value from 0..11 in a non-leap year, and from 0..12 in a
            // leap year.
            int32_t m = get(UCAL_MONTH, status); // 0-based month
            if (U_FAILURE(status)) break;
            if (isLeapYear) { // (member variable)
                if (get(UCAL_IS_LEAP_MONTH, status) == 1) {
                    ++m;
                } else {
                    // Check for a prior leap month.  (In the
                    // following, month 0 is the first month of the
                    // year.)  Month 0 is never followed by a leap
                    // month, and we know month m is not a leap month.
                    // moon1 will be the start of month 0 if there is
                    // no leap month between month 0 and month m;
                    // otherwise it will be the start of month 1.
                    int moon1 = moon -
                        (int) (CalendarAstronomer::SYNODIC_MONTH * (m - 0.5));
                    moon1 = newMoonNear(moon1, TRUE);
                    if (isLeapMonthBetween(moon1, moon)) {
                        ++m;
                    }
                }
                if (U_FAILURE(status)) break;
            }

            // Now do the standard roll computation on m, with the
            // allowed range of 0..n-1, where n is 12 or 13.
            int32_t n = isLeapYear ? 13 : 12; // Months in this year
            int32_t newM = (m + amount) % n;
            if (newM < 0) {
                newM += n;
            }

            if (newM != m) {
                offsetMonth(moon, dom, newM - m);
            }
        }
        break;
    default:
        Calendar::roll(field, amount, status);
        break;
    }
}

void ChineseCalendar::roll(EDateFields field, int32_t amount, UErrorCode& status) {
    roll((UCalendarDateFields)field, amount, status);
}


//------------------------------------------------------------------
// Support methods and constants
//------------------------------------------------------------------

/**
 * Convert local days to UTC epoch milliseconds.
 * This is not an accurate conversion in that getTimezoneOffset
 * takes the milliseconds in GMT (not local time). In theory, more
 * accurate algorithm can be implemented but practically we do not need
 * to go through that complication as long as the historical timezone
 * changes did not happen around the 'tricky' new moon (new moon around
 * midnight).
 *
 * @param days days after January 1, 1970 0:00 in the astronomical base zone
 * @return milliseconds after January 1, 1970 0:00 GMT
 */
double ChineseCalendar::daysToMillis(double days) const {
    double millis = days * (double)kOneDay;
    if (fZoneAstroCalc != NULL) {
        int32_t rawOffset, dstOffset;
        UErrorCode status = U_ZERO_ERROR;
        fZoneAstroCalc->getOffset(millis, FALSE, rawOffset, dstOffset, status);
        if (U_SUCCESS(status)) {
        	return millis - (double)(rawOffset + dstOffset);
        }
    }
    return millis - (double)CHINA_OFFSET;
}

/**
 * Convert UTC epoch milliseconds to local days.
 * @param millis milliseconds after January 1, 1970 0:00 GMT
 * @return days after January 1, 1970 0:00 in the astronomical base zone
 */
double ChineseCalendar::millisToDays(double millis) const {
    if (fZoneAstroCalc != NULL) {
        int32_t rawOffset, dstOffset;
        UErrorCode status = U_ZERO_ERROR;
        fZoneAstroCalc->getOffset(millis, FALSE, rawOffset, dstOffset, status);
        if (U_SUCCESS(status)) {
        	return ClockMath::floorDivide(millis + (double)(rawOffset + dstOffset), kOneDay);
        }
    }
    return ClockMath::floorDivide(millis + (double)CHINA_OFFSET, kOneDay);
}

//------------------------------------------------------------------
// Astronomical computations
//------------------------------------------------------------------


/**
 * Return the major solar term on or after December 15 of the given
 * Gregorian year, that is, the winter solstice of the given year.
 * Computations are relative to Asia/Shanghai time zone.
 * @param gyear a Gregorian year
 * @return days after January 1, 1970 0:00 Asia/Shanghai of the
 * winter solstice of the given year
 */
int32_t ChineseCalendar::winterSolstice(int32_t gyear) const {

    UErrorCode status = U_ZERO_ERROR;
    int32_t cacheValue = CalendarCache::get(&gChineseCalendarWinterSolsticeCache, gyear, status);

    if (cacheValue == 0) {
        // In books December 15 is used, but it fails for some years
        // using our algorithms, e.g.: 1298 1391 1492 1553 1560.  That
        // is, winterSolstice(1298) starts search at Dec 14 08:00:00
        // PST 1298 with a final result of Dec 14 10:31:59 PST 1299.
        double ms = daysToMillis(Grego::fieldsToDay(gyear, UCAL_DECEMBER, 1));

        umtx_lock(&astroLock);
        if(gChineseCalendarAstro == NULL) {
            gChineseCalendarAstro = new CalendarAstronomer();
            ucln_i18n_registerCleanup(UCLN_I18N_CHINESE_CALENDAR, calendar_chinese_cleanup);
        }
        gChineseCalendarAstro->setTime(ms);
        UDate solarLong = gChineseCalendarAstro->getSunTime(CalendarAstronomer::WINTER_SOLSTICE(), TRUE);
        umtx_unlock(&astroLock);

        // Winter solstice is 270 degrees solar longitude aka Dongzhi
        cacheValue = (int32_t)millisToDays(solarLong);
        CalendarCache::put(&gChineseCalendarWinterSolsticeCache, gyear, cacheValue, status);
    }
    if(U_FAILURE(status)) {
        cacheValue = 0;
    }
    return cacheValue;
}

/**
 * Return the closest new moon to the given date, searching either
 * forward or backward in time.
 * @param days days after January 1, 1970 0:00 Asia/Shanghai
 * @param after if true, search for a new moon on or after the given
 * date; otherwise, search for a new moon before it
 * @return days after January 1, 1970 0:00 Asia/Shanghai of the nearest
 * new moon after or before <code>days</code>
 */
int32_t ChineseCalendar::newMoonNear(double days, UBool after) const {

    umtx_lock(&astroLock);
    if(gChineseCalendarAstro == NULL) {
        gChineseCalendarAstro = new CalendarAstronomer();
        ucln_i18n_registerCleanup(UCLN_I18N_CHINESE_CALENDAR, calendar_chinese_cleanup);
    }
    gChineseCalendarAstro->setTime(daysToMillis(days));
    UDate newMoon = gChineseCalendarAstro->getMoonTime(CalendarAstronomer::NEW_MOON(), after);
    umtx_unlock(&astroLock);

    return (int32_t) millisToDays(newMoon);
}

/**
 * Return the nearest integer number of synodic months between
 * two dates.
 * @param day1 days after January 1, 1970 0:00 Asia/Shanghai
 * @param day2 days after January 1, 1970 0:00 Asia/Shanghai
 * @return the nearest integer number of months between day1 and day2
 */
int32_t ChineseCalendar::synodicMonthsBetween(int32_t day1, int32_t day2) const {
    double roundme = ((day2 - day1) / CalendarAstronomer::SYNODIC_MONTH);
    return (int32_t) (roundme + (roundme >= 0 ? .5 : -.5));
}

/**
 * Return the major solar term on or before a given date.  This
 * will be an integer from 1..12, with 1 corresponding to 330 degrees,
 * 2 to 0 degrees, 3 to 30 degrees,..., and 12 to 300 degrees.
 * @param days days after January 1, 1970 0:00 Asia/Shanghai
 */
int32_t ChineseCalendar::majorSolarTerm(int32_t days) const {

    umtx_lock(&astroLock);
    if(gChineseCalendarAstro == NULL) {
        gChineseCalendarAstro = new CalendarAstronomer();
        ucln_i18n_registerCleanup(UCLN_I18N_CHINESE_CALENDAR, calendar_chinese_cleanup);
    }
    gChineseCalendarAstro->setTime(daysToMillis(days));
    UDate solarLongitude = gChineseCalendarAstro->getSunLongitude();
    umtx_unlock(&astroLock);

    // Compute (floor(solarLongitude / (pi/6)) + 2) % 12
    int32_t term = ( ((int32_t)(6 * solarLongitude / CalendarAstronomer::PI)) + 2 ) % 12;
    if (term < 1) {
        term += 12;
    }
    return term;
}

/**
 * Return true if the given month lacks a major solar term.
 * @param newMoon days after January 1, 1970 0:00 Asia/Shanghai of a new
 * moon
 */
UBool ChineseCalendar::hasNoMajorSolarTerm(int32_t newMoon) const {
    return majorSolarTerm(newMoon) ==
        majorSolarTerm(newMoonNear(newMoon + SYNODIC_GAP, TRUE));
}


//------------------------------------------------------------------
// Time to fields
//------------------------------------------------------------------

/**
 * Return true if there is a leap month on or after month newMoon1 and
 * at or before month newMoon2.
 * @param newMoon1 days after January 1, 1970 0:00 astronomical base zone
 * of a new moon
 * @param newMoon2 days after January 1, 1970 0:00 astronomical base zone
 * of a new moon
 */
UBool ChineseCalendar::isLeapMonthBetween(int32_t newMoon1, int32_t newMoon2) const {

#ifdef U_DEBUG_CHNSECAL
    // This is only needed to debug the timeOfAngle divergence bug.
    // Remove this later. Liu 11/9/00
    if (synodicMonthsBetween(newMoon1, newMoon2) >= 50) {
        U_DEBUG_CHNSECAL_MSG((
            "isLeapMonthBetween(%d, %d): Invalid parameters", newMoon1, newMoon2
            ));
    }
#endif

    return (newMoon2 >= newMoon1) &&
        (isLeapMonthBetween(newMoon1, newMoonNear(newMoon2 - SYNODIC_GAP, FALSE)) ||
         hasNoMajorSolarTerm(newMoon2));
}

/**
 * Compute fields for the Chinese calendar system.  This method can
 * either set all relevant fields, as required by
 * <code>handleComputeFields()</code>, or it can just set the MONTH and
 * IS_LEAP_MONTH fields, as required by
 * <code>handleComputeMonthStart()</code>.
 *
 * <p>As a side effect, this method sets {@link #isLeapYear}.
 * @param days days after January 1, 1970 0:00 astronomical base zone
 * of the date to compute fields for
 * @param gyear the Gregorian year of the given date
 * @param gmonth the Gregorian month of the given date
 * @param setAllFields if true, set the EXTENDED_YEAR, ERA, YEAR,
 * DAY_OF_MONTH, and DAY_OF_YEAR fields.  In either case set the MONTH
 * and IS_LEAP_MONTH fields.
 */
void ChineseCalendar::computeChineseFields(int32_t days, int32_t gyear, int32_t gmonth,
                                  UBool setAllFields) {

    // Find the winter solstices before and after the target date.
    // These define the boundaries of this Chinese year, specifically,
    // the position of month 11, which always contains the solstice.
    // We want solsticeBefore <= date < solsticeAfter.
    int32_t solsticeBefore;
    int32_t solsticeAfter = winterSolstice(gyear);
    if (days < solsticeAfter) {
        solsticeBefore = winterSolstice(gyear - 1);
    } else {
        solsticeBefore = solsticeAfter;
        solsticeAfter = winterSolstice(gyear + 1);
    }

    // Find the start of the month after month 11.  This will be either
    // the prior month 12 or leap month 11 (very rare).  Also find the
    // start of the following month 11.
    int32_t firstMoon = newMoonNear(solsticeBefore + 1, TRUE);
    int32_t lastMoon = newMoonNear(solsticeAfter + 1, FALSE);
    int32_t thisMoon = newMoonNear(days + 1, FALSE); // Start of this month
    // Note: isLeapYear is a member variable
    isLeapYear = synodicMonthsBetween(firstMoon, lastMoon) == 12;

    int32_t month = synodicMonthsBetween(firstMoon, thisMoon);
    if (isLeapYear && isLeapMonthBetween(firstMoon, thisMoon)) {
        month--;
    }
    if (month < 1) {
        month += 12;
    }

    UBool isLeapMonth = isLeapYear &&
        hasNoMajorSolarTerm(thisMoon) &&
        !isLeapMonthBetween(firstMoon, newMoonNear(thisMoon - SYNODIC_GAP, FALSE));

    internalSet(UCAL_MONTH, month-1); // Convert from 1-based to 0-based
    internalSet(UCAL_IS_LEAP_MONTH, isLeapMonth?1:0);

    if (setAllFields) {

        // Extended year and cycle year is based on the epoch year

        int32_t extended_year = gyear - fEpochYear;
        int cycle_year = gyear - CHINESE_EPOCH_YEAR;
        if (month < 11 ||
            gmonth >= UCAL_JULY) {
            extended_year++;
            cycle_year++;
        }
        int32_t dayOfMonth = days - thisMoon + 1;

        internalSet(UCAL_EXTENDED_YEAR, extended_year);

        // 0->0,60  1->1,1  60->1,60  61->2,1  etc.
        int32_t yearOfCycle;
        int32_t cycle = ClockMath::floorDivide(cycle_year - 1, 60, yearOfCycle);
        internalSet(UCAL_ERA, cycle + 1);
        internalSet(UCAL_YEAR, yearOfCycle + 1);

        internalSet(UCAL_DAY_OF_MONTH, dayOfMonth);

        // Days will be before the first new year we compute if this
        // date is in month 11, leap 11, 12.  There is never a leap 12.
        // New year computations are cached so this should be cheap in
        // the long run.
        int32_t theNewYear = newYear(gyear);
        if (days < theNewYear) {
            theNewYear = newYear(gyear-1);
        }
        internalSet(UCAL_DAY_OF_YEAR, days - theNewYear + 1);
    }
}


//------------------------------------------------------------------
// Fields to time
//------------------------------------------------------------------

/**
 * Return the Chinese new year of the given Gregorian year.
 * @param gyear a Gregorian year
 * @return days after January 1, 1970 0:00 astronomical base zone of the
 * Chinese new year of the given year (this will be a new moon)
 */
int32_t ChineseCalendar::newYear(int32_t gyear) const {
    UErrorCode status = U_ZERO_ERROR;
    int32_t cacheValue = CalendarCache::get(&gChineseCalendarNewYearCache, gyear, status);

    if (cacheValue == 0) {

        int32_t solsticeBefore= winterSolstice(gyear - 1);
        int32_t solsticeAfter = winterSolstice(gyear);
        int32_t newMoon1 = newMoonNear(solsticeBefore + 1, TRUE);
        int32_t newMoon2 = newMoonNear(newMoon1 + SYNODIC_GAP, TRUE);
        int32_t newMoon11 = newMoonNear(solsticeAfter + 1, FALSE);

        if (synodicMonthsBetween(newMoon1, newMoon11) == 12 &&
            (hasNoMajorSolarTerm(newMoon1) || hasNoMajorSolarTerm(newMoon2))) {
            cacheValue = newMoonNear(newMoon2 + SYNODIC_GAP, TRUE);
        } else {
            cacheValue = newMoon2;
        }

        CalendarCache::put(&gChineseCalendarNewYearCache, gyear, cacheValue, status);
    }
    if(U_FAILURE(status)) {
        cacheValue = 0;
    }
    return cacheValue;
}

/**
 * Adjust this calendar to be delta months before or after a given
 * start position, pinning the day of month if necessary.  The start
 * position is given as a local days number for the start of the month
 * and a day-of-month.  Used by add() and roll().
 * @param newMoon the local days of the first day of the month of the
 * start position (days after January 1, 1970 0:00 Asia/Shanghai)
 * @param dom the 1-based day-of-month of the start position
 * @param delta the number of months to move forward or backward from
 * the start position
 */
void ChineseCalendar::offsetMonth(int32_t newMoon, int32_t dom, int32_t delta) {
    UErrorCode status = U_ZERO_ERROR;

    // Move to the middle of the month before our target month.
    newMoon += (int32_t) (CalendarAstronomer::SYNODIC_MONTH * (delta - 0.5));

    // Search forward to the target month's new moon
    newMoon = newMoonNear(newMoon, TRUE);

    // Find the target dom
    int32_t jd = newMoon + kEpochStartAsJulianDay - 1 + dom;

    // Pin the dom.  In this calendar all months are 29 or 30 days
    // so pinning just means handling dom 30.
    if (dom > 29) {
        set(UCAL_JULIAN_DAY, jd-1);
        // TODO Fix this.  We really shouldn't ever have to
        // explicitly call complete().  This is either a bug in
        // this method, in ChineseCalendar, or in
        // Calendar.getActualMaximum().  I suspect the last.
        complete(status);
        if (U_FAILURE(status)) return;
        if (getActualMaximum(UCAL_DAY_OF_MONTH, status) >= dom) {
            if (U_FAILURE(status)) return;
            set(UCAL_JULIAN_DAY, jd);
        }
    } else {
        set(UCAL_JULIAN_DAY, jd);
    }
}


UBool
ChineseCalendar::inDaylightTime(UErrorCode& status) const
{
    // copied from GregorianCalendar
    if (U_FAILURE(status) || !getTimeZone().useDaylightTime())
        return FALSE;

    // Force an update of the state of the Calendar.
    ((ChineseCalendar*)this)->complete(status); // cast away const

    return (UBool)(U_SUCCESS(status) ? (internalGet(UCAL_DST_OFFSET) != 0) : FALSE);
}

// default century

static UDate     chnsecal_gSystemDefaultCenturyStart       = DBL_MIN;
static int32_t   chnsecal_gSystemDefaultCenturyStartYear   = -1;
static icu::UInitOnce chnsecal_gSystemDefaultCenturyInitOnce = U_INITONCE_INITIALIZER;


UBool ChineseCalendar::haveDefaultCentury() const
{
    return TRUE;
}

UDate ChineseCalendar::defaultCenturyStart() const
{
    return internalGetDefaultCenturyStart();
}

int32_t ChineseCalendar::defaultCenturyStartYear() const
{
    return internalGetDefaultCenturyStartYear();
}

static void U_CALLCONV initializeChnseCalSystemDefaultCentury()
{
    // initialize systemDefaultCentury and systemDefaultCenturyYear based
    // on the current time.  They'll be set to 80 years before
    // the current time.
    UErrorCode status = U_ZERO_ERROR;
    ChineseCalendar calendar(Locale("@calendar=chinese"),status);
    if (U_SUCCESS(status)) {
        calendar.setTime(Calendar::getNow(), status);
        calendar.add(UCAL_YEAR, -80, status);
        chnsecal_gSystemDefaultCenturyStart     = calendar.getTime(status);
        chnsecal_gSystemDefaultCenturyStartYear = calendar.get(UCAL_YEAR, status);
    }
    // We have no recourse upon failure unless we want to propagate the failure
    // out.
}

UDate
ChineseCalendar::internalGetDefaultCenturyStart() const
{
    // lazy-evaluate systemDefaultCenturyStart
    umtx_initOnce(chnsecal_gSystemDefaultCenturyInitOnce, &initializeChnseCalSystemDefaultCentury);
    return chnsecal_gSystemDefaultCenturyStart;
}

int32_t
ChineseCalendar::internalGetDefaultCenturyStartYear() const
{
    // lazy-evaluate systemDefaultCenturyStartYear
    umtx_initOnce(chnsecal_gSystemDefaultCenturyInitOnce, &initializeChnseCalSystemDefaultCentury);
    return    chnsecal_gSystemDefaultCenturyStartYear;
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(ChineseCalendar)

U_NAMESPACE_END

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2013, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* File CHOICFMT.CPP
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/20/97    helena      Finished first cut of implementation and got rid
*                           of nextDouble/previousDouble and replaced with
*                           boolean array.
*   4/10/97     aliu        Clean up.  Modified to work on AIX.
*   06/04/97    helena      Fixed applyPattern(), toPattern() and not to include
*                           wchar.h.
*   07/09/97    helena      Made ParsePosition into a class.
*   08/06/97    nos         removed overloaded constructor, fixed 'format(array)'
*   07/22/98    stephen     JDK 1.2 Sync - removed UBool array (doubleFlags)
*   02/22/99    stephen     Removed character literals for EBCDIC safety
********************************************************************************
*/



#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 1997-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File CHOICFMT.H
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/20/97    helena      Finished first cut of implementation and got rid
*                           of nextDouble/previousDouble and replaced with
*                           boolean array.
*   4/10/97     aliu        Clean up.  Modified to work on AIX.
*   8/6/97      nos         Removed overloaded constructor, member var 'buffer'.
*   07/22/98    stephen     Removed operator!= (implemented in Format)
********************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Choice Format.
 */

#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ********************************************************************************
 *   Copyright (C) 1997-2006, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 ********************************************************************************
 *
 * File FIELDPOS.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   02/25/97    aliu        Converted from java.
 *   03/17/97    clhuang     Updated per Format implementation.
 *    07/17/98    stephen        Added default/copy ctors, and operators =, ==, !=
 ********************************************************************************
 */

// *****************************************************************************
// This file was generated from the java source file FieldPosition.java
// *****************************************************************************
 



#if U_SHOW_CPLUSPLUS_API

/**
 * \file 
 * \brief C++ API: FieldPosition identifies the fields in a formatted output.
 */

#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

/**
 * <code>FieldPosition</code> is a simple class used by <code>Format</code>
 * and its subclasses to identify fields in formatted output. Fields are
 * identified by constants, whose names typically end with <code>_FIELD</code>,
 * defined in the various subclasses of <code>Format</code>. See
 * <code>ERA_FIELD</code> and its friends in <code>DateFormat</code> for
 * an example.
 *
 * <p>
 * <code>FieldPosition</code> keeps track of the position of the
 * field within the formatted output with two indices: the index
 * of the first character of the field and the index of the last
 * character of the field.
 *
 * <p>
 * One version of the <code>format</code> method in the various
 * <code>Format</code> classes requires a <code>FieldPosition</code>
 * object as an argument. You use this <code>format</code> method
 * to perform partial formatting or to get information about the
 * formatted output (such as the position of a field).
 *
 * The FieldPosition class is not intended for public subclassing.
 *
 * <p>
 * Below is an example of using <code>FieldPosition</code> to aid
 * alignment of an array of formatted floating-point numbers on
 * their decimal points:
 * <pre>
 * \code
 *       double doubleNum[] = {123456789.0, -12345678.9, 1234567.89, -123456.789,
 *                  12345.6789, -1234.56789, 123.456789, -12.3456789, 1.23456789};
 *       int dNumSize = (int)(sizeof(doubleNum)/sizeof(double));
 *       
 *       UErrorCode status = U_ZERO_ERROR;
 *       DecimalFormat* fmt = (DecimalFormat*) NumberFormat::createInstance(status);
 *       fmt->setDecimalSeparatorAlwaysShown(true);
 *       
 *       const int tempLen = 20;
 *       char temp[tempLen];
 *       
 *       for (int i=0; i<dNumSize; i++) {
 *           FieldPosition pos(NumberFormat::INTEGER_FIELD);
 *           UnicodeString buf;
 *           char fmtText[tempLen];
 *           ToCharString(fmt->format(doubleNum[i], buf, pos), fmtText);
 *           for (int j=0; j<tempLen; j++) temp[j] = ' '; // clear with spaces
 *           temp[__min(tempLen, tempLen-pos.getEndIndex())] = '\0';
 *           cout << temp << fmtText   << endl;
 *       }
 *       delete fmt;
 * \endcode
 * </pre>
 * <p>
 * The code will generate the following output:
 * <pre>
 * \code
 *           123,456,789.000
 *           -12,345,678.900
 *             1,234,567.880
 *              -123,456.789
 *                12,345.678
 *                -1,234.567
 *                   123.456
 *                   -12.345
 *                     1.234
 *  \endcode
 * </pre>
 */
class U_I18N_API FieldPosition : public UObject {
public:
    /**
     * DONT_CARE may be specified as the field to indicate that the
     * caller doesn't need to specify a field.
     * @stable ICU 2.0
     */
    enum { DONT_CARE = -1 };

    /**
     * Creates a FieldPosition object with a non-specified field.
     * @stable ICU 2.0
     */
    FieldPosition() 
        : UObject(), fField(DONT_CARE), fBeginIndex(0), fEndIndex(0) {}

    /**
     * Creates a FieldPosition object for the given field.  Fields are
     * identified by constants, whose names typically end with _FIELD,
     * in the various subclasses of Format.
     *
     * @see NumberFormat#INTEGER_FIELD
     * @see NumberFormat#FRACTION_FIELD
     * @see DateFormat#YEAR_FIELD
     * @see DateFormat#MONTH_FIELD
     * @stable ICU 2.0
     */
    FieldPosition(int32_t field) 
        : UObject(), fField(field), fBeginIndex(0), fEndIndex(0) {}

    /**
     * Copy constructor
     * @param copy the object to be copied from.
     * @stable ICU 2.0
     */
    FieldPosition(const FieldPosition& copy) 
        : UObject(copy), fField(copy.fField), fBeginIndex(copy.fBeginIndex), fEndIndex(copy.fEndIndex) {}

    /**
     * Destructor
     * @stable ICU 2.0
     */
    virtual ~FieldPosition();

    /**
     * Assignment operator
     * @param copy the object to be copied from.
     * @stable ICU 2.0
     */
    FieldPosition&      operator=(const FieldPosition& copy);

    /** 
     * Equality operator.
     * @param that    the object to be compared with.
     * @return        TRUE if the two field positions are equal, FALSE otherwise.
     * @stable ICU 2.0
     */
    UBool              operator==(const FieldPosition& that) const;

    /** 
     * Equality operator.
     * @param that    the object to be compared with.
     * @return        TRUE if the two field positions are not equal, FALSE otherwise.
     * @stable ICU 2.0
     */
    UBool              operator!=(const FieldPosition& that) const;

    /**
     * Clone this object.
     * Clones can be used concurrently in multiple threads.
     * If an error occurs, then NULL is returned.
     * The caller must delete the clone.
     *
     * @return a clone of this object
     *
     * @see getDynamicClassID
     * @stable ICU 2.8
     */
    FieldPosition *clone() const;

    /**
     * Retrieve the field identifier.
     * @return    the field identifier.
     * @stable ICU 2.0
     */
    int32_t getField(void) const { return fField; }

    /**
     * Retrieve the index of the first character in the requested field.
     * @return    the index of the first character in the requested field.
     * @stable ICU 2.0
     */
    int32_t getBeginIndex(void) const { return fBeginIndex; }

    /**
     * Retrieve the index of the character following the last character in the
     * requested field.
     * @return    the index of the character following the last character in the
     *            requested field.
     * @stable ICU 2.0
     */
    int32_t getEndIndex(void) const { return fEndIndex; }
 
    /**
     * Set the field.
     * @param f    the new value of the field.
     * @stable ICU 2.0
     */
    void setField(int32_t f) { fField = f; }

    /**
     * Set the begin index.  For use by subclasses of Format.
     * @param bi    the new value of the begin index
     * @stable ICU 2.0
     */
    void setBeginIndex(int32_t bi) { fBeginIndex = bi; }

    /**
     * Set the end index.  For use by subclasses of Format.
     * @param ei    the new value of the end index
     * @stable ICU 2.0
     */
    void setEndIndex(int32_t ei) { fEndIndex = ei; }
    
    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 2.2
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 2.2
     */
    static UClassID U_EXPORT2 getStaticClassID();

private:
    /**
     * Input: Desired field to determine start and end offsets for.
     * The meaning depends on the subclass of Format.
     */
    int32_t fField;

    /**
     * Output: Start offset of field in text.
     * If the field does not occur in the text, 0 is returned.
     */
    int32_t fBeginIndex;

    /**
     * Output: End offset of field in text.
     * If the field does not occur in the text, 0 is returned.
     */
    int32_t fEndIndex;
};

inline FieldPosition&
FieldPosition::operator=(const FieldPosition& copy)
{
    fField         = copy.fField;
    fEndIndex     = copy.fEndIndex;
    fBeginIndex = copy.fBeginIndex;
    return *this;
}

inline UBool
FieldPosition::operator==(const FieldPosition& copy) const
{
    return (fField == copy.fField &&
        fEndIndex == copy.fEndIndex &&
        fBeginIndex == copy.fBeginIndex);
}

inline UBool
FieldPosition::operator!=(const FieldPosition& copy) const
{
    return !operator==(copy);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
* Copyright (C) 1997-2011, International Business Machines Corporation and others.
* All Rights Reserved.
********************************************************************************
*
* File FORMAT.H
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/17/97    clhuang     Updated per C++ implementation.
*   03/27/97    helena      Updated to pass the simple test after code review.
********************************************************************************
*/
// *****************************************************************************
// This file was generated from the java source file Format.java
// *****************************************************************************





#if U_SHOW_CPLUSPLUS_API

/**
 * \file 
 * \brief C++ API: Base class for all formats. 
 */

#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 1997-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File FMTABLE.H
*
* Modification History:
*
*   Date        Name        Description
*   02/29/97    aliu        Creation.
********************************************************************************
*/



#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Formattable is a thin wrapper for primitive types used for formatting and parsing
 */

#if !UCONFIG_NO_FORMATTING



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
* Copyright (C) 2013-2014, International Business Machines Corporation and others.
* All Rights Reserved.
********************************************************************************
*
* File UFORMATTABLE.H
*
* Modification History:
*
*   Date        Name        Description
*   2013 Jun 7  srl         New
********************************************************************************
*/

/**
 * \file
 * \brief C API: UFormattable is a thin wrapper for primitive types used for formatting and parsing.
 *
 * This is a C interface to the icu::Formattable class. Static functions on this class convert
 * to and from this interface (via reinterpret_cast).  Note that Formattables (and thus UFormattables)
 * are mutable, and many operations (even getters) may actually modify the internal state. For this
 * reason, UFormattables are not thread safe, and should not be shared between threads.
 *
 * See {@link unum_parseToUFormattable} for example code.
 */




#if !UCONFIG_NO_FORMATTING



/**
 * Enum designating the type of a UFormattable instance.
 * Practically, this indicates which of the getters would return without conversion
 * or error.
 * @see icu::Formattable::Type
 * @stable ICU 52
 */
typedef enum UFormattableType {
  UFMT_DATE = 0, /**< ufmt_getDate() will return without conversion. @see ufmt_getDate*/
  UFMT_DOUBLE,   /**< ufmt_getDouble() will return without conversion.  @see ufmt_getDouble*/
  UFMT_LONG,     /**< ufmt_getLong() will return without conversion. @see ufmt_getLong */
  UFMT_STRING,   /**< ufmt_getUChars() will return without conversion.  @see ufmt_getUChars*/
  UFMT_ARRAY,    /**< ufmt_countArray() and ufmt_getArray() will return the value.  @see ufmt_getArrayItemByIndex */
  UFMT_INT64,    /**< ufmt_getInt64() will return without conversion. @see ufmt_getInt64 */
  UFMT_OBJECT,   /**< ufmt_getObject() will return without conversion.  @see ufmt_getObject*/
#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UFormattableType value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UFMT_COUNT
#endif  /* U_HIDE_DEPRECATED_API */
} UFormattableType;


/**
 * Opaque type representing various types of data which may be used for formatting
 * and parsing operations.
 * @see icu::Formattable
 * @stable ICU 52
 */
typedef void *UFormattable;

/**
 * Initialize a UFormattable, to type UNUM_LONG, value 0
 * may return error if memory allocation failed.
 * parameter status error code.
 * See {@link unum_parseToUFormattable} for example code.
 * @stable ICU 52
 * @return the new UFormattable
 * @see ufmt_close
 * @see icu::Formattable::Formattable()
 */
U_STABLE UFormattable* U_EXPORT2
ufmt_open(UErrorCode* status);

/**
 * Cleanup any additional memory allocated by this UFormattable.
 * @param fmt the formatter
 * @stable ICU 52
 * @see ufmt_open
 */
U_STABLE void U_EXPORT2
ufmt_close(UFormattable* fmt);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUFormattablePointer
 * "Smart pointer" class, closes a UFormattable via ufmt_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 52
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUFormattablePointer, UFormattable, ufmt_close);

U_NAMESPACE_END

#endif

/**
 * Return the type of this object
 * @param fmt the UFormattable object
 * @param status status code - U_ILLEGAL_ARGUMENT_ERROR is returned if the UFormattable contains data not supported by
 * the API
 * @return the value as a UFormattableType
 * @see ufmt_isNumeric
 * @see icu::Formattable::getType() const
 * @stable ICU 52
 */
U_STABLE UFormattableType U_EXPORT2
ufmt_getType(const UFormattable* fmt, UErrorCode *status);

/**
 * Return whether the object is numeric.
 * @param fmt the UFormattable object
 * @return true if the object is a double, long, or int64 value, else false.
 * @see ufmt_getType
 * @see icu::Formattable::isNumeric() const
 * @stable ICU 52
 */
U_STABLE UBool U_EXPORT2
ufmt_isNumeric(const UFormattable* fmt);

/**
 * Gets the UDate value of this object.  If the type is not of type UFMT_DATE,
 * status is set to U_INVALID_FORMAT_ERROR and the return value is
 * undefined.
 * @param fmt the UFormattable object
 * @param status the error code - any conversion or format errors
 * @return the value
 * @stable ICU 52
 * @see icu::Formattable::getDate(UErrorCode&) const
 */
U_STABLE UDate U_EXPORT2
ufmt_getDate(const UFormattable* fmt, UErrorCode *status);

/**
 * Gets the double value of this object. If the type is not a UFMT_DOUBLE, or
 * if there are additional significant digits than fit in a double type,
 * a conversion is performed with  possible loss of precision.
 * If the type is UFMT_OBJECT and the
 * object is a Measure, then the result of
 * getNumber().getDouble(status) is returned.  If this object is
 * neither a numeric type nor a Measure, then 0 is returned and
 * the status is set to U_INVALID_FORMAT_ERROR.
 * @param fmt the UFormattable object
 * @param status the error code - any conversion or format errors
 * @return the value
 * @stable ICU 52
 * @see icu::Formattable::getDouble(UErrorCode&) const
 */
U_STABLE double U_EXPORT2
ufmt_getDouble(UFormattable* fmt, UErrorCode *status);

/**
 * Gets the long (int32_t) value of this object. If the magnitude is too
 * large to fit in a long, then the maximum or minimum long value,
 * as appropriate, is returned and the status is set to
 * U_INVALID_FORMAT_ERROR.  If this object is of type UFMT_INT64 and
 * it fits within a long, then no precision is lost.  If it is of
 * type kDouble or kDecimalNumber, then a conversion is peformed, with
 * truncation of any fractional part.  If the type is UFMT_OBJECT and
 * the object is a Measure, then the result of
 * getNumber().getLong(status) is returned.  If this object is
 * neither a numeric type nor a Measure, then 0 is returned and
 * the status is set to U_INVALID_FORMAT_ERROR.
 * @param fmt the UFormattable object
 * @param status the error code - any conversion or format errors
 * @return the value
 * @stable ICU 52
 * @see icu::Formattable::getLong(UErrorCode&) const
 */
U_STABLE int32_t U_EXPORT2
ufmt_getLong(UFormattable* fmt, UErrorCode *status);


/**
 * Gets the int64_t value of this object. If this object is of a numeric
 * type and the magnitude is too large to fit in an int64, then
 * the maximum or minimum int64 value, as appropriate, is returned
 * and the status is set to U_INVALID_FORMAT_ERROR.  If the
 * magnitude fits in an int64, then a casting conversion is
 * peformed, with truncation of any fractional part.  If the type
 * is UFMT_OBJECT and the object is a Measure, then the result of
 * getNumber().getDouble(status) is returned.  If this object is
 * neither a numeric type nor a Measure, then 0 is returned and
 * the status is set to U_INVALID_FORMAT_ERROR.
 * @param fmt the UFormattable object
 * @param status the error code - any conversion or format errors
 * @return the value
 * @stable ICU 52
 * @see icu::Formattable::getInt64(UErrorCode&) const
 */
U_STABLE int64_t U_EXPORT2
ufmt_getInt64(UFormattable* fmt, UErrorCode *status);

/**
 * Returns a pointer to the UObject contained within this
 * formattable (as a const void*), or NULL if this object
 * is not of type UFMT_OBJECT.
 * @param fmt the UFormattable object
 * @param status the error code - any conversion or format errors
 * @return the value as a const void*. It is a polymorphic C++ object.
 * @stable ICU 52
 * @see icu::Formattable::getObject() const
 */
U_STABLE const void *U_EXPORT2
ufmt_getObject(const UFormattable* fmt, UErrorCode *status);

/**
 * Gets the string value of this object as a UChar string. If the type is not a
 * string, status is set to U_INVALID_FORMAT_ERROR and a NULL pointer is returned.
 * This function is not thread safe and may modify the UFormattable if need be to terminate the string.
 * The returned pointer is not valid if any other functions are called on this UFormattable, or if the UFormattable is closed.
 * @param fmt the UFormattable object
 * @param status the error code - any conversion or format errors
 * @param len if non null, contains the string length on return
 * @return the null terminated string value - must not be referenced after any other functions are called on this UFormattable.
 * @stable ICU 52
 * @see icu::Formattable::getString(UnicodeString&)const
 */
U_STABLE const UChar* U_EXPORT2
ufmt_getUChars(UFormattable* fmt, int32_t *len, UErrorCode *status);

/**
 * Get the number of array objects contained, if an array type UFMT_ARRAY
 * @param fmt the UFormattable object
 * @param status the error code - any conversion or format errors. U_ILLEGAL_ARGUMENT_ERROR if not an array type.
 * @return the number of array objects or undefined if not an array type
 * @stable ICU 52
 * @see ufmt_getArrayItemByIndex
 */
U_STABLE int32_t U_EXPORT2
ufmt_getArrayLength(const UFormattable* fmt, UErrorCode *status);

/**
 * Get the specified value from the array of UFormattables. Invalid if the object is not an array type UFMT_ARRAY
 * @param fmt the UFormattable object
 * @param n the number of the array to return (0 based).
 * @param status the error code - any conversion or format errors. Returns an error if n is out of bounds.
 * @return the nth array value, only valid while the containing UFormattable is valid. NULL if not an array.
 * @stable ICU 52
 * @see icu::Formattable::getArray(int32_t&, UErrorCode&) const
 */
U_STABLE UFormattable * U_EXPORT2
ufmt_getArrayItemByIndex(UFormattable* fmt, int32_t n, UErrorCode *status);

/**
 * Returns a numeric string representation of the number contained within this
 * formattable, or NULL if this object does not contain numeric type.
 * For values obtained by parsing, the returned decimal number retains
 * the full precision and range of the original input, unconstrained by
 * the limits of a double floating point or a 64 bit int.
 *
 * This function is not thread safe, and therfore is not declared const,
 * even though it is logically const.
 * The resulting buffer is owned by the UFormattable and is invalid if any other functions are
 * called on the UFormattable.
 *
 * Possible errors include U_MEMORY_ALLOCATION_ERROR, and
 * U_INVALID_STATE if the formattable object has not been set to
 * a numeric type.
 * @param fmt the UFormattable object
 * @param len if non-null, on exit contains the string length (not including the terminating null)
 * @param status the error code
 * @return the character buffer as a NULL terminated string, which is owned by the object and must not be accessed if any other functions are called on this object.
 * @stable ICU 52
 * @see icu::Formattable::getDecimalNumber(UErrorCode&)
 */
U_STABLE const char * U_EXPORT2
ufmt_getDecNumChars(UFormattable *fmt, int32_t *len, UErrorCode *status);

#endif



U_NAMESPACE_BEGIN

class CharString;
namespace number {
namespace impl {
class DecimalQuantity;
}
}

/**
 * Formattable objects can be passed to the Format class or
 * its subclasses for formatting.  Formattable is a thin wrapper
 * class which interconverts between the primitive numeric types
 * (double, long, etc.) as well as UDate and UnicodeString.
 *
 * <p>Internally, a Formattable object is a union of primitive types.
 * As such, it can only store one flavor of data at a time.  To
 * determine what flavor of data it contains, use the getType method.
 *
 * <p>As of ICU 3.0, Formattable may also wrap a UObject pointer,
 * which it owns.  This allows an instance of any ICU class to be
 * encapsulated in a Formattable.  For legacy reasons and for
 * efficiency, primitive numeric types are still stored directly
 * within a Formattable.
 *
 * <p>The Formattable class is not suitable for subclassing.
 *
 * <p>See UFormattable for a C wrapper.
 */
class U_I18N_API Formattable : public UObject {
public:
    /**
     * This enum is only used to let callers distinguish between
     * the Formattable(UDate) constructor and the Formattable(double)
     * constructor; the compiler cannot distinguish the signatures,
     * since UDate is currently typedefed to be either double or long.
     * If UDate is changed later to be a bonafide class
     * or struct, then we no longer need this enum.
     * @stable ICU 2.4
     */
    enum ISDATE { kIsDate };

    /**
     * Default constructor
     * @stable ICU 2.4
     */
    Formattable(); // Type kLong, value 0

    /**
     * Creates a Formattable object with a UDate instance.
     * @param d the UDate instance.
     * @param flag the flag to indicate this is a date. Always set it to kIsDate
     * @stable ICU 2.0
     */
    Formattable(UDate d, ISDATE flag);

    /**
     * Creates a Formattable object with a double number.
     * @param d the double number.
     * @stable ICU 2.0
     */
    Formattable(double d);

    /**
     * Creates a Formattable object with a long number.
     * @param l the long number.
     * @stable ICU 2.0
     */
    Formattable(int32_t l);

    /**
     * Creates a Formattable object with an int64_t number
     * @param ll the int64_t number.
     * @stable ICU 2.8
     */
    Formattable(int64_t ll);

#if !UCONFIG_NO_CONVERSION
    /**
     * Creates a Formattable object with a char string pointer.
     * Assumes that the char string is null terminated.
     * @param strToCopy the char string.
     * @stable ICU 2.0
     */
    Formattable(const char* strToCopy);
#endif

    /**
     * Creates a Formattable object of an appropriate numeric type from a
     * a decimal number in string form.  The Formattable will retain the
     * full precision of the input in decimal format, even when it exceeds
     * what can be represented by a double or int64_t.
     *
     * @param number  the unformatted (not localized) string representation
     *                     of the Decimal number.
     * @param status  the error code.  Possible errors include U_INVALID_FORMAT_ERROR
     *                if the format of the string does not conform to that of a
     *                decimal number.
     * @stable ICU 4.4
     */
    Formattable(StringPiece number, UErrorCode &status);

    /**
     * Creates a Formattable object with a UnicodeString object to copy from.
     * @param strToCopy the UnicodeString string.
     * @stable ICU 2.0
     */
    Formattable(const UnicodeString& strToCopy);

    /**
     * Creates a Formattable object with a UnicodeString object to adopt from.
     * @param strToAdopt the UnicodeString string.
     * @stable ICU 2.0
     */
    Formattable(UnicodeString* strToAdopt);

    /**
     * Creates a Formattable object with an array of Formattable objects.
     * @param arrayToCopy the Formattable object array.
     * @param count the array count.
     * @stable ICU 2.0
     */
    Formattable(const Formattable* arrayToCopy, int32_t count);

    /**
     * Creates a Formattable object that adopts the given UObject.
     * @param objectToAdopt the UObject to set this object to
     * @stable ICU 3.0
     */
    Formattable(UObject* objectToAdopt);

    /**
     * Copy constructor.
     * @stable ICU 2.0
     */
    Formattable(const Formattable&);

    /**
     * Assignment operator.
     * @param rhs   The Formattable object to copy into this object.
     * @stable ICU 2.0
     */
    Formattable&    operator=(const Formattable &rhs);

    /**
     * Equality comparison.
     * @param other    the object to be compared with.
     * @return        TRUE if other are equal to this, FALSE otherwise.
     * @stable ICU 2.0
     */
    UBool          operator==(const Formattable &other) const;

    /**
     * Equality operator.
     * @param other    the object to be compared with.
     * @return        TRUE if other are unequal to this, FALSE otherwise.
     * @stable ICU 2.0
     */
    UBool          operator!=(const Formattable& other) const
      { return !operator==(other); }

    /**
     * Destructor.
     * @stable ICU 2.0
     */
    virtual         ~Formattable();

    /**
     * Clone this object.
     * Clones can be used concurrently in multiple threads.
     * If an error occurs, then NULL is returned.
     * The caller must delete the clone.
     *
     * @return a clone of this object
     *
     * @see getDynamicClassID
     * @stable ICU 2.8
     */
    Formattable *clone() const;

    /**
     * Selector for flavor of data type contained within a
     * Formattable object.  Formattable is a union of several
     * different types, and at any time contains exactly one type.
     * @stable ICU 2.4
     */
    enum Type {
        /**
         * Selector indicating a UDate value.  Use getDate to retrieve
         * the value.
         * @stable ICU 2.4
         */
        kDate,

        /**
         * Selector indicating a double value.  Use getDouble to
         * retrieve the value.
         * @stable ICU 2.4
         */
        kDouble,

        /**
         * Selector indicating a 32-bit integer value.  Use getLong to
         * retrieve the value.
         * @stable ICU 2.4
         */
        kLong,

        /**
         * Selector indicating a UnicodeString value.  Use getString
         * to retrieve the value.
         * @stable ICU 2.4
         */
        kString,

        /**
         * Selector indicating an array of Formattables.  Use getArray
         * to retrieve the value.
         * @stable ICU 2.4
         */
        kArray,

        /**
         * Selector indicating a 64-bit integer value.  Use getInt64
         * to retrieve the value.
         * @stable ICU 2.8
         */
        kInt64,

        /**
         * Selector indicating a UObject value.  Use getObject to
         * retrieve the value.
         * @stable ICU 3.0
         */
        kObject
   };

    /**
     * Gets the data type of this Formattable object.
     * @return    the data type of this Formattable object.
     * @stable ICU 2.0
     */
    Type            getType(void) const;

    /**
     * Returns TRUE if the data type of this Formattable object
     * is kDouble, kLong, or kInt64
     * @return TRUE if this is a pure numeric object
     * @stable ICU 3.0
     */
    UBool           isNumeric() const;

    /**
     * Gets the double value of this object. If this object is not of type
     * kDouble then the result is undefined.
     * @return    the double value of this object.
     * @stable ICU 2.0
     */
    double          getDouble(void) const { return fValue.fDouble; }

    /**
     * Gets the double value of this object. If this object is of type
     * long, int64 or Decimal Number then a conversion is peformed, with
     * possible loss of precision.  If the type is kObject and the
     * object is a Measure, then the result of
     * getNumber().getDouble(status) is returned.  If this object is
     * neither a numeric type nor a Measure, then 0 is returned and
     * the status is set to U_INVALID_FORMAT_ERROR.
     * @param status the error code
     * @return the double value of this object.
     * @stable ICU 3.0
     */
    double          getDouble(UErrorCode& status) const;

    /**
     * Gets the long value of this object. If this object is not of type
     * kLong then the result is undefined.
     * @return    the long value of this object.
     * @stable ICU 2.0
     */
    int32_t         getLong(void) const { return (int32_t)fValue.fInt64; }

    /**
     * Gets the long value of this object. If the magnitude is too
     * large to fit in a long, then the maximum or minimum long value,
     * as appropriate, is returned and the status is set to
     * U_INVALID_FORMAT_ERROR.  If this object is of type kInt64 and
     * it fits within a long, then no precision is lost.  If it is of
     * type kDouble, then a conversion is peformed, with
     * truncation of any fractional part.  If the type is kObject and
     * the object is a Measure, then the result of
     * getNumber().getLong(status) is returned.  If this object is
     * neither a numeric type nor a Measure, then 0 is returned and
     * the status is set to U_INVALID_FORMAT_ERROR.
     * @param status the error code
     * @return    the long value of this object.
     * @stable ICU 3.0
     */
    int32_t         getLong(UErrorCode& status) const;

    /**
     * Gets the int64 value of this object. If this object is not of type
     * kInt64 then the result is undefined.
     * @return    the int64 value of this object.
     * @stable ICU 2.8
     */
    int64_t         getInt64(void) const { return fValue.fInt64; }

    /**
     * Gets the int64 value of this object. If this object is of a numeric
     * type and the magnitude is too large to fit in an int64, then
     * the maximum or minimum int64 value, as appropriate, is returned
     * and the status is set to U_INVALID_FORMAT_ERROR.  If the
     * magnitude fits in an int64, then a casting conversion is
     * peformed, with truncation of any fractional part.  If the type
     * is kObject and the object is a Measure, then the result of
     * getNumber().getDouble(status) is returned.  If this object is
     * neither a numeric type nor a Measure, then 0 is returned and
     * the status is set to U_INVALID_FORMAT_ERROR.
     * @param status the error code
     * @return    the int64 value of this object.
     * @stable ICU 3.0
     */
    int64_t         getInt64(UErrorCode& status) const;

    /**
     * Gets the Date value of this object. If this object is not of type
     * kDate then the result is undefined.
     * @return    the Date value of this object.
     * @stable ICU 2.0
     */
    UDate           getDate() const { return fValue.fDate; }

    /**
     * Gets the Date value of this object.  If the type is not a date,
     * status is set to U_INVALID_FORMAT_ERROR and the return value is
     * undefined.
     * @param status the error code.
     * @return    the Date value of this object.
     * @stable ICU 3.0
     */
     UDate          getDate(UErrorCode& status) const;

    /**
     * Gets the string value of this object. If this object is not of type
     * kString then the result is undefined.
     * @param result    Output param to receive the Date value of this object.
     * @return          A reference to 'result'.
     * @stable ICU 2.0
     */
    UnicodeString&  getString(UnicodeString& result) const
      { result=*fValue.fString; return result; }

    /**
     * Gets the string value of this object. If the type is not a
     * string, status is set to U_INVALID_FORMAT_ERROR and a bogus
     * string is returned.
     * @param result    Output param to receive the Date value of this object.
     * @param status    the error code.
     * @return          A reference to 'result'.
     * @stable ICU 3.0
     */
    UnicodeString&  getString(UnicodeString& result, UErrorCode& status) const;

    /**
     * Gets a const reference to the string value of this object. If
     * this object is not of type kString then the result is
     * undefined.
     * @return   a const reference to the string value of this object.
     * @stable ICU 2.0
     */
    inline const UnicodeString& getString(void) const;

    /**
     * Gets a const reference to the string value of this object.  If
     * the type is not a string, status is set to
     * U_INVALID_FORMAT_ERROR and the result is a bogus string.
     * @param status    the error code.
     * @return   a const reference to the string value of this object.
     * @stable ICU 3.0
     */
    const UnicodeString& getString(UErrorCode& status) const;

    /**
     * Gets a reference to the string value of this object. If this
     * object is not of type kString then the result is undefined.
     * @return   a reference to the string value of this object.
     * @stable ICU 2.0
     */
    inline UnicodeString& getString(void);

    /**
     * Gets a reference to the string value of this object. If the
     * type is not a string, status is set to U_INVALID_FORMAT_ERROR
     * and the result is a bogus string.
     * @param status    the error code.
     * @return   a reference to the string value of this object.
     * @stable ICU 3.0
     */
    UnicodeString& getString(UErrorCode& status);

    /**
     * Gets the array value and count of this object. If this object
     * is not of type kArray then the result is undefined.
     * @param count    fill-in with the count of this object.
     * @return         the array value of this object.
     * @stable ICU 2.0
     */
    const Formattable* getArray(int32_t& count) const
      { count=fValue.fArrayAndCount.fCount; return fValue.fArrayAndCount.fArray; }

    /**
     * Gets the array value and count of this object. If the type is
     * not an array, status is set to U_INVALID_FORMAT_ERROR, count is
     * set to 0, and the result is NULL.
     * @param count    fill-in with the count of this object.
     * @param status the error code.
     * @return         the array value of this object.
     * @stable ICU 3.0
     */
    const Formattable* getArray(int32_t& count, UErrorCode& status) const;

    /**
     * Accesses the specified element in the array value of this
     * Formattable object. If this object is not of type kArray then
     * the result is undefined.
     * @param index the specified index.
     * @return the accessed element in the array.
     * @stable ICU 2.0
     */
    Formattable&    operator[](int32_t index) { return fValue.fArrayAndCount.fArray[index]; }

    /**
     * Returns a pointer to the UObject contained within this
     * formattable, or NULL if this object does not contain a UObject.
     * @return a UObject pointer, or NULL
     * @stable ICU 3.0
     */
    const UObject*  getObject() const;

    /**
     * Returns a numeric string representation of the number contained within this
     * formattable, or NULL if this object does not contain numeric type.
     * For values obtained by parsing, the returned decimal number retains
     * the full precision and range of the original input, unconstrained by
     * the limits of a double floating point or a 64 bit int.
     *
     * This function is not thread safe, and therfore is not declared const,
     * even though it is logically const.
     *
     * Possible errors include U_MEMORY_ALLOCATION_ERROR, and
     * U_INVALID_STATE if the formattable object has not been set to
     * a numeric type.
     *
     * @param status the error code.
     * @return the unformatted string representation of a number.
     * @stable ICU 4.4
     */
    StringPiece getDecimalNumber(UErrorCode &status);

     /**
     * Sets the double value of this object and changes the type to
     * kDouble.
     * @param d    the new double value to be set.
     * @stable ICU 2.0
     */
    void            setDouble(double d);

    /**
     * Sets the long value of this object and changes the type to
     * kLong.
     * @param l    the new long value to be set.
     * @stable ICU 2.0
     */
    void            setLong(int32_t l);

    /**
     * Sets the int64 value of this object and changes the type to
     * kInt64.
     * @param ll    the new int64 value to be set.
     * @stable ICU 2.8
     */
    void            setInt64(int64_t ll);

    /**
     * Sets the Date value of this object and changes the type to
     * kDate.
     * @param d    the new Date value to be set.
     * @stable ICU 2.0
     */
    void            setDate(UDate d);

    /**
     * Sets the string value of this object and changes the type to
     * kString.
     * @param stringToCopy    the new string value to be set.
     * @stable ICU 2.0
     */
    void            setString(const UnicodeString& stringToCopy);

    /**
     * Sets the array value and count of this object and changes the
     * type to kArray.
     * @param array    the array value.
     * @param count    the number of array elements to be copied.
     * @stable ICU 2.0
     */
    void            setArray(const Formattable* array, int32_t count);

    /**
     * Sets and adopts the string value and count of this object and
     * changes the type to kArray.
     * @param stringToAdopt    the new string value to be adopted.
     * @stable ICU 2.0
     */
    void            adoptString(UnicodeString* stringToAdopt);

    /**
     * Sets and adopts the array value and count of this object and
     * changes the type to kArray.
     * @stable ICU 2.0
     */
    void            adoptArray(Formattable* array, int32_t count);

    /**
     * Sets and adopts the UObject value of this object and changes
     * the type to kObject.  After this call, the caller must not
     * delete the given object.
     * @param objectToAdopt the UObject value to be adopted
     * @stable ICU 3.0
     */
    void            adoptObject(UObject* objectToAdopt);

    /**
     * Sets the the numeric value from a decimal number string, and changes
     * the type to to a numeric type appropriate for the number.
     * The syntax of the number is a "numeric string"
     * as defined in the Decimal Arithmetic Specification, available at
     * http://speleotrove.com/decimal
     * The full precision and range of the input number will be retained,
     * even when it exceeds what can be represented by a double or an int64.
     *
     * @param numberString  a string representation of the unformatted decimal number.
     * @param status        the error code.  Set to U_INVALID_FORMAT_ERROR if the
     *                      incoming string is not a valid decimal number.
     * @stable ICU 4.4
     */
    void             setDecimalNumber(StringPiece numberString,
                                      UErrorCode &status);

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 2.2
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 2.2
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * Convert the UFormattable to a Formattable.  Internally, this is a reinterpret_cast.
     * @param fmt a valid UFormattable
     * @return the UFormattable as a Formattable object pointer.  This is an alias to the original
     * UFormattable, and so is only valid while the original argument remains in scope.
     * @stable ICU 52
     */
    static inline Formattable *fromUFormattable(UFormattable *fmt);

    /**
     * Convert the const UFormattable to a const Formattable.  Internally, this is a reinterpret_cast.
     * @param fmt a valid UFormattable
     * @return the UFormattable as a Formattable object pointer.  This is an alias to the original
     * UFormattable, and so is only valid while the original argument remains in scope.
     * @stable ICU 52
     */
    static inline const Formattable *fromUFormattable(const UFormattable *fmt);

    /**
     * Convert this object pointer to a UFormattable.
     * @return this object as a UFormattable pointer.   This is an alias to this object,
     * and so is only valid while this object remains in scope.
     * @stable ICU 52
     */
    inline UFormattable *toUFormattable();

    /**
     * Convert this object pointer to a UFormattable.
     * @return this object as a UFormattable pointer.   This is an alias to this object,
     * and so is only valid while this object remains in scope.
     * @stable ICU 52
     */
    inline const UFormattable *toUFormattable() const;

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Deprecated variant of getLong(UErrorCode&).
     * @param status the error code
     * @return the long value of this object.
     * @deprecated ICU 3.0 use getLong(UErrorCode&) instead
     */
    inline int32_t getLong(UErrorCode* status) const;
#endif  /* U_HIDE_DEPRECATED_API */

#ifndef U_HIDE_INTERNAL_API
    /**
     * Internal function, do not use.
     * TODO:  figure out how to make this be non-public.
     *        NumberFormat::format(Formattable, ...
     *        needs to get at the DecimalQuantity, if it exists, for
     *        big decimal formatting.
     *  @internal
     */
    number::impl::DecimalQuantity *getDecimalQuantity() const { return fDecimalQuantity;}

    /**
     * Export the value of this Formattable to a DecimalQuantity.
     * @internal
     */
    void populateDecimalQuantity(number::impl::DecimalQuantity& output, UErrorCode& status) const;

    /**
     *  Adopt, and set value from, a DecimalQuantity
     *     Internal Function, do not use.
     *  @param dq the DecimalQuantity to be adopted
     *  @internal
     */
    void adoptDecimalQuantity(number::impl::DecimalQuantity *dq);

    /**
     * Internal function to return the CharString pointer.
     * @param status error code
     * @return pointer to the CharString - may become invalid if the object is modified
     * @internal
     */
    CharString *internalGetCharString(UErrorCode &status);

#endif  /* U_HIDE_INTERNAL_API */

private:
    /**
     * Cleans up the memory for unwanted values.  For example, the adopted
     * string or array objects.
     */
    void            dispose(void);

    /**
     * Common initialization, for use by constructors.
     */
    void            init();

    UnicodeString* getBogus() const;

    union {
        UObject*        fObject;
        UnicodeString*  fString;
        double          fDouble;
        int64_t         fInt64;
        UDate           fDate;
        struct {
          Formattable*  fArray;
          int32_t       fCount;
        }               fArrayAndCount;
    } fValue;

    CharString           *fDecimalStr;

    number::impl::DecimalQuantity *fDecimalQuantity;

    Type                fType;
    UnicodeString       fBogus; // Bogus string when it's needed.
};

inline UDate Formattable::getDate(UErrorCode& status) const {
    if (fType != kDate) {
        if (U_SUCCESS(status)) {
            status = U_INVALID_FORMAT_ERROR;
        }
        return 0;
    }
    return fValue.fDate;
}

inline const UnicodeString& Formattable::getString(void) const {
    return *fValue.fString;
}

inline UnicodeString& Formattable::getString(void) {
    return *fValue.fString;
}

#ifndef U_HIDE_DEPRECATED_API
inline int32_t Formattable::getLong(UErrorCode* status) const {
    return getLong(*status);
}
#endif  /* U_HIDE_DEPRECATED_API */

inline UFormattable* Formattable::toUFormattable() {
  return reinterpret_cast<UFormattable*>(this);
}

inline const UFormattable* Formattable::toUFormattable() const {
  return reinterpret_cast<const UFormattable*>(this);
}

inline Formattable* Formattable::fromUFormattable(UFormattable *fmt) {
  return reinterpret_cast<Formattable *>(fmt);
}

inline const Formattable* Formattable::fromUFormattable(const UFormattable *fmt) {
  return reinterpret_cast<const Formattable *>(fmt);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 2010-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File attiter.h
*
* Modification History:
*
*   Date        Name        Description
*   12/15/2009  dougfelt    Created
********************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API



/**
 * \file
 * \brief C++ API: FieldPosition Iterator.
 */

#if UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

/*
 * Allow the declaration of APIs with pointers to FieldPositionIterator
 * even when formatting is removed from the build.
 */
class FieldPositionIterator;

U_NAMESPACE_END

#else




U_NAMESPACE_BEGIN

class UVector32;

/**
 * FieldPositionIterator returns the field ids and their start/limit positions generated
 * by a call to Format::format.  See Format, NumberFormat, DecimalFormat.
 * @stable ICU 4.4
 */
class U_I18N_API FieldPositionIterator : public UObject {
public:
    /**
     * Destructor.
     * @stable ICU 4.4
     */
    ~FieldPositionIterator();

    /**
     * Constructs a new, empty iterator.
     * @stable ICU 4.4
     */
    FieldPositionIterator(void);

    /**
     * Copy constructor.  If the copy failed for some reason, the new iterator will
     * be empty.
     * @stable ICU 4.4
     */
    FieldPositionIterator(const FieldPositionIterator&);

    /**
     * Return true if another object is semantically equal to this
     * one.
     * <p>
     * Return true if this FieldPositionIterator is at the same position in an
     * equal array of run values.
     * @stable ICU 4.4
     */
    UBool operator==(const FieldPositionIterator&) const;

    /**
     * Returns the complement of the result of operator==
     * @param rhs The FieldPositionIterator to be compared for inequality
     * @return the complement of the result of operator==
     * @stable ICU 4.4
     */
    UBool operator!=(const FieldPositionIterator& rhs) const { return !operator==(rhs); }

    /**
     * If the current position is valid, updates the FieldPosition values, advances the iterator,
     * and returns TRUE, otherwise returns FALSE.
     * @stable ICU 4.4
     */
    UBool next(FieldPosition& fp);

private:
    /**
     * Sets the data used by the iterator, and resets the position.
     * Returns U_ILLEGAL_ARGUMENT_ERROR in status if the data is not valid 
     * (length is not a multiple of 3, or start >= limit for any run).
     */
    void setData(UVector32 *adopt, UErrorCode& status);

    friend class FieldPositionIteratorHandler;

    UVector32 *data;
    int32_t pos;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



 


U_NAMESPACE_BEGIN

/**
 * Base class for all formats.  This is an abstract base class which
 * specifies the protocol for classes which convert other objects or
 * values, such as numeric values and dates, and their string
 * representations.  In some cases these representations may be
 * localized or contain localized characters or strings.  For example,
 * a numeric formatter such as DecimalFormat may convert a numeric
 * value such as 12345 to the string "$12,345".  It may also parse
 * the string back into a numeric value.  A date and time formatter
 * like SimpleDateFormat may represent a specific date, encoded
 * numerically, as a string such as "Wednesday, February 26, 1997 AD".
 * <P>
 * Many of the concrete subclasses of Format employ the notion of
 * a pattern.  A pattern is a string representation of the rules which
 * govern the interconversion between values and strings.  For example,
 * a DecimalFormat object may be associated with the pattern
 * "$#,##0.00;($#,##0.00)", which is a common US English format for
 * currency values, yielding strings such as "$1,234.45" for 1234.45,
 * and "($987.65)" for 987.6543.  The specific syntax of a pattern
 * is defined by each subclass.
 * <P>
 * Even though many subclasses use patterns, the notion of a pattern
 * is not inherent to Format classes in general, and is not part of
 * the explicit base class protocol.
 * <P>
 * Two complex formatting classes bear mentioning.  These are
 * MessageFormat and ChoiceFormat.  ChoiceFormat is a subclass of
 * NumberFormat which allows the user to format different number ranges
 * as strings.  For instance, 0 may be represented as "no files", 1 as
 * "one file", and any number greater than 1 as "many files".
 * MessageFormat is a formatter which utilizes other Format objects to
 * format a string containing with multiple values.  For instance,
 * A MessageFormat object might produce the string "There are no files
 * on the disk MyDisk on February 27, 1997." given the arguments 0,
 * "MyDisk", and the date value of 2/27/97.  See the ChoiceFormat
 * and MessageFormat headers for further information.
 * <P>
 * If formatting is unsuccessful, a failing UErrorCode is returned when
 * the Format cannot format the type of object, otherwise if there is
 * something illformed about the the Unicode replacement character
 * 0xFFFD is returned.
 * <P>
 * If there is no match when parsing, a parse failure UErrorCode is
 * retured for methods which take no ParsePosition.  For the method
 * that takes a ParsePosition, the index parameter is left unchanged.
 * <P>
 * <em>User subclasses are not supported.</em> While clients may write
 * subclasses, such code will not necessarily work and will not be
 * guaranteed to work stably from release to release.
 */
class U_I18N_API Format : public UObject {
public:

    /** Destructor
     * @stable ICU 2.4
     */
    virtual ~Format();

    /**
     * Return true if the given Format objects are semantically equal.
     * Objects of different subclasses are considered unequal.
     * @param other    the object to be compared with.
     * @return         Return true if the given Format objects are semantically equal.
     *                 Objects of different subclasses are considered unequal.
     * @stable ICU 2.0
     */
    virtual UBool operator==(const Format& other) const = 0;

    /**
     * Return true if the given Format objects are not semantically
     * equal.
     * @param other    the object to be compared with.
     * @return         Return true if the given Format objects are not semantically.
     * @stable ICU 2.0
     */
    UBool operator!=(const Format& other) const { return !operator==(other); }

    /**
     * Clone this object polymorphically.  The caller is responsible
     * for deleting the result when done.
     * @return    A copy of the object
     * @stable ICU 2.0
     */
    virtual Format* clone() const = 0;

    /**
     * Formats an object to produce a string.
     *
     * @param obj       The object to format.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param status    Output parameter filled in with success or failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    UnicodeString& format(const Formattable& obj,
                          UnicodeString& appendTo,
                          UErrorCode& status) const;

    /**
     * Format an object to produce a string.  This is a pure virtual method which
     * subclasses must implement. This method allows polymorphic formatting
     * of Formattable objects. If a subclass of Format receives a Formattable
     * object type it doesn't handle (e.g., if a numeric Formattable is passed
     * to a DateFormat object) then it returns a failing UErrorCode.
     *
     * @param obj       The object to format.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    virtual UnicodeString& format(const Formattable& obj,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode& status) const = 0;
    /**
     * Format an object to produce a string.  Subclasses should override this
     * method. This method allows polymorphic formatting of Formattable objects.
     * If a subclass of Format receives a Formattable object type it doesn't
     * handle (e.g., if a numeric Formattable is passed to a DateFormat object)
     * then it returns a failing UErrorCode.
     *
     * @param obj       The object to format.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    virtual UnicodeString& format(const Formattable& obj,
                                  UnicodeString& appendTo,
                                  FieldPositionIterator* posIter,
                                  UErrorCode& status) const;

    /**
     * Parse a string to produce an object.  This is a pure virtual
     * method which subclasses must implement.  This method allows
     * polymorphic parsing of strings into Formattable objects.
     * <P>
     * Before calling, set parse_pos.index to the offset you want to
     * start parsing at in the source.  After calling, parse_pos.index
     * is the end of the text you parsed.  If error occurs, index is
     * unchanged.
     * <P>
     * When parsing, leading whitespace is discarded (with successful
     * parse), while trailing whitespace is left as is.
     * <P>
     * Example:
     * <P>
     * Parsing "_12_xy" (where _ represents a space) for a number,
     * with index == 0 will result in the number 12, with
     * parse_pos.index updated to 3 (just before the second space).
     * Parsing a second time will result in a failing UErrorCode since
     * "xy" is not a number, and leave index at 3.
     * <P>
     * Subclasses will typically supply specific parse methods that
     * return different types of values. Since methods can't overload
     * on return types, these will typically be named "parse", while
     * this polymorphic method will always be called parseObject.  Any
     * parse method that does not take a parse_pos should set status
     * to an error value when no text in the required format is at the
     * start position.
     *
     * @param source    The string to be parsed into an object.
     * @param result    Formattable to be set to the parse result.
     *                  If parse fails, return contents are undefined.
     * @param parse_pos The position to start parsing at. Upon return
     *                  this param is set to the position after the
     *                  last character successfully parsed. If the
     *                  source is not parsed successfully, this param
     *                  will remain unchanged.
     * @stable ICU 2.0
     */
    virtual void parseObject(const UnicodeString& source,
                             Formattable& result,
                             ParsePosition& parse_pos) const = 0;

    /**
     * Parses a string to produce an object. This is a convenience method
     * which calls the pure virtual parseObject() method, and returns a
     * failure UErrorCode if the ParsePosition indicates failure.
     *
     * @param source    The string to be parsed into an object.
     * @param result    Formattable to be set to the parse result.
     *                  If parse fails, return contents are undefined.
     * @param status    Output param to be filled with success/failure
     *                  result code.
     * @stable ICU 2.0
     */
    void parseObject(const UnicodeString& source,
                     Formattable& result,
                     UErrorCode& status) const;

    /** Get the locale for this format object. You can choose between valid and actual locale.
     *  @param type type of the locale we're looking for (valid or actual) 
     *  @param status error code for the operation
     *  @return the locale
     *  @stable ICU 2.8
     */
    Locale getLocale(ULocDataLocaleType type, UErrorCode& status) const;

#ifndef U_HIDE_INTERNAL_API
    /** Get the locale for this format object. You can choose between valid and actual locale.
     *  @param type type of the locale we're looking for (valid or actual) 
     *  @param status error code for the operation
     *  @return the locale
     *  @internal
     */
    const char* getLocaleID(ULocDataLocaleType type, UErrorCode &status) const;
#endif  /* U_HIDE_INTERNAL_API */

 protected:
    /** @stable ICU 2.8 */
    void setLocaleIDs(const char* valid, const char* actual);

protected:
    /**
     * Default constructor for subclass use only.  Does nothing.
     * @stable ICU 2.0
     */
    Format();

    /**
     * @stable ICU 2.0
     */
    Format(const Format&); // Does nothing; for subclasses only

    /**
     * @stable ICU 2.0
     */
    Format& operator=(const Format&); // Does nothing; for subclasses

       
    /**
     * Simple function for initializing a UParseError from a UnicodeString.
     *
     * @param pattern The pattern to copy into the parseError
     * @param pos The position in pattern where the error occured
     * @param parseError The UParseError object to fill in
     * @stable ICU 2.4
     */
    static void syntaxError(const UnicodeString& pattern,
                            int32_t pos,
                            UParseError& parseError);

 private:
    char actualLocale[ULOC_FULLNAME_CAPACITY];
    char validLocale[ULOC_FULLNAME_CAPACITY];
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
* Copyright (C) 1997-2016, International Business Machines Corporation and others.
* All Rights Reserved.
********************************************************************************
*
* File NUMFMT.H
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/18/97    clhuang     Updated per C++ implementation.
*   04/17/97    aliu        Changed DigitCount to int per code review.
*    07/20/98    stephen        JDK 1.2 sync up. Added scientific support.
*                            Changed naming conventions to match C++ guidelines
*                            Derecated Java style constants (eg, INTEGER_FIELD)
********************************************************************************
*/





#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Compatibility APIs for number formatting.
 */

#if !UCONFIG_NO_FORMATTING



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2015, International Business Machines Corporation and others.
* All Rights Reserved.
* Modification History:
*
*   Date        Name        Description
*   06/24/99    helena      Integrated Alan's NF enhancements and Java2 bug fixes
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING








// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2015-2016, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/




#if !UCONFIG_NO_FORMATTING



/**
 * \file
 * \brief C API: UFieldPositionIterator for use with format APIs.
 *
 * Usage:
 * ufieldpositer_open creates an empty (unset) UFieldPositionIterator.
 * This can be passed to format functions such as {@link #udat_formatForFields},
 * which will set it to apply to the fields in a particular formatted string.
 * ufieldpositer_next can then be used to iterate over those fields,
 * providing for each field its type (using values that are specific to the
 * particular format type, such as date or number formats), as well as the
 * start and end positions of the field in the formatted string.
 * A given UFieldPositionIterator can be re-used for different format calls;
 * each such call resets it to apply to that format string.
 * ufieldpositer_close should be called to dispose of the UFieldPositionIterator
 * when it is no longer needed.
 *
 * @see FieldPositionIterator
 */

/**
 * Opaque UFieldPositionIterator object for use in C.
 * @stable ICU 55
 */
struct UFieldPositionIterator;
typedef struct UFieldPositionIterator UFieldPositionIterator;  /**< C typedef for struct UFieldPositionIterator. @stable ICU 55 */

/**
 * Open a new, unset UFieldPositionIterator object.
 * @param status
 *          A pointer to a UErrorCode to receive any errors.
 * @return
 *          A pointer to an empty (unset) UFieldPositionIterator object,
 *          or NULL if an error occurred.
 * @stable ICU 55
 */
U_STABLE UFieldPositionIterator* U_EXPORT2
ufieldpositer_open(UErrorCode* status);

/**
 * Close a UFieldPositionIterator object. Once closed it may no longer be used.
 * @param fpositer
 *          A pointer to the UFieldPositionIterator object to close.
 * @stable ICU 55
 */
U_STABLE void U_EXPORT2
ufieldpositer_close(UFieldPositionIterator *fpositer);


#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUFieldPositionIteratorPointer
 * "Smart pointer" class, closes a UFieldPositionIterator via ufieldpositer_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 55
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUFieldPositionIteratorPointer, UFieldPositionIterator, ufieldpositer_close);

U_NAMESPACE_END

#endif

/**
 * Get information for the next field in the formatted string to which this
 * UFieldPositionIterator currently applies, or return a negative value if there
 * are no more fields.
 * @param fpositer
 *          A pointer to the UFieldPositionIterator object containing iteration
 *          state for the format fields.
 * @param beginIndex
 *          A pointer to an int32_t to receive information about the start offset
 *          of the field in the formatted string (undefined if the function
 *          returns a negative value). May be NULL if this information is not needed.
 * @param endIndex
 *          A pointer to an int32_t to receive information about the end offset
 *          of the field in the formatted string (undefined if the function
 *          returns a negative value). May be NULL if this information is not needed.
 * @return
 *          The field type (non-negative value), or a negative value if there are
 *          no more fields for which to provide information. If negative, then any
 *          values pointed to by beginIndex and endIndex are undefined.
 *
 *          The values for field type depend on what type of formatter the
 *          UFieldPositionIterator has been set by; for a date formatter, the
 *          values from the UDateFormatField enum. For more information, see the
 *          descriptions of format functions that take a UFieldPositionIterator*
 *          parameter, such as {@link #udat_formatForFields}.
 *
 * @stable ICU 55
 */
U_STABLE int32_t U_EXPORT2
ufieldpositer_next(UFieldPositionIterator *fpositer,
                   int32_t *beginIndex, int32_t *endIndex);

#endif /* #if !UCONFIG_NO_FORMATTING */



/**
 * \file
 * \brief C API: Compatibility APIs for number formatting.
 *
 * <h2> Number Format C API </h2>
 *
 * <p><strong>IMPORTANT:</strong> New users with are strongly encouraged to
 * see if unumberformatter.h fits their use case.  Although not deprecated,
 * this header is provided for backwards compatibility only.
 *
 * Number Format C API  Provides functions for
 * formatting and parsing a number.  Also provides methods for
 * determining which locales have number formats, and what their names
 * are.
 * <P>
 * UNumberFormat helps you to format and parse numbers for any locale.
 * Your code can be completely independent of the locale conventions
 * for decimal points, thousands-separators, or even the particular
 * decimal digits used, or whether the number format is even decimal.
 * There are different number format styles like decimal, currency,
 * percent and spellout.
 * <P>
 * To format a number for the current Locale, use one of the static
 * factory methods:
 * <pre>
 * \code
 *    UChar myString[20];
 *    double myNumber = 7.0;
 *    UErrorCode status = U_ZERO_ERROR;
 *    UNumberFormat* nf = unum_open(UNUM_DEFAULT, NULL, -1, NULL, NULL, &status);
 *    unum_formatDouble(nf, myNumber, myString, 20, NULL, &status);
 *    printf(" Example 1: %s\n", austrdup(myString) ); //austrdup( a function used to convert UChar* to char*)
 * \endcode
 * </pre>
 * If you are formatting multiple numbers, it is more efficient to get
 * the format and use it multiple times so that the system doesn't
 * have to fetch the information about the local language and country
 * conventions multiple times.
 * <pre>
 * \code
 * uint32_t i, resultlength, reslenneeded;
 * UErrorCode status = U_ZERO_ERROR;
 * UFieldPosition pos;
 * uint32_t a[] = { 123, 3333, -1234567 };
 * const uint32_t a_len = sizeof(a) / sizeof(a[0]);
 * UNumberFormat* nf;
 * UChar* result = NULL;
 *
 * nf = unum_open(UNUM_DEFAULT, NULL, -1, NULL, NULL, &status);
 * for (i = 0; i < a_len; i++) {
 *    resultlength=0;
 *    reslenneeded=unum_format(nf, a[i], NULL, resultlength, &pos, &status);
 *    result = NULL;
 *    if(status==U_BUFFER_OVERFLOW_ERROR){
 *       status=U_ZERO_ERROR;
 *       resultlength=reslenneeded+1;
 *       result=(UChar*)malloc(sizeof(UChar) * resultlength);
 *       unum_format(nf, a[i], result, resultlength, &pos, &status);
 *    }
 *    printf( " Example 2: %s\n", austrdup(result));
 *    free(result);
 * }
 * \endcode
 * </pre>
 * To format a number for a different Locale, specify it in the
 * call to unum_open().
 * <pre>
 * \code
 *     UNumberFormat* nf = unum_open(UNUM_DEFAULT, NULL, -1, "fr_FR", NULL, &success)
 * \endcode
 * </pre>
 * You can use a NumberFormat API unum_parse() to parse.
 * <pre>
 * \code
 *    UErrorCode status = U_ZERO_ERROR;
 *    int32_t pos=0;
 *    int32_t num;
 *    num = unum_parse(nf, str, u_strlen(str), &pos, &status);
 * \endcode
 * </pre>
 * Use UNUM_DECIMAL to get the normal number format for that country.
 * There are other static options available.  Use UNUM_CURRENCY
 * to get the currency number format for that country.  Use UNUM_PERCENT
 * to get a format for displaying percentages. With this format, a
 * fraction from 0.53 is displayed as 53%.
 * <P>
 * Use a pattern to create either a DecimalFormat or a RuleBasedNumberFormat
 * formatter.  The pattern must conform to the syntax defined for those
 * formatters.
 * <P>
 * You can also control the display of numbers with such function as
 * unum_getAttributes() and unum_setAttributes(), which let you set the
 * minimum fraction digits, grouping, etc.
 * @see UNumberFormatAttributes for more details
 * <P>
 * You can also use forms of the parse and format methods with
 * ParsePosition and UFieldPosition to allow you to:
 * <ul type=round>
 *   <li>(a) progressively parse through pieces of a string.
 *   <li>(b) align the decimal point and other areas.
 * </ul>
 * <p>
 * It is also possible to change or set the symbols used for a particular
 * locale like the currency symbol, the grouping separator , monetary separator
 * etc by making use of functions unum_setSymbols() and unum_getSymbols().
 */

/** A number formatter.
 *  For usage in C programs.
 *  @stable ICU 2.0
 */
typedef void* UNumberFormat;

/** The possible number format styles.
 *  @stable ICU 2.0
 */
typedef enum UNumberFormatStyle {
    /**
     * Decimal format defined by a pattern string.
     * @stable ICU 3.0
     */
    UNUM_PATTERN_DECIMAL=0,
    /**
     * Decimal format ("normal" style).
     * @stable ICU 2.0
     */
    UNUM_DECIMAL=1,
    /**
     * Currency format (generic).
     * Defaults to UNUM_CURRENCY_STANDARD style
     * (using currency symbol, e.g., "$1.00", with non-accounting
     * style for negative values e.g. using minus sign).
     * The specific style may be specified using the -cf- locale key.
     * @stable ICU 2.0
     */
    UNUM_CURRENCY=2,
    /**
     * Percent format
     * @stable ICU 2.0
     */
    UNUM_PERCENT=3,
    /**
     * Scientific format
     * @stable ICU 2.1
     */
    UNUM_SCIENTIFIC=4,
    /**
     * Spellout rule-based format. The default ruleset can be specified/changed using
     * unum_setTextAttribute with UNUM_DEFAULT_RULESET; the available public rulesets
     * can be listed using unum_getTextAttribute with UNUM_PUBLIC_RULESETS.
     * @stable ICU 2.0
     */
    UNUM_SPELLOUT=5,
    /**
     * Ordinal rule-based format . The default ruleset can be specified/changed using
     * unum_setTextAttribute with UNUM_DEFAULT_RULESET; the available public rulesets
     * can be listed using unum_getTextAttribute with UNUM_PUBLIC_RULESETS.
     * @stable ICU 3.0
     */
    UNUM_ORDINAL=6,
    /**
     * Duration rule-based format
     * @stable ICU 3.0
     */
    UNUM_DURATION=7,
    /**
     * Numbering system rule-based format
     * @stable ICU 4.2
     */
    UNUM_NUMBERING_SYSTEM=8,
    /**
     * Rule-based format defined by a pattern string.
     * @stable ICU 3.0
     */
    UNUM_PATTERN_RULEBASED=9,
    /**
     * Currency format with an ISO currency code, e.g., "USD1.00".
     * @stable ICU 4.8
     */
    UNUM_CURRENCY_ISO=10,
    /**
     * Currency format with a pluralized currency name,
     * e.g., "1.00 US dollar" and "3.00 US dollars".
     * @stable ICU 4.8
     */
    UNUM_CURRENCY_PLURAL=11,
    /**
     * Currency format for accounting, e.g., "($3.00)" for
     * negative currency amount instead of "-$3.00" ({@link #UNUM_CURRENCY}).
     * Overrides any style specified using -cf- key in locale.
     * @stable ICU 53
     */
    UNUM_CURRENCY_ACCOUNTING=12,
    /**
     * Currency format with a currency symbol given CASH usage, e.g.,
     * "NT$3" instead of "NT$3.23".
     * @stable ICU 54
     */
    UNUM_CASH_CURRENCY=13,
    /**
     * Decimal format expressed using compact notation
     * (short form, corresponds to UNumberCompactStyle=UNUM_SHORT)
     * e.g. "23K", "45B"
     * @stable ICU 56
     */
    UNUM_DECIMAL_COMPACT_SHORT=14,
    /**
     * Decimal format expressed using compact notation
     * (long form, corresponds to UNumberCompactStyle=UNUM_LONG)
     * e.g. "23 thousand", "45 billion"
     * @stable ICU 56
     */
    UNUM_DECIMAL_COMPACT_LONG=15,
    /**
     * Currency format with a currency symbol, e.g., "$1.00",
     * using non-accounting style for negative values (e.g. minus sign).
     * Overrides any style specified using -cf- key in locale.
     * @stable ICU 56
     */
    UNUM_CURRENCY_STANDARD=16,

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UNumberFormatStyle value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UNUM_FORMAT_STYLE_COUNT=17,
#endif  /* U_HIDE_DEPRECATED_API */

    /**
     * Default format
     * @stable ICU 2.0
     */
    UNUM_DEFAULT = UNUM_DECIMAL,
    /**
     * Alias for UNUM_PATTERN_DECIMAL
     * @stable ICU 3.0
     */
    UNUM_IGNORE = UNUM_PATTERN_DECIMAL
} UNumberFormatStyle;

/** The possible number format rounding modes.
 *
 * <p>
 * For more detail on rounding modes, see:
 * http://userguide.icu-project.org/formatparse/numbers/rounding-modes
 *
 * @stable ICU 2.0
 */
typedef enum UNumberFormatRoundingMode {
    UNUM_ROUND_CEILING,
    UNUM_ROUND_FLOOR,
    UNUM_ROUND_DOWN,
    UNUM_ROUND_UP,
    /**
     * Half-even rounding
     * @stable, ICU 3.8
     */
    UNUM_ROUND_HALFEVEN,
#ifndef U_HIDE_DEPRECATED_API
    /**
     * Half-even rounding, misspelled name
     * @deprecated, ICU 3.8
     */
    UNUM_FOUND_HALFEVEN = UNUM_ROUND_HALFEVEN,
#endif  /* U_HIDE_DEPRECATED_API */
    UNUM_ROUND_HALFDOWN = UNUM_ROUND_HALFEVEN + 1,
    UNUM_ROUND_HALFUP,
    /**
      * ROUND_UNNECESSARY reports an error if formatted result is not exact.
      * @stable ICU 4.8
      */
    UNUM_ROUND_UNNECESSARY
} UNumberFormatRoundingMode;

/** The possible number format pad positions.
 *  @stable ICU 2.0
 */
typedef enum UNumberFormatPadPosition {
    UNUM_PAD_BEFORE_PREFIX,
    UNUM_PAD_AFTER_PREFIX,
    UNUM_PAD_BEFORE_SUFFIX,
    UNUM_PAD_AFTER_SUFFIX
} UNumberFormatPadPosition;

/**
 * Constants for specifying short or long format.
 * @stable ICU 51
 */
typedef enum UNumberCompactStyle {
  /** @stable ICU 51 */
  UNUM_SHORT,
  /** @stable ICU 51 */
  UNUM_LONG
  /** @stable ICU 51 */
} UNumberCompactStyle;

/**
 * Constants for specifying currency spacing
 * @stable ICU 4.8
 */
enum UCurrencySpacing {
    /** @stable ICU 4.8 */
    UNUM_CURRENCY_MATCH,
    /** @stable ICU 4.8 */
    UNUM_CURRENCY_SURROUNDING_MATCH,
    /** @stable ICU 4.8 */
    UNUM_CURRENCY_INSERT,

    /* Do not conditionalize the following with #ifndef U_HIDE_DEPRECATED_API,
     * it is needed for layout of DecimalFormatSymbols object. */
#ifndef U_FORCE_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UCurrencySpacing value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UNUM_CURRENCY_SPACING_COUNT
#endif  // U_FORCE_HIDE_DEPRECATED_API
};
typedef enum UCurrencySpacing UCurrencySpacing; /**< @stable ICU 4.8 */


/**
 * FieldPosition and UFieldPosition selectors for format fields
 * defined by NumberFormat and UNumberFormat.
 * @stable ICU 49
 */
typedef enum UNumberFormatFields {
    /** @stable ICU 49 */
    UNUM_INTEGER_FIELD,
    /** @stable ICU 49 */
    UNUM_FRACTION_FIELD,
    /** @stable ICU 49 */
    UNUM_DECIMAL_SEPARATOR_FIELD,
    /** @stable ICU 49 */
    UNUM_EXPONENT_SYMBOL_FIELD,
    /** @stable ICU 49 */
    UNUM_EXPONENT_SIGN_FIELD,
    /** @stable ICU 49 */
    UNUM_EXPONENT_FIELD,
    /** @stable ICU 49 */
    UNUM_GROUPING_SEPARATOR_FIELD,
    /** @stable ICU 49 */
    UNUM_CURRENCY_FIELD,
    /** @stable ICU 49 */
    UNUM_PERCENT_FIELD,
    /** @stable ICU 49 */
    UNUM_PERMILL_FIELD,
    /** @stable ICU 49 */
    UNUM_SIGN_FIELD,
#ifndef U_HIDE_DRAFT_API
    /** @draft ICU 64 */
    UNUM_MEASURE_UNIT_FIELD,
    /** @draft ICU 64 */
    UNUM_COMPACT_FIELD,
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UNumberFormatFields value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UNUM_FIELD_COUNT = UNUM_SIGN_FIELD + 3
#endif  /* U_HIDE_DEPRECATED_API */
} UNumberFormatFields;


/**
 * Create and return a new UNumberFormat for formatting and parsing
 * numbers.  A UNumberFormat may be used to format numbers by calling
 * {@link #unum_format }, and to parse numbers by calling {@link #unum_parse }.
 * The caller must call {@link #unum_close } when done to release resources
 * used by this object.
 * @param style The type of number format to open: one of
 * UNUM_DECIMAL, UNUM_CURRENCY, UNUM_PERCENT, UNUM_SCIENTIFIC,
 * UNUM_CURRENCY_ISO, UNUM_CURRENCY_PLURAL, UNUM_SPELLOUT,
 * UNUM_ORDINAL, UNUM_DURATION, UNUM_NUMBERING_SYSTEM,
 * UNUM_PATTERN_DECIMAL, UNUM_PATTERN_RULEBASED, or UNUM_DEFAULT.
 * If UNUM_PATTERN_DECIMAL or UNUM_PATTERN_RULEBASED is passed then the
 * number format is opened using the given pattern, which must conform
 * to the syntax described in DecimalFormat or RuleBasedNumberFormat,
 * respectively.
 *
 * <p><strong>NOTE::</strong> New users with are strongly encouraged to
 * use unumf_openForSkeletonAndLocale instead of unum_open.
 *
 * @param pattern A pattern specifying the format to use.
 * This parameter is ignored unless the style is
 * UNUM_PATTERN_DECIMAL or UNUM_PATTERN_RULEBASED.
 * @param patternLength The number of characters in the pattern, or -1
 * if null-terminated. This parameter is ignored unless the style is
 * UNUM_PATTERN.
 * @param locale A locale identifier to use to determine formatting
 * and parsing conventions, or NULL to use the default locale.
 * @param parseErr A pointer to a UParseError struct to receive the
 * details of any parsing errors, or NULL if no parsing error details
 * are desired.
 * @param status A pointer to an input-output UErrorCode.
 * @return A pointer to a newly created UNumberFormat, or NULL if an
 * error occurred.
 * @see unum_close
 * @see DecimalFormat
 * @stable ICU 2.0
 */
U_STABLE UNumberFormat* U_EXPORT2
unum_open(  UNumberFormatStyle    style,
            const    UChar*    pattern,
            int32_t            patternLength,
            const    char*     locale,
            UParseError*       parseErr,
            UErrorCode*        status);


/**
* Close a UNumberFormat.
* Once closed, a UNumberFormat may no longer be used.
* @param fmt The formatter to close.
* @stable ICU 2.0
*/
U_STABLE void U_EXPORT2
unum_close(UNumberFormat* fmt);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUNumberFormatPointer
 * "Smart pointer" class, closes a UNumberFormat via unum_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUNumberFormatPointer, UNumberFormat, unum_close);

U_NAMESPACE_END

#endif

/**
 * Open a copy of a UNumberFormat.
 * This function performs a deep copy.
 * @param fmt The format to copy
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return A pointer to a UNumberFormat identical to fmt.
 * @stable ICU 2.0
 */
U_STABLE UNumberFormat* U_EXPORT2
unum_clone(const UNumberFormat *fmt,
       UErrorCode *status);

/**
* Format an integer using a UNumberFormat.
* The integer will be formatted according to the UNumberFormat's locale.
* @param fmt The formatter to use.
* @param number The number to format.
* @param result A pointer to a buffer to receive the NULL-terminated formatted number. If
* the formatted number fits into dest but cannot be NULL-terminated (length == resultLength)
* then the error code is set to U_STRING_NOT_TERMINATED_WARNING. If the formatted number
* doesn't fit into result then the error code is set to U_BUFFER_OVERFLOW_ERROR.
* @param resultLength The maximum size of result.
* @param pos    A pointer to a UFieldPosition.  On input, position->field
* is read.  On output, position->beginIndex and position->endIndex indicate
* the beginning and ending indices of field number position->field, if such
* a field exists.  This parameter may be NULL, in which case no field
* @param status A pointer to an UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see unum_formatInt64
* @see unum_formatDouble
* @see unum_parse
* @see unum_parseInt64
* @see unum_parseDouble
* @see UFieldPosition
* @stable ICU 2.0
*/
U_STABLE int32_t U_EXPORT2
unum_format(    const    UNumberFormat*    fmt,
        int32_t            number,
        UChar*            result,
        int32_t            resultLength,
        UFieldPosition    *pos,
        UErrorCode*        status);

/**
* Format an int64 using a UNumberFormat.
* The int64 will be formatted according to the UNumberFormat's locale.
* @param fmt The formatter to use.
* @param number The number to format.
* @param result A pointer to a buffer to receive the NULL-terminated formatted number. If
* the formatted number fits into dest but cannot be NULL-terminated (length == resultLength)
* then the error code is set to U_STRING_NOT_TERMINATED_WARNING. If the formatted number
* doesn't fit into result then the error code is set to U_BUFFER_OVERFLOW_ERROR.
* @param resultLength The maximum size of result.
* @param pos    A pointer to a UFieldPosition.  On input, position->field
* is read.  On output, position->beginIndex and position->endIndex indicate
* the beginning and ending indices of field number position->field, if such
* a field exists.  This parameter may be NULL, in which case no field
* @param status A pointer to an UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see unum_format
* @see unum_formatDouble
* @see unum_parse
* @see unum_parseInt64
* @see unum_parseDouble
* @see UFieldPosition
* @stable ICU 2.0
*/
U_STABLE int32_t U_EXPORT2
unum_formatInt64(const UNumberFormat *fmt,
        int64_t         number,
        UChar*          result,
        int32_t         resultLength,
        UFieldPosition *pos,
        UErrorCode*     status);

/**
* Format a double using a UNumberFormat.
* The double will be formatted according to the UNumberFormat's locale.
* @param fmt The formatter to use.
* @param number The number to format.
* @param result A pointer to a buffer to receive the NULL-terminated formatted number. If
* the formatted number fits into dest but cannot be NULL-terminated (length == resultLength)
* then the error code is set to U_STRING_NOT_TERMINATED_WARNING. If the formatted number
* doesn't fit into result then the error code is set to U_BUFFER_OVERFLOW_ERROR.
* @param resultLength The maximum size of result.
* @param pos    A pointer to a UFieldPosition.  On input, position->field
* is read.  On output, position->beginIndex and position->endIndex indicate
* the beginning and ending indices of field number position->field, if such
* a field exists.  This parameter may be NULL, in which case no field
* @param status A pointer to an UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see unum_format
* @see unum_formatInt64
* @see unum_parse
* @see unum_parseInt64
* @see unum_parseDouble
* @see UFieldPosition
* @stable ICU 2.0
*/
U_STABLE int32_t U_EXPORT2
unum_formatDouble(    const    UNumberFormat*  fmt,
            double          number,
            UChar*          result,
            int32_t         resultLength,
            UFieldPosition  *pos, /* 0 if ignore */
            UErrorCode*     status);

/**
* Format a double using a UNumberFormat according to the UNumberFormat's locale,
* and initialize a UFieldPositionIterator that enumerates the subcomponents of
* the resulting string.
*
* @param format
*          The formatter to use.
* @param number
*          The number to format.
* @param result
*          A pointer to a buffer to receive the NULL-terminated formatted
*          number. If the formatted number fits into dest but cannot be
*          NULL-terminated (length == resultLength) then the error code is set
*          to U_STRING_NOT_TERMINATED_WARNING. If the formatted number doesn't
*          fit into result then the error code is set to
*          U_BUFFER_OVERFLOW_ERROR.
* @param resultLength
*          The maximum size of result.
* @param fpositer
*          A pointer to a UFieldPositionIterator created by {@link #ufieldpositer_open}
*          (may be NULL if field position information is not needed, but in this
*          case it's preferable to use {@link #unum_formatDouble}). Iteration
*          information already present in the UFieldPositionIterator is deleted,
*          and the iterator is reset to apply to the fields in the formatted
*          string created by this function call. The field values and indexes
*          returned by {@link #ufieldpositer_next} represent fields denoted by
*          the UNumberFormatFields enum. Fields are not returned in a guaranteed
*          order. Fields cannot overlap, but they may nest. For example, 1234
*          could format as "1,234" which might consist of a grouping separator
*          field for ',' and an integer field encompassing the entire string.
* @param status
*          A pointer to an UErrorCode to receive any errors
* @return
*          The total buffer size needed; if greater than resultLength, the
*          output was truncated.
* @see unum_formatDouble
* @see unum_parse
* @see unum_parseDouble
* @see UFieldPositionIterator
* @see UNumberFormatFields
* @stable ICU 59
*/
U_STABLE int32_t U_EXPORT2
unum_formatDoubleForFields(const UNumberFormat* format,
                           double number,
                           UChar* result,
                           int32_t resultLength,
                           UFieldPositionIterator* fpositer,
                           UErrorCode* status);


/**
* Format a decimal number using a UNumberFormat.
* The number will be formatted according to the UNumberFormat's locale.
* The syntax of the input number is a "numeric string"
* as defined in the Decimal Arithmetic Specification, available at
* http://speleotrove.com/decimal
* @param fmt The formatter to use.
* @param number The number to format.
* @param length The length of the input number, or -1 if the input is nul-terminated.
* @param result A pointer to a buffer to receive the NULL-terminated formatted number. If
* the formatted number fits into dest but cannot be NULL-terminated (length == resultLength)
* then the error code is set to U_STRING_NOT_TERMINATED_WARNING. If the formatted number
* doesn't fit into result then the error code is set to U_BUFFER_OVERFLOW_ERROR.
* @param resultLength The maximum size of result.
* @param pos    A pointer to a UFieldPosition.  On input, position->field
*               is read.  On output, position->beginIndex and position->endIndex indicate
*               the beginning and ending indices of field number position->field, if such
*               a field exists.  This parameter may be NULL, in which case it is ignored.
* @param status A pointer to an UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see unum_format
* @see unum_formatInt64
* @see unum_parse
* @see unum_parseInt64
* @see unum_parseDouble
* @see UFieldPosition
* @stable ICU 4.4
*/
U_STABLE int32_t U_EXPORT2
unum_formatDecimal(    const    UNumberFormat*  fmt,
            const char *    number,
            int32_t         length,
            UChar*          result,
            int32_t         resultLength,
            UFieldPosition  *pos, /* 0 if ignore */
            UErrorCode*     status);

/**
 * Format a double currency amount using a UNumberFormat.
 * The double will be formatted according to the UNumberFormat's locale.
 * @param fmt the formatter to use
 * @param number the number to format
 * @param currency the 3-letter null-terminated ISO 4217 currency code
 * @param result A pointer to a buffer to receive the NULL-terminated formatted number. If
 * the formatted number fits into dest but cannot be NULL-terminated (length == resultLength)
 * then the error code is set to U_STRING_NOT_TERMINATED_WARNING. If the formatted number
 * doesn't fit into result then the error code is set to U_BUFFER_OVERFLOW_ERROR.
 * @param resultLength the maximum number of UChars to write to result
 * @param pos a pointer to a UFieldPosition.  On input,
 * position->field is read.  On output, position->beginIndex and
 * position->endIndex indicate the beginning and ending indices of
 * field number position->field, if such a field exists.  This
 * parameter may be NULL, in which case it is ignored.
 * @param status a pointer to an input-output UErrorCode
 * @return the total buffer size needed; if greater than resultLength,
 * the output was truncated.
 * @see unum_formatDouble
 * @see unum_parseDoubleCurrency
 * @see UFieldPosition
 * @stable ICU 3.0
 */
U_STABLE int32_t U_EXPORT2
unum_formatDoubleCurrency(const UNumberFormat* fmt,
                          double number,
                          UChar* currency,
                          UChar* result,
                          int32_t resultLength,
                          UFieldPosition* pos,
                          UErrorCode* status);

/**
 * Format a UFormattable into a string.
 * @param fmt the formatter to use
 * @param number the number to format, as a UFormattable
 * @param result A pointer to a buffer to receive the NULL-terminated formatted number. If
 * the formatted number fits into dest but cannot be NULL-terminated (length == resultLength)
 * then the error code is set to U_STRING_NOT_TERMINATED_WARNING. If the formatted number
 * doesn't fit into result then the error code is set to U_BUFFER_OVERFLOW_ERROR.
 * @param resultLength the maximum number of UChars to write to result
 * @param pos a pointer to a UFieldPosition.  On input,
 * position->field is read.  On output, position->beginIndex and
 * position->endIndex indicate the beginning and ending indices of
 * field number position->field, if such a field exists.  This
 * parameter may be NULL, in which case it is ignored.
 * @param status a pointer to an input-output UErrorCode
 * @return the total buffer size needed; if greater than resultLength,
 * the output was truncated. Will return 0 on error.
 * @see unum_parseToUFormattable
 * @stable ICU 52
 */
U_STABLE int32_t U_EXPORT2
unum_formatUFormattable(const UNumberFormat* fmt,
                        const UFormattable *number,
                        UChar *result,
                        int32_t resultLength,
                        UFieldPosition *pos,
                        UErrorCode *status);

/**
* Parse a string into an integer using a UNumberFormat.
* The string will be parsed according to the UNumberFormat's locale.
* Note: parsing is not supported for styles UNUM_DECIMAL_COMPACT_SHORT
* and UNUM_DECIMAL_COMPACT_LONG.
* @param fmt The formatter to use.
* @param text The text to parse.
* @param textLength The length of text, or -1 if null-terminated.
* @param parsePos If not NULL, on input a pointer to an integer specifying the offset at which
* to begin parsing.  If not NULL, on output the offset at which parsing ended.
* @param status A pointer to an UErrorCode to receive any errors
* @return The value of the parsed integer
* @see unum_parseInt64
* @see unum_parseDouble
* @see unum_format
* @see unum_formatInt64
* @see unum_formatDouble
* @stable ICU 2.0
*/
U_STABLE int32_t U_EXPORT2
unum_parse(    const   UNumberFormat*  fmt,
        const   UChar*          text,
        int32_t         textLength,
        int32_t         *parsePos /* 0 = start */,
        UErrorCode      *status);

/**
* Parse a string into an int64 using a UNumberFormat.
* The string will be parsed according to the UNumberFormat's locale.
* Note: parsing is not supported for styles UNUM_DECIMAL_COMPACT_SHORT
* and UNUM_DECIMAL_COMPACT_LONG.
* @param fmt The formatter to use.
* @param text The text to parse.
* @param textLength The length of text, or -1 if null-terminated.
* @param parsePos If not NULL, on input a pointer to an integer specifying the offset at which
* to begin parsing.  If not NULL, on output the offset at which parsing ended.
* @param status A pointer to an UErrorCode to receive any errors
* @return The value of the parsed integer
* @see unum_parse
* @see unum_parseDouble
* @see unum_format
* @see unum_formatInt64
* @see unum_formatDouble
* @stable ICU 2.8
*/
U_STABLE int64_t U_EXPORT2
unum_parseInt64(const UNumberFormat*  fmt,
        const UChar*  text,
        int32_t       textLength,
        int32_t       *parsePos /* 0 = start */,
        UErrorCode    *status);

/**
* Parse a string into a double using a UNumberFormat.
* The string will be parsed according to the UNumberFormat's locale.
* Note: parsing is not supported for styles UNUM_DECIMAL_COMPACT_SHORT
* and UNUM_DECIMAL_COMPACT_LONG.
* @param fmt The formatter to use.
* @param text The text to parse.
* @param textLength The length of text, or -1 if null-terminated.
* @param parsePos If not NULL, on input a pointer to an integer specifying the offset at which
* to begin parsing.  If not NULL, on output the offset at which parsing ended.
* @param status A pointer to an UErrorCode to receive any errors
* @return The value of the parsed double
* @see unum_parse
* @see unum_parseInt64
* @see unum_format
* @see unum_formatInt64
* @see unum_formatDouble
* @stable ICU 2.0
*/
U_STABLE double U_EXPORT2
unum_parseDouble(    const   UNumberFormat*  fmt,
            const   UChar*          text,
            int32_t         textLength,
            int32_t         *parsePos /* 0 = start */,
            UErrorCode      *status);


/**
* Parse a number from a string into an unformatted numeric string using a UNumberFormat.
* The input string will be parsed according to the UNumberFormat's locale.
* The syntax of the output is a "numeric string"
* as defined in the Decimal Arithmetic Specification, available at
* http://speleotrove.com/decimal
* Note: parsing is not supported for styles UNUM_DECIMAL_COMPACT_SHORT
* and UNUM_DECIMAL_COMPACT_LONG.
* @param fmt The formatter to use.
* @param text The text to parse.
* @param textLength The length of text, or -1 if null-terminated.
* @param parsePos If not NULL, on input a pointer to an integer specifying the offset at which
*                 to begin parsing.  If not NULL, on output the offset at which parsing ended.
* @param outBuf A (char *) buffer to receive the parsed number as a string.  The output string
*               will be nul-terminated if there is sufficient space.
* @param outBufLength The size of the output buffer.  May be zero, in which case
*               the outBuf pointer may be NULL, and the function will return the
*               size of the output string.
* @param status A pointer to an UErrorCode to receive any errors
* @return the length of the output string, not including any terminating nul.
* @see unum_parse
* @see unum_parseInt64
* @see unum_format
* @see unum_formatInt64
* @see unum_formatDouble
* @stable ICU 4.4
*/
U_STABLE int32_t U_EXPORT2
unum_parseDecimal(const   UNumberFormat*  fmt,
                 const   UChar*          text,
                         int32_t         textLength,
                         int32_t         *parsePos /* 0 = start */,
                         char            *outBuf,
                         int32_t         outBufLength,
                         UErrorCode      *status);

/**
 * Parse a string into a double and a currency using a UNumberFormat.
 * The string will be parsed according to the UNumberFormat's locale.
 * @param fmt the formatter to use
 * @param text the text to parse
 * @param textLength the length of text, or -1 if null-terminated
 * @param parsePos a pointer to an offset index into text at which to
 * begin parsing. On output, *parsePos will point after the last
 * parsed character.  This parameter may be NULL, in which case parsing
 * begins at offset 0.
 * @param currency a pointer to the buffer to receive the parsed null-
 * terminated currency.  This buffer must have a capacity of at least
 * 4 UChars.
 * @param status a pointer to an input-output UErrorCode
 * @return the parsed double
 * @see unum_parseDouble
 * @see unum_formatDoubleCurrency
 * @stable ICU 3.0
 */
U_STABLE double U_EXPORT2
unum_parseDoubleCurrency(const UNumberFormat* fmt,
                         const UChar* text,
                         int32_t textLength,
                         int32_t* parsePos, /* 0 = start */
                         UChar* currency,
                         UErrorCode* status);

/**
 * Parse a UChar string into a UFormattable.
 * Example code:
 * \snippet test/cintltst/cnumtst.c unum_parseToUFormattable
 * Note: parsing is not supported for styles UNUM_DECIMAL_COMPACT_SHORT
 * and UNUM_DECIMAL_COMPACT_LONG.
 * @param fmt the formatter to use
 * @param result the UFormattable to hold the result. If NULL, a new UFormattable will be allocated (which the caller must close with ufmt_close).
 * @param text the text to parse
 * @param textLength the length of text, or -1 if null-terminated
 * @param parsePos a pointer to an offset index into text at which to
 * begin parsing. On output, *parsePos will point after the last
 * parsed character.  This parameter may be NULL in which case parsing
 * begins at offset 0.
 * @param status a pointer to an input-output UErrorCode
 * @return the UFormattable.  Will be ==result unless NULL was passed in for result, in which case it will be the newly opened UFormattable.
 * @see ufmt_getType
 * @see ufmt_close
 * @stable ICU 52
 */
U_STABLE UFormattable* U_EXPORT2
unum_parseToUFormattable(const UNumberFormat* fmt,
                         UFormattable *result,
                         const UChar* text,
                         int32_t textLength,
                         int32_t* parsePos, /* 0 = start */
                         UErrorCode* status);

/**
 * Set the pattern used by a UNumberFormat.  This can only be used
 * on a DecimalFormat, other formats return U_UNSUPPORTED_ERROR
 * in the status.
 * @param format The formatter to set.
 * @param localized TRUE if the pattern is localized, FALSE otherwise.
 * @param pattern The new pattern
 * @param patternLength The length of pattern, or -1 if null-terminated.
 * @param parseError A pointer to UParseError to receive information
 * about errors occurred during parsing, or NULL if no parse error
 * information is desired.
 * @param status A pointer to an input-output UErrorCode.
 * @see unum_toPattern
 * @see DecimalFormat
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2
unum_applyPattern(          UNumberFormat  *format,
                            UBool          localized,
                    const   UChar          *pattern,
                            int32_t         patternLength,
                            UParseError    *parseError,
                            UErrorCode     *status
                                    );

/**
* Get a locale for which decimal formatting patterns are available.
* A UNumberFormat in a locale returned by this function will perform the correct
* formatting and parsing for the locale.  The results of this call are not
* valid for rule-based number formats.
* @param localeIndex The index of the desired locale.
* @return A locale for which number formatting patterns are available, or 0 if none.
* @see unum_countAvailable
* @stable ICU 2.0
*/
U_STABLE const char* U_EXPORT2
unum_getAvailable(int32_t localeIndex);

/**
* Determine how many locales have decimal formatting patterns available.  The
* results of this call are not valid for rule-based number formats.
* This function is useful for determining the loop ending condition for
* calls to {@link #unum_getAvailable }.
* @return The number of locales for which decimal formatting patterns are available.
* @see unum_getAvailable
* @stable ICU 2.0
*/
U_STABLE int32_t U_EXPORT2
unum_countAvailable(void);

#if UCONFIG_HAVE_PARSEALLINPUT
/* The UNumberFormatAttributeValue type cannot be #ifndef U_HIDE_INTERNAL_API, needed for .h variable declaration */
/**
 * @internal
 */
typedef enum UNumberFormatAttributeValue {
#ifndef U_HIDE_INTERNAL_API
  /** @internal */
  UNUM_NO = 0,
  /** @internal */
  UNUM_YES = 1,
  /** @internal */
  UNUM_MAYBE = 2
#else
  /** @internal */
  UNUM_FORMAT_ATTRIBUTE_VALUE_HIDDEN
#endif /* U_HIDE_INTERNAL_API */
} UNumberFormatAttributeValue;
#endif

/** The possible UNumberFormat numeric attributes @stable ICU 2.0 */
typedef enum UNumberFormatAttribute {
  /** Parse integers only */
  UNUM_PARSE_INT_ONLY,
  /** Use grouping separator */
  UNUM_GROUPING_USED,
  /** Always show decimal point */
  UNUM_DECIMAL_ALWAYS_SHOWN,
  /** Maximum integer digits */
  UNUM_MAX_INTEGER_DIGITS,
  /** Minimum integer digits */
  UNUM_MIN_INTEGER_DIGITS,
  /** Integer digits */
  UNUM_INTEGER_DIGITS,
  /** Maximum fraction digits */
  UNUM_MAX_FRACTION_DIGITS,
  /** Minimum fraction digits */
  UNUM_MIN_FRACTION_DIGITS,
  /** Fraction digits */
  UNUM_FRACTION_DIGITS,
  /** Multiplier */
  UNUM_MULTIPLIER,
  /** Grouping size */
  UNUM_GROUPING_SIZE,
  /** Rounding Mode */
  UNUM_ROUNDING_MODE,
  /** Rounding increment */
  UNUM_ROUNDING_INCREMENT,
  /** The width to which the output of <code>format()</code> is padded. */
  UNUM_FORMAT_WIDTH,
  /** The position at which padding will take place. */
  UNUM_PADDING_POSITION,
  /** Secondary grouping size */
  UNUM_SECONDARY_GROUPING_SIZE,
  /** Use significant digits
   * @stable ICU 3.0 */
  UNUM_SIGNIFICANT_DIGITS_USED,
  /** Minimum significant digits
   * @stable ICU 3.0 */
  UNUM_MIN_SIGNIFICANT_DIGITS,
  /** Maximum significant digits
   * @stable ICU 3.0 */
  UNUM_MAX_SIGNIFICANT_DIGITS,
  /** Lenient parse mode used by rule-based formats.
   * @stable ICU 3.0
   */
  UNUM_LENIENT_PARSE,
#if UCONFIG_HAVE_PARSEALLINPUT
  /** Consume all input. (may use fastpath). Set to UNUM_YES (require fastpath), UNUM_NO (skip fastpath), or UNUM_MAYBE (heuristic).
   * This is an internal ICU API. Do not use.
   * @internal
   */
  UNUM_PARSE_ALL_INPUT = 20,
#endif
  /**
    * Scale, which adjusts the position of the
    * decimal point when formatting.  Amounts will be multiplied by 10 ^ (scale)
    * before they are formatted.  The default value for the scale is 0 ( no adjustment ).
    *
    * <p>Example: setting the scale to 3, 123 formats as "123,000"
    * <p>Example: setting the scale to -4, 123 formats as "0.0123"
    *
    * This setting is analogous to getMultiplierScale() and setMultiplierScale() in decimfmt.h.
    *
   * @stable ICU 51 */
  UNUM_SCALE = 21,

#ifndef U_HIDE_DRAFT_API
  /**
   * Minimum grouping digits; most commonly set to 2 to print "1000" instead of "1,000".
   * See DecimalFormat::getMinimumGroupingDigits().
   *
   * For better control over grouping strategies, use UNumberFormatter.
   *
   * @draft ICU 64
   */
  UNUM_MINIMUM_GROUPING_DIGITS = 22,
#endif /* U_HIDE_DRAFT_API */

  /**
   * if this attribute is set to 0, it is set to UNUM_CURRENCY_STANDARD purpose,
   * otherwise it is UNUM_CURRENCY_CASH purpose
   * Default: 0 (UNUM_CURRENCY_STANDARD purpose)
   * @stable ICU 54
   */
  UNUM_CURRENCY_USAGE = 23,

#ifndef U_HIDE_INTERNAL_API
  /** One below the first bitfield-boolean item.
   * All items after this one are stored in boolean form.
   * @internal */
  UNUM_MAX_NONBOOLEAN_ATTRIBUTE = 0x0FFF,
#endif /* U_HIDE_INTERNAL_API */

  /** If 1, specifies that if setting the "max integer digits" attribute would truncate a value, set an error status rather than silently truncating.
   * For example,  formatting the value 1234 with 4 max int digits would succeed, but formatting 12345 would fail. There is no effect on parsing.
   * Default: 0 (not set)
   * @stable ICU 50
   */
  UNUM_FORMAT_FAIL_IF_MORE_THAN_MAX_DIGITS = 0x1000,
  /**
   * if this attribute is set to 1, specifies that, if the pattern doesn't contain an exponent, the exponent will not be parsed. If the pattern does contain an exponent, this attribute has no effect.
   * Has no effect on formatting.
   * Default: 0 (unset)
   * @stable ICU 50
   */
  UNUM_PARSE_NO_EXPONENT = 0x1001,

  /**
   * if this attribute is set to 1, specifies that, if the pattern contains a
   * decimal mark the input is required to have one. If this attribute is set to 0,
   * specifies that input does not have to contain a decimal mark.
   * Has no effect on formatting.
   * Default: 0 (unset)
   * @stable ICU 54
   */
  UNUM_PARSE_DECIMAL_MARK_REQUIRED = 0x1002,

#ifndef U_HIDE_DRAFT_API

  /**
   * Parsing: if set to 1, parsing is sensitive to case (lowercase/uppercase).
   *
   * @draft ICU 64
   */
  UNUM_PARSE_CASE_SENSITIVE = 0x1003,

  /**
   * Formatting: if set to 1, whether to show the plus sign on non-negative numbers.
   *
   * For better control over sign display, use UNumberFormatter.
   *
   * @draft ICU 64
   */
  UNUM_SIGN_ALWAYS_SHOWN = 0x1004,

#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_INTERNAL_API
  /** Limit of boolean attributes. (value should
   * not depend on U_HIDE conditionals)
   * @internal */
  UNUM_LIMIT_BOOLEAN_ATTRIBUTE = 0x1005,
#endif /* U_HIDE_INTERNAL_API */

} UNumberFormatAttribute;

/**
* Get a numeric attribute associated with a UNumberFormat.
* An example of a numeric attribute is the number of integer digits a formatter will produce.
* @param fmt The formatter to query.
* @param attr The attribute to query; one of UNUM_PARSE_INT_ONLY, UNUM_GROUPING_USED,
* UNUM_DECIMAL_ALWAYS_SHOWN, UNUM_MAX_INTEGER_DIGITS, UNUM_MIN_INTEGER_DIGITS, UNUM_INTEGER_DIGITS,
* UNUM_MAX_FRACTION_DIGITS, UNUM_MIN_FRACTION_DIGITS, UNUM_FRACTION_DIGITS, UNUM_MULTIPLIER,
* UNUM_GROUPING_SIZE, UNUM_ROUNDING_MODE, UNUM_FORMAT_WIDTH, UNUM_PADDING_POSITION, UNUM_SECONDARY_GROUPING_SIZE,
* UNUM_SCALE, UNUM_MINIMUM_GROUPING_DIGITS.
* @return The value of attr.
* @see unum_setAttribute
* @see unum_getDoubleAttribute
* @see unum_setDoubleAttribute
* @see unum_getTextAttribute
* @see unum_setTextAttribute
* @stable ICU 2.0
*/
U_STABLE int32_t U_EXPORT2
unum_getAttribute(const UNumberFormat*          fmt,
          UNumberFormatAttribute  attr);

/**
* Set a numeric attribute associated with a UNumberFormat.
* An example of a numeric attribute is the number of integer digits a formatter will produce.  If the
* formatter does not understand the attribute, the call is ignored.  Rule-based formatters only understand
* the lenient-parse attribute.
* @param fmt The formatter to set.
* @param attr The attribute to set; one of UNUM_PARSE_INT_ONLY, UNUM_GROUPING_USED,
* UNUM_DECIMAL_ALWAYS_SHOWN, UNUM_MAX_INTEGER_DIGITS, UNUM_MIN_INTEGER_DIGITS, UNUM_INTEGER_DIGITS,
* UNUM_MAX_FRACTION_DIGITS, UNUM_MIN_FRACTION_DIGITS, UNUM_FRACTION_DIGITS, UNUM_MULTIPLIER,
* UNUM_GROUPING_SIZE, UNUM_ROUNDING_MODE, UNUM_FORMAT_WIDTH, UNUM_PADDING_POSITION, UNUM_SECONDARY_GROUPING_SIZE,
* UNUM_LENIENT_PARSE, UNUM_SCALE, UNUM_MINIMUM_GROUPING_DIGITS.
* @param newValue The new value of attr.
* @see unum_getAttribute
* @see unum_getDoubleAttribute
* @see unum_setDoubleAttribute
* @see unum_getTextAttribute
* @see unum_setTextAttribute
* @stable ICU 2.0
*/
U_STABLE void U_EXPORT2
unum_setAttribute(    UNumberFormat*          fmt,
            UNumberFormatAttribute  attr,
            int32_t                 newValue);


/**
* Get a numeric attribute associated with a UNumberFormat.
* An example of a numeric attribute is the number of integer digits a formatter will produce.
* If the formatter does not understand the attribute, -1 is returned.
* @param fmt The formatter to query.
* @param attr The attribute to query; e.g. UNUM_ROUNDING_INCREMENT.
* @return The value of attr.
* @see unum_getAttribute
* @see unum_setAttribute
* @see unum_setDoubleAttribute
* @see unum_getTextAttribute
* @see unum_setTextAttribute
* @stable ICU 2.0
*/
U_STABLE double U_EXPORT2
unum_getDoubleAttribute(const UNumberFormat*          fmt,
          UNumberFormatAttribute  attr);

/**
* Set a numeric attribute associated with a UNumberFormat.
* An example of a numeric attribute is the number of integer digits a formatter will produce.
* If the formatter does not understand the attribute, this call is ignored.
* @param fmt The formatter to set.
* @param attr The attribute to set; e.g. UNUM_ROUNDING_INCREMENT.
* @param newValue The new value of attr.
* @see unum_getAttribute
* @see unum_setAttribute
* @see unum_getDoubleAttribute
* @see unum_getTextAttribute
* @see unum_setTextAttribute
* @stable ICU 2.0
*/
U_STABLE void U_EXPORT2
unum_setDoubleAttribute(    UNumberFormat*          fmt,
            UNumberFormatAttribute  attr,
            double                 newValue);

/** The possible UNumberFormat text attributes @stable ICU 2.0*/
typedef enum UNumberFormatTextAttribute {
  /** Positive prefix */
  UNUM_POSITIVE_PREFIX,
  /** Positive suffix */
  UNUM_POSITIVE_SUFFIX,
  /** Negative prefix */
  UNUM_NEGATIVE_PREFIX,
  /** Negative suffix */
  UNUM_NEGATIVE_SUFFIX,
  /** The character used to pad to the format width. */
  UNUM_PADDING_CHARACTER,
  /** The ISO currency code */
  UNUM_CURRENCY_CODE,
  /**
   * The default rule set, such as "%spellout-numbering-year:", "%spellout-cardinal:",
   * "%spellout-ordinal-masculine-plural:", "%spellout-ordinal-feminine:", or
   * "%spellout-ordinal-neuter:". The available public rulesets can be listed using
   * unum_getTextAttribute with UNUM_PUBLIC_RULESETS. This is only available with
   * rule-based formatters.
   * @stable ICU 3.0
   */
  UNUM_DEFAULT_RULESET,
  /**
   * The public rule sets.  This is only available with rule-based formatters.
   * This is a read-only attribute.  The public rulesets are returned as a
   * single string, with each ruleset name delimited by ';' (semicolon). See the
   * CLDR LDML spec for more information about RBNF rulesets:
   * http://www.unicode.org/reports/tr35/tr35-numbers.html#Rule-Based_Number_Formatting
   * @stable ICU 3.0
   */
  UNUM_PUBLIC_RULESETS
} UNumberFormatTextAttribute;

/**
* Get a text attribute associated with a UNumberFormat.
* An example of a text attribute is the suffix for positive numbers.  If the formatter
* does not understand the attribute, U_UNSUPPORTED_ERROR is returned as the status.
* Rule-based formatters only understand UNUM_DEFAULT_RULESET and UNUM_PUBLIC_RULESETS.
* @param fmt The formatter to query.
* @param tag The attribute to query; one of UNUM_POSITIVE_PREFIX, UNUM_POSITIVE_SUFFIX,
* UNUM_NEGATIVE_PREFIX, UNUM_NEGATIVE_SUFFIX, UNUM_PADDING_CHARACTER, UNUM_CURRENCY_CODE,
* UNUM_DEFAULT_RULESET, or UNUM_PUBLIC_RULESETS.
* @param result A pointer to a buffer to receive the attribute.
* @param resultLength The maximum size of result.
* @param status A pointer to an UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see unum_setTextAttribute
* @see unum_getAttribute
* @see unum_setAttribute
* @stable ICU 2.0
*/
U_STABLE int32_t U_EXPORT2
unum_getTextAttribute(    const    UNumberFormat*                    fmt,
            UNumberFormatTextAttribute      tag,
            UChar*                            result,
            int32_t                            resultLength,
            UErrorCode*                        status);

/**
* Set a text attribute associated with a UNumberFormat.
* An example of a text attribute is the suffix for positive numbers.  Rule-based formatters
* only understand UNUM_DEFAULT_RULESET.
* @param fmt The formatter to set.
* @param tag The attribute to set; one of UNUM_POSITIVE_PREFIX, UNUM_POSITIVE_SUFFIX,
* UNUM_NEGATIVE_PREFIX, UNUM_NEGATIVE_SUFFIX, UNUM_PADDING_CHARACTER, UNUM_CURRENCY_CODE,
* or UNUM_DEFAULT_RULESET.
* @param newValue The new value of attr.
* @param newValueLength The length of newValue, or -1 if null-terminated.
* @param status A pointer to an UErrorCode to receive any errors
* @see unum_getTextAttribute
* @see unum_getAttribute
* @see unum_setAttribute
* @stable ICU 2.0
*/
U_STABLE void U_EXPORT2
unum_setTextAttribute(    UNumberFormat*                    fmt,
            UNumberFormatTextAttribute      tag,
            const    UChar*                            newValue,
            int32_t                            newValueLength,
            UErrorCode                        *status);

/**
 * Extract the pattern from a UNumberFormat.  The pattern will follow
 * the DecimalFormat pattern syntax.
 * @param fmt The formatter to query.
 * @param isPatternLocalized TRUE if the pattern should be localized,
 * FALSE otherwise.  This is ignored if the formatter is a rule-based
 * formatter.
 * @param result A pointer to a buffer to receive the pattern.
 * @param resultLength The maximum size of result.
 * @param status A pointer to an input-output UErrorCode.
 * @return The total buffer size needed; if greater than resultLength,
 * the output was truncated.
 * @see unum_applyPattern
 * @see DecimalFormat
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2
unum_toPattern(    const    UNumberFormat*          fmt,
        UBool                  isPatternLocalized,
        UChar*                  result,
        int32_t                 resultLength,
        UErrorCode*             status);


/**
 * Constants for specifying a number format symbol.
 * @stable ICU 2.0
 */
typedef enum UNumberFormatSymbol {
  /** The decimal separator */
  UNUM_DECIMAL_SEPARATOR_SYMBOL = 0,
  /** The grouping separator */
  UNUM_GROUPING_SEPARATOR_SYMBOL = 1,
  /** The pattern separator */
  UNUM_PATTERN_SEPARATOR_SYMBOL = 2,
  /** The percent sign */
  UNUM_PERCENT_SYMBOL = 3,
  /** Zero*/
  UNUM_ZERO_DIGIT_SYMBOL = 4,
  /** Character representing a digit in the pattern */
  UNUM_DIGIT_SYMBOL = 5,
  /** The minus sign */
  UNUM_MINUS_SIGN_SYMBOL = 6,
  /** The plus sign */
  UNUM_PLUS_SIGN_SYMBOL = 7,
  /** The currency symbol */
  UNUM_CURRENCY_SYMBOL = 8,
  /** The international currency symbol */
  UNUM_INTL_CURRENCY_SYMBOL = 9,
  /** The monetary separator */
  UNUM_MONETARY_SEPARATOR_SYMBOL = 10,
  /** The exponential symbol */
  UNUM_EXPONENTIAL_SYMBOL = 11,
  /** Per mill symbol */
  UNUM_PERMILL_SYMBOL = 12,
  /** Escape padding character */
  UNUM_PAD_ESCAPE_SYMBOL = 13,
  /** Infinity symbol */
  UNUM_INFINITY_SYMBOL = 14,
  /** Nan symbol */
  UNUM_NAN_SYMBOL = 15,
  /** Significant digit symbol
   * @stable ICU 3.0 */
  UNUM_SIGNIFICANT_DIGIT_SYMBOL = 16,
  /** The monetary grouping separator
   * @stable ICU 3.6
   */
  UNUM_MONETARY_GROUPING_SEPARATOR_SYMBOL = 17,
  /** One
   * @stable ICU 4.6
   */
  UNUM_ONE_DIGIT_SYMBOL = 18,
  /** Two
   * @stable ICU 4.6
   */
  UNUM_TWO_DIGIT_SYMBOL = 19,
  /** Three
   * @stable ICU 4.6
   */
  UNUM_THREE_DIGIT_SYMBOL = 20,
  /** Four
   * @stable ICU 4.6
   */
  UNUM_FOUR_DIGIT_SYMBOL = 21,
  /** Five
   * @stable ICU 4.6
   */
  UNUM_FIVE_DIGIT_SYMBOL = 22,
  /** Six
   * @stable ICU 4.6
   */
  UNUM_SIX_DIGIT_SYMBOL = 23,
  /** Seven
    * @stable ICU 4.6
   */
  UNUM_SEVEN_DIGIT_SYMBOL = 24,
  /** Eight
   * @stable ICU 4.6
   */
  UNUM_EIGHT_DIGIT_SYMBOL = 25,
  /** Nine
   * @stable ICU 4.6
   */
  UNUM_NINE_DIGIT_SYMBOL = 26,

  /** Multiplication sign
   * @stable ICU 54
   */
  UNUM_EXPONENT_MULTIPLICATION_SYMBOL = 27,

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UNumberFormatSymbol value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
  UNUM_FORMAT_SYMBOL_COUNT = 28
#endif  /* U_HIDE_DEPRECATED_API */
} UNumberFormatSymbol;

/**
* Get a symbol associated with a UNumberFormat.
* A UNumberFormat uses symbols to represent the special locale-dependent
* characters in a number, for example the percent sign. This API is not
* supported for rule-based formatters.
* @param fmt The formatter to query.
* @param symbol The UNumberFormatSymbol constant for the symbol to get
* @param buffer The string buffer that will receive the symbol string;
*               if it is NULL, then only the length of the symbol is returned
* @param size The size of the string buffer
* @param status A pointer to an UErrorCode to receive any errors
* @return The length of the symbol; the buffer is not modified if
*         <code>length&gt;=size</code>
* @see unum_setSymbol
* @stable ICU 2.0
*/
U_STABLE int32_t U_EXPORT2
unum_getSymbol(const UNumberFormat *fmt,
               UNumberFormatSymbol symbol,
               UChar *buffer,
               int32_t size,
               UErrorCode *status);

/**
* Set a symbol associated with a UNumberFormat.
* A UNumberFormat uses symbols to represent the special locale-dependent
* characters in a number, for example the percent sign.  This API is not
* supported for rule-based formatters.
* @param fmt The formatter to set.
* @param symbol The UNumberFormatSymbol constant for the symbol to set
* @param value The string to set the symbol to
* @param length The length of the string, or -1 for a zero-terminated string
* @param status A pointer to an UErrorCode to receive any errors.
* @see unum_getSymbol
* @stable ICU 2.0
*/
U_STABLE void U_EXPORT2
unum_setSymbol(UNumberFormat *fmt,
               UNumberFormatSymbol symbol,
               const UChar *value,
               int32_t length,
               UErrorCode *status);


/**
 * Get the locale for this number format object.
 * You can choose between valid and actual locale.
 * @param fmt The formatter to get the locale from
 * @param type type of the locale we're looking for (valid or actual)
 * @param status error code for the operation
 * @return the locale name
 * @stable ICU 2.8
 */
U_STABLE const char* U_EXPORT2
unum_getLocaleByType(const UNumberFormat *fmt,
                     ULocDataLocaleType type,
                     UErrorCode* status);

/**
 * Set a particular UDisplayContext value in the formatter, such as
 * UDISPCTX_CAPITALIZATION_FOR_STANDALONE.
 * @param fmt The formatter for which to set a UDisplayContext value.
 * @param value The UDisplayContext value to set.
 * @param status A pointer to an UErrorCode to receive any errors
 * @stable ICU 53
 */
U_STABLE void U_EXPORT2
unum_setContext(UNumberFormat* fmt, UDisplayContext value, UErrorCode* status);

/**
 * Get the formatter's UDisplayContext value for the specified UDisplayContextType,
 * such as UDISPCTX_TYPE_CAPITALIZATION.
 * @param fmt The formatter to query.
 * @param type The UDisplayContextType whose value to return
 * @param status A pointer to an UErrorCode to receive any errors
 * @return The UDisplayContextValue for the specified type.
 * @stable ICU 53
 */
U_STABLE UDisplayContext U_EXPORT2
unum_getContext(const UNumberFormat *fmt, UDisplayContextType type, UErrorCode* status);

#endif /* #if !UCONFIG_NO_FORMATTING */

 // UNumberFormatStyle


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2006, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 26, 2004
* Since: ICU 3.0
**********************************************************************
*/



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 26, 2004
* Since: ICU 3.0
**********************************************************************
*/



#if U_SHOW_CPLUSPLUS_API

/**
 * \file 
 * \brief C++ API: MeasureUnit object.
 */
 
#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

class MeasureUnit;

/**
 * An amount of a specified unit, consisting of a number and a Unit.
 * For example, a length measure consists of a number and a length
 * unit, such as feet or meters.
 *
 * <p>Measure objects are formatted by MeasureFormat.
 *
 * <p>Measure objects are immutable.
 *
 * @author Alan Liu
 * @stable ICU 3.0
 */
class U_I18N_API Measure: public UObject {
 public:
    /**
     * Construct an object with the given numeric amount and the given
     * unit.  After this call, the caller must not delete the given
     * unit object.
     * @param number a numeric object; amount.isNumeric() must be TRUE
     * @param adoptedUnit the unit object, which must not be NULL
     * @param ec input-output error code. If the amount or the unit
     * is invalid, then this will be set to a failing value.
     * @stable ICU 3.0
     */
    Measure(const Formattable& number, MeasureUnit* adoptedUnit,
            UErrorCode& ec);

    /**
     * Copy constructor
     * @stable ICU 3.0
     */
    Measure(const Measure& other);

    /**
     * Assignment operator
     * @stable ICU 3.0
     */
    Measure& operator=(const Measure& other);

    /**
     * Return a polymorphic clone of this object.  The result will
     * have the same class as returned by getDynamicClassID().
     * @stable ICU 3.0
     */
    virtual Measure* clone() const;

    /**
     * Destructor
     * @stable ICU 3.0
     */
    virtual ~Measure();
    
    /**
     * Equality operator.  Return true if this object is equal
     * to the given object.
     * @stable ICU 3.0
     */
    UBool operator==(const UObject& other) const;

    /**
     * Return a reference to the numeric value of this object.  The
     * numeric value may be of any numeric type supported by
     * Formattable.
     * @stable ICU 3.0
     */
    inline const Formattable& getNumber() const;

    /**
     * Return a reference to the unit of this object.
     * @stable ICU 3.0
     */
    inline const MeasureUnit& getUnit() const;

    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 53
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 53
     */
    virtual UClassID getDynamicClassID(void) const;

 protected:
    /**
     * Default constructor.
     * @stable ICU 3.0
     */
    Measure();

 private:
    /**
     * The numeric value of this object, e.g. 2.54 or 100.
     */
    Formattable number;

    /**
     * The unit of this object, e.g., "millimeter" or "JPY".  This is
     * owned by this object.
     */
    MeasureUnit* unit;
};

inline const Formattable& Measure::getNumber() const {
    return number;
}

inline const MeasureUnit& Measure::getUnit() const {
    return *unit;
}

U_NAMESPACE_END

#endif // !UCONFIG_NO_FORMATTING

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 26, 2004
* Since: ICU 3.0
**********************************************************************
*/



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 26, 2004
* Since: ICU 3.0
**********************************************************************
*/



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING



/**
 * \file 
 * \brief C++ API: A unit for measuring a quantity.
 */
 
U_NAMESPACE_BEGIN

class StringEnumeration;

/**
 * A unit such as length, mass, volume, currency, etc.  A unit is
 * coupled with a numeric amount to produce a Measure.
 *
 * @author Alan Liu
 * @stable ICU 3.0
 */
class U_I18N_API MeasureUnit: public UObject {
 public:

    /**
     * Default constructor.
     * Populates the instance with the base dimensionless unit.
     * @stable ICU 3.0
     */
    MeasureUnit();
    
    /**
     * Copy constructor.
     * @stable ICU 3.0
     */
    MeasureUnit(const MeasureUnit &other);
        
    /**
     * Assignment operator.
     * @stable ICU 3.0
     */
    MeasureUnit &operator=(const MeasureUnit &other);

    /**
     * Returns a polymorphic clone of this object.  The result will
     * have the same class as returned by getDynamicClassID().
     * @stable ICU 3.0
     */
    virtual MeasureUnit* clone() const;

    /**
     * Destructor
     * @stable ICU 3.0
     */
    virtual ~MeasureUnit();

    /**
     * Equality operator.  Return true if this object is equal
     * to the given object.
     * @stable ICU 3.0
     */
    virtual UBool operator==(const UObject& other) const;

    /**
     * Inequality operator.  Return true if this object is not equal
     * to the given object.
     * @stable ICU 53
     */
    UBool operator!=(const UObject& other) const {
        return !(*this == other);
    }

    /**
     * Get the type.
     * @stable ICU 53
     */
    const char *getType() const;

    /**
     * Get the sub type.
     * @stable ICU 53
     */
    const char *getSubtype() const;

    /**
     * getAvailable gets all of the available units.
     * If there are too many units to fit into destCapacity then the
     * error code is set to U_BUFFER_OVERFLOW_ERROR.
     *
     * @param destArray destination buffer.
     * @param destCapacity number of MeasureUnit instances available at dest.
     * @param errorCode ICU error code.
     * @return number of available units.
     * @stable ICU 53
     */
    static int32_t getAvailable(
            MeasureUnit *destArray,
            int32_t destCapacity,
            UErrorCode &errorCode);

    /**
     * getAvailable gets all of the available units for a specific type.
     * If there are too many units to fit into destCapacity then the
     * error code is set to U_BUFFER_OVERFLOW_ERROR.
     *
     * @param type the type
     * @param destArray destination buffer.
     * @param destCapacity number of MeasureUnit instances available at dest.
     * @param errorCode ICU error code.
     * @return number of available units for type.
     * @stable ICU 53
     */
    static int32_t getAvailable(
            const char *type,
            MeasureUnit *destArray,
            int32_t destCapacity,
            UErrorCode &errorCode);

    /**
     * getAvailableTypes gets all of the available types. Caller owns the
     * returned StringEnumeration and must delete it when finished using it.
     *
     * @param errorCode ICU error code.
     * @return the types.
     * @stable ICU 53
     */
    static StringEnumeration* getAvailableTypes(UErrorCode &errorCode);

    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       Derived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 53
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 53
     */
    virtual UClassID getDynamicClassID(void) const;

#ifndef U_HIDE_INTERNAL_API
    /**
     * ICU use only.
     * Returns associated array index for this measure unit. Only valid for
     * non-currency measure units.
     * @internal
     */
    int32_t getIndex() const;

    /**
     * ICU use only.
     * Returns maximum value from getIndex plus 1.
     * @internal
     */
    static int32_t getIndexCount();

    /**
     * ICU use only.
     * @return the unit.getIndex() of the unit which has this unit.getType() and unit.getSubtype(),
     *         or a negative value if there is no such unit
     * @internal
     */
    static int32_t internalGetIndexForTypeAndSubtype(const char *type, const char *subtype);

    /**
     * ICU use only.
     * @internal
     */
    static MeasureUnit resolveUnitPerUnit(
            const MeasureUnit &unit, const MeasureUnit &perUnit, bool* isResolved);
#endif /* U_HIDE_INTERNAL_API */

// All code between the "Start generated createXXX methods" comment and
// the "End generated createXXX methods" comment is auto generated code
// and must not be edited manually. For instructions on how to correctly
// update this code, refer to:
// http://site.icu-project.org/design/formatting/measureformat/updating-measure-unit
//
// Start generated createXXX methods

    /**
     * Returns by pointer, unit of acceleration: g-force.
     * Caller owns returned value and must free it.
     * Also see {@link #getGForce()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createGForce(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of acceleration: g-force.
     * Also see {@link #createGForce()}.
     * @draft ICU 64
     */
    static MeasureUnit getGForce();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of acceleration: meter-per-second-squared.
     * Caller owns returned value and must free it.
     * Also see {@link #getMeterPerSecondSquared()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMeterPerSecondSquared(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of acceleration: meter-per-second-squared.
     * Also see {@link #createMeterPerSecondSquared()}.
     * @draft ICU 64
     */
    static MeasureUnit getMeterPerSecondSquared();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of angle: arc-minute.
     * Caller owns returned value and must free it.
     * Also see {@link #getArcMinute()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createArcMinute(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of angle: arc-minute.
     * Also see {@link #createArcMinute()}.
     * @draft ICU 64
     */
    static MeasureUnit getArcMinute();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of angle: arc-second.
     * Caller owns returned value and must free it.
     * Also see {@link #getArcSecond()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createArcSecond(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of angle: arc-second.
     * Also see {@link #createArcSecond()}.
     * @draft ICU 64
     */
    static MeasureUnit getArcSecond();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of angle: degree.
     * Caller owns returned value and must free it.
     * Also see {@link #getDegree()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createDegree(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of angle: degree.
     * Also see {@link #createDegree()}.
     * @draft ICU 64
     */
    static MeasureUnit getDegree();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of angle: radian.
     * Caller owns returned value and must free it.
     * Also see {@link #getRadian()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createRadian(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of angle: radian.
     * Also see {@link #createRadian()}.
     * @draft ICU 64
     */
    static MeasureUnit getRadian();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of angle: revolution.
     * Caller owns returned value and must free it.
     * Also see {@link #getRevolutionAngle()}.
     * @param status ICU error code.
     * @stable ICU 56
     */
    static MeasureUnit *createRevolutionAngle(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of angle: revolution.
     * Also see {@link #createRevolutionAngle()}.
     * @draft ICU 64
     */
    static MeasureUnit getRevolutionAngle();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of area: acre.
     * Caller owns returned value and must free it.
     * Also see {@link #getAcre()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createAcre(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of area: acre.
     * Also see {@link #createAcre()}.
     * @draft ICU 64
     */
    static MeasureUnit getAcre();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of area: dunam.
     * Caller owns returned value and must free it.
     * Also see {@link #getDunam()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createDunam(UErrorCode &status);

    /**
     * Returns by value, unit of area: dunam.
     * Also see {@link #createDunam()}.
     * @draft ICU 64
     */
    static MeasureUnit getDunam();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of area: hectare.
     * Caller owns returned value and must free it.
     * Also see {@link #getHectare()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createHectare(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of area: hectare.
     * Also see {@link #createHectare()}.
     * @draft ICU 64
     */
    static MeasureUnit getHectare();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of area: square-centimeter.
     * Caller owns returned value and must free it.
     * Also see {@link #getSquareCentimeter()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createSquareCentimeter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of area: square-centimeter.
     * Also see {@link #createSquareCentimeter()}.
     * @draft ICU 64
     */
    static MeasureUnit getSquareCentimeter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of area: square-foot.
     * Caller owns returned value and must free it.
     * Also see {@link #getSquareFoot()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createSquareFoot(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of area: square-foot.
     * Also see {@link #createSquareFoot()}.
     * @draft ICU 64
     */
    static MeasureUnit getSquareFoot();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of area: square-inch.
     * Caller owns returned value and must free it.
     * Also see {@link #getSquareInch()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createSquareInch(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of area: square-inch.
     * Also see {@link #createSquareInch()}.
     * @draft ICU 64
     */
    static MeasureUnit getSquareInch();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of area: square-kilometer.
     * Caller owns returned value and must free it.
     * Also see {@link #getSquareKilometer()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createSquareKilometer(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of area: square-kilometer.
     * Also see {@link #createSquareKilometer()}.
     * @draft ICU 64
     */
    static MeasureUnit getSquareKilometer();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of area: square-meter.
     * Caller owns returned value and must free it.
     * Also see {@link #getSquareMeter()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createSquareMeter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of area: square-meter.
     * Also see {@link #createSquareMeter()}.
     * @draft ICU 64
     */
    static MeasureUnit getSquareMeter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of area: square-mile.
     * Caller owns returned value and must free it.
     * Also see {@link #getSquareMile()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createSquareMile(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of area: square-mile.
     * Also see {@link #createSquareMile()}.
     * @draft ICU 64
     */
    static MeasureUnit getSquareMile();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of area: square-yard.
     * Caller owns returned value and must free it.
     * Also see {@link #getSquareYard()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createSquareYard(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of area: square-yard.
     * Also see {@link #createSquareYard()}.
     * @draft ICU 64
     */
    static MeasureUnit getSquareYard();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of concentr: karat.
     * Caller owns returned value and must free it.
     * Also see {@link #getKarat()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createKarat(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of concentr: karat.
     * Also see {@link #createKarat()}.
     * @draft ICU 64
     */
    static MeasureUnit getKarat();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of concentr: milligram-per-deciliter.
     * Caller owns returned value and must free it.
     * Also see {@link #getMilligramPerDeciliter()}.
     * @param status ICU error code.
     * @stable ICU 57
     */
    static MeasureUnit *createMilligramPerDeciliter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of concentr: milligram-per-deciliter.
     * Also see {@link #createMilligramPerDeciliter()}.
     * @draft ICU 64
     */
    static MeasureUnit getMilligramPerDeciliter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of concentr: millimole-per-liter.
     * Caller owns returned value and must free it.
     * Also see {@link #getMillimolePerLiter()}.
     * @param status ICU error code.
     * @stable ICU 57
     */
    static MeasureUnit *createMillimolePerLiter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of concentr: millimole-per-liter.
     * Also see {@link #createMillimolePerLiter()}.
     * @draft ICU 64
     */
    static MeasureUnit getMillimolePerLiter();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of concentr: mole.
     * Caller owns returned value and must free it.
     * Also see {@link #getMole()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createMole(UErrorCode &status);

    /**
     * Returns by value, unit of concentr: mole.
     * Also see {@link #createMole()}.
     * @draft ICU 64
     */
    static MeasureUnit getMole();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of concentr: part-per-million.
     * Caller owns returned value and must free it.
     * Also see {@link #getPartPerMillion()}.
     * @param status ICU error code.
     * @stable ICU 57
     */
    static MeasureUnit *createPartPerMillion(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of concentr: part-per-million.
     * Also see {@link #createPartPerMillion()}.
     * @draft ICU 64
     */
    static MeasureUnit getPartPerMillion();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of concentr: percent.
     * Caller owns returned value and must free it.
     * Also see {@link #getPercent()}.
     * @param status ICU error code.
     * @stable ICU 63
     */
    static MeasureUnit *createPercent(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of concentr: percent.
     * Also see {@link #createPercent()}.
     * @draft ICU 64
     */
    static MeasureUnit getPercent();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of concentr: permille.
     * Caller owns returned value and must free it.
     * Also see {@link #getPermille()}.
     * @param status ICU error code.
     * @stable ICU 63
     */
    static MeasureUnit *createPermille(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of concentr: permille.
     * Also see {@link #createPermille()}.
     * @draft ICU 64
     */
    static MeasureUnit getPermille();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of concentr: permyriad.
     * Caller owns returned value and must free it.
     * Also see {@link #getPermyriad()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createPermyriad(UErrorCode &status);

    /**
     * Returns by value, unit of concentr: permyriad.
     * Also see {@link #createPermyriad()}.
     * @draft ICU 64
     */
    static MeasureUnit getPermyriad();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of consumption: liter-per-100kilometers.
     * Caller owns returned value and must free it.
     * Also see {@link #getLiterPer100Kilometers()}.
     * @param status ICU error code.
     * @stable ICU 56
     */
    static MeasureUnit *createLiterPer100Kilometers(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of consumption: liter-per-100kilometers.
     * Also see {@link #createLiterPer100Kilometers()}.
     * @draft ICU 64
     */
    static MeasureUnit getLiterPer100Kilometers();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of consumption: liter-per-kilometer.
     * Caller owns returned value and must free it.
     * Also see {@link #getLiterPerKilometer()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createLiterPerKilometer(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of consumption: liter-per-kilometer.
     * Also see {@link #createLiterPerKilometer()}.
     * @draft ICU 64
     */
    static MeasureUnit getLiterPerKilometer();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of consumption: mile-per-gallon.
     * Caller owns returned value and must free it.
     * Also see {@link #getMilePerGallon()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMilePerGallon(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of consumption: mile-per-gallon.
     * Also see {@link #createMilePerGallon()}.
     * @draft ICU 64
     */
    static MeasureUnit getMilePerGallon();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of consumption: mile-per-gallon-imperial.
     * Caller owns returned value and must free it.
     * Also see {@link #getMilePerGallonImperial()}.
     * @param status ICU error code.
     * @stable ICU 57
     */
    static MeasureUnit *createMilePerGallonImperial(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of consumption: mile-per-gallon-imperial.
     * Also see {@link #createMilePerGallonImperial()}.
     * @draft ICU 64
     */
    static MeasureUnit getMilePerGallonImperial();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: bit.
     * Caller owns returned value and must free it.
     * Also see {@link #getBit()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createBit(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: bit.
     * Also see {@link #createBit()}.
     * @draft ICU 64
     */
    static MeasureUnit getBit();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: byte.
     * Caller owns returned value and must free it.
     * Also see {@link #getByte()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createByte(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: byte.
     * Also see {@link #createByte()}.
     * @draft ICU 64
     */
    static MeasureUnit getByte();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: gigabit.
     * Caller owns returned value and must free it.
     * Also see {@link #getGigabit()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createGigabit(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: gigabit.
     * Also see {@link #createGigabit()}.
     * @draft ICU 64
     */
    static MeasureUnit getGigabit();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: gigabyte.
     * Caller owns returned value and must free it.
     * Also see {@link #getGigabyte()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createGigabyte(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: gigabyte.
     * Also see {@link #createGigabyte()}.
     * @draft ICU 64
     */
    static MeasureUnit getGigabyte();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: kilobit.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilobit()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createKilobit(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: kilobit.
     * Also see {@link #createKilobit()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilobit();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: kilobyte.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilobyte()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createKilobyte(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: kilobyte.
     * Also see {@link #createKilobyte()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilobyte();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: megabit.
     * Caller owns returned value and must free it.
     * Also see {@link #getMegabit()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMegabit(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: megabit.
     * Also see {@link #createMegabit()}.
     * @draft ICU 64
     */
    static MeasureUnit getMegabit();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: megabyte.
     * Caller owns returned value and must free it.
     * Also see {@link #getMegabyte()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMegabyte(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: megabyte.
     * Also see {@link #createMegabyte()}.
     * @draft ICU 64
     */
    static MeasureUnit getMegabyte();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: petabyte.
     * Caller owns returned value and must free it.
     * Also see {@link #getPetabyte()}.
     * @param status ICU error code.
     * @stable ICU 63
     */
    static MeasureUnit *createPetabyte(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: petabyte.
     * Also see {@link #createPetabyte()}.
     * @draft ICU 64
     */
    static MeasureUnit getPetabyte();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: terabit.
     * Caller owns returned value and must free it.
     * Also see {@link #getTerabit()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createTerabit(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: terabit.
     * Also see {@link #createTerabit()}.
     * @draft ICU 64
     */
    static MeasureUnit getTerabit();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of digital: terabyte.
     * Caller owns returned value and must free it.
     * Also see {@link #getTerabyte()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createTerabyte(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of digital: terabyte.
     * Also see {@link #createTerabyte()}.
     * @draft ICU 64
     */
    static MeasureUnit getTerabyte();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: century.
     * Caller owns returned value and must free it.
     * Also see {@link #getCentury()}.
     * @param status ICU error code.
     * @stable ICU 56
     */
    static MeasureUnit *createCentury(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: century.
     * Also see {@link #createCentury()}.
     * @draft ICU 64
     */
    static MeasureUnit getCentury();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: day.
     * Caller owns returned value and must free it.
     * Also see {@link #getDay()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createDay(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: day.
     * Also see {@link #createDay()}.
     * @draft ICU 64
     */
    static MeasureUnit getDay();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of duration: day-person.
     * Caller owns returned value and must free it.
     * Also see {@link #getDayPerson()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createDayPerson(UErrorCode &status);

    /**
     * Returns by value, unit of duration: day-person.
     * Also see {@link #createDayPerson()}.
     * @draft ICU 64
     */
    static MeasureUnit getDayPerson();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of duration: decade.
     * Caller owns returned value and must free it.
     * Also see {@link #getDecade()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createDecade(UErrorCode &status);

    /**
     * Returns by value, unit of duration: decade.
     * Also see {@link #createDecade()}.
     * @draft ICU 65
     */
    static MeasureUnit getDecade();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: hour.
     * Caller owns returned value and must free it.
     * Also see {@link #getHour()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createHour(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: hour.
     * Also see {@link #createHour()}.
     * @draft ICU 64
     */
    static MeasureUnit getHour();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: microsecond.
     * Caller owns returned value and must free it.
     * Also see {@link #getMicrosecond()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMicrosecond(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: microsecond.
     * Also see {@link #createMicrosecond()}.
     * @draft ICU 64
     */
    static MeasureUnit getMicrosecond();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: millisecond.
     * Caller owns returned value and must free it.
     * Also see {@link #getMillisecond()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createMillisecond(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: millisecond.
     * Also see {@link #createMillisecond()}.
     * @draft ICU 64
     */
    static MeasureUnit getMillisecond();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: minute.
     * Caller owns returned value and must free it.
     * Also see {@link #getMinute()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createMinute(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: minute.
     * Also see {@link #createMinute()}.
     * @draft ICU 64
     */
    static MeasureUnit getMinute();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: month.
     * Caller owns returned value and must free it.
     * Also see {@link #getMonth()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createMonth(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: month.
     * Also see {@link #createMonth()}.
     * @draft ICU 64
     */
    static MeasureUnit getMonth();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of duration: month-person.
     * Caller owns returned value and must free it.
     * Also see {@link #getMonthPerson()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createMonthPerson(UErrorCode &status);

    /**
     * Returns by value, unit of duration: month-person.
     * Also see {@link #createMonthPerson()}.
     * @draft ICU 64
     */
    static MeasureUnit getMonthPerson();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: nanosecond.
     * Caller owns returned value and must free it.
     * Also see {@link #getNanosecond()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createNanosecond(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: nanosecond.
     * Also see {@link #createNanosecond()}.
     * @draft ICU 64
     */
    static MeasureUnit getNanosecond();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: second.
     * Caller owns returned value and must free it.
     * Also see {@link #getSecond()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createSecond(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: second.
     * Also see {@link #createSecond()}.
     * @draft ICU 64
     */
    static MeasureUnit getSecond();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: week.
     * Caller owns returned value and must free it.
     * Also see {@link #getWeek()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createWeek(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: week.
     * Also see {@link #createWeek()}.
     * @draft ICU 64
     */
    static MeasureUnit getWeek();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of duration: week-person.
     * Caller owns returned value and must free it.
     * Also see {@link #getWeekPerson()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createWeekPerson(UErrorCode &status);

    /**
     * Returns by value, unit of duration: week-person.
     * Also see {@link #createWeekPerson()}.
     * @draft ICU 64
     */
    static MeasureUnit getWeekPerson();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of duration: year.
     * Caller owns returned value and must free it.
     * Also see {@link #getYear()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createYear(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of duration: year.
     * Also see {@link #createYear()}.
     * @draft ICU 64
     */
    static MeasureUnit getYear();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of duration: year-person.
     * Caller owns returned value and must free it.
     * Also see {@link #getYearPerson()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createYearPerson(UErrorCode &status);

    /**
     * Returns by value, unit of duration: year-person.
     * Also see {@link #createYearPerson()}.
     * @draft ICU 64
     */
    static MeasureUnit getYearPerson();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of electric: ampere.
     * Caller owns returned value and must free it.
     * Also see {@link #getAmpere()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createAmpere(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of electric: ampere.
     * Also see {@link #createAmpere()}.
     * @draft ICU 64
     */
    static MeasureUnit getAmpere();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of electric: milliampere.
     * Caller owns returned value and must free it.
     * Also see {@link #getMilliampere()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMilliampere(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of electric: milliampere.
     * Also see {@link #createMilliampere()}.
     * @draft ICU 64
     */
    static MeasureUnit getMilliampere();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of electric: ohm.
     * Caller owns returned value and must free it.
     * Also see {@link #getOhm()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createOhm(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of electric: ohm.
     * Also see {@link #createOhm()}.
     * @draft ICU 64
     */
    static MeasureUnit getOhm();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of electric: volt.
     * Caller owns returned value and must free it.
     * Also see {@link #getVolt()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createVolt(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of electric: volt.
     * Also see {@link #createVolt()}.
     * @draft ICU 64
     */
    static MeasureUnit getVolt();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of energy: british-thermal-unit.
     * Caller owns returned value and must free it.
     * Also see {@link #getBritishThermalUnit()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createBritishThermalUnit(UErrorCode &status);

    /**
     * Returns by value, unit of energy: british-thermal-unit.
     * Also see {@link #createBritishThermalUnit()}.
     * @draft ICU 64
     */
    static MeasureUnit getBritishThermalUnit();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of energy: calorie.
     * Caller owns returned value and must free it.
     * Also see {@link #getCalorie()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createCalorie(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of energy: calorie.
     * Also see {@link #createCalorie()}.
     * @draft ICU 64
     */
    static MeasureUnit getCalorie();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of energy: electronvolt.
     * Caller owns returned value and must free it.
     * Also see {@link #getElectronvolt()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createElectronvolt(UErrorCode &status);

    /**
     * Returns by value, unit of energy: electronvolt.
     * Also see {@link #createElectronvolt()}.
     * @draft ICU 64
     */
    static MeasureUnit getElectronvolt();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of energy: foodcalorie.
     * Caller owns returned value and must free it.
     * Also see {@link #getFoodcalorie()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createFoodcalorie(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of energy: foodcalorie.
     * Also see {@link #createFoodcalorie()}.
     * @draft ICU 64
     */
    static MeasureUnit getFoodcalorie();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of energy: joule.
     * Caller owns returned value and must free it.
     * Also see {@link #getJoule()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createJoule(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of energy: joule.
     * Also see {@link #createJoule()}.
     * @draft ICU 64
     */
    static MeasureUnit getJoule();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of energy: kilocalorie.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilocalorie()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createKilocalorie(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of energy: kilocalorie.
     * Also see {@link #createKilocalorie()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilocalorie();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of energy: kilojoule.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilojoule()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createKilojoule(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of energy: kilojoule.
     * Also see {@link #createKilojoule()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilojoule();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of energy: kilowatt-hour.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilowattHour()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createKilowattHour(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of energy: kilowatt-hour.
     * Also see {@link #createKilowattHour()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilowattHour();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of energy: therm-us.
     * Caller owns returned value and must free it.
     * Also see {@link #getThermUs()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createThermUs(UErrorCode &status);

    /**
     * Returns by value, unit of energy: therm-us.
     * Also see {@link #createThermUs()}.
     * @draft ICU 65
     */
    static MeasureUnit getThermUs();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of force: newton.
     * Caller owns returned value and must free it.
     * Also see {@link #getNewton()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createNewton(UErrorCode &status);

    /**
     * Returns by value, unit of force: newton.
     * Also see {@link #createNewton()}.
     * @draft ICU 64
     */
    static MeasureUnit getNewton();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of force: pound-force.
     * Caller owns returned value and must free it.
     * Also see {@link #getPoundForce()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createPoundForce(UErrorCode &status);

    /**
     * Returns by value, unit of force: pound-force.
     * Also see {@link #createPoundForce()}.
     * @draft ICU 64
     */
    static MeasureUnit getPoundForce();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of frequency: gigahertz.
     * Caller owns returned value and must free it.
     * Also see {@link #getGigahertz()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createGigahertz(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of frequency: gigahertz.
     * Also see {@link #createGigahertz()}.
     * @draft ICU 64
     */
    static MeasureUnit getGigahertz();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of frequency: hertz.
     * Caller owns returned value and must free it.
     * Also see {@link #getHertz()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createHertz(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of frequency: hertz.
     * Also see {@link #createHertz()}.
     * @draft ICU 64
     */
    static MeasureUnit getHertz();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of frequency: kilohertz.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilohertz()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createKilohertz(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of frequency: kilohertz.
     * Also see {@link #createKilohertz()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilohertz();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of frequency: megahertz.
     * Caller owns returned value and must free it.
     * Also see {@link #getMegahertz()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMegahertz(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of frequency: megahertz.
     * Also see {@link #createMegahertz()}.
     * @draft ICU 64
     */
    static MeasureUnit getMegahertz();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of graphics: dot-per-centimeter.
     * Caller owns returned value and must free it.
     * Also see {@link #getDotPerCentimeter()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createDotPerCentimeter(UErrorCode &status);

    /**
     * Returns by value, unit of graphics: dot-per-centimeter.
     * Also see {@link #createDotPerCentimeter()}.
     * @draft ICU 65
     */
    static MeasureUnit getDotPerCentimeter();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of graphics: dot-per-inch.
     * Caller owns returned value and must free it.
     * Also see {@link #getDotPerInch()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createDotPerInch(UErrorCode &status);

    /**
     * Returns by value, unit of graphics: dot-per-inch.
     * Also see {@link #createDotPerInch()}.
     * @draft ICU 65
     */
    static MeasureUnit getDotPerInch();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of graphics: em.
     * Caller owns returned value and must free it.
     * Also see {@link #getEm()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createEm(UErrorCode &status);

    /**
     * Returns by value, unit of graphics: em.
     * Also see {@link #createEm()}.
     * @draft ICU 65
     */
    static MeasureUnit getEm();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of graphics: megapixel.
     * Caller owns returned value and must free it.
     * Also see {@link #getMegapixel()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createMegapixel(UErrorCode &status);

    /**
     * Returns by value, unit of graphics: megapixel.
     * Also see {@link #createMegapixel()}.
     * @draft ICU 65
     */
    static MeasureUnit getMegapixel();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of graphics: pixel.
     * Caller owns returned value and must free it.
     * Also see {@link #getPixel()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createPixel(UErrorCode &status);

    /**
     * Returns by value, unit of graphics: pixel.
     * Also see {@link #createPixel()}.
     * @draft ICU 65
     */
    static MeasureUnit getPixel();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of graphics: pixel-per-centimeter.
     * Caller owns returned value and must free it.
     * Also see {@link #getPixelPerCentimeter()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createPixelPerCentimeter(UErrorCode &status);

    /**
     * Returns by value, unit of graphics: pixel-per-centimeter.
     * Also see {@link #createPixelPerCentimeter()}.
     * @draft ICU 65
     */
    static MeasureUnit getPixelPerCentimeter();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of graphics: pixel-per-inch.
     * Caller owns returned value and must free it.
     * Also see {@link #getPixelPerInch()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createPixelPerInch(UErrorCode &status);

    /**
     * Returns by value, unit of graphics: pixel-per-inch.
     * Also see {@link #createPixelPerInch()}.
     * @draft ICU 65
     */
    static MeasureUnit getPixelPerInch();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: astronomical-unit.
     * Caller owns returned value and must free it.
     * Also see {@link #getAstronomicalUnit()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createAstronomicalUnit(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: astronomical-unit.
     * Also see {@link #createAstronomicalUnit()}.
     * @draft ICU 64
     */
    static MeasureUnit getAstronomicalUnit();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: centimeter.
     * Caller owns returned value and must free it.
     * Also see {@link #getCentimeter()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createCentimeter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: centimeter.
     * Also see {@link #createCentimeter()}.
     * @draft ICU 64
     */
    static MeasureUnit getCentimeter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: decimeter.
     * Caller owns returned value and must free it.
     * Also see {@link #getDecimeter()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createDecimeter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: decimeter.
     * Also see {@link #createDecimeter()}.
     * @draft ICU 64
     */
    static MeasureUnit getDecimeter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: fathom.
     * Caller owns returned value and must free it.
     * Also see {@link #getFathom()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createFathom(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: fathom.
     * Also see {@link #createFathom()}.
     * @draft ICU 64
     */
    static MeasureUnit getFathom();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: foot.
     * Caller owns returned value and must free it.
     * Also see {@link #getFoot()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createFoot(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: foot.
     * Also see {@link #createFoot()}.
     * @draft ICU 64
     */
    static MeasureUnit getFoot();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: furlong.
     * Caller owns returned value and must free it.
     * Also see {@link #getFurlong()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createFurlong(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: furlong.
     * Also see {@link #createFurlong()}.
     * @draft ICU 64
     */
    static MeasureUnit getFurlong();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: inch.
     * Caller owns returned value and must free it.
     * Also see {@link #getInch()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createInch(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: inch.
     * Also see {@link #createInch()}.
     * @draft ICU 64
     */
    static MeasureUnit getInch();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: kilometer.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilometer()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createKilometer(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: kilometer.
     * Also see {@link #createKilometer()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilometer();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: light-year.
     * Caller owns returned value and must free it.
     * Also see {@link #getLightYear()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createLightYear(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: light-year.
     * Also see {@link #createLightYear()}.
     * @draft ICU 64
     */
    static MeasureUnit getLightYear();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: meter.
     * Caller owns returned value and must free it.
     * Also see {@link #getMeter()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createMeter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: meter.
     * Also see {@link #createMeter()}.
     * @draft ICU 64
     */
    static MeasureUnit getMeter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: micrometer.
     * Caller owns returned value and must free it.
     * Also see {@link #getMicrometer()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMicrometer(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: micrometer.
     * Also see {@link #createMicrometer()}.
     * @draft ICU 64
     */
    static MeasureUnit getMicrometer();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: mile.
     * Caller owns returned value and must free it.
     * Also see {@link #getMile()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createMile(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: mile.
     * Also see {@link #createMile()}.
     * @draft ICU 64
     */
    static MeasureUnit getMile();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: mile-scandinavian.
     * Caller owns returned value and must free it.
     * Also see {@link #getMileScandinavian()}.
     * @param status ICU error code.
     * @stable ICU 56
     */
    static MeasureUnit *createMileScandinavian(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: mile-scandinavian.
     * Also see {@link #createMileScandinavian()}.
     * @draft ICU 64
     */
    static MeasureUnit getMileScandinavian();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: millimeter.
     * Caller owns returned value and must free it.
     * Also see {@link #getMillimeter()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createMillimeter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: millimeter.
     * Also see {@link #createMillimeter()}.
     * @draft ICU 64
     */
    static MeasureUnit getMillimeter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: nanometer.
     * Caller owns returned value and must free it.
     * Also see {@link #getNanometer()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createNanometer(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: nanometer.
     * Also see {@link #createNanometer()}.
     * @draft ICU 64
     */
    static MeasureUnit getNanometer();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: nautical-mile.
     * Caller owns returned value and must free it.
     * Also see {@link #getNauticalMile()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createNauticalMile(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: nautical-mile.
     * Also see {@link #createNauticalMile()}.
     * @draft ICU 64
     */
    static MeasureUnit getNauticalMile();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: parsec.
     * Caller owns returned value and must free it.
     * Also see {@link #getParsec()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createParsec(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: parsec.
     * Also see {@link #createParsec()}.
     * @draft ICU 64
     */
    static MeasureUnit getParsec();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: picometer.
     * Caller owns returned value and must free it.
     * Also see {@link #getPicometer()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createPicometer(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: picometer.
     * Also see {@link #createPicometer()}.
     * @draft ICU 64
     */
    static MeasureUnit getPicometer();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: point.
     * Caller owns returned value and must free it.
     * Also see {@link #getPoint()}.
     * @param status ICU error code.
     * @stable ICU 59
     */
    static MeasureUnit *createPoint(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: point.
     * Also see {@link #createPoint()}.
     * @draft ICU 64
     */
    static MeasureUnit getPoint();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of length: solar-radius.
     * Caller owns returned value and must free it.
     * Also see {@link #getSolarRadius()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createSolarRadius(UErrorCode &status);

    /**
     * Returns by value, unit of length: solar-radius.
     * Also see {@link #createSolarRadius()}.
     * @draft ICU 64
     */
    static MeasureUnit getSolarRadius();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of length: yard.
     * Caller owns returned value and must free it.
     * Also see {@link #getYard()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createYard(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of length: yard.
     * Also see {@link #createYard()}.
     * @draft ICU 64
     */
    static MeasureUnit getYard();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of light: lux.
     * Caller owns returned value and must free it.
     * Also see {@link #getLux()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createLux(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of light: lux.
     * Also see {@link #createLux()}.
     * @draft ICU 64
     */
    static MeasureUnit getLux();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of light: solar-luminosity.
     * Caller owns returned value and must free it.
     * Also see {@link #getSolarLuminosity()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createSolarLuminosity(UErrorCode &status);

    /**
     * Returns by value, unit of light: solar-luminosity.
     * Also see {@link #createSolarLuminosity()}.
     * @draft ICU 64
     */
    static MeasureUnit getSolarLuminosity();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: carat.
     * Caller owns returned value and must free it.
     * Also see {@link #getCarat()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createCarat(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: carat.
     * Also see {@link #createCarat()}.
     * @draft ICU 64
     */
    static MeasureUnit getCarat();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of mass: dalton.
     * Caller owns returned value and must free it.
     * Also see {@link #getDalton()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createDalton(UErrorCode &status);

    /**
     * Returns by value, unit of mass: dalton.
     * Also see {@link #createDalton()}.
     * @draft ICU 64
     */
    static MeasureUnit getDalton();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of mass: earth-mass.
     * Caller owns returned value and must free it.
     * Also see {@link #getEarthMass()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createEarthMass(UErrorCode &status);

    /**
     * Returns by value, unit of mass: earth-mass.
     * Also see {@link #createEarthMass()}.
     * @draft ICU 64
     */
    static MeasureUnit getEarthMass();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: gram.
     * Caller owns returned value and must free it.
     * Also see {@link #getGram()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createGram(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: gram.
     * Also see {@link #createGram()}.
     * @draft ICU 64
     */
    static MeasureUnit getGram();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: kilogram.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilogram()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createKilogram(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: kilogram.
     * Also see {@link #createKilogram()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilogram();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: metric-ton.
     * Caller owns returned value and must free it.
     * Also see {@link #getMetricTon()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMetricTon(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: metric-ton.
     * Also see {@link #createMetricTon()}.
     * @draft ICU 64
     */
    static MeasureUnit getMetricTon();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: microgram.
     * Caller owns returned value and must free it.
     * Also see {@link #getMicrogram()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMicrogram(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: microgram.
     * Also see {@link #createMicrogram()}.
     * @draft ICU 64
     */
    static MeasureUnit getMicrogram();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: milligram.
     * Caller owns returned value and must free it.
     * Also see {@link #getMilligram()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMilligram(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: milligram.
     * Also see {@link #createMilligram()}.
     * @draft ICU 64
     */
    static MeasureUnit getMilligram();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: ounce.
     * Caller owns returned value and must free it.
     * Also see {@link #getOunce()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createOunce(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: ounce.
     * Also see {@link #createOunce()}.
     * @draft ICU 64
     */
    static MeasureUnit getOunce();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: ounce-troy.
     * Caller owns returned value and must free it.
     * Also see {@link #getOunceTroy()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createOunceTroy(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: ounce-troy.
     * Also see {@link #createOunceTroy()}.
     * @draft ICU 64
     */
    static MeasureUnit getOunceTroy();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: pound.
     * Caller owns returned value and must free it.
     * Also see {@link #getPound()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createPound(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: pound.
     * Also see {@link #createPound()}.
     * @draft ICU 64
     */
    static MeasureUnit getPound();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of mass: solar-mass.
     * Caller owns returned value and must free it.
     * Also see {@link #getSolarMass()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createSolarMass(UErrorCode &status);

    /**
     * Returns by value, unit of mass: solar-mass.
     * Also see {@link #createSolarMass()}.
     * @draft ICU 64
     */
    static MeasureUnit getSolarMass();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: stone.
     * Caller owns returned value and must free it.
     * Also see {@link #getStone()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createStone(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: stone.
     * Also see {@link #createStone()}.
     * @draft ICU 64
     */
    static MeasureUnit getStone();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of mass: ton.
     * Caller owns returned value and must free it.
     * Also see {@link #getTon()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createTon(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of mass: ton.
     * Also see {@link #createTon()}.
     * @draft ICU 64
     */
    static MeasureUnit getTon();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of power: gigawatt.
     * Caller owns returned value and must free it.
     * Also see {@link #getGigawatt()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createGigawatt(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of power: gigawatt.
     * Also see {@link #createGigawatt()}.
     * @draft ICU 64
     */
    static MeasureUnit getGigawatt();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of power: horsepower.
     * Caller owns returned value and must free it.
     * Also see {@link #getHorsepower()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createHorsepower(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of power: horsepower.
     * Also see {@link #createHorsepower()}.
     * @draft ICU 64
     */
    static MeasureUnit getHorsepower();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of power: kilowatt.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilowatt()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createKilowatt(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of power: kilowatt.
     * Also see {@link #createKilowatt()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilowatt();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of power: megawatt.
     * Caller owns returned value and must free it.
     * Also see {@link #getMegawatt()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMegawatt(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of power: megawatt.
     * Also see {@link #createMegawatt()}.
     * @draft ICU 64
     */
    static MeasureUnit getMegawatt();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of power: milliwatt.
     * Caller owns returned value and must free it.
     * Also see {@link #getMilliwatt()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMilliwatt(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of power: milliwatt.
     * Also see {@link #createMilliwatt()}.
     * @draft ICU 64
     */
    static MeasureUnit getMilliwatt();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of power: watt.
     * Caller owns returned value and must free it.
     * Also see {@link #getWatt()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createWatt(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of power: watt.
     * Also see {@link #createWatt()}.
     * @draft ICU 64
     */
    static MeasureUnit getWatt();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of pressure: atmosphere.
     * Caller owns returned value and must free it.
     * Also see {@link #getAtmosphere()}.
     * @param status ICU error code.
     * @stable ICU 63
     */
    static MeasureUnit *createAtmosphere(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of pressure: atmosphere.
     * Also see {@link #createAtmosphere()}.
     * @draft ICU 64
     */
    static MeasureUnit getAtmosphere();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of pressure: bar.
     * Caller owns returned value and must free it.
     * Also see {@link #getBar()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createBar(UErrorCode &status);

    /**
     * Returns by value, unit of pressure: bar.
     * Also see {@link #createBar()}.
     * @draft ICU 65
     */
    static MeasureUnit getBar();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of pressure: hectopascal.
     * Caller owns returned value and must free it.
     * Also see {@link #getHectopascal()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createHectopascal(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of pressure: hectopascal.
     * Also see {@link #createHectopascal()}.
     * @draft ICU 64
     */
    static MeasureUnit getHectopascal();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of pressure: inch-hg.
     * Caller owns returned value and must free it.
     * Also see {@link #getInchHg()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createInchHg(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of pressure: inch-hg.
     * Also see {@link #createInchHg()}.
     * @draft ICU 64
     */
    static MeasureUnit getInchHg();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of pressure: kilopascal.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilopascal()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createKilopascal(UErrorCode &status);

    /**
     * Returns by value, unit of pressure: kilopascal.
     * Also see {@link #createKilopascal()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilopascal();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of pressure: megapascal.
     * Caller owns returned value and must free it.
     * Also see {@link #getMegapascal()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createMegapascal(UErrorCode &status);

    /**
     * Returns by value, unit of pressure: megapascal.
     * Also see {@link #createMegapascal()}.
     * @draft ICU 64
     */
    static MeasureUnit getMegapascal();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of pressure: millibar.
     * Caller owns returned value and must free it.
     * Also see {@link #getMillibar()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createMillibar(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of pressure: millibar.
     * Also see {@link #createMillibar()}.
     * @draft ICU 64
     */
    static MeasureUnit getMillibar();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of pressure: millimeter-of-mercury.
     * Caller owns returned value and must free it.
     * Also see {@link #getMillimeterOfMercury()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMillimeterOfMercury(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of pressure: millimeter-of-mercury.
     * Also see {@link #createMillimeterOfMercury()}.
     * @draft ICU 64
     */
    static MeasureUnit getMillimeterOfMercury();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of pressure: pascal.
     * Caller owns returned value and must free it.
     * Also see {@link #getPascal()}.
     * @param status ICU error code.
     * @draft ICU 65
     */
    static MeasureUnit *createPascal(UErrorCode &status);

    /**
     * Returns by value, unit of pressure: pascal.
     * Also see {@link #createPascal()}.
     * @draft ICU 65
     */
    static MeasureUnit getPascal();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of pressure: pound-per-square-inch.
     * Caller owns returned value and must free it.
     * Also see {@link #getPoundPerSquareInch()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createPoundPerSquareInch(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of pressure: pound-per-square-inch.
     * Also see {@link #createPoundPerSquareInch()}.
     * @draft ICU 64
     */
    static MeasureUnit getPoundPerSquareInch();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of speed: kilometer-per-hour.
     * Caller owns returned value and must free it.
     * Also see {@link #getKilometerPerHour()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createKilometerPerHour(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of speed: kilometer-per-hour.
     * Also see {@link #createKilometerPerHour()}.
     * @draft ICU 64
     */
    static MeasureUnit getKilometerPerHour();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of speed: knot.
     * Caller owns returned value and must free it.
     * Also see {@link #getKnot()}.
     * @param status ICU error code.
     * @stable ICU 56
     */
    static MeasureUnit *createKnot(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of speed: knot.
     * Also see {@link #createKnot()}.
     * @draft ICU 64
     */
    static MeasureUnit getKnot();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of speed: meter-per-second.
     * Caller owns returned value and must free it.
     * Also see {@link #getMeterPerSecond()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createMeterPerSecond(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of speed: meter-per-second.
     * Also see {@link #createMeterPerSecond()}.
     * @draft ICU 64
     */
    static MeasureUnit getMeterPerSecond();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of speed: mile-per-hour.
     * Caller owns returned value and must free it.
     * Also see {@link #getMilePerHour()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createMilePerHour(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of speed: mile-per-hour.
     * Also see {@link #createMilePerHour()}.
     * @draft ICU 64
     */
    static MeasureUnit getMilePerHour();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of temperature: celsius.
     * Caller owns returned value and must free it.
     * Also see {@link #getCelsius()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createCelsius(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of temperature: celsius.
     * Also see {@link #createCelsius()}.
     * @draft ICU 64
     */
    static MeasureUnit getCelsius();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of temperature: fahrenheit.
     * Caller owns returned value and must free it.
     * Also see {@link #getFahrenheit()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createFahrenheit(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of temperature: fahrenheit.
     * Also see {@link #createFahrenheit()}.
     * @draft ICU 64
     */
    static MeasureUnit getFahrenheit();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of temperature: generic.
     * Caller owns returned value and must free it.
     * Also see {@link #getGenericTemperature()}.
     * @param status ICU error code.
     * @stable ICU 56
     */
    static MeasureUnit *createGenericTemperature(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of temperature: generic.
     * Also see {@link #createGenericTemperature()}.
     * @draft ICU 64
     */
    static MeasureUnit getGenericTemperature();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of temperature: kelvin.
     * Caller owns returned value and must free it.
     * Also see {@link #getKelvin()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createKelvin(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of temperature: kelvin.
     * Also see {@link #createKelvin()}.
     * @draft ICU 64
     */
    static MeasureUnit getKelvin();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of torque: newton-meter.
     * Caller owns returned value and must free it.
     * Also see {@link #getNewtonMeter()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createNewtonMeter(UErrorCode &status);

    /**
     * Returns by value, unit of torque: newton-meter.
     * Also see {@link #createNewtonMeter()}.
     * @draft ICU 64
     */
    static MeasureUnit getNewtonMeter();
#endif /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of torque: pound-foot.
     * Caller owns returned value and must free it.
     * Also see {@link #getPoundFoot()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createPoundFoot(UErrorCode &status);

    /**
     * Returns by value, unit of torque: pound-foot.
     * Also see {@link #createPoundFoot()}.
     * @draft ICU 64
     */
    static MeasureUnit getPoundFoot();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: acre-foot.
     * Caller owns returned value and must free it.
     * Also see {@link #getAcreFoot()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createAcreFoot(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: acre-foot.
     * Also see {@link #createAcreFoot()}.
     * @draft ICU 64
     */
    static MeasureUnit getAcreFoot();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of volume: barrel.
     * Caller owns returned value and must free it.
     * Also see {@link #getBarrel()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createBarrel(UErrorCode &status);

    /**
     * Returns by value, unit of volume: barrel.
     * Also see {@link #createBarrel()}.
     * @draft ICU 64
     */
    static MeasureUnit getBarrel();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: bushel.
     * Caller owns returned value and must free it.
     * Also see {@link #getBushel()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createBushel(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: bushel.
     * Also see {@link #createBushel()}.
     * @draft ICU 64
     */
    static MeasureUnit getBushel();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: centiliter.
     * Caller owns returned value and must free it.
     * Also see {@link #getCentiliter()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createCentiliter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: centiliter.
     * Also see {@link #createCentiliter()}.
     * @draft ICU 64
     */
    static MeasureUnit getCentiliter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: cubic-centimeter.
     * Caller owns returned value and must free it.
     * Also see {@link #getCubicCentimeter()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createCubicCentimeter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: cubic-centimeter.
     * Also see {@link #createCubicCentimeter()}.
     * @draft ICU 64
     */
    static MeasureUnit getCubicCentimeter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: cubic-foot.
     * Caller owns returned value and must free it.
     * Also see {@link #getCubicFoot()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createCubicFoot(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: cubic-foot.
     * Also see {@link #createCubicFoot()}.
     * @draft ICU 64
     */
    static MeasureUnit getCubicFoot();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: cubic-inch.
     * Caller owns returned value and must free it.
     * Also see {@link #getCubicInch()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createCubicInch(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: cubic-inch.
     * Also see {@link #createCubicInch()}.
     * @draft ICU 64
     */
    static MeasureUnit getCubicInch();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: cubic-kilometer.
     * Caller owns returned value and must free it.
     * Also see {@link #getCubicKilometer()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createCubicKilometer(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: cubic-kilometer.
     * Also see {@link #createCubicKilometer()}.
     * @draft ICU 64
     */
    static MeasureUnit getCubicKilometer();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: cubic-meter.
     * Caller owns returned value and must free it.
     * Also see {@link #getCubicMeter()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createCubicMeter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: cubic-meter.
     * Also see {@link #createCubicMeter()}.
     * @draft ICU 64
     */
    static MeasureUnit getCubicMeter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: cubic-mile.
     * Caller owns returned value and must free it.
     * Also see {@link #getCubicMile()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createCubicMile(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: cubic-mile.
     * Also see {@link #createCubicMile()}.
     * @draft ICU 64
     */
    static MeasureUnit getCubicMile();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: cubic-yard.
     * Caller owns returned value and must free it.
     * Also see {@link #getCubicYard()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createCubicYard(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: cubic-yard.
     * Also see {@link #createCubicYard()}.
     * @draft ICU 64
     */
    static MeasureUnit getCubicYard();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: cup.
     * Caller owns returned value and must free it.
     * Also see {@link #getCup()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createCup(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: cup.
     * Also see {@link #createCup()}.
     * @draft ICU 64
     */
    static MeasureUnit getCup();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: cup-metric.
     * Caller owns returned value and must free it.
     * Also see {@link #getCupMetric()}.
     * @param status ICU error code.
     * @stable ICU 56
     */
    static MeasureUnit *createCupMetric(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: cup-metric.
     * Also see {@link #createCupMetric()}.
     * @draft ICU 64
     */
    static MeasureUnit getCupMetric();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: deciliter.
     * Caller owns returned value and must free it.
     * Also see {@link #getDeciliter()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createDeciliter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: deciliter.
     * Also see {@link #createDeciliter()}.
     * @draft ICU 64
     */
    static MeasureUnit getDeciliter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: fluid-ounce.
     * Caller owns returned value and must free it.
     * Also see {@link #getFluidOunce()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createFluidOunce(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: fluid-ounce.
     * Also see {@link #createFluidOunce()}.
     * @draft ICU 64
     */
    static MeasureUnit getFluidOunce();
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by pointer, unit of volume: fluid-ounce-imperial.
     * Caller owns returned value and must free it.
     * Also see {@link #getFluidOunceImperial()}.
     * @param status ICU error code.
     * @draft ICU 64
     */
    static MeasureUnit *createFluidOunceImperial(UErrorCode &status);

    /**
     * Returns by value, unit of volume: fluid-ounce-imperial.
     * Also see {@link #createFluidOunceImperial()}.
     * @draft ICU 64
     */
    static MeasureUnit getFluidOunceImperial();
#endif /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: gallon.
     * Caller owns returned value and must free it.
     * Also see {@link #getGallon()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createGallon(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: gallon.
     * Also see {@link #createGallon()}.
     * @draft ICU 64
     */
    static MeasureUnit getGallon();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: gallon-imperial.
     * Caller owns returned value and must free it.
     * Also see {@link #getGallonImperial()}.
     * @param status ICU error code.
     * @stable ICU 57
     */
    static MeasureUnit *createGallonImperial(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: gallon-imperial.
     * Also see {@link #createGallonImperial()}.
     * @draft ICU 64
     */
    static MeasureUnit getGallonImperial();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: hectoliter.
     * Caller owns returned value and must free it.
     * Also see {@link #getHectoliter()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createHectoliter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: hectoliter.
     * Also see {@link #createHectoliter()}.
     * @draft ICU 64
     */
    static MeasureUnit getHectoliter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: liter.
     * Caller owns returned value and must free it.
     * Also see {@link #getLiter()}.
     * @param status ICU error code.
     * @stable ICU 53
     */
    static MeasureUnit *createLiter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: liter.
     * Also see {@link #createLiter()}.
     * @draft ICU 64
     */
    static MeasureUnit getLiter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: megaliter.
     * Caller owns returned value and must free it.
     * Also see {@link #getMegaliter()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMegaliter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: megaliter.
     * Also see {@link #createMegaliter()}.
     * @draft ICU 64
     */
    static MeasureUnit getMegaliter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: milliliter.
     * Caller owns returned value and must free it.
     * Also see {@link #getMilliliter()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createMilliliter(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: milliliter.
     * Also see {@link #createMilliliter()}.
     * @draft ICU 64
     */
    static MeasureUnit getMilliliter();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: pint.
     * Caller owns returned value and must free it.
     * Also see {@link #getPint()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createPint(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: pint.
     * Also see {@link #createPint()}.
     * @draft ICU 64
     */
    static MeasureUnit getPint();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: pint-metric.
     * Caller owns returned value and must free it.
     * Also see {@link #getPintMetric()}.
     * @param status ICU error code.
     * @stable ICU 56
     */
    static MeasureUnit *createPintMetric(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: pint-metric.
     * Also see {@link #createPintMetric()}.
     * @draft ICU 64
     */
    static MeasureUnit getPintMetric();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: quart.
     * Caller owns returned value and must free it.
     * Also see {@link #getQuart()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createQuart(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: quart.
     * Also see {@link #createQuart()}.
     * @draft ICU 64
     */
    static MeasureUnit getQuart();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: tablespoon.
     * Caller owns returned value and must free it.
     * Also see {@link #getTablespoon()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createTablespoon(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: tablespoon.
     * Also see {@link #createTablespoon()}.
     * @draft ICU 64
     */
    static MeasureUnit getTablespoon();
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Returns by pointer, unit of volume: teaspoon.
     * Caller owns returned value and must free it.
     * Also see {@link #getTeaspoon()}.
     * @param status ICU error code.
     * @stable ICU 54
     */
    static MeasureUnit *createTeaspoon(UErrorCode &status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns by value, unit of volume: teaspoon.
     * Also see {@link #createTeaspoon()}.
     * @draft ICU 64
     */
    static MeasureUnit getTeaspoon();
#endif  /* U_HIDE_DRAFT_API */


// End generated createXXX methods

 protected:

#ifndef U_HIDE_INTERNAL_API
    /**
     * For ICU use only.
     * @internal
     */
    void initTime(const char *timeId);

    /**
     * For ICU use only.
     * @internal
     */
    void initCurrency(const char *isoCurrency);

    /**
     * For ICU use only.
     * @internal
     */
    void initNoUnit(const char *subtype);

#endif  /* U_HIDE_INTERNAL_API */

private:
    int32_t fTypeId;
    int32_t fSubTypeId;
    char fCurrency[4];

    MeasureUnit(int32_t typeId, int32_t subTypeId) : fTypeId(typeId), fSubTypeId(subTypeId) {
        fCurrency[0] = 0;
    }
    void setTo(int32_t typeId, int32_t subTypeId);
    int32_t getOffset() const;
    static MeasureUnit *create(int typeId, int subTypeId, UErrorCode &status);
};

U_NAMESPACE_END

#endif // !UNCONFIG_NO_FORMATTING

#endif /* U_SHOW_CPLUSPLUS_API */



/**
 * \file 
 * \brief C++ API: Currency Unit Information.
 */
 
U_NAMESPACE_BEGIN

/**
 * A unit of currency, such as USD (U.S. dollars) or JPY (Japanese
 * yen).  This class is a thin wrapper over a char16_t string that
 * subclasses MeasureUnit, for use with Measure and MeasureFormat.
 *
 * @author Alan Liu
 * @stable ICU 3.0
 */
class U_I18N_API CurrencyUnit: public MeasureUnit {
 public:
    /**
     * Default constructor.  Initializes currency code to "XXX" (no currency).
     * @stable ICU 60
     */
    CurrencyUnit();

    /**
     * Construct an object with the given ISO currency code.
     *
     * @param isoCode the 3-letter ISO 4217 currency code; must have
     * length 3 and need not be NUL-terminated. If NULL, the currency
     * is initialized to the unknown currency XXX.
     * @param ec input-output error code. If the isoCode is invalid,
     * then this will be set to a failing value.
     * @stable ICU 3.0
     */
    CurrencyUnit(ConstChar16Ptr isoCode, UErrorCode &ec);

#ifndef U_HIDE_DRAFT_API
    /**
     * Construct an object with the given ISO currency code.
     *
     * @param isoCode the 3-letter ISO 4217 currency code; must have
     * length 3. If invalid, the currency is initialized to XXX.
     * @param ec input-output error code. If the isoCode is invalid,
     * then this will be set to a failing value.
     * @draft ICU 64
     */
    CurrencyUnit(StringPiece isoCode, UErrorCode &ec);
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Copy constructor
     * @stable ICU 3.0
     */
    CurrencyUnit(const CurrencyUnit& other);

    /**
     * Copy constructor from MeasureUnit. This constructor allows you to
     * restore a CurrencyUnit that was sliced to MeasureUnit.
     *
     * @param measureUnit The MeasureUnit to copy from.
     * @param ec Set to a failing value if the MeasureUnit is not a currency.
     * @stable ICU 60
     */
    CurrencyUnit(const MeasureUnit& measureUnit, UErrorCode &ec);

    /**
     * Assignment operator
     * @stable ICU 3.0
     */
    CurrencyUnit& operator=(const CurrencyUnit& other);

    /**
     * Return a polymorphic clone of this object.  The result will
     * have the same class as returned by getDynamicClassID().
     * @stable ICU 3.0
     */
    virtual CurrencyUnit* clone() const;

    /**
     * Destructor
     * @stable ICU 3.0
     */
    virtual ~CurrencyUnit();

    /**
     * Returns a unique class ID for this object POLYMORPHICALLY.
     * This method implements a simple form of RTTI used by ICU.
     * @return The class ID for this object. All objects of a given
     * class have the same class ID.  Objects of other classes have
     * different class IDs.
     * @stable ICU 3.0
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * Returns the class ID for this class. This is used to compare to
     * the return value of getDynamicClassID().
     * @return The class ID for all objects of this class.
     * @stable ICU 3.0
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * Return the ISO currency code of this object.
     * @stable ICU 3.0
     */
    inline const char16_t* getISOCurrency() const;

 private:
    /**
     * The ISO 4217 code of this object.
     */
    char16_t isoCode[4];
};

inline const char16_t* CurrencyUnit::getISOCurrency() const {
    return isoCode;
}

U_NAMESPACE_END

#endif // !UCONFIG_NO_FORMATTING

#endif /* U_SHOW_CPLUSPLUS_API */



/**
 * \file 
 * \brief C++ API: Currency Amount Object.
 */
 
U_NAMESPACE_BEGIN

/**
 *
 * A currency together with a numeric amount, such as 200 USD.
 *
 * @author Alan Liu
 * @stable ICU 3.0
 */
class U_I18N_API CurrencyAmount: public Measure {
 public:
    /**
     * Construct an object with the given numeric amount and the given
     * ISO currency code.
     * @param amount a numeric object; amount.isNumeric() must be TRUE
     * @param isoCode the 3-letter ISO 4217 currency code; must not be
     * NULL and must have length 3
     * @param ec input-output error code. If the amount or the isoCode
     * is invalid, then this will be set to a failing value.
     * @stable ICU 3.0
     */
    CurrencyAmount(const Formattable& amount, ConstChar16Ptr isoCode,
                   UErrorCode &ec);

    /**
     * Construct an object with the given numeric amount and the given
     * ISO currency code.
     * @param amount the amount of the given currency
     * @param isoCode the 3-letter ISO 4217 currency code; must not be
     * NULL and must have length 3
     * @param ec input-output error code. If the isoCode is invalid,
     * then this will be set to a failing value.
     * @stable ICU 3.0
     */
    CurrencyAmount(double amount, ConstChar16Ptr isoCode,
                   UErrorCode &ec);

    /**
     * Copy constructor
     * @stable ICU 3.0
     */
    CurrencyAmount(const CurrencyAmount& other);
 
    /**
     * Assignment operator
     * @stable ICU 3.0
     */
    CurrencyAmount& operator=(const CurrencyAmount& other);

    /**
     * Return a polymorphic clone of this object.  The result will
     * have the same class as returned by getDynamicClassID().
     * @stable ICU 3.0
     */
    virtual CurrencyAmount* clone() const;

    /**
     * Destructor
     * @stable ICU 3.0
     */
    virtual ~CurrencyAmount();
    
    /**
     * Returns a unique class ID for this object POLYMORPHICALLY.
     * This method implements a simple form of RTTI used by ICU.
     * @return The class ID for this object. All objects of a given
     * class have the same class ID.  Objects of other classes have
     * different class IDs.
     * @stable ICU 3.0
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * Returns the class ID for this class. This is used to compare to
     * the return value of getDynamicClassID().
     * @return The class ID for all objects of this class.
     * @stable ICU 3.0
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * Return the currency unit object of this object.
     * @stable ICU 3.0
     */
    inline const CurrencyUnit& getCurrency() const;

    /**
     * Return the ISO currency code of this object.
     * @stable ICU 3.0
     */
    inline const char16_t* getISOCurrency() const;
};

inline const CurrencyUnit& CurrencyAmount::getCurrency() const {
    return (const CurrencyUnit&) getUnit();
}

inline const char16_t* CurrencyAmount::getISOCurrency() const {
    return getCurrency().getISOCurrency();
}

U_NAMESPACE_END

#endif // !UCONFIG_NO_FORMATTING

#endif /* U_SHOW_CPLUSPLUS_API */




class NumberFormatTest;

U_NAMESPACE_BEGIN

class SharedNumberFormat;

#if !UCONFIG_NO_SERVICE
class NumberFormatFactory;
class StringEnumeration;
#endif

/**
 * <p><strong>IMPORTANT:</strong> New users are strongly encouraged to see if
 * numberformatter.h fits their use case.  Although not deprecated, this header
 * is provided for backwards compatibility only.
 *
 * Abstract base class for all number formats.  Provides interface for
 * formatting and parsing a number.  Also provides methods for
 * determining which locales have number formats, and what their names
 * are.
 *
 * \headerfile unicode/numfmt.h "unicode/numfmt.h"
 * <P>
 * NumberFormat helps you to format and parse numbers for any locale.
 * Your code can be completely independent of the locale conventions
 * for decimal points, thousands-separators, or even the particular
 * decimal digits used, or whether the number format is even decimal.
 * <P>
 * To format a number for the current Locale, use one of the static
 * factory methods:
 * \code
 *    #include <iostream>
 *    
 *    
 *    #include "unicode/ustream.h"
 *    using namespace std;
 *
 *    int main() {
 *        double myNumber = 7.0;
 *        UnicodeString myString;
 *        UErrorCode success = U_ZERO_ERROR;
 *        NumberFormat* nf = NumberFormat::createInstance(success);
 *        nf->format(myNumber, myString);
 *        cout << " Example 1: " << myString << endl;
 *    }
 * \endcode
 * Note that there are additional factory methods within subclasses of
 * NumberFormat.
 * <P>
 * If you are formatting multiple numbers, it is more efficient to get
 * the format and use it multiple times so that the system doesn't
 * have to fetch the information about the local language and country
 * conventions multiple times.
 * \code
 *     UnicodeString myString;
 *     UErrorCode success = U_ZERO_ERROR;
 *     NumberFormat *nf = NumberFormat::createInstance( success );
 *     for (int32_t number: {123, 3333, -1234567}) {
 *         nf->format(number, myString);
 *         myString += "; ";
 *     }
 *     cout << " Example 2: " << myString << endl;
 * \endcode
 * To format a number for a different Locale, specify it in the
 * call to \c createInstance().
 * \code
 *     nf = NumberFormat::createInstance(Locale::getFrench(), success);
 * \endcode
 * You can use a \c NumberFormat to parse also.
 * \code
 *    UErrorCode success;
 *    Formattable result(-999);  // initialized with error code
 *    nf->parse(myString, result, success);
 * \endcode
 * Use \c createInstance() to get the normal number format for a \c Locale.
 * There are other static factory methods available.  Use \c createCurrencyInstance()
 * to get the currency number format for that country.  Use \c createPercentInstance()
 * to get a format for displaying percentages. With this format, a
 * fraction from 0.53 is displayed as 53%.
 * <P>
 * The type of number formatting can be specified by passing a 'style' parameter to \c createInstance().
 * For example, use\n
 * \c createInstance(locale, UNUM_DECIMAL, errorCode) to get the normal number format,\n
 * \c createInstance(locale, UNUM_PERCENT, errorCode) to get a format for displaying percentage,\n
 * \c createInstance(locale, UNUM_SCIENTIFIC, errorCode) to get a format for displaying scientific number,\n
 * \c createInstance(locale, UNUM_CURRENCY, errorCode) to get the currency number format,
 * in which the currency is represented by its symbol, for example, "$3.00".\n
 * \c createInstance(locale, UNUM_CURRENCY_ISO, errorCode)  to get the currency number format,
 * in which the currency is represented by its ISO code, for example "USD3.00".\n
 * \c createInstance(locale, UNUM_CURRENCY_PLURAL, errorCode) to get the currency number format,
 * in which the currency is represented by its full name in plural format,
 * for example, "3.00 US dollars" or "1.00 US dollar".
 * <P>
 * You can also control the display of numbers with such methods as
 * \c getMinimumFractionDigits().  If you want even more control over the
 * format or parsing, or want to give your users more control, you can
 * try dynamic_casting the \c NumberFormat you get from the factory methods to a
 * \c DecimalFormat. This will work for the vast majority of
 * countries; just remember to test for NULL in case you
 * encounter an unusual one.
 * <P>
 * You can also use forms of the parse and format methods with
 * \c ParsePosition and \c FieldPosition to allow you to:
 * <ul type=round>
 *   <li>(a) progressively parse through pieces of a string.
 *   <li>(b) align the decimal point and other areas.
 * </ul>
 * For example, you can align numbers in two ways.
 * <P>
 * If you are using a monospaced font with spacing for alignment, you
 * can pass the \c FieldPosition in your format call, with field =
 * \c UNUM_INTEGER_FIELD. On output, \c getEndIndex will be set to the offset
 * between the last character of the integer and the decimal. Add
 * (desiredSpaceCount - getEndIndex) spaces at the front of the
 * string.
 * <P>
 * If you are using proportional fonts, instead of padding with
 * spaces, measure the width of the string in pixels from the start to
 * getEndIndex.  Then move the pen by (desiredPixelWidth -
 * widthToAlignmentPoint) before drawing the text.  It also works
 * where there is no decimal, but possibly additional characters at
 * the end, e.g. with parentheses in negative numbers: "(12)" for -12.
 * <p>
 * <em>User subclasses are not supported.</em> While clients may write
 * subclasses, such code will not necessarily work and will not be
 * guaranteed to work stably from release to release.
 *
 * @stable ICU 2.0
 */
class U_I18N_API NumberFormat : public Format {
public:
    /**
     * Rounding mode.
     *
     * <p>
     * For more detail on rounding modes, see:
     * http://userguide.icu-project.org/formatparse/numbers/rounding-modes
     *
     * @stable ICU 2.4
     */
    enum ERoundingMode {
        kRoundCeiling,  /**< Round towards positive infinity */
        kRoundFloor,    /**< Round towards negative infinity */
        kRoundDown,     /**< Round towards zero */
        kRoundUp,       /**< Round away from zero */
        kRoundHalfEven, /**< Round towards the nearest integer, or
                             towards the nearest even integer if equidistant */
        kRoundHalfDown, /**< Round towards the nearest integer, or
                             towards zero if equidistant */
        kRoundHalfUp,   /**< Round towards the nearest integer, or
                             away from zero if equidistant */
        /**
          *  Return U_FORMAT_INEXACT_ERROR if number does not format exactly.
          *  @stable ICU 4.8
          */
        kRoundUnnecessary
    };

    /**
     * Alignment Field constants used to construct a FieldPosition object.
     * Signifies that the position of the integer part or fraction part of
     * a formatted number should be returned.
     *
     * Note: as of ICU 4.4, the values in this enum have been extended to
     * support identification of all number format fields, not just those
     * pertaining to alignment.
     *
     * These constants are provided for backwards compatibility only.
     * Please use the C style constants defined in the header file unum.h.
     *
     * @see FieldPosition
     * @stable ICU 2.0
     */
    enum EAlignmentFields {
        /** @stable ICU 2.0 */
        kIntegerField = UNUM_INTEGER_FIELD,
        /** @stable ICU 2.0 */
        kFractionField = UNUM_FRACTION_FIELD,
        /** @stable ICU 2.0 */
        kDecimalSeparatorField = UNUM_DECIMAL_SEPARATOR_FIELD,
        /** @stable ICU 2.0 */
        kExponentSymbolField = UNUM_EXPONENT_SYMBOL_FIELD,
        /** @stable ICU 2.0 */
        kExponentSignField = UNUM_EXPONENT_SIGN_FIELD,
        /** @stable ICU 2.0 */
        kExponentField = UNUM_EXPONENT_FIELD,
        /** @stable ICU 2.0 */
        kGroupingSeparatorField = UNUM_GROUPING_SEPARATOR_FIELD,
        /** @stable ICU 2.0 */
        kCurrencyField = UNUM_CURRENCY_FIELD,
        /** @stable ICU 2.0 */
        kPercentField = UNUM_PERCENT_FIELD,
        /** @stable ICU 2.0 */
        kPermillField = UNUM_PERMILL_FIELD,
        /** @stable ICU 2.0 */
        kSignField = UNUM_SIGN_FIELD,
#ifndef U_HIDE_DRAFT_API
        /** @draft ICU 64 */
        kMeasureUnitField = UNUM_MEASURE_UNIT_FIELD,
        /** @draft ICU 64 */
        kCompactField = UNUM_COMPACT_FIELD,
#endif  // U_HIDE_DRAFT_API

    /**
     * These constants are provided for backwards compatibility only.
     * Please use the constants defined in the header file unum.h.
     */
        /** @stable ICU 2.0 */
        INTEGER_FIELD        = UNUM_INTEGER_FIELD,
        /** @stable ICU 2.0 */
        FRACTION_FIELD       = UNUM_FRACTION_FIELD
    };

    /**
     * Destructor.
     * @stable ICU 2.0
     */
    virtual ~NumberFormat();

    /**
     * Clones this object polymorphically.
     * The caller owns the result and should delete it when done.
     * @return clone, or nullptr if an error occurred
     * @stable ICU 2.0
     */
    virtual NumberFormat* clone() const = 0;

    /**
     * Return true if the given Format objects are semantically equal.
     * Objects of different subclasses are considered unequal.
     * @return    true if the given Format objects are semantically equal.
     * @stable ICU 2.0
     */
    virtual UBool operator==(const Format& other) const;


    using Format::format;

    /**
     * Format an object to produce a string.  This method handles
     * Formattable objects with numeric types. If the Formattable
     * object type is not a numeric type, then it returns a failing
     * UErrorCode.
     *
     * @param obj       The object to format.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    virtual UnicodeString& format(const Formattable& obj,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode& status) const;

    /**
     * Format an object to produce a string.  This method handles
     * Formattable objects with numeric types. If the Formattable
     * object type is not a numeric type, then it returns a failing
     * UErrorCode.
     *
     * @param obj       The object to format.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.  Can be
     *                  NULL.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    virtual UnicodeString& format(const Formattable& obj,
                                  UnicodeString& appendTo,
                                  FieldPositionIterator* posIter,
                                  UErrorCode& status) const;

    /**
     * Parse a string to produce an object.  This methods handles
     * parsing of numeric strings into Formattable objects with numeric
     * types.
     * <P>
     * Before calling, set parse_pos.index to the offset you want to
     * start parsing at in the source. After calling, parse_pos.index
     * indicates the position after the successfully parsed text.  If
     * an error occurs, parse_pos.index is unchanged.
     * <P>
     * When parsing, leading whitespace is discarded (with successful
     * parse), while trailing whitespace is left as is.
     * <P>
     * See Format::parseObject() for more.
     *
     * @param source    The string to be parsed into an object.
     * @param result    Formattable to be set to the parse result.
     *                  If parse fails, return contents are undefined.
     * @param parse_pos The position to start parsing at. Upon return
     *                  this param is set to the position after the
     *                  last character successfully parsed. If the
     *                  source is not parsed successfully, this param
     *                  will remain unchanged.
     * @return          A newly created Formattable* object, or NULL
     *                  on failure.  The caller owns this and should
     *                  delete it when done.
     * @stable ICU 2.0
     */
    virtual void parseObject(const UnicodeString& source,
                             Formattable& result,
                             ParsePosition& parse_pos) const;

    /**
     * Format a double number. These methods call the NumberFormat
     * pure virtual format() methods with the default FieldPosition.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    UnicodeString& format(  double number,
                            UnicodeString& appendTo) const;

    /**
     * Format a long number. These methods call the NumberFormat
     * pure virtual format() methods with the default FieldPosition.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    UnicodeString& format(  int32_t number,
                            UnicodeString& appendTo) const;

    /**
     * Format an int64 number. These methods call the NumberFormat
     * pure virtual format() methods with the default FieldPosition.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.8
     */
    UnicodeString& format(  int64_t number,
                            UnicodeString& appendTo) const;

    /**
     * Format a double number. Concrete subclasses must implement
     * these pure virtual methods.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    virtual UnicodeString& format(double number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos) const = 0;
    /**
     * Format a double number. By default, the parent function simply
     * calls the base class and does not return an error status.
     * Therefore, the status may be ignored in some subclasses.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    error status
     * @return          Reference to 'appendTo' parameter.
     * @internal
     */
    virtual UnicodeString& format(double number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode &status) const;
    /**
     * Format a double number. Subclasses must implement
     * this method.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     *                  Can be NULL.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    virtual UnicodeString& format(double number,
                                  UnicodeString& appendTo,
                                  FieldPositionIterator* posIter,
                                  UErrorCode& status) const;
    /**
     * Format a long number. Concrete subclasses must implement
     * these pure virtual methods.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
    */
    virtual UnicodeString& format(int32_t number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos) const = 0;

    /**
     * Format a long number. Concrete subclasses may override
     * this function to provide status return.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status the output status.
     * @return          Reference to 'appendTo' parameter.
     * @internal
    */
    virtual UnicodeString& format(int32_t number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode &status) const;

    /**
     * Format an int32 number. Subclasses must implement
     * this method.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     *                  Can be NULL.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    virtual UnicodeString& format(int32_t number,
                                  UnicodeString& appendTo,
                                  FieldPositionIterator* posIter,
                                  UErrorCode& status) const;
    /**
     * Format an int64 number. (Not abstract to retain compatibility
     * with earlier releases, however subclasses should override this
     * method as it just delegates to format(int32_t number...);
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.8
    */
    virtual UnicodeString& format(int64_t number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos) const;

    /**
     * Format an int64 number. (Not abstract to retain compatibility
     * with earlier releases, however subclasses should override this
     * method as it just delegates to format(int32_t number...);
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @internal
    */
    virtual UnicodeString& format(int64_t number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode& status) const;
    /**
     * Format an int64 number. Subclasses must implement
     * this method.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     *                  Can be NULL.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    virtual UnicodeString& format(int64_t number,
                                  UnicodeString& appendTo,
                                  FieldPositionIterator* posIter,
                                  UErrorCode& status) const;

    /**
     * Format a decimal number. Subclasses must implement
     * this method.  The syntax of the unformatted number is a "numeric string"
     * as defined in the Decimal Arithmetic Specification, available at
     * http://speleotrove.com/decimal
     *
     * @param number    The unformatted number, as a string, to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     *                  Can be NULL.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    virtual UnicodeString& format(StringPiece number,
                                  UnicodeString& appendTo,
                                  FieldPositionIterator* posIter,
                                  UErrorCode& status) const;

// Can't use #ifndef U_HIDE_INTERNAL_API because these are virtual methods

    /**
     * Format a decimal number.
     * The number is a DecimalQuantity wrapper onto a floating point decimal number.
     * The default implementation in NumberFormat converts the decimal number
     * to a double and formats that.  Subclasses of NumberFormat that want
     * to specifically handle big decimal numbers must override this method.
     * class DecimalFormat does so.
     *
     * @param number    The number, a DecimalQuantity format Decimal Floating Point.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @internal
     */
    virtual UnicodeString& format(const number::impl::DecimalQuantity &number,
                                  UnicodeString& appendTo,
                                  FieldPositionIterator* posIter,
                                  UErrorCode& status) const;

    /**
     * Format a decimal number.
     * The number is a DecimalQuantity wrapper onto a floating point decimal number.
     * The default implementation in NumberFormat converts the decimal number
     * to a double and formats that.  Subclasses of NumberFormat that want
     * to specifically handle big decimal numbers must override this method.
     * class DecimalFormat does so.
     *
     * @param number    The number, a DecimalQuantity format Decimal Floating Point.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @internal
     */
    virtual UnicodeString& format(const number::impl::DecimalQuantity &number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode& status) const;

   /**
    * Return a long if possible (e.g. within range LONG_MAX,
    * LONG_MAX], and with no decimals), otherwise a double.  If
    * IntegerOnly is set, will stop at a decimal point (or equivalent;
    * e.g. for rational numbers "1 2/3", will stop after the 1).
    * <P>
    * If no object can be parsed, index is unchanged, and NULL is
    * returned.
    * <P>
    * This is a pure virtual which concrete subclasses must implement.
    *
    * @param text           The text to be parsed.
    * @param result         Formattable to be set to the parse result.
    *                       If parse fails, return contents are undefined.
    * @param parsePosition  The position to start parsing at on input.
    *                       On output, moved to after the last successfully
    *                       parse character. On parse failure, does not change.
    * @stable ICU 2.0
    */
    virtual void parse(const UnicodeString& text,
                       Formattable& result,
                       ParsePosition& parsePosition) const = 0;

    /**
     * Parse a string as a numeric value, and return a Formattable
     * numeric object. This method parses integers only if IntegerOnly
     * is set.
     *
     * @param text          The text to be parsed.
     * @param result        Formattable to be set to the parse result.
     *                      If parse fails, return contents are undefined.
     * @param status        Output parameter set to a failure error code
     *                      when a failure occurs. The error code when the
     *                      string fails to parse is U_INVALID_FORMAT_ERROR,
     *                      unless overridden by a subclass.
     * @see                 NumberFormat::isParseIntegerOnly
     * @stable ICU 2.0
     */
    virtual void parse(const UnicodeString& text,
                       Formattable& result,
                       UErrorCode& status) const;

    /**
     * Parses text from the given string as a currency amount.  Unlike
     * the parse() method, this method will attempt to parse a generic
     * currency name, searching for a match of this object's locale's
     * currency display names, or for a 3-letter ISO currency code.
     * This method will fail if this format is not a currency format,
     * that is, if it does not contain the currency pattern symbol
     * (U+00A4) in its prefix or suffix.
     *
     * @param text the string to parse
     * @param pos  input-output position; on input, the position within text
     *             to match; must have 0 <= pos.getIndex() < text.length();
     *             on output, the position after the last matched character.
     *             If the parse fails, the position in unchanged upon output.
     * @return     if parse succeeds, a pointer to a newly-created CurrencyAmount
     *             object (owned by the caller) containing information about
     *             the parsed currency; if parse fails, this is NULL.
     * @stable ICU 49
     */
    virtual CurrencyAmount* parseCurrency(const UnicodeString& text,
                                          ParsePosition& pos) const;

    /**
     * Return true if this format will parse numbers as integers
     * only.  For example in the English locale, with ParseIntegerOnly
     * true, the string "1234." would be parsed as the integer value
     * 1234 and parsing would stop at the "." character.  Of course,
     * the exact format accepted by the parse operation is locale
     * dependant and determined by sub-classes of NumberFormat.
     * @return    true if this format will parse numbers as integers
     *            only.
     * @stable ICU 2.0
     */
    UBool isParseIntegerOnly(void) const;

    /**
     * Sets whether or not numbers should be parsed as integers only.
     * @param value    set True, this format will parse numbers as integers
     *                 only.
     * @see isParseIntegerOnly
     * @stable ICU 2.0
     */
    virtual void setParseIntegerOnly(UBool value);

    /**
     * Sets whether lenient parsing should be enabled (it is off by default).
     *
     * @param enable \c TRUE if lenient parsing should be used,
     *               \c FALSE otherwise.
     * @stable ICU 4.8
     */
    virtual void setLenient(UBool enable);

    /**
     * Returns whether lenient parsing is enabled (it is off by default).
     *
     * @return \c TRUE if lenient parsing is enabled,
     *         \c FALSE otherwise.
     * @see #setLenient
     * @stable ICU 4.8
     */
    virtual UBool isLenient(void) const;

    /**
     * Create a default style NumberFormat for the current default locale.
     * The default formatting style is locale dependent.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @stable ICU 2.0
     */
    static NumberFormat* U_EXPORT2 createInstance(UErrorCode&);

    /**
     * Create a default style NumberFormat for the specified locale.
     * The default formatting style is locale dependent.
     * @param inLocale    the given locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @stable ICU 2.0
     */
    static NumberFormat* U_EXPORT2 createInstance(const Locale& inLocale,
                                        UErrorCode&);

    /**
     * Create a specific style NumberFormat for the specified locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @param desiredLocale    the given locale.
     * @param style            the given style.
     * @param errorCode        Output param filled with success/failure status.
     * @return                 A new NumberFormat instance.
     * @stable ICU 4.8
     */
    static NumberFormat* U_EXPORT2 createInstance(const Locale& desiredLocale,
                                                  UNumberFormatStyle style,
                                                  UErrorCode& errorCode);

#ifndef U_HIDE_INTERNAL_API

    /**
     * ICU use only.
     * Creates NumberFormat instance without using the cache.
     * @internal
     */
    static NumberFormat* internalCreateInstance(
            const Locale& desiredLocale,
            UNumberFormatStyle style,
            UErrorCode& errorCode);

    /**
     * ICU use only.
     * Returns handle to the shared, cached NumberFormat instance for given
     * locale. On success, caller must call removeRef() on returned value
     * once it is done with the shared instance.
     * @internal
     */
    static const SharedNumberFormat* U_EXPORT2 createSharedInstance(
            const Locale& inLocale, UNumberFormatStyle style, UErrorCode& status);

#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Returns a currency format for the current default locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @stable ICU 2.0
     */
    static NumberFormat* U_EXPORT2 createCurrencyInstance(UErrorCode&);

    /**
     * Returns a currency format for the specified locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @param inLocale    the given locale.
     * @stable ICU 2.0
     */
    static NumberFormat* U_EXPORT2 createCurrencyInstance(const Locale& inLocale,
                                                UErrorCode&);

    /**
     * Returns a percentage format for the current default locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @stable ICU 2.0
     */
    static NumberFormat* U_EXPORT2 createPercentInstance(UErrorCode&);

    /**
     * Returns a percentage format for the specified locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @param inLocale    the given locale.
     * @stable ICU 2.0
     */
    static NumberFormat* U_EXPORT2 createPercentInstance(const Locale& inLocale,
                                               UErrorCode&);

    /**
     * Returns a scientific format for the current default locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @stable ICU 2.0
     */
    static NumberFormat* U_EXPORT2 createScientificInstance(UErrorCode&);

    /**
     * Returns a scientific format for the specified locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @param inLocale    the given locale.
     * @stable ICU 2.0
     */
    static NumberFormat* U_EXPORT2 createScientificInstance(const Locale& inLocale,
                                                UErrorCode&);

    /**
     * Get the set of Locales for which NumberFormats are installed.
     * @param count    Output param to receive the size of the locales
     * @stable ICU 2.0
     */
    static const Locale* U_EXPORT2 getAvailableLocales(int32_t& count);

#if !UCONFIG_NO_SERVICE
    /**
     * Register a new NumberFormatFactory.  The factory will be adopted.
     * Because ICU may choose to cache NumberFormat objects internally,
     * this must be called at application startup, prior to any calls to
     * NumberFormat::createInstance to avoid undefined behavior.
     * @param toAdopt the NumberFormatFactory instance to be adopted
     * @param status the in/out status code, no special meanings are assigned
     * @return a registry key that can be used to unregister this factory
     * @stable ICU 2.6
     */
    static URegistryKey U_EXPORT2 registerFactory(NumberFormatFactory* toAdopt, UErrorCode& status);

    /**
     * Unregister a previously-registered NumberFormatFactory using the key returned from the
     * register call.  Key becomes invalid after a successful call and should not be used again.
     * The NumberFormatFactory corresponding to the key will be deleted.
     * Because ICU may choose to cache NumberFormat objects internally,
     * this should be called during application shutdown, after all calls to
     * NumberFormat::createInstance to avoid undefined behavior.
     * @param key the registry key returned by a previous call to registerFactory
     * @param status the in/out status code, no special meanings are assigned
     * @return TRUE if the factory for the key was successfully unregistered
     * @stable ICU 2.6
     */
    static UBool U_EXPORT2 unregister(URegistryKey key, UErrorCode& status);

    /**
     * Return a StringEnumeration over the locales available at the time of the call,
     * including registered locales.
     * @return a StringEnumeration over the locales available at the time of the call
     * @stable ICU 2.6
     */
    static StringEnumeration* U_EXPORT2 getAvailableLocales(void);
#endif /* UCONFIG_NO_SERVICE */

    /**
     * Returns true if grouping is used in this format. For example,
     * in the English locale, with grouping on, the number 1234567
     * might be formatted as "1,234,567". The grouping separator as
     * well as the size of each group is locale dependent and is
     * determined by sub-classes of NumberFormat.
     * @see setGroupingUsed
     * @stable ICU 2.0
     */
    UBool isGroupingUsed(void) const;

    /**
     * Set whether or not grouping will be used in this format.
     * @param newValue    True, grouping will be used in this format.
     * @see getGroupingUsed
     * @stable ICU 2.0
     */
    virtual void setGroupingUsed(UBool newValue);

    /**
     * Returns the maximum number of digits allowed in the integer portion of a
     * number.
     * @return     the maximum number of digits allowed in the integer portion of a
     *             number.
     * @see setMaximumIntegerDigits
     * @stable ICU 2.0
     */
    int32_t getMaximumIntegerDigits(void) const;

    /**
     * Sets the maximum number of digits allowed in the integer portion of a
     * number. maximumIntegerDigits must be >= minimumIntegerDigits.  If the
     * new value for maximumIntegerDigits is less than the current value
     * of minimumIntegerDigits, then minimumIntegerDigits will also be set to
     * the new value.
     *
     * @param newValue    the new value for the maximum number of digits
     *                    allowed in the integer portion of a number.
     * @see getMaximumIntegerDigits
     * @stable ICU 2.0
     */
    virtual void setMaximumIntegerDigits(int32_t newValue);

    /**
     * Returns the minimum number of digits allowed in the integer portion of a
     * number.
     * @return    the minimum number of digits allowed in the integer portion of a
     *            number.
     * @see setMinimumIntegerDigits
     * @stable ICU 2.0
     */
    int32_t getMinimumIntegerDigits(void) const;

    /**
     * Sets the minimum number of digits allowed in the integer portion of a
     * number. minimumIntegerDigits must be &lt;= maximumIntegerDigits.  If the
     * new value for minimumIntegerDigits exceeds the current value
     * of maximumIntegerDigits, then maximumIntegerDigits will also be set to
     * the new value.
     * @param newValue    the new value to be set.
     * @see getMinimumIntegerDigits
     * @stable ICU 2.0
     */
    virtual void setMinimumIntegerDigits(int32_t newValue);

    /**
     * Returns the maximum number of digits allowed in the fraction portion of a
     * number.
     * @return    the maximum number of digits allowed in the fraction portion of a
     *            number.
     * @see setMaximumFractionDigits
     * @stable ICU 2.0
     */
    int32_t getMaximumFractionDigits(void) const;

    /**
     * Sets the maximum number of digits allowed in the fraction portion of a
     * number. maximumFractionDigits must be >= minimumFractionDigits.  If the
     * new value for maximumFractionDigits is less than the current value
     * of minimumFractionDigits, then minimumFractionDigits will also be set to
     * the new value.
     * @param newValue    the new value to be set.
     * @see getMaximumFractionDigits
     * @stable ICU 2.0
     */
    virtual void setMaximumFractionDigits(int32_t newValue);

    /**
     * Returns the minimum number of digits allowed in the fraction portion of a
     * number.
     * @return    the minimum number of digits allowed in the fraction portion of a
     *            number.
     * @see setMinimumFractionDigits
     * @stable ICU 2.0
     */
    int32_t getMinimumFractionDigits(void) const;

    /**
     * Sets the minimum number of digits allowed in the fraction portion of a
     * number. minimumFractionDigits must be &lt;= maximumFractionDigits.   If the
     * new value for minimumFractionDigits exceeds the current value
     * of maximumFractionDigits, then maximumIntegerDigits will also be set to
     * the new value
     * @param newValue    the new value to be set.
     * @see getMinimumFractionDigits
     * @stable ICU 2.0
     */
    virtual void setMinimumFractionDigits(int32_t newValue);

    /**
     * Sets the currency used to display currency
     * amounts.  This takes effect immediately, if this format is a
     * currency format.  If this format is not a currency format, then
     * the currency is used if and when this object becomes a
     * currency format.
     * @param theCurrency a 3-letter ISO code indicating new currency
     * to use.  It need not be null-terminated.  May be the empty
     * string or NULL to indicate no currency.
     * @param ec input-output error code
     * @stable ICU 3.0
     */
    virtual void setCurrency(const char16_t* theCurrency, UErrorCode& ec);

    /**
     * Gets the currency used to display currency
     * amounts.  This may be an empty string for some subclasses.
     * @return a 3-letter null-terminated ISO code indicating
     * the currency in use, or a pointer to the empty string.
     * @stable ICU 2.6
     */
    const char16_t* getCurrency() const;

    /**
     * Set a particular UDisplayContext value in the formatter, such as
     * UDISPCTX_CAPITALIZATION_FOR_STANDALONE.
     * @param value The UDisplayContext value to set.
     * @param status Input/output status. If at entry this indicates a failure
     *               status, the function will do nothing; otherwise this will be
     *               updated with any new status from the function.
     * @stable ICU 53
     */
    virtual void setContext(UDisplayContext value, UErrorCode& status);

    /**
     * Get the formatter's UDisplayContext value for the specified UDisplayContextType,
     * such as UDISPCTX_TYPE_CAPITALIZATION.
     * @param type The UDisplayContextType whose value to return
     * @param status Input/output status. If at entry this indicates a failure
     *               status, the function will do nothing; otherwise this will be
     *               updated with any new status from the function.
     * @return The UDisplayContextValue for the specified type.
     * @stable ICU 53
     */
    virtual UDisplayContext getContext(UDisplayContextType type, UErrorCode& status) const;

    /**
     * Get the rounding mode. This will always return NumberFormat::ERoundingMode::kRoundUnnecessary
     * if the subclass does not support rounding.
     * @return A rounding mode
     * @stable ICU 60
     */
    virtual ERoundingMode getRoundingMode(void) const;

    /**
     * Set the rounding mode. If a subclass does not support rounding, this will do nothing.
     * @param roundingMode A rounding mode
     * @stable ICU 60
     */
    virtual void setRoundingMode(ERoundingMode roundingMode);

public:

    /**
     * Return the class ID for this class.  This is useful for
     * comparing to a return value from getDynamicClassID(). Note that,
     * because NumberFormat is an abstract base class, no fully constructed object
     * will have the class ID returned by NumberFormat::getStaticClassID().
     * @return The class ID for all objects of this class.
     * @stable ICU 2.0
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY.  Pure virtual override.
     * This method is to implement a simple version of RTTI, since not all
     * C++ compilers support genuine RTTI.  Polymorphic operator==() and
     * clone() methods call this method.
     * <P>
     * @return The class ID for this object. All objects of a
     * given class have the same class ID.  Objects of
     * other classes have different class IDs.
     * @stable ICU 2.0
     */
    virtual UClassID getDynamicClassID(void) const = 0;

protected:

    /**
     * Default constructor for subclass use only.
     * @stable ICU 2.0
     */
    NumberFormat();

    /**
     * Copy constructor.
     * @stable ICU 2.0
     */
    NumberFormat(const NumberFormat&);

    /**
     * Assignment operator.
     * @stable ICU 2.0
     */
    NumberFormat& operator=(const NumberFormat&);

    /**
     * Returns the currency in effect for this formatter.  Subclasses
     * should override this method as needed.  Unlike getCurrency(),
     * this method should never return "".
     * @result output parameter for null-terminated result, which must
     * have a capacity of at least 4
     * @internal
     */
    virtual void getEffectiveCurrency(char16_t* result, UErrorCode& ec) const;

#ifndef U_HIDE_INTERNAL_API
    /**
     * Creates the specified number format style of the desired locale.
     * If mustBeDecimalFormat is TRUE, then the returned pointer is
     * either a DecimalFormat or it is NULL.
     * @internal
     */
    static NumberFormat* makeInstance(const Locale& desiredLocale,
                                      UNumberFormatStyle style,
                                      UBool mustBeDecimalFormat,
                                      UErrorCode& errorCode);
#endif  /* U_HIDE_INTERNAL_API */

private:

    static UBool isStyleSupported(UNumberFormatStyle style);

    /**
     * Creates the specified decimal format style of the desired locale.
     * @param desiredLocale    the given locale.
     * @param style            the given style.
     * @param errorCode        Output param filled with success/failure status.
     * @return                 A new NumberFormat instance.
     */
    static NumberFormat* makeInstance(const Locale& desiredLocale,
                                      UNumberFormatStyle style,
                                      UErrorCode& errorCode);

    UBool       fGroupingUsed;
    int32_t     fMaxIntegerDigits;
    int32_t     fMinIntegerDigits;
    int32_t     fMaxFractionDigits;
    int32_t     fMinFractionDigits;

  protected:
    /** \internal */
    static const int32_t gDefaultMaxIntegerDigits;
    /** \internal */
    static const int32_t gDefaultMinIntegerDigits;

  private:
    UBool      fParseIntegerOnly;
    UBool      fLenient; // TRUE => lenient parse is enabled

    // ISO currency code
    char16_t      fCurrency[4];

    UDisplayContext fCapitalizationContext;

    friend class ICUNumberFormatFactory; // access to makeInstance
    friend class ICUNumberFormatService;
    friend class ::NumberFormatTest;  // access to isStyleSupported()
};

#if !UCONFIG_NO_SERVICE
/**
 * A NumberFormatFactory is used to register new number formats.  The factory
 * should be able to create any of the predefined formats for each locale it
 * supports.  When registered, the locales it supports extend or override the
 * locale already supported by ICU.
 *
 * @stable ICU 2.6
 */
class U_I18N_API NumberFormatFactory : public UObject {
public:

    /**
     * Destructor
     * @stable ICU 3.0
     */
    virtual ~NumberFormatFactory();

    /**
     * Return true if this factory will be visible.  Default is true.
     * If not visible, the locales supported by this factory will not
     * be listed by getAvailableLocales.
     * @stable ICU 2.6
     */
    virtual UBool visible(void) const = 0;

    /**
     * Return the locale names directly supported by this factory.  The number of names
     * is returned in count;
     * @stable ICU 2.6
     */
    virtual const UnicodeString * getSupportedIDs(int32_t &count, UErrorCode& status) const = 0;

    /**
     * Return a number format of the appropriate type.  If the locale
     * is not supported, return null.  If the locale is supported, but
     * the type is not provided by this service, return null.  Otherwise
     * return an appropriate instance of NumberFormat.
     * @stable ICU 2.6
     */
    virtual NumberFormat* createFormat(const Locale& loc, UNumberFormatStyle formatType) = 0;
};

/**
 * A NumberFormatFactory that supports a single locale.  It can be visible or invisible.
 * @stable ICU 2.6
 */
class U_I18N_API SimpleNumberFormatFactory : public NumberFormatFactory {
protected:
    /**
     * True if the locale supported by this factory is visible.
     * @stable ICU 2.6
     */
    const UBool _visible;

    /**
     * The locale supported by this factory, as a UnicodeString.
     * @stable ICU 2.6
     */
    UnicodeString _id;

public:
    /**
     * @stable ICU 2.6
     */
    SimpleNumberFormatFactory(const Locale& locale, UBool visible = TRUE);

    /**
     * @stable ICU 3.0
     */
    virtual ~SimpleNumberFormatFactory();

    /**
     * @stable ICU 2.6
     */
    virtual UBool visible(void) const;

    /**
     * @stable ICU 2.6
     */
    virtual const UnicodeString * getSupportedIDs(int32_t &count, UErrorCode& status) const;
};
#endif /* #if !UCONFIG_NO_SERVICE */

// -------------------------------------

inline UBool
NumberFormat::isParseIntegerOnly() const
{
    return fParseIntegerOnly;
}

inline UBool
NumberFormat::isLenient() const
{
    return fLenient;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */




#ifndef U_HIDE_DEPRECATED_API

U_NAMESPACE_BEGIN

class MessageFormat;

/**
 * ChoiceFormat converts between ranges of numeric values and strings for those ranges.
 * The strings must conform to the MessageFormat pattern syntax.
 *
 * <p><em><code>ChoiceFormat</code> is probably not what you need.
 * Please use <code>MessageFormat</code>
 * with <code>plural</code> arguments for proper plural selection,
 * and <code>select</code> arguments for simple selection among a fixed set of choices!</em></p>
 *
 * <p>A <code>ChoiceFormat</code> splits
 * the real number line \htmlonly<code>-&#x221E;</code> to
 * <code>+&#x221E;</code>\endhtmlonly into two
 * or more contiguous ranges. Each range is mapped to a
 * string.</p>
 *
 * <p><code>ChoiceFormat</code> was originally intended
 * for displaying grammatically correct
 * plurals such as &quot;There is one file.&quot; vs. &quot;There are 2 files.&quot;
 * <em>However,</em> plural rules for many languages
 * are too complex for the capabilities of ChoiceFormat,
 * and its requirement of specifying the precise rules for each message
 * is unmanageable for translators.</p>
 *
 * <p>There are two methods of defining a <code>ChoiceFormat</code>; both
 * are equivalent.  The first is by using a string pattern. This is the
 * preferred method in most cases.  The second method is through direct
 * specification of the arrays that logically make up the
 * <code>ChoiceFormat</code>.</p>
 *
 * <p>Note: Typically, choice formatting is done (if done at all) via <code>MessageFormat</code>
 * with a <code>choice</code> argument type,
 * rather than using a stand-alone <code>ChoiceFormat</code>.</p>
 *
 * <h5>Patterns and Their Interpretation</h5>
 *
 * <p>The pattern string defines the range boundaries and the strings for each number range.
 * Syntax:
 * <pre>
 * choiceStyle = number separator message ('|' number separator message)*
 * number = normal_number | ['-'] \htmlonly&#x221E;\endhtmlonly (U+221E, infinity)
 * normal_number = double value (unlocalized ASCII string)
 * separator = less_than | less_than_or_equal
 * less_than = '<'
 * less_than_or_equal = '#' | \htmlonly&#x2264;\endhtmlonly (U+2264)
 * message: see {@link MessageFormat}
 * </pre>
 * Pattern_White_Space between syntax elements is ignored, except
 * around each range's sub-message.</p>
 *
 * <p>Each numeric sub-range extends from the current range's number
 * to the next range's number.
 * The number itself is included in its range if a <code>less_than_or_equal</code> sign is used,
 * and excluded from its range (and instead included in the previous range)
 * if a <code>less_than</code> sign is used.</p>
 *
 * <p>When a <code>ChoiceFormat</code> is constructed from
 * arrays of numbers, closure flags and strings,
 * they are interpreted just like
 * the sequence of <code>(number separator string)</code> in an equivalent pattern string.
 * <code>closure[i]==TRUE</code> corresponds to a <code>less_than</code> separator sign.
 * The equivalent pattern string will be constructed automatically.</p>
 *
 * <p>During formatting, a number is mapped to the first range
 * where the number is not greater than the range's upper limit.
 * That range's message string is returned. A NaN maps to the very first range.</p>
 *
 * <p>During parsing, a range is selected for the longest match of
 * any range's message. That range's number is returned, ignoring the separator/closure.
 * Only a simple string match is performed, without parsing of arguments that
 * might be specified in the message strings.</p>
 *
 * <p>Note that the first range's number is ignored in formatting
 * but may be returned from parsing.</p>
 *
 * <h5>Examples</h5>
 *
 * <p>Here is an example of two arrays that map the number
 * <code>1..7</code> to the English day of the week abbreviations
 * <code>Sun..Sat</code>. No closures array is given; this is the same as
 * specifying all closures to be <code>FALSE</code>.</p>
 *
 * <pre>    {1,2,3,4,5,6,7},
 *     {&quot;Sun&quot;,&quot;Mon&quot;,&quot;Tue&quot;,&quot;Wed&quot;,&quot;Thur&quot;,&quot;Fri&quot;,&quot;Sat&quot;}</pre>
 *
 * <p>Here is an example that maps the ranges [-Inf, 1), [1, 1], and (1,
 * +Inf] to three strings. That is, the number line is split into three
 * ranges: x &lt; 1.0, x = 1.0, and x &gt; 1.0.
 * (The round parentheses in the notation above indicate an exclusive boundary,
 * like the turned bracket in European notation: [-Inf, 1) == [-Inf, 1[  )</p>
 *
 * <pre>    {0, 1, 1},
 *     {FALSE, FALSE, TRUE},
 *     {&quot;no files&quot;, &quot;one file&quot;, &quot;many files&quot;}</pre>
 *
 * <p>Here is an example that shows formatting and parsing: </p>
 *
 * \code
 *   #include <unicode/choicfmt.h>
 *   #include <unicode/unistr.h>
 *   #include <iostream.h>
 *
 *   int main(int argc, char *argv[]) {
 *       double limits[] = {1,2,3,4,5,6,7};
 *       UnicodeString monthNames[] = {
 *           "Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
 *       ChoiceFormat fmt(limits, monthNames, 7);
 *       UnicodeString str;
 *       char buf[256];
 *       for (double x = 1.0; x <= 8.0; x += 1.0) {
 *           fmt.format(x, str);
 *           str.extract(0, str.length(), buf, 256, "");
 *           str.truncate(0);
 *           cout << x << " -> "
 *                << buf << endl;
 *       }
 *       cout << endl;
 *       return 0;
 *   }
 * \endcode
 *
 * <p><em>User subclasses are not supported.</em> While clients may write
 * subclasses, such code will not necessarily work and will not be
 * guaranteed to work stably from release to release.
 *
 * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
 */
class U_I18N_API ChoiceFormat: public NumberFormat {
public:
    /**
     * Constructs a new ChoiceFormat from the pattern string.
     *
     * @param pattern   Pattern used to construct object.
     * @param status    Output param to receive success code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    ChoiceFormat(const UnicodeString& pattern,
                 UErrorCode& status);


    /**
     * Constructs a new ChoiceFormat with the given limits and message strings.
     * All closure flags default to <code>FALSE</code>,
     * equivalent to <code>less_than_or_equal</code> separators.
     *
     * Copies the limits and formats instead of adopting them.
     *
     * @param limits    Array of limit values.
     * @param formats   Array of formats.
     * @param count     Size of 'limits' and 'formats' arrays.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    ChoiceFormat(const double* limits,
                 const UnicodeString* formats,
                 int32_t count );

    /**
     * Constructs a new ChoiceFormat with the given limits, closure flags and message strings.
     *
     * Copies the limits and formats instead of adopting them.
     *
     * @param limits Array of limit values
     * @param closures Array of booleans specifying whether each
     * element of 'limits' is open or closed.  If FALSE, then the
     * corresponding limit number is a member of its range.
     * If TRUE, then the limit number belongs to the previous range it.
     * @param formats Array of formats
     * @param count Size of 'limits', 'closures', and 'formats' arrays
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    ChoiceFormat(const double* limits,
                 const UBool* closures,
                 const UnicodeString* formats,
                 int32_t count);

    /**
     * Copy constructor.
     *
     * @param that   ChoiceFormat object to be copied from
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    ChoiceFormat(const ChoiceFormat& that);

    /**
     * Assignment operator.
     *
     * @param that   ChoiceFormat object to be copied
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    const ChoiceFormat& operator=(const ChoiceFormat& that);

    /**
     * Destructor.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual ~ChoiceFormat();

    /**
     * Clones this Format object. The caller owns the
     * result and must delete it when done.
     *
     * @return a copy of this object
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual ChoiceFormat* clone() const;

    /**
     * Returns true if the given Format objects are semantically equal.
     * Objects of different subclasses are considered unequal.
     *
     * @param other    ChoiceFormat object to be compared
     * @return         true if other is the same as this.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual UBool operator==(const Format& other) const;

    /**
     * Sets the pattern.
     * @param pattern   The pattern to be applied.
     * @param status    Output param set to success/failure code on
     *                  exit. If the pattern is invalid, this will be
     *                  set to a failure result.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual void applyPattern(const UnicodeString& pattern,
                              UErrorCode& status);

    /**
     * Sets the pattern.
     * @param pattern    The pattern to be applied.
     * @param parseError Struct to receive information on position
     *                   of error if an error is encountered
     * @param status     Output param set to success/failure code on
     *                   exit. If the pattern is invalid, this will be
     *                   set to a failure result.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual void applyPattern(const UnicodeString& pattern,
                             UParseError& parseError,
                             UErrorCode& status);
    /**
     * Gets the pattern.
     *
     * @param pattern    Output param which will receive the pattern
     *                   Previous contents are deleted.
     * @return    A reference to 'pattern'
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual UnicodeString& toPattern(UnicodeString &pattern) const;

    /**
     * Sets the choices to be used in formatting.
     * For details see the constructor with the same parameter list.
     *
     * @param limitsToCopy      Contains the top value that you want
     *                          parsed with that format,and should be in
     *                          ascending sorted order. When formatting X,
     *                          the choice will be the i, where limit[i]
     *                          &lt;= X &lt; limit[i+1].
     * @param formatsToCopy     The format strings you want to use for each limit.
     * @param count             The size of the above arrays.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual void setChoices(const double* limitsToCopy,
                            const UnicodeString* formatsToCopy,
                            int32_t count );

    /**
     * Sets the choices to be used in formatting.
     * For details see the constructor with the same parameter list.
     *
     * @param limits Array of limits
     * @param closures Array of limit booleans
     * @param formats Array of format string
     * @param count The size of the above arrays
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual void setChoices(const double* limits,
                            const UBool* closures,
                            const UnicodeString* formats,
                            int32_t count);

    /**
     * Returns NULL and 0.
     * Before ICU 4.8, this used to return the choice limits array.
     *
     * @param count Will be set to 0.
     * @return NULL
     * @deprecated ICU 4.8 Use the MessagePattern class to analyze a ChoiceFormat pattern.
     */
    virtual const double* getLimits(int32_t& count) const;

    /**
     * Returns NULL and 0.
     * Before ICU 4.8, this used to return the limit booleans array.
     *
     * @param count Will be set to 0.
     * @return NULL
     * @deprecated ICU 4.8 Use the MessagePattern class to analyze a ChoiceFormat pattern.
     */
    virtual const UBool* getClosures(int32_t& count) const;

    /**
     * Returns NULL and 0.
     * Before ICU 4.8, this used to return the array of choice strings.
     *
     * @param count Will be set to 0.
     * @return NULL
     * @deprecated ICU 4.8 Use the MessagePattern class to analyze a ChoiceFormat pattern.
     */
    virtual const UnicodeString* getFormats(int32_t& count) const;


    using NumberFormat::format;

    /**
     * Formats a double number using this object's choices.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual UnicodeString& format(double number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos) const;
    /**
     * Formats an int32_t number using this object's choices.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual UnicodeString& format(int32_t number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos) const;

    /**
     * Formats an int64_t number using this object's choices.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual UnicodeString& format(int64_t number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos) const;

    /**
     * Formats an array of objects using this object's choices.
     *
     * @param objs      The array of objects to be formatted.
     * @param cnt       The size of objs.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param success   Output param set to success/failure code on
     *                  exit.
     * @return          Reference to 'appendTo' parameter.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual UnicodeString& format(const Formattable* objs,
                                  int32_t cnt,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode& success) const;

   using NumberFormat::parse;

   /**
    * Looks for the longest match of any message string on the input text and,
    * if there is a match, sets the result object to the corresponding range's number.
    *
    * If no string matches, then the parsePosition is unchanged.
    *
    * @param text           The text to be parsed.
    * @param result         Formattable to be set to the parse result.
    *                       If parse fails, return contents are undefined.
    * @param parsePosition  The position to start parsing at on input.
    *                       On output, moved to after the last successfully
    *                       parse character. On parse failure, does not change.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
    */
    virtual void parse(const UnicodeString& text,
                       Formattable& result,
                       ParsePosition& parsePosition) const;

    /**
     * Returns a unique class ID POLYMORPHICALLY. Part of ICU's "poor man's RTTI".
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Returns the class ID for this class.  This is useful only for
     * comparing to a return value from getDynamicClassID().  For example:
     * <pre>
     * .       Base* polymorphic_pointer = createPolymorphicObject();
     * .       if (polymorphic_pointer->getDynamicClassID() ==
     * .           Derived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @deprecated ICU 49 Use MessageFormat instead, with plural and select arguments.
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

private:
    /**
     * Converts a double value to a string.
     * @param value the double number to be converted.
     * @param string the result string.
     * @return the converted string.
     */
    static UnicodeString& dtos(double value, UnicodeString& string);

    ChoiceFormat(); // default constructor not implemented

    /**
     * Construct a new ChoiceFormat with the limits and the corresponding formats
     * based on the pattern.
     *
     * @param newPattern   Pattern used to construct object.
     * @param parseError   Struct to receive information on position
     *                     of error if an error is encountered.
     * @param status       Output param to receive success code.  If the
     *                     pattern cannot be parsed, set to failure code.
     */
    ChoiceFormat(const UnicodeString& newPattern,
                 UParseError& parseError,
                 UErrorCode& status);

    friend class MessageFormat;

    virtual void setChoices(const double* limits,
                            const UBool* closures,
                            const UnicodeString* formats,
                            int32_t count,
                            UErrorCode &errorCode);

    /**
     * Finds the ChoiceFormat sub-message for the given number.
     * @param pattern A MessagePattern.
     * @param partIndex the index of the first ChoiceFormat argument style part.
     * @param number a number to be mapped to one of the ChoiceFormat argument's intervals
     * @return the sub-message start part index.
     */
    static int32_t findSubMessage(const MessagePattern &pattern, int32_t partIndex, double number);

    static double parseArgument(
            const MessagePattern &pattern, int32_t partIndex,
            const UnicodeString &source, ParsePosition &pos);

    /**
     * Matches the pattern string from the end of the partIndex to
     * the beginning of the limitPartIndex,
     * including all syntax except SKIP_SYNTAX,
     * against the source string starting at sourceOffset.
     * If they match, returns the length of the source string match.
     * Otherwise returns -1.
     */
    static int32_t matchStringUntilLimitPart(
            const MessagePattern &pattern, int32_t partIndex, int32_t limitPartIndex,
            const UnicodeString &source, int32_t sourceOffset);

    /**
     * Some of the ChoiceFormat constructors do not have a UErrorCode paramater.
     * We need _some_ way to provide one for the MessagePattern constructor.
     * Alternatively, the MessagePattern could be a pointer field, but that is
     * not nice either.
     */
    UErrorCode constructorErrorCode;

    /**
     * The MessagePattern which contains the parsed structure of the pattern string.
     *
     * Starting with ICU 4.8, the MessagePattern contains a sequence of
     * numeric/selector/message parts corresponding to the parsed pattern.
     * For details see the MessagePattern class API docs.
     */
    MessagePattern msgPattern;

    /**
     * Docs & fields from before ICU 4.8, before MessagePattern was used.
     * Commented out, and left only for explanation of semantics.
     * --------
     * Each ChoiceFormat divides the range -Inf..+Inf into fCount
     * intervals.  The intervals are:
     *
     *         0: fChoiceLimits[0]..fChoiceLimits[1]
     *         1: fChoiceLimits[1]..fChoiceLimits[2]
     *        ...
     *  fCount-2: fChoiceLimits[fCount-2]..fChoiceLimits[fCount-1]
     *  fCount-1: fChoiceLimits[fCount-1]..+Inf
     *
     * Interval 0 is special; during formatting (mapping numbers to
     * strings), it also contains all numbers less than
     * fChoiceLimits[0], as well as NaN values.
     *
     * Interval i maps to and from string fChoiceFormats[i].  When
     * parsing (mapping strings to numbers), then intervals map to
     * their lower limit, that is, interval i maps to fChoiceLimit[i].
     *
     * The intervals may be closed, half open, or open.  This affects
     * formatting but does not affect parsing.  Interval i is affected
     * by fClosures[i] and fClosures[i+1].  If fClosures[i]
     * is FALSE, then the value fChoiceLimits[i] is in interval i.
     * That is, intervals i and i are:
     *
     *  i-1:                 ... x < fChoiceLimits[i]
     *    i: fChoiceLimits[i] <= x ...
     *
     * If fClosures[i] is TRUE, then the value fChoiceLimits[i] is
     * in interval i-1.  That is, intervals i-1 and i are:
     *
     *  i-1:                ... x <= fChoiceLimits[i]
     *    i: fChoiceLimits[i] < x ...
     *
     * Because of the nature of interval 0, fClosures[0] has no
     * effect.
     */
    // double*         fChoiceLimits;
    // UBool*          fClosures;
    // UnicodeString*  fChoiceFormats;
    // int32_t         fCount;
};


U_NAMESPACE_END

#endif  // U_HIDE_DEPRECATED_API
#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */









#include <stdio.h>
#include <float.h>

// *****************************************************************************
// class ChoiceFormat
// *****************************************************************************

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(ChoiceFormat)

// Special characters used by ChoiceFormat.  There are two characters
// used interchangeably to indicate <=.  Either is parsed, but only
// CHOICE_FORMAT_LESS_EQUAL is generated by toPattern().
#define CHOICE_FORMAT_SINGLE_QUOTE ((UChar)0x0027)   /*'*/
#define CHOICE_FORMAT_LESS_THAN    ((UChar)0x003C)   /*<*/
#define CHOICE_FORMAT_LESS_EQUAL   ((UChar)0x0023)   /*#*/
#define CHOICE_FORMAT_LESS_EQUAL2  ((UChar)0x2264)
#define CHOICE_FORMAT_VERTICAL_BAR ((UChar)0x007C)   /*|*/
#define CHOICE_FORMAT_MINUS        ((UChar)0x002D)   /*-*/

static const UChar LEFT_CURLY_BRACE = 0x7B;     /*{*/
static const UChar RIGHT_CURLY_BRACE = 0x7D;    /*}*/

#ifdef CHOICE_FORMAT_INFINITY
#undef CHOICE_FORMAT_INFINITY
#endif
#define CHOICE_FORMAT_INFINITY     ((UChar)0x221E)

//static const UChar gPositiveInfinity[] = {CHOICE_FORMAT_INFINITY, 0};
//static const UChar gNegativeInfinity[] = {CHOICE_FORMAT_MINUS, CHOICE_FORMAT_INFINITY, 0};
#define CHOICE_FORMAT_POSITIVE_INF_STRLEN 1
#define CHOICE_FORMAT_NEGATIVE_INF_STRLEN 2

// -------------------------------------
// Creates a ChoiceFormat instance based on the pattern.

ChoiceFormat::ChoiceFormat(const UnicodeString& newPattern,
                           UErrorCode& status)
: constructorErrorCode(status),
  msgPattern(status)
{
    applyPattern(newPattern, status);
}

// -------------------------------------
// Creates a ChoiceFormat instance with the limit array and
// format strings for each limit.

ChoiceFormat::ChoiceFormat(const double* limits,
                           const UnicodeString* formats,
                           int32_t cnt )
: constructorErrorCode(U_ZERO_ERROR),
  msgPattern(constructorErrorCode)
{
    setChoices(limits, NULL, formats, cnt, constructorErrorCode);
}

// -------------------------------------

ChoiceFormat::ChoiceFormat(const double* limits,
                           const UBool* closures,
                           const UnicodeString* formats,
                           int32_t cnt )
: constructorErrorCode(U_ZERO_ERROR),
  msgPattern(constructorErrorCode)
{
    setChoices(limits, closures, formats, cnt, constructorErrorCode);
}

// -------------------------------------
// copy constructor

ChoiceFormat::ChoiceFormat(const    ChoiceFormat&   that)
: NumberFormat(that),
  constructorErrorCode(that.constructorErrorCode),
  msgPattern(that.msgPattern)
{
}

// -------------------------------------
// Private constructor that creates a
// ChoiceFormat instance based on the
// pattern and populates UParseError

ChoiceFormat::ChoiceFormat(const UnicodeString& newPattern,
                           UParseError& parseError,
                           UErrorCode& status)
: constructorErrorCode(status),
  msgPattern(status)
{
    applyPattern(newPattern,parseError, status);
}
// -------------------------------------

UBool
ChoiceFormat::operator==(const Format& that) const
{
    if (this == &that) return TRUE;
    if (!NumberFormat::operator==(that)) return FALSE;
    ChoiceFormat& thatAlias = (ChoiceFormat&)that;
    return msgPattern == thatAlias.msgPattern;
}

// -------------------------------------
// copy constructor

const ChoiceFormat&
ChoiceFormat::operator=(const   ChoiceFormat& that)
{
    if (this != &that) {
        NumberFormat::operator=(that);
        constructorErrorCode = that.constructorErrorCode;
        msgPattern = that.msgPattern;
    }
    return *this;
}

// -------------------------------------

ChoiceFormat::~ChoiceFormat()
{
}

// -------------------------------------

/**
 * Convert a double value to a string without the overhead of NumberFormat.
 */
UnicodeString&
ChoiceFormat::dtos(double value,
                   UnicodeString& string)
{
    /* Buffer to contain the digits and any extra formatting stuff. */
    char temp[DBL_DIG + 16];
    char *itrPtr = temp;
    char *expPtr;

    sprintf(temp, "%.*g", DBL_DIG, value);

    /* Find and convert the decimal point.
       Using setlocale on some machines will cause sprintf to use a comma for certain locales.
    */
    while (*itrPtr && (*itrPtr == '-' || isdigit(*itrPtr))) {
        itrPtr++;
    }
    if (*itrPtr != 0 && *itrPtr != 'e') {
        /* We reached something that looks like a decimal point.
        In case someone used setlocale(), which changes the decimal point. */
        *itrPtr = '.';
        itrPtr++;
    }
    /* Search for the exponent */
    while (*itrPtr && *itrPtr != 'e') {
        itrPtr++;
    }
    if (*itrPtr == 'e') {
        itrPtr++;
        /* Verify the exponent sign */
        if (*itrPtr == '+' || *itrPtr == '-') {
            itrPtr++;
        }
        /* Remove leading zeros. You will see this on Windows machines. */
        expPtr = itrPtr;
        while (*itrPtr == '0') {
            itrPtr++;
        }
        if (*itrPtr && expPtr != itrPtr) {
            /* Shift the exponent without zeros. */
            while (*itrPtr) {
                *(expPtr++)  = *(itrPtr++);
            }
            // NULL terminate
            *expPtr = 0;
        }
    }

    string = UnicodeString(temp, -1, US_INV);    /* invariant codepage */
    return string;
}

// -------------------------------------
// calls the overloaded applyPattern method.

void
ChoiceFormat::applyPattern(const UnicodeString& pattern,
                           UErrorCode& status)
{
    msgPattern.parseChoiceStyle(pattern, NULL, status);
    constructorErrorCode = status;
}

// -------------------------------------
// Applies the pattern to this ChoiceFormat instance.

void
ChoiceFormat::applyPattern(const UnicodeString& pattern,
                           UParseError& parseError,
                           UErrorCode& status)
{
    msgPattern.parseChoiceStyle(pattern, &parseError, status);
    constructorErrorCode = status;
}
// -------------------------------------
// Returns the input pattern string.

UnicodeString&
ChoiceFormat::toPattern(UnicodeString& result) const
{
    return result = msgPattern.getPatternString();
}

// -------------------------------------
// Sets the limit and format arrays.
void
ChoiceFormat::setChoices(  const double* limits,
                           const UnicodeString* formats,
                           int32_t cnt )
{
    UErrorCode errorCode = U_ZERO_ERROR;
    setChoices(limits, NULL, formats, cnt, errorCode);
}

// -------------------------------------
// Sets the limit and format arrays.
void
ChoiceFormat::setChoices(  const double* limits,
                           const UBool* closures,
                           const UnicodeString* formats,
                           int32_t cnt )
{
    UErrorCode errorCode = U_ZERO_ERROR;
    setChoices(limits, closures, formats, cnt, errorCode);
}

void
ChoiceFormat::setChoices(const double* limits,
                         const UBool* closures,
                         const UnicodeString* formats,
                         int32_t count,
                         UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) {
        return;
    }
    if (limits == NULL || formats == NULL) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    // Reconstruct the original input pattern.
    // Modified version of the pre-ICU 4.8 toPattern() implementation.
    UnicodeString result;
    for (int32_t i = 0; i < count; ++i) {
        if (i != 0) {
            result += CHOICE_FORMAT_VERTICAL_BAR;
        }
        UnicodeString buf;
        if (uprv_isPositiveInfinity(limits[i])) {
            result += CHOICE_FORMAT_INFINITY;
        } else if (uprv_isNegativeInfinity(limits[i])) {
            result += CHOICE_FORMAT_MINUS;
            result += CHOICE_FORMAT_INFINITY;
        } else {
            result += dtos(limits[i], buf);
        }
        if (closures != NULL && closures[i]) {
            result += CHOICE_FORMAT_LESS_THAN;
        } else {
            result += CHOICE_FORMAT_LESS_EQUAL;
        }
        // Append formats[i], using quotes if there are special
        // characters.  Single quotes themselves must be escaped in
        // either case.
        const UnicodeString& text = formats[i];
        int32_t textLength = text.length();
        int32_t nestingLevel = 0;
        for (int32_t j = 0; j < textLength; ++j) {
            UChar c = text[j];
            if (c == CHOICE_FORMAT_SINGLE_QUOTE && nestingLevel == 0) {
                // Double each top-level apostrophe.
                result.append(c);
            } else if (c == CHOICE_FORMAT_VERTICAL_BAR && nestingLevel == 0) {
                // Surround each pipe symbol with apostrophes for quoting.
                // If the next character is an apostrophe, then that will be doubled,
                // and although the parser will see the apostrophe pairs beginning
                // and ending one character earlier than our doubling, the result
                // is as desired.
                //   | -> '|'
                //   |' -> '|'''
                //   |'' -> '|''''' etc.
                result.append(CHOICE_FORMAT_SINGLE_QUOTE).append(c).append(CHOICE_FORMAT_SINGLE_QUOTE);
                continue;  // Skip the append(c) at the end of the loop body.
            } else if (c == LEFT_CURLY_BRACE) {
                ++nestingLevel;
            } else if (c == RIGHT_CURLY_BRACE && nestingLevel > 0) {
                --nestingLevel;
            }
            result.append(c);
        }
    }
    // Apply the reconstructed pattern.
    applyPattern(result, errorCode);
}

// -------------------------------------
// Gets the limit array.

const double*
ChoiceFormat::getLimits(int32_t& cnt) const
{
    cnt = 0;
    return NULL;
}

// -------------------------------------
// Gets the closures array.

const UBool*
ChoiceFormat::getClosures(int32_t& cnt) const
{
    cnt = 0;
    return NULL;
}

// -------------------------------------
// Gets the format array.

const UnicodeString*
ChoiceFormat::getFormats(int32_t& cnt) const
{
    cnt = 0;
    return NULL;
}

// -------------------------------------
// Formats an int64 number, it's actually formatted as
// a double.  The returned format string may differ
// from the input number because of this.

UnicodeString&
ChoiceFormat::format(int64_t number,
                     UnicodeString& appendTo,
                     FieldPosition& status) const
{
    return format((double) number, appendTo, status);
}

// -------------------------------------
// Formats an int32_t number, it's actually formatted as
// a double.

UnicodeString&
ChoiceFormat::format(int32_t number,
                     UnicodeString& appendTo,
                     FieldPosition& status) const
{
    return format((double) number, appendTo, status);
}

// -------------------------------------
// Formats a double number.

UnicodeString&
ChoiceFormat::format(double number,
                     UnicodeString& appendTo,
                     FieldPosition& /*pos*/) const
{
    if (msgPattern.countParts() == 0) {
        // No pattern was applied, or it failed.
        return appendTo;
    }
    // Get the appropriate sub-message.
    int32_t msgStart = findSubMessage(msgPattern, 0, number);
    if (!MessageImpl::jdkAposMode(msgPattern)) {
        int32_t patternStart = msgPattern.getPart(msgStart).getLimit();
        int32_t msgLimit = msgPattern.getLimitPartIndex(msgStart);
        appendTo.append(msgPattern.getPatternString(),
                        patternStart,
                        msgPattern.getPatternIndex(msgLimit) - patternStart);
        return appendTo;
    }
    // JDK compatibility mode: Remove SKIP_SYNTAX.
    return MessageImpl::appendSubMessageWithoutSkipSyntax(msgPattern, msgStart, appendTo);
}

int32_t
ChoiceFormat::findSubMessage(const MessagePattern &pattern, int32_t partIndex, double number) {
    int32_t count = pattern.countParts();
    int32_t msgStart;
    // Iterate over (ARG_INT|DOUBLE, ARG_SELECTOR, message) tuples
    // until ARG_LIMIT or end of choice-only pattern.
    // Ignore the first number and selector and start the loop on the first message.
    partIndex += 2;
    for (;;) {
        // Skip but remember the current sub-message.
        msgStart = partIndex;
        partIndex = pattern.getLimitPartIndex(partIndex);
        if (++partIndex >= count) {
            // Reached the end of the choice-only pattern.
            // Return with the last sub-message.
            break;
        }
        const MessagePattern::Part &part = pattern.getPart(partIndex++);
        UMessagePatternPartType type = part.getType();
        if (type == UMSGPAT_PART_TYPE_ARG_LIMIT) {
            // Reached the end of the ChoiceFormat style.
            // Return with the last sub-message.
            break;
        }
        // part is an ARG_INT or ARG_DOUBLE
        U_ASSERT(MessagePattern::Part::hasNumericValue(type));
        double boundary = pattern.getNumericValue(part);
        // Fetch the ARG_SELECTOR character.
        int32_t selectorIndex = pattern.getPatternIndex(partIndex++);
        UChar boundaryChar = pattern.getPatternString().charAt(selectorIndex);
        if (boundaryChar == CHOICE_FORMAT_LESS_THAN ? !(number > boundary) : !(number >= boundary)) {
            // The number is in the interval between the previous boundary and the current one.
            // Return with the sub-message between them.
            // The !(a>b) and !(a>=b) comparisons are equivalent to
            // (a<=b) and (a<b) except they "catch" NaN.
            break;
        }
    }
    return msgStart;
}

// -------------------------------------
// Formats an array of objects. Checks if the data type of the objects
// to get the right value for formatting.

UnicodeString&
ChoiceFormat::format(const Formattable* objs,
                     int32_t cnt,
                     UnicodeString& appendTo,
                     FieldPosition& pos,
                     UErrorCode& status) const
{
    if(cnt < 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return appendTo;
    }
    if (msgPattern.countParts() == 0) {
        status = U_INVALID_STATE_ERROR;
        return appendTo;
    }

    for (int32_t i = 0; i < cnt; i++) {
        double objDouble = objs[i].getDouble(status);
        if (U_SUCCESS(status)) {
            format(objDouble, appendTo, pos);
        }
    }

    return appendTo;
}

// -------------------------------------

void
ChoiceFormat::parse(const UnicodeString& text,
                    Formattable& result,
                    ParsePosition& pos) const
{
    result.setDouble(parseArgument(msgPattern, 0, text, pos));
}

double
ChoiceFormat::parseArgument(
        const MessagePattern &pattern, int32_t partIndex,
        const UnicodeString &source, ParsePosition &pos) {
    // find the best number (defined as the one with the longest parse)
    int32_t start = pos.getIndex();
    int32_t furthest = start;
    double bestNumber = uprv_getNaN();
    double tempNumber = 0.0;
    int32_t count = pattern.countParts();
    while (partIndex < count && pattern.getPartType(partIndex) != UMSGPAT_PART_TYPE_ARG_LIMIT) {
        tempNumber = pattern.getNumericValue(pattern.getPart(partIndex));
        partIndex += 2;  // skip the numeric part and ignore the ARG_SELECTOR
        int32_t msgLimit = pattern.getLimitPartIndex(partIndex);
        int32_t len = matchStringUntilLimitPart(pattern, partIndex, msgLimit, source, start);
        if (len >= 0) {
            int32_t newIndex = start + len;
            if (newIndex > furthest) {
                furthest = newIndex;
                bestNumber = tempNumber;
                if (furthest == source.length()) {
                    break;
                }
            }
        }
        partIndex = msgLimit + 1;
    }
    if (furthest == start) {
        pos.setErrorIndex(start);
    } else {
        pos.setIndex(furthest);
    }
    return bestNumber;
}

int32_t
ChoiceFormat::matchStringUntilLimitPart(
        const MessagePattern &pattern, int32_t partIndex, int32_t limitPartIndex,
        const UnicodeString &source, int32_t sourceOffset) {
    int32_t matchingSourceLength = 0;
    const UnicodeString &msgString = pattern.getPatternString();
    int32_t prevIndex = pattern.getPart(partIndex).getLimit();
    for (;;) {
        const MessagePattern::Part &part = pattern.getPart(++partIndex);
        if (partIndex == limitPartIndex || part.getType() == UMSGPAT_PART_TYPE_SKIP_SYNTAX) {
            int32_t index = part.getIndex();
            int32_t length = index - prevIndex;
            if (length != 0 && 0 != source.compare(sourceOffset, length, msgString, prevIndex, length)) {
                return -1;  // mismatch
            }
            matchingSourceLength += length;
            if (partIndex == limitPartIndex) {
                return matchingSourceLength;
            }
            prevIndex = part.getLimit();  // SKIP_SYNTAX
        }
    }
}

// -------------------------------------

ChoiceFormat*
ChoiceFormat::clone() const
{
    ChoiceFormat *aCopy = new ChoiceFormat(*this);
    return aCopy;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1996-2014, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

/*
* File coleitr.cpp
*
* Created by: Helena Shih
*
* Modification History:
*
*  Date      Name        Description
*
*  6/23/97   helena      Adding comments to make code more readable.
* 08/03/98   erm         Synched with 1.2 version of CollationElementIterator.java
* 12/10/99   aliu        Ported Thai collation support from Java.
* 01/25/01   swquek      Modified to a C++ wrapper calling C APIs (ucoliter.h)
* 02/19/01   swquek      Removed CollationElementIterator() since it is 
*                        private constructor and no calls are made to it
* 2012-2014  markus      Rewritten in C++ again.
*/



#if !UCONFIG_NO_COLLATION


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ******************************************************************************
 *   Copyright (C) 1997-2014, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 ******************************************************************************
 */

/**
 * \file 
 * \brief C++ API: Collation Element Iterator.
 */

/**
* File coleitr.h
*
* Created by: Helena Shih
*
* Modification History:
*
*  Date       Name        Description
*
*  8/18/97    helena      Added internal API documentation.
* 08/03/98    erm         Synched with 1.2 version CollationElementIterator.java
* 12/10/99    aliu        Ported Thai collation support from Java.
* 01/25/01    swquek      Modified into a C++ wrapper calling C APIs (ucoliter.h)
* 02/19/01    swquek      Removed CollationElementsIterator() since it is 
*                         private constructor and no calls are made to it
* 2012-2014   markus      Rewritten in C++ again.
*/




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_COLLATION




struct UCollationElements;
struct UHashtable;

U_NAMESPACE_BEGIN

struct CollationData;

class CharacterIterator;
class CollationIterator;
class RuleBasedCollator;
class UCollationPCE;
class UVector32;

/**
* The CollationElementIterator class is used as an iterator to walk through     
* each character of an international string. Use the iterator to return the
* ordering priority of the positioned character. The ordering priority of a 
* character, which we refer to as a key, defines how a character is collated in 
* the given collation object.
* For example, consider the following in Slovak and in traditional Spanish collation:
* <pre>
*        "ca" -> the first key is key('c') and second key is key('a').
*        "cha" -> the first key is key('ch') and second key is key('a').</pre>
* And in German phonebook collation,
* <pre> \htmlonly       "&#x00E6;b"-> the first key is key('a'), the second key is key('e'), and
*        the third key is key('b'). \endhtmlonly </pre>
* The key of a character, is an integer composed of primary order(short),
* secondary order(char), and tertiary order(char). Java strictly defines the 
* size and signedness of its primitive data types. Therefore, the static
* functions primaryOrder(), secondaryOrder(), and tertiaryOrder() return 
* int32_t to ensure the correctness of the key value.
* <p>Example of the iterator usage: (without error checking)
* <pre>
* \code
*   void CollationElementIterator_Example()
*   {
*       UnicodeString str = "This is a test";
*       UErrorCode success = U_ZERO_ERROR;
*       RuleBasedCollator* rbc =
*           (RuleBasedCollator*) RuleBasedCollator::createInstance(success);
*       CollationElementIterator* c =
*           rbc->createCollationElementIterator( str );
*       int32_t order = c->next(success);
*       c->reset();
*       order = c->previous(success);
*       delete c;
*       delete rbc;
*   }
* \endcode
* </pre>
* <p>
* The method next() returns the collation order of the next character based on
* the comparison level of the collator. The method previous() returns the
* collation order of the previous character based on the comparison level of
* the collator. The Collation Element Iterator moves only in one direction
* between calls to reset(), setOffset(), or setText(). That is, next() 
* and previous() can not be inter-used. Whenever previous() is to be called after 
* next() or vice versa, reset(), setOffset() or setText() has to be called first
* to reset the status, shifting pointers to either the end or the start of
* the string (reset() or setText()), or the specified position (setOffset()).
* Hence at the next call of next() or previous(), the first or last collation order,
* or collation order at the spefcifieid position will be returned. If a change of
* direction is done without one of these calls, the result is undefined.
* <p>
* The result of a forward iterate (next()) and reversed result of the backward
* iterate (previous()) on the same string are equivalent, if collation orders
* with the value 0 are ignored.
* Character based on the comparison level of the collator.  A collation order 
* consists of primary order, secondary order and tertiary order.  The data 
* type of the collation order is <strong>int32_t</strong>. 
*
* Note, CollationElementIterator should not be subclassed.
* @see     Collator
* @see     RuleBasedCollator
* @version 1.8 Jan 16 2001
*/
class U_I18N_API CollationElementIterator U_FINAL : public UObject {
public: 

    // CollationElementIterator public data member ------------------------------

    enum {
        /**
         * NULLORDER indicates that an error has occured while processing
         * @stable ICU 2.0
         */
        NULLORDER = (int32_t)0xffffffff
    };

    // CollationElementIterator public constructor/destructor -------------------

    /**
    * Copy constructor.
    *
    * @param other    the object to be copied from
    * @stable ICU 2.0
    */
    CollationElementIterator(const CollationElementIterator& other);

    /** 
    * Destructor
    * @stable ICU 2.0
    */
    virtual ~CollationElementIterator();

    // CollationElementIterator public methods ----------------------------------

    /**
    * Returns true if "other" is the same as "this"
    *
    * @param other    the object to be compared
    * @return         true if "other" is the same as "this"
    * @stable ICU 2.0
    */
    UBool operator==(const CollationElementIterator& other) const;

    /**
    * Returns true if "other" is not the same as "this".
    *
    * @param other    the object to be compared
    * @return         true if "other" is not the same as "this"
    * @stable ICU 2.0
    */
    UBool operator!=(const CollationElementIterator& other) const;

    /**
    * Resets the cursor to the beginning of the string.
    * @stable ICU 2.0
    */
    void reset(void);

    /**
    * Gets the ordering priority of the next character in the string.
    * @param status the error code status.
    * @return the next character's ordering. otherwise returns NULLORDER if an 
    *         error has occured or if the end of string has been reached
    * @stable ICU 2.0
    */
    int32_t next(UErrorCode& status);

    /**
    * Get the ordering priority of the previous collation element in the string.
    * @param status the error code status.
    * @return the previous element's ordering. otherwise returns NULLORDER if an 
    *         error has occured or if the start of string has been reached
    * @stable ICU 2.0
    */
    int32_t previous(UErrorCode& status);

    /**
    * Gets the primary order of a collation order.
    * @param order the collation order
    * @return the primary order of a collation order.
    * @stable ICU 2.0
    */
    static inline int32_t primaryOrder(int32_t order);

    /**
    * Gets the secondary order of a collation order.
    * @param order the collation order
    * @return the secondary order of a collation order.
    * @stable ICU 2.0
    */
    static inline int32_t secondaryOrder(int32_t order);

    /**
    * Gets the tertiary order of a collation order.
    * @param order the collation order
    * @return the tertiary order of a collation order.
    * @stable ICU 2.0
    */
    static inline int32_t tertiaryOrder(int32_t order);

    /**
    * Return the maximum length of any expansion sequences that end with the 
    * specified comparison order.
    * @param order a collation order returned by previous or next.
    * @return maximum size of the expansion sequences ending with the collation 
    *         element or 1 if collation element does not occur at the end of any 
    *         expansion sequence
    * @stable ICU 2.0
    */
    int32_t getMaxExpansion(int32_t order) const;

    /**
    * Gets the comparison order in the desired strength. Ignore the other
    * differences.
    * @param order The order value
    * @stable ICU 2.0
    */
    int32_t strengthOrder(int32_t order) const;

    /**
    * Sets the source string.
    * @param str the source string.
    * @param status the error code status.
    * @stable ICU 2.0
    */
    void setText(const UnicodeString& str, UErrorCode& status);

    /**
    * Sets the source string.
    * @param str the source character iterator.
    * @param status the error code status.
    * @stable ICU 2.0
    */
    void setText(CharacterIterator& str, UErrorCode& status);

    /**
    * Checks if a comparison order is ignorable.
    * @param order the collation order.
    * @return TRUE if a character is ignorable, FALSE otherwise.
    * @stable ICU 2.0
    */
    static inline UBool isIgnorable(int32_t order);

    /**
    * Gets the offset of the currently processed character in the source string.
    * @return the offset of the character.
    * @stable ICU 2.0
    */
    int32_t getOffset(void) const;

    /**
    * Sets the offset of the currently processed character in the source string.
    * @param newOffset the new offset.
    * @param status the error code status.
    * @return the offset of the character.
    * @stable ICU 2.0
    */
    void setOffset(int32_t newOffset, UErrorCode& status);

    /**
    * ICU "poor man's RTTI", returns a UClassID for the actual class.
    *
    * @stable ICU 2.2
    */
    virtual UClassID getDynamicClassID() const;

    /**
    * ICU "poor man's RTTI", returns a UClassID for this class.
    *
    * @stable ICU 2.2
    */
    static UClassID U_EXPORT2 getStaticClassID();

#ifndef U_HIDE_INTERNAL_API
    /** @internal */
    static inline CollationElementIterator *fromUCollationElements(UCollationElements *uc) {
        return reinterpret_cast<CollationElementIterator *>(uc);
    }
    /** @internal */
    static inline const CollationElementIterator *fromUCollationElements(const UCollationElements *uc) {
        return reinterpret_cast<const CollationElementIterator *>(uc);
    }
    /** @internal */
    inline UCollationElements *toUCollationElements() {
        return reinterpret_cast<UCollationElements *>(this);
    }
    /** @internal */
    inline const UCollationElements *toUCollationElements() const {
        return reinterpret_cast<const UCollationElements *>(this);
    }
#endif  // U_HIDE_INTERNAL_API

private:
    friend class RuleBasedCollator;
    friend class UCollationPCE;

    /**
    * CollationElementIterator constructor. This takes the source string and the 
    * collation object. The cursor will walk thru the source string based on the 
    * predefined collation rules. If the source string is empty, NULLORDER will 
    * be returned on the calls to next().
    * @param sourceText    the source string.
    * @param order         the collation object.
    * @param status        the error code status.
    */
    CollationElementIterator(const UnicodeString& sourceText,
        const RuleBasedCollator* order, UErrorCode& status);
    // Note: The constructors should take settings & tailoring, not a collator,
    // to avoid circular dependencies.
    // However, for operator==() we would need to be able to compare tailoring data for equality
    // without making CollationData or CollationTailoring depend on TailoredSet.
    // (See the implementation of RuleBasedCollator::operator==().)
    // That might require creating an intermediate class that would be used
    // by both CollationElementIterator and RuleBasedCollator
    // but only contain the part of RBC== related to data and rules.

    /**
    * CollationElementIterator constructor. This takes the source string and the 
    * collation object.  The cursor will walk thru the source string based on the 
    * predefined collation rules.  If the source string is empty, NULLORDER will 
    * be returned on the calls to next().
    * @param sourceText    the source string.
    * @param order         the collation object.
    * @param status        the error code status.
    */
    CollationElementIterator(const CharacterIterator& sourceText,
        const RuleBasedCollator* order, UErrorCode& status);

    /**
    * Assignment operator
    *
    * @param other    the object to be copied
    */
    const CollationElementIterator&
        operator=(const CollationElementIterator& other);

    CollationElementIterator(); // default constructor not implemented

    /** Normalizes dir_=1 (just after setOffset()) to dir_=0 (just after reset()). */
    inline int8_t normalizeDir() const { return dir_ == 1 ? 0 : dir_; }

    static UHashtable *computeMaxExpansions(const CollationData *data, UErrorCode &errorCode);

    static int32_t getMaxExpansion(const UHashtable *maxExpansions, int32_t order);

    // CollationElementIterator private data members ----------------------------

    CollationIterator *iter_;  // owned
    const RuleBasedCollator *rbc_;  // aliased
    uint32_t otherHalf_;
    /**
     * <0: backwards; 0: just after reset() (previous() begins from end);
     * 1: just after setOffset(); >1: forward
     */
    int8_t dir_;
    /**
     * Stores offsets from expansions and from unsafe-backwards iteration,
     * so that getOffset() returns intermediate offsets for the CEs
     * that are consistent with forward iteration.
     */
    UVector32 *offsets_;

    UnicodeString string_;
};

// CollationElementIterator inline method definitions --------------------------

inline int32_t CollationElementIterator::primaryOrder(int32_t order)
{
    return (order >> 16) & 0xffff;
}

inline int32_t CollationElementIterator::secondaryOrder(int32_t order)
{
    return (order >> 8) & 0xff;
}

inline int32_t CollationElementIterator::tertiaryOrder(int32_t order)
{
    return order & 0xff;
}

inline UBool CollationElementIterator::isIgnorable(int32_t order)
{
    return (order & 0xffff0000) == 0;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_COLLATION */

#endif /* U_SHOW_CPLUSPLUS_API */





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collation.h
*
* created on: 2010oct27
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION

U_NAMESPACE_BEGIN

/**
 * Collation v2 basic definitions and static helper functions.
 *
 * Data structures except for expansion tables store 32-bit CEs which are
 * either specials (see tags below) or are compact forms of 64-bit CEs.
 */
class U_I18N_API Collation {
public:
    // Special sort key bytes for all levels.
    static const uint8_t TERMINATOR_BYTE = 0;
    static const uint8_t LEVEL_SEPARATOR_BYTE = 1;

    /** The secondary/tertiary lower limit for tailoring before any root elements. */
    static const uint32_t BEFORE_WEIGHT16 = 0x0100;

    /**
     * Merge-sort-key separator.
     * Same as the unique primary and identical-level weights of U+FFFE.
     * Must not be used as primary compression low terminator.
     * Otherwise usable.
     */
    static const uint8_t MERGE_SEPARATOR_BYTE = 2;
    static const uint32_t MERGE_SEPARATOR_PRIMARY = 0x02000000;  // U+FFFE
    static const uint32_t MERGE_SEPARATOR_CE32 = 0x02000505;  // U+FFFE

    /**
     * Primary compression low terminator, must be greater than MERGE_SEPARATOR_BYTE.
     * Reserved value in primary second byte if the lead byte is compressible.
     * Otherwise usable in all CE weight bytes.
     */
    static const uint8_t PRIMARY_COMPRESSION_LOW_BYTE = 3;
    /**
     * Primary compression high terminator.
     * Reserved value in primary second byte if the lead byte is compressible.
     * Otherwise usable in all CE weight bytes.
     */
    static const uint8_t PRIMARY_COMPRESSION_HIGH_BYTE = 0xff;

    /** Default secondary/tertiary weight lead byte. */
    static const uint8_t COMMON_BYTE = 5;
    static const uint32_t COMMON_WEIGHT16 = 0x0500;
    /** Middle 16 bits of a CE with a common secondary weight. */
    static const uint32_t COMMON_SECONDARY_CE = 0x05000000;
    /** Lower 16 bits of a CE with a common tertiary weight. */
    static const uint32_t COMMON_TERTIARY_CE = 0x0500;
    /** Lower 32 bits of a CE with common secondary and tertiary weights. */
    static const uint32_t COMMON_SEC_AND_TER_CE = 0x05000500;

    static const uint32_t SECONDARY_MASK = 0xffff0000;
    static const uint32_t CASE_MASK = 0xc000;
    static const uint32_t SECONDARY_AND_CASE_MASK = SECONDARY_MASK | CASE_MASK;
    /** Only the 2*6 bits for the pure tertiary weight. */
    static const uint32_t ONLY_TERTIARY_MASK = 0x3f3f;
    /** Only the secondary & tertiary bits; no case, no quaternary. */
    static const uint32_t ONLY_SEC_TER_MASK = SECONDARY_MASK | ONLY_TERTIARY_MASK;
    /** Case bits and tertiary bits. */
    static const uint32_t CASE_AND_TERTIARY_MASK = CASE_MASK | ONLY_TERTIARY_MASK;
    static const uint32_t QUATERNARY_MASK = 0xc0;
    /** Case bits and quaternary bits. */
    static const uint32_t CASE_AND_QUATERNARY_MASK = CASE_MASK | QUATERNARY_MASK;

    static const uint8_t UNASSIGNED_IMPLICIT_BYTE = 0xfe;  // compressible
    /**
     * First unassigned: AlphabeticIndex overflow boundary.
     * We want a 3-byte primary so that it fits into the root elements table.
     *
     * This 3-byte primary will not collide with
     * any unassigned-implicit 4-byte primaries because
     * the first few hundred Unicode code points all have real mappings.
     */
    static const uint32_t FIRST_UNASSIGNED_PRIMARY = 0xfe040200;

    static const uint8_t TRAIL_WEIGHT_BYTE = 0xff;  // not compressible
    static const uint32_t FIRST_TRAILING_PRIMARY = 0xff020200;  // [first trailing]
    static const uint32_t MAX_PRIMARY = 0xffff0000;  // U+FFFF
    static const uint32_t MAX_REGULAR_CE32 = 0xffff0505;  // U+FFFF

    // CE32 value for U+FFFD as well as illegal UTF-8 byte sequences (which behave like U+FFFD).
    // We use the third-highest primary weight for U+FFFD (as in UCA 6.3+).
    static const uint32_t FFFD_PRIMARY = MAX_PRIMARY - 0x20000;
    static const uint32_t FFFD_CE32 = MAX_REGULAR_CE32 - 0x20000;

    /**
     * A CE32 is special if its low byte is this or greater.
     * Impossible case bits 11 mark special CE32s.
     * This value itself is used to indicate a fallback to the base collator.
     */
    static const uint8_t SPECIAL_CE32_LOW_BYTE = 0xc0;
    static const uint32_t FALLBACK_CE32 = SPECIAL_CE32_LOW_BYTE;
    /**
     * Low byte of a long-primary special CE32.
     */
    static const uint8_t LONG_PRIMARY_CE32_LOW_BYTE = 0xc1;  // SPECIAL_CE32_LOW_BYTE | LONG_PRIMARY_TAG

    static const uint32_t UNASSIGNED_CE32 = 0xffffffff;  // Compute an unassigned-implicit CE.

    static const uint32_t NO_CE32 = 1;

    /** No CE: End of input. Only used in runtime code, not stored in data. */
    static const uint32_t NO_CE_PRIMARY = 1;  // not a left-adjusted weight
    static const uint32_t NO_CE_WEIGHT16 = 0x0100;  // weight of LEVEL_SEPARATOR_BYTE
    static const int64_t NO_CE = INT64_C(0x101000100);  // NO_CE_PRIMARY, NO_CE_WEIGHT16, NO_CE_WEIGHT16

    /** Sort key levels. */
    enum Level {
        /** Unspecified level. */
        NO_LEVEL,
        PRIMARY_LEVEL,
        SECONDARY_LEVEL,
        CASE_LEVEL,
        TERTIARY_LEVEL,
        QUATERNARY_LEVEL,
        IDENTICAL_LEVEL,
        /** Beyond sort key bytes. */
        ZERO_LEVEL
    };

    /**
     * Sort key level flags: xx_FLAG = 1 << xx_LEVEL.
     * In Java, use enum Level with flag() getters, or use EnumSet rather than hand-made bit sets.
     */
    static const uint32_t NO_LEVEL_FLAG = 1;
    static const uint32_t PRIMARY_LEVEL_FLAG = 2;
    static const uint32_t SECONDARY_LEVEL_FLAG = 4;
    static const uint32_t CASE_LEVEL_FLAG = 8;
    static const uint32_t TERTIARY_LEVEL_FLAG = 0x10;
    static const uint32_t QUATERNARY_LEVEL_FLAG = 0x20;
    static const uint32_t IDENTICAL_LEVEL_FLAG = 0x40;
    static const uint32_t ZERO_LEVEL_FLAG = 0x80;

    /**
     * Special-CE32 tags, from bits 3..0 of a special 32-bit CE.
     * Bits 31..8 are available for tag-specific data.
     * Bits  5..4: Reserved. May be used in the future to indicate lccc!=0 and tccc!=0.
     */
    enum {
        /**
         * Fall back to the base collator.
         * This is the tag value in SPECIAL_CE32_LOW_BYTE and FALLBACK_CE32.
         * Bits 31..8: Unused, 0.
         */
        FALLBACK_TAG = 0,
        /**
         * Long-primary CE with COMMON_SEC_AND_TER_CE.
         * Bits 31..8: Three-byte primary.
         */
        LONG_PRIMARY_TAG = 1,
        /**
         * Long-secondary CE with zero primary.
         * Bits 31..16: Secondary weight.
         * Bits 15.. 8: Tertiary weight.
         */
        LONG_SECONDARY_TAG = 2,
        /**
         * Unused.
         * May be used in the future for single-byte secondary CEs (SHORT_SECONDARY_TAG),
         * storing the secondary in bits 31..24, the ccc in bits 23..16,
         * and the tertiary in bits 15..8.
         */
        RESERVED_TAG_3 = 3,
        /**
         * Latin mini expansions of two simple CEs [pp, 05, tt] [00, ss, 05].
         * Bits 31..24: Single-byte primary weight pp of the first CE.
         * Bits 23..16: Tertiary weight tt of the first CE.
         * Bits 15.. 8: Secondary weight ss of the second CE.
         */
        LATIN_EXPANSION_TAG = 4,
        /**
         * Points to one or more simple/long-primary/long-secondary 32-bit CE32s.
         * Bits 31..13: Index into uint32_t table.
         * Bits 12.. 8: Length=1..31.
         */
        EXPANSION32_TAG = 5,
        /**
         * Points to one or more 64-bit CEs.
         * Bits 31..13: Index into CE table.
         * Bits 12.. 8: Length=1..31.
         */
        EXPANSION_TAG = 6,
        /**
         * Builder data, used only in the CollationDataBuilder, not in runtime data.
         *
         * If bit 8 is 0: Builder context, points to a list of context-sensitive mappings.
         * Bits 31..13: Index to the builder's list of ConditionalCE32 for this character.
         * Bits 12.. 9: Unused, 0.
         *
         * If bit 8 is 1 (IS_BUILDER_JAMO_CE32): Builder-only jamoCE32 value.
         * The builder fetches the Jamo CE32 from the trie.
         * Bits 31..13: Jamo code point.
         * Bits 12.. 9: Unused, 0.
         */
        BUILDER_DATA_TAG = 7,
        /**
         * Points to prefix trie.
         * Bits 31..13: Index into prefix/contraction data.
         * Bits 12.. 8: Unused, 0.
         */
        PREFIX_TAG = 8,
        /**
         * Points to contraction data.
         * Bits 31..13: Index into prefix/contraction data.
         * Bits 12..11: Unused, 0.
         * Bit      10: CONTRACT_TRAILING_CCC flag.
         * Bit       9: CONTRACT_NEXT_CCC flag.
         * Bit       8: CONTRACT_SINGLE_CP_NO_MATCH flag.
         */
        CONTRACTION_TAG = 9,
        /**
         * Decimal digit.
         * Bits 31..13: Index into uint32_t table for non-numeric-collation CE32.
         * Bit      12: Unused, 0.
         * Bits 11.. 8: Digit value 0..9.
         */
        DIGIT_TAG = 10,
        /**
         * Tag for U+0000, for moving the NUL-termination handling
         * from the regular fastpath into specials-handling code.
         * Bits 31..8: Unused, 0.
         */
        U0000_TAG = 11,
        /**
         * Tag for a Hangul syllable.
         * Bits 31..9: Unused, 0.
         * Bit      8: HANGUL_NO_SPECIAL_JAMO flag.
         */
        HANGUL_TAG = 12,
        /**
         * Tag for a lead surrogate code unit.
         * Optional optimization for UTF-16 string processing.
         * Bits 31..10: Unused, 0.
         *       9.. 8: =0: All associated supplementary code points are unassigned-implict.
         *              =1: All associated supplementary code points fall back to the base data.
         *              else: (Normally 2) Look up the data for the supplementary code point.
         */
        LEAD_SURROGATE_TAG = 13,
        /**
         * Tag for CEs with primary weights in code point order.
         * Bits 31..13: Index into CE table, for one data "CE".
         * Bits 12.. 8: Unused, 0.
         *
         * This data "CE" has the following bit fields:
         * Bits 63..32: Three-byte primary pppppp00.
         *      31.. 8: Start/base code point of the in-order range.
         *           7: Flag isCompressible primary.
         *       6.. 0: Per-code point primary-weight increment.
         */
        OFFSET_TAG = 14,
        /**
         * Implicit CE tag. Compute an unassigned-implicit CE.
         * All bits are set (UNASSIGNED_CE32=0xffffffff).
         */
        IMPLICIT_TAG = 15
    };

    static UBool isAssignedCE32(uint32_t ce32) {
        return ce32 != FALLBACK_CE32 && ce32 != UNASSIGNED_CE32;
    }

    /**
     * We limit the number of CEs in an expansion
     * so that we can use a small number of length bits in the data structure,
     * and so that an implementation can copy CEs at runtime without growing a destination buffer.
     */
    static const int32_t MAX_EXPANSION_LENGTH = 31;
    static const int32_t MAX_INDEX = 0x7ffff;

    /**
     * Set if there is no match for the single (no-suffix) character itself.
     * This is only possible if there is a prefix.
     * In this case, discontiguous contraction matching cannot add combining marks
     * starting from an empty suffix.
     * The default CE32 is used anyway if there is no suffix match.
     */
    static const uint32_t CONTRACT_SINGLE_CP_NO_MATCH = 0x100;
    /** Set if the first character of every contraction suffix has lccc!=0. */
    static const uint32_t CONTRACT_NEXT_CCC = 0x200;
    /** Set if any contraction suffix ends with lccc!=0. */
    static const uint32_t CONTRACT_TRAILING_CCC = 0x400;

    /** For HANGUL_TAG: None of its Jamo CE32s isSpecialCE32(). */
    static const uint32_t HANGUL_NO_SPECIAL_JAMO = 0x100;

    static const uint32_t LEAD_ALL_UNASSIGNED = 0;
    static const uint32_t LEAD_ALL_FALLBACK = 0x100;
    static const uint32_t LEAD_MIXED = 0x200;
    static const uint32_t LEAD_TYPE_MASK = 0x300;

    static uint32_t makeLongPrimaryCE32(uint32_t p) { return p | LONG_PRIMARY_CE32_LOW_BYTE; }

    /** Turns the long-primary CE32 into a primary weight pppppp00. */
    static inline uint32_t primaryFromLongPrimaryCE32(uint32_t ce32) {
        return ce32 & 0xffffff00;
    }
    static inline int64_t ceFromLongPrimaryCE32(uint32_t ce32) {
        return ((int64_t)(ce32 & 0xffffff00) << 32) | COMMON_SEC_AND_TER_CE;
    }

    static uint32_t makeLongSecondaryCE32(uint32_t lower32) {
        return lower32 | SPECIAL_CE32_LOW_BYTE | LONG_SECONDARY_TAG;
    }
    static inline int64_t ceFromLongSecondaryCE32(uint32_t ce32) {
        return ce32 & 0xffffff00;
    }

    /** Makes a special CE32 with tag, index and length. */
    static uint32_t makeCE32FromTagIndexAndLength(int32_t tag, int32_t index, int32_t length) {
        return (index << 13) | (length << 8) | SPECIAL_CE32_LOW_BYTE | tag;
    }
    /** Makes a special CE32 with only tag and index. */
    static uint32_t makeCE32FromTagAndIndex(int32_t tag, int32_t index) {
        return (index << 13) | SPECIAL_CE32_LOW_BYTE | tag;
    }

    static inline UBool isSpecialCE32(uint32_t ce32) {
        return (ce32 & 0xff) >= SPECIAL_CE32_LOW_BYTE;
    }

    static inline int32_t tagFromCE32(uint32_t ce32) {
        return (int32_t)(ce32 & 0xf);
    }

    static inline UBool hasCE32Tag(uint32_t ce32, int32_t tag) {
        return isSpecialCE32(ce32) && tagFromCE32(ce32) == tag;
    }

    static inline UBool isLongPrimaryCE32(uint32_t ce32) {
        return hasCE32Tag(ce32, LONG_PRIMARY_TAG);
    }

    static UBool isSimpleOrLongCE32(uint32_t ce32) {
        return !isSpecialCE32(ce32) ||
                tagFromCE32(ce32) == LONG_PRIMARY_TAG ||
                tagFromCE32(ce32) == LONG_SECONDARY_TAG;
    }

    /**
     * @return TRUE if the ce32 yields one or more CEs without further data lookups
     */
    static UBool isSelfContainedCE32(uint32_t ce32) {
        return !isSpecialCE32(ce32) ||
                tagFromCE32(ce32) == LONG_PRIMARY_TAG ||
                tagFromCE32(ce32) == LONG_SECONDARY_TAG ||
                tagFromCE32(ce32) == LATIN_EXPANSION_TAG;
    }

    static inline UBool isPrefixCE32(uint32_t ce32) {
        return hasCE32Tag(ce32, PREFIX_TAG);
    }

    static inline UBool isContractionCE32(uint32_t ce32) {
        return hasCE32Tag(ce32, CONTRACTION_TAG);
    }

    static inline UBool ce32HasContext(uint32_t ce32) {
        return isSpecialCE32(ce32) &&
                (tagFromCE32(ce32) == PREFIX_TAG ||
                tagFromCE32(ce32) == CONTRACTION_TAG);
    }

    /**
     * Get the first of the two Latin-expansion CEs encoded in ce32.
     * @see LATIN_EXPANSION_TAG
     */
    static inline int64_t latinCE0FromCE32(uint32_t ce32) {
        return ((int64_t)(ce32 & 0xff000000) << 32) | COMMON_SECONDARY_CE | ((ce32 & 0xff0000) >> 8);
    }

    /**
     * Get the second of the two Latin-expansion CEs encoded in ce32.
     * @see LATIN_EXPANSION_TAG
     */
    static inline int64_t latinCE1FromCE32(uint32_t ce32) {
        return ((ce32 & 0xff00) << 16) | COMMON_TERTIARY_CE;
    }

    /**
     * Returns the data index from a special CE32.
     */
    static inline int32_t indexFromCE32(uint32_t ce32) {
        return (int32_t)(ce32 >> 13);
    }

    /**
     * Returns the data length from a ce32.
     */
    static inline int32_t lengthFromCE32(uint32_t ce32) {
        return (ce32 >> 8) & 31;
    }

    /**
     * Returns the digit value from a DIGIT_TAG ce32.
     */
    static inline char digitFromCE32(uint32_t ce32) {
        return (char)((ce32 >> 8) & 0xf);
    }

    /** Returns a 64-bit CE from a simple CE32 (not special). */
    static inline int64_t ceFromSimpleCE32(uint32_t ce32) {
        // normal form ppppsstt -> pppp0000ss00tt00
        // assert (ce32 & 0xff) < SPECIAL_CE32_LOW_BYTE
        return ((int64_t)(ce32 & 0xffff0000) << 32) | ((ce32 & 0xff00) << 16) | ((ce32 & 0xff) << 8);
    }

    /** Returns a 64-bit CE from a simple/long-primary/long-secondary CE32. */
    static inline int64_t ceFromCE32(uint32_t ce32) {
        uint32_t tertiary = ce32 & 0xff;
        if(tertiary < SPECIAL_CE32_LOW_BYTE) {
            // normal form ppppsstt -> pppp0000ss00tt00
            return ((int64_t)(ce32 & 0xffff0000) << 32) | ((ce32 & 0xff00) << 16) | (tertiary << 8);
        } else {
            ce32 -= tertiary;
            if((tertiary & 0xf) == LONG_PRIMARY_TAG) {
                // long-primary form ppppppC1 -> pppppp00050000500
                return ((int64_t)ce32 << 32) | COMMON_SEC_AND_TER_CE;
            } else {
                // long-secondary form ssssttC2 -> 00000000sssstt00
                // assert (tertiary & 0xf) == LONG_SECONDARY_TAG
                return ce32;
            }
        }
    }

    /** Creates a CE from a primary weight. */
    static inline int64_t makeCE(uint32_t p) {
        return ((int64_t)p << 32) | COMMON_SEC_AND_TER_CE;
    }
    /**
     * Creates a CE from a primary weight,
     * 16-bit secondary/tertiary weights, and a 2-bit quaternary.
     */
    static inline int64_t makeCE(uint32_t p, uint32_t s, uint32_t t, uint32_t q) {
        return ((int64_t)p << 32) | (s << 16) | t | (q << 6);
    }

    /**
     * Increments a 2-byte primary by a code point offset.
     */
    static uint32_t incTwoBytePrimaryByOffset(uint32_t basePrimary, UBool isCompressible,
                                              int32_t offset);

    /**
     * Increments a 3-byte primary by a code point offset.
     */
    static uint32_t incThreeBytePrimaryByOffset(uint32_t basePrimary, UBool isCompressible,
                                                int32_t offset);

    /**
     * Decrements a 2-byte primary by one range step (1..0x7f).
     */
    static uint32_t decTwoBytePrimaryByOneStep(uint32_t basePrimary, UBool isCompressible, int32_t step);

    /**
     * Decrements a 3-byte primary by one range step (1..0x7f).
     */
    static uint32_t decThreeBytePrimaryByOneStep(uint32_t basePrimary, UBool isCompressible, int32_t step);

    /**
     * Computes a 3-byte primary for c's OFFSET_TAG data "CE".
     */
    static uint32_t getThreeBytePrimaryForOffsetData(UChar32 c, int64_t dataCE);

    /**
     * Returns the unassigned-character implicit primary weight for any valid code point c.
     */
    static uint32_t unassignedPrimaryFromCodePoint(UChar32 c);

    static inline int64_t unassignedCEFromCodePoint(UChar32 c) {
        return makeCE(unassignedPrimaryFromCodePoint(c));
    }

private:
    Collation();  // No instantiation.
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationdata.h
*
* created on: 2010oct27
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION







struct UDataMemory;

U_NAMESPACE_BEGIN

class UVector32;

/**
 * Collation data container.
 * Immutable data created by a CollationDataBuilder, or loaded from a file,
 * or deserialized from API-provided binary data.
 *
 * Includes data for the collation base (root/default), aliased if this is not the base.
 */
struct U_I18N_API CollationData : public UMemory {
    // Note: The ucadata.icu loader could discover the reserved ranges by setting an array
    // parallel with the ranges, and resetting ranges that are indexed.
    // The reordering builder code could clone the resulting template array.
    enum {
        REORDER_RESERVED_BEFORE_LATIN = UCOL_REORDER_CODE_FIRST + 14,
        REORDER_RESERVED_AFTER_LATIN
    };

    enum {
        MAX_NUM_SPECIAL_REORDER_CODES = 8,
        /** C++ only, data reader check scriptStartsLength. */
        MAX_NUM_SCRIPT_RANGES = 256
    };

    CollationData(const Normalizer2Impl &nfc)
            : trie(NULL),
              ce32s(NULL), ces(NULL), contexts(NULL), base(NULL),
              jamoCE32s(NULL),
              nfcImpl(nfc),
              numericPrimary(0x12000000),
              ce32sLength(0), cesLength(0), contextsLength(0),
              compressibleBytes(NULL),
              unsafeBackwardSet(NULL),
              fastLatinTable(NULL), fastLatinTableLength(0),
              numScripts(0), scriptsIndex(NULL), scriptStarts(NULL), scriptStartsLength(0),
              rootElements(NULL), rootElementsLength(0) {}

    uint32_t getCE32(UChar32 c) const {
        return UTRIE2_GET32(trie, c);
    }

    uint32_t getCE32FromSupplementary(UChar32 c) const {
        return UTRIE2_GET32_FROM_SUPP(trie, c);
    }

    UBool isDigit(UChar32 c) const {
        return c < 0x660 ? c <= 0x39 && 0x30 <= c :
                Collation::hasCE32Tag(getCE32(c), Collation::DIGIT_TAG);
    }

    UBool isUnsafeBackward(UChar32 c, UBool numeric) const {
        return unsafeBackwardSet->contains(c) || (numeric && isDigit(c));
    }

    UBool isCompressibleLeadByte(uint32_t b) const {
        return compressibleBytes[b];
    }

    inline UBool isCompressiblePrimary(uint32_t p) const {
        return isCompressibleLeadByte(p >> 24);
    }

    /**
     * Returns the CE32 from two contexts words.
     * Access to the defaultCE32 for contraction and prefix matching.
     */
    static uint32_t readCE32(const UChar *p) {
        return ((uint32_t)p[0] << 16) | p[1];
    }

    /**
     * Returns the CE32 for an indirect special CE32 (e.g., with DIGIT_TAG).
     * Requires that ce32 is special.
     */
    uint32_t getIndirectCE32(uint32_t ce32) const;
    /**
     * Returns the CE32 for an indirect special CE32 (e.g., with DIGIT_TAG),
     * if ce32 is special.
     */
    uint32_t getFinalCE32(uint32_t ce32) const;

    /**
     * Computes a CE from c's ce32 which has the OFFSET_TAG.
     */
    int64_t getCEFromOffsetCE32(UChar32 c, uint32_t ce32) const {
        int64_t dataCE = ces[Collation::indexFromCE32(ce32)];
        return Collation::makeCE(Collation::getThreeBytePrimaryForOffsetData(c, dataCE));
    }

    /**
     * Returns the single CE that c maps to.
     * Sets U_UNSUPPORTED_ERROR if c does not map to a single CE.
     */
    int64_t getSingleCE(UChar32 c, UErrorCode &errorCode) const;

    /**
     * Returns the FCD16 value for code point c. c must be >= 0.
     */
    uint16_t getFCD16(UChar32 c) const {
        return nfcImpl.getFCD16(c);
    }

    /**
     * Returns the first primary for the script's reordering group.
     * @return the primary with only the first primary lead byte of the group
     *         (not necessarily an actual root collator primary weight),
     *         or 0 if the script is unknown
     */
    uint32_t getFirstPrimaryForGroup(int32_t script) const;

    /**
     * Returns the last primary for the script's reordering group.
     * @return the last primary of the group
     *         (not an actual root collator primary weight),
     *         or 0 if the script is unknown
     */
    uint32_t getLastPrimaryForGroup(int32_t script) const;

    /**
     * Finds the reordering group which contains the primary weight.
     * @return the first script of the group, or -1 if the weight is beyond the last group
     */
    int32_t getGroupForPrimary(uint32_t p) const;

    int32_t getEquivalentScripts(int32_t script,
                                 int32_t dest[], int32_t capacity, UErrorCode &errorCode) const;

    /**
     * Writes the permutation of primary-weight ranges
     * for the given reordering of scripts and groups.
     * The caller checks for illegal arguments and
     * takes care of [DEFAULT] and memory allocation.
     *
     * Each list element will be a (limit, offset) pair as described
     * for the CollationSettings::reorderRanges.
     * The list will be empty if no ranges are reordered.
     */
    void makeReorderRanges(const int32_t *reorder, int32_t length,
                           UVector32 &ranges, UErrorCode &errorCode) const;

    /** @see jamoCE32s */
    static const int32_t JAMO_CE32S_LENGTH = 19 + 21 + 27;

    /** Main lookup trie. */
    const UTrie2 *trie;
    /**
     * Array of CE32 values.
     * At index 0 there must be CE32(U+0000)
     * to support U+0000's special-tag for NUL-termination handling.
     */
    const uint32_t *ce32s;
    /** Array of CE values for expansions and OFFSET_TAG. */
    const int64_t *ces;
    /** Array of prefix and contraction-suffix matching data. */
    const UChar *contexts;
    /** Base collation data, or NULL if this data itself is a base. */
    const CollationData *base;
    /**
     * Simple array of JAMO_CE32S_LENGTH=19+21+27 CE32s, one per canonical Jamo L/V/T.
     * They are normally simple CE32s, rarely expansions.
     * For fast handling of HANGUL_TAG.
     */
    const uint32_t *jamoCE32s;
    const Normalizer2Impl &nfcImpl;
    /** The single-byte primary weight (xx000000) for numeric collation. */
    uint32_t numericPrimary;

    int32_t ce32sLength;
    int32_t cesLength;
    int32_t contextsLength;

    /** 256 flags for which primary-weight lead bytes are compressible. */
    const UBool *compressibleBytes;
    /**
     * Set of code points that are unsafe for starting string comparison after an identical prefix,
     * or in backwards CE iteration.
     */
    const UnicodeSet *unsafeBackwardSet;

    /**
     * Fast Latin table for common-Latin-text string comparisons.
     * Data structure see class CollationFastLatin.
     */
    const uint16_t *fastLatinTable;
    int32_t fastLatinTableLength;

    /**
     * Data for scripts and reordering groups.
     * Uses include building a reordering permutation table and
     * providing script boundaries to AlphabeticIndex.
     */
    int32_t numScripts;
    /**
     * The length of scriptsIndex is numScripts+16.
     * It maps from a UScriptCode or a special reorder code to an entry in scriptStarts.
     * 16 special reorder codes (not all used) are mapped starting at numScripts.
     * Up to MAX_NUM_SPECIAL_REORDER_CODES are codes for special groups like space/punct/digit.
     * There are special codes at the end for reorder-reserved primary ranges.
     *
     * Multiple scripts may share a range and index, for example Hira & Kana.
     */
    const uint16_t *scriptsIndex;
    /**
     * Start primary weight (top 16 bits only) for a group/script/reserved range
     * indexed by scriptsIndex.
     * The first range (separators & terminators) and the last range (trailing weights)
     * are not reorderable, and no scriptsIndex entry points to them.
     */
    const uint16_t *scriptStarts;
    int32_t scriptStartsLength;

    /**
     * Collation elements in the root collator.
     * Used by the CollationRootElements class. The data structure is described there.
     * NULL in a tailoring.
     */
    const uint32_t *rootElements;
    int32_t rootElementsLength;

private:
    int32_t getScriptIndex(int32_t script) const;
    void makeReorderRanges(const int32_t *reorder, int32_t length,
                           UBool latinMustMove,
                           UVector32 &ranges, UErrorCode &errorCode) const;
    int32_t addLowScriptRange(uint8_t table[], int32_t index, int32_t lowStart) const;
    int32_t addHighScriptRange(uint8_t table[], int32_t index, int32_t highLimit) const;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationiterator.h
*
* created on: 2010oct27
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION





U_NAMESPACE_BEGIN

class SkippedState;
class UCharsTrie;
class UVector32;

/* Large enough for CEs of most short strings. */
#define CEBUFFER_INITIAL_CAPACITY 40

// Export an explicit template instantiation of the MaybeStackArray that
//    is used as a data member of CEBuffer.
//
//    When building DLLs for Windows this is required even though
//    no direct access to the MaybeStackArray leaks out of the i18n library.
//
// See digitlst.h, pluralaffix.h, datefmt.h, and others for similar examples.
//
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN
template class U_I18N_API MaybeStackArray<int64_t, CEBUFFER_INITIAL_CAPACITY>;
#endif

/**
 * Collation element iterator and abstract character iterator.
 *
 * When a method returns a code point value, it must be in 0..10FFFF,
 * except it can be negative as a sentinel value.
 */
class U_I18N_API CollationIterator : public UObject {
private:
    class U_I18N_API CEBuffer {
    private:
        /** Large enough for CEs of most short strings. */
        static const int32_t INITIAL_CAPACITY = CEBUFFER_INITIAL_CAPACITY;
    public:
        CEBuffer() : length(0) {}
        ~CEBuffer();

        inline void append(int64_t ce, UErrorCode &errorCode) {
            if(length < INITIAL_CAPACITY || ensureAppendCapacity(1, errorCode)) {
                buffer[length++] = ce;
            }
        }

        inline void appendUnsafe(int64_t ce) {
            buffer[length++] = ce;
        }

        UBool ensureAppendCapacity(int32_t appCap, UErrorCode &errorCode);

        inline UBool incLength(UErrorCode &errorCode) {
            // Use INITIAL_CAPACITY for a very simple fastpath.
            // (Rather than buffer.getCapacity().)
            if(length < INITIAL_CAPACITY || ensureAppendCapacity(1, errorCode)) {
                ++length;
                return TRUE;
            } else {
                return FALSE;
            }
        }

        inline int64_t set(int32_t i, int64_t ce) {
            return buffer[i] = ce;
        }
        inline int64_t get(int32_t i) const { return buffer[i]; }

        const int64_t *getCEs() const { return buffer.getAlias(); }

        int32_t length;

    private:
        CEBuffer(const CEBuffer &);
        void operator=(const CEBuffer &);

        MaybeStackArray<int64_t, INITIAL_CAPACITY> buffer;
    };

public:
    CollationIterator(const CollationData *d, UBool numeric)
            : trie(d->trie),
              data(d),
              cesIndex(0),
              skipped(NULL),
              numCpFwd(-1),
              isNumeric(numeric) {}

    virtual ~CollationIterator();

    virtual UBool operator==(const CollationIterator &other) const;
    inline UBool operator!=(const CollationIterator &other) const {
        return !operator==(other);
    }

    /**
     * Resets the iterator state and sets the position to the specified offset.
     * Subclasses must implement, and must call the parent class method,
     * or CollationIterator::reset().
     */
    virtual void resetToOffset(int32_t newOffset) = 0;

    virtual int32_t getOffset() const = 0;

    /**
     * Returns the next collation element.
     */
    inline int64_t nextCE(UErrorCode &errorCode) {
        if(cesIndex < ceBuffer.length) {
            // Return the next buffered CE.
            return ceBuffer.get(cesIndex++);
        }
        // assert cesIndex == ceBuffer.length;
        if(!ceBuffer.incLength(errorCode)) {
            return Collation::NO_CE;
        }
        UChar32 c;
        uint32_t ce32 = handleNextCE32(c, errorCode);
        uint32_t t = ce32 & 0xff;
        if(t < Collation::SPECIAL_CE32_LOW_BYTE) {  // Forced-inline of isSpecialCE32(ce32).
            // Normal CE from the main data.
            // Forced-inline of ceFromSimpleCE32(ce32).
            return ceBuffer.set(cesIndex++,
                    ((int64_t)(ce32 & 0xffff0000) << 32) | ((ce32 & 0xff00) << 16) | (t << 8));
        }
        const CollationData *d;
        // The compiler should be able to optimize the previous and the following
        // comparisons of t with the same constant.
        if(t == Collation::SPECIAL_CE32_LOW_BYTE) {
            if(c < 0) {
                return ceBuffer.set(cesIndex++, Collation::NO_CE);
            }
            d = data->base;
            ce32 = d->getCE32(c);
            t = ce32 & 0xff;
            if(t < Collation::SPECIAL_CE32_LOW_BYTE) {
                // Normal CE from the base data.
                return ceBuffer.set(cesIndex++,
                        ((int64_t)(ce32 & 0xffff0000) << 32) | ((ce32 & 0xff00) << 16) | (t << 8));
            }
        } else {
            d = data;
        }
        if(t == Collation::LONG_PRIMARY_CE32_LOW_BYTE) {
            // Forced-inline of ceFromLongPrimaryCE32(ce32).
            return ceBuffer.set(cesIndex++,
                    ((int64_t)(ce32 - t) << 32) | Collation::COMMON_SEC_AND_TER_CE);
        }
        return nextCEFromCE32(d, c, ce32, errorCode);
    }

    /**
     * Fetches all CEs.
     * @return getCEsLength()
     */
    int32_t fetchCEs(UErrorCode &errorCode);

    /**
     * Overwrites the current CE (the last one returned by nextCE()).
     */
    void setCurrentCE(int64_t ce) {
        // assert cesIndex > 0;
        ceBuffer.set(cesIndex - 1, ce);
    }

    /**
     * Returns the previous collation element.
     */
    int64_t previousCE(UVector32 &offsets, UErrorCode &errorCode);

    inline int32_t getCEsLength() const {
        return ceBuffer.length;
    }

    inline int64_t getCE(int32_t i) const {
        return ceBuffer.get(i);
    }

    const int64_t *getCEs() const {
        return ceBuffer.getCEs();
    }

    void clearCEs() {
        cesIndex = ceBuffer.length = 0;
    }

    void clearCEsIfNoneRemaining() {
        if(cesIndex == ceBuffer.length) { clearCEs(); }
    }

    /**
     * Returns the next code point (with post-increment).
     * Public for identical-level comparison and for testing.
     */
    virtual UChar32 nextCodePoint(UErrorCode &errorCode) = 0;

    /**
     * Returns the previous code point (with pre-decrement).
     * Public for identical-level comparison and for testing.
     */
    virtual UChar32 previousCodePoint(UErrorCode &errorCode) = 0;

protected:
    CollationIterator(const CollationIterator &other);

    void reset();

    /**
     * Returns the next code point and its local CE32 value.
     * Returns Collation::FALLBACK_CE32 at the end of the text (c<0)
     * or when c's CE32 value is to be looked up in the base data (fallback).
     *
     * The code point is used for fallbacks, context and implicit weights.
     * It is ignored when the returned CE32 is not special (e.g., FFFD_CE32).
     */
    virtual uint32_t handleNextCE32(UChar32 &c, UErrorCode &errorCode);

    /**
     * Called when handleNextCE32() returns a LEAD_SURROGATE_TAG for a lead surrogate code unit.
     * Returns the trail surrogate in that case and advances past it,
     * if a trail surrogate follows the lead surrogate.
     * Otherwise returns any other code unit and does not advance.
     */
    virtual UChar handleGetTrailSurrogate();

    /**
     * Called when handleNextCE32() returns with c==0, to see whether it is a NUL terminator.
     * (Not needed in Java.)
     */
    virtual UBool foundNULTerminator();

    /**
     * @return FALSE if surrogate code points U+D800..U+DFFF
     *         map to their own implicit primary weights (for UTF-16),
     *         or TRUE if they map to CE(U+FFFD) (for UTF-8)
     */
    virtual UBool forbidSurrogateCodePoints() const;

    virtual void forwardNumCodePoints(int32_t num, UErrorCode &errorCode) = 0;

    virtual void backwardNumCodePoints(int32_t num, UErrorCode &errorCode) = 0;

    /**
     * Returns the CE32 from the data trie.
     * Normally the same as data->getCE32(), but overridden in the builder.
     * Call this only when the faster data->getCE32() cannot be used.
     */
    virtual uint32_t getDataCE32(UChar32 c) const;

    virtual uint32_t getCE32FromBuilderData(uint32_t ce32, UErrorCode &errorCode);

    void appendCEsFromCE32(const CollationData *d, UChar32 c, uint32_t ce32,
                           UBool forward, UErrorCode &errorCode);

    // Main lookup trie of the data object.
    const UTrie2 *trie;
    const CollationData *data;

private:
    int64_t nextCEFromCE32(const CollationData *d, UChar32 c, uint32_t ce32,
                           UErrorCode &errorCode);

    uint32_t getCE32FromPrefix(const CollationData *d, uint32_t ce32,
                               UErrorCode &errorCode);

    UChar32 nextSkippedCodePoint(UErrorCode &errorCode);

    void backwardNumSkipped(int32_t n, UErrorCode &errorCode);

    uint32_t nextCE32FromContraction(
            const CollationData *d, uint32_t contractionCE32,
            const UChar *p, uint32_t ce32, UChar32 c,
            UErrorCode &errorCode);

    uint32_t nextCE32FromDiscontiguousContraction(
            const CollationData *d, UCharsTrie &suffixes, uint32_t ce32,
            int32_t lookAhead, UChar32 c,
            UErrorCode &errorCode);

    /**
     * Returns the previous CE when data->isUnsafeBackward(c, isNumeric).
     */
    int64_t previousCEUnsafe(UChar32 c, UVector32 &offsets, UErrorCode &errorCode);

    /**
     * Turns a string of digits (bytes 0..9)
     * into a sequence of CEs that will sort in numeric order.
     *
     * Starts from this ce32's digit value and consumes the following/preceding digits.
     * The digits string must not be empty and must not have leading zeros.
     */
    void appendNumericCEs(uint32_t ce32, UBool forward, UErrorCode &errorCode);

    /**
     * Turns 1..254 digits into a sequence of CEs.
     * Called by appendNumericCEs() for each segment of at most 254 digits.
     */
    void appendNumericSegmentCEs(const char *digits, int32_t length, UErrorCode &errorCode);

    CEBuffer ceBuffer;
    int32_t cesIndex;

    SkippedState *skipped;

    // Number of code points to read forward, or -1.
    // Used as a forward iteration limit in previousCEUnsafe().
    int32_t numCpFwd;
    // Numeric collation (CollationSettings::NUMERIC).
    UBool isNumeric;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationsets.h
*
* created on: 2013feb09
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION




U_NAMESPACE_BEGIN

struct CollationData;

/**
 * Finds the set of characters and strings that sort differently in the tailoring
 * from the base data.
 *
 * Every mapping in the tailoring needs to be compared to the base,
 * because some mappings are copied for optimization, and
 * all contractions for a character are copied if any contractions for that character
 * are added, modified or removed.
 *
 * It might be simpler to re-parse the rule string, but:
 * - That would require duplicating some of the from-rules builder code.
 * - That would make the runtime code depend on the builder.
 * - That would only work if we have the rule string, and we allow users to
 *   omit the rule string from data files.
 */
class TailoredSet : public UMemory {
public:
    TailoredSet(UnicodeSet *t)
            : data(NULL), baseData(NULL),
              tailored(t),
              suffix(NULL),
              errorCode(U_ZERO_ERROR) {}

    void forData(const CollationData *d, UErrorCode &errorCode);

    /**
     * @return U_SUCCESS(errorCode) in C++, void in Java
     * @internal only public for access by callback
     */
    UBool handleCE32(UChar32 start, UChar32 end, uint32_t ce32);

private:
    void compare(UChar32 c, uint32_t ce32, uint32_t baseCE32);
    void comparePrefixes(UChar32 c, const UChar *p, const UChar *q);
    void compareContractions(UChar32 c, const UChar *p, const UChar *q);

    void addPrefixes(const CollationData *d, UChar32 c, const UChar *p);
    void addPrefix(const CollationData *d, const UnicodeString &pfx, UChar32 c, uint32_t ce32);
    void addContractions(UChar32 c, const UChar *p);
    void addSuffix(UChar32 c, const UnicodeString &sfx);
    void add(UChar32 c);

    /** Prefixes are reversed in the data structure. */
    void setPrefix(const UnicodeString &pfx) {
        unreversedPrefix = pfx;
        unreversedPrefix.reverse();
    }
    void resetPrefix() {
        unreversedPrefix.remove();
    }

    const CollationData *data;
    const CollationData *baseData;
    UnicodeSet *tailored;
    UnicodeString unreversedPrefix;
    const UnicodeString *suffix;
    UErrorCode errorCode;
};

class ContractionsAndExpansions : public UMemory {
public:
    class CESink : public UMemory {
    public:
        virtual ~CESink();
        virtual void handleCE(int64_t ce) = 0;
        virtual void handleExpansion(const int64_t ces[], int32_t length) = 0;
    };

    ContractionsAndExpansions(UnicodeSet *con, UnicodeSet *exp, CESink *s, UBool prefixes)
            : data(NULL),
              contractions(con), expansions(exp),
              sink(s),
              addPrefixes(prefixes),
              checkTailored(0),
              suffix(NULL),
              errorCode(U_ZERO_ERROR) {}

    void forData(const CollationData *d, UErrorCode &errorCode);
    void forCodePoint(const CollationData *d, UChar32 c, UErrorCode &ec);

    // all following: @internal, only public for access by callback

    void handleCE32(UChar32 start, UChar32 end, uint32_t ce32);

    void handlePrefixes(UChar32 start, UChar32 end, uint32_t ce32);
    void handleContractions(UChar32 start, UChar32 end, uint32_t ce32);

    void addExpansions(UChar32 start, UChar32 end);
    void addStrings(UChar32 start, UChar32 end, UnicodeSet *set);

    /** Prefixes are reversed in the data structure. */
    void setPrefix(const UnicodeString &pfx) {
        unreversedPrefix = pfx;
        unreversedPrefix.reverse();
    }
    void resetPrefix() {
        unreversedPrefix.remove();
    }

    const CollationData *data;
    UnicodeSet *contractions;
    UnicodeSet *expansions;
    CESink *sink;
    UBool addPrefixes;
    int8_t checkTailored;  // -1: collected tailored  +1: exclude tailored
    UnicodeSet tailored;
    UnicodeSet ranges;
    UnicodeString unreversedPrefix;
    const UnicodeString *suffix;
    int64_t ces[Collation::MAX_EXPANSION_LENGTH];
    UErrorCode errorCode;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationtailoring.h
*
* created on: 2013mar12
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationsettings.h
*
* created on: 2013feb07
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION






U_NAMESPACE_BEGIN

struct CollationData;

/**
 * Collation settings/options/attributes.
 * These are the values that can be changed via API.
 */
struct U_I18N_API CollationSettings : public SharedObject {
    /**
     * Options bit 0: Perform the FCD check on the input text and deliver normalized text.
     */
    static const int32_t CHECK_FCD = 1;
    /**
     * Options bit 1: Numeric collation.
     * Also known as CODAN = COllate Digits As Numbers.
     *
     * Treat digit sequences as numbers with CE sequences in numeric order,
     * rather than returning a normal CE for each digit.
     */
    static const int32_t NUMERIC = 2;
    /**
     * "Shifted" alternate handling, see ALTERNATE_MASK.
     */
    static const int32_t SHIFTED = 4;
    /**
     * Options bits 3..2: Alternate-handling mask. 0 for non-ignorable.
     * Reserve values 8 and 0xc for shift-trimmed and blanked.
     */
    static const int32_t ALTERNATE_MASK = 0xc;
    /**
     * Options bits 6..4: The 3-bit maxVariable value bit field is shifted by this value.
     */
    static const int32_t MAX_VARIABLE_SHIFT = 4;
    /** maxVariable options bit mask before shifting. */
    static const int32_t MAX_VARIABLE_MASK = 0x70;
    /** Options bit 7: Reserved/unused/0. */
    /**
     * Options bit 8: Sort uppercase first if caseLevel or caseFirst is on.
     */
    static const int32_t UPPER_FIRST = 0x100;
    /**
     * Options bit 9: Keep the case bits in the tertiary weight (they trump other tertiary values)
     * unless case level is on (when they are *moved* into the separate case level).
     * By default, the case bits are removed from the tertiary weight (ignored).
     *
     * When CASE_FIRST is off, UPPER_FIRST must be off too, corresponding to
     * the tri-value UCOL_CASE_FIRST attribute: UCOL_OFF vs. UCOL_LOWER_FIRST vs. UCOL_UPPER_FIRST.
     */
    static const int32_t CASE_FIRST = 0x200;
    /**
     * Options bit mask for caseFirst and upperFirst, before shifting.
     * Same value as caseFirst==upperFirst.
     */
    static const int32_t CASE_FIRST_AND_UPPER_MASK = CASE_FIRST | UPPER_FIRST;
    /**
     * Options bit 10: Insert the case level between the secondary and tertiary levels.
     */
    static const int32_t CASE_LEVEL = 0x400;
    /**
     * Options bit 11: Compare secondary weights backwards. ("French secondary")
     */
    static const int32_t BACKWARD_SECONDARY = 0x800;
    /**
     * Options bits 15..12: The 4-bit strength value bit field is shifted by this value.
     * It is the top used bit field in the options. (No need to mask after shifting.)
     */
    static const int32_t STRENGTH_SHIFT = 12;
    /** Strength options bit mask before shifting. */
    static const int32_t STRENGTH_MASK = 0xf000;

    /** maxVariable values */
    enum MaxVariable {
        MAX_VAR_SPACE,
        MAX_VAR_PUNCT,
        MAX_VAR_SYMBOL,
        MAX_VAR_CURRENCY
    };

    CollationSettings()
            : options((UCOL_DEFAULT_STRENGTH << STRENGTH_SHIFT) |
                      (MAX_VAR_PUNCT << MAX_VARIABLE_SHIFT)),
              variableTop(0),
              reorderTable(NULL),
              minHighNoReorder(0),
              reorderRanges(NULL), reorderRangesLength(0),
              reorderCodes(NULL), reorderCodesLength(0), reorderCodesCapacity(0),
              fastLatinOptions(-1) {}

    CollationSettings(const CollationSettings &other);
    virtual ~CollationSettings();

    UBool operator==(const CollationSettings &other) const;

    inline UBool operator!=(const CollationSettings &other) const {
        return !operator==(other);
    }

    int32_t hashCode() const;

    void resetReordering();
    void aliasReordering(const CollationData &data, const int32_t *codes, int32_t length,
                         const uint32_t *ranges, int32_t rangesLength,
                         const uint8_t *table, UErrorCode &errorCode);
    void setReordering(const CollationData &data, const int32_t *codes, int32_t codesLength,
                       UErrorCode &errorCode);
    void copyReorderingFrom(const CollationSettings &other, UErrorCode &errorCode);

    inline UBool hasReordering() const { return reorderTable != NULL; }
    static UBool reorderTableHasSplitBytes(const uint8_t table[256]);
    inline uint32_t reorder(uint32_t p) const {
        uint8_t b = reorderTable[p >> 24];
        if(b != 0 || p <= Collation::NO_CE_PRIMARY) {
            return ((uint32_t)b << 24) | (p & 0xffffff);
        } else {
            return reorderEx(p);
        }
    }

    void setStrength(int32_t value, int32_t defaultOptions, UErrorCode &errorCode);

    static int32_t getStrength(int32_t options) {
        return options >> STRENGTH_SHIFT;
    }

    int32_t getStrength() const {
        return getStrength(options);
    }

    /** Sets the options bit for an on/off attribute. */
    void setFlag(int32_t bit, UColAttributeValue value,
                 int32_t defaultOptions, UErrorCode &errorCode);

    UColAttributeValue getFlag(int32_t bit) const {
        return ((options & bit) != 0) ? UCOL_ON : UCOL_OFF;
    }

    void setCaseFirst(UColAttributeValue value, int32_t defaultOptions, UErrorCode &errorCode);

    UColAttributeValue getCaseFirst() const {
        int32_t option = options & CASE_FIRST_AND_UPPER_MASK;
        return (option == 0) ? UCOL_OFF :
                (option == CASE_FIRST) ? UCOL_LOWER_FIRST : UCOL_UPPER_FIRST;
    }

    void setAlternateHandling(UColAttributeValue value,
                              int32_t defaultOptions, UErrorCode &errorCode);

    UColAttributeValue getAlternateHandling() const {
        return ((options & ALTERNATE_MASK) == 0) ? UCOL_NON_IGNORABLE : UCOL_SHIFTED;
    }

    void setMaxVariable(int32_t value, int32_t defaultOptions, UErrorCode &errorCode);

    MaxVariable getMaxVariable() const {
        return (MaxVariable)((options & MAX_VARIABLE_MASK) >> MAX_VARIABLE_SHIFT);
    }

    /**
     * Include case bits in the tertiary level if caseLevel=off and caseFirst!=off.
     */
    static inline UBool isTertiaryWithCaseBits(int32_t options) {
        return (options & (CASE_LEVEL | CASE_FIRST)) == CASE_FIRST;
    }
    static uint32_t getTertiaryMask(int32_t options) {
        // Remove the case bits from the tertiary weight when caseLevel is on or caseFirst is off.
        return isTertiaryWithCaseBits(options) ?
                Collation::CASE_AND_TERTIARY_MASK : Collation::ONLY_TERTIARY_MASK;
    }

    static UBool sortsTertiaryUpperCaseFirst(int32_t options) {
        // On tertiary level, consider case bits and sort uppercase first
        // if caseLevel is off and caseFirst==upperFirst.
        return (options & (CASE_LEVEL | CASE_FIRST_AND_UPPER_MASK)) == CASE_FIRST_AND_UPPER_MASK;
    }

    inline UBool dontCheckFCD() const {
        return (options & CHECK_FCD) == 0;
    }

    inline UBool hasBackwardSecondary() const {
        return (options & BACKWARD_SECONDARY) != 0;
    }

    inline UBool isNumeric() const {
        return (options & NUMERIC) != 0;
    }

    /** CHECK_FCD etc. */
    int32_t options;
    /** Variable-top primary weight. */
    uint32_t variableTop;
    /**
     * 256-byte table for reordering permutation of primary lead bytes; NULL if no reordering.
     * A 0 entry at a non-zero index means that the primary lead byte is "split"
     * (there are different offsets for primaries that share that lead byte)
     * and the reordering offset must be determined via the reorderRanges.
     */
    const uint8_t *reorderTable;
    /** Limit of last reordered range. 0 if no reordering or no split bytes. */
    uint32_t minHighNoReorder;
    /**
     * Primary-weight ranges for script reordering,
     * to be used by reorder(p) for split-reordered primary lead bytes.
     *
     * Each entry is a (limit, offset) pair.
     * The upper 16 bits of the entry are the upper 16 bits of the
     * exclusive primary limit of a range.
     * Primaries between the previous limit and this one have their lead bytes
     * modified by the signed offset (-0xff..+0xff) stored in the lower 16 bits.
     *
     * CollationData::makeReorderRanges() writes a full list where the first range
     * (at least for terminators and separators) has a 0 offset.
     * The last range has a non-zero offset.
     * minHighNoReorder is set to the limit of that last range.
     *
     * In the settings object, the initial ranges before the first split lead byte
     * are omitted for efficiency; they are handled by reorder(p) via the reorderTable.
     * If there are no split-reordered lead bytes, then no ranges are needed.
     */
    const uint32_t *reorderRanges;
    int32_t reorderRangesLength;
    /** Array of reorder codes; ignored if reorderCodesLength == 0. */
    const int32_t *reorderCodes;
    /** Number of reorder codes; 0 if no reordering. */
    int32_t reorderCodesLength;
    /**
     * Capacity of reorderCodes.
     * If 0, then the codes, the ranges, and the table are aliases.
     * Otherwise, this object owns the memory via the reorderCodes pointer;
     * the codes, the ranges, and the table are in the same memory block, in that order.
     */
    int32_t reorderCodesCapacity;

    /** Options for CollationFastLatin. Negative if disabled. */
    int32_t fastLatinOptions;
    uint16_t fastLatinPrimaries[0x180];

private:
    void setReorderArrays(const int32_t *codes, int32_t codesLength,
                          const uint32_t *ranges, int32_t rangesLength,
                          const uint8_t *table, UErrorCode &errorCode);
    uint32_t reorderEx(uint32_t p) const;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION




struct UDataMemory;
struct UResourceBundle;
struct UTrie2;

U_NAMESPACE_BEGIN

struct CollationData;

class UnicodeSet;

/**
 * Collation tailoring data & settings.
 * This is a container of values for a collation tailoring
 * built from rules or deserialized from binary data.
 *
 * It is logically immutable: Do not modify its values.
 * The fields are public for convenience.
 *
 * It is shared, reference-counted, and auto-deleted; see SharedObject.
 */
struct U_I18N_API CollationTailoring : public SharedObject {
    CollationTailoring(const CollationSettings *baseSettings);
    virtual ~CollationTailoring();

    /**
     * Returns TRUE if the constructor could not initialize properly.
     */
    UBool isBogus() { return settings == NULL; }

    UBool ensureOwnedData(UErrorCode &errorCode);

    static void makeBaseVersion(const UVersionInfo ucaVersion, UVersionInfo version);
    void setVersion(const UVersionInfo baseVersion, const UVersionInfo rulesVersion);
    int32_t getUCAVersion() const;

    // data for sorting etc.
    const CollationData *data;  // == base data or ownedData
    const CollationSettings *settings;  // reference-counted
    UnicodeString rules;
    // The locale is bogus when built from rules or constructed from a binary blob.
    // It can then be set by the service registration code which is thread-safe.
    mutable Locale actualLocale;
    // UCA version u.v.w & rules version r.s.t.q:
    // version[0]: builder version (runtime version is mixed in at runtime)
    // version[1]: bits 7..3=u, bits 2..0=v
    // version[2]: bits 7..6=w, bits 5..0=r
    // version[3]= (s<<5)+(s>>3)+t+(q<<4)+(q>>4)
    UVersionInfo version;

    // owned objects
    CollationData *ownedData;
    UObject *builder;
    UDataMemory *memory;
    UResourceBundle *bundle;
    UTrie2 *trie;
    UnicodeSet *unsafeBackwardSet;
    mutable UHashtable *maxExpansions;
    mutable UInitOnce maxExpansionsInitOnce;

private:
    /**
     * No copy constructor: A CollationTailoring cannot be copied.
     * It is immutable, and the data trie cannot be copied either.
     */
    CollationTailoring(const CollationTailoring &other);
};

struct CollationCacheEntry : public SharedObject {
    CollationCacheEntry(const Locale &loc, const CollationTailoring *t)
            : validLocale(loc), tailoring(t) {
        if(t != NULL) {
            t->addRef();
        }
    }
    ~CollationCacheEntry();

    Locale validLocale;
    const CollationTailoring *tailoring;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* utf16collationiterator.h
*
* created on: 2010oct27
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION







U_NAMESPACE_BEGIN

/**
 * UTF-16 collation element and character iterator.
 * Handles normalized UTF-16 text inline, with length or NUL-terminated.
 * Unnormalized text is handled by a subclass.
 */
class U_I18N_API UTF16CollationIterator : public CollationIterator {
public:
    UTF16CollationIterator(const CollationData *d, UBool numeric,
                           const UChar *s, const UChar *p, const UChar *lim)
            : CollationIterator(d, numeric),
              start(s), pos(p), limit(lim) {}

    UTF16CollationIterator(const UTF16CollationIterator &other, const UChar *newText);

    virtual ~UTF16CollationIterator();

    virtual UBool operator==(const CollationIterator &other) const;

    virtual void resetToOffset(int32_t newOffset);

    virtual int32_t getOffset() const;

    void setText(const UChar *s, const UChar *lim) {
        reset();
        start = pos = s;
        limit = lim;
    }

    virtual UChar32 nextCodePoint(UErrorCode &errorCode);

    virtual UChar32 previousCodePoint(UErrorCode &errorCode);

protected:
    // Copy constructor only for subclasses which set the pointers.
    UTF16CollationIterator(const UTF16CollationIterator &other)
            : CollationIterator(other),
              start(NULL), pos(NULL), limit(NULL) {}

    virtual uint32_t handleNextCE32(UChar32 &c, UErrorCode &errorCode);

    virtual UChar handleGetTrailSurrogate();

    virtual UBool foundNULTerminator();

    virtual void forwardNumCodePoints(int32_t num, UErrorCode &errorCode);

    virtual void backwardNumCodePoints(int32_t num, UErrorCode &errorCode);

    // UTF-16 string pointers.
    // limit can be NULL for NUL-terminated strings.
    const UChar *start, *pos, *limit;
};

/**
 * Incrementally checks the input text for FCD and normalizes where necessary.
 */
class U_I18N_API FCDUTF16CollationIterator : public UTF16CollationIterator {
public:
    FCDUTF16CollationIterator(const CollationData *data, UBool numeric,
                              const UChar *s, const UChar *p, const UChar *lim)
            : UTF16CollationIterator(data, numeric, s, p, lim),
              rawStart(s), segmentStart(p), segmentLimit(NULL), rawLimit(lim),
              nfcImpl(data->nfcImpl),
              checkDir(1) {}

    FCDUTF16CollationIterator(const FCDUTF16CollationIterator &other, const UChar *newText);

    virtual ~FCDUTF16CollationIterator();

    virtual UBool operator==(const CollationIterator &other) const;

    virtual void resetToOffset(int32_t newOffset);

    virtual int32_t getOffset() const;

    virtual UChar32 nextCodePoint(UErrorCode &errorCode);

    virtual UChar32 previousCodePoint(UErrorCode &errorCode);

protected:
    virtual uint32_t handleNextCE32(UChar32 &c, UErrorCode &errorCode);

    virtual UBool foundNULTerminator();

    virtual void forwardNumCodePoints(int32_t num, UErrorCode &errorCode);

    virtual void backwardNumCodePoints(int32_t num, UErrorCode &errorCode);

private:
    /**
     * Switches to forward checking if possible.
     * To be called when checkDir < 0 || (checkDir == 0 && pos == limit).
     * Returns with checkDir > 0 || (checkDir == 0 && pos != limit).
     */
    void switchToForward();

    /**
     * Extend the FCD text segment forward or normalize around pos.
     * To be called when checkDir > 0 && pos != limit.
     * @return TRUE if success, checkDir == 0 and pos != limit
     */
    UBool nextSegment(UErrorCode &errorCode);

    /**
     * Switches to backward checking.
     * To be called when checkDir > 0 || (checkDir == 0 && pos == start).
     * Returns with checkDir < 0 || (checkDir == 0 && pos != start).
     */
    void switchToBackward();

    /**
     * Extend the FCD text segment backward or normalize around pos.
     * To be called when checkDir < 0 && pos != start.
     * @return TRUE if success, checkDir == 0 and pos != start
     */
    UBool previousSegment(UErrorCode &errorCode);

    UBool normalize(const UChar *from, const UChar *to, UErrorCode &errorCode);

    // Text pointers: The input text is [rawStart, rawLimit[
    // where rawLimit can be NULL for NUL-terminated text.
    //
    // checkDir > 0:
    //
    // The input text [segmentStart..pos[ passes the FCD check.
    // Moving forward checks incrementally.
    // segmentLimit is undefined. limit == rawLimit.
    //
    // checkDir < 0:
    // The input text [pos..segmentLimit[ passes the FCD check.
    // Moving backward checks incrementally.
    // segmentStart is undefined, start == rawStart.
    //
    // checkDir == 0:
    //
    // The input text [segmentStart..segmentLimit[ is being processed.
    // These pointers are at FCD boundaries.
    // Either this text segment already passes the FCD check
    // and segmentStart==start<=pos<=limit==segmentLimit,
    // or the current segment had to be normalized so that
    // [segmentStart..segmentLimit[ turned into the normalized string,
    // corresponding to normalized.getBuffer()==start<=pos<=limit==start+normalized.length().
    const UChar *rawStart;
    const UChar *segmentStart;
    const UChar *segmentLimit;
    // rawLimit==NULL for a NUL-terminated string.
    const UChar *rawLimit;

    const Normalizer2Impl &nfcImpl;
    UnicodeString normalized;
    // Direction of incremental FCD check. See comments before rawStart.
    int8_t checkDir;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION



/* Constants --------------------------------------------------------------- */

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(CollationElementIterator)

/* CollationElementIterator public constructor/destructor ------------------ */

CollationElementIterator::CollationElementIterator(
                                         const CollationElementIterator& other) 
        : UObject(other), iter_(NULL), rbc_(NULL), otherHalf_(0), dir_(0), offsets_(NULL) {
    *this = other;
}

CollationElementIterator::~CollationElementIterator()
{
    delete iter_;
    delete offsets_;
}

/* CollationElementIterator public methods --------------------------------- */

namespace {

uint32_t getFirstHalf(uint32_t p, uint32_t lower32) {
    return (p & 0xffff0000) | ((lower32 >> 16) & 0xff00) | ((lower32 >> 8) & 0xff);
}
uint32_t getSecondHalf(uint32_t p, uint32_t lower32) {
    return (p << 16) | ((lower32 >> 8) & 0xff00) | (lower32 & 0x3f);
}
UBool ceNeedsTwoParts(int64_t ce) {
    return (ce & INT64_C(0xffff00ff003f)) != 0;
}

}  // namespace

int32_t CollationElementIterator::getOffset() const
{
    if (dir_ < 0 && offsets_ != NULL && !offsets_->isEmpty()) {
        // CollationIterator::previousCE() decrements the CEs length
        // while it pops CEs from its internal buffer.
        int32_t i = iter_->getCEsLength();
        if (otherHalf_ != 0) {
            // Return the trailing CE offset while we are in the middle of a 64-bit CE.
            ++i;
        }
        U_ASSERT(i < offsets_->size());
        return offsets_->elementAti(i);
    }
    return iter_->getOffset();
}

/**
* Get the ordering priority of the next character in the string.
* @return the next character's ordering. Returns NULLORDER if an error has 
*         occured or if the end of string has been reached
*/
int32_t CollationElementIterator::next(UErrorCode& status)
{
    if (U_FAILURE(status)) { return NULLORDER; }
    if (dir_ > 1) {
        // Continue forward iteration. Test this first.
        if (otherHalf_ != 0) {
            uint32_t oh = otherHalf_;
            otherHalf_ = 0;
            return oh;
        }
    } else if (dir_ == 1) {
        // next() after setOffset()
        dir_ = 2;
    } else if (dir_ == 0) {
        // The iter_ is already reset to the start of the text.
        dir_ = 2;
    } else /* dir_ < 0 */ {
        // illegal change of direction
        status = U_INVALID_STATE_ERROR;
        return NULLORDER;
    }
    // No need to keep all CEs in the buffer when we iterate.
    iter_->clearCEsIfNoneRemaining();
    int64_t ce = iter_->nextCE(status);
    if (ce == Collation::NO_CE) { return NULLORDER; }
    // Turn the 64-bit CE into two old-style 32-bit CEs, without quaternary bits.
    uint32_t p = (uint32_t)(ce >> 32);
    uint32_t lower32 = (uint32_t)ce;
    uint32_t firstHalf = getFirstHalf(p, lower32);
    uint32_t secondHalf = getSecondHalf(p, lower32);
    if (secondHalf != 0) {
        otherHalf_ = secondHalf | 0xc0;  // continuation CE
    }
    return firstHalf;
}

UBool CollationElementIterator::operator!=(
                                  const CollationElementIterator& other) const
{
    return !(*this == other);
}

UBool CollationElementIterator::operator==(
                                    const CollationElementIterator& that) const
{
    if (this == &that) {
        return TRUE;
    }

    return
        (rbc_ == that.rbc_ || *rbc_ == *that.rbc_) &&
        otherHalf_ == that.otherHalf_ &&
        normalizeDir() == that.normalizeDir() &&
        string_ == that.string_ &&
        *iter_ == *that.iter_;
}

/**
* Get the ordering priority of the previous collation element in the string.
* @param status the error code status.
* @return the previous element's ordering. Returns NULLORDER if an error has 
*         occured or if the start of string has been reached.
*/
int32_t CollationElementIterator::previous(UErrorCode& status)
{
    if (U_FAILURE(status)) { return NULLORDER; }
    if (dir_ < 0) {
        // Continue backwards iteration. Test this first.
        if (otherHalf_ != 0) {
            uint32_t oh = otherHalf_;
            otherHalf_ = 0;
            return oh;
        }
    } else if (dir_ == 0) {
        iter_->resetToOffset(string_.length());
        dir_ = -1;
    } else if (dir_ == 1) {
        // previous() after setOffset()
        dir_ = -1;
    } else /* dir_ > 1 */ {
        // illegal change of direction
        status = U_INVALID_STATE_ERROR;
        return NULLORDER;
    }
    if (offsets_ == NULL) {
        offsets_ = new UVector32(status);
        if (offsets_ == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return NULLORDER;
        }
    }
    // If we already have expansion CEs, then we also have offsets.
    // Otherwise remember the trailing offset in case we need to
    // write offsets for an artificial expansion.
    int32_t limitOffset = iter_->getCEsLength() == 0 ? iter_->getOffset() : 0;
    int64_t ce = iter_->previousCE(*offsets_, status);
    if (ce == Collation::NO_CE) { return NULLORDER; }
    // Turn the 64-bit CE into two old-style 32-bit CEs, without quaternary bits.
    uint32_t p = (uint32_t)(ce >> 32);
    uint32_t lower32 = (uint32_t)ce;
    uint32_t firstHalf = getFirstHalf(p, lower32);
    uint32_t secondHalf = getSecondHalf(p, lower32);
    if (secondHalf != 0) {
        if (offsets_->isEmpty()) {
            // When we convert a single 64-bit CE into two 32-bit CEs,
            // we need to make this artificial expansion behave like a normal expansion.
            // See CollationIterator::previousCE().
            offsets_->addElement(iter_->getOffset(), status);
            offsets_->addElement(limitOffset, status);
        }
        otherHalf_ = firstHalf;
        return secondHalf | 0xc0;  // continuation CE
    }
    return firstHalf;
}

/**
* Resets the cursor to the beginning of the string.
*/
void CollationElementIterator::reset()
{
    iter_ ->resetToOffset(0);
    otherHalf_ = 0;
    dir_ = 0;
}

void CollationElementIterator::setOffset(int32_t newOffset, 
                                         UErrorCode& status)
{
    if (U_FAILURE(status)) { return; }
    if (0 < newOffset && newOffset < string_.length()) {
        int32_t offset = newOffset;
        do {
            UChar c = string_.charAt(offset);
            if (!rbc_->isUnsafe(c) ||
                    (U16_IS_LEAD(c) && !rbc_->isUnsafe(string_.char32At(offset)))) {
                break;
            }
            // Back up to before this unsafe character.
            --offset;
        } while (offset > 0);
        if (offset < newOffset) {
            // We might have backed up more than necessary.
            // For example, contractions "ch" and "cu" make both 'h' and 'u' unsafe,
            // but for text "chu" setOffset(2) should remain at 2
            // although we initially back up to offset 0.
            // Find the last safe offset no greater than newOffset by iterating forward.
            int32_t lastSafeOffset = offset;
            do {
                iter_->resetToOffset(lastSafeOffset);
                do {
                    iter_->nextCE(status);
                    if (U_FAILURE(status)) { return; }
                } while ((offset = iter_->getOffset()) == lastSafeOffset);
                if (offset <= newOffset) {
                    lastSafeOffset = offset;
                }
            } while (offset < newOffset);
            newOffset = lastSafeOffset;
        }
    }
    iter_->resetToOffset(newOffset);
    otherHalf_ = 0;
    dir_ = 1;
}

/**
* Sets the source to the new source string.
*/
void CollationElementIterator::setText(const UnicodeString& source,
                                       UErrorCode& status)
{
    if (U_FAILURE(status)) {
        return;
    }

    string_ = source;
    const UChar *s = string_.getBuffer();
    CollationIterator *newIter;
    UBool numeric = rbc_->settings->isNumeric();
    if (rbc_->settings->dontCheckFCD()) {
        newIter = new UTF16CollationIterator(rbc_->data, numeric, s, s, s + string_.length());
    } else {
        newIter = new FCDUTF16CollationIterator(rbc_->data, numeric, s, s, s + string_.length());
    }
    if (newIter == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    delete iter_;
    iter_ = newIter;
    otherHalf_ = 0;
    dir_ = 0;
}

// Sets the source to the new character iterator.
void CollationElementIterator::setText(CharacterIterator& source, 
                                       UErrorCode& status)
{
    if (U_FAILURE(status)) 
        return;

    source.getText(string_);
    setText(string_, status);
}

int32_t CollationElementIterator::strengthOrder(int32_t order) const
{
    UColAttributeValue s = (UColAttributeValue)rbc_->settings->getStrength();
    // Mask off the unwanted differences.
    if (s == UCOL_PRIMARY) {
        order &= 0xffff0000;
    }
    else if (s == UCOL_SECONDARY) {
        order &= 0xffffff00;
    }

    return order;
}

/* CollationElementIterator private constructors/destructors --------------- */

/** 
* This is the "real" constructor for this class; it constructs an iterator
* over the source text using the specified collator
*/
CollationElementIterator::CollationElementIterator(
                                               const UnicodeString &source,
                                               const RuleBasedCollator *coll,
                                               UErrorCode &status)
        : iter_(NULL), rbc_(coll), otherHalf_(0), dir_(0), offsets_(NULL) {
    setText(source, status);
}

/** 
* This is the "real" constructor for this class; it constructs an iterator over 
* the source text using the specified collator
*/
CollationElementIterator::CollationElementIterator(
                                           const CharacterIterator &source,
                                           const RuleBasedCollator *coll,
                                           UErrorCode &status)
        : iter_(NULL), rbc_(coll), otherHalf_(0), dir_(0), offsets_(NULL) {
    // We only call source.getText() which should be const anyway.
    setText(const_cast<CharacterIterator &>(source), status);
}

/* CollationElementIterator private methods -------------------------------- */

const CollationElementIterator& CollationElementIterator::operator=(
                                         const CollationElementIterator& other)
{
    if (this == &other) {
        return *this;
    }

    CollationIterator *newIter;
    const FCDUTF16CollationIterator *otherFCDIter =
            dynamic_cast<const FCDUTF16CollationIterator *>(other.iter_);
    if(otherFCDIter != NULL) {
        newIter = new FCDUTF16CollationIterator(*otherFCDIter, string_.getBuffer());
    } else {
        const UTF16CollationIterator *otherIter =
                dynamic_cast<const UTF16CollationIterator *>(other.iter_);
        if(otherIter != NULL) {
            newIter = new UTF16CollationIterator(*otherIter, string_.getBuffer());
        } else {
            newIter = NULL;
        }
    }
    if(newIter != NULL) {
        delete iter_;
        iter_ = newIter;
        rbc_ = other.rbc_;
        otherHalf_ = other.otherHalf_;
        dir_ = other.dir_;

        string_ = other.string_;
    }
    if(other.dir_ < 0 && other.offsets_ != NULL && !other.offsets_->isEmpty()) {
        UErrorCode errorCode = U_ZERO_ERROR;
        if(offsets_ == NULL) {
            offsets_ = new UVector32(other.offsets_->size(), errorCode);
        }
        if(offsets_ != NULL) {
            offsets_->assign(*other.offsets_, errorCode);
        }
    }
    return *this;
}

namespace {

class MaxExpSink : public ContractionsAndExpansions::CESink {
public:
    MaxExpSink(UHashtable *h, UErrorCode &ec) : maxExpansions(h), errorCode(ec) {}
    virtual ~MaxExpSink();
    virtual void handleCE(int64_t /*ce*/) {}
    virtual void handleExpansion(const int64_t ces[], int32_t length) {
        if (length <= 1) {
            // We do not need to add single CEs into the map.
            return;
        }
        int32_t count = 0;  // number of CE "halves"
        for (int32_t i = 0; i < length; ++i) {
            count += ceNeedsTwoParts(ces[i]) ? 2 : 1;
        }
        // last "half" of the last CE
        int64_t ce = ces[length - 1];
        uint32_t p = (uint32_t)(ce >> 32);
        uint32_t lower32 = (uint32_t)ce;
        uint32_t lastHalf = getSecondHalf(p, lower32);
        if (lastHalf == 0) {
            lastHalf = getFirstHalf(p, lower32);
            U_ASSERT(lastHalf != 0);
        } else {
            lastHalf |= 0xc0;  // old-style continuation CE
        }
        if (count > uhash_igeti(maxExpansions, (int32_t)lastHalf)) {
            uhash_iputi(maxExpansions, (int32_t)lastHalf, count, &errorCode);
        }
    }

private:
    UHashtable *maxExpansions;
    UErrorCode &errorCode;
};

MaxExpSink::~MaxExpSink() {}

}  // namespace

UHashtable *
CollationElementIterator::computeMaxExpansions(const CollationData *data, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return NULL; }
    UHashtable *maxExpansions = uhash_open(uhash_hashLong, uhash_compareLong,
                                           uhash_compareLong, &errorCode);
    if (U_FAILURE(errorCode)) { return NULL; }
    MaxExpSink sink(maxExpansions, errorCode);
    ContractionsAndExpansions(NULL, NULL, &sink, TRUE).forData(data, errorCode);
    if (U_FAILURE(errorCode)) {
        uhash_close(maxExpansions);
        return NULL;
    }
    return maxExpansions;
}

int32_t
CollationElementIterator::getMaxExpansion(int32_t order) const {
    return getMaxExpansion(rbc_->tailoring->maxExpansions, order);
}

int32_t
CollationElementIterator::getMaxExpansion(const UHashtable *maxExpansions, int32_t order) {
    if (order == 0) { return 1; }
    int32_t max;
    if(maxExpansions != NULL && (max = uhash_igeti(maxExpansions, order)) != 0) {
        return max;
    }
    if ((order & 0xc0) == 0xc0) {
        // old-style continuation CE
        return 2;
    } else {
        return 1;
    }
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_COLLATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ******************************************************************************
 * Copyright (C) 1996-2014, International Business Machines Corporation and
 * others. All Rights Reserved.
 ******************************************************************************
 */

/**
 * File coll.cpp
 *
 * Created by: Helena Shih
 *
 * Modification History:
 *
 *  Date        Name        Description
 *  2/5/97      aliu        Modified createDefault to load collation data from
 *                          binary files when possible.  Added related methods
 *                          createCollationFromFile, chopLocale, createPathName.
 *  2/11/97     aliu        Added methods addToCache, findInCache, which implement
 *                          a Collation cache.  Modified createDefault to look in
 *                          cache first, and also to store newly created Collation
 *                          objects in the cache.  Modified to not use gLocPath.
 *  2/12/97     aliu        Modified to create objects from RuleBasedCollator cache.
 *                          Moved cache out of Collation class.
 *  2/13/97     aliu        Moved several methods out of this class and into
 *                          RuleBasedCollator, with modifications.  Modified
 *                          createDefault() to call new RuleBasedCollator(Locale&)
 *                          constructor.  General clean up and documentation.
 *  2/20/97     helena      Added clone, operator==, operator!=, operator=, and copy
 *                          constructor.
 * 05/06/97     helena      Added memory allocation error detection.
 * 05/08/97     helena      Added createInstance().
 *  6/20/97     helena      Java class name change.
 * 04/23/99     stephen     Removed EDecompositionMode, merged with
 *                          Normalizer::EMode
 * 11/23/9      srl         Inlining of some critical functions
 * 01/29/01     synwee      Modified into a C++ wrapper calling C APIs (ucol.h)
 * 2012-2014    markus      Rewritten in C++ again.
 */

  // for 'typeid' to work



#if !UCONFIG_NO_COLLATION




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationroot.h
*
* created on: 2012dec17
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION

U_NAMESPACE_BEGIN

struct CollationCacheEntry;
struct CollationData;
struct CollationSettings;
struct CollationTailoring;

/**
 * Collation root provider.
 */
class U_I18N_API CollationRoot {  // purely static
public:
    static const CollationCacheEntry *getRootCacheEntry(UErrorCode &errorCode);
    static const CollationTailoring *getRoot(UErrorCode &errorCode);
    static const CollationData *getData(UErrorCode &errorCode);
    static const CollationSettings *getSettings(UErrorCode &errorCode);

private:
    static void U_CALLCONV load(UErrorCode &errorCode);
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1998-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*
* Private implementation header for C collation
*   file name:  ucol_imp.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2000dec11
*   created by: Vladimir Weinstein
*
* Modification history
* Date        Name      Comments
* 02/16/2001  synwee    Added UCOL_GETPREVCE for the use in ucoleitr
* 02/27/2001  synwee    Added getMaxExpansion data structure in UCollator
* 03/02/2001  synwee    Added UCOL_IMPLICIT_CE
* 03/12/2001  synwee    Added pointer start to collIterate.
*/




#if !UCONFIG_NO_COLLATION

// This part needs to compile as plain C code, for cintltst.



/** Check whether two collators are equal. Collators are considered equal if they
 *  will sort strings the same. This means that both the current attributes and the
 *  rules must be equivalent.
 *  @param source first collator
 *  @param target second collator
 *  @return TRUE or FALSE
 *  @internal ICU 3.0
 */
U_INTERNAL UBool U_EXPORT2
ucol_equals(const UCollator *source, const UCollator *target);

/**
 * Convenience string denoting the Collation data tree
 */
#define U_ICUDATA_COLL U_ICUDATA_NAME U_TREE_SEPARATOR_STRING "coll"

#ifdef __cplusplus




U_NAMESPACE_BEGIN

struct CollationCacheEntry;

class Locale;
class UnicodeString;
class UnifiedCache;

/** Implemented in ucol_res.cpp. */
class CollationLoader {
public:
    static void appendRootRules(UnicodeString &s);
    static void loadRules(const char *localeID, const char *collationType,
                          UnicodeString &rules, UErrorCode &errorCode);
    // Adds a reference to returned value.
    static const CollationCacheEntry *loadTailoring(const Locale &locale, UErrorCode &errorCode);

    // Cache callback. Adds a reference to returned value.
    const CollationCacheEntry *createCacheEntry(UErrorCode &errorCode);

private:
    static void U_CALLCONV loadRootRules(UErrorCode &errorCode);

    // The following members are used by loadTailoring()
    // and the cache callback.
    static const uint32_t TRIED_SEARCH = 1;
    static const uint32_t TRIED_DEFAULT = 2;
    static const uint32_t TRIED_STANDARD = 4;

    CollationLoader(const CollationCacheEntry *re, const Locale &requested, UErrorCode &errorCode);
    ~CollationLoader();

    // All loadFromXXX methods add a reference to the returned value.
    const CollationCacheEntry *loadFromLocale(UErrorCode &errorCode);
    const CollationCacheEntry *loadFromBundle(UErrorCode &errorCode);
    const CollationCacheEntry *loadFromCollations(UErrorCode &errorCode);
    const CollationCacheEntry *loadFromData(UErrorCode &errorCode);

    // Adds a reference to returned value.
    const CollationCacheEntry *getCacheEntry(UErrorCode &errorCode);

    /**
     * Returns the rootEntry (with one addRef()) if loc==root,
     * or else returns a new cache entry with ref count 1 for the loc and
     * the root tailoring.
     */
    const CollationCacheEntry *makeCacheEntryFromRoot(
            const Locale &loc, UErrorCode &errorCode) const;

    /**
     * Returns the entryFromCache as is if loc==validLocale,
     * or else returns a new cache entry with ref count 1 for the loc and
     * the same tailoring. In the latter case, a ref count is removed from
     * entryFromCache.
     */
    static const CollationCacheEntry *makeCacheEntry(
            const Locale &loc,
            const CollationCacheEntry *entryFromCache,
            UErrorCode &errorCode);

    const UnifiedCache *cache;
    const CollationCacheEntry *rootEntry;
    Locale validLocale;
    Locale locale;
    char type[16];
    char defaultType[16];
    uint32_t typesTried;
    UBool typeFallback;
    UResourceBundle *bundle;
    UResourceBundle *collations;
    UResourceBundle *data;
};

U_NAMESPACE_END

#endif  /* __cplusplus */

#endif /* #if !UCONFIG_NO_COLLATION */











static icu::Locale* availableLocaleList = NULL;
static int32_t  availableLocaleListCount;
#if !UCONFIG_NO_SERVICE
static icu::ICULocaleService* gService = NULL;
static icu::UInitOnce gServiceInitOnce = U_INITONCE_INITIALIZER;
#endif
static icu::UInitOnce gAvailableLocaleListInitOnce = U_INITONCE_INITIALIZER;

/**
 * Release all static memory held by collator.
 */
U_CDECL_BEGIN
static UBool U_CALLCONV collator_cleanup(void) {
#if !UCONFIG_NO_SERVICE
    if (gService) {
        delete gService;
        gService = NULL;
    }
    gServiceInitOnce.reset();
#endif
    if (availableLocaleList) {
        delete []availableLocaleList;
        availableLocaleList = NULL;
    }
    availableLocaleListCount = 0;
    gAvailableLocaleListInitOnce.reset();
    return TRUE;
}

U_CDECL_END

U_NAMESPACE_BEGIN

#if !UCONFIG_NO_SERVICE

// ------------------------------------------
//
// Registration
//

//-------------------------------------------

CollatorFactory::~CollatorFactory() {}

//-------------------------------------------

UBool
CollatorFactory::visible(void) const {
    return TRUE;
}

//-------------------------------------------

UnicodeString&
CollatorFactory::getDisplayName(const Locale& objectLocale,
                                const Locale& displayLocale,
                                UnicodeString& result)
{
  return objectLocale.getDisplayName(displayLocale, result);
}

// -------------------------------------

class ICUCollatorFactory : public ICUResourceBundleFactory {
 public:
    ICUCollatorFactory() : ICUResourceBundleFactory(UnicodeString(U_ICUDATA_COLL, -1, US_INV)) { }
    virtual ~ICUCollatorFactory();
 protected:
    virtual UObject* create(const ICUServiceKey& key, const ICUService* service, UErrorCode& status) const;
};

ICUCollatorFactory::~ICUCollatorFactory() {}

UObject*
ICUCollatorFactory::create(const ICUServiceKey& key, const ICUService* /* service */, UErrorCode& status) const {
    if (handlesKey(key, status)) {
        const LocaleKey& lkey = (const LocaleKey&)key;
        Locale loc;
        // make sure the requested locale is correct
        // default LocaleFactory uses currentLocale since that's the one vetted by handlesKey
        // but for ICU rb resources we use the actual one since it will fallback again
        lkey.canonicalLocale(loc);

        return Collator::makeInstance(loc, status);
    }
    return NULL;
}

// -------------------------------------

class ICUCollatorService : public ICULocaleService {
public:
    ICUCollatorService()
        : ICULocaleService(UNICODE_STRING_SIMPLE("Collator"))
    {
        UErrorCode status = U_ZERO_ERROR;
        registerFactory(new ICUCollatorFactory(), status);
    }

    virtual ~ICUCollatorService();

    virtual UObject* cloneInstance(UObject* instance) const {
        return ((Collator*)instance)->clone();
    }

    virtual UObject* handleDefault(const ICUServiceKey& key, UnicodeString* actualID, UErrorCode& status) const {
        LocaleKey& lkey = (LocaleKey&)key;
        if (actualID) {
            // Ugly Hack Alert! We return an empty actualID to signal
            // to callers that this is a default object, not a "real"
            // service-created object. (TODO remove in 3.0) [aliu]
            actualID->truncate(0);
        }
        Locale loc("");
        lkey.canonicalLocale(loc);
        return Collator::makeInstance(loc, status);
    }

    virtual UObject* getKey(ICUServiceKey& key, UnicodeString* actualReturn, UErrorCode& status) const {
        UnicodeString ar;
        if (actualReturn == NULL) {
            actualReturn = &ar;
        }
        return (Collator*)ICULocaleService::getKey(key, actualReturn, status);
    }

    virtual UBool isDefault() const {
        return countFactories() == 1;
    }
};

ICUCollatorService::~ICUCollatorService() {}

// -------------------------------------

static void U_CALLCONV initService() {
    gService = new ICUCollatorService();
    ucln_i18n_registerCleanup(UCLN_I18N_COLLATOR, collator_cleanup);
}


static ICULocaleService*
getService(void)
{
    umtx_initOnce(gServiceInitOnce, &initService);
    return gService;
}

// -------------------------------------

static inline UBool
hasService(void)
{
    UBool retVal = !gServiceInitOnce.isReset() && (getService() != NULL);
    return retVal;
}

#endif /* UCONFIG_NO_SERVICE */

static void U_CALLCONV
initAvailableLocaleList(UErrorCode &status) {
    U_ASSERT(availableLocaleListCount == 0);
    U_ASSERT(availableLocaleList == NULL);
    // for now, there is a hardcoded list, so just walk through that list and set it up.
    UResourceBundle *index = NULL;
    StackUResourceBundle installed;
    int32_t i = 0;

    index = ures_openDirect(U_ICUDATA_COLL, "res_index", &status);
    ures_getByKey(index, "InstalledLocales", installed.getAlias(), &status);

    if(U_SUCCESS(status)) {
        availableLocaleListCount = ures_getSize(installed.getAlias());
        availableLocaleList = new Locale[availableLocaleListCount];

        if (availableLocaleList != NULL) {
            ures_resetIterator(installed.getAlias());
            while(ures_hasNext(installed.getAlias())) {
                const char *tempKey = NULL;
                ures_getNextString(installed.getAlias(), NULL, &tempKey, &status);
                availableLocaleList[i++] = Locale(tempKey);
            }
        }
        U_ASSERT(availableLocaleListCount == i);
    }
    ures_close(index);
    ucln_i18n_registerCleanup(UCLN_I18N_COLLATOR, collator_cleanup);
}

static UBool isAvailableLocaleListInitialized(UErrorCode &status) {
    umtx_initOnce(gAvailableLocaleListInitOnce, &initAvailableLocaleList, status);
    return U_SUCCESS(status);
}


// Collator public methods -----------------------------------------------

namespace {

static const struct {
    const char *name;
    UColAttribute attr;
} collAttributes[] = {
    { "colStrength", UCOL_STRENGTH },
    { "colBackwards", UCOL_FRENCH_COLLATION },
    { "colCaseLevel", UCOL_CASE_LEVEL },
    { "colCaseFirst", UCOL_CASE_FIRST },
    { "colAlternate", UCOL_ALTERNATE_HANDLING },
    { "colNormalization", UCOL_NORMALIZATION_MODE },
    { "colNumeric", UCOL_NUMERIC_COLLATION }
};

static const struct {
    const char *name;
    UColAttributeValue value;
} collAttributeValues[] = {
    { "primary", UCOL_PRIMARY },
    { "secondary", UCOL_SECONDARY },
    { "tertiary", UCOL_TERTIARY },
    { "quaternary", UCOL_QUATERNARY },
    // Note: Not supporting typo "quarternary" because it was never supported in locale IDs.
    { "identical", UCOL_IDENTICAL },
    { "no", UCOL_OFF },
    { "yes", UCOL_ON },
    { "shifted", UCOL_SHIFTED },
    { "non-ignorable", UCOL_NON_IGNORABLE },
    { "lower", UCOL_LOWER_FIRST },
    { "upper", UCOL_UPPER_FIRST }
};

static const char *collReorderCodes[UCOL_REORDER_CODE_LIMIT - UCOL_REORDER_CODE_FIRST] = {
    "space", "punct", "symbol", "currency", "digit"
};

int32_t getReorderCode(const char *s) {
    for (int32_t i = 0; i < UPRV_LENGTHOF(collReorderCodes); ++i) {
        if (uprv_stricmp(s, collReorderCodes[i]) == 0) {
            return UCOL_REORDER_CODE_FIRST + i;
        }
    }
    // Not supporting "others" = UCOL_REORDER_CODE_OTHERS
    // as a synonym for Zzzz = USCRIPT_UNKNOWN for now:
    // Avoid introducing synonyms/aliases.
    return -1;
}

/**
 * Sets collation attributes according to locale keywords. See
 * http://www.unicode.org/reports/tr35/tr35-collation.html#Collation_Settings
 *
 * Using "alias" keywords and values where defined:
 * http://www.unicode.org/reports/tr35/tr35.html#Old_Locale_Extension_Syntax
 * http://unicode.org/repos/cldr/trunk/common/bcp47/collation.xml
 */
void setAttributesFromKeywords(const Locale &loc, Collator &coll, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) {
        return;
    }
    if (uprv_strcmp(loc.getName(), loc.getBaseName()) == 0) {
        // No keywords.
        return;
    }
    char value[1024];  // The reordering value could be long.
    // Check for collation keywords that were already deprecated
    // before any were supported in createInstance() (except for "collation").
    int32_t length = loc.getKeywordValue("colHiraganaQuaternary", value, UPRV_LENGTHOF(value), errorCode);
    if (U_FAILURE(errorCode)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if (length != 0) {
        errorCode = U_UNSUPPORTED_ERROR;
        return;
    }
    length = loc.getKeywordValue("variableTop", value, UPRV_LENGTHOF(value), errorCode);
    if (U_FAILURE(errorCode)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if (length != 0) {
        errorCode = U_UNSUPPORTED_ERROR;
        return;
    }
    // Parse known collation keywords, ignore others.
    if (errorCode == U_STRING_NOT_TERMINATED_WARNING) {
        errorCode = U_ZERO_ERROR;
    }
    for (int32_t i = 0; i < UPRV_LENGTHOF(collAttributes); ++i) {
        length = loc.getKeywordValue(collAttributes[i].name, value, UPRV_LENGTHOF(value), errorCode);
        if (U_FAILURE(errorCode) || errorCode == U_STRING_NOT_TERMINATED_WARNING) {
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        if (length == 0) { continue; }
        for (int32_t j = 0;; ++j) {
            if (j == UPRV_LENGTHOF(collAttributeValues)) {
                errorCode = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
            if (uprv_stricmp(value, collAttributeValues[j].name) == 0) {
                coll.setAttribute(collAttributes[i].attr, collAttributeValues[j].value, errorCode);
                break;
            }
        }
    }
    length = loc.getKeywordValue("colReorder", value, UPRV_LENGTHOF(value), errorCode);
    if (U_FAILURE(errorCode) || errorCode == U_STRING_NOT_TERMINATED_WARNING) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if (length != 0) {
        int32_t codes[USCRIPT_CODE_LIMIT + UCOL_REORDER_CODE_LIMIT - UCOL_REORDER_CODE_FIRST];
        int32_t codesLength = 0;
        char *scriptName = value;
        for (;;) {
            if (codesLength == UPRV_LENGTHOF(codes)) {
                errorCode = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
            char *limit = scriptName;
            char c;
            while ((c = *limit) != 0 && c != '-') { ++limit; }
            *limit = 0;
            int32_t code;
            if ((limit - scriptName) == 4) {
                // Strict parsing, accept only 4-letter script codes, not long names.
                code = u_getPropertyValueEnum(UCHAR_SCRIPT, scriptName);
            } else {
                code = getReorderCode(scriptName);
            }
            if (code < 0) {
                errorCode = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
            codes[codesLength++] = code;
            if (c == 0) { break; }
            scriptName = limit + 1;
        }
        coll.setReorderCodes(codes, codesLength, errorCode);
    }
    length = loc.getKeywordValue("kv", value, UPRV_LENGTHOF(value), errorCode);
    if (U_FAILURE(errorCode) || errorCode == U_STRING_NOT_TERMINATED_WARNING) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if (length != 0) {
        int32_t code = getReorderCode(value);
        if (code < 0) {
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        coll.setMaxVariable((UColReorderCode)code, errorCode);
    }
    if (U_FAILURE(errorCode)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

}  // namespace

Collator* U_EXPORT2 Collator::createInstance(UErrorCode& success)
{
    return createInstance(Locale::getDefault(), success);
}

Collator* U_EXPORT2 Collator::createInstance(const Locale& desiredLocale,
                                   UErrorCode& status)
{
    if (U_FAILURE(status))
        return 0;
    if (desiredLocale.isBogus()) {
        // Locale constructed from malformed locale ID or language tag.
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    Collator* coll;
#if !UCONFIG_NO_SERVICE
    if (hasService()) {
        Locale actualLoc;
        coll = (Collator*)gService->get(desiredLocale, &actualLoc, status);
    } else
#endif
    {
        coll = makeInstance(desiredLocale, status);
        // Either returns NULL with U_FAILURE(status), or non-NULL with U_SUCCESS(status)
    }
    // The use of *coll in setAttributesFromKeywords can cause the NULL check to be
    // optimized out of the delete even though setAttributesFromKeywords returns
    // immediately if U_FAILURE(status), so we add a check here.
    if (U_FAILURE(status)) {
        return NULL;
    }
    setAttributesFromKeywords(desiredLocale, *coll, status);
    if (U_FAILURE(status)) {
        delete coll;
        return NULL;
    }
    return coll;
}


Collator* Collator::makeInstance(const Locale&  desiredLocale, UErrorCode& status) {
    const CollationCacheEntry *entry = CollationLoader::loadTailoring(desiredLocale, status);
    if (U_SUCCESS(status)) {
        Collator *result = new RuleBasedCollator(entry);
        if (result != NULL) {
            // Both the unified cache's get() and the RBC constructor
            // did addRef(). Undo one of them.
            entry->removeRef();
            return result;
        }
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    if (entry != NULL) {
        // Undo the addRef() from the cache.get().
        entry->removeRef();
    }
    return NULL;
}

Collator *
Collator::safeClone() const {
    return clone();
}

// implement deprecated, previously abstract method
Collator::EComparisonResult Collator::compare(const UnicodeString& source,
                                    const UnicodeString& target) const
{
    UErrorCode ec = U_ZERO_ERROR;
    return (EComparisonResult)compare(source, target, ec);
}

// implement deprecated, previously abstract method
Collator::EComparisonResult Collator::compare(const UnicodeString& source,
                                    const UnicodeString& target,
                                    int32_t length) const
{
    UErrorCode ec = U_ZERO_ERROR;
    return (EComparisonResult)compare(source, target, length, ec);
}

// implement deprecated, previously abstract method
Collator::EComparisonResult Collator::compare(const UChar* source, int32_t sourceLength,
                                    const UChar* target, int32_t targetLength)
                                    const
{
    UErrorCode ec = U_ZERO_ERROR;
    return (EComparisonResult)compare(source, sourceLength, target, targetLength, ec);
}

UCollationResult Collator::compare(UCharIterator &/*sIter*/,
                                   UCharIterator &/*tIter*/,
                                   UErrorCode &status) const {
    if(U_SUCCESS(status)) {
        // Not implemented in the base class.
        status = U_UNSUPPORTED_ERROR;
    }
    return UCOL_EQUAL;
}

UCollationResult Collator::compareUTF8(const StringPiece &source,
                                       const StringPiece &target,
                                       UErrorCode &status) const {
    if(U_FAILURE(status)) {
        return UCOL_EQUAL;
    }
    UCharIterator sIter, tIter;
    uiter_setUTF8(&sIter, source.data(), source.length());
    uiter_setUTF8(&tIter, target.data(), target.length());
    return compare(sIter, tIter, status);
}

UBool Collator::equals(const UnicodeString& source,
                       const UnicodeString& target) const
{
    UErrorCode ec = U_ZERO_ERROR;
    return (compare(source, target, ec) == UCOL_EQUAL);
}

UBool Collator::greaterOrEqual(const UnicodeString& source,
                               const UnicodeString& target) const
{
    UErrorCode ec = U_ZERO_ERROR;
    return (compare(source, target, ec) != UCOL_LESS);
}

UBool Collator::greater(const UnicodeString& source,
                        const UnicodeString& target) const
{
    UErrorCode ec = U_ZERO_ERROR;
    return (compare(source, target, ec) == UCOL_GREATER);
}

// this API  ignores registered collators, since it returns an
// array of indefinite lifetime
const Locale* U_EXPORT2 Collator::getAvailableLocales(int32_t& count)
{
    UErrorCode status = U_ZERO_ERROR;
    Locale *result = NULL;
    count = 0;
    if (isAvailableLocaleListInitialized(status))
    {
        result = availableLocaleList;
        count = availableLocaleListCount;
    }
    return result;
}

UnicodeString& U_EXPORT2 Collator::getDisplayName(const Locale& objectLocale,
                                        const Locale& displayLocale,
                                        UnicodeString& name)
{
#if !UCONFIG_NO_SERVICE
    if (hasService()) {
        UnicodeString locNameStr;
        LocaleUtility::initNameFromLocale(objectLocale, locNameStr);
        return gService->getDisplayName(locNameStr, name, displayLocale);
    }
#endif
    return objectLocale.getDisplayName(displayLocale, name);
}

UnicodeString& U_EXPORT2 Collator::getDisplayName(const Locale& objectLocale,
                                        UnicodeString& name)
{
    return getDisplayName(objectLocale, Locale::getDefault(), name);
}

/* This is useless information */
/*void Collator::getVersion(UVersionInfo versionInfo) const
{
  if (versionInfo!=NULL)
    uprv_memcpy(versionInfo, fVersion, U_MAX_VERSION_LENGTH);
}
*/

// UCollator protected constructor destructor ----------------------------

/**
* Default constructor.
* Constructor is different from the old default Collator constructor.
* The task for determing the default collation strength and normalization mode
* is left to the child class.
*/
Collator::Collator()
: UObject()
{
}

/**
* Constructor.
* Empty constructor, does not handle the arguments.
* This constructor is done for backward compatibility with 1.7 and 1.8.
* The task for handling the argument collation strength and normalization
* mode is left to the child class.
* @param collationStrength collation strength
* @param decompositionMode
* @deprecated 2.4 use the default constructor instead
*/
Collator::Collator(UCollationStrength, UNormalizationMode )
: UObject()
{
}

Collator::~Collator()
{
}

Collator::Collator(const Collator &other)
    : UObject(other)
{
}

UBool Collator::operator==(const Collator& other) const
{
    // Subclasses: Call this method and then add more specific checks.
    return typeid(*this) == typeid(other);
}

UBool Collator::operator!=(const Collator& other) const
{
    return (UBool)!(*this == other);
}

int32_t U_EXPORT2 Collator::getBound(const uint8_t       *source,
                           int32_t             sourceLength,
                           UColBoundMode       boundType,
                           uint32_t            noOfLevels,
                           uint8_t             *result,
                           int32_t             resultLength,
                           UErrorCode          &status)
{
    return ucol_getBound(source, sourceLength, boundType, noOfLevels, result, resultLength, &status);
}

void
Collator::setLocales(const Locale& /* requestedLocale */, const Locale& /* validLocale */, const Locale& /*actualLocale*/) {
}

UnicodeSet *Collator::getTailoredSet(UErrorCode &status) const
{
    if(U_FAILURE(status)) {
        return NULL;
    }
    // everything can be changed
    return new UnicodeSet(0, 0x10FFFF);
}

// -------------------------------------

#if !UCONFIG_NO_SERVICE
URegistryKey U_EXPORT2
Collator::registerInstance(Collator* toAdopt, const Locale& locale, UErrorCode& status)
{
    if (U_SUCCESS(status)) {
        // Set the collator locales while registering so that createInstance()
        // need not guess whether the collator's locales are already set properly
        // (as they are by the data loader).
        toAdopt->setLocales(locale, locale, locale);
        return getService()->registerInstance(toAdopt, locale, status);
    }
    return NULL;
}

// -------------------------------------

class CFactory : public LocaleKeyFactory {
private:
    CollatorFactory* _delegate;
    Hashtable* _ids;

public:
    CFactory(CollatorFactory* delegate, UErrorCode& status)
        : LocaleKeyFactory(delegate->visible() ? VISIBLE : INVISIBLE)
        , _delegate(delegate)
        , _ids(NULL)
    {
        if (U_SUCCESS(status)) {
            int32_t count = 0;
            _ids = new Hashtable(status);
            if (_ids) {
                const UnicodeString * idlist = _delegate->getSupportedIDs(count, status);
                for (int i = 0; i < count; ++i) {
                    _ids->put(idlist[i], (void*)this, status);
                    if (U_FAILURE(status)) {
                        delete _ids;
                        _ids = NULL;
                        return;
                    }
                }
            } else {
                status = U_MEMORY_ALLOCATION_ERROR;
            }
        }
    }

    virtual ~CFactory();

    virtual UObject* create(const ICUServiceKey& key, const ICUService* service, UErrorCode& status) const;

protected:
    virtual const Hashtable* getSupportedIDs(UErrorCode& status) const
    {
        if (U_SUCCESS(status)) {
            return _ids;
        }
        return NULL;
    }

    virtual UnicodeString&
        getDisplayName(const UnicodeString& id, const Locale& locale, UnicodeString& result) const;
};

CFactory::~CFactory()
{
    delete _delegate;
    delete _ids;
}

UObject*
CFactory::create(const ICUServiceKey& key, const ICUService* /* service */, UErrorCode& status) const
{
    if (handlesKey(key, status)) {
        const LocaleKey& lkey = (const LocaleKey&)key;
        Locale validLoc;
        lkey.currentLocale(validLoc);
        return _delegate->createCollator(validLoc);
    }
    return NULL;
}

UnicodeString&
CFactory::getDisplayName(const UnicodeString& id, const Locale& locale, UnicodeString& result) const
{
    if ((_coverage & 0x1) == 0) {
        UErrorCode status = U_ZERO_ERROR;
        const Hashtable* ids = getSupportedIDs(status);
        if (ids && (ids->get(id) != NULL)) {
            Locale loc;
            LocaleUtility::initLocaleFromName(id, loc);
            return _delegate->getDisplayName(loc, locale, result);
        }
    }
    result.setToBogus();
    return result;
}

URegistryKey U_EXPORT2
Collator::registerFactory(CollatorFactory* toAdopt, UErrorCode& status)
{
    if (U_SUCCESS(status)) {
        CFactory* f = new CFactory(toAdopt, status);
        if (f) {
            return getService()->registerFactory(f, status);
        }
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return NULL;
}

// -------------------------------------

UBool U_EXPORT2
Collator::unregister(URegistryKey key, UErrorCode& status)
{
    if (U_SUCCESS(status)) {
        if (hasService()) {
            return gService->unregister(key, status);
        }
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return FALSE;
}
#endif /* UCONFIG_NO_SERVICE */

class CollationLocaleListEnumeration : public StringEnumeration {
private:
    int32_t index;
public:
    static UClassID U_EXPORT2 getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
public:
    CollationLocaleListEnumeration()
        : index(0)
    {
        // The global variables should already be initialized.
        //isAvailableLocaleListInitialized(status);
    }

    virtual ~CollationLocaleListEnumeration();

    virtual StringEnumeration * clone() const
    {
        CollationLocaleListEnumeration *result = new CollationLocaleListEnumeration();
        if (result) {
            result->index = index;
        }
        return result;
    }

    virtual int32_t count(UErrorCode &/*status*/) const {
        return availableLocaleListCount;
    }

    virtual const char* next(int32_t* resultLength, UErrorCode& /*status*/) {
        const char* result;
        if(index < availableLocaleListCount) {
            result = availableLocaleList[index++].getName();
            if(resultLength != NULL) {
                *resultLength = (int32_t)uprv_strlen(result);
            }
        } else {
            if(resultLength != NULL) {
                *resultLength = 0;
            }
            result = NULL;
        }
        return result;
    }

    virtual const UnicodeString* snext(UErrorCode& status) {
        int32_t resultLength = 0;
        const char *s = next(&resultLength, status);
        return setChars(s, resultLength, status);
    }

    virtual void reset(UErrorCode& /*status*/) {
        index = 0;
    }
};

CollationLocaleListEnumeration::~CollationLocaleListEnumeration() {}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(CollationLocaleListEnumeration)


// -------------------------------------

StringEnumeration* U_EXPORT2
Collator::getAvailableLocales(void)
{
#if !UCONFIG_NO_SERVICE
    if (hasService()) {
        return getService()->getAvailableLocales();
    }
#endif /* UCONFIG_NO_SERVICE */
    UErrorCode status = U_ZERO_ERROR;
    if (isAvailableLocaleListInitialized(status)) {
        return new CollationLocaleListEnumeration();
    }
    return NULL;
}

StringEnumeration* U_EXPORT2
Collator::getKeywords(UErrorCode& status) {
    return UStringEnumeration::fromUEnumeration(
            ucol_getKeywords(&status), status);
}

StringEnumeration* U_EXPORT2
Collator::getKeywordValues(const char *keyword, UErrorCode& status) {
    return UStringEnumeration::fromUEnumeration(
            ucol_getKeywordValues(keyword, &status), status);
}

StringEnumeration* U_EXPORT2
Collator::getKeywordValuesForLocale(const char* key, const Locale& locale,
                                    UBool commonlyUsed, UErrorCode& status) {
    return UStringEnumeration::fromUEnumeration(
            ucol_getKeywordValuesForLocale(
                    key, locale.getName(), commonlyUsed, &status),
            status);
}

Locale U_EXPORT2
Collator::getFunctionalEquivalent(const char* keyword, const Locale& locale,
                                  UBool& isAvailable, UErrorCode& status) {
    // This is a wrapper over ucol_getFunctionalEquivalent
    char loc[ULOC_FULLNAME_CAPACITY];
    /*int32_t len =*/ ucol_getFunctionalEquivalent(loc, sizeof(loc),
                    keyword, locale.getName(), &isAvailable, &status);
    if (U_FAILURE(status)) {
        *loc = 0; // root
    }
    return Locale::createFromName(loc);
}

Collator::ECollationStrength
Collator::getStrength(void) const {
    UErrorCode intStatus = U_ZERO_ERROR;
    return (ECollationStrength)getAttribute(UCOL_STRENGTH, intStatus);
}

void
Collator::setStrength(ECollationStrength newStrength) {
    UErrorCode intStatus = U_ZERO_ERROR;
    setAttribute(UCOL_STRENGTH, (UColAttributeValue)newStrength, intStatus);
}

Collator &
Collator::setMaxVariable(UColReorderCode /*group*/, UErrorCode &errorCode) {
    if (U_SUCCESS(errorCode)) {
        errorCode = U_UNSUPPORTED_ERROR;
    }
    return *this;
}

UColReorderCode
Collator::getMaxVariable() const {
    return UCOL_REORDER_CODE_PUNCTUATION;
}

int32_t
Collator::getReorderCodes(int32_t* /* dest*/,
                          int32_t /* destCapacity*/,
                          UErrorCode& status) const
{
    if (U_SUCCESS(status)) {
        status = U_UNSUPPORTED_ERROR;
    }
    return 0;
}

void
Collator::setReorderCodes(const int32_t* /* reorderCodes */,
                          int32_t /* reorderCodesLength */,
                          UErrorCode& status)
{
    if (U_SUCCESS(status)) {
        status = U_UNSUPPORTED_ERROR;
    }
}

int32_t
Collator::getEquivalentReorderCodes(int32_t reorderCode,
                                    int32_t *dest, int32_t capacity,
                                    UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    if(capacity < 0 || (dest == NULL && capacity > 0)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    const CollationData *baseData = CollationRoot::getData(errorCode);
    if(U_FAILURE(errorCode)) { return 0; }
    return baseData->getEquivalentScripts(reorderCode, dest, capacity, errorCode);
}

int32_t
Collator::internalGetShortDefinitionString(const char * /*locale*/,
                                                             char * /*buffer*/,
                                                             int32_t /*capacity*/,
                                                             UErrorCode &status) const {
  if(U_SUCCESS(status)) {
    status = U_UNSUPPORTED_ERROR; /* Shouldn't happen, internal function */
  }
  return 0;
}

UCollationResult
Collator::internalCompareUTF8(const char *left, int32_t leftLength,
                              const char *right, int32_t rightLength,
                              UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return UCOL_EQUAL; }
    if((left == NULL && leftLength != 0) || (right == NULL && rightLength != 0)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return UCOL_EQUAL;
    }
    return compareUTF8(
            StringPiece(left, (leftLength < 0) ? static_cast<int32_t>(uprv_strlen(left)) : leftLength),
            StringPiece(right, (rightLength < 0) ? static_cast<int32_t>(uprv_strlen(right)) : rightLength),
            errorCode);
}

int32_t
Collator::internalNextSortKeyPart(UCharIterator * /*iter*/, uint32_t /*state*/[2],
                                  uint8_t * /*dest*/, int32_t /*count*/, UErrorCode &errorCode) const {
    if (U_SUCCESS(errorCode)) {
        errorCode = U_UNSUPPORTED_ERROR;
    }
    return 0;
}

// UCollator private data members ----------------------------------------

/* This is useless information */
/*const UVersionInfo Collator::fVersion = {1, 1, 0, 0};*/

// -------------------------------------

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_COLLATION */

/* eof */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collation.cpp
*
* created on: 2010oct27
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION




U_NAMESPACE_BEGIN

// Some compilers don't care if constants are defined in the .cpp file.
// MS Visual C++ does not like it, but gcc requires it. clang does not care.
#ifndef _MSC_VER
const uint8_t Collation::LEVEL_SEPARATOR_BYTE;
const uint8_t Collation::MERGE_SEPARATOR_BYTE;
const uint32_t Collation::ONLY_TERTIARY_MASK;
const uint32_t Collation::CASE_AND_TERTIARY_MASK;
#endif

uint32_t
Collation::incTwoBytePrimaryByOffset(uint32_t basePrimary, UBool isCompressible, int32_t offset) {
    // Extract the second byte, minus the minimum byte value,
    // plus the offset, modulo the number of usable byte values, plus the minimum.
    // Reserve the PRIMARY_COMPRESSION_LOW_BYTE and high byte if necessary.
    uint32_t primary;
    if(isCompressible) {
        offset += ((int32_t)(basePrimary >> 16) & 0xff) - 4;
        primary = (uint32_t)((offset % 251) + 4) << 16;
        offset /= 251;
    } else {
        offset += ((int32_t)(basePrimary >> 16) & 0xff) - 2;
        primary = (uint32_t)((offset % 254) + 2) << 16;
        offset /= 254;
    }
    // First byte, assume no further overflow.
    return primary | ((basePrimary & 0xff000000) + (uint32_t)(offset << 24));
}

uint32_t
Collation::incThreeBytePrimaryByOffset(uint32_t basePrimary, UBool isCompressible, int32_t offset) {
    // Extract the third byte, minus the minimum byte value,
    // plus the offset, modulo the number of usable byte values, plus the minimum.
    offset += ((int32_t)(basePrimary >> 8) & 0xff) - 2;
    uint32_t primary = (uint32_t)((offset % 254) + 2) << 8;
    offset /= 254;
    // Same with the second byte,
    // but reserve the PRIMARY_COMPRESSION_LOW_BYTE and high byte if necessary.
    if(isCompressible) {
        offset += ((int32_t)(basePrimary >> 16) & 0xff) - 4;
        primary |= (uint32_t)((offset % 251) + 4) << 16;
        offset /= 251;
    } else {
        offset += ((int32_t)(basePrimary >> 16) & 0xff) - 2;
        primary |= (uint32_t)((offset % 254) + 2) << 16;
        offset /= 254;
    }
    // First byte, assume no further overflow.
    return primary | ((basePrimary & 0xff000000) + (uint32_t)(offset << 24));
}

uint32_t
Collation::decTwoBytePrimaryByOneStep(uint32_t basePrimary, UBool isCompressible, int32_t step) {
    // Extract the second byte, minus the minimum byte value,
    // minus the step, modulo the number of usable byte values, plus the minimum.
    // Reserve the PRIMARY_COMPRESSION_LOW_BYTE and high byte if necessary.
    // Assume no further underflow for the first byte.
    U_ASSERT(0 < step && step <= 0x7f);
    int32_t byte2 = ((int32_t)(basePrimary >> 16) & 0xff) - step;
    if(isCompressible) {
        if(byte2 < 4) {
            byte2 += 251;
            basePrimary -= 0x1000000;
        }
    } else {
        if(byte2 < 2) {
            byte2 += 254;
            basePrimary -= 0x1000000;
        }
    }
    return (basePrimary & 0xff000000) | ((uint32_t)byte2 << 16);
}

uint32_t
Collation::decThreeBytePrimaryByOneStep(uint32_t basePrimary, UBool isCompressible, int32_t step) {
    // Extract the third byte, minus the minimum byte value,
    // minus the step, modulo the number of usable byte values, plus the minimum.
    U_ASSERT(0 < step && step <= 0x7f);
    int32_t byte3 = ((int32_t)(basePrimary >> 8) & 0xff) - step;
    if(byte3 >= 2) {
        return (basePrimary & 0xffff0000) | ((uint32_t)byte3 << 8);
    }
    byte3 += 254;
    // Same with the second byte,
    // but reserve the PRIMARY_COMPRESSION_LOW_BYTE and high byte if necessary.
    int32_t byte2 = ((int32_t)(basePrimary >> 16) & 0xff) - 1;
    if(isCompressible) {
        if(byte2 < 4) {
            byte2 = 0xfe;
            basePrimary -= 0x1000000;
        }
    } else {
        if(byte2 < 2) {
            byte2 = 0xff;
            basePrimary -= 0x1000000;
        }
    }
    // First byte, assume no further underflow.
    return (basePrimary & 0xff000000) | ((uint32_t)byte2 << 16) | ((uint32_t)byte3 << 8);
}

uint32_t
Collation::getThreeBytePrimaryForOffsetData(UChar32 c, int64_t dataCE) {
    uint32_t p = (uint32_t)(dataCE >> 32);  // three-byte primary pppppp00
    int32_t lower32 = (int32_t)dataCE;  // base code point b & step s: bbbbbbss (bit 7: isCompressible)
    int32_t offset = (c - (lower32 >> 8)) * (lower32 & 0x7f);  // delta * increment
    UBool isCompressible = (lower32 & 0x80) != 0;
    return Collation::incThreeBytePrimaryByOffset(p, isCompressible, offset);
}

uint32_t
Collation::unassignedPrimaryFromCodePoint(UChar32 c) {
    // Create a gap before U+0000. Use c=-1 for [first unassigned].
    ++c;
    // Fourth byte: 18 values, every 14th byte value (gap of 13).
    uint32_t primary = 2 + (c % 18) * 14;
    c /= 18;
    // Third byte: 254 values.
    primary |= (2 + (c % 254)) << 8;
    c /= 254;
    // Second byte: 251 values 04..FE excluding the primary compression bytes.
    primary |= (4 + (c % 251)) << 16;
    // One lead byte covers all code points (c < 0x1182B4 = 1*251*254*18).
    return primary | (UNASSIGNED_IMPLICIT_BYTE << 24);
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationbuilder.cpp
*
* (replaced the former ucol_bld.cpp)
*
* created on: 2013may06
* created by: Markus W. Scherer
*/

#ifdef DEBUG_COLLATION_BUILDER
#include <stdio.h>
#endif



#if !UCONFIG_NO_COLLATION













// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationbuilder.h
*
* created on: 2013may06
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationrootelements.h
*
* created on: 2013mar01
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION




U_NAMESPACE_BEGIN

/**
 * Container and access methods for collation elements and weights
 * that occur in the root collator.
 * Needed for finding boundaries for building a tailoring.
 *
 * This class takes and returns 16-bit secondary and tertiary weights.
 */
class U_I18N_API CollationRootElements : public UMemory {
public:
    CollationRootElements(const uint32_t *rootElements, int32_t rootElementsLength)
            : elements(rootElements), length(rootElementsLength) {}

    /**
     * Higher than any root primary.
     */
    static const uint32_t PRIMARY_SENTINEL = 0xffffff00;

    /**
     * Flag in a root element, set if the element contains secondary & tertiary weights,
     * rather than a primary.
     */
    static const uint32_t SEC_TER_DELTA_FLAG = 0x80;
    /**
     * Mask for getting the primary range step value from a primary-range-end element.
     */
    static const uint8_t PRIMARY_STEP_MASK = 0x7f;

    enum {
        /**
         * Index of the first CE with a non-zero tertiary weight.
         * Same as the start of the compact root elements table.
         */
        IX_FIRST_TERTIARY_INDEX,
        /**
         * Index of the first CE with a non-zero secondary weight.
         */
        IX_FIRST_SECONDARY_INDEX,
        /**
         * Index of the first CE with a non-zero primary weight.
         */
        IX_FIRST_PRIMARY_INDEX,
        /**
         * Must match Collation::COMMON_SEC_AND_TER_CE.
         */
        IX_COMMON_SEC_AND_TER_CE,
        /**
         * Secondary & tertiary boundaries.
         * Bits 31..24: [fixed last secondary common byte 45]
         * Bits 23..16: [fixed first ignorable secondary byte 80]
         * Bits 15.. 8: reserved, 0
         * Bits  7.. 0: [fixed first ignorable tertiary byte 3C]
         */
        IX_SEC_TER_BOUNDARIES,
        /**
         * The current number of indexes.
         * Currently the same as elements[IX_FIRST_TERTIARY_INDEX].
         */
        IX_COUNT
    };

    /**
     * Returns the boundary between tertiary weights of primary/secondary CEs
     * and those of tertiary CEs.
     * This is the upper limit for tertiaries of primary/secondary CEs.
     * This minus one is the lower limit for tertiaries of tertiary CEs.
     */
    uint32_t getTertiaryBoundary() const {
        return (elements[IX_SEC_TER_BOUNDARIES] << 8) & 0xff00;
    }

    /**
     * Returns the first assigned tertiary CE.
     */
    uint32_t getFirstTertiaryCE() const {
        return elements[elements[IX_FIRST_TERTIARY_INDEX]] & ~SEC_TER_DELTA_FLAG;
    }

    /**
     * Returns the last assigned tertiary CE.
     */
    uint32_t getLastTertiaryCE() const {
        return elements[elements[IX_FIRST_SECONDARY_INDEX] - 1] & ~SEC_TER_DELTA_FLAG;
    }

    /**
     * Returns the last common secondary weight.
     * This is the lower limit for secondaries of primary CEs.
     */
    uint32_t getLastCommonSecondary() const {
        return (elements[IX_SEC_TER_BOUNDARIES] >> 16) & 0xff00;
    }

    /**
     * Returns the boundary between secondary weights of primary CEs
     * and those of secondary CEs.
     * This is the upper limit for secondaries of primary CEs.
     * This minus one is the lower limit for secondaries of secondary CEs.
     */
    uint32_t getSecondaryBoundary() const {
        return (elements[IX_SEC_TER_BOUNDARIES] >> 8) & 0xff00;
    }

    /**
     * Returns the first assigned secondary CE.
     */
    uint32_t getFirstSecondaryCE() const {
        return elements[elements[IX_FIRST_SECONDARY_INDEX]] & ~SEC_TER_DELTA_FLAG;
    }

    /**
     * Returns the last assigned secondary CE.
     */
    uint32_t getLastSecondaryCE() const {
        return elements[elements[IX_FIRST_PRIMARY_INDEX] - 1] & ~SEC_TER_DELTA_FLAG;
    }

    /**
     * Returns the first assigned primary weight.
     */
    uint32_t getFirstPrimary() const {
        return elements[elements[IX_FIRST_PRIMARY_INDEX]];  // step=0: cannot be a range end
    }

    /**
     * Returns the first assigned primary CE.
     */
    int64_t getFirstPrimaryCE() const {
        return Collation::makeCE(getFirstPrimary());
    }

    /**
     * Returns the last root CE with a primary weight before p.
     * Intended only for reordering group boundaries.
     */
    int64_t lastCEWithPrimaryBefore(uint32_t p) const;

    /**
     * Returns the first root CE with a primary weight of at least p.
     * Intended only for reordering group boundaries.
     */
    int64_t firstCEWithPrimaryAtLeast(uint32_t p) const;

    /**
     * Returns the primary weight before p.
     * p must be greater than the first root primary.
     */
    uint32_t getPrimaryBefore(uint32_t p, UBool isCompressible) const;

    /** Returns the secondary weight before [p, s]. */
    uint32_t getSecondaryBefore(uint32_t p, uint32_t s) const;

    /** Returns the tertiary weight before [p, s, t]. */
    uint32_t getTertiaryBefore(uint32_t p, uint32_t s, uint32_t t) const;

    /**
     * Finds the index of the input primary.
     * p must occur as a root primary, and must not be 0.
     */
    int32_t findPrimary(uint32_t p) const;

    /**
     * Returns the primary weight after p where index=findPrimary(p).
     * p must be at least the first root primary.
     */
    uint32_t getPrimaryAfter(uint32_t p, int32_t index, UBool isCompressible) const;
    /**
     * Returns the secondary weight after [p, s] where index=findPrimary(p)
     * except use index=0 for p=0.
     *
     * Must return a weight for every root [p, s] as well as for every weight
     * returned by getSecondaryBefore(). If p!=0 then s can be BEFORE_WEIGHT16.
     *
     * Exception: [0, 0] is handled by the CollationBuilder:
     * Both its lower and upper boundaries are special.
     */
    uint32_t getSecondaryAfter(int32_t index, uint32_t s) const;
    /**
     * Returns the tertiary weight after [p, s, t] where index=findPrimary(p)
     * except use index=0 for p=0.
     *
     * Must return a weight for every root [p, s, t] as well as for every weight
     * returned by getTertiaryBefore(). If s!=0 then t can be BEFORE_WEIGHT16.
     *
     * Exception: [0, 0, 0] is handled by the CollationBuilder:
     * Both its lower and upper boundaries are special.
     */
    uint32_t getTertiaryAfter(int32_t index, uint32_t s, uint32_t t) const;

private:
    /**
     * Returns the first secondary & tertiary weights for p where index=findPrimary(p)+1.
     */
    uint32_t getFirstSecTerForPrimary(int32_t index) const;

    /**
     * Finds the largest index i where elements[i]<=p.
     * Requires first primary<=p<0xffffff00 (PRIMARY_SENTINEL).
     * Does not require that p is a root collator primary.
     */
    int32_t findP(uint32_t p) const;

    static inline UBool isEndOfPrimaryRange(uint32_t q) {
        return (q & SEC_TER_DELTA_FLAG) == 0 && (q & PRIMARY_STEP_MASK) != 0;
    }

    /**
     * Data structure:
     *
     * The first few entries are indexes, up to elements[IX_FIRST_TERTIARY_INDEX].
     * See the comments on the IX_ constants.
     *
     * All other elements are a compact form of the root collator CEs
     * in mostly collation order.
     *
     * A sequence of one or more root CEs with the same primary weight is stored as
     * one element with the primary weight, with the SEC_TER_DELTA_FLAG flag not set,
     * followed by elements with only the secondary/tertiary weights,
     * each with that flag set.
     * If the lowest secondary/tertiary combination is Collation::COMMON_SEC_AND_TER_CE,
     * then the element for that combination is omitted.
     *
     * Note: If the first actual secondary/tertiary combination is higher than
     * Collation::COMMON_SEC_AND_TER_CE (which is unusual),
     * the runtime code will assume anyway that Collation::COMMON_SEC_AND_TER_CE is present.
     *
     * A range of only-primary CEs with a consistent "step" increment
     * from each primary to the next may be stored as a range.
     * Only the first and last primary are stored, and the last has the step
     * value in the low bits (PRIMARY_STEP_MASK).
     *
     * An range-end element may also either start a new range or be followed by
     * elements with secondary/tertiary deltas.
     *
     * A primary element that is not a range end has zero step bits.
     *
     * There is no element for the completely ignorable CE (all weights 0).
     *
     * Before elements[IX_FIRST_PRIMARY_INDEX], all elements are secondary/tertiary deltas,
     * for all of the ignorable root CEs.
     *
     * There are no elements for unassigned-implicit primary CEs.
     * All primaries stored here are at most 3 bytes long.
     */
    const uint32_t *elements;
    int32_t length;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationruleparser.h
*
* created on: 2013apr10
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION





struct UParseError;

U_NAMESPACE_BEGIN

struct CollationData;
struct CollationTailoring;

class Locale;
class Normalizer2;

struct CollationSettings;

class U_I18N_API CollationRuleParser : public UMemory {
public:
    /** Special reset positions. */
    enum Position {
        FIRST_TERTIARY_IGNORABLE,
        LAST_TERTIARY_IGNORABLE,
        FIRST_SECONDARY_IGNORABLE,
        LAST_SECONDARY_IGNORABLE,
        FIRST_PRIMARY_IGNORABLE,
        LAST_PRIMARY_IGNORABLE,
        FIRST_VARIABLE,
        LAST_VARIABLE,
        FIRST_REGULAR,
        LAST_REGULAR,
        FIRST_IMPLICIT,
        LAST_IMPLICIT,
        FIRST_TRAILING,
        LAST_TRAILING
    };

    /**
     * First character of contractions that encode special reset positions.
     * U+FFFE cannot be tailored via rule syntax.
     *
     * The second contraction character is POS_BASE + Position.
     */
    static const UChar POS_LEAD = 0xfffe;
    /**
     * Base for the second character of contractions that encode special reset positions.
     * Braille characters U+28xx are printable and normalization-inert.
     * @see POS_LEAD
     */
    static const UChar POS_BASE = 0x2800;

    class U_I18N_API Sink : public UObject {
    public:
        virtual ~Sink();
        /**
         * Adds a reset.
         * strength=UCOL_IDENTICAL for &str.
         * strength=UCOL_PRIMARY/UCOL_SECONDARY/UCOL_TERTIARY for &[before n]str where n=1/2/3.
         */
        virtual void addReset(int32_t strength, const UnicodeString &str,
                              const char *&errorReason, UErrorCode &errorCode) = 0;
        /**
         * Adds a relation with strength and prefix | str / extension.
         */
        virtual void addRelation(int32_t strength, const UnicodeString &prefix,
                                 const UnicodeString &str, const UnicodeString &extension,
                                 const char *&errorReason, UErrorCode &errorCode) = 0;

        virtual void suppressContractions(const UnicodeSet &set, const char *&errorReason,
                                          UErrorCode &errorCode);

        virtual void optimize(const UnicodeSet &set, const char *&errorReason,
                              UErrorCode &errorCode);
    };

    class U_I18N_API Importer : public UObject {
    public:
        virtual ~Importer();
        virtual void getRules(
                const char *localeID, const char *collationType,
                UnicodeString &rules,
                const char *&errorReason, UErrorCode &errorCode) = 0;
    };

    /**
     * Constructor.
     * The Sink must be set before parsing.
     * The Importer can be set, otherwise [import locale] syntax is not supported.
     */
    CollationRuleParser(const CollationData *base, UErrorCode &errorCode);
    ~CollationRuleParser();

    /**
     * Sets the pointer to a Sink object.
     * The pointer is aliased: Pointer copy without cloning or taking ownership.
     */
    void setSink(Sink *sinkAlias) {
        sink = sinkAlias;
    }

    /**
     * Sets the pointer to an Importer object.
     * The pointer is aliased: Pointer copy without cloning or taking ownership.
     */
    void setImporter(Importer *importerAlias) {
        importer = importerAlias;
    }

    void parse(const UnicodeString &ruleString,
               CollationSettings &outSettings,
               UParseError *outParseError,
               UErrorCode &errorCode);

    const char *getErrorReason() const { return errorReason; }

    /**
     * Gets a script or reorder code from its string representation.
     * @return the script/reorder code, or
     * -1 if not recognized
     */
    static int32_t getReorderCode(const char *word);

private:
    /** UCOL_PRIMARY=0 .. UCOL_IDENTICAL=15 */
    static const int32_t STRENGTH_MASK = 0xf;
    static const int32_t STARRED_FLAG = 0x10;
    static const int32_t OFFSET_SHIFT = 8;

    void parse(const UnicodeString &ruleString, UErrorCode &errorCode);
    void parseRuleChain(UErrorCode &errorCode);
    int32_t parseResetAndPosition(UErrorCode &errorCode);
    int32_t parseRelationOperator(UErrorCode &errorCode);
    void parseRelationStrings(int32_t strength, int32_t i, UErrorCode &errorCode);
    void parseStarredCharacters(int32_t strength, int32_t i, UErrorCode &errorCode);
    int32_t parseTailoringString(int32_t i, UnicodeString &raw, UErrorCode &errorCode);
    int32_t parseString(int32_t i, UnicodeString &raw, UErrorCode &errorCode);

    /**
     * Sets str to a contraction of U+FFFE and (U+2800 + Position).
     * @return rule index after the special reset position
     */
    int32_t parseSpecialPosition(int32_t i, UnicodeString &str, UErrorCode &errorCode);
    void parseSetting(UErrorCode &errorCode);
    void parseReordering(const UnicodeString &raw, UErrorCode &errorCode);
    static UColAttributeValue getOnOffValue(const UnicodeString &s);

    int32_t parseUnicodeSet(int32_t i, UnicodeSet &set, UErrorCode &errorCode);
    int32_t readWords(int32_t i, UnicodeString &raw) const;
    int32_t skipComment(int32_t i) const;

    void setParseError(const char *reason, UErrorCode &errorCode);
    void setErrorContext();

    /**
     * ASCII [:P:] and [:S:]:
     * [\u0021-\u002F \u003A-\u0040 \u005B-\u0060 \u007B-\u007E]
     */
    static UBool isSyntaxChar(UChar32 c);
    int32_t skipWhiteSpace(int32_t i) const;

    const Normalizer2 &nfd, &nfc;

    const UnicodeString *rules;
    const CollationData *const baseData;
    CollationSettings *settings;
    UParseError *parseError;
    const char *errorReason;

    Sink *sink;
    Importer *importer;

    int32_t ruleIndex;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION




struct UParseError;

U_NAMESPACE_BEGIN

struct CollationData;
struct CollationTailoring;

class CEFinalizer;
class CollationDataBuilder;
class Normalizer2;
class Normalizer2Impl;

class U_I18N_API CollationBuilder : public CollationRuleParser::Sink {
public:
    CollationBuilder(const CollationTailoring *base, UErrorCode &errorCode);
    virtual ~CollationBuilder();

    void disableFastLatin() { fastLatinEnabled = FALSE; }

    CollationTailoring *parseAndBuild(const UnicodeString &ruleString,
                                      const UVersionInfo rulesVersion,
                                      CollationRuleParser::Importer *importer,
                                      UParseError *outParseError,
                                      UErrorCode &errorCode);

    const char *getErrorReason() const { return errorReason; }

private:
    friend class CEFinalizer;

    /** Implements CollationRuleParser::Sink. */
    virtual void addReset(int32_t strength, const UnicodeString &str,
                          const char *&errorReason, UErrorCode &errorCode);
    /**
     * Returns the secondary or tertiary weight preceding the current node's weight.
     * node=nodes[index].
     */
    uint32_t getWeight16Before(int32_t index, int64_t node, int32_t level);

    int64_t getSpecialResetPosition(const UnicodeString &str,
                                    const char *&parserErrorReason, UErrorCode &errorCode);

    /** Implements CollationRuleParser::Sink. */
    virtual void addRelation(int32_t strength, const UnicodeString &prefix,
                             const UnicodeString &str, const UnicodeString &extension,
                             const char *&errorReason, UErrorCode &errorCode);

    /**
     * Picks one of the current CEs and finds or inserts a node in the graph
     * for the CE + strength.
     */
    int32_t findOrInsertNodeForCEs(int32_t strength, const char *&parserErrorReason,
                                   UErrorCode &errorCode);
    int32_t findOrInsertNodeForRootCE(int64_t ce, int32_t strength, UErrorCode &errorCode);
    /** Finds or inserts the node for a root CE's primary weight. */
    int32_t findOrInsertNodeForPrimary(uint32_t p, UErrorCode &errorCode);
    /** Finds or inserts the node for a secondary or tertiary weight. */
    int32_t findOrInsertWeakNode(int32_t index, uint32_t weight16, int32_t level,
                                 UErrorCode &errorCode);

    /**
     * Makes and inserts a new tailored node into the list, after the one at index.
     * Skips over nodes of weaker strength to maintain collation order
     * ("postpone insertion").
     * @return the new node's index
     */
    int32_t insertTailoredNodeAfter(int32_t index, int32_t strength, UErrorCode &errorCode);

    /**
     * Inserts a new node into the list, between list-adjacent items.
     * The node's previous and next indexes must not be set yet.
     * @return the new node's index
     */
    int32_t insertNodeBetween(int32_t index, int32_t nextIndex, int64_t node,
                              UErrorCode &errorCode);

    /**
     * Finds the node which implies or contains a common=05 weight of the given strength
     * (secondary or tertiary), if the current node is stronger.
     * Skips weaker nodes and tailored nodes if the current node is stronger
     * and is followed by an explicit-common-weight node.
     * Always returns the input index if that node is no stronger than the given strength.
     */
    int32_t findCommonNode(int32_t index, int32_t strength) const;

    void setCaseBits(const UnicodeString &nfdString,
                     const char *&parserErrorReason, UErrorCode &errorCode);

    /** Implements CollationRuleParser::Sink. */
    virtual void suppressContractions(const UnicodeSet &set, const char *&parserErrorReason,
                                      UErrorCode &errorCode);

    /** Implements CollationRuleParser::Sink. */
    virtual void optimize(const UnicodeSet &set, const char *&parserErrorReason,
                          UErrorCode &errorCode);

    /**
     * Adds the mapping and its canonical closure.
     * Takes ce32=dataBuilder->encodeCEs(...) so that the data builder
     * need not re-encode the CEs multiple times.
     */
    uint32_t addWithClosure(const UnicodeString &nfdPrefix, const UnicodeString &nfdString,
                            const int64_t newCEs[], int32_t newCEsLength, uint32_t ce32,
                            UErrorCode &errorCode);
    uint32_t addOnlyClosure(const UnicodeString &nfdPrefix, const UnicodeString &nfdString,
                            const int64_t newCEs[], int32_t newCEsLength, uint32_t ce32,
                            UErrorCode &errorCode);
    void addTailComposites(const UnicodeString &nfdPrefix, const UnicodeString &nfdString,
                           UErrorCode &errorCode);
    UBool mergeCompositeIntoString(const UnicodeString &nfdString, int32_t indexAfterLastStarter,
                                   UChar32 composite, const UnicodeString &decomp,
                                   UnicodeString &newNFDString, UnicodeString &newString,
                                   UErrorCode &errorCode) const;

    UBool ignorePrefix(const UnicodeString &s, UErrorCode &errorCode) const;
    UBool ignoreString(const UnicodeString &s, UErrorCode &errorCode) const;
    UBool isFCD(const UnicodeString &s, UErrorCode &errorCode) const;

    void closeOverComposites(UErrorCode &errorCode);

    uint32_t addIfDifferent(const UnicodeString &prefix, const UnicodeString &str,
                            const int64_t newCEs[], int32_t newCEsLength, uint32_t ce32,
                            UErrorCode &errorCode);
    static UBool sameCEs(const int64_t ces1[], int32_t ces1Length,
                         const int64_t ces2[], int32_t ces2Length);

    /**
     * Walks the tailoring graph and overwrites tailored nodes with new CEs.
     * After this, the graph is destroyed.
     * The nodes array can then be used only as a source of tailored CEs.
     */
    void makeTailoredCEs(UErrorCode &errorCode);
    /**
     * Counts the tailored nodes of the given strength up to the next node
     * which is either stronger or has an explicit weight of this strength.
     */
    static int32_t countTailoredNodes(const int64_t *nodesArray, int32_t i, int32_t strength);

    /** Replaces temporary CEs with the final CEs they point to. */
    void finalizeCEs(UErrorCode &errorCode);

    /**
     * Encodes "temporary CE" data into a CE that fits into the CE32 data structure,
     * with 2-byte primary, 1-byte secondary and 6-bit tertiary,
     * with valid CE byte values.
     *
     * The index must not exceed 20 bits (0xfffff).
     * The strength must fit into 2 bits (UCOL_PRIMARY..UCOL_QUATERNARY).
     *
     * Temporary CEs are distinguished from real CEs by their use of
     * secondary weights 06..45 which are otherwise reserved for compressed sort keys.
     *
     * The case bits are unused and available.
     */
    static inline int64_t tempCEFromIndexAndStrength(int32_t index, int32_t strength) {
        return
            // CE byte offsets, to ensure valid CE bytes, and case bits 11
            INT64_C(0x4040000006002000) +
            // index bits 19..13 -> primary byte 1 = CE bits 63..56 (byte values 40..BF)
            ((int64_t)(index & 0xfe000) << 43) +
            // index bits 12..6 -> primary byte 2 = CE bits 55..48 (byte values 40..BF)
            ((int64_t)(index & 0x1fc0) << 42) +
            // index bits 5..0 -> secondary byte 1 = CE bits 31..24 (byte values 06..45)
            ((index & 0x3f) << 24) +
            // strength bits 1..0 -> tertiary byte 1 = CE bits 13..8 (byte values 20..23)
            (strength << 8);
    }
    static inline int32_t indexFromTempCE(int64_t tempCE) {
        tempCE -= INT64_C(0x4040000006002000);
        return
            ((int32_t)(tempCE >> 43) & 0xfe000) |
            ((int32_t)(tempCE >> 42) & 0x1fc0) |
            ((int32_t)(tempCE >> 24) & 0x3f);
    }
    static inline int32_t strengthFromTempCE(int64_t tempCE) {
        return ((int32_t)tempCE >> 8) & 3;
    }
    static inline UBool isTempCE(int64_t ce) {
        uint32_t sec = (uint32_t)ce >> 24;
        return 6 <= sec && sec <= 0x45;
    }

    static inline int32_t indexFromTempCE32(uint32_t tempCE32) {
        tempCE32 -= 0x40400620;
        return
            ((int32_t)(tempCE32 >> 11) & 0xfe000) |
            ((int32_t)(tempCE32 >> 10) & 0x1fc0) |
            ((int32_t)(tempCE32 >> 8) & 0x3f);
    }
    static inline UBool isTempCE32(uint32_t ce32) {
        return
            (ce32 & 0xff) >= 2 &&  // not a long-primary/long-secondary CE32
            6 <= ((ce32 >> 8) & 0xff) && ((ce32 >> 8) & 0xff) <= 0x45;
    }

    static int32_t ceStrength(int64_t ce);

    /** At most 1M nodes, limited by the 20 bits in node bit fields. */
    static const int32_t MAX_INDEX = 0xfffff;
    /**
     * Node bit 6 is set on a primary node if there are nodes
     * with secondary values below the common secondary weight (05).
     */
    static const int32_t HAS_BEFORE2 = 0x40;
    /**
     * Node bit 5 is set on a primary or secondary node if there are nodes
     * with tertiary values below the common tertiary weight (05).
     */
    static const int32_t HAS_BEFORE3 = 0x20;
    /**
     * Node bit 3 distinguishes a tailored node, which has no weight value,
     * from a node with an explicit (root or default) weight.
     */
    static const int32_t IS_TAILORED = 8;

    static inline int64_t nodeFromWeight32(uint32_t weight32) {
        return (int64_t)weight32 << 32;
    }
    static inline int64_t nodeFromWeight16(uint32_t weight16) {
        return (int64_t)weight16 << 48;
    }
    static inline int64_t nodeFromPreviousIndex(int32_t previous) {
        return (int64_t)previous << 28;
    }
    static inline int64_t nodeFromNextIndex(int32_t next) {
        return next << 8;
    }
    static inline int64_t nodeFromStrength(int32_t strength) {
        return strength;
    }

    static inline uint32_t weight32FromNode(int64_t node) {
        return (uint32_t)(node >> 32);
    }
    static inline uint32_t weight16FromNode(int64_t node) {
        return (uint32_t)(node >> 48) & 0xffff;
    }
    static inline int32_t previousIndexFromNode(int64_t node) {
        return (int32_t)(node >> 28) & MAX_INDEX;
    }
    static inline int32_t nextIndexFromNode(int64_t node) {
        return ((int32_t)node >> 8) & MAX_INDEX;
    }
    static inline int32_t strengthFromNode(int64_t node) {
        return (int32_t)node & 3;
    }

    static inline UBool nodeHasBefore2(int64_t node) {
        return (node & HAS_BEFORE2) != 0;
    }
    static inline UBool nodeHasBefore3(int64_t node) {
        return (node & HAS_BEFORE3) != 0;
    }
    static inline UBool nodeHasAnyBefore(int64_t node) {
        return (node & (HAS_BEFORE2 | HAS_BEFORE3)) != 0;
    }
    static inline UBool isTailoredNode(int64_t node) {
        return (node & IS_TAILORED) != 0;
    }

    static inline int64_t changeNodePreviousIndex(int64_t node, int32_t previous) {
        return (node & INT64_C(0xffff00000fffffff)) | nodeFromPreviousIndex(previous);
    }
    static inline int64_t changeNodeNextIndex(int64_t node, int32_t next) {
        return (node & INT64_C(0xfffffffff00000ff)) | nodeFromNextIndex(next);
    }

    const Normalizer2 &nfd, &fcd;
    const Normalizer2Impl &nfcImpl;

    const CollationTailoring *base;
    const CollationData *baseData;
    const CollationRootElements rootElements;
    uint32_t variableTop;

    CollationDataBuilder *dataBuilder;
    UBool fastLatinEnabled;
    UnicodeSet optimizeSet;
    const char *errorReason;

    int64_t ces[Collation::MAX_EXPANSION_LENGTH];
    int32_t cesLength;

    /**
     * Indexes of nodes with root primary weights, sorted by primary.
     * Compact form of a TreeMap from root primary to node index.
     *
     * This is a performance optimization for finding reset positions.
     * Without this, we would have to search through the entire nodes list.
     * It also allows storing root primary weights in list head nodes,
     * without previous index, leaving room in root primary nodes for 32-bit primary weights.
     */
    UVector32 rootPrimaryIndexes;
    /**
     * Data structure for assigning tailored weights and CEs.
     * Doubly-linked lists of nodes in mostly collation order.
     * Each list starts with a root primary node and ends with a nextIndex of 0.
     *
     * When there are any nodes in the list, then there is always a root primary node at index 0.
     * This allows some code not to have to check explicitly for nextIndex==0.
     *
     * Root primary nodes have 32-bit weights but do not have previous indexes.
     * All other nodes have at most 16-bit weights and do have previous indexes.
     *
     * Nodes with explicit weights store root collator weights,
     * or default weak weights (e.g., secondary 05) for stronger nodes.
     * "Tailored" nodes, with the IS_TAILORED bit set,
     * do not store explicit weights but rather
     * create a difference of a certain strength from the preceding node.
     *
     * A root node is followed by either
     * - a root/default node of the same strength, or
     * - a root/default node of the next-weaker strength, or
     * - a tailored node of the same strength.
     *
     * A node of a given strength normally implies "common" weights on weaker levels.
     *
     * A node with HAS_BEFORE2 must be immediately followed by
     * a secondary node with an explicit below-common weight, then a secondary tailored node,
     * and later an explicit common-secondary node.
     * The below-common weight can be a root weight,
     * or it can be BEFORE_WEIGHT16 for tailoring before an implied common weight
     * or before the lowest root weight.
     * (&[before 2] resets to an explicit secondary node so that
     * the following addRelation(secondary) tailors right after that.
     * If we did not have this node and instead were to reset on the primary node,
     * then addRelation(secondary) would skip forward to the the COMMON_WEIGHT16 node.)
     *
     * If the flag is not set, then there are no explicit secondary nodes
     * with the common or lower weights.
     *
     * Same for HAS_BEFORE3 for tertiary nodes and weights.
     * A node must not have both flags set.
     *
     * Tailored CEs are initially represented in a CollationDataBuilder as temporary CEs
     * which point to stable indexes in this list,
     * and temporary CEs stored in a CollationDataBuilder only point to tailored nodes.
     *
     * A temporary CE in the ces[] array may point to a non-tailored reset-before-position node,
     * until the next relation is added.
     *
     * At the end, the tailored weights are allocated as necessary,
     * then the tailored nodes are replaced with final CEs,
     * and the CollationData is rewritten by replacing temporary CEs with final ones.
     *
     * We cannot simply insert new nodes in the middle of the array
     * because that would invalidate the indexes stored in existing temporary CEs.
     * We need to use a linked graph with stable indexes to existing nodes.
     * A doubly-linked list seems easiest to maintain.
     *
     * Each node is stored as an int64_t, with its fields stored as bit fields.
     *
     * Root primary node:
     * - primary weight: 32 bits 63..32
     * - reserved/unused/zero: 4 bits 31..28
     *
     * Weaker root nodes & tailored nodes:
     * - a weight: 16 bits 63..48
     *   + a root or default weight for a non-tailored node
     *   + unused/zero for a tailored node
     * - index to the previous node: 20 bits 47..28
     *
     * All types of nodes:
     * - index to the next node: 20 bits 27..8
     *   + nextIndex=0 in last node per root-primary list
     * - reserved/unused/zero bits: bits 7, 4, 2
     * - HAS_BEFORE2: bit 6
     * - HAS_BEFORE3: bit 5
     * - IS_TAILORED: bit 3
     * - the difference strength (primary/secondary/tertiary/quaternary): 2 bits 1..0
     *
     * We could allocate structs with pointers, but we would have to store them
     * in a pointer list so that they can be indexed from temporary CEs,
     * and they would require more memory allocations.
     */
    UVector64 nodes;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationdatabuilder.h
*
* created on: 2012apr01
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION













U_NAMESPACE_BEGIN

struct ConditionalCE32;

class CollationFastLatinBuilder;
class CopyHelper;
class DataBuilderCollationIterator;
class UCharsTrieBuilder;

/**
 * Low-level CollationData builder.
 * Takes (character, CE) pairs and builds them into runtime data structures.
 * Supports characters with context prefixes and contraction suffixes.
 */
class U_I18N_API CollationDataBuilder : public UObject {
public:
    /**
     * Collation element modifier. Interface class for a modifier
     * that changes a tailoring builder's temporary CEs to final CEs.
     * Called for every non-special CE32 and every expansion CE.
     */
    class CEModifier : public UObject {
    public:
        virtual ~CEModifier();
        /** Returns a new CE to replace the non-special input CE32, or else Collation::NO_CE. */
        virtual int64_t modifyCE32(uint32_t ce32) const = 0;
        /** Returns a new CE to replace the input CE, or else Collation::NO_CE. */
        virtual int64_t modifyCE(int64_t ce) const = 0;
    };

    CollationDataBuilder(UErrorCode &errorCode);

    virtual ~CollationDataBuilder();

    void initForTailoring(const CollationData *b, UErrorCode &errorCode);

    virtual UBool isCompressibleLeadByte(uint32_t b) const;

    inline UBool isCompressiblePrimary(uint32_t p) const {
        return isCompressibleLeadByte(p >> 24);
    }

    /**
     * @return TRUE if this builder has mappings (e.g., add() has been called)
     */
    UBool hasMappings() const { return modified; }

    /**
     * @return TRUE if c has CEs in this builder
     */
    UBool isAssigned(UChar32 c) const;

    /**
     * @return the three-byte primary if c maps to a single such CE and has no context data,
     * otherwise returns 0.
     */
    uint32_t getLongPrimaryIfSingleCE(UChar32 c) const;

    /**
     * @return the single CE for c.
     * Sets an error code if c does not have a single CE.
     */
    int64_t getSingleCE(UChar32 c, UErrorCode &errorCode) const;

    void add(const UnicodeString &prefix, const UnicodeString &s,
             const int64_t ces[], int32_t cesLength,
             UErrorCode &errorCode);

    /**
     * Encodes the ces as either the returned ce32 by itself,
     * or by storing an expansion, with the returned ce32 referring to that.
     *
     * add(p, s, ces, cesLength) = addCE32(p, s, encodeCEs(ces, cesLength))
     */
    virtual uint32_t encodeCEs(const int64_t ces[], int32_t cesLength, UErrorCode &errorCode);
    void addCE32(const UnicodeString &prefix, const UnicodeString &s,
                 uint32_t ce32, UErrorCode &errorCode);

    /**
     * Sets three-byte-primary CEs for a range of code points in code point order,
     * if it is worth doing; otherwise no change is made.
     * None of the code points in the range should have complex mappings so far
     * (expansions/contractions/prefixes).
     * @param start first code point
     * @param end last code point (inclusive)
     * @param primary primary weight for 'start'
     * @param step per-code point primary-weight increment
     * @param errorCode ICU in/out error code
     * @return TRUE if an OFFSET_TAG range was used for start..end
     */
    UBool maybeSetPrimaryRange(UChar32 start, UChar32 end,
                               uint32_t primary, int32_t step,
                               UErrorCode &errorCode);

    /**
     * Sets three-byte-primary CEs for a range of code points in code point order.
     * Sets range values if that is worth doing, or else individual values.
     * None of the code points in the range should have complex mappings so far
     * (expansions/contractions/prefixes).
     * @param start first code point
     * @param end last code point (inclusive)
     * @param primary primary weight for 'start'
     * @param step per-code point primary-weight increment
     * @param errorCode ICU in/out error code
     * @return the next primary after 'end': start primary incremented by ((end-start)+1)*step
     */
    uint32_t setPrimaryRangeAndReturnNext(UChar32 start, UChar32 end,
                                          uint32_t primary, int32_t step,
                                          UErrorCode &errorCode);

    /**
     * Copies all mappings from the src builder, with modifications.
     * This builder here must not be built yet, and should be empty.
     */
    void copyFrom(const CollationDataBuilder &src, const CEModifier &modifier,
                  UErrorCode &errorCode);

    void optimize(const UnicodeSet &set, UErrorCode &errorCode);
    void suppressContractions(const UnicodeSet &set, UErrorCode &errorCode);

    void enableFastLatin() { fastLatinEnabled = TRUE; }
    virtual void build(CollationData &data, UErrorCode &errorCode);

    /**
     * Looks up CEs for s and appends them to the ces array.
     * Does not handle normalization: s should be in FCD form.
     *
     * Does not write completely ignorable CEs.
     * Does not write beyond Collation::MAX_EXPANSION_LENGTH.
     *
     * @return incremented cesLength
     */
    int32_t getCEs(const UnicodeString &s, int64_t ces[], int32_t cesLength);
    int32_t getCEs(const UnicodeString &prefix, const UnicodeString &s,
                   int64_t ces[], int32_t cesLength);

protected:
    friend class CopyHelper;
    friend class DataBuilderCollationIterator;

    uint32_t getCE32FromOffsetCE32(UBool fromBase, UChar32 c, uint32_t ce32) const;

    int32_t addCE(int64_t ce, UErrorCode &errorCode);
    int32_t addCE32(uint32_t ce32, UErrorCode &errorCode);
    int32_t addConditionalCE32(const UnicodeString &context, uint32_t ce32, UErrorCode &errorCode);

    inline ConditionalCE32 *getConditionalCE32(int32_t index) const {
        return static_cast<ConditionalCE32 *>(conditionalCE32s[index]);
    }
    inline ConditionalCE32 *getConditionalCE32ForCE32(uint32_t ce32) const {
        return getConditionalCE32(Collation::indexFromCE32(ce32));
    }

    static uint32_t makeBuilderContextCE32(int32_t index) {
        return Collation::makeCE32FromTagAndIndex(Collation::BUILDER_DATA_TAG, index);
    }
    static inline UBool isBuilderContextCE32(uint32_t ce32) {
        return Collation::hasCE32Tag(ce32, Collation::BUILDER_DATA_TAG);
    }

    static uint32_t encodeOneCEAsCE32(int64_t ce);
    uint32_t encodeOneCE(int64_t ce, UErrorCode &errorCode);
    uint32_t encodeExpansion(const int64_t ces[], int32_t length, UErrorCode &errorCode);
    uint32_t encodeExpansion32(const int32_t newCE32s[], int32_t length, UErrorCode &errorCode);

    uint32_t copyFromBaseCE32(UChar32 c, uint32_t ce32, UBool withContext, UErrorCode &errorCode);
    /**
     * Copies base contractions to a list of ConditionalCE32.
     * Sets cond->next to the index of the first new item
     * and returns the index of the last new item.
     */
    int32_t copyContractionsFromBaseCE32(UnicodeString &context, UChar32 c, uint32_t ce32,
                                         ConditionalCE32 *cond, UErrorCode &errorCode);

    UBool getJamoCE32s(uint32_t jamoCE32s[], UErrorCode &errorCode);
    void setDigitTags(UErrorCode &errorCode);
    void setLeadSurrogates(UErrorCode &errorCode);

    void buildMappings(CollationData &data, UErrorCode &errorCode);

    void clearContexts();
    void buildContexts(UErrorCode &errorCode);
    uint32_t buildContext(ConditionalCE32 *head, UErrorCode &errorCode);
    int32_t addContextTrie(uint32_t defaultCE32, UCharsTrieBuilder &trieBuilder,
                           UErrorCode &errorCode);

    void buildFastLatinTable(CollationData &data, UErrorCode &errorCode);

    int32_t getCEs(const UnicodeString &s, int32_t start, int64_t ces[], int32_t cesLength);

    static UChar32 jamoCpFromIndex(int32_t i) {
        // 0 <= i < CollationData::JAMO_CE32S_LENGTH = 19 + 21 + 27
        if(i < Hangul::JAMO_L_COUNT) { return Hangul::JAMO_L_BASE + i; }
        i -= Hangul::JAMO_L_COUNT;
        if(i < Hangul::JAMO_V_COUNT) { return Hangul::JAMO_V_BASE + i; }
        i -= Hangul::JAMO_V_COUNT;
        // i < 27
        return Hangul::JAMO_T_BASE + 1 + i;
    }

    /** @see Collation::BUILDER_DATA_TAG */
    static const uint32_t IS_BUILDER_JAMO_CE32 = 0x100;

    const Normalizer2Impl &nfcImpl;
    const CollationData *base;
    const CollationSettings *baseSettings;
    UTrie2 *trie;
    UVector32 ce32s;
    UVector64 ce64s;
    UVector conditionalCE32s;  // vector of ConditionalCE32
    // Characters that have context (prefixes or contraction suffixes).
    UnicodeSet contextChars;
    // Serialized UCharsTrie structures for finalized contexts.
    UnicodeString contexts;
    UnicodeSet unsafeBackwardSet;
    UBool modified;

    UBool fastLatinEnabled;
    CollationFastLatinBuilder *fastLatinBuilder;

    DataBuilderCollationIterator *collIter;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationfastlatin.h
*
* created on: 2013aug09
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION

U_NAMESPACE_BEGIN

struct CollationData;
struct CollationSettings;

class U_I18N_API CollationFastLatin /* all static */ {
public:
    /**
     * Fast Latin format version (one byte 1..FF).
     * Must be incremented for any runtime-incompatible changes,
     * in particular, for changes to any of the following constants.
     *
     * When the major version number of the main data format changes,
     * we can reset this fast Latin version to 1.
     */
    static const uint16_t VERSION = 2;

    static const int32_t LATIN_MAX = 0x17f;
    static const int32_t LATIN_LIMIT = LATIN_MAX + 1;

    static const int32_t LATIN_MAX_UTF8_LEAD = 0xc5;  // UTF-8 lead byte of LATIN_MAX

    static const int32_t PUNCT_START = 0x2000;
    static const int32_t PUNCT_LIMIT = 0x2040;

    // excludes U+FFFE & U+FFFF
    static const int32_t NUM_FAST_CHARS = LATIN_LIMIT + (PUNCT_LIMIT - PUNCT_START);

    // Note on the supported weight ranges:
    // Analysis of UCA 6.3 and CLDR 23 non-search tailorings shows that
    // the CEs for characters in the above ranges, excluding expansions with length >2,
    // excluding contractions of >2 characters, and other restrictions
    // (see the builder's getCEsFromCE32()),
    // use at most about 150 primary weights,
    // where about 94 primary weights are possibly-variable (space/punct/symbol/currency),
    // at most 4 secondary before-common weights,
    // at most 4 secondary after-common weights,
    // at most 16 secondary high weights (in secondary CEs), and
    // at most 4 tertiary after-common weights.
    // The following ranges are designed to support slightly more weights than that.
    // (en_US_POSIX is unusual: It creates about 64 variable + 116 Latin primaries.)

    // Digits may use long primaries (preserving more short ones)
    // or short primaries (faster) without changing this data structure.
    // (If we supported numeric collation, then digits would have to have long primaries
    // so that special handling does not affect the fast path.)

    static const uint32_t SHORT_PRIMARY_MASK = 0xfc00;  // bits 15..10
    static const uint32_t INDEX_MASK = 0x3ff;  // bits 9..0 for expansions & contractions
    static const uint32_t SECONDARY_MASK = 0x3e0;  // bits 9..5
    static const uint32_t CASE_MASK = 0x18;  // bits 4..3
    static const uint32_t LONG_PRIMARY_MASK = 0xfff8;  // bits 15..3
    static const uint32_t TERTIARY_MASK = 7;  // bits 2..0
    static const uint32_t CASE_AND_TERTIARY_MASK = CASE_MASK | TERTIARY_MASK;

    static const uint32_t TWO_SHORT_PRIMARIES_MASK =
            (SHORT_PRIMARY_MASK << 16) | SHORT_PRIMARY_MASK;  // 0xfc00fc00
    static const uint32_t TWO_LONG_PRIMARIES_MASK =
            (LONG_PRIMARY_MASK << 16) | LONG_PRIMARY_MASK;  // 0xfff8fff8
    static const uint32_t TWO_SECONDARIES_MASK =
            (SECONDARY_MASK << 16) | SECONDARY_MASK;  // 0x3e003e0
    static const uint32_t TWO_CASES_MASK =
            (CASE_MASK << 16) | CASE_MASK;  // 0x180018
    static const uint32_t TWO_TERTIARIES_MASK =
            (TERTIARY_MASK << 16) | TERTIARY_MASK;  // 0x70007

    /**
     * Contraction with one fast Latin character.
     * Use INDEX_MASK to find the start of the contraction list after the fixed table.
     * The first entry contains the default mapping.
     * Otherwise use CONTR_CHAR_MASK for the contraction character index
     * (in ascending order).
     * Use CONTR_LENGTH_SHIFT for the length of the entry
     * (1=BAIL_OUT, 2=one CE, 3=two CEs).
     *
     * Also, U+0000 maps to a contraction entry, so that the fast path need not
     * check for NUL termination.
     * It usually maps to a contraction list with only the completely ignorable default value.
     */
    static const uint32_t CONTRACTION = 0x400;
    /**
     * An expansion encodes two CEs.
     * Use INDEX_MASK to find the pair of CEs after the fixed table.
     *
     * The higher a mini CE value, the easier it is to process.
     * For expansions and higher, no context needs to be considered.
     */
    static const uint32_t EXPANSION = 0x800;
    /**
     * Encodes one CE with a long/low mini primary (there are 128).
     * All potentially-variable primaries must be in this range,
     * to make the short-primary path as fast as possible.
     */
    static const uint32_t MIN_LONG = 0xc00;
    static const uint32_t LONG_INC = 8;
    static const uint32_t MAX_LONG = 0xff8;
    /**
     * Encodes one CE with a short/high primary (there are 60),
     * plus a secondary CE if the secondary weight is high.
     * Fast handling: At least all letter primaries should be in this range.
     */
    static const uint32_t MIN_SHORT = 0x1000;
    static const uint32_t SHORT_INC = 0x400;
    /** The highest primary weight is reserved for U+FFFF. */
    static const uint32_t MAX_SHORT = SHORT_PRIMARY_MASK;

    static const uint32_t MIN_SEC_BEFORE = 0;  // must add SEC_OFFSET
    static const uint32_t SEC_INC = 0x20;
    static const uint32_t MAX_SEC_BEFORE = MIN_SEC_BEFORE + 4 * SEC_INC;  // 5 before common
    static const uint32_t COMMON_SEC = MAX_SEC_BEFORE + SEC_INC;
    static const uint32_t MIN_SEC_AFTER = COMMON_SEC + SEC_INC;
    static const uint32_t MAX_SEC_AFTER = MIN_SEC_AFTER + 5 * SEC_INC;  // 6 after common
    static const uint32_t MIN_SEC_HIGH = MAX_SEC_AFTER + SEC_INC;  // 20 high secondaries
    static const uint32_t MAX_SEC_HIGH = SECONDARY_MASK;

    /**
     * Lookup: Add this offset to secondary weights, except for completely ignorable CEs.
     * Must be greater than any special value, e.g., MERGE_WEIGHT.
     * The exact value is not relevant for the format version.
     */
    static const uint32_t SEC_OFFSET = SEC_INC;
    static const uint32_t COMMON_SEC_PLUS_OFFSET = COMMON_SEC + SEC_OFFSET;

    static const uint32_t TWO_SEC_OFFSETS =
            (SEC_OFFSET << 16) | SEC_OFFSET;  // 0x200020
    static const uint32_t TWO_COMMON_SEC_PLUS_OFFSET =
            (COMMON_SEC_PLUS_OFFSET << 16) | COMMON_SEC_PLUS_OFFSET;

    static const uint32_t LOWER_CASE = 8;  // case bits include this offset
    static const uint32_t TWO_LOWER_CASES = (LOWER_CASE << 16) | LOWER_CASE;  // 0x80008

    static const uint32_t COMMON_TER = 0;  // must add TER_OFFSET
    static const uint32_t MAX_TER_AFTER = 7;  // 7 after common

    /**
     * Lookup: Add this offset to tertiary weights, except for completely ignorable CEs.
     * Must be greater than any special value, e.g., MERGE_WEIGHT.
     * Must be greater than case bits as well, so that with combined case+tertiary weights
     * plus the offset the tertiary bits does not spill over into the case bits.
     * The exact value is not relevant for the format version.
     */
    static const uint32_t TER_OFFSET = SEC_OFFSET;
    static const uint32_t COMMON_TER_PLUS_OFFSET = COMMON_TER + TER_OFFSET;

    static const uint32_t TWO_TER_OFFSETS = (TER_OFFSET << 16) | TER_OFFSET;
    static const uint32_t TWO_COMMON_TER_PLUS_OFFSET =
            (COMMON_TER_PLUS_OFFSET << 16) | COMMON_TER_PLUS_OFFSET;

    static const uint32_t MERGE_WEIGHT = 3;
    static const uint32_t EOS = 2;  // end of string
    static const uint32_t BAIL_OUT = 1;

    /**
     * Contraction result first word bits 8..0 contain the
     * second contraction character, as a char index 0..NUM_FAST_CHARS-1.
     * Each contraction list is terminated with a word containing CONTR_CHAR_MASK.
     */
    static const uint32_t CONTR_CHAR_MASK = 0x1ff;
    /**
     * Contraction result first word bits 10..9 contain the result length:
     * 1=bail out, 2=one mini CE, 3=two mini CEs
     */
    static const uint32_t CONTR_LENGTH_SHIFT = 9;

    /**
     * Comparison return value when the regular comparison must be used.
     * The exact value is not relevant for the format version.
     */
    static const int32_t BAIL_OUT_RESULT = -2;

    static inline int32_t getCharIndex(UChar c) {
        if(c <= LATIN_MAX) {
            return c;
        } else if(PUNCT_START <= c && c < PUNCT_LIMIT) {
            return c - (PUNCT_START - LATIN_LIMIT);
        } else {
            // Not a fast Latin character.
            // Note: U+FFFE & U+FFFF are forbidden in tailorings
            // and thus do not occur in any contractions.
            return -1;
        }
    }

    /**
     * Computes the options value for the compare functions
     * and writes the precomputed primary weights.
     * Returns -1 if the Latin fastpath is not supported for the data and settings.
     * The capacity must be LATIN_LIMIT.
     */
    static int32_t getOptions(const CollationData *data, const CollationSettings &settings,
                              uint16_t *primaries, int32_t capacity);

    static int32_t compareUTF16(const uint16_t *table, const uint16_t *primaries, int32_t options,
                                const UChar *left, int32_t leftLength,
                                const UChar *right, int32_t rightLength);

    static int32_t compareUTF8(const uint16_t *table, const uint16_t *primaries, int32_t options,
                               const uint8_t *left, int32_t leftLength,
                               const uint8_t *right, int32_t rightLength);

private:
    static uint32_t lookup(const uint16_t *table, UChar32 c);
    static uint32_t lookupUTF8(const uint16_t *table, UChar32 c,
                               const uint8_t *s8, int32_t &sIndex, int32_t sLength);
    static uint32_t lookupUTF8Unsafe(const uint16_t *table, UChar32 c,
                                     const uint8_t *s8, int32_t &sIndex);

    static uint32_t nextPair(const uint16_t *table, UChar32 c, uint32_t ce,
                             const UChar *s16, const uint8_t *s8, int32_t &sIndex, int32_t &sLength);

    static inline uint32_t getPrimaries(uint32_t variableTop, uint32_t pair) {
        uint32_t ce = pair & 0xffff;
        if(ce >= MIN_SHORT) { return pair & TWO_SHORT_PRIMARIES_MASK; }
        if(ce > variableTop) { return pair & TWO_LONG_PRIMARIES_MASK; }
        if(ce >= MIN_LONG) { return 0; }  // variable
        return pair;  // special mini CE
    }
    static inline uint32_t getSecondariesFromOneShortCE(uint32_t ce) {
        ce &= SECONDARY_MASK;
        if(ce < MIN_SEC_HIGH) {
            return ce + SEC_OFFSET;
        } else {
            return ((ce + SEC_OFFSET) << 16) | COMMON_SEC_PLUS_OFFSET;
        }
    }
    static uint32_t getSecondaries(uint32_t variableTop, uint32_t pair);
    static uint32_t getCases(uint32_t variableTop, UBool strengthIsPrimary, uint32_t pair);
    static uint32_t getTertiaries(uint32_t variableTop, UBool withCaseBits, uint32_t pair);
    static uint32_t getQuaternaries(uint32_t variableTop, uint32_t pair);

private:
    CollationFastLatin();  // no constructor
};

/*
 * Format of the CollationFastLatin data table.
 * CollationFastLatin::VERSION = 2.
 *
 * This table contains data for a Latin-text collation fastpath.
 * The data is stored as an array of uint16_t which contains the following parts.
 *
 * uint16_t  -- version & header length
 *   Bits 15..8: version, must match the VERSION
 *         7..0: length of the header
 *
 * uint16_t varTops[header length - 1]
 *   Version 2:
 *   varTops[m] is the highest CollationFastLatin long-primary weight
 *   of supported maxVariable group m
 *   (special reorder group space, punct, symbol, currency).
 *
 *   Version 1:
 *   Each of these values maps the variable top lead byte of a supported maxVariable group
 *   to the highest CollationFastLatin long-primary weight.
 *   The values are stored in ascending order.
 *   Bits 15..7: max fast-Latin long-primary weight (bits 11..3 shifted left by 4 bits)
 *         6..0: regular primary lead byte
 *
 * uint16_t miniCEs[0x1c0]
 *   A mini collation element for each character U+0000..U+017F and U+2000..U+203F.
 *   Each value encodes one or two mini CEs (two are possible if the first one
 *   has a short mini primary and the second one is a secondary CE, i.e., primary == 0),
 *   or points to an expansion or to a contraction table.
 *   U+0000 always has a contraction entry,
 *   so that NUL-termination need not be tested in the fastpath.
 *   If the collation elements for a character or contraction cannot be encoded in this format,
 *   then the BAIL_OUT value is stored.
 *   For details see the comments for the class constants.
 *
 * uint16_t expansions[variable length];
 *   Expansion mini CEs contain an offset relative to just after the miniCEs table.
 *   An expansions contains exactly 2 mini CEs.
 *
 * uint16_t contractions[variable length];
 *   Contraction mini CEs contain an offset relative to just after the miniCEs table.
 *   It points to a list of tuples which map from a contraction suffix character to a result.
 *   First uint16_t of each tuple:
 *     Bits 10..9: Length of the result (1..3), see comments on CONTR_LENGTH_SHIFT.
 *     Bits  8..0: Contraction character, see comments on CONTR_CHAR_MASK.
 *   This is followed by 0, 1, or 2 uint16_t according to the length.
 *   Each list is terminated by an entry with CONTR_CHAR_MASK.
 *   Each list starts with such an entry which also contains the default result
 *   for when there is no contraction match.
 *
 * -----------------
 * Changes for version 2 (ICU 55)
 *
 * Special reorder groups do not necessarily start on whole primary lead bytes any more.
 * Therefore, the varTops data has a new format:
 * Version 1 stored the lead bytes of the highest root primaries for
 * the maxVariable-supported special reorder groups.
 * Now the top 16 bits would need to be stored,
 * and it is simpler to store only the fast-Latin weights.
 */

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION






// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*  
*******************************************************************************
*
*   Copyright (C) 1999-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  collationweights.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001mar08 as ucol_wgt.h
*   created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION



U_NAMESPACE_BEGIN

/**
 * Allocates n collation element weights between two exclusive limits.
 * Used only internally by the collation tailoring builder.
 */
class U_I18N_API CollationWeights : public UMemory {
public:
    CollationWeights();

    static inline int32_t lengthOfWeight(uint32_t weight) {
        if((weight&0xffffff)==0) {
            return 1;
        } else if((weight&0xffff)==0) {
            return 2;
        } else if((weight&0xff)==0) {
            return 3;
        } else {
            return 4;
        }
    }

    void initForPrimary(UBool compressible);
    void initForSecondary();
    void initForTertiary();

    /**
     * Determine heuristically
     * what ranges to use for a given number of weights between (excluding)
     * two limits.
     *
     * @param lowerLimit A collation element weight; the ranges will be filled to cover
     *                   weights greater than this one.
     * @param upperLimit A collation element weight; the ranges will be filled to cover
     *                   weights less than this one.
     * @param n          The number of collation element weights w necessary such that
     *                   lowerLimit<w<upperLimit in lexical order.
     * @return TRUE if it is possible to fit n elements between the limits
     */
    UBool allocWeights(uint32_t lowerLimit, uint32_t upperLimit, int32_t n);

    /**
     * Given a set of ranges calculated by allocWeights(),
     * iterate through the weights.
     * The ranges are modified to keep the current iteration state.
     *
     * @return The next weight in the ranges, or 0xffffffff if there is none left.
     */
    uint32_t nextWeight();

    /** @internal */
    struct WeightRange {
        uint32_t start, end;
        int32_t length, count;
    };

private:
    /** @return number of usable byte values for byte idx */
    inline int32_t countBytes(int32_t idx) const {
        return (int32_t)(maxBytes[idx] - minBytes[idx] + 1);
    }

    uint32_t incWeight(uint32_t weight, int32_t length) const;
    uint32_t incWeightByOffset(uint32_t weight, int32_t length, int32_t offset) const;
    void lengthenRange(WeightRange &range) const;
    /**
     * Takes two CE weights and calculates the
     * possible ranges of weights between the two limits, excluding them.
     * For weights with up to 4 bytes there are up to 2*4-1=7 ranges.
     */
    UBool getWeightRanges(uint32_t lowerLimit, uint32_t upperLimit);
    UBool allocWeightsInShortRanges(int32_t n, int32_t minLength);
    UBool allocWeightsInMinLengthRanges(int32_t n, int32_t minLength);

    int32_t middleLength;
    uint32_t minBytes[5];  // for byte 1, 2, 3, 4
    uint32_t maxBytes[5];
    WeightRange ranges[7];
    int32_t rangeIndex;
    int32_t rangeCount;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION






U_NAMESPACE_BEGIN

namespace {

class BundleImporter : public CollationRuleParser::Importer {
public:
    BundleImporter() {}
    virtual ~BundleImporter();
    virtual void getRules(
            const char *localeID, const char *collationType,
            UnicodeString &rules,
            const char *&errorReason, UErrorCode &errorCode);
};

BundleImporter::~BundleImporter() {}

void
BundleImporter::getRules(
        const char *localeID, const char *collationType,
        UnicodeString &rules,
        const char *& /*errorReason*/, UErrorCode &errorCode) {
    CollationLoader::loadRules(localeID, collationType, rules, errorCode);
}

}  // namespace

// RuleBasedCollator implementation ---------------------------------------- ***

// These methods are here, rather than in rulebasedcollator.cpp,
// for modularization:
// Most code using Collator does not need to build a Collator from rules.
// By moving these constructors and helper methods to a separate file,
// most code will not have a static dependency on the builder code.

RuleBasedCollator::RuleBasedCollator()
        : data(NULL),
          settings(NULL),
          tailoring(NULL),
          cacheEntry(NULL),
          validLocale(""),
          explicitlySetAttributes(0),
          actualLocaleIsSameAsValid(FALSE) {
}

RuleBasedCollator::RuleBasedCollator(const UnicodeString &rules, UErrorCode &errorCode)
        : data(NULL),
          settings(NULL),
          tailoring(NULL),
          cacheEntry(NULL),
          validLocale(""),
          explicitlySetAttributes(0),
          actualLocaleIsSameAsValid(FALSE) {
    internalBuildTailoring(rules, UCOL_DEFAULT, UCOL_DEFAULT, NULL, NULL, errorCode);
}

RuleBasedCollator::RuleBasedCollator(const UnicodeString &rules, ECollationStrength strength,
                                     UErrorCode &errorCode)
        : data(NULL),
          settings(NULL),
          tailoring(NULL),
          cacheEntry(NULL),
          validLocale(""),
          explicitlySetAttributes(0),
          actualLocaleIsSameAsValid(FALSE) {
    internalBuildTailoring(rules, strength, UCOL_DEFAULT, NULL, NULL, errorCode);
}

RuleBasedCollator::RuleBasedCollator(const UnicodeString &rules,
                                     UColAttributeValue decompositionMode,
                                     UErrorCode &errorCode)
        : data(NULL),
          settings(NULL),
          tailoring(NULL),
          cacheEntry(NULL),
          validLocale(""),
          explicitlySetAttributes(0),
          actualLocaleIsSameAsValid(FALSE) {
    internalBuildTailoring(rules, UCOL_DEFAULT, decompositionMode, NULL, NULL, errorCode);
}

RuleBasedCollator::RuleBasedCollator(const UnicodeString &rules,
                                     ECollationStrength strength,
                                     UColAttributeValue decompositionMode,
                                     UErrorCode &errorCode)
        : data(NULL),
          settings(NULL),
          tailoring(NULL),
          cacheEntry(NULL),
          validLocale(""),
          explicitlySetAttributes(0),
          actualLocaleIsSameAsValid(FALSE) {
    internalBuildTailoring(rules, strength, decompositionMode, NULL, NULL, errorCode);
}

RuleBasedCollator::RuleBasedCollator(const UnicodeString &rules,
                                     UParseError &parseError, UnicodeString &reason,
                                     UErrorCode &errorCode)
        : data(NULL),
          settings(NULL),
          tailoring(NULL),
          cacheEntry(NULL),
          validLocale(""),
          explicitlySetAttributes(0),
          actualLocaleIsSameAsValid(FALSE) {
    internalBuildTailoring(rules, UCOL_DEFAULT, UCOL_DEFAULT, &parseError, &reason, errorCode);
}

void
RuleBasedCollator::internalBuildTailoring(const UnicodeString &rules,
                                          int32_t strength,
                                          UColAttributeValue decompositionMode,
                                          UParseError *outParseError, UnicodeString *outReason,
                                          UErrorCode &errorCode) {
    const CollationTailoring *base = CollationRoot::getRoot(errorCode);
    if(U_FAILURE(errorCode)) { return; }
    if(outReason != NULL) { outReason->remove(); }
    CollationBuilder builder(base, errorCode);
    UVersionInfo noVersion = { 0, 0, 0, 0 };
    BundleImporter importer;
    LocalPointer<CollationTailoring> t(builder.parseAndBuild(rules, noVersion,
                                                             &importer,
                                                             outParseError, errorCode));
    if(U_FAILURE(errorCode)) {
        const char *reason = builder.getErrorReason();
        if(reason != NULL && outReason != NULL) {
            *outReason = UnicodeString(reason, -1, US_INV);
        }
        return;
    }
    t->actualLocale.setToBogus();
    adoptTailoring(t.orphan(), errorCode);
    // Set attributes after building the collator,
    // to keep the default settings consistent with the rule string.
    if(strength != UCOL_DEFAULT) {
        setAttribute(UCOL_STRENGTH, (UColAttributeValue)strength, errorCode);
    }
    if(decompositionMode != UCOL_DEFAULT) {
        setAttribute(UCOL_NORMALIZATION_MODE, decompositionMode, errorCode);
    }
}

// CollationBuilder implementation ----------------------------------------- ***

// Some compilers don't care if constants are defined in the .cpp file.
// MS Visual C++ does not like it, but gcc requires it. clang does not care.
#ifndef _MSC_VER
const int32_t CollationBuilder::HAS_BEFORE2;
const int32_t CollationBuilder::HAS_BEFORE3;
#endif

CollationBuilder::CollationBuilder(const CollationTailoring *b, UErrorCode &errorCode)
        : nfd(*Normalizer2::getNFDInstance(errorCode)),
          fcd(*Normalizer2Factory::getFCDInstance(errorCode)),
          nfcImpl(*Normalizer2Factory::getNFCImpl(errorCode)),
          base(b),
          baseData(b->data),
          rootElements(b->data->rootElements, b->data->rootElementsLength),
          variableTop(0),
          dataBuilder(new CollationDataBuilder(errorCode)), fastLatinEnabled(TRUE),
          errorReason(NULL),
          cesLength(0),
          rootPrimaryIndexes(errorCode), nodes(errorCode) {
    nfcImpl.ensureCanonIterData(errorCode);
    if(U_FAILURE(errorCode)) {
        errorReason = "CollationBuilder fields initialization failed";
        return;
    }
    if(dataBuilder == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    dataBuilder->initForTailoring(baseData, errorCode);
    if(U_FAILURE(errorCode)) {
        errorReason = "CollationBuilder initialization failed";
    }
}

CollationBuilder::~CollationBuilder() {
    delete dataBuilder;
}

CollationTailoring *
CollationBuilder::parseAndBuild(const UnicodeString &ruleString,
                                const UVersionInfo rulesVersion,
                                CollationRuleParser::Importer *importer,
                                UParseError *outParseError,
                                UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return NULL; }
    if(baseData->rootElements == NULL) {
        errorCode = U_MISSING_RESOURCE_ERROR;
        errorReason = "missing root elements data, tailoring not supported";
        return NULL;
    }
    LocalPointer<CollationTailoring> tailoring(new CollationTailoring(base->settings));
    if(tailoring.isNull() || tailoring->isBogus()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    CollationRuleParser parser(baseData, errorCode);
    if(U_FAILURE(errorCode)) { return NULL; }
    // Note: This always bases &[last variable] and &[first regular]
    // on the root collator's maxVariable/variableTop.
    // If we wanted this to change after [maxVariable x], then we would keep
    // the tailoring.settings pointer here and read its variableTop when we need it.
    // See http://unicode.org/cldr/trac/ticket/6070
    variableTop = base->settings->variableTop;
    parser.setSink(this);
    parser.setImporter(importer);
    CollationSettings &ownedSettings = *SharedObject::copyOnWrite(tailoring->settings);
    parser.parse(ruleString, ownedSettings, outParseError, errorCode);
    errorReason = parser.getErrorReason();
    if(U_FAILURE(errorCode)) { return NULL; }
    if(dataBuilder->hasMappings()) {
        makeTailoredCEs(errorCode);
        closeOverComposites(errorCode);
        finalizeCEs(errorCode);
        // Copy all of ASCII, and Latin-1 letters, into each tailoring.
        optimizeSet.add(0, 0x7f);
        optimizeSet.add(0xc0, 0xff);
        // Hangul is decomposed on the fly during collation,
        // and the tailoring data is always built with HANGUL_TAG specials.
        optimizeSet.remove(Hangul::HANGUL_BASE, Hangul::HANGUL_END);
        dataBuilder->optimize(optimizeSet, errorCode);
        tailoring->ensureOwnedData(errorCode);
        if(U_FAILURE(errorCode)) { return NULL; }
        if(fastLatinEnabled) { dataBuilder->enableFastLatin(); }
        dataBuilder->build(*tailoring->ownedData, errorCode);
        tailoring->builder = dataBuilder;
        dataBuilder = NULL;
    } else {
        tailoring->data = baseData;
    }
    if(U_FAILURE(errorCode)) { return NULL; }
    ownedSettings.fastLatinOptions = CollationFastLatin::getOptions(
        tailoring->data, ownedSettings,
        ownedSettings.fastLatinPrimaries, UPRV_LENGTHOF(ownedSettings.fastLatinPrimaries));
    tailoring->rules = ruleString;
    tailoring->rules.getTerminatedBuffer();  // ensure NUL-termination
    tailoring->setVersion(base->version, rulesVersion);
    return tailoring.orphan();
}

void
CollationBuilder::addReset(int32_t strength, const UnicodeString &str,
                           const char *&parserErrorReason, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    U_ASSERT(!str.isEmpty());
    if(str.charAt(0) == CollationRuleParser::POS_LEAD) {
        ces[0] = getSpecialResetPosition(str, parserErrorReason, errorCode);
        cesLength = 1;
        if(U_FAILURE(errorCode)) { return; }
        U_ASSERT((ces[0] & Collation::CASE_AND_QUATERNARY_MASK) == 0);
    } else {
        // normal reset to a character or string
        UnicodeString nfdString = nfd.normalize(str, errorCode);
        if(U_FAILURE(errorCode)) {
            parserErrorReason = "normalizing the reset position";
            return;
        }
        cesLength = dataBuilder->getCEs(nfdString, ces, 0);
        if(cesLength > Collation::MAX_EXPANSION_LENGTH) {
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            parserErrorReason = "reset position maps to too many collation elements (more than 31)";
            return;
        }
    }
    if(strength == UCOL_IDENTICAL) { return; }  // simple reset-at-position

    // &[before strength]position
    U_ASSERT(UCOL_PRIMARY <= strength && strength <= UCOL_TERTIARY);
    int32_t index = findOrInsertNodeForCEs(strength, parserErrorReason, errorCode);
    if(U_FAILURE(errorCode)) { return; }

    int64_t node = nodes.elementAti(index);
    // If the index is for a "weaker" node,
    // then skip backwards over this and further "weaker" nodes.
    while(strengthFromNode(node) > strength) {
        index = previousIndexFromNode(node);
        node = nodes.elementAti(index);
    }

    // Find or insert a node whose index we will put into a temporary CE.
    if(strengthFromNode(node) == strength && isTailoredNode(node)) {
        // Reset to just before this same-strength tailored node.
        index = previousIndexFromNode(node);
    } else if(strength == UCOL_PRIMARY) {
        // root primary node (has no previous index)
        uint32_t p = weight32FromNode(node);
        if(p == 0) {
            errorCode = U_UNSUPPORTED_ERROR;
            parserErrorReason = "reset primary-before ignorable not possible";
            return;
        }
        if(p <= rootElements.getFirstPrimary()) {
            // There is no primary gap between ignorables and the space-first-primary.
            errorCode = U_UNSUPPORTED_ERROR;
            parserErrorReason = "reset primary-before first non-ignorable not supported";
            return;
        }
        if(p == Collation::FIRST_TRAILING_PRIMARY) {
            // We do not support tailoring to an unassigned-implicit CE.
            errorCode = U_UNSUPPORTED_ERROR;
            parserErrorReason = "reset primary-before [first trailing] not supported";
            return;
        }
        p = rootElements.getPrimaryBefore(p, baseData->isCompressiblePrimary(p));
        index = findOrInsertNodeForPrimary(p, errorCode);
        // Go to the last node in this list:
        // Tailor after the last node between adjacent root nodes.
        for(;;) {
            node = nodes.elementAti(index);
            int32_t nextIndex = nextIndexFromNode(node);
            if(nextIndex == 0) { break; }
            index = nextIndex;
        }
    } else {
        // &[before 2] or &[before 3]
        index = findCommonNode(index, UCOL_SECONDARY);
        if(strength >= UCOL_TERTIARY) {
            index = findCommonNode(index, UCOL_TERTIARY);
        }
        // findCommonNode() stayed on the stronger node or moved to
        // an explicit common-weight node of the reset-before strength.
        node = nodes.elementAti(index);
        if(strengthFromNode(node) == strength) {
            // Found a same-strength node with an explicit weight.
            uint32_t weight16 = weight16FromNode(node);
            if(weight16 == 0) {
                errorCode = U_UNSUPPORTED_ERROR;
                if(strength == UCOL_SECONDARY) {
                    parserErrorReason = "reset secondary-before secondary ignorable not possible";
                } else {
                    parserErrorReason = "reset tertiary-before completely ignorable not possible";
                }
                return;
            }
            U_ASSERT(weight16 > Collation::BEFORE_WEIGHT16);
            // Reset to just before this node.
            // Insert the preceding same-level explicit weight if it is not there already.
            // Which explicit weight immediately precedes this one?
            weight16 = getWeight16Before(index, node, strength);
            // Does this preceding weight have a node?
            uint32_t previousWeight16;
            int32_t previousIndex = previousIndexFromNode(node);
            for(int32_t i = previousIndex;; i = previousIndexFromNode(node)) {
                node = nodes.elementAti(i);
                int32_t previousStrength = strengthFromNode(node);
                if(previousStrength < strength) {
                    U_ASSERT(weight16 >= Collation::COMMON_WEIGHT16 || i == previousIndex);
                    // Either the reset element has an above-common weight and
                    // the parent node provides the implied common weight,
                    // or the reset element has a weight<=common in the node
                    // right after the parent, and we need to insert the preceding weight.
                    previousWeight16 = Collation::COMMON_WEIGHT16;
                    break;
                } else if(previousStrength == strength && !isTailoredNode(node)) {
                    previousWeight16 = weight16FromNode(node);
                    break;
                }
                // Skip weaker nodes and same-level tailored nodes.
            }
            if(previousWeight16 == weight16) {
                // The preceding weight has a node,
                // maybe with following weaker or tailored nodes.
                // Reset to the last of them.
                index = previousIndex;
            } else {
                // Insert a node with the preceding weight, reset to that.
                node = nodeFromWeight16(weight16) | nodeFromStrength(strength);
                index = insertNodeBetween(previousIndex, index, node, errorCode);
            }
        } else {
            // Found a stronger node with implied strength-common weight.
            uint32_t weight16 = getWeight16Before(index, node, strength);
            index = findOrInsertWeakNode(index, weight16, strength, errorCode);
        }
        // Strength of the temporary CE = strength of its reset position.
        // Code above raises an error if the before-strength is stronger.
        strength = ceStrength(ces[cesLength - 1]);
    }
    if(U_FAILURE(errorCode)) {
        parserErrorReason = "inserting reset position for &[before n]";
        return;
    }
    ces[cesLength - 1] = tempCEFromIndexAndStrength(index, strength);
}

uint32_t
CollationBuilder::getWeight16Before(int32_t index, int64_t node, int32_t level) {
    U_ASSERT(strengthFromNode(node) < level || !isTailoredNode(node));
    // Collect the root CE weights if this node is for a root CE.
    // If it is not, then return the low non-primary boundary for a tailored CE.
    uint32_t t;
    if(strengthFromNode(node) == UCOL_TERTIARY) {
        t = weight16FromNode(node);
    } else {
        t = Collation::COMMON_WEIGHT16;  // Stronger node with implied common weight.
    }
    while(strengthFromNode(node) > UCOL_SECONDARY) {
        index = previousIndexFromNode(node);
        node = nodes.elementAti(index);
    }
    if(isTailoredNode(node)) {
        return Collation::BEFORE_WEIGHT16;
    }
    uint32_t s;
    if(strengthFromNode(node) == UCOL_SECONDARY) {
        s = weight16FromNode(node);
    } else {
        s = Collation::COMMON_WEIGHT16;  // Stronger node with implied common weight.
    }
    while(strengthFromNode(node) > UCOL_PRIMARY) {
        index = previousIndexFromNode(node);
        node = nodes.elementAti(index);
    }
    if(isTailoredNode(node)) {
        return Collation::BEFORE_WEIGHT16;
    }
    // [p, s, t] is a root CE. Return the preceding weight for the requested level.
    uint32_t p = weight32FromNode(node);
    uint32_t weight16;
    if(level == UCOL_SECONDARY) {
        weight16 = rootElements.getSecondaryBefore(p, s);
    } else {
        weight16 = rootElements.getTertiaryBefore(p, s, t);
        U_ASSERT((weight16 & ~Collation::ONLY_TERTIARY_MASK) == 0);
    }
    return weight16;
}

int64_t
CollationBuilder::getSpecialResetPosition(const UnicodeString &str,
                                          const char *&parserErrorReason, UErrorCode &errorCode) {
    U_ASSERT(str.length() == 2);
    int64_t ce;
    int32_t strength = UCOL_PRIMARY;
    UBool isBoundary = FALSE;
    UChar32 pos = str.charAt(1) - CollationRuleParser::POS_BASE;
    U_ASSERT(0 <= pos && pos <= CollationRuleParser::LAST_TRAILING);
    switch(pos) {
    case CollationRuleParser::FIRST_TERTIARY_IGNORABLE:
        // Quaternary CEs are not supported.
        // Non-zero quaternary weights are possible only on tertiary or stronger CEs.
        return 0;
    case CollationRuleParser::LAST_TERTIARY_IGNORABLE:
        return 0;
    case CollationRuleParser::FIRST_SECONDARY_IGNORABLE: {
        // Look for a tailored tertiary node after [0, 0, 0].
        int32_t index = findOrInsertNodeForRootCE(0, UCOL_TERTIARY, errorCode);
        if(U_FAILURE(errorCode)) { return 0; }
        int64_t node = nodes.elementAti(index);
        if((index = nextIndexFromNode(node)) != 0) {
            node = nodes.elementAti(index);
            U_ASSERT(strengthFromNode(node) <= UCOL_TERTIARY);
            if(isTailoredNode(node) && strengthFromNode(node) == UCOL_TERTIARY) {
                return tempCEFromIndexAndStrength(index, UCOL_TERTIARY);
            }
        }
        return rootElements.getFirstTertiaryCE();
        // No need to look for nodeHasAnyBefore() on a tertiary node.
    }
    case CollationRuleParser::LAST_SECONDARY_IGNORABLE:
        ce = rootElements.getLastTertiaryCE();
        strength = UCOL_TERTIARY;
        break;
    case CollationRuleParser::FIRST_PRIMARY_IGNORABLE: {
        // Look for a tailored secondary node after [0, 0, *].
        int32_t index = findOrInsertNodeForRootCE(0, UCOL_SECONDARY, errorCode);
        if(U_FAILURE(errorCode)) { return 0; }
        int64_t node = nodes.elementAti(index);
        while((index = nextIndexFromNode(node)) != 0) {
            node = nodes.elementAti(index);
            strength = strengthFromNode(node);
            if(strength < UCOL_SECONDARY) { break; }
            if(strength == UCOL_SECONDARY) {
                if(isTailoredNode(node)) {
                    if(nodeHasBefore3(node)) {
                        index = nextIndexFromNode(nodes.elementAti(nextIndexFromNode(node)));
                        U_ASSERT(isTailoredNode(nodes.elementAti(index)));
                    }
                    return tempCEFromIndexAndStrength(index, UCOL_SECONDARY);
                } else {
                    break;
                }
            }
        }
        ce = rootElements.getFirstSecondaryCE();
        strength = UCOL_SECONDARY;
        break;
    }
    case CollationRuleParser::LAST_PRIMARY_IGNORABLE:
        ce = rootElements.getLastSecondaryCE();
        strength = UCOL_SECONDARY;
        break;
    case CollationRuleParser::FIRST_VARIABLE:
        ce = rootElements.getFirstPrimaryCE();
        isBoundary = TRUE;  // FractionalUCA.txt: FDD1 00A0, SPACE first primary
        break;
    case CollationRuleParser::LAST_VARIABLE:
        ce = rootElements.lastCEWithPrimaryBefore(variableTop + 1);
        break;
    case CollationRuleParser::FIRST_REGULAR:
        ce = rootElements.firstCEWithPrimaryAtLeast(variableTop + 1);
        isBoundary = TRUE;  // FractionalUCA.txt: FDD1 263A, SYMBOL first primary
        break;
    case CollationRuleParser::LAST_REGULAR:
        // Use the Hani-first-primary rather than the actual last "regular" CE before it,
        // for backward compatibility with behavior before the introduction of
        // script-first-primary CEs in the root collator.
        ce = rootElements.firstCEWithPrimaryAtLeast(
            baseData->getFirstPrimaryForGroup(USCRIPT_HAN));
        break;
    case CollationRuleParser::FIRST_IMPLICIT:
        ce = baseData->getSingleCE(0x4e00, errorCode);
        break;
    case CollationRuleParser::LAST_IMPLICIT:
        // We do not support tailoring to an unassigned-implicit CE.
        errorCode = U_UNSUPPORTED_ERROR;
        parserErrorReason = "reset to [last implicit] not supported";
        return 0;
    case CollationRuleParser::FIRST_TRAILING:
        ce = Collation::makeCE(Collation::FIRST_TRAILING_PRIMARY);
        isBoundary = TRUE;  // trailing first primary (there is no mapping for it)
        break;
    case CollationRuleParser::LAST_TRAILING:
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        parserErrorReason = "LDML forbids tailoring to U+FFFF";
        return 0;
    default:
        UPRV_UNREACHABLE;
    }

    int32_t index = findOrInsertNodeForRootCE(ce, strength, errorCode);
    if(U_FAILURE(errorCode)) { return 0; }
    int64_t node = nodes.elementAti(index);
    if((pos & 1) == 0) {
        // even pos = [first xyz]
        if(!nodeHasAnyBefore(node) && isBoundary) {
            // A <group> first primary boundary is artificially added to FractionalUCA.txt.
            // It is reachable via its special contraction, but is not normally used.
            // Find the first character tailored after the boundary CE,
            // or the first real root CE after it.
            if((index = nextIndexFromNode(node)) != 0) {
                // If there is a following node, then it must be tailored
                // because there are no root CEs with a boundary primary
                // and non-common secondary/tertiary weights.
                node = nodes.elementAti(index);
                U_ASSERT(isTailoredNode(node));
                ce = tempCEFromIndexAndStrength(index, strength);
            } else {
                U_ASSERT(strength == UCOL_PRIMARY);
                uint32_t p = (uint32_t)(ce >> 32);
                int32_t pIndex = rootElements.findPrimary(p);
                UBool isCompressible = baseData->isCompressiblePrimary(p);
                p = rootElements.getPrimaryAfter(p, pIndex, isCompressible);
                ce = Collation::makeCE(p);
                index = findOrInsertNodeForRootCE(ce, UCOL_PRIMARY, errorCode);
                if(U_FAILURE(errorCode)) { return 0; }
                node = nodes.elementAti(index);
            }
        }
        if(nodeHasAnyBefore(node)) {
            // Get the first node that was tailored before this one at a weaker strength.
            if(nodeHasBefore2(node)) {
                index = nextIndexFromNode(nodes.elementAti(nextIndexFromNode(node)));
                node = nodes.elementAti(index);
            }
            if(nodeHasBefore3(node)) {
                index = nextIndexFromNode(nodes.elementAti(nextIndexFromNode(node)));
            }
            U_ASSERT(isTailoredNode(nodes.elementAti(index)));
            ce = tempCEFromIndexAndStrength(index, strength);
        }
    } else {
        // odd pos = [last xyz]
        // Find the last node that was tailored after the [last xyz]
        // at a strength no greater than the position's strength.
        for(;;) {
            int32_t nextIndex = nextIndexFromNode(node);
            if(nextIndex == 0) { break; }
            int64_t nextNode = nodes.elementAti(nextIndex);
            if(strengthFromNode(nextNode) < strength) { break; }
            index = nextIndex;
            node = nextNode;
        }
        // Do not make a temporary CE for a root node.
        // This last node might be the node for the root CE itself,
        // or a node with a common secondary or tertiary weight.
        if(isTailoredNode(node)) {
            ce = tempCEFromIndexAndStrength(index, strength);
        }
    }
    return ce;
}

void
CollationBuilder::addRelation(int32_t strength, const UnicodeString &prefix,
                              const UnicodeString &str, const UnicodeString &extension,
                              const char *&parserErrorReason, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    UnicodeString nfdPrefix;
    if(!prefix.isEmpty()) {
        nfd.normalize(prefix, nfdPrefix, errorCode);
        if(U_FAILURE(errorCode)) {
            parserErrorReason = "normalizing the relation prefix";
            return;
        }
    }
    UnicodeString nfdString = nfd.normalize(str, errorCode);
    if(U_FAILURE(errorCode)) {
        parserErrorReason = "normalizing the relation string";
        return;
    }

    // The runtime code decomposes Hangul syllables on the fly,
    // with recursive processing but without making the Jamo pieces visible for matching.
    // It does not work with certain types of contextual mappings.
    int32_t nfdLength = nfdString.length();
    if(nfdLength >= 2) {
        UChar c = nfdString.charAt(0);
        if(Hangul::isJamoL(c) || Hangul::isJamoV(c)) {
            // While handling a Hangul syllable, contractions starting with Jamo L or V
            // would not see the following Jamo of that syllable.
            errorCode = U_UNSUPPORTED_ERROR;
            parserErrorReason = "contractions starting with conjoining Jamo L or V not supported";
            return;
        }
        c = nfdString.charAt(nfdLength - 1);
        if(Hangul::isJamoL(c) ||
                (Hangul::isJamoV(c) && Hangul::isJamoL(nfdString.charAt(nfdLength - 2)))) {
            // A contraction ending with Jamo L or L+V would require
            // generating Hangul syllables in addTailComposites() (588 for a Jamo L),
            // or decomposing a following Hangul syllable on the fly, during contraction matching.
            errorCode = U_UNSUPPORTED_ERROR;
            parserErrorReason = "contractions ending with conjoining Jamo L or L+V not supported";
            return;
        }
        // A Hangul syllable completely inside a contraction is ok.
    }
    // Note: If there is a prefix, then the parser checked that
    // both the prefix and the string beging with NFC boundaries (not Jamo V or T).
    // Therefore: prefix.isEmpty() || !isJamoVOrT(nfdString.charAt(0))
    // (While handling a Hangul syllable, prefixes on Jamo V or T
    // would not see the previous Jamo of that syllable.)

    if(strength != UCOL_IDENTICAL) {
        // Find the node index after which we insert the new tailored node.
        int32_t index = findOrInsertNodeForCEs(strength, parserErrorReason, errorCode);
        U_ASSERT(cesLength > 0);
        int64_t ce = ces[cesLength - 1];
        if(strength == UCOL_PRIMARY && !isTempCE(ce) && (uint32_t)(ce >> 32) == 0) {
            // There is no primary gap between ignorables and the space-first-primary.
            errorCode = U_UNSUPPORTED_ERROR;
            parserErrorReason = "tailoring primary after ignorables not supported";
            return;
        }
        if(strength == UCOL_QUATERNARY && ce == 0) {
            // The CE data structure does not support non-zero quaternary weights
            // on tertiary ignorables.
            errorCode = U_UNSUPPORTED_ERROR;
            parserErrorReason = "tailoring quaternary after tertiary ignorables not supported";
            return;
        }
        // Insert the new tailored node.
        index = insertTailoredNodeAfter(index, strength, errorCode);
        if(U_FAILURE(errorCode)) {
            parserErrorReason = "modifying collation elements";
            return;
        }
        // Strength of the temporary CE:
        // The new relation may yield a stronger CE but not a weaker one.
        int32_t tempStrength = ceStrength(ce);
        if(strength < tempStrength) { tempStrength = strength; }
        ces[cesLength - 1] = tempCEFromIndexAndStrength(index, tempStrength);
    }

    setCaseBits(nfdString, parserErrorReason, errorCode);
    if(U_FAILURE(errorCode)) { return; }

    int32_t cesLengthBeforeExtension = cesLength;
    if(!extension.isEmpty()) {
        UnicodeString nfdExtension = nfd.normalize(extension, errorCode);
        if(U_FAILURE(errorCode)) {
            parserErrorReason = "normalizing the relation extension";
            return;
        }
        cesLength = dataBuilder->getCEs(nfdExtension, ces, cesLength);
        if(cesLength > Collation::MAX_EXPANSION_LENGTH) {
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            parserErrorReason =
                "extension string adds too many collation elements (more than 31 total)";
            return;
        }
    }
    uint32_t ce32 = Collation::UNASSIGNED_CE32;
    if((prefix != nfdPrefix || str != nfdString) &&
            !ignorePrefix(prefix, errorCode) && !ignoreString(str, errorCode)) {
        // Map from the original input to the CEs.
        // We do this in case the canonical closure is incomplete,
        // so that it is possible to explicitly provide the missing mappings.
        ce32 = addIfDifferent(prefix, str, ces, cesLength, ce32, errorCode);
    }
    addWithClosure(nfdPrefix, nfdString, ces, cesLength, ce32, errorCode);
    if(U_FAILURE(errorCode)) {
        parserErrorReason = "writing collation elements";
        return;
    }
    cesLength = cesLengthBeforeExtension;
}

int32_t
CollationBuilder::findOrInsertNodeForCEs(int32_t strength, const char *&parserErrorReason,
                                         UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    U_ASSERT(UCOL_PRIMARY <= strength && strength <= UCOL_QUATERNARY);

    // Find the last CE that is at least as "strong" as the requested difference.
    // Note: Stronger is smaller (UCOL_PRIMARY=0).
    int64_t ce;
    for(;; --cesLength) {
        if(cesLength == 0) {
            ce = ces[0] = 0;
            cesLength = 1;
            break;
        } else {
            ce = ces[cesLength - 1];
        }
        if(ceStrength(ce) <= strength) { break; }
    }

    if(isTempCE(ce)) {
        // No need to findCommonNode() here for lower levels
        // because insertTailoredNodeAfter() will do that anyway.
        return indexFromTempCE(ce);
    }

    // root CE
    if((uint8_t)(ce >> 56) == Collation::UNASSIGNED_IMPLICIT_BYTE) {
        errorCode = U_UNSUPPORTED_ERROR;
        parserErrorReason = "tailoring relative to an unassigned code point not supported";
        return 0;
    }
    return findOrInsertNodeForRootCE(ce, strength, errorCode);
}

int32_t
CollationBuilder::findOrInsertNodeForRootCE(int64_t ce, int32_t strength, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    U_ASSERT((uint8_t)(ce >> 56) != Collation::UNASSIGNED_IMPLICIT_BYTE);

    // Find or insert the node for each of the root CE's weights,
    // down to the requested level/strength.
    // Root CEs must have common=zero quaternary weights (for which we never insert any nodes).
    U_ASSERT((ce & 0xc0) == 0);
    int32_t index = findOrInsertNodeForPrimary((uint32_t)(ce >> 32), errorCode);
    if(strength >= UCOL_SECONDARY) {
        uint32_t lower32 = (uint32_t)ce;
        index = findOrInsertWeakNode(index, lower32 >> 16, UCOL_SECONDARY, errorCode);
        if(strength >= UCOL_TERTIARY) {
            index = findOrInsertWeakNode(index, lower32 & Collation::ONLY_TERTIARY_MASK,
                                         UCOL_TERTIARY, errorCode);
        }
    }
    return index;
}

namespace {

/**
 * Like Java Collections.binarySearch(List, key, Comparator).
 *
 * @return the index>=0 where the item was found,
 *         or the index<0 for inserting the string at ~index in sorted order
 *         (index into rootPrimaryIndexes)
 */
int32_t
binarySearchForRootPrimaryNode(const int32_t *rootPrimaryIndexes, int32_t length,
                               const int64_t *nodes, uint32_t p) {
    if(length == 0) { return ~0; }
    int32_t start = 0;
    int32_t limit = length;
    for (;;) {
        int32_t i = (start + limit) / 2;
        int64_t node = nodes[rootPrimaryIndexes[i]];
        uint32_t nodePrimary = (uint32_t)(node >> 32);  // weight32FromNode(node)
        if (p == nodePrimary) {
            return i;
        } else if (p < nodePrimary) {
            if (i == start) {
                return ~start;  // insert s before i
            }
            limit = i;
        } else {
            if (i == start) {
                return ~(start + 1);  // insert s after i
            }
            start = i;
        }
    }
}

}  // namespace

int32_t
CollationBuilder::findOrInsertNodeForPrimary(uint32_t p, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }

    int32_t rootIndex = binarySearchForRootPrimaryNode(
        rootPrimaryIndexes.getBuffer(), rootPrimaryIndexes.size(), nodes.getBuffer(), p);
    if(rootIndex >= 0) {
        return rootPrimaryIndexes.elementAti(rootIndex);
    } else {
        // Start a new list of nodes with this primary.
        int32_t index = nodes.size();
        nodes.addElement(nodeFromWeight32(p), errorCode);
        rootPrimaryIndexes.insertElementAt(index, ~rootIndex, errorCode);
        return index;
    }
}

int32_t
CollationBuilder::findOrInsertWeakNode(int32_t index, uint32_t weight16, int32_t level, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    U_ASSERT(0 <= index && index < nodes.size());
    U_ASSERT(UCOL_SECONDARY <= level && level <= UCOL_TERTIARY);

    if(weight16 == Collation::COMMON_WEIGHT16) {
        return findCommonNode(index, level);
    }

    // If this will be the first below-common weight for the parent node,
    // then we will also need to insert a common weight after it.
    int64_t node = nodes.elementAti(index);
    U_ASSERT(strengthFromNode(node) < level);  // parent node is stronger
    if(weight16 != 0 && weight16 < Collation::COMMON_WEIGHT16) {
        int32_t hasThisLevelBefore = level == UCOL_SECONDARY ? HAS_BEFORE2 : HAS_BEFORE3;
        if((node & hasThisLevelBefore) == 0) {
            // The parent node has an implied level-common weight.
            int64_t commonNode =
                nodeFromWeight16(Collation::COMMON_WEIGHT16) | nodeFromStrength(level);
            if(level == UCOL_SECONDARY) {
                // Move the HAS_BEFORE3 flag from the parent node
                // to the new secondary common node.
                commonNode |= node & HAS_BEFORE3;
                node &= ~(int64_t)HAS_BEFORE3;
            }
            nodes.setElementAt(node | hasThisLevelBefore, index);
            // Insert below-common-weight node.
            int32_t nextIndex = nextIndexFromNode(node);
            node = nodeFromWeight16(weight16) | nodeFromStrength(level);
            index = insertNodeBetween(index, nextIndex, node, errorCode);
            // Insert common-weight node.
            insertNodeBetween(index, nextIndex, commonNode, errorCode);
            // Return index of below-common-weight node.
            return index;
        }
    }

    // Find the root CE's weight for this level.
    // Postpone insertion if not found:
    // Insert the new root node before the next stronger node,
    // or before the next root node with the same strength and a larger weight.
    int32_t nextIndex;
    while((nextIndex = nextIndexFromNode(node)) != 0) {
        node = nodes.elementAti(nextIndex);
        int32_t nextStrength = strengthFromNode(node);
        if(nextStrength <= level) {
            // Insert before a stronger node.
            if(nextStrength < level) { break; }
            // nextStrength == level
            if(!isTailoredNode(node)) {
                uint32_t nextWeight16 = weight16FromNode(node);
                if(nextWeight16 == weight16) {
                    // Found the node for the root CE up to this level.
                    return nextIndex;
                }
                // Insert before a node with a larger same-strength weight.
                if(nextWeight16 > weight16) { break; }
            }
        }
        // Skip the next node.
        index = nextIndex;
    }
    node = nodeFromWeight16(weight16) | nodeFromStrength(level);
    return insertNodeBetween(index, nextIndex, node, errorCode);
}

int32_t
CollationBuilder::insertTailoredNodeAfter(int32_t index, int32_t strength, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    U_ASSERT(0 <= index && index < nodes.size());
    if(strength >= UCOL_SECONDARY) {
        index = findCommonNode(index, UCOL_SECONDARY);
        if(strength >= UCOL_TERTIARY) {
            index = findCommonNode(index, UCOL_TERTIARY);
        }
    }
    // Postpone insertion:
    // Insert the new node before the next one with a strength at least as strong.
    int64_t node = nodes.elementAti(index);
    int32_t nextIndex;
    while((nextIndex = nextIndexFromNode(node)) != 0) {
        node = nodes.elementAti(nextIndex);
        if(strengthFromNode(node) <= strength) { break; }
        // Skip the next node which has a weaker (larger) strength than the new one.
        index = nextIndex;
    }
    node = IS_TAILORED | nodeFromStrength(strength);
    return insertNodeBetween(index, nextIndex, node, errorCode);
}

int32_t
CollationBuilder::insertNodeBetween(int32_t index, int32_t nextIndex, int64_t node,
                                    UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    U_ASSERT(previousIndexFromNode(node) == 0);
    U_ASSERT(nextIndexFromNode(node) == 0);
    U_ASSERT(nextIndexFromNode(nodes.elementAti(index)) == nextIndex);
    // Append the new node and link it to the existing nodes.
    int32_t newIndex = nodes.size();
    node |= nodeFromPreviousIndex(index) | nodeFromNextIndex(nextIndex);
    nodes.addElement(node, errorCode);
    if(U_FAILURE(errorCode)) { return 0; }
    // nodes[index].nextIndex = newIndex
    node = nodes.elementAti(index);
    nodes.setElementAt(changeNodeNextIndex(node, newIndex), index);
    // nodes[nextIndex].previousIndex = newIndex
    if(nextIndex != 0) {
        node = nodes.elementAti(nextIndex);
        nodes.setElementAt(changeNodePreviousIndex(node, newIndex), nextIndex);
    }
    return newIndex;
}

int32_t
CollationBuilder::findCommonNode(int32_t index, int32_t strength) const {
    U_ASSERT(UCOL_SECONDARY <= strength && strength <= UCOL_TERTIARY);
    int64_t node = nodes.elementAti(index);
    if(strengthFromNode(node) >= strength) {
        // The current node is no stronger.
        return index;
    }
    if(strength == UCOL_SECONDARY ? !nodeHasBefore2(node) : !nodeHasBefore3(node)) {
        // The current node implies the strength-common weight.
        return index;
    }
    index = nextIndexFromNode(node);
    node = nodes.elementAti(index);
    U_ASSERT(!isTailoredNode(node) && strengthFromNode(node) == strength &&
            weight16FromNode(node) < Collation::COMMON_WEIGHT16);
    // Skip to the explicit common node.
    do {
        index = nextIndexFromNode(node);
        node = nodes.elementAti(index);
        U_ASSERT(strengthFromNode(node) >= strength);
    } while(isTailoredNode(node) || strengthFromNode(node) > strength ||
            weight16FromNode(node) < Collation::COMMON_WEIGHT16);
    U_ASSERT(weight16FromNode(node) == Collation::COMMON_WEIGHT16);
    return index;
}

void
CollationBuilder::setCaseBits(const UnicodeString &nfdString,
                              const char *&parserErrorReason, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    int32_t numTailoredPrimaries = 0;
    for(int32_t i = 0; i < cesLength; ++i) {
        if(ceStrength(ces[i]) == UCOL_PRIMARY) { ++numTailoredPrimaries; }
    }
    // We should not be able to get too many case bits because
    // cesLength<=31==MAX_EXPANSION_LENGTH.
    // 31 pairs of case bits fit into an int64_t without setting its sign bit.
    U_ASSERT(numTailoredPrimaries <= 31);

    int64_t cases = 0;
    if(numTailoredPrimaries > 0) {
        const UChar *s = nfdString.getBuffer();
        UTF16CollationIterator baseCEs(baseData, FALSE, s, s, s + nfdString.length());
        int32_t baseCEsLength = baseCEs.fetchCEs(errorCode) - 1;
        if(U_FAILURE(errorCode)) {
            parserErrorReason = "fetching root CEs for tailored string";
            return;
        }
        U_ASSERT(baseCEsLength >= 0 && baseCEs.getCE(baseCEsLength) == Collation::NO_CE);

        uint32_t lastCase = 0;
        int32_t numBasePrimaries = 0;
        for(int32_t i = 0; i < baseCEsLength; ++i) {
            int64_t ce = baseCEs.getCE(i);
            if((ce >> 32) != 0) {
                ++numBasePrimaries;
                uint32_t c = ((uint32_t)ce >> 14) & 3;
                U_ASSERT(c == 0 || c == 2);  // lowercase or uppercase, no mixed case in any base CE
                if(numBasePrimaries < numTailoredPrimaries) {
                    cases |= (int64_t)c << ((numBasePrimaries - 1) * 2);
                } else if(numBasePrimaries == numTailoredPrimaries) {
                    lastCase = c;
                } else if(c != lastCase) {
                    // There are more base primary CEs than tailored primaries.
                    // Set mixed case if the case bits of the remainder differ.
                    lastCase = 1;
                    // Nothing more can change.
                    break;
                }
            }
        }
        if(numBasePrimaries >= numTailoredPrimaries) {
            cases |= (int64_t)lastCase << ((numTailoredPrimaries - 1) * 2);
        }
    }

    for(int32_t i = 0; i < cesLength; ++i) {
        int64_t ce = ces[i] & INT64_C(0xffffffffffff3fff);  // clear old case bits
        int32_t strength = ceStrength(ce);
        if(strength == UCOL_PRIMARY) {
            ce |= (cases & 3) << 14;
            cases >>= 2;
        } else if(strength == UCOL_TERTIARY) {
            // Tertiary CEs must have uppercase bits.
            // See the LDML spec, and comments in class CollationCompare.
            ce |= 0x8000;
        }
        // Tertiary ignorable CEs must have 0 case bits.
        // We set 0 case bits for secondary CEs too
        // since currently only U+0345 is cased and maps to a secondary CE,
        // and it is lowercase. Other secondaries are uncased.
        // See [[:Cased:]&[:uca1=:]] where uca1 queries the root primary weight.
        ces[i] = ce;
    }
}

void
CollationBuilder::suppressContractions(const UnicodeSet &set, const char *&parserErrorReason,
                                       UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    dataBuilder->suppressContractions(set, errorCode);
    if(U_FAILURE(errorCode)) {
        parserErrorReason = "application of [suppressContractions [set]] failed";
    }
}

void
CollationBuilder::optimize(const UnicodeSet &set, const char *& /* parserErrorReason */,
                           UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    optimizeSet.addAll(set);
}

uint32_t
CollationBuilder::addWithClosure(const UnicodeString &nfdPrefix, const UnicodeString &nfdString,
                                 const int64_t newCEs[], int32_t newCEsLength, uint32_t ce32,
                                 UErrorCode &errorCode) {
    // Map from the NFD input to the CEs.
    ce32 = addIfDifferent(nfdPrefix, nfdString, newCEs, newCEsLength, ce32, errorCode);
    ce32 = addOnlyClosure(nfdPrefix, nfdString, newCEs, newCEsLength, ce32, errorCode);
    addTailComposites(nfdPrefix, nfdString, errorCode);
    return ce32;
}

uint32_t
CollationBuilder::addOnlyClosure(const UnicodeString &nfdPrefix, const UnicodeString &nfdString,
                                 const int64_t newCEs[], int32_t newCEsLength, uint32_t ce32,
                                 UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return ce32; }

    // Map from canonically equivalent input to the CEs. (But not from the all-NFD input.)
    if(nfdPrefix.isEmpty()) {
        CanonicalIterator stringIter(nfdString, errorCode);
        if(U_FAILURE(errorCode)) { return ce32; }
        UnicodeString prefix;
        for(;;) {
            UnicodeString str = stringIter.next();
            if(str.isBogus()) { break; }
            if(ignoreString(str, errorCode) || str == nfdString) { continue; }
            ce32 = addIfDifferent(prefix, str, newCEs, newCEsLength, ce32, errorCode);
            if(U_FAILURE(errorCode)) { return ce32; }
        }
    } else {
        CanonicalIterator prefixIter(nfdPrefix, errorCode);
        CanonicalIterator stringIter(nfdString, errorCode);
        if(U_FAILURE(errorCode)) { return ce32; }
        for(;;) {
            UnicodeString prefix = prefixIter.next();
            if(prefix.isBogus()) { break; }
            if(ignorePrefix(prefix, errorCode)) { continue; }
            UBool samePrefix = prefix == nfdPrefix;
            for(;;) {
                UnicodeString str = stringIter.next();
                if(str.isBogus()) { break; }
                if(ignoreString(str, errorCode) || (samePrefix && str == nfdString)) { continue; }
                ce32 = addIfDifferent(prefix, str, newCEs, newCEsLength, ce32, errorCode);
                if(U_FAILURE(errorCode)) { return ce32; }
            }
            stringIter.reset();
        }
    }
    return ce32;
}

void
CollationBuilder::addTailComposites(const UnicodeString &nfdPrefix, const UnicodeString &nfdString,
                                    UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }

    // Look for the last starter in the NFD string.
    UChar32 lastStarter;
    int32_t indexAfterLastStarter = nfdString.length();
    for(;;) {
        if(indexAfterLastStarter == 0) { return; }  // no starter at all
        lastStarter = nfdString.char32At(indexAfterLastStarter - 1);
        if(nfd.getCombiningClass(lastStarter) == 0) { break; }
        indexAfterLastStarter -= U16_LENGTH(lastStarter);
    }
    // No closure to Hangul syllables since we decompose them on the fly.
    if(Hangul::isJamoL(lastStarter)) { return; }

    // Are there any composites whose decomposition starts with the lastStarter?
    // Note: Normalizer2Impl does not currently return start sets for NFC_QC=Maybe characters.
    // We might find some more equivalent mappings here if it did.
    UnicodeSet composites;
    if(!nfcImpl.getCanonStartSet(lastStarter, composites)) { return; }

    UnicodeString decomp;
    UnicodeString newNFDString, newString;
    int64_t newCEs[Collation::MAX_EXPANSION_LENGTH];
    UnicodeSetIterator iter(composites);
    while(iter.next()) {
        U_ASSERT(!iter.isString());
        UChar32 composite = iter.getCodepoint();
        nfd.getDecomposition(composite, decomp);
        if(!mergeCompositeIntoString(nfdString, indexAfterLastStarter, composite, decomp,
                                     newNFDString, newString, errorCode)) {
            continue;
        }
        int32_t newCEsLength = dataBuilder->getCEs(nfdPrefix, newNFDString, newCEs, 0);
        if(newCEsLength > Collation::MAX_EXPANSION_LENGTH) {
            // Ignore mappings that we cannot store.
            continue;
        }
        // Note: It is possible that the newCEs do not make use of the mapping
        // for which we are adding the tail composites, in which case we might be adding
        // unnecessary mappings.
        // For example, when we add tail composites for ae^ (^=combining circumflex),
        // UCA discontiguous-contraction matching does not find any matches
        // for ae_^ (_=any combining diacritic below) *unless* there is also
        // a contraction mapping for ae.
        // Thus, if there is no ae contraction, then the ae^ mapping is ignored
        // while fetching the newCEs for ae_^.
        // TODO: Try to detect this effectively.
        // (Alternatively, print a warning when prefix contractions are missing.)

        // We do not need an explicit mapping for the NFD strings.
        // It is fine if the NFD input collates like this via a sequence of mappings.
        // It also saves a little bit of space, and may reduce the set of characters with contractions.
        uint32_t ce32 = addIfDifferent(nfdPrefix, newString,
                                       newCEs, newCEsLength, Collation::UNASSIGNED_CE32, errorCode);
        if(ce32 != Collation::UNASSIGNED_CE32) {
            // was different, was added
            addOnlyClosure(nfdPrefix, newNFDString, newCEs, newCEsLength, ce32, errorCode);
        }
    }
}

UBool
CollationBuilder::mergeCompositeIntoString(const UnicodeString &nfdString,
                                           int32_t indexAfterLastStarter,
                                           UChar32 composite, const UnicodeString &decomp,
                                           UnicodeString &newNFDString, UnicodeString &newString,
                                           UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return FALSE; }
    U_ASSERT(nfdString.char32At(indexAfterLastStarter - 1) == decomp.char32At(0));
    int32_t lastStarterLength = decomp.moveIndex32(0, 1);
    if(lastStarterLength == decomp.length()) {
        // Singleton decompositions should be found by addWithClosure()
        // and the CanonicalIterator, so we can ignore them here.
        return FALSE;
    }
    if(nfdString.compare(indexAfterLastStarter, 0x7fffffff,
                         decomp, lastStarterLength, 0x7fffffff) == 0) {
        // same strings, nothing new to be found here
        return FALSE;
    }

    // Make new FCD strings that combine a composite, or its decomposition,
    // into the nfdString's last starter and the combining marks following it.
    // Make an NFD version, and a version with the composite.
    newNFDString.setTo(nfdString, 0, indexAfterLastStarter);
    newString.setTo(nfdString, 0, indexAfterLastStarter - lastStarterLength).append(composite);

    // The following is related to discontiguous contraction matching,
    // but builds only FCD strings (or else returns FALSE).
    int32_t sourceIndex = indexAfterLastStarter;
    int32_t decompIndex = lastStarterLength;
    // Small optimization: We keep the source character across loop iterations
    // because we do not always consume it,
    // and then need not fetch it again nor look up its combining class again.
    UChar32 sourceChar = U_SENTINEL;
    // The cc variables need to be declared before the loop so that at the end
    // they are set to the last combining classes seen.
    uint8_t sourceCC = 0;
    uint8_t decompCC = 0;
    for(;;) {
        if(sourceChar < 0) {
            if(sourceIndex >= nfdString.length()) { break; }
            sourceChar = nfdString.char32At(sourceIndex);
            sourceCC = nfd.getCombiningClass(sourceChar);
            U_ASSERT(sourceCC != 0);
        }
        // We consume a decomposition character in each iteration.
        if(decompIndex >= decomp.length()) { break; }
        UChar32 decompChar = decomp.char32At(decompIndex);
        decompCC = nfd.getCombiningClass(decompChar);
        // Compare the two characters and their combining classes.
        if(decompCC == 0) {
            // Unable to merge because the source contains a non-zero combining mark
            // but the composite's decomposition contains another starter.
            // The strings would not be equivalent.
            return FALSE;
        } else if(sourceCC < decompCC) {
            // Composite + sourceChar would not be FCD.
            return FALSE;
        } else if(decompCC < sourceCC) {
            newNFDString.append(decompChar);
            decompIndex += U16_LENGTH(decompChar);
        } else if(decompChar != sourceChar) {
            // Blocked because same combining class.
            return FALSE;
        } else {  // match: decompChar == sourceChar
            newNFDString.append(decompChar);
            decompIndex += U16_LENGTH(decompChar);
            sourceIndex += U16_LENGTH(decompChar);
            sourceChar = U_SENTINEL;
        }
    }
    // We are at the end of at least one of the two inputs.
    if(sourceChar >= 0) {  // more characters from nfdString but not from decomp
        if(sourceCC < decompCC) {
            // Appending the next source character to the composite would not be FCD.
            return FALSE;
        }
        newNFDString.append(nfdString, sourceIndex, 0x7fffffff);
        newString.append(nfdString, sourceIndex, 0x7fffffff);
    } else if(decompIndex < decomp.length()) {  // more characters from decomp, not from nfdString
        newNFDString.append(decomp, decompIndex, 0x7fffffff);
    }
    U_ASSERT(nfd.isNormalized(newNFDString, errorCode));
    U_ASSERT(fcd.isNormalized(newString, errorCode));
    U_ASSERT(nfd.normalize(newString, errorCode) == newNFDString);  // canonically equivalent
    return TRUE;
}

UBool
CollationBuilder::ignorePrefix(const UnicodeString &s, UErrorCode &errorCode) const {
    // Do not map non-FCD prefixes.
    return !isFCD(s, errorCode);
}

UBool
CollationBuilder::ignoreString(const UnicodeString &s, UErrorCode &errorCode) const {
    // Do not map non-FCD strings.
    // Do not map strings that start with Hangul syllables: We decompose those on the fly.
    return !isFCD(s, errorCode) || Hangul::isHangul(s.charAt(0));
}

UBool
CollationBuilder::isFCD(const UnicodeString &s, UErrorCode &errorCode) const {
    return U_SUCCESS(errorCode) && fcd.isNormalized(s, errorCode);
}

void
CollationBuilder::closeOverComposites(UErrorCode &errorCode) {
    UnicodeSet composites(UNICODE_STRING_SIMPLE("[:NFD_QC=N:]"), errorCode);  // Java: static final
    if(U_FAILURE(errorCode)) { return; }
    // Hangul is decomposed on the fly during collation.
    composites.remove(Hangul::HANGUL_BASE, Hangul::HANGUL_END);
    UnicodeString prefix;  // empty
    UnicodeString nfdString;
    UnicodeSetIterator iter(composites);
    while(iter.next()) {
        U_ASSERT(!iter.isString());
        nfd.getDecomposition(iter.getCodepoint(), nfdString);
        cesLength = dataBuilder->getCEs(nfdString, ces, 0);
        if(cesLength > Collation::MAX_EXPANSION_LENGTH) {
            // Too many CEs from the decomposition (unusual), ignore this composite.
            // We could add a capacity parameter to getCEs() and reallocate if necessary.
            // However, this can only really happen in contrived cases.
            continue;
        }
        const UnicodeString &composite(iter.getString());
        addIfDifferent(prefix, composite, ces, cesLength, Collation::UNASSIGNED_CE32, errorCode);
    }
}

uint32_t
CollationBuilder::addIfDifferent(const UnicodeString &prefix, const UnicodeString &str,
                                 const int64_t newCEs[], int32_t newCEsLength, uint32_t ce32,
                                 UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return ce32; }
    int64_t oldCEs[Collation::MAX_EXPANSION_LENGTH];
    int32_t oldCEsLength = dataBuilder->getCEs(prefix, str, oldCEs, 0);
    if(!sameCEs(newCEs, newCEsLength, oldCEs, oldCEsLength)) {
        if(ce32 == Collation::UNASSIGNED_CE32) {
            ce32 = dataBuilder->encodeCEs(newCEs, newCEsLength, errorCode);
        }
        dataBuilder->addCE32(prefix, str, ce32, errorCode);
    }
    return ce32;
}

UBool
CollationBuilder::sameCEs(const int64_t ces1[], int32_t ces1Length,
                          const int64_t ces2[], int32_t ces2Length) {
    if(ces1Length != ces2Length) {
        return FALSE;
    }
    U_ASSERT(ces1Length <= Collation::MAX_EXPANSION_LENGTH);
    for(int32_t i = 0; i < ces1Length; ++i) {
        if(ces1[i] != ces2[i]) { return FALSE; }
    }
    return TRUE;
}

#ifdef DEBUG_COLLATION_BUILDER

uint32_t
alignWeightRight(uint32_t w) {
    if(w != 0) {
        while((w & 0xff) == 0) { w >>= 8; }
    }
    return w;
}

#endif

void
CollationBuilder::makeTailoredCEs(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }

    CollationWeights primaries, secondaries, tertiaries;
    int64_t *nodesArray = nodes.getBuffer();
#ifdef DEBUG_COLLATION_BUILDER
        puts("\nCollationBuilder::makeTailoredCEs()");
#endif

    for(int32_t rpi = 0; rpi < rootPrimaryIndexes.size(); ++rpi) {
        int32_t i = rootPrimaryIndexes.elementAti(rpi);
        int64_t node = nodesArray[i];
        uint32_t p = weight32FromNode(node);
        uint32_t s = p == 0 ? 0 : Collation::COMMON_WEIGHT16;
        uint32_t t = s;
        uint32_t q = 0;
        UBool pIsTailored = FALSE;
        UBool sIsTailored = FALSE;
        UBool tIsTailored = FALSE;
#ifdef DEBUG_COLLATION_BUILDER
        printf("\nprimary     %lx\n", (long)alignWeightRight(p));
#endif
        int32_t pIndex = p == 0 ? 0 : rootElements.findPrimary(p);
        int32_t nextIndex = nextIndexFromNode(node);
        while(nextIndex != 0) {
            i = nextIndex;
            node = nodesArray[i];
            nextIndex = nextIndexFromNode(node);
            int32_t strength = strengthFromNode(node);
            if(strength == UCOL_QUATERNARY) {
                U_ASSERT(isTailoredNode(node));
#ifdef DEBUG_COLLATION_BUILDER
                printf("      quat+     ");
#endif
                if(q == 3) {
                    errorCode = U_BUFFER_OVERFLOW_ERROR;
                    errorReason = "quaternary tailoring gap too small";
                    return;
                }
                ++q;
            } else {
                if(strength == UCOL_TERTIARY) {
                    if(isTailoredNode(node)) {
#ifdef DEBUG_COLLATION_BUILDER
                        printf("    ter+        ");
#endif
                        if(!tIsTailored) {
                            // First tailored tertiary node for [p, s].
                            int32_t tCount = countTailoredNodes(nodesArray, nextIndex,
                                                                UCOL_TERTIARY) + 1;
                            uint32_t tLimit;
                            if(t == 0) {
                                // Gap at the beginning of the tertiary CE range.
                                t = rootElements.getTertiaryBoundary() - 0x100;
                                tLimit = rootElements.getFirstTertiaryCE() & Collation::ONLY_TERTIARY_MASK;
                            } else if(!pIsTailored && !sIsTailored) {
                                // p and s are root weights.
                                tLimit = rootElements.getTertiaryAfter(pIndex, s, t);
                            } else if(t == Collation::BEFORE_WEIGHT16) {
                                tLimit = Collation::COMMON_WEIGHT16;
                            } else {
                                // [p, s] is tailored.
                                U_ASSERT(t == Collation::COMMON_WEIGHT16);
                                tLimit = rootElements.getTertiaryBoundary();
                            }
                            U_ASSERT(tLimit == 0x4000 || (tLimit & ~Collation::ONLY_TERTIARY_MASK) == 0);
                            tertiaries.initForTertiary();
                            if(!tertiaries.allocWeights(t, tLimit, tCount)) {
                                errorCode = U_BUFFER_OVERFLOW_ERROR;
                                errorReason = "tertiary tailoring gap too small";
                                return;
                            }
                            tIsTailored = TRUE;
                        }
                        t = tertiaries.nextWeight();
                        U_ASSERT(t != 0xffffffff);
                    } else {
                        t = weight16FromNode(node);
                        tIsTailored = FALSE;
#ifdef DEBUG_COLLATION_BUILDER
                        printf("    ter     %lx\n", (long)alignWeightRight(t));
#endif
                    }
                } else {
                    if(strength == UCOL_SECONDARY) {
                        if(isTailoredNode(node)) {
#ifdef DEBUG_COLLATION_BUILDER
                            printf("  sec+          ");
#endif
                            if(!sIsTailored) {
                                // First tailored secondary node for p.
                                int32_t sCount = countTailoredNodes(nodesArray, nextIndex,
                                                                    UCOL_SECONDARY) + 1;
                                uint32_t sLimit;
                                if(s == 0) {
                                    // Gap at the beginning of the secondary CE range.
                                    s = rootElements.getSecondaryBoundary() - 0x100;
                                    sLimit = rootElements.getFirstSecondaryCE() >> 16;
                                } else if(!pIsTailored) {
                                    // p is a root primary.
                                    sLimit = rootElements.getSecondaryAfter(pIndex, s);
                                } else if(s == Collation::BEFORE_WEIGHT16) {
                                    sLimit = Collation::COMMON_WEIGHT16;
                                } else {
                                    // p is a tailored primary.
                                    U_ASSERT(s == Collation::COMMON_WEIGHT16);
                                    sLimit = rootElements.getSecondaryBoundary();
                                }
                                if(s == Collation::COMMON_WEIGHT16) {
                                    // Do not tailor into the getSortKey() range of
                                    // compressed common secondaries.
                                    s = rootElements.getLastCommonSecondary();
                                }
                                secondaries.initForSecondary();
                                if(!secondaries.allocWeights(s, sLimit, sCount)) {
                                    errorCode = U_BUFFER_OVERFLOW_ERROR;
                                    errorReason = "secondary tailoring gap too small";
#ifdef DEBUG_COLLATION_BUILDER
                                    printf("!secondaries.allocWeights(%lx, %lx, sCount=%ld)\n",
                                           (long)alignWeightRight(s), (long)alignWeightRight(sLimit),
                                           (long)alignWeightRight(sCount));
#endif
                                    return;
                                }
                                sIsTailored = TRUE;
                            }
                            s = secondaries.nextWeight();
                            U_ASSERT(s != 0xffffffff);
                        } else {
                            s = weight16FromNode(node);
                            sIsTailored = FALSE;
#ifdef DEBUG_COLLATION_BUILDER
                            printf("  sec       %lx\n", (long)alignWeightRight(s));
#endif
                        }
                    } else /* UCOL_PRIMARY */ {
                        U_ASSERT(isTailoredNode(node));
#ifdef DEBUG_COLLATION_BUILDER
                        printf("pri+            ");
#endif
                        if(!pIsTailored) {
                            // First tailored primary node in this list.
                            int32_t pCount = countTailoredNodes(nodesArray, nextIndex,
                                                                UCOL_PRIMARY) + 1;
                            UBool isCompressible = baseData->isCompressiblePrimary(p);
                            uint32_t pLimit =
                                rootElements.getPrimaryAfter(p, pIndex, isCompressible);
                            primaries.initForPrimary(isCompressible);
                            if(!primaries.allocWeights(p, pLimit, pCount)) {
                                errorCode = U_BUFFER_OVERFLOW_ERROR;  // TODO: introduce a more specific UErrorCode?
                                errorReason = "primary tailoring gap too small";
                                return;
                            }
                            pIsTailored = TRUE;
                        }
                        p = primaries.nextWeight();
                        U_ASSERT(p != 0xffffffff);
                        s = Collation::COMMON_WEIGHT16;
                        sIsTailored = FALSE;
                    }
                    t = s == 0 ? 0 : Collation::COMMON_WEIGHT16;
                    tIsTailored = FALSE;
                }
                q = 0;
            }
            if(isTailoredNode(node)) {
                nodesArray[i] = Collation::makeCE(p, s, t, q);
#ifdef DEBUG_COLLATION_BUILDER
                printf("%016llx\n", (long long)nodesArray[i]);
#endif
            }
        }
    }
}

int32_t
CollationBuilder::countTailoredNodes(const int64_t *nodesArray, int32_t i, int32_t strength) {
    int32_t count = 0;
    for(;;) {
        if(i == 0) { break; }
        int64_t node = nodesArray[i];
        if(strengthFromNode(node) < strength) { break; }
        if(strengthFromNode(node) == strength) {
            if(isTailoredNode(node)) {
                ++count;
            } else {
                break;
            }
        }
        i = nextIndexFromNode(node);
    }
    return count;
}

class CEFinalizer : public CollationDataBuilder::CEModifier {
public:
    CEFinalizer(const int64_t *ces) : finalCEs(ces) {}
    virtual ~CEFinalizer();
    virtual int64_t modifyCE32(uint32_t ce32) const {
        U_ASSERT(!Collation::isSpecialCE32(ce32));
        if(CollationBuilder::isTempCE32(ce32)) {
            // retain case bits
            return finalCEs[CollationBuilder::indexFromTempCE32(ce32)] | ((ce32 & 0xc0) << 8);
        } else {
            return Collation::NO_CE;
        }
    }
    virtual int64_t modifyCE(int64_t ce) const {
        if(CollationBuilder::isTempCE(ce)) {
            // retain case bits
            return finalCEs[CollationBuilder::indexFromTempCE(ce)] | (ce & 0xc000);
        } else {
            return Collation::NO_CE;
        }
    }

private:
    const int64_t *finalCEs;
};

CEFinalizer::~CEFinalizer() {}

void
CollationBuilder::finalizeCEs(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    LocalPointer<CollationDataBuilder> newBuilder(new CollationDataBuilder(errorCode), errorCode);
    if(U_FAILURE(errorCode)) {
        return;
    }
    newBuilder->initForTailoring(baseData, errorCode);
    CEFinalizer finalizer(nodes.getBuffer());
    newBuilder->copyFrom(*dataBuilder, finalizer, errorCode);
    if(U_FAILURE(errorCode)) { return; }
    delete dataBuilder;
    dataBuilder = newBuilder.orphan();
}

int32_t
CollationBuilder::ceStrength(int64_t ce) {
    return
        isTempCE(ce) ? strengthFromTempCE(ce) :
        (ce & INT64_C(0xff00000000000000)) != 0 ? UCOL_PRIMARY :
        ((uint32_t)ce & 0xff000000) != 0 ? UCOL_SECONDARY :
        ce != 0 ? UCOL_TERTIARY :
        UCOL_IDENTICAL;
}

U_NAMESPACE_END

U_NAMESPACE_USE

U_CAPI UCollator * U_EXPORT2
ucol_openRules(const UChar *rules, int32_t rulesLength,
               UColAttributeValue normalizationMode, UCollationStrength strength,
               UParseError *parseError, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) { return NULL; }
    if(rules == NULL && rulesLength != 0) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    RuleBasedCollator *coll = new RuleBasedCollator();
    if(coll == NULL) {
        *pErrorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    UnicodeString r((UBool)(rulesLength < 0), rules, rulesLength);
    coll->internalBuildTailoring(r, strength, normalizationMode, parseError, NULL, *pErrorCode);
    if(U_FAILURE(*pErrorCode)) {
        delete coll;
        return NULL;
    }
    return coll->toUCollator();
}

static const int32_t collationbuilder_internalBufferSize = 512;

// The @internal ucol_getUnsafeSet() was moved here from ucol_sit.cpp
// because it calls UnicodeSet "builder" code that depends on all Unicode properties,
// and the rest of the collation "runtime" code only depends on normalization.
// This function is not related to the collation builder,
// but it did not seem worth moving it into its own .cpp file,
// nor rewriting it to use lower-level UnicodeSet and Normalizer2Impl methods.
U_CAPI int32_t U_EXPORT2
ucol_getUnsafeSet( const UCollator *coll,
                  USet *unsafe,
                  UErrorCode *status)
{
    UChar buffer[collationbuilder_internalBufferSize];
    int32_t len = 0;

    uset_clear(unsafe);

    // cccpattern = "[[:^tccc=0:][:^lccc=0:]]", unfortunately variant
    static const UChar cccpattern[25] = { 0x5b, 0x5b, 0x3a, 0x5e, 0x74, 0x63, 0x63, 0x63, 0x3d, 0x30, 0x3a, 0x5d,
                                    0x5b, 0x3a, 0x5e, 0x6c, 0x63, 0x63, 0x63, 0x3d, 0x30, 0x3a, 0x5d, 0x5d, 0x00 };

    // add chars that fail the fcd check
    uset_applyPattern(unsafe, cccpattern, 24, USET_IGNORE_SPACE, status);

    // add lead/trail surrogates
    // (trail surrogates should need to be unsafe only if the caller tests for UTF-16 code *units*,
    // not when testing code *points*)
    uset_addRange(unsafe, 0xd800, 0xdfff);

    USet *contractions = uset_open(0,0);

    int32_t i = 0, j = 0;
    ucol_getContractionsAndExpansions(coll, contractions, NULL, FALSE, status);
    int32_t contsSize = uset_size(contractions);
    UChar32 c = 0;
    // Contraction set consists only of strings
    // to get unsafe code points, we need to
    // break the strings apart and add them to the unsafe set
    for(i = 0; i < contsSize; i++) {
        len = uset_getItem(contractions, i, NULL, NULL, buffer, collationbuilder_internalBufferSize, status);
        if(len > 0) {
            j = 0;
            while(j < len) {
                U16_NEXT(buffer, j, len, c);
                if(j < len) {
                    uset_add(unsafe, c);
                }
            }
        }
    }

    uset_close(contractions);

    return uset_size(unsafe);
}

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1996-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationcompare.cpp
*
* created on: 2012feb14 with new and old collation code
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1996-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationcompare.h
*
* created on: 2012feb14 with new and old collation code
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION



U_NAMESPACE_BEGIN

class CollationIterator;
struct CollationSettings;

class U_I18N_API CollationCompare /* not : public UObject because all methods are static */ {
public:
    static UCollationResult compareUpToQuaternary(CollationIterator &left, CollationIterator &right,
                                                  const CollationSettings &settings,
                                                  UErrorCode &errorCode);
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION





U_NAMESPACE_BEGIN

UCollationResult
CollationCompare::compareUpToQuaternary(CollationIterator &left, CollationIterator &right,
                                        const CollationSettings &settings,
                                        UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return UCOL_EQUAL; }

    int32_t options = settings.options;
    uint32_t variableTop;
    if((options & CollationSettings::ALTERNATE_MASK) == 0) {
        variableTop = 0;
    } else {
        // +1 so that we can use "<" and primary ignorables test out early.
        variableTop = settings.variableTop + 1;
    }
    UBool anyVariable = FALSE;

    // Fetch CEs, compare primaries, store secondary & tertiary weights.
    for(;;) {
        // We fetch CEs until we get a non-ignorable primary or reach the end.
        uint32_t leftPrimary;
        do {
            int64_t ce = left.nextCE(errorCode);
            leftPrimary = (uint32_t)(ce >> 32);
            if(leftPrimary < variableTop && leftPrimary > Collation::MERGE_SEPARATOR_PRIMARY) {
                // Variable CE, shift it to quaternary level.
                // Ignore all following primary ignorables, and shift further variable CEs.
                anyVariable = TRUE;
                do {
                    // Store only the primary of the variable CE.
                    left.setCurrentCE(ce & INT64_C(0xffffffff00000000));
                    for(;;) {
                        ce = left.nextCE(errorCode);
                        leftPrimary = (uint32_t)(ce >> 32);
                        if(leftPrimary == 0) {
                            left.setCurrentCE(0);
                        } else {
                            break;
                        }
                    }
                } while(leftPrimary < variableTop &&
                        leftPrimary > Collation::MERGE_SEPARATOR_PRIMARY);
            }
        } while(leftPrimary == 0);

        uint32_t rightPrimary;
        do {
            int64_t ce = right.nextCE(errorCode);
            rightPrimary = (uint32_t)(ce >> 32);
            if(rightPrimary < variableTop && rightPrimary > Collation::MERGE_SEPARATOR_PRIMARY) {
                // Variable CE, shift it to quaternary level.
                // Ignore all following primary ignorables, and shift further variable CEs.
                anyVariable = TRUE;
                do {
                    // Store only the primary of the variable CE.
                    right.setCurrentCE(ce & INT64_C(0xffffffff00000000));
                    for(;;) {
                        ce = right.nextCE(errorCode);
                        rightPrimary = (uint32_t)(ce >> 32);
                        if(rightPrimary == 0) {
                            right.setCurrentCE(0);
                        } else {
                            break;
                        }
                    }
                } while(rightPrimary < variableTop &&
                        rightPrimary > Collation::MERGE_SEPARATOR_PRIMARY);
            }
        } while(rightPrimary == 0);

        if(leftPrimary != rightPrimary) {
            // Return the primary difference, with script reordering.
            if(settings.hasReordering()) {
                leftPrimary = settings.reorder(leftPrimary);
                rightPrimary = settings.reorder(rightPrimary);
            }
            return (leftPrimary < rightPrimary) ? UCOL_LESS : UCOL_GREATER;
        }
        if(leftPrimary == Collation::NO_CE_PRIMARY) { break; }
    }
    if(U_FAILURE(errorCode)) { return UCOL_EQUAL; }

    // Compare the buffered secondary & tertiary weights.
    // We might skip the secondary level but continue with the case level
    // which is turned on separately.
    if(CollationSettings::getStrength(options) >= UCOL_SECONDARY) {
        if((options & CollationSettings::BACKWARD_SECONDARY) == 0) {
            int32_t leftIndex = 0;
            int32_t rightIndex = 0;
            for(;;) {
                uint32_t leftSecondary;
                do {
                    leftSecondary = ((uint32_t)left.getCE(leftIndex++)) >> 16;
                } while(leftSecondary == 0);

                uint32_t rightSecondary;
                do {
                    rightSecondary = ((uint32_t)right.getCE(rightIndex++)) >> 16;
                } while(rightSecondary == 0);

                if(leftSecondary != rightSecondary) {
                    return (leftSecondary < rightSecondary) ? UCOL_LESS : UCOL_GREATER;
                }
                if(leftSecondary == Collation::NO_CE_WEIGHT16) { break; }
            }
        } else {
            // The backwards secondary level compares secondary weights backwards
            // within segments separated by the merge separator (U+FFFE, weight 02).
            int32_t leftStart = 0;
            int32_t rightStart = 0;
            for(;;) {
                // Find the merge separator or the NO_CE terminator.
                uint32_t p;
                int32_t leftLimit = leftStart;
                while((p = (uint32_t)(left.getCE(leftLimit) >> 32)) >
                            Collation::MERGE_SEPARATOR_PRIMARY ||
                        p == 0) {
                    ++leftLimit;
                }
                int32_t rightLimit = rightStart;
                while((p = (uint32_t)(right.getCE(rightLimit) >> 32)) >
                            Collation::MERGE_SEPARATOR_PRIMARY ||
                        p == 0) {
                    ++rightLimit;
                }

                // Compare the segments.
                int32_t leftIndex = leftLimit;
                int32_t rightIndex = rightLimit;
                for(;;) {
                    int32_t leftSecondary = 0;
                    while(leftSecondary == 0 && leftIndex > leftStart) {
                        leftSecondary = ((uint32_t)left.getCE(--leftIndex)) >> 16;
                    }

                    int32_t rightSecondary = 0;
                    while(rightSecondary == 0 && rightIndex > rightStart) {
                        rightSecondary = ((uint32_t)right.getCE(--rightIndex)) >> 16;
                    }

                    if(leftSecondary != rightSecondary) {
                        return (leftSecondary < rightSecondary) ? UCOL_LESS : UCOL_GREATER;
                    }
                    if(leftSecondary == 0) { break; }
                }

                // Did we reach the end of either string?
                // Both strings have the same number of merge separators,
                // or else there would have been a primary-level difference.
                U_ASSERT(left.getCE(leftLimit) == right.getCE(rightLimit));
                if(p == Collation::NO_CE_PRIMARY) { break; }
                // Skip both merge separators and continue.
                leftStart = leftLimit + 1;
                rightStart = rightLimit + 1;
            }
        }
    }

    if((options & CollationSettings::CASE_LEVEL) != 0) {
        int32_t strength = CollationSettings::getStrength(options);
        int32_t leftIndex = 0;
        int32_t rightIndex = 0;
        for(;;) {
            uint32_t leftCase, leftLower32, rightCase;
            if(strength == UCOL_PRIMARY) {
                // Primary+caseLevel: Ignore case level weights of primary ignorables.
                // Otherwise we would get a-umlaut > a
                // which is not desirable for accent-insensitive sorting.
                // Check for (lower 32 bits) == 0 as well because variable CEs are stored
                // with only primary weights.
                int64_t ce;
                do {
                    ce = left.getCE(leftIndex++);
                    leftCase = (uint32_t)ce;
                } while((uint32_t)(ce >> 32) == 0 || leftCase == 0);
                leftLower32 = leftCase;
                leftCase &= 0xc000;

                do {
                    ce = right.getCE(rightIndex++);
                    rightCase = (uint32_t)ce;
                } while((uint32_t)(ce >> 32) == 0 || rightCase == 0);
                rightCase &= 0xc000;
            } else {
                // Secondary+caseLevel: By analogy with the above,
                // ignore case level weights of secondary ignorables.
                //
                // Note: A tertiary CE has uppercase case bits (0.0.ut)
                // to keep tertiary+caseFirst well-formed.
                //
                // Tertiary+caseLevel: Also ignore case level weights of secondary ignorables.
                // Otherwise a tertiary CE's uppercase would be no greater than
                // a primary/secondary CE's uppercase.
                // (See UCA well-formedness condition 2.)
                // We could construct a special case weight higher than uppercase,
                // but it's simpler to always ignore case weights of secondary ignorables,
                // turning 0.0.ut into 0.0.0.t.
                // (See LDML Collation, Case Parameters.)
                do {
                    leftCase = (uint32_t)left.getCE(leftIndex++);
                } while(leftCase <= 0xffff);
                leftLower32 = leftCase;
                leftCase &= 0xc000;

                do {
                    rightCase = (uint32_t)right.getCE(rightIndex++);
                } while(rightCase <= 0xffff);
                rightCase &= 0xc000;
            }

            // No need to handle NO_CE and MERGE_SEPARATOR specially:
            // There is one case weight for each previous-level weight,
            // so level length differences were handled there.
            if(leftCase != rightCase) {
                if((options & CollationSettings::UPPER_FIRST) == 0) {
                    return (leftCase < rightCase) ? UCOL_LESS : UCOL_GREATER;
                } else {
                    return (leftCase < rightCase) ? UCOL_GREATER : UCOL_LESS;
                }
            }
            if((leftLower32 >> 16) == Collation::NO_CE_WEIGHT16) { break; }
        }
    }
    if(CollationSettings::getStrength(options) <= UCOL_SECONDARY) { return UCOL_EQUAL; }

    uint32_t tertiaryMask = CollationSettings::getTertiaryMask(options);

    int32_t leftIndex = 0;
    int32_t rightIndex = 0;
    uint32_t anyQuaternaries = 0;
    for(;;) {
        uint32_t leftLower32, leftTertiary;
        do {
            leftLower32 = (uint32_t)left.getCE(leftIndex++);
            anyQuaternaries |= leftLower32;
            U_ASSERT((leftLower32 & Collation::ONLY_TERTIARY_MASK) != 0 ||
                     (leftLower32 & 0xc0c0) == 0);
            leftTertiary = leftLower32 & tertiaryMask;
        } while(leftTertiary == 0);

        uint32_t rightLower32, rightTertiary;
        do {
            rightLower32 = (uint32_t)right.getCE(rightIndex++);
            anyQuaternaries |= rightLower32;
            U_ASSERT((rightLower32 & Collation::ONLY_TERTIARY_MASK) != 0 ||
                     (rightLower32 & 0xc0c0) == 0);
            rightTertiary = rightLower32 & tertiaryMask;
        } while(rightTertiary == 0);

        if(leftTertiary != rightTertiary) {
            if(CollationSettings::sortsTertiaryUpperCaseFirst(options)) {
                // Pass through NO_CE and keep real tertiary weights larger than that.
                // Do not change the artificial uppercase weight of a tertiary CE (0.0.ut),
                // to keep tertiary CEs well-formed.
                // Their case+tertiary weights must be greater than those of
                // primary and secondary CEs.
                if(leftTertiary > Collation::NO_CE_WEIGHT16) {
                    if(leftLower32 > 0xffff) {
                        leftTertiary ^= 0xc000;
                    } else {
                        leftTertiary += 0x4000;
                    }
                }
                if(rightTertiary > Collation::NO_CE_WEIGHT16) {
                    if(rightLower32 > 0xffff) {
                        rightTertiary ^= 0xc000;
                    } else {
                        rightTertiary += 0x4000;
                    }
                }
            }
            return (leftTertiary < rightTertiary) ? UCOL_LESS : UCOL_GREATER;
        }
        if(leftTertiary == Collation::NO_CE_WEIGHT16) { break; }
    }
    if(CollationSettings::getStrength(options) <= UCOL_TERTIARY) { return UCOL_EQUAL; }

    if(!anyVariable && (anyQuaternaries & 0xc0) == 0) {
        // If there are no "variable" CEs and no non-zero quaternary weights,
        // then there are no quaternary differences.
        return UCOL_EQUAL;
    }

    leftIndex = 0;
    rightIndex = 0;
    for(;;) {
        uint32_t leftQuaternary;
        do {
            int64_t ce = left.getCE(leftIndex++);
            leftQuaternary = (uint32_t)ce & 0xffff;
            if(leftQuaternary <= Collation::NO_CE_WEIGHT16) {
                // Variable primary or completely ignorable or NO_CE.
                leftQuaternary = (uint32_t)(ce >> 32);
            } else {
                // Regular CE, not tertiary ignorable.
                // Preserve the quaternary weight in bits 7..6.
                leftQuaternary |= 0xffffff3f;
            }
        } while(leftQuaternary == 0);

        uint32_t rightQuaternary;
        do {
            int64_t ce = right.getCE(rightIndex++);
            rightQuaternary = (uint32_t)ce & 0xffff;
            if(rightQuaternary <= Collation::NO_CE_WEIGHT16) {
                // Variable primary or completely ignorable or NO_CE.
                rightQuaternary = (uint32_t)(ce >> 32);
            } else {
                // Regular CE, not tertiary ignorable.
                // Preserve the quaternary weight in bits 7..6.
                rightQuaternary |= 0xffffff3f;
            }
        } while(rightQuaternary == 0);

        if(leftQuaternary != rightQuaternary) {
            // Return the difference, with script reordering.
            if(settings.hasReordering()) {
                leftQuaternary = settings.reorder(leftQuaternary);
                rightQuaternary = settings.reorder(rightQuaternary);
            }
            return (leftQuaternary < rightQuaternary) ? UCOL_LESS : UCOL_GREATER;
        }
        if(leftQuaternary == Collation::NO_CE_PRIMARY) { break; }
    }
    return UCOL_EQUAL;
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationdata.cpp
*
* created on: 2012jul28
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION











U_NAMESPACE_BEGIN

uint32_t
CollationData::getIndirectCE32(uint32_t ce32) const {
    U_ASSERT(Collation::isSpecialCE32(ce32));
    int32_t tag = Collation::tagFromCE32(ce32);
    if(tag == Collation::DIGIT_TAG) {
        // Fetch the non-numeric-collation CE32.
        ce32 = ce32s[Collation::indexFromCE32(ce32)];
    } else if(tag == Collation::LEAD_SURROGATE_TAG) {
        ce32 = Collation::UNASSIGNED_CE32;
    } else if(tag == Collation::U0000_TAG) {
        // Fetch the normal ce32 for U+0000.
        ce32 = ce32s[0];
    }
    return ce32;
}

uint32_t
CollationData::getFinalCE32(uint32_t ce32) const {
    if(Collation::isSpecialCE32(ce32)) {
        ce32 = getIndirectCE32(ce32);
    }
    return ce32;
}

int64_t
CollationData::getSingleCE(UChar32 c, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return 0; }
    // Keep parallel with CollationDataBuilder::getSingleCE().
    const CollationData *d;
    uint32_t ce32 = getCE32(c);
    if(ce32 == Collation::FALLBACK_CE32) {
        d = base;
        ce32 = base->getCE32(c);
    } else {
        d = this;
    }
    while(Collation::isSpecialCE32(ce32)) {
        switch(Collation::tagFromCE32(ce32)) {
        case Collation::LATIN_EXPANSION_TAG:
        case Collation::BUILDER_DATA_TAG:
        case Collation::PREFIX_TAG:
        case Collation::CONTRACTION_TAG:
        case Collation::HANGUL_TAG:
        case Collation::LEAD_SURROGATE_TAG:
            errorCode = U_UNSUPPORTED_ERROR;
            return 0;
        case Collation::FALLBACK_TAG:
        case Collation::RESERVED_TAG_3:
            errorCode = U_INTERNAL_PROGRAM_ERROR;
            return 0;
        case Collation::LONG_PRIMARY_TAG:
            return Collation::ceFromLongPrimaryCE32(ce32);
        case Collation::LONG_SECONDARY_TAG:
            return Collation::ceFromLongSecondaryCE32(ce32);
        case Collation::EXPANSION32_TAG:
            if(Collation::lengthFromCE32(ce32) == 1) {
                ce32 = d->ce32s[Collation::indexFromCE32(ce32)];
                break;
            } else {
                errorCode = U_UNSUPPORTED_ERROR;
                return 0;
            }
        case Collation::EXPANSION_TAG: {
            if(Collation::lengthFromCE32(ce32) == 1) {
                return d->ces[Collation::indexFromCE32(ce32)];
            } else {
                errorCode = U_UNSUPPORTED_ERROR;
                return 0;
            }
        }
        case Collation::DIGIT_TAG:
            // Fetch the non-numeric-collation CE32 and continue.
            ce32 = d->ce32s[Collation::indexFromCE32(ce32)];
            break;
        case Collation::U0000_TAG:
            U_ASSERT(c == 0);
            // Fetch the normal ce32 for U+0000 and continue.
            ce32 = d->ce32s[0];
            break;
        case Collation::OFFSET_TAG:
            return d->getCEFromOffsetCE32(c, ce32);
        case Collation::IMPLICIT_TAG:
            return Collation::unassignedCEFromCodePoint(c);
        }
    }
    return Collation::ceFromSimpleCE32(ce32);
}

uint32_t
CollationData::getFirstPrimaryForGroup(int32_t script) const {
    int32_t index = getScriptIndex(script);
    return index == 0 ? 0 : (uint32_t)scriptStarts[index] << 16;
}

uint32_t
CollationData::getLastPrimaryForGroup(int32_t script) const {
    int32_t index = getScriptIndex(script);
    if(index == 0) {
        return 0;
    }
    uint32_t limit = scriptStarts[index + 1];
    return (limit << 16) - 1;
}

int32_t
CollationData::getGroupForPrimary(uint32_t p) const {
    p >>= 16;
    if(p < scriptStarts[1] || scriptStarts[scriptStartsLength - 1] <= p) {
        return -1;
    }
    int32_t index = 1;
    while(p >= scriptStarts[index + 1]) { ++index; }
    for(int32_t i = 0; i < numScripts; ++i) {
        if(scriptsIndex[i] == index) {
            return i;
        }
    }
    for(int32_t i = 0; i < MAX_NUM_SPECIAL_REORDER_CODES; ++i) {
        if(scriptsIndex[numScripts + i] == index) {
            return UCOL_REORDER_CODE_FIRST + i;
        }
    }
    return -1;
}

int32_t
CollationData::getScriptIndex(int32_t script) const {
    if(script < 0) {
        return 0;
    } else if(script < numScripts) {
        return scriptsIndex[script];
    } else if(script < UCOL_REORDER_CODE_FIRST) {
        return 0;
    } else {
        script -= UCOL_REORDER_CODE_FIRST;
        if(script < MAX_NUM_SPECIAL_REORDER_CODES) {
            return scriptsIndex[numScripts + script];
        } else {
            return 0;
        }
    }
}

int32_t
CollationData::getEquivalentScripts(int32_t script,
                                    int32_t dest[], int32_t capacity,
                                    UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return 0; }
    int32_t index = getScriptIndex(script);
    if(index == 0) { return 0; }
    if(script >= UCOL_REORDER_CODE_FIRST) {
        // Special groups have no aliases.
        if(capacity > 0) {
            dest[0] = script;
        } else {
            errorCode = U_BUFFER_OVERFLOW_ERROR;
        }
        return 1;
    }

    int32_t length = 0;
    for(int32_t i = 0; i < numScripts; ++i) {
        if(scriptsIndex[i] == index) {
            if(length < capacity) {
                dest[length] = i;
            }
            ++length;
        }
    }
    if(length > capacity) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
    }
    return length;
}

void
CollationData::makeReorderRanges(const int32_t *reorder, int32_t length,
                                 UVector32 &ranges, UErrorCode &errorCode) const {
    makeReorderRanges(reorder, length, FALSE, ranges, errorCode);
}

void
CollationData::makeReorderRanges(const int32_t *reorder, int32_t length,
                                 UBool latinMustMove,
                                 UVector32 &ranges, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return; }
    ranges.removeAllElements();
    if(length == 0 || (length == 1 && reorder[0] == USCRIPT_UNKNOWN)) {
        return;
    }

    // Maps each script-or-group range to a new lead byte.
    uint8_t table[MAX_NUM_SCRIPT_RANGES];
    uprv_memset(table, 0, sizeof(table));

    {
        // Set "don't care" values for reserved ranges.
        int32_t index = scriptsIndex[
                numScripts + REORDER_RESERVED_BEFORE_LATIN - UCOL_REORDER_CODE_FIRST];
        if(index != 0) {
            table[index] = 0xff;
        }
        index = scriptsIndex[
                numScripts + REORDER_RESERVED_AFTER_LATIN - UCOL_REORDER_CODE_FIRST];
        if(index != 0) {
            table[index] = 0xff;
        }
    }

    // Never reorder special low and high primary lead bytes.
    U_ASSERT(scriptStartsLength >= 2);
    U_ASSERT(scriptStarts[0] == 0);
    int32_t lowStart = scriptStarts[1];
    U_ASSERT(lowStart == ((Collation::MERGE_SEPARATOR_BYTE + 1) << 8));
    int32_t highLimit = scriptStarts[scriptStartsLength - 1];
    U_ASSERT(highLimit == (Collation::TRAIL_WEIGHT_BYTE << 8));

    // Get the set of special reorder codes in the input list.
    // This supports a fixed number of special reorder codes;
    // it works for data with codes beyond UCOL_REORDER_CODE_LIMIT.
    uint32_t specials = 0;
    for(int32_t i = 0; i < length; ++i) {
        int32_t reorderCode = reorder[i] - UCOL_REORDER_CODE_FIRST;
        if(0 <= reorderCode && reorderCode < MAX_NUM_SPECIAL_REORDER_CODES) {
            specials |= (uint32_t)1 << reorderCode;
        }
    }

    // Start the reordering with the special low reorder codes that do not occur in the input.
    for(int32_t i = 0; i < MAX_NUM_SPECIAL_REORDER_CODES; ++i) {
        int32_t index = scriptsIndex[numScripts + i];
        if(index != 0 && (specials & ((uint32_t)1 << i)) == 0) {
            lowStart = addLowScriptRange(table, index, lowStart);
        }
    }

    // Skip the reserved range before Latin if Latin is the first script,
    // so that we do not move it unnecessarily.
    int32_t skippedReserved = 0;
    if(specials == 0 && reorder[0] == USCRIPT_LATIN && !latinMustMove) {
        int32_t index = scriptsIndex[USCRIPT_LATIN];
        U_ASSERT(index != 0);
        int32_t start = scriptStarts[index];
        U_ASSERT(lowStart <= start);
        skippedReserved = start - lowStart;
        lowStart = start;
    }

    // Reorder according to the input scripts, continuing from the bottom of the primary range.
    int32_t originalLength = length;  // length will be decremented if "others" is in the list.
    UBool hasReorderToEnd = FALSE;
    for(int32_t i = 0; i < length;) {
        int32_t script = reorder[i++];
        if(script == USCRIPT_UNKNOWN) {
            // Put the remaining scripts at the top.
            hasReorderToEnd = TRUE;
            while(i < length) {
                script = reorder[--length];
                if(script == USCRIPT_UNKNOWN ||  // Must occur at most once.
                        script == UCOL_REORDER_CODE_DEFAULT) {
                    errorCode = U_ILLEGAL_ARGUMENT_ERROR;
                    return;
                }
                int32_t index = getScriptIndex(script);
                if(index == 0) { continue; }
                if(table[index] != 0) {  // Duplicate or equivalent script.
                    errorCode = U_ILLEGAL_ARGUMENT_ERROR;
                    return;
                }
                highLimit = addHighScriptRange(table, index, highLimit);
            }
            break;
        }
        if(script == UCOL_REORDER_CODE_DEFAULT) {
            // The default code must be the only one in the list, and that is handled by the caller.
            // Otherwise it must not be used.
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        int32_t index = getScriptIndex(script);
        if(index == 0) { continue; }
        if(table[index] != 0) {  // Duplicate or equivalent script.
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        lowStart = addLowScriptRange(table, index, lowStart);
    }

    // Put all remaining scripts into the middle.
    for(int32_t i = 1; i < scriptStartsLength - 1; ++i) {
        int32_t leadByte = table[i];
        if(leadByte != 0) { continue; }
        int32_t start = scriptStarts[i];
        if(!hasReorderToEnd && start > lowStart) {
            // No need to move this script.
            lowStart = start;
        }
        lowStart = addLowScriptRange(table, i, lowStart);
    }
    if(lowStart > highLimit) {
        if((lowStart - (skippedReserved & 0xff00)) <= highLimit) {
            // Try not skipping the before-Latin reserved range.
            makeReorderRanges(reorder, originalLength, TRUE, ranges, errorCode);
            return;
        }
        // We need more primary lead bytes than available, despite the reserved ranges.
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return;
    }

    // Turn lead bytes into a list of (limit, offset) pairs.
    // Encode each pair in one list element:
    // Upper 16 bits = limit, lower 16 = signed lead byte offset.
    int32_t offset = 0;
    for(int32_t i = 1;; ++i) {
        int32_t nextOffset = offset;
        while(i < scriptStartsLength - 1) {
            int32_t newLeadByte = table[i];
            if(newLeadByte == 0xff) {
                // "Don't care" lead byte for reserved range, continue with current offset.
            } else {
                nextOffset = newLeadByte - (scriptStarts[i] >> 8);
                if(nextOffset != offset) { break; }
            }
            ++i;
        }
        if(offset != 0 || i < scriptStartsLength - 1) {
            ranges.addElement(((int32_t)scriptStarts[i] << 16) | (offset & 0xffff), errorCode);
        }
        if(i == scriptStartsLength - 1) { break; }
        offset = nextOffset;
    }
}

int32_t
CollationData::addLowScriptRange(uint8_t table[], int32_t index, int32_t lowStart) const {
    int32_t start = scriptStarts[index];
    if((start & 0xff) < (lowStart & 0xff)) {
        lowStart += 0x100;
    }
    table[index] = (uint8_t)(lowStart >> 8);
    int32_t limit = scriptStarts[index + 1];
    lowStart = ((lowStart & 0xff00) + ((limit & 0xff00) - (start & 0xff00))) | (limit & 0xff);
    return lowStart;
}

int32_t
CollationData::addHighScriptRange(uint8_t table[], int32_t index, int32_t highLimit) const {
    int32_t limit = scriptStarts[index + 1];
    if((limit & 0xff) > (highLimit & 0xff)) {
        highLimit -= 0x100;
    }
    int32_t start = scriptStarts[index];
    highLimit = ((highLimit & 0xff00) - ((limit & 0xff00) - (start & 0xff00))) | (start & 0xff);
    table[index] = (uint8_t)(highLimit >> 8);
    return highLimit;
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationdatabuilder.cpp
*
* (replaced the former ucol_elm.cpp)
*
* created on: 2012apr01
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION













// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationfastlatinbuilder.h
*
* created on: 2013aug09
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION








U_NAMESPACE_BEGIN

struct CollationData;

class U_I18N_API CollationFastLatinBuilder : public UObject {
public:
    CollationFastLatinBuilder(UErrorCode &errorCode);
    ~CollationFastLatinBuilder();

    UBool forData(const CollationData &data, UErrorCode &errorCode);

    const uint16_t *getTable() const {
        return reinterpret_cast<const uint16_t *>(result.getBuffer());
    }
    int32_t lengthOfTable() const { return result.length(); }

private:
    // space, punct, symbol, currency (not digit)
    enum { NUM_SPECIAL_GROUPS = UCOL_REORDER_CODE_CURRENCY - UCOL_REORDER_CODE_FIRST + 1 };

    UBool loadGroups(const CollationData &data, UErrorCode &errorCode);
    UBool inSameGroup(uint32_t p, uint32_t q) const;

    void resetCEs();
    void getCEs(const CollationData &data, UErrorCode &errorCode);
    UBool getCEsFromCE32(const CollationData &data, UChar32 c, uint32_t ce32,
                         UErrorCode &errorCode);
    UBool getCEsFromContractionCE32(const CollationData &data, uint32_t ce32,
                                    UErrorCode &errorCode);
    void addContractionEntry(int32_t x, int64_t cce0, int64_t cce1, UErrorCode &errorCode);
    void addUniqueCE(int64_t ce, UErrorCode &errorCode);
    uint32_t getMiniCE(int64_t ce) const;
    UBool encodeUniqueCEs(UErrorCode &errorCode);
    UBool encodeCharCEs(UErrorCode &errorCode);
    UBool encodeContractions(UErrorCode &errorCode);
    uint32_t encodeTwoCEs(int64_t first, int64_t second) const;

    static UBool isContractionCharCE(int64_t ce) {
        return (uint32_t)(ce >> 32) == Collation::NO_CE_PRIMARY && ce != Collation::NO_CE;
    }

    static const uint32_t CONTRACTION_FLAG = 0x80000000;

    // temporary "buffer"
    int64_t ce0, ce1;

    int64_t charCEs[CollationFastLatin::NUM_FAST_CHARS][2];

    UVector64 contractionCEs;
    UVector64 uniqueCEs;

    /** One 16-bit mini CE per unique CE. */
    uint16_t *miniCEs;

    // These are constant for a given root collator.
    uint32_t lastSpecialPrimaries[NUM_SPECIAL_GROUPS];
    uint32_t firstDigitPrimary;
    uint32_t firstLatinPrimary;
    uint32_t lastLatinPrimary;
    // This determines the first normal primary weight which is mapped to
    // a short mini primary. It must be >=firstDigitPrimary.
    uint32_t firstShortPrimary;

    UBool shortPrimaryOverflow;

    UnicodeString result;
    int32_t headerLength;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION








U_NAMESPACE_BEGIN

CollationDataBuilder::CEModifier::~CEModifier() {}

/**
 * Build-time context and CE32 for a code point.
 * If a code point has contextual mappings, then the default (no-context) mapping
 * and all conditional mappings are stored in a singly-linked list
 * of ConditionalCE32, sorted by context strings.
 *
 * Context strings sort by prefix length, then by prefix, then by contraction suffix.
 * Context strings must be unique and in ascending order.
 */
struct ConditionalCE32 : public UMemory {
    ConditionalCE32()
            : context(),
              ce32(0), defaultCE32(Collation::NO_CE32), builtCE32(Collation::NO_CE32),
              next(-1) {}
    ConditionalCE32(const UnicodeString &ct, uint32_t ce)
            : context(ct),
              ce32(ce), defaultCE32(Collation::NO_CE32), builtCE32(Collation::NO_CE32),
              next(-1) {}

    inline UBool hasContext() const { return context.length() > 1; }
    inline int32_t prefixLength() const { return context.charAt(0); }

    /**
     * "\0" for the first entry for any code point, with its default CE32.
     *
     * Otherwise one unit with the length of the prefix string,
     * then the prefix string, then the contraction suffix.
     */
    UnicodeString context;
    /**
     * CE32 for the code point and its context.
     * Can be special (e.g., for an expansion) but not contextual (prefix or contraction tag).
     */
    uint32_t ce32;
    /**
     * Default CE32 for all contexts with this same prefix.
     * Initially NO_CE32. Set only while building runtime data structures,
     * and only on one of the nodes of a sub-list with the same prefix.
     */
    uint32_t defaultCE32;
    /**
     * CE32 for the built contexts.
     * When fetching CEs from the builder, the contexts are built into their runtime form
     * so that the normal collation implementation can process them.
     * The result is cached in the list head. It is reset when the contexts are modified.
     */
    uint32_t builtCE32;
    /**
     * Index of the next ConditionalCE32.
     * Negative for the end of the list.
     */
    int32_t next;
};

U_CDECL_BEGIN

U_CAPI void U_CALLCONV
uprv_deleteConditionalCE32(void *obj) {
    delete static_cast<ConditionalCE32 *>(obj);
}

U_CDECL_END

/**
 * Build-time collation element and character iterator.
 * Uses the runtime CollationIterator for fetching CEs for a string
 * but reads from the builder's unfinished data structures.
 * In particular, this class reads from the unfinished trie
 * and has to avoid CollationIterator::nextCE() and redirect other
 * calls to data->getCE32() and data->getCE32FromSupplementary().
 *
 * We do this so that we need not implement the collation algorithm
 * again for the builder and make it behave exactly like the runtime code.
 * That would be more difficult to test and maintain than this indirection.
 *
 * Some CE32 tags (for example, the DIGIT_TAG) do not occur in the builder data,
 * so the data accesses from those code paths need not be modified.
 *
 * This class iterates directly over whole code points
 * so that the CollationIterator does not need the finished trie
 * for handling the LEAD_SURROGATE_TAG.
 */
class DataBuilderCollationIterator : public CollationIterator {
public:
    DataBuilderCollationIterator(CollationDataBuilder &b);

    virtual ~DataBuilderCollationIterator();

    int32_t fetchCEs(const UnicodeString &str, int32_t start, int64_t ces[], int32_t cesLength);

    virtual void resetToOffset(int32_t newOffset);
    virtual int32_t getOffset() const;

    virtual UChar32 nextCodePoint(UErrorCode &errorCode);
    virtual UChar32 previousCodePoint(UErrorCode &errorCode);

protected:
    virtual void forwardNumCodePoints(int32_t num, UErrorCode &errorCode);
    virtual void backwardNumCodePoints(int32_t num, UErrorCode &errorCode);

    virtual uint32_t getDataCE32(UChar32 c) const;
    virtual uint32_t getCE32FromBuilderData(uint32_t ce32, UErrorCode &errorCode);

    CollationDataBuilder &builder;
    CollationData builderData;
    uint32_t jamoCE32s[CollationData::JAMO_CE32S_LENGTH];
    const UnicodeString *s;
    int32_t pos;
};

DataBuilderCollationIterator::DataBuilderCollationIterator(CollationDataBuilder &b)
        : CollationIterator(&builderData, /*numeric=*/ FALSE),
          builder(b), builderData(b.nfcImpl),
          s(NULL), pos(0) {
    builderData.base = builder.base;
    // Set all of the jamoCE32s[] to indirection CE32s.
    for(int32_t j = 0; j < CollationData::JAMO_CE32S_LENGTH; ++j) {  // Count across Jamo types.
        UChar32 jamo = CollationDataBuilder::jamoCpFromIndex(j);
        jamoCE32s[j] = Collation::makeCE32FromTagAndIndex(Collation::BUILDER_DATA_TAG, jamo) |
                CollationDataBuilder::IS_BUILDER_JAMO_CE32;
    }
    builderData.jamoCE32s = jamoCE32s;
}

DataBuilderCollationIterator::~DataBuilderCollationIterator() {}

int32_t
DataBuilderCollationIterator::fetchCEs(const UnicodeString &str, int32_t start,
                                       int64_t ces[], int32_t cesLength) {
    // Set the pointers each time, in case they changed due to reallocation.
    builderData.ce32s = reinterpret_cast<const uint32_t *>(builder.ce32s.getBuffer());
    builderData.ces = builder.ce64s.getBuffer();
    builderData.contexts = builder.contexts.getBuffer();
    // Modified copy of CollationIterator::nextCE() and CollationIterator::nextCEFromCE32().
    reset();
    s = &str;
    pos = start;
    UErrorCode errorCode = U_ZERO_ERROR;
    while(U_SUCCESS(errorCode) && pos < s->length()) {
        // No need to keep all CEs in the iterator buffer.
        clearCEs();
        UChar32 c = s->char32At(pos);
        pos += U16_LENGTH(c);
        uint32_t ce32 = utrie2_get32(builder.trie, c);
        const CollationData *d;
        if(ce32 == Collation::FALLBACK_CE32) {
            d = builder.base;
            ce32 = builder.base->getCE32(c);
        } else {
            d = &builderData;
        }
        appendCEsFromCE32(d, c, ce32, /*forward=*/ TRUE, errorCode);
        U_ASSERT(U_SUCCESS(errorCode));
        for(int32_t i = 0; i < getCEsLength(); ++i) {
            int64_t ce = getCE(i);
            if(ce != 0) {
                if(cesLength < Collation::MAX_EXPANSION_LENGTH) {
                    ces[cesLength] = ce;
                }
                ++cesLength;
            }
        }
    }
    return cesLength;
}

void
DataBuilderCollationIterator::resetToOffset(int32_t newOffset) {
    reset();
    pos = newOffset;
}

int32_t
DataBuilderCollationIterator::getOffset() const {
    return pos;
}

UChar32
DataBuilderCollationIterator::nextCodePoint(UErrorCode & /*errorCode*/) {
    if(pos == s->length()) {
        return U_SENTINEL;
    }
    UChar32 c = s->char32At(pos);
    pos += U16_LENGTH(c);
    return c;
}

UChar32
DataBuilderCollationIterator::previousCodePoint(UErrorCode & /*errorCode*/) {
    if(pos == 0) {
        return U_SENTINEL;
    }
    UChar32 c = s->char32At(pos - 1);
    pos -= U16_LENGTH(c);
    return c;
}

void
DataBuilderCollationIterator::forwardNumCodePoints(int32_t num, UErrorCode & /*errorCode*/) {
    pos = s->moveIndex32(pos, num);
}

void
DataBuilderCollationIterator::backwardNumCodePoints(int32_t num, UErrorCode & /*errorCode*/) {
    pos = s->moveIndex32(pos, -num);
}

uint32_t
DataBuilderCollationIterator::getDataCE32(UChar32 c) const {
    return utrie2_get32(builder.trie, c);
}

uint32_t
DataBuilderCollationIterator::getCE32FromBuilderData(uint32_t ce32, UErrorCode &errorCode) {
    U_ASSERT(Collation::hasCE32Tag(ce32, Collation::BUILDER_DATA_TAG));
    if((ce32 & CollationDataBuilder::IS_BUILDER_JAMO_CE32) != 0) {
        UChar32 jamo = Collation::indexFromCE32(ce32);
        return utrie2_get32(builder.trie, jamo);
    } else {
        ConditionalCE32 *cond = builder.getConditionalCE32ForCE32(ce32);
        if(cond->builtCE32 == Collation::NO_CE32) {
            // Build the context-sensitive mappings into their runtime form and cache the result.
            cond->builtCE32 = builder.buildContext(cond, errorCode);
            if(errorCode == U_BUFFER_OVERFLOW_ERROR) {
                errorCode = U_ZERO_ERROR;
                builder.clearContexts();
                cond->builtCE32 = builder.buildContext(cond, errorCode);
            }
            builderData.contexts = builder.contexts.getBuffer();
        }
        return cond->builtCE32;
    }
}

// ------------------------------------------------------------------------- ***

CollationDataBuilder::CollationDataBuilder(UErrorCode &errorCode)
        : nfcImpl(*Normalizer2Factory::getNFCImpl(errorCode)),
          base(NULL), baseSettings(NULL),
          trie(NULL),
          ce32s(errorCode), ce64s(errorCode), conditionalCE32s(errorCode),
          modified(FALSE),
          fastLatinEnabled(FALSE), fastLatinBuilder(NULL),
          collIter(NULL) {
    // Reserve the first CE32 for U+0000.
    ce32s.addElement(0, errorCode);
    conditionalCE32s.setDeleter(uprv_deleteConditionalCE32);
}

CollationDataBuilder::~CollationDataBuilder() {
    utrie2_close(trie);
    delete fastLatinBuilder;
    delete collIter;
}

void
CollationDataBuilder::initForTailoring(const CollationData *b, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(trie != NULL) {
        errorCode = U_INVALID_STATE_ERROR;
        return;
    }
    if(b == NULL) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    base = b;

    // For a tailoring, the default is to fall back to the base.
    trie = utrie2_open(Collation::FALLBACK_CE32, Collation::FFFD_CE32, &errorCode);

    // Set the Latin-1 letters block so that it is allocated first in the data array,
    // to try to improve locality of reference when sorting Latin-1 text.
    // Do not use utrie2_setRange32() since that will not actually allocate blocks
    // that are filled with the default value.
    // ASCII (0..7F) is already preallocated anyway.
    for(UChar32 c = 0xc0; c <= 0xff; ++c) {
        utrie2_set32(trie, c, Collation::FALLBACK_CE32, &errorCode);
    }

    // Hangul syllables are not tailorable (except via tailoring Jamos).
    // Always set the Hangul tag to help performance.
    // Do this here, rather than in buildMappings(),
    // so that we see the HANGUL_TAG in various assertions.
    uint32_t hangulCE32 = Collation::makeCE32FromTagAndIndex(Collation::HANGUL_TAG, 0);
    utrie2_setRange32(trie, Hangul::HANGUL_BASE, Hangul::HANGUL_END, hangulCE32, TRUE, &errorCode);

    // Copy the set contents but don't copy/clone the set as a whole because
    // that would copy the isFrozen state too.
    unsafeBackwardSet.addAll(*b->unsafeBackwardSet);

    if(U_FAILURE(errorCode)) { return; }
}

UBool
CollationDataBuilder::maybeSetPrimaryRange(UChar32 start, UChar32 end,
                                           uint32_t primary, int32_t step,
                                           UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    U_ASSERT(start <= end);
    // TODO: Do we need to check what values are currently set for start..end?
    // An offset range is worth it only if we can achieve an overlap between
    // adjacent UTrie2 blocks of 32 code points each.
    // An offset CE is also a little more expensive to look up and compute
    // than a simple CE.
    // If the range spans at least three UTrie2 block boundaries (> 64 code points),
    // then we take it.
    // If the range spans one or two block boundaries and there are
    // at least 4 code points on either side, then we take it.
    // (We could additionally require a minimum range length of, say, 16.)
    int32_t blockDelta = (end >> 5) - (start >> 5);
    if(2 <= step && step <= 0x7f &&
            (blockDelta >= 3 ||
            (blockDelta > 0 && (start & 0x1f) <= 0x1c && (end & 0x1f) >= 3))) {
        int64_t dataCE = ((int64_t)primary << 32) | (start << 8) | step;
        if(isCompressiblePrimary(primary)) { dataCE |= 0x80; }
        int32_t index = addCE(dataCE, errorCode);
        if(U_FAILURE(errorCode)) { return 0; }
        if(index > Collation::MAX_INDEX) {
            errorCode = U_BUFFER_OVERFLOW_ERROR;
            return 0;
        }
        uint32_t offsetCE32 = Collation::makeCE32FromTagAndIndex(Collation::OFFSET_TAG, index);
        utrie2_setRange32(trie, start, end, offsetCE32, TRUE, &errorCode);
        modified = TRUE;
        return TRUE;
    } else {
        return FALSE;
    }
}

uint32_t
CollationDataBuilder::setPrimaryRangeAndReturnNext(UChar32 start, UChar32 end,
                                                   uint32_t primary, int32_t step,
                                                   UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    UBool isCompressible = isCompressiblePrimary(primary);
    if(maybeSetPrimaryRange(start, end, primary, step, errorCode)) {
        return Collation::incThreeBytePrimaryByOffset(primary, isCompressible,
                                                      (end - start + 1) * step);
    } else {
        // Short range: Set individual CE32s.
        for(;;) {
            utrie2_set32(trie, start, Collation::makeLongPrimaryCE32(primary), &errorCode);
            ++start;
            primary = Collation::incThreeBytePrimaryByOffset(primary, isCompressible, step);
            if(start > end) { return primary; }
        }
        modified = TRUE;
    }
}

uint32_t
CollationDataBuilder::getCE32FromOffsetCE32(UBool fromBase, UChar32 c, uint32_t ce32) const {
    int32_t i = Collation::indexFromCE32(ce32);
    int64_t dataCE = fromBase ? base->ces[i] : ce64s.elementAti(i);
    uint32_t p = Collation::getThreeBytePrimaryForOffsetData(c, dataCE);
    return Collation::makeLongPrimaryCE32(p);
}

UBool
CollationDataBuilder::isCompressibleLeadByte(uint32_t b) const {
    return base->isCompressibleLeadByte(b);
}

UBool
CollationDataBuilder::isAssigned(UChar32 c) const {
    return Collation::isAssignedCE32(utrie2_get32(trie, c));
}

uint32_t
CollationDataBuilder::getLongPrimaryIfSingleCE(UChar32 c) const {
    uint32_t ce32 = utrie2_get32(trie, c);
    if(Collation::isLongPrimaryCE32(ce32)) {
        return Collation::primaryFromLongPrimaryCE32(ce32);
    } else {
        return 0;
    }
}

int64_t
CollationDataBuilder::getSingleCE(UChar32 c, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return 0; }
    // Keep parallel with CollationData::getSingleCE().
    UBool fromBase = FALSE;
    uint32_t ce32 = utrie2_get32(trie, c);
    if(ce32 == Collation::FALLBACK_CE32) {
        fromBase = TRUE;
        ce32 = base->getCE32(c);
    }
    while(Collation::isSpecialCE32(ce32)) {
        switch(Collation::tagFromCE32(ce32)) {
        case Collation::LATIN_EXPANSION_TAG:
        case Collation::BUILDER_DATA_TAG:
        case Collation::PREFIX_TAG:
        case Collation::CONTRACTION_TAG:
        case Collation::HANGUL_TAG:
        case Collation::LEAD_SURROGATE_TAG:
            errorCode = U_UNSUPPORTED_ERROR;
            return 0;
        case Collation::FALLBACK_TAG:
        case Collation::RESERVED_TAG_3:
            errorCode = U_INTERNAL_PROGRAM_ERROR;
            return 0;
        case Collation::LONG_PRIMARY_TAG:
            return Collation::ceFromLongPrimaryCE32(ce32);
        case Collation::LONG_SECONDARY_TAG:
            return Collation::ceFromLongSecondaryCE32(ce32);
        case Collation::EXPANSION32_TAG:
            if(Collation::lengthFromCE32(ce32) == 1) {
                int32_t i = Collation::indexFromCE32(ce32);
                ce32 = fromBase ? base->ce32s[i] : ce32s.elementAti(i);
                break;
            } else {
                errorCode = U_UNSUPPORTED_ERROR;
                return 0;
            }
        case Collation::EXPANSION_TAG: {
            if(Collation::lengthFromCE32(ce32) == 1) {
                int32_t i = Collation::indexFromCE32(ce32);
                return fromBase ? base->ces[i] : ce64s.elementAti(i);
            } else {
                errorCode = U_UNSUPPORTED_ERROR;
                return 0;
            }
        }
        case Collation::DIGIT_TAG:
            // Fetch the non-numeric-collation CE32 and continue.
            ce32 = ce32s.elementAti(Collation::indexFromCE32(ce32));
            break;
        case Collation::U0000_TAG:
            U_ASSERT(c == 0);
            // Fetch the normal ce32 for U+0000 and continue.
            ce32 = fromBase ? base->ce32s[0] : ce32s.elementAti(0);
            break;
        case Collation::OFFSET_TAG:
            ce32 = getCE32FromOffsetCE32(fromBase, c, ce32);
            break;
        case Collation::IMPLICIT_TAG:
            return Collation::unassignedCEFromCodePoint(c);
        }
    }
    return Collation::ceFromSimpleCE32(ce32);
}

int32_t
CollationDataBuilder::addCE(int64_t ce, UErrorCode &errorCode) {
    int32_t length = ce64s.size();
    for(int32_t i = 0; i < length; ++i) {
        if(ce == ce64s.elementAti(i)) { return i; }
    }
    ce64s.addElement(ce, errorCode);
    return length;
}

int32_t
CollationDataBuilder::addCE32(uint32_t ce32, UErrorCode &errorCode) {
    int32_t length = ce32s.size();
    for(int32_t i = 0; i < length; ++i) {
        if(ce32 == (uint32_t)ce32s.elementAti(i)) { return i; }
    }
    ce32s.addElement((int32_t)ce32, errorCode);  
    return length;
}

int32_t
CollationDataBuilder::addConditionalCE32(const UnicodeString &context, uint32_t ce32,
                                         UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return -1; }
    U_ASSERT(!context.isEmpty());
    int32_t index = conditionalCE32s.size();
    if(index > Collation::MAX_INDEX) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return -1;
    }
    ConditionalCE32 *cond = new ConditionalCE32(context, ce32);
    if(cond == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return -1;
    }
    conditionalCE32s.addElement(cond, errorCode);
    return index;
}

void
CollationDataBuilder::add(const UnicodeString &prefix, const UnicodeString &s,
                          const int64_t ces[], int32_t cesLength,
                          UErrorCode &errorCode) {
    uint32_t ce32 = encodeCEs(ces, cesLength, errorCode);
    addCE32(prefix, s, ce32, errorCode);
}

void
CollationDataBuilder::addCE32(const UnicodeString &prefix, const UnicodeString &s,
                              uint32_t ce32, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(s.isEmpty()) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if(trie == NULL || utrie2_isFrozen(trie)) {
        errorCode = U_INVALID_STATE_ERROR;
        return;
    }
    UChar32 c = s.char32At(0);
    int32_t cLength = U16_LENGTH(c);
    uint32_t oldCE32 = utrie2_get32(trie, c);
    UBool hasContext = !prefix.isEmpty() || s.length() > cLength;
    if(oldCE32 == Collation::FALLBACK_CE32) {
        // First tailoring for c.
        // If c has contextual base mappings or if we add a contextual mapping,
        // then copy the base mappings.
        // Otherwise we just override the base mapping.
        uint32_t baseCE32 = base->getFinalCE32(base->getCE32(c));
        if(hasContext || Collation::ce32HasContext(baseCE32)) {
            oldCE32 = copyFromBaseCE32(c, baseCE32, TRUE, errorCode);
            utrie2_set32(trie, c, oldCE32, &errorCode);
            if(U_FAILURE(errorCode)) { return; }
        }
    }
    if(!hasContext) {
        // No prefix, no contraction.
        if(!isBuilderContextCE32(oldCE32)) {
            utrie2_set32(trie, c, ce32, &errorCode);
        } else {
            ConditionalCE32 *cond = getConditionalCE32ForCE32(oldCE32);
            cond->builtCE32 = Collation::NO_CE32;
            cond->ce32 = ce32;
        }
    } else {
        ConditionalCE32 *cond;
        if(!isBuilderContextCE32(oldCE32)) {
            // Replace the simple oldCE32 with a builder context CE32
            // pointing to a new ConditionalCE32 list head.
            int32_t index = addConditionalCE32(UnicodeString((UChar)0), oldCE32, errorCode);
            if(U_FAILURE(errorCode)) { return; }
            uint32_t contextCE32 = makeBuilderContextCE32(index);
            utrie2_set32(trie, c, contextCE32, &errorCode);
            contextChars.add(c);
            cond = getConditionalCE32(index);
        } else {
            cond = getConditionalCE32ForCE32(oldCE32);
            cond->builtCE32 = Collation::NO_CE32;
        }
        UnicodeString suffix(s, cLength);
        UnicodeString context((UChar)prefix.length());
        context.append(prefix).append(suffix);
        unsafeBackwardSet.addAll(suffix);
        for(;;) {
            // invariant: context > cond->context
            int32_t next = cond->next;
            if(next < 0) {
                // Append a new ConditionalCE32 after cond.
                int32_t index = addConditionalCE32(context, ce32, errorCode);
                if(U_FAILURE(errorCode)) { return; }
                cond->next = index;
                break;
            }
            ConditionalCE32 *nextCond = getConditionalCE32(next);
            int8_t cmp = context.compare(nextCond->context);
            if(cmp < 0) {
                // Insert a new ConditionalCE32 between cond and nextCond.
                int32_t index = addConditionalCE32(context, ce32, errorCode);
                if(U_FAILURE(errorCode)) { return; }
                cond->next = index;
                getConditionalCE32(index)->next = next;
                break;
            } else if(cmp == 0) {
                // Same context as before, overwrite its ce32.
                nextCond->ce32 = ce32;
                break;
            }
            cond = nextCond;
        }
    }
    modified = TRUE;
}

uint32_t
CollationDataBuilder::encodeOneCEAsCE32(int64_t ce) {
    uint32_t p = (uint32_t)(ce >> 32);
    uint32_t lower32 = (uint32_t)ce;
    uint32_t t = (uint32_t)(ce & 0xffff);
    U_ASSERT((t & 0xc000) != 0xc000);  // Impossible case bits 11 mark special CE32s.
    if((ce & INT64_C(0xffff00ff00ff)) == 0) {
        // normal form ppppsstt
        return p | (lower32 >> 16) | (t >> 8);
    } else if((ce & INT64_C(0xffffffffff)) == Collation::COMMON_SEC_AND_TER_CE) {
        // long-primary form ppppppC1
        return Collation::makeLongPrimaryCE32(p);
    } else if(p == 0 && (t & 0xff) == 0) {
        // long-secondary form ssssttC2
        return Collation::makeLongSecondaryCE32(lower32);
    }
    return Collation::NO_CE32;
}

uint32_t
CollationDataBuilder::encodeOneCE(int64_t ce, UErrorCode &errorCode) {
    // Try to encode one CE as one CE32.
    uint32_t ce32 = encodeOneCEAsCE32(ce);
    if(ce32 != Collation::NO_CE32) { return ce32; }
    int32_t index = addCE(ce, errorCode);
    if(U_FAILURE(errorCode)) { return 0; }
    if(index > Collation::MAX_INDEX) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return 0;
    }
    return Collation::makeCE32FromTagIndexAndLength(Collation::EXPANSION_TAG, index, 1);
}

uint32_t
CollationDataBuilder::encodeCEs(const int64_t ces[], int32_t cesLength,
                                UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    if(cesLength < 0 || cesLength > Collation::MAX_EXPANSION_LENGTH) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    if(trie == NULL || utrie2_isFrozen(trie)) {
        errorCode = U_INVALID_STATE_ERROR;
        return 0;
    }
    if(cesLength == 0) {
        // Convenience: We cannot map to nothing, but we can map to a completely ignorable CE.
        // Do this here so that callers need not do it.
        return encodeOneCEAsCE32(0);
    } else if(cesLength == 1) {
        return encodeOneCE(ces[0], errorCode);
    } else if(cesLength == 2) {
        // Try to encode two CEs as one CE32.
        int64_t ce0 = ces[0];
        int64_t ce1 = ces[1];
        uint32_t p0 = (uint32_t)(ce0 >> 32);
        if((ce0 & INT64_C(0xffffffffff00ff)) == Collation::COMMON_SECONDARY_CE &&
                (ce1 & INT64_C(0xffffffff00ffffff)) == Collation::COMMON_TERTIARY_CE &&
                p0 != 0) {
            // Latin mini expansion
            return
                p0 |
                (((uint32_t)ce0 & 0xff00u) << 8) |
                (uint32_t)(ce1 >> 16) |
                Collation::SPECIAL_CE32_LOW_BYTE |
                Collation::LATIN_EXPANSION_TAG;
        }
    }
    // Try to encode two or more CEs as CE32s.
    int32_t newCE32s[Collation::MAX_EXPANSION_LENGTH];
    for(int32_t i = 0;; ++i) {
        if(i == cesLength) {
            return encodeExpansion32(newCE32s, cesLength, errorCode);
        }
        uint32_t ce32 = encodeOneCEAsCE32(ces[i]);
        if(ce32 == Collation::NO_CE32) { break; }
        newCE32s[i] = (int32_t)ce32;
    }
    return encodeExpansion(ces, cesLength, errorCode);
}

uint32_t
CollationDataBuilder::encodeExpansion(const int64_t ces[], int32_t length, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    // See if this sequence of CEs has already been stored.
    int64_t first = ces[0];
    int32_t ce64sMax = ce64s.size() - length;
    for(int32_t i = 0; i <= ce64sMax; ++i) {
        if(first == ce64s.elementAti(i)) {
            if(i > Collation::MAX_INDEX) {
                errorCode = U_BUFFER_OVERFLOW_ERROR;
                return 0;
            }
            for(int32_t j = 1;; ++j) {
                if(j == length) {
                    return Collation::makeCE32FromTagIndexAndLength(
                            Collation::EXPANSION_TAG, i, length);
                }
                if(ce64s.elementAti(i + j) != ces[j]) { break; }
            }
        }
    }
    // Store the new sequence.
    int32_t i = ce64s.size();
    if(i > Collation::MAX_INDEX) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return 0;
    }
    for(int32_t j = 0; j < length; ++j) {
        ce64s.addElement(ces[j], errorCode);
    }
    return Collation::makeCE32FromTagIndexAndLength(Collation::EXPANSION_TAG, i, length);
}

uint32_t
CollationDataBuilder::encodeExpansion32(const int32_t newCE32s[], int32_t length,
                                        UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    // See if this sequence of CE32s has already been stored.
    int32_t first = newCE32s[0];
    int32_t ce32sMax = ce32s.size() - length;
    for(int32_t i = 0; i <= ce32sMax; ++i) {
        if(first == ce32s.elementAti(i)) {
            if(i > Collation::MAX_INDEX) {
                errorCode = U_BUFFER_OVERFLOW_ERROR;
                return 0;
            }
            for(int32_t j = 1;; ++j) {
                if(j == length) {
                    return Collation::makeCE32FromTagIndexAndLength(
                            Collation::EXPANSION32_TAG, i, length);
                }
                if(ce32s.elementAti(i + j) != newCE32s[j]) { break; }
            }
        }
    }
    // Store the new sequence.
    int32_t i = ce32s.size();
    if(i > Collation::MAX_INDEX) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return 0;
    }
    for(int32_t j = 0; j < length; ++j) {
        ce32s.addElement(newCE32s[j], errorCode);
    }
    return Collation::makeCE32FromTagIndexAndLength(Collation::EXPANSION32_TAG, i, length);
}

uint32_t
CollationDataBuilder::copyFromBaseCE32(UChar32 c, uint32_t ce32, UBool withContext,
                                       UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    if(!Collation::isSpecialCE32(ce32)) { return ce32; }
    switch(Collation::tagFromCE32(ce32)) {
    case Collation::LONG_PRIMARY_TAG:
    case Collation::LONG_SECONDARY_TAG:
    case Collation::LATIN_EXPANSION_TAG:
        // copy as is
        break;
    case Collation::EXPANSION32_TAG: {
        const uint32_t *baseCE32s = base->ce32s + Collation::indexFromCE32(ce32);
        int32_t length = Collation::lengthFromCE32(ce32);
        ce32 = encodeExpansion32(
            reinterpret_cast<const int32_t *>(baseCE32s), length, errorCode);
        break;
    }
    case Collation::EXPANSION_TAG: {
        const int64_t *baseCEs = base->ces + Collation::indexFromCE32(ce32);
        int32_t length = Collation::lengthFromCE32(ce32);
        ce32 = encodeExpansion(baseCEs, length, errorCode);
        break;
    }
    case Collation::PREFIX_TAG: {
        // Flatten prefixes and nested suffixes (contractions)
        // into a linear list of ConditionalCE32.
        const UChar *p = base->contexts + Collation::indexFromCE32(ce32);
        ce32 = CollationData::readCE32(p);  // Default if no prefix match.
        if(!withContext) {
            return copyFromBaseCE32(c, ce32, FALSE, errorCode);
        }
        ConditionalCE32 head;
        UnicodeString context((UChar)0);
        int32_t index;
        if(Collation::isContractionCE32(ce32)) {
            index = copyContractionsFromBaseCE32(context, c, ce32, &head, errorCode);
        } else {
            ce32 = copyFromBaseCE32(c, ce32, TRUE, errorCode);
            head.next = index = addConditionalCE32(context, ce32, errorCode);
        }
        if(U_FAILURE(errorCode)) { return 0; }
        ConditionalCE32 *cond = getConditionalCE32(index);  // the last ConditionalCE32 so far
        UCharsTrie::Iterator prefixes(p + 2, 0, errorCode);
        while(prefixes.next(errorCode)) {
            context = prefixes.getString();
            context.reverse();
            context.insert(0, (UChar)context.length());
            ce32 = (uint32_t)prefixes.getValue();
            if(Collation::isContractionCE32(ce32)) {
                index = copyContractionsFromBaseCE32(context, c, ce32, cond, errorCode);
            } else {
                ce32 = copyFromBaseCE32(c, ce32, TRUE, errorCode);
                cond->next = index = addConditionalCE32(context, ce32, errorCode);
            }
            if(U_FAILURE(errorCode)) { return 0; }
            cond = getConditionalCE32(index);
        }
        ce32 = makeBuilderContextCE32(head.next);
        contextChars.add(c);
        break;
    }
    case Collation::CONTRACTION_TAG: {
        if(!withContext) {
            const UChar *p = base->contexts + Collation::indexFromCE32(ce32);
            ce32 = CollationData::readCE32(p);  // Default if no suffix match.
            return copyFromBaseCE32(c, ce32, FALSE, errorCode);
        }
        ConditionalCE32 head;
        UnicodeString context((UChar)0);
        copyContractionsFromBaseCE32(context, c, ce32, &head, errorCode);
        ce32 = makeBuilderContextCE32(head.next);
        contextChars.add(c);
        break;
    }
    case Collation::HANGUL_TAG:
        errorCode = U_UNSUPPORTED_ERROR;  // We forbid tailoring of Hangul syllables.
        break;
    case Collation::OFFSET_TAG:
        ce32 = getCE32FromOffsetCE32(TRUE, c, ce32);
        break;
    case Collation::IMPLICIT_TAG:
        ce32 = encodeOneCE(Collation::unassignedCEFromCodePoint(c), errorCode);
        break;
    default:
        UPRV_UNREACHABLE;  // require ce32 == base->getFinalCE32(ce32)
    }
    return ce32;
}

int32_t
CollationDataBuilder::copyContractionsFromBaseCE32(UnicodeString &context, UChar32 c, uint32_t ce32,
                                                   ConditionalCE32 *cond, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    const UChar *p = base->contexts + Collation::indexFromCE32(ce32);
    int32_t index;
    if((ce32 & Collation::CONTRACT_SINGLE_CP_NO_MATCH) != 0) {
        // No match on the single code point.
        // We are underneath a prefix, and the default mapping is just
        // a fallback to the mappings for a shorter prefix.
        U_ASSERT(context.length() > 1);
        index = -1;
    } else {
        ce32 = CollationData::readCE32(p);  // Default if no suffix match.
        U_ASSERT(!Collation::isContractionCE32(ce32));
        ce32 = copyFromBaseCE32(c, ce32, TRUE, errorCode);
        cond->next = index = addConditionalCE32(context, ce32, errorCode);
        if(U_FAILURE(errorCode)) { return 0; }
        cond = getConditionalCE32(index);
    }

    int32_t suffixStart = context.length();
    UCharsTrie::Iterator suffixes(p + 2, 0, errorCode);
    while(suffixes.next(errorCode)) {
        context.append(suffixes.getString());
        ce32 = copyFromBaseCE32(c, (uint32_t)suffixes.getValue(), TRUE, errorCode);
        cond->next = index = addConditionalCE32(context, ce32, errorCode);
        if(U_FAILURE(errorCode)) { return 0; }
        // No need to update the unsafeBackwardSet because the tailoring set
        // is already a copy of the base set.
        cond = getConditionalCE32(index);
        context.truncate(suffixStart);
    }
    U_ASSERT(index >= 0);
    return index;
}

class CopyHelper {
public:
    CopyHelper(const CollationDataBuilder &s, CollationDataBuilder &d,
               const CollationDataBuilder::CEModifier &m, UErrorCode &initialErrorCode)
            : src(s), dest(d), modifier(m),
              errorCode(initialErrorCode) {}

    UBool copyRangeCE32(UChar32 start, UChar32 end, uint32_t ce32) {
        ce32 = copyCE32(ce32);
        utrie2_setRange32(dest.trie, start, end, ce32, TRUE, &errorCode);
        if(CollationDataBuilder::isBuilderContextCE32(ce32)) {
            dest.contextChars.add(start, end);
        }
        return U_SUCCESS(errorCode);
    }

    uint32_t copyCE32(uint32_t ce32) {
        if(!Collation::isSpecialCE32(ce32)) {
            int64_t ce = modifier.modifyCE32(ce32);
            if(ce != Collation::NO_CE) {
                ce32 = dest.encodeOneCE(ce, errorCode);
            }
        } else {
            int32_t tag = Collation::tagFromCE32(ce32);
            if(tag == Collation::EXPANSION32_TAG) {
                const uint32_t *srcCE32s = reinterpret_cast<uint32_t *>(src.ce32s.getBuffer());
                srcCE32s += Collation::indexFromCE32(ce32);
                int32_t length = Collation::lengthFromCE32(ce32);
                // Inspect the source CE32s. Just copy them if none are modified.
                // Otherwise copy to modifiedCEs, with modifications.
                UBool isModified = FALSE;
                for(int32_t i = 0; i < length; ++i) {
                    ce32 = srcCE32s[i];
                    int64_t ce;
                    if(Collation::isSpecialCE32(ce32) ||
                            (ce = modifier.modifyCE32(ce32)) == Collation::NO_CE) {
                        if(isModified) {
                            modifiedCEs[i] = Collation::ceFromCE32(ce32);
                        }
                    } else {
                        if(!isModified) {
                            for(int32_t j = 0; j < i; ++j) {
                                modifiedCEs[j] = Collation::ceFromCE32(srcCE32s[j]);
                            }
                            isModified = TRUE;
                        }
                        modifiedCEs[i] = ce;
                    }
                }
                if(isModified) {
                    ce32 = dest.encodeCEs(modifiedCEs, length, errorCode);
                } else {
                    ce32 = dest.encodeExpansion32(
                        reinterpret_cast<const int32_t *>(srcCE32s), length, errorCode);
                }
            } else if(tag == Collation::EXPANSION_TAG) {
                const int64_t *srcCEs = src.ce64s.getBuffer();
                srcCEs += Collation::indexFromCE32(ce32);
                int32_t length = Collation::lengthFromCE32(ce32);
                // Inspect the source CEs. Just copy them if none are modified.
                // Otherwise copy to modifiedCEs, with modifications.
                UBool isModified = FALSE;
                for(int32_t i = 0; i < length; ++i) {
                    int64_t srcCE = srcCEs[i];
                    int64_t ce = modifier.modifyCE(srcCE);
                    if(ce == Collation::NO_CE) {
                        if(isModified) {
                            modifiedCEs[i] = srcCE;
                        }
                    } else {
                        if(!isModified) {
                            for(int32_t j = 0; j < i; ++j) {
                                modifiedCEs[j] = srcCEs[j];
                            }
                            isModified = TRUE;
                        }
                        modifiedCEs[i] = ce;
                    }
                }
                if(isModified) {
                    ce32 = dest.encodeCEs(modifiedCEs, length, errorCode);
                } else {
                    ce32 = dest.encodeExpansion(srcCEs, length, errorCode);
                }
            } else if(tag == Collation::BUILDER_DATA_TAG) {
                // Copy the list of ConditionalCE32.
                ConditionalCE32 *cond = src.getConditionalCE32ForCE32(ce32);
                U_ASSERT(!cond->hasContext());
                int32_t destIndex = dest.addConditionalCE32(
                        cond->context, copyCE32(cond->ce32), errorCode);
                ce32 = CollationDataBuilder::makeBuilderContextCE32(destIndex);
                while(cond->next >= 0) {
                    cond = src.getConditionalCE32(cond->next);
                    ConditionalCE32 *prevDestCond = dest.getConditionalCE32(destIndex);
                    destIndex = dest.addConditionalCE32(
                            cond->context, copyCE32(cond->ce32), errorCode);
                    int32_t suffixStart = cond->prefixLength() + 1;
                    dest.unsafeBackwardSet.addAll(cond->context.tempSubString(suffixStart));
                    prevDestCond->next = destIndex;
                }
            } else {
                // Just copy long CEs and Latin mini expansions (and other expected values) as is,
                // assuming that the modifier would not modify them.
                U_ASSERT(tag == Collation::LONG_PRIMARY_TAG ||
                        tag == Collation::LONG_SECONDARY_TAG ||
                        tag == Collation::LATIN_EXPANSION_TAG ||
                        tag == Collation::HANGUL_TAG);
            }
        }
        return ce32;
    }

    const CollationDataBuilder &src;
    CollationDataBuilder &dest;
    const CollationDataBuilder::CEModifier &modifier;
    int64_t modifiedCEs[Collation::MAX_EXPANSION_LENGTH];
    UErrorCode errorCode;
};

U_CDECL_BEGIN

static UBool U_CALLCONV
enumRangeForCopy(const void *context, UChar32 start, UChar32 end, uint32_t value) {
    return
        value == Collation::UNASSIGNED_CE32 || value == Collation::FALLBACK_CE32 ||
        ((CopyHelper *)context)->copyRangeCE32(start, end, value);
}

U_CDECL_END

void
CollationDataBuilder::copyFrom(const CollationDataBuilder &src, const CEModifier &modifier,
                               UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(trie == NULL || utrie2_isFrozen(trie)) {
        errorCode = U_INVALID_STATE_ERROR;
        return;
    }
    CopyHelper helper(src, *this, modifier, errorCode);
    utrie2_enum(src.trie, NULL, enumRangeForCopy, &helper);
    errorCode = helper.errorCode;
    // Update the contextChars and the unsafeBackwardSet while copying,
    // in case a character had conditional mappings in the source builder
    // and they were removed later.
    modified |= src.modified;
}

void
CollationDataBuilder::optimize(const UnicodeSet &set, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode) || set.isEmpty()) { return; }
    UnicodeSetIterator iter(set);
    while(iter.next() && !iter.isString()) {
        UChar32 c = iter.getCodepoint();
        uint32_t ce32 = utrie2_get32(trie, c);
        if(ce32 == Collation::FALLBACK_CE32) {
            ce32 = base->getFinalCE32(base->getCE32(c));
            ce32 = copyFromBaseCE32(c, ce32, TRUE, errorCode);
            utrie2_set32(trie, c, ce32, &errorCode);
        }
    }
    modified = TRUE;
}

void
CollationDataBuilder::suppressContractions(const UnicodeSet &set, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode) || set.isEmpty()) { return; }
    UnicodeSetIterator iter(set);
    while(iter.next() && !iter.isString()) {
        UChar32 c = iter.getCodepoint();
        uint32_t ce32 = utrie2_get32(trie, c);
        if(ce32 == Collation::FALLBACK_CE32) {
            ce32 = base->getFinalCE32(base->getCE32(c));
            if(Collation::ce32HasContext(ce32)) {
                ce32 = copyFromBaseCE32(c, ce32, FALSE /* without context */, errorCode);
                utrie2_set32(trie, c, ce32, &errorCode);
            }
        } else if(isBuilderContextCE32(ce32)) {
            ce32 = getConditionalCE32ForCE32(ce32)->ce32;
            // Simply abandon the list of ConditionalCE32.
            // The caller will copy this builder in the end,
            // eliminating unreachable data.
            utrie2_set32(trie, c, ce32, &errorCode);
            contextChars.remove(c);
        }
    }
    modified = TRUE;
}

UBool
CollationDataBuilder::getJamoCE32s(uint32_t jamoCE32s[], UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    UBool anyJamoAssigned = base == NULL;  // always set jamoCE32s in the base data
    UBool needToCopyFromBase = FALSE;
    for(int32_t j = 0; j < CollationData::JAMO_CE32S_LENGTH; ++j) {  // Count across Jamo types.
        UChar32 jamo = jamoCpFromIndex(j);
        UBool fromBase = FALSE;
        uint32_t ce32 = utrie2_get32(trie, jamo);
        anyJamoAssigned |= Collation::isAssignedCE32(ce32);
        // TODO: Try to prevent [optimize [Jamo]] from counting as anyJamoAssigned.
        // (As of CLDR 24 [2013] the Korean tailoring does not optimize conjoining Jamo.)
        if(ce32 == Collation::FALLBACK_CE32) {
            fromBase = TRUE;
            ce32 = base->getCE32(jamo);
        }
        if(Collation::isSpecialCE32(ce32)) {
            switch(Collation::tagFromCE32(ce32)) {
            case Collation::LONG_PRIMARY_TAG:
            case Collation::LONG_SECONDARY_TAG:
            case Collation::LATIN_EXPANSION_TAG:
                // Copy the ce32 as-is.
                break;
            case Collation::EXPANSION32_TAG:
            case Collation::EXPANSION_TAG:
            case Collation::PREFIX_TAG:
            case Collation::CONTRACTION_TAG:
                if(fromBase) {
                    // Defer copying until we know if anyJamoAssigned.
                    ce32 = Collation::FALLBACK_CE32;
                    needToCopyFromBase = TRUE;
                }
                break;
            case Collation::IMPLICIT_TAG:
                // An unassigned Jamo should only occur in tests with incomplete bases.
                U_ASSERT(fromBase);
                ce32 = Collation::FALLBACK_CE32;
                needToCopyFromBase = TRUE;
                break;
            case Collation::OFFSET_TAG:
                ce32 = getCE32FromOffsetCE32(fromBase, jamo, ce32);
                break;
            case Collation::FALLBACK_TAG:
            case Collation::RESERVED_TAG_3:
            case Collation::BUILDER_DATA_TAG:
            case Collation::DIGIT_TAG:
            case Collation::U0000_TAG:
            case Collation::HANGUL_TAG:
            case Collation::LEAD_SURROGATE_TAG:
                errorCode = U_INTERNAL_PROGRAM_ERROR;
                return FALSE;
            }
        }
        jamoCE32s[j] = ce32;
    }
    if(anyJamoAssigned && needToCopyFromBase) {
        for(int32_t j = 0; j < CollationData::JAMO_CE32S_LENGTH; ++j) {
            if(jamoCE32s[j] == Collation::FALLBACK_CE32) {
                UChar32 jamo = jamoCpFromIndex(j);
                jamoCE32s[j] = copyFromBaseCE32(jamo, base->getCE32(jamo),
                                                /*withContext=*/ TRUE, errorCode);
            }
        }
    }
    return anyJamoAssigned && U_SUCCESS(errorCode);
}

void
CollationDataBuilder::setDigitTags(UErrorCode &errorCode) {
    UnicodeSet digits(UNICODE_STRING_SIMPLE("[:Nd:]"), errorCode);
    if(U_FAILURE(errorCode)) { return; }
    UnicodeSetIterator iter(digits);
    while(iter.next()) {
        U_ASSERT(!iter.isString());
        UChar32 c = iter.getCodepoint();
        uint32_t ce32 = utrie2_get32(trie, c);
        if(ce32 != Collation::FALLBACK_CE32 && ce32 != Collation::UNASSIGNED_CE32) {
            int32_t index = addCE32(ce32, errorCode);
            if(U_FAILURE(errorCode)) { return; }
            if(index > Collation::MAX_INDEX) {
                errorCode = U_BUFFER_OVERFLOW_ERROR;
                return;
            }
            ce32 = Collation::makeCE32FromTagIndexAndLength(
                    Collation::DIGIT_TAG, index, u_charDigitValue(c));
            utrie2_set32(trie, c, ce32, &errorCode);
        }
    }
}

U_CDECL_BEGIN

static UBool U_CALLCONV
enumRangeLeadValue(const void *context, UChar32 /*start*/, UChar32 /*end*/, uint32_t value) {
    int32_t *pValue = (int32_t *)context;
    if(value == Collation::UNASSIGNED_CE32) {
        value = Collation::LEAD_ALL_UNASSIGNED;
    } else if(value == Collation::FALLBACK_CE32) {
        value = Collation::LEAD_ALL_FALLBACK;
    } else {
        *pValue = Collation::LEAD_MIXED;
        return FALSE;
    }
    if(*pValue < 0) {
        *pValue = (int32_t)value;
    } else if(*pValue != (int32_t)value) {
        *pValue = Collation::LEAD_MIXED;
        return FALSE;
    }
    return TRUE;
}

U_CDECL_END

void
CollationDataBuilder::setLeadSurrogates(UErrorCode &errorCode) {
    for(UChar lead = 0xd800; lead < 0xdc00; ++lead) {
        int32_t value = -1;
        utrie2_enumForLeadSurrogate(trie, lead, NULL, enumRangeLeadValue, &value);
        utrie2_set32ForLeadSurrogateCodeUnit(
            trie, lead,
            Collation::makeCE32FromTagAndIndex(Collation::LEAD_SURROGATE_TAG, 0) | (uint32_t)value,
            &errorCode);
    }
}

void
CollationDataBuilder::build(CollationData &data, UErrorCode &errorCode) {
    buildMappings(data, errorCode);
    if(base != NULL) {
        data.numericPrimary = base->numericPrimary;
        data.compressibleBytes = base->compressibleBytes;
        data.numScripts = base->numScripts;
        data.scriptsIndex = base->scriptsIndex;
        data.scriptStarts = base->scriptStarts;
        data.scriptStartsLength = base->scriptStartsLength;
    }
    buildFastLatinTable(data, errorCode);
}

void
CollationDataBuilder::buildMappings(CollationData &data, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(trie == NULL || utrie2_isFrozen(trie)) {
        errorCode = U_INVALID_STATE_ERROR;
        return;
    }

    buildContexts(errorCode);

    uint32_t jamoCE32s[CollationData::JAMO_CE32S_LENGTH];
    int32_t jamoIndex = -1;
    if(getJamoCE32s(jamoCE32s, errorCode)) {
        jamoIndex = ce32s.size();
        for(int32_t i = 0; i < CollationData::JAMO_CE32S_LENGTH; ++i) {
            ce32s.addElement((int32_t)jamoCE32s[i], errorCode);
        }
        // Small optimization: Use a bit in the Hangul ce32
        // to indicate that none of the Jamo CE32s are isSpecialCE32()
        // (as it should be in the root collator).
        // It allows CollationIterator to avoid recursive function calls and per-Jamo tests.
        // In order to still have good trie compression and keep this code simple,
        // we only set this flag if a whole block of 588 Hangul syllables starting with
        // a common leading consonant (Jamo L) has this property.
        UBool isAnyJamoVTSpecial = FALSE;
        for(int32_t i = Hangul::JAMO_L_COUNT; i < CollationData::JAMO_CE32S_LENGTH; ++i) {
            if(Collation::isSpecialCE32(jamoCE32s[i])) {
                isAnyJamoVTSpecial = TRUE;
                break;
            }
        }
        uint32_t hangulCE32 = Collation::makeCE32FromTagAndIndex(Collation::HANGUL_TAG, 0);
        UChar32 c = Hangul::HANGUL_BASE;
        for(int32_t i = 0; i < Hangul::JAMO_L_COUNT; ++i) {  // iterate over the Jamo L
            uint32_t ce32 = hangulCE32;
            if(!isAnyJamoVTSpecial && !Collation::isSpecialCE32(jamoCE32s[i])) {
                ce32 |= Collation::HANGUL_NO_SPECIAL_JAMO;
            }
            UChar32 limit = c + Hangul::JAMO_VT_COUNT;
            utrie2_setRange32(trie, c, limit - 1, ce32, TRUE, &errorCode);
            c = limit;
        }
    } else {
        // Copy the Hangul CE32s from the base in blocks per Jamo L,
        // assuming that HANGUL_NO_SPECIAL_JAMO is set or not set for whole blocks.
        for(UChar32 c = Hangul::HANGUL_BASE; c < Hangul::HANGUL_LIMIT;) {
            uint32_t ce32 = base->getCE32(c);
            U_ASSERT(Collation::hasCE32Tag(ce32, Collation::HANGUL_TAG));
            UChar32 limit = c + Hangul::JAMO_VT_COUNT;
            utrie2_setRange32(trie, c, limit - 1, ce32, TRUE, &errorCode);
            c = limit;
        }
    }

    setDigitTags(errorCode);
    setLeadSurrogates(errorCode);

    // For U+0000, move its normal ce32 into CE32s[0] and set U0000_TAG.
    ce32s.setElementAt((int32_t)utrie2_get32(trie, 0), 0);
    utrie2_set32(trie, 0, Collation::makeCE32FromTagAndIndex(Collation::U0000_TAG, 0), &errorCode);

    utrie2_freeze(trie, UTRIE2_32_VALUE_BITS, &errorCode);
    if(U_FAILURE(errorCode)) { return; }

    // Mark each lead surrogate as "unsafe"
    // if any of its 1024 associated supplementary code points is "unsafe".
    UChar32 c = 0x10000;
    for(UChar lead = 0xd800; lead < 0xdc00; ++lead, c += 0x400) {
        if(unsafeBackwardSet.containsSome(c, c + 0x3ff)) {
            unsafeBackwardSet.add(lead);
        }
    }
    unsafeBackwardSet.freeze();

    data.trie = trie;
    data.ce32s = reinterpret_cast<const uint32_t *>(ce32s.getBuffer());
    data.ces = ce64s.getBuffer();
    data.contexts = contexts.getBuffer();

    data.ce32sLength = ce32s.size();
    data.cesLength = ce64s.size();
    data.contextsLength = contexts.length();

    data.base = base;
    if(jamoIndex >= 0) {
        data.jamoCE32s = data.ce32s + jamoIndex;
    } else {
        data.jamoCE32s = base->jamoCE32s;
    }
    data.unsafeBackwardSet = &unsafeBackwardSet;
}

void
CollationDataBuilder::clearContexts() {
    contexts.remove();
    UnicodeSetIterator iter(contextChars);
    while(iter.next()) {
        U_ASSERT(!iter.isString());
        uint32_t ce32 = utrie2_get32(trie, iter.getCodepoint());
        U_ASSERT(isBuilderContextCE32(ce32));
        getConditionalCE32ForCE32(ce32)->builtCE32 = Collation::NO_CE32;
    }
}

void
CollationDataBuilder::buildContexts(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    // Ignore abandoned lists and the cached builtCE32,
    // and build all contexts from scratch.
    contexts.remove();
    UnicodeSetIterator iter(contextChars);
    while(U_SUCCESS(errorCode) && iter.next()) {
        U_ASSERT(!iter.isString());
        UChar32 c = iter.getCodepoint();
        uint32_t ce32 = utrie2_get32(trie, c);
        if(!isBuilderContextCE32(ce32)) {
            // Impossible: No context data for c in contextChars.
            errorCode = U_INTERNAL_PROGRAM_ERROR;
            return;
        }
        ConditionalCE32 *cond = getConditionalCE32ForCE32(ce32);
        ce32 = buildContext(cond, errorCode);
        utrie2_set32(trie, c, ce32, &errorCode);
    }
}

uint32_t
CollationDataBuilder::buildContext(ConditionalCE32 *head, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    // The list head must have no context.
    U_ASSERT(!head->hasContext());
    // The list head must be followed by one or more nodes that all do have context.
    U_ASSERT(head->next >= 0);
    UCharsTrieBuilder prefixBuilder(errorCode);
    UCharsTrieBuilder contractionBuilder(errorCode);
    for(ConditionalCE32 *cond = head;; cond = getConditionalCE32(cond->next)) {
        // After the list head, the prefix or suffix can be empty, but not both.
        U_ASSERT(cond == head || cond->hasContext());
        int32_t prefixLength = cond->prefixLength();
        UnicodeString prefix(cond->context, 0, prefixLength + 1);
        // Collect all contraction suffixes for one prefix.
        ConditionalCE32 *firstCond = cond;
        ConditionalCE32 *lastCond = cond;
        while(cond->next >= 0 &&
                (cond = getConditionalCE32(cond->next))->context.startsWith(prefix)) {
            lastCond = cond;
        }
        uint32_t ce32;
        int32_t suffixStart = prefixLength + 1;  // == prefix.length()
        if(lastCond->context.length() == suffixStart) {
            // One prefix without contraction suffix.
            U_ASSERT(firstCond == lastCond);
            ce32 = lastCond->ce32;
            cond = lastCond;
        } else {
            // Build the contractions trie.
            contractionBuilder.clear();
            // Entry for an empty suffix, to be stored before the trie.
            uint32_t emptySuffixCE32 = 0;
            uint32_t flags = 0;
            if(firstCond->context.length() == suffixStart) {
                // There is a mapping for the prefix and the single character c. (p|c)
                // If no other suffix matches, then we return this value.
                emptySuffixCE32 = firstCond->ce32;
                cond = getConditionalCE32(firstCond->next);
            } else {
                // There is no mapping for the prefix and just the single character.
                // (There is no p|c, only p|cd, p|ce etc.)
                flags |= Collation::CONTRACT_SINGLE_CP_NO_MATCH;
                // When the prefix matches but none of the prefix-specific suffixes,
                // then we fall back to the mappings with the next-longest prefix,
                // and ultimately to mappings with no prefix.
                // Each fallback might be another set of contractions.
                // For example, if there are mappings for ch, p|cd, p|ce, but not for p|c,
                // then in text "pch" we find the ch contraction.
                for(cond = head;; cond = getConditionalCE32(cond->next)) {
                    int32_t length = cond->prefixLength();
                    if(length == prefixLength) { break; }
                    if(cond->defaultCE32 != Collation::NO_CE32 &&
                            (length==0 || prefix.endsWith(cond->context, 1, length))) {
                        emptySuffixCE32 = cond->defaultCE32;
                    }
                }
                cond = firstCond;
            }
            // Optimization: Set a flag when
            // the first character of every contraction suffix has lccc!=0.
            // Short-circuits contraction matching when a normal letter follows.
            flags |= Collation::CONTRACT_NEXT_CCC;
            // Add all of the non-empty suffixes into the contraction trie.
            for(;;) {
                UnicodeString suffix(cond->context, suffixStart);
                uint16_t fcd16 = nfcImpl.getFCD16(suffix.char32At(0));
                if(fcd16 <= 0xff) {
                    flags &= ~Collation::CONTRACT_NEXT_CCC;
                }
                fcd16 = nfcImpl.getFCD16(suffix.char32At(suffix.length() - 1));
                if(fcd16 > 0xff) {
                    // The last suffix character has lccc!=0, allowing for discontiguous contractions.
                    flags |= Collation::CONTRACT_TRAILING_CCC;
                }
                contractionBuilder.add(suffix, (int32_t)cond->ce32, errorCode);
                if(cond == lastCond) { break; }
                cond = getConditionalCE32(cond->next);
            }
            int32_t index = addContextTrie(emptySuffixCE32, contractionBuilder, errorCode);
            if(U_FAILURE(errorCode)) { return 0; }
            if(index > Collation::MAX_INDEX) {
                errorCode = U_BUFFER_OVERFLOW_ERROR;
                return 0;
            }
            ce32 = Collation::makeCE32FromTagAndIndex(Collation::CONTRACTION_TAG, index) | flags;
        }
        U_ASSERT(cond == lastCond);
        firstCond->defaultCE32 = ce32;
        if(prefixLength == 0) {
            if(cond->next < 0) {
                // No non-empty prefixes, only contractions.
                return ce32;
            }
        } else {
            prefix.remove(0, 1);  // Remove the length unit.
            prefix.reverse();
            prefixBuilder.add(prefix, (int32_t)ce32, errorCode);
            if(cond->next < 0) { break; }
        }
    }
    U_ASSERT(head->defaultCE32 != Collation::NO_CE32);
    int32_t index = addContextTrie(head->defaultCE32, prefixBuilder, errorCode);
    if(U_FAILURE(errorCode)) { return 0; }
    if(index > Collation::MAX_INDEX) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return 0;
    }
    return Collation::makeCE32FromTagAndIndex(Collation::PREFIX_TAG, index);
}

int32_t
CollationDataBuilder::addContextTrie(uint32_t defaultCE32, UCharsTrieBuilder &trieBuilder,
                                     UErrorCode &errorCode) {
    UnicodeString context;
    context.append((UChar)(defaultCE32 >> 16)).append((UChar)defaultCE32);
    UnicodeString trieString;
    context.append(trieBuilder.buildUnicodeString(USTRINGTRIE_BUILD_SMALL, trieString, errorCode));
    if(U_FAILURE(errorCode)) { return -1; }
    int32_t index = contexts.indexOf(context);
    if(index < 0) {
        index = contexts.length();
        contexts.append(context);
    }
    return index;
}

void
CollationDataBuilder::buildFastLatinTable(CollationData &data, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode) || !fastLatinEnabled) { return; }

    delete fastLatinBuilder;
    fastLatinBuilder = new CollationFastLatinBuilder(errorCode);
    if(fastLatinBuilder == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    if(fastLatinBuilder->forData(data, errorCode)) {
        const uint16_t *table = fastLatinBuilder->getTable();
        int32_t length = fastLatinBuilder->lengthOfTable();
        if(base != NULL && length == base->fastLatinTableLength &&
                uprv_memcmp(table, base->fastLatinTable, length * 2) == 0) {
            // Same fast Latin table as in the base, use that one instead.
            delete fastLatinBuilder;
            fastLatinBuilder = NULL;
            table = base->fastLatinTable;
        }
        data.fastLatinTable = table;
        data.fastLatinTableLength = length;
    } else {
        delete fastLatinBuilder;
        fastLatinBuilder = NULL;
    }
}

int32_t
CollationDataBuilder::getCEs(const UnicodeString &s, int64_t ces[], int32_t cesLength) {
    return getCEs(s, 0, ces, cesLength);
}

int32_t
CollationDataBuilder::getCEs(const UnicodeString &prefix, const UnicodeString &s,
                             int64_t ces[], int32_t cesLength) {
    int32_t prefixLength = prefix.length();
    if(prefixLength == 0) {
        return getCEs(s, 0, ces, cesLength);
    } else {
        return getCEs(prefix + s, prefixLength, ces, cesLength);
    }
}

int32_t
CollationDataBuilder::getCEs(const UnicodeString &s, int32_t start,
                             int64_t ces[], int32_t cesLength) {
    if(collIter == NULL) {
        collIter = new DataBuilderCollationIterator(*this);
        if(collIter == NULL) { return 0; }
    }
    return collIter->fetchCEs(s, start, ces, cesLength);
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationdatareader.cpp
*
* created on: 2013feb07
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION







// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationdatareader.h
*
* created on: 2013feb07
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION



struct UDataMemory;

U_NAMESPACE_BEGIN

struct CollationTailoring;

/**
 * Collation binary data reader.
 */
struct U_I18N_API CollationDataReader /* all static */ {
    // The following constants are also copied into source/common/ucol_swp.cpp.
    // Keep them in sync!
    enum {
        /**
         * Number of int32_t indexes.
         *
         * Can be 2 if there are only options.
         * Can be 7 or 8 if there are only options and a script reordering.
         * The loader treats any index>=indexes[IX_INDEXES_LENGTH] as 0.
         */
        IX_INDEXES_LENGTH,  // 0
        /**
         * Bits 31..24: numericPrimary, for numeric collation
         *      23..16: fast Latin format version (0 = no fast Latin table)
         *      15.. 0: options bit set
         */
        IX_OPTIONS,
        IX_RESERVED2,
        IX_RESERVED3,

        /** Array offset to Jamo CE32s in ce32s[], or <0 if none. */
        IX_JAMO_CE32S_START,  // 4

        // Byte offsets from the start of the data, after the generic header.
        // The indexes[] are at byte offset 0, other data follows.
        // Each data item is aligned properly.
        // The data items should be in descending order of unit size,
        // to minimize the need for padding.
        // Each item's byte length is given by the difference between its offset and
        // the next index/offset value.
        /** Byte offset to int32_t reorderCodes[]. */
        IX_REORDER_CODES_OFFSET,
        /**
         * Byte offset to uint8_t reorderTable[].
         * Empty table if <256 bytes (padding only).
         * Otherwise 256 bytes or more (with padding).
         */
        IX_REORDER_TABLE_OFFSET,
        /** Byte offset to the collation trie. Its length is a multiple of 8 bytes. */
        IX_TRIE_OFFSET,

        IX_RESERVED8_OFFSET,  // 8
        /** Byte offset to int64_t ces[]. */
        IX_CES_OFFSET,
        IX_RESERVED10_OFFSET,
        /** Byte offset to uint32_t ce32s[]. */
        IX_CE32S_OFFSET,

        /** Byte offset to uint32_t rootElements[]. */
        IX_ROOT_ELEMENTS_OFFSET,  // 12
        /** Byte offset to UChar *contexts[]. */
        IX_CONTEXTS_OFFSET,
        /** Byte offset to uint16_t [] with serialized unsafeBackwardSet. */
        IX_UNSAFE_BWD_OFFSET,
        /** Byte offset to uint16_t fastLatinTable[]. */
        IX_FAST_LATIN_TABLE_OFFSET,

        /** Byte offset to uint16_t scripts[]. */
        IX_SCRIPTS_OFFSET,  // 16
        /**
         * Byte offset to UBool compressibleBytes[].
         * Empty table if <256 bytes (padding only).
         * Otherwise 256 bytes or more (with padding).
         */
        IX_COMPRESSIBLE_BYTES_OFFSET,
        IX_RESERVED18_OFFSET,
        IX_TOTAL_SIZE
    };

    static void read(const CollationTailoring *base, const uint8_t *inBytes, int32_t inLength,
                     CollationTailoring &tailoring, UErrorCode &errorCode);

    static UBool U_CALLCONV
    isAcceptable(void *context, const char *type, const char *name, const UDataInfo *pInfo);

private:
    CollationDataReader();  // no constructor
};

/*
 * Format of collation data (ucadata.icu, binary data in coll/ *.res files).
 * Format version 5.
 *
 * The root collation data is stored in the ucadata.icu file.
 * Tailorings are stored inside .res resource bundle files, with a complete file header.
 *
 * Collation data begins with a standard ICU data file header
 * (DataHeader, see ucmndata.h and unicode/udata.h).
 * The UDataInfo.dataVersion field contains the UCA and other version numbers,
 * see the comments for CollationTailoring.version.
 *
 * After the header, the file contains the following parts.
 * Constants are defined as enum values of the CollationDataReader class.
 * See also the Collation class.
 *
 * int32_t indexes[indexesLength];
 *      The indexes array has variable length.
 *      Some tailorings only need the length and the options,
 *      others only add reorderCodes and the reorderTable,
 *      some need to store mappings.
 *      Only as many indexes are stored as needed to read all of the data.
 *
 *      Index 0: indexesLength
 *      Index 1: numericPrimary, CollationFastLatin::VERSION, and options: see IX_OPTIONS
 *      Index 2..3: Unused/reserved/0.
 *      Index 4: Index into the ce32s array where the CE32s of the conjoining Jamo
 *               are stored in a short, contiguous part of the ce32s array.
 *
 *      Indexes 5..19 are byte offsets in ascending order.
 *      Each byte offset marks the start of the next part in the data file,
 *      and the end of the previous one.
 *      When two consecutive byte offsets are the same (or too short),
 *      then the corresponding part is empty.
 *      Byte offsets are offsets from after the header,
 *      that is, from the beginning of the indexes[].
 *      Each part starts at an offset with proper alignment for its data.
 *      If necessary, the previous part may include padding bytes to achieve this alignment.
 *      The last byte offset that is stored in the indexes indicates the total size of the data
 *      (starting with the indexes).
 *
 * int32_t reorderCodes[]; -- empty in root
 *      The list of script and reordering codes.
 *
 *      Beginning with format version 5, this array may optionally
 *      have trailing entries with a full list of reorder ranges
 *      as described for CollationSettings::reorderRanges.
 *
 *      Script or reorder codes are first and do not exceed 16-bit values.
 *      Range limits are stored in the upper 16 bits, and are never 0.
 *      Split this array into reorder codes and ranges at the first entry
 *      with non-zero upper 16 bits.
 *
 *      If the ranges are missing but needed for split-reordered primary lead bytes,
 *      then they are regenerated at load time.
 *
 * uint8_t reorderTable[256]; -- empty in root; can be longer to include padding bytes
 *      Primary-weight lead byte permutation table.
 *      Normally present when the reorderCodes are, but can be built at load time.
 *
 *      Beginning with format version 5, a 0 entry at a non-zero index
 *      (which is otherwise an illegal value)
 *      means that the primary lead byte is "split"
 *      (there are different offsets for primaries that share that lead byte)
 *      and the reordering offset must be determined via the reorder ranges
 *      that are either stored as part of the reorderCodes array
 *      or regenerated at load time.
 *
 * UTrie2 trie; -- see utrie2_impl.h and utrie2.h
 *      The trie holds the main collation data. Each code point is mapped to a 32-bit value.
 *      It encodes a simple collation element (CE) in compact form, unless bits 7..6 are both set,
 *      in which case it is a special CE32 and contains a 4-bit tag and further data.
 *      See the Collation class for details.
 *
 *      The trie has a value for each lead surrogate code unit with some bits encoding
 *      collective properties of the 1024 supplementary characters whose UTF-16 form starts with
 *      the lead surrogate. See Collation::LEAD_SURROGATE_TAG..
 *
 * int64_t ces[];
 *      64-bit CEs and expansions that cannot be stored in a more compact form.
 *
 * uint32_t ce32s[];
 *      CE32s for expansions in compact form, and for characters whose trie values
 *      contain special data.
 *
 * uint32_t rootElements[]; -- empty in all tailorings
 *      Compact storage for all of the CEs that occur in the root collation.
 *      See the CollationRootElements class.
 *
 * UChar *contexts[];
 *      Serialized UCharsTrie structures with prefix (pre-context) and contraction mappings.
 *
 * uint16_t unsafeBackwardSet[]; -- see UnicodeSet::serialize()
 *      Serialized form of characters that are unsafe when iterating backwards,
 *      and at the end of an identical string prefix.
 *      Back up to a safe character.
 *      Lead surrogates are "unsafe" when any of their corresponding supplementary
 *      code points are unsafe.
 *      Does not include [:^lccc=0:][:^tccc=0:].
 *      For each tailoring, the root unsafeBackwardSet is subtracted.
 *      (As a result, in many tailorings no set needs to be stored.)
 *
 * uint16_t fastLatinTable[];
 *      Optional optimization for Latin text.
 *      See the CollationFastLatin class.
 *
 * uint16_t scripts[]; -- empty in all tailorings
 *      Format version 5:
 *      uint16_t numScripts;
 *      uint16_t scriptsIndex[numScripts+16];
 *      uint16_t scriptStarts[];
 *      See CollationData::numScripts etc.
 *
 *      Format version 4:
 *      Table of the reordering groups with their first and last lead bytes,
 *      and their script and reordering codes.
 *      See CollationData::scripts.
 *
 * UBool compressibleBytes[]; -- empty in all tailorings
 *      Flag for getSortKey(), indicating primary weight lead bytes that are compressible.
 *
 * -----------------
 * Changes for formatVersion 5 (ICU 55)
 *
 * Reordering moves single scripts, not groups of scripts.
 * Reorder ranges are optionally appended to the reorderCodes,
 * and a 0 entry in the reorderTable indicates a split lead byte.
 * The scripts data has a new format.
 *
 * The rootElements may contain secondary and tertiary weights below common=05.
 * (Used for small Hiragana letters.)
 * Where is occurs, there is also an explicit unit with common secondary & tertiary weights.
 * There are no other data structure changes, but builder code needs to be able to handle such data.
 *
 * The collation element for the merge separator code point U+FFFE
 * does not necessarily have special, unique secondary/tertiary weights any more.
 */

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationkeys.h
*
* created on: 2012sep02
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION






U_NAMESPACE_BEGIN

class CollationIterator;
struct CollationDataReader;
struct CollationSettings;

class SortKeyByteSink : public ByteSink {
public:
    SortKeyByteSink(char *dest, int32_t destCapacity)
            : buffer_(dest), capacity_(destCapacity),
              appended_(0), ignore_(0) {}
    virtual ~SortKeyByteSink();

    void IgnoreBytes(int32_t numIgnore) { ignore_ = numIgnore; }

    virtual void Append(const char *bytes, int32_t n);
    void Append(uint32_t b) {
        if (ignore_ > 0) {
            --ignore_;
        } else {
            if (appended_ < capacity_ || Resize(1, appended_)) {
                buffer_[appended_] = (char)b;
            }
            ++appended_;
        }
    }
    virtual char *GetAppendBuffer(int32_t min_capacity,
                                  int32_t desired_capacity_hint,
                                  char *scratch, int32_t scratch_capacity,
                                  int32_t *result_capacity);
    int32_t NumberOfBytesAppended() const { return appended_; }

    /**
     * @return how many bytes can be appended (including ignored ones)
     *         without reallocation
     */
    int32_t GetRemainingCapacity() const {
        // Either ignore_ or appended_ should be 0.
        return ignore_ + capacity_ - appended_;
    }

    UBool Overflowed() const { return appended_ > capacity_; }
    /** @return FALSE if memory allocation failed */
    UBool IsOk() const { return buffer_ != NULL; }

protected:
    virtual void AppendBeyondCapacity(const char *bytes, int32_t n, int32_t length) = 0;
    virtual UBool Resize(int32_t appendCapacity, int32_t length) = 0;

    void SetNotOk() {
        buffer_ = NULL;
        capacity_ = 0;
    }

    char *buffer_;
    int32_t capacity_;
    int32_t appended_;
    int32_t ignore_;

private:
    SortKeyByteSink(const SortKeyByteSink &); // copy constructor not implemented
    SortKeyByteSink &operator=(const SortKeyByteSink &); // assignment operator not implemented
};

class U_I18N_API CollationKeys /* not : public UObject because all methods are static */ {
public:
    class LevelCallback : public UMemory {
    public:
        virtual ~LevelCallback();
        /**
         * @param level The next level about to be written to the ByteSink.
         * @return TRUE if the level is to be written
         *         (the base class implementation always returns TRUE)
         */
        virtual UBool needToWrite(Collation::Level level);
    };

    /**
     * Writes the sort key bytes for minLevel up to the iterator data's strength.
     * Optionally writes the case level.
     * Stops writing levels when callback.needToWrite(level) returns FALSE.
     * Separates levels with the LEVEL_SEPARATOR_BYTE
     * but does not write a TERMINATOR_BYTE.
     */
    static void writeSortKeyUpToQuaternary(CollationIterator &iter,
                                           const UBool *compressibleBytes,
                                           const CollationSettings &settings,
                                           SortKeyByteSink &sink,
                                           Collation::Level minLevel, LevelCallback &callback,
                                           UBool preflight, UErrorCode &errorCode);
private:
    friend struct CollationDataReader;

    CollationKeys();  // no instantiation

    // Secondary level: Compress up to 33 common weights as 05..25 or 25..45.
    static const uint32_t SEC_COMMON_LOW = Collation::COMMON_BYTE;
    static const uint32_t SEC_COMMON_MIDDLE = SEC_COMMON_LOW + 0x20;
    static const uint32_t SEC_COMMON_HIGH = SEC_COMMON_LOW + 0x40;
    static const int32_t SEC_COMMON_MAX_COUNT = 0x21;

    // Case level, lowerFirst: Compress up to 7 common weights as 1..7 or 7..13.
    static const uint32_t CASE_LOWER_FIRST_COMMON_LOW = 1;
    static const uint32_t CASE_LOWER_FIRST_COMMON_MIDDLE = 7;
    static const uint32_t CASE_LOWER_FIRST_COMMON_HIGH = 13;
    static const int32_t CASE_LOWER_FIRST_COMMON_MAX_COUNT = 7;

    // Case level, upperFirst: Compress up to 13 common weights as 3..15.
    static const uint32_t CASE_UPPER_FIRST_COMMON_LOW = 3;
    static const uint32_t CASE_UPPER_FIRST_COMMON_HIGH = 15;
    static const int32_t CASE_UPPER_FIRST_COMMON_MAX_COUNT = 13;

    // Tertiary level only (no case): Compress up to 97 common weights as 05..65 or 65..C5.
    static const uint32_t TER_ONLY_COMMON_LOW = Collation::COMMON_BYTE;
    static const uint32_t TER_ONLY_COMMON_MIDDLE = TER_ONLY_COMMON_LOW + 0x60;
    static const uint32_t TER_ONLY_COMMON_HIGH = TER_ONLY_COMMON_LOW + 0xc0;
    static const int32_t TER_ONLY_COMMON_MAX_COUNT = 0x61;

    // Tertiary with case, lowerFirst: Compress up to 33 common weights as 05..25 or 25..45.
    static const uint32_t TER_LOWER_FIRST_COMMON_LOW = Collation::COMMON_BYTE;
    static const uint32_t TER_LOWER_FIRST_COMMON_MIDDLE = TER_LOWER_FIRST_COMMON_LOW + 0x20;
    static const uint32_t TER_LOWER_FIRST_COMMON_HIGH = TER_LOWER_FIRST_COMMON_LOW + 0x40;
    static const int32_t TER_LOWER_FIRST_COMMON_MAX_COUNT = 0x21;

    // Tertiary with case, upperFirst: Compress up to 33 common weights as 85..A5 or A5..C5.
    static const uint32_t TER_UPPER_FIRST_COMMON_LOW = Collation::COMMON_BYTE + 0x80;
    static const uint32_t TER_UPPER_FIRST_COMMON_MIDDLE = TER_UPPER_FIRST_COMMON_LOW + 0x20;
    static const uint32_t TER_UPPER_FIRST_COMMON_HIGH = TER_UPPER_FIRST_COMMON_LOW + 0x40;
    static const int32_t TER_UPPER_FIRST_COMMON_MAX_COUNT = 0x21;

    // Quaternary level: Compress up to 113 common weights as 1C..8C or 8C..FC.
    static const uint32_t QUAT_COMMON_LOW = 0x1c;
    static const uint32_t QUAT_COMMON_MIDDLE = QUAT_COMMON_LOW + 0x70;
    static const uint32_t QUAT_COMMON_HIGH = QUAT_COMMON_LOW + 0xE0;
    static const int32_t QUAT_COMMON_MAX_COUNT = 0x71;
    // Primary weights shifted to quaternary level must be encoded with
    // a lead byte below the common-weight compression range.
    static const uint32_t QUAT_SHIFTED_LIMIT_BYTE = QUAT_COMMON_LOW - 1;  // 0x1b
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
// collunsafe.h
//  Copyright (C) 2015-2016, International Business Machines Corporation and others.
//  All Rights Reserved.

// To be included by collationdatareader.cpp, and generated by gencolusb.
// Machine generated, do not edit.




#define COLLUNSAFE_ICU_VERSION "56.0.1"
#define COLLUNSAFE_COLL_VERSION "9.64"
#define COLLUNSAFE_SERIALIZE 1
static const int32_t unsafe_serializedCount = 850;
static const uint16_t unsafe_serializedData[850] = { 
0x8350, 0x01B8, 0x0034, 0x0035, 0x004C, 0x004D, 0x00A0, 0x00A1,  // 8
0x0300, 0x034F, 0x0350, 0x0370, 0x03A9, 0x03AA, 0x03E2, 0x03E3,  // 16
0x042F, 0x0430, 0x0483, 0x0488, 0x0531, 0x0532, 0x0591, 0x05BE,  // 24
0x05BF, 0x05C0, 0x05C1, 0x05C3, 0x05C4, 0x05C6, 0x05C7, 0x05C8,  // 32
0x05D0, 0x05D1, 0x0610, 0x061B, 0x0628, 0x0629, 0x064B, 0x0660,  // 40
0x0670, 0x0671, 0x06D6, 0x06DD, 0x06DF, 0x06E5, 0x06E7, 0x06E9,  // 48
0x06EA, 0x06EE, 0x0710, 0x0712, 0x0730, 0x074B, 0x078C, 0x078D,  // 56
0x07D8, 0x07D9, 0x07EB, 0x07F4, 0x0800, 0x0801, 0x0816, 0x081A,  // 64
0x081B, 0x0824, 0x0825, 0x0828, 0x0829, 0x082E, 0x0840, 0x0841,  // 72
0x0859, 0x085C, 0x08E3, 0x0900, 0x0905, 0x0906, 0x093C, 0x093D,  // 80
0x094D, 0x094E, 0x0951, 0x0955, 0x0995, 0x0996, 0x09BC, 0x09BD,  // 88
0x09BE, 0x09BF, 0x09CD, 0x09CE, 0x09D7, 0x09D8, 0x0A15, 0x0A16,  // 96
0x0A3C, 0x0A3D, 0x0A4D, 0x0A4E, 0x0A95, 0x0A96, 0x0ABC, 0x0ABD,  // 104
0x0ACD, 0x0ACE, 0x0B15, 0x0B16, 0x0B3C, 0x0B3D, 0x0B3E, 0x0B3F,  // 112
0x0B4D, 0x0B4E, 0x0B56, 0x0B58, 0x0B95, 0x0B96, 0x0BBE, 0x0BBF,  // 120
0x0BCD, 0x0BCE, 0x0BD7, 0x0BD8, 0x0C15, 0x0C16, 0x0C4D, 0x0C4E,  // 128
0x0C55, 0x0C57, 0x0C95, 0x0C96, 0x0CBC, 0x0CBD, 0x0CC2, 0x0CC3,  // 136
0x0CCD, 0x0CCE, 0x0CD5, 0x0CD7, 0x0D15, 0x0D16, 0x0D3E, 0x0D3F,  // 144
0x0D4D, 0x0D4E, 0x0D57, 0x0D58, 0x0D85, 0x0D86, 0x0DCA, 0x0DCB,  // 152
0x0DCF, 0x0DD0, 0x0DDF, 0x0DE0, 0x0E01, 0x0E2F, 0x0E32, 0x0E33,  // 160
0x0E38, 0x0E3B, 0x0E48, 0x0E4C, 0x0E81, 0x0E83, 0x0E84, 0x0E85,  // 168
0x0E87, 0x0E89, 0x0E8A, 0x0E8B, 0x0E8D, 0x0E8E, 0x0E94, 0x0E98,  // 176
0x0E99, 0x0EA0, 0x0EA1, 0x0EA4, 0x0EA5, 0x0EA6, 0x0EA7, 0x0EA8,  // 184
0x0EAA, 0x0EAC, 0x0EAD, 0x0EAF, 0x0EB2, 0x0EB3, 0x0EB8, 0x0EBA,  // 192
0x0EC8, 0x0ECC, 0x0EDC, 0x0EE0, 0x0F18, 0x0F1A, 0x0F35, 0x0F36,  // 200
0x0F37, 0x0F38, 0x0F39, 0x0F3A, 0x0F40, 0x0F41, 0x0F71, 0x0F76,  // 208
0x0F7A, 0x0F7E, 0x0F80, 0x0F85, 0x0F86, 0x0F88, 0x0FC6, 0x0FC7,  // 216
0x1000, 0x1001, 0x102E, 0x102F, 0x1037, 0x1038, 0x1039, 0x103B,  // 224
0x108D, 0x108E, 0x10D3, 0x10D4, 0x12A0, 0x12A1, 0x135D, 0x1360,  // 232
0x13C4, 0x13C5, 0x14C0, 0x14C1, 0x168F, 0x1690, 0x16A0, 0x16A1,  // 240
0x1703, 0x1704, 0x1714, 0x1715, 0x1723, 0x1724, 0x1734, 0x1735,  // 248
0x1743, 0x1744, 0x1763, 0x1764, 0x1780, 0x1781, 0x17D2, 0x17D3,  // 256
0x17DD, 0x17DE, 0x1826, 0x1827, 0x18A9, 0x18AA, 0x1900, 0x1901,  // 264
0x1939, 0x193C, 0x1950, 0x1951, 0x1980, 0x19AC, 0x1A00, 0x1A01,  // 272
0x1A17, 0x1A19, 0x1A20, 0x1A21, 0x1A60, 0x1A61, 0x1A75, 0x1A7D,  // 280
0x1A7F, 0x1A80, 0x1AB0, 0x1ABE, 0x1B05, 0x1B06, 0x1B34, 0x1B36,  // 288
0x1B44, 0x1B45, 0x1B6B, 0x1B74, 0x1B83, 0x1B84, 0x1BAA, 0x1BAC,  // 296
0x1BC0, 0x1BC1, 0x1BE6, 0x1BE7, 0x1BF2, 0x1BF4, 0x1C00, 0x1C01,  // 304
0x1C37, 0x1C38, 0x1C5A, 0x1C5B, 0x1CD0, 0x1CD3, 0x1CD4, 0x1CE1,  // 312
0x1CE2, 0x1CE9, 0x1CED, 0x1CEE, 0x1CF4, 0x1CF5, 0x1CF8, 0x1CFA,  // 320
0x1DC0, 0x1DF6, 0x1DFC, 0x1E00, 0x201C, 0x201D, 0x20AC, 0x20AD,  // 328
0x20D0, 0x20DD, 0x20E1, 0x20E2, 0x20E5, 0x20F1, 0x263A, 0x263B,  // 336
0x2C00, 0x2C01, 0x2CEF, 0x2CF2, 0x2D5E, 0x2D5F, 0x2D7F, 0x2D80,  // 344
0x2DE0, 0x2E00, 0x302A, 0x3030, 0x304B, 0x304C, 0x3099, 0x309B,  // 352
0x30AB, 0x30AC, 0x3105, 0x3106, 0x5B57, 0x5B58, 0xA288, 0xA289,  // 360
0xA4E8, 0xA4E9, 0xA549, 0xA54A, 0xA66F, 0xA670, 0xA674, 0xA67E,  // 368
0xA69E, 0xA6A1, 0xA6F0, 0xA6F2, 0xA800, 0xA801, 0xA806, 0xA807,  // 376
0xA840, 0xA841, 0xA882, 0xA883, 0xA8C4, 0xA8C5, 0xA8E0, 0xA8F2,  // 384
0xA90A, 0xA90B, 0xA92B, 0xA92E, 0xA930, 0xA931, 0xA953, 0xA954,  // 392
0xA984, 0xA985, 0xA9B3, 0xA9B4, 0xA9C0, 0xA9C1, 0xAA00, 0xAA01,  // 400
0xAA80, 0xAAB1, 0xAAB2, 0xAAB5, 0xAAB7, 0xAAB9, 0xAABE, 0xAAC0,  // 408
0xAAC1, 0xAAC2, 0xAAF6, 0xAAF7, 0xABC0, 0xABC1, 0xABED, 0xABEE,  // 416
0xAC00, 0xAC01, 0xD800, 0xD807, 0xD808, 0xD809, 0xD80C, 0xD80D,  // 424
0xD811, 0xD812, 0xD81A, 0xD81C, 0xD82F, 0xD830, 0xD834, 0xD835,  // 432
0xD83A, 0xD83B, 0xDC00, 0xE000, 0xFB1E, 0xFB1F, 0xFDD0, 0xFDD1,  // 440
0xFE20, 0xFE30, 0x0001, 0x0000, 0x0001, 0x0001, 0x0001, 0x01FD,  // 448
0x0001, 0x01FE, 0x0001, 0x0280, 0x0001, 0x0281, 0x0001, 0x02B7,  // 456
0x0001, 0x02B8, 0x0001, 0x02E0, 0x0001, 0x02E1, 0x0001, 0x0308,  // 464
0x0001, 0x0309, 0x0001, 0x0330, 0x0001, 0x0331, 0x0001, 0x036B,  // 472
0x0001, 0x036C, 0x0001, 0x0376, 0x0001, 0x037B, 0x0001, 0x0380,  // 480
0x0001, 0x0381, 0x0001, 0x03A0, 0x0001, 0x03A1, 0x0001, 0x0414,  // 488
0x0001, 0x0415, 0x0001, 0x0450, 0x0001, 0x0451, 0x0001, 0x0480,  // 496
0x0001, 0x0481, 0x0001, 0x0500, 0x0001, 0x0501, 0x0001, 0x0537,  // 504
0x0001, 0x0538, 0x0001, 0x0647, 0x0001, 0x0648, 0x0001, 0x0800,  // 512
0x0001, 0x0801, 0x0001, 0x0840, 0x0001, 0x0841, 0x0001, 0x0873,  // 520
0x0001, 0x0874, 0x0001, 0x0896, 0x0001, 0x0897, 0x0001, 0x08F4,  // 528
0x0001, 0x08F5, 0x0001, 0x0900, 0x0001, 0x0901, 0x0001, 0x0920,  // 536
0x0001, 0x0921, 0x0001, 0x0980, 0x0001, 0x0981, 0x0001, 0x09A0,  // 544
0x0001, 0x09A1, 0x0001, 0x0A00, 0x0001, 0x0A01, 0x0001, 0x0A0D,  // 552
0x0001, 0x0A0E, 0x0001, 0x0A0F, 0x0001, 0x0A10, 0x0001, 0x0A38,  // 560
0x0001, 0x0A3B, 0x0001, 0x0A3F, 0x0001, 0x0A40, 0x0001, 0x0A60,  // 568
0x0001, 0x0A61, 0x0001, 0x0A95, 0x0001, 0x0A96, 0x0001, 0x0AC1,  // 576
0x0001, 0x0AC2, 0x0001, 0x0AE5, 0x0001, 0x0AE7, 0x0001, 0x0B00,  // 584
0x0001, 0x0B01, 0x0001, 0x0B40, 0x0001, 0x0B41, 0x0001, 0x0B60,  // 592
0x0001, 0x0B61, 0x0001, 0x0B8F, 0x0001, 0x0B90, 0x0001, 0x0C00,  // 600
0x0001, 0x0C01, 0x0001, 0x0CA1, 0x0001, 0x0CA2, 0x0001, 0x1005,  // 608
0x0001, 0x1006, 0x0001, 0x1046, 0x0001, 0x1047, 0x0001, 0x107F,  // 616
0x0001, 0x1080, 0x0001, 0x1083, 0x0001, 0x1084, 0x0001, 0x10B9,  // 624
0x0001, 0x10BB, 0x0001, 0x10D0, 0x0001, 0x10D1, 0x0001, 0x1100,  // 632
0x0001, 0x1104, 0x0001, 0x1127, 0x0001, 0x1128, 0x0001, 0x1133,  // 640
0x0001, 0x1135, 0x0001, 0x1152, 0x0001, 0x1153, 0x0001, 0x1173,  // 648
0x0001, 0x1174, 0x0001, 0x1183, 0x0001, 0x1184, 0x0001, 0x11C0,  // 656
0x0001, 0x11C1, 0x0001, 0x11CA, 0x0001, 0x11CB, 0x0001, 0x1208,  // 664
0x0001, 0x1209, 0x0001, 0x1235, 0x0001, 0x1237, 0x0001, 0x128F,  // 672
0x0001, 0x1290, 0x0001, 0x12BE, 0x0001, 0x12BF, 0x0001, 0x12E9,  // 680
0x0001, 0x12EB, 0x0001, 0x1315, 0x0001, 0x1316, 0x0001, 0x133C,  // 688
0x0001, 0x133D, 0x0001, 0x133E, 0x0001, 0x133F, 0x0001, 0x134D,  // 696
0x0001, 0x134E, 0x0001, 0x1357, 0x0001, 0x1358, 0x0001, 0x1366,  // 704
0x0001, 0x136D, 0x0001, 0x1370, 0x0001, 0x1375, 0x0001, 0x1484,  // 712
0x0001, 0x1485, 0x0001, 0x14B0, 0x0001, 0x14B1, 0x0001, 0x14BA,  // 720
0x0001, 0x14BB, 0x0001, 0x14BD, 0x0001, 0x14BE, 0x0001, 0x14C2,  // 728
0x0001, 0x14C4, 0x0001, 0x158E, 0x0001, 0x158F, 0x0001, 0x15AF,  // 736
0x0001, 0x15B0, 0x0001, 0x15BF, 0x0001, 0x15C1, 0x0001, 0x160E,  // 744
0x0001, 0x160F, 0x0001, 0x163F, 0x0001, 0x1640, 0x0001, 0x1680,  // 752
0x0001, 0x1681, 0x0001, 0x16B6, 0x0001, 0x16B8, 0x0001, 0x1717,  // 760
0x0001, 0x1718, 0x0001, 0x172B, 0x0001, 0x172C, 0x0001, 0x18B4,  // 768
0x0001, 0x18B5, 0x0001, 0x1AC0, 0x0001, 0x1AC1, 0x0001, 0x2000,  // 776
0x0001, 0x2001, 0x0001, 0x3153, 0x0001, 0x3154, 0x0001, 0x4400,  // 784
0x0001, 0x4401, 0x0001, 0x6A4F, 0x0001, 0x6A50, 0x0001, 0x6AE6,  // 792
0x0001, 0x6AE7, 0x0001, 0x6AF0, 0x0001, 0x6AF5, 0x0001, 0x6B1C,  // 800
0x0001, 0x6B1D, 0x0001, 0x6B30, 0x0001, 0x6B37, 0x0001, 0x6F00,  // 808
0x0001, 0x6F01, 0x0001, 0xBC20, 0x0001, 0xBC21, 0x0001, 0xBC9E,  // 816
0x0001, 0xBC9F, 0x0001, 0xD165, 0x0001, 0xD16A, 0x0001, 0xD16D,  // 824
0x0001, 0xD173, 0x0001, 0xD17B, 0x0001, 0xD183, 0x0001, 0xD185,  // 832
0x0001, 0xD18C, 0x0001, 0xD1AA, 0x0001, 0xD1AE, 0x0001, 0xD242,  // 840
0x0001, 0xD245, 0x0001, 0xE802, 0x0001, 0xE803, 0x0001, 0xE8D0,  // 848
0x0001, 0xE8D7};






U_NAMESPACE_BEGIN

namespace {

int32_t getIndex(const int32_t *indexes, int32_t length, int32_t i) {
    return (i < length) ? indexes[i] : -1;
}

}  // namespace

void
CollationDataReader::read(const CollationTailoring *base, const uint8_t *inBytes, int32_t inLength,
                          CollationTailoring &tailoring, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(base != NULL) {
        if(inBytes == NULL || (0 <= inLength && inLength < 24)) {
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        const DataHeader *header = reinterpret_cast<const DataHeader *>(inBytes);
        if(!(header->dataHeader.magic1 == 0xda && header->dataHeader.magic2 == 0x27 &&
                isAcceptable(tailoring.version, NULL, NULL, &header->info))) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        if(base->getUCAVersion() != tailoring.getUCAVersion()) {
            errorCode = U_COLLATOR_VERSION_MISMATCH;
            return;
        }
        int32_t headerLength = header->dataHeader.headerSize;
        inBytes += headerLength;
        if(inLength >= 0) {
            inLength -= headerLength;
        }
    }

    if(inBytes == NULL || (0 <= inLength && inLength < 8)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    const int32_t *inIndexes = reinterpret_cast<const int32_t *>(inBytes);
    int32_t indexesLength = inIndexes[IX_INDEXES_LENGTH];
    if(indexesLength < 2 || (0 <= inLength && inLength < indexesLength * 4)) {
        errorCode = U_INVALID_FORMAT_ERROR;  // Not enough indexes.
        return;
    }

    // Assume that the tailoring data is in initial state,
    // with NULL pointers and 0 lengths.

    // Set pointers to non-empty data parts.
    // Do this in order of their byte offsets. (Should help porting to Java.)

    int32_t index;  // one of the indexes[] slots
    int32_t offset;  // byte offset for the index part
    int32_t length;  // number of bytes in the index part

    if(indexesLength > IX_TOTAL_SIZE) {
        length = inIndexes[IX_TOTAL_SIZE];
    } else if(indexesLength > IX_REORDER_CODES_OFFSET) {
        length = inIndexes[indexesLength - 1];
    } else {
        length = 0;  // only indexes, and inLength was already checked for them
    }
    if(0 <= inLength && inLength < length) {
        errorCode = U_INVALID_FORMAT_ERROR;
        return;
    }

    const CollationData *baseData = base == NULL ? NULL : base->data;
    const int32_t *reorderCodes = NULL;
    int32_t reorderCodesLength = 0;
    const uint32_t *reorderRanges = NULL;
    int32_t reorderRangesLength = 0;
    index = IX_REORDER_CODES_OFFSET;
    offset = getIndex(inIndexes, indexesLength, index);
    length = getIndex(inIndexes, indexesLength, index + 1) - offset;
    if(length >= 4) {
        if(baseData == NULL) {
            // We assume for collation settings that
            // the base data does not have a reordering.
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        reorderCodes = reinterpret_cast<const int32_t *>(inBytes + offset);
        reorderCodesLength = length / 4;

        // The reorderRanges (if any) are the trailing reorderCodes entries.
        // Split the array at the boundary.
        // Script or reorder codes do not exceed 16-bit values.
        // Range limits are stored in the upper 16 bits, and are never 0.
        while(reorderRangesLength < reorderCodesLength &&
                (reorderCodes[reorderCodesLength - reorderRangesLength - 1] & 0xffff0000) != 0) {
            ++reorderRangesLength;
        }
        U_ASSERT(reorderRangesLength < reorderCodesLength);
        if(reorderRangesLength != 0) {
            reorderCodesLength -= reorderRangesLength;
            reorderRanges = reinterpret_cast<const uint32_t *>(reorderCodes + reorderCodesLength);
        }
    }

    // There should be a reorder table only if there are reorder codes.
    // However, when there are reorder codes the reorder table may be omitted to reduce
    // the data size.
    const uint8_t *reorderTable = NULL;
    index = IX_REORDER_TABLE_OFFSET;
    offset = getIndex(inIndexes, indexesLength, index);
    length = getIndex(inIndexes, indexesLength, index + 1) - offset;
    if(length >= 256) {
        if(reorderCodesLength == 0) {
            errorCode = U_INVALID_FORMAT_ERROR;  // Reordering table without reordering codes.
            return;
        }
        reorderTable = inBytes + offset;
    } else {
        // If we have reorder codes, then build the reorderTable at the end,
        // when the CollationData is otherwise complete.
    }

    if(baseData != NULL && baseData->numericPrimary != (inIndexes[IX_OPTIONS] & 0xff000000)) {
        errorCode = U_INVALID_FORMAT_ERROR;
        return;
    }
    CollationData *data = NULL;  // Remains NULL if there are no mappings.

    index = IX_TRIE_OFFSET;
    offset = getIndex(inIndexes, indexesLength, index);
    length = getIndex(inIndexes, indexesLength, index + 1) - offset;
    if(length >= 8) {
        if(!tailoring.ensureOwnedData(errorCode)) { return; }
        data = tailoring.ownedData;
        data->base = baseData;
        data->numericPrimary = inIndexes[IX_OPTIONS] & 0xff000000;
        data->trie = tailoring.trie = utrie2_openFromSerialized(
            UTRIE2_32_VALUE_BITS, inBytes + offset, length, NULL,
            &errorCode);
        if(U_FAILURE(errorCode)) { return; }
    } else if(baseData != NULL) {
        // Use the base data. Only the settings are tailored.
        tailoring.data = baseData;
    } else {
        errorCode = U_INVALID_FORMAT_ERROR;  // No mappings.
        return;
    }

    index = IX_CES_OFFSET;
    offset = getIndex(inIndexes, indexesLength, index);
    length = getIndex(inIndexes, indexesLength, index + 1) - offset;
    if(length >= 8) {
        if(data == NULL) {
            errorCode = U_INVALID_FORMAT_ERROR;  // Tailored ces without tailored trie.
            return;
        }
        data->ces = reinterpret_cast<const int64_t *>(inBytes + offset);
        data->cesLength = length / 8;
    }

    index = IX_CE32S_OFFSET;
    offset = getIndex(inIndexes, indexesLength, index);
    length = getIndex(inIndexes, indexesLength, index + 1) - offset;
    if(length >= 4) {
        if(data == NULL) {
            errorCode = U_INVALID_FORMAT_ERROR;  // Tailored ce32s without tailored trie.
            return;
        }
        data->ce32s = reinterpret_cast<const uint32_t *>(inBytes + offset);
        data->ce32sLength = length / 4;
    }

    int32_t jamoCE32sStart = getIndex(inIndexes, indexesLength, IX_JAMO_CE32S_START);
    if(jamoCE32sStart >= 0) {
        if(data == NULL || data->ce32s == NULL) {
            errorCode = U_INVALID_FORMAT_ERROR;  // Index into non-existent ce32s[].
            return;
        }
        data->jamoCE32s = data->ce32s + jamoCE32sStart;
    } else if(data == NULL) {
        // Nothing to do.
    } else if(baseData != NULL) {
        data->jamoCE32s = baseData->jamoCE32s;
    } else {
        errorCode = U_INVALID_FORMAT_ERROR;  // No Jamo CE32s for Hangul processing.
        return;
    }

    index = IX_ROOT_ELEMENTS_OFFSET;
    offset = getIndex(inIndexes, indexesLength, index);
    length = getIndex(inIndexes, indexesLength, index + 1) - offset;
    if(length >= 4) {
        length /= 4;
        if(data == NULL || length <= CollationRootElements::IX_SEC_TER_BOUNDARIES) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        data->rootElements = reinterpret_cast<const uint32_t *>(inBytes + offset);
        data->rootElementsLength = length;
        uint32_t commonSecTer = data->rootElements[CollationRootElements::IX_COMMON_SEC_AND_TER_CE];
        if(commonSecTer != Collation::COMMON_SEC_AND_TER_CE) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        uint32_t secTerBoundaries = data->rootElements[CollationRootElements::IX_SEC_TER_BOUNDARIES];
        if((secTerBoundaries >> 24) < CollationKeys::SEC_COMMON_HIGH) {
            // [fixed last secondary common byte] is too low,
            // and secondary weights would collide with compressed common secondaries.
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
    }

    index = IX_CONTEXTS_OFFSET;
    offset = getIndex(inIndexes, indexesLength, index);
    length = getIndex(inIndexes, indexesLength, index + 1) - offset;
    if(length >= 2) {
        if(data == NULL) {
            errorCode = U_INVALID_FORMAT_ERROR;  // Tailored contexts without tailored trie.
            return;
        }
        data->contexts = reinterpret_cast<const UChar *>(inBytes + offset);
        data->contextsLength = length / 2;
    }

    index = IX_UNSAFE_BWD_OFFSET;
    offset = getIndex(inIndexes, indexesLength, index);
    length = getIndex(inIndexes, indexesLength, index + 1) - offset;
    if(length >= 2) {
        if(data == NULL) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        if(baseData == NULL) {
#if defined(COLLUNSAFE_COLL_VERSION) && defined (COLLUNSAFE_SERIALIZE)
          tailoring.unsafeBackwardSet = new UnicodeSet(unsafe_serializedData, unsafe_serializedCount, UnicodeSet::kSerialized, errorCode);
          if(tailoring.unsafeBackwardSet == NULL) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
          } else if (U_FAILURE(errorCode)) {
            return;
          }
#else
            // Create the unsafe-backward set for the root collator.
            // Include all non-zero combining marks and trail surrogates.
            // We do this at load time, rather than at build time,
            // to simplify Unicode version bootstrapping:
            // The root data builder only needs the new FractionalUCA.txt data,
            // but it need not be built with a version of ICU already updated to
            // the corresponding new Unicode Character Database.
            //
            // The following is an optimized version of
            // new UnicodeSet("[[:^lccc=0:][\\udc00-\\udfff]]").
            // It is faster and requires fewer code dependencies.
            tailoring.unsafeBackwardSet = new UnicodeSet(0xdc00, 0xdfff);  // trail surrogates
            if(tailoring.unsafeBackwardSet == NULL) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            data->nfcImpl.addLcccChars(*tailoring.unsafeBackwardSet);
#endif // !COLLUNSAFE_SERIALIZE || !COLLUNSAFE_COLL_VERSION
        } else {
            // Clone the root collator's set contents.
            tailoring.unsafeBackwardSet = static_cast<UnicodeSet *>(
                baseData->unsafeBackwardSet->cloneAsThawed());
            if(tailoring.unsafeBackwardSet == NULL) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
        }
        // Add the ranges from the data file to the unsafe-backward set.
        USerializedSet sset;
        const uint16_t *unsafeData = reinterpret_cast<const uint16_t *>(inBytes + offset);
        if(!uset_getSerializedSet(&sset, unsafeData, length / 2)) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        int32_t count = uset_getSerializedRangeCount(&sset);
        for(int32_t i = 0; i < count; ++i) {
            UChar32 start, end;
            uset_getSerializedRange(&sset, i, &start, &end);
            tailoring.unsafeBackwardSet->add(start, end);
        }
        // Mark each lead surrogate as "unsafe"
        // if any of its 1024 associated supplementary code points is "unsafe".
        UChar32 c = 0x10000;
        for(UChar lead = 0xd800; lead < 0xdc00; ++lead, c += 0x400) {
            if(!tailoring.unsafeBackwardSet->containsNone(c, c + 0x3ff)) {
                tailoring.unsafeBackwardSet->add(lead);
            }
        }
        tailoring.unsafeBackwardSet->freeze();
        data->unsafeBackwardSet = tailoring.unsafeBackwardSet;
    } else if(data == NULL) {
        // Nothing to do.
    } else if(baseData != NULL) {
        // No tailoring-specific data: Alias the root collator's set.
        data->unsafeBackwardSet = baseData->unsafeBackwardSet;
    } else {
        errorCode = U_INVALID_FORMAT_ERROR;  // No unsafeBackwardSet.
        return;
    }

    // If the fast Latin format version is different,
    // or the version is set to 0 for "no fast Latin table",
    // then just always use the normal string comparison path.
    if(data != NULL) {
        data->fastLatinTable = NULL;
        data->fastLatinTableLength = 0;
        if(((inIndexes[IX_OPTIONS] >> 16) & 0xff) == CollationFastLatin::VERSION) {
            index = IX_FAST_LATIN_TABLE_OFFSET;
            offset = getIndex(inIndexes, indexesLength, index);
            length = getIndex(inIndexes, indexesLength, index + 1) - offset;
            if(length >= 2) {
                data->fastLatinTable = reinterpret_cast<const uint16_t *>(inBytes + offset);
                data->fastLatinTableLength = length / 2;
                if((*data->fastLatinTable >> 8) != CollationFastLatin::VERSION) {
                    errorCode = U_INVALID_FORMAT_ERROR;  // header vs. table version mismatch
                    return;
                }
            } else if(baseData != NULL) {
                data->fastLatinTable = baseData->fastLatinTable;
                data->fastLatinTableLength = baseData->fastLatinTableLength;
            }
        }
    }

    index = IX_SCRIPTS_OFFSET;
    offset = getIndex(inIndexes, indexesLength, index);
    length = getIndex(inIndexes, indexesLength, index + 1) - offset;
    if(length >= 2) {
        if(data == NULL) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        const uint16_t *scripts = reinterpret_cast<const uint16_t *>(inBytes + offset);
        int32_t scriptsLength = length / 2;
        data->numScripts = scripts[0];
        // There must be enough entries for both arrays, including more than two range starts.
        data->scriptStartsLength = scriptsLength - (1 + data->numScripts + 16);
        if(data->scriptStartsLength <= 2 ||
                CollationData::MAX_NUM_SCRIPT_RANGES < data->scriptStartsLength) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        data->scriptsIndex = scripts + 1;
        data->scriptStarts = scripts + 1 + data->numScripts + 16;
        if(!(data->scriptStarts[0] == 0 &&
                data->scriptStarts[1] == ((Collation::MERGE_SEPARATOR_BYTE + 1) << 8) &&
                data->scriptStarts[data->scriptStartsLength - 1] ==
                        (Collation::TRAIL_WEIGHT_BYTE << 8))) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
    } else if(data == NULL) {
        // Nothing to do.
    } else if(baseData != NULL) {
        data->numScripts = baseData->numScripts;
        data->scriptsIndex = baseData->scriptsIndex;
        data->scriptStarts = baseData->scriptStarts;
        data->scriptStartsLength = baseData->scriptStartsLength;
    }

    index = IX_COMPRESSIBLE_BYTES_OFFSET;
    offset = getIndex(inIndexes, indexesLength, index);
    length = getIndex(inIndexes, indexesLength, index + 1) - offset;
    if(length >= 256) {
        if(data == NULL) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        data->compressibleBytes = reinterpret_cast<const UBool *>(inBytes + offset);
    } else if(data == NULL) {
        // Nothing to do.
    } else if(baseData != NULL) {
        data->compressibleBytes = baseData->compressibleBytes;
    } else {
        errorCode = U_INVALID_FORMAT_ERROR;  // No compressibleBytes[].
        return;
    }

    const CollationSettings &ts = *tailoring.settings;
    int32_t options = inIndexes[IX_OPTIONS] & 0xffff;
    uint16_t fastLatinPrimaries[CollationFastLatin::LATIN_LIMIT];
    int32_t fastLatinOptions = CollationFastLatin::getOptions(
            tailoring.data, ts, fastLatinPrimaries, UPRV_LENGTHOF(fastLatinPrimaries));
    if(options == ts.options && ts.variableTop != 0 &&
            reorderCodesLength == ts.reorderCodesLength &&
            (reorderCodesLength == 0 ||
                uprv_memcmp(reorderCodes, ts.reorderCodes, reorderCodesLength * 4) == 0) &&
            fastLatinOptions == ts.fastLatinOptions &&
            (fastLatinOptions < 0 ||
                uprv_memcmp(fastLatinPrimaries, ts.fastLatinPrimaries,
                            sizeof(fastLatinPrimaries)) == 0)) {
        return;
    }

    CollationSettings *settings = SharedObject::copyOnWrite(tailoring.settings);
    if(settings == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    settings->options = options;
    // Set variableTop from options and scripts data.
    settings->variableTop = tailoring.data->getLastPrimaryForGroup(
            UCOL_REORDER_CODE_FIRST + settings->getMaxVariable());
    if(settings->variableTop == 0) {
        errorCode = U_INVALID_FORMAT_ERROR;
        return;
    }

    if(reorderCodesLength != 0) {
        settings->aliasReordering(*baseData, reorderCodes, reorderCodesLength,
                                  reorderRanges, reorderRangesLength,
                                  reorderTable, errorCode);
    }

    settings->fastLatinOptions = CollationFastLatin::getOptions(
        tailoring.data, *settings,
        settings->fastLatinPrimaries, UPRV_LENGTHOF(settings->fastLatinPrimaries));
}

UBool U_CALLCONV
CollationDataReader::isAcceptable(void *context,
                                  const char * /* type */, const char * /*name*/,
                                  const UDataInfo *pInfo) {
    if(
        pInfo->size >= 20 &&
        pInfo->isBigEndian == U_IS_BIG_ENDIAN &&
        pInfo->charsetFamily == U_CHARSET_FAMILY &&
        pInfo->dataFormat[0] == 0x55 &&  // dataFormat="UCol"
        pInfo->dataFormat[1] == 0x43 &&
        pInfo->dataFormat[2] == 0x6f &&
        pInfo->dataFormat[3] == 0x6c &&
        pInfo->formatVersion[0] == 5
    ) {
        UVersionInfo *version = static_cast<UVersionInfo *>(context);
        if(version != NULL) {
            uprv_memcpy(version, pInfo->dataVersion, 4);
        }
        return TRUE;
    } else {
        return FALSE;
    }
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationdatawriter.cpp
*
* created on: 2013aug06
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION








// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationdatawriter.h
*
* created on: 2013aug06
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION

U_NAMESPACE_BEGIN

struct CollationData;
struct CollationSettings;
struct CollationTailoring;

/**
 * Collation-related code for tools & demos.
 */
class U_I18N_API CollationDataWriter /* all static */ {
public:
    static int32_t writeBase(const CollationData &data, const CollationSettings &settings,
                             const void *rootElements, int32_t rootElementsLength,
                             int32_t indexes[], uint8_t *dest, int32_t capacity,
                             UErrorCode &errorCode);

    static int32_t writeTailoring(const CollationTailoring &t, const CollationSettings &settings,
                                  int32_t indexes[], uint8_t *dest, int32_t capacity,
                                  UErrorCode &errorCode);

private:
    CollationDataWriter();  // no constructor

    static int32_t write(UBool isBase, const UVersionInfo dataVersion,
                         const CollationData &data, const CollationSettings &settings,
                         const void *rootElements, int32_t rootElementsLength,
                         int32_t indexes[], uint8_t *dest, int32_t capacity,
                         UErrorCode &errorCode);

    static void copyData(const int32_t indexes[], int32_t startIndex,
                         const void *src, uint8_t *dest);
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION







U_NAMESPACE_BEGIN

uint8_t *
RuleBasedCollator::cloneRuleData(int32_t &length, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return NULL; }
    LocalMemory<uint8_t> buffer((uint8_t *)uprv_malloc(20000));
    if(buffer.isNull()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    length = cloneBinary(buffer.getAlias(), 20000, errorCode);
    if(errorCode == U_BUFFER_OVERFLOW_ERROR) {
        if(buffer.allocateInsteadAndCopy(length, 0) == NULL) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        errorCode = U_ZERO_ERROR;
        length = cloneBinary(buffer.getAlias(), length, errorCode);
    }
    if(U_FAILURE(errorCode)) { return NULL; }
    return buffer.orphan();
}

int32_t
RuleBasedCollator::cloneBinary(uint8_t *dest, int32_t capacity, UErrorCode &errorCode) const {
    int32_t indexes[CollationDataReader::IX_TOTAL_SIZE + 1];
    return CollationDataWriter::writeTailoring(
            *tailoring, *settings, indexes, dest, capacity,
            errorCode);
}

static const UDataInfo dataInfo = {
    sizeof(UDataInfo),
    0,

    U_IS_BIG_ENDIAN,
    U_CHARSET_FAMILY,
    U_SIZEOF_UCHAR,
    0,

    { 0x55, 0x43, 0x6f, 0x6c },         // dataFormat="UCol"
    { 5, 0, 0, 0 },                     // formatVersion
    { 6, 3, 0, 0 }                      // dataVersion
};

int32_t
CollationDataWriter::writeBase(const CollationData &data, const CollationSettings &settings,
                               const void *rootElements, int32_t rootElementsLength,
                               int32_t indexes[], uint8_t *dest, int32_t capacity,
                               UErrorCode &errorCode) {
    return write(TRUE, NULL,
                 data, settings,
                 rootElements, rootElementsLength,
                 indexes, dest, capacity, errorCode);
}

int32_t
CollationDataWriter::writeTailoring(const CollationTailoring &t, const CollationSettings &settings,
                                    int32_t indexes[], uint8_t *dest, int32_t capacity,
                                    UErrorCode &errorCode) {
    return write(FALSE, t.version,
                 *t.data, settings,
                 NULL, 0,
                 indexes, dest, capacity, errorCode);
}

int32_t
CollationDataWriter::write(UBool isBase, const UVersionInfo dataVersion,
                           const CollationData &data, const CollationSettings &settings,
                           const void *rootElements, int32_t rootElementsLength,
                           int32_t indexes[], uint8_t *dest, int32_t capacity,
                           UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    if(capacity < 0 || (capacity > 0 && dest == NULL)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    // Figure out which data items to write before settling on
    // the indexes length and writing offsets.
    // For any data item, we need to write the start and limit offsets,
    // so the indexes length must be at least index-of-start-offset + 2.
    int32_t indexesLength;
    UBool hasMappings;
    UnicodeSet unsafeBackwardSet;
    const CollationData *baseData = data.base;

    int32_t fastLatinVersion;
    if(data.fastLatinTable != NULL) {
        fastLatinVersion = (int32_t)CollationFastLatin::VERSION << 16;
    } else {
        fastLatinVersion = 0;
    }
    int32_t fastLatinTableLength = 0;

    if(isBase) {
        // For the root collator, we write an even number of indexes
        // so that we start with an 8-aligned offset.
        indexesLength = CollationDataReader::IX_TOTAL_SIZE + 1;
        U_ASSERT(settings.reorderCodesLength == 0);
        hasMappings = TRUE;
        unsafeBackwardSet = *data.unsafeBackwardSet;
        fastLatinTableLength = data.fastLatinTableLength;
    } else if(baseData == NULL) {
        hasMappings = FALSE;
        if(settings.reorderCodesLength == 0) {
            // only options
            indexesLength = CollationDataReader::IX_OPTIONS + 1;  // no limit offset here
        } else {
            // only options, reorder codes, and the reorder table
            indexesLength = CollationDataReader::IX_REORDER_TABLE_OFFSET + 2;
        }
    } else {
        hasMappings = TRUE;
        // Tailored mappings, and what else?
        // Check in ascending order of optional tailoring data items.
        indexesLength = CollationDataReader::IX_CE32S_OFFSET + 2;
        if(data.contextsLength != 0) {
            indexesLength = CollationDataReader::IX_CONTEXTS_OFFSET + 2;
        }
        unsafeBackwardSet.addAll(*data.unsafeBackwardSet).removeAll(*baseData->unsafeBackwardSet);
        if(!unsafeBackwardSet.isEmpty()) {
            indexesLength = CollationDataReader::IX_UNSAFE_BWD_OFFSET + 2;
        }
        if(data.fastLatinTable != baseData->fastLatinTable) {
            fastLatinTableLength = data.fastLatinTableLength;
            indexesLength = CollationDataReader::IX_FAST_LATIN_TABLE_OFFSET + 2;
        }
    }

    UVector32 codesAndRanges(errorCode);
    const int32_t *reorderCodes = settings.reorderCodes;
    int32_t reorderCodesLength = settings.reorderCodesLength;
    if(settings.hasReordering() &&
            CollationSettings::reorderTableHasSplitBytes(settings.reorderTable)) {
        // Rebuild the full list of reorder ranges.
        // The list in the settings is truncated for efficiency.
        data.makeReorderRanges(reorderCodes, reorderCodesLength, codesAndRanges, errorCode);
        // Write the codes, then the ranges.
        for(int32_t i = 0; i < reorderCodesLength; ++i) {
            codesAndRanges.insertElementAt(reorderCodes[i], i, errorCode);
        }
        if(U_FAILURE(errorCode)) { return 0; }
        reorderCodes = codesAndRanges.getBuffer();
        reorderCodesLength = codesAndRanges.size();
    }

    int32_t headerSize;
    if(isBase) {
        headerSize = 0;  // udata_create() writes the header
    } else {
        DataHeader header;
        header.dataHeader.magic1 = 0xda;
        header.dataHeader.magic2 = 0x27;
        uprv_memcpy(&header.info, &dataInfo, sizeof(UDataInfo));
        uprv_memcpy(header.info.dataVersion, dataVersion, sizeof(UVersionInfo));
        headerSize = (int32_t)sizeof(header);
        U_ASSERT((headerSize & 3) == 0);  // multiple of 4 bytes
        if(hasMappings && data.cesLength != 0) {
            // Sum of the sizes of the data items which are
            // not automatically multiples of 8 bytes and which are placed before the CEs.
            int32_t sum = headerSize + (indexesLength + reorderCodesLength) * 4;
            if((sum & 7) != 0) {
                // We need to add padding somewhere so that the 64-bit CEs are 8-aligned.
                // We add to the header size here.
                // Alternatively, we could increment the indexesLength
                // or add a few bytes to the reorderTable.
                headerSize += 4;
            }
        }
        header.dataHeader.headerSize = (uint16_t)headerSize;
        if(headerSize <= capacity) {
            uprv_memcpy(dest, &header, sizeof(header));
            // Write 00 bytes so that the padding is not mistaken for a copyright string.
            uprv_memset(dest + sizeof(header), 0, headerSize - (int32_t)sizeof(header));
            dest += headerSize;
            capacity -= headerSize;
        } else {
            dest = NULL;
            capacity = 0;
        }
    }

    indexes[CollationDataReader::IX_INDEXES_LENGTH] = indexesLength;
    U_ASSERT((settings.options & ~0xffff) == 0);
    indexes[CollationDataReader::IX_OPTIONS] =
            data.numericPrimary | fastLatinVersion | settings.options;
    indexes[CollationDataReader::IX_RESERVED2] = 0;
    indexes[CollationDataReader::IX_RESERVED3] = 0;

    // Byte offsets of data items all start from the start of the indexes.
    // We add the headerSize at the very end.
    int32_t totalSize = indexesLength * 4;

    if(hasMappings && (isBase || data.jamoCE32s != baseData->jamoCE32s)) {
        indexes[CollationDataReader::IX_JAMO_CE32S_START] = static_cast<int32_t>(data.jamoCE32s - data.ce32s);
    } else {
        indexes[CollationDataReader::IX_JAMO_CE32S_START] = -1;
    }

    indexes[CollationDataReader::IX_REORDER_CODES_OFFSET] = totalSize;
    totalSize += reorderCodesLength * 4;

    indexes[CollationDataReader::IX_REORDER_TABLE_OFFSET] = totalSize;
    if(settings.reorderTable != NULL) {
        totalSize += 256;
    }

    indexes[CollationDataReader::IX_TRIE_OFFSET] = totalSize;
    if(hasMappings) {
        UErrorCode errorCode2 = U_ZERO_ERROR;
        int32_t length;
        if(totalSize < capacity) {
            length = utrie2_serialize(data.trie, dest + totalSize,
                                      capacity - totalSize, &errorCode2);
        } else {
            length = utrie2_serialize(data.trie, NULL, 0, &errorCode2);
        }
        if(U_FAILURE(errorCode2) && errorCode2 != U_BUFFER_OVERFLOW_ERROR) {
            errorCode = errorCode2;
            return 0;
        }
        // The trie size should be a multiple of 8 bytes due to the way
        // compactIndex2(UNewTrie2 *trie) currently works.
        U_ASSERT((length & 7) == 0);
        totalSize += length;
    }

    indexes[CollationDataReader::IX_RESERVED8_OFFSET] = totalSize;
    indexes[CollationDataReader::IX_CES_OFFSET] = totalSize;
    if(hasMappings && data.cesLength != 0) {
        U_ASSERT(((headerSize + totalSize) & 7) == 0);
        totalSize += data.cesLength * 8;
    }

    indexes[CollationDataReader::IX_RESERVED10_OFFSET] = totalSize;
    indexes[CollationDataReader::IX_CE32S_OFFSET] = totalSize;
    if(hasMappings) {
        totalSize += data.ce32sLength * 4;
    }

    indexes[CollationDataReader::IX_ROOT_ELEMENTS_OFFSET] = totalSize;
    totalSize += rootElementsLength * 4;

    indexes[CollationDataReader::IX_CONTEXTS_OFFSET] = totalSize;
    if(hasMappings) {
        totalSize += data.contextsLength * 2;
    }

    indexes[CollationDataReader::IX_UNSAFE_BWD_OFFSET] = totalSize;
    if(hasMappings && !unsafeBackwardSet.isEmpty()) {
        UErrorCode errorCode2 = U_ZERO_ERROR;
        int32_t length;
        if(totalSize < capacity) {
            uint16_t *p = reinterpret_cast<uint16_t *>(dest + totalSize);
            length = unsafeBackwardSet.serialize(
                    p, (capacity - totalSize) / 2, errorCode2);
        } else {
            length = unsafeBackwardSet.serialize(NULL, 0, errorCode2);
        }
        if(U_FAILURE(errorCode2) && errorCode2 != U_BUFFER_OVERFLOW_ERROR) {
            errorCode = errorCode2;
            return 0;
        }
        totalSize += length * 2;
    }

    indexes[CollationDataReader::IX_FAST_LATIN_TABLE_OFFSET] = totalSize;
    totalSize += fastLatinTableLength * 2;

    UnicodeString scripts;
    indexes[CollationDataReader::IX_SCRIPTS_OFFSET] = totalSize;
    if(isBase) {
        scripts.append((UChar)data.numScripts);
        scripts.append(reinterpret_cast<const UChar *>(data.scriptsIndex), data.numScripts + 16);
        scripts.append(reinterpret_cast<const UChar *>(data.scriptStarts), data.scriptStartsLength);
        totalSize += scripts.length() * 2;
    }

    indexes[CollationDataReader::IX_COMPRESSIBLE_BYTES_OFFSET] = totalSize;
    if(isBase) {
        totalSize += 256;
    }

    indexes[CollationDataReader::IX_RESERVED18_OFFSET] = totalSize;
    indexes[CollationDataReader::IX_TOTAL_SIZE] = totalSize;

    if(totalSize > capacity) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return headerSize + totalSize;
    }

    uprv_memcpy(dest, indexes, indexesLength * 4);
    copyData(indexes, CollationDataReader::IX_REORDER_CODES_OFFSET, reorderCodes, dest);
    copyData(indexes, CollationDataReader::IX_REORDER_TABLE_OFFSET, settings.reorderTable, dest);
    // The trie has already been serialized into the dest buffer.
    copyData(indexes, CollationDataReader::IX_CES_OFFSET, data.ces, dest);
    copyData(indexes, CollationDataReader::IX_CE32S_OFFSET, data.ce32s, dest);
    copyData(indexes, CollationDataReader::IX_ROOT_ELEMENTS_OFFSET, rootElements, dest);
    copyData(indexes, CollationDataReader::IX_CONTEXTS_OFFSET, data.contexts, dest);
    // The unsafeBackwardSet has already been serialized into the dest buffer.
    copyData(indexes, CollationDataReader::IX_FAST_LATIN_TABLE_OFFSET, data.fastLatinTable, dest);
    copyData(indexes, CollationDataReader::IX_SCRIPTS_OFFSET, scripts.getBuffer(), dest);
    copyData(indexes, CollationDataReader::IX_COMPRESSIBLE_BYTES_OFFSET, data.compressibleBytes, dest);

    return headerSize + totalSize;
}

void
CollationDataWriter::copyData(const int32_t indexes[], int32_t startIndex,
                              const void *src, uint8_t *dest) {
    int32_t start = indexes[startIndex];
    int32_t limit = indexes[startIndex + 1];
    if(start < limit) {
        uprv_memcpy(dest + start, src, limit - start);
    }
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationfastlatin.cpp
*
* created on: 2013aug18
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION







U_NAMESPACE_BEGIN

int32_t
CollationFastLatin::getOptions(const CollationData *data, const CollationSettings &settings,
                               uint16_t *primaries, int32_t capacity) {
    const uint16_t *table = data->fastLatinTable;
    if(table == NULL) { return -1; }
    U_ASSERT(capacity == LATIN_LIMIT);
    if(capacity != LATIN_LIMIT) { return -1; }

    uint32_t miniVarTop;
    if((settings.options & CollationSettings::ALTERNATE_MASK) == 0) {
        // No mini primaries are variable, set a variableTop just below the
        // lowest long mini primary.
        miniVarTop = MIN_LONG - 1;
    } else {
        int32_t headerLength = *table & 0xff;
        int32_t i = 1 + settings.getMaxVariable();
        if(i >= headerLength) {
            return -1;  // variableTop >= digits, should not occur
        }
        miniVarTop = table[i];
    }

    UBool digitsAreReordered = FALSE;
    if(settings.hasReordering()) {
        uint32_t prevStart = 0;
        uint32_t beforeDigitStart = 0;
        uint32_t digitStart = 0;
        uint32_t afterDigitStart = 0;
        for(int32_t group = UCOL_REORDER_CODE_FIRST;
                group < UCOL_REORDER_CODE_FIRST + CollationData::MAX_NUM_SPECIAL_REORDER_CODES;
                ++group) {
            uint32_t start = data->getFirstPrimaryForGroup(group);
            start = settings.reorder(start);
            if(group == UCOL_REORDER_CODE_DIGIT) {
                beforeDigitStart = prevStart;
                digitStart = start;
            } else if(start != 0) {
                if(start < prevStart) {
                    // The permutation affects the groups up to Latin.
                    return -1;
                }
                // In the future, there might be a special group between digits & Latin.
                if(digitStart != 0 && afterDigitStart == 0 && prevStart == beforeDigitStart) {
                    afterDigitStart = start;
                }
                prevStart = start;
            }
        }
        uint32_t latinStart = data->getFirstPrimaryForGroup(USCRIPT_LATIN);
        latinStart = settings.reorder(latinStart);
        if(latinStart < prevStart) {
            return -1;
        }
        if(afterDigitStart == 0) {
            afterDigitStart = latinStart;
        }
        if(!(beforeDigitStart < digitStart && digitStart < afterDigitStart)) {
            digitsAreReordered = TRUE;
        }
    }

    table += (table[0] & 0xff);  // skip the header
    for(UChar32 c = 0; c < LATIN_LIMIT; ++c) {
        uint32_t p = table[c];
        if(p >= MIN_SHORT) {
            p &= SHORT_PRIMARY_MASK;
        } else if(p > miniVarTop) {
            p &= LONG_PRIMARY_MASK;
        } else {
            p = 0;
        }
        primaries[c] = (uint16_t)p;
    }
    if(digitsAreReordered || (settings.options & CollationSettings::NUMERIC) != 0) {
        // Bail out for digits.
        for(UChar32 c = 0x30; c <= 0x39; ++c) { primaries[c] = 0; }
    }

    // Shift the miniVarTop above other options.
    return ((int32_t)miniVarTop << 16) | settings.options;
}

int32_t
CollationFastLatin::compareUTF16(const uint16_t *table, const uint16_t *primaries, int32_t options,
                                 const UChar *left, int32_t leftLength,
                                 const UChar *right, int32_t rightLength) {
    // This is a modified copy of CollationCompare::compareUpToQuaternary(),
    // optimized for common Latin text.
    // Keep them in sync!
    // Keep compareUTF16() and compareUTF8() in sync very closely!

    U_ASSERT((table[0] >> 8) == VERSION);
    table += (table[0] & 0xff);  // skip the header
    uint32_t variableTop = (uint32_t)options >> 16;  // see getOptions()
    options &= 0xffff;  // needed for CollationSettings::getStrength() to work

    // Check for supported characters, fetch mini CEs, and compare primaries.
    int32_t leftIndex = 0, rightIndex = 0;
    /**
     * Single mini CE or a pair.
     * The current mini CE is in the lower 16 bits, the next one is in the upper 16 bits.
     * If there is only one, then it is in the lower bits, and the upper bits are 0.
     */
    uint32_t leftPair = 0, rightPair = 0;
    for(;;) {
        // We fetch CEs until we get a non-ignorable primary or reach the end.
        while(leftPair == 0) {
            if(leftIndex == leftLength) {
                leftPair = EOS;
                break;
            }
            UChar32 c = left[leftIndex++];
            if(c <= LATIN_MAX) {
                leftPair = primaries[c];
                if(leftPair != 0) { break; }
                if(c <= 0x39 && c >= 0x30 && (options & CollationSettings::NUMERIC) != 0) {
                    return BAIL_OUT_RESULT;
                }
                leftPair = table[c];
            } else if(PUNCT_START <= c && c < PUNCT_LIMIT) {
                leftPair = table[c - PUNCT_START + LATIN_LIMIT];
            } else {
                leftPair = lookup(table, c);
            }
            if(leftPair >= MIN_SHORT) {
                leftPair &= SHORT_PRIMARY_MASK;
                break;
            } else if(leftPair > variableTop) {
                leftPair &= LONG_PRIMARY_MASK;
                break;
            } else {
                leftPair = nextPair(table, c, leftPair, left, NULL, leftIndex, leftLength);
                if(leftPair == BAIL_OUT) { return BAIL_OUT_RESULT; }
                leftPair = getPrimaries(variableTop, leftPair);
            }
        }

        while(rightPair == 0) {
            if(rightIndex == rightLength) {
                rightPair = EOS;
                break;
            }
            UChar32 c = right[rightIndex++];
            if(c <= LATIN_MAX) {
                rightPair = primaries[c];
                if(rightPair != 0) { break; }
                if(c <= 0x39 && c >= 0x30 && (options & CollationSettings::NUMERIC) != 0) {
                    return BAIL_OUT_RESULT;
                }
                rightPair = table[c];
            } else if(PUNCT_START <= c && c < PUNCT_LIMIT) {
                rightPair = table[c - PUNCT_START + LATIN_LIMIT];
            } else {
                rightPair = lookup(table, c);
            }
            if(rightPair >= MIN_SHORT) {
                rightPair &= SHORT_PRIMARY_MASK;
                break;
            } else if(rightPair > variableTop) {
                rightPair &= LONG_PRIMARY_MASK;
                break;
            } else {
                rightPair = nextPair(table, c, rightPair, right, NULL, rightIndex, rightLength);
                if(rightPair == BAIL_OUT) { return BAIL_OUT_RESULT; }
                rightPair = getPrimaries(variableTop, rightPair);
            }
        }

        if(leftPair == rightPair) {
            if(leftPair == EOS) { break; }
            leftPair = rightPair = 0;
            continue;
        }
        uint32_t leftPrimary = leftPair & 0xffff;
        uint32_t rightPrimary = rightPair & 0xffff;
        if(leftPrimary != rightPrimary) {
            // Return the primary difference.
            return (leftPrimary < rightPrimary) ? UCOL_LESS : UCOL_GREATER;
        }
        if(leftPair == EOS) { break; }
        leftPair >>= 16;
        rightPair >>= 16;
    }
    // In the following, we need to re-fetch each character because we did not buffer the CEs,
    // but we know that the string is well-formed and
    // only contains supported characters and mappings.

    // We might skip the secondary level but continue with the case level
    // which is turned on separately.
    if(CollationSettings::getStrength(options) >= UCOL_SECONDARY) {
        leftIndex = rightIndex = 0;
        leftPair = rightPair = 0;
        for(;;) {
            while(leftPair == 0) {
                if(leftIndex == leftLength) {
                    leftPair = EOS;
                    break;
                }
                UChar32 c = left[leftIndex++];
                if(c <= LATIN_MAX) {
                    leftPair = table[c];
                } else if(PUNCT_START <= c && c < PUNCT_LIMIT) {
                    leftPair = table[c - PUNCT_START + LATIN_LIMIT];
                } else {
                    leftPair = lookup(table, c);
                }
                if(leftPair >= MIN_SHORT) {
                    leftPair = getSecondariesFromOneShortCE(leftPair);
                    break;
                } else if(leftPair > variableTop) {
                    leftPair = COMMON_SEC_PLUS_OFFSET;
                    break;
                } else {
                    leftPair = nextPair(table, c, leftPair, left, NULL, leftIndex, leftLength);
                    leftPair = getSecondaries(variableTop, leftPair);
                }
            }

            while(rightPair == 0) {
                if(rightIndex == rightLength) {
                    rightPair = EOS;
                    break;
                }
                UChar32 c = right[rightIndex++];
                if(c <= LATIN_MAX) {
                    rightPair = table[c];
                } else if(PUNCT_START <= c && c < PUNCT_LIMIT) {
                    rightPair = table[c - PUNCT_START + LATIN_LIMIT];
                } else {
                    rightPair = lookup(table, c);
                }
                if(rightPair >= MIN_SHORT) {
                    rightPair = getSecondariesFromOneShortCE(rightPair);
                    break;
                } else if(rightPair > variableTop) {
                    rightPair = COMMON_SEC_PLUS_OFFSET;
                    break;
                } else {
                    rightPair = nextPair(table, c, rightPair, right, NULL, rightIndex, rightLength);
                    rightPair = getSecondaries(variableTop, rightPair);
                }
            }

            if(leftPair == rightPair) {
                if(leftPair == EOS) { break; }
                leftPair = rightPair = 0;
                continue;
            }
            uint32_t leftSecondary = leftPair & 0xffff;
            uint32_t rightSecondary = rightPair & 0xffff;
            if(leftSecondary != rightSecondary) {
                if((options & CollationSettings::BACKWARD_SECONDARY) != 0) {
                    // Full support for backwards secondary requires backwards contraction matching
                    // and moving backwards between merge separators.
                    return BAIL_OUT_RESULT;
                }
                return (leftSecondary < rightSecondary) ? UCOL_LESS : UCOL_GREATER;
            }
            if(leftPair == EOS) { break; }
            leftPair >>= 16;
            rightPair >>= 16;
        }
    }

    if((options & CollationSettings::CASE_LEVEL) != 0) {
        UBool strengthIsPrimary = CollationSettings::getStrength(options) == UCOL_PRIMARY;
        leftIndex = rightIndex = 0;
        leftPair = rightPair = 0;
        for(;;) {
            while(leftPair == 0) {
                if(leftIndex == leftLength) {
                    leftPair = EOS;
                    break;
                }
                UChar32 c = left[leftIndex++];
                leftPair = (c <= LATIN_MAX) ? table[c] : lookup(table, c);
                if(leftPair < MIN_LONG) {
                    leftPair = nextPair(table, c, leftPair, left, NULL, leftIndex, leftLength);
                }
                leftPair = getCases(variableTop, strengthIsPrimary, leftPair);
            }

            while(rightPair == 0) {
                if(rightIndex == rightLength) {
                    rightPair = EOS;
                    break;
                }
                UChar32 c = right[rightIndex++];
                rightPair = (c <= LATIN_MAX) ? table[c] : lookup(table, c);
                if(rightPair < MIN_LONG) {
                    rightPair = nextPair(table, c, rightPair, right, NULL, rightIndex, rightLength);
                }
                rightPair = getCases(variableTop, strengthIsPrimary, rightPair);
            }

            if(leftPair == rightPair) {
                if(leftPair == EOS) { break; }
                leftPair = rightPair = 0;
                continue;
            }
            uint32_t leftCase = leftPair & 0xffff;
            uint32_t rightCase = rightPair & 0xffff;
            if(leftCase != rightCase) {
                if((options & CollationSettings::UPPER_FIRST) == 0) {
                    return (leftCase < rightCase) ? UCOL_LESS : UCOL_GREATER;
                } else {
                    return (leftCase < rightCase) ? UCOL_GREATER : UCOL_LESS;
                }
            }
            if(leftPair == EOS) { break; }
            leftPair >>= 16;
            rightPair >>= 16;
        }
    }
    if(CollationSettings::getStrength(options) <= UCOL_SECONDARY) { return UCOL_EQUAL; }

    // Remove the case bits from the tertiary weight when caseLevel is on or caseFirst is off.
    UBool withCaseBits = CollationSettings::isTertiaryWithCaseBits(options);

    leftIndex = rightIndex = 0;
    leftPair = rightPair = 0;
    for(;;) {
        while(leftPair == 0) {
            if(leftIndex == leftLength) {
                leftPair = EOS;
                break;
            }
            UChar32 c = left[leftIndex++];
            leftPair = (c <= LATIN_MAX) ? table[c] : lookup(table, c);
            if(leftPair < MIN_LONG) {
                leftPair = nextPair(table, c, leftPair, left, NULL, leftIndex, leftLength);
            }
            leftPair = getTertiaries(variableTop, withCaseBits, leftPair);
        }

        while(rightPair == 0) {
            if(rightIndex == rightLength) {
                rightPair = EOS;
                break;
            }
            UChar32 c = right[rightIndex++];
            rightPair = (c <= LATIN_MAX) ? table[c] : lookup(table, c);
            if(rightPair < MIN_LONG) {
                rightPair = nextPair(table, c, rightPair, right, NULL, rightIndex, rightLength);
            }
            rightPair = getTertiaries(variableTop, withCaseBits, rightPair);
        }

        if(leftPair == rightPair) {
            if(leftPair == EOS) { break; }
            leftPair = rightPair = 0;
            continue;
        }
        uint32_t leftTertiary = leftPair & 0xffff;
        uint32_t rightTertiary = rightPair & 0xffff;
        if(leftTertiary != rightTertiary) {
            if(CollationSettings::sortsTertiaryUpperCaseFirst(options)) {
                // Pass through EOS and MERGE_WEIGHT
                // and keep real tertiary weights larger than the MERGE_WEIGHT.
                // Tertiary CEs (secondary ignorables) are not supported in fast Latin.
                if(leftTertiary > MERGE_WEIGHT) {
                    leftTertiary ^= CASE_MASK;
                }
                if(rightTertiary > MERGE_WEIGHT) {
                    rightTertiary ^= CASE_MASK;
                }
            }
            return (leftTertiary < rightTertiary) ? UCOL_LESS : UCOL_GREATER;
        }
        if(leftPair == EOS) { break; }
        leftPair >>= 16;
        rightPair >>= 16;
    }
    if(CollationSettings::getStrength(options) <= UCOL_TERTIARY) { return UCOL_EQUAL; }

    leftIndex = rightIndex = 0;
    leftPair = rightPair = 0;
    for(;;) {
        while(leftPair == 0) {
            if(leftIndex == leftLength) {
                leftPair = EOS;
                break;
            }
            UChar32 c = left[leftIndex++];
            leftPair = (c <= LATIN_MAX) ? table[c] : lookup(table, c);
            if(leftPair < MIN_LONG) {
                leftPair = nextPair(table, c, leftPair, left, NULL, leftIndex, leftLength);
            }
            leftPair = getQuaternaries(variableTop, leftPair);
        }

        while(rightPair == 0) {
            if(rightIndex == rightLength) {
                rightPair = EOS;
                break;
            }
            UChar32 c = right[rightIndex++];
            rightPair = (c <= LATIN_MAX) ? table[c] : lookup(table, c);
            if(rightPair < MIN_LONG) {
                rightPair = nextPair(table, c, rightPair, right, NULL, rightIndex, rightLength);
            }
            rightPair = getQuaternaries(variableTop, rightPair);
        }

        if(leftPair == rightPair) {
            if(leftPair == EOS) { break; }
            leftPair = rightPair = 0;
            continue;
        }
        uint32_t leftQuaternary = leftPair & 0xffff;
        uint32_t rightQuaternary = rightPair & 0xffff;
        if(leftQuaternary != rightQuaternary) {
            return (leftQuaternary < rightQuaternary) ? UCOL_LESS : UCOL_GREATER;
        }
        if(leftPair == EOS) { break; }
        leftPair >>= 16;
        rightPair >>= 16;
    }
    return UCOL_EQUAL;
}

int32_t
CollationFastLatin::compareUTF8(const uint16_t *table, const uint16_t *primaries, int32_t options,
                                 const uint8_t *left, int32_t leftLength,
                                 const uint8_t *right, int32_t rightLength) {
    // Keep compareUTF16() and compareUTF8() in sync very closely!

    U_ASSERT((table[0] >> 8) == VERSION);
    table += (table[0] & 0xff);  // skip the header
    uint32_t variableTop = (uint32_t)options >> 16;  // see RuleBasedCollator::getFastLatinOptions()
    options &= 0xffff;  // needed for CollationSettings::getStrength() to work

    // Check for supported characters, fetch mini CEs, and compare primaries.
    int32_t leftIndex = 0, rightIndex = 0;
    /**
     * Single mini CE or a pair.
     * The current mini CE is in the lower 16 bits, the next one is in the upper 16 bits.
     * If there is only one, then it is in the lower bits, and the upper bits are 0.
     */
    uint32_t leftPair = 0, rightPair = 0;
    // Note: There is no need to assemble the code point.
    // We only need to look up the table entry for the character,
    // and nextPair() looks for whether c==0.
    for(;;) {
        // We fetch CEs until we get a non-ignorable primary or reach the end.
        while(leftPair == 0) {
            if(leftIndex == leftLength) {
                leftPair = EOS;
                break;
            }
            UChar32 c = left[leftIndex++];
            uint8_t t;
            if(c <= 0x7f) {
                leftPair = primaries[c];
                if(leftPair != 0) { break; }
                if(c <= 0x39 && c >= 0x30 && (options & CollationSettings::NUMERIC) != 0) {
                    return BAIL_OUT_RESULT;
                }
                leftPair = table[c];
            } else if(c <= LATIN_MAX_UTF8_LEAD && 0xc2 <= c && leftIndex != leftLength &&
                    0x80 <= (t = left[leftIndex]) && t <= 0xbf) {
                ++leftIndex;
                c = ((c - 0xc2) << 6) + t;
                leftPair = primaries[c];
                if(leftPair != 0) { break; }
                leftPair = table[c];
            } else {
                leftPair = lookupUTF8(table, c, left, leftIndex, leftLength);
            }
            if(leftPair >= MIN_SHORT) {
                leftPair &= SHORT_PRIMARY_MASK;
                break;
            } else if(leftPair > variableTop) {
                leftPair &= LONG_PRIMARY_MASK;
                break;
            } else {
                leftPair = nextPair(table, c, leftPair, NULL, left, leftIndex, leftLength);
                if(leftPair == BAIL_OUT) { return BAIL_OUT_RESULT; }
                leftPair = getPrimaries(variableTop, leftPair);
            }
        }

        while(rightPair == 0) {
            if(rightIndex == rightLength) {
                rightPair = EOS;
                break;
            }
            UChar32 c = right[rightIndex++];
            uint8_t t;
            if(c <= 0x7f) {
                rightPair = primaries[c];
                if(rightPair != 0) { break; }
                if(c <= 0x39 && c >= 0x30 && (options & CollationSettings::NUMERIC) != 0) {
                    return BAIL_OUT_RESULT;
                }
                rightPair = table[c];
            } else if(c <= LATIN_MAX_UTF8_LEAD && 0xc2 <= c && rightIndex != rightLength &&
                    0x80 <= (t = right[rightIndex]) && t <= 0xbf) {
                ++rightIndex;
                c = ((c - 0xc2) << 6) + t;
                rightPair = primaries[c];
                if(rightPair != 0) { break; }
                rightPair = table[c];
            } else {
                rightPair = lookupUTF8(table, c, right, rightIndex, rightLength);
            }
            if(rightPair >= MIN_SHORT) {
                rightPair &= SHORT_PRIMARY_MASK;
                break;
            } else if(rightPair > variableTop) {
                rightPair &= LONG_PRIMARY_MASK;
                break;
            } else {
                rightPair = nextPair(table, c, rightPair, NULL, right, rightIndex, rightLength);
                if(rightPair == BAIL_OUT) { return BAIL_OUT_RESULT; }
                rightPair = getPrimaries(variableTop, rightPair);
            }
        }

        if(leftPair == rightPair) {
            if(leftPair == EOS) { break; }
            leftPair = rightPair = 0;
            continue;
        }
        uint32_t leftPrimary = leftPair & 0xffff;
        uint32_t rightPrimary = rightPair & 0xffff;
        if(leftPrimary != rightPrimary) {
            // Return the primary difference.
            return (leftPrimary < rightPrimary) ? UCOL_LESS : UCOL_GREATER;
        }
        if(leftPair == EOS) { break; }
        leftPair >>= 16;
        rightPair >>= 16;
    }
    // In the following, we need to re-fetch each character because we did not buffer the CEs,
    // but we know that the string is well-formed and
    // only contains supported characters and mappings.

    // We might skip the secondary level but continue with the case level
    // which is turned on separately.
    if(CollationSettings::getStrength(options) >= UCOL_SECONDARY) {
        leftIndex = rightIndex = 0;
        leftPair = rightPair = 0;
        for(;;) {
            while(leftPair == 0) {
                if(leftIndex == leftLength) {
                    leftPair = EOS;
                    break;
                }
                UChar32 c = left[leftIndex++];
                if(c <= 0x7f) {
                    leftPair = table[c];
                } else if(c <= LATIN_MAX_UTF8_LEAD) {
                    leftPair = table[((c - 0xc2) << 6) + left[leftIndex++]];
                } else {
                    leftPair = lookupUTF8Unsafe(table, c, left, leftIndex);
                }
                if(leftPair >= MIN_SHORT) {
                    leftPair = getSecondariesFromOneShortCE(leftPair);
                    break;
                } else if(leftPair > variableTop) {
                    leftPair = COMMON_SEC_PLUS_OFFSET;
                    break;
                } else {
                    leftPair = nextPair(table, c, leftPair, NULL, left, leftIndex, leftLength);
                    leftPair = getSecondaries(variableTop, leftPair);
                }
            }

            while(rightPair == 0) {
                if(rightIndex == rightLength) {
                    rightPair = EOS;
                    break;
                }
                UChar32 c = right[rightIndex++];
                if(c <= 0x7f) {
                    rightPair = table[c];
                } else if(c <= LATIN_MAX_UTF8_LEAD) {
                    rightPair = table[((c - 0xc2) << 6) + right[rightIndex++]];
                } else {
                    rightPair = lookupUTF8Unsafe(table, c, right, rightIndex);
                }
                if(rightPair >= MIN_SHORT) {
                    rightPair = getSecondariesFromOneShortCE(rightPair);
                    break;
                } else if(rightPair > variableTop) {
                    rightPair = COMMON_SEC_PLUS_OFFSET;
                    break;
                } else {
                    rightPair = nextPair(table, c, rightPair, NULL, right, rightIndex, rightLength);
                    rightPair = getSecondaries(variableTop, rightPair);
                }
            }

            if(leftPair == rightPair) {
                if(leftPair == EOS) { break; }
                leftPair = rightPair = 0;
                continue;
            }
            uint32_t leftSecondary = leftPair & 0xffff;
            uint32_t rightSecondary = rightPair & 0xffff;
            if(leftSecondary != rightSecondary) {
                if((options & CollationSettings::BACKWARD_SECONDARY) != 0) {
                    // Full support for backwards secondary requires backwards contraction matching
                    // and moving backwards between merge separators.
                    return BAIL_OUT_RESULT;
                }
                return (leftSecondary < rightSecondary) ? UCOL_LESS : UCOL_GREATER;
            }
            if(leftPair == EOS) { break; }
            leftPair >>= 16;
            rightPair >>= 16;
        }
    }

    if((options & CollationSettings::CASE_LEVEL) != 0) {
        UBool strengthIsPrimary = CollationSettings::getStrength(options) == UCOL_PRIMARY;
        leftIndex = rightIndex = 0;
        leftPair = rightPair = 0;
        for(;;) {
            while(leftPair == 0) {
                if(leftIndex == leftLength) {
                    leftPair = EOS;
                    break;
                }
                UChar32 c = left[leftIndex++];
                leftPair = (c <= 0x7f) ? table[c] : lookupUTF8Unsafe(table, c, left, leftIndex);
                if(leftPair < MIN_LONG) {
                    leftPair = nextPair(table, c, leftPair, NULL, left, leftIndex, leftLength);
                }
                leftPair = getCases(variableTop, strengthIsPrimary, leftPair);
            }

            while(rightPair == 0) {
                if(rightIndex == rightLength) {
                    rightPair = EOS;
                    break;
                }
                UChar32 c = right[rightIndex++];
                rightPair = (c <= 0x7f) ? table[c] : lookupUTF8Unsafe(table, c, right, rightIndex);
                if(rightPair < MIN_LONG) {
                    rightPair = nextPair(table, c, rightPair, NULL, right, rightIndex, rightLength);
                }
                rightPair = getCases(variableTop, strengthIsPrimary, rightPair);
            }

            if(leftPair == rightPair) {
                if(leftPair == EOS) { break; }
                leftPair = rightPair = 0;
                continue;
            }
            uint32_t leftCase = leftPair & 0xffff;
            uint32_t rightCase = rightPair & 0xffff;
            if(leftCase != rightCase) {
                if((options & CollationSettings::UPPER_FIRST) == 0) {
                    return (leftCase < rightCase) ? UCOL_LESS : UCOL_GREATER;
                } else {
                    return (leftCase < rightCase) ? UCOL_GREATER : UCOL_LESS;
                }
            }
            if(leftPair == EOS) { break; }
            leftPair >>= 16;
            rightPair >>= 16;
        }
    }
    if(CollationSettings::getStrength(options) <= UCOL_SECONDARY) { return UCOL_EQUAL; }

    // Remove the case bits from the tertiary weight when caseLevel is on or caseFirst is off.
    UBool withCaseBits = CollationSettings::isTertiaryWithCaseBits(options);

    leftIndex = rightIndex = 0;
    leftPair = rightPair = 0;
    for(;;) {
        while(leftPair == 0) {
            if(leftIndex == leftLength) {
                leftPair = EOS;
                break;
            }
            UChar32 c = left[leftIndex++];
            leftPair = (c <= 0x7f) ? table[c] : lookupUTF8Unsafe(table, c, left, leftIndex);
            if(leftPair < MIN_LONG) {
                leftPair = nextPair(table, c, leftPair, NULL, left, leftIndex, leftLength);
            }
            leftPair = getTertiaries(variableTop, withCaseBits, leftPair);
        }

        while(rightPair == 0) {
            if(rightIndex == rightLength) {
                rightPair = EOS;
                break;
            }
            UChar32 c = right[rightIndex++];
            rightPair = (c <= 0x7f) ? table[c] : lookupUTF8Unsafe(table, c, right, rightIndex);
            if(rightPair < MIN_LONG) {
                rightPair = nextPair(table, c, rightPair, NULL, right, rightIndex, rightLength);
            }
            rightPair = getTertiaries(variableTop, withCaseBits, rightPair);
        }

        if(leftPair == rightPair) {
            if(leftPair == EOS) { break; }
            leftPair = rightPair = 0;
            continue;
        }
        uint32_t leftTertiary = leftPair & 0xffff;
        uint32_t rightTertiary = rightPair & 0xffff;
        if(leftTertiary != rightTertiary) {
            if(CollationSettings::sortsTertiaryUpperCaseFirst(options)) {
                // Pass through EOS and MERGE_WEIGHT
                // and keep real tertiary weights larger than the MERGE_WEIGHT.
                // Tertiary CEs (secondary ignorables) are not supported in fast Latin.
                if(leftTertiary > MERGE_WEIGHT) {
                    leftTertiary ^= CASE_MASK;
                }
                if(rightTertiary > MERGE_WEIGHT) {
                    rightTertiary ^= CASE_MASK;
                }
            }
            return (leftTertiary < rightTertiary) ? UCOL_LESS : UCOL_GREATER;
        }
        if(leftPair == EOS) { break; }
        leftPair >>= 16;
        rightPair >>= 16;
    }
    if(CollationSettings::getStrength(options) <= UCOL_TERTIARY) { return UCOL_EQUAL; }

    leftIndex = rightIndex = 0;
    leftPair = rightPair = 0;
    for(;;) {
        while(leftPair == 0) {
            if(leftIndex == leftLength) {
                leftPair = EOS;
                break;
            }
            UChar32 c = left[leftIndex++];
            leftPair = (c <= 0x7f) ? table[c] : lookupUTF8Unsafe(table, c, left, leftIndex);
            if(leftPair < MIN_LONG) {
                leftPair = nextPair(table, c, leftPair, NULL, left, leftIndex, leftLength);
            }
            leftPair = getQuaternaries(variableTop, leftPair);
        }

        while(rightPair == 0) {
            if(rightIndex == rightLength) {
                rightPair = EOS;
                break;
            }
            UChar32 c = right[rightIndex++];
            rightPair = (c <= 0x7f) ? table[c] : lookupUTF8Unsafe(table, c, right, rightIndex);
            if(rightPair < MIN_LONG) {
                rightPair = nextPair(table, c, rightPair, NULL, right, rightIndex, rightLength);
            }
            rightPair = getQuaternaries(variableTop, rightPair);
        }

        if(leftPair == rightPair) {
            if(leftPair == EOS) { break; }
            leftPair = rightPair = 0;
            continue;
        }
        uint32_t leftQuaternary = leftPair & 0xffff;
        uint32_t rightQuaternary = rightPair & 0xffff;
        if(leftQuaternary != rightQuaternary) {
            return (leftQuaternary < rightQuaternary) ? UCOL_LESS : UCOL_GREATER;
        }
        if(leftPair == EOS) { break; }
        leftPair >>= 16;
        rightPair >>= 16;
    }
    return UCOL_EQUAL;
}

uint32_t
CollationFastLatin::lookup(const uint16_t *table, UChar32 c) {
    U_ASSERT(c > LATIN_MAX);
    if(PUNCT_START <= c && c < PUNCT_LIMIT) {
        return table[c - PUNCT_START + LATIN_LIMIT];
    } else if(c == 0xfffe) {
        return MERGE_WEIGHT;
    } else if(c == 0xffff) {
        return MAX_SHORT | COMMON_SEC | LOWER_CASE | COMMON_TER;
    } else {
        return BAIL_OUT;
    }
}

uint32_t
CollationFastLatin::lookupUTF8(const uint16_t *table, UChar32 c,
                               const uint8_t *s8, int32_t &sIndex, int32_t sLength) {
    // The caller handled ASCII and valid/supported Latin.
    U_ASSERT(c > 0x7f);
    int32_t i2 = sIndex + 1;
    if(i2 < sLength || sLength < 0) {
        uint8_t t1 = s8[sIndex];
        uint8_t t2 = s8[i2];
        sIndex += 2;
        if(c == 0xe2 && t1 == 0x80 && 0x80 <= t2 && t2 <= 0xbf) {
            return table[(LATIN_LIMIT - 0x80) + t2];  // 2000..203F -> 0180..01BF
        } else if(c == 0xef && t1 == 0xbf) {
            if(t2 == 0xbe) {
                return MERGE_WEIGHT;  // U+FFFE
            } else if(t2 == 0xbf) {
                return MAX_SHORT | COMMON_SEC | LOWER_CASE | COMMON_TER;  // U+FFFF
            }
        }
    }
    return BAIL_OUT;
}

uint32_t
CollationFastLatin::lookupUTF8Unsafe(const uint16_t *table, UChar32 c,
                                     const uint8_t *s8, int32_t &sIndex) {
    // The caller handled ASCII.
    // The string is well-formed and contains only supported characters.
    U_ASSERT(c > 0x7f);
    if(c <= LATIN_MAX_UTF8_LEAD) {
        return table[((c - 0xc2) << 6) + s8[sIndex++]];  // 0080..017F
    }
    uint8_t t2 = s8[sIndex + 1];
    sIndex += 2;
    if(c == 0xe2) {
        return table[(LATIN_LIMIT - 0x80) + t2];  // 2000..203F -> 0180..01BF
    } else if(t2 == 0xbe) {
        return MERGE_WEIGHT;  // U+FFFE
    } else {
        return MAX_SHORT | COMMON_SEC | LOWER_CASE | COMMON_TER;  // U+FFFF
    }
}

uint32_t
CollationFastLatin::nextPair(const uint16_t *table, UChar32 c, uint32_t ce,
                             const UChar *s16, const uint8_t *s8, int32_t &sIndex, int32_t &sLength) {
    if(ce >= MIN_LONG || ce < CONTRACTION) {
        return ce;  // simple or special mini CE
    } else if(ce >= EXPANSION) {
        int32_t index = NUM_FAST_CHARS + (ce & INDEX_MASK);
        return ((uint32_t)table[index + 1] << 16) | table[index];
    } else /* ce >= CONTRACTION */ {
        if(c == 0 && sLength < 0) {
            sLength = sIndex - 1;
            return EOS;
        }
        // Contraction list: Default mapping followed by
        // 0 or more single-character contraction suffix mappings.
        int32_t index = NUM_FAST_CHARS + (ce & INDEX_MASK);
        if(sIndex != sLength) {
            // Read the next character.
            int32_t c2;
            int32_t nextIndex = sIndex;
            if(s16 != NULL) {
                c2 = s16[nextIndex++];
                if(c2 > LATIN_MAX) {
                    if(PUNCT_START <= c2 && c2 < PUNCT_LIMIT) {
                        c2 = c2 - PUNCT_START + LATIN_LIMIT;  // 2000..203F -> 0180..01BF
                    } else if(c2 == 0xfffe || c2 == 0xffff) {
                        c2 = -1;  // U+FFFE & U+FFFF cannot occur in contractions.
                    } else {
                        return BAIL_OUT;
                    }
                }
            } else {
                c2 = s8[nextIndex++];
                if(c2 > 0x7f) {
                    uint8_t t;
                    if(c2 <= 0xc5 && 0xc2 <= c2 && nextIndex != sLength &&
                            0x80 <= (t = s8[nextIndex]) && t <= 0xbf) {
                        c2 = ((c2 - 0xc2) << 6) + t;  // 0080..017F
                        ++nextIndex;
                    } else {
                        int32_t i2 = nextIndex + 1;
                        if(i2 < sLength || sLength < 0) {
                            if(c2 == 0xe2 && s8[nextIndex] == 0x80 &&
                                    0x80 <= (t = s8[i2]) && t <= 0xbf) {
                                c2 = (LATIN_LIMIT - 0x80) + t;  // 2000..203F -> 0180..01BF
                            } else if(c2 == 0xef && s8[nextIndex] == 0xbf &&
                                    ((t = s8[i2]) == 0xbe || t == 0xbf)) {
                                c2 = -1;  // U+FFFE & U+FFFF cannot occur in contractions.
                            } else {
                                return BAIL_OUT;
                            }
                        } else {
                            return BAIL_OUT;
                        }
                        nextIndex += 2;
                    }
                }
            }
            if(c2 == 0 && sLength < 0) {
                sLength = sIndex;
                c2 = -1;
            }
            // Look for the next character in the contraction suffix list,
            // which is in ascending order of single suffix characters.
            int32_t i = index;
            int32_t head = table[i];  // first skip the default mapping
            int32_t x;
            do {
                i += head >> CONTR_LENGTH_SHIFT;
                head = table[i];
                x = head & CONTR_CHAR_MASK;
            } while(x < c2);
            if(x == c2) {
                index = i;
                sIndex = nextIndex;
            }
        }
        // Return the CE or CEs for the default or contraction mapping.
        int32_t length = table[index] >> CONTR_LENGTH_SHIFT;
        if(length == 1) {
            return BAIL_OUT;
        }
        ce = table[index + 1];
        if(length == 2) {
            return ce;
        } else {
            return ((uint32_t)table[index + 2] << 16) | ce;
        }
    }
}

uint32_t
CollationFastLatin::getSecondaries(uint32_t variableTop, uint32_t pair) {
    if(pair <= 0xffff) {
        // one mini CE
        if(pair >= MIN_SHORT) {
            pair = getSecondariesFromOneShortCE(pair);
        } else if(pair > variableTop) {
            pair = COMMON_SEC_PLUS_OFFSET;
        } else if(pair >= MIN_LONG) {
            pair = 0;  // variable
        }
        // else special mini CE
    } else {
        uint32_t ce = pair & 0xffff;
        if(ce >= MIN_SHORT) {
            pair = (pair & TWO_SECONDARIES_MASK) + TWO_SEC_OFFSETS;
        } else if(ce > variableTop) {
            pair = TWO_COMMON_SEC_PLUS_OFFSET;
        } else {
            U_ASSERT(ce >= MIN_LONG);
            pair = 0;  // variable
        }
    }
    return pair;
}

uint32_t
CollationFastLatin::getCases(uint32_t variableTop, UBool strengthIsPrimary, uint32_t pair) {
    // Primary+caseLevel: Ignore case level weights of primary ignorables.
    // Otherwise: Ignore case level weights of secondary ignorables.
    // For details see the comments in the CollationCompare class.
    // Tertiary CEs (secondary ignorables) are not supported in fast Latin.
    if(pair <= 0xffff) {
        // one mini CE
        if(pair >= MIN_SHORT) {
            // A high secondary weight means we really have two CEs,
            // a primary CE and a secondary CE.
            uint32_t ce = pair;
            pair &= CASE_MASK;  // explicit weight of primary CE
            if(!strengthIsPrimary && (ce & SECONDARY_MASK) >= MIN_SEC_HIGH) {
                pair |= LOWER_CASE << 16;  // implied weight of secondary CE
            }
        } else if(pair > variableTop) {
            pair = LOWER_CASE;
        } else if(pair >= MIN_LONG) {
            pair = 0;  // variable
        }
        // else special mini CE
    } else {
        // two mini CEs, same primary groups, neither expands like above
        uint32_t ce = pair & 0xffff;
        if(ce >= MIN_SHORT) {
            if(strengthIsPrimary && (pair & (SHORT_PRIMARY_MASK << 16)) == 0) {
                pair &= CASE_MASK;
            } else {
                pair &= TWO_CASES_MASK;
            }
        } else if(ce > variableTop) {
            pair = TWO_LOWER_CASES;
        } else {
            U_ASSERT(ce >= MIN_LONG);
            pair = 0;  // variable
        }
    }
    return pair;
}

uint32_t
CollationFastLatin::getTertiaries(uint32_t variableTop, UBool withCaseBits, uint32_t pair) {
    if(pair <= 0xffff) {
        // one mini CE
        if(pair >= MIN_SHORT) {
            // A high secondary weight means we really have two CEs,
            // a primary CE and a secondary CE.
            uint32_t ce = pair;
            if(withCaseBits) {
                pair = (pair & CASE_AND_TERTIARY_MASK) + TER_OFFSET;
                if((ce & SECONDARY_MASK) >= MIN_SEC_HIGH) {
                    pair |= (LOWER_CASE | COMMON_TER_PLUS_OFFSET) << 16;
                }
            } else {
                pair = (pair & TERTIARY_MASK) + TER_OFFSET;
                if((ce & SECONDARY_MASK) >= MIN_SEC_HIGH) {
                    pair |= COMMON_TER_PLUS_OFFSET << 16;
                }
            }
        } else if(pair > variableTop) {
            pair = (pair & TERTIARY_MASK) + TER_OFFSET;
            if(withCaseBits) {
                pair |= LOWER_CASE;
            }
        } else if(pair >= MIN_LONG) {
            pair = 0;  // variable
        }
        // else special mini CE
    } else {
        // two mini CEs, same primary groups, neither expands like above
        uint32_t ce = pair & 0xffff;
        if(ce >= MIN_SHORT) {
            if(withCaseBits) {
                pair &= TWO_CASES_MASK | TWO_TERTIARIES_MASK;
            } else {
                pair &= TWO_TERTIARIES_MASK;
            }
            pair += TWO_TER_OFFSETS;
        } else if(ce > variableTop) {
            pair = (pair & TWO_TERTIARIES_MASK) + TWO_TER_OFFSETS;
            if(withCaseBits) {
                pair |= TWO_LOWER_CASES;
            }
        } else {
            U_ASSERT(ce >= MIN_LONG);
            pair = 0;  // variable
        }
    }
    return pair;
}

uint32_t
CollationFastLatin::getQuaternaries(uint32_t variableTop, uint32_t pair) {
    // Return the primary weight of a variable CE,
    // or the maximum primary weight for a non-variable, not-completely-ignorable CE.
    if(pair <= 0xffff) {
        // one mini CE
        if(pair >= MIN_SHORT) {
            // A high secondary weight means we really have two CEs,
            // a primary CE and a secondary CE.
            if((pair & SECONDARY_MASK) >= MIN_SEC_HIGH) {
                pair = TWO_SHORT_PRIMARIES_MASK;
            } else {
                pair = SHORT_PRIMARY_MASK;
            }
        } else if(pair > variableTop) {
            pair = SHORT_PRIMARY_MASK;
        } else if(pair >= MIN_LONG) {
            pair &= LONG_PRIMARY_MASK;  // variable
        }
        // else special mini CE
    } else {
        // two mini CEs, same primary groups, neither expands like above
        uint32_t ce = pair & 0xffff;
        if(ce > variableTop) {
            pair = TWO_SHORT_PRIMARIES_MASK;
        } else {
            U_ASSERT(ce >= MIN_LONG);
            pair &= TWO_LONG_PRIMARIES_MASK;  // variable
        }
    }
    return pair;
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationfastlatinbuilder.cpp
*
* created on: 2013aug09
* created by: Markus W. Scherer
*/

#define DEBUG_COLLATION_FAST_LATIN_BUILDER 0  // 0 or 1 or 2
#if DEBUG_COLLATION_FAST_LATIN_BUILDER
#include <stdio.h>
#include <string>
#endif



#if !UCONFIG_NO_COLLATION














U_NAMESPACE_BEGIN

struct CollationData;

namespace {

/**
 * Compare two signed int64_t values as if they were unsigned.
 */
int32_t
compareInt64AsUnsigned(int64_t a, int64_t b) {
    if((uint64_t)a < (uint64_t)b) {
        return -1;
    } else if((uint64_t)a > (uint64_t)b) {
        return 1;
    } else {
        return 0;
    }
}

// TODO: Merge this with the near-identical version in collationbasedatabuilder.cpp
/**
 * Like Java Collections.binarySearch(List, String, Comparator).
 *
 * @return the index>=0 where the item was found,
 *         or the index<0 for inserting the string at ~index in sorted order
 */
int32_t
binarySearch(const int64_t list[], int32_t limit, int64_t ce) {
    if (limit == 0) { return ~0; }
    int32_t start = 0;
    for (;;) {
        int32_t i = (start + limit) / 2;
        int32_t cmp = compareInt64AsUnsigned(ce, list[i]);
        if (cmp == 0) {
            return i;
        } else if (cmp < 0) {
            if (i == start) {
                return ~start;  // insert ce before i
            }
            limit = i;
        } else {
            if (i == start) {
                return ~(start + 1);  // insert ce after i
            }
            start = i;
        }
    }
}

}  // namespace

CollationFastLatinBuilder::CollationFastLatinBuilder(UErrorCode &errorCode)
        : ce0(0), ce1(0),
          contractionCEs(errorCode), uniqueCEs(errorCode),
          miniCEs(NULL),
          firstDigitPrimary(0), firstLatinPrimary(0), lastLatinPrimary(0),
          firstShortPrimary(0), shortPrimaryOverflow(FALSE),
          headerLength(0) {
}

CollationFastLatinBuilder::~CollationFastLatinBuilder() {
    uprv_free(miniCEs);
}

UBool
CollationFastLatinBuilder::forData(const CollationData &data, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    if(!result.isEmpty()) {  // This builder is not reusable.
        errorCode = U_INVALID_STATE_ERROR;
        return FALSE;
    }
    if(!loadGroups(data, errorCode)) { return FALSE; }

    // Fast handling of digits.
    firstShortPrimary = firstDigitPrimary;
    getCEs(data, errorCode);
    if(!encodeUniqueCEs(errorCode)) { return FALSE; }
    if(shortPrimaryOverflow) {
        // Give digits long mini primaries,
        // so that there are more short primaries for letters.
        firstShortPrimary = firstLatinPrimary;
        resetCEs();
        getCEs(data, errorCode);
        if(!encodeUniqueCEs(errorCode)) { return FALSE; }
    }
    // Note: If we still have a short-primary overflow but not a long-primary overflow,
    // then we could calculate how many more long primaries would fit,
    // and set the firstShortPrimary to that many after the current firstShortPrimary,
    // and try again.
    // However, this might only benefit the en_US_POSIX tailoring,
    // and it is simpler to suppress building fast Latin data for it in genrb,
    // or by returning FALSE here if shortPrimaryOverflow.

    UBool ok = !shortPrimaryOverflow &&
            encodeCharCEs(errorCode) && encodeContractions(errorCode);
    contractionCEs.removeAllElements();  // might reduce heap memory usage
    uniqueCEs.removeAllElements();
    return ok;
}

UBool
CollationFastLatinBuilder::loadGroups(const CollationData &data, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    headerLength = 1 + NUM_SPECIAL_GROUPS;
    uint32_t r0 = (CollationFastLatin::VERSION << 8) | headerLength;
    result.append((UChar)r0);
    // The first few reordering groups should be special groups
    // (space, punct, ..., digit) followed by Latn, then Grek and other scripts.
    for(int32_t i = 0; i < NUM_SPECIAL_GROUPS; ++i) {
        lastSpecialPrimaries[i] = data.getLastPrimaryForGroup(UCOL_REORDER_CODE_FIRST + i);
        if(lastSpecialPrimaries[i] == 0) {
            // missing data
            return FALSE;
        }
        result.append((UChar)0);  // reserve a slot for this group
    }

    firstDigitPrimary = data.getFirstPrimaryForGroup(UCOL_REORDER_CODE_DIGIT);
    firstLatinPrimary = data.getFirstPrimaryForGroup(USCRIPT_LATIN);
    lastLatinPrimary = data.getLastPrimaryForGroup(USCRIPT_LATIN);
    if(firstDigitPrimary == 0 || firstLatinPrimary == 0) {
        // missing data
        return FALSE;
    }
    return TRUE;
}

UBool
CollationFastLatinBuilder::inSameGroup(uint32_t p, uint32_t q) const {
    // Both or neither need to be encoded as short primaries,
    // so that we can test only one and use the same bit mask.
    if(p >= firstShortPrimary) {
        return q >= firstShortPrimary;
    } else if(q >= firstShortPrimary) {
        return FALSE;
    }
    // Both or neither must be potentially-variable,
    // so that we can test only one and determine if both are variable.
    uint32_t lastVariablePrimary = lastSpecialPrimaries[NUM_SPECIAL_GROUPS - 1];
    if(p > lastVariablePrimary) {
        return q > lastVariablePrimary;
    } else if(q > lastVariablePrimary) {
        return FALSE;
    }
    // Both will be encoded with long mini primaries.
    // They must be in the same special reordering group,
    // so that we can test only one and determine if both are variable.
    U_ASSERT(p != 0 && q != 0);
    for(int32_t i = 0;; ++i) {  // will terminate
        uint32_t lastPrimary = lastSpecialPrimaries[i];
        if(p <= lastPrimary) {
            return q <= lastPrimary;
        } else if(q <= lastPrimary) {
            return FALSE;
        }
    }
}

void
CollationFastLatinBuilder::resetCEs() {
    contractionCEs.removeAllElements();
    uniqueCEs.removeAllElements();
    shortPrimaryOverflow = FALSE;
    result.truncate(headerLength);
}

void
CollationFastLatinBuilder::getCEs(const CollationData &data, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    int32_t i = 0;
    for(UChar c = 0;; ++i, ++c) {
        if(c == CollationFastLatin::LATIN_LIMIT) {
            c = CollationFastLatin::PUNCT_START;
        } else if(c == CollationFastLatin::PUNCT_LIMIT) {
            break;
        }
        const CollationData *d;
        uint32_t ce32 = data.getCE32(c);
        if(ce32 == Collation::FALLBACK_CE32) {
            d = data.base;
            ce32 = d->getCE32(c);
        } else {
            d = &data;
        }
        if(getCEsFromCE32(*d, c, ce32, errorCode)) {
            charCEs[i][0] = ce0;
            charCEs[i][1] = ce1;
            addUniqueCE(ce0, errorCode);
            addUniqueCE(ce1, errorCode);
        } else {
            // bail out for c
            charCEs[i][0] = ce0 = Collation::NO_CE;
            charCEs[i][1] = ce1 = 0;
        }
        if(c == 0 && !isContractionCharCE(ce0)) {
            // Always map U+0000 to a contraction.
            // Write a contraction list with only a default value if there is no real contraction.
            U_ASSERT(contractionCEs.isEmpty());
            addContractionEntry(CollationFastLatin::CONTR_CHAR_MASK, ce0, ce1, errorCode);
            charCEs[0][0] = ((int64_t)Collation::NO_CE_PRIMARY << 32) | CONTRACTION_FLAG;
            charCEs[0][1] = 0;
        }
    }
    // Terminate the last contraction list.
    contractionCEs.addElement(CollationFastLatin::CONTR_CHAR_MASK, errorCode);
}

UBool
CollationFastLatinBuilder::getCEsFromCE32(const CollationData &data, UChar32 c, uint32_t ce32,
                                          UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    ce32 = data.getFinalCE32(ce32);
    ce1 = 0;
    if(Collation::isSimpleOrLongCE32(ce32)) {
        ce0 = Collation::ceFromCE32(ce32);
    } else {
        switch(Collation::tagFromCE32(ce32)) {
        case Collation::LATIN_EXPANSION_TAG:
            ce0 = Collation::latinCE0FromCE32(ce32);
            ce1 = Collation::latinCE1FromCE32(ce32);
            break;
        case Collation::EXPANSION32_TAG: {
            const uint32_t *ce32s = data.ce32s + Collation::indexFromCE32(ce32);
            int32_t length = Collation::lengthFromCE32(ce32);
            if(length <= 2) {
                ce0 = Collation::ceFromCE32(ce32s[0]);
                if(length == 2) {
                    ce1 = Collation::ceFromCE32(ce32s[1]);
                }
                break;
            } else {
                return FALSE;
            }
        }
        case Collation::EXPANSION_TAG: {
            const int64_t *ces = data.ces + Collation::indexFromCE32(ce32);
            int32_t length = Collation::lengthFromCE32(ce32);
            if(length <= 2) {
                ce0 = ces[0];
                if(length == 2) {
                    ce1 = ces[1];
                }
                break;
            } else {
                return FALSE;
            }
        }
        // Note: We could support PREFIX_TAG (assert c>=0)
        // by recursing on its default CE32 and checking that none of the prefixes starts
        // with a fast Latin character.
        // However, currently (2013) there are only the L-before-middle-dot
        // prefix mappings in the Latin range, and those would be rejected anyway.
        case Collation::CONTRACTION_TAG:
            U_ASSERT(c >= 0);
            return getCEsFromContractionCE32(data, ce32, errorCode);
        case Collation::OFFSET_TAG:
            U_ASSERT(c >= 0);
            ce0 = data.getCEFromOffsetCE32(c, ce32);
            break;
        default:
            return FALSE;
        }
    }
    // A mapping can be completely ignorable.
    if(ce0 == 0) { return ce1 == 0; }
    // We do not support an ignorable ce0 unless it is completely ignorable.
    uint32_t p0 = (uint32_t)(ce0 >> 32);
    if(p0 == 0) { return FALSE; }
    // We only support primaries up to the Latin script.
    if(p0 > lastLatinPrimary) { return FALSE; }
    // We support non-common secondary and case weights only together with short primaries.
    uint32_t lower32_0 = (uint32_t)ce0;
    if(p0 < firstShortPrimary) {
        uint32_t sc0 = lower32_0 & Collation::SECONDARY_AND_CASE_MASK;
        if(sc0 != Collation::COMMON_SECONDARY_CE) { return FALSE; }
    }
    // No below-common tertiary weights.
    if((lower32_0 & Collation::ONLY_TERTIARY_MASK) < Collation::COMMON_WEIGHT16) { return FALSE; }
    if(ce1 != 0) {
        // Both primaries must be in the same group,
        // or both must get short mini primaries,
        // or a short-primary CE is followed by a secondary CE.
        // This is so that we can test the first primary and use the same mask for both,
        // and determine for both whether they are variable.
        uint32_t p1 = (uint32_t)(ce1 >> 32);
        if(p1 == 0 ? p0 < firstShortPrimary : !inSameGroup(p0, p1)) { return FALSE; }
        uint32_t lower32_1 = (uint32_t)ce1;
        // No tertiary CEs.
        if((lower32_1 >> 16) == 0) { return FALSE; }
        // We support non-common secondary and case weights
        // only for secondary CEs or together with short primaries.
        if(p1 != 0 && p1 < firstShortPrimary) {
            uint32_t sc1 = lower32_1 & Collation::SECONDARY_AND_CASE_MASK;
            if(sc1 != Collation::COMMON_SECONDARY_CE) { return FALSE; }
        }
        // No below-common tertiary weights.
        if((lower32_1 & Collation::ONLY_TERTIARY_MASK) < Collation::COMMON_WEIGHT16) { return FALSE; }
    }
    // No quaternary weights.
    if(((ce0 | ce1) & Collation::QUATERNARY_MASK) != 0) { return FALSE; }
    return TRUE;
}

UBool
CollationFastLatinBuilder::getCEsFromContractionCE32(const CollationData &data, uint32_t ce32,
                                                     UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    const UChar *p = data.contexts + Collation::indexFromCE32(ce32);
    ce32 = CollationData::readCE32(p);  // Default if no suffix match.
    // Since the original ce32 is not a prefix mapping,
    // the default ce32 must not be another contraction.
    U_ASSERT(!Collation::isContractionCE32(ce32));
    int32_t contractionIndex = contractionCEs.size();
    if(getCEsFromCE32(data, U_SENTINEL, ce32, errorCode)) {
        addContractionEntry(CollationFastLatin::CONTR_CHAR_MASK, ce0, ce1, errorCode);
    } else {
        // Bail out for c-without-contraction.
        addContractionEntry(CollationFastLatin::CONTR_CHAR_MASK, Collation::NO_CE, 0, errorCode);
    }
    // Handle an encodable contraction unless the next contraction is too long
    // and starts with the same character.
    int32_t prevX = -1;
    UBool addContraction = FALSE;
    UCharsTrie::Iterator suffixes(p + 2, 0, errorCode);
    while(suffixes.next(errorCode)) {
        const UnicodeString &suffix = suffixes.getString();
        int32_t x = CollationFastLatin::getCharIndex(suffix.charAt(0));
        if(x < 0) { continue; }  // ignore anything but fast Latin text
        if(x == prevX) {
            if(addContraction) {
                // Bail out for all contractions starting with this character.
                addContractionEntry(x, Collation::NO_CE, 0, errorCode);
                addContraction = FALSE;
            }
            continue;
        }
        if(addContraction) {
            addContractionEntry(prevX, ce0, ce1, errorCode);
        }
        ce32 = (uint32_t)suffixes.getValue();
        if(suffix.length() == 1 && getCEsFromCE32(data, U_SENTINEL, ce32, errorCode)) {
            addContraction = TRUE;
        } else {
            addContractionEntry(x, Collation::NO_CE, 0, errorCode);
            addContraction = FALSE;
        }
        prevX = x;
    }
    if(addContraction) {
        addContractionEntry(prevX, ce0, ce1, errorCode);
    }
    if(U_FAILURE(errorCode)) { return FALSE; }
    // Note: There might not be any fast Latin contractions, but
    // we need to enter contraction handling anyway so that we can bail out
    // when there is a non-fast-Latin character following.
    // For example: Danish &Y<<u+umlaut, when we compare Y vs. u\u0308 we need to see the
    // following umlaut and bail out, rather than return the difference of Y vs. u.
    ce0 = ((int64_t)Collation::NO_CE_PRIMARY << 32) | CONTRACTION_FLAG | contractionIndex;
    ce1 = 0;
    return TRUE;
}

void
CollationFastLatinBuilder::addContractionEntry(int32_t x, int64_t cce0, int64_t cce1,
                                               UErrorCode &errorCode) {
    contractionCEs.addElement(x, errorCode);
    contractionCEs.addElement(cce0, errorCode);
    contractionCEs.addElement(cce1, errorCode);
    addUniqueCE(cce0, errorCode);
    addUniqueCE(cce1, errorCode);
}

void
CollationFastLatinBuilder::addUniqueCE(int64_t ce, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(ce == 0 || (uint32_t)(ce >> 32) == Collation::NO_CE_PRIMARY) { return; }
    ce &= ~(int64_t)Collation::CASE_MASK;  // blank out case bits
    int32_t i = binarySearch(uniqueCEs.getBuffer(), uniqueCEs.size(), ce);
    if(i < 0) {
        uniqueCEs.insertElementAt(ce, ~i, errorCode);
    }
}

uint32_t
CollationFastLatinBuilder::getMiniCE(int64_t ce) const {
    ce &= ~(int64_t)Collation::CASE_MASK;  // blank out case bits
    int32_t index = binarySearch(uniqueCEs.getBuffer(), uniqueCEs.size(), ce);
    U_ASSERT(index >= 0);
    return miniCEs[index];
}

UBool
CollationFastLatinBuilder::encodeUniqueCEs(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    uprv_free(miniCEs);
    miniCEs = (uint16_t *)uprv_malloc(uniqueCEs.size() * 2);
    if(miniCEs == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    int32_t group = 0;
    uint32_t lastGroupPrimary = lastSpecialPrimaries[group];
    // The lowest unique CE must be at least a secondary CE.
    U_ASSERT(((uint32_t)uniqueCEs.elementAti(0) >> 16) != 0);
    uint32_t prevPrimary = 0;
    uint32_t prevSecondary = 0;
    uint32_t pri = 0;
    uint32_t sec = 0;
    uint32_t ter = CollationFastLatin::COMMON_TER;
    for(int32_t i = 0; i < uniqueCEs.size(); ++i) {
        int64_t ce = uniqueCEs.elementAti(i);
        // Note: At least one of the p/s/t weights changes from one unique CE to the next.
        // (uniqueCEs does not store case bits.)
        uint32_t p = (uint32_t)(ce >> 32);
        if(p != prevPrimary) {
            while(p > lastGroupPrimary) {
                U_ASSERT(pri <= CollationFastLatin::MAX_LONG);
                // Set the group's header entry to the
                // last "long primary" in or before the group.
                result.setCharAt(1 + group, (UChar)pri);
                if(++group < NUM_SPECIAL_GROUPS) {
                    lastGroupPrimary = lastSpecialPrimaries[group];
                } else {
                    lastGroupPrimary = 0xffffffff;
                    break;
                }
            }
            if(p < firstShortPrimary) {
                if(pri == 0) {
                    pri = CollationFastLatin::MIN_LONG;
                } else if(pri < CollationFastLatin::MAX_LONG) {
                    pri += CollationFastLatin::LONG_INC;
                } else {
#if DEBUG_COLLATION_FAST_LATIN_BUILDER
                    printf("long-primary overflow for %08x\n", p);
#endif
                    miniCEs[i] = CollationFastLatin::BAIL_OUT;
                    continue;
                }
            } else {
                if(pri < CollationFastLatin::MIN_SHORT) {
                    pri = CollationFastLatin::MIN_SHORT;
                } else if(pri < (CollationFastLatin::MAX_SHORT - CollationFastLatin::SHORT_INC)) {
                    // Reserve the highest primary weight for U+FFFF.
                    pri += CollationFastLatin::SHORT_INC;
                } else {
#if DEBUG_COLLATION_FAST_LATIN_BUILDER
                    printf("short-primary overflow for %08x\n", p);
#endif
                    shortPrimaryOverflow = TRUE;
                    miniCEs[i] = CollationFastLatin::BAIL_OUT;
                    continue;
                }
            }
            prevPrimary = p;
            prevSecondary = Collation::COMMON_WEIGHT16;
            sec = CollationFastLatin::COMMON_SEC;
            ter = CollationFastLatin::COMMON_TER;
        }
        uint32_t lower32 = (uint32_t)ce;
        uint32_t s = lower32 >> 16;
        if(s != prevSecondary) {
            if(pri == 0) {
                if(sec == 0) {
                    sec = CollationFastLatin::MIN_SEC_HIGH;
                } else if(sec < CollationFastLatin::MAX_SEC_HIGH) {
                    sec += CollationFastLatin::SEC_INC;
                } else {
                    miniCEs[i] = CollationFastLatin::BAIL_OUT;
                    continue;
                }
                prevSecondary = s;
                ter = CollationFastLatin::COMMON_TER;
            } else if(s < Collation::COMMON_WEIGHT16) {
                if(sec == CollationFastLatin::COMMON_SEC) {
                    sec = CollationFastLatin::MIN_SEC_BEFORE;
                } else if(sec < CollationFastLatin::MAX_SEC_BEFORE) {
                    sec += CollationFastLatin::SEC_INC;
                } else {
                    miniCEs[i] = CollationFastLatin::BAIL_OUT;
                    continue;
                }
            } else if(s == Collation::COMMON_WEIGHT16) {
                sec = CollationFastLatin::COMMON_SEC;
            } else {
                if(sec < CollationFastLatin::MIN_SEC_AFTER) {
                    sec = CollationFastLatin::MIN_SEC_AFTER;
                } else if(sec < CollationFastLatin::MAX_SEC_AFTER) {
                    sec += CollationFastLatin::SEC_INC;
                } else {
                    miniCEs[i] = CollationFastLatin::BAIL_OUT;
                    continue;
                }
            }
            prevSecondary = s;
            ter = CollationFastLatin::COMMON_TER;
        }
        U_ASSERT((lower32 & Collation::CASE_MASK) == 0);  // blanked out in uniqueCEs
        uint32_t t = lower32 & Collation::ONLY_TERTIARY_MASK;
        if(t > Collation::COMMON_WEIGHT16) {
            if(ter < CollationFastLatin::MAX_TER_AFTER) {
                ++ter;
            } else {
                miniCEs[i] = CollationFastLatin::BAIL_OUT;
                continue;
            }
        }
        if(CollationFastLatin::MIN_LONG <= pri && pri <= CollationFastLatin::MAX_LONG) {
            U_ASSERT(sec == CollationFastLatin::COMMON_SEC);
            miniCEs[i] = (uint16_t)(pri | ter);
        } else {
            miniCEs[i] = (uint16_t)(pri | sec | ter);
        }
    }
#if DEBUG_COLLATION_FAST_LATIN_BUILDER
    printf("last mini primary: %04x\n", pri);
#endif
#if DEBUG_COLLATION_FAST_LATIN_BUILDER >= 2
    for(int32_t i = 0; i < uniqueCEs.size(); ++i) {
        int64_t ce = uniqueCEs.elementAti(i);
        printf("unique CE 0x%016lx -> 0x%04x\n", ce, miniCEs[i]);
    }
#endif
    return U_SUCCESS(errorCode);
}

UBool
CollationFastLatinBuilder::encodeCharCEs(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    int32_t miniCEsStart = result.length();
    for(int32_t i = 0; i < CollationFastLatin::NUM_FAST_CHARS; ++i) {
        result.append((UChar)0);  // initialize to completely ignorable
    }
    int32_t indexBase = result.length();
    for(int32_t i = 0; i < CollationFastLatin::NUM_FAST_CHARS; ++i) {
        int64_t ce = charCEs[i][0];
        if(isContractionCharCE(ce)) { continue; }  // defer contraction
        uint32_t miniCE = encodeTwoCEs(ce, charCEs[i][1]);
        if(miniCE > 0xffff) {
            // Note: There is a chance that this new expansion is the same as a previous one,
            // and if so, then we could reuse the other expansion.
            // However, that seems unlikely.
            int32_t expansionIndex = result.length() - indexBase;
            if(expansionIndex > (int32_t)CollationFastLatin::INDEX_MASK) {
                miniCE = CollationFastLatin::BAIL_OUT;
            } else {
                result.append((UChar)(miniCE >> 16)).append((UChar)miniCE);
                miniCE = CollationFastLatin::EXPANSION | expansionIndex;
            }
        }
        result.setCharAt(miniCEsStart + i, (UChar)miniCE);
    }
    return U_SUCCESS(errorCode);
}

UBool
CollationFastLatinBuilder::encodeContractions(UErrorCode &errorCode) {
    // We encode all contraction lists so that the first word of a list
    // terminates the previous list, and we only need one additional terminator at the end.
    if(U_FAILURE(errorCode)) { return FALSE; }
    int32_t indexBase = headerLength + CollationFastLatin::NUM_FAST_CHARS;
    int32_t firstContractionIndex = result.length();
    for(int32_t i = 0; i < CollationFastLatin::NUM_FAST_CHARS; ++i) {
        int64_t ce = charCEs[i][0];
        if(!isContractionCharCE(ce)) { continue; }
        int32_t contractionIndex = result.length() - indexBase;
        if(contractionIndex > (int32_t)CollationFastLatin::INDEX_MASK) {
            result.setCharAt(headerLength + i, CollationFastLatin::BAIL_OUT);
            continue;
        }
        UBool firstTriple = TRUE;
        for(int32_t index = (int32_t)ce & 0x7fffffff;; index += 3) {
            int32_t x = static_cast<int32_t>(contractionCEs.elementAti(index));
            if((uint32_t)x == CollationFastLatin::CONTR_CHAR_MASK && !firstTriple) { break; }
            int64_t cce0 = contractionCEs.elementAti(index + 1);
            int64_t cce1 = contractionCEs.elementAti(index + 2);
            uint32_t miniCE = encodeTwoCEs(cce0, cce1);
            if(miniCE == CollationFastLatin::BAIL_OUT) {
                result.append((UChar)(x | (1 << CollationFastLatin::CONTR_LENGTH_SHIFT)));
            } else if(miniCE <= 0xffff) {
                result.append((UChar)(x | (2 << CollationFastLatin::CONTR_LENGTH_SHIFT)));
                result.append((UChar)miniCE);
            } else {
                result.append((UChar)(x | (3 << CollationFastLatin::CONTR_LENGTH_SHIFT)));
                result.append((UChar)(miniCE >> 16)).append((UChar)miniCE);
            }
            firstTriple = FALSE;
        }
        // Note: There is a chance that this new contraction list is the same as a previous one,
        // and if so, then we could truncate the result and reuse the other list.
        // However, that seems unlikely.
        result.setCharAt(headerLength + i,
                         (UChar)(CollationFastLatin::CONTRACTION | contractionIndex));
    }
    if(result.length() > firstContractionIndex) {
        // Terminate the last contraction list.
        result.append((UChar)CollationFastLatin::CONTR_CHAR_MASK);
    }
    if(result.isBogus()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
#if DEBUG_COLLATION_FAST_LATIN_BUILDER
    printf("** fast Latin %d * 2 = %d bytes\n", result.length(), result.length() * 2);
    puts("   header & below-digit groups map");
    int32_t i = 0;
    for(; i < headerLength; ++i) {
        printf(" %04x", result[i]);
    }
    printf("\n   char mini CEs");
    U_ASSERT(CollationFastLatin::NUM_FAST_CHARS % 16 == 0);
    for(; i < indexBase; i += 16) {
        UChar32 c = i - headerLength;
        if(c >= CollationFastLatin::LATIN_LIMIT) {
            c = CollationFastLatin::PUNCT_START + c - CollationFastLatin::LATIN_LIMIT;
        }
        printf("\n %04x:", c);
        for(int32_t j = 0; j < 16; ++j) {
            printf(" %04x", result[i + j]);
        }
    }
    printf("\n   expansions & contractions");
    for(; i < result.length(); ++i) {
        if((i - indexBase) % 16 == 0) { puts(""); }
        printf(" %04x", result[i]);
    }
    puts("");
#endif
    return TRUE;
}

uint32_t
CollationFastLatinBuilder::encodeTwoCEs(int64_t first, int64_t second) const {
    if(first == 0) {
        return 0;  // completely ignorable
    }
    if(first == Collation::NO_CE) {
        return CollationFastLatin::BAIL_OUT;
    }
    U_ASSERT((uint32_t)(first >> 32) != Collation::NO_CE_PRIMARY);

    uint32_t miniCE = getMiniCE(first);
    if(miniCE == CollationFastLatin::BAIL_OUT) { return miniCE; }
    if(miniCE >= CollationFastLatin::MIN_SHORT) {
        // Extract & copy the case bits.
        // Shift them from normal CE bits 15..14 to mini CE bits 4..3.
        uint32_t c = (((uint32_t)first & Collation::CASE_MASK) >> (14 - 3));
        // Only in mini CEs: Ignorable case bits = 0, lowercase = 1.
        c += CollationFastLatin::LOWER_CASE;
        miniCE |= c;
    }
    if(second == 0) { return miniCE; }

    uint32_t miniCE1 = getMiniCE(second);
    if(miniCE1 == CollationFastLatin::BAIL_OUT) { return miniCE1; }

    uint32_t case1 = (uint32_t)second & Collation::CASE_MASK;
    if(miniCE >= CollationFastLatin::MIN_SHORT &&
            (miniCE & CollationFastLatin::SECONDARY_MASK) == CollationFastLatin::COMMON_SEC) {
        // Try to combine the two mini CEs into one.
        uint32_t sec1 = miniCE1 & CollationFastLatin::SECONDARY_MASK;
        uint32_t ter1 = miniCE1 & CollationFastLatin::TERTIARY_MASK;
        if(sec1 >= CollationFastLatin::MIN_SEC_HIGH && case1 == 0 &&
                ter1 == CollationFastLatin::COMMON_TER) {
            // sec1>=sec_high implies pri1==0.
            return (miniCE & ~CollationFastLatin::SECONDARY_MASK) | sec1;
        }
    }

    if(miniCE1 <= CollationFastLatin::SECONDARY_MASK || CollationFastLatin::MIN_SHORT <= miniCE1) {
        // Secondary CE, or a CE with a short primary, copy the case bits.
        case1 = (case1 >> (14 - 3)) + CollationFastLatin::LOWER_CASE;
        miniCE1 |= case1;
    }
    return (miniCE << 16) | miniCE1;
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// Copyright (C) 1999-2016, International Business Machines
// Corporation and others.  All Rights Reserved.
//
// file name: collationfcd.cpp
//
// machine-generated by: icu/tools/unicode/c/genuca/genuca.cpp




#if !UCONFIG_NO_COLLATION

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationfcd.h
*
* created on: 2012aug18
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION



U_NAMESPACE_BEGIN

/**
 * Data and functions for the FCD check fast path.
 *
 * The fast path looks at a pair of 16-bit code units and checks
 * whether there is an FCD boundary between them;
 * there is if the first unit has a trailing ccc=0 (!hasTccc(first))
 * or the second unit has a leading ccc=0 (!hasLccc(second)),
 * or both.
 * When the fast path finds a possible non-boundary,
 * then the FCD check slow path looks at the actual sequence of FCD values.
 *
 * This is a pure optimization.
 * The fast path must at least find all possible non-boundaries.
 * If the fast path is too pessimistic, it costs performance.
 *
 * For a pair of BMP characters, the fast path tests are precise (1 bit per character).
 *
 * For a supplementary code point, the two units are its lead and trail surrogates.
 * We set hasTccc(lead)=true if any of its 1024 associated supplementary code points
 * has lccc!=0 or tccc!=0.
 * We set hasLccc(trail)=true for all trail surrogates.
 * As a result, we leave the fast path if the lead surrogate might start a
 * supplementary code point that is not FCD-inert.
 * (So the fast path need not detect that there is a surrogate pair,
 * nor look ahead to the next full code point.)
 *
 * hasLccc(lead)=true if any of its 1024 associated supplementary code points
 * has lccc!=0, for fast boundary checking between BMP & supplementary.
 *
 * hasTccc(trail)=false:
 * It should only be tested for unpaired trail surrogates which are FCD-inert.
 */
class U_I18N_API CollationFCD {
public:
    static inline UBool hasLccc(UChar32 c) {
        // assert c <= 0xffff
        // c can be negative, e.g., U_SENTINEL from UCharIterator;
        // that is handled in the first test.
        int32_t i;
        return
            // U+0300 is the first character with lccc!=0.
            c >= 0x300 &&
            (i = lcccIndex[c >> 5]) != 0 &&
            (lcccBits[i] & ((uint32_t)1 << (c & 0x1f))) != 0;
    }

    static inline UBool hasTccc(UChar32 c) {
        // assert c <= 0xffff
        // c can be negative, e.g., U_SENTINEL from UCharIterator;
        // that is handled in the first test.
        int32_t i;
        return
            // U+00C0 is the first character with tccc!=0.
            c >= 0xc0 &&
            (i = tcccIndex[c >> 5]) != 0 &&
            (tcccBits[i] & ((uint32_t)1 << (c & 0x1f))) != 0;
    }

    static inline UBool mayHaveLccc(UChar32 c) {
        // Handles all of Unicode 0..10FFFF.
        // c can be negative, e.g., U_SENTINEL.
        // U+0300 is the first character with lccc!=0.
        if(c < 0x300) { return FALSE; }
        if(c > 0xffff) { c = U16_LEAD(c); }
        int32_t i;
        return
            (i = lcccIndex[c >> 5]) != 0 &&
            (lcccBits[i] & ((uint32_t)1 << (c & 0x1f))) != 0;
    }

    /**
     * Tibetan composite vowel signs (U+0F73, U+0F75, U+0F81)
     * must be decomposed before reaching the core collation code,
     * or else some sequences including them, even ones passing the FCD check,
     * do not yield canonically equivalent results.
     *
     * This is a fast and imprecise test.
     *
     * @param c a code point
     * @return TRUE if c is U+0F73, U+0F75 or U+0F81 or one of several other Tibetan characters
     */
    static inline UBool maybeTibetanCompositeVowel(UChar32 c) {
        return (c & 0x1fff01) == 0xf01;
    }

    /**
     * Tibetan composite vowel signs (U+0F73, U+0F75, U+0F81)
     * must be decomposed before reaching the core collation code,
     * or else some sequences including them, even ones passing the FCD check,
     * do not yield canonically equivalent results.
     *
     * They have distinct lccc/tccc combinations: 129/130 or 129/132.
     *
     * @param fcd16 the FCD value (lccc/tccc combination) of a code point
     * @return TRUE if fcd16 is from U+0F73, U+0F75 or U+0F81
     */
    static inline UBool isFCD16OfTibetanCompositeVowel(uint16_t fcd16) {
        return fcd16 == 0x8182 || fcd16 == 0x8184;
    }

private:
    CollationFCD();  // No instantiation.

    static const uint8_t lcccIndex[2048];
    static const uint8_t tcccIndex[2048];
    static const uint32_t lcccBits[];
    static const uint32_t tcccBits[];
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION


U_NAMESPACE_BEGIN

const uint8_t CollationFCD::lcccIndex[2048]={
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1,1,2,3,0,0,0,0,
0,0,0,0,4,0,0,0,0,0,0,0,5,6,7,0,
8,0,9,0xa,0,0,0xb,0xc,0xd,0xe,0xf,0,0,0,0,0x10,
0x11,0x12,0x13,0,0,0,0x14,0x15,0,0x16,0x17,0,0,0x16,0x18,0x19,
0,0x16,0x18,0,0,0x16,0x18,0,0,0x16,0x18,0,0,0,0x18,0,
0,0,0x1a,0,0,0x16,0x18,0,0,0x1b,0x18,0,0,0,0x1c,0,
0,0x1d,0x1e,0,0,0x1d,0x1e,0,0x1f,0x20,0,0x21,0x22,0,0x23,0,
0,0x24,0,0,0x18,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0x25,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x26,0x26,0,0,0,0,0x27,0,
0,0,0,0,0,0x28,0,0,0,0x13,0,0,0,0,0,0,
0x29,0,0,0x2a,0,0x2b,0x2c,0,0,0x26,0x2d,0x2e,0,0x2f,0,0x30,
0,0x31,0,0,0,0,0x32,0x33,0,0,0,0,0,0,1,0x34,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0x35,0x36,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0x37,0,0,0,0x38,0,0,0,1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0x39,0,0,0x3a,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0x3b,0x3c,0,0,0x3d,0,0,0,0,0,0,0,0,
0x23,0x3e,0,0,0,0,0x2d,0x3f,0,0x40,0x41,0,0,0x41,0x2c,0,
0,0,0,0,0,0x42,0x43,0x44,0,0,0,0,0,0,0,0x18,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x45,0x46,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x19,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};

const uint32_t CollationFCD::lcccBits[71]={
0,0xffffffff,0xffff7fff,0xffff,0xf8,0xfffe0000,0xbfffffff,0xb6,0x7ff0000,0xfffff800,0x10000,0x9fc00000,0x3d9f,0x20000,0xffff0000,0x7ff,
0x200ff800,0xfbc00000,0x3eef,0xe000000,0xfff80000,0xfffffffb,0x10000000,0x1e2000,0x2000,0x40000000,0x602000,0x18000000,0x400,0x7000000,0xf00,0x3000000,
0x2a00000,0x3c3e0000,0xdf,0x40,0x6800000,0xe0000000,0x100000,0x20040000,0x200,0x1800000,0x9fe00001,0xbfff0000,1,0x10,0xff800,0xc00,
0xc0040,0x800000,0xfff70000,0x31021fd,0xfbffffff,0x1fff0000,0x1ffe2,0x38000,0x80000000,0xfc00,0x6000000,0x3ff08000,0xc0000000,0x30000,0x1000,0x3ffff,
0x3800,0x80000,0xc19d0000,2,0x400000,0xc0000fd,0x5108000
};

const uint8_t CollationFCD::tcccIndex[2048]={
0,0,0,0,0,0,2,3,4,5,6,7,0,8,9,0xa,
0xb,0xc,0,0,0,0,0,0,1,1,0xd,0xe,0xf,0x10,0x11,0,
0x12,0x13,0x14,0x15,0x16,0,0x17,0x18,0,0,0,0,0x19,0x1a,0x1b,0,
0x1c,0x1d,0x1e,0x1f,0,0,0x20,0x21,0x22,0x23,0x24,0,0,0,0,0x25,
0x26,0x27,0x28,0,0,0,0x29,0x2a,0,0x2b,0x2c,0,0,0x2d,0x2e,0x2f,
0,0x30,0x31,0,0,0x2d,0x32,0,0,0x2d,0x33,0,0,0,0x32,0,
0,0,0x34,0,0,0x2d,0x32,0,0,0x35,0x32,0,0,0,0x36,0,
0,0x37,0x38,0,0,0x37,0x38,0,0x39,0x3a,0,0x3b,0x3c,0,0x3d,0,
0,0x3e,0,0,0x32,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0x3f,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x40,0x40,0,0,0,0,0x41,0,
0,0,0,0,0,0x42,0,0,0,0x28,0,0,0,0,0,0,
0x43,0,0,0x44,0,0x45,0x46,0,0,0x40,0x47,0x48,0,0x49,0,0x4a,
0,0x4b,0,0,0,0,0x4c,0x4d,0,0,0,0,0,0,1,0x4e,
1,1,1,1,0x4f,1,1,0x50,0x51,1,0x52,0x53,1,0x54,0x55,0x56,
0,0,0,0,0,0,0x57,0x58,0,0x59,0,0,0x5a,0x5b,0x5c,0,
0x5d,0x5e,0x5f,0x60,0x61,0x62,0,0x63,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0x2d,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0x64,0,0,0,0x65,0,0,0,1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0x66,0x67,0x68,0x69,0x67,0x68,0x6a,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0x6b,0x6c,0,0,0x6d,0,0,0,0,0,0,0,0,
0x3d,0x6e,0,0,0,0,0x47,0x6f,0,0x70,0x71,0,0,0x71,0x46,0,
0,0,0,0,0,0x72,0x73,0x74,0,0,0,0,0,0,0,0x32,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0x75,0x76,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0x3f,0x77,0x78,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0xe,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};

const uint32_t CollationFCD::tcccBits[121]={
0,0xffffffff,0x3e7effbf,0xbe7effbf,0xfffcffff,0x7ef1ff3f,0xfff3f1f8,0x7fffff3f,0x18003,0xdfffe000,0xff31ffcf,0xcfffffff,0xfffc0,0xffff7fff,0xffff,0x1d760,
0x1fc00,0x187c00,0x200708b,0x2000000,0x708b0000,0xc00000,0xf8,0xfccf0006,0x33ffcfc,0xfffe0000,0xbfffffff,0xb6,0x7ff0000,0x7c,0xfffff800,0x10000,
0x9fc80005,0x3d9f,0x20000,0xffff0000,0x7ff,0x200ff800,0xfbc00000,0x3eef,0xe000000,0xfff80000,0xfffffffb,0x10120200,0xff1e2000,0x10000000,0xb0002000,0x40000000,
0x10480000,0x4e002000,0x2000,0x30002000,0x602100,0x18000000,0x24000400,0x7000000,0xf00,0x3000000,0x2a00000,0x3d7e0000,0xdf,0x40,0x6800000,0xe0000000,
0x100000,0x20040000,0x200,0x1800000,0x9fe00001,0xbfff0000,1,0x10,0xff800,0xc00,0xc0040,0x800000,0xfff70000,0x31021fd,0xfbffffff,0xbffffff,
0x3ffffff,0x3f3fffff,0xaaff3f3f,0x3fffffff,0x1fdfffff,0xefcfffde,0x1fdc7fff,0x1fff0000,0x1ffe2,0x800,0xc000000,0x4000,0xe000,0x1210,0x50,0x292,
0x333e005,0x333,0xf000,0x3c0f,0x38000,0x80000000,0xfc00,0x55555000,0x36db02a5,0x46100000,0x47900000,0x3ff08000,0xc0000000,0x30000,0x1000,0x3ffff,
0x3800,0x80000,0xc19d0000,2,0x400000,0xc0000fd,0x5108000,0x5f7ffc00,0x7fdb
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationiterator.cpp
*
* created on: 2010oct27
* created by: Markus W. Scherer
*/

  // for 'typeid' to work



#if !UCONFIG_NO_COLLATION













U_NAMESPACE_BEGIN

CollationIterator::CEBuffer::~CEBuffer() {}

UBool
CollationIterator::CEBuffer::ensureAppendCapacity(int32_t appCap, UErrorCode &errorCode) {
    int32_t capacity = buffer.getCapacity();
    if((length + appCap) <= capacity) { return TRUE; }
    if(U_FAILURE(errorCode)) { return FALSE; }
    do {
        if(capacity < 1000) {
            capacity *= 4;
        } else {
            capacity *= 2;
        }
    } while(capacity < (length + appCap));
    int64_t *p = buffer.resize(capacity, length);
    if(p == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    return TRUE;
}

// State of combining marks skipped in discontiguous contraction.
// We create a state object on first use and keep it around deactivated between uses.
class SkippedState : public UMemory {
public:
    // Born active but empty.
    SkippedState() : pos(0), skipLengthAtMatch(0) {}
    void clear() {
        oldBuffer.remove();
        pos = 0;
        // The newBuffer is reset by setFirstSkipped().
    }

    UBool isEmpty() const { return oldBuffer.isEmpty(); }

    UBool hasNext() const { return pos < oldBuffer.length(); }

    // Requires hasNext().
    UChar32 next() {
        UChar32 c = oldBuffer.char32At(pos);
        pos += U16_LENGTH(c);
        return c;
    }

    // Accounts for one more input code point read beyond the end of the marks buffer.
    void incBeyond() {
        U_ASSERT(!hasNext());
        ++pos;
    }

    // Goes backward through the skipped-marks buffer.
    // Returns the number of code points read beyond the skipped marks
    // that need to be backtracked through normal input.
    int32_t backwardNumCodePoints(int32_t n) {
        int32_t length = oldBuffer.length();
        int32_t beyond = pos - length;
        if(beyond > 0) {
            if(beyond >= n) {
                // Not back far enough to re-enter the oldBuffer.
                pos -= n;
                return n;
            } else {
                // Back out all beyond-oldBuffer code points and re-enter the buffer.
                pos = oldBuffer.moveIndex32(length, beyond - n);
                return beyond;
            }
        } else {
            // Go backwards from inside the oldBuffer.
            pos = oldBuffer.moveIndex32(pos, -n);
            return 0;
        }
    }

    void setFirstSkipped(UChar32 c) {
        skipLengthAtMatch = 0;
        newBuffer.setTo(c);
    }

    void skip(UChar32 c) {
        newBuffer.append(c);
    }

    void recordMatch() { skipLengthAtMatch = newBuffer.length(); }

    // Replaces the characters we consumed with the newly skipped ones.
    void replaceMatch() {
        // Note: UnicodeString.replace() pins pos to at most length().
        oldBuffer.replace(0, pos, newBuffer, 0, skipLengthAtMatch);
        pos = 0;
    }

    void saveTrieState(const UCharsTrie &trie) { trie.saveState(state); }
    void resetToTrieState(UCharsTrie &trie) const { trie.resetToState(state); }

private:
    // Combining marks skipped in previous discontiguous-contraction matching.
    // After that discontiguous contraction was completed, we start reading them from here.
    UnicodeString oldBuffer;
    // Combining marks newly skipped in current discontiguous-contraction matching.
    // These might have been read from the normal text or from the oldBuffer.
    UnicodeString newBuffer;
    // Reading index in oldBuffer,
    // or counter for how many code points have been read beyond oldBuffer (pos-oldBuffer.length()).
    int32_t pos;
    // newBuffer.length() at the time of the last matching character.
    // When a partial match fails, we back out skipped and partial-matching input characters.
    int32_t skipLengthAtMatch;
    // We save the trie state before we attempt to match a character,
    // so that we can skip it and try the next one.
    UCharsTrie::State state;
};

CollationIterator::CollationIterator(const CollationIterator &other)
        : UObject(other),
          trie(other.trie),
          data(other.data),
          cesIndex(other.cesIndex),
          skipped(NULL),
          numCpFwd(other.numCpFwd),
          isNumeric(other.isNumeric) {
    UErrorCode errorCode = U_ZERO_ERROR;
    int32_t length = other.ceBuffer.length;
    if(length > 0 && ceBuffer.ensureAppendCapacity(length, errorCode)) {
        for(int32_t i = 0; i < length; ++i) {
            ceBuffer.set(i, other.ceBuffer.get(i));
        }
        ceBuffer.length = length;
    } else {
        cesIndex = 0;
    }
}

CollationIterator::~CollationIterator() {
    delete skipped;
}

UBool
CollationIterator::operator==(const CollationIterator &other) const {
    // Subclasses: Call this method and then add more specific checks.
    // Compare the iterator state but not the collation data (trie & data fields):
    // Assume that the caller compares the data.
    // Ignore skipped since that should be unused between calls to nextCE().
    // (It only stays around to avoid another memory allocation.)
    if(!(typeid(*this) == typeid(other) &&
            ceBuffer.length == other.ceBuffer.length &&
            cesIndex == other.cesIndex &&
            numCpFwd == other.numCpFwd &&
            isNumeric == other.isNumeric)) {
        return FALSE;
    }
    for(int32_t i = 0; i < ceBuffer.length; ++i) {
        if(ceBuffer.get(i) != other.ceBuffer.get(i)) { return FALSE; }
    }
    return TRUE;
}

void
CollationIterator::reset() {
    cesIndex = ceBuffer.length = 0;
    if(skipped != NULL) { skipped->clear(); }
}

int32_t
CollationIterator::fetchCEs(UErrorCode &errorCode) {
    while(U_SUCCESS(errorCode) && nextCE(errorCode) != Collation::NO_CE) {
        // No need to loop for each expansion CE.
        cesIndex = ceBuffer.length;
    }
    return ceBuffer.length;
}

uint32_t
CollationIterator::handleNextCE32(UChar32 &c, UErrorCode &errorCode) {
    c = nextCodePoint(errorCode);
    return (c < 0) ? Collation::FALLBACK_CE32 : data->getCE32(c);
}

UChar
CollationIterator::handleGetTrailSurrogate() {
    return 0;
}

UBool
CollationIterator::foundNULTerminator() {
    return FALSE;
}

UBool
CollationIterator::forbidSurrogateCodePoints() const {
    return FALSE;
}

uint32_t
CollationIterator::getDataCE32(UChar32 c) const {
    return data->getCE32(c);
}

uint32_t
CollationIterator::getCE32FromBuilderData(uint32_t /*ce32*/, UErrorCode &errorCode) {
    if(U_SUCCESS(errorCode)) { errorCode = U_INTERNAL_PROGRAM_ERROR; }
    return 0;
}

int64_t
CollationIterator::nextCEFromCE32(const CollationData *d, UChar32 c, uint32_t ce32,
                                  UErrorCode &errorCode) {
    --ceBuffer.length;  // Undo ceBuffer.incLength().
    appendCEsFromCE32(d, c, ce32, TRUE, errorCode);
    if(U_SUCCESS(errorCode)) {
        return ceBuffer.get(cesIndex++);
    } else {
        return Collation::NO_CE_PRIMARY;
    }
}

void
CollationIterator::appendCEsFromCE32(const CollationData *d, UChar32 c, uint32_t ce32,
                                     UBool forward, UErrorCode &errorCode) {
    while(Collation::isSpecialCE32(ce32)) {
        switch(Collation::tagFromCE32(ce32)) {
        case Collation::FALLBACK_TAG:
        case Collation::RESERVED_TAG_3:
            if(U_SUCCESS(errorCode)) { errorCode = U_INTERNAL_PROGRAM_ERROR; }
            return;
        case Collation::LONG_PRIMARY_TAG:
            ceBuffer.append(Collation::ceFromLongPrimaryCE32(ce32), errorCode);
            return;
        case Collation::LONG_SECONDARY_TAG:
            ceBuffer.append(Collation::ceFromLongSecondaryCE32(ce32), errorCode);
            return;
        case Collation::LATIN_EXPANSION_TAG:
            if(ceBuffer.ensureAppendCapacity(2, errorCode)) {
                ceBuffer.set(ceBuffer.length, Collation::latinCE0FromCE32(ce32));
                ceBuffer.set(ceBuffer.length + 1, Collation::latinCE1FromCE32(ce32));
                ceBuffer.length += 2;
            }
            return;
        case Collation::EXPANSION32_TAG: {
            const uint32_t *ce32s = d->ce32s + Collation::indexFromCE32(ce32);
            int32_t length = Collation::lengthFromCE32(ce32);
            if(ceBuffer.ensureAppendCapacity(length, errorCode)) {
                do {
                    ceBuffer.appendUnsafe(Collation::ceFromCE32(*ce32s++));
                } while(--length > 0);
            }
            return;
        }
        case Collation::EXPANSION_TAG: {
            const int64_t *ces = d->ces + Collation::indexFromCE32(ce32);
            int32_t length = Collation::lengthFromCE32(ce32);
            if(ceBuffer.ensureAppendCapacity(length, errorCode)) {
                do {
                    ceBuffer.appendUnsafe(*ces++);
                } while(--length > 0);
            }
            return;
        }
        case Collation::BUILDER_DATA_TAG:
            ce32 = getCE32FromBuilderData(ce32, errorCode);
            if(U_FAILURE(errorCode)) { return; }
            if(ce32 == Collation::FALLBACK_CE32) {
                d = data->base;
                ce32 = d->getCE32(c);
            }
            break;
        case Collation::PREFIX_TAG:
            if(forward) { backwardNumCodePoints(1, errorCode); }
            ce32 = getCE32FromPrefix(d, ce32, errorCode);
            if(forward) { forwardNumCodePoints(1, errorCode); }
            break;
        case Collation::CONTRACTION_TAG: {
            const UChar *p = d->contexts + Collation::indexFromCE32(ce32);
            uint32_t defaultCE32 = CollationData::readCE32(p);  // Default if no suffix match.
            if(!forward) {
                // Backward contractions are handled by previousCEUnsafe().
                // c has contractions but they were not found.
                ce32 = defaultCE32;
                break;
            }
            UChar32 nextCp;
            if(skipped == NULL && numCpFwd < 0) {
                // Some portion of nextCE32FromContraction() pulled out here as an ASCII fast path,
                // avoiding the function call and the nextSkippedCodePoint() overhead.
                nextCp = nextCodePoint(errorCode);
                if(nextCp < 0) {
                    // No more text.
                    ce32 = defaultCE32;
                    break;
                } else if((ce32 & Collation::CONTRACT_NEXT_CCC) != 0 &&
                        !CollationFCD::mayHaveLccc(nextCp)) {
                    // All contraction suffixes start with characters with lccc!=0
                    // but the next code point has lccc==0.
                    backwardNumCodePoints(1, errorCode);
                    ce32 = defaultCE32;
                    break;
                }
            } else {
                nextCp = nextSkippedCodePoint(errorCode);
                if(nextCp < 0) {
                    // No more text.
                    ce32 = defaultCE32;
                    break;
                } else if((ce32 & Collation::CONTRACT_NEXT_CCC) != 0 &&
                        !CollationFCD::mayHaveLccc(nextCp)) {
                    // All contraction suffixes start with characters with lccc!=0
                    // but the next code point has lccc==0.
                    backwardNumSkipped(1, errorCode);
                    ce32 = defaultCE32;
                    break;
                }
            }
            ce32 = nextCE32FromContraction(d, ce32, p + 2, defaultCE32, nextCp, errorCode);
            if(ce32 == Collation::NO_CE32) {
                // CEs from a discontiguous contraction plus the skipped combining marks
                // have been appended already.
                return;
            }
            break;
        }
        case Collation::DIGIT_TAG:
            if(isNumeric) {
                appendNumericCEs(ce32, forward, errorCode);
                return;
            } else {
                // Fetch the non-numeric-collation CE32 and continue.
                ce32 = d->ce32s[Collation::indexFromCE32(ce32)];
                break;
            }
        case Collation::U0000_TAG:
            U_ASSERT(c == 0);
            if(forward && foundNULTerminator()) {
                // Handle NUL-termination. (Not needed in Java.)
                ceBuffer.append(Collation::NO_CE, errorCode);
                return;
            } else {
                // Fetch the normal ce32 for U+0000 and continue.
                ce32 = d->ce32s[0];
                break;
            }
        case Collation::HANGUL_TAG: {
            const uint32_t *jamoCE32s = d->jamoCE32s;
            c -= Hangul::HANGUL_BASE;
            UChar32 t = c % Hangul::JAMO_T_COUNT;
            c /= Hangul::JAMO_T_COUNT;
            UChar32 v = c % Hangul::JAMO_V_COUNT;
            c /= Hangul::JAMO_V_COUNT;
            if((ce32 & Collation::HANGUL_NO_SPECIAL_JAMO) != 0) {
                // None of the Jamo CE32s are isSpecialCE32().
                // Avoid recursive function calls and per-Jamo tests.
                if(ceBuffer.ensureAppendCapacity(t == 0 ? 2 : 3, errorCode)) {
                    ceBuffer.set(ceBuffer.length, Collation::ceFromCE32(jamoCE32s[c]));
                    ceBuffer.set(ceBuffer.length + 1, Collation::ceFromCE32(jamoCE32s[19 + v]));
                    ceBuffer.length += 2;
                    if(t != 0) {
                        ceBuffer.appendUnsafe(Collation::ceFromCE32(jamoCE32s[39 + t]));
                    }
                }
                return;
            } else {
                // We should not need to compute each Jamo code point.
                // In particular, there should be no offset or implicit ce32.
                appendCEsFromCE32(d, U_SENTINEL, jamoCE32s[c], forward, errorCode);
                appendCEsFromCE32(d, U_SENTINEL, jamoCE32s[19 + v], forward, errorCode);
                if(t == 0) { return; }
                // offset 39 = 19 + 21 - 1:
                // 19 = JAMO_L_COUNT
                // 21 = JAMO_T_COUNT
                // -1 = omit t==0
                ce32 = jamoCE32s[39 + t];
                c = U_SENTINEL;
                break;
            }
        }
        case Collation::LEAD_SURROGATE_TAG: {
            U_ASSERT(forward);  // Backward iteration should never see lead surrogate code _unit_ data.
            U_ASSERT(U16_IS_LEAD(c));
            UChar trail;
            if(U16_IS_TRAIL(trail = handleGetTrailSurrogate())) {
                c = U16_GET_SUPPLEMENTARY(c, trail);
                ce32 &= Collation::LEAD_TYPE_MASK;
                if(ce32 == Collation::LEAD_ALL_UNASSIGNED) {
                    ce32 = Collation::UNASSIGNED_CE32;  // unassigned-implicit
                } else if(ce32 == Collation::LEAD_ALL_FALLBACK ||
                        (ce32 = d->getCE32FromSupplementary(c)) == Collation::FALLBACK_CE32) {
                    // fall back to the base data
                    d = d->base;
                    ce32 = d->getCE32FromSupplementary(c);
                }
            } else {
                // c is an unpaired surrogate.
                ce32 = Collation::UNASSIGNED_CE32;
            }
            break;
        }
        case Collation::OFFSET_TAG:
            U_ASSERT(c >= 0);
            ceBuffer.append(d->getCEFromOffsetCE32(c, ce32), errorCode);
            return;
        case Collation::IMPLICIT_TAG:
            U_ASSERT(c >= 0);
            if(U_IS_SURROGATE(c) && forbidSurrogateCodePoints()) {
                ce32 = Collation::FFFD_CE32;
                break;
            } else {
                ceBuffer.append(Collation::unassignedCEFromCodePoint(c), errorCode);
                return;
            }
        }
    }
    ceBuffer.append(Collation::ceFromSimpleCE32(ce32), errorCode);
}

uint32_t
CollationIterator::getCE32FromPrefix(const CollationData *d, uint32_t ce32,
                                     UErrorCode &errorCode) {
    const UChar *p = d->contexts + Collation::indexFromCE32(ce32);
    ce32 = CollationData::readCE32(p);  // Default if no prefix match.
    p += 2;
    // Number of code points read before the original code point.
    int32_t lookBehind = 0;
    UCharsTrie prefixes(p);
    for(;;) {
        UChar32 c = previousCodePoint(errorCode);
        if(c < 0) { break; }
        ++lookBehind;
        UStringTrieResult match = prefixes.nextForCodePoint(c);
        if(USTRINGTRIE_HAS_VALUE(match)) {
            ce32 = (uint32_t)prefixes.getValue();
        }
        if(!USTRINGTRIE_HAS_NEXT(match)) { break; }
    }
    forwardNumCodePoints(lookBehind, errorCode);
    return ce32;
}

UChar32
CollationIterator::nextSkippedCodePoint(UErrorCode &errorCode) {
    if(skipped != NULL && skipped->hasNext()) { return skipped->next(); }
    if(numCpFwd == 0) { return U_SENTINEL; }
    UChar32 c = nextCodePoint(errorCode);
    if(skipped != NULL && !skipped->isEmpty() && c >= 0) { skipped->incBeyond(); }
    if(numCpFwd > 0 && c >= 0) { --numCpFwd; }
    return c;
}

void
CollationIterator::backwardNumSkipped(int32_t n, UErrorCode &errorCode) {
    if(skipped != NULL && !skipped->isEmpty()) {
        n = skipped->backwardNumCodePoints(n);
    }
    backwardNumCodePoints(n, errorCode);
    if(numCpFwd >= 0) { numCpFwd += n; }
}

uint32_t
CollationIterator::nextCE32FromContraction(const CollationData *d, uint32_t contractionCE32,
                                           const UChar *p, uint32_t ce32, UChar32 c,
                                           UErrorCode &errorCode) {
    // c: next code point after the original one

    // Number of code points read beyond the original code point.
    // Needed for discontiguous contraction matching.
    int32_t lookAhead = 1;
    // Number of code points read since the last match (initially only c).
    int32_t sinceMatch = 1;
    // Normally we only need a contiguous match,
    // and therefore need not remember the suffixes state from before a mismatch for retrying.
    // If we are already processing skipped combining marks, then we do track the state.
    UCharsTrie suffixes(p);
    if(skipped != NULL && !skipped->isEmpty()) { skipped->saveTrieState(suffixes); }
    UStringTrieResult match = suffixes.firstForCodePoint(c);
    for(;;) {
        UChar32 nextCp;
        if(USTRINGTRIE_HAS_VALUE(match)) {
            ce32 = (uint32_t)suffixes.getValue();
            if(!USTRINGTRIE_HAS_NEXT(match) || (c = nextSkippedCodePoint(errorCode)) < 0) {
                return ce32;
            }
            if(skipped != NULL && !skipped->isEmpty()) { skipped->saveTrieState(suffixes); }
            sinceMatch = 1;
        } else if(match == USTRINGTRIE_NO_MATCH || (nextCp = nextSkippedCodePoint(errorCode)) < 0) {
            // No match for c, or partial match (USTRINGTRIE_NO_VALUE) and no further text.
            // Back up if necessary, and try a discontiguous contraction.
            if((contractionCE32 & Collation::CONTRACT_TRAILING_CCC) != 0 &&
                    // Discontiguous contraction matching extends an existing match.
                    // If there is no match yet, then there is nothing to do.
                    ((contractionCE32 & Collation::CONTRACT_SINGLE_CP_NO_MATCH) == 0 ||
                        sinceMatch < lookAhead)) {
                // The last character of at least one suffix has lccc!=0,
                // allowing for discontiguous contractions.
                // UCA S2.1.1 only processes non-starters immediately following
                // "a match in the table" (sinceMatch=1).
                if(sinceMatch > 1) {
                    // Return to the state after the last match.
                    // (Return to sinceMatch=0 and re-fetch the first partially-matched character.)
                    backwardNumSkipped(sinceMatch, errorCode);
                    c = nextSkippedCodePoint(errorCode);
                    lookAhead -= sinceMatch - 1;
                    sinceMatch = 1;
                }
                if(d->getFCD16(c) > 0xff) {
                    return nextCE32FromDiscontiguousContraction(
                        d, suffixes, ce32, lookAhead, c, errorCode);
                }
            }
            break;
        } else {
            // Continue after partial match (USTRINGTRIE_NO_VALUE) for c.
            // It does not have a result value, therefore it is not itself "a match in the table".
            // If a partially-matched c has ccc!=0 then
            // it might be skipped in discontiguous contraction.
            c = nextCp;
            ++sinceMatch;
        }
        ++lookAhead;
        match = suffixes.nextForCodePoint(c);
    }
    backwardNumSkipped(sinceMatch, errorCode);
    return ce32;
}

uint32_t
CollationIterator::nextCE32FromDiscontiguousContraction(
        const CollationData *d, UCharsTrie &suffixes, uint32_t ce32,
        int32_t lookAhead, UChar32 c,
        UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }

    // UCA section 3.3.2 Contractions:
    // Contractions that end with non-starter characters
    // are known as discontiguous contractions.
    // ... discontiguous contractions must be detected in input text
    // whenever the final sequence of non-starter characters could be rearranged
    // so as to make a contiguous matching sequence that is canonically equivalent.

    // UCA: http://www.unicode.org/reports/tr10/#S2.1
    // S2.1 Find the longest initial substring S at each point that has a match in the table.
    // S2.1.1 If there are any non-starters following S, process each non-starter C.
    // S2.1.2 If C is not blocked from S, find if S + C has a match in the table.
    //     Note: A non-starter in a string is called blocked
    //     if there is another non-starter of the same canonical combining class or zero
    //     between it and the last character of canonical combining class 0.
    // S2.1.3 If there is a match, replace S by S + C, and remove C.

    // First: Is a discontiguous contraction even possible?
    uint16_t fcd16 = d->getFCD16(c);
    U_ASSERT(fcd16 > 0xff);  // The caller checked this already, as a shortcut.
    UChar32 nextCp = nextSkippedCodePoint(errorCode);
    if(nextCp < 0) {
        // No further text.
        backwardNumSkipped(1, errorCode);
        return ce32;
    }
    ++lookAhead;
    uint8_t prevCC = (uint8_t)fcd16;
    fcd16 = d->getFCD16(nextCp);
    if(fcd16 <= 0xff) {
        // The next code point after c is a starter (S2.1.1 "process each non-starter").
        backwardNumSkipped(2, errorCode);
        return ce32;
    }

    // We have read and matched (lookAhead-2) code points,
    // read non-matching c and peeked ahead at nextCp.
    // Return to the state before the mismatch and continue matching with nextCp.
    if(skipped == NULL || skipped->isEmpty()) {
        if(skipped == NULL) {
            skipped = new SkippedState();
            if(skipped == NULL) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return 0;
            }
        }
        suffixes.reset();
        if(lookAhead > 2) {
            // Replay the partial match so far.
            backwardNumCodePoints(lookAhead, errorCode);
            suffixes.firstForCodePoint(nextCodePoint(errorCode));
            for(int32_t i = 3; i < lookAhead; ++i) {
                suffixes.nextForCodePoint(nextCodePoint(errorCode));
            }
            // Skip c (which did not match) and nextCp (which we will try now).
            forwardNumCodePoints(2, errorCode);
        }
        skipped->saveTrieState(suffixes);
    } else {
        // Reset to the trie state before the failed match of c.
        skipped->resetToTrieState(suffixes);
    }

    skipped->setFirstSkipped(c);
    // Number of code points read since the last match (at this point: c and nextCp).
    int32_t sinceMatch = 2;
    c = nextCp;
    for(;;) {
        UStringTrieResult match;
        // "If C is not blocked from S, find if S + C has a match in the table." (S2.1.2)
        if(prevCC < (fcd16 >> 8) && USTRINGTRIE_HAS_VALUE(match = suffixes.nextForCodePoint(c))) {
            // "If there is a match, replace S by S + C, and remove C." (S2.1.3)
            // Keep prevCC unchanged.
            ce32 = (uint32_t)suffixes.getValue();
            sinceMatch = 0;
            skipped->recordMatch();
            if(!USTRINGTRIE_HAS_NEXT(match)) { break; }
            skipped->saveTrieState(suffixes);
        } else {
            // No match for "S + C", skip C.
            skipped->skip(c);
            skipped->resetToTrieState(suffixes);
            prevCC = (uint8_t)fcd16;
        }
        if((c = nextSkippedCodePoint(errorCode)) < 0) { break; }
        ++sinceMatch;
        fcd16 = d->getFCD16(c);
        if(fcd16 <= 0xff) {
            // The next code point after c is a starter (S2.1.1 "process each non-starter").
            break;
        }
    }
    backwardNumSkipped(sinceMatch, errorCode);
    UBool isTopDiscontiguous = skipped->isEmpty();
    skipped->replaceMatch();
    if(isTopDiscontiguous && !skipped->isEmpty()) {
        // We did get a match after skipping one or more combining marks,
        // and we are not in a recursive discontiguous contraction.
        // Append CEs from the contraction ce32
        // and then from the combining marks that we skipped before the match.
        c = U_SENTINEL;
        for(;;) {
            appendCEsFromCE32(d, c, ce32, TRUE, errorCode);
            // Fetch CE32s for skipped combining marks from the normal data, with fallback,
            // rather than from the CollationData where we found the contraction.
            if(!skipped->hasNext()) { break; }
            c = skipped->next();
            ce32 = getDataCE32(c);
            if(ce32 == Collation::FALLBACK_CE32) {
                d = data->base;
                ce32 = d->getCE32(c);
            } else {
                d = data;
            }
            // Note: A nested discontiguous-contraction match
            // replaces consumed combining marks with newly skipped ones
            // and resets the reading position to the beginning.
        }
        skipped->clear();
        ce32 = Collation::NO_CE32;  // Signal to the caller that the result is in the ceBuffer.
    }
    return ce32;
}

void
CollationIterator::appendNumericCEs(uint32_t ce32, UBool forward, UErrorCode &errorCode) {
    // Collect digits.
    CharString digits;
    if(forward) {
        for(;;) {
            char digit = Collation::digitFromCE32(ce32);
            digits.append(digit, errorCode);
            if(numCpFwd == 0) { break; }
            UChar32 c = nextCodePoint(errorCode);
            if(c < 0) { break; }
            ce32 = data->getCE32(c);
            if(ce32 == Collation::FALLBACK_CE32) {
                ce32 = data->base->getCE32(c);
            }
            if(!Collation::hasCE32Tag(ce32, Collation::DIGIT_TAG)) {
                backwardNumCodePoints(1, errorCode);
                break;
            }
            if(numCpFwd > 0) { --numCpFwd; }
        }
    } else {
        for(;;) {
            char digit = Collation::digitFromCE32(ce32);
            digits.append(digit, errorCode);
            UChar32 c = previousCodePoint(errorCode);
            if(c < 0) { break; }
            ce32 = data->getCE32(c);
            if(ce32 == Collation::FALLBACK_CE32) {
                ce32 = data->base->getCE32(c);
            }
            if(!Collation::hasCE32Tag(ce32, Collation::DIGIT_TAG)) {
                forwardNumCodePoints(1, errorCode);
                break;
            }
        }
        // Reverse the digit string.
        char *p = digits.data();
        char *q = p + digits.length() - 1;
        while(p < q) {
            char digit = *p;
            *p++ = *q;
            *q-- = digit;
        }
    }
    if(U_FAILURE(errorCode)) { return; }
    int32_t pos = 0;
    do {
        // Skip leading zeros.
        while(pos < (digits.length() - 1) && digits[pos] == 0) { ++pos; }
        // Write a sequence of CEs for at most 254 digits at a time.
        int32_t segmentLength = digits.length() - pos;
        if(segmentLength > 254) { segmentLength = 254; }
        appendNumericSegmentCEs(digits.data() + pos, segmentLength, errorCode);
        pos += segmentLength;
    } while(U_SUCCESS(errorCode) && pos < digits.length());
}

void
CollationIterator::appendNumericSegmentCEs(const char *digits, int32_t length, UErrorCode &errorCode) {
    U_ASSERT(1 <= length && length <= 254);
    U_ASSERT(length == 1 || digits[0] != 0);
    uint32_t numericPrimary = data->numericPrimary;
    // Note: We use primary byte values 2..255: digits are not compressible.
    if(length <= 7) {
        // Very dense encoding for small numbers.
        int32_t value = digits[0];
        for(int32_t i = 1; i < length; ++i) {
            value = value * 10 + digits[i];
        }
        // Primary weight second byte values:
        //     74 byte values   2.. 75 for small numbers in two-byte primary weights.
        //     40 byte values  76..115 for medium numbers in three-byte primary weights.
        //     16 byte values 116..131 for large numbers in four-byte primary weights.
        //    124 byte values 132..255 for very large numbers with 4..127 digit pairs.
        int32_t firstByte = 2;
        int32_t numBytes = 74;
        if(value < numBytes) {
            // Two-byte primary for 0..73, good for day & month numbers etc.
            uint32_t primary = numericPrimary | ((firstByte + value) << 16);
            ceBuffer.append(Collation::makeCE(primary), errorCode);
            return;
        }
        value -= numBytes;
        firstByte += numBytes;
        numBytes = 40;
        if(value < numBytes * 254) {
            // Three-byte primary for 74..10233=74+40*254-1, good for year numbers and more.
            uint32_t primary = numericPrimary |
                ((firstByte + value / 254) << 16) | ((2 + value % 254) << 8);
            ceBuffer.append(Collation::makeCE(primary), errorCode);
            return;
        }
        value -= numBytes * 254;
        firstByte += numBytes;
        numBytes = 16;
        if(value < numBytes * 254 * 254) {
            // Four-byte primary for 10234..1042489=10234+16*254*254-1.
            uint32_t primary = numericPrimary | (2 + value % 254);
            value /= 254;
            primary |= (2 + value % 254) << 8;
            value /= 254;
            primary |= (firstByte + value % 254) << 16;
            ceBuffer.append(Collation::makeCE(primary), errorCode);
            return;
        }
        // original value > 1042489
    }
    U_ASSERT(length >= 7);

    // The second primary byte value 132..255 indicates the number of digit pairs (4..127),
    // then we generate primary bytes with those pairs.
    // Omit trailing 00 pairs.
    // Decrement the value for the last pair.

    // Set the exponent. 4 pairs->132, 5 pairs->133, ..., 127 pairs->255.
    int32_t numPairs = (length + 1) / 2;
    uint32_t primary = numericPrimary | ((132 - 4 + numPairs) << 16);
    // Find the length without trailing 00 pairs.
    while(digits[length - 1] == 0 && digits[length - 2] == 0) {
        length -= 2;
    }
    // Read the first pair.
    uint32_t pair;
    int32_t pos;
    if(length & 1) {
        // Only "half a pair" if we have an odd number of digits.
        pair = digits[0];
        pos = 1;
    } else {
        pair = digits[0] * 10 + digits[1];
        pos = 2;
    }
    pair = 11 + 2 * pair;
    // Add the pairs of digits between pos and length.
    int32_t shift = 8;
    while(pos < length) {
        if(shift == 0) {
            // Every three pairs/bytes we need to store a 4-byte-primary CE
            // and start with a new CE with the '0' primary lead byte.
            primary |= pair;
            ceBuffer.append(Collation::makeCE(primary), errorCode);
            primary = numericPrimary;
            shift = 16;
        } else {
            primary |= pair << shift;
            shift -= 8;
        }
        pair = 11 + 2 * (digits[pos] * 10 + digits[pos + 1]);
        pos += 2;
    }
    primary |= (pair - 1) << shift;
    ceBuffer.append(Collation::makeCE(primary), errorCode);
}

int64_t
CollationIterator::previousCE(UVector32 &offsets, UErrorCode &errorCode) {
    if(ceBuffer.length > 0) {
        // Return the previous buffered CE.
        return ceBuffer.get(--ceBuffer.length);
    }
    offsets.removeAllElements();
    int32_t limitOffset = getOffset();
    UChar32 c = previousCodePoint(errorCode);
    if(c < 0) { return Collation::NO_CE; }
    if(data->isUnsafeBackward(c, isNumeric)) {
        return previousCEUnsafe(c, offsets, errorCode);
    }
    // Simple, safe-backwards iteration:
    // Get a CE going backwards, handle prefixes but no contractions.
    uint32_t ce32 = data->getCE32(c);
    const CollationData *d;
    if(ce32 == Collation::FALLBACK_CE32) {
        d = data->base;
        ce32 = d->getCE32(c);
    } else {
        d = data;
    }
    if(Collation::isSimpleOrLongCE32(ce32)) {
        return Collation::ceFromCE32(ce32);
    }
    appendCEsFromCE32(d, c, ce32, FALSE, errorCode);
    if(U_SUCCESS(errorCode)) {
        if(ceBuffer.length > 1) {
            offsets.addElement(getOffset(), errorCode);
            // For an expansion, the offset of each non-initial CE is the limit offset,
            // consistent with forward iteration.
            while(offsets.size() <= ceBuffer.length) {
                offsets.addElement(limitOffset, errorCode);
            }
        }
        return ceBuffer.get(--ceBuffer.length);
    } else {
        return Collation::NO_CE_PRIMARY;
    }
}

int64_t
CollationIterator::previousCEUnsafe(UChar32 c, UVector32 &offsets, UErrorCode &errorCode) {
    // We just move through the input counting safe and unsafe code points
    // without collecting the unsafe-backward substring into a buffer and
    // switching to it.
    // This is to keep the logic simple. Otherwise we would have to handle
    // prefix matching going before the backward buffer, switching
    // to iteration and back, etc.
    // In the most important case of iterating over a normal string,
    // reading from the string itself is already maximally fast.
    // The only drawback there is that after getting the CEs we always
    // skip backward to the safe character rather than switching out
    // of a backwardBuffer.
    // But this should not be the common case for previousCE(),
    // and correctness and maintainability are more important than
    // complex optimizations.
    // Find the first safe character before c.
    int32_t numBackward = 1;
    while((c = previousCodePoint(errorCode)) >= 0) {
        ++numBackward;
        if(!data->isUnsafeBackward(c, isNumeric)) {
            break;
        }
    }
    // Set the forward iteration limit.
    // Note: This counts code points.
    // We cannot enforce a limit in the middle of a surrogate pair or similar.
    numCpFwd = numBackward;
    // Reset the forward iterator.
    cesIndex = 0;
    U_ASSERT(ceBuffer.length == 0);
    // Go forward and collect the CEs.
    int32_t offset = getOffset();
    while(numCpFwd > 0) {
        // nextCE() normally reads one code point.
        // Contraction matching and digit specials read more and check numCpFwd.
        --numCpFwd;
        // Append one or more CEs to the ceBuffer.
        (void)nextCE(errorCode);
        U_ASSERT(U_FAILURE(errorCode) || ceBuffer.get(ceBuffer.length - 1) != Collation::NO_CE);
        // No need to loop for getting each expansion CE from nextCE().
        cesIndex = ceBuffer.length;
        // However, we need to write an offset for each CE.
        // This is for CollationElementIterator::getOffset() to return
        // intermediate offsets from the unsafe-backwards segment.
        U_ASSERT(offsets.size() < ceBuffer.length);
        offsets.addElement(offset, errorCode);
        // For an expansion, the offset of each non-initial CE is the limit offset,
        // consistent with forward iteration.
        offset = getOffset();
        while(offsets.size() < ceBuffer.length) {
            offsets.addElement(offset, errorCode);
        }
    }
    U_ASSERT(offsets.size() == ceBuffer.length);
    // End offset corresponding to just after the unsafe-backwards segment.
    offsets.addElement(offset, errorCode);
    // Reset the forward iteration limit
    // and move backward to before the segment for which we fetched CEs.
    numCpFwd = -1;
    backwardNumCodePoints(numBackward, errorCode);
    // Use the collected CEs and return the last one.
    cesIndex = 0;  // Avoid cesIndex > ceBuffer.length when that gets decremented.
    if(U_SUCCESS(errorCode)) {
        return ceBuffer.get(--ceBuffer.length);
    } else {
        return Collation::NO_CE_PRIMARY;
    }
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationkeys.cpp
*
* created on: 2012sep02
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION








U_NAMESPACE_BEGIN

SortKeyByteSink::~SortKeyByteSink() {}

void
SortKeyByteSink::Append(const char *bytes, int32_t n) {
    if (n <= 0 || bytes == NULL) {
        return;
    }
    if (ignore_ > 0) {
        int32_t ignoreRest = ignore_ - n;
        if (ignoreRest >= 0) {
            ignore_ = ignoreRest;
            return;
        } else {
            bytes += ignore_;
            n = -ignoreRest;
            ignore_ = 0;
        }
    }
    int32_t length = appended_;
    appended_ += n;
    if ((buffer_ + length) == bytes) {
        return;  // the caller used GetAppendBuffer() and wrote the bytes already
    }
    int32_t available = capacity_ - length;
    if (n <= available) {
        uprv_memcpy(buffer_ + length, bytes, n);
    } else {
        AppendBeyondCapacity(bytes, n, length);
    }
}

char *
SortKeyByteSink::GetAppendBuffer(int32_t min_capacity,
                                 int32_t desired_capacity_hint,
                                 char *scratch,
                                 int32_t scratch_capacity,
                                 int32_t *result_capacity) {
    if (min_capacity < 1 || scratch_capacity < min_capacity) {
        *result_capacity = 0;
        return NULL;
    }
    if (ignore_ > 0) {
        // Do not write ignored bytes right at the end of the buffer.
        *result_capacity = scratch_capacity;
        return scratch;
    }
    int32_t available = capacity_ - appended_;
    if (available >= min_capacity) {
        *result_capacity = available;
        return buffer_ + appended_;
    } else if (Resize(desired_capacity_hint, appended_)) {
        *result_capacity = capacity_ - appended_;
        return buffer_ + appended_;
    } else {
        *result_capacity = scratch_capacity;
        return scratch;
    }
}

namespace {

/**
 * uint8_t byte buffer, similar to CharString but simpler.
 */
class SortKeyLevel : public UMemory {
public:
    SortKeyLevel() : len(0), ok(TRUE) {}
    ~SortKeyLevel() {}

    /** @return FALSE if memory allocation failed */
    UBool isOk() const { return ok; }
    UBool isEmpty() const { return len == 0; }
    int32_t length() const { return len; }
    const uint8_t *data() const { return buffer.getAlias(); }
    uint8_t operator[](int32_t index) const { return buffer[index]; }

    uint8_t *data() { return buffer.getAlias(); }

    void appendByte(uint32_t b);
    void appendWeight16(uint32_t w);
    void appendWeight32(uint32_t w);
    void appendReverseWeight16(uint32_t w);

    /** Appends all but the last byte to the sink. The last byte should be the 01 terminator. */
    void appendTo(ByteSink &sink) const {
        U_ASSERT(len > 0 && buffer[len - 1] == 1);
        sink.Append(reinterpret_cast<const char *>(buffer.getAlias()), len - 1);
    }

private:
    MaybeStackArray<uint8_t, 40> buffer;
    int32_t len;
    UBool ok;

    UBool ensureCapacity(int32_t appendCapacity);

    SortKeyLevel(const SortKeyLevel &other); // forbid copying of this class
    SortKeyLevel &operator=(const SortKeyLevel &other); // forbid copying of this class
};

void SortKeyLevel::appendByte(uint32_t b) {
    if(len < buffer.getCapacity() || ensureCapacity(1)) {
        buffer[len++] = (uint8_t)b;
    }
}

void
SortKeyLevel::appendWeight16(uint32_t w) {
    U_ASSERT((w & 0xffff) != 0);
    uint8_t b0 = (uint8_t)(w >> 8);
    uint8_t b1 = (uint8_t)w;
    int32_t appendLength = (b1 == 0) ? 1 : 2;
    if((len + appendLength) <= buffer.getCapacity() || ensureCapacity(appendLength)) {
        buffer[len++] = b0;
        if(b1 != 0) {
            buffer[len++] = b1;
        }
    }
}

void
SortKeyLevel::appendWeight32(uint32_t w) {
    U_ASSERT(w != 0);
    uint8_t bytes[4] = { (uint8_t)(w >> 24), (uint8_t)(w >> 16), (uint8_t)(w >> 8), (uint8_t)w };
    int32_t appendLength = (bytes[1] == 0) ? 1 : (bytes[2] == 0) ? 2 : (bytes[3] == 0) ? 3 : 4;
    if((len + appendLength) <= buffer.getCapacity() || ensureCapacity(appendLength)) {
        buffer[len++] = bytes[0];
        if(bytes[1] != 0) {
            buffer[len++] = bytes[1];
            if(bytes[2] != 0) {
                buffer[len++] = bytes[2];
                if(bytes[3] != 0) {
                    buffer[len++] = bytes[3];
                }
            }
        }
    }
}

void
SortKeyLevel::appendReverseWeight16(uint32_t w) {
    U_ASSERT((w & 0xffff) != 0);
    uint8_t b0 = (uint8_t)(w >> 8);
    uint8_t b1 = (uint8_t)w;
    int32_t appendLength = (b1 == 0) ? 1 : 2;
    if((len + appendLength) <= buffer.getCapacity() || ensureCapacity(appendLength)) {
        if(b1 == 0) {
            buffer[len++] = b0;
        } else {
            buffer[len] = b1;
            buffer[len + 1] = b0;
            len += 2;
        }
    }
}

UBool SortKeyLevel::ensureCapacity(int32_t appendCapacity) {
    if(!ok) {
        return FALSE;
    }
    int32_t newCapacity = 2 * buffer.getCapacity();
    int32_t altCapacity = len + 2 * appendCapacity;
    if (newCapacity < altCapacity) {
        newCapacity = altCapacity;
    }
    if (newCapacity < 200) {
        newCapacity = 200;
    }
    if(buffer.resize(newCapacity, len)==NULL) {
        return ok = FALSE;
    }
    return TRUE;
}

}  // namespace

CollationKeys::LevelCallback::~LevelCallback() {}

UBool
CollationKeys::LevelCallback::needToWrite(Collation::Level /*level*/) { return TRUE; }

/**
 * Map from collation strength (UColAttributeValue)
 * to a mask of Collation::Level bits up to that strength,
 * excluding the CASE_LEVEL which is independent of the strength,
 * and excluding IDENTICAL_LEVEL which this function does not write.
 */
static const uint32_t levelMasks[UCOL_STRENGTH_LIMIT] = {
    2,          // UCOL_PRIMARY -> PRIMARY_LEVEL
    6,          // UCOL_SECONDARY -> up to SECONDARY_LEVEL
    0x16,       // UCOL_TERTIARY -> up to TERTIARY_LEVEL
    0x36,       // UCOL_QUATERNARY -> up to QUATERNARY_LEVEL
    0, 0, 0, 0,
    0, 0, 0, 0,
    0, 0, 0,
    0x36        // UCOL_IDENTICAL -> up to QUATERNARY_LEVEL
};

void
CollationKeys::writeSortKeyUpToQuaternary(CollationIterator &iter,
                                          const UBool *compressibleBytes,
                                          const CollationSettings &settings,
                                          SortKeyByteSink &sink,
                                          Collation::Level minLevel, LevelCallback &callback,
                                          UBool preflight, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }

    int32_t options = settings.options;
    // Set of levels to process and write.
    uint32_t levels = levelMasks[CollationSettings::getStrength(options)];
    if((options & CollationSettings::CASE_LEVEL) != 0) {
        levels |= Collation::CASE_LEVEL_FLAG;
    }
    // Minus the levels below minLevel.
    levels &= ~(((uint32_t)1 << minLevel) - 1);
    if(levels == 0) { return; }

    uint32_t variableTop;
    if((options & CollationSettings::ALTERNATE_MASK) == 0) {
        variableTop = 0;
    } else {
        // +1 so that we can use "<" and primary ignorables test out early.
        variableTop = settings.variableTop + 1;
    }

    uint32_t tertiaryMask = CollationSettings::getTertiaryMask(options);

    SortKeyLevel cases;
    SortKeyLevel secondaries;
    SortKeyLevel tertiaries;
    SortKeyLevel quaternaries;

    uint32_t prevReorderedPrimary = 0;  // 0==no compression
    int32_t commonCases = 0;
    int32_t commonSecondaries = 0;
    int32_t commonTertiaries = 0;
    int32_t commonQuaternaries = 0;

    uint32_t prevSecondary = 0;
    int32_t secSegmentStart = 0;

    for(;;) {
        // No need to keep all CEs in the buffer when we write a sort key.
        iter.clearCEsIfNoneRemaining();
        int64_t ce = iter.nextCE(errorCode);
        uint32_t p = (uint32_t)(ce >> 32);
        if(p < variableTop && p > Collation::MERGE_SEPARATOR_PRIMARY) {
            // Variable CE, shift it to quaternary level.
            // Ignore all following primary ignorables, and shift further variable CEs.
            if(commonQuaternaries != 0) {
                --commonQuaternaries;
                while(commonQuaternaries >= QUAT_COMMON_MAX_COUNT) {
                    quaternaries.appendByte(QUAT_COMMON_MIDDLE);
                    commonQuaternaries -= QUAT_COMMON_MAX_COUNT;
                }
                // Shifted primary weights are lower than the common weight.
                quaternaries.appendByte(QUAT_COMMON_LOW + commonQuaternaries);
                commonQuaternaries = 0;
            }
            do {
                if((levels & Collation::QUATERNARY_LEVEL_FLAG) != 0) {
                    if(settings.hasReordering()) {
                        p = settings.reorder(p);
                    }
                    if((p >> 24) >= QUAT_SHIFTED_LIMIT_BYTE) {
                        // Prevent shifted primary lead bytes from
                        // overlapping with the common compression range.
                        quaternaries.appendByte(QUAT_SHIFTED_LIMIT_BYTE);
                    }
                    quaternaries.appendWeight32(p);
                }
                do {
                    ce = iter.nextCE(errorCode);
                    p = (uint32_t)(ce >> 32);
                } while(p == 0);
            } while(p < variableTop && p > Collation::MERGE_SEPARATOR_PRIMARY);
        }
        // ce could be primary ignorable, or NO_CE, or the merge separator,
        // or a regular primary CE, but it is not variable.
        // If ce==NO_CE, then write nothing for the primary level but
        // terminate compression on all levels and then exit the loop.
        if(p > Collation::NO_CE_PRIMARY && (levels & Collation::PRIMARY_LEVEL_FLAG) != 0) {
            // Test the un-reordered primary for compressibility.
            UBool isCompressible = compressibleBytes[p >> 24];
            if(settings.hasReordering()) {
                p = settings.reorder(p);
            }
            uint32_t p1 = p >> 24;
            if(!isCompressible || p1 != (prevReorderedPrimary >> 24)) {
                if(prevReorderedPrimary != 0) {
                    if(p < prevReorderedPrimary) {
                        // No primary compression terminator
                        // at the end of the level or merged segment.
                        if(p1 > Collation::MERGE_SEPARATOR_BYTE) {
                            sink.Append(Collation::PRIMARY_COMPRESSION_LOW_BYTE);
                        }
                    } else {
                        sink.Append(Collation::PRIMARY_COMPRESSION_HIGH_BYTE);
                    }
                }
                sink.Append(p1);
                if(isCompressible) {
                    prevReorderedPrimary = p;
                } else {
                    prevReorderedPrimary = 0;
                }
            }
            char p2 = (char)(p >> 16);
            if(p2 != 0) {
                char buffer[3] = { p2, (char)(p >> 8), (char)p };
                sink.Append(buffer, (buffer[1] == 0) ? 1 : (buffer[2] == 0) ? 2 : 3);
            }
            // Optimization for internalNextSortKeyPart():
            // When the primary level overflows we can stop because we need not
            // calculate (preflight) the whole sort key length.
            if(!preflight && sink.Overflowed()) {
                if(U_SUCCESS(errorCode) && !sink.IsOk()) {
                    errorCode = U_MEMORY_ALLOCATION_ERROR;
                }
                return;
            }
        }

        uint32_t lower32 = (uint32_t)ce;
        if(lower32 == 0) { continue; }  // completely ignorable, no secondary/case/tertiary/quaternary

        if((levels & Collation::SECONDARY_LEVEL_FLAG) != 0) {
            uint32_t s = lower32 >> 16;
            if(s == 0) {
                // secondary ignorable
            } else if(s == Collation::COMMON_WEIGHT16 &&
                    ((options & CollationSettings::BACKWARD_SECONDARY) == 0 ||
                        p != Collation::MERGE_SEPARATOR_PRIMARY)) {
                // s is a common secondary weight, and
                // backwards-secondary is off or the ce is not the merge separator.
                ++commonSecondaries;
            } else if((options & CollationSettings::BACKWARD_SECONDARY) == 0) {
                if(commonSecondaries != 0) {
                    --commonSecondaries;
                    while(commonSecondaries >= SEC_COMMON_MAX_COUNT) {
                        secondaries.appendByte(SEC_COMMON_MIDDLE);
                        commonSecondaries -= SEC_COMMON_MAX_COUNT;
                    }
                    uint32_t b;
                    if(s < Collation::COMMON_WEIGHT16) {
                        b = SEC_COMMON_LOW + commonSecondaries;
                    } else {
                        b = SEC_COMMON_HIGH - commonSecondaries;
                    }
                    secondaries.appendByte(b);
                    commonSecondaries = 0;
                }
                secondaries.appendWeight16(s);
            } else {
                if(commonSecondaries != 0) {
                    --commonSecondaries;
                    // Append reverse weights. The level will be re-reversed later.
                    int32_t remainder = commonSecondaries % SEC_COMMON_MAX_COUNT;
                    uint32_t b;
                    if(prevSecondary < Collation::COMMON_WEIGHT16) {
                        b = SEC_COMMON_LOW + remainder;
                    } else {
                        b = SEC_COMMON_HIGH - remainder;
                    }
                    secondaries.appendByte(b);
                    commonSecondaries -= remainder;
                    // commonSecondaries is now a multiple of SEC_COMMON_MAX_COUNT.
                    while(commonSecondaries > 0) {  // same as >= SEC_COMMON_MAX_COUNT
                        secondaries.appendByte(SEC_COMMON_MIDDLE);
                        commonSecondaries -= SEC_COMMON_MAX_COUNT;
                    }
                    // commonSecondaries == 0
                }
                if(0 < p && p <= Collation::MERGE_SEPARATOR_PRIMARY) {
                    // The backwards secondary level compares secondary weights backwards
                    // within segments separated by the merge separator (U+FFFE).
                    uint8_t *secs = secondaries.data();
                    int32_t last = secondaries.length() - 1;
                    if(secSegmentStart < last) {
                        uint8_t *q = secs + secSegmentStart;
                        uint8_t *r = secs + last;
                        do {
                            uint8_t b = *q;
                            *q++ = *r;
                            *r-- = b;
                        } while(q < r);
                    }
                    secondaries.appendByte(p == Collation::NO_CE_PRIMARY ?
                        Collation::LEVEL_SEPARATOR_BYTE : Collation::MERGE_SEPARATOR_BYTE);
                    prevSecondary = 0;
                    secSegmentStart = secondaries.length();
                } else {
                    secondaries.appendReverseWeight16(s);
                    prevSecondary = s;
                }
            }
        }

        if((levels & Collation::CASE_LEVEL_FLAG) != 0) {
            if((CollationSettings::getStrength(options) == UCOL_PRIMARY) ?
                    p == 0 : lower32 <= 0xffff) {
                // Primary+caseLevel: Ignore case level weights of primary ignorables.
                // Otherwise: Ignore case level weights of secondary ignorables.
                // For details see the comments in the CollationCompare class.
            } else {
                uint32_t c = (lower32 >> 8) & 0xff;  // case bits & tertiary lead byte
                U_ASSERT((c & 0xc0) != 0xc0);
                if((c & 0xc0) == 0 && c > Collation::LEVEL_SEPARATOR_BYTE) {
                    ++commonCases;
                } else {
                    if((options & CollationSettings::UPPER_FIRST) == 0) {
                        // lowerFirst: Compress common weights to nibbles 1..7..13, mixed=14, upper=15.
                        // If there are only common (=lowest) weights in the whole level,
                        // then we need not write anything.
                        // Level length differences are handled already on the next-higher level.
                        if(commonCases != 0 &&
                                (c > Collation::LEVEL_SEPARATOR_BYTE || !cases.isEmpty())) {
                            --commonCases;
                            while(commonCases >= CASE_LOWER_FIRST_COMMON_MAX_COUNT) {
                                cases.appendByte(CASE_LOWER_FIRST_COMMON_MIDDLE << 4);
                                commonCases -= CASE_LOWER_FIRST_COMMON_MAX_COUNT;
                            }
                            uint32_t b;
                            if(c <= Collation::LEVEL_SEPARATOR_BYTE) {
                                b = CASE_LOWER_FIRST_COMMON_LOW + commonCases;
                            } else {
                                b = CASE_LOWER_FIRST_COMMON_HIGH - commonCases;
                            }
                            cases.appendByte(b << 4);
                            commonCases = 0;
                        }
                        if(c > Collation::LEVEL_SEPARATOR_BYTE) {
                            c = (CASE_LOWER_FIRST_COMMON_HIGH + (c >> 6)) << 4;  // 14 or 15
                        }
                    } else {
                        // upperFirst: Compress common weights to nibbles 3..15, mixed=2, upper=1.
                        // The compressed common case weights only go up from the "low" value
                        // because with upperFirst the common weight is the highest one.
                        if(commonCases != 0) {
                            --commonCases;
                            while(commonCases >= CASE_UPPER_FIRST_COMMON_MAX_COUNT) {
                                cases.appendByte(CASE_UPPER_FIRST_COMMON_LOW << 4);
                                commonCases -= CASE_UPPER_FIRST_COMMON_MAX_COUNT;
                            }
                            cases.appendByte((CASE_UPPER_FIRST_COMMON_LOW + commonCases) << 4);
                            commonCases = 0;
                        }
                        if(c > Collation::LEVEL_SEPARATOR_BYTE) {
                            c = (CASE_UPPER_FIRST_COMMON_LOW - (c >> 6)) << 4;  // 2 or 1
                        }
                    }
                    // c is a separator byte 01,
                    // or a left-shifted nibble 0x10, 0x20, ... 0xf0.
                    cases.appendByte(c);
                }
            }
        }

        if((levels & Collation::TERTIARY_LEVEL_FLAG) != 0) {
            uint32_t t = lower32 & tertiaryMask;
            U_ASSERT((lower32 & 0xc000) != 0xc000);
            if(t == Collation::COMMON_WEIGHT16) {
                ++commonTertiaries;
            } else if((tertiaryMask & 0x8000) == 0) {
                // Tertiary weights without case bits.
                // Move lead bytes 06..3F to C6..FF for a large common-weight range.
                if(commonTertiaries != 0) {
                    --commonTertiaries;
                    while(commonTertiaries >= TER_ONLY_COMMON_MAX_COUNT) {
                        tertiaries.appendByte(TER_ONLY_COMMON_MIDDLE);
                        commonTertiaries -= TER_ONLY_COMMON_MAX_COUNT;
                    }
                    uint32_t b;
                    if(t < Collation::COMMON_WEIGHT16) {
                        b = TER_ONLY_COMMON_LOW + commonTertiaries;
                    } else {
                        b = TER_ONLY_COMMON_HIGH - commonTertiaries;
                    }
                    tertiaries.appendByte(b);
                    commonTertiaries = 0;
                }
                if(t > Collation::COMMON_WEIGHT16) { t += 0xc000; }
                tertiaries.appendWeight16(t);
            } else if((options & CollationSettings::UPPER_FIRST) == 0) {
                // Tertiary weights with caseFirst=lowerFirst.
                // Move lead bytes 06..BF to 46..FF for the common-weight range.
                if(commonTertiaries != 0) {
                    --commonTertiaries;
                    while(commonTertiaries >= TER_LOWER_FIRST_COMMON_MAX_COUNT) {
                        tertiaries.appendByte(TER_LOWER_FIRST_COMMON_MIDDLE);
                        commonTertiaries -= TER_LOWER_FIRST_COMMON_MAX_COUNT;
                    }
                    uint32_t b;
                    if(t < Collation::COMMON_WEIGHT16) {
                        b = TER_LOWER_FIRST_COMMON_LOW + commonTertiaries;
                    } else {
                        b = TER_LOWER_FIRST_COMMON_HIGH - commonTertiaries;
                    }
                    tertiaries.appendByte(b);
                    commonTertiaries = 0;
                }
                if(t > Collation::COMMON_WEIGHT16) { t += 0x4000; }
                tertiaries.appendWeight16(t);
            } else {
                // Tertiary weights with caseFirst=upperFirst.
                // Do not change the artificial uppercase weight of a tertiary CE (0.0.ut),
                // to keep tertiary CEs well-formed.
                // Their case+tertiary weights must be greater than those of
                // primary and secondary CEs.
                //
                // Separator         01 -> 01      (unchanged)
                // Lowercase     02..04 -> 82..84  (includes uncased)
                // Common weight     05 -> 85..C5  (common-weight compression range)
                // Lowercase     06..3F -> C6..FF
                // Mixed case    42..7F -> 42..7F
                // Uppercase     82..BF -> 02..3F
                // Tertiary CE   86..BF -> C6..FF
                if(t <= Collation::NO_CE_WEIGHT16) {
                    // Keep separators unchanged.
                } else if(lower32 > 0xffff) {
                    // Invert case bits of primary & secondary CEs.
                    t ^= 0xc000;
                    if(t < (TER_UPPER_FIRST_COMMON_HIGH << 8)) {
                        t -= 0x4000;
                    }
                } else {
                    // Keep uppercase bits of tertiary CEs.
                    U_ASSERT(0x8600 <= t && t <= 0xbfff);
                    t += 0x4000;
                }
                if(commonTertiaries != 0) {
                    --commonTertiaries;
                    while(commonTertiaries >= TER_UPPER_FIRST_COMMON_MAX_COUNT) {
                        tertiaries.appendByte(TER_UPPER_FIRST_COMMON_MIDDLE);
                        commonTertiaries -= TER_UPPER_FIRST_COMMON_MAX_COUNT;
                    }
                    uint32_t b;
                    if(t < (TER_UPPER_FIRST_COMMON_LOW << 8)) {
                        b = TER_UPPER_FIRST_COMMON_LOW + commonTertiaries;
                    } else {
                        b = TER_UPPER_FIRST_COMMON_HIGH - commonTertiaries;
                    }
                    tertiaries.appendByte(b);
                    commonTertiaries = 0;
                }
                tertiaries.appendWeight16(t);
            }
        }

        if((levels & Collation::QUATERNARY_LEVEL_FLAG) != 0) {
            uint32_t q = lower32 & 0xffff;
            if((q & 0xc0) == 0 && q > Collation::NO_CE_WEIGHT16) {
                ++commonQuaternaries;
            } else if(q == Collation::NO_CE_WEIGHT16 &&
                    (options & CollationSettings::ALTERNATE_MASK) == 0 &&
                    quaternaries.isEmpty()) {
                // If alternate=non-ignorable and there are only common quaternary weights,
                // then we need not write anything.
                // The only weights greater than the merge separator and less than the common weight
                // are shifted primary weights, which are not generated for alternate=non-ignorable.
                // There are also exactly as many quaternary weights as tertiary weights,
                // so level length differences are handled already on tertiary level.
                // Any above-common quaternary weight will compare greater regardless.
                quaternaries.appendByte(Collation::LEVEL_SEPARATOR_BYTE);
            } else {
                if(q == Collation::NO_CE_WEIGHT16) {
                    q = Collation::LEVEL_SEPARATOR_BYTE;
                } else {
                    q = 0xfc + ((q >> 6) & 3);
                }
                if(commonQuaternaries != 0) {
                    --commonQuaternaries;
                    while(commonQuaternaries >= QUAT_COMMON_MAX_COUNT) {
                        quaternaries.appendByte(QUAT_COMMON_MIDDLE);
                        commonQuaternaries -= QUAT_COMMON_MAX_COUNT;
                    }
                    uint32_t b;
                    if(q < QUAT_COMMON_LOW) {
                        b = QUAT_COMMON_LOW + commonQuaternaries;
                    } else {
                        b = QUAT_COMMON_HIGH - commonQuaternaries;
                    }
                    quaternaries.appendByte(b);
                    commonQuaternaries = 0;
                }
                quaternaries.appendByte(q);
            }
        }

        if((lower32 >> 24) == Collation::LEVEL_SEPARATOR_BYTE) { break; }  // ce == NO_CE
    }

    if(U_FAILURE(errorCode)) { return; }

    // Append the beyond-primary levels.
    UBool ok = TRUE;
    if((levels & Collation::SECONDARY_LEVEL_FLAG) != 0) {
        if(!callback.needToWrite(Collation::SECONDARY_LEVEL)) { return; }
        ok &= secondaries.isOk();
        sink.Append(Collation::LEVEL_SEPARATOR_BYTE);
        secondaries.appendTo(sink);
    }

    if((levels & Collation::CASE_LEVEL_FLAG) != 0) {
        if(!callback.needToWrite(Collation::CASE_LEVEL)) { return; }
        ok &= cases.isOk();
        sink.Append(Collation::LEVEL_SEPARATOR_BYTE);
        // Write pairs of nibbles as bytes, except separator bytes as themselves.
        int32_t length = cases.length() - 1;  // Ignore the trailing NO_CE.
        uint8_t b = 0;
        for(int32_t i = 0; i < length; ++i) {
            uint8_t c = (uint8_t)cases[i];
            U_ASSERT((c & 0xf) == 0 && c != 0);
            if(b == 0) {
                b = c;
            } else {
                sink.Append(b | (c >> 4));
                b = 0;
            }
        }
        if(b != 0) {
            sink.Append(b);
        }
    }

    if((levels & Collation::TERTIARY_LEVEL_FLAG) != 0) {
        if(!callback.needToWrite(Collation::TERTIARY_LEVEL)) { return; }
        ok &= tertiaries.isOk();
        sink.Append(Collation::LEVEL_SEPARATOR_BYTE);
        tertiaries.appendTo(sink);
    }

    if((levels & Collation::QUATERNARY_LEVEL_FLAG) != 0) {
        if(!callback.needToWrite(Collation::QUATERNARY_LEVEL)) { return; }
        ok &= quaternaries.isOk();
        sink.Append(Collation::LEVEL_SEPARATOR_BYTE);
        quaternaries.appendTo(sink);
    }

    if(!ok || !sink.IsOk()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
    }
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationroot.cpp
*
* created on: 2012dec17
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION














U_NAMESPACE_BEGIN

namespace {

static const CollationCacheEntry *rootSingleton = NULL;
static UInitOnce collationroot_initOnce = U_INITONCE_INITIALIZER;

}  // namespace

U_CDECL_BEGIN

static UBool U_CALLCONV uprv_collation_root_cleanup() {
    SharedObject::clearPtr(rootSingleton);
    collationroot_initOnce.reset();
    return TRUE;
}

U_CDECL_END

void U_CALLCONV
CollationRoot::load(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    LocalPointer<CollationTailoring> t(new CollationTailoring(NULL));
    if(t.isNull() || t->isBogus()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    t->memory = udata_openChoice(U_ICUDATA_NAME U_TREE_SEPARATOR_STRING "coll",
                                 "icu", "ucadata",
                                 CollationDataReader::isAcceptable, t->version, &errorCode);
    if(U_FAILURE(errorCode)) { return; }
    const uint8_t *inBytes = static_cast<const uint8_t *>(udata_getMemory(t->memory));
    CollationDataReader::read(NULL, inBytes, udata_getLength(t->memory), *t, errorCode);
    if(U_FAILURE(errorCode)) { return; }
    ucln_i18n_registerCleanup(UCLN_I18N_COLLATION_ROOT, uprv_collation_root_cleanup);
    CollationCacheEntry *entry = new CollationCacheEntry(Locale::getRoot(), t.getAlias());
    if(entry != NULL) {
        t.orphan();  // The rootSingleton took ownership of the tailoring.
        entry->addRef();
        rootSingleton = entry;
    }
}

const CollationCacheEntry *
CollationRoot::getRootCacheEntry(UErrorCode &errorCode) {
    umtx_initOnce(collationroot_initOnce, CollationRoot::load, errorCode);
    if(U_FAILURE(errorCode)) { return NULL; }
    return rootSingleton;
}

const CollationTailoring *
CollationRoot::getRoot(UErrorCode &errorCode) {
    umtx_initOnce(collationroot_initOnce, CollationRoot::load, errorCode);
    if(U_FAILURE(errorCode)) { return NULL; }
    return rootSingleton->tailoring;
}

const CollationData *
CollationRoot::getData(UErrorCode &errorCode) {
    const CollationTailoring *root = getRoot(errorCode);
    if(U_FAILURE(errorCode)) { return NULL; }
    return root->data;
}

const CollationSettings *
CollationRoot::getSettings(UErrorCode &errorCode) {
    const CollationTailoring *root = getRoot(errorCode);
    if(U_FAILURE(errorCode)) { return NULL; }
    return root->settings;
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationrootelements.cpp
*
* created on: 2013mar05
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION





U_NAMESPACE_BEGIN

int64_t
CollationRootElements::lastCEWithPrimaryBefore(uint32_t p) const {
    if(p == 0) { return 0; }
    U_ASSERT(p > elements[elements[IX_FIRST_PRIMARY_INDEX]]);
    int32_t index = findP(p);
    uint32_t q = elements[index];
    uint32_t secTer;
    if(p == (q & 0xffffff00)) {
        // p == elements[index] is a root primary. Find the CE before it.
        // We must not be in a primary range.
        U_ASSERT((q & PRIMARY_STEP_MASK) == 0);
        secTer = elements[index - 1];
        if((secTer & SEC_TER_DELTA_FLAG) == 0) {
            // Primary CE just before p.
            p = secTer & 0xffffff00;
            secTer = Collation::COMMON_SEC_AND_TER_CE;
        } else {
            // secTer = last secondary & tertiary for the previous primary
            index -= 2;
            for(;;) {
                p = elements[index];
                if((p & SEC_TER_DELTA_FLAG) == 0) {
                    p &= 0xffffff00;
                    break;
                }
                --index;
            }
        }
    } else {
        // p > elements[index] which is the previous primary.
        // Find the last secondary & tertiary weights for it.
        p = q & 0xffffff00;
        secTer = Collation::COMMON_SEC_AND_TER_CE;
        for(;;) {
            q = elements[++index];
            if((q & SEC_TER_DELTA_FLAG) == 0) {
                // We must not be in a primary range.
                U_ASSERT((q & PRIMARY_STEP_MASK) == 0);
                break;
            }
            secTer = q;
        }
    }
    return ((int64_t)p << 32) | (secTer & ~SEC_TER_DELTA_FLAG);
}

int64_t
CollationRootElements::firstCEWithPrimaryAtLeast(uint32_t p) const {
    if(p == 0) { return 0; }
    int32_t index = findP(p);
    if(p != (elements[index] & 0xffffff00)) {
        for(;;) {
            p = elements[++index];
            if((p & SEC_TER_DELTA_FLAG) == 0) {
                // First primary after p. We must not be in a primary range.
                U_ASSERT((p & PRIMARY_STEP_MASK) == 0);
                break;
            }
        }
    }
    // The code above guarantees that p has at most 3 bytes: (p & 0xff) == 0.
    return ((int64_t)p << 32) | Collation::COMMON_SEC_AND_TER_CE;
}

uint32_t
CollationRootElements::getPrimaryBefore(uint32_t p, UBool isCompressible) const {
    int32_t index = findPrimary(p);
    int32_t step;
    uint32_t q = elements[index];
    if(p == (q & 0xffffff00)) {
        // Found p itself. Return the previous primary.
        // See if p is at the end of a previous range.
        step = (int32_t)q & PRIMARY_STEP_MASK;
        if(step == 0) {
            // p is not at the end of a range. Look for the previous primary.
            do {
                p = elements[--index];
            } while((p & SEC_TER_DELTA_FLAG) != 0);
            return p & 0xffffff00;
        }
    } else {
        // p is in a range, and not at the start.
        uint32_t nextElement = elements[index + 1];
        U_ASSERT(isEndOfPrimaryRange(nextElement));
        step = (int32_t)nextElement & PRIMARY_STEP_MASK;
    }
    // Return the previous range primary.
    if((p & 0xffff) == 0) {
        return Collation::decTwoBytePrimaryByOneStep(p, isCompressible, step);
    } else {
        return Collation::decThreeBytePrimaryByOneStep(p, isCompressible, step);
    }
}

uint32_t
CollationRootElements::getSecondaryBefore(uint32_t p, uint32_t s) const {
    int32_t index;
    uint32_t previousSec, sec;
    if(p == 0) {
        index = (int32_t)elements[IX_FIRST_SECONDARY_INDEX];
        // Gap at the beginning of the secondary CE range.
        previousSec = 0;
        sec = elements[index] >> 16;
    } else {
        index = findPrimary(p) + 1;
        previousSec = Collation::BEFORE_WEIGHT16;
        sec = getFirstSecTerForPrimary(index) >> 16;
    }
    U_ASSERT(s >= sec);
    while(s > sec) {
        previousSec = sec;
        U_ASSERT((elements[index] & SEC_TER_DELTA_FLAG) != 0);
        sec = elements[index++] >> 16;
    }
    U_ASSERT(sec == s);
    return previousSec;
}

uint32_t
CollationRootElements::getTertiaryBefore(uint32_t p, uint32_t s, uint32_t t) const {
    U_ASSERT((t & ~Collation::ONLY_TERTIARY_MASK) == 0);
    int32_t index;
    uint32_t previousTer, secTer;
    if(p == 0) {
        if(s == 0) {
            index = (int32_t)elements[IX_FIRST_TERTIARY_INDEX];
            // Gap at the beginning of the tertiary CE range.
            previousTer = 0;
        } else {
            index = (int32_t)elements[IX_FIRST_SECONDARY_INDEX];
            previousTer = Collation::BEFORE_WEIGHT16;
        }
        secTer = elements[index] & ~SEC_TER_DELTA_FLAG;
    } else {
        index = findPrimary(p) + 1;
        previousTer = Collation::BEFORE_WEIGHT16;
        secTer = getFirstSecTerForPrimary(index);
    }
    uint32_t st = (s << 16) | t;
    while(st > secTer) {
        if((secTer >> 16) == s) { previousTer = secTer; }
        U_ASSERT((elements[index] & SEC_TER_DELTA_FLAG) != 0);
        secTer = elements[index++] & ~SEC_TER_DELTA_FLAG;
    }
    U_ASSERT(secTer == st);
    return previousTer & 0xffff;
}

uint32_t
CollationRootElements::getPrimaryAfter(uint32_t p, int32_t index, UBool isCompressible) const {
    U_ASSERT(p == (elements[index] & 0xffffff00) || isEndOfPrimaryRange(elements[index + 1]));
    uint32_t q = elements[++index];
    int32_t step;
    if((q & SEC_TER_DELTA_FLAG) == 0 && (step = (int32_t)q & PRIMARY_STEP_MASK) != 0) {
        // Return the next primary in this range.
        if((p & 0xffff) == 0) {
            return Collation::incTwoBytePrimaryByOffset(p, isCompressible, step);
        } else {
            return Collation::incThreeBytePrimaryByOffset(p, isCompressible, step);
        }
    } else {
        // Return the next primary in the list.
        while((q & SEC_TER_DELTA_FLAG) != 0) {
            q = elements[++index];
        }
        U_ASSERT((q & PRIMARY_STEP_MASK) == 0);
        return q;
    }
}

uint32_t
CollationRootElements::getSecondaryAfter(int32_t index, uint32_t s) const {
    uint32_t secTer;
    uint32_t secLimit;
    if(index == 0) {
        // primary = 0
        U_ASSERT(s != 0);
        index = (int32_t)elements[IX_FIRST_SECONDARY_INDEX];
        secTer = elements[index];
        // Gap at the end of the secondary CE range.
        secLimit = 0x10000;
    } else {
        U_ASSERT(index >= (int32_t)elements[IX_FIRST_PRIMARY_INDEX]);
        secTer = getFirstSecTerForPrimary(index + 1);
        // If this is an explicit sec/ter unit, then it will be read once more.
        // Gap for secondaries of primary CEs.
        secLimit = getSecondaryBoundary();
    }
    for(;;) {
        uint32_t sec = secTer >> 16;
        if(sec > s) { return sec; }
        secTer = elements[++index];
        if((secTer & SEC_TER_DELTA_FLAG) == 0) { return secLimit; }
    }
}

uint32_t
CollationRootElements::getTertiaryAfter(int32_t index, uint32_t s, uint32_t t) const {
    uint32_t secTer;
    uint32_t terLimit;
    if(index == 0) {
        // primary = 0
        if(s == 0) {
            U_ASSERT(t != 0);
            index = (int32_t)elements[IX_FIRST_TERTIARY_INDEX];
            // Gap at the end of the tertiary CE range.
            terLimit = 0x4000;
        } else {
            index = (int32_t)elements[IX_FIRST_SECONDARY_INDEX];
            // Gap for tertiaries of primary/secondary CEs.
            terLimit = getTertiaryBoundary();
        }
        secTer = elements[index] & ~SEC_TER_DELTA_FLAG;
    } else {
        U_ASSERT(index >= (int32_t)elements[IX_FIRST_PRIMARY_INDEX]);
        secTer = getFirstSecTerForPrimary(index + 1);
        // If this is an explicit sec/ter unit, then it will be read once more.
        terLimit = getTertiaryBoundary();
    }
    uint32_t st = (s << 16) | t;
    for(;;) {
        if(secTer > st) {
            U_ASSERT((secTer >> 16) == s);
            return secTer & 0xffff;
        }
        secTer = elements[++index];
        // No tertiary greater than t for this primary+secondary.
        if((secTer & SEC_TER_DELTA_FLAG) == 0 || (secTer >> 16) > s) { return terLimit; }
        secTer &= ~SEC_TER_DELTA_FLAG;
    }
}

uint32_t
CollationRootElements::getFirstSecTerForPrimary(int32_t index) const {
    uint32_t secTer = elements[index];
    if((secTer & SEC_TER_DELTA_FLAG) == 0) {
        // No sec/ter delta.
        return Collation::COMMON_SEC_AND_TER_CE;
    }
    secTer &= ~SEC_TER_DELTA_FLAG;
    if(secTer > Collation::COMMON_SEC_AND_TER_CE) {
        // Implied sec/ter.
        return Collation::COMMON_SEC_AND_TER_CE;
    }
    // Explicit sec/ter below common/common.
    return secTer;
}

int32_t
CollationRootElements::findPrimary(uint32_t p) const {
    // Requirement: p must occur as a root primary.
    U_ASSERT((p & 0xff) == 0);  // at most a 3-byte primary
    int32_t index = findP(p);
    // If p is in a range, then we just assume that p is an actual primary in this range.
    // (Too cumbersome/expensive to check.)
    // Otherwise, it must be an exact match.
    U_ASSERT(isEndOfPrimaryRange(elements[index + 1]) || p == (elements[index] & 0xffffff00));
    return index;
}

int32_t
CollationRootElements::findP(uint32_t p) const {
    // p need not occur as a root primary.
    // For example, it might be a reordering group boundary.
    U_ASSERT((p >> 24) != Collation::UNASSIGNED_IMPLICIT_BYTE);
    // modified binary search
    int32_t start = (int32_t)elements[IX_FIRST_PRIMARY_INDEX];
    U_ASSERT(p >= elements[start]);
    int32_t limit = length - 1;
    U_ASSERT(elements[limit] >= PRIMARY_SENTINEL);
    U_ASSERT(p < elements[limit]);
    while((start + 1) < limit) {
        // Invariant: elements[start] and elements[limit] are primaries,
        // and elements[start]<=p<=elements[limit].
        int32_t i = (start + limit) / 2;
        uint32_t q = elements[i];
        if((q & SEC_TER_DELTA_FLAG) != 0) {
            // Find the next primary.
            int32_t j = i + 1;
            for(;;) {
                if(j == limit) { break; }
                q = elements[j];
                if((q & SEC_TER_DELTA_FLAG) == 0) {
                    i = j;
                    break;
                }
                ++j;
            }
            if((q & SEC_TER_DELTA_FLAG) != 0) {
                // Find the preceding primary.
                j = i - 1;
                for(;;) {
                    if(j == start) { break; }
                    q = elements[j];
                    if((q & SEC_TER_DELTA_FLAG) == 0) {
                        i = j;
                        break;
                    }
                    --j;
                }
                if((q & SEC_TER_DELTA_FLAG) != 0) {
                    // No primary between start and limit.
                    break;
                }
            }
        }
        if(p < (q & 0xffffff00)) {  // Reset the "step" bits of a range end primary.
            limit = i;
        } else {
            start = i;
        }
    }
    return start;
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationruleparser.cpp
*
* (replaced the former ucol_tok.cpp)
*
* created on: 2013apr10
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION




















U_NAMESPACE_BEGIN

namespace {

static const UChar BEFORE[] = { 0x5b, 0x62, 0x65, 0x66, 0x6f, 0x72, 0x65, 0 };  // "[before"
const int32_t BEFORE_LENGTH = 7;

}  // namespace

CollationRuleParser::Sink::~Sink() {}

void
CollationRuleParser::Sink::suppressContractions(const UnicodeSet &, const char *&, UErrorCode &) {}

void
CollationRuleParser::Sink::optimize(const UnicodeSet &, const char *&, UErrorCode &) {}

CollationRuleParser::Importer::~Importer() {}

CollationRuleParser::CollationRuleParser(const CollationData *base, UErrorCode &errorCode)
        : nfd(*Normalizer2::getNFDInstance(errorCode)),
          nfc(*Normalizer2::getNFCInstance(errorCode)),
          rules(NULL), baseData(base), settings(NULL),
          parseError(NULL), errorReason(NULL),
          sink(NULL), importer(NULL),
          ruleIndex(0) {
}

CollationRuleParser::~CollationRuleParser() {
}

void
CollationRuleParser::parse(const UnicodeString &ruleString,
                           CollationSettings &outSettings,
                           UParseError *outParseError,
                           UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    settings = &outSettings;
    parseError = outParseError;
    if(parseError != NULL) {
        parseError->line = 0;
        parseError->offset = -1;
        parseError->preContext[0] = 0;
        parseError->postContext[0] = 0;
    }
    errorReason = NULL;
    parse(ruleString, errorCode);
}

void
CollationRuleParser::parse(const UnicodeString &ruleString, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    rules = &ruleString;
    ruleIndex = 0;

    while(ruleIndex < rules->length()) {
        UChar c = rules->charAt(ruleIndex);
        if(PatternProps::isWhiteSpace(c)) {
            ++ruleIndex;
            continue;
        }
        switch(c) {
        case 0x26:  // '&'
            parseRuleChain(errorCode);
            break;
        case 0x5b:  // '['
            parseSetting(errorCode);
            break;
        case 0x23:  // '#' starts a comment, until the end of the line
            ruleIndex = skipComment(ruleIndex + 1);
            break;
        case 0x40:  // '@' is equivalent to [backwards 2]
            settings->setFlag(CollationSettings::BACKWARD_SECONDARY,
                              UCOL_ON, 0, errorCode);
            ++ruleIndex;
            break;
        case 0x21:  // '!' used to turn on Thai/Lao character reversal
            // Accept but ignore. The root collator has contractions
            // that are equivalent to the character reversal, where appropriate.
            ++ruleIndex;
            break;
        default:
            setParseError("expected a reset or setting or comment", errorCode);
            break;
        }
        if(U_FAILURE(errorCode)) { return; }
    }
}

void
CollationRuleParser::parseRuleChain(UErrorCode &errorCode) {
    int32_t resetStrength = parseResetAndPosition(errorCode);
    UBool isFirstRelation = TRUE;
    for(;;) {
        int32_t result = parseRelationOperator(errorCode);
        if(U_FAILURE(errorCode)) { return; }
        if(result < 0) {
            if(ruleIndex < rules->length() && rules->charAt(ruleIndex) == 0x23) {
                // '#' starts a comment, until the end of the line
                ruleIndex = skipComment(ruleIndex + 1);
                continue;
            }
            if(isFirstRelation) {
                setParseError("reset not followed by a relation", errorCode);
            }
            return;
        }
        int32_t strength = result & STRENGTH_MASK;
        if(resetStrength < UCOL_IDENTICAL) {
            // reset-before rule chain
            if(isFirstRelation) {
                if(strength != resetStrength) {
                    setParseError("reset-before strength differs from its first relation", errorCode);
                    return;
                }
            } else {
                if(strength < resetStrength) {
                    setParseError("reset-before strength followed by a stronger relation", errorCode);
                    return;
                }
            }
        }
        int32_t i = ruleIndex + (result >> OFFSET_SHIFT);  // skip over the relation operator
        if((result & STARRED_FLAG) == 0) {
            parseRelationStrings(strength, i, errorCode);
        } else {
            parseStarredCharacters(strength, i, errorCode);
        }
        if(U_FAILURE(errorCode)) { return; }
        isFirstRelation = FALSE;
    }
}

int32_t
CollationRuleParser::parseResetAndPosition(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return UCOL_DEFAULT; }
    int32_t i = skipWhiteSpace(ruleIndex + 1);
    int32_t j;
    UChar c;
    int32_t resetStrength;
    if(rules->compare(i, BEFORE_LENGTH, BEFORE, 0, BEFORE_LENGTH) == 0 &&
            (j = i + BEFORE_LENGTH) < rules->length() &&
            PatternProps::isWhiteSpace(rules->charAt(j)) &&
            ((j = skipWhiteSpace(j + 1)) + 1) < rules->length() &&
            0x31 <= (c = rules->charAt(j)) && c <= 0x33 &&
            rules->charAt(j + 1) == 0x5d) {
        // &[before n] with n=1 or 2 or 3
        resetStrength = UCOL_PRIMARY + (c - 0x31);
        i = skipWhiteSpace(j + 2);
    } else {
        resetStrength = UCOL_IDENTICAL;
    }
    if(i >= rules->length()) {
        setParseError("reset without position", errorCode);
        return UCOL_DEFAULT;
    }
    UnicodeString str;
    if(rules->charAt(i) == 0x5b) {  // '['
        i = parseSpecialPosition(i, str, errorCode);
    } else {
        i = parseTailoringString(i, str, errorCode);
    }
    sink->addReset(resetStrength, str, errorReason, errorCode);
    if(U_FAILURE(errorCode)) { setErrorContext(); }
    ruleIndex = i;
    return resetStrength;
}

int32_t
CollationRuleParser::parseRelationOperator(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return UCOL_DEFAULT; }
    ruleIndex = skipWhiteSpace(ruleIndex);
    if(ruleIndex >= rules->length()) { return UCOL_DEFAULT; }
    int32_t strength;
    int32_t i = ruleIndex;
    UChar c = rules->charAt(i++);
    switch(c) {
    case 0x3c:  // '<'
        if(i < rules->length() && rules->charAt(i) == 0x3c) {  // <<
            ++i;
            if(i < rules->length() && rules->charAt(i) == 0x3c) {  // <<<
                ++i;
                if(i < rules->length() && rules->charAt(i) == 0x3c) {  // <<<<
                    ++i;
                    strength = UCOL_QUATERNARY;
                } else {
                    strength = UCOL_TERTIARY;
                }
            } else {
                strength = UCOL_SECONDARY;
            }
        } else {
            strength = UCOL_PRIMARY;
        }
        if(i < rules->length() && rules->charAt(i) == 0x2a) {  // '*'
            ++i;
            strength |= STARRED_FLAG;
        }
        break;
    case 0x3b:  // ';' same as <<
        strength = UCOL_SECONDARY;
        break;
    case 0x2c:  // ',' same as <<<
        strength = UCOL_TERTIARY;
        break;
    case 0x3d:  // '='
        strength = UCOL_IDENTICAL;
        if(i < rules->length() && rules->charAt(i) == 0x2a) {  // '*'
            ++i;
            strength |= STARRED_FLAG;
        }
        break;
    default:
        return UCOL_DEFAULT;
    }
    return ((i - ruleIndex) << OFFSET_SHIFT) | strength;
}

void
CollationRuleParser::parseRelationStrings(int32_t strength, int32_t i, UErrorCode &errorCode) {
    // Parse
    //     prefix | str / extension
    // where prefix and extension are optional.
    UnicodeString prefix, str, extension;
    i = parseTailoringString(i, str, errorCode);
    if(U_FAILURE(errorCode)) { return; }
    UChar next = (i < rules->length()) ? rules->charAt(i) : 0;
    if(next == 0x7c) {  // '|' separates the context prefix from the string.
        prefix = str;
        i = parseTailoringString(i + 1, str, errorCode);
        if(U_FAILURE(errorCode)) { return; }
        next = (i < rules->length()) ? rules->charAt(i) : 0;
    }
    if(next == 0x2f) {  // '/' separates the string from the extension.
        i = parseTailoringString(i + 1, extension, errorCode);
    }
    if(!prefix.isEmpty()) {
        UChar32 prefix0 = prefix.char32At(0);
        UChar32 c = str.char32At(0);
        if(!nfc.hasBoundaryBefore(prefix0) || !nfc.hasBoundaryBefore(c)) {
            setParseError("in 'prefix|str', prefix and str must each start with an NFC boundary",
                          errorCode);
            return;
        }
    }
    sink->addRelation(strength, prefix, str, extension, errorReason, errorCode);
    if(U_FAILURE(errorCode)) { setErrorContext(); }
    ruleIndex = i;
}

void
CollationRuleParser::parseStarredCharacters(int32_t strength, int32_t i, UErrorCode &errorCode) {
    UnicodeString empty, raw;
    i = parseString(skipWhiteSpace(i), raw, errorCode);
    if(U_FAILURE(errorCode)) { return; }
    if(raw.isEmpty()) {
        setParseError("missing starred-relation string", errorCode);
        return;
    }
    UChar32 prev = -1;
    int32_t j = 0;
    for(;;) {
        while(j < raw.length()) {
            UChar32 c = raw.char32At(j);
            if(!nfd.isInert(c)) {
                setParseError("starred-relation string is not all NFD-inert", errorCode);
                return;
            }
            sink->addRelation(strength, empty, UnicodeString(c), empty, errorReason, errorCode);
            if(U_FAILURE(errorCode)) {
                setErrorContext();
                return;
            }
            j += U16_LENGTH(c);
            prev = c;
        }
        if(i >= rules->length() || rules->charAt(i) != 0x2d) {  // '-'
            break;
        }
        if(prev < 0) {
            setParseError("range without start in starred-relation string", errorCode);
            return;
        }
        i = parseString(i + 1, raw, errorCode);
        if(U_FAILURE(errorCode)) { return; }
        if(raw.isEmpty()) {
            setParseError("range without end in starred-relation string", errorCode);
            return;
        }
        UChar32 c = raw.char32At(0);
        if(c < prev) {
            setParseError("range start greater than end in starred-relation string", errorCode);
            return;
        }
        // range prev-c
        UnicodeString s;
        while(++prev <= c) {
            if(!nfd.isInert(prev)) {
                setParseError("starred-relation string range is not all NFD-inert", errorCode);
                return;
            }
            if(U_IS_SURROGATE(prev)) {
                setParseError("starred-relation string range contains a surrogate", errorCode);
                return;
            }
            if(0xfffd <= prev && prev <= 0xffff) {
                setParseError("starred-relation string range contains U+FFFD, U+FFFE or U+FFFF", errorCode);
                return;
            }
            s.setTo(prev);
            sink->addRelation(strength, empty, s, empty, errorReason, errorCode);
            if(U_FAILURE(errorCode)) {
                setErrorContext();
                return;
            }
        }
        prev = -1;
        j = U16_LENGTH(c);
    }
    ruleIndex = skipWhiteSpace(i);
}

int32_t
CollationRuleParser::parseTailoringString(int32_t i, UnicodeString &raw, UErrorCode &errorCode) {
    i = parseString(skipWhiteSpace(i), raw, errorCode);
    if(U_SUCCESS(errorCode) && raw.isEmpty()) {
        setParseError("missing relation string", errorCode);
    }
    return skipWhiteSpace(i);
}

int32_t
CollationRuleParser::parseString(int32_t i, UnicodeString &raw, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return i; }
    raw.remove();
    while(i < rules->length()) {
        UChar32 c = rules->charAt(i++);
        if(isSyntaxChar(c)) {
            if(c == 0x27) {  // apostrophe
                if(i < rules->length() && rules->charAt(i) == 0x27) {
                    // Double apostrophe, encodes a single one.
                    raw.append((UChar)0x27);
                    ++i;
                    continue;
                }
                // Quote literal text until the next single apostrophe.
                for(;;) {
                    if(i == rules->length()) {
                        setParseError("quoted literal text missing terminating apostrophe", errorCode);
                        return i;
                    }
                    c = rules->charAt(i++);
                    if(c == 0x27) {
                        if(i < rules->length() && rules->charAt(i) == 0x27) {
                            // Double apostrophe inside quoted literal text,
                            // still encodes a single apostrophe.
                            ++i;
                        } else {
                            break;
                        }
                    }
                    raw.append((UChar)c);
                }
            } else if(c == 0x5c) {  // backslash
                if(i == rules->length()) {
                    setParseError("backslash escape at the end of the rule string", errorCode);
                    return i;
                }
                c = rules->char32At(i);
                raw.append(c);
                i += U16_LENGTH(c);
            } else {
                // Any other syntax character terminates a string.
                --i;
                break;
            }
        } else if(PatternProps::isWhiteSpace(c)) {
            // Unquoted white space terminates a string.
            --i;
            break;
        } else {
            raw.append((UChar)c);
        }
    }
    for(int32_t j = 0; j < raw.length();) {
        UChar32 c = raw.char32At(j);
        if(U_IS_SURROGATE(c)) {
            setParseError("string contains an unpaired surrogate", errorCode);
            return i;
        }
        if(0xfffd <= c && c <= 0xffff) {
            setParseError("string contains U+FFFD, U+FFFE or U+FFFF", errorCode);
            return i;
        }
        j += U16_LENGTH(c);
    }
    return i;
}

namespace {

static const char *const positions[] = {
    "first tertiary ignorable",
    "last tertiary ignorable",
    "first secondary ignorable",
    "last secondary ignorable",
    "first primary ignorable",
    "last primary ignorable",
    "first variable",
    "last variable",
    "first regular",
    "last regular",
    "first implicit",
    "last implicit",
    "first trailing",
    "last trailing"
};

}  // namespace

int32_t
CollationRuleParser::parseSpecialPosition(int32_t i, UnicodeString &str, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    UnicodeString raw;
    int32_t j = readWords(i + 1, raw);
    if(j > i && rules->charAt(j) == 0x5d && !raw.isEmpty()) {  // words end with ]
        ++j;
        for(int32_t pos = 0; pos < UPRV_LENGTHOF(positions); ++pos) {
            if(raw == UnicodeString(positions[pos], -1, US_INV)) {
                str.setTo((UChar)POS_LEAD).append((UChar)(POS_BASE + pos));
                return j;
            }
        }
        if(raw == UNICODE_STRING_SIMPLE("top")) {
            str.setTo((UChar)POS_LEAD).append((UChar)(POS_BASE + LAST_REGULAR));
            return j;
        }
        if(raw == UNICODE_STRING_SIMPLE("variable top")) {
            str.setTo((UChar)POS_LEAD).append((UChar)(POS_BASE + LAST_VARIABLE));
            return j;
        }
    }
    setParseError("not a valid special reset position", errorCode);
    return i;
}

void
CollationRuleParser::parseSetting(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    UnicodeString raw;
    int32_t i = ruleIndex + 1;
    int32_t j = readWords(i, raw);
    if(j <= i || raw.isEmpty()) {
        setParseError("expected a setting/option at '['", errorCode);
    }
    if(rules->charAt(j) == 0x5d) {  // words end with ]
        ++j;
        if(raw.startsWith(UNICODE_STRING_SIMPLE("reorder")) &&
                (raw.length() == 7 || raw.charAt(7) == 0x20)) {
            parseReordering(raw, errorCode);
            ruleIndex = j;
            return;
        }
        if(raw == UNICODE_STRING_SIMPLE("backwards 2")) {
            settings->setFlag(CollationSettings::BACKWARD_SECONDARY,
                              UCOL_ON, 0, errorCode);
            ruleIndex = j;
            return;
        }
        UnicodeString v;
        int32_t valueIndex = raw.lastIndexOf((UChar)0x20);
        if(valueIndex >= 0) {
            v.setTo(raw, valueIndex + 1);
            raw.truncate(valueIndex);
        }
        if(raw == UNICODE_STRING_SIMPLE("strength") && v.length() == 1) {
            int32_t value = UCOL_DEFAULT;
            UChar c = v.charAt(0);
            if(0x31 <= c && c <= 0x34) {  // 1..4
                value = UCOL_PRIMARY + (c - 0x31);
            } else if(c == 0x49) {  // 'I'
                value = UCOL_IDENTICAL;
            }
            if(value != UCOL_DEFAULT) {
                settings->setStrength(value, 0, errorCode);
                ruleIndex = j;
                return;
            }
        } else if(raw == UNICODE_STRING_SIMPLE("alternate")) {
            UColAttributeValue value = UCOL_DEFAULT;
            if(v == UNICODE_STRING_SIMPLE("non-ignorable")) {
                value = UCOL_NON_IGNORABLE;
            } else if(v == UNICODE_STRING_SIMPLE("shifted")) {
                value = UCOL_SHIFTED;
            }
            if(value != UCOL_DEFAULT) {
                settings->setAlternateHandling(value, 0, errorCode);
                ruleIndex = j;
                return;
            }
        } else if(raw == UNICODE_STRING_SIMPLE("maxVariable")) {
            int32_t value = UCOL_DEFAULT;
            if(v == UNICODE_STRING_SIMPLE("space")) {
                value = CollationSettings::MAX_VAR_SPACE;
            } else if(v == UNICODE_STRING_SIMPLE("punct")) {
                value = CollationSettings::MAX_VAR_PUNCT;
            } else if(v == UNICODE_STRING_SIMPLE("symbol")) {
                value = CollationSettings::MAX_VAR_SYMBOL;
            } else if(v == UNICODE_STRING_SIMPLE("currency")) {
                value = CollationSettings::MAX_VAR_CURRENCY;
            }
            if(value != UCOL_DEFAULT) {
                settings->setMaxVariable(value, 0, errorCode);
                settings->variableTop = baseData->getLastPrimaryForGroup(
                    UCOL_REORDER_CODE_FIRST + value);
                U_ASSERT(settings->variableTop != 0);
                ruleIndex = j;
                return;
            }
        } else if(raw == UNICODE_STRING_SIMPLE("caseFirst")) {
            UColAttributeValue value = UCOL_DEFAULT;
            if(v == UNICODE_STRING_SIMPLE("off")) {
                value = UCOL_OFF;
            } else if(v == UNICODE_STRING_SIMPLE("lower")) {
                value = UCOL_LOWER_FIRST;
            } else if(v == UNICODE_STRING_SIMPLE("upper")) {
                value = UCOL_UPPER_FIRST;
            }
            if(value != UCOL_DEFAULT) {
                settings->setCaseFirst(value, 0, errorCode);
                ruleIndex = j;
                return;
            }
        } else if(raw == UNICODE_STRING_SIMPLE("caseLevel")) {
            UColAttributeValue value = getOnOffValue(v);
            if(value != UCOL_DEFAULT) {
                settings->setFlag(CollationSettings::CASE_LEVEL, value, 0, errorCode);
                ruleIndex = j;
                return;
            }
        } else if(raw == UNICODE_STRING_SIMPLE("normalization")) {
            UColAttributeValue value = getOnOffValue(v);
            if(value != UCOL_DEFAULT) {
                settings->setFlag(CollationSettings::CHECK_FCD, value, 0, errorCode);
                ruleIndex = j;
                return;
            }
        } else if(raw == UNICODE_STRING_SIMPLE("numericOrdering")) {
            UColAttributeValue value = getOnOffValue(v);
            if(value != UCOL_DEFAULT) {
                settings->setFlag(CollationSettings::NUMERIC, value, 0, errorCode);
                ruleIndex = j;
                return;
            }
        } else if(raw == UNICODE_STRING_SIMPLE("hiraganaQ")) {
            UColAttributeValue value = getOnOffValue(v);
            if(value != UCOL_DEFAULT) {
                if(value == UCOL_ON) {
                    setParseError("[hiraganaQ on] is not supported", errorCode);
                }
                ruleIndex = j;
                return;
            }
        } else if(raw == UNICODE_STRING_SIMPLE("import")) {
            CharString lang;
            lang.appendInvariantChars(v, errorCode);
            if(errorCode == U_MEMORY_ALLOCATION_ERROR) { return; }
            // BCP 47 language tag -> ICU locale ID
            char localeID[ULOC_FULLNAME_CAPACITY];
            int32_t parsedLength;
            int32_t length = uloc_forLanguageTag(lang.data(), localeID, ULOC_FULLNAME_CAPACITY,
                                                 &parsedLength, &errorCode);
            if(U_FAILURE(errorCode) ||
                    parsedLength != lang.length() || length >= ULOC_FULLNAME_CAPACITY) {
                errorCode = U_ZERO_ERROR;
                setParseError("expected language tag in [import langTag]", errorCode);
                return;
            }
            // localeID minus all keywords
            char baseID[ULOC_FULLNAME_CAPACITY];
            length = uloc_getBaseName(localeID, baseID, ULOC_FULLNAME_CAPACITY, &errorCode);
            if(U_FAILURE(errorCode) || length >= ULOC_KEYWORDS_CAPACITY) {
                errorCode = U_ZERO_ERROR;
                setParseError("expected language tag in [import langTag]", errorCode);
                return;
            }
            if(length == 0) {
                uprv_strcpy(baseID, "root");
            } else if(*baseID == '_') {
                uprv_memmove(baseID + 3, baseID, length + 1);
                uprv_memcpy(baseID, "und", 3);
            }
            // @collation=type, or length=0 if not specified
            char collationType[ULOC_KEYWORDS_CAPACITY];
            length = uloc_getKeywordValue(localeID, "collation",
                                          collationType, ULOC_KEYWORDS_CAPACITY,
                                          &errorCode);
            if(U_FAILURE(errorCode) || length >= ULOC_KEYWORDS_CAPACITY) {
                errorCode = U_ZERO_ERROR;
                setParseError("expected language tag in [import langTag]", errorCode);
                return;
            }
            if(importer == NULL) {
                setParseError("[import langTag] is not supported", errorCode);
            } else {
                UnicodeString importedRules;
                importer->getRules(baseID, length > 0 ? collationType : "standard",
                                   importedRules, errorReason, errorCode);
                if(U_FAILURE(errorCode)) {
                    if(errorReason == NULL) {
                        errorReason = "[import langTag] failed";
                    }
                    setErrorContext();
                    return;
                }
                const UnicodeString *outerRules = rules;
                int32_t outerRuleIndex = ruleIndex;
                parse(importedRules, errorCode);
                if(U_FAILURE(errorCode)) {
                    if(parseError != NULL) {
                        parseError->offset = outerRuleIndex;
                    }
                }
                rules = outerRules;
                ruleIndex = j;
            }
            return;
        }
    } else if(rules->charAt(j) == 0x5b) {  // words end with [
        UnicodeSet set;
        j = parseUnicodeSet(j, set, errorCode);
        if(U_FAILURE(errorCode)) { return; }
        if(raw == UNICODE_STRING_SIMPLE("optimize")) {
            sink->optimize(set, errorReason, errorCode);
            if(U_FAILURE(errorCode)) { setErrorContext(); }
            ruleIndex = j;
            return;
        } else if(raw == UNICODE_STRING_SIMPLE("suppressContractions")) {
            sink->suppressContractions(set, errorReason, errorCode);
            if(U_FAILURE(errorCode)) { setErrorContext(); }
            ruleIndex = j;
            return;
        }
    }
    setParseError("not a valid setting/option", errorCode);
}

void
CollationRuleParser::parseReordering(const UnicodeString &raw, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    int32_t i = 7;  // after "reorder"
    if(i == raw.length()) {
        // empty [reorder] with no codes
        settings->resetReordering();
        return;
    }
    // Parse the codes in [reorder aa bb cc].
    UVector32 reorderCodes(errorCode);
    if(U_FAILURE(errorCode)) { return; }
    CharString word;
    while(i < raw.length()) {
        ++i;  // skip the word-separating space
        int32_t limit = raw.indexOf((UChar)0x20, i);
        if(limit < 0) { limit = raw.length(); }
        word.clear().appendInvariantChars(raw.tempSubStringBetween(i, limit), errorCode);
        if(U_FAILURE(errorCode)) { return; }
        int32_t code = getReorderCode(word.data());
        if(code < 0) {
            setParseError("unknown script or reorder code", errorCode);
            return;
        }
        reorderCodes.addElement(code, errorCode);
        if(U_FAILURE(errorCode)) { return; }
        i = limit;
    }
    settings->setReordering(*baseData, reorderCodes.getBuffer(), reorderCodes.size(), errorCode);
}

static const char *const gSpecialReorderCodes[] = {
    "space", "punct", "symbol", "currency", "digit"
};

int32_t
CollationRuleParser::getReorderCode(const char *word) {
    for(int32_t i = 0; i < UPRV_LENGTHOF(gSpecialReorderCodes); ++i) {
        if(uprv_stricmp(word, gSpecialReorderCodes[i]) == 0) {
            return UCOL_REORDER_CODE_FIRST + i;
        }
    }
    int32_t script = u_getPropertyValueEnum(UCHAR_SCRIPT, word);
    if(script >= 0) {
        return script;
    }
    if(uprv_stricmp(word, "others") == 0) {
        return UCOL_REORDER_CODE_OTHERS;  // same as Zzzz = USCRIPT_UNKNOWN
    }
    return -1;
}

UColAttributeValue
CollationRuleParser::getOnOffValue(const UnicodeString &s) {
    if(s == UNICODE_STRING_SIMPLE("on")) {
        return UCOL_ON;
    } else if(s == UNICODE_STRING_SIMPLE("off")) {
        return UCOL_OFF;
    } else {
        return UCOL_DEFAULT;
    }
}

int32_t
CollationRuleParser::parseUnicodeSet(int32_t i, UnicodeSet &set, UErrorCode &errorCode) {
    // Collect a UnicodeSet pattern between a balanced pair of [brackets].
    int32_t level = 0;
    int32_t j = i;
    for(;;) {
        if(j == rules->length()) {
            setParseError("unbalanced UnicodeSet pattern brackets", errorCode);
            return j;
        }
        UChar c = rules->charAt(j++);
        if(c == 0x5b) {  // '['
            ++level;
        } else if(c == 0x5d) {  // ']'
            if(--level == 0) { break; }
        }
    }
    set.applyPattern(rules->tempSubStringBetween(i, j), errorCode);
    if(U_FAILURE(errorCode)) {
        errorCode = U_ZERO_ERROR;
        setParseError("not a valid UnicodeSet pattern", errorCode);
        return j;
    }
    j = skipWhiteSpace(j);
    if(j == rules->length() || rules->charAt(j) != 0x5d) {
        setParseError("missing option-terminating ']' after UnicodeSet pattern", errorCode);
        return j;
    }
    return ++j;
}

int32_t
CollationRuleParser::readWords(int32_t i, UnicodeString &raw) const {
    static const UChar sp = 0x20;
    raw.remove();
    i = skipWhiteSpace(i);
    for(;;) {
        if(i >= rules->length()) { return 0; }
        UChar c = rules->charAt(i);
        if(isSyntaxChar(c) && c != 0x2d && c != 0x5f) {  // syntax except -_
            if(raw.isEmpty()) { return i; }
            if(raw.endsWith(&sp, 1)) {  // remove trailing space
                raw.truncate(raw.length() - 1);
            }
            return i;
        }
        if(PatternProps::isWhiteSpace(c)) {
            raw.append(sp);
            i = skipWhiteSpace(i + 1);
        } else {
            raw.append(c);
            ++i;
        }
    }
}

int32_t
CollationRuleParser::skipComment(int32_t i) const {
    // skip to past the newline
    while(i < rules->length()) {
        UChar c = rules->charAt(i++);
        // LF or FF or CR or NEL or LS or PS
        if(c == 0xa || c == 0xc || c == 0xd || c == 0x85 || c == 0x2028 || c == 0x2029) {
            // Unicode Newline Guidelines: "A readline function should stop at NLF, LS, FF, or PS."
            // NLF (new line function) = CR or LF or CR+LF or NEL.
            // No need to collect all of CR+LF because a following LF will be ignored anyway.
            break;
        }
    }
    return i;
}

void
CollationRuleParser::setParseError(const char *reason, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    // Error code consistent with the old parser (from ca. 2001),
    // rather than U_PARSE_ERROR;
    errorCode = U_INVALID_FORMAT_ERROR;
    errorReason = reason;
    if(parseError != NULL) { setErrorContext(); }
}

void
CollationRuleParser::setErrorContext() {
    if(parseError == NULL) { return; }

    // Note: This relies on the calling code maintaining the ruleIndex
    // at a position that is useful for debugging.
    // For example, at the beginning of a reset or relation etc.
    parseError->offset = ruleIndex;
    parseError->line = 0;  // We are not counting line numbers.

    // before ruleIndex
    int32_t start = ruleIndex - (U_PARSE_CONTEXT_LEN - 1);
    if(start < 0) {
        start = 0;
    } else if(start > 0 && U16_IS_TRAIL(rules->charAt(start))) {
        ++start;
    }
    int32_t length = ruleIndex - start;
    rules->extract(start, length, parseError->preContext);
    parseError->preContext[length] = 0;

    // starting from ruleIndex
    length = rules->length() - ruleIndex;
    if(length >= U_PARSE_CONTEXT_LEN) {
        length = U_PARSE_CONTEXT_LEN - 1;
        if(U16_IS_LEAD(rules->charAt(ruleIndex + length - 1))) {
            --length;
        }
    }
    rules->extract(ruleIndex, length, parseError->postContext);
    parseError->postContext[length] = 0;
}

UBool
CollationRuleParser::isSyntaxChar(UChar32 c) {
    return 0x21 <= c && c <= 0x7e &&
            (c <= 0x2f || (0x3a <= c && c <= 0x40) ||
            (0x5b <= c && c <= 0x60) || (0x7b <= c));
}

int32_t
CollationRuleParser::skipWhiteSpace(int32_t i) const {
    while(i < rules->length() && PatternProps::isWhiteSpace(rules->charAt(i))) {
        ++i;
    }
    return i;
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationsets.cpp
*
* created on: 2013feb09
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION













U_NAMESPACE_BEGIN

U_CDECL_BEGIN

static UBool U_CALLCONV
enumTailoredRange(const void *context, UChar32 start, UChar32 end, uint32_t ce32) {
    if(ce32 == Collation::FALLBACK_CE32) {
        return TRUE;  // fallback to base, not tailored
    }
    TailoredSet *ts = (TailoredSet *)context;
    return ts->handleCE32(start, end, ce32);
}

U_CDECL_END

void
TailoredSet::forData(const CollationData *d, UErrorCode &ec) {
    if(U_FAILURE(ec)) { return; }
    errorCode = ec;  // Preserve info & warning codes.
    data = d;
    baseData = d->base;
    U_ASSERT(baseData != NULL);
    utrie2_enum(data->trie, NULL, enumTailoredRange, this);
    ec = errorCode;
}

UBool
TailoredSet::handleCE32(UChar32 start, UChar32 end, uint32_t ce32) {
    U_ASSERT(ce32 != Collation::FALLBACK_CE32);
    if(Collation::isSpecialCE32(ce32)) {
        ce32 = data->getIndirectCE32(ce32);
        if(ce32 == Collation::FALLBACK_CE32) {
            return U_SUCCESS(errorCode);
        }
    }
    do {
        uint32_t baseCE32 = baseData->getFinalCE32(baseData->getCE32(start));
        // Do not just continue if ce32 == baseCE32 because
        // contractions and expansions in different data objects
        // normally differ even if they have the same data offsets.
        if(Collation::isSelfContainedCE32(ce32) && Collation::isSelfContainedCE32(baseCE32)) {
            // fastpath
            if(ce32 != baseCE32) {
                tailored->add(start);
            }
        } else {
            compare(start, ce32, baseCE32);
        }
    } while(++start <= end);
    return U_SUCCESS(errorCode);
}

void
TailoredSet::compare(UChar32 c, uint32_t ce32, uint32_t baseCE32) {
    if(Collation::isPrefixCE32(ce32)) {
        const UChar *p = data->contexts + Collation::indexFromCE32(ce32);
        ce32 = data->getFinalCE32(CollationData::readCE32(p));
        if(Collation::isPrefixCE32(baseCE32)) {
            const UChar *q = baseData->contexts + Collation::indexFromCE32(baseCE32);
            baseCE32 = baseData->getFinalCE32(CollationData::readCE32(q));
            comparePrefixes(c, p + 2, q + 2);
        } else {
            addPrefixes(data, c, p + 2);
        }
    } else if(Collation::isPrefixCE32(baseCE32)) {
        const UChar *q = baseData->contexts + Collation::indexFromCE32(baseCE32);
        baseCE32 = baseData->getFinalCE32(CollationData::readCE32(q));
        addPrefixes(baseData, c, q + 2);
    }

    if(Collation::isContractionCE32(ce32)) {
        const UChar *p = data->contexts + Collation::indexFromCE32(ce32);
        if((ce32 & Collation::CONTRACT_SINGLE_CP_NO_MATCH) != 0) {
            ce32 = Collation::NO_CE32;
        } else {
            ce32 = data->getFinalCE32(CollationData::readCE32(p));
        }
        if(Collation::isContractionCE32(baseCE32)) {
            const UChar *q = baseData->contexts + Collation::indexFromCE32(baseCE32);
            if((baseCE32 & Collation::CONTRACT_SINGLE_CP_NO_MATCH) != 0) {
                baseCE32 = Collation::NO_CE32;
            } else {
                baseCE32 = baseData->getFinalCE32(CollationData::readCE32(q));
            }
            compareContractions(c, p + 2, q + 2);
        } else {
            addContractions(c, p + 2);
        }
    } else if(Collation::isContractionCE32(baseCE32)) {
        const UChar *q = baseData->contexts + Collation::indexFromCE32(baseCE32);
        baseCE32 = baseData->getFinalCE32(CollationData::readCE32(q));
        addContractions(c, q + 2);
    }

    int32_t tag;
    if(Collation::isSpecialCE32(ce32)) {
        tag = Collation::tagFromCE32(ce32);
        U_ASSERT(tag != Collation::PREFIX_TAG);
        U_ASSERT(tag != Collation::CONTRACTION_TAG);
        // Currently, the tailoring data builder does not write offset tags.
        // They might be useful for saving space,
        // but they would complicate the builder,
        // and in tailorings we assume that performance of tailored characters is more important.
        U_ASSERT(tag != Collation::OFFSET_TAG);
    } else {
        tag = -1;
    }
    int32_t baseTag;
    if(Collation::isSpecialCE32(baseCE32)) {
        baseTag = Collation::tagFromCE32(baseCE32);
        U_ASSERT(baseTag != Collation::PREFIX_TAG);
        U_ASSERT(baseTag != Collation::CONTRACTION_TAG);
    } else {
        baseTag = -1;
    }

    // Non-contextual mappings, expansions, etc.
    if(baseTag == Collation::OFFSET_TAG) {
        // We might be comparing a tailoring CE which is a copy of
        // a base offset-tag CE, via the [optimize [set]] syntax
        // or when a single-character mapping was copied for tailored contractions.
        // Offset tags always result in long-primary CEs,
        // with common secondary/tertiary weights.
        if(!Collation::isLongPrimaryCE32(ce32)) {
            add(c);
            return;
        }
        int64_t dataCE = baseData->ces[Collation::indexFromCE32(baseCE32)];
        uint32_t p = Collation::getThreeBytePrimaryForOffsetData(c, dataCE);
        if(Collation::primaryFromLongPrimaryCE32(ce32) != p) {
            add(c);
            return;
        }
    }

    if(tag != baseTag) {
        add(c);
        return;
    }

    if(tag == Collation::EXPANSION32_TAG) {
        const uint32_t *ce32s = data->ce32s + Collation::indexFromCE32(ce32);
        int32_t length = Collation::lengthFromCE32(ce32);

        const uint32_t *baseCE32s = baseData->ce32s + Collation::indexFromCE32(baseCE32);
        int32_t baseLength = Collation::lengthFromCE32(baseCE32);

        if(length != baseLength) {
            add(c);
            return;
        }
        for(int32_t i = 0; i < length; ++i) {
            if(ce32s[i] != baseCE32s[i]) {
                add(c);
                break;
            }
        }
    } else if(tag == Collation::EXPANSION_TAG) {
        const int64_t *ces = data->ces + Collation::indexFromCE32(ce32);
        int32_t length = Collation::lengthFromCE32(ce32);

        const int64_t *baseCEs = baseData->ces + Collation::indexFromCE32(baseCE32);
        int32_t baseLength = Collation::lengthFromCE32(baseCE32);

        if(length != baseLength) {
            add(c);
            return;
        }
        for(int32_t i = 0; i < length; ++i) {
            if(ces[i] != baseCEs[i]) {
                add(c);
                break;
            }
        }
    } else if(tag == Collation::HANGUL_TAG) {
        UChar jamos[3];
        int32_t length = Hangul::decompose(c, jamos);
        if(tailored->contains(jamos[0]) || tailored->contains(jamos[1]) ||
                (length == 3 && tailored->contains(jamos[2]))) {
            add(c);
        }
    } else if(ce32 != baseCE32) {
        add(c);
    }
}

void
TailoredSet::comparePrefixes(UChar32 c, const UChar *p, const UChar *q) {
    // Parallel iteration over prefixes of both tables.
    UCharsTrie::Iterator prefixes(p, 0, errorCode);
    UCharsTrie::Iterator basePrefixes(q, 0, errorCode);
    const UnicodeString *tp = NULL;  // Tailoring prefix.
    const UnicodeString *bp = NULL;  // Base prefix.
    // Use a string with a U+FFFF as the limit sentinel.
    // U+FFFF is untailorable and will not occur in prefixes.
    UnicodeString none((UChar)0xffff);
    for(;;) {
        if(tp == NULL) {
            if(prefixes.next(errorCode)) {
                tp = &prefixes.getString();
            } else {
                tp = &none;
            }
        }
        if(bp == NULL) {
            if(basePrefixes.next(errorCode)) {
                bp = &basePrefixes.getString();
            } else {
                bp = &none;
            }
        }
        if(tp == &none && bp == &none) { break; }
        int32_t cmp = tp->compare(*bp);
        if(cmp < 0) {
            // tp occurs in the tailoring but not in the base.
            addPrefix(data, *tp, c, (uint32_t)prefixes.getValue());
            tp = NULL;
        } else if(cmp > 0) {
            // bp occurs in the base but not in the tailoring.
            addPrefix(baseData, *bp, c, (uint32_t)basePrefixes.getValue());
            bp = NULL;
        } else {
            setPrefix(*tp);
            compare(c, (uint32_t)prefixes.getValue(), (uint32_t)basePrefixes.getValue());
            resetPrefix();
            tp = NULL;
            bp = NULL;
        }
    }
}

void
TailoredSet::compareContractions(UChar32 c, const UChar *p, const UChar *q) {
    // Parallel iteration over suffixes of both tables.
    UCharsTrie::Iterator suffixes(p, 0, errorCode);
    UCharsTrie::Iterator baseSuffixes(q, 0, errorCode);
    const UnicodeString *ts = NULL;  // Tailoring suffix.
    const UnicodeString *bs = NULL;  // Base suffix.
    // Use a string with two U+FFFF as the limit sentinel.
    // U+FFFF is untailorable and will not occur in contractions except maybe
    // as a single suffix character for a root-collator boundary contraction.
    UnicodeString none((UChar)0xffff);
    none.append((UChar)0xffff);
    for(;;) {
        if(ts == NULL) {
            if(suffixes.next(errorCode)) {
                ts = &suffixes.getString();
            } else {
                ts = &none;
            }
        }
        if(bs == NULL) {
            if(baseSuffixes.next(errorCode)) {
                bs = &baseSuffixes.getString();
            } else {
                bs = &none;
            }
        }
        if(ts == &none && bs == &none) { break; }
        int32_t cmp = ts->compare(*bs);
        if(cmp < 0) {
            // ts occurs in the tailoring but not in the base.
            addSuffix(c, *ts);
            ts = NULL;
        } else if(cmp > 0) {
            // bs occurs in the base but not in the tailoring.
            addSuffix(c, *bs);
            bs = NULL;
        } else {
            suffix = ts;
            compare(c, (uint32_t)suffixes.getValue(), (uint32_t)baseSuffixes.getValue());
            suffix = NULL;
            ts = NULL;
            bs = NULL;
        }
    }
}

void
TailoredSet::addPrefixes(const CollationData *d, UChar32 c, const UChar *p) {
    UCharsTrie::Iterator prefixes(p, 0, errorCode);
    while(prefixes.next(errorCode)) {
        addPrefix(d, prefixes.getString(), c, (uint32_t)prefixes.getValue());
    }
}

void
TailoredSet::addPrefix(const CollationData *d, const UnicodeString &pfx, UChar32 c, uint32_t ce32) {
    setPrefix(pfx);
    ce32 = d->getFinalCE32(ce32);
    if(Collation::isContractionCE32(ce32)) {
        const UChar *p = d->contexts + Collation::indexFromCE32(ce32);
        addContractions(c, p + 2);
    }
    tailored->add(UnicodeString(unreversedPrefix).append(c));
    resetPrefix();
}

void
TailoredSet::addContractions(UChar32 c, const UChar *p) {
    UCharsTrie::Iterator suffixes(p, 0, errorCode);
    while(suffixes.next(errorCode)) {
        addSuffix(c, suffixes.getString());
    }
}

void
TailoredSet::addSuffix(UChar32 c, const UnicodeString &sfx) {
    tailored->add(UnicodeString(unreversedPrefix).append(c).append(sfx));
}

void
TailoredSet::add(UChar32 c) {
    if(unreversedPrefix.isEmpty() && suffix == NULL) {
        tailored->add(c);
    } else {
        UnicodeString s(unreversedPrefix);
        s.append(c);
        if(suffix != NULL) {
            s.append(*suffix);
        }
        tailored->add(s);
    }
}

ContractionsAndExpansions::CESink::~CESink() {}

U_CDECL_BEGIN

static UBool U_CALLCONV
enumCnERange(const void *context, UChar32 start, UChar32 end, uint32_t ce32) {
    ContractionsAndExpansions *cne = (ContractionsAndExpansions *)context;
    if(cne->checkTailored == 0) {
        // There is no tailoring.
        // No need to collect nor check the tailored set.
    } else if(cne->checkTailored < 0) {
        // Collect the set of code points with mappings in the tailoring data.
        if(ce32 == Collation::FALLBACK_CE32) {
            return TRUE;  // fallback to base, not tailored
        } else {
            cne->tailored.add(start, end);
        }
        // checkTailored > 0: Exclude tailored ranges from the base data enumeration.
    } else if(start == end) {
        if(cne->tailored.contains(start)) {
            return TRUE;
        }
    } else if(cne->tailored.containsSome(start, end)) {
        cne->ranges.set(start, end).removeAll(cne->tailored);
        int32_t count = cne->ranges.getRangeCount();
        for(int32_t i = 0; i < count; ++i) {
            cne->handleCE32(cne->ranges.getRangeStart(i), cne->ranges.getRangeEnd(i), ce32);
        }
        return U_SUCCESS(cne->errorCode);
    }
    cne->handleCE32(start, end, ce32);
    return U_SUCCESS(cne->errorCode);
}

U_CDECL_END

void
ContractionsAndExpansions::forData(const CollationData *d, UErrorCode &ec) {
    if(U_FAILURE(ec)) { return; }
    errorCode = ec;  // Preserve info & warning codes.
    // Add all from the data, can be tailoring or base.
    if(d->base != NULL) {
        checkTailored = -1;
    }
    data = d;
    utrie2_enum(data->trie, NULL, enumCnERange, this);
    if(d->base == NULL || U_FAILURE(errorCode)) {
        ec = errorCode;
        return;
    }
    // Add all from the base data but only for un-tailored code points.
    tailored.freeze();
    checkTailored = 1;
    data = d->base;
    utrie2_enum(data->trie, NULL, enumCnERange, this);
    ec = errorCode;
}

void
ContractionsAndExpansions::forCodePoint(const CollationData *d, UChar32 c, UErrorCode &ec) {
    if(U_FAILURE(ec)) { return; }
    errorCode = ec;  // Preserve info & warning codes.
    uint32_t ce32 = d->getCE32(c);
    if(ce32 == Collation::FALLBACK_CE32) {
        d = d->base;
        ce32 = d->getCE32(c);
    }
    data = d;
    handleCE32(c, c, ce32);
    ec = errorCode;
}

void
ContractionsAndExpansions::handleCE32(UChar32 start, UChar32 end, uint32_t ce32) {
    for(;;) {
        if((ce32 & 0xff) < Collation::SPECIAL_CE32_LOW_BYTE) {
            // !isSpecialCE32()
            if(sink != NULL) {
                sink->handleCE(Collation::ceFromSimpleCE32(ce32));
            }
            return;
        }
        switch(Collation::tagFromCE32(ce32)) {
        case Collation::FALLBACK_TAG:
            return;
        case Collation::RESERVED_TAG_3:
        case Collation::BUILDER_DATA_TAG:
        case Collation::LEAD_SURROGATE_TAG:
            if(U_SUCCESS(errorCode)) { errorCode = U_INTERNAL_PROGRAM_ERROR; }
            return;
        case Collation::LONG_PRIMARY_TAG:
            if(sink != NULL) {
                sink->handleCE(Collation::ceFromLongPrimaryCE32(ce32));
            }
            return;
        case Collation::LONG_SECONDARY_TAG:
            if(sink != NULL) {
                sink->handleCE(Collation::ceFromLongSecondaryCE32(ce32));
            }
            return;
        case Collation::LATIN_EXPANSION_TAG:
            if(sink != NULL) {
                ces[0] = Collation::latinCE0FromCE32(ce32);
                ces[1] = Collation::latinCE1FromCE32(ce32);
                sink->handleExpansion(ces, 2);
            }
            // Optimization: If we have a prefix,
            // then the relevant strings have been added already.
            if(unreversedPrefix.isEmpty()) {
                addExpansions(start, end);
            }
            return;
        case Collation::EXPANSION32_TAG:
            if(sink != NULL) {
                const uint32_t *ce32s = data->ce32s + Collation::indexFromCE32(ce32);
                int32_t length = Collation::lengthFromCE32(ce32);
                for(int32_t i = 0; i < length; ++i) {
                    ces[i] = Collation::ceFromCE32(*ce32s++);
                }
                sink->handleExpansion(ces, length);
            }
            // Optimization: If we have a prefix,
            // then the relevant strings have been added already.
            if(unreversedPrefix.isEmpty()) {
                addExpansions(start, end);
            }
            return;
        case Collation::EXPANSION_TAG:
            if(sink != NULL) {
                int32_t length = Collation::lengthFromCE32(ce32);
                sink->handleExpansion(data->ces + Collation::indexFromCE32(ce32), length);
            }
            // Optimization: If we have a prefix,
            // then the relevant strings have been added already.
            if(unreversedPrefix.isEmpty()) {
                addExpansions(start, end);
            }
            return;
        case Collation::PREFIX_TAG:
            handlePrefixes(start, end, ce32);
            return;
        case Collation::CONTRACTION_TAG:
            handleContractions(start, end, ce32);
            return;
        case Collation::DIGIT_TAG:
            // Fetch the non-numeric-collation CE32 and continue.
            ce32 = data->ce32s[Collation::indexFromCE32(ce32)];
            break;
        case Collation::U0000_TAG:
            U_ASSERT(start == 0 && end == 0);
            // Fetch the normal ce32 for U+0000 and continue.
            ce32 = data->ce32s[0];
            break;
        case Collation::HANGUL_TAG:
            if(sink != NULL) {
                // TODO: This should be optimized,
                // especially if [start..end] is the complete Hangul range. (assert that)
                UTF16CollationIterator iter(data, FALSE, NULL, NULL, NULL);
                UChar hangul[1] = { 0 };
                for(UChar32 c = start; c <= end; ++c) {
                    hangul[0] = (UChar)c;
                    iter.setText(hangul, hangul + 1);
                    int32_t length = iter.fetchCEs(errorCode);
                    if(U_FAILURE(errorCode)) { return; }
                    // Ignore the terminating non-CE.
                    U_ASSERT(length >= 2 && iter.getCE(length - 1) == Collation::NO_CE);
                    sink->handleExpansion(iter.getCEs(), length - 1);
                }
            }
            // Optimization: If we have a prefix,
            // then the relevant strings have been added already.
            if(unreversedPrefix.isEmpty()) {
                addExpansions(start, end);
            }
            return;
        case Collation::OFFSET_TAG:
            // Currently no need to send offset CEs to the sink.
            return;
        case Collation::IMPLICIT_TAG:
            // Currently no need to send implicit CEs to the sink.
            return;
        }
    }
}

void
ContractionsAndExpansions::handlePrefixes(
        UChar32 start, UChar32 end, uint32_t ce32) {
    const UChar *p = data->contexts + Collation::indexFromCE32(ce32);
    ce32 = CollationData::readCE32(p);  // Default if no prefix match.
    handleCE32(start, end, ce32);
    if(!addPrefixes) { return; }
    UCharsTrie::Iterator prefixes(p + 2, 0, errorCode);
    while(prefixes.next(errorCode)) {
        setPrefix(prefixes.getString());
        // Prefix/pre-context mappings are special kinds of contractions
        // that always yield expansions.
        addStrings(start, end, contractions);
        addStrings(start, end, expansions);
        handleCE32(start, end, (uint32_t)prefixes.getValue());
    }
    resetPrefix();
}

void
ContractionsAndExpansions::handleContractions(
        UChar32 start, UChar32 end, uint32_t ce32) {
    const UChar *p = data->contexts + Collation::indexFromCE32(ce32);
    if((ce32 & Collation::CONTRACT_SINGLE_CP_NO_MATCH) != 0) {
        // No match on the single code point.
        // We are underneath a prefix, and the default mapping is just
        // a fallback to the mappings for a shorter prefix.
        U_ASSERT(!unreversedPrefix.isEmpty());
    } else {
        ce32 = CollationData::readCE32(p);  // Default if no suffix match.
        U_ASSERT(!Collation::isContractionCE32(ce32));
        handleCE32(start, end, ce32);
    }
    UCharsTrie::Iterator suffixes(p + 2, 0, errorCode);
    while(suffixes.next(errorCode)) {
        suffix = &suffixes.getString();
        addStrings(start, end, contractions);
        if(!unreversedPrefix.isEmpty()) {
            addStrings(start, end, expansions);
        }
        handleCE32(start, end, (uint32_t)suffixes.getValue());
    }
    suffix = NULL;
}

void
ContractionsAndExpansions::addExpansions(UChar32 start, UChar32 end) {
    if(unreversedPrefix.isEmpty() && suffix == NULL) {
        if(expansions != NULL) {
            expansions->add(start, end);
        }
    } else {
        addStrings(start, end, expansions);
    }
}

void
ContractionsAndExpansions::addStrings(UChar32 start, UChar32 end, UnicodeSet *set) {
    if(set == NULL) { return; }
    UnicodeString s(unreversedPrefix);
    do {
        s.append(start);
        if(suffix != NULL) {
            s.append(*suffix);
        }
        set->add(s);
        s.truncate(unreversedPrefix.length());
    } while(++start <= end);
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationsettings.cpp
*
* created on: 2013feb07
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION











U_NAMESPACE_BEGIN

CollationSettings::CollationSettings(const CollationSettings &other)
        : SharedObject(other),
          options(other.options), variableTop(other.variableTop),
          reorderTable(NULL),
          minHighNoReorder(other.minHighNoReorder),
          reorderRanges(NULL), reorderRangesLength(0),
          reorderCodes(NULL), reorderCodesLength(0), reorderCodesCapacity(0),
          fastLatinOptions(other.fastLatinOptions) {
    UErrorCode errorCode = U_ZERO_ERROR;
    copyReorderingFrom(other, errorCode);
    if(fastLatinOptions >= 0) {
        uprv_memcpy(fastLatinPrimaries, other.fastLatinPrimaries, sizeof(fastLatinPrimaries));
    }
}

CollationSettings::~CollationSettings() {
    if(reorderCodesCapacity != 0) {
        uprv_free(const_cast<int32_t *>(reorderCodes));
    }
}

UBool
CollationSettings::operator==(const CollationSettings &other) const {
    if(options != other.options) { return FALSE; }
    if((options & ALTERNATE_MASK) != 0 && variableTop != other.variableTop) { return FALSE; }
    if(reorderCodesLength != other.reorderCodesLength) { return FALSE; }
    for(int32_t i = 0; i < reorderCodesLength; ++i) {
        if(reorderCodes[i] != other.reorderCodes[i]) { return FALSE; }
    }
    return TRUE;
}

int32_t
CollationSettings::hashCode() const {
    int32_t h = options << 8;
    if((options & ALTERNATE_MASK) != 0) { h ^= variableTop; }
    h ^= reorderCodesLength;
    for(int32_t i = 0; i < reorderCodesLength; ++i) {
        h ^= (reorderCodes[i] << i);
    }
    return h;
}

void
CollationSettings::resetReordering() {
    // When we turn off reordering, we want to set a NULL permutation
    // rather than a no-op permutation.
    // Keep the memory via reorderCodes and its capacity.
    reorderTable = NULL;
    minHighNoReorder = 0;
    reorderRangesLength = 0;
    reorderCodesLength = 0;
}

void
CollationSettings::aliasReordering(const CollationData &data, const int32_t *codes, int32_t length,
                                   const uint32_t *ranges, int32_t rangesLength,
                                   const uint8_t *table, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(table != NULL &&
            (rangesLength == 0 ?
                    !reorderTableHasSplitBytes(table) :
                    rangesLength >= 2 &&
                    // The first offset must be 0. The last offset must not be 0.
                    (ranges[0] & 0xffff) == 0 && (ranges[rangesLength - 1] & 0xffff) != 0)) {
        // We need to release the memory before setting the alias pointer.
        if(reorderCodesCapacity != 0) {
            uprv_free(const_cast<int32_t *>(reorderCodes));
            reorderCodesCapacity = 0;
        }
        reorderTable = table;
        reorderCodes = codes;
        reorderCodesLength = length;
        // Drop ranges before the first split byte. They are reordered by the table.
        // This then speeds up reordering of the remaining ranges.
        int32_t firstSplitByteRangeIndex = 0;
        while(firstSplitByteRangeIndex < rangesLength &&
                (ranges[firstSplitByteRangeIndex] & 0xff0000) == 0) {
            // The second byte of the primary limit is 0.
            ++firstSplitByteRangeIndex;
        }
        if(firstSplitByteRangeIndex == rangesLength) {
            U_ASSERT(!reorderTableHasSplitBytes(table));
            minHighNoReorder = 0;
            reorderRanges = NULL;
            reorderRangesLength = 0;
        } else {
            U_ASSERT(table[ranges[firstSplitByteRangeIndex] >> 24] == 0);
            minHighNoReorder = ranges[rangesLength - 1] & 0xffff0000;
            reorderRanges = ranges + firstSplitByteRangeIndex;
            reorderRangesLength = rangesLength - firstSplitByteRangeIndex;
        }
        return;
    }
    // Regenerate missing data.
    setReordering(data, codes, length, errorCode);
}

void
CollationSettings::setReordering(const CollationData &data,
                                 const int32_t *codes, int32_t codesLength,
                                 UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(codesLength == 0 || (codesLength == 1 && codes[0] == UCOL_REORDER_CODE_NONE)) {
        resetReordering();
        return;
    }
    UVector32 rangesList(errorCode);
    data.makeReorderRanges(codes, codesLength, rangesList, errorCode);
    if(U_FAILURE(errorCode)) { return; }
    int32_t rangesLength = rangesList.size();
    if(rangesLength == 0) {
        resetReordering();
        return;
    }
    const uint32_t *ranges = reinterpret_cast<uint32_t *>(rangesList.getBuffer());
    // ranges[] contains at least two (limit, offset) pairs.
    // The first offset must be 0. The last offset must not be 0.
    // Separators (at the low end) and trailing weights (at the high end)
    // are never reordered.
    U_ASSERT(rangesLength >= 2);
    U_ASSERT((ranges[0] & 0xffff) == 0 && (ranges[rangesLength - 1] & 0xffff) != 0);
    minHighNoReorder = ranges[rangesLength - 1] & 0xffff0000;

    // Write the lead byte permutation table.
    // Set a 0 for each lead byte that has a range boundary in the middle.
    uint8_t table[256];
    int32_t b = 0;
    int32_t firstSplitByteRangeIndex = -1;
    for(int32_t i = 0; i < rangesLength; ++i) {
        uint32_t pair = ranges[i];
        int32_t limit1 = (int32_t)(pair >> 24);
        while(b < limit1) {
            table[b] = (uint8_t)(b + pair);
            ++b;
        }
        // Check the second byte of the limit.
        if((pair & 0xff0000) != 0) {
            table[limit1] = 0;
            b = limit1 + 1;
            if(firstSplitByteRangeIndex < 0) {
                firstSplitByteRangeIndex = i;
            }
        }
    }
    while(b <= 0xff) {
        table[b] = (uint8_t)b;
        ++b;
    }
    if(firstSplitByteRangeIndex < 0) {
        // The lead byte permutation table alone suffices for reordering.
        rangesLength = 0;
    } else {
        // Remove the ranges below the first split byte.
        ranges += firstSplitByteRangeIndex;
        rangesLength -= firstSplitByteRangeIndex;
    }
    setReorderArrays(codes, codesLength, ranges, rangesLength, table, errorCode);
}

void
CollationSettings::setReorderArrays(const int32_t *codes, int32_t codesLength,
                                    const uint32_t *ranges, int32_t rangesLength,
                                    const uint8_t *table, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    int32_t *ownedCodes;
    int32_t totalLength = codesLength + rangesLength;
    U_ASSERT(totalLength > 0);
    if(totalLength <= reorderCodesCapacity) {
        ownedCodes = const_cast<int32_t *>(reorderCodes);
    } else {
        // Allocate one memory block for the codes, the ranges, and the 16-aligned table.
        int32_t capacity = (totalLength + 3) & ~3;  // round up to a multiple of 4 ints
        ownedCodes = (int32_t *)uprv_malloc(capacity * 4 + 256);
        if(ownedCodes == NULL) {
            resetReordering();
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        if(reorderCodesCapacity != 0) {
            uprv_free(const_cast<int32_t *>(reorderCodes));
        }
        reorderCodes = ownedCodes;
        reorderCodesCapacity = capacity;
    }
    uprv_memcpy(ownedCodes + reorderCodesCapacity, table, 256);
    uprv_memcpy(ownedCodes, codes, codesLength * 4);
    uprv_memcpy(ownedCodes + codesLength, ranges, rangesLength * 4);
    reorderTable = reinterpret_cast<const uint8_t *>(reorderCodes + reorderCodesCapacity);
    reorderCodesLength = codesLength;
    reorderRanges = reinterpret_cast<uint32_t *>(ownedCodes) + codesLength;
    reorderRangesLength = rangesLength;
}

void
CollationSettings::copyReorderingFrom(const CollationSettings &other, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(!other.hasReordering()) {
        resetReordering();
        return;
    }
    minHighNoReorder = other.minHighNoReorder;
    if(other.reorderCodesCapacity == 0) {
        // The reorder arrays are aliased to memory-mapped data.
        reorderTable = other.reorderTable;
        reorderRanges = other.reorderRanges;
        reorderRangesLength = other.reorderRangesLength;
        reorderCodes = other.reorderCodes;
        reorderCodesLength = other.reorderCodesLength;
    } else {
        setReorderArrays(other.reorderCodes, other.reorderCodesLength,
                         other.reorderRanges, other.reorderRangesLength,
                         other.reorderTable, errorCode);
    }
}

UBool
CollationSettings::reorderTableHasSplitBytes(const uint8_t table[256]) {
    U_ASSERT(table[0] == 0);
    for(int32_t i = 1; i < 256; ++i) {
        if(table[i] == 0) {
            return TRUE;
        }
    }
    return FALSE;
}

uint32_t
CollationSettings::reorderEx(uint32_t p) const {
    if(p >= minHighNoReorder) { return p; }
    // Round up p so that its lower 16 bits are >= any offset bits.
    // Then compare q directly with (limit, offset) pairs.
    uint32_t q = p | 0xffff;
    uint32_t r;
    const uint32_t *ranges = reorderRanges;
    while(q >= (r = *ranges)) { ++ranges; }
    return p + (r << 24);
}

void
CollationSettings::setStrength(int32_t value, int32_t defaultOptions, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    int32_t noStrength = options & ~STRENGTH_MASK;
    switch(value) {
    case UCOL_PRIMARY:
    case UCOL_SECONDARY:
    case UCOL_TERTIARY:
    case UCOL_QUATERNARY:
    case UCOL_IDENTICAL:
        options = noStrength | (value << STRENGTH_SHIFT);
        break;
    case UCOL_DEFAULT:
        options = noStrength | (defaultOptions & STRENGTH_MASK);
        break;
    default:
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        break;
    }
}

void
CollationSettings::setFlag(int32_t bit, UColAttributeValue value,
                           int32_t defaultOptions, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    switch(value) {
    case UCOL_ON:
        options |= bit;
        break;
    case UCOL_OFF:
        options &= ~bit;
        break;
    case UCOL_DEFAULT:
        options = (options & ~bit) | (defaultOptions & bit);
        break;
    default:
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        break;
    }
}

void
CollationSettings::setCaseFirst(UColAttributeValue value,
                                int32_t defaultOptions, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    int32_t noCaseFirst = options & ~CASE_FIRST_AND_UPPER_MASK;
    switch(value) {
    case UCOL_OFF:
        options = noCaseFirst;
        break;
    case UCOL_LOWER_FIRST:
        options = noCaseFirst | CASE_FIRST;
        break;
    case UCOL_UPPER_FIRST:
        options = noCaseFirst | CASE_FIRST_AND_UPPER_MASK;
        break;
    case UCOL_DEFAULT:
        options = noCaseFirst | (defaultOptions & CASE_FIRST_AND_UPPER_MASK);
        break;
    default:
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        break;
    }
}

void
CollationSettings::setAlternateHandling(UColAttributeValue value,
                                        int32_t defaultOptions, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    int32_t noAlternate = options & ~ALTERNATE_MASK;
    switch(value) {
    case UCOL_NON_IGNORABLE:
        options = noAlternate;
        break;
    case UCOL_SHIFTED:
        options = noAlternate | SHIFTED;
        break;
    case UCOL_DEFAULT:
        options = noAlternate | (defaultOptions & ALTERNATE_MASK);
        break;
    default:
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        break;
    }
}

void
CollationSettings::setMaxVariable(int32_t value, int32_t defaultOptions, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    int32_t noMax = options & ~MAX_VARIABLE_MASK;
    switch(value) {
    case MAX_VAR_SPACE:
    case MAX_VAR_PUNCT:
    case MAX_VAR_SYMBOL:
    case MAX_VAR_CURRENCY:
        options = noMax | (value << MAX_VARIABLE_SHIFT);
        break;
    case UCOL_DEFAULT:
        options = noMax | (defaultOptions & MAX_VARIABLE_MASK);
        break;
    default:
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        break;
    }
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* collationtailoring.cpp
*
* created on: 2013mar12
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION
















U_NAMESPACE_BEGIN

CollationTailoring::CollationTailoring(const CollationSettings *baseSettings)
        : data(NULL), settings(baseSettings),
          actualLocale(""),
          ownedData(NULL),
          builder(NULL), memory(NULL), bundle(NULL),
          trie(NULL), unsafeBackwardSet(NULL),
          maxExpansions(NULL) {
    if(baseSettings != NULL) {
        U_ASSERT(baseSettings->reorderCodesLength == 0);
        U_ASSERT(baseSettings->reorderTable == NULL);
        U_ASSERT(baseSettings->minHighNoReorder == 0);
    } else {
        settings = new CollationSettings();
    }
    if(settings != NULL) {
        settings->addRef();
    }
    rules.getTerminatedBuffer();  // ensure NUL-termination
    version[0] = version[1] = version[2] = version[3] = 0;
    maxExpansionsInitOnce.reset();
}

CollationTailoring::~CollationTailoring() {
    SharedObject::clearPtr(settings);
    delete ownedData;
    delete builder;
    udata_close(memory);
    ures_close(bundle);
    utrie2_close(trie);
    delete unsafeBackwardSet;
    uhash_close(maxExpansions);
    maxExpansionsInitOnce.reset();
}

UBool
CollationTailoring::ensureOwnedData(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    if(ownedData == NULL) {
        const Normalizer2Impl *nfcImpl = Normalizer2Factory::getNFCImpl(errorCode);
        if(U_FAILURE(errorCode)) { return FALSE; }
        ownedData = new CollationData(*nfcImpl);
        if(ownedData == NULL) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return FALSE;
        }
    }
    data = ownedData;
    return TRUE;
}

void
CollationTailoring::makeBaseVersion(const UVersionInfo ucaVersion, UVersionInfo version) {
    version[0] = UCOL_BUILDER_VERSION;
    version[1] = (ucaVersion[0] << 3) + ucaVersion[1];
    version[2] = ucaVersion[2] << 6;
    version[3] = 0;
}

void
CollationTailoring::setVersion(const UVersionInfo baseVersion, const UVersionInfo rulesVersion) {
    version[0] = UCOL_BUILDER_VERSION;
    version[1] = baseVersion[1];
    version[2] = (baseVersion[2] & 0xc0) + ((rulesVersion[0] + (rulesVersion[0] >> 6)) & 0x3f);
    version[3] = (rulesVersion[1] << 3) + (rulesVersion[1] >> 5) + rulesVersion[2] +
            (rulesVersion[3] << 4) + (rulesVersion[3] >> 4);
}

int32_t
CollationTailoring::getUCAVersion() const {
    return ((int32_t)version[1] << 4) | (version[2] >> 6);
}

CollationCacheEntry::~CollationCacheEntry() {
    SharedObject::clearPtr(tailoring);
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*  
*******************************************************************************
*
*   Copyright (C) 1999-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  collationweights.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001mar08 as ucol_wgt.cpp
*   created by: Markus W. Scherer
*
*   This file contains code for allocating n collation element weights
*   between two exclusive limits.
*   It is used only internally by the collation tailoring builder.
*/



#if !UCONFIG_NO_COLLATION







#ifdef UCOL_DEBUG
#   include <stdio.h>
#endif

U_NAMESPACE_BEGIN

/* collation element weight allocation -------------------------------------- */

/* helper functions for CE weights */

static inline uint32_t
getWeightTrail(uint32_t weight, int32_t length) {
    return (uint32_t)(weight>>(8*(4-length)))&0xff;
}

static inline uint32_t
setWeightTrail(uint32_t weight, int32_t length, uint32_t trail) {
    length=8*(4-length);
    return (uint32_t)((weight&(0xffffff00<<length))|(trail<<length));
}

static inline uint32_t
getWeightByte(uint32_t weight, int32_t idx) {
    return getWeightTrail(weight, idx); /* same calculation */
}

static inline uint32_t
setWeightByte(uint32_t weight, int32_t idx, uint32_t byte) {
    uint32_t mask; /* 0xffffffff except a 00 "hole" for the index-th byte */

    idx*=8;
    if(idx<32) {
        mask=((uint32_t)0xffffffff)>>idx;
    } else {
        // Do not use uint32_t>>32 because on some platforms that does not shift at all
        // while we need it to become 0.
        // PowerPC: 0xffffffff>>32 = 0           (wanted)
        // x86:     0xffffffff>>32 = 0xffffffff  (not wanted)
        //
        // ANSI C99 6.5.7 Bitwise shift operators:
        // "If the value of the right operand is negative
        // or is greater than or equal to the width of the promoted left operand,
        // the behavior is undefined."
        mask=0;
    }
    idx=32-idx;
    mask|=0xffffff00<<idx;
    return (uint32_t)((weight&mask)|(byte<<idx));
}

static inline uint32_t
truncateWeight(uint32_t weight, int32_t length) {
    return (uint32_t)(weight&(0xffffffff<<(8*(4-length))));
}

static inline uint32_t
incWeightTrail(uint32_t weight, int32_t length) {
    return (uint32_t)(weight+(1UL<<(8*(4-length))));
}

static inline uint32_t
decWeightTrail(uint32_t weight, int32_t length) {
    return (uint32_t)(weight-(1UL<<(8*(4-length))));
}

CollationWeights::CollationWeights()
        : middleLength(0), rangeIndex(0), rangeCount(0) {
    for(int32_t i = 0; i < 5; ++i) {
        minBytes[i] = maxBytes[i] = 0;
    }
}

void
CollationWeights::initForPrimary(UBool compressible) {
    middleLength=1;
    minBytes[1] = Collation::MERGE_SEPARATOR_BYTE + 1;
    maxBytes[1] = Collation::TRAIL_WEIGHT_BYTE;
    if(compressible) {
        minBytes[2] = Collation::PRIMARY_COMPRESSION_LOW_BYTE + 1;
        maxBytes[2] = Collation::PRIMARY_COMPRESSION_HIGH_BYTE - 1;
    } else {
        minBytes[2] = 2;
        maxBytes[2] = 0xff;
    }
    minBytes[3] = 2;
    maxBytes[3] = 0xff;
    minBytes[4] = 2;
    maxBytes[4] = 0xff;
}

void
CollationWeights::initForSecondary() {
    // We use only the lower 16 bits for secondary weights.
    middleLength=3;
    minBytes[1] = 0;
    maxBytes[1] = 0;
    minBytes[2] = 0;
    maxBytes[2] = 0;
    minBytes[3] = Collation::LEVEL_SEPARATOR_BYTE + 1;
    maxBytes[3] = 0xff;
    minBytes[4] = 2;
    maxBytes[4] = 0xff;
}

void
CollationWeights::initForTertiary() {
    // We use only the lower 16 bits for tertiary weights.
    middleLength=3;
    minBytes[1] = 0;
    maxBytes[1] = 0;
    minBytes[2] = 0;
    maxBytes[2] = 0;
    // We use only 6 bits per byte.
    // The other bits are used for case & quaternary weights.
    minBytes[3] = Collation::LEVEL_SEPARATOR_BYTE + 1;
    maxBytes[3] = 0x3f;
    minBytes[4] = 2;
    maxBytes[4] = 0x3f;
}

uint32_t
CollationWeights::incWeight(uint32_t weight, int32_t length) const {
    for(;;) {
        uint32_t byte=getWeightByte(weight, length);
        if(byte<maxBytes[length]) {
            return setWeightByte(weight, length, byte+1);
        } else {
            // Roll over, set this byte to the minimum and increment the previous one.
            weight=setWeightByte(weight, length, minBytes[length]);
            --length;
            U_ASSERT(length > 0);
        }
    }
}

uint32_t
CollationWeights::incWeightByOffset(uint32_t weight, int32_t length, int32_t offset) const {
    for(;;) {
        offset += getWeightByte(weight, length);
        if((uint32_t)offset <= maxBytes[length]) {
            return setWeightByte(weight, length, offset);
        } else {
            // Split the offset between this byte and the previous one.
            offset -= minBytes[length];
            weight = setWeightByte(weight, length, minBytes[length] + offset % countBytes(length));
            offset /= countBytes(length);
            --length;
            U_ASSERT(length > 0);
        }
    }
}

void
CollationWeights::lengthenRange(WeightRange &range) const {
    int32_t length=range.length+1;
    range.start=setWeightTrail(range.start, length, minBytes[length]);
    range.end=setWeightTrail(range.end, length, maxBytes[length]);
    range.count*=countBytes(length);
    range.length=length;
}

/* for uprv_sortArray: sort ranges in weight order */
static int32_t U_CALLCONV
compareRanges(const void * /*context*/, const void *left, const void *right) {
    uint32_t l, r;

    l=((const CollationWeights::WeightRange *)left)->start;
    r=((const CollationWeights::WeightRange *)right)->start;
    if(l<r) {
        return -1;
    } else if(l>r) {
        return 1;
    } else {
        return 0;
    }
}

UBool
CollationWeights::getWeightRanges(uint32_t lowerLimit, uint32_t upperLimit) {
    U_ASSERT(lowerLimit != 0);
    U_ASSERT(upperLimit != 0);

    /* get the lengths of the limits */
    int32_t lowerLength=lengthOfWeight(lowerLimit);
    int32_t upperLength=lengthOfWeight(upperLimit);

#ifdef UCOL_DEBUG
    printf("length of lower limit 0x%08lx is %ld\n", lowerLimit, lowerLength);
    printf("length of upper limit 0x%08lx is %ld\n", upperLimit, upperLength);
#endif
    U_ASSERT(lowerLength>=middleLength);
    // Permit upperLength<middleLength: The upper limit for secondaries is 0x10000.

    if(lowerLimit>=upperLimit) {
#ifdef UCOL_DEBUG
        printf("error: no space between lower & upper limits\n");
#endif
        return FALSE;
    }

    /* check that neither is a prefix of the other */
    if(lowerLength<upperLength) {
        if(lowerLimit==truncateWeight(upperLimit, lowerLength)) {
#ifdef UCOL_DEBUG
            printf("error: lower limit 0x%08lx is a prefix of upper limit 0x%08lx\n", lowerLimit, upperLimit);
#endif
            return FALSE;
        }
    }
    /* if the upper limit is a prefix of the lower limit then the earlier test lowerLimit>=upperLimit has caught it */

    WeightRange lower[5], middle, upper[5]; /* [0] and [1] are not used - this simplifies indexing */
    uprv_memset(lower, 0, sizeof(lower));
    uprv_memset(&middle, 0, sizeof(middle));
    uprv_memset(upper, 0, sizeof(upper));

    /*
     * With the limit lengths of 1..4, there are up to 7 ranges for allocation:
     * range     minimum length
     * lower[4]  4
     * lower[3]  3
     * lower[2]  2
     * middle    1
     * upper[2]  2
     * upper[3]  3
     * upper[4]  4
     *
     * We are now going to calculate up to 7 ranges.
     * Some of them will typically overlap, so we will then have to merge and eliminate ranges.
     */
    uint32_t weight=lowerLimit;
    for(int32_t length=lowerLength; length>middleLength; --length) {
        uint32_t trail=getWeightTrail(weight, length);
        if(trail<maxBytes[length]) {
            lower[length].start=incWeightTrail(weight, length);
            lower[length].end=setWeightTrail(weight, length, maxBytes[length]);
            lower[length].length=length;
            lower[length].count=maxBytes[length]-trail;
        }
        weight=truncateWeight(weight, length-1);
    }
    if(weight<0xff000000) {
        middle.start=incWeightTrail(weight, middleLength);
    } else {
        // Prevent overflow for primary lead byte FF
        // which would yield a middle range starting at 0.
        middle.start=0xffffffff;  // no middle range
    }

    weight=upperLimit;
    for(int32_t length=upperLength; length>middleLength; --length) {
        uint32_t trail=getWeightTrail(weight, length);
        if(trail>minBytes[length]) {
            upper[length].start=setWeightTrail(weight, length, minBytes[length]);
            upper[length].end=decWeightTrail(weight, length);
            upper[length].length=length;
            upper[length].count=trail-minBytes[length];
        }
        weight=truncateWeight(weight, length-1);
    }
    middle.end=decWeightTrail(weight, middleLength);

    /* set the middle range */
    middle.length=middleLength;
    if(middle.end>=middle.start) {
        middle.count=(int32_t)((middle.end-middle.start)>>(8*(4-middleLength)))+1;
    } else {
        /* no middle range, eliminate overlaps */
        for(int32_t length=4; length>middleLength; --length) {
            if(lower[length].count>0 && upper[length].count>0) {
                // Note: The lowerEnd and upperStart weights are versions of
                // lowerLimit and upperLimit (which are lowerLimit<upperLimit),
                // truncated (still less-or-equal)
                // and then with their last bytes changed to the
                // maxByte (for lowerEnd) or minByte (for upperStart).
                const uint32_t lowerEnd=lower[length].end;
                const uint32_t upperStart=upper[length].start;
                UBool merged=FALSE;

                if(lowerEnd>upperStart) {
                    // These two lower and upper ranges collide.
                    // Since lowerLimit<upperLimit and lowerEnd and upperStart
                    // are versions with only their last bytes modified
                    // (and following ones removed/reset to 0),
                    // lowerEnd>upperStart is only possible
                    // if the leading bytes are equal
                    // and lastByte(lowerEnd)>lastByte(upperStart).
                    U_ASSERT(truncateWeight(lowerEnd, length-1)==
                            truncateWeight(upperStart, length-1));
                    // Intersect these two ranges.
                    lower[length].end=upper[length].end;
                    lower[length].count=
                            (int32_t)getWeightTrail(lower[length].end, length)-
                            (int32_t)getWeightTrail(lower[length].start, length)+1;
                    // count might be <=0 in which case there is no room,
                    // and the range-collecting code below will ignore this range.
                    merged=TRUE;
                } else if(lowerEnd==upperStart) {
                    // Not possible, unless minByte==maxByte which is not allowed.
                    U_ASSERT(minBytes[length]<maxBytes[length]);
                } else /* lowerEnd<upperStart */ {
                    if(incWeight(lowerEnd, length)==upperStart) {
                        // Merge adjacent ranges.
                        lower[length].end=upper[length].end;
                        lower[length].count+=upper[length].count;  // might be >countBytes
                        merged=TRUE;
                    }
                }
                if(merged) {
                    // Remove all shorter ranges.
                    // There was no room available for them between the ranges we just merged.
                    upper[length].count=0;
                    while(--length>middleLength) {
                        lower[length].count=upper[length].count=0;
                    }
                    break;
                }
            }
        }
    }

#ifdef UCOL_DEBUG
    /* print ranges */
    for(int32_t length=4; length>=2; --length) {
        if(lower[length].count>0) {
            printf("lower[%ld] .start=0x%08lx .end=0x%08lx .count=%ld\n", length, lower[length].start, lower[length].end, lower[length].count);
        }
    }
    if(middle.count>0) {
        printf("middle   .start=0x%08lx .end=0x%08lx .count=%ld\n", middle.start, middle.end, middle.count);
    }
    for(int32_t length=2; length<=4; ++length) {
        if(upper[length].count>0) {
            printf("upper[%ld] .start=0x%08lx .end=0x%08lx .count=%ld\n", length, upper[length].start, upper[length].end, upper[length].count);
        }
    }
#endif

    /* copy the ranges, shortest first, into the result array */
    rangeCount=0;
    if(middle.count>0) {
        uprv_memcpy(ranges, &middle, sizeof(WeightRange));
        rangeCount=1;
    }
    for(int32_t length=middleLength+1; length<=4; ++length) {
        /* copy upper first so that later the middle range is more likely the first one to use */
        if(upper[length].count>0) {
            uprv_memcpy(ranges+rangeCount, upper+length, sizeof(WeightRange));
            ++rangeCount;
        }
        if(lower[length].count>0) {
            uprv_memcpy(ranges+rangeCount, lower+length, sizeof(WeightRange));
            ++rangeCount;
        }
    }
    return rangeCount>0;
}

UBool
CollationWeights::allocWeightsInShortRanges(int32_t n, int32_t minLength) {
    // See if the first few minLength and minLength+1 ranges have enough weights.
    for(int32_t i = 0; i < rangeCount && ranges[i].length <= (minLength + 1); ++i) {
        if(n <= ranges[i].count) {
            // Use the first few minLength and minLength+1 ranges.
            if(ranges[i].length > minLength) {
                // Reduce the number of weights from the last minLength+1 range
                // which might sort before some minLength ranges,
                // so that we use all weights in the minLength ranges.
                ranges[i].count = n;
            }
            rangeCount = i + 1;
#ifdef UCOL_DEBUG
            printf("take first %ld ranges\n", rangeCount);
#endif

            if(rangeCount>1) {
                /* sort the ranges by weight values */
                UErrorCode errorCode=U_ZERO_ERROR;
                uprv_sortArray(ranges, rangeCount, sizeof(WeightRange),
                               compareRanges, NULL, FALSE, &errorCode);
                /* ignore error code: we know that the internal sort function will not fail here */
            }
            return TRUE;
        }
        n -= ranges[i].count;  // still >0
    }
    return FALSE;
}

UBool
CollationWeights::allocWeightsInMinLengthRanges(int32_t n, int32_t minLength) {
    // See if the minLength ranges have enough weights
    // when we split one and lengthen the following ones.
    int32_t count = 0;
    int32_t minLengthRangeCount;
    for(minLengthRangeCount = 0;
            minLengthRangeCount < rangeCount &&
                ranges[minLengthRangeCount].length == minLength;
            ++minLengthRangeCount) {
        count += ranges[minLengthRangeCount].count;
    }

    int32_t nextCountBytes = countBytes(minLength + 1);
    if(n > count * nextCountBytes) { return FALSE; }

    // Use the minLength ranges. Merge them, and then split again as necessary.
    uint32_t start = ranges[0].start;
    uint32_t end = ranges[0].end;
    for(int32_t i = 1; i < minLengthRangeCount; ++i) {
        if(ranges[i].start < start) { start = ranges[i].start; }
        if(ranges[i].end > end) { end = ranges[i].end; }
    }

    // Calculate how to split the range between minLength (count1) and minLength+1 (count2).
    // Goal:
    //   count1 + count2 * nextCountBytes = n
    //   count1 + count2 = count
    // These turn into
    //   (count - count2) + count2 * nextCountBytes = n
    // and then into the following count1 & count2 computations.
    int32_t count2 = (n - count) / (nextCountBytes - 1);  // number of weights to be lengthened
    int32_t count1 = count - count2;  // number of minLength weights
    if(count2 == 0 || (count1 + count2 * nextCountBytes) < n) {
        // round up
        ++count2;
        --count1;
        U_ASSERT((count1 + count2 * nextCountBytes) >= n);
    }

    ranges[0].start = start;

    if(count1 == 0) {
        // Make one long range.
        ranges[0].end = end;
        ranges[0].count = count;
        lengthenRange(ranges[0]);
        rangeCount = 1;
    } else {
        // Split the range, lengthen the second part.
#ifdef UCOL_DEBUG
        printf("split the range number %ld (out of %ld minLength ranges) by %ld:%ld\n",
               splitRange, rangeCount, count1, count2);
#endif

        // Next start = start + count1. First end = 1 before that.
        ranges[0].end = incWeightByOffset(start, minLength, count1 - 1);
        ranges[0].count = count1;

        ranges[1].start = incWeight(ranges[0].end, minLength);
        ranges[1].end = end;
        ranges[1].length = minLength;  // +1 when lengthened
        ranges[1].count = count2;  // *countBytes when lengthened
        lengthenRange(ranges[1]);
        rangeCount = 2;
    }
    return TRUE;
}

/*
 * call getWeightRanges and then determine heuristically
 * which ranges to use for a given number of weights between (excluding)
 * two limits
 */
UBool
CollationWeights::allocWeights(uint32_t lowerLimit, uint32_t upperLimit, int32_t n) {
#ifdef UCOL_DEBUG
    puts("");
#endif

    if(!getWeightRanges(lowerLimit, upperLimit)) {
#ifdef UCOL_DEBUG
        printf("error: unable to get Weight ranges\n");
#endif
        return FALSE;
    }

    /* try until we find suitably large ranges */
    for(;;) {
        /* get the smallest number of bytes in a range */
        int32_t minLength=ranges[0].length;

        if(allocWeightsInShortRanges(n, minLength)) { break; }

        if(minLength == 4) {
#ifdef UCOL_DEBUG
            printf("error: the maximum number of %ld weights is insufficient for n=%ld\n",
                   minLengthCount, n);
#endif
            return FALSE;
        }

        if(allocWeightsInMinLengthRanges(n, minLength)) { break; }

        /* no good match, lengthen all minLength ranges and iterate */
#ifdef UCOL_DEBUG
        printf("lengthen the short ranges from %ld bytes to %ld and iterate\n", minLength, minLength+1);
#endif
        for(int32_t i=0; i<rangeCount && ranges[i].length==minLength; ++i) {
            lengthenRange(ranges[i]);
        }
    }

#ifdef UCOL_DEBUG
    puts("final ranges:");
    for(int32_t i=0; i<rangeCount; ++i) {
        printf("ranges[%ld] .start=0x%08lx .end=0x%08lx .length=%ld .count=%ld\n",
               i, ranges[i].start, ranges[i].end, ranges[i].length, ranges[i].count);
    }
#endif

    rangeIndex = 0;
    return TRUE;
}

uint32_t
CollationWeights::nextWeight() {
    if(rangeIndex >= rangeCount) {
        return 0xffffffff;
    } else {
        /* get the next weight */
        WeightRange &range = ranges[rangeIndex];
        uint32_t weight = range.start;
        if(--range.count == 0) {
            /* this range is finished */
            ++rangeIndex;
        } else {
            /* increment the weight for the next value */
            range.start = incWeight(weight, range.length);
            U_ASSERT(range.start <= range.end);
        }

        return weight;
    }
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_COLLATION */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 2012-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File COMPACTDECIMALFORMAT.H
********************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Compatibility APIs for compact decimal number formatting.
 */

#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File DECIMFMT.H
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/20/97    clhuang     Updated per C++ implementation.
*   04/03/97    aliu        Rewrote parsing and formatting completely, and
*                           cleaned up and debugged.  Actually works now.
*   04/17/97    aliu        Changed DigitCount to int per code review.
*   07/10/97    helena      Made ParsePosition a class and get rid of the function
*                           hiding problems.
*   09/09/97    aliu        Ported over support for exponential formats.
*   07/20/98    stephen     Changed documentation
*   01/30/13    emmons      Added Scaling methods
********************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Compatibility APIs for decimal formatting.
 */

#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File DCFMTSYM.H
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/18/97    clhuang     Updated per C++ implementation.
*   03/27/97    helena      Updated to pass the simple test after code review.
*   08/26/97    aliu        Added currency/intl currency symbol support.
*   07/22/98    stephen     Changed to match C++ style
*                            currencySymbol -> fCurrencySymbol
*                            Constants changed from CAPS to kCaps
*   06/24/99    helena      Integrated Alan's NF enhancements and Java2 bug fixes
*   09/22/00    grhoten     Marked deprecation tags with a pointer to replacement
*                            functions.
********************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2014, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
*
* File NUMSYS.H
*
* Modification History:*
*   Date        Name        Description
*
********************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: NumberingSystem object
 */

#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

// can't be #ifndef U_HIDE_INTERNAL_API; needed for char[] field size
/**
 * Size of a numbering system name.
 * @internal
 */
constexpr const size_t kInternalNumSysNameCapacity = 8;

/**
 * Defines numbering systems. A numbering system describes the scheme by which
 * numbers are to be presented to the end user.  In its simplest form, a numbering
 * system describes the set of digit characters that are to be used to display
 * numbers, such as Western digits, Thai digits, Arabic-Indic digits, etc., in a
 * positional numbering system with a specified radix (typically 10).
 * More complicated numbering systems are algorithmic in nature, and require use
 * of an RBNF formatter ( rule based number formatter ), in order to calculate
 * the characters to be displayed for a given number.  Examples of algorithmic
 * numbering systems include Roman numerals, Chinese numerals, and Hebrew numerals.
 * Formatting rules for many commonly used numbering systems are included in
 * the ICU package, based on the numbering system rules defined in CLDR.
 * Alternate numbering systems can be specified to a locale by using the
 * numbers locale keyword.
 */

class U_I18N_API NumberingSystem : public UObject {
public:

    /**
     * Default Constructor.
     *
     * @stable ICU 4.2
     */
    NumberingSystem();

    /**
     * Copy constructor.
     * @stable ICU 4.2
     */
    NumberingSystem(const NumberingSystem& other);

    /**
     * Destructor.
     * @stable ICU 4.2
     */
    virtual ~NumberingSystem();

    /**
     * Create the default numbering system associated with the specified locale.
     * @param inLocale The given locale.
     * @param status ICU status
     * @stable ICU 4.2
     */
    static NumberingSystem* U_EXPORT2 createInstance(const Locale & inLocale, UErrorCode& status);

    /**
     * Create the default numbering system associated with the default locale.
     * @stable ICU 4.2
     */
    static NumberingSystem* U_EXPORT2 createInstance(UErrorCode& status);

    /**
     * Create a numbering system using the specified radix, type, and description.
     * @param radix         The radix (base) for this numbering system.
     * @param isAlgorithmic TRUE if the numbering system is algorithmic rather than numeric.
     * @param description   The string representing the set of digits used in a numeric system, or the name of the RBNF
     *                      ruleset to be used in an algorithmic system.
     * @param status ICU status
     * @stable ICU 4.2
     */
    static NumberingSystem* U_EXPORT2 createInstance(int32_t radix, UBool isAlgorithmic, const UnicodeString& description, UErrorCode& status );

    /**
     * Return a StringEnumeration over all the names of numbering systems known to ICU.
     * The numbering system names will be in alphabetical (invariant) order.
     *
     * The returned StringEnumeration is owned by the caller, who must delete it when
     * finished with it.
     *
     * @stable ICU 4.2
     */
     static StringEnumeration * U_EXPORT2 getAvailableNames(UErrorCode& status);

    /**
     * Create a numbering system from one of the predefined numbering systems specified
     * by CLDR and known to ICU, such as "latn", "arabext", or "hanidec"; the full list
     * is returned by unumsys_openAvailableNames. Note that some of the names listed at
     * http://unicode.org/repos/cldr/tags/latest/common/bcp47/number.xml - e.g.
     * default, native, traditional, finance - do not identify specific numbering systems,
     * but rather key values that may only be used as part of a locale, which in turn
     * defines how they are mapped to a specific numbering system such as "latn" or "hant".
     *
     * @param name   The name of the numbering system.
     * @param status ICU status; set to U_UNSUPPORTED_ERROR if numbering system not found.
     * @return The NumberingSystem instance, or nullptr if not found.
     * @stable ICU 4.2
     */
    static NumberingSystem* U_EXPORT2 createInstanceByName(const char* name, UErrorCode& status);


    /**
     * Returns the radix of this numbering system. Simple positional numbering systems
     * typically have radix 10, but might have a radix of e.g. 16 for hexadecimal. The
     * radix is less well-defined for non-positional algorithmic systems.
     * @stable ICU 4.2
     */
    int32_t getRadix() const;

    /**
     * Returns the name of this numbering system if it was created using one of the predefined names
     * known to ICU.  Otherwise, returns NULL.
     * The predefined names are identical to the numbering system names as defined by
     * the BCP47 definition in Unicode CLDR.
     * See also, http://www.unicode.org/repos/cldr/tags/latest/common/bcp47/number.xml
     * @stable ICU 4.6
     */
    const char * getName() const;

    /**
     * Returns the description string of this numbering system. For simple
     * positional systems this is the ordered string of digits (with length matching
     * the radix), e.g. "\u3007\u4E00\u4E8C\u4E09\u56DB\u4E94\u516D\u4E03\u516B\u4E5D"
     * for "hanidec"; it would be "0123456789ABCDEF" for hexadecimal. For
     * algorithmic systems this is the name of the RBNF ruleset used for formatting,
     * e.g. "zh/SpelloutRules/%spellout-cardinal" for "hans" or "%greek-upper" for
     * "grek".
     * @stable ICU 4.2
     */
    virtual UnicodeString getDescription() const;



    /**
     * Returns TRUE if the given numbering system is algorithmic
     *
     * @return         TRUE if the numbering system is algorithmic.
     *                 Otherwise, return FALSE.
     * @stable ICU 4.2
     */
    UBool isAlgorithmic() const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 4.2
     *
    */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 4.2
     */
    virtual UClassID getDynamicClassID() const;


private:
    UnicodeString   desc;
    int32_t         radix;
    UBool           algorithmic;
    char            name[kInternalNumSysNameCapacity+1];

    void setRadix(int32_t radix);

    void setAlgorithmic(UBool algorithmic);

    void setDesc(const UnicodeString &desc);

    void setName(const char* name);

    static UBool isValidDigitString(const UnicodeString &str);

    UBool hasContiguousDecimalDigits() const;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */





/**
 * \file
 * \brief C++ API: Symbols for formatting numbers.
 */


U_NAMESPACE_BEGIN

/**
 * This class represents the set of symbols needed by DecimalFormat
 * to format numbers. DecimalFormat creates for itself an instance of
 * DecimalFormatSymbols from its locale data.  If you need to change any
 * of these symbols, you can get the DecimalFormatSymbols object from
 * your DecimalFormat and modify it.
 * <P>
 * Here are the special characters used in the parts of the
 * subpattern, with notes on their usage.
 * <pre>
 * \code
 *        Symbol   Meaning
 *          0      a digit
 *          #      a digit, zero shows as absent
 *          .      placeholder for decimal separator
 *          ,      placeholder for grouping separator.
 *          ;      separates formats.
 *          -      default negative prefix.
 *          %      divide by 100 and show as percentage
 *          X      any other characters can be used in the prefix or suffix
 *          '      used to quote special characters in a prefix or suffix.
 * \endcode
 *  </pre>
 * [Notes]
 * <P>
 * If there is no explicit negative subpattern, - is prefixed to the
 * positive form. That is, "0.00" alone is equivalent to "0.00;-0.00".
 * <P>
 * The grouping separator is commonly used for thousands, but in some
 * countries for ten-thousands. The interval is a constant number of
 * digits between the grouping characters, such as 100,000,000 or 1,0000,0000.
 * If you supply a pattern with multiple grouping characters, the interval
 * between the last one and the end of the integer is the one that is
 * used. So "#,##,###,####" == "######,####" == "##,####,####".
 */
class U_I18N_API DecimalFormatSymbols : public UObject {
public:
    /**
     * Constants for specifying a number format symbol.
     * @stable ICU 2.0
     */
    enum ENumberFormatSymbol {
        /** The decimal separator */
        kDecimalSeparatorSymbol,
        /** The grouping separator */
        kGroupingSeparatorSymbol,
        /** The pattern separator */
        kPatternSeparatorSymbol,
        /** The percent sign */
        kPercentSymbol,
        /** Zero*/
        kZeroDigitSymbol,
        /** Character representing a digit in the pattern */
        kDigitSymbol,
        /** The minus sign */
        kMinusSignSymbol,
        /** The plus sign */
        kPlusSignSymbol,
        /** The currency symbol */
        kCurrencySymbol,
        /** The international currency symbol */
        kIntlCurrencySymbol,
        /** The monetary separator */
        kMonetarySeparatorSymbol,
        /** The exponential symbol */
        kExponentialSymbol,
        /** Per mill symbol - replaces kPermillSymbol */
        kPerMillSymbol,
        /** Escape padding character */
        kPadEscapeSymbol,
        /** Infinity symbol */
        kInfinitySymbol,
        /** Nan symbol */
        kNaNSymbol,
        /** Significant digit symbol
         * @stable ICU 3.0 */
        kSignificantDigitSymbol,
        /** The monetary grouping separator
         * @stable ICU 3.6
         */
        kMonetaryGroupingSeparatorSymbol,
        /** One
         * @stable ICU 4.6
         */
        kOneDigitSymbol,
        /** Two
         * @stable ICU 4.6
         */
        kTwoDigitSymbol,
        /** Three
         * @stable ICU 4.6
         */
        kThreeDigitSymbol,
        /** Four
         * @stable ICU 4.6
         */
        kFourDigitSymbol,
        /** Five
         * @stable ICU 4.6
         */
        kFiveDigitSymbol,
        /** Six
         * @stable ICU 4.6
         */
        kSixDigitSymbol,
        /** Seven
         * @stable ICU 4.6
         */
        kSevenDigitSymbol,
        /** Eight
         * @stable ICU 4.6
         */
        kEightDigitSymbol,
        /** Nine
         * @stable ICU 4.6
         */
        kNineDigitSymbol,
        /** Multiplication sign.
         * @stable ICU 54
         */
        kExponentMultiplicationSymbol,
        /** count symbol constants */
        kFormatSymbolCount = kNineDigitSymbol + 2
    };

    /**
     * Create a DecimalFormatSymbols object for the given locale.
     *
     * @param locale    The locale to get symbols for.
     * @param status    Input/output parameter, set to success or
     *                  failure code upon return.
     * @stable ICU 2.0
     */
    DecimalFormatSymbols(const Locale& locale, UErrorCode& status);

    /**
     * Creates a DecimalFormatSymbols instance for the given locale with digits and symbols
     * corresponding to the given NumberingSystem.
     *
     * This constructor behaves equivalently to the normal constructor called with a locale having a
     * "numbers=xxxx" keyword specifying the numbering system by name.
     *
     * In this constructor, the NumberingSystem argument will be used even if the locale has its own
     * "numbers=xxxx" keyword.
     *
     * @param locale    The locale to get symbols for.
     * @param ns        The numbering system.
     * @param status    Input/output parameter, set to success or
     *                  failure code upon return.
     * @stable ICU 60
     */
    DecimalFormatSymbols(const Locale& locale, const NumberingSystem& ns, UErrorCode& status);

    /**
     * Create a DecimalFormatSymbols object for the default locale.
     * This constructor will not fail.  If the resource file data is
     * not available, it will use hard-coded last-resort data and
     * set status to U_USING_FALLBACK_ERROR.
     *
     * @param status    Input/output parameter, set to success or
     *                  failure code upon return.
     * @stable ICU 2.0
     */
    DecimalFormatSymbols(UErrorCode& status);

    /**
     * Creates a DecimalFormatSymbols object with last-resort data.
     * Intended for callers who cache the symbols data and
     * set all symbols on the resulting object.
     *
     * The last-resort symbols are similar to those for the root data,
     * except that the grouping separators are empty,
     * the NaN symbol is U+FFFD rather than "NaN",
     * and the CurrencySpacing patterns are empty.
     *
     * @param status    Input/output parameter, set to success or
     *                  failure code upon return.
     * @return last-resort symbols
     * @stable ICU 52
     */
    static DecimalFormatSymbols* createWithLastResortData(UErrorCode& status);

    /**
     * Copy constructor.
     * @stable ICU 2.0
     */
    DecimalFormatSymbols(const DecimalFormatSymbols&);

    /**
     * Assignment operator.
     * @stable ICU 2.0
     */
    DecimalFormatSymbols& operator=(const DecimalFormatSymbols&);

    /**
     * Destructor.
     * @stable ICU 2.0
     */
    virtual ~DecimalFormatSymbols();

    /**
     * Return true if another object is semantically equal to this one.
     *
     * @param other    the object to be compared with.
     * @return         true if another object is semantically equal to this one.
     * @stable ICU 2.0
     */
    UBool operator==(const DecimalFormatSymbols& other) const;

    /**
     * Return true if another object is semantically unequal to this one.
     *
     * @param other    the object to be compared with.
     * @return         true if another object is semantically unequal to this one.
     * @stable ICU 2.0
     */
    UBool operator!=(const DecimalFormatSymbols& other) const { return !operator==(other); }

    /**
     * Get one of the format symbols by its enum constant.
     * Each symbol is stored as a string so that graphemes
     * (characters with modifier letters) can be used.
     *
     * @param symbol    Constant to indicate a number format symbol.
     * @return    the format symbols by the param 'symbol'
     * @stable ICU 2.0
     */
    inline UnicodeString getSymbol(ENumberFormatSymbol symbol) const;

    /**
     * Set one of the format symbols by its enum constant.
     * Each symbol is stored as a string so that graphemes
     * (characters with modifier letters) can be used.
     *
     * @param symbol    Constant to indicate a number format symbol.
     * @param value     value of the format symbol
     * @param propogateDigits If false, setting the zero digit will not automatically set 1-9.
     *     The default behavior is to automatically set 1-9 if zero is being set and the value
     *     it is being set to corresponds to a known Unicode zero digit.
     * @stable ICU 2.0
     */
    void setSymbol(ENumberFormatSymbol symbol, const UnicodeString &value, const UBool propogateDigits);

    /**
     * Returns the locale for which this object was constructed.
     * @stable ICU 2.6
     */
    inline Locale getLocale() const;

    /**
     * Returns the locale for this object. Two flavors are available:
     * valid and actual locale.
     * @stable ICU 2.8
     */
    Locale getLocale(ULocDataLocaleType type, UErrorCode& status) const;

    /**
      * Get pattern string for 'CurrencySpacing' that can be applied to
      * currency format.
      * This API gets the CurrencySpacing data from ResourceBundle. The pattern can
      * be empty if there is no data from current locale and its parent locales.
      *
      * @param type :  UNUM_CURRENCY_MATCH, UNUM_CURRENCY_SURROUNDING_MATCH or UNUM_CURRENCY_INSERT.
      * @param beforeCurrency : true if the pattern is for before currency symbol.
      *                         false if the pattern is for after currency symbol.
      * @param status: Input/output parameter, set to success or
      *                  failure code upon return.
      * @return pattern string for currencyMatch, surroundingMatch or spaceInsert.
      *     Return empty string if there is no data for this locale and its parent
      *     locales.
      * @stable ICU 4.8
      */
     const UnicodeString& getPatternForCurrencySpacing(UCurrencySpacing type,
                                                 UBool beforeCurrency,
                                                 UErrorCode& status) const;
     /**
       * Set pattern string for 'CurrencySpacing' that can be applied to
       * currency format.
       *
       * @param type : UNUM_CURRENCY_MATCH, UNUM_CURRENCY_SURROUNDING_MATCH or UNUM_CURRENCY_INSERT.
       * @param beforeCurrency : true if the pattern is for before currency symbol.
       *                         false if the pattern is for after currency symbol.
       * @param pattern : pattern string to override current setting.
       * @stable ICU 4.8
       */
     void setPatternForCurrencySpacing(UCurrencySpacing type,
                                       UBool beforeCurrency,
                                       const UnicodeString& pattern);

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 2.2
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 2.2
     */
    static UClassID U_EXPORT2 getStaticClassID();

private:
    DecimalFormatSymbols();

    /**
     * Initializes the symbols from the LocaleElements resource bundle.
     * Note: The organization of LocaleElements badly needs to be
     * cleaned up.
     *
     * @param locale               The locale to get symbols for.
     * @param success              Input/output parameter, set to success or
     *                             failure code upon return.
     * @param useLastResortData    determine if use last resort data
     * @param ns                   The NumberingSystem to use; otherwise, fall
     *                             back to the locale.
     */
    void initialize(const Locale& locale, UErrorCode& success,
        UBool useLastResortData = FALSE, const NumberingSystem* ns = nullptr);

    /**
     * Initialize the symbols with default values.
     */
    void initialize();

    void setCurrencyForSymbols();

public:

#ifndef U_HIDE_INTERNAL_API
    /**
     * @internal For ICU use only
     */
    inline UBool isCustomCurrencySymbol() const {
        return fIsCustomCurrencySymbol;
    }

    /**
     * @internal For ICU use only
     */
    inline UBool isCustomIntlCurrencySymbol() const {
        return fIsCustomIntlCurrencySymbol;
    }

    /**
     * @internal For ICU use only
     */
    inline UChar32 getCodePointZero() const {
        return fCodePointZero;
    }
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * _Internal_ function - more efficient version of getSymbol,
     * returning a const reference to one of the symbol strings.
     * The returned reference becomes invalid when the symbol is changed
     * or when the DecimalFormatSymbols are destroyed.
     * Note: moved \#ifndef U_HIDE_INTERNAL_API after this, since this is needed for inline in DecimalFormat
     *
     * This is not currently stable API, but if you think it should be stable,
     * post a comment on the following ticket and the ICU team will take a look:
     * http://bugs.icu-project.org/trac/ticket/13580
     *
     * @param symbol Constant to indicate a number format symbol.
     * @return the format symbol by the param 'symbol'
     * @internal
     */
    inline const UnicodeString& getConstSymbol(ENumberFormatSymbol symbol) const;

#ifndef U_HIDE_INTERNAL_API
    /**
     * Returns the const UnicodeString reference, like getConstSymbol,
     * corresponding to the digit with the given value.  This is equivalent
     * to accessing the symbol from getConstSymbol with the corresponding
     * key, such as kZeroDigitSymbol or kOneDigitSymbol.
     *
     * This is not currently stable API, but if you think it should be stable,
     * post a comment on the following ticket and the ICU team will take a look:
     * http://bugs.icu-project.org/trac/ticket/13580
     *
     * @param digit The digit, an integer between 0 and 9 inclusive.
     *              If outside the range 0 to 9, the zero digit is returned.
     * @return the format symbol for the given digit.
     * @internal This API is currently for ICU use only.
     */
    inline const UnicodeString& getConstDigitSymbol(int32_t digit) const;

    /**
     * Returns that pattern stored in currecy info. Internal API for use by NumberFormat API.
     * @internal
     */
    inline const char16_t* getCurrencyPattern(void) const;
#endif  /* U_HIDE_INTERNAL_API */

private:
    /**
     * Private symbol strings.
     * They are either loaded from a resource bundle or otherwise owned.
     * setSymbol() clones the symbol string.
     * Readonly aliases can only come from a resource bundle, so that we can always
     * use fastCopyFrom() with them.
     *
     * If DecimalFormatSymbols becomes subclassable and the status of fSymbols changes
     * from private to protected,
     * or when fSymbols can be set any other way that allows them to be readonly aliases
     * to non-resource bundle strings,
     * then regular UnicodeString copies must be used instead of fastCopyFrom().
     *
     */
    UnicodeString fSymbols[kFormatSymbolCount];

    /**
     * Non-symbol variable for getConstSymbol(). Always empty.
     */
    UnicodeString fNoSymbol;

    /**
     * Dealing with code points is faster than dealing with strings when formatting. Because of
     * this, we maintain a value containing the zero code point that is used whenever digitStrings
     * represents a sequence of ten code points in order.
     *
     * <p>If the value stored here is positive, it means that the code point stored in this value
     * corresponds to the digitStrings array, and codePointZero can be used instead of the
     * digitStrings array for the purposes of efficient formatting; if -1, then digitStrings does
     * *not* contain a sequence of code points, and it must be used directly.
     *
     * <p>It is assumed that codePointZero always shadows the value in digitStrings. codePointZero
     * should never be set directly; rather, it should be updated only when digitStrings mutates.
     * That is, the flow of information is digitStrings -> codePointZero, not the other way.
     */
    UChar32 fCodePointZero;

    Locale locale;

    char actualLocale[ULOC_FULLNAME_CAPACITY];
    char validLocale[ULOC_FULLNAME_CAPACITY];
    const char16_t* currPattern;

    UnicodeString currencySpcBeforeSym[UNUM_CURRENCY_SPACING_COUNT];
    UnicodeString currencySpcAfterSym[UNUM_CURRENCY_SPACING_COUNT];
    UBool fIsCustomCurrencySymbol;
    UBool fIsCustomIntlCurrencySymbol;
};

// -------------------------------------

inline UnicodeString
DecimalFormatSymbols::getSymbol(ENumberFormatSymbol symbol) const {
    const UnicodeString *strPtr;
    if(symbol < kFormatSymbolCount) {
        strPtr = &fSymbols[symbol];
    } else {
        strPtr = &fNoSymbol;
    }
    return *strPtr;
}

// See comments above for this function. Not hidden with #ifdef U_HIDE_INTERNAL_API
inline const UnicodeString &
DecimalFormatSymbols::getConstSymbol(ENumberFormatSymbol symbol) const {
    const UnicodeString *strPtr;
    if(symbol < kFormatSymbolCount) {
        strPtr = &fSymbols[symbol];
    } else {
        strPtr = &fNoSymbol;
    }
    return *strPtr;
}

#ifndef U_HIDE_INTERNAL_API
inline const UnicodeString& DecimalFormatSymbols::getConstDigitSymbol(int32_t digit) const {
    if (digit < 0 || digit > 9) {
        digit = 0;
    }
    if (digit == 0) {
        return fSymbols[kZeroDigitSymbol];
    }
    ENumberFormatSymbol key = static_cast<ENumberFormatSymbol>(kOneDigitSymbol + digit - 1);
    return fSymbols[key];
}
#endif /* U_HIDE_INTERNAL_API */

// -------------------------------------

inline void
DecimalFormatSymbols::setSymbol(ENumberFormatSymbol symbol, const UnicodeString &value, const UBool propogateDigits = TRUE) {
    if (symbol == kCurrencySymbol) {
        fIsCustomCurrencySymbol = TRUE;
    }
    else if (symbol == kIntlCurrencySymbol) {
        fIsCustomIntlCurrencySymbol = TRUE;
    }
    if(symbol<kFormatSymbolCount) {
        fSymbols[symbol]=value;
    }

    // If the zero digit is being set to a known zero digit according to Unicode,
    // then we automatically set the corresponding 1-9 digits
    // Also record updates to fCodePointZero. Be conservative if in doubt.
    if (symbol == kZeroDigitSymbol) {
        UChar32 sym = value.char32At(0);
        if ( propogateDigits && u_charDigitValue(sym) == 0 && value.countChar32() == 1 ) {
            fCodePointZero = sym;
            for ( int8_t i = 1 ; i<= 9 ; i++ ) {
                sym++;
                fSymbols[(int)kOneDigitSymbol+i-1] = UnicodeString(sym);
            }
        } else {
            fCodePointZero = -1;
        }
    } else if (symbol >= kOneDigitSymbol && symbol <= kNineDigitSymbol) {
        fCodePointZero = -1;
    }
}

// -------------------------------------

inline Locale
DecimalFormatSymbols::getLocale() const {
    return locale;
}

#ifndef U_HIDE_INTERNAL_API
inline const char16_t*
DecimalFormatSymbols::getCurrencyPattern() const {
    return currPattern;
}
#endif /* U_HIDE_INTERNAL_API */

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */







// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 2012,2014 International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*/

/**
 * \file
 * \brief C++: internal template EnumSet<>
 */




#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/* Can't use #ifndef U_HIDE_INTERNAL_API for the entire EnumSet class, needed in .h file declarations */
/**
 * enum bitset for boolean fields. Similar to Java EnumSet<>. 
 * Needs to range check. Used for private instance variables.
 * @internal
 * \cond
 */
template<typename T, uint32_t minValue, uint32_t limitValue>
class EnumSet {
public:
    inline EnumSet() : fBools(0) {}
    inline EnumSet(const EnumSet<T,minValue,limitValue>& other) : fBools(other.fBools) {}
    inline ~EnumSet() {}
#ifndef U_HIDE_INTERNAL_API
    inline void clear() { fBools=0; }
    inline void add(T toAdd) { set(toAdd, 1); }
    inline void remove(T toRemove) { set(toRemove, 0); }
    inline int32_t contains(T toCheck) const { return get(toCheck); }
    inline void set(T toSet, int32_t v) { fBools=(fBools&(~flag(toSet)))|(v?(flag(toSet)):0); }
    inline int32_t get(T toCheck) const { return (fBools & flag(toCheck))?1:0; }
    inline UBool isValidEnum(T toCheck) const {  return (toCheck>=minValue&&toCheck<limitValue); }
    inline UBool isValidValue(int32_t v) const { return (v==0||v==1); }
    inline const EnumSet<T,minValue,limitValue>& operator=(const EnumSet<T,minValue,limitValue>& other) {
        fBools = other.fBools;
        return *this;
    }
  
    inline uint32_t getAll() const {
        return fBools; 
    }
#endif /* U_HIDE_INTERNAL_API */

private:
    inline uint32_t flag(T toCheck) const { return (1<<(toCheck-minValue)); }
private:
    uint32_t fBools;
};

/** \endcond */

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */


U_NAMESPACE_BEGIN

class CurrencyPluralInfo;
class CompactDecimalFormat;

namespace number {
class LocalizedNumberFormatter;
class FormattedNumber;
namespace impl {
class DecimalQuantity;
struct DecimalFormatFields;
}
}

namespace numparse {
namespace impl {
class NumberParserImpl;
}
}

/**
 * **IMPORTANT:** New users are strongly encouraged to see if
 * numberformatter.h fits their use case.  Although not deprecated, this header
 * is provided for backwards compatibility only.
 *
 * DecimalFormat is a concrete subclass of NumberFormat that formats decimal
 * numbers. It has a variety of features designed to make it possible to parse
 * and format numbers in any locale, including support for Western, Arabic, or
 * Indic digits.  It also supports different flavors of numbers, including
 * integers ("123"), fixed-point numbers ("123.4"), scientific notation
 * ("1.23E4"), percentages ("12%"), and currency amounts ("$123", "USD123",
 * "123 US dollars").  All of these flavors can be easily localized.
 *
 * To obtain a NumberFormat for a specific locale (including the default
 * locale) call one of NumberFormat's factory methods such as
 * createInstance(). Do not call the DecimalFormat constructors directly, unless
 * you know what you are doing, since the NumberFormat factory methods may
 * return subclasses other than DecimalFormat.
 *
 * **Example Usage**
 *
 * \code
 *     // Normally we would have a GUI with a menu for this
 *     int32_t locCount;
 *     const Locale* locales = NumberFormat::getAvailableLocales(locCount);
 *
 *     double myNumber = -1234.56;
 *     UErrorCode success = U_ZERO_ERROR;
 *     NumberFormat* form;
 *
 *     // Print out a number with the localized number, currency and percent
 *     // format for each locale.
 *     UnicodeString countryName;
 *     UnicodeString displayName;
 *     UnicodeString str;
 *     UnicodeString pattern;
 *     Formattable fmtable;
 *     for (int32_t j = 0; j < 3; ++j) {
 *         cout << endl << "FORMAT " << j << endl;
 *         for (int32_t i = 0; i < locCount; ++i) {
 *             if (locales[i].getCountry(countryName).size() == 0) {
 *                 // skip language-only
 *                 continue;
 *             }
 *             switch (j) {
 *             case 0:
 *                 form = NumberFormat::createInstance(locales[i], success ); break;
 *             case 1:
 *                 form = NumberFormat::createCurrencyInstance(locales[i], success ); break;
 *             default:
 *                 form = NumberFormat::createPercentInstance(locales[i], success ); break;
 *             }
 *             if (form) {
 *                 str.remove();
 *                 pattern = ((DecimalFormat*)form)->toPattern(pattern);
 *                 cout << locales[i].getDisplayName(displayName) << ": " << pattern;
 *                 cout << "  ->  " << form->format(myNumber,str) << endl;
 *                 form->parse(form->format(myNumber,str), fmtable, success);
 *                 delete form;
 *             }
 *         }
 *     }
 * \endcode
 *
 * **Another example use createInstance(style)**
 *
 * \code
 * // Print out a number using the localized number, currency,
 * // percent, scientific, integer, iso currency, and plural currency
 * // format for each locale</strong>
 * Locale* locale = new Locale("en", "US");
 * double myNumber = 1234.56;
 * UErrorCode success = U_ZERO_ERROR;
 * UnicodeString str;
 * Formattable fmtable;
 * for (int j=NumberFormat::kNumberStyle;
 *      j<=NumberFormat::kPluralCurrencyStyle;
 *      ++j) {
 *     NumberFormat* form = NumberFormat::createInstance(locale, j, success);
 *     str.remove();
 *     cout << "format result " << form->format(myNumber, str) << endl;
 *     format->parse(form->format(myNumber, str), fmtable, success);
 *     delete form;
 * }
 * \endcode
 *
 *
 * <p><strong>Patterns</strong>
 *
 * <p>A DecimalFormat consists of a <em>pattern</em> and a set of
 * <em>symbols</em>.  The pattern may be set directly using
 * applyPattern(), or indirectly using other API methods which
 * manipulate aspects of the pattern, such as the minimum number of integer
 * digits.  The symbols are stored in a DecimalFormatSymbols
 * object.  When using the NumberFormat factory methods, the
 * pattern and symbols are read from ICU's locale data.
 *
 * <p><strong>Special Pattern Characters</strong>
 *
 * <p>Many characters in a pattern are taken literally; they are matched during
 * parsing and output unchanged during formatting.  Special characters, on the
 * other hand, stand for other characters, strings, or classes of characters.
 * For example, the '#' character is replaced by a localized digit.  Often the
 * replacement character is the same as the pattern character; in the U.S. locale,
 * the ',' grouping character is replaced by ','.  However, the replacement is
 * still happening, and if the symbols are modified, the grouping character
 * changes.  Some special characters affect the behavior of the formatter by
 * their presence; for example, if the percent character is seen, then the
 * value is multiplied by 100 before being displayed.
 *
 * <p>To insert a special character in a pattern as a literal, that is, without
 * any special meaning, the character must be quoted.  There are some exceptions to
 * this which are noted below.
 *
 * <p>The characters listed here are used in non-localized patterns.  Localized
 * patterns use the corresponding characters taken from this formatter's
 * DecimalFormatSymbols object instead, and these characters lose
 * their special status.  Two exceptions are the currency sign and quote, which
 * are not localized.
 *
 * <table border=0 cellspacing=3 cellpadding=0>
 *   <tr bgcolor="#ccccff">
 *     <td align=left><strong>Symbol</strong>
 *     <td align=left><strong>Location</strong>
 *     <td align=left><strong>Localized?</strong>
 *     <td align=left><strong>Meaning</strong>
 *   <tr valign=top>
 *     <td><code>0</code>
 *     <td>Number
 *     <td>Yes
 *     <td>Digit
 *   <tr valign=top bgcolor="#eeeeff">
 *     <td><code>1-9</code>
 *     <td>Number
 *     <td>Yes
 *     <td>'1' through '9' indicate rounding.
 *   <tr valign=top>
 *     <td><code>\htmlonly&#x40;\endhtmlonly</code> <!--doxygen doesn't like @-->
 *     <td>Number
 *     <td>No
 *     <td>Significant digit
 *   <tr valign=top bgcolor="#eeeeff">
 *     <td><code>#</code>
 *     <td>Number
 *     <td>Yes
 *     <td>Digit, zero shows as absent
 *   <tr valign=top>
 *     <td><code>.</code>
 *     <td>Number
 *     <td>Yes
 *     <td>Decimal separator or monetary decimal separator
 *   <tr valign=top bgcolor="#eeeeff">
 *     <td><code>-</code>
 *     <td>Number
 *     <td>Yes
 *     <td>Minus sign
 *   <tr valign=top>
 *     <td><code>,</code>
 *     <td>Number
 *     <td>Yes
 *     <td>Grouping separator
 *   <tr valign=top bgcolor="#eeeeff">
 *     <td><code>E</code>
 *     <td>Number
 *     <td>Yes
 *     <td>Separates mantissa and exponent in scientific notation.
 *         <em>Need not be quoted in prefix or suffix.</em>
 *   <tr valign=top>
 *     <td><code>+</code>
 *     <td>Exponent
 *     <td>Yes
 *     <td>Prefix positive exponents with localized plus sign.
 *         <em>Need not be quoted in prefix or suffix.</em>
 *   <tr valign=top bgcolor="#eeeeff">
 *     <td><code>;</code>
 *     <td>Subpattern boundary
 *     <td>Yes
 *     <td>Separates positive and negative subpatterns
 *   <tr valign=top>
 *     <td><code>\%</code>
 *     <td>Prefix or suffix
 *     <td>Yes
 *     <td>Multiply by 100 and show as percentage
 *   <tr valign=top bgcolor="#eeeeff">
 *     <td><code>\\u2030</code>
 *     <td>Prefix or suffix
 *     <td>Yes
 *     <td>Multiply by 1000 and show as per mille
 *   <tr valign=top>
 *     <td><code>\htmlonly&curren;\endhtmlonly</code> (<code>\\u00A4</code>)
 *     <td>Prefix or suffix
 *     <td>No
 *     <td>Currency sign, replaced by currency symbol.  If
 *         doubled, replaced by international currency symbol.
 *         If tripled, replaced by currency plural names, for example,
 *         "US dollar" or "US dollars" for America.
 *         If present in a pattern, the monetary decimal separator
 *         is used instead of the decimal separator.
 *   <tr valign=top bgcolor="#eeeeff">
 *     <td><code>'</code>
 *     <td>Prefix or suffix
 *     <td>No
 *     <td>Used to quote special characters in a prefix or suffix,
 *         for example, <code>"'#'#"</code> formats 123 to
 *         <code>"#123"</code>.  To create a single quote
 *         itself, use two in a row: <code>"# o''clock"</code>.
 *   <tr valign=top>
 *     <td><code>*</code>
 *     <td>Prefix or suffix boundary
 *     <td>Yes
 *     <td>Pad escape, precedes pad character
 * </table>
 *
 * <p>A DecimalFormat pattern contains a positive and negative
 * subpattern, for example, "#,##0.00;(#,##0.00)".  Each subpattern has a
 * prefix, a numeric part, and a suffix.  If there is no explicit negative
 * subpattern, the negative subpattern is the localized minus sign prefixed to the
 * positive subpattern. That is, "0.00" alone is equivalent to "0.00;-0.00".  If there
 * is an explicit negative subpattern, it serves only to specify the negative
 * prefix and suffix; the number of digits, minimal digits, and other
 * characteristics are ignored in the negative subpattern. That means that
 * "#,##0.0#;(#)" has precisely the same result as "#,##0.0#;(#,##0.0#)".
 *
 * <p>The prefixes, suffixes, and various symbols used for infinity, digits,
 * thousands separators, decimal separators, etc. may be set to arbitrary
 * values, and they will appear properly during formatting.  However, care must
 * be taken that the symbols and strings do not conflict, or parsing will be
 * unreliable.  For example, either the positive and negative prefixes or the
 * suffixes must be distinct for parse() to be able
 * to distinguish positive from negative values.  Another example is that the
 * decimal separator and thousands separator should be distinct characters, or
 * parsing will be impossible.
 *
 * <p>The <em>grouping separator</em> is a character that separates clusters of
 * integer digits to make large numbers more legible.  It commonly used for
 * thousands, but in some locales it separates ten-thousands.  The <em>grouping
 * size</em> is the number of digits between the grouping separators, such as 3
 * for "100,000,000" or 4 for "1 0000 0000". There are actually two different
 * grouping sizes: One used for the least significant integer digits, the
 * <em>primary grouping size</em>, and one used for all others, the
 * <em>secondary grouping size</em>.  In most locales these are the same, but
 * sometimes they are different. For example, if the primary grouping interval
 * is 3, and the secondary is 2, then this corresponds to the pattern
 * "#,##,##0", and the number 123456789 is formatted as "12,34,56,789".  If a
 * pattern contains multiple grouping separators, the interval between the last
 * one and the end of the integer defines the primary grouping size, and the
 * interval between the last two defines the secondary grouping size. All others
 * are ignored, so "#,##,###,####" == "###,###,####" == "##,#,###,####".
 *
 * <p>Illegal patterns, such as "#.#.#" or "#.###,###", will cause
 * DecimalFormat to set a failing UErrorCode.
 *
 * <p><strong>Pattern BNF</strong>
 *
 * <pre>
 * pattern    := subpattern (';' subpattern)?
 * subpattern := prefix? number exponent? suffix?
 * number     := (integer ('.' fraction)?) | sigDigits
 * prefix     := '\\u0000'..'\\uFFFD' - specialCharacters
 * suffix     := '\\u0000'..'\\uFFFD' - specialCharacters
 * integer    := '#'* '0'* '0'
 * fraction   := '0'* '#'*
 * sigDigits  := '#'* '@' '@'* '#'*
 * exponent   := 'E' '+'? '0'* '0'
 * padSpec    := '*' padChar
 * padChar    := '\\u0000'..'\\uFFFD' - quote
 * &nbsp;
 * Notation:
 *   X*       0 or more instances of X
 *   X?       0 or 1 instances of X
 *   X|Y      either X or Y
 *   C..D     any character from C up to D, inclusive
 *   S-T      characters in S, except those in T
 * </pre>
 * The first subpattern is for positive numbers. The second (optional)
 * subpattern is for negative numbers.
 *
 * <p>Not indicated in the BNF syntax above:
 *
 * <ul><li>The grouping separator ',' can occur inside the integer and
 * sigDigits elements, between any two pattern characters of that
 * element, as long as the integer or sigDigits element is not
 * followed by the exponent element.
 *
 * <li>Two grouping intervals are recognized: That between the
 *     decimal point and the first grouping symbol, and that
 *     between the first and second grouping symbols. These
 *     intervals are identical in most locales, but in some
 *     locales they differ. For example, the pattern
 *     &quot;#,##,###&quot; formats the number 123456789 as
 *     &quot;12,34,56,789&quot;.</li>
 *
 * <li>The pad specifier <code>padSpec</code> may appear before the prefix,
 * after the prefix, before the suffix, after the suffix, or not at all.
 *
 * <li>In place of '0', the digits '1' through '9' may be used to
 * indicate a rounding increment.
 * </ul>
 *
 * <p><strong>Parsing</strong>
 *
 * <p>DecimalFormat parses all Unicode characters that represent
 * decimal digits, as defined by u_charDigitValue().  In addition,
 * DecimalFormat also recognizes as digits the ten consecutive
 * characters starting with the localized zero digit defined in the
 * DecimalFormatSymbols object.  During formatting, the
 * DecimalFormatSymbols-based digits are output.
 *
 * <p>During parsing, grouping separators are ignored if in lenient mode;
 * otherwise, if present, they must be in appropriate positions.
 *
 * <p>For currency parsing, the formatter is able to parse every currency
 * style formats no matter which style the formatter is constructed with.
 * For example, a formatter instance gotten from
 * NumberFormat.getInstance(ULocale, NumberFormat.CURRENCYSTYLE) can parse
 * formats such as "USD1.00" and "3.00 US dollars".
 *
 * <p>If parse(UnicodeString&,Formattable&,ParsePosition&)
 * fails to parse a string, it leaves the parse position unchanged.
 * The convenience method parse(UnicodeString&,Formattable&,UErrorCode&)
 * indicates parse failure by setting a failing
 * UErrorCode.
 *
 * <p><strong>Formatting</strong>
 *
 * <p>Formatting is guided by several parameters, all of which can be
 * specified either using a pattern or using the API.  The following
 * description applies to formats that do not use <a href="#sci">scientific
 * notation</a> or <a href="#sigdig">significant digits</a>.
 *
 * <ul><li>If the number of actual integer digits exceeds the
 * <em>maximum integer digits</em>, then only the least significant
 * digits are shown.  For example, 1997 is formatted as "97" if the
 * maximum integer digits is set to 2.
 *
 * <li>If the number of actual integer digits is less than the
 * <em>minimum integer digits</em>, then leading zeros are added.  For
 * example, 1997 is formatted as "01997" if the minimum integer digits
 * is set to 5.
 *
 * <li>If the number of actual fraction digits exceeds the <em>maximum
 * fraction digits</em>, then rounding is performed to the
 * maximum fraction digits.  For example, 0.125 is formatted as "0.12"
 * if the maximum fraction digits is 2.  This behavior can be changed
 * by specifying a rounding increment and/or a rounding mode.
 *
 * <li>If the number of actual fraction digits is less than the
 * <em>minimum fraction digits</em>, then trailing zeros are added.
 * For example, 0.125 is formatted as "0.1250" if the minimum fraction
 * digits is set to 4.
 *
 * <li>Trailing fractional zeros are not displayed if they occur
 * <em>j</em> positions after the decimal, where <em>j</em> is less
 * than the maximum fraction digits. For example, 0.10004 is
 * formatted as "0.1" if the maximum fraction digits is four or less.
 * </ul>
 *
 * <p><strong>Special Values</strong>
 *
 * <p><code>NaN</code> is represented as a single character, typically
 * <code>\\uFFFD</code>.  This character is determined by the
 * DecimalFormatSymbols object.  This is the only value for which
 * the prefixes and suffixes are not used.
 *
 * <p>Infinity is represented as a single character, typically
 * <code>\\u221E</code>, with the positive or negative prefixes and suffixes
 * applied.  The infinity character is determined by the
 * DecimalFormatSymbols object.
 *
 * <a name="sci"><strong>Scientific Notation</strong></a>
 *
 * <p>Numbers in scientific notation are expressed as the product of a mantissa
 * and a power of ten, for example, 1234 can be expressed as 1.234 x 10<sup>3</sup>. The
 * mantissa is typically in the half-open interval [1.0, 10.0) or sometimes [0.0, 1.0),
 * but it need not be.  DecimalFormat supports arbitrary mantissas.
 * DecimalFormat can be instructed to use scientific
 * notation through the API or through the pattern.  In a pattern, the exponent
 * character immediately followed by one or more digit characters indicates
 * scientific notation.  Example: "0.###E0" formats the number 1234 as
 * "1.234E3".
 *
 * <ul>
 * <li>The number of digit characters after the exponent character gives the
 * minimum exponent digit count.  There is no maximum.  Negative exponents are
 * formatted using the localized minus sign, <em>not</em> the prefix and suffix
 * from the pattern.  This allows patterns such as "0.###E0 m/s".  To prefix
 * positive exponents with a localized plus sign, specify '+' between the
 * exponent and the digits: "0.###E+0" will produce formats "1E+1", "1E+0",
 * "1E-1", etc.  (In localized patterns, use the localized plus sign rather than
 * '+'.)
 *
 * <li>The minimum number of integer digits is achieved by adjusting the
 * exponent.  Example: 0.00123 formatted with "00.###E0" yields "12.3E-4".  This
 * only happens if there is no maximum number of integer digits.  If there is a
 * maximum, then the minimum number of integer digits is fixed at one.
 *
 * <li>The maximum number of integer digits, if present, specifies the exponent
 * grouping.  The most common use of this is to generate <em>engineering
 * notation</em>, in which the exponent is a multiple of three, e.g.,
 * "##0.###E0".  The number 12345 is formatted using "##0.####E0" as "12.345E3".
 *
 * <li>When using scientific notation, the formatter controls the
 * digit counts using significant digits logic.  The maximum number of
 * significant digits limits the total number of integer and fraction
 * digits that will be shown in the mantissa; it does not affect
 * parsing.  For example, 12345 formatted with "##0.##E0" is "12.3E3".
 * See the section on significant digits for more details.
 *
 * <li>The number of significant digits shown is determined as
 * follows: If areSignificantDigitsUsed() returns false, then the
 * minimum number of significant digits shown is one, and the maximum
 * number of significant digits shown is the sum of the <em>minimum
 * integer</em> and <em>maximum fraction</em> digits, and is
 * unaffected by the maximum integer digits.  If this sum is zero,
 * then all significant digits are shown.  If
 * areSignificantDigitsUsed() returns true, then the significant digit
 * counts are specified by getMinimumSignificantDigits() and
 * getMaximumSignificantDigits().  In this case, the number of
 * integer digits is fixed at one, and there is no exponent grouping.
 *
 * <li>Exponential patterns may not contain grouping separators.
 * </ul>
 *
 * <a name="sigdig"><strong>Significant Digits</strong></a>
 *
 * <code>DecimalFormat</code> has two ways of controlling how many
 * digits are shows: (a) significant digits counts, or (b) integer and
 * fraction digit counts.  Integer and fraction digit counts are
 * described above.  When a formatter is using significant digits
 * counts, the number of integer and fraction digits is not specified
 * directly, and the formatter settings for these counts are ignored.
 * Instead, the formatter uses however many integer and fraction
 * digits are required to display the specified number of significant
 * digits.  Examples:
 *
 * <table border=0 cellspacing=3 cellpadding=0>
 *   <tr bgcolor="#ccccff">
 *     <td align=left>Pattern
 *     <td align=left>Minimum significant digits
 *     <td align=left>Maximum significant digits
 *     <td align=left>Number
 *     <td align=left>Output of format()
 *   <tr valign=top>
 *     <td><code>\@\@\@</code>
 *     <td>3
 *     <td>3
 *     <td>12345
 *     <td><code>12300</code>
 *   <tr valign=top bgcolor="#eeeeff">
 *     <td><code>\@\@\@</code>
 *     <td>3
 *     <td>3
 *     <td>0.12345
 *     <td><code>0.123</code>
 *   <tr valign=top>
 *     <td><code>\@\@##</code>
 *     <td>2
 *     <td>4
 *     <td>3.14159
 *     <td><code>3.142</code>
 *   <tr valign=top bgcolor="#eeeeff">
 *     <td><code>\@\@##</code>
 *     <td>2
 *     <td>4
 *     <td>1.23004
 *     <td><code>1.23</code>
 * </table>
 *
 * <ul>
 * <li>Significant digit counts may be expressed using patterns that
 * specify a minimum and maximum number of significant digits.  These
 * are indicated by the <code>'@'</code> and <code>'#'</code>
 * characters.  The minimum number of significant digits is the number
 * of <code>'@'</code> characters.  The maximum number of significant
 * digits is the number of <code>'@'</code> characters plus the number
 * of <code>'#'</code> characters following on the right.  For
 * example, the pattern <code>"@@@"</code> indicates exactly 3
 * significant digits.  The pattern <code>"@##"</code> indicates from
 * 1 to 3 significant digits.  Trailing zero digits to the right of
 * the decimal separator are suppressed after the minimum number of
 * significant digits have been shown.  For example, the pattern
 * <code>"@##"</code> formats the number 0.1203 as
 * <code>"0.12"</code>.
 *
 * <li>If a pattern uses significant digits, it may not contain a
 * decimal separator, nor the <code>'0'</code> pattern character.
 * Patterns such as <code>"@00"</code> or <code>"@.###"</code> are
 * disallowed.
 *
 * <li>Any number of <code>'#'</code> characters may be prepended to
 * the left of the leftmost <code>'@'</code> character.  These have no
 * effect on the minimum and maximum significant digits counts, but
 * may be used to position grouping separators.  For example,
 * <code>"#,#@#"</code> indicates a minimum of one significant digits,
 * a maximum of two significant digits, and a grouping size of three.
 *
 * <li>In order to enable significant digits formatting, use a pattern
 * containing the <code>'@'</code> pattern character.  Alternatively,
 * call setSignificantDigitsUsed(TRUE).
 *
 * <li>In order to disable significant digits formatting, use a
 * pattern that does not contain the <code>'@'</code> pattern
 * character. Alternatively, call setSignificantDigitsUsed(FALSE).
 *
 * <li>The number of significant digits has no effect on parsing.
 *
 * <li>Significant digits may be used together with exponential notation. Such
 * patterns are equivalent to a normal exponential pattern with a minimum and
 * maximum integer digit count of one, a minimum fraction digit count of
 * <code>getMinimumSignificantDigits() - 1</code>, and a maximum fraction digit
 * count of <code>getMaximumSignificantDigits() - 1</code>. For example, the
 * pattern <code>"@@###E0"</code> is equivalent to <code>"0.0###E0"</code>.
 *
 * <li>If significant digits are in use, then the integer and fraction
 * digit counts, as set via the API, are ignored.  If significant
 * digits are not in use, then the significant digit counts, as set via
 * the API, are ignored.
 *
 * </ul>
 *
 * <p><strong>Padding</strong>
 *
 * <p>DecimalFormat supports padding the result of
 * format() to a specific width.  Padding may be specified either
 * through the API or through the pattern syntax.  In a pattern the pad escape
 * character, followed by a single pad character, causes padding to be parsed
 * and formatted.  The pad escape character is '*' in unlocalized patterns, and
 * can be localized using DecimalFormatSymbols::setSymbol() with a
 * DecimalFormatSymbols::kPadEscapeSymbol
 * selector.  For example, <code>"$*x#,##0.00"</code> formats 123 to
 * <code>"$xx123.00"</code>, and 1234 to <code>"$1,234.00"</code>.
 *
 * <ul>
 * <li>When padding is in effect, the width of the positive subpattern,
 * including prefix and suffix, determines the format width.  For example, in
 * the pattern <code>"* #0 o''clock"</code>, the format width is 10.
 *
 * <li>The width is counted in 16-bit code units (char16_ts).
 *
 * <li>Some parameters which usually do not matter have meaning when padding is
 * used, because the pattern width is significant with padding.  In the pattern
 * "* ##,##,#,##0.##", the format width is 14.  The initial characters "##,##,"
 * do not affect the grouping size or maximum integer digits, but they do affect
 * the format width.
 *
 * <li>Padding may be inserted at one of four locations: before the prefix,
 * after the prefix, before the suffix, or after the suffix.  If padding is
 * specified in any other location, applyPattern()
 * sets a failing UErrorCode.  If there is no prefix,
 * before the prefix and after the prefix are equivalent, likewise for the
 * suffix.
 *
 * <li>When specified in a pattern, the 32-bit code point immediately
 * following the pad escape is the pad character. This may be any character,
 * including a special pattern character. That is, the pad escape
 * <em>escapes</em> the following character. If there is no character after
 * the pad escape, then the pattern is illegal.
 *
 * </ul>
 *
 * <p><strong>Rounding</strong>
 *
 * <p>DecimalFormat supports rounding to a specific increment.  For
 * example, 1230 rounded to the nearest 50 is 1250.  1.234 rounded to the
 * nearest 0.65 is 1.3.  The rounding increment may be specified through the API
 * or in a pattern.  To specify a rounding increment in a pattern, include the
 * increment in the pattern itself.  "#,#50" specifies a rounding increment of
 * 50.  "#,##0.05" specifies a rounding increment of 0.05.
 *
 * <p>In the absence of an explicit rounding increment numbers are
 * rounded to their formatted width.
 *
 * <ul>
 * <li>Rounding only affects the string produced by formatting.  It does
 * not affect parsing or change any numerical values.
 *
 * <li>A <em>rounding mode</em> determines how values are rounded; see
 * DecimalFormat::ERoundingMode.  The default rounding mode is
 * DecimalFormat::kRoundHalfEven.  The rounding mode can only be set
 * through the API; it can not be set with a pattern.
 *
 * <li>Some locales use rounding in their currency formats to reflect the
 * smallest currency denomination.
 *
 * <li>In a pattern, digits '1' through '9' specify rounding, but otherwise
 * behave identically to digit '0'.
 * </ul>
 *
 * <p><strong>Synchronization</strong>
 *
 * <p>DecimalFormat objects are not synchronized.  Multiple
 * threads should not access one formatter concurrently.
 *
 * <p><strong>Subclassing</strong>
 *
 * <p><em>User subclasses are not supported.</em> While clients may write
 * subclasses, such code will not necessarily work and will not be
 * guaranteed to work stably from release to release.
 */
class U_I18N_API DecimalFormat : public NumberFormat {
  public:
    /**
     * Pad position.
     * @stable ICU 2.4
     */
    enum EPadPosition {
        kPadBeforePrefix, kPadAfterPrefix, kPadBeforeSuffix, kPadAfterSuffix
    };

    /**
     * Create a DecimalFormat using the default pattern and symbols
     * for the default locale. This is a convenient way to obtain a
     * DecimalFormat when internationalization is not the main concern.
     * <P>
     * To obtain standard formats for a given locale, use the factory methods
     * on NumberFormat such as createInstance. These factories will
     * return the most appropriate sub-class of NumberFormat for a given
     * locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * #icu::number::NumberFormatter instead of DecimalFormat.
     * @param status    Output param set to success/failure code. If the
     *                  pattern is invalid this will be set to a failure code.
     * @stable ICU 2.0
     */
    DecimalFormat(UErrorCode& status);

    /**
     * Create a DecimalFormat from the given pattern and the symbols
     * for the default locale. This is a convenient way to obtain a
     * DecimalFormat when internationalization is not the main concern.
     * <P>
     * To obtain standard formats for a given locale, use the factory methods
     * on NumberFormat such as createInstance. These factories will
     * return the most appropriate sub-class of NumberFormat for a given
     * locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * #icu::number::NumberFormatter instead of DecimalFormat.
     * @param pattern   A non-localized pattern string.
     * @param status    Output param set to success/failure code. If the
     *                  pattern is invalid this will be set to a failure code.
     * @stable ICU 2.0
     */
    DecimalFormat(const UnicodeString& pattern, UErrorCode& status);

    /**
     * Create a DecimalFormat from the given pattern and symbols.
     * Use this constructor when you need to completely customize the
     * behavior of the format.
     * <P>
     * To obtain standard formats for a given
     * locale, use the factory methods on NumberFormat such as
     * createInstance or createCurrencyInstance. If you need only minor adjustments
     * to a standard format, you can modify the format returned by
     * a NumberFormat factory method.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * #icu::number::NumberFormatter instead of DecimalFormat.
     *
     * @param pattern           a non-localized pattern string
     * @param symbolsToAdopt    the set of symbols to be used.  The caller should not
     *                          delete this object after making this call.
     * @param status            Output param set to success/failure code. If the
     *                          pattern is invalid this will be set to a failure code.
     * @stable ICU 2.0
     */
    DecimalFormat(const UnicodeString& pattern, DecimalFormatSymbols* symbolsToAdopt, UErrorCode& status);

#ifndef U_HIDE_INTERNAL_API

    /**
     * This API is for ICU use only.
     * Create a DecimalFormat from the given pattern, symbols, and style.
     *
     * @param pattern           a non-localized pattern string
     * @param symbolsToAdopt    the set of symbols to be used.  The caller should not
     *                          delete this object after making this call.
     * @param style             style of decimal format
     * @param status            Output param set to success/failure code. If the
     *                          pattern is invalid this will be set to a failure code.
     * @internal
     */
    DecimalFormat(const UnicodeString& pattern, DecimalFormatSymbols* symbolsToAdopt,
                  UNumberFormatStyle style, UErrorCode& status);

#if UCONFIG_HAVE_PARSEALLINPUT

    /**
     * @internal
     */
    void setParseAllInput(UNumberFormatAttributeValue value);

#endif

#endif  /* U_HIDE_INTERNAL_API */

  private:

    /**
     * Internal constructor for DecimalFormat; sets up internal fields. All public constructors should
     * call this constructor.
     */
    DecimalFormat(const DecimalFormatSymbols* symbolsToAdopt, UErrorCode& status);

  public:

    /**
     * Set an integer attribute on this DecimalFormat.
     * May return U_UNSUPPORTED_ERROR if this instance does not support
     * the specified attribute.
     * @param attr the attribute to set
     * @param newValue new value
     * @param status the error type
     * @return *this - for chaining (example: format.setAttribute(...).setAttribute(...) )
     * @stable ICU 51
     */
    virtual DecimalFormat& setAttribute(UNumberFormatAttribute attr, int32_t newValue, UErrorCode& status);

    /**
     * Get an integer
     * May return U_UNSUPPORTED_ERROR if this instance does not support
     * the specified attribute.
     * @param attr the attribute to set
     * @param status the error type
     * @return the attribute value. Undefined if there is an error.
     * @stable ICU 51
     */
    virtual int32_t getAttribute(UNumberFormatAttribute attr, UErrorCode& status) const;


    /**
     * Set whether or not grouping will be used in this format.
     * @param newValue    True, grouping will be used in this format.
     * @see getGroupingUsed
     * @stable ICU 53
     */
    void setGroupingUsed(UBool newValue) U_OVERRIDE;

    /**
     * Sets whether or not numbers should be parsed as integers only.
     * @param value    set True, this format will parse numbers as integers
     *                 only.
     * @see isParseIntegerOnly
     * @stable ICU 53
     */
    void setParseIntegerOnly(UBool value) U_OVERRIDE;

    /**
     * Sets whether lenient parsing should be enabled (it is off by default).
     *
     * @param enable \c TRUE if lenient parsing should be used,
     *               \c FALSE otherwise.
     * @stable ICU 4.8
     */
    void setLenient(UBool enable) U_OVERRIDE;

    /**
     * Create a DecimalFormat from the given pattern and symbols.
     * Use this constructor when you need to completely customize the
     * behavior of the format.
     * <P>
     * To obtain standard formats for a given
     * locale, use the factory methods on NumberFormat such as
     * createInstance or createCurrencyInstance. If you need only minor adjustments
     * to a standard format, you can modify the format returned by
     * a NumberFormat factory method.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * #icu::number::NumberFormatter instead of DecimalFormat.
     *
     * @param pattern           a non-localized pattern string
     * @param symbolsToAdopt    the set of symbols to be used.  The caller should not
     *                          delete this object after making this call.
     * @param parseError        Output param to receive errors occurred during parsing
     * @param status            Output param set to success/failure code. If the
     *                          pattern is invalid this will be set to a failure code.
     * @stable ICU 2.0
     */
    DecimalFormat(const UnicodeString& pattern, DecimalFormatSymbols* symbolsToAdopt,
                  UParseError& parseError, UErrorCode& status);

    /**
     * Create a DecimalFormat from the given pattern and symbols.
     * Use this constructor when you need to completely customize the
     * behavior of the format.
     * <P>
     * To obtain standard formats for a given
     * locale, use the factory methods on NumberFormat such as
     * createInstance or createCurrencyInstance. If you need only minor adjustments
     * to a standard format, you can modify the format returned by
     * a NumberFormat factory method.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * #icu::number::NumberFormatter instead of DecimalFormat.
     *
     * @param pattern           a non-localized pattern string
     * @param symbols   the set of symbols to be used
     * @param status            Output param set to success/failure code. If the
     *                          pattern is invalid this will be set to a failure code.
     * @stable ICU 2.0
     */
    DecimalFormat(const UnicodeString& pattern, const DecimalFormatSymbols& symbols, UErrorCode& status);

    /**
     * Copy constructor.
     *
     * @param source    the DecimalFormat object to be copied from.
     * @stable ICU 2.0
     */
    DecimalFormat(const DecimalFormat& source);

    /**
     * Assignment operator.
     *
     * @param rhs    the DecimalFormat object to be copied.
     * @stable ICU 2.0
     */
    DecimalFormat& operator=(const DecimalFormat& rhs);

    /**
     * Destructor.
     * @stable ICU 2.0
     */
    ~DecimalFormat() U_OVERRIDE;

    /**
     * Clone this Format object polymorphically. The caller owns the
     * result and should delete it when done.
     *
     * @return    a polymorphic copy of this DecimalFormat.
     * @stable ICU 2.0
     */
    DecimalFormat* clone() const U_OVERRIDE;

    /**
     * Return true if the given Format objects are semantically equal.
     * Objects of different subclasses are considered unequal.
     *
     * @param other    the object to be compared with.
     * @return         true if the given Format objects are semantically equal.
     * @stable ICU 2.0
     */
    UBool operator==(const Format& other) const U_OVERRIDE;


    using NumberFormat::format;

    /**
     * Format a double or long number using base-10 representation.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    UnicodeString& format(double number, UnicodeString& appendTo, FieldPosition& pos) const U_OVERRIDE;

#ifndef U_HIDE_INTERNAL_API
    /**
     * Format a double or long number using base-10 representation.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status
     * @return          Reference to 'appendTo' parameter.
     * @internal
     */
    UnicodeString& format(double number, UnicodeString& appendTo, FieldPosition& pos,
                          UErrorCode& status) const U_OVERRIDE;
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Format a double or long number using base-10 representation.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     *                  Can be NULL.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    UnicodeString& format(double number, UnicodeString& appendTo, FieldPositionIterator* posIter,
                          UErrorCode& status) const U_OVERRIDE;

    /**
     * Format a long number using base-10 representation.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    UnicodeString& format(int32_t number, UnicodeString& appendTo, FieldPosition& pos) const U_OVERRIDE;

#ifndef U_HIDE_INTERNAL_API
    /**
     * Format a long number using base-10 representation.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @internal
     */
    UnicodeString& format(int32_t number, UnicodeString& appendTo, FieldPosition& pos,
                          UErrorCode& status) const U_OVERRIDE;
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Format a long number using base-10 representation.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     *                  Can be NULL.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    UnicodeString& format(int32_t number, UnicodeString& appendTo, FieldPositionIterator* posIter,
                          UErrorCode& status) const U_OVERRIDE;

    /**
     * Format an int64 number using base-10 representation.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.8
     */
    UnicodeString& format(int64_t number, UnicodeString& appendTo, FieldPosition& pos) const U_OVERRIDE;

#ifndef U_HIDE_INTERNAL_API
    /**
     * Format an int64 number using base-10 representation.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @internal
     */
    UnicodeString& format(int64_t number, UnicodeString& appendTo, FieldPosition& pos,
                          UErrorCode& status) const U_OVERRIDE;
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Format an int64 number using base-10 representation.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     *                  Can be NULL.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    UnicodeString& format(int64_t number, UnicodeString& appendTo, FieldPositionIterator* posIter,
                          UErrorCode& status) const U_OVERRIDE;

    /**
     * Format a decimal number.
     * The syntax of the unformatted number is a "numeric string"
     * as defined in the Decimal Arithmetic Specification, available at
     * http://speleotrove.com/decimal
     *
     * @param number    The unformatted number, as a string.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     *                  Can be NULL.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    UnicodeString& format(StringPiece number, UnicodeString& appendTo, FieldPositionIterator* posIter,
                          UErrorCode& status) const U_OVERRIDE;

#ifndef U_HIDE_INTERNAL_API

    /**
     * Format a decimal number.
     * The number is a DecimalQuantity wrapper onto a floating point decimal number.
     * The default implementation in NumberFormat converts the decimal number
     * to a double and formats that.
     *
     * @param number    The number, a DecimalQuantity format Decimal Floating Point.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @internal
     */
    UnicodeString& format(const number::impl::DecimalQuantity& number, UnicodeString& appendTo,
                          FieldPositionIterator* posIter, UErrorCode& status) const U_OVERRIDE;

    /**
     * Format a decimal number.
     * The number is a DecimalQuantity wrapper onto a floating point decimal number.
     * The default implementation in NumberFormat converts the decimal number
     * to a double and formats that.
     *
     * @param number    The number, a DecimalQuantity format Decimal Floating Point.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @internal
     */
    UnicodeString& format(const number::impl::DecimalQuantity& number, UnicodeString& appendTo,
                          FieldPosition& pos, UErrorCode& status) const U_OVERRIDE;

#endif // U_HIDE_INTERNAL_API

    using NumberFormat::parse;

    /**
     * Parse the given string using this object's choices. The method
     * does string comparisons to try to find an optimal match.
     * If no object can be parsed, index is unchanged, and NULL is
     * returned.  The result is returned as the most parsimonious
     * type of Formattable that will accommodate all of the
     * necessary precision.  For example, if the result is exactly 12,
     * it will be returned as a long.  However, if it is 1.5, it will
     * be returned as a double.
     *
     * @param text           The text to be parsed.
     * @param result         Formattable to be set to the parse result.
     *                       If parse fails, return contents are undefined.
     * @param parsePosition  The position to start parsing at on input.
     *                       On output, moved to after the last successfully
     *                       parse character. On parse failure, does not change.
     * @see Formattable
     * @stable ICU 2.0
     */
    void parse(const UnicodeString& text, Formattable& result,
               ParsePosition& parsePosition) const U_OVERRIDE;

    /**
     * Parses text from the given string as a currency amount.  Unlike
     * the parse() method, this method will attempt to parse a generic
     * currency name, searching for a match of this object's locale's
     * currency display names, or for a 3-letter ISO currency code.
     * This method will fail if this format is not a currency format,
     * that is, if it does not contain the currency pattern symbol
     * (U+00A4) in its prefix or suffix.
     *
     * @param text the string to parse
     * @param pos  input-output position; on input, the position within text
     *             to match; must have 0 <= pos.getIndex() < text.length();
     *             on output, the position after the last matched character.
     *             If the parse fails, the position in unchanged upon output.
     * @return     if parse succeeds, a pointer to a newly-created CurrencyAmount
     *             object (owned by the caller) containing information about
     *             the parsed currency; if parse fails, this is NULL.
     * @stable ICU 49
     */
    CurrencyAmount* parseCurrency(const UnicodeString& text, ParsePosition& pos) const U_OVERRIDE;

    /**
     * Returns the decimal format symbols, which is generally not changed
     * by the programmer or user.
     * @return desired DecimalFormatSymbols
     * @see DecimalFormatSymbols
     * @stable ICU 2.0
     */
    virtual const DecimalFormatSymbols* getDecimalFormatSymbols(void) const;

    /**
     * Sets the decimal format symbols, which is generally not changed
     * by the programmer or user.
     * @param symbolsToAdopt DecimalFormatSymbols to be adopted.
     * @stable ICU 2.0
     */
    virtual void adoptDecimalFormatSymbols(DecimalFormatSymbols* symbolsToAdopt);

    /**
     * Sets the decimal format symbols, which is generally not changed
     * by the programmer or user.
     * @param symbols DecimalFormatSymbols.
     * @stable ICU 2.0
     */
    virtual void setDecimalFormatSymbols(const DecimalFormatSymbols& symbols);


    /**
     * Returns the currency plural format information,
     * which is generally not changed by the programmer or user.
     * @return desired CurrencyPluralInfo
     * @stable ICU 4.2
     */
    virtual const CurrencyPluralInfo* getCurrencyPluralInfo(void) const;

    /**
     * Sets the currency plural format information,
     * which is generally not changed by the programmer or user.
     * @param toAdopt CurrencyPluralInfo to be adopted.
     * @stable ICU 4.2
     */
    virtual void adoptCurrencyPluralInfo(CurrencyPluralInfo* toAdopt);

    /**
     * Sets the currency plural format information,
     * which is generally not changed by the programmer or user.
     * @param info Currency Plural Info.
     * @stable ICU 4.2
     */
    virtual void setCurrencyPluralInfo(const CurrencyPluralInfo& info);


    /**
     * Get the positive prefix.
     *
     * @param result    Output param which will receive the positive prefix.
     * @return          A reference to 'result'.
     * Examples: +123, $123, sFr123
     * @stable ICU 2.0
     */
    UnicodeString& getPositivePrefix(UnicodeString& result) const;

    /**
     * Set the positive prefix.
     *
     * @param newValue    the new value of the the positive prefix to be set.
     * Examples: +123, $123, sFr123
     * @stable ICU 2.0
     */
    virtual void setPositivePrefix(const UnicodeString& newValue);

    /**
     * Get the negative prefix.
     *
     * @param result    Output param which will receive the negative prefix.
     * @return          A reference to 'result'.
     * Examples: -123, ($123) (with negative suffix), sFr-123
     * @stable ICU 2.0
     */
    UnicodeString& getNegativePrefix(UnicodeString& result) const;

    /**
     * Set the negative prefix.
     *
     * @param newValue    the new value of the the negative prefix to be set.
     * Examples: -123, ($123) (with negative suffix), sFr-123
     * @stable ICU 2.0
     */
    virtual void setNegativePrefix(const UnicodeString& newValue);

    /**
     * Get the positive suffix.
     *
     * @param result    Output param which will receive the positive suffix.
     * @return          A reference to 'result'.
     * Example: 123%
     * @stable ICU 2.0
     */
    UnicodeString& getPositiveSuffix(UnicodeString& result) const;

    /**
     * Set the positive suffix.
     *
     * @param newValue    the new value of the positive suffix to be set.
     * Example: 123%
     * @stable ICU 2.0
     */
    virtual void setPositiveSuffix(const UnicodeString& newValue);

    /**
     * Get the negative suffix.
     *
     * @param result    Output param which will receive the negative suffix.
     * @return          A reference to 'result'.
     * Examples: -123%, ($123) (with positive suffixes)
     * @stable ICU 2.0
     */
    UnicodeString& getNegativeSuffix(UnicodeString& result) const;

    /**
     * Set the negative suffix.
     *
     * @param newValue    the new value of the negative suffix to be set.
     * Examples: 123%
     * @stable ICU 2.0
     */
    virtual void setNegativeSuffix(const UnicodeString& newValue);

#ifndef U_HIDE_DRAFT_API
    /**
     * Whether to show the plus sign on positive (non-negative) numbers; for example, "+12"
     *
     * For more control over sign display, use NumberFormatter.
     *
     * @return Whether the sign is shown on positive numbers and zero.
     * @draft ICU 64
     */
    UBool isSignAlwaysShown() const;

    /**
     * Set whether to show the plus sign on positive (non-negative) numbers; for example, "+12".
     *
     * For more control over sign display, use NumberFormatter.
     *
     * @param value true to always show a sign; false to hide the sign on positive numbers and zero.
     * @draft ICU 64
     */
    void setSignAlwaysShown(UBool value);
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Get the multiplier for use in percent, permill, etc.
     * For a percentage, set the suffixes to have "%" and the multiplier to be 100.
     * (For Arabic, use arabic percent symbol).
     * For a permill, set the suffixes to have "\\u2031" and the multiplier to be 1000.
     *
     * The number may also be multiplied by a power of ten; see getMultiplierScale().
     *
     * @return    the multiplier for use in percent, permill, etc.
     * Examples: with 100, 1.23 -> "123", and "123" -> 1.23
     * @stable ICU 2.0
     */
    int32_t getMultiplier(void) const;

    /**
     * Set the multiplier for use in percent, permill, etc.
     * For a percentage, set the suffixes to have "%" and the multiplier to be 100.
     * (For Arabic, use arabic percent symbol).
     * For a permill, set the suffixes to have "\\u2031" and the multiplier to be 1000.
     *
     * This method only supports integer multipliers. To multiply by a non-integer, pair this
     * method with setMultiplierScale().
     *
     * @param newValue    the new value of the multiplier for use in percent, permill, etc.
     * Examples: with 100, 1.23 -> "123", and "123" -> 1.23
     * @stable ICU 2.0
     */
    virtual void setMultiplier(int32_t newValue);

    /**
     * Gets the power of ten by which number should be multiplied before formatting, which
     * can be combined with setMultiplier() to multiply by any arbitrary decimal value.
     *
     * A multiplier scale of 2 corresponds to multiplication by 100, and a multiplier scale
     * of -2 corresponds to multiplication by 0.01.
     *
     * This method is analogous to UNUM_SCALE in getAttribute.
     *
     * @return    the current value of the power-of-ten multiplier.
     * @stable ICU 62
     */
    int32_t getMultiplierScale(void) const;

    /**
     * Sets a power of ten by which number should be multiplied before formatting, which
     * can be combined with setMultiplier() to multiply by any arbitrary decimal value.
     *
     * A multiplier scale of 2 corresponds to multiplication by 100, and a multiplier scale
     * of -2 corresponds to multiplication by 0.01.
     *
     * For example, to multiply numbers by 0.5 before formatting, you can do:
     *
     * <pre>
     * df.setMultiplier(5);
     * df.setMultiplierScale(-1);
     * </pre>
     *
     * This method is analogous to UNUM_SCALE in setAttribute.
     *
     * @param newValue    the new value of the power-of-ten multiplier.
     * @stable ICU 62
     */
    void setMultiplierScale(int32_t newValue);

    /**
     * Get the rounding increment.
     * @return A positive rounding increment, or 0.0 if a custom rounding
     * increment is not in effect.
     * @see #setRoundingIncrement
     * @see #getRoundingMode
     * @see #setRoundingMode
     * @stable ICU 2.0
     */
    virtual double getRoundingIncrement(void) const;

    /**
     * Set the rounding increment.  In the absence of a rounding increment,
     *    numbers will be rounded to the number of digits displayed.
     * @param newValue A positive rounding increment, or 0.0 to
     * use the default rounding increment.
     * Negative increments are equivalent to 0.0.
     * @see #getRoundingIncrement
     * @see #getRoundingMode
     * @see #setRoundingMode
     * @stable ICU 2.0
     */
    virtual void setRoundingIncrement(double newValue);

    /**
     * Get the rounding mode.
     * @return A rounding mode
     * @see #setRoundingIncrement
     * @see #getRoundingIncrement
     * @see #setRoundingMode
     * @stable ICU 2.0
     */
    virtual ERoundingMode getRoundingMode(void) const U_OVERRIDE;

    /**
     * Set the rounding mode.
     * @param roundingMode A rounding mode
     * @see #setRoundingIncrement
     * @see #getRoundingIncrement
     * @see #getRoundingMode
     * @stable ICU 2.0
     */
    virtual void setRoundingMode(ERoundingMode roundingMode) U_OVERRIDE;

    /**
     * Get the width to which the output of format() is padded.
     * The width is counted in 16-bit code units.
     * @return the format width, or zero if no padding is in effect
     * @see #setFormatWidth
     * @see #getPadCharacterString
     * @see #setPadCharacter
     * @see #getPadPosition
     * @see #setPadPosition
     * @stable ICU 2.0
     */
    virtual int32_t getFormatWidth(void) const;

    /**
     * Set the width to which the output of format() is padded.
     * The width is counted in 16-bit code units.
     * This method also controls whether padding is enabled.
     * @param width the width to which to pad the result of
     * format(), or zero to disable padding.  A negative
     * width is equivalent to 0.
     * @see #getFormatWidth
     * @see #getPadCharacterString
     * @see #setPadCharacter
     * @see #getPadPosition
     * @see #setPadPosition
     * @stable ICU 2.0
     */
    virtual void setFormatWidth(int32_t width);

    /**
     * Get the pad character used to pad to the format width.  The
     * default is ' '.
     * @return a string containing the pad character. This will always
     * have a length of one 32-bit code point.
     * @see #setFormatWidth
     * @see #getFormatWidth
     * @see #setPadCharacter
     * @see #getPadPosition
     * @see #setPadPosition
     * @stable ICU 2.0
     */
    virtual UnicodeString getPadCharacterString() const;

    /**
     * Set the character used to pad to the format width.  If padding
     * is not enabled, then this will take effect if padding is later
     * enabled.
     * @param padChar a string containing the pad character. If the string
     * has length 0, then the pad character is set to ' '.  Otherwise
     * padChar.char32At(0) will be used as the pad character.
     * @see #setFormatWidth
     * @see #getFormatWidth
     * @see #getPadCharacterString
     * @see #getPadPosition
     * @see #setPadPosition
     * @stable ICU 2.0
     */
    virtual void setPadCharacter(const UnicodeString& padChar);

    /**
     * Get the position at which padding will take place.  This is the location
     * at which padding will be inserted if the result of format()
     * is shorter than the format width.
     * @return the pad position, one of kPadBeforePrefix,
     * kPadAfterPrefix, kPadBeforeSuffix, or
     * kPadAfterSuffix.
     * @see #setFormatWidth
     * @see #getFormatWidth
     * @see #setPadCharacter
     * @see #getPadCharacterString
     * @see #setPadPosition
     * @see #EPadPosition
     * @stable ICU 2.0
     */
    virtual EPadPosition getPadPosition(void) const;

    /**
     * Set the position at which padding will take place.  This is the location
     * at which padding will be inserted if the result of format()
     * is shorter than the format width.  This has no effect unless padding is
     * enabled.
     * @param padPos the pad position, one of kPadBeforePrefix,
     * kPadAfterPrefix, kPadBeforeSuffix, or
     * kPadAfterSuffix.
     * @see #setFormatWidth
     * @see #getFormatWidth
     * @see #setPadCharacter
     * @see #getPadCharacterString
     * @see #getPadPosition
     * @see #EPadPosition
     * @stable ICU 2.0
     */
    virtual void setPadPosition(EPadPosition padPos);

    /**
     * Return whether or not scientific notation is used.
     * @return TRUE if this object formats and parses scientific notation
     * @see #setScientificNotation
     * @see #getMinimumExponentDigits
     * @see #setMinimumExponentDigits
     * @see #isExponentSignAlwaysShown
     * @see #setExponentSignAlwaysShown
     * @stable ICU 2.0
     */
    virtual UBool isScientificNotation(void) const;

    /**
     * Set whether or not scientific notation is used. When scientific notation
     * is used, the effective maximum number of integer digits is <= 8.  If the
     * maximum number of integer digits is set to more than 8, the effective
     * maximum will be 1.  This allows this call to generate a 'default' scientific
     * number format without additional changes.
     * @param useScientific TRUE if this object formats and parses scientific
     * notation
     * @see #isScientificNotation
     * @see #getMinimumExponentDigits
     * @see #setMinimumExponentDigits
     * @see #isExponentSignAlwaysShown
     * @see #setExponentSignAlwaysShown
     * @stable ICU 2.0
     */
    virtual void setScientificNotation(UBool useScientific);

    /**
     * Return the minimum exponent digits that will be shown.
     * @return the minimum exponent digits that will be shown
     * @see #setScientificNotation
     * @see #isScientificNotation
     * @see #setMinimumExponentDigits
     * @see #isExponentSignAlwaysShown
     * @see #setExponentSignAlwaysShown
     * @stable ICU 2.0
     */
    virtual int8_t getMinimumExponentDigits(void) const;

    /**
     * Set the minimum exponent digits that will be shown.  This has no
     * effect unless scientific notation is in use.
     * @param minExpDig a value >= 1 indicating the fewest exponent digits
     * that will be shown.  Values less than 1 will be treated as 1.
     * @see #setScientificNotation
     * @see #isScientificNotation
     * @see #getMinimumExponentDigits
     * @see #isExponentSignAlwaysShown
     * @see #setExponentSignAlwaysShown
     * @stable ICU 2.0
     */
    virtual void setMinimumExponentDigits(int8_t minExpDig);

    /**
     * Return whether the exponent sign is always shown.
     * @return TRUE if the exponent is always prefixed with either the
     * localized minus sign or the localized plus sign, false if only negative
     * exponents are prefixed with the localized minus sign.
     * @see #setScientificNotation
     * @see #isScientificNotation
     * @see #setMinimumExponentDigits
     * @see #getMinimumExponentDigits
     * @see #setExponentSignAlwaysShown
     * @stable ICU 2.0
     */
    virtual UBool isExponentSignAlwaysShown(void) const;

    /**
     * Set whether the exponent sign is always shown.  This has no effect
     * unless scientific notation is in use.
     * @param expSignAlways TRUE if the exponent is always prefixed with either
     * the localized minus sign or the localized plus sign, false if only
     * negative exponents are prefixed with the localized minus sign.
     * @see #setScientificNotation
     * @see #isScientificNotation
     * @see #setMinimumExponentDigits
     * @see #getMinimumExponentDigits
     * @see #isExponentSignAlwaysShown
     * @stable ICU 2.0
     */
    virtual void setExponentSignAlwaysShown(UBool expSignAlways);

    /**
     * Return the grouping size. Grouping size is the number of digits between
     * grouping separators in the integer portion of a number.  For example,
     * in the number "123,456.78", the grouping size is 3.
     *
     * @return    the grouping size.
     * @see setGroupingSize
     * @see NumberFormat::isGroupingUsed
     * @see DecimalFormatSymbols::getGroupingSeparator
     * @stable ICU 2.0
     */
    int32_t getGroupingSize(void) const;

    /**
     * Set the grouping size. Grouping size is the number of digits between
     * grouping separators in the integer portion of a number.  For example,
     * in the number "123,456.78", the grouping size is 3.
     *
     * @param newValue    the new value of the grouping size.
     * @see getGroupingSize
     * @see NumberFormat::setGroupingUsed
     * @see DecimalFormatSymbols::setGroupingSeparator
     * @stable ICU 2.0
     */
    virtual void setGroupingSize(int32_t newValue);

    /**
     * Return the secondary grouping size. In some locales one
     * grouping interval is used for the least significant integer
     * digits (the primary grouping size), and another is used for all
     * others (the secondary grouping size).  A formatter supporting a
     * secondary grouping size will return a positive integer unequal
     * to the primary grouping size returned by
     * getGroupingSize().  For example, if the primary
     * grouping size is 4, and the secondary grouping size is 2, then
     * the number 123456789 formats as "1,23,45,6789", and the pattern
     * appears as "#,##,###0".
     * @return the secondary grouping size, or a value less than
     * one if there is none
     * @see setSecondaryGroupingSize
     * @see NumberFormat::isGroupingUsed
     * @see DecimalFormatSymbols::getGroupingSeparator
     * @stable ICU 2.4
     */
    int32_t getSecondaryGroupingSize(void) const;

    /**
     * Set the secondary grouping size. If set to a value less than 1,
     * then secondary grouping is turned off, and the primary grouping
     * size is used for all intervals, not just the least significant.
     *
     * @param newValue    the new value of the secondary grouping size.
     * @see getSecondaryGroupingSize
     * @see NumberFormat#setGroupingUsed
     * @see DecimalFormatSymbols::setGroupingSeparator
     * @stable ICU 2.4
     */
    virtual void setSecondaryGroupingSize(int32_t newValue);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns the minimum number of grouping digits.
     * Grouping separators are output if there are at least this many
     * digits to the left of the first (rightmost) grouping separator,
     * that is, there are at least (minimum grouping + grouping size) integer digits.
     * (Subject to isGroupingUsed().)
     *
     * For example, if this value is 2, and the grouping size is 3, then
     * 9999 -> "9999" and 10000 -> "10,000"
     *
     * The default value for this attribute is 0.
     * A value of 1, 0, or lower, means that the use of grouping separators
     * only depends on the grouping size (and on isGroupingUsed()).
     *
     * NOTE: The CLDR data is used in NumberFormatter but not in DecimalFormat.
     * This is for backwards compatibility reasons.
     *
     * For more control over grouping strategies, use NumberFormatter.
     *
     * @see setMinimumGroupingDigits
     * @see getGroupingSize
     * @draft ICU 64
     */
    int32_t getMinimumGroupingDigits() const;

    /**
     * Sets the minimum grouping digits. Setting to a value less than or
     * equal to 1 turns off minimum grouping digits.
     *
     * For more control over grouping strategies, use NumberFormatter.
     *
     * @param newValue the new value of minimum grouping digits.
     * @see getMinimumGroupingDigits
     * @draft ICU 64
     */
    void setMinimumGroupingDigits(int32_t newValue);
#endif  /* U_HIDE_DRAFT_API */


    /**
     * Allows you to get the behavior of the decimal separator with integers.
     * (The decimal separator will always appear with decimals.)
     *
     * @return    TRUE if the decimal separator always appear with decimals.
     * Example: Decimal ON: 12345 -> 12345.; OFF: 12345 -> 12345
     * @stable ICU 2.0
     */
    UBool isDecimalSeparatorAlwaysShown(void) const;

    /**
     * Allows you to set the behavior of the decimal separator with integers.
     * (The decimal separator will always appear with decimals.)
     *
     * @param newValue    set TRUE if the decimal separator will always appear with decimals.
     * Example: Decimal ON: 12345 -> 12345.; OFF: 12345 -> 12345
     * @stable ICU 2.0
     */
    virtual void setDecimalSeparatorAlwaysShown(UBool newValue);

    /**
     * Allows you to get the parse behavior of the pattern decimal mark.
     *
     * @return    TRUE if input must contain a match to decimal mark in pattern
     * @stable ICU 54
     */
    UBool isDecimalPatternMatchRequired(void) const;

    /**
     * Allows you to set the parse behavior of the pattern decimal mark.
     *
     * if TRUE, the input must have a decimal mark if one was specified in the pattern. When
     * FALSE the decimal mark may be omitted from the input.
     *
     * @param newValue    set TRUE if input must contain a match to decimal mark in pattern
     * @stable ICU 54
     */
    virtual void setDecimalPatternMatchRequired(UBool newValue);

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns whether to ignore exponents when parsing.
     *
     * @return Whether to ignore exponents when parsing.
     * @see #setParseNoExponent
     * @draft ICU 64
     */
    UBool isParseNoExponent() const;

    /**
     * Specifies whether to stop parsing when an exponent separator is encountered. For
     * example, parses "123E4" to 123 (with parse position 3) instead of 1230000 (with parse position
     * 5).
     *
     * @param value true to prevent exponents from being parsed; false to allow them to be parsed.
     * @draft ICU 64
     */
    void setParseNoExponent(UBool value);

    /**
     * Returns whether parsing is sensitive to case (lowercase/uppercase).
     *
     * @return Whether parsing is case-sensitive.
     * @see #setParseCaseSensitive
     * @draft ICU 64
     */
    UBool isParseCaseSensitive() const;

    /**
     * Whether to pay attention to case when parsing; default is to ignore case (perform
     * case-folding). For example, "A" == "a" in case-insensitive but not case-sensitive mode.
     *
     * Currency symbols are never case-folded. For example, "us$1.00" will not parse in case-insensitive
     * mode, even though "US$1.00" parses.
     *
     * @param value true to enable case-sensitive parsing (the default); false to force
     *              case-sensitive parsing behavior.
     * @draft ICU 64
     */
    void setParseCaseSensitive(UBool value);

    /**
     * Returns whether truncation of high-order integer digits should result in an error.
     * By default, setMaximumIntegerDigits truncates high-order digits silently.
     *
     * @return Whether an error code is set if high-order digits are truncated.
     * @see setFormatFailIfMoreThanMaxDigits
     * @draft ICU 64
     */
    UBool isFormatFailIfMoreThanMaxDigits() const;

    /**
     * Sets whether truncation of high-order integer digits should result in an error.
     * By default, setMaximumIntegerDigits truncates high-order digits silently.
     *
     * @param value Whether to set an error code if high-order digits are truncated.
     * @draft ICU 64
     */
    void setFormatFailIfMoreThanMaxDigits(UBool value);
#endif  /* U_HIDE_DRAFT_API */


    /**
     * Synthesizes a pattern string that represents the current state
     * of this Format object.
     *
     * @param result    Output param which will receive the pattern.
     *                  Previous contents are deleted.
     * @return          A reference to 'result'.
     * @see applyPattern
     * @stable ICU 2.0
     */
    virtual UnicodeString& toPattern(UnicodeString& result) const;

    /**
     * Synthesizes a localized pattern string that represents the current
     * state of this Format object.
     *
     * @param result    Output param which will receive the localized pattern.
     *                  Previous contents are deleted.
     * @return          A reference to 'result'.
     * @see applyPattern
     * @stable ICU 2.0
     */
    virtual UnicodeString& toLocalizedPattern(UnicodeString& result) const;

    /**
     * Apply the given pattern to this Format object.  A pattern is a
     * short-hand specification for the various formatting properties.
     * These properties can also be changed individually through the
     * various setter methods.
     * <P>
     * There is no limit to integer digits are set
     * by this routine, since that is the typical end-user desire;
     * use setMaximumInteger if you want to set a real value.
     * For negative numbers, use a second pattern, separated by a semicolon
     * <pre>
     * .      Example "#,#00.0#" -> 1,234.56
     * </pre>
     * This means a minimum of 2 integer digits, 1 fraction digit, and
     * a maximum of 2 fraction digits.
     * <pre>
     * .      Example: "#,#00.0#;(#,#00.0#)" for negatives in parantheses.
     * </pre>
     * In negative patterns, the minimum and maximum counts are ignored;
     * these are presumed to be set in the positive pattern.
     *
     * @param pattern    The pattern to be applied.
     * @param parseError Struct to recieve information on position
     *                   of error if an error is encountered
     * @param status     Output param set to success/failure code on
     *                   exit. If the pattern is invalid, this will be
     *                   set to a failure result.
     * @stable ICU 2.0
     */
    virtual void applyPattern(const UnicodeString& pattern, UParseError& parseError, UErrorCode& status);

    /**
     * Sets the pattern.
     * @param pattern   The pattern to be applied.
     * @param status    Output param set to success/failure code on
     *                  exit. If the pattern is invalid, this will be
     *                  set to a failure result.
     * @stable ICU 2.0
     */
    virtual void applyPattern(const UnicodeString& pattern, UErrorCode& status);

    /**
     * Apply the given pattern to this Format object.  The pattern
     * is assumed to be in a localized notation. A pattern is a
     * short-hand specification for the various formatting properties.
     * These properties can also be changed individually through the
     * various setter methods.
     * <P>
     * There is no limit to integer digits are set
     * by this routine, since that is the typical end-user desire;
     * use setMaximumInteger if you want to set a real value.
     * For negative numbers, use a second pattern, separated by a semicolon
     * <pre>
     * .      Example "#,#00.0#" -> 1,234.56
     * </pre>
     * This means a minimum of 2 integer digits, 1 fraction digit, and
     * a maximum of 2 fraction digits.
     *
     * Example: "#,#00.0#;(#,#00.0#)" for negatives in parantheses.
     *
     * In negative patterns, the minimum and maximum counts are ignored;
     * these are presumed to be set in the positive pattern.
     *
     * @param pattern   The localized pattern to be applied.
     * @param parseError Struct to recieve information on position
     *                   of error if an error is encountered
     * @param status    Output param set to success/failure code on
     *                  exit. If the pattern is invalid, this will be
     *                  set to a failure result.
     * @stable ICU 2.0
     */
    virtual void applyLocalizedPattern(const UnicodeString& pattern, UParseError& parseError,
                                       UErrorCode& status);

    /**
     * Apply the given pattern to this Format object.
     *
     * @param pattern   The localized pattern to be applied.
     * @param status    Output param set to success/failure code on
     *                  exit. If the pattern is invalid, this will be
     *                  set to a failure result.
     * @stable ICU 2.0
     */
    virtual void applyLocalizedPattern(const UnicodeString& pattern, UErrorCode& status);


    /**
     * Sets the maximum number of digits allowed in the integer portion of a
     * number. This override limits the integer digit count to 309.
     *
     * @param newValue    the new value of the maximum number of digits
     *                      allowed in the integer portion of a number.
     * @see NumberFormat#setMaximumIntegerDigits
     * @stable ICU 2.0
     */
    void setMaximumIntegerDigits(int32_t newValue) U_OVERRIDE;

    /**
     * Sets the minimum number of digits allowed in the integer portion of a
     * number. This override limits the integer digit count to 309.
     *
     * @param newValue    the new value of the minimum number of digits
     *                      allowed in the integer portion of a number.
     * @see NumberFormat#setMinimumIntegerDigits
     * @stable ICU 2.0
     */
    void setMinimumIntegerDigits(int32_t newValue) U_OVERRIDE;

    /**
     * Sets the maximum number of digits allowed in the fraction portion of a
     * number. This override limits the fraction digit count to 340.
     *
     * @param newValue    the new value of the maximum number of digits
     *                    allowed in the fraction portion of a number.
     * @see NumberFormat#setMaximumFractionDigits
     * @stable ICU 2.0
     */
    void setMaximumFractionDigits(int32_t newValue) U_OVERRIDE;

    /**
     * Sets the minimum number of digits allowed in the fraction portion of a
     * number. This override limits the fraction digit count to 340.
     *
     * @param newValue    the new value of the minimum number of digits
     *                    allowed in the fraction portion of a number.
     * @see NumberFormat#setMinimumFractionDigits
     * @stable ICU 2.0
     */
    void setMinimumFractionDigits(int32_t newValue) U_OVERRIDE;

    /**
     * Returns the minimum number of significant digits that will be
     * displayed. This value has no effect unless areSignificantDigitsUsed()
     * returns true.
     * @return the fewest significant digits that will be shown
     * @stable ICU 3.0
     */
    int32_t getMinimumSignificantDigits() const;

    /**
     * Returns the maximum number of significant digits that will be
     * displayed. This value has no effect unless areSignificantDigitsUsed()
     * returns true.
     * @return the most significant digits that will be shown
     * @stable ICU 3.0
     */
    int32_t getMaximumSignificantDigits() const;

    /**
     * Sets the minimum number of significant digits that will be
     * displayed.  If <code>min</code> is less than one then it is set
     * to one.  If the maximum significant digits count is less than
     * <code>min</code>, then it is set to <code>min</code>.
     * This function also enables the use of significant digits
     * by this formatter - areSignificantDigitsUsed() will return TRUE.
     * @see #areSignificantDigitsUsed
     * @param min the fewest significant digits to be shown
     * @stable ICU 3.0
     */
    void setMinimumSignificantDigits(int32_t min);

    /**
     * Sets the maximum number of significant digits that will be
     * displayed.  If <code>max</code> is less than one then it is set
     * to one.  If the minimum significant digits count is greater
     * than <code>max</code>, then it is set to <code>max</code>.
     * This function also enables the use of significant digits
     * by this formatter - areSignificantDigitsUsed() will return TRUE.
     * @see #areSignificantDigitsUsed
     * @param max the most significant digits to be shown
     * @stable ICU 3.0
     */
    void setMaximumSignificantDigits(int32_t max);

    /**
     * Returns true if significant digits are in use, or false if
     * integer and fraction digit counts are in use.
     * @return true if significant digits are in use
     * @stable ICU 3.0
     */
    UBool areSignificantDigitsUsed() const;

    /**
     * Sets whether significant digits are in use, or integer and
     * fraction digit counts are in use.
     * @param useSignificantDigits true to use significant digits, or
     * false to use integer and fraction digit counts
     * @stable ICU 3.0
     */
    void setSignificantDigitsUsed(UBool useSignificantDigits);

    /**
     * Sets the currency used to display currency
     * amounts.  This takes effect immediately, if this format is a
     * currency format.  If this format is not a currency format, then
     * the currency is used if and when this object becomes a
     * currency format through the application of a new pattern.
     * @param theCurrency a 3-letter ISO code indicating new currency
     * to use.  It need not be null-terminated.  May be the empty
     * string or NULL to indicate no currency.
     * @param ec input-output error code
     * @stable ICU 3.0
     */
    void setCurrency(const char16_t* theCurrency, UErrorCode& ec) U_OVERRIDE;

#ifndef U_FORCE_HIDE_DEPRECATED_API
    /**
     * Sets the currency used to display currency amounts.  See
     * setCurrency(const char16_t*, UErrorCode&).
     * @deprecated ICU 3.0. Use setCurrency(const char16_t*, UErrorCode&).
     */
    virtual void setCurrency(const char16_t* theCurrency);
#endif  // U_FORCE_HIDE_DEPRECATED_API

    /**
     * Sets the `Currency Usage` object used to display currency.
     * This takes effect immediately, if this format is a
     * currency format.
     * @param newUsage new currency usage object to use.
     * @param ec input-output error code
     * @stable ICU 54
     */
    void setCurrencyUsage(UCurrencyUsage newUsage, UErrorCode* ec);

    /**
     * Returns the `Currency Usage` object used to display currency
     * @stable ICU 54
     */
    UCurrencyUsage getCurrencyUsage() const;

#ifndef U_HIDE_INTERNAL_API

    /**
     *  Format a number and save it into the given DecimalQuantity.
     *  Internal, not intended for public use.
     *  @internal
     */
    void formatToDecimalQuantity(double number, number::impl::DecimalQuantity& output,
                                 UErrorCode& status) const;

    /**
     *  Get a DecimalQuantity corresponding to a formattable as it would be
     *  formatted by this DecimalFormat.
     *  Internal, not intended for public use.
     *  @internal
     */
    void formatToDecimalQuantity(const Formattable& number, number::impl::DecimalQuantity& output,
                                 UErrorCode& status) const;

#endif  /* U_HIDE_INTERNAL_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Converts this DecimalFormat to a (Localized)NumberFormatter. Starting
     * in ICU 60, NumberFormatter is the recommended way to format numbers.
     * You can use the returned LocalizedNumberFormatter to format numbers and
     * get a FormattedNumber, which contains a string as well as additional
     * annotations about the formatted value.
     * 
     * If a memory allocation failure occurs, the return value of this method
     * might be null. If you are concerned about correct recovery from
     * out-of-memory situations, use this pattern:
     *
     * <pre>
     * FormattedNumber result;
     * if (auto* ptr = df->toNumberFormatter(status)) {
     *     result = ptr->formatDouble(123, status);
     * }
     * </pre>
     *
     * If you are not concerned about out-of-memory situations, or if your
     * environment throws exceptions when memory allocation failure occurs,
     * you can chain the methods, like this:
     *
     * <pre>
     * FormattedNumber result = df
     *     ->toNumberFormatter(status)
     *     ->formatDouble(123, status);
     * </pre>
     *
     * NOTE: The returned LocalizedNumberFormatter is owned by this DecimalFormat.
     * If a non-const method is called on the DecimalFormat, or if the DecimalFormat
     * is deleted, the object becomes invalid. If you plan to keep the return value
     * beyond the lifetime of the DecimalFormat, copy it to a local variable:
     *
     * <pre>
     * LocalizedNumberFormatter lnf;
     * if (auto* ptr = df->toNumberFormatter(status)) {
     *     lnf = *ptr;
     * }
     * </pre>
     *
     * @param status Set on failure, like U_MEMORY_ALLOCATION_ERROR.
     * @return A pointer to an internal object, or nullptr on failure.
     *         Do not delete the return value!
     * @draft ICU 64
     */
    const number::LocalizedNumberFormatter* toNumberFormatter(UErrorCode& status) const;
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Return the class ID for this class.  This is useful only for
     * comparing to a return value from getDynamicClassID().  For example:
     * <pre>
     * .      Base* polymorphic_pointer = createPolymorphicObject();
     * .      if (polymorphic_pointer->getDynamicClassID() ==
     * .          Derived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 2.0
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY.  Pure virtual override.
     * This method is to implement a simple version of RTTI, since not all
     * C++ compilers support genuine RTTI.  Polymorphic operator==() and
     * clone() methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 2.0
     */
    UClassID getDynamicClassID(void) const U_OVERRIDE;

  private:

    /** Rebuilds the formatter object from the property bag. */
    void touch(UErrorCode& status);

    /** Rebuilds the formatter object, ignoring any error code. */
    void touchNoError();

    /**
     * Updates the property bag with settings from the given pattern.
     *
     * @param pattern The pattern string to parse.
     * @param ignoreRounding Whether to leave out rounding information (minFrac, maxFrac, and rounding
     *     increment) when parsing the pattern. This may be desirable if a custom rounding mode, such
     *     as CurrencyUsage, is to be used instead. One of {@link
     *     PatternStringParser#IGNORE_ROUNDING_ALWAYS}, {@link PatternStringParser#IGNORE_ROUNDING_IF_CURRENCY},
     *     or {@link PatternStringParser#IGNORE_ROUNDING_NEVER}.
     * @see PatternAndPropertyUtils#parseToExistingProperties
     */
    void setPropertiesFromPattern(const UnicodeString& pattern, int32_t ignoreRounding,
                                  UErrorCode& status);

    const numparse::impl::NumberParserImpl* getParser(UErrorCode& status) const;

    const numparse::impl::NumberParserImpl* getCurrencyParser(UErrorCode& status) const;

    static void fieldPositionHelper(const number::FormattedNumber& formatted, FieldPosition& fieldPosition,
                                    int32_t offset, UErrorCode& status);

    static void fieldPositionIteratorHelper(const number::FormattedNumber& formatted,
                                            FieldPositionIterator* fpi, int32_t offset, UErrorCode& status);

    void setupFastFormat();

    bool fastFormatDouble(double input, UnicodeString& output) const;

    bool fastFormatInt64(int64_t input, UnicodeString& output) const;

    void doFastFormatInt32(int32_t input, bool isNegative, UnicodeString& output) const;

    //=====================================================================================//
    //                                   INSTANCE FIELDS                                   //
    //=====================================================================================//


    // One instance field for the implementation, keep all fields inside of an implementation
    // class defined in number_mapper.h
    number::impl::DecimalFormatFields* fields = nullptr;

    // Allow child class CompactDecimalFormat to access fProperties:
    friend class CompactDecimalFormat;

    // Allow MeasureFormat to use fieldPositionHelper:
    friend class MeasureFormat;

};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



struct UHashtable;

U_NAMESPACE_BEGIN

class PluralRules;

/**
 * **IMPORTANT:** New users are strongly encouraged to see if
 * numberformatter.h fits their use case.  Although not deprecated, this header
 * is provided for backwards compatibility only.
 *
 * -----------------------------------------------------------------------------
 *
 * The CompactDecimalFormat produces abbreviated numbers, suitable for display in
 * environments will limited real estate. For example, 'Hits: 1.2B' instead of
 * 'Hits: 1,200,000,000'. The format will be appropriate for the given language,
 * such as "1,2 Mrd." for German.
 *
 * For numbers under 1000 trillion (under 10^15, such as 123,456,789,012,345),
 * the result will be short for supported languages. However, the result may
 * sometimes exceed 7 characters, such as when there are combining marks or thin
 * characters. In such cases, the visual width in fonts should still be short.
 *
 * By default, there are 3 significant digits. After creation, if more than
 * three significant digits are set (with setMaximumSignificantDigits), or if a
 * fixed number of digits are set (with setMaximumIntegerDigits or
 * setMaximumFractionDigits), then result may be wider.
 *
 * At this time, parsing is not supported, and will produce a U_UNSUPPORTED_ERROR.
 * Resetting the pattern prefixes or suffixes is not supported; the method calls
 * are ignored.
 *
 * @stable ICU 51
 */
class U_I18N_API CompactDecimalFormat : public DecimalFormat {
public:

     /**
      * Returns a compact decimal instance for specified locale.
      *
      * **NOTE:** New users are strongly encouraged to use
      * `number::NumberFormatter` instead of NumberFormat.
      * @param inLocale the given locale.
      * @param style whether to use short or long style.
      * @param status error code returned  here.
      * @stable ICU 51
      */
     static CompactDecimalFormat* U_EXPORT2 createInstance(
          const Locale& inLocale, UNumberCompactStyle style, UErrorCode& status);

    /**
     * Copy constructor.
     *
     * @param source    the DecimalFormat object to be copied from.
     * @stable ICU 51
      */
    CompactDecimalFormat(const CompactDecimalFormat& source);

    /**
     * Destructor.
     * @stable ICU 51
     */
    ~CompactDecimalFormat() U_OVERRIDE;

    /**
     * Assignment operator.
     *
     * @param rhs    the DecimalFormat object to be copied.
     * @stable ICU 51
     */
    CompactDecimalFormat& operator=(const CompactDecimalFormat& rhs);

    /**
     * Clone this Format object polymorphically. The caller owns the
     * result and should delete it when done.
     *
     * @return    a polymorphic copy of this CompactDecimalFormat.
     * @stable ICU 51
     */
    CompactDecimalFormat* clone() const U_OVERRIDE;

    using DecimalFormat::format;

    /**
     * CompactDecimalFormat does not support parsing. This implementation
     * does nothing.
     * @param text           Unused.
     * @param result         Does not change.
     * @param parsePosition  Does not change.
     * @see Formattable
     * @stable ICU 51
     */
    void parse(const UnicodeString& text, Formattable& result,
               ParsePosition& parsePosition) const U_OVERRIDE;

    /**
     * CompactDecimalFormat does not support parsing. This implementation
     * sets status to U_UNSUPPORTED_ERROR
     *
     * @param text      Unused.
     * @param result    Does not change.
     * @param status    Always set to U_UNSUPPORTED_ERROR.
     * @stable ICU 51
     */
    void parse(const UnicodeString& text, Formattable& result, UErrorCode& status) const U_OVERRIDE;

#ifndef U_HIDE_INTERNAL_API
    /**
     * Parses text from the given string as a currency amount.  Unlike
     * the parse() method, this method will attempt to parse a generic
     * currency name, searching for a match of this object's locale's
     * currency display names, or for a 3-letter ISO currency code.
     * This method will fail if this format is not a currency format,
     * that is, if it does not contain the currency pattern symbol
     * (U+00A4) in its prefix or suffix. This implementation always returns
     * NULL.
     *
     * @param text the string to parse
     * @param pos  input-output position; on input, the position within text
     *             to match; must have 0 <= pos.getIndex() < text.length();
     *             on output, the position after the last matched character.
     *             If the parse fails, the position in unchanged upon output.
     * @return     if parse succeeds, a pointer to a newly-created CurrencyAmount
     *             object (owned by the caller) containing information about
     *             the parsed currency; if parse fails, this is NULL.
     * @internal
     */
    CurrencyAmount* parseCurrency(const UnicodeString& text, ParsePosition& pos) const U_OVERRIDE;
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Return the class ID for this class.  This is useful only for
     * comparing to a return value from getDynamicClassID().  For example:
     * <pre>
     * .      Base* polymorphic_pointer = createPolymorphicObject();
     * .      if (polymorphic_pointer->getDynamicClassID() ==
     * .          Derived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 51
     */
    static UClassID U_EXPORT2 getStaticClassID();

    /**
     * Returns a unique class ID POLYMORPHICALLY.  Pure virtual override.
     * This method is to implement a simple version of RTTI, since not all
     * C++ compilers support genuine RTTI.  Polymorphic operator==() and
     * clone() methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 51
     */
    UClassID getDynamicClassID() const U_OVERRIDE;

  private:
    CompactDecimalFormat(const Locale& inLocale, UNumberCompactStyle style, UErrorCode& status);
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

#include <atomic>
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

#include <cstdint>



// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING






// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING




// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html




#if !UCONFIG_NO_FORMATTING



#ifndef U_HIDE_DRAFT_API

/**
 * \file
 * \brief C API: Abstract operations for localized strings.
 * 
 * This file contains declarations for classes that deal with formatted strings. A number
 * of APIs throughout ICU use these classes for expressing their localized output.
 */


/**
 * All possible field categories in ICU. Every entry in this enum corresponds
 * to another enum that exists in ICU.
 * 
 * In the APIs that take a UFieldCategory, an int32_t type is used. Field
 * categories having any of the top four bits turned on are reserved as
 * private-use for external APIs implementing FormattedValue. This means that
 * categories 2^28 and higher or below zero (with the highest bit turned on)
 * are private-use and will not be used by ICU in the future.
 *
 * @draft ICU 64
 */
typedef enum UFieldCategory {
    /**
     * For an undefined field category.
     * 
     * @draft ICU 64
     */
    UFIELD_CATEGORY_UNDEFINED = 0,

    /**
     * For fields in UDateFormatField (udat.h), from ICU 3.0.
     *
     * @draft ICU 64
     */
    UFIELD_CATEGORY_DATE,

    /**
     * For fields in UNumberFormatFields (unum.h), from ICU 49.
     *
     * @draft ICU 64
     */
    UFIELD_CATEGORY_NUMBER,

    /**
     * For fields in UListFormatterField (ulistformatter.h), from ICU 63.
     *
     * @draft ICU 64
     */
    UFIELD_CATEGORY_LIST,

    /**
     * For fields in URelativeDateTimeFormatterField (ureldatefmt.h), from ICU 64.
     *
     * @draft ICU 64
     */
    UFIELD_CATEGORY_RELATIVE_DATETIME,

    /**
     * Reserved for possible future fields in UDateIntervalFormatField.
     *
     * @internal
     */
    UFIELD_CATEGORY_DATE_INTERVAL,

#ifndef U_HIDE_INTERNAL_API
    /** @internal */
    UFIELD_CATEGORY_COUNT,
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Category for spans in a list.
     *
     * @draft ICU 64
     */
    UFIELD_CATEGORY_LIST_SPAN = 0x1000 + UFIELD_CATEGORY_LIST,

    /**
     * Category for spans in a date interval.
     *
     * @draft ICU 64
     */
    UFIELD_CATEGORY_DATE_INTERVAL_SPAN = 0x1000 + UFIELD_CATEGORY_DATE_INTERVAL,

} UFieldCategory;


struct UConstrainedFieldPosition;
/**
 * Represents a span of a string containing a given field.
 *
 * This struct differs from UFieldPosition in the following ways:
 *
 *   1. It has information on the field category.
 *   2. It allows you to set constraints to use when iterating over field positions.
 *   3. It is used for the newer FormattedValue APIs.
 *
 * @draft ICU 64
 */
typedef struct UConstrainedFieldPosition UConstrainedFieldPosition;


/**
 * Creates a new UConstrainedFieldPosition.
 *
 * By default, the UConstrainedFieldPosition has no iteration constraints.
 *
 * @param ec Set if an error occurs.
 * @return The new object, or NULL if an error occurs.
 * @draft ICU 64
 */
U_DRAFT UConstrainedFieldPosition* U_EXPORT2
ucfpos_open(UErrorCode* ec);


/**
 * Resets a UConstrainedFieldPosition to its initial state, as if it were newly created.
 *
 * Removes any constraints that may have been set on the instance.
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @param ec Set if an error occurs.
 * @draft ICU 64
 */
U_DRAFT void U_EXPORT2
ucfpos_reset(
    UConstrainedFieldPosition* ucfpos,
    UErrorCode* ec);


/**
 * Destroys a UConstrainedFieldPosition and releases its memory.
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @draft ICU 64
 */
U_DRAFT void U_EXPORT2
ucfpos_close(UConstrainedFieldPosition* ucfpos);


/**
 * Sets a constraint on the field category.
 * 
 * When this instance of UConstrainedFieldPosition is passed to ufmtval_nextPosition,
 * positions are skipped unless they have the given category.
 *
 * Any previously set constraints are cleared.
 *
 * For example, to loop over only the number-related fields:
 *
 *     UConstrainedFieldPosition* ucfpos = ucfpos_open(ec);
 *     ucfpos_constrainCategory(ucfpos, UFIELDCATEGORY_NUMBER_FORMAT, ec);
 *     while (ufmtval_nextPosition(ufmtval, ucfpos, ec)) {
 *         // handle the number-related field position
 *     }
 *     ucfpos_close(ucfpos);
 *
 * Changing the constraint while in the middle of iterating over a FormattedValue
 * does not generally have well-defined behavior.
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @param category The field category to fix when iterating.
 * @param ec Set if an error occurs.
 * @draft ICU 64
 */
U_DRAFT void U_EXPORT2
ucfpos_constrainCategory(
    UConstrainedFieldPosition* ucfpos,
    int32_t category,
    UErrorCode* ec);


/**
 * Sets a constraint on the category and field.
 * 
 * When this instance of UConstrainedFieldPosition is passed to ufmtval_nextPosition,
 * positions are skipped unless they have the given category and field.
 *
 * Any previously set constraints are cleared.
 *
 * For example, to loop over all grouping separators:
 *
 *     UConstrainedFieldPosition* ucfpos = ucfpos_open(ec);
 *     ucfpos_constrainField(ucfpos, UFIELDCATEGORY_NUMBER_FORMAT, UNUM_GROUPING_SEPARATOR_FIELD, ec);
 *     while (ufmtval_nextPosition(ufmtval, ucfpos, ec)) {
 *         // handle the grouping separator position
 *     }
 *     ucfpos_close(ucfpos);
 *
 * Changing the constraint while in the middle of iterating over a FormattedValue
 * does not generally have well-defined behavior.
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @param category The field category to fix when iterating.
 * @param field The field to fix when iterating.
 * @param ec Set if an error occurs.
 * @draft ICU 64
 */
U_DRAFT void U_EXPORT2
ucfpos_constrainField(
    UConstrainedFieldPosition* ucfpos,
    int32_t category,
    int32_t field,
    UErrorCode* ec);


/**
 * Gets the field category for the current position.
 *
 * If a category or field constraint was set, this function returns the constrained
 * category. Otherwise, the return value is well-defined only after
 * ufmtval_nextPosition returns TRUE.
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @param ec Set if an error occurs.
 * @return The field category saved in the instance.
 * @draft ICU 64
 */
U_DRAFT int32_t U_EXPORT2
ucfpos_getCategory(
    const UConstrainedFieldPosition* ucfpos,
    UErrorCode* ec);


/**
 * Gets the field for the current position.
 *
 * If a field constraint was set, this function returns the constrained
 * field. Otherwise, the return value is well-defined only after
 * ufmtval_nextPosition returns TRUE.
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @param ec Set if an error occurs.
 * @return The field saved in the instance.
 * @draft ICU 64
 */
U_DRAFT int32_t U_EXPORT2
ucfpos_getField(
    const UConstrainedFieldPosition* ucfpos,
    UErrorCode* ec);


/**
 * Gets the INCLUSIVE start and EXCLUSIVE end index stored for the current position.
 *
 * The output values are well-defined only after ufmtval_nextPosition returns TRUE.
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @param pStart Set to the start index saved in the instance. Ignored if nullptr.
 * @param pLimit Set to the end index saved in the instance. Ignored if nullptr.
 * @param ec Set if an error occurs.
 * @draft ICU 64
 */
U_DRAFT void U_EXPORT2
ucfpos_getIndexes(
    const UConstrainedFieldPosition* ucfpos,
    int32_t* pStart,
    int32_t* pLimit,
    UErrorCode* ec);


/**
 * Gets an int64 that FormattedValue implementations may use for storage.
 *
 * The initial value is zero.
 *
 * Users of FormattedValue should not need to call this method.
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @param ec Set if an error occurs.
 * @return The current iteration context from ucfpos_setInt64IterationContext.
 * @draft ICU 64
 */
U_DRAFT int64_t U_EXPORT2
ucfpos_getInt64IterationContext(
    const UConstrainedFieldPosition* ucfpos,
    UErrorCode* ec);


/**
 * Sets an int64 that FormattedValue implementations may use for storage.
 *
 * Intended to be used by FormattedValue implementations.
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @param context The new iteration context.
 * @param ec Set if an error occurs.
 * @draft ICU 64
 */
U_DRAFT void U_EXPORT2
ucfpos_setInt64IterationContext(
    UConstrainedFieldPosition* ucfpos,
    int64_t context,
    UErrorCode* ec);


/**
 * Determines whether a given field should be included given the
 * constraints.
 *
 * Intended to be used by FormattedValue implementations.
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @param category The category to test.
 * @param field The field to test.
 * @param ec Set if an error occurs.
 * @draft ICU 64
 */
U_DRAFT UBool U_EXPORT2
ucfpos_matchesField(
    const UConstrainedFieldPosition* ucfpos,
    int32_t category,
    int32_t field,
    UErrorCode* ec);


/**
 * Sets new values for the primary public getters.
 *
 * Intended to be used by FormattedValue implementations.
 *
 * It is up to the implementation to ensure that the user-requested
 * constraints are satisfied. This method does not check!
 *
 * @param ucfpos The instance of UConstrainedFieldPosition.
 * @param category The new field category.
 * @param field The new field.
 * @param start The new inclusive start index.
 * @param limit The new exclusive end index.
 * @param ec Set if an error occurs.
 * @draft ICU 64
 */
U_DRAFT void U_EXPORT2
ucfpos_setState(
    UConstrainedFieldPosition* ucfpos,
    int32_t category,
    int32_t field,
    int32_t start,
    int32_t limit,
    UErrorCode* ec);


struct UFormattedValue;
/**
 * An abstract formatted value: a string with associated field attributes.
 * Many formatters format to types compatible with UFormattedValue.
 *
 * @draft ICU 64
 */
typedef struct UFormattedValue UFormattedValue;


/**
 * Returns a pointer to the formatted string. The pointer is owned by the UFormattedValue. The
 * return value is valid only as long as the UFormattedValue is present and unchanged in memory.
 *
 * The return value is NUL-terminated but could contain internal NULs.
 *
 * @param ufmtval
 *         The object containing the formatted string and attributes.
 * @param pLength Output variable for the length of the string. Ignored if NULL.
 * @param ec Set if an error occurs.
 * @return A NUL-terminated char16 string owned by the UFormattedValue.
 * @draft ICU 64
 */
U_DRAFT const UChar* U_EXPORT2
ufmtval_getString(
    const UFormattedValue* ufmtval,
    int32_t* pLength,
    UErrorCode* ec);


/**
 * Iterates over field positions in the UFormattedValue. This lets you determine the position
 * of specific types of substrings, like a month or a decimal separator.
 *
 * To loop over all field positions:
 *
 *     UConstrainedFieldPosition* ucfpos = ucfpos_open(ec);
 *     while (ufmtval_nextPosition(ufmtval, ucfpos, ec)) {
 *         // handle the field position; get information from ucfpos
 *     }
 *     ucfpos_close(ucfpos);
 *
 * @param ufmtval
 *         The object containing the formatted string and attributes.
 * @param ucfpos
 *         The object used for iteration state; can provide constraints to iterate over only
 *         one specific category or field;
 *         see ucfpos_constrainCategory
 *         and ucfpos_constrainField.
 * @param ec Set if an error occurs.
 * @return TRUE if another position was found; FALSE otherwise.
 * @draft ICU 64
 */
U_DRAFT UBool U_EXPORT2
ufmtval_nextPosition(
    const UFormattedValue* ufmtval,
    UConstrainedFieldPosition* ucfpos,
    UErrorCode* ec);


#if U_SHOW_CPLUSPLUS_API
U_NAMESPACE_BEGIN

/**
 * \class LocalUConstrainedFieldPositionPointer
 * "Smart pointer" class; closes a UConstrainedFieldPosition via ucfpos_close().
 * For most methods see the LocalPointerBase base class.
 *
 * Usage:
 *
 *     LocalUConstrainedFieldPositionPointer ucfpos(ucfpos_open(ec));
 *     // no need to explicitly call ucfpos_close()
 *
 * @draft ICU 64
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUConstrainedFieldPositionPointer,
    UConstrainedFieldPosition,
    ucfpos_close);

U_NAMESPACE_END
#endif // U_SHOW_CPLUSPLUS_API


#endif  /* U_HIDE_DRAFT_API */
#endif /* #if !UCONFIG_NO_FORMATTING */


U_NAMESPACE_BEGIN

/**
 * \file
 * \brief C++ API: Abstract operations for localized strings.
 * 
 * This file contains declarations for classes that deal with formatted strings. A number
 * of APIs throughout ICU use these classes for expressing their localized output.
 */


// The following cannot have #ifndef U_HIDE_DRAFT_API because
// class FormattedValue depends on it, and FormattedValue cannot be
// hidden becauseclass FormattedNumber (stable ICU 60) depends on it.
#ifndef U_FORCE_HIDE_DRAFT_API
/**
 * Represents a span of a string containing a given field.
 *
 * This class differs from FieldPosition in the following ways:
 *
 *   1. It has information on the field category.
 *   2. It allows you to set constraints to use when iterating over field positions.
 *   3. It is used for the newer FormattedValue APIs.
 *
 * This class is not intended for public subclassing.
 *
 * @draft ICU 64
 */
class U_I18N_API ConstrainedFieldPosition : public UMemory {
  public:

    /**
     * Initializes a ConstrainedFieldPosition.
     *
     * By default, the ConstrainedFieldPosition has no iteration constraints.
     *
     * @draft ICU 64
     */
    ConstrainedFieldPosition();

    /** @draft ICU 64 */
    ~ConstrainedFieldPosition();

#ifndef U_HIDE_DRAFT_API
    /**
     * Resets this ConstrainedFieldPosition to its initial state, as if it were newly created:
     *
     * - Removes any constraints that may have been set on the instance.
     * - Resets the iteration position.
     *
     * @draft ICU 64
     */
    void reset();

    /**
     * Sets a constraint on the field category.
     * 
     * When this instance of ConstrainedFieldPosition is passed to FormattedValue#nextPosition,
     * positions are skipped unless they have the given category.
     *
     * Any previously set constraints are cleared.
     *
     * For example, to loop over only the number-related fields:
     *
     *     ConstrainedFieldPosition cfpos;
     *     cfpos.constrainCategory(UFIELDCATEGORY_NUMBER_FORMAT);
     *     while (fmtval.nextPosition(cfpos, status)) {
     *         // handle the number-related field position
     *     }
     *
     * Changing the constraint while in the middle of iterating over a FormattedValue
     * does not generally have well-defined behavior.
     *
     * @param category The field category to fix when iterating.
     * @draft ICU 64
     */
    void constrainCategory(int32_t category);

    /**
     * Sets a constraint on the category and field.
     * 
     * When this instance of ConstrainedFieldPosition is passed to FormattedValue#nextPosition,
     * positions are skipped unless they have the given category and field.
     *
     * Any previously set constraints are cleared.
     *
     * For example, to loop over all grouping separators:
     *
     *     ConstrainedFieldPosition cfpos;
     *     cfpos.constrainField(UFIELDCATEGORY_NUMBER_FORMAT, UNUM_GROUPING_SEPARATOR_FIELD);
     *     while (fmtval.nextPosition(cfpos, status)) {
     *         // handle the grouping separator position
     *     }
     *
     * Changing the constraint while in the middle of iterating over a FormattedValue
     * does not generally have well-defined behavior.
     *
     * @param category The field category to fix when iterating.
     * @param field The field to fix when iterating.
     * @draft ICU 64
     */
    void constrainField(int32_t category, int32_t field);

    /**
     * Gets the field category for the current position.
     *
     * The return value is well-defined only after
     * FormattedValue#nextPosition returns TRUE.
     *
     * @return The field category saved in the instance.
     * @draft ICU 64
     */
    inline int32_t getCategory() const {
        return fCategory;
    }

    /**
     * Gets the field for the current position.
     *
     * The return value is well-defined only after
     * FormattedValue#nextPosition returns TRUE.
     *
     * @return The field saved in the instance.
     * @draft ICU 64
     */
    inline int32_t getField() const {
        return fField;
    }

    /**
     * Gets the INCLUSIVE start index for the current position.
     *
     * The return value is well-defined only after FormattedValue#nextPosition returns TRUE.
     *
     * @return The start index saved in the instance.
     * @draft ICU 64
     */
    inline int32_t getStart() const {
        return fStart;
    }

    /**
     * Gets the EXCLUSIVE end index stored for the current position.
     *
     * The return value is well-defined only after FormattedValue#nextPosition returns TRUE.
     *
     * @return The end index saved in the instance.
     * @draft ICU 64
     */
    inline int32_t getLimit() const {
        return fLimit;
    }

    ////////////////////////////////////////////////////////////////////
    //// The following methods are for FormattedValue implementers; ////
    //// most users can ignore them.                                ////
    ////////////////////////////////////////////////////////////////////

    /**
     * Gets an int64 that FormattedValue implementations may use for storage.
     *
     * The initial value is zero.
     *
     * Users of FormattedValue should not need to call this method.
     *
     * @return The current iteration context from {@link #setInt64IterationContext}.
     * @draft ICU 64
     */
    inline int64_t getInt64IterationContext() const {
        return fContext;
    }

    /**
     * Sets an int64 that FormattedValue implementations may use for storage.
     *
     * Intended to be used by FormattedValue implementations.
     *
     * @param context The new iteration context.
     * @draft ICU 64
     */
    void setInt64IterationContext(int64_t context);

    /**
     * Determines whether a given field should be included given the
     * constraints.
     *
     * Intended to be used by FormattedValue implementations.
     *
     * @param category The category to test.
     * @param field The field to test.
     * @draft ICU 64
     */
    UBool matchesField(int32_t category, int32_t field) const;

    /**
     * Sets new values for the primary public getters.
     *
     * Intended to be used by FormattedValue implementations.
     *
     * It is up to the implementation to ensure that the user-requested
     * constraints are satisfied. This method does not check!
     *
     * @param category The new field category.
     * @param field The new field.
     * @param start The new inclusive start index.
     * @param limit The new exclusive end index.
     * @draft ICU 64
     */
    void setState(
        int32_t category,
        int32_t field,
        int32_t start,
        int32_t limit);
#endif  /* U_HIDE_DRAFT_API */

  private:
    int64_t fContext = 0LL;
    int32_t fField = 0;
    int32_t fStart = 0;
    int32_t fLimit = 0;
#ifndef U_HIDE_DRAFT_API
    int32_t fCategory = UFIELD_CATEGORY_UNDEFINED;
#else   /* U_HIDE_DRAFT_API */
    int32_t fCategory = 0;
#endif  /* U_HIDE_DRAFT_API */
    int8_t fConstraint = 0;
};

// The following cannot have #ifndef U_HIDE_DRAFT_API because
// class FormattedNumber (stable ICU 60) depends on it.
/**
 * An abstract formatted value: a string with associated field attributes.
 * Many formatters format to classes implementing FormattedValue.
 *
 * @draft ICU 64
 */
class U_I18N_API FormattedValue /* not : public UObject because this is an interface/mixin class */ {
  public:
    /** @draft ICU 64 */
    virtual ~FormattedValue();

    /**
     * Returns the formatted string as a self-contained UnicodeString.
     *
     * If you need the string within the current scope only, consider #toTempString.
     *
     * @param status Set if an error occurs.
     * @return a UnicodeString containing the formatted string.
     *
     * @draft ICU 64
     */
    virtual UnicodeString toString(UErrorCode& status) const = 0;

    /**
     * Returns the formatted string as a read-only alias to memory owned by the FormattedValue.
     *
     * The return value is valid only as long as this FormattedValue is present and unchanged in
     * memory. If you need the string outside the current scope, consider #toString.
     *
     * The buffer returned by calling UnicodeString#getBuffer() on the return value is
     * guaranteed to be NUL-terminated.
     *
     * @param status Set if an error occurs.
     * @return a temporary UnicodeString containing the formatted string.
     *
     * @draft ICU 64
     */
    virtual UnicodeString toTempString(UErrorCode& status) const = 0;

    /**
     * Appends the formatted string to an Appendable.
     *
     * @param appendable
     *         The Appendable to which to append the string output.
     * @param status Set if an error occurs.
     * @return The same Appendable, for chaining.
     *
     * @draft ICU 64
     * @see Appendable
     */
    virtual Appendable& appendTo(Appendable& appendable, UErrorCode& status) const = 0;

    /**
     * Iterates over field positions in the FormattedValue. This lets you determine the position
     * of specific types of substrings, like a month or a decimal separator.
     *
     * To loop over all field positions:
     *
     *     ConstrainedFieldPosition cfpos;
     *     while (fmtval.nextPosition(cfpos, status)) {
     *         // handle the field position; get information from cfpos
     *     }
     *
     * @param cfpos
     *         The object used for iteration state. This can provide constraints to iterate over
     *         only one specific category or field;
     *         see ConstrainedFieldPosition#constrainCategory
     *         and ConstrainedFieldPosition#constrainField.
     * @param status Set if an error occurs.
     * @return TRUE if a new occurrence of the field was found;
     *         FALSE otherwise or if an error was set.
     *
     * @draft ICU 64
     */
    virtual UBool nextPosition(ConstrainedFieldPosition& cfpos, UErrorCode& status) const = 0;
};
#endif  // U_FORCE_HIDE_DRAFT_API

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */




// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2009-2017, International Business Machines Corporation,       *
 * Google, and others. All Rights Reserved.                                    *
 *******************************************************************************
 */




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING



#ifndef U_HIDE_DRAFT_API

/**
 * \file
 * \brief C++ API: units for percent and permille
 */

U_NAMESPACE_BEGIN

/**
 * Dimensionless unit for percent and permille.
 * @see NumberFormatter
 * @draft ICU 60
 */
class U_I18N_API NoUnit: public MeasureUnit {
public:
    /**
     * Returns an instance for the base unit (dimensionless and no scaling).
     *
     * @return               a NoUnit instance
     * @draft ICU 60
     */
    static NoUnit U_EXPORT2 base();

    /**
     * Returns an instance for percent, or 1/100 of a base unit.
     *
     * @return               a NoUnit instance
     * @draft ICU 60
     */
    static NoUnit U_EXPORT2 percent();

    /**
     * Returns an instance for permille, or 1/1000 of a base unit.
     *
     * @return               a NoUnit instance
     * @draft ICU 60
     */
    static NoUnit U_EXPORT2 permille();

    /**
     * Copy operator.
     * @draft ICU 60
     */
    NoUnit(const NoUnit& other);

    /**
     * Destructor.
     * @draft ICU 60
     */
    virtual ~NoUnit();

    /**
     * Return a polymorphic clone of this object.  The result will
     * have the same class as returned by getDynamicClassID().
     * @draft ICU 60
     */
    virtual NoUnit* clone() const;

    /**
     * Returns a unique class ID for this object POLYMORPHICALLY.
     * This method implements a simple form of RTTI used by ICU.
     * @return The class ID for this object. All objects of a given
     * class have the same class ID.  Objects of other classes have
     * different class IDs.
     * @draft ICU 60
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * Returns the class ID for this class. This is used to compare to
     * the return value of getDynamicClassID().
     * @return The class ID for all objects of this class.
     * @draft ICU 60
     */
    static UClassID U_EXPORT2 getStaticClassID();

private:
    /**
     * Constructor
     * @internal (private)
     */
    NoUnit(const char* subtype);

};

U_NAMESPACE_END

#endif  /* U_HIDE_DRAFT_API */
#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2008-2015, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
*
* File PLURRULE.H
*
* Modification History:*
*   Date        Name        Description
*
********************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: PluralRules object
 */

#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2010-2013, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/




#if !UCONFIG_NO_FORMATTING



#ifndef U_HIDE_INTERNAL_API

#endif  /* U_HIDE_INTERNAL_API */

// Forward-declaration
struct UFormattedNumber;

/**
 * \file
 * \brief C API: Plural rules, select plural keywords for numeric values.
 *
 * A UPluralRules object defines rules for mapping non-negative numeric
 * values onto a small set of keywords. Rules are constructed from a text
 * description, consisting of a series of keywords and conditions.
 * The uplrules_select function examines each condition in order and
 * returns the keyword for the first condition that matches the number.
 * If none match, the default rule(other) is returned.
 *
 * For more information, see the LDML spec, C.11 Language Plural Rules:
 * http://www.unicode.org/reports/tr35/#Language_Plural_Rules
 *
 * Keywords: ICU locale data has 6 predefined values -
 * 'zero', 'one', 'two', 'few', 'many' and 'other'. Callers need to check
 * the value of keyword returned by the uplrules_select function.
 *
 * These are based on CLDR <i>Language Plural Rules</i>. For these
 * predefined rules, see the CLDR page at
 * http://unicode.org/repos/cldr-tmp/trunk/diff/supplemental/language_plural_rules.html
 */

/**
 * Type of plurals and PluralRules.
 * @stable ICU 50
 */
enum UPluralType {
    /**
     * Plural rules for cardinal numbers: 1 file vs. 2 files.
     * @stable ICU 50
     */
    UPLURAL_TYPE_CARDINAL,
    /**
     * Plural rules for ordinal numbers: 1st file, 2nd file, 3rd file, 4th file, etc.
     * @stable ICU 50
     */
    UPLURAL_TYPE_ORDINAL,
#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UPluralType value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UPLURAL_TYPE_COUNT
#endif  /* U_HIDE_DEPRECATED_API */
};
/**
 * @stable ICU 50
 */
typedef enum UPluralType UPluralType;

/**
 * Opaque UPluralRules object for use in C programs.
 * @stable ICU 4.8
 */
struct UPluralRules;
typedef struct UPluralRules UPluralRules;  /**< C typedef for struct UPluralRules. @stable ICU 4.8 */

/**
 * Opens a new UPluralRules object using the predefined cardinal-number plural rules for a
 * given locale.
 * Same as uplrules_openForType(locale, UPLURAL_TYPE_CARDINAL, status).
 * @param locale The locale for which the rules are desired.
 * @param status A pointer to a UErrorCode to receive any errors.
 * @return A UPluralRules for the specified locale, or NULL if an error occurred.
 * @stable ICU 4.8
 */
U_CAPI UPluralRules* U_EXPORT2
uplrules_open(const char *locale, UErrorCode *status);

/**
 * Opens a new UPluralRules object using the predefined plural rules for a
 * given locale and the plural type.
 * @param locale The locale for which the rules are desired.
 * @param type The plural type (e.g., cardinal or ordinal).
 * @param status A pointer to a UErrorCode to receive any errors.
 * @return A UPluralRules for the specified locale, or NULL if an error occurred.
 * @stable ICU 50
 */
U_CAPI UPluralRules* U_EXPORT2
uplrules_openForType(const char *locale, UPluralType type, UErrorCode *status);

/**
 * Closes a UPluralRules object. Once closed it may no longer be used.
 * @param uplrules The UPluralRules object to close.
 * @stable ICU 4.8
 */
U_CAPI void U_EXPORT2
uplrules_close(UPluralRules *uplrules);


#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUPluralRulesPointer
 * "Smart pointer" class, closes a UPluralRules via uplrules_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.8
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUPluralRulesPointer, UPluralRules, uplrules_close);

U_NAMESPACE_END

#endif


/**
 * Given a floating-point number, returns the keyword of the first rule that
 * applies to the number, according to the supplied UPluralRules object.
 * @param uplrules The UPluralRules object specifying the rules.
 * @param number The number for which the rule has to be determined.
 * @param keyword An output buffer to write the keyword of the rule that
 *         applies to number.
 * @param capacity The capacity of the keyword buffer.
 * @param status A pointer to a UErrorCode to receive any errors.
 * @return The length of the keyword.
 * @stable ICU 4.8
 */
U_CAPI int32_t U_EXPORT2
uplrules_select(const UPluralRules *uplrules,
               double number,
               UChar *keyword, int32_t capacity,
               UErrorCode *status);

#ifndef U_HIDE_DRAFT_API
/**
 * Given a formatted number, returns the keyword of the first rule
 * that applies to the number, according to the supplied UPluralRules object.
 *
 * A UFormattedNumber allows you to specify an exponent or trailing zeros,
 * which can affect the plural category. To get a UFormattedNumber, see
 * {@link UNumberFormatter}.
 *
 * @param uplrules The UPluralRules object specifying the rules.
 * @param number The formatted number for which the rule has to be determined.
 * @param keyword The destination buffer for the keyword of the rule that
 *         applies to number.
 * @param capacity The capacity of the keyword buffer.
 * @param status A pointer to a UErrorCode to receive any errors.
 * @return The length of the keyword.
 * @draft ICU 64
 */
U_CAPI int32_t U_EXPORT2
uplrules_selectFormatted(const UPluralRules *uplrules,
               const struct UFormattedNumber* number,
               UChar *keyword, int32_t capacity,
               UErrorCode *status);
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_INTERNAL_API
/**
 * Given a number, returns the keyword of the first rule that applies to the
 * number, according to the UPluralRules object and given the number format
 * specified by the UNumberFormat object.
 * Note: This internal preview interface may be removed in the future if
 * an architecturally cleaner solution reaches stable status.
 * @param uplrules The UPluralRules object specifying the rules.
 * @param number The number for which the rule has to be determined.
 * @param fmt The UNumberFormat specifying how the number will be formatted
 *        (this can affect the plural form, e.g. "1 dollar" vs "1.0 dollars").
 *        If this is NULL, the function behaves like uplrules_select.
 * @param keyword An output buffer to write the keyword of the rule that
 *         applies to number.
 * @param capacity The capacity of the keyword buffer.
 * @param status A pointer to a UErrorCode to receive any errors.
 * @return The length of keyword.
 * @internal ICU 59 technology preview, may be removed in the future
 */
U_INTERNAL int32_t U_EXPORT2
uplrules_selectWithFormat(const UPluralRules *uplrules,
                          double number,
                          const UNumberFormat *fmt,
                          UChar *keyword, int32_t capacity,
                          UErrorCode *status);

#endif  /* U_HIDE_INTERNAL_API */

/**
 * Creates a string enumeration of all plural rule keywords used in this
 * UPluralRules object. The rule "other" is always present by default.
 * @param uplrules The UPluralRules object specifying the rules for
 *        a given locale.
 * @param status A pointer to a UErrorCode to receive any errors.
 * @return a string enumeration over plural rule keywords, or NULL
 * upon error. The caller is responsible for closing the result.
 * @stable ICU 59
 */
U_STABLE UEnumeration* U_EXPORT2
uplrules_getKeywords(const UPluralRules *uplrules,
                     UErrorCode *status);

#endif /* #if !UCONFIG_NO_FORMATTING */


#ifndef U_HIDE_INTERNAL_API

#endif  /* U_HIDE_INTERNAL_API */

/**
 * Value returned by PluralRules::getUniqueKeywordValue() when there is no
 * unique value to return.
 * @stable ICU 4.8
 */
#define UPLRULES_NO_UNIQUE_VALUE ((double)-0.00123456777)

U_NAMESPACE_BEGIN

class Hashtable;
class IFixedDecimal;
class RuleChain;
class PluralRuleParser;
class PluralKeywordEnumeration;
class AndConstraint;
class SharedPluralRules;

namespace number {
class FormattedNumber;
}

/**
 * Defines rules for mapping non-negative numeric values onto a small set of
 * keywords. Rules are constructed from a text description, consisting
 * of a series of keywords and conditions.  The {@link #select} method
 * examines each condition in order and returns the keyword for the
 * first condition that matches the number.  If none match,
 * default rule(other) is returned.
 *
 * For more information, details, and tips for writing rules, see the
 * LDML spec, C.11 Language Plural Rules:
 * http://www.unicode.org/draft/reports/tr35/tr35.html#Language_Plural_Rules
 *
 * Examples:<pre>
 *   "one: n is 1; few: n in 2..4"</pre>
 *  This defines two rules, for 'one' and 'few'.  The condition for
 *  'one' is "n is 1" which means that the number must be equal to
 *  1 for this condition to pass.  The condition for 'few' is
 *  "n in 2..4" which means that the number must be between 2 and
 *  4 inclusive for this condition to pass.  All other numbers
 *  are assigned the keyword "other" by the default rule.
 *  </p><pre>
 *    "zero: n is 0; one: n is 1; zero: n mod 100 in 1..19"</pre>
 *  This illustrates that the same keyword can be defined multiple times.
 *  Each rule is examined in order, and the first keyword whose condition
 *  passes is the one returned.  Also notes that a modulus is applied
 *  to n in the last rule.  Thus its condition holds for 119, 219, 319...
 *  </p><pre>
 *    "one: n is 1; few: n mod 10 in 2..4 and n mod 100 not in 12..14"</pre>
 *  This illustrates conjunction and negation.  The condition for 'few'
 *  has two parts, both of which must be met: "n mod 10 in 2..4" and
 *  "n mod 100 not in 12..14".  The first part applies a modulus to n
 *  before the test as in the previous example.  The second part applies
 *  a different modulus and also uses negation, thus it matches all
 *  numbers _not_ in 12, 13, 14, 112, 113, 114, 212, 213, 214...
 *  </p>
 *  <p>
 * Syntax:<pre>
 * \code
 * rules         = rule (';' rule)*
 * rule          = keyword ':' condition
 * keyword       = <identifier>
 * condition     = and_condition ('or' and_condition)*
 * and_condition = relation ('and' relation)*
 * relation      = is_relation | in_relation | within_relation | 'n' <EOL>
 * is_relation   = expr 'is' ('not')? value
 * in_relation   = expr ('not')? 'in' range_list
 * within_relation = expr ('not')? 'within' range
 * expr          = ('n' | 'i' | 'f' | 'v' | 'j') ('mod' value)?
 * range_list    = (range | value) (',' range_list)*
 * value         = digit+  ('.' digit+)?
 * digit         = 0|1|2|3|4|5|6|7|8|9
 * range         = value'..'value
 * \endcode
 * </pre></p>
 * <p>
 * <p>
 * The i, f, and v values are defined as follows:
 * </p>
 * <ul>
 * <li>i to be the integer digits.</li>
 * <li>f to be the visible fractional digits, as an integer.</li>
 * <li>v to be the number of visible fraction digits.</li>
 * <li>j is defined to only match integers. That is j is 3 fails if v != 0 (eg for 3.1 or 3.0).</li>
 * </ul>
 * <p>
 * Examples are in the following table:
 * </p>
 * <table border='1' style="border-collapse:collapse">
 * <tr>
 * <th>n</th>
 * <th>i</th>
 * <th>f</th>
 * <th>v</th>
 * </tr>
 * <tr>
 * <td>1.0</td>
 * <td>1</td>
 * <td align="right">0</td>
 * <td>1</td>
 * </tr>
 * <tr>
 * <td>1.00</td>
 * <td>1</td>
 * <td align="right">0</td>
 * <td>2</td>
 * </tr>
 * <tr>
 * <td>1.3</td>
 * <td>1</td>
 * <td align="right">3</td>
 * <td>1</td>
 * </tr>
 * <tr>
 * <td>1.03</td>
 * <td>1</td>
 * <td align="right">3</td>
 * <td>2</td>
 * </tr>
 * <tr>
 * <td>1.23</td>
 * <td>1</td>
 * <td align="right">23</td>
 * <td>2</td>
 * </tr>
 * </table>
 * <p>
 * The difference between 'in' and 'within' is that 'in' only includes integers in the specified range, while 'within'
 * includes all values. Using 'within' with a range_list consisting entirely of values is the same as using 'in' (it's
 * not an error).
 * </p>

 * An "identifier" is a sequence of characters that do not have the
 * Unicode Pattern_Syntax or Pattern_White_Space properties.
 * <p>
 * The difference between 'in' and 'within' is that 'in' only includes
 * integers in the specified range, while 'within' includes all values.
 * Using 'within' with a range_list consisting entirely of values is the
 * same as using 'in' (it's not an error).
 *</p>
 * <p>
 * Keywords
 * could be defined by users or from ICU locale data. There are 6
 * predefined values in ICU - 'zero', 'one', 'two', 'few', 'many' and
 * 'other'. Callers need to check the value of keyword returned by
 * {@link #select} method.
 * </p>
 *
 * Examples:<pre>
 * UnicodeString keyword = pl->select(number);
 * if (keyword== UnicodeString("one") {
 *     ...
 * }
 * else if ( ... )
 * </pre>
 * <strong>Note:</strong><br>
 *  <p>
 *   ICU defines plural rules for many locales based on CLDR <i>Language Plural Rules</i>.
 *   For these predefined rules, see CLDR page at
 *    http://unicode.org/repos/cldr-tmp/trunk/diff/supplemental/language_plural_rules.html
 * </p>
 */
class U_I18N_API PluralRules : public UObject {
public:

    /**
     * Constructor.
     * @param status  Output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     *
     * @stable ICU 4.0
     */
    PluralRules(UErrorCode& status);

    /**
     * Copy constructor.
     * @stable ICU 4.0
     */
    PluralRules(const PluralRules& other);

    /**
     * Destructor.
     * @stable ICU 4.0
     */
    virtual ~PluralRules();

    /**
     * Clone
     * @stable ICU 4.0
     */
    PluralRules* clone() const;

    /**
      * Assignment operator.
      * @stable ICU 4.0
      */
    PluralRules& operator=(const PluralRules&);

    /**
     * Creates a PluralRules from a description if it is parsable, otherwise
     * returns NULL.
     *
     * @param description rule description
     * @param status      Output param set to success/failure code on exit, which
     *                    must not indicate a failure before the function call.
     * @return            new PluralRules pointer. NULL if there is an error.
     * @stable ICU 4.0
     */
    static PluralRules* U_EXPORT2 createRules(const UnicodeString& description,
                                              UErrorCode& status);

    /**
     * The default rules that accept any number.
     *
     * @param status  Output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @return        new PluralRules pointer. NULL if there is an error.
     * @stable ICU 4.0
     */
    static PluralRules* U_EXPORT2 createDefaultRules(UErrorCode& status);

    /**
     * Provides access to the predefined cardinal-number <code>PluralRules</code> for a given
     * locale.
     * Same as forLocale(locale, UPLURAL_TYPE_CARDINAL, status).
     *
     * @param locale  The locale for which a <code>PluralRules</code> object is
     *                returned.
     * @param status  Output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @return        The predefined <code>PluralRules</code> object pointer for
     *                this locale. If there's no predefined rules for this locale,
     *                the rules for the closest parent in the locale hierarchy
     *                that has one will  be returned.  The final fallback always
     *                returns the default 'other' rules.
     * @stable ICU 4.0
     */
    static PluralRules* U_EXPORT2 forLocale(const Locale& locale, UErrorCode& status);

    /**
     * Provides access to the predefined <code>PluralRules</code> for a given
     * locale and the plural type.
     *
     * @param locale  The locale for which a <code>PluralRules</code> object is
     *                returned.
     * @param type    The plural type (e.g., cardinal or ordinal).
     * @param status  Output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @return        The predefined <code>PluralRules</code> object pointer for
     *                this locale. If there's no predefined rules for this locale,
     *                the rules for the closest parent in the locale hierarchy
     *                that has one will  be returned.  The final fallback always
     *                returns the default 'other' rules.
     * @stable ICU 50
     */
    static PluralRules* U_EXPORT2 forLocale(const Locale& locale, UPluralType type, UErrorCode& status);

#ifndef U_HIDE_INTERNAL_API
    /**
     * Return a StringEnumeration over the locales for which there is plurals data.
     * @return a StringEnumeration over the locales available.
     * @internal
     */
    static StringEnumeration* U_EXPORT2 getAvailableLocales(UErrorCode &status);

    /**
     * Returns whether or not there are overrides.
     * @param locale       the locale to check.
     * @return
     * @internal
     */
    static UBool hasOverride(const Locale &locale);

    /**
     * For ICU use only.
     * creates a  SharedPluralRules object
     * @internal
     */
    static PluralRules* U_EXPORT2 internalForLocale(const Locale& locale, UPluralType type, UErrorCode& status);

    /**
     * For ICU use only.
     * Returns handle to the shared, cached PluralRules instance.
     * Caller must call removeRef() on returned value once it is done with
     * the shared instance.
     * @internal
     */
    static const SharedPluralRules* U_EXPORT2 createSharedInstance(
            const Locale& locale, UPluralType type, UErrorCode& status);


#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Given an integer, returns the keyword of the first rule
     * that applies to  the number.  This function can be used with
     * isKeyword* functions to determine the keyword for default plural rules.
     *
     * @param number  The number for which the rule has to be determined.
     * @return        The keyword of the selected rule.
     * @stable ICU 4.0
     */
    UnicodeString select(int32_t number) const;

    /**
     * Given a floating-point number, returns the keyword of the first rule
     * that applies to  the number.  This function can be used with
     * isKeyword* functions to determine the keyword for default plural rules.
     *
     * @param number  The number for which the rule has to be determined.
     * @return        The keyword of the selected rule.
     * @stable ICU 4.0
     */
    UnicodeString select(double number) const;

#ifndef U_HIDE_DRAFT_API
    /**
     * Given a formatted number, returns the keyword of the first rule
     * that applies to  the number.  This function can be used with
     * isKeyword* functions to determine the keyword for default plural rules.
     *
     * A FormattedNumber allows you to specify an exponent or trailing zeros,
     * which can affect the plural category. To get a FormattedNumber, see
     * NumberFormatter.
     *
     * @param number  The number for which the rule has to be determined.
     * @param status  Set if an error occurs while selecting plural keyword.
     *                This could happen if the FormattedNumber is invalid.
     * @return        The keyword of the selected rule.
     * @draft ICU 64
     */
    UnicodeString select(const number::FormattedNumber& number, UErrorCode& status) const;
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_INTERNAL_API
    /**
      * @internal
      */
    UnicodeString select(const IFixedDecimal &number) const;
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Returns a list of all rule keywords used in this <code>PluralRules</code>
     * object.  The rule 'other' is always present by default.
     *
     * @param status Output param set to success/failure code on exit, which
     *               must not indicate a failure before the function call.
     * @return       StringEnumeration with the keywords.
     *               The caller must delete the object.
     * @stable ICU 4.0
     */
    StringEnumeration* getKeywords(UErrorCode& status) const;

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Deprecated Function, does not return useful results.
     *
     * Originally intended to return a unique value for this keyword if it exists,
     * else the constant UPLRULES_NO_UNIQUE_VALUE.
     *
     * @param keyword The keyword.
     * @return        Stub deprecated function returns UPLRULES_NO_UNIQUE_VALUE always.
     * @deprecated ICU 55
     */
    double getUniqueKeywordValue(const UnicodeString& keyword);

    /**
     * Deprecated Function, does not produce useful results.
     *
     * Originally intended to return all the values for which select() would return the keyword.
     * If the keyword is unknown, returns no values, but this is not an error.  If
     * the number of values is unlimited, returns no values and -1 as the
     * count.
     *
     * The number of returned values is typically small.
     *
     * @param keyword      The keyword.
     * @param dest         Array into which to put the returned values.  May
     *                     be NULL if destCapacity is 0.
     * @param destCapacity The capacity of the array, must be at least 0.
     * @param status       The error code. Deprecated function, always sets U_UNSUPPORTED_ERROR.
     * @return             The count of values available, or -1.  This count
     *                     can be larger than destCapacity, but no more than
     *                     destCapacity values will be written.
     * @deprecated ICU 55
     */
    int32_t getAllKeywordValues(const UnicodeString &keyword,
                                double *dest, int32_t destCapacity,
                                UErrorCode& status);
#endif  /* U_HIDE_DEPRECATED_API */

    /**
     * Returns sample values for which select() would return the keyword.  If
     * the keyword is unknown, returns no values, but this is not an error.
     *
     * The number of returned values is typically small.
     *
     * @param keyword      The keyword.
     * @param dest         Array into which to put the returned values.  May
     *                     be NULL if destCapacity is 0.
     * @param destCapacity The capacity of the array, must be at least 0.
     * @param status       The error code.
     * @return             The count of values written.
     *                     If more than destCapacity samples are available, then
     *                     only destCapacity are written, and destCapacity is returned as the count,
     *                     rather than setting a U_BUFFER_OVERFLOW_ERROR.
     *                     (The actual number of keyword values could be unlimited.)
     * @stable ICU 4.8
     */
    int32_t getSamples(const UnicodeString &keyword,
                       double *dest, int32_t destCapacity,
                       UErrorCode& status);

    /**
     * Returns TRUE if the given keyword is defined in this
     * <code>PluralRules</code> object.
     *
     * @param keyword  the input keyword.
     * @return         TRUE if the input keyword is defined.
     *                 Otherwise, return FALSE.
     * @stable ICU 4.0
     */
    UBool isKeyword(const UnicodeString& keyword) const;


    /**
     * Returns keyword for default plural form.
     *
     * @return         keyword for default plural form.
     * @stable ICU 4.0
     */
    UnicodeString getKeywordOther() const;

#ifndef U_HIDE_INTERNAL_API
    /**
     *
     * @internal
     */
     UnicodeString getRules() const;
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Compares the equality of two PluralRules objects.
     *
     * @param other The other PluralRules object to be compared with.
     * @return      True if the given PluralRules is the same as this
     *              PluralRules; false otherwise.
     * @stable ICU 4.0
     */
    virtual UBool operator==(const PluralRules& other) const;

    /**
     * Compares the inequality of two PluralRules objects.
     *
     * @param other The PluralRules object to be compared with.
     * @return      True if the given PluralRules is not the same as this
     *              PluralRules; false otherwise.
     * @stable ICU 4.0
     */
    UBool operator!=(const PluralRules& other) const  {return !operator==(other);}


    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 4.0
     *
    */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 4.0
     */
    virtual UClassID getDynamicClassID() const;


private:
    RuleChain  *mRules;

    PluralRules();   // default constructor not implemented
    void            parseDescription(const UnicodeString& ruleData, UErrorCode &status);
    int32_t         getNumberValue(const UnicodeString& token) const;
    UnicodeString   getRuleFromResource(const Locale& locale, UPluralType type, UErrorCode& status);
    RuleChain      *rulesForKeyword(const UnicodeString &keyword) const;

    /**
    * An internal status variable used to indicate that the object is in an 'invalid' state.
    * Used by copy constructor, the assignment operator and the clone method.
    */
    UErrorCode mInternalStatus;

    friend class PluralRuleParser;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */




// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING







/**
 * \file
 * \brief C-compatible API for localized number formatting; not recommended for C++.
 *
 * This is the C-compatible version of the NumberFormatter API introduced in ICU 60. C++ users should
 * include unicode/numberformatter.h and use the proper C++ APIs.
 *
 * The C API accepts a number skeleton string for specifying the settings for formatting, which covers a
 * very large subset of all possible number formatting features. For more information on number skeleton
 * strings, see unicode/numberformatter.h.
 *
 * When using UNumberFormatter, which is treated as immutable, the results are exported to a mutable
 * UFormattedNumber object, which you subsequently use for populating your string buffer or iterating over
 * the fields.
 *
 * Example code:
 * <pre>
 * // Setup:
 * UErrorCode ec = U_ZERO_ERROR;
 * UNumberFormatter* uformatter = unumf_openForSkeletonAndLocale(u"precision-integer", -1, "en", &ec);
 * UFormattedNumber* uresult = unumf_openResult(&ec);
 * if (U_FAILURE(ec)) { return; }
 *
 * // Format a double:
 * unumf_formatDouble(uformatter, 5142.3, uresult, &ec);
 * if (U_FAILURE(ec)) { return; }
 *
 * // Export the string to a malloc'd buffer:
 * int32_t len = unumf_resultToString(uresult, NULL, 0, &ec);
 * // at this point, ec == U_BUFFER_OVERFLOW_ERROR
 * ec = U_ZERO_ERROR;
 * UChar* buffer = (UChar*) malloc((len+1)*sizeof(UChar));
 * unumf_resultToString(uresult, buffer, len+1, &ec);
 * if (U_FAILURE(ec)) { return; }
 * // buffer should equal "5,142"
 *
 * // Cleanup:
 * unumf_close(uformatter);
 * unumf_closeResult(uresult);
 * free(buffer);
 * </pre>
 *
 * If you are a C++ user linking against the C libraries, you can use the LocalPointer versions of these
 * APIs. The following example uses LocalPointer with the decimal number and field position APIs:
 *
 * <pre>
 * // Setup:
 * LocalUNumberFormatterPointer uformatter(unumf_openForSkeletonAndLocale(u"percent", -1, "en", &ec));
 * LocalUFormattedNumberPointer uresult(unumf_openResult(&ec));
 * if (U_FAILURE(ec)) { return; }
 *
 * // Format a decimal number:
 * unumf_formatDecimal(uformatter.getAlias(), "9.87E-3", -1, uresult.getAlias(), &ec);
 * if (U_FAILURE(ec)) { return; }
 *
 * // Get the location of the percent sign:
 * UFieldPosition ufpos = {UNUM_PERCENT_FIELD, 0, 0};
 * unumf_resultNextFieldPosition(uresult.getAlias(), &ufpos, &ec);
 * // ufpos should contain beginIndex=7 and endIndex=8 since the string is "0.00987%"
 *
 * // No need to do any cleanup since we are using LocalPointer.
 * </pre>
 */

/**
 * An enum declaring how to render units, including currencies. Example outputs when formatting 123 USD and 123
 * meters in <em>en-CA</em>:
 *
 * <p>
 * <ul>
 * <li>NARROW*: "$123.00" and "123 m"
 * <li>SHORT: "US$ 123.00" and "123 m"
 * <li>FULL_NAME: "123.00 US dollars" and "123 meters"
 * <li>ISO_CODE: "USD 123.00" and undefined behavior
 * <li>HIDDEN: "123.00" and "123"
 * </ul>
 *
 * <p>
 * This enum is similar to {@link UMeasureFormatWidth}.
 *
 * @stable ICU 60
 */
typedef enum UNumberUnitWidth {
    /**
     * Print an abbreviated version of the unit name. Similar to SHORT, but always use the shortest available
     * abbreviation or symbol. This option can be used when the context hints at the identity of the unit. For more
     * information on the difference between NARROW and SHORT, see SHORT.
     *
     * <p>
     * In CLDR, this option corresponds to the "Narrow" format for measure units and the "¤¤¤¤¤" placeholder for
     * currencies.
     *
     * @stable ICU 60
     */
            UNUM_UNIT_WIDTH_NARROW,

    /**
     * Print an abbreviated version of the unit name. Similar to NARROW, but use a slightly wider abbreviation or
     * symbol when there may be ambiguity. This is the default behavior.
     *
     * <p>
     * For example, in <em>es-US</em>, the SHORT form for Fahrenheit is "{0} °F", but the NARROW form is "{0}°",
     * since Fahrenheit is the customary unit for temperature in that locale.
     *
     * <p>
     * In CLDR, this option corresponds to the "Short" format for measure units and the "¤" placeholder for
     * currencies.
     *
     * @stable ICU 60
     */
            UNUM_UNIT_WIDTH_SHORT,

    /**
     * Print the full name of the unit, without any abbreviations.
     *
     * <p>
     * In CLDR, this option corresponds to the default format for measure units and the "¤¤¤" placeholder for
     * currencies.
     *
     * @stable ICU 60
     */
            UNUM_UNIT_WIDTH_FULL_NAME,

    /**
     * Use the three-digit ISO XXX code in place of the symbol for displaying currencies. The behavior of this
     * option is currently undefined for use with measure units.
     *
     * <p>
     * In CLDR, this option corresponds to the "¤¤" placeholder for currencies.
     *
     * @stable ICU 60
     */
            UNUM_UNIT_WIDTH_ISO_CODE,

    /**
     * Format the number according to the specified unit, but do not display the unit. For currencies, apply
     * monetary symbols and formats as with SHORT, but omit the currency symbol. For measure units, the behavior is
     * equivalent to not specifying the unit at all.
     *
     * @stable ICU 60
     */
            UNUM_UNIT_WIDTH_HIDDEN,

    /**
     * One more than the highest UNumberUnitWidth value.
     *
     * @internal ICU 60: The numeric value may change over time; see ICU ticket #12420.
     */
            UNUM_UNIT_WIDTH_COUNT
} UNumberUnitWidth;

/**
 * An enum declaring the strategy for when and how to display grouping separators (i.e., the
 * separator, often a comma or period, after every 2-3 powers of ten). The choices are several
 * pre-built strategies for different use cases that employ locale data whenever possible. Example
 * outputs for 1234 and 1234567 in <em>en-IN</em>:
 *
 * <ul>
 * <li>OFF: 1234 and 12345
 * <li>MIN2: 1234 and 12,34,567
 * <li>AUTO: 1,234 and 12,34,567
 * <li>ON_ALIGNED: 1,234 and 12,34,567
 * <li>THOUSANDS: 1,234 and 1,234,567
 * </ul>
 *
 * <p>
 * The default is AUTO, which displays grouping separators unless the locale data says that grouping
 * is not customary. To force grouping for all numbers greater than 1000 consistently across locales,
 * use ON_ALIGNED. On the other hand, to display grouping less frequently than the default, use MIN2
 * or OFF. See the docs of each option for details.
 *
 * <p>
 * Note: This enum specifies the strategy for grouping sizes. To set which character to use as the
 * grouping separator, use the "symbols" setter.
 *
 * @stable ICU 63
 */
typedef enum UNumberGroupingStrategy {
    /**
     * Do not display grouping separators in any locale.
     *
     * @stable ICU 61
     */
            UNUM_GROUPING_OFF,

    /**
     * Display grouping using locale defaults, except do not show grouping on values smaller than
     * 10000 (such that there is a <em>minimum of two digits</em> before the first separator).
     *
     * <p>
     * Note that locales may restrict grouping separators to be displayed only on 1 million or
     * greater (for example, ee and hu) or disable grouping altogether (for example, bg currency).
     *
     * <p>
     * Locale data is used to determine whether to separate larger numbers into groups of 2
     * (customary in South Asia) or groups of 3 (customary in Europe and the Americas).
     *
     * @stable ICU 61
     */
            UNUM_GROUPING_MIN2,

    /**
     * Display grouping using the default strategy for all locales. This is the default behavior.
     *
     * <p>
     * Note that locales may restrict grouping separators to be displayed only on 1 million or
     * greater (for example, ee and hu) or disable grouping altogether (for example, bg currency).
     *
     * <p>
     * Locale data is used to determine whether to separate larger numbers into groups of 2
     * (customary in South Asia) or groups of 3 (customary in Europe and the Americas).
     *
     * @stable ICU 61
     */
            UNUM_GROUPING_AUTO,

    /**
     * Always display the grouping separator on values of at least 1000.
     *
     * <p>
     * This option ignores the locale data that restricts or disables grouping, described in MIN2 and
     * AUTO. This option may be useful to normalize the alignment of numbers, such as in a
     * spreadsheet.
     *
     * <p>
     * Locale data is used to determine whether to separate larger numbers into groups of 2
     * (customary in South Asia) or groups of 3 (customary in Europe and the Americas).
     *
     * @stable ICU 61
     */
            UNUM_GROUPING_ON_ALIGNED,

    /**
     * Use the Western defaults: groups of 3 and enabled for all numbers 1000 or greater. Do not use
     * locale data for determining the grouping strategy.
     *
     * @stable ICU 61
     */
            UNUM_GROUPING_THOUSANDS

#ifndef U_HIDE_INTERNAL_API
    ,
    /**
     * One more than the highest UNumberGroupingStrategy value.
     *
     * @internal ICU 62: The numeric value may change over time; see ICU ticket #12420.
     */
            UNUM_GROUPING_COUNT
#endif  /* U_HIDE_INTERNAL_API */

} UNumberGroupingStrategy;

/**
 * An enum declaring how to denote positive and negative numbers. Example outputs when formatting
 * 123, 0, and -123 in <em>en-US</em>:
 *
 * <ul>
 * <li>AUTO: "123", "0", and "-123"
 * <li>ALWAYS: "+123", "+0", and "-123"
 * <li>NEVER: "123", "0", and "123"
 * <li>ACCOUNTING: "$123", "$0", and "($123)"
 * <li>ACCOUNTING_ALWAYS: "+$123", "+$0", and "($123)"
 * <li>EXCEPT_ZERO: "+123", "0", and "-123"
 * <li>ACCOUNTING_EXCEPT_ZERO: "+$123", "$0", and "($123)"
 * </ul>
 *
 * <p>
 * The exact format, including the position and the code point of the sign, differ by locale.
 *
 * @stable ICU 60
 */
typedef enum UNumberSignDisplay {
    /**
     * Show the minus sign on negative numbers, and do not show the sign on positive numbers. This is the default
     * behavior.
     *
     * @stable ICU 60
     */
            UNUM_SIGN_AUTO,

    /**
     * Show the minus sign on negative numbers and the plus sign on positive numbers, including zero.
     * To hide the sign on zero, see {@link UNUM_SIGN_EXCEPT_ZERO}.
     *
     * @stable ICU 60
     */
            UNUM_SIGN_ALWAYS,

    /**
     * Do not show the sign on positive or negative numbers.
     *
     * @stable ICU 60
     */
            UNUM_SIGN_NEVER,

    /**
     * Use the locale-dependent accounting format on negative numbers, and do not show the sign on positive numbers.
     *
     * <p>
     * The accounting format is defined in CLDR and varies by locale; in many Western locales, the format is a pair
     * of parentheses around the number.
     *
     * <p>
     * Note: Since CLDR defines the accounting format in the monetary context only, this option falls back to the
     * AUTO sign display strategy when formatting without a currency unit. This limitation may be lifted in the
     * future.
     *
     * @stable ICU 60
     */
            UNUM_SIGN_ACCOUNTING,

    /**
     * Use the locale-dependent accounting format on negative numbers, and show the plus sign on
     * positive numbers, including zero. For more information on the accounting format, see the
     * ACCOUNTING sign display strategy. To hide the sign on zero, see
     * {@link UNUM_SIGN_ACCOUNTING_EXCEPT_ZERO}.
     *
     * @stable ICU 60
     */
            UNUM_SIGN_ACCOUNTING_ALWAYS,

    /**
     * Show the minus sign on negative numbers and the plus sign on positive numbers. Do not show a
     * sign on zero or NaN, unless the sign bit is set (-0.0 gets a sign).
     *
     * @stable ICU 61
     */
            UNUM_SIGN_EXCEPT_ZERO,

    /**
     * Use the locale-dependent accounting format on negative numbers, and show the plus sign on
     * positive numbers. Do not show a sign on zero or NaN, unless the sign bit is set (-0.0 gets a
     * sign). For more information on the accounting format, see the ACCOUNTING sign display
     * strategy.
     *
     * @stable ICU 61
     */
            UNUM_SIGN_ACCOUNTING_EXCEPT_ZERO,

    /**
     * One more than the highest UNumberSignDisplay value.
     *
     * @internal ICU 60: The numeric value may change over time; see ICU ticket #12420.
     */
            UNUM_SIGN_COUNT
} UNumberSignDisplay;

/**
 * An enum declaring how to render the decimal separator.
 *
 * <p>
 * <ul>
 * <li>UNUM_DECIMAL_SEPARATOR_AUTO: "1", "1.1"
 * <li>UNUM_DECIMAL_SEPARATOR_ALWAYS: "1.", "1.1"
 * </ul>
 *
 * @stable ICU 60
 */
typedef enum UNumberDecimalSeparatorDisplay {
    /**
     * Show the decimal separator when there are one or more digits to display after the separator, and do not show
     * it otherwise. This is the default behavior.
     *
     * @stable ICU 60
     */
            UNUM_DECIMAL_SEPARATOR_AUTO,

    /**
     * Always show the decimal separator, even if there are no digits to display after the separator.
     *
     * @stable ICU 60
     */
            UNUM_DECIMAL_SEPARATOR_ALWAYS,

    /**
     * One more than the highest UNumberDecimalSeparatorDisplay value.
     *
     * @internal ICU 60: The numeric value may change over time; see ICU ticket #12420.
     */
            UNUM_DECIMAL_SEPARATOR_COUNT
} UNumberDecimalSeparatorDisplay;

struct UNumberFormatter;
/**
 * C-compatible version of icu::number::LocalizedNumberFormatter.
 *
 * NOTE: This is a C-compatible API; C++ users should build against numberformatter.h instead.
 *
 * @stable ICU 62
 */
typedef struct UNumberFormatter UNumberFormatter;

struct UFormattedNumber;
/**
 * C-compatible version of icu::number::FormattedNumber.
 *
 * NOTE: This is a C-compatible API; C++ users should build against numberformatter.h instead.
 *
 * @stable ICU 62
 */
typedef struct UFormattedNumber UFormattedNumber;


/**
 * Creates a new UNumberFormatter for the given skeleton string and locale. This is currently the only
 * method for creating a new UNumberFormatter.
 *
 * Objects of type UNumberFormatter returned by this method are threadsafe.
 *
 * For more details on skeleton strings, see the documentation in numberformatter.h. For more details on
 * the usage of this API, see the documentation at the top of unumberformatter.h.
 *
 * NOTE: This is a C-compatible API; C++ users should build against numberformatter.h instead.
 *
 * @param skeleton The skeleton string, like u"percent precision-integer"
 * @param skeletonLen The number of UChars in the skeleton string, or -1 it it is NUL-terminated.
 * @param locale The NUL-terminated locale ID.
 * @param ec Set if an error occurs.
 * @stable ICU 62
 */
U_STABLE UNumberFormatter* U_EXPORT2
unumf_openForSkeletonAndLocale(const UChar* skeleton, int32_t skeletonLen, const char* locale,
                               UErrorCode* ec);


#ifndef U_HIDE_DRAFT_API
/**
 * Like unumf_openForSkeletonAndLocale, but accepts a UParseError, which will be populated with the
 * location of a skeleton syntax error if such a syntax error exists.
 *
 * @param skeleton The skeleton string, like u"percent precision-integer"
 * @param skeletonLen The number of UChars in the skeleton string, or -1 it it is NUL-terminated.
 * @param locale The NUL-terminated locale ID.
 * @param perror A parse error struct populated if an error occurs when parsing. Can be NULL.
 *               If no error occurs, perror->offset will be set to -1.
 * @param ec Set if an error occurs.
 * @draft ICU 64
 */
U_DRAFT UNumberFormatter* U_EXPORT2
unumf_openForSkeletonAndLocaleWithError(
       const UChar* skeleton, int32_t skeletonLen, const char* locale, UParseError* perror, UErrorCode* ec);
#endif  // U_HIDE_DRAFT_API


/**
 * Creates an object to hold the result of a UNumberFormatter
 * operation. The object can be used repeatedly; it is cleared whenever
 * passed to a format function.
 *
 * @param ec Set if an error occurs.
 * @stable ICU 62
 */
U_STABLE UFormattedNumber* U_EXPORT2
unumf_openResult(UErrorCode* ec);


/**
 * Uses a UNumberFormatter to format an integer to a UFormattedNumber. A string, field position, and other
 * information can be retrieved from the UFormattedNumber.
 *
 * The UNumberFormatter can be shared between threads. Each thread should have its own local
 * UFormattedNumber, however, for storing the result of the formatting operation.
 *
 * NOTE: This is a C-compatible API; C++ users should build against numberformatter.h instead.
 *
 * @param uformatter A formatter object created by unumf_openForSkeletonAndLocale or similar.
 * @param value The number to be formatted.
 * @param uresult The object that will be mutated to store the result; see unumf_openResult.
 * @param ec Set if an error occurs.
 * @stable ICU 62
 */
U_STABLE void U_EXPORT2
unumf_formatInt(const UNumberFormatter* uformatter, int64_t value, UFormattedNumber* uresult,
                UErrorCode* ec);


/**
 * Uses a UNumberFormatter to format a double to a UFormattedNumber. A string, field position, and other
 * information can be retrieved from the UFormattedNumber.
 *
 * The UNumberFormatter can be shared between threads. Each thread should have its own local
 * UFormattedNumber, however, for storing the result of the formatting operation.
 *
 * NOTE: This is a C-compatible API; C++ users should build against numberformatter.h instead.
 *
 * @param uformatter A formatter object created by unumf_openForSkeletonAndLocale or similar.
 * @param value The number to be formatted.
 * @param uresult The object that will be mutated to store the result; see unumf_openResult.
 * @param ec Set if an error occurs.
 * @stable ICU 62
 */
U_STABLE void U_EXPORT2
unumf_formatDouble(const UNumberFormatter* uformatter, double value, UFormattedNumber* uresult,
                   UErrorCode* ec);


/**
 * Uses a UNumberFormatter to format a decimal number to a UFormattedNumber. A string, field position, and
 * other information can be retrieved from the UFormattedNumber.
 *
 * The UNumberFormatter can be shared between threads. Each thread should have its own local
 * UFormattedNumber, however, for storing the result of the formatting operation.
 *
 * The syntax of the unformatted number is a "numeric string" as defined in the Decimal Arithmetic
 * Specification, available at http://speleotrove.com/decimal
 *
 * NOTE: This is a C-compatible API; C++ users should build against numberformatter.h instead.
 *
 * @param uformatter A formatter object created by unumf_openForSkeletonAndLocale or similar.
 * @param value The numeric string to be formatted.
 * @param valueLen The length of the numeric string, or -1 if it is NUL-terminated.
 * @param uresult The object that will be mutated to store the result; see unumf_openResult.
 * @param ec Set if an error occurs.
 * @stable ICU 62
 */
U_STABLE void U_EXPORT2
unumf_formatDecimal(const UNumberFormatter* uformatter, const char* value, int32_t valueLen,
                    UFormattedNumber* uresult, UErrorCode* ec);

#ifndef U_HIDE_DRAFT_API
/**
 * Returns a representation of a UFormattedNumber as a UFormattedValue,
 * which can be subsequently passed to any API requiring that type.
 *
 * The returned object is owned by the UFormattedNumber and is valid
 * only as long as the UFormattedNumber is present and unchanged in memory.
 *
 * You can think of this method as a cast between types.
 *
 * @param uresult The object containing the formatted string.
 * @param ec Set if an error occurs.
 * @return A UFormattedValue owned by the input object.
 * @draft ICU 64
 */
U_DRAFT const UFormattedValue* U_EXPORT2
unumf_resultAsValue(const UFormattedNumber* uresult, UErrorCode* ec);
#endif  /* U_HIDE_DRAFT_API */


/**
 * Extracts the result number string out of a UFormattedNumber to a UChar buffer if possible.
 * If bufferCapacity is greater than the required length, a terminating NUL is written.
 * If bufferCapacity is less than the required length, an error code is set.
 *
 * Also see ufmtval_getString, which returns a NUL-terminated string:
 *
 *     int32_t len;
 *     const UChar* str = ufmtval_getString(unumf_resultAsValue(uresult, &ec), &len, &ec);
 *
 * NOTE: This is a C-compatible API; C++ users should build against numberformatter.h instead.
 *
 * @param uresult The object containing the formatted number.
 * @param buffer Where to save the string output.
 * @param bufferCapacity The number of UChars available in the buffer.
 * @param ec Set if an error occurs.
 * @return The required length.
 * @stable ICU 62
 */
U_STABLE int32_t U_EXPORT2
unumf_resultToString(const UFormattedNumber* uresult, UChar* buffer, int32_t bufferCapacity,
                     UErrorCode* ec);


/**
 * Determines the start and end indices of the next occurrence of the given <em>field</em> in the
 * output string. This allows you to determine the locations of, for example, the integer part,
 * fraction part, or symbols.
 *
 * This is a simpler but less powerful alternative to {@link ufmtval_nextPosition}.
 *
 * If a field occurs just once, calling this method will find that occurrence and return it. If a
 * field occurs multiple times, this method may be called repeatedly with the following pattern:
 *
 * <pre>
 * UFieldPosition ufpos = {UNUM_GROUPING_SEPARATOR_FIELD, 0, 0};
 * while (unumf_resultNextFieldPosition(uresult, ufpos, &ec)) {
 *   // do something with ufpos.
 * }
 * </pre>
 *
 * This method is useful if you know which field to query. If you want all available field position
 * information, use unumf_resultGetAllFieldPositions().
 *
 * NOTE: All fields of the UFieldPosition must be initialized before calling this method.
 *
 * @param uresult The object containing the formatted number.
 * @param ufpos
 *            Input+output variable. On input, the "field" property determines which field to look up,
 *            and the "endIndex" property determines where to begin the search. On output, the
 *            "beginIndex" field is set to the beginning of the first occurrence of the field after the
 *            input "endIndex", and "endIndex" is set to the end of that occurrence of the field
 *            (exclusive index). If a field position is not found, the FieldPosition is not changed and
 *            the method returns FALSE.
 * @param ec Set if an error occurs.
 * @stable ICU 62
 */
U_STABLE UBool U_EXPORT2
unumf_resultNextFieldPosition(const UFormattedNumber* uresult, UFieldPosition* ufpos, UErrorCode* ec);


/**
 * Populates the given iterator with all fields in the formatted output string. This allows you to
 * determine the locations of the integer part, fraction part, and sign.
 *
 * This is an alternative to the more powerful {@link ufmtval_nextPosition} API.
 *
 * If you need information on only one field, use {@link ufmtval_nextPosition} or
 * {@link unumf_resultNextFieldPosition}.
 *
 * @param uresult The object containing the formatted number.
 * @param ufpositer
 *         A pointer to a UFieldPositionIterator created by {@link #ufieldpositer_open}. Iteration
 *         information already present in the UFieldPositionIterator is deleted, and the iterator is reset
 *         to apply to the fields in the formatted string created by this function call. The field values
 *         and indexes returned by {@link #ufieldpositer_next} represent fields denoted by
 *         the UNumberFormatFields enum. Fields are not returned in a guaranteed order. Fields cannot
 *         overlap, but they may nest. For example, 1234 could format as "1,234" which might consist of a
 *         grouping separator field for ',' and an integer field encompassing the entire string.
 * @param ec Set if an error occurs.
 * @stable ICU 62
 */
U_STABLE void U_EXPORT2
unumf_resultGetAllFieldPositions(const UFormattedNumber* uresult, UFieldPositionIterator* ufpositer,
                                 UErrorCode* ec);


// TODO(ICU-20775): Propose this as API.
// NOTE: This is not currently implemented.
// U_DRAFT int32_t U_EXPORT2
// unumf_resultToDecimalNumber(const UFormattedNumber* uresult, char* buffer, int32_t bufferCapacity,
//                             UErrorCode* ec);


/**
 * Releases the UNumberFormatter created by unumf_openForSkeletonAndLocale().
 *
 * @param uformatter An object created by unumf_openForSkeletonAndLocale().
 * @stable ICU 62
 */
U_STABLE void U_EXPORT2
unumf_close(UNumberFormatter* uformatter);


/**
 * Releases the UFormattedNumber created by unumf_openResult().
 *
 * @param uresult An object created by unumf_openResult().
 * @stable ICU 62
 */
U_STABLE void U_EXPORT2
unumf_closeResult(UFormattedNumber* uresult);


#if U_SHOW_CPLUSPLUS_API
U_NAMESPACE_BEGIN

/**
 * \class LocalUNumberFormatterPointer
 * "Smart pointer" class; closes a UNumberFormatter via unumf_close().
 * For most methods see the LocalPointerBase base class.
 *
 * Usage:
 * <pre>
 * LocalUNumberFormatterPointer uformatter(unumf_openForSkeletonAndLocale(...));
 * // no need to explicitly call unumf_close()
 * </pre>
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 62
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUNumberFormatterPointer, UNumberFormatter, unumf_close);

/**
 * \class LocalUFormattedNumberPointer
 * "Smart pointer" class; closes a UFormattedNumber via unumf_closeResult().
 * For most methods see the LocalPointerBase base class.
 *
 * Usage:
 * <pre>
 * LocalUFormattedNumberPointer uformatter(unumf_openResult(...));
 * // no need to explicitly call unumf_closeResult()
 * </pre>
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 62
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUFormattedNumberPointer, UFormattedNumber, unumf_closeResult);

U_NAMESPACE_END
#endif // U_SHOW_CPLUSPLUS_API

#endif //__UNUMBERFORMATTER_H__



/**
 * \file
 * \brief C++ API: Library for localized number formatting introduced in ICU 60.
 *
 * This library was introduced in ICU 60 to simplify the process of formatting localized number strings.
 * Basic usage examples:
 *
 * <pre>
 * // Most basic usage:
 * NumberFormatter::withLocale(...).format(123).toString();  // 1,234 in en-US
 *
 * // Custom notation, unit, and rounding precision:
 * NumberFormatter::with()
 *     .notation(Notation::compactShort())
 *     .unit(CurrencyUnit("EUR", status))
 *     .precision(Precision::maxDigits(2))
 *     .locale(...)
 *     .format(1234)
 *     .toString();  // €1.2K in en-US
 *
 * // Create a formatter in a singleton by value for use later:
 * static const LocalizedNumberFormatter formatter = NumberFormatter::withLocale(...)
 *     .unit(NoUnit::percent())
 *     .precision(Precision::fixedFraction(3));
 * formatter.format(5.9831).toString();  // 5.983% in en-US
 *
 * // Create a "template" in a singleton unique_ptr but without setting a locale until the call site:
 * std::unique_ptr<UnlocalizedNumberFormatter> template = NumberFormatter::with()
 *     .sign(UNumberSignDisplay::UNUM_SIGN_ALWAYS)
 *     .unit(MeasureUnit::getMeter())
 *     .unitWidth(UNumberUnitWidth::UNUM_UNIT_WIDTH_FULL_NAME)
 *     .clone();
 * template->locale(...).format(1234).toString();  // +1,234 meters in en-US
 * </pre>
 *
 * <p>
 * This API offers more features than DecimalFormat and is geared toward new users of ICU.
 *
 * <p>
 * NumberFormatter instances (i.e., LocalizedNumberFormatter and UnlocalizedNumberFormatter)
 * are immutable and thread safe. This means that invoking a configuration method has no
 * effect on the receiving instance; you must store and use the new number formatter instance it returns instead.
 *
 * <pre>
 * UnlocalizedNumberFormatter formatter = UnlocalizedNumberFormatter::with().notation(Notation::scientific());
 * formatter.precision(Precision.maxFraction(2)); // does nothing!
 * formatter.locale(Locale.getEnglish()).format(9.8765).toString(); // prints "9.8765E0", not "9.88E0"
 * </pre>
 *
 * <p>
 * This API is based on the <em>fluent</em> design pattern popularized by libraries such as Google's Guava. For
 * extensive details on the design of this API, read <a href="https://goo.gl/szi5VB">the design doc</a>.
 *
 * @author Shane Carr
 */

U_NAMESPACE_BEGIN

// Forward declarations:
class IFixedDecimal;
class FieldPositionIteratorHandler;
class FormattedStringBuilder;

namespace numparse {
namespace impl {

// Forward declarations:
class NumberParserImpl;
class MultiplierParseHandler;

}
}

namespace number {  // icu::number

// Forward declarations:
class UnlocalizedNumberFormatter;
class LocalizedNumberFormatter;
class FormattedNumber;
class Notation;
class ScientificNotation;
class Precision;
class FractionPrecision;
class CurrencyPrecision;
class IncrementPrecision;
class IntegerWidth;

namespace impl {

// can't be #ifndef U_HIDE_INTERNAL_API; referenced throughout this file in public classes
/**
 * Datatype for minimum/maximum fraction digits. Must be able to hold kMaxIntFracSig.
 *
 * @internal
 */
typedef int16_t digits_t;

// can't be #ifndef U_HIDE_INTERNAL_API; needed for struct initialization
/**
 * Use a default threshold of 3. This means that the third time .format() is called, the data structures get built
 * using the "safe" code path. The first two calls to .format() will trigger the unsafe code path.
 *
 * @internal
 */
static constexpr int32_t kInternalDefaultThreshold = 3;

// Forward declarations:
class Padder;
struct MacroProps;
struct MicroProps;
class DecimalQuantity;
class UFormattedNumberData;
class NumberFormatterImpl;
struct ParsedPatternInfo;
class ScientificModifier;
class MultiplierProducer;
class RoundingImpl;
class ScientificHandler;
class Modifier;
class AffixPatternProvider;
class NumberPropertyMapper;
struct DecimalFormatProperties;
class MultiplierFormatHandler;
class CurrencySymbols;
class GeneratorHelpers;
class DecNum;
class NumberRangeFormatterImpl;
struct RangeMacroProps;
struct UFormattedNumberImpl;

/**
 * Used for NumberRangeFormatter and implemented in numrange_fluent.cpp.
 * Declared here so it can be friended.
 *
 * @internal
 */
void touchRangeLocales(impl::RangeMacroProps& macros);

} // namespace impl

/**
 * Extra name reserved in case it is needed in the future.
 *
 * @stable ICU 63
 */
typedef Notation CompactNotation;

/**
 * Extra name reserved in case it is needed in the future.
 *
 * @stable ICU 63
 */
typedef Notation SimpleNotation;

/**
 * A class that defines the notation style to be used when formatting numbers in NumberFormatter.
 *
 * @stable ICU 60
 */
class U_I18N_API Notation : public UMemory {
  public:
    /**
     * Print the number using scientific notation (also known as scientific form, standard index form, or standard form
     * in the UK). The format for scientific notation varies by locale; for example, many Western locales display the
     * number in the form "#E0", where the number is displayed with one digit before the decimal separator, zero or more
     * digits after the decimal separator, and the corresponding power of 10 displayed after the "E".
     *
     * <p>
     * Example outputs in <em>en-US</em> when printing 8.765E4 through 8.765E-3:
     *
     * <pre>
     * 8.765E4
     * 8.765E3
     * 8.765E2
     * 8.765E1
     * 8.765E0
     * 8.765E-1
     * 8.765E-2
     * 8.765E-3
     * 0E0
     * </pre>
     *
     * @return A ScientificNotation for chaining or passing to the NumberFormatter notation() setter.
     * @stable ICU 60
     */
    static ScientificNotation scientific();

    /**
     * Print the number using engineering notation, a variant of scientific notation in which the exponent must be
     * divisible by 3.
     *
     * <p>
     * Example outputs in <em>en-US</em> when printing 8.765E4 through 8.765E-3:
     *
     * <pre>
     * 87.65E3
     * 8.765E3
     * 876.5E0
     * 87.65E0
     * 8.765E0
     * 876.5E-3
     * 87.65E-3
     * 8.765E-3
     * 0E0
     * </pre>
     *
     * @return A ScientificNotation for chaining or passing to the NumberFormatter notation() setter.
     * @stable ICU 60
     */
    static ScientificNotation engineering();

    /**
     * Print the number using short-form compact notation.
     *
     * <p>
     * <em>Compact notation</em>, defined in Unicode Technical Standard #35 Part 3 Section 2.4.1, prints numbers with
     * localized prefixes or suffixes corresponding to different powers of ten. Compact notation is similar to
     * engineering notation in how it scales numbers.
     *
     * <p>
     * Compact notation is ideal for displaying large numbers (over ~1000) to humans while at the same time minimizing
     * screen real estate.
     *
     * <p>
     * In short form, the powers of ten are abbreviated. In <em>en-US</em>, the abbreviations are "K" for thousands, "M"
     * for millions, "B" for billions, and "T" for trillions. Example outputs in <em>en-US</em> when printing 8.765E7
     * through 8.765E0:
     *
     * <pre>
     * 88M
     * 8.8M
     * 876K
     * 88K
     * 8.8K
     * 876
     * 88
     * 8.8
     * </pre>
     *
     * <p>
     * When compact notation is specified without an explicit rounding precision, numbers are rounded off to the closest
     * integer after scaling the number by the corresponding power of 10, but with a digit shown after the decimal
     * separator if there is only one digit before the decimal separator. The default compact notation rounding precision
     * is equivalent to:
     *
     * <pre>
     * Precision::integer().withMinDigits(2)
     * </pre>
     *
     * @return A CompactNotation for passing to the NumberFormatter notation() setter.
     * @stable ICU 60
     */
    static CompactNotation compactShort();

    /**
     * Print the number using long-form compact notation. For more information on compact notation, see
     * {@link #compactShort}.
     *
     * <p>
     * In long form, the powers of ten are spelled out fully. Example outputs in <em>en-US</em> when printing 8.765E7
     * through 8.765E0:
     *
     * <pre>
     * 88 million
     * 8.8 million
     * 876 thousand
     * 88 thousand
     * 8.8 thousand
     * 876
     * 88
     * 8.8
     * </pre>
     *
     * @return A CompactNotation for passing to the NumberFormatter notation() setter.
     * @stable ICU 60
     */
    static CompactNotation compactLong();

    /**
     * Print the number using simple notation without any scaling by powers of ten. This is the default behavior.
     *
     * <p>
     * Since this is the default behavior, this method needs to be called only when it is necessary to override a
     * previous setting.
     *
     * <p>
     * Example outputs in <em>en-US</em> when printing 8.765E7 through 8.765E0:
     *
     * <pre>
     * 87,650,000
     * 8,765,000
     * 876,500
     * 87,650
     * 8,765
     * 876.5
     * 87.65
     * 8.765
     * </pre>
     *
     * @return A SimpleNotation for passing to the NumberFormatter notation() setter.
     * @stable ICU 60
     */
    static SimpleNotation simple();

  private:
    enum NotationType {
        NTN_SCIENTIFIC, NTN_COMPACT, NTN_SIMPLE, NTN_ERROR
    } fType;

    union NotationUnion {
        // For NTN_SCIENTIFIC
        /** @internal */
        struct ScientificSettings {
            /** @internal */
            int8_t fEngineeringInterval;
            /** @internal */
            bool fRequireMinInt;
            /** @internal */
            number::impl::digits_t fMinExponentDigits;
            /** @internal */
            UNumberSignDisplay fExponentSignDisplay;
        } scientific;

        // For NTN_COMPACT
        UNumberCompactStyle compactStyle;

        // For NTN_ERROR
        UErrorCode errorCode;
    } fUnion;

    typedef NotationUnion::ScientificSettings ScientificSettings;

    Notation(const NotationType &type, const NotationUnion &union_) : fType(type), fUnion(union_) {}

    Notation(UErrorCode errorCode) : fType(NTN_ERROR) {
        fUnion.errorCode = errorCode;
    }

    Notation() : fType(NTN_SIMPLE), fUnion() {}

    UBool copyErrorTo(UErrorCode &status) const {
        if (fType == NTN_ERROR) {
            status = fUnion.errorCode;
            return TRUE;
        }
        return FALSE;
    }

    // To allow MacroProps to initialize empty instances:
    friend struct number::impl::MacroProps;
    friend class ScientificNotation;

    // To allow implementation to access internal types:
    friend class number::impl::NumberFormatterImpl;
    friend class number::impl::ScientificModifier;
    friend class number::impl::ScientificHandler;

    // To allow access to the skeleton generation code:
    friend class number::impl::GeneratorHelpers;
};

/**
 * A class that defines the scientific notation style to be used when formatting numbers in NumberFormatter.
 *
 * <p>
 * To create a ScientificNotation, use one of the factory methods in {@link Notation}.
 *
 * @stable ICU 60
 */
class U_I18N_API ScientificNotation : public Notation {
  public:
    /**
     * Sets the minimum number of digits to show in the exponent of scientific notation, padding with zeros if
     * necessary. Useful for fixed-width display.
     *
     * <p>
     * For example, with minExponentDigits=2, the number 123 will be printed as "1.23E02" in <em>en-US</em> instead of
     * the default "1.23E2".
     *
     * @param minExponentDigits
     *            The minimum number of digits to show in the exponent.
     * @return A ScientificNotation, for chaining.
     * @stable ICU 60
     */
    ScientificNotation withMinExponentDigits(int32_t minExponentDigits) const;

    /**
     * Sets whether to show the sign on positive and negative exponents in scientific notation. The default is AUTO,
     * showing the minus sign but not the plus sign.
     *
     * <p>
     * For example, with exponentSignDisplay=ALWAYS, the number 123 will be printed as "1.23E+2" in <em>en-US</em>
     * instead of the default "1.23E2".
     *
     * @param exponentSignDisplay
     *            The strategy for displaying the sign in the exponent.
     * @return A ScientificNotation, for chaining.
     * @stable ICU 60
     */
    ScientificNotation withExponentSignDisplay(UNumberSignDisplay exponentSignDisplay) const;

  private:
    // Inherit constructor
    using Notation::Notation;

    // Raw constructor for NumberPropertyMapper
    ScientificNotation(int8_t fEngineeringInterval, bool fRequireMinInt, number::impl::digits_t fMinExponentDigits,
                       UNumberSignDisplay fExponentSignDisplay);

    friend class Notation;

    // So that NumberPropertyMapper can create instances
    friend class number::impl::NumberPropertyMapper;
};

/**
 * Extra name reserved in case it is needed in the future.
 *
 * @stable ICU 63
 */
typedef Precision SignificantDigitsPrecision;

/**
 * A class that defines the rounding precision to be used when formatting numbers in NumberFormatter.
 *
 * <p>
 * To create a Precision, use one of the factory methods.
 *
 * @stable ICU 60
 */
class U_I18N_API Precision : public UMemory {

  public:
    /**
     * Show all available digits to full precision.
     *
     * <p>
     * <strong>NOTE:</strong> When formatting a <em>double</em>, this method, along with {@link #minFraction} and
     * {@link #minSignificantDigits}, will trigger complex algorithm similar to <em>Dragon4</em> to determine the
     * low-order digits and the number of digits to display based on the value of the double.
     * If the number of fraction places or significant digits can be bounded, consider using {@link #maxFraction}
     * or {@link #maxSignificantDigits} instead to maximize performance.
     * For more information, read the following blog post.
     *
     * <p>
     * http://www.serpentine.com/blog/2011/06/29/here-be-dragons-advances-in-problems-you-didnt-even-know-you-had/
     *
     * @return A Precision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    static Precision unlimited();

    /**
     * Show numbers rounded if necessary to the nearest integer.
     *
     * @return A FractionPrecision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    static FractionPrecision integer();

    /**
     * Show numbers rounded if necessary to a certain number of fraction places (numerals after the decimal separator).
     * Additionally, pad with zeros to ensure that this number of places are always shown.
     *
     * <p>
     * Example output with minMaxFractionPlaces = 3:
     *
     * <p>
     * 87,650.000<br>
     * 8,765.000<br>
     * 876.500<br>
     * 87.650<br>
     * 8.765<br>
     * 0.876<br>
     * 0.088<br>
     * 0.009<br>
     * 0.000 (zero)
     *
     * <p>
     * This method is equivalent to {@link #minMaxFraction} with both arguments equal.
     *
     * @param minMaxFractionPlaces
     *            The minimum and maximum number of numerals to display after the decimal separator (rounding if too
     *            long or padding with zeros if too short).
     * @return A FractionPrecision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    static FractionPrecision fixedFraction(int32_t minMaxFractionPlaces);

    /**
     * Always show at least a certain number of fraction places after the decimal separator, padding with zeros if
     * necessary. Do not perform rounding (display numbers to their full precision).
     *
     * <p>
     * <strong>NOTE:</strong> If you are formatting <em>doubles</em>, see the performance note in {@link #unlimited}.
     *
     * @param minFractionPlaces
     *            The minimum number of numerals to display after the decimal separator (padding with zeros if
     *            necessary).
     * @return A FractionPrecision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    static FractionPrecision minFraction(int32_t minFractionPlaces);

    /**
     * Show numbers rounded if necessary to a certain number of fraction places (numerals after the decimal separator).
     * Unlike the other fraction rounding strategies, this strategy does <em>not</em> pad zeros to the end of the
     * number.
     *
     * @param maxFractionPlaces
     *            The maximum number of numerals to display after the decimal mark (rounding if necessary).
     * @return A FractionPrecision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    static FractionPrecision maxFraction(int32_t maxFractionPlaces);

    /**
     * Show numbers rounded if necessary to a certain number of fraction places (numerals after the decimal separator);
     * in addition, always show at least a certain number of places after the decimal separator, padding with zeros if
     * necessary.
     *
     * @param minFractionPlaces
     *            The minimum number of numerals to display after the decimal separator (padding with zeros if
     *            necessary).
     * @param maxFractionPlaces
     *            The maximum number of numerals to display after the decimal separator (rounding if necessary).
     * @return A FractionPrecision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    static FractionPrecision minMaxFraction(int32_t minFractionPlaces, int32_t maxFractionPlaces);

    /**
     * Show numbers rounded if necessary to a certain number of significant digits or significant figures. Additionally,
     * pad with zeros to ensure that this number of significant digits/figures are always shown.
     *
     * <p>
     * This method is equivalent to {@link #minMaxSignificantDigits} with both arguments equal.
     *
     * @param minMaxSignificantDigits
     *            The minimum and maximum number of significant digits to display (rounding if too long or padding with
     *            zeros if too short).
     * @return A precision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 62
     */
    static SignificantDigitsPrecision fixedSignificantDigits(int32_t minMaxSignificantDigits);

    /**
     * Always show at least a certain number of significant digits/figures, padding with zeros if necessary. Do not
     * perform rounding (display numbers to their full precision).
     *
     * <p>
     * <strong>NOTE:</strong> If you are formatting <em>doubles</em>, see the performance note in {@link #unlimited}.
     *
     * @param minSignificantDigits
     *            The minimum number of significant digits to display (padding with zeros if too short).
     * @return A precision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 62
     */
    static SignificantDigitsPrecision minSignificantDigits(int32_t minSignificantDigits);

    /**
     * Show numbers rounded if necessary to a certain number of significant digits/figures.
     *
     * @param maxSignificantDigits
     *            The maximum number of significant digits to display (rounding if too long).
     * @return A precision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 62
     */
    static SignificantDigitsPrecision maxSignificantDigits(int32_t maxSignificantDigits);

    /**
     * Show numbers rounded if necessary to a certain number of significant digits/figures; in addition, always show at
     * least a certain number of significant digits, padding with zeros if necessary.
     *
     * @param minSignificantDigits
     *            The minimum number of significant digits to display (padding with zeros if necessary).
     * @param maxSignificantDigits
     *            The maximum number of significant digits to display (rounding if necessary).
     * @return A precision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 62
     */
    static SignificantDigitsPrecision minMaxSignificantDigits(int32_t minSignificantDigits,
                                                              int32_t maxSignificantDigits);

    /**
     * Show numbers rounded if necessary to the closest multiple of a certain rounding increment. For example, if the
     * rounding increment is 0.5, then round 1.2 to 1 and round 1.3 to 1.5.
     *
     * <p>
     * In order to ensure that numbers are padded to the appropriate number of fraction places, call
     * withMinFraction() on the return value of this method.
     * For example, to round to the nearest 0.5 and always display 2 numerals after the
     * decimal separator (to display 1.2 as "1.00" and 1.3 as "1.50"), you can run:
     *
     * <pre>
     * Precision::increment(0.5).withMinFraction(2)
     * </pre>
     *
     * @param roundingIncrement
     *            The increment to which to round numbers.
     * @return A precision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    static IncrementPrecision increment(double roundingIncrement);

    /**
     * Show numbers rounded and padded according to the rules for the currency unit. The most common
     * rounding precision settings for currencies include <code>Precision::fixedFraction(2)</code>,
     * <code>Precision::integer()</code>, and <code>Precision::increment(0.05)</code> for cash transactions
     * ("nickel rounding").
     *
     * <p>
     * The exact rounding details will be resolved at runtime based on the currency unit specified in the
     * NumberFormatter chain. To round according to the rules for one currency while displaying the symbol for another
     * currency, the withCurrency() method can be called on the return value of this method.
     *
     * @param currencyUsage
     *            Either STANDARD (for digital transactions) or CASH (for transactions where the rounding increment may
     *            be limited by the available denominations of cash or coins).
     * @return A CurrencyPrecision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    static CurrencyPrecision currency(UCurrencyUsage currencyUsage);

  private:
    enum PrecisionType {
        RND_BOGUS,
        RND_NONE,
        RND_FRACTION,
        RND_SIGNIFICANT,
        RND_FRACTION_SIGNIFICANT,

        // Used for strange increments like 3.14.
        RND_INCREMENT,

        // Used for increments with 1 as the only digit. This is different than fraction
        // rounding because it supports having additional trailing zeros. For example, this
        // class is used to round with the increment 0.010.
        RND_INCREMENT_ONE,

        // Used for increments with 5 as the only digit (nickel rounding).
        RND_INCREMENT_FIVE,

        RND_CURRENCY,
        RND_ERROR
    } fType;

    union PrecisionUnion {
        /** @internal */
        struct FractionSignificantSettings {
            // For RND_FRACTION, RND_SIGNIFICANT, and RND_FRACTION_SIGNIFICANT
            /** @internal */
            number::impl::digits_t fMinFrac;
            /** @internal */
            number::impl::digits_t fMaxFrac;
            /** @internal */
            number::impl::digits_t fMinSig;
            /** @internal */
            number::impl::digits_t fMaxSig;
        } fracSig;
        /** @internal */
        struct IncrementSettings {
            // For RND_INCREMENT, RND_INCREMENT_ONE, and RND_INCREMENT_FIVE
            /** @internal */
            double fIncrement;
            /** @internal */
            number::impl::digits_t fMinFrac;
            /** @internal */
            number::impl::digits_t fMaxFrac;
        } increment;
        UCurrencyUsage currencyUsage; // For RND_CURRENCY
        UErrorCode errorCode; // For RND_ERROR
    } fUnion;

    typedef PrecisionUnion::FractionSignificantSettings FractionSignificantSettings;
    typedef PrecisionUnion::IncrementSettings IncrementSettings;

    /** The Precision encapsulates the RoundingMode when used within the implementation. */
    UNumberFormatRoundingMode fRoundingMode;

    Precision(const PrecisionType& type, const PrecisionUnion& union_,
              UNumberFormatRoundingMode roundingMode)
            : fType(type), fUnion(union_), fRoundingMode(roundingMode) {}

    Precision(UErrorCode errorCode) : fType(RND_ERROR) {
        fUnion.errorCode = errorCode;
    }

    Precision() : fType(RND_BOGUS) {}

    bool isBogus() const {
        return fType == RND_BOGUS;
    }

    UBool copyErrorTo(UErrorCode &status) const {
        if (fType == RND_ERROR) {
            status = fUnion.errorCode;
            return TRUE;
        }
        return FALSE;
    }

    // On the parent type so that this method can be called internally on Precision instances.
    Precision withCurrency(const CurrencyUnit &currency, UErrorCode &status) const;

    static FractionPrecision constructFraction(int32_t minFrac, int32_t maxFrac);

    static Precision constructSignificant(int32_t minSig, int32_t maxSig);

    static Precision
    constructFractionSignificant(const FractionPrecision &base, int32_t minSig, int32_t maxSig);

    static IncrementPrecision constructIncrement(double increment, int32_t minFrac);

    static CurrencyPrecision constructCurrency(UCurrencyUsage usage);

    static Precision constructPassThrough();

    // To allow MacroProps/MicroProps to initialize bogus instances:
    friend struct number::impl::MacroProps;
    friend struct number::impl::MicroProps;

    // To allow NumberFormatterImpl to access isBogus() and other internal methods:
    friend class number::impl::NumberFormatterImpl;

    // To allow NumberPropertyMapper to create instances from DecimalFormatProperties:
    friend class number::impl::NumberPropertyMapper;

    // To allow access to the main implementation class:
    friend class number::impl::RoundingImpl;

    // To allow child classes to call private methods:
    friend class FractionPrecision;
    friend class CurrencyPrecision;
    friend class IncrementPrecision;

    // To allow access to the skeleton generation code:
    friend class number::impl::GeneratorHelpers;
};

/**
 * A class that defines a rounding precision based on a number of fraction places and optionally significant digits to be
 * used when formatting numbers in NumberFormatter.
 *
 * <p>
 * To create a FractionPrecision, use one of the factory methods on Precision.
 *
 * @stable ICU 60
 */
class U_I18N_API FractionPrecision : public Precision {
  public:
    /**
     * Ensure that no less than this number of significant digits are retained when rounding according to fraction
     * rules.
     *
     * <p>
     * For example, with integer rounding, the number 3.141 becomes "3". However, with minimum figures set to 2, 3.141
     * becomes "3.1" instead.
     *
     * <p>
     * This setting does not affect the number of trailing zeros. For example, 3.01 would print as "3", not "3.0".
     *
     * @param minSignificantDigits
     *            The number of significant figures to guarantee.
     * @return A precision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    Precision withMinDigits(int32_t minSignificantDigits) const;

    /**
     * Ensure that no more than this number of significant digits are retained when rounding according to fraction
     * rules.
     *
     * <p>
     * For example, with integer rounding, the number 123.4 becomes "123". However, with maximum figures set to 2, 123.4
     * becomes "120" instead.
     *
     * <p>
     * This setting does not affect the number of trailing zeros. For example, with fixed fraction of 2, 123.4 would
     * become "120.00".
     *
     * @param maxSignificantDigits
     *            Round the number to no more than this number of significant figures.
     * @return A precision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    Precision withMaxDigits(int32_t maxSignificantDigits) const;

  private:
    // Inherit constructor
    using Precision::Precision;

    // To allow parent class to call this class's constructor:
    friend class Precision;
};

/**
 * A class that defines a rounding precision parameterized by a currency to be used when formatting numbers in
 * NumberFormatter.
 *
 * <p>
 * To create a CurrencyPrecision, use one of the factory methods on Precision.
 *
 * @stable ICU 60
 */
class U_I18N_API CurrencyPrecision : public Precision {
  public:
    /**
      * Associates a currency with this rounding precision.
      *
      * <p>
      * <strong>Calling this method is <em>not required</em></strong>, because the currency specified in unit()
      * is automatically applied to currency rounding precisions. However,
      * this method enables you to override that automatic association.
      *
      * <p>
      * This method also enables numbers to be formatted using currency rounding rules without explicitly using a
      * currency format.
      *
      * @param currency
      *            The currency to associate with this rounding precision.
      * @return A precision for chaining or passing to the NumberFormatter precision() setter.
      * @stable ICU 60
      */
    Precision withCurrency(const CurrencyUnit &currency) const;

  private:
    // Inherit constructor
    using Precision::Precision;

    // To allow parent class to call this class's constructor:
    friend class Precision;
};

/**
 * A class that defines a rounding precision parameterized by a rounding increment to be used when formatting numbers in
 * NumberFormatter.
 *
 * <p>
 * To create an IncrementPrecision, use one of the factory methods on Precision.
 *
 * @stable ICU 60
 */
class U_I18N_API IncrementPrecision : public Precision {
  public:
    /**
     * Specifies the minimum number of fraction digits to render after the decimal separator, padding with zeros if
     * necessary.  By default, no trailing zeros are added.
     *
     * <p>
     * For example, if the rounding increment is 0.5 and minFrac is 2, then the resulting strings include "0.00",
     * "0.50", "1.00", and "1.50".
     *
     * <p>
     * Note: In ICU4J, this functionality is accomplished via the scale of the BigDecimal rounding increment.
     *
     * @param minFrac The minimum number of digits after the decimal separator.
     * @return A precision for chaining or passing to the NumberFormatter precision() setter.
     * @stable ICU 60
     */
    Precision withMinFraction(int32_t minFrac) const;

  private:
    // Inherit constructor
    using Precision::Precision;

    // To allow parent class to call this class's constructor:
    friend class Precision;
};

/**
 * A class that defines the strategy for padding and truncating integers before the decimal separator.
 *
 * <p>
 * To create an IntegerWidth, use one of the factory methods.
 *
 * @stable ICU 60
 * @see NumberFormatter
 */
class U_I18N_API IntegerWidth : public UMemory {
  public:
    /**
     * Pad numbers at the beginning with zeros to guarantee a certain number of numerals before the decimal separator.
     *
     * <p>
     * For example, with minInt=3, the number 55 will get printed as "055".
     *
     * @param minInt
     *            The minimum number of places before the decimal separator.
     * @return An IntegerWidth for chaining or passing to the NumberFormatter integerWidth() setter.
     * @stable ICU 60
     */
    static IntegerWidth zeroFillTo(int32_t minInt);

    /**
     * Truncate numbers exceeding a certain number of numerals before the decimal separator.
     *
     * For example, with maxInt=3, the number 1234 will get printed as "234".
     *
     * @param maxInt
     *            The maximum number of places before the decimal separator. maxInt == -1 means no
     *            truncation.
     * @return An IntegerWidth for passing to the NumberFormatter integerWidth() setter.
     * @stable ICU 60
     */
    IntegerWidth truncateAt(int32_t maxInt);

  private:
    union {
        struct {
            number::impl::digits_t fMinInt;
            number::impl::digits_t fMaxInt;
            bool fFormatFailIfMoreThanMaxDigits;
        } minMaxInt;
        UErrorCode errorCode;
    } fUnion;
    bool fHasError = false;

    IntegerWidth(impl::digits_t minInt, number::impl::digits_t maxInt, bool formatFailIfMoreThanMaxDigits);

    IntegerWidth(UErrorCode errorCode) { // NOLINT
        fUnion.errorCode = errorCode;
        fHasError = true;
    }

    IntegerWidth() { // NOLINT
        fUnion.minMaxInt.fMinInt = -1;
    }

    /** Returns the default instance. */
    static IntegerWidth standard() {
        return IntegerWidth::zeroFillTo(1);
    }

    bool isBogus() const {
        return !fHasError && fUnion.minMaxInt.fMinInt == -1;
    }

    UBool copyErrorTo(UErrorCode &status) const {
        if (fHasError) {
            status = fUnion.errorCode;
            return TRUE;
        }
        return FALSE;
    }

    void apply(number::impl::DecimalQuantity &quantity, UErrorCode &status) const;

    bool operator==(const IntegerWidth& other) const;

    // To allow MacroProps/MicroProps to initialize empty instances:
    friend struct number::impl::MacroProps;
    friend struct number::impl::MicroProps;

    // To allow NumberFormatterImpl to access isBogus() and perform other operations:
    friend class number::impl::NumberFormatterImpl;

    // So that NumberPropertyMapper can create instances
    friend class number::impl::NumberPropertyMapper;

    // To allow access to the skeleton generation code:
    friend class number::impl::GeneratorHelpers;
};

/**
 * A class that defines a quantity by which a number should be multiplied when formatting.
 *
 * <p>
 * To create a Scale, use one of the factory methods.
 *
 * @stable ICU 62
 */
class U_I18N_API Scale : public UMemory {
  public:
    /**
     * Do not change the value of numbers when formatting or parsing.
     *
     * @return A Scale to prevent any multiplication.
     * @stable ICU 62
     */
    static Scale none();

    /**
     * Multiply numbers by a power of ten before formatting. Useful for combining with a percent unit:
     *
     * <pre>
     * NumberFormatter::with().unit(NoUnit::percent()).multiplier(Scale::powerOfTen(2))
     * </pre>
     *
     * @return A Scale for passing to the setter in NumberFormatter.
     * @stable ICU 62
     */
    static Scale powerOfTen(int32_t power);

    /**
     * Multiply numbers by an arbitrary value before formatting. Useful for unit conversions.
     *
     * This method takes a string in a decimal number format with syntax
     * as defined in the Decimal Arithmetic Specification, available at
     * http://speleotrove.com/decimal
     *
     * Also see the version of this method that takes a double.
     *
     * @return A Scale for passing to the setter in NumberFormatter.
     * @stable ICU 62
     */
    static Scale byDecimal(StringPiece multiplicand);

    /**
     * Multiply numbers by an arbitrary value before formatting. Useful for unit conversions.
     *
     * This method takes a double; also see the version of this method that takes an exact decimal.
     *
     * @return A Scale for passing to the setter in NumberFormatter.
     * @stable ICU 62
     */
    static Scale byDouble(double multiplicand);

    /**
     * Multiply a number by both a power of ten and by an arbitrary double value.
     *
     * @return A Scale for passing to the setter in NumberFormatter.
     * @stable ICU 62
     */
    static Scale byDoubleAndPowerOfTen(double multiplicand, int32_t power);

    // We need a custom destructor for the DecNum, which means we need to declare
    // the copy/move constructor/assignment quartet.

    /** @stable ICU 62 */
    Scale(const Scale& other);

    /** @stable ICU 62 */
    Scale& operator=(const Scale& other);

    /** @stable ICU 62 */
    Scale(Scale&& src) U_NOEXCEPT;

    /** @stable ICU 62 */
    Scale& operator=(Scale&& src) U_NOEXCEPT;

    /** @stable ICU 62 */
    ~Scale();

#ifndef U_HIDE_INTERNAL_API
    /** @internal */
    Scale(int32_t magnitude, number::impl::DecNum* arbitraryToAdopt);
#endif  /* U_HIDE_INTERNAL_API */

  private:
    int32_t fMagnitude;
    number::impl::DecNum* fArbitrary;
    UErrorCode fError;

    Scale(UErrorCode error) : fMagnitude(0), fArbitrary(nullptr), fError(error) {}

    Scale() : fMagnitude(0), fArbitrary(nullptr), fError(U_ZERO_ERROR) {}

    bool isValid() const {
        return fMagnitude != 0 || fArbitrary != nullptr;
    }

    UBool copyErrorTo(UErrorCode &status) const {
        if (fError != U_ZERO_ERROR) {
            status = fError;
            return TRUE;
        }
        return FALSE;
    }

    void applyTo(impl::DecimalQuantity& quantity) const;

    void applyReciprocalTo(impl::DecimalQuantity& quantity) const;

    // To allow MacroProps/MicroProps to initialize empty instances:
    friend struct number::impl::MacroProps;
    friend struct number::impl::MicroProps;

    // To allow NumberFormatterImpl to access isBogus() and perform other operations:
    friend class number::impl::NumberFormatterImpl;

    // To allow the helper class MultiplierFormatHandler access to private fields:
    friend class number::impl::MultiplierFormatHandler;

    // To allow access to the skeleton generation code:
    friend class number::impl::GeneratorHelpers;

    // To allow access to parsing code:
    friend class ::icu::numparse::impl::NumberParserImpl;
    friend class ::icu::numparse::impl::MultiplierParseHandler;
};

namespace impl {

// Do not enclose entire SymbolsWrapper with #ifndef U_HIDE_INTERNAL_API, needed for a protected field
/** @internal */
class U_I18N_API SymbolsWrapper : public UMemory {
  public:
    /** @internal */
    SymbolsWrapper() : fType(SYMPTR_NONE), fPtr{nullptr} {}

    /** @internal */
    SymbolsWrapper(const SymbolsWrapper &other);

    /** @internal */
    SymbolsWrapper &operator=(const SymbolsWrapper &other);

    /** @internal */
    SymbolsWrapper(SymbolsWrapper&& src) U_NOEXCEPT;

    /** @internal */
    SymbolsWrapper &operator=(SymbolsWrapper&& src) U_NOEXCEPT;

    /** @internal */
    ~SymbolsWrapper();

#ifndef U_HIDE_INTERNAL_API

    /**
     * The provided object is copied, but we do not adopt it.
     * @internal
     */
    void setTo(const DecimalFormatSymbols &dfs);

    /**
     * Adopt the provided object.
     * @internal
     */
    void setTo(const NumberingSystem *ns);

    /**
     * Whether the object is currently holding a DecimalFormatSymbols.
     * @internal
     */
    bool isDecimalFormatSymbols() const;

    /**
     * Whether the object is currently holding a NumberingSystem.
     * @internal
     */
    bool isNumberingSystem() const;

    /**
     * Get the DecimalFormatSymbols pointer. No ownership change.
     * @internal
     */
    const DecimalFormatSymbols *getDecimalFormatSymbols() const;

    /**
     * Get the NumberingSystem pointer. No ownership change.
     * @internal
     */
    const NumberingSystem *getNumberingSystem() const;

#endif  // U_HIDE_INTERNAL_API

    /** @internal */
    UBool copyErrorTo(UErrorCode &status) const {
        if (fType == SYMPTR_DFS && fPtr.dfs == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return TRUE;
        } else if (fType == SYMPTR_NS && fPtr.ns == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return TRUE;
        }
        return FALSE;
    }

  private:
    enum SymbolsPointerType {
        SYMPTR_NONE, SYMPTR_DFS, SYMPTR_NS
    } fType;

    union {
        const DecimalFormatSymbols *dfs;
        const NumberingSystem *ns;
    } fPtr;

    void doCopyFrom(const SymbolsWrapper &other);

    void doMoveFrom(SymbolsWrapper&& src);

    void doCleanup();
};

// Do not enclose entire Grouper with #ifndef U_HIDE_INTERNAL_API, needed for a protected field
/** @internal */
class U_I18N_API Grouper : public UMemory {
  public:
#ifndef U_HIDE_INTERNAL_API
    /** @internal */
    static Grouper forStrategy(UNumberGroupingStrategy grouping);

    /**
     * Resolve the values in Properties to a Grouper object.
     * @internal
     */
    static Grouper forProperties(const DecimalFormatProperties& properties);

    // Future: static Grouper forProperties(DecimalFormatProperties& properties);

    /** @internal */
    Grouper(int16_t grouping1, int16_t grouping2, int16_t minGrouping, UNumberGroupingStrategy strategy)
            : fGrouping1(grouping1),
              fGrouping2(grouping2),
              fMinGrouping(minGrouping),
              fStrategy(strategy) {}
#endif  // U_HIDE_INTERNAL_API

    /** @internal */
    int16_t getPrimary() const;

    /** @internal */
    int16_t getSecondary() const;

  private:
    /**
     * The grouping sizes, with the following special values:
     * <ul>
     * <li>-1 = no grouping
     * <li>-2 = needs locale data
     * <li>-4 = fall back to Western grouping if not in locale
     * </ul>
     */
    int16_t fGrouping1;
    int16_t fGrouping2;

    /**
     * The minimum grouping size, with the following special values:
     * <ul>
     * <li>-2 = needs locale data
     * <li>-3 = no less than 2
     * </ul>
     */
    int16_t fMinGrouping;

    /**
     * The UNumberGroupingStrategy that was used to create this Grouper, or UNUM_GROUPING_COUNT if this
     * was not created from a UNumberGroupingStrategy.
     */
    UNumberGroupingStrategy fStrategy;

    Grouper() : fGrouping1(-3) {}

    bool isBogus() const {
        return fGrouping1 == -3;
    }

    /** NON-CONST: mutates the current instance. */
    void setLocaleData(const number::impl::ParsedPatternInfo &patternInfo, const Locale& locale);

    bool groupAtPosition(int32_t position, const number::impl::DecimalQuantity &value) const;

    // To allow MacroProps/MicroProps to initialize empty instances:
    friend struct MacroProps;
    friend struct MicroProps;

    // To allow NumberFormatterImpl to access isBogus() and perform other operations:
    friend class NumberFormatterImpl;

    // To allow NumberParserImpl to perform setLocaleData():
    friend class ::icu::numparse::impl::NumberParserImpl;

    // To allow access to the skeleton generation code:
    friend class number::impl::GeneratorHelpers;
};

// Do not enclose entire Padder with #ifndef U_HIDE_INTERNAL_API, needed for a protected field
/** @internal */
class U_I18N_API Padder : public UMemory {
  public:
#ifndef U_HIDE_INTERNAL_API
    /** @internal */
    static Padder none();

    /** @internal */
    static Padder codePoints(UChar32 cp, int32_t targetWidth, UNumberFormatPadPosition position);
#endif  // U_HIDE_INTERNAL_API

    /** @internal */
    static Padder forProperties(const DecimalFormatProperties& properties);

  private:
    UChar32 fWidth;  // -3 = error; -2 = bogus; -1 = no padding
    union {
        struct {
            int32_t fCp;
            UNumberFormatPadPosition fPosition;
        } padding;
        UErrorCode errorCode;
    } fUnion;

    Padder(UChar32 cp, int32_t width, UNumberFormatPadPosition position);

    Padder(int32_t width);

    Padder(UErrorCode errorCode) : fWidth(-3) { // NOLINT
        fUnion.errorCode = errorCode;
    }

    Padder() : fWidth(-2) {} // NOLINT

    bool isBogus() const {
        return fWidth == -2;
    }

    UBool copyErrorTo(UErrorCode &status) const {
        if (fWidth == -3) {
            status = fUnion.errorCode;
            return TRUE;
        }
        return FALSE;
    }

    bool isValid() const {
        return fWidth > 0;
    }

    int32_t padAndApply(const number::impl::Modifier &mod1, const number::impl::Modifier &mod2,
                        FormattedStringBuilder &string, int32_t leftIndex, int32_t rightIndex,
                        UErrorCode &status) const;

    // To allow MacroProps/MicroProps to initialize empty instances:
    friend struct MacroProps;
    friend struct MicroProps;

    // To allow NumberFormatterImpl to access isBogus() and perform other operations:
    friend class number::impl::NumberFormatterImpl;

    // To allow access to the skeleton generation code:
    friend class number::impl::GeneratorHelpers;
};

// Do not enclose entire MacroProps with #ifndef U_HIDE_INTERNAL_API, needed for a protected field
/** @internal */
struct U_I18N_API MacroProps : public UMemory {
    /** @internal */
    Notation notation;

    /** @internal */
    MeasureUnit unit; // = NoUnit::base();

    /** @internal */
    MeasureUnit perUnit; // = NoUnit::base();

    /** @internal */
    Precision precision;  // = Precision();  (bogus)

    /** @internal */
    UNumberFormatRoundingMode roundingMode = UNUM_ROUND_HALFEVEN;

    /** @internal */
    Grouper grouper;  // = Grouper();  (bogus)

    /** @internal */
    Padder padder;    // = Padder();   (bogus)

    /** @internal */
    IntegerWidth integerWidth; // = IntegerWidth(); (bogus)

    /** @internal */
    SymbolsWrapper symbols;

    // UNUM_XYZ_COUNT denotes null (bogus) values.

    /** @internal */
    UNumberUnitWidth unitWidth = UNUM_UNIT_WIDTH_COUNT;

    /** @internal */
    UNumberSignDisplay sign = UNUM_SIGN_COUNT;

    /** @internal */
    UNumberDecimalSeparatorDisplay decimal = UNUM_DECIMAL_SEPARATOR_COUNT;

    /** @internal */
    Scale scale;  // = Scale();  (benign value)

    /** @internal */
    const AffixPatternProvider* affixProvider = nullptr;  // no ownership

    /** @internal */
    const PluralRules* rules = nullptr;  // no ownership

    /** @internal */
    const CurrencySymbols* currencySymbols = nullptr;  // no ownership

    /** @internal */
    int32_t threshold = kInternalDefaultThreshold;

    /** @internal */
    Locale locale;

    // NOTE: Uses default copy and move constructors.

    /**
     * Check all members for errors.
     * @internal
     */
    bool copyErrorTo(UErrorCode &status) const {
        return notation.copyErrorTo(status) || precision.copyErrorTo(status) ||
               padder.copyErrorTo(status) || integerWidth.copyErrorTo(status) ||
               symbols.copyErrorTo(status) || scale.copyErrorTo(status);
    }
};

} // namespace impl

/**
 * An abstract base class for specifying settings related to number formatting. This class is implemented by
 * {@link UnlocalizedNumberFormatter} and {@link LocalizedNumberFormatter}. This class is not intended for
 * public subclassing.
 */
template<typename Derived>
class U_I18N_API NumberFormatterSettings {
  public:
    /**
     * Specifies the notation style (simple, scientific, or compact) for rendering numbers.
     *
     * <ul>
     * <li>Simple notation: "12,300"
     * <li>Scientific notation: "1.23E4"
     * <li>Compact notation: "12K"
     * </ul>
     *
     * <p>
     * All notation styles will be properly localized with locale data, and all notation styles are compatible with
     * units, rounding precisions, and other number formatter settings.
     *
     * <p>
     * Pass this method the return value of a {@link Notation} factory method. For example:
     *
     * <pre>
     * NumberFormatter::with().notation(Notation::compactShort())
     * </pre>
     *
     * The default is to use simple notation.
     *
     * @param notation
     *            The notation strategy to use.
     * @return The fluent chain.
     * @see Notation
     * @stable ICU 60
     */
    Derived notation(const Notation &notation) const &;

    /**
     * Overload of notation() for use on an rvalue reference.
     *
     * @param notation
     *            The notation strategy to use.
     * @return The fluent chain.
     * @see #notation
     * @stable ICU 62
     */
    Derived notation(const Notation &notation) &&;

    /**
     * Specifies the unit (unit of measure, currency, or percent) to associate with rendered numbers.
     *
     * <ul>
     * <li>Unit of measure: "12.3 meters"
     * <li>Currency: "$12.30"
     * <li>Percent: "12.3%"
     * </ul>
     *
     * All units will be properly localized with locale data, and all units are compatible with notation styles,
     * rounding precisions, and other number formatter settings.
     *
     * Pass this method any instance of {@link MeasureUnit}. For units of measure:
     *
     * <pre>
     * NumberFormatter::with().unit(MeasureUnit::getMeter())
     * </pre>
     *
     * Currency:
     *
     * <pre>
     * NumberFormatter::with().unit(CurrencyUnit(u"USD", status))
     * </pre>
     *
     * Percent:
     *
     * <pre>
     * NumberFormatter::with().unit(NoUnit.percent())
     * </pre>
     *
     * See {@link #perUnit} for information on how to format strings like "5 meters per second".
     *
     * The default is to render without units (equivalent to NoUnit.base()).
     *
     * @param unit
     *            The unit to render.
     * @return The fluent chain.
     * @see MeasureUnit
     * @see Currency
     * @see NoUnit
     * @see #perUnit
     * @stable ICU 60
     */
    Derived unit(const icu::MeasureUnit &unit) const &;

    /**
     * Overload of unit() for use on an rvalue reference.
     *
     * @param unit
     *            The unit to render.
     * @return The fluent chain.
     * @see #unit
     * @stable ICU 62
     */
    Derived unit(const icu::MeasureUnit &unit) &&;

    /**
     * Like unit(), but takes ownership of a pointer.  Convenient for use with the MeasureFormat factory
     * methods that return pointers that need ownership.
     *
     * Note: consider using the MeasureFormat factory methods that return by value.
     *
     * @param unit
     *            The unit to render.
     * @return The fluent chain.
     * @see #unit
     * @see MeasureUnit
     * @stable ICU 60
     */
    Derived adoptUnit(icu::MeasureUnit *unit) const &;

    /**
     * Overload of adoptUnit() for use on an rvalue reference.
     *
     * @param unit
     *            The unit to render.
     * @return The fluent chain.
     * @see #adoptUnit
     * @stable ICU 62
     */
    Derived adoptUnit(icu::MeasureUnit *unit) &&;

    /**
     * Sets a unit to be used in the denominator. For example, to format "3 m/s", pass METER to the unit and SECOND to
     * the perUnit.
     *
     * Pass this method any instance of {@link MeasureUnit}. Example:
     *
     * <pre>
     * NumberFormatter::with()
     *      .unit(MeasureUnit::getMeter())
     *      .perUnit(MeasureUnit::getSecond())
     * </pre>
     *
     * The default is not to display any unit in the denominator.
     *
     * If a per-unit is specified without a primary unit via {@link #unit}, the behavior is undefined.
     *
     * @param perUnit
     *            The unit to render in the denominator.
     * @return The fluent chain
     * @see #unit
     * @stable ICU 61
     */
    Derived perUnit(const icu::MeasureUnit &perUnit) const &;

    /**
     * Overload of perUnit() for use on an rvalue reference.
     *
     * @param perUnit
     *            The unit to render in the denominator.
     * @return The fluent chain.
     * @see #perUnit
     * @stable ICU 62
     */
    Derived perUnit(const icu::MeasureUnit &perUnit) &&;

    /**
     * Like perUnit(), but takes ownership of a pointer.  Convenient for use with the MeasureFormat factory
     * methods that return pointers that need ownership.
     *
     * Note: consider using the MeasureFormat factory methods that return by value.
     *
     * @param perUnit
     *            The unit to render in the denominator.
     * @return The fluent chain.
     * @see #perUnit
     * @see MeasureUnit
     * @stable ICU 61
     */
    Derived adoptPerUnit(icu::MeasureUnit *perUnit) const &;

    /**
     * Overload of adoptPerUnit() for use on an rvalue reference.
     *
     * @param perUnit
     *            The unit to render in the denominator.
     * @return The fluent chain.
     * @see #adoptPerUnit
     * @stable ICU 62
     */
    Derived adoptPerUnit(icu::MeasureUnit *perUnit) &&;

    /**
     * Specifies the rounding precision to use when formatting numbers.
     *
     * <ul>
     * <li>Round to 3 decimal places: "3.142"
     * <li>Round to 3 significant figures: "3.14"
     * <li>Round to the closest nickel: "3.15"
     * <li>Do not perform rounding: "3.1415926..."
     * </ul>
     *
     * <p>
     * Pass this method the return value of one of the factory methods on {@link Precision}. For example:
     *
     * <pre>
     * NumberFormatter::with().precision(Precision::fixedFraction(2))
     * </pre>
     *
     * <p>
     * In most cases, the default rounding strategy is to round to 6 fraction places; i.e.,
     * <code>Precision.maxFraction(6)</code>. The exceptions are if compact notation is being used, then the compact
     * notation rounding strategy is used (see {@link Notation#compactShort} for details), or if the unit is a currency,
     * then standard currency rounding is used, which varies from currency to currency (see {@link Precision#currency} for
     * details).
     *
     * @param precision
     *            The rounding precision to use.
     * @return The fluent chain.
     * @see Precision
     * @stable ICU 62
     */
    Derived precision(const Precision& precision) const &;

    /**
     * Overload of precision() for use on an rvalue reference.
     *
     * @param precision
     *            The rounding precision to use.
     * @return The fluent chain.
     * @see #precision
     * @stable ICU 62
     */
    Derived precision(const Precision& precision) &&;

    /**
     * Specifies how to determine the direction to round a number when it has more digits than fit in the
     * desired precision.  When formatting 1.235:
     *
     * <ul>
     * <li>Ceiling rounding mode with integer precision: "2"
     * <li>Half-down rounding mode with 2 fixed fraction digits: "1.23"
     * <li>Half-up rounding mode with 2 fixed fraction digits: "1.24"
     * </ul>
     *
     * The default is HALF_EVEN. For more information on rounding mode, see the ICU userguide here:
     *
     * http://userguide.icu-project.org/formatparse/numbers/rounding-modes
     *
     * @param roundingMode The rounding mode to use.
     * @return The fluent chain.
     * @stable ICU 62
     */
    Derived roundingMode(UNumberFormatRoundingMode roundingMode) const &;

    /**
     * Overload of roundingMode() for use on an rvalue reference.
     *
     * @param roundingMode The rounding mode to use.
     * @return The fluent chain.
     * @see #roundingMode
     * @stable ICU 62
     */
    Derived roundingMode(UNumberFormatRoundingMode roundingMode) &&;

    /**
     * Specifies the grouping strategy to use when formatting numbers.
     *
     * <ul>
     * <li>Default grouping: "12,300" and "1,230"
     * <li>Grouping with at least 2 digits: "12,300" and "1230"
     * <li>No grouping: "12300" and "1230"
     * </ul>
     *
     * <p>
     * The exact grouping widths will be chosen based on the locale.
     *
     * <p>
     * Pass this method an element from the {@link UNumberGroupingStrategy} enum. For example:
     *
     * <pre>
     * NumberFormatter::with().grouping(UNUM_GROUPING_MIN2)
     * </pre>
     *
     * The default is to perform grouping according to locale data; most locales, but not all locales,
     * enable it by default.
     *
     * @param strategy
     *            The grouping strategy to use.
     * @return The fluent chain.
     * @stable ICU 61
     */
    Derived grouping(UNumberGroupingStrategy strategy) const &;

    /**
     * Overload of grouping() for use on an rvalue reference.
     *
     * @param strategy
     *            The grouping strategy to use.
     * @return The fluent chain.
     * @see #grouping
     * @stable ICU 62
     */
    Derived grouping(UNumberGroupingStrategy strategy) &&;

    /**
     * Specifies the minimum and maximum number of digits to render before the decimal mark.
     *
     * <ul>
     * <li>Zero minimum integer digits: ".08"
     * <li>One minimum integer digit: "0.08"
     * <li>Two minimum integer digits: "00.08"
     * </ul>
     *
     * <p>
     * Pass this method the return value of {@link IntegerWidth#zeroFillTo}. For example:
     *
     * <pre>
     * NumberFormatter::with().integerWidth(IntegerWidth::zeroFillTo(2))
     * </pre>
     *
     * The default is to have one minimum integer digit.
     *
     * @param style
     *            The integer width to use.
     * @return The fluent chain.
     * @see IntegerWidth
     * @stable ICU 60
     */
    Derived integerWidth(const IntegerWidth &style) const &;

    /**
     * Overload of integerWidth() for use on an rvalue reference.
     *
     * @param style
     *            The integer width to use.
     * @return The fluent chain.
     * @see #integerWidth
     * @stable ICU 62
     */
    Derived integerWidth(const IntegerWidth &style) &&;

    /**
     * Specifies the symbols (decimal separator, grouping separator, percent sign, numerals, etc.) to use when rendering
     * numbers.
     *
     * <ul>
     * <li><em>en_US</em> symbols: "12,345.67"
     * <li><em>fr_FR</em> symbols: "12&nbsp;345,67"
     * <li><em>de_CH</em> symbols: "12’345.67"
     * <li><em>my_MY</em> symbols: "၁၂,၃၄၅.၆၇"
     * </ul>
     *
     * <p>
     * Pass this method an instance of {@link DecimalFormatSymbols}. For example:
     *
     * <pre>
     * NumberFormatter::with().symbols(DecimalFormatSymbols(Locale("de_CH"), status))
     * </pre>
     *
     * <p>
     * <strong>Note:</strong> DecimalFormatSymbols automatically chooses the best numbering system based on the locale.
     * In the examples above, the first three are using the Latin numbering system, and the fourth is using the Myanmar
     * numbering system.
     *
     * <p>
     * <strong>Note:</strong> The instance of DecimalFormatSymbols will be copied: changes made to the symbols object
     * after passing it into the fluent chain will not be seen.
     *
     * <p>
     * <strong>Note:</strong> Calling this method will override any previously specified DecimalFormatSymbols
     * or NumberingSystem.
     *
     * <p>
     * The default is to choose the symbols based on the locale specified in the fluent chain.
     *
     * @param symbols
     *            The DecimalFormatSymbols to use.
     * @return The fluent chain.
     * @see DecimalFormatSymbols
     * @stable ICU 60
     */
    Derived symbols(const DecimalFormatSymbols &symbols) const &;

    /**
     * Overload of symbols() for use on an rvalue reference.
     *
     * @param symbols
     *            The DecimalFormatSymbols to use.
     * @return The fluent chain.
     * @see #symbols
     * @stable ICU 62
     */
    Derived symbols(const DecimalFormatSymbols &symbols) &&;

    /**
     * Specifies that the given numbering system should be used when fetching symbols.
     *
     * <ul>
     * <li>Latin numbering system: "12,345"
     * <li>Myanmar numbering system: "၁၂,၃၄၅"
     * <li>Math Sans Bold numbering system: "𝟭𝟮,𝟯𝟰𝟱"
     * </ul>
     *
     * <p>
     * Pass this method an instance of {@link NumberingSystem}. For example, to force the locale to always use the Latin
     * alphabet numbering system (ASCII digits):
     *
     * <pre>
     * NumberFormatter::with().adoptSymbols(NumberingSystem::createInstanceByName("latn", status))
     * </pre>
     *
     * <p>
     * <strong>Note:</strong> Calling this method will override any previously specified DecimalFormatSymbols
     * or NumberingSystem.
     *
     * <p>
     * The default is to choose the best numbering system for the locale.
     *
     * <p>
     * This method takes ownership of a pointer in order to work nicely with the NumberingSystem factory methods.
     *
     * @param symbols
     *            The NumberingSystem to use.
     * @return The fluent chain.
     * @see NumberingSystem
     * @stable ICU 60
     */
    Derived adoptSymbols(NumberingSystem *symbols) const &;

    /**
     * Overload of adoptSymbols() for use on an rvalue reference.
     *
     * @param symbols
     *            The NumberingSystem to use.
     * @return The fluent chain.
     * @see #adoptSymbols
     * @stable ICU 62
     */
    Derived adoptSymbols(NumberingSystem *symbols) &&;

    /**
     * Sets the width of the unit (measure unit or currency).  Most common values:
     *
     * <ul>
     * <li>Short: "$12.00", "12 m"
     * <li>ISO Code: "USD 12.00"
     * <li>Full name: "12.00 US dollars", "12 meters"
     * </ul>
     *
     * <p>
     * Pass an element from the {@link UNumberUnitWidth} enum to this setter. For example:
     *
     * <pre>
     * NumberFormatter::with().unitWidth(UNumberUnitWidth::UNUM_UNIT_WIDTH_FULL_NAME)
     * </pre>
     *
     * <p>
     * The default is the SHORT width.
     *
     * @param width
     *            The width to use when rendering numbers.
     * @return The fluent chain
     * @see UNumberUnitWidth
     * @stable ICU 60
     */
    Derived unitWidth(UNumberUnitWidth width) const &;

    /**
     * Overload of unitWidth() for use on an rvalue reference.
     *
     * @param width
     *            The width to use when rendering numbers.
     * @return The fluent chain.
     * @see #unitWidth
     * @stable ICU 62
     */
    Derived unitWidth(UNumberUnitWidth width) &&;

    /**
     * Sets the plus/minus sign display strategy. Most common values:
     *
     * <ul>
     * <li>Auto: "123", "-123"
     * <li>Always: "+123", "-123"
     * <li>Accounting: "$123", "($123)"
     * </ul>
     *
     * <p>
     * Pass an element from the {@link UNumberSignDisplay} enum to this setter. For example:
     *
     * <pre>
     * NumberFormatter::with().sign(UNumberSignDisplay::UNUM_SIGN_ALWAYS)
     * </pre>
     *
     * <p>
     * The default is AUTO sign display.
     *
     * @param style
     *            The sign display strategy to use when rendering numbers.
     * @return The fluent chain
     * @see UNumberSignDisplay
     * @stable ICU 60
     */
    Derived sign(UNumberSignDisplay style) const &;

    /**
     * Overload of sign() for use on an rvalue reference.
     *
     * @param style
     *            The sign display strategy to use when rendering numbers.
     * @return The fluent chain.
     * @see #sign
     * @stable ICU 62
     */
    Derived sign(UNumberSignDisplay style) &&;

    /**
     * Sets the decimal separator display strategy. This affects integer numbers with no fraction part. Most common
     * values:
     *
     * <ul>
     * <li>Auto: "1"
     * <li>Always: "1."
     * </ul>
     *
     * <p>
     * Pass an element from the {@link UNumberDecimalSeparatorDisplay} enum to this setter. For example:
     *
     * <pre>
     * NumberFormatter::with().decimal(UNumberDecimalSeparatorDisplay::UNUM_DECIMAL_SEPARATOR_ALWAYS)
     * </pre>
     *
     * <p>
     * The default is AUTO decimal separator display.
     *
     * @param style
     *            The decimal separator display strategy to use when rendering numbers.
     * @return The fluent chain
     * @see UNumberDecimalSeparatorDisplay
     * @stable ICU 60
     */
    Derived decimal(UNumberDecimalSeparatorDisplay style) const &;

    /**
     * Overload of decimal() for use on an rvalue reference.
     *
     * @param style
     *            The decimal separator display strategy to use when rendering numbers.
     * @return The fluent chain.
     * @see #decimal
     * @stable ICU 62
     */
    Derived decimal(UNumberDecimalSeparatorDisplay style) &&;

    /**
     * Sets a scale (multiplier) to be used to scale the number by an arbitrary amount before formatting.
     * Most common values:
     *
     * <ul>
     * <li>Multiply by 100: useful for percentages.
     * <li>Multiply by an arbitrary value: useful for unit conversions.
     * </ul>
     *
     * <p>
     * Pass an element from a {@link Scale} factory method to this setter. For example:
     *
     * <pre>
     * NumberFormatter::with().scale(Scale::powerOfTen(2))
     * </pre>
     *
     * <p>
     * The default is to not apply any multiplier.
     *
     * @param scale
     *            The scale to apply when rendering numbers.
     * @return The fluent chain
     * @stable ICU 62
     */
    Derived scale(const Scale &scale) const &;

    /**
     * Overload of scale() for use on an rvalue reference.
     *
     * @param scale
     *            The scale to apply when rendering numbers.
     * @return The fluent chain.
     * @see #scale
     * @stable ICU 62
     */
    Derived scale(const Scale &scale) &&;

#ifndef U_HIDE_INTERNAL_API

    /**
     * Set the padding strategy. May be added in the future; see #13338.
     *
     * @internal ICU 60: This API is ICU internal only.
     */
    Derived padding(const number::impl::Padder &padder) const &;

    /** @internal */
    Derived padding(const number::impl::Padder &padder) &&;

    /**
     * Internal fluent setter to support a custom regulation threshold. A threshold of 1 causes the data structures to
     * be built right away. A threshold of 0 prevents the data structures from being built.
     *
     * @internal ICU 60: This API is ICU internal only.
     */
    Derived threshold(int32_t threshold) const &;

    /** @internal */
    Derived threshold(int32_t threshold) &&;

    /**
     * Internal fluent setter to overwrite the entire macros object.
     *
     * @internal ICU 60: This API is ICU internal only.
     */
    Derived macros(const number::impl::MacroProps& macros) const &;

    /** @internal */
    Derived macros(const number::impl::MacroProps& macros) &&;

    /** @internal */
    Derived macros(impl::MacroProps&& macros) const &;

    /** @internal */
    Derived macros(impl::MacroProps&& macros) &&;

#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Creates a skeleton string representation of this number formatter. A skeleton string is a
     * locale-agnostic serialized form of a number formatter.
     *
     * Not all options are capable of being represented in the skeleton string; for example, a
     * DecimalFormatSymbols object. If any such option is encountered, the error code is set to
     * U_UNSUPPORTED_ERROR.
     *
     * The returned skeleton is in normalized form, such that two number formatters with equivalent
     * behavior should produce the same skeleton.
     *
     * @return A number skeleton string with behavior corresponding to this number formatter.
     * @stable ICU 62
     */
    UnicodeString toSkeleton(UErrorCode& status) const;

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns the current (Un)LocalizedNumberFormatter as a LocalPointer
     * wrapping a heap-allocated copy of the current object.
     *
     * This is equivalent to new-ing the move constructor with a value object
     * as the argument.
     *
     * @return A wrapped (Un)LocalizedNumberFormatter pointer, or a wrapped
     *         nullptr on failure.
     * @draft ICU 64
     */
    LocalPointer<Derived> clone() const &;

    /**
     * Overload of clone for use on an rvalue reference.
     *
     * @return A wrapped (Un)LocalizedNumberFormatter pointer, or a wrapped
     *         nullptr on failure.
     * @draft ICU 64
     */
    LocalPointer<Derived> clone() &&;
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Sets the UErrorCode if an error occurred in the fluent chain.
     * Preserves older error codes in the outErrorCode.
     * @return TRUE if U_FAILURE(outErrorCode)
     * @stable ICU 60
     */
    UBool copyErrorTo(UErrorCode &outErrorCode) const {
        if (U_FAILURE(outErrorCode)) {
            // Do not overwrite the older error code
            return TRUE;
        }
        fMacros.copyErrorTo(outErrorCode);
        return U_FAILURE(outErrorCode);
    }

    // NOTE: Uses default copy and move constructors.

  private:
    number::impl::MacroProps fMacros;

    // Don't construct me directly!  Use (Un)LocalizedNumberFormatter.
    NumberFormatterSettings() = default;

    friend class LocalizedNumberFormatter;
    friend class UnlocalizedNumberFormatter;

    // Give NumberRangeFormatter access to the MacroProps
    friend void number::impl::touchRangeLocales(impl::RangeMacroProps& macros);
    friend class number::impl::NumberRangeFormatterImpl;
};

/**
 * A NumberFormatter that does not yet have a locale. In order to format numbers, a locale must be specified.
 *
 * Instances of this class are immutable and thread-safe.
 *
 * @see NumberFormatter
 * @stable ICU 60
 */
class U_I18N_API UnlocalizedNumberFormatter
        : public NumberFormatterSettings<UnlocalizedNumberFormatter>, public UMemory {

  public:
    /**
     * Associate the given locale with the number formatter. The locale is used for picking the appropriate symbols,
     * formats, and other data for number display.
     *
     * @param locale
     *            The locale to use when loading data for number formatting.
     * @return The fluent chain.
     * @stable ICU 60
     */
    LocalizedNumberFormatter locale(const icu::Locale &locale) const &;

    /**
     * Overload of locale() for use on an rvalue reference.
     *
     * @param locale
     *            The locale to use when loading data for number formatting.
     * @return The fluent chain.
     * @see #locale
     * @stable ICU 62
     */
    LocalizedNumberFormatter locale(const icu::Locale &locale) &&;

    /**
     * Default constructor: puts the formatter into a valid but undefined state.
     *
     * @stable ICU 62
     */
    UnlocalizedNumberFormatter() = default;

    /**
     * Returns a copy of this UnlocalizedNumberFormatter.
     * @stable ICU 60
     */
    UnlocalizedNumberFormatter(const UnlocalizedNumberFormatter &other);

    /**
     * Move constructor:
     * The source UnlocalizedNumberFormatter will be left in a valid but undefined state.
     * @stable ICU 62
     */
    UnlocalizedNumberFormatter(UnlocalizedNumberFormatter&& src) U_NOEXCEPT;

    /**
     * Copy assignment operator.
     * @stable ICU 62
     */
    UnlocalizedNumberFormatter& operator=(const UnlocalizedNumberFormatter& other);

    /**
     * Move assignment operator:
     * The source UnlocalizedNumberFormatter will be left in a valid but undefined state.
     * @stable ICU 62
     */
    UnlocalizedNumberFormatter& operator=(UnlocalizedNumberFormatter&& src) U_NOEXCEPT;

  private:
    explicit UnlocalizedNumberFormatter(const NumberFormatterSettings<UnlocalizedNumberFormatter>& other);

    explicit UnlocalizedNumberFormatter(
            NumberFormatterSettings<UnlocalizedNumberFormatter>&& src) U_NOEXCEPT;

    // To give the fluent setters access to this class's constructor:
    friend class NumberFormatterSettings<UnlocalizedNumberFormatter>;

    // To give NumberFormatter::with() access to this class's constructor:
    friend class NumberFormatter;
};

/**
 * A NumberFormatter that has a locale associated with it; this means .format() methods are available.
 *
 * Instances of this class are immutable and thread-safe.
 *
 * @see NumberFormatter
 * @stable ICU 60
 */
class U_I18N_API LocalizedNumberFormatter
        : public NumberFormatterSettings<LocalizedNumberFormatter>, public UMemory {
  public:
    /**
     * Format the given integer number to a string using the settings specified in the NumberFormatter fluent
     * setting chain.
     *
     * @param value
     *            The number to format.
     * @param status
     *            Set to an ErrorCode if one occurred in the setter chain or during formatting.
     * @return A FormattedNumber object; call .toString() to get the string.
     * @stable ICU 60
     */
    FormattedNumber formatInt(int64_t value, UErrorCode &status) const;

    /**
     * Format the given float or double to a string using the settings specified in the NumberFormatter fluent setting
     * chain.
     *
     * @param value
     *            The number to format.
     * @param status
     *            Set to an ErrorCode if one occurred in the setter chain or during formatting.
     * @return A FormattedNumber object; call .toString() to get the string.
     * @stable ICU 60
     */
    FormattedNumber formatDouble(double value, UErrorCode &status) const;

    /**
     * Format the given decimal number to a string using the settings
     * specified in the NumberFormatter fluent setting chain.
     * The syntax of the unformatted number is a "numeric string"
     * as defined in the Decimal Arithmetic Specification, available at
     * http://speleotrove.com/decimal
     *
     * @param value
     *            The number to format.
     * @param status
     *            Set to an ErrorCode if one occurred in the setter chain or during formatting.
     * @return A FormattedNumber object; call .toString() to get the string.
     * @stable ICU 60
     */
    FormattedNumber formatDecimal(StringPiece value, UErrorCode& status) const;

#ifndef U_HIDE_INTERNAL_API

    /** Internal method.
     * @internal
     */
    FormattedNumber formatDecimalQuantity(const number::impl::DecimalQuantity& dq, UErrorCode& status) const;

    /** Internal method for DecimalFormat compatibility.
     * @internal
     */
    void getAffixImpl(bool isPrefix, bool isNegative, UnicodeString& result, UErrorCode& status) const;

    /**
     * Internal method for testing.
     * @internal
     */
    const number::impl::NumberFormatterImpl* getCompiled() const;

    /**
     * Internal method for testing.
     * @internal
     */
    int32_t getCallCount() const;

#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Creates a representation of this LocalizedNumberFormat as an icu::Format, enabling the use
     * of this number formatter with APIs that need an object of that type, such as MessageFormat.
     *
     * This API is not intended to be used other than for enabling API compatibility. The formatDouble,
     * formatInt, and formatDecimal methods should normally be used when formatting numbers, not the Format
     * object returned by this method.
     *
     * The caller owns the returned object and must delete it when finished.
     *
     * @return A Format wrapping this LocalizedNumberFormatter.
     * @stable ICU 62
     */
    Format* toFormat(UErrorCode& status) const;

    /**
     * Default constructor: puts the formatter into a valid but undefined state.
     *
     * @stable ICU 62
     */
    LocalizedNumberFormatter() = default;

    /**
     * Returns a copy of this LocalizedNumberFormatter.
     * @stable ICU 60
     */
    LocalizedNumberFormatter(const LocalizedNumberFormatter &other);

    /**
     * Move constructor:
     * The source LocalizedNumberFormatter will be left in a valid but undefined state.
     * @stable ICU 62
     */
    LocalizedNumberFormatter(LocalizedNumberFormatter&& src) U_NOEXCEPT;

    /**
     * Copy assignment operator.
     * @stable ICU 62
     */
    LocalizedNumberFormatter& operator=(const LocalizedNumberFormatter& other);

    /**
     * Move assignment operator:
     * The source LocalizedNumberFormatter will be left in a valid but undefined state.
     * @stable ICU 62
     */
    LocalizedNumberFormatter& operator=(LocalizedNumberFormatter&& src) U_NOEXCEPT;

#ifndef U_HIDE_INTERNAL_API

    /**
     * This is the core entrypoint to the number formatting pipeline. It performs self-regulation: a static code path
     * for the first few calls, and compiling a more efficient data structure if called repeatedly.
     *
     * <p>
     * This function is very hot, being called in every call to the number formatting pipeline.
     *
     * @param results
     *            The results object. This method will mutate it to save the results.
     * @param status
     * @internal
     */
    void formatImpl(impl::UFormattedNumberData *results, UErrorCode &status) const;

#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Destruct this LocalizedNumberFormatter, cleaning up any memory it might own.
     * @stable ICU 60
     */
    ~LocalizedNumberFormatter();

  private:
    // Note: fCompiled can't be a LocalPointer because number::impl::NumberFormatterImpl is defined in an internal
    // header, and LocalPointer needs the full class definition in order to delete the instance.
    const number::impl::NumberFormatterImpl* fCompiled {nullptr};
    char fUnsafeCallCount[8] {};  // internally cast to u_atomic_int32_t

    explicit LocalizedNumberFormatter(const NumberFormatterSettings<LocalizedNumberFormatter>& other);

    explicit LocalizedNumberFormatter(NumberFormatterSettings<LocalizedNumberFormatter>&& src) U_NOEXCEPT;

    LocalizedNumberFormatter(const number::impl::MacroProps &macros, const Locale &locale);

    LocalizedNumberFormatter(impl::MacroProps &&macros, const Locale &locale);

    void clear();

    void lnfMoveHelper(LocalizedNumberFormatter&& src);

    /**
     * @return true if the compiled formatter is available.
     */
    bool computeCompiled(UErrorCode& status) const;

    // To give the fluent setters access to this class's constructor:
    friend class NumberFormatterSettings<UnlocalizedNumberFormatter>;
    friend class NumberFormatterSettings<LocalizedNumberFormatter>;

    // To give UnlocalizedNumberFormatter::locale() access to this class's constructor:
    friend class UnlocalizedNumberFormatter;
};

/**
 * The result of a number formatting operation. This class allows the result to be exported in several data types,
 * including a UnicodeString and a FieldPositionIterator.
 *
 * Instances of this class are immutable and thread-safe.
 *
 * @stable ICU 60
 */
class U_I18N_API FormattedNumber : public UMemory, public FormattedValue {
  public:

    // Default constructor cannot have #ifndef U_HIDE_DRAFT_API
#ifndef U_FORCE_HIDE_DRAFT_API
    /**
     * Default constructor; makes an empty FormattedNumber.
     * @draft ICU 64
     */
    FormattedNumber()
        : fData(nullptr), fErrorCode(U_INVALID_STATE_ERROR) {}
#endif  // U_FORCE_HIDE_DRAFT_API

    /**
     * Move constructor: Leaves the source FormattedNumber in an undefined state.
     * @stable ICU 62
     */
    FormattedNumber(FormattedNumber&& src) U_NOEXCEPT;

    /**
     * Destruct an instance of FormattedNumber.
     * @stable ICU 60
     */
    virtual ~FormattedNumber() U_OVERRIDE;

    /** Copying not supported; use move constructor instead. */
    FormattedNumber(const FormattedNumber&) = delete;

    /** Copying not supported; use move assignment instead. */
    FormattedNumber& operator=(const FormattedNumber&) = delete;

    /**
     * Move assignment: Leaves the source FormattedNumber in an undefined state.
     * @stable ICU 62
     */
    FormattedNumber& operator=(FormattedNumber&& src) U_NOEXCEPT;

    // Copybrief: this method is older than the parent method
    /**
     * @copybrief FormattedValue::toString()
     *
     * For more information, see FormattedValue::toString()
     *
     * @stable ICU 62
     */
    UnicodeString toString(UErrorCode& status) const U_OVERRIDE;

    // Copydoc: this method is new in ICU 64
    /** @copydoc FormattedValue::toTempString() */
    UnicodeString toTempString(UErrorCode& status) const U_OVERRIDE;

    // Copybrief: this method is older than the parent method
    /**
     * @copybrief FormattedValue::appendTo()
     *
     * For more information, see FormattedValue::appendTo()
     *
     * @stable ICU 62
     */
    Appendable &appendTo(Appendable& appendable, UErrorCode& status) const U_OVERRIDE;

    // Copydoc: this method is new in ICU 64
    /** @copydoc FormattedValue::nextPosition() */
    UBool nextPosition(ConstrainedFieldPosition& cfpos, UErrorCode& status) const U_OVERRIDE;

#ifndef U_HIDE_DRAFT_API
    /**
     * Determines the start (inclusive) and end (exclusive) indices of the next occurrence of the given
     * <em>field</em> in the output string. This allows you to determine the locations of, for example,
     * the integer part, fraction part, or symbols.
     *
     * This is a simpler but less powerful alternative to {@link #nextPosition}.
     *
     * If a field occurs just once, calling this method will find that occurrence and return it. If a
     * field occurs multiple times, this method may be called repeatedly with the following pattern:
     *
     * <pre>
     * FieldPosition fpos(UNUM_GROUPING_SEPARATOR_FIELD);
     * while (formattedNumber.nextFieldPosition(fpos, status)) {
     *   // do something with fpos.
     * }
     * </pre>
     *
     * This method is useful if you know which field to query. If you want all available field position
     * information, use {@link #nextPosition} or {@link #getAllFieldPositions}.
     *
     * @param fieldPosition
     *            Input+output variable. On input, the "field" property determines which field to look
     *            up, and the "beginIndex" and "endIndex" properties determine where to begin the search.
     *            On output, the "beginIndex" is set to the beginning of the first occurrence of the
     *            field with either begin or end indices after the input indices; "endIndex" is set to
     *            the end of that occurrence of the field (exclusive index). If a field position is not
     *            found, the method returns FALSE and the FieldPosition may or may not be changed.
     * @param status
     *            Set if an error occurs while populating the FieldPosition.
     * @return TRUE if a new occurrence of the field was found; FALSE otherwise.
     * @draft ICU 62
     * @see UNumberFormatFields
     */
    UBool nextFieldPosition(FieldPosition& fieldPosition, UErrorCode& status) const;

    /**
     * Export the formatted number to a FieldPositionIterator. This allows you to determine which characters in
     * the output string correspond to which <em>fields</em>, such as the integer part, fraction part, and sign.
     *
     * This is an alternative to the more powerful #nextPosition() API.
     *
     * If information on only one field is needed, use #nextPosition() or #nextFieldPosition() instead.
     *
     * @param iterator
     *            The FieldPositionIterator to populate with all of the fields present in the formatted number.
     * @param status
     *            Set if an error occurs while populating the FieldPositionIterator.
     * @draft ICU 62
     * @see UNumberFormatFields
     */
    void getAllFieldPositions(FieldPositionIterator &iterator, UErrorCode &status) const;
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
    /**
     * Export the formatted number as a "numeric string" conforming to the
     * syntax defined in the Decimal Arithmetic Specification, available at
     * http://speleotrove.com/decimal
     *
     * This endpoint is useful for obtaining the exact number being printed
     * after scaling and rounding have been applied by the number formatter.
     *
     * Example call site:
     *
     *     auto decimalNumber = fn.toDecimalNumber<std::string>(status);
     *
     * @tparam StringClass A string class compatible with StringByteSink;
     *         for example, std::string.
     * @param status Set if an error occurs.
     * @return A StringClass containing the numeric string.
     * @draft ICU 65
     */
    template<typename StringClass>
    inline StringClass toDecimalNumber(UErrorCode& status) const;
#endif // U_HIDE_DRAFT_API

#ifndef U_HIDE_INTERNAL_API

    /**
     *  Gets the raw DecimalQuantity for plural rule selection.
     *  @internal
     */
    void getDecimalQuantity(number::impl::DecimalQuantity& output, UErrorCode& status) const;

    /**
     * Populates the mutable builder type FieldPositionIteratorHandler.
     * @internal
     */
    void getAllFieldPositionsImpl(FieldPositionIteratorHandler& fpih, UErrorCode& status) const;

#endif  /* U_HIDE_INTERNAL_API */

  private:
    // Can't use LocalPointer because UFormattedNumberData is forward-declared
    const number::impl::UFormattedNumberData *fData;

    // Error code for the terminal methods
    UErrorCode fErrorCode;

    /**
     * Internal constructor from data type. Adopts the data pointer.
     * @internal
     */
    explicit FormattedNumber(impl::UFormattedNumberData *results)
        : fData(results), fErrorCode(U_ZERO_ERROR) {}

    explicit FormattedNumber(UErrorCode errorCode)
        : fData(nullptr), fErrorCode(errorCode) {}

    // TODO(ICU-20775): Propose this as API.
    void toDecimalNumber(ByteSink& sink, UErrorCode& status) const;

    // To give LocalizedNumberFormatter format methods access to this class's constructor:
    friend class LocalizedNumberFormatter;

    // To give C API access to internals
    friend struct number::impl::UFormattedNumberImpl;
};

#ifndef U_HIDE_DRAFT_API
// Note: This is draft ICU 65
template<typename StringClass>
StringClass FormattedNumber::toDecimalNumber(UErrorCode& status) const {
    StringClass result;
    StringByteSink<StringClass> sink(&result);
    toDecimalNumber(sink, status);
    return result;
}
#endif // U_HIDE_DRAFT_API

/**
 * See the main description in numberformatter.h for documentation and examples.
 *
 * @stable ICU 60
 */
class U_I18N_API NumberFormatter final {
  public:
    /**
     * Call this method at the beginning of a NumberFormatter fluent chain in which the locale is not currently known at
     * the call site.
     *
     * @return An {@link UnlocalizedNumberFormatter}, to be used for chaining.
     * @stable ICU 60
     */
    static UnlocalizedNumberFormatter with();

    /**
     * Call this method at the beginning of a NumberFormatter fluent chain in which the locale is known at the call
     * site.
     *
     * @param locale
     *            The locale from which to load formats and symbols for number formatting.
     * @return A {@link LocalizedNumberFormatter}, to be used for chaining.
     * @stable ICU 60
     */
    static LocalizedNumberFormatter withLocale(const Locale &locale);

    /**
     * Call this method at the beginning of a NumberFormatter fluent chain to create an instance based
     * on a given number skeleton string.
     *
     * It is possible for an error to occur while parsing. See the overload of this method if you are
     * interested in the location of a possible parse error.
     *
     * @param skeleton
     *            The skeleton string off of which to base this NumberFormatter.
     * @param status
     *            Set to U_NUMBER_SKELETON_SYNTAX_ERROR if the skeleton was invalid.
     * @return An UnlocalizedNumberFormatter, to be used for chaining.
     * @stable ICU 62
     */
    static UnlocalizedNumberFormatter forSkeleton(const UnicodeString& skeleton, UErrorCode& status);

#ifndef U_HIDE_DRAFT_API
    /**
     * Call this method at the beginning of a NumberFormatter fluent chain to create an instance based
     * on a given number skeleton string.
     *
     * If an error occurs while parsing the skeleton string, the offset into the skeleton string at
     * which the error occurred will be saved into the UParseError, if provided.
     *
     * @param skeleton
     *            The skeleton string off of which to base this NumberFormatter.
     * @param perror
     *            A parse error struct populated if an error occurs when parsing.
 *                If no error occurs, perror.offset will be set to -1.
     * @param status
     *            Set to U_NUMBER_SKELETON_SYNTAX_ERROR if the skeleton was invalid.
     * @return An UnlocalizedNumberFormatter, to be used for chaining.
     * @draft ICU 64
     */
    static UnlocalizedNumberFormatter forSkeleton(const UnicodeString& skeleton,
                                                  UParseError& perror, UErrorCode& status);
#endif

    /**
     * Use factory methods instead of the constructor to create a NumberFormatter.
     */
    NumberFormatter() = delete;
};

}  // namespace number
U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */






// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2015, International Business Machines Corporation
 * and others. All Rights Reserved.
 *******************************************************************************
 * standardplural.h
 *
 * created on: 2015dec14
 * created by: Markus W. Scherer
 */




#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

class UnicodeString;

/**
 * Standard CLDR plural form/category constants.
 * See http://www.unicode.org/reports/tr35/tr35-numbers.html#Language_Plural_Rules
 */
class U_I18N_API StandardPlural {
public:
    enum Form {
        ZERO,
        ONE,
        TWO,
        FEW,
        MANY,
        OTHER,
        COUNT
    };

    /**
     * @return the lowercase CLDR keyword string for the plural form
     */
    static const char *getKeyword(Form p);

    /**
     * @param keyword for example "few" or "other"
     * @return the plural form corresponding to the keyword, or OTHER
     */
    static Form orOtherFromString(const char *keyword) {
        return static_cast<Form>(indexOrOtherIndexFromString(keyword));
    }

    /**
     * @param keyword for example "few" or "other"
     * @return the plural form corresponding to the keyword, or OTHER
     */
    static Form orOtherFromString(const UnicodeString &keyword) {
        return static_cast<Form>(indexOrOtherIndexFromString(keyword));
    }

    /**
     * Sets U_ILLEGAL_ARGUMENT_ERROR if the keyword is not a plural form.
     *
     * @param keyword for example "few" or "other"
     * @return the plural form corresponding to the keyword
     */
    static Form fromString(const char *keyword, UErrorCode &errorCode) {
        return static_cast<Form>(indexFromString(keyword, errorCode));
    }

    /**
     * Sets U_ILLEGAL_ARGUMENT_ERROR if the keyword is not a plural form.
     *
     * @param keyword for example "few" or "other"
     * @return the plural form corresponding to the keyword
     */
    static Form fromString(const UnicodeString &keyword, UErrorCode &errorCode) {
        return static_cast<Form>(indexFromString(keyword, errorCode));
    }

    /**
     * @param keyword for example "few" or "other"
     * @return the index of the plural form corresponding to the keyword, or a negative value
     */
    static int32_t indexOrNegativeFromString(const char *keyword);

    /**
     * @param keyword for example "few" or "other"
     * @return the index of the plural form corresponding to the keyword, or a negative value
     */
    static int32_t indexOrNegativeFromString(const UnicodeString &keyword);

    /**
     * @param keyword for example "few" or "other"
     * @return the index of the plural form corresponding to the keyword, or OTHER
     */
    static int32_t indexOrOtherIndexFromString(const char *keyword) {
        int32_t i = indexOrNegativeFromString(keyword);
        return i >= 0 ? i : OTHER;
    }

    /**
     * @param keyword for example "few" or "other"
     * @return the index of the plural form corresponding to the keyword, or OTHER
     */
    static int32_t indexOrOtherIndexFromString(const UnicodeString &keyword) {
        int32_t i = indexOrNegativeFromString(keyword);
        return i >= 0 ? i : OTHER;
    }

    /**
     * Sets U_ILLEGAL_ARGUMENT_ERROR if the keyword is not a plural form.
     *
     * @param keyword for example "few" or "other"
     * @return the index of the plural form corresponding to the keyword
     */
    static int32_t indexFromString(const char *keyword, UErrorCode &errorCode);

    /**
     * Sets U_ILLEGAL_ARGUMENT_ERROR if the keyword is not a plural form.
     *
     * @param keyword for example "few" or "other"
     * @return the index of the plural form corresponding to the keyword
     */
    static int32_t indexFromString(const UnicodeString &keyword, UErrorCode &errorCode);
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_FORMATTING

// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING


#include <cstdint>
 // for UNUM_FIELD_COUNT


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2015, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING





U_NAMESPACE_BEGIN

// utility FieldPositionHandler
// base class, null implementation

class U_I18N_API FieldPositionHandler: public UMemory {
 protected:
  int32_t fShift = 0;

 public:
  virtual ~FieldPositionHandler();
  virtual void addAttribute(int32_t id, int32_t start, int32_t limit) = 0;
  virtual void shiftLast(int32_t delta) = 0;
  virtual UBool isRecording(void) const = 0;

  void setShift(int32_t delta);
};


// utility subclass FieldPositionOnlyHandler

class FieldPositionOnlyHandler : public FieldPositionHandler {
  FieldPosition& pos;
  UBool acceptFirstOnly = FALSE;
  UBool seenFirst = FALSE;

 public:
  FieldPositionOnlyHandler(FieldPosition& pos);
  virtual ~FieldPositionOnlyHandler();

  void addAttribute(int32_t id, int32_t start, int32_t limit) U_OVERRIDE;
  void shiftLast(int32_t delta) U_OVERRIDE;
  UBool isRecording(void) const U_OVERRIDE;

  /**
   * Enable this option to lock in the FieldPosition value after seeing the
   * first occurrence of the field. The default behavior is to take the last
   * occurrence.
   */
  void setAcceptFirstOnly(UBool acceptFirstOnly);
};


// utility subclass FieldPositionIteratorHandler

class FieldPositionIteratorHandler : public FieldPositionHandler {
  FieldPositionIterator* iter; // can be NULL
  UVector32* vec;
  UErrorCode status;
  UFieldCategory fCategory;

  // Note, we keep a reference to status, so if status is on the stack, we have
  // to be destroyed before status goes out of scope.  Easiest thing is to
  // allocate us on the stack in the same (or narrower) scope as status has.
  // This attempts to encourage that by blocking heap allocation.
  static void* U_EXPORT2 operator new(size_t) U_NOEXCEPT = delete;
  static void* U_EXPORT2 operator new[](size_t) U_NOEXCEPT = delete;
#if U_HAVE_PLACEMENT_NEW
  static void* U_EXPORT2 operator new(size_t, void*) U_NOEXCEPT = delete;
#endif

 public:
  FieldPositionIteratorHandler(FieldPositionIterator* posIter, UErrorCode& status);
  /** If using this constructor, you must call getError() when done formatting! */
  FieldPositionIteratorHandler(UVector32* vec, UErrorCode& status);
  ~FieldPositionIteratorHandler();

  void addAttribute(int32_t id, int32_t start, int32_t limit) U_OVERRIDE;
  void shiftLast(int32_t delta) U_OVERRIDE;
  UBool isRecording(void) const U_OVERRIDE;

  /** Copies a failed error code into _status. */
  inline void getError(UErrorCode& _status) {
    if (U_SUCCESS(_status) && U_FAILURE(status)) {
      _status = status;
    }
  }

  inline void setCategory(UFieldCategory category) {
    fCategory = category;
  }
};

U_NAMESPACE_END

#endif /* !UCONFIG_NO_FORMATTING */



U_NAMESPACE_BEGIN

class FormattedValueStringBuilderImpl;

/**
 * A StringBuilder optimized for formatting. It implements the following key
 * features beyond a UnicodeString:
 *
 * <ol>
 * <li>Efficient prepend as well as append.
 * <li>Keeps tracks of Fields in an efficient manner.
 * </ol>
 *
 * See also FormattedValueStringBuilderImpl.
 *
 * @author sffc (Shane Carr)
 */
class U_I18N_API FormattedStringBuilder : public UMemory {
  private:
    static const int32_t DEFAULT_CAPACITY = 40;

    template<typename T>
    union ValueOrHeapArray {
        T value[DEFAULT_CAPACITY];
        struct {
            T *ptr;
            int32_t capacity;
        } heap;
    };

  public:
    FormattedStringBuilder();

    ~FormattedStringBuilder();

    FormattedStringBuilder(const FormattedStringBuilder &other);

    // Convention: bottom 4 bits for field, top 4 bits for field category.
    // Field category 0 implies the number category so that the number field
    // literals can be directly passed as a Field type.
    // See the helper functions in "StringBuilderFieldUtils" below.
    typedef uint8_t Field;

    FormattedStringBuilder &operator=(const FormattedStringBuilder &other);

    int32_t length() const;

    int32_t codePointCount() const;

    inline char16_t charAt(int32_t index) const {
        U_ASSERT(index >= 0);
        U_ASSERT(index < fLength);
        return getCharPtr()[fZero + index];
    }

    inline Field fieldAt(int32_t index) const {
        U_ASSERT(index >= 0);
        U_ASSERT(index < fLength);
        return getFieldPtr()[fZero + index];
    }

    UChar32 getFirstCodePoint() const;

    UChar32 getLastCodePoint() const;

    UChar32 codePointAt(int32_t index) const;

    UChar32 codePointBefore(int32_t index) const;

    FormattedStringBuilder &clear();

    /** Appends a UTF-16 code unit. */
    inline int32_t appendChar16(char16_t codeUnit, Field field, UErrorCode& status) {
        // appendCodePoint handles both code units and code points.
        return insertCodePoint(fLength, codeUnit, field, status);
    }

    /** Inserts a UTF-16 code unit. Note: insert at index 0 is very efficient. */
    inline int32_t insertChar16(int32_t index, char16_t codeUnit, Field field, UErrorCode& status) {
        // insertCodePoint handles both code units and code points.
        return insertCodePoint(index, codeUnit, field, status);
    }

    /** Appends a Unicode code point. */
    inline int32_t appendCodePoint(UChar32 codePoint, Field field, UErrorCode &status) {
        return insertCodePoint(fLength, codePoint, field, status);
    }

    /** Inserts a Unicode code point. Note: insert at index 0 is very efficient. */
    int32_t insertCodePoint(int32_t index, UChar32 codePoint, Field field, UErrorCode &status);

    /** Appends a string. */
    inline int32_t append(const UnicodeString &unistr, Field field, UErrorCode &status) {
        return insert(fLength, unistr, field, status);
    }

    /** Inserts a string. Note: insert at index 0 is very efficient. */
    int32_t insert(int32_t index, const UnicodeString &unistr, Field field, UErrorCode &status);

    /** Inserts a substring. Note: insert at index 0 is very efficient.
     *
     * @param start Start index of the substring of unistr to be inserted.
     * @param end End index of the substring of unistr to be inserted (exclusive).
     */
    int32_t insert(int32_t index, const UnicodeString &unistr, int32_t start, int32_t end, Field field,
                   UErrorCode &status);

    /** Deletes a substring and then inserts a string at that same position.
     * Similar to JavaScript Array.prototype.splice().
     *
     * @param startThis Start of the span to delete.
     * @param endThis End of the span to delete (exclusive).
     * @param unistr The string to insert at the deletion position.
     * @param startOther Start index of the substring of unistr to be inserted.
     * @param endOther End index of the substring of unistr to be inserted (exclusive).
     */
    int32_t splice(int32_t startThis, int32_t endThis,  const UnicodeString &unistr,
                   int32_t startOther, int32_t endOther, Field field, UErrorCode& status);

    /** Appends a formatted string. */
    int32_t append(const FormattedStringBuilder &other, UErrorCode &status);

    /** Inserts a formatted string. Note: insert at index 0 is very efficient. */
    int32_t insert(int32_t index, const FormattedStringBuilder &other, UErrorCode &status);

    /**
     * Ensures that the string buffer contains a NUL terminator. The NUL terminator does
     * not count toward the string length. Any further changes to the string (insert or
     * append) may invalidate the NUL terminator.
     *
     * You should call this method after the formatted string is completely built if you
     * plan to return a pointer to the string from a C API.
     */
    void writeTerminator(UErrorCode& status);

    /**
     * Gets a "safe" UnicodeString that can be used even after the FormattedStringBuilder is destructed.
     */
    UnicodeString toUnicodeString() const;

    /**
     * Gets an "unsafe" UnicodeString that is valid only as long as the FormattedStringBuilder is alive and
     * unchanged. Slightly faster than toUnicodeString().
     */
    const UnicodeString toTempUnicodeString() const;

    UnicodeString toDebugString() const;

    const char16_t *chars() const;

    bool contentEquals(const FormattedStringBuilder &other) const;

    bool containsField(Field field) const;

  private:
    bool fUsingHeap = false;
    ValueOrHeapArray<char16_t> fChars;
    ValueOrHeapArray<Field> fFields;
    int32_t fZero = DEFAULT_CAPACITY / 2;
    int32_t fLength = 0;

    inline char16_t *getCharPtr() {
        return fUsingHeap ? fChars.heap.ptr : fChars.value;
    }

    inline const char16_t *getCharPtr() const {
        return fUsingHeap ? fChars.heap.ptr : fChars.value;
    }

    inline Field *getFieldPtr() {
        return fUsingHeap ? fFields.heap.ptr : fFields.value;
    }

    inline const Field *getFieldPtr() const {
        return fUsingHeap ? fFields.heap.ptr : fFields.value;
    }

    inline int32_t getCapacity() const {
        return fUsingHeap ? fChars.heap.capacity : DEFAULT_CAPACITY;
    }

    int32_t prepareForInsert(int32_t index, int32_t count, UErrorCode &status);

    int32_t prepareForInsertHelper(int32_t index, int32_t count, UErrorCode &status);

    int32_t remove(int32_t index, int32_t count);

    friend class FormattedValueStringBuilderImpl;
};

/**
 * Helper functions for dealing with the Field typedef, which stores fields
 * in a compressed format.
 */
class StringBuilderFieldUtils {
public:
    struct CategoryFieldPair {
        int32_t category;
        int32_t field;
    };

    /** Compile-time function to construct a Field from a category and a field */
    template <int32_t category, int32_t field>
    static constexpr FormattedStringBuilder::Field compress() {
        static_assert(category != 0, "cannot use Undefined category in FieldUtils");
        static_assert(category <= 0xf, "only 4 bits for category");
        static_assert(field <= 0xf, "only 4 bits for field");
        return static_cast<int8_t>((category << 4) | field);
    }

    /** Runtime inline function to unpack the category and field from the Field */
    static inline CategoryFieldPair expand(FormattedStringBuilder::Field field) {
        if (field == UNUM_FIELD_COUNT) {
            return {UFIELD_CATEGORY_UNDEFINED, 0};
        }
        CategoryFieldPair ret = {
            (field >> 4),
            (field & 0xf)
        };
        if (ret.category == 0) {
            ret.category = UFIELD_CATEGORY_NUMBER;
        }
        return ret;
    }

    static inline bool isNumericField(FormattedStringBuilder::Field field) {
        int8_t category = field >> 4;
        return category == 0 || category == UFIELD_CATEGORY_NUMBER;
    }
};

U_NAMESPACE_END


#endif //__NUMBER_STRINGBUILDER_H__



U_NAMESPACE_BEGIN
namespace number {
namespace impl {

// For convenience and historical reasons, import the Field typedef to the namespace.
typedef FormattedStringBuilder::Field Field;

// Typedef several enums for brevity and for easier comparison to Java.

typedef UNumberFormatRoundingMode RoundingMode;

typedef UNumberFormatPadPosition PadPosition;

typedef UNumberCompactStyle CompactStyle;

// ICU4J Equivalent: RoundingUtils.MAX_INT_FRAC_SIG
static constexpr int32_t kMaxIntFracSig = 999;

// ICU4J Equivalent: RoundingUtils.DEFAULT_ROUNDING_MODE
static constexpr RoundingMode kDefaultMode = RoundingMode::UNUM_FOUND_HALFEVEN;

// ICU4J Equivalent: Padder.FALLBACK_PADDING_STRING
static constexpr char16_t kFallbackPaddingString[] = u" ";

// Forward declarations:

class Modifier;
class MutablePatternModifier;
class DecimalQuantity;
class ModifierStore;
struct MicroProps;


enum AffixPatternType {
    // Represents a literal character; the value is stored in the code point field.
            TYPE_CODEPOINT = 0,

    // Represents a minus sign symbol '-'.
            TYPE_MINUS_SIGN = -1,

    // Represents a plus sign symbol '+'.
            TYPE_PLUS_SIGN = -2,

    // Represents a percent sign symbol '%'.
            TYPE_PERCENT = -3,

    // Represents a permille sign symbol '‰'.
            TYPE_PERMILLE = -4,

    // Represents a single currency symbol '¤'.
            TYPE_CURRENCY_SINGLE = -5,

    // Represents a double currency symbol '¤¤'.
            TYPE_CURRENCY_DOUBLE = -6,

    // Represents a triple currency symbol '¤¤¤'.
            TYPE_CURRENCY_TRIPLE = -7,

    // Represents a quadruple currency symbol '¤¤¤¤'.
            TYPE_CURRENCY_QUAD = -8,

    // Represents a quintuple currency symbol '¤¤¤¤¤'.
            TYPE_CURRENCY_QUINT = -9,

    // Represents a sequence of six or more currency symbols.
            TYPE_CURRENCY_OVERFLOW = -15
};

enum CompactType {
    TYPE_DECIMAL, TYPE_CURRENCY
};

enum Signum {
    SIGNUM_NEG = -1,
    SIGNUM_ZERO = 0,
    SIGNUM_POS = 1
};


class U_I18N_API AffixPatternProvider {
  public:
    static const int32_t AFFIX_PLURAL_MASK = 0xff;
    static const int32_t AFFIX_PREFIX = 0x100;
    static const int32_t AFFIX_NEGATIVE_SUBPATTERN = 0x200;
    static const int32_t AFFIX_PADDING = 0x400;

    // Convenience compound flags
    static const int32_t AFFIX_POS_PREFIX = AFFIX_PREFIX;
    static const int32_t AFFIX_POS_SUFFIX = 0;
    static const int32_t AFFIX_NEG_PREFIX = AFFIX_PREFIX | AFFIX_NEGATIVE_SUBPATTERN;
    static const int32_t AFFIX_NEG_SUFFIX = AFFIX_NEGATIVE_SUBPATTERN;

    virtual ~AffixPatternProvider();

    virtual char16_t charAt(int flags, int i) const = 0;

    virtual int length(int flags) const = 0;

    virtual UnicodeString getString(int flags) const = 0;

    virtual bool hasCurrencySign() const = 0;

    virtual bool positiveHasPlusSign() const = 0;

    virtual bool hasNegativeSubpattern() const = 0;

    virtual bool negativeHasMinusSign() const = 0;

    virtual bool containsSymbolType(AffixPatternType, UErrorCode&) const = 0;

    /**
     * True if the pattern has a number placeholder like "0" or "#,##0.00"; false if the pattern does not
     * have one. This is used in cases like compact notation, where the pattern replaces the entire
     * number instead of rendering the number.
     */
    virtual bool hasBody() const = 0;
};


/**
 * A Modifier is an object that can be passed through the formatting pipeline until it is finally applied to the string
 * builder. A Modifier usually contains a prefix and a suffix that are applied, but it could contain something else,
 * like a {@link com.ibm.icu.text.SimpleFormatter} pattern.
 *
 * A Modifier is usually immutable, except in cases such as {@link MutablePatternModifier}, which are mutable for performance
 * reasons.
 *
 * Exported as U_I18N_API because it is a base class for other exported types
 */
class U_I18N_API Modifier {
  public:
    virtual ~Modifier();

    /**
     * Apply this Modifier to the string builder.
     *
     * @param output
     *            The string builder to which to apply this modifier.
     * @param leftIndex
     *            The left index of the string within the builder. Equal to 0 when only one number is being formatted.
     * @param rightIndex
     *            The right index of the string within the string builder. Equal to length when only one number is being
     *            formatted.
     * @return The number of characters (UTF-16 code units) that were added to the string builder.
     */
    virtual int32_t apply(FormattedStringBuilder& output, int leftIndex, int rightIndex,
                          UErrorCode& status) const = 0;

    /**
     * Gets the length of the prefix. This information can be used in combination with {@link #apply} to extract the
     * prefix and suffix strings.
     *
     * @return The number of characters (UTF-16 code units) in the prefix.
     */
    virtual int32_t getPrefixLength() const = 0;

    /**
     * Returns the number of code points in the modifier, prefix plus suffix.
     */
    virtual int32_t getCodePointCount() const = 0;

    /**
     * Whether this modifier is strong. If a modifier is strong, it should always be applied immediately and not allowed
     * to bubble up. With regard to padding, strong modifiers are considered to be on the inside of the prefix and
     * suffix.
     *
     * @return Whether the modifier is strong.
     */
    virtual bool isStrong() const = 0;

    /**
     * Whether the modifier contains at least one occurrence of the given field.
     */
    virtual bool containsField(UNumberFormatFields field) const = 0;

    /**
     * A fill-in for getParameters(). obj will always be set; if non-null, the other
     * two fields are also safe to read.
     */
    struct U_I18N_API Parameters {
        const ModifierStore* obj = nullptr;
        Signum signum;
        StandardPlural::Form plural;

        Parameters();
        Parameters(const ModifierStore* _obj, Signum _signum, StandardPlural::Form _plural);
    };

    /**
     * Gets a set of "parameters" for this Modifier.
     *
     * TODO: Make this return a `const Parameters*` more like Java?
     */
    virtual void getParameters(Parameters& output) const = 0;

    /**
     * Returns whether this Modifier is *semantically equivalent* to the other Modifier;
     * in many cases, this is the same as equal, but parameters should be ignored.
     */
    virtual bool semanticallyEquivalent(const Modifier& other) const = 0;
};


/**
 * This is *not* a modifier; rather, it is an object that can return modifiers
 * based on given parameters.
 *
 * Exported as U_I18N_API because it is a base class for other exported types.
 */
class U_I18N_API ModifierStore {
  public:
    virtual ~ModifierStore();

    /**
     * Returns a Modifier with the given parameters (best-effort).
     */
    virtual const Modifier* getModifier(Signum signum, StandardPlural::Form plural) const = 0;
};


/**
 * This interface is used when all number formatting settings, including the locale, are known, except for the quantity
 * itself. The {@link #processQuantity} method performs the final step in the number processing pipeline: it uses the
 * quantity to generate a finalized {@link MicroProps}, which can be used to render the number to output.
 *
 * <p>
 * In other words, this interface is used for the parts of number processing that are <em>quantity-dependent</em>.
 *
 * <p>
 * In order to allow for multiple different objects to all mutate the same MicroProps, a "chain" of MicroPropsGenerators
 * are linked together, and each one is responsible for manipulating a certain quantity-dependent part of the
 * MicroProps. At the tail of the linked list is a base instance of {@link MicroProps} with properties that are not
 * quantity-dependent. Each element in the linked list calls {@link #processQuantity} on its "parent", then does its
 * work, and then returns the result.
 *
 * Exported as U_I18N_API because it is a base class for other exported types
 *
 */
class U_I18N_API MicroPropsGenerator {
  public:
    virtual ~MicroPropsGenerator();

    /**
     * Considers the given {@link DecimalQuantity}, optionally mutates it, and returns a {@link MicroProps}.
     *
     * @param quantity
     *            The quantity for consideration and optional mutation.
     * @param micros
     *            The MicroProps instance to populate.
     * @return A MicroProps instance resolved for the quantity.
     */
    virtual void processQuantity(DecimalQuantity& quantity, MicroProps& micros,
                                 UErrorCode& status) const = 0;
};

/**
 * An interface used by compact notation and scientific notation to choose a multiplier while rounding.
 */
class MultiplierProducer {
  public:
    virtual ~MultiplierProducer();

    /**
     * Maps a magnitude to a multiplier in powers of ten. For example, in compact notation in English, a magnitude of 5
     * (e.g., 100,000) should return a multiplier of -3, since the number is displayed in thousands.
     *
     * @param magnitude
     *            The power of ten of the input number.
     * @return The shift in powers of ten.
     */
    virtual int32_t getMultiplier(int32_t magnitude) const = 0;
};

// Exported as U_I18N_API because it is a public member field of exported DecimalFormatProperties
template<typename T>
class U_I18N_API NullableValue {
  public:
    NullableValue()
            : fNull(true) {}

    NullableValue(const NullableValue<T>& other) = default;

    explicit NullableValue(const T& other) {
        fValue = other;
        fNull = false;
    }

    NullableValue<T>& operator=(const NullableValue<T>& other) {
        fNull = other.fNull;
        if (!fNull) {
            fValue = other.fValue;
        }
        return *this;
    }

    NullableValue<T>& operator=(const T& other) {
        fValue = other;
        fNull = false;
        return *this;
    }

    bool operator==(const NullableValue& other) const {
        // "fValue == other.fValue" returns UBool, not bool (causes compiler warnings)
        return fNull ? other.fNull : (other.fNull ? false : static_cast<bool>(fValue == other.fValue));
    }

    void nullify() {
        // TODO: It might be nice to call the destructor here.
        fNull = true;
    }

    bool isNull() const {
        return fNull;
    }

    T get(UErrorCode& status) const {
        if (fNull) {
            status = U_UNDEFINED_VARIABLE;
        }
        return fValue;
    }

    T getNoError() const {
        return fValue;
    }

    T getOrDefault(T defaultValue) const {
        return fNull ? defaultValue : fValue;
    }

  private:
    bool fNull;
    T fValue;
};


} // namespace impl
} // namespace number
U_NAMESPACE_END

#endif //__NUMBER_TYPES_H__


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2009-2015, International Business Machines Corporation and    *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 */



#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Currency Plural Information used by Decimal Format
 */

#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

class Locale;
class PluralRules;
class Hashtable;

/**
 * This class represents the information needed by 
 * DecimalFormat to format currency plural, 
 * such as "3.00 US dollars" or "1.00 US dollar". 
 * DecimalFormat creates for itself an instance of
 * CurrencyPluralInfo from its locale data.  
 * If you need to change any of these symbols, you can get the
 * CurrencyPluralInfo object from your 
 * DecimalFormat and modify it.
 *
 * Following are the information needed for currency plural format and parse:
 * locale information,
 * plural rule of the locale,
 * currency plural pattern of the locale.
 *
 * @stable ICU 4.2
 */
class  U_I18N_API CurrencyPluralInfo : public UObject {
public:

    /**
     * Create a CurrencyPluralInfo object for the default locale.
     * @param status output param set to success/failure code on exit
     * @stable ICU 4.2
     */
    CurrencyPluralInfo(UErrorCode& status);

    /**
     * Create a CurrencyPluralInfo object for the given locale.
     * @param locale the locale
     * @param status output param set to success/failure code on exit
     * @stable ICU 4.2
     */
    CurrencyPluralInfo(const Locale& locale, UErrorCode& status); 

    /**
     * Copy constructor
     *
     * @stable ICU 4.2
     */
    CurrencyPluralInfo(const CurrencyPluralInfo& info);


    /**
     * Assignment operator
     *
     * @stable ICU 4.2
     */
    CurrencyPluralInfo& operator=(const CurrencyPluralInfo& info);


    /**
     * Destructor
     *
     * @stable ICU 4.2
     */
    virtual ~CurrencyPluralInfo();


    /**
     * Equal operator.
     *
     * @stable ICU 4.2
     */
    UBool operator==(const CurrencyPluralInfo& info) const;


    /**
     * Not equal operator
     *
     * @stable ICU 4.2
     */
    UBool operator!=(const CurrencyPluralInfo& info) const;


    /**
     * Clone
     *
     * @stable ICU 4.2
     */
    CurrencyPluralInfo* clone() const;


    /**
     * Gets plural rules of this locale, used for currency plural format
     *
     * @return plural rule
     * @stable ICU 4.2
     */
    const PluralRules* getPluralRules() const;

    /**
     * Given a plural count, gets currency plural pattern of this locale, 
     * used for currency plural format
     *
     * @param  pluralCount currency plural count
     * @param  result      output param to receive the pattern
     * @return a currency plural pattern based on plural count
     * @stable ICU 4.2
     */
    UnicodeString& getCurrencyPluralPattern(const UnicodeString& pluralCount,
                                            UnicodeString& result) const; 

    /**
     * Get locale 
     *
     * @return locale
     * @stable ICU 4.2
     */
    const Locale& getLocale() const;

    /**
     * Set plural rules.
     * The plural rule is set when CurrencyPluralInfo
     * instance is created.
     * You can call this method to reset plural rules only if you want
     * to modify the default plural rule of the locale.
     *
     * @param ruleDescription new plural rule description
     * @param status output param set to success/failure code on exit
     * @stable ICU 4.2
     */
    void setPluralRules(const UnicodeString& ruleDescription,
                        UErrorCode& status);

    /**
     * Set currency plural pattern.
     * The currency plural pattern is set when CurrencyPluralInfo
     * instance is created.
     * You can call this method to reset currency plural pattern only if 
     * you want to modify the default currency plural pattern of the locale.
     *
     * @param pluralCount the plural count for which the currency pattern will 
     *                    be overridden.
     * @param pattern     the new currency plural pattern
     * @param status      output param set to success/failure code on exit
     * @stable ICU 4.2
     */
    void setCurrencyPluralPattern(const UnicodeString& pluralCount, 
                                  const UnicodeString& pattern,
                                  UErrorCode& status);

    /**
     * Set locale
     *
     * @param loc     the new locale to set
     * @param status  output param set to success/failure code on exit
     * @stable ICU 4.2
     */
    void setLocale(const Locale& loc, UErrorCode& status);

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 4.2
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 4.2
     */
    static UClassID U_EXPORT2 getStaticClassID();

private:
    friend class DecimalFormat;
    friend class DecimalFormatImpl;

    void initialize(const Locale& loc, UErrorCode& status);
   
    void setupCurrencyPluralPattern(const Locale& loc, UErrorCode& status);

    /*
     * delete hash table
     *
     * @param hTable  hash table to be deleted
     */
    void deleteHash(Hashtable* hTable);


    /*
     * initialize hash table
     *
     * @param status   output param set to success/failure code on exit
     * @return         hash table initialized
     */
    Hashtable* initHash(UErrorCode& status);



    /**
     * copy hash table
     *
     * @param source   the source to copy from
     * @param target   the target to copy to
     * @param status   error code
     */
    void copyHash(const Hashtable* source, Hashtable* target, UErrorCode& status);

    //-------------------- private data member ---------------------
    // map from plural count to currency plural pattern, for example
    // a plural pattern defined in "CurrencyUnitPatterns" is
    // "one{{0} {1}}", in which "one" is a plural count
    // and "{0} {1}" is a currency plural pattern".
    // The currency plural pattern saved in this mapping is the pattern
    // defined in "CurrencyUnitPattern" by replacing
    // {0} with the number format pattern,
    // and {1} with 3 currency sign.
    Hashtable* fPluralCountToCurrencyUnitPattern;

    /*
     * The plural rule is used to format currency plural name,
     * for example: "3.00 US Dollars".
     * If there are 3 currency signs in the currency pattern,
     * the 3 currency signs will be replaced by currency plural name.
     */
    PluralRules* fPluralRules;

    // locale
    Locale* fLocale;

private:
    /**
    * An internal status variable used to indicate that the object is in an 'invalid' state.
    * Used by copy constructor, the assignment operator and the clone method.
    */
    UErrorCode fInternalStatus;
};


inline UBool
CurrencyPluralInfo::operator!=(const CurrencyPluralInfo& info) const {
    return !operator==(info);
}  

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING


#include <cstdint>



// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

#include <cstdint>


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File PLURRULE_IMPL.H
*
*******************************************************************************
*/



// Internal definitions for the PluralRules implementation.



#if !UCONFIG_NO_FORMATTING










class PluralRulesTest;

U_NAMESPACE_BEGIN

class AndConstraint;
class RuleChain;
class Digiplurrule_token_tInterval;
class PluralRules;
class VisibleDigits;

// namespace pluralimpl {

// TODO: Remove this and replace with u"" literals. Was for EBCDIC compatibility.

static const UChar PLURRULE_DOT = ((UChar) 0x002E);
// static const UChar PLURRULE_SINGLE_QUOTE = ((UChar) 0x0027);
// static const UChar PLURRULE_SLASH = ((UChar) 0x002F);
// static const UChar PLURRULE_BACKSLASH = ((UChar) 0x005C);
static const UChar PLURRULE_SPACE = ((UChar) 0x0020);
static const UChar PLURRULE_EXCLAMATION = ((UChar) 0x0021);
// static const UChar PLURRULE_QUOTATION_MARK = ((UChar) 0x0022);
// static const UChar PLURRULE_NUMBER_SIGN = ((UChar) 0x0023);
static const UChar PLURRULE_PERCENT_SIGN = ((UChar) 0x0025);
// static const UChar PLURRULE_ASTERISK = ((UChar) 0x002A);
static const UChar PLURRULE_COMMA = ((UChar) 0x002C);
// static const UChar PLURRULE_HYPHEN = ((UChar) 0x002D);
static const UChar PLURRULE_U_ZERO = ((UChar) 0x0030);
// static const UChar PLURRULE_U_ONE = ((UChar) 0x0031);
// static const UChar PLURRULE_U_TWO = ((UChar) 0x0032);
// static const UChar PLURRULE_U_THREE = ((UChar) 0x0033);
// static const UChar PLURRULE_U_FOUR = ((UChar) 0x0034);
// static const UChar PLURRULE_U_FIVE = ((UChar) 0x0035);
// static const UChar PLURRULE_U_SIX = ((UChar) 0x0036);
// static const UChar PLURRULE_U_SEVEN = ((UChar) 0x0037);
// static const UChar PLURRULE_U_EIGHT = ((UChar) 0x0038);
static const UChar PLURRULE_U_NINE = ((UChar) 0x0039);
static const UChar PLURRULE_COLON = ((UChar) 0x003A);
static const UChar PLURRULE_SEMI_COLON = ((UChar) 0x003B);
static const UChar PLURRULE_EQUALS = ((UChar) 0x003D);
static const UChar PLURRULE_AT = ((UChar) 0x0040);
// static const UChar PLURRULE_CAP_A = ((UChar) 0x0041);
// static const UChar PLURRULE_CAP_B = ((UChar) 0x0042);
// static const UChar PLURRULE_CAP_R = ((UChar) 0x0052);
// static const UChar PLURRULE_CAP_Z = ((UChar) 0x005A);
// static const UChar PLURRULE_LOWLINE = ((UChar) 0x005F);
// static const UChar PLURRULE_LEFTBRACE = ((UChar) 0x007B);
// static const UChar PLURRULE_RIGHTBRACE = ((UChar) 0x007D);
static const UChar PLURRULE_TILDE = ((UChar) 0x007E);
static const UChar PLURRULE_ELLIPSIS = ((UChar) 0x2026);

static const UChar PLURRULE_LOW_A = ((UChar) 0x0061);
// static const UChar PLURRULE_LOW_B = ((UChar) 0x0062);
static const UChar PLURRULE_LOW_C = ((UChar) 0x0063);
static const UChar PLURRULE_LOW_D = ((UChar) 0x0064);
static const UChar PLURRULE_LOW_E = ((UChar) 0x0065);
static const UChar PLURRULE_LOW_F = ((UChar) 0x0066);
static const UChar PLURRULE_LOW_G = ((UChar) 0x0067);
static const UChar PLURRULE_LOW_H = ((UChar) 0x0068);
static const UChar PLURRULE_LOW_I = ((UChar) 0x0069);
// static const UChar PLURRULE_LOW_J = ((UChar) 0x006a);
// static const UChar PLURRULE_LOW_K = ((UChar) 0x006B);
static const UChar PLURRULE_LOW_L = ((UChar) 0x006C);
static const UChar PLURRULE_LOW_M = ((UChar) 0x006D);
static const UChar PLURRULE_LOW_N = ((UChar) 0x006E);
static const UChar PLURRULE_LOW_O = ((UChar) 0x006F);
// static const UChar PLURRULE_LOW_P = ((UChar) 0x0070);
// static const UChar PLURRULE_LOW_Q = ((UChar) 0x0071);
static const UChar PLURRULE_LOW_R = ((UChar) 0x0072);
static const UChar PLURRULE_LOW_S = ((UChar) 0x0073);
static const UChar PLURRULE_LOW_T = ((UChar) 0x0074);
// static const UChar PLURRULE_LOW_U = ((UChar) 0x0075);
static const UChar PLURRULE_LOW_V = ((UChar) 0x0076);
static const UChar PLURRULE_LOW_W = ((UChar) 0x0077);
// static const UChar PLURRULE_LOW_Y = ((UChar) 0x0079);
static const UChar PLURRULE_LOW_Z = ((UChar) 0x007A);

// }


static const int32_t PLURAL_RANGE_HIGH = 0x7fffffff;

enum tokenType {
  plurrule_token_none,
  plurrule_token_tNumber,
  plurrule_token_tComma,
  plurrule_token_tSemiColon,
  plurrule_token_tSpace,
  plurrule_token_tColon,
  plurrule_token_tAt,           // '@'
  plurrule_token_tDot,
  plurrule_token_tDot2,
  plurrule_token_tEllipsis,
  plurrule_token_tKeyword,
  plurrule_token_tAnd,
  plurrule_token_tOr,
  plurrule_token_tMod,          // 'mod' or '%'
  plurrule_token_tNot,          //  'not' only.
  plurrule_token_tIn,           //  'in'  only.
  plurrule_token_tEqual,        //  '='   only.
  plurrule_token_tNotEqual,     //  '!='
  plurrule_token_tTilde,
  plurrule_token_tWithin,
  plurrule_token_tIs,
  plurrule_token_tVariableN,
  plurrule_token_tVariableI,
  plurrule_token_tVariableF,
  plurrule_token_tVariableV,
  plurrule_token_tVariableT,
  plurrule_token_tDecimal,
  plurrule_token_tInteger,
  plurrule_token_tEOF
};


class PluralRuleParser: public UMemory {
public:
    PluralRuleParser();
    virtual ~PluralRuleParser();

    void parse(const UnicodeString &rules, PluralRules *dest, UErrorCode &status);
    void getNextToken(UErrorCode &status);
    void checkSyntax(UErrorCode &status);
    static int32_t getNumberValue(const UnicodeString &token);

private:
    static tokenType getKeyType(const UnicodeString& token, tokenType type);
    static tokenType charType(UChar ch);
    static UBool isValidKeyword(const UnicodeString& token);

    const UnicodeString  *ruleSrc;  // The rules string.
    int32_t        ruleIndex;       // String index in the input rules, the current parse position.
    UnicodeString  token;           // Token most recently scanned.
    tokenType      type;
    tokenType      prevType;

                                    // The items currently being parsed & built.
                                    // Note: currentChain may not be the last RuleChain in the
                                    //       list because the "other" chain is forced to the end.
    AndConstraint *curAndConstraint;
    RuleChain     *currentChain;

    int32_t        rangeLowIdx;     // Indices in the UVector of ranges of the
    int32_t        rangeHiIdx;      //    low and hi values currently being parsed.

    enum EParseState {
       kKeyword,
       kExpr,
       kValue,
       kRangeList,
       kSamples
    };
};

enum PluralOperand {
    /**
    * The double value of the entire number.
    */
    PLURAL_OPERAND_N,

    /**
     * The integer value, with the fraction digits truncated off.
     */
    PLURAL_OPERAND_I,

    /**
     * All visible fraction digits as an integer, including trailing zeros.
     */
    PLURAL_OPERAND_F,

    /**
     * Visible fraction digits as an integer, not including trailing zeros.
     */
    PLURAL_OPERAND_T,

    /**
     * Number of visible fraction digits.
     */
    PLURAL_OPERAND_V,

    /**
     * Number of visible fraction digits, not including trailing zeros.
     */
    PLURAL_OPERAND_W,

    /**
     * THIS OPERAND IS DEPRECATED AND HAS BEEN REMOVED FROM THE SPEC.
     *
     * <p>Returns the integer value, but will fail if the number has fraction digits.
     * That is, using "j" instead of "i" is like implicitly adding "v is 0".
     *
     * <p>For example, "j is 3" is equivalent to "i is 3 and v is 0": it matches
     * "3" but not "3.1" or "3.0".
     */
    PLURAL_OPERAND_J
};

/**
 * Converts from the tokenType enum to PluralOperand. Asserts that the given
 * tokenType can be mapped to a PluralOperand.
 */
PluralOperand tokenTypeToPluralOperand(tokenType tt);

/**
 * An interface to FixedDecimal, allowing for other implementations.
 * @internal
 */
class U_I18N_API IFixedDecimal {
  public:
    virtual ~IFixedDecimal();

    /**
     * Returns the value corresponding to the specified operand (n, i, f, t, v, or w).
     * If the operand is 'n', returns a double; otherwise, returns an integer.
     */
    virtual double getPluralOperand(PluralOperand operand) const = 0;

    virtual bool isNaN() const = 0;

    virtual bool isInfinite() const = 0;

    /** Whether the number has no nonzero fraction digits. */
    virtual bool hasIntegerValue() const = 0;
};

/**
 * class FixedDecimal serves to communicate the properties
 * of a formatted number from a decimal formatter to PluralRules::select()
 *
 * see DecimalFormat::getFixedDecimal()
 * @internal
 */
class U_I18N_API FixedDecimal: public IFixedDecimal, public UObject {
  public:
    /**
      * @param n   the number, e.g. 12.345
      * @param v   The number of visible fraction digits, e.g. 3
      * @param f   The fraction digits, e.g. 345
      */
    FixedDecimal(double  n, int32_t v, int64_t f);
    FixedDecimal(double n, int32_t);
    explicit FixedDecimal(double n);
    FixedDecimal();
    ~FixedDecimal() U_OVERRIDE;
    FixedDecimal(const UnicodeString &s, UErrorCode &ec);
    FixedDecimal(const FixedDecimal &other);

    double getPluralOperand(PluralOperand operand) const U_OVERRIDE;
    bool isNaN() const U_OVERRIDE;
    bool isInfinite() const U_OVERRIDE;
    bool hasIntegerValue() const U_OVERRIDE;

    bool isNanOrInfinity() const;  // used in decimfmtimpl.cpp

    int32_t getVisibleFractionDigitCount() const;

    void init(double n, int32_t v, int64_t f);
    void init(double n);
    UBool quickInit(double n);  // Try a fast-path only initialization,
                                //    return TRUE if successful.
    void adjustForMinFractionDigits(int32_t min);
    static int64_t getFractionalDigits(double n, int32_t v);
    static int32_t decimals(double n);

    double      source;
    int32_t     visibleDecimalDigitCount;
    int64_t     decimalDigits;
    int64_t     decimalDigitsWithoutTrailingZeros;
    int64_t     intValue;
    UBool       _hasIntegerValue;
    UBool       isNegative;
    UBool       _isNaN;
    UBool       _isInfinite;
};

class AndConstraint : public UMemory  {
public:
    typedef enum RuleOp {
        NONE,
        MOD
    } RuleOp;
    RuleOp op = AndConstraint::NONE;
    int32_t opNum = -1;             // for mod expressions, the right operand of the mod.
    int32_t value = -1;             // valid for 'is' rules only.
    UVector32 *rangeList = nullptr; // for 'in', 'within' rules. Null otherwise.
    UBool negated = FALSE;          // TRUE for negated rules.
    UBool integerOnly = FALSE;      // TRUE for 'within' rules.
    tokenType digitsType = plurrule_token_none;    // n | i | v | f constraint.
    AndConstraint *next = nullptr;
    // Internal error status, used for errors that occur during the copy constructor.
    UErrorCode fInternalStatus = U_ZERO_ERROR;

    AndConstraint() = default;
    AndConstraint(const AndConstraint& other);
    virtual ~AndConstraint();
    AndConstraint* add(UErrorCode& status);
    // UBool isFulfilled(double number);
    UBool isFulfilled(const IFixedDecimal &number);
};

class OrConstraint : public UMemory  {
public:
    AndConstraint *childNode = nullptr;
    OrConstraint *next = nullptr;
    // Internal error status, used for errors that occur during the copy constructor.
    UErrorCode fInternalStatus = U_ZERO_ERROR;

    OrConstraint() = default;
    OrConstraint(const OrConstraint& other);
    virtual ~OrConstraint();
    AndConstraint* add(UErrorCode& status);
    // UBool isFulfilled(double number);
    UBool isFulfilled(const IFixedDecimal &number);
};

class RuleChain : public UMemory  {
public:
    UnicodeString   fKeyword;
    RuleChain      *fNext = nullptr;
    OrConstraint   *ruleHeader = nullptr;
    UnicodeString   fDecimalSamples;  // Samples strings from rule source
    UnicodeString   fIntegerSamples;  //   without @decimal or @integer, otherwise unprocessed.
    UBool           fDecimalSamplesUnbounded = FALSE;
    UBool           fIntegerSamplesUnbounded = FALSE;
    // Internal error status, used for errors that occur during the copy constructor.
    UErrorCode      fInternalStatus = U_ZERO_ERROR;

    RuleChain() = default;
    RuleChain(const RuleChain& other);
    virtual ~RuleChain();

    UnicodeString select(const IFixedDecimal &number) const;
    void          dumpRules(UnicodeString& result);
    UErrorCode    getKeywords(int32_t maxArraySize, UnicodeString *keywords, int32_t& arraySize) const;
    UBool         isKeyword(const UnicodeString& keyword) const;
};

class PluralKeywordEnumeration : public StringEnumeration {
public:
    PluralKeywordEnumeration(RuleChain *header, UErrorCode& status);
    virtual ~PluralKeywordEnumeration();
    static UClassID U_EXPORT2 getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
    virtual const UnicodeString* snext(UErrorCode& status);
    virtual void reset(UErrorCode& status);
    virtual int32_t count(UErrorCode& status) const;
private:
    int32_t         pos;
    UVector         fKeywordNames;
};


class U_I18N_API PluralAvailableLocalesEnumeration: public StringEnumeration {
  public:
    PluralAvailableLocalesEnumeration(UErrorCode &status);
    virtual ~PluralAvailableLocalesEnumeration();
    virtual const char* next(int32_t *resultLength, UErrorCode& status);
    virtual void reset(UErrorCode& status);
    virtual int32_t count(UErrorCode& status) const;
  private:
    UErrorCode      fOpenStatus;
    UResourceBundle *fLocales = nullptr;
    UResourceBundle *fRes = nullptr;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */




U_NAMESPACE_BEGIN namespace number {
namespace impl {

// Forward-declare (maybe don't want number_utils.h included here):
class DecNum;

/**
 * An class for representing a number to be processed by the decimal formatting pipeline. Includes
 * methods for rounding, plural rules, and decimal digit extraction.
 *
 * <p>By design, this is NOT IMMUTABLE and NOT THREAD SAFE. It is intended to be an intermediate
 * object holding state during a pass through the decimal formatting pipeline.
 *
 * <p>Represents numbers and digit display properties using Binary Coded Decimal (BCD).
 *
 * <p>Java has multiple implementations for testing, but C++ has only one implementation.
 */
class U_I18N_API DecimalQuantity : public IFixedDecimal, public UMemory {
  public:
    /** Copy constructor. */
    DecimalQuantity(const DecimalQuantity &other);

    /** Move constructor. */
    DecimalQuantity(DecimalQuantity &&src) U_NOEXCEPT;

    DecimalQuantity();

    ~DecimalQuantity() override;

    /**
     * Sets this instance to be equal to another instance.
     *
     * @param other The instance to copy from.
     */
    DecimalQuantity &operator=(const DecimalQuantity &other);

    /** Move assignment */
    DecimalQuantity &operator=(DecimalQuantity&& src) U_NOEXCEPT;

    /**
     * Sets the minimum integer digits that this {@link DecimalQuantity} should generate.
     * This method does not perform rounding.
     *
     * @param minInt The minimum number of integer digits.
     */
    void setMinInteger(int32_t minInt);

    /**
     * Sets the minimum fraction digits that this {@link DecimalQuantity} should generate.
     * This method does not perform rounding.
     *
     * @param minFrac The minimum number of fraction digits.
     */
    void setMinFraction(int32_t minFrac);

    /**
     * Truncates digits from the upper magnitude of the number in order to satisfy the
     * specified maximum number of integer digits.
     *
     * @param maxInt The maximum number of integer digits.
     */
    void applyMaxInteger(int32_t maxInt);

    /**
     * Rounds the number to a specified interval, such as 0.05.
     *
     * <p>If rounding to a power of ten, use the more efficient {@link #roundToMagnitude} instead.
     *
     * @param roundingIncrement The increment to which to round.
     * @param roundingMode The {@link RoundingMode} to use if rounding is necessary.
     */
    void roundToIncrement(double roundingIncrement, RoundingMode roundingMode,
                          UErrorCode& status);

    /** Removes all fraction digits. */
    void truncate();

    /**
     * Rounds the number to the nearest multiple of 5 at the specified magnitude.
     * For example, when magnitude == -2, this performs rounding to the nearest 0.05.
     *
     * @param magnitude The magnitude at which the digit should become either 0 or 5.
     * @param roundingMode Rounding strategy.
     */
    void roundToNickel(int32_t magnitude, RoundingMode roundingMode, UErrorCode& status);

    /**
     * Rounds the number to a specified magnitude (power of ten).
     *
     * @param roundingMagnitude The power of ten to which to round. For example, a value of -2 will
     *     round to 2 decimal places.
     * @param roundingMode The {@link RoundingMode} to use if rounding is necessary.
     */
    void roundToMagnitude(int32_t magnitude, RoundingMode roundingMode, UErrorCode& status);

    /**
     * Rounds the number to an infinite number of decimal points. This has no effect except for
     * forcing the double in {@link DecimalQuantity_AbstractBCD} to adopt its exact representation.
     */
    void roundToInfinity();

    /**
     * Multiply the internal value. Uses decNumber.
     *
     * @param multiplicand The value by which to multiply.
     */
    void multiplyBy(const DecNum& multiplicand, UErrorCode& status);

    /**
     * Divide the internal value. Uses decNumber.
     *
     * @param multiplicand The value by which to multiply.
     */
    void divideBy(const DecNum& divisor, UErrorCode& status);

    /** Flips the sign from positive to negative and back. */
    void negate();

    /**
     * Scales the number by a power of ten. For example, if the value is currently "1234.56", calling
     * this method with delta=-3 will change the value to "1.23456".
     *
     * @param delta The number of magnitudes of ten to change by.
     * @return true if integer overflow occured; false otherwise.
     */
    bool adjustMagnitude(int32_t delta);

    /**
     * @return The power of ten corresponding to the most significant nonzero digit.
     * The number must not be zero.
     */
    int32_t getMagnitude() const;

    /**
     * @return Whether the value represented by this {@link DecimalQuantity} is
     * zero, infinity, or NaN.
     */
    bool isZeroish() const;

    /** @return Whether the value represented by this {@link DecimalQuantity} is less than zero. */
    bool isNegative() const;

    /** @return The appropriate value from the Signum enum. */
    Signum signum() const;

    /** @return Whether the value represented by this {@link DecimalQuantity} is infinite. */
    bool isInfinite() const U_OVERRIDE;

    /** @return Whether the value represented by this {@link DecimalQuantity} is not a number. */
    bool isNaN() const U_OVERRIDE;

    /** @param truncateIfOverflow if false and the number does NOT fit, fails with an assertion error. */
    int64_t toLong(bool truncateIfOverflow = false) const;

    uint64_t toFractionLong(bool includeTrailingZeros) const;

    /**
     * Returns whether or not a Long can fully represent the value stored in this DecimalQuantity.
     * @param ignoreFraction if true, silently ignore digits after the decimal place.
     */
    bool fitsInLong(bool ignoreFraction = false) const;

    /** @return The value contained in this {@link DecimalQuantity} approximated as a double. */
    double toDouble() const;

    /** Computes a DecNum representation of this DecimalQuantity, saving it to the output parameter. */
    void toDecNum(DecNum& output, UErrorCode& status) const;

    DecimalQuantity &setToInt(int32_t n);

    DecimalQuantity &setToLong(int64_t n);

    DecimalQuantity &setToDouble(double n);

    /** decNumber is similar to BigDecimal in Java. */
    DecimalQuantity &setToDecNumber(StringPiece n, UErrorCode& status);

    /** Internal method if the caller already has a DecNum. */
    DecimalQuantity &setToDecNum(const DecNum& n, UErrorCode& status);

    /**
     * Appends a digit, optionally with one or more leading zeros, to the end of the value represented
     * by this DecimalQuantity.
     *
     * <p>The primary use of this method is to construct numbers during a parsing loop. It allows
     * parsing to take advantage of the digit list infrastructure primarily designed for formatting.
     *
     * @param value The digit to append.
     * @param leadingZeros The number of zeros to append before the digit. For example, if the value
     *     in this instance starts as 12.3, and you append a 4 with 1 leading zero, the value becomes
     *     12.304.
     * @param appendAsInteger If true, increase the magnitude of existing digits to make room for the
     *     new digit. If false, append to the end like a fraction digit. If true, there must not be
     *     any fraction digits already in the number.
     * @internal
     * @deprecated This API is ICU internal only.
     */
    void appendDigit(int8_t value, int32_t leadingZeros, bool appendAsInteger);

    double getPluralOperand(PluralOperand operand) const U_OVERRIDE;

    bool hasIntegerValue() const U_OVERRIDE;

    /**
     * Gets the digit at the specified magnitude. For example, if the represented number is 12.3,
     * getDigit(-1) returns 3, since 3 is the digit corresponding to 10^-1.
     *
     * @param magnitude The magnitude of the digit.
     * @return The digit at the specified magnitude.
     */
    int8_t getDigit(int32_t magnitude) const;

    /**
     * Gets the largest power of ten that needs to be displayed. The value returned by this function
     * will be bounded between minInt and maxInt.
     *
     * @return The highest-magnitude digit to be displayed.
     */
    int32_t getUpperDisplayMagnitude() const;

    /**
     * Gets the smallest power of ten that needs to be displayed. The value returned by this function
     * will be bounded between -minFrac and -maxFrac.
     *
     * @return The lowest-magnitude digit to be displayed.
     */
    int32_t getLowerDisplayMagnitude() const;

    int32_t fractionCount() const;

    int32_t fractionCountWithoutTrailingZeros() const;

    void clear();

    /** This method is for internal testing only. */
    uint64_t getPositionFingerprint() const;

//    /**
//     * If the given {@link FieldPosition} is a {@link UFieldPosition}, populates it with the fraction
//     * length and fraction long value. If the argument is not a {@link UFieldPosition}, nothing
//     * happens.
//     *
//     * @param fp The {@link UFieldPosition} to populate.
//     */
//    void populateUFieldPosition(FieldPosition fp);

    /**
     * Checks whether the bytes stored in this instance are all valid. For internal unit testing only.
     *
     * @return An error message if this instance is invalid, or null if this instance is healthy.
     */
    const char16_t* checkHealth() const;

    UnicodeString toString() const;

    /** Returns the string in standard exponential notation. */
    UnicodeString toScientificString() const;

    /** Returns the string without exponential notation. Slightly slower than toScientificString(). */
    UnicodeString toPlainString() const;

    /** Visible for testing */
    inline bool isUsingBytes() { return usingBytes; }

    /** Visible for testing */
    inline bool isExplicitExactDouble() { return explicitExactDouble; }

    bool operator==(const DecimalQuantity& other) const;

    inline bool operator!=(const DecimalQuantity& other) const {
        return !(*this == other);
    }

    /**
     * Bogus flag for when a DecimalQuantity is stored on the stack.
     */
    bool bogus = false;

  private:
    /**
     * The power of ten corresponding to the least significant digit in the BCD. For example, if this
     * object represents the number "3.14", the BCD will be "0x314" and the scale will be -2.
     *
     * <p>Note that in {@link java.math.BigDecimal}, the scale is defined differently: the number of
     * digits after the decimal place, which is the negative of our definition of scale.
     */
    int32_t scale;

    /**
     * The number of digits in the BCD. For example, "1007" has BCD "0x1007" and precision 4. The
     * maximum precision is 16 since a long can hold only 16 digits.
     *
     * <p>This value must be re-calculated whenever the value in bcd changes by using {@link
     * #computePrecisionAndCompact()}.
     */
    int32_t precision;

    /**
     * A bitmask of properties relating to the number represented by this object.
     *
     * @see #NEGATIVE_FLAG
     * @see #INFINITY_FLAG
     * @see #NAN_FLAG
     */
    int8_t flags;

    // The following three fields relate to the double-to-ascii fast path algorithm.
    // When a double is given to DecimalQuantityBCD, it is converted to using a fast algorithm. The
    // fast algorithm guarantees correctness to only the first ~12 digits of the double. The process
    // of rounding the number ensures that the converted digits are correct, falling back to a slow-
    // path algorithm if required.  Therefore, if a DecimalQuantity is constructed from a double, it
    // is *required* that roundToMagnitude(), roundToIncrement(), or roundToInfinity() is called. If
    // you don't round, assertions will fail in certain other methods if you try calling them.

    /**
     * Whether the value in the BCD comes from the double fast path without having been rounded to
     * ensure correctness
     */
    UBool isApproximate;

    /**
     * The original number provided by the user and which is represented in BCD. Used when we need to
     * re-compute the BCD for an exact double representation.
     */
    double origDouble;

    /**
     * The change in magnitude relative to the original double. Used when we need to re-compute the
     * BCD for an exact double representation.
     */
    int32_t origDelta;

    // Positions to keep track of leading and trailing zeros.
    // lReqPos is the magnitude of the first required leading zero.
    // rReqPos is the magnitude of the last required trailing zero.
    int32_t lReqPos = 0;
    int32_t rReqPos = 0;

    /**
     * The BCD of the 16 digits of the number represented by this object. Every 4 bits of the long map
     * to one digit. For example, the number "12345" in BCD is "0x12345".
     *
     * <p>Whenever bcd changes internally, {@link #compact()} must be called, except in special cases
     * like setting the digit to zero.
     */
    union {
        struct {
            int8_t *ptr;
            int32_t len;
        } bcdBytes;
        uint64_t bcdLong;
    } fBCD;

    bool usingBytes = false;

    /**
     * Whether this {@link DecimalQuantity} has been explicitly converted to an exact double. true if
     * backed by a double that was explicitly converted via convertToAccurateDouble; false otherwise.
     * Used for testing.
     */
    bool explicitExactDouble = false;

    void roundToMagnitude(int32_t magnitude, RoundingMode roundingMode, bool nickel, UErrorCode& status);

    /**
     * Returns a single digit from the BCD list. No internal state is changed by calling this method.
     *
     * @param position The position of the digit to pop, counted in BCD units from the least
     *     significant digit. If outside the range supported by the implementation, zero is returned.
     * @return The digit at the specified location.
     */
    int8_t getDigitPos(int32_t position) const;

    /**
     * Sets the digit in the BCD list. This method only sets the digit; it is the caller's
     * responsibility to call {@link #compact} after setting the digit.
     *
     * @param position The position of the digit to pop, counted in BCD units from the least
     *     significant digit. If outside the range supported by the implementation, an AssertionError
     *     is thrown.
     * @param value The digit to set at the specified location.
     */
    void setDigitPos(int32_t position, int8_t value);

    /**
     * Adds zeros to the end of the BCD list. This will result in an invalid BCD representation; it is
     * the caller's responsibility to do further manipulation and then call {@link #compact}.
     *
     * @param numDigits The number of zeros to add.
     */
    void shiftLeft(int32_t numDigits);

    /**
     * Directly removes digits from the end of the BCD list.
     * Updates the scale and precision.
     *
     * CAUTION: it is the caller's responsibility to call {@link #compact} after this method.
     */
    void shiftRight(int32_t numDigits);

    /**
     * Directly removes digits from the front of the BCD list.
     * Updates precision.
     *
     * CAUTION: it is the caller's responsibility to call {@link #compact} after this method.
     */
    void popFromLeft(int32_t numDigits);

    /**
     * Sets the internal representation to zero. Clears any values stored in scale, precision,
     * hasDouble, origDouble, origDelta, and BCD data.
     */
    void setBcdToZero();

    /**
     * Sets the internal BCD state to represent the value in the given int. The int is guaranteed to
     * be either positive. The internal state is guaranteed to be empty when this method is called.
     *
     * @param n The value to consume.
     */
    void readIntToBcd(int32_t n);

    /**
     * Sets the internal BCD state to represent the value in the given long. The long is guaranteed to
     * be either positive. The internal state is guaranteed to be empty when this method is called.
     *
     * @param n The value to consume.
     */
    void readLongToBcd(int64_t n);

    void readDecNumberToBcd(const DecNum& dn);

    void readDoubleConversionToBcd(const char* buffer, int32_t length, int32_t point);

    void copyFieldsFrom(const DecimalQuantity& other);

    void copyBcdFrom(const DecimalQuantity &other);

    void moveBcdFrom(DecimalQuantity& src);

    /**
     * Removes trailing zeros from the BCD (adjusting the scale as required) and then computes the
     * precision. The precision is the number of digits in the number up through the greatest nonzero
     * digit.
     *
     * <p>This method must always be called when bcd changes in order for assumptions to be correct in
     * methods like {@link #fractionCount()}.
     */
    void compact();

    void _setToInt(int32_t n);

    void _setToLong(int64_t n);

    void _setToDoubleFast(double n);

    void _setToDecNum(const DecNum& dn, UErrorCode& status);

    void convertToAccurateDouble();

    /** Ensure that a byte array of at least 40 digits is allocated. */
    void ensureCapacity();

    void ensureCapacity(int32_t capacity);

    /** Switches the internal storage mechanism between the 64-bit long and the byte array. */
    void switchStorage();
};

} // namespace impl
} // namespace number
U_NAMESPACE_END


#endif //__NUMBER_DECIMALQUANTITY_H__


// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING


#include <cstdint>






U_NAMESPACE_BEGIN

// Export an explicit template instantiation of the LocalPointer that is used as a
// data member of CurrencyPluralInfoWrapper.
// (When building DLLs for Windows this is required.)
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN
#if defined(_MSC_VER)
// Ignore warning 4661 as LocalPointerBase does not use operator== or operator!=
#pragma warning(push)
#pragma warning(disable: 4661)
#endif
template class U_I18N_API LocalPointerBase<CurrencyPluralInfo>;
template class U_I18N_API LocalPointer<CurrencyPluralInfo>;
#if defined(_MSC_VER)
#pragma warning(pop)
#endif
#endif

namespace number {
namespace impl {

// Exported as U_I18N_API because it is a public member field of exported DecimalFormatProperties
// Using this wrapper is rather unfortunate, but is needed on Windows platforms in order to allow
// for DLL-exporting an fully specified template instantiation.
class U_I18N_API CurrencyPluralInfoWrapper {
public:
    LocalPointer<CurrencyPluralInfo> fPtr;

    CurrencyPluralInfoWrapper() = default;

    CurrencyPluralInfoWrapper(const CurrencyPluralInfoWrapper& other) {
        if (!other.fPtr.isNull()) {
            fPtr.adoptInstead(new CurrencyPluralInfo(*other.fPtr));
        }
    }

    CurrencyPluralInfoWrapper& operator=(const CurrencyPluralInfoWrapper& other) {
        if (!other.fPtr.isNull()) {
            fPtr.adoptInstead(new CurrencyPluralInfo(*other.fPtr));
        }
        return *this;
    }
};

/** Controls the set of rules for parsing a string from the old DecimalFormat API. */
enum ParseMode {
    /**
     * Lenient mode should be used if you want to accept malformed user input. It will use heuristics
     * to attempt to parse through typographical errors in the string.
     */
            PARSE_MODE_LENIENT,

    /**
     * Strict mode should be used if you want to require that the input is well-formed. More
     * specifically, it differs from lenient mode in the following ways:
     *
     * <ul>
     * <li>Grouping widths must match the grouping settings. For example, "12,3,45" will fail if the
     * grouping width is 3, as in the pattern "#,##0".
     * <li>The string must contain a complete prefix and suffix. For example, if the pattern is
     * "{#};(#)", then "{123}" or "(123)" would match, but "{123", "123}", and "123" would all fail.
     * (The latter strings would be accepted in lenient mode.)
     * <li>Whitespace may not appear at arbitrary places in the string. In lenient mode, whitespace
     * is allowed to occur arbitrarily before and after prefixes and exponent separators.
     * <li>Leading grouping separators are not allowed, as in ",123".
     * <li>Minus and plus signs can only appear if specified in the pattern. In lenient mode, a plus
     * or minus sign can always precede a number.
     * <li>The set of characters that can be interpreted as a decimal or grouping separator is
     * smaller.
     * <li><strong>If currency parsing is enabled,</strong> currencies must only appear where
     * specified in either the current pattern string or in a valid pattern string for the current
     * locale. For example, if the pattern is "¤0.00", then "$1.23" would match, but "1.23$" would
     * fail to match.
     * </ul>
     */
            PARSE_MODE_STRICT,
};

// Exported as U_I18N_API because it is needed for the unit test PatternStringTest
struct U_I18N_API DecimalFormatProperties : public UMemory {

  public:
    NullableValue<UNumberCompactStyle> compactStyle;
    NullableValue<CurrencyUnit> currency;
    CurrencyPluralInfoWrapper currencyPluralInfo;
    NullableValue<UCurrencyUsage> currencyUsage;
    bool decimalPatternMatchRequired;
    bool decimalSeparatorAlwaysShown;
    bool exponentSignAlwaysShown;
    bool formatFailIfMoreThanMaxDigits; // ICU4C-only
    int32_t formatWidth;
    int32_t groupingSize;
    bool groupingUsed;
    int32_t magnitudeMultiplier; // internal field like multiplierScale but separate to avoid conflict
    int32_t maximumFractionDigits;
    int32_t maximumIntegerDigits;
    int32_t maximumSignificantDigits;
    int32_t minimumExponentDigits;
    int32_t minimumFractionDigits;
    int32_t minimumGroupingDigits;
    int32_t minimumIntegerDigits;
    int32_t minimumSignificantDigits;
    int32_t multiplier;
    int32_t multiplierScale; // ICU4C-only
    UnicodeString negativePrefix;
    UnicodeString negativePrefixPattern;
    UnicodeString negativeSuffix;
    UnicodeString negativeSuffixPattern;
    NullableValue<PadPosition> padPosition;
    UnicodeString padString;
    bool parseCaseSensitive;
    bool parseIntegerOnly;
    NullableValue<ParseMode> parseMode;
    bool parseNoExponent;
    bool parseToBigDecimal; // TODO: Not needed in ICU4C?
    UNumberFormatAttributeValue parseAllInput; // ICU4C-only
    //PluralRules pluralRules;
    UnicodeString positivePrefix;
    UnicodeString positivePrefixPattern;
    UnicodeString positiveSuffix;
    UnicodeString positiveSuffixPattern;
    double roundingIncrement;
    NullableValue<RoundingMode> roundingMode;
    int32_t secondaryGroupingSize;
    bool signAlwaysShown;

    DecimalFormatProperties();

    inline bool operator==(const DecimalFormatProperties& other) const {
        return _equals(other, false);
    }

    void clear();

    /**
     * Checks for equality to the default DecimalFormatProperties, but ignores the prescribed set of
     * options for fast-path formatting.
     */
    bool equalsDefaultExceptFastFormat() const;

    /**
     * Returns the default DecimalFormatProperties instance.
     */
    static const DecimalFormatProperties& getDefault();

  private:
    bool _equals(const DecimalFormatProperties& other, bool ignoreForFastFormat) const;
};

} // namespace impl
} // namespace number
U_NAMESPACE_END


#endif //__NUMBER_DECIMFMTPROPS_H__


// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

#include <cstdint>






U_NAMESPACE_BEGIN namespace number {
namespace impl {

enum AffixPatternState {
    STATE_BASE = 0,
    STATE_FIRST_QUOTE = 1,
    STATE_INSIDE_QUOTE = 2,
    STATE_AFTER_QUOTE = 3,
    STATE_FIRST_CURR = 4,
    STATE_SECOND_CURR = 5,
    STATE_THIRD_CURR = 6,
    STATE_FOURTH_CURR = 7,
    STATE_FIFTH_CURR = 8,
    STATE_OVERFLOW_CURR = 9
};

// enum AffixPatternType defined in internals.h

struct AffixTag {
    int32_t offset;
    UChar32 codePoint;
    AffixPatternState state;
    AffixPatternType type;

    AffixTag()
            : offset(0), state(STATE_BASE) {}

    AffixTag(int32_t offset)
            : offset(offset) {}

    AffixTag(int32_t offset, UChar32 codePoint, AffixPatternState state, AffixPatternType type)
            : offset(offset), codePoint(codePoint), state(state), type(type) {}
};

class TokenConsumer {
  public:
    virtual ~TokenConsumer();

    virtual void consumeToken(AffixPatternType type, UChar32 cp, UErrorCode& status) = 0;
};

// Exported as U_I18N_API because it is a base class for other exported types
class U_I18N_API SymbolProvider {
  public:
    virtual ~SymbolProvider();

    // TODO: Could this be more efficient if it returned by reference?
    virtual UnicodeString getSymbol(AffixPatternType type) const = 0;
};

/**
 * Performs manipulations on affix patterns: the prefix and suffix strings associated with a decimal
 * format pattern. For example:
 *
 * <table>
 * <tr><th>Affix Pattern</th><th>Example Unescaped (Formatted) String</th></tr>
 * <tr><td>abc</td><td>abc</td></tr>
 * <tr><td>ab-</td><td>ab−</td></tr>
 * <tr><td>ab'-'</td><td>ab-</td></tr>
 * <tr><td>ab''</td><td>ab'</td></tr>
 * </table>
 *
 * To manually iterate over tokens in a literal string, use the following pattern, which is designed
 * to be efficient.
 *
 * <pre>
 * long tag = 0L;
 * while (AffixPatternUtils.hasNext(tag, patternString)) {
 *   tag = AffixPatternUtils.nextToken(tag, patternString);
 *   int typeOrCp = AffixPatternUtils.getTypeOrCp(tag);
 *   switch (typeOrCp) {
 *     case AffixPatternUtils.TYPE_MINUS_SIGN:
 *       // Current token is a minus sign.
 *       break;
 *     case AffixPatternUtils.TYPE_PLUS_SIGN:
 *       // Current token is a plus sign.
 *       break;
 *     case AffixPatternUtils.TYPE_PERCENT:
 *       // Current token is a percent sign.
 *       break;
 *     // ... other types ...
 *     default:
 *       // Current token is an arbitrary code point.
 *       // The variable typeOrCp is the code point.
 *       break;
 *   }
 * }
 * </pre>
 */
class U_I18N_API AffixUtils {

  public:

    /**
     * Estimates the number of code points present in an unescaped version of the affix pattern string
     * (one that would be returned by {@link #unescape}), assuming that all interpolated symbols
     * consume one code point and that currencies consume as many code points as their symbol width.
     * Used for computing padding width.
     *
     * @param patternString The original string whose width will be estimated.
     * @return The length of the unescaped string.
     */
    static int32_t estimateLength(const UnicodeString& patternString, UErrorCode& status);

    /**
     * Takes a string and escapes (quotes) characters that have special meaning in the affix pattern
     * syntax. This function does not reverse-lookup symbols.
     *
     * <p>Example input: "-$x"; example output: "'-'$x"
     *
     * @param input The string to be escaped.
     * @return The resulting UnicodeString.
     */
    static UnicodeString escape(const UnicodeString& input);

    static Field getFieldForType(AffixPatternType type);

    /**
     * Executes the unescape state machine. Replaces the unquoted characters "-", "+", "%", "‰", and
     * "¤" with the corresponding symbols provided by the {@link SymbolProvider}, and inserts the
     * result into the FormattedStringBuilder at the requested location.
     *
     * <p>Example input: "'-'¤x"; example output: "-$x"
     *
     * @param affixPattern The original string to be unescaped.
     * @param output The FormattedStringBuilder to mutate with the result.
     * @param position The index into the FormattedStringBuilder to insert the string.
     * @param provider An object to generate locale symbols.
     */
    static int32_t unescape(const UnicodeString& affixPattern, FormattedStringBuilder& output,
                            int32_t position, const SymbolProvider& provider, Field field,
                            UErrorCode& status);

    /**
   * Sames as {@link #unescape}, but only calculates the code point count.  More efficient than {@link #unescape}
   * if you only need the length but not the string itself.
     *
     * @param affixPattern The original string to be unescaped.
     * @param provider An object to generate locale symbols.
     * @return The same return value as if you called {@link #unescape}.
     */
    static int32_t unescapedCodePointCount(const UnicodeString& affixPattern,
                                           const SymbolProvider& provider, UErrorCode& status);

    /**
     * Checks whether the given affix pattern contains at least one token of the given type, which is
     * one of the constants "TYPE_" in {@link AffixPatternUtils}.
     *
     * @param affixPattern The affix pattern to check.
     * @param type The token type.
     * @return true if the affix pattern contains the given token type; false otherwise.
     */
    static bool containsType(const UnicodeString& affixPattern, AffixPatternType type, UErrorCode& status);

    /**
     * Checks whether the specified affix pattern has any unquoted currency symbols ("¤").
     *
     * @param affixPattern The string to check for currency symbols.
     * @return true if the literal has at least one unquoted currency symbol; false otherwise.
     */
    static bool hasCurrencySymbols(const UnicodeString& affixPattern, UErrorCode& status);

    /**
     * Replaces all occurrences of tokens with the given type with the given replacement char.
     *
     * @param affixPattern The source affix pattern (does not get modified).
     * @param type The token type.
     * @param replacementChar The char to substitute in place of chars of the given token type.
     * @return A string containing the new affix pattern.
     */
    static UnicodeString replaceType(const UnicodeString& affixPattern, AffixPatternType type,
                                     char16_t replacementChar, UErrorCode& status);

    /**
     * Returns whether the given affix pattern contains only symbols and ignorables as defined by the
     * given ignorables set.
     */
    static bool containsOnlySymbolsAndIgnorables(const UnicodeString& affixPattern,
                                                 const UnicodeSet& ignorables, UErrorCode& status);

    /**
     * Iterates over the affix pattern, calling the TokenConsumer for each token.
     */
    static void iterateWithConsumer(const UnicodeString& affixPattern, TokenConsumer& consumer,
                                    UErrorCode& status);

    /**
     * Returns the next token from the affix pattern.
     *
     * @param tag A bitmask used for keeping track of state from token to token. The initial value
     *     should be 0L.
     * @param patternString The affix pattern.
     * @return The bitmask tag to pass to the next call of this method to retrieve the following token
     *     (never negative), or -1 if there were no more tokens in the affix pattern.
     * @see #hasNext
     */
    static AffixTag nextToken(AffixTag tag, const UnicodeString& patternString, UErrorCode& status);

    /**
     * Returns whether the affix pattern string has any more tokens to be retrieved from a call to
     * {@link #nextToken}.
     *
     * @param tag The bitmask tag of the previous token, as returned by {@link #nextToken}.
     * @param string The affix pattern.
     * @return true if there are more tokens to consume; false otherwise.
     */
    static bool hasNext(const AffixTag& tag, const UnicodeString& string);

  private:
    /**
     * Encodes the given values into a tag struct.
     * The order of the arguments is consistent with Java, but the order of the stored
     * fields is not necessarily the same.
     */
    static inline AffixTag makeTag(int32_t offset, AffixPatternType type, AffixPatternState state,
                                   UChar32 cp) {
        return {offset, cp, state, type};
    }
};

} // namespace impl
} // namespace number
U_NAMESPACE_END


#endif //__NUMBER_AFFIXUTILS_H__



U_NAMESPACE_BEGIN namespace number {
namespace impl {

// Forward declaration
class PatternParser;

// Exported as U_I18N_API because it is a public member field of exported ParsedSubpatternInfo
struct U_I18N_API Endpoints {
    int32_t start = 0;
    int32_t end = 0;
};

// Exported as U_I18N_API because it is a public member field of exported ParsedPatternInfo
struct U_I18N_API ParsedSubpatternInfo {
    uint64_t groupingSizes = 0x0000ffffffff0000L;
    int32_t integerLeadingHashSigns = 0;
    int32_t integerTrailingHashSigns = 0;
    int32_t integerNumerals = 0;
    int32_t integerAtSigns = 0;
    int32_t integerTotal = 0; // for convenience
    int32_t fractionNumerals = 0;
    int32_t fractionHashSigns = 0;
    int32_t fractionTotal = 0; // for convenience
    bool hasDecimal = false;
    int32_t widthExceptAffixes = 0;
    // Note: NullableValue causes issues here with std::move.
    bool hasPadding = false;
    UNumberFormatPadPosition paddingLocation = UNUM_PAD_BEFORE_PREFIX;
    DecimalQuantity rounding;
    bool exponentHasPlusSign = false;
    int32_t exponentZeros = 0;
    bool hasPercentSign = false;
    bool hasPerMilleSign = false;
    bool hasCurrencySign = false;
    bool hasMinusSign = false;
    bool hasPlusSign = false;

    Endpoints prefixEndpoints;
    Endpoints suffixEndpoints;
    Endpoints paddingEndpoints;
};

// Exported as U_I18N_API because it is needed for the unit test PatternStringTest
struct U_I18N_API ParsedPatternInfo : public AffixPatternProvider, public UMemory {
    UnicodeString pattern;
    ParsedSubpatternInfo positive;
    ParsedSubpatternInfo negative;

    ParsedPatternInfo()
            : state(this->pattern), currentSubpattern(nullptr) {}

    ~ParsedPatternInfo() U_OVERRIDE = default;

    // Need to declare this explicitly because of the destructor
    ParsedPatternInfo& operator=(ParsedPatternInfo&& src) U_NOEXCEPT = default;

    static int32_t getLengthFromEndpoints(const Endpoints& endpoints);

    char16_t charAt(int32_t flags, int32_t index) const U_OVERRIDE;

    int32_t length(int32_t flags) const U_OVERRIDE;

    UnicodeString getString(int32_t flags) const U_OVERRIDE;

    bool positiveHasPlusSign() const U_OVERRIDE;

    bool hasNegativeSubpattern() const U_OVERRIDE;

    bool negativeHasMinusSign() const U_OVERRIDE;

    bool hasCurrencySign() const U_OVERRIDE;

    bool containsSymbolType(AffixPatternType type, UErrorCode& status) const U_OVERRIDE;

    bool hasBody() const U_OVERRIDE;

  private:
    struct U_I18N_API ParserState {
        const UnicodeString& pattern; // reference to the parent
        int32_t offset = 0;

        explicit ParserState(const UnicodeString& _pattern)
                : pattern(_pattern) {}

        ParserState& operator=(ParserState&& src) U_NOEXCEPT {
            // Leave pattern reference alone; it will continue to point to the same place in memory,
            // which gets overwritten by ParsedPatternInfo's implicit move assignment.
            offset = src.offset;
            return *this;
        }

        UChar32 peek();

        UChar32 next();

        // TODO: We don't currently do anything with the message string.
        // This method is here as a shell for Java compatibility.
        inline void toParseException(const char16_t* message) { (void) message; }
    } state;

    // NOTE: In Java, these are written as pure functions.
    // In C++, they're written as methods.
    // The behavior is the same.

    // Mutable transient pointer:
    ParsedSubpatternInfo* currentSubpattern;

    // In Java, "negative == null" tells us whether or not we had a negative subpattern.
    // In C++, we need to remember in another boolean.
    bool fHasNegativeSubpattern = false;

    const Endpoints& getEndpoints(int32_t flags) const;

    /** Run the recursive descent parser. */
    void consumePattern(const UnicodeString& patternString, UErrorCode& status);

    void consumeSubpattern(UErrorCode& status);

    void consumePadding(PadPosition paddingLocation, UErrorCode& status);

    void consumeAffix(Endpoints& endpoints, UErrorCode& status);

    void consumeLiteral(UErrorCode& status);

    void consumeFormat(UErrorCode& status);

    void consumeIntegerFormat(UErrorCode& status);

    void consumeFractionFormat(UErrorCode& status);

    void consumeExponent(UErrorCode& status);

    friend class PatternParser;
};

enum IgnoreRounding {
    IGNORE_ROUNDING_NEVER = 0, IGNORE_ROUNDING_IF_CURRENCY = 1, IGNORE_ROUNDING_ALWAYS = 2
};

class U_I18N_API PatternParser {
  public:
    /**
     * Runs the recursive descent parser on the given pattern string, returning a data structure with raw information
     * about the pattern string.
     *
     * <p>
     * To obtain a more useful form of the data, consider using {@link #parseToProperties} instead.
     *
     * TODO: Change argument type to const char16_t* instead of UnicodeString?
     *
     * @param patternString
     *            The LDML decimal format pattern (Excel-style pattern) to parse.
     * @return The results of the parse.
     */
    static void parseToPatternInfo(const UnicodeString& patternString, ParsedPatternInfo& patternInfo,
                                   UErrorCode& status);

    /**
     * Parses a pattern string into a new property bag.
     *
     * @param pattern
     *            The pattern string, like "#,##0.00"
     * @param ignoreRounding
     *            Whether to leave out rounding information (minFrac, maxFrac, and rounding increment) when parsing the
     *            pattern. This may be desirable if a custom rounding mode, such as CurrencyUsage, is to be used
     *            instead.
     * @return A property bag object.
     * @throws IllegalArgumentException
     *             If there is a syntax error in the pattern string.
     */
    static DecimalFormatProperties parseToProperties(const UnicodeString& pattern,
                                                     IgnoreRounding ignoreRounding, UErrorCode& status);

    static DecimalFormatProperties parseToProperties(const UnicodeString& pattern, UErrorCode& status);

    /**
     * Parses a pattern string into an existing property bag. All properties that can be encoded into a pattern string
     * will be overwritten with either their default value or with the value coming from the pattern string. Properties
     * that cannot be encoded into a pattern string, such as rounding mode, are not modified.
     *
     * @param pattern
     *            The pattern string, like "#,##0.00"
     * @param properties
     *            The property bag object to overwrite.
     * @param ignoreRounding
     *            See {@link #parseToProperties(String pattern, int ignoreRounding)}.
     * @throws IllegalArgumentException
     *             If there was a syntax error in the pattern string.
     */
    static void parseToExistingProperties(const UnicodeString& pattern,
                                          DecimalFormatProperties& properties,
                                          IgnoreRounding ignoreRounding, UErrorCode& status);

  private:
    static void parseToExistingPropertiesImpl(const UnicodeString& pattern,
                                              DecimalFormatProperties& properties,
                                              IgnoreRounding ignoreRounding, UErrorCode& status);

    /** Finalizes the temporary data stored in the ParsedPatternInfo to the Properties. */
    static void patternInfoToProperties(DecimalFormatProperties& properties,
                                        ParsedPatternInfo& patternInfo, IgnoreRounding _ignoreRounding,
                                        UErrorCode& status);
};

class U_I18N_API PatternStringUtils {
  public:
    /**
     * Determine whether a given roundingIncrement should be ignored for formatting
     * based on the current maxFrac value (maximum fraction digits). For example a
     * roundingIncrement of 0.01 should be ignored if maxFrac is 1, but not if maxFrac
     * is 2 or more. Note that roundingIncrements are rounded up in significance, so
     * a roundingIncrement of 0.006 is treated like 0.01 for this determination, i.e.
     * it should not be ignored if maxFrac is 2 or more (but a roundingIncrement of
     * 0.005 is treated like 0.001 for significance).
     *
     * This test is needed for both NumberPropertyMapper::oldToNew and 
     * PatternStringUtils::propertiesToPatternString. In Java it cannot be
     * exported by NumberPropertyMapper (package provate) so it is in
     * PatternStringUtils, do the same in C.
     *
     * @param roundIncr
     *            The roundingIncrement to be checked. Must be non-zero.
     * @param maxFrac
     *            The current maximum fraction digits value.
     * @return true if roundIncr should be ignored for formatting.
     */
     static bool ignoreRoundingIncrement(double roundIncr, int32_t maxFrac);

    /**
     * Creates a pattern string from a property bag.
     *
     * <p>
     * Since pattern strings support only a subset of the functionality available in a property bag, a new property bag
     * created from the string returned by this function may not be the same as the original property bag.
     *
     * @param properties
     *            The property bag to serialize.
     * @return A pattern string approximately serializing the property bag.
     */
    static UnicodeString propertiesToPatternString(const DecimalFormatProperties& properties,
                                                   UErrorCode& status);


    /**
     * Converts a pattern between standard notation and localized notation. Localized notation means that instead of
     * using generic placeholders in the pattern, you use the corresponding locale-specific characters instead. For
     * example, in locale <em>fr-FR</em>, the period in the pattern "0.000" means "decimal" in standard notation (as it
     * does in every other locale), but it means "grouping" in localized notation.
     *
     * <p>
     * A greedy string-substitution strategy is used to substitute locale symbols. If two symbols are ambiguous or have
     * the same prefix, the result is not well-defined.
     *
     * <p>
     * Locale symbols are not allowed to contain the ASCII quote character.
     *
     * <p>
     * This method is provided for backwards compatibility and should not be used in any new code.
     *
     * TODO(C++): This method is not yet implemented.
     *
     * @param input
     *            The pattern to convert.
     * @param symbols
     *            The symbols corresponding to the localized pattern.
     * @param toLocalized
     *            true to convert from standard to localized notation; false to convert from localized to standard
     *            notation.
     * @return The pattern expressed in the other notation.
     */
    static UnicodeString convertLocalized(const UnicodeString& input, const DecimalFormatSymbols& symbols,
                                          bool toLocalized, UErrorCode& status);

    /**
     * This method contains the heart of the logic for rendering LDML affix strings. It handles
     * sign-always-shown resolution, whether to use the positive or negative subpattern, permille
     * substitution, and plural forms for CurrencyPluralInfo.
     */
    static void patternInfoToStringBuilder(const AffixPatternProvider& patternInfo, bool isPrefix,
                                           Signum signum, UNumberSignDisplay signDisplay,
                                           StandardPlural::Form plural, bool perMilleReplacesPercent,
                                           UnicodeString& output);

  private:
    /** @return The number of chars inserted. */
    static int escapePaddingString(UnicodeString input, UnicodeString& output, int startIndex,
                                   UErrorCode& status);
};

} // namespace impl
} // namespace number
U_NAMESPACE_END


#endif //__NUMBER_PATTERNSTRING_H__


// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING



// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN


/**
 * A mutable UnicodeString wrapper with a variable offset and length and
 * support for case folding. The charAt, length, and subSequence methods all
 * operate relative to the fixed offset into the UnicodeString.
 *
 * Intended to be useful for parsing.
 *
 * CAUTION: Since this class is mutable, it must not be used anywhere that an
 * immutable object is required, like in a cache or as the key of a hash map.
 *
 * @author sffc (Shane Carr)
 */
// Exported as U_I18N_API for tests
class U_I18N_API StringSegment : public UMemory {
  public:
    StringSegment(const UnicodeString& str, bool ignoreCase);

    int32_t getOffset() const;

    void setOffset(int32_t start);

    /**
     * Equivalent to <code>setOffset(getOffset()+delta)</code>.
     *
     * <p>
     * This method is usually called by a Matcher to register that a char was consumed. If the char is
     * strong (it usually is, except for things like whitespace), follow this with a call to
     * {@link ParsedNumber#setCharsConsumed}. For more information on strong chars, see that method.
     */
    void adjustOffset(int32_t delta);

    /**
     * Adjusts the offset by the width of the current code point, either 1 or 2 chars.
     */
    void adjustOffsetByCodePoint();

    void setLength(int32_t length);

    void resetLength();

    int32_t length() const;

    char16_t charAt(int32_t index) const;

    UChar32 codePointAt(int32_t index) const;

    UnicodeString toUnicodeString() const;

    const UnicodeString toTempUnicodeString() const;

    /**
     * Returns the first code point in the string segment, or -1 if the string starts with an invalid
     * code point.
     *
     * <p>
     * <strong>Important:</strong> Most of the time, you should use {@link #startsWith}, which handles case
     * folding logic, instead of this method.
     */
    UChar32 getCodePoint() const;

    /**
     * Returns true if the first code point of this StringSegment equals the given code point.
     *
     * <p>
     * This method will perform case folding if case folding is enabled for the parser.
     */
    bool startsWith(UChar32 otherCp) const;

    /**
     * Returns true if the first code point of this StringSegment is in the given UnicodeSet.
     */
    bool startsWith(const UnicodeSet& uniset) const;

    /**
     * Returns true if there is at least one code point of overlap between this StringSegment and the
     * given UnicodeString.
     */
    bool startsWith(const UnicodeString& other) const;

    /**
     * Returns the length of the prefix shared by this StringSegment and the given UnicodeString. For
     * example, if this string segment is "aab", and the char sequence is "aac", this method returns 2,
     * since the first 2 characters are the same.
     *
     * <p>
     * This method only returns offsets along code point boundaries.
     *
     * <p>
     * This method will perform case folding if case folding was enabled in the constructor.
     *
     * <p>
     * IMPORTANT: The given UnicodeString must not be empty! It is the caller's responsibility to check.
     */
    int32_t getCommonPrefixLength(const UnicodeString& other);

    /**
     * Like {@link #getCommonPrefixLength}, but never performs case folding, even if case folding is
     * enabled for the parser.
     */
    int32_t getCaseSensitivePrefixLength(const UnicodeString& other);

    bool operator==(const UnicodeString& other) const;

  private:
    const UnicodeString& fStr;
    int32_t fStart;
    int32_t fEnd;
    bool fFoldCase;

    int32_t getPrefixLengthInternal(const UnicodeString& other, bool foldCase);

    static bool codePointsEqual(UChar32 cp1, UChar32 cp2, bool foldCase);
};


U_NAMESPACE_END

#endif //__NUMPARSE_STRINGSEGMENT_H__


U_NAMESPACE_BEGIN
namespace numparse {
namespace impl {

// Forward-declarations
class ParsedNumber;

typedef int32_t result_flags_t;
typedef int32_t parse_flags_t;

/** Flags for the type result_flags_t */
enum ResultFlags {
    FLAG_NEGATIVE = 0x0001,
    FLAG_PERCENT = 0x0002,
    FLAG_PERMILLE = 0x0004,
    FLAG_HAS_EXPONENT = 0x0008,
    // FLAG_HAS_DEFAULT_CURRENCY = 0x0010, // no longer used
    FLAG_HAS_DECIMAL_SEPARATOR = 0x0020,
    FLAG_NAN = 0x0040,
    FLAG_INFINITY = 0x0080,
    FLAG_FAIL = 0x0100,
};

/** Flags for the type parse_flags_t */
enum ParseFlags {
    PARSE_FLAG_IGNORE_CASE = 0x0001,
    PARSE_FLAG_MONETARY_SEPARATORS = 0x0002,
    PARSE_FLAG_STRICT_SEPARATORS = 0x0004,
    PARSE_FLAG_STRICT_GROUPING_SIZE = 0x0008,
    PARSE_FLAG_INTEGER_ONLY = 0x0010,
    PARSE_FLAG_GROUPING_DISABLED = 0x0020,
    // PARSE_FLAG_FRACTION_GROUPING_ENABLED = 0x0040, // see #10794
    PARSE_FLAG_INCLUDE_UNPAIRED_AFFIXES = 0x0080,
    PARSE_FLAG_USE_FULL_AFFIXES = 0x0100,
    PARSE_FLAG_EXACT_AFFIX = 0x0200,
    PARSE_FLAG_PLUS_SIGN_ALLOWED = 0x0400,
    // PARSE_FLAG_OPTIMIZE = 0x0800, // no longer used
    // PARSE_FLAG_FORCE_BIG_DECIMAL = 0x1000, // not used in ICU4C
    PARSE_FLAG_NO_FOREIGN_CURRENCY = 0x2000,
    PARSE_FLAG_ALLOW_INFINITE_RECURSION = 0x4000,
    PARSE_FLAG_STRICT_IGNORABLES = 0x8000,
};


// TODO: Is this class worthwhile?
template<int32_t stackCapacity>
class CompactUnicodeString {
  public:
    CompactUnicodeString() {
        static_assert(stackCapacity > 0, "cannot have zero space on stack");
        fBuffer[0] = 0;
    }

    CompactUnicodeString(const UnicodeString& text)
            : fBuffer(text.length() + 1) {
        uprv_memcpy(fBuffer.getAlias(), text.getBuffer(), sizeof(UChar) * text.length());
        fBuffer[text.length()] = 0;
    }

    inline UnicodeString toAliasedUnicodeString() const {
        return UnicodeString(TRUE, fBuffer.getAlias(), -1);
    }

    bool operator==(const CompactUnicodeString& other) const {
        // Use the alias-only constructor and then call UnicodeString operator==
        return toAliasedUnicodeString() == other.toAliasedUnicodeString();
    }

  private:
    MaybeStackArray<UChar, stackCapacity> fBuffer;
};


/**
 * Struct-like class to hold the results of a parsing routine.
 *
 * @author sffc
 */
// Exported as U_I18N_API for tests
class U_I18N_API ParsedNumber {
  public:

    /**
     * The numerical value that was parsed.
     */
    ::icu::number::impl::DecimalQuantity quantity;

    /**
     * The index of the last char consumed during parsing. If parsing started at index 0, this is equal
     * to the number of chars consumed. This is NOT necessarily the same as the StringSegment offset;
     * "weak" chars, like whitespace, change the offset, but the charsConsumed is not touched until a
     * "strong" char is encountered.
     */
    int32_t charEnd;

    /**
     * Boolean flags (see constants above).
     */
    result_flags_t flags;

    /**
     * The pattern string corresponding to the prefix that got consumed.
     */
    UnicodeString prefix;

    /**
     * The pattern string corresponding to the suffix that got consumed.
     */
    UnicodeString suffix;

    /**
     * The currency that got consumed.
     */
    UChar currencyCode[4];

    ParsedNumber();

    ParsedNumber(const ParsedNumber& other) = default;

    ParsedNumber& operator=(const ParsedNumber& other) = default;

    void clear();

    /**
     * Call this method to register that a "strong" char was consumed. This should be done after calling
     * {@link StringSegment#setOffset} or {@link StringSegment#adjustOffset} except when the char is
     * "weak", like whitespace.
     *
     * <p>
     * <strong>What is a strong versus weak char?</strong> The behavior of number parsing is to "stop"
     * after reading the number, even if there is other content following the number. For example, after
     * parsing the string "123 " (123 followed by a space), the cursor should be set to 3, not 4, even
     * though there are matchers that accept whitespace. In this example, the digits are strong, whereas
     * the whitespace is weak. Grouping separators are weak, whereas decimal separators are strong. Most
     * other chars are strong.
     *
     * @param segment
     *            The current StringSegment, usually immediately following a call to setOffset.
     */
    void setCharsConsumed(const StringSegment& segment);

    /** Apply certain number-related flags to the DecimalQuantity. */
    void postProcess();

    /**
     * Returns whether this the parse was successful. To be successful, at least one char must have been
     * consumed, and the failure flag must not be set.
     */
    bool success() const;

    bool seenNumber() const;

    double getDouble(UErrorCode& status) const;

    void populateFormattable(Formattable& output, parse_flags_t parseFlags) const;

    bool isBetterThan(const ParsedNumber& other);
};


/**
 * The core interface implemented by all matchers used for number parsing.
 *
 * Given a string, there should NOT be more than one way to consume the string with the same matcher
 * applied multiple times. If there is, the non-greedy parsing algorithm will be unhappy and may enter an
 * exponential-time loop. For example, consider the "A Matcher" that accepts "any number of As". Given
 * the string "AAAA", there are 2^N = 8 ways to apply the A Matcher to this string: you could have the A
 * Matcher apply 4 times to each character; you could have it apply just once to all the characters; you
 * could have it apply to the first 2 characters and the second 2 characters; and so on. A better version
 * of the "A Matcher" would be for it to accept exactly one A, and allow the algorithm to run it
 * repeatedly to consume a string of multiple As. The A Matcher can implement the Flexible interface
 * below to signal that it can be applied multiple times in a row.
 *
 * @author sffc
 */
// Exported as U_I18N_API for tests
class U_I18N_API NumberParseMatcher {
  public:
    virtual ~NumberParseMatcher();

    /**
     * Matchers can override this method to return true to indicate that they are optional and can be run
     * repeatedly. Used by SeriesMatcher, primarily in the context of IgnorablesMatcher.
     */
    virtual bool isFlexible() const {
        return false;
    }

    /**
     * Runs this matcher starting at the beginning of the given StringSegment. If this matcher finds
     * something interesting in the StringSegment, it should update the offset of the StringSegment
     * corresponding to how many chars were matched.
     *
     * This method is thread-safe.
     *
     * @param segment
     *            The StringSegment to match against. Matches always start at the beginning of the
     *            segment. The segment is guaranteed to contain at least one char.
     * @param result
     *            The data structure to store results if the match succeeds.
     * @return Whether this matcher thinks there may be more interesting chars beyond the end of the
     *         string segment.
     */
    virtual bool match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const = 0;

    /**
     * Performs a fast "smoke check" for whether or not this matcher could possibly match against the
     * given string segment. The test should be as fast as possible but also as restrictive as possible.
     * For example, matchers can maintain a UnicodeSet of all code points that count possibly start a
     * match. Matchers should use the {@link StringSegment#startsWith} method in order to correctly
     * handle case folding.
     *
     * @param segment
     *            The segment to check against.
     * @return true if the matcher might be able to match against this segment; false if it definitely
     *         will not be able to match.
     */
    virtual bool smokeTest(const StringSegment& segment) const = 0;

    /**
     * Method called at the end of a parse, after all matchers have failed to consume any more chars.
     * Allows a matcher to make final modifications to the result given the knowledge that no more
     * matches are possible.
     *
     * @param result
     *            The data structure to store results.
     */
    virtual void postProcess(ParsedNumber&) const {
        // Default implementation: no-op
    }

    // String for debugging
    virtual UnicodeString toString() const = 0;

  protected:
    // No construction except by subclasses!
    NumberParseMatcher() = default;
};


/**
 * Interface for use in arguments.
 */
// Exported as U_I18N_API for tests
class U_I18N_API MutableMatcherCollection {
  public:
    virtual ~MutableMatcherCollection() = default;

    virtual void addMatcher(NumberParseMatcher& matcher) = 0;
};


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__NUMPARSE_TYPES_H__




U_NAMESPACE_BEGIN namespace number {
namespace impl {


// Exported as U_I18N_API for tests
class U_I18N_API CurrencySymbols : public UMemory {
  public:
    CurrencySymbols() = default; // default constructor: leaves class in valid but undefined state

    /** Creates an instance in which all symbols are loaded from data. */
    CurrencySymbols(CurrencyUnit currency, const Locale& locale, UErrorCode& status);

    /** Creates an instance in which some symbols might be pre-populated. */
    CurrencySymbols(CurrencyUnit currency, const Locale& locale, const DecimalFormatSymbols& symbols,
                    UErrorCode& status);

    const char16_t* getIsoCode() const;

    UnicodeString getNarrowCurrencySymbol(UErrorCode& status) const;

    UnicodeString getCurrencySymbol(UErrorCode& status) const;

    UnicodeString getIntlCurrencySymbol(UErrorCode& status) const;

    UnicodeString getPluralName(StandardPlural::Form plural, UErrorCode& status) const;

  protected:
    // Required fields:
    CurrencyUnit fCurrency;
    CharString fLocaleName;

    // Optional fields:
    UnicodeString fCurrencySymbol;
    UnicodeString fIntlCurrencySymbol;

    UnicodeString loadSymbol(UCurrNameStyle selector, UErrorCode& status) const;
};


/**
 * Resolves the effective currency from the property bag.
 */
CurrencyUnit
resolveCurrency(const DecimalFormatProperties& properties, const Locale& locale, UErrorCode& status);


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__SOURCE_NUMBER_CURRENCYSYMBOLS_H__

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING


// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN namespace numparse {
namespace impl {

using ::icu::number::impl::Grouper;

class DecimalMatcher : public NumberParseMatcher, public UMemory {
  public:
    DecimalMatcher() = default;  // WARNING: Leaves the object in an unusable state

    DecimalMatcher(const DecimalFormatSymbols& symbols, const Grouper& grouper,
                   parse_flags_t parseFlags);

    bool match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const override;

    bool
    match(StringSegment& segment, ParsedNumber& result, int8_t exponentSign, UErrorCode& status) const;

    bool smokeTest(const StringSegment& segment) const override;

    UnicodeString toString() const override;

  private:
    /** If true, only accept strings whose grouping sizes match the locale */
    bool requireGroupingMatch;

    /** If true, do not accept grouping separators at all */
    bool groupingDisabled;

    // Fraction grouping parsing is disabled for now but could be enabled later.
    // See http://bugs.icu-project.org/trac/ticket/10794
    // bool fractionGrouping;

    /** If true, do not accept numbers in the fraction */
    bool integerOnly;

    int16_t grouping1;
    int16_t grouping2;

    UnicodeString groupingSeparator;
    UnicodeString decimalSeparator;

    // Assumption: these sets all consist of single code points. If this assumption needs to be broken,
    // fix getLeadCodePoints() as well as matching logic. Be careful of the performance impact.
    const UnicodeSet* groupingUniSet;
    const UnicodeSet* decimalUniSet;
    const UnicodeSet* separatorSet;
    const UnicodeSet* leadSet;

    // Make this class the owner of a few objects that could be allocated.
    // The first three LocalPointers are used for assigning ownership only.
    LocalPointer<const UnicodeSet> fLocalDecimalUniSet;
    LocalPointer<const UnicodeSet> fLocalSeparatorSet;
    LocalArray<const UnicodeString> fLocalDigitStrings;

    bool validateGroup(int32_t sepType, int32_t count, bool isPrimary) const;
};


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__NUMPARSE_DECIMAL_H__

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING





U_NAMESPACE_BEGIN namespace numparse {
namespace impl {


/**
 * A base class for many matchers that performs a simple match against a UnicodeString and/or UnicodeSet.
 *
 * @author sffc
 */
// Exported as U_I18N_API for tests
class U_I18N_API SymbolMatcher : public NumberParseMatcher, public UMemory {
  public:
    SymbolMatcher() = default;  // WARNING: Leaves the object in an unusable state

    const UnicodeSet* getSet() const;

    bool match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const override;

    bool smokeTest(const StringSegment& segment) const override;

    UnicodeString toString() const override;

    virtual bool isDisabled(const ParsedNumber& result) const = 0;

    virtual void accept(StringSegment& segment, ParsedNumber& result) const = 0;

  protected:
    UnicodeString fString;
    const UnicodeSet* fUniSet; // a reference from numparse_unisets.h; never owned

    SymbolMatcher(const UnicodeString& symbolString, unisets::Key key);
};


// Exported as U_I18N_API for tests
class U_I18N_API IgnorablesMatcher : public SymbolMatcher {
  public:
    IgnorablesMatcher() = default;  // WARNING: Leaves the object in an unusable state

    IgnorablesMatcher(parse_flags_t parseFlags);

    bool isFlexible() const override;

    UnicodeString toString() const override;

  protected:
    bool isDisabled(const ParsedNumber& result) const override;

    void accept(StringSegment& segment, ParsedNumber& result) const override;
};


class InfinityMatcher : public SymbolMatcher {
  public:
    InfinityMatcher() = default;  // WARNING: Leaves the object in an unusable state

    InfinityMatcher(const DecimalFormatSymbols& dfs);

  protected:
    bool isDisabled(const ParsedNumber& result) const override;

    void accept(StringSegment& segment, ParsedNumber& result) const override;
};


// Exported as U_I18N_API for tests
class U_I18N_API MinusSignMatcher : public SymbolMatcher {
  public:
    MinusSignMatcher() = default;  // WARNING: Leaves the object in an unusable state

    MinusSignMatcher(const DecimalFormatSymbols& dfs, bool allowTrailing);

  protected:
    bool isDisabled(const ParsedNumber& result) const override;

    void accept(StringSegment& segment, ParsedNumber& result) const override;

  private:
    bool fAllowTrailing;
};


class NanMatcher : public SymbolMatcher {
  public:
    NanMatcher() = default;  // WARNING: Leaves the object in an unusable state

    NanMatcher(const DecimalFormatSymbols& dfs);

  protected:
    bool isDisabled(const ParsedNumber& result) const override;

    void accept(StringSegment& segment, ParsedNumber& result) const override;
};


class PaddingMatcher : public SymbolMatcher {
  public:
    PaddingMatcher() = default;  // WARNING: Leaves the object in an unusable state

    PaddingMatcher(const UnicodeString& padString);

    bool isFlexible() const override;

  protected:
    bool isDisabled(const ParsedNumber& result) const override;

    void accept(StringSegment& segment, ParsedNumber& result) const override;
};


// Exported as U_I18N_API for tests
class U_I18N_API PercentMatcher : public SymbolMatcher {
  public:
    PercentMatcher() = default;  // WARNING: Leaves the object in an unusable state

    PercentMatcher(const DecimalFormatSymbols& dfs);

  protected:
    bool isDisabled(const ParsedNumber& result) const override;

    void accept(StringSegment& segment, ParsedNumber& result) const override;
};

// Exported as U_I18N_API for tests
class U_I18N_API PermilleMatcher : public SymbolMatcher {
  public:
    PermilleMatcher() = default;  // WARNING: Leaves the object in an unusable state

    PermilleMatcher(const DecimalFormatSymbols& dfs);

  protected:
    bool isDisabled(const ParsedNumber& result) const override;

    void accept(StringSegment& segment, ParsedNumber& result) const override;
};


// Exported as U_I18N_API for tests
class U_I18N_API PlusSignMatcher : public SymbolMatcher {
  public:
    PlusSignMatcher() = default;  // WARNING: Leaves the object in an unusable state

    PlusSignMatcher(const DecimalFormatSymbols& dfs, bool allowTrailing);

  protected:
    bool isDisabled(const ParsedNumber& result) const override;

    void accept(StringSegment& segment, ParsedNumber& result) const override;

  private:
    bool fAllowTrailing;
};


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__NUMPARSE_SYMBOLS_H__

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING






using icu::number::impl::Grouper;

U_NAMESPACE_BEGIN namespace numparse {
namespace impl {


class ScientificMatcher : public NumberParseMatcher, public UMemory {
  public:
    ScientificMatcher() = default;  // WARNING: Leaves the object in an unusable state

    ScientificMatcher(const DecimalFormatSymbols& dfs, const Grouper& grouper);

    bool match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const override;

    bool smokeTest(const StringSegment& segment) const override;

    UnicodeString toString() const override;

  private:
    UnicodeString fExponentSeparatorString;
    DecimalMatcher fExponentMatcher;
    IgnorablesMatcher fIgnorablesMatcher;
    UnicodeString fCustomMinusSign;
    UnicodeString fCustomPlusSign;
};


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__NUMPARSE_SCIENTIFIC_H__


// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING


// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html




#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

// Export an explicit template instantiation of the MaybeStackArray that is used as a data member of ArraySeriesMatcher.
// When building DLLs for Windows this is required even though no direct access to the MaybeStackArray leaks out of the i18n library.
// (See digitlst.h, pluralaffix.h, datefmt.h, and others for similar examples.)
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN
template class U_I18N_API MaybeStackArray<const numparse::impl::NumberParseMatcher*, 3>;
#endif

namespace numparse {
namespace impl {

/**
 * Base class for AnyMatcher and SeriesMatcher.
 */
// Exported as U_I18N_API for tests
class U_I18N_API CompositionMatcher : public NumberParseMatcher {
  protected:
    // No construction except by subclasses!
    CompositionMatcher() = default;

    // To be overridden by subclasses (used for iteration):
    virtual const NumberParseMatcher* const* begin() const = 0;

    // To be overridden by subclasses (used for iteration):
    virtual const NumberParseMatcher* const* end() const = 0;
};


// NOTE: AnyMatcher is no longer being used. The previous definition is shown below.
// The implementation can be found in SVN source control, deleted around March 30, 2018.
///**
// * Composes a number of matchers, and succeeds if any of the matchers succeed. Always greedily chooses
// * the first matcher in the list to succeed.
// *
// * NOTE: In C++, this is a base class, unlike ICU4J, which uses a factory-style interface.
// *
// * @author sffc
// * @see SeriesMatcher
// */
//class AnyMatcher : public CompositionMatcher {
//  public:
//    bool match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const override;
//
//    bool smokeTest(const StringSegment& segment) const override;
//
//    void postProcess(ParsedNumber& result) const override;
//
//  protected:
//    // No construction except by subclasses!
//    AnyMatcher() = default;
//};


/**
 * Composes a number of matchers, running one after another. Matches the input string only if all of the
 * matchers in the series succeed. Performs greedy matches within the context of the series.
 *
 * @author sffc
 * @see AnyMatcher
 */
// Exported as U_I18N_API for tests
class U_I18N_API SeriesMatcher : public CompositionMatcher {
  public:
    bool match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const override;

    bool smokeTest(const StringSegment& segment) const override;

    void postProcess(ParsedNumber& result) const override;

    virtual int32_t length() const = 0;

  protected:
    // No construction except by subclasses!
    SeriesMatcher() = default;
};

/**
 * An implementation of SeriesMatcher that references an array of matchers.
 *
 * The object adopts the array, but NOT the matchers contained inside the array.
 */
// Exported as U_I18N_API for tests
class U_I18N_API ArraySeriesMatcher : public SeriesMatcher {
  public:
    ArraySeriesMatcher();  // WARNING: Leaves the object in an unusable state

    typedef MaybeStackArray<const NumberParseMatcher*, 3> MatcherArray;

    /** The array is std::move'd */
    ArraySeriesMatcher(MatcherArray& matchers, int32_t matchersLen);

    UnicodeString toString() const override;

    int32_t length() const override;

  protected:
    const NumberParseMatcher* const* begin() const override;

    const NumberParseMatcher* const* end() const override;

  private:
    MatcherArray fMatchers;
    int32_t fMatchersLen;
};


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__SOURCE_NUMPARSE_COMPOSITIONS__





U_NAMESPACE_BEGIN namespace numparse {
namespace impl {

using ::icu::number::impl::CurrencySymbols;

/**
 * Matches a currency, either a custom currency or one from the data bundle. The class is called
 * "combined" to emphasize that the currency string may come from one of multiple sources.
 *
 * Will match currency spacing either before or after the number depending on whether we are currently in
 * the prefix or suffix.
 *
 * The implementation of this class is slightly different between J and C. See #13584 for a follow-up.
 *
 * @author sffc
 */
// Exported as U_I18N_API for tests
class U_I18N_API CombinedCurrencyMatcher : public NumberParseMatcher, public UMemory {
  public:
    CombinedCurrencyMatcher() = default;  // WARNING: Leaves the object in an unusable state

    CombinedCurrencyMatcher(const CurrencySymbols& currencySymbols, const DecimalFormatSymbols& dfs,
                            parse_flags_t parseFlags, UErrorCode& status);

    bool match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const override;

    bool smokeTest(const StringSegment& segment) const override;

    UnicodeString toString() const override;

  private:
    UChar fCurrencyCode[4];
    UnicodeString fCurrency1;
    UnicodeString fCurrency2;

    bool fUseFullCurrencyData;
    UnicodeString fLocalLongNames[StandardPlural::COUNT];

    UnicodeString afterPrefixInsert;
    UnicodeString beforeSuffixInsert;

    // We could use Locale instead of CharString here, but
    // Locale has a non-trivial default constructor.
    CharString fLocaleName;

    // TODO: See comments in constructor in numparse_currency.cpp
    // UnicodeSet fLeadCodePoints;

    /** Matches the currency string without concern for currency spacing. */
    bool matchCurrency(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const;
};


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__NUMPARSE_CURRENCY_H__

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING









U_NAMESPACE_BEGIN
namespace numparse {
namespace impl {

// Forward-declaration of implementation classes for friending
class AffixPatternMatcherBuilder;
class AffixPatternMatcher;

using ::icu::number::impl::AffixPatternProvider;
using ::icu::number::impl::TokenConsumer;
using ::icu::number::impl::CurrencySymbols;


class CodePointMatcher : public NumberParseMatcher, public UMemory {
  public:
    CodePointMatcher() = default;  // WARNING: Leaves the object in an unusable state

    CodePointMatcher(UChar32 cp);

    bool match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const override;

    bool smokeTest(const StringSegment& segment) const override;

    UnicodeString toString() const override;

  private:
    UChar32 fCp;
};

} // namespace impl
} // namespace numparse

// Export a explicit template instantiations of MaybeStackArray, MemoryPool and CompactUnicodeString.
// When building DLLs for Windows this is required even though no direct access leaks out of the i18n library.
// (See digitlst.h, pluralaffix.h, datefmt.h, and others for similar examples.)
// Note: These need to be outside of the numparse::impl namespace, or Clang will generate a compile error.
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN
template class U_I18N_API MaybeStackArray<numparse::impl::CodePointMatcher*, 8>; 
template class U_I18N_API MaybeStackArray<UChar, 4>;
template class U_I18N_API MemoryPool<numparse::impl::CodePointMatcher, 8>;
template class U_I18N_API numparse::impl::CompactUnicodeString<4>;
#endif

namespace numparse {
namespace impl {

struct AffixTokenMatcherSetupData {
    const CurrencySymbols& currencySymbols;
    const DecimalFormatSymbols& dfs;
    IgnorablesMatcher& ignorables;
    const Locale& locale;
    parse_flags_t parseFlags;
};


/**
 * Small helper class that generates matchers for individual tokens for AffixPatternMatcher.
 *
 * In Java, this is called AffixTokenMatcherFactory (a "factory"). However, in C++, it is called a
 * "warehouse", because in addition to generating the matchers, it also retains ownership of them. The
 * warehouse must stay in scope for the whole lifespan of the AffixPatternMatcher that uses matchers from
 * the warehouse.
 *
 * @author sffc
 */
// Exported as U_I18N_API for tests
class U_I18N_API AffixTokenMatcherWarehouse : public UMemory {
  public:
    AffixTokenMatcherWarehouse() = default;  // WARNING: Leaves the object in an unusable state

    AffixTokenMatcherWarehouse(const AffixTokenMatcherSetupData* setupData);

    NumberParseMatcher& minusSign();

    NumberParseMatcher& plusSign();

    NumberParseMatcher& percent();

    NumberParseMatcher& permille();

    NumberParseMatcher& currency(UErrorCode& status);

    IgnorablesMatcher& ignorables();

    NumberParseMatcher* nextCodePointMatcher(UChar32 cp, UErrorCode& status);

  private:
    // NOTE: The following field may be unsafe to access after construction is done!
    const AffixTokenMatcherSetupData* fSetupData;

    // NOTE: These are default-constructed and should not be used until initialized.
    MinusSignMatcher fMinusSign;
    PlusSignMatcher fPlusSign;
    PercentMatcher fPercent;
    PermilleMatcher fPermille;
    CombinedCurrencyMatcher fCurrency;

    // Use a child class for code point matchers, since it requires non-default operators.
    MemoryPool<CodePointMatcher> fCodePoints;

    friend class AffixPatternMatcherBuilder;
    friend class AffixPatternMatcher;
};


class AffixPatternMatcherBuilder : public TokenConsumer, public MutableMatcherCollection {
  public:
    AffixPatternMatcherBuilder(const UnicodeString& pattern, AffixTokenMatcherWarehouse& warehouse,
                               IgnorablesMatcher* ignorables);

    void consumeToken(::icu::number::impl::AffixPatternType type, UChar32 cp, UErrorCode& status) override;

    /** NOTE: You can build only once! */
    AffixPatternMatcher build();

  private:
    ArraySeriesMatcher::MatcherArray fMatchers;
    int32_t fMatchersLen;
    int32_t fLastTypeOrCp;

    const UnicodeString& fPattern;
    AffixTokenMatcherWarehouse& fWarehouse;
    IgnorablesMatcher* fIgnorables;

    void addMatcher(NumberParseMatcher& matcher) override;
};


// Exported as U_I18N_API for tests
class U_I18N_API AffixPatternMatcher : public ArraySeriesMatcher {
  public:
    AffixPatternMatcher() = default;  // WARNING: Leaves the object in an unusable state

    static AffixPatternMatcher fromAffixPattern(const UnicodeString& affixPattern,
                                                AffixTokenMatcherWarehouse& warehouse,
                                                parse_flags_t parseFlags, bool* success,
                                                UErrorCode& status);

    UnicodeString getPattern() const;

    bool operator==(const AffixPatternMatcher& other) const;

  private:
    CompactUnicodeString<4> fPattern;

    AffixPatternMatcher(MatcherArray& matchers, int32_t matchersLen, const UnicodeString& pattern);

    friend class AffixPatternMatcherBuilder;
};


class AffixMatcher : public NumberParseMatcher, public UMemory {
  public:
    AffixMatcher() = default;  // WARNING: Leaves the object in an unusable state

    AffixMatcher(AffixPatternMatcher* prefix, AffixPatternMatcher* suffix, result_flags_t flags);

    bool match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const override;

    void postProcess(ParsedNumber& result) const override;

    bool smokeTest(const StringSegment& segment) const override;

    int8_t compareTo(const AffixMatcher& rhs) const;

    UnicodeString toString() const override;

  private:
    AffixPatternMatcher* fPrefix;
    AffixPatternMatcher* fSuffix;
    result_flags_t fFlags;
};


/**
 * A C++-only class to retain ownership of the AffixMatchers needed for parsing.
 */
class AffixMatcherWarehouse {
  public:
    AffixMatcherWarehouse() = default;  // WARNING: Leaves the object in an unusable state

    AffixMatcherWarehouse(AffixTokenMatcherWarehouse* tokenWarehouse);

    void createAffixMatchers(const AffixPatternProvider& patternInfo, MutableMatcherCollection& output,
                             const IgnorablesMatcher& ignorables, parse_flags_t parseFlags,
                             UErrorCode& status);

  private:
    // 9 is the limit: positive, zero, and negative, each with prefix, suffix, and prefix+suffix
    AffixMatcher fAffixMatchers[9];
    // 6 is the limit: positive, zero, and negative, a prefix and a suffix for each
    AffixPatternMatcher fAffixPatternMatchers[6];
    // Reference to the warehouse for tokens used by the AffixPatternMatchers
    AffixTokenMatcherWarehouse* fTokenWarehouse;

    friend class AffixMatcher;

    static bool isInteresting(const AffixPatternProvider& patternInfo, const IgnorablesMatcher& ignorables,
                              parse_flags_t parseFlags, UErrorCode& status);
};


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__NUMPARSE_AFFIXES_H__



// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN namespace numparse {
namespace impl {


class ValidationMatcher : public NumberParseMatcher {
  public:
    bool match(StringSegment&, ParsedNumber&, UErrorCode&) const U_OVERRIDE {
        // No-op
        return false;
    }

    bool smokeTest(const StringSegment&) const U_OVERRIDE {
        // No-op
        return false;
    }

    void postProcess(ParsedNumber& result) const U_OVERRIDE = 0;
};


class RequireAffixValidator : public ValidationMatcher, public UMemory {
  public:
    void postProcess(ParsedNumber& result) const U_OVERRIDE;

    UnicodeString toString() const U_OVERRIDE;
};


class RequireCurrencyValidator : public ValidationMatcher, public UMemory {
  public:
    void postProcess(ParsedNumber& result) const U_OVERRIDE;

    UnicodeString toString() const U_OVERRIDE;
};


class RequireDecimalSeparatorValidator : public ValidationMatcher, public UMemory {
  public:
    RequireDecimalSeparatorValidator() = default;  // leaves instance in valid but undefined state

    RequireDecimalSeparatorValidator(bool patternHasDecimalSeparator);

    void postProcess(ParsedNumber& result) const U_OVERRIDE;

    UnicodeString toString() const U_OVERRIDE;

  private:
    bool fPatternHasDecimalSeparator;
};


class RequireNumberValidator : public ValidationMatcher, public UMemory {
  public:
    void postProcess(ParsedNumber& result) const U_OVERRIDE;

    UnicodeString toString() const U_OVERRIDE;
};


/**
 * Wraps a {@link Multiplier} for use in the number parsing pipeline.
 */
class MultiplierParseHandler : public ValidationMatcher, public UMemory {
  public:
    MultiplierParseHandler() = default;  // leaves instance in valid but undefined state

    MultiplierParseHandler(::icu::number::Scale multiplier);

    void postProcess(ParsedNumber& result) const U_OVERRIDE;

    UnicodeString toString() const U_OVERRIDE;

  private:
    ::icu::number::Scale fMultiplier;
};


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__SOURCE_NUMPARSE_VALIDATORS_H__

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN namespace number {
namespace impl {


/**
 * Wraps a {@link Multiplier} for use in the number formatting pipeline.
 */
// Exported as U_I18N_API for tests
class U_I18N_API MultiplierFormatHandler : public MicroPropsGenerator, public UMemory {
  public:
    MultiplierFormatHandler() = default; // WARNING: Leaves object in an unusable state; call setAndChain()

    void setAndChain(const Scale& multiplier, const MicroPropsGenerator* parent);

    void processQuantity(DecimalQuantity& quantity, MicroProps& micros,
                         UErrorCode& status) const U_OVERRIDE;

  private:
    Scale fMultiplier;
    const MicroPropsGenerator *fParent;
};


/** Gets a Scale from a DecimalFormatProperties. In Java, defined in RoundingUtils.java */
static inline Scale scaleFromProperties(const DecimalFormatProperties& properties) {
    int32_t magnitudeMultiplier = properties.magnitudeMultiplier + properties.multiplierScale;
    int32_t arbitraryMultiplier = properties.multiplier;
    if (magnitudeMultiplier != 0 && arbitraryMultiplier != 1) {
        return Scale::byDoubleAndPowerOfTen(arbitraryMultiplier, magnitudeMultiplier);
    } else if (magnitudeMultiplier != 0) {
        return Scale::powerOfTen(magnitudeMultiplier);
    } else if (arbitraryMultiplier != 1) {
        return Scale::byDouble(arbitraryMultiplier);
    } else {
        return Scale::none();
    }
}


} // namespace impl
} // namespace number
U_NAMESPACE_END

#endif //__SOURCE_NUMBER_MULTIPLIER_H__



U_NAMESPACE_BEGIN

// Export an explicit template instantiation of the MaybeStackArray that is used as a data member of NumberParserImpl.
// When building DLLs for Windows this is required even though no direct access to the MaybeStackArray leaks out of the i18n library.
// (See numparse_compositions.h, numparse_affixes.h, datefmt.h, and others for similar examples.)
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN
template class U_I18N_API MaybeStackArray<const numparse::impl::NumberParseMatcher*, 10>;
#endif

namespace numparse {
namespace impl {

// Exported as U_I18N_API for tests
class U_I18N_API NumberParserImpl : public MutableMatcherCollection, public UMemory {
  public:
    virtual ~NumberParserImpl();

    static NumberParserImpl* createSimpleParser(const Locale& locale, const UnicodeString& patternString,
                                                parse_flags_t parseFlags, UErrorCode& status);

    static NumberParserImpl* createParserFromProperties(
            const number::impl::DecimalFormatProperties& properties, const DecimalFormatSymbols& symbols,
            bool parseCurrency, UErrorCode& status);

    /**
     * Does NOT take ownership of the matcher. The matcher MUST remain valid for the lifespan of the
     * NumberParserImpl.
     * @param matcher The matcher to reference.
     */
    void addMatcher(NumberParseMatcher& matcher) override;

    void freeze();

    parse_flags_t getParseFlags() const;

    void parse(const UnicodeString& input, bool greedy, ParsedNumber& result, UErrorCode& status) const;

    void parse(const UnicodeString& input, int32_t start, bool greedy, ParsedNumber& result,
               UErrorCode& status) const;

    UnicodeString toString() const;

  private:
    parse_flags_t fParseFlags;
    int32_t fNumMatchers = 0;
    // NOTE: The stack capacity for fMatchers and fLeads should be the same
    MaybeStackArray<const NumberParseMatcher*, 10> fMatchers;
    bool fFrozen = false;

    // WARNING: All of these matchers start in an undefined state (default-constructed).
    // You must use an assignment operator on them before using.
    struct {
        IgnorablesMatcher ignorables;
        InfinityMatcher infinity;
        MinusSignMatcher minusSign;
        NanMatcher nan;
        PaddingMatcher padding;
        PercentMatcher percent;
        PermilleMatcher permille;
        PlusSignMatcher plusSign;
        DecimalMatcher decimal;
        ScientificMatcher scientific;
        CombinedCurrencyMatcher currency;
        AffixMatcherWarehouse affixMatcherWarehouse;
        AffixTokenMatcherWarehouse affixTokenMatcherWarehouse;
    } fLocalMatchers;
    struct {
        RequireAffixValidator affix;
        RequireCurrencyValidator currency;
        RequireDecimalSeparatorValidator decimalSeparator;
        RequireNumberValidator number;
        MultiplierParseHandler multiplier;
    } fLocalValidators;

    explicit NumberParserImpl(parse_flags_t parseFlags);

    void parseGreedy(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const;

    void parseLongestRecursive(
        StringSegment& segment, ParsedNumber& result, int32_t recursionLevels, UErrorCode& status) const;
};


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__NUMPARSE_IMPL_H__


U_NAMESPACE_BEGIN
namespace number {
namespace impl {


class PropertiesAffixPatternProvider : public AffixPatternProvider, public UMemory {
  public:
    bool isBogus() const {
        return fBogus;
    }

    void setToBogus() {
        fBogus = true;
    }

    void setTo(const DecimalFormatProperties& properties, UErrorCode& status);

    PropertiesAffixPatternProvider() = default; // puts instance in valid but undefined state

    PropertiesAffixPatternProvider(const DecimalFormatProperties& properties, UErrorCode& status) {
        setTo(properties, status);
    }

    // AffixPatternProvider Methods:

    char16_t charAt(int32_t flags, int32_t i) const U_OVERRIDE;

    int32_t length(int32_t flags) const U_OVERRIDE;

    UnicodeString getString(int32_t flags) const U_OVERRIDE;

    bool hasCurrencySign() const U_OVERRIDE;

    bool positiveHasPlusSign() const U_OVERRIDE;

    bool hasNegativeSubpattern() const U_OVERRIDE;

    bool negativeHasMinusSign() const U_OVERRIDE;

    bool containsSymbolType(AffixPatternType, UErrorCode&) const U_OVERRIDE;

    bool hasBody() const U_OVERRIDE;

  private:
    UnicodeString posPrefix;
    UnicodeString posSuffix;
    UnicodeString negPrefix;
    UnicodeString negSuffix;
    bool isCurrencyPattern;

    const UnicodeString& getStringInternal(int32_t flags) const;

    bool fBogus{true};
};


class CurrencyPluralInfoAffixProvider : public AffixPatternProvider, public UMemory {
  public:
    bool isBogus() const {
        return fBogus;
    }

    void setToBogus() {
        fBogus = true;
    }

    void setTo(const CurrencyPluralInfo& cpi, const DecimalFormatProperties& properties,
               UErrorCode& status);

    // AffixPatternProvider Methods:

    char16_t charAt(int32_t flags, int32_t i) const U_OVERRIDE;

    int32_t length(int32_t flags) const U_OVERRIDE;

    UnicodeString getString(int32_t flags) const U_OVERRIDE;

    bool hasCurrencySign() const U_OVERRIDE;

    bool positiveHasPlusSign() const U_OVERRIDE;

    bool hasNegativeSubpattern() const U_OVERRIDE;

    bool negativeHasMinusSign() const U_OVERRIDE;

    bool containsSymbolType(AffixPatternType, UErrorCode&) const U_OVERRIDE;

    bool hasBody() const U_OVERRIDE;

  private:
    PropertiesAffixPatternProvider affixesByPlural[StandardPlural::COUNT];

    bool fBogus{true};
};


/**
 * A struct for ownership of a few objects needed for formatting.
 */
struct DecimalFormatWarehouse {
    PropertiesAffixPatternProvider propertiesAPP;
    CurrencyPluralInfoAffixProvider currencyPluralInfoAPP;
    CurrencySymbols currencySymbols;
};


/**
* Internal fields for DecimalFormat.
* TODO: Make some of these fields by value instead of by LocalPointer?
*/
struct DecimalFormatFields : public UMemory {

    DecimalFormatFields() {}

    DecimalFormatFields(const DecimalFormatProperties& propsToCopy)
        : properties(propsToCopy) {}

    /** The property bag corresponding to user-specified settings and settings from the pattern string. */
    DecimalFormatProperties properties;

    /** The symbols for the current locale. */
    LocalPointer<const DecimalFormatSymbols> symbols;

    /**
    * The pre-computed formatter object. Setters cause this to be re-computed atomically. The {@link
    * #format} method uses the formatter directly without needing to synchronize.
    */
    LocalizedNumberFormatter formatter;

    /** The lazy-computed parser for .parse() */
    std::atomic<::icu::numparse::impl::NumberParserImpl*> atomicParser = {};

    /** The lazy-computed parser for .parseCurrency() */
    std::atomic<::icu::numparse::impl::NumberParserImpl*> atomicCurrencyParser = {};

    /** Small object ownership warehouse for the formatter and parser */
    DecimalFormatWarehouse warehouse;

    /** The effective properties as exported from the formatter object. Used by some getters. */
    DecimalFormatProperties exportedProperties;

    // Data for fastpath
    bool canUseFastFormat = false;
    struct FastFormatData {
        char16_t cpZero;
        char16_t cpGroupingSeparator;
        char16_t cpMinusSign;
        int8_t minInt;
        int8_t maxInt;
    } fastData;
};


/**
 * Utilities for converting between a DecimalFormatProperties and a MacroProps.
 */
class NumberPropertyMapper {
  public:
    /** Convenience method to create a NumberFormatter directly from Properties. */
    static UnlocalizedNumberFormatter create(const DecimalFormatProperties& properties,
                                             const DecimalFormatSymbols& symbols,
                                             DecimalFormatWarehouse& warehouse, UErrorCode& status);

    /** Convenience method to create a NumberFormatter directly from Properties. */
    static UnlocalizedNumberFormatter create(const DecimalFormatProperties& properties,
                                             const DecimalFormatSymbols& symbols,
                                             DecimalFormatWarehouse& warehouse,
                                             DecimalFormatProperties& exportedProperties,
                                             UErrorCode& status);

    /**
     * Creates a new {@link MacroProps} object based on the content of a {@link DecimalFormatProperties}
     * object. In other words, maps Properties to MacroProps. This function is used by the
     * JDK-compatibility API to call into the ICU 60 fluent number formatting pipeline.
     *
     * @param properties
     *            The property bag to be mapped.
     * @param symbols
     *            The symbols associated with the property bag.
     * @param exportedProperties
     *            A property bag in which to store validated properties. Used by some DecimalFormat
     *            getters.
     * @return A new MacroProps containing all of the information in the Properties.
     */
    static MacroProps oldToNew(const DecimalFormatProperties& properties,
                               const DecimalFormatSymbols& symbols, DecimalFormatWarehouse& warehouse,
                               DecimalFormatProperties* exportedProperties, UErrorCode& status);
};


} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__NUMBER_MAPPER_H__



using namespace icu;


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(CompactDecimalFormat)


CompactDecimalFormat*
CompactDecimalFormat::createInstance(const Locale& inLocale, UNumberCompactStyle style,
                                     UErrorCode& status) {
    return new CompactDecimalFormat(inLocale, style, status);
}

CompactDecimalFormat::CompactDecimalFormat(const Locale& inLocale, UNumberCompactStyle style,
                                           UErrorCode& status)
        : DecimalFormat(new DecimalFormatSymbols(inLocale, status), status) {
    if (U_FAILURE(status)) return;
    // Minimal properties: let the non-shim code path do most of the logic for us.
    fields->properties.compactStyle = style;
    fields->properties.groupingSize = -2; // do not forward grouping information
    fields->properties.minimumGroupingDigits = 2;
    touch(status);
}

CompactDecimalFormat::CompactDecimalFormat(const CompactDecimalFormat& source) = default;

CompactDecimalFormat::~CompactDecimalFormat() = default;

CompactDecimalFormat& CompactDecimalFormat::operator=(const CompactDecimalFormat& rhs) {
    DecimalFormat::operator=(rhs);
    return *this;
}

CompactDecimalFormat* CompactDecimalFormat::clone() const {
    return new CompactDecimalFormat(*this);
}

void
CompactDecimalFormat::parse(
        const UnicodeString& /* text */,
        Formattable& /* result */,
        ParsePosition& /* parsePosition */) const {
}

void
CompactDecimalFormat::parse(
        const UnicodeString& /* text */,
        Formattable& /* result */,
        UErrorCode& status) const {
    status = U_UNSUPPORTED_ERROR;
}

CurrencyAmount*
CompactDecimalFormat::parseCurrency(
        const UnicodeString& /* text */,
        ParsePosition& /* pos */) const {
    return nullptr;
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2003 - 2013, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING




#include <float.h>

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(CopticCalendar)

static const int32_t COPTIC_JD_EPOCH_OFFSET  = 1824665;

//-------------------------------------------------------------------------
// Constructors...
//-------------------------------------------------------------------------

CopticCalendar::CopticCalendar(const Locale& aLocale, UErrorCode& success)
: CECalendar(aLocale, success)
{
}

CopticCalendar::CopticCalendar (const CopticCalendar& other)
: CECalendar(other)
{
}

CopticCalendar::~CopticCalendar()
{
}

CopticCalendar*
CopticCalendar::clone() const
{
    return new CopticCalendar(*this);
}

const char*
CopticCalendar::getType() const
{
    return "coptic";
}

//-------------------------------------------------------------------------
// Calendar framework
//-------------------------------------------------------------------------

int32_t
CopticCalendar::handleGetExtendedYear()
{
    int32_t eyear;
    if (newerField(UCAL_EXTENDED_YEAR, UCAL_YEAR) == UCAL_EXTENDED_YEAR) {
        eyear = internalGet(UCAL_EXTENDED_YEAR, 1); // Default to year 1
    } else {
        // The year defaults to the epoch start, the era to CE
        int32_t era = internalGet(UCAL_ERA, CE);
        if (era == BCE) {
            eyear = 1 - internalGet(UCAL_YEAR, 1); // Convert to extended year
        } else {
            eyear = internalGet(UCAL_YEAR, 1); // Default to year 1
        }
    }
    return eyear;
}

void
CopticCalendar::handleComputeFields(int32_t julianDay, UErrorCode &/*status*/)
{
    int32_t eyear, month, day, era, year;
    jdToCE(julianDay, getJDEpochOffset(), eyear, month, day);

    if (eyear <= 0) {
        era = BCE;
        year = 1 - eyear;
    } else {
        era = CE;
        year = eyear;
    }

    internalSet(UCAL_EXTENDED_YEAR, eyear);
    internalSet(UCAL_ERA, era);
    internalSet(UCAL_YEAR, year);
    internalSet(UCAL_MONTH, month);
    internalSet(UCAL_DATE, day);
    internalSet(UCAL_DAY_OF_YEAR, (30 * month) + day);
}

/**
 * The system maintains a static default century start date and Year.  They are
 * initialized the first time they are used.  Once the system default century date
 * and year are set, they do not change.
 */
static UDate           gSystemDefaultCenturyStart       = DBL_MIN;
static int32_t         gSystemDefaultCenturyStartYear   = -1;
static icu::UInitOnce  gSystemDefaultCenturyInit        = U_INITONCE_INITIALIZER;


static void U_CALLCONV coptccal_initializeSystemDefaultCentury() {
    UErrorCode status = U_ZERO_ERROR;
    CopticCalendar calendar(Locale("@calendar=coptic"), status);
    if (U_SUCCESS(status)) {
        calendar.setTime(Calendar::getNow(), status);
        calendar.add(UCAL_YEAR, -80, status);
        gSystemDefaultCenturyStart = calendar.getTime(status);
        gSystemDefaultCenturyStartYear = calendar.get(UCAL_YEAR, status);
    }
    // We have no recourse upon failure unless we want to propagate the failure
    // out.
}

UDate
CopticCalendar::defaultCenturyStart() const
{
    // lazy-evaluate systemDefaultCenturyStart
    umtx_initOnce(gSystemDefaultCenturyInit, &coptccal_initializeSystemDefaultCentury);
    return gSystemDefaultCenturyStart;
}

int32_t
CopticCalendar::defaultCenturyStartYear() const
{
    // lazy-evaluate systemDefaultCenturyStart
    umtx_initOnce(gSystemDefaultCenturyInit, &coptccal_initializeSystemDefaultCentury);
    return gSystemDefaultCenturyStartYear;
}


int32_t
CopticCalendar::getJDEpochOffset() const
{
    return COPTIC_JD_EPOCH_OFFSET;
}


#if 0
// We do not want to introduce this API in ICU4C.
// It was accidentally introduced in ICU4J as a public API.

//-------------------------------------------------------------------------
// Calendar system Conversion methods...
//-------------------------------------------------------------------------

int32_t
CopticCalendar::copticToJD(int32_t year, int32_t month, int32_t day)
{
    return CECalendar::ceToJD(year, month, day, COPTIC_JD_EPOCH_OFFSET);
}
#endif

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 26, 2004
* Since: ICU 3.0
**********************************************************************
*/


#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

CurrencyAmount::CurrencyAmount(const Formattable& amount, ConstChar16Ptr isoCode,
                               UErrorCode& ec) :
    Measure(amount, new CurrencyUnit(isoCode, ec), ec) {
}

CurrencyAmount::CurrencyAmount(double amount, ConstChar16Ptr isoCode,
                               UErrorCode& ec) :
    Measure(Formattable(amount), new CurrencyUnit(isoCode, ec), ec) {
}

CurrencyAmount::CurrencyAmount(const CurrencyAmount& other) :
    Measure(other) {
}

CurrencyAmount& CurrencyAmount::operator=(const CurrencyAmount& other) {
    Measure::operator=(other);
    return *this;
}

CurrencyAmount* CurrencyAmount::clone() const {
    return new CurrencyAmount(*this);
}

CurrencyAmount::~CurrencyAmount() {
}
    
UOBJECT_DEFINE_RTTI_IMPLEMENTATION(CurrencyAmount)

U_NAMESPACE_END

#endif // !UCONFIG_NO_FORMATTING
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2014 International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 20, 2004
* Since: ICU 3.0
**********************************************************************
*/


#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 20, 2004
* Since: ICU 3.0
**********************************************************************
*/



#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 20, 2004
* Since: ICU 3.0
**********************************************************************
*/



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 1996-2016, International Business Machines
 * Corporation and others. All Rights Reserved.
 *******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING






/**
 * \file
 * \brief C API: DateFormat
 *
 * <h2> Date Format C API</h2>
 *
 * Date Format C API  consists of functions that convert dates and
 * times from their internal representations to textual form and back again in a
 * language-independent manner. Converting from the internal representation (milliseconds
 * since midnight, January 1, 1970) to text is known as "formatting," and converting
 * from text to millis is known as "parsing."  We currently define only one concrete
 * structure UDateFormat, which can handle pretty much all normal
 * date formatting and parsing actions.
 * <P>
 * Date Format helps you to format and parse dates for any locale. Your code can
 * be completely independent of the locale conventions for months, days of the
 * week, or even the calendar format: lunar vs. solar.
 * <P>
 * To format a date for the current Locale with default time and date style,
 * use one of the static factory methods:
 * <pre>
 * \code
 *  UErrorCode status = U_ZERO_ERROR;
 *  UChar *myString;
 *  int32_t myStrlen = 0;
 *  UDateFormat* dfmt = udat_open(UDAT_DEFAULT, UDAT_DEFAULT, NULL, NULL, -1, NULL, -1, &status);
 *  myStrlen = udat_format(dfmt, myDate, NULL, myStrlen, NULL, &status);
 *  if (status==U_BUFFER_OVERFLOW_ERROR){
 *      status=U_ZERO_ERROR;
 *      myString=(UChar*)malloc(sizeof(UChar) * (myStrlen+1) );
 *      udat_format(dfmt, myDate, myString, myStrlen+1, NULL, &status);
 *  }
 * \endcode
 * </pre>
 * If you are formatting multiple numbers, it is more efficient to get the
 * format and use it multiple times so that the system doesn't have to fetch the
 * information about the local language and country conventions multiple times.
 * <pre>
 * \code
 *  UErrorCode status = U_ZERO_ERROR;
 *  int32_t i, myStrlen = 0;
 *  UChar* myString;
 *  char buffer[1024];
 *  UDate myDateArr[] = { 0.0, 100000000.0, 2000000000.0 }; // test values
 *  UDateFormat* df = udat_open(UDAT_DEFAULT, UDAT_DEFAULT, NULL, NULL, -1, NULL, 0, &status);
 *  for (i = 0; i < 3; i++) {
 *      myStrlen = udat_format(df, myDateArr[i], NULL, myStrlen, NULL, &status);
 *      if(status == U_BUFFER_OVERFLOW_ERROR){
 *          status = U_ZERO_ERROR;
 *          myString = (UChar*)malloc(sizeof(UChar) * (myStrlen+1) );
 *          udat_format(df, myDateArr[i], myString, myStrlen+1, NULL, &status);
 *          printf("%s\n", u_austrcpy(buffer, myString) );
 *          free(myString);
 *      }
 *  }
 * \endcode
 * </pre>
 * To get specific fields of a date, you can use UFieldPosition to
 * get specific fields.
 * <pre>
 * \code
 *  UErrorCode status = U_ZERO_ERROR;
 *  UFieldPosition pos;
 *  UChar *myString;
 *  int32_t myStrlen = 0;
 *  char buffer[1024];
 *
 *  pos.field = 1;  // Same as the DateFormat::EField enum
 *  UDateFormat* dfmt = udat_open(UDAT_DEFAULT, UDAT_DEFAULT, NULL, -1, NULL, 0, &status);
 *  myStrlen = udat_format(dfmt, myDate, NULL, myStrlen, &pos, &status);
 *  if (status==U_BUFFER_OVERFLOW_ERROR){
 *      status=U_ZERO_ERROR;
 *      myString=(UChar*)malloc(sizeof(UChar) * (myStrlen+1) );
 *      udat_format(dfmt, myDate, myString, myStrlen+1, &pos, &status);
 *  }
 *  printf("date format: %s\n", u_austrcpy(buffer, myString));
 *  buffer[pos.endIndex] = 0;   // NULL terminate the string.
 *  printf("UFieldPosition position equals %s\n", &buffer[pos.beginIndex]);
 * \endcode
 * </pre>
 * To format a date for a different Locale, specify it in the call to
 * udat_open()
 * <pre>
 * \code
 *        UDateFormat* df = udat_open(UDAT_SHORT, UDAT_SHORT, "fr_FR", NULL, -1, NULL, 0, &status);
 * \endcode
 * </pre>
 * You can use a DateFormat API udat_parse() to parse.
 * <pre>
 * \code
 *  UErrorCode status = U_ZERO_ERROR;
 *  int32_t parsepos=0;
 *  UDate myDate = udat_parse(df, myString, u_strlen(myString), &parsepos, &status);
 * \endcode
 * </pre>
 *  You can pass in different options for the arguments for date and time style
 *  to control the length of the result; from SHORT to MEDIUM to LONG to FULL.
 *  The exact result depends on the locale, but generally:
 *  see UDateFormatStyle for more details
 * <ul type=round>
 *   <li>   UDAT_SHORT is completely numeric, such as 12/13/52 or 3:30pm
 *   <li>   UDAT_MEDIUM is longer, such as Jan 12, 1952
 *   <li>   UDAT_LONG is longer, such as January 12, 1952 or 3:30:32pm
 *   <li>   UDAT_FULL is pretty completely specified, such as
 *          Tuesday, April 12, 1952 AD or 3:30:42pm PST.
 * </ul>
 * You can also set the time zone on the format if you wish.
 * <P>
 * You can also use forms of the parse and format methods with Parse Position and
 * UFieldPosition to allow you to
 * <ul type=round>
 *   <li>   Progressively parse through pieces of a string.
 *   <li>   Align any particular field, or find out where it is for selection
 *          on the screen.
 * </ul>
 * <p><strong>Date and Time Patterns:</strong></p>
 *
 * <p>Date and time formats are specified by <em>date and time pattern</em> strings.
 * Within date and time pattern strings, all unquoted ASCII letters [A-Za-z] are reserved
 * as pattern letters representing calendar fields. <code>UDateFormat</code> supports
 * the date and time formatting algorithm and pattern letters defined by
 * <a href="http://www.unicode.org/reports/tr35/tr35-dates.html#Date_Field_Symbol_Table">UTS#35
 * Unicode Locale Data Markup Language (LDML)</a> and further documented for ICU in the
 * <a href="https://sites.google.com/site/icuprojectuserguide/formatparse/datetime?pli=1#TOC-Date-Field-Symbol-Table">ICU
 * User Guide</a>.</p>
 */

/** A date formatter.
 *  For usage in C programs.
 *  @stable ICU 2.6
 */
typedef void* UDateFormat;

/** The possible date/time format styles
 *  @stable ICU 2.6
 */
typedef enum UDateFormatStyle {
    /** Full style */
    UDAT_FULL,
    /** Long style */
    UDAT_LONG,
    /** Medium style */
    UDAT_MEDIUM,
    /** Short style */
    UDAT_SHORT,
    /** Default style */
    UDAT_DEFAULT = UDAT_MEDIUM,

    /** Bitfield for relative date */
    UDAT_RELATIVE = (1 << 7),

    UDAT_FULL_RELATIVE = UDAT_FULL | UDAT_RELATIVE,

    UDAT_LONG_RELATIVE = UDAT_LONG | UDAT_RELATIVE,

    UDAT_MEDIUM_RELATIVE = UDAT_MEDIUM | UDAT_RELATIVE,

    UDAT_SHORT_RELATIVE = UDAT_SHORT | UDAT_RELATIVE,


    /** No style */
    UDAT_NONE = -1,

    /**
     * Use the pattern given in the parameter to udat_open
     * @see udat_open
     * @stable ICU 50
     */
    UDAT_PATTERN = -2,

#ifndef U_HIDE_INTERNAL_API
    /** @internal alias to UDAT_PATTERN */
    UDAT_IGNORE = UDAT_PATTERN
#endif /* U_HIDE_INTERNAL_API */
} UDateFormatStyle;

/* Skeletons for dates. */

/**
 * Constant for date skeleton with year.
 * @stable ICU 4.0
 */
#define UDAT_YEAR                       "y"
/**
 * Constant for date skeleton with quarter.
 * @stable ICU 51
 */
#define UDAT_QUARTER                    "QQQQ"
/**
 * Constant for date skeleton with abbreviated quarter.
 * @stable ICU 51
 */
#define UDAT_ABBR_QUARTER               "QQQ"
/**
 * Constant for date skeleton with year and quarter.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_QUARTER               "yQQQQ"
/**
 * Constant for date skeleton with year and abbreviated quarter.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_ABBR_QUARTER          "yQQQ"
/**
 * Constant for date skeleton with month.
 * @stable ICU 4.0
 */
#define UDAT_MONTH                      "MMMM"
/**
 * Constant for date skeleton with abbreviated month.
 * @stable ICU 4.0
 */
#define UDAT_ABBR_MONTH                 "MMM"
/**
 * Constant for date skeleton with numeric month.
 * @stable ICU 4.0
 */
#define UDAT_NUM_MONTH                  "M"
/**
 * Constant for date skeleton with year and month.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_MONTH                 "yMMMM"
/**
 * Constant for date skeleton with year and abbreviated month.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_ABBR_MONTH            "yMMM"
/**
 * Constant for date skeleton with year and numeric month.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_NUM_MONTH             "yM"
/**
 * Constant for date skeleton with day.
 * @stable ICU 4.0
 */
#define UDAT_DAY                        "d"
/**
 * Constant for date skeleton with year, month, and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_MONTH_DAY             "yMMMMd"
/**
 * Constant for date skeleton with year, abbreviated month, and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_ABBR_MONTH_DAY        "yMMMd"
/**
 * Constant for date skeleton with year, numeric month, and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_NUM_MONTH_DAY         "yMd"
/**
 * Constant for date skeleton with weekday.
 * @stable ICU 51
 */
#define UDAT_WEEKDAY                    "EEEE"
/**
 * Constant for date skeleton with abbreviated weekday.
 * @stable ICU 51
 */
#define UDAT_ABBR_WEEKDAY               "E"
/**
 * Constant for date skeleton with year, month, weekday, and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_MONTH_WEEKDAY_DAY     "yMMMMEEEEd"
/**
 * Constant for date skeleton with year, abbreviated month, weekday, and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_ABBR_MONTH_WEEKDAY_DAY "yMMMEd"
/**
 * Constant for date skeleton with year, numeric month, weekday, and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_YEAR_NUM_MONTH_WEEKDAY_DAY "yMEd"
/**
 * Constant for date skeleton with long month and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_MONTH_DAY                  "MMMMd"
/**
 * Constant for date skeleton with abbreviated month and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_ABBR_MONTH_DAY             "MMMd"
/**
 * Constant for date skeleton with numeric month and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_NUM_MONTH_DAY              "Md"
/**
 * Constant for date skeleton with month, weekday, and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_MONTH_WEEKDAY_DAY          "MMMMEEEEd"
/**
 * Constant for date skeleton with abbreviated month, weekday, and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_ABBR_MONTH_WEEKDAY_DAY     "MMMEd"
/**
 * Constant for date skeleton with numeric month, weekday, and day.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_NUM_MONTH_WEEKDAY_DAY      "MEd"

/* Skeletons for times. */

/**
 * Constant for date skeleton with hour, with the locale's preferred hour format (12 or 24).
 * @stable ICU 4.0
 */
#define UDAT_HOUR                       "j"
/**
 * Constant for date skeleton with hour in 24-hour presentation.
 * @stable ICU 51
 */
#define UDAT_HOUR24                     "H"
/**
 * Constant for date skeleton with minute.
 * @stable ICU 51
 */
#define UDAT_MINUTE                     "m"
/**
 * Constant for date skeleton with hour and minute, with the locale's preferred hour format (12 or 24).
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_HOUR_MINUTE                "jm"
/**
 * Constant for date skeleton with hour and minute in 24-hour presentation.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_HOUR24_MINUTE              "Hm"
/**
 * Constant for date skeleton with second.
 * @stable ICU 51
 */
#define UDAT_SECOND                     "s"
/**
 * Constant for date skeleton with hour, minute, and second,
 * with the locale's preferred hour format (12 or 24).
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_HOUR_MINUTE_SECOND         "jms"
/**
 * Constant for date skeleton with hour, minute, and second in
 * 24-hour presentation.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_HOUR24_MINUTE_SECOND       "Hms"
/**
 * Constant for date skeleton with minute and second.
 * Used in combinations date + time, date + time + zone, or time + zone.
 * @stable ICU 4.0
 */
#define UDAT_MINUTE_SECOND              "ms"

/* Skeletons for time zones. */

/**
 * Constant for <i>generic location format</i>, such as Los Angeles Time;
 * used in combinations date + time + zone, or time + zone.
 * @see <a href="http://unicode.org/reports/tr35/#Date_Format_Patterns">LDML Date Format Patterns</a>
 * @see <a href="http://unicode.org/reports/tr35/#Time_Zone_Fallback">LDML Time Zone Fallback</a>
 * @stable ICU 51
 */
#define UDAT_LOCATION_TZ "VVVV"
/**
 * Constant for <i>generic non-location format</i>, such as Pacific Time;
 * used in combinations date + time + zone, or time + zone.
 * @see <a href="http://unicode.org/reports/tr35/#Date_Format_Patterns">LDML Date Format Patterns</a>
 * @see <a href="http://unicode.org/reports/tr35/#Time_Zone_Fallback">LDML Time Zone Fallback</a>
 * @stable ICU 51
 */
#define UDAT_GENERIC_TZ "vvvv"
/**
 * Constant for <i>generic non-location format</i>, abbreviated if possible, such as PT;
 * used in combinations date + time + zone, or time + zone.
 * @see <a href="http://unicode.org/reports/tr35/#Date_Format_Patterns">LDML Date Format Patterns</a>
 * @see <a href="http://unicode.org/reports/tr35/#Time_Zone_Fallback">LDML Time Zone Fallback</a>
 * @stable ICU 51
 */
#define UDAT_ABBR_GENERIC_TZ "v"
/**
 * Constant for <i>specific non-location format</i>, such as Pacific Daylight Time;
 * used in combinations date + time + zone, or time + zone.
 * @see <a href="http://unicode.org/reports/tr35/#Date_Format_Patterns">LDML Date Format Patterns</a>
 * @see <a href="http://unicode.org/reports/tr35/#Time_Zone_Fallback">LDML Time Zone Fallback</a>
 * @stable ICU 51
 */
#define UDAT_SPECIFIC_TZ "zzzz"
/**
 * Constant for <i>specific non-location format</i>, abbreviated if possible, such as PDT;
 * used in combinations date + time + zone, or time + zone.
 * @see <a href="http://unicode.org/reports/tr35/#Date_Format_Patterns">LDML Date Format Patterns</a>
 * @see <a href="http://unicode.org/reports/tr35/#Time_Zone_Fallback">LDML Time Zone Fallback</a>
 * @stable ICU 51
 */
#define UDAT_ABBR_SPECIFIC_TZ "z"
/**
 * Constant for <i>localized GMT/UTC format</i>, such as GMT+8:00 or HPG-8:00;
 * used in combinations date + time + zone, or time + zone.
 * @see <a href="http://unicode.org/reports/tr35/#Date_Format_Patterns">LDML Date Format Patterns</a>
 * @see <a href="http://unicode.org/reports/tr35/#Time_Zone_Fallback">LDML Time Zone Fallback</a>
 * @stable ICU 51
 */
#define UDAT_ABBR_UTC_TZ "ZZZZ"

/* deprecated skeleton constants */

#ifndef U_HIDE_DEPRECATED_API
/**
 * Constant for date skeleton with standalone month.
 * @deprecated ICU 50 Use UDAT_MONTH instead.
 */
#define UDAT_STANDALONE_MONTH           "LLLL"
/**
 * Constant for date skeleton with standalone abbreviated month.
 * @deprecated ICU 50 Use UDAT_ABBR_MONTH instead.
 */
#define UDAT_ABBR_STANDALONE_MONTH      "LLL"

/**
 * Constant for date skeleton with hour, minute, and generic timezone.
 * @deprecated ICU 50 Use instead UDAT_HOUR_MINUTE UDAT_ABBR_GENERIC_TZ or some other timezone presentation.
 */
#define UDAT_HOUR_MINUTE_GENERIC_TZ     "jmv"
/**
 * Constant for date skeleton with hour, minute, and timezone.
 * @deprecated ICU 50 Use instead UDAT_HOUR_MINUTE UDAT_ABBR_SPECIFIC_TZ or some other timezone presentation.
 */
#define UDAT_HOUR_MINUTE_TZ             "jmz"
/**
 * Constant for date skeleton with hour and generic timezone.
 * @deprecated ICU 50 Use instead UDAT_HOUR UDAT_ABBR_GENERIC_TZ or some other timezone presentation.
 */
#define UDAT_HOUR_GENERIC_TZ            "jv"
/**
 * Constant for date skeleton with hour and timezone.
 * @deprecated ICU 50 Use instead UDAT_HOUR UDAT_ABBR_SPECIFIC_TZ or some other timezone presentation.
 */
#define UDAT_HOUR_TZ                    "jz"
#endif  /* U_HIDE_DEPRECATED_API */

#ifndef U_HIDE_INTERNAL_API
/**
 * Constant for Unicode string name of new (in 2019) Japanese calendar era,
 * root/English abbreviated version (ASCII-range characters).
 * @internal
 */
#define JP_ERA_2019_ROOT                "Reiwa"
/**
 * Constant for Unicode string name of new (in 2019) Japanese calendar era,
 * Japanese abbreviated version (Han, or fullwidth Latin for testing).
 * @internal
 */
#define JP_ERA_2019_JA                  "\\u4EE4\\u548C"
/**
 * Constant for Unicode string name of new (in 2019) Japanese calendar era,
 * root and Japanese narrow version (ASCII-range characters).
 * @internal
 */
#define JP_ERA_2019_NARROW              "R"
#endif  // U_HIDE_INTERNAL_API

/**
 * FieldPosition and UFieldPosition selectors for format fields
 * defined by DateFormat and UDateFormat.
 * @stable ICU 3.0
 */
typedef enum UDateFormatField {
    /**
     * FieldPosition and UFieldPosition selector for 'G' field alignment,
     * corresponding to the UCAL_ERA field.
     * @stable ICU 3.0
     */
    UDAT_ERA_FIELD = 0,

    /**
     * FieldPosition and UFieldPosition selector for 'y' field alignment,
     * corresponding to the UCAL_YEAR field.
     * @stable ICU 3.0
     */
    UDAT_YEAR_FIELD = 1,

    /**
     * FieldPosition and UFieldPosition selector for 'M' field alignment,
     * corresponding to the UCAL_MONTH field.
     * @stable ICU 3.0
     */
    UDAT_MONTH_FIELD = 2,

    /**
     * FieldPosition and UFieldPosition selector for 'd' field alignment,
     * corresponding to the UCAL_DATE field.
     * @stable ICU 3.0
     */
    UDAT_DATE_FIELD = 3,

    /**
     * FieldPosition and UFieldPosition selector for 'k' field alignment,
     * corresponding to the UCAL_HOUR_OF_DAY field.
     * UDAT_HOUR_OF_DAY1_FIELD is used for the one-based 24-hour clock.
     * For example, 23:59 + 01:00 results in 24:59.
     * @stable ICU 3.0
     */
    UDAT_HOUR_OF_DAY1_FIELD = 4,

    /**
     * FieldPosition and UFieldPosition selector for 'H' field alignment,
     * corresponding to the UCAL_HOUR_OF_DAY field.
     * UDAT_HOUR_OF_DAY0_FIELD is used for the zero-based 24-hour clock.
     * For example, 23:59 + 01:00 results in 00:59.
     * @stable ICU 3.0
     */
    UDAT_HOUR_OF_DAY0_FIELD = 5,

    /**
     * FieldPosition and UFieldPosition selector for 'm' field alignment,
     * corresponding to the UCAL_MINUTE field.
     * @stable ICU 3.0
     */
    UDAT_MINUTE_FIELD = 6,

    /**
     * FieldPosition and UFieldPosition selector for 's' field alignment,
     * corresponding to the UCAL_SECOND field.
     * @stable ICU 3.0
     */
    UDAT_SECOND_FIELD = 7,

    /**
     * FieldPosition and UFieldPosition selector for 'S' field alignment,
     * corresponding to the UCAL_MILLISECOND field.
     *
     * Note: Time formats that use 'S' can display a maximum of three
     * significant digits for fractional seconds, corresponding to millisecond
     * resolution and a fractional seconds sub-pattern of SSS. If the
     * sub-pattern is S or SS, the fractional seconds value will be truncated
     * (not rounded) to the number of display places specified. If the
     * fractional seconds sub-pattern is longer than SSS, the additional
     * display places will be filled with zeros.
     * @stable ICU 3.0
     */
    UDAT_FRACTIONAL_SECOND_FIELD = 8,

    /**
     * FieldPosition and UFieldPosition selector for 'E' field alignment,
     * corresponding to the UCAL_DAY_OF_WEEK field.
     * @stable ICU 3.0
     */
    UDAT_DAY_OF_WEEK_FIELD = 9,

    /**
     * FieldPosition and UFieldPosition selector for 'D' field alignment,
     * corresponding to the UCAL_DAY_OF_YEAR field.
     * @stable ICU 3.0
     */
    UDAT_DAY_OF_YEAR_FIELD = 10,

    /**
     * FieldPosition and UFieldPosition selector for 'F' field alignment,
     * corresponding to the UCAL_DAY_OF_WEEK_IN_MONTH field.
     * @stable ICU 3.0
     */
    UDAT_DAY_OF_WEEK_IN_MONTH_FIELD = 11,

    /**
     * FieldPosition and UFieldPosition selector for 'w' field alignment,
     * corresponding to the UCAL_WEEK_OF_YEAR field.
     * @stable ICU 3.0
     */
    UDAT_WEEK_OF_YEAR_FIELD = 12,

    /**
     * FieldPosition and UFieldPosition selector for 'W' field alignment,
     * corresponding to the UCAL_WEEK_OF_MONTH field.
     * @stable ICU 3.0
     */
    UDAT_WEEK_OF_MONTH_FIELD = 13,

    /**
     * FieldPosition and UFieldPosition selector for 'a' field alignment,
     * corresponding to the UCAL_AM_PM field.
     * @stable ICU 3.0
     */
    UDAT_AM_PM_FIELD = 14,

    /**
     * FieldPosition and UFieldPosition selector for 'h' field alignment,
     * corresponding to the UCAL_HOUR field.
     * UDAT_HOUR1_FIELD is used for the one-based 12-hour clock.
     * For example, 11:30 PM + 1 hour results in 12:30 AM.
     * @stable ICU 3.0
     */
    UDAT_HOUR1_FIELD = 15,

    /**
     * FieldPosition and UFieldPosition selector for 'K' field alignment,
     * corresponding to the UCAL_HOUR field.
     * UDAT_HOUR0_FIELD is used for the zero-based 12-hour clock.
     * For example, 11:30 PM + 1 hour results in 00:30 AM.
     * @stable ICU 3.0
     */
    UDAT_HOUR0_FIELD = 16,

    /**
     * FieldPosition and UFieldPosition selector for 'z' field alignment,
     * corresponding to the UCAL_ZONE_OFFSET and
     * UCAL_DST_OFFSET fields.
     * @stable ICU 3.0
     */
    UDAT_TIMEZONE_FIELD = 17,

    /**
     * FieldPosition and UFieldPosition selector for 'Y' field alignment,
     * corresponding to the UCAL_YEAR_WOY field.
     * @stable ICU 3.0
     */
    UDAT_YEAR_WOY_FIELD = 18,

    /**
     * FieldPosition and UFieldPosition selector for 'e' field alignment,
     * corresponding to the UCAL_DOW_LOCAL field.
     * @stable ICU 3.0
     */
    UDAT_DOW_LOCAL_FIELD = 19,

    /**
     * FieldPosition and UFieldPosition selector for 'u' field alignment,
     * corresponding to the UCAL_EXTENDED_YEAR field.
     * @stable ICU 3.0
     */
    UDAT_EXTENDED_YEAR_FIELD = 20,

    /**
     * FieldPosition and UFieldPosition selector for 'g' field alignment,
     * corresponding to the UCAL_JULIAN_DAY field.
     * @stable ICU 3.0
     */
    UDAT_JULIAN_DAY_FIELD = 21,

    /**
     * FieldPosition and UFieldPosition selector for 'A' field alignment,
     * corresponding to the UCAL_MILLISECONDS_IN_DAY field.
     * @stable ICU 3.0
     */
    UDAT_MILLISECONDS_IN_DAY_FIELD = 22,

    /**
     * FieldPosition and UFieldPosition selector for 'Z' field alignment,
     * corresponding to the UCAL_ZONE_OFFSET and
     * UCAL_DST_OFFSET fields.
     * @stable ICU 3.0
     */
    UDAT_TIMEZONE_RFC_FIELD = 23,

    /**
     * FieldPosition and UFieldPosition selector for 'v' field alignment,
     * corresponding to the UCAL_ZONE_OFFSET field.
     * @stable ICU 3.4
     */
    UDAT_TIMEZONE_GENERIC_FIELD = 24,
    /**
     * FieldPosition selector for 'c' field alignment,
     * corresponding to the {@link #UCAL_DOW_LOCAL} field.
     * This displays the stand alone day name, if available.
     * @stable ICU 3.4
     */
    UDAT_STANDALONE_DAY_FIELD = 25,

    /**
     * FieldPosition selector for 'L' field alignment,
     * corresponding to the {@link #UCAL_MONTH} field.
     * This displays the stand alone month name, if available.
     * @stable ICU 3.4
     */
    UDAT_STANDALONE_MONTH_FIELD = 26,

    /**
     * FieldPosition selector for "Q" field alignment,
     * corresponding to quarters. This is implemented
     * using the {@link #UCAL_MONTH} field. This
     * displays the quarter.
     * @stable ICU 3.6
     */
    UDAT_QUARTER_FIELD = 27,

    /**
     * FieldPosition selector for the "q" field alignment,
     * corresponding to stand-alone quarters. This is
     * implemented using the {@link #UCAL_MONTH} field.
     * This displays the stand-alone quarter.
     * @stable ICU 3.6
     */
    UDAT_STANDALONE_QUARTER_FIELD = 28,

    /**
     * FieldPosition and UFieldPosition selector for 'V' field alignment,
     * corresponding to the UCAL_ZONE_OFFSET field.
     * @stable ICU 3.8
     */
    UDAT_TIMEZONE_SPECIAL_FIELD = 29,

    /**
     * FieldPosition selector for "U" field alignment,
     * corresponding to cyclic year names. This is implemented
     * using the {@link #UCAL_YEAR} field. This displays
     * the cyclic year name, if available.
     * @stable ICU 49
     */
    UDAT_YEAR_NAME_FIELD = 30,

    /**
     * FieldPosition selector for 'O' field alignment,
     * corresponding to the UCAL_ZONE_OFFSET and UCAL_DST_OFFSETfields.
     * This displays the localized GMT format.
     * @stable ICU 51
     */
    UDAT_TIMEZONE_LOCALIZED_GMT_OFFSET_FIELD = 31,

    /**
     * FieldPosition selector for 'X' field alignment,
     * corresponding to the UCAL_ZONE_OFFSET and UCAL_DST_OFFSETfields.
     * This displays the ISO 8601 local time offset format or UTC indicator ("Z").
     * @stable ICU 51
     */
    UDAT_TIMEZONE_ISO_FIELD = 32,

    /**
     * FieldPosition selector for 'x' field alignment,
     * corresponding to the UCAL_ZONE_OFFSET and UCAL_DST_OFFSET fields.
     * This displays the ISO 8601 local time offset format.
     * @stable ICU 51
     */
    UDAT_TIMEZONE_ISO_LOCAL_FIELD = 33,

#ifndef U_HIDE_INTERNAL_API
    /**
     * FieldPosition and UFieldPosition selector for 'r' field alignment,
     * no directly corresponding UCAL_ field.
     * @internal ICU 53
     */
    UDAT_RELATED_YEAR_FIELD = 34,
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * FieldPosition selector for 'b' field alignment.
     * Displays midnight and noon for 12am and 12pm, respectively, if available;
     * otherwise fall back to AM / PM.
     * @stable ICU 57
     */
    UDAT_AM_PM_MIDNIGHT_NOON_FIELD = 35,

    /* FieldPosition selector for 'B' field alignment.
     * Displays flexible day periods, such as "in the morning", if available.
     * @stable ICU 57
     */
    UDAT_FLEXIBLE_DAY_PERIOD_FIELD = 36,

#ifndef U_HIDE_INTERNAL_API
    /**
     * FieldPosition and UFieldPosition selector for time separator,
     * no corresponding UCAL_ field. No pattern character is currently
     * defined for this.
     * @internal
     */
    UDAT_TIME_SEPARATOR_FIELD = 37,
#endif  /* U_HIDE_INTERNAL_API */

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Number of FieldPosition and UFieldPosition selectors for
     * DateFormat and UDateFormat.
     * Valid selectors range from 0 to UDAT_FIELD_COUNT-1.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UDAT_FIELD_COUNT = 38
#endif  /* U_HIDE_DEPRECATED_API */
} UDateFormatField;


#ifndef U_HIDE_INTERNAL_API
/**
 * Is a pattern character defined for UDAT_TIME_SEPARATOR_FIELD?
 * In ICU 55 it was COLON, but that was withdrawn in ICU 56.
 * @internal ICU 56
 */
#define UDAT_HAS_PATTERN_CHAR_FOR_TIME_SEPARATOR 0
#endif /* U_HIDE_INTERNAL_API */


/**
 * Maps from a UDateFormatField to the corresponding UCalendarDateFields.
 * Note: since the mapping is many-to-one, there is no inverse mapping.
 * @param field the UDateFormatField.
 * @return the UCalendarDateField.  This will be UCAL_FIELD_COUNT in case
 * of error (e.g., the input field is UDAT_FIELD_COUNT).
 * @stable ICU 4.4
 */
U_CAPI UCalendarDateFields U_EXPORT2
udat_toCalendarDateField(UDateFormatField field);


/**
 * Open a new UDateFormat for formatting and parsing dates and times.
 * A UDateFormat may be used to format dates in calls to {@link #udat_format },
 * and to parse dates in calls to {@link #udat_parse }.
 * @param timeStyle The style used to format times; one of UDAT_FULL, UDAT_LONG,
 * UDAT_MEDIUM, UDAT_SHORT, UDAT_DEFAULT, or UDAT_NONE (relative time styles
 * are not currently supported).
 * When the pattern parameter is used, pass in UDAT_PATTERN for both timeStyle and dateStyle.
 * @param dateStyle The style used to format dates; one of UDAT_FULL, UDAT_LONG,
 * UDAT_MEDIUM, UDAT_SHORT, UDAT_DEFAULT, UDAT_FULL_RELATIVE, UDAT_LONG_RELATIVE,
 * UDAT_MEDIUM_RELATIVE, UDAT_SHORT_RELATIVE, or UDAT_NONE.
 * When the pattern parameter is used, pass in UDAT_PATTERN for both timeStyle and dateStyle.
 * As currently implemented,
 * relative date formatting only affects a limited range of calendar days before or
 * after the current date, based on the CLDR &lt;field type="day"&gt;/&lt;relative&gt; data: For
 * example, in English, "Yesterday", "Today", and "Tomorrow". Outside of this range,
 * dates are formatted using the corresponding non-relative style.
 * @param locale The locale specifying the formatting conventions
 * @param tzID A timezone ID specifying the timezone to use.  If 0, use
 * the default timezone.
 * @param tzIDLength The length of tzID, or -1 if null-terminated.
 * @param pattern A pattern specifying the format to use.
 * @param patternLength The number of characters in the pattern, or -1 if null-terminated.
 * @param status A pointer to an UErrorCode to receive any errors
 * @return A pointer to a UDateFormat to use for formatting dates and times, or 0 if
 * an error occurred.
 * @stable ICU 2.0
 */
U_CAPI UDateFormat* U_EXPORT2
udat_open(UDateFormatStyle  timeStyle,
          UDateFormatStyle  dateStyle,
          const char        *locale,
          const UChar       *tzID,
          int32_t           tzIDLength,
          const UChar       *pattern,
          int32_t           patternLength,
          UErrorCode        *status);


/**
* Close a UDateFormat.
* Once closed, a UDateFormat may no longer be used.
* @param format The formatter to close.
* @stable ICU 2.0
*/
U_CAPI void U_EXPORT2
udat_close(UDateFormat* format);


/**
 * DateFormat boolean attributes
 *
 * @stable ICU 53
 */
typedef enum UDateFormatBooleanAttribute {
   /**
     * indicates whether whitespace is allowed. Includes trailing dot tolerance.
     * @stable ICU 53
     */
    UDAT_PARSE_ALLOW_WHITESPACE = 0,
    /**
     * indicates tolerance of numeric data when String data may be assumed. eg: UDAT_YEAR_NAME_FIELD,
     * UDAT_STANDALONE_MONTH_FIELD, UDAT_DAY_OF_WEEK_FIELD
     * @stable ICU 53
     */
    UDAT_PARSE_ALLOW_NUMERIC = 1,
    /**
     * indicates tolerance of a partial literal match
     * e.g. accepting "--mon-02-march-2011" for a pattern of "'--: 'EEE-WW-MMMM-yyyy"
     * @stable ICU 56
     */
    UDAT_PARSE_PARTIAL_LITERAL_MATCH = 2,
    /**
     * indicates tolerance of pattern mismatch between input data and specified format pattern.
     * e.g. accepting "September" for a month pattern of MMM ("Sep")
     * @stable ICU 56
     */
    UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH = 3,

    /* Do not conditionalize the following with #ifndef U_HIDE_DEPRECATED_API,
     * it is needed for layout of DateFormat object. */
#ifndef U_FORCE_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UDateFormatBooleanAttribute value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UDAT_BOOLEAN_ATTRIBUTE_COUNT = 4
#endif  // U_FORCE_HIDE_DEPRECATED_API
} UDateFormatBooleanAttribute;

/**
 * Get a boolean attribute associated with a UDateFormat.
 * An example would be a true value for a key of UDAT_PARSE_ALLOW_WHITESPACE indicating allowing whitespace leniency.
 * If the formatter does not understand the attribute, -1 is returned.
 * @param fmt The formatter to query.
 * @param attr The attribute to query; e.g. UDAT_PARSE_ALLOW_WHITESPACE.
 * @param status A pointer to an UErrorCode to receive any errors
 * @return The value of attr.
 * @stable ICU 53
 */
U_CAPI UBool U_EXPORT2
udat_getBooleanAttribute(const UDateFormat* fmt, UDateFormatBooleanAttribute attr, UErrorCode* status);

/**
 * Set a boolean attribute associated with a UDateFormat.
 * An example of a boolean attribute is parse leniency control.  If the formatter does not understand
 * the attribute, the call is ignored.
 * @param fmt The formatter to set.
 * @param attr The attribute to set; one of UDAT_PARSE_ALLOW_WHITESPACE or UDAT_PARSE_ALLOW_NUMERIC
 * @param newValue The new value of attr.
 * @param status A pointer to an UErrorCode to receive any errors
 * @stable ICU 53
 */
U_CAPI void U_EXPORT2
udat_setBooleanAttribute(UDateFormat *fmt, UDateFormatBooleanAttribute attr, UBool newValue, UErrorCode* status);



#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUDateFormatPointer
 * "Smart pointer" class, closes a UDateFormat via udat_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUDateFormatPointer, UDateFormat, udat_close);

U_NAMESPACE_END

#endif

/**
 * Open a copy of a UDateFormat.
 * This function performs a deep copy.
 * @param fmt The format to copy
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return A pointer to a UDateFormat identical to fmt.
 * @stable ICU 2.0
 */
U_CAPI UDateFormat* U_EXPORT2
udat_clone(const UDateFormat *fmt,
       UErrorCode *status);

/**
* Format a date using a UDateFormat.
* The date will be formatted using the conventions specified in {@link #udat_open }
* @param format The formatter to use
* @param dateToFormat The date to format
* @param result A pointer to a buffer to receive the formatted number.
* @param resultLength The maximum size of result.
* @param position A pointer to a UFieldPosition.  On input, position->field
* is read.  On output, position->beginIndex and position->endIndex indicate
* the beginning and ending indices of field number position->field, if such
* a field exists.  This parameter may be NULL, in which case no field
* position data is returned.
* @param status A pointer to an UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see udat_parse
* @see UFieldPosition
* @stable ICU 2.0
*/
U_CAPI int32_t U_EXPORT2
udat_format(    const    UDateFormat*    format,
                        UDate           dateToFormat,
                        UChar*          result,
                        int32_t         resultLength,
                        UFieldPosition* position,
                        UErrorCode*     status);

/**
* Format a date using an UDateFormat.
* The date will be formatted using the conventions specified in {@link #udat_open }
* @param format The formatter to use
* @param calendar The calendar to format. The calendar instance might be
*                 mutated if fields are not yet fully calculated, though
*                 the function won't change the logical date and time held
*                 by the instance.
* @param result A pointer to a buffer to receive the formatted number.
* @param capacity The maximum size of result.
* @param position A pointer to a UFieldPosition.  On input, position->field
* is read.  On output, position->beginIndex and position->endIndex indicate
* the beginning and ending indices of field number position->field, if such
* a field exists.  This parameter may be NULL, in which case no field
* position data is returned.
* @param status A pointer to an UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see udat_format
* @see udat_parseCalendar
* @see UFieldPosition
* @stable ICU 55
*/
U_CAPI int32_t U_EXPORT2
udat_formatCalendar(    const UDateFormat*  format,
                        UCalendar*      calendar,
                        UChar*          result,
                        int32_t         capacity,
                        UFieldPosition* position,
                        UErrorCode*     status);

/**
* Format a date using a UDateFormat.
* The date will be formatted using the conventions specified in {@link #udat_open}
* @param format
*          The formatter to use
* @param dateToFormat
*          The date to format
* @param result
*          A pointer to a buffer to receive the formatted number.
* @param resultLength
*          The maximum size of result.
* @param fpositer
*          A pointer to a UFieldPositionIterator created by {@link #ufieldpositer_open}
*          (may be NULL if field position information is not needed). Any
*          iteration information already present in the UFieldPositionIterator
*          will be deleted, and the iterator will be reset to apply to the
*          fields in the formatted string created by this function call; the
*          field values provided by {@link #ufieldpositer_next} will be from the
*          UDateFormatField enum.
* @param status
*          A pointer to a UErrorCode to receive any errors
* @return
*          The total buffer size needed; if greater than resultLength, the output was truncated.
* @see udat_parse
* @see UFieldPositionIterator
* @stable ICU 55
*/
U_CAPI int32_t U_EXPORT2
udat_formatForFields(   const UDateFormat* format,
                        UDate           dateToFormat,
                        UChar*          result,
                        int32_t         resultLength,
                        UFieldPositionIterator* fpositer,
                        UErrorCode*     status);

/**
* Format a date using a UDateFormat.
* The date will be formatted using the conventions specified in {@link #udat_open }
* @param format
*          The formatter to use
* @param calendar
*          The calendar to format. The calendar instance might be mutated if fields
*          are not yet fully calculated, though the function won't change the logical
*          date and time held by the instance.
* @param result
*          A pointer to a buffer to receive the formatted number.
* @param capacity
*          The maximum size of result.
* @param fpositer
*          A pointer to a UFieldPositionIterator created by {@link #ufieldpositer_open}
*          (may be NULL if field position information is not needed). Any
*          iteration information already present in the UFieldPositionIterator
*          will be deleted, and the iterator will be reset to apply to the
*          fields in the formatted string created by this function call; the
*          field values provided by {@link #ufieldpositer_next} will be from the
*          UDateFormatField enum.
* @param status
*          A pointer to a UErrorCode to receive any errors
* @return
*          The total buffer size needed; if greater than resultLength, the output was truncated.
* @see udat_format
* @see udat_parseCalendar
* @see UFieldPositionIterator
* @stable ICU 55
*/
U_CAPI int32_t U_EXPORT2
udat_formatCalendarForFields( const UDateFormat* format,
                        UCalendar*      calendar,
                        UChar*          result,
                        int32_t         capacity,
                        UFieldPositionIterator* fpositer,
                        UErrorCode*     status);


/**
* Parse a string into an date/time using a UDateFormat.
* The date will be parsed using the conventions specified in {@link #udat_open }.
* <P>
* Note that the normal date formats associated with some calendars - such
* as the Chinese lunar calendar - do not specify enough fields to enable
* dates to be parsed unambiguously. In the case of the Chinese lunar
* calendar, while the year within the current 60-year cycle is specified,
* the number of such cycles since the start date of the calendar (in the
* UCAL_ERA field of the UCalendar object) is not normally part of the format,
* and parsing may assume the wrong era. For cases such as this it is
* recommended that clients parse using udat_parseCalendar with the UCalendar
* passed in set to the current date, or to a date within the era/cycle that
* should be assumed if absent in the format.
*
* @param format The formatter to use.
* @param text The text to parse.
* @param textLength The length of text, or -1 if null-terminated.
* @param parsePos If not 0, on input a pointer to an integer specifying the offset at which
* to begin parsing.  If not 0, on output the offset at which parsing ended.
* @param status A pointer to an UErrorCode to receive any errors
* @return The value of the parsed date/time
* @see udat_format
* @stable ICU 2.0
*/
U_CAPI UDate U_EXPORT2
udat_parse(const    UDateFormat*    format,
           const    UChar*          text,
                    int32_t         textLength,
                    int32_t         *parsePos,
                    UErrorCode      *status);

/**
* Parse a string into an date/time using a UDateFormat.
* The date will be parsed using the conventions specified in {@link #udat_open }.
* @param format The formatter to use.
* @param calendar A calendar set on input to the date and time to be used for
*                 missing values in the date/time string being parsed, and set
*                 on output to the parsed date/time. When the calendar type is
*                 different from the internal calendar held by the UDateFormat
*                 instance, the internal calendar will be cloned to a work
*                 calendar set to the same milliseconds and time zone as this
*                 calendar parameter, field values will be parsed based on the
*                 work calendar, then the result (milliseconds and time zone)
*                 will be set in this calendar.
* @param text The text to parse.
* @param textLength The length of text, or -1 if null-terminated.
* @param parsePos If not 0, on input a pointer to an integer specifying the offset at which
* to begin parsing.  If not 0, on output the offset at which parsing ended.
* @param status A pointer to an UErrorCode to receive any errors
* @see udat_format
* @stable ICU 2.0
*/
U_CAPI void U_EXPORT2
udat_parseCalendar(const    UDateFormat*    format,
                            UCalendar*      calendar,
                   const    UChar*          text,
                            int32_t         textLength,
                            int32_t         *parsePos,
                            UErrorCode      *status);

/**
* Determine if an UDateFormat will perform lenient parsing.
* With lenient parsing, the parser may use heuristics to interpret inputs that do not
* precisely match the pattern. With strict parsing, inputs must match the pattern.
* @param fmt The formatter to query
* @return TRUE if fmt is set to perform lenient parsing, FALSE otherwise.
* @see udat_setLenient
* @stable ICU 2.0
*/
U_CAPI UBool U_EXPORT2
udat_isLenient(const UDateFormat* fmt);

/**
* Specify whether an UDateFormat will perform lenient parsing.
* With lenient parsing, the parser may use heuristics to interpret inputs that do not
* precisely match the pattern. With strict parsing, inputs must match the pattern.
* @param fmt The formatter to set
* @param isLenient TRUE if fmt should perform lenient parsing, FALSE otherwise.
* @see dat_isLenient
* @stable ICU 2.0
*/
U_CAPI void U_EXPORT2
udat_setLenient(    UDateFormat*    fmt,
                    UBool          isLenient);

/**
* Get the UCalendar associated with an UDateFormat.
* A UDateFormat uses a UCalendar to convert a raw value to, for example,
* the day of the week.
* @param fmt The formatter to query.
* @return A pointer to the UCalendar used by fmt.
* @see udat_setCalendar
* @stable ICU 2.0
*/
U_CAPI const UCalendar* U_EXPORT2
udat_getCalendar(const UDateFormat* fmt);

/**
* Set the UCalendar associated with an UDateFormat.
* A UDateFormat uses a UCalendar to convert a raw value to, for example,
* the day of the week.
* @param fmt The formatter to set.
* @param calendarToSet A pointer to an UCalendar to be used by fmt.
* @see udat_setCalendar
* @stable ICU 2.0
*/
U_CAPI void U_EXPORT2
udat_setCalendar(            UDateFormat*    fmt,
                    const   UCalendar*      calendarToSet);

/**
* Get the UNumberFormat associated with an UDateFormat.
* A UDateFormat uses a UNumberFormat to format numbers within a date,
* for example the day number.
* @param fmt The formatter to query.
* @return A pointer to the UNumberFormat used by fmt to format numbers.
* @see udat_setNumberFormat
* @stable ICU 2.0
*/
U_CAPI const UNumberFormat* U_EXPORT2
udat_getNumberFormat(const UDateFormat* fmt);

/**
* Get the UNumberFormat for specific field associated with an UDateFormat.
* For example: 'y' for year and 'M' for month
* @param fmt The formatter to query.
* @param field the field to query
* @return A pointer to the UNumberFormat used by fmt to format field numbers.
* @see udat_setNumberFormatForField
* @stable ICU 54
*/
U_CAPI const UNumberFormat* U_EXPORT2
udat_getNumberFormatForField(const UDateFormat* fmt, UChar field);

/**
* Set the UNumberFormat for specific field associated with an UDateFormat.
* It can be a single field like: "y"(year) or "M"(month)
* It can be several field combined together: "yM"(year and month)
* Note:
* 1 symbol field is enough for multiple symbol field (so "y" will override "yy", "yyy")
* If the field is not numeric, then override has no effect (like "MMM" will use abbreviation, not numerical field)
*
* @param fields the fields to set
* @param fmt The formatter to set.
* @param numberFormatToSet A pointer to the UNumberFormat to be used by fmt to format numbers.
* @param status error code passed around (memory allocation or invalid fields)
* @see udat_getNumberFormatForField
* @stable ICU 54
*/
U_CAPI void U_EXPORT2
udat_adoptNumberFormatForFields(  UDateFormat* fmt,
                            const UChar* fields,
                                  UNumberFormat*  numberFormatToSet,
                                  UErrorCode* status);
/**
* Set the UNumberFormat associated with an UDateFormat.
* A UDateFormat uses a UNumberFormat to format numbers within a date,
* for example the day number.
* This method also clears per field NumberFormat instances previously
* set by {@see udat_setNumberFormatForField}
* @param fmt The formatter to set.
* @param numberFormatToSet A pointer to the UNumberFormat to be used by fmt to format numbers.
* @see udat_getNumberFormat
* @see udat_setNumberFormatForField
* @stable ICU 2.0
*/
U_CAPI void U_EXPORT2
udat_setNumberFormat(            UDateFormat*    fmt,
                        const   UNumberFormat*  numberFormatToSet);
/**
* Adopt the UNumberFormat associated with an UDateFormat.
* A UDateFormat uses a UNumberFormat to format numbers within a date,
* for example the day number.
* @param fmt The formatter to set.
* @param numberFormatToAdopt A pointer to the UNumberFormat to be used by fmt to format numbers.
* @see udat_getNumberFormat
* @stable ICU 54
*/
U_CAPI void U_EXPORT2
udat_adoptNumberFormat(            UDateFormat*    fmt,
                                   UNumberFormat*  numberFormatToAdopt);
/**
* Get a locale for which date/time formatting patterns are available.
* A UDateFormat in a locale returned by this function will perform the correct
* formatting and parsing for the locale.
* @param localeIndex The index of the desired locale.
* @return A locale for which date/time formatting patterns are available, or 0 if none.
* @see udat_countAvailable
* @stable ICU 2.0
*/
U_CAPI const char* U_EXPORT2
udat_getAvailable(int32_t localeIndex);

/**
* Determine how many locales have date/time  formatting patterns available.
* This function is most useful as determining the loop ending condition for
* calls to {@link #udat_getAvailable }.
* @return The number of locales for which date/time formatting patterns are available.
* @see udat_getAvailable
* @stable ICU 2.0
*/
U_CAPI int32_t U_EXPORT2
udat_countAvailable(void);

/**
* Get the year relative to which all 2-digit years are interpreted.
* For example, if the 2-digit start year is 2100, the year 99 will be
* interpreted as 2199.
* @param fmt The formatter to query.
* @param status A pointer to an UErrorCode to receive any errors
* @return The year relative to which all 2-digit years are interpreted.
* @see udat_Set2DigitYearStart
* @stable ICU 2.0
*/
U_CAPI UDate U_EXPORT2
udat_get2DigitYearStart(    const   UDateFormat     *fmt,
                                    UErrorCode      *status);

/**
* Set the year relative to which all 2-digit years will be interpreted.
* For example, if the 2-digit start year is 2100, the year 99 will be
* interpreted as 2199.
* @param fmt The formatter to set.
* @param d The year relative to which all 2-digit years will be interpreted.
* @param status A pointer to an UErrorCode to receive any errors
* @see udat_Set2DigitYearStart
* @stable ICU 2.0
*/
U_CAPI void U_EXPORT2
udat_set2DigitYearStart(    UDateFormat     *fmt,
                            UDate           d,
                            UErrorCode      *status);

/**
* Extract the pattern from a UDateFormat.
* The pattern will follow the pattern syntax rules.
* @param fmt The formatter to query.
* @param localized TRUE if the pattern should be localized, FALSE otherwise.
* @param result A pointer to a buffer to receive the pattern.
* @param resultLength The maximum size of result.
* @param status A pointer to an UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see udat_applyPattern
* @stable ICU 2.0
*/
U_CAPI int32_t U_EXPORT2
udat_toPattern(    const   UDateFormat     *fmt,
                        UBool          localized,
                        UChar           *result,
                        int32_t         resultLength,
                        UErrorCode      *status);

/**
* Set the pattern used by an UDateFormat.
* The pattern should follow the pattern syntax rules.
* @param format The formatter to set.
* @param localized TRUE if the pattern is localized, FALSE otherwise.
* @param pattern The new pattern
* @param patternLength The length of pattern, or -1 if null-terminated.
* @see udat_toPattern
* @stable ICU 2.0
*/
U_CAPI void U_EXPORT2
udat_applyPattern(            UDateFormat     *format,
                            UBool          localized,
                    const   UChar           *pattern,
                            int32_t         patternLength);

/**
 * The possible types of date format symbols
 * @stable ICU 2.6
 */
typedef enum UDateFormatSymbolType {
    /** The era names, for example AD */
    UDAT_ERAS,
    /** The month names, for example February */
    UDAT_MONTHS,
    /** The short month names, for example Feb. */
    UDAT_SHORT_MONTHS,
    /** The CLDR-style format "wide" weekday names, for example Monday */
    UDAT_WEEKDAYS,
    /**
     * The CLDR-style format "abbreviated" (not "short") weekday names, for example "Mon."
     * For the CLDR-style format "short" weekday names, use UDAT_SHORTER_WEEKDAYS.
     */
    UDAT_SHORT_WEEKDAYS,
    /** The AM/PM names, for example AM */
    UDAT_AM_PMS,
    /** The localized characters */
    UDAT_LOCALIZED_CHARS,
    /** The long era names, for example Anno Domini */
    UDAT_ERA_NAMES,
    /** The narrow month names, for example F */
    UDAT_NARROW_MONTHS,
    /** The CLDR-style format "narrow" weekday names, for example "M" */
    UDAT_NARROW_WEEKDAYS,
    /** Standalone context versions of months */
    UDAT_STANDALONE_MONTHS,
    UDAT_STANDALONE_SHORT_MONTHS,
    UDAT_STANDALONE_NARROW_MONTHS,
    /** The CLDR-style stand-alone "wide" weekday names */
    UDAT_STANDALONE_WEEKDAYS,
    /**
     * The CLDR-style stand-alone "abbreviated" (not "short") weekday names.
     * For the CLDR-style stand-alone "short" weekday names, use UDAT_STANDALONE_SHORTER_WEEKDAYS.
     */
    UDAT_STANDALONE_SHORT_WEEKDAYS,
    /** The CLDR-style stand-alone "narrow" weekday names */
    UDAT_STANDALONE_NARROW_WEEKDAYS,
    /** The quarters, for example 1st Quarter */
    UDAT_QUARTERS,
    /** The short quarter names, for example Q1 */
    UDAT_SHORT_QUARTERS,
    /** Standalone context versions of quarters */
    UDAT_STANDALONE_QUARTERS,
    UDAT_STANDALONE_SHORT_QUARTERS,
    /**
     * The CLDR-style short weekday names, e.g. "Su", Mo", etc.
     * These are named "SHORTER" to contrast with the constants using _SHORT_
     * above, which actually get the CLDR-style *abbreviated* versions of the
     * corresponding names.
     * @stable ICU 51
     */
    UDAT_SHORTER_WEEKDAYS,
    /**
     * Standalone version of UDAT_SHORTER_WEEKDAYS.
     * @stable ICU 51
     */
    UDAT_STANDALONE_SHORTER_WEEKDAYS,
    /**
     * Cyclic year names (only supported for some calendars, and only for FORMAT usage;
     * udat_setSymbols not supported for UDAT_CYCLIC_YEARS_WIDE)
     * @stable ICU 54
     */
    UDAT_CYCLIC_YEARS_WIDE,
    /**
     * Cyclic year names (only supported for some calendars, and only for FORMAT usage)
     * @stable ICU 54
     */
    UDAT_CYCLIC_YEARS_ABBREVIATED,
    /**
     * Cyclic year names (only supported for some calendars, and only for FORMAT usage;
     * udat_setSymbols not supported for UDAT_CYCLIC_YEARS_NARROW)
     * @stable ICU 54
     */
    UDAT_CYCLIC_YEARS_NARROW,
    /**
     * Calendar zodiac  names (only supported for some calendars, and only for FORMAT usage;
     * udat_setSymbols not supported for UDAT_ZODIAC_NAMES_WIDE)
     * @stable ICU 54
     */
    UDAT_ZODIAC_NAMES_WIDE,
    /**
     * Calendar zodiac  names (only supported for some calendars, and only for FORMAT usage)
     * @stable ICU 54
     */
    UDAT_ZODIAC_NAMES_ABBREVIATED,
    /**
     * Calendar zodiac  names (only supported for some calendars, and only for FORMAT usage;
     * udat_setSymbols not supported for UDAT_ZODIAC_NAMES_NARROW)
     * @stable ICU 54
     */
    UDAT_ZODIAC_NAMES_NARROW
} UDateFormatSymbolType;

struct UDateFormatSymbols;
/** Date format symbols.
 *  For usage in C programs.
 *  @stable ICU 2.6
 */
typedef struct UDateFormatSymbols UDateFormatSymbols;

/**
* Get the symbols associated with an UDateFormat.
* The symbols are what a UDateFormat uses to represent locale-specific data,
* for example month or day names.
* @param fmt The formatter to query.
* @param type The type of symbols to get.  One of UDAT_ERAS, UDAT_MONTHS, UDAT_SHORT_MONTHS,
* UDAT_WEEKDAYS, UDAT_SHORT_WEEKDAYS, UDAT_AM_PMS, or UDAT_LOCALIZED_CHARS
* @param symbolIndex The desired symbol of type type.
* @param result A pointer to a buffer to receive the pattern.
* @param resultLength The maximum size of result.
* @param status A pointer to an UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see udat_countSymbols
* @see udat_setSymbols
* @stable ICU 2.0
*/
U_CAPI int32_t U_EXPORT2
udat_getSymbols(const   UDateFormat             *fmt,
                        UDateFormatSymbolType   type,
                        int32_t                 symbolIndex,
                        UChar                   *result,
                        int32_t                 resultLength,
                        UErrorCode              *status);

/**
* Count the number of particular symbols for an UDateFormat.
* This function is most useful as for detemining the loop termination condition
* for calls to {@link #udat_getSymbols }.
* @param fmt The formatter to query.
* @param type The type of symbols to count.  One of UDAT_ERAS, UDAT_MONTHS, UDAT_SHORT_MONTHS,
* UDAT_WEEKDAYS, UDAT_SHORT_WEEKDAYS, UDAT_AM_PMS, or UDAT_LOCALIZED_CHARS
* @return The number of symbols of type type.
* @see udat_getSymbols
* @see udat_setSymbols
* @stable ICU 2.0
*/
U_CAPI int32_t U_EXPORT2
udat_countSymbols(    const    UDateFormat                *fmt,
                            UDateFormatSymbolType    type);

/**
* Set the symbols associated with an UDateFormat.
* The symbols are what a UDateFormat uses to represent locale-specific data,
* for example month or day names.
* @param format The formatter to set
* @param type The type of symbols to set.  One of UDAT_ERAS, UDAT_MONTHS, UDAT_SHORT_MONTHS,
* UDAT_WEEKDAYS, UDAT_SHORT_WEEKDAYS, UDAT_AM_PMS, or UDAT_LOCALIZED_CHARS
* @param symbolIndex The index of the symbol to set of type type.
* @param value The new value
* @param valueLength The length of value, or -1 if null-terminated
* @param status A pointer to an UErrorCode to receive any errors
* @see udat_getSymbols
* @see udat_countSymbols
* @stable ICU 2.0
*/
U_CAPI void U_EXPORT2
udat_setSymbols(    UDateFormat             *format,
                    UDateFormatSymbolType   type,
                    int32_t                 symbolIndex,
                    UChar                   *value,
                    int32_t                 valueLength,
                    UErrorCode              *status);

/**
 * Get the locale for this date format object.
 * You can choose between valid and actual locale.
 * @param fmt The formatter to get the locale from
 * @param type type of the locale we're looking for (valid or actual)
 * @param status error code for the operation
 * @return the locale name
 * @stable ICU 2.8
 */
U_CAPI const char* U_EXPORT2
udat_getLocaleByType(const UDateFormat *fmt,
                     ULocDataLocaleType type,
                     UErrorCode* status);

/**
 * Set a particular UDisplayContext value in the formatter, such as
 * UDISPCTX_CAPITALIZATION_FOR_STANDALONE.
 * @param fmt The formatter for which to set a UDisplayContext value.
 * @param value The UDisplayContext value to set.
 * @param status A pointer to an UErrorCode to receive any errors
 * @stable ICU 51
 */
U_CAPI void U_EXPORT2
udat_setContext(UDateFormat* fmt, UDisplayContext value, UErrorCode* status);

/**
 * Get the formatter's UDisplayContext value for the specified UDisplayContextType,
 * such as UDISPCTX_TYPE_CAPITALIZATION.
 * @param fmt The formatter to query.
 * @param type The UDisplayContextType whose value to return
 * @param status A pointer to an UErrorCode to receive any errors
 * @return The UDisplayContextValue for the specified type.
 * @stable ICU 53
 */
U_CAPI UDisplayContext U_EXPORT2
udat_getContext(const UDateFormat* fmt, UDisplayContextType type, UErrorCode* status);

#ifndef U_HIDE_INTERNAL_API
/**
* Extract the date pattern from a UDateFormat set for relative date formatting.
* The pattern will follow the pattern syntax rules.
* @param fmt The formatter to query.
* @param result A pointer to a buffer to receive the pattern.
* @param resultLength The maximum size of result.
* @param status A pointer to a UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see udat_applyPatternRelative
* @internal ICU 4.2 technology preview
*/
U_INTERNAL int32_t U_EXPORT2
udat_toPatternRelativeDate(const UDateFormat *fmt,
                           UChar             *result,
                           int32_t           resultLength,
                           UErrorCode        *status);

/**
* Extract the time pattern from a UDateFormat set for relative date formatting.
* The pattern will follow the pattern syntax rules.
* @param fmt The formatter to query.
* @param result A pointer to a buffer to receive the pattern.
* @param resultLength The maximum size of result.
* @param status A pointer to a UErrorCode to receive any errors
* @return The total buffer size needed; if greater than resultLength, the output was truncated.
* @see udat_applyPatternRelative
* @internal ICU 4.2 technology preview
*/
U_INTERNAL int32_t U_EXPORT2
udat_toPatternRelativeTime(const UDateFormat *fmt,
                           UChar             *result,
                           int32_t           resultLength,
                           UErrorCode        *status);

/**
* Set the date & time patterns used by a UDateFormat set for relative date formatting.
* The patterns should follow the pattern syntax rules.
* @param format The formatter to set.
* @param datePattern The new date pattern
* @param datePatternLength The length of datePattern, or -1 if null-terminated.
* @param timePattern The new time pattern
* @param timePatternLength The length of timePattern, or -1 if null-terminated.
* @param status A pointer to a UErrorCode to receive any errors
* @see udat_toPatternRelativeDate, udat_toPatternRelativeTime
* @internal ICU 4.2 technology preview
*/
U_INTERNAL void U_EXPORT2
udat_applyPatternRelative(UDateFormat *format,
                          const UChar *datePattern,
                          int32_t     datePatternLength,
                          const UChar *timePattern,
                          int32_t     timePatternLength,
                          UErrorCode  *status);

/**
 * @internal
 * @see udat_open
 */
typedef UDateFormat* (U_EXPORT2 *UDateFormatOpener) (UDateFormatStyle  timeStyle,
                                                    UDateFormatStyle  dateStyle,
                                                    const char        *locale,
                                                    const UChar       *tzID,
                                                    int32_t           tzIDLength,
                                                    const UChar       *pattern,
                                                    int32_t           patternLength,
                                                    UErrorCode        *status);

/**
 * Register a provider factory
 * @internal ICU 49
 */
U_INTERNAL void U_EXPORT2
udat_registerOpener(UDateFormatOpener opener, UErrorCode *status);

/**
 * Un-Register a provider factory
 * @internal ICU 49
 */
U_INTERNAL UDateFormatOpener U_EXPORT2
udat_unregisterOpener(UDateFormatOpener opener, UErrorCode *status);
#endif  /* U_HIDE_INTERNAL_API */


#endif /* #if !UCONFIG_NO_FORMATTING */



/**
 * \file 
 * \brief C++ API: Compatibility APIs for measure formatting.
 */

/**
 * Constants for various widths.
 * There are 4 widths: Wide, Short, Narrow, Numeric.
 * For example, for English, when formatting "3 hours"
 * Wide is "3 hours"; short is "3 hrs"; narrow is "3h";
 * formatting "3 hours 17 minutes" as numeric give "3:17"
 * @stable ICU 53
 */
enum UMeasureFormatWidth {

    // Wide, short, and narrow must be first and in this order.
    /**
     * Spell out measure units.
     * @stable ICU 53 
     */
    UMEASFMT_WIDTH_WIDE,
 
    /**
     * Abbreviate measure units.
     * @stable ICU 53
     */
    UMEASFMT_WIDTH_SHORT,

    /**
     * Use symbols for measure units when possible.
     * @stable ICU 53
     */
    UMEASFMT_WIDTH_NARROW,

    /**
     * Completely omit measure units when possible. For example, format
     * '5 hours, 37 minutes' as '5:37'
     * @stable ICU 53
     */
    UMEASFMT_WIDTH_NUMERIC,

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UMeasureFormatWidth value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UMEASFMT_WIDTH_COUNT = 4
#endif  // U_HIDE_DEPRECATED_API
};
/** @stable ICU 53 */
typedef enum UMeasureFormatWidth UMeasureFormatWidth; 

U_NAMESPACE_BEGIN

class Measure;
class MeasureUnit;
class NumberFormat;
class PluralRules;
class MeasureFormatCacheData;
class SharedNumberFormat;
class SharedPluralRules;
class QuantityFormatter;
class SimpleFormatter;
class ListFormatter;
class DateFormat;

/**
 * <p><strong>IMPORTANT:</strong> New users are strongly encouraged to see if
 * numberformatter.h fits their use case.  Although not deprecated, this header
 * is provided for backwards compatibility only.
 *
 * @see Format
 * @author Alan Liu
 * @stable ICU 3.0
 */
class U_I18N_API MeasureFormat : public Format {
 public:
    using Format::parseObject;
    using Format::format;

    /**
     * Constructor.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @stable ICU 53
     */
    MeasureFormat(
            const Locale &locale, UMeasureFormatWidth width, UErrorCode &status);

    /**
     * Constructor.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @stable ICU 53
     */
    MeasureFormat(
            const Locale &locale,
            UMeasureFormatWidth width,
            NumberFormat *nfToAdopt,
            UErrorCode &status);

    /**
     * Copy constructor.
     * @stable ICU 3.0
     */
    MeasureFormat(const MeasureFormat &other);

    /**
     * Assignment operator.
     * @stable ICU 3.0
     */
    MeasureFormat &operator=(const MeasureFormat &rhs);

    /**
     * Destructor.
     * @stable ICU 3.0
     */
    virtual ~MeasureFormat();

    /**
     * Return true if given Format objects are semantically equal.
     * @stable ICU 53
     */
    virtual UBool operator==(const Format &other) const;

    /**
     * Clones this object polymorphically.
     * @stable ICU 53
     */
    virtual MeasureFormat *clone() const;

    /**
     * Formats object to produce a string.
     * @stable ICU 53
     */
    virtual UnicodeString &format(
            const Formattable &obj,
            UnicodeString &appendTo,
            FieldPosition &pos,
            UErrorCode &status) const;

#ifndef U_FORCE_HIDE_DRAFT_API
    /**
     * Parse a string to produce an object. This implementation sets
     * status to U_UNSUPPORTED_ERROR.
     *
     * @draft ICU 53
     */
    virtual void parseObject(
            const UnicodeString &source,
            Formattable &reslt,
            ParsePosition &pos) const;
#endif  // U_FORCE_HIDE_DRAFT_API

    /**
     * Formats measure objects to produce a string. An example of such a
     * formatted string is 3 meters, 3.5 centimeters. Measure objects appear
     * in the formatted string in the same order they appear in the "measures"
     * array. The NumberFormat of this object is used only to format the amount
     * of the very last measure. The other amounts are formatted with zero
     * decimal places while rounding toward zero.
     * @param measures array of measure objects.
     * @param measureCount the number of measure objects.
     * @param appendTo formatted string appended here.
     * @param pos the field position.
     * @param status the error.
     * @return appendTo reference
     *
     * @stable ICU 53
     */
    UnicodeString &formatMeasures(
            const Measure *measures,
            int32_t measureCount,
            UnicodeString &appendTo,
            FieldPosition &pos,
            UErrorCode &status) const;

    /**
     * Formats a single measure per unit. An example of such a
     * formatted string is 3.5 meters per second.
     * @param measure The measure object. In above example, 3.5 meters.
     * @param perUnit The per unit. In above example, it is
     *        `*%MeasureUnit::createSecond(status)`.
     * @param appendTo formatted string appended here.
     * @param pos the field position.
     * @param status the error.
     * @return appendTo reference
     *
     * @stable ICU 55
     */
    UnicodeString &formatMeasurePerUnit(
            const Measure &measure,
            const MeasureUnit &perUnit,
            UnicodeString &appendTo,
            FieldPosition &pos,
            UErrorCode &status) const;

    /**
     * Gets the display name of the specified {@link MeasureUnit} corresponding to the current
     * locale and format width.
     * @param unit  The unit for which to get a display name.
     * @param status the error.
     * @return  The display name in the locale and width specified in
     *          the MeasureFormat constructor, or null if there is no display name available
     *          for the specified unit.
     *
     * @stable ICU 58
     */
    UnicodeString getUnitDisplayName(const MeasureUnit& unit, UErrorCode &status) const;


    /**
     * Return a formatter for CurrencyAmount objects in the given
     * locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @param locale desired locale
     * @param ec input-output error code
     * @return a formatter object, or NULL upon error
     * @stable ICU 3.0
     */
    static MeasureFormat* U_EXPORT2 createCurrencyFormat(const Locale& locale,
                                               UErrorCode& ec);

    /**
     * Return a formatter for CurrencyAmount objects in the default
     * locale.
     * <p>
     * <strong>NOTE:</strong> New users are strongly encouraged to use
     * {@link icu::number::NumberFormatter} instead of NumberFormat.
     * @param ec input-output error code
     * @return a formatter object, or NULL upon error
     * @stable ICU 3.0
     */
    static MeasureFormat* U_EXPORT2 createCurrencyFormat(UErrorCode& ec);

    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 53
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 53
     */
    virtual UClassID getDynamicClassID(void) const;

 protected:
    /**
     * Default constructor.
     * @stable ICU 3.0
     */
    MeasureFormat();

#ifndef U_HIDE_INTERNAL_API 

    /**
     * ICU use only.
     * Initialize or change MeasureFormat class from subclass.
     * @internal.
     */
    void initMeasureFormat(
            const Locale &locale,
            UMeasureFormatWidth width,
            NumberFormat *nfToAdopt,
            UErrorCode &status);
    /**
     * ICU use only.
     * Allows subclass to change locale. Note that this method also changes
     * the NumberFormat object. Returns TRUE if locale changed; FALSE if no
     * change was made.
     * @internal.
     */
    UBool setMeasureFormatLocale(const Locale &locale, UErrorCode &status);

    /**
     * ICU use only.
     * Let subclass change NumberFormat.
     * @internal.
     */
    void adoptNumberFormat(NumberFormat *nfToAdopt, UErrorCode &status);

    /**
     * ICU use only.
     * @internal.
     */
    const NumberFormat &getNumberFormatInternal() const;

    /**
     * ICU use only.
     * Always returns the short form currency formatter.
     * @internal.
     */
    const NumberFormat& getCurrencyFormatInternal() const;

    /**
     * ICU use only.
     * @internal.
     */
    const PluralRules &getPluralRules() const;

    /**
     * ICU use only.
     * @internal.
     */
    Locale getLocale(UErrorCode &status) const;

    /**
     * ICU use only.
     * @internal.
     */
    const char *getLocaleID(UErrorCode &status) const;

#endif /* U_HIDE_INTERNAL_API */

 private:
    const MeasureFormatCacheData *cache;
    const SharedNumberFormat *numberFormat;
    const SharedPluralRules *pluralRules;
    UMeasureFormatWidth fWidth;    

    // Declared outside of MeasureFormatSharedData because ListFormatter
    // objects are relatively cheap to copy; therefore, they don't need to be
    // shared across instances.
    ListFormatter *listFormatter;

    UnicodeString &formatMeasure(
        const Measure &measure,
        const NumberFormat &nf,
        UnicodeString &appendTo,
        FieldPosition &pos,
        UErrorCode &status) const;

    UnicodeString &formatMeasuresSlowTrack(
        const Measure *measures,
        int32_t measureCount,
        UnicodeString& appendTo,
        FieldPosition& pos,
        UErrorCode& status) const;

    UnicodeString &formatNumeric(
        const Formattable *hms,  // always length 3: [0] is hour; [1] is
                                 // minute; [2] is second.
        int32_t bitMap,   // 1=hour set, 2=minute set, 4=second set
        UnicodeString &appendTo,
        UErrorCode &status) const;
};

U_NAMESPACE_END

#endif // #if !UCONFIG_NO_FORMATTING

#endif /* U_SHOW_CPLUSPLUS_API */



U_NAMESPACE_BEGIN

class NumberFormat;

/**
 * Temporary internal concrete subclass of MeasureFormat implementing
 * parsing and formatting of currency amount objects.  This class is
 * likely to be redesigned and rewritten in the near future.
 *
 * <p>This class currently delegates to DecimalFormat for parsing and
 * formatting.
 *
 * @see MeasureFormat
 * @author Alan Liu
 * @internal
 */
class CurrencyFormat : public MeasureFormat {

 public:

    /**
     * Construct a CurrencyFormat for the given locale.
     */
    CurrencyFormat(const Locale& locale, UErrorCode& ec);

    /**
     * Copy constructor.
     */
    CurrencyFormat(const CurrencyFormat& other);

    /**
     * Destructor.
     */
    virtual ~CurrencyFormat();

    /**
     * Override Format API.
     */
    virtual CurrencyFormat* clone() const;


    using MeasureFormat::format;

    /**
     * Override Format API.
     */
    virtual UnicodeString& format(const Formattable& obj,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode& ec) const;

    /**
     * Override Format API.
     */
    virtual void parseObject(const UnicodeString& source,
                             Formattable& result,
                             ParsePosition& pos) const;

    /**
     * Override Format API.
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * Returns the class ID for this class.
     */
    static UClassID U_EXPORT2 getStaticClassID();
};

U_NAMESPACE_END

#endif // #if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

CurrencyFormat::CurrencyFormat(const Locale& locale, UErrorCode& ec) :
    MeasureFormat(locale, UMEASFMT_WIDTH_WIDE, ec)
{
}

CurrencyFormat::CurrencyFormat(const CurrencyFormat& other) :
    MeasureFormat(other)
{
}

CurrencyFormat::~CurrencyFormat() {
}

CurrencyFormat* CurrencyFormat::clone() const {
    return new CurrencyFormat(*this);
}

UnicodeString& CurrencyFormat::format(const Formattable& obj,
                                      UnicodeString& appendTo,
                                      FieldPosition& pos,
                                      UErrorCode& ec) const
{
    return getCurrencyFormatInternal().format(obj, appendTo, pos, ec);
}

void CurrencyFormat::parseObject(const UnicodeString& source,
                                 Formattable& result,
                                 ParsePosition& pos) const
{
    CurrencyAmount* currAmt = getCurrencyFormatInternal().parseCurrency(source, pos);
    if (currAmt != NULL) {
        result.adoptObject(currAmt);
    }
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(CurrencyFormat)

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2009-2014, International Business Machines Corporation and
 * others. All Rights Reserved.
 *******************************************************************************
 */



#if !UCONFIG_NO_FORMATTING

//#define CURRENCY_PLURAL_INFO_DEBUG 1

#ifdef CURRENCY_PLURAL_INFO_DEBUG
#include <iostream>
#endif











U_NAMESPACE_BEGIN

static const UChar gNumberPatternSeparator = 0x3B; // ;

U_CDECL_BEGIN

/**
 * @internal ICU 4.2
 */
static UBool U_CALLCONV ValueComparator(UHashTok val1, UHashTok val2);

UBool
U_CALLCONV ValueComparator(UHashTok val1, UHashTok val2) {
    const UnicodeString* affix_1 = (UnicodeString*)val1.pointer;
    const UnicodeString* affix_2 = (UnicodeString*)val2.pointer;
    return  *affix_1 == *affix_2;
}

U_CDECL_END


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(CurrencyPluralInfo)

static const UChar currpinf_gDefaultCurrencyPluralPattern[] = {'0', '.', '#', '#', ' ', 0xA4, 0xA4, 0xA4, 0};
static const UChar currpinf_gTripleCurrencySign[] = {0xA4, 0xA4, 0xA4, 0};
static const UChar currpinf_gPluralCountOther[] = {0x6F, 0x74, 0x68, 0x65, 0x72, 0};
static const UChar currpinf_gPart0[] = {0x7B, 0x30, 0x7D, 0};
static const UChar currpinf_gPart1[] = {0x7B, 0x31, 0x7D, 0};

static const char currpinf_gNumberElementsTag[]="NumberElements";
static const char currpinf_gLatnTag[]="latn";
static const char currpinf_gPatternsTag[]="patterns";
static const char currpinf_gDecimalFormatTag[]="decimalFormat";
static const char currpinf_gCurrUnitPtnTag[]="CurrencyUnitPatterns";

CurrencyPluralInfo::CurrencyPluralInfo(UErrorCode& status)
:   fPluralCountToCurrencyUnitPattern(nullptr),
    fPluralRules(nullptr),
    fLocale(nullptr),
    fInternalStatus(U_ZERO_ERROR) {
    initialize(Locale::getDefault(), status);
}

CurrencyPluralInfo::CurrencyPluralInfo(const Locale& locale, UErrorCode& status)
:   fPluralCountToCurrencyUnitPattern(nullptr),
    fPluralRules(nullptr),
    fLocale(nullptr),
    fInternalStatus(U_ZERO_ERROR) {
    initialize(locale, status);
}

CurrencyPluralInfo::CurrencyPluralInfo(const CurrencyPluralInfo& info)
:   UObject(info),
    fPluralCountToCurrencyUnitPattern(nullptr),
    fPluralRules(nullptr),
    fLocale(nullptr),
    fInternalStatus(U_ZERO_ERROR) {
    *this = info;
}

CurrencyPluralInfo&
CurrencyPluralInfo::operator=(const CurrencyPluralInfo& info) {
    if (this == &info) {
        return *this;
    }

    fInternalStatus = info.fInternalStatus;
    if (U_FAILURE(fInternalStatus)) {
        // bail out early if the object we were copying from was already 'invalid'.
        return *this;
    }

    deleteHash(fPluralCountToCurrencyUnitPattern);
    fPluralCountToCurrencyUnitPattern = initHash(fInternalStatus);
    copyHash(info.fPluralCountToCurrencyUnitPattern,
             fPluralCountToCurrencyUnitPattern, fInternalStatus);
    if ( U_FAILURE(fInternalStatus) ) {
        return *this;
    }

    delete fPluralRules;
    fPluralRules = nullptr;
    delete fLocale;
    fLocale = nullptr;

    if (info.fPluralRules != nullptr) {
        fPluralRules = info.fPluralRules->clone();
        if (fPluralRules == nullptr) {
            fInternalStatus = U_MEMORY_ALLOCATION_ERROR;
            return *this;
        }
    }
    if (info.fLocale != nullptr) {
        fLocale = info.fLocale->clone();
        if (fLocale == nullptr) {
            // Note: If clone had an error parameter, then we could check/set that instead.
            fInternalStatus = U_MEMORY_ALLOCATION_ERROR;
            return *this;
        }
        // If the other locale wasn't bogus, but our clone'd locale is bogus, then OOM happened
        // during the call to clone().
        if (!info.fLocale->isBogus() && fLocale->isBogus()) {
            fInternalStatus = U_MEMORY_ALLOCATION_ERROR;
            return *this;
        }
    }
    return *this;
}

CurrencyPluralInfo::~CurrencyPluralInfo() {
    deleteHash(fPluralCountToCurrencyUnitPattern);
    fPluralCountToCurrencyUnitPattern = nullptr;
    delete fPluralRules;
    delete fLocale;
    fPluralRules = nullptr;
    fLocale = nullptr;
}

UBool
CurrencyPluralInfo::operator==(const CurrencyPluralInfo& info) const {
#ifdef CURRENCY_PLURAL_INFO_DEBUG
    if (*fPluralRules == *info.fPluralRules) {
        std::cout << "same plural rules\n";
    }
    if (*fLocale == *info.fLocale) {
        std::cout << "same locale\n";
    }
    if (fPluralCountToCurrencyUnitPattern->equals(*info.fPluralCountToCurrencyUnitPattern)) {
        std::cout << "same pattern\n";
    }
#endif
    return *fPluralRules == *info.fPluralRules &&
           *fLocale == *info.fLocale &&
           fPluralCountToCurrencyUnitPattern->equals(*info.fPluralCountToCurrencyUnitPattern);
}


CurrencyPluralInfo*
CurrencyPluralInfo::clone() const {
    CurrencyPluralInfo* newObj = new CurrencyPluralInfo(*this);
    // Since clone doesn't have a 'status' parameter, the best we can do is return nullptr
    // if the new object was not full constructed properly (an error occurred).
    if (newObj != nullptr && U_FAILURE(newObj->fInternalStatus)) {
        delete newObj;
        newObj = nullptr;
    }
    return newObj;
}

const PluralRules*
CurrencyPluralInfo::getPluralRules() const {
    return fPluralRules;
}

UnicodeString&
CurrencyPluralInfo::getCurrencyPluralPattern(const UnicodeString&  pluralCount,
                                             UnicodeString& result) const {
    const UnicodeString* currencyPluralPattern =
        (UnicodeString*)fPluralCountToCurrencyUnitPattern->get(pluralCount);
    if (currencyPluralPattern == nullptr) {
        // fall back to "other"
        if (pluralCount.compare(currpinf_gPluralCountOther, 5)) {
            currencyPluralPattern =
                (UnicodeString*)fPluralCountToCurrencyUnitPattern->get(UnicodeString(TRUE, currpinf_gPluralCountOther, 5));
        }
        if (currencyPluralPattern == nullptr) {
            // no currencyUnitPatterns defined,
            // fallback to predefined default.
            // This should never happen when ICU resource files are
            // available, since currencyUnitPattern of "other" is always
            // defined in root.
            result = UnicodeString(currpinf_gDefaultCurrencyPluralPattern);
            return result;
        }
    }
    result = *currencyPluralPattern;
    return result;
}

const Locale&
CurrencyPluralInfo::getLocale() const {
    return *fLocale;
}

void
CurrencyPluralInfo::setPluralRules(const UnicodeString& ruleDescription,
                                   UErrorCode& status) {
    if (U_SUCCESS(status)) {
        delete fPluralRules;
        fPluralRules = PluralRules::createRules(ruleDescription, status);
    }
}

void
CurrencyPluralInfo::setCurrencyPluralPattern(const UnicodeString& pluralCount,
                                             const UnicodeString& pattern,
                                             UErrorCode& status) {
    if (U_SUCCESS(status)) {
        UnicodeString* oldValue = static_cast<UnicodeString*>(
            fPluralCountToCurrencyUnitPattern->get(pluralCount));
        delete oldValue;
        LocalPointer<UnicodeString> p(new UnicodeString(pattern), status);
        if (U_SUCCESS(status)) {
            // the p object allocated above will be owned by fPluralCountToCurrencyUnitPattern
            // after the call to put(), even if the method returns failure.
            fPluralCountToCurrencyUnitPattern->put(pluralCount, p.orphan(), status);
        }
    }
}

void
CurrencyPluralInfo::setLocale(const Locale& loc, UErrorCode& status) {
    initialize(loc, status);
}

void
CurrencyPluralInfo::initialize(const Locale& loc, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    delete fLocale;
    fLocale = nullptr;
    delete fPluralRules;
    fPluralRules = nullptr;

    fLocale = loc.clone();
    if (fLocale == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    // If the locale passed in wasn't bogus, but our clone'd locale is bogus, then OOM happened
    // during the call to loc.clone().
    if (!loc.isBogus() && fLocale->isBogus()) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    fPluralRules = PluralRules::forLocale(loc, status);
    setupCurrencyPluralPattern(loc, status);
}

void
CurrencyPluralInfo::setupCurrencyPluralPattern(const Locale& loc, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }

    deleteHash(fPluralCountToCurrencyUnitPattern);
    fPluralCountToCurrencyUnitPattern = initHash(status);
    if (U_FAILURE(status)) {
        return;
    }

    LocalPointer<NumberingSystem> ns(NumberingSystem::createInstance(loc, status), status);
    if (U_FAILURE(status)) {
        return;
    }
    UErrorCode ec = U_ZERO_ERROR;
    LocalUResourceBundlePointer rb(ures_open(nullptr, loc.getName(), &ec));
    LocalUResourceBundlePointer numElements(ures_getByKeyWithFallback(rb.getAlias(), currpinf_gNumberElementsTag, nullptr, &ec));
    ures_getByKeyWithFallback(numElements.getAlias(), ns->getName(), rb.getAlias(), &ec);
    ures_getByKeyWithFallback(rb.getAlias(), currpinf_gPatternsTag, rb.getAlias(), &ec);
    int32_t ptnLen;
    const UChar* numberStylePattern = ures_getStringByKeyWithFallback(rb.getAlias(), currpinf_gDecimalFormatTag, &ptnLen, &ec);
    // Fall back to "latn" if num sys specific pattern isn't there.
    if ( ec == U_MISSING_RESOURCE_ERROR && (uprv_strcmp(ns->getName(), currpinf_gLatnTag) != 0)) {
        ec = U_ZERO_ERROR;
        ures_getByKeyWithFallback(numElements.getAlias(), currpinf_gLatnTag, rb.getAlias(), &ec);
        ures_getByKeyWithFallback(rb.getAlias(), currpinf_gPatternsTag, rb.getAlias(), &ec);
        numberStylePattern = ures_getStringByKeyWithFallback(rb.getAlias(), currpinf_gDecimalFormatTag, &ptnLen, &ec);
    }
    int32_t numberStylePatternLen = ptnLen;
    const UChar* negNumberStylePattern = nullptr;
    int32_t negNumberStylePatternLen = 0;
    // TODO: Java
    // parse to check whether there is ";" separator in the numberStylePattern
    UBool hasSeparator = false;
    if (U_SUCCESS(ec)) {
        for (int32_t styleCharIndex = 0; styleCharIndex < ptnLen; ++styleCharIndex) {
            if (numberStylePattern[styleCharIndex] == gNumberPatternSeparator) {
                hasSeparator = true;
                // split the number style pattern into positive and negative
                negNumberStylePattern = numberStylePattern + styleCharIndex + 1;
                negNumberStylePatternLen = ptnLen - styleCharIndex - 1;
                numberStylePatternLen = styleCharIndex;
            }
        }
    }

    if (U_FAILURE(ec)) {
        // If OOM occurred during the above code, then we want to report that back to the caller.
        if (ec == U_MEMORY_ALLOCATION_ERROR) {
            status = ec;
        }
        return;
    }

    LocalUResourceBundlePointer currRb(ures_open(U_ICUDATA_CURR, loc.getName(), &ec));
    LocalUResourceBundlePointer currencyRes(ures_getByKeyWithFallback(currRb.getAlias(), currpinf_gCurrUnitPtnTag, nullptr, &ec));

#ifdef CURRENCY_PLURAL_INFO_DEBUG
    std::cout << "in set up\n";
#endif
    LocalPointer<StringEnumeration> keywords(fPluralRules->getKeywords(ec), ec);
    if (U_SUCCESS(ec)) {
        const char* pluralCount;
        while (((pluralCount = keywords->next(nullptr, ec)) != nullptr) && U_SUCCESS(ec)) {
            int32_t ptnLength;
            UErrorCode err = U_ZERO_ERROR;
            const UChar* patternChars = ures_getStringByKeyWithFallback(currencyRes.getAlias(), pluralCount, &ptnLength, &err);
            if (err == U_MEMORY_ALLOCATION_ERROR || patternChars == nullptr) {
                ec = err;
                break;
            }
            if (U_SUCCESS(err) && ptnLength > 0) {
                UnicodeString* pattern = new UnicodeString(patternChars, ptnLength);
                if (pattern == nullptr) {
                    ec = U_MEMORY_ALLOCATION_ERROR;
                    break;
                }
#ifdef CURRENCY_PLURAL_INFO_DEBUG
                char result_1[1000];
                pattern->extract(0, pattern->length(), result_1, "UTF-8");
                std::cout << "pluralCount: " << pluralCount << "; pattern: " << result_1 << "\n";
#endif
                pattern->findAndReplace(UnicodeString(TRUE, currpinf_gPart0, 3),
                    UnicodeString(numberStylePattern, numberStylePatternLen));
                pattern->findAndReplace(UnicodeString(TRUE, currpinf_gPart1, 3), UnicodeString(TRUE, currpinf_gTripleCurrencySign, 3));

                if (hasSeparator) {
                    UnicodeString negPattern(patternChars, ptnLength);
                    negPattern.findAndReplace(UnicodeString(TRUE, currpinf_gPart0, 3),
                        UnicodeString(negNumberStylePattern, negNumberStylePatternLen));
                    negPattern.findAndReplace(UnicodeString(TRUE, currpinf_gPart1, 3), UnicodeString(TRUE, currpinf_gTripleCurrencySign, 3));
                    pattern->append(gNumberPatternSeparator);
                    pattern->append(negPattern);
                }
#ifdef CURRENCY_PLURAL_INFO_DEBUG
                pattern->extract(0, pattern->length(), result_1, "UTF-8");
                std::cout << "pluralCount: " << pluralCount << "; pattern: " << result_1 << "\n";
#endif
                // the 'pattern' object allocated above will be owned by the fPluralCountToCurrencyUnitPattern after the call to
                // put(), even if the method returns failure.
                fPluralCountToCurrencyUnitPattern->put(UnicodeString(pluralCount, -1, US_INV), pattern, status);
            }
        }
    }
    // If OOM occurred during the above code, then we want to report that back to the caller.
    if (ec == U_MEMORY_ALLOCATION_ERROR) {
        status = ec;
    }
}

void
CurrencyPluralInfo::deleteHash(Hashtable* hTable) {
    if ( hTable == nullptr ) {
        return;
    }
    int32_t pos = UHASH_FIRST;
    const UHashElement* element = nullptr;
    while ( (element = hTable->nextElement(pos)) != nullptr ) {
        const UHashTok valueTok = element->value;
        const UnicodeString* value = (UnicodeString*)valueTok.pointer;
        delete value;
    }
    delete hTable;
    hTable = nullptr;
}

Hashtable*
CurrencyPluralInfo::initHash(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    LocalPointer<Hashtable> hTable(new Hashtable(TRUE, status), status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    hTable->setValueComparator(ValueComparator);
    return hTable.orphan();
}

void
CurrencyPluralInfo::copyHash(const Hashtable* source,
                           Hashtable* target,
                           UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    int32_t pos = UHASH_FIRST;
    const UHashElement* element = nullptr;
    if (source) {
        while ( (element = source->nextElement(pos)) != nullptr ) {
            const UHashTok keyTok = element->key;
            const UnicodeString* key = (UnicodeString*)keyTok.pointer;
            const UHashTok valueTok = element->value;
            const UnicodeString* value = (UnicodeString*)valueTok.pointer;
            LocalPointer<UnicodeString> copy(new UnicodeString(*value), status);
            if (U_FAILURE(status)) {
                return;
            }
            // The HashTable owns the 'copy' object after the call to put().
            target->put(UnicodeString(*key), copy.orphan(), status);
            if (U_FAILURE(status)) {
                return;
            }
        }
    }
}

U_NAMESPACE_END

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 26, 2004
* Since: ICU 3.0
**********************************************************************
*/


#if !UCONFIG_NO_FORMATTING







static constexpr char16_t kDefaultCurrency[] = u"XXX";
static constexpr char kDefaultCurrency8[] = "XXX";

U_NAMESPACE_BEGIN

CurrencyUnit::CurrencyUnit(ConstChar16Ptr _isoCode, UErrorCode& ec) {
    // The constructor always leaves the CurrencyUnit in a valid state (with a 3-character currency code).
    // Note: in ICU4J Currency.getInstance(), we check string length for 3, but in ICU4C we allow a
    // non-NUL-terminated string to be passed as an argument, so it is not possible to check length.
    // However, we allow a NUL-terminated empty string, which should have the same behavior as nullptr.
    // Consider NUL-terminated strings of length 1 or 2 as invalid.
    const char16_t* isoCodeToUse;
    if (U_FAILURE(ec) || _isoCode == nullptr || _isoCode[0] == 0) {
        isoCodeToUse = kDefaultCurrency;
    } else if (_isoCode[1] == 0 || _isoCode[2] == 0) {
        isoCodeToUse = kDefaultCurrency;
        ec = U_ILLEGAL_ARGUMENT_ERROR;
    } else if (!uprv_isInvariantUString(_isoCode, 3)) {
        // TODO: Perform a more strict ASCII check like in ICU4J isAlpha3Code?
        isoCodeToUse = kDefaultCurrency;
        ec = U_INVARIANT_CONVERSION_ERROR;
    } else {
        isoCodeToUse = _isoCode;
    }
    // TODO: Perform uppercasing here like in ICU4J Currency.getInstance()?
    uprv_memcpy(isoCode, isoCodeToUse, sizeof(UChar) * 3);
    isoCode[3] = 0;
    char simpleIsoCode[4];
    u_UCharsToChars(isoCode, simpleIsoCode, 4);
    initCurrency(simpleIsoCode);
}

CurrencyUnit::CurrencyUnit(StringPiece _isoCode, UErrorCode& ec) {
    // Note: unlike the old constructor, reject empty arguments with an error.
    char isoCodeBuffer[4];
    const char* isoCodeToUse;
    // uprv_memchr checks that the string contains no internal NULs
    if (_isoCode.length() != 3 || uprv_memchr(_isoCode.data(), 0, 3) != nullptr) {
        isoCodeToUse = kDefaultCurrency8;
        ec = U_ILLEGAL_ARGUMENT_ERROR;
    } else if (!uprv_isInvariantString(_isoCode.data(), 3)) {
        // TODO: Perform a more strict ASCII check like in ICU4J isAlpha3Code?
        isoCodeToUse = kDefaultCurrency8;
        ec = U_INVARIANT_CONVERSION_ERROR;
    } else {
        // Have to use isoCodeBuffer to ensure the string is NUL-terminated
        uprv_strncpy(isoCodeBuffer, _isoCode.data(), 3);
        isoCodeBuffer[3] = 0;
        isoCodeToUse = isoCodeBuffer;
    }
    // TODO: Perform uppercasing here like in ICU4J Currency.getInstance()?
    u_charsToUChars(isoCodeToUse, isoCode, 3);
    isoCode[3] = 0;
    initCurrency(isoCodeToUse);
}

CurrencyUnit::CurrencyUnit(const CurrencyUnit& other) : MeasureUnit(other) {
    u_strcpy(isoCode, other.isoCode);
}

CurrencyUnit::CurrencyUnit(const MeasureUnit& other, UErrorCode& ec) : MeasureUnit(other) {
    // Make sure this is a currency.
    // OK to hard-code the string because we are comparing against another hard-coded string.
    if (uprv_strcmp("currency", getType()) != 0) {
        ec = U_ILLEGAL_ARGUMENT_ERROR;
        isoCode[0] = 0;
    } else {
        // Get the ISO Code from the subtype field.
        u_charsToUChars(getSubtype(), isoCode, 4);
        isoCode[3] = 0; // make 100% sure it is NUL-terminated
    }
}

CurrencyUnit::CurrencyUnit() : MeasureUnit() {
    u_strcpy(isoCode, kDefaultCurrency);
    char simpleIsoCode[4];
    u_UCharsToChars(isoCode, simpleIsoCode, 4);
    initCurrency(simpleIsoCode);
}

CurrencyUnit& CurrencyUnit::operator=(const CurrencyUnit& other) {
    if (this == &other) {
        return *this;
    }
    MeasureUnit::operator=(other);
    u_strcpy(isoCode, other.isoCode);
    return *this;
}

CurrencyUnit* CurrencyUnit::clone() const {
    return new CurrencyUnit(*this);
}

CurrencyUnit::~CurrencyUnit() {
}
    
UOBJECT_DEFINE_RTTI_IMPLEMENTATION(CurrencyUnit)

U_NAMESPACE_END

#endif // !UCONFIG_NO_FORMATTING
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ******************************************************************************
 * Copyright (C) 2013, International Business Machines Corporation
 * and others. All Rights Reserved.
 ******************************************************************************
 *
 * File DANGICAL.CPP
 *****************************************************************************
 */




#if !UCONFIG_NO_FORMATTING

 // Math






// --- The cache --
static icu::TimeZone *gDangiCalendarZoneAstroCalc = NULL;
static icu::UInitOnce gDangiCalendarInitOnce = U_INITONCE_INITIALIZER;

/**
 * The start year of the Korean traditional calendar (Dan-gi) is the inaugural
 * year of Dan-gun (BC 2333).
 */
static const int32_t DANGI_EPOCH_YEAR = -2332; // Gregorian year

U_CDECL_BEGIN
static UBool calendar_dangi_cleanup(void) {
    if (gDangiCalendarZoneAstroCalc) {
        delete gDangiCalendarZoneAstroCalc;
        gDangiCalendarZoneAstroCalc = NULL;
    }
    gDangiCalendarInitOnce.reset();
    return TRUE;
}
U_CDECL_END

U_NAMESPACE_BEGIN

// Implementation of the DangiCalendar class

//-------------------------------------------------------------------------
// Constructors...
//-------------------------------------------------------------------------

DangiCalendar::DangiCalendar(const Locale& aLocale, UErrorCode& success)
:   ChineseCalendar(aLocale, DANGI_EPOCH_YEAR, getDangiCalZoneAstroCalc(), success)
{
}

DangiCalendar::DangiCalendar (const DangiCalendar& other) 
: ChineseCalendar(other)
{
}

DangiCalendar::~DangiCalendar()
{
}

DangiCalendar*
DangiCalendar::clone() const
{
    return new DangiCalendar(*this);
}

const char *DangiCalendar::getType() const { 
    return "dangi";
}

/**
 * The time zone used for performing astronomical computations for
 * Dangi calendar. In Korea various timezones have been used historically 
 * (cf. http://www.math.snu.ac.kr/~kye/others/lunar.html): 
 *  
 *            - 1908/04/01: GMT+8 
 * 1908/04/01 - 1911/12/31: GMT+8.5 
 * 1912/01/01 - 1954/03/20: GMT+9 
 * 1954/03/21 - 1961/08/09: GMT+8.5 
 * 1961/08/10 -           : GMT+9 
 *  
 * Note that, in 1908-1911, the government did not apply the timezone change 
 * but used GMT+8. In addition, 1954-1961's timezone change does not affect 
 * the lunar date calculation. Therefore, the following simpler rule works: 
 *   
 * -1911: GMT+8 
 * 1912-: GMT+9 
 *  
 * Unfortunately, our astronomer's approximation doesn't agree with the 
 * references (http://www.math.snu.ac.kr/~kye/others/lunar.html and 
 * http://astro.kasi.re.kr/Life/ConvertSolarLunarForm.aspx?MenuID=115) 
 * in 1897/7/30. So the following ad hoc fix is used here: 
 *  
 *     -1896: GMT+8 
 *      1897: GMT+7 
 * 1898-1911: GMT+8 
 * 1912-    : GMT+9 
 */
static void U_CALLCONV initDangiCalZoneAstroCalc(void) {
    U_ASSERT(gDangiCalendarZoneAstroCalc == NULL);
    const UDate millis1897[] = { (UDate)((1897 - 1970) * 365 * kOneDay) }; // some days of error is not a problem here
    const UDate millis1898[] = { (UDate)((1898 - 1970) * 365 * kOneDay) }; // some days of error is not a problem here
    const UDate millis1912[] = { (UDate)((1912 - 1970) * 365 * kOneDay) }; // this doesn't create an issue for 1911/12/20
    InitialTimeZoneRule* initialTimeZone = new InitialTimeZoneRule(UNICODE_STRING_SIMPLE("GMT+8"), 8*kOneHour, 0);
    TimeZoneRule* rule1897 = new TimeArrayTimeZoneRule(UNICODE_STRING_SIMPLE("Korean 1897"), 7*kOneHour, 0, millis1897, 1, DateTimeRule::STANDARD_TIME);
    TimeZoneRule* rule1898to1911 = new TimeArrayTimeZoneRule(UNICODE_STRING_SIMPLE("Korean 1898-1911"), 8*kOneHour, 0, millis1898, 1, DateTimeRule::STANDARD_TIME);
    TimeZoneRule* ruleFrom1912 = new TimeArrayTimeZoneRule(UNICODE_STRING_SIMPLE("Korean 1912-"), 9*kOneHour, 0, millis1912, 1, DateTimeRule::STANDARD_TIME);
    UErrorCode status = U_ZERO_ERROR;
    RuleBasedTimeZone* dangiCalZoneAstroCalc = new RuleBasedTimeZone(UNICODE_STRING_SIMPLE("KOREA_ZONE"), initialTimeZone); // adopts initialTimeZone
    dangiCalZoneAstroCalc->addTransitionRule(rule1897, status); // adopts rule1897
    dangiCalZoneAstroCalc->addTransitionRule(rule1898to1911, status);
    dangiCalZoneAstroCalc->addTransitionRule(ruleFrom1912, status);
    dangiCalZoneAstroCalc->complete(status);
    if (U_SUCCESS(status)) {
        gDangiCalendarZoneAstroCalc = dangiCalZoneAstroCalc;
    } else {
        delete dangiCalZoneAstroCalc;
        gDangiCalendarZoneAstroCalc = NULL;
    }
    ucln_i18n_registerCleanup(UCLN_I18N_DANGI_CALENDAR, calendar_dangi_cleanup);
}

const TimeZone* DangiCalendar::getDangiCalZoneAstroCalc(void) const {
    umtx_initOnce(gDangiCalendarInitOnce, &initDangiCalZoneAstroCalc);
    return gDangiCalendarZoneAstroCalc;
}


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DangiCalendar)

U_NAMESPACE_END

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 1997-2015, International Business Machines Corporation and    *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 *
 * File DATEFMT.CPP
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   02/19/97    aliu        Converted from java.
 *   03/31/97    aliu        Modified extensively to work with 50 locales.
 *   04/01/97    aliu        Added support for centuries.
 *   08/12/97    aliu        Fixed operator== to use Calendar::equivalentTo.
 *   07/20/98    stephen     Changed ParsePosition initialization
 ********************************************************************************
 */



#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ********************************************************************************
 *   Copyright (C) 1997-2016, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 ********************************************************************************
 *
 * File DATEFMT.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   02/19/97    aliu        Converted from java.
 *   04/01/97    aliu        Added support for centuries.
 *   07/23/98    stephen     JDK 1.2 sync
 *   11/15/99    weiv        Added support for week of year/day of week formatting
 ********************************************************************************
 */




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING









/**
 * \file
 * \brief C++ API: Abstract class for converting dates.
 */

U_NAMESPACE_BEGIN

class TimeZone;
class DateTimePatternGenerator;

/**
 * \cond
 * Export an explicit template instantiation. (See digitlst.h, datefmt.h, and others.)
 * (When building DLLs for Windows this is required.)
 */
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN && !defined(U_IN_DOXYGEN)
template class U_I18N_API EnumSet<UDateFormatBooleanAttribute,
            0,
            UDAT_BOOLEAN_ATTRIBUTE_COUNT>;
#endif
/** \endcond */

/**
 * DateFormat is an abstract class for a family of classes that convert dates and
 * times from their internal representations to textual form and back again in a
 * language-independent manner. Converting from the internal representation (milliseconds
 * since midnight, January 1, 1970) to text is known as "formatting," and converting
 * from text to millis is known as "parsing."  We currently define only one concrete
 * subclass of DateFormat: SimpleDateFormat, which can handle pretty much all normal
 * date formatting and parsing actions.
 * <P>
 * DateFormat helps you to format and parse dates for any locale. Your code can
 * be completely independent of the locale conventions for months, days of the
 * week, or even the calendar format: lunar vs. solar.
 * <P>
 * To format a date for the current Locale, use one of the static factory
 * methods:
 * <pre>
 * \code
 *      DateFormat* dfmt = DateFormat::createDateInstance();
 *      UDate myDate = Calendar::getNow();
 *      UnicodeString myString;
 *      myString = dfmt->format( myDate, myString );
 * \endcode
 * </pre>
 * If you are formatting multiple numbers, it is more efficient to get the
 * format and use it multiple times so that the system doesn't have to fetch the
 * information about the local language and country conventions multiple times.
 * <pre>
 * \code
 *      DateFormat* df = DateFormat::createDateInstance();
 *      UnicodeString myString;
 *      UDate myDateArr[] = { 0.0, 100000000.0, 2000000000.0 }; // test values
 *      for (int32_t i = 0; i < 3; ++i) {
 *          myString.remove();
 *          cout << df->format( myDateArr[i], myString ) << endl;
 *      }
 * \endcode
 * </pre>
 * To get specific fields of a date, you can use UFieldPosition to
 * get specific fields.
 * <pre>
 * \code
 *      DateFormat* dfmt = DateFormat::createDateInstance();
 *      FieldPosition pos(DateFormat::YEAR_FIELD);
 *      UnicodeString myString;
 *      myString = dfmt->format( myDate, myString );
 *      cout << myString << endl;
 *      cout << pos.getBeginIndex() << "," << pos. getEndIndex() << endl;
 * \endcode
 * </pre>
 * To format a date for a different Locale, specify it in the call to
 * createDateInstance().
 * <pre>
 * \code
 *       DateFormat* df =
 *           DateFormat::createDateInstance( DateFormat::SHORT, Locale::getFrance());
 * \endcode
 * </pre>
 * You can use a DateFormat to parse also.
 * <pre>
 * \code
 *       UErrorCode status = U_ZERO_ERROR;
 *       UDate myDate = df->parse(myString, status);
 * \endcode
 * </pre>
 * Use createDateInstance() to produce the normal date format for that country.
 * There are other static factory methods available. Use createTimeInstance()
 * to produce the normal time format for that country. Use createDateTimeInstance()
 * to produce a DateFormat that formats both date and time. You can pass in
 * different options to these factory methods to control the length of the
 * result; from SHORT to MEDIUM to LONG to FULL. The exact result depends on the
 * locale, but generally:
 * <ul type=round>
 *   <li>   SHORT is completely numeric, such as 12/13/52 or 3:30pm
 *   <li>   MEDIUM is longer, such as Jan 12, 1952
 *   <li>   LONG is longer, such as January 12, 1952 or 3:30:32pm
 *   <li>   FULL is pretty completely specified, such as
 *          Tuesday, April 12, 1952 AD or 3:30:42pm PST.
 * </ul>
 * You can also set the time zone on the format if you wish. If you want even
 * more control over the format or parsing, (or want to give your users more
 * control), you can try casting the DateFormat you get from the factory methods
 * to a SimpleDateFormat. This will work for the majority of countries; just
 * remember to chck getDynamicClassID() before carrying out the cast.
 * <P>
 * You can also use forms of the parse and format methods with ParsePosition and
 * FieldPosition to allow you to
 * <ul type=round>
 *   <li>   Progressively parse through pieces of a string.
 *   <li>   Align any particular field, or find out where it is for selection
 *          on the screen.
 * </ul>
 *
 * <p><em>User subclasses are not supported.</em> While clients may write
 * subclasses, such code will not necessarily work and will not be
 * guaranteed to work stably from release to release.
 */
class U_I18N_API DateFormat : public Format {
public:

    /**
     * Constants for various style patterns. These reflect the order of items in
     * the DateTimePatterns resource. There are 4 time patterns, 4 date patterns,
     * the default date-time pattern, and 4 date-time patterns. Each block of 4 values
     * in the resource occurs in the order full, long, medium, short.
     * @stable ICU 2.4
     */
    enum EStyle
    {
        kNone   = -1,

        kFull   = 0,
        kLong   = 1,
        kMedium = 2,
        kShort  = 3,

        kDateOffset   = kShort + 1,
     // kFull   + kDateOffset = 4
     // kLong   + kDateOffset = 5
     // kMedium + kDateOffset = 6
     // kShort  + kDateOffset = 7

        kDateTime             = 8,
     // Default DateTime

        kDateTimeOffset = kDateTime + 1,
     // kFull   + kDateTimeOffset = 9
     // kLong   + kDateTimeOffset = 10
     // kMedium + kDateTimeOffset = 11
     // kShort  + kDateTimeOffset = 12

        // relative dates
        kRelative = (1 << 7),

        kFullRelative = (kFull | kRelative),

        kLongRelative = kLong | kRelative,

        kMediumRelative = kMedium | kRelative,

        kShortRelative = kShort | kRelative,


        kDefault      = kMedium,



    /**
     * These constants are provided for backwards compatibility only.
     * Please use the C++ style constants defined above.
     */
        FULL        = kFull,
        LONG        = kLong,
        MEDIUM        = kMedium,
        SHORT        = kShort,
        DEFAULT        = kDefault,
        DATE_OFFSET    = kDateOffset,
        NONE        = kNone,
        DATE_TIME    = kDateTime
    };

    /**
     * Destructor.
     * @stable ICU 2.0
     */
    virtual ~DateFormat();

    /**
     * Clones this object polymorphically.
     * The caller owns the result and should delete it when done.
     * @return clone, or nullptr if an error occurred
     * @stable ICU 2.0
     */
    virtual DateFormat* clone() const = 0;

    /**
     * Equality operator.  Returns true if the two formats have the same behavior.
     * @stable ICU 2.0
     */
    virtual UBool operator==(const Format&) const;


    using Format::format;

    /**
     * Format an object to produce a string. This method handles Formattable
     * objects with a UDate type. If a the Formattable object type is not a Date,
     * then it returns a failing UErrorCode.
     *
     * @param obj       The object to format. Must be a Date.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    virtual UnicodeString& format(const Formattable& obj,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode& status) const;

    /**
     * Format an object to produce a string. This method handles Formattable
     * objects with a UDate type. If a the Formattable object type is not a Date,
     * then it returns a failing UErrorCode.
     *
     * @param obj       The object to format. Must be a Date.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.  Field values
     *                  are defined in UDateFormatField.  Can be NULL.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    virtual UnicodeString& format(const Formattable& obj,
                                  UnicodeString& appendTo,
                                  FieldPositionIterator* posIter,
                                  UErrorCode& status) const;
    /**
     * Formats a date into a date/time string. This is an abstract method which
     * concrete subclasses must implement.
     * <P>
     * On input, the FieldPosition parameter may have its "field" member filled with
     * an enum value specifying a field.  On output, the FieldPosition will be filled
     * in with the text offsets for that field.
     * <P> For example, given a time text
     * "1996.07.10 AD at 15:08:56 PDT", if the given fieldPosition.field is
     * UDAT_YEAR_FIELD, the offsets fieldPosition.beginIndex and
     * statfieldPositionus.getEndIndex will be set to 0 and 4, respectively.
     * <P> Notice
     * that if the same time field appears more than once in a pattern, the status will
     * be set for the first occurence of that time field. For instance,
     * formatting a UDate to the time string "1 PM PDT (Pacific Daylight Time)"
     * using the pattern "h a z (zzzz)" and the alignment field
     * DateFormat::TIMEZONE_FIELD, the offsets fieldPosition.beginIndex and
     * fieldPosition.getEndIndex will be set to 5 and 8, respectively, for the first
     * occurence of the timezone pattern character 'z'.
     *
     * @param cal           Calendar set to the date and time to be formatted
     *                      into a date/time string.  When the calendar type is
     *                      different from the internal calendar held by this
     *                      DateFormat instance, the date and the time zone will
     *                      be inherited from the input calendar, but other calendar
     *                      field values will be calculated by the internal calendar.
     * @param appendTo      Output parameter to receive result.
     *                      Result is appended to existing contents.
     * @param fieldPosition On input: an alignment field, if desired (see examples above)
     *                      On output: the offsets of the alignment field (see examples above)
     * @return              Reference to 'appendTo' parameter.
     * @stable ICU 2.1
     */
    virtual UnicodeString& format(  Calendar& cal,
                                    UnicodeString& appendTo,
                                    FieldPosition& fieldPosition) const = 0;

    /**
     * Formats a date into a date/time string. Subclasses should implement this method.
     *
     * @param cal       Calendar set to the date and time to be formatted
     *                  into a date/time string.  When the calendar type is
     *                  different from the internal calendar held by this
     *                  DateFormat instance, the date and the time zone will
     *                  be inherited from the input calendar, but other calendar
     *                  field values will be calculated by the internal calendar.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.  Field values
     *                  are defined in UDateFormatField.  Can be NULL.
     * @param status    error status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    virtual UnicodeString& format(Calendar& cal,
                                  UnicodeString& appendTo,
                                  FieldPositionIterator* posIter,
                                  UErrorCode& status) const;
    /**
     * Formats a UDate into a date/time string.
     * <P>
     * On input, the FieldPosition parameter may have its "field" member filled with
     * an enum value specifying a field.  On output, the FieldPosition will be filled
     * in with the text offsets for that field.
     * <P> For example, given a time text
     * "1996.07.10 AD at 15:08:56 PDT", if the given fieldPosition.field is
     * UDAT_YEAR_FIELD, the offsets fieldPosition.beginIndex and
     * statfieldPositionus.getEndIndex will be set to 0 and 4, respectively.
     * <P> Notice
     * that if the same time field appears more than once in a pattern, the status will
     * be set for the first occurence of that time field. For instance,
     * formatting a UDate to the time string "1 PM PDT (Pacific Daylight Time)"
     * using the pattern "h a z (zzzz)" and the alignment field
     * DateFormat::TIMEZONE_FIELD, the offsets fieldPosition.beginIndex and
     * fieldPosition.getEndIndex will be set to 5 and 8, respectively, for the first
     * occurence of the timezone pattern character 'z'.
     *
     * @param date          UDate to be formatted into a date/time string.
     * @param appendTo      Output parameter to receive result.
     *                      Result is appended to existing contents.
     * @param fieldPosition On input: an alignment field, if desired (see examples above)
     *                      On output: the offsets of the alignment field (see examples above)
     * @return              Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    UnicodeString& format(  UDate date,
                            UnicodeString& appendTo,
                            FieldPosition& fieldPosition) const;

    /**
     * Formats a UDate into a date/time string.
     *
     * @param date      UDate to be formatted into a date/time string.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.  Field values
     *                  are defined in UDateFormatField.  Can be NULL.
     * @param status    error status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    UnicodeString& format(UDate date,
                          UnicodeString& appendTo,
                          FieldPositionIterator* posIter,
                          UErrorCode& status) const;
    /**
     * Formats a UDate into a date/time string. If there is a problem, you won't
     * know, using this method. Use the overloaded format() method which takes a
     * FieldPosition& to detect formatting problems.
     *
     * @param date      The UDate value to be formatted into a string.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    UnicodeString& format(UDate date, UnicodeString& appendTo) const;

    /**
     * Parse a date/time string. For example, a time text "07/10/96 4:5 PM, PDT"
     * will be parsed into a UDate that is equivalent to Date(837039928046).
     * Parsing begins at the beginning of the string and proceeds as far as
     * possible.  Assuming no parse errors were encountered, this function
     * doesn't return any information about how much of the string was consumed
     * by the parsing.  If you need that information, use the version of
     * parse() that takes a ParsePosition.
     * <P>
     * By default, parsing is lenient: If the input is not in the form used by
     * this object's format method but can still be parsed as a date, then the
     * parse succeeds. Clients may insist on strict adherence to the format by
     * calling setLenient(false).
     * @see DateFormat::setLenient(boolean)
     * <P>
     * Note that the normal date formats associated with some calendars - such
     * as the Chinese lunar calendar - do not specify enough fields to enable
     * dates to be parsed unambiguously. In the case of the Chinese lunar
     * calendar, while the year within the current 60-year cycle is specified,
     * the number of such cycles since the start date of the calendar (in the
     * ERA field of the Calendar object) is not normally part of the format,
     * and parsing may assume the wrong era. For cases such as this it is
     * recommended that clients parse using the method
     * parse(const UnicodeString&, Calendar& cal, ParsePosition&)
     * with the Calendar passed in set to the current date, or to a date
     * within the era/cycle that should be assumed if absent in the format.
     *
     * @param text      The date/time string to be parsed into a UDate value.
     * @param status    Output param to be set to success/failure code. If
     *                  'text' cannot be parsed, it will be set to a failure
     *                  code.
     * @return          The parsed UDate value, if successful.
     * @stable ICU 2.0
     */
    virtual UDate parse( const UnicodeString& text,
                        UErrorCode& status) const;

    /**
     * Parse a date/time string beginning at the given parse position. For
     * example, a time text "07/10/96 4:5 PM, PDT" will be parsed into a Date
     * that is equivalent to Date(837039928046).
     * <P>
     * By default, parsing is lenient: If the input is not in the form used by
     * this object's format method but can still be parsed as a date, then the
     * parse succeeds. Clients may insist on strict adherence to the format by
     * calling setLenient(false).
     * @see DateFormat::setLenient(boolean)
     *
     * @param text  The date/time string to be parsed.
     * @param cal   A Calendar set on input to the date and time to be used for
     *              missing values in the date/time string being parsed, and set
     *              on output to the parsed date/time. When the calendar type is
     *              different from the internal calendar held by this DateFormat
     *              instance, the internal calendar will be cloned to a work
     *              calendar set to the same milliseconds and time zone as the
     *              cal parameter, field values will be parsed based on the work
     *              calendar, then the result (milliseconds and time zone) will
     *              be set in this calendar.
     * @param pos   On input, the position at which to start parsing; on
     *              output, the position at which parsing terminated, or the
     *              start position if the parse failed.
     * @stable ICU 2.1
     */
    virtual void parse( const UnicodeString& text,
                        Calendar& cal,
                        ParsePosition& pos) const = 0;

    /**
     * Parse a date/time string beginning at the given parse position. For
     * example, a time text "07/10/96 4:5 PM, PDT" will be parsed into a Date
     * that is equivalent to Date(837039928046).
     * <P>
     * By default, parsing is lenient: If the input is not in the form used by
     * this object's format method but can still be parsed as a date, then the
     * parse succeeds. Clients may insist on strict adherence to the format by
     * calling setLenient(false).
     * @see DateFormat::setLenient(boolean)
     * <P>
     * Note that the normal date formats associated with some calendars - such
     * as the Chinese lunar calendar - do not specify enough fields to enable
     * dates to be parsed unambiguously. In the case of the Chinese lunar
     * calendar, while the year within the current 60-year cycle is specified,
     * the number of such cycles since the start date of the calendar (in the
     * ERA field of the Calendar object) is not normally part of the format,
     * and parsing may assume the wrong era. For cases such as this it is
     * recommended that clients parse using the method
     * parse(const UnicodeString&, Calendar& cal, ParsePosition&)
     * with the Calendar passed in set to the current date, or to a date
     * within the era/cycle that should be assumed if absent in the format.
     *
     * @param text  The date/time string to be parsed into a UDate value.
     * @param pos   On input, the position at which to start parsing; on
     *              output, the position at which parsing terminated, or the
     *              start position if the parse failed.
     * @return      A valid UDate if the input could be parsed.
     * @stable ICU 2.0
     */
    UDate parse( const UnicodeString& text,
                 ParsePosition& pos) const;

    /**
     * Parse a string to produce an object. This methods handles parsing of
     * date/time strings into Formattable objects with UDate types.
     * <P>
     * Before calling, set parse_pos.index to the offset you want to start
     * parsing at in the source. After calling, parse_pos.index is the end of
     * the text you parsed. If error occurs, index is unchanged.
     * <P>
     * When parsing, leading whitespace is discarded (with a successful parse),
     * while trailing whitespace is left as is.
     * <P>
     * See Format::parseObject() for more.
     *
     * @param source    The string to be parsed into an object.
     * @param result    Formattable to be set to the parse result.
     *                  If parse fails, return contents are undefined.
     * @param parse_pos The position to start parsing at. Upon return
     *                  this param is set to the position after the
     *                  last character successfully parsed. If the
     *                  source is not parsed successfully, this param
     *                  will remain unchanged.
     * @stable ICU 2.0
     */
    virtual void parseObject(const UnicodeString& source,
                             Formattable& result,
                             ParsePosition& parse_pos) const;

    /**
     * Create a default date/time formatter that uses the SHORT style for both
     * the date and the time.
     *
     * @return A date/time formatter which the caller owns.
     * @stable ICU 2.0
     */
    static DateFormat* U_EXPORT2 createInstance(void);

    /**
     * Creates a time formatter with the given formatting style for the given
     * locale.
     *
     * @param style     The given formatting style. For example,
     *                  SHORT for "h:mm a" in the US locale. Relative
     *                  time styles are not currently supported.
     * @param aLocale   The given locale.
     * @return          A time formatter which the caller owns.
     * @stable ICU 2.0
     */
    static DateFormat* U_EXPORT2 createTimeInstance(EStyle style = kDefault,
                                          const Locale& aLocale = Locale::getDefault());

    /**
     * Creates a date formatter with the given formatting style for the given
     * const locale.
     *
     * @param style     The given formatting style. For example, SHORT for "M/d/yy" in the
     *                  US locale. As currently implemented, relative date formatting only
     *                  affects a limited range of calendar days before or after the
     *                  current date, based on the CLDR &lt;field type="day"&gt;/&lt;relative&gt; data:
     *                  For example, in English, "Yesterday", "Today", and "Tomorrow".
     *                  Outside of this range, dates are formatted using the corresponding
     *                  non-relative style.
     * @param aLocale   The given locale.
     * @return          A date formatter which the caller owns.
     * @stable ICU 2.0
     */
    static DateFormat* U_EXPORT2 createDateInstance(EStyle style = kDefault,
                                          const Locale& aLocale = Locale::getDefault());

    /**
     * Creates a date/time formatter with the given formatting styles for the
     * given locale.
     *
     * @param dateStyle The given formatting style for the date portion of the result.
     *                  For example, SHORT for "M/d/yy" in the US locale. As currently
     *                  implemented, relative date formatting only affects a limited range
     *                  of calendar days before or after the current date, based on the
     *                  CLDR &lt;field type="day"&gt;/&lt;relative&gt; data: For example, in English,
     *                  "Yesterday", "Today", and "Tomorrow". Outside of this range, dates
     *                  are formatted using the corresponding non-relative style.
     * @param timeStyle The given formatting style for the time portion of the result.
     *                  For example, SHORT for "h:mm a" in the US locale. Relative
     *                  time styles are not currently supported.
     * @param aLocale   The given locale.
     * @return          A date/time formatter which the caller owns.
     * @stable ICU 2.0
     */
    static DateFormat* U_EXPORT2 createDateTimeInstance(EStyle dateStyle = kDefault,
                                              EStyle timeStyle = kDefault,
                                              const Locale& aLocale = Locale::getDefault());

#ifndef U_HIDE_INTERNAL_API
    /**
     * Returns the best pattern given a skeleton and locale.
     * @param locale the locale
     * @param skeleton the skeleton
     * @param status ICU error returned here
     * @return the best pattern.
     * @internal For ICU use only.
     */
    static UnicodeString getBestPattern(
            const Locale &locale,
            const UnicodeString &skeleton,
            UErrorCode &status);
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Creates a date/time formatter for the given skeleton and
     * default locale.
     *
     * @param skeleton The skeleton e.g "yMMMMd." Fields in the skeleton can
     *                 be in any order, and this method uses the locale to
     *                 map the skeleton to a pattern that includes locale
     *                 specific separators with the fields in the appropriate
     *                 order for that locale.
     * @param status   Any error returned here.
     * @return         A date/time formatter which the caller owns.
     * @stable ICU 55
     */
    static DateFormat* U_EXPORT2 createInstanceForSkeleton(
            const UnicodeString& skeleton,
            UErrorCode &status);

    /**
     * Creates a date/time formatter for the given skeleton and locale.
     *
     * @param skeleton The skeleton e.g "yMMMMd." Fields in the skeleton can
     *                 be in any order, and this method uses the locale to
     *                 map the skeleton to a pattern that includes locale
     *                 specific separators with the fields in the appropriate
     *                 order for that locale.
     * @param locale  The given locale.
     * @param status   Any error returned here.
     * @return         A date/time formatter which the caller owns.
     * @stable ICU 55
     */
    static DateFormat* U_EXPORT2 createInstanceForSkeleton(
            const UnicodeString& skeleton,
            const Locale &locale,
            UErrorCode &status);

    /**
     * Creates a date/time formatter for the given skeleton and locale.
     *
     * @param calendarToAdopt the calendar returned DateFormat is to use.
     * @param skeleton The skeleton e.g "yMMMMd." Fields in the skeleton can
     *                 be in any order, and this method uses the locale to
     *                 map the skeleton to a pattern that includes locale
     *                 specific separators with the fields in the appropriate
     *                 order for that locale.
     * @param locale  The given locale.
     * @param status   Any error returned here.
     * @return         A date/time formatter which the caller owns.
     * @stable ICU 55
     */
    static DateFormat* U_EXPORT2 createInstanceForSkeleton(
            Calendar *calendarToAdopt,
            const UnicodeString& skeleton,
            const Locale &locale,
            UErrorCode &status);


    /**
     * Gets the set of locales for which DateFormats are installed.
     * @param count Filled in with the number of locales in the list that is returned.
     * @return the set of locales for which DateFormats are installed.  The caller
     *  does NOT own this list and must not delete it.
     * @stable ICU 2.0
     */
    static const Locale* U_EXPORT2 getAvailableLocales(int32_t& count);

    /**
     * Returns whether both date/time parsing in the encapsulated Calendar object and DateFormat whitespace &
     * numeric processing is lenient.
     * @stable ICU 2.0
     */
    virtual UBool isLenient(void) const;

    /**
     * Specifies whether date/time parsing is to be lenient.  With
     * lenient parsing, the parser may use heuristics to interpret inputs that
     * do not precisely match this object's format.  Without lenient parsing,
     * inputs must match this object's format more closely.
     *
     * Note: ICU 53 introduced finer grained control of leniency (and added
     * new control points) making the preferred method a combination of
     * setCalendarLenient() & setBooleanAttribute() calls.
     * This method supports prior functionality but may not support all
     * future leniency control & behavior of DateFormat. For control of pre 53 leniency,
     * Calendar and DateFormat whitespace & numeric tolerance, this method is safe to
     * use. However, mixing leniency control via this method and modification of the
     * newer attributes via setBooleanAttribute() may produce undesirable
     * results.
     *
     * @param lenient  True specifies date/time interpretation to be lenient.
     * @see Calendar::setLenient
     * @stable ICU 2.0
     */
    virtual void setLenient(UBool lenient);


    /**
     * Returns whether date/time parsing in the encapsulated Calendar object processing is lenient.
     * @stable ICU 53
     */
    virtual UBool isCalendarLenient(void) const;


    /**
     * Specifies whether encapsulated Calendar date/time parsing is to be lenient.  With
     * lenient parsing, the parser may use heuristics to interpret inputs that
     * do not precisely match this object's format.  Without lenient parsing,
     * inputs must match this object's format more closely.
     * @param lenient when true, parsing is lenient
     * @see com.ibm.icu.util.Calendar#setLenient
     * @stable ICU 53
     */
    virtual void setCalendarLenient(UBool lenient);


    /**
     * Gets the calendar associated with this date/time formatter.
     * The calendar is owned by the formatter and must not be modified.
     * Also, the calendar does not reflect the results of a parse operation.
     * To parse to a calendar, use {@link #parse(const UnicodeString&, Calendar& cal, ParsePosition&) const parse(const UnicodeString&, Calendar& cal, ParsePosition&)}
     * @return the calendar associated with this date/time formatter.
     * @stable ICU 2.0
     */
    virtual const Calendar* getCalendar(void) const;

    /**
     * Set the calendar to be used by this date format. Initially, the default
     * calendar for the specified or default locale is used.  The caller should
     * not delete the Calendar object after it is adopted by this call.
     * Adopting a new calendar will change to the default symbols.
     *
     * @param calendarToAdopt    Calendar object to be adopted.
     * @stable ICU 2.0
     */
    virtual void adoptCalendar(Calendar* calendarToAdopt);

    /**
     * Set the calendar to be used by this date format. Initially, the default
     * calendar for the specified or default locale is used.
     *
     * @param newCalendar Calendar object to be set.
     * @stable ICU 2.0
     */
    virtual void setCalendar(const Calendar& newCalendar);


    /**
     * Gets the number formatter which this date/time formatter uses to format
     * and parse the numeric portions of the pattern.
     * @return the number formatter which this date/time formatter uses.
     * @stable ICU 2.0
     */
    virtual const NumberFormat* getNumberFormat(void) const;

    /**
     * Allows you to set the number formatter.  The caller should
     * not delete the NumberFormat object after it is adopted by this call.
     * @param formatToAdopt     NumberFormat object to be adopted.
     * @stable ICU 2.0
     */
    virtual void adoptNumberFormat(NumberFormat* formatToAdopt);

    /**
     * Allows you to set the number formatter.
     * @param newNumberFormat  NumberFormat object to be set.
     * @stable ICU 2.0
     */
    virtual void setNumberFormat(const NumberFormat& newNumberFormat);

    /**
     * Returns a reference to the TimeZone used by this DateFormat's calendar.
     * @return the time zone associated with the calendar of DateFormat.
     * @stable ICU 2.0
     */
    virtual const TimeZone& getTimeZone(void) const;

    /**
     * Sets the time zone for the calendar of this DateFormat object. The caller
     * no longer owns the TimeZone object and should not delete it after this call.
     * @param zoneToAdopt the TimeZone to be adopted.
     * @stable ICU 2.0
     */
    virtual void adoptTimeZone(TimeZone* zoneToAdopt);

    /**
     * Sets the time zone for the calendar of this DateFormat object.
     * @param zone the new time zone.
     * @stable ICU 2.0
     */
    virtual void setTimeZone(const TimeZone& zone);

    /**
     * Set a particular UDisplayContext value in the formatter, such as
     * UDISPCTX_CAPITALIZATION_FOR_STANDALONE.
     * @param value The UDisplayContext value to set.
     * @param status Input/output status. If at entry this indicates a failure
     *               status, the function will do nothing; otherwise this will be
     *               updated with any new status from the function.
     * @stable ICU 53
     */
    virtual void setContext(UDisplayContext value, UErrorCode& status);

    /**
     * Get the formatter's UDisplayContext value for the specified UDisplayContextType,
     * such as UDISPCTX_TYPE_CAPITALIZATION.
     * @param type The UDisplayContextType whose value to return
     * @param status Input/output status. If at entry this indicates a failure
     *               status, the function will do nothing; otherwise this will be
     *               updated with any new status from the function.
     * @return The UDisplayContextValue for the specified type.
     * @stable ICU 53
     */
    virtual UDisplayContext getContext(UDisplayContextType type, UErrorCode& status) const;

   /**
     * Sets an boolean attribute on this DateFormat.
     * May return U_UNSUPPORTED_ERROR if this instance does not support
     * the specified attribute.
     * @param attr the attribute to set
     * @param newvalue new value
     * @param status the error type
     * @return *this - for chaining (example: format.setAttribute(...).setAttribute(...) )
     * @stable ICU 53
     */

    virtual DateFormat&  U_EXPORT2 setBooleanAttribute(UDateFormatBooleanAttribute attr,
    									UBool newvalue,
    									UErrorCode &status);

    /**
     * Returns a boolean from this DateFormat
     * May return U_UNSUPPORTED_ERROR if this instance does not support
     * the specified attribute.
     * @param attr the attribute to set
     * @param status the error type
     * @return the attribute value. Undefined if there is an error.
     * @stable ICU 53
     */
    virtual UBool U_EXPORT2 getBooleanAttribute(UDateFormatBooleanAttribute attr, UErrorCode &status) const;

protected:
    /**
     * Default constructor.  Creates a DateFormat with no Calendar or NumberFormat
     * associated with it.  This constructor depends on the subclasses to fill in
     * the calendar and numberFormat fields.
     * @stable ICU 2.0
     */
    DateFormat();

    /**
     * Copy constructor.
     * @stable ICU 2.0
     */
    DateFormat(const DateFormat&);

    /**
     * Default assignment operator.
     * @stable ICU 2.0
     */
    DateFormat& operator=(const DateFormat&);

    /**
     * The calendar that DateFormat uses to produce the time field values needed
     * to implement date/time formatting. Subclasses should generally initialize
     * this to the default calendar for the locale associated with this DateFormat.
     * @stable ICU 2.4
     */
    Calendar* fCalendar;

    /**
     * The number formatter that DateFormat uses to format numbers in dates and
     * times. Subclasses should generally initialize this to the default number
     * format for the locale associated with this DateFormat.
     * @stable ICU 2.4
     */
    NumberFormat* fNumberFormat;


private:

    /**
     * Gets the date/time formatter with the given formatting styles for the
     * given locale.
     * @param dateStyle the given date formatting style.
     * @param timeStyle the given time formatting style.
     * @param inLocale the given locale.
     * @return a date/time formatter, or 0 on failure.
     */
    static DateFormat* U_EXPORT2 create(EStyle timeStyle, EStyle dateStyle, const Locale& inLocale);


    /**
     * enum set of active boolean attributes for this instance
     */
    EnumSet<UDateFormatBooleanAttribute, 0, UDAT_BOOLEAN_ATTRIBUTE_COUNT> fBoolFlags;


    UDisplayContext fCapitalizationContext;
    friend class DateFmtKeyByStyle;

public:
#ifndef U_HIDE_OBSOLETE_API
    /**
     * Field selector for FieldPosition for DateFormat fields.
     * @obsolete ICU 3.4 use UDateFormatField instead, since this API will be
     * removed in that release
     */
    enum EField
    {
        // Obsolete; use UDateFormatField instead
        kEraField = UDAT_ERA_FIELD,
        kYearField = UDAT_YEAR_FIELD,
        kMonthField = UDAT_MONTH_FIELD,
        kDateField = UDAT_DATE_FIELD,
        kHourOfDay1Field = UDAT_HOUR_OF_DAY1_FIELD,
        kHourOfDay0Field = UDAT_HOUR_OF_DAY0_FIELD,
        kMinuteField = UDAT_MINUTE_FIELD,
        kSecondField = UDAT_SECOND_FIELD,
        kMillisecondField = UDAT_FRACTIONAL_SECOND_FIELD,
        kDayOfWeekField = UDAT_DAY_OF_WEEK_FIELD,
        kDayOfYearField = UDAT_DAY_OF_YEAR_FIELD,
        kDayOfWeekInMonthField = UDAT_DAY_OF_WEEK_IN_MONTH_FIELD,
        kWeekOfYearField = UDAT_WEEK_OF_YEAR_FIELD,
        kWeekOfMonthField = UDAT_WEEK_OF_MONTH_FIELD,
        kAmPmField = UDAT_AM_PM_FIELD,
        kHour1Field = UDAT_HOUR1_FIELD,
        kHour0Field = UDAT_HOUR0_FIELD,
        kTimezoneField = UDAT_TIMEZONE_FIELD,
        kYearWOYField = UDAT_YEAR_WOY_FIELD,
        kDOWLocalField = UDAT_DOW_LOCAL_FIELD,
        kExtendedYearField = UDAT_EXTENDED_YEAR_FIELD,
        kJulianDayField = UDAT_JULIAN_DAY_FIELD,
        kMillisecondsInDayField = UDAT_MILLISECONDS_IN_DAY_FIELD,

        // Obsolete; use UDateFormatField instead
        ERA_FIELD = UDAT_ERA_FIELD,
        YEAR_FIELD = UDAT_YEAR_FIELD,
        MONTH_FIELD = UDAT_MONTH_FIELD,
        DATE_FIELD = UDAT_DATE_FIELD,
        HOUR_OF_DAY1_FIELD = UDAT_HOUR_OF_DAY1_FIELD,
        HOUR_OF_DAY0_FIELD = UDAT_HOUR_OF_DAY0_FIELD,
        MINUTE_FIELD = UDAT_MINUTE_FIELD,
        SECOND_FIELD = UDAT_SECOND_FIELD,
        MILLISECOND_FIELD = UDAT_FRACTIONAL_SECOND_FIELD,
        DAY_OF_WEEK_FIELD = UDAT_DAY_OF_WEEK_FIELD,
        DAY_OF_YEAR_FIELD = UDAT_DAY_OF_YEAR_FIELD,
        DAY_OF_WEEK_IN_MONTH_FIELD = UDAT_DAY_OF_WEEK_IN_MONTH_FIELD,
        WEEK_OF_YEAR_FIELD = UDAT_WEEK_OF_YEAR_FIELD,
        WEEK_OF_MONTH_FIELD = UDAT_WEEK_OF_MONTH_FIELD,
        AM_PM_FIELD = UDAT_AM_PM_FIELD,
        HOUR1_FIELD = UDAT_HOUR1_FIELD,
        HOUR0_FIELD = UDAT_HOUR0_FIELD,
        TIMEZONE_FIELD = UDAT_TIMEZONE_FIELD
    };
#endif  /* U_HIDE_OBSOLETE_API */
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
* Copyright (C) 1997-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File SMPDTFMT.H
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   07/09/97    helena      Make ParsePosition into a class.
*   07/21/98    stephen     Added GMT_PLUS, GMT_MINUS
*                            Changed setTwoDigitStartDate to set2DigitYearStart
*                            Changed getTwoDigitStartDate to get2DigitYearStart
*                            Removed subParseLong
*                            Removed getZoneIndex (added in DateFormatSymbols)
*   06/14/99    stephen     Removed fgTimeZoneDataSuffix
*   10/14/99    aliu        Updated class doc to describe 2-digit year parsing
*                           {j28 4182066}.
*******************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Format and parse dates in a language-independent manner.
 */

#if !UCONFIG_NO_FORMATTING



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2011-2015, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

/**
 * \file
 * \brief C++ API: TimeZoneFormat
 */



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2011-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

/**
 * \file
 * \brief C++ API: TimeZoneNames
 */


#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING




U_CDECL_BEGIN

/**
 * Constants for time zone display name types.
 * @stable ICU 50
 */
typedef enum UTimeZoneNameType {
    /**
     * Unknown display name type.
     * @stable ICU 50
     */
    UTZNM_UNKNOWN           = 0x00,
    /**
     * Long display name, such as "Eastern Time".
     * @stable ICU 50
     */
    UTZNM_LONG_GENERIC      = 0x01,
    /**
     * Long display name for standard time, such as "Eastern Standard Time".
     * @stable ICU 50
     */
    UTZNM_LONG_STANDARD     = 0x02,
    /**
     * Long display name for daylight saving time, such as "Eastern Daylight Time".
     * @stable ICU 50
     */
    UTZNM_LONG_DAYLIGHT     = 0x04,
    /**
     * Short display name, such as "ET".
     * @stable ICU 50
     */
    UTZNM_SHORT_GENERIC     = 0x08,
    /**
     * Short display name for standard time, such as "EST".
     * @stable ICU 50
     */
    UTZNM_SHORT_STANDARD    = 0x10,
    /**
     * Short display name for daylight saving time, such as "EDT".
     * @stable ICU 50
     */
    UTZNM_SHORT_DAYLIGHT    = 0x20,
    /**
     * Exemplar location name, such as "Los Angeles".
     * @stable ICU 51
     */
    UTZNM_EXEMPLAR_LOCATION = 0x40
} UTimeZoneNameType;

U_CDECL_END

U_NAMESPACE_BEGIN

class UVector;
struct MatchInfo;

/**
 * <code>TimeZoneNames</code> is an abstract class representing the time zone display name data model defined
 * by <a href="http://www.unicode.org/reports/tr35/">UTS#35 Unicode Locale Data Markup Language (LDML)</a>.
 * The model defines meta zone, which is used for storing a set of display names. A meta zone can be shared
 * by multiple time zones. Also a time zone may have multiple meta zone historic mappings.
 * <p>
 * For example, people in the United States refer the zone used by the east part of North America as "Eastern Time".
 * The tz database contains multiple time zones "America/New_York", "America/Detroit", "America/Montreal" and some
 * others that belong to "Eastern Time". However, assigning different display names to these time zones does not make
 * much sense for most of people.
 * <p>
 * In <a href="http://cldr.unicode.org/">CLDR</a> (which uses LDML for representing locale data), the display name
 * "Eastern Time" is stored as long generic display name of a meta zone identified by the ID "America_Eastern".
 * Then, there is another table maintaining the historic mapping to meta zones for each time zone. The time zones in
 * the above example ("America/New_York", "America/Detroit"...) are mapped to the meta zone "America_Eastern".
 * <p>
 * Sometimes, a time zone is mapped to a different time zone in the past. For example, "America/Indiana/Knox"
 * had been moving "Eastern Time" and "Central Time" back and forth. Therefore, it is necessary that time zone
 * to meta zones mapping data are stored by date range.
 *
 * <p><b>Note:</b>
 * The methods in this class assume that time zone IDs are already canonicalized. For example, you may not get proper
 * result returned by a method with time zone ID "America/Indiana/Indianapolis", because it's not a canonical time zone
 * ID (the canonical time zone ID for the time zone is "America/Indianapolis". See
 * {@link TimeZone#getCanonicalID(const UnicodeString& id, UnicodeString& canonicalID, UErrorCode& status)} about ICU
 * canonical time zone IDs.
 *
 * <p>
 * In CLDR, most of time zone display names except location names are provided through meta zones. But a time zone may
 * have a specific name that is not shared with other time zones.
 *
 * For example, time zone "Europe/London" has English long name for standard time "Greenwich Mean Time", which is also
 * shared with other time zones. However, the long name for daylight saving time is "British Summer Time", which is only
 * used for "Europe/London".
 *
 * <p>
 * {@link #getTimeZoneDisplayName} is designed for accessing a name only used by a single time zone.
 * But is not necessarily mean that a subclass implementation use the same model with CLDR. A subclass implementation
 * may provide time zone names only through {@link #getTimeZoneDisplayName}, or only through {@link #getMetaZoneDisplayName},
 * or both.
 *
 * <p>
 * The default <code>TimeZoneNames</code> implementation returned by {@link #createInstance}
 * uses the locale data imported from CLDR. In CLDR, set of meta zone IDs and mappings between zone IDs and meta zone
 * IDs are shared by all locales. Therefore, the behavior of {@link #getAvailableMetaZoneIDs},
 * {@link #getMetaZoneID}, and {@link #getReferenceZoneID} won't be changed no matter
 * what locale is used for getting an instance of <code>TimeZoneNames</code>.
 *
 * @stable ICU 50
 */
class U_I18N_API TimeZoneNames : public UObject {
public:
    /**
     * Destructor.
     * @stable ICU 50
     */
    virtual ~TimeZoneNames();

    /**
     * Return true if the given TimeZoneNames objects are semantically equal.
     * @param other the object to be compared with.
     * @return Return TRUE if the given Format objects are semantically equal.
     * @stable ICU 50
     */
    virtual UBool operator==(const TimeZoneNames& other) const = 0;

    /**
     * Return true if the given TimeZoneNames objects are not semantically
     * equal.
     * @param other the object to be compared with.
     * @return Return TRUE if the given Format objects are not semantically equal.
     * @stable ICU 50
     */
    UBool operator!=(const TimeZoneNames& other) const { return !operator==(other); }

    /**
     * Clone this object polymorphically.  The caller is responsible
     * for deleting the result when done.
     * @return A copy of the object
     * @stable ICU 50
     */
    virtual TimeZoneNames* clone() const = 0;

    /**
     * Returns an instance of <code>TimeZoneNames</code> for the specified locale.
     *
     * @param locale The locale.
     * @param status Receives the status.
     * @return An instance of <code>TimeZoneNames</code>
     * @stable ICU 50
     */
    static TimeZoneNames* U_EXPORT2 createInstance(const Locale& locale, UErrorCode& status);

    /**
     * Returns an instance of <code>TimeZoneNames</code> containing only short specific
     * zone names (SHORT_STANDARD and SHORT_DAYLIGHT),
     * compatible with the IANA tz database's zone abbreviations (not localized).
     * <br>
     * Note: The input locale is used for resolving ambiguous names (e.g. "IST" is parsed
     * as Israel Standard Time for Israel, while it is parsed as India Standard Time for
     * all other regions). The zone names returned by this instance are not localized.
     * @stable ICU 54
     */
     static TimeZoneNames* U_EXPORT2 createTZDBInstance(const Locale& locale, UErrorCode& status);

    /**
     * Returns an enumeration of all available meta zone IDs.
     * @param status Receives the status.
     * @return an enumeration object, owned by the caller.
     * @stable ICU 50
     */
    virtual StringEnumeration* getAvailableMetaZoneIDs(UErrorCode& status) const = 0;

    /**
     * Returns an enumeration of all available meta zone IDs used by the given time zone.
     * @param tzID The canoical tiem zone ID.
     * @param status Receives the status.
     * @return an enumeration object, owned by the caller.
     * @stable ICU 50
     */
    virtual StringEnumeration* getAvailableMetaZoneIDs(const UnicodeString& tzID, UErrorCode& status) const = 0;

    /**
     * Returns the meta zone ID for the given canonical time zone ID at the given date.
     * @param tzID The canonical time zone ID.
     * @param date The date.
     * @param mzID Receives the meta zone ID for the given time zone ID at the given date. If the time zone does not have a
     *          corresponding meta zone at the given date or the implementation does not support meta zones, "bogus" state
     *          is set.
     * @return A reference to the result.
     * @stable ICU 50
     */
    virtual UnicodeString& getMetaZoneID(const UnicodeString& tzID, UDate date, UnicodeString& mzID) const = 0;

    /**
     * Returns the reference zone ID for the given meta zone ID for the region.
     *
     * Note: Each meta zone must have a reference zone associated with a special region "001" (world).
     * Some meta zones may have region specific reference zone IDs other than the special region
     * "001". When a meta zone does not have any region specific reference zone IDs, this method
     * return the reference zone ID for the special region "001" (world).
     *
     * @param mzID The meta zone ID.
     * @param region The region.
     * @param tzID Receives the reference zone ID ("golden zone" in the LDML specification) for the given time zone ID for the
     *          region. If the meta zone is unknown or the implementation does not support meta zones, "bogus" state
     *          is set.
     * @return A reference to the result.
     * @stable ICU 50
     */
    virtual UnicodeString& getReferenceZoneID(const UnicodeString& mzID, const char* region, UnicodeString& tzID) const = 0;

    /**
     * Returns the display name of the meta zone.
     * @param mzID The meta zone ID.
     * @param type The display name type. See {@link #UTimeZoneNameType}.
     * @param name Receives the display name of the meta zone. When this object does not have a localized display name for the given
     *         meta zone with the specified type or the implementation does not provide any display names associated
     *         with meta zones, "bogus" state is set.
     * @return A reference to the result.
     * @stable ICU 50
     */
    virtual UnicodeString& getMetaZoneDisplayName(const UnicodeString& mzID, UTimeZoneNameType type, UnicodeString& name) const = 0;

    /**
     * Returns the display name of the time zone. Unlike {@link #getDisplayName},
     * this method does not get a name from a meta zone used by the time zone.
     * @param tzID The canonical time zone ID.
     * @param type The display name type. See {@link #UTimeZoneNameType}.
     * @param name Receives the display name for the time zone. When this object does not have a localized display name for the given
     *         time zone with the specified type, "bogus" state is set.
     * @return A reference to the result.
     * @stable ICU 50
     */
    virtual UnicodeString& getTimeZoneDisplayName(const UnicodeString& tzID, UTimeZoneNameType type, UnicodeString& name) const = 0;

    /**
     * Returns the exemplar location name for the given time zone. When this object does not have a localized location
     * name, the default implementation may still returns a programmatically generated name with the logic described
     * below.
     * <ol>
     * <li>Check if the ID contains "/". If not, return null.
     * <li>Check if the ID does not start with "Etc/" or "SystemV/". If it does, return null.
     * <li>Extract a substring after the last occurrence of "/".
     * <li>Replace "_" with " ".
     * </ol>
     * For example, "New York" is returned for the time zone ID "America/New_York" when this object does not have the
     * localized location name.
     *
     * @param tzID The canonical time zone ID
     * @param name Receives the exemplar location name for the given time zone, or "bogus" state is set when a localized
     *          location name is not available and the fallback logic described above cannot extract location from the ID.
     * @return A reference to the result.
     * @stable ICU 50
     */
    virtual UnicodeString& getExemplarLocationName(const UnicodeString& tzID, UnicodeString& name) const;

    /**
     * Returns the display name of the time zone at the given date.
     * <p>
     * <b>Note:</b> This method calls the subclass's {@link #getTimeZoneDisplayName} first. When the
     * result is bogus, this method calls {@link #getMetaZoneID} to get the meta zone ID mapped from the
     * time zone, then calls {@link #getMetaZoneDisplayName}.
     *
     * @param tzID The canonical time zone ID.
     * @param type The display name type. See {@link #UTimeZoneNameType}.
     * @param date The date.
     * @param name Receives the display name for the time zone at the given date. When this object does not have a localized display
     *          name for the time zone with the specified type and date, "bogus" state is set.
     * @return A reference to the result.
     * @stable ICU 50
     */
    virtual UnicodeString& getDisplayName(const UnicodeString& tzID, UTimeZoneNameType type, UDate date, UnicodeString& name) const;

    /**
     * @internal ICU internal only, for specific users only until proposed publicly.
     */
    virtual void loadAllDisplayNames(UErrorCode& status);

    /**
     * @internal ICU internal only, for specific users only until proposed publicly.
     */
    virtual void getDisplayNames(const UnicodeString& tzID, const UTimeZoneNameType types[], int32_t numTypes, UDate date, UnicodeString dest[], UErrorCode& status) const;

    /**
     * <code>MatchInfoCollection</code> represents a collection of time zone name matches used by
     * {@link TimeZoneNames#find}.
     * @internal
     */
    class U_I18N_API MatchInfoCollection : public UMemory {
    public:
        /**
         * Constructor.
         * @internal
         */
        MatchInfoCollection();
        /**
         * Destructor.
         * @internal
         */
        virtual ~MatchInfoCollection();

#ifndef U_HIDE_INTERNAL_API
        /**
         * Adds a zone match.
         * @param nameType The name type.
         * @param matchLength The match length.
         * @param tzID The time zone ID.
         * @param status Receives the status
         * @internal
         */
        void addZone(UTimeZoneNameType nameType, int32_t matchLength,
            const UnicodeString& tzID, UErrorCode& status);

        /**
         * Adds a meata zone match.
         * @param nameType The name type.
         * @param matchLength The match length.
         * @param mzID The metazone ID.
         * @param status Receives the status
         * @internal
         */
        void addMetaZone(UTimeZoneNameType nameType, int32_t matchLength,
            const UnicodeString& mzID, UErrorCode& status);

        /**
         * Returns the number of entries available in this object.
         * @return The number of entries.
         * @internal
         */
        int32_t size() const;

        /**
         * Returns the time zone name type of a match at the specified index.
         * @param idx The index
         * @return The time zone name type. If the specified idx is out of range,
         *      it returns UTZNM_UNKNOWN.
         * @see UTimeZoneNameType
         * @internal
         */
        UTimeZoneNameType getNameTypeAt(int32_t idx) const;

        /**
         * Returns the match length of a match at the specified index.
         * @param idx The index
         * @return The match length. If the specified idx is out of range,
         *      it returns 0.
         * @internal
         */
        int32_t getMatchLengthAt(int32_t idx) const;

        /**
         * Gets the zone ID of a match at the specified index.
         * @param idx The index
         * @param tzID Receives the zone ID.
         * @return TRUE if the zone ID was set to tzID.
         * @internal
         */
        UBool getTimeZoneIDAt(int32_t idx, UnicodeString& tzID) const;

        /**
         * Gets the metazone ID of a match at the specified index.
         * @param idx The index
         * @param mzID Receives the metazone ID
         * @return TRUE if the meta zone ID was set to mzID.
         * @internal
         */
        UBool getMetaZoneIDAt(int32_t idx, UnicodeString& mzID) const;
#endif  /* U_HIDE_INTERNAL_API */

    private:
        UVector* fMatches;  // vector of MatchEntry

        UVector* matches(UErrorCode& status);
    };

    /**
     * Finds time zone name prefix matches for the input text at the
     * given offset and returns a collection of the matches.
     * @param text The text.
     * @param start The starting offset within the text.
     * @param types The set of name types represented by bitwise flags of UTimeZoneNameType enums,
     *              or UTZNM_UNKNOWN for all name types.
     * @param status Receives the status.
     * @return A collection of matches (owned by the caller), or NULL if no matches are found.
     * @see UTimeZoneNameType
     * @see MatchInfoCollection
     * @internal
     */
    virtual MatchInfoCollection* find(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const = 0;
};

U_NAMESPACE_END

#endif

#endif /* U_SHOW_CPLUSPLUS_API */



U_CDECL_BEGIN
/**
 * Constants for time zone display format style used by format/parse APIs
 * in TimeZoneFormat.
 * @stable ICU 50
 */
typedef enum UTimeZoneFormatStyle {
    /**
     * Generic location format, such as "United States Time (New York)", "Italy Time"
     * @stable ICU 50
     */
    UTZFMT_STYLE_GENERIC_LOCATION,
    /**
     * Generic long non-location format, such as "Eastern Time".
     * @stable ICU 50
     */
    UTZFMT_STYLE_GENERIC_LONG,
    /**
     * Generic short non-location format, such as "ET".
     * @stable ICU 50
     */
    UTZFMT_STYLE_GENERIC_SHORT,
    /**
     * Specific long format, such as "Eastern Standard Time".
     * @stable ICU 50
     */
    UTZFMT_STYLE_SPECIFIC_LONG,
    /**
     * Specific short format, such as "EST", "PDT".
     * @stable ICU 50
     */
    UTZFMT_STYLE_SPECIFIC_SHORT,
    /**
     * Localized GMT offset format, such as "GMT-05:00", "UTC+0100"
     * @stable ICU 50
     */
    UTZFMT_STYLE_LOCALIZED_GMT,
    /**
     * Short localized GMT offset format, such as "GMT-5", "UTC+1:30"
     * This style is equivalent to the LDML date format pattern "O".
     * @stable ICU 51
     */
    UTZFMT_STYLE_LOCALIZED_GMT_SHORT,
    /**
     * Short ISO 8601 local time difference (basic format) or the UTC indicator.
     * For example, "-05", "+0530", and "Z"(UTC).
     * This style is equivalent to the LDML date format pattern "X".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ISO_BASIC_SHORT,
    /**
     * Short ISO 8601 locale time difference (basic format).
     * For example, "-05" and "+0530".
     * This style is equivalent to the LDML date format pattern "x".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ISO_BASIC_LOCAL_SHORT,
    /**
     * Fixed width ISO 8601 local time difference (basic format) or the UTC indicator.
     * For example, "-0500", "+0530", and "Z"(UTC).
     * This style is equivalent to the LDML date format pattern "XX".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ISO_BASIC_FIXED,
    /**
     * Fixed width ISO 8601 local time difference (basic format).
     * For example, "-0500" and "+0530".
     * This style is equivalent to the LDML date format pattern "xx".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ISO_BASIC_LOCAL_FIXED,
    /**
     * ISO 8601 local time difference (basic format) with optional seconds field, or the UTC indicator.
     * For example, "-0500", "+052538", and "Z"(UTC).
     * This style is equivalent to the LDML date format pattern "XXXX".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ISO_BASIC_FULL,
    /**
     * ISO 8601 local time difference (basic format) with optional seconds field.
     * For example, "-0500" and "+052538".
     * This style is equivalent to the LDML date format pattern "xxxx".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ISO_BASIC_LOCAL_FULL,
    /**
     * Fixed width ISO 8601 local time difference (extended format) or the UTC indicator.
     * For example, "-05:00", "+05:30", and "Z"(UTC).
     * This style is equivalent to the LDML date format pattern "XXX".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ISO_EXTENDED_FIXED,
    /**
     * Fixed width ISO 8601 local time difference (extended format).
     * For example, "-05:00" and "+05:30".
     * This style is equivalent to the LDML date format pattern "xxx" and "ZZZZZ".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FIXED,
    /**
     * ISO 8601 local time difference (extended format) with optional seconds field, or the UTC indicator.
     * For example, "-05:00", "+05:25:38", and "Z"(UTC).
     * This style is equivalent to the LDML date format pattern "XXXXX".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ISO_EXTENDED_FULL,
    /**
     * ISO 8601 local time difference (extended format) with optional seconds field.
     * For example, "-05:00" and "+05:25:38".
     * This style is equivalent to the LDML date format pattern "xxxxx".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FULL,
    /**
     * Time Zone ID, such as "America/Los_Angeles".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ZONE_ID,
    /**
     * Short Time Zone ID (BCP 47 Unicode location extension, time zone type value), such as "uslax".
     * @stable ICU 51
     */
    UTZFMT_STYLE_ZONE_ID_SHORT,
    /**
     * Exemplar location, such as "Los Angeles" and "Paris".
     * @stable ICU 51
     */
    UTZFMT_STYLE_EXEMPLAR_LOCATION
} UTimeZoneFormatStyle;

/**
 * Constants for GMT offset pattern types.
 * @stable ICU 50
 */
typedef enum UTimeZoneFormatGMTOffsetPatternType {
    /**
     * Positive offset with hours and minutes fields
     * @stable ICU 50
     */
    UTZFMT_PAT_POSITIVE_HM,
    /**
     * Positive offset with hours, minutes and seconds fields
     * @stable ICU 50
     */
    UTZFMT_PAT_POSITIVE_HMS,
    /**
     * Negative offset with hours and minutes fields
     * @stable ICU 50
     */
    UTZFMT_PAT_NEGATIVE_HM,
    /**
     * Negative offset with hours, minutes and seconds fields
     * @stable ICU 50
     */
    UTZFMT_PAT_NEGATIVE_HMS,
    /**
     * Positive offset with hours field
     * @stable ICU 51
     */
    UTZFMT_PAT_POSITIVE_H,
    /**
     * Negative offset with hours field
     * @stable ICU 51
     */
    UTZFMT_PAT_NEGATIVE_H,

    /* The following cannot be #ifndef U_HIDE_INTERNAL_API, needed for other .h declarations */
    /**
     * Number of UTimeZoneFormatGMTOffsetPatternType types.
     * @internal
     */
    UTZFMT_PAT_COUNT = 6
} UTimeZoneFormatGMTOffsetPatternType;

/**
 * Constants for time types used by TimeZoneFormat APIs for
 * receiving time type (standard time, daylight time or unknown).
 * @stable ICU 50
 */
typedef enum UTimeZoneFormatTimeType {
    /**
     * Unknown
     * @stable ICU 50
     */
    UTZFMT_TIME_TYPE_UNKNOWN,
    /**
     * Standard time
     * @stable ICU 50
     */
    UTZFMT_TIME_TYPE_STANDARD,
    /**
     * Daylight saving time
     * @stable ICU 50
     */
    UTZFMT_TIME_TYPE_DAYLIGHT
} UTimeZoneFormatTimeType;

/**
 * Constants for parse option flags, used for specifying optional parse behavior.
 * @stable ICU 50
 */
typedef enum UTimeZoneFormatParseOption {
    /**
     * No option.
     * @stable ICU 50
     */
    UTZFMT_PARSE_OPTION_NONE        = 0x00,
    /**
     * When a time zone display name is not found within a set of display names
     * used for the specified style, look for the name from display names used
     * by other styles.
     * @stable ICU 50
     */
    UTZFMT_PARSE_OPTION_ALL_STYLES  = 0x01,
     /**
      * When parsing a time zone display name in \link UTZFMT_STYLE_SPECIFIC_SHORT \endlink,
      * look for the IANA tz database compatible zone abbreviations in addition
      * to the localized names coming from the icu::TimeZoneNames currently
      * used by the icu::TimeZoneFormat.
      * @stable ICU 54
      */
    UTZFMT_PARSE_OPTION_TZ_DATABASE_ABBREVIATIONS = 0x02
} UTimeZoneFormatParseOption;

U_CDECL_END

U_NAMESPACE_BEGIN

class TimeZoneGenericNames;
class TZDBTimeZoneNames;
class UVector;

/**
 * <code>TimeZoneFormat</code> supports time zone display name formatting and parsing.
 * An instance of TimeZoneFormat works as a subformatter of {@link SimpleDateFormat},
 * but you can also directly get a new instance of <code>TimeZoneFormat</code> and
 * formatting/parsing time zone display names.
 * <p>
 * ICU implements the time zone display names defined by <a href="http://www.unicode.org/reports/tr35/">UTS#35
 * Unicode Locale Data Markup Language (LDML)</a>. {@link TimeZoneNames} represents the
 * time zone display name data model and this class implements the algorithm for actual
 * formatting and parsing.
 *
 * @see SimpleDateFormat
 * @see TimeZoneNames
 * @stable ICU 50
 */
class U_I18N_API TimeZoneFormat : public Format {
public:
    /**
     * Copy constructor.
     * @stable ICU 50
     */
    TimeZoneFormat(const TimeZoneFormat& other);

    /**
     * Destructor.
     * @stable ICU 50
     */
    virtual ~TimeZoneFormat();

    /**
     * Assignment operator.
     * @stable ICU 50
     */
    TimeZoneFormat& operator=(const TimeZoneFormat& other);

    /**
     * Return true if the given Format objects are semantically equal.
     * Objects of different subclasses are considered unequal.
     * @param other The object to be compared with.
     * @return Return TRUE if the given Format objects are semantically equal.
     *                Objects of different subclasses are considered unequal.
     * @stable ICU 50
     */
    virtual UBool operator==(const Format& other) const;

    /**
     * Clone this object polymorphically. The caller is responsible
     * for deleting the result when done.
     * @return A copy of the object
     * @stable ICU 50
     */
    virtual TimeZoneFormat* clone() const;

    /**
     * Creates an instance of <code>TimeZoneFormat</code> for the given locale.
     * @param locale The locale.
     * @param status Receives the status.
     * @return An instance of <code>TimeZoneFormat</code> for the given locale,
     *          owned by the caller.
     * @stable ICU 50
     */
    static TimeZoneFormat* U_EXPORT2 createInstance(const Locale& locale, UErrorCode& status);

    /**
     * Returns the time zone display name data used by this instance.
     * @return The time zone display name data.
     * @stable ICU 50
     */
    const TimeZoneNames* getTimeZoneNames() const;

    /**
     * Sets the time zone display name data to this format instnace.
     * The caller should not delete the TimeZoenNames object after it is adopted
     * by this call.
     * @param tznames TimeZoneNames object to be adopted.
     * @stable ICU 50
     */
    void adoptTimeZoneNames(TimeZoneNames *tznames);

    /**
     * Sets the time zone display name data to this format instnace.
     * @param tznames TimeZoneNames object to be set.
     * @stable ICU 50
     */
    void setTimeZoneNames(const TimeZoneNames &tznames);

    /**
     * Returns the localized GMT format pattern.
     * @param pattern Receives the localized GMT format pattern.
     * @return A reference to the result pattern.
     * @see #setGMTPattern
     * @stable ICU 50
     */
    UnicodeString& getGMTPattern(UnicodeString& pattern) const;

    /**
     * Sets the localized GMT format pattern. The pattern must contain
     * a single argument {0}, for example "GMT {0}".
     * @param pattern The localized GMT format pattern to be used by this object.
     * @param status Recieves the status.
     * @see #getGMTPattern
     * @stable ICU 50
     */
    void setGMTPattern(const UnicodeString& pattern, UErrorCode& status);

    /**
     * Returns the offset pattern used for localized GMT format.
     * @param type The offset pattern type enum.
     * @param pattern Receives the offset pattern.
     * @return A reference to the result pattern.
     * @see #setGMTOffsetPattern
     * @stable ICU 50
     */
    UnicodeString& getGMTOffsetPattern(UTimeZoneFormatGMTOffsetPatternType type, UnicodeString& pattern) const;

    /**
     * Sets the offset pattern for the given offset type.
     * @param type The offset pattern type enum.
     * @param pattern The offset pattern used for localized GMT format for the type.
     * @param status Receives the status.
     * @see #getGMTOffsetPattern
     * @stable ICU 50
     */
    void setGMTOffsetPattern(UTimeZoneFormatGMTOffsetPatternType type, const UnicodeString& pattern, UErrorCode& status);

    /**
     * Returns the decimal digit characters used for localized GMT format.
     * The return string contains exactly 10 code points (may include Unicode
     * supplementary character) representing digit 0 to digit 9 in the ascending
     * order.
     * @param digits Receives the decimal digits used for localized GMT format.
     * @see #setGMTOffsetDigits
     * @stable ICU 50
     */
    UnicodeString& getGMTOffsetDigits(UnicodeString& digits) const;

    /**
     * Sets the decimal digit characters used for localized GMT format.
     * The input <code>digits</code> must contain exactly 10 code points
     * (Unicode supplementary characters are also allowed) representing
     * digit 0 to digit 9 in the ascending order. When the input <code>digits</code>
     * does not satisfy the condition, <code>U_ILLEGAL_ARGUMENT_ERROR</code>
     * will be set to the return status.
     * @param digits The decimal digits used for localized GMT format.
     * @param status Receives the status.
     * @see #getGMTOffsetDigits
     * @stable ICU 50
     */
    void setGMTOffsetDigits(const UnicodeString& digits, UErrorCode& status);

    /**
     * Returns the localized GMT format string for GMT(UTC) itself (GMT offset is 0).
     * @param gmtZeroFormat Receives the localized GMT string string for GMT(UTC) itself.
     * @return A reference to the result GMT string.
     * @see #setGMTZeroFormat
     * @stable ICU 50
     */
    UnicodeString& getGMTZeroFormat(UnicodeString& gmtZeroFormat) const;

    /**
     * Sets the localized GMT format string for GMT(UTC) itself (GMT offset is 0).
     * @param gmtZeroFormat The localized GMT format string for GMT(UTC).
     * @param status Receives the status.
     * @see #getGMTZeroFormat
     * @stable ICU 50
     */
    void setGMTZeroFormat(const UnicodeString& gmtZeroFormat, UErrorCode& status);

    /**
     * Returns the bitwise flags of UTimeZoneFormatParseOption representing the default parse
     * options used by this object.
     * @return the default parse options.
     * @see ParseOption
     * @stable ICU 50
     */
    uint32_t getDefaultParseOptions(void) const;

    /**
     * Sets the default parse options.
     * <p><b>Note</b>: By default, an instance of <code>TimeZoneFormat</code>
     * created by {@link #createInstance} has no parse options set (UTZFMT_PARSE_OPTION_NONE).
     * To specify multipe options, use bitwise flags of UTimeZoneFormatParseOption.
     * @see #UTimeZoneFormatParseOption
     * @stable ICU 50
     */
    void setDefaultParseOptions(uint32_t flags);

    /**
     * Returns the ISO 8601 basic time zone string for the given offset.
     * For example, "-08", "-0830" and "Z"
     *
     * @param offset the offset from GMT(UTC) in milliseconds.
     * @param useUtcIndicator true if ISO 8601 UTC indicator "Z" is used when the offset is 0.
     * @param isShort true if shortest form is used.
     * @param ignoreSeconds true if non-zero offset seconds is appended.
     * @param result Receives the ISO format string.
     * @param status Receives the status
     * @return the ISO 8601 basic format.
     * @see #formatOffsetISO8601Extended
     * @see #parseOffsetISO8601
     * @stable ICU 51
     */
    UnicodeString& formatOffsetISO8601Basic(int32_t offset, UBool useUtcIndicator, UBool isShort, UBool ignoreSeconds,
        UnicodeString& result, UErrorCode& status) const;

    /**
     * Returns the ISO 8601 extended time zone string for the given offset.
     * For example, "-08:00", "-08:30" and "Z"
     *
     * @param offset the offset from GMT(UTC) in milliseconds.
     * @param useUtcIndicator true if ISO 8601 UTC indicator "Z" is used when the offset is 0.
     * @param isShort true if shortest form is used.
     * @param ignoreSeconds true if non-zero offset seconds is appended.
     * @param result Receives the ISO format string.
     * @param status Receives the status
     * @return the ISO 8601 basic format.
     * @see #formatOffsetISO8601Extended
     * @see #parseOffsetISO8601
     * @stable ICU 51
     */
    UnicodeString& formatOffsetISO8601Extended(int32_t offset, UBool useUtcIndicator, UBool isShort, UBool ignoreSeconds,
        UnicodeString& result, UErrorCode& status) const;

    /**
     * Returns the localized GMT(UTC) offset format for the given offset.
     * The localized GMT offset is defined by;
     * <ul>
     * <li>GMT format pattern (e.g. "GMT {0}" - see {@link #getGMTPattern})
     * <li>Offset time pattern (e.g. "+HH:mm" - see {@link #getGMTOffsetPattern})
     * <li>Offset digits (e.g. "0123456789" - see {@link #getGMTOffsetDigits})
     * <li>GMT zero format (e.g. "GMT" - see {@link #getGMTZeroFormat})
     * </ul>
     * This format always uses 2 digit hours and minutes. When the given offset has non-zero
     * seconds, 2 digit seconds field will be appended. For example,
     * GMT+05:00 and GMT+05:28:06.
     * @param offset the offset from GMT(UTC) in milliseconds.
     * @param status Receives the status
     * @param result Receives the localized GMT format string.
     * @return A reference to the result.
     * @see #parseOffsetLocalizedGMT
     * @stable ICU 50
     */
    UnicodeString& formatOffsetLocalizedGMT(int32_t offset, UnicodeString& result, UErrorCode& status) const;

    /**
     * Returns the short localized GMT(UTC) offset format for the given offset.
     * The short localized GMT offset is defined by;
     * <ul>
     * <li>GMT format pattern (e.g. "GMT {0}" - see {@link #getGMTPattern})
     * <li>Offset time pattern (e.g. "+HH:mm" - see {@link #getGMTOffsetPattern})
     * <li>Offset digits (e.g. "0123456789" - see {@link #getGMTOffsetDigits})
     * <li>GMT zero format (e.g. "GMT" - see {@link #getGMTZeroFormat})
     * </ul>
     * This format uses the shortest representation of offset. The hours field does not
     * have leading zero and lower fields with zero will be truncated. For example,
     * GMT+5 and GMT+530.
     * @param offset the offset from GMT(UTC) in milliseconds.
     * @param status Receives the status
     * @param result Receives the short localized GMT format string.
     * @return A reference to the result.
     * @see #parseOffsetShortLocalizedGMT
     * @stable ICU 51
     */
    UnicodeString& formatOffsetShortLocalizedGMT(int32_t offset, UnicodeString& result, UErrorCode& status) const;

    using Format::format;

    /**
     * Returns the display name of the time zone at the given date for the style.
     * @param style The style (e.g. <code>UTZFMT_STYLE_GENERIC_LONG</code>, <code>UTZFMT_STYLE_LOCALIZED_GMT</code>...)
     * @param tz The time zone.
     * @param date The date.
     * @param name Receives the display name.
     * @param timeType the output argument for receiving the time type (standard/daylight/unknown)
     * used for the display name, or NULL if the information is not necessary.
     * @return A reference to the result
     * @see #UTimeZoneFormatStyle
     * @see #UTimeZoneFormatTimeType
     * @stable ICU 50
     */
    virtual UnicodeString& format(UTimeZoneFormatStyle style, const TimeZone& tz, UDate date,
        UnicodeString& name, UTimeZoneFormatTimeType* timeType = NULL) const;

    /**
     * Returns offset from GMT(UTC) in milliseconds for the given ISO 8601
     * style time zone string. When the given string is not an ISO 8601 time zone
     * string, this method sets the current position as the error index
     * to <code>ParsePosition pos</code> and returns 0.
     * @param text The text contains ISO8601 style time zone string (e.g. "-08:00", "Z")
     *              at the position.
     * @param pos The ParsePosition object.
     * @return The offset from GMT(UTC) in milliseconds for the given ISO 8601 style
     *              time zone string.
     * @see #formatOffsetISO8601Basic
     * @see #formatOffsetISO8601Extended
     * @stable ICU 50
     */
    int32_t parseOffsetISO8601(const UnicodeString& text, ParsePosition& pos) const;

    /**
     * Returns offset from GMT(UTC) in milliseconds for the given localized GMT
     * offset format string. When the given string cannot be parsed, this method
     * sets the current position as the error index to <code>ParsePosition pos</code>
     * and returns 0.
     * @param text The text contains a localized GMT offset string at the position.
     * @param pos The ParsePosition object.
     * @return The offset from GMT(UTC) in milliseconds for the given localized GMT
     *          offset format string.
     * @see #formatOffsetLocalizedGMT
     * @stable ICU 50
     */
    int32_t parseOffsetLocalizedGMT(const UnicodeString& text, ParsePosition& pos) const;

    /**
     * Returns offset from GMT(UTC) in milliseconds for the given short localized GMT
     * offset format string. When the given string cannot be parsed, this method
     * sets the current position as the error index to <code>ParsePosition pos</code>
     * and returns 0.
     * @param text The text contains a short localized GMT offset string at the position.
     * @param pos The ParsePosition object.
     * @return The offset from GMT(UTC) in milliseconds for the given short localized GMT
     *          offset format string.
     * @see #formatOffsetShortLocalizedGMT
     * @stable ICU 51
     */
    int32_t parseOffsetShortLocalizedGMT(const UnicodeString& text, ParsePosition& pos) const;

    /**
     * Returns a <code>TimeZone</code> by parsing the time zone string according to
     * the given parse position, the specified format style and parse options.
     *
     * @param text The text contains a time zone string at the position.
     * @param style The format style
     * @param pos The position.
     * @param parseOptions The parse options repesented by bitwise flags of UTimeZoneFormatParseOption.
     * @param timeType The output argument for receiving the time type (standard/daylight/unknown),
     * or NULL if the information is not necessary.
     * @return A <code>TimeZone</code>, or null if the input could not be parsed.
     * @see UTimeZoneFormatStyle
     * @see UTimeZoneFormatParseOption
     * @see UTimeZoneFormatTimeType
     * @stable ICU 50
     */
    virtual TimeZone* parse(UTimeZoneFormatStyle style, const UnicodeString& text, ParsePosition& pos,
        int32_t parseOptions, UTimeZoneFormatTimeType* timeType = NULL) const;

    /**
     * Returns a <code>TimeZone</code> by parsing the time zone string according to
     * the given parse position, the specified format style and the default parse options.
     *
     * @param text The text contains a time zone string at the position.
     * @param style The format style
     * @param pos The position.
     * @param timeType The output argument for receiving the time type (standard/daylight/unknown),
     * or NULL if the information is not necessary.
     * @return A <code>TimeZone</code>, or null if the input could not be parsed.
     * @see UTimeZoneFormatStyle
     * @see UTimeZoneFormatParseOption
     * @see UTimeZoneFormatTimeType
     * @stable ICU 50
     */
    TimeZone* parse(UTimeZoneFormatStyle style, const UnicodeString& text, ParsePosition& pos,
        UTimeZoneFormatTimeType* timeType = NULL) const;

    /* ----------------------------------------------
     * Format APIs
     * ---------------------------------------------- */

    /**
     * Format an object to produce a time zone display string using localized GMT offset format.
     * This method handles Formattable objects with a <code>TimeZone</code>. If a the Formattable
     * object type is not a <code>TimeZone</code>, then it returns a failing UErrorCode.
     * @param obj The object to format. Must be a <code>TimeZone</code>.
     * @param appendTo Output parameter to receive result. Result is appended to existing contents.
     * @param pos On input: an alignment field, if desired. On output: the offsets of the alignment field.
     * @param status Output param filled with success/failure status.
     * @return Reference to 'appendTo' parameter.
     * @stable ICU 50
     */
    virtual UnicodeString& format(const Formattable& obj, UnicodeString& appendTo,
        FieldPosition& pos, UErrorCode& status) const;

    /**
     * Parse a string to produce an object. This methods handles parsing of
     * time zone display strings into Formattable objects with <code>TimeZone</code>.
     * @param source The string to be parsed into an object.
     * @param result Formattable to be set to the parse result. If parse fails, return contents are undefined.
     * @param parse_pos The position to start parsing at. Upon return this param is set to the position after the
     *                  last character successfully parsed. If the source is not parsed successfully, this param
     *                  will remain unchanged.
     * @return A newly created Formattable* object, or NULL on failure.  The caller owns this and should
     *                 delete it when done.
     * @stable ICU 50
     */
    virtual void parseObject(const UnicodeString& source, Formattable& result, ParsePosition& parse_pos) const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     * @stable ICU 50
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     * @stable ICU 50
     */
    virtual UClassID getDynamicClassID() const;

protected:
    /**
     * Constructs a TimeZoneFormat object for the specified locale.
     * @param locale the locale
     * @param status receives the status.
     * @stable ICU 50
     */
    TimeZoneFormat(const Locale& locale, UErrorCode& status);

private:
    /* Locale of this object */
    Locale fLocale;

    /* Stores the region (could be implicit default) */
    char fTargetRegion[ULOC_COUNTRY_CAPACITY];

    /* TimeZoneNames object used by this formatter */
    TimeZoneNames* fTimeZoneNames;

    /* TimeZoneGenericNames object used by this formatter - lazily instantiated */
    TimeZoneGenericNames* fTimeZoneGenericNames;

    /* Localized GMT format pattern - e.g. "GMT{0}" */
    UnicodeString fGMTPattern;

    /* Array of offset patterns used by Localized GMT format - e.g. "+HH:mm" */
    UnicodeString fGMTOffsetPatterns[UTZFMT_PAT_COUNT];

    /* Localized decimal digits used by Localized GMT format */
    UChar32 fGMTOffsetDigits[10];

    /* Localized GMT zero format - e.g. "GMT" */
    UnicodeString fGMTZeroFormat;

    /* Bit flags representing parse options */
    uint32_t fDefParseOptionFlags;

    /* Constant parts of GMT format pattern, populated from localized GMT format pattern*/
    UnicodeString fGMTPatternPrefix;    /* Substring before {0} */
    UnicodeString fGMTPatternSuffix;    /* Substring after {0} */

    /* Compiled offset patterns generated from fGMTOffsetPatterns[] */
    UVector* fGMTOffsetPatternItems[UTZFMT_PAT_COUNT];

    UBool fAbuttingOffsetHoursAndMinutes;

    /* TZDBTimeZoneNames object used for parsing */
    TZDBTimeZoneNames* fTZDBTimeZoneNames;

    /**
     * Returns the time zone's specific format string.
     * @param tz the time zone
     * @param stdType the name type used for standard time
     * @param dstType the name type used for daylight time
     * @param date the date
     * @param name receives the time zone's specific format name string
     * @param timeType when null, actual time type is set
     * @return a reference to name.
     */
    UnicodeString& formatSpecific(const TimeZone& tz, UTimeZoneNameType stdType, UTimeZoneNameType dstType,
        UDate date, UnicodeString& name, UTimeZoneFormatTimeType *timeType) const;

    /**
     * Returns the time zone's generic format string.
     * @param tz the time zone
     * @param genType the generic name type
     * @param date the date
     * @param name receives the time zone's generic format name string
     * @return a reference to name.
     */
    UnicodeString& formatGeneric(const TimeZone& tz, int32_t genType, UDate date, UnicodeString& name) const;

    /**
     * Lazily create a TimeZoneGenericNames instance
     * @param status receives the status
     * @return the cached TimeZoneGenericNames.
     */
    const TimeZoneGenericNames* getTimeZoneGenericNames(UErrorCode& status) const;

    /**
     * Lazily create a TZDBTimeZoneNames instance
     * @param status receives the status
     * @return the cached TZDBTimeZoneNames.
     */
    const TZDBTimeZoneNames* getTZDBTimeZoneNames(UErrorCode& status) const;

    /**
     * Private method returning the time zone's exemplar location string.
     * This method will never return empty.
     * @param tz the time zone
     * @param name receives the time zone's exemplar location name
     * @return a reference to name.
     */
    UnicodeString& formatExemplarLocation(const TimeZone& tz, UnicodeString& name) const;

    /**
     * Private enum specifying a combination of offset fields
     */
    enum OffsetFields {
        FIELDS_H,
        FIELDS_HM,
        FIELDS_HMS
    };

    /**
     * Parses the localized GMT pattern string and initialize
     * localized gmt pattern fields.
     * @param gmtPattern the localized GMT pattern string such as "GMT {0}"
     * @param status U_ILLEGAL_ARGUMENT_ERROR is set when the specified pattern does not
     *               contain an argument "{0}".
     */
    void initGMTPattern(const UnicodeString& gmtPattern, UErrorCode& status);

    /**
     * Parse the GMT offset pattern into runtime optimized format.
     * @param pattern the offset pattern string
     * @param required the required set of fields, such as FIELDS_HM
     * @param status U_ILLEGAL_ARGUMENT is set when the specified pattern does not contain
     *               pattern letters for the required fields.
     * @return A list of GMTOffsetField objects, or NULL on error.
     */
    static UVector* parseOffsetPattern(const UnicodeString& pattern, OffsetFields required, UErrorCode& status);

    /**
     * Appends seconds field to the offset pattern with hour/minute
     * Note: This code will be obsoleted once we add hour-minute-second pattern data in CLDR.
     * @param offsetHM the offset pattern including hours and minutes fields
     * @param result the output offset pattern including hour, minute and seconds fields
     * @param status receives the status
     * @return a reference to result
     */
    static UnicodeString& expandOffsetPattern(const UnicodeString& offsetHM, UnicodeString& result, UErrorCode& status);

    /**
     * Truncates minutes field to the offset pattern with hour/minute
     * Note: This code will be obsoleted once we add hour pattern data in CLDR.
     * @param offsetHM the offset pattern including hours and minutes fields
     * @param result the output offset pattern including only hours field
     * @param status receives the status
     * @return a reference to result
     */
    static UnicodeString& truncateOffsetPattern(const UnicodeString& offsetHM, UnicodeString& result, UErrorCode& status);

    /**
     * Break input string into UChar32[]. Each array element represents
     * a code point. This method is used for parsing localized digit
     * characters and support characters in Unicode supplemental planes.
     * @param str the string
     * @param codeArray receives the result
     * @param capacity the capacity of codeArray
     * @return TRUE when the specified code array is fully filled with code points
     *         (no under/overflow).
     */
    static UBool toCodePoints(const UnicodeString& str, UChar32* codeArray, int32_t capacity);

    /**
     * Private method supprting all of ISO8601 formats
     * @param offset the offset from GMT(UTC) in milliseconds.
     * @param useUtcIndicator true if ISO 8601 UTC indicator "Z" is used when the offset is 0.
     * @param isShort true if shortest form is used.
     * @param ignoreSeconds true if non-zero offset seconds is appended.
     * @param result Receives the result
     * @param status Receives the status
     * @return the ISO 8601 basic format.
     */
    UnicodeString& formatOffsetISO8601(int32_t offset, UBool isBasic, UBool useUtcIndicator,
        UBool isShort, UBool ignoreSeconds, UnicodeString& result, UErrorCode& status) const;

    /**
     * Private method used for localized GMT formatting.
     * @param offset the zone's UTC offset
     * @param isShort true if the short localized GMT format is desired.
     * @param result receives the localized GMT format string
     * @param status receives the status
     */
    UnicodeString& formatOffsetLocalizedGMT(int32_t offset, UBool isShort, UnicodeString& result, UErrorCode& status) const;

    /**
     * Returns offset from GMT(UTC) in milliseconds for the given ISO 8601 style
     * (extended format) time zone string. When the given string is not an ISO 8601 time
     * zone string, this method sets the current position as the error index
     * to <code>ParsePosition pos</code> and returns 0.
     * @param text the text contains ISO 8601 style time zone string (e.g. "-08:00", "Z")
     *      at the position.
     * @param pos the position, non-negative error index will be set on failure.
     * @param extendedOnly TRUE if parsing the text as ISO 8601 extended offset format (e.g. "-08:00"),
     *      or FALSE to evaluate the text as basic format.
     * @param hasDigitOffset receiving if the parsed zone string contains offset digits.
     * @return the offset from GMT(UTC) in milliseconds for the given ISO 8601 style
     *      time zone string.
     */
    int32_t parseOffsetISO8601(const UnicodeString& text, ParsePosition& pos, UBool extendedOnly,
        UBool* hasDigitOffset = NULL) const;

    /**
     * Appends localized digits to the buffer.
     * This code assumes that the input number is 0 - 59
     * @param buf the target buffer
     * @param n the integer number
     * @param minDigits the minimum digits width
     */
    void appendOffsetDigits(UnicodeString& buf, int32_t n, uint8_t minDigits) const;

    /**
     * Returns offset from GMT(UTC) in milliseconds for the given localized GMT
     * offset format string. When the given string cannot be parsed, this method
     * sets the current position as the error index to <code>ParsePosition pos</code>
     * and returns 0.
     * @param text the text contains a localized GMT offset string at the position.
     * @param pos the position, non-negative error index will be set on failure.
     * @param isShort true if this parser to try the short format first
     * @param hasDigitOffset receiving if the parsed zone string contains offset digits.
     * @return the offset from GMT(UTC) in milliseconds for the given localized GMT
     *      offset format string.
     */
    int32_t parseOffsetLocalizedGMT(const UnicodeString& text, ParsePosition& pos,
        UBool isShort, UBool* hasDigitOffset) const;

    /**
     * Parse localized GMT format generated by the patter used by this formatter, except
     * GMT Zero format.
     * @param text the input text
     * @param start the start index
     * @param isShort true if the short localized format is parsed.
     * @param parsedLen receives the parsed length
     * @return the parsed offset in milliseconds
     */
    int32_t parseOffsetLocalizedGMTPattern(const UnicodeString& text, int32_t start,
        UBool isShort, int32_t& parsedLen) const;

    /**
     * Parses localized GMT offset fields into offset.
     * @param text the input text
     * @param start the start index
     * @param isShort true if this is a short format - currently not used
     * @param parsedLen the parsed length, or 0 on failure.
     * @return the parsed offset in milliseconds.
     */
    int32_t parseOffsetFields(const UnicodeString& text, int32_t start, UBool isShort, int32_t& parsedLen) const;

    /**
     * Parse localized GMT offset fields with the given pattern.
     * @param text the input text
     * @param start the start index
     * @param pattenItems the pattern (already itemized)
     * @param forceSingleHourDigit true if hours field is parsed as a single digit
     * @param hour receives the hour offset field
     * @param min receives the minute offset field
     * @param sec receives the second offset field
     * @return the parsed length
     */
    int32_t parseOffsetFieldsWithPattern(const UnicodeString& text, int32_t start,
        UVector* patternItems, UBool forceSingleHourDigit, int32_t& hour, int32_t& min, int32_t& sec) const;

    /**
     * Parses abutting localized GMT offset fields (such as 0800) into offset.
     * @param text the input text
     * @param start the start index
     * @param parsedLen the parsed length, or 0 on failure
     * @return the parsed offset in milliseconds.
     */
    int32_t parseAbuttingOffsetFields(const UnicodeString& text, int32_t start, int32_t& parsedLen) const;

    /**
     * Parses the input text using the default format patterns (e.g. "UTC{0}").
     * @param text the input text
     * @param start the start index
     * @param parsedLen the parsed length, or 0 on failure
     * @return the parsed offset in milliseconds.
     */
    int32_t parseOffsetDefaultLocalizedGMT(const UnicodeString& text, int start, int32_t& parsedLen) const;

    /**
     * Parses the input GMT offset fields with the default offset pattern.
     * @param text the input text
     * @param start the start index
     * @param separator the separator character, e.g. ':'
     * @param parsedLen the parsed length, or 0 on failure.
     * @return the parsed offset in milliseconds.
     */
    int32_t parseDefaultOffsetFields(const UnicodeString& text, int32_t start, char16_t separator,
        int32_t& parsedLen) const;

    /**
     * Reads an offset field value. This method will stop parsing when
     * 1) number of digits reaches <code>maxDigits</code>
     * 2) just before already parsed number exceeds <code>maxVal</code>
     *
     * @param text the text
     * @param start the start offset
     * @param minDigits the minimum number of required digits
     * @param maxDigits the maximum number of digits
     * @param minVal the minimum value
     * @param maxVal the maximum value
     * @param parsedLen the actual parsed length.
     * @return the integer value parsed
     */
    int32_t parseOffsetFieldWithLocalizedDigits(const UnicodeString& text, int32_t start,
        uint8_t minDigits, uint8_t maxDigits, uint16_t minVal, uint16_t maxVal, int32_t& parsedLen) const;

    /**
     * Reads a single decimal digit, either localized digits used by this object
     * or any Unicode numeric character.
     * @param text the text
     * @param start the start index
     * @param len the actual length read from the text
     * the start index is not a decimal number.
     * @return the integer value of the parsed digit, or -1 on failure.
     */
    int32_t parseSingleLocalizedDigit(const UnicodeString& text, int32_t start, int32_t& len) const;

    /**
     * Formats offset using ASCII digits. The input offset range must be
     * within +/-24 hours (exclusive).
     * @param offset The offset
     * @param sep The field separator character or 0 if not required
     * @param minFields The minimum fields
     * @param maxFields The maximum fields
     * @return The offset string
     */
    static UnicodeString& formatOffsetWithAsciiDigits(int32_t offset, char16_t sep,
        OffsetFields minFields, OffsetFields maxFields, UnicodeString& result);

    /**
     * Parses offset represented by contiguous ASCII digits.
     * <p>
     * Note: This method expects the input position is already at the start of
     * ASCII digits and does not parse sign (+/-).
     * @param text The text contains a sequence of ASCII digits
     * @param pos The parse position
     * @param minFields The minimum Fields to be parsed
     * @param maxFields The maximum Fields to be parsed
     * @param fixedHourWidth true if hours field must be width of 2
     * @return Parsed offset, 0 or positive number.
     */
    static int32_t parseAbuttingAsciiOffsetFields(const UnicodeString& text, ParsePosition& pos,
        OffsetFields minFields, OffsetFields maxFields, UBool fixedHourWidth);

    /**
     * Parses offset represented by ASCII digits and separators.
     * <p>
     * Note: This method expects the input position is already at the start of
     * ASCII digits and does not parse sign (+/-).
     * @param text The text
     * @param pos The parse position
     * @param sep The separator character
     * @param minFields The minimum Fields to be parsed
     * @param maxFields The maximum Fields to be parsed
     * @return Parsed offset, 0 or positive number.
     */
    static int32_t parseAsciiOffsetFields(const UnicodeString& text, ParsePosition& pos, char16_t sep,
        OffsetFields minFields, OffsetFields maxFields);

    /**
     * Unquotes the message format style pattern.
     * @param pattern the pattern
     * @param result receive the unquoted pattern.
     * @return A reference to result.
     */
    static UnicodeString& unquote(const UnicodeString& pattern, UnicodeString& result);

    /**
     * Initialize localized GMT format offset hour/min/sec patterns.
     * This method parses patterns into optimized run-time format.
     * @param status receives the status.
     */
    void initGMTOffsetPatterns(UErrorCode& status);

    /**
     * Check if there are any GMT format offset patterns without
     * any separators between hours field and minutes field and update
     * fAbuttingOffsetHoursAndMinutes field. This method must be called
     * after all patterns are parsed into pattern items.
     */
    void checkAbuttingHoursAndMinutes();

    /**
     * Creates an instance of TimeZone for the given offset
     * @param offset the offset
     * @return A TimeZone with the given offset
     */
    TimeZone* createTimeZoneForOffset(int32_t offset) const;

    /**
     * Returns the time type for the given name type
     * @param nameType the name type
     * @return the time type (unknown/standard/daylight)
     */
    static UTimeZoneFormatTimeType getTimeType(UTimeZoneNameType nameType);

    /**
     * Returns the time zone ID of a match at the specified index within
     * the MatchInfoCollection.
     * @param matches the collection of matches
     * @param idx the index withing matches
     * @param tzID receives the resolved time zone ID
     * @return a reference to tzID.
     */
    UnicodeString& getTimeZoneID(const TimeZoneNames::MatchInfoCollection* matches, int32_t idx, UnicodeString& tzID) const;


    /**
     * Parse a zone ID.
     * @param text the text contains a time zone ID string at the position.
     * @param pos the position
     * @param tzID receives the zone ID
     * @return a reference to tzID
     */
    UnicodeString& parseZoneID(const UnicodeString& text, ParsePosition& pos, UnicodeString& tzID) const;

    /**
     * Parse a short zone ID.
     * @param text the text contains a short time zone ID string at the position.
     * @param pos the position
     * @param tzID receives the short zone ID
     * @return a reference to tzID
     */
    UnicodeString& parseShortZoneID(const UnicodeString& text, ParsePosition& pos, UnicodeString& tzID) const;

    /**
     * Parse an exemplar location string.
     * @param text the text contains an exemplar location string at the position.
     * @param pos the position.
     * @param tzID receives the time zone ID
     * @return a reference to tzID
     */
    UnicodeString& parseExemplarLocation(const UnicodeString& text, ParsePosition& pos, UnicodeString& tzID) const;
};

U_NAMESPACE_END

#endif /* !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */

  /* for UTimeZoneFormatTimeType */
// #include "unicode/brkiter.h"

U_NAMESPACE_BEGIN

class DateFormatSymbols;
class DateFormat;
class MessageFormat;
class FieldPositionHandler;
class TimeZoneFormat;
class SharedNumberFormat;
class SimpleDateFormatMutableNFs;
class DateIntervalFormat;

namespace number {
class LocalizedNumberFormatter;
}

/**
 *
 * SimpleDateFormat is a concrete class for formatting and parsing dates in a
 * language-independent manner. It allows for formatting (millis -> text),
 * parsing (text -> millis), and normalization. Formats/Parses a date or time,
 * which is the standard milliseconds since 24:00 GMT, Jan 1, 1970.
 * <P>
 * Clients are encouraged to create a date-time formatter using DateFormat::getInstance(),
 * getDateInstance(), getDateInstance(), or getDateTimeInstance() rather than
 * explicitly constructing an instance of SimpleDateFormat.  This way, the client
 * is guaranteed to get an appropriate formatting pattern for whatever locale the
 * program is running in.  However, if the client needs something more unusual than
 * the default patterns in the locales, he can construct a SimpleDateFormat directly
 * and give it an appropriate pattern (or use one of the factory methods on DateFormat
 * and modify the pattern after the fact with toPattern() and applyPattern().
 *
 * <p><strong>Date and Time Patterns:</strong></p>
 *
 * <p>Date and time formats are specified by <em>date and time pattern</em> strings.
 * Within date and time pattern strings, all unquoted ASCII letters [A-Za-z] are reserved
 * as pattern letters representing calendar fields. <code>SimpleDateFormat</code> supports
 * the date and time formatting algorithm and pattern letters defined by
 * <a href="http://www.unicode.org/reports/tr35/tr35-dates.html#Date_Field_Symbol_Table">UTS#35
 * Unicode Locale Data Markup Language (LDML)</a> and further documented for ICU in the
 * <a href="https://sites.google.com/site/icuprojectuserguide/formatparse/datetime?pli=1#TOC-Date-Field-Symbol-Table">ICU
 * User Guide</a>. The following pattern letters are currently available (note that the actual
 * values depend on CLDR and may change from the examples shown here):</p>
 *
 * <table border="1">
 *     <tr>
 *         <th>Field</th>
 *         <th style="text-align: center">Sym.</th>
 *         <th style="text-align: center">No.</th>
 *         <th>Example</th>
 *         <th>Description</th>
 *     </tr>
 *     <tr>
 *         <th rowspan="3">era</th>
 *         <td style="text-align: center" rowspan="3">G</td>
 *         <td style="text-align: center">1..3</td>
 *         <td>AD</td>
 *         <td rowspan="3">Era - Replaced with the Era string for the current date. One to three letters for the
 *         abbreviated form, four letters for the long (wide) form, five for the narrow form.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>Anno Domini</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">5</td>
 *         <td>A</td>
 *     </tr>
 *     <tr>
 *         <th rowspan="6">year</th>
 *         <td style="text-align: center">y</td>
 *         <td style="text-align: center">1..n</td>
 *         <td>1996</td>
 *         <td>Year. Normally the length specifies the padding, but for two letters it also specifies the maximum
 *         length. Example:<div align="center">
 *             <center>
 *             <table border="1" cellpadding="2" cellspacing="0">
 *                 <tr>
 *                     <th>Year</th>
 *                     <th style="text-align: right">y</th>
 *                     <th style="text-align: right">yy</th>
 *                     <th style="text-align: right">yyy</th>
 *                     <th style="text-align: right">yyyy</th>
 *                     <th style="text-align: right">yyyyy</th>
 *                 </tr>
 *                 <tr>
 *                     <td>AD 1</td>
 *                     <td style="text-align: right">1</td>
 *                     <td style="text-align: right">01</td>
 *                     <td style="text-align: right">001</td>
 *                     <td style="text-align: right">0001</td>
 *                     <td style="text-align: right">00001</td>
 *                 </tr>
 *                 <tr>
 *                     <td>AD 12</td>
 *                     <td style="text-align: right">12</td>
 *                     <td style="text-align: right">12</td>
 *                     <td style="text-align: right">012</td>
 *                     <td style="text-align: right">0012</td>
 *                     <td style="text-align: right">00012</td>
 *                 </tr>
 *                 <tr>
 *                     <td>AD 123</td>
 *                     <td style="text-align: right">123</td>
 *                     <td style="text-align: right">23</td>
 *                     <td style="text-align: right">123</td>
 *                     <td style="text-align: right">0123</td>
 *                     <td style="text-align: right">00123</td>
 *                 </tr>
 *                 <tr>
 *                     <td>AD 1234</td>
 *                     <td style="text-align: right">1234</td>
 *                     <td style="text-align: right">34</td>
 *                     <td style="text-align: right">1234</td>
 *                     <td style="text-align: right">1234</td>
 *                     <td style="text-align: right">01234</td>
 *                 </tr>
 *                 <tr>
 *                     <td>AD 12345</td>
 *                     <td style="text-align: right">12345</td>
 *                     <td style="text-align: right">45</td>
 *                     <td style="text-align: right">12345</td>
 *                     <td style="text-align: right">12345</td>
 *                     <td style="text-align: right">12345</td>
 *                 </tr>
 *             </table>
 *             </center></div>
 *         </td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">Y</td>
 *         <td style="text-align: center">1..n</td>
 *         <td>1997</td>
 *         <td>Year (in "Week of Year" based calendars). Normally the length specifies the padding,
 *         but for two letters it also specifies the maximum length. This year designation is used in ISO
 *         year-week calendar as defined by ISO 8601, but can be used in non-Gregorian based calendar systems
 *         where week date processing is desired. May not always be the same value as calendar year.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">u</td>
 *         <td style="text-align: center">1..n</td>
 *         <td>4601</td>
 *         <td>Extended year. This is a single number designating the year of this calendar system, encompassing
 *         all supra-year fields. For example, for the Julian calendar system, year numbers are positive, with an
 *         era of BCE or CE. An extended year value for the Julian calendar system assigns positive values to CE
 *         years and negative values to BCE years, with 1 BCE being year 0.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center" rowspan="3">U</td>
 *         <td style="text-align: center">1..3</td>
 *         <td>&#30002;&#23376;</td>
 *         <td rowspan="3">Cyclic year name. Calendars such as the Chinese lunar calendar (and related calendars)
 *         and the Hindu calendars use 60-year cycles of year names. Use one through three letters for the abbreviated
 *         name, four for the full (wide) name, or five for the narrow name (currently the data only provides abbreviated names,
 *         which will be used for all requested name widths). If the calendar does not provide cyclic year name data,
 *         or if the year value to be formatted is out of the range of years for which cyclic name data is provided,
 *         then numeric formatting is used (behaves like 'y').</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>(currently also &#30002;&#23376;)</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">5</td>
 *         <td>(currently also &#30002;&#23376;)</td>
 *     </tr>
 *     <tr>
 *         <th rowspan="6">quarter</th>
 *         <td rowspan="3" style="text-align: center">Q</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>02</td>
 *         <td rowspan="3">Quarter - Use one or two for the numerical quarter, three for the abbreviation, or four for the
 *         full (wide) name (five for the narrow name is not yet supported).</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">3</td>
 *         <td>Q2</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>2nd quarter</td>
 *     </tr>
 *     <tr>
 *         <td rowspan="3" style="text-align: center">q</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>02</td>
 *         <td rowspan="3"><b>Stand-Alone</b> Quarter - Use one or two for the numerical quarter, three for the abbreviation,
 *         or four for the full name (five for the narrow name is not yet supported).</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">3</td>
 *         <td>Q2</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>2nd quarter</td>
 *     </tr>
 *     <tr>
 *         <th rowspan="8">month</th>
 *         <td rowspan="4" style="text-align: center">M</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>09</td>
 *         <td rowspan="4">Month - Use one or two for the numerical month, three for the abbreviation, four for
 *         the full (wide) name, or five for the narrow name. With two ("MM"), the month number is zero-padded
 *         if necessary (e.g. "08")</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">3</td>
 *         <td>Sep</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>September</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">5</td>
 *         <td>S</td>
 *     </tr>
 *     <tr>
 *         <td rowspan="4" style="text-align: center">L</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>09</td>
 *         <td rowspan="4"><b>Stand-Alone</b> Month - Use one or two for the numerical month, three for the abbreviation,
 *         four for the full (wide) name, or 5 for the narrow name. With two ("LL"), the month number is zero-padded if
 *         necessary (e.g. "08")</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">3</td>
 *         <td>Sep</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>September</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">5</td>
 *         <td>S</td>
 *     </tr>
 *     <tr>
 *         <th rowspan="2">week</th>
 *         <td style="text-align: center">w</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>27</td>
 *         <td>Week of Year. Use "w" to show the minimum number of digits, or "ww" to always show two digits
 *         (zero-padding if necessary, e.g. "08").</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">W</td>
 *         <td style="text-align: center">1</td>
 *         <td>3</td>
 *         <td>Week of Month</td>
 *     </tr>
 *     <tr>
 *         <th rowspan="4">day</th>
 *         <td style="text-align: center">d</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>1</td>
 *         <td>Date - Day of the month. Use "d" to show the minimum number of digits, or "dd" to always show
 *         two digits (zero-padding if necessary, e.g. "08").</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">D</td>
 *         <td style="text-align: center">1..3</td>
 *         <td>345</td>
 *         <td>Day of year</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">F</td>
 *         <td style="text-align: center">1</td>
 *         <td>2</td>
 *         <td>Day of Week in Month. The example is for the 2nd Wed in July</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">g</td>
 *         <td style="text-align: center">1..n</td>
 *         <td>2451334</td>
 *         <td>Modified Julian day. This is different from the conventional Julian day number in two regards.
 *         First, it demarcates days at local zone midnight, rather than noon GMT. Second, it is a local number;
 *         that is, it depends on the local time zone. It can be thought of as a single number that encompasses
 *         all the date-related fields.</td>
 *     </tr>
 *     <tr>
 *         <th rowspan="14">week<br>
 *         day</th>
 *         <td rowspan="4" style="text-align: center">E</td>
 *         <td style="text-align: center">1..3</td>
 *         <td>Tue</td>
 *         <td rowspan="4">Day of week - Use one through three letters for the short day, four for the full (wide) name,
 *         five for the narrow name, or six for the short name.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>Tuesday</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">5</td>
 *         <td>T</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">6</td>
 *         <td>Tu</td>
 *     </tr>
 *     <tr>
 *         <td rowspan="5" style="text-align: center">e</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>2</td>
 *         <td rowspan="5">Local day of week. Same as E except adds a numeric value that will depend on the local
 *         starting day of the week, using one or two letters. For this example, Monday is the first day of the week.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">3</td>
 *         <td>Tue</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>Tuesday</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">5</td>
 *         <td>T</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">6</td>
 *         <td>Tu</td>
 *     </tr>
 *     <tr>
 *         <td rowspan="5" style="text-align: center">c</td>
 *         <td style="text-align: center">1</td>
 *         <td>2</td>
 *         <td rowspan="5"><b>Stand-Alone</b> local day of week - Use one letter for the local numeric value (same
 *         as 'e'), three for the short day, four for the full (wide) name, five for the narrow name, or six for
 *         the short name.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">3</td>
 *         <td>Tue</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>Tuesday</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">5</td>
 *         <td>T</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">6</td>
 *         <td>Tu</td>
 *     </tr>
 *     <tr>
 *         <th>period</th>
 *         <td style="text-align: center">a</td>
 *         <td style="text-align: center">1</td>
 *         <td>AM</td>
 *         <td>AM or PM</td>
 *     </tr>
 *     <tr>
 *         <th rowspan="4">hour</th>
 *         <td style="text-align: center">h</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>11</td>
 *         <td>Hour [1-12]. When used in skeleton data or in a skeleton passed in an API for flexible data pattern
 *         generation, it should match the 12-hour-cycle format preferred by the locale (h or K); it should not match
 *         a 24-hour-cycle format (H or k). Use hh for zero padding.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">H</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>13</td>
 *         <td>Hour [0-23]. When used in skeleton data or in a skeleton passed in an API for flexible data pattern
 *         generation, it should match the 24-hour-cycle format preferred by the locale (H or k); it should not match a
 *         12-hour-cycle format (h or K). Use HH for zero padding.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">K</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>0</td>
 *         <td>Hour [0-11]. When used in a skeleton, only matches K or h, see above. Use KK for zero padding.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">k</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>24</td>
 *         <td>Hour [1-24]. When used in a skeleton, only matches k or H, see above. Use kk for zero padding.</td>
 *     </tr>
 *     <tr>
 *         <th>minute</th>
 *         <td style="text-align: center">m</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>59</td>
 *         <td>Minute. Use "m" to show the minimum number of digits, or "mm" to always show two digits
 *         (zero-padding if necessary, e.g. "08").</td>
 *     </tr>
 *     <tr>
 *         <th rowspan="3">second</th>
 *         <td style="text-align: center">s</td>
 *         <td style="text-align: center">1..2</td>
 *         <td>12</td>
 *         <td>Second. Use "s" to show the minimum number of digits, or "ss" to always show two digits
 *         (zero-padding if necessary, e.g. "08").</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">S</td>
 *         <td style="text-align: center">1..n</td>
 *         <td>3450</td>
 *         <td>Fractional Second - truncates (like other time fields) to the count of letters when formatting.
 *         Appends zeros if more than 3 letters specified. Truncates at three significant digits when parsing.
 *         (example shows display using pattern SSSS for seconds value 12.34567)</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">A</td>
 *         <td style="text-align: center">1..n</td>
 *         <td>69540000</td>
 *         <td>Milliseconds in day. This field behaves <i>exactly</i> like a composite of all time-related fields,
 *         not including the zone fields. As such, it also reflects discontinuities of those fields on DST transition
 *         days. On a day of DST onset, it will jump forward. On a day of DST cessation, it will jump backward. This
 *         reflects the fact that is must be combined with the offset field to obtain a unique local time value.</td>
 *     </tr>
 *     <tr>
 *         <th rowspan="23">zone</th>
 *         <td rowspan="2" style="text-align: center">z</td>
 *         <td style="text-align: center">1..3</td>
 *         <td>PDT</td>
 *         <td>The <i>short specific non-location format</i>.
 *         Where that is unavailable, falls back to the <i>short localized GMT format</i> ("O").</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>Pacific Daylight Time</td>
 *         <td>The <i>long specific non-location format</i>.
 *         Where that is unavailable, falls back to the <i>long localized GMT format</i> ("OOOO").</td>
 *     </tr>
 *     <tr>
 *         <td rowspan="3" style="text-align: center">Z</td>
 *         <td style="text-align: center">1..3</td>
 *         <td>-0800</td>
 *         <td>The <i>ISO8601 basic format</i> with hours, minutes and optional seconds fields.
 *         The format is equivalent to RFC 822 zone format (when optional seconds field is absent).
 *         This is equivalent to the "xxxx" specifier.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>GMT-8:00</td>
 *         <td>The <i>long localized GMT format</i>.
 *         This is equivalent to the "OOOO" specifier.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">5</td>
 *         <td>-08:00<br>
 *         -07:52:58</td>
 *         <td>The <i>ISO8601 extended format</i> with hours, minutes and optional seconds fields.
 *         The ISO8601 UTC indicator "Z" is used when local time offset is 0.
 *         This is equivalent to the "XXXXX" specifier.</td>
 *     </tr>
 *     <tr>
 *         <td rowspan="2" style="text-align: center">O</td>
 *         <td style="text-align: center">1</td>
 *         <td>GMT-8</td>
 *         <td>The <i>short localized GMT format</i>.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>GMT-08:00</td>
 *         <td>The <i>long localized GMT format</i>.</td>
 *     </tr>
 *     <tr>
 *         <td rowspan="2" style="text-align: center">v</td>
 *         <td style="text-align: center">1</td>
 *         <td>PT</td>
 *         <td>The <i>short generic non-location format</i>.
 *         Where that is unavailable, falls back to the <i>generic location format</i> ("VVVV"),
 *         then the <i>short localized GMT format</i> as the final fallback.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>Pacific Time</td>
 *         <td>The <i>long generic non-location format</i>.
 *         Where that is unavailable, falls back to <i>generic location format</i> ("VVVV").
 *     </tr>
 *     <tr>
 *         <td rowspan="4" style="text-align: center">V</td>
 *         <td style="text-align: center">1</td>
 *         <td>uslax</td>
 *         <td>The short time zone ID.
 *         Where that is unavailable, the special short time zone ID <i>unk</i> (Unknown Zone) is used.<br>
 *         <i><b>Note</b>: This specifier was originally used for a variant of the short specific non-location format,
 *         but it was deprecated in the later version of the LDML specification. In CLDR 23/ICU 51, the definition of
 *         the specifier was changed to designate a short time zone ID.</i></td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">2</td>
 *         <td>America/Los_Angeles</td>
 *         <td>The long time zone ID.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">3</td>
 *         <td>Los Angeles</td>
 *         <td>The exemplar city (location) for the time zone.
 *         Where that is unavailable, the localized exemplar city name for the special zone <i>Etc/Unknown</i> is used
 *         as the fallback (for example, "Unknown City"). </td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>Los Angeles Time</td>
 *         <td>The <i>generic location format</i>.
 *         Where that is unavailable, falls back to the <i>long localized GMT format</i> ("OOOO";
 *         Note: Fallback is only necessary with a GMT-style Time Zone ID, like Etc/GMT-830.)<br>
 *         This is especially useful when presenting possible timezone choices for user selection,
 *         since the naming is more uniform than the "v" format.</td>
 *     </tr>
 *     <tr>
 *         <td rowspan="5" style="text-align: center">X</td>
 *         <td style="text-align: center">1</td>
 *         <td>-08<br>
 *         +0530<br>
 *         Z</td>
 *         <td>The <i>ISO8601 basic format</i> with hours field and optional minutes field.
 *         The ISO8601 UTC indicator "Z" is used when local time offset is 0.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">2</td>
 *         <td>-0800<br>
 *         Z</td>
 *         <td>The <i>ISO8601 basic format</i> with hours and minutes fields.
 *         The ISO8601 UTC indicator "Z" is used when local time offset is 0.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">3</td>
 *         <td>-08:00<br>
 *         Z</td>
 *         <td>The <i>ISO8601 extended format</i> with hours and minutes fields.
 *         The ISO8601 UTC indicator "Z" is used when local time offset is 0.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>-0800<br>
 *         -075258<br>
 *         Z</td>
 *         <td>The <i>ISO8601 basic format</i> with hours, minutes and optional seconds fields.
 *         (Note: The seconds field is not supported by the ISO8601 specification.)
 *         The ISO8601 UTC indicator "Z" is used when local time offset is 0.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">5</td>
 *         <td>-08:00<br>
 *         -07:52:58<br>
 *         Z</td>
 *         <td>The <i>ISO8601 extended format</i> with hours, minutes and optional seconds fields.
 *         (Note: The seconds field is not supported by the ISO8601 specification.)
 *         The ISO8601 UTC indicator "Z" is used when local time offset is 0.</td>
 *     </tr>
 *     <tr>
 *         <td rowspan="5" style="text-align: center">x</td>
 *         <td style="text-align: center">1</td>
 *         <td>-08<br>
 *         +0530</td>
 *         <td>The <i>ISO8601 basic format</i> with hours field and optional minutes field.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">2</td>
 *         <td>-0800</td>
 *         <td>The <i>ISO8601 basic format</i> with hours and minutes fields.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">3</td>
 *         <td>-08:00</td>
 *         <td>The <i>ISO8601 extended format</i> with hours and minutes fields.</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">4</td>
 *         <td>-0800<br>
 *         -075258</td>
 *         <td>The <i>ISO8601 basic format</i> with hours, minutes and optional seconds fields.
 *         (Note: The seconds field is not supported by the ISO8601 specification.)</td>
 *     </tr>
 *     <tr>
 *         <td style="text-align: center">5</td>
 *         <td>-08:00<br>
 *         -07:52:58</td>
 *         <td>The <i>ISO8601 extended format</i> with hours, minutes and optional seconds fields.
 *         (Note: The seconds field is not supported by the ISO8601 specification.)</td>
 *     </tr>
 * </table>
 *
 * <P>
 * Any characters in the pattern that are not in the ranges of ['a'..'z'] and
 * ['A'..'Z'] will be treated as quoted text. For instance, characters
 * like ':', '.', ' ', '#' and '@' will appear in the resulting time text
 * even they are not embraced within single quotes.
 * <P>
 * A pattern containing any invalid pattern letter will result in a failing
 * UErrorCode result during formatting or parsing.
 * <P>
 * Examples using the US locale:
 * <pre>
 * \code
 *    Format Pattern                         Result
 *    --------------                         -------
 *    "yyyy.MM.dd G 'at' HH:mm:ss vvvv" ->>  1996.07.10 AD at 15:08:56 Pacific Time
 *    "EEE, MMM d, ''yy"                ->>  Wed, July 10, '96
 *    "h:mm a"                          ->>  12:08 PM
 *    "hh 'o''clock' a, zzzz"           ->>  12 o'clock PM, Pacific Daylight Time
 *    "K:mm a, vvv"                     ->>  0:00 PM, PT
 *    "yyyyy.MMMMM.dd GGG hh:mm aaa"    ->>  1996.July.10 AD 12:08 PM
 * \endcode
 * </pre>
 * Code Sample:
 * <pre>
 * \code
 *     UErrorCode success = U_ZERO_ERROR;
 *     SimpleTimeZone* pdt = new SimpleTimeZone(-8 * 60 * 60 * 1000, "PST");
 *     pdt->setStartRule( Calendar::APRIL, 1, Calendar::SUNDAY, 2*60*60*1000);
 *     pdt->setEndRule( Calendar::OCTOBER, -1, Calendar::SUNDAY, 2*60*60*1000);
 *
 *     // Format the current time.
 *     SimpleDateFormat* formatter
 *         = new SimpleDateFormat ("yyyy.MM.dd G 'at' hh:mm:ss a zzz", success );
 *     GregorianCalendar cal(success);
 *     UDate currentTime_1 = cal.getTime(success);
 *     FieldPosition fp(FieldPosition::DONT_CARE);
 *     UnicodeString dateString;
 *     formatter->format( currentTime_1, dateString, fp );
 *     cout << "result: " << dateString << endl;
 *
 *     // Parse the previous string back into a Date.
 *     ParsePosition pp(0);
 *     UDate currentTime_2 = formatter->parse(dateString, pp );
 * \endcode
 * </pre>
 * In the above example, the time value "currentTime_2" obtained from parsing
 * will be equal to currentTime_1. However, they may not be equal if the am/pm
 * marker 'a' is left out from the format pattern while the "hour in am/pm"
 * pattern symbol is used. This information loss can happen when formatting the
 * time in PM.
 *
 * <p>
 * When parsing a date string using the abbreviated year pattern ("y" or "yy"),
 * SimpleDateFormat must interpret the abbreviated year
 * relative to some century.  It does this by adjusting dates to be
 * within 80 years before and 20 years after the time the SimpleDateFormat
 * instance is created. For example, using a pattern of "MM/dd/yy" and a
 * SimpleDateFormat instance created on Jan 1, 1997,  the string
 * "01/11/12" would be interpreted as Jan 11, 2012 while the string "05/04/64"
 * would be interpreted as May 4, 1964.
 * During parsing, only strings consisting of exactly two digits, as defined by
 * <code>Unicode::isDigit()</code>, will be parsed into the default century.
 * Any other numeric string, such as a one digit string, a three or more digit
 * string, or a two digit string that isn't all digits (for example, "-1"), is
 * interpreted literally.  So "01/02/3" or "01/02/003" are parsed (for the
 * Gregorian calendar), using the same pattern, as Jan 2, 3 AD.  Likewise (but
 * only in lenient parse mode, the default) "01/02/-3" is parsed as Jan 2, 4 BC.
 *
 * <p>
 * If the year pattern has more than two 'y' characters, the year is
 * interpreted literally, regardless of the number of digits.  So using the
 * pattern "MM/dd/yyyy", "01/11/12" parses to Jan 11, 12 A.D.
 *
 * <p>
 * When numeric fields abut one another directly, with no intervening delimiter
 * characters, they constitute a run of abutting numeric fields.  Such runs are
 * parsed specially.  For example, the format "HHmmss" parses the input text
 * "123456" to 12:34:56, parses the input text "12345" to 1:23:45, and fails to
 * parse "1234".  In other words, the leftmost field of the run is flexible,
 * while the others keep a fixed width.  If the parse fails anywhere in the run,
 * then the leftmost field is shortened by one character, and the entire run is
 * parsed again. This is repeated until either the parse succeeds or the
 * leftmost field is one character in length.  If the parse still fails at that
 * point, the parse of the run fails.
 *
 * <P>
 * For time zones that have no names, SimpleDateFormat uses strings GMT+hours:minutes or
 * GMT-hours:minutes.
 * <P>
 * The calendar defines what is the first day of the week, the first week of the
 * year, whether hours are zero based or not (0 vs 12 or 24), and the timezone.
 * There is one common number format to handle all the numbers; the digit count
 * is handled programmatically according to the pattern.
 *
 * <p><em>User subclasses are not supported.</em> While clients may write
 * subclasses, such code will not necessarily work and will not be
 * guaranteed to work stably from release to release.
 */
class U_I18N_API SimpleDateFormat: public DateFormat {
public:
    /**
     * Construct a SimpleDateFormat using the default pattern for the default
     * locale.
     * <P>
     * [Note:] Not all locales support SimpleDateFormat; for full generality,
     * use the factory methods in the DateFormat class.
     * @param status    Output param set to success/failure code.
     * @stable ICU 2.0
     */
    SimpleDateFormat(UErrorCode& status);

    /**
     * Construct a SimpleDateFormat using the given pattern and the default locale.
     * The locale is used to obtain the symbols used in formatting (e.g., the
     * names of the months), but not to provide the pattern.
     * <P>
     * [Note:] Not all locales support SimpleDateFormat; for full generality,
     * use the factory methods in the DateFormat class.
     * @param pattern    the pattern for the format.
     * @param status     Output param set to success/failure code.
     * @stable ICU 2.0
     */
    SimpleDateFormat(const UnicodeString& pattern,
                     UErrorCode& status);

    /**
     * Construct a SimpleDateFormat using the given pattern, numbering system override, and the default locale.
     * The locale is used to obtain the symbols used in formatting (e.g., the
     * names of the months), but not to provide the pattern.
     * <P>
     * A numbering system override is a string containing either the name of a known numbering system,
     * or a set of field and numbering system pairs that specify which fields are to be formattied with
     * the alternate numbering system.  For example, to specify that all numeric fields in the specified
     * date or time pattern are to be rendered using Thai digits, simply specify the numbering system override
     * as "thai".  To specify that just the year portion of the date be formatted using Hebrew numbering,
     * use the override string "y=hebrew".  Numbering system overrides can be combined using a semi-colon
     * character in the override string, such as "d=decimal;M=arabic;y=hebrew", etc.
     *
     * <P>
     * [Note:] Not all locales support SimpleDateFormat; for full generality,
     * use the factory methods in the DateFormat class.
     * @param pattern    the pattern for the format.
     * @param override   the override string.
     * @param status     Output param set to success/failure code.
     * @stable ICU 4.2
     */
    SimpleDateFormat(const UnicodeString& pattern,
                     const UnicodeString& override,
                     UErrorCode& status);

    /**
     * Construct a SimpleDateFormat using the given pattern and locale.
     * The locale is used to obtain the symbols used in formatting (e.g., the
     * names of the months), but not to provide the pattern.
     * <P>
     * [Note:] Not all locales support SimpleDateFormat; for full generality,
     * use the factory methods in the DateFormat class.
     * @param pattern    the pattern for the format.
     * @param locale     the given locale.
     * @param status     Output param set to success/failure code.
     * @stable ICU 2.0
     */
    SimpleDateFormat(const UnicodeString& pattern,
                     const Locale& locale,
                     UErrorCode& status);

    /**
     * Construct a SimpleDateFormat using the given pattern, numbering system override, and locale.
     * The locale is used to obtain the symbols used in formatting (e.g., the
     * names of the months), but not to provide the pattern.
     * <P>
     * A numbering system override is a string containing either the name of a known numbering system,
     * or a set of field and numbering system pairs that specify which fields are to be formattied with
     * the alternate numbering system.  For example, to specify that all numeric fields in the specified
     * date or time pattern are to be rendered using Thai digits, simply specify the numbering system override
     * as "thai".  To specify that just the year portion of the date be formatted using Hebrew numbering,
     * use the override string "y=hebrew".  Numbering system overrides can be combined using a semi-colon
     * character in the override string, such as "d=decimal;M=arabic;y=hebrew", etc.
     * <P>
     * [Note:] Not all locales support SimpleDateFormat; for full generality,
     * use the factory methods in the DateFormat class.
     * @param pattern    the pattern for the format.
     * @param override   the numbering system override.
     * @param locale     the given locale.
     * @param status     Output param set to success/failure code.
     * @stable ICU 4.2
     */
    SimpleDateFormat(const UnicodeString& pattern,
                     const UnicodeString& override,
                     const Locale& locale,
                     UErrorCode& status);

    /**
     * Construct a SimpleDateFormat using the given pattern and locale-specific
     * symbol data.  The formatter takes ownership of the DateFormatSymbols object;
     * the caller is no longer responsible for deleting it.
     * @param pattern           the given pattern for the format.
     * @param formatDataToAdopt the symbols to be adopted.
     * @param status            Output param set to success/faulure code.
     * @stable ICU 2.0
     */
    SimpleDateFormat(const UnicodeString& pattern,
                     DateFormatSymbols* formatDataToAdopt,
                     UErrorCode& status);

    /**
     * Construct a SimpleDateFormat using the given pattern and locale-specific
     * symbol data.  The DateFormatSymbols object is NOT adopted; the caller
     * remains responsible for deleting it.
     * @param pattern           the given pattern for the format.
     * @param formatData        the formatting symbols to be use.
     * @param status            Output param set to success/faulure code.
     * @stable ICU 2.0
     */
    SimpleDateFormat(const UnicodeString& pattern,
                     const DateFormatSymbols& formatData,
                     UErrorCode& status);

    /**
     * Copy constructor.
     * @stable ICU 2.0
     */
    SimpleDateFormat(const SimpleDateFormat&);

    /**
     * Assignment operator.
     * @stable ICU 2.0
     */
    SimpleDateFormat& operator=(const SimpleDateFormat&);

    /**
     * Destructor.
     * @stable ICU 2.0
     */
    virtual ~SimpleDateFormat();

    /**
     * Clone this Format object polymorphically. The caller owns the result and
     * should delete it when done.
     * @return    A copy of the object.
     * @stable ICU 2.0
     */
    virtual SimpleDateFormat* clone() const;

    /**
     * Return true if the given Format objects are semantically equal. Objects
     * of different subclasses are considered unequal.
     * @param other    the object to be compared with.
     * @return         true if the given Format objects are semantically equal.
     * @stable ICU 2.0
     */
    virtual UBool operator==(const Format& other) const;


    using DateFormat::format;

    /**
     * Format a date or time, which is the standard millis since 24:00 GMT, Jan
     * 1, 1970. Overrides DateFormat pure virtual method.
     * <P>
     * Example: using the US locale: "yyyy.MM.dd e 'at' HH:mm:ss zzz" ->>
     * 1996.07.10 AD at 15:08:56 PDT
     *
     * @param cal       Calendar set to the date and time to be formatted
     *                  into a date/time string.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       The formatting position. On input: an alignment field,
     *                  if desired. On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.1
     */
    virtual UnicodeString& format(  Calendar& cal,
                                    UnicodeString& appendTo,
                                    FieldPosition& pos) const;

    /**
     * Format a date or time, which is the standard millis since 24:00 GMT, Jan
     * 1, 1970. Overrides DateFormat pure virtual method.
     * <P>
     * Example: using the US locale: "yyyy.MM.dd e 'at' HH:mm:ss zzz" ->>
     * 1996.07.10 AD at 15:08:56 PDT
     *
     * @param cal       Calendar set to the date and time to be formatted
     *                  into a date/time string.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param posIter   On return, can be used to iterate over positions
     *                  of fields generated by this format call.  Field values
     *                  are defined in UDateFormatField.
     * @param status    Input/output param set to success/failure code.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    virtual UnicodeString& format(  Calendar& cal,
                                    UnicodeString& appendTo,
                                    FieldPositionIterator* posIter,
                                    UErrorCode& status) const;

    using DateFormat::parse;

    /**
     * Parse a date/time string beginning at the given parse position. For
     * example, a time text "07/10/96 4:5 PM, PDT" will be parsed into a Date
     * that is equivalent to Date(837039928046).
     * <P>
     * By default, parsing is lenient: If the input is not in the form used by
     * this object's format method but can still be parsed as a date, then the
     * parse succeeds. Clients may insist on strict adherence to the format by
     * calling setLenient(false).
     * @see DateFormat::setLenient(boolean)
     *
     * @param text  The date/time string to be parsed
     * @param cal   A Calendar set on input to the date and time to be used for
     *              missing values in the date/time string being parsed, and set
     *              on output to the parsed date/time. When the calendar type is
     *              different from the internal calendar held by this SimpleDateFormat
     *              instance, the internal calendar will be cloned to a work
     *              calendar set to the same milliseconds and time zone as the
     *              cal parameter, field values will be parsed based on the work
     *              calendar, then the result (milliseconds and time zone) will
     *              be set in this calendar.
     * @param pos   On input, the position at which to start parsing; on
     *              output, the position at which parsing terminated, or the
     *              start position if the parse failed.
     * @stable ICU 2.1
     */
    virtual void parse( const UnicodeString& text,
                        Calendar& cal,
                        ParsePosition& pos) const;


    /**
     * Set the start UDate used to interpret two-digit year strings.
     * When dates are parsed having 2-digit year strings, they are placed within
     * a assumed range of 100 years starting on the two digit start date.  For
     * example, the string "24-Jan-17" may be in the year 1817, 1917, 2017, or
     * some other year.  SimpleDateFormat chooses a year so that the resultant
     * date is on or after the two digit start date and within 100 years of the
     * two digit start date.
     * <P>
     * By default, the two digit start date is set to 80 years before the current
     * time at which a SimpleDateFormat object is created.
     * @param d      start UDate used to interpret two-digit year strings.
     * @param status Filled in with U_ZERO_ERROR if the parse was successful, and with
     *               an error value if there was a parse error.
     * @stable ICU 2.0
     */
    virtual void set2DigitYearStart(UDate d, UErrorCode& status);

    /**
     * Get the start UDate used to interpret two-digit year strings.
     * When dates are parsed having 2-digit year strings, they are placed within
     * a assumed range of 100 years starting on the two digit start date.  For
     * example, the string "24-Jan-17" may be in the year 1817, 1917, 2017, or
     * some other year.  SimpleDateFormat chooses a year so that the resultant
     * date is on or after the two digit start date and within 100 years of the
     * two digit start date.
     * <P>
     * By default, the two digit start date is set to 80 years before the current
     * time at which a SimpleDateFormat object is created.
     * @param status Filled in with U_ZERO_ERROR if the parse was successful, and with
     *               an error value if there was a parse error.
     * @stable ICU 2.0
     */
    UDate get2DigitYearStart(UErrorCode& status) const;

    /**
     * Return a pattern string describing this date format.
     * @param result Output param to receive the pattern.
     * @return       A reference to 'result'.
     * @stable ICU 2.0
     */
    virtual UnicodeString& toPattern(UnicodeString& result) const;

    /**
     * Return a localized pattern string describing this date format.
     * In most cases, this will return the same thing as toPattern(),
     * but a locale can specify characters to use in pattern descriptions
     * in place of the ones described in this class's class documentation.
     * (Presumably, letters that would be more mnemonic in that locale's
     * language.)  This function would produce a pattern using those
     * letters.
     * <p>
     * <b>Note:</b> This implementation depends on DateFormatSymbols::getLocalPatternChars()
     * to get localized format pattern characters. ICU does not include
     * localized pattern character data, therefore, unless user sets localized
     * pattern characters manually, this method returns the same result as
     * toPattern().
     *
     * @param result    Receives the localized pattern.
     * @param status    Output param set to success/failure code on
     *                  exit. If the pattern is invalid, this will be
     *                  set to a failure result.
     * @return          A reference to 'result'.
     * @stable ICU 2.0
     */
    virtual UnicodeString& toLocalizedPattern(UnicodeString& result,
                                              UErrorCode& status) const;

    /**
     * Apply the given unlocalized pattern string to this date format.
     * (i.e., after this call, this formatter will format dates according to
     * the new pattern)
     *
     * @param pattern   The pattern to be applied.
     * @stable ICU 2.0
     */
    virtual void applyPattern(const UnicodeString& pattern);

    /**
     * Apply the given localized pattern string to this date format.
     * (see toLocalizedPattern() for more information on localized patterns.)
     *
     * @param pattern   The localized pattern to be applied.
     * @param status    Output param set to success/failure code on
     *                  exit. If the pattern is invalid, this will be
     *                  set to a failure result.
     * @stable ICU 2.0
     */
    virtual void applyLocalizedPattern(const UnicodeString& pattern,
                                       UErrorCode& status);

    /**
     * Gets the date/time formatting symbols (this is an object carrying
     * the various strings and other symbols used in formatting: e.g., month
     * names and abbreviations, time zone names, AM/PM strings, etc.)
     * @return a copy of the date-time formatting data associated
     * with this date-time formatter.
     * @stable ICU 2.0
     */
    virtual const DateFormatSymbols* getDateFormatSymbols(void) const;

    /**
     * Set the date/time formatting symbols.  The caller no longer owns the
     * DateFormatSymbols object and should not delete it after making this call.
     * @param newFormatSymbols the given date-time formatting symbols to copy.
     * @stable ICU 2.0
     */
    virtual void adoptDateFormatSymbols(DateFormatSymbols* newFormatSymbols);

    /**
     * Set the date/time formatting data.
     * @param newFormatSymbols the given date-time formatting symbols to copy.
     * @stable ICU 2.0
     */
    virtual void setDateFormatSymbols(const DateFormatSymbols& newFormatSymbols);

    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 2.0
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 2.0
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Set the calendar to be used by this date format. Initially, the default
     * calendar for the specified or default locale is used.  The caller should
     * not delete the Calendar object after it is adopted by this call.
     * Adopting a new calendar will change to the default symbols.
     *
     * @param calendarToAdopt    Calendar object to be adopted.
     * @stable ICU 2.0
     */
    virtual void adoptCalendar(Calendar* calendarToAdopt);

    /* Cannot use #ifndef U_HIDE_INTERNAL_API for the following methods since they are virtual */
    /**
     * Sets the TimeZoneFormat to be used by this date/time formatter.
     * The caller should not delete the TimeZoneFormat object after
     * it is adopted by this call.
     * @param timeZoneFormatToAdopt The TimeZoneFormat object to be adopted.
     * @internal ICU 49 technology preview
     */
    virtual void adoptTimeZoneFormat(TimeZoneFormat* timeZoneFormatToAdopt);

    /**
     * Sets the TimeZoneFormat to be used by this date/time formatter.
     * @param newTimeZoneFormat The TimeZoneFormat object to copy.
     * @internal ICU 49 technology preview
     */
    virtual void setTimeZoneFormat(const TimeZoneFormat& newTimeZoneFormat);

    /**
     * Gets the time zone format object associated with this date/time formatter.
     * @return the time zone format associated with this date/time formatter.
     * @internal ICU 49 technology preview
     */
    virtual const TimeZoneFormat* getTimeZoneFormat(void) const;

    /**
     * Set a particular UDisplayContext value in the formatter, such as
     * UDISPCTX_CAPITALIZATION_FOR_STANDALONE. Note: For getContext, see
     * DateFormat.
     * @param value The UDisplayContext value to set.
     * @param status Input/output status. If at entry this indicates a failure
     *               status, the function will do nothing; otherwise this will be
     *               updated with any new status from the function.
     * @stable ICU 53
     */
    virtual void setContext(UDisplayContext value, UErrorCode& status);

    /**
     * Overrides base class method and
     * This method clears per field NumberFormat instances
     * previously set by {@see adoptNumberFormat(const UnicodeString&, NumberFormat*, UErrorCode)}
     * @param formatToAdopt the NumbeferFormat used
     * @stable ICU 54
     */
    void adoptNumberFormat(NumberFormat *formatToAdopt);

    /**
     * Allow the user to set the NumberFormat for several fields
     * It can be a single field like: "y"(year) or "M"(month)
     * It can be several field combined together: "yM"(year and month)
     * Note:
     * 1 symbol field is enough for multiple symbol field (so "y" will override "yy", "yyy")
     * If the field is not numeric, then override has no effect (like "MMM" will use abbreviation, not numerical field)
     * Per field NumberFormat can also be cleared in {@see DateFormat::setNumberFormat(const NumberFormat& newNumberFormat)}
     *
     * @param fields  the fields to override(like y)
     * @param formatToAdopt the NumbeferFormat used
     * @param status  Receives a status code, which will be U_ZERO_ERROR
     *                if the operation succeeds.
     * @stable ICU 54
     */
    void adoptNumberFormat(const UnicodeString& fields, NumberFormat *formatToAdopt, UErrorCode &status);

    /**
     * Get the numbering system to be used for a particular field.
     * @param field The UDateFormatField to get
     * @stable ICU 54
     */
    const NumberFormat * getNumberFormatForField(char16_t field) const;

#ifndef U_HIDE_INTERNAL_API
    /**
     * This is for ICU internal use only. Please do not use.
     * Check whether the 'field' is smaller than all the fields covered in
     * pattern, return TRUE if it is. The sequence of calendar field,
     * from large to small is: ERA, YEAR, MONTH, DATE, AM_PM, HOUR, MINUTE,...
     * @param field    the calendar field need to check against
     * @return         TRUE if the 'field' is smaller than all the fields
     *                 covered in pattern. FALSE otherwise.
     * @internal ICU 4.0
     */
    UBool isFieldUnitIgnored(UCalendarDateFields field) const;


    /**
     * This is for ICU internal use only. Please do not use.
     * Check whether the 'field' is smaller than all the fields covered in
     * pattern, return TRUE if it is. The sequence of calendar field,
     * from large to small is: ERA, YEAR, MONTH, DATE, AM_PM, HOUR, MINUTE,...
     * @param pattern  the pattern to check against
     * @param field    the calendar field need to check against
     * @return         TRUE if the 'field' is smaller than all the fields
     *                 covered in pattern. FALSE otherwise.
     * @internal ICU 4.0
     */
    static UBool isFieldUnitIgnored(const UnicodeString& pattern,
                                    UCalendarDateFields field);

    /**
     * This is for ICU internal use only. Please do not use.
     * Get the locale of this simple date formatter.
     * It is used in DateIntervalFormat.
     *
     * @return   locale in this simple date formatter
     * @internal ICU 4.0
     */
    const Locale& getSmpFmtLocale(void) const;
#endif  /* U_HIDE_INTERNAL_API */

private:
    friend class DateFormat;
    friend class DateIntervalFormat;

    void initializeDefaultCentury(void);

    void initializeBooleanAttributes(void);

    SimpleDateFormat(); // default constructor not implemented

    /**
     * Used by the DateFormat factory methods to construct a SimpleDateFormat.
     * @param timeStyle the time style.
     * @param dateStyle the date style.
     * @param locale    the given locale.
     * @param status    Output param set to success/failure code on
     *                  exit.
     */
    SimpleDateFormat(EStyle timeStyle, EStyle dateStyle, const Locale& locale, UErrorCode& status);

    /**
     * Construct a SimpleDateFormat for the given locale.  If no resource data
     * is available, create an object of last resort, using hard-coded strings.
     * This is an internal method, called by DateFormat.  It should never fail.
     * @param locale    the given locale.
     * @param status    Output param set to success/failure code on
     *                  exit.
     */
    SimpleDateFormat(const Locale& locale, UErrorCode& status); // Use default pattern

    /**
     * Hook called by format(... FieldPosition& ...) and format(...FieldPositionIterator&...)
     */
    UnicodeString& _format(Calendar& cal, UnicodeString& appendTo, FieldPositionHandler& handler, UErrorCode& status) const;

    /**
     * Called by format() to format a single field.
     *
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param ch        The format character we encountered in the pattern.
     * @param count     Number of characters in the current pattern symbol (e.g.,
     *                  "yyyy" in the pattern would result in a call to this function
     *                  with ch equal to 'y' and count equal to 4)
     * @param capitalizationContext Capitalization context for this date format.
     * @param fieldNum  Zero-based numbering of current field within the overall format.
     * @param handler   Records information about field positions.
     * @param cal       Calendar to use
     * @param status    Receives a status code, which will be U_ZERO_ERROR if the operation
     *                  succeeds.
     */
    void subFormat(UnicodeString &appendTo,
                   char16_t ch,
                   int32_t count,
                   UDisplayContext capitalizationContext,
                   int32_t fieldNum,
                   FieldPositionHandler& handler,
                   Calendar& cal,
                   UErrorCode& status) const; // in case of illegal argument

    /**
     * Used by subFormat() to format a numeric value.
     * Appends to toAppendTo a string representation of "value"
     * having a number of digits between "minDigits" and
     * "maxDigits".  Uses the DateFormat's NumberFormat.
     *
     * @param currentNumberFormat
     * @param appendTo  Output parameter to receive result.
     *                  Formatted number is appended to existing contents.
     * @param value     Value to format.
     * @param minDigits Minimum number of digits the result should have
     * @param maxDigits Maximum number of digits the result should have
     */
    void zeroPaddingNumber(const NumberFormat *currentNumberFormat,
                           UnicodeString &appendTo,
                           int32_t value,
                           int32_t minDigits,
                           int32_t maxDigits) const;

    /**
     * Return true if the given format character, occuring count
     * times, represents a numeric field.
     */
    static UBool isNumeric(char16_t formatChar, int32_t count);

    /**
     * Returns TRUE if the patternOffset is at the start of a numeric field.
     */
    static UBool isAtNumericField(const UnicodeString &pattern, int32_t patternOffset);

    /**
     * Returns TRUE if the patternOffset is right after a non-numeric field.
     */
    static UBool isAfterNonNumericField(const UnicodeString &pattern, int32_t patternOffset);

    /**
     * initializes fCalendar from parameters.  Returns fCalendar as a convenience.
     * @param adoptZone  Zone to be adopted, or NULL for TimeZone::createDefault().
     * @param locale Locale of the calendar
     * @param status Error code
     * @return the newly constructed fCalendar
     */
    Calendar *initializeCalendar(TimeZone* adoptZone, const Locale& locale, UErrorCode& status);

    /**
     * Called by several of the constructors to load pattern data and formatting symbols
     * out of a resource bundle and initialize the locale based on it.
     * @param timeStyle     The time style, as passed to DateFormat::createDateInstance().
     * @param dateStyle     The date style, as passed to DateFormat::createTimeInstance().
     * @param locale        The locale to load the patterns from.
     * @param status        Filled in with an error code if loading the data from the
     *                      resources fails.
     */
    void construct(EStyle timeStyle, EStyle dateStyle, const Locale& locale, UErrorCode& status);

    /**
     * Called by construct() and the various constructors to set up the SimpleDateFormat's
     * Calendar and NumberFormat objects.
     * @param locale    The locale for which we want a Calendar and a NumberFormat.
     * @param status    Filled in with an error code if creating either subobject fails.
     */
    void initialize(const Locale& locale, UErrorCode& status);

    /**
     * Private code-size reduction function used by subParse.
     * @param text the time text being parsed.
     * @param start where to start parsing.
     * @param field the date field being parsed.
     * @param stringArray the string array to parsed.
     * @param stringArrayCount the size of the array.
     * @param monthPattern pointer to leap month pattern, or NULL if none.
     * @param cal a Calendar set to the date and time to be formatted
     *            into a date/time string.
     * @return the new start position if matching succeeded; a negative number
     * indicating matching failure, otherwise.
     */
    int32_t matchString(const UnicodeString& text, int32_t start, UCalendarDateFields field,
                        const UnicodeString* stringArray, int32_t stringArrayCount,
                        const UnicodeString* monthPattern, Calendar& cal) const;

    /**
     * Private code-size reduction function used by subParse.
     * @param text the time text being parsed.
     * @param start where to start parsing.
     * @param field the date field being parsed.
     * @param stringArray the string array to parsed.
     * @param stringArrayCount the size of the array.
     * @param cal a Calendar set to the date and time to be formatted
     *            into a date/time string.
     * @return the new start position if matching succeeded; a negative number
     * indicating matching failure, otherwise.
     */
    int32_t matchQuarterString(const UnicodeString& text, int32_t start, UCalendarDateFields field,
                               const UnicodeString* stringArray, int32_t stringArrayCount, Calendar& cal) const;

    /**
     * Used by subParse() to match localized day period strings.
     */
    int32_t matchDayPeriodStrings(const UnicodeString& text, int32_t start,
                                  const UnicodeString* stringArray, int32_t stringArrayCount,
                                  int32_t &dayPeriod) const;

    /**
     * Private function used by subParse to match literal pattern text.
     *
     * @param pattern the pattern string
     * @param patternOffset the starting offset into the pattern text. On
     *        outupt will be set the offset of the first non-literal character in the pattern
     * @param text the text being parsed
     * @param textOffset the starting offset into the text. On output
     *                   will be set to the offset of the character after the match
     * @param whitespaceLenient <code>TRUE</code> if whitespace parse is lenient, <code>FALSE</code> otherwise.
     * @param partialMatchLenient <code>TRUE</code> if partial match parse is lenient, <code>FALSE</code> otherwise.
     * @param oldLeniency <code>TRUE</code> if old leniency control is lenient, <code>FALSE</code> otherwise.
     *
     * @return <code>TRUE</code> if the literal text could be matched, <code>FALSE</code> otherwise.
     */
    static UBool matchLiterals(const UnicodeString &pattern, int32_t &patternOffset,
                               const UnicodeString &text, int32_t &textOffset,
                               UBool whitespaceLenient, UBool partialMatchLenient, UBool oldLeniency);

    /**
     * Private member function that converts the parsed date strings into
     * timeFields. Returns -start (for ParsePosition) if failed.
     * @param text the time text to be parsed.
     * @param start where to start parsing.
     * @param ch the pattern character for the date field text to be parsed.
     * @param count the count of a pattern character.
     * @param obeyCount if true then the count is strictly obeyed.
     * @param allowNegative
     * @param ambiguousYear If true then the two-digit year == the default start year.
     * @param saveHebrewMonth Used to hang onto month until year is known.
     * @param cal a Calendar set to the date and time to be formatted
     *            into a date/time string.
     * @param patLoc
     * @param numericLeapMonthFormatter If non-null, used to parse numeric leap months.
     * @param tzTimeType the type of parsed time zone - standard, daylight or unknown (output).
     *      This parameter can be NULL if caller does not need the information.
     * @return the new start position if matching succeeded; a negative number
     * indicating matching failure, otherwise.
     */
    int32_t subParse(const UnicodeString& text, int32_t& start, char16_t ch, int32_t count,
                     UBool obeyCount, UBool allowNegative, UBool ambiguousYear[], int32_t& saveHebrewMonth, Calendar& cal,
                     int32_t patLoc, MessageFormat * numericLeapMonthFormatter, UTimeZoneFormatTimeType *tzTimeType,
                     int32_t *dayPeriod=NULL) const;

    void parseInt(const UnicodeString& text,
                  Formattable& number,
                  ParsePosition& pos,
                  UBool allowNegative,
                  const NumberFormat *fmt) const;

    void parseInt(const UnicodeString& text,
                  Formattable& number,
                  int32_t maxDigits,
                  ParsePosition& pos,
                  UBool allowNegative,
                  const NumberFormat *fmt) const;

    int32_t checkIntSuffix(const UnicodeString& text, int32_t start,
                           int32_t patLoc, UBool isNegative) const;

    /**
     * Counts number of digit code points in the specified text.
     *
     * @param text  input text
     * @param start start index, inclusive
     * @param end   end index, exclusive
     * @return  number of digits found in the text in the specified range.
    */
    int32_t countDigits(const UnicodeString& text, int32_t start, int32_t end) const;

    /**
     * Translate a pattern, mapping each character in the from string to the
     * corresponding character in the to string. Return an error if the original
     * pattern contains an unmapped character, or if a quote is unmatched.
     * Quoted (single quotes only) material is not translated.
     * @param originalPattern   the original pattern.
     * @param translatedPattern Output param to receive the translited pattern.
     * @param from              the characters to be translited from.
     * @param to                the characters to be translited to.
     * @param status            Receives a status code, which will be U_ZERO_ERROR
     *                          if the operation succeeds.
     */
    static void translatePattern(const UnicodeString& originalPattern,
                                UnicodeString& translatedPattern,
                                const UnicodeString& from,
                                const UnicodeString& to,
                                UErrorCode& status);

    /**
     * Sets the starting date of the 100-year window that dates with 2-digit years
     * are considered to fall within.
     * @param startDate the start date
     * @param status    Receives a status code, which will be U_ZERO_ERROR
     *                  if the operation succeeds.
     */
    void         parseAmbiguousDatesAsAfter(UDate startDate, UErrorCode& status);

    /**
     * Return the length matched by the given affix, or -1 if none.
     * Runs of white space in the affix, match runs of white space in
     * the input.
     * @param affix pattern string, taken as a literal
     * @param input input text
     * @param pos offset into input at which to begin matching
     * @return length of input that matches, or -1 if match failure
     */
    int32_t compareSimpleAffix(const UnicodeString& affix,
                   const UnicodeString& input,
                   int32_t pos) const;

    /**
     * Skip over a run of zero or more Pattern_White_Space characters at
     * pos in text.
     */
    int32_t skipPatternWhiteSpace(const UnicodeString& text, int32_t pos) const;

    /**
     * Skip over a run of zero or more isUWhiteSpace() characters at pos
     * in text.
     */
    int32_t skipUWhiteSpace(const UnicodeString& text, int32_t pos) const;

    /**
     * Initialize LocalizedNumberFormatter instances used for speedup.
     */
    void initFastNumberFormatters(UErrorCode& status);

    /**
     * Delete the LocalizedNumberFormatter instances used for speedup.
     */
    void freeFastNumberFormatters();

    /**
     * Initialize NumberFormat instances used for numbering system overrides.
     */
    void initNumberFormatters(const Locale &locale,UErrorCode &status);

    /**
     * Parse the given override string and set up structures for number formats
     */
    void processOverrideString(const Locale &locale, const UnicodeString &str, int8_t type, UErrorCode &status);

    /**
     * Used to map pattern characters to Calendar field identifiers.
     */
    static const UCalendarDateFields fgPatternIndexToCalendarField[];

    /**
     * Map index into pattern character string to DateFormat field number
     */
    static const UDateFormatField fgPatternIndexToDateFormatField[];

    /**
     * Lazy TimeZoneFormat instantiation, semantically const
     */
    TimeZoneFormat *tzFormat(UErrorCode &status) const;

    const NumberFormat* getNumberFormatByIndex(UDateFormatField index) const;

    /**
     * Used to map Calendar field to field level.
     * The larger the level, the smaller the field unit.
     * For example, UCAL_ERA level is 0, UCAL_YEAR level is 10,
     * UCAL_MONTH level is 20.
     */
    static const int32_t fgCalendarFieldToLevel[];

    /**
     * Map calendar field letter into calendar field level.
     */
    static int32_t getLevelFromChar(char16_t ch);

    /**
     * Tell if a character can be used to define a field in a format string.
     */
    static UBool isSyntaxChar(char16_t ch);

    /**
     * The formatting pattern for this formatter.
     */
    UnicodeString       fPattern;

    /**
     * The numbering system override for dates.
     */
    UnicodeString       fDateOverride;

    /**
     * The numbering system override for times.
     */
    UnicodeString       fTimeOverride;


    /**
     * The original locale used (for reloading symbols)
     */
    Locale              fLocale;

    /**
     * A pointer to an object containing the strings to use in formatting (e.g.,
     * month and day names, AM and PM strings, time zone names, etc.)
     */
    DateFormatSymbols*  fSymbols;   // Owned

    /**
     * The time zone formatter
     */
    TimeZoneFormat* fTimeZoneFormat;

    /**
     * If dates have ambiguous years, we map them into the century starting
     * at defaultCenturyStart, which may be any date.  If defaultCenturyStart is
     * set to SYSTEM_DEFAULT_CENTURY, which it is by default, then the system
     * values are used.  The instance values defaultCenturyStart and
     * defaultCenturyStartYear are only used if explicitly set by the user
     * through the API method parseAmbiguousDatesAsAfter().
     */
    UDate                fDefaultCenturyStart;

    UBool                fHasMinute;
    UBool                fHasSecond;
    UBool                fHasHanYearChar; // pattern contains the Han year character \u5E74

    /**
     * Sets fHasMinutes and fHasSeconds.
     */
    void                 parsePattern();

    /**
     * See documentation for defaultCenturyStart.
     */
    /*transient*/ int32_t   fDefaultCenturyStartYear;

    struct NSOverride : public UMemory {
        const SharedNumberFormat *snf;
        int32_t hash;
        NSOverride *next;
        void free();
        NSOverride() : snf(NULL), hash(0), next(NULL) {
        }
        ~NSOverride();
    };

    /**
     * The number format in use for each date field. NULL means fall back
     * to fNumberFormat in DateFormat.
     */
    const SharedNumberFormat    **fSharedNumberFormatters;

    enum NumberFormatterKey {
        SMPDTFMT_NF_1x10,
        SMPDTFMT_NF_2x10,
        SMPDTFMT_NF_3x10,
        SMPDTFMT_NF_4x10,
        SMPDTFMT_NF_2x2,
        SMPDTFMT_NF_COUNT
    };

    /**
     * Number formatters pre-allocated for fast performance on the most common integer lengths.
     */
    const number::LocalizedNumberFormatter* fFastNumberFormatters[SMPDTFMT_NF_COUNT] = {};

    UBool fHaveDefaultCentury;

    void* fCapitalizationBrkIter;
};

inline UDate
SimpleDateFormat::get2DigitYearStart(UErrorCode& /*status*/) const
{
    return fDefaultCenturyStart;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File DTPTNGEN.H
*
*******************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2007-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  udatpg.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2007jul30
*   created by: Markus W. Scherer
*/






/**
 * \file
 * \brief C API: Wrapper for icu::DateTimePatternGenerator (unicode/dtptngen.h).
 *
 * UDateTimePatternGenerator provides flexible generation of date format patterns, 
 * like "yy-MM-dd". The user can build up the generator by adding successive 
 * patterns. Once that is done, a query can be made using a "skeleton", which is 
 * a pattern which just includes the desired fields and lengths. The generator 
 * will return the "best fit" pattern corresponding to that skeleton.
 * <p>The main method people will use is udatpg_getBestPattern, since normally
 * UDateTimePatternGenerator is pre-built with data from a particular locale. 
 * However, generators can be built directly from other data as well.
 * <p><i>Issue: may be useful to also have a function that returns the list of 
 * fields in a pattern, in order, since we have that internally.
 * That would be useful for getting the UI order of field elements.</i>
 */

/**
 * Opaque type for a date/time pattern generator object.
 * @stable ICU 3.8
 */
typedef void *UDateTimePatternGenerator;

/**
 * Field number constants for udatpg_getAppendItemFormats() and similar functions.
 * These constants are separate from UDateFormatField despite semantic overlap
 * because some fields are merged for the date/time pattern generator.
 * @stable ICU 3.8
 */
typedef enum UDateTimePatternField {
    /** @stable ICU 3.8 */
    UDATPG_ERA_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_YEAR_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_QUARTER_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_MONTH_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_WEEK_OF_YEAR_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_WEEK_OF_MONTH_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_WEEKDAY_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_DAY_OF_YEAR_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_DAY_OF_WEEK_IN_MONTH_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_DAY_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_DAYPERIOD_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_HOUR_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_MINUTE_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_SECOND_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_FRACTIONAL_SECOND_FIELD,
    /** @stable ICU 3.8 */
    UDATPG_ZONE_FIELD,

    /* Do not conditionalize the following with #ifndef U_HIDE_DEPRECATED_API,
     * it is needed for layout of DateTimePatternGenerator object. */
#ifndef U_FORCE_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UDateTimePatternField value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UDATPG_FIELD_COUNT
#endif  // U_FORCE_HIDE_DEPRECATED_API
} UDateTimePatternField;

/**
 * Field display name width constants for udatpg_getFieldDisplayName().
 * @stable ICU 61
 */
typedef enum UDateTimePGDisplayWidth {
    /** @stable ICU 61 */
    UDATPG_WIDE,
    /** @stable ICU 61 */
    UDATPG_ABBREVIATED,
    /** @stable ICU 61 */
    UDATPG_NARROW
} UDateTimePGDisplayWidth;

/**
 * Masks to control forcing the length of specified fields in the returned
 * pattern to match those in the skeleton (when this would not happen
 * otherwise). These may be combined to force the length of multiple fields.
 * Used with udatpg_getBestPatternWithOptions, udatpg_replaceFieldTypesWithOptions.
 * @stable ICU 4.4
 */
typedef enum UDateTimePatternMatchOptions {
    /** @stable ICU 4.4 */
    UDATPG_MATCH_NO_OPTIONS = 0,
    /** @stable ICU 4.4 */
    UDATPG_MATCH_HOUR_FIELD_LENGTH = 1 << UDATPG_HOUR_FIELD,
#ifndef U_HIDE_INTERNAL_API
    /** @internal ICU 4.4 */
    UDATPG_MATCH_MINUTE_FIELD_LENGTH = 1 << UDATPG_MINUTE_FIELD,
    /** @internal ICU 4.4 */
    UDATPG_MATCH_SECOND_FIELD_LENGTH = 1 << UDATPG_SECOND_FIELD,
#endif  /* U_HIDE_INTERNAL_API */
    /** @stable ICU 4.4 */
    UDATPG_MATCH_ALL_FIELDS_LENGTH = (1 << UDATPG_FIELD_COUNT) - 1
} UDateTimePatternMatchOptions;

/**
 * Status return values from udatpg_addPattern().
 * @stable ICU 3.8
 */
typedef enum UDateTimePatternConflict {
    /** @stable ICU 3.8 */
    UDATPG_NO_CONFLICT,
    /** @stable ICU 3.8 */
    UDATPG_BASE_CONFLICT,
    /** @stable ICU 3.8 */
    UDATPG_CONFLICT,
#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UDateTimePatternConflict value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UDATPG_CONFLICT_COUNT
#endif  // U_HIDE_DEPRECATED_API
} UDateTimePatternConflict;

/**
  * Open a generator according to a given locale.
  * @param locale
  * @param pErrorCode a pointer to the UErrorCode which must not indicate a
  *                   failure before the function call.
  * @return a pointer to UDateTimePatternGenerator.
  * @stable ICU 3.8
  */
U_STABLE UDateTimePatternGenerator * U_EXPORT2
udatpg_open(const char *locale, UErrorCode *pErrorCode);

/**
  * Open an empty generator, to be constructed with udatpg_addPattern(...) etc.
  * @param pErrorCode a pointer to the UErrorCode which must not indicate a
  *                   failure before the function call.
  * @return a pointer to UDateTimePatternGenerator.
  * @stable ICU 3.8
  */
U_STABLE UDateTimePatternGenerator * U_EXPORT2
udatpg_openEmpty(UErrorCode *pErrorCode);

/**
  * Close a generator.
  * @param dtpg a pointer to UDateTimePatternGenerator.
  * @stable ICU 3.8
  */
U_STABLE void U_EXPORT2
udatpg_close(UDateTimePatternGenerator *dtpg);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUDateTimePatternGeneratorPointer
 * "Smart pointer" class, closes a UDateTimePatternGenerator via udatpg_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUDateTimePatternGeneratorPointer, UDateTimePatternGenerator, udatpg_close);

U_NAMESPACE_END

#endif

/**
  * Create a copy pf a generator.
  * @param dtpg a pointer to UDateTimePatternGenerator to be copied.
  * @param pErrorCode a pointer to the UErrorCode which must not indicate a
  *                   failure before the function call.
  * @return a pointer to a new UDateTimePatternGenerator.
  * @stable ICU 3.8
 */
U_STABLE UDateTimePatternGenerator * U_EXPORT2
udatpg_clone(const UDateTimePatternGenerator *dtpg, UErrorCode *pErrorCode);

/**
 * Get the best pattern matching the input skeleton. It is guaranteed to
 * have all of the fields in the skeleton.
 * 
 * Note that this function uses a non-const UDateTimePatternGenerator:
 * It uses a stateful pattern parser which is set up for each generator object,
 * rather than creating one for each function call.
 * Consecutive calls to this function do not affect each other,
 * but this function cannot be used concurrently on a single generator object.
 * 
 * @param dtpg a pointer to UDateTimePatternGenerator.
 * @param skeleton
 *            The skeleton is a pattern containing only the variable fields.
 *            For example, "MMMdd" and "mmhh" are skeletons.
 * @param length the length of skeleton
 * @param bestPattern
 *            The best pattern found from the given skeleton.
 * @param capacity the capacity of bestPattern.
 * @param pErrorCode a pointer to the UErrorCode which must not indicate a
 *                   failure before the function call.
 * @return the length of bestPattern.
 * @stable ICU 3.8
 */
U_STABLE int32_t U_EXPORT2
udatpg_getBestPattern(UDateTimePatternGenerator *dtpg,
                      const UChar *skeleton, int32_t length,
                      UChar *bestPattern, int32_t capacity,
                      UErrorCode *pErrorCode);

/**
 * Get the best pattern matching the input skeleton. It is guaranteed to
 * have all of the fields in the skeleton.
 * 
 * Note that this function uses a non-const UDateTimePatternGenerator:
 * It uses a stateful pattern parser which is set up for each generator object,
 * rather than creating one for each function call.
 * Consecutive calls to this function do not affect each other,
 * but this function cannot be used concurrently on a single generator object.
 * 
 * @param dtpg a pointer to UDateTimePatternGenerator.
 * @param skeleton
 *            The skeleton is a pattern containing only the variable fields.
 *            For example, "MMMdd" and "mmhh" are skeletons.
 * @param length the length of skeleton
 * @param options
 *            Options for forcing the length of specified fields in the
 *            returned pattern to match those in the skeleton (when this
 *            would not happen otherwise). For default behavior, use
 *            UDATPG_MATCH_NO_OPTIONS.
 * @param bestPattern
 *            The best pattern found from the given skeleton.
 * @param capacity
 *            the capacity of bestPattern.
 * @param pErrorCode
 *            a pointer to the UErrorCode which must not indicate a
 *            failure before the function call.
 * @return the length of bestPattern.
 * @stable ICU 4.4
 */
U_STABLE int32_t U_EXPORT2
udatpg_getBestPatternWithOptions(UDateTimePatternGenerator *dtpg,
                                 const UChar *skeleton, int32_t length,
                                 UDateTimePatternMatchOptions options,
                                 UChar *bestPattern, int32_t capacity,
                                 UErrorCode *pErrorCode);

/**
  * Get a unique skeleton from a given pattern. For example,
  * both "MMM-dd" and "dd/MMM" produce the skeleton "MMMdd".
  * 
  * Note that this function uses a non-const UDateTimePatternGenerator:
  * It uses a stateful pattern parser which is set up for each generator object,
  * rather than creating one for each function call.
  * Consecutive calls to this function do not affect each other,
  * but this function cannot be used concurrently on a single generator object.
  *
  * @param unusedDtpg     a pointer to UDateTimePatternGenerator.
  *    This parameter is no longer used. Callers may pass NULL.
  * @param pattern  input pattern, such as "dd/MMM".
  * @param length   the length of pattern.
  * @param skeleton such as "MMMdd"
  * @param capacity the capacity of skeleton.
  * @param pErrorCode a pointer to the UErrorCode which must not indicate a
  *                  failure before the function call.
  * @return the length of skeleton.
  * @stable ICU 3.8
  */
U_STABLE int32_t U_EXPORT2
udatpg_getSkeleton(UDateTimePatternGenerator *unusedDtpg,
                   const UChar *pattern, int32_t length,
                   UChar *skeleton, int32_t capacity,
                   UErrorCode *pErrorCode);

/**
 * Get a unique base skeleton from a given pattern. This is the same
 * as the skeleton, except that differences in length are minimized so
 * as to only preserve the difference between string and numeric form. So
 * for example, both "MMM-dd" and "d/MMM" produce the skeleton "MMMd"
 * (notice the single d).
 *
 * Note that this function uses a non-const UDateTimePatternGenerator:
 * It uses a stateful pattern parser which is set up for each generator object,
 * rather than creating one for each function call.
 * Consecutive calls to this function do not affect each other,
 * but this function cannot be used concurrently on a single generator object.
 *
 * @param unusedDtpg     a pointer to UDateTimePatternGenerator.
 *    This parameter is no longer used. Callers may pass NULL.
 * @param pattern  input pattern, such as "dd/MMM".
 * @param length   the length of pattern.
 * @param baseSkeleton such as "Md"
 * @param capacity the capacity of base skeleton.
 * @param pErrorCode a pointer to the UErrorCode which must not indicate a
 *                  failure before the function call.
 * @return the length of baseSkeleton.
 * @stable ICU 3.8
 */
U_STABLE int32_t U_EXPORT2
udatpg_getBaseSkeleton(UDateTimePatternGenerator *unusedDtpg,
                       const UChar *pattern, int32_t length,
                       UChar *baseSkeleton, int32_t capacity,
                       UErrorCode *pErrorCode);

/**
 * Adds a pattern to the generator. If the pattern has the same skeleton as
 * an existing pattern, and the override parameter is set, then the previous
 * value is overriden. Otherwise, the previous value is retained. In either
 * case, the conflicting status is set and previous vale is stored in 
 * conflicting pattern.
 * <p>
 * Note that single-field patterns (like "MMM") are automatically added, and
 * don't need to be added explicitly!
 *
 * @param dtpg     a pointer to UDateTimePatternGenerator.
 * @param pattern  input pattern, such as "dd/MMM"
 * @param patternLength the length of pattern.
 * @param override  When existing values are to be overridden use true, 
 *                  otherwise use false.
 * @param conflictingPattern  Previous pattern with the same skeleton.
 * @param capacity the capacity of conflictingPattern.
 * @param pLength a pointer to the length of conflictingPattern.
 * @param pErrorCode a pointer to the UErrorCode which must not indicate a
 *                  failure before the function call.
 * @return conflicting status. The value could be UDATPG_NO_CONFLICT, 
 *                  UDATPG_BASE_CONFLICT or UDATPG_CONFLICT.
 * @stable ICU 3.8
 */
U_STABLE UDateTimePatternConflict U_EXPORT2
udatpg_addPattern(UDateTimePatternGenerator *dtpg,
                  const UChar *pattern, int32_t patternLength,
                  UBool override,
                  UChar *conflictingPattern, int32_t capacity, int32_t *pLength,
                  UErrorCode *pErrorCode);

/**
  * An AppendItem format is a pattern used to append a field if there is no
  * good match. For example, suppose that the input skeleton is "GyyyyMMMd",
  * and there is no matching pattern internally, but there is a pattern
  * matching "yyyyMMMd", say "d-MM-yyyy". Then that pattern is used, plus the
  * G. The way these two are conjoined is by using the AppendItemFormat for G
  * (era). So if that value is, say "{0}, {1}" then the final resulting
  * pattern is "d-MM-yyyy, G".
  * <p>
  * There are actually three available variables: {0} is the pattern so far,
  * {1} is the element we are adding, and {2} is the name of the element.
  * <p>
  * This reflects the way that the CLDR data is organized.
  *
  * @param dtpg   a pointer to UDateTimePatternGenerator.
  * @param field  UDateTimePatternField, such as UDATPG_ERA_FIELD
  * @param value  pattern, such as "{0}, {1}"
  * @param length the length of value.
  * @stable ICU 3.8
  */
U_STABLE void U_EXPORT2
udatpg_setAppendItemFormat(UDateTimePatternGenerator *dtpg,
                           UDateTimePatternField field,
                           const UChar *value, int32_t length);

/**
 * Getter corresponding to setAppendItemFormat. Values below 0 or at or
 * above UDATPG_FIELD_COUNT are illegal arguments.
 *
 * @param dtpg   A pointer to UDateTimePatternGenerator.
 * @param field  UDateTimePatternField, such as UDATPG_ERA_FIELD
 * @param pLength A pointer that will receive the length of appendItemFormat.
 * @return appendItemFormat for field.
 * @stable ICU 3.8
 */
U_STABLE const UChar * U_EXPORT2
udatpg_getAppendItemFormat(const UDateTimePatternGenerator *dtpg,
                           UDateTimePatternField field,
                           int32_t *pLength);

/**
   * Set the name of field, eg "era" in English for ERA. These are only
   * used if the corresponding AppendItemFormat is used, and if it contains a
   * {2} variable.
   * <p>
   * This reflects the way that the CLDR data is organized.
   *
   * @param dtpg   a pointer to UDateTimePatternGenerator.
   * @param field  UDateTimePatternField
   * @param value  name for the field.
   * @param length the length of value.
   * @stable ICU 3.8
   */
U_STABLE void U_EXPORT2
udatpg_setAppendItemName(UDateTimePatternGenerator *dtpg,
                         UDateTimePatternField field,
                         const UChar *value, int32_t length);

/**
 * Getter corresponding to setAppendItemNames. Values below 0 or at or above
 * UDATPG_FIELD_COUNT are illegal arguments. Note: The more general function
 * for getting date/time field display names is udatpg_getFieldDisplayName.
 *
 * @param dtpg   a pointer to UDateTimePatternGenerator.
 * @param field  UDateTimePatternField, such as UDATPG_ERA_FIELD
 * @param pLength A pointer that will receive the length of the name for field.
 * @return name for field
 * @see udatpg_getFieldDisplayName
 * @stable ICU 3.8
 */
U_STABLE const UChar * U_EXPORT2
udatpg_getAppendItemName(const UDateTimePatternGenerator *dtpg,
                         UDateTimePatternField field,
                         int32_t *pLength);

/**
 * The general interface to get a display name for a particular date/time field,
 * in one of several possible display widths.
 *
 * @param dtpg
 *          A pointer to the UDateTimePatternGenerator object with the localized
 *          display names.
 * @param field
 *          The desired UDateTimePatternField, such as UDATPG_ERA_FIELD.
 * @param width
 *          The desired UDateTimePGDisplayWidth, such as UDATPG_ABBREVIATED.
 * @param fieldName
 *          A pointer to a buffer to receive the NULL-terminated display name. If the name
 *          fits into fieldName but cannot be  NULL-terminated (length == capacity) then
 *          the error code is set to U_STRING_NOT_TERMINATED_WARNING. If the name doesn't
 *          fit into fieldName then the error code is set to U_BUFFER_OVERFLOW_ERROR.
 * @param capacity
 *          The size of fieldName (in UChars).
 * @param pErrorCode
 *          A pointer to a UErrorCode to receive any errors
 * @return
 *         The full length of the name; if greater than capacity, fieldName contains a
 *         truncated result.
 * @stable ICU 61
 */
U_STABLE int32_t U_EXPORT2
udatpg_getFieldDisplayName(const UDateTimePatternGenerator *dtpg,
                           UDateTimePatternField field,
                           UDateTimePGDisplayWidth width,
                           UChar *fieldName, int32_t capacity,
                           UErrorCode *pErrorCode);

/**
 * The DateTimeFormat is a message format pattern used to compose date and
 * time patterns. The default pattern in the root locale is "{1} {0}", where
 * {1} will be replaced by the date pattern and {0} will be replaced by the
 * time pattern; however, other locales may specify patterns such as
 * "{1}, {0}" or "{1} 'at' {0}", etc.
 * <p>
 * This is used when the input skeleton contains both date and time fields,
 * but there is not a close match among the added patterns. For example,
 * suppose that this object was created by adding "dd-MMM" and "hh:mm", and
 * its DateTimeFormat is the default "{1} {0}". Then if the input skeleton
 * is "MMMdhmm", there is not an exact match, so the input skeleton is
 * broken up into two components "MMMd" and "hmm". There are close matches
 * for those two skeletons, so the result is put together with this pattern,
 * resulting in "d-MMM h:mm".
 *
 * @param dtpg a pointer to UDateTimePatternGenerator.
 * @param dtFormat
 *            message format pattern, here {1} will be replaced by the date
 *            pattern and {0} will be replaced by the time pattern.
 * @param length the length of dtFormat.
 * @stable ICU 3.8
 */
U_STABLE void U_EXPORT2
udatpg_setDateTimeFormat(const UDateTimePatternGenerator *dtpg,
                         const UChar *dtFormat, int32_t length);

/**
 * Getter corresponding to setDateTimeFormat.
 * @param dtpg   a pointer to UDateTimePatternGenerator.
 * @param pLength A pointer that will receive the length of the format
 * @return dateTimeFormat.
 * @stable ICU 3.8
 */
U_STABLE const UChar * U_EXPORT2
udatpg_getDateTimeFormat(const UDateTimePatternGenerator *dtpg,
                         int32_t *pLength);

/**
 * The decimal value is used in formatting fractions of seconds. If the
 * skeleton contains fractional seconds, then this is used with the
 * fractional seconds. For example, suppose that the input pattern is
 * "hhmmssSSSS", and the best matching pattern internally is "H:mm:ss", and
 * the decimal string is ",". Then the resulting pattern is modified to be
 * "H:mm:ss,SSSS"
 *
 * @param dtpg a pointer to UDateTimePatternGenerator.
 * @param decimal
 * @param length the length of decimal.
 * @stable ICU 3.8
 */
U_STABLE void U_EXPORT2
udatpg_setDecimal(UDateTimePatternGenerator *dtpg,
                  const UChar *decimal, int32_t length);

/**
 * Getter corresponding to setDecimal.
 * 
 * @param dtpg a pointer to UDateTimePatternGenerator.
 * @param pLength A pointer that will receive the length of the decimal string.
 * @return corresponding to the decimal point.
 * @stable ICU 3.8
 */
U_STABLE const UChar * U_EXPORT2
udatpg_getDecimal(const UDateTimePatternGenerator *dtpg,
                  int32_t *pLength);

/**
 * Adjusts the field types (width and subtype) of a pattern to match what is
 * in a skeleton. That is, if you supply a pattern like "d-M H:m", and a
 * skeleton of "MMMMddhhmm", then the input pattern is adjusted to be
 * "dd-MMMM hh:mm". This is used internally to get the best match for the
 * input skeleton, but can also be used externally.
 *
 * Note that this function uses a non-const UDateTimePatternGenerator:
 * It uses a stateful pattern parser which is set up for each generator object,
 * rather than creating one for each function call.
 * Consecutive calls to this function do not affect each other,
 * but this function cannot be used concurrently on a single generator object.
 *
 * @param dtpg a pointer to UDateTimePatternGenerator.
 * @param pattern Input pattern
 * @param patternLength the length of input pattern.
 * @param skeleton
 * @param skeletonLength the length of input skeleton.
 * @param dest  pattern adjusted to match the skeleton fields widths and subtypes.
 * @param destCapacity the capacity of dest.
 * @param pErrorCode a pointer to the UErrorCode which must not indicate a
 *                  failure before the function call.
 * @return the length of dest.
 * @stable ICU 3.8
 */
U_STABLE int32_t U_EXPORT2
udatpg_replaceFieldTypes(UDateTimePatternGenerator *dtpg,
                         const UChar *pattern, int32_t patternLength,
                         const UChar *skeleton, int32_t skeletonLength,
                         UChar *dest, int32_t destCapacity,
                         UErrorCode *pErrorCode);

/**
 * Adjusts the field types (width and subtype) of a pattern to match what is
 * in a skeleton. That is, if you supply a pattern like "d-M H:m", and a
 * skeleton of "MMMMddhhmm", then the input pattern is adjusted to be
 * "dd-MMMM hh:mm". This is used internally to get the best match for the
 * input skeleton, but can also be used externally.
 *
 * Note that this function uses a non-const UDateTimePatternGenerator:
 * It uses a stateful pattern parser which is set up for each generator object,
 * rather than creating one for each function call.
 * Consecutive calls to this function do not affect each other,
 * but this function cannot be used concurrently on a single generator object.
 *
 * @param dtpg a pointer to UDateTimePatternGenerator.
 * @param pattern Input pattern
 * @param patternLength the length of input pattern.
 * @param skeleton
 * @param skeletonLength the length of input skeleton.
 * @param options
 *            Options controlling whether the length of specified fields in the
 *            pattern are adjusted to match those in the skeleton (when this
 *            would not happen otherwise). For default behavior, use
 *            UDATPG_MATCH_NO_OPTIONS.
 * @param dest  pattern adjusted to match the skeleton fields widths and subtypes.
 * @param destCapacity the capacity of dest.
 * @param pErrorCode a pointer to the UErrorCode which must not indicate a
 *                  failure before the function call.
 * @return the length of dest.
 * @stable ICU 4.4
 */
U_STABLE int32_t U_EXPORT2
udatpg_replaceFieldTypesWithOptions(UDateTimePatternGenerator *dtpg,
                                    const UChar *pattern, int32_t patternLength,
                                    const UChar *skeleton, int32_t skeletonLength,
                                    UDateTimePatternMatchOptions options,
                                    UChar *dest, int32_t destCapacity,
                                    UErrorCode *pErrorCode);

/**
 * Return a UEnumeration list of all the skeletons in canonical form.
 * Call udatpg_getPatternForSkeleton() to get the corresponding pattern.
 * 
 * @param dtpg a pointer to UDateTimePatternGenerator.
 * @param pErrorCode a pointer to the UErrorCode which must not indicate a
 *                  failure before the function call
 * @return a UEnumeration list of all the skeletons
 *         The caller must close the object.
 * @stable ICU 3.8
 */
U_STABLE UEnumeration * U_EXPORT2
udatpg_openSkeletons(const UDateTimePatternGenerator *dtpg, UErrorCode *pErrorCode);

/**
 * Return a UEnumeration list of all the base skeletons in canonical form.
 *
 * @param dtpg a pointer to UDateTimePatternGenerator.
 * @param pErrorCode a pointer to the UErrorCode which must not indicate a
 *             failure before the function call.
 * @return a UEnumeration list of all the base skeletons
 *             The caller must close the object.
 * @stable ICU 3.8
 */
U_STABLE UEnumeration * U_EXPORT2
udatpg_openBaseSkeletons(const UDateTimePatternGenerator *dtpg, UErrorCode *pErrorCode);

/**
 * Get the pattern corresponding to a given skeleton.
 * 
 * @param dtpg a pointer to UDateTimePatternGenerator.
 * @param skeleton 
 * @param skeletonLength pointer to the length of skeleton.
 * @param pLength pointer to the length of return pattern.
 * @return pattern corresponding to a given skeleton.
 * @stable ICU 3.8
 */
U_STABLE const UChar * U_EXPORT2
udatpg_getPatternForSkeleton(const UDateTimePatternGenerator *dtpg,
                             const UChar *skeleton, int32_t skeletonLength,
                             int32_t *pLength);




U_NAMESPACE_BEGIN

/**
 * \file
 * \brief C++ API: Date/Time Pattern Generator
 */


class CharString;
class Hashtable;
class FormatParser;
class DateTimeMatcher;
class DistanceInfo;
class PatternMap;
class PtnSkeleton;
class SharedDateTimePatternGenerator;

/**
 * This class provides flexible generation of date format patterns, like "yy-MM-dd".
 * The user can build up the generator by adding successive patterns. Once that
 * is done, a query can be made using a "skeleton", which is a pattern which just
 * includes the desired fields and lengths. The generator will return the "best fit"
 * pattern corresponding to that skeleton.
 * <p>The main method people will use is getBestPattern(String skeleton),
 * since normally this class is pre-built with data from a particular locale.
 * However, generators can be built directly from other data as well.
 * <p><i>Issue: may be useful to also have a function that returns the list of
 * fields in a pattern, in order, since we have that internally.
 * That would be useful for getting the UI order of field elements.</i>
 * @stable ICU 3.8
**/
class U_I18N_API DateTimePatternGenerator : public UObject {
public:
    /**
     * Construct a flexible generator according to default locale.
     * @param status  Output param set to success/failure code on exit,
     *               which must not indicate a failure before the function call.
     * @stable ICU 3.8
     */
    static DateTimePatternGenerator* U_EXPORT2 createInstance(UErrorCode& status);

    /**
     * Construct a flexible generator according to data for a given locale.
     * @param uLocale
     * @param status  Output param set to success/failure code on exit,
     *               which must not indicate a failure before the function call.
     * @stable ICU 3.8
     */
    static DateTimePatternGenerator* U_EXPORT2 createInstance(const Locale& uLocale, UErrorCode& status);

#ifndef U_HIDE_INTERNAL_API

    /**
     * For ICU use only
     *
     * @internal
     */
    static DateTimePatternGenerator* U_EXPORT2 internalMakeInstance(const Locale& uLocale, UErrorCode& status);

#endif /* U_HIDE_INTERNAL_API */

    /**
     * Create an empty generator, to be constructed with addPattern(...) etc.
     * @param status  Output param set to success/failure code on exit,
     *               which must not indicate a failure before the function call.
     * @stable ICU 3.8
     */
     static DateTimePatternGenerator* U_EXPORT2 createEmptyInstance(UErrorCode& status);

    /**
     * Destructor.
     * @stable ICU 3.8
     */
    virtual ~DateTimePatternGenerator();

    /**
     * Clone DateTimePatternGenerator object. Clients are responsible for
     * deleting the DateTimePatternGenerator object cloned.
     * @stable ICU 3.8
     */
    DateTimePatternGenerator* clone() const;

     /**
      * Return true if another object is semantically equal to this one.
      *
      * @param other    the DateTimePatternGenerator object to be compared with.
      * @return         true if other is semantically equal to this.
      * @stable ICU 3.8
      */
    UBool operator==(const DateTimePatternGenerator& other) const;

    /**
     * Return true if another object is semantically unequal to this one.
     *
     * @param other    the DateTimePatternGenerator object to be compared with.
     * @return         true if other is semantically unequal to this.
     * @stable ICU 3.8
     */
    UBool operator!=(const DateTimePatternGenerator& other) const;

    /**
     * Utility to return a unique skeleton from a given pattern. For example,
     * both "MMM-dd" and "dd/MMM" produce the skeleton "MMMdd".
     *
     * @param pattern   Input pattern, such as "dd/MMM"
     * @param status  Output param set to success/failure code on exit,
     *                  which must not indicate a failure before the function call.
     * @return skeleton such as "MMMdd"
     * @stable ICU 56
     */
    static UnicodeString staticGetSkeleton(const UnicodeString& pattern, UErrorCode& status);

    /**
     * Utility to return a unique skeleton from a given pattern. For example,
     * both "MMM-dd" and "dd/MMM" produce the skeleton "MMMdd".
     * getSkeleton() works exactly like staticGetSkeleton().
     * Use staticGetSkeleton() instead of getSkeleton().
     *
     * @param pattern   Input pattern, such as "dd/MMM"
     * @param status  Output param set to success/failure code on exit,
     *                  which must not indicate a failure before the function call.
     * @return skeleton such as "MMMdd"
     * @stable ICU 3.8
     */
    UnicodeString getSkeleton(const UnicodeString& pattern, UErrorCode& status); /* {
        The function is commented out because it is a stable API calling a draft API.
        After staticGetSkeleton becomes stable, staticGetSkeleton can be used and
        these comments and the definition of getSkeleton in dtptngen.cpp should be removed.
        return staticGetSkeleton(pattern, status);
    }*/

    /**
     * Utility to return a unique base skeleton from a given pattern. This is
     * the same as the skeleton, except that differences in length are minimized
     * so as to only preserve the difference between string and numeric form. So
     * for example, both "MMM-dd" and "d/MMM" produce the skeleton "MMMd"
     * (notice the single d).
     *
     * @param pattern  Input pattern, such as "dd/MMM"
     * @param status  Output param set to success/failure code on exit,
     *               which must not indicate a failure before the function call.
     * @return base skeleton, such as "MMMd"
     * @stable ICU 56
     */
    static UnicodeString staticGetBaseSkeleton(const UnicodeString& pattern, UErrorCode& status);

    /**
     * Utility to return a unique base skeleton from a given pattern. This is
     * the same as the skeleton, except that differences in length are minimized
     * so as to only preserve the difference between string and numeric form. So
     * for example, both "MMM-dd" and "d/MMM" produce the skeleton "MMMd"
     * (notice the single d).
     * getBaseSkeleton() works exactly like staticGetBaseSkeleton().
     * Use staticGetBaseSkeleton() instead of getBaseSkeleton().
     *
     * @param pattern  Input pattern, such as "dd/MMM"
     * @param status  Output param set to success/failure code on exit,
     *               which must not indicate a failure before the function call.
     * @return base skeleton, such as "MMMd"
     * @stable ICU 3.8
     */
    UnicodeString getBaseSkeleton(const UnicodeString& pattern, UErrorCode& status); /* {
        The function is commented out because it is a stable API calling a draft API.
        After staticGetBaseSkeleton becomes stable, staticGetBaseSkeleton can be used and
        these comments and the definition of getBaseSkeleton in dtptngen.cpp should be removed.
        return staticGetBaseSkeleton(pattern, status);
    }*/

    /**
     * Adds a pattern to the generator. If the pattern has the same skeleton as
     * an existing pattern, and the override parameter is set, then the previous
     * value is overriden. Otherwise, the previous value is retained. In either
     * case, the conflicting status is set and previous vale is stored in
     * conflicting pattern.
     * <p>
     * Note that single-field patterns (like "MMM") are automatically added, and
     * don't need to be added explicitly!
     *
     * @param pattern   Input pattern, such as "dd/MMM"
     * @param override  When existing values are to be overridden use true,
     *                   otherwise use false.
     * @param conflictingPattern  Previous pattern with the same skeleton.
     * @param status  Output param set to success/failure code on exit,
     *               which must not indicate a failure before the function call.
     * @return conflicting status.  The value could be UDATPG_NO_CONFLICT,
     *                             UDATPG_BASE_CONFLICT or UDATPG_CONFLICT.
     * @stable ICU 3.8
     * <p>
     * <h4>Sample code</h4>
     * \snippet samples/dtptngsample/dtptngsample.cpp getBestPatternExample1
     * \snippet samples/dtptngsample/dtptngsample.cpp addPatternExample
     * <p>
     */
    UDateTimePatternConflict addPattern(const UnicodeString& pattern,
                                        UBool override,
                                        UnicodeString& conflictingPattern,
                                        UErrorCode& status);

    /**
     * An AppendItem format is a pattern used to append a field if there is no
     * good match. For example, suppose that the input skeleton is "GyyyyMMMd",
     * and there is no matching pattern internally, but there is a pattern
     * matching "yyyyMMMd", say "d-MM-yyyy". Then that pattern is used, plus the
     * G. The way these two are conjoined is by using the AppendItemFormat for G
     * (era). So if that value is, say "{0}, {1}" then the final resulting
     * pattern is "d-MM-yyyy, G".
     * <p>
     * There are actually three available variables: {0} is the pattern so far,
     * {1} is the element we are adding, and {2} is the name of the element.
     * <p>
     * This reflects the way that the CLDR data is organized.
     *
     * @param field  such as UDATPG_ERA_FIELD.
     * @param value  pattern, such as "{0}, {1}"
     * @stable ICU 3.8
     */
    void setAppendItemFormat(UDateTimePatternField field, const UnicodeString& value);

    /**
     * Getter corresponding to setAppendItemFormat. Values below 0 or at or
     * above UDATPG_FIELD_COUNT are illegal arguments.
     *
     * @param  field  such as UDATPG_ERA_FIELD.
     * @return append pattern for field
     * @stable ICU 3.8
     */
    const UnicodeString& getAppendItemFormat(UDateTimePatternField field) const;

    /**
     * Sets the names of field, eg "era" in English for ERA. These are only
     * used if the corresponding AppendItemFormat is used, and if it contains a
     * {2} variable.
     * <p>
     * This reflects the way that the CLDR data is organized.
     *
     * @param field   such as UDATPG_ERA_FIELD.
     * @param value   name of the field
     * @stable ICU 3.8
     */
    void setAppendItemName(UDateTimePatternField field, const UnicodeString& value);

    /**
     * Getter corresponding to setAppendItemNames. Values below 0 or at or above
     * UDATPG_FIELD_COUNT are illegal arguments. Note: The more general method
     * for getting date/time field display names is getFieldDisplayName.
     *
     * @param field  such as UDATPG_ERA_FIELD.
     * @return name for field
     * @see getFieldDisplayName
     * @stable ICU 3.8
     */
    const UnicodeString& getAppendItemName(UDateTimePatternField field) const;

    /**
     * The general interface to get a display name for a particular date/time field,
     * in one of several possible display widths.
     *
     * @param field  The desired UDateTimePatternField, such as UDATPG_ERA_FIELD.
     * @param width  The desired UDateTimePGDisplayWidth, such as UDATPG_ABBREVIATED.
     * @return       The display name for field
     * @stable ICU 61
     */
    UnicodeString getFieldDisplayName(UDateTimePatternField field, UDateTimePGDisplayWidth width) const;

    /**
     * The DateTimeFormat is a message format pattern used to compose date and
     * time patterns. The default pattern in the root locale is "{1} {0}", where
     * {1} will be replaced by the date pattern and {0} will be replaced by the
     * time pattern; however, other locales may specify patterns such as
     * "{1}, {0}" or "{1} 'at' {0}", etc.
     * <p>
     * This is used when the input skeleton contains both date and time fields,
     * but there is not a close match among the added patterns. For example,
     * suppose that this object was created by adding "dd-MMM" and "hh:mm", and
     * its datetimeFormat is the default "{1} {0}". Then if the input skeleton
     * is "MMMdhmm", there is not an exact match, so the input skeleton is
     * broken up into two components "MMMd" and "hmm". There are close matches
     * for those two skeletons, so the result is put together with this pattern,
     * resulting in "d-MMM h:mm".
     *
     * @param dateTimeFormat
     *            message format pattern, here {1} will be replaced by the date
     *            pattern and {0} will be replaced by the time pattern.
     * @stable ICU 3.8
     */
    void setDateTimeFormat(const UnicodeString& dateTimeFormat);

    /**
     * Getter corresponding to setDateTimeFormat.
     * @return DateTimeFormat.
     * @stable ICU 3.8
     */
    const UnicodeString& getDateTimeFormat() const;

    /**
     * Return the best pattern matching the input skeleton. It is guaranteed to
     * have all of the fields in the skeleton.
     *
     * @param skeleton
     *            The skeleton is a pattern containing only the variable fields.
     *            For example, "MMMdd" and "mmhh" are skeletons.
     * @param status  Output param set to success/failure code on exit,
     *               which must not indicate a failure before the function call.
     * @return bestPattern
     *            The best pattern found from the given skeleton.
     * @stable ICU 3.8
     * <p>
     * <h4>Sample code</h4>
     * \snippet samples/dtptngsample/dtptngsample.cpp getBestPatternExample1
     * \snippet samples/dtptngsample/dtptngsample.cpp getBestPatternExample
     * <p>
     */
     UnicodeString getBestPattern(const UnicodeString& skeleton, UErrorCode& status);


    /**
     * Return the best pattern matching the input skeleton. It is guaranteed to
     * have all of the fields in the skeleton.
     *
     * @param skeleton
     *            The skeleton is a pattern containing only the variable fields.
     *            For example, "MMMdd" and "mmhh" are skeletons.
     * @param options
     *            Options for forcing the length of specified fields in the
     *            returned pattern to match those in the skeleton (when this
     *            would not happen otherwise). For default behavior, use
     *            UDATPG_MATCH_NO_OPTIONS.
     * @param status
     *            Output param set to success/failure code on exit,
     *            which must not indicate a failure before the function call.
     * @return bestPattern
     *            The best pattern found from the given skeleton.
     * @stable ICU 4.4
     */
     UnicodeString getBestPattern(const UnicodeString& skeleton,
                                  UDateTimePatternMatchOptions options,
                                  UErrorCode& status);


    /**
     * Adjusts the field types (width and subtype) of a pattern to match what is
     * in a skeleton. That is, if you supply a pattern like "d-M H:m", and a
     * skeleton of "MMMMddhhmm", then the input pattern is adjusted to be
     * "dd-MMMM hh:mm". This is used internally to get the best match for the
     * input skeleton, but can also be used externally.
     *
     * @param pattern Input pattern
     * @param skeleton
     *            The skeleton is a pattern containing only the variable fields.
     *            For example, "MMMdd" and "mmhh" are skeletons.
     * @param status  Output param set to success/failure code on exit,
     *               which must not indicate a failure before the function call.
     * @return pattern adjusted to match the skeleton fields widths and subtypes.
     * @stable ICU 3.8
     * <p>
     * <h4>Sample code</h4>
     * \snippet samples/dtptngsample/dtptngsample.cpp getBestPatternExample1
     * \snippet samples/dtptngsample/dtptngsample.cpp replaceFieldTypesExample
     * <p>
     */
     UnicodeString replaceFieldTypes(const UnicodeString& pattern,
                                     const UnicodeString& skeleton,
                                     UErrorCode& status);

    /**
     * Adjusts the field types (width and subtype) of a pattern to match what is
     * in a skeleton. That is, if you supply a pattern like "d-M H:m", and a
     * skeleton of "MMMMddhhmm", then the input pattern is adjusted to be
     * "dd-MMMM hh:mm". This is used internally to get the best match for the
     * input skeleton, but can also be used externally.
     *
     * @param pattern Input pattern
     * @param skeleton
     *            The skeleton is a pattern containing only the variable fields.
     *            For example, "MMMdd" and "mmhh" are skeletons.
     * @param options
     *            Options controlling whether the length of specified fields in the
     *            pattern are adjusted to match those in the skeleton (when this
     *            would not happen otherwise). For default behavior, use
     *            UDATPG_MATCH_NO_OPTIONS.
     * @param status
     *            Output param set to success/failure code on exit,
     *            which must not indicate a failure before the function call.
     * @return pattern adjusted to match the skeleton fields widths and subtypes.
     * @stable ICU 4.4
     */
     UnicodeString replaceFieldTypes(const UnicodeString& pattern,
                                     const UnicodeString& skeleton,
                                     UDateTimePatternMatchOptions options,
                                     UErrorCode& status);

    /**
     * Return a list of all the skeletons (in canonical form) from this class.
     *
     * Call getPatternForSkeleton() to get the corresponding pattern.
     *
     * @param status  Output param set to success/failure code on exit,
     *               which must not indicate a failure before the function call.
     * @return StringEnumeration with the skeletons.
     *         The caller must delete the object.
     * @stable ICU 3.8
     */
     StringEnumeration* getSkeletons(UErrorCode& status) const;

     /**
      * Get the pattern corresponding to a given skeleton.
      * @param skeleton
      * @return pattern corresponding to a given skeleton.
      * @stable ICU 3.8
      */
     const UnicodeString& getPatternForSkeleton(const UnicodeString& skeleton) const;

    /**
     * Return a list of all the base skeletons (in canonical form) from this class.
     *
     * @param status  Output param set to success/failure code on exit,
     *               which must not indicate a failure before the function call.
     * @return a StringEnumeration with the base skeletons.
     *         The caller must delete the object.
     * @stable ICU 3.8
     */
     StringEnumeration* getBaseSkeletons(UErrorCode& status) const;

#ifndef U_HIDE_INTERNAL_API
     /**
      * Return a list of redundant patterns are those which if removed, make no
      * difference in the resulting getBestPattern values. This method returns a
      * list of them, to help check the consistency of the patterns used to build
      * this generator.
      *
      * @param status  Output param set to success/failure code on exit,
      *               which must not indicate a failure before the function call.
      * @return a StringEnumeration with the redundant pattern.
      *         The caller must delete the object.
      * @internal ICU 3.8
      */
     StringEnumeration* getRedundants(UErrorCode& status);
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * The decimal value is used in formatting fractions of seconds. If the
     * skeleton contains fractional seconds, then this is used with the
     * fractional seconds. For example, suppose that the input pattern is
     * "hhmmssSSSS", and the best matching pattern internally is "H:mm:ss", and
     * the decimal string is ",". Then the resulting pattern is modified to be
     * "H:mm:ss,SSSS"
     *
     * @param decimal
     * @stable ICU 3.8
     */
    void setDecimal(const UnicodeString& decimal);

    /**
     * Getter corresponding to setDecimal.
     * @return UnicodeString corresponding to the decimal point
     * @stable ICU 3.8
     */
    const UnicodeString& getDecimal() const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 3.8
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 3.8
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

private:
    /**
     * Constructor.
     */
    DateTimePatternGenerator(UErrorCode & status);

    /**
     * Constructor.
     */
    DateTimePatternGenerator(const Locale& locale, UErrorCode & status);

    /**
     * Copy constructor.
     * @param other DateTimePatternGenerator to copy
     */
    DateTimePatternGenerator(const DateTimePatternGenerator& other);

    /**
     * Default assignment operator.
     * @param other DateTimePatternGenerator to copy
     */
    DateTimePatternGenerator& operator=(const DateTimePatternGenerator& other);

    // TODO(ticket:13619): re-enable when UDATPG_NARROW no longer in  draft mode.
    // static const int32_t UDATPG_WIDTH_COUNT = UDATPG_NARROW + 1;

    Locale pLocale;  // pattern locale
    FormatParser *fp;
    DateTimeMatcher* dtMatcher;
    DistanceInfo *distanceInfo;
    PatternMap *patternMap;
    UnicodeString appendItemFormats[UDATPG_FIELD_COUNT];
    // TODO(ticket:13619): [3] -> UDATPG_WIDTH_COUNT
    UnicodeString fieldDisplayNames[UDATPG_FIELD_COUNT][3];
    UnicodeString dateTimeFormat;
    UnicodeString decimal;
    DateTimeMatcher *skipMatcher;
    Hashtable *fAvailableFormatKeyHash;
    UnicodeString emptyString;
    char16_t fDefaultHourFormatChar;

    int32_t fAllowedHourFormats[7];  // Actually an array of AllowedHourFormat enum type, ending with UNKNOWN.

    // Internal error code used for recording/reporting errors that occur during methods that do not
    // have a UErrorCode parameter. For example: the Copy Constructor, or the ::clone() method.
    // When this is set to an error the object is in an invalid state.
    UErrorCode internalErrorCode;

    /* internal flags masks for adjustFieldTypes etc. */
    enum {
        kDTPGNoFlags = 0,
        kDTPGFixFractionalSeconds = 1,
        kDTPGSkeletonUsesCapJ = 2
        // with #13183, no longer need flags for b, B
    };

    void initData(const Locale &locale, UErrorCode &status);
    void addCanonicalItems(UErrorCode &status);
    void addICUPatterns(const Locale& locale, UErrorCode& status);
    void hackTimes(const UnicodeString& hackPattern, UErrorCode& status);
    void getCalendarTypeToUse(const Locale& locale, CharString& destination, UErrorCode& err);
    void consumeShortTimePattern(const UnicodeString& shortTimePattern, UErrorCode& status);
    void addCLDRData(const Locale& locale, UErrorCode& status);
    UDateTimePatternConflict addPatternWithSkeleton(const UnicodeString& pattern, const UnicodeString * skeletonToUse, UBool override, UnicodeString& conflictingPattern, UErrorCode& status);
    void initHashtable(UErrorCode& status);
    void setDateTimeFromCalendar(const Locale& locale, UErrorCode& status);
    void setDecimalSymbols(const Locale& locale, UErrorCode& status);
    UDateTimePatternField getAppendFormatNumber(const char* field) const;
    // Note for the next 3: UDateTimePGDisplayWidth is now stable ICU 61
    UDateTimePatternField getFieldAndWidthIndices(const char* key, UDateTimePGDisplayWidth* widthP) const;
    void setFieldDisplayName(UDateTimePatternField field, UDateTimePGDisplayWidth width, const UnicodeString& value);
    UnicodeString& getMutableFieldDisplayName(UDateTimePatternField field, UDateTimePGDisplayWidth width);
    void getAppendName(UDateTimePatternField field, UnicodeString& value);
    UnicodeString mapSkeletonMetacharacters(const UnicodeString& patternForm, int32_t* flags, UErrorCode& status);
    const UnicodeString* getBestRaw(DateTimeMatcher& source, int32_t includeMask, DistanceInfo* missingFields, UErrorCode& status, const PtnSkeleton** specifiedSkeletonPtr = 0);
    UnicodeString adjustFieldTypes(const UnicodeString& pattern, const PtnSkeleton* specifiedSkeleton, int32_t flags, UDateTimePatternMatchOptions options = UDATPG_MATCH_NO_OPTIONS);
    UnicodeString getBestAppending(int32_t missingFields, int32_t flags, UErrorCode& status, UDateTimePatternMatchOptions options = UDATPG_MATCH_NO_OPTIONS);
    int32_t getTopBitNumber(int32_t foundMask) const;
    void setAvailableFormat(const UnicodeString &key, UErrorCode& status);
    UBool isAvailableFormatSet(const UnicodeString &key) const;
    void copyHashtable(Hashtable *other, UErrorCode &status);
    UBool isCanonicalItem(const UnicodeString& item) const;
    static void U_CALLCONV loadAllowedHourFormatsData(UErrorCode &status);
    void getAllowedHourFormats(const Locale &locale, UErrorCode &status);

    struct AppendItemFormatsSink;
    struct AppendItemNamesSink;
    struct AvailableFormatsSink;
} ;// end class DateTimePatternGenerator

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2016, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/




/**
 * \file
 * \brief C++ API: Format and parse relative dates and times.
 */

#if !UCONFIG_NO_FORMATTING



// #include "unicode/brkiter.h"

U_NAMESPACE_BEGIN

// forward declarations
class DateFormatSymbols;
class SimpleFormatter;

// internal structure used for caching strings
struct URelativeString;

/**
 * This class is normally accessed using the kRelative or k...Relative values of EStyle as
 * parameters to DateFormat::createDateInstance.
 *
 * Example:
 *     DateFormat *fullrelative = DateFormat::createDateInstance(DateFormat::kFullRelative, loc);
 *
 * @internal ICU 3.8
 */

class RelativeDateFormat : public DateFormat {
public:
    RelativeDateFormat( UDateFormatStyle timeStyle, UDateFormatStyle dateStyle, const Locale& locale, UErrorCode& status);

    // overrides
    /**
     * Copy constructor.
     * @internal ICU 3.8
     */
    RelativeDateFormat(const RelativeDateFormat&);

    /**
     * Assignment operator.
     * @internal ICU 3.8
     */
    RelativeDateFormat& operator=(const RelativeDateFormat&);

    /**
     * Destructor.
     * @internal ICU 3.8
     */
    virtual ~RelativeDateFormat();

    /**
     * Clone this Format object polymorphically. The caller owns the result and
     * should delete it when done.
     * @return    A copy of the object.
     * @internal ICU 3.8
     */
    virtual RelativeDateFormat* clone() const;

    /**
     * Return true if the given Format objects are semantically equal. Objects
     * of different subclasses are considered unequal.
     * @param other    the object to be compared with.
     * @return         true if the given Format objects are semantically equal.
     * @internal ICU 3.8
     */
    virtual UBool operator==(const Format& other) const;


    using DateFormat::format;

    /**
     * Format a date or time, which is the standard millis since 24:00 GMT, Jan
     * 1, 1970. Overrides DateFormat pure virtual method.
     * <P>
     * Example: using the US locale: "yyyy.MM.dd e 'at' HH:mm:ss zzz" ->>
     * 1996.07.10 AD at 15:08:56 PDT
     *
     * @param cal       Calendar set to the date and time to be formatted
     *                  into a date/time string.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       The formatting position. On input: an alignment field,
     *                  if desired. On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     * @internal ICU 3.8
     */
    virtual UnicodeString& format(  Calendar& cal,
                                    UnicodeString& appendTo,
                                    FieldPosition& pos) const;

    /**
     * Format an object to produce a string. This method handles Formattable
     * objects with a UDate type. If a the Formattable object type is not a Date,
     * then it returns a failing UErrorCode.
     *
     * @param obj       The object to format. Must be a Date.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @internal ICU 3.8
     */
    virtual UnicodeString& format(const Formattable& obj,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode& status) const;


    /**
     * Parse a date/time string beginning at the given parse position. For
     * example, a time text "07/10/96 4:5 PM, PDT" will be parsed into a Date
     * that is equivalent to Date(837039928046).
     * <P>
     * By default, parsing is lenient: If the input is not in the form used by
     * this object's format method but can still be parsed as a date, then the
     * parse succeeds. Clients may insist on strict adherence to the format by
     * calling setLenient(false).
     *
     * @param text  The date/time string to be parsed
     * @param cal   a Calendar set to the date and time to be formatted
     *              into a date/time string.
     * @param pos   On input, the position at which to start parsing; on
     *              output, the position at which parsing terminated, or the
     *              start position if the parse failed.
     * @return      A valid UDate if the input could be parsed.
     * @internal ICU 3.8
     */
    virtual void parse( const UnicodeString& text,
                        Calendar& cal,
                        ParsePosition& pos) const;

    /**
     * Parse a date/time string starting at the given parse position. For
     * example, a time text "07/10/96 4:5 PM, PDT" will be parsed into a Date
     * that is equivalent to Date(837039928046).
     * <P>
     * By default, parsing is lenient: If the input is not in the form used by
     * this object's format method but can still be parsed as a date, then the
     * parse succeeds. Clients may insist on strict adherence to the format by
     * calling setLenient(false).
     *
     * @see DateFormat::setLenient(boolean)
     *
     * @param text  The date/time string to be parsed
     * @param pos   On input, the position at which to start parsing; on
     *              output, the position at which parsing terminated, or the
     *              start position if the parse failed.
     * @return      A valid UDate if the input could be parsed.
     * @internal ICU 3.8
     */
    UDate parse( const UnicodeString& text,
                 ParsePosition& pos) const;


    /**
     * Parse a date/time string. For example, a time text "07/10/96 4:5 PM, PDT"
     * will be parsed into a UDate that is equivalent to Date(837039928046).
     * Parsing begins at the beginning of the string and proceeds as far as
     * possible.  Assuming no parse errors were encountered, this function
     * doesn't return any information about how much of the string was consumed
     * by the parsing.  If you need that information, use the version of
     * parse() that takes a ParsePosition.
     *
     * @param text  The date/time string to be parsed
     * @param status Filled in with U_ZERO_ERROR if the parse was successful, and with
     *              an error value if there was a parse error.
     * @return      A valid UDate if the input could be parsed.
     * @internal ICU 3.8
     */
    virtual UDate parse( const UnicodeString& text,
                        UErrorCode& status) const;

    /**
     * Return a single pattern string generated by combining the patterns for the
     * date and time formatters associated with this object.
     * @param result Output param to receive the pattern.
     * @return       A reference to 'result'.
     * @internal ICU 4.2 technology preview
     */
    virtual UnicodeString& toPattern(UnicodeString& result, UErrorCode& status) const;

    /**
     * Get the date pattern for the the date formatter associated with this object.
     * @param result Output param to receive the date pattern.
     * @return       A reference to 'result'.
     * @internal ICU 4.2 technology preview
     */
    virtual UnicodeString& toPatternDate(UnicodeString& result, UErrorCode& status) const;

    /**
     * Get the time pattern for the the time formatter associated with this object.
     * @param result Output param to receive the time pattern.
     * @return       A reference to 'result'.
     * @internal ICU 4.2 technology preview
     */
    virtual UnicodeString& toPatternTime(UnicodeString& result, UErrorCode& status) const;

    /**
     * Apply the given unlocalized date & time pattern strings to this relative date format.
     * (i.e., after this call, this formatter will format dates and times according to
     * the new patterns)
     *
     * @param datePattern   The date pattern to be applied.
     * @param timePattern   The time pattern to be applied.
     * @internal ICU 4.2 technology preview
     */
    virtual void applyPatterns(const UnicodeString& datePattern, const UnicodeString& timePattern, UErrorCode &status);

    /**
     * Gets the date/time formatting symbols (this is an object carrying
     * the various strings and other symbols used in formatting: e.g., month
     * names and abbreviations, time zone names, AM/PM strings, etc.)
     * @return a copy of the date-time formatting data associated
     * with this date-time formatter.
     * @internal ICU 4.8
     */
    virtual const DateFormatSymbols* getDateFormatSymbols(void) const;

    /**
     * Set a particular UDisplayContext value in the formatter, such as
     * UDISPCTX_CAPITALIZATION_FOR_STANDALONE. Note: For getContext, see
     * DateFormat.
     * @param value The UDisplayContext value to set.
     * @param status Input/output status. If at entry this indicates a failure
     *               status, the function will do nothing; otherwise this will be
     *               updated with any new status from the function.
     * @internal ICU 53
     */
    virtual void setContext(UDisplayContext value, UErrorCode& status);

private:
    SimpleDateFormat *fDateTimeFormatter;
    UnicodeString fDatePattern;
    UnicodeString fTimePattern;
    SimpleFormatter *fCombinedFormat;  // the {0} {1} format.

    UDateFormatStyle fDateStyle;
    Locale  fLocale;

    int32_t fDatesLen;    // Length of array
    URelativeString *fDates; // array of strings

    UBool fCombinedHasDateAtStart;
    UBool fCapitalizationInfoSet;
    UBool fCapitalizationOfRelativeUnitsForUIListMenu;
    UBool fCapitalizationOfRelativeUnitsForStandAlone;
#if !UCONFIG_NO_BREAK_ITERATION
    BreakIterator* fCapitalizationBrkIter;
#else
    UObject* fCapitalizationBrkIter;
#endif

    /**
     * Get the string at a specific offset.
     * @param day day offset ( -1, 0, 1, etc.. )
     * @param len on output, length of string.
     * @return the string, or NULL if none at that location.
     */
    const UChar *getStringForDay(int32_t day, int32_t &len, UErrorCode &status) const;

    /**
     * Load the Date string array
     */
    void loadDates(UErrorCode &status);

    /**
     * Set fCapitalizationOfRelativeUnitsForUIListMenu, fCapitalizationOfRelativeUnitsForStandAlone
     */
    void initCapitalizationContextInfo(const Locale& thelocale);

    /**
     * @return the number of days in "until-now"
     */
    static int32_t dayDifference(Calendar &until, UErrorCode &status);

    /**
     * initializes fCalendar from parameters.  Returns fCalendar as a convenience.
     * @param adoptZone  Zone to be adopted, or NULL for TimeZone::createDefault().
     * @param locale Locale of the calendar
     * @param status Error code
     * @return the newly constructed fCalendar
     * @internal ICU 3.8
     */
    Calendar* initializeCalendar(TimeZone* adoptZone, const Locale& locale, UErrorCode& status);

public:
    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @internal ICU 3.8
     */
    U_I18N_API static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @internal ICU 3.8
     */
    virtual UClassID getDynamicClassID(void) const;
};


U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */







// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 2005-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File WINDTFMT.H
*
********************************************************************************
*/




#if U_PLATFORM_USES_ONLY_WIN32_API

#if !UCONFIG_NO_FORMATTING







/**
 * \file
 * \brief C++ API: Format dates using Windows API.
 */

U_CDECL_BEGIN
// Forward declarations for Windows types...
typedef struct _SYSTEMTIME SYSTEMTIME;
typedef struct _TIME_ZONE_INFORMATION TIME_ZONE_INFORMATION;
U_CDECL_END

U_NAMESPACE_BEGIN

class Win32DateFormat : public DateFormat
{
public:
    Win32DateFormat(DateFormat::EStyle timeStyle, DateFormat::EStyle dateStyle, const Locale &locale, UErrorCode &status);

    Win32DateFormat(const Win32DateFormat &other);

    virtual ~Win32DateFormat();

    virtual Win32DateFormat *clone() const;

    Win32DateFormat &operator=(const Win32DateFormat &other);

    UnicodeString &format(Calendar &cal, UnicodeString &appendTo, FieldPosition &pos) const;

    using DateFormat::format;

    void parse(const UnicodeString& text, Calendar& cal, ParsePosition& pos) const;

    /**
     * Set the calendar to be used by this date format. Initially, the default
     * calendar for the specified or default locale is used.  The caller should
     * not delete the Calendar object after it is adopted by this call.
     *
     * @param calendarToAdopt    Calendar object to be adopted.
     */
    virtual void adoptCalendar(Calendar* calendarToAdopt);

    /**
     * Set the calendar to be used by this date format. Initially, the default
     * calendar for the specified or default locale is used.
     *
     * @param newCalendar Calendar object to be set.
     */
    virtual void setCalendar(const Calendar& newCalendar);

    /**
     * Sets the time zone for the calendar of this DateFormat object. The caller
     * no longer owns the TimeZone object and should not delete it after this call.
     *
     * @param zoneToAdopt the TimeZone to be adopted.
     */
    virtual void adoptTimeZone(TimeZone* zoneToAdopt);

    /**
     * Sets the time zone for the calendar of this DateFormat object.
     * @param zone the new time zone.
     */
    virtual void setTimeZone(const TimeZone& zone);

    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       derived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     */
    U_I18N_API static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     */
    virtual UClassID getDynamicClassID(void) const;

private:
    void formatDate(const SYSTEMTIME *st, UnicodeString &appendTo) const;
    void formatTime(const SYSTEMTIME *st, UnicodeString &appendTo) const;

    UnicodeString setTimeZoneInfo(TIME_ZONE_INFORMATION *tzi, const TimeZone &zone) const;
    UnicodeString* getTimeDateFormat(const Calendar *cal, const Locale *locale, UErrorCode &status) const;

    UnicodeString *fDateTimeMsg;
    DateFormat::EStyle fTimeStyle;
    DateFormat::EStyle fDateStyle;
    Locale fLocale;
    UnicodeString fZoneID;
    TIME_ZONE_INFORMATION *fTZI;

    UnicodeString* fWindowsLocaleName; // Stores the equivalent Windows locale name.
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif // U_PLATFORM_USES_ONLY_WIN32_API



#if defined( U_DEBUG_CALSVC ) || defined (U_DEBUG_CAL)
#include <stdio.h>
#endif

// *****************************************************************************
// class DateFormat
// *****************************************************************************

U_NAMESPACE_BEGIN

class U_I18N_API DateFmtBestPattern : public SharedObject {
public:
    UnicodeString fPattern;

    DateFmtBestPattern(const UnicodeString &pattern)
            : fPattern(pattern) { }
    ~DateFmtBestPattern();
};

DateFmtBestPattern::~DateFmtBestPattern() {
}

template<> U_I18N_API
const DateFmtBestPattern *LocaleCacheKey<DateFmtBestPattern>::createObject(
        const void * /*creationContext*/, UErrorCode &status) const {
    status = U_UNSUPPORTED_ERROR;
    return NULL;
}

class U_I18N_API DateFmtBestPatternKey : public LocaleCacheKey<DateFmtBestPattern> { 
private:
    UnicodeString fSkeleton;
public:
    DateFmtBestPatternKey(
        const Locale &loc,
        const UnicodeString &skeleton,
        UErrorCode &status)
            : LocaleCacheKey<DateFmtBestPattern>(loc),
              fSkeleton(DateTimePatternGenerator::staticGetSkeleton(skeleton, status)) { }
    DateFmtBestPatternKey(const DateFmtBestPatternKey &other) :
            LocaleCacheKey<DateFmtBestPattern>(other),
            fSkeleton(other.fSkeleton) { }
    virtual ~DateFmtBestPatternKey();
    virtual int32_t hashCode() const {
        return (int32_t)(37u * (uint32_t)LocaleCacheKey<DateFmtBestPattern>::hashCode() + (uint32_t)fSkeleton.hashCode());
    }
    virtual UBool operator==(const CacheKeyBase &other) const {
       // reflexive
       if (this == &other) { 	
           return TRUE;
       }
       if (!LocaleCacheKey<DateFmtBestPattern>::operator==(other)) {
           return FALSE;
       }
       // We know that this and other are of same class if we get this far.
       const DateFmtBestPatternKey &realOther =
               static_cast<const DateFmtBestPatternKey &>(other);
       return (realOther.fSkeleton == fSkeleton);
    }
    virtual CacheKeyBase *clone() const {
        return new DateFmtBestPatternKey(*this);
    }
    virtual const DateFmtBestPattern *createObject(
            const void * /*unused*/, UErrorCode &status) const {
        LocalPointer<DateTimePatternGenerator> dtpg(
                    DateTimePatternGenerator::createInstance(fLoc, status));
        if (U_FAILURE(status)) {
            return NULL;
        }
  
        LocalPointer<DateFmtBestPattern> pattern(
                new DateFmtBestPattern(
                        dtpg->getBestPattern(fSkeleton, status)),
                status);
        if (U_FAILURE(status)) {
            return NULL;
        }
        DateFmtBestPattern *result = pattern.orphan();
        result->addRef();
        return result;
    }
};

DateFmtBestPatternKey::~DateFmtBestPatternKey() { }


DateFormat::DateFormat()
:   fCalendar(0),
    fNumberFormat(0),
    fCapitalizationContext(UDISPCTX_CAPITALIZATION_NONE)
{
}

//----------------------------------------------------------------------

DateFormat::DateFormat(const DateFormat& other)
:   Format(other),
    fCalendar(0),
    fNumberFormat(0),
    fCapitalizationContext(UDISPCTX_CAPITALIZATION_NONE)
{
    *this = other;
}

//----------------------------------------------------------------------

DateFormat& DateFormat::operator=(const DateFormat& other)
{
    if (this != &other)
    {
        delete fCalendar;
        delete fNumberFormat;
        if(other.fCalendar) {
          fCalendar = other.fCalendar->clone();
        } else {
          fCalendar = NULL;
        }
        if(other.fNumberFormat) {
          fNumberFormat = other.fNumberFormat->clone();
        } else {
          fNumberFormat = NULL;
        }
        fBoolFlags = other.fBoolFlags;
        fCapitalizationContext = other.fCapitalizationContext;
    }
    return *this;
}

//----------------------------------------------------------------------

DateFormat::~DateFormat()
{
    delete fCalendar;
    delete fNumberFormat;
}

//----------------------------------------------------------------------

UBool
DateFormat::operator==(const Format& other) const
{
    // This protected comparison operator should only be called by subclasses
    // which have confirmed that the other object being compared against is
    // an instance of a sublcass of DateFormat.  THIS IS IMPORTANT.

    // Format::operator== guarantees that this cast is safe
    DateFormat* fmt = (DateFormat*)&other;

    return (this == fmt) ||
        (Format::operator==(other) &&
         fCalendar&&(fCalendar->isEquivalentTo(*fmt->fCalendar)) &&
         (fNumberFormat && *fNumberFormat == *fmt->fNumberFormat) &&
         (fCapitalizationContext == fmt->fCapitalizationContext) );
}

//----------------------------------------------------------------------

UnicodeString&
DateFormat::format(const Formattable& obj,
                   UnicodeString& appendTo,
                   FieldPosition& fieldPosition,
                   UErrorCode& status) const
{
    if (U_FAILURE(status)) return appendTo;

    // if the type of the Formattable is double or long, treat it as if it were a Date
    UDate date = 0;
    switch (obj.getType())
    {
    case Formattable::kDate:
        date = obj.getDate();
        break;
    case Formattable::kDouble:
        date = (UDate)obj.getDouble();
        break;
    case Formattable::kLong:
        date = (UDate)obj.getLong();
        break;
    default:
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return appendTo;
    }

    // Is this right?
    //if (fieldPosition.getBeginIndex() == fieldPosition.getEndIndex())
    //  status = U_ILLEGAL_ARGUMENT_ERROR;

    return format(date, appendTo, fieldPosition);
}

//----------------------------------------------------------------------

UnicodeString&
DateFormat::format(const Formattable& obj,
                   UnicodeString& appendTo,
                   FieldPositionIterator* posIter,
                   UErrorCode& status) const
{
    if (U_FAILURE(status)) return appendTo;

    // if the type of the Formattable is double or long, treat it as if it were a Date
    UDate date = 0;
    switch (obj.getType())
    {
    case Formattable::kDate:
        date = obj.getDate();
        break;
    case Formattable::kDouble:
        date = (UDate)obj.getDouble();
        break;
    case Formattable::kLong:
        date = (UDate)obj.getLong();
        break;
    default:
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return appendTo;
    }

    // Is this right?
    //if (fieldPosition.getBeginIndex() == fieldPosition.getEndIndex())
    //  status = U_ILLEGAL_ARGUMENT_ERROR;

    return format(date, appendTo, posIter, status);
}

//----------------------------------------------------------------------

// Default implementation for backwards compatibility, subclasses should implement.
UnicodeString&
DateFormat::format(Calendar& /* unused cal */,
                   UnicodeString& appendTo,
                   FieldPositionIterator* /* unused posIter */,
                   UErrorCode& status) const {
    if (U_SUCCESS(status)) {
        status = U_UNSUPPORTED_ERROR;
    }
    return appendTo;
}

//----------------------------------------------------------------------

UnicodeString&
DateFormat::format(UDate date, UnicodeString& appendTo, FieldPosition& fieldPosition) const {
    if (fCalendar != NULL) {
        // Use a clone of our calendar instance
        Calendar* calClone = fCalendar->clone();
        if (calClone != NULL) {
            UErrorCode ec = U_ZERO_ERROR;
            calClone->setTime(date, ec);
            if (U_SUCCESS(ec)) {
                format(*calClone, appendTo, fieldPosition);
            }
            delete calClone;
        }
    }
    return appendTo;
}

//----------------------------------------------------------------------

UnicodeString&
DateFormat::format(UDate date, UnicodeString& appendTo, FieldPositionIterator* posIter,
                   UErrorCode& status) const {
    if (fCalendar != NULL) {
        Calendar* calClone = fCalendar->clone();
        if (calClone != NULL) {
            calClone->setTime(date, status);
            if (U_SUCCESS(status)) {
               format(*calClone, appendTo, posIter, status);
            }
            delete calClone;
        }
    }
    return appendTo;
}

//----------------------------------------------------------------------

UnicodeString&
DateFormat::format(UDate date, UnicodeString& appendTo) const
{
    // Note that any error information is just lost.  That's okay
    // for this convenience method.
    FieldPosition fpos(FieldPosition::DONT_CARE);
    return format(date, appendTo, fpos);
}

//----------------------------------------------------------------------

UDate
DateFormat::parse(const UnicodeString& text,
                  ParsePosition& pos) const
{
    UDate d = 0; // Error return UDate is 0 (the epoch)
    if (fCalendar != NULL) {
        Calendar* calClone = fCalendar->clone();
        if (calClone != NULL) {
            int32_t start = pos.getIndex();
            calClone->clear();
            parse(text, *calClone, pos);
            if (pos.getIndex() != start) {
                UErrorCode ec = U_ZERO_ERROR;
                d = calClone->getTime(ec);
                if (U_FAILURE(ec)) {
                    // We arrive here if fCalendar => calClone is non-lenient and
                    // there is an out-of-range field.  We don't know which field
                    // was illegal so we set the error index to the start.
                    pos.setIndex(start);
                    pos.setErrorIndex(start);
                    d = 0;
                }
            }
            delete calClone;
        }
    }
    return d;
}

//----------------------------------------------------------------------

UDate
DateFormat::parse(const UnicodeString& text,
                  UErrorCode& status) const
{
    if (U_FAILURE(status)) return 0;

    ParsePosition pos(0);
    UDate result = parse(text, pos);
    if (pos.getIndex() == 0) {
#if defined (U_DEBUG_CAL)
      fprintf(stderr, "%s:%d - - failed to parse  - err index %d\n"
              , __FILE__, __LINE__, pos.getErrorIndex() );
#endif
      status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return result;
}

//----------------------------------------------------------------------

void
DateFormat::parseObject(const UnicodeString& source,
                        Formattable& result,
                        ParsePosition& pos) const
{
    result.setDate(parse(source, pos));
}

//----------------------------------------------------------------------

DateFormat* U_EXPORT2
DateFormat::createTimeInstance(DateFormat::EStyle style,
                               const Locale& aLocale)
{
    return createDateTimeInstance(kNone, style, aLocale);
}

//----------------------------------------------------------------------

DateFormat* U_EXPORT2
DateFormat::createDateInstance(DateFormat::EStyle style,
                               const Locale& aLocale)
{
    return createDateTimeInstance(style, kNone, aLocale);
}

//----------------------------------------------------------------------

DateFormat* U_EXPORT2
DateFormat::createDateTimeInstance(EStyle dateStyle,
                                   EStyle timeStyle,
                                   const Locale& aLocale)
{
   if(dateStyle != kNone)
   {
       dateStyle = (EStyle) (dateStyle + kDateOffset);
   }
   return create(timeStyle, dateStyle, aLocale);
}

//----------------------------------------------------------------------

DateFormat* U_EXPORT2
DateFormat::createInstance()
{
    return createDateTimeInstance(kShort, kShort, Locale::getDefault());
}

//----------------------------------------------------------------------

UnicodeString U_EXPORT2
DateFormat::getBestPattern(
        const Locale &locale,
        const UnicodeString &skeleton,
        UErrorCode &status) {
    UnifiedCache *cache = UnifiedCache::getInstance(status);
    if (U_FAILURE(status)) {
        return UnicodeString();
    }
    DateFmtBestPatternKey key(locale, skeleton, status);
    const DateFmtBestPattern *patternPtr = NULL;
    cache->get(key, patternPtr, status);
    if (U_FAILURE(status)) {
        return UnicodeString();
    }
    UnicodeString result(patternPtr->fPattern);
    patternPtr->removeRef();
    return result;
}

DateFormat* U_EXPORT2
DateFormat::createInstanceForSkeleton(
        Calendar *calendarToAdopt,
        const UnicodeString& skeleton,
        const Locale &locale,
        UErrorCode &status) {
    LocalPointer<Calendar> calendar(calendarToAdopt);
    if (U_FAILURE(status)) {
        return NULL;
    }
    if (calendar.isNull()) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    Locale localeWithCalendar = locale;
    localeWithCalendar.setKeywordValue("calendar", calendar->getType(), status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    DateFormat *result = createInstanceForSkeleton(skeleton, localeWithCalendar, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    result->adoptCalendar(calendar.orphan());
    return result;
}

DateFormat* U_EXPORT2
DateFormat::createInstanceForSkeleton(
        const UnicodeString& skeleton,
        const Locale &locale,
        UErrorCode &status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    LocalPointer<DateFormat> df(
        new SimpleDateFormat(
            getBestPattern(locale, skeleton, status),
            locale, status),
        status);
    return U_SUCCESS(status) ? df.orphan() : NULL;
}

DateFormat* U_EXPORT2
DateFormat::createInstanceForSkeleton(
        const UnicodeString& skeleton,
        UErrorCode &status) {
    return createInstanceForSkeleton(
            skeleton, Locale::getDefault(), status);
}

//----------------------------------------------------------------------

DateFormat* U_EXPORT2
DateFormat::create(EStyle timeStyle, EStyle dateStyle, const Locale& locale)
{
    UErrorCode status = U_ZERO_ERROR;
#if U_PLATFORM_USES_ONLY_WIN32_API
    char buffer[8];
    int32_t count = locale.getKeywordValue("compat", buffer, sizeof(buffer), status);

    // if the locale has "@compat=host", create a host-specific DateFormat...
    if (count > 0 && uprv_strcmp(buffer, "host") == 0) {
        Win32DateFormat *f = new Win32DateFormat(timeStyle, dateStyle, locale, status);

        if (U_SUCCESS(status)) {
            return f;
        }

        delete f;
    }
#endif

    // is it relative?
    if(/*((timeStyle!=UDAT_NONE)&&(timeStyle & UDAT_RELATIVE)) || */((dateStyle!=kNone)&&((dateStyle-kDateOffset) & UDAT_RELATIVE))) {
        RelativeDateFormat *r = new RelativeDateFormat((UDateFormatStyle)timeStyle, (UDateFormatStyle)(dateStyle-kDateOffset), locale, status);
        if(U_SUCCESS(status)) return r;
        delete r;
        status = U_ZERO_ERROR;
    }

    // Try to create a SimpleDateFormat of the desired style.
    SimpleDateFormat *f = new SimpleDateFormat(timeStyle, dateStyle, locale, status);
    if (U_SUCCESS(status)) return f;
    delete f;

    // If that fails, try to create a format using the default pattern and
    // the DateFormatSymbols for this locale.
    status = U_ZERO_ERROR;
    f = new SimpleDateFormat(locale, status);
    if (U_SUCCESS(status)) return f;
    delete f;

    // This should never really happen, because the preceding constructor
    // should always succeed.  If the resource data is unavailable, a last
    // resort object should be returned.
    return 0;
}

//----------------------------------------------------------------------

const Locale* U_EXPORT2
DateFormat::getAvailableLocales(int32_t& count)
{
    // Get the list of installed locales.
    // Even if root has the correct date format for this locale,
    // it's still a valid locale (we don't worry about data fallbacks).
    return Locale::getAvailableLocales(count);
}

//----------------------------------------------------------------------

void
DateFormat::adoptCalendar(Calendar* newCalendar)
{
    delete fCalendar;
    fCalendar = newCalendar;
}

//----------------------------------------------------------------------
void
DateFormat::setCalendar(const Calendar& newCalendar)
{
    Calendar* newCalClone = newCalendar.clone();
    if (newCalClone != NULL) {
        adoptCalendar(newCalClone);
    }
}

//----------------------------------------------------------------------

const Calendar*
DateFormat::getCalendar() const
{
    return fCalendar;
}

//----------------------------------------------------------------------

void
DateFormat::adoptNumberFormat(NumberFormat* newNumberFormat)
{
    delete fNumberFormat;
    fNumberFormat = newNumberFormat;
    newNumberFormat->setParseIntegerOnly(TRUE);
    newNumberFormat->setGroupingUsed(FALSE);
}
//----------------------------------------------------------------------

void
DateFormat::setNumberFormat(const NumberFormat& newNumberFormat)
{
    NumberFormat* newNumFmtClone = newNumberFormat.clone();
    if (newNumFmtClone != NULL) {
        adoptNumberFormat(newNumFmtClone);
    }
}

//----------------------------------------------------------------------

const NumberFormat*
DateFormat::getNumberFormat() const
{
    return fNumberFormat;
}

//----------------------------------------------------------------------

void
DateFormat::adoptTimeZone(TimeZone* zone)
{
    if (fCalendar != NULL) {
        fCalendar->adoptTimeZone(zone);
    }
}
//----------------------------------------------------------------------

void
DateFormat::setTimeZone(const TimeZone& zone)
{
    if (fCalendar != NULL) {
        fCalendar->setTimeZone(zone);
    }
}

//----------------------------------------------------------------------

const TimeZone&
DateFormat::getTimeZone() const
{
    if (fCalendar != NULL) {
        return fCalendar->getTimeZone();
    }
    // If calendar doesn't exists, create default timezone.
    // fCalendar is rarely null
    return *(TimeZone::createDefault());
}

//----------------------------------------------------------------------

void
DateFormat::setLenient(UBool lenient)
{
    if (fCalendar != NULL) {
        fCalendar->setLenient(lenient);
    }
    UErrorCode status = U_ZERO_ERROR;
    setBooleanAttribute(UDAT_PARSE_ALLOW_WHITESPACE, lenient, status);
    setBooleanAttribute(UDAT_PARSE_ALLOW_NUMERIC, lenient, status);
}

//----------------------------------------------------------------------

UBool
DateFormat::isLenient() const
{
    UBool lenient = TRUE;
    if (fCalendar != NULL) {
        lenient = fCalendar->isLenient();
    }
    UErrorCode status = U_ZERO_ERROR;
    return lenient
        && getBooleanAttribute(UDAT_PARSE_ALLOW_WHITESPACE, status)
        && getBooleanAttribute(UDAT_PARSE_ALLOW_NUMERIC, status);
}

void
DateFormat::setCalendarLenient(UBool lenient)
{
    if (fCalendar != NULL) {
        fCalendar->setLenient(lenient);
    }
}

//----------------------------------------------------------------------

UBool
DateFormat::isCalendarLenient() const
{
    if (fCalendar != NULL) {
        return fCalendar->isLenient();
    }
    // fCalendar is rarely null
    return FALSE;
}


//----------------------------------------------------------------------


void DateFormat::setContext(UDisplayContext value, UErrorCode& status)
{
    if (U_FAILURE(status))
        return;
    if ( (UDisplayContextType)((uint32_t)value >> 8) == UDISPCTX_TYPE_CAPITALIZATION ) {
        fCapitalizationContext = value;
    } else {
        status = U_ILLEGAL_ARGUMENT_ERROR;
   }
}


//----------------------------------------------------------------------


UDisplayContext DateFormat::getContext(UDisplayContextType type, UErrorCode& status) const
{
    if (U_FAILURE(status))
        return (UDisplayContext)0;
    if (type != UDISPCTX_TYPE_CAPITALIZATION) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return (UDisplayContext)0;
    }
    return fCapitalizationContext;
}


//----------------------------------------------------------------------


DateFormat& 
DateFormat::setBooleanAttribute(UDateFormatBooleanAttribute attr,
    									UBool newValue,
    									UErrorCode &status) {
    if(!fBoolFlags.isValidValue(newValue)) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
    } else {
        fBoolFlags.set(attr, newValue);
    }

    return *this;
}

//----------------------------------------------------------------------

UBool 
DateFormat::getBooleanAttribute(UDateFormatBooleanAttribute attr, UErrorCode &/*status*/) const {

    return static_cast<UBool>(fBoolFlags.get(attr));
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2016, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* dayperiodrules.cpp
*
* created on: 2016-01-20
* created by: kazede
*/

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2016, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* dayperiodrules.h
*
* created on: 2016-01-20
* created by: kazede
*/











U_NAMESPACE_BEGIN

struct DayPeriodRulesDataSink;

class DayPeriodRules : public UMemory {
    friend struct DayPeriodRulesDataSink;
public:
    enum DayPeriod {
        DAYPERIOD_UNKNOWN = -1,
        DAYPERIOD_MIDNIGHT,
        DAYPERIOD_NOON,
        DAYPERIOD_MORNING1,
        DAYPERIOD_AFTERNOON1,
        DAYPERIOD_EVENING1,
        DAYPERIOD_NIGHT1,
        DAYPERIOD_MORNING2,
        DAYPERIOD_AFTERNOON2,
        DAYPERIOD_EVENING2,
        DAYPERIOD_NIGHT2,
        DAYPERIOD_AM,
        DAYPERIOD_PM
    };

    static const DayPeriodRules *getInstance(const Locale &locale, UErrorCode &errorCode);

    UBool hasMidnight() const { return fHasMidnight; }
    UBool hasNoon() const { return fHasNoon; }
    DayPeriod getDayPeriodForHour(int32_t hour) const { return fDayPeriodForHour[hour]; }

    // Returns the center of dayPeriod. Half hours are indicated with a .5 .
    double getMidPointForDayPeriod(DayPeriod dayPeriod, UErrorCode &errorCode) const;

private:
    DayPeriodRules();

    // Translates "morning1" to DAYPERIOD_MORNING1, for example.
    static DayPeriod getDayPeriodFromString(const char *type_str);

    static void U_CALLCONV load(UErrorCode &errorCode);

    // Sets period type for all hours in [startHour, limitHour).
    void add(int32_t startHour, int32_t limitHour, DayPeriod period);

    // Returns TRUE if for all i, DayPeriodForHour[i] has a type other than UNKNOWN.
    // Values of HasNoon and HasMidnight do not affect the return value.
    UBool allHoursAreSet();

    // Returns the hour that starts dayPeriod. Returns 0 for MIDNIGHT and 12 for NOON.
    int32_t getStartHourForDayPeriod(DayPeriod dayPeriod, UErrorCode &errorCode) const;

    // Returns the hour that ends dayPeriod, i.e. that starts the next period.
    // E.g. if fDayPeriodForHour[13] thru [16] are AFTERNOON1, then this function returns 17 if
    // queried with AFTERNOON1.
    // Returns 0 for MIDNIGHT and 12 for NOON.
    int32_t getEndHourForDayPeriod(DayPeriod dayPeriod, UErrorCode &errorCode) const;

    UBool fHasMidnight;
    UBool fHasNoon;
    DayPeriod fDayPeriodForHour[24];
};

U_NAMESPACE_END












U_NAMESPACE_BEGIN

struct DayPeriodRulesData : public UMemory {
    DayPeriodRulesData() : localeToRuleSetNumMap(NULL), rules(NULL), maxRuleSetNum(0) {}

    UHashtable *localeToRuleSetNumMap;
    DayPeriodRules *rules;
    int32_t maxRuleSetNum;
} *data = NULL;

enum CutoffType {
    CUTOFF_TYPE_UNKNOWN = -1,
    CUTOFF_TYPE_BEFORE,
    CUTOFF_TYPE_AFTER,  // TODO: AFTER is deprecated in CLDR 29. Remove.
    CUTOFF_TYPE_FROM,
    CUTOFF_TYPE_AT
};

struct DayPeriodRulesDataSink : public ResourceSink {
    DayPeriodRulesDataSink() {
        for (int32_t i = 0; i < UPRV_LENGTHOF(cutoffs); ++i) { cutoffs[i] = 0; }
    }
    virtual ~DayPeriodRulesDataSink();

    virtual void put(const char *key, ResourceValue &value, UBool, UErrorCode &errorCode) {
        ResourceTable dayPeriodData = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }

        for (int32_t i = 0; dayPeriodData.getKeyAndValue(i, key, value); ++i) {
            if (uprv_strcmp(key, "locales") == 0) {
                ResourceTable locales = value.getTable(errorCode);
                if (U_FAILURE(errorCode)) { return; }

                for (int32_t j = 0; locales.getKeyAndValue(j, key, value); ++j) {
                    UnicodeString setNum_str = value.getUnicodeString(errorCode);
                    int32_t setNum = parseSetNum(setNum_str, errorCode);
                    uhash_puti(data->localeToRuleSetNumMap, const_cast<char *>(key), setNum, &errorCode);
                }
            } else if (uprv_strcmp(key, "rules") == 0) {
                // Allocate one more than needed to skip [0]. See comment in parseSetNum().
                data->rules = new DayPeriodRules[data->maxRuleSetNum + 1];
                if (data->rules == NULL) {
                    errorCode = U_MEMORY_ALLOCATION_ERROR;
                    return;
                }
                ResourceTable rules = value.getTable(errorCode);
                processRules(rules, key, value, errorCode);
                if (U_FAILURE(errorCode)) { return; }
            }
        }
    }

    void processRules(const ResourceTable &rules, const char *key,
                      ResourceValue &value, UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return; }

        for (int32_t i = 0; rules.getKeyAndValue(i, key, value); ++i) {
            ruleSetNum = parseSetNum(key, errorCode);
            ResourceTable ruleSet = value.getTable(errorCode);
            if (U_FAILURE(errorCode)) { return; }

            for (int32_t j = 0; ruleSet.getKeyAndValue(j, key, value); ++j) {
                period = DayPeriodRules::getDayPeriodFromString(key);
                if (period == DayPeriodRules::DAYPERIOD_UNKNOWN) {
                    errorCode = U_INVALID_FORMAT_ERROR;
                    return;
                }
                ResourceTable periodDefinition = value.getTable(errorCode);
                if (U_FAILURE(errorCode)) { return; }

                for (int32_t k = 0; periodDefinition.getKeyAndValue(k, key, value); ++k) {
                    if (value.getType() == URES_STRING) {
                        // Key-value pairs (e.g. before{6:00}).
                        CutoffType type = getCutoffTypeFromString(key);
                        addCutoff(type, value.getUnicodeString(errorCode), errorCode);
                        if (U_FAILURE(errorCode)) { return; }
                    } else {
                        // Arrays (e.g. before{6:00, 24:00}).
                        cutoffType = getCutoffTypeFromString(key);
                        ResourceArray cutoffArray = value.getArray(errorCode);
                        if (U_FAILURE(errorCode)) { return; }

                        int32_t length = cutoffArray.getSize();
                        for (int32_t l = 0; l < length; ++l) {
                            cutoffArray.getValue(l, value);
                            addCutoff(cutoffType, value.getUnicodeString(errorCode), errorCode);
                            if (U_FAILURE(errorCode)) { return; }
                        }
                    }
                }
                setDayPeriodForHoursFromCutoffs(errorCode);
                for (int32_t k = 0; k < UPRV_LENGTHOF(cutoffs); ++k) {
                    cutoffs[k] = 0;
                }
            }

            if (!data->rules[ruleSetNum].allHoursAreSet()) {
                errorCode = U_INVALID_FORMAT_ERROR;
                return;
            }
        }
    }

    // Members.
    int32_t cutoffs[25];  // [0] thru [24]: 24 is allowed in "before 24".

    // "Path" to data.
    int32_t ruleSetNum;
    DayPeriodRules::DayPeriod period;
    CutoffType cutoffType;

    // Helpers.
    static int32_t parseSetNum(const UnicodeString &setNumStr, UErrorCode &errorCode) {
        CharString cs;
        cs.appendInvariantChars(setNumStr, errorCode);
        return parseSetNum(cs.data(), errorCode);
    }

    static int32_t parseSetNum(const char *setNumStr, UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return -1; }

        if (uprv_strncmp(setNumStr, "set", 3) != 0) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return -1;
        }

        int32_t i = 3;
        int32_t setNum = 0;
        while (setNumStr[i] != 0) {
            int32_t digit = setNumStr[i] - '0';
            if (digit < 0 || 9 < digit) {
                errorCode = U_INVALID_FORMAT_ERROR;
                return -1;
            }
            setNum = 10 * setNum + digit;
            ++i;
        }

        // Rule set number must not be zero. (0 is used to indicate "not found" by hashmap.)
        // Currently ICU data conveniently starts numbering rule sets from 1.
        if (setNum == 0) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return -1;
        } else {
            return setNum;
        }
    }

    void addCutoff(CutoffType type, const UnicodeString &hour_str, UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return; }

        if (type == CUTOFF_TYPE_UNKNOWN) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }

        int32_t hour = parseHour(hour_str, errorCode);
        if (U_FAILURE(errorCode)) { return; }

        cutoffs[hour] |= 1 << type;
    }

    // Translate the cutoffs[] array to day period rules.
    void setDayPeriodForHoursFromCutoffs(UErrorCode &errorCode) {
        DayPeriodRules &rule = data->rules[ruleSetNum];

        for (int32_t startHour = 0; startHour <= 24; ++startHour) {
            // AT cutoffs must be either midnight or noon.
            if (cutoffs[startHour] & (1 << CUTOFF_TYPE_AT)) {
                if (startHour == 0 && period == DayPeriodRules::DAYPERIOD_MIDNIGHT) {
                    rule.fHasMidnight = TRUE;
                } else if (startHour == 12 && period == DayPeriodRules::DAYPERIOD_NOON) {
                    rule.fHasNoon = TRUE;
                } else {
                    errorCode = U_INVALID_FORMAT_ERROR;  // Bad data.
                    return;
                }
            }

            // FROM/AFTER and BEFORE must come in a pair.
            if (cutoffs[startHour] & (1 << CUTOFF_TYPE_FROM) ||
                    cutoffs[startHour] & (1 << CUTOFF_TYPE_AFTER)) {
                for (int32_t hour = startHour + 1;; ++hour) {
                    if (hour == startHour) {
                        // We've gone around the array once and can't find a BEFORE.
                        errorCode = U_INVALID_FORMAT_ERROR;
                        return;
                    }
                    if (hour == 25) { hour = 0; }
                    if (cutoffs[hour] & (1 << CUTOFF_TYPE_BEFORE)) {
                        rule.add(startHour, hour, period);
                        break;
                    }
                }
            }
        }
    }

    // Translate "before" to CUTOFF_TYPE_BEFORE, for example.
    static CutoffType getCutoffTypeFromString(const char *type_str) {
        if (uprv_strcmp(type_str, "from") == 0) {
            return CUTOFF_TYPE_FROM;
        } else if (uprv_strcmp(type_str, "before") == 0) {
            return CUTOFF_TYPE_BEFORE;
        } else if (uprv_strcmp(type_str, "after") == 0) {
            return CUTOFF_TYPE_AFTER;
        } else if (uprv_strcmp(type_str, "at") == 0) {
            return CUTOFF_TYPE_AT;
        } else {
            return CUTOFF_TYPE_UNKNOWN;
        }
    }

    // Gets the numerical value of the hour from the Unicode string.
    static int32_t parseHour(const UnicodeString &time, UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) {
            return 0;
        }

        int32_t hourLimit = time.length() - 3;
        // `time` must look like "x:00" or "xx:00".
        // If length is wrong or `time` doesn't end with ":00", error out.
        if ((hourLimit != 1 && hourLimit != 2) ||
                time[hourLimit] != 0x3A || time[hourLimit + 1] != 0x30 ||
                time[hourLimit + 2] != 0x30) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return 0;
        }

        // If `time` doesn't begin with a number in [0, 24], error out.
        // Note: "24:00" is possible in "before 24:00".
        int32_t hour = time[0] - 0x30;
        if (hour < 0 || 9 < hour) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return 0;
        }
        if (hourLimit == 2) {
            int32_t hourDigit2 = time[1] - 0x30;
            if (hourDigit2 < 0 || 9 < hourDigit2) {
                errorCode = U_INVALID_FORMAT_ERROR;
                return 0;
            }
            hour = hour * 10 + hourDigit2;
            if (hour > 24) {
                errorCode = U_INVALID_FORMAT_ERROR;
                return 0;
            }
        }

        return hour;
    }
};  // struct DayPeriodRulesDataSink

struct DayPeriodRulesCountSink : public ResourceSink {
    virtual ~DayPeriodRulesCountSink();

    virtual void put(const char *key, ResourceValue &value, UBool, UErrorCode &errorCode) {
        ResourceTable rules = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }

        for (int32_t i = 0; rules.getKeyAndValue(i, key, value); ++i) {
            int32_t setNum = DayPeriodRulesDataSink::parseSetNum(key, errorCode);
            if (setNum > data->maxRuleSetNum) {
                data->maxRuleSetNum = setNum;
            }
        }
    }
};

// Out-of-line virtual destructors.
DayPeriodRulesDataSink::~DayPeriodRulesDataSink() {}
DayPeriodRulesCountSink::~DayPeriodRulesCountSink() {}

namespace {

UInitOnce initOnce = U_INITONCE_INITIALIZER;

U_CFUNC UBool U_CALLCONV dayPeriodRulesCleanup() {
    delete[] data->rules;
    uhash_close(data->localeToRuleSetNumMap);
    delete data;
    data = NULL;
    return TRUE;
}

}  // namespace

void U_CALLCONV DayPeriodRules::load(UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) {
        return;
    }

    data = new DayPeriodRulesData();
    data->localeToRuleSetNumMap = uhash_open(uhash_hashChars, uhash_compareChars, NULL, &errorCode);
    LocalUResourceBundlePointer rb_dayPeriods(ures_openDirect(NULL, "dayPeriods", &errorCode));

    // Get the largest rule set number (so we allocate enough objects).
    DayPeriodRulesCountSink countSink;
    ures_getAllItemsWithFallback(rb_dayPeriods.getAlias(), "rules", countSink, errorCode);

    // Populate rules.
    DayPeriodRulesDataSink sink;
    ures_getAllItemsWithFallback(rb_dayPeriods.getAlias(), "", sink, errorCode);

    ucln_i18n_registerCleanup(UCLN_I18N_DAYPERIODRULES, dayPeriodRulesCleanup);
}

const DayPeriodRules *DayPeriodRules::getInstance(const Locale &locale, UErrorCode &errorCode) {
    umtx_initOnce(initOnce, DayPeriodRules::load, errorCode);

    // If the entire day period rules data doesn't conform to spec (even if the part we want
    // does), return NULL.
    if(U_FAILURE(errorCode)) { return NULL; }

    const char *localeCode = locale.getBaseName();
    char name[ULOC_FULLNAME_CAPACITY];
    char parentName[ULOC_FULLNAME_CAPACITY];

    if (uprv_strlen(localeCode) < ULOC_FULLNAME_CAPACITY) {
        uprv_strcpy(name, localeCode);

        // Treat empty string as root.
        if (*name == '\0') {
            uprv_strcpy(name, "root");
        }
    } else {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return NULL;
    }

    int32_t ruleSetNum = 0;  // NB there is no rule set 0 and 0 is returned upon lookup failure.
    while (*name != '\0') {
        ruleSetNum = uhash_geti(data->localeToRuleSetNumMap, name);
        if (ruleSetNum == 0) {
            // name and parentName can't be the same pointer, so fill in parent then copy to child.
            uloc_getParent(name, parentName, ULOC_FULLNAME_CAPACITY, &errorCode);
            if (*parentName == '\0') {
                // Saves a lookup in the hash table.
                break;
            }
            uprv_strcpy(name, parentName);
        } else {
            break;
        }
    }

    if (ruleSetNum <= 0 || data->rules[ruleSetNum].getDayPeriodForHour(0) == DAYPERIOD_UNKNOWN) {
        // If day period for hour 0 is UNKNOWN then day period for all hours are UNKNOWN.
        // Data doesn't exist even with fallback.
        return NULL;
    } else {
        return &data->rules[ruleSetNum];
    }
}

DayPeriodRules::DayPeriodRules() : fHasMidnight(FALSE), fHasNoon(FALSE) {
    for (int32_t i = 0; i < 24; ++i) {
        fDayPeriodForHour[i] = DayPeriodRules::DAYPERIOD_UNKNOWN;
    }
}

double DayPeriodRules::getMidPointForDayPeriod(
        DayPeriodRules::DayPeriod dayPeriod, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) { return -1; }

    int32_t startHour = getStartHourForDayPeriod(dayPeriod, errorCode);
    int32_t endHour = getEndHourForDayPeriod(dayPeriod, errorCode);
    // Can't obtain startHour or endHour; bail out.
    if (U_FAILURE(errorCode)) { return -1; }

    double midPoint = (startHour + endHour) / 2.0;

    if (startHour > endHour) {
        // dayPeriod wraps around midnight. Shift midPoint by 12 hours, in the direction that
        // lands it in [0, 24).
        midPoint += 12;
        if (midPoint >= 24) {
            midPoint -= 24;
        }
    }

    return midPoint;
}

int32_t DayPeriodRules::getStartHourForDayPeriod(
        DayPeriodRules::DayPeriod dayPeriod, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) { return -1; }

    if (dayPeriod == DAYPERIOD_MIDNIGHT) { return 0; }
    if (dayPeriod == DAYPERIOD_NOON) { return 12; }

    if (fDayPeriodForHour[0] == dayPeriod && fDayPeriodForHour[23] == dayPeriod) {
        // dayPeriod wraps around midnight. Start hour is later than end hour.
        for (int32_t i = 22; i >= 1; --i) {
            if (fDayPeriodForHour[i] != dayPeriod) {
                return (i + 1);
            }
        }
    } else {
        for (int32_t i = 0; i <= 23; ++i) {
            if (fDayPeriodForHour[i] == dayPeriod) {
                return i;
            }
        }
    }

    // dayPeriod doesn't exist in rule set; set error and exit.
    errorCode = U_ILLEGAL_ARGUMENT_ERROR;
    return -1;
}

int32_t DayPeriodRules::getEndHourForDayPeriod(
        DayPeriodRules::DayPeriod dayPeriod, UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) { return -1; }

    if (dayPeriod == DAYPERIOD_MIDNIGHT) { return 0; }
    if (dayPeriod == DAYPERIOD_NOON) { return 12; }

    if (fDayPeriodForHour[0] == dayPeriod && fDayPeriodForHour[23] == dayPeriod) {
        // dayPeriod wraps around midnight. End hour is before start hour.
        for (int32_t i = 1; i <= 22; ++i) {
            if (fDayPeriodForHour[i] != dayPeriod) {
                // i o'clock is when a new period starts, therefore when the old period ends.
                return i;
            }
        }
    } else {
        for (int32_t i = 23; i >= 0; --i) {
            if (fDayPeriodForHour[i] == dayPeriod) {
                return (i + 1);
            }
        }
    }

    // dayPeriod doesn't exist in rule set; set error and exit.
    errorCode = U_ILLEGAL_ARGUMENT_ERROR;
    return -1;
}

DayPeriodRules::DayPeriod DayPeriodRules::getDayPeriodFromString(const char *type_str) {
    if (uprv_strcmp(type_str, "midnight") == 0) {
        return DAYPERIOD_MIDNIGHT;
    } else if (uprv_strcmp(type_str, "noon") == 0) {
        return DAYPERIOD_NOON;
    } else if (uprv_strcmp(type_str, "morning1") == 0) {
        return DAYPERIOD_MORNING1;
    } else if (uprv_strcmp(type_str, "afternoon1") == 0) {
        return DAYPERIOD_AFTERNOON1;
    } else if (uprv_strcmp(type_str, "evening1") == 0) {
        return DAYPERIOD_EVENING1;
    } else if (uprv_strcmp(type_str, "night1") == 0) {
        return DAYPERIOD_NIGHT1;
    } else if (uprv_strcmp(type_str, "morning2") == 0) {
        return DAYPERIOD_MORNING2;
    } else if (uprv_strcmp(type_str, "afternoon2") == 0) {
        return DAYPERIOD_AFTERNOON2;
    } else if (uprv_strcmp(type_str, "evening2") == 0) {
        return DAYPERIOD_EVENING2;
    } else if (uprv_strcmp(type_str, "night2") == 0) {
        return DAYPERIOD_NIGHT2;
    } else if (uprv_strcmp(type_str, "am") == 0) {
        return DAYPERIOD_AM;
    } else if (uprv_strcmp(type_str, "pm") == 0) {
        return DAYPERIOD_PM;
    } else {
        return DAYPERIOD_UNKNOWN;
    }
}

void DayPeriodRules::add(int32_t startHour, int32_t limitHour, DayPeriod period) {
    for (int32_t i = startHour; i != limitHour; ++i) {
        if (i == 24) { i = 0; }
        fDayPeriodForHour[i] = period;
    }
}

UBool DayPeriodRules::allHoursAreSet() {
    for (int32_t i = 0; i < 24; ++i) {
        if (fDayPeriodForHour[i] == DAYPERIOD_UNKNOWN) { return FALSE; }
    }

    return TRUE;
}



U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File DCFMTSYM.CPP
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/18/97    clhuang     Implemented with C++ APIs.
*   03/27/97    helena      Updated to pass the simple test after code review.
*   08/26/97    aliu        Added currency/intl currency symbol support.
*   07/20/98    stephen     Slightly modified initialization of monetarySeparator
********************************************************************************
*/



#if !UCONFIG_NO_FORMATTING










// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2002-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
*/







/**
 * Internal method.  Given a currency ISO code and a locale, return
 * the "static" currency name.  This is usually the same as the
 * UCURR_SYMBOL_NAME, but if the latter is a choice format, then the
 * format is applied to the number 2.0 (to yield the more common
 * plural) to return a static name.
 *
 * This is used for backward compatibility with old currency logic in
 * DecimalFormat and DecimalFormatSymbols.
 */
U_CAPI void
uprv_getStaticCurrencyName(const UChar* iso, const char* loc,
                           icu::UnicodeString& result, UErrorCode& ec);

/**
 * Attempt to parse the given string as a currency, either as a
 * display name in the given locale, or as a 3-letter ISO 4217
 * code.  If multiple display names match, then the longest one is
 * selected.  If both a display name and a 3-letter ISO code
 * match, then the display name is preferred, unless it's length
 * is less than 3.
 *
 * The parameters must not be NULL.
 *
 * @param locale the locale of the display names to match
 * @param text the text to parse
 * @param pos input-output position; on input, the position within
 * text to match; must have 0 <= pos.getIndex() < text.length();
 * on output, the position after the last matched character. If
 * the parse fails, the position in unchanged upon output.
 * @param type currency type to parse against, LONG_NAME only or not
 * @param partialMatchLen The length of the longest matching prefix;
 * this may be nonzero even if no full currency was matched.
 * @return the ISO 4217 code, as a string, of the best match, or
 * null if there is no match
 *
 * @internal
 */
U_CAPI void
uprv_parseCurrency(const char* locale,
                   const icu::UnicodeString& text,
                   icu::ParsePosition& pos,
                   int8_t type,
                   int32_t* partialMatchLen,
                   UChar* result,
                   UErrorCode& ec);

/**
 * Puts all possible first-characters of a currency into the
 * specified UnicodeSet.
 *
 * @param locale the locale of the display names of interest
 * @param result the UnicodeSet to which to add the starting characters
 */
void uprv_currencyLeads(const char* locale, icu::UnicodeSet& result, UErrorCode& ec);











// *****************************************************************************
// class DecimalFormatSymbols
// *****************************************************************************

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DecimalFormatSymbols)

static const char dcfmtsym_gNumberElements[] = "NumberElements";
static const char dcfmtsym_gCurrencySpacingTag[] = "currencySpacing";
static const char dcfmtsym_gBeforeCurrencyTag[] = "beforeCurrency";
static const char dcfmtsym_gAfterCurrencyTag[] = "afterCurrency";
static const char dcfmtsym_gCurrencyMatchTag[] = "currencyMatch";
static const char dcfmtsym_gCurrencySudMatchTag[] = "surroundingMatch";
static const char dcfmtsym_gCurrencyInsertBtnTag[] = "insertBetween";
static const char dcfmtsym_gLatn[] =  "latn";
static const char dcfmtsym_gSymbols[] = "symbols";
static const char dcfmtsym_gNumberElementsLatnSymbols[] = "NumberElements/latn/symbols";

static const UChar INTL_CURRENCY_SYMBOL_STR[] = {0xa4, 0xa4, 0};

// List of field names to be loaded from the data files.
// These are parallel with the enum ENumberFormatSymbol in unicode/dcfmtsym.h.
static const char *gNumberElementKeys[DecimalFormatSymbols::kFormatSymbolCount] = {
    "decimal",
    "group",
    NULL, /* #11897: the <list> symbol is NOT the pattern separator symbol */
    "percentSign",
    NULL, /* Native zero digit is deprecated from CLDR - get it from the numbering system */
    NULL, /* Pattern digit character is deprecated from CLDR - use # by default always */
    "minusSign",
    "plusSign",
    NULL, /* currency symbol - Wait until we know the currency before loading from CLDR */
    NULL, /* intl currency symbol - Wait until we know the currency before loading from CLDR */
    "currencyDecimal",
    "exponential",
    "perMille",
    NULL, /* Escape padding character - not in CLDR */
    "infinity",
    "nan",
    NULL, /* Significant digit symbol - not in CLDR */
    "currencyGroup",
    NULL, /* one digit - get it from the numbering system */
    NULL, /* two digit - get it from the numbering system */
    NULL, /* three digit - get it from the numbering system */
    NULL, /* four digit - get it from the numbering system */
    NULL, /* five digit - get it from the numbering system */
    NULL, /* six digit - get it from the numbering system */
    NULL, /* seven digit - get it from the numbering system */
    NULL, /* eight digit - get it from the numbering system */
    NULL, /* nine digit - get it from the numbering system */
    "superscriptingExponent", /* Multiplication (x) symbol for exponents */
};

// -------------------------------------
// Initializes this with the decimal format symbols in the default locale.

DecimalFormatSymbols::DecimalFormatSymbols(UErrorCode& status)
        : UObject(), locale(), currPattern(NULL) {
    initialize(locale, status, TRUE);
}

// -------------------------------------
// Initializes this with the decimal format symbols in the desired locale.

DecimalFormatSymbols::DecimalFormatSymbols(const Locale& loc, UErrorCode& status)
        : UObject(), locale(loc), currPattern(NULL) {
    initialize(locale, status);
}

DecimalFormatSymbols::DecimalFormatSymbols(const Locale& loc, const NumberingSystem& ns, UErrorCode& status)
        : UObject(), locale(loc), currPattern(NULL) {
    initialize(locale, status, FALSE, &ns);
}

DecimalFormatSymbols::DecimalFormatSymbols()
        : UObject(), locale(Locale::getRoot()), currPattern(NULL) {
    *validLocale = *actualLocale = 0;
    initialize();
}

DecimalFormatSymbols*
DecimalFormatSymbols::createWithLastResortData(UErrorCode& status) {
    if (U_FAILURE(status)) { return NULL; }
    DecimalFormatSymbols* sym = new DecimalFormatSymbols();
    if (sym == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return sym;
}

// -------------------------------------

DecimalFormatSymbols::~DecimalFormatSymbols()
{
}

// -------------------------------------
// copy constructor

DecimalFormatSymbols::DecimalFormatSymbols(const DecimalFormatSymbols &source)
    : UObject(source)
{
    *this = source;
}

// -------------------------------------
// assignment operator

DecimalFormatSymbols&
DecimalFormatSymbols::operator=(const DecimalFormatSymbols& rhs)
{
    if (this != &rhs) {
        for(int32_t i = 0; i < (int32_t)kFormatSymbolCount; ++i) {
            // fastCopyFrom is safe, see docs on fSymbols
            fSymbols[(ENumberFormatSymbol)i].fastCopyFrom(rhs.fSymbols[(ENumberFormatSymbol)i]);
        }
        for(int32_t i = 0; i < (int32_t)UNUM_CURRENCY_SPACING_COUNT; ++i) {
            currencySpcBeforeSym[i].fastCopyFrom(rhs.currencySpcBeforeSym[i]);
            currencySpcAfterSym[i].fastCopyFrom(rhs.currencySpcAfterSym[i]);
        }
        locale = rhs.locale;
        uprv_strcpy(validLocale, rhs.validLocale);
        uprv_strcpy(actualLocale, rhs.actualLocale);
        fIsCustomCurrencySymbol = rhs.fIsCustomCurrencySymbol;
        fIsCustomIntlCurrencySymbol = rhs.fIsCustomIntlCurrencySymbol;
        fCodePointZero = rhs.fCodePointZero;
    }
    return *this;
}

// -------------------------------------

UBool
DecimalFormatSymbols::operator==(const DecimalFormatSymbols& that) const
{
    if (this == &that) {
        return TRUE;
    }
    if (fIsCustomCurrencySymbol != that.fIsCustomCurrencySymbol) {
        return FALSE;
    }
    if (fIsCustomIntlCurrencySymbol != that.fIsCustomIntlCurrencySymbol) {
        return FALSE;
    }
    for(int32_t i = 0; i < (int32_t)kFormatSymbolCount; ++i) {
        if(fSymbols[(ENumberFormatSymbol)i] != that.fSymbols[(ENumberFormatSymbol)i]) {
            return FALSE;
        }
    }
    for(int32_t i = 0; i < (int32_t)UNUM_CURRENCY_SPACING_COUNT; ++i) {
        if(currencySpcBeforeSym[i] != that.currencySpcBeforeSym[i]) {
            return FALSE;
        }
        if(currencySpcAfterSym[i] != that.currencySpcAfterSym[i]) {
            return FALSE;
        }
    }
    // No need to check fCodePointZero since it is based on fSymbols
    return locale == that.locale &&
        uprv_strcmp(validLocale, that.validLocale) == 0 &&
        uprv_strcmp(actualLocale, that.actualLocale) == 0;
}

// -------------------------------------

namespace {

/**
 * Sink for enumerating all of the decimal format symbols (more specifically, anything
 * under the "NumberElements.symbols" tree).
 *
 * More specific bundles (en_GB) are enumerated before their parents (en_001, en, root):
 * Only store a value if it is still missing, that is, it has not been overridden.
 */
struct DecFmtSymDataSink : public ResourceSink {

    // Destination for data, modified via setters.
    DecimalFormatSymbols& dfs;
    // Boolean array of whether or not we have seen a particular symbol yet.
    // Can't simpy check fSymbols because it is pre-populated with defaults.
    UBool seenSymbol[DecimalFormatSymbols::kFormatSymbolCount];

    // Constructor/Destructor
    DecFmtSymDataSink(DecimalFormatSymbols& _dfs) : dfs(_dfs) {
        uprv_memset(seenSymbol, FALSE, sizeof(seenSymbol));
    }
    virtual ~DecFmtSymDataSink();

    virtual void put(const char *key, ResourceValue &value, UBool /*noFallback*/,
            UErrorCode &errorCode) {
        ResourceTable symbolsTable = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }
        for (int32_t j = 0; symbolsTable.getKeyAndValue(j, key, value); ++j) {
            for (int32_t i=0; i<DecimalFormatSymbols::kFormatSymbolCount; i++) {
                if (gNumberElementKeys[i] != NULL && uprv_strcmp(key, gNumberElementKeys[i]) == 0) {
                    if (!seenSymbol[i]) {
                        seenSymbol[i] = TRUE;
                        dfs.setSymbol(
                            (DecimalFormatSymbols::ENumberFormatSymbol) i,
                            value.getUnicodeString(errorCode));
                        if (U_FAILURE(errorCode)) { return; }
                    }
                    break;
                }
            }
        }
    }

    // Returns true if all the symbols have been seen.
    UBool seenAll() {
        for (int32_t i=0; i<DecimalFormatSymbols::kFormatSymbolCount; i++) {
            if (!seenSymbol[i]) {
                return FALSE;
            }
        }
        return TRUE;
    }

    // If monetary decimal or grouping were not explicitly set, then set them to be the
    // same as their non-monetary counterparts.
    void resolveMissingMonetarySeparators(const UnicodeString* fSymbols) {
        if (!seenSymbol[DecimalFormatSymbols::kMonetarySeparatorSymbol]) {
            dfs.setSymbol(
                DecimalFormatSymbols::kMonetarySeparatorSymbol,
                fSymbols[DecimalFormatSymbols::kDecimalSeparatorSymbol]);
        }
        if (!seenSymbol[DecimalFormatSymbols::kMonetaryGroupingSeparatorSymbol]) {
            dfs.setSymbol(
                DecimalFormatSymbols::kMonetaryGroupingSeparatorSymbol,
                fSymbols[DecimalFormatSymbols::kGroupingSeparatorSymbol]);
        }
    }
};

struct CurrencySpacingSink : public ResourceSink {
    DecimalFormatSymbols& dfs;
    UBool hasBeforeCurrency;
    UBool hasAfterCurrency;

    CurrencySpacingSink(DecimalFormatSymbols& _dfs)
        : dfs(_dfs), hasBeforeCurrency(FALSE), hasAfterCurrency(FALSE) {}
    virtual ~CurrencySpacingSink();

    virtual void put(const char *key, ResourceValue &value, UBool /*noFallback*/,
            UErrorCode &errorCode) {
        ResourceTable spacingTypesTable = value.getTable(errorCode);
        for (int32_t i = 0; spacingTypesTable.getKeyAndValue(i, key, value); ++i) {
            UBool beforeCurrency;
            if (uprv_strcmp(key, dcfmtsym_gBeforeCurrencyTag) == 0) {
                beforeCurrency = TRUE;
                hasBeforeCurrency = TRUE;
            } else if (uprv_strcmp(key, dcfmtsym_gAfterCurrencyTag) == 0) {
                beforeCurrency = FALSE;
                hasAfterCurrency = TRUE;
            } else {
                continue;
            }

            ResourceTable patternsTable = value.getTable(errorCode);
            for (int32_t j = 0; patternsTable.getKeyAndValue(j, key, value); ++j) {
                UCurrencySpacing pattern;
                if (uprv_strcmp(key, dcfmtsym_gCurrencyMatchTag) == 0) {
                    pattern = UNUM_CURRENCY_MATCH;
                } else if (uprv_strcmp(key, dcfmtsym_gCurrencySudMatchTag) == 0) {
                    pattern = UNUM_CURRENCY_SURROUNDING_MATCH;
                } else if (uprv_strcmp(key, dcfmtsym_gCurrencyInsertBtnTag) == 0) {
                    pattern = UNUM_CURRENCY_INSERT;
                } else {
                    continue;
                }

                const UnicodeString& current = dfs.getPatternForCurrencySpacing(
                    pattern, beforeCurrency, errorCode);
                if (current.isEmpty()) {
                    dfs.setPatternForCurrencySpacing(
                        pattern, beforeCurrency, value.getUnicodeString(errorCode));
                }
            }
        }
    }

    void resolveMissing() {
        // For consistency with Java, this method overwrites everything with the defaults unless
        // both beforeCurrency and afterCurrency were found in CLDR.
        static const char* defaults[] = { "[:letter:]", "[:digit:]", " " };
        if (!hasBeforeCurrency || !hasAfterCurrency) {
            for (UBool beforeCurrency = 0; beforeCurrency <= TRUE; beforeCurrency++) {
                for (int32_t pattern = 0; pattern < UNUM_CURRENCY_SPACING_COUNT; pattern++) {
                    dfs.setPatternForCurrencySpacing((UCurrencySpacing)pattern,
                        beforeCurrency, UnicodeString(defaults[pattern], -1, US_INV));
                }
            }
        }
    }
};

// Virtual destructors must be defined out of line.
DecFmtSymDataSink::~DecFmtSymDataSink() {}
CurrencySpacingSink::~CurrencySpacingSink() {}

} // namespace

void
DecimalFormatSymbols::initialize(const Locale& loc, UErrorCode& status,
    UBool useLastResortData, const NumberingSystem* ns)
{
    if (U_FAILURE(status)) { return; }
    *validLocale = *actualLocale = 0;

    // First initialize all the symbols to the fallbacks for anything we can't find
    initialize();

    //
    // Next get the numbering system for this locale and set zero digit
    // and the digit string based on the numbering system for the locale
    //
    LocalPointer<NumberingSystem> nsLocal;
    if (ns == nullptr) {
        // Use the numbering system according to the locale.
        // Save it into a LocalPointer so it gets cleaned up.
        nsLocal.adoptInstead(NumberingSystem::createInstance(loc, status));
        ns = nsLocal.getAlias();
    }
    const char *nsName;
    if (U_SUCCESS(status) && ns->getRadix() == 10 && !ns->isAlgorithmic()) {
        nsName = ns->getName();
        UnicodeString digitString(ns->getDescription());
        int32_t digitIndex = 0;
        UChar32 digit = digitString.char32At(0);
        fSymbols[kZeroDigitSymbol].setTo(digit);
        for (int32_t i = kOneDigitSymbol; i <= kNineDigitSymbol; ++i) {
            digitIndex += U16_LENGTH(digit);
            digit = digitString.char32At(digitIndex);
            fSymbols[i].setTo(digit);
        }
    } else {
        nsName = dcfmtsym_gLatn;
    }

    // Open resource bundles
    const char* locStr = loc.getName();
    LocalUResourceBundlePointer resource(ures_open(NULL, locStr, &status));
    LocalUResourceBundlePointer numberElementsRes(
        ures_getByKeyWithFallback(resource.getAlias(), dcfmtsym_gNumberElements, NULL, &status));

    if (U_FAILURE(status)) {
        if ( useLastResortData ) {
            status = U_USING_DEFAULT_WARNING;
            initialize();
        }
        return;
    }

    // Set locale IDs
    // TODO: Is there a way to do this without depending on the resource bundle instance?
    U_LOCALE_BASED(locBased, *this);
    locBased.setLocaleIDs(
        ures_getLocaleByType(
            numberElementsRes.getAlias(),
            ULOC_VALID_LOCALE, &status),
        ures_getLocaleByType(
            numberElementsRes.getAlias(),
            ULOC_ACTUAL_LOCALE, &status));

    // Now load the rest of the data from the data sink.
    // Start with loading this nsName if it is not Latin.
    DecFmtSymDataSink sink(*this);
    if (uprv_strcmp(nsName, dcfmtsym_gLatn) != 0) {
        CharString path;
        path.append(dcfmtsym_gNumberElements, status)
            .append('/', status)
            .append(nsName, status)
            .append('/', status)
            .append(dcfmtsym_gSymbols, status);
        ures_getAllItemsWithFallback(resource.getAlias(), path.data(), sink, status);

        // If no symbols exist for the given nsName and resource bundle, silently ignore
        // and fall back to Latin.
        if (status == U_MISSING_RESOURCE_ERROR) {
            status = U_ZERO_ERROR;
        } else if (U_FAILURE(status)) {
            return;
        }
    }

    // Continue with Latin if necessary.
    if (!sink.seenAll()) {
        ures_getAllItemsWithFallback(resource.getAlias(), dcfmtsym_gNumberElementsLatnSymbols, sink, status);
        if (U_FAILURE(status)) { return; }
    }

    // Let the monetary number separators equal the default number separators if necessary.
    sink.resolveMissingMonetarySeparators(fSymbols);

    // Resolve codePointZero
    UChar32 tempCodePointZero = -1;
    for (int32_t i=0; i<=9; i++) {
        const UnicodeString& stringDigit = getConstDigitSymbol(i);
        if (stringDigit.countChar32() != 1) {
            tempCodePointZero = -1;
            break;
        }
        UChar32 cp = stringDigit.char32At(0);
        if (i == 0) {
            tempCodePointZero = cp;
        } else if (cp != tempCodePointZero + i) {
            tempCodePointZero = -1;
            break;
        }
    }
    fCodePointZero = tempCodePointZero;

    // Obtain currency data from the currency API.  This is strictly
    // for backward compatibility; we don't use DecimalFormatSymbols
    // for currency data anymore.
    UErrorCode internalStatus = U_ZERO_ERROR; // don't propagate failures out
    UChar curriso[4];
    UnicodeString tempStr;
    int32_t currisoLength = ucurr_forLocale(locStr, curriso, UPRV_LENGTHOF(curriso), &internalStatus);
    if (U_SUCCESS(internalStatus) && currisoLength == 3) {
        uprv_getStaticCurrencyName(curriso, locStr, tempStr, internalStatus);
        if (U_SUCCESS(internalStatus)) {
            fSymbols[kIntlCurrencySymbol].setTo(curriso, currisoLength);
            fSymbols[kCurrencySymbol] = tempStr;
        }
    }
    /* else use the default values. */

    //load the currency data
    UChar ucc[4]={0}; //Currency Codes are always 3 chars long
    int32_t uccLen = 4;
    const char* locName = loc.getName();
    UErrorCode localStatus = U_ZERO_ERROR;
    uccLen = ucurr_forLocale(locName, ucc, uccLen, &localStatus);

    // TODO: Currency pattern data loading is duplicated in number_formatimpl.cpp
    if(U_SUCCESS(localStatus) && uccLen > 0) {
        char cc[4]={0};
        u_UCharsToChars(ucc, cc, uccLen);
        /* An explicit currency was requested */
        LocalUResourceBundlePointer currencyResource(ures_open(U_ICUDATA_CURR, locStr, &localStatus));
        LocalUResourceBundlePointer currency(
            ures_getByKeyWithFallback(currencyResource.getAlias(), "Currencies", NULL, &localStatus));
        ures_getByKeyWithFallback(currency.getAlias(), cc, currency.getAlias(), &localStatus);
        if(U_SUCCESS(localStatus) && ures_getSize(currency.getAlias())>2) { // the length is 3 if more data is present
            ures_getByIndex(currency.getAlias(), 2, currency.getAlias(), &localStatus);
            int32_t currPatternLen = 0;
            currPattern =
                ures_getStringByIndex(currency.getAlias(), (int32_t)0, &currPatternLen, &localStatus);
            UnicodeString decimalSep =
                ures_getUnicodeStringByIndex(currency.getAlias(), (int32_t)1, &localStatus);
            UnicodeString groupingSep =
                ures_getUnicodeStringByIndex(currency.getAlias(), (int32_t)2, &localStatus);
            if(U_SUCCESS(localStatus)){
                fSymbols[kMonetaryGroupingSeparatorSymbol] = groupingSep;
                fSymbols[kMonetarySeparatorSymbol] = decimalSep;
                //pattern.setTo(TRUE, currPattern, currPatternLen);
                status = localStatus;
            }
        }
        /* else An explicit currency was requested and is unknown or locale data is malformed. */
        /* ucurr_* API will get the correct value later on. */
    }
        // else ignore the error if no currency

    // Currency Spacing.
    LocalUResourceBundlePointer currencyResource(ures_open(U_ICUDATA_CURR, locStr, &status));
    CurrencySpacingSink currencySink(*this);
    ures_getAllItemsWithFallback(currencyResource.getAlias(), dcfmtsym_gCurrencySpacingTag, currencySink, status);
    currencySink.resolveMissing();
    if (U_FAILURE(status)) { return; }
}

void
DecimalFormatSymbols::initialize() {
    /*
     * These strings used to be in static arrays, but the HP/UX aCC compiler
     * cannot initialize a static array with class constructors.
     *  markus 2000may25
     */
    fSymbols[kDecimalSeparatorSymbol] = (UChar)0x2e;    // '.' decimal separator
    fSymbols[kGroupingSeparatorSymbol].remove();        //     group (thousands) separator
    fSymbols[kPatternSeparatorSymbol] = (UChar)0x3b;    // ';' pattern separator
    fSymbols[kPercentSymbol] = (UChar)0x25;             // '%' percent sign
    fSymbols[kZeroDigitSymbol] = (UChar)0x30;           // '0' native 0 digit
    fSymbols[kOneDigitSymbol] = (UChar)0x31;            // '1' native 1 digit
    fSymbols[kTwoDigitSymbol] = (UChar)0x32;            // '2' native 2 digit
    fSymbols[kThreeDigitSymbol] = (UChar)0x33;          // '3' native 3 digit
    fSymbols[kFourDigitSymbol] = (UChar)0x34;           // '4' native 4 digit
    fSymbols[kFiveDigitSymbol] = (UChar)0x35;           // '5' native 5 digit
    fSymbols[kSixDigitSymbol] = (UChar)0x36;            // '6' native 6 digit
    fSymbols[kSevenDigitSymbol] = (UChar)0x37;          // '7' native 7 digit
    fSymbols[kEightDigitSymbol] = (UChar)0x38;          // '8' native 8 digit
    fSymbols[kNineDigitSymbol] = (UChar)0x39;           // '9' native 9 digit
    fSymbols[kDigitSymbol] = (UChar)0x23;               // '#' pattern digit
    fSymbols[kPlusSignSymbol] = (UChar)0x002b;          // '+' plus sign
    fSymbols[kMinusSignSymbol] = (UChar)0x2d;           // '-' minus sign
    fSymbols[kCurrencySymbol] = (UChar)0xa4;            // 'OX' currency symbol
    fSymbols[kIntlCurrencySymbol].setTo(TRUE, INTL_CURRENCY_SYMBOL_STR, 2);
    fSymbols[kMonetarySeparatorSymbol] = (UChar)0x2e;   // '.' monetary decimal separator
    fSymbols[kExponentialSymbol] = (UChar)0x45;         // 'E' exponential
    fSymbols[kPerMillSymbol] = (UChar)0x2030;           // '%o' per mill
    fSymbols[kPadEscapeSymbol] = (UChar)0x2a;           // '*' pad escape symbol
    fSymbols[kInfinitySymbol] = (UChar)0x221e;          // 'oo' infinite
    fSymbols[kNaNSymbol] = (UChar)0xfffd;               // SUB NaN
    fSymbols[kSignificantDigitSymbol] = (UChar)0x0040;  // '@' significant digit
    fSymbols[kMonetaryGroupingSeparatorSymbol].remove(); //
    fSymbols[kExponentMultiplicationSymbol] = (UChar)0xd7; // 'x' multiplication symbol for exponents
    fIsCustomCurrencySymbol = FALSE;
    fIsCustomIntlCurrencySymbol = FALSE;
    fCodePointZero = 0x30;
    U_ASSERT(fCodePointZero == fSymbols[kZeroDigitSymbol].char32At(0));

}

Locale
DecimalFormatSymbols::getLocale(ULocDataLocaleType type, UErrorCode& status) const {
    U_LOCALE_BASED(locBased, *this);
    return locBased.getLocale(type, status);
}

const UnicodeString&
DecimalFormatSymbols::getPatternForCurrencySpacing(UCurrencySpacing type,
                                                 UBool beforeCurrency,
                                                 UErrorCode& status) const {
    if (U_FAILURE(status)) {
      return fNoSymbol;  // always empty.
    }
    if (beforeCurrency) {
      return currencySpcBeforeSym[(int32_t)type];
    } else {
      return currencySpcAfterSym[(int32_t)type];
    }
}

void
DecimalFormatSymbols::setPatternForCurrencySpacing(UCurrencySpacing type,
                                                   UBool beforeCurrency,
                                             const UnicodeString& pattern) {
  if (beforeCurrency) {
    currencySpcBeforeSym[(int32_t)type] = pattern;
  } else {
    currencySpcAfterSym[(int32_t)type] =  pattern;
  }
}
U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/* ------------------------------------------------------------------ */
/* Decimal Context module                                             */
/* ------------------------------------------------------------------ */
/* Copyright (c) IBM Corporation, 2000-2012.  All rights reserved.    */
/*                                                                    */
/* This software is made available under the terms of the             */
/* ICU License -- ICU 1.8.1 and later.                                */
/*                                                                    */
/* The description and User's Guide ("The decNumber C Library") for   */
/* this software is called decNumber.pdf.  This document is           */
/* available, together with arithmetic and format specifications,     */
/* testcases, and Web links, on the General Decimal Arithmetic page.  */
/*                                                                    */
/* Please send comments, suggestions, and corrections to the author:  */
/*   mfc@uk.ibm.com                                                   */
/*   Mike Cowlishaw, IBM Fellow                                       */
/*   IBM UK, PO Box 31, Birmingham Road, Warwick CV34 5JL, UK         */
/* ------------------------------------------------------------------ */
/* This module comprises the routines for handling arithmetic         */
/* context structures.                                                */
/* ------------------------------------------------------------------ */

#include <string.h>           /* for strcmp  */
#include <stdio.h>            /* for printf if DECCHECK  */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/* ------------------------------------------------------------------ */
/* Decimal Context module header                                      */
/* ------------------------------------------------------------------ */
/* Copyright (c) IBM Corporation, 2000-2011.   All rights reserved.   */
/*                                                                    */
/* This software is made available under the terms of the             */
/* ICU License -- ICU 1.8.1 and later.                                */
/*                                                                    */
/* The description and User's Guide ("The decNumber C Library") for   */
/* this software is called decNumber.pdf.  This document is           */
/* available, together with arithmetic and format specifications,     */
/* testcases, and Web links, on the General Decimal Arithmetic page.  */
/*                                                                    */
/* Please send comments, suggestions, and corrections to the author:  */
/*   mfc@uk.ibm.com                                                   */
/*   Mike Cowlishaw, IBM Fellow                                       */
/*   IBM UK, PO Box 31, Birmingham Road, Warwick CV34 5JL, UK         */
/* ------------------------------------------------------------------ */

/* Modified version, for use from within ICU.
 *    Renamed public functions, to avoid an unwanted export of the
 *    standard names from the ICU library.
 *
 *    Use ICU's uprv_malloc() and uprv_free()
 *
 *    Revert comment syntax to plain C
 *
 *    Remove a few compiler warnings.
 */



/*                                                                    */
/* Context variables must always have valid values:                   */
/*                                                                    */
/*  status   -- [any bits may be cleared, but not set, by user]       */
/*  round    -- must be one of the enumerated rounding modes          */
/*                                                                    */
/* The following variables are implied for fixed size formats (i.e.,  */
/* they are ignored) but should still be set correctly in case used   */
/* with decNumber functions:                                          */
/*                                                                    */
/*  clamp    -- must be either 0 or 1                                 */
/*  digits   -- must be in the range 1 through 999999999              */
/*  emax     -- must be in the range 0 through 999999999              */
/*  emin     -- must be in the range 0 through -999999999             */
/*  extended -- must be either 0 or 1 [present only if DECSUBSET]     */
/*  traps    -- only defined bits may be set                          */
/*                                                                    */
/* ------------------------------------------------------------------ */

  #define DECCNAME     "decContext"                     /* Short name */
  #define DECCFULLNAME "Decimal Context Descriptor"   /* Verbose name */
  #define DECCAUTHOR   "Mike Cowlishaw"               /* Who to blame */

  #if !defined(int32_t)
/* #include <stdint.h>   */         /* C99 standard integers           */
  #endif
#include <stdio.h>               /* for printf, etc.                */
#include <signal.h>              /* for traps                       */

  /* Extended flags setting -- set this to 0 to use only IEEE flags   */
  #if !defined(DECEXTFLAG)
  #define DECEXTFLAG 1             /* 1=enable extended flags         */
  #endif

  /* Conditional code flag -- set this to 0 for best performance      */
  #if !defined(DECSUBSET)
  #define DECSUBSET  0             /* 1=enable subset arithmetic      */
  #endif

  /* Context for operations, with associated constants                */
  enum rounding {
    DEC_ROUND_CEILING,             /* round towards +infinity         */
    DEC_ROUND_UP,                  /* round away from 0               */
    DEC_ROUND_HALF_UP,             /* 0.5 rounds up                   */
    DEC_ROUND_HALF_EVEN,           /* 0.5 rounds to nearest even      */
    DEC_ROUND_HALF_DOWN,           /* 0.5 rounds down                 */
    DEC_ROUND_DOWN,                /* round towards 0 (truncate)      */
    DEC_ROUND_FLOOR,               /* round towards -infinity         */
    DEC_ROUND_05UP,                /* round for reround               */
    DEC_ROUND_MAX                  /* enum must be less than this     */
    };
  #define DEC_ROUND_DEFAULT DEC_ROUND_HALF_EVEN;

  typedef struct {
    int32_t  digits;               /* working precision               */
    int32_t  emax;                 /* maximum positive exponent       */
    int32_t  emin;                 /* minimum negative exponent       */
    enum     rounding round;       /* rounding mode                   */
    uint32_t traps;                /* trap-enabler flags              */
    uint32_t status;               /* status flags                    */
    uint8_t  clamp;                /* flag: apply IEEE exponent clamp */
    #if DECSUBSET
    uint8_t  extended;             /* flag: special-values allowed    */
    #endif
    } decContext;

  /* Maxima and Minima for context settings                           */
  #define DEC_MAX_DIGITS 999999999
  #define DEC_MIN_DIGITS         1
  #define DEC_MAX_EMAX   999999999
  #define DEC_MIN_EMAX           0
  #define DEC_MAX_EMIN           0
  #define DEC_MIN_EMIN  -999999999
  #define DEC_MAX_MATH      999999 /* max emax, etc., for math funcs. */

  /* Classifications for decimal numbers, aligned with 754 (note that */
  /* 'normal' and 'subnormal' are meaningful only with a decContext   */
  /* or a fixed size format).                                         */
  enum decClass {
    DEC_CLASS_SNAN,
    DEC_CLASS_QNAN,
    DEC_CLASS_NEG_INF,
    DEC_CLASS_NEG_NORMAL,
    DEC_CLASS_NEG_SUBNORMAL,
    DEC_CLASS_NEG_ZERO,
    DEC_CLASS_POS_ZERO,
    DEC_CLASS_POS_SUBNORMAL,
    DEC_CLASS_POS_NORMAL,
    DEC_CLASS_POS_INF
    };
  /* Strings for the decClasses */
  #define DEC_ClassString_SN  "sNaN"
  #define DEC_ClassString_QN  "NaN"
  #define DEC_ClassString_NI  "-Infinity"
  #define DEC_ClassString_NN  "-Normal"
  #define DEC_ClassString_NS  "-Subnormal"
  #define DEC_ClassString_NZ  "-Zero"
  #define DEC_ClassString_PZ  "+Zero"
  #define DEC_ClassString_PS  "+Subnormal"
  #define DEC_ClassString_PN  "+Normal"
  #define DEC_ClassString_PI  "+Infinity"
  #define DEC_ClassString_UN  "Invalid"

  /* Trap-enabler and Status flags (exceptional conditions), and      */
  /* their names.  The top byte is reserved for internal use          */
  #if DECEXTFLAG
    /* Extended flags */
    #define DEC_Conversion_syntax    0x00000001
    #define DEC_Division_by_zero     0x00000002
    #define DEC_Division_impossible  0x00000004
    #define DEC_Division_undefined   0x00000008
    #define DEC_Insufficient_storage 0x00000010 /* [when malloc fails]  */
    #define DEC_Inexact              0x00000020
    #define DEC_Invalid_context      0x00000040
    #define DEC_Invalid_operation    0x00000080
    #if DECSUBSET
    #define DEC_Lost_digits          0x00000100
    #endif
    #define DEC_Overflow             0x00000200
    #define DEC_Clamped              0x00000400
    #define DEC_Rounded              0x00000800
    #define DEC_Subnormal            0x00001000
    #define DEC_Underflow            0x00002000
  #else
    /* IEEE flags only */
    #define DEC_Conversion_syntax    0x00000010
    #define DEC_Division_by_zero     0x00000002
    #define DEC_Division_impossible  0x00000010
    #define DEC_Division_undefined   0x00000010
    #define DEC_Insufficient_storage 0x00000010 /* [when malloc fails]  */
    #define DEC_Inexact              0x00000001
    #define DEC_Invalid_context      0x00000010
    #define DEC_Invalid_operation    0x00000010
    #if DECSUBSET
    #define DEC_Lost_digits          0x00000000
    #endif
    #define DEC_Overflow             0x00000008
    #define DEC_Clamped              0x00000000
    #define DEC_Rounded              0x00000000
    #define DEC_Subnormal            0x00000000
    #define DEC_Underflow            0x00000004
  #endif

  /* IEEE 754 groupings for the flags                                 */
  /* [DEC_Clamped, DEC_Lost_digits, DEC_Rounded, and DEC_Subnormal    */
  /* are not in IEEE 754]                                             */
  #define DEC_IEEE_754_Division_by_zero  (DEC_Division_by_zero)
  #if DECSUBSET
  #define DEC_IEEE_754_Inexact           (DEC_Inexact | DEC_Lost_digits)
  #else
  #define DEC_IEEE_754_Inexact           (DEC_Inexact)
  #endif
  #define DEC_IEEE_754_Invalid_operation (DEC_Conversion_syntax |     \
                                          DEC_Division_impossible |   \
                                          DEC_Division_undefined |    \
                                          DEC_Insufficient_storage |  \
                                          DEC_Invalid_context |       \
                                          DEC_Invalid_operation)
  #define DEC_IEEE_754_Overflow          (DEC_Overflow)
  #define DEC_IEEE_754_Underflow         (DEC_Underflow)

  /* flags which are normally errors (result is qNaN, infinite, or 0) */
  #define DEC_Errors (DEC_IEEE_754_Division_by_zero |                 \
                      DEC_IEEE_754_Invalid_operation |                \
                      DEC_IEEE_754_Overflow | DEC_IEEE_754_Underflow)
  /* flags which cause a result to become qNaN                        */
  #define DEC_NaNs    DEC_IEEE_754_Invalid_operation

  /* flags which are normally for information only (finite results)   */
  #if DECSUBSET
  #define DEC_Information (DEC_Clamped | DEC_Rounded | DEC_Inexact    \
                          | DEC_Lost_digits)
  #else
  #define DEC_Information (DEC_Clamped | DEC_Rounded | DEC_Inexact)
  #endif

  /* IEEE 854 names (for compatibility with older decNumber versions) */
  #define DEC_IEEE_854_Division_by_zero  DEC_IEEE_754_Division_by_zero
  #define DEC_IEEE_854_Inexact           DEC_IEEE_754_Inexact
  #define DEC_IEEE_854_Invalid_operation DEC_IEEE_754_Invalid_operation
  #define DEC_IEEE_854_Overflow          DEC_IEEE_754_Overflow
  #define DEC_IEEE_854_Underflow         DEC_IEEE_754_Underflow

  /* Name strings for the exceptional conditions                      */
  #define DEC_Condition_CS "Conversion syntax"
  #define DEC_Condition_DZ "Division by zero"
  #define DEC_Condition_DI "Division impossible"
  #define DEC_Condition_DU "Division undefined"
  #define DEC_Condition_IE "Inexact"
  #define DEC_Condition_IS "Insufficient storage"
  #define DEC_Condition_IC "Invalid context"
  #define DEC_Condition_IO "Invalid operation"
  #if DECSUBSET
  #define DEC_Condition_LD "Lost digits"
  #endif
  #define DEC_Condition_OV "Overflow"
  #define DEC_Condition_PA "Clamped"
  #define DEC_Condition_RO "Rounded"
  #define DEC_Condition_SU "Subnormal"
  #define DEC_Condition_UN "Underflow"
  #define DEC_Condition_ZE "No status"
  #define DEC_Condition_MU "Multiple status"
  #define DEC_Condition_Length 21  /* length of the longest string,   */
                                   /* including terminator            */

  /* Initialization descriptors, used by decContextDefault            */
  #define DEC_INIT_BASE         0
  #define DEC_INIT_DECIMAL32   32
  #define DEC_INIT_DECIMAL64   64
  #define DEC_INIT_DECIMAL128 128
  /* Synonyms */
  #define DEC_INIT_DECSINGLE  DEC_INIT_DECIMAL32
  #define DEC_INIT_DECDOUBLE  DEC_INIT_DECIMAL64
  #define DEC_INIT_DECQUAD    DEC_INIT_DECIMAL128

  /* decContext routines                                              */
  U_INTERNAL decContext  * U_EXPORT2 uprv_decContextClearStatus(decContext *, uint32_t);
  U_INTERNAL decContext  * U_EXPORT2 uprv_decContextDefault(decContext *, int32_t);
  U_INTERNAL enum rounding U_EXPORT2 uprv_decContextGetRounding(decContext *);
  U_INTERNAL uint32_t      U_EXPORT2 uprv_decContextGetStatus(decContext *);
  U_INTERNAL decContext  * U_EXPORT2 uprv_decContextRestoreStatus(decContext *, uint32_t, uint32_t);
  U_INTERNAL uint32_t      U_EXPORT2 uprv_decContextSaveStatus(decContext *, uint32_t);
  U_INTERNAL decContext  * U_EXPORT2 uprv_decContextSetRounding(decContext *, enum rounding);
  U_INTERNAL decContext  * U_EXPORT2 uprv_decContextSetStatus(decContext *, uint32_t);
  U_INTERNAL decContext  * U_EXPORT2 uprv_decContextSetStatusFromString(decContext *, const char *);
  U_INTERNAL decContext  * U_EXPORT2 uprv_decContextSetStatusFromStringQuiet(decContext *, const char *);
  U_INTERNAL decContext  * U_EXPORT2 uprv_decContextSetStatusQuiet(decContext *, uint32_t);
  U_INTERNAL const char  * U_EXPORT2 uprv_decContextStatusToString(const decContext *);
  U_INTERNAL int32_t       U_EXPORT2 uprv_decContextTestEndian(uint8_t);
  U_INTERNAL uint32_t      U_EXPORT2 uprv_decContextTestSavedStatus(uint32_t, uint32_t);
  U_INTERNAL uint32_t      U_EXPORT2 uprv_decContextTestStatus(decContext *, uint32_t);
  U_INTERNAL decContext  * U_EXPORT2 uprv_decContextZeroStatus(decContext *);

       /* context and base types  */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/* ------------------------------------------------------------------ */
/* decNumber package local type, tuning, and macro definitions        */
/* ------------------------------------------------------------------ */
/* Copyright (c) IBM Corporation, 2000-2016.   All rights reserved.   */
/*                                                                    */
/* This software is made available under the terms of the             */
/* ICU License -- ICU 1.8.1 and later.                                */
/*                                                                    */
/* The description and User's Guide ("The decNumber C Library") for   */
/* this software is called decNumber.pdf.  This document is           */
/* available, together with arithmetic and format specifications,     */
/* testcases, and Web links, on the General Decimal Arithmetic page.  */
/*                                                                    */
/* Please send comments, suggestions, and corrections to the author:  */
/*   mfc@uk.ibm.com                                                   */
/*   Mike Cowlishaw, IBM Fellow                                       */
/*   IBM UK, PO Box 31, Birmingham Road, Warwick CV34 5JL, UK         */
/* ------------------------------------------------------------------ */
/* This header file is included by all modules in the decNumber       */
/* library, and contains local type definitions, tuning parameters,   */
/* etc.  It should not need to be used by application programs.       */
/* decNumber.h or one of decDouble (etc.) must be included first.     */
/* ------------------------------------------------------------------ */

#define DECVERSION    "decNumber 3.61" /* Package Version [16 max.] */

#include <stdlib.h>         /* for abs                              */
#include <string.h>         /* for memset, strcpy                   */


  /* Conditional code flag -- set this to match hardware platform     */
  #if !defined(DECLITEND)
  #define DECLITEND 1         /* 1=little-endian, 0=big-endian        */
  #endif

  /* Conditional code flag -- set this to 1 for best performance      */
  #if !defined(DECUSE64)
  #define DECUSE64  1         /* 1=use int64s, 0=int32 & smaller only */
  #endif

  /* Conditional check flags -- set these to 0 for best performance   */
  #if !defined(DECCHECK)
  #define DECCHECK  0         /* 1 to enable robust checking          */
  #endif
  #if !defined(DECALLOC)
  #define DECALLOC  0         /* 1 to enable memory accounting        */
  #endif
  #if !defined(DECTRACE)
  #define DECTRACE  0         /* 1 to trace certain internals, etc.   */
  #endif

  /* Tuning parameter for decNumber (arbitrary precision) module      */
  #if !defined(DECBUFFER)
  #define DECBUFFER 36        /* Size basis for local buffers.  This  */
                              /* should be a common maximum precision */
                              /* rounded up to a multiple of 4; must  */
                              /* be zero or positive.                 */
  #endif

  /* ---------------------------------------------------------------- */
  /* Definitions for all modules (general-purpose)                    */
  /* ---------------------------------------------------------------- */

  /* Local names for common types -- for safety, decNumber modules do */
  /* not use int or long directly.                                    */
  #define Flag   uint8_t
  #define Byte   int8_t
  #define uByte  uint8_t
  #define Short  int16_t
  #define uShort uint16_t
  #define Int    int32_t
  #define uInt   uint32_t
  #define Unit   decNumberUnit
  #if DECUSE64
  #define Long   int64_t
  #define uLong  uint64_t
  #endif

  /* Development-use definitions                                      */
  typedef long int LI;        /* for printf arguments only            */
  #define DECNOINT  0         /* 1 to check no internal use of 'int'  */
                              /*   or stdint types                    */
  #if DECNOINT
    /* if these interfere with your C includes, do not set DECNOINT   */
    #define int     ?         /* enable to ensure that plain C 'int'  */
    #define long    ??        /* .. or 'long' types are not used      */
  #endif

  /* LONGMUL32HI -- set w=(u*v)>>32, where w, u, and v are uInts      */
  /* (that is, sets w to be the high-order word of the 64-bit result; */
  /* the low-order word is simply u*v.)                               */
  /* This version is derived from Knuth via Hacker's Delight;         */
  /* it seems to optimize better than some others tried               */
  #define LONGMUL32HI(w, u, v) {             \
    uInt u0, u1, v0, v1, w0, w1, w2, t;      \
    u0=u & 0xffff; u1=u>>16;                 \
    v0=v & 0xffff; v1=v>>16;                 \
    w0=u0*v0;                                \
    t=u1*v0 + (w0>>16);                      \
    w1=t & 0xffff; w2=t>>16;                 \
    w1=u0*v1 + w1;                           \
    (w)=u1*v1 + w2 + (w1>>16);}

  /* ROUNDUP -- round an integer up to a multiple of n                */
  #define ROUNDUP(i, n) ((((i)+(n)-1)/n)*n)
  #define ROUNDUP4(i)   (((i)+3)&~3)    /* special for n=4            */

  /* ROUNDDOWN -- round an integer down to a multiple of n            */
  #define ROUNDDOWN(i, n) (((i)/n)*n)
  #define ROUNDDOWN4(i)   ((i)&~3)      /* special for n=4            */

  /* References to multi-byte sequences under different sizes; these  */
  /* require locally declared variables, but do not violate strict    */
  /* aliasing or alignment (as did the UINTAT simple cast to uInt).   */
  /* Variables needed are uswork, uiwork, etc. [so do not use at same */
  /* level in an expression, e.g., UBTOUI(x)==UBTOUI(y) may fail].    */

  /* Return a uInt, etc., from bytes starting at a char* or uByte*    */
  #define UBTOUS(b)  (memcpy((void *)&uswork, b, 2), uswork)
  #define UBTOUI(b)  (memcpy((void *)&uiwork, b, 4), uiwork)

  /* Store a uInt, etc., into bytes starting at a char* or uByte*.    */
  /* Returns i, evaluated, for convenience; has to use uiwork because */
  /* i may be an expression.                                          */
  #define UBFROMUS(b, i)  (uswork=(i), memcpy(b, (void *)&uswork, 2), uswork)
  #define UBFROMUI(b, i)  (uiwork=(i), memcpy(b, (void *)&uiwork, 4), uiwork)

  /* X10 and X100 -- multiply integer i by 10 or 100                  */
  /* [shifts are usually faster than multiply; could be conditional]  */
  #define X10(i)  (((i)<<1)+((i)<<3))
  #define X100(i) (((i)<<2)+((i)<<5)+((i)<<6))

  /* MAXI and MINI -- general max & min (not in ANSI) for integers    */
  #define MAXI(x,y) ((x)<(y)?(y):(x))
  #define MINI(x,y) ((x)>(y)?(y):(x))

  /* Useful constants                                                 */
  #define BILLION      1000000000            /* 10**9                 */
  /* CHARMASK: 0x30303030 for ASCII/UTF8; 0xF0F0F0F0 for EBCDIC       */
  #define CHARMASK ((((((((uInt)'0')<<8)+'0')<<8)+'0')<<8)+'0')


  /* ---------------------------------------------------------------- */
  /* Definitions for arbitary-precision modules (only valid after     */
  /* decNumber.h has been included)                                   */
  /* ---------------------------------------------------------------- */

  /* Limits and constants                                             */
  #define DECNUMMAXP 999999999  /* maximum precision code can handle  */
  #define DECNUMMAXE 999999999  /* maximum adjusted exponent ditto    */
  #define DECNUMMINE -999999999 /* minimum adjusted exponent ditto    */
  #if (DECNUMMAXP != DEC_MAX_DIGITS)
    #error Maximum digits mismatch
  #endif
  #if (DECNUMMAXE != DEC_MAX_EMAX)
    #error Maximum exponent mismatch
  #endif
  #if (DECNUMMINE != DEC_MIN_EMIN)
    #error Minimum exponent mismatch
  #endif

  /* Set DECDPUNMAX -- the maximum integer that fits in DECDPUN       */
  /* digits, and D2UTABLE -- the initializer for the D2U table        */
  // #ifndef DECDPUN
    // no-op
  // #elif   DECDPUN==1
    #define DECDPUNMAX 9
    #define D2UTABLE {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,  \
                      18,19,20,21,22,23,24,25,26,27,28,29,30,31,32, \
                      33,34,35,36,37,38,39,40,41,42,43,44,45,46,47, \
                      48,49}

  /* ----- Shared data (in decNumber.c) ----- */
  /* Public lookup table used by the D2U macro (see below)            */
  #define DECMAXD2U 49
  /*extern const uByte d2utable[DECMAXD2U+1];*/

  /* ----- Macros ----- */
  /* ISZERO -- return true if decNumber dn is a zero                  */
  /* [performance-critical in some situations]                        */
  #define ISZERO(dn) decNumberIsZero(dn)     /* now just a local name */

  /* D2U -- return the number of Units needed to hold d digits        */
  /* (runtime version, with table lookaside for small d)              */
  #if defined(DECDPUN) && DECDPUN==8
    #define D2U(d) ((unsigned)((d)<=DECMAXD2U?d2utable[d]:((d)+7)>>3))
  #elif defined(DECDPUN) && DECDPUN==4
    #define D2U(d) ((unsigned)((d)<=DECMAXD2U?d2utable[d]:((d)+3)>>2))
  #else
    #define D2U(d) ((d)<=DECMAXD2U?d2utable[d]:((d)+DECDPUN-1)/DECDPUN)
  #endif
  /* SD2U -- static D2U macro (for compile-time calculation)          */
  #define SD2U(d) (((d)+DECDPUN-1)/DECDPUN)

  /* MSUDIGITS -- returns digits in msu, from digits, calculated      */
  /* using D2U                                                        */
  #define MSUDIGITS(d) ((d)-(D2U(d)-1)*DECDPUN)

  /* D2N -- return the number of decNumber structs that would be      */
  /* needed to contain that number of digits (and the initial         */
  /* decNumber struct) safely.  Note that one Unit is included in the */
  /* initial structure.  Used for allocating space that is aligned on */
  /* a decNumber struct boundary. */
  #define D2N(d) \
    ((((SD2U(d)-1)*sizeof(Unit))+sizeof(decNumber)*2-1)/sizeof(decNumber))

  /* TODIGIT -- macro to remove the leading digit from the unsigned   */
  /* integer u at column cut (counting from the right, LSD=0) and     */
  /* place it as an ASCII character into the character pointed to by  */
  /* c.  Note that cut must be <= 9, and the maximum value for u is   */
  /* 2,000,000,000 (as is needed for negative exponents of            */
  /* subnormals).  The unsigned integer pow is used as a temporary    */
  /* variable. */
  #define TODIGIT(u, cut, c, pow) UPRV_BLOCK_MACRO_BEGIN { \
    *(c)='0';                             \
    pow=DECPOWERS[cut]*2;                 \
    if ((u)>pow) {                        \
      pow*=4;                             \
      if ((u)>=pow) {(u)-=pow; *(c)+=8;}  \
      pow/=2;                             \
      if ((u)>=pow) {(u)-=pow; *(c)+=4;}  \
      pow/=2;                             \
      }                                   \
    if ((u)>=pow) {(u)-=pow; *(c)+=2;}    \
    pow/=2;                               \
    if ((u)>=pow) {(u)-=pow; *(c)+=1;}    \
    } UPRV_BLOCK_MACRO_END

  /* ---------------------------------------------------------------- */
  /* Definitions for fixed-precision modules (only valid after        */
  /* decSingle.h, decDouble.h, or decQuad.h has been included)        */
  /* ---------------------------------------------------------------- */

  /* bcdnum -- a structure describing a format-independent finite     */
  /* number, whose coefficient is a string of bcd8 uBytes             */
  typedef struct {
    uByte   *msd;             /* -> most significant digit            */
    uByte   *lsd;             /* -> least ditto                       */
    uInt     sign;            /* 0=positive, DECFLOAT_Sign=negative   */
    Int      exponent;        /* Unadjusted signed exponent (q), or   */
                              /* DECFLOAT_NaN etc. for a special      */
    } bcdnum;

  /* Test if exponent or bcdnum exponent must be a special, etc.      */
  #define EXPISSPECIAL(exp) ((exp)>=DECFLOAT_MinSp)
  #define EXPISINF(exp) (exp==DECFLOAT_Inf)
  #define EXPISNAN(exp) (exp==DECFLOAT_qNaN || exp==DECFLOAT_sNaN)
  #define NUMISSPECIAL(num) (EXPISSPECIAL((num)->exponent))

  /* Refer to a 32-bit word or byte in a decFloat (df) by big-endian  */
  /* (array) notation (the 0 word or byte contains the sign bit),     */
  /* automatically adjusting for endianness; similarly address a word */
  /* in the next-wider format (decFloatWider, or dfw)                 */
  #define DECWORDS  (DECBYTES/4)
  #define DECWWORDS (DECWBYTES/4)
  #if DECLITEND
    #define DFBYTE(df, off)   ((df)->bytes[DECBYTES-1-(off)])
    #define DFWORD(df, off)   ((df)->words[DECWORDS-1-(off)])
    #define DFWWORD(dfw, off) ((dfw)->words[DECWWORDS-1-(off)])
  #else
    #define DFBYTE(df, off)   ((df)->bytes[off])
    #define DFWORD(df, off)   ((df)->words[off])
    #define DFWWORD(dfw, off) ((dfw)->words[off])
  #endif

  /* Tests for sign or specials, directly on DECFLOATs                */
  #define DFISSIGNED(df)   (DFWORD(df, 0)&0x80000000)
  #define DFISSPECIAL(df) ((DFWORD(df, 0)&0x78000000)==0x78000000)
  #define DFISINF(df)     ((DFWORD(df, 0)&0x7c000000)==0x78000000)
  #define DFISNAN(df)     ((DFWORD(df, 0)&0x7c000000)==0x7c000000)
  #define DFISQNAN(df)    ((DFWORD(df, 0)&0x7e000000)==0x7c000000)
  #define DFISSNAN(df)    ((DFWORD(df, 0)&0x7e000000)==0x7e000000)

  /* Shared lookup tables                                             */
  extern const uInt   DECCOMBMSD[64];   /* Combination field -> MSD   */
  extern const uInt   DECCOMBFROM[48];  /* exp+msd -> Combination     */

  /* Private generic (utility) routine                                */
  #if DECCHECK || DECTRACE
    extern void decShowNum(const bcdnum *, const char *);
  #endif

// #else
//   #error decNumberLocal included more than once
   /* decNumber local types, etc.  */

#if 0  /* ICU: No need to test endianness at runtime. */
/* compile-time endian tester [assumes sizeof(Int)>1] */
static  const  Int mfcone=1;                 /* constant 1  */
static  const  Flag *mfctop=(Flag *)&mfcone; /* -> top byte  */
#define LITEND *mfctop             /* named flag; 1=little-endian  */
#endif

/* ------------------------------------------------------------------ */
/* decContextClearStatus -- clear bits in current status              */
/*                                                                    */
/*  context is the context structure to be queried                    */
/*  mask indicates the bits to be cleared (the status bit that        */
/*    corresponds to each 1 bit in the mask is cleared)               */
/*  returns context                                                   */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI decContext * U_EXPORT2 uprv_decContextClearStatus(decContext *context, uInt mask) {
  context->status&=~mask;
  return context;
  } /* decContextClearStatus  */

/* ------------------------------------------------------------------ */
/* decContextDefault -- initialize a context structure                */
/*                                                                    */
/*  context is the structure to be initialized                        */
/*  kind selects the required set of default values, one of:          */
/*      DEC_INIT_BASE       -- select ANSI X3-274 defaults            */
/*      DEC_INIT_DECIMAL32  -- select IEEE 754 defaults, 32-bit       */
/*      DEC_INIT_DECIMAL64  -- select IEEE 754 defaults, 64-bit       */
/*      DEC_INIT_DECIMAL128 -- select IEEE 754 defaults, 128-bit      */
/*      For any other value a valid context is returned, but with     */
/*      Invalid_operation set in the status field.                    */
/*  returns a context structure with the appropriate initial values.  */
/* ------------------------------------------------------------------ */
U_CAPI decContext *  U_EXPORT2 uprv_decContextDefault(decContext *context, Int kind) {
  /* set defaults...  */
  context->digits=9;                         /* 9 digits  */
  context->emax=DEC_MAX_EMAX;                /* 9-digit exponents  */
  context->emin=DEC_MIN_EMIN;                /* .. balanced  */
  context->round=DEC_ROUND_HALF_UP;          /* 0.5 rises  */
  context->traps=DEC_Errors;                 /* all but informational  */
  context->status=0;                         /* cleared  */
  context->clamp=0;                          /* no clamping  */
  #if DECSUBSET
  context->extended=0;                       /* cleared  */
  #endif
  switch (kind) {
    case DEC_INIT_BASE:
      /* [use defaults]  */
      break;
    case DEC_INIT_DECIMAL32:
      context->digits=7;                     /* digits  */
      context->emax=96;                      /* Emax  */
      context->emin=-95;                     /* Emin  */
      context->round=DEC_ROUND_HALF_EVEN;    /* 0.5 to nearest even  */
      context->traps=0;                      /* no traps set  */
      context->clamp=1;                      /* clamp exponents  */
      #if DECSUBSET
      context->extended=1;                   /* set  */
      #endif
      break;
    case DEC_INIT_DECIMAL64:
      context->digits=16;                    /* digits  */
      context->emax=384;                     /* Emax  */
      context->emin=-383;                    /* Emin  */
      context->round=DEC_ROUND_HALF_EVEN;    /* 0.5 to nearest even  */
      context->traps=0;                      /* no traps set  */
      context->clamp=1;                      /* clamp exponents  */
      #if DECSUBSET
      context->extended=1;                   /* set  */
      #endif
      break;
    case DEC_INIT_DECIMAL128:
      context->digits=34;                    /* digits  */
      context->emax=6144;                    /* Emax  */
      context->emin=-6143;                   /* Emin  */
      context->round=DEC_ROUND_HALF_EVEN;    /* 0.5 to nearest even  */
      context->traps=0;                      /* no traps set  */
      context->clamp=1;                      /* clamp exponents  */
      #if DECSUBSET
      context->extended=1;                   /* set  */
      #endif
      break;

    default:                                 /* invalid Kind  */
      /* use defaults, and ..  */
      uprv_decContextSetStatus(context, DEC_Invalid_operation); /* trap  */
    }

  return context;} /* decContextDefault  */

/* ------------------------------------------------------------------ */
/* decContextGetRounding -- return current rounding mode              */
/*                                                                    */
/*  context is the context structure to be queried                    */
/*  returns the rounding mode                                         */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI enum rounding  U_EXPORT2 uprv_decContextGetRounding(decContext *context) {
  return context->round;
  } /* decContextGetRounding  */

/* ------------------------------------------------------------------ */
/* decContextGetStatus -- return current status                       */
/*                                                                    */
/*  context is the context structure to be queried                    */
/*  returns status                                                    */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI uInt  U_EXPORT2 uprv_decContextGetStatus(decContext *context) {
  return context->status;
  } /* decContextGetStatus  */

/* ------------------------------------------------------------------ */
/* decContextRestoreStatus -- restore bits in current status          */
/*                                                                    */
/*  context is the context structure to be updated                    */
/*  newstatus is the source for the bits to be restored               */
/*  mask indicates the bits to be restored (the status bit that       */
/*    corresponds to each 1 bit in the mask is set to the value of    */
/*    the correspnding bit in newstatus)                              */
/*  returns context                                                   */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI decContext * U_EXPORT2 uprv_decContextRestoreStatus(decContext *context,
                                    uInt newstatus, uInt mask) {
  context->status&=~mask;               /* clear the selected bits  */
  context->status|=(mask&newstatus);    /* or in the new bits  */
  return context;
  } /* decContextRestoreStatus  */

/* ------------------------------------------------------------------ */
/* decContextSaveStatus -- save bits in current status                */
/*                                                                    */
/*  context is the context structure to be queried                    */
/*  mask indicates the bits to be saved (the status bits that         */
/*    correspond to each 1 bit in the mask are saved)                 */
/*  returns the AND of the mask and the current status                */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI uInt  U_EXPORT2 uprv_decContextSaveStatus(decContext *context, uInt mask) {
  return context->status&mask;
  } /* decContextSaveStatus  */

/* ------------------------------------------------------------------ */
/* decContextSetRounding -- set current rounding mode                 */
/*                                                                    */
/*  context is the context structure to be updated                    */
/*  newround is the value which will replace the current mode         */
/*  returns context                                                   */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI decContext * U_EXPORT2 uprv_decContextSetRounding(decContext *context,
                                  enum rounding newround) {
  context->round=newround;
  return context;
  } /* decContextSetRounding  */

/* ------------------------------------------------------------------ */
/* decContextSetStatus -- set status and raise trap if appropriate    */
/*                                                                    */
/*  context is the context structure to be updated                    */
/*  status  is the DEC_ exception code                                */
/*  returns the context structure                                     */
/*                                                                    */
/* Control may never return from this routine, if there is a signal   */
/* handler and it takes a long jump.                                  */
/* ------------------------------------------------------------------ */
U_CAPI decContext *  U_EXPORT2 uprv_decContextSetStatus(decContext *context, uInt status) {
  context->status|=status;
#if 0  /* ICU: Do not raise signals. */
  if (status & context->traps) raise(SIGFPE);
#endif
  return context;} /* decContextSetStatus  */

/* ------------------------------------------------------------------ */
/* decContextSetStatusFromString -- set status from a string + trap   */
/*                                                                    */
/*  context is the context structure to be updated                    */
/*  string is a string exactly equal to one that might be returned    */
/*            by decContextStatusToString                             */
/*                                                                    */
/*  The status bit corresponding to the string is set, and a trap     */
/*  is raised if appropriate.                                         */
/*                                                                    */
/*  returns the context structure, unless the string is equal to      */
/*    DEC_Condition_MU or is not recognized.  In these cases NULL is  */
/*    returned.                                                       */
/* ------------------------------------------------------------------ */
U_CAPI decContext *  U_EXPORT2 uprv_decContextSetStatusFromString(decContext *context,
                                           const char *string) {
  if (strcmp(string, DEC_Condition_CS)==0)
    return uprv_decContextSetStatus(context, DEC_Conversion_syntax);
  if (strcmp(string, DEC_Condition_DZ)==0)
    return uprv_decContextSetStatus(context, DEC_Division_by_zero);
  if (strcmp(string, DEC_Condition_DI)==0)
    return uprv_decContextSetStatus(context, DEC_Division_impossible);
  if (strcmp(string, DEC_Condition_DU)==0)
    return uprv_decContextSetStatus(context, DEC_Division_undefined);
  if (strcmp(string, DEC_Condition_IE)==0)
    return uprv_decContextSetStatus(context, DEC_Inexact);
  if (strcmp(string, DEC_Condition_IS)==0)
    return uprv_decContextSetStatus(context, DEC_Insufficient_storage);
  if (strcmp(string, DEC_Condition_IC)==0)
    return uprv_decContextSetStatus(context, DEC_Invalid_context);
  if (strcmp(string, DEC_Condition_IO)==0)
    return uprv_decContextSetStatus(context, DEC_Invalid_operation);
  #if DECSUBSET
  if (strcmp(string, DEC_Condition_LD)==0)
    return uprv_decContextSetStatus(context, DEC_Lost_digits);
  #endif
  if (strcmp(string, DEC_Condition_OV)==0)
    return uprv_decContextSetStatus(context, DEC_Overflow);
  if (strcmp(string, DEC_Condition_PA)==0)
    return uprv_decContextSetStatus(context, DEC_Clamped);
  if (strcmp(string, DEC_Condition_RO)==0)
    return uprv_decContextSetStatus(context, DEC_Rounded);
  if (strcmp(string, DEC_Condition_SU)==0)
    return uprv_decContextSetStatus(context, DEC_Subnormal);
  if (strcmp(string, DEC_Condition_UN)==0)
    return uprv_decContextSetStatus(context, DEC_Underflow);
  if (strcmp(string, DEC_Condition_ZE)==0)
    return context;
  return NULL;  /* Multiple status, or unknown  */
  } /* decContextSetStatusFromString  */

/* ------------------------------------------------------------------ */
/* decContextSetStatusFromStringQuiet -- set status from a string     */
/*                                                                    */
/*  context is the context structure to be updated                    */
/*  string is a string exactly equal to one that might be returned    */
/*            by decContextStatusToString                             */
/*                                                                    */
/*  The status bit corresponding to the string is set; no trap is     */
/*  raised.                                                           */
/*                                                                    */
/*  returns the context structure, unless the string is equal to      */
/*    DEC_Condition_MU or is not recognized.  In these cases NULL is  */
/*    returned.                                                       */
/* ------------------------------------------------------------------ */
U_CAPI decContext *  U_EXPORT2 uprv_decContextSetStatusFromStringQuiet(decContext *context,
                                                const char *string) {
  if (strcmp(string, DEC_Condition_CS)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Conversion_syntax);
  if (strcmp(string, DEC_Condition_DZ)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Division_by_zero);
  if (strcmp(string, DEC_Condition_DI)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Division_impossible);
  if (strcmp(string, DEC_Condition_DU)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Division_undefined);
  if (strcmp(string, DEC_Condition_IE)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Inexact);
  if (strcmp(string, DEC_Condition_IS)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Insufficient_storage);
  if (strcmp(string, DEC_Condition_IC)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Invalid_context);
  if (strcmp(string, DEC_Condition_IO)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Invalid_operation);
  #if DECSUBSET
  if (strcmp(string, DEC_Condition_LD)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Lost_digits);
  #endif
  if (strcmp(string, DEC_Condition_OV)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Overflow);
  if (strcmp(string, DEC_Condition_PA)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Clamped);
  if (strcmp(string, DEC_Condition_RO)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Rounded);
  if (strcmp(string, DEC_Condition_SU)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Subnormal);
  if (strcmp(string, DEC_Condition_UN)==0)
    return uprv_decContextSetStatusQuiet(context, DEC_Underflow);
  if (strcmp(string, DEC_Condition_ZE)==0)
    return context;
  return NULL;  /* Multiple status, or unknown  */
  } /* decContextSetStatusFromStringQuiet  */

/* ------------------------------------------------------------------ */
/* decContextSetStatusQuiet -- set status without trap                */
/*                                                                    */
/*  context is the context structure to be updated                    */
/*  status  is the DEC_ exception code                                */
/*  returns the context structure                                     */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI decContext *  U_EXPORT2 uprv_decContextSetStatusQuiet(decContext *context, uInt status) {
  context->status|=status;
  return context;} /* decContextSetStatusQuiet  */

/* ------------------------------------------------------------------ */
/* decContextStatusToString -- convert status flags to a string       */
/*                                                                    */
/*  context is a context with valid status field                      */
/*                                                                    */
/*  returns a constant string describing the condition.  If multiple  */
/*    (or no) flags are set, a generic constant message is returned.  */
/* ------------------------------------------------------------------ */
U_CAPI const char * U_EXPORT2 uprv_decContextStatusToString(const decContext *context) {
  Int status=context->status;

  /* test the five IEEE first, as some of the others are ambiguous when  */
  /* DECEXTFLAG=0  */
  if (status==DEC_Invalid_operation    ) return DEC_Condition_IO;
  if (status==DEC_Division_by_zero     ) return DEC_Condition_DZ;
  if (status==DEC_Overflow             ) return DEC_Condition_OV;
  if (status==DEC_Underflow            ) return DEC_Condition_UN;
  if (status==DEC_Inexact              ) return DEC_Condition_IE;

  if (status==DEC_Division_impossible  ) return DEC_Condition_DI;
  if (status==DEC_Division_undefined   ) return DEC_Condition_DU;
  if (status==DEC_Rounded              ) return DEC_Condition_RO;
  if (status==DEC_Clamped              ) return DEC_Condition_PA;
  if (status==DEC_Subnormal            ) return DEC_Condition_SU;
  if (status==DEC_Conversion_syntax    ) return DEC_Condition_CS;
  if (status==DEC_Insufficient_storage ) return DEC_Condition_IS;
  if (status==DEC_Invalid_context      ) return DEC_Condition_IC;
  #if DECSUBSET
  if (status==DEC_Lost_digits          ) return DEC_Condition_LD;
  #endif
  if (status==0                        ) return DEC_Condition_ZE;
  return DEC_Condition_MU;  /* Multiple errors  */
  } /* decContextStatusToString  */

/* ------------------------------------------------------------------ */
/* decContextTestEndian -- test whether DECLITEND is set correctly    */
/*                                                                    */
/*  quiet is 1 to suppress message; 0 otherwise                       */
/*  returns 0 if DECLITEND is correct                                 */
/*          1 if DECLITEND is incorrect and should be 1               */
/*         -1 if DECLITEND is incorrect and should be 0               */
/*                                                                    */
/* A message is displayed if the return value is not 0 and quiet==0.  */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
#if 0  /* ICU: Unused function. Anyway, do not call printf(). */
U_CAPI Int  U_EXPORT2 uprv_decContextTestEndian(Flag quiet) {
  Int res=0;                  /* optimist  */
  uInt dle=(uInt)DECLITEND;   /* unsign  */
  if (dle>1) dle=1;           /* ensure 0 or 1  */

  if (LITEND!=DECLITEND) {
    const char *adj;
    if (!quiet) {
      if (LITEND) adj="little";
             else adj="big";
      printf("Warning: DECLITEND is set to %d, but this computer appears to be %s-endian\n",
             DECLITEND, adj);
      }
    res=(Int)LITEND-dle;
    }
  return res;
  } /* decContextTestEndian  */
#endif

/* ------------------------------------------------------------------ */
/* decContextTestSavedStatus -- test bits in saved status             */
/*                                                                    */
/*  oldstatus is the status word to be tested                         */
/*  mask indicates the bits to be tested (the oldstatus bits that     */
/*    correspond to each 1 bit in the mask are tested)                */
/*  returns 1 if any of the tested bits are 1, or 0 otherwise         */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI  uInt U_EXPORT2 uprv_decContextTestSavedStatus(uInt oldstatus, uInt mask) {
  return (oldstatus&mask)!=0;
  } /* decContextTestSavedStatus  */

/* ------------------------------------------------------------------ */
/* decContextTestStatus -- test bits in current status                */
/*                                                                    */
/*  context is the context structure to be updated                    */
/*  mask indicates the bits to be tested (the status bits that        */
/*    correspond to each 1 bit in the mask are tested)                */
/*  returns 1 if any of the tested bits are 1, or 0 otherwise         */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI uInt  U_EXPORT2 uprv_decContextTestStatus(decContext *context, uInt mask) {
  return (context->status&mask)!=0;
  } /* decContextTestStatus  */

/* ------------------------------------------------------------------ */
/* decContextZeroStatus -- clear all status bits                      */
/*                                                                    */
/*  context is the context structure to be updated                    */
/*  returns context                                                   */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI decContext * U_EXPORT2 uprv_decContextZeroStatus(decContext *context) {
  context->status=0;
  return context;
  } /* decContextZeroStatus  */

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/* ------------------------------------------------------------------ */
/* Decimal Number arithmetic module                                   */
/* ------------------------------------------------------------------ */
/* Copyright (c) IBM Corporation, 2000-2014.  All rights reserved.    */
/*                                                                    */
/* This software is made available under the terms of the             */
/* ICU License -- ICU 1.8.1 and later.                                */
/*                                                                    */
/* The description and User's Guide ("The decNumber C Library") for   */
/* this software is called decNumber.pdf.  This document is           */
/* available, together with arithmetic and format specifications,     */
/* testcases, and Web links, on the General Decimal Arithmetic page.  */
/*                                                                    */
/* Please send comments, suggestions, and corrections to the author:  */
/*   mfc@uk.ibm.com                                                   */
/*   Mike Cowlishaw, IBM Fellow                                       */
/*   IBM UK, PO Box 31, Birmingham Road, Warwick CV34 5JL, UK         */
/* ------------------------------------------------------------------ */

/* Modified version, for use from within ICU.
 *    Renamed public functions, to avoid an unwanted export of the
 *    standard names from the ICU library.
 *
 *    Use ICU's uprv_malloc() and uprv_free()
 *
 *    Revert comment syntax to plain C
 *
 *    Remove a few compiler warnings.
 */

/* This module comprises the routines for arbitrary-precision General */
/* Decimal Arithmetic as defined in the specification which may be    */
/* found on the General Decimal Arithmetic pages.  It implements both */
/* the full ('extended') arithmetic and the simpler ('subset')        */
/* arithmetic.                                                        */
/*                                                                    */
/* Usage notes:                                                       */
/*                                                                    */
/* 1. This code is ANSI C89 except:                                   */
/*                                                                    */
/*    a) C99 line comments (double forward slash) are used.  (Most C  */
/*       compilers accept these.  If yours does not, a simple script  */
/*       can be used to convert them to ANSI C comments.)             */
/*                                                                    */
/*    b) Types from C99 stdint.h are used.  If you do not have this   */
/*       header file, see the User's Guide section of the decNumber   */
/*       documentation; this lists the necessary definitions.         */
/*                                                                    */
/*    c) If DECDPUN>4 or DECUSE64=1, the C99 64-bit int64_t and       */
/*       uint64_t types may be used.  To avoid these, set DECUSE64=0  */
/*       and DECDPUN<=4 (see documentation).                          */
/*                                                                    */
/*    The code also conforms to C99 restrictions; in particular,      */
/*    strict aliasing rules are observed.                             */
/*                                                                    */
/* 2. The decNumber format which this library uses is optimized for   */
/*    efficient processing of relatively short numbers; in particular */
/*    it allows the use of fixed sized structures and minimizes copy  */
/*    and move operations.  It does, however, support arbitrary       */
/*    precision (up to 999,999,999 digits) and arbitrary exponent     */
/*    range (Emax in the range 0 through 999,999,999 and Emin in the  */
/*    range -999,999,999 through 0).  Mathematical functions (for     */
/*    example decNumberExp) as identified below are restricted more   */
/*    tightly: digits, emax, and -emin in the context must be <=      */
/*    DEC_MAX_MATH (999999), and their operand(s) must be within      */
/*    these bounds.                                                   */
/*                                                                    */
/* 3. Logical functions are further restricted; their operands must   */
/*    be finite, positive, have an exponent of zero, and all digits   */
/*    must be either 0 or 1.  The result will only contain digits     */
/*    which are 0 or 1 (and will have exponent=0 and a sign of 0).    */
/*                                                                    */
/* 4. Operands to operator functions are never modified unless they   */
/*    are also specified to be the result number (which is always     */
/*    permitted).  Other than that case, operands must not overlap.   */
/*                                                                    */
/* 5. Error handling: the type of the error is ORed into the status   */
/*    flags in the current context (decContext structure).  The       */
/*    SIGFPE signal is then raised if the corresponding trap-enabler  */
/*    flag in the decContext is set (is 1).                           */
/*                                                                    */
/*    It is the responsibility of the caller to clear the status      */
/*    flags as required.                                              */
/*                                                                    */
/*    The result of any routine which returns a number will always    */
/*    be a valid number (which may be a special value, such as an     */
/*    Infinity or NaN).                                               */
/*                                                                    */
/* 6. The decNumber format is not an exchangeable concrete            */
/*    representation as it comprises fields which may be machine-     */
/*    dependent (packed or unpacked, or special length, for example). */
/*    Canonical conversions to and from strings are provided; other   */
/*    conversions are available in separate modules.                  */
/*                                                                    */
/* 7. Normally, input operands are assumed to be valid.  Set DECCHECK */
/*    to 1 for extended operand checking (including NULL operands).   */
/*    Results are undefined if a badly-formed structure (or a NULL    */
/*    pointer to a structure) is provided, though with DECCHECK       */
/*    enabled the operator routines are protected against exceptions. */
/*    (Except if the result pointer is NULL, which is unrecoverable.) */
/*                                                                    */
/*    However, the routines will never cause exceptions if they are   */
/*    given well-formed operands, even if the value of the operands   */
/*    is inappropriate for the operation and DECCHECK is not set.     */
/*    (Except for SIGFPE, as and where documented.)                   */
/*                                                                    */
/* 8. Subset arithmetic is available only if DECSUBSET is set to 1.   */
/* ------------------------------------------------------------------ */
/* Implementation notes for maintenance of this module:               */
/*                                                                    */
/* 1. Storage leak protection:  Routines which use malloc are not     */
/*    permitted to use return for fastpath or error exits (i.e.,      */
/*    they follow strict structured programming conventions).         */
/*    Instead they have a do{}while(0); construct surrounding the     */
/*    code which is protected -- break may be used to exit this.      */
/*    Other routines can safely use the return statement inline.      */
/*                                                                    */
/*    Storage leak accounting can be enabled using DECALLOC.          */
/*                                                                    */
/* 2. All loops use the for(;;) construct.  Any do construct does     */
/*    not loop; it is for allocation protection as just described.    */
/*                                                                    */
/* 3. Setting status in the context must always be the very last      */
/*    action in a routine, as non-0 status may raise a trap and hence */
/*    the call to set status may not return (if the handler uses long */
/*    jump).  Therefore all cleanup must be done first.  In general,  */
/*    to achieve this status is accumulated and is only applied just  */
/*    before return by calling decContextSetStatus (via decStatus).   */
/*                                                                    */
/*    Routines which allocate storage cannot, in general, use the     */
/*    'top level' routines which could cause a non-returning          */
/*    transfer of control.  The decXxxxOp routines are safe (do not   */
/*    call decStatus even if traps are set in the context) and should */
/*    be used instead (they are also a little faster).                */
/*                                                                    */
/* 4. Exponent checking is minimized by allowing the exponent to      */
/*    grow outside its limits during calculations, provided that      */
/*    the decFinalize function is called later.  Multiplication and   */
/*    division, and intermediate calculations in exponentiation,      */
/*    require more careful checks because of the risk of 31-bit       */
/*    overflow (the most negative valid exponent is -1999999997, for  */
/*    a 999999999-digit number with adjusted exponent of -999999999). */
/*                                                                    */
/* 5. Rounding is deferred until finalization of results, with any    */
/*    'off to the right' data being represented as a single digit     */
/*    residue (in the range -1 through 9).  This avoids any double-   */
/*    rounding when more than one shortening takes place (for         */
/*    example, when a result is subnormal).                           */
/*                                                                    */
/* 6. The digits count is allowed to rise to a multiple of DECDPUN    */
/*    during many operations, so whole Units are handled and exact    */
/*    accounting of digits is not needed.  The correct digits value   */
/*    is found by decGetDigits, which accounts for leading zeros.     */
/*    This must be called before any rounding if the number of digits */
/*    is not known exactly.                                           */
/*                                                                    */
/* 7. The multiply-by-reciprocal 'trick' is used for partitioning     */
/*    numbers up to four digits, using appropriate constants.  This   */
/*    is not useful for longer numbers because overflow of 32 bits    */
/*    would lead to 4 multiplies, which is almost as expensive as     */
/*    a divide (unless a floating-point or 64-bit multiply is         */
/*    assumed to be available).                                       */
/*                                                                    */
/* 8. Unusual abbreviations that may be used in the commentary:       */
/*      lhs -- left hand side (operand, of an operation)              */
/*      lsd -- least significant digit (of coefficient)               */
/*      lsu -- least significant Unit (of coefficient)                */
/*      msd -- most significant digit (of coefficient)                */
/*      msi -- most significant item (in an array)                    */
/*      msu -- most significant Unit (of coefficient)                 */
/*      rhs -- right hand side (operand, of an operation)             */
/*      +ve -- positive                                               */
/*      -ve -- negative                                               */
/*      **  -- raise to the power                                     */
/* ------------------------------------------------------------------ */

#include <stdlib.h>                /* for malloc, free, etc.  */
/*  #include <stdio.h>   */        /* for printf [if needed]  */
#include <string.h>                /* for strcpy  */
#include <ctype.h>                 /* for lower  */
               /* for uprv_malloc, etc., in ICU */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/* ------------------------------------------------------------------ */
/* Decimal Number arithmetic module header                            */
/* ------------------------------------------------------------------ */
/* Copyright (c) IBM Corporation, 2000-2010.   All rights reserved.   */
/*                                                                    */
/* This software is made available under the terms of the             */
/* ICU License -- ICU 1.8.1 and later.                                */
/*                                                                    */
/* The description and User's Guide ("The decNumber C Library") for   */
/* this software is called decNumber.pdf.  This document is           */
/* available, together with arithmetic and format specifications,     */
/* testcases, and Web links, on the General Decimal Arithmetic page.  */
/*                                                                    */
/* Please send comments, suggestions, and corrections to the author:  */
/*   mfc@uk.ibm.com                                                   */
/*   Mike Cowlishaw, IBM Fellow                                       */
/*   IBM UK, PO Box 31, Birmingham Road, Warwick CV34 5JL, UK         */
/* ------------------------------------------------------------------ */

/* Modified version, for use from within ICU.
 *    Renamed public functions, to avoid an unwanted export of the
 *    standard names from the ICU library.
 *
 *    Use ICU's uprv_malloc() and uprv_free()
 *
 *    Revert comment syntax to plain C
 *
 *    Remove a few compiler warnings.
 */


  #define DECNAME     "decNumber"                       /* Short name */
  #define DECFULLNAME "Decimal Number Module"         /* Verbose name */
  #define DECAUTHOR   "Mike Cowlishaw"                /* Who to blame */

#if !defined(DECCONTEXT)

#endif

  /* Bit settings for decNumber.bits                                  */
  #define DECNEG    0x80      /* Sign; 1=negative, 0=positive or zero */
  #define DECINF    0x40      /* 1=Infinity                           */
  #define DECNAN    0x20      /* 1=NaN                                */
  #define DECSNAN   0x10      /* 1=sNaN                               */
  /* The remaining bits are reserved; they must be 0                  */
  #define DECSPECIAL (DECINF|DECNAN|DECSNAN) /* any special value     */

  /* Define the decNumber data structure.  The size and shape of the  */
  /* units array in the structure is determined by the following      */
  /* constant.  This must not be changed without recompiling the      */
  /* decNumber library modules. */

  /* For ICU, use one digit per byte, to make it easier to emulate the
   * old DigitList interface on top of a decNumber
   */
  #define DECDPUN 1           /* DECimal Digits Per UNit [must be >0  */
                              /* and <10; 3 or powers of 2 are best]. */

  /* DECNUMDIGITS is the default number of digits that can be held in */
  /* the structure.  If undefined, 1 is assumed and it is assumed     */
  /* that the structure will be immediately followed by extra space,  */
  /* as required.  DECNUMDIGITS is always >0.                         */
  #if !defined(DECNUMDIGITS)
    #define DECNUMDIGITS 1
  #endif

  /* The size (integer data type) of each unit is determined by the   */
  /* number of digits it will hold.                                   */
  #if   DECDPUN<=2
    #define decNumberUnit uint8_t
  #elif DECDPUN<=4
    #define decNumberUnit uint16_t
  #else
    #define decNumberUnit uint32_t
  #endif
  /* The number of units needed is ceil(DECNUMDIGITS/DECDPUN)         */
  #define DECNUMUNITS ((DECNUMDIGITS+DECDPUN-1)/DECDPUN)

  /* The data structure... */
  typedef struct {
    int32_t digits;      /* Count of digits in the coefficient; >0    */
    int32_t exponent;    /* Unadjusted exponent, unbiased, in         */
                         /* range: -1999999997 through 999999999      */
    uint8_t bits;        /* Indicator bits (see above)                */
                         /* Coefficient, from least significant unit  */
    decNumberUnit lsu[DECNUMUNITS];
    } decNumber;

  /* Notes:                                                           */
  /* 1. If digits is > DECDPUN then there will one or more            */
  /*    decNumberUnits immediately following the first element of lsu.*/
  /*    These contain the remaining (more significant) digits of the  */
  /*    number, and may be in the lsu array, or may be guaranteed by  */
  /*    some other mechanism (such as being contained in another      */
  /*    structure, or being overlaid on dynamically allocated         */
  /*    storage).                                                     */
  /*                                                                  */
  /*    Each integer of the coefficient (except potentially the last) */
  /*    contains DECDPUN digits (e.g., a value in the range 0 through */
  /*    99999999 if DECDPUN is 8, or 0 through 999 if DECDPUN is 3).  */
  /*                                                                  */
  /* 2. A decNumber converted to a string may need up to digits+14    */
  /*    characters.  The worst cases (non-exponential and exponential */
  /*    formats) are -0.00000{9...}# and -9.{9...}E+999999999#        */
  /*    (where # is '\0')                                             */


  /* ---------------------------------------------------------------- */
  /* decNumber public functions and macros                            */
  /* ---------------------------------------------------------------- */
  /* Conversions                                                      */
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberFromInt32(decNumber *, int32_t);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberFromUInt32(decNumber *, uint32_t);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberFromString(decNumber *, const char *, decContext *);
  U_INTERNAL char      * U_EXPORT2 uprv_decNumberToString(const decNumber *, char *);
  U_INTERNAL char      * U_EXPORT2 uprv_decNumberToEngString(const decNumber *, char *);
  U_INTERNAL uint32_t    U_EXPORT2 uprv_decNumberToUInt32(const decNumber *, decContext *);
  U_INTERNAL int32_t     U_EXPORT2 uprv_decNumberToInt32(const decNumber *, decContext *);
  U_INTERNAL uint8_t   * U_EXPORT2 uprv_decNumberGetBCD(const decNumber *, uint8_t *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberSetBCD(decNumber *, const uint8_t *, uint32_t);

  /* Operators and elementary functions                               */
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberAbs(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberAdd(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberAnd(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberCompare(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberCompareSignal(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberCompareTotal(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberCompareTotalMag(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberDivide(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberDivideInteger(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberExp(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberFMA(decNumber *, const decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberInvert(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberLn(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberLogB(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberLog10(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberMax(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberMaxMag(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberMin(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberMinMag(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberMinus(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberMultiply(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberNormalize(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberOr(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberPlus(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberPower(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberQuantize(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberReduce(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberRemainder(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberRemainderNear(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberRescale(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberRotate(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberSameQuantum(decNumber *, const decNumber *, const decNumber *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberScaleB(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberShift(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberSquareRoot(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberSubtract(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberToIntegralExact(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberToIntegralValue(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber * U_EXPORT2 uprv_decNumberXor(decNumber *, const decNumber *, const decNumber *, decContext *);

  /* Utilities                                                        */
  enum decClass uprv_decNumberClass(const decNumber *, decContext *);
  U_INTERNAL const char * U_EXPORT2 uprv_decNumberClassToString(enum decClass);
  U_INTERNAL decNumber  * U_EXPORT2 uprv_decNumberCopy(decNumber *, const decNumber *);
  U_INTERNAL decNumber  * U_EXPORT2 uprv_decNumberCopyAbs(decNumber *, const decNumber *);
  U_INTERNAL decNumber  * U_EXPORT2 uprv_decNumberCopyNegate(decNumber *, const decNumber *);
  U_INTERNAL decNumber  * U_EXPORT2 uprv_decNumberCopySign(decNumber *, const decNumber *, const decNumber *);
  U_INTERNAL decNumber  * U_EXPORT2 uprv_decNumberNextMinus(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber  * U_EXPORT2 uprv_decNumberNextPlus(decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber  * U_EXPORT2 uprv_decNumberNextToward(decNumber *, const decNumber *, const decNumber *, decContext *);
  U_INTERNAL decNumber  * U_EXPORT2 uprv_decNumberTrim(decNumber *);
  U_INTERNAL const char * U_EXPORT2 uprv_decNumberVersion(void);
  U_INTERNAL decNumber  * U_EXPORT2 uprv_decNumberZero(decNumber *);

  /* Functions for testing decNumbers (normality depends on context)  */
  U_INTERNAL int32_t U_EXPORT2 uprv_decNumberIsNormal(const decNumber *, decContext *);
  U_INTERNAL int32_t U_EXPORT2 uprv_decNumberIsSubnormal(const decNumber *, decContext *);

  /* Macros for testing decNumber *dn                                 */
  #define decNumberIsCanonical(dn) (1)  /* All decNumbers are saintly */
  #define decNumberIsFinite(dn)    (((dn)->bits&DECSPECIAL)==0)
  #define decNumberIsInfinite(dn)  (((dn)->bits&DECINF)!=0)
  #define decNumberIsNaN(dn)       (((dn)->bits&(DECNAN|DECSNAN))!=0)
  #define decNumberIsNegative(dn)  (((dn)->bits&DECNEG)!=0)
  #define decNumberIsQNaN(dn)      (((dn)->bits&(DECNAN))!=0)
  #define decNumberIsSNaN(dn)      (((dn)->bits&(DECSNAN))!=0)
  #define decNumberIsSpecial(dn)   (((dn)->bits&DECSPECIAL)!=0)
  #define decNumberIsZero(dn)      (*(dn)->lsu==0 \
                                    && (dn)->digits==1 \
                                    && (((dn)->bits&DECSPECIAL)==0))
  #define decNumberRadix(dn)       (10)

             /* base number library  */
        /* decNumber local types, etc.  */


/* Constants */
/* Public lookup table used by the D2U macro  */
static const uByte d2utable[DECMAXD2U+1]=D2UTABLE;

#define DECNUMBER_DECVERB     1              /* set to 1 for verbose DECCHECK  */
#define DECNUMBER_powers      DECPOWERS      /* old internal name  */

/* Local constants  */
#define DECNUMBER_DIVIDE      0x80           /* Divide operators  */
#define DECNUMBER_REMAINDER   0x40           /* ..  */
#define DECNUMBER_DIVIDEINT   0x20           /* ..  */
#define DECNUMBER_REMNEAR     0x10           /* ..  */
#define DECNUMBER_COMPARE     0x01           /* Compare operators  */
#define DECNUMBER_COMPMAX     0x02           /* ..  */
#define DECNUMBER_COMPMIN     0x03           /* ..  */
#define DECNUMBER_COMPTOTAL   0x04           /* ..  */
#define DECNUMBER_COMPNAN     0x05           /* .. [NaN processing]  */
#define DECNUMBER_COMPSIG     0x06           /* .. [signaling DECNUMBER_COMPARE]  */
#define DECNUMBER_COMPMAXMAG  0x07           /* ..  */
#define DECNUMBER_COMPMINMAG  0x08           /* ..  */

#define DECNUMBER_DEC_sNaN     0x40000000    /* local status: sNaN signal  */
#define DECNUMBER_BADINT  (Int)0x80000000    /* most-negative Int; error indicator  */
/* Next two indicate an integer >= 10**6, and its parity (bottom bit)  */
#define DECNUMBER_BIGEVEN (Int)0x80000002
#define DECNUMBER_BIGODD  (Int)0x80000003

static const Unit uarrone[1]={1};   /* Unit array of 1, used for incrementing  */

/* ------------------------------------------------------------------ */
/* round-for-reround digits                                           */
/* ------------------------------------------------------------------ */
#if 0
static const uByte DECSTICKYTAB[10]={1,1,2,3,4,6,6,7,8,9}; /* used if sticky */
#endif

/* ------------------------------------------------------------------ */
/* Powers of ten (DECNUMBER_powers[n]==10**n, 0<=n<=9)                          */
/* ------------------------------------------------------------------ */
static const uInt DECPOWERS[10]={1, 10, 100, 1000, 10000, 100000, 1000000,
                          10000000, 100000000, 1000000000};


/* Granularity-dependent code */
#if DECDPUN<=4
  #define eInt  Int           /* extended integer  */
  #define ueInt uInt          /* unsigned extended integer  */
  /* Constant multipliers for divide-by-power-of five using reciprocal  */
  /* multiply, after removing DECNUMBER_powers of 2 by shifting, and final shift  */
  /* of 17 [we only need up to **4]  */
  static const uInt multies[]={131073, 26215, 5243, 1049, 210};
  /* QUOT10 -- macro to return the quotient of unit u divided by 10**n  */
  #define QUOT10(u, n) ((((uInt)(u)>>(n))*multies[n])>>17)
#else
  /* For DECDPUN>4 non-ANSI-89 64-bit types are needed.  */
  #if !DECUSE64
    #error decNumber.c: DECUSE64 must be 1 when DECDPUN>4
  #endif
  #define eInt  Long          /* extended integer  */
  #define ueInt uLong         /* unsigned extended integer  */
#endif

/* Local routines */
static decNumber * decAddOp(decNumber *, const decNumber *, const decNumber *,
                              decContext *, uByte, uInt *);
static Flag        decBiStr(const char *, const char *, const char *);
static uInt        decCheckMath(const decNumber *, decContext *, uInt *);
static void        decApplyRound(decNumber *, decContext *, Int, uInt *);
static Int         decCompare(const decNumber *lhs, const decNumber *rhs, Flag);
static decNumber * decCompareOp(decNumber *, const decNumber *,
                              const decNumber *, decContext *,
                              Flag, uInt *);
static void        decCopyFit(decNumber *, const decNumber *, decContext *,
                              Int *, uInt *);
static decNumber * decDecap(decNumber *, Int);
static decNumber * decDivideOp(decNumber *, const decNumber *,
                              const decNumber *, decContext *, Flag, uInt *);
static decNumber * decExpOp(decNumber *, const decNumber *,
                              decContext *, uInt *);
static void        decFinalize(decNumber *, decContext *, Int *, uInt *);
static Int         decGetDigits(Unit *, Int);
static Int         decGetInt(const decNumber *);
static decNumber * decLnOp(decNumber *, const decNumber *,
                              decContext *, uInt *);
static decNumber * decMultiplyOp(decNumber *, const decNumber *,
                              const decNumber *, decContext *,
                              uInt *);
static decNumber * decNaNs(decNumber *, const decNumber *,
                              const decNumber *, decContext *, uInt *);
static decNumber * decQuantizeOp(decNumber *, const decNumber *,
                              const decNumber *, decContext *, Flag,
                              uInt *);
static void        decReverse(Unit *, Unit *);
static void        decSetCoeff(decNumber *, decContext *, const Unit *,
                              Int, Int *, uInt *);
static void        decSetMaxValue(decNumber *, decContext *);
static void        decSetOverflow(decNumber *, decContext *, uInt *);
static void        decSetSubnormal(decNumber *, decContext *, Int *, uInt *);
static Int         decShiftToLeast(Unit *, Int, Int);
static Int         decShiftToMost(Unit *, Int, Int);
static void        decStatus(decNumber *, uInt, decContext *);
static void        decToString(const decNumber *, char[], Flag);
static decNumber * decTrim(decNumber *, decContext *, Flag, Flag, Int *);
static Int         decUnitAddSub(const Unit *, Int, const Unit *, Int, Int,
                              Unit *, Int);
static Int         decUnitCompare(const Unit *, Int, const Unit *, Int, Int);

#if !DECSUBSET
/* decFinish == decFinalize when no subset arithmetic needed */
#define decFinish(a,b,c,d) decFinalize(a,b,c,d)
#else
static void        decFinish(decNumber *, decContext *, Int *, uInt *);
static decNumber * decRoundOperand(const decNumber *, decContext *, uInt *);
#endif

/* Local macros */
/* masked special-values bits  */
#define SPECIALARG  (rhs->bits & DECSPECIAL)
#define SPECIALARGS ((lhs->bits | rhs->bits) & DECSPECIAL)

// /* For use in ICU */
// #define malloc(a) uprv_malloc(a)
// #define free(a) uprv_free(a)

/* Diagnostic macros, etc. */
#if DECALLOC
/* Handle malloc/free accounting.  If enabled, our accountable routines  */
/* are used; otherwise the code just goes straight to the system malloc  */
// /* and free routines.  */
// #define malloc(a) decMalloc(a)
// #define free(a) decFree(a)
#define DECFENCE 0x5a              /* corruption detector  */
/* 'Our' malloc and free:  */
// static void *decMalloc(size_t);
// static void  decFree(void *);
uInt decAllocBytes=0;              /* count of bytes allocated  */
/* Note that DECALLOC code only checks for storage buffer overflow.  */
/* To check for memory leaks, the decAllocBytes variable must be  */
/* checked to be 0 at appropriate times (e.g., after the test  */
/* harness completes a set of tests).  This checking may be unreliable  */
/* if the testing is done in a multi-thread environment.  */
#endif

#if DECCHECK
/* Optional checking routines.  Enabling these means that decNumber  */
/* and decContext operands to operator routines are checked for  */
/* correctness.  This roughly doubles the execution time of the  */
/* fastest routines (and adds 600+ bytes), so should not normally be  */
/* used in 'production'.  */
/* decCheckInexact is used to check that inexact results have a full  */
/* complement of digits (where appropriate -- this is not the case  */
/* for Quantize, for example)  */
#define DECUNRESU ((decNumber *)(void *)0xffffffff)
#define DECUNUSED ((const decNumber *)(void *)0xffffffff)
#define DECUNCONT ((decContext *)(void *)(0xffffffff))
static Flag decCheckOperands(decNumber *, const decNumber *,
                             const decNumber *, decContext *);
static Flag decCheckNumber(const decNumber *);
static void decCheckInexact(const decNumber *, decContext *);
#endif

#if DECTRACE || DECCHECK
/* Optional trace/debugging routines (may or may not be used)  */
void decNumberShow(const decNumber *);  /* displays the components of a number  */
static void decDumpAr(char, const Unit *, Int);
#endif

/* ================================================================== */
/* Conversions                                                        */
/* ================================================================== */

/* ------------------------------------------------------------------ */
/* from-int32 -- conversion from Int or uInt                          */
/*                                                                    */
/*  dn is the decNumber to receive the integer                        */
/*  in or uin is the integer to be converted                          */
/*  returns dn                                                        */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberFromInt32(decNumber *dn, Int in) {
  uInt unsig;
  if (in>=0) unsig=in;
   else {                               /* negative (possibly DECNUMBER_BADINT)  */
    if (in==DECNUMBER_BADINT) unsig=(uInt)1073741824*2; /* special case  */
     else unsig=-in;                    /* invert  */
    }
  /* in is now positive  */
  uprv_decNumberFromUInt32(dn, unsig);
  if (in<0) dn->bits=DECNEG;            /* sign needed  */
  return dn;
  } /* decNumberFromInt32  */

U_CAPI decNumber * U_EXPORT2 uprv_decNumberFromUInt32(decNumber *dn, uInt uin) {
  Unit *up;                             /* work pointer  */
  uprv_decNumberZero(dn);                    /* clean  */
  if (uin==0) return dn;                /* [or decGetDigits bad call]  */
  for (up=dn->lsu; uin>0; up++) {
    *up=(Unit)(uin%(DECDPUNMAX+1));
    uin=uin/(DECDPUNMAX+1);
    }
  dn->digits=decGetDigits(dn->lsu, static_cast<int32_t>(up - dn->lsu));
  return dn;
  } /* decNumberFromUInt32  */

/* ------------------------------------------------------------------ */
/* to-int32 -- conversion to Int or uInt                              */
/*                                                                    */
/*  dn is the decNumber to convert                                    */
/*  set is the context for reporting errors                           */
/*  returns the converted decNumber, or 0 if Invalid is set           */
/*                                                                    */
/* Invalid is set if the decNumber does not have exponent==0 or if    */
/* it is a NaN, Infinite, or out-of-range.                            */
/* ------------------------------------------------------------------ */
U_CAPI Int U_EXPORT2 uprv_decNumberToInt32(const decNumber *dn, decContext *set) {
  #if DECCHECK
  if (decCheckOperands(DECUNRESU, DECUNUSED, dn, set)) return 0;
  #endif

  /* special or too many digits, or bad exponent  */
  if (dn->bits&DECSPECIAL || dn->digits>10 || dn->exponent!=0) ; /* bad  */
   else { /* is a finite integer with 10 or fewer digits  */
    Int d;                         /* work  */
    const Unit *up;                /* ..  */
    uInt hi=0, lo;                 /* ..  */
    up=dn->lsu;                    /* -> lsu  */
    lo=*up;                        /* get 1 to 9 digits  */
    #if DECDPUN>1                  /* split to higher  */
      hi=lo/10;
      lo=lo%10;
    #endif
    up++;
    /* collect remaining Units, if any, into hi  */
    for (d=DECDPUN; d<dn->digits; up++, d+=DECDPUN) hi+=*up*DECNUMBER_powers[d-1];
    /* now low has the lsd, hi the remainder  */
    if (hi>214748364 || (hi==214748364 && lo>7)) { /* out of range?  */
      /* most-negative is a reprieve  */
      if (dn->bits&DECNEG && hi==214748364 && lo==8) return 0x80000000;
      /* bad -- drop through  */
      }
     else { /* in-range always  */
      Int i=X10(hi)+lo;
      if (dn->bits&DECNEG) return -i;
      return i;
      }
    } /* integer  */
  uprv_decContextSetStatus(set, DEC_Invalid_operation); /* [may not return]  */
  return 0;
  } /* decNumberToInt32  */

U_CAPI uInt U_EXPORT2 uprv_decNumberToUInt32(const decNumber *dn, decContext *set) {
  #if DECCHECK
  if (decCheckOperands(DECUNRESU, DECUNUSED, dn, set)) return 0;
  #endif
  /* special or too many digits, or bad exponent, or negative (<0)  */
  if (dn->bits&DECSPECIAL || dn->digits>10 || dn->exponent!=0
    || (dn->bits&DECNEG && !ISZERO(dn)));                   /* bad  */
   else { /* is a finite integer with 10 or fewer digits  */
    Int d;                         /* work  */
    const Unit *up;                /* ..  */
    uInt hi=0, lo;                 /* ..  */
    up=dn->lsu;                    /* -> lsu  */
    lo=*up;                        /* get 1 to 9 digits  */
    #if DECDPUN>1                  /* split to higher  */
      hi=lo/10;
      lo=lo%10;
    #endif
    up++;
    /* collect remaining Units, if any, into hi  */
    for (d=DECDPUN; d<dn->digits; up++, d+=DECDPUN) hi+=*up*DECNUMBER_powers[d-1];

    /* now low has the lsd, hi the remainder  */
    if (hi>429496729 || (hi==429496729 && lo>5)) ; /* no reprieve possible  */
     else return X10(hi)+lo;
    } /* integer  */
  uprv_decContextSetStatus(set, DEC_Invalid_operation); /* [may not return]  */
  return 0;
  } /* decNumberToUInt32  */

/* ------------------------------------------------------------------ */
/* to-scientific-string -- conversion to numeric string               */
/* to-engineering-string -- conversion to numeric string              */
/*                                                                    */
/*   decNumberToString(dn, string);                                   */
/*   decNumberToEngString(dn, string);                                */
/*                                                                    */
/*  dn is the decNumber to convert                                    */
/*  string is the string where the result will be laid out            */
/*                                                                    */
/*  string must be at least dn->digits+14 characters long             */
/*                                                                    */
/*  No error is possible, and no status can be set.                   */
/* ------------------------------------------------------------------ */
U_CAPI char * U_EXPORT2 uprv_decNumberToString(const decNumber *dn, char *string){
  decToString(dn, string, 0);
  return string;
  } /* DecNumberToString  */

U_CAPI char * U_EXPORT2 uprv_decNumberToEngString(const decNumber *dn, char *string){
  decToString(dn, string, 1);
  return string;
  } /* DecNumberToEngString  */

/* ------------------------------------------------------------------ */
/* to-number -- conversion from numeric string                        */
/*                                                                    */
/* decNumberFromString -- convert string to decNumber                 */
/*   dn        -- the number structure to fill                        */
/*   chars[]   -- the string to convert ('\0' terminated)             */
/*   set       -- the context used for processing any error,          */
/*                determining the maximum precision available         */
/*                (set.digits), determining the maximum and minimum   */
/*                exponent (set.emax and set.emin), determining if    */
/*                extended values are allowed, and checking the       */
/*                rounding mode if overflow occurs or rounding is     */
/*                needed.                                             */
/*                                                                    */
/* The length of the coefficient and the size of the exponent are     */
/* checked by this routine, so the correct error (Underflow or        */
/* Overflow) can be reported or rounding applied, as necessary.       */
/*                                                                    */
/* If bad syntax is detected, the result will be a quiet NaN.         */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberFromString(decNumber *dn, const char chars[],
                                decContext *set) {
  Int   exponent=0;                /* working exponent [assume 0]  */
  uByte bits=0;                    /* working flags [assume +ve]  */
  Unit  *res;                      /* where result will be built  */
  Unit  resbuff[SD2U(DECBUFFER+9)];/* local buffer in case need temporary  */
                                   /* [+9 allows for ln() constants]  */
  Unit  *allocres=NULL;            /* -> allocated result, iff allocated  */
  Int   d=0;                       /* count of digits found in decimal part  */
  const char *dotchar=NULL;        /* where dot was found  */
  const char *cfirst=chars;        /* -> first character of decimal part  */
  const char *last=NULL;           /* -> last digit of decimal part  */
  const char *c;                   /* work  */
  Unit  *up;                       /* ..  */
  #if DECDPUN>1
  Int   cut, out;                  /* ..  */
  #endif
  Int   residue;                   /* rounding residue  */
  uInt  status=0;                  /* error code  */

  #if DECCHECK
  if (decCheckOperands(DECUNRESU, DECUNUSED, DECUNUSED, set))
    return uprv_decNumberZero(dn);
  #endif

  do {                             /* status & malloc protection  */
    for (c=chars;; c++) {          /* -> input character  */
      if (*c>='0' && *c<='9') {    /* test for Arabic digit  */
        last=c;
        d++;                       /* count of real digits  */
        continue;                  /* still in decimal part  */
        }
      if (*c=='.' && dotchar==NULL) { /* first '.'  */
        dotchar=c;                 /* record offset into decimal part  */
        if (c==cfirst) cfirst++;   /* first digit must follow  */
        continue;}
      if (c==chars) {              /* first in string...  */
        if (*c=='-') {             /* valid - sign  */
          cfirst++;
          bits=DECNEG;
          continue;}
        if (*c=='+') {             /* valid + sign  */
          cfirst++;
          continue;}
        }
      /* *c is not a digit, or a valid +, -, or '.'  */
      break;
      } /* c  */

    if (last==NULL) {              /* no digits yet  */
      status=DEC_Conversion_syntax;/* assume the worst  */
      if (*c=='\0') break;         /* and no more to come...  */
      #if DECSUBSET
      /* if subset then infinities and NaNs are not allowed  */
      if (!set->extended) break;   /* hopeless  */
      #endif
      /* Infinities and NaNs are possible, here  */
      if (dotchar!=NULL) break;    /* .. unless had a dot  */
      uprv_decNumberZero(dn);           /* be optimistic  */
      if (decBiStr(c, "infinity", "INFINITY")
       || decBiStr(c, "inf", "INF")) {
        dn->bits=bits | DECINF;
        status=0;                  /* is OK  */
        break; /* all done  */
        }
      /* a NaN expected  */
      /* 2003.09.10 NaNs are now permitted to have a sign  */
      dn->bits=bits | DECNAN;      /* assume simple NaN  */
      if (*c=='s' || *c=='S') {    /* looks like an sNaN  */
        c++;
        dn->bits=bits | DECSNAN;
        }
      if (*c!='n' && *c!='N') break;    /* check caseless "NaN"  */
      c++;
      if (*c!='a' && *c!='A') break;    /* ..  */
      c++;
      if (*c!='n' && *c!='N') break;    /* ..  */
      c++;
      /* now either nothing, or nnnn payload, expected  */
      /* -> start of integer and skip leading 0s [including plain 0]  */
      for (cfirst=c; *cfirst=='0';) cfirst++;
      if (*cfirst=='\0') {         /* "NaN" or "sNaN", maybe with all 0s  */
        status=0;                  /* it's good  */
        break;                     /* ..  */
        }
      /* something other than 0s; setup last and d as usual [no dots]  */
      for (c=cfirst;; c++, d++) {
        if (*c<'0' || *c>'9') break; /* test for Arabic digit  */
        last=c;
        }
      if (*c!='\0') break;         /* not all digits  */
      if (d>set->digits-1) {
        /* [NB: payload in a decNumber can be full length unless  */
        /* clamped, in which case can only be digits-1]  */
        if (set->clamp) break;
        if (d>set->digits) break;
        } /* too many digits?  */
      /* good; drop through to convert the integer to coefficient  */
      status=0;                    /* syntax is OK  */
      bits=dn->bits;               /* for copy-back  */
      } /* last==NULL  */

     else if (*c!='\0') {          /* more to process...  */
      /* had some digits; exponent is only valid sequence now  */
      Flag nege;                   /* 1=negative exponent  */
      const char *firstexp;        /* -> first significant exponent digit  */
      status=DEC_Conversion_syntax;/* assume the worst  */
      if (*c!='e' && *c!='E') break;
      /* Found 'e' or 'E' -- now process explicit exponent */
      /* 1998.07.11: sign no longer required  */
      nege=0;
      c++;                         /* to (possible) sign  */
      if (*c=='-') {nege=1; c++;}
       else if (*c=='+') c++;
      if (*c=='\0') break;

      for (; *c=='0' && *(c+1)!='\0';) c++;  /* strip insignificant zeros  */
      firstexp=c;                            /* save exponent digit place  */
      uInt uexponent = 0;   /* Avoid undefined behavior on signed int overflow */
      for (; ;c++) {
        if (*c<'0' || *c>'9') break;         /* not a digit  */
        uexponent=X10(uexponent)+(uInt)*c-(uInt)'0';
        } /* c  */
      exponent = (Int)uexponent;
      /* if not now on a '\0', *c must not be a digit  */
      if (*c!='\0') break;

      /* (this next test must be after the syntax checks)  */
      /* if it was too long the exponent may have wrapped, so check  */
      /* carefully and set it to a certain overflow if wrap possible  */
      if (c>=firstexp+9+1) {
        if (c>firstexp+9+1 || *firstexp>'1') exponent=DECNUMMAXE*2;
        /* [up to 1999999999 is OK, for example 1E-1000000998]  */
        }
      if (nege) exponent=-exponent;     /* was negative  */
      status=0;                         /* is OK  */
      } /* stuff after digits  */

    /* Here when whole string has been inspected; syntax is good  */
    /* cfirst->first digit (never dot), last->last digit (ditto)  */

    /* strip leading zeros/dot [leave final 0 if all 0's]  */
    if (*cfirst=='0') {                 /* [cfirst has stepped over .]  */
      for (c=cfirst; c<last; c++, cfirst++) {
        if (*c=='.') continue;          /* ignore dots  */
        if (*c!='0') break;             /* non-zero found  */
        d--;                            /* 0 stripped  */
        } /* c  */
      #if DECSUBSET
      /* make a rapid exit for easy zeros if !extended  */
      if (*cfirst=='0' && !set->extended) {
        uprv_decNumberZero(dn);              /* clean result  */
        break;                          /* [could be return]  */
        }
      #endif
      } /* at least one leading 0  */

    /* Handle decimal point...  */
    if (dotchar!=NULL && dotchar<last)  /* non-trailing '.' found?  */
      exponent -= static_cast<int32_t>(last-dotchar);         /* adjust exponent  */
    /* [we can now ignore the .]  */

    /* OK, the digits string is good.  Assemble in the decNumber, or in  */
    /* a temporary units array if rounding is needed  */
    if (d<=set->digits) res=dn->lsu;    /* fits into supplied decNumber  */
     else {                             /* rounding needed  */
      Int needbytes=D2U(d)*sizeof(Unit);/* bytes needed  */
      res=resbuff;                      /* assume use local buffer  */
      if (needbytes>(Int)sizeof(resbuff)) { /* too big for local  */
        allocres=(Unit *)malloc(needbytes);
        if (allocres==NULL) {status|=DEC_Insufficient_storage; break;}
        res=allocres;
        }
      }
    /* res now -> number lsu, buffer, or allocated storage for Unit array  */

    /* Place the coefficient into the selected Unit array  */
    /* [this is often 70% of the cost of this function when DECDPUN>1]  */
    #if DECDPUN>1
    out=0;                         /* accumulator  */
    up=res+D2U(d)-1;               /* -> msu  */
    cut=d-(up-res)*DECDPUN;        /* digits in top unit  */
    for (c=cfirst;; c++) {         /* along the digits  */
      if (*c=='.') continue;       /* ignore '.' [don't decrement cut]  */
      out=X10(out)+(Int)*c-(Int)'0';
      if (c==last) break;          /* done [never get to trailing '.']  */
      cut--;
      if (cut>0) continue;         /* more for this unit  */
      *up=(Unit)out;               /* write unit  */
      up--;                        /* prepare for unit below..  */
      cut=DECDPUN;                 /* ..  */
      out=0;                       /* ..  */
      } /* c  */
    *up=(Unit)out;                 /* write lsu  */

    #else
    /* DECDPUN==1  */
    up=res;                        /* -> lsu  */
    for (c=last; c>=cfirst; c--) { /* over each character, from least  */
      if (*c=='.') continue;       /* ignore . [don't step up]  */
      *up=(Unit)((Int)*c-(Int)'0');
      up++;
      } /* c  */
    #endif

    dn->bits=bits;
    dn->exponent=exponent;
    dn->digits=d;

    /* if not in number (too long) shorten into the number  */
    if (d>set->digits) {
      residue=0;
      decSetCoeff(dn, set, res, d, &residue, &status);
      /* always check for overflow or subnormal and round as needed  */
      decFinalize(dn, set, &residue, &status);
      }
     else { /* no rounding, but may still have overflow or subnormal  */
      /* [these tests are just for performance; finalize repeats them]  */
      if ((dn->exponent-1<set->emin-dn->digits)
       || (dn->exponent-1>set->emax-set->digits)) {
        residue=0;
        decFinalize(dn, set, &residue, &status);
        }
      }
    /* decNumberShow(dn);  */
    } while(0);                         /* [for break]  */

  if (allocres!=NULL) free(allocres);   /* drop any storage used  */
  if (status!=0) decStatus(dn, status, set);
  return dn;
  } /* decNumberFromString */

/* ================================================================== */
/* Operators                                                          */
/* ================================================================== */

/* ------------------------------------------------------------------ */
/* decNumberAbs -- absolute value operator                            */
/*                                                                    */
/*   This computes C = abs(A)                                         */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context                                               */
/*                                                                    */
/* See also decNumberCopyAbs for a quiet bitwise version of this.     */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
/* This has the same effect as decNumberPlus unless A is negative,    */
/* in which case it has the same effect as decNumberMinus.            */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberAbs(decNumber *res, const decNumber *rhs,
                         decContext *set) {
  decNumber dzero;                      /* for 0  */
  uInt status=0;                        /* accumulator  */

  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  uprv_decNumberZero(&dzero);                /* set 0  */
  dzero.exponent=rhs->exponent;         /* [no coefficient expansion]  */
  decAddOp(res, &dzero, rhs, set, (uByte)(rhs->bits & DECNEG), &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberAbs  */

/* ------------------------------------------------------------------ */
/* decNumberAdd -- add two Numbers                                    */
/*                                                                    */
/*   This computes C = A + B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X+X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
/* This just calls the routine shared with Subtract                   */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberAdd(decNumber *res, const decNumber *lhs,
                         const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decAddOp(res, lhs, rhs, set, 0, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberAdd  */

/* ------------------------------------------------------------------ */
/* decNumberAnd -- AND two Numbers, digitwise                         */
/*                                                                    */
/*   This computes C = A & B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X&X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context (used for result length and error report)     */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Logical function restrictions apply (see above); a NaN is          */
/* returned with Invalid_operation if a restriction is violated.      */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberAnd(decNumber *res, const decNumber *lhs,
                         const decNumber *rhs, decContext *set) {
  const Unit *ua, *ub;                  /* -> operands  */
  const Unit *msua, *msub;              /* -> operand msus  */
  Unit *uc,  *msuc;                     /* -> result and its msu  */
  Int   msudigs;                        /* digits in res msu  */
  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  if (lhs->exponent!=0 || decNumberIsSpecial(lhs) || decNumberIsNegative(lhs)
   || rhs->exponent!=0 || decNumberIsSpecial(rhs) || decNumberIsNegative(rhs)) {
    decStatus(res, DEC_Invalid_operation, set);
    return res;
    }

  /* operands are valid  */
  ua=lhs->lsu;                          /* bottom-up  */
  ub=rhs->lsu;                          /* ..  */
  uc=res->lsu;                          /* ..  */
  msua=ua+D2U(lhs->digits)-1;           /* -> msu of lhs  */
  msub=ub+D2U(rhs->digits)-1;           /* -> msu of rhs  */
  msuc=uc+D2U(set->digits)-1;           /* -> msu of result  */
  msudigs=MSUDIGITS(set->digits);       /* [faster than remainder]  */
  for (; uc<=msuc; ua++, ub++, uc++) {  /* Unit loop  */
    Unit a, b;                          /* extract units  */
    if (ua>msua) a=0;
     else a=*ua;
    if (ub>msub) b=0;
     else b=*ub;
    *uc=0;                              /* can now write back  */
    if (a|b) {                          /* maybe 1 bits to examine  */
      Int i, j;
      *uc=0;                            /* can now write back  */
      /* This loop could be unrolled and/or use BIN2BCD tables  */
      for (i=0; i<DECDPUN; i++) {
        if (a&b&1) *uc=*uc+(Unit)DECNUMBER_powers[i];  /* effect AND  */
        j=a%10;
        a=a/10;
        j|=b%10;
        b=b/10;
        if (j>1) {
          decStatus(res, DEC_Invalid_operation, set);
          return res;
          }
        if (uc==msuc && i==msudigs-1) break; /* just did final digit  */
        } /* each digit  */
      } /* both OK  */
    } /* each unit  */
  /* [here uc-1 is the msu of the result]  */
  res->digits=decGetDigits(res->lsu, static_cast<int32_t>(uc - res->lsu));
  res->exponent=0;                      /* integer  */
  res->bits=0;                          /* sign=0  */
  return res;  /* [no status to set]  */
  } /* decNumberAnd  */

/* ------------------------------------------------------------------ */
/* decNumberCompare -- compare two Numbers                            */
/*                                                                    */
/*   This computes C = A ? B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X?X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for one digit (or NaN).                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberCompare(decNumber *res, const decNumber *lhs,
                             const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decCompareOp(res, lhs, rhs, set, DECNUMBER_COMPARE, &status);
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberCompare  */

/* ------------------------------------------------------------------ */
/* decNumberCompareSignal -- compare, signalling on all NaNs          */
/*                                                                    */
/*   This computes C = A ? B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X?X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for one digit (or NaN).                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberCompareSignal(decNumber *res, const decNumber *lhs,
                                   const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decCompareOp(res, lhs, rhs, set, DECNUMBER_COMPSIG, &status);
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberCompareSignal  */

/* ------------------------------------------------------------------ */
/* decNumberCompareTotal -- compare two Numbers, using total ordering */
/*                                                                    */
/*   This computes C = A ? B, under total ordering                    */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X?X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for one digit; the result will always be one of  */
/* -1, 0, or 1.                                                       */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberCompareTotal(decNumber *res, const decNumber *lhs,
                                  const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decCompareOp(res, lhs, rhs, set, DECNUMBER_COMPTOTAL, &status);
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberCompareTotal  */

/* ------------------------------------------------------------------ */
/* decNumberCompareTotalMag -- compare, total ordering of magnitudes  */
/*                                                                    */
/*   This computes C = |A| ? |B|, under total ordering                */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X?X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for one digit; the result will always be one of  */
/* -1, 0, or 1.                                                       */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberCompareTotalMag(decNumber *res, const decNumber *lhs,
                                     const decNumber *rhs, decContext *set) {
  uInt status=0;                   /* accumulator  */
  uInt needbytes;                  /* for space calculations  */
  decNumber bufa[D2N(DECBUFFER+1)];/* +1 in case DECBUFFER=0  */
  decNumber *allocbufa=NULL;       /* -> allocated bufa, iff allocated  */
  decNumber bufb[D2N(DECBUFFER+1)];
  decNumber *allocbufb=NULL;       /* -> allocated bufb, iff allocated  */
  decNumber *a, *b;                /* temporary pointers  */

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  do {                                  /* protect allocated storage  */
    /* if either is negative, take a copy and absolute  */
    if (decNumberIsNegative(lhs)) {     /* lhs<0  */
      a=bufa;
      needbytes=sizeof(decNumber)+(D2U(lhs->digits)-1)*sizeof(Unit);
      if (needbytes>sizeof(bufa)) {     /* need malloc space  */
        allocbufa=(decNumber *)malloc(needbytes);
        if (allocbufa==NULL) {          /* hopeless -- abandon  */
          status|=DEC_Insufficient_storage;
          break;}
        a=allocbufa;                    /* use the allocated space  */
        }
      uprv_decNumberCopy(a, lhs);            /* copy content  */
      a->bits&=~DECNEG;                 /* .. and clear the sign  */
      lhs=a;                            /* use copy from here on  */
      }
    if (decNumberIsNegative(rhs)) {     /* rhs<0  */
      b=bufb;
      needbytes=sizeof(decNumber)+(D2U(rhs->digits)-1)*sizeof(Unit);
      if (needbytes>sizeof(bufb)) {     /* need malloc space  */
        allocbufb=(decNumber *)malloc(needbytes);
        if (allocbufb==NULL) {          /* hopeless -- abandon  */
          status|=DEC_Insufficient_storage;
          break;}
        b=allocbufb;                    /* use the allocated space  */
        }
      uprv_decNumberCopy(b, rhs);            /* copy content  */
      b->bits&=~DECNEG;                 /* .. and clear the sign  */
      rhs=b;                            /* use copy from here on  */
      }
    decCompareOp(res, lhs, rhs, set, DECNUMBER_COMPTOTAL, &status);
    } while(0);                         /* end protected  */

  if (allocbufa!=NULL) free(allocbufa); /* drop any storage used  */
  if (allocbufb!=NULL) free(allocbufb); /* ..  */
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberCompareTotalMag  */

/* ------------------------------------------------------------------ */
/* decNumberDivide -- divide one number by another                    */
/*                                                                    */
/*   This computes C = A / B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X/X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberDivide(decNumber *res, const decNumber *lhs,
                            const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decDivideOp(res, lhs, rhs, set, DECNUMBER_DIVIDE, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberDivide  */

/* ------------------------------------------------------------------ */
/* decNumberDivideInteger -- divide and return integer quotient       */
/*                                                                    */
/*   This computes C = A # B, where # is the integer divide operator  */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X#X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberDivideInteger(decNumber *res, const decNumber *lhs,
                                   const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decDivideOp(res, lhs, rhs, set, DECNUMBER_DIVIDEINT, &status);
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberDivideInteger  */

/* ------------------------------------------------------------------ */
/* decNumberExp -- exponentiation                                     */
/*                                                                    */
/*   This computes C = exp(A)                                         */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context; note that rounding mode has no effect        */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Mathematical function restrictions apply (see above); a NaN is     */
/* returned with Invalid_operation if a restriction is violated.      */
/*                                                                    */
/* Finite results will always be full precision and Inexact, except   */
/* when A is a zero or -Infinity (giving 1 or 0 respectively).        */
/*                                                                    */
/* An Inexact result is rounded using DEC_ROUND_HALF_EVEN; it will    */
/* almost always be correctly rounded, but may be up to 1 ulp in      */
/* error in rare cases.                                               */
/* ------------------------------------------------------------------ */
/* This is a wrapper for decExpOp which can handle the slightly wider */
/* (double) range needed by Ln (which has to be able to calculate     */
/* exp(-a) where a can be the tiniest number (Ntiny).                 */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberExp(decNumber *res, const decNumber *rhs,
                         decContext *set) {
  uInt status=0;                        /* accumulator  */
  #if DECSUBSET
  decNumber *allocrhs=NULL;        /* non-NULL if rounded rhs allocated  */
  #endif

  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  /* Check restrictions; these restrictions ensure that if h=8 (see  */
  /* decExpOp) then the result will either overflow or underflow to 0.  */
  /* Other math functions restrict the input range, too, for inverses.  */
  /* If not violated then carry out the operation.  */
  if (!decCheckMath(rhs, set, &status)) do { /* protect allocation  */
    #if DECSUBSET
    if (!set->extended) {
      /* reduce operand and set lostDigits status, as needed  */
      if (rhs->digits>set->digits) {
        allocrhs=decRoundOperand(rhs, set, &status);
        if (allocrhs==NULL) break;
        rhs=allocrhs;
        }
      }
    #endif
    decExpOp(res, rhs, set, &status);
    } while(0);                         /* end protected  */

  #if DECSUBSET
  if (allocrhs !=NULL) free(allocrhs);  /* drop any storage used  */
  #endif
  /* apply significant status  */
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberExp  */

/* ------------------------------------------------------------------ */
/* decNumberFMA -- fused multiply add                                 */
/*                                                                    */
/*   This computes D = (A * B) + C with only one rounding             */
/*                                                                    */
/*   res is D, the result.  D may be A or B or C (e.g., X=FMA(X,X,X)) */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   fhs is C [far hand side]                                         */
/*   set is the context                                               */
/*                                                                    */
/* Mathematical function restrictions apply (see above); a NaN is     */
/* returned with Invalid_operation if a restriction is violated.      */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberFMA(decNumber *res, const decNumber *lhs,
                         const decNumber *rhs, const decNumber *fhs,
                         decContext *set) {
  uInt status=0;                   /* accumulator  */
  decContext dcmul;                /* context for the multiplication  */
  uInt needbytes;                  /* for space calculations  */
  decNumber bufa[D2N(DECBUFFER*2+1)];
  decNumber *allocbufa=NULL;       /* -> allocated bufa, iff allocated  */
  decNumber *acc;                  /* accumulator pointer  */
  decNumber dzero;                 /* work  */

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  if (decCheckOperands(res, fhs, DECUNUSED, set)) return res;
  #endif

  do {                                  /* protect allocated storage  */
    #if DECSUBSET
    if (!set->extended) {               /* [undefined if subset]  */
      status|=DEC_Invalid_operation;
      break;}
    #endif
    /* Check math restrictions [these ensure no overflow or underflow]  */
    if ((!decNumberIsSpecial(lhs) && decCheckMath(lhs, set, &status))
     || (!decNumberIsSpecial(rhs) && decCheckMath(rhs, set, &status))
     || (!decNumberIsSpecial(fhs) && decCheckMath(fhs, set, &status))) break;
    /* set up context for multiply  */
    dcmul=*set;
    dcmul.digits=lhs->digits+rhs->digits; /* just enough  */
    /* [The above may be an over-estimate for subset arithmetic, but that's OK]  */
    dcmul.emax=DEC_MAX_EMAX;            /* effectively unbounded ..  */
    dcmul.emin=DEC_MIN_EMIN;            /* [thanks to Math restrictions]  */
    /* set up decNumber space to receive the result of the multiply  */
    acc=bufa;                           /* may fit  */
    needbytes=sizeof(decNumber)+(D2U(dcmul.digits)-1)*sizeof(Unit);
    if (needbytes>sizeof(bufa)) {       /* need malloc space  */
      allocbufa=(decNumber *)malloc(needbytes);
      if (allocbufa==NULL) {            /* hopeless -- abandon  */
        status|=DEC_Insufficient_storage;
        break;}
      acc=allocbufa;                    /* use the allocated space  */
      }
    /* multiply with extended range and necessary precision  */
    /*printf("emin=%ld\n", dcmul.emin);  */
    decMultiplyOp(acc, lhs, rhs, &dcmul, &status);
    /* Only Invalid operation (from sNaN or Inf * 0) is possible in  */
    /* status; if either is seen than ignore fhs (in case it is  */
    /* another sNaN) and set acc to NaN unless we had an sNaN  */
    /* [decMultiplyOp leaves that to caller]  */
    /* Note sNaN has to go through addOp to shorten payload if  */
    /* necessary  */
    if ((status&DEC_Invalid_operation)!=0) {
      if (!(status&DECNUMBER_DEC_sNaN)) {         /* but be true invalid  */
        uprv_decNumberZero(res);             /* acc not yet set  */
        res->bits=DECNAN;
        break;
        }
      uprv_decNumberZero(&dzero);            /* make 0 (any non-NaN would do)  */
      fhs=&dzero;                       /* use that  */
      }
    #if DECCHECK
     else { /* multiply was OK  */
      if (status!=0) printf("Status=%08lx after FMA multiply\n", (LI)status);
      }
    #endif
    /* add the third operand and result -> res, and all is done  */
    decAddOp(res, acc, fhs, set, 0, &status);
    } while(0);                         /* end protected  */

  if (allocbufa!=NULL) free(allocbufa); /* drop any storage used  */
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberFMA  */

/* ------------------------------------------------------------------ */
/* decNumberInvert -- invert a Number, digitwise                      */
/*                                                                    */
/*   This computes C = ~A                                             */
/*                                                                    */
/*   res is C, the result.  C may be A (e.g., X=~X)                   */
/*   rhs is A                                                         */
/*   set is the context (used for result length and error report)     */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Logical function restrictions apply (see above); a NaN is          */
/* returned with Invalid_operation if a restriction is violated.      */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberInvert(decNumber *res, const decNumber *rhs,
                            decContext *set) {
  const Unit *ua, *msua;                /* -> operand and its msu  */
  Unit  *uc, *msuc;                     /* -> result and its msu  */
  Int   msudigs;                        /* digits in res msu  */
  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  if (rhs->exponent!=0 || decNumberIsSpecial(rhs) || decNumberIsNegative(rhs)) {
    decStatus(res, DEC_Invalid_operation, set);
    return res;
    }
  /* operand is valid  */
  ua=rhs->lsu;                          /* bottom-up  */
  uc=res->lsu;                          /* ..  */
  msua=ua+D2U(rhs->digits)-1;           /* -> msu of rhs  */
  msuc=uc+D2U(set->digits)-1;           /* -> msu of result  */
  msudigs=MSUDIGITS(set->digits);       /* [faster than remainder]  */
  for (; uc<=msuc; ua++, uc++) {        /* Unit loop  */
    Unit a;                             /* extract unit  */
    Int  i, j;                          /* work  */
    if (ua>msua) a=0;
     else a=*ua;
    *uc=0;                              /* can now write back  */
    /* always need to examine all bits in rhs  */
    /* This loop could be unrolled and/or use BIN2BCD tables  */
    for (i=0; i<DECDPUN; i++) {
      if ((~a)&1) *uc=*uc+(Unit)DECNUMBER_powers[i];   /* effect INVERT  */
      j=a%10;
      a=a/10;
      if (j>1) {
        decStatus(res, DEC_Invalid_operation, set);
        return res;
        }
      if (uc==msuc && i==msudigs-1) break;   /* just did final digit  */
      } /* each digit  */
    } /* each unit  */
  /* [here uc-1 is the msu of the result]  */
  res->digits=decGetDigits(res->lsu, static_cast<int32_t>(uc - res->lsu));
  res->exponent=0;                      /* integer  */
  res->bits=0;                          /* sign=0  */
  return res;  /* [no status to set]  */
  } /* decNumberInvert  */

/* ------------------------------------------------------------------ */
/* decNumberLn -- natural logarithm                                   */
/*                                                                    */
/*   This computes C = ln(A)                                          */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context; note that rounding mode has no effect        */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Notable cases:                                                     */
/*   A<0 -> Invalid                                                   */
/*   A=0 -> -Infinity (Exact)                                         */
/*   A=+Infinity -> +Infinity (Exact)                                 */
/*   A=1 exactly -> 0 (Exact)                                         */
/*                                                                    */
/* Mathematical function restrictions apply (see above); a NaN is     */
/* returned with Invalid_operation if a restriction is violated.      */
/*                                                                    */
/* An Inexact result is rounded using DEC_ROUND_HALF_EVEN; it will    */
/* almost always be correctly rounded, but may be up to 1 ulp in      */
/* error in rare cases.                                               */
/* ------------------------------------------------------------------ */
/* This is a wrapper for decLnOp which can handle the slightly wider  */
/* (+11) range needed by Ln, Log10, etc. (which may have to be able   */
/* to calculate at p+e+2).                                            */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberLn(decNumber *res, const decNumber *rhs,
                        decContext *set) {
  uInt status=0;                   /* accumulator  */
  #if DECSUBSET
  decNumber *allocrhs=NULL;        /* non-NULL if rounded rhs allocated  */
  #endif

  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  /* Check restrictions; this is a math function; if not violated  */
  /* then carry out the operation.  */
  if (!decCheckMath(rhs, set, &status)) do { /* protect allocation  */
    #if DECSUBSET
    if (!set->extended) {
      /* reduce operand and set lostDigits status, as needed  */
      if (rhs->digits>set->digits) {
        allocrhs=decRoundOperand(rhs, set, &status);
        if (allocrhs==NULL) break;
        rhs=allocrhs;
        }
      /* special check in subset for rhs=0  */
      if (ISZERO(rhs)) {                /* +/- zeros -> error  */
        status|=DEC_Invalid_operation;
        break;}
      } /* extended=0  */
    #endif
    decLnOp(res, rhs, set, &status);
    } while(0);                         /* end protected  */

  #if DECSUBSET
  if (allocrhs !=NULL) free(allocrhs);  /* drop any storage used  */
  #endif
  /* apply significant status  */
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberLn  */

/* ------------------------------------------------------------------ */
/* decNumberLogB - get adjusted exponent, by 754 rules                */
/*                                                                    */
/*   This computes C = adjustedexponent(A)                            */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context, used only for digits and status              */
/*                                                                    */
/* C must have space for 10 digits (A might have 10**9 digits and     */
/* an exponent of +999999999, or one digit and an exponent of         */
/* -1999999999).                                                      */
/*                                                                    */
/* This returns the adjusted exponent of A after (in theory) padding  */
/* with zeros on the right to set->digits digits while keeping the    */
/* same value.  The exponent is not limited by emin/emax.             */
/*                                                                    */
/* Notable cases:                                                     */
/*   A<0 -> Use |A|                                                   */
/*   A=0 -> -Infinity (Division by zero)                              */
/*   A=Infinite -> +Infinity (Exact)                                  */
/*   A=1 exactly -> 0 (Exact)                                         */
/*   NaNs are propagated as usual                                     */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberLogB(decNumber *res, const decNumber *rhs,
                          decContext *set) {
  uInt status=0;                   /* accumulator  */

  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  /* NaNs as usual; Infinities return +Infinity; 0->oops  */
  if (decNumberIsNaN(rhs)) decNaNs(res, rhs, NULL, set, &status);
   else if (decNumberIsInfinite(rhs)) uprv_decNumberCopyAbs(res, rhs);
   else if (decNumberIsZero(rhs)) {
    uprv_decNumberZero(res);                 /* prepare for Infinity  */
    res->bits=DECNEG|DECINF;            /* -Infinity  */
    status|=DEC_Division_by_zero;       /* as per 754  */
    }
   else { /* finite non-zero  */
    Int ae=rhs->exponent+rhs->digits-1; /* adjusted exponent  */
    uprv_decNumberFromInt32(res, ae);        /* lay it out  */
    }

  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberLogB  */

/* ------------------------------------------------------------------ */
/* decNumberLog10 -- logarithm in base 10                             */
/*                                                                    */
/*   This computes C = log10(A)                                       */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context; note that rounding mode has no effect        */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Notable cases:                                                     */
/*   A<0 -> Invalid                                                   */
/*   A=0 -> -Infinity (Exact)                                         */
/*   A=+Infinity -> +Infinity (Exact)                                 */
/*   A=10**n (if n is an integer) -> n (Exact)                        */
/*                                                                    */
/* Mathematical function restrictions apply (see above); a NaN is     */
/* returned with Invalid_operation if a restriction is violated.      */
/*                                                                    */
/* An Inexact result is rounded using DEC_ROUND_HALF_EVEN; it will    */
/* almost always be correctly rounded, but may be up to 1 ulp in      */
/* error in rare cases.                                               */
/* ------------------------------------------------------------------ */
/* This calculates ln(A)/ln(10) using appropriate precision.  For     */
/* ln(A) this is the max(p, rhs->digits + t) + 3, where p is the      */
/* requested digits and t is the number of digits in the exponent     */
/* (maximum 6).  For ln(10) it is p + 3; this is often handled by the */
/* fastpath in decLnOp.  The final division is done to the requested  */
/* precision.                                                         */
/* ------------------------------------------------------------------ */
#if defined(__clang__) || U_GCC_MAJOR_MINOR >= 406
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Warray-bounds"
#endif
U_CAPI decNumber * U_EXPORT2 uprv_decNumberLog10(decNumber *res, const decNumber *rhs,
                          decContext *set) {
  uInt status=0, ignore=0;         /* status accumulators  */
  uInt needbytes;                  /* for space calculations  */
  Int p;                           /* working precision  */
  Int t;                           /* digits in exponent of A  */

  /* buffers for a and b working decimals  */
  /* (adjustment calculator, same size)  */
  decNumber bufa[D2N(DECBUFFER+2)];
  decNumber *allocbufa=NULL;       /* -> allocated bufa, iff allocated  */
  decNumber *a=bufa;               /* temporary a  */
  decNumber bufb[D2N(DECBUFFER+2)];
  decNumber *allocbufb=NULL;       /* -> allocated bufb, iff allocated  */
  decNumber *b=bufb;               /* temporary b  */
  decNumber bufw[D2N(10)];         /* working 2-10 digit number  */
  decNumber *w=bufw;               /* ..  */
  #if DECSUBSET
  decNumber *allocrhs=NULL;        /* non-NULL if rounded rhs allocated  */
  #endif

  decContext aset;                 /* working context  */

  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  /* Check restrictions; this is a math function; if not violated  */
  /* then carry out the operation.  */
  if (!decCheckMath(rhs, set, &status)) do { /* protect malloc  */
    #if DECSUBSET
    if (!set->extended) {
      /* reduce operand and set lostDigits status, as needed  */
      if (rhs->digits>set->digits) {
        allocrhs=decRoundOperand(rhs, set, &status);
        if (allocrhs==NULL) break;
        rhs=allocrhs;
        }
      /* special check in subset for rhs=0  */
      if (ISZERO(rhs)) {                /* +/- zeros -> error  */
        status|=DEC_Invalid_operation;
        break;}
      } /* extended=0  */
    #endif

    uprv_decContextDefault(&aset, DEC_INIT_DECIMAL64); /* clean context  */

    /* handle exact DECNUMBER_powers of 10; only check if +ve finite  */
    if (!(rhs->bits&(DECNEG|DECSPECIAL)) && !ISZERO(rhs)) {
      Int residue=0;               /* (no residue)  */
      uInt copystat=0;             /* clean status  */

      /* round to a single digit...  */
      aset.digits=1;
      decCopyFit(w, rhs, &aset, &residue, &copystat); /* copy & shorten  */
      /* if exact and the digit is 1, rhs is a power of 10  */
      if (!(copystat&DEC_Inexact) && w->lsu[0]==1) {
        /* the exponent, conveniently, is the power of 10; making  */
        /* this the result needs a little care as it might not fit,  */
        /* so first convert it into the working number, and then move  */
        /* to res  */
        uprv_decNumberFromInt32(w, w->exponent);
        residue=0;
        decCopyFit(res, w, set, &residue, &status); /* copy & round  */
        decFinish(res, set, &residue, &status);     /* cleanup/set flags  */
        break;
        } /* not a power of 10  */
      } /* not a candidate for exact  */

    /* simplify the information-content calculation to use 'total  */
    /* number of digits in a, including exponent' as compared to the  */
    /* requested digits, as increasing this will only rarely cost an  */
    /* iteration in ln(a) anyway  */
    t=6;                                /* it can never be >6  */

    /* allocate space when needed...  */
    p=(rhs->digits+t>set->digits?rhs->digits+t:set->digits)+3;
    needbytes=sizeof(decNumber)+(D2U(p)-1)*sizeof(Unit);
    if (needbytes>sizeof(bufa)) {       /* need malloc space  */
      allocbufa=(decNumber *)malloc(needbytes);
      if (allocbufa==NULL) {            /* hopeless -- abandon  */
        status|=DEC_Insufficient_storage;
        break;}
      a=allocbufa;                      /* use the allocated space  */
      }
    aset.digits=p;                      /* as calculated  */
    aset.emax=DEC_MAX_MATH;             /* usual bounds  */
    aset.emin=-DEC_MAX_MATH;            /* ..  */
    aset.clamp=0;                       /* and no concrete format  */
    decLnOp(a, rhs, &aset, &status);    /* a=ln(rhs)  */

    /* skip the division if the result so far is infinite, NaN, or  */
    /* zero, or there was an error; note NaN from sNaN needs copy  */
    if (status&DEC_NaNs && !(status&DECNUMBER_DEC_sNaN)) break;
    if (a->bits&DECSPECIAL || ISZERO(a)) {
      uprv_decNumberCopy(res, a);            /* [will fit]  */
      break;}

    /* for ln(10) an extra 3 digits of precision are needed  */
    p=set->digits+3;
    needbytes=sizeof(decNumber)+(D2U(p)-1)*sizeof(Unit);
    if (needbytes>sizeof(bufb)) {       /* need malloc space  */
      allocbufb=(decNumber *)malloc(needbytes);
      if (allocbufb==NULL) {            /* hopeless -- abandon  */
        status|=DEC_Insufficient_storage;
        break;}
      b=allocbufb;                      /* use the allocated space  */
      }
    uprv_decNumberZero(w);                   /* set up 10...  */
    #if DECDPUN==1
    w->lsu[1]=1; w->lsu[0]=0;           /* ..  */
    #else
    w->lsu[0]=10;                       /* ..  */
    #endif
    w->digits=2;                        /* ..  */

    aset.digits=p;
    decLnOp(b, w, &aset, &ignore);      /* b=ln(10)  */

    aset.digits=set->digits;            /* for final divide  */
    decDivideOp(res, a, b, &aset, DECNUMBER_DIVIDE, &status); /* into result  */
    } while(0);                         /* [for break]  */

  if (allocbufa!=NULL) free(allocbufa); /* drop any storage used  */
  if (allocbufb!=NULL) free(allocbufb); /* ..  */
  #if DECSUBSET
  if (allocrhs !=NULL) free(allocrhs);  /* ..  */
  #endif
  /* apply significant status  */
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberLog10  */
#if defined(__clang__) || U_GCC_MAJOR_MINOR >= 406
#pragma GCC diagnostic pop
#endif

/* ------------------------------------------------------------------ */
/* decNumberMax -- compare two Numbers and return the maximum         */
/*                                                                    */
/*   This computes C = A ? B, returning the maximum by 754 rules      */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X?X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberMax(decNumber *res, const decNumber *lhs,
                         const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decCompareOp(res, lhs, rhs, set, DECNUMBER_COMPMAX, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberMax  */

/* ------------------------------------------------------------------ */
/* decNumberMaxMag -- compare and return the maximum by magnitude     */
/*                                                                    */
/*   This computes C = A ? B, returning the maximum by 754 rules      */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X?X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberMaxMag(decNumber *res, const decNumber *lhs,
                         const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decCompareOp(res, lhs, rhs, set, DECNUMBER_COMPMAXMAG, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberMaxMag  */

/* ------------------------------------------------------------------ */
/* decNumberMin -- compare two Numbers and return the minimum         */
/*                                                                    */
/*   This computes C = A ? B, returning the minimum by 754 rules      */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X?X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberMin(decNumber *res, const decNumber *lhs,
                         const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decCompareOp(res, lhs, rhs, set, DECNUMBER_COMPMIN, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberMin  */

/* ------------------------------------------------------------------ */
/* decNumberMinMag -- compare and return the minimum by magnitude     */
/*                                                                    */
/*   This computes C = A ? B, returning the minimum by 754 rules      */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X?X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberMinMag(decNumber *res, const decNumber *lhs,
                         const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decCompareOp(res, lhs, rhs, set, DECNUMBER_COMPMINMAG, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberMinMag  */

/* ------------------------------------------------------------------ */
/* decNumberMinus -- prefix minus operator                            */
/*                                                                    */
/*   This computes C = 0 - A                                          */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context                                               */
/*                                                                    */
/* See also decNumberCopyNegate for a quiet bitwise version of this.  */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
/* Simply use AddOp for the subtract, which will do the necessary.    */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberMinus(decNumber *res, const decNumber *rhs,
                           decContext *set) {
  decNumber dzero;
  uInt status=0;                        /* accumulator  */

  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  uprv_decNumberZero(&dzero);                /* make 0  */
  dzero.exponent=rhs->exponent;         /* [no coefficient expansion]  */
  decAddOp(res, &dzero, rhs, set, DECNEG, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberMinus  */

/* ------------------------------------------------------------------ */
/* decNumberNextMinus -- next towards -Infinity                       */
/*                                                                    */
/*   This computes C = A - infinitesimal, rounded towards -Infinity   */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context                                               */
/*                                                                    */
/* This is a generalization of 754 NextDown.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberNextMinus(decNumber *res, const decNumber *rhs,
                               decContext *set) {
  decNumber dtiny;                           /* constant  */
  decContext workset=*set;                   /* work  */
  uInt status=0;                             /* accumulator  */
  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  /* +Infinity is the special case  */
  if ((rhs->bits&(DECINF|DECNEG))==DECINF) {
    decSetMaxValue(res, set);                /* is +ve  */
    /* there is no status to set  */
    return res;
    }
  uprv_decNumberZero(&dtiny);                     /* start with 0  */
  dtiny.lsu[0]=1;                            /* make number that is ..  */
  dtiny.exponent=DEC_MIN_EMIN-1;             /* .. smaller than tiniest  */
  workset.round=DEC_ROUND_FLOOR;
  decAddOp(res, rhs, &dtiny, &workset, DECNEG, &status);
  status&=DEC_Invalid_operation|DECNUMBER_DEC_sNaN;    /* only sNaN Invalid please  */
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberNextMinus  */

/* ------------------------------------------------------------------ */
/* decNumberNextPlus -- next towards +Infinity                        */
/*                                                                    */
/*   This computes C = A + infinitesimal, rounded towards +Infinity   */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context                                               */
/*                                                                    */
/* This is a generalization of 754 NextUp.                            */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberNextPlus(decNumber *res, const decNumber *rhs,
                              decContext *set) {
  decNumber dtiny;                           /* constant  */
  decContext workset=*set;                   /* work  */
  uInt status=0;                             /* accumulator  */
  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  /* -Infinity is the special case  */
  if ((rhs->bits&(DECINF|DECNEG))==(DECINF|DECNEG)) {
    decSetMaxValue(res, set);
    res->bits=DECNEG;                        /* negative  */
    /* there is no status to set  */
    return res;
    }
  uprv_decNumberZero(&dtiny);                     /* start with 0  */
  dtiny.lsu[0]=1;                            /* make number that is ..  */
  dtiny.exponent=DEC_MIN_EMIN-1;             /* .. smaller than tiniest  */
  workset.round=DEC_ROUND_CEILING;
  decAddOp(res, rhs, &dtiny, &workset, 0, &status);
  status&=DEC_Invalid_operation|DECNUMBER_DEC_sNaN;    /* only sNaN Invalid please  */
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberNextPlus  */

/* ------------------------------------------------------------------ */
/* decNumberNextToward -- next towards rhs                            */
/*                                                                    */
/*   This computes C = A +/- infinitesimal, rounded towards           */
/*   +/-Infinity in the direction of B, as per 754-1985 nextafter     */
/*   modified during revision but dropped from 754-2008.              */
/*                                                                    */
/*   res is C, the result.  C may be A or B.                          */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* This is a generalization of 754-1985 NextAfter.                    */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberNextToward(decNumber *res, const decNumber *lhs,
                                const decNumber *rhs, decContext *set) {
  decNumber dtiny;                           /* constant  */
  decContext workset=*set;                   /* work  */
  Int result;                                /* ..  */
  uInt status=0;                             /* accumulator  */
  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  if (decNumberIsNaN(lhs) || decNumberIsNaN(rhs)) {
    decNaNs(res, lhs, rhs, set, &status);
    }
   else { /* Is numeric, so no chance of sNaN Invalid, etc.  */
    result=decCompare(lhs, rhs, 0);     /* sign matters  */
    if (result==DECNUMBER_BADINT) status|=DEC_Insufficient_storage; /* rare  */
     else { /* valid compare  */
      if (result==0) uprv_decNumberCopySign(res, lhs, rhs); /* easy  */
       else { /* differ: need NextPlus or NextMinus  */
        uByte sub;                      /* add or subtract  */
        if (result<0) {                 /* lhs<rhs, do nextplus  */
          /* -Infinity is the special case  */
          if ((lhs->bits&(DECINF|DECNEG))==(DECINF|DECNEG)) {
            decSetMaxValue(res, set);
            res->bits=DECNEG;           /* negative  */
            return res;                 /* there is no status to set  */
            }
          workset.round=DEC_ROUND_CEILING;
          sub=0;                        /* add, please  */
          } /* plus  */
         else {                         /* lhs>rhs, do nextminus  */
          /* +Infinity is the special case  */
          if ((lhs->bits&(DECINF|DECNEG))==DECINF) {
            decSetMaxValue(res, set);
            return res;                 /* there is no status to set  */
            }
          workset.round=DEC_ROUND_FLOOR;
          sub=DECNEG;                   /* subtract, please  */
          } /* minus  */
        uprv_decNumberZero(&dtiny);          /* start with 0  */
        dtiny.lsu[0]=1;                 /* make number that is ..  */
        dtiny.exponent=DEC_MIN_EMIN-1;  /* .. smaller than tiniest  */
        decAddOp(res, lhs, &dtiny, &workset, sub, &status); /* + or -  */
        /* turn off exceptions if the result is a normal number  */
        /* (including Nmin), otherwise let all status through  */
        if (uprv_decNumberIsNormal(res, set)) status=0;
        } /* unequal  */
      } /* compare OK  */
    } /* numeric  */
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberNextToward  */

/* ------------------------------------------------------------------ */
/* decNumberOr -- OR two Numbers, digitwise                           */
/*                                                                    */
/*   This computes C = A | B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X|X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context (used for result length and error report)     */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Logical function restrictions apply (see above); a NaN is          */
/* returned with Invalid_operation if a restriction is violated.      */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberOr(decNumber *res, const decNumber *lhs,
                        const decNumber *rhs, decContext *set) {
  const Unit *ua, *ub;                  /* -> operands  */
  const Unit *msua, *msub;              /* -> operand msus  */
  Unit  *uc, *msuc;                     /* -> result and its msu  */
  Int   msudigs;                        /* digits in res msu  */
  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  if (lhs->exponent!=0 || decNumberIsSpecial(lhs) || decNumberIsNegative(lhs)
   || rhs->exponent!=0 || decNumberIsSpecial(rhs) || decNumberIsNegative(rhs)) {
    decStatus(res, DEC_Invalid_operation, set);
    return res;
    }
  /* operands are valid  */
  ua=lhs->lsu;                          /* bottom-up  */
  ub=rhs->lsu;                          /* ..  */
  uc=res->lsu;                          /* ..  */
  msua=ua+D2U(lhs->digits)-1;           /* -> msu of lhs  */
  msub=ub+D2U(rhs->digits)-1;           /* -> msu of rhs  */
  msuc=uc+D2U(set->digits)-1;           /* -> msu of result  */
  msudigs=MSUDIGITS(set->digits);       /* [faster than remainder]  */
  for (; uc<=msuc; ua++, ub++, uc++) {  /* Unit loop  */
    Unit a, b;                          /* extract units  */
    if (ua>msua) a=0;
     else a=*ua;
    if (ub>msub) b=0;
     else b=*ub;
    *uc=0;                              /* can now write back  */
    if (a|b) {                          /* maybe 1 bits to examine  */
      Int i, j;
      /* This loop could be unrolled and/or use BIN2BCD tables  */
      for (i=0; i<DECDPUN; i++) {
        if ((a|b)&1) *uc=*uc+(Unit)DECNUMBER_powers[i];     /* effect OR  */
        j=a%10;
        a=a/10;
        j|=b%10;
        b=b/10;
        if (j>1) {
          decStatus(res, DEC_Invalid_operation, set);
          return res;
          }
        if (uc==msuc && i==msudigs-1) break;      /* just did final digit  */
        } /* each digit  */
      } /* non-zero  */
    } /* each unit  */
  /* [here uc-1 is the msu of the result]  */
  res->digits=decGetDigits(res->lsu, static_cast<int32_t>(uc-res->lsu));
  res->exponent=0;                      /* integer  */
  res->bits=0;                          /* sign=0  */
  return res;  /* [no status to set]  */
  } /* decNumberOr  */

/* ------------------------------------------------------------------ */
/* decNumberPlus -- prefix plus operator                              */
/*                                                                    */
/*   This computes C = 0 + A                                          */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context                                               */
/*                                                                    */
/* See also decNumberCopy for a quiet bitwise version of this.        */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
/* This simply uses AddOp; Add will take fast path after preparing A. */
/* Performance is a concern here, as this routine is often used to    */
/* check operands and apply rounding and overflow/underflow testing.  */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberPlus(decNumber *res, const decNumber *rhs,
                          decContext *set) {
  decNumber dzero;
  uInt status=0;                        /* accumulator  */
  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  uprv_decNumberZero(&dzero);                /* make 0  */
  dzero.exponent=rhs->exponent;         /* [no coefficient expansion]  */
  decAddOp(res, &dzero, rhs, set, 0, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberPlus  */

/* ------------------------------------------------------------------ */
/* decNumberMultiply -- multiply two Numbers                          */
/*                                                                    */
/*   This computes C = A x B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X+X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberMultiply(decNumber *res, const decNumber *lhs,
                              const decNumber *rhs, decContext *set) {
  uInt status=0;                   /* accumulator  */
  decMultiplyOp(res, lhs, rhs, set, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberMultiply  */

/* ------------------------------------------------------------------ */
/* decNumberPower -- raise a number to a power                        */
/*                                                                    */
/*   This computes C = A ** B                                         */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X**X)        */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Mathematical function restrictions apply (see above); a NaN is     */
/* returned with Invalid_operation if a restriction is violated.      */
/*                                                                    */
/* However, if 1999999997<=B<=999999999 and B is an integer then the  */
/* restrictions on A and the context are relaxed to the usual bounds, */
/* for compatibility with the earlier (integer power only) version    */
/* of this function.                                                  */
/*                                                                    */
/* When B is an integer, the result may be exact, even if rounded.    */
/*                                                                    */
/* The final result is rounded according to the context; it will      */
/* almost always be correctly rounded, but may be up to 1 ulp in      */
/* error in rare cases.                                               */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberPower(decNumber *res, const decNumber *lhs,
                           const decNumber *rhs, decContext *set) {
  #if DECSUBSET
  decNumber *alloclhs=NULL;        /* non-NULL if rounded lhs allocated  */
  decNumber *allocrhs=NULL;        /* .., rhs  */
  #endif
  decNumber *allocdac=NULL;        /* -> allocated acc buffer, iff used  */
  decNumber *allocinv=NULL;        /* -> allocated 1/x buffer, iff used  */
  Int   reqdigits=set->digits;     /* requested DIGITS  */
  Int   n;                         /* rhs in binary  */
  Flag  rhsint=0;                  /* 1 if rhs is an integer  */
  Flag  useint=0;                  /* 1 if can use integer calculation  */
  Flag  isoddint=0;                /* 1 if rhs is an integer and odd  */
  Int   i;                         /* work  */
  #if DECSUBSET
  Int   dropped;                   /* ..  */
  #endif
  uInt  needbytes;                 /* buffer size needed  */
  Flag  seenbit;                   /* seen a bit while powering  */
  Int   residue=0;                 /* rounding residue  */
  uInt  status=0;                  /* accumulators  */
  uByte bits=0;                    /* result sign if errors  */
  decContext aset;                 /* working context  */
  decNumber dnOne;                 /* work value 1...  */
  /* local accumulator buffer [a decNumber, with digits+elength+1 digits]  */
  decNumber dacbuff[D2N(DECBUFFER+9)];
  decNumber *dac=dacbuff;          /* -> result accumulator  */
  /* same again for possible 1/lhs calculation  */
  decNumber invbuff[D2N(DECBUFFER+9)];

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  do {                             /* protect allocated storage  */
    #if DECSUBSET
    if (!set->extended) { /* reduce operands and set status, as needed  */
      if (lhs->digits>reqdigits) {
        alloclhs=decRoundOperand(lhs, set, &status);
        if (alloclhs==NULL) break;
        lhs=alloclhs;
        }
      if (rhs->digits>reqdigits) {
        allocrhs=decRoundOperand(rhs, set, &status);
        if (allocrhs==NULL) break;
        rhs=allocrhs;
        }
      }
    #endif
    /* [following code does not require input rounding]  */

    /* handle NaNs and rhs Infinity (lhs infinity is harder)  */
    if (SPECIALARGS) {
      if (decNumberIsNaN(lhs) || decNumberIsNaN(rhs)) { /* NaNs  */
        decNaNs(res, lhs, rhs, set, &status);
        break;}
      if (decNumberIsInfinite(rhs)) {   /* rhs Infinity  */
        Flag rhsneg=rhs->bits&DECNEG;   /* save rhs sign  */
        if (decNumberIsNegative(lhs)    /* lhs<0  */
         && !decNumberIsZero(lhs))      /* ..  */
          status|=DEC_Invalid_operation;
         else {                         /* lhs >=0  */
          uprv_decNumberZero(&dnOne);        /* set up 1  */
          dnOne.lsu[0]=1;
          uprv_decNumberCompare(dac, lhs, &dnOne, set); /* lhs ? 1  */
          uprv_decNumberZero(res);           /* prepare for 0/1/Infinity  */
          if (decNumberIsNegative(dac)) {    /* lhs<1  */
            if (rhsneg) res->bits|=DECINF;   /* +Infinity [else is +0]  */
            }
           else if (dac->lsu[0]==0) {        /* lhs=1  */
            /* 1**Infinity is inexact, so return fully-padded 1.0000  */
            Int shift=set->digits-1;
            *res->lsu=1;                     /* was 0, make int 1  */
            res->digits=decShiftToMost(res->lsu, 1, shift);
            res->exponent=-shift;            /* make 1.0000...  */
            status|=DEC_Inexact|DEC_Rounded; /* deemed inexact  */
            }
           else {                            /* lhs>1  */
            if (!rhsneg) res->bits|=DECINF;  /* +Infinity [else is +0]  */
            }
          } /* lhs>=0  */
        break;}
      /* [lhs infinity drops through]  */
      } /* specials  */

    /* Original rhs may be an integer that fits and is in range  */
    n=decGetInt(rhs);
    if (n!=DECNUMBER_BADINT) {                    /* it is an integer  */
      rhsint=1;                         /* record the fact for 1**n  */
      isoddint=(Flag)n&1;               /* [works even if big]  */
      if (n!=DECNUMBER_BIGEVEN && n!=DECNUMBER_BIGODD)      /* can use integer path?  */
        useint=1;                       /* looks good  */
      }

    if (decNumberIsNegative(lhs)        /* -x ..  */
      && isoddint) bits=DECNEG;         /* .. to an odd power  */

    /* handle LHS infinity  */
    if (decNumberIsInfinite(lhs)) {     /* [NaNs already handled]  */
      uByte rbits=rhs->bits;            /* save  */
      uprv_decNumberZero(res);               /* prepare  */
      if (n==0) *res->lsu=1;            /* [-]Inf**0 => 1  */
       else {
        /* -Inf**nonint -> error  */
        if (!rhsint && decNumberIsNegative(lhs)) {
          status|=DEC_Invalid_operation;     /* -Inf**nonint is error  */
          break;}
        if (!(rbits & DECNEG)) bits|=DECINF; /* was not a **-n  */
        /* [otherwise will be 0 or -0]  */
        res->bits=bits;
        }
      break;}

    /* similarly handle LHS zero  */
    if (decNumberIsZero(lhs)) {
      if (n==0) {                            /* 0**0 => Error  */
        #if DECSUBSET
        if (!set->extended) {                /* [unless subset]  */
          uprv_decNumberZero(res);
          *res->lsu=1;                       /* return 1  */
          break;}
        #endif
        status|=DEC_Invalid_operation;
        }
       else {                                /* 0**x  */
        uByte rbits=rhs->bits;               /* save  */
        if (rbits & DECNEG) {                /* was a 0**(-n)  */
          #if DECSUBSET
          if (!set->extended) {              /* [bad if subset]  */
            status|=DEC_Invalid_operation;
            break;}
          #endif
          bits|=DECINF;
          }
        uprv_decNumberZero(res);                  /* prepare  */
        /* [otherwise will be 0 or -0]  */
        res->bits=bits;
        }
      break;}

    /* here both lhs and rhs are finite; rhs==0 is handled in the  */
    /* integer path.  Next handle the non-integer cases  */
    if (!useint) {                      /* non-integral rhs  */
      /* any -ve lhs is bad, as is either operand or context out of  */
      /* bounds  */
      if (decNumberIsNegative(lhs)) {
        status|=DEC_Invalid_operation;
        break;}
      if (decCheckMath(lhs, set, &status)
       || decCheckMath(rhs, set, &status)) break; /* variable status  */

      uprv_decContextDefault(&aset, DEC_INIT_DECIMAL64); /* clean context  */
      aset.emax=DEC_MAX_MATH;           /* usual bounds  */
      aset.emin=-DEC_MAX_MATH;          /* ..  */
      aset.clamp=0;                     /* and no concrete format  */

      /* calculate the result using exp(ln(lhs)*rhs), which can  */
      /* all be done into the accumulator, dac.  The precision needed  */
      /* is enough to contain the full information in the lhs (which  */
      /* is the total digits, including exponent), or the requested  */
      /* precision, if larger, + 4; 6 is used for the exponent  */
      /* maximum length, and this is also used when it is shorter  */
      /* than the requested digits as it greatly reduces the >0.5 ulp  */
      /* cases at little cost (because Ln doubles digits each  */
      /* iteration so a few extra digits rarely causes an extra  */
      /* iteration)  */
      aset.digits=MAXI(lhs->digits, set->digits)+6+4;
      } /* non-integer rhs  */

     else { /* rhs is in-range integer  */
      if (n==0) {                       /* x**0 = 1  */
        /* (0**0 was handled above)  */
        uprv_decNumberZero(res);             /* result=1  */
        *res->lsu=1;                    /* ..  */
        break;}
      /* rhs is a non-zero integer  */
      if (n<0) n=-n;                    /* use abs(n)  */

      aset=*set;                        /* clone the context  */
      aset.round=DEC_ROUND_HALF_EVEN;   /* internally use balanced  */
      /* calculate the working DIGITS  */
      aset.digits=reqdigits+(rhs->digits+rhs->exponent)+2;
      #if DECSUBSET
      if (!set->extended) aset.digits--;     /* use classic precision  */
      #endif
      /* it's an error if this is more than can be handled  */
      if (aset.digits>DECNUMMAXP) {status|=DEC_Invalid_operation; break;}
      } /* integer path  */

    /* aset.digits is the count of digits for the accumulator needed  */
    /* if accumulator is too long for local storage, then allocate  */
    needbytes=sizeof(decNumber)+(D2U(aset.digits)-1)*sizeof(Unit);
    /* [needbytes also used below if 1/lhs needed]  */
    if (needbytes>sizeof(dacbuff)) {
      allocdac=(decNumber *)malloc(needbytes);
      if (allocdac==NULL) {   /* hopeless -- abandon  */
        status|=DEC_Insufficient_storage;
        break;}
      dac=allocdac;           /* use the allocated space  */
      }
    /* here, aset is set up and accumulator is ready for use  */

    if (!useint) {                           /* non-integral rhs  */
      /* x ** y; special-case x=1 here as it will otherwise always  */
      /* reduce to integer 1; decLnOp has a fastpath which detects  */
      /* the case of x=1  */
      decLnOp(dac, lhs, &aset, &status);     /* dac=ln(lhs)  */
      /* [no error possible, as lhs 0 already handled]  */
      if (ISZERO(dac)) {                     /* x==1, 1.0, etc.  */
        /* need to return fully-padded 1.0000 etc., but rhsint->1  */
        *dac->lsu=1;                         /* was 0, make int 1  */
        if (!rhsint) {                       /* add padding  */
          Int shift=set->digits-1;
          dac->digits=decShiftToMost(dac->lsu, 1, shift);
          dac->exponent=-shift;              /* make 1.0000...  */
          status|=DEC_Inexact|DEC_Rounded;   /* deemed inexact  */
          }
        }
       else {
        decMultiplyOp(dac, dac, rhs, &aset, &status);  /* dac=dac*rhs  */
        decExpOp(dac, dac, &aset, &status);            /* dac=exp(dac)  */
        }
      /* and drop through for final rounding  */
      } /* non-integer rhs  */

     else {                             /* carry on with integer  */
      uprv_decNumberZero(dac);               /* acc=1  */
      *dac->lsu=1;                      /* ..  */

      /* if a negative power the constant 1 is needed, and if not subset  */
      /* invert the lhs now rather than inverting the result later  */
      if (decNumberIsNegative(rhs)) {   /* was a **-n [hence digits>0]  */
        decNumber *inv=invbuff;         /* asssume use fixed buffer  */
        uprv_decNumberCopy(&dnOne, dac);     /* dnOne=1;  [needed now or later]  */
        #if DECSUBSET
        if (set->extended) {            /* need to calculate 1/lhs  */
        #endif
          /* divide lhs into 1, putting result in dac [dac=1/dac]  */
          decDivideOp(dac, &dnOne, lhs, &aset, DECNUMBER_DIVIDE, &status);
          /* now locate or allocate space for the inverted lhs  */
          if (needbytes>sizeof(invbuff)) {
            allocinv=(decNumber *)malloc(needbytes);
            if (allocinv==NULL) {       /* hopeless -- abandon  */
              status|=DEC_Insufficient_storage;
              break;}
            inv=allocinv;               /* use the allocated space  */
            }
          /* [inv now points to big-enough buffer or allocated storage]  */
          uprv_decNumberCopy(inv, dac);      /* copy the 1/lhs  */
          uprv_decNumberCopy(dac, &dnOne);   /* restore acc=1  */
          lhs=inv;                      /* .. and go forward with new lhs  */
        #if DECSUBSET
          }
        #endif
        }

      /* Raise-to-the-power loop...  */
      seenbit=0;                   /* set once a 1-bit is encountered  */
      for (i=1;;i++){              /* for each bit [top bit ignored]  */
        /* abandon if had overflow or terminal underflow  */
        if (status & (DEC_Overflow|DEC_Underflow)) { /* interesting?  */
          if (status&DEC_Overflow || ISZERO(dac)) break;
          }
        /* [the following two lines revealed an optimizer bug in a C++  */
        /* compiler, with symptom: 5**3 -> 25, when n=n+n was used]  */
        n=n<<1;                    /* move next bit to testable position  */
        if (n<0) {                 /* top bit is set  */
          seenbit=1;               /* OK, significant bit seen  */
          decMultiplyOp(dac, dac, lhs, &aset, &status); /* dac=dac*x  */
          }
        if (i==31) break;          /* that was the last bit  */
        if (!seenbit) continue;    /* no need to square 1  */
        decMultiplyOp(dac, dac, dac, &aset, &status); /* dac=dac*dac [square]  */
        } /*i*/ /* 32 bits  */

      /* complete internal overflow or underflow processing  */
      if (status & (DEC_Overflow|DEC_Underflow)) {
        #if DECSUBSET
        /* If subset, and power was negative, reverse the kind of -erflow  */
        /* [1/x not yet done]  */
        if (!set->extended && decNumberIsNegative(rhs)) {
          if (status & DEC_Overflow)
            status^=DEC_Overflow | DEC_Underflow | DEC_Subnormal;
           else { /* trickier -- Underflow may or may not be set  */
            status&=~(DEC_Underflow | DEC_Subnormal); /* [one or both]  */
            status|=DEC_Overflow;
            }
          }
        #endif
        dac->bits=(dac->bits & ~DECNEG) | bits; /* force correct sign  */
        /* round subnormals [to set.digits rather than aset.digits]  */
        /* or set overflow result similarly as required  */
        decFinalize(dac, set, &residue, &status);
        uprv_decNumberCopy(res, dac);   /* copy to result (is now OK length)  */
        break;
        }

      #if DECSUBSET
      if (!set->extended &&                  /* subset math  */
          decNumberIsNegative(rhs)) {        /* was a **-n [hence digits>0]  */
        /* so divide result into 1 [dac=1/dac]  */
        decDivideOp(dac, &dnOne, dac, &aset, DECNUMBER_DIVIDE, &status);
        }
      #endif
      } /* rhs integer path  */

    /* reduce result to the requested length and copy to result  */
    decCopyFit(res, dac, set, &residue, &status);
    decFinish(res, set, &residue, &status);  /* final cleanup  */
    #if DECSUBSET
    if (!set->extended) decTrim(res, set, 0, 1, &dropped); /* trailing zeros  */
    #endif
    } while(0);                         /* end protected  */

  if (allocdac!=NULL) free(allocdac);   /* drop any storage used  */
  if (allocinv!=NULL) free(allocinv);   /* ..  */
  #if DECSUBSET
  if (alloclhs!=NULL) free(alloclhs);   /* ..  */
  if (allocrhs!=NULL) free(allocrhs);   /* ..  */
  #endif
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberPower  */

/* ------------------------------------------------------------------ */
/* decNumberQuantize -- force exponent to requested value             */
/*                                                                    */
/*   This computes C = op(A, B), where op adjusts the coefficient     */
/*   of C (by rounding or shifting) such that the exponent (-scale)   */
/*   of C has exponent of B.  The numerical value of C will equal A,  */
/*   except for the effects of any rounding that occurred.            */
/*                                                                    */
/*   res is C, the result.  C may be A or B                           */
/*   lhs is A, the number to adjust                                   */
/*   rhs is B, the number with exponent to match                      */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Unless there is an error or the result is infinite, the exponent   */
/* after the operation is guaranteed to be equal to that of B.        */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberQuantize(decNumber *res, const decNumber *lhs,
                              const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decQuantizeOp(res, lhs, rhs, set, 1, &status);
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberQuantize  */

/* ------------------------------------------------------------------ */
/* decNumberReduce -- remove trailing zeros                           */
/*                                                                    */
/*   This computes C = 0 + A, and normalizes the result               */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
/* Previously known as Normalize  */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberNormalize(decNumber *res, const decNumber *rhs,
                               decContext *set) {
  return uprv_decNumberReduce(res, rhs, set);
  } /* decNumberNormalize  */

U_CAPI decNumber * U_EXPORT2 uprv_decNumberReduce(decNumber *res, const decNumber *rhs,
                            decContext *set) {
  #if DECSUBSET
  decNumber *allocrhs=NULL;        /* non-NULL if rounded rhs allocated  */
  #endif
  uInt status=0;                   /* as usual  */
  Int  residue=0;                  /* as usual  */
  Int  dropped;                    /* work  */

  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  do {                             /* protect allocated storage  */
    #if DECSUBSET
    if (!set->extended) {
      /* reduce operand and set lostDigits status, as needed  */
      if (rhs->digits>set->digits) {
        allocrhs=decRoundOperand(rhs, set, &status);
        if (allocrhs==NULL) break;
        rhs=allocrhs;
        }
      }
    #endif
    /* [following code does not require input rounding]  */

    /* Infinities copy through; NaNs need usual treatment  */
    if (decNumberIsNaN(rhs)) {
      decNaNs(res, rhs, NULL, set, &status);
      break;
      }

    /* reduce result to the requested length and copy to result  */
    decCopyFit(res, rhs, set, &residue, &status); /* copy & round  */
    decFinish(res, set, &residue, &status);       /* cleanup/set flags  */
    decTrim(res, set, 1, 0, &dropped);            /* normalize in place  */
                                                  /* [may clamp]  */
    } while(0);                              /* end protected  */

  #if DECSUBSET
  if (allocrhs !=NULL) free(allocrhs);       /* ..  */
  #endif
  if (status!=0) decStatus(res, status, set);/* then report status  */
  return res;
  } /* decNumberReduce  */

/* ------------------------------------------------------------------ */
/* decNumberRescale -- force exponent to requested value              */
/*                                                                    */
/*   This computes C = op(A, B), where op adjusts the coefficient     */
/*   of C (by rounding or shifting) such that the exponent (-scale)   */
/*   of C has the value B.  The numerical value of C will equal A,    */
/*   except for the effects of any rounding that occurred.            */
/*                                                                    */
/*   res is C, the result.  C may be A or B                           */
/*   lhs is A, the number to adjust                                   */
/*   rhs is B, the requested exponent                                 */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Unless there is an error or the result is infinite, the exponent   */
/* after the operation is guaranteed to be equal to B.                */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberRescale(decNumber *res, const decNumber *lhs,
                             const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decQuantizeOp(res, lhs, rhs, set, 0, &status);
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberRescale  */

/* ------------------------------------------------------------------ */
/* decNumberRemainder -- divide and return remainder                  */
/*                                                                    */
/*   This computes C = A % B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X%X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberRemainder(decNumber *res, const decNumber *lhs,
                               const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decDivideOp(res, lhs, rhs, set, DECNUMBER_REMAINDER, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberRemainder  */

/* ------------------------------------------------------------------ */
/* decNumberRemainderNear -- divide and return remainder from nearest */
/*                                                                    */
/*   This computes C = A % B, where % is the IEEE remainder operator  */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X%X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberRemainderNear(decNumber *res, const decNumber *lhs,
                                   const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */
  decDivideOp(res, lhs, rhs, set, DECNUMBER_REMNEAR, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberRemainderNear  */

/* ------------------------------------------------------------------ */
/* decNumberRotate -- rotate the coefficient of a Number left/right   */
/*                                                                    */
/*   This computes C = A rot B  (in base ten and rotating set->digits */
/*   digits).                                                         */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=XrotX)       */
/*   lhs is A                                                         */
/*   rhs is B, the number of digits to rotate (-ve to right)          */
/*   set is the context                                               */
/*                                                                    */
/* The digits of the coefficient of A are rotated to the left (if B   */
/* is positive) or to the right (if B is negative) without adjusting  */
/* the exponent or the sign of A.  If lhs->digits is less than        */
/* set->digits the coefficient is padded with zeros on the left       */
/* before the rotate.  Any leading zeros in the result are removed    */
/* as usual.                                                          */
/*                                                                    */
/* B must be an integer (q=0) and in the range -set->digits through   */
/* +set->digits.                                                      */
/* C must have space for set->digits digits.                          */
/* NaNs are propagated as usual.  Infinities are unaffected (but      */
/* B must be valid).  No status is set unless B is invalid or an      */
/* operand is an sNaN.                                                */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberRotate(decNumber *res, const decNumber *lhs,
                           const decNumber *rhs, decContext *set) {
  uInt status=0;              /* accumulator  */
  Int  rotate;                /* rhs as an Int  */

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  /* NaNs propagate as normal  */
  if (decNumberIsNaN(lhs) || decNumberIsNaN(rhs))
    decNaNs(res, lhs, rhs, set, &status);
   /* rhs must be an integer  */
   else if (decNumberIsInfinite(rhs) || rhs->exponent!=0)
    status=DEC_Invalid_operation;
   else { /* both numeric, rhs is an integer  */
    rotate=decGetInt(rhs);                   /* [cannot fail]  */
    if (rotate==DECNUMBER_BADINT                       /* something bad ..  */
     || rotate==DECNUMBER_BIGODD || rotate==DECNUMBER_BIGEVEN    /* .. very big ..  */
     || abs(rotate)>set->digits)             /* .. or out of range  */
      status=DEC_Invalid_operation;
     else {                                  /* rhs is OK  */
      uprv_decNumberCopy(res, lhs);
      /* convert -ve rotate to equivalent positive rotation  */
      if (rotate<0) rotate=set->digits+rotate;
      if (rotate!=0 && rotate!=set->digits   /* zero or full rotation  */
       && !decNumberIsInfinite(res)) {       /* lhs was infinite  */
        /* left-rotate to do; 0 < rotate < set->digits  */
        uInt units, shift;                   /* work  */
        uInt msudigits;                      /* digits in result msu  */
        Unit *msu=res->lsu+D2U(res->digits)-1;    /* current msu  */
        Unit *msumax=res->lsu+D2U(set->digits)-1; /* rotation msu  */
        for (msu++; msu<=msumax; msu++) *msu=0;   /* ensure high units=0  */
        res->digits=set->digits;                  /* now full-length  */
        msudigits=MSUDIGITS(res->digits);         /* actual digits in msu  */

        /* rotation here is done in-place, in three steps  */
        /* 1. shift all to least up to one unit to unit-align final  */
        /*    lsd [any digits shifted out are rotated to the left,  */
        /*    abutted to the original msd (which may require split)]  */
        /*  */
        /*    [if there are no whole units left to rotate, the  */
        /*    rotation is now complete]  */
        /*  */
        /* 2. shift to least, from below the split point only, so that  */
        /*    the final msd is in the right place in its Unit [any  */
        /*    digits shifted out will fit exactly in the current msu,  */
        /*    left aligned, no split required]  */
        /*  */
        /* 3. rotate all the units by reversing left part, right  */
        /*    part, and then whole  */
        /*  */
        /* example: rotate right 8 digits (2 units + 2), DECDPUN=3.  */
        /*  */
        /*   start: 00a bcd efg hij klm npq  */
        /*  */
        /*      1a  000 0ab cde fgh|ijk lmn [pq saved]  */
        /*      1b  00p qab cde fgh|ijk lmn  */
        /*  */
        /*      2a  00p qab cde fgh|00i jkl [mn saved]  */
        /*      2b  mnp qab cde fgh|00i jkl  */
        /*  */
        /*      3a  fgh cde qab mnp|00i jkl  */
        /*      3b  fgh cde qab mnp|jkl 00i  */
        /*      3c  00i jkl mnp qab cde fgh  */

        /* Step 1: amount to shift is the partial right-rotate count  */
        rotate=set->digits-rotate;      /* make it right-rotate  */
        units=rotate/DECDPUN;           /* whole units to rotate  */
        shift=rotate%DECDPUN;           /* left-over digits count  */
        if (shift>0) {                  /* not an exact number of units  */
          uInt save=res->lsu[0]%DECNUMBER_powers[shift];    /* save low digit(s)  */
          decShiftToLeast(res->lsu, D2U(res->digits), shift);
          if (shift>msudigits) {        /* msumax-1 needs >0 digits  */
            uInt rem=save%DECNUMBER_powers[shift-msudigits];/* split save  */
            *msumax=(Unit)(save/DECNUMBER_powers[shift-msudigits]); /* and insert  */
            *(msumax-1)=*(msumax-1)
                       +(Unit)(rem*DECNUMBER_powers[DECDPUN-(shift-msudigits)]); /* ..  */
            }
           else { /* all fits in msumax  */
            *msumax=*msumax+(Unit)(save*DECNUMBER_powers[msudigits-shift]); /* [maybe *1]  */
            }
          } /* digits shift needed  */

        /* If whole units to rotate...  */
        if (units>0) {                  /* some to do  */
          /* Step 2: the units to touch are the whole ones in rotate,  */
          /*   if any, and the shift is DECDPUN-msudigits (which may be  */
          /*   0, again)  */
          shift=DECDPUN-msudigits;
          if (shift>0) {                /* not an exact number of units  */
            uInt save=res->lsu[0]%DECNUMBER_powers[shift];  /* save low digit(s)  */
            decShiftToLeast(res->lsu, units, shift);
            *msumax=*msumax+(Unit)(save*DECNUMBER_powers[msudigits]);
            } /* partial shift needed  */

          /* Step 3: rotate the units array using triple reverse  */
          /* (reversing is easy and fast)  */
          decReverse(res->lsu+units, msumax);     /* left part  */
          decReverse(res->lsu, res->lsu+units-1); /* right part  */
          decReverse(res->lsu, msumax);           /* whole  */
          } /* whole units to rotate  */
        /* the rotation may have left an undetermined number of zeros  */
        /* on the left, so true length needs to be calculated  */
        res->digits=decGetDigits(res->lsu, static_cast<int32_t>(msumax-res->lsu+1));
        } /* rotate needed  */
      } /* rhs OK  */
    } /* numerics  */
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberRotate  */

/* ------------------------------------------------------------------ */
/* decNumberSameQuantum -- test for equal exponents                   */
/*                                                                    */
/*   res is the result number, which will contain either 0 or 1       */
/*   lhs is a number to test                                          */
/*   rhs is the second (usually a pattern)                            */
/*                                                                    */
/* No errors are possible and no context is needed.                   */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberSameQuantum(decNumber *res, const decNumber *lhs,
                                 const decNumber *rhs) {
  Unit ret=0;                      /* return value  */

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, DECUNCONT)) return res;
  #endif

  if (SPECIALARGS) {
    if (decNumberIsNaN(lhs) && decNumberIsNaN(rhs)) ret=1;
     else if (decNumberIsInfinite(lhs) && decNumberIsInfinite(rhs)) ret=1;
     /* [anything else with a special gives 0]  */
    }
   else if (lhs->exponent==rhs->exponent) ret=1;

  uprv_decNumberZero(res);              /* OK to overwrite an operand now  */
  *res->lsu=ret;
  return res;
  } /* decNumberSameQuantum  */

/* ------------------------------------------------------------------ */
/* decNumberScaleB -- multiply by a power of 10                       */
/*                                                                    */
/* This computes C = A x 10**B where B is an integer (q=0) with       */
/* maximum magnitude 2*(emax+digits)                                  */
/*                                                                    */
/*   res is C, the result.  C may be A or B                           */
/*   lhs is A, the number to adjust                                   */
/*   rhs is B, the requested power of ten to use                      */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* The result may underflow or overflow.                              */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberScaleB(decNumber *res, const decNumber *lhs,
                            const decNumber *rhs, decContext *set) {
  Int  reqexp;                /* requested exponent change [B]  */
  uInt status=0;              /* accumulator  */
  Int  residue;               /* work  */

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  /* Handle special values except lhs infinite  */
  if (decNumberIsNaN(lhs) || decNumberIsNaN(rhs))
    decNaNs(res, lhs, rhs, set, &status);
    /* rhs must be an integer  */
   else if (decNumberIsInfinite(rhs) || rhs->exponent!=0)
    status=DEC_Invalid_operation;
   else {
    /* lhs is a number; rhs is a finite with q==0  */
    reqexp=decGetInt(rhs);                   /* [cannot fail]  */
    if (reqexp==DECNUMBER_BADINT                       /* something bad ..  */
     || reqexp==DECNUMBER_BIGODD || reqexp==DECNUMBER_BIGEVEN    /* .. very big ..  */
     || abs(reqexp)>(2*(set->digits+set->emax))) /* .. or out of range  */
      status=DEC_Invalid_operation;
     else {                                  /* rhs is OK  */
      uprv_decNumberCopy(res, lhs);               /* all done if infinite lhs  */
      if (!decNumberIsInfinite(res)) {       /* prepare to scale  */
        res->exponent+=reqexp;               /* adjust the exponent  */
        residue=0;
        decFinalize(res, set, &residue, &status); /* .. and check  */
        } /* finite LHS  */
      } /* rhs OK  */
    } /* rhs finite  */
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberScaleB  */

/* ------------------------------------------------------------------ */
/* decNumberShift -- shift the coefficient of a Number left or right  */
/*                                                                    */
/*   This computes C = A << B or C = A >> -B  (in base ten).          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X<<X)        */
/*   lhs is A                                                         */
/*   rhs is B, the number of digits to shift (-ve to right)           */
/*   set is the context                                               */
/*                                                                    */
/* The digits of the coefficient of A are shifted to the left (if B   */
/* is positive) or to the right (if B is negative) without adjusting  */
/* the exponent or the sign of A.                                     */
/*                                                                    */
/* B must be an integer (q=0) and in the range -set->digits through   */
/* +set->digits.                                                      */
/* C must have space for set->digits digits.                          */
/* NaNs are propagated as usual.  Infinities are unaffected (but      */
/* B must be valid).  No status is set unless B is invalid or an      */
/* operand is an sNaN.                                                */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberShift(decNumber *res, const decNumber *lhs,
                           const decNumber *rhs, decContext *set) {
  uInt status=0;              /* accumulator  */
  Int  shift;                 /* rhs as an Int  */

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  /* NaNs propagate as normal  */
  if (decNumberIsNaN(lhs) || decNumberIsNaN(rhs))
    decNaNs(res, lhs, rhs, set, &status);
   /* rhs must be an integer  */
   else if (decNumberIsInfinite(rhs) || rhs->exponent!=0)
    status=DEC_Invalid_operation;
   else { /* both numeric, rhs is an integer  */
    shift=decGetInt(rhs);                    /* [cannot fail]  */
    if (shift==DECNUMBER_BADINT                        /* something bad ..  */
     || shift==DECNUMBER_BIGODD || shift==DECNUMBER_BIGEVEN      /* .. very big ..  */
     || abs(shift)>set->digits)              /* .. or out of range  */
      status=DEC_Invalid_operation;
     else {                                  /* rhs is OK  */
      uprv_decNumberCopy(res, lhs);
      if (shift!=0 && !decNumberIsInfinite(res)) { /* something to do  */
        if (shift>0) {                       /* to left  */
          if (shift==set->digits) {          /* removing all  */
            *res->lsu=0;                     /* so place 0  */
            res->digits=1;                   /* ..  */
            }
           else {                            /*  */
            /* first remove leading digits if necessary  */
            if (res->digits+shift>set->digits) {
              decDecap(res, res->digits+shift-set->digits);
              /* that updated res->digits; may have gone to 1 (for a  */
              /* single digit or for zero  */
              }
            if (res->digits>1 || *res->lsu)  /* if non-zero..  */
              res->digits=decShiftToMost(res->lsu, res->digits, shift);
            } /* partial left  */
          } /* left  */
         else { /* to right  */
          if (-shift>=res->digits) {         /* discarding all  */
            *res->lsu=0;                     /* so place 0  */
            res->digits=1;                   /* ..  */
            }
           else {
            decShiftToLeast(res->lsu, D2U(res->digits), -shift);
            res->digits-=(-shift);
            }
          } /* to right  */
        } /* non-0 non-Inf shift  */
      } /* rhs OK  */
    } /* numerics  */
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberShift  */

/* ------------------------------------------------------------------ */
/* decNumberSquareRoot -- square root operator                        */
/*                                                                    */
/*   This computes C = squareroot(A)                                  */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context; note that rounding mode has no effect        */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
/* This uses the following varying-precision algorithm in:            */
/*                                                                    */
/*   Properly Rounded Variable Precision Square Root, T. E. Hull and  */
/*   A. Abrham, ACM Transactions on Mathematical Software, Vol 11 #3, */
/*   pp229-237, ACM, September 1985.                                  */
/*                                                                    */
/* The square-root is calculated using Newton's method, after which   */
/* a check is made to ensure the result is correctly rounded.         */
/*                                                                    */
/* % [Reformatted original Numerical Turing source code follows.]     */
/* function sqrt(x : real) : real                                     */
/* % sqrt(x) returns the properly rounded approximation to the square */
/* % root of x, in the precision of the calling environment, or it    */
/* % fails if x < 0.                                                  */
/* % t e hull and a abrham, august, 1984                              */
/* if x <= 0 then                                                     */
/*   if x < 0 then                                                    */
/*     assert false                                                   */
/*   else                                                             */
/*     result 0                                                       */
/*   end if                                                           */
/* end if                                                             */
/* var f := setexp(x, 0)  % fraction part of x   [0.1 <= x < 1]       */
/* var e := getexp(x)     % exponent part of x                        */
/* var approx : real                                                  */
/* if e mod 2 = 0  then                                               */
/*   approx := .259 + .819 * f   % approx to root of f                */
/* else                                                               */
/*   f := f/l0                   % adjustments                        */
/*   e := e + 1                  %   for odd                          */
/*   approx := .0819 + 2.59 * f  %   exponent                         */
/* end if                                                             */
/*                                                                    */
/* var p:= 3                                                          */
/* const maxp := currentprecision + 2                                 */
/* loop                                                               */
/*   p := min(2*p - 2, maxp)     % p = 4,6,10, . . . , maxp           */
/*   precision p                                                      */
/*   approx := .5 * (approx + f/approx)                               */
/*   exit when p = maxp                                               */
/* end loop                                                           */
/*                                                                    */
/* % approx is now within 1 ulp of the properly rounded square root   */
/* % of f; to ensure proper rounding, compare squares of (approx -    */
/* % l/2 ulp) and (approx + l/2 ulp) with f.                          */
/* p := currentprecision                                              */
/* begin                                                              */
/*   precision p + 2                                                  */
/*   const approxsubhalf := approx - setexp(.5, -p)                   */
/*   if mulru(approxsubhalf, approxsubhalf) > f then                  */
/*     approx := approx - setexp(.l, -p + 1)                          */
/*   else                                                             */
/*     const approxaddhalf := approx + setexp(.5, -p)                 */
/*     if mulrd(approxaddhalf, approxaddhalf) < f then                */
/*       approx := approx + setexp(.l, -p + 1)                        */
/*     end if                                                         */
/*   end if                                                           */
/* end                                                                */
/* result setexp(approx, e div 2)  % fix exponent                     */
/* end sqrt                                                           */
/* ------------------------------------------------------------------ */
#if defined(__clang__) || U_GCC_MAJOR_MINOR >= 406
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Warray-bounds"
#endif
U_CAPI decNumber * U_EXPORT2 uprv_decNumberSquareRoot(decNumber *res, const decNumber *rhs,
                                decContext *set) {
  decContext workset, approxset;   /* work contexts  */
  decNumber dzero;                 /* used for constant zero  */
  Int  maxp;                       /* largest working precision  */
  Int  workp;                      /* working precision  */
  Int  residue=0;                  /* rounding residue  */
  uInt status=0, ignore=0;         /* status accumulators  */
  uInt rstatus;                    /* ..  */
  Int  exp;                        /* working exponent  */
  Int  ideal;                      /* ideal (preferred) exponent  */
  Int  needbytes;                  /* work  */
  Int  dropped;                    /* ..  */

  #if DECSUBSET
  decNumber *allocrhs=NULL;        /* non-NULL if rounded rhs allocated  */
  #endif
  /* buffer for f [needs +1 in case DECBUFFER 0]  */
  decNumber buff[D2N(DECBUFFER+1)];
  /* buffer for a [needs +2 to match likely maxp]  */
  decNumber bufa[D2N(DECBUFFER+2)];
  /* buffer for temporary, b [must be same size as a]  */
  decNumber bufb[D2N(DECBUFFER+2)];
  decNumber *allocbuff=NULL;       /* -> allocated buff, iff allocated  */
  decNumber *allocbufa=NULL;       /* -> allocated bufa, iff allocated  */
  decNumber *allocbufb=NULL;       /* -> allocated bufb, iff allocated  */
  decNumber *f=buff;               /* reduced fraction  */
  decNumber *a=bufa;               /* approximation to result  */
  decNumber *b=bufb;               /* intermediate result  */
  /* buffer for temporary variable, up to 3 digits  */
  decNumber buft[D2N(3)];
  decNumber *t=buft;               /* up-to-3-digit constant or work  */

  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  do {                             /* protect allocated storage  */
    #if DECSUBSET
    if (!set->extended) {
      /* reduce operand and set lostDigits status, as needed  */
      if (rhs->digits>set->digits) {
        allocrhs=decRoundOperand(rhs, set, &status);
        if (allocrhs==NULL) break;
        /* [Note: 'f' allocation below could reuse this buffer if  */
        /* used, but as this is rare they are kept separate for clarity.]  */
        rhs=allocrhs;
        }
      }
    #endif
    /* [following code does not require input rounding]  */

    /* handle infinities and NaNs  */
    if (SPECIALARG) {
      if (decNumberIsInfinite(rhs)) {         /* an infinity  */
        if (decNumberIsNegative(rhs)) status|=DEC_Invalid_operation;
         else uprv_decNumberCopy(res, rhs);        /* +Infinity  */
        }
       else decNaNs(res, rhs, NULL, set, &status); /* a NaN  */
      break;
      }

    /* calculate the ideal (preferred) exponent [floor(exp/2)]  */
    /* [It would be nicer to write: ideal=rhs->exponent>>1, but this  */
    /* generates a compiler warning.  Generated code is the same.]  */
    ideal=(rhs->exponent&~1)/2;         /* target  */

    /* handle zeros  */
    if (ISZERO(rhs)) {
      uprv_decNumberCopy(res, rhs);          /* could be 0 or -0  */
      res->exponent=ideal;              /* use the ideal [safe]  */
      /* use decFinish to clamp any out-of-range exponent, etc.  */
      decFinish(res, set, &residue, &status);
      break;
      }

    /* any other -x is an oops  */
    if (decNumberIsNegative(rhs)) {
      status|=DEC_Invalid_operation;
      break;
      }

    /* space is needed for three working variables  */
    /*   f -- the same precision as the RHS, reduced to 0.01->0.99...  */
    /*   a -- Hull's approximation -- precision, when assigned, is  */
    /*        currentprecision+1 or the input argument precision,  */
    /*        whichever is larger (+2 for use as temporary)  */
    /*   b -- intermediate temporary result (same size as a)  */
    /* if any is too long for local storage, then allocate  */
    workp=MAXI(set->digits+1, rhs->digits);  /* actual rounding precision  */
    workp=MAXI(workp, 7);                    /* at least 7 for low cases  */
    maxp=workp+2;                            /* largest working precision  */

    needbytes=sizeof(decNumber)+(D2U(rhs->digits)-1)*sizeof(Unit);
    if (needbytes>(Int)sizeof(buff)) {
      allocbuff=(decNumber *)malloc(needbytes);
      if (allocbuff==NULL) {  /* hopeless -- abandon  */
        status|=DEC_Insufficient_storage;
        break;}
      f=allocbuff;            /* use the allocated space  */
      }
    /* a and b both need to be able to hold a maxp-length number  */
    needbytes=sizeof(decNumber)+(D2U(maxp)-1)*sizeof(Unit);
    if (needbytes>(Int)sizeof(bufa)) {            /* [same applies to b]  */
      allocbufa=(decNumber *)malloc(needbytes);
      allocbufb=(decNumber *)malloc(needbytes);
      if (allocbufa==NULL || allocbufb==NULL) {   /* hopeless  */
        status|=DEC_Insufficient_storage;
        break;}
      a=allocbufa;            /* use the allocated spaces  */
      b=allocbufb;            /* ..  */
      }

    /* copy rhs -> f, save exponent, and reduce so 0.1 <= f < 1  */
    uprv_decNumberCopy(f, rhs);
    exp=f->exponent+f->digits;               /* adjusted to Hull rules  */
    f->exponent=-(f->digits);                /* to range  */

    /* set up working context  */
    uprv_decContextDefault(&workset, DEC_INIT_DECIMAL64);
    workset.emax=DEC_MAX_EMAX;
    workset.emin=DEC_MIN_EMIN;

    /* [Until further notice, no error is possible and status bits  */
    /* (Rounded, etc.) should be ignored, not accumulated.]  */

    /* Calculate initial approximation, and allow for odd exponent  */
    workset.digits=workp;                    /* p for initial calculation  */
    t->bits=0; t->digits=3;
    a->bits=0; a->digits=3;
    if ((exp & 1)==0) {                      /* even exponent  */
      /* Set t=0.259, a=0.819  */
      t->exponent=-3;
      a->exponent=-3;
      #if DECDPUN>=3
        t->lsu[0]=259;
        a->lsu[0]=819;
      #elif DECDPUN==2
        t->lsu[0]=59; t->lsu[1]=2;
        a->lsu[0]=19; a->lsu[1]=8;
      #else
        t->lsu[0]=9; t->lsu[1]=5; t->lsu[2]=2;
        a->lsu[0]=9; a->lsu[1]=1; a->lsu[2]=8;
      #endif
      }
     else {                                  /* odd exponent  */
      /* Set t=0.0819, a=2.59  */
      f->exponent--;                         /* f=f/10  */
      exp++;                                 /* e=e+1  */
      t->exponent=-4;
      a->exponent=-2;
      #if DECDPUN>=3
        t->lsu[0]=819;
        a->lsu[0]=259;
      #elif DECDPUN==2
        t->lsu[0]=19; t->lsu[1]=8;
        a->lsu[0]=59; a->lsu[1]=2;
      #else
        t->lsu[0]=9; t->lsu[1]=1; t->lsu[2]=8;
        a->lsu[0]=9; a->lsu[1]=5; a->lsu[2]=2;
      #endif
      }

    decMultiplyOp(a, a, f, &workset, &ignore);    /* a=a*f  */
    decAddOp(a, a, t, &workset, 0, &ignore);      /* ..+t  */
    /* [a is now the initial approximation for sqrt(f), calculated with  */
    /* currentprecision, which is also a's precision.]  */

    /* the main calculation loop  */
    uprv_decNumberZero(&dzero);                   /* make 0  */
    uprv_decNumberZero(t);                        /* set t = 0.5  */
    t->lsu[0]=5;                             /* ..  */
    t->exponent=-1;                          /* ..  */
    workset.digits=3;                        /* initial p  */
    for (; workset.digits<maxp;) {
      /* set p to min(2*p - 2, maxp)  [hence 3; or: 4, 6, 10, ... , maxp]  */
      workset.digits=MINI(workset.digits*2-2, maxp);
      /* a = 0.5 * (a + f/a)  */
      /* [calculated at p then rounded to currentprecision]  */
      decDivideOp(b, f, a, &workset, DECNUMBER_DIVIDE, &ignore); /* b=f/a  */
      decAddOp(b, b, a, &workset, 0, &ignore);         /* b=b+a  */
      decMultiplyOp(a, b, t, &workset, &ignore);       /* a=b*0.5  */
      } /* loop  */

    /* Here, 0.1 <= a < 1 [Hull], and a has maxp digits  */
    /* now reduce to length, etc.; this needs to be done with a  */
    /* having the correct exponent so as to handle subnormals  */
    /* correctly  */
    approxset=*set;                          /* get emin, emax, etc.  */
    approxset.round=DEC_ROUND_HALF_EVEN;
    a->exponent+=exp/2;                      /* set correct exponent  */
    rstatus=0;                               /* clear status  */
    residue=0;                               /* .. and accumulator  */
    decCopyFit(a, a, &approxset, &residue, &rstatus);  /* reduce (if needed)  */
    decFinish(a, &approxset, &residue, &rstatus);      /* clean and finalize  */

    /* Overflow was possible if the input exponent was out-of-range,  */
    /* in which case quit  */
    if (rstatus&DEC_Overflow) {
      status=rstatus;                        /* use the status as-is  */
      uprv_decNumberCopy(res, a);                 /* copy to result  */
      break;
      }

    /* Preserve status except Inexact/Rounded  */
    status|=(rstatus & ~(DEC_Rounded|DEC_Inexact));

    /* Carry out the Hull correction  */
    a->exponent-=exp/2;                      /* back to 0.1->1  */

    /* a is now at final precision and within 1 ulp of the properly  */
    /* rounded square root of f; to ensure proper rounding, compare  */
    /* squares of (a - l/2 ulp) and (a + l/2 ulp) with f.  */
    /* Here workset.digits=maxp and t=0.5, and a->digits determines  */
    /* the ulp  */
    workset.digits--;                             /* maxp-1 is OK now  */
    t->exponent=-a->digits-1;                     /* make 0.5 ulp  */
    decAddOp(b, a, t, &workset, DECNEG, &ignore); /* b = a - 0.5 ulp  */
    workset.round=DEC_ROUND_UP;
    decMultiplyOp(b, b, b, &workset, &ignore);    /* b = mulru(b, b)  */
    decCompareOp(b, f, b, &workset, DECNUMBER_COMPARE, &ignore); /* b ? f, reversed  */
    if (decNumberIsNegative(b)) {                 /* f < b [i.e., b > f]  */
      /* this is the more common adjustment, though both are rare  */
      t->exponent++;                              /* make 1.0 ulp  */
      t->lsu[0]=1;                                /* ..  */
      decAddOp(a, a, t, &workset, DECNEG, &ignore); /* a = a - 1 ulp  */
      /* assign to approx [round to length]  */
      approxset.emin-=exp/2;                      /* adjust to match a  */
      approxset.emax-=exp/2;
      decAddOp(a, &dzero, a, &approxset, 0, &ignore);
      }
     else {
      decAddOp(b, a, t, &workset, 0, &ignore);    /* b = a + 0.5 ulp  */
      workset.round=DEC_ROUND_DOWN;
      decMultiplyOp(b, b, b, &workset, &ignore);  /* b = mulrd(b, b)  */
      decCompareOp(b, b, f, &workset, DECNUMBER_COMPARE, &ignore);   /* b ? f  */
      if (decNumberIsNegative(b)) {               /* b < f  */
        t->exponent++;                            /* make 1.0 ulp  */
        t->lsu[0]=1;                              /* ..  */
        decAddOp(a, a, t, &workset, 0, &ignore);  /* a = a + 1 ulp  */
        /* assign to approx [round to length]  */
        approxset.emin-=exp/2;                    /* adjust to match a  */
        approxset.emax-=exp/2;
        decAddOp(a, &dzero, a, &approxset, 0, &ignore);
        }
      }
    /* [no errors are possible in the above, and rounding/inexact during  */
    /* estimation are irrelevant, so status was not accumulated]  */

    /* Here, 0.1 <= a < 1  (still), so adjust back  */
    a->exponent+=exp/2;                      /* set correct exponent  */

    /* count droppable zeros [after any subnormal rounding] by  */
    /* trimming a copy  */
    uprv_decNumberCopy(b, a);
    decTrim(b, set, 1, 1, &dropped);         /* [drops trailing zeros]  */

    /* Set Inexact and Rounded.  The answer can only be exact if  */
    /* it is short enough so that squaring it could fit in workp  */
    /* digits, so this is the only (relatively rare) condition that  */
    /* a careful check is needed  */
    if (b->digits*2-1 > workp) {             /* cannot fit  */
      status|=DEC_Inexact|DEC_Rounded;
      }
     else {                                  /* could be exact/unrounded  */
      uInt mstatus=0;                        /* local status  */
      decMultiplyOp(b, b, b, &workset, &mstatus); /* try the multiply  */
      if (mstatus&DEC_Overflow) {            /* result just won't fit  */
        status|=DEC_Inexact|DEC_Rounded;
        }
       else {                                /* plausible  */
        decCompareOp(t, b, rhs, &workset, DECNUMBER_COMPARE, &mstatus); /* b ? rhs  */
        if (!ISZERO(t)) status|=DEC_Inexact|DEC_Rounded; /* not equal  */
         else {                              /* is Exact  */
          /* here, dropped is the count of trailing zeros in 'a'  */
          /* use closest exponent to ideal...  */
          Int todrop=ideal-a->exponent;      /* most that can be dropped  */
          if (todrop<0) status|=DEC_Rounded; /* ideally would add 0s  */
           else {                            /* unrounded  */
            /* there are some to drop, but emax may not allow all  */
            Int maxexp=set->emax-set->digits+1;
            Int maxdrop=maxexp-a->exponent;
            if (todrop>maxdrop && set->clamp) { /* apply clamping  */
              todrop=maxdrop;
              status|=DEC_Clamped;
              }
            if (dropped<todrop) {            /* clamp to those available  */
              todrop=dropped;
              status|=DEC_Clamped;
              }
            if (todrop>0) {                  /* have some to drop  */
              decShiftToLeast(a->lsu, D2U(a->digits), todrop);
              a->exponent+=todrop;           /* maintain numerical value  */
              a->digits-=todrop;             /* new length  */
              }
            }
          }
        }
      }

    /* double-check Underflow, as perhaps the result could not have  */
    /* been subnormal (initial argument too big), or it is now Exact  */
    if (status&DEC_Underflow) {
      Int ae=rhs->exponent+rhs->digits-1;    /* adjusted exponent  */
      /* check if truly subnormal  */
      #if DECEXTFLAG                         /* DEC_Subnormal too  */
        if (ae>=set->emin*2) status&=~(DEC_Subnormal|DEC_Underflow);
      #else
        if (ae>=set->emin*2) status&=~DEC_Underflow;
      #endif
      /* check if truly inexact  */
      if (!(status&DEC_Inexact)) status&=~DEC_Underflow;
      }

    uprv_decNumberCopy(res, a);                   /* a is now the result  */
    } while(0);                              /* end protected  */

  if (allocbuff!=NULL) free(allocbuff);      /* drop any storage used  */
  if (allocbufa!=NULL) free(allocbufa);      /* ..  */
  if (allocbufb!=NULL) free(allocbufb);      /* ..  */
  #if DECSUBSET
  if (allocrhs !=NULL) free(allocrhs);       /* ..  */
  #endif
  if (status!=0) decStatus(res, status, set);/* then report status  */
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberSquareRoot  */
#if defined(__clang__) || U_GCC_MAJOR_MINOR >= 406
#pragma GCC diagnostic pop
#endif

/* ------------------------------------------------------------------ */
/* decNumberSubtract -- subtract two Numbers                          */
/*                                                                    */
/*   This computes C = A - B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X-X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberSubtract(decNumber *res, const decNumber *lhs,
                              const decNumber *rhs, decContext *set) {
  uInt status=0;                        /* accumulator  */

  decAddOp(res, lhs, rhs, set, DECNEG, &status);
  if (status!=0) decStatus(res, status, set);
  #if DECCHECK
  decCheckInexact(res, set);
  #endif
  return res;
  } /* decNumberSubtract  */

/* ------------------------------------------------------------------ */
/* decNumberToIntegralExact -- round-to-integral-value with InExact   */
/* decNumberToIntegralValue -- round-to-integral-value                */
/*                                                                    */
/*   res is the result                                                */
/*   rhs is input number                                              */
/*   set is the context                                               */
/*                                                                    */
/* res must have space for any value of rhs.                          */
/*                                                                    */
/* This implements the IEEE special operators and therefore treats    */
/* special values as valid.  For finite numbers it returns            */
/* rescale(rhs, 0) if rhs->exponent is <0.                            */
/* Otherwise the result is rhs (so no error is possible, except for   */
/* sNaN).                                                             */
/*                                                                    */
/* The context is used for rounding mode and status after sNaN, but   */
/* the digits setting is ignored.  The Exact version will signal      */
/* Inexact if the result differs numerically from rhs; the other      */
/* never signals Inexact.                                             */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberToIntegralExact(decNumber *res, const decNumber *rhs,
                                     decContext *set) {
  decNumber dn;
  decContext workset;              /* working context  */
  uInt status=0;                   /* accumulator  */

  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  /* handle infinities and NaNs  */
  if (SPECIALARG) {
    if (decNumberIsInfinite(rhs)) uprv_decNumberCopy(res, rhs); /* an Infinity  */
     else decNaNs(res, rhs, NULL, set, &status); /* a NaN  */
    }
   else { /* finite  */
    /* have a finite number; no error possible (res must be big enough)  */
    if (rhs->exponent>=0) return uprv_decNumberCopy(res, rhs);
    /* that was easy, but if negative exponent there is work to do...  */
    workset=*set;                  /* clone rounding, etc.  */
    workset.digits=rhs->digits;    /* no length rounding  */
    workset.traps=0;               /* no traps  */
    uprv_decNumberZero(&dn);            /* make a number with exponent 0  */
    uprv_decNumberQuantize(res, rhs, &dn, &workset);
    status|=workset.status;
    }
  if (status!=0) decStatus(res, status, set);
  return res;
  } /* decNumberToIntegralExact  */

U_CAPI decNumber * U_EXPORT2 uprv_decNumberToIntegralValue(decNumber *res, const decNumber *rhs,
                                     decContext *set) {
  decContext workset=*set;         /* working context  */
  workset.traps=0;                 /* no traps  */
  uprv_decNumberToIntegralExact(res, rhs, &workset);
  /* this never affects set, except for sNaNs; NaN will have been set  */
  /* or propagated already, so no need to call decStatus  */
  set->status|=workset.status&DEC_Invalid_operation;
  return res;
  } /* decNumberToIntegralValue  */

/* ------------------------------------------------------------------ */
/* decNumberXor -- XOR two Numbers, digitwise                         */
/*                                                                    */
/*   This computes C = A ^ B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X^X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context (used for result length and error report)     */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Logical function restrictions apply (see above); a NaN is          */
/* returned with Invalid_operation if a restriction is violated.      */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberXor(decNumber *res, const decNumber *lhs,
                         const decNumber *rhs, decContext *set) {
  const Unit *ua, *ub;                  /* -> operands  */
  const Unit *msua, *msub;              /* -> operand msus  */
  Unit  *uc, *msuc;                     /* -> result and its msu  */
  Int   msudigs;                        /* digits in res msu  */
  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  if (lhs->exponent!=0 || decNumberIsSpecial(lhs) || decNumberIsNegative(lhs)
   || rhs->exponent!=0 || decNumberIsSpecial(rhs) || decNumberIsNegative(rhs)) {
    decStatus(res, DEC_Invalid_operation, set);
    return res;
    }
  /* operands are valid  */
  ua=lhs->lsu;                          /* bottom-up  */
  ub=rhs->lsu;                          /* ..  */
  uc=res->lsu;                          /* ..  */
  msua=ua+D2U(lhs->digits)-1;           /* -> msu of lhs  */
  msub=ub+D2U(rhs->digits)-1;           /* -> msu of rhs  */
  msuc=uc+D2U(set->digits)-1;           /* -> msu of result  */
  msudigs=MSUDIGITS(set->digits);       /* [faster than remainder]  */
  for (; uc<=msuc; ua++, ub++, uc++) {  /* Unit loop  */
    Unit a, b;                          /* extract units  */
    if (ua>msua) a=0;
     else a=*ua;
    if (ub>msub) b=0;
     else b=*ub;
    *uc=0;                              /* can now write back  */
    if (a|b) {                          /* maybe 1 bits to examine  */
      Int i, j;
      /* This loop could be unrolled and/or use BIN2BCD tables  */
      for (i=0; i<DECDPUN; i++) {
        if ((a^b)&1) *uc=*uc+(Unit)DECNUMBER_powers[i];     /* effect XOR  */
        j=a%10;
        a=a/10;
        j|=b%10;
        b=b/10;
        if (j>1) {
          decStatus(res, DEC_Invalid_operation, set);
          return res;
          }
        if (uc==msuc && i==msudigs-1) break;      /* just did final digit  */
        } /* each digit  */
      } /* non-zero  */
    } /* each unit  */
  /* [here uc-1 is the msu of the result]  */
  res->digits=decGetDigits(res->lsu, static_cast<int32_t>(uc-res->lsu));
  res->exponent=0;                      /* integer  */
  res->bits=0;                          /* sign=0  */
  return res;  /* [no status to set]  */
  } /* decNumberXor  */


/* ================================================================== */
/* Utility routines                                                   */
/* ================================================================== */

/* ------------------------------------------------------------------ */
/* decNumberClass -- return the decClass of a decNumber               */
/*   dn -- the decNumber to test                                      */
/*   set -- the context to use for Emin                               */
/*   returns the decClass enum                                        */
/* ------------------------------------------------------------------ */
enum decClass uprv_decNumberClass(const decNumber *dn, decContext *set) {
  if (decNumberIsSpecial(dn)) {
    if (decNumberIsQNaN(dn)) return DEC_CLASS_QNAN;
    if (decNumberIsSNaN(dn)) return DEC_CLASS_SNAN;
    /* must be an infinity  */
    if (decNumberIsNegative(dn)) return DEC_CLASS_NEG_INF;
    return DEC_CLASS_POS_INF;
    }
  /* is finite  */
  if (uprv_decNumberIsNormal(dn, set)) { /* most common  */
    if (decNumberIsNegative(dn)) return DEC_CLASS_NEG_NORMAL;
    return DEC_CLASS_POS_NORMAL;
    }
  /* is subnormal or zero  */
  if (decNumberIsZero(dn)) {    /* most common  */
    if (decNumberIsNegative(dn)) return DEC_CLASS_NEG_ZERO;
    return DEC_CLASS_POS_ZERO;
    }
  if (decNumberIsNegative(dn)) return DEC_CLASS_NEG_SUBNORMAL;
  return DEC_CLASS_POS_SUBNORMAL;
  } /* decNumberClass  */

/* ------------------------------------------------------------------ */
/* decNumberClassToString -- convert decClass to a string             */
/*                                                                    */
/*  eclass is a valid decClass                                        */
/*  returns a constant string describing the class (max 13+1 chars)   */
/* ------------------------------------------------------------------ */
const char *uprv_decNumberClassToString(enum decClass eclass) {
  if (eclass==DEC_CLASS_POS_NORMAL)    return DEC_ClassString_PN;
  if (eclass==DEC_CLASS_NEG_NORMAL)    return DEC_ClassString_NN;
  if (eclass==DEC_CLASS_POS_ZERO)      return DEC_ClassString_PZ;
  if (eclass==DEC_CLASS_NEG_ZERO)      return DEC_ClassString_NZ;
  if (eclass==DEC_CLASS_POS_SUBNORMAL) return DEC_ClassString_PS;
  if (eclass==DEC_CLASS_NEG_SUBNORMAL) return DEC_ClassString_NS;
  if (eclass==DEC_CLASS_POS_INF)       return DEC_ClassString_PI;
  if (eclass==DEC_CLASS_NEG_INF)       return DEC_ClassString_NI;
  if (eclass==DEC_CLASS_QNAN)          return DEC_ClassString_QN;
  if (eclass==DEC_CLASS_SNAN)          return DEC_ClassString_SN;
  return DEC_ClassString_UN;           /* Unknown  */
  } /* decNumberClassToString  */

/* ------------------------------------------------------------------ */
/* decNumberCopy -- copy a number                                     */
/*                                                                    */
/*   dest is the target decNumber                                     */
/*   src  is the source decNumber                                     */
/*   returns dest                                                     */
/*                                                                    */
/* (dest==src is allowed and is a no-op)                              */
/* All fields are updated as required.  This is a utility operation,  */
/* so special values are unchanged and no error is possible.          */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberCopy(decNumber *dest, const decNumber *src) {

  #if DECCHECK
  if (src==NULL) return uprv_decNumberZero(dest);
  #endif

  if (dest==src) return dest;                /* no copy required  */

  /* Use explicit assignments here as structure assignment could copy  */
  /* more than just the lsu (for small DECDPUN).  This would not affect  */
  /* the value of the results, but could disturb test harness spill  */
  /* checking.  */
  dest->bits=src->bits;
  dest->exponent=src->exponent;
  dest->digits=src->digits;
  dest->lsu[0]=src->lsu[0];
  if (src->digits>DECDPUN) {                 /* more Units to come  */
    const Unit *smsup, *s;                   /* work  */
    Unit  *d;                                /* ..  */
    /* memcpy for the remaining Units would be safe as they cannot  */
    /* overlap.  However, this explicit loop is faster in short cases.  */
    d=dest->lsu+1;                           /* -> first destination  */
    smsup=src->lsu+D2U(src->digits);         /* -> source msu+1  */
    for (s=src->lsu+1; s<smsup; s++, d++) *d=*s;
    }
  return dest;
  } /* decNumberCopy  */

/* ------------------------------------------------------------------ */
/* decNumberCopyAbs -- quiet absolute value operator                  */
/*                                                                    */
/*   This sets C = abs(A)                                             */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* No exception or error can occur; this is a quiet bitwise operation.*/
/* See also decNumberAbs for a checking version of this.              */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberCopyAbs(decNumber *res, const decNumber *rhs) {
  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, DECUNCONT)) return res;
  #endif
  uprv_decNumberCopy(res, rhs);
  res->bits&=~DECNEG;                   /* turn off sign  */
  return res;
  } /* decNumberCopyAbs  */

/* ------------------------------------------------------------------ */
/* decNumberCopyNegate -- quiet negate value operator                 */
/*                                                                    */
/*   This sets C = negate(A)                                          */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* No exception or error can occur; this is a quiet bitwise operation.*/
/* See also decNumberMinus for a checking version of this.            */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberCopyNegate(decNumber *res, const decNumber *rhs) {
  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, DECUNCONT)) return res;
  #endif
  uprv_decNumberCopy(res, rhs);
  res->bits^=DECNEG;                    /* invert the sign  */
  return res;
  } /* decNumberCopyNegate  */

/* ------------------------------------------------------------------ */
/* decNumberCopySign -- quiet copy and set sign operator              */
/*                                                                    */
/*   This sets C = A with the sign of B                               */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* No exception or error can occur; this is a quiet bitwise operation.*/
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberCopySign(decNumber *res, const decNumber *lhs,
                              const decNumber *rhs) {
  uByte sign;                           /* rhs sign  */
  #if DECCHECK
  if (decCheckOperands(res, DECUNUSED, rhs, DECUNCONT)) return res;
  #endif
  sign=rhs->bits & DECNEG;              /* save sign bit  */
  uprv_decNumberCopy(res, lhs);
  res->bits&=~DECNEG;                   /* clear the sign  */
  res->bits|=sign;                      /* set from rhs  */
  return res;
  } /* decNumberCopySign  */

/* ------------------------------------------------------------------ */
/* decNumberGetBCD -- get the coefficient in BCD8                     */
/*   dn is the source decNumber                                       */
/*   bcd is the uInt array that will receive dn->digits BCD bytes,    */
/*     most-significant at offset 0                                   */
/*   returns bcd                                                      */
/*                                                                    */
/* bcd must have at least dn->digits bytes.  No error is possible; if */
/* dn is a NaN or Infinite, digits must be 1 and the coefficient 0.   */
/* ------------------------------------------------------------------ */
U_CAPI uByte * U_EXPORT2 uprv_decNumberGetBCD(const decNumber *dn, uByte *bcd) {
  uByte *ub=bcd+dn->digits-1;      /* -> lsd  */
  const Unit *up=dn->lsu;          /* Unit pointer, -> lsu  */

  #if DECDPUN==1                   /* trivial simple copy  */
    for (; ub>=bcd; ub--, up++) *ub=*up;
  #else                            /* chopping needed  */
    uInt u=*up;                    /* work  */
    uInt cut=DECDPUN;              /* downcounter through unit  */
    for (; ub>=bcd; ub--) {
      *ub=(uByte)(u%10);           /* [*6554 trick inhibits, here]  */
      u=u/10;
      cut--;
      if (cut>0) continue;         /* more in this unit  */
      up++;
      u=*up;
      cut=DECDPUN;
      }
  #endif
  return bcd;
  } /* decNumberGetBCD  */

/* ------------------------------------------------------------------ */
/* decNumberSetBCD -- set (replace) the coefficient from BCD8         */
/*   dn is the target decNumber                                       */
/*   bcd is the uInt array that will source n BCD bytes, most-        */
/*     significant at offset 0                                        */
/*   n is the number of digits in the source BCD array (bcd)          */
/*   returns dn                                                       */
/*                                                                    */
/* dn must have space for at least n digits.  No error is possible;   */
/* if dn is a NaN, or Infinite, or is to become a zero, n must be 1   */
/* and bcd[0] zero.                                                   */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberSetBCD(decNumber *dn, const uByte *bcd, uInt n) {
  Unit *up=dn->lsu+D2U(dn->digits)-1;   /* -> msu [target pointer]  */
  const uByte *ub=bcd;                  /* -> source msd  */

  #if DECDPUN==1                        /* trivial simple copy  */
    for (; ub<bcd+n; ub++, up--) *up=*ub;
  #else                                 /* some assembly needed  */
    /* calculate how many digits in msu, and hence first cut  */
    Int cut=MSUDIGITS(n);               /* [faster than remainder]  */
    for (;up>=dn->lsu; up--) {          /* each Unit from msu  */
      *up=0;                            /* will take <=DECDPUN digits  */
      for (; cut>0; ub++, cut--) *up=X10(*up)+*ub;
      cut=DECDPUN;                      /* next Unit has all digits  */
      }
  #endif
  dn->digits=n;                         /* set digit count  */
  return dn;
  } /* decNumberSetBCD  */

/* ------------------------------------------------------------------ */
/* decNumberIsNormal -- test normality of a decNumber                 */
/*   dn is the decNumber to test                                      */
/*   set is the context to use for Emin                               */
/*   returns 1 if |dn| is finite and >=Nmin, 0 otherwise              */
/* ------------------------------------------------------------------ */
Int uprv_decNumberIsNormal(const decNumber *dn, decContext *set) {
  Int ae;                               /* adjusted exponent  */
  #if DECCHECK
  if (decCheckOperands(DECUNRESU, DECUNUSED, dn, set)) return 0;
  #endif

  if (decNumberIsSpecial(dn)) return 0; /* not finite  */
  if (decNumberIsZero(dn)) return 0;    /* not non-zero  */

  ae=dn->exponent+dn->digits-1;         /* adjusted exponent  */
  if (ae<set->emin) return 0;           /* is subnormal  */
  return 1;
  } /* decNumberIsNormal  */

/* ------------------------------------------------------------------ */
/* decNumberIsSubnormal -- test subnormality of a decNumber           */
/*   dn is the decNumber to test                                      */
/*   set is the context to use for Emin                               */
/*   returns 1 if |dn| is finite, non-zero, and <Nmin, 0 otherwise    */
/* ------------------------------------------------------------------ */
Int uprv_decNumberIsSubnormal(const decNumber *dn, decContext *set) {
  Int ae;                               /* adjusted exponent  */
  #if DECCHECK
  if (decCheckOperands(DECUNRESU, DECUNUSED, dn, set)) return 0;
  #endif

  if (decNumberIsSpecial(dn)) return 0; /* not finite  */
  if (decNumberIsZero(dn)) return 0;    /* not non-zero  */

  ae=dn->exponent+dn->digits-1;         /* adjusted exponent  */
  if (ae<set->emin) return 1;           /* is subnormal  */
  return 0;
  } /* decNumberIsSubnormal  */

/* ------------------------------------------------------------------ */
/* decNumberTrim -- remove insignificant zeros                        */
/*                                                                    */
/*   dn is the number to trim                                         */
/*   returns dn                                                       */
/*                                                                    */
/* All fields are updated as required.  This is a utility operation,  */
/* so special values are unchanged and no error is possible.  The     */
/* zeros are removed unconditionally.                                 */
/* ------------------------------------------------------------------ */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberTrim(decNumber *dn) {
  Int  dropped;                    /* work  */
  decContext set;                  /* ..  */
  #if DECCHECK
  if (decCheckOperands(DECUNRESU, DECUNUSED, dn, DECUNCONT)) return dn;
  #endif
  uprv_decContextDefault(&set, DEC_INIT_BASE);    /* clamp=0  */
  return decTrim(dn, &set, 0, 1, &dropped);
  } /* decNumberTrim  */

/* ------------------------------------------------------------------ */
/* decNumberVersion -- return the name and version of this module     */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
const char * uprv_decNumberVersion(void) {
  return DECVERSION;
  } /* decNumberVersion  */

/* ------------------------------------------------------------------ */
/* decNumberZero -- set a number to 0                                 */
/*                                                                    */
/*   dn is the number to set, with space for one digit                */
/*   returns dn                                                       */
/*                                                                    */
/* No error is possible.                                              */
/* ------------------------------------------------------------------ */
/* Memset is not used as it is much slower in some environments.  */
U_CAPI decNumber * U_EXPORT2 uprv_decNumberZero(decNumber *dn) {

  #if DECCHECK
  if (decCheckOperands(dn, DECUNUSED, DECUNUSED, DECUNCONT)) return dn;
  #endif

  dn->bits=0;
  dn->exponent=0;
  dn->digits=1;
  dn->lsu[0]=0;
  return dn;
  } /* decNumberZero  */

/* ================================================================== */
/* Local routines                                                     */
/* ================================================================== */

/* ------------------------------------------------------------------ */
/* decToString -- lay out a number into a string                      */
/*                                                                    */
/*   dn     is the number to lay out                                  */
/*   string is where to lay out the number                            */
/*   eng    is 1 if Engineering, 0 if Scientific                      */
/*                                                                    */
/* string must be at least dn->digits+14 characters long              */
/* No error is possible.                                              */
/*                                                                    */
/* Note that this routine can generate a -0 or 0.000.  These are      */
/* never generated in subset to-number or arithmetic, but can occur   */
/* in non-subset arithmetic (e.g., -1*0 or 1.234-1.234).              */
/* ------------------------------------------------------------------ */
/* If DECCHECK is enabled the string "?" is returned if a number is  */
/* invalid.  */
static void decToString(const decNumber *dn, char *string, Flag eng) {
  Int exp=dn->exponent;       /* local copy  */
  Int e;                      /* E-part value  */
  Int pre;                    /* digits before the '.'  */
  Int cut;                    /* for counting digits in a Unit  */
  char *c=string;             /* work [output pointer]  */
  const Unit *up=dn->lsu+D2U(dn->digits)-1; /* -> msu [input pointer]  */
  uInt u, pow;                /* work  */

  #if DECCHECK
  if (decCheckOperands(DECUNRESU, dn, DECUNUSED, DECUNCONT)) {
    strcpy(string, "?");
    return;}
  #endif

  if (decNumberIsNegative(dn)) {   /* Negatives get a minus  */
    *c='-';
    c++;
    }
  if (dn->bits&DECSPECIAL) {       /* Is a special value  */
    if (decNumberIsInfinite(dn)) {
      strcpy(c,   "Inf");
      strcpy(c+3, "inity");
      return;}
    /* a NaN  */
    if (dn->bits&DECSNAN) {        /* signalling NaN  */
      *c='s';
      c++;
      }
    strcpy(c, "NaN");
    c+=3;                          /* step past  */
    /* if not a clean non-zero coefficient, that's all there is in a  */
    /* NaN string  */
    if (exp!=0 || (*dn->lsu==0 && dn->digits==1)) return;
    /* [drop through to add integer]  */
    }

  /* calculate how many digits in msu, and hence first cut  */
  cut=MSUDIGITS(dn->digits);       /* [faster than remainder]  */
  cut--;                           /* power of ten for digit  */

  if (exp==0) {                    /* simple integer [common fastpath]  */
    for (;up>=dn->lsu; up--) {     /* each Unit from msu  */
      u=*up;                       /* contains DECDPUN digits to lay out  */
      for (; cut>=0; c++, cut--) TODIGIT(u, cut, c, pow);
      cut=DECDPUN-1;               /* next Unit has all digits  */
      }
    *c='\0';                       /* terminate the string  */
    return;}

  /* non-0 exponent -- assume plain form */
  pre=dn->digits+exp;              /* digits before '.'  */
  e=0;                             /* no E  */
  if ((exp>0) || (pre<-5)) {       /* need exponential form  */
    e=exp+dn->digits-1;            /* calculate E value  */
    pre=1;                         /* assume one digit before '.'  */
    if (eng && (e!=0)) {           /* engineering: may need to adjust  */
      Int adj;                     /* adjustment  */
      /* The C remainder operator is undefined for negative numbers, so  */
      /* a positive remainder calculation must be used here  */
      if (e<0) {
        adj=(-e)%3;
        if (adj!=0) adj=3-adj;
        }
       else { /* e>0  */
        adj=e%3;
        }
      e=e-adj;
      /* if dealing with zero still produce an exponent which is a  */
      /* multiple of three, as expected, but there will only be the  */
      /* one zero before the E, still.  Otherwise note the padding.  */
      if (!ISZERO(dn)) pre+=adj;
       else {  /* is zero  */
        if (adj!=0) {              /* 0.00Esnn needed  */
          e=e+3;
          pre=-(2-adj);
          }
        } /* zero  */
      } /* eng  */
    } /* need exponent  */

  /* lay out the digits of the coefficient, adding 0s and . as needed */
  u=*up;
  if (pre>0) {                     /* xxx.xxx or xx00 (engineering) form  */
    Int n=pre;
    for (; pre>0; pre--, c++, cut--) {
      if (cut<0) {                 /* need new Unit  */
        if (up==dn->lsu) break;    /* out of input digits (pre>digits)  */
        up--;
        cut=DECDPUN-1;
        u=*up;
        }
      TODIGIT(u, cut, c, pow);
      }
    if (n<dn->digits) {            /* more to come, after '.'  */
      *c='.'; c++;
      for (;; c++, cut--) {
        if (cut<0) {               /* need new Unit  */
          if (up==dn->lsu) break;  /* out of input digits  */
          up--;
          cut=DECDPUN-1;
          u=*up;
          }
        TODIGIT(u, cut, c, pow);
        }
      }
     else for (; pre>0; pre--, c++) *c='0'; /* 0 padding (for engineering) needed  */
    }
   else {                          /* 0.xxx or 0.000xxx form  */
    *c='0'; c++;
    *c='.'; c++;
    for (; pre<0; pre++, c++) *c='0';   /* add any 0's after '.'  */
    for (; ; c++, cut--) {
      if (cut<0) {                 /* need new Unit  */
        if (up==dn->lsu) break;    /* out of input digits  */
        up--;
        cut=DECDPUN-1;
        u=*up;
        }
      TODIGIT(u, cut, c, pow);
      }
    }

  /* Finally add the E-part, if needed.  It will never be 0, has a
     base maximum and minimum of +999999999 through -999999999, but
     could range down to -1999999998 for anormal numbers */
  if (e!=0) {
    Flag had=0;               /* 1=had non-zero  */
    *c='E'; c++;
    *c='+'; c++;              /* assume positive  */
    u=e;                      /* ..  */
    if (e<0) {
      *(c-1)='-';             /* oops, need -  */
      u=-e;                   /* uInt, please  */
      }
    /* lay out the exponent [_itoa or equivalent is not ANSI C]  */
    for (cut=9; cut>=0; cut--) {
      TODIGIT(u, cut, c, pow);
      if (*c=='0' && !had) continue;    /* skip leading zeros  */
      had=1;                            /* had non-0  */
      c++;                              /* step for next  */
      } /* cut  */
    }
  *c='\0';          /* terminate the string (all paths)  */
  return;
  } /* decToString  */

/* ------------------------------------------------------------------ */
/* decAddOp -- add/subtract operation                                 */
/*                                                                    */
/*   This computes C = A + B                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X+X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*   negate is DECNEG if rhs should be negated, or 0 otherwise        */
/*   status accumulates status for the caller                         */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/* Inexact in status must be 0 for correct Exact zero sign in result  */
/* ------------------------------------------------------------------ */
/* If possible, the coefficient is calculated directly into C.        */
/* However, if:                                                       */
/*   -- a digits+1 calculation is needed because the numbers are      */
/*      unaligned and span more than set->digits digits               */
/*   -- a carry to digits+1 digits looks possible                     */
/*   -- C is the same as A or B, and the result would destructively   */
/*      overlap the A or B coefficient                                */
/* then the result must be calculated into a temporary buffer.  In    */
/* this case a local (stack) buffer is used if possible, and only if  */
/* too long for that does malloc become the final resort.             */
/*                                                                    */
/* Misalignment is handled as follows:                                */
/*   Apad: (AExp>BExp) Swap operands and proceed as for BExp>AExp.    */
/*   BPad: Apply the padding by a combination of shifting (whole      */
/*         units) and multiplication (part units).                    */
/*                                                                    */
/* Addition, especially x=x+1, is speed-critical.                     */
/* The static buffer is larger than might be expected to allow for    */
/* calls from higher-level funtions (notable exp).                    */
/* ------------------------------------------------------------------ */
static decNumber * decAddOp(decNumber *res, const decNumber *lhs,
                            const decNumber *rhs, decContext *set,
                            uByte negate, uInt *status) {
  #if DECSUBSET
  decNumber *alloclhs=NULL;        /* non-NULL if rounded lhs allocated  */
  decNumber *allocrhs=NULL;        /* .., rhs  */
  #endif
  Int   rhsshift;                  /* working shift (in Units)  */
  Int   maxdigits;                 /* longest logical length  */
  Int   mult;                      /* multiplier  */
  Int   residue;                   /* rounding accumulator  */
  uByte bits;                      /* result bits  */
  Flag  diffsign;                  /* non-0 if arguments have different sign  */
  Unit  *acc;                      /* accumulator for result  */
  Unit  accbuff[SD2U(DECBUFFER*2+20)]; /* local buffer [*2+20 reduces many  */
                                   /* allocations when called from  */
                                   /* other operations, notable exp]  */
  Unit  *allocacc=NULL;            /* -> allocated acc buffer, iff allocated  */
  Int   reqdigits=set->digits;     /* local copy; requested DIGITS  */
  Int   padding;                   /* work  */

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  do {                             /* protect allocated storage  */
    #if DECSUBSET
    if (!set->extended) {
      /* reduce operands and set lostDigits status, as needed  */
      if (lhs->digits>reqdigits) {
        alloclhs=decRoundOperand(lhs, set, status);
        if (alloclhs==NULL) break;
        lhs=alloclhs;
        }
      if (rhs->digits>reqdigits) {
        allocrhs=decRoundOperand(rhs, set, status);
        if (allocrhs==NULL) break;
        rhs=allocrhs;
        }
      }
    #endif
    /* [following code does not require input rounding]  */

    /* note whether signs differ [used all paths]  */
    diffsign=(Flag)((lhs->bits^rhs->bits^negate)&DECNEG);

    /* handle infinities and NaNs  */
    if (SPECIALARGS) {                  /* a special bit set  */
      if (SPECIALARGS & (DECSNAN | DECNAN))  /* a NaN  */
        decNaNs(res, lhs, rhs, set, status);
       else { /* one or two infinities  */
        if (decNumberIsInfinite(lhs)) { /* LHS is infinity  */
          /* two infinities with different signs is invalid  */
          if (decNumberIsInfinite(rhs) && diffsign) {
            *status|=DEC_Invalid_operation;
            break;
            }
          bits=lhs->bits & DECNEG;      /* get sign from LHS  */
          }
         else bits=(rhs->bits^negate) & DECNEG;/* RHS must be Infinity  */
        bits|=DECINF;
        uprv_decNumberZero(res);
        res->bits=bits;                 /* set +/- infinity  */
        } /* an infinity  */
      break;
      }

    /* Quick exit for add 0s; return the non-0, modified as need be  */
    if (ISZERO(lhs)) {
      Int adjust;                       /* work  */
      Int lexp=lhs->exponent;           /* save in case LHS==RES  */
      bits=lhs->bits;                   /* ..  */
      residue=0;                        /* clear accumulator  */
      decCopyFit(res, rhs, set, &residue, status); /* copy (as needed)  */
      res->bits^=negate;                /* flip if rhs was negated  */
      #if DECSUBSET
      if (set->extended) {              /* exponents on zeros count  */
      #endif
        /* exponent will be the lower of the two  */
        adjust=lexp-res->exponent;      /* adjustment needed [if -ve]  */
        if (ISZERO(res)) {              /* both 0: special IEEE 754 rules  */
          if (adjust<0) res->exponent=lexp;  /* set exponent  */
          /* 0-0 gives +0 unless rounding to -infinity, and -0-0 gives -0  */
          if (diffsign) {
            if (set->round!=DEC_ROUND_FLOOR) res->bits=0;
             else res->bits=DECNEG;     /* preserve 0 sign  */
            }
          }
         else { /* non-0 res  */
          if (adjust<0) {     /* 0-padding needed  */
            if ((res->digits-adjust)>set->digits) {
              adjust=res->digits-set->digits;     /* to fit exactly  */
              *status|=DEC_Rounded;               /* [but exact]  */
              }
            res->digits=decShiftToMost(res->lsu, res->digits, -adjust);
            res->exponent+=adjust;                /* set the exponent.  */
            }
          } /* non-0 res  */
      #if DECSUBSET
        } /* extended  */
      #endif
      decFinish(res, set, &residue, status);      /* clean and finalize  */
      break;}

    if (ISZERO(rhs)) {                  /* [lhs is non-zero]  */
      Int adjust;                       /* work  */
      Int rexp=rhs->exponent;           /* save in case RHS==RES  */
      bits=rhs->bits;                   /* be clean  */
      residue=0;                        /* clear accumulator  */
      decCopyFit(res, lhs, set, &residue, status); /* copy (as needed)  */
      #if DECSUBSET
      if (set->extended) {              /* exponents on zeros count  */
      #endif
        /* exponent will be the lower of the two  */
        /* [0-0 case handled above]  */
        adjust=rexp-res->exponent;      /* adjustment needed [if -ve]  */
        if (adjust<0) {     /* 0-padding needed  */
          if ((res->digits-adjust)>set->digits) {
            adjust=res->digits-set->digits;     /* to fit exactly  */
            *status|=DEC_Rounded;               /* [but exact]  */
            }
          res->digits=decShiftToMost(res->lsu, res->digits, -adjust);
          res->exponent+=adjust;                /* set the exponent.  */
          }
      #if DECSUBSET
        } /* extended  */
      #endif
      decFinish(res, set, &residue, status);      /* clean and finalize  */
      break;}

    /* [NB: both fastpath and mainpath code below assume these cases  */
    /* (notably 0-0) have already been handled]  */

    /* calculate the padding needed to align the operands  */
    padding=rhs->exponent-lhs->exponent;

    /* Fastpath cases where the numbers are aligned and normal, the RHS  */
    /* is all in one unit, no operand rounding is needed, and no carry,  */
    /* lengthening, or borrow is needed  */
    if (padding==0
        && rhs->digits<=DECDPUN
        && rhs->exponent>=set->emin     /* [some normals drop through]  */
        && rhs->exponent<=set->emax-set->digits+1 /* [could clamp]  */
        && rhs->digits<=reqdigits
        && lhs->digits<=reqdigits) {
      Int partial=*lhs->lsu;
      if (!diffsign) {                  /* adding  */
        partial+=*rhs->lsu;
        if ((partial<=DECDPUNMAX)       /* result fits in unit  */
         && (lhs->digits>=DECDPUN ||    /* .. and no digits-count change  */
             partial<(Int)DECNUMBER_powers[lhs->digits])) { /* ..  */
          if (res!=lhs) uprv_decNumberCopy(res, lhs);  /* not in place  */
          *res->lsu=(Unit)partial;      /* [copy could have overwritten RHS]  */
          break;
          }
        /* else drop out for careful add  */
        }
       else {                           /* signs differ  */
        partial-=*rhs->lsu;
        if (partial>0) { /* no borrow needed, and non-0 result  */
          if (res!=lhs) uprv_decNumberCopy(res, lhs);  /* not in place  */
          *res->lsu=(Unit)partial;
          /* this could have reduced digits [but result>0]  */
          res->digits=decGetDigits(res->lsu, D2U(res->digits));
          break;
          }
        /* else drop out for careful subtract  */
        }
      }

    /* Now align (pad) the lhs or rhs so they can be added or  */
    /* subtracted, as necessary.  If one number is much larger than  */
    /* the other (that is, if in plain form there is a least one  */
    /* digit between the lowest digit of one and the highest of the  */
    /* other) padding with up to DIGITS-1 trailing zeros may be  */
    /* needed; then apply rounding (as exotic rounding modes may be  */
    /* affected by the residue).  */
    rhsshift=0;               /* rhs shift to left (padding) in Units  */
    bits=lhs->bits;           /* assume sign is that of LHS  */
    mult=1;                   /* likely multiplier  */

    /* [if padding==0 the operands are aligned; no padding is needed]  */
    if (padding!=0) {
      /* some padding needed; always pad the RHS, as any required  */
      /* padding can then be effected by a simple combination of  */
      /* shifts and a multiply  */
      Flag swapped=0;
      if (padding<0) {                  /* LHS needs the padding  */
        const decNumber *t;
        padding=-padding;               /* will be +ve  */
        bits=(uByte)(rhs->bits^negate); /* assumed sign is now that of RHS  */
        t=lhs; lhs=rhs; rhs=t;
        swapped=1;
        }

      /* If, after pad, rhs would be longer than lhs by digits+1 or  */
      /* more then lhs cannot affect the answer, except as a residue,  */
      /* so only need to pad up to a length of DIGITS+1.  */
      if (rhs->digits+padding > lhs->digits+reqdigits+1) {
        /* The RHS is sufficient  */
        /* for residue use the relative sign indication...  */
        Int shift=reqdigits-rhs->digits;     /* left shift needed  */
        residue=1;                           /* residue for rounding  */
        if (diffsign) residue=-residue;      /* signs differ  */
        /* copy, shortening if necessary  */
        decCopyFit(res, rhs, set, &residue, status);
        /* if it was already shorter, then need to pad with zeros  */
        if (shift>0) {
          res->digits=decShiftToMost(res->lsu, res->digits, shift);
          res->exponent-=shift;              /* adjust the exponent.  */
          }
        /* flip the result sign if unswapped and rhs was negated  */
        if (!swapped) res->bits^=negate;
        decFinish(res, set, &residue, status);    /* done  */
        break;}

      /* LHS digits may affect result  */
      rhsshift=D2U(padding+1)-1;        /* this much by Unit shift ..  */
      mult=DECNUMBER_powers[padding-(rhsshift*DECDPUN)]; /* .. this by multiplication  */
      } /* padding needed  */

    if (diffsign) mult=-mult;           /* signs differ  */

    /* determine the longer operand  */
    maxdigits=rhs->digits+padding;      /* virtual length of RHS  */
    if (lhs->digits>maxdigits) maxdigits=lhs->digits;

    /* Decide on the result buffer to use; if possible place directly  */
    /* into result.  */
    acc=res->lsu;                       /* assume add direct to result  */
    /* If destructive overlap, or the number is too long, or a carry or  */
    /* borrow to DIGITS+1 might be possible, a buffer must be used.  */
    /* [Might be worth more sophisticated tests when maxdigits==reqdigits]  */
    if ((maxdigits>=reqdigits)          /* is, or could be, too large  */
     || (res==rhs && rhsshift>0)) {     /* destructive overlap  */
      /* buffer needed, choose it; units for maxdigits digits will be  */
      /* needed, +1 Unit for carry or borrow  */
      Int need=D2U(maxdigits)+1;
      acc=accbuff;                      /* assume use local buffer  */
      if (need*sizeof(Unit)>sizeof(accbuff)) {
        /* printf("malloc add %ld %ld\n", need, sizeof(accbuff));  */
        allocacc=(Unit *)malloc(need*sizeof(Unit));
        if (allocacc==NULL) {           /* hopeless -- abandon  */
          *status|=DEC_Insufficient_storage;
          break;}
        acc=allocacc;
        }
      }

    res->bits=(uByte)(bits&DECNEG);     /* it's now safe to overwrite..  */
    res->exponent=lhs->exponent;        /* .. operands (even if aliased)  */

    #if DECTRACE
      decDumpAr('A', lhs->lsu, D2U(lhs->digits));
      decDumpAr('B', rhs->lsu, D2U(rhs->digits));
      printf("  :h: %ld %ld\n", rhsshift, mult);
    #endif

    /* add [A+B*m] or subtract [A+B*(-m)]  */
    U_ASSERT(rhs->digits > 0);
    U_ASSERT(lhs->digits > 0);
    res->digits=decUnitAddSub(lhs->lsu, D2U(lhs->digits),
                              rhs->lsu, D2U(rhs->digits),
                              rhsshift, acc, mult)
               *DECDPUN;           /* [units -> digits]  */
    if (res->digits<0) {           /* borrowed...  */
      res->digits=-res->digits;
      res->bits^=DECNEG;           /* flip the sign  */
      }
    #if DECTRACE
      decDumpAr('+', acc, D2U(res->digits));
    #endif

    /* If a buffer was used the result must be copied back, possibly  */
    /* shortening.  (If no buffer was used then the result must have  */
    /* fit, so can't need rounding and residue must be 0.)  */
    residue=0;                     /* clear accumulator  */
    if (acc!=res->lsu) {
      #if DECSUBSET
      if (set->extended) {         /* round from first significant digit  */
      #endif
        /* remove leading zeros that were added due to rounding up to  */
        /* integral Units -- before the test for rounding.  */
        if (res->digits>reqdigits)
          res->digits=decGetDigits(acc, D2U(res->digits));
        decSetCoeff(res, set, acc, res->digits, &residue, status);
      #if DECSUBSET
        }
       else { /* subset arithmetic rounds from original significant digit  */
        /* May have an underestimate.  This only occurs when both  */
        /* numbers fit in DECDPUN digits and are padding with a  */
        /* negative multiple (-10, -100...) and the top digit(s) become  */
        /* 0.  (This only matters when using X3.274 rules where the  */
        /* leading zero could be included in the rounding.)  */
        if (res->digits<maxdigits) {
          *(acc+D2U(res->digits))=0; /* ensure leading 0 is there  */
          res->digits=maxdigits;
          }
         else {
          /* remove leading zeros that added due to rounding up to  */
          /* integral Units (but only those in excess of the original  */
          /* maxdigits length, unless extended) before test for rounding.  */
          if (res->digits>reqdigits) {
            res->digits=decGetDigits(acc, D2U(res->digits));
            if (res->digits<maxdigits) res->digits=maxdigits;
            }
          }
        decSetCoeff(res, set, acc, res->digits, &residue, status);
        /* Now apply rounding if needed before removing leading zeros.  */
        /* This is safe because subnormals are not a possibility  */
        if (residue!=0) {
          decApplyRound(res, set, residue, status);
          residue=0;                 /* did what needed to be done  */
          }
        } /* subset  */
      #endif
      } /* used buffer  */

    /* strip leading zeros [these were left on in case of subset subtract]  */
    res->digits=decGetDigits(res->lsu, D2U(res->digits));

    /* apply checks and rounding  */
    decFinish(res, set, &residue, status);

    /* "When the sum of two operands with opposite signs is exactly  */
    /* zero, the sign of that sum shall be '+' in all rounding modes  */
    /* except round toward -Infinity, in which mode that sign shall be  */
    /* '-'."  [Subset zeros also never have '-', set by decFinish.]  */
    if (ISZERO(res) && diffsign
     #if DECSUBSET
     && set->extended
     #endif
     && (*status&DEC_Inexact)==0) {
      if (set->round==DEC_ROUND_FLOOR) res->bits|=DECNEG;   /* sign -  */
                                  else res->bits&=~DECNEG;  /* sign +  */
      }
    } while(0);                              /* end protected  */

  if (allocacc!=NULL) free(allocacc);        /* drop any storage used  */
  #if DECSUBSET
  if (allocrhs!=NULL) free(allocrhs);        /* ..  */
  if (alloclhs!=NULL) free(alloclhs);        /* ..  */
  #endif
  return res;
  } /* decAddOp  */

/* ------------------------------------------------------------------ */
/* decDivideOp -- division operation                                  */
/*                                                                    */
/*  This routine performs the calculations for all four division      */
/*  operators (divide, divideInteger, remainder, remainderNear).      */
/*                                                                    */
/*  C=A op B                                                          */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X/X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*   op  is DECNUMBER_DIVIDE, DECNUMBER_DIVIDEINT, DECNUMBER_REMAINDER, or DECNUMBER_REMNEAR respectively.    */
/*   status is the usual accumulator                                  */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* ------------------------------------------------------------------ */
/*   The underlying algorithm of this routine is the same as in the   */
/*   1981 S/370 implementation, that is, non-restoring long division  */
/*   with bi-unit (rather than bi-digit) estimation for each unit     */
/*   multiplier.  In this pseudocode overview, complications for the  */
/*   Remainder operators and division residues for exact rounding are */
/*   omitted for clarity.                                             */
/*                                                                    */
/*     Prepare operands and handle special values                     */
/*     Test for x/0 and then 0/x                                      */
/*     Exp =Exp1 - Exp2                                               */
/*     Exp =Exp +len(var1) -len(var2)                                 */
/*     Sign=Sign1 * Sign2                                             */
/*     Pad accumulator (Var1) to double-length with 0's (pad1)        */
/*     Pad Var2 to same length as Var1                                */
/*     msu2pair/plus=1st 2 or 1 units of var2, +1 to allow for round  */
/*     have=0                                                         */
/*     Do until (have=digits+1 OR residue=0)                          */
/*       if exp<0 then if integer divide/residue then leave           */
/*       this_unit=0                                                  */
/*       Do forever                                                   */
/*          compare numbers                                           */
/*          if <0 then leave inner_loop                               */
/*          if =0 then (* quick exit without subtract *) do           */
/*             this_unit=this_unit+1; output this_unit                */
/*             leave outer_loop; end                                  */
/*          Compare lengths of numbers (mantissae):                   */
/*          If same then tops2=msu2pair -- {units 1&2 of var2}        */
/*                  else tops2=msu2plus -- {0, unit 1 of var2}        */
/*          tops1=first_unit_of_Var1*10**DECDPUN +second_unit_of_var1 */
/*          mult=tops1/tops2  -- Good and safe guess at divisor       */
/*          if mult=0 then mult=1                                     */
/*          this_unit=this_unit+mult                                  */
/*          subtract                                                  */
/*          end inner_loop                                            */
/*        if have\=0 | this_unit\=0 then do                           */
/*          output this_unit                                          */
/*          have=have+1; end                                          */
/*        var2=var2/10                                                */
/*        exp=exp-1                                                   */
/*        end outer_loop                                              */
/*     exp=exp+1   -- set the proper exponent                         */
/*     if have=0 then generate answer=0                               */
/*     Return (Result is defined by Var1)                             */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* Two working buffers are needed during the division; one (digits+   */
/* 1) to accumulate the result, and the other (up to 2*digits+1) for  */
/* long subtractions.  These are acc and var1 respectively.           */
/* var1 is a copy of the lhs coefficient, var2 is the rhs coefficient.*/
/* The static buffers may be larger than might be expected to allow   */
/* for calls from higher-level funtions (notable exp).                */
/* ------------------------------------------------------------------ */
static decNumber * decDivideOp(decNumber *res,
                               const decNumber *lhs, const decNumber *rhs,
                               decContext *set, Flag op, uInt *status) {
  #if DECSUBSET
  decNumber *alloclhs=NULL;        /* non-NULL if rounded lhs allocated  */
  decNumber *allocrhs=NULL;        /* .., rhs  */
  #endif
  Unit  accbuff[SD2U(DECBUFFER+DECDPUN+10)]; /* local buffer  */
  Unit  *acc=accbuff;              /* -> accumulator array for result  */
  Unit  *allocacc=NULL;            /* -> allocated buffer, iff allocated  */
  Unit  *accnext;                  /* -> where next digit will go  */
  Int   acclength;                 /* length of acc needed [Units]  */
  Int   accunits;                  /* count of units accumulated  */
  Int   accdigits;                 /* count of digits accumulated  */

  Unit  varbuff[SD2U(DECBUFFER*2+DECDPUN)];  /* buffer for var1  */
  Unit  *var1=varbuff;             /* -> var1 array for long subtraction  */
  Unit  *varalloc=NULL;            /* -> allocated buffer, iff used  */
  Unit  *msu1;                     /* -> msu of var1  */

  const Unit *var2;                /* -> var2 array  */
  const Unit *msu2;                /* -> msu of var2  */
  Int   msu2plus;                  /* msu2 plus one [does not vary]  */
  eInt  msu2pair;                  /* msu2 pair plus one [does not vary]  */

  Int   var1units, var2units;      /* actual lengths  */
  Int   var2ulen;                  /* logical length (units)  */
  Int   var1initpad=0;             /* var1 initial padding (digits)  */
  Int   maxdigits;                 /* longest LHS or required acc length  */
  Int   mult;                      /* multiplier for subtraction  */
  Unit  thisunit;                  /* current unit being accumulated  */
  Int   residue;                   /* for rounding  */
  Int   reqdigits=set->digits;     /* requested DIGITS  */
  Int   exponent;                  /* working exponent  */
  Int   maxexponent=0;             /* DECNUMBER_DIVIDE maximum exponent if unrounded  */
  uByte bits;                      /* working sign  */
  Unit  *target;                   /* work  */
  const Unit *source;              /* ..  */
  uInt  const *pow;                /* ..  */
  Int   shift, cut;                /* ..  */
  #if DECSUBSET
  Int   dropped;                   /* work  */
  #endif

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  do {                             /* protect allocated storage  */
    #if DECSUBSET
    if (!set->extended) {
      /* reduce operands and set lostDigits status, as needed  */
      if (lhs->digits>reqdigits) {
        alloclhs=decRoundOperand(lhs, set, status);
        if (alloclhs==NULL) break;
        lhs=alloclhs;
        }
      if (rhs->digits>reqdigits) {
        allocrhs=decRoundOperand(rhs, set, status);
        if (allocrhs==NULL) break;
        rhs=allocrhs;
        }
      }
    #endif
    /* [following code does not require input rounding]  */

    bits=(lhs->bits^rhs->bits)&DECNEG;  /* assumed sign for divisions  */

    /* handle infinities and NaNs  */
    if (SPECIALARGS) {                  /* a special bit set  */
      if (SPECIALARGS & (DECSNAN | DECNAN)) { /* one or two NaNs  */
        decNaNs(res, lhs, rhs, set, status);
        break;
        }
      /* one or two infinities  */
      if (decNumberIsInfinite(lhs)) {   /* LHS (dividend) is infinite  */
        if (decNumberIsInfinite(rhs) || /* two infinities are invalid ..  */
            op & (DECNUMBER_REMAINDER | DECNUMBER_REMNEAR)) { /* as is remainder of infinity  */
          *status|=DEC_Invalid_operation;
          break;
          }
        /* [Note that infinity/0 raises no exceptions]  */
        uprv_decNumberZero(res);
        res->bits=bits|DECINF;          /* set +/- infinity  */
        break;
        }
       else {                           /* RHS (divisor) is infinite  */
        residue=0;
        if (op&(DECNUMBER_REMAINDER|DECNUMBER_REMNEAR)) {
          /* result is [finished clone of] lhs  */
          decCopyFit(res, lhs, set, &residue, status);
          }
         else {  /* a division  */
          uprv_decNumberZero(res);
          res->bits=bits;               /* set +/- zero  */
          /* for DECNUMBER_DIVIDEINT the exponent is always 0.  For DECNUMBER_DIVIDE, result  */
          /* is a 0 with infinitely negative exponent, clamped to minimum  */
          if (op&DECNUMBER_DIVIDE) {
            res->exponent=set->emin-set->digits+1;
            *status|=DEC_Clamped;
            }
          }
        decFinish(res, set, &residue, status);
        break;
        }
      }

    /* handle 0 rhs (x/0)  */
    if (ISZERO(rhs)) {                  /* x/0 is always exceptional  */
      if (ISZERO(lhs)) {
        uprv_decNumberZero(res);             /* [after lhs test]  */
        *status|=DEC_Division_undefined;/* 0/0 will become NaN  */
        }
       else {
        uprv_decNumberZero(res);
        if (op&(DECNUMBER_REMAINDER|DECNUMBER_REMNEAR)) *status|=DEC_Invalid_operation;
         else {
          *status|=DEC_Division_by_zero; /* x/0  */
          res->bits=bits|DECINF;         /* .. is +/- Infinity  */
          }
        }
      break;}

    /* handle 0 lhs (0/x)  */
    if (ISZERO(lhs)) {                  /* 0/x [x!=0]  */
      #if DECSUBSET
      if (!set->extended) uprv_decNumberZero(res);
       else {
      #endif
        if (op&DECNUMBER_DIVIDE) {
          residue=0;
          exponent=lhs->exponent-rhs->exponent; /* ideal exponent  */
          uprv_decNumberCopy(res, lhs);      /* [zeros always fit]  */
          res->bits=bits;               /* sign as computed  */
          res->exponent=exponent;       /* exponent, too  */
          decFinalize(res, set, &residue, status);   /* check exponent  */
          }
         else if (op&DECNUMBER_DIVIDEINT) {
          uprv_decNumberZero(res);           /* integer 0  */
          res->bits=bits;               /* sign as computed  */
          }
         else {                         /* a remainder  */
          exponent=rhs->exponent;       /* [save in case overwrite]  */
          uprv_decNumberCopy(res, lhs);      /* [zeros always fit]  */
          if (exponent<res->exponent) res->exponent=exponent; /* use lower  */
          }
      #if DECSUBSET
        }
      #endif
      break;}

    /* Precalculate exponent.  This starts off adjusted (and hence fits  */
    /* in 31 bits) and becomes the usual unadjusted exponent as the  */
    /* division proceeds.  The order of evaluation is important, here,  */
    /* to avoid wrap.  */
    exponent=(lhs->exponent+lhs->digits)-(rhs->exponent+rhs->digits);

    /* If the working exponent is -ve, then some quick exits are  */
    /* possible because the quotient is known to be <1  */
    /* [for DECNUMBER_REMNEAR, it needs to be < -1, as -0.5 could need work]  */
    if (exponent<0 && !(op==DECNUMBER_DIVIDE)) {
      if (op&DECNUMBER_DIVIDEINT) {
        uprv_decNumberZero(res);                  /* integer part is 0  */
        #if DECSUBSET
        if (set->extended)
        #endif
          res->bits=bits;                    /* set +/- zero  */
        break;}
      /* fastpath remainders so long as the lhs has the smaller  */
      /* (or equal) exponent  */
      if (lhs->exponent<=rhs->exponent) {
        if (op&DECNUMBER_REMAINDER || exponent<-1) {
          /* It is DECNUMBER_REMAINDER or safe DECNUMBER_REMNEAR; result is [finished  */
          /* clone of] lhs  (r = x - 0*y)  */
          residue=0;
          decCopyFit(res, lhs, set, &residue, status);
          decFinish(res, set, &residue, status);
          break;
          }
        /* [unsafe DECNUMBER_REMNEAR drops through]  */
        }
      } /* fastpaths  */

    /* Long (slow) division is needed; roll up the sleeves... */

    /* The accumulator will hold the quotient of the division.  */
    /* If it needs to be too long for stack storage, then allocate.  */
    acclength=D2U(reqdigits+DECDPUN);   /* in Units  */
    if (acclength*sizeof(Unit)>sizeof(accbuff)) {
      /* printf("malloc dvacc %ld units\n", acclength);  */
      allocacc=(Unit *)malloc(acclength*sizeof(Unit));
      if (allocacc==NULL) {             /* hopeless -- abandon  */
        *status|=DEC_Insufficient_storage;
        break;}
      acc=allocacc;                     /* use the allocated space  */
      }

    /* var1 is the padded LHS ready for subtractions.  */
    /* If it needs to be too long for stack storage, then allocate.  */
    /* The maximum units needed for var1 (long subtraction) is:  */
    /* Enough for  */
    /*     (rhs->digits+reqdigits-1) -- to allow full slide to right  */
    /* or  (lhs->digits)             -- to allow for long lhs  */
    /* whichever is larger  */
    /*   +1                -- for rounding of slide to right  */
    /*   +1                -- for leading 0s  */
    /*   +1                -- for pre-adjust if a remainder or DECNUMBER_DIVIDEINT  */
    /* [Note: unused units do not participate in decUnitAddSub data]  */
    maxdigits=rhs->digits+reqdigits-1;
    if (lhs->digits>maxdigits) maxdigits=lhs->digits;
    var1units=D2U(maxdigits)+2;
    /* allocate a guard unit above msu1 for DECNUMBER_REMAINDERNEAR  */
    if (!(op&DECNUMBER_DIVIDE)) var1units++;
    if ((var1units+1)*sizeof(Unit)>sizeof(varbuff)) {
      /* printf("malloc dvvar %ld units\n", var1units+1);  */
      varalloc=(Unit *)malloc((var1units+1)*sizeof(Unit));
      if (varalloc==NULL) {             /* hopeless -- abandon  */
        *status|=DEC_Insufficient_storage;
        break;}
      var1=varalloc;                    /* use the allocated space  */
      }

    /* Extend the lhs and rhs to full long subtraction length.  The lhs  */
    /* is truly extended into the var1 buffer, with 0 padding, so a  */
    /* subtract in place is always possible.  The rhs (var2) has  */
    /* virtual padding (implemented by decUnitAddSub).  */
    /* One guard unit was allocated above msu1 for rem=rem+rem in  */
    /* DECNUMBER_REMAINDERNEAR.  */
    msu1=var1+var1units-1;              /* msu of var1  */
    source=lhs->lsu+D2U(lhs->digits)-1; /* msu of input array  */
    for (target=msu1; source>=lhs->lsu; source--, target--) *target=*source;
    for (; target>=var1; target--) *target=0;

    /* rhs (var2) is left-aligned with var1 at the start  */
    var2ulen=var1units;                 /* rhs logical length (units)  */
    var2units=D2U(rhs->digits);         /* rhs actual length (units)  */
    var2=rhs->lsu;                      /* -> rhs array  */
    msu2=var2+var2units-1;              /* -> msu of var2 [never changes]  */
    /* now set up the variables which will be used for estimating the  */
    /* multiplication factor.  If these variables are not exact, add  */
    /* 1 to make sure that the multiplier is never overestimated.  */
    msu2plus=*msu2;                     /* it's value ..  */
    if (var2units>1) msu2plus++;        /* .. +1 if any more  */
    msu2pair=(eInt)*msu2*(DECDPUNMAX+1);/* top two pair ..  */
    if (var2units>1) {                  /* .. [else treat 2nd as 0]  */
      msu2pair+=*(msu2-1);              /* ..  */
      if (var2units>2) msu2pair++;      /* .. +1 if any more  */
      }

    /* The calculation is working in units, which may have leading zeros,  */
    /* but the exponent was calculated on the assumption that they are  */
    /* both left-aligned.  Adjust the exponent to compensate: add the  */
    /* number of leading zeros in var1 msu and subtract those in var2 msu.  */
    /* [This is actually done by counting the digits and negating, as  */
    /* lead1=DECDPUN-digits1, and similarly for lead2.]  */
    for (pow=&DECNUMBER_powers[1]; *msu1>=*pow; pow++) exponent--;
    for (pow=&DECNUMBER_powers[1]; *msu2>=*pow; pow++) exponent++;

    /* Now, if doing an integer divide or remainder, ensure that  */
    /* the result will be Unit-aligned.  To do this, shift the var1  */
    /* accumulator towards least if need be.  (It's much easier to  */
    /* do this now than to reassemble the residue afterwards, if  */
    /* doing a remainder.)  Also ensure the exponent is not negative.  */
    if (!(op&DECNUMBER_DIVIDE)) {
      Unit *u;                          /* work  */
      /* save the initial 'false' padding of var1, in digits  */
      var1initpad=(var1units-D2U(lhs->digits))*DECDPUN;
      /* Determine the shift to do.  */
      if (exponent<0) cut=-exponent;
       else cut=DECDPUN-exponent%DECDPUN;
      decShiftToLeast(var1, var1units, cut);
      exponent+=cut;                    /* maintain numerical value  */
      var1initpad-=cut;                 /* .. and reduce padding  */
      /* clean any most-significant units which were just emptied  */
      for (u=msu1; cut>=DECDPUN; cut-=DECDPUN, u--) *u=0;
      } /* align  */
     else { /* is DECNUMBER_DIVIDE  */
      maxexponent=lhs->exponent-rhs->exponent;    /* save  */
      /* optimization: if the first iteration will just produce 0,  */
      /* preadjust to skip it [valid for DECNUMBER_DIVIDE only]  */
      if (*msu1<*msu2) {
        var2ulen--;                     /* shift down  */
        exponent-=DECDPUN;              /* update the exponent  */
        }
      }

    /* ---- start the long-division loops ------------------------------  */
    accunits=0;                         /* no units accumulated yet  */
    accdigits=0;                        /* .. or digits  */
    accnext=acc+acclength-1;            /* -> msu of acc [NB: allows digits+1]  */
    for (;;) {                          /* outer forever loop  */
      thisunit=0;                       /* current unit assumed 0  */
      /* find the next unit  */
      for (;;) {                        /* inner forever loop  */
        /* strip leading zero units [from either pre-adjust or from  */
        /* subtract last time around].  Leave at least one unit.  */
        for (; *msu1==0 && msu1>var1; msu1--) var1units--;

        if (var1units<var2ulen) break;       /* var1 too low for subtract  */
        if (var1units==var2ulen) {           /* unit-by-unit compare needed  */
          /* compare the two numbers, from msu  */
          const Unit *pv1, *pv2;
          Unit v2;                           /* units to compare  */
          pv2=msu2;                          /* -> msu  */
          for (pv1=msu1; ; pv1--, pv2--) {
            /* v1=*pv1 -- always OK  */
            v2=0;                            /* assume in padding  */
            if (pv2>=var2) v2=*pv2;          /* in range  */
            if (*pv1!=v2) break;             /* no longer the same  */
            if (pv1==var1) break;            /* done; leave pv1 as is  */
            }
          /* here when all inspected or a difference seen  */
          if (*pv1<v2) break;                /* var1 too low to subtract  */
          if (*pv1==v2) {                    /* var1 == var2  */
            /* reach here if var1 and var2 are identical; subtraction  */
            /* would increase digit by one, and the residue will be 0 so  */
            /* the calculation is done; leave the loop with residue=0.  */
            thisunit++;                      /* as though subtracted  */
            *var1=0;                         /* set var1 to 0  */
            var1units=1;                     /* ..  */
            break;  /* from inner  */
            } /* var1 == var2  */
          /* *pv1>v2.  Prepare for real subtraction; the lengths are equal  */
          /* Estimate the multiplier (there's always a msu1-1)...  */
          /* Bring in two units of var2 to provide a good estimate.  */
          mult=(Int)(((eInt)*msu1*(DECDPUNMAX+1)+*(msu1-1))/msu2pair);
          } /* lengths the same  */
         else { /* var1units > var2ulen, so subtraction is safe  */
          /* The var2 msu is one unit towards the lsu of the var1 msu,  */
          /* so only one unit for var2 can be used.  */
          mult=(Int)(((eInt)*msu1*(DECDPUNMAX+1)+*(msu1-1))/msu2plus);
          }
        if (mult==0) mult=1;                 /* must always be at least 1  */
        /* subtraction needed; var1 is > var2  */
        thisunit=(Unit)(thisunit+mult);      /* accumulate  */
        /* subtract var1-var2, into var1; only the overlap needs  */
        /* processing, as this is an in-place calculation  */
        shift=var2ulen-var2units;
        #if DECTRACE
          decDumpAr('1', &var1[shift], var1units-shift);
          decDumpAr('2', var2, var2units);
          printf("m=%ld\n", -mult);
        #endif
        decUnitAddSub(&var1[shift], var1units-shift,
                      var2, var2units, 0,
                      &var1[shift], -mult);
        #if DECTRACE
          decDumpAr('#', &var1[shift], var1units-shift);
        #endif
        /* var1 now probably has leading zeros; these are removed at the  */
        /* top of the inner loop.  */
        } /* inner loop  */

      /* The next unit has been calculated in full; unless it's a  */
      /* leading zero, add to acc  */
      if (accunits!=0 || thisunit!=0) {      /* is first or non-zero  */
        *accnext=thisunit;                   /* store in accumulator  */
        /* account exactly for the new digits  */
        if (accunits==0) {
          accdigits++;                       /* at least one  */
          for (pow=&DECNUMBER_powers[1]; thisunit>=*pow; pow++) accdigits++;
          }
         else accdigits+=DECDPUN;
        accunits++;                          /* update count  */
        accnext--;                           /* ready for next  */
        if (accdigits>reqdigits) break;      /* have enough digits  */
        }

      /* if the residue is zero, the operation is done (unless divide  */
      /* or divideInteger and still not enough digits yet)  */
      if (*var1==0 && var1units==1) {        /* residue is 0  */
        if (op&(DECNUMBER_REMAINDER|DECNUMBER_REMNEAR)) break;
        if ((op&DECNUMBER_DIVIDE) && (exponent<=maxexponent)) break;
        /* [drop through if divideInteger]  */
        }
      /* also done enough if calculating remainder or integer  */
      /* divide and just did the last ('units') unit  */
      if (exponent==0 && !(op&DECNUMBER_DIVIDE)) break;

      /* to get here, var1 is less than var2, so divide var2 by the per-  */
      /* Unit power of ten and go for the next digit  */
      var2ulen--;                            /* shift down  */
      exponent-=DECDPUN;                     /* update the exponent  */
      } /* outer loop  */

    /* ---- division is complete ---------------------------------------  */
    /* here: acc      has at least reqdigits+1 of good results (or fewer  */
    /*                if early stop), starting at accnext+1 (its lsu)  */
    /*       var1     has any residue at the stopping point  */
    /*       accunits is the number of digits collected in acc  */
    if (accunits==0) {             /* acc is 0  */
      accunits=1;                  /* show have a unit ..  */
      accdigits=1;                 /* ..  */
      *accnext=0;                  /* .. whose value is 0  */
      }
     else accnext++;               /* back to last placed  */
    /* accnext now -> lowest unit of result  */

    residue=0;                     /* assume no residue  */
    if (op&DECNUMBER_DIVIDE) {
      /* record the presence of any residue, for rounding  */
      if (*var1!=0 || var1units>1) residue=1;
       else { /* no residue  */
        /* Had an exact division; clean up spurious trailing 0s.  */
        /* There will be at most DECDPUN-1, from the final multiply,  */
        /* and then only if the result is non-0 (and even) and the  */
        /* exponent is 'loose'.  */
        #if DECDPUN>1
        Unit lsu=*accnext;
        if (!(lsu&0x01) && (lsu!=0)) {
          /* count the trailing zeros  */
          Int drop=0;
          for (;; drop++) {    /* [will terminate because lsu!=0]  */
            if (exponent>=maxexponent) break;     /* don't chop real 0s  */
            #if DECDPUN<=4
              if ((lsu-QUOT10(lsu, drop+1)
                  *DECNUMBER_powers[drop+1])!=0) break;     /* found non-0 digit  */
            #else
              if (lsu%DECNUMBER_powers[drop+1]!=0) break;   /* found non-0 digit  */
            #endif
            exponent++;
            }
          if (drop>0) {
            accunits=decShiftToLeast(accnext, accunits, drop);
            accdigits=decGetDigits(accnext, accunits);
            accunits=D2U(accdigits);
            /* [exponent was adjusted in the loop]  */
            }
          } /* neither odd nor 0  */
        #endif
        } /* exact divide  */
      } /* divide  */
     else /* op!=DECNUMBER_DIVIDE */ {
      /* check for coefficient overflow  */
      if (accdigits+exponent>reqdigits) {
        *status|=DEC_Division_impossible;
        break;
        }
      if (op & (DECNUMBER_REMAINDER|DECNUMBER_REMNEAR)) {
        /* [Here, the exponent will be 0, because var1 was adjusted  */
        /* appropriately.]  */
        Int postshift;                       /* work  */
        Flag wasodd=0;                       /* integer was odd  */
        Unit *quotlsu;                       /* for save  */
        Int  quotdigits;                     /* ..  */

        bits=lhs->bits;                      /* remainder sign is always as lhs  */

        /* Fastpath when residue is truly 0 is worthwhile [and  */
        /* simplifies the code below]  */
        if (*var1==0 && var1units==1) {      /* residue is 0  */
          Int exp=lhs->exponent;             /* save min(exponents)  */
          if (rhs->exponent<exp) exp=rhs->exponent;
          uprv_decNumberZero(res);                /* 0 coefficient  */
          #if DECSUBSET
          if (set->extended)
          #endif
          res->exponent=exp;                 /* .. with proper exponent  */
          res->bits=(uByte)(bits&DECNEG);          /* [cleaned]  */
          decFinish(res, set, &residue, status);   /* might clamp  */
          break;
          }
        /* note if the quotient was odd  */
        if (*accnext & 0x01) wasodd=1;       /* acc is odd  */
        quotlsu=accnext;                     /* save in case need to reinspect  */
        quotdigits=accdigits;                /* ..  */

        /* treat the residue, in var1, as the value to return, via acc  */
        /* calculate the unused zero digits.  This is the smaller of:  */
        /*   var1 initial padding (saved above)  */
        /*   var2 residual padding, which happens to be given by:  */
        postshift=var1initpad+exponent-lhs->exponent+rhs->exponent;
        /* [the 'exponent' term accounts for the shifts during divide]  */
        if (var1initpad<postshift) postshift=var1initpad;

        /* shift var1 the requested amount, and adjust its digits  */
        var1units=decShiftToLeast(var1, var1units, postshift);
        accnext=var1;
        accdigits=decGetDigits(var1, var1units);
        accunits=D2U(accdigits);

        exponent=lhs->exponent;         /* exponent is smaller of lhs & rhs  */
        if (rhs->exponent<exponent) exponent=rhs->exponent;

        /* Now correct the result if doing remainderNear; if it  */
        /* (looking just at coefficients) is > rhs/2, or == rhs/2 and  */
        /* the integer was odd then the result should be rem-rhs.  */
        if (op&DECNUMBER_REMNEAR) {
          Int compare, tarunits;        /* work  */
          Unit *up;                     /* ..  */
          /* calculate remainder*2 into the var1 buffer (which has  */
          /* 'headroom' of an extra unit and hence enough space)  */
          /* [a dedicated 'double' loop would be faster, here]  */
          tarunits=decUnitAddSub(accnext, accunits, accnext, accunits,
                                 0, accnext, 1);
          /* decDumpAr('r', accnext, tarunits);  */

          /* Here, accnext (var1) holds tarunits Units with twice the  */
          /* remainder's coefficient, which must now be compared to the  */
          /* RHS.  The remainder's exponent may be smaller than the RHS's.  */
          compare=decUnitCompare(accnext, tarunits, rhs->lsu, D2U(rhs->digits),
                                 rhs->exponent-exponent);
          if (compare==DECNUMBER_BADINT) {             /* deep trouble  */
            *status|=DEC_Insufficient_storage;
            break;}

          /* now restore the remainder by dividing by two; the lsu  */
          /* is known to be even.  */
          for (up=accnext; up<accnext+tarunits; up++) {
            Int half;              /* half to add to lower unit  */
            half=*up & 0x01;
            *up/=2;                /* [shift]  */
            if (!half) continue;
            *(up-1)+=(DECDPUNMAX+1)/2;
            }
          /* [accunits still describes the original remainder length]  */

          if (compare>0 || (compare==0 && wasodd)) { /* adjustment needed  */
            Int exp, expunits, exprem;       /* work  */
            /* This is effectively causing round-up of the quotient,  */
            /* so if it was the rare case where it was full and all  */
            /* nines, it would overflow and hence division-impossible  */
            /* should be raised  */
            Flag allnines=0;                 /* 1 if quotient all nines  */
            if (quotdigits==reqdigits) {     /* could be borderline  */
              for (up=quotlsu; ; up++) {
                if (quotdigits>DECDPUN) {
                  if (*up!=DECDPUNMAX) break;/* non-nines  */
                  }
                 else {                      /* this is the last Unit  */
                  if (*up==DECNUMBER_powers[quotdigits]-1) allnines=1;
                  break;
                  }
                quotdigits-=DECDPUN;         /* checked those digits  */
                } /* up  */
              } /* borderline check  */
            if (allnines) {
              *status|=DEC_Division_impossible;
              break;}

            /* rem-rhs is needed; the sign will invert.  Again, var1  */
            /* can safely be used for the working Units array.  */
            exp=rhs->exponent-exponent;      /* RHS padding needed  */
            /* Calculate units and remainder from exponent.  */
            expunits=exp/DECDPUN;
            exprem=exp%DECDPUN;
            /* subtract [A+B*(-m)]; the result will always be negative  */
            accunits=-decUnitAddSub(accnext, accunits,
                                    rhs->lsu, D2U(rhs->digits),
                                    expunits, accnext, -(Int)DECNUMBER_powers[exprem]);
            accdigits=decGetDigits(accnext, accunits); /* count digits exactly  */
            accunits=D2U(accdigits);    /* and recalculate the units for copy  */
            /* [exponent is as for original remainder]  */
            bits^=DECNEG;               /* flip the sign  */
            }
          } /* DECNUMBER_REMNEAR  */
        } /* DECNUMBER_REMAINDER or DECNUMBER_REMNEAR  */
      } /* not DECNUMBER_DIVIDE  */

    /* Set exponent and bits  */
    res->exponent=exponent;
    res->bits=(uByte)(bits&DECNEG);          /* [cleaned]  */

    /* Now the coefficient.  */
    decSetCoeff(res, set, accnext, accdigits, &residue, status);

    decFinish(res, set, &residue, status);   /* final cleanup  */

    #if DECSUBSET
    /* If a divide then strip trailing zeros if subset [after round]  */
    if (!set->extended && (op==DECNUMBER_DIVIDE)) decTrim(res, set, 0, 1, &dropped);
    #endif
    } while(0);                              /* end protected  */

  if (varalloc!=NULL) free(varalloc);   /* drop any storage used  */
  if (allocacc!=NULL) free(allocacc);   /* ..  */
  #if DECSUBSET
  if (allocrhs!=NULL) free(allocrhs);   /* ..  */
  if (alloclhs!=NULL) free(alloclhs);   /* ..  */
  #endif
  return res;
  } /* decDivideOp  */

/* ------------------------------------------------------------------ */
/* decMultiplyOp -- multiplication operation                          */
/*                                                                    */
/*  This routine performs the multiplication C=A x B.                 */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X*X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*   status is the usual accumulator                                  */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* ------------------------------------------------------------------ */
/* 'Classic' multiplication is used rather than Karatsuba, as the     */
/* latter would give only a minor improvement for the short numbers   */
/* expected to be handled most (and uses much more memory).           */
/*                                                                    */
/* There are two major paths here: the general-purpose ('old code')   */
/* path which handles all DECDPUN values, and a fastpath version      */
/* which is used if 64-bit ints are available, DECDPUN<=4, and more   */
/* than two calls to decUnitAddSub would be made.                     */
/*                                                                    */
/* The fastpath version lumps units together into 8-digit or 9-digit  */
/* chunks, and also uses a lazy carry strategy to minimise expensive  */
/* 64-bit divisions.  The chunks are then broken apart again into     */
/* units for continuing processing.  Despite this overhead, the       */
/* fastpath can speed up some 16-digit operations by 10x (and much    */
/* more for higher-precision calculations).                           */
/*                                                                    */
/* A buffer always has to be used for the accumulator; in the         */
/* fastpath, buffers are also always needed for the chunked copies of */
/* of the operand coefficients.                                       */
/* Static buffers are larger than needed just for multiply, to allow  */
/* for calls from other operations (notably exp).                     */
/* ------------------------------------------------------------------ */
#define FASTMUL (DECUSE64 && DECDPUN<5)
static decNumber * decMultiplyOp(decNumber *res, const decNumber *lhs,
                                 const decNumber *rhs, decContext *set,
                                 uInt *status) {
  Int    accunits;                 /* Units of accumulator in use  */
  Int    exponent;                 /* work  */
  Int    residue=0;                /* rounding residue  */
  uByte  bits;                     /* result sign  */
  Unit  *acc;                      /* -> accumulator Unit array  */
  Int    needbytes;                /* size calculator  */
  void  *allocacc=NULL;            /* -> allocated accumulator, iff allocated  */
  Unit  accbuff[SD2U(DECBUFFER*4+1)]; /* buffer (+1 for DECBUFFER==0,  */
                                   /* *4 for calls from other operations)  */
  const Unit *mer, *mermsup;       /* work  */
  Int   madlength;                 /* Units in multiplicand  */
  Int   shift;                     /* Units to shift multiplicand by  */

  #if FASTMUL
    /* if DECDPUN is 1 or 3 work in base 10**9, otherwise  */
    /* (DECDPUN is 2 or 4) then work in base 10**8  */
    #if DECDPUN & 1                /* odd  */
      #define FASTBASE 1000000000  /* base  */
      #define FASTDIGS          9  /* digits in base  */
      #define FASTLAZY         18  /* carry resolution point [1->18]  */
    #else
      #define FASTBASE  100000000
      #define FASTDIGS          8
      #define FASTLAZY       1844  /* carry resolution point [1->1844]  */
    #endif
    /* three buffers are used, two for chunked copies of the operands  */
    /* (base 10**8 or base 10**9) and one base 2**64 accumulator with  */
    /* lazy carry evaluation  */
    uInt   zlhibuff[(DECBUFFER*2+1)/8+1]; /* buffer (+1 for DECBUFFER==0)  */
    uInt  *zlhi=zlhibuff;                 /* -> lhs array  */
    uInt  *alloclhi=NULL;                 /* -> allocated buffer, iff allocated  */
    uInt   zrhibuff[(DECBUFFER*2+1)/8+1]; /* buffer (+1 for DECBUFFER==0)  */
    uInt  *zrhi=zrhibuff;                 /* -> rhs array  */
    uInt  *allocrhi=NULL;                 /* -> allocated buffer, iff allocated  */
    uLong  zaccbuff[(DECBUFFER*2+1)/4+2]; /* buffer (+1 for DECBUFFER==0)  */
    /* [allocacc is shared for both paths, as only one will run]  */
    uLong *zacc=zaccbuff;          /* -> accumulator array for exact result  */
    #if DECDPUN==1
    Int    zoff;                   /* accumulator offset  */
    #endif
    uInt  *lip, *rip;              /* item pointers  */
    uInt  *lmsi, *rmsi;            /* most significant items  */
    Int    ilhs, irhs, iacc;       /* item counts in the arrays  */
    Int    lazy;                   /* lazy carry counter  */
    uLong  lcarry;                 /* uLong carry  */
    uInt   carry;                  /* carry (NB not uLong)  */
    Int    count;                  /* work  */
    const  Unit *cup;              /* ..  */
    Unit  *up;                     /* ..  */
    uLong *lp;                     /* ..  */
    Int    p;                      /* ..  */
  #endif

  #if DECSUBSET
    decNumber *alloclhs=NULL;      /* -> allocated buffer, iff allocated  */
    decNumber *allocrhs=NULL;      /* -> allocated buffer, iff allocated  */
  #endif

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  /* precalculate result sign  */
  bits=(uByte)((lhs->bits^rhs->bits)&DECNEG);

  /* handle infinities and NaNs  */
  if (SPECIALARGS) {               /* a special bit set  */
    if (SPECIALARGS & (DECSNAN | DECNAN)) { /* one or two NaNs  */
      decNaNs(res, lhs, rhs, set, status);
      return res;}
    /* one or two infinities; Infinity * 0 is invalid  */
    if (((lhs->bits & DECINF)==0 && ISZERO(lhs))
      ||((rhs->bits & DECINF)==0 && ISZERO(rhs))) {
      *status|=DEC_Invalid_operation;
      return res;}
    uprv_decNumberZero(res);
    res->bits=bits|DECINF;         /* infinity  */
    return res;}

  /* For best speed, as in DMSRCN [the original Rexx numerics  */
  /* module], use the shorter number as the multiplier (rhs) and  */
  /* the longer as the multiplicand (lhs) to minimise the number of  */
  /* adds (partial products)  */
  if (lhs->digits<rhs->digits) {   /* swap...  */
    const decNumber *hold=lhs;
    lhs=rhs;
    rhs=hold;
    }

  do {                             /* protect allocated storage  */
    #if DECSUBSET
    if (!set->extended) {
      /* reduce operands and set lostDigits status, as needed  */
      if (lhs->digits>set->digits) {
        alloclhs=decRoundOperand(lhs, set, status);
        if (alloclhs==NULL) break;
        lhs=alloclhs;
        }
      if (rhs->digits>set->digits) {
        allocrhs=decRoundOperand(rhs, set, status);
        if (allocrhs==NULL) break;
        rhs=allocrhs;
        }
      }
    #endif
    /* [following code does not require input rounding]  */

    #if FASTMUL                    /* fastpath can be used  */
    /* use the fast path if there are enough digits in the shorter  */
    /* operand to make the setup and takedown worthwhile  */
    #define NEEDTWO (DECDPUN*2)    /* within two decUnitAddSub calls  */
    if (rhs->digits>NEEDTWO) {     /* use fastpath...  */
      /* calculate the number of elements in each array  */
      ilhs=(lhs->digits+FASTDIGS-1)/FASTDIGS; /* [ceiling]  */
      irhs=(rhs->digits+FASTDIGS-1)/FASTDIGS; /* ..  */
      iacc=ilhs+irhs;

      /* allocate buffers if required, as usual  */
      needbytes=ilhs*sizeof(uInt);
      if (needbytes>(Int)sizeof(zlhibuff)) {
        alloclhi=(uInt *)malloc(needbytes);
        zlhi=alloclhi;}
      needbytes=irhs*sizeof(uInt);
      if (needbytes>(Int)sizeof(zrhibuff)) {
        allocrhi=(uInt *)malloc(needbytes);
        zrhi=allocrhi;}

      /* Allocating the accumulator space needs a special case when  */
      /* DECDPUN=1 because when converting the accumulator to Units  */
      /* after the multiplication each 8-byte item becomes 9 1-byte  */
      /* units.  Therefore iacc extra bytes are needed at the front  */
      /* (rounded up to a multiple of 8 bytes), and the uLong  */
      /* accumulator starts offset the appropriate number of units  */
      /* to the right to avoid overwrite during the unchunking.  */

      /* Make sure no signed int overflow below. This is always true */
      /* if the given numbers have less digits than DEC_MAX_DIGITS. */
      U_ASSERT((uint32_t)iacc <= INT32_MAX/sizeof(uLong));
      needbytes=iacc*sizeof(uLong);
      #if DECDPUN==1
      zoff=(iacc+7)/8;        /* items to offset by  */
      needbytes+=zoff*8;
      #endif
      if (needbytes>(Int)sizeof(zaccbuff)) {
        allocacc=(uLong *)malloc(needbytes);
        zacc=(uLong *)allocacc;}
      if (zlhi==NULL||zrhi==NULL||zacc==NULL) {
        *status|=DEC_Insufficient_storage;
        break;}

      acc=(Unit *)zacc;       /* -> target Unit array  */
      #if DECDPUN==1
      zacc+=zoff;             /* start uLong accumulator to right  */
      #endif

      /* assemble the chunked copies of the left and right sides  */
      for (count=lhs->digits, cup=lhs->lsu, lip=zlhi; count>0; lip++)
        for (p=0, *lip=0; p<FASTDIGS && count>0;
             p+=DECDPUN, cup++, count-=DECDPUN)
          *lip+=*cup*DECNUMBER_powers[p];
      lmsi=lip-1;     /* save -> msi  */
      for (count=rhs->digits, cup=rhs->lsu, rip=zrhi; count>0; rip++)
        for (p=0, *rip=0; p<FASTDIGS && count>0;
             p+=DECDPUN, cup++, count-=DECDPUN)
          *rip+=*cup*DECNUMBER_powers[p];
      rmsi=rip-1;     /* save -> msi  */

      /* zero the accumulator  */
      for (lp=zacc; lp<zacc+iacc; lp++) *lp=0;

      /* Start the multiplication */
      /* Resolving carries can dominate the cost of accumulating the  */
      /* partial products, so this is only done when necessary.  */
      /* Each uLong item in the accumulator can hold values up to  */
      /* 2**64-1, and each partial product can be as large as  */
      /* (10**FASTDIGS-1)**2.  When FASTDIGS=9, this can be added to  */
      /* itself 18.4 times in a uLong without overflowing, so during  */
      /* the main calculation resolution is carried out every 18th  */
      /* add -- every 162 digits.  Similarly, when FASTDIGS=8, the  */
      /* partial products can be added to themselves 1844.6 times in  */
      /* a uLong without overflowing, so intermediate carry  */
      /* resolution occurs only every 14752 digits.  Hence for common  */
      /* short numbers usually only the one final carry resolution  */
      /* occurs.  */
      /* (The count is set via FASTLAZY to simplify experiments to  */
      /* measure the value of this approach: a 35% improvement on a  */
      /* [34x34] multiply.)  */
      lazy=FASTLAZY;                         /* carry delay count  */
      for (rip=zrhi; rip<=rmsi; rip++) {     /* over each item in rhs  */
        lp=zacc+(rip-zrhi);                  /* where to add the lhs  */
        for (lip=zlhi; lip<=lmsi; lip++, lp++) { /* over each item in lhs  */
          *lp+=(uLong)(*lip)*(*rip);         /* [this should in-line]  */
          } /* lip loop  */
        lazy--;
        if (lazy>0 && rip!=rmsi) continue;
        lazy=FASTLAZY;                       /* reset delay count  */
        /* spin up the accumulator resolving overflows  */
        for (lp=zacc; lp<zacc+iacc; lp++) {
          if (*lp<FASTBASE) continue;        /* it fits  */
          lcarry=*lp/FASTBASE;               /* top part [slow divide]  */
          /* lcarry can exceed 2**32-1, so check again; this check  */
          /* and occasional extra divide (slow) is well worth it, as  */
          /* it allows FASTLAZY to be increased to 18 rather than 4  */
          /* in the FASTDIGS=9 case  */
          if (lcarry<FASTBASE) carry=(uInt)lcarry;  /* [usual]  */
           else { /* two-place carry [fairly rare]  */
            uInt carry2=(uInt)(lcarry/FASTBASE);    /* top top part  */
            *(lp+2)+=carry2;                        /* add to item+2  */
            *lp-=((uLong)FASTBASE*FASTBASE*carry2); /* [slow]  */
            carry=(uInt)(lcarry-((uLong)FASTBASE*carry2)); /* [inline]  */
            }
          *(lp+1)+=carry;                    /* add to item above [inline]  */
          *lp-=((uLong)FASTBASE*carry);      /* [inline]  */
          } /* carry resolution  */
        } /* rip loop  */

      /* The multiplication is complete; time to convert back into  */
      /* units.  This can be done in-place in the accumulator and in  */
      /* 32-bit operations, because carries were resolved after the  */
      /* final add.  This needs N-1 divides and multiplies for  */
      /* each item in the accumulator (which will become up to N  */
      /* units, where 2<=N<=9).  */
      for (lp=zacc, up=acc; lp<zacc+iacc; lp++) {
        uInt item=(uInt)*lp;                 /* decapitate to uInt  */
        for (p=0; p<FASTDIGS-DECDPUN; p+=DECDPUN, up++) {
          uInt part=item/(DECDPUNMAX+1);
          *up=(Unit)(item-(part*(DECDPUNMAX+1)));
          item=part;
          } /* p  */
        *up=(Unit)item; up++;                /* [final needs no division]  */
        } /* lp  */
      accunits = static_cast<int32_t>(up-acc);                       /* count of units  */
      }
     else { /* here to use units directly, without chunking ['old code']  */
    #endif

      /* if accumulator will be too long for local storage, then allocate  */
      acc=accbuff;                 /* -> assume buffer for accumulator  */
      needbytes=(D2U(lhs->digits)+D2U(rhs->digits))*sizeof(Unit);
      if (needbytes>(Int)sizeof(accbuff)) {
        allocacc=(Unit *)malloc(needbytes);
        if (allocacc==NULL) {*status|=DEC_Insufficient_storage; break;}
        acc=(Unit *)allocacc;                /* use the allocated space  */
        }

      /* Now the main long multiplication loop */
      /* Unlike the equivalent in the IBM Java implementation, there  */
      /* is no advantage in calculating from msu to lsu.  So, do it  */
      /* by the book, as it were.  */
      /* Each iteration calculates ACC=ACC+MULTAND*MULT  */
      accunits=1;                  /* accumulator starts at '0'  */
      *acc=0;                      /* .. (lsu=0)  */
      shift=0;                     /* no multiplicand shift at first  */
      madlength=D2U(lhs->digits);  /* this won't change  */
      mermsup=rhs->lsu+D2U(rhs->digits); /* -> msu+1 of multiplier  */

      for (mer=rhs->lsu; mer<mermsup; mer++) {
        /* Here, *mer is the next Unit in the multiplier to use  */
        /* If non-zero [optimization] add it...  */
        if (*mer!=0) accunits=decUnitAddSub(&acc[shift], accunits-shift,
                                            lhs->lsu, madlength, 0,
                                            &acc[shift], *mer)
                                            + shift;
         else { /* extend acc with a 0; it will be used shortly  */
          *(acc+accunits)=0;       /* [this avoids length of <=0 later]  */
          accunits++;
          }
        /* multiply multiplicand by 10**DECDPUN for next Unit to left  */
        shift++;                   /* add this for 'logical length'  */
        } /* n  */
    #if FASTMUL
      } /* unchunked units  */
    #endif
    /* common end-path  */
    #if DECTRACE
      decDumpAr('*', acc, accunits);         /* Show exact result  */
    #endif

    /* acc now contains the exact result of the multiplication,  */
    /* possibly with a leading zero unit; build the decNumber from  */
    /* it, noting if any residue  */
    res->bits=bits;                          /* set sign  */
    res->digits=decGetDigits(acc, accunits); /* count digits exactly  */

    /* There can be a 31-bit wrap in calculating the exponent.  */
    /* This can only happen if both input exponents are negative and  */
    /* both their magnitudes are large.  If there was a wrap, set a  */
    /* safe very negative exponent, from which decFinalize() will  */
    /* raise a hard underflow shortly.  */
    exponent=lhs->exponent+rhs->exponent;    /* calculate exponent  */
    if (lhs->exponent<0 && rhs->exponent<0 && exponent>0)
      exponent=-2*DECNUMMAXE;                /* force underflow  */
    res->exponent=exponent;                  /* OK to overwrite now  */


    /* Set the coefficient.  If any rounding, residue records  */
    decSetCoeff(res, set, acc, res->digits, &residue, status);
    decFinish(res, set, &residue, status);   /* final cleanup  */
    } while(0);                         /* end protected  */

  if (allocacc!=NULL) free(allocacc);   /* drop any storage used  */
  #if DECSUBSET
  if (allocrhs!=NULL) free(allocrhs);   /* ..  */
  if (alloclhs!=NULL) free(alloclhs);   /* ..  */
  #endif
  #if FASTMUL
  if (allocrhi!=NULL) free(allocrhi);   /* ..  */
  if (alloclhi!=NULL) free(alloclhi);   /* ..  */
  #endif
  return res;
  } /* decMultiplyOp  */

/* ------------------------------------------------------------------ */
/* decExpOp -- effect exponentiation                                  */
/*                                                                    */
/*   This computes C = exp(A)                                         */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context; note that rounding mode has no effect        */
/*                                                                    */
/* C must have space for set->digits digits. status is updated but    */
/* not set.                                                           */
/*                                                                    */
/* Restrictions:                                                      */
/*                                                                    */
/*   digits, emax, and -emin in the context must be less than         */
/*   2*DEC_MAX_MATH (1999998), and the rhs must be within these       */
/*   bounds or a zero.  This is an internal routine, so these         */
/*   restrictions are contractual and not enforced.                   */
/*                                                                    */
/* A finite result is rounded using DEC_ROUND_HALF_EVEN; it will      */
/* almost always be correctly rounded, but may be up to 1 ulp in      */
/* error in rare cases.                                               */
/*                                                                    */
/* Finite results will always be full precision and Inexact, except   */
/* when A is a zero or -Infinity (giving 1 or 0 respectively).        */
/* ------------------------------------------------------------------ */
/* This approach used here is similar to the algorithm described in   */
/*                                                                    */
/*   Variable Precision Exponential Function, T. E. Hull and          */
/*   A. Abrham, ACM Transactions on Mathematical Software, Vol 12 #2, */
/*   pp79-91, ACM, June 1986.                                         */
/*                                                                    */
/* with the main difference being that the iterations in the series   */
/* evaluation are terminated dynamically (which does not require the  */
/* extra variable-precision variables which are expensive in this     */
/* context).                                                          */
/*                                                                    */
/* The error analysis in Hull & Abrham's paper applies except for the */
/* round-off error accumulation during the series evaluation.  This   */
/* code does not precalculate the number of iterations and so cannot  */
/* use Horner's scheme.  Instead, the accumulation is done at double- */
/* precision, which ensures that the additions of the terms are exact */
/* and do not accumulate round-off (and any round-off errors in the   */
/* terms themselves move 'to the right' faster than they can          */
/* accumulate).  This code also extends the calculation by allowing,  */
/* in the spirit of other decNumber operators, the input to be more   */
/* precise than the result (the precision used is based on the more   */
/* precise of the input or requested result).                         */
/*                                                                    */
/* Implementation notes:                                              */
/*                                                                    */
/* 1. This is separated out as decExpOp so it can be called from      */
/*    other Mathematical functions (notably Ln) with a wider range    */
/*    than normal.  In particular, it can handle the slightly wider   */
/*    (double) range needed by Ln (which has to be able to calculate  */
/*    exp(-x) where x can be the tiniest number (Ntiny).              */
/*                                                                    */
/* 2. Normalizing x to be <=0.1 (instead of <=1) reduces loop         */
/*    iterations by appoximately a third with additional (although    */
/*    diminishing) returns as the range is reduced to even smaller    */
/*    fractions.  However, h (the power of 10 used to correct the     */
/*    result at the end, see below) must be kept <=8 as otherwise     */
/*    the final result cannot be computed.  Hence the leverage is a   */
/*    sliding value (8-h), where potentially the range is reduced     */
/*    more for smaller values.                                        */
/*                                                                    */
/*    The leverage that can be applied in this way is severely        */
/*    limited by the cost of the raise-to-the power at the end,       */
/*    which dominates when the number of iterations is small (less    */
/*    than ten) or when rhs is short.  As an example, the adjustment  */
/*    x**10,000,000 needs 31 multiplications, all but one full-width. */
/*                                                                    */
/* 3. The restrictions (especially precision) could be raised with    */
/*    care, but the full decNumber range seems very hard within the   */
/*    32-bit limits.                                                  */
/*                                                                    */
/* 4. The working precisions for the static buffers are twice the     */
/*    obvious size to allow for calls from decNumberPower.            */
/* ------------------------------------------------------------------ */
decNumber * decExpOp(decNumber *res, const decNumber *rhs,
                         decContext *set, uInt *status) {
  uInt ignore=0;                   /* working status  */
  Int h;                           /* adjusted exponent for 0.xxxx  */
  Int p;                           /* working precision  */
  Int residue;                     /* rounding residue  */
  uInt needbytes;                  /* for space calculations  */
  const decNumber *x=rhs;          /* (may point to safe copy later)  */
  decContext aset, tset, dset;     /* working contexts  */
  Int comp;                        /* work  */

  /* the argument is often copied to normalize it, so (unusually) it  */
  /* is treated like other buffers, using DECBUFFER, +1 in case  */
  /* DECBUFFER is 0  */
  decNumber bufr[D2N(DECBUFFER*2+1)];
  decNumber *allocrhs=NULL;        /* non-NULL if rhs buffer allocated  */

  /* the working precision will be no more than set->digits+8+1  */
  /* so for on-stack buffers DECBUFFER+9 is used, +1 in case DECBUFFER  */
  /* is 0 (and twice that for the accumulator)  */

  /* buffer for t, term (working precision plus)  */
  decNumber buft[D2N(DECBUFFER*2+9+1)];
  decNumber *allocbuft=NULL;       /* -> allocated buft, iff allocated  */
  decNumber *t=buft;               /* term  */
  /* buffer for a, accumulator (working precision * 2), at least 9  */
  decNumber bufa[D2N(DECBUFFER*4+18+1)];
  decNumber *allocbufa=NULL;       /* -> allocated bufa, iff allocated  */
  decNumber *a=bufa;               /* accumulator  */
  /* decNumber for the divisor term; this needs at most 9 digits  */
  /* and so can be fixed size [16 so can use standard context]  */
  decNumber bufd[D2N(16)];
  decNumber *d=bufd;               /* divisor  */
  decNumber numone;                /* constant 1  */

  #if DECCHECK
  Int iterations=0;                /* for later sanity check  */
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  do {                                  /* protect allocated storage  */
    if (SPECIALARG) {                   /* handle infinities and NaNs  */
      if (decNumberIsInfinite(rhs)) {   /* an infinity  */
        if (decNumberIsNegative(rhs))   /* -Infinity -> +0  */
          uprv_decNumberZero(res);
         else uprv_decNumberCopy(res, rhs);  /* +Infinity -> self  */
        }
       else decNaNs(res, rhs, NULL, set, status); /* a NaN  */
      break;}

    if (ISZERO(rhs)) {                  /* zeros -> exact 1  */
      uprv_decNumberZero(res);               /* make clean 1  */
      *res->lsu=1;                      /* ..  */
      break;}                           /* [no status to set]  */

    /* e**x when 0 < x < 0.66 is < 1+3x/2, hence can fast-path  */
    /* positive and negative tiny cases which will result in inexact  */
    /* 1.  This also allows the later add-accumulate to always be  */
    /* exact (because its length will never be more than twice the  */
    /* working precision).  */
    /* The comparator (tiny) needs just one digit, so use the  */
    /* decNumber d for it (reused as the divisor, etc., below); its  */
    /* exponent is such that if x is positive it will have  */
    /* set->digits-1 zeros between the decimal point and the digit,  */
    /* which is 4, and if x is negative one more zero there as the  */
    /* more precise result will be of the form 0.9999999 rather than  */
    /* 1.0000001.  Hence, tiny will be 0.0000004  if digits=7 and x>0  */
    /* or 0.00000004 if digits=7 and x<0.  If RHS not larger than  */
    /* this then the result will be 1.000000  */
    uprv_decNumberZero(d);                   /* clean  */
    *d->lsu=4;                          /* set 4 ..  */
    d->exponent=-set->digits;           /* * 10**(-d)  */
    if (decNumberIsNegative(rhs)) d->exponent--;  /* negative case  */
    comp=decCompare(d, rhs, 1);         /* signless compare  */
    if (comp==DECNUMBER_BADINT) {
      *status|=DEC_Insufficient_storage;
      break;}
    if (comp>=0) {                      /* rhs < d  */
      Int shift=set->digits-1;
      uprv_decNumberZero(res);               /* set 1  */
      *res->lsu=1;                      /* ..  */
      res->digits=decShiftToMost(res->lsu, 1, shift);
      res->exponent=-shift;                  /* make 1.0000...  */
      *status|=DEC_Inexact | DEC_Rounded;    /* .. inexactly  */
      break;} /* tiny  */

    /* set up the context to be used for calculating a, as this is  */
    /* used on both paths below  */
    uprv_decContextDefault(&aset, DEC_INIT_DECIMAL64);
    /* accumulator bounds are as requested (could underflow)  */
    aset.emax=set->emax;                /* usual bounds  */
    aset.emin=set->emin;                /* ..  */
    aset.clamp=0;                       /* and no concrete format  */

    /* calculate the adjusted (Hull & Abrham) exponent (where the  */
    /* decimal point is just to the left of the coefficient msd)  */
    h=rhs->exponent+rhs->digits;
    /* if h>8 then 10**h cannot be calculated safely; however, when  */
    /* h=8 then exp(|rhs|) will be at least exp(1E+7) which is at  */
    /* least 6.59E+4342944, so (due to the restriction on Emax/Emin)  */
    /* overflow (or underflow to 0) is guaranteed -- so this case can  */
    /* be handled by simply forcing the appropriate excess  */
    if (h>8) {                          /* overflow/underflow  */
      /* set up here so Power call below will over or underflow to  */
      /* zero; set accumulator to either 2 or 0.02  */
      /* [stack buffer for a is always big enough for this]  */
      uprv_decNumberZero(a);
      *a->lsu=2;                        /* not 1 but < exp(1)  */
      if (decNumberIsNegative(rhs)) a->exponent=-2; /* make 0.02  */
      h=8;                              /* clamp so 10**h computable  */
      p=9;                              /* set a working precision  */
      }
     else {                             /* h<=8  */
      Int maxlever=(rhs->digits>8?1:0);
      /* [could/should increase this for precisions >40 or so, too]  */

      /* if h is 8, cannot normalize to a lower upper limit because  */
      /* the final result will not be computable (see notes above),  */
      /* but leverage can be applied whenever h is less than 8.  */
      /* Apply as much as possible, up to a MAXLEVER digits, which  */
      /* sets the tradeoff against the cost of the later a**(10**h).  */
      /* As h is increased, the working precision below also  */
      /* increases to compensate for the "constant digits at the  */
      /* front" effect.  */
      Int lever=MINI(8-h, maxlever);    /* leverage attainable  */
      Int use=-rhs->digits-lever;       /* exponent to use for RHS  */
      h+=lever;                         /* apply leverage selected  */
      if (h<0) {                        /* clamp  */
        use+=h;                         /* [may end up subnormal]  */
        h=0;
        }
      /* Take a copy of RHS if it needs normalization (true whenever x>=1)  */
      if (rhs->exponent!=use) {
        decNumber *newrhs=bufr;         /* assume will fit on stack  */
        needbytes=sizeof(decNumber)+(D2U(rhs->digits)-1)*sizeof(Unit);
        if (needbytes>sizeof(bufr)) {   /* need malloc space  */
          allocrhs=(decNumber *)malloc(needbytes);
          if (allocrhs==NULL) {         /* hopeless -- abandon  */
            *status|=DEC_Insufficient_storage;
            break;}
          newrhs=allocrhs;              /* use the allocated space  */
          }
        uprv_decNumberCopy(newrhs, rhs);     /* copy to safe space  */
        newrhs->exponent=use;           /* normalize; now <1  */
        x=newrhs;                       /* ready for use  */
        /* decNumberShow(x);  */
        }

      /* Now use the usual power series to evaluate exp(x).  The  */
      /* series starts as 1 + x + x^2/2 ... so prime ready for the  */
      /* third term by setting the term variable t=x, the accumulator  */
      /* a=1, and the divisor d=2.  */

      /* First determine the working precision.  From Hull & Abrham  */
      /* this is set->digits+h+2.  However, if x is 'over-precise' we  */
      /* need to allow for all its digits to potentially participate  */
      /* (consider an x where all the excess digits are 9s) so in  */
      /* this case use x->digits+h+2  */
      p=MAXI(x->digits, set->digits)+h+2;    /* [h<=8]  */

      /* a and t are variable precision, and depend on p, so space  */
      /* must be allocated for them if necessary  */

      /* the accumulator needs to be able to hold 2p digits so that  */
      /* the additions on the second and subsequent iterations are  */
      /* sufficiently exact.  */
      needbytes=sizeof(decNumber)+(D2U(p*2)-1)*sizeof(Unit);
      if (needbytes>sizeof(bufa)) {     /* need malloc space  */
        allocbufa=(decNumber *)malloc(needbytes);
        if (allocbufa==NULL) {          /* hopeless -- abandon  */
          *status|=DEC_Insufficient_storage;
          break;}
        a=allocbufa;                    /* use the allocated space  */
        }
      /* the term needs to be able to hold p digits (which is  */
      /* guaranteed to be larger than x->digits, so the initial copy  */
      /* is safe); it may also be used for the raise-to-power  */
      /* calculation below, which needs an extra two digits  */
      needbytes=sizeof(decNumber)+(D2U(p+2)-1)*sizeof(Unit);
      if (needbytes>sizeof(buft)) {     /* need malloc space  */
        allocbuft=(decNumber *)malloc(needbytes);
        if (allocbuft==NULL) {          /* hopeless -- abandon  */
          *status|=DEC_Insufficient_storage;
          break;}
        t=allocbuft;                    /* use the allocated space  */
        }

      uprv_decNumberCopy(t, x);              /* term=x  */
      uprv_decNumberZero(a); *a->lsu=1;      /* accumulator=1  */
      uprv_decNumberZero(d); *d->lsu=2;      /* divisor=2  */
      uprv_decNumberZero(&numone); *numone.lsu=1; /* constant 1 for increment  */

      /* set up the contexts for calculating a, t, and d  */
      uprv_decContextDefault(&tset, DEC_INIT_DECIMAL64);
      dset=tset;
      /* accumulator bounds are set above, set precision now  */
      aset.digits=p*2;                  /* double  */
      /* term bounds avoid any underflow or overflow  */
      tset.digits=p;
      tset.emin=DEC_MIN_EMIN;           /* [emax is plenty]  */
      /* [dset.digits=16, etc., are sufficient]  */

      /* finally ready to roll  */
      for (;;) {
        #if DECCHECK
        iterations++;
        #endif
        /* only the status from the accumulation is interesting  */
        /* [but it should remain unchanged after first add]  */
        decAddOp(a, a, t, &aset, 0, status);           /* a=a+t  */
        decMultiplyOp(t, t, x, &tset, &ignore);        /* t=t*x  */
        decDivideOp(t, t, d, &tset, DECNUMBER_DIVIDE, &ignore);  /* t=t/d  */
        /* the iteration ends when the term cannot affect the result,  */
        /* if rounded to p digits, which is when its value is smaller  */
        /* than the accumulator by p+1 digits.  There must also be  */
        /* full precision in a.  */
        if (((a->digits+a->exponent)>=(t->digits+t->exponent+p+1))
            && (a->digits>=p)) break;
        decAddOp(d, d, &numone, &dset, 0, &ignore);    /* d=d+1  */
        } /* iterate  */

      #if DECCHECK
      /* just a sanity check; comment out test to show always  */
      if (iterations>p+3)
        printf("Exp iterations=%ld, status=%08lx, p=%ld, d=%ld\n",
               (LI)iterations, (LI)*status, (LI)p, (LI)x->digits);
      #endif
      } /* h<=8  */

    /* apply postconditioning: a=a**(10**h) -- this is calculated  */
    /* at a slightly higher precision than Hull & Abrham suggest  */
    if (h>0) {
      Int seenbit=0;               /* set once a 1-bit is seen  */
      Int i;                       /* counter  */
      Int n=DECNUMBER_powers[h];             /* always positive  */
      aset.digits=p+2;             /* sufficient precision  */
      /* avoid the overhead and many extra digits of decNumberPower  */
      /* as all that is needed is the short 'multipliers' loop; here  */
      /* accumulate the answer into t  */
      uprv_decNumberZero(t); *t->lsu=1; /* acc=1  */
      for (i=1;;i++){              /* for each bit [top bit ignored]  */
        /* abandon if have had overflow or terminal underflow  */
        if (*status & (DEC_Overflow|DEC_Underflow)) { /* interesting?  */
          if (*status&DEC_Overflow || ISZERO(t)) break;}
        n=n<<1;                    /* move next bit to testable position  */
        if (n<0) {                 /* top bit is set  */
          seenbit=1;               /* OK, have a significant bit  */
          decMultiplyOp(t, t, a, &aset, status); /* acc=acc*x  */
          }
        if (i==31) break;          /* that was the last bit  */
        if (!seenbit) continue;    /* no need to square 1  */
        decMultiplyOp(t, t, t, &aset, status); /* acc=acc*acc [square]  */
        } /*i*/ /* 32 bits  */
      /* decNumberShow(t);  */
      a=t;                         /* and carry on using t instead of a  */
      }

    /* Copy and round the result to res  */
    residue=1;                          /* indicate dirt to right ..  */
    if (ISZERO(a)) residue=0;           /* .. unless underflowed to 0  */
    aset.digits=set->digits;            /* [use default rounding]  */
    decCopyFit(res, a, &aset, &residue, status); /* copy & shorten  */
    decFinish(res, set, &residue, status);       /* cleanup/set flags  */
    } while(0);                         /* end protected  */

  if (allocrhs !=NULL) free(allocrhs);  /* drop any storage used  */
  if (allocbufa!=NULL) free(allocbufa); /* ..  */
  if (allocbuft!=NULL) free(allocbuft); /* ..  */
  /* [status is handled by caller]  */
  return res;
  } /* decExpOp  */

/* ------------------------------------------------------------------ */
/* Initial-estimate natural logarithm table                           */
/*                                                                    */
/*   LNnn -- 90-entry 16-bit table for values from .10 through .99.   */
/*           The result is a 4-digit encode of the coefficient (c=the */
/*           top 14 bits encoding 0-9999) and a 2-digit encode of the */
/*           exponent (e=the bottom 2 bits encoding 0-3)              */
/*                                                                    */
/*           The resulting value is given by:                         */
/*                                                                    */
/*             v = -c * 10**(-e-3)                                    */
/*                                                                    */
/*           where e and c are extracted from entry k = LNnn[x-10]    */
/*           where x is truncated (NB) into the range 10 through 99,  */
/*           and then c = k>>2 and e = k&3.                           */
/* ------------------------------------------------------------------ */
static const uShort LNnn[90]={9016,  8652,  8316,  8008,  7724,  7456,  7208,
  6972,  6748,  6540,  6340,  6148,  5968,  5792,  5628,  5464,  5312,
  5164,  5020,  4884,  4748,  4620,  4496,  4376,  4256,  4144,  4032,
 39233, 38181, 37157, 36157, 35181, 34229, 33297, 32389, 31501, 30629,
 29777, 28945, 28129, 27329, 26545, 25777, 25021, 24281, 23553, 22837,
 22137, 21445, 20769, 20101, 19445, 18801, 18165, 17541, 16925, 16321,
 15721, 15133, 14553, 13985, 13421, 12865, 12317, 11777, 11241, 10717,
 10197,  9685,  9177,  8677,  8185,  7697,  7213,  6737,  6269,  5801,
  5341,  4889,  4437, 39930, 35534, 31186, 26886, 22630, 18418, 14254,
 10130,  6046, 20055};

/* ------------------------------------------------------------------ */
/* decLnOp -- effect natural logarithm                                */
/*                                                                    */
/*   This computes C = ln(A)                                          */
/*                                                                    */
/*   res is C, the result.  C may be A                                */
/*   rhs is A                                                         */
/*   set is the context; note that rounding mode has no effect        */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Notable cases:                                                     */
/*   A<0 -> Invalid                                                   */
/*   A=0 -> -Infinity (Exact)                                         */
/*   A=+Infinity -> +Infinity (Exact)                                 */
/*   A=1 exactly -> 0 (Exact)                                         */
/*                                                                    */
/* Restrictions (as for Exp):                                         */
/*                                                                    */
/*   digits, emax, and -emin in the context must be less than         */
/*   DEC_MAX_MATH+11 (1000010), and the rhs must be within these      */
/*   bounds or a zero.  This is an internal routine, so these         */
/*   restrictions are contractual and not enforced.                   */
/*                                                                    */
/* A finite result is rounded using DEC_ROUND_HALF_EVEN; it will      */
/* almost always be correctly rounded, but may be up to 1 ulp in      */
/* error in rare cases.                                               */
/* ------------------------------------------------------------------ */
/* The result is calculated using Newton's method, with each          */
/* iteration calculating a' = a + x * exp(-a) - 1.  See, for example, */
/* Epperson 1989.                                                     */
/*                                                                    */
/* The iteration ends when the adjustment x*exp(-a)-1 is tiny enough. */
/* This has to be calculated at the sum of the precision of x and the */
/* working precision.                                                 */
/*                                                                    */
/* Implementation notes:                                              */
/*                                                                    */
/* 1. This is separated out as decLnOp so it can be called from       */
/*    other Mathematical functions (e.g., Log 10) with a wider range  */
/*    than normal.  In particular, it can handle the slightly wider   */
/*    (+9+2) range needed by a power function.                        */
/*                                                                    */
/* 2. The speed of this function is about 10x slower than exp, as     */
/*    it typically needs 4-6 iterations for short numbers, and the    */
/*    extra precision needed adds a squaring effect, twice.           */
/*                                                                    */
/* 3. Fastpaths are included for ln(10) and ln(2), up to length 40,   */
/*    as these are common requests.  ln(10) is used by log10(x).      */
/*                                                                    */
/* 4. An iteration might be saved by widening the LNnn table, and     */
/*    would certainly save at least one if it were made ten times     */
/*    bigger, too (for truncated fractions 0.100 through 0.999).      */
/*    However, for most practical evaluations, at least four or five  */
/*    iterations will be neede -- so this would only speed up by      */
/*    20-25% and that probably does not justify increasing the table  */
/*    size.                                                           */
/*                                                                    */
/* 5. The static buffers are larger than might be expected to allow   */
/*    for calls from decNumberPower.                                  */
/* ------------------------------------------------------------------ */
#if defined(__clang__) || U_GCC_MAJOR_MINOR >= 406
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Warray-bounds"
#endif
decNumber * decLnOp(decNumber *res, const decNumber *rhs,
                    decContext *set, uInt *status) {
  uInt ignore=0;                   /* working status accumulator  */
  uInt needbytes;                  /* for space calculations  */
  Int residue;                     /* rounding residue  */
  Int r;                           /* rhs=f*10**r [see below]  */
  Int p;                           /* working precision  */
  Int pp;                          /* precision for iteration  */
  Int t;                           /* work  */

  /* buffers for a (accumulator, typically precision+2) and b  */
  /* (adjustment calculator, same size)  */
  decNumber bufa[D2N(DECBUFFER+12)];
  decNumber *allocbufa=NULL;       /* -> allocated bufa, iff allocated  */
  decNumber *a=bufa;               /* accumulator/work  */
  decNumber bufb[D2N(DECBUFFER*2+2)];
  decNumber *allocbufb=NULL;       /* -> allocated bufa, iff allocated  */
  decNumber *b=bufb;               /* adjustment/work  */

  decNumber  numone;               /* constant 1  */
  decNumber  cmp;                  /* work  */
  decContext aset, bset;           /* working contexts  */

  #if DECCHECK
  Int iterations=0;                /* for later sanity check  */
  if (decCheckOperands(res, DECUNUSED, rhs, set)) return res;
  #endif

  do {                                  /* protect allocated storage  */
    if (SPECIALARG) {                   /* handle infinities and NaNs  */
      if (decNumberIsInfinite(rhs)) {   /* an infinity  */
        if (decNumberIsNegative(rhs))   /* -Infinity -> error  */
          *status|=DEC_Invalid_operation;
         else uprv_decNumberCopy(res, rhs);  /* +Infinity -> self  */
        }
       else decNaNs(res, rhs, NULL, set, status); /* a NaN  */
      break;}

    if (ISZERO(rhs)) {                  /* +/- zeros -> -Infinity  */
      uprv_decNumberZero(res);               /* make clean  */
      res->bits=DECINF|DECNEG;          /* set - infinity  */
      break;}                           /* [no status to set]  */

    /* Non-zero negatives are bad...  */
    if (decNumberIsNegative(rhs)) {     /* -x -> error  */
      *status|=DEC_Invalid_operation;
      break;}

    /* Here, rhs is positive, finite, and in range  */

    /* lookaside fastpath code for ln(2) and ln(10) at common lengths  */
    if (rhs->exponent==0 && set->digits<=40) {
      #if DECDPUN==1
      if (rhs->lsu[0]==0 && rhs->lsu[1]==1 && rhs->digits==2) { /* ln(10)  */
      #else
      if (rhs->lsu[0]==10 && rhs->digits==2) {                  /* ln(10)  */
      #endif
        aset=*set; aset.round=DEC_ROUND_HALF_EVEN;
        #define LN10 "2.302585092994045684017991454684364207601"
        uprv_decNumberFromString(res, LN10, &aset);
        *status|=(DEC_Inexact | DEC_Rounded); /* is inexact  */
        break;}
      if (rhs->lsu[0]==2 && rhs->digits==1) { /* ln(2)  */
        aset=*set; aset.round=DEC_ROUND_HALF_EVEN;
        #define LN2 "0.6931471805599453094172321214581765680755"
        uprv_decNumberFromString(res, LN2, &aset);
        *status|=(DEC_Inexact | DEC_Rounded);
        break;}
      } /* integer and short  */

    /* Determine the working precision.  This is normally the  */
    /* requested precision + 2, with a minimum of 9.  However, if  */
    /* the rhs is 'over-precise' then allow for all its digits to  */
    /* potentially participate (consider an rhs where all the excess  */
    /* digits are 9s) so in this case use rhs->digits+2.  */
    p=MAXI(rhs->digits, MAXI(set->digits, 7))+2;

    /* Allocate space for the accumulator and the high-precision  */
    /* adjustment calculator, if necessary.  The accumulator must  */
    /* be able to hold p digits, and the adjustment up to  */
    /* rhs->digits+p digits.  They are also made big enough for 16  */
    /* digits so that they can be used for calculating the initial  */
    /* estimate.  */
    needbytes=sizeof(decNumber)+(D2U(MAXI(p,16))-1)*sizeof(Unit);
    if (needbytes>sizeof(bufa)) {     /* need malloc space  */
      allocbufa=(decNumber *)malloc(needbytes);
      if (allocbufa==NULL) {          /* hopeless -- abandon  */
        *status|=DEC_Insufficient_storage;
        break;}
      a=allocbufa;                    /* use the allocated space  */
      }
    pp=p+rhs->digits;
    needbytes=sizeof(decNumber)+(D2U(MAXI(pp,16))-1)*sizeof(Unit);
    if (needbytes>sizeof(bufb)) {     /* need malloc space  */
      allocbufb=(decNumber *)malloc(needbytes);
      if (allocbufb==NULL) {          /* hopeless -- abandon  */
        *status|=DEC_Insufficient_storage;
        break;}
      b=allocbufb;                    /* use the allocated space  */
      }

    /* Prepare an initial estimate in acc. Calculate this by  */
    /* considering the coefficient of x to be a normalized fraction,  */
    /* f, with the decimal point at far left and multiplied by  */
    /* 10**r.  Then, rhs=f*10**r and 0.1<=f<1, and  */
    /*   ln(x) = ln(f) + ln(10)*r  */
    /* Get the initial estimate for ln(f) from a small lookup  */
    /* table (see above) indexed by the first two digits of f,  */
    /* truncated.  */

    uprv_decContextDefault(&aset, DEC_INIT_DECIMAL64); /* 16-digit extended  */
    r=rhs->exponent+rhs->digits;        /* 'normalised' exponent  */
    uprv_decNumberFromInt32(a, r);           /* a=r  */
    uprv_decNumberFromInt32(b, 2302585);     /* b=ln(10) (2.302585)  */
    b->exponent=-6;                     /*  ..  */
    decMultiplyOp(a, a, b, &aset, &ignore);  /* a=a*b  */
    /* now get top two digits of rhs into b by simple truncate and  */
    /* force to integer  */
    residue=0;                          /* (no residue)  */
    aset.digits=2; aset.round=DEC_ROUND_DOWN;
    decCopyFit(b, rhs, &aset, &residue, &ignore); /* copy & shorten  */
    b->exponent=0;                      /* make integer  */
    t=decGetInt(b);                     /* [cannot fail]  */
    if (t<10) t=X10(t);                 /* adjust single-digit b  */
    t=LNnn[t-10];                       /* look up ln(b)  */
    uprv_decNumberFromInt32(b, t>>2);        /* b=ln(b) coefficient  */
    b->exponent=-(t&3)-3;               /* set exponent  */
    b->bits=DECNEG;                     /* ln(0.10)->ln(0.99) always -ve  */
    aset.digits=16; aset.round=DEC_ROUND_HALF_EVEN; /* restore  */
    decAddOp(a, a, b, &aset, 0, &ignore); /* acc=a+b  */
    /* the initial estimate is now in a, with up to 4 digits correct.  */
    /* When rhs is at or near Nmax the estimate will be low, so we  */
    /* will approach it from below, avoiding overflow when calling exp.  */

    uprv_decNumberZero(&numone); *numone.lsu=1;   /* constant 1 for adjustment  */

    /* accumulator bounds are as requested (could underflow, but  */
    /* cannot overflow)  */
    aset.emax=set->emax;
    aset.emin=set->emin;
    aset.clamp=0;                       /* no concrete format  */
    /* set up a context to be used for the multiply and subtract  */
    bset=aset;
    bset.emax=DEC_MAX_MATH*2;           /* use double bounds for the  */
    bset.emin=-DEC_MAX_MATH*2;          /* adjustment calculation  */
                                        /* [see decExpOp call below]  */
    /* for each iteration double the number of digits to calculate,  */
    /* up to a maximum of p  */
    pp=9;                               /* initial precision  */
    /* [initially 9 as then the sequence starts 7+2, 16+2, and  */
    /* 34+2, which is ideal for standard-sized numbers]  */
    aset.digits=pp;                     /* working context  */
    bset.digits=pp+rhs->digits;         /* wider context  */
    for (;;) {                          /* iterate  */
      #if DECCHECK
      iterations++;
      if (iterations>24) break;         /* consider 9 * 2**24  */
      #endif
      /* calculate the adjustment (exp(-a)*x-1) into b.  This is a  */
      /* catastrophic subtraction but it really is the difference  */
      /* from 1 that is of interest.  */
      /* Use the internal entry point to Exp as it allows the double  */
      /* range for calculating exp(-a) when a is the tiniest subnormal.  */
      a->bits^=DECNEG;                  /* make -a  */
      decExpOp(b, a, &bset, &ignore);   /* b=exp(-a)  */
      a->bits^=DECNEG;                  /* restore sign of a  */
      /* now multiply by rhs and subtract 1, at the wider precision  */
      decMultiplyOp(b, b, rhs, &bset, &ignore);        /* b=b*rhs  */
      decAddOp(b, b, &numone, &bset, DECNEG, &ignore); /* b=b-1  */

      /* the iteration ends when the adjustment cannot affect the  */
      /* result by >=0.5 ulp (at the requested digits), which  */
      /* is when its value is smaller than the accumulator by  */
      /* set->digits+1 digits (or it is zero) -- this is a looser  */
      /* requirement than for Exp because all that happens to the  */
      /* accumulator after this is the final rounding (but note that  */
      /* there must also be full precision in a, or a=0).  */

      if (decNumberIsZero(b) ||
          (a->digits+a->exponent)>=(b->digits+b->exponent+set->digits+1)) {
        if (a->digits==p) break;
        if (decNumberIsZero(a)) {
          decCompareOp(&cmp, rhs, &numone, &aset, DECNUMBER_COMPARE, &ignore); /* rhs=1 ?  */
          if (cmp.lsu[0]==0) a->exponent=0;            /* yes, exact 0  */
           else *status|=(DEC_Inexact | DEC_Rounded);  /* no, inexact  */
          break;
          }
        /* force padding if adjustment has gone to 0 before full length  */
        if (decNumberIsZero(b)) b->exponent=a->exponent-p;
        }

      /* not done yet ...  */
      decAddOp(a, a, b, &aset, 0, &ignore);  /* a=a+b for next estimate  */
      if (pp==p) continue;                   /* precision is at maximum  */
      /* lengthen the next calculation  */
      pp=pp*2;                               /* double precision  */
      if (pp>p) pp=p;                        /* clamp to maximum  */
      aset.digits=pp;                        /* working context  */
      bset.digits=pp+rhs->digits;            /* wider context  */
      } /* Newton's iteration  */

    #if DECCHECK
    /* just a sanity check; remove the test to show always  */
    if (iterations>24)
      printf("Ln iterations=%ld, status=%08lx, p=%ld, d=%ld\n",
            (LI)iterations, (LI)*status, (LI)p, (LI)rhs->digits);
    #endif

    /* Copy and round the result to res  */
    residue=1;                          /* indicate dirt to right  */
    if (ISZERO(a)) residue=0;           /* .. unless underflowed to 0  */
    aset.digits=set->digits;            /* [use default rounding]  */
    decCopyFit(res, a, &aset, &residue, status); /* copy & shorten  */
    decFinish(res, set, &residue, status);       /* cleanup/set flags  */
    } while(0);                         /* end protected  */

  if (allocbufa!=NULL) free(allocbufa); /* drop any storage used  */
  if (allocbufb!=NULL) free(allocbufb); /* ..  */
  /* [status is handled by caller]  */
  return res;
  } /* decLnOp  */
#if defined(__clang__) || U_GCC_MAJOR_MINOR >= 406
#pragma GCC diagnostic pop
#endif

/* ------------------------------------------------------------------ */
/* decQuantizeOp  -- force exponent to requested value                */
/*                                                                    */
/*   This computes C = op(A, B), where op adjusts the coefficient     */
/*   of C (by rounding or shifting) such that the exponent (-scale)   */
/*   of C has the value B or matches the exponent of B.               */
/*   The numerical value of C will equal A, except for the effects of */
/*   any rounding that occurred.                                      */
/*                                                                    */
/*   res is C, the result.  C may be A or B                           */
/*   lhs is A, the number to adjust                                   */
/*   rhs is B, the requested exponent                                 */
/*   set is the context                                               */
/*   quant is 1 for quantize or 0 for rescale                         */
/*   status is the status accumulator (this can be called without     */
/*          risk of control loss)                                     */
/*                                                                    */
/* C must have space for set->digits digits.                          */
/*                                                                    */
/* Unless there is an error or the result is infinite, the exponent   */
/* after the operation is guaranteed to be that requested.            */
/* ------------------------------------------------------------------ */
static decNumber * decQuantizeOp(decNumber *res, const decNumber *lhs,
                                 const decNumber *rhs, decContext *set,
                                 Flag quant, uInt *status) {
  #if DECSUBSET
  decNumber *alloclhs=NULL;        /* non-NULL if rounded lhs allocated  */
  decNumber *allocrhs=NULL;        /* .., rhs  */
  #endif
  const decNumber *inrhs=rhs;      /* save original rhs  */
  Int   reqdigits=set->digits;     /* requested DIGITS  */
  Int   reqexp;                    /* requested exponent [-scale]  */
  Int   residue=0;                 /* rounding residue  */
  Int   etiny=set->emin-(reqdigits-1);

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  do {                             /* protect allocated storage  */
    #if DECSUBSET
    if (!set->extended) {
      /* reduce operands and set lostDigits status, as needed  */
      if (lhs->digits>reqdigits) {
        alloclhs=decRoundOperand(lhs, set, status);
        if (alloclhs==NULL) break;
        lhs=alloclhs;
        }
      if (rhs->digits>reqdigits) { /* [this only checks lostDigits]  */
        allocrhs=decRoundOperand(rhs, set, status);
        if (allocrhs==NULL) break;
        rhs=allocrhs;
        }
      }
    #endif
    /* [following code does not require input rounding]  */

    /* Handle special values  */
    if (SPECIALARGS) {
      /* NaNs get usual processing  */
      if (SPECIALARGS & (DECSNAN | DECNAN))
        decNaNs(res, lhs, rhs, set, status);
      /* one infinity but not both is bad  */
      else if ((lhs->bits ^ rhs->bits) & DECINF)
        *status|=DEC_Invalid_operation;
      /* both infinity: return lhs  */
      else uprv_decNumberCopy(res, lhs);          /* [nop if in place]  */
      break;
      }

    /* set requested exponent  */
    if (quant) reqexp=inrhs->exponent;  /* quantize -- match exponents  */
     else {                             /* rescale -- use value of rhs  */
      /* Original rhs must be an integer that fits and is in range,  */
      /* which could be from -1999999997 to +999999999, thanks to  */
      /* subnormals  */
      reqexp=decGetInt(inrhs);               /* [cannot fail]  */
      }

    #if DECSUBSET
    if (!set->extended) etiny=set->emin;     /* no subnormals  */
    #endif

    if (reqexp==DECNUMBER_BADINT                       /* bad (rescale only) or ..  */
     || reqexp==DECNUMBER_BIGODD || reqexp==DECNUMBER_BIGEVEN    /* very big (ditto) or ..  */
     || (reqexp<etiny)                       /* < lowest  */
     || (reqexp>set->emax)) {                /* > emax  */
      *status|=DEC_Invalid_operation;
      break;}

    /* the RHS has been processed, so it can be overwritten now if necessary  */
    if (ISZERO(lhs)) {                       /* zero coefficient unchanged  */
      uprv_decNumberCopy(res, lhs);               /* [nop if in place]  */
      res->exponent=reqexp;                  /* .. just set exponent  */
      #if DECSUBSET
      if (!set->extended) res->bits=0;       /* subset specification; no -0  */
      #endif
      }
     else {                                  /* non-zero lhs  */
      Int adjust=reqexp-lhs->exponent;       /* digit adjustment needed  */
      /* if adjusted coefficient will definitely not fit, give up now  */
      if ((lhs->digits-adjust)>reqdigits) {
        *status|=DEC_Invalid_operation;
        break;
        }

      if (adjust>0) {                        /* increasing exponent  */
        /* this will decrease the length of the coefficient by adjust  */
        /* digits, and must round as it does so  */
        decContext workset;                  /* work  */
        workset=*set;                        /* clone rounding, etc.  */
        workset.digits=lhs->digits-adjust;   /* set requested length  */
        /* [note that the latter can be <1, here]  */
        decCopyFit(res, lhs, &workset, &residue, status); /* fit to result  */
        decApplyRound(res, &workset, residue, status);    /* .. and round  */
        residue=0;                                        /* [used]  */
        /* If just rounded a 999s case, exponent will be off by one;  */
        /* adjust back (after checking space), if so.  */
        if (res->exponent>reqexp) {
          /* re-check needed, e.g., for quantize(0.9999, 0.001) under  */
          /* set->digits==3  */
          if (res->digits==reqdigits) {      /* cannot shift by 1  */
            *status&=~(DEC_Inexact | DEC_Rounded); /* [clean these]  */
            *status|=DEC_Invalid_operation;
            break;
            }
          res->digits=decShiftToMost(res->lsu, res->digits, 1); /* shift  */
          res->exponent--;                   /* (re)adjust the exponent.  */
          }
        #if DECSUBSET
        if (ISZERO(res) && !set->extended) res->bits=0; /* subset; no -0  */
        #endif
        } /* increase  */
       else /* adjust<=0 */ {                /* decreasing or = exponent  */
        /* this will increase the length of the coefficient by -adjust  */
        /* digits, by adding zero or more trailing zeros; this is  */
        /* already checked for fit, above  */
        uprv_decNumberCopy(res, lhs);             /* [it will fit]  */
        /* if padding needed (adjust<0), add it now...  */
        if (adjust<0) {
          res->digits=decShiftToMost(res->lsu, res->digits, -adjust);
          res->exponent+=adjust;             /* adjust the exponent  */
          }
        } /* decrease  */
      } /* non-zero  */

    /* Check for overflow [do not use Finalize in this case, as an  */
    /* overflow here is a "don't fit" situation]  */
    if (res->exponent>set->emax-res->digits+1) {  /* too big  */
      *status|=DEC_Invalid_operation;
      break;
      }
     else {
      decFinalize(res, set, &residue, status);    /* set subnormal flags  */
      *status&=~DEC_Underflow;          /* suppress Underflow [as per 754]  */
      }
    } while(0);                         /* end protected  */

  #if DECSUBSET
  if (allocrhs!=NULL) free(allocrhs);   /* drop any storage used  */
  if (alloclhs!=NULL) free(alloclhs);   /* ..  */
  #endif
  return res;
  } /* decQuantizeOp  */

/* ------------------------------------------------------------------ */
/* decCompareOp -- compare, min, or max two Numbers                   */
/*                                                                    */
/*   This computes C = A ? B and carries out one of four operations:  */
/*     DECNUMBER_COMPARE    -- returns the signum (as a number) giving the      */
/*                   result of a comparison unless one or both        */
/*                   operands is a NaN (in which case a NaN results)  */
/*     DECNUMBER_COMPSIG    -- as DECNUMBER_COMPARE except that a quiet NaN raises        */
/*                   Invalid operation.                               */
/*     DECNUMBER_COMPMAX    -- returns the larger of the operands, using the    */
/*                   754 maxnum operation                             */
/*     DECNUMBER_COMPMAXMAG -- ditto, comparing absolute values                 */
/*     DECNUMBER_COMPMIN    -- the 754 minnum operation                         */
/*     DECNUMBER_COMPMINMAG -- ditto, comparing absolute values                 */
/*     COMTOTAL   -- returns the signum (as a number) giving the      */
/*                   result of a comparison using 754 total ordering  */
/*                                                                    */
/*   res is C, the result.  C may be A and/or B (e.g., X=X?X)         */
/*   lhs is A                                                         */
/*   rhs is B                                                         */
/*   set is the context                                               */
/*   op  is the operation flag                                        */
/*   status is the usual accumulator                                  */
/*                                                                    */
/* C must have space for one digit for DECNUMBER_COMPARE or set->digits for     */
/* DECNUMBER_COMPMAX, DECNUMBER_COMPMIN, DECNUMBER_COMPMAXMAG, or DECNUMBER_COMPMINMAG.                       */
/* ------------------------------------------------------------------ */
/* The emphasis here is on speed for common cases, and avoiding       */
/* coefficient comparison if possible.                                */
/* ------------------------------------------------------------------ */
static decNumber * decCompareOp(decNumber *res, const decNumber *lhs,
                         const decNumber *rhs, decContext *set,
                         Flag op, uInt *status) {
  #if DECSUBSET
  decNumber *alloclhs=NULL;        /* non-NULL if rounded lhs allocated  */
  decNumber *allocrhs=NULL;        /* .., rhs  */
  #endif
  Int   result=0;                  /* default result value  */
  uByte merged;                    /* work  */

  #if DECCHECK
  if (decCheckOperands(res, lhs, rhs, set)) return res;
  #endif

  do {                             /* protect allocated storage  */
    #if DECSUBSET
    if (!set->extended) {
      /* reduce operands and set lostDigits status, as needed  */
      if (lhs->digits>set->digits) {
        alloclhs=decRoundOperand(lhs, set, status);
        if (alloclhs==NULL) {result=DECNUMBER_BADINT; break;}
        lhs=alloclhs;
        }
      if (rhs->digits>set->digits) {
        allocrhs=decRoundOperand(rhs, set, status);
        if (allocrhs==NULL) {result=DECNUMBER_BADINT; break;}
        rhs=allocrhs;
        }
      }
    #endif
    /* [following code does not require input rounding]  */

    /* If total ordering then handle differing signs 'up front'  */
    if (op==DECNUMBER_COMPTOTAL) {                /* total ordering  */
      if (decNumberIsNegative(lhs) && !decNumberIsNegative(rhs)) {
        result=-1;
        break;
        }
      if (!decNumberIsNegative(lhs) && decNumberIsNegative(rhs)) {
        result=+1;
        break;
        }
      }

    /* handle NaNs specially; let infinities drop through  */
    /* This assumes sNaN (even just one) leads to NaN.  */
    merged=(lhs->bits | rhs->bits) & (DECSNAN | DECNAN);
    if (merged) {                       /* a NaN bit set  */
      if (op==DECNUMBER_COMPARE);                 /* result will be NaN  */
       else if (op==DECNUMBER_COMPSIG)            /* treat qNaN as sNaN  */
        *status|=DEC_Invalid_operation | DECNUMBER_DEC_sNaN;
       else if (op==DECNUMBER_COMPTOTAL) {        /* total ordering, always finite  */
        /* signs are known to be the same; compute the ordering here  */
        /* as if the signs are both positive, then invert for negatives  */
        if (!decNumberIsNaN(lhs)) result=-1;
         else if (!decNumberIsNaN(rhs)) result=+1;
         /* here if both NaNs  */
         else if (decNumberIsSNaN(lhs) && decNumberIsQNaN(rhs)) result=-1;
         else if (decNumberIsQNaN(lhs) && decNumberIsSNaN(rhs)) result=+1;
         else { /* both NaN or both sNaN  */
          /* now it just depends on the payload  */
          result=decUnitCompare(lhs->lsu, D2U(lhs->digits),
                                rhs->lsu, D2U(rhs->digits), 0);
          /* [Error not possible, as these are 'aligned']  */
          } /* both same NaNs  */
        if (decNumberIsNegative(lhs)) result=-result;
        break;
        } /* total order  */

       else if (merged & DECSNAN);           /* sNaN -> qNaN  */
       else { /* here if MIN or MAX and one or two quiet NaNs  */
        /* min or max -- 754 rules ignore single NaN  */
        if (!decNumberIsNaN(lhs) || !decNumberIsNaN(rhs)) {
          /* just one NaN; force choice to be the non-NaN operand  */
          op=DECNUMBER_COMPMAX;
          if (lhs->bits & DECNAN) result=-1; /* pick rhs  */
                             else result=+1; /* pick lhs  */
          break;
          }
        } /* max or min  */
      op=DECNUMBER_COMPNAN;                            /* use special path  */
      decNaNs(res, lhs, rhs, set, status);   /* propagate NaN  */
      break;
      }
    /* have numbers  */
    if (op==DECNUMBER_COMPMAXMAG || op==DECNUMBER_COMPMINMAG) result=decCompare(lhs, rhs, 1);
     else result=decCompare(lhs, rhs, 0);    /* sign matters  */
    } while(0);                              /* end protected  */

  if (result==DECNUMBER_BADINT) *status|=DEC_Insufficient_storage; /* rare  */
   else {
    if (op==DECNUMBER_COMPARE || op==DECNUMBER_COMPSIG ||op==DECNUMBER_COMPTOTAL) { /* returning signum  */
      if (op==DECNUMBER_COMPTOTAL && result==0) {
        /* operands are numerically equal or same NaN (and same sign,  */
        /* tested first); if identical, leave result 0  */
        if (lhs->exponent!=rhs->exponent) {
          if (lhs->exponent<rhs->exponent) result=-1;
           else result=+1;
          if (decNumberIsNegative(lhs)) result=-result;
          } /* lexp!=rexp  */
        } /* total-order by exponent  */
      uprv_decNumberZero(res);               /* [always a valid result]  */
      if (result!=0) {                  /* must be -1 or +1  */
        *res->lsu=1;
        if (result<0) res->bits=DECNEG;
        }
      }
     else if (op==DECNUMBER_COMPNAN);             /* special, drop through  */
     else {                             /* MAX or MIN, non-NaN result  */
      Int residue=0;                    /* rounding accumulator  */
      /* choose the operand for the result  */
      const decNumber *choice;
      if (result==0) { /* operands are numerically equal  */
        /* choose according to sign then exponent (see 754)  */
        uByte slhs=(lhs->bits & DECNEG);
        uByte srhs=(rhs->bits & DECNEG);
        #if DECSUBSET
        if (!set->extended) {           /* subset: force left-hand  */
          op=DECNUMBER_COMPMAX;
          result=+1;
          }
        else
        #endif
        if (slhs!=srhs) {          /* signs differ  */
          if (slhs) result=-1;     /* rhs is max  */
               else result=+1;     /* lhs is max  */
          }
         else if (slhs && srhs) {  /* both negative  */
          if (lhs->exponent<rhs->exponent) result=+1;
                                      else result=-1;
          /* [if equal, use lhs, technically identical]  */
          }
         else {                    /* both positive  */
          if (lhs->exponent>rhs->exponent) result=+1;
                                      else result=-1;
          /* [ditto]  */
          }
        } /* numerically equal  */
      /* here result will be non-0; reverse if looking for MIN  */
      if (op==DECNUMBER_COMPMIN || op==DECNUMBER_COMPMINMAG) result=-result;
      choice=(result>0 ? lhs : rhs);    /* choose  */
      /* copy chosen to result, rounding if need be  */
      decCopyFit(res, choice, set, &residue, status);
      decFinish(res, set, &residue, status);
      }
    }
  #if DECSUBSET
  if (allocrhs!=NULL) free(allocrhs);   /* free any storage used  */
  if (alloclhs!=NULL) free(alloclhs);   /* ..  */
  #endif
  return res;
  } /* decCompareOp  */

/* ------------------------------------------------------------------ */
/* decCompare -- compare two decNumbers by numerical value            */
/*                                                                    */
/*  This routine compares A ? B without altering them.                */
/*                                                                    */
/*  Arg1 is A, a decNumber which is not a NaN                         */
/*  Arg2 is B, a decNumber which is not a NaN                         */
/*  Arg3 is 1 for a sign-independent compare, 0 otherwise             */
/*                                                                    */
/*  returns -1, 0, or 1 for A<B, A==B, or A>B, or DECNUMBER_BADINT if failure   */
/*  (the only possible failure is an allocation error)                */
/* ------------------------------------------------------------------ */
static Int decCompare(const decNumber *lhs, const decNumber *rhs,
                      Flag abs_c) {
  Int   result;                    /* result value  */
  Int   sigr;                      /* rhs signum  */
  Int   compare;                   /* work  */

  result=1;                                  /* assume signum(lhs)  */
  if (ISZERO(lhs)) result=0;
  if (abs_c) {
    if (ISZERO(rhs)) return result;          /* LHS wins or both 0  */
    /* RHS is non-zero  */
    if (result==0) return -1;                /* LHS is 0; RHS wins  */
    /* [here, both non-zero, result=1]  */
    }
   else {                                    /* signs matter  */
    if (result && decNumberIsNegative(lhs)) result=-1;
    sigr=1;                                  /* compute signum(rhs)  */
    if (ISZERO(rhs)) sigr=0;
     else if (decNumberIsNegative(rhs)) sigr=-1;
    if (result > sigr) return +1;            /* L > R, return 1  */
    if (result < sigr) return -1;            /* L < R, return -1  */
    if (result==0) return 0;                   /* both 0  */
    }

  /* signums are the same; both are non-zero  */
  if ((lhs->bits | rhs->bits) & DECINF) {    /* one or more infinities  */
    if (decNumberIsInfinite(rhs)) {
      if (decNumberIsInfinite(lhs)) result=0;/* both infinite  */
       else result=-result;                  /* only rhs infinite  */
      }
    return result;
    }
  /* must compare the coefficients, allowing for exponents  */
  if (lhs->exponent>rhs->exponent) {         /* LHS exponent larger  */
    /* swap sides, and sign  */
    const decNumber *temp=lhs;
    lhs=rhs;
    rhs=temp;
    result=-result;
    }
  compare=decUnitCompare(lhs->lsu, D2U(lhs->digits),
                         rhs->lsu, D2U(rhs->digits),
                         rhs->exponent-lhs->exponent);
  if (compare!=DECNUMBER_BADINT) compare*=result;      /* comparison succeeded  */
  return compare;
  } /* decCompare  */

/* ------------------------------------------------------------------ */
/* decUnitCompare -- compare two >=0 integers in Unit arrays          */
/*                                                                    */
/*  This routine compares A ? B*10**E where A and B are unit arrays   */
/*  A is a plain integer                                              */
/*  B has an exponent of E (which must be non-negative)               */
/*                                                                    */
/*  Arg1 is A first Unit (lsu)                                        */
/*  Arg2 is A length in Units                                         */
/*  Arg3 is B first Unit (lsu)                                        */
/*  Arg4 is B length in Units                                         */
/*  Arg5 is E (0 if the units are aligned)                            */
/*                                                                    */
/*  returns -1, 0, or 1 for A<B, A==B, or A>B, or DECNUMBER_BADINT if failure   */
/*  (the only possible failure is an allocation error, which can      */
/*  only occur if E!=0)                                               */
/* ------------------------------------------------------------------ */
static Int decUnitCompare(const Unit *a, Int alength,
                          const Unit *b, Int blength, Int exp) {
  Unit  *acc;                      /* accumulator for result  */
  Unit  accbuff[SD2U(DECBUFFER*2+1)]; /* local buffer  */
  Unit  *allocacc=NULL;            /* -> allocated acc buffer, iff allocated  */
  Int   accunits, need;            /* units in use or needed for acc  */
  const Unit *l, *r, *u;           /* work  */
  Int   expunits, exprem, result;  /* ..  */

  if (exp==0) {                    /* aligned; fastpath  */
    if (alength>blength) return 1;
    if (alength<blength) return -1;
    /* same number of units in both -- need unit-by-unit compare  */
    l=a+alength-1;
    r=b+alength-1;
    for (;l>=a; l--, r--) {
      if (*l>*r) return 1;
      if (*l<*r) return -1;
      }
    return 0;                      /* all units match  */
    } /* aligned  */

  /* Unaligned.  If one is >1 unit longer than the other, padded  */
  /* approximately, then can return easily  */
  if (alength>blength+(Int)D2U(exp)) return 1;
  if (alength+1<blength+(Int)D2U(exp)) return -1;

  /* Need to do a real subtract.  For this, a result buffer is needed  */
  /* even though only the sign is of interest.  Its length needs  */
  /* to be the larger of alength and padded blength, +2  */
  need=blength+D2U(exp);                /* maximum real length of B  */
  if (need<alength) need=alength;
  need+=2;
  acc=accbuff;                          /* assume use local buffer  */
  if (need*sizeof(Unit)>sizeof(accbuff)) {
    allocacc=(Unit *)malloc(need*sizeof(Unit));
    if (allocacc==NULL) return DECNUMBER_BADINT;  /* hopeless -- abandon  */
    acc=allocacc;
    }
  /* Calculate units and remainder from exponent.  */
  expunits=exp/DECDPUN;
  exprem=exp%DECDPUN;
  /* subtract [A+B*(-m)]  */
  accunits=decUnitAddSub(a, alength, b, blength, expunits, acc,
                         -(Int)DECNUMBER_powers[exprem]);
  /* [UnitAddSub result may have leading zeros, even on zero]  */
  if (accunits<0) result=-1;            /* negative result  */
   else {                               /* non-negative result  */
    /* check units of the result before freeing any storage  */
    for (u=acc; u<acc+accunits-1 && *u==0;) u++;
    result=(*u==0 ? 0 : +1);
    }
  /* clean up and return the result  */
  if (allocacc!=NULL) free(allocacc);   /* drop any storage used  */
  return result;
  } /* decUnitCompare  */

/* ------------------------------------------------------------------ */
/* decUnitAddSub -- add or subtract two >=0 integers in Unit arrays   */
/*                                                                    */
/*  This routine performs the calculation:                            */
/*                                                                    */
/*  C=A+(B*M)                                                         */
/*                                                                    */
/*  Where M is in the range -DECDPUNMAX through +DECDPUNMAX.          */
/*                                                                    */
/*  A may be shorter or longer than B.                                */
/*                                                                    */
/*  Leading zeros are not removed after a calculation.  The result is */
/*  either the same length as the longer of A and B (adding any       */
/*  shift), or one Unit longer than that (if a Unit carry occurred).  */
/*                                                                    */
/*  A and B content are not altered unless C is also A or B.          */
/*  C may be the same array as A or B, but only if no zero padding is */
/*  requested (that is, C may be B only if bshift==0).                */
/*  C is filled from the lsu; only those units necessary to complete  */
/*  the calculation are referenced.                                   */
/*                                                                    */
/*  Arg1 is A first Unit (lsu)                                        */
/*  Arg2 is A length in Units                                         */
/*  Arg3 is B first Unit (lsu)                                        */
/*  Arg4 is B length in Units                                         */
/*  Arg5 is B shift in Units  (>=0; pads with 0 units if positive)    */
/*  Arg6 is C first Unit (lsu)                                        */
/*  Arg7 is M, the multiplier                                         */
/*                                                                    */
/*  returns the count of Units written to C, which will be non-zero   */
/*  and negated if the result is negative.  That is, the sign of the  */
/*  returned Int is the sign of the result (positive for zero) and    */
/*  the absolute value of the Int is the count of Units.              */
/*                                                                    */
/*  It is the caller's responsibility to make sure that C size is     */
/*  safe, allowing space if necessary for a one-Unit carry.           */
/*                                                                    */
/*  This routine is severely performance-critical; *any* change here  */
/*  must be measured (timed) to assure no performance degradation.    */
/*  In particular, trickery here tends to be counter-productive, as   */
/*  increased complexity of code hurts register optimizations on      */
/*  register-poor architectures.  Avoiding divisions is nearly        */
/*  always a Good Idea, however.                                      */
/*                                                                    */
/* Special thanks to Rick McGuire (IBM Cambridge, MA) and Dave Clark  */
/* (IBM Warwick, UK) for some of the ideas used in this routine.      */
/* ------------------------------------------------------------------ */
static Int decUnitAddSub(const Unit *a, Int alength,
                         const Unit *b, Int blength, Int bshift,
                         Unit *c, Int m) {
  const Unit *alsu=a;              /* A lsu [need to remember it]  */
  Unit *clsu=c;                    /* C ditto  */
  Unit *minC;                      /* low water mark for C  */
  Unit *maxC;                      /* high water mark for C  */
  eInt carry=0;                    /* carry integer (could be Long)  */
  Int  add;                        /* work  */
  #if DECDPUN<=4                   /* myriadal, millenary, etc.  */
  Int  est;                        /* estimated quotient  */
  #endif

  #if DECTRACE
  if (alength<1 || blength<1)
    printf("decUnitAddSub: alen blen m %ld %ld [%ld]\n", alength, blength, m);
  #endif

  maxC=c+alength;                  /* A is usually the longer  */
  minC=c+blength;                  /* .. and B the shorter  */
  if (bshift!=0) {                 /* B is shifted; low As copy across  */
    minC+=bshift;
    /* if in place [common], skip copy unless there's a gap [rare]  */
    if (a==c && bshift<=alength) {
      c+=bshift;
      a+=bshift;
      }
     else for (; c<clsu+bshift; a++, c++) {  /* copy needed  */
      if (a<alsu+alength) *c=*a;
       else *c=0;
      }
    }
  if (minC>maxC) { /* swap  */
    Unit *hold=minC;
    minC=maxC;
    maxC=hold;
    }

  /* For speed, do the addition as two loops; the first where both A  */
  /* and B contribute, and the second (if necessary) where only one or  */
  /* other of the numbers contribute.  */
  /* Carry handling is the same (i.e., duplicated) in each case.  */
  for (; c<minC; c++) {
    carry+=*a;
    a++;
    carry+=((eInt)*b)*m;                /* [special-casing m=1/-1  */
    b++;                                /* here is not a win]  */
    /* here carry is new Unit of digits; it could be +ve or -ve  */
    if ((ueInt)carry<=DECDPUNMAX) {     /* fastpath 0-DECDPUNMAX  */
      *c=(Unit)carry;
      carry=0;
      continue;
      }
    #if DECDPUN==4                           /* use divide-by-multiply  */
      if (carry>=0) {
        est=(((ueInt)carry>>11)*53687)>>18;
        *c=(Unit)(carry-est*(DECDPUNMAX+1)); /* remainder  */
        carry=est;                           /* likely quotient [89%]  */
        if (*c<DECDPUNMAX+1) continue;       /* estimate was correct  */
        carry++;
        *c-=DECDPUNMAX+1;
        continue;
        }
      /* negative case  */
      carry=carry+(eInt)(DECDPUNMAX+1)*(DECDPUNMAX+1); /* make positive  */
      est=(((ueInt)carry>>11)*53687)>>18;
      *c=(Unit)(carry-est*(DECDPUNMAX+1));
      carry=est-(DECDPUNMAX+1);              /* correctly negative  */
      if (*c<DECDPUNMAX+1) continue;         /* was OK  */
      carry++;
      *c-=DECDPUNMAX+1;
    #elif DECDPUN==3
      if (carry>=0) {
        est=(((ueInt)carry>>3)*16777)>>21;
        *c=(Unit)(carry-est*(DECDPUNMAX+1)); /* remainder  */
        carry=est;                           /* likely quotient [99%]  */
        if (*c<DECDPUNMAX+1) continue;       /* estimate was correct  */
        carry++;
        *c-=DECDPUNMAX+1;
        continue;
        }
      /* negative case  */
      carry=carry+(eInt)(DECDPUNMAX+1)*(DECDPUNMAX+1); /* make positive  */
      est=(((ueInt)carry>>3)*16777)>>21;
      *c=(Unit)(carry-est*(DECDPUNMAX+1));
      carry=est-(DECDPUNMAX+1);              /* correctly negative  */
      if (*c<DECDPUNMAX+1) continue;         /* was OK  */
      carry++;
      *c-=DECDPUNMAX+1;
    #elif DECDPUN<=2
      /* Can use QUOT10 as carry <= 4 digits  */
      if (carry>=0) {
        est=QUOT10(carry, DECDPUN);
        *c=(Unit)(carry-est*(DECDPUNMAX+1)); /* remainder  */
        carry=est;                           /* quotient  */
        continue;
        }
      /* negative case  */
      carry=carry+(eInt)(DECDPUNMAX+1)*(DECDPUNMAX+1); /* make positive  */
      est=QUOT10(carry, DECDPUN);
      *c=(Unit)(carry-est*(DECDPUNMAX+1));
      carry=est-(DECDPUNMAX+1);              /* correctly negative  */
    #else
      /* remainder operator is undefined if negative, so must test  */
      if ((ueInt)carry<(DECDPUNMAX+1)*2) {   /* fastpath carry +1  */
        *c=(Unit)(carry-(DECDPUNMAX+1));     /* [helps additions]  */
        carry=1;
        continue;
        }
      if (carry>=0) {
        *c=(Unit)(carry%(DECDPUNMAX+1));
        carry=carry/(DECDPUNMAX+1);
        continue;
        }
      /* negative case  */
      carry=carry+(eInt)(DECDPUNMAX+1)*(DECDPUNMAX+1); /* make positive  */
      *c=(Unit)(carry%(DECDPUNMAX+1));
      carry=carry/(DECDPUNMAX+1)-(DECDPUNMAX+1);
    #endif
    } /* c  */

  /* now may have one or other to complete  */
  /* [pretest to avoid loop setup/shutdown]  */
  if (c<maxC) for (; c<maxC; c++) {
    if (a<alsu+alength) {               /* still in A  */
      carry+=*a;
      a++;
      }
     else {                             /* inside B  */
      carry+=((eInt)*b)*m;
      b++;
      }
    /* here carry is new Unit of digits; it could be +ve or -ve and  */
    /* magnitude up to DECDPUNMAX squared  */
    if ((ueInt)carry<=DECDPUNMAX) {     /* fastpath 0-DECDPUNMAX  */
      *c=(Unit)carry;
      carry=0;
      continue;
      }
    /* result for this unit is negative or >DECDPUNMAX  */
    #if DECDPUN==4                           /* use divide-by-multiply  */
      if (carry>=0) {
        est=(((ueInt)carry>>11)*53687)>>18;
        *c=(Unit)(carry-est*(DECDPUNMAX+1)); /* remainder  */
        carry=est;                           /* likely quotient [79.7%]  */
        if (*c<DECDPUNMAX+1) continue;       /* estimate was correct  */
        carry++;
        *c-=DECDPUNMAX+1;
        continue;
        }
      /* negative case  */
      carry=carry+(eInt)(DECDPUNMAX+1)*(DECDPUNMAX+1); /* make positive  */
      est=(((ueInt)carry>>11)*53687)>>18;
      *c=(Unit)(carry-est*(DECDPUNMAX+1));
      carry=est-(DECDPUNMAX+1);              /* correctly negative  */
      if (*c<DECDPUNMAX+1) continue;         /* was OK  */
      carry++;
      *c-=DECDPUNMAX+1;
    #elif DECDPUN==3
      if (carry>=0) {
        est=(((ueInt)carry>>3)*16777)>>21;
        *c=(Unit)(carry-est*(DECDPUNMAX+1)); /* remainder  */
        carry=est;                           /* likely quotient [99%]  */
        if (*c<DECDPUNMAX+1) continue;       /* estimate was correct  */
        carry++;
        *c-=DECDPUNMAX+1;
        continue;
        }
      /* negative case  */
      carry=carry+(eInt)(DECDPUNMAX+1)*(DECDPUNMAX+1); /* make positive  */
      est=(((ueInt)carry>>3)*16777)>>21;
      *c=(Unit)(carry-est*(DECDPUNMAX+1));
      carry=est-(DECDPUNMAX+1);              /* correctly negative  */
      if (*c<DECDPUNMAX+1) continue;         /* was OK  */
      carry++;
      *c-=DECDPUNMAX+1;
    #elif DECDPUN<=2
      if (carry>=0) {
        est=QUOT10(carry, DECDPUN);
        *c=(Unit)(carry-est*(DECDPUNMAX+1)); /* remainder  */
        carry=est;                           /* quotient  */
        continue;
        }
      /* negative case  */
      carry=carry+(eInt)(DECDPUNMAX+1)*(DECDPUNMAX+1); /* make positive  */
      est=QUOT10(carry, DECDPUN);
      *c=(Unit)(carry-est*(DECDPUNMAX+1));
      carry=est-(DECDPUNMAX+1);              /* correctly negative  */
    #else
      if ((ueInt)carry<(DECDPUNMAX+1)*2){    /* fastpath carry 1  */
        *c=(Unit)(carry-(DECDPUNMAX+1));
        carry=1;
        continue;
        }
      /* remainder operator is undefined if negative, so must test  */
      if (carry>=0) {
        *c=(Unit)(carry%(DECDPUNMAX+1));
        carry=carry/(DECDPUNMAX+1);
        continue;
        }
      /* negative case  */
      carry=carry+(eInt)(DECDPUNMAX+1)*(DECDPUNMAX+1); /* make positive  */
      *c=(Unit)(carry%(DECDPUNMAX+1));
      carry=carry/(DECDPUNMAX+1)-(DECDPUNMAX+1);
    #endif
    } /* c  */

  /* OK, all A and B processed; might still have carry or borrow  */
  /* return number of Units in the result, negated if a borrow  */
  if (carry==0) return static_cast<int32_t>(c-clsu);     /* no carry, so no more to do  */
  if (carry>0) {                   /* positive carry  */
    *c=(Unit)carry;                /* place as new unit  */
    c++;                           /* ..  */
    return static_cast<int32_t>(c-clsu);
    }
  /* -ve carry: it's a borrow; complement needed  */
  add=1;                           /* temporary carry...  */
  for (c=clsu; c<maxC; c++) {
    add=DECDPUNMAX+add-*c;
    if (add<=DECDPUNMAX) {
      *c=(Unit)add;
      add=0;
      }
     else {
      *c=0;
      add=1;
      }
    }
  /* add an extra unit iff it would be non-zero  */
  #if DECTRACE
    printf("UAS borrow: add %ld, carry %ld\n", add, carry);
  #endif
  if ((add-carry-1)!=0) {
    *c=(Unit)(add-carry-1);
    c++;                      /* interesting, include it  */
    }
  return static_cast<int32_t>(clsu-c);              /* -ve result indicates borrowed  */
  } /* decUnitAddSub  */

/* ------------------------------------------------------------------ */
/* decTrim -- trim trailing zeros or normalize                        */
/*                                                                    */
/*   dn is the number to trim or normalize                            */
/*   set is the context to use to check for clamp                     */
/*   all is 1 to remove all trailing zeros, 0 for just fraction ones  */
/*   noclamp is 1 to unconditional (unclamped) trim                   */
/*   dropped returns the number of discarded trailing zeros           */
/*   returns dn                                                       */
/*                                                                    */
/* If clamp is set in the context then the number of zeros trimmed    */
/* may be limited if the exponent is high.                            */
/* All fields are updated as required.  This is a utility operation,  */
/* so special values are unchanged and no error is possible.          */
/* ------------------------------------------------------------------ */
static decNumber * decTrim(decNumber *dn, decContext *set, Flag all,
                           Flag noclamp, Int *dropped) {
  Int   d, exp;                    /* work  */
  uInt  cut;                       /* ..  */
  Unit  *up;                       /* -> current Unit  */

  #if DECCHECK
  if (decCheckOperands(dn, DECUNUSED, DECUNUSED, DECUNCONT)) return dn;
  #endif

  *dropped=0;                           /* assume no zeros dropped  */
  if ((dn->bits & DECSPECIAL)           /* fast exit if special ..  */
    || (*dn->lsu & 0x01)) return dn;    /* .. or odd  */
  if (ISZERO(dn)) {                     /* .. or 0  */
    dn->exponent=0;                     /* (sign is preserved)  */
    return dn;
    }

  /* have a finite number which is even  */
  exp=dn->exponent;
  cut=1;                           /* digit (1-DECDPUN) in Unit  */
  up=dn->lsu;                      /* -> current Unit  */
  for (d=0; d<dn->digits-1; d++) { /* [don't strip the final digit]  */
    /* slice by DECNUMBER_powers  */
    #if DECDPUN<=4
      uInt quot=QUOT10(*up, cut);
      if ((*up-quot*DECNUMBER_powers[cut])!=0) break;  /* found non-0 digit  */
    #else
      if (*up%DECNUMBER_powers[cut]!=0) break;         /* found non-0 digit  */
    #endif
    /* have a trailing 0  */
    if (!all) {                    /* trimming  */
      /* [if exp>0 then all trailing 0s are significant for trim]  */
      if (exp<=0) {                /* if digit might be significant  */
        if (exp==0) break;         /* then quit  */
        exp++;                     /* next digit might be significant  */
        }
      }
    cut++;                         /* next power  */
    if (cut>DECDPUN) {             /* need new Unit  */
      up++;
      cut=1;
      }
    } /* d  */
  if (d==0) return dn;             /* none to drop  */

  /* may need to limit drop if clamping  */
  if (set->clamp && !noclamp) {
    Int maxd=set->emax-set->digits+1-dn->exponent;
    if (maxd<=0) return dn;        /* nothing possible  */
    if (d>maxd) d=maxd;
    }

  /* effect the drop  */
  decShiftToLeast(dn->lsu, D2U(dn->digits), d);
  dn->exponent+=d;                 /* maintain numerical value  */
  dn->digits-=d;                   /* new length  */
  *dropped=d;                      /* report the count  */
  return dn;
  } /* decTrim  */

/* ------------------------------------------------------------------ */
/* decReverse -- reverse a Unit array in place                        */
/*                                                                    */
/*   ulo    is the start of the array                                 */
/*   uhi    is the end of the array (highest Unit to include)         */
/*                                                                    */
/* The units ulo through uhi are reversed in place (if the number     */
/* of units is odd, the middle one is untouched).  Note that the      */
/* digit(s) in each unit are unaffected.                              */
/* ------------------------------------------------------------------ */
static void decReverse(Unit *ulo, Unit *uhi) {
  Unit temp;
  for (; ulo<uhi; ulo++, uhi--) {
    temp=*ulo;
    *ulo=*uhi;
    *uhi=temp;
    }
  return;
  } /* decReverse  */

/* ------------------------------------------------------------------ */
/* decShiftToMost -- shift digits in array towards most significant   */
/*                                                                    */
/*   uar    is the array                                              */
/*   digits is the count of digits in use in the array                */
/*   shift  is the number of zeros to pad with (least significant);   */
/*     it must be zero or positive                                    */
/*                                                                    */
/*   returns the new length of the integer in the array, in digits    */
/*                                                                    */
/* No overflow is permitted (that is, the uar array must be known to  */
/* be large enough to hold the result, after shifting).               */
/* ------------------------------------------------------------------ */
static Int decShiftToMost(Unit *uar, Int digits, Int shift) {
  Unit  *target, *source, *first;  /* work  */
  Int   cut;                       /* odd 0's to add  */
  uInt  next;                      /* work  */

  if (shift==0) return digits;     /* [fastpath] nothing to do  */
  if ((digits+shift)<=DECDPUN) {   /* [fastpath] single-unit case  */
    *uar=(Unit)(*uar*DECNUMBER_powers[shift]);
    return digits+shift;
    }

  next=0;                          /* all paths  */
  source=uar+D2U(digits)-1;        /* where msu comes from  */
  target=source+D2U(shift);        /* where upper part of first cut goes  */
  cut=DECDPUN-MSUDIGITS(shift);    /* where to slice  */
  if (cut==0) {                    /* unit-boundary case  */
    for (; source>=uar; source--, target--) *target=*source;
    }
   else {
    first=uar+D2U(digits+shift)-1; /* where msu of source will end up  */
    for (; source>=uar; source--, target--) {
      /* split the source Unit and accumulate remainder for next  */
      #if DECDPUN<=4
        uInt quot=QUOT10(*source, cut);
        uInt rem=*source-quot*DECNUMBER_powers[cut];
        next+=quot;
      #else
        uInt rem=*source%DECNUMBER_powers[cut];
        next+=*source/DECNUMBER_powers[cut];
      #endif
      if (target<=first) *target=(Unit)next;   /* write to target iff valid  */
      next=rem*DECNUMBER_powers[DECDPUN-cut];            /* save remainder for next Unit  */
      }
    } /* shift-move  */

  /* propagate any partial unit to one below and clear the rest  */
  for (; target>=uar; target--) {
    *target=(Unit)next;
    next=0;
    }
  return digits+shift;
  } /* decShiftToMost  */

/* ------------------------------------------------------------------ */
/* decShiftToLeast -- shift digits in array towards least significant */
/*                                                                    */
/*   uar   is the array                                               */
/*   units is length of the array, in units                           */
/*   shift is the number of digits to remove from the lsu end; it     */
/*     must be zero or positive and <= than units*DECDPUN.            */
/*                                                                    */
/*   returns the new length of the integer in the array, in units     */
/*                                                                    */
/* Removed digits are discarded (lost).  Units not required to hold   */
/* the final result are unchanged.                                    */
/* ------------------------------------------------------------------ */
static Int decShiftToLeast(Unit *uar, Int units, Int shift) {
  Unit  *target, *up;              /* work  */
  Int   cut, count;                /* work  */
  Int   quot, rem;                 /* for division  */

  if (shift==0) return units;      /* [fastpath] nothing to do  */
  if (shift==units*DECDPUN) {      /* [fastpath] little to do  */
    *uar=0;                        /* all digits cleared gives zero  */
    return 1;                      /* leaves just the one  */
    }

  target=uar;                      /* both paths  */
  cut=MSUDIGITS(shift);
  if (cut==DECDPUN) {              /* unit-boundary case; easy  */
    up=uar+D2U(shift);
    for (; up<uar+units; target++, up++) *target=*up;
    return static_cast<int32_t>(target-uar);
    }

  /* messier  */
  up=uar+D2U(shift-cut);           /* source; correct to whole Units  */
  count=units*DECDPUN-shift;       /* the maximum new length  */
  #if DECDPUN<=4
    quot=QUOT10(*up, cut);
  #else
    quot=*up/DECNUMBER_powers[cut];
  #endif
  for (; ; target++) {
    *target=(Unit)quot;
    count-=(DECDPUN-cut);
    if (count<=0) break;
    up++;
    quot=*up;
    #if DECDPUN<=4
      quot=QUOT10(quot, cut);
      rem=*up-quot*DECNUMBER_powers[cut];
    #else
      rem=quot%DECNUMBER_powers[cut];
      quot=quot/DECNUMBER_powers[cut];
    #endif
    *target=(Unit)(*target+rem*DECNUMBER_powers[DECDPUN-cut]);
    count-=cut;
    if (count<=0) break;
    }
  return static_cast<int32_t>(target-uar+1);
  } /* decShiftToLeast  */

#if DECSUBSET
/* ------------------------------------------------------------------ */
/* decRoundOperand -- round an operand  [used for subset only]        */
/*                                                                    */
/*   dn is the number to round (dn->digits is > set->digits)          */
/*   set is the relevant context                                      */
/*   status is the status accumulator                                 */
/*                                                                    */
/*   returns an allocated decNumber with the rounded result.          */
/*                                                                    */
/* lostDigits and other status may be set by this.                    */
/*                                                                    */
/* Since the input is an operand, it must not be modified.            */
/* Instead, return an allocated decNumber, rounded as required.       */
/* It is the caller's responsibility to free the allocated storage.   */
/*                                                                    */
/* If no storage is available then the result cannot be used, so NULL */
/* is returned.                                                       */
/* ------------------------------------------------------------------ */
static decNumber *decRoundOperand(const decNumber *dn, decContext *set,
                                  uInt *status) {
  decNumber *res;                       /* result structure  */
  uInt newstatus=0;                     /* status from round  */
  Int  residue=0;                       /* rounding accumulator  */

  /* Allocate storage for the returned decNumber, big enough for the  */
  /* length specified by the context  */
  res=(decNumber *)malloc(sizeof(decNumber)
                          +(D2U(set->digits)-1)*sizeof(Unit));
  if (res==NULL) {
    *status|=DEC_Insufficient_storage;
    return NULL;
    }
  decCopyFit(res, dn, set, &residue, &newstatus);
  decApplyRound(res, set, residue, &newstatus);

  /* If that set Inexact then "lost digits" is raised...  */
  if (newstatus & DEC_Inexact) newstatus|=DEC_Lost_digits;
  *status|=newstatus;
  return res;
  } /* decRoundOperand  */
#endif

/* ------------------------------------------------------------------ */
/* decCopyFit -- copy a number, truncating the coefficient if needed  */
/*                                                                    */
/*   dest is the target decNumber                                     */
/*   src  is the source decNumber                                     */
/*   set is the context [used for length (digits) and rounding mode]  */
/*   residue is the residue accumulator                               */
/*   status contains the current status to be updated                 */
/*                                                                    */
/* (dest==src is allowed and will be a no-op if fits)                 */
/* All fields are updated as required.                                */
/* ------------------------------------------------------------------ */
static void decCopyFit(decNumber *dest, const decNumber *src,
                       decContext *set, Int *residue, uInt *status) {
  dest->bits=src->bits;
  dest->exponent=src->exponent;
  decSetCoeff(dest, set, src->lsu, src->digits, residue, status);
  } /* decCopyFit  */

/* ------------------------------------------------------------------ */
/* decSetCoeff -- set the coefficient of a number                     */
/*                                                                    */
/*   dn    is the number whose coefficient array is to be set.        */
/*         It must have space for set->digits digits                  */
/*   set   is the context [for size]                                  */
/*   lsu   -> lsu of the source coefficient [may be dn->lsu]          */
/*   len   is digits in the source coefficient [may be dn->digits]    */
/*   residue is the residue accumulator.  This has values as in       */
/*         decApplyRound, and will be unchanged unless the            */
/*         target size is less than len.  In this case, the           */
/*         coefficient is truncated and the residue is updated to     */
/*         reflect the previous residue and the dropped digits.       */
/*   status is the status accumulator, as usual                       */
/*                                                                    */
/* The coefficient may already be in the number, or it can be an      */
/* external intermediate array.  If it is in the number, lsu must ==  */
/* dn->lsu and len must == dn->digits.                                */
/*                                                                    */
/* Note that the coefficient length (len) may be < set->digits, and   */
/* in this case this merely copies the coefficient (or is a no-op     */
/* if dn->lsu==lsu).                                                  */
/*                                                                    */
/* Note also that (only internally, from decQuantizeOp and            */
/* decSetSubnormal) the value of set->digits may be less than one,    */
/* indicating a round to left.  This routine handles that case        */
/* correctly; caller ensures space.                                   */
/*                                                                    */
/* dn->digits, dn->lsu (and as required), and dn->exponent are        */
/* updated as necessary.   dn->bits (sign) is unchanged.              */
/*                                                                    */
/* DEC_Rounded status is set if any digits are discarded.             */
/* DEC_Inexact status is set if any non-zero digits are discarded, or */
/*                       incoming residue was non-0 (implies rounded) */
/* ------------------------------------------------------------------ */
/* mapping array: maps 0-9 to canonical residues, so that a residue  */
/* can be adjusted in the range [-1, +1] and achieve correct rounding  */
/*                             0  1  2  3  4  5  6  7  8  9  */
static const uByte resmap[10]={0, 3, 3, 3, 3, 5, 7, 7, 7, 7};
static void decSetCoeff(decNumber *dn, decContext *set, const Unit *lsu,
                        Int len, Int *residue, uInt *status) {
  Int   discard;              /* number of digits to discard  */
  uInt  cut;                  /* cut point in Unit  */
  const Unit *up;             /* work  */
  Unit  *target;              /* ..  */
  Int   count;                /* ..  */
  #if DECDPUN<=4
  uInt  temp;                 /* ..  */
  #endif

  discard=len-set->digits;    /* digits to discard  */
  if (discard<=0) {           /* no digits are being discarded  */
    if (dn->lsu!=lsu) {       /* copy needed  */
      /* copy the coefficient array to the result number; no shift needed  */
      count=len;              /* avoids D2U  */
      up=lsu;
      for (target=dn->lsu; count>0; target++, up++, count-=DECDPUN)
        *target=*up;
      dn->digits=len;         /* set the new length  */
      }
    /* dn->exponent and residue are unchanged, record any inexactitude  */
    if (*residue!=0) *status|=(DEC_Inexact | DEC_Rounded);
    return;
    }

  /* some digits must be discarded ...  */
  dn->exponent+=discard;      /* maintain numerical value  */
  *status|=DEC_Rounded;       /* accumulate Rounded status  */
  if (*residue>1) *residue=1; /* previous residue now to right, so reduce  */

  if (discard>len) {          /* everything, +1, is being discarded  */
    /* guard digit is 0  */
    /* residue is all the number [NB could be all 0s]  */
    if (*residue<=0) {        /* not already positive  */
      count=len;              /* avoids D2U  */
      for (up=lsu; count>0; up++, count-=DECDPUN) if (*up!=0) { /* found non-0  */
        *residue=1;
        break;                /* no need to check any others  */
        }
      }
    if (*residue!=0) *status|=DEC_Inexact; /* record inexactitude  */
    *dn->lsu=0;               /* coefficient will now be 0  */
    dn->digits=1;             /* ..  */
    return;
    } /* total discard  */

  /* partial discard [most common case]  */
  /* here, at least the first (most significant) discarded digit exists  */

  /* spin up the number, noting residue during the spin, until get to  */
  /* the Unit with the first discarded digit.  When reach it, extract  */
  /* it and remember its position  */
  count=0;
  for (up=lsu;; up++) {
    count+=DECDPUN;
    if (count>=discard) break; /* full ones all checked  */
    if (*up!=0) *residue=1;
    } /* up  */

  /* here up -> Unit with first discarded digit  */
  cut=discard-(count-DECDPUN)-1;
  if (cut==DECDPUN-1) {       /* unit-boundary case (fast)  */
    Unit half=(Unit)DECNUMBER_powers[DECDPUN]>>1;
    /* set residue directly  */
    if (*up>=half) {
      if (*up>half) *residue=7;
      else *residue+=5;       /* add sticky bit  */
      }
     else { /* <half  */
      if (*up!=0) *residue=3; /* [else is 0, leave as sticky bit]  */
      }
    if (set->digits<=0) {     /* special for Quantize/Subnormal :-(  */
      *dn->lsu=0;             /* .. result is 0  */
      dn->digits=1;           /* ..  */
      }
     else {                   /* shift to least  */
      count=set->digits;      /* now digits to end up with  */
      dn->digits=count;       /* set the new length  */
      up++;                   /* move to next  */
      /* on unit boundary, so shift-down copy loop is simple  */
      for (target=dn->lsu; count>0; target++, up++, count-=DECDPUN)
        *target=*up;
      }
    } /* unit-boundary case  */

   else { /* discard digit is in low digit(s), and not top digit  */
    uInt  discard1;                /* first discarded digit  */
    uInt  quot, rem;               /* for divisions  */
    if (cut==0) quot=*up;          /* is at bottom of unit  */
     else /* cut>0 */ {            /* it's not at bottom of unit  */
      #if DECDPUN<=4
        U_ASSERT(/* cut >= 0 &&*/ cut <= 4);
        quot=QUOT10(*up, cut);
        rem=*up-quot*DECNUMBER_powers[cut];
      #else
        rem=*up%DECNUMBER_powers[cut];
        quot=*up/DECNUMBER_powers[cut];
      #endif
      if (rem!=0) *residue=1;
      }
    /* discard digit is now at bottom of quot  */
    #if DECDPUN<=4
      temp=(quot*6554)>>16;        /* fast /10  */
      /* Vowels algorithm here not a win (9 instructions)  */
      discard1=quot-X10(temp);
      quot=temp;
    #else
      discard1=quot%10;
      quot=quot/10;
    #endif
    /* here, discard1 is the guard digit, and residue is everything  */
    /* else [use mapping array to accumulate residue safely]  */
    *residue+=resmap[discard1];
    cut++;                         /* update cut  */
    /* here: up -> Unit of the array with bottom digit  */
    /*       cut is the division point for each Unit  */
    /*       quot holds the uncut high-order digits for the current unit  */
    if (set->digits<=0) {          /* special for Quantize/Subnormal :-(  */
      *dn->lsu=0;                  /* .. result is 0  */
      dn->digits=1;                /* ..  */
      }
     else {                        /* shift to least needed  */
      count=set->digits;           /* now digits to end up with  */
      dn->digits=count;            /* set the new length  */
      /* shift-copy the coefficient array to the result number  */
      for (target=dn->lsu; ; target++) {
        *target=(Unit)quot;
        count-=(DECDPUN-cut);
        if (count<=0) break;
        up++;
        quot=*up;
        #if DECDPUN<=4
          quot=QUOT10(quot, cut);
          rem=*up-quot*DECNUMBER_powers[cut];
        #else
          rem=quot%DECNUMBER_powers[cut];
          quot=quot/DECNUMBER_powers[cut];
        #endif
        *target=(Unit)(*target+rem*DECNUMBER_powers[DECDPUN-cut]);
        count-=cut;
        if (count<=0) break;
        } /* shift-copy loop  */
      } /* shift to least  */
    } /* not unit boundary  */

  if (*residue!=0) *status|=DEC_Inexact; /* record inexactitude  */
  return;
  } /* decSetCoeff  */

/* ------------------------------------------------------------------ */
/* decApplyRound -- apply pending rounding to a number                */
/*                                                                    */
/*   dn    is the number, with space for set->digits digits           */
/*   set   is the context [for size and rounding mode]                */
/*   residue indicates pending rounding, being any accumulated        */
/*         guard and sticky information.  It may be:                  */
/*         6-9: rounding digit is >5                                  */
/*         5:   rounding digit is exactly half-way                    */
/*         1-4: rounding digit is <5 and >0                           */
/*         0:   the coefficient is exact                              */
/*        -1:   as 1, but the hidden digits are subtractive, that     */
/*              is, of the opposite sign to dn.  In this case the     */
/*              coefficient must be non-0.  This case occurs when     */
/*              subtracting a small number (which can be reduced to   */
/*              a sticky bit); see decAddOp.                          */
/*   status is the status accumulator, as usual                       */
/*                                                                    */
/* This routine applies rounding while keeping the length of the      */
/* coefficient constant.  The exponent and status are unchanged       */
/* except if:                                                         */
/*                                                                    */
/*   -- the coefficient was increased and is all nines (in which      */
/*      case Overflow could occur, and is handled directly here so    */
/*      the caller does not need to re-test for overflow)             */
/*                                                                    */
/*   -- the coefficient was decreased and becomes all nines (in which */
/*      case Underflow could occur, and is also handled directly).    */
/*                                                                    */
/* All fields in dn are updated as required.                          */
/*                                                                    */
/* ------------------------------------------------------------------ */
static void decApplyRound(decNumber *dn, decContext *set, Int residue,
                          uInt *status) {
  Int  bump;                  /* 1 if coefficient needs to be incremented  */
                              /* -1 if coefficient needs to be decremented  */

  if (residue==0) return;     /* nothing to apply  */

  bump=0;                     /* assume a smooth ride  */

  /* now decide whether, and how, to round, depending on mode  */
  switch (set->round) {
    case DEC_ROUND_05UP: {    /* round zero or five up (for reround)  */
      /* This is the same as DEC_ROUND_DOWN unless there is a  */
      /* positive residue and the lsd of dn is 0 or 5, in which case  */
      /* it is bumped; when residue is <0, the number is therefore  */
      /* bumped down unless the final digit was 1 or 6 (in which  */
      /* case it is bumped down and then up -- a no-op)  */
      Int lsd5=*dn->lsu%5;     /* get lsd and quintate  */
      if (residue<0 && lsd5!=1) bump=-1;
       else if (residue>0 && lsd5==0) bump=1;
      /* [bump==1 could be applied directly; use common path for clarity]  */
      break;} /* r-05  */

    case DEC_ROUND_DOWN: {
      /* no change, except if negative residue  */
      if (residue<0) bump=-1;
      break;} /* r-d  */

    case DEC_ROUND_HALF_DOWN: {
      if (residue>5) bump=1;
      break;} /* r-h-d  */

    case DEC_ROUND_HALF_EVEN: {
      if (residue>5) bump=1;            /* >0.5 goes up  */
       else if (residue==5) {           /* exactly 0.5000...  */
        /* 0.5 goes up iff [new] lsd is odd  */
        if (*dn->lsu & 0x01) bump=1;
        }
      break;} /* r-h-e  */

    case DEC_ROUND_HALF_UP: {
      if (residue>=5) bump=1;
      break;} /* r-h-u  */

    case DEC_ROUND_UP: {
      if (residue>0) bump=1;
      break;} /* r-u  */

    case DEC_ROUND_CEILING: {
      /* same as _UP for positive numbers, and as _DOWN for negatives  */
      /* [negative residue cannot occur on 0]  */
      if (decNumberIsNegative(dn)) {
        if (residue<0) bump=-1;
        }
       else {
        if (residue>0) bump=1;
        }
      break;} /* r-c  */

    case DEC_ROUND_FLOOR: {
      /* same as _UP for negative numbers, and as _DOWN for positive  */
      /* [negative residue cannot occur on 0]  */
      if (!decNumberIsNegative(dn)) {
        if (residue<0) bump=-1;
        }
       else {
        if (residue>0) bump=1;
        }
      break;} /* r-f  */

    default: {      /* e.g., DEC_ROUND_MAX  */
      *status|=DEC_Invalid_context;
      #if DECTRACE || (DECCHECK && DECNUMBER_DECVERB)
      printf("Unknown rounding mode: %d\n", set->round);
      #endif
      break;}
    } /* switch  */

  /* now bump the number, up or down, if need be  */
  if (bump==0) return;                       /* no action required  */

  /* Simply use decUnitAddSub unless bumping up and the number is  */
  /* all nines.  In this special case set to 100... explicitly  */
  /* and adjust the exponent by one (as otherwise could overflow  */
  /* the array)  */
  /* Similarly handle all-nines result if bumping down.  */
  if (bump>0) {
    Unit *up;                                /* work  */
    uInt count=dn->digits;                   /* digits to be checked  */
    for (up=dn->lsu; ; up++) {
      if (count<=DECDPUN) {
        /* this is the last Unit (the msu)  */
        if (*up!=DECNUMBER_powers[count]-1) break;     /* not still 9s  */
        /* here if it, too, is all nines  */
        *up=(Unit)DECNUMBER_powers[count-1];           /* here 999 -> 100 etc.  */
        for (up=up-1; up>=dn->lsu; up--) *up=0; /* others all to 0  */
        dn->exponent++;                      /* and bump exponent  */
        /* [which, very rarely, could cause Overflow...]  */
        if ((dn->exponent+dn->digits)>set->emax+1) {
          decSetOverflow(dn, set, status);
          }
        return;                              /* done  */
        }
      /* a full unit to check, with more to come  */
      if (*up!=DECDPUNMAX) break;            /* not still 9s  */
      count-=DECDPUN;
      } /* up  */
    } /* bump>0  */
   else {                                    /* -1  */
    /* here checking for a pre-bump of 1000... (leading 1, all  */
    /* other digits zero)  */
    Unit *up, *sup;                          /* work  */
    uInt count=dn->digits;                   /* digits to be checked  */
    for (up=dn->lsu; ; up++) {
      if (count<=DECDPUN) {
        /* this is the last Unit (the msu)  */
        if (*up!=DECNUMBER_powers[count-1]) break;     /* not 100..  */
        /* here if have the 1000... case  */
        sup=up;                              /* save msu pointer  */
        *up=(Unit)DECNUMBER_powers[count]-1;           /* here 100 in msu -> 999  */
        /* others all to all-nines, too  */
        for (up=up-1; up>=dn->lsu; up--) *up=(Unit)DECNUMBER_powers[DECDPUN]-1;
        dn->exponent--;                      /* and bump exponent  */

        /* iff the number was at the subnormal boundary (exponent=etiny)  */
        /* then the exponent is now out of range, so it will in fact get  */
        /* clamped to etiny and the final 9 dropped.  */
        /* printf(">> emin=%d exp=%d sdig=%d\n", set->emin,  */
        /*        dn->exponent, set->digits);  */
        if (dn->exponent+1==set->emin-set->digits+1) {
          if (count==1 && dn->digits==1) *sup=0;  /* here 9 -> 0[.9]  */
           else {
            *sup=(Unit)DECNUMBER_powers[count-1]-1;    /* here 999.. in msu -> 99..  */
            dn->digits--;
            }
          dn->exponent++;
          *status|=DEC_Underflow | DEC_Subnormal | DEC_Inexact | DEC_Rounded;
          }
        return;                              /* done  */
        }

      /* a full unit to check, with more to come  */
      if (*up!=0) break;                     /* not still 0s  */
      count-=DECDPUN;
      } /* up  */

    } /* bump<0  */

  /* Actual bump needed.  Do it.  */
  decUnitAddSub(dn->lsu, D2U(dn->digits), uarrone, 1, 0, dn->lsu, bump);
  } /* decApplyRound  */

#if DECSUBSET
/* ------------------------------------------------------------------ */
/* decFinish -- finish processing a number                            */
/*                                                                    */
/*   dn is the number                                                 */
/*   set is the context                                               */
/*   residue is the rounding accumulator (as in decApplyRound)        */
/*   status is the accumulator                                        */
/*                                                                    */
/* This finishes off the current number by:                           */
/*    1. If not extended:                                             */
/*       a. Converting a zero result to clean '0'                     */
/*       b. Reducing positive exponents to 0, if would fit in digits  */
/*    2. Checking for overflow and subnormals (always)                */
/* Note this is just Finalize when no subset arithmetic.              */
/* All fields are updated as required.                                */
/* ------------------------------------------------------------------ */
static void decFinish(decNumber *dn, decContext *set, Int *residue,
                      uInt *status) {
  if (!set->extended) {
    if ISZERO(dn) {                /* value is zero  */
      dn->exponent=0;              /* clean exponent ..  */
      dn->bits=0;                  /* .. and sign  */
      return;                      /* no error possible  */
      }
    if (dn->exponent>=0) {         /* non-negative exponent  */
      /* >0; reduce to integer if possible  */
      if (set->digits >= (dn->exponent+dn->digits)) {
        dn->digits=decShiftToMost(dn->lsu, dn->digits, dn->exponent);
        dn->exponent=0;
        }
      }
    } /* !extended  */

  decFinalize(dn, set, residue, status);
  } /* decFinish  */
#endif

/* ------------------------------------------------------------------ */
/* decFinalize -- final check, clamp, and round of a number           */
/*                                                                    */
/*   dn is the number                                                 */
/*   set is the context                                               */
/*   residue is the rounding accumulator (as in decApplyRound)        */
/*   status is the status accumulator                                 */
/*                                                                    */
/* This finishes off the current number by checking for subnormal     */
/* results, applying any pending rounding, checking for overflow,     */
/* and applying any clamping.                                         */
/* Underflow and overflow conditions are raised as appropriate.       */
/* All fields are updated as required.                                */
/* ------------------------------------------------------------------ */
static void decFinalize(decNumber *dn, decContext *set, Int *residue,
                        uInt *status) {
  Int shift;                            /* shift needed if clamping  */
  Int tinyexp=set->emin-dn->digits+1;   /* precalculate subnormal boundary  */

  /* Must be careful, here, when checking the exponent as the  */
  /* adjusted exponent could overflow 31 bits [because it may already  */
  /* be up to twice the expected].  */

  /* First test for subnormal.  This must be done before any final  */
  /* round as the result could be rounded to Nmin or 0.  */
  if (dn->exponent<=tinyexp) {          /* prefilter  */
    Int comp;
    decNumber nmin;
    /* A very nasty case here is dn == Nmin and residue<0  */
    if (dn->exponent<tinyexp) {
      /* Go handle subnormals; this will apply round if needed.  */
      decSetSubnormal(dn, set, residue, status);
      return;
      }
    /* Equals case: only subnormal if dn=Nmin and negative residue  */
    uprv_decNumberZero(&nmin);
    nmin.lsu[0]=1;
    nmin.exponent=set->emin;
    comp=decCompare(dn, &nmin, 1);                /* (signless compare)  */
    if (comp==DECNUMBER_BADINT) {                           /* oops  */
      *status|=DEC_Insufficient_storage;          /* abandon...  */
      return;
      }
    if (*residue<0 && comp==0) {                  /* neg residue and dn==Nmin  */
      decApplyRound(dn, set, *residue, status);   /* might force down  */
      decSetSubnormal(dn, set, residue, status);
      return;
      }
    }

  /* now apply any pending round (this could raise overflow).  */
  if (*residue!=0) decApplyRound(dn, set, *residue, status);

  /* Check for overflow [redundant in the 'rare' case] or clamp  */
  if (dn->exponent<=set->emax-set->digits+1) return;   /* neither needed  */


  /* here when might have an overflow or clamp to do  */
  if (dn->exponent>set->emax-dn->digits+1) {           /* too big  */
    decSetOverflow(dn, set, status);
    return;
    }
  /* here when the result is normal but in clamp range  */
  if (!set->clamp) return;

  /* here when need to apply the IEEE exponent clamp (fold-down)  */
  shift=dn->exponent-(set->emax-set->digits+1);

  /* shift coefficient (if non-zero)  */
  if (!ISZERO(dn)) {
    dn->digits=decShiftToMost(dn->lsu, dn->digits, shift);
    }
  dn->exponent-=shift;   /* adjust the exponent to match  */
  *status|=DEC_Clamped;  /* and record the dirty deed  */
  return;
  } /* decFinalize  */

/* ------------------------------------------------------------------ */
/* decSetOverflow -- set number to proper overflow value              */
/*                                                                    */
/*   dn is the number (used for sign [only] and result)               */
/*   set is the context [used for the rounding mode, etc.]            */
/*   status contains the current status to be updated                 */
/*                                                                    */
/* This sets the sign of a number and sets its value to either        */
/* Infinity or the maximum finite value, depending on the sign of     */
/* dn and the rounding mode, following IEEE 754 rules.                */
/* ------------------------------------------------------------------ */
static void decSetOverflow(decNumber *dn, decContext *set, uInt *status) {
  Flag needmax=0;                  /* result is maximum finite value  */
  uByte sign=dn->bits&DECNEG;      /* clean and save sign bit  */

  if (ISZERO(dn)) {                /* zero does not overflow magnitude  */
    Int emax=set->emax;                      /* limit value  */
    if (set->clamp) emax-=set->digits-1;     /* lower if clamping  */
    if (dn->exponent>emax) {                 /* clamp required  */
      dn->exponent=emax;
      *status|=DEC_Clamped;
      }
    return;
    }

  uprv_decNumberZero(dn);
  switch (set->round) {
    case DEC_ROUND_DOWN: {
      needmax=1;                   /* never Infinity  */
      break;} /* r-d  */
    case DEC_ROUND_05UP: {
      needmax=1;                   /* never Infinity  */
      break;} /* r-05  */
    case DEC_ROUND_CEILING: {
      if (sign) needmax=1;         /* Infinity if non-negative  */
      break;} /* r-c  */
    case DEC_ROUND_FLOOR: {
      if (!sign) needmax=1;        /* Infinity if negative  */
      break;} /* r-f  */
    default: break;                /* Infinity in all other cases  */
    }
  if (needmax) {
    decSetMaxValue(dn, set);
    dn->bits=sign;                 /* set sign  */
    }
   else dn->bits=sign|DECINF;      /* Value is +/-Infinity  */
  *status|=DEC_Overflow | DEC_Inexact | DEC_Rounded;
  } /* decSetOverflow  */

/* ------------------------------------------------------------------ */
/* decSetMaxValue -- set number to +Nmax (maximum normal value)       */
/*                                                                    */
/*   dn is the number to set                                          */
/*   set is the context [used for digits and emax]                    */
/*                                                                    */
/* This sets the number to the maximum positive value.                */
/* ------------------------------------------------------------------ */
static void decSetMaxValue(decNumber *dn, decContext *set) {
  Unit *up;                        /* work  */
  Int count=set->digits;           /* nines to add  */
  dn->digits=count;
  /* fill in all nines to set maximum value  */
  for (up=dn->lsu; ; up++) {
    if (count>DECDPUN) *up=DECDPUNMAX;  /* unit full o'nines  */
     else {                             /* this is the msu  */
      *up=(Unit)(DECNUMBER_powers[count]-1);
      break;
      }
    count-=DECDPUN;                /* filled those digits  */
    } /* up  */
  dn->bits=0;                      /* + sign  */
  dn->exponent=set->emax-set->digits+1;
  } /* decSetMaxValue  */

/* ------------------------------------------------------------------ */
/* decSetSubnormal -- process value whose exponent is <Emin           */
/*                                                                    */
/*   dn is the number (used as input as well as output; it may have   */
/*         an allowed subnormal value, which may need to be rounded)  */
/*   set is the context [used for the rounding mode]                  */
/*   residue is any pending residue                                   */
/*   status contains the current status to be updated                 */
/*                                                                    */
/* If subset mode, set result to zero and set Underflow flags.        */
/*                                                                    */
/* Value may be zero with a low exponent; this does not set Subnormal */
/* but the exponent will be clamped to Etiny.                         */
/*                                                                    */
/* Otherwise ensure exponent is not out of range, and round as        */
/* necessary.  Underflow is set if the result is Inexact.             */
/* ------------------------------------------------------------------ */
static void decSetSubnormal(decNumber *dn, decContext *set, Int *residue,
                            uInt *status) {
  decContext workset;         /* work  */
  Int        etiny, adjust;   /* ..  */

  #if DECSUBSET
  /* simple set to zero and 'hard underflow' for subset  */
  if (!set->extended) {
    uprv_decNumberZero(dn);
    /* always full overflow  */
    *status|=DEC_Underflow | DEC_Subnormal | DEC_Inexact | DEC_Rounded;
    return;
    }
  #endif

  /* Full arithmetic -- allow subnormals, rounded to minimum exponent  */
  /* (Etiny) if needed  */
  etiny=set->emin-(set->digits-1);      /* smallest allowed exponent  */

  if ISZERO(dn) {                       /* value is zero  */
    /* residue can never be non-zero here  */
    #if DECCHECK
      if (*residue!=0) {
        printf("++ Subnormal 0 residue %ld\n", (LI)*residue);
        *status|=DEC_Invalid_operation;
        }
    #endif
    if (dn->exponent<etiny) {           /* clamp required  */
      dn->exponent=etiny;
      *status|=DEC_Clamped;
      }
    return;
    }

  *status|=DEC_Subnormal;               /* have a non-zero subnormal  */
  adjust=etiny-dn->exponent;            /* calculate digits to remove  */
  if (adjust<=0) {                      /* not out of range; unrounded  */
    /* residue can never be non-zero here, except in the Nmin-residue  */
    /* case (which is a subnormal result), so can take fast-path here  */
    /* it may already be inexact (from setting the coefficient)  */
    if (*status&DEC_Inexact) *status|=DEC_Underflow;
    return;
    }

  /* adjust>0, so need to rescale the result so exponent becomes Etiny  */
  /* [this code is similar to that in rescale]  */
  workset=*set;                         /* clone rounding, etc.  */
  workset.digits=dn->digits-adjust;     /* set requested length  */
  workset.emin-=adjust;                 /* and adjust emin to match  */
  /* [note that the latter can be <1, here, similar to Rescale case]  */
  decSetCoeff(dn, &workset, dn->lsu, dn->digits, residue, status);
  decApplyRound(dn, &workset, *residue, status);

  /* Use 754 default rule: Underflow is set iff Inexact  */
  /* [independent of whether trapped]  */
  if (*status&DEC_Inexact) *status|=DEC_Underflow;

  /* if rounded up a 999s case, exponent will be off by one; adjust  */
  /* back if so [it will fit, because it was shortened earlier]  */
  if (dn->exponent>etiny) {
    dn->digits=decShiftToMost(dn->lsu, dn->digits, 1);
    dn->exponent--;                     /* (re)adjust the exponent.  */
    }

  /* if rounded to zero, it is by definition clamped...  */
  if (ISZERO(dn)) *status|=DEC_Clamped;
  } /* decSetSubnormal  */

/* ------------------------------------------------------------------ */
/* decCheckMath - check entry conditions for a math function          */
/*                                                                    */
/*   This checks the context and the operand                          */
/*                                                                    */
/*   rhs is the operand to check                                      */
/*   set is the context to check                                      */
/*   status is unchanged if both are good                             */
/*                                                                    */
/* returns non-zero if status is changed, 0 otherwise                 */
/*                                                                    */
/* Restrictions enforced:                                             */
/*                                                                    */
/*   digits, emax, and -emin in the context must be less than         */
/*   DEC_MAX_MATH (999999), and A must be within these bounds if      */
/*   non-zero.  Invalid_operation is set in the status if a           */
/*   restriction is violated.                                         */
/* ------------------------------------------------------------------ */
static uInt decCheckMath(const decNumber *rhs, decContext *set,
                         uInt *status) {
  uInt save=*status;                         /* record  */
  if (set->digits>DEC_MAX_MATH
   || set->emax>DEC_MAX_MATH
   || -set->emin>DEC_MAX_MATH) *status|=DEC_Invalid_context;
   else if ((rhs->digits>DEC_MAX_MATH
     || rhs->exponent+rhs->digits>DEC_MAX_MATH+1
     || rhs->exponent+rhs->digits<2*(1-DEC_MAX_MATH))
     && !ISZERO(rhs)) *status|=DEC_Invalid_operation;
  return (*status!=save);
  } /* decCheckMath  */

/* ------------------------------------------------------------------ */
/* decGetInt -- get integer from a number                             */
/*                                                                    */
/*   dn is the number [which will not be altered]                     */
/*                                                                    */
/*   returns one of:                                                  */
/*     DECNUMBER_BADINT if there is a non-zero fraction                         */
/*     the converted integer                                          */
/*     DECNUMBER_BIGEVEN if the integer is even and magnitude > 2*10**9         */
/*     DECNUMBER_BIGODD  if the integer is odd  and magnitude > 2*10**9         */
/*                                                                    */
/* This checks and gets a whole number from the input decNumber.      */
/* The sign can be determined from dn by the caller when DECNUMBER_BIGEVEN or   */
/* DECNUMBER_BIGODD is returned.                                                */
/* ------------------------------------------------------------------ */
static Int decGetInt(const decNumber *dn) {
  Int  theInt;                          /* result accumulator  */
  const Unit *up;                       /* work  */
  Int  got;                             /* digits (real or not) processed  */
  Int  ilength=dn->digits+dn->exponent; /* integral length  */
  Flag neg=decNumberIsNegative(dn);     /* 1 if -ve  */

  /* The number must be an integer that fits in 10 digits  */
  /* Assert, here, that 10 is enough for any rescale Etiny  */
  #if DEC_MAX_EMAX > 999999999
    #error GetInt may need updating [for Emax]
  #endif
  #if DEC_MIN_EMIN < -999999999
    #error GetInt may need updating [for Emin]
  #endif
  if (ISZERO(dn)) return 0;             /* zeros are OK, with any exponent  */

  up=dn->lsu;                           /* ready for lsu  */
  theInt=0;                             /* ready to accumulate  */
  if (dn->exponent>=0) {                /* relatively easy  */
    /* no fractional part [usual]; allow for positive exponent  */
    got=dn->exponent;
    }
   else { /* -ve exponent; some fractional part to check and discard  */
    Int count=-dn->exponent;            /* digits to discard  */
    /* spin up whole units until reach the Unit with the unit digit  */
    for (; count>=DECDPUN; up++) {
      if (*up!=0) return DECNUMBER_BADINT;        /* non-zero Unit to discard  */
      count-=DECDPUN;
      }
    if (count==0) got=0;                /* [a multiple of DECDPUN]  */
     else {                             /* [not multiple of DECDPUN]  */
      Int rem;                          /* work  */
      /* slice off fraction digits and check for non-zero  */
      #if DECDPUN<=4
        theInt=QUOT10(*up, count);
        rem=*up-theInt*DECNUMBER_powers[count];
      #else
        rem=*up%DECNUMBER_powers[count];          /* slice off discards  */
        theInt=*up/DECNUMBER_powers[count];
      #endif
      if (rem!=0) return DECNUMBER_BADINT;        /* non-zero fraction  */
      /* it looks good  */
      got=DECDPUN-count;                /* number of digits so far  */
      up++;                             /* ready for next  */
      }
    }
  /* now it's known there's no fractional part  */

  /* tricky code now, to accumulate up to 9.3 digits  */
  if (got==0) {theInt=*up; got+=DECDPUN; up++;} /* ensure lsu is there  */

  if (ilength<11) {
    Int save=theInt;
    /* collect any remaining unit(s)  */
    for (; got<ilength; up++) {
      theInt+=*up*DECNUMBER_powers[got];
      got+=DECDPUN;
      }
    if (ilength==10) {                  /* need to check for wrap  */
      if (theInt/(Int)DECNUMBER_powers[got-DECDPUN]!=(Int)*(up-1)) ilength=11;
         /* [that test also disallows the DECNUMBER_BADINT result case]  */
       else if (neg && theInt>1999999997) ilength=11;
       else if (!neg && theInt>999999999) ilength=11;
      if (ilength==11) theInt=save;     /* restore correct low bit  */
      }
    }

  if (ilength>10) {                     /* too big  */
    if (theInt&1) return DECNUMBER_BIGODD;        /* bottom bit 1  */
    return DECNUMBER_BIGEVEN;                     /* bottom bit 0  */
    }

  if (neg) theInt=-theInt;              /* apply sign  */
  return theInt;
  } /* decGetInt  */

/* ------------------------------------------------------------------ */
/* decDecap -- decapitate the coefficient of a number                 */
/*                                                                    */
/*   dn   is the number to be decapitated                             */
/*   drop is the number of digits to be removed from the left of dn;  */
/*     this must be <= dn->digits (if equal, the coefficient is       */
/*     set to 0)                                                      */
/*                                                                    */
/* Returns dn; dn->digits will be <= the initial digits less drop     */
/* (after removing drop digits there may be leading zero digits       */
/* which will also be removed).  Only dn->lsu and dn->digits change.  */
/* ------------------------------------------------------------------ */
static decNumber *decDecap(decNumber *dn, Int drop) {
  Unit *msu;                            /* -> target cut point  */
  Int cut;                              /* work  */
  if (drop>=dn->digits) {               /* losing the whole thing  */
    #if DECCHECK
    if (drop>dn->digits)
      printf("decDecap called with drop>digits [%ld>%ld]\n",
             (LI)drop, (LI)dn->digits);
    #endif
    dn->lsu[0]=0;
    dn->digits=1;
    return dn;
    }
  msu=dn->lsu+D2U(dn->digits-drop)-1;   /* -> likely msu  */
  cut=MSUDIGITS(dn->digits-drop);       /* digits to be in use in msu  */
  if (cut!=DECDPUN) *msu%=DECNUMBER_powers[cut];  /* clear left digits  */
  /* that may have left leading zero digits, so do a proper count...  */
  dn->digits=decGetDigits(dn->lsu, static_cast<int32_t>(msu-dn->lsu+1));
  return dn;
  } /* decDecap  */

/* ------------------------------------------------------------------ */
/* decBiStr -- compare string with pairwise options                   */
/*                                                                    */
/*   targ is the string to compare                                    */
/*   str1 is one of the strings to compare against (length may be 0)  */
/*   str2 is the other; it must be the same length as str1            */
/*                                                                    */
/*   returns 1 if strings compare equal, (that is, it is the same     */
/*   length as str1 and str2, and each character of targ is in either */
/*   str1 or str2 in the corresponding position), or 0 otherwise      */
/*                                                                    */
/* This is used for generic caseless compare, including the awkward   */
/* case of the Turkish dotted and dotless Is.  Use as (for example):  */
/*   if (decBiStr(test, "mike", "MIKE")) ...                          */
/* ------------------------------------------------------------------ */
static Flag decBiStr(const char *targ, const char *str1, const char *str2) {
  for (;;targ++, str1++, str2++) {
    if (*targ!=*str1 && *targ!=*str2) return 0;
    /* *targ has a match in one (or both, if terminator)  */
    if (*targ=='\0') break;
    } /* forever  */
  return 1;
  } /* decBiStr  */

/* ------------------------------------------------------------------ */
/* decNaNs -- handle NaN operand or operands                          */
/*                                                                    */
/*   res     is the result number                                     */
/*   lhs     is the first operand                                     */
/*   rhs     is the second operand, or NULL if none                   */
/*   context is used to limit payload length                          */
/*   status  contains the current status                              */
/*   returns res in case convenient                                   */
/*                                                                    */
/* Called when one or both operands is a NaN, and propagates the      */
/* appropriate result to res.  When an sNaN is found, it is changed   */
/* to a qNaN and Invalid operation is set.                            */
/* ------------------------------------------------------------------ */
static decNumber * decNaNs(decNumber *res, const decNumber *lhs,
                           const decNumber *rhs, decContext *set,
                           uInt *status) {
  /* This decision tree ends up with LHS being the source pointer,  */
  /* and status updated if need be  */
  if (lhs->bits & DECSNAN)
    *status|=DEC_Invalid_operation | DECNUMBER_DEC_sNaN;
   else if (rhs==NULL);
   else if (rhs->bits & DECSNAN) {
    lhs=rhs;
    *status|=DEC_Invalid_operation | DECNUMBER_DEC_sNaN;
    }
   else if (lhs->bits & DECNAN);
   else lhs=rhs;

  /* propagate the payload  */
  if (lhs->digits<=set->digits) uprv_decNumberCopy(res, lhs); /* easy  */
   else { /* too long  */
    const Unit *ul;
    Unit *ur, *uresp1;
    /* copy safe number of units, then decapitate  */
    res->bits=lhs->bits;                /* need sign etc.  */
    uresp1=res->lsu+D2U(set->digits);
    for (ur=res->lsu, ul=lhs->lsu; ur<uresp1; ur++, ul++) *ur=*ul;
    res->digits=D2U(set->digits)*DECDPUN;
    /* maybe still too long  */
    if (res->digits>set->digits) decDecap(res, res->digits-set->digits);
    }

  res->bits&=~DECSNAN;        /* convert any sNaN to NaN, while  */
  res->bits|=DECNAN;          /* .. preserving sign  */
  res->exponent=0;            /* clean exponent  */
                              /* [coefficient was copied/decapitated]  */
  return res;
  } /* decNaNs  */

/* ------------------------------------------------------------------ */
/* decStatus -- apply non-zero status                                 */
/*                                                                    */
/*   dn     is the number to set if error                             */
/*   status contains the current status (not yet in context)          */
/*   set    is the context                                            */
/*                                                                    */
/* If the status is an error status, the number is set to a NaN,      */
/* unless the error was an overflow, divide-by-zero, or underflow,    */
/* in which case the number will have already been set.               */
/*                                                                    */
/* The context status is then updated with the new status.  Note that */
/* this may raise a signal, so control may never return from this     */
/* routine (hence resources must be recovered before it is called).   */
/* ------------------------------------------------------------------ */
static void decStatus(decNumber *dn, uInt status, decContext *set) {
  if (status & DEC_NaNs) {              /* error status -> NaN  */
    /* if cause was an sNaN, clear and propagate [NaN is already set up]  */
    if (status & DECNUMBER_DEC_sNaN) status&=~DECNUMBER_DEC_sNaN;
     else {
      uprv_decNumberZero(dn);                /* other error: clean throughout  */
      dn->bits=DECNAN;                  /* and make a quiet NaN  */
      }
    }
  uprv_decContextSetStatus(set, status);     /* [may not return]  */
  return;
  } /* decStatus  */

/* ------------------------------------------------------------------ */
/* decGetDigits -- count digits in a Units array                      */
/*                                                                    */
/*   uar is the Unit array holding the number (this is often an       */
/*          accumulator of some sort)                                 */
/*   len is the length of the array in units [>=1]                    */
/*                                                                    */
/*   returns the number of (significant) digits in the array          */
/*                                                                    */
/* All leading zeros are excluded, except the last if the array has   */
/* only zero Units.                                                   */
/* ------------------------------------------------------------------ */
/* This may be called twice during some operations.  */
static Int decGetDigits(Unit *uar, Int len) {
  Unit *up=uar+(len-1);            /* -> msu  */
  Int  digits=(len-1)*DECDPUN+1;   /* possible digits excluding msu  */
  #if DECDPUN>4
  uInt const *pow;                 /* work  */
  #endif
                                   /* (at least 1 in final msu)  */
  #if DECCHECK
  if (len<1) printf("decGetDigits called with len<1 [%ld]\n", (LI)len);
  #endif

  for (; up>=uar; up--) {
    if (*up==0) {                  /* unit is all 0s  */
      if (digits==1) break;        /* a zero has one digit  */
      digits-=DECDPUN;             /* adjust for 0 unit  */
      continue;}
    /* found the first (most significant) non-zero Unit  */
    #if DECDPUN>1                  /* not done yet  */
    if (*up<10) break;             /* is 1-9  */
    digits++;
    #if DECDPUN>2                  /* not done yet  */
    if (*up<100) break;            /* is 10-99  */
    digits++;
    #if DECDPUN>3                  /* not done yet  */
    if (*up<1000) break;           /* is 100-999  */
    digits++;
    #if DECDPUN>4                  /* count the rest ...  */
    for (pow=&DECNUMBER_powers[4]; *up>=*pow; pow++) digits++;
    #endif
    #endif
    #endif
    #endif
    break;
    } /* up  */
  return digits;
  } /* decGetDigits  */

#if DECTRACE | DECCHECK
/* ------------------------------------------------------------------ */
/* decNumberShow -- display a number [debug aid]                      */
/*   dn is the number to show                                         */
/*                                                                    */
/* Shows: sign, exponent, coefficient (msu first), digits             */
/*    or: sign, special-value                                         */
/* ------------------------------------------------------------------ */
/* this is public so other modules can use it  */
void uprv_decNumberShow(const decNumber *dn) {
  const Unit *up;                  /* work  */
  uInt u, d;                       /* ..  */
  Int cut;                         /* ..  */
  char isign='+';                  /* main sign  */
  if (dn==NULL) {
    printf("NULL\n");
    return;}
  if (decNumberIsNegative(dn)) isign='-';
  printf(" >> %c ", isign);
  if (dn->bits&DECSPECIAL) {       /* Is a special value  */
    if (decNumberIsInfinite(dn)) printf("Infinity");
     else {                                  /* a NaN  */
      if (dn->bits&DECSNAN) printf("sNaN");  /* signalling NaN  */
       else printf("NaN");
      }
    /* if coefficient and exponent are 0, no more to do  */
    if (dn->exponent==0 && dn->digits==1 && *dn->lsu==0) {
      printf("\n");
      return;}
    /* drop through to report other information  */
    printf(" ");
    }

  /* now carefully display the coefficient  */
  up=dn->lsu+D2U(dn->digits)-1;         /* msu  */
  printf("%ld", (LI)*up);
  for (up=up-1; up>=dn->lsu; up--) {
    u=*up;
    printf(":");
    for (cut=DECDPUN-1; cut>=0; cut--) {
      d=u/DECNUMBER_powers[cut];
      u-=d*DECNUMBER_powers[cut];
      printf("%ld", (LI)d);
      } /* cut  */
    } /* up  */
  if (dn->exponent!=0) {
    char esign='+';
    if (dn->exponent<0) esign='-';
    printf(" E%c%ld", esign, (LI)abs(dn->exponent));
    }
  printf(" [%ld]\n", (LI)dn->digits);
  } /* decNumberShow  */
#endif

#if DECTRACE || DECCHECK
/* ------------------------------------------------------------------ */
/* decDumpAr -- display a unit array [debug/check aid]                */
/*   name is a single-character tag name                              */
/*   ar   is the array to display                                     */
/*   len  is the length of the array in Units                         */
/* ------------------------------------------------------------------ */
static void decDumpAr(char name, const Unit *ar, Int len) {
  Int i;
  const char *spec;
  #if DECDPUN==9
    spec="%09d ";
  #elif DECDPUN==8
    spec="%08d ";
  #elif DECDPUN==7
    spec="%07d ";
  #elif DECDPUN==6
    spec="%06d ";
  #elif DECDPUN==5
    spec="%05d ";
  #elif DECDPUN==4
    spec="%04d ";
  #elif DECDPUN==3
    spec="%03d ";
  #elif DECDPUN==2
    spec="%02d ";
  #else
    spec="%d ";
  #endif
  printf("  :%c: ", name);
  for (i=len-1; i>=0; i--) {
    if (i==len-1) printf("%ld ", (LI)ar[i]);
     else printf(spec, ar[i]);
    }
  printf("\n");
  return;}
#endif

#if DECCHECK
/* ------------------------------------------------------------------ */
/* decCheckOperands -- check operand(s) to a routine                  */
/*   res is the result structure (not checked; it will be set to      */
/*          quiet NaN if error found (and it is not NULL))            */
/*   lhs is the first operand (may be DECUNRESU)                      */
/*   rhs is the second (may be DECUNUSED)                             */
/*   set is the context (may be DECUNCONT)                            */
/*   returns 0 if both operands, and the context are clean, or 1      */
/*     otherwise (in which case the context will show an error,       */
/*     unless NULL).  Note that res is not cleaned; caller should     */
/*     handle this so res=NULL case is safe.                          */
/* The caller is expected to abandon immediately if 1 is returned.    */
/* ------------------------------------------------------------------ */
static Flag decCheckOperands(decNumber *res, const decNumber *lhs,
                             const decNumber *rhs, decContext *set) {
  Flag bad=0;
  if (set==NULL) {                 /* oops; hopeless  */
    #if DECTRACE || DECNUMBER_DECVERB
    printf("Reference to context is NULL.\n");
    #endif
    bad=1;
    return 1;}
   else if (set!=DECUNCONT
     && (set->digits<1 || set->round>=DEC_ROUND_MAX)) {
    bad=1;
    #if DECTRACE || DECNUMBER_DECVERB
    printf("Bad context [digits=%ld round=%ld].\n",
           (LI)set->digits, (LI)set->round);
    #endif
    }
   else {
    if (res==NULL) {
      bad=1;
      #if DECTRACE
      /* this one not DECNUMBER_DECVERB as standard tests include NULL  */
      printf("Reference to result is NULL.\n");
      #endif
      }
    if (!bad && lhs!=DECUNUSED) bad=(decCheckNumber(lhs));
    if (!bad && rhs!=DECUNUSED) bad=(decCheckNumber(rhs));
    }
  if (bad) {
    if (set!=DECUNCONT) uprv_decContextSetStatus(set, DEC_Invalid_operation);
    if (res!=DECUNRESU && res!=NULL) {
      uprv_decNumberZero(res);
      res->bits=DECNAN;       /* qNaN  */
      }
    }
  return bad;
  } /* decCheckOperands  */

/* ------------------------------------------------------------------ */
/* decCheckNumber -- check a number                                   */
/*   dn is the number to check                                        */
/*   returns 0 if the number is clean, or 1 otherwise                 */
/*                                                                    */
/* The number is considered valid if it could be a result from some   */
/* operation in some valid context.                                   */
/* ------------------------------------------------------------------ */
static Flag decCheckNumber(const decNumber *dn) {
  const Unit *up;             /* work  */
  uInt maxuint;               /* ..  */
  Int ae, d, digits;          /* ..  */
  Int emin, emax;             /* ..  */

  if (dn==NULL) {             /* hopeless  */
    #if DECTRACE
    /* this one not DECNUMBER_DECVERB as standard tests include NULL  */
    printf("Reference to decNumber is NULL.\n");
    #endif
    return 1;}

  /* check special values  */
  if (dn->bits & DECSPECIAL) {
    if (dn->exponent!=0) {
      #if DECTRACE || DECNUMBER_DECVERB
      printf("Exponent %ld (not 0) for a special value [%02x].\n",
             (LI)dn->exponent, dn->bits);
      #endif
      return 1;}

    /* 2003.09.08: NaNs may now have coefficients, so next tests Inf only  */
    if (decNumberIsInfinite(dn)) {
      if (dn->digits!=1) {
        #if DECTRACE || DECNUMBER_DECVERB
        printf("Digits %ld (not 1) for an infinity.\n", (LI)dn->digits);
        #endif
        return 1;}
      if (*dn->lsu!=0) {
        #if DECTRACE || DECNUMBER_DECVERB
        printf("LSU %ld (not 0) for an infinity.\n", (LI)*dn->lsu);
        #endif
        decDumpAr('I', dn->lsu, D2U(dn->digits));
        return 1;}
      } /* Inf  */
    /* 2002.12.26: negative NaNs can now appear through proposed IEEE  */
    /*             concrete formats (decimal64, etc.).  */
    return 0;
    }

  /* check the coefficient  */
  if (dn->digits<1 || dn->digits>DECNUMMAXP) {
    #if DECTRACE || DECNUMBER_DECVERB
    printf("Digits %ld in number.\n", (LI)dn->digits);
    #endif
    return 1;}

  d=dn->digits;

  for (up=dn->lsu; d>0; up++) {
    if (d>DECDPUN) maxuint=DECDPUNMAX;
     else {                   /* reached the msu  */
      maxuint=DECNUMBER_powers[d]-1;
      if (dn->digits>1 && *up<DECNUMBER_powers[d-1]) {
        #if DECTRACE || DECNUMBER_DECVERB
        printf("Leading 0 in number.\n");
        uprv_decNumberShow(dn);
        #endif
        return 1;}
      }
    if (*up>maxuint) {
      #if DECTRACE || DECNUMBER_DECVERB
      printf("Bad Unit [%08lx] in %ld-digit number at offset %ld [maxuint %ld].\n",
              (LI)*up, (LI)dn->digits, (LI)(up-dn->lsu), (LI)maxuint);
      #endif
      return 1;}
    d-=DECDPUN;
    }

  /* check the exponent.  Note that input operands can have exponents  */
  /* which are out of the set->emin/set->emax and set->digits range  */
  /* (just as they can have more digits than set->digits).  */
  ae=dn->exponent+dn->digits-1;    /* adjusted exponent  */
  emax=DECNUMMAXE;
  emin=DECNUMMINE;
  digits=DECNUMMAXP;
  if (ae<emin-(digits-1)) {
    #if DECTRACE || DECNUMBER_DECVERB
    printf("Adjusted exponent underflow [%ld].\n", (LI)ae);
    uprv_decNumberShow(dn);
    #endif
    return 1;}
  if (ae>+emax) {
    #if DECTRACE || DECNUMBER_DECVERB
    printf("Adjusted exponent overflow [%ld].\n", (LI)ae);
    uprv_decNumberShow(dn);
    #endif
    return 1;}

  return 0;              /* it's OK  */
  } /* decCheckNumber  */

/* ------------------------------------------------------------------ */
/* decCheckInexact -- check a normal finite inexact result has digits */
/*   dn is the number to check                                        */
/*   set is the context (for status and precision)                    */
/*   sets Invalid operation, etc., if some digits are missing         */
/* [this check is not made for DECSUBSET compilation or when          */
/* subnormal is not set]                                              */
/* ------------------------------------------------------------------ */
static void decCheckInexact(const decNumber *dn, decContext *set) {
  #if !DECSUBSET && DECEXTFLAG
    if ((set->status & (DEC_Inexact|DEC_Subnormal))==DEC_Inexact
     && (set->digits!=dn->digits) && !(dn->bits & DECSPECIAL)) {
      #if DECTRACE || DECNUMBER_DECVERB
      printf("Insufficient digits [%ld] on normal Inexact result.\n",
             (LI)dn->digits);
      uprv_decNumberShow(dn);
      #endif
      uprv_decContextSetStatus(set, DEC_Invalid_operation);
      }
  #else
    /* next is a noop for quiet compiler  */
    if (dn!=NULL && dn->digits==0) set->status|=DEC_Invalid_operation;
  #endif
  return;
  } /* decCheckInexact  */
#endif

#if DECALLOC
#undef malloc
#undef free
/* ------------------------------------------------------------------ */
/* decMalloc -- accountable allocation routine                        */
/*   n is the number of bytes to allocate                             */
/*                                                                    */
/* Semantics is the same as the stdlib malloc routine, but bytes      */
/* allocated are accounted for globally, and corruption fences are    */
/* added before and after the 'actual' storage.                       */
/* ------------------------------------------------------------------ */
/* This routine allocates storage with an extra twelve bytes; 8 are   */
/* at the start and hold:                                             */
/*   0-3 the original length requested                                */
/*   4-7 buffer corruption detection fence (DECFENCE, x4)             */
/* The 4 bytes at the end also hold a corruption fence (DECFENCE, x4) */
/* ------------------------------------------------------------------ */
static void *decMalloc(size_t n) {
  uInt  size=n+12;                 /* true size  */
  void  *alloc;                    /* -> allocated storage  */
  uByte *b, *b0;                   /* work  */
  uInt  uiwork;                    /* for macros  */

  alloc=malloc(size);              /* -> allocated storage  */
  if (alloc==NULL) return NULL;    /* out of strorage  */
  b0=(uByte *)alloc;               /* as bytes  */
  decAllocBytes+=n;                /* account for storage  */
  UBFROMUI(alloc, n);              /* save n  */
  /* printf(" alloc ++ dAB: %ld (%ld)\n", (LI)decAllocBytes, (LI)n);  */
  for (b=b0+4; b<b0+8; b++) *b=DECFENCE;
  for (b=b0+n+8; b<b0+n+12; b++) *b=DECFENCE;
  return b0+8;                     /* -> play area  */
  } /* decMalloc  */

/* ------------------------------------------------------------------ */
/* decFree -- accountable free routine                                */
/*   alloc is the storage to free                                     */
/*                                                                    */
/* Semantics is the same as the stdlib malloc routine, except that    */
/* the global storage accounting is updated and the fences are        */
/* checked to ensure that no routine has written 'out of bounds'.     */
/* ------------------------------------------------------------------ */
/* This routine first checks that the fences have not been corrupted. */
/* It then frees the storage using the 'truw' storage address (that   */
/* is, offset by 8).                                                  */
/* ------------------------------------------------------------------ */
static void decFree(void *alloc) {
  uInt  n;                         /* original length  */
  uByte *b, *b0;                   /* work  */
  uInt  uiwork;                    /* for macros  */

  if (alloc==NULL) return;         /* allowed; it's a nop  */
  b0=(uByte *)alloc;               /* as bytes  */
  b0-=8;                           /* -> true start of storage  */
  n=UBTOUI(b0);                    /* lift length  */
  for (b=b0+4; b<b0+8; b++) if (*b!=DECFENCE)
    printf("=== Corrupt byte [%02x] at offset %d from %ld ===\n", *b,
           b-b0-8, (LI)b0);
  for (b=b0+n+8; b<b0+n+12; b++) if (*b!=DECFENCE)
    printf("=== Corrupt byte [%02x] at offset +%d from %ld, n=%ld ===\n", *b,
           b-b0-8, (LI)b0, (LI)n);
  free(b0);                        /* drop the storage  */
  decAllocBytes-=n;                /* account for storage  */
  /* printf(" free -- dAB: %d (%d)\n", decAllocBytes, -n);  */
  } /* decFree  */
#define malloc(a) decMalloc(a)
#define free(a) decFree(a)
#endif
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif


#include <cmath>
#include <cstdlib>
#include <stdlib.h>








// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING




// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN namespace number {
namespace impl {

// Forward-declare
class ScientificHandler;

class U_I18N_API ScientificModifier : public UMemory, public Modifier {
  public:
    ScientificModifier();

    void set(int32_t exponent, const ScientificHandler *handler);

    int32_t apply(FormattedStringBuilder &output, int32_t leftIndex, int32_t rightIndex,
                  UErrorCode &status) const U_OVERRIDE;

    int32_t getPrefixLength() const U_OVERRIDE;

    int32_t getCodePointCount() const U_OVERRIDE;

    bool isStrong() const U_OVERRIDE;

    bool containsField(UNumberFormatFields field) const U_OVERRIDE;

    void getParameters(Parameters& output) const U_OVERRIDE;

    bool semanticallyEquivalent(const Modifier& other) const U_OVERRIDE;

  private:
    int32_t fExponent;
    const ScientificHandler *fHandler;
};

class ScientificHandler : public UMemory, public MicroPropsGenerator, public MultiplierProducer {
  public:
    ScientificHandler(const Notation *notation, const DecimalFormatSymbols *symbols,
                      const MicroPropsGenerator *parent);

    void
    processQuantity(DecimalQuantity &quantity, MicroProps &micros, UErrorCode &status) const U_OVERRIDE;

    int32_t getMultiplier(int32_t magnitude) const U_OVERRIDE;

  private:
    const Notation::ScientificSettings& fSettings;
    const DecimalFormatSymbols *fSymbols;
    const MicroPropsGenerator *fParent;

    friend class ScientificModifier;
};

} // namespace impl
} // namespace number
U_NAMESPACE_END

#endif //__NUMBER_SCIENTIFIC_H__



// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

#include <algorithm>
#include <cstdint>






U_NAMESPACE_BEGIN namespace number {
namespace impl {

/**
 * The canonical implementation of {@link Modifier}, containing a prefix and suffix string.
 * TODO: This is not currently being used by real code and could be removed.
 */
class U_I18N_API ConstantAffixModifier : public Modifier, public UObject {
  public:
    ConstantAffixModifier(const UnicodeString &prefix, const UnicodeString &suffix, Field field,
                          bool strong)
            : fPrefix(prefix), fSuffix(suffix), fField(field), fStrong(strong) {}

    int32_t apply(FormattedStringBuilder &output, int32_t leftIndex, int32_t rightIndex,
                  UErrorCode &status) const U_OVERRIDE;

    int32_t getPrefixLength() const U_OVERRIDE;

    int32_t getCodePointCount() const U_OVERRIDE;

    bool isStrong() const U_OVERRIDE;

    bool containsField(UNumberFormatFields field) const U_OVERRIDE;

    void getParameters(Parameters& output) const U_OVERRIDE;

    bool semanticallyEquivalent(const Modifier& other) const U_OVERRIDE;

  private:
    UnicodeString fPrefix;
    UnicodeString fSuffix;
    Field fField;
    bool fStrong;
};

/**
 * The second primary implementation of {@link Modifier}, this one consuming a {@link SimpleFormatter}
 * pattern.
 */
class U_I18N_API SimpleModifier : public Modifier, public UMemory {
  public:
    SimpleModifier(const SimpleFormatter &simpleFormatter, Field field, bool strong);

    SimpleModifier(const SimpleFormatter &simpleFormatter, Field field, bool strong,
                   const Modifier::Parameters parameters);

    // Default constructor for LongNameHandler.h
    SimpleModifier();

    int32_t apply(FormattedStringBuilder &output, int32_t leftIndex, int32_t rightIndex,
                  UErrorCode &status) const U_OVERRIDE;

    int32_t getPrefixLength() const U_OVERRIDE;

    int32_t getCodePointCount() const U_OVERRIDE;

    bool isStrong() const U_OVERRIDE;

    bool containsField(UNumberFormatFields field) const U_OVERRIDE;

    void getParameters(Parameters& output) const U_OVERRIDE;

    bool semanticallyEquivalent(const Modifier& other) const U_OVERRIDE;

    /**
     * TODO: This belongs in SimpleFormatterImpl. The only reason I haven't moved it there yet is because
     * FormattedStringBuilder is an internal class and SimpleFormatterImpl feels like it should not depend on it.
     *
     * <p>
     * Formats a value that is already stored inside the StringBuilder <code>result</code> between the indices
     * <code>startIndex</code> and <code>endIndex</code> by inserting characters before the start index and after the
     * end index.
     *
     * <p>
     * This is well-defined only for patterns with exactly one argument.
     *
     * @param result
     *            The StringBuilder containing the value argument.
     * @param startIndex
     *            The left index of the value within the string builder.
     * @param endIndex
     *            The right index of the value within the string builder.
     * @return The number of characters (UTF-16 code points) that were added to the StringBuilder.
     */
    int32_t
    formatAsPrefixSuffix(FormattedStringBuilder& result, int32_t startIndex, int32_t endIndex,
                         UErrorCode& status) const;

    /**
     * TODO: Like above, this belongs with the rest of the SimpleFormatterImpl code.
     * I put it here so that the SimpleFormatter uses in FormattedStringBuilder are near each other.
     *
     * <p>
     * Applies the compiled two-argument pattern to the FormattedStringBuilder.
     *
     * <p>
     * This method is optimized for the case where the prefix and suffix are often empty, such as
     * in the range pattern like "{0}-{1}".
     */
    static int32_t
    formatTwoArgPattern(const SimpleFormatter& compiled, FormattedStringBuilder& result,
                        int32_t index, int32_t* outPrefixLength, int32_t* outSuffixLength,
                        Field field, UErrorCode& status);

  private:
    UnicodeString fCompiledPattern;
    Field fField;
    bool fStrong = false;
    int32_t fPrefixLength = 0;
    int32_t fSuffixOffset = -1;
    int32_t fSuffixLength = 0;
    Modifier::Parameters fParameters;
};

/**
 * An implementation of {@link Modifier} that allows for multiple types of fields in the same modifier. Constructed
 * based on the contents of two {@link FormattedStringBuilder} instances (one for the prefix, one for the suffix).
 */
class U_I18N_API ConstantMultiFieldModifier : public Modifier, public UMemory {
  public:
    ConstantMultiFieldModifier(
            const FormattedStringBuilder &prefix,
            const FormattedStringBuilder &suffix,
            bool overwrite,
            bool strong,
            const Modifier::Parameters parameters)
      : fPrefix(prefix),
        fSuffix(suffix),
        fOverwrite(overwrite),
        fStrong(strong),
        fParameters(parameters) {}

    ConstantMultiFieldModifier(
            const FormattedStringBuilder &prefix,
            const FormattedStringBuilder &suffix,
            bool overwrite,
            bool strong)
      : fPrefix(prefix),
        fSuffix(suffix),
        fOverwrite(overwrite),
        fStrong(strong) {}

    int32_t apply(FormattedStringBuilder &output, int32_t leftIndex, int32_t rightIndex,
                  UErrorCode &status) const U_OVERRIDE;

    int32_t getPrefixLength() const U_OVERRIDE;

    int32_t getCodePointCount() const U_OVERRIDE;

    bool isStrong() const U_OVERRIDE;

    bool containsField(UNumberFormatFields field) const U_OVERRIDE;

    void getParameters(Parameters& output) const U_OVERRIDE;

    bool semanticallyEquivalent(const Modifier& other) const U_OVERRIDE;

  protected:
    // NOTE: In Java, these are stored as array pointers. In C++, the FormattedStringBuilder is stored by
    // value and is treated internally as immutable.
    FormattedStringBuilder fPrefix;
    FormattedStringBuilder fSuffix;
    bool fOverwrite;
    bool fStrong;
    Modifier::Parameters fParameters;
};

/** Identical to {@link ConstantMultiFieldModifier}, but supports currency spacing. */
class U_I18N_API CurrencySpacingEnabledModifier : public ConstantMultiFieldModifier {
  public:
    /** Safe code path */
    CurrencySpacingEnabledModifier(
            const FormattedStringBuilder &prefix,
            const FormattedStringBuilder &suffix,
            bool overwrite,
            bool strong,
            const DecimalFormatSymbols &symbols,
            UErrorCode &status);

    int32_t apply(FormattedStringBuilder &output, int32_t leftIndex, int32_t rightIndex,
                  UErrorCode &status) const U_OVERRIDE;

    /** Unsafe code path */
    static int32_t
    applyCurrencySpacing(FormattedStringBuilder &output, int32_t prefixStart, int32_t prefixLen,
                         int32_t suffixStart, int32_t suffixLen, const DecimalFormatSymbols &symbols,
                         UErrorCode &status);

  private:
    UnicodeSet fAfterPrefixUnicodeSet;
    UnicodeString fAfterPrefixInsert;
    UnicodeSet fBeforeSuffixUnicodeSet;
    UnicodeString fBeforeSuffixInsert;

    enum EAffix {
        PREFIX, SUFFIX
    };

    enum EPosition {
        IN_CURRENCY, IN_NUMBER
    };

    /** Unsafe code path */
    static int32_t applyCurrencySpacingAffix(FormattedStringBuilder &output, int32_t index, EAffix affix,
                                             const DecimalFormatSymbols &symbols, UErrorCode &status);

    static UnicodeSet
    getUnicodeSet(const DecimalFormatSymbols &symbols, EPosition position, EAffix affix,
                  UErrorCode &status);

    static UnicodeString
    getInsertString(const DecimalFormatSymbols &symbols, EAffix affix, UErrorCode &status);
};

/** A Modifier that does not do anything. */
class U_I18N_API EmptyModifier : public Modifier, public UMemory {
  public:
    explicit EmptyModifier(bool isStrong) : fStrong(isStrong) {}

    int32_t apply(FormattedStringBuilder &output, int32_t leftIndex, int32_t rightIndex,
                  UErrorCode &status) const U_OVERRIDE {
        (void)output;
        (void)leftIndex;
        (void)rightIndex;
        (void)status;
        return 0;
    }

    int32_t getPrefixLength() const U_OVERRIDE {
        return 0;
    }

    int32_t getCodePointCount() const U_OVERRIDE {
        return 0;
    }

    bool isStrong() const U_OVERRIDE {
        return fStrong;
    }

    bool containsField(UNumberFormatFields field) const U_OVERRIDE {
        (void)field;
        return false;
    }

    void getParameters(Parameters& output) const U_OVERRIDE {
        output.obj = nullptr;
    }

    bool semanticallyEquivalent(const Modifier& other) const U_OVERRIDE {
        return other.getCodePointCount() == 0;
    }

  private:
    bool fStrong;
};

/**
 * This implementation of ModifierStore adopts Modifer pointers.
 */
class U_I18N_API AdoptingModifierStore : public ModifierStore, public UMemory {
  public:
    virtual ~AdoptingModifierStore();

    static constexpr StandardPlural::Form DEFAULT_STANDARD_PLURAL = StandardPlural::OTHER;

    AdoptingModifierStore() = default;

    // No copying!
    AdoptingModifierStore(const AdoptingModifierStore &other) = delete;

    /**
     * Sets the Modifier with the specified signum and plural form.
     */
    void adoptModifier(Signum signum, StandardPlural::Form plural, const Modifier *mod) {
        U_ASSERT(mods[getModIndex(signum, plural)] == nullptr);
        mods[getModIndex(signum, plural)] = mod;
    }

    /**
     * Sets the Modifier with the specified signum.
     * The modifier will apply to all plural forms.
     */
    void adoptModifierWithoutPlural(Signum signum, const Modifier *mod) {
        U_ASSERT(mods[getModIndex(signum, DEFAULT_STANDARD_PLURAL)] == nullptr);
        mods[getModIndex(signum, DEFAULT_STANDARD_PLURAL)] = mod;
    }

    /** Returns a reference to the modifier; no ownership change. */
    const Modifier *getModifier(Signum signum, StandardPlural::Form plural) const U_OVERRIDE {
        const Modifier* modifier = mods[getModIndex(signum, plural)];
        if (modifier == nullptr && plural != DEFAULT_STANDARD_PLURAL) {
            modifier = mods[getModIndex(signum, DEFAULT_STANDARD_PLURAL)];
        }
        return modifier;
    }

    /** Returns a reference to the modifier; no ownership change. */
    const Modifier *getModifierWithoutPlural(Signum signum) const {
        return mods[getModIndex(signum, DEFAULT_STANDARD_PLURAL)];
    }

  private:
    // NOTE: mods is zero-initialized (to nullptr)
    const Modifier *mods[3 * StandardPlural::COUNT] = {};

    inline static int32_t getModIndex(Signum signum, StandardPlural::Form plural) {
        U_ASSERT(signum >= -1 && signum <= 1);
        U_ASSERT(plural >= 0 && plural < StandardPlural::COUNT);
        return static_cast<int32_t>(plural) * 3 + (signum + 1);
    }
};

} // namespace impl
} // namespace number
U_NAMESPACE_END


#endif //__NUMBER_MODIFIERS_H__



// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN
namespace number {
namespace impl {
namespace roundingutils {

enum Section {
    SECTION_LOWER_EDGE = -1,
    SECTION_UPPER_EDGE = -2,
    SECTION_LOWER = 1,
    SECTION_MIDPOINT = 2,
    SECTION_UPPER = 3
};

/**
 * Converts a rounding mode and metadata about the quantity being rounded to a boolean determining
 * whether the value should be rounded toward infinity or toward zero.
 *
 * <p>The parameters are of type int because benchmarks on an x86-64 processor against OpenJDK
 * showed that ints were demonstrably faster than enums in switch statements.
 *
 * @param isEven Whether the digit immediately before the rounding magnitude is even.
 * @param isNegative Whether the quantity is negative.
 * @param section Whether the part of the quantity to the right of the rounding magnitude is
 *     exactly halfway between two digits, whether it is in the lower part (closer to zero), or
 *     whether it is in the upper part (closer to infinity). See {@link #SECTION_LOWER}, {@link
 *     #SECTION_MIDPOINT}, and {@link #SECTION_UPPER}.
 * @param roundingMode The integer version of the {@link RoundingMode}, which you can get via
 *     {@link RoundingMode#ordinal}.
 * @param status Error code, set to U_FORMAT_INEXACT_ERROR if the rounding mode is kRoundUnnecessary.
 * @return true if the number should be rounded toward zero; false if it should be rounded toward
 *     infinity.
 */
inline bool
getRoundingDirection(bool isEven, bool isNegative, Section section, RoundingMode roundingMode,
                     UErrorCode &status) {
    switch (roundingMode) {
        case RoundingMode::UNUM_ROUND_UP:
            // round away from zero
            return false;

        case RoundingMode::UNUM_ROUND_DOWN:
            // round toward zero
            return true;

        case RoundingMode::UNUM_ROUND_CEILING:
            // round toward positive infinity
            return isNegative;

        case RoundingMode::UNUM_ROUND_FLOOR:
            // round toward negative infinity
            return !isNegative;

        case RoundingMode::UNUM_ROUND_HALFUP:
            switch (section) {
                case SECTION_MIDPOINT:
                    return false;
                case SECTION_LOWER:
                    return true;
                case SECTION_UPPER:
                    return false;
                default:
                    break;
            }
            break;

        case RoundingMode::UNUM_ROUND_HALFDOWN:
            switch (section) {
                case SECTION_MIDPOINT:
                    return true;
                case SECTION_LOWER:
                    return true;
                case SECTION_UPPER:
                    return false;
                default:
                    break;
            }
            break;

        case RoundingMode::UNUM_ROUND_HALFEVEN:
            switch (section) {
                case SECTION_MIDPOINT:
                    return isEven;
                case SECTION_LOWER:
                    return true;
                case SECTION_UPPER:
                    return false;
                default:
                    break;
            }
            break;

        default:
            break;
    }

    status = U_FORMAT_INEXACT_ERROR;
    return false;
}

/**
 * Gets whether the given rounding mode's rounding boundary is at the midpoint. The rounding
 * boundary is the point at which a number switches from being rounded down to being rounded up.
 * For example, with rounding mode HALF_EVEN, HALF_UP, or HALF_DOWN, the rounding boundary is at
 * the midpoint, and this function would return true. However, for UP, DOWN, CEILING, and FLOOR,
 * the rounding boundary is at the "edge", and this function would return false.
 *
 * @param roundingMode The integer version of the {@link RoundingMode}.
 * @return true if rounding mode is HALF_EVEN, HALF_UP, or HALF_DOWN; false otherwise.
 */
inline bool roundsAtMidpoint(int roundingMode) {
    switch (roundingMode) {
        case RoundingMode::UNUM_ROUND_UP:
        case RoundingMode::UNUM_ROUND_DOWN:
        case RoundingMode::UNUM_ROUND_CEILING:
        case RoundingMode::UNUM_ROUND_FLOOR:
            return false;

        default:
            return true;
    }
}

/**
 * Computes the number of fraction digits in a double. Used for computing maxFrac for an increment.
 * Calls into the DoubleToStringConverter library to do so.
 *
 * @param singleDigit An output parameter; set to a number if that is the
 *        only digit in the double, or -1 if there is more than one digit.
 */
digits_t doubleFractionLength(double input, int8_t* singleDigit);

} // namespace roundingutils


/**
 * Encapsulates a Precision and a RoundingMode and performs rounding on a DecimalQuantity.
 *
 * This class does not exist in Java: instead, the base Precision class is used.
 */
class RoundingImpl {
  public:
    RoundingImpl() = default;  // default constructor: leaves object in undefined state

    RoundingImpl(const Precision& precision, UNumberFormatRoundingMode roundingMode,
                 const CurrencyUnit& currency, UErrorCode& status);

    static RoundingImpl passThrough();

    /** Required for ScientificFormatter */
    bool isSignificantDigits() const;

    /**
     * Rounding endpoint used by Engineering and Compact notation. Chooses the most appropriate multiplier (magnitude
     * adjustment), applies the adjustment, rounds, and returns the chosen multiplier.
     *
     * <p>
     * In most cases, this is simple. However, when rounding the number causes it to cross a multiplier boundary, we
     * need to re-do the rounding. For example, to display 999,999 in Engineering notation with 2 sigfigs, first you
     * guess the multiplier to be -3. However, then you end up getting 1000E3, which is not the correct output. You then
     * change your multiplier to be -6, and you get 1.0E6, which is correct.
     *
     * @param input The quantity to process.
     * @param producer Function to call to return a multiplier based on a magnitude.
     * @return The number of orders of magnitude the input was adjusted by this method.
     */
    int32_t
    chooseMultiplierAndApply(number::impl::DecimalQuantity &input, const number::impl::MultiplierProducer &producer,
                             UErrorCode &status);

    void apply(number::impl::DecimalQuantity &value, UErrorCode &status) const;

    /** Version of {@link #apply} that obeys minInt constraints. Used for scientific notation compatibility mode. */
    void apply(number::impl::DecimalQuantity &value, int32_t minInt, UErrorCode status);

  private:
    Precision fPrecision;
    UNumberFormatRoundingMode fRoundingMode;
    bool fPassThrough;
};


} // namespace impl
} // namespace number
U_NAMESPACE_END

#endif //__NUMBER_ROUNDINGUTILS_H__






U_NAMESPACE_BEGIN

namespace number {
namespace impl {

enum CldrPatternStyle {
    CLDR_PATTERN_STYLE_DECIMAL,
    CLDR_PATTERN_STYLE_CURRENCY,
    CLDR_PATTERN_STYLE_ACCOUNTING,
    CLDR_PATTERN_STYLE_PERCENT,
    CLDR_PATTERN_STYLE_SCIENTIFIC,
    CLDR_PATTERN_STYLE_COUNT,
};

// Namespace for naked functions
namespace utils {

inline int32_t insertDigitFromSymbols(FormattedStringBuilder& output, int32_t index, int8_t digit,
                                      const DecimalFormatSymbols& symbols, Field field,
                                      UErrorCode& status) {
    if (symbols.getCodePointZero() != -1) {
        return output.insertCodePoint(index, symbols.getCodePointZero() + digit, field, status);
    }
    return output.insert(index, symbols.getConstDigitSymbol(digit), field, status);
}

inline bool unitIsCurrency(const MeasureUnit& unit) {
    return uprv_strcmp("currency", unit.getType()) == 0;
}

inline bool unitIsNoUnit(const MeasureUnit& unit) {
    return uprv_strcmp("none", unit.getType()) == 0;
}

inline bool unitIsPercent(const MeasureUnit& unit) {
    return uprv_strcmp("percent", unit.getSubtype()) == 0;
}

inline bool unitIsPermille(const MeasureUnit& unit) {
    return uprv_strcmp("permille", unit.getSubtype()) == 0;
}

// NOTE: In Java, this method is in NumberFormat.java
const char16_t*
getPatternForStyle(const Locale& locale, const char* nsName, CldrPatternStyle style, UErrorCode& status);

/**
 * Computes the plural form for this number based on the specified set of rules.
 *
 * @param rules A {@link PluralRules} object representing the set of rules.
 * @return The {@link StandardPlural} according to the PluralRules. If the plural form is not in
 *     the set of standard plurals, {@link StandardPlural#OTHER} is returned instead.
 */
inline StandardPlural::Form getStandardPlural(const PluralRules *rules,
                                              const IFixedDecimal &fdec) {
    if (rules == nullptr) {
        // Fail gracefully if the user didn't provide a PluralRules
        return StandardPlural::Form::OTHER;
    } else {
        UnicodeString ruleString = rules->select(fdec);
        return StandardPlural::orOtherFromString(ruleString);
    }
}

/**
 * Computes the plural form after copying the number and applying rounding rules.
 */
inline StandardPlural::Form getPluralSafe(
        const RoundingImpl& rounder,
        const PluralRules* rules,
        const DecimalQuantity& dq,
        UErrorCode& status) {
    // TODO(ICU-20500): Avoid the copy?
    DecimalQuantity copy(dq);
    rounder.apply(copy, status);
    if (U_FAILURE(status)) {
        return StandardPlural::Form::OTHER;
    }
    return getStandardPlural(rules, copy);
}

} // namespace utils

} // namespace impl
} // namespace number

U_NAMESPACE_END

#endif //__NUMBER_UTILS_H__


// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING





// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// This file contains compliant implementations of FormattedValue which can be
// leveraged by ICU formatters.
//
// Each implementation is defined in its own cpp file in order to split
// dependencies more modularly.


// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html




U_NAMESPACE_BEGIN

/**
 * An internal helper class to help convert between C and C++ APIs.
 */
template<typename CType, typename CPPType, int32_t kMagic>
class IcuCApiHelper {
  public:
    /**
     * Convert from the C type to the C++ type (const version).
     */
    static const CPPType* validate(const CType* input, UErrorCode& status);

    /**
     * Convert from the C type to the C++ type (non-const version).
     */
    static CPPType* validate(CType* input, UErrorCode& status);

    /**
     * Convert from the C++ type to the C type (const version).
     */
    const CType* exportConstForC() const;

    /**
     * Convert from the C++ type to the C type (non-const version).
     */
    CType* exportForC();

    /**
     * Invalidates the object.
     */
    ~IcuCApiHelper();

  private:
    /**
     * While the object is valid, fMagic equals kMagic.
     */
    int32_t fMagic = kMagic;
};


template<typename CType, typename CPPType, int32_t kMagic>
const CPPType*
IcuCApiHelper<CType, CPPType, kMagic>::validate(const CType* input, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (input == nullptr) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    auto* impl = reinterpret_cast<const CPPType*>(input);
    if (static_cast<const IcuCApiHelper<CType, CPPType, kMagic>*>(impl)->fMagic != kMagic) {
        status = U_INVALID_FORMAT_ERROR;
        return nullptr;
    }
    return impl;
}

template<typename CType, typename CPPType, int32_t kMagic>
CPPType*
IcuCApiHelper<CType, CPPType, kMagic>::validate(CType* input, UErrorCode& status) {
    auto* constInput = static_cast<const CType*>(input);
    auto* validated = validate(constInput, status);
    return const_cast<CPPType*>(validated);
}

template<typename CType, typename CPPType, int32_t kMagic>
const CType*
IcuCApiHelper<CType, CPPType, kMagic>::exportConstForC() const {
    return reinterpret_cast<const CType*>(static_cast<const CPPType*>(this));
}

template<typename CType, typename CPPType, int32_t kMagic>
CType*
IcuCApiHelper<CType, CPPType, kMagic>::exportForC() {
    return reinterpret_cast<CType*>(static_cast<CPPType*>(this));
}

template<typename CType, typename CPPType, int32_t kMagic>
IcuCApiHelper<CType, CPPType, kMagic>::~IcuCApiHelper() {
    // head off application errors by preventing use of of deleted objects.
    fMagic = 0;
}


U_NAMESPACE_END








/**
 * Represents the type of constraint for ConstrainedFieldPosition.
 *
 * Constraints are used to control the behavior of iteration in FormattedValue.
 *
 * @internal
 */
typedef enum UCFPosConstraintType {
    /**
     * Represents the lack of a constraint.
     *
     * This is the value of fConstraint if no "constrain" methods were called.
     *
     * @internal
     */
    UCFPOS_CONSTRAINT_NONE = 0,

    /**
     * Represents that the field category is constrained.
     *
     * This is the value of fConstraint if constraintCategory was called.
     *
     * FormattedValue implementations should not change the field category
     * while this constraint is active.
     *
     * @internal
     */
    UCFPOS_CONSTRAINT_CATEGORY,

    /**
     * Represents that the field and field category are constrained.
     *
     * This is the value of fConstraint if constraintField was called.
     *
     * FormattedValue implementations should not change the field or field category
     * while this constraint is active.
     *
     * @internal
     */
    UCFPOS_CONSTRAINT_FIELD
} UCFPosConstraintType;


U_NAMESPACE_BEGIN


/**
 * Implementation of FormattedValue using FieldPositionHandler to accept fields.
 */
class FormattedValueFieldPositionIteratorImpl : public UMemory, public FormattedValue {
public:

    /** @param initialFieldCapacity Initially allocate space for this many fields. */
    FormattedValueFieldPositionIteratorImpl(int32_t initialFieldCapacity, UErrorCode& status);

    virtual ~FormattedValueFieldPositionIteratorImpl();

    // Implementation of FormattedValue (const):

    UnicodeString toString(UErrorCode& status) const U_OVERRIDE;
    UnicodeString toTempString(UErrorCode& status) const U_OVERRIDE;
    Appendable& appendTo(Appendable& appendable, UErrorCode& status) const U_OVERRIDE;
    UBool nextPosition(ConstrainedFieldPosition& cfpos, UErrorCode& status) const U_OVERRIDE;

    // Additional methods used during construction phase only (non-const):

    FieldPositionIteratorHandler getHandler(UErrorCode& status);
    void appendString(UnicodeString string, UErrorCode& status);

    /**
     * Computes the spans for duplicated values.
     * For example, if the string has fields:
     * 
     *     ...aa..[b.cc]..d.[bb.e.c]..a..
     *
     * then the spans will be the bracketed regions.
     *
     * Assumes that the currently known fields are sorted
     * and all in the same category.
     */
    void addOverlapSpans(UFieldCategory spanCategory, int8_t firstIndex, UErrorCode& status);

    /**
     * Sorts the fields: start index first, length second.
     */
    void sort();

private:
    UnicodeString fString;
    UVector32 fFields;
};


/**
 * Implementation of FormattedValue based on FormattedStringBuilder.
 *
 * The implementation currently revolves around numbers and number fields.
 * However, it can be generalized in the future when there is a need.
 *
 * @author sffc (Shane Carr)
 */
// Exported as U_I18N_API for tests
class U_I18N_API FormattedValueStringBuilderImpl : public UMemory, public FormattedValue {
public:

    FormattedValueStringBuilderImpl(FormattedStringBuilder::Field numericField);

    virtual ~FormattedValueStringBuilderImpl();

    // Implementation of FormattedValue (const):

    UnicodeString toString(UErrorCode& status) const U_OVERRIDE;
    UnicodeString toTempString(UErrorCode& status) const U_OVERRIDE;
    Appendable& appendTo(Appendable& appendable, UErrorCode& status) const U_OVERRIDE;
    UBool nextPosition(ConstrainedFieldPosition& cfpos, UErrorCode& status) const U_OVERRIDE;

    // Additional helper functions:
    UBool nextFieldPosition(FieldPosition& fp, UErrorCode& status) const;
    void getAllFieldPositions(FieldPositionIteratorHandler& fpih, UErrorCode& status) const;
    inline FormattedStringBuilder& getStringRef() {
        return fString;
    }
    inline const FormattedStringBuilder& getStringRef() const {
        return fString;
    }

private:
    FormattedStringBuilder fString;
    FormattedStringBuilder::Field fNumericField;

    bool nextPositionImpl(ConstrainedFieldPosition& cfpos, FormattedStringBuilder::Field numericField, UErrorCode& status) const;
    static bool isIntOrGroup(FormattedStringBuilder::Field field);
    static bool isNumericField(FormattedStringBuilder::Field field);
    int32_t trimBack(int32_t limit) const;
    int32_t trimFront(int32_t start) const;
};


// C API Helpers for FormattedValue
// Magic number as ASCII == "UFV"
struct UFormattedValueImpl;
typedef IcuCApiHelper<UFormattedValue, UFormattedValueImpl, 0x55465600> UFormattedValueApiHelper;
struct UFormattedValueImpl : public UMemory, public UFormattedValueApiHelper {
    // This pointer should be set by the child class.
    FormattedValue* fFormattedValue = nullptr;
};


/** Boilerplate to check for valid status before dereferencing the fData pointer. */
#define UPRV_FORMATTED_VALUE_METHOD_GUARD(returnExpression) \
    if (U_FAILURE(status)) { \
        return returnExpression; \
    } \
    if (fData == nullptr) { \
        status = fErrorCode; \
        return returnExpression; \
    } \


/** Implementation of the methods from U_FORMATTED_VALUE_SUBCLASS_AUTO. */
#define UPRV_FORMATTED_VALUE_SUBCLASS_AUTO_IMPL(Name) \
    Name::Name(Name&& src) U_NOEXCEPT \
            : fData(src.fData), fErrorCode(src.fErrorCode) { \
        src.fData = nullptr; \
        src.fErrorCode = U_INVALID_STATE_ERROR; \
    } \
    Name::~Name() { \
        delete fData; \
        fData = nullptr; \
    } \
    Name& Name::operator=(Name&& src) U_NOEXCEPT { \
        delete fData; \
        fData = src.fData; \
        src.fData = nullptr; \
        fErrorCode = src.fErrorCode; \
        src.fErrorCode = U_INVALID_STATE_ERROR; \
        return *this; \
    } \
    UnicodeString Name::toString(UErrorCode& status) const { \
        UPRV_FORMATTED_VALUE_METHOD_GUARD(ICU_Utility::makeBogusString()) \
        return fData->toString(status); \
    } \
    UnicodeString Name::toTempString(UErrorCode& status) const { \
        UPRV_FORMATTED_VALUE_METHOD_GUARD(ICU_Utility::makeBogusString()) \
        return fData->toTempString(status); \
    } \
    Appendable& Name::appendTo(Appendable& appendable, UErrorCode& status) const { \
        UPRV_FORMATTED_VALUE_METHOD_GUARD(appendable) \
        return fData->appendTo(appendable, status); \
    } \
    UBool Name::nextPosition(ConstrainedFieldPosition& cfpos, UErrorCode& status) const { \
        UPRV_FORMATTED_VALUE_METHOD_GUARD(FALSE) \
        return fData->nextPosition(cfpos, status); \
    }


/** Like UPRV_FORMATTED_VALUE_CAPI_AUTO_IMPL but without impl type declarations. */
#define UPRV_FORMATTED_VALUE_CAPI_NO_IMPLTYPE_AUTO_IMPL(CType, ImplType, HelperType, Prefix) \
    U_CAPI CType* U_EXPORT2 \
    Prefix ## _openResult (UErrorCode* ec) { \
        if (U_FAILURE(*ec)) { \
            return nullptr; \
        } \
        ImplType* impl = new ImplType(); \
        if (impl == nullptr) { \
            *ec = U_MEMORY_ALLOCATION_ERROR; \
            return nullptr; \
        } \
        return static_cast<HelperType*>(impl)->exportForC(); \
    } \
    U_DRAFT const UFormattedValue* U_EXPORT2 \
    Prefix ## _resultAsValue (const CType* uresult, UErrorCode* ec) { \
        const ImplType* result = HelperType::validate(uresult, *ec); \
        if (U_FAILURE(*ec)) { return nullptr; } \
        return static_cast<const UFormattedValueApiHelper*>(result)->exportConstForC(); \
    } \
    U_CAPI void U_EXPORT2 \
    Prefix ## _closeResult (CType* uresult) { \
        UErrorCode localStatus = U_ZERO_ERROR; \
        const ImplType* impl = HelperType::validate(uresult, localStatus); \
        delete impl; \
    }


/**
 * Implementation of the standard methods for a UFormattedValue "subclass" C API.
 * @param CPPType The public C++ type, like FormattedList
 * @param CType The public C type, like UFormattedList
 * @param ImplType A name to use for the implementation class
 * @param HelperType A name to use for the "mixin" typedef for C API conversion
 * @param Prefix The C API prefix, like ulistfmt
 * @param MagicNumber A unique 32-bit number to use to identify this type
 */
#define UPRV_FORMATTED_VALUE_CAPI_AUTO_IMPL(CPPType, CType, ImplType, HelperType, Prefix, MagicNumber) \
    U_NAMESPACE_BEGIN \
    class ImplType; \
    typedef IcuCApiHelper<CType, ImplType, MagicNumber> HelperType; \
    class ImplType : public UFormattedValueImpl, public HelperType { \
    public: \
        ImplType(); \
        ~ImplType(); \
        CPPType fImpl; \
    }; \
    ImplType::ImplType() { \
        fFormattedValue = &fImpl; \
    } \
    ImplType::~ImplType() {} \
    U_NAMESPACE_END \
    UPRV_FORMATTED_VALUE_CAPI_NO_IMPLTYPE_AUTO_IMPL(CType, ImplType, HelperType, Prefix)


U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */


U_NAMESPACE_BEGIN namespace number {
namespace impl {


/** Helper function used in upluralrules.cpp */
const DecimalQuantity* validateUFormattedNumberToDecimalQuantity(
    const UFormattedNumber* uresult, UErrorCode& status);


/**
 * Struct for data used by FormattedNumber.
 *
 * This struct is held internally by the C++ version FormattedNumber since the member types are not
 * declared in the public header file.
 *
 * The DecimalQuantity is not currently being used by FormattedNumber, but at some point it could be used
 * to add a toDecNumber() or similar method.
 */
class UFormattedNumberData : public FormattedValueStringBuilderImpl {
public:
    UFormattedNumberData() : FormattedValueStringBuilderImpl(0) {}
    virtual ~UFormattedNumberData();

    DecimalQuantity quantity;
};


} // namespace impl
} // namespace number
U_NAMESPACE_END

#endif //__SOURCE_NUMBER_UTYPES_H__


using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;
using namespace icu::numparse;
using namespace icu::numparse::impl;
using ERoundingMode = icu::DecimalFormat::ERoundingMode;
using EPadPosition = icu::DecimalFormat::EPadPosition;

// MSVC VS2015 warns C4805 when comparing bool with UBool, VS2017 no longer emits this warning.
// TODO: Move this macro into a better place?
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN
#define UBOOL_TO_BOOL(b) static_cast<bool>(b)
#else
#define UBOOL_TO_BOOL(b) b
#endif


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DecimalFormat)


DecimalFormat::DecimalFormat(UErrorCode& status)
        : DecimalFormat(nullptr, status) {
    if (U_FAILURE(status)) { return; }
    // Use the default locale and decimal pattern.
    const char* localeName = Locale::getDefault().getName();
    LocalPointer<NumberingSystem> ns(NumberingSystem::createInstance(status));
    UnicodeString patternString = utils::getPatternForStyle(
            localeName,
            ns->getName(),
            CLDR_PATTERN_STYLE_DECIMAL,
            status);
    setPropertiesFromPattern(patternString, IGNORE_ROUNDING_IF_CURRENCY, status);
    touch(status);
}

DecimalFormat::DecimalFormat(const UnicodeString& pattern, UErrorCode& status)
        : DecimalFormat(nullptr, status) {
    if (U_FAILURE(status)) { return; }
    setPropertiesFromPattern(pattern, IGNORE_ROUNDING_IF_CURRENCY, status);
    touch(status);
}

DecimalFormat::DecimalFormat(const UnicodeString& pattern, DecimalFormatSymbols* symbolsToAdopt,
                             UErrorCode& status)
        : DecimalFormat(symbolsToAdopt, status) {
    if (U_FAILURE(status)) { return; }
    setPropertiesFromPattern(pattern, IGNORE_ROUNDING_IF_CURRENCY, status);
    touch(status);
}

DecimalFormat::DecimalFormat(const UnicodeString& pattern, DecimalFormatSymbols* symbolsToAdopt,
                             UNumberFormatStyle style, UErrorCode& status)
        : DecimalFormat(symbolsToAdopt, status) {
    if (U_FAILURE(status)) { return; }
    // If choice is a currency type, ignore the rounding information.
    if (style == UNumberFormatStyle::UNUM_CURRENCY ||
        style == UNumberFormatStyle::UNUM_CURRENCY_ISO ||
        style == UNumberFormatStyle::UNUM_CURRENCY_ACCOUNTING ||
        style == UNumberFormatStyle::UNUM_CASH_CURRENCY ||
        style == UNumberFormatStyle::UNUM_CURRENCY_STANDARD ||
        style == UNumberFormatStyle::UNUM_CURRENCY_PLURAL) {
        setPropertiesFromPattern(pattern, IGNORE_ROUNDING_ALWAYS, status);
    } else {
        setPropertiesFromPattern(pattern, IGNORE_ROUNDING_IF_CURRENCY, status);
    }
    // Note: in Java, CurrencyPluralInfo is set in NumberFormat.java, but in C++, it is not set there,
    // so we have to set it here.
    if (style == UNumberFormatStyle::UNUM_CURRENCY_PLURAL) {
        LocalPointer<CurrencyPluralInfo> cpi(
                new CurrencyPluralInfo(fields->symbols->getLocale(), status),
                status);
        if (U_FAILURE(status)) { return; }
        fields->properties.currencyPluralInfo.fPtr.adoptInstead(cpi.orphan());
    }
    touch(status);
}

DecimalFormat::DecimalFormat(const DecimalFormatSymbols* symbolsToAdopt, UErrorCode& status) {
    // we must take ownership of symbolsToAdopt, even in a failure case.
    LocalPointer<const DecimalFormatSymbols> adoptedSymbols(symbolsToAdopt);
    if (U_FAILURE(status)) {
        return;
    }
    fields = new DecimalFormatFields();
    if (fields == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    if (adoptedSymbols.isNull()) {
        fields->symbols.adoptInsteadAndCheckErrorCode(new DecimalFormatSymbols(status), status);
    } else {
        fields->symbols.adoptInsteadAndCheckErrorCode(adoptedSymbols.orphan(), status);
    }
    if (U_FAILURE(status)) {
        delete fields;
        fields = nullptr;
    }
}

#if UCONFIG_HAVE_PARSEALLINPUT

void DecimalFormat::setParseAllInput(UNumberFormatAttributeValue value) {
    if (fields == nullptr) { return; }
    if (value == fields->properties.parseAllInput) { return; }
    fields->properties.parseAllInput = value;
}

#endif

DecimalFormat&
DecimalFormat::setAttribute(UNumberFormatAttribute attr, int32_t newValue, UErrorCode& status) {
    if (U_FAILURE(status)) { return *this; }

    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        return *this;
    }

    switch (attr) {
        case UNUM_LENIENT_PARSE:
            setLenient(newValue != 0);
            break;

        case UNUM_PARSE_INT_ONLY:
            setParseIntegerOnly(newValue != 0);
            break;

        case UNUM_GROUPING_USED:
            setGroupingUsed(newValue != 0);
            break;

        case UNUM_DECIMAL_ALWAYS_SHOWN:
            setDecimalSeparatorAlwaysShown(newValue != 0);
            break;

        case UNUM_MAX_INTEGER_DIGITS:
            setMaximumIntegerDigits(newValue);
            break;

        case UNUM_MIN_INTEGER_DIGITS:
            setMinimumIntegerDigits(newValue);
            break;

        case UNUM_INTEGER_DIGITS:
            setMinimumIntegerDigits(newValue);
            setMaximumIntegerDigits(newValue);
            break;

        case UNUM_MAX_FRACTION_DIGITS:
            setMaximumFractionDigits(newValue);
            break;

        case UNUM_MIN_FRACTION_DIGITS:
            setMinimumFractionDigits(newValue);
            break;

        case UNUM_FRACTION_DIGITS:
            setMinimumFractionDigits(newValue);
            setMaximumFractionDigits(newValue);
            break;

        case UNUM_SIGNIFICANT_DIGITS_USED:
            setSignificantDigitsUsed(newValue != 0);
            break;

        case UNUM_MAX_SIGNIFICANT_DIGITS:
            setMaximumSignificantDigits(newValue);
            break;

        case UNUM_MIN_SIGNIFICANT_DIGITS:
            setMinimumSignificantDigits(newValue);
            break;

        case UNUM_MULTIPLIER:
            setMultiplier(newValue);
            break;

        case UNUM_SCALE:
            setMultiplierScale(newValue);
            break;

        case UNUM_GROUPING_SIZE:
            setGroupingSize(newValue);
            break;

        case UNUM_ROUNDING_MODE:
            setRoundingMode((DecimalFormat::ERoundingMode) newValue);
            break;

        case UNUM_FORMAT_WIDTH:
            setFormatWidth(newValue);
            break;

        case UNUM_PADDING_POSITION:
            /** The position at which padding will take place. */
            setPadPosition((DecimalFormat::EPadPosition) newValue);
            break;

        case UNUM_SECONDARY_GROUPING_SIZE:
            setSecondaryGroupingSize(newValue);
            break;

#if UCONFIG_HAVE_PARSEALLINPUT
        case UNUM_PARSE_ALL_INPUT:
            setParseAllInput((UNumberFormatAttributeValue) newValue);
            break;
#endif

        case UNUM_PARSE_NO_EXPONENT:
            setParseNoExponent((UBool) newValue);
            break;

        case UNUM_PARSE_DECIMAL_MARK_REQUIRED:
            setDecimalPatternMatchRequired((UBool) newValue);
            break;

        case UNUM_CURRENCY_USAGE:
            setCurrencyUsage((UCurrencyUsage) newValue, &status);
            break;

        case UNUM_MINIMUM_GROUPING_DIGITS:
            setMinimumGroupingDigits(newValue);
            break;

        case UNUM_PARSE_CASE_SENSITIVE:
            setParseCaseSensitive(static_cast<UBool>(newValue));
            break;

        case UNUM_SIGN_ALWAYS_SHOWN:
            setSignAlwaysShown(static_cast<UBool>(newValue));
            break;

        case UNUM_FORMAT_FAIL_IF_MORE_THAN_MAX_DIGITS:
            setFormatFailIfMoreThanMaxDigits(static_cast<UBool>(newValue));
            break;

        default:
            status = U_UNSUPPORTED_ERROR;
            break;
    }
    return *this;
}

int32_t DecimalFormat::getAttribute(UNumberFormatAttribute attr, UErrorCode& status) const {
    if (U_FAILURE(status)) { return -1; }

    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        return -1;
    }

    switch (attr) {
        case UNUM_LENIENT_PARSE:
            return isLenient();

        case UNUM_PARSE_INT_ONLY:
            return isParseIntegerOnly();

        case UNUM_GROUPING_USED:
            return isGroupingUsed();

        case UNUM_DECIMAL_ALWAYS_SHOWN:
            return isDecimalSeparatorAlwaysShown();

        case UNUM_MAX_INTEGER_DIGITS:
            return getMaximumIntegerDigits();

        case UNUM_MIN_INTEGER_DIGITS:
            return getMinimumIntegerDigits();

        case UNUM_INTEGER_DIGITS:
            // TBD: what should this return?
            return getMinimumIntegerDigits();

        case UNUM_MAX_FRACTION_DIGITS:
            return getMaximumFractionDigits();

        case UNUM_MIN_FRACTION_DIGITS:
            return getMinimumFractionDigits();

        case UNUM_FRACTION_DIGITS:
            // TBD: what should this return?
            return getMinimumFractionDigits();

        case UNUM_SIGNIFICANT_DIGITS_USED:
            return areSignificantDigitsUsed();

        case UNUM_MAX_SIGNIFICANT_DIGITS:
            return getMaximumSignificantDigits();

        case UNUM_MIN_SIGNIFICANT_DIGITS:
            return getMinimumSignificantDigits();

        case UNUM_MULTIPLIER:
            return getMultiplier();

        case UNUM_SCALE:
            return getMultiplierScale();

        case UNUM_GROUPING_SIZE:
            return getGroupingSize();

        case UNUM_ROUNDING_MODE:
            return getRoundingMode();

        case UNUM_FORMAT_WIDTH:
            return getFormatWidth();

        case UNUM_PADDING_POSITION:
            return getPadPosition();

        case UNUM_SECONDARY_GROUPING_SIZE:
            return getSecondaryGroupingSize();

        case UNUM_PARSE_NO_EXPONENT:
            return isParseNoExponent();

        case UNUM_PARSE_DECIMAL_MARK_REQUIRED:
            return isDecimalPatternMatchRequired();

        case UNUM_CURRENCY_USAGE:
            return getCurrencyUsage();

        case UNUM_MINIMUM_GROUPING_DIGITS:
            return getMinimumGroupingDigits();

        case UNUM_PARSE_CASE_SENSITIVE:
            return isParseCaseSensitive();

        case UNUM_SIGN_ALWAYS_SHOWN:
            return isSignAlwaysShown();

        case UNUM_FORMAT_FAIL_IF_MORE_THAN_MAX_DIGITS:
            return isFormatFailIfMoreThanMaxDigits();

        default:
            status = U_UNSUPPORTED_ERROR;
            break;
    }

    return -1; /* undefined */
}

void DecimalFormat::setGroupingUsed(UBool enabled) {
    if (fields == nullptr) {
        return;
    }
    if (UBOOL_TO_BOOL(enabled) == fields->properties.groupingUsed) { return; }
    NumberFormat::setGroupingUsed(enabled); // to set field for compatibility
    fields->properties.groupingUsed = enabled;
    touchNoError();
}

void DecimalFormat::setParseIntegerOnly(UBool value) {
    if (fields == nullptr) {
        return;
    }
    if (UBOOL_TO_BOOL(value) == fields->properties.parseIntegerOnly) { return; }
    NumberFormat::setParseIntegerOnly(value); // to set field for compatibility
    fields->properties.parseIntegerOnly = value;
    touchNoError();
}

void DecimalFormat::setLenient(UBool enable) {
    if (fields == nullptr) {
        return;
    }
    ParseMode mode = enable ? PARSE_MODE_LENIENT : PARSE_MODE_STRICT;
    if (!fields->properties.parseMode.isNull() && mode == fields->properties.parseMode.getNoError()) { return; }
    NumberFormat::setLenient(enable); // to set field for compatibility
    fields->properties.parseMode = mode;
    touchNoError();
}

DecimalFormat::DecimalFormat(const UnicodeString& pattern, DecimalFormatSymbols* symbolsToAdopt,
                             UParseError&, UErrorCode& status)
        : DecimalFormat(symbolsToAdopt, status) {
    if (U_FAILURE(status)) { return; }
    // TODO: What is parseError for?
    setPropertiesFromPattern(pattern, IGNORE_ROUNDING_IF_CURRENCY, status);
    touch(status);
}

DecimalFormat::DecimalFormat(const UnicodeString& pattern, const DecimalFormatSymbols& symbols,
                             UErrorCode& status)
        : DecimalFormat(nullptr, status) {
    if (U_FAILURE(status)) { return; }
    LocalPointer<DecimalFormatSymbols> dfs(new DecimalFormatSymbols(symbols), status);
    if (U_FAILURE(status)) {
        // If we failed to allocate DecimalFormatSymbols, then release fields and its members.
        // We must have a fully complete fields object, we cannot have partially populated members.
        delete fields;
        fields = nullptr;
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    fields->symbols.adoptInstead(dfs.orphan());
    setPropertiesFromPattern(pattern, IGNORE_ROUNDING_IF_CURRENCY, status);
    touch(status);
}

DecimalFormat::DecimalFormat(const DecimalFormat& source) : NumberFormat(source) {
    // If the object that we are copying from is invalid, no point in going further.
    if (source.fields == nullptr) {
        return;
    }
    // Note: it is not safe to copy fields->formatter or fWarehouse directly because fields->formatter might have
    // dangling pointers to fields inside fWarehouse. The safe thing is to re-construct fields->formatter from
    // the property bag, despite being somewhat slower.
    fields = new DecimalFormatFields(source.fields->properties);
    if (fields == nullptr) {
        return; // no way to report an error.
    }
    UErrorCode status = U_ZERO_ERROR;
    fields->symbols.adoptInsteadAndCheckErrorCode(new DecimalFormatSymbols(*source.fields->symbols), status);
    // In order to simplify error handling logic in the various getters/setters/etc, we do not allow
    // any partially populated DecimalFormatFields object. We must have a fully complete fields object
    // or else we set it to nullptr.
    if (U_FAILURE(status)) {
        delete fields;
        fields = nullptr;
        return;
    }
    touch(status);
}

DecimalFormat& DecimalFormat::operator=(const DecimalFormat& rhs) {
    // guard against self-assignment
    if (this == &rhs) {
        return *this;
    }
    // Make sure both objects are valid.
    if (fields == nullptr || rhs.fields == nullptr) {
        return *this; // unfortunately, no way to report an error.
    }
    fields->properties = rhs.fields->properties;
    fields->exportedProperties.clear();
    UErrorCode status = U_ZERO_ERROR;
    LocalPointer<DecimalFormatSymbols> dfs(new DecimalFormatSymbols(*rhs.fields->symbols), status);
    if (U_FAILURE(status)) {
        // We failed to allocate DecimalFormatSymbols, release fields and its members.
        // We must have a fully complete fields object, we cannot have partially populated members.
        delete fields;
        fields = nullptr;
        return *this;
    }
    fields->symbols.adoptInstead(dfs.orphan());
    touch(status);

    return *this;
}

DecimalFormat::~DecimalFormat() {
    if (fields == nullptr) { return; }

    delete fields->atomicParser.exchange(nullptr);
    delete fields->atomicCurrencyParser.exchange(nullptr);
    delete fields;
}

DecimalFormat* DecimalFormat::clone() const {
    // can only clone valid objects.
    if (fields == nullptr) {
        return nullptr;
    }
    LocalPointer<DecimalFormat> df(new DecimalFormat(*this));
    if (df.isValid() && df->fields != nullptr) {
        return df.orphan();
    }
    return nullptr;
}

UBool DecimalFormat::operator==(const Format& other) const {
    auto* otherDF = dynamic_cast<const DecimalFormat*>(&other);
    if (otherDF == nullptr) {
        return false;
    }
    // If either object is in an invalid state, prevent dereferencing nullptr below.
    // Additionally, invalid objects should not be considered equal to anything.
    if (fields == nullptr || otherDF->fields == nullptr) {
        return false;
    }
    return fields->properties == otherDF->fields->properties && *fields->symbols == *otherDF->fields->symbols;
}

UnicodeString& DecimalFormat::format(double number, UnicodeString& appendTo, FieldPosition& pos) const {
    if (fields == nullptr) {
        appendTo.setToBogus();
        return appendTo;
    }
    if (pos.getField() == FieldPosition::DONT_CARE && fastFormatDouble(number, appendTo)) {
        return appendTo;
    }
    UErrorCode localStatus = U_ZERO_ERROR;
    FormattedNumber output = fields->formatter.formatDouble(number, localStatus);
    fieldPositionHelper(output, pos, appendTo.length(), localStatus);
    auto appendable = UnicodeStringAppendable(appendTo);
    output.appendTo(appendable, localStatus);
    return appendTo;
}

UnicodeString& DecimalFormat::format(double number, UnicodeString& appendTo, FieldPosition& pos,
                                     UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo; // don't overwrite status if it's already a failure.
    }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        appendTo.setToBogus();
        return appendTo;
    }
    if (pos.getField() == FieldPosition::DONT_CARE && fastFormatDouble(number, appendTo)) {
        return appendTo;
    }
    FormattedNumber output = fields->formatter.formatDouble(number, status);
    fieldPositionHelper(output, pos, appendTo.length(), status);
    auto appendable = UnicodeStringAppendable(appendTo);
    output.appendTo(appendable, status);
    return appendTo;
}

UnicodeString&
DecimalFormat::format(double number, UnicodeString& appendTo, FieldPositionIterator* posIter,
                      UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo; // don't overwrite status if it's already a failure.
    }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        appendTo.setToBogus();
        return appendTo;
    }
    if (posIter == nullptr && fastFormatDouble(number, appendTo)) {
        return appendTo;
    }
    FormattedNumber output = fields->formatter.formatDouble(number, status);
    fieldPositionIteratorHelper(output, posIter, appendTo.length(), status);
    auto appendable = UnicodeStringAppendable(appendTo);
    output.appendTo(appendable, status);
    return appendTo;
}

UnicodeString& DecimalFormat::format(int32_t number, UnicodeString& appendTo, FieldPosition& pos) const {
    return format(static_cast<int64_t> (number), appendTo, pos);
}

UnicodeString& DecimalFormat::format(int32_t number, UnicodeString& appendTo, FieldPosition& pos,
                                     UErrorCode& status) const {
    return format(static_cast<int64_t> (number), appendTo, pos, status);
}

UnicodeString&
DecimalFormat::format(int32_t number, UnicodeString& appendTo, FieldPositionIterator* posIter,
                      UErrorCode& status) const {
    return format(static_cast<int64_t> (number), appendTo, posIter, status);
}

UnicodeString& DecimalFormat::format(int64_t number, UnicodeString& appendTo, FieldPosition& pos) const {
    if (fields == nullptr) {
        appendTo.setToBogus();
        return appendTo;
    }
    if (pos.getField() == FieldPosition::DONT_CARE && fastFormatInt64(number, appendTo)) {
        return appendTo;
    }
    UErrorCode localStatus = U_ZERO_ERROR;
    FormattedNumber output = fields->formatter.formatInt(number, localStatus);
    fieldPositionHelper(output, pos, appendTo.length(), localStatus);
    auto appendable = UnicodeStringAppendable(appendTo);
    output.appendTo(appendable, localStatus);
    return appendTo;
}

UnicodeString& DecimalFormat::format(int64_t number, UnicodeString& appendTo, FieldPosition& pos,
                                     UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo; // don't overwrite status if it's already a failure.
    }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        appendTo.setToBogus();
        return appendTo;
    }
    if (pos.getField() == FieldPosition::DONT_CARE && fastFormatInt64(number, appendTo)) {
        return appendTo;
    }
    FormattedNumber output = fields->formatter.formatInt(number, status);
    fieldPositionHelper(output, pos, appendTo.length(), status);
    auto appendable = UnicodeStringAppendable(appendTo);
    output.appendTo(appendable, status);
    return appendTo;
}

UnicodeString&
DecimalFormat::format(int64_t number, UnicodeString& appendTo, FieldPositionIterator* posIter,
                      UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo; // don't overwrite status if it's already a failure.
    }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        appendTo.setToBogus();
        return appendTo;
    }
    if (posIter == nullptr && fastFormatInt64(number, appendTo)) {
        return appendTo;
    }
    FormattedNumber output = fields->formatter.formatInt(number, status);
    fieldPositionIteratorHelper(output, posIter, appendTo.length(), status);
    auto appendable = UnicodeStringAppendable(appendTo);
    output.appendTo(appendable, status);
    return appendTo;
}

UnicodeString&
DecimalFormat::format(StringPiece number, UnicodeString& appendTo, FieldPositionIterator* posIter,
                      UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo; // don't overwrite status if it's already a failure.
    }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        appendTo.setToBogus();
        return appendTo;
    }
    FormattedNumber output = fields->formatter.formatDecimal(number, status);
    fieldPositionIteratorHelper(output, posIter, appendTo.length(), status);
    auto appendable = UnicodeStringAppendable(appendTo);
    output.appendTo(appendable, status);
    return appendTo;
}

UnicodeString& DecimalFormat::format(const DecimalQuantity& number, UnicodeString& appendTo,
                                     FieldPositionIterator* posIter, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo; // don't overwrite status if it's already a failure.
    }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        appendTo.setToBogus();
        return appendTo;
    }
    FormattedNumber output = fields->formatter.formatDecimalQuantity(number, status);
    fieldPositionIteratorHelper(output, posIter, appendTo.length(), status);
    auto appendable = UnicodeStringAppendable(appendTo);
    output.appendTo(appendable, status);
    return appendTo;
}

UnicodeString&
DecimalFormat::format(const DecimalQuantity& number, UnicodeString& appendTo, FieldPosition& pos,
                      UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo; // don't overwrite status if it's already a failure.
    }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        appendTo.setToBogus();
        return appendTo;
    }
    FormattedNumber output = fields->formatter.formatDecimalQuantity(number, status);
    fieldPositionHelper(output, pos, appendTo.length(), status);
    auto appendable = UnicodeStringAppendable(appendTo);
    output.appendTo(appendable, status);
    return appendTo;
}

void DecimalFormat::parse(const UnicodeString& text, Formattable& output,
                          ParsePosition& parsePosition) const {
    if (fields == nullptr) {
        return;
    }
    if (parsePosition.getIndex() < 0 || parsePosition.getIndex() >= text.length()) {
        if (parsePosition.getIndex() == text.length()) {
            // If there is nothing to parse, it is an error
            parsePosition.setErrorIndex(parsePosition.getIndex());
        }
        return;
    }

    ErrorCode status;
    ParsedNumber result;
    // Note: if this is a currency instance, currencies will be matched despite the fact that we are not in the
    // parseCurrency method (backwards compatibility)
    int32_t startIndex = parsePosition.getIndex();
    const NumberParserImpl* parser = getParser(status);
    if (U_FAILURE(status)) {
        return; // unfortunately no way to report back the error.
    }
    parser->parse(text, startIndex, true, result, status);
    if (U_FAILURE(status)) {
        return; // unfortunately no way to report back the error.
    }
    // TODO: Do we need to check for fImpl->properties->parseAllInput (UCONFIG_HAVE_PARSEALLINPUT) here?
    if (result.success()) {
        parsePosition.setIndex(result.charEnd);
        result.populateFormattable(output, parser->getParseFlags());
    } else {
        parsePosition.setErrorIndex(startIndex + result.charEnd);
    }
}

CurrencyAmount* DecimalFormat::parseCurrency(const UnicodeString& text, ParsePosition& parsePosition) const {
    if (fields == nullptr) {
        return nullptr;
    }
    if (parsePosition.getIndex() < 0 || parsePosition.getIndex() >= text.length()) {
        return nullptr;
    }

    ErrorCode status;
    ParsedNumber result;
    // Note: if this is a currency instance, currencies will be matched despite the fact that we are not in the
    // parseCurrency method (backwards compatibility)
    int32_t startIndex = parsePosition.getIndex();
    const NumberParserImpl* parser = getCurrencyParser(status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    parser->parse(text, startIndex, true, result, status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    // TODO: Do we need to check for fImpl->properties->parseAllInput (UCONFIG_HAVE_PARSEALLINPUT) here?
    if (result.success()) {
        parsePosition.setIndex(result.charEnd);
        Formattable formattable;
        result.populateFormattable(formattable, parser->getParseFlags());
        LocalPointer<CurrencyAmount> currencyAmount(
            new CurrencyAmount(formattable, result.currencyCode, status), status);
        if (U_FAILURE(status)) {
            return nullptr;
        }
        return currencyAmount.orphan();
    } else {
        parsePosition.setErrorIndex(startIndex + result.charEnd);
        return nullptr;
    }
}

const DecimalFormatSymbols* DecimalFormat::getDecimalFormatSymbols(void) const {
    if (fields == nullptr) {
        return nullptr;
    }
    return fields->symbols.getAlias();
}

void DecimalFormat::adoptDecimalFormatSymbols(DecimalFormatSymbols* symbolsToAdopt) {
    if (symbolsToAdopt == nullptr) {
        return; // do not allow caller to set fields->symbols to NULL
    }
    // we must take ownership of symbolsToAdopt, even in a failure case.
    LocalPointer<DecimalFormatSymbols> dfs(symbolsToAdopt);
    if (fields == nullptr) {
        return;
    }
    fields->symbols.adoptInstead(dfs.orphan());
    touchNoError();
}

void DecimalFormat::setDecimalFormatSymbols(const DecimalFormatSymbols& symbols) {
    if (fields == nullptr) {
        return;
    }
    UErrorCode status = U_ZERO_ERROR;
    LocalPointer<DecimalFormatSymbols> dfs(new DecimalFormatSymbols(symbols), status);
    if (U_FAILURE(status)) {
        // We failed to allocate DecimalFormatSymbols, release fields and its members.
        // We must have a fully complete fields object, we cannot have partially populated members.
        delete fields;
        fields = nullptr;
        return;
    }
    fields->symbols.adoptInstead(dfs.orphan());
    touchNoError();
}

const CurrencyPluralInfo* DecimalFormat::getCurrencyPluralInfo(void) const {
    if (fields == nullptr) {
        return nullptr;
    }
    return fields->properties.currencyPluralInfo.fPtr.getAlias();
}

void DecimalFormat::adoptCurrencyPluralInfo(CurrencyPluralInfo* toAdopt) {
    // TODO: should we guard against nullptr input, like in adoptDecimalFormatSymbols?
    // we must take ownership of toAdopt, even in a failure case.
    LocalPointer<CurrencyPluralInfo> cpi(toAdopt);
    if (fields == nullptr) {
        return;
    }
    fields->properties.currencyPluralInfo.fPtr.adoptInstead(cpi.orphan());
    touchNoError();
}

void DecimalFormat::setCurrencyPluralInfo(const CurrencyPluralInfo& info) {
    if (fields == nullptr) {
        return;
    }
    if (fields->properties.currencyPluralInfo.fPtr.isNull()) {
        // Note: clone() can fail with OOM error, but we have no way to report it. :(
        fields->properties.currencyPluralInfo.fPtr.adoptInstead(info.clone());
    } else {
        *fields->properties.currencyPluralInfo.fPtr = info; // copy-assignment operator
    }
    touchNoError();
}

UnicodeString& DecimalFormat::getPositivePrefix(UnicodeString& result) const {
    if (fields == nullptr) {
        result.setToBogus();
        return result;
    }
    UErrorCode status = U_ZERO_ERROR;
    fields->formatter.getAffixImpl(true, false, result, status);
    if (U_FAILURE(status)) { result.setToBogus(); }
    return result;
}

void DecimalFormat::setPositivePrefix(const UnicodeString& newValue) {
    if (fields == nullptr) {
        return;
    }
    if (newValue == fields->properties.positivePrefix) { return; }
    fields->properties.positivePrefix = newValue;
    touchNoError();
}

UnicodeString& DecimalFormat::getNegativePrefix(UnicodeString& result) const {
    if (fields == nullptr) {
        result.setToBogus();
        return result;
    }
    UErrorCode status = U_ZERO_ERROR;
    fields->formatter.getAffixImpl(true, true, result, status);
    if (U_FAILURE(status)) { result.setToBogus(); }
    return result;
}

void DecimalFormat::setNegativePrefix(const UnicodeString& newValue) {
    if (fields == nullptr) {
        return;
    }
    if (newValue == fields->properties.negativePrefix) { return; }
    fields->properties.negativePrefix = newValue;
    touchNoError();
}

UnicodeString& DecimalFormat::getPositiveSuffix(UnicodeString& result) const {
    if (fields == nullptr) {
        result.setToBogus();
        return result;
    }
    UErrorCode status = U_ZERO_ERROR;
    fields->formatter.getAffixImpl(false, false, result, status);
    if (U_FAILURE(status)) { result.setToBogus(); }
    return result;
}

void DecimalFormat::setPositiveSuffix(const UnicodeString& newValue) {
    if (fields == nullptr) {
        return;
    }
    if (newValue == fields->properties.positiveSuffix) { return; }
    fields->properties.positiveSuffix = newValue;
    touchNoError();
}

UnicodeString& DecimalFormat::getNegativeSuffix(UnicodeString& result) const {
    if (fields == nullptr) {
        result.setToBogus();
        return result;
    }
    UErrorCode status = U_ZERO_ERROR;
    fields->formatter.getAffixImpl(false, true, result, status);
    if (U_FAILURE(status)) { result.setToBogus(); }
    return result;
}

void DecimalFormat::setNegativeSuffix(const UnicodeString& newValue) {
    if (fields == nullptr) {
        return;
    }
    if (newValue == fields->properties.negativeSuffix) { return; }
    fields->properties.negativeSuffix = newValue;
    touchNoError();
}

UBool DecimalFormat::isSignAlwaysShown() const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        return DecimalFormatProperties::getDefault().signAlwaysShown;
    }
    return fields->properties.signAlwaysShown;
}

void DecimalFormat::setSignAlwaysShown(UBool value) {
    if (fields == nullptr) { return; }
    if (UBOOL_TO_BOOL(value) == fields->properties.signAlwaysShown) { return; }
    fields->properties.signAlwaysShown = value;
    touchNoError();
}

int32_t DecimalFormat::getMultiplier(void) const {
    const DecimalFormatProperties *dfp;
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        dfp = &(DecimalFormatProperties::getDefault());
    } else {
        dfp = &fields->properties;
    }
    if (dfp->multiplier != 1) {
        return dfp->multiplier;
    } else if (dfp->magnitudeMultiplier != 0) {
        return static_cast<int32_t>(uprv_pow10(dfp->magnitudeMultiplier));
    } else {
        return 1;
    }
}

void DecimalFormat::setMultiplier(int32_t multiplier) {
    if (fields == nullptr) {
         return;
    }
    if (multiplier == 0) {
        multiplier = 1;     // one being the benign default value for a multiplier.
    }

    // Try to convert to a magnitude multiplier first
    int delta = 0;
    int value = multiplier;
    while (value != 1) {
        delta++;
        int temp = value / 10;
        if (temp * 10 != value) {
            delta = -1;
            break;
        }
        value = temp;
    }
    if (delta != -1) {
        fields->properties.magnitudeMultiplier = delta;
        fields->properties.multiplier = 1;
    } else {
        fields->properties.magnitudeMultiplier = 0;
        fields->properties.multiplier = multiplier;
    }
    touchNoError();
}

int32_t DecimalFormat::getMultiplierScale() const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().multiplierScale;
    }
    return fields->properties.multiplierScale;
}

void DecimalFormat::setMultiplierScale(int32_t newValue) {
    if (fields == nullptr) { return; }
    if (newValue == fields->properties.multiplierScale) { return; }
    fields->properties.multiplierScale = newValue;
    touchNoError();
}

double DecimalFormat::getRoundingIncrement(void) const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().roundingIncrement;
    }
    return fields->exportedProperties.roundingIncrement;
}

void DecimalFormat::setRoundingIncrement(double newValue) {
    if (fields == nullptr) { return; }
    if (newValue == fields->properties.roundingIncrement) { return; }
    fields->properties.roundingIncrement = newValue;
    touchNoError();
}

ERoundingMode DecimalFormat::getRoundingMode(void) const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return static_cast<ERoundingMode>(DecimalFormatProperties::getDefault().roundingMode.getNoError());
    }
    // UNumberFormatRoundingMode and ERoundingMode have the same values.
    return static_cast<ERoundingMode>(fields->exportedProperties.roundingMode.getNoError());
}

void DecimalFormat::setRoundingMode(ERoundingMode roundingMode) {
    if (fields == nullptr) { return; }
    auto uRoundingMode = static_cast<UNumberFormatRoundingMode>(roundingMode);
    if (!fields->properties.roundingMode.isNull() && uRoundingMode == fields->properties.roundingMode.getNoError()) {
        return;
    }
    NumberFormat::setMaximumIntegerDigits(roundingMode); // to set field for compatibility
    fields->properties.roundingMode = uRoundingMode;
    touchNoError();
}

int32_t DecimalFormat::getFormatWidth(void) const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().formatWidth;
    }
    return fields->properties.formatWidth;
}

void DecimalFormat::setFormatWidth(int32_t width) {
    if (fields == nullptr) { return; }
    if (width == fields->properties.formatWidth) { return; }
    fields->properties.formatWidth = width;
    touchNoError();
}

UnicodeString DecimalFormat::getPadCharacterString() const {
    if (fields == nullptr || fields->properties.padString.isBogus()) {
        // Readonly-alias the static string kFallbackPaddingString
        return {TRUE, kFallbackPaddingString, -1};
    } else {
        return fields->properties.padString;
    }
}

void DecimalFormat::setPadCharacter(const UnicodeString& padChar) {
    if (fields == nullptr) { return; }
    if (padChar == fields->properties.padString) { return; }
    if (padChar.length() > 0) {
        fields->properties.padString = UnicodeString(padChar.char32At(0));
    } else {
        fields->properties.padString.setToBogus();
    }
    touchNoError();
}

EPadPosition DecimalFormat::getPadPosition(void) const {
    if (fields == nullptr || fields->properties.padPosition.isNull()) {
        return EPadPosition::kPadBeforePrefix;
    } else {
        // UNumberFormatPadPosition and EPadPosition have the same values.
        return static_cast<EPadPosition>(fields->properties.padPosition.getNoError());
    }
}

void DecimalFormat::setPadPosition(EPadPosition padPos) {
    if (fields == nullptr) { return; }
    auto uPadPos = static_cast<UNumberFormatPadPosition>(padPos);
    if (!fields->properties.padPosition.isNull() && uPadPos == fields->properties.padPosition.getNoError()) {
        return;
    }
    fields->properties.padPosition = uPadPos;
    touchNoError();
}

UBool DecimalFormat::isScientificNotation(void) const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return (DecimalFormatProperties::getDefault().minimumExponentDigits != -1);
    }
    return (fields->properties.minimumExponentDigits != -1);
}

void DecimalFormat::setScientificNotation(UBool useScientific) {
    if (fields == nullptr) { return; }
    int32_t minExp = useScientific ? 1 : -1;
    if (fields->properties.minimumExponentDigits == minExp) { return; }
    if (useScientific) {
        fields->properties.minimumExponentDigits = 1;
    } else {
        fields->properties.minimumExponentDigits = -1;
    }
    touchNoError();
}

int8_t DecimalFormat::getMinimumExponentDigits(void) const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return static_cast<int8_t>(DecimalFormatProperties::getDefault().minimumExponentDigits);
    }
    return static_cast<int8_t>(fields->properties.minimumExponentDigits);
}

void DecimalFormat::setMinimumExponentDigits(int8_t minExpDig) {
    if (fields == nullptr) { return; }
    if (minExpDig == fields->properties.minimumExponentDigits) { return; }
    fields->properties.minimumExponentDigits = minExpDig;
    touchNoError();
}

UBool DecimalFormat::isExponentSignAlwaysShown(void) const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().exponentSignAlwaysShown;
    }
    return fields->properties.exponentSignAlwaysShown;
}

void DecimalFormat::setExponentSignAlwaysShown(UBool expSignAlways) {
    if (fields == nullptr) { return; }
    if (UBOOL_TO_BOOL(expSignAlways) == fields->properties.exponentSignAlwaysShown) { return; }
    fields->properties.exponentSignAlwaysShown = expSignAlways;
    touchNoError();
}

int32_t DecimalFormat::getGroupingSize(void) const {
    int32_t groupingSize;
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        groupingSize = DecimalFormatProperties::getDefault().groupingSize;
    } else {
        groupingSize = fields->properties.groupingSize;
    }
    if (groupingSize < 0) {
        return 0;
    }
    return groupingSize;
}

void DecimalFormat::setGroupingSize(int32_t newValue) {
    if (fields == nullptr) { return; }
    if (newValue == fields->properties.groupingSize) { return; }
    fields->properties.groupingSize = newValue;
    touchNoError();
}

int32_t DecimalFormat::getSecondaryGroupingSize(void) const {
    int32_t grouping2;
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        grouping2 = DecimalFormatProperties::getDefault().secondaryGroupingSize;
    } else {
        grouping2 = fields->properties.secondaryGroupingSize;
    }
    if (grouping2 < 0) {
        return 0;
    }
    return grouping2;
}

void DecimalFormat::setSecondaryGroupingSize(int32_t newValue) {
    if (fields == nullptr) { return; }
    if (newValue == fields->properties.secondaryGroupingSize) { return; }
    fields->properties.secondaryGroupingSize = newValue;
    touchNoError();
}

int32_t DecimalFormat::getMinimumGroupingDigits() const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().minimumGroupingDigits;
    }
    return fields->properties.minimumGroupingDigits;
}

void DecimalFormat::setMinimumGroupingDigits(int32_t newValue) {
    if (fields == nullptr) { return; }
    if (newValue == fields->properties.minimumGroupingDigits) { return; }
    fields->properties.minimumGroupingDigits = newValue;
    touchNoError();
}

UBool DecimalFormat::isDecimalSeparatorAlwaysShown(void) const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().decimalSeparatorAlwaysShown;
    }
    return fields->properties.decimalSeparatorAlwaysShown;
}

void DecimalFormat::setDecimalSeparatorAlwaysShown(UBool newValue) {
    if (fields == nullptr) { return; }
    if (UBOOL_TO_BOOL(newValue) == fields->properties.decimalSeparatorAlwaysShown) { return; }
    fields->properties.decimalSeparatorAlwaysShown = newValue;
    touchNoError();
}

UBool DecimalFormat::isDecimalPatternMatchRequired(void) const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().decimalPatternMatchRequired;
    }
    return fields->properties.decimalPatternMatchRequired;
}

void DecimalFormat::setDecimalPatternMatchRequired(UBool newValue) {
    if (fields == nullptr) { return; }
    if (UBOOL_TO_BOOL(newValue) == fields->properties.decimalPatternMatchRequired) { return; }
    fields->properties.decimalPatternMatchRequired = newValue;
    touchNoError();
}

UBool DecimalFormat::isParseNoExponent() const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().parseNoExponent;
    }
    return fields->properties.parseNoExponent;
}

void DecimalFormat::setParseNoExponent(UBool value) {
    if (fields == nullptr) { return; }
    if (UBOOL_TO_BOOL(value) == fields->properties.parseNoExponent) { return; }
    fields->properties.parseNoExponent = value;
    touchNoError();
}

UBool DecimalFormat::isParseCaseSensitive() const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().parseCaseSensitive;
    }
    return fields->properties.parseCaseSensitive;
}

void DecimalFormat::setParseCaseSensitive(UBool value) {
    if (fields == nullptr) { return; }
    if (UBOOL_TO_BOOL(value) == fields->properties.parseCaseSensitive) { return; }
    fields->properties.parseCaseSensitive = value;
    touchNoError();
}

UBool DecimalFormat::isFormatFailIfMoreThanMaxDigits() const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().formatFailIfMoreThanMaxDigits;
    }
    return fields->properties.formatFailIfMoreThanMaxDigits;
}

void DecimalFormat::setFormatFailIfMoreThanMaxDigits(UBool value) {
    if (fields == nullptr) { return; }
    if (UBOOL_TO_BOOL(value) == fields->properties.formatFailIfMoreThanMaxDigits) { return; }
    fields->properties.formatFailIfMoreThanMaxDigits = value;
    touchNoError();
}

UnicodeString& DecimalFormat::toPattern(UnicodeString& result) const {
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        result.setToBogus();
        return result;
    }
    // Pull some properties from exportedProperties and others from properties
    // to keep affix patterns intact.  In particular, pull rounding properties
    // so that CurrencyUsage is reflected properly.
    // TODO: Consider putting this logic in number_patternstring.cpp instead.
    ErrorCode localStatus;
    DecimalFormatProperties tprops(fields->properties);
    bool useCurrency = (
        !tprops.currency.isNull() ||
        !tprops.currencyPluralInfo.fPtr.isNull() ||
        !tprops.currencyUsage.isNull() ||
        AffixUtils::hasCurrencySymbols(tprops.positivePrefixPattern, localStatus) ||
        AffixUtils::hasCurrencySymbols(tprops.positiveSuffixPattern, localStatus) ||
        AffixUtils::hasCurrencySymbols(tprops.negativePrefixPattern, localStatus) ||
        AffixUtils::hasCurrencySymbols(tprops.negativeSuffixPattern, localStatus));
    if (useCurrency) {
        tprops.minimumFractionDigits = fields->exportedProperties.minimumFractionDigits;
        tprops.maximumFractionDigits = fields->exportedProperties.maximumFractionDigits;
        tprops.roundingIncrement = fields->exportedProperties.roundingIncrement;
    }
    result = PatternStringUtils::propertiesToPatternString(tprops, localStatus);
    return result;
}

UnicodeString& DecimalFormat::toLocalizedPattern(UnicodeString& result) const {
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        result.setToBogus();
        return result;
    }
    ErrorCode localStatus;
    result = toPattern(result);
    result = PatternStringUtils::convertLocalized(result, *fields->symbols, true, localStatus);
    return result;
}

void DecimalFormat::applyPattern(const UnicodeString& pattern, UParseError&, UErrorCode& status) {
    // TODO: What is parseError for?
    applyPattern(pattern, status);
}

void DecimalFormat::applyPattern(const UnicodeString& pattern, UErrorCode& status) {
    // don't overwrite status if it's already a failure.
    if (U_FAILURE(status)) { return; }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    setPropertiesFromPattern(pattern, IGNORE_ROUNDING_NEVER, status);
    touch(status);
}

void DecimalFormat::applyLocalizedPattern(const UnicodeString& localizedPattern, UParseError&,
                                          UErrorCode& status) {
    // TODO: What is parseError for?
    applyLocalizedPattern(localizedPattern, status);
}

void DecimalFormat::applyLocalizedPattern(const UnicodeString& localizedPattern, UErrorCode& status) {
    // don't overwrite status if it's already a failure.
    if (U_FAILURE(status)) { return; }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    UnicodeString pattern = PatternStringUtils::convertLocalized(
            localizedPattern, *fields->symbols, false, status);
    applyPattern(pattern, status);
}

void DecimalFormat::setMaximumIntegerDigits(int32_t newValue) {
    if (fields == nullptr) { return; }
    if (newValue == fields->properties.maximumIntegerDigits) { return; }
    // For backwards compatibility, conflicting min/max need to keep the most recent setting.
    int32_t min = fields->properties.minimumIntegerDigits;
    if (min >= 0 && min > newValue) {
        fields->properties.minimumIntegerDigits = newValue;
    }
    fields->properties.maximumIntegerDigits = newValue;
    touchNoError();
}

void DecimalFormat::setMinimumIntegerDigits(int32_t newValue) {
    if (fields == nullptr) { return; }
    if (newValue == fields->properties.minimumIntegerDigits) { return; }
    // For backwards compatibility, conflicting min/max need to keep the most recent setting.
    int32_t max = fields->properties.maximumIntegerDigits;
    if (max >= 0 && max < newValue) {
        fields->properties.maximumIntegerDigits = newValue;
    }
    fields->properties.minimumIntegerDigits = newValue;
    touchNoError();
}

void DecimalFormat::setMaximumFractionDigits(int32_t newValue) {
    if (fields == nullptr) { return; }
    if (newValue == fields->properties.maximumFractionDigits) { return; }
    // cap for backward compatibility, formerly 340, now 999
    if (newValue > kMaxIntFracSig) {
        newValue = kMaxIntFracSig;
    }
    // For backwards compatibility, conflicting min/max need to keep the most recent setting.
    int32_t min = fields->properties.minimumFractionDigits;
    if (min >= 0 && min > newValue) {
        fields->properties.minimumFractionDigits = newValue;
    }
    fields->properties.maximumFractionDigits = newValue;
    touchNoError();
}

void DecimalFormat::setMinimumFractionDigits(int32_t newValue) {
    if (fields == nullptr) { return; }
    if (newValue == fields->properties.minimumFractionDigits) { return; }
    // For backwards compatibility, conflicting min/max need to keep the most recent setting.
    int32_t max = fields->properties.maximumFractionDigits;
    if (max >= 0 && max < newValue) {
        fields->properties.maximumFractionDigits = newValue;
    }
    fields->properties.minimumFractionDigits = newValue;
    touchNoError();
}

int32_t DecimalFormat::getMinimumSignificantDigits() const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().minimumSignificantDigits;
    }
    return fields->exportedProperties.minimumSignificantDigits;
}

int32_t DecimalFormat::getMaximumSignificantDigits() const {
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        return DecimalFormatProperties::getDefault().maximumSignificantDigits;
    }
    return fields->exportedProperties.maximumSignificantDigits;
}

void DecimalFormat::setMinimumSignificantDigits(int32_t value) {
    if (fields == nullptr) { return; }
    if (value == fields->properties.minimumSignificantDigits) { return; }
    int32_t max = fields->properties.maximumSignificantDigits;
    if (max >= 0 && max < value) {
        fields->properties.maximumSignificantDigits = value;
    }
    fields->properties.minimumSignificantDigits = value;
    touchNoError();
}

void DecimalFormat::setMaximumSignificantDigits(int32_t value) {
    if (fields == nullptr) { return; }
    if (value == fields->properties.maximumSignificantDigits) { return; }
    int32_t min = fields->properties.minimumSignificantDigits;
    if (min >= 0 && min > value) {
        fields->properties.minimumSignificantDigits = value;
    }
    fields->properties.maximumSignificantDigits = value;
    touchNoError();
}

UBool DecimalFormat::areSignificantDigitsUsed() const {
    const DecimalFormatProperties* dfp;
    // Not much we can do to report an error.
    if (fields == nullptr) {
        // Fallback to using the default instance of DecimalFormatProperties.
        dfp = &(DecimalFormatProperties::getDefault());
    } else {
        dfp = &fields->properties;
    }
    return dfp->minimumSignificantDigits != -1 || dfp->maximumSignificantDigits != -1;
}

void DecimalFormat::setSignificantDigitsUsed(UBool useSignificantDigits) {
    if (fields == nullptr) { return; }

    // These are the default values from the old implementation.
    if (useSignificantDigits) {
        if (fields->properties.minimumSignificantDigits != -1 ||
            fields->properties.maximumSignificantDigits != -1) {
            return;
        }
    } else {
        if (fields->properties.minimumSignificantDigits == -1 &&
            fields->properties.maximumSignificantDigits == -1) {
            return;
        }
    }
    int32_t minSig = useSignificantDigits ? 1 : -1;
    int32_t maxSig = useSignificantDigits ? 6 : -1;
    fields->properties.minimumSignificantDigits = minSig;
    fields->properties.maximumSignificantDigits = maxSig;
    touchNoError();
}

void DecimalFormat::setCurrency(const char16_t* theCurrency, UErrorCode& ec) {
    // don't overwrite ec if it's already a failure.
    if (U_FAILURE(ec)) { return; }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        ec = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    CurrencyUnit currencyUnit(theCurrency, ec);
    if (U_FAILURE(ec)) { return; }
    if (!fields->properties.currency.isNull() && fields->properties.currency.getNoError() == currencyUnit) {
        return;
    }
    NumberFormat::setCurrency(theCurrency, ec); // to set field for compatibility
    fields->properties.currency = currencyUnit;
    // TODO: Set values in fields->symbols, too?
    touchNoError();
}

void DecimalFormat::setCurrency(const char16_t* theCurrency) {
    ErrorCode localStatus;
    setCurrency(theCurrency, localStatus);
}

void DecimalFormat::setCurrencyUsage(UCurrencyUsage newUsage, UErrorCode* ec) {
    // don't overwrite ec if it's already a failure.
    if (U_FAILURE(*ec)) { return; }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        *ec = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    if (!fields->properties.currencyUsage.isNull() && newUsage == fields->properties.currencyUsage.getNoError()) {
        return;
    }
    fields->properties.currencyUsage = newUsage;
    touch(*ec);
}

UCurrencyUsage DecimalFormat::getCurrencyUsage() const {
    // CurrencyUsage is not exported, so we have to get it from the input property bag.
    // TODO: Should we export CurrencyUsage instead?
    if (fields == nullptr || fields->properties.currencyUsage.isNull()) {
        return UCURR_USAGE_STANDARD;
    }
    return fields->properties.currencyUsage.getNoError();
}

void
DecimalFormat::formatToDecimalQuantity(double number, DecimalQuantity& output, UErrorCode& status) const {
    // don't overwrite status if it's already a failure.
    if (U_FAILURE(status)) { return; }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    fields->formatter.formatDouble(number, status).getDecimalQuantity(output, status);
}

void DecimalFormat::formatToDecimalQuantity(const Formattable& number, DecimalQuantity& output,
                                            UErrorCode& status) const {
    // don't overwrite status if it's already a failure.
    if (U_FAILURE(status)) { return; }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    UFormattedNumberData obj;
    number.populateDecimalQuantity(obj.quantity, status);
    fields->formatter.formatImpl(&obj, status);
    output = std::move(obj.quantity);
}

const number::LocalizedNumberFormatter* DecimalFormat::toNumberFormatter(UErrorCode& status) const {
    // We sometimes need to return nullptr here (see ICU-20380)
    if (U_FAILURE(status)) { return nullptr; }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    return &fields->formatter;
}

/** Rebuilds the formatter object from the property bag. */
void DecimalFormat::touch(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    if (fields == nullptr) {
        // We only get here if an OOM error happend during construction, copy construction, assignment, or modification.
        // For regular construction, the caller should have checked the status variable for errors.
        // For copy construction, there is unfortunately nothing to report the error, so we need to guard against
        // this possible bad state here and set the status to an error.
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    // In C++, fields->symbols is the source of truth for the locale.
    Locale locale = fields->symbols->getLocale();

    // Note: The formatter is relatively cheap to create, and we need it to populate fields->exportedProperties,
    // so automatically recompute it here. The parser is a bit more expensive and is not needed until the
    // parse method is called, so defer that until needed.
    // TODO: Only update the pieces that changed instead of re-computing the whole formatter?

    // Since memory has already been allocated for the formatter, we can move assign a stack-allocated object
    // and don't need to call new. (Which is slower and could possibly fail).
    fields->formatter = NumberPropertyMapper::create(
        fields->properties, *fields->symbols, fields->warehouse, fields->exportedProperties, status
    ).locale(locale);

    // Do this after fields->exportedProperties are set up
    setupFastFormat();

    // Delete the parsers if they were made previously
    delete fields->atomicParser.exchange(nullptr);
    delete fields->atomicCurrencyParser.exchange(nullptr);

    // In order for the getters to work, we need to populate some fields in NumberFormat.
    NumberFormat::setCurrency(fields->exportedProperties.currency.get(status).getISOCurrency(), status);
    NumberFormat::setMaximumIntegerDigits(fields->exportedProperties.maximumIntegerDigits);
    NumberFormat::setMinimumIntegerDigits(fields->exportedProperties.minimumIntegerDigits);
    NumberFormat::setMaximumFractionDigits(fields->exportedProperties.maximumFractionDigits);
    NumberFormat::setMinimumFractionDigits(fields->exportedProperties.minimumFractionDigits);
    // fImpl->properties, not fields->exportedProperties, since this information comes from the pattern:
    NumberFormat::setGroupingUsed(fields->properties.groupingUsed);
}

void DecimalFormat::touchNoError() {
    UErrorCode localStatus = U_ZERO_ERROR;
    touch(localStatus);
}

void DecimalFormat::setPropertiesFromPattern(const UnicodeString& pattern, int32_t ignoreRounding,
                                             UErrorCode& status) {
    if (U_SUCCESS(status)) {
        // Cast workaround to get around putting the enum in the public header file
        auto actualIgnoreRounding = static_cast<IgnoreRounding>(ignoreRounding);
        PatternParser::parseToExistingProperties(pattern, fields->properties,  actualIgnoreRounding, status);
    }
}

const numparse::impl::NumberParserImpl* DecimalFormat::getParser(UErrorCode& status) const {
    // TODO: Move this into umutex.h? (similar logic also in numrange_fluent.cpp)
    // See ICU-20146

    if (U_FAILURE(status)) {
        return nullptr;
    }

    // First try to get the pre-computed parser
    auto* ptr = fields->atomicParser.load();
    if (ptr != nullptr) {
        return ptr;
    }

    // Try computing the parser on our own
    auto* temp = NumberParserImpl::createParserFromProperties(fields->properties, *fields->symbols, false, status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (temp == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }

    // Note: ptr starts as nullptr; during compare_exchange,
    // it is set to what is actually stored in the atomic
    // if another thread beat us to computing the parser object.
    auto* nonConstThis = const_cast<DecimalFormat*>(this);
    if (!nonConstThis->fields->atomicParser.compare_exchange_strong(ptr, temp)) {
        // Another thread beat us to computing the parser
        delete temp;
        return ptr;
    } else {
        // Our copy of the parser got stored in the atomic
        return temp;
    }
}

const numparse::impl::NumberParserImpl* DecimalFormat::getCurrencyParser(UErrorCode& status) const {
    if (U_FAILURE(status)) { return nullptr; }

    // First try to get the pre-computed parser
    auto* ptr = fields->atomicCurrencyParser.load();
    if (ptr != nullptr) {
        return ptr;
    }

    // Try computing the parser on our own
    auto* temp = NumberParserImpl::createParserFromProperties(fields->properties, *fields->symbols, true, status);
    if (temp == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        // although we may still dereference, call sites should be guarded
    }

    // Note: ptr starts as nullptr; during compare_exchange, it is set to what is actually stored in the
    // atomic if another thread beat us to computing the parser object.
    auto* nonConstThis = const_cast<DecimalFormat*>(this);
    if (!nonConstThis->fields->atomicCurrencyParser.compare_exchange_strong(ptr, temp)) {
        // Another thread beat us to computing the parser
        delete temp;
        return ptr;
    } else {
        // Our copy of the parser got stored in the atomic
        return temp;
    }
}

void
DecimalFormat::fieldPositionHelper(const number::FormattedNumber& formatted, FieldPosition& fieldPosition,
                                   int32_t offset, UErrorCode& status) {
    if (U_FAILURE(status)) { return; }
    // always return first occurrence:
    fieldPosition.setBeginIndex(0);
    fieldPosition.setEndIndex(0);
    bool found = formatted.nextFieldPosition(fieldPosition, status);
    if (found && offset != 0) {
        FieldPositionOnlyHandler fpoh(fieldPosition);
        fpoh.shiftLast(offset);
    }
}

void
DecimalFormat::fieldPositionIteratorHelper(const number::FormattedNumber& formatted, FieldPositionIterator* fpi,
                                           int32_t offset, UErrorCode& status) {
    if (U_SUCCESS(status) && (fpi != nullptr)) {
        FieldPositionIteratorHandler fpih(fpi, status);
        fpih.setShift(offset);
        formatted.getAllFieldPositionsImpl(fpih, status);
    }
}

// To debug fast-format, change void(x) to printf(x)
#define trace(x) void(x)

void DecimalFormat::setupFastFormat() {
    // Check the majority of properties:
    if (!fields->properties.equalsDefaultExceptFastFormat()) {
        trace("no fast format: equality\n");
        fields->canUseFastFormat = false;
        return;
    }

    // Now check the remaining properties.
    // Nontrivial affixes:
    UBool trivialPP = fields->properties.positivePrefixPattern.isEmpty();
    UBool trivialPS = fields->properties.positiveSuffixPattern.isEmpty();
    UBool trivialNP = fields->properties.negativePrefixPattern.isBogus() || (
            fields->properties.negativePrefixPattern.length() == 1 &&
            fields->properties.negativePrefixPattern.charAt(0) == u'-');
    UBool trivialNS = fields->properties.negativeSuffixPattern.isEmpty();
    if (!trivialPP || !trivialPS || !trivialNP || !trivialNS) {
        trace("no fast format: affixes\n");
        fields->canUseFastFormat = false;
        return;
    }

    // Grouping (secondary grouping is forbidden in equalsDefaultExceptFastFormat):
    bool groupingUsed = fields->properties.groupingUsed;
    int32_t groupingSize = fields->properties.groupingSize;
    bool unusualGroupingSize = groupingSize > 0 && groupingSize != 3;
    const UnicodeString& groupingString = fields->symbols->getConstSymbol(DecimalFormatSymbols::kGroupingSeparatorSymbol);
    if (groupingUsed && (unusualGroupingSize || groupingString.length() != 1)) {
        trace("no fast format: grouping\n");
        fields->canUseFastFormat = false;
        return;
    }

    // Integer length:
    int32_t minInt = fields->exportedProperties.minimumIntegerDigits;
    int32_t maxInt = fields->exportedProperties.maximumIntegerDigits;
    // Fastpath supports up to only 10 digits (length of INT32_MIN)
    if (minInt > 10) {
        trace("no fast format: integer\n");
        fields->canUseFastFormat = false;
        return;
    }

    // Fraction length (no fraction part allowed in fast path):
    int32_t minFrac = fields->exportedProperties.minimumFractionDigits;
    if (minFrac > 0) {
        trace("no fast format: fraction\n");
        fields->canUseFastFormat = false;
        return;
    }

    // Other symbols:
    const UnicodeString& minusSignString = fields->symbols->getConstSymbol(DecimalFormatSymbols::kMinusSignSymbol);
    UChar32 codePointZero = fields->symbols->getCodePointZero();
    if (minusSignString.length() != 1 || U16_LENGTH(codePointZero) != 1) {
        trace("no fast format: symbols\n");
        fields->canUseFastFormat = false;
        return;
    }

    // Good to go!
    trace("can use fast format!\n");
    fields->canUseFastFormat = true;
    fields->fastData.cpZero = static_cast<char16_t>(codePointZero);
    fields->fastData.cpGroupingSeparator = groupingUsed && groupingSize == 3 ? groupingString.charAt(0) : 0;
    fields->fastData.cpMinusSign = minusSignString.charAt(0);
    fields->fastData.minInt = (minInt < 0 || minInt > 127) ? 0 : static_cast<int8_t>(minInt);
    fields->fastData.maxInt = (maxInt < 0 || maxInt > 127) ? 127 : static_cast<int8_t>(maxInt);
}

bool DecimalFormat::fastFormatDouble(double input, UnicodeString& output) const {
    if (!fields->canUseFastFormat) {
        return false;
    }
    if (std::isnan(input)
            || std::trunc(input) != input
            || input <= INT32_MIN
            || input > INT32_MAX) {
        return false;
    }
    doFastFormatInt32(static_cast<int32_t>(input), std::signbit(input), output);
    return true;
}

bool DecimalFormat::fastFormatInt64(int64_t input, UnicodeString& output) const {
    if (!fields->canUseFastFormat) {
        return false;
    }
    if (input <= INT32_MIN || input > INT32_MAX) {
        return false;
    }
    doFastFormatInt32(static_cast<int32_t>(input), input < 0, output);
    return true;
}

void DecimalFormat::doFastFormatInt32(int32_t input, bool isNegative, UnicodeString& output) const {
    U_ASSERT(fields->canUseFastFormat);
    if (isNegative) {
        output.append(fields->fastData.cpMinusSign);
        U_ASSERT(input != INT32_MIN);  // handled by callers
        input = -input;
    }
    // Cap at int32_t to make the buffer small and operations fast.
    // Longest string: "2,147,483,648" (13 chars in length)
    static constexpr int32_t localCapacity = 13;
    char16_t localBuffer[localCapacity];
    char16_t* ptr = localBuffer + localCapacity;
    int8_t group = 0;
    for (int8_t i = 0; i < fields->fastData.maxInt && (input != 0 || i < fields->fastData.minInt); i++) {
        if (group++ == 3 && fields->fastData.cpGroupingSeparator != 0) {
            *(--ptr) = fields->fastData.cpGroupingSeparator;
            group = 1;
        }
        std::div_t res = std::div(input, 10);
        *(--ptr) = static_cast<char16_t>(fields->fastData.cpZero + res.rem);
        input = res.quot;
    }
    int32_t len = localCapacity - static_cast<int32_t>(ptr - localBuffer);
    output.append(ptr, len);
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING

#include <cmath>

// ICU PATCH: Customize header file paths for ICU.

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


// ICU PATCH: Customize header file paths for ICU.

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


#include <cstdlib>
#include <cstring>

// ICU PATCH: Use U_ASSERT instead of <assert.h>

#ifndef DOUBLE_CONVERSION_ASSERT
#define DOUBLE_CONVERSION_ASSERT(condition)         \
    U_ASSERT(condition);
#endif
#ifndef DOUBLE_CONVERSION_UNIMPLEMENTED
#define DOUBLE_CONVERSION_UNIMPLEMENTED() (abort())
#endif
#ifndef DOUBLE_CONVERSION_NO_RETURN
#ifdef _MSC_VER
#define DOUBLE_CONVERSION_NO_RETURN __declspec(noreturn)
#else
#define DOUBLE_CONVERSION_NO_RETURN __attribute__((noreturn))
#endif
#endif
#ifndef DOUBLE_CONVERSION_UNREACHABLE
#ifdef _MSC_VER
void DOUBLE_CONVERSION_NO_RETURN abort_noreturn();
inline void abort_noreturn() { abort(); }
#define DOUBLE_CONVERSION_UNREACHABLE()   (abort_noreturn())
#else
#define DOUBLE_CONVERSION_UNREACHABLE()   (abort())
#endif
#endif

#ifndef DOUBLE_CONVERSION_UNUSED
#ifdef __GNUC__
#define DOUBLE_CONVERSION_UNUSED __attribute__((unused))
#else
#define DOUBLE_CONVERSION_UNUSED
#endif
#endif

// Double operations detection based on target architecture.
// Linux uses a 80bit wide floating point stack on x86. This induces double
// rounding, which in turn leads to wrong results.
// An easy way to test if the floating-point operations are correct is to
// evaluate: 89255.0/1e22. If the floating-point stack is 64 bits wide then
// the result is equal to 89255e-22.
// The best way to test this, is to create a division-function and to compare
// the output of the division with the expected result. (Inlining must be
// disabled.)
// On Linux,x86 89255e-22 != Div_double(89255.0/1e22)
//
// For example:
/*
// -- in div.c
double Div_double(double x, double y) { return x / y; }

// -- in main.c
double Div_double(double x, double y);  // Forward declaration.

int main(int argc, char** argv) {
  return Div_double(89255.0, 1e22) == 89255e-22;
}
*/
// Run as follows ./main || echo "correct"
//
// If it prints "correct" then the architecture should be here, in the "correct" section.
#if defined(_M_X64) || defined(__x86_64__) || \
    defined(__ARMEL__) || defined(__avr32__) || defined(_M_ARM) || defined(_M_ARM64) || \
    defined(__hppa__) || defined(__ia64__) || \
    defined(__mips__) || \
    defined(__powerpc__) || defined(__ppc__) || defined(__ppc64__) || \
    defined(_POWER) || defined(_ARCH_PPC) || defined(_ARCH_PPC64) || \
    defined(__sparc__) || defined(__sparc) || defined(__s390__) || \
    defined(__SH4__) || defined(__alpha__) || \
    defined(_MIPS_ARCH_MIPS32R2) || defined(__ARMEB__) ||\
    defined(__AARCH64EL__) || defined(__aarch64__) || defined(__AARCH64EB__) || \
    defined(__riscv) || defined(__e2k__) || \
    defined(__or1k__) || defined(__arc__) || \
    defined(__EMSCRIPTEN__)
#define DOUBLE_CONVERSION_CORRECT_DOUBLE_OPERATIONS 1
#elif defined(__mc68000__) || \
    defined(__pnacl__) || defined(__native_client__)
#undef DOUBLE_CONVERSION_CORRECT_DOUBLE_OPERATIONS
#elif defined(_M_IX86) || defined(__i386__) || defined(__i386)
#if defined(_WIN32)
// Windows uses a 64bit wide floating point stack.
#define DOUBLE_CONVERSION_CORRECT_DOUBLE_OPERATIONS 1
#else
#undef DOUBLE_CONVERSION_CORRECT_DOUBLE_OPERATIONS
#endif  // _WIN32
#else
#error Target architecture was not detected as supported by Double-Conversion.
#endif

#if defined(_WIN32) && !defined(__MINGW32__)

typedef signed char int8_t;
typedef unsigned char uint8_t;
typedef short int16_t;  // NOLINT
typedef unsigned short uint16_t;  // NOLINT
typedef int int32_t;
typedef unsigned int uint32_t;
typedef __int64 int64_t;
typedef unsigned __int64 uint64_t;
// intptr_t and friends are defined in crtdefs.h through stdio.h.

#else

#include <stdint.h>

#endif

typedef uint16_t uc16;

// The following macro works on both 32 and 64-bit platforms.
// Usage: instead of writing 0x1234567890123456
//      write DOUBLE_CONVERSION_UINT64_2PART_C(0x12345678,90123456);
#define DOUBLE_CONVERSION_UINT64_2PART_C(a, b) (((static_cast<uint64_t>(a) << 32) + 0x##b##u))


// The expression DOUBLE_CONVERSION_ARRAY_SIZE(a) is a compile-time constant of type
// size_t which represents the number of elements of the given
// array. You should only use DOUBLE_CONVERSION_ARRAY_SIZE on statically allocated
// arrays.
#ifndef DOUBLE_CONVERSION_ARRAY_SIZE
#define DOUBLE_CONVERSION_ARRAY_SIZE(a)                                   \
  ((sizeof(a) / sizeof(*(a))) /                         \
  static_cast<size_t>(!(sizeof(a) % sizeof(*(a)))))
#endif

// A macro to disallow the evil copy constructor and operator= functions
// This should be used in the private: declarations for a class
#ifndef DOUBLE_CONVERSION_DISALLOW_COPY_AND_ASSIGN
#define DOUBLE_CONVERSION_DISALLOW_COPY_AND_ASSIGN(TypeName)      \
  TypeName(const TypeName&);                    \
  void operator=(const TypeName&)
#endif

// A macro to disallow all the implicit constructors, namely the
// default constructor, copy constructor and operator= functions.
//
// This should be used in the private: declarations for a class
// that wants to prevent anyone from instantiating it. This is
// especially useful for classes containing only static methods.
#ifndef DOUBLE_CONVERSION_DISALLOW_IMPLICIT_CONSTRUCTORS
#define DOUBLE_CONVERSION_DISALLOW_IMPLICIT_CONSTRUCTORS(TypeName) \
  TypeName();                                    \
  DOUBLE_CONVERSION_DISALLOW_COPY_AND_ASSIGN(TypeName)
#endif

// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

inline int StrLength(const char* string) {
  size_t length = strlen(string);
  DOUBLE_CONVERSION_ASSERT(length == static_cast<size_t>(static_cast<int>(length)));
  return static_cast<int>(length);
}

// This is a simplified version of V8's Vector class.
template <typename T>
class Vector {
 public:
  Vector() : start_(NULL), length_(0) {}
  Vector(T* data, int len) : start_(data), length_(len) {
    DOUBLE_CONVERSION_ASSERT(len == 0 || (len > 0 && data != NULL));
  }

  // Returns a vector using the same backing storage as this one,
  // spanning from and including 'from', to but not including 'to'.
  Vector<T> SubVector(int from, int to) {
    DOUBLE_CONVERSION_ASSERT(to <= length_);
    DOUBLE_CONVERSION_ASSERT(from < to);
    DOUBLE_CONVERSION_ASSERT(0 <= from);
    return Vector<T>(start() + from, to - from);
  }

  // Returns the length of the vector.
  int length() const { return length_; }

  // Returns whether or not the vector is empty.
  bool is_empty() const { return length_ == 0; }

  // Returns the pointer to the start of the data in the vector.
  T* start() const { return start_; }

  // Access individual vector elements - checks bounds in debug mode.
  T& operator[](int index) const {
    DOUBLE_CONVERSION_ASSERT(0 <= index && index < length_);
    return start_[index];
  }

  T& first() { return start_[0]; }

  T& last() { return start_[length_ - 1]; }

  void pop_back() {
    DOUBLE_CONVERSION_ASSERT(!is_empty());
    --length_;
  }

 private:
  T* start_;
  int length_;
};


// Helper class for building result strings in a character buffer. The
// purpose of the class is to use safe operations that checks the
// buffer bounds on all operations in debug mode.
class StringBuilder {
 public:
  StringBuilder(char* buffer, int buffer_size)
      : buffer_(buffer, buffer_size), position_(0) { }

  ~StringBuilder() { if (!is_finalized()) Finalize(); }

  int size() const { return buffer_.length(); }

  // Get the current position in the builder.
  int position() const {
    DOUBLE_CONVERSION_ASSERT(!is_finalized());
    return position_;
  }

  // Reset the position.
  void Reset() { position_ = 0; }

  // Add a single character to the builder. It is not allowed to add
  // 0-characters; use the Finalize() method to terminate the string
  // instead.
  void AddCharacter(char c) {
    DOUBLE_CONVERSION_ASSERT(c != '\0');
    DOUBLE_CONVERSION_ASSERT(!is_finalized() && position_ < buffer_.length());
    buffer_[position_++] = c;
  }

  // Add an entire string to the builder. Uses strlen() internally to
  // compute the length of the input string.
  void AddString(const char* s) {
    AddSubstring(s, StrLength(s));
  }

  // Add the first 'n' characters of the given string 's' to the
  // builder. The input string must have enough characters.
  void AddSubstring(const char* s, int n) {
    DOUBLE_CONVERSION_ASSERT(!is_finalized() && position_ + n < buffer_.length());
    DOUBLE_CONVERSION_ASSERT(static_cast<size_t>(n) <= strlen(s));
    memmove(&buffer_[position_], s, n);
    position_ += n;
  }


  // Add character padding to the builder. If count is non-positive,
  // nothing is added to the builder.
  void AddPadding(char c, int count) {
    for (int i = 0; i < count; i++) {
      AddCharacter(c);
    }
  }

  // Finalize the string by 0-terminating it and returning the buffer.
  char* Finalize() {
    DOUBLE_CONVERSION_ASSERT(!is_finalized() && position_ < buffer_.length());
    buffer_[position_] = '\0';
    // Make sure nobody managed to add a 0-character to the
    // buffer while building the string.
    DOUBLE_CONVERSION_ASSERT(strlen(buffer_.start()) == static_cast<size_t>(position_));
    position_ = -1;
    DOUBLE_CONVERSION_ASSERT(is_finalized());
    return buffer_.start();
  }

 private:
  Vector<char> buffer_;
  int position_;

  bool is_finalized() const { return position_ < 0; }

  DOUBLE_CONVERSION_DISALLOW_IMPLICIT_CONSTRUCTORS(StringBuilder);
};

// The type-based aliasing rule allows the compiler to assume that pointers of
// different types (for some definition of different) never alias each other.
// Thus the following code does not work:
//
// float f = foo();
// int fbits = *(int*)(&f);
//
// The compiler 'knows' that the int pointer can't refer to f since the types
// don't match, so the compiler may cache f in a register, leaving random data
// in fbits.  Using C++ style casts makes no difference, however a pointer to
// char data is assumed to alias any other pointer.  This is the 'memcpy
// exception'.
//
// Bit_cast uses the memcpy exception to move the bits from a variable of one
// type of a variable of another type.  Of course the end result is likely to
// be implementation dependent.  Most compilers (gcc-4.2 and MSVC 2005)
// will completely optimize BitCast away.
//
// There is an additional use for BitCast.
// Recent gccs will warn when they see casts that may result in breakage due to
// the type-based aliasing rule.  If you have checked that there is no breakage
// you can use BitCast to cast one pointer type to another.  This confuses gcc
// enough that it can no longer see that you have cast one pointer type to
// another thus avoiding the warning.
template <class Dest, class Source>
Dest BitCast(const Source& source) {
  // Compile time assertion: sizeof(Dest) == sizeof(Source)
  // A compile error here means your Dest and Source have different sizes.
#if __cplusplus >= 201103L
  static_assert(sizeof(Dest) == sizeof(Source),
                "source and destination size mismatch");
#else
  DOUBLE_CONVERSION_UNUSED
  typedef char VerifySizesAreEqual[sizeof(Dest) == sizeof(Source) ? 1 : -1];
#endif

  Dest dest;
  memmove(&dest, &source, sizeof(dest));
  return dest;
}

template <class Dest, class Source>
Dest BitCast(Source* source) {
  return BitCast<Dest>(reinterpret_cast<uintptr_t>(source));
}

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END

#endif  // DOUBLE_CONVERSION_UTILS_H_


// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

enum BignumDtoaMode {
  // Return the shortest correct representation.
  // For example the output of 0.299999999999999988897 is (the less accurate but
  // correct) 0.3.
  BIGNUM_DTOA_SHORTEST,
  // Same as BIGNUM_DTOA_SHORTEST but for single-precision floats.
  BIGNUM_DTOA_SHORTEST_SINGLE,
  // Return a fixed number of digits after the decimal point.
  // For instance fixed(0.1, 4) becomes 0.1000
  // If the input number is big, the output will be big.
  BIGNUM_DTOA_FIXED,
  // Return a fixed number of digits, no matter what the exponent is.
  BIGNUM_DTOA_PRECISION
};

// Converts the given double 'v' to ascii.
// The result should be interpreted as buffer * 10^(point-length).
// The buffer will be null-terminated.
//
// The input v must be > 0 and different from NaN, and Infinity.
//
// The output depends on the given mode:
//  - SHORTEST: produce the least amount of digits for which the internal
//   identity requirement is still satisfied. If the digits are printed
//   (together with the correct exponent) then reading this number will give
//   'v' again. The buffer will choose the representation that is closest to
//   'v'. If there are two at the same distance, than the number is round up.
//   In this mode the 'requested_digits' parameter is ignored.
//  - FIXED: produces digits necessary to print a given number with
//   'requested_digits' digits after the decimal point. The produced digits
//   might be too short in which case the caller has to fill the gaps with '0's.
//   Example: toFixed(0.001, 5) is allowed to return buffer="1", point=-2.
//   Halfway cases are rounded up. The call toFixed(0.15, 2) thus returns
//     buffer="2", point=0.
//   Note: the length of the returned buffer has no meaning wrt the significance
//   of its digits. That is, just because it contains '0's does not mean that
//   any other digit would not satisfy the internal identity requirement.
//  - PRECISION: produces 'requested_digits' where the first digit is not '0'.
//   Even though the length of produced digits usually equals
//   'requested_digits', the function is allowed to return fewer digits, in
//   which case the caller has to fill the missing digits with '0's.
//   Halfway cases are again rounded up.
// 'BignumDtoa' expects the given buffer to be big enough to hold all digits
// and a terminating null-character.
void BignumDtoa(double v, BignumDtoaMode mode, int requested_digits,
                Vector<char> buffer, int* length, int* point);

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END

#endif  // DOUBLE_CONVERSION_BIGNUM_DTOA_H_


// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


// ICU PATCH: Customize header file paths for ICU.



// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

class Bignum {
 public:
  // 3584 = 128 * 28. We can represent 2^3584 > 10^1000 accurately.
  // This bignum can encode much bigger numbers, since it contains an
  // exponent.
  static const int kMaxSignificantBits = 3584;

  Bignum() : used_bigits_(0), exponent_(0) {}

  void AssignUInt16(const uint16_t value);
  void AssignUInt64(uint64_t value);
  void AssignBignum(const Bignum& other);

  void AssignDecimalString(const Vector<const char> value);
  void AssignHexString(const Vector<const char> value);

  void AssignPowerUInt16(uint16_t base, const int exponent);

  void AddUInt64(const uint64_t operand);
  void AddBignum(const Bignum& other);
  // Precondition: this >= other.
  void SubtractBignum(const Bignum& other);

  void Square();
  void ShiftLeft(const int shift_amount);
  void MultiplyByUInt32(const uint32_t factor);
  void MultiplyByUInt64(const uint64_t factor);
  void MultiplyByPowerOfTen(const int exponent);
  void Times10() { return MultiplyByUInt32(10); }
  // Pseudocode:
  //  int result = this / other;
  //  this = this % other;
  // In the worst case this function is in O(this/other).
  uint16_t DivideModuloIntBignum(const Bignum& other);

  bool ToHexString(char* buffer, const int buffer_size) const;

  // Returns
  //  -1 if a < b,
  //   0 if a == b, and
  //  +1 if a > b.
  static int Compare(const Bignum& a, const Bignum& b);
  static bool Equal(const Bignum& a, const Bignum& b) {
    return Compare(a, b) == 0;
  }
  static bool LessEqual(const Bignum& a, const Bignum& b) {
    return Compare(a, b) <= 0;
  }
  static bool Less(const Bignum& a, const Bignum& b) {
    return Compare(a, b) < 0;
  }
  // Returns Compare(a + b, c);
  static int PlusCompare(const Bignum& a, const Bignum& b, const Bignum& c);
  // Returns a + b == c
  static bool PlusEqual(const Bignum& a, const Bignum& b, const Bignum& c) {
    return PlusCompare(a, b, c) == 0;
  }
  // Returns a + b <= c
  static bool PlusLessEqual(const Bignum& a, const Bignum& b, const Bignum& c) {
    return PlusCompare(a, b, c) <= 0;
  }
  // Returns a + b < c
  static bool PlusLess(const Bignum& a, const Bignum& b, const Bignum& c) {
    return PlusCompare(a, b, c) < 0;
  }
 private:
  typedef uint32_t Chunk;
  typedef uint64_t DoubleChunk;

  static const int kChunkSize = sizeof(Chunk) * 8;
  static const int kDoubleChunkSize = sizeof(DoubleChunk) * 8;
  // With bigit size of 28 we loose some bits, but a double still fits easily
  // into two chunks, and more importantly we can use the Comba multiplication.
  static const int kBigitSize = 28;
  static const Chunk kBigitMask = (1 << kBigitSize) - 1;
  // Every instance allocates kBigitLength chunks on the stack. Bignums cannot
  // grow. There are no checks if the stack-allocated space is sufficient.
  static const int kBigitCapacity = kMaxSignificantBits / kBigitSize;

  static void EnsureCapacity(const int size) {
    if (size > kBigitCapacity) {
      DOUBLE_CONVERSION_UNREACHABLE();
    }
  }
  void Align(const Bignum& other);
  void Clamp();
  bool IsClamped() const {
    return used_bigits_ == 0 || RawBigit(used_bigits_ - 1) != 0;
  }
  void Zero() {
    used_bigits_ = 0;
    exponent_ = 0;
  }
  // Requires this to have enough capacity (no tests done).
  // Updates used_bigits_ if necessary.
  // shift_amount must be < kBigitSize.
  void BigitsShiftLeft(const int shift_amount);
  // BigitLength includes the "hidden" bigits encoded in the exponent.
  int BigitLength() const { return used_bigits_ + exponent_; }
  Chunk& RawBigit(const int index);
  const Chunk& RawBigit(const int index) const;
  Chunk BigitOrZero(const int index) const;
  void SubtractTimes(const Bignum& other, const int factor);

  // The Bignum's value is value(bigits_buffer_) * 2^(exponent_ * kBigitSize),
  // where the value of the buffer consists of the lower kBigitSize bits of
  // the first used_bigits_ Chunks in bigits_buffer_, first chunk has lowest
  // significant bits.
  int16_t used_bigits_;
  int16_t exponent_;
  Chunk bigits_buffer_[kBigitCapacity];

  DOUBLE_CONVERSION_DISALLOW_COPY_AND_ASSIGN(Bignum);
};

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END

#endif  // DOUBLE_CONVERSION_BIGNUM_H_

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2012 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


// ICU PATCH: Customize header file paths for ICU.

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


// ICU PATCH: Customize header file paths for ICU.



// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

// This "Do It Yourself Floating Point" class implements a floating-point number
// with a uint64 significand and an int exponent. Normalized DiyFp numbers will
// have the most significant bit of the significand set.
// Multiplication and Subtraction do not normalize their results.
// DiyFp store only non-negative numbers and are not designed to contain special
// doubles (NaN and Infinity).
class DiyFp {
 public:
  static const int kSignificandSize = 64;

  DiyFp() : f_(0), e_(0) {}
  DiyFp(const uint64_t significand, const int32_t exponent) : f_(significand), e_(exponent) {}

  // this -= other.
  // The exponents of both numbers must be the same and the significand of this
  // must be greater or equal than the significand of other.
  // The result will not be normalized.
  void Subtract(const DiyFp& other) {
    DOUBLE_CONVERSION_ASSERT(e_ == other.e_);
    DOUBLE_CONVERSION_ASSERT(f_ >= other.f_);
    f_ -= other.f_;
  }

  // Returns a - b.
  // The exponents of both numbers must be the same and a must be greater
  // or equal than b. The result will not be normalized.
  static DiyFp Minus(const DiyFp& a, const DiyFp& b) {
    DiyFp result = a;
    result.Subtract(b);
    return result;
  }

  // this *= other.
  void Multiply(const DiyFp& other) {
    // Simply "emulates" a 128 bit multiplication.
    // However: the resulting number only contains 64 bits. The least
    // significant 64 bits are only used for rounding the most significant 64
    // bits.
    const uint64_t kM32 = 0xFFFFFFFFU;
    const uint64_t a = f_ >> 32;
    const uint64_t b = f_ & kM32;
    const uint64_t c = other.f_ >> 32;
    const uint64_t d = other.f_ & kM32;
    const uint64_t ac = a * c;
    const uint64_t bc = b * c;
    const uint64_t ad = a * d;
    const uint64_t bd = b * d;
    // By adding 1U << 31 to tmp we round the final result.
    // Halfway cases will be rounded up.
    const uint64_t tmp = (bd >> 32) + (ad & kM32) + (bc & kM32) + (1U << 31);
    e_ += other.e_ + 64;
    f_ = ac + (ad >> 32) + (bc >> 32) + (tmp >> 32);
  }

  // returns a * b;
  static DiyFp Times(const DiyFp& a, const DiyFp& b) {
    DiyFp result = a;
    result.Multiply(b);
    return result;
  }

  void Normalize() {
    DOUBLE_CONVERSION_ASSERT(f_ != 0);
    uint64_t significand = f_;
    int32_t exponent = e_;

    // This method is mainly called for normalizing boundaries. In general,
    // boundaries need to be shifted by 10 bits, and we optimize for this case.
    const uint64_t k10MSBits = DOUBLE_CONVERSION_UINT64_2PART_C(0xFFC00000, 00000000);
    while ((significand & k10MSBits) == 0) {
      significand <<= 10;
      exponent -= 10;
    }
    while ((significand & kUint64MSB) == 0) {
      significand <<= 1;
      exponent--;
    }
    f_ = significand;
    e_ = exponent;
  }

  static DiyFp Normalize(const DiyFp& a) {
    DiyFp result = a;
    result.Normalize();
    return result;
  }

  uint64_t f() const { return f_; }
  int32_t e() const { return e_; }

  void set_f(uint64_t new_value) { f_ = new_value; }
  void set_e(int32_t new_value) { e_ = new_value; }

 private:
  static const uint64_t kUint64MSB = DOUBLE_CONVERSION_UINT64_2PART_C(0x80000000, 00000000);

  uint64_t f_;
  int32_t e_;
};

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END

#endif  // DOUBLE_CONVERSION_DIY_FP_H_


// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

// We assume that doubles and uint64_t have the same endianness.
static uint64_t double_to_uint64(double d) { return BitCast<uint64_t>(d); }
static double uint64_to_double(uint64_t d64) { return BitCast<double>(d64); }
static uint32_t float_to_uint32(float f) { return BitCast<uint32_t>(f); }
static float uint32_to_float(uint32_t d32) { return BitCast<float>(d32); }

// Helper functions for doubles.
class Double {
 public:
  static const uint64_t kSignMask = DOUBLE_CONVERSION_UINT64_2PART_C(0x80000000, 00000000);
  static const uint64_t kExponentMask = DOUBLE_CONVERSION_UINT64_2PART_C(0x7FF00000, 00000000);
  static const uint64_t kSignificandMask = DOUBLE_CONVERSION_UINT64_2PART_C(0x000FFFFF, FFFFFFFF);
  static const uint64_t kHiddenBit = DOUBLE_CONVERSION_UINT64_2PART_C(0x00100000, 00000000);
  static const int kPhysicalSignificandSize = 52;  // Excludes the hidden bit.
  static const int kSignificandSize = 53;
  static const int kExponentBias = 0x3FF + kPhysicalSignificandSize;
  static const int kMaxExponent = 0x7FF - kExponentBias;

  Double() : d64_(0) {}
  explicit Double(double d) : d64_(double_to_uint64(d)) {}
  explicit Double(uint64_t d64) : d64_(d64) {}
  explicit Double(DiyFp diy_fp)
    : d64_(DiyFpToUint64(diy_fp)) {}

  // The value encoded by this Double must be greater or equal to +0.0.
  // It must not be special (infinity, or NaN).
  DiyFp AsDiyFp() const {
    DOUBLE_CONVERSION_ASSERT(Sign() > 0);
    DOUBLE_CONVERSION_ASSERT(!IsSpecial());
    return DiyFp(Significand(), Exponent());
  }

  // The value encoded by this Double must be strictly greater than 0.
  DiyFp AsNormalizedDiyFp() const {
    DOUBLE_CONVERSION_ASSERT(value() > 0.0);
    uint64_t f = Significand();
    int e = Exponent();

    // The current double could be a denormal.
    while ((f & kHiddenBit) == 0) {
      f <<= 1;
      e--;
    }
    // Do the final shifts in one go.
    f <<= DiyFp::kSignificandSize - kSignificandSize;
    e -= DiyFp::kSignificandSize - kSignificandSize;
    return DiyFp(f, e);
  }

  // Returns the double's bit as uint64.
  uint64_t AsUint64() const {
    return d64_;
  }

  // Returns the next greater double. Returns +infinity on input +infinity.
  double NextDouble() const {
    if (d64_ == kInfinity) return Double(kInfinity).value();
    if (Sign() < 0 && Significand() == 0) {
      // -0.0
      return 0.0;
    }
    if (Sign() < 0) {
      return Double(d64_ - 1).value();
    } else {
      return Double(d64_ + 1).value();
    }
  }

  double PreviousDouble() const {
    if (d64_ == (kInfinity | kSignMask)) return -Infinity();
    if (Sign() < 0) {
      return Double(d64_ + 1).value();
    } else {
      if (Significand() == 0) return -0.0;
      return Double(d64_ - 1).value();
    }
  }

  int Exponent() const {
    if (IsDenormal()) return kDenormalExponent;

    uint64_t d64 = AsUint64();
    int biased_e =
        static_cast<int>((d64 & kExponentMask) >> kPhysicalSignificandSize);
    return biased_e - kExponentBias;
  }

  uint64_t Significand() const {
    uint64_t d64 = AsUint64();
    uint64_t significand = d64 & kSignificandMask;
    if (!IsDenormal()) {
      return significand + kHiddenBit;
    } else {
      return significand;
    }
  }

  // Returns true if the double is a denormal.
  bool IsDenormal() const {
    uint64_t d64 = AsUint64();
    return (d64 & kExponentMask) == 0;
  }

  // We consider denormals not to be special.
  // Hence only Infinity and NaN are special.
  bool IsSpecial() const {
    uint64_t d64 = AsUint64();
    return (d64 & kExponentMask) == kExponentMask;
  }

  bool IsNan() const {
    uint64_t d64 = AsUint64();
    return ((d64 & kExponentMask) == kExponentMask) &&
        ((d64 & kSignificandMask) != 0);
  }

  bool IsInfinite() const {
    uint64_t d64 = AsUint64();
    return ((d64 & kExponentMask) == kExponentMask) &&
        ((d64 & kSignificandMask) == 0);
  }

  int Sign() const {
    uint64_t d64 = AsUint64();
    return (d64 & kSignMask) == 0? 1: -1;
  }

  // Precondition: the value encoded by this Double must be greater or equal
  // than +0.0.
  DiyFp UpperBoundary() const {
    DOUBLE_CONVERSION_ASSERT(Sign() > 0);
    return DiyFp(Significand() * 2 + 1, Exponent() - 1);
  }

  // Computes the two boundaries of this.
  // The bigger boundary (m_plus) is normalized. The lower boundary has the same
  // exponent as m_plus.
  // Precondition: the value encoded by this Double must be greater than 0.
  void NormalizedBoundaries(DiyFp* out_m_minus, DiyFp* out_m_plus) const {
    DOUBLE_CONVERSION_ASSERT(value() > 0.0);
    DiyFp v = this->AsDiyFp();
    DiyFp m_plus = DiyFp::Normalize(DiyFp((v.f() << 1) + 1, v.e() - 1));
    DiyFp m_minus;
    if (LowerBoundaryIsCloser()) {
      m_minus = DiyFp((v.f() << 2) - 1, v.e() - 2);
    } else {
      m_minus = DiyFp((v.f() << 1) - 1, v.e() - 1);
    }
    m_minus.set_f(m_minus.f() << (m_minus.e() - m_plus.e()));
    m_minus.set_e(m_plus.e());
    *out_m_plus = m_plus;
    *out_m_minus = m_minus;
  }

  bool LowerBoundaryIsCloser() const {
    // The boundary is closer if the significand is of the form f == 2^p-1 then
    // the lower boundary is closer.
    // Think of v = 1000e10 and v- = 9999e9.
    // Then the boundary (== (v - v-)/2) is not just at a distance of 1e9 but
    // at a distance of 1e8.
    // The only exception is for the smallest normal: the largest denormal is
    // at the same distance as its successor.
    // Note: denormals have the same exponent as the smallest normals.
    bool physical_significand_is_zero = ((AsUint64() & kSignificandMask) == 0);
    return physical_significand_is_zero && (Exponent() != kDenormalExponent);
  }

  double value() const { return uint64_to_double(d64_); }

  // Returns the significand size for a given order of magnitude.
  // If v = f*2^e with 2^p-1 <= f <= 2^p then p+e is v's order of magnitude.
  // This function returns the number of significant binary digits v will have
  // once it's encoded into a double. In almost all cases this is equal to
  // kSignificandSize. The only exceptions are denormals. They start with
  // leading zeroes and their effective significand-size is hence smaller.
  static int SignificandSizeForOrderOfMagnitude(int order) {
    if (order >= (kDenormalExponent + kSignificandSize)) {
      return kSignificandSize;
    }
    if (order <= kDenormalExponent) return 0;
    return order - kDenormalExponent;
  }

  static double Infinity() {
    return Double(kInfinity).value();
  }

  static double NaN() {
    return Double(kNaN).value();
  }

 private:
  static const int kDenormalExponent = -kExponentBias + 1;
  static const uint64_t kInfinity = DOUBLE_CONVERSION_UINT64_2PART_C(0x7FF00000, 00000000);
  static const uint64_t kNaN = DOUBLE_CONVERSION_UINT64_2PART_C(0x7FF80000, 00000000);

  const uint64_t d64_;

  static uint64_t DiyFpToUint64(DiyFp diy_fp) {
    uint64_t significand = diy_fp.f();
    int exponent = diy_fp.e();
    while (significand > kHiddenBit + kSignificandMask) {
      significand >>= 1;
      exponent++;
    }
    if (exponent >= kMaxExponent) {
      return kInfinity;
    }
    if (exponent < kDenormalExponent) {
      return 0;
    }
    while (exponent > kDenormalExponent && (significand & kHiddenBit) == 0) {
      significand <<= 1;
      exponent--;
    }
    uint64_t biased_exponent;
    if (exponent == kDenormalExponent && (significand & kHiddenBit) == 0) {
      biased_exponent = 0;
    } else {
      biased_exponent = static_cast<uint64_t>(exponent + kExponentBias);
    }
    return (significand & kSignificandMask) |
        (biased_exponent << kPhysicalSignificandSize);
  }

  DOUBLE_CONVERSION_DISALLOW_COPY_AND_ASSIGN(Double);
};

class Single {
 public:
  static const uint32_t kSignMask = 0x80000000;
  static const uint32_t kExponentMask = 0x7F800000;
  static const uint32_t kSignificandMask = 0x007FFFFF;
  static const uint32_t kHiddenBit = 0x00800000;
  static const int kPhysicalSignificandSize = 23;  // Excludes the hidden bit.
  static const int kSignificandSize = 24;

  Single() : d32_(0) {}
  explicit Single(float f) : d32_(float_to_uint32(f)) {}
  explicit Single(uint32_t d32) : d32_(d32) {}

  // The value encoded by this Single must be greater or equal to +0.0.
  // It must not be special (infinity, or NaN).
  DiyFp AsDiyFp() const {
    DOUBLE_CONVERSION_ASSERT(Sign() > 0);
    DOUBLE_CONVERSION_ASSERT(!IsSpecial());
    return DiyFp(Significand(), Exponent());
  }

  // Returns the single's bit as uint64.
  uint32_t AsUint32() const {
    return d32_;
  }

  int Exponent() const {
    if (IsDenormal()) return kDenormalExponent;

    uint32_t d32 = AsUint32();
    int biased_e =
        static_cast<int>((d32 & kExponentMask) >> kPhysicalSignificandSize);
    return biased_e - kExponentBias;
  }

  uint32_t Significand() const {
    uint32_t d32 = AsUint32();
    uint32_t significand = d32 & kSignificandMask;
    if (!IsDenormal()) {
      return significand + kHiddenBit;
    } else {
      return significand;
    }
  }

  // Returns true if the single is a denormal.
  bool IsDenormal() const {
    uint32_t d32 = AsUint32();
    return (d32 & kExponentMask) == 0;
  }

  // We consider denormals not to be special.
  // Hence only Infinity and NaN are special.
  bool IsSpecial() const {
    uint32_t d32 = AsUint32();
    return (d32 & kExponentMask) == kExponentMask;
  }

  bool IsNan() const {
    uint32_t d32 = AsUint32();
    return ((d32 & kExponentMask) == kExponentMask) &&
        ((d32 & kSignificandMask) != 0);
  }

  bool IsInfinite() const {
    uint32_t d32 = AsUint32();
    return ((d32 & kExponentMask) == kExponentMask) &&
        ((d32 & kSignificandMask) == 0);
  }

  int Sign() const {
    uint32_t d32 = AsUint32();
    return (d32 & kSignMask) == 0? 1: -1;
  }

  // Computes the two boundaries of this.
  // The bigger boundary (m_plus) is normalized. The lower boundary has the same
  // exponent as m_plus.
  // Precondition: the value encoded by this Single must be greater than 0.
  void NormalizedBoundaries(DiyFp* out_m_minus, DiyFp* out_m_plus) const {
    DOUBLE_CONVERSION_ASSERT(value() > 0.0);
    DiyFp v = this->AsDiyFp();
    DiyFp m_plus = DiyFp::Normalize(DiyFp((v.f() << 1) + 1, v.e() - 1));
    DiyFp m_minus;
    if (LowerBoundaryIsCloser()) {
      m_minus = DiyFp((v.f() << 2) - 1, v.e() - 2);
    } else {
      m_minus = DiyFp((v.f() << 1) - 1, v.e() - 1);
    }
    m_minus.set_f(m_minus.f() << (m_minus.e() - m_plus.e()));
    m_minus.set_e(m_plus.e());
    *out_m_plus = m_plus;
    *out_m_minus = m_minus;
  }

  // Precondition: the value encoded by this Single must be greater or equal
  // than +0.0.
  DiyFp UpperBoundary() const {
    DOUBLE_CONVERSION_ASSERT(Sign() > 0);
    return DiyFp(Significand() * 2 + 1, Exponent() - 1);
  }

  bool LowerBoundaryIsCloser() const {
    // The boundary is closer if the significand is of the form f == 2^p-1 then
    // the lower boundary is closer.
    // Think of v = 1000e10 and v- = 9999e9.
    // Then the boundary (== (v - v-)/2) is not just at a distance of 1e9 but
    // at a distance of 1e8.
    // The only exception is for the smallest normal: the largest denormal is
    // at the same distance as its successor.
    // Note: denormals have the same exponent as the smallest normals.
    bool physical_significand_is_zero = ((AsUint32() & kSignificandMask) == 0);
    return physical_significand_is_zero && (Exponent() != kDenormalExponent);
  }

  float value() const { return uint32_to_float(d32_); }

  static float Infinity() {
    return Single(kInfinity).value();
  }

  static float NaN() {
    return Single(kNaN).value();
  }

 private:
  static const int kExponentBias = 0x7F + kPhysicalSignificandSize;
  static const int kDenormalExponent = -kExponentBias + 1;
  static const int kMaxExponent = 0xFF - kExponentBias;
  static const uint32_t kInfinity = 0x7F800000;
  static const uint32_t kNaN = 0x7FC00000;

  const uint32_t d32_;

  DOUBLE_CONVERSION_DISALLOW_COPY_AND_ASSIGN(Single);
};

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END

#endif  // DOUBLE_CONVERSION_DOUBLE_H_


// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

static int NormalizedExponent(uint64_t significand, int exponent) {
  DOUBLE_CONVERSION_ASSERT(significand != 0);
  while ((significand & Double::kHiddenBit) == 0) {
    significand = significand << 1;
    exponent = exponent - 1;
  }
  return exponent;
}


// Forward declarations:
// Returns an estimation of k such that 10^(k-1) <= v < 10^k.
static int EstimatePower(int exponent);
// Computes v / 10^estimated_power exactly, as a ratio of two bignums, numerator
// and denominator.
static void InitialScaledStartValues(uint64_t significand,
                                     int exponent,
                                     bool lower_boundary_is_closer,
                                     int estimated_power,
                                     bool need_boundary_deltas,
                                     Bignum* numerator,
                                     Bignum* denominator,
                                     Bignum* delta_minus,
                                     Bignum* delta_plus);
// Multiplies numerator/denominator so that its values lies in the range 1-10.
// Returns decimal_point s.t.
//  v = numerator'/denominator' * 10^(decimal_point-1)
//     where numerator' and denominator' are the values of numerator and
//     denominator after the call to this function.
static void FixupMultiply10(int estimated_power, bool is_even,
                            int* decimal_point,
                            Bignum* numerator, Bignum* denominator,
                            Bignum* delta_minus, Bignum* delta_plus);
// Generates digits from the left to the right and stops when the generated
// digits yield the shortest decimal representation of v.
static void GenerateShortestDigits(Bignum* numerator, Bignum* denominator,
                                   Bignum* delta_minus, Bignum* delta_plus,
                                   bool is_even,
                                   Vector<char> buffer, int* length);
// Generates 'requested_digits' after the decimal point.
static void BignumToFixed(int requested_digits, int* decimal_point,
                          Bignum* numerator, Bignum* denominator,
                          Vector<char> buffer, int* length);
// Generates 'count' digits of numerator/denominator.
// Once 'count' digits have been produced rounds the result depending on the
// remainder (remainders of exactly .5 round upwards). Might update the
// decimal_point when rounding up (for example for 0.9999).
static void GenerateCountedDigits(int count, int* decimal_point,
                                  Bignum* numerator, Bignum* denominator,
                                  Vector<char> buffer, int* length);


void BignumDtoa(double v, BignumDtoaMode mode, int requested_digits,
                Vector<char> buffer, int* length, int* decimal_point) {
  DOUBLE_CONVERSION_ASSERT(v > 0);
  DOUBLE_CONVERSION_ASSERT(!Double(v).IsSpecial());
  uint64_t significand;
  int exponent;
  bool lower_boundary_is_closer;
  if (mode == BIGNUM_DTOA_SHORTEST_SINGLE) {
    float f = static_cast<float>(v);
    DOUBLE_CONVERSION_ASSERT(f == v);
    significand = Single(f).Significand();
    exponent = Single(f).Exponent();
    lower_boundary_is_closer = Single(f).LowerBoundaryIsCloser();
  } else {
    significand = Double(v).Significand();
    exponent = Double(v).Exponent();
    lower_boundary_is_closer = Double(v).LowerBoundaryIsCloser();
  }
  bool need_boundary_deltas =
      (mode == BIGNUM_DTOA_SHORTEST || mode == BIGNUM_DTOA_SHORTEST_SINGLE);

  bool is_even = (significand & 1) == 0;
  int normalized_exponent = NormalizedExponent(significand, exponent);
  // estimated_power might be too low by 1.
  int estimated_power = EstimatePower(normalized_exponent);

  // Shortcut for Fixed.
  // The requested digits correspond to the digits after the point. If the
  // number is much too small, then there is no need in trying to get any
  // digits.
  if (mode == BIGNUM_DTOA_FIXED && -estimated_power - 1 > requested_digits) {
    buffer[0] = '\0';
    *length = 0;
    // Set decimal-point to -requested_digits. This is what Gay does.
    // Note that it should not have any effect anyways since the string is
    // empty.
    *decimal_point = -requested_digits;
    return;
  }

  Bignum numerator;
  Bignum denominator;
  Bignum delta_minus;
  Bignum delta_plus;
  // Make sure the bignum can grow large enough. The smallest double equals
  // 4e-324. In this case the denominator needs fewer than 324*4 binary digits.
  // The maximum double is 1.7976931348623157e308 which needs fewer than
  // 308*4 binary digits.
  DOUBLE_CONVERSION_ASSERT(Bignum::kMaxSignificantBits >= 324*4);
  InitialScaledStartValues(significand, exponent, lower_boundary_is_closer,
                           estimated_power, need_boundary_deltas,
                           &numerator, &denominator,
                           &delta_minus, &delta_plus);
  // We now have v = (numerator / denominator) * 10^estimated_power.
  FixupMultiply10(estimated_power, is_even, decimal_point,
                  &numerator, &denominator,
                  &delta_minus, &delta_plus);
  // We now have v = (numerator / denominator) * 10^(decimal_point-1), and
  //  1 <= (numerator + delta_plus) / denominator < 10
  switch (mode) {
    case BIGNUM_DTOA_SHORTEST:
    case BIGNUM_DTOA_SHORTEST_SINGLE:
      GenerateShortestDigits(&numerator, &denominator,
                             &delta_minus, &delta_plus,
                             is_even, buffer, length);
      break;
    case BIGNUM_DTOA_FIXED:
      BignumToFixed(requested_digits, decimal_point,
                    &numerator, &denominator,
                    buffer, length);
      break;
    case BIGNUM_DTOA_PRECISION:
      GenerateCountedDigits(requested_digits, decimal_point,
                            &numerator, &denominator,
                            buffer, length);
      break;
    default:
      DOUBLE_CONVERSION_UNREACHABLE();
  }
  buffer[*length] = '\0';
}


// The procedure starts generating digits from the left to the right and stops
// when the generated digits yield the shortest decimal representation of v. A
// decimal representation of v is a number lying closer to v than to any other
// double, so it converts to v when read.
//
// This is true if d, the decimal representation, is between m- and m+, the
// upper and lower boundaries. d must be strictly between them if !is_even.
//           m- := (numerator - delta_minus) / denominator
//           m+ := (numerator + delta_plus) / denominator
//
// Precondition: 0 <= (numerator+delta_plus) / denominator < 10.
//   If 1 <= (numerator+delta_plus) / denominator < 10 then no leading 0 digit
//   will be produced. This should be the standard precondition.
static void GenerateShortestDigits(Bignum* numerator, Bignum* denominator,
                                   Bignum* delta_minus, Bignum* delta_plus,
                                   bool is_even,
                                   Vector<char> buffer, int* length) {
  // Small optimization: if delta_minus and delta_plus are the same just reuse
  // one of the two bignums.
  if (Bignum::Equal(*delta_minus, *delta_plus)) {
    delta_plus = delta_minus;
  }
  *length = 0;
  for (;;) {
    uint16_t digit;
    digit = numerator->DivideModuloIntBignum(*denominator);
    DOUBLE_CONVERSION_ASSERT(digit <= 9);  // digit is a uint16_t and therefore always positive.
    // digit = numerator / denominator (integer division).
    // numerator = numerator % denominator.
    buffer[(*length)++] = static_cast<char>(digit + '0');

    // Can we stop already?
    // If the remainder of the division is less than the distance to the lower
    // boundary we can stop. In this case we simply round down (discarding the
    // remainder).
    // Similarly we test if we can round up (using the upper boundary).
    bool in_delta_room_minus;
    bool in_delta_room_plus;
    if (is_even) {
      in_delta_room_minus = Bignum::LessEqual(*numerator, *delta_minus);
    } else {
      in_delta_room_minus = Bignum::Less(*numerator, *delta_minus);
    }
    if (is_even) {
      in_delta_room_plus =
          Bignum::PlusCompare(*numerator, *delta_plus, *denominator) >= 0;
    } else {
      in_delta_room_plus =
          Bignum::PlusCompare(*numerator, *delta_plus, *denominator) > 0;
    }
    if (!in_delta_room_minus && !in_delta_room_plus) {
      // Prepare for next iteration.
      numerator->Times10();
      delta_minus->Times10();
      // We optimized delta_plus to be equal to delta_minus (if they share the
      // same value). So don't multiply delta_plus if they point to the same
      // object.
      if (delta_minus != delta_plus) {
        delta_plus->Times10();
      }
    } else if (in_delta_room_minus && in_delta_room_plus) {
      // Let's see if 2*numerator < denominator.
      // If yes, then the next digit would be < 5 and we can round down.
      int compare = Bignum::PlusCompare(*numerator, *numerator, *denominator);
      if (compare < 0) {
        // Remaining digits are less than .5. -> Round down (== do nothing).
      } else if (compare > 0) {
        // Remaining digits are more than .5 of denominator. -> Round up.
        // Note that the last digit could not be a '9' as otherwise the whole
        // loop would have stopped earlier.
        // We still have an assert here in case the preconditions were not
        // satisfied.
        DOUBLE_CONVERSION_ASSERT(buffer[(*length) - 1] != '9');
        buffer[(*length) - 1]++;
      } else {
        // Halfway case.
        // TODO(floitsch): need a way to solve half-way cases.
        //   For now let's round towards even (since this is what Gay seems to
        //   do).

        if ((buffer[(*length) - 1] - '0') % 2 == 0) {
          // Round down => Do nothing.
        } else {
          DOUBLE_CONVERSION_ASSERT(buffer[(*length) - 1] != '9');
          buffer[(*length) - 1]++;
        }
      }
      return;
    } else if (in_delta_room_minus) {
      // Round down (== do nothing).
      return;
    } else {  // in_delta_room_plus
      // Round up.
      // Note again that the last digit could not be '9' since this would have
      // stopped the loop earlier.
      // We still have an DOUBLE_CONVERSION_ASSERT here, in case the preconditions were not
      // satisfied.
      DOUBLE_CONVERSION_ASSERT(buffer[(*length) -1] != '9');
      buffer[(*length) - 1]++;
      return;
    }
  }
}


// Let v = numerator / denominator < 10.
// Then we generate 'count' digits of d = x.xxxxx... (without the decimal point)
// from left to right. Once 'count' digits have been produced we decide wether
// to round up or down. Remainders of exactly .5 round upwards. Numbers such
// as 9.999999 propagate a carry all the way, and change the
// exponent (decimal_point), when rounding upwards.
static void GenerateCountedDigits(int count, int* decimal_point,
                                  Bignum* numerator, Bignum* denominator,
                                  Vector<char> buffer, int* length) {
  DOUBLE_CONVERSION_ASSERT(count >= 0);
  for (int i = 0; i < count - 1; ++i) {
    uint16_t digit;
    digit = numerator->DivideModuloIntBignum(*denominator);
    DOUBLE_CONVERSION_ASSERT(digit <= 9);  // digit is a uint16_t and therefore always positive.
    // digit = numerator / denominator (integer division).
    // numerator = numerator % denominator.
    buffer[i] = static_cast<char>(digit + '0');
    // Prepare for next iteration.
    numerator->Times10();
  }
  // Generate the last digit.
  uint16_t digit;
  digit = numerator->DivideModuloIntBignum(*denominator);
  if (Bignum::PlusCompare(*numerator, *numerator, *denominator) >= 0) {
    digit++;
  }
  DOUBLE_CONVERSION_ASSERT(digit <= 10);
  buffer[count - 1] = static_cast<char>(digit + '0');
  // Correct bad digits (in case we had a sequence of '9's). Propagate the
  // carry until we hat a non-'9' or til we reach the first digit.
  for (int i = count - 1; i > 0; --i) {
    if (buffer[i] != '0' + 10) break;
    buffer[i] = '0';
    buffer[i - 1]++;
  }
  if (buffer[0] == '0' + 10) {
    // Propagate a carry past the top place.
    buffer[0] = '1';
    (*decimal_point)++;
  }
  *length = count;
}


// Generates 'requested_digits' after the decimal point. It might omit
// trailing '0's. If the input number is too small then no digits at all are
// generated (ex.: 2 fixed digits for 0.00001).
//
// Input verifies:  1 <= (numerator + delta) / denominator < 10.
static void BignumToFixed(int requested_digits, int* decimal_point,
                          Bignum* numerator, Bignum* denominator,
                          Vector<char> buffer, int* length) {
  // Note that we have to look at more than just the requested_digits, since
  // a number could be rounded up. Example: v=0.5 with requested_digits=0.
  // Even though the power of v equals 0 we can't just stop here.
  if (-(*decimal_point) > requested_digits) {
    // The number is definitively too small.
    // Ex: 0.001 with requested_digits == 1.
    // Set decimal-point to -requested_digits. This is what Gay does.
    // Note that it should not have any effect anyways since the string is
    // empty.
    *decimal_point = -requested_digits;
    *length = 0;
    return;
  } else if (-(*decimal_point) == requested_digits) {
    // We only need to verify if the number rounds down or up.
    // Ex: 0.04 and 0.06 with requested_digits == 1.
    DOUBLE_CONVERSION_ASSERT(*decimal_point == -requested_digits);
    // Initially the fraction lies in range (1, 10]. Multiply the denominator
    // by 10 so that we can compare more easily.
    denominator->Times10();
    if (Bignum::PlusCompare(*numerator, *numerator, *denominator) >= 0) {
      // If the fraction is >= 0.5 then we have to include the rounded
      // digit.
      buffer[0] = '1';
      *length = 1;
      (*decimal_point)++;
    } else {
      // Note that we caught most of similar cases earlier.
      *length = 0;
    }
    return;
  } else {
    // The requested digits correspond to the digits after the point.
    // The variable 'needed_digits' includes the digits before the point.
    int needed_digits = (*decimal_point) + requested_digits;
    GenerateCountedDigits(needed_digits, decimal_point,
                          numerator, denominator,
                          buffer, length);
  }
}


// Returns an estimation of k such that 10^(k-1) <= v < 10^k where
// v = f * 2^exponent and 2^52 <= f < 2^53.
// v is hence a normalized double with the given exponent. The output is an
// approximation for the exponent of the decimal approimation .digits * 10^k.
//
// The result might undershoot by 1 in which case 10^k <= v < 10^k+1.
// Note: this property holds for v's upper boundary m+ too.
//    10^k <= m+ < 10^k+1.
//   (see explanation below).
//
// Examples:
//  EstimatePower(0)   => 16
//  EstimatePower(-52) => 0
//
// Note: e >= 0 => EstimatedPower(e) > 0. No similar claim can be made for e<0.
static int EstimatePower(int exponent) {
  // This function estimates log10 of v where v = f*2^e (with e == exponent).
  // Note that 10^floor(log10(v)) <= v, but v <= 10^ceil(log10(v)).
  // Note that f is bounded by its container size. Let p = 53 (the double's
  // significand size). Then 2^(p-1) <= f < 2^p.
  //
  // Given that log10(v) == log2(v)/log2(10) and e+(len(f)-1) is quite close
  // to log2(v) the function is simplified to (e+(len(f)-1)/log2(10)).
  // The computed number undershoots by less than 0.631 (when we compute log3
  // and not log10).
  //
  // Optimization: since we only need an approximated result this computation
  // can be performed on 64 bit integers. On x86/x64 architecture the speedup is
  // not really measurable, though.
  //
  // Since we want to avoid overshooting we decrement by 1e10 so that
  // floating-point imprecisions don't affect us.
  //
  // Explanation for v's boundary m+: the computation takes advantage of
  // the fact that 2^(p-1) <= f < 2^p. Boundaries still satisfy this requirement
  // (even for denormals where the delta can be much more important).

  const double k1Log10 = 0.30102999566398114;  // 1/lg(10)

  // For doubles len(f) == 53 (don't forget the hidden bit).
  const int kSignificandSize = Double::kSignificandSize;
  double estimate = ceil((exponent + kSignificandSize - 1) * k1Log10 - 1e-10);
  return static_cast<int>(estimate);
}


// See comments for InitialScaledStartValues.
static void InitialScaledStartValuesPositiveExponent(
    uint64_t significand, int exponent,
    int estimated_power, bool need_boundary_deltas,
    Bignum* numerator, Bignum* denominator,
    Bignum* delta_minus, Bignum* delta_plus) {
  // A positive exponent implies a positive power.
  DOUBLE_CONVERSION_ASSERT(estimated_power >= 0);
  // Since the estimated_power is positive we simply multiply the denominator
  // by 10^estimated_power.

  // numerator = v.
  numerator->AssignUInt64(significand);
  numerator->ShiftLeft(exponent);
  // denominator = 10^estimated_power.
  denominator->AssignPowerUInt16(10, estimated_power);

  if (need_boundary_deltas) {
    // Introduce a common denominator so that the deltas to the boundaries are
    // integers.
    denominator->ShiftLeft(1);
    numerator->ShiftLeft(1);
    // Let v = f * 2^e, then m+ - v = 1/2 * 2^e; With the common
    // denominator (of 2) delta_plus equals 2^e.
    delta_plus->AssignUInt16(1);
    delta_plus->ShiftLeft(exponent);
    // Same for delta_minus. The adjustments if f == 2^p-1 are done later.
    delta_minus->AssignUInt16(1);
    delta_minus->ShiftLeft(exponent);
  }
}


// See comments for InitialScaledStartValues
static void InitialScaledStartValuesNegativeExponentPositivePower(
    uint64_t significand, int exponent,
    int estimated_power, bool need_boundary_deltas,
    Bignum* numerator, Bignum* denominator,
    Bignum* delta_minus, Bignum* delta_plus) {
  // v = f * 2^e with e < 0, and with estimated_power >= 0.
  // This means that e is close to 0 (have a look at how estimated_power is
  // computed).

  // numerator = significand
  //  since v = significand * 2^exponent this is equivalent to
  //  numerator = v * / 2^-exponent
  numerator->AssignUInt64(significand);
  // denominator = 10^estimated_power * 2^-exponent (with exponent < 0)
  denominator->AssignPowerUInt16(10, estimated_power);
  denominator->ShiftLeft(-exponent);

  if (need_boundary_deltas) {
    // Introduce a common denominator so that the deltas to the boundaries are
    // integers.
    denominator->ShiftLeft(1);
    numerator->ShiftLeft(1);
    // Let v = f * 2^e, then m+ - v = 1/2 * 2^e; With the common
    // denominator (of 2) delta_plus equals 2^e.
    // Given that the denominator already includes v's exponent the distance
    // to the boundaries is simply 1.
    delta_plus->AssignUInt16(1);
    // Same for delta_minus. The adjustments if f == 2^p-1 are done later.
    delta_minus->AssignUInt16(1);
  }
}


// See comments for InitialScaledStartValues
static void InitialScaledStartValuesNegativeExponentNegativePower(
    uint64_t significand, int exponent,
    int estimated_power, bool need_boundary_deltas,
    Bignum* numerator, Bignum* denominator,
    Bignum* delta_minus, Bignum* delta_plus) {
  // Instead of multiplying the denominator with 10^estimated_power we
  // multiply all values (numerator and deltas) by 10^-estimated_power.

  // Use numerator as temporary container for power_ten.
  Bignum* power_ten = numerator;
  power_ten->AssignPowerUInt16(10, -estimated_power);

  if (need_boundary_deltas) {
    // Since power_ten == numerator we must make a copy of 10^estimated_power
    // before we complete the computation of the numerator.
    // delta_plus = delta_minus = 10^estimated_power
    delta_plus->AssignBignum(*power_ten);
    delta_minus->AssignBignum(*power_ten);
  }

  // numerator = significand * 2 * 10^-estimated_power
  //  since v = significand * 2^exponent this is equivalent to
  // numerator = v * 10^-estimated_power * 2 * 2^-exponent.
  // Remember: numerator has been abused as power_ten. So no need to assign it
  //  to itself.
  DOUBLE_CONVERSION_ASSERT(numerator == power_ten);
  numerator->MultiplyByUInt64(significand);

  // denominator = 2 * 2^-exponent with exponent < 0.
  denominator->AssignUInt16(1);
  denominator->ShiftLeft(-exponent);

  if (need_boundary_deltas) {
    // Introduce a common denominator so that the deltas to the boundaries are
    // integers.
    numerator->ShiftLeft(1);
    denominator->ShiftLeft(1);
    // With this shift the boundaries have their correct value, since
    // delta_plus = 10^-estimated_power, and
    // delta_minus = 10^-estimated_power.
    // These assignments have been done earlier.
    // The adjustments if f == 2^p-1 (lower boundary is closer) are done later.
  }
}


// Let v = significand * 2^exponent.
// Computes v / 10^estimated_power exactly, as a ratio of two bignums, numerator
// and denominator. The functions GenerateShortestDigits and
// GenerateCountedDigits will then convert this ratio to its decimal
// representation d, with the required accuracy.
// Then d * 10^estimated_power is the representation of v.
// (Note: the fraction and the estimated_power might get adjusted before
// generating the decimal representation.)
//
// The initial start values consist of:
//  - a scaled numerator: s.t. numerator/denominator == v / 10^estimated_power.
//  - a scaled (common) denominator.
//  optionally (used by GenerateShortestDigits to decide if it has the shortest
//  decimal converting back to v):
//  - v - m-: the distance to the lower boundary.
//  - m+ - v: the distance to the upper boundary.
//
// v, m+, m-, and therefore v - m- and m+ - v all share the same denominator.
//
// Let ep == estimated_power, then the returned values will satisfy:
//  v / 10^ep = numerator / denominator.
//  v's boundarys m- and m+:
//    m- / 10^ep == v / 10^ep - delta_minus / denominator
//    m+ / 10^ep == v / 10^ep + delta_plus / denominator
//  Or in other words:
//    m- == v - delta_minus * 10^ep / denominator;
//    m+ == v + delta_plus * 10^ep / denominator;
//
// Since 10^(k-1) <= v < 10^k    (with k == estimated_power)
//  or       10^k <= v < 10^(k+1)
//  we then have 0.1 <= numerator/denominator < 1
//           or    1 <= numerator/denominator < 10
//
// It is then easy to kickstart the digit-generation routine.
//
// The boundary-deltas are only filled if the mode equals BIGNUM_DTOA_SHORTEST
// or BIGNUM_DTOA_SHORTEST_SINGLE.

static void InitialScaledStartValues(uint64_t significand,
                                     int exponent,
                                     bool lower_boundary_is_closer,
                                     int estimated_power,
                                     bool need_boundary_deltas,
                                     Bignum* numerator,
                                     Bignum* denominator,
                                     Bignum* delta_minus,
                                     Bignum* delta_plus) {
  if (exponent >= 0) {
    InitialScaledStartValuesPositiveExponent(
        significand, exponent, estimated_power, need_boundary_deltas,
        numerator, denominator, delta_minus, delta_plus);
  } else if (estimated_power >= 0) {
    InitialScaledStartValuesNegativeExponentPositivePower(
        significand, exponent, estimated_power, need_boundary_deltas,
        numerator, denominator, delta_minus, delta_plus);
  } else {
    InitialScaledStartValuesNegativeExponentNegativePower(
        significand, exponent, estimated_power, need_boundary_deltas,
        numerator, denominator, delta_minus, delta_plus);
  }

  if (need_boundary_deltas && lower_boundary_is_closer) {
    // The lower boundary is closer at half the distance of "normal" numbers.
    // Increase the common denominator and adapt all but the delta_minus.
    denominator->ShiftLeft(1);  // *2
    numerator->ShiftLeft(1);    // *2
    delta_plus->ShiftLeft(1);   // *2
  }
}


// This routine multiplies numerator/denominator so that its values lies in the
// range 1-10. That is after a call to this function we have:
//    1 <= (numerator + delta_plus) /denominator < 10.
// Let numerator the input before modification and numerator' the argument
// after modification, then the output-parameter decimal_point is such that
//  numerator / denominator * 10^estimated_power ==
//    numerator' / denominator' * 10^(decimal_point - 1)
// In some cases estimated_power was too low, and this is already the case. We
// then simply adjust the power so that 10^(k-1) <= v < 10^k (with k ==
// estimated_power) but do not touch the numerator or denominator.
// Otherwise the routine multiplies the numerator and the deltas by 10.
static void FixupMultiply10(int estimated_power, bool is_even,
                            int* decimal_point,
                            Bignum* numerator, Bignum* denominator,
                            Bignum* delta_minus, Bignum* delta_plus) {
  bool in_range;
  if (is_even) {
    // For IEEE doubles half-way cases (in decimal system numbers ending with 5)
    // are rounded to the closest floating-point number with even significand.
    in_range = Bignum::PlusCompare(*numerator, *delta_plus, *denominator) >= 0;
  } else {
    in_range = Bignum::PlusCompare(*numerator, *delta_plus, *denominator) > 0;
  }
  if (in_range) {
    // Since numerator + delta_plus >= denominator we already have
    // 1 <= numerator/denominator < 10. Simply update the estimated_power.
    *decimal_point = estimated_power + 1;
  } else {
    *decimal_point = estimated_power;
    numerator->Times10();
    if (Bignum::Equal(*delta_minus, *delta_plus)) {
      delta_minus->Times10();
      delta_plus->AssignBignum(*delta_minus);
    } else {
      delta_minus->Times10();
      delta_plus->Times10();
    }
  }
}

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END
#endif // ICU PATCH: close #if !UCONFIG_NO_FORMATTING
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING

#include <algorithm>
#include <cstring>

// ICU PATCH: Customize header file paths for ICU.




// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

Bignum::Chunk& Bignum::RawBigit(const int index) {
  DOUBLE_CONVERSION_ASSERT(static_cast<unsigned>(index) < kBigitCapacity);
  return bigits_buffer_[index];
}


const Bignum::Chunk& Bignum::RawBigit(const int index) const {
  DOUBLE_CONVERSION_ASSERT(static_cast<unsigned>(index) < kBigitCapacity);
  return bigits_buffer_[index];
}


template<typename S>
static int BitSize(const S value) {
  (void) value;  // Mark variable as used.
  return 8 * sizeof(value);
}

// Guaranteed to lie in one Bigit.
void Bignum::AssignUInt16(const uint16_t value) {
  DOUBLE_CONVERSION_ASSERT(kBigitSize >= BitSize(value));
  Zero();
  if (value > 0) {
    RawBigit(0) = value;
    used_bigits_ = 1;
  }
}


void Bignum::AssignUInt64(uint64_t value) {
  Zero();
  for(int i = 0; value > 0; ++i) {
    RawBigit(i) = value & kBigitMask;
    value >>= kBigitSize;
    ++used_bigits_;
  }
}


void Bignum::AssignBignum(const Bignum& other) {
  exponent_ = other.exponent_;
  for (int i = 0; i < other.used_bigits_; ++i) {
    RawBigit(i) = other.RawBigit(i);
  }
  used_bigits_ = other.used_bigits_;
}


static uint64_t ReadUInt64(const Vector<const char> buffer,
                           const int from,
                           const int digits_to_read) {
  uint64_t result = 0;
  for (int i = from; i < from + digits_to_read; ++i) {
    const int digit = buffer[i] - '0';
    DOUBLE_CONVERSION_ASSERT(0 <= digit && digit <= 9);
    result = result * 10 + digit;
  }
  return result;
}


void Bignum::AssignDecimalString(const Vector<const char> value) {
  // 2^64 = 18446744073709551616 > 10^19
  static const int kMaxUint64DecimalDigits = 19;
  Zero();
  int length = value.length();
  unsigned pos = 0;
  // Let's just say that each digit needs 4 bits.
  while (length >= kMaxUint64DecimalDigits) {
    const uint64_t digits = ReadUInt64(value, pos, kMaxUint64DecimalDigits);
    pos += kMaxUint64DecimalDigits;
    length -= kMaxUint64DecimalDigits;
    MultiplyByPowerOfTen(kMaxUint64DecimalDigits);
    AddUInt64(digits);
  }
  const uint64_t digits = ReadUInt64(value, pos, length);
  MultiplyByPowerOfTen(length);
  AddUInt64(digits);
  Clamp();
}


static uint64_t HexCharValue(const int c) {
  if ('0' <= c && c <= '9') {
    return c - '0';
  }
  if ('a' <= c && c <= 'f') {
    return 10 + c - 'a';
  }
  DOUBLE_CONVERSION_ASSERT('A' <= c && c <= 'F');
  return 10 + c - 'A';
}


// Unlike AssignDecimalString(), this function is "only" used
// for unit-tests and therefore not performance critical.
void Bignum::AssignHexString(Vector<const char> value) {
  Zero();
  // Required capacity could be reduced by ignoring leading zeros.
  EnsureCapacity(((value.length() * 4) + kBigitSize - 1) / kBigitSize);
  DOUBLE_CONVERSION_ASSERT(sizeof(uint64_t) * 8 >= kBigitSize + 4);  // TODO: static_assert
  // Accumulates converted hex digits until at least kBigitSize bits.
  // Works with non-factor-of-four kBigitSizes.
  uint64_t tmp = 0;  // Accumulates converted hex digits until at least
  for (int cnt = 0; !value.is_empty(); value.pop_back()) {
    tmp |= (HexCharValue(value.last()) << cnt);
    if ((cnt += 4) >= kBigitSize) {
      RawBigit(used_bigits_++) = (tmp & kBigitMask);
      cnt -= kBigitSize;
      tmp >>= kBigitSize;
    }
  }
  if (tmp > 0) {
    RawBigit(used_bigits_++) = tmp;
  }
  Clamp();
}


void Bignum::AddUInt64(const uint64_t operand) {
  if (operand == 0) {
    return;
  }
  Bignum other;
  other.AssignUInt64(operand);
  AddBignum(other);
}


void Bignum::AddBignum(const Bignum& other) {
  DOUBLE_CONVERSION_ASSERT(IsClamped());
  DOUBLE_CONVERSION_ASSERT(other.IsClamped());

  // If this has a greater exponent than other append zero-bigits to this.
  // After this call exponent_ <= other.exponent_.
  Align(other);

  // There are two possibilities:
  //   aaaaaaaaaaa 0000  (where the 0s represent a's exponent)
  //     bbbbb 00000000
  //   ----------------
  //   ccccccccccc 0000
  // or
  //    aaaaaaaaaa 0000
  //  bbbbbbbbb 0000000
  //  -----------------
  //  cccccccccccc 0000
  // In both cases we might need a carry bigit.

  EnsureCapacity(1 + (std::max)(BigitLength(), other.BigitLength()) - exponent_);
  Chunk carry = 0;
  int bigit_pos = other.exponent_ - exponent_;
  DOUBLE_CONVERSION_ASSERT(bigit_pos >= 0);
  for (int i = used_bigits_; i < bigit_pos; ++i) {
    RawBigit(i) = 0;
  }
  for (int i = 0; i < other.used_bigits_; ++i) {
    const Chunk my = (bigit_pos < used_bigits_) ? RawBigit(bigit_pos) : 0;
    const Chunk sum = my + other.RawBigit(i) + carry;
    RawBigit(bigit_pos) = sum & kBigitMask;
    carry = sum >> kBigitSize;
    ++bigit_pos;
  }
  while (carry != 0) {
    const Chunk my = (bigit_pos < used_bigits_) ? RawBigit(bigit_pos) : 0;
    const Chunk sum = my + carry;
    RawBigit(bigit_pos) = sum & kBigitMask;
    carry = sum >> kBigitSize;
    ++bigit_pos;
  }
  used_bigits_ = (std::max)(bigit_pos, static_cast<int>(used_bigits_));
  DOUBLE_CONVERSION_ASSERT(IsClamped());
}


void Bignum::SubtractBignum(const Bignum& other) {
  DOUBLE_CONVERSION_ASSERT(IsClamped());
  DOUBLE_CONVERSION_ASSERT(other.IsClamped());
  // We require this to be bigger than other.
  DOUBLE_CONVERSION_ASSERT(LessEqual(other, *this));

  Align(other);

  const int offset = other.exponent_ - exponent_;
  Chunk borrow = 0;
  int i;
  for (i = 0; i < other.used_bigits_; ++i) {
    DOUBLE_CONVERSION_ASSERT((borrow == 0) || (borrow == 1));
    const Chunk difference = RawBigit(i + offset) - other.RawBigit(i) - borrow;
    RawBigit(i + offset) = difference & kBigitMask;
    borrow = difference >> (kChunkSize - 1);
  }
  while (borrow != 0) {
    const Chunk difference = RawBigit(i + offset) - borrow;
    RawBigit(i + offset) = difference & kBigitMask;
    borrow = difference >> (kChunkSize - 1);
    ++i;
  }
  Clamp();
}


void Bignum::ShiftLeft(const int shift_amount) {
  if (used_bigits_ == 0) {
    return;
  }
  exponent_ += (shift_amount / kBigitSize);
  const int local_shift = shift_amount % kBigitSize;
  EnsureCapacity(used_bigits_ + 1);
  BigitsShiftLeft(local_shift);
}


void Bignum::MultiplyByUInt32(const uint32_t factor) {
  if (factor == 1) {
    return;
  }
  if (factor == 0) {
    Zero();
    return;
  }
  if (used_bigits_ == 0) {
    return;
  }
  // The product of a bigit with the factor is of size kBigitSize + 32.
  // Assert that this number + 1 (for the carry) fits into double chunk.
  DOUBLE_CONVERSION_ASSERT(kDoubleChunkSize >= kBigitSize + 32 + 1);
  DoubleChunk carry = 0;
  for (int i = 0; i < used_bigits_; ++i) {
    const DoubleChunk product = static_cast<DoubleChunk>(factor) * RawBigit(i) + carry;
    RawBigit(i) = static_cast<Chunk>(product & kBigitMask);
    carry = (product >> kBigitSize);
  }
  while (carry != 0) {
    EnsureCapacity(used_bigits_ + 1);
    RawBigit(used_bigits_) = carry & kBigitMask;
    used_bigits_++;
    carry >>= kBigitSize;
  }
}


void Bignum::MultiplyByUInt64(const uint64_t factor) {
  if (factor == 1) {
    return;
  }
  if (factor == 0) {
    Zero();
    return;
  }
  if (used_bigits_ == 0) {
    return;
  }
  DOUBLE_CONVERSION_ASSERT(kBigitSize < 32);
  uint64_t carry = 0;
  const uint64_t low = factor & 0xFFFFFFFF;
  const uint64_t high = factor >> 32;
  for (int i = 0; i < used_bigits_; ++i) {
    const uint64_t product_low = low * RawBigit(i);
    const uint64_t product_high = high * RawBigit(i);
    const uint64_t tmp = (carry & kBigitMask) + product_low;
    RawBigit(i) = tmp & kBigitMask;
    carry = (carry >> kBigitSize) + (tmp >> kBigitSize) +
        (product_high << (32 - kBigitSize));
  }
  while (carry != 0) {
    EnsureCapacity(used_bigits_ + 1);
    RawBigit(used_bigits_) = carry & kBigitMask;
    used_bigits_++;
    carry >>= kBigitSize;
  }
}


void Bignum::MultiplyByPowerOfTen(const int exponent) {
  static const uint64_t kFive27 = DOUBLE_CONVERSION_UINT64_2PART_C(0x6765c793, fa10079d);
  static const uint16_t kFive1 = 5;
  static const uint16_t kFive2 = kFive1 * 5;
  static const uint16_t kFive3 = kFive2 * 5;
  static const uint16_t kFive4 = kFive3 * 5;
  static const uint16_t kFive5 = kFive4 * 5;
  static const uint16_t kFive6 = kFive5 * 5;
  static const uint32_t kFive7 = kFive6 * 5;
  static const uint32_t kFive8 = kFive7 * 5;
  static const uint32_t kFive9 = kFive8 * 5;
  static const uint32_t kFive10 = kFive9 * 5;
  static const uint32_t kFive11 = kFive10 * 5;
  static const uint32_t kFive12 = kFive11 * 5;
  static const uint32_t kFive13 = kFive12 * 5;
  static const uint32_t kFive1_to_12[] =
      { kFive1, kFive2, kFive3, kFive4, kFive5, kFive6,
        kFive7, kFive8, kFive9, kFive10, kFive11, kFive12 };

  DOUBLE_CONVERSION_ASSERT(exponent >= 0);

  if (exponent == 0) {
    return;
  }
  if (used_bigits_ == 0) {
    return;
  }
  // We shift by exponent at the end just before returning.
  int remaining_exponent = exponent;
  while (remaining_exponent >= 27) {
    MultiplyByUInt64(kFive27);
    remaining_exponent -= 27;
  }
  while (remaining_exponent >= 13) {
    MultiplyByUInt32(kFive13);
    remaining_exponent -= 13;
  }
  if (remaining_exponent > 0) {
    MultiplyByUInt32(kFive1_to_12[remaining_exponent - 1]);
  }
  ShiftLeft(exponent);
}


void Bignum::Square() {
  DOUBLE_CONVERSION_ASSERT(IsClamped());
  const int product_length = 2 * used_bigits_;
  EnsureCapacity(product_length);

  // Comba multiplication: compute each column separately.
  // Example: r = a2a1a0 * b2b1b0.
  //    r =  1    * a0b0 +
  //        10    * (a1b0 + a0b1) +
  //        100   * (a2b0 + a1b1 + a0b2) +
  //        1000  * (a2b1 + a1b2) +
  //        10000 * a2b2
  //
  // In the worst case we have to accumulate nb-digits products of digit*digit.
  //
  // Assert that the additional number of bits in a DoubleChunk are enough to
  // sum up used_digits of Bigit*Bigit.
  if ((1 << (2 * (kChunkSize - kBigitSize))) <= used_bigits_) {
    DOUBLE_CONVERSION_UNIMPLEMENTED();
  }
  DoubleChunk accumulator = 0;
  // First shift the digits so we don't overwrite them.
  const int copy_offset = used_bigits_;
  for (int i = 0; i < used_bigits_; ++i) {
    RawBigit(copy_offset + i) = RawBigit(i);
  }
  // We have two loops to avoid some 'if's in the loop.
  for (int i = 0; i < used_bigits_; ++i) {
    // Process temporary digit i with power i.
    // The sum of the two indices must be equal to i.
    int bigit_index1 = i;
    int bigit_index2 = 0;
    // Sum all of the sub-products.
    while (bigit_index1 >= 0) {
      const Chunk chunk1 = RawBigit(copy_offset + bigit_index1);
      const Chunk chunk2 = RawBigit(copy_offset + bigit_index2);
      accumulator += static_cast<DoubleChunk>(chunk1) * chunk2;
      bigit_index1--;
      bigit_index2++;
    }
    RawBigit(i) = static_cast<Chunk>(accumulator) & kBigitMask;
    accumulator >>= kBigitSize;
  }
  for (int i = used_bigits_; i < product_length; ++i) {
    int bigit_index1 = used_bigits_ - 1;
    int bigit_index2 = i - bigit_index1;
    // Invariant: sum of both indices is again equal to i.
    // Inner loop runs 0 times on last iteration, emptying accumulator.
    while (bigit_index2 < used_bigits_) {
      const Chunk chunk1 = RawBigit(copy_offset + bigit_index1);
      const Chunk chunk2 = RawBigit(copy_offset + bigit_index2);
      accumulator += static_cast<DoubleChunk>(chunk1) * chunk2;
      bigit_index1--;
      bigit_index2++;
    }
    // The overwritten RawBigit(i) will never be read in further loop iterations,
    // because bigit_index1 and bigit_index2 are always greater
    // than i - used_bigits_.
    RawBigit(i) = static_cast<Chunk>(accumulator) & kBigitMask;
    accumulator >>= kBigitSize;
  }
  // Since the result was guaranteed to lie inside the number the
  // accumulator must be 0 now.
  DOUBLE_CONVERSION_ASSERT(accumulator == 0);

  // Don't forget to update the used_digits and the exponent.
  used_bigits_ = product_length;
  exponent_ *= 2;
  Clamp();
}


void Bignum::AssignPowerUInt16(uint16_t base, const int power_exponent) {
  DOUBLE_CONVERSION_ASSERT(base != 0);
  DOUBLE_CONVERSION_ASSERT(power_exponent >= 0);
  if (power_exponent == 0) {
    AssignUInt16(1);
    return;
  }
  Zero();
  int shifts = 0;
  // We expect base to be in range 2-32, and most often to be 10.
  // It does not make much sense to implement different algorithms for counting
  // the bits.
  while ((base & 1) == 0) {
    base >>= 1;
    shifts++;
  }
  int bit_size = 0;
  int tmp_base = base;
  while (tmp_base != 0) {
    tmp_base >>= 1;
    bit_size++;
  }
  const int final_size = bit_size * power_exponent;
  // 1 extra bigit for the shifting, and one for rounded final_size.
  EnsureCapacity(final_size / kBigitSize + 2);

  // Left to Right exponentiation.
  int mask = 1;
  while (power_exponent >= mask) mask <<= 1;

  // The mask is now pointing to the bit above the most significant 1-bit of
  // power_exponent.
  // Get rid of first 1-bit;
  mask >>= 2;
  uint64_t this_value = base;

  bool delayed_multiplication = false;
  const uint64_t max_32bits = 0xFFFFFFFF;
  while (mask != 0 && this_value <= max_32bits) {
    this_value = this_value * this_value;
    // Verify that there is enough space in this_value to perform the
    // multiplication.  The first bit_size bits must be 0.
    if ((power_exponent & mask) != 0) {
      DOUBLE_CONVERSION_ASSERT(bit_size > 0);
      const uint64_t base_bits_mask =
        ~((static_cast<uint64_t>(1) << (64 - bit_size)) - 1);
      const bool high_bits_zero = (this_value & base_bits_mask) == 0;
      if (high_bits_zero) {
        this_value *= base;
      } else {
        delayed_multiplication = true;
      }
    }
    mask >>= 1;
  }
  AssignUInt64(this_value);
  if (delayed_multiplication) {
    MultiplyByUInt32(base);
  }

  // Now do the same thing as a bignum.
  while (mask != 0) {
    Square();
    if ((power_exponent & mask) != 0) {
      MultiplyByUInt32(base);
    }
    mask >>= 1;
  }

  // And finally add the saved shifts.
  ShiftLeft(shifts * power_exponent);
}


// Precondition: this/other < 16bit.
uint16_t Bignum::DivideModuloIntBignum(const Bignum& other) {
  DOUBLE_CONVERSION_ASSERT(IsClamped());
  DOUBLE_CONVERSION_ASSERT(other.IsClamped());
  DOUBLE_CONVERSION_ASSERT(other.used_bigits_ > 0);

  // Easy case: if we have less digits than the divisor than the result is 0.
  // Note: this handles the case where this == 0, too.
  if (BigitLength() < other.BigitLength()) {
    return 0;
  }

  Align(other);

  uint16_t result = 0;

  // Start by removing multiples of 'other' until both numbers have the same
  // number of digits.
  while (BigitLength() > other.BigitLength()) {
    // This naive approach is extremely inefficient if `this` divided by other
    // is big. This function is implemented for doubleToString where
    // the result should be small (less than 10).
    DOUBLE_CONVERSION_ASSERT(other.RawBigit(other.used_bigits_ - 1) >= ((1 << kBigitSize) / 16));
    DOUBLE_CONVERSION_ASSERT(RawBigit(used_bigits_ - 1) < 0x10000);
    // Remove the multiples of the first digit.
    // Example this = 23 and other equals 9. -> Remove 2 multiples.
    result += static_cast<uint16_t>(RawBigit(used_bigits_ - 1));
    SubtractTimes(other, RawBigit(used_bigits_ - 1));
  }

  DOUBLE_CONVERSION_ASSERT(BigitLength() == other.BigitLength());

  // Both bignums are at the same length now.
  // Since other has more than 0 digits we know that the access to
  // RawBigit(used_bigits_ - 1) is safe.
  const Chunk this_bigit = RawBigit(used_bigits_ - 1);
  const Chunk other_bigit = other.RawBigit(other.used_bigits_ - 1);

  if (other.used_bigits_ == 1) {
    // Shortcut for easy (and common) case.
    int quotient = this_bigit / other_bigit;
    RawBigit(used_bigits_ - 1) = this_bigit - other_bigit * quotient;
    DOUBLE_CONVERSION_ASSERT(quotient < 0x10000);
    result += static_cast<uint16_t>(quotient);
    Clamp();
    return result;
  }

  const int division_estimate = this_bigit / (other_bigit + 1);
  DOUBLE_CONVERSION_ASSERT(division_estimate < 0x10000);
  result += static_cast<uint16_t>(division_estimate);
  SubtractTimes(other, division_estimate);

  if (other_bigit * (division_estimate + 1) > this_bigit) {
    // No need to even try to subtract. Even if other's remaining digits were 0
    // another subtraction would be too much.
    return result;
  }

  while (LessEqual(other, *this)) {
    SubtractBignum(other);
    result++;
  }
  return result;
}


template<typename S>
static int SizeInHexChars(S number) {
  DOUBLE_CONVERSION_ASSERT(number > 0);
  int result = 0;
  while (number != 0) {
    number >>= 4;
    result++;
  }
  return result;
}


static char HexCharOfValue(const int value) {
  DOUBLE_CONVERSION_ASSERT(0 <= value && value <= 16);
  if (value < 10) {
    return static_cast<char>(value + '0');
  }
  return static_cast<char>(value - 10 + 'A');
}


bool Bignum::ToHexString(char* buffer, const int buffer_size) const {
  DOUBLE_CONVERSION_ASSERT(IsClamped());
  // Each bigit must be printable as separate hex-character.
  DOUBLE_CONVERSION_ASSERT(kBigitSize % 4 == 0);
  static const int kHexCharsPerBigit = kBigitSize / 4;

  if (used_bigits_ == 0) {
    if (buffer_size < 2) {
      return false;
    }
    buffer[0] = '0';
    buffer[1] = '\0';
    return true;
  }
  // We add 1 for the terminating '\0' character.
  const int needed_chars = (BigitLength() - 1) * kHexCharsPerBigit +
    SizeInHexChars(RawBigit(used_bigits_ - 1)) + 1;
  if (needed_chars > buffer_size) {
    return false;
  }
  int string_index = needed_chars - 1;
  buffer[string_index--] = '\0';
  for (int i = 0; i < exponent_; ++i) {
    for (int j = 0; j < kHexCharsPerBigit; ++j) {
      buffer[string_index--] = '0';
    }
  }
  for (int i = 0; i < used_bigits_ - 1; ++i) {
    Chunk current_bigit = RawBigit(i);
    for (int j = 0; j < kHexCharsPerBigit; ++j) {
      buffer[string_index--] = HexCharOfValue(current_bigit & 0xF);
      current_bigit >>= 4;
    }
  }
  // And finally the last bigit.
  Chunk most_significant_bigit = RawBigit(used_bigits_ - 1);
  while (most_significant_bigit != 0) {
    buffer[string_index--] = HexCharOfValue(most_significant_bigit & 0xF);
    most_significant_bigit >>= 4;
  }
  return true;
}


Bignum::Chunk Bignum::BigitOrZero(const int index) const {
  if (index >= BigitLength()) {
    return 0;
  }
  if (index < exponent_) {
    return 0;
  }
  return RawBigit(index - exponent_);
}


int Bignum::Compare(const Bignum& a, const Bignum& b) {
  DOUBLE_CONVERSION_ASSERT(a.IsClamped());
  DOUBLE_CONVERSION_ASSERT(b.IsClamped());
  const int bigit_length_a = a.BigitLength();
  const int bigit_length_b = b.BigitLength();
  if (bigit_length_a < bigit_length_b) {
    return -1;
  }
  if (bigit_length_a > bigit_length_b) {
    return +1;
  }
  for (int i = bigit_length_a - 1; i >= (std::min)(a.exponent_, b.exponent_); --i) {
    const Chunk bigit_a = a.BigitOrZero(i);
    const Chunk bigit_b = b.BigitOrZero(i);
    if (bigit_a < bigit_b) {
      return -1;
    }
    if (bigit_a > bigit_b) {
      return +1;
    }
    // Otherwise they are equal up to this digit. Try the next digit.
  }
  return 0;
}


int Bignum::PlusCompare(const Bignum& a, const Bignum& b, const Bignum& c) {
  DOUBLE_CONVERSION_ASSERT(a.IsClamped());
  DOUBLE_CONVERSION_ASSERT(b.IsClamped());
  DOUBLE_CONVERSION_ASSERT(c.IsClamped());
  if (a.BigitLength() < b.BigitLength()) {
    return PlusCompare(b, a, c);
  }
  if (a.BigitLength() + 1 < c.BigitLength()) {
    return -1;
  }
  if (a.BigitLength() > c.BigitLength()) {
    return +1;
  }
  // The exponent encodes 0-bigits. So if there are more 0-digits in 'a' than
  // 'b' has digits, then the bigit-length of 'a'+'b' must be equal to the one
  // of 'a'.
  if (a.exponent_ >= b.BigitLength() && a.BigitLength() < c.BigitLength()) {
    return -1;
  }

  Chunk borrow = 0;
  // Starting at min_exponent all digits are == 0. So no need to compare them.
  const int min_exponent = (std::min)((std::min)(a.exponent_, b.exponent_), c.exponent_);
  for (int i = c.BigitLength() - 1; i >= min_exponent; --i) {
    const Chunk chunk_a = a.BigitOrZero(i);
    const Chunk chunk_b = b.BigitOrZero(i);
    const Chunk chunk_c = c.BigitOrZero(i);
    const Chunk sum = chunk_a + chunk_b;
    if (sum > chunk_c + borrow) {
      return +1;
    } else {
      borrow = chunk_c + borrow - sum;
      if (borrow > 1) {
        return -1;
      }
      borrow <<= kBigitSize;
    }
  }
  if (borrow == 0) {
    return 0;
  }
  return -1;
}


void Bignum::Clamp() {
  while (used_bigits_ > 0 && RawBigit(used_bigits_ - 1) == 0) {
    used_bigits_--;
  }
  if (used_bigits_ == 0) {
    // Zero.
    exponent_ = 0;
  }
}


void Bignum::Align(const Bignum& other) {
  if (exponent_ > other.exponent_) {
    // If "X" represents a "hidden" bigit (by the exponent) then we are in the
    // following case (a == this, b == other):
    // a:  aaaaaaXXXX   or a:   aaaaaXXX
    // b:     bbbbbbX      b: bbbbbbbbXX
    // We replace some of the hidden digits (X) of a with 0 digits.
    // a:  aaaaaa000X   or a:   aaaaa0XX
    const int zero_bigits = exponent_ - other.exponent_;
    EnsureCapacity(used_bigits_ + zero_bigits);
    for (int i = used_bigits_ - 1; i >= 0; --i) {
      RawBigit(i + zero_bigits) = RawBigit(i);
    }
    for (int i = 0; i < zero_bigits; ++i) {
      RawBigit(i) = 0;
    }
    used_bigits_ += zero_bigits;
    exponent_ -= zero_bigits;

    DOUBLE_CONVERSION_ASSERT(used_bigits_ >= 0);
    DOUBLE_CONVERSION_ASSERT(exponent_ >= 0);
  }
}


void Bignum::BigitsShiftLeft(const int shift_amount) {
  DOUBLE_CONVERSION_ASSERT(shift_amount < kBigitSize);
  DOUBLE_CONVERSION_ASSERT(shift_amount >= 0);
  Chunk carry = 0;
  for (int i = 0; i < used_bigits_; ++i) {
    const Chunk new_carry = RawBigit(i) >> (kBigitSize - shift_amount);
    RawBigit(i) = ((RawBigit(i) << shift_amount) + carry) & kBigitMask;
    carry = new_carry;
  }
  if (carry != 0) {
    RawBigit(used_bigits_) = carry;
    used_bigits_++;
  }
}


void Bignum::SubtractTimes(const Bignum& other, const int factor) {
  DOUBLE_CONVERSION_ASSERT(exponent_ <= other.exponent_);
  if (factor < 3) {
    for (int i = 0; i < factor; ++i) {
      SubtractBignum(other);
    }
    return;
  }
  Chunk borrow = 0;
  const int exponent_diff = other.exponent_ - exponent_;
  for (int i = 0; i < other.used_bigits_; ++i) {
    const DoubleChunk product = static_cast<DoubleChunk>(factor) * other.RawBigit(i);
    const DoubleChunk remove = borrow + product;
    const Chunk difference = RawBigit(i + exponent_diff) - (remove & kBigitMask);
    RawBigit(i + exponent_diff) = difference & kBigitMask;
    borrow = static_cast<Chunk>((difference >> (kChunkSize - 1)) +
                                (remove >> kBigitSize));
  }
  for (int i = other.used_bigits_ + exponent_diff; i < used_bigits_; ++i) {
    if (borrow == 0) {
      return;
    }
    const Chunk difference = RawBigit(i) - borrow;
    RawBigit(i) = difference & kBigitMask;
    borrow = difference >> (kChunkSize - 1);
  }
  Clamp();
}


}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END
#endif // ICU PATCH: close #if !UCONFIG_NO_FORMATTING
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2006-2008 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING

#include <climits>
#include <cmath>
#include <cstdarg>

// ICU PATCH: Customize header file paths for ICU.



// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


// ICU PATCH: Customize header file paths for ICU.



// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

namespace PowersOfTenCache {

  // Not all powers of ten are cached. The decimal exponent of two neighboring
  // cached numbers will differ by kDecimalExponentDistance.
  static const int kDecimalExponentDistance = 8;

  static const int kMinDecimalExponent = -348;
#if U_DEBUG
  static const int kMaxDecimalExponent = 340;
#endif

  // Returns a cached power-of-ten with a binary exponent in the range
  // [min_exponent; max_exponent] (boundaries included).
  void GetCachedPowerForBinaryExponentRange(int min_exponent,
                                            int max_exponent,
                                            DiyFp* power,
                                            int* decimal_exponent);

  // Returns a cached power of ten x ~= 10^k such that
  //   k <= decimal_exponent < k + kCachedPowersDecimalDistance.
  // The given decimal_exponent must satisfy
  //   kMinDecimalExponent <= requested_exponent, and
  //   requested_exponent < kMaxDecimalExponent + kDecimalExponentDistance.
  void GetCachedPowerForDecimalExponent(int requested_exponent,
                                        DiyFp* power,
                                        int* found_exponent);

}  // namespace PowersOfTenCache

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END

#endif  // DOUBLE_CONVERSION_CACHED_POWERS_H_


// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

namespace PowersOfTenCache {

struct CachedPower {
  uint64_t significand;
  int16_t binary_exponent;
  int16_t decimal_exponent;
};

static const CachedPower kCachedPowers[] = {
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xfa8fd5a0, 081c0288), -1220, -348},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xbaaee17f, a23ebf76), -1193, -340},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x8b16fb20, 3055ac76), -1166, -332},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xcf42894a, 5dce35ea), -1140, -324},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x9a6bb0aa, 55653b2d), -1113, -316},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xe61acf03, 3d1a45df), -1087, -308},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xab70fe17, c79ac6ca), -1060, -300},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xff77b1fc, bebcdc4f), -1034, -292},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xbe5691ef, 416bd60c), -1007, -284},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x8dd01fad, 907ffc3c), -980, -276},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xd3515c28, 31559a83), -954, -268},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x9d71ac8f, ada6c9b5), -927, -260},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xea9c2277, 23ee8bcb), -901, -252},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xaecc4991, 4078536d), -874, -244},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x823c1279, 5db6ce57), -847, -236},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xc2109436, 4dfb5637), -821, -228},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x9096ea6f, 3848984f), -794, -220},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xd77485cb, 25823ac7), -768, -212},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xa086cfcd, 97bf97f4), -741, -204},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xef340a98, 172aace5), -715, -196},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xb23867fb, 2a35b28e), -688, -188},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x84c8d4df, d2c63f3b), -661, -180},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xc5dd4427, 1ad3cdba), -635, -172},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x936b9fce, bb25c996), -608, -164},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xdbac6c24, 7d62a584), -582, -156},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xa3ab6658, 0d5fdaf6), -555, -148},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xf3e2f893, dec3f126), -529, -140},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xb5b5ada8, aaff80b8), -502, -132},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x87625f05, 6c7c4a8b), -475, -124},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xc9bcff60, 34c13053), -449, -116},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x964e858c, 91ba2655), -422, -108},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xdff97724, 70297ebd), -396, -100},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xa6dfbd9f, b8e5b88f), -369, -92},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xf8a95fcf, 88747d94), -343, -84},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xb9447093, 8fa89bcf), -316, -76},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x8a08f0f8, bf0f156b), -289, -68},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xcdb02555, 653131b6), -263, -60},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x993fe2c6, d07b7fac), -236, -52},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xe45c10c4, 2a2b3b06), -210, -44},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xaa242499, 697392d3), -183, -36},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xfd87b5f2, 8300ca0e), -157, -28},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xbce50864, 92111aeb), -130, -20},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x8cbccc09, 6f5088cc), -103, -12},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xd1b71758, e219652c), -77, -4},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x9c400000, 00000000), -50, 4},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xe8d4a510, 00000000), -24, 12},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xad78ebc5, ac620000), 3, 20},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x813f3978, f8940984), 30, 28},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xc097ce7b, c90715b3), 56, 36},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x8f7e32ce, 7bea5c70), 83, 44},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xd5d238a4, abe98068), 109, 52},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x9f4f2726, 179a2245), 136, 60},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xed63a231, d4c4fb27), 162, 68},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xb0de6538, 8cc8ada8), 189, 76},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x83c7088e, 1aab65db), 216, 84},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xc45d1df9, 42711d9a), 242, 92},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x924d692c, a61be758), 269, 100},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xda01ee64, 1a708dea), 295, 108},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xa26da399, 9aef774a), 322, 116},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xf209787b, b47d6b85), 348, 124},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xb454e4a1, 79dd1877), 375, 132},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x865b8692, 5b9bc5c2), 402, 140},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xc83553c5, c8965d3d), 428, 148},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x952ab45c, fa97a0b3), 455, 156},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xde469fbd, 99a05fe3), 481, 164},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xa59bc234, db398c25), 508, 172},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xf6c69a72, a3989f5c), 534, 180},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xb7dcbf53, 54e9bece), 561, 188},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x88fcf317, f22241e2), 588, 196},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xcc20ce9b, d35c78a5), 614, 204},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x98165af3, 7b2153df), 641, 212},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xe2a0b5dc, 971f303a), 667, 220},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xa8d9d153, 5ce3b396), 694, 228},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xfb9b7cd9, a4a7443c), 720, 236},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xbb764c4c, a7a44410), 747, 244},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x8bab8eef, b6409c1a), 774, 252},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xd01fef10, a657842c), 800, 260},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x9b10a4e5, e9913129), 827, 268},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xe7109bfb, a19c0c9d), 853, 276},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xac2820d9, 623bf429), 880, 284},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x80444b5e, 7aa7cf85), 907, 292},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xbf21e440, 03acdd2d), 933, 300},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x8e679c2f, 5e44ff8f), 960, 308},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xd433179d, 9c8cb841), 986, 316},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0x9e19db92, b4e31ba9), 1013, 324},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xeb96bf6e, badf77d9), 1039, 332},
  {DOUBLE_CONVERSION_UINT64_2PART_C(0xaf87023b, 9bf0ee6b), 1066, 340},
};

static const int kCachedPowersOffset = 348;  // -1 * the first decimal_exponent.
static const double kD_1_LOG2_10 = 0.30102999566398114;  //  1 / lg(10)

void GetCachedPowerForBinaryExponentRange(
    int min_exponent,
    int max_exponent,
    DiyFp* power,
    int* decimal_exponent) {
  int kQ = DiyFp::kSignificandSize;
  double k = ceil((min_exponent + kQ - 1) * kD_1_LOG2_10);
  int foo = kCachedPowersOffset;
  int index =
      (foo + static_cast<int>(k) - 1) / kDecimalExponentDistance + 1;
  DOUBLE_CONVERSION_ASSERT(0 <= index && index < static_cast<int>(DOUBLE_CONVERSION_ARRAY_SIZE(kCachedPowers)));
  CachedPower cached_power = kCachedPowers[index];
  DOUBLE_CONVERSION_ASSERT(min_exponent <= cached_power.binary_exponent);
  (void) max_exponent;  // Mark variable as used.
  DOUBLE_CONVERSION_ASSERT(cached_power.binary_exponent <= max_exponent);
  *decimal_exponent = cached_power.decimal_exponent;
  *power = DiyFp(cached_power.significand, cached_power.binary_exponent);
}


void GetCachedPowerForDecimalExponent(int requested_exponent,
                                      DiyFp* power,
                                      int* found_exponent) {
  DOUBLE_CONVERSION_ASSERT(kMinDecimalExponent <= requested_exponent);
  DOUBLE_CONVERSION_ASSERT(requested_exponent < kMaxDecimalExponent + kDecimalExponentDistance);
  int index =
      (requested_exponent + kCachedPowersOffset) / kDecimalExponentDistance;
  CachedPower cached_power = kCachedPowers[index];
  *power = DiyFp(cached_power.significand, cached_power.binary_exponent);
  *found_exponent = cached_power.decimal_exponent;
  DOUBLE_CONVERSION_ASSERT(*found_exponent <= requested_exponent);
  DOUBLE_CONVERSION_ASSERT(requested_exponent < *found_exponent + kDecimalExponentDistance);
}

}  // namespace PowersOfTenCache

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END
#endif // ICU PATCH: close #if !UCONFIG_NO_FORMATTING
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING

#include <algorithm>
#include <climits>
#include <cmath>

// ICU PATCH: Customize header file paths for ICU.
// The file fixed-dtoa.h is not needed.

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2012 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


// ICU PATCH: Customize header file paths for ICU.



// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

class DoubleToStringConverter {
 public:
#if 0 // not needed for ICU
  // When calling ToFixed with a double > 10^kMaxFixedDigitsBeforePoint
  // or a requested_digits parameter > kMaxFixedDigitsAfterPoint then the
  // function returns false.
  static const int kMaxFixedDigitsBeforePoint = 60;
  static const int kMaxFixedDigitsAfterPoint = 60;

  // When calling ToExponential with a requested_digits
  // parameter > kMaxExponentialDigits then the function returns false.
  static const int kMaxExponentialDigits = 120;

  // When calling ToPrecision with a requested_digits
  // parameter < kMinPrecisionDigits or requested_digits > kMaxPrecisionDigits
  // then the function returns false.
  static const int kMinPrecisionDigits = 1;
  static const int kMaxPrecisionDigits = 120;

  enum Flags {
    NO_FLAGS = 0,
    EMIT_POSITIVE_EXPONENT_SIGN = 1,
    EMIT_TRAILING_DECIMAL_POINT = 2,
    EMIT_TRAILING_ZERO_AFTER_POINT = 4,
    UNIQUE_ZERO = 8
  };

  // Flags should be a bit-or combination of the possible Flags-enum.
  //  - NO_FLAGS: no special flags.
  //  - EMIT_POSITIVE_EXPONENT_SIGN: when the number is converted into exponent
  //    form, emits a '+' for positive exponents. Example: 1.2e+2.
  //  - EMIT_TRAILING_DECIMAL_POINT: when the input number is an integer and is
  //    converted into decimal format then a trailing decimal point is appended.
  //    Example: 2345.0 is converted to "2345.".
  //  - EMIT_TRAILING_ZERO_AFTER_POINT: in addition to a trailing decimal point
  //    emits a trailing '0'-character. This flag requires the
  //    EXMIT_TRAILING_DECIMAL_POINT flag.
  //    Example: 2345.0 is converted to "2345.0".
  //  - UNIQUE_ZERO: "-0.0" is converted to "0.0".
  //
  // Infinity symbol and nan_symbol provide the string representation for these
  // special values. If the string is NULL and the special value is encountered
  // then the conversion functions return false.
  //
  // The exponent_character is used in exponential representations. It is
  // usually 'e' or 'E'.
  //
  // When converting to the shortest representation the converter will
  // represent input numbers in decimal format if they are in the interval
  // [10^decimal_in_shortest_low; 10^decimal_in_shortest_high[
  //    (lower boundary included, greater boundary excluded).
  // Example: with decimal_in_shortest_low = -6 and
  //               decimal_in_shortest_high = 21:
  //   ToShortest(0.000001)  -> "0.000001"
  //   ToShortest(0.0000001) -> "1e-7"
  //   ToShortest(111111111111111111111.0)  -> "111111111111111110000"
  //   ToShortest(100000000000000000000.0)  -> "100000000000000000000"
  //   ToShortest(1111111111111111111111.0) -> "1.1111111111111111e+21"
  //
  // When converting to precision mode the converter may add
  // max_leading_padding_zeroes before returning the number in exponential
  // format.
  // Example with max_leading_padding_zeroes_in_precision_mode = 6.
  //   ToPrecision(0.0000012345, 2) -> "0.0000012"
  //   ToPrecision(0.00000012345, 2) -> "1.2e-7"
  // Similarily the converter may add up to
  // max_trailing_padding_zeroes_in_precision_mode in precision mode to avoid
  // returning an exponential representation. A zero added by the
  // EMIT_TRAILING_ZERO_AFTER_POINT flag is counted for this limit.
  // Examples for max_trailing_padding_zeroes_in_precision_mode = 1:
  //   ToPrecision(230.0, 2) -> "230"
  //   ToPrecision(230.0, 2) -> "230."  with EMIT_TRAILING_DECIMAL_POINT.
  //   ToPrecision(230.0, 2) -> "2.3e2" with EMIT_TRAILING_ZERO_AFTER_POINT.
  //
  // The min_exponent_width is used for exponential representations.
  // The converter adds leading '0's to the exponent until the exponent
  // is at least min_exponent_width digits long.
  // The min_exponent_width is clamped to 5.
  // As such, the exponent may never have more than 5 digits in total.
  DoubleToStringConverter(int flags,
                          const char* infinity_symbol,
                          const char* nan_symbol,
                          char exponent_character,
                          int decimal_in_shortest_low,
                          int decimal_in_shortest_high,
                          int max_leading_padding_zeroes_in_precision_mode,
                          int max_trailing_padding_zeroes_in_precision_mode,
                          int min_exponent_width = 0)
      : flags_(flags),
        infinity_symbol_(infinity_symbol),
        nan_symbol_(nan_symbol),
        exponent_character_(exponent_character),
        decimal_in_shortest_low_(decimal_in_shortest_low),
        decimal_in_shortest_high_(decimal_in_shortest_high),
        max_leading_padding_zeroes_in_precision_mode_(
            max_leading_padding_zeroes_in_precision_mode),
        max_trailing_padding_zeroes_in_precision_mode_(
            max_trailing_padding_zeroes_in_precision_mode),
        min_exponent_width_(min_exponent_width) {
    // When 'trailing zero after the point' is set, then 'trailing point'
    // must be set too.
    DOUBLE_CONVERSION_ASSERT(((flags & EMIT_TRAILING_DECIMAL_POINT) != 0) ||
        !((flags & EMIT_TRAILING_ZERO_AFTER_POINT) != 0));
  }

  // Returns a converter following the EcmaScript specification.
  static const DoubleToStringConverter& EcmaScriptConverter();

  // Computes the shortest string of digits that correctly represent the input
  // number. Depending on decimal_in_shortest_low and decimal_in_shortest_high
  // (see constructor) it then either returns a decimal representation, or an
  // exponential representation.
  // Example with decimal_in_shortest_low = -6,
  //              decimal_in_shortest_high = 21,
  //              EMIT_POSITIVE_EXPONENT_SIGN activated, and
  //              EMIT_TRAILING_DECIMAL_POINT deactived:
  //   ToShortest(0.000001)  -> "0.000001"
  //   ToShortest(0.0000001) -> "1e-7"
  //   ToShortest(111111111111111111111.0)  -> "111111111111111110000"
  //   ToShortest(100000000000000000000.0)  -> "100000000000000000000"
  //   ToShortest(1111111111111111111111.0) -> "1.1111111111111111e+21"
  //
  // Note: the conversion may round the output if the returned string
  // is accurate enough to uniquely identify the input-number.
  // For example the most precise representation of the double 9e59 equals
  // "899999999999999918767229449717619953810131273674690656206848", but
  // the converter will return the shorter (but still correct) "9e59".
  //
  // Returns true if the conversion succeeds. The conversion always succeeds
  // except when the input value is special and no infinity_symbol or
  // nan_symbol has been given to the constructor.
  bool ToShortest(double value, StringBuilder* result_builder) const {
    return ToShortestIeeeNumber(value, result_builder, SHORTEST);
  }

  // Same as ToShortest, but for single-precision floats.
  bool ToShortestSingle(float value, StringBuilder* result_builder) const {
    return ToShortestIeeeNumber(value, result_builder, SHORTEST_SINGLE);
  }


  // Computes a decimal representation with a fixed number of digits after the
  // decimal point. The last emitted digit is rounded.
  //
  // Examples:
  //   ToFixed(3.12, 1) -> "3.1"
  //   ToFixed(3.1415, 3) -> "3.142"
  //   ToFixed(1234.56789, 4) -> "1234.5679"
  //   ToFixed(1.23, 5) -> "1.23000"
  //   ToFixed(0.1, 4) -> "0.1000"
  //   ToFixed(1e30, 2) -> "1000000000000000019884624838656.00"
  //   ToFixed(0.1, 30) -> "0.100000000000000005551115123126"
  //   ToFixed(0.1, 17) -> "0.10000000000000001"
  //
  // If requested_digits equals 0, then the tail of the result depends on
  // the EMIT_TRAILING_DECIMAL_POINT and EMIT_TRAILING_ZERO_AFTER_POINT.
  // Examples, for requested_digits == 0,
  //   let EMIT_TRAILING_DECIMAL_POINT and EMIT_TRAILING_ZERO_AFTER_POINT be
  //    - false and false: then 123.45 -> 123
  //                             0.678 -> 1
  //    - true and false: then 123.45 -> 123.
  //                            0.678 -> 1.
  //    - true and true: then 123.45 -> 123.0
  //                           0.678 -> 1.0
  //
  // Returns true if the conversion succeeds. The conversion always succeeds
  // except for the following cases:
  //   - the input value is special and no infinity_symbol or nan_symbol has
  //     been provided to the constructor,
  //   - 'value' > 10^kMaxFixedDigitsBeforePoint, or
  //   - 'requested_digits' > kMaxFixedDigitsAfterPoint.
  // The last two conditions imply that the result will never contain more than
  // 1 + kMaxFixedDigitsBeforePoint + 1 + kMaxFixedDigitsAfterPoint characters
  // (one additional character for the sign, and one for the decimal point).
  bool ToFixed(double value,
               int requested_digits,
               StringBuilder* result_builder) const;

  // Computes a representation in exponential format with requested_digits
  // after the decimal point. The last emitted digit is rounded.
  // If requested_digits equals -1, then the shortest exponential representation
  // is computed.
  //
  // Examples with EMIT_POSITIVE_EXPONENT_SIGN deactivated, and
  //               exponent_character set to 'e'.
  //   ToExponential(3.12, 1) -> "3.1e0"
  //   ToExponential(5.0, 3) -> "5.000e0"
  //   ToExponential(0.001, 2) -> "1.00e-3"
  //   ToExponential(3.1415, -1) -> "3.1415e0"
  //   ToExponential(3.1415, 4) -> "3.1415e0"
  //   ToExponential(3.1415, 3) -> "3.142e0"
  //   ToExponential(123456789000000, 3) -> "1.235e14"
  //   ToExponential(1000000000000000019884624838656.0, -1) -> "1e30"
  //   ToExponential(1000000000000000019884624838656.0, 32) ->
  //                     "1.00000000000000001988462483865600e30"
  //   ToExponential(1234, 0) -> "1e3"
  //
  // Returns true if the conversion succeeds. The conversion always succeeds
  // except for the following cases:
  //   - the input value is special and no infinity_symbol or nan_symbol has
  //     been provided to the constructor,
  //   - 'requested_digits' > kMaxExponentialDigits.
  // The last condition implies that the result will never contain more than
  // kMaxExponentialDigits + 8 characters (the sign, the digit before the
  // decimal point, the decimal point, the exponent character, the
  // exponent's sign, and at most 3 exponent digits).
  bool ToExponential(double value,
                     int requested_digits,
                     StringBuilder* result_builder) const;

  // Computes 'precision' leading digits of the given 'value' and returns them
  // either in exponential or decimal format, depending on
  // max_{leading|trailing}_padding_zeroes_in_precision_mode (given to the
  // constructor).
  // The last computed digit is rounded.
  //
  // Example with max_leading_padding_zeroes_in_precision_mode = 6.
  //   ToPrecision(0.0000012345, 2) -> "0.0000012"
  //   ToPrecision(0.00000012345, 2) -> "1.2e-7"
  // Similarily the converter may add up to
  // max_trailing_padding_zeroes_in_precision_mode in precision mode to avoid
  // returning an exponential representation. A zero added by the
  // EMIT_TRAILING_ZERO_AFTER_POINT flag is counted for this limit.
  // Examples for max_trailing_padding_zeroes_in_precision_mode = 1:
  //   ToPrecision(230.0, 2) -> "230"
  //   ToPrecision(230.0, 2) -> "230."  with EMIT_TRAILING_DECIMAL_POINT.
  //   ToPrecision(230.0, 2) -> "2.3e2" with EMIT_TRAILING_ZERO_AFTER_POINT.
  // Examples for max_trailing_padding_zeroes_in_precision_mode = 3, and no
  //    EMIT_TRAILING_ZERO_AFTER_POINT:
  //   ToPrecision(123450.0, 6) -> "123450"
  //   ToPrecision(123450.0, 5) -> "123450"
  //   ToPrecision(123450.0, 4) -> "123500"
  //   ToPrecision(123450.0, 3) -> "123000"
  //   ToPrecision(123450.0, 2) -> "1.2e5"
  //
  // Returns true if the conversion succeeds. The conversion always succeeds
  // except for the following cases:
  //   - the input value is special and no infinity_symbol or nan_symbol has
  //     been provided to the constructor,
  //   - precision < kMinPericisionDigits
  //   - precision > kMaxPrecisionDigits
  // The last condition implies that the result will never contain more than
  // kMaxPrecisionDigits + 7 characters (the sign, the decimal point, the
  // exponent character, the exponent's sign, and at most 3 exponent digits).
  bool ToPrecision(double value,
                   int precision,
                   StringBuilder* result_builder) const;
#endif // not needed for ICU

  enum DtoaMode {
    // Produce the shortest correct representation.
    // For example the output of 0.299999999999999988897 is (the less accurate
    // but correct) 0.3.
    SHORTEST,
    // Same as SHORTEST, but for single-precision floats.
    SHORTEST_SINGLE,
    // Produce a fixed number of digits after the decimal point.
    // For instance fixed(0.1, 4) becomes 0.1000
    // If the input number is big, the output will be big.
    FIXED,
    // Fixed number of digits (independent of the decimal point).
    PRECISION
  };

  // The maximal number of digits that are needed to emit a double in base 10.
  // A higher precision can be achieved by using more digits, but the shortest
  // accurate representation of any double will never use more digits than
  // kBase10MaximalLength.
  // Note that DoubleToAscii null-terminates its input. So the given buffer
  // should be at least kBase10MaximalLength + 1 characters long.
  static const int kBase10MaximalLength = 17;

  // Converts the given double 'v' to digit characters. 'v' must not be NaN,
  // +Infinity, or -Infinity. In SHORTEST_SINGLE-mode this restriction also
  // applies to 'v' after it has been casted to a single-precision float. That
  // is, in this mode static_cast<float>(v) must not be NaN, +Infinity or
  // -Infinity.
  //
  // The result should be interpreted as buffer * 10^(point-length).
  //
  // The digits are written to the buffer in the platform's charset, which is
  // often UTF-8 (with ASCII-range digits) but may be another charset, such
  // as EBCDIC.
  //
  // The output depends on the given mode:
  //  - SHORTEST: produce the least amount of digits for which the internal
  //   identity requirement is still satisfied. If the digits are printed
  //   (together with the correct exponent) then reading this number will give
  //   'v' again. The buffer will choose the representation that is closest to
  //   'v'. If there are two at the same distance, than the one farther away
  //   from 0 is chosen (halfway cases - ending with 5 - are rounded up).
  //   In this mode the 'requested_digits' parameter is ignored.
  //  - SHORTEST_SINGLE: same as SHORTEST but with single-precision.
  //  - FIXED: produces digits necessary to print a given number with
  //   'requested_digits' digits after the decimal point. The produced digits
  //   might be too short in which case the caller has to fill the remainder
  //   with '0's.
  //   Example: toFixed(0.001, 5) is allowed to return buffer="1", point=-2.
  //   Halfway cases are rounded towards +/-Infinity (away from 0). The call
  //   toFixed(0.15, 2) thus returns buffer="2", point=0.
  //   The returned buffer may contain digits that would be truncated from the
  //   shortest representation of the input.
  //  - PRECISION: produces 'requested_digits' where the first digit is not '0'.
  //   Even though the length of produced digits usually equals
  //   'requested_digits', the function is allowed to return fewer digits, in
  //   which case the caller has to fill the missing digits with '0's.
  //   Halfway cases are again rounded away from 0.
  // DoubleToAscii expects the given buffer to be big enough to hold all
  // digits and a terminating null-character. In SHORTEST-mode it expects a
  // buffer of at least kBase10MaximalLength + 1. In all other modes the
  // requested_digits parameter and the padding-zeroes limit the size of the
  // output. Don't forget the decimal point, the exponent character and the
  // terminating null-character when computing the maximal output size.
  // The given length is only used in debug mode to ensure the buffer is big
  // enough.
  // ICU PATCH: Export this as U_I18N_API for unit tests.
  static void U_I18N_API DoubleToAscii(double v,
                            DtoaMode mode,
                            int requested_digits,
                            char* buffer,
                            int buffer_length,
                            bool* sign,
                            int* length,
                            int* point);

#if 0 // not needed for ICU
 private:
  // Implementation for ToShortest and ToShortestSingle.
  bool ToShortestIeeeNumber(double value,
                            StringBuilder* result_builder,
                            DtoaMode mode) const;

  // If the value is a special value (NaN or Infinity) constructs the
  // corresponding string using the configured infinity/nan-symbol.
  // If either of them is NULL or the value is not special then the
  // function returns false.
  bool HandleSpecialValues(double value, StringBuilder* result_builder) const;
  // Constructs an exponential representation (i.e. 1.234e56).
  // The given exponent assumes a decimal point after the first decimal digit.
  void CreateExponentialRepresentation(const char* decimal_digits,
                                       int length,
                                       int exponent,
                                       StringBuilder* result_builder) const;
  // Creates a decimal representation (i.e 1234.5678).
  void CreateDecimalRepresentation(const char* decimal_digits,
                                   int length,
                                   int decimal_point,
                                   int digits_after_point,
                                   StringBuilder* result_builder) const;

  const int flags_;
  const char* const infinity_symbol_;
  const char* const nan_symbol_;
  const char exponent_character_;
  const int decimal_in_shortest_low_;
  const int decimal_in_shortest_high_;
  const int max_leading_padding_zeroes_in_precision_mode_;
  const int max_trailing_padding_zeroes_in_precision_mode_;
  const int min_exponent_width_;
#endif // not needed for ICU

  DOUBLE_CONVERSION_DISALLOW_IMPLICIT_CONSTRUCTORS(DoubleToStringConverter);
};

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END

#endif  // DOUBLE_CONVERSION_DOUBLE_TO_STRING_H_



// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


// ICU PATCH: Customize header file paths for ICU.



// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

enum FastDtoaMode {
  // Computes the shortest representation of the given input. The returned
  // result will be the most accurate number of this length. Longer
  // representations might be more accurate.
  FAST_DTOA_SHORTEST,
  // Same as FAST_DTOA_SHORTEST but for single-precision floats.
  FAST_DTOA_SHORTEST_SINGLE,
  // Computes a representation where the precision (number of digits) is
  // given as input. The precision is independent of the decimal point.
  FAST_DTOA_PRECISION
};

// FastDtoa will produce at most kFastDtoaMaximalLength digits. This does not
// include the terminating '\0' character.
// static const int kFastDtoaMaximalLength = 17;
// Same for single-precision numbers.
// static const int kFastDtoaMaximalSingleLength = 9;

// Provides a decimal representation of v.
// The result should be interpreted as buffer * 10^(point - length).
//
// Precondition:
//   * v must be a strictly positive finite double.
//
// Returns true if it succeeds, otherwise the result can not be trusted.
// There will be *length digits inside the buffer followed by a null terminator.
// If the function returns true and mode equals
//   - FAST_DTOA_SHORTEST, then
//     the parameter requested_digits is ignored.
//     The result satisfies
//         v == (double) (buffer * 10^(point - length)).
//     The digits in the buffer are the shortest representation possible. E.g.
//     if 0.099999999999 and 0.1 represent the same double then "1" is returned
//     with point = 0.
//     The last digit will be closest to the actual v. That is, even if several
//     digits might correctly yield 'v' when read again, the buffer will contain
//     the one closest to v.
//   - FAST_DTOA_PRECISION, then
//     the buffer contains requested_digits digits.
//     the difference v - (buffer * 10^(point-length)) is closest to zero for
//     all possible representations of requested_digits digits.
//     If there are two values that are equally close, then FastDtoa returns
//     false.
// For both modes the buffer must be large enough to hold the result.
bool FastDtoa(double d,
              FastDtoaMode mode,
              int requested_digits,
              Vector<char> buffer,
              int* length,
              int* decimal_point);

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END

#endif  // DOUBLE_CONVERSION_FAST_DTOA_H_




// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

#if 0  // not needed for ICU
const DoubleToStringConverter& DoubleToStringConverter::EcmaScriptConverter() {
  int flags = UNIQUE_ZERO | EMIT_POSITIVE_EXPONENT_SIGN;
  static DoubleToStringConverter converter(flags,
                                           "Infinity",
                                           "NaN",
                                           'e',
                                           -6, 21,
                                           6, 0);
  return converter;
}


bool DoubleToStringConverter::HandleSpecialValues(
    double value,
    StringBuilder* result_builder) const {
  Double double_inspect(value);
  if (double_inspect.IsInfinite()) {
    if (infinity_symbol_ == NULL) return false;
    if (value < 0) {
      result_builder->AddCharacter('-');
    }
    result_builder->AddString(infinity_symbol_);
    return true;
  }
  if (double_inspect.IsNan()) {
    if (nan_symbol_ == NULL) return false;
    result_builder->AddString(nan_symbol_);
    return true;
  }
  return false;
}


void DoubleToStringConverter::CreateExponentialRepresentation(
    const char* decimal_digits,
    int length,
    int exponent,
    StringBuilder* result_builder) const {
  DOUBLE_CONVERSION_ASSERT(length != 0);
  result_builder->AddCharacter(decimal_digits[0]);
  if (length != 1) {
    result_builder->AddCharacter('.');
    result_builder->AddSubstring(&decimal_digits[1], length-1);
  }
  result_builder->AddCharacter(exponent_character_);
  if (exponent < 0) {
    result_builder->AddCharacter('-');
    exponent = -exponent;
  } else {
    if ((flags_ & EMIT_POSITIVE_EXPONENT_SIGN) != 0) {
      result_builder->AddCharacter('+');
    }
  }
  if (exponent == 0) {
    result_builder->AddCharacter('0');
    return;
  }
  DOUBLE_CONVERSION_ASSERT(exponent < 1e4);
  // Changing this constant requires updating the comment of DoubleToStringConverter constructor
  const int kMaxExponentLength = 5;
  char buffer[kMaxExponentLength + 1];
  buffer[kMaxExponentLength] = '\0';
  int first_char_pos = kMaxExponentLength;
  while (exponent > 0) {
    buffer[--first_char_pos] = '0' + (exponent % 10);
    exponent /= 10;
  }
  // Add prefix '0' to make exponent width >= min(min_exponent_with_, kMaxExponentLength)
  // For example: convert 1e+9 -> 1e+09, if min_exponent_with_ is set to 2
  while(kMaxExponentLength - first_char_pos < std::min(min_exponent_width_, kMaxExponentLength)) {
    buffer[--first_char_pos] = '0';
  }
  result_builder->AddSubstring(&buffer[first_char_pos],
                               kMaxExponentLength - first_char_pos);
}


void DoubleToStringConverter::CreateDecimalRepresentation(
    const char* decimal_digits,
    int length,
    int decimal_point,
    int digits_after_point,
    StringBuilder* result_builder) const {
  // Create a representation that is padded with zeros if needed.
  if (decimal_point <= 0) {
      // "0.00000decimal_rep" or "0.000decimal_rep00".
    result_builder->AddCharacter('0');
    if (digits_after_point > 0) {
      result_builder->AddCharacter('.');
      result_builder->AddPadding('0', -decimal_point);
      DOUBLE_CONVERSION_ASSERT(length <= digits_after_point - (-decimal_point));
      result_builder->AddSubstring(decimal_digits, length);
      int remaining_digits = digits_after_point - (-decimal_point) - length;
      result_builder->AddPadding('0', remaining_digits);
    }
  } else if (decimal_point >= length) {
    // "decimal_rep0000.00000" or "decimal_rep.0000".
    result_builder->AddSubstring(decimal_digits, length);
    result_builder->AddPadding('0', decimal_point - length);
    if (digits_after_point > 0) {
      result_builder->AddCharacter('.');
      result_builder->AddPadding('0', digits_after_point);
    }
  } else {
    // "decima.l_rep000".
    DOUBLE_CONVERSION_ASSERT(digits_after_point > 0);
    result_builder->AddSubstring(decimal_digits, decimal_point);
    result_builder->AddCharacter('.');
    DOUBLE_CONVERSION_ASSERT(length - decimal_point <= digits_after_point);
    result_builder->AddSubstring(&decimal_digits[decimal_point],
                                 length - decimal_point);
    int remaining_digits = digits_after_point - (length - decimal_point);
    result_builder->AddPadding('0', remaining_digits);
  }
  if (digits_after_point == 0) {
    if ((flags_ & EMIT_TRAILING_DECIMAL_POINT) != 0) {
      result_builder->AddCharacter('.');
    }
    if ((flags_ & EMIT_TRAILING_ZERO_AFTER_POINT) != 0) {
      result_builder->AddCharacter('0');
    }
  }
}


bool DoubleToStringConverter::ToShortestIeeeNumber(
    double value,
    StringBuilder* result_builder,
    DoubleToStringConverter::DtoaMode mode) const {
  DOUBLE_CONVERSION_ASSERT(mode == SHORTEST || mode == SHORTEST_SINGLE);
  if (Double(value).IsSpecial()) {
    return HandleSpecialValues(value, result_builder);
  }

  int decimal_point;
  bool sign;
  const int kDecimalRepCapacity = kBase10MaximalLength + 1;
  char decimal_rep[kDecimalRepCapacity];
  int decimal_rep_length;

  DoubleToAscii(value, mode, 0, decimal_rep, kDecimalRepCapacity,
                &sign, &decimal_rep_length, &decimal_point);

  bool unique_zero = (flags_ & UNIQUE_ZERO) != 0;
  if (sign && (value != 0.0 || !unique_zero)) {
    result_builder->AddCharacter('-');
  }

  int exponent = decimal_point - 1;
  if ((decimal_in_shortest_low_ <= exponent) &&
      (exponent < decimal_in_shortest_high_)) {
    CreateDecimalRepresentation(decimal_rep, decimal_rep_length,
                                decimal_point,
                                (std::max)(0, decimal_rep_length - decimal_point),
                                result_builder);
  } else {
    CreateExponentialRepresentation(decimal_rep, decimal_rep_length, exponent,
                                    result_builder);
  }
  return true;
}


bool DoubleToStringConverter::ToFixed(double value,
                                      int requested_digits,
                                      StringBuilder* result_builder) const {
  DOUBLE_CONVERSION_ASSERT(kMaxFixedDigitsBeforePoint == 60);
  const double kFirstNonFixed = 1e60;

  if (Double(value).IsSpecial()) {
    return HandleSpecialValues(value, result_builder);
  }

  if (requested_digits > kMaxFixedDigitsAfterPoint) return false;
  if (value >= kFirstNonFixed || value <= -kFirstNonFixed) return false;

  // Find a sufficiently precise decimal representation of n.
  int decimal_point;
  bool sign;
  // Add space for the '\0' byte.
  const int kDecimalRepCapacity =
      kMaxFixedDigitsBeforePoint + kMaxFixedDigitsAfterPoint + 1;
  char decimal_rep[kDecimalRepCapacity];
  int decimal_rep_length;
  DoubleToAscii(value, FIXED, requested_digits,
                decimal_rep, kDecimalRepCapacity,
                &sign, &decimal_rep_length, &decimal_point);

  bool unique_zero = ((flags_ & UNIQUE_ZERO) != 0);
  if (sign && (value != 0.0 || !unique_zero)) {
    result_builder->AddCharacter('-');
  }

  CreateDecimalRepresentation(decimal_rep, decimal_rep_length, decimal_point,
                              requested_digits, result_builder);
  return true;
}


bool DoubleToStringConverter::ToExponential(
    double value,
    int requested_digits,
    StringBuilder* result_builder) const {
  if (Double(value).IsSpecial()) {
    return HandleSpecialValues(value, result_builder);
  }

  if (requested_digits < -1) return false;
  if (requested_digits > kMaxExponentialDigits) return false;

  int decimal_point;
  bool sign;
  // Add space for digit before the decimal point and the '\0' character.
  const int kDecimalRepCapacity = kMaxExponentialDigits + 2;
  DOUBLE_CONVERSION_ASSERT(kDecimalRepCapacity > kBase10MaximalLength);
  char decimal_rep[kDecimalRepCapacity];
#ifndef NDEBUG
  // Problem: there is an assert in StringBuilder::AddSubstring() that
  // will pass this buffer to strlen(), and this buffer is not generally
  // null-terminated.
  memset(decimal_rep, 0, sizeof(decimal_rep));
#endif
  int decimal_rep_length;

  if (requested_digits == -1) {
    DoubleToAscii(value, SHORTEST, 0,
                  decimal_rep, kDecimalRepCapacity,
                  &sign, &decimal_rep_length, &decimal_point);
  } else {
    DoubleToAscii(value, PRECISION, requested_digits + 1,
                  decimal_rep, kDecimalRepCapacity,
                  &sign, &decimal_rep_length, &decimal_point);
    DOUBLE_CONVERSION_ASSERT(decimal_rep_length <= requested_digits + 1);

    for (int i = decimal_rep_length; i < requested_digits + 1; ++i) {
      decimal_rep[i] = '0';
    }
    decimal_rep_length = requested_digits + 1;
  }

  bool unique_zero = ((flags_ & UNIQUE_ZERO) != 0);
  if (sign && (value != 0.0 || !unique_zero)) {
    result_builder->AddCharacter('-');
  }

  int exponent = decimal_point - 1;
  CreateExponentialRepresentation(decimal_rep,
                                  decimal_rep_length,
                                  exponent,
                                  result_builder);
  return true;
}


bool DoubleToStringConverter::ToPrecision(double value,
                                          int precision,
                                          StringBuilder* result_builder) const {
  if (Double(value).IsSpecial()) {
    return HandleSpecialValues(value, result_builder);
  }

  if (precision < kMinPrecisionDigits || precision > kMaxPrecisionDigits) {
    return false;
  }

  // Find a sufficiently precise decimal representation of n.
  int decimal_point;
  bool sign;
  // Add one for the terminating null character.
  const int kDecimalRepCapacity = kMaxPrecisionDigits + 1;
  char decimal_rep[kDecimalRepCapacity];
  int decimal_rep_length;

  DoubleToAscii(value, PRECISION, precision,
                decimal_rep, kDecimalRepCapacity,
                &sign, &decimal_rep_length, &decimal_point);
  DOUBLE_CONVERSION_ASSERT(decimal_rep_length <= precision);

  bool unique_zero = ((flags_ & UNIQUE_ZERO) != 0);
  if (sign && (value != 0.0 || !unique_zero)) {
    result_builder->AddCharacter('-');
  }

  // The exponent if we print the number as x.xxeyyy. That is with the
  // decimal point after the first digit.
  int exponent = decimal_point - 1;

  int extra_zero = ((flags_ & EMIT_TRAILING_ZERO_AFTER_POINT) != 0) ? 1 : 0;
  if ((-decimal_point + 1 > max_leading_padding_zeroes_in_precision_mode_) ||
      (decimal_point - precision + extra_zero >
       max_trailing_padding_zeroes_in_precision_mode_)) {
    // Fill buffer to contain 'precision' digits.
    // Usually the buffer is already at the correct length, but 'DoubleToAscii'
    // is allowed to return less characters.
    for (int i = decimal_rep_length; i < precision; ++i) {
      decimal_rep[i] = '0';
    }

    CreateExponentialRepresentation(decimal_rep,
                                    precision,
                                    exponent,
                                    result_builder);
  } else {
    CreateDecimalRepresentation(decimal_rep, decimal_rep_length, decimal_point,
                                (std::max)(0, precision - decimal_point),
                                result_builder);
  }
  return true;
}
#endif // not needed for ICU


static BignumDtoaMode DtoaToBignumDtoaMode(
    DoubleToStringConverter::DtoaMode dtoa_mode) {
  switch (dtoa_mode) {
    case DoubleToStringConverter::SHORTEST:  return BIGNUM_DTOA_SHORTEST;
    case DoubleToStringConverter::SHORTEST_SINGLE:
        return BIGNUM_DTOA_SHORTEST_SINGLE;
    case DoubleToStringConverter::FIXED:     return BIGNUM_DTOA_FIXED;
    case DoubleToStringConverter::PRECISION: return BIGNUM_DTOA_PRECISION;
    default:
      DOUBLE_CONVERSION_UNREACHABLE();
  }
}


void DoubleToStringConverter::DoubleToAscii(double v,
                                            DtoaMode mode,
                                            int requested_digits,
                                            char* buffer,
                                            int buffer_length,
                                            bool* sign,
                                            int* length,
                                            int* point) {
  Vector<char> vector(buffer, buffer_length);
  DOUBLE_CONVERSION_ASSERT(!Double(v).IsSpecial());
  DOUBLE_CONVERSION_ASSERT(mode == SHORTEST || mode == SHORTEST_SINGLE || requested_digits >= 0);

  if (Double(v).Sign() < 0) {
    *sign = true;
    v = -v;
  } else {
    *sign = false;
  }

  if (mode == PRECISION && requested_digits == 0) {
    vector[0] = '\0';
    *length = 0;
    return;
  }

  if (v == 0) {
    vector[0] = '0';
    vector[1] = '\0';
    *length = 1;
    *point = 1;
    return;
  }

  bool fast_worked;
  switch (mode) {
    case SHORTEST:
      fast_worked = FastDtoa(v, FAST_DTOA_SHORTEST, 0, vector, length, point);
      break;
#if 0 // not needed for ICU
    case SHORTEST_SINGLE:
      fast_worked = FastDtoa(v, FAST_DTOA_SHORTEST_SINGLE, 0,
                             vector, length, point);
      break;
    case FIXED:
      fast_worked = FastFixedDtoa(v, requested_digits, vector, length, point);
      break;
    case PRECISION:
      fast_worked = FastDtoa(v, FAST_DTOA_PRECISION, requested_digits,
                             vector, length, point);
      break;
#endif // not needed for ICU
    default:
      fast_worked = false;
      DOUBLE_CONVERSION_UNREACHABLE();
  }
  if (fast_worked) return;

  // If the fast dtoa didn't succeed use the slower bignum version.
  BignumDtoaMode bignum_mode = DtoaToBignumDtoaMode(mode);
  BignumDtoa(v, bignum_mode, requested_digits, vector, length, point);
  vector[*length] = '\0';
}

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END
#endif // ICU PATCH: close #if !UCONFIG_NO_FORMATTING
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2012 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING

// ICU PATCH: Customize header file paths for ICU.







// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

// The minimal and maximal target exponent define the range of w's binary
// exponent, where 'w' is the result of multiplying the input by a cached power
// of ten.
//
// A different range might be chosen on a different platform, to optimize digit
// generation, but a smaller range requires more powers of ten to be cached.
static const int kMinimalTargetExponent = -60;
static const int kMaximalTargetExponent = -32;


// Adjusts the last digit of the generated number, and screens out generated
// solutions that may be inaccurate. A solution may be inaccurate if it is
// outside the safe interval, or if we cannot prove that it is closer to the
// input than a neighboring representation of the same length.
//
// Input: * buffer containing the digits of too_high / 10^kappa
//        * the buffer's length
//        * distance_too_high_w == (too_high - w).f() * unit
//        * unsafe_interval == (too_high - too_low).f() * unit
//        * rest = (too_high - buffer * 10^kappa).f() * unit
//        * ten_kappa = 10^kappa * unit
//        * unit = the common multiplier
// Output: returns true if the buffer is guaranteed to contain the closest
//    representable number to the input.
//  Modifies the generated digits in the buffer to approach (round towards) w.
static bool RoundWeed(Vector<char> buffer,
                      int length,
                      uint64_t distance_too_high_w,
                      uint64_t unsafe_interval,
                      uint64_t rest,
                      uint64_t ten_kappa,
                      uint64_t unit) {
  uint64_t small_distance = distance_too_high_w - unit;
  uint64_t big_distance = distance_too_high_w + unit;
  // Let w_low  = too_high - big_distance, and
  //     w_high = too_high - small_distance.
  // Note: w_low < w < w_high
  //
  // The real w (* unit) must lie somewhere inside the interval
  // ]w_low; w_high[ (often written as "(w_low; w_high)")

  // Basically the buffer currently contains a number in the unsafe interval
  // ]too_low; too_high[ with too_low < w < too_high
  //
  //  too_high - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //                     ^v 1 unit            ^      ^                 ^      ^
  //  boundary_high ---------------------     .      .                 .      .
  //                     ^v 1 unit            .      .                 .      .
  //   - - - - - - - - - - - - - - - - - - -  +  - - + - - - - - -     .      .
  //                                          .      .         ^       .      .
  //                                          .  big_distance  .       .      .
  //                                          .      .         .       .    rest
  //                              small_distance     .         .       .      .
  //                                          v      .         .       .      .
  //  w_high - - - - - - - - - - - - - - - - - -     .         .       .      .
  //                     ^v 1 unit                   .         .       .      .
  //  w ----------------------------------------     .         .       .      .
  //                     ^v 1 unit                   v         .       .      .
  //  w_low  - - - - - - - - - - - - - - - - - - - - -         .       .      .
  //                                                           .       .      v
  //  buffer --------------------------------------------------+-------+--------
  //                                                           .       .
  //                                                  safe_interval    .
  //                                                           v       .
  //   - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -     .
  //                     ^v 1 unit                                     .
  //  boundary_low -------------------------                     unsafe_interval
  //                     ^v 1 unit                                     v
  //  too_low  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //
  //
  // Note that the value of buffer could lie anywhere inside the range too_low
  // to too_high.
  //
  // boundary_low, boundary_high and w are approximations of the real boundaries
  // and v (the input number). They are guaranteed to be precise up to one unit.
  // In fact the error is guaranteed to be strictly less than one unit.
  //
  // Anything that lies outside the unsafe interval is guaranteed not to round
  // to v when read again.
  // Anything that lies inside the safe interval is guaranteed to round to v
  // when read again.
  // If the number inside the buffer lies inside the unsafe interval but not
  // inside the safe interval then we simply do not know and bail out (returning
  // false).
  //
  // Similarly we have to take into account the imprecision of 'w' when finding
  // the closest representation of 'w'. If we have two potential
  // representations, and one is closer to both w_low and w_high, then we know
  // it is closer to the actual value v.
  //
  // By generating the digits of too_high we got the largest (closest to
  // too_high) buffer that is still in the unsafe interval. In the case where
  // w_high < buffer < too_high we try to decrement the buffer.
  // This way the buffer approaches (rounds towards) w.
  // There are 3 conditions that stop the decrementation process:
  //   1) the buffer is already below w_high
  //   2) decrementing the buffer would make it leave the unsafe interval
  //   3) decrementing the buffer would yield a number below w_high and farther
  //      away than the current number. In other words:
  //              (buffer{-1} < w_high) && w_high - buffer{-1} > buffer - w_high
  // Instead of using the buffer directly we use its distance to too_high.
  // Conceptually rest ~= too_high - buffer
  // We need to do the following tests in this order to avoid over- and
  // underflows.
  DOUBLE_CONVERSION_ASSERT(rest <= unsafe_interval);
  while (rest < small_distance &&  // Negated condition 1
         unsafe_interval - rest >= ten_kappa &&  // Negated condition 2
         (rest + ten_kappa < small_distance ||  // buffer{-1} > w_high
          small_distance - rest >= rest + ten_kappa - small_distance)) {
    buffer[length - 1]--;
    rest += ten_kappa;
  }

  // We have approached w+ as much as possible. We now test if approaching w-
  // would require changing the buffer. If yes, then we have two possible
  // representations close to w, but we cannot decide which one is closer.
  if (rest < big_distance &&
      unsafe_interval - rest >= ten_kappa &&
      (rest + ten_kappa < big_distance ||
       big_distance - rest > rest + ten_kappa - big_distance)) {
    return false;
  }

  // Weeding test.
  //   The safe interval is [too_low + 2 ulp; too_high - 2 ulp]
  //   Since too_low = too_high - unsafe_interval this is equivalent to
  //      [too_high - unsafe_interval + 4 ulp; too_high - 2 ulp]
  //   Conceptually we have: rest ~= too_high - buffer
  return (2 * unit <= rest) && (rest <= unsafe_interval - 4 * unit);
}


// Rounds the buffer upwards if the result is closer to v by possibly adding
// 1 to the buffer. If the precision of the calculation is not sufficient to
// round correctly, return false.
// The rounding might shift the whole buffer in which case the kappa is
// adjusted. For example "99", kappa = 3 might become "10", kappa = 4.
//
// If 2*rest > ten_kappa then the buffer needs to be round up.
// rest can have an error of +/- 1 unit. This function accounts for the
// imprecision and returns false, if the rounding direction cannot be
// unambiguously determined.
//
// Precondition: rest < ten_kappa.
static bool RoundWeedCounted(Vector<char> buffer,
                             int length,
                             uint64_t rest,
                             uint64_t ten_kappa,
                             uint64_t unit,
                             int* kappa) {
  DOUBLE_CONVERSION_ASSERT(rest < ten_kappa);
  // The following tests are done in a specific order to avoid overflows. They
  // will work correctly with any uint64 values of rest < ten_kappa and unit.
  //
  // If the unit is too big, then we don't know which way to round. For example
  // a unit of 50 means that the real number lies within rest +/- 50. If
  // 10^kappa == 40 then there is no way to tell which way to round.
  if (unit >= ten_kappa) return false;
  // Even if unit is just half the size of 10^kappa we are already completely
  // lost. (And after the previous test we know that the expression will not
  // over/underflow.)
  if (ten_kappa - unit <= unit) return false;
  // If 2 * (rest + unit) <= 10^kappa we can safely round down.
  if ((ten_kappa - rest > rest) && (ten_kappa - 2 * rest >= 2 * unit)) {
    return true;
  }
  // If 2 * (rest - unit) >= 10^kappa, then we can safely round up.
  if ((rest > unit) && (ten_kappa - (rest - unit) <= (rest - unit))) {
    // Increment the last digit recursively until we find a non '9' digit.
    buffer[length - 1]++;
    for (int i = length - 1; i > 0; --i) {
      if (buffer[i] != '0' + 10) break;
      buffer[i] = '0';
      buffer[i - 1]++;
    }
    // If the first digit is now '0'+ 10 we had a buffer with all '9's. With the
    // exception of the first digit all digits are now '0'. Simply switch the
    // first digit to '1' and adjust the kappa. Example: "99" becomes "10" and
    // the power (the kappa) is increased.
    if (buffer[0] == '0' + 10) {
      buffer[0] = '1';
      (*kappa) += 1;
    }
    return true;
  }
  return false;
}

// Returns the biggest power of ten that is less than or equal to the given
// number. We furthermore receive the maximum number of bits 'number' has.
//
// Returns power == 10^(exponent_plus_one-1) such that
//    power <= number < power * 10.
// If number_bits == 0 then 0^(0-1) is returned.
// The number of bits must be <= 32.
// Precondition: number < (1 << (number_bits + 1)).

// Inspired by the method for finding an integer log base 10 from here:
// http://graphics.stanford.edu/~seander/bithacks.html#IntegerLog10
static unsigned int const kSmallPowersOfTen[] =
    {0, 1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000,
     1000000000};

static void BiggestPowerTen(uint32_t number,
                            int number_bits,
                            uint32_t* power,
                            int* exponent_plus_one) {
  DOUBLE_CONVERSION_ASSERT(number < (1u << (number_bits + 1)));
  // 1233/4096 is approximately 1/lg(10).
  int exponent_plus_one_guess = ((number_bits + 1) * 1233 >> 12);
  // We increment to skip over the first entry in the kPowersOf10 table.
  // Note: kPowersOf10[i] == 10^(i-1).
  exponent_plus_one_guess++;
  // We don't have any guarantees that 2^number_bits <= number.
  if (number < kSmallPowersOfTen[exponent_plus_one_guess]) {
    exponent_plus_one_guess--;
  }
  *power = kSmallPowersOfTen[exponent_plus_one_guess];
  *exponent_plus_one = exponent_plus_one_guess;
}

// Generates the digits of input number w.
// w is a floating-point number (DiyFp), consisting of a significand and an
// exponent. Its exponent is bounded by kMinimalTargetExponent and
// kMaximalTargetExponent.
//       Hence -60 <= w.e() <= -32.
//
// Returns false if it fails, in which case the generated digits in the buffer
// should not be used.
// Preconditions:
//  * low, w and high are correct up to 1 ulp (unit in the last place). That
//    is, their error must be less than a unit of their last digits.
//  * low.e() == w.e() == high.e()
//  * low < w < high, and taking into account their error: low~ <= high~
//  * kMinimalTargetExponent <= w.e() <= kMaximalTargetExponent
// Postconditions: returns false if procedure fails.
//   otherwise:
//     * buffer is not null-terminated, but len contains the number of digits.
//     * buffer contains the shortest possible decimal digit-sequence
//       such that LOW < buffer * 10^kappa < HIGH, where LOW and HIGH are the
//       correct values of low and high (without their error).
//     * if more than one decimal representation gives the minimal number of
//       decimal digits then the one closest to W (where W is the correct value
//       of w) is chosen.
// Remark: this procedure takes into account the imprecision of its input
//   numbers. If the precision is not enough to guarantee all the postconditions
//   then false is returned. This usually happens rarely (~0.5%).
//
// Say, for the sake of example, that
//   w.e() == -48, and w.f() == 0x1234567890abcdef
// w's value can be computed by w.f() * 2^w.e()
// We can obtain w's integral digits by simply shifting w.f() by -w.e().
//  -> w's integral part is 0x1234
//  w's fractional part is therefore 0x567890abcdef.
// Printing w's integral part is easy (simply print 0x1234 in decimal).
// In order to print its fraction we repeatedly multiply the fraction by 10 and
// get each digit. Example the first digit after the point would be computed by
//   (0x567890abcdef * 10) >> 48. -> 3
// The whole thing becomes slightly more complicated because we want to stop
// once we have enough digits. That is, once the digits inside the buffer
// represent 'w' we can stop. Everything inside the interval low - high
// represents w. However we have to pay attention to low, high and w's
// imprecision.
static bool DigitGen(DiyFp low,
                     DiyFp w,
                     DiyFp high,
                     Vector<char> buffer,
                     int* length,
                     int* kappa) {
  DOUBLE_CONVERSION_ASSERT(low.e() == w.e() && w.e() == high.e());
  DOUBLE_CONVERSION_ASSERT(low.f() + 1 <= high.f() - 1);
  DOUBLE_CONVERSION_ASSERT(kMinimalTargetExponent <= w.e() && w.e() <= kMaximalTargetExponent);
  // low, w and high are imprecise, but by less than one ulp (unit in the last
  // place).
  // If we remove (resp. add) 1 ulp from low (resp. high) we are certain that
  // the new numbers are outside of the interval we want the final
  // representation to lie in.
  // Inversely adding (resp. removing) 1 ulp from low (resp. high) would yield
  // numbers that are certain to lie in the interval. We will use this fact
  // later on.
  // We will now start by generating the digits within the uncertain
  // interval. Later we will weed out representations that lie outside the safe
  // interval and thus _might_ lie outside the correct interval.
  uint64_t unit = 1;
  DiyFp too_low = DiyFp(low.f() - unit, low.e());
  DiyFp too_high = DiyFp(high.f() + unit, high.e());
  // too_low and too_high are guaranteed to lie outside the interval we want the
  // generated number in.
  DiyFp unsafe_interval = DiyFp::Minus(too_high, too_low);
  // We now cut the input number into two parts: the integral digits and the
  // fractionals. We will not write any decimal separator though, but adapt
  // kappa instead.
  // Reminder: we are currently computing the digits (stored inside the buffer)
  // such that:   too_low < buffer * 10^kappa < too_high
  // We use too_high for the digit_generation and stop as soon as possible.
  // If we stop early we effectively round down.
  DiyFp one = DiyFp(static_cast<uint64_t>(1) << -w.e(), w.e());
  // Division by one is a shift.
  uint32_t integrals = static_cast<uint32_t>(too_high.f() >> -one.e());
  // Modulo by one is an and.
  uint64_t fractionals = too_high.f() & (one.f() - 1);
  uint32_t divisor;
  int divisor_exponent_plus_one;
  BiggestPowerTen(integrals, DiyFp::kSignificandSize - (-one.e()),
                  &divisor, &divisor_exponent_plus_one);
  *kappa = divisor_exponent_plus_one;
  *length = 0;
  // Loop invariant: buffer = too_high / 10^kappa  (integer division)
  // The invariant holds for the first iteration: kappa has been initialized
  // with the divisor exponent + 1. And the divisor is the biggest power of ten
  // that is smaller than integrals.
  while (*kappa > 0) {
    int digit = integrals / divisor;
    DOUBLE_CONVERSION_ASSERT(digit <= 9);
    buffer[*length] = static_cast<char>('0' + digit);
    (*length)++;
    integrals %= divisor;
    (*kappa)--;
    // Note that kappa now equals the exponent of the divisor and that the
    // invariant thus holds again.
    uint64_t rest =
        (static_cast<uint64_t>(integrals) << -one.e()) + fractionals;
    // Invariant: too_high = buffer * 10^kappa + DiyFp(rest, one.e())
    // Reminder: unsafe_interval.e() == one.e()
    if (rest < unsafe_interval.f()) {
      // Rounding down (by not emitting the remaining digits) yields a number
      // that lies within the unsafe interval.
      return RoundWeed(buffer, *length, DiyFp::Minus(too_high, w).f(),
                       unsafe_interval.f(), rest,
                       static_cast<uint64_t>(divisor) << -one.e(), unit);
    }
    divisor /= 10;
  }

  // The integrals have been generated. We are at the point of the decimal
  // separator. In the following loop we simply multiply the remaining digits by
  // 10 and divide by one. We just need to pay attention to multiply associated
  // data (like the interval or 'unit'), too.
  // Note that the multiplication by 10 does not overflow, because w.e >= -60
  // and thus one.e >= -60.
  DOUBLE_CONVERSION_ASSERT(one.e() >= -60);
  DOUBLE_CONVERSION_ASSERT(fractionals < one.f());
  DOUBLE_CONVERSION_ASSERT(DOUBLE_CONVERSION_UINT64_2PART_C(0xFFFFFFFF, FFFFFFFF) / 10 >= one.f());
  for (;;) {
    fractionals *= 10;
    unit *= 10;
    unsafe_interval.set_f(unsafe_interval.f() * 10);
    // Integer division by one.
    int digit = static_cast<int>(fractionals >> -one.e());
    DOUBLE_CONVERSION_ASSERT(digit <= 9);
    buffer[*length] = static_cast<char>('0' + digit);
    (*length)++;
    fractionals &= one.f() - 1;  // Modulo by one.
    (*kappa)--;
    if (fractionals < unsafe_interval.f()) {
      return RoundWeed(buffer, *length, DiyFp::Minus(too_high, w).f() * unit,
                       unsafe_interval.f(), fractionals, one.f(), unit);
    }
  }
}



// Generates (at most) requested_digits digits of input number w.
// w is a floating-point number (DiyFp), consisting of a significand and an
// exponent. Its exponent is bounded by kMinimalTargetExponent and
// kMaximalTargetExponent.
//       Hence -60 <= w.e() <= -32.
//
// Returns false if it fails, in which case the generated digits in the buffer
// should not be used.
// Preconditions:
//  * w is correct up to 1 ulp (unit in the last place). That
//    is, its error must be strictly less than a unit of its last digit.
//  * kMinimalTargetExponent <= w.e() <= kMaximalTargetExponent
//
// Postconditions: returns false if procedure fails.
//   otherwise:
//     * buffer is not null-terminated, but length contains the number of
//       digits.
//     * the representation in buffer is the most precise representation of
//       requested_digits digits.
//     * buffer contains at most requested_digits digits of w. If there are less
//       than requested_digits digits then some trailing '0's have been removed.
//     * kappa is such that
//            w = buffer * 10^kappa + eps with |eps| < 10^kappa / 2.
//
// Remark: This procedure takes into account the imprecision of its input
//   numbers. If the precision is not enough to guarantee all the postconditions
//   then false is returned. This usually happens rarely, but the failure-rate
//   increases with higher requested_digits.
static bool DigitGenCounted(DiyFp w,
                            int requested_digits,
                            Vector<char> buffer,
                            int* length,
                            int* kappa) {
  DOUBLE_CONVERSION_ASSERT(kMinimalTargetExponent <= w.e() && w.e() <= kMaximalTargetExponent);
  DOUBLE_CONVERSION_ASSERT(kMinimalTargetExponent >= -60);
  DOUBLE_CONVERSION_ASSERT(kMaximalTargetExponent <= -32);
  // w is assumed to have an error less than 1 unit. Whenever w is scaled we
  // also scale its error.
  uint64_t w_error = 1;
  // We cut the input number into two parts: the integral digits and the
  // fractional digits. We don't emit any decimal separator, but adapt kappa
  // instead. Example: instead of writing "1.2" we put "12" into the buffer and
  // increase kappa by 1.
  DiyFp one = DiyFp(static_cast<uint64_t>(1) << -w.e(), w.e());
  // Division by one is a shift.
  uint32_t integrals = static_cast<uint32_t>(w.f() >> -one.e());
  // Modulo by one is an and.
  uint64_t fractionals = w.f() & (one.f() - 1);
  uint32_t divisor;
  int divisor_exponent_plus_one;
  BiggestPowerTen(integrals, DiyFp::kSignificandSize - (-one.e()),
                  &divisor, &divisor_exponent_plus_one);
  *kappa = divisor_exponent_plus_one;
  *length = 0;

  // Loop invariant: buffer = w / 10^kappa  (integer division)
  // The invariant holds for the first iteration: kappa has been initialized
  // with the divisor exponent + 1. And the divisor is the biggest power of ten
  // that is smaller than 'integrals'.
  while (*kappa > 0) {
    int digit = integrals / divisor;
    DOUBLE_CONVERSION_ASSERT(digit <= 9);
    buffer[*length] = static_cast<char>('0' + digit);
    (*length)++;
    requested_digits--;
    integrals %= divisor;
    (*kappa)--;
    // Note that kappa now equals the exponent of the divisor and that the
    // invariant thus holds again.
    if (requested_digits == 0) break;
    divisor /= 10;
  }

  if (requested_digits == 0) {
    uint64_t rest =
        (static_cast<uint64_t>(integrals) << -one.e()) + fractionals;
    return RoundWeedCounted(buffer, *length, rest,
                            static_cast<uint64_t>(divisor) << -one.e(), w_error,
                            kappa);
  }

  // The integrals have been generated. We are at the point of the decimal
  // separator. In the following loop we simply multiply the remaining digits by
  // 10 and divide by one. We just need to pay attention to multiply associated
  // data (the 'unit'), too.
  // Note that the multiplication by 10 does not overflow, because w.e >= -60
  // and thus one.e >= -60.
  DOUBLE_CONVERSION_ASSERT(one.e() >= -60);
  DOUBLE_CONVERSION_ASSERT(fractionals < one.f());
  DOUBLE_CONVERSION_ASSERT(DOUBLE_CONVERSION_UINT64_2PART_C(0xFFFFFFFF, FFFFFFFF) / 10 >= one.f());
  while (requested_digits > 0 && fractionals > w_error) {
    fractionals *= 10;
    w_error *= 10;
    // Integer division by one.
    int digit = static_cast<int>(fractionals >> -one.e());
    DOUBLE_CONVERSION_ASSERT(digit <= 9);
    buffer[*length] = static_cast<char>('0' + digit);
    (*length)++;
    requested_digits--;
    fractionals &= one.f() - 1;  // Modulo by one.
    (*kappa)--;
  }
  if (requested_digits != 0) return false;
  return RoundWeedCounted(buffer, *length, fractionals, one.f(), w_error,
                          kappa);
}


// Provides a decimal representation of v.
// Returns true if it succeeds, otherwise the result cannot be trusted.
// There will be *length digits inside the buffer (not null-terminated).
// If the function returns true then
//        v == (double) (buffer * 10^decimal_exponent).
// The digits in the buffer are the shortest representation possible: no
// 0.09999999999999999 instead of 0.1. The shorter representation will even be
// chosen even if the longer one would be closer to v.
// The last digit will be closest to the actual v. That is, even if several
// digits might correctly yield 'v' when read again, the closest will be
// computed.
static bool Grisu3(double v,
                   FastDtoaMode mode,
                   Vector<char> buffer,
                   int* length,
                   int* decimal_exponent) {
  DiyFp w = Double(v).AsNormalizedDiyFp();
  // boundary_minus and boundary_plus are the boundaries between v and its
  // closest floating-point neighbors. Any number strictly between
  // boundary_minus and boundary_plus will round to v when convert to a double.
  // Grisu3 will never output representations that lie exactly on a boundary.
  DiyFp boundary_minus, boundary_plus;
  if (mode == FAST_DTOA_SHORTEST) {
    Double(v).NormalizedBoundaries(&boundary_minus, &boundary_plus);
  } else {
    DOUBLE_CONVERSION_ASSERT(mode == FAST_DTOA_SHORTEST_SINGLE);
    float single_v = static_cast<float>(v);
    Single(single_v).NormalizedBoundaries(&boundary_minus, &boundary_plus);
  }
  DOUBLE_CONVERSION_ASSERT(boundary_plus.e() == w.e());
  DiyFp ten_mk;  // Cached power of ten: 10^-k
  int mk;        // -k
  int ten_mk_minimal_binary_exponent =
     kMinimalTargetExponent - (w.e() + DiyFp::kSignificandSize);
  int ten_mk_maximal_binary_exponent =
     kMaximalTargetExponent - (w.e() + DiyFp::kSignificandSize);
  PowersOfTenCache::GetCachedPowerForBinaryExponentRange(
      ten_mk_minimal_binary_exponent,
      ten_mk_maximal_binary_exponent,
      &ten_mk, &mk);
  DOUBLE_CONVERSION_ASSERT((kMinimalTargetExponent <= w.e() + ten_mk.e() +
          DiyFp::kSignificandSize) &&
         (kMaximalTargetExponent >= w.e() + ten_mk.e() +
          DiyFp::kSignificandSize));
  // Note that ten_mk is only an approximation of 10^-k. A DiyFp only contains a
  // 64 bit significand and ten_mk is thus only precise up to 64 bits.

  // The DiyFp::Times procedure rounds its result, and ten_mk is approximated
  // too. The variable scaled_w (as well as scaled_boundary_minus/plus) are now
  // off by a small amount.
  // In fact: scaled_w - w*10^k < 1ulp (unit in the last place) of scaled_w.
  // In other words: let f = scaled_w.f() and e = scaled_w.e(), then
  //           (f-1) * 2^e < w*10^k < (f+1) * 2^e
  DiyFp scaled_w = DiyFp::Times(w, ten_mk);
  DOUBLE_CONVERSION_ASSERT(scaled_w.e() ==
         boundary_plus.e() + ten_mk.e() + DiyFp::kSignificandSize);
  // In theory it would be possible to avoid some recomputations by computing
  // the difference between w and boundary_minus/plus (a power of 2) and to
  // compute scaled_boundary_minus/plus by subtracting/adding from
  // scaled_w. However the code becomes much less readable and the speed
  // enhancements are not terriffic.
  DiyFp scaled_boundary_minus = DiyFp::Times(boundary_minus, ten_mk);
  DiyFp scaled_boundary_plus  = DiyFp::Times(boundary_plus,  ten_mk);

  // DigitGen will generate the digits of scaled_w. Therefore we have
  // v == (double) (scaled_w * 10^-mk).
  // Set decimal_exponent == -mk and pass it to DigitGen. If scaled_w is not an
  // integer than it will be updated. For instance if scaled_w == 1.23 then
  // the buffer will be filled with "123" und the decimal_exponent will be
  // decreased by 2.
  int kappa;
  bool result = DigitGen(scaled_boundary_minus, scaled_w, scaled_boundary_plus,
                         buffer, length, &kappa);
  *decimal_exponent = -mk + kappa;
  return result;
}


// The "counted" version of grisu3 (see above) only generates requested_digits
// number of digits. This version does not generate the shortest representation,
// and with enough requested digits 0.1 will at some point print as 0.9999999...
// Grisu3 is too imprecise for real halfway cases (1.5 will not work) and
// therefore the rounding strategy for halfway cases is irrelevant.
static bool Grisu3Counted(double v,
                          int requested_digits,
                          Vector<char> buffer,
                          int* length,
                          int* decimal_exponent) {
  DiyFp w = Double(v).AsNormalizedDiyFp();
  DiyFp ten_mk;  // Cached power of ten: 10^-k
  int mk;        // -k
  int ten_mk_minimal_binary_exponent =
     kMinimalTargetExponent - (w.e() + DiyFp::kSignificandSize);
  int ten_mk_maximal_binary_exponent =
     kMaximalTargetExponent - (w.e() + DiyFp::kSignificandSize);
  PowersOfTenCache::GetCachedPowerForBinaryExponentRange(
      ten_mk_minimal_binary_exponent,
      ten_mk_maximal_binary_exponent,
      &ten_mk, &mk);
  DOUBLE_CONVERSION_ASSERT((kMinimalTargetExponent <= w.e() + ten_mk.e() +
          DiyFp::kSignificandSize) &&
         (kMaximalTargetExponent >= w.e() + ten_mk.e() +
          DiyFp::kSignificandSize));
  // Note that ten_mk is only an approximation of 10^-k. A DiyFp only contains a
  // 64 bit significand and ten_mk is thus only precise up to 64 bits.

  // The DiyFp::Times procedure rounds its result, and ten_mk is approximated
  // too. The variable scaled_w (as well as scaled_boundary_minus/plus) are now
  // off by a small amount.
  // In fact: scaled_w - w*10^k < 1ulp (unit in the last place) of scaled_w.
  // In other words: let f = scaled_w.f() and e = scaled_w.e(), then
  //           (f-1) * 2^e < w*10^k < (f+1) * 2^e
  DiyFp scaled_w = DiyFp::Times(w, ten_mk);

  // We now have (double) (scaled_w * 10^-mk).
  // DigitGen will generate the first requested_digits digits of scaled_w and
  // return together with a kappa such that scaled_w ~= buffer * 10^kappa. (It
  // will not always be exactly the same since DigitGenCounted only produces a
  // limited number of digits.)
  int kappa;
  bool result = DigitGenCounted(scaled_w, requested_digits,
                                buffer, length, &kappa);
  *decimal_exponent = -mk + kappa;
  return result;
}


bool FastDtoa(double v,
              FastDtoaMode mode,
              int requested_digits,
              Vector<char> buffer,
              int* length,
              int* decimal_point) {
  DOUBLE_CONVERSION_ASSERT(v > 0);
  DOUBLE_CONVERSION_ASSERT(!Double(v).IsSpecial());

  bool result = false;
  int decimal_exponent = 0;
  switch (mode) {
    case FAST_DTOA_SHORTEST:
    case FAST_DTOA_SHORTEST_SINGLE:
      result = Grisu3(v, mode, buffer, length, &decimal_exponent);
      break;
    case FAST_DTOA_PRECISION:
      result = Grisu3Counted(v, requested_digits,
                             buffer, length, &decimal_exponent);
      break;
    default:
      DOUBLE_CONVERSION_UNREACHABLE();
  }
  if (result) {
    *decimal_point = *length + decimal_exponent;
    buffer[*length] = '\0';
  }
  return result;
}

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END
#endif // ICU PATCH: close #if !UCONFIG_NO_FORMATTING
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING

// ICU PATCH: Do not include std::locale.

#include <climits>
// #include <locale>
#include <cmath>

// ICU PATCH: Customize header file paths for ICU.

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2012 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


// ICU PATCH: Customize header file paths for ICU.



// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

class StringToDoubleConverter {
 public:
  // Enumeration for allowing octals and ignoring junk when converting
  // strings to numbers.
  enum Flags {
    NO_FLAGS = 0,
    ALLOW_HEX = 1,
    ALLOW_OCTALS = 2,
    ALLOW_TRAILING_JUNK = 4,
    ALLOW_LEADING_SPACES = 8,
    ALLOW_TRAILING_SPACES = 16,
    ALLOW_SPACES_AFTER_SIGN = 32,
    ALLOW_CASE_INSENSITIVITY = 64,
    ALLOW_CASE_INSENSIBILITY = 64,  // Deprecated
    ALLOW_HEX_FLOATS = 128,
  };

  static const uc16 kNoSeparator = '\0';

  // Flags should be a bit-or combination of the possible Flags-enum.
  //  - NO_FLAGS: no special flags.
  //  - ALLOW_HEX: recognizes the prefix "0x". Hex numbers may only be integers.
  //      Ex: StringToDouble("0x1234") -> 4660.0
  //          In StringToDouble("0x1234.56") the characters ".56" are trailing
  //          junk. The result of the call is hence dependent on
  //          the ALLOW_TRAILING_JUNK flag and/or the junk value.
  //      With this flag "0x" is a junk-string. Even with ALLOW_TRAILING_JUNK,
  //      the string will not be parsed as "0" followed by junk.
  //
  //  - ALLOW_OCTALS: recognizes the prefix "0" for octals:
  //      If a sequence of octal digits starts with '0', then the number is
  //      read as octal integer. Octal numbers may only be integers.
  //      Ex: StringToDouble("01234") -> 668.0
  //          StringToDouble("012349") -> 12349.0  // Not a sequence of octal
  //                                               // digits.
  //          In StringToDouble("01234.56") the characters ".56" are trailing
  //          junk. The result of the call is hence dependent on
  //          the ALLOW_TRAILING_JUNK flag and/or the junk value.
  //          In StringToDouble("01234e56") the characters "e56" are trailing
  //          junk, too.
  //  - ALLOW_TRAILING_JUNK: ignore trailing characters that are not part of
  //      a double literal.
  //  - ALLOW_LEADING_SPACES: skip over leading whitespace, including spaces,
  //                          new-lines, and tabs.
  //  - ALLOW_TRAILING_SPACES: ignore trailing whitespace.
  //  - ALLOW_SPACES_AFTER_SIGN: ignore whitespace after the sign.
  //       Ex: StringToDouble("-   123.2") -> -123.2.
  //           StringToDouble("+   123.2") -> 123.2
  //  - ALLOW_CASE_INSENSITIVITY: ignore case of characters for special values:
  //      infinity and nan.
  //  - ALLOW_HEX_FLOATS: allows hexadecimal float literals.
  //      This *must* start with "0x" and separate the exponent with "p".
  //      Examples: 0x1.2p3 == 9.0
  //                0x10.1p0 == 16.0625
  //      ALLOW_HEX and ALLOW_HEX_FLOATS are indendent.
  //
  // empty_string_value is returned when an empty string is given as input.
  // If ALLOW_LEADING_SPACES or ALLOW_TRAILING_SPACES are set, then a string
  // containing only spaces is converted to the 'empty_string_value', too.
  //
  // junk_string_value is returned when
  //  a) ALLOW_TRAILING_JUNK is not set, and a junk character (a character not
  //     part of a double-literal) is found.
  //  b) ALLOW_TRAILING_JUNK is set, but the string does not start with a
  //     double literal.
  //
  // infinity_symbol and nan_symbol are strings that are used to detect
  // inputs that represent infinity and NaN. They can be null, in which case
  // they are ignored.
  // The conversion routine first reads any possible signs. Then it compares the
  // following character of the input-string with the first character of
  // the infinity, and nan-symbol. If either matches, the function assumes, that
  // a match has been found, and expects the following input characters to match
  // the remaining characters of the special-value symbol.
  // This means that the following restrictions apply to special-value symbols:
  //  - they must not start with signs ('+', or '-'),
  //  - they must not have the same first character.
  //  - they must not start with digits.
  //
  // If the separator character is not kNoSeparator, then that specific
  // character is ignored when in between two valid digits of the significant.
  // It is not allowed to appear in the exponent.
  // It is not allowed to lead or trail the number.
  // It is not allowed to appear twice next to each other.
  //
  // Examples:
  //  flags = ALLOW_HEX | ALLOW_TRAILING_JUNK,
  //  empty_string_value = 0.0,
  //  junk_string_value = NaN,
  //  infinity_symbol = "infinity",
  //  nan_symbol = "nan":
  //    StringToDouble("0x1234") -> 4660.0.
  //    StringToDouble("0x1234K") -> 4660.0.
  //    StringToDouble("") -> 0.0  // empty_string_value.
  //    StringToDouble(" ") -> NaN  // junk_string_value.
  //    StringToDouble(" 1") -> NaN  // junk_string_value.
  //    StringToDouble("0x") -> NaN  // junk_string_value.
  //    StringToDouble("-123.45") -> -123.45.
  //    StringToDouble("--123.45") -> NaN  // junk_string_value.
  //    StringToDouble("123e45") -> 123e45.
  //    StringToDouble("123E45") -> 123e45.
  //    StringToDouble("123e+45") -> 123e45.
  //    StringToDouble("123E-45") -> 123e-45.
  //    StringToDouble("123e") -> 123.0  // trailing junk ignored.
  //    StringToDouble("123e-") -> 123.0  // trailing junk ignored.
  //    StringToDouble("+NaN") -> NaN  // NaN string literal.
  //    StringToDouble("-infinity") -> -inf.  // infinity literal.
  //    StringToDouble("Infinity") -> NaN  // junk_string_value.
  //
  //  flags = ALLOW_OCTAL | ALLOW_LEADING_SPACES,
  //  empty_string_value = 0.0,
  //  junk_string_value = NaN,
  //  infinity_symbol = NULL,
  //  nan_symbol = NULL:
  //    StringToDouble("0x1234") -> NaN  // junk_string_value.
  //    StringToDouble("01234") -> 668.0.
  //    StringToDouble("") -> 0.0  // empty_string_value.
  //    StringToDouble(" ") -> 0.0  // empty_string_value.
  //    StringToDouble(" 1") -> 1.0
  //    StringToDouble("0x") -> NaN  // junk_string_value.
  //    StringToDouble("0123e45") -> NaN  // junk_string_value.
  //    StringToDouble("01239E45") -> 1239e45.
  //    StringToDouble("-infinity") -> NaN  // junk_string_value.
  //    StringToDouble("NaN") -> NaN  // junk_string_value.
  //
  //  flags = NO_FLAGS,
  //  separator = ' ':
  //    StringToDouble("1 2 3 4") -> 1234.0
  //    StringToDouble("1  2") -> NaN // junk_string_value
  //    StringToDouble("1 000 000.0") -> 1000000.0
  //    StringToDouble("1.000 000") -> 1.0
  //    StringToDouble("1.0e1 000") -> NaN // junk_string_value
  StringToDoubleConverter(int flags,
                          double empty_string_value,
                          double junk_string_value,
                          const char* infinity_symbol,
                          const char* nan_symbol,
                          uc16 separator = kNoSeparator)
      : flags_(flags),
        empty_string_value_(empty_string_value),
        junk_string_value_(junk_string_value),
        infinity_symbol_(infinity_symbol),
        nan_symbol_(nan_symbol),
        separator_(separator) {
  }

  // Performs the conversion.
  // The output parameter 'processed_characters_count' is set to the number
  // of characters that have been processed to read the number.
  // Spaces than are processed with ALLOW_{LEADING|TRAILING}_SPACES are included
  // in the 'processed_characters_count'. Trailing junk is never included.
  double StringToDouble(const char* buffer,
                        int length,
                        int* processed_characters_count) const;

  // Same as StringToDouble above but for 16 bit characters.
  double StringToDouble(const uc16* buffer,
                        int length,
                        int* processed_characters_count) const;

  // Same as StringToDouble but reads a float.
  // Note that this is not equivalent to static_cast<float>(StringToDouble(...))
  // due to potential double-rounding.
  float StringToFloat(const char* buffer,
                      int length,
                      int* processed_characters_count) const;

  // Same as StringToFloat above but for 16 bit characters.
  float StringToFloat(const uc16* buffer,
                      int length,
                      int* processed_characters_count) const;

 private:
  const int flags_;
  const double empty_string_value_;
  const double junk_string_value_;
  const char* const infinity_symbol_;
  const char* const nan_symbol_;
  const uc16 separator_;

  template <class Iterator>
  double StringToIeee(Iterator start_pointer,
                      int length,
                      bool read_as_double,
                      int* processed_characters_count) const;

  DOUBLE_CONVERSION_DISALLOW_IMPLICIT_CONSTRUCTORS(StringToDoubleConverter);
};

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END

#endif  // DOUBLE_CONVERSION_STRING_TO_DOUBLE_H_



// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


// ICU PATCH: Customize header file paths for ICU.



// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

// The buffer must only contain digits in the range [0-9]. It must not
// contain a dot or a sign. It must not start with '0', and must not be empty.
double Strtod(Vector<const char> buffer, int exponent);

// The buffer must only contain digits in the range [0-9]. It must not
// contain a dot or a sign. It must not start with '0', and must not be empty.
float Strtof(Vector<const char> buffer, int exponent);

// For special use cases, the heart of the Strtod() function is also available
// separately, it assumes that 'trimmed' is as produced by TrimAndCut(), i.e.
// no leading or trailing zeros, also no lone zero, and not 'too many' digits.
double StrtodTrimmed(Vector<const char> trimmed, int exponent);

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END

#endif  // DOUBLE_CONVERSION_STRTOD_H_



// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

namespace {

inline char ToLower(char ch) {
#if 0  // do not include std::locale in ICU
  static const std::ctype<char>& cType =
      std::use_facet<std::ctype<char> >(std::locale::classic());
  return cType.tolower(ch);
#else
  (void)ch;
  DOUBLE_CONVERSION_UNREACHABLE();
#endif
}

inline char Pass(char ch) {
  return ch;
}

template <class Iterator, class Converter>
static inline bool ConsumeSubStringImpl(Iterator* current,
                                        Iterator end,
                                        const char* substring,
                                        Converter converter) {
  DOUBLE_CONVERSION_ASSERT(converter(**current) == *substring);
  for (substring++; *substring != '\0'; substring++) {
    ++*current;
    if (*current == end || converter(**current) != *substring) {
      return false;
    }
  }
  ++*current;
  return true;
}

// Consumes the given substring from the iterator.
// Returns false, if the substring does not match.
template <class Iterator>
static bool ConsumeSubString(Iterator* current,
                             Iterator end,
                             const char* substring,
                             bool allow_case_insensitivity) {
  if (allow_case_insensitivity) {
    return ConsumeSubStringImpl(current, end, substring, ToLower);
  } else {
    return ConsumeSubStringImpl(current, end, substring, Pass);
  }
}

// Consumes first character of the str is equal to ch
inline bool ConsumeFirstCharacter(char ch,
                                         const char* str,
                                         bool case_insensitivity) {
  return case_insensitivity ? ToLower(ch) == str[0] : ch == str[0];
}
}  // namespace

// Maximum number of significant digits in decimal representation.
// The longest possible double in decimal representation is
// (2^53 - 1) * 2 ^ -1074 that is (2 ^ 53 - 1) * 5 ^ 1074 / 10 ^ 1074
// (768 digits). If we parse a number whose first digits are equal to a
// mean of 2 adjacent doubles (that could have up to 769 digits) the result
// must be rounded to the bigger one unless the tail consists of zeros, so
// we don't need to preserve all the digits.
const int kMaxSignificantDigits = 772;


static const char kWhitespaceTable7[] = { 32, 13, 10, 9, 11, 12 };
static const int kWhitespaceTable7Length = DOUBLE_CONVERSION_ARRAY_SIZE(kWhitespaceTable7);


static const uc16 kWhitespaceTable16[] = {
  160, 8232, 8233, 5760, 6158, 8192, 8193, 8194, 8195,
  8196, 8197, 8198, 8199, 8200, 8201, 8202, 8239, 8287, 12288, 65279
};
static const int kWhitespaceTable16Length = DOUBLE_CONVERSION_ARRAY_SIZE(kWhitespaceTable16);


static bool isWhitespace(int x) {
  if (x < 128) {
    for (int i = 0; i < kWhitespaceTable7Length; i++) {
      if (kWhitespaceTable7[i] == x) return true;
    }
  } else {
    for (int i = 0; i < kWhitespaceTable16Length; i++) {
      if (kWhitespaceTable16[i] == x) return true;
    }
  }
  return false;
}


// Returns true if a nonspace found and false if the end has reached.
template <class Iterator>
static inline bool AdvanceToNonspace(Iterator* current, Iterator end) {
  while (*current != end) {
    if (!isWhitespace(**current)) return true;
    ++*current;
  }
  return false;
}


static bool isDigit(int x, int radix) {
  return (x >= '0' && x <= '9' && x < '0' + radix)
      || (radix > 10 && x >= 'a' && x < 'a' + radix - 10)
      || (radix > 10 && x >= 'A' && x < 'A' + radix - 10);
}


static double SignedZero(bool sign) {
  return sign ? -0.0 : 0.0;
}


// Returns true if 'c' is a decimal digit that is valid for the given radix.
//
// The function is small and could be inlined, but VS2012 emitted a warning
// because it constant-propagated the radix and concluded that the last
// condition was always true. By moving it into a separate function the
// compiler wouldn't warn anymore.
#ifdef _MSC_VER
#pragma optimize("",off)
static bool IsDecimalDigitForRadix(int c, int radix) {
  return '0' <= c && c <= '9' && (c - '0') < radix;
}
#pragma optimize("",on)
#else
static bool inline IsDecimalDigitForRadix(int c, int radix) {
  return '0' <= c && c <= '9' && (c - '0') < radix;
}
#endif
// Returns true if 'c' is a character digit that is valid for the given radix.
// The 'a_character' should be 'a' or 'A'.
//
// The function is small and could be inlined, but VS2012 emitted a warning
// because it constant-propagated the radix and concluded that the first
// condition was always false. By moving it into a separate function the
// compiler wouldn't warn anymore.
static bool IsCharacterDigitForRadix(int c, int radix, char a_character) {
  return radix > 10 && c >= a_character && c < a_character + radix - 10;
}

// Returns true, when the iterator is equal to end.
template<class Iterator>
static bool Advance (Iterator* it, uc16 separator, int base, Iterator& end) {
  if (separator == StringToDoubleConverter::kNoSeparator) {
    ++(*it);
    return *it == end;
  }
  if (!isDigit(**it, base)) {
    ++(*it);
    return *it == end;
  }
  ++(*it);
  if (*it == end) return true;
  if (*it + 1 == end) return false;
  if (**it == separator && isDigit(*(*it + 1), base)) {
    ++(*it);
  }
  return *it == end;
}

// Checks whether the string in the range start-end is a hex-float string.
// This function assumes that the leading '0x'/'0X' is already consumed.
//
// Hex float strings are of one of the following forms:
//   - hex_digits+ 'p' ('+'|'-')? exponent_digits+
//   - hex_digits* '.' hex_digits+ 'p' ('+'|'-')? exponent_digits+
//   - hex_digits+ '.' 'p' ('+'|'-')? exponent_digits+
template<class Iterator>
static bool IsHexFloatString(Iterator start,
                             Iterator end,
                             uc16 separator,
                             bool allow_trailing_junk) {
  DOUBLE_CONVERSION_ASSERT(start != end);

  Iterator current = start;

  bool saw_digit = false;
  while (isDigit(*current, 16)) {
    saw_digit = true;
    if (Advance(&current, separator, 16, end)) return false;
  }
  if (*current == '.') {
    if (Advance(&current, separator, 16, end)) return false;
    while (isDigit(*current, 16)) {
      saw_digit = true;
      if (Advance(&current, separator, 16, end)) return false;
    }
  }
  if (!saw_digit) return false;
  if (*current != 'p' && *current != 'P') return false;
  if (Advance(&current, separator, 16, end)) return false;
  if (*current == '+' || *current == '-') {
    if (Advance(&current, separator, 16, end)) return false;
  }
  if (!isDigit(*current, 10)) return false;
  if (Advance(&current, separator, 16, end)) return true;
  while (isDigit(*current, 10)) {
    if (Advance(&current, separator, 16, end)) return true;
  }
  return allow_trailing_junk || !AdvanceToNonspace(&current, end);
}


// Parsing integers with radix 2, 4, 8, 16, 32. Assumes current != end.
//
// If parse_as_hex_float is true, then the string must be a valid
// hex-float.
template <int radix_log_2, class Iterator>
static double RadixStringToIeee(Iterator* current,
                                Iterator end,
                                bool sign,
                                uc16 separator,
                                bool parse_as_hex_float,
                                bool allow_trailing_junk,
                                double junk_string_value,
                                bool read_as_double,
                                bool* result_is_junk) {
  DOUBLE_CONVERSION_ASSERT(*current != end);
  DOUBLE_CONVERSION_ASSERT(!parse_as_hex_float ||
      IsHexFloatString(*current, end, separator, allow_trailing_junk));

  const int kDoubleSize = Double::kSignificandSize;
  const int kSingleSize = Single::kSignificandSize;
  const int kSignificandSize = read_as_double? kDoubleSize: kSingleSize;

  *result_is_junk = true;

  int64_t number = 0;
  int exponent = 0;
  const int radix = (1 << radix_log_2);
  // Whether we have encountered a '.' and are parsing the decimal digits.
  // Only relevant if parse_as_hex_float is true.
  bool post_decimal = false;

  // Skip leading 0s.
  while (**current == '0') {
    if (Advance(current, separator, radix, end)) {
      *result_is_junk = false;
      return SignedZero(sign);
    }
  }

  while (true) {
    int digit;
    if (IsDecimalDigitForRadix(**current, radix)) {
      digit = static_cast<char>(**current) - '0';
      if (post_decimal) exponent -= radix_log_2;
    } else if (IsCharacterDigitForRadix(**current, radix, 'a')) {
      digit = static_cast<char>(**current) - 'a' + 10;
      if (post_decimal) exponent -= radix_log_2;
    } else if (IsCharacterDigitForRadix(**current, radix, 'A')) {
      digit = static_cast<char>(**current) - 'A' + 10;
      if (post_decimal) exponent -= radix_log_2;
    } else if (parse_as_hex_float && **current == '.') {
      post_decimal = true;
      Advance(current, separator, radix, end);
      DOUBLE_CONVERSION_ASSERT(*current != end);
      continue;
    } else if (parse_as_hex_float && (**current == 'p' || **current == 'P')) {
      break;
    } else {
      if (allow_trailing_junk || !AdvanceToNonspace(current, end)) {
        break;
      } else {
        return junk_string_value;
      }
    }

    number = number * radix + digit;
    int overflow = static_cast<int>(number >> kSignificandSize);
    if (overflow != 0) {
      // Overflow occurred. Need to determine which direction to round the
      // result.
      int overflow_bits_count = 1;
      while (overflow > 1) {
        overflow_bits_count++;
        overflow >>= 1;
      }

      int dropped_bits_mask = ((1 << overflow_bits_count) - 1);
      int dropped_bits = static_cast<int>(number) & dropped_bits_mask;
      number >>= overflow_bits_count;
      exponent += overflow_bits_count;

      bool zero_tail = true;
      for (;;) {
        if (Advance(current, separator, radix, end)) break;
        if (parse_as_hex_float && **current == '.') {
          // Just run over the '.'. We are just trying to see whether there is
          // a non-zero digit somewhere.
          Advance(current, separator, radix, end);
          DOUBLE_CONVERSION_ASSERT(*current != end);
          post_decimal = true;
        }
        if (!isDigit(**current, radix)) break;
        zero_tail = zero_tail && **current == '0';
        if (!post_decimal) exponent += radix_log_2;
      }

      if (!parse_as_hex_float &&
          !allow_trailing_junk &&
          AdvanceToNonspace(current, end)) {
        return junk_string_value;
      }

      int middle_value = (1 << (overflow_bits_count - 1));
      if (dropped_bits > middle_value) {
        number++;  // Rounding up.
      } else if (dropped_bits == middle_value) {
        // Rounding to even to consistency with decimals: half-way case rounds
        // up if significant part is odd and down otherwise.
        if ((number & 1) != 0 || !zero_tail) {
          number++;  // Rounding up.
        }
      }

      // Rounding up may cause overflow.
      if ((number & ((int64_t)1 << kSignificandSize)) != 0) {
        exponent++;
        number >>= 1;
      }
      break;
    }
    if (Advance(current, separator, radix, end)) break;
  }

  DOUBLE_CONVERSION_ASSERT(number < ((int64_t)1 << kSignificandSize));
  DOUBLE_CONVERSION_ASSERT(static_cast<int64_t>(static_cast<double>(number)) == number);

  *result_is_junk = false;

  if (parse_as_hex_float) {
    DOUBLE_CONVERSION_ASSERT(**current == 'p' || **current == 'P');
    Advance(current, separator, radix, end);
    DOUBLE_CONVERSION_ASSERT(*current != end);
    bool is_negative = false;
    if (**current == '+') {
      Advance(current, separator, radix, end);
      DOUBLE_CONVERSION_ASSERT(*current != end);
    } else if (**current == '-') {
      is_negative = true;
      Advance(current, separator, radix, end);
      DOUBLE_CONVERSION_ASSERT(*current != end);
    }
    int written_exponent = 0;
    while (IsDecimalDigitForRadix(**current, 10)) {
      // No need to read exponents if they are too big. That could potentially overflow
      // the `written_exponent` variable.
      if (abs(written_exponent) <= 100 * Double::kMaxExponent) {
        written_exponent = 10 * written_exponent + **current - '0';
      }
      if (Advance(current, separator, radix, end)) break;
    }
    if (is_negative) written_exponent = -written_exponent;
    exponent += written_exponent;
  }

  if (exponent == 0 || number == 0) {
    if (sign) {
      if (number == 0) return -0.0;
      number = -number;
    }
    return static_cast<double>(number);
  }

  DOUBLE_CONVERSION_ASSERT(number != 0);
  double result = Double(DiyFp(number, exponent)).value();
  return sign ? -result : result;
}

template <class Iterator>
double StringToDoubleConverter::StringToIeee(
    Iterator input,
    int length,
    bool read_as_double,
    int* processed_characters_count) const {
  Iterator current = input;
  Iterator end = input + length;

  *processed_characters_count = 0;

  const bool allow_trailing_junk = (flags_ & ALLOW_TRAILING_JUNK) != 0;
  const bool allow_leading_spaces = (flags_ & ALLOW_LEADING_SPACES) != 0;
  const bool allow_trailing_spaces = (flags_ & ALLOW_TRAILING_SPACES) != 0;
  const bool allow_spaces_after_sign = (flags_ & ALLOW_SPACES_AFTER_SIGN) != 0;
  const bool allow_case_insensitivity = (flags_ & ALLOW_CASE_INSENSITIVITY) != 0;

  // To make sure that iterator dereferencing is valid the following
  // convention is used:
  // 1. Each '++current' statement is followed by check for equality to 'end'.
  // 2. If AdvanceToNonspace returned false then current == end.
  // 3. If 'current' becomes equal to 'end' the function returns or goes to
  // 'parsing_done'.
  // 4. 'current' is not dereferenced after the 'parsing_done' label.
  // 5. Code before 'parsing_done' may rely on 'current != end'.
  if (current == end) return empty_string_value_;

  if (allow_leading_spaces || allow_trailing_spaces) {
    if (!AdvanceToNonspace(&current, end)) {
      *processed_characters_count = static_cast<int>(current - input);
      return empty_string_value_;
    }
    if (!allow_leading_spaces && (input != current)) {
      // No leading spaces allowed, but AdvanceToNonspace moved forward.
      return junk_string_value_;
    }
  }

  // The longest form of simplified number is: "-<significant digits>.1eXXX\0".
  const int kBufferSize = kMaxSignificantDigits + 10;
  char buffer[kBufferSize];  // NOLINT: size is known at compile time.
  int buffer_pos = 0;

  // Exponent will be adjusted if insignificant digits of the integer part
  // or insignificant leading zeros of the fractional part are dropped.
  int exponent = 0;
  int significant_digits = 0;
  int insignificant_digits = 0;
  bool nonzero_digit_dropped = false;

  bool sign = false;

  if (*current == '+' || *current == '-') {
    sign = (*current == '-');
    ++current;
    Iterator next_non_space = current;
    // Skip following spaces (if allowed).
    if (!AdvanceToNonspace(&next_non_space, end)) return junk_string_value_;
    if (!allow_spaces_after_sign && (current != next_non_space)) {
      return junk_string_value_;
    }
    current = next_non_space;
  }

  if (infinity_symbol_ != NULL) {
    if (ConsumeFirstCharacter(*current, infinity_symbol_, allow_case_insensitivity)) {
      if (!ConsumeSubString(&current, end, infinity_symbol_, allow_case_insensitivity)) {
        return junk_string_value_;
      }

      if (!(allow_trailing_spaces || allow_trailing_junk) && (current != end)) {
        return junk_string_value_;
      }
      if (!allow_trailing_junk && AdvanceToNonspace(&current, end)) {
        return junk_string_value_;
      }

      DOUBLE_CONVERSION_ASSERT(buffer_pos == 0);
      *processed_characters_count = static_cast<int>(current - input);
      return sign ? -Double::Infinity() : Double::Infinity();
    }
  }

  if (nan_symbol_ != NULL) {
    if (ConsumeFirstCharacter(*current, nan_symbol_, allow_case_insensitivity)) {
      if (!ConsumeSubString(&current, end, nan_symbol_, allow_case_insensitivity)) {
        return junk_string_value_;
      }

      if (!(allow_trailing_spaces || allow_trailing_junk) && (current != end)) {
        return junk_string_value_;
      }
      if (!allow_trailing_junk && AdvanceToNonspace(&current, end)) {
        return junk_string_value_;
      }

      DOUBLE_CONVERSION_ASSERT(buffer_pos == 0);
      *processed_characters_count = static_cast<int>(current - input);
      return sign ? -Double::NaN() : Double::NaN();
    }
  }

  bool leading_zero = false;
  if (*current == '0') {
    if (Advance(&current, separator_, 10, end)) {
      *processed_characters_count = static_cast<int>(current - input);
      return SignedZero(sign);
    }

    leading_zero = true;

    // It could be hexadecimal value.
    if (((flags_ & ALLOW_HEX) || (flags_ & ALLOW_HEX_FLOATS)) &&
        (*current == 'x' || *current == 'X')) {
      ++current;

      if (current == end) return junk_string_value_;  // "0x"

      bool parse_as_hex_float = (flags_ & ALLOW_HEX_FLOATS) &&
                IsHexFloatString(current, end, separator_, allow_trailing_junk);

      if (!parse_as_hex_float && !isDigit(*current, 16)) {
        return junk_string_value_;
      }

      bool result_is_junk;
      double result = RadixStringToIeee<4>(&current,
                                           end,
                                           sign,
                                           separator_,
                                           parse_as_hex_float,
                                           allow_trailing_junk,
                                           junk_string_value_,
                                           read_as_double,
                                           &result_is_junk);
      if (!result_is_junk) {
        if (allow_trailing_spaces) AdvanceToNonspace(&current, end);
        *processed_characters_count = static_cast<int>(current - input);
      }
      return result;
    }

    // Ignore leading zeros in the integer part.
    while (*current == '0') {
      if (Advance(&current, separator_, 10, end)) {
        *processed_characters_count = static_cast<int>(current - input);
        return SignedZero(sign);
      }
    }
  }

  bool octal = leading_zero && (flags_ & ALLOW_OCTALS) != 0;

  // Copy significant digits of the integer part (if any) to the buffer.
  while (*current >= '0' && *current <= '9') {
    if (significant_digits < kMaxSignificantDigits) {
      DOUBLE_CONVERSION_ASSERT(buffer_pos < kBufferSize);
      buffer[buffer_pos++] = static_cast<char>(*current);
      significant_digits++;
      // Will later check if it's an octal in the buffer.
    } else {
      insignificant_digits++;  // Move the digit into the exponential part.
      nonzero_digit_dropped = nonzero_digit_dropped || *current != '0';
    }
    octal = octal && *current < '8';
    if (Advance(&current, separator_, 10, end)) goto parsing_done;
  }

  if (significant_digits == 0) {
    octal = false;
  }

  if (*current == '.') {
    if (octal && !allow_trailing_junk) return junk_string_value_;
    if (octal) goto parsing_done;

    if (Advance(&current, separator_, 10, end)) {
      if (significant_digits == 0 && !leading_zero) {
        return junk_string_value_;
      } else {
        goto parsing_done;
      }
    }

    if (significant_digits == 0) {
      // octal = false;
      // Integer part consists of 0 or is absent. Significant digits start after
      // leading zeros (if any).
      while (*current == '0') {
        if (Advance(&current, separator_, 10, end)) {
          *processed_characters_count = static_cast<int>(current - input);
          return SignedZero(sign);
        }
        exponent--;  // Move this 0 into the exponent.
      }
    }

    // There is a fractional part.
    // We don't emit a '.', but adjust the exponent instead.
    while (*current >= '0' && *current <= '9') {
      if (significant_digits < kMaxSignificantDigits) {
        DOUBLE_CONVERSION_ASSERT(buffer_pos < kBufferSize);
        buffer[buffer_pos++] = static_cast<char>(*current);
        significant_digits++;
        exponent--;
      } else {
        // Ignore insignificant digits in the fractional part.
        nonzero_digit_dropped = nonzero_digit_dropped || *current != '0';
      }
      if (Advance(&current, separator_, 10, end)) goto parsing_done;
    }
  }

  if (!leading_zero && exponent == 0 && significant_digits == 0) {
    // If leading_zeros is true then the string contains zeros.
    // If exponent < 0 then string was [+-]\.0*...
    // If significant_digits != 0 the string is not equal to 0.
    // Otherwise there are no digits in the string.
    return junk_string_value_;
  }

  // Parse exponential part.
  if (*current == 'e' || *current == 'E') {
    if (octal && !allow_trailing_junk) return junk_string_value_;
    if (octal) goto parsing_done;
    Iterator junk_begin = current;
    ++current;
    if (current == end) {
      if (allow_trailing_junk) {
        current = junk_begin;
        goto parsing_done;
      } else {
        return junk_string_value_;
      }
    }
    char exponen_sign = '+';
    if (*current == '+' || *current == '-') {
      exponen_sign = static_cast<char>(*current);
      ++current;
      if (current == end) {
        if (allow_trailing_junk) {
          current = junk_begin;
          goto parsing_done;
        } else {
          return junk_string_value_;
        }
      }
    }

    if (current == end || *current < '0' || *current > '9') {
      if (allow_trailing_junk) {
        current = junk_begin;
        goto parsing_done;
      } else {
        return junk_string_value_;
      }
    }

    const int max_exponent = INT_MAX / 2;
    DOUBLE_CONVERSION_ASSERT(-max_exponent / 2 <= exponent && exponent <= max_exponent / 2);
    int num = 0;
    do {
      // Check overflow.
      int digit = *current - '0';
      if (num >= max_exponent / 10
          && !(num == max_exponent / 10 && digit <= max_exponent % 10)) {
        num = max_exponent;
      } else {
        num = num * 10 + digit;
      }
      ++current;
    } while (current != end && *current >= '0' && *current <= '9');

    exponent += (exponen_sign == '-' ? -num : num);
  }

  if (!(allow_trailing_spaces || allow_trailing_junk) && (current != end)) {
    return junk_string_value_;
  }
  if (!allow_trailing_junk && AdvanceToNonspace(&current, end)) {
    return junk_string_value_;
  }
  if (allow_trailing_spaces) {
    AdvanceToNonspace(&current, end);
  }

  parsing_done:
  exponent += insignificant_digits;

  if (octal) {
    double result;
    bool result_is_junk;
    char* start = buffer;
    result = RadixStringToIeee<3>(&start,
                                  buffer + buffer_pos,
                                  sign,
                                  separator_,
                                  false, // Don't parse as hex_float.
                                  allow_trailing_junk,
                                  junk_string_value_,
                                  read_as_double,
                                  &result_is_junk);
    DOUBLE_CONVERSION_ASSERT(!result_is_junk);
    *processed_characters_count = static_cast<int>(current - input);
    return result;
  }

  if (nonzero_digit_dropped) {
    buffer[buffer_pos++] = '1';
    exponent--;
  }

  DOUBLE_CONVERSION_ASSERT(buffer_pos < kBufferSize);
  buffer[buffer_pos] = '\0';

  double converted;
  if (read_as_double) {
    converted = Strtod(Vector<const char>(buffer, buffer_pos), exponent);
  } else {
    converted = Strtof(Vector<const char>(buffer, buffer_pos), exponent);
  }
  *processed_characters_count = static_cast<int>(current - input);
  return sign? -converted: converted;
}


double StringToDoubleConverter::StringToDouble(
    const char* buffer,
    int length,
    int* processed_characters_count) const {
  return StringToIeee(buffer, length, true, processed_characters_count);
}


double StringToDoubleConverter::StringToDouble(
    const uc16* buffer,
    int length,
    int* processed_characters_count) const {
  return StringToIeee(buffer, length, true, processed_characters_count);
}


float StringToDoubleConverter::StringToFloat(
    const char* buffer,
    int length,
    int* processed_characters_count) const {
  return static_cast<float>(StringToIeee(buffer, length, false,
                                         processed_characters_count));
}


float StringToDoubleConverter::StringToFloat(
    const uc16* buffer,
    int length,
    int* processed_characters_count) const {
  return static_cast<float>(StringToIeee(buffer, length, false,
                                         processed_characters_count));
}

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END
#endif // ICU PATCH: close #if !UCONFIG_NO_FORMATTING
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING

#include <climits>
#include <cstdarg>

// ICU PATCH: Customize header file paths for ICU.






// ICU PATCH: Wrap in ICU namespace
U_NAMESPACE_BEGIN

namespace double_conversion {

// 2^53 = 9007199254740992.
// Any integer with at most 15 decimal digits will hence fit into a double
// (which has a 53bit significand) without loss of precision.
static const int kMaxExactDoubleIntegerDecimalDigits = 15;
// 2^64 = 18446744073709551616 > 10^19
static const int kMaxUint64DecimalDigits = 19;

// Max double: 1.7976931348623157 x 10^308
// Min non-zero double: 4.9406564584124654 x 10^-324
// Any x >= 10^309 is interpreted as +infinity.
// Any x <= 10^-324 is interpreted as 0.
// Note that 2.5e-324 (despite being smaller than the min double) will be read
// as non-zero (equal to the min non-zero double).
static const int kMaxDecimalPower = 309;
static const int kMinDecimalPower = -324;

// 2^64 = 18446744073709551616
static const uint64_t kMaxUint64 = DOUBLE_CONVERSION_UINT64_2PART_C(0xFFFFFFFF, FFFFFFFF);


static const double exact_powers_of_ten[] = {
  1.0,  // 10^0
  10.0,
  100.0,
  1000.0,
  10000.0,
  100000.0,
  1000000.0,
  10000000.0,
  100000000.0,
  1000000000.0,
  10000000000.0,  // 10^10
  100000000000.0,
  1000000000000.0,
  10000000000000.0,
  100000000000000.0,
  1000000000000000.0,
  10000000000000000.0,
  100000000000000000.0,
  1000000000000000000.0,
  10000000000000000000.0,
  100000000000000000000.0,  // 10^20
  1000000000000000000000.0,
  // 10^22 = 0x21e19e0c9bab2400000 = 0x878678326eac9 * 2^22
  10000000000000000000000.0
};
static const int kExactPowersOfTenSize = DOUBLE_CONVERSION_ARRAY_SIZE(exact_powers_of_ten);

// Maximum number of significant digits in the decimal representation.
// In fact the value is 772 (see conversions.cc), but to give us some margin
// we round up to 780.
static const int kMaxSignificantDecimalDigits = 780;

static Vector<const char> TrimLeadingZeros(Vector<const char> buffer) {
  for (int i = 0; i < buffer.length(); i++) {
    if (buffer[i] != '0') {
      return buffer.SubVector(i, buffer.length());
    }
  }
  return Vector<const char>(buffer.start(), 0);
}


static Vector<const char> TrimTrailingZeros(Vector<const char> buffer) {
  for (int i = buffer.length() - 1; i >= 0; --i) {
    if (buffer[i] != '0') {
      return buffer.SubVector(0, i + 1);
    }
  }
  return Vector<const char>(buffer.start(), 0);
}


static void CutToMaxSignificantDigits(Vector<const char> buffer,
                                       int exponent,
                                       char* significant_buffer,
                                       int* significant_exponent) {
  for (int i = 0; i < kMaxSignificantDecimalDigits - 1; ++i) {
    significant_buffer[i] = buffer[i];
  }
  // The input buffer has been trimmed. Therefore the last digit must be
  // different from '0'.
  DOUBLE_CONVERSION_ASSERT(buffer[buffer.length() - 1] != '0');
  // Set the last digit to be non-zero. This is sufficient to guarantee
  // correct rounding.
  significant_buffer[kMaxSignificantDecimalDigits - 1] = '1';
  *significant_exponent =
      exponent + (buffer.length() - kMaxSignificantDecimalDigits);
}


// Trims the buffer and cuts it to at most kMaxSignificantDecimalDigits.
// If possible the input-buffer is reused, but if the buffer needs to be
// modified (due to cutting), then the input needs to be copied into the
// buffer_copy_space.
static void TrimAndCut(Vector<const char> buffer, int exponent,
                       char* buffer_copy_space, int space_size,
                       Vector<const char>* trimmed, int* updated_exponent) {
  Vector<const char> left_trimmed = TrimLeadingZeros(buffer);
  Vector<const char> right_trimmed = TrimTrailingZeros(left_trimmed);
  exponent += left_trimmed.length() - right_trimmed.length();
  if (right_trimmed.length() > kMaxSignificantDecimalDigits) {
    (void) space_size;  // Mark variable as used.
    DOUBLE_CONVERSION_ASSERT(space_size >= kMaxSignificantDecimalDigits);
    CutToMaxSignificantDigits(right_trimmed, exponent,
                              buffer_copy_space, updated_exponent);
    *trimmed = Vector<const char>(buffer_copy_space,
                                 kMaxSignificantDecimalDigits);
  } else {
    *trimmed = right_trimmed;
    *updated_exponent = exponent;
  }
}


// Reads digits from the buffer and converts them to a uint64.
// Reads in as many digits as fit into a uint64.
// When the string starts with "1844674407370955161" no further digit is read.
// Since 2^64 = 18446744073709551616 it would still be possible read another
// digit if it was less or equal than 6, but this would complicate the code.
static uint64_t ReadUint64(Vector<const char> buffer,
                           int* number_of_read_digits) {
  uint64_t result = 0;
  int i = 0;
  while (i < buffer.length() && result <= (kMaxUint64 / 10 - 1)) {
    int digit = buffer[i++] - '0';
    DOUBLE_CONVERSION_ASSERT(0 <= digit && digit <= 9);
    result = 10 * result + digit;
  }
  *number_of_read_digits = i;
  return result;
}


// Reads a DiyFp from the buffer.
// The returned DiyFp is not necessarily normalized.
// If remaining_decimals is zero then the returned DiyFp is accurate.
// Otherwise it has been rounded and has error of at most 1/2 ulp.
static void ReadDiyFp(Vector<const char> buffer,
                      DiyFp* result,
                      int* remaining_decimals) {
  int read_digits;
  uint64_t significand = ReadUint64(buffer, &read_digits);
  if (buffer.length() == read_digits) {
    *result = DiyFp(significand, 0);
    *remaining_decimals = 0;
  } else {
    // Round the significand.
    if (buffer[read_digits] >= '5') {
      significand++;
    }
    // Compute the binary exponent.
    int exponent = 0;
    *result = DiyFp(significand, exponent);
    *remaining_decimals = buffer.length() - read_digits;
  }
}


static bool DoubleStrtod(Vector<const char> trimmed,
                         int exponent,
                         double* result) {
#if !defined(DOUBLE_CONVERSION_CORRECT_DOUBLE_OPERATIONS)
  // On x86 the floating-point stack can be 64 or 80 bits wide. If it is
  // 80 bits wide (as is the case on Linux) then double-rounding occurs and the
  // result is not accurate.
  // We know that Windows32 uses 64 bits and is therefore accurate.
  // Note that the ARM simulator is compiled for 32bits. It therefore exhibits
  // the same problem.
  return false;
#else
  if (trimmed.length() <= kMaxExactDoubleIntegerDecimalDigits) {
    int read_digits;
    // The trimmed input fits into a double.
    // If the 10^exponent (resp. 10^-exponent) fits into a double too then we
    // can compute the result-double simply by multiplying (resp. dividing) the
    // two numbers.
    // This is possible because IEEE guarantees that floating-point operations
    // return the best possible approximation.
    if (exponent < 0 && -exponent < kExactPowersOfTenSize) {
      // 10^-exponent fits into a double.
      *result = static_cast<double>(ReadUint64(trimmed, &read_digits));
      DOUBLE_CONVERSION_ASSERT(read_digits == trimmed.length());
      *result /= exact_powers_of_ten[-exponent];
      return true;
    }
    if (0 <= exponent && exponent < kExactPowersOfTenSize) {
      // 10^exponent fits into a double.
      *result = static_cast<double>(ReadUint64(trimmed, &read_digits));
      DOUBLE_CONVERSION_ASSERT(read_digits == trimmed.length());
      *result *= exact_powers_of_ten[exponent];
      return true;
    }
    int remaining_digits =
        kMaxExactDoubleIntegerDecimalDigits - trimmed.length();
    if ((0 <= exponent) &&
        (exponent - remaining_digits < kExactPowersOfTenSize)) {
      // The trimmed string was short and we can multiply it with
      // 10^remaining_digits. As a result the remaining exponent now fits
      // into a double too.
      *result = static_cast<double>(ReadUint64(trimmed, &read_digits));
      DOUBLE_CONVERSION_ASSERT(read_digits == trimmed.length());
      *result *= exact_powers_of_ten[remaining_digits];
      *result *= exact_powers_of_ten[exponent - remaining_digits];
      return true;
    }
  }
  return false;
#endif
}


// Returns 10^exponent as an exact DiyFp.
// The given exponent must be in the range [1; kDecimalExponentDistance[.
static DiyFp AdjustmentPowerOfTen(int exponent) {
  DOUBLE_CONVERSION_ASSERT(0 < exponent);
  DOUBLE_CONVERSION_ASSERT(exponent < PowersOfTenCache::kDecimalExponentDistance);
  // Simply hardcode the remaining powers for the given decimal exponent
  // distance.
  DOUBLE_CONVERSION_ASSERT(PowersOfTenCache::kDecimalExponentDistance == 8);
  switch (exponent) {
    case 1: return DiyFp(DOUBLE_CONVERSION_UINT64_2PART_C(0xa0000000, 00000000), -60);
    case 2: return DiyFp(DOUBLE_CONVERSION_UINT64_2PART_C(0xc8000000, 00000000), -57);
    case 3: return DiyFp(DOUBLE_CONVERSION_UINT64_2PART_C(0xfa000000, 00000000), -54);
    case 4: return DiyFp(DOUBLE_CONVERSION_UINT64_2PART_C(0x9c400000, 00000000), -50);
    case 5: return DiyFp(DOUBLE_CONVERSION_UINT64_2PART_C(0xc3500000, 00000000), -47);
    case 6: return DiyFp(DOUBLE_CONVERSION_UINT64_2PART_C(0xf4240000, 00000000), -44);
    case 7: return DiyFp(DOUBLE_CONVERSION_UINT64_2PART_C(0x98968000, 00000000), -40);
    default:
      DOUBLE_CONVERSION_UNREACHABLE();
  }
}


// If the function returns true then the result is the correct double.
// Otherwise it is either the correct double or the double that is just below
// the correct double.
static bool DiyFpStrtod(Vector<const char> buffer,
                        int exponent,
                        double* result) {
  DiyFp input;
  int remaining_decimals;
  ReadDiyFp(buffer, &input, &remaining_decimals);
  // Since we may have dropped some digits the input is not accurate.
  // If remaining_decimals is different than 0 than the error is at most
  // .5 ulp (unit in the last place).
  // We don't want to deal with fractions and therefore keep a common
  // denominator.
  const int kDenominatorLog = 3;
  const int kDenominator = 1 << kDenominatorLog;
  // Move the remaining decimals into the exponent.
  exponent += remaining_decimals;
  uint64_t error = (remaining_decimals == 0 ? 0 : kDenominator / 2);

  int old_e = input.e();
  input.Normalize();
  error <<= old_e - input.e();

  DOUBLE_CONVERSION_ASSERT(exponent <= PowersOfTenCache::kMaxDecimalExponent);
  if (exponent < PowersOfTenCache::kMinDecimalExponent) {
    *result = 0.0;
    return true;
  }
  DiyFp cached_power;
  int cached_decimal_exponent;
  PowersOfTenCache::GetCachedPowerForDecimalExponent(exponent,
                                                     &cached_power,
                                                     &cached_decimal_exponent);

  if (cached_decimal_exponent != exponent) {
    int adjustment_exponent = exponent - cached_decimal_exponent;
    DiyFp adjustment_power = AdjustmentPowerOfTen(adjustment_exponent);
    input.Multiply(adjustment_power);
    if (kMaxUint64DecimalDigits - buffer.length() >= adjustment_exponent) {
      // The product of input with the adjustment power fits into a 64 bit
      // integer.
      DOUBLE_CONVERSION_ASSERT(DiyFp::kSignificandSize == 64);
    } else {
      // The adjustment power is exact. There is hence only an error of 0.5.
      error += kDenominator / 2;
    }
  }

  input.Multiply(cached_power);
  // The error introduced by a multiplication of a*b equals
  //   error_a + error_b + error_a*error_b/2^64 + 0.5
  // Substituting a with 'input' and b with 'cached_power' we have
  //   error_b = 0.5  (all cached powers have an error of less than 0.5 ulp),
  //   error_ab = 0 or 1 / kDenominator > error_a*error_b/ 2^64
  int error_b = kDenominator / 2;
  int error_ab = (error == 0 ? 0 : 1);  // We round up to 1.
  int fixed_error = kDenominator / 2;
  error += error_b + error_ab + fixed_error;

  old_e = input.e();
  input.Normalize();
  error <<= old_e - input.e();

  // See if the double's significand changes if we add/subtract the error.
  int order_of_magnitude = DiyFp::kSignificandSize + input.e();
  int effective_significand_size =
      Double::SignificandSizeForOrderOfMagnitude(order_of_magnitude);
  int precision_digits_count =
      DiyFp::kSignificandSize - effective_significand_size;
  if (precision_digits_count + kDenominatorLog >= DiyFp::kSignificandSize) {
    // This can only happen for very small denormals. In this case the
    // half-way multiplied by the denominator exceeds the range of an uint64.
    // Simply shift everything to the right.
    int shift_amount = (precision_digits_count + kDenominatorLog) -
        DiyFp::kSignificandSize + 1;
    input.set_f(input.f() >> shift_amount);
    input.set_e(input.e() + shift_amount);
    // We add 1 for the lost precision of error, and kDenominator for
    // the lost precision of input.f().
    error = (error >> shift_amount) + 1 + kDenominator;
    precision_digits_count -= shift_amount;
  }
  // We use uint64_ts now. This only works if the DiyFp uses uint64_ts too.
  DOUBLE_CONVERSION_ASSERT(DiyFp::kSignificandSize == 64);
  DOUBLE_CONVERSION_ASSERT(precision_digits_count < 64);
  uint64_t one64 = 1;
  uint64_t precision_bits_mask = (one64 << precision_digits_count) - 1;
  uint64_t precision_bits = input.f() & precision_bits_mask;
  uint64_t half_way = one64 << (precision_digits_count - 1);
  precision_bits *= kDenominator;
  half_way *= kDenominator;
  DiyFp rounded_input(input.f() >> precision_digits_count,
                      input.e() + precision_digits_count);
  if (precision_bits >= half_way + error) {
    rounded_input.set_f(rounded_input.f() + 1);
  }
  // If the last_bits are too close to the half-way case than we are too
  // inaccurate and round down. In this case we return false so that we can
  // fall back to a more precise algorithm.

  *result = Double(rounded_input).value();
  if (half_way - error < precision_bits && precision_bits < half_way + error) {
    // Too imprecise. The caller will have to fall back to a slower version.
    // However the returned number is guaranteed to be either the correct
    // double, or the next-lower double.
    return false;
  } else {
    return true;
  }
}


// Returns
//   - -1 if buffer*10^exponent < diy_fp.
//   -  0 if buffer*10^exponent == diy_fp.
//   - +1 if buffer*10^exponent > diy_fp.
// Preconditions:
//   buffer.length() + exponent <= kMaxDecimalPower + 1
//   buffer.length() + exponent > kMinDecimalPower
//   buffer.length() <= kMaxDecimalSignificantDigits
static int CompareBufferWithDiyFp(Vector<const char> buffer,
                                  int exponent,
                                  DiyFp diy_fp) {
  DOUBLE_CONVERSION_ASSERT(buffer.length() + exponent <= kMaxDecimalPower + 1);
  DOUBLE_CONVERSION_ASSERT(buffer.length() + exponent > kMinDecimalPower);
  DOUBLE_CONVERSION_ASSERT(buffer.length() <= kMaxSignificantDecimalDigits);
  // Make sure that the Bignum will be able to hold all our numbers.
  // Our Bignum implementation has a separate field for exponents. Shifts will
  // consume at most one bigit (< 64 bits).
  // ln(10) == 3.3219...
  DOUBLE_CONVERSION_ASSERT(((kMaxDecimalPower + 1) * 333 / 100) < Bignum::kMaxSignificantBits);
  Bignum buffer_bignum;
  Bignum diy_fp_bignum;
  buffer_bignum.AssignDecimalString(buffer);
  diy_fp_bignum.AssignUInt64(diy_fp.f());
  if (exponent >= 0) {
    buffer_bignum.MultiplyByPowerOfTen(exponent);
  } else {
    diy_fp_bignum.MultiplyByPowerOfTen(-exponent);
  }
  if (diy_fp.e() > 0) {
    diy_fp_bignum.ShiftLeft(diy_fp.e());
  } else {
    buffer_bignum.ShiftLeft(-diy_fp.e());
  }
  return Bignum::Compare(buffer_bignum, diy_fp_bignum);
}


// Returns true if the guess is the correct double.
// Returns false, when guess is either correct or the next-lower double.
static bool ComputeGuess(Vector<const char> trimmed, int exponent,
                         double* guess) {
  if (trimmed.length() == 0) {
    *guess = 0.0;
    return true;
  }
  if (exponent + trimmed.length() - 1 >= kMaxDecimalPower) {
    *guess = Double::Infinity();
    return true;
  }
  if (exponent + trimmed.length() <= kMinDecimalPower) {
    *guess = 0.0;
    return true;
  }

  if (DoubleStrtod(trimmed, exponent, guess) ||
      DiyFpStrtod(trimmed, exponent, guess)) {
    return true;
  }
  if (*guess == Double::Infinity()) {
    return true;
  }
  return false;
}

#if U_DEBUG // needed for ICU only in debug mode
static bool IsDigit(const char d) {
  return ('0' <= d) && (d <= '9');
}

static bool IsNonZeroDigit(const char d) {
  return ('1' <= d) && (d <= '9');
}

static bool AssertTrimmedDigits(const Vector<const char>& buffer) {
  for(int i = 0; i < buffer.length(); ++i) {
    if(!IsDigit(buffer[i])) {
      return false;
    }
  }
  return (buffer.length() == 0) || (IsNonZeroDigit(buffer[0]) && IsNonZeroDigit(buffer[buffer.length()-1]));
}
#endif // needed for ICU only in debug mode

double StrtodTrimmed(Vector<const char> trimmed, int exponent) {
  DOUBLE_CONVERSION_ASSERT(trimmed.length() <= kMaxSignificantDecimalDigits);
  DOUBLE_CONVERSION_ASSERT(AssertTrimmedDigits(trimmed));
  double guess;
  const bool is_correct = ComputeGuess(trimmed, exponent, &guess);
  if (is_correct) {
    return guess;
  }
  DiyFp upper_boundary = Double(guess).UpperBoundary();
  int comparison = CompareBufferWithDiyFp(trimmed, exponent, upper_boundary);
  if (comparison < 0) {
    return guess;
  } else if (comparison > 0) {
    return Double(guess).NextDouble();
  } else if ((Double(guess).Significand() & 1) == 0) {
    // Round towards even.
    return guess;
  } else {
    return Double(guess).NextDouble();
  }
}

double Strtod(Vector<const char> buffer, int exponent) {
  char copy_buffer[kMaxSignificantDecimalDigits];
  Vector<const char> trimmed;
  int updated_exponent;
  TrimAndCut(buffer, exponent, copy_buffer, kMaxSignificantDecimalDigits,
             &trimmed, &updated_exponent);
  return StrtodTrimmed(trimmed, updated_exponent);
}

static float SanitizedDoubletof(double d) {
  DOUBLE_CONVERSION_ASSERT(d >= 0.0);
  // ASAN has a sanitize check that disallows casting doubles to floats if
  // they are too big.
  // https://clang.llvm.org/docs/UndefinedBehaviorSanitizer.html#available-checks
  // The behavior should be covered by IEEE 754, but some projects use this
  // flag, so work around it.
  float max_finite = 3.4028234663852885981170418348451692544e+38;
  // The half-way point between the max-finite and infinity value.
  // Since infinity has an even significand everything equal or greater than
  // this value should become infinity.
  double half_max_finite_infinity =
      3.40282356779733661637539395458142568448e+38;
  if (d >= max_finite) {
    if (d >= half_max_finite_infinity) {
      return Single::Infinity();
    } else {
      return max_finite;
    }
  } else {
    return static_cast<float>(d);
  }
}

float Strtof(Vector<const char> buffer, int exponent) {
  char copy_buffer[kMaxSignificantDecimalDigits];
  Vector<const char> trimmed;
  int updated_exponent;
  TrimAndCut(buffer, exponent, copy_buffer, kMaxSignificantDecimalDigits,
             &trimmed, &updated_exponent);
  exponent = updated_exponent;

  double double_guess;
  bool is_correct = ComputeGuess(trimmed, exponent, &double_guess);

  float float_guess = SanitizedDoubletof(double_guess);
  if (float_guess == double_guess) {
    // This shortcut triggers for integer values.
    return float_guess;
  }

  // We must catch double-rounding. Say the double has been rounded up, and is
  // now a boundary of a float, and rounds up again. This is why we have to
  // look at previous too.
  // Example (in decimal numbers):
  //    input: 12349
  //    high-precision (4 digits): 1235
  //    low-precision (3 digits):
  //       when read from input: 123
  //       when rounded from high precision: 124.
  // To do this we simply look at the neigbors of the correct result and see
  // if they would round to the same float. If the guess is not correct we have
  // to look at four values (since two different doubles could be the correct
  // double).

  double double_next = Double(double_guess).NextDouble();
  double double_previous = Double(double_guess).PreviousDouble();

  float f1 = SanitizedDoubletof(double_previous);
  float f2 = float_guess;
  float f3 = SanitizedDoubletof(double_next);
  float f4;
  if (is_correct) {
    f4 = f3;
  } else {
    double double_next2 = Double(double_next).NextDouble();
    f4 = SanitizedDoubletof(double_next2);
  }
  (void) f2;  // Mark variable as used.
  DOUBLE_CONVERSION_ASSERT(f1 <= f2 && f2 <= f3 && f3 <= f4);

  // If the guess doesn't lie near a single-precision boundary we can simply
  // return its float-value.
  if (f1 == f4) {
    return float_guess;
  }

  DOUBLE_CONVERSION_ASSERT((f1 != f2 && f2 == f3 && f3 == f4) ||
         (f1 == f2 && f2 != f3 && f3 == f4) ||
         (f1 == f2 && f2 == f3 && f3 != f4));

  // guess and next are the two possible candidates (in the same way that
  // double_guess was the lower candidate for a double-precision guess).
  float guess = f1;
  float next = f4;
  DiyFp upper_boundary;
  if (guess == 0.0f) {
    float min_float = 1e-45f;
    upper_boundary = Double(static_cast<double>(min_float) / 2).AsDiyFp();
  } else {
    upper_boundary = Single(guess).UpperBoundary();
  }
  int comparison = CompareBufferWithDiyFp(trimmed, exponent, upper_boundary);
  if (comparison < 0) {
    return guess;
  } else if (comparison > 0) {
    return next;
  } else if ((Single(guess).Significand() & 1) == 0) {
    // Round towards even.
    return guess;
  } else {
    return next;
  }
}

}  // namespace double_conversion

// ICU PATCH: Close ICU namespace
U_NAMESPACE_END
#endif // ICU PATCH: close #if !UCONFIG_NO_FORMATTING
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2016, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* File DTFMTSYM.CPP
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   07/21/98    stephen     Added getZoneIndex
*                            Changed weekdays/short weekdays to be one-based
*   06/14/99    stephen     Removed SimpleDateFormat::fgTimeZoneDataSuffix
*   11/16/99    weiv        Added 'Y' and 'e' to fgPatternChars
*   03/27/00    weiv        Keeping resource bundle around!
*   06/30/05    emmons      Added eraNames, narrow month/day, standalone context
*   10/12/05    emmons      Added setters for eraNames, month/day by width/context
*******************************************************************************
*/

#include <utility>



#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File DTFMTSYM.H
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*    07/21/98    stephen        Added getZoneIndex()
*                            Changed to match C++ conventions
********************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING








/**
 * \file
 * \brief C++ API: Symbols for formatting dates.
 */

U_NAMESPACE_BEGIN

/* forward declaration */
class SimpleDateFormat;
class Hashtable;

/**
 * DateFormatSymbols is a public class for encapsulating localizable date-time
 * formatting data -- including timezone data. DateFormatSymbols is used by
 * DateFormat and SimpleDateFormat.
 * <P>
 * Rather than first creating a DateFormatSymbols to get a date-time formatter
 * by using a SimpleDateFormat constructor, clients are encouraged to create a
 * date-time formatter using the getTimeInstance(), getDateInstance(), or
 * getDateTimeInstance() method in DateFormat. Each of these methods can return a
 * date/time formatter initialized with a default format pattern along with the
 * date-time formatting data for a given or default locale. After a formatter is
 * created, clients may modify the format pattern using the setPattern function
 * as so desired. For more information on using these formatter factory
 * functions, see DateFormat.
 * <P>
 * If clients decide to create a date-time formatter with a particular format
 * pattern and locale, they can do so with new SimpleDateFormat(aPattern,
 * new DateFormatSymbols(aLocale)).  This will load the appropriate date-time
 * formatting data from the locale.
 * <P>
 * DateFormatSymbols objects are clonable. When clients obtain a
 * DateFormatSymbols object, they can feel free to modify the date-time
 * formatting data as necessary. For instance, clients can
 * replace the localized date-time format pattern characters with the ones that
 * they feel easy to remember. Or they can change the representative cities
 * originally picked by default to using their favorite ones.
 * <P>
 * DateFormatSymbols are not expected to be subclassed. Data for a calendar is
 * loaded out of resource bundles.  The 'type' parameter indicates the type of
 * calendar, for example, "gregorian" or "japanese".  If the type is not gregorian
 * (or NULL, or an empty string) then the type is appended to the resource name,
 * for example,  'Eras_japanese' instead of 'Eras'.   If the resource 'Eras_japanese' did
 * not exist (even in root), then this class will fall back to just 'Eras', that is,
 * Gregorian data.  Therefore, the calendar implementor MUST ensure that the root
 * locale at least contains any resources that are to be particularized for the
 * calendar type.
 */
class U_I18N_API DateFormatSymbols U_FINAL : public UObject  {
public:
    /**
     * Construct a DateFormatSymbols object by loading format data from
     * resources for the default locale, in the default calendar (Gregorian).
     * <P>
     * NOTE: This constructor will never fail; if it cannot get resource
     * data for the default locale, it will return a last-resort object
     * based on hard-coded strings.
     *
     * @param status    Status code.  Failure
     *                  results if the resources for the default cannot be
     *                  found or cannot be loaded
     * @stable ICU 2.0
     */
    DateFormatSymbols(UErrorCode& status);

    /**
     * Construct a DateFormatSymbols object by loading format data from
     * resources for the given locale, in the default calendar (Gregorian).
     *
     * @param locale    Locale to load format data from.
     * @param status    Status code.  Failure
     *                  results if the resources for the locale cannot be
     *                  found or cannot be loaded
     * @stable ICU 2.0
     */
    DateFormatSymbols(const Locale& locale,
                      UErrorCode& status);

#ifndef U_HIDE_INTERNAL_API
    /**
     * Construct a DateFormatSymbols object by loading format data from
     * resources for the default locale, in the default calendar (Gregorian).
     * <P>
     * NOTE: This constructor will never fail; if it cannot get resource
     * data for the default locale, it will return a last-resort object
     * based on hard-coded strings.
     *
     * @param type      Type of calendar (as returned by Calendar::getType).
     *                  Will be used to access the correct set of strings.
     *                  (NULL or empty string defaults to "gregorian".)
     * @param status    Status code.  Failure
     *                  results if the resources for the default cannot be
     *                  found or cannot be loaded
     * @internal
     */
    DateFormatSymbols(const char *type, UErrorCode& status);

    /**
     * Construct a DateFormatSymbols object by loading format data from
     * resources for the given locale, in the default calendar (Gregorian).
     *
     * @param locale    Locale to load format data from.
     * @param type      Type of calendar (as returned by Calendar::getType).
     *                  Will be used to access the correct set of strings.
     *                  (NULL or empty string defaults to "gregorian".)
     * @param status    Status code.  Failure
     *                  results if the resources for the locale cannot be
     *                  found or cannot be loaded
     * @internal
     */
    DateFormatSymbols(const Locale& locale,
                      const char *type,
                      UErrorCode& status);
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Copy constructor.
     * @stable ICU 2.0
     */
    DateFormatSymbols(const DateFormatSymbols&);

    /**
     * Assignment operator.
     * @stable ICU 2.0
     */
    DateFormatSymbols& operator=(const DateFormatSymbols&);

    /**
     * Destructor. This is nonvirtual because this class is not designed to be
     * subclassed.
     * @stable ICU 2.0
     */
    virtual ~DateFormatSymbols();

    /**
     * Return true if another object is semantically equal to this one.
     *
     * @param other    the DateFormatSymbols object to be compared with.
     * @return         true if other is semantically equal to this.
     * @stable ICU 2.0
     */
    UBool operator==(const DateFormatSymbols& other) const;

    /**
     * Return true if another object is semantically unequal to this one.
     *
     * @param other    the DateFormatSymbols object to be compared with.
     * @return         true if other is semantically unequal to this.
     * @stable ICU 2.0
     */
    UBool operator!=(const DateFormatSymbols& other) const { return !operator==(other); }

    /**
     * Gets abbreviated era strings. For example: "AD" and "BC".
     *
     * @param count    Filled in with length of the array.
     * @return         the era strings.
     * @stable ICU 2.0
     */
    const UnicodeString* getEras(int32_t& count) const;

    /**
     * Sets abbreviated era strings. For example: "AD" and "BC".
     * @param eras  Array of era strings (DateFormatSymbols retains ownership.)
     * @param count Filled in with length of the array.
     * @stable ICU 2.0
     */
    void setEras(const UnicodeString* eras, int32_t count);

    /**
     * Gets era name strings. For example: "Anno Domini" and "Before Christ".
     *
     * @param count    Filled in with length of the array.
     * @return         the era name strings.
     * @stable ICU 3.4
     */
    const UnicodeString* getEraNames(int32_t& count) const;

    /**
     * Sets era name strings. For example: "Anno Domini" and "Before Christ".
     * @param eraNames  Array of era name strings (DateFormatSymbols retains ownership.)
     * @param count Filled in with length of the array.
     * @stable ICU 3.6
     */
    void setEraNames(const UnicodeString* eraNames, int32_t count);

    /**
     * Gets narrow era strings. For example: "A" and "B".
     *
     * @param count    Filled in with length of the array.
     * @return         the narrow era strings.
     * @stable ICU 4.2
     */
    const UnicodeString* getNarrowEras(int32_t& count) const;

    /**
     * Sets narrow era strings. For example: "A" and "B".
     * @param narrowEras  Array of narrow era strings (DateFormatSymbols retains ownership.)
     * @param count Filled in with length of the array.
     * @stable ICU 4.2
     */
    void setNarrowEras(const UnicodeString* narrowEras, int32_t count);

    /**
     * Gets month strings. For example: "January", "February", etc.
     * @param count Filled in with length of the array.
     * @return the month strings. (DateFormatSymbols retains ownership.)
     * @stable ICU 2.0
     */
    const UnicodeString* getMonths(int32_t& count) const;

    /**
     * Sets month strings. For example: "January", "February", etc.
     *
     * @param months    the new month strings. (not adopted; caller retains ownership)
     * @param count     Filled in with length of the array.
     * @stable ICU 2.0
     */
    void setMonths(const UnicodeString* months, int32_t count);

    /**
     * Gets short month strings. For example: "Jan", "Feb", etc.
     *
     * @param count Filled in with length of the array.
     * @return the short month strings. (DateFormatSymbols retains ownership.)
     * @stable ICU 2.0
     */
    const UnicodeString* getShortMonths(int32_t& count) const;

    /**
     * Sets short month strings. For example: "Jan", "Feb", etc.
     * @param count        Filled in with length of the array.
     * @param shortMonths  the new short month strings. (not adopted; caller retains ownership)
     * @stable ICU 2.0
     */
    void setShortMonths(const UnicodeString* shortMonths, int32_t count);

    /**
     * Selector for date formatting context
     * @stable ICU 3.6
     */
    enum DtContextType {
        FORMAT,
        STANDALONE,
#ifndef U_HIDE_DEPRECATED_API
        /**
         * One more than the highest normal DtContextType value.
         * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
         */
        DT_CONTEXT_COUNT
#endif  // U_HIDE_DEPRECATED_API
    };

    /**
     * Selector for date formatting width
     * @stable ICU 3.6
     */
    enum DtWidthType {
        ABBREVIATED,
        WIDE,
        NARROW,
        /**
         * Short width is currently only supported for weekday names.
         * @stable ICU 51
         */
        SHORT,
#ifndef U_HIDE_DEPRECATED_API
        /**
         * One more than the highest normal DtWidthType value.
         * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
         */
        DT_WIDTH_COUNT = 4
#endif  // U_HIDE_DEPRECATED_API
    };

    /**
     * Gets month strings by width and context. For example: "January", "February", etc.
     * @param count Filled in with length of the array.
     * @param context The formatting context, either FORMAT or STANDALONE
     * @param width   The width of returned strings, either WIDE, ABBREVIATED, or NARROW.
     * @return the month strings. (DateFormatSymbols retains ownership.)
     * @stable ICU 3.4
     */
    const UnicodeString* getMonths(int32_t& count, DtContextType context, DtWidthType width) const;

    /**
     * Sets month strings by width and context. For example: "January", "February", etc.
     *
     * @param months  The new month strings. (not adopted; caller retains ownership)
     * @param count   Filled in with length of the array.
     * @param context The formatting context, either FORMAT or STANDALONE
     * @param width   The width of returned strings, either WIDE, ABBREVIATED, or NARROW.
     * @stable ICU 3.6
     */
    void setMonths(const UnicodeString* months, int32_t count, DtContextType context, DtWidthType width);

    /**
     * Gets wide weekday strings. For example: "Sunday", "Monday", etc.
     * @param count        Filled in with length of the array.
     * @return the weekday strings. (DateFormatSymbols retains ownership.)
     * @stable ICU 2.0
     */
    const UnicodeString* getWeekdays(int32_t& count) const;


    /**
     * Sets wide weekday strings. For example: "Sunday", "Monday", etc.
     * @param weekdays     the new weekday strings. (not adopted; caller retains ownership)
     * @param count        Filled in with length of the array.
     * @stable ICU 2.0
     */
    void setWeekdays(const UnicodeString* weekdays, int32_t count);

    /**
     * Gets abbreviated weekday strings. For example: "Sun", "Mon", etc. (Note: The method name is
     * misleading; it does not get the CLDR-style "short" weekday strings, e.g. "Su", "Mo", etc.)
     * @param count        Filled in with length of the array.
     * @return             the abbreviated weekday strings. (DateFormatSymbols retains ownership.)
     * @stable ICU 2.0
     */
    const UnicodeString* getShortWeekdays(int32_t& count) const;

    /**
     * Sets abbreviated weekday strings. For example: "Sun", "Mon", etc. (Note: The method name is
     * misleading; it does not set the CLDR-style "short" weekday strings, e.g. "Su", "Mo", etc.)
     * @param abbrevWeekdays  the new abbreviated weekday strings. (not adopted; caller retains ownership)
     * @param count           Filled in with length of the array.
     * @stable ICU 2.0
     */
    void setShortWeekdays(const UnicodeString* abbrevWeekdays, int32_t count);

    /**
     * Gets weekday strings by width and context. For example: "Sunday", "Monday", etc.
     * @param count   Filled in with length of the array.
     * @param context The formatting context, either FORMAT or STANDALONE
     * @param width   The width of returned strings, either WIDE, ABBREVIATED, SHORT, or NARROW
     * @return the month strings. (DateFormatSymbols retains ownership.)
     * @stable ICU 3.4
     */
    const UnicodeString* getWeekdays(int32_t& count, DtContextType context, DtWidthType width) const;

    /**
     * Sets weekday strings by width and context. For example: "Sunday", "Monday", etc.
     * @param weekdays  The new weekday strings. (not adopted; caller retains ownership)
     * @param count     Filled in with length of the array.
     * @param context   The formatting context, either FORMAT or STANDALONE
     * @param width     The width of returned strings, either WIDE, ABBREVIATED, SHORT, or NARROW
     * @stable ICU 3.6
     */
    void setWeekdays(const UnicodeString* weekdays, int32_t count, DtContextType context, DtWidthType width);

    /**
     * Gets quarter strings by width and context. For example: "1st Quarter", "2nd Quarter", etc.
     * @param count Filled in with length of the array.
     * @param context The formatting context, either FORMAT or STANDALONE
     * @param width   The width of returned strings, either WIDE or ABBREVIATED. There
     *                are no NARROW quarters.
     * @return the quarter strings. (DateFormatSymbols retains ownership.)
     * @stable ICU 3.6
     */
    const UnicodeString* getQuarters(int32_t& count, DtContextType context, DtWidthType width) const;

    /**
     * Sets quarter strings by width and context. For example: "1st Quarter", "2nd Quarter", etc.
     *
     * @param quarters  The new quarter strings. (not adopted; caller retains ownership)
     * @param count   Filled in with length of the array.
     * @param context The formatting context, either FORMAT or STANDALONE
     * @param width   The width of returned strings, either WIDE or ABBREVIATED. There
     *                are no NARROW quarters.
     * @stable ICU 3.6
     */
    void setQuarters(const UnicodeString* quarters, int32_t count, DtContextType context, DtWidthType width);

    /**
     * Gets AM/PM strings. For example: "AM" and "PM".
     * @param count        Filled in with length of the array.
     * @return             the weekday strings. (DateFormatSymbols retains ownership.)
     * @stable ICU 2.0
     */
    const UnicodeString* getAmPmStrings(int32_t& count) const;

    /**
     * Sets ampm strings. For example: "AM" and "PM".
     * @param ampms        the new ampm strings. (not adopted; caller retains ownership)
     * @param count        Filled in with length of the array.
     * @stable ICU 2.0
     */
    void setAmPmStrings(const UnicodeString* ampms, int32_t count);

#ifndef U_HIDE_INTERNAL_API
    /**
     * This default time separator is used for formatting when the locale
     * doesn't specify any time separator, and always recognized when parsing.
     * @internal
     */
    static const char16_t DEFAULT_TIME_SEPARATOR = 0x003a;  // ':'

    /**
     * This alternate time separator is always recognized when parsing.
     * @internal
     */
    static const char16_t ALTERNATE_TIME_SEPARATOR = 0x002e;  // '.'

    /**
     * Gets the time separator string. For example: ":".
     * @param result Output param which will receive the time separator string.
     * @return       A reference to 'result'.
     * @internal
     */
    UnicodeString& getTimeSeparatorString(UnicodeString& result) const;

    /**
     * Sets the time separator string. For example: ":".
     * @param newTimeSeparator the new time separator string.
     * @internal
     */
    void setTimeSeparatorString(const UnicodeString& newTimeSeparator);
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Gets cyclic year name strings if the calendar has them, by width and context.
     * For example: "jia-zi", "yi-chou", etc.
     * @param count     Filled in with length of the array.
     * @param context   The usage context: FORMAT, STANDALONE.
     * @param width     The requested name width: WIDE, ABBREVIATED, NARROW.
     * @return          The year name strings (DateFormatSymbols retains ownership),
     *                  or null if they are not available for this calendar.
     * @stable ICU 54
     */
    const UnicodeString* getYearNames(int32_t& count,
                            DtContextType context, DtWidthType width) const;

    /**
     * Sets cyclic year name strings by width and context. For example: "jia-zi", "yi-chou", etc.
     *
     * @param yearNames The new cyclic year name strings (not adopted; caller retains ownership).
     * @param count     The length of the array.
     * @param context   The usage context: FORMAT, STANDALONE (currently only FORMAT is supported).
     * @param width     The name width: WIDE, ABBREVIATED, NARROW (currently only ABBREVIATED is supported).
     * @stable ICU 54
     */
    void setYearNames(const UnicodeString* yearNames, int32_t count,
                            DtContextType context, DtWidthType width);

    /**
     * Gets calendar zodiac name strings if the calendar has them, by width and context.
     * For example: "Rat", "Ox", "Tiger", etc.
     * @param count     Filled in with length of the array.
     * @param context   The usage context: FORMAT, STANDALONE.
     * @param width     The requested name width: WIDE, ABBREVIATED, NARROW.
     * @return          The zodiac name strings (DateFormatSymbols retains ownership),
     *                  or null if they are not available for this calendar.
     * @stable ICU 54
     */
    const UnicodeString* getZodiacNames(int32_t& count,
                            DtContextType context, DtWidthType width) const;

    /**
     * Sets calendar zodiac name strings by width and context. For example: "Rat", "Ox", "Tiger", etc.
     *
     * @param zodiacNames The new zodiac name strings (not adopted; caller retains ownership).
     * @param count     The length of the array.
     * @param context   The usage context: FORMAT, STANDALONE (currently only FORMAT is supported).
     * @param width     The name width: WIDE, ABBREVIATED, NARROW (currently only ABBREVIATED is supported).
     * @stable ICU 54
     */
    void setZodiacNames(const UnicodeString* zodiacNames, int32_t count,
                            DtContextType context, DtWidthType width);

#ifndef U_HIDE_INTERNAL_API
    /**
     * Somewhat temporary constants for leap month pattern types, adequate for supporting
     * just leap month patterns as needed for Chinese lunar calendar.
     * Eventually we will add full support for different month pattern types (needed for
     * other calendars such as Hindu) at which point this approach will be replaced by a
     * more complete approach.
     * @internal
     */
    enum EMonthPatternType
    {
        kLeapMonthPatternFormatWide,
        kLeapMonthPatternFormatAbbrev,
        kLeapMonthPatternFormatNarrow,
        kLeapMonthPatternStandaloneWide,
        kLeapMonthPatternStandaloneAbbrev,
        kLeapMonthPatternStandaloneNarrow,
        kLeapMonthPatternNumeric,
        kMonthPatternsCount
    };

    /**
     * Somewhat temporary function for getting complete set of leap month patterns for all
     * contexts & widths, indexed by EMonthPatternType values. Returns NULL if calendar
     * does not have leap month patterns. Note, there is currently no setter for this.
     * Eventually we will add full support for different month pattern types (needed for
     * other calendars such as Hindu) at which point this approach will be replaced by a
     * more complete approach.
     * @param count        Filled in with length of the array (may be 0).
     * @return             The leap month patterns (DateFormatSymbols retains ownership).
     *                     May be NULL if there are no leap month patterns for this calendar.
     * @internal
     */
    const UnicodeString* getLeapMonthPatterns(int32_t& count) const;

#endif  /* U_HIDE_INTERNAL_API */

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Gets timezone strings. These strings are stored in a 2-dimensional array.
     * @param rowCount      Output param to receive number of rows.
     * @param columnCount   Output param to receive number of columns.
     * @return              The timezone strings as a 2-d array. (DateFormatSymbols retains ownership.)
     * @deprecated ICU 3.6
     */
    const UnicodeString** getZoneStrings(int32_t& rowCount, int32_t& columnCount) const;
#endif  /* U_HIDE_DEPRECATED_API */

    /**
     * Sets timezone strings. These strings are stored in a 2-dimensional array.
     * <p><b>Note:</b> SimpleDateFormat no longer use the zone strings stored in
     * a DateFormatSymbols. Therefore, the time zone strings set by this mthod
     * have no effects in an instance of SimpleDateFormat for formatting time
     * zones.
     * @param strings       The timezone strings as a 2-d array to be copied. (not adopted; caller retains ownership)
     * @param rowCount      The number of rows (count of first index).
     * @param columnCount   The number of columns (count of second index).
     * @stable ICU 2.0
     */
    void setZoneStrings(const UnicodeString* const* strings, int32_t rowCount, int32_t columnCount);

    /**
     * Get the non-localized date-time pattern characters.
     * @return    the non-localized date-time pattern characters
     * @stable ICU 2.0
     */
    static const char16_t * U_EXPORT2 getPatternUChars(void);

    /**
     * Gets localized date-time pattern characters. For example: 'u', 't', etc.
     * <p>
     * Note: ICU no longer provides localized date-time pattern characters for a locale
     * starting ICU 3.8.  This method returns the non-localized date-time pattern
     * characters unless user defined localized data is set by setLocalPatternChars.
     * @param result    Output param which will receive the localized date-time pattern characters.
     * @return          A reference to 'result'.
     * @stable ICU 2.0
     */
    UnicodeString& getLocalPatternChars(UnicodeString& result) const;

    /**
     * Sets localized date-time pattern characters. For example: 'u', 't', etc.
     * @param newLocalPatternChars the new localized date-time
     * pattern characters.
     * @stable ICU 2.0
     */
    void setLocalPatternChars(const UnicodeString& newLocalPatternChars);

    /**
     * Returns the locale for this object. Two flavors are available:
     * valid and actual locale.
     * @stable ICU 2.8
     */
    Locale getLocale(ULocDataLocaleType type, UErrorCode& status) const;

    /* The following type and kCapContextUsageTypeCount cannot be #ifndef U_HIDE_INTERNAL_API,
       they are needed for .h file declarations. */ 
    /**
     * Constants for capitalization context usage types.
     * @internal
     */
    enum ECapitalizationContextUsageType
    {
#ifndef U_HIDE_INTERNAL_API
        kCapContextUsageOther = 0,
        kCapContextUsageMonthFormat,     /* except narrow */
        kCapContextUsageMonthStandalone, /* except narrow */
        kCapContextUsageMonthNarrow,
        kCapContextUsageDayFormat,     /* except narrow */
        kCapContextUsageDayStandalone, /* except narrow */
        kCapContextUsageDayNarrow,
        kCapContextUsageEraWide,
        kCapContextUsageEraAbbrev,
        kCapContextUsageEraNarrow,
        kCapContextUsageZoneLong,
        kCapContextUsageZoneShort,
        kCapContextUsageMetazoneLong,
        kCapContextUsageMetazoneShort,
#endif /* U_HIDE_INTERNAL_API */
        kCapContextUsageTypeCount = 14
    };

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 2.2
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 2.2
     */
    static UClassID U_EXPORT2 getStaticClassID();

private:

    friend class SimpleDateFormat;
    friend class DateFormatSymbolsSingleSetter; // see udat.cpp

    /**
     * Abbreviated era strings. For example: "AD" and "BC".
     */
    UnicodeString*  fEras;
    int32_t         fErasCount;

    /**
     * Era name strings. For example: "Anno Domini" and "Before Christ".
     */
    UnicodeString*  fEraNames;
    int32_t         fEraNamesCount;

    /**
     * Narrow era strings. For example: "A" and "B".
     */
    UnicodeString*  fNarrowEras;
    int32_t         fNarrowErasCount;

    /**
     * Month strings. For example: "January", "February", etc.
     */
    UnicodeString*  fMonths;
    int32_t         fMonthsCount;

    /**
     * Short month strings. For example: "Jan", "Feb", etc.
     */
    UnicodeString*  fShortMonths;
    int32_t         fShortMonthsCount;

    /**
     * Narrow month strings. For example: "J", "F", etc.
     */
    UnicodeString*  fNarrowMonths;
    int32_t         fNarrowMonthsCount;

    /**
     * Standalone Month strings. For example: "January", "February", etc.
     */
    UnicodeString*  fStandaloneMonths;
    int32_t         fStandaloneMonthsCount;

    /**
     * Standalone Short month strings. For example: "Jan", "Feb", etc.
     */
    UnicodeString*  fStandaloneShortMonths;
    int32_t         fStandaloneShortMonthsCount;

    /**
     * Standalone Narrow month strings. For example: "J", "F", etc.
     */
    UnicodeString*  fStandaloneNarrowMonths;
    int32_t         fStandaloneNarrowMonthsCount;

    /**
     * CLDR-style format wide weekday strings. For example: "Sunday", "Monday", etc.
     */
    UnicodeString*  fWeekdays;
    int32_t         fWeekdaysCount;

    /**
     * CLDR-style format abbreviated (not short) weekday strings. For example: "Sun", "Mon", etc.
     */
    UnicodeString*  fShortWeekdays;
    int32_t         fShortWeekdaysCount;

    /**
     * CLDR-style format short weekday strings. For example: "Su", "Mo", etc.
     */
    UnicodeString*  fShorterWeekdays;
    int32_t         fShorterWeekdaysCount;

    /**
     * CLDR-style format narrow weekday strings. For example: "S", "M", etc.
     */
    UnicodeString*  fNarrowWeekdays;
    int32_t         fNarrowWeekdaysCount;

    /**
     * CLDR-style standalone wide weekday strings. For example: "Sunday", "Monday", etc.
     */
    UnicodeString*  fStandaloneWeekdays;
    int32_t         fStandaloneWeekdaysCount;

    /**
     * CLDR-style standalone abbreviated (not short) weekday strings. For example: "Sun", "Mon", etc.
     */
    UnicodeString*  fStandaloneShortWeekdays;
    int32_t         fStandaloneShortWeekdaysCount;

    /**
     * CLDR-style standalone short weekday strings. For example: "Su", "Mo", etc.
     */
    UnicodeString*  fStandaloneShorterWeekdays;
    int32_t         fStandaloneShorterWeekdaysCount;

    /**
     * Standalone Narrow weekday strings. For example: "Sun", "Mon", etc.
     */
    UnicodeString*  fStandaloneNarrowWeekdays;
    int32_t         fStandaloneNarrowWeekdaysCount;

    /**
     * Ampm strings. For example: "AM" and "PM".
     */
    UnicodeString*  fAmPms;
    int32_t         fAmPmsCount;

    /**
     * Narrow Ampm strings. For example: "a" and "p".
     */
    UnicodeString*  fNarrowAmPms;
    int32_t         fNarrowAmPmsCount;

    /**
     * Time separator string. For example: ":".
     */
    UnicodeString   fTimeSeparator;

    /**
     * Quarter strings. For example: "1st quarter", "2nd quarter", etc.
     */
    UnicodeString  *fQuarters;
    int32_t         fQuartersCount;

    /**
     * Short quarters. For example: "Q1", "Q2", etc.
     */
    UnicodeString  *fShortQuarters;
    int32_t         fShortQuartersCount;

    /**
     * Standalone quarter strings. For example: "1st quarter", "2nd quarter", etc.
     */
    UnicodeString  *fStandaloneQuarters;
    int32_t         fStandaloneQuartersCount;

    /**
     * Standalone short quarter strings. For example: "Q1", "Q2", etc.
     */
    UnicodeString  *fStandaloneShortQuarters;
    int32_t         fStandaloneShortQuartersCount;

    /**
     * All leap month patterns, for example "{0}bis".
     */
    UnicodeString  *fLeapMonthPatterns;
    int32_t         fLeapMonthPatternsCount;

    /**
     * Cyclic year names, for example: "jia-zi", "yi-chou", ... "gui-hai";
     * currently we only have data for format/abbreviated.
     * For the others, just get from format/abbreviated, ignore set.
     */
    UnicodeString  *fShortYearNames;
    int32_t         fShortYearNamesCount;

    /**
     * Cyclic zodiac names, for example "Rat", "Ox", "Tiger", etc.;
     * currently we only have data for format/abbreviated.
     * For the others, just get from format/abbreviated, ignore set.
     */
    UnicodeString  *fShortZodiacNames;
    int32_t         fShortZodiacNamesCount;

    /**
     * Localized names of time zones in this locale.  This is a
     * two-dimensional array of strings of size n by m,
     * where m is at least 5 and up to 7.  Each of the n rows is an
     * entry containing the localized names for a single TimeZone.
     *
     * Each such row contains (with i ranging from 0..n-1):
     * 
     * zoneStrings[i][0] - time zone ID
     *  example: America/Los_Angeles
     * zoneStrings[i][1] - long name of zone in standard time
     *  example: Pacific Standard Time
     * zoneStrings[i][2] - short name of zone in standard time
     *  example: PST
     * zoneStrings[i][3] - long name of zone in daylight savings time
     *  example: Pacific Daylight Time
     * zoneStrings[i][4] - short name of zone in daylight savings time
     *  example: PDT
     * zoneStrings[i][5] - location name of zone
     *  example: United States (Los Angeles)
     * zoneStrings[i][6] - long generic name of zone
     *  example: Pacific Time
     * zoneStrings[i][7] - short generic of zone
     *  example: PT
     *
     * The zone ID is not localized; it corresponds to the ID
     * value associated with a system time zone object.  All other entries
     * are localized names.  If a zone does not implement daylight savings
     * time, the daylight savings time names are ignored.
     *
     * Note:CLDR 1.5 introduced metazone and its historical mappings.
     * This simple two-dimensional array is no longer sufficient to represent
     * localized names and its historic changes.  Since ICU 3.8.1, localized
     * zone names extracted from ICU locale data is stored in a ZoneStringFormat
     * instance.  But we still need to support the old way of customizing
     * localized zone names, so we keep this field for the purpose.
     */
    UnicodeString   **fZoneStrings;         // Zone string array set by setZoneStrings
    UnicodeString   **fLocaleZoneStrings;   // Zone string array created by the locale
    int32_t         fZoneStringsRowCount;
    int32_t         fZoneStringsColCount;

    Locale                  fZSFLocale;         // Locale used for getting ZoneStringFormat

    /**
     * Localized date-time pattern characters. For example: use 'u' as 'y'.
     */
    UnicodeString   fLocalPatternChars;

    /**
     * Capitalization transforms. For each usage type, the first array element indicates
     * whether to titlecase for uiListOrMenu context, the second indicates whether to
     * titlecase for stand-alone context.
     */
     UBool fCapitalization[kCapContextUsageTypeCount][2];

    /**
     * Abbreviated (== short) day period strings.
     */
    UnicodeString  *fAbbreviatedDayPeriods;
    int32_t         fAbbreviatedDayPeriodsCount;

    /**
     * Wide day period strings.
     */
    UnicodeString  *fWideDayPeriods;
    int32_t         fWideDayPeriodsCount;

    /**
     * Narrow day period strings.
     */
    UnicodeString  *fNarrowDayPeriods;
    int32_t         fNarrowDayPeriodsCount;

    /**
     * Stand-alone abbreviated (== short) day period strings.
     */
    UnicodeString  *fStandaloneAbbreviatedDayPeriods;
    int32_t         fStandaloneAbbreviatedDayPeriodsCount;

    /**
     * Stand-alone wide day period strings.
     */
    UnicodeString  *fStandaloneWideDayPeriods;
    int32_t         fStandaloneWideDayPeriodsCount;

    /**
     * Stand-alone narrow day period strings.
     */
    UnicodeString  *fStandaloneNarrowDayPeriods;
    int32_t         fStandaloneNarrowDayPeriodsCount;

private:
    /** valid/actual locale information 
     *  these are always ICU locales, so the length should not be a problem
     */
    char validLocale[ULOC_FULLNAME_CAPACITY];
    char actualLocale[ULOC_FULLNAME_CAPACITY];

    DateFormatSymbols(); // default constructor not implemented

    /**
     * Called by the constructors to actually load data from the resources
     *
     * @param locale               The locale to get symbols for.
     * @param type                 Calendar Type (as from Calendar::getType())
     * @param status               Input/output parameter, set to success or
     *                             failure code upon return.
     * @param useLastResortData    determine if use last resort data
     */
    void initializeData(const Locale& locale, const char *type, UErrorCode& status, UBool useLastResortData = FALSE);

    /**
     * Copy or alias an array in another object, as appropriate.
     *
     * @param dstArray    the copy destination array.
     * @param dstCount    fill in with the lenth of 'dstArray'.
     * @param srcArray    the source array to be copied.
     * @param srcCount    the length of items to be copied from the 'srcArray'.
     */
    static void assignArray(UnicodeString*& dstArray,
                            int32_t& dstCount,
                            const UnicodeString* srcArray,
                            int32_t srcCount);

    /**
     * Return true if the given arrays' contents are equal, or if the arrays are
     * identical (pointers are equal).
     *
     * @param array1   one array to be compared with.
     * @param array2   another array to be compared with.
     * @param count    the length of items to be copied.
     * @return         true if the given arrays' contents are equal, or if the arrays are
     *                 identical (pointers are equal).
     */
    static UBool arrayCompare(const UnicodeString* array1,
                             const UnicodeString* array2,
                             int32_t count);

    /**
     * Create a copy, in fZoneStrings, of the given zone strings array. The
     * member variables fZoneStringsRowCount and fZoneStringsColCount should be
     * set already by the caller.
     */
    void createZoneStrings(const UnicodeString *const * otherStrings);

    /**
     * Delete all the storage owned by this object.
     */
    void dispose(void);

    /**
     * Copy all of the other's data to this.
     * @param other the object to be copied.
     */
    void copyData(const DateFormatSymbols& other);

    /**
     * Create zone strings array by locale if not yet available
     */
    void initZoneStringsArray(void);

    /**
     * Delete just the zone strings.
     */
    void disposeZoneStrings(void);

    /**
     * Returns the date format field index of the pattern character c,
     * or UDAT_FIELD_COUNT if c is not a pattern character.
     */
    static UDateFormatField U_EXPORT2 getPatternCharIndex(char16_t c);

    /**
     * Returns TRUE if f (with its pattern character repeated count times) is a numeric field.
     */
    static UBool U_EXPORT2 isNumericField(UDateFormatField f, int32_t count);

    /**
     * Returns TRUE if c (repeated count times) is the pattern character for a numeric field.
     */
    static UBool U_EXPORT2 isNumericPatternChar(char16_t c, int32_t count);
public:
#ifndef U_HIDE_INTERNAL_API
    /**
     * Gets a DateFormatSymbols by locale.
     * Unlike the constructors which always use gregorian calendar, this
     * method uses the calendar in the locale. If the locale contains no
     * explicit calendar, this method uses the default calendar for that
     * locale.
     * @param locale the locale.
     * @param status error returned here.
     * @return the new DateFormatSymbols which the caller owns.
     * @internal For ICU use only.
     */
    static DateFormatSymbols * U_EXPORT2 createForLocale(
            const Locale &locale, UErrorCode &status);
#endif  /* U_HIDE_INTERNAL_API */
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
* Copyright (C) 2007-2013, International Business Machines Corporation and
* others. All Rights Reserved.
********************************************************************************
*
* File MSGFMT.H
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/20/97    helena      Finished first cut of implementation.
*   07/22/98    stephen     Removed operator!= (defined in Format)
*   08/19/2002  srl         Removing Javaisms
*******************************************************************************/




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Formats messages in a language-neutral way.
 */

#if !UCONFIG_NO_FORMATTING





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2014, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*

* File PLURFMT.H
********************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: PluralFormat object
 */

#if !UCONFIG_NO_FORMATTING





U_NAMESPACE_BEGIN

class Hashtable;
class NFRule;

/**
 * <p>
 * <code>PluralFormat</code> supports the creation of internationalized
 * messages with plural inflection. It is based on <i>plural
 * selection</i>, i.e. the caller specifies messages for each
 * plural case that can appear in the user's language and the
 * <code>PluralFormat</code> selects the appropriate message based on
 * the number.
 * </p>
 * <h4>The Problem of Plural Forms in Internationalized Messages</h4>
 * <p>
 * Different languages have different ways to inflect
 * plurals. Creating internationalized messages that include plural
 * forms is only feasible when the framework is able to handle plural
 * forms of <i>all</i> languages correctly. <code>ChoiceFormat</code>
 * doesn't handle this well, because it attaches a number interval to
 * each message and selects the message whose interval contains a
 * given number. This can only handle a finite number of
 * intervals. But in some languages, like Polish, one plural case
 * applies to infinitely many intervals (e.g., the plural case applies to
 * numbers ending with 2, 3, or 4 except those ending with 12, 13, or
 * 14). Thus <code>ChoiceFormat</code> is not adequate.
 * </p><p>
 * <code>PluralFormat</code> deals with this by breaking the problem
 * into two parts:
 * <ul>
 * <li>It uses <code>PluralRules</code> that can define more complex
 *     conditions for a plural case than just a single interval. These plural
 *     rules define both what plural cases exist in a language, and to
 *     which numbers these cases apply.
 * <li>It provides predefined plural rules for many languages. Thus, the programmer
 *     need not worry about the plural cases of a language and
 *     does not have to define the plural cases; they can simply
 *     use the predefined keywords. The whole plural formatting of messages can
 *     be done using localized patterns from resource bundles. For predefined plural
 *     rules, see the CLDR <i>Language Plural Rules</i> page at
 *    http://unicode.org/repos/cldr-tmp/trunk/diff/supplemental/language_plural_rules.html
 * </ul>
 * </p>
 * <h4>Usage of <code>PluralFormat</code></h4>
 * <p>Note: Typically, plural formatting is done via <code>MessageFormat</code>
 * with a <code>plural</code> argument type,
 * rather than using a stand-alone <code>PluralFormat</code>.
 * </p><p>
 * This discussion assumes that you use <code>PluralFormat</code> with
 * a predefined set of plural rules. You can create one using one of
 * the constructors that takes a <code>locale</code> object. To
 * specify the message pattern, you can either pass it to the
 * constructor or set it explicitly using the
 * <code>applyPattern()</code> method. The <code>format()</code>
 * method takes a number object and selects the message of the
 * matching plural case. This message will be returned.
 * </p>
 * <h5>Patterns and Their Interpretation</h5>
 * <p>
 * The pattern text defines the message output for each plural case of the
 * specified locale. Syntax:
 * <pre>
 * pluralStyle = [offsetValue] (selector '{' message '}')+
 * offsetValue = "offset:" number
 * selector = explicitValue | keyword
 * explicitValue = '=' number  // adjacent, no white space in between
 * keyword = [^[[:Pattern_Syntax:][:Pattern_White_Space:]]]+
 * message: see {@link MessageFormat}
 * </pre>
 * Pattern_White_Space between syntax elements is ignored, except
 * between the {curly braces} and their sub-message,
 * and between the '=' and the number of an explicitValue.
 *
 * </p><p>
 * There are 6 predefined casekeyword in CLDR/ICU - 'zero', 'one', 'two', 'few', 'many' and
 * 'other'. You always have to define a message text for the default plural case
 * <code>other</code> which is contained in every rule set.
 * If you do not specify a message text for a particular plural case, the
 * message text of the plural case <code>other</code> gets assigned to this
 * plural case.
 * </p><p>
 * When formatting, the input number is first matched against the explicitValue clauses.
 * If there is no exact-number match, then a keyword is selected by calling
 * the <code>PluralRules</code> with the input number <em>minus the offset</em>.
 * (The offset defaults to 0 if it is omitted from the pattern string.)
 * If there is no clause with that keyword, then the "other" clauses is returned.
 * </p><p>
 * An unquoted pound sign (<code>#</code>) in the selected sub-message
 * itself (i.e., outside of arguments nested in the sub-message)
 * is replaced by the input number minus the offset.
 * The number-minus-offset value is formatted using a
 * <code>NumberFormat</code> for the <code>PluralFormat</code>'s locale. If you
 * need special number formatting, you have to use a <code>MessageFormat</code>
 * and explicitly specify a <code>NumberFormat</code> argument.
 * <strong>Note:</strong> That argument is formatting without subtracting the offset!
 * If you need a custom format and have a non-zero offset, then you need to pass the
 * number-minus-offset value as a separate parameter.
 * </p>
 * For a usage example, see the {@link MessageFormat} class documentation.
 *
 * <h4>Defining Custom Plural Rules</h4>
 * <p>If you need to use <code>PluralFormat</code> with custom rules, you can
 * create a <code>PluralRules</code> object and pass it to
 * <code>PluralFormat</code>'s constructor. If you also specify a locale in this
 * constructor, this locale will be used to format the number in the message
 * texts.
 * </p><p>
 * For more information about <code>PluralRules</code>, see
 * {@link PluralRules}.
 * </p>
 *
 * ported from Java
 * @stable ICU 4.0
 */

class U_I18N_API PluralFormat : public Format {
public:

    /**
     * Creates a new cardinal-number <code>PluralFormat</code> for the default locale.
     * This locale will be used to get the set of plural rules and for standard
     * number formatting.
     * @param status  output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @stable ICU 4.0
     */
    PluralFormat(UErrorCode& status);

    /**
     * Creates a new cardinal-number <code>PluralFormat</code> for a given locale.
     * @param locale the <code>PluralFormat</code> will be configured with
     *               rules for this locale. This locale will also be used for
     *               standard number formatting.
     * @param status output param set to success/failure code on exit, which
     *               must not indicate a failure before the function call.
     * @stable ICU 4.0
     */
    PluralFormat(const Locale& locale, UErrorCode& status);

    /**
     * Creates a new <code>PluralFormat</code> for a given set of rules.
     * The standard number formatting will be done using the default locale.
     * @param rules   defines the behavior of the <code>PluralFormat</code>
     *                object.
     * @param status  output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @stable ICU 4.0
     */
    PluralFormat(const PluralRules& rules, UErrorCode& status);

    /**
     * Creates a new <code>PluralFormat</code> for a given set of rules.
     * The standard number formatting will be done using the given locale.
     * @param locale  the default number formatting will be done using this
     *                locale.
     * @param rules   defines the behavior of the <code>PluralFormat</code>
     *                object.
     * @param status  output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @stable ICU 4.0
	 * <p>
	 * <h4>Sample code</h4>
	 * \snippet samples/plurfmtsample/plurfmtsample.cpp PluralFormatExample1
	 * \snippet samples/plurfmtsample/plurfmtsample.cpp PluralFormatExample
	 * <p>
     */
    PluralFormat(const Locale& locale, const PluralRules& rules, UErrorCode& status);

    /**
     * Creates a new <code>PluralFormat</code> for the plural type.
     * The standard number formatting will be done using the given locale.
     * @param locale  the default number formatting will be done using this
     *                locale.
     * @param type    The plural type (e.g., cardinal or ordinal).
     * @param status  output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @stable ICU 50
     */
    PluralFormat(const Locale& locale, UPluralType type, UErrorCode& status);

    /**
     * Creates a new cardinal-number <code>PluralFormat</code> for a given pattern string.
     * The default locale will be used to get the set of plural rules and for
     * standard number formatting.
     * @param  pattern the pattern for this <code>PluralFormat</code>.
     *                 errors are returned to status if the pattern is invalid.
     * @param status   output param set to success/failure code on exit, which
     *                 must not indicate a failure before the function call.
     * @stable ICU 4.0
     */
    PluralFormat(const UnicodeString& pattern, UErrorCode& status);

    /**
     * Creates a new cardinal-number <code>PluralFormat</code> for a given pattern string and
     * locale.
     * The locale will be used to get the set of plural rules and for
     * standard number formatting.
     * @param locale   the <code>PluralFormat</code> will be configured with
     *                 rules for this locale. This locale will also be used for
     *                 standard number formatting.
     * @param pattern  the pattern for this <code>PluralFormat</code>.
     *                 errors are returned to status if the pattern is invalid.
     * @param status   output param set to success/failure code on exit, which
     *                 must not indicate a failure before the function call.
     * @stable ICU 4.0
     */
    PluralFormat(const Locale& locale, const UnicodeString& pattern, UErrorCode& status);

    /**
     * Creates a new <code>PluralFormat</code> for a given set of rules, a
     * pattern and a locale.
     * @param rules    defines the behavior of the <code>PluralFormat</code>
     *                 object.
     * @param pattern  the pattern for this <code>PluralFormat</code>.
     *                 errors are returned to status if the pattern is invalid.
     * @param status   output param set to success/failure code on exit, which
     *                 must not indicate a failure before the function call.
     * @stable ICU 4.0
     */
    PluralFormat(const PluralRules& rules,
                 const UnicodeString& pattern,
                 UErrorCode& status);

    /**
     * Creates a new <code>PluralFormat</code> for a given set of rules, a
     * pattern and a locale.
     * @param locale  the <code>PluralFormat</code> will be configured with
     *                rules for this locale. This locale will also be used for
     *                standard number formatting.
     * @param rules   defines the behavior of the <code>PluralFormat</code>
     *                object.
     * @param pattern the pattern for this <code>PluralFormat</code>.
     *                errors are returned to status if the pattern is invalid.
     * @param status  output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @stable ICU 4.0
     */
    PluralFormat(const Locale& locale,
                 const PluralRules& rules,
                 const UnicodeString& pattern,
                 UErrorCode& status);

    /**
     * Creates a new <code>PluralFormat</code> for a plural type, a
     * pattern and a locale.
     * @param locale  the <code>PluralFormat</code> will be configured with
     *                rules for this locale. This locale will also be used for
     *                standard number formatting.
     * @param type    The plural type (e.g., cardinal or ordinal).
     * @param pattern the pattern for this <code>PluralFormat</code>.
     *                errors are returned to status if the pattern is invalid.
     * @param status  output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @stable ICU 50
     */
    PluralFormat(const Locale& locale,
                 UPluralType type,
                 const UnicodeString& pattern,
                 UErrorCode& status);

    /**
      * copy constructor.
      * @stable ICU 4.0
      */
    PluralFormat(const PluralFormat& other);

    /**
     * Destructor.
     * @stable ICU 4.0
     */
    virtual ~PluralFormat();

    /**
     * Sets the pattern used by this plural format.
     * The method parses the pattern and creates a map of format strings
     * for the plural rules.
     * Patterns and their interpretation are specified in the class description.
     *
     * @param pattern the pattern for this plural format
     *                errors are returned to status if the pattern is invalid.
     * @param status  output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @stable ICU 4.0
     */
    void applyPattern(const UnicodeString& pattern, UErrorCode& status);


    using Format::format;

    /**
     * Formats a plural message for a given number.
     *
     * @param number  a number for which the plural message should be formatted
     *                for. If no pattern has been applied to this
     *                <code>PluralFormat</code> object yet, the formatted number
     *                will be returned.
     * @param status  output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @return        the string containing the formatted plural message.
     * @stable ICU 4.0
     */
    UnicodeString format(int32_t number, UErrorCode& status) const;

    /**
     * Formats a plural message for a given number.
     *
     * @param number  a number for which the plural message should be formatted
     *                for. If no pattern has been applied to this
     *                PluralFormat object yet, the formatted number
     *                will be returned.
     * @param status  output param set to success or failure code on exit, which
     *                must not indicate a failure before the function call.
     * @return        the string containing the formatted plural message.
     * @stable ICU 4.0
     */
    UnicodeString format(double number, UErrorCode& status) const;

    /**
     * Formats a plural message for a given number.
     *
     * @param number   a number for which the plural message should be formatted
     *                 for. If no pattern has been applied to this
     *                 <code>PluralFormat</code> object yet, the formatted number
     *                 will be returned.
     * @param appendTo output parameter to receive result.
     *                 result is appended to existing contents.
     * @param pos      On input: an alignment field, if desired.
     *                 On output: the offsets of the alignment field.
     * @param status   output param set to success/failure code on exit, which
     *                 must not indicate a failure before the function call.
     * @return         the string containing the formatted plural message.
     * @stable ICU 4.0
     */
    UnicodeString& format(int32_t number,
                          UnicodeString& appendTo,
                          FieldPosition& pos,
                          UErrorCode& status) const;

    /**
     * Formats a plural message for a given number.
     *
     * @param number   a number for which the plural message should be formatted
     *                 for. If no pattern has been applied to this
     *                 PluralFormat object yet, the formatted number
     *                 will be returned.
     * @param appendTo output parameter to receive result.
     *                 result is appended to existing contents.
     * @param pos      On input: an alignment field, if desired.
     *                 On output: the offsets of the alignment field.
     * @param status   output param set to success/failure code on exit, which
     *                 must not indicate a failure before the function call.
     * @return         the string containing the formatted plural message.
     * @stable ICU 4.0
     */
    UnicodeString& format(double number,
                          UnicodeString& appendTo,
                          FieldPosition& pos,
                          UErrorCode& status) const;

#ifndef U_HIDE_DEPRECATED_API
    /**
     * Sets the locale used by this <code>PluraFormat</code> object.
     * Note: Calling this method resets this <code>PluraFormat</code> object,
     *     i.e., a pattern that was applied previously will be removed,
     *     and the NumberFormat is set to the default number format for
     *     the locale.  The resulting format behaves the same as one
     *     constructed from {@link #PluralFormat(const Locale& locale, UPluralType type, UErrorCode& status)}
     *     with UPLURAL_TYPE_CARDINAL.
     * @param locale  the <code>locale</code> to use to configure the formatter.
     * @param status  output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @deprecated ICU 50 This method clears the pattern and might create
     *             a different kind of PluralRules instance;
     *             use one of the constructors to create a new instance instead.
     */
    void setLocale(const Locale& locale, UErrorCode& status);
#endif  /* U_HIDE_DEPRECATED_API */

    /**
      * Sets the number format used by this formatter.  You only need to
      * call this if you want a different number format than the default
      * formatter for the locale.
      * @param format  the number format to use.
      * @param status  output param set to success/failure code on exit, which
      *                must not indicate a failure before the function call.
      * @stable ICU 4.0
      */
    void setNumberFormat(const NumberFormat* format, UErrorCode& status);

    /**
       * Assignment operator
       *
       * @param other    the PluralFormat object to copy from.
       * @stable ICU 4.0
       */
    PluralFormat& operator=(const PluralFormat& other);

    /**
      * Return true if another object is semantically equal to this one.
      *
      * @param other    the PluralFormat object to be compared with.
      * @return         true if other is semantically equal to this.
      * @stable ICU 4.0
      */
    virtual UBool operator==(const Format& other) const;

    /**
     * Return true if another object is semantically unequal to this one.
     *
     * @param other    the PluralFormat object to be compared with.
     * @return         true if other is semantically unequal to this.
     * @stable ICU 4.0
     */
    virtual UBool operator!=(const Format& other) const;

    /**
     * Clones this Format object polymorphically.  The caller owns the
     * result and should delete it when done.
     * @stable ICU 4.0
     */
    virtual PluralFormat* clone() const;

   /**
    * Formats a plural message for a number taken from a Formattable object.
    *
    * @param obj       The object containing a number for which the
    *                  plural message should be formatted.
    *                  The object must be of a numeric type.
    * @param appendTo  output parameter to receive result.
    *                  Result is appended to existing contents.
    * @param pos       On input: an alignment field, if desired.
    *                  On output: the offsets of the alignment field.
    * @param status    output param filled with success/failure status.
    * @return          Reference to 'appendTo' parameter.
    * @stable ICU 4.0
    */
   UnicodeString& format(const Formattable& obj,
                         UnicodeString& appendTo,
                         FieldPosition& pos,
                         UErrorCode& status) const;

   /**
    * Returns the pattern from applyPattern() or constructor().
    *
    * @param  appendTo  output parameter to receive result.
     *                  Result is appended to existing contents.
    * @return the UnicodeString with inserted pattern.
    * @stable ICU 4.0
    */
   UnicodeString& toPattern(UnicodeString& appendTo);

   /**
    * This method is not yet supported by <code>PluralFormat</code>.
    * <P>
    * Before calling, set parse_pos.index to the offset you want to start
    * parsing at in the source. After calling, parse_pos.index is the end of
    * the text you parsed. If error occurs, index is unchanged.
    * <P>
    * When parsing, leading whitespace is discarded (with a successful parse),
    * while trailing whitespace is left as is.
    * <P>
    * See Format::parseObject() for more.
    *
    * @param source    The string to be parsed into an object.
    * @param result    Formattable to be set to the parse result.
    *                  If parse fails, return contents are undefined.
    * @param parse_pos The position to start parsing at. Upon return
    *                  this param is set to the position after the
    *                  last character successfully parsed. If the
    *                  source is not parsed successfully, this param
    *                  will remain unchanged.
    * @stable ICU 4.0
    */
   virtual void parseObject(const UnicodeString& source,
                            Formattable& result,
                            ParsePosition& parse_pos) const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 4.0
     *
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 4.0
     */
     virtual UClassID getDynamicClassID() const;

private:
     /**
      * @internal (private)
      */
    class U_I18N_API PluralSelector : public UMemory {
      public:
        virtual ~PluralSelector();
        /**
         * Given a number, returns the appropriate PluralFormat keyword.
         *
         * @param context worker object for the selector.
         * @param number The number to be plural-formatted.
         * @param ec Error code.
         * @return The selected PluralFormat keyword.
         * @internal (private)
         */
        virtual UnicodeString select(void *context, double number, UErrorCode& ec) const = 0;
    };

    class U_I18N_API PluralSelectorAdapter : public PluralSelector {
      public:
        PluralSelectorAdapter() : pluralRules(NULL) {
        }

        virtual ~PluralSelectorAdapter();

        virtual UnicodeString select(void *context, double number, UErrorCode& /*ec*/) const;

        void reset();

        PluralRules* pluralRules;
    };

    Locale  locale;
    MessagePattern msgPattern;
    NumberFormat*  numberFormat;
    double offset;
    PluralSelectorAdapter pluralRulesWrapper;

    PluralFormat();   // default constructor not implemented
    void init(const PluralRules* rules, UPluralType type, UErrorCode& status);
    /**
     * Copies dynamically allocated values (pointer fields).
     * Others are copied using their copy constructors and assignment operators.
     */
    void copyObjects(const PluralFormat& other);

    UnicodeString& format(const Formattable& numberObject, double number,
                          UnicodeString& appendTo,
                          FieldPosition& pos,
                          UErrorCode& status) const;

    /**
     * Finds the PluralFormat sub-message for the given number, or the "other" sub-message.
     * @param pattern A MessagePattern.
     * @param partIndex the index of the first PluralFormat argument style part.
     * @param selector the PluralSelector for mapping the number (minus offset) to a keyword.
     * @param context worker object for the selector.
     * @param number a number to be matched to one of the PluralFormat argument's explicit values,
     *        or mapped via the PluralSelector.
     * @param ec ICU error code.
     * @return the sub-message start part index.
     */
    static int32_t findSubMessage(
         const MessagePattern& pattern, int32_t partIndex,
         const PluralSelector& selector, void *context, double number, UErrorCode& ec); /**< @internal */

    void parseType(const UnicodeString& source, const NFRule *rbnfLenientScanner,
        Formattable& result, FieldPosition& pos) const;

    friend class MessageFormat;
    friend class NFRule;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */




U_CDECL_BEGIN
// Forward declaration.
struct UHashtable;
typedef struct UHashtable UHashtable; /**< @internal */
U_CDECL_END

U_NAMESPACE_BEGIN

class AppendableWrapper;
class DateFormat;
class NumberFormat;

/**
 * <p>MessageFormat prepares strings for display to users,
 * with optional arguments (variables/placeholders).
 * The arguments can occur in any order, which is necessary for translation
 * into languages with different grammars.
 *
 * <p>A MessageFormat is constructed from a <em>pattern</em> string
 * with arguments in {curly braces} which will be replaced by formatted values.
 *
 * <p><code>MessageFormat</code> differs from the other <code>Format</code>
 * classes in that you create a <code>MessageFormat</code> object with one
 * of its constructors (not with a <code>createInstance</code> style factory
 * method). Factory methods aren't necessary because <code>MessageFormat</code>
 * itself doesn't implement locale-specific behavior. Any locale-specific
 * behavior is defined by the pattern that you provide and the
 * subformats used for inserted arguments.
 *
 * <p>Arguments can be named (using identifiers) or numbered (using small ASCII-digit integers).
 * Some of the API methods work only with argument numbers and throw an exception
 * if the pattern has named arguments (see {@link #usesNamedArguments()}).
 *
 * <p>An argument might not specify any format type. In this case,
 * a numeric value is formatted with a default (for the locale) NumberFormat,
 * and a date/time value is formatted with a default (for the locale) DateFormat.
 *
 * <p>An argument might specify a "simple" type for which the specified
 * Format object is created, cached and used.
 *
 * <p>An argument might have a "complex" type with nested MessageFormat sub-patterns.
 * During formatting, one of these sub-messages is selected according to the argument value
 * and recursively formatted.
 *
 * <p>After construction, a custom Format object can be set for
 * a top-level argument, overriding the default formatting and parsing behavior
 * for that argument.
 * However, custom formatting can be achieved more simply by writing
 * a typeless argument in the pattern string
 * and supplying it with a preformatted string value.
 *
 * <p>When formatting, MessageFormat takes a collection of argument values
 * and writes an output string.
 * The argument values may be passed as an array
 * (when the pattern contains only numbered arguments)
 * or as an array of names and and an array of arguments (which works for both named
 * and numbered arguments).
 *
 * <p>Each argument is matched with one of the input values by array index or argument name
 * and formatted according to its pattern specification
 * (or using a custom Format object if one was set).
 * A numbered pattern argument is matched with an argument name that contains that number
 * as an ASCII-decimal-digit string (without leading zero).
 *
 * <h4><a name="patterns">Patterns and Their Interpretation</a></h4>
 *
 * <code>MessageFormat</code> uses patterns of the following form:
 * <pre>
 * message = messageText (argument messageText)*
 * argument = noneArg | simpleArg | complexArg
 * complexArg = choiceArg | pluralArg | selectArg | selectordinalArg
 *
 * noneArg = '{' argNameOrNumber '}'
 * simpleArg = '{' argNameOrNumber ',' argType [',' argStyle] '}'
 * choiceArg = '{' argNameOrNumber ',' "choice" ',' choiceStyle '}'
 * pluralArg = '{' argNameOrNumber ',' "plural" ',' pluralStyle '}'
 * selectArg = '{' argNameOrNumber ',' "select" ',' selectStyle '}'
 * selectordinalArg = '{' argNameOrNumber ',' "selectordinal" ',' pluralStyle '}'
 *
 * choiceStyle: see {@link ChoiceFormat}
 * pluralStyle: see {@link PluralFormat}
 * selectStyle: see {@link SelectFormat}
 *
 * argNameOrNumber = argName | argNumber
 * argName = [^[[:Pattern_Syntax:][:Pattern_White_Space:]]]+
 * argNumber = '0' | ('1'..'9' ('0'..'9')*)
 *
 * argType = "number" | "date" | "time" | "spellout" | "ordinal" | "duration"
 * argStyle = "short" | "medium" | "long" | "full" | "integer" | "currency" | "percent" | argStyleText | "::" argSkeletonText
 * </pre>
 *
 * <ul>
 *   <li>messageText can contain quoted literal strings including syntax characters.
 *       A quoted literal string begins with an ASCII apostrophe and a syntax character
 *       (usually a {curly brace}) and continues until the next single apostrophe.
 *       A double ASCII apostrohpe inside or outside of a quoted string represents
 *       one literal apostrophe.
 *   <li>Quotable syntax characters are the {curly braces} in all messageText parts,
 *       plus the '#' sign in a messageText immediately inside a pluralStyle,
 *       and the '|' symbol in a messageText immediately inside a choiceStyle.
 *   <li>See also {@link #UMessagePatternApostropheMode}
 *   <li>In argStyleText, every single ASCII apostrophe begins and ends quoted literal text,
 *       and unquoted {curly braces} must occur in matched pairs.
 * </ul>
 *
 * <p>Recommendation: Use the real apostrophe (single quote) character
 * \htmlonly&#x2019;\endhtmlonly (U+2019) for
 * human-readable text, and use the ASCII apostrophe ' (U+0027)
 * only in program syntax, like quoting in MessageFormat.
 * See the annotations for U+0027 Apostrophe in The Unicode Standard.
 *
 * <p>The <code>choice</code> argument type is deprecated.
 * Use <code>plural</code> arguments for proper plural selection,
 * and <code>select</code> arguments for simple selection among a fixed set of choices.
 *
 * <p>The <code>argType</code> and <code>argStyle</code> values are used to create
 * a <code>Format</code> instance for the format element. The following
 * table shows how the values map to Format instances. Combinations not
 * shown in the table are illegal. Any <code>argStyleText</code> must
 * be a valid pattern string for the Format subclass used.
 *
 * <p><table border=1>
 *    <tr>
 *       <th>argType
 *       <th>argStyle
 *       <th>resulting Format object
 *    <tr>
 *       <td colspan=2><i>(none)</i>
 *       <td><code>null</code>
 *    <tr>
 *       <td rowspan=6><code>number</code>
 *       <td><i>(none)</i>
 *       <td><code>NumberFormat.createInstance(getLocale(), status)</code>
 *    <tr>
 *       <td><code>integer</code>
 *       <td><code>NumberFormat.createInstance(getLocale(), kNumberStyle, status)</code>
 *    <tr>
 *       <td><code>currency</code>
 *       <td><code>NumberFormat.createCurrencyInstance(getLocale(), status)</code>
 *    <tr>
 *       <td><code>percent</code>
 *       <td><code>NumberFormat.createPercentInstance(getLocale(), status)</code>
 *    <tr>
 *       <td><i>argStyleText</i>
 *       <td><code>new DecimalFormat(argStyleText, new DecimalFormatSymbols(getLocale(), status), status)</code>
 *    <tr>
 *       <td><i>argSkeletonText</i>
 *       <td><code>NumberFormatter::forSkeleton(argSkeletonText, status).locale(getLocale()).toFormat(status)</code>
 *    <tr>
 *       <td rowspan=7><code>date</code>
 *       <td><i>(none)</i>
 *       <td><code>DateFormat.createDateInstance(kDefault, getLocale(), status)</code>
 *    <tr>
 *       <td><code>short</code>
 *       <td><code>DateFormat.createDateInstance(kShort, getLocale(), status)</code>
 *    <tr>
 *       <td><code>medium</code>
 *       <td><code>DateFormat.createDateInstance(kDefault, getLocale(), status)</code>
 *    <tr>
 *       <td><code>long</code>
 *       <td><code>DateFormat.createDateInstance(kLong, getLocale(), status)</code>
 *    <tr>
 *       <td><code>full</code>
 *       <td><code>DateFormat.createDateInstance(kFull, getLocale(), status)</code>
 *    <tr>
 *       <td><i>argStyleText</i>
 *       <td><code>new SimpleDateFormat(argStyleText, getLocale(), status)</code>
 *    <tr>
 *       <td><i>argSkeletonText</i>
 *       <td><code>DateFormat::createInstanceForSkeleton(argSkeletonText, getLocale(), status)</code>
 *    <tr>
 *       <td rowspan=6><code>time</code>
 *       <td><i>(none)</i>
 *       <td><code>DateFormat.createTimeInstance(kDefault, getLocale(), status)</code>
 *    <tr>
 *       <td><code>short</code>
 *       <td><code>DateFormat.createTimeInstance(kShort, getLocale(), status)</code>
 *    <tr>
 *       <td><code>medium</code>
 *       <td><code>DateFormat.createTimeInstance(kDefault, getLocale(), status)</code>
 *    <tr>
 *       <td><code>long</code>
 *       <td><code>DateFormat.createTimeInstance(kLong, getLocale(), status)</code>
 *    <tr>
 *       <td><code>full</code>
 *       <td><code>DateFormat.createTimeInstance(kFull, getLocale(), status)</code>
 *    <tr>
 *       <td><i>argStyleText</i>
 *       <td><code>new SimpleDateFormat(argStyleText, getLocale(), status)</code>
 *    <tr>
 *       <td><code>spellout</code>
 *       <td><i>argStyleText (optional)</i>
 *       <td><code>new RuleBasedNumberFormat(URBNF_SPELLOUT, getLocale(), status)
 *           <br/>&nbsp;&nbsp;&nbsp;&nbsp;.setDefaultRuleset(argStyleText, status);</code>
 *    <tr>
 *       <td><code>ordinal</code>
 *       <td><i>argStyleText (optional)</i>
 *       <td><code>new RuleBasedNumberFormat(URBNF_ORDINAL, getLocale(), status)
 *           <br/>&nbsp;&nbsp;&nbsp;&nbsp;.setDefaultRuleset(argStyleText, status);</code>
 *    <tr>
 *       <td><code>duration</code>
 *       <td><i>argStyleText (optional)</i>
 *       <td><code>new RuleBasedNumberFormat(URBNF_DURATION, getLocale(), status)
 *           <br/>&nbsp;&nbsp;&nbsp;&nbsp;.setDefaultRuleset(argStyleText, status);</code>
 * </table>
 * <p>
 *
 * <h4>Argument formatting</h4>
 *
 * <p>Arguments are formatted according to their type, using the default
 * ICU formatters for those types, unless otherwise specified.</p>
 *
 * <p>There are also several ways to control the formatting.</p>
 *
 * <p>We recommend you use default styles, predefined style values, skeletons,
 * or preformatted values, but not pattern strings or custom format objects.</p>
 *
 * <p>For more details, see the
 * <a href="http://userguide.icu-project.org/formatparse/messages">ICU User Guide</a>.</p>
 *
 * <h4>Usage Information</h4>
 *
 * <p>Here are some examples of usage:
 * Example 1:
 *
 * <pre>
 * \code
 *     UErrorCode success = U_ZERO_ERROR;
 *     GregorianCalendar cal(success);
 *     Formattable arguments[] = {
 *         7L,
 *         Formattable( (Date) cal.getTime(success), Formattable::kIsDate),
 *         "a disturbance in the Force"
 *     };
 *
 *     UnicodeString result;
 *     MessageFormat::format(
 *          "At {1,time,::jmm} on {1,date,::dMMMM}, there was {2} on planet {0,number}.",
 *          arguments, 3, result, success );
 *
 *     cout << "result: " << result << endl;
 *     //<output>: At 4:34 PM on March 23, there was a disturbance
 *     //             in the Force on planet 7.
 * \endcode
 * </pre>
 *
 * Typically, the message format will come from resources, and the
 * arguments will be dynamically set at runtime.
 *
 * <p>Example 2:
 *
 * <pre>
 *  \code
 *     success = U_ZERO_ERROR;
 *     Formattable testArgs[] = {3L, "MyDisk"};
 *
 *     MessageFormat form(
 *         "The disk \"{1}\" contains {0} file(s).", success );
 *
 *     UnicodeString string;
 *     FieldPosition fpos = 0;
 *     cout << "format: " << form.format(testArgs, 2, string, fpos, success ) << endl;
 *
 *     // output, with different testArgs:
 *     // output: The disk "MyDisk" contains 0 file(s).
 *     // output: The disk "MyDisk" contains 1 file(s).
 *     // output: The disk "MyDisk" contains 1,273 file(s).
 *  \endcode
 *  </pre>
 *
 *
 * <p>For messages that include plural forms, you can use a plural argument:
 * <pre>
 * \code
 *  success = U_ZERO_ERROR;
 *  MessageFormat msgFmt(
 *       "{num_files, plural, "
 *       "=0{There are no files on disk \"{disk_name}\".}"
 *       "=1{There is one file on disk \"{disk_name}\".}"
 *       "other{There are # files on disk \"{disk_name}\".}}",
 *      Locale("en"),
 *      success);
 *  FieldPosition fpos = 0;
 *  Formattable testArgs[] = {0L, "MyDisk"};
 *  UnicodeString testArgsNames[] = {"num_files", "disk_name"};
 *  UnicodeString result;
 *  cout << msgFmt.format(testArgs, testArgsNames, 2, result, fpos, 0, success);
 *  testArgs[0] = 3L;
 *  cout << msgFmt.format(testArgs, testArgsNames, 2, result, fpos, 0, success);
 * \endcode
 * <em>output</em>:
 * There are no files on disk "MyDisk".
 * There are 3 files on "MyDisk".
 * </pre>
 * See {@link PluralFormat} and {@link PluralRules} for details.
 *
 * <h4><a name="synchronization">Synchronization</a></h4>
 *
 * <p>MessageFormats are not synchronized.
 * It is recommended to create separate format instances for each thread.
 * If multiple threads access a format concurrently, it must be synchronized
 * externally.
 *
 * @stable ICU 2.0
 */
class U_I18N_API MessageFormat : public Format {
public:
#ifndef U_HIDE_OBSOLETE_API
    /**
     * Enum type for kMaxFormat.
     * @obsolete ICU 3.0.  The 10-argument limit was removed as of ICU 2.6,
     * rendering this enum type obsolete.
     */
    enum EFormatNumber {
        /**
         * The maximum number of arguments.
         * @obsolete ICU 3.0.  The 10-argument limit was removed as of ICU 2.6,
         * rendering this constant obsolete.
         */
        kMaxFormat = 10
    };
#endif  /* U_HIDE_OBSOLETE_API */

    /**
     * Constructs a new MessageFormat using the given pattern and the
     * default locale.
     *
     * @param pattern   Pattern used to construct object.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @stable ICU 2.0
     */
    MessageFormat(const UnicodeString& pattern,
                  UErrorCode &status);

    /**
     * Constructs a new MessageFormat using the given pattern and locale.
     * @param pattern   Pattern used to construct object.
     * @param newLocale The locale to use for formatting dates and numbers.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @stable ICU 2.0
     */
    MessageFormat(const UnicodeString& pattern,
                  const Locale& newLocale,
                        UErrorCode& status);
    /**
     * Constructs a new MessageFormat using the given pattern and locale.
     * @param pattern   Pattern used to construct object.
     * @param newLocale The locale to use for formatting dates and numbers.
     * @param parseError Struct to receive information on the position
     *                   of an error within the pattern.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @stable ICU 2.0
     */
    MessageFormat(const UnicodeString& pattern,
                  const Locale& newLocale,
                  UParseError& parseError,
                  UErrorCode& status);
    /**
     * Constructs a new MessageFormat from an existing one.
     * @stable ICU 2.0
     */
    MessageFormat(const MessageFormat&);

    /**
     * Assignment operator.
     * @stable ICU 2.0
     */
    const MessageFormat& operator=(const MessageFormat&);

    /**
     * Destructor.
     * @stable ICU 2.0
     */
    virtual ~MessageFormat();

    /**
     * Clones this Format object polymorphically.  The caller owns the
     * result and should delete it when done.
     * @stable ICU 2.0
     */
    virtual MessageFormat* clone() const;

    /**
     * Returns true if the given Format objects are semantically equal.
     * Objects of different subclasses are considered unequal.
     * @param other  the object to be compared with.
     * @return       true if the given Format objects are semantically equal.
     * @stable ICU 2.0
     */
    virtual UBool operator==(const Format& other) const;

    /**
     * Sets the locale to be used for creating argument Format objects.
     * @param theLocale    the new locale value to be set.
     * @stable ICU 2.0
     */
    virtual void setLocale(const Locale& theLocale);

    /**
     * Gets the locale used for creating argument Format objects.
     * format information.
     * @return    the locale of the object.
     * @stable ICU 2.0
     */
    virtual const Locale& getLocale(void) const;

    /**
     * Applies the given pattern string to this message format.
     *
     * @param pattern   The pattern to be applied.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @stable ICU 2.0
     */
    virtual void applyPattern(const UnicodeString& pattern,
                              UErrorCode& status);
    /**
     * Applies the given pattern string to this message format.
     *
     * @param pattern    The pattern to be applied.
     * @param parseError Struct to receive information on the position
     *                   of an error within the pattern.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @stable ICU 2.0
     */
    virtual void applyPattern(const UnicodeString& pattern,
                             UParseError& parseError,
                             UErrorCode& status);

    /**
     * Sets the UMessagePatternApostropheMode and the pattern used by this message format.
     * Parses the pattern and caches Format objects for simple argument types.
     * Patterns and their interpretation are specified in the
     * <a href="#patterns">class description</a>.
     * <p>
     * This method is best used only once on a given object to avoid confusion about the mode,
     * and after constructing the object with an empty pattern string to minimize overhead.
     *
     * @param pattern    The pattern to be applied.
     * @param aposMode   The new apostrophe mode.
     * @param parseError Struct to receive information on the position
     *                   of an error within the pattern.
     *                   Can be NULL.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @stable ICU 4.8
     */
    virtual void applyPattern(const UnicodeString& pattern,
                              UMessagePatternApostropheMode aposMode,
                              UParseError* parseError,
                              UErrorCode& status);

    /**
     * @return this instance's UMessagePatternApostropheMode.
     * @stable ICU 4.8
     */
    UMessagePatternApostropheMode getApostropheMode() const {
        return msgPattern.getApostropheMode();
    }

    /**
     * Returns a pattern that can be used to recreate this object.
     *
     * @param appendTo  Output parameter to receive the pattern.
     *                  Result is appended to existing contents.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    virtual UnicodeString& toPattern(UnicodeString& appendTo) const;

    /**
     * Sets subformats.
     * See the class description about format numbering.
     * The caller should not delete the Format objects after this call.
     * <EM>The array formatsToAdopt is not itself adopted.</EM> Its
     * ownership is retained by the caller. If the call fails because
     * memory cannot be allocated, then the formats will be deleted
     * by this method, and this object will remain unchanged.
     *
     * <p>If this format uses named arguments, the new formats are discarded
     * and this format remains unchanged.
     *
     * @stable ICU 2.0
     * @param formatsToAdopt    the format to be adopted.
     * @param count             the size of the array.
     */
    virtual void adoptFormats(Format** formatsToAdopt, int32_t count);

    /**
     * Sets subformats.
     * See the class description about format numbering.
     * Each item in the array is cloned into the internal array.
     * If the call fails because memory cannot be allocated, then this
     * object will remain unchanged.
     *
     * <p>If this format uses named arguments, the new formats are discarded
     * and this format remains unchanged.
     *
     * @stable ICU 2.0
     * @param newFormats the new format to be set.
     * @param cnt        the size of the array.
     */
    virtual void setFormats(const Format** newFormats, int32_t cnt);


    /**
     * Sets one subformat.
     * See the class description about format numbering.
     * The caller should not delete the Format object after this call.
     * If the number is over the number of formats already set,
     * the item will be deleted and ignored.
     *
     * <p>If this format uses named arguments, the new format is discarded
     * and this format remains unchanged.
     *
     * @stable ICU 2.0
     * @param formatNumber     index of the subformat.
     * @param formatToAdopt    the format to be adopted.
     */
    virtual void adoptFormat(int32_t formatNumber, Format* formatToAdopt);

    /**
     * Sets one subformat.
     * See the class description about format numbering.
     * If the number is over the number of formats already set,
     * the item will be ignored.
     * @param formatNumber     index of the subformat.
     * @param format    the format to be set.
     * @stable ICU 2.0
     */
    virtual void setFormat(int32_t formatNumber, const Format& format);

    /**
     * Gets format names. This function returns formatNames in StringEnumerations
     * which can be used with getFormat() and setFormat() to export formattable
     * array from current MessageFormat to another.  It is the caller's responsibility
     * to delete the returned formatNames.
     * @param status  output param set to success/failure code.
     * @stable ICU 4.0
     */
    virtual StringEnumeration* getFormatNames(UErrorCode& status);

    /**
     * Gets subformat pointer for given format name.
     * This function supports both named and numbered
     * arguments. If numbered, the formatName is the
     * corresponding UnicodeStrings (e.g. "0", "1", "2"...).
     * The returned Format object should not be deleted by the caller,
     * nor should the ponter of other object .  The pointer and its
     * contents remain valid only until the next call to any method
     * of this class is made with this object.
     * @param formatName the name or number specifying a format
     * @param status  output param set to success/failure code.
     * @stable ICU 4.0
     */
    virtual Format* getFormat(const UnicodeString& formatName, UErrorCode& status);

    /**
     * Sets one subformat for given format name.
     * See the class description about format name.
     * This function supports both named and numbered
     * arguments-- if numbered, the formatName is the
     * corresponding UnicodeStrings (e.g. "0", "1", "2"...).
     * If there is no matched formatName or wrong type,
     * the item will be ignored.
     * @param formatName  Name of the subformat.
     * @param format      the format to be set.
     * @param status  output param set to success/failure code.
     * @stable ICU 4.0
     */
    virtual void setFormat(const UnicodeString& formatName, const Format& format, UErrorCode& status);

    /**
     * Sets one subformat for given format name.
     * See the class description about format name.
     * This function supports both named and numbered
     * arguments-- if numbered, the formatName is the
     * corresponding UnicodeStrings (e.g. "0", "1", "2"...).
     * If there is no matched formatName or wrong type,
     * the item will be ignored.
     * The caller should not delete the Format object after this call.
     * @param formatName  Name of the subformat.
     * @param formatToAdopt  Format to be adopted.
     * @param status      output param set to success/failure code.
     * @stable ICU 4.0
     */
    virtual void adoptFormat(const UnicodeString& formatName, Format* formatToAdopt, UErrorCode& status);

    /**
     * Gets an array of subformats of this object.  The returned array
     * should not be deleted by the caller, nor should the pointers
     * within the array.  The array and its contents remain valid only
     * until the next call to this format. See the class description
     * about format numbering.
     *
     * @param count output parameter to receive the size of the array
     * @return an array of count Format* objects, or NULL if out of
     * memory.  Any or all of the array elements may be NULL.
     * @stable ICU 2.0
     */
    virtual const Format** getFormats(int32_t& count) const;


    using Format::format;

    /**
     * Formats the given array of arguments into a user-readable string.
     * Does not take ownership of the Formattable* array or its contents.
     *
     * <p>If this format uses named arguments, appendTo is unchanged and
     * status is set to U_ILLEGAL_ARGUMENT_ERROR.
     *
     * @param source    An array of objects to be formatted.
     * @param count     The number of elements of 'source'.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param ignore    Not used; inherited from base class API.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    UnicodeString& format(const Formattable* source,
                          int32_t count,
                          UnicodeString& appendTo,
                          FieldPosition& ignore,
                          UErrorCode& status) const;

    /**
     * Formats the given array of arguments into a user-readable string
     * using the given pattern.
     *
     * <p>If this format uses named arguments, appendTo is unchanged and
     * status is set to U_ILLEGAL_ARGUMENT_ERROR.
     *
     * @param pattern   The pattern.
     * @param arguments An array of objects to be formatted.
     * @param count     The number of elements of 'source'.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    static UnicodeString& format(const UnicodeString& pattern,
                                 const Formattable* arguments,
                                 int32_t count,
                                 UnicodeString& appendTo,
                                 UErrorCode& status);

    /**
     * Formats the given array of arguments into a user-readable
     * string.  The array must be stored within a single Formattable
     * object of type kArray. If the Formattable object type is not of
     * type kArray, then returns a failing UErrorCode.
     *
     * <p>If this format uses named arguments, appendTo is unchanged and
     * status is set to U_ILLEGAL_ARGUMENT_ERROR.
     *
     * @param obj       A Formattable of type kArray containing
     *                  arguments to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 2.0
     */
    virtual UnicodeString& format(const Formattable& obj,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode& status) const;

    /**
     * Formats the given array of arguments into a user-defined argument name
     * array. This function supports both named and numbered
     * arguments-- if numbered, the formatName is the
     * corresponding UnicodeStrings (e.g. "0", "1", "2"...).
     *
     * @param argumentNames argument name array
     * @param arguments An array of objects to be formatted.
     * @param count     The number of elements of 'argumentNames' and
     *                  arguments.  The number of argumentNames and arguments
     *                  must be the same.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.0
     */
    UnicodeString& format(const UnicodeString* argumentNames,
                          const Formattable* arguments,
                          int32_t count,
                          UnicodeString& appendTo,
                          UErrorCode& status) const;
    /**
     * Parses the given string into an array of output arguments.
     *
     * @param source    String to be parsed.
     * @param pos       On input, starting position for parse. On output,
     *                  final position after parse.  Unchanged if parse
     *                  fails.
     * @param count     Output parameter to receive the number of arguments
     *                  parsed.
     * @return an array of parsed arguments.  The caller owns both
     * the array and its contents.
     * @stable ICU 2.0
     */
    virtual Formattable* parse(const UnicodeString& source,
                               ParsePosition& pos,
                               int32_t& count) const;

    /**
     * Parses the given string into an array of output arguments.
     *
     * <p>If this format uses named arguments, status is set to
     * U_ARGUMENT_TYPE_MISMATCH.
     *
     * @param source    String to be parsed.
     * @param count     Output param to receive size of returned array.
     * @param status    Input/output error code.  If the
     *                  pattern cannot be parsed, set to failure code.
     * @return an array of parsed arguments.  The caller owns both
     * the array and its contents. Returns NULL if status is not U_ZERO_ERROR.
     *
     * @stable ICU 2.0
     */
    virtual Formattable* parse(const UnicodeString& source,
                               int32_t& count,
                               UErrorCode& status) const;

    /**
     * Parses the given string into an array of output arguments
     * stored within a single Formattable of type kArray.
     *
     * @param source    The string to be parsed into an object.
     * @param result    Formattable to be set to the parse result.
     *                  If parse fails, return contents are undefined.
     * @param pos       On input, starting position for parse. On output,
     *                  final position after parse.  Unchanged if parse
     *                  fails.
     * @stable ICU 2.0
     */
    virtual void parseObject(const UnicodeString& source,
                             Formattable& result,
                             ParsePosition& pos) const;

    /**
     * Convert an 'apostrophe-friendly' pattern into a standard
     * pattern.  Standard patterns treat all apostrophes as
     * quotes, which is problematic in some languages, e.g.
     * French, where apostrophe is commonly used.  This utility
     * assumes that only an unpaired apostrophe immediately before
     * a brace is a true quote.  Other unpaired apostrophes are paired,
     * and the resulting standard pattern string is returned.
     *
     * <p><b>Note</b> it is not guaranteed that the returned pattern
     * is indeed a valid pattern.  The only effect is to convert
     * between patterns having different quoting semantics.
     *
     * @param pattern the 'apostrophe-friendly' patttern to convert
     * @param status    Input/output error code.  If the pattern
     *                  cannot be parsed, the failure code is set.
     * @return the standard equivalent of the original pattern
     * @stable ICU 3.4
     */
    static UnicodeString autoQuoteApostrophe(const UnicodeString& pattern,
        UErrorCode& status);


    /**
     * Returns true if this MessageFormat uses named arguments,
     * and false otherwise.  See class description.
     *
     * @return true if named arguments are used.
     * @stable ICU 4.0
     */
    UBool usesNamedArguments() const;


#ifndef U_HIDE_INTERNAL_API
    /**
     * This API is for ICU internal use only.
     * Please do not use it.
     *
     * Returns argument types count in the parsed pattern.
     * Used to distinguish pattern "{0} d" and "d".
     *
     * @return           The number of formattable types in the pattern
     * @internal
     */
    int32_t getArgTypeCount() const;
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Returns a unique class ID POLYMORPHICALLY.  Pure virtual override.
     * This method is to implement a simple version of RTTI, since not all
     * C++ compilers support genuine RTTI.  Polymorphic operator==() and
     * clone() methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 2.0
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Return the class ID for this class.  This is useful only for
     * comparing to a return value from getDynamicClassID().  For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .      Derived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 2.0
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

#ifndef U_HIDE_INTERNAL_API
    /**
     * Compares two Format objects. This is used for constructing the hash
     * tables.
     *
     * @param left pointer to a Format object. Must not be NULL.
     * @param right pointer to a Format object. Must not be NULL.
     *
     * @return whether the two objects are the same
     * @internal
     */
    static UBool equalFormats(const void* left, const void* right);
#endif  /* U_HIDE_INTERNAL_API */

private:

    Locale              fLocale;
    MessagePattern      msgPattern;
    Format**            formatAliases; // see getFormats
    int32_t             formatAliasesCapacity;

    MessageFormat(); // default constructor not implemented

     /**
      * This provider helps defer instantiation of a PluralRules object
      * until we actually need to select a keyword.
      * For example, if the number matches an explicit-value selector like "=1"
      * we do not need any PluralRules.
      */
    class U_I18N_API PluralSelectorProvider : public PluralFormat::PluralSelector {
    public:
        PluralSelectorProvider(const MessageFormat &mf, UPluralType type);
        virtual ~PluralSelectorProvider();
        virtual UnicodeString select(void *ctx, double number, UErrorCode& ec) const;

        void reset();
    private:
        const MessageFormat &msgFormat;
        PluralRules* rules;
        UPluralType type;
    };

    /**
     * A MessageFormat formats an array of arguments.  Each argument
     * has an expected type, based on the pattern.  For example, if
     * the pattern contains the subformat "{3,number,integer}", then
     * we expect argument 3 to have type Formattable::kLong.  This
     * array needs to grow dynamically if the MessageFormat is
     * modified.
     */
    Formattable::Type* argTypes;
    int32_t            argTypeCount;
    int32_t            argTypeCapacity;

    /**
     * TRUE if there are different argTypes for the same argument.
     * This only matters when the MessageFormat is used in the plain C (umsg_xxx) API
     * where the pattern argTypes determine how the va_arg list is read.
     */
    UBool hasArgTypeConflicts;

    // Variable-size array management
    UBool allocateArgTypes(int32_t capacity, UErrorCode& status);

    /**
     * Default Format objects used when no format is specified and a
     * numeric or date argument is formatted.  These are volatile
     * cache objects maintained only for performance.  They do not
     * participate in operator=(), copy constructor(), nor
     * operator==().
     */
    NumberFormat* defaultNumberFormat;
    DateFormat*   defaultDateFormat;

    UHashtable* cachedFormatters;
    UHashtable* customFormatArgStarts;

    PluralSelectorProvider pluralProvider;
    PluralSelectorProvider ordinalProvider;

    /**
     * Method to retrieve default formats (or NULL on failure).
     * These are semantically const, but may modify *this.
     */
    const NumberFormat* getDefaultNumberFormat(UErrorCode&) const;
    const DateFormat*   getDefaultDateFormat(UErrorCode&) const;

    /**
     * Finds the word s, in the keyword list and returns the located index.
     * @param s the keyword to be searched for.
     * @param list the list of keywords to be searched with.
     * @return the index of the list which matches the keyword s.
     */
    static int32_t findKeyword( const UnicodeString& s,
                                const char16_t * const *list);

    /**
     * Thin wrapper around the format(... AppendableWrapper ...) variant.
     * Wraps the destination UnicodeString into an AppendableWrapper and
     * supplies default values for some other parameters.
     */
    UnicodeString& format(const Formattable* arguments,
                          const UnicodeString *argumentNames,
                          int32_t cnt,
                          UnicodeString& appendTo,
                          FieldPosition* pos,
                          UErrorCode& status) const;

    /**
     * Formats the arguments and writes the result into the
     * AppendableWrapper, updates the field position.
     *
     * @param msgStart      Index to msgPattern part to start formatting from.
     * @param plNumber      NULL except when formatting a plural argument sub-message
     *                      where a '#' is replaced by the format string for this number.
     * @param arguments     The formattable objects array. (Must not be NULL.)
     * @param argumentNames NULL if numbered values are used. Otherwise the same
     *                      length as "arguments", and each entry is the name of the
     *                      corresponding argument in "arguments".
     * @param cnt           The length of arguments (and of argumentNames if that is not NULL).
     * @param appendTo      Output parameter to receive the result.
     *                      The result string is appended to existing contents.
     * @param pos           Field position status.
     * @param success       The error code status.
     */
    void format(int32_t msgStart,
                const void *plNumber,
                const Formattable* arguments,
                const UnicodeString *argumentNames,
                int32_t cnt,
                AppendableWrapper& appendTo,
                FieldPosition* pos,
                UErrorCode& success) const;

    UnicodeString getArgName(int32_t partIndex);

    void setArgStartFormat(int32_t argStart, Format* formatter, UErrorCode& status);

    void setCustomArgStartFormat(int32_t argStart, Format* formatter, UErrorCode& status);

    int32_t nextTopLevelArgStart(int32_t partIndex) const;

    UBool argNameMatches(int32_t partIndex, const UnicodeString& argName, int32_t argNumber);

    void cacheExplicitFormats(UErrorCode& status);

    int32_t skipLeadingSpaces(UnicodeString& style);

    Format* createAppropriateFormat(UnicodeString& type,
                                    UnicodeString& style,
                                    Formattable::Type& formattableType,
                                    UParseError& parseError,
                                    UErrorCode& ec);

    const Formattable* getArgFromListByName(const Formattable* arguments,
                                            const UnicodeString *argumentNames,
                                            int32_t cnt, UnicodeString& name) const;

    Formattable* parse(int32_t msgStart,
                       const UnicodeString& source,
                       ParsePosition& pos,
                       int32_t& count,
                       UErrorCode& ec) const;

    FieldPosition* updateMetaData(AppendableWrapper& dest, int32_t prevLength,
                                  FieldPosition* fp, const Formattable* argId) const;

    /**
     * Finds the "other" sub-message.
     * @param partIndex the index of the first PluralFormat argument style part.
     * @return the "other" sub-message start part index.
     */
    int32_t findOtherSubMessage(int32_t partIndex) const;

    /**
     * Returns the ARG_START index of the first occurrence of the plural number in a sub-message.
     * Returns -1 if it is a REPLACE_NUMBER.
     * Returns 0 if there is neither.
     */
    int32_t findFirstPluralNumberArg(int32_t msgStart, const UnicodeString &argName) const;

    Format* getCachedFormatter(int32_t argumentNumber) const;

    UnicodeString getLiteralStringUntilNextArgument(int32_t from) const;

    void copyObjects(const MessageFormat& that, UErrorCode& ec);

    void formatComplexSubMessage(int32_t msgStart,
                                 const void *plNumber,
                                 const Formattable* arguments,
                                 const UnicodeString *argumentNames,
                                 int32_t cnt,
                                 AppendableWrapper& appendTo,
                                 UErrorCode& success) const;

    /**
     * Convenience method that ought to be in NumberFormat
     */
    NumberFormat* createIntegerFormat(const Locale& locale, UErrorCode& status) const;

    /**
     * Returns array of argument types in the parsed pattern
     * for use in C API.  Only for the use of umsg_vformat().  Not
     * for public consumption.
     * @param listCount  Output parameter to receive the size of array
     * @return           The array of formattable types in the pattern
     */
    const Formattable::Type* getArgTypeList(int32_t& listCount) const {
        listCount = argTypeCount;
        return argTypes;
    }

    /**
     * Resets the internal MessagePattern, and other associated caches.
     */
    void resetPattern();

    /**
     * A DummyFormatter that we use solely to store a NULL value. UHash does
     * not support storing NULL values.
     */
    class U_I18N_API DummyFormat : public Format {
    public:
        virtual UBool operator==(const Format&) const;
        virtual DummyFormat* clone() const;
        virtual UnicodeString& format(const Formattable& obj,
                              UnicodeString& appendTo,
                              UErrorCode& status) const;
        virtual UnicodeString& format(const Formattable&,
                                      UnicodeString& appendTo,
                                      FieldPosition&,
                                      UErrorCode& status) const;
        virtual UnicodeString& format(const Formattable& obj,
                                      UnicodeString& appendTo,
                                      FieldPositionIterator* posIter,
                                      UErrorCode& status) const;
        virtual void parseObject(const UnicodeString&,
                                 Formattable&,
                                 ParsePosition&) const;
    };

    friend class MessageFormatAdapter; // getFormatTypeList() access
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */









// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File dt_impl.h
*
*******************************************************************************
*/



/**
 * \file
 * \brief C++ API: Defines macros for interval format implementation
 */

#if !UCONFIG_NO_FORMATTING




#define dt_impl_QUOTE             ((UChar)0x0027)
#define dt_impl_LOW_LINE          ((UChar)0x005F)
#define dt_impl_COLON             ((UChar)0x003A)
#define dt_impl_LEFT_CURLY_BRACKET  ((UChar)0x007B)
#define dt_impl_RIGHT_CURLY_BRACKET ((UChar)0x007D)
#define dt_impl_SPACE             ((UChar)0x0020)
#define dt_impl_EN_DASH           ((UChar)0x2013)
#define dt_impl_SOLIDUS           ((UChar)0x002F)
#define dt_impl_PERCENT           ((UChar)0x0025)

#define dt_impl_DIGIT_ZERO        ((UChar)0x0030)
#define dt_impl_DIGIT_ONE         ((UChar)0x0031)

#define dt_impl_LOW_A             ((UChar)0x0061)
#define dt_impl_LOW_B             ((UChar)0x0062)
#define dt_impl_LOW_C             ((UChar)0x0063)
#define dt_impl_LOW_D             ((UChar)0x0064)
#define dt_impl_LOW_E             ((UChar)0x0065)
#define dt_impl_LOW_F             ((UChar)0x0066)
#define dt_impl_LOW_G             ((UChar)0x0067)
#define dt_impl_LOW_H             ((UChar)0x0068)
#define dt_impl_LOW_I             ((UChar)0x0069)
#define dt_impl_LOW_J             ((UChar)0x006a)
#define dt_impl_LOW_K             ((UChar)0x006B)
#define dt_impl_LOW_L             ((UChar)0x006C)
#define dt_impl_LOW_M             ((UChar)0x006D)
#define dt_impl_LOW_N             ((UChar)0x006E)
#define dt_impl_LOW_O             ((UChar)0x006F)
#define dt_impl_LOW_P             ((UChar)0x0070)
#define dt_impl_LOW_Q             ((UChar)0x0071)
#define dt_impl_LOW_R             ((UChar)0x0072)
#define dt_impl_LOW_S             ((UChar)0x0073)
#define dt_impl_LOW_T             ((UChar)0x0074)
#define dt_impl_LOW_U             ((UChar)0x0075)
#define dt_impl_LOW_V             ((UChar)0x0076)
#define dt_impl_LOW_W             ((UChar)0x0077)
#define dt_impl_LOW_Y             ((UChar)0x0079)
#define dt_impl_LOW_Z             ((UChar)0x007A)

#define dt_impl_CAP_A             ((UChar)0x0041)
#define dt_impl_CAP_C             ((UChar)0x0043)
#define dt_impl_CAP_D             ((UChar)0x0044)
#define dt_impl_CAP_E             ((UChar)0x0045)
#define dt_impl_CAP_F             ((UChar)0x0046)
#define dt_impl_CAP_G             ((UChar)0x0047)
#define dt_impl_CAP_H             ((UChar)0x0048)
#define dt_impl_CAP_K             ((UChar)0x004B)
#define dt_impl_CAP_L             ((UChar)0x004C)
#define dt_impl_CAP_M             ((UChar)0x004D)
#define dt_impl_CAP_N             ((UChar)0x004E)
#define dt_impl_CAP_O             ((UChar)0x004F)
#define dt_impl_CAP_P             ((UChar)0x0050)
#define dt_impl_CAP_Q             ((UChar)0x0051)
#define dt_impl_CAP_S             ((UChar)0x0053)
#define dt_impl_CAP_T             ((UChar)0x0054)
#define dt_impl_CAP_U             ((UChar)0x0055)
#define dt_impl_CAP_V             ((UChar)0x0056)
#define dt_impl_CAP_W             ((UChar)0x0057)
#define dt_impl_CAP_Y             ((UChar)0x0059)
#define dt_impl_CAP_Z             ((UChar)0x005A)

#endif /* #if !UCONFIG_NO_FORMATTING */









// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2014, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* shareddateformatsymbols.h
*/




#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN


class U_I18N_API SharedDateFormatSymbols : public SharedObject {
public:
    SharedDateFormatSymbols(
            const Locale &loc, const char *type, UErrorCode &status)
            : dfs(loc, type, status) { }
    virtual ~SharedDateFormatSymbols();
    const DateFormatSymbols &get() const { return dfs; }
private:
    DateFormatSymbols dfs;
    SharedDateFormatSymbols(const SharedDateFormatSymbols &);
    SharedDateFormatSymbols &operator=(const SharedDateFormatSymbols &);
};

U_NAMESPACE_END

#endif /* !UCONFIG_NO_FORMATTING */





// *****************************************************************************
// class DateFormatSymbols
// *****************************************************************************

/**
 * These are static arrays we use only in the case where we have no
 * resource data.
 */

#if UDAT_HAS_PATTERN_CHAR_FOR_TIME_SEPARATOR
#define PATTERN_CHARS_LEN 38
#else
#define PATTERN_CHARS_LEN 37
#endif

/**
 * Unlocalized date-time pattern characters. For example: 'y', 'd', etc. All
 * locales use the same these unlocalized pattern characters.
 */
static const UChar gPatternChars[] = {
    // if UDAT_HAS_PATTERN_CHAR_FOR_TIME_SEPARATOR:
    //   GyMdkHmsSEDFwWahKzYeugAZvcLQqVUOXxrbB:
    // else:
    //   GyMdkHmsSEDFwWahKzYeugAZvcLQqVUOXxrbB

    0x47, 0x79, 0x4D, 0x64, 0x6B, 0x48, 0x6D, 0x73, 0x53, 0x45,
    0x44, 0x46, 0x77, 0x57, 0x61, 0x68, 0x4B, 0x7A, 0x59, 0x65,
    0x75, 0x67, 0x41, 0x5A, 0x76, 0x63, 0x4c, 0x51, 0x71, 0x56,
    0x55, 0x4F, 0x58, 0x78, 0x72, 0x62, 0x42,
#if UDAT_HAS_PATTERN_CHAR_FOR_TIME_SEPARATOR
    0x3a,
#endif
    0
};

//------------------------------------------------------
// Strings of last resort.  These are only used if we have no resource
// files.  They aren't designed for actual use, just for backup.

// These are the month names and abbreviations of last resort.
static const UChar gLastResortMonthNames[13][3] =
{
    {0x0030, 0x0031, 0x0000}, /* "01" */
    {0x0030, 0x0032, 0x0000}, /* "02" */
    {0x0030, 0x0033, 0x0000}, /* "03" */
    {0x0030, 0x0034, 0x0000}, /* "04" */
    {0x0030, 0x0035, 0x0000}, /* "05" */
    {0x0030, 0x0036, 0x0000}, /* "06" */
    {0x0030, 0x0037, 0x0000}, /* "07" */
    {0x0030, 0x0038, 0x0000}, /* "08" */
    {0x0030, 0x0039, 0x0000}, /* "09" */
    {0x0031, 0x0030, 0x0000}, /* "10" */
    {0x0031, 0x0031, 0x0000}, /* "11" */
    {0x0031, 0x0032, 0x0000}, /* "12" */
    {0x0031, 0x0033, 0x0000}  /* "13" */
};

// These are the weekday names and abbreviations of last resort.
static const UChar gLastResortDayNames[8][2] =
{
    {0x0030, 0x0000}, /* "0" */
    {0x0031, 0x0000}, /* "1" */
    {0x0032, 0x0000}, /* "2" */
    {0x0033, 0x0000}, /* "3" */
    {0x0034, 0x0000}, /* "4" */
    {0x0035, 0x0000}, /* "5" */
    {0x0036, 0x0000}, /* "6" */
    {0x0037, 0x0000}  /* "7" */
};

// These are the quarter names and abbreviations of last resort.
static const UChar gLastResortQuarters[4][2] =
{
    {0x0031, 0x0000}, /* "1" */
    {0x0032, 0x0000}, /* "2" */
    {0x0033, 0x0000}, /* "3" */
    {0x0034, 0x0000}, /* "4" */
};

// These are the am/pm and BC/AD markers of last resort.
static const UChar gLastResortAmPmMarkers[2][3] =
{
    {0x0041, 0x004D, 0x0000}, /* "AM" */
    {0x0050, 0x004D, 0x0000}  /* "PM" */
};

static const UChar gLastResortEras[2][3] =
{
    {0x0042, 0x0043, 0x0000}, /* "BC" */
    {0x0041, 0x0044, 0x0000}  /* "AD" */
};

/* Sizes for the last resort string arrays */
typedef enum LastResortSize {
    kMonthNum = 13,
    kMonthLen = 3,

    kDayNum = 8,
    kDayLen = 2,

    kAmPmNum = 2,
    kAmPmLen = 3,

    kQuarterNum = 4,
    kQuarterLen = 2,

    kEraNum = 2,
    kEraLen = 3,

    kZoneNum = 5,
    kZoneLen = 4,

    kGmtHourNum = 4,
    kGmtHourLen = 10
} LastResortSize;

U_NAMESPACE_BEGIN

SharedDateFormatSymbols::~SharedDateFormatSymbols() {
}

template<> U_I18N_API
const SharedDateFormatSymbols *
        LocaleCacheKey<SharedDateFormatSymbols>::createObject(
                const void * /*unusedContext*/, UErrorCode &status) const {
    char type[256];
    Calendar::getCalendarTypeFromLocale(fLoc, type, UPRV_LENGTHOF(type), status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    SharedDateFormatSymbols *shared
            = new SharedDateFormatSymbols(fLoc, type, status);
    if (shared == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    if (U_FAILURE(status)) {
        delete shared;
        return NULL;
    }
    shared->addRef();
    return shared;
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DateFormatSymbols)

#define kSUPPLEMENTAL "supplementalData"

/**
 * These are the tags we expect to see in normal resource bundle files associated
 * with a locale and calendar
 */
static const char dtfmtsym_gCalendarTag[]="calendar";
static const char dtfmtsym_gGregorianTag[]="gregorian";
static const char dtfmtsym_gErasTag[]="eras";
static const char dtfmtsym_gCyclicNameSetsTag[]="cyclicNameSets";
static const char dtfmtsym_gNameSetYearsTag[]="years";
static const char dtfmtsym_gNameSetZodiacsTag[]="zodiacs";
static const char dtfmtsym_gMonthNamesTag[]="monthNames";
static const char dtfmtsym_gMonthPatternsTag[]="monthPatterns";
static const char dtfmtsym_gDayNamesTag[]="dayNames";
static const char dtfmtsym_gNamesWideTag[]="wide";
static const char dtfmtsym_gNamesAbbrTag[]="abbreviated";
static const char dtfmtsym_gNamesShortTag[]="short";
static const char dtfmtsym_gNamesNarrowTag[]="narrow";
static const char dtfmtsym_gNamesAllTag[]="all";
static const char dtfmtsym_gNamesFormatTag[]="format";
static const char dtfmtsym_gNamesStandaloneTag[]="stand-alone";
static const char dtfmtsym_gNamesNumericTag[]="numeric";
static const char dtfmtsym_gAmPmMarkersTag[]="AmPmMarkers";
static const char dtfmtsym_gAmPmMarkersAbbrTag[]="AmPmMarkersAbbr";
static const char dtfmtsym_gAmPmMarkersNarrowTag[]="AmPmMarkersNarrow";
static const char dtfmtsym_gQuartersTag[]="quarters";
static const char dtfmtsym_gNumberElementsTag[]="NumberElements";
static const char dtfmtsym_gSymbolsTag[]="symbols";
static const char dtfmtsym_gTimeSeparatorTag[]="timeSeparator";
static const char dtfmtsym_gDayPeriodTag[]="dayPeriod";

// static const char gZoneStringsTag[]="zoneStrings";

// static const char gLocalPatternCharsTag[]="localPatternChars";

static const char gContextTransformsTag[]="contextTransforms";

/**
 * Jitterbug 2974: MSVC has a bug whereby new X[0] behaves badly.
 * Work around this.
 */
static inline UnicodeString* newUnicodeStringArray(size_t count) {
    return new UnicodeString[count ? count : 1];
}

//------------------------------------------------------

DateFormatSymbols * U_EXPORT2
DateFormatSymbols::createForLocale(
        const Locale& locale, UErrorCode &status) {
    const SharedDateFormatSymbols *shared = NULL;
    UnifiedCache::getByLocale(locale, shared, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    DateFormatSymbols *result = new DateFormatSymbols(shared->get());
    shared->removeRef();
    if (result == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    return result;
}

DateFormatSymbols::DateFormatSymbols(const Locale& locale,
                                     UErrorCode& status)
    : UObject()
{
  initializeData(locale, NULL,  status);
}

DateFormatSymbols::DateFormatSymbols(UErrorCode& status)
    : UObject()
{
  initializeData(Locale::getDefault(), NULL, status, TRUE);
}


DateFormatSymbols::DateFormatSymbols(const Locale& locale,
                                     const char *type,
                                     UErrorCode& status)
    : UObject()
{
  initializeData(locale, type,  status);
}

DateFormatSymbols::DateFormatSymbols(const char *type, UErrorCode& status)
    : UObject()
{
  initializeData(Locale::getDefault(), type, status, TRUE);
}

DateFormatSymbols::DateFormatSymbols(const DateFormatSymbols& other)
    : UObject(other)
{
    copyData(other);
}

void
DateFormatSymbols::assignArray(UnicodeString*& dstArray,
                               int32_t& dstCount,
                               const UnicodeString* srcArray,
                               int32_t srcCount)
{
    // assignArray() is only called by copyData() and initializeData(), which in turn
    // implements the copy constructor and the assignment operator.
    // All strings in a DateFormatSymbols object are created in one of the following
    // three ways that all allow to safely use UnicodeString::fastCopyFrom():
    // - readonly-aliases from resource bundles
    // - readonly-aliases or allocated strings from constants
    // - safely cloned strings (with owned buffers) from setXYZ() functions
    //
    // Note that this is true for as long as DateFormatSymbols can be constructed
    // only from a locale bundle or set via the cloning API,
    // *and* for as long as all the strings are in *private* fields, preventing
    // a subclass from creating these strings in an "unsafe" way (with respect to fastCopyFrom()).
    dstCount = srcCount;
    dstArray = newUnicodeStringArray(srcCount);
    if(dstArray != NULL) {
        int32_t i;
        for(i=0; i<srcCount; ++i) {
            dstArray[i].fastCopyFrom(srcArray[i]);
        }
    }
}

/**
 * Create a copy, in fZoneStrings, of the given zone strings array.  The
 * member variables fZoneStringsRowCount and fZoneStringsColCount should
 * be set already by the caller.
 */
void
DateFormatSymbols::createZoneStrings(const UnicodeString *const * otherStrings)
{
    int32_t row, col;
    UBool failed = FALSE;

    fZoneStrings = (UnicodeString **)uprv_malloc(fZoneStringsRowCount * sizeof(UnicodeString *));
    if (fZoneStrings != NULL) {
        for (row=0; row<fZoneStringsRowCount; ++row)
        {
            fZoneStrings[row] = newUnicodeStringArray(fZoneStringsColCount);
            if (fZoneStrings[row] == NULL) {
                failed = TRUE;
                break;
            }
            for (col=0; col<fZoneStringsColCount; ++col) {
                // fastCopyFrom() - see assignArray comments
                fZoneStrings[row][col].fastCopyFrom(otherStrings[row][col]);
            }
        }
    }
    // If memory allocation failed, roll back and delete fZoneStrings
    if (failed) {
        for (int i = row; i >= 0; i--) {
            delete[] fZoneStrings[i];
        }
        uprv_free(fZoneStrings);
        fZoneStrings = NULL;
    }
}

/**
 * Copy all of the other's data to this.
 */
void
DateFormatSymbols::copyData(const DateFormatSymbols& other) {
    UErrorCode status = U_ZERO_ERROR;
    U_LOCALE_BASED(locBased, *this);
    locBased.setLocaleIDs(
        other.getLocale(ULOC_VALID_LOCALE, status),
        other.getLocale(ULOC_ACTUAL_LOCALE, status));
    assignArray(fEras, fErasCount, other.fEras, other.fErasCount);
    assignArray(fEraNames, fEraNamesCount, other.fEraNames, other.fEraNamesCount);
    assignArray(fNarrowEras, fNarrowErasCount, other.fNarrowEras, other.fNarrowErasCount);
    assignArray(fMonths, fMonthsCount, other.fMonths, other.fMonthsCount);
    assignArray(fShortMonths, fShortMonthsCount, other.fShortMonths, other.fShortMonthsCount);
    assignArray(fNarrowMonths, fNarrowMonthsCount, other.fNarrowMonths, other.fNarrowMonthsCount);
    assignArray(fStandaloneMonths, fStandaloneMonthsCount, other.fStandaloneMonths, other.fStandaloneMonthsCount);
    assignArray(fStandaloneShortMonths, fStandaloneShortMonthsCount, other.fStandaloneShortMonths, other.fStandaloneShortMonthsCount);
    assignArray(fStandaloneNarrowMonths, fStandaloneNarrowMonthsCount, other.fStandaloneNarrowMonths, other.fStandaloneNarrowMonthsCount);
    assignArray(fWeekdays, fWeekdaysCount, other.fWeekdays, other.fWeekdaysCount);
    assignArray(fShortWeekdays, fShortWeekdaysCount, other.fShortWeekdays, other.fShortWeekdaysCount);
    assignArray(fShorterWeekdays, fShorterWeekdaysCount, other.fShorterWeekdays, other.fShorterWeekdaysCount);
    assignArray(fNarrowWeekdays, fNarrowWeekdaysCount, other.fNarrowWeekdays, other.fNarrowWeekdaysCount);
    assignArray(fStandaloneWeekdays, fStandaloneWeekdaysCount, other.fStandaloneWeekdays, other.fStandaloneWeekdaysCount);
    assignArray(fStandaloneShortWeekdays, fStandaloneShortWeekdaysCount, other.fStandaloneShortWeekdays, other.fStandaloneShortWeekdaysCount);
    assignArray(fStandaloneShorterWeekdays, fStandaloneShorterWeekdaysCount, other.fStandaloneShorterWeekdays, other.fStandaloneShorterWeekdaysCount);
    assignArray(fStandaloneNarrowWeekdays, fStandaloneNarrowWeekdaysCount, other.fStandaloneNarrowWeekdays, other.fStandaloneNarrowWeekdaysCount);
    assignArray(fAmPms, fAmPmsCount, other.fAmPms, other.fAmPmsCount);
    assignArray(fNarrowAmPms, fNarrowAmPmsCount, other.fNarrowAmPms, other.fNarrowAmPmsCount );
    fTimeSeparator.fastCopyFrom(other.fTimeSeparator);  // fastCopyFrom() - see assignArray comments
    assignArray(fQuarters, fQuartersCount, other.fQuarters, other.fQuartersCount);
    assignArray(fShortQuarters, fShortQuartersCount, other.fShortQuarters, other.fShortQuartersCount);
    assignArray(fStandaloneQuarters, fStandaloneQuartersCount, other.fStandaloneQuarters, other.fStandaloneQuartersCount);
    assignArray(fStandaloneShortQuarters, fStandaloneShortQuartersCount, other.fStandaloneShortQuarters, other.fStandaloneShortQuartersCount);
    assignArray(fWideDayPeriods, fWideDayPeriodsCount,
                other.fWideDayPeriods, other.fWideDayPeriodsCount);
    assignArray(fNarrowDayPeriods, fNarrowDayPeriodsCount,
                other.fNarrowDayPeriods, other.fNarrowDayPeriodsCount);
    assignArray(fAbbreviatedDayPeriods, fAbbreviatedDayPeriodsCount,
                other.fAbbreviatedDayPeriods, other.fAbbreviatedDayPeriodsCount);
    assignArray(fStandaloneWideDayPeriods, fStandaloneWideDayPeriodsCount,
                other.fStandaloneWideDayPeriods, other.fStandaloneWideDayPeriodsCount);
    assignArray(fStandaloneNarrowDayPeriods, fStandaloneNarrowDayPeriodsCount,
                other.fStandaloneNarrowDayPeriods, other.fStandaloneNarrowDayPeriodsCount);
    assignArray(fStandaloneAbbreviatedDayPeriods, fStandaloneAbbreviatedDayPeriodsCount,
                other.fStandaloneAbbreviatedDayPeriods, other.fStandaloneAbbreviatedDayPeriodsCount);
    if (other.fLeapMonthPatterns != NULL) {
        assignArray(fLeapMonthPatterns, fLeapMonthPatternsCount, other.fLeapMonthPatterns, other.fLeapMonthPatternsCount);
    } else {
        fLeapMonthPatterns = NULL;
        fLeapMonthPatternsCount = 0;
    }
    if (other.fShortYearNames != NULL) {
        assignArray(fShortYearNames, fShortYearNamesCount, other.fShortYearNames, other.fShortYearNamesCount);
    } else {
        fShortYearNames = NULL;
        fShortYearNamesCount = 0;
    }
    if (other.fShortZodiacNames != NULL) {
        assignArray(fShortZodiacNames, fShortZodiacNamesCount, other.fShortZodiacNames, other.fShortZodiacNamesCount);
    } else {
        fShortZodiacNames = NULL;
        fShortZodiacNamesCount = 0;
    }

    if (other.fZoneStrings != NULL) {
        fZoneStringsColCount = other.fZoneStringsColCount;
        fZoneStringsRowCount = other.fZoneStringsRowCount;
        createZoneStrings((const UnicodeString**)other.fZoneStrings);

    } else {
        fZoneStrings = NULL;
        fZoneStringsColCount = 0;
        fZoneStringsRowCount = 0;
    }
    fZSFLocale = other.fZSFLocale;
    // Other zone strings data is created on demand
    fLocaleZoneStrings = NULL;

    // fastCopyFrom() - see assignArray comments
    fLocalPatternChars.fastCopyFrom(other.fLocalPatternChars);

    uprv_memcpy(fCapitalization, other.fCapitalization, sizeof(fCapitalization));
}

/**
 * Assignment operator.
 */
DateFormatSymbols& DateFormatSymbols::operator=(const DateFormatSymbols& other)
{
    dispose();
    copyData(other);

    return *this;
}

DateFormatSymbols::~DateFormatSymbols()
{
    dispose();
}

void DateFormatSymbols::dispose()
{
    delete[] fEras;
    delete[] fEraNames;
    delete[] fNarrowEras;
    delete[] fMonths;
    delete[] fShortMonths;
    delete[] fNarrowMonths;
    delete[] fStandaloneMonths;
    delete[] fStandaloneShortMonths;
    delete[] fStandaloneNarrowMonths;
    delete[] fWeekdays;
    delete[] fShortWeekdays;
    delete[] fShorterWeekdays;
    delete[] fNarrowWeekdays;
    delete[] fStandaloneWeekdays;
    delete[] fStandaloneShortWeekdays;
    delete[] fStandaloneShorterWeekdays;
    delete[] fStandaloneNarrowWeekdays;
    delete[] fAmPms;
    delete[] fNarrowAmPms;
    delete[] fQuarters;
    delete[] fShortQuarters;
    delete[] fStandaloneQuarters;
    delete[] fStandaloneShortQuarters;
    delete[] fLeapMonthPatterns;
    delete[] fShortYearNames;
    delete[] fShortZodiacNames;
    delete[] fAbbreviatedDayPeriods;
    delete[] fWideDayPeriods;
    delete[] fNarrowDayPeriods;
    delete[] fStandaloneAbbreviatedDayPeriods;
    delete[] fStandaloneWideDayPeriods;
    delete[] fStandaloneNarrowDayPeriods;

    disposeZoneStrings();
}

void DateFormatSymbols::disposeZoneStrings()
{
    if (fZoneStrings) {
        for (int32_t row = 0; row < fZoneStringsRowCount; ++row) {
            delete[] fZoneStrings[row];
        }
        uprv_free(fZoneStrings);
    }
    if (fLocaleZoneStrings) {
        for (int32_t row = 0; row < fZoneStringsRowCount; ++row) {
            delete[] fLocaleZoneStrings[row];
        }
        uprv_free(fLocaleZoneStrings);
    }

    fZoneStrings = NULL;
    fLocaleZoneStrings = NULL;
    fZoneStringsRowCount = 0;
    fZoneStringsColCount = 0;
}

UBool
DateFormatSymbols::arrayCompare(const UnicodeString* array1,
                                const UnicodeString* array2,
                                int32_t count)
{
    if (array1 == array2) return TRUE;
    while (count>0)
    {
        --count;
        if (array1[count] != array2[count]) return FALSE;
    }
    return TRUE;
}

UBool
DateFormatSymbols::operator==(const DateFormatSymbols& other) const
{
    // First do cheap comparisons
    if (this == &other) {
        return TRUE;
    }
    if (fErasCount == other.fErasCount &&
        fEraNamesCount == other.fEraNamesCount &&
        fNarrowErasCount == other.fNarrowErasCount &&
        fMonthsCount == other.fMonthsCount &&
        fShortMonthsCount == other.fShortMonthsCount &&
        fNarrowMonthsCount == other.fNarrowMonthsCount &&
        fStandaloneMonthsCount == other.fStandaloneMonthsCount &&
        fStandaloneShortMonthsCount == other.fStandaloneShortMonthsCount &&
        fStandaloneNarrowMonthsCount == other.fStandaloneNarrowMonthsCount &&
        fWeekdaysCount == other.fWeekdaysCount &&
        fShortWeekdaysCount == other.fShortWeekdaysCount &&
        fShorterWeekdaysCount == other.fShorterWeekdaysCount &&
        fNarrowWeekdaysCount == other.fNarrowWeekdaysCount &&
        fStandaloneWeekdaysCount == other.fStandaloneWeekdaysCount &&
        fStandaloneShortWeekdaysCount == other.fStandaloneShortWeekdaysCount &&
        fStandaloneShorterWeekdaysCount == other.fStandaloneShorterWeekdaysCount &&
        fStandaloneNarrowWeekdaysCount == other.fStandaloneNarrowWeekdaysCount &&
        fAmPmsCount == other.fAmPmsCount &&
        fNarrowAmPmsCount == other.fNarrowAmPmsCount &&
        fQuartersCount == other.fQuartersCount &&
        fShortQuartersCount == other.fShortQuartersCount &&
        fStandaloneQuartersCount == other.fStandaloneQuartersCount &&
        fStandaloneShortQuartersCount == other.fStandaloneShortQuartersCount &&
        fLeapMonthPatternsCount == other.fLeapMonthPatternsCount &&
        fShortYearNamesCount == other.fShortYearNamesCount &&
        fShortZodiacNamesCount == other.fShortZodiacNamesCount &&
        fAbbreviatedDayPeriodsCount == other.fAbbreviatedDayPeriodsCount &&
        fWideDayPeriodsCount == other.fWideDayPeriodsCount &&
        fNarrowDayPeriodsCount == other.fNarrowDayPeriodsCount &&
        fStandaloneAbbreviatedDayPeriodsCount == other.fStandaloneAbbreviatedDayPeriodsCount &&
        fStandaloneWideDayPeriodsCount == other.fStandaloneWideDayPeriodsCount &&
        fStandaloneNarrowDayPeriodsCount == other.fStandaloneNarrowDayPeriodsCount &&
        (uprv_memcmp(fCapitalization, other.fCapitalization, sizeof(fCapitalization))==0))
    {
        // Now compare the arrays themselves
        if (arrayCompare(fEras, other.fEras, fErasCount) &&
            arrayCompare(fEraNames, other.fEraNames, fEraNamesCount) &&
            arrayCompare(fNarrowEras, other.fNarrowEras, fNarrowErasCount) &&
            arrayCompare(fMonths, other.fMonths, fMonthsCount) &&
            arrayCompare(fShortMonths, other.fShortMonths, fShortMonthsCount) &&
            arrayCompare(fNarrowMonths, other.fNarrowMonths, fNarrowMonthsCount) &&
            arrayCompare(fStandaloneMonths, other.fStandaloneMonths, fStandaloneMonthsCount) &&
            arrayCompare(fStandaloneShortMonths, other.fStandaloneShortMonths, fStandaloneShortMonthsCount) &&
            arrayCompare(fStandaloneNarrowMonths, other.fStandaloneNarrowMonths, fStandaloneNarrowMonthsCount) &&
            arrayCompare(fWeekdays, other.fWeekdays, fWeekdaysCount) &&
            arrayCompare(fShortWeekdays, other.fShortWeekdays, fShortWeekdaysCount) &&
            arrayCompare(fShorterWeekdays, other.fShorterWeekdays, fShorterWeekdaysCount) &&
            arrayCompare(fNarrowWeekdays, other.fNarrowWeekdays, fNarrowWeekdaysCount) &&
            arrayCompare(fStandaloneWeekdays, other.fStandaloneWeekdays, fStandaloneWeekdaysCount) &&
            arrayCompare(fStandaloneShortWeekdays, other.fStandaloneShortWeekdays, fStandaloneShortWeekdaysCount) &&
            arrayCompare(fStandaloneShorterWeekdays, other.fStandaloneShorterWeekdays, fStandaloneShorterWeekdaysCount) &&
            arrayCompare(fStandaloneNarrowWeekdays, other.fStandaloneNarrowWeekdays, fStandaloneNarrowWeekdaysCount) &&
            arrayCompare(fAmPms, other.fAmPms, fAmPmsCount) &&
            arrayCompare(fNarrowAmPms, other.fNarrowAmPms, fNarrowAmPmsCount) &&
            fTimeSeparator == other.fTimeSeparator &&
            arrayCompare(fQuarters, other.fQuarters, fQuartersCount) &&
            arrayCompare(fShortQuarters, other.fShortQuarters, fShortQuartersCount) &&
            arrayCompare(fStandaloneQuarters, other.fStandaloneQuarters, fStandaloneQuartersCount) &&
            arrayCompare(fStandaloneShortQuarters, other.fStandaloneShortQuarters, fStandaloneShortQuartersCount) &&
            arrayCompare(fLeapMonthPatterns, other.fLeapMonthPatterns, fLeapMonthPatternsCount) &&
            arrayCompare(fShortYearNames, other.fShortYearNames, fShortYearNamesCount) &&
            arrayCompare(fShortZodiacNames, other.fShortZodiacNames, fShortZodiacNamesCount) &&
            arrayCompare(fAbbreviatedDayPeriods, other.fAbbreviatedDayPeriods, fAbbreviatedDayPeriodsCount) &&
            arrayCompare(fWideDayPeriods, other.fWideDayPeriods, fWideDayPeriodsCount) &&
            arrayCompare(fNarrowDayPeriods, other.fNarrowDayPeriods, fNarrowDayPeriodsCount) &&
            arrayCompare(fStandaloneAbbreviatedDayPeriods, other.fStandaloneAbbreviatedDayPeriods,
                         fStandaloneAbbreviatedDayPeriodsCount) &&
            arrayCompare(fStandaloneWideDayPeriods, other.fStandaloneWideDayPeriods,
                         fStandaloneWideDayPeriodsCount) &&
            arrayCompare(fStandaloneNarrowDayPeriods, other.fStandaloneNarrowDayPeriods,
                         fStandaloneWideDayPeriodsCount))
        {
            // Compare the contents of fZoneStrings
            if (fZoneStrings == NULL && other.fZoneStrings == NULL) {
                if (fZSFLocale == other.fZSFLocale) {
                    return TRUE;
                }
            } else if (fZoneStrings != NULL && other.fZoneStrings != NULL) {
                if (fZoneStringsRowCount == other.fZoneStringsRowCount
                    && fZoneStringsColCount == other.fZoneStringsColCount) {
                    UBool cmpres = TRUE;
                    for (int32_t i = 0; (i < fZoneStringsRowCount) && cmpres; i++) {
                        cmpres = arrayCompare(fZoneStrings[i], other.fZoneStrings[i], fZoneStringsColCount);
                    }
                    return cmpres;
                }
            }
            return FALSE;
        }
    }
    return FALSE;
}

//------------------------------------------------------

const UnicodeString*
DateFormatSymbols::getEras(int32_t &count) const
{
    count = fErasCount;
    return fEras;
}

const UnicodeString*
DateFormatSymbols::getEraNames(int32_t &count) const
{
    count = fEraNamesCount;
    return fEraNames;
}

const UnicodeString*
DateFormatSymbols::getNarrowEras(int32_t &count) const
{
    count = fNarrowErasCount;
    return fNarrowEras;
}

const UnicodeString*
DateFormatSymbols::getMonths(int32_t &count) const
{
    count = fMonthsCount;
    return fMonths;
}

const UnicodeString*
DateFormatSymbols::getShortMonths(int32_t &count) const
{
    count = fShortMonthsCount;
    return fShortMonths;
}

const UnicodeString*
DateFormatSymbols::getMonths(int32_t &count, DtContextType context, DtWidthType width ) const
{
    UnicodeString *returnValue = NULL;

    switch (context) {
    case FORMAT :
        switch(width) {
        case WIDE :
            count = fMonthsCount;
            returnValue = fMonths;
            break;
        case ABBREVIATED :
        case SHORT : // no month data for this, defaults to ABBREVIATED
            count = fShortMonthsCount;
            returnValue = fShortMonths;
            break;
        case NARROW :
            count = fNarrowMonthsCount;
            returnValue = fNarrowMonths;
            break;
        case DT_WIDTH_COUNT :
            break;
        }
        break;
    case STANDALONE :
        switch(width) {
        case WIDE :
            count = fStandaloneMonthsCount;
            returnValue = fStandaloneMonths;
            break;
        case ABBREVIATED :
        case SHORT : // no month data for this, defaults to ABBREVIATED
            count = fStandaloneShortMonthsCount;
            returnValue = fStandaloneShortMonths;
            break;
        case NARROW :
            count = fStandaloneNarrowMonthsCount;
            returnValue = fStandaloneNarrowMonths;
            break;
        case DT_WIDTH_COUNT :
            break;
        }
        break;
    case DT_CONTEXT_COUNT :
        break;
    }
    return returnValue;
}

const UnicodeString*
DateFormatSymbols::getWeekdays(int32_t &count) const
{
    count = fWeekdaysCount;
    return fWeekdays;
}

const UnicodeString*
DateFormatSymbols::getShortWeekdays(int32_t &count) const
{
    count = fShortWeekdaysCount;
    return fShortWeekdays;
}

const UnicodeString*
DateFormatSymbols::getWeekdays(int32_t &count, DtContextType context, DtWidthType width) const
{
    UnicodeString *returnValue = NULL;
    switch (context) {
    case FORMAT :
        switch(width) {
            case WIDE :
                count = fWeekdaysCount;
                returnValue = fWeekdays;
                break;
            case ABBREVIATED :
                count = fShortWeekdaysCount;
                returnValue = fShortWeekdays;
                break;
            case SHORT :
                count = fShorterWeekdaysCount;
                returnValue = fShorterWeekdays;
                break;
            case NARROW :
                count = fNarrowWeekdaysCount;
                returnValue = fNarrowWeekdays;
                break;
            case DT_WIDTH_COUNT :
                break;
        }
        break;
    case STANDALONE :
        switch(width) {
            case WIDE :
                count = fStandaloneWeekdaysCount;
                returnValue = fStandaloneWeekdays;
                break;
            case ABBREVIATED :
                count = fStandaloneShortWeekdaysCount;
                returnValue = fStandaloneShortWeekdays;
                break;
            case SHORT :
                count = fStandaloneShorterWeekdaysCount;
                returnValue = fStandaloneShorterWeekdays;
                break;
            case NARROW :
                count = fStandaloneNarrowWeekdaysCount;
                returnValue = fStandaloneNarrowWeekdays;
                break;
            case DT_WIDTH_COUNT :
                break;
        }
        break;
    case DT_CONTEXT_COUNT :
        break;
    }
    return returnValue;
}

const UnicodeString*
DateFormatSymbols::getQuarters(int32_t &count, DtContextType context, DtWidthType width ) const
{
    UnicodeString *returnValue = NULL;

    switch (context) {
    case FORMAT :
        switch(width) {
        case WIDE :
            count = fQuartersCount;
            returnValue = fQuarters;
            break;
        case ABBREVIATED :
        case SHORT : // no quarter data for this, defaults to ABBREVIATED
            count = fShortQuartersCount;
            returnValue = fShortQuarters;
            break;
        case NARROW :
            count = 0;
            returnValue = NULL;
            break;
        case DT_WIDTH_COUNT :
            break;
        }
        break;
    case STANDALONE :
        switch(width) {
        case WIDE :
            count = fStandaloneQuartersCount;
            returnValue = fStandaloneQuarters;
            break;
        case ABBREVIATED :
        case SHORT : // no quarter data for this, defaults to ABBREVIATED
            count = fStandaloneShortQuartersCount;
            returnValue = fStandaloneShortQuarters;
            break;
        case NARROW :
            count = 0;
            returnValue = NULL;
            break;
        case DT_WIDTH_COUNT :
            break;
        }
        break;
    case DT_CONTEXT_COUNT :
        break;
    }
    return returnValue;
}

UnicodeString&
DateFormatSymbols::getTimeSeparatorString(UnicodeString& result) const
{
    // fastCopyFrom() - see assignArray comments
    return result.fastCopyFrom(fTimeSeparator);
}

const UnicodeString*
DateFormatSymbols::getAmPmStrings(int32_t &count) const
{
    count = fAmPmsCount;
    return fAmPms;
}

const UnicodeString*
DateFormatSymbols::getLeapMonthPatterns(int32_t &count) const
{
    count = fLeapMonthPatternsCount;
    return fLeapMonthPatterns;
}

const UnicodeString*
DateFormatSymbols::getYearNames(int32_t& count,
                                DtContextType /*ignored*/, DtWidthType /*ignored*/) const
{
    count = fShortYearNamesCount;
    return fShortYearNames;
}

void
DateFormatSymbols::setYearNames(const UnicodeString* yearNames, int32_t count,
                                DtContextType context, DtWidthType width)
{
    if (context == FORMAT && width == ABBREVIATED) {
        if (fShortYearNames) {
            delete[] fShortYearNames;
        }
        fShortYearNames = newUnicodeStringArray(count);
        uprv_arrayCopy(yearNames, fShortYearNames, count);
        fShortYearNamesCount = count;
    }
}

const UnicodeString*
DateFormatSymbols::getZodiacNames(int32_t& count,
                                DtContextType /*ignored*/, DtWidthType /*ignored*/) const
{
    count = fShortZodiacNamesCount;
    return fShortZodiacNames;
}

void
DateFormatSymbols::setZodiacNames(const UnicodeString* zodiacNames, int32_t count,
                                DtContextType context, DtWidthType width)
{
    if (context == FORMAT && width == ABBREVIATED) {
        if (fShortZodiacNames) {
            delete[] fShortZodiacNames;
        }
        fShortZodiacNames = newUnicodeStringArray(count);
        uprv_arrayCopy(zodiacNames, fShortZodiacNames, count);
        fShortZodiacNamesCount = count;
    }
}

//------------------------------------------------------

void
DateFormatSymbols::setEras(const UnicodeString* erasArray, int32_t count)
{
    // delete the old list if we own it
    if (fEras)
        delete[] fEras;

    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)
    fEras = newUnicodeStringArray(count);
    uprv_arrayCopy(erasArray,fEras,  count);
    fErasCount = count;
}

void
DateFormatSymbols::setEraNames(const UnicodeString* eraNamesArray, int32_t count)
{
    // delete the old list if we own it
    if (fEraNames)
        delete[] fEraNames;

    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)
    fEraNames = newUnicodeStringArray(count);
    uprv_arrayCopy(eraNamesArray,fEraNames,  count);
    fEraNamesCount = count;
}

void
DateFormatSymbols::setNarrowEras(const UnicodeString* narrowErasArray, int32_t count)
{
    // delete the old list if we own it
    if (fNarrowEras)
        delete[] fNarrowEras;

    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)
    fNarrowEras = newUnicodeStringArray(count);
    uprv_arrayCopy(narrowErasArray,fNarrowEras,  count);
    fNarrowErasCount = count;
}

void
DateFormatSymbols::setMonths(const UnicodeString* monthsArray, int32_t count)
{
    // delete the old list if we own it
    if (fMonths)
        delete[] fMonths;

    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)
    fMonths = newUnicodeStringArray(count);
    uprv_arrayCopy( monthsArray,fMonths,count);
    fMonthsCount = count;
}

void
DateFormatSymbols::setShortMonths(const UnicodeString* shortMonthsArray, int32_t count)
{
    // delete the old list if we own it
    if (fShortMonths)
        delete[] fShortMonths;

    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)
    fShortMonths = newUnicodeStringArray(count);
    uprv_arrayCopy(shortMonthsArray,fShortMonths,  count);
    fShortMonthsCount = count;
}

void
DateFormatSymbols::setMonths(const UnicodeString* monthsArray, int32_t count, DtContextType context, DtWidthType width)
{
    // delete the old list if we own it
    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)

    switch (context) {
    case FORMAT :
        switch (width) {
        case WIDE :
            if (fMonths)
                delete[] fMonths;
            fMonths = newUnicodeStringArray(count);
            uprv_arrayCopy( monthsArray,fMonths,count);
            fMonthsCount = count;
            break;
        case ABBREVIATED :
            if (fShortMonths)
                delete[] fShortMonths;
            fShortMonths = newUnicodeStringArray(count);
            uprv_arrayCopy( monthsArray,fShortMonths,count);
            fShortMonthsCount = count;
            break;
        case NARROW :
            if (fNarrowMonths)
                delete[] fNarrowMonths;
            fNarrowMonths = newUnicodeStringArray(count);
            uprv_arrayCopy( monthsArray,fNarrowMonths,count);
            fNarrowMonthsCount = count;
            break;
        default :
            break;
        }
        break;
    case STANDALONE :
        switch (width) {
        case WIDE :
            if (fStandaloneMonths)
                delete[] fStandaloneMonths;
            fStandaloneMonths = newUnicodeStringArray(count);
            uprv_arrayCopy( monthsArray,fStandaloneMonths,count);
            fStandaloneMonthsCount = count;
            break;
        case ABBREVIATED :
            if (fStandaloneShortMonths)
                delete[] fStandaloneShortMonths;
            fStandaloneShortMonths = newUnicodeStringArray(count);
            uprv_arrayCopy( monthsArray,fStandaloneShortMonths,count);
            fStandaloneShortMonthsCount = count;
            break;
        case NARROW :
           if (fStandaloneNarrowMonths)
                delete[] fStandaloneNarrowMonths;
            fStandaloneNarrowMonths = newUnicodeStringArray(count);
            uprv_arrayCopy( monthsArray,fStandaloneNarrowMonths,count);
            fStandaloneNarrowMonthsCount = count;
            break;
        default :
            break;
        }
        break;
    case DT_CONTEXT_COUNT :
        break;
    }
}

void DateFormatSymbols::setWeekdays(const UnicodeString* weekdaysArray, int32_t count)
{
    // delete the old list if we own it
    if (fWeekdays)
        delete[] fWeekdays;

    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)
    fWeekdays = newUnicodeStringArray(count);
    uprv_arrayCopy(weekdaysArray,fWeekdays,count);
    fWeekdaysCount = count;
}

void
DateFormatSymbols::setShortWeekdays(const UnicodeString* shortWeekdaysArray, int32_t count)
{
    // delete the old list if we own it
    if (fShortWeekdays)
        delete[] fShortWeekdays;

    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)
    fShortWeekdays = newUnicodeStringArray(count);
    uprv_arrayCopy(shortWeekdaysArray, fShortWeekdays, count);
    fShortWeekdaysCount = count;
}

void
DateFormatSymbols::setWeekdays(const UnicodeString* weekdaysArray, int32_t count, DtContextType context, DtWidthType width)
{
    // delete the old list if we own it
    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)

    switch (context) {
    case FORMAT :
        switch (width) {
        case WIDE :
            if (fWeekdays)
                delete[] fWeekdays;
            fWeekdays = newUnicodeStringArray(count);
            uprv_arrayCopy(weekdaysArray, fWeekdays, count);
            fWeekdaysCount = count;
            break;
        case ABBREVIATED :
            if (fShortWeekdays)
                delete[] fShortWeekdays;
            fShortWeekdays = newUnicodeStringArray(count);
            uprv_arrayCopy(weekdaysArray, fShortWeekdays, count);
            fShortWeekdaysCount = count;
            break;
        case SHORT :
            if (fShorterWeekdays)
                delete[] fShorterWeekdays;
            fShorterWeekdays = newUnicodeStringArray(count);
            uprv_arrayCopy(weekdaysArray, fShorterWeekdays, count);
            fShorterWeekdaysCount = count;
            break;
        case NARROW :
            if (fNarrowWeekdays)
                delete[] fNarrowWeekdays;
            fNarrowWeekdays = newUnicodeStringArray(count);
            uprv_arrayCopy(weekdaysArray, fNarrowWeekdays, count);
            fNarrowWeekdaysCount = count;
            break;
        case DT_WIDTH_COUNT :
            break;
        }
        break;
    case STANDALONE :
        switch (width) {
        case WIDE :
            if (fStandaloneWeekdays)
                delete[] fStandaloneWeekdays;
            fStandaloneWeekdays = newUnicodeStringArray(count);
            uprv_arrayCopy(weekdaysArray, fStandaloneWeekdays, count);
            fStandaloneWeekdaysCount = count;
            break;
        case ABBREVIATED :
            if (fStandaloneShortWeekdays)
                delete[] fStandaloneShortWeekdays;
            fStandaloneShortWeekdays = newUnicodeStringArray(count);
            uprv_arrayCopy(weekdaysArray, fStandaloneShortWeekdays, count);
            fStandaloneShortWeekdaysCount = count;
            break;
        case SHORT :
            if (fStandaloneShorterWeekdays)
                delete[] fStandaloneShorterWeekdays;
            fStandaloneShorterWeekdays = newUnicodeStringArray(count);
            uprv_arrayCopy(weekdaysArray, fStandaloneShorterWeekdays, count);
            fStandaloneShorterWeekdaysCount = count;
            break;
        case NARROW :
            if (fStandaloneNarrowWeekdays)
                delete[] fStandaloneNarrowWeekdays;
            fStandaloneNarrowWeekdays = newUnicodeStringArray(count);
            uprv_arrayCopy(weekdaysArray, fStandaloneNarrowWeekdays, count);
            fStandaloneNarrowWeekdaysCount = count;
            break;
        case DT_WIDTH_COUNT :
            break;
        }
        break;
    case DT_CONTEXT_COUNT :
        break;
    }
}

void
DateFormatSymbols::setQuarters(const UnicodeString* quartersArray, int32_t count, DtContextType context, DtWidthType width)
{
    // delete the old list if we own it
    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)

    switch (context) {
    case FORMAT :
        switch (width) {
        case WIDE :
            if (fQuarters)
                delete[] fQuarters;
            fQuarters = newUnicodeStringArray(count);
            uprv_arrayCopy( quartersArray,fQuarters,count);
            fQuartersCount = count;
            break;
        case ABBREVIATED :
            if (fShortQuarters)
                delete[] fShortQuarters;
            fShortQuarters = newUnicodeStringArray(count);
            uprv_arrayCopy( quartersArray,fShortQuarters,count);
            fShortQuartersCount = count;
            break;
        case NARROW :
        /*
            if (fNarrowQuarters)
                delete[] fNarrowQuarters;
            fNarrowQuarters = newUnicodeStringArray(count);
            uprv_arrayCopy( quartersArray,fNarrowQuarters,count);
            fNarrowQuartersCount = count;
        */
            break;
        default :
            break;
        }
        break;
    case STANDALONE :
        switch (width) {
        case WIDE :
            if (fStandaloneQuarters)
                delete[] fStandaloneQuarters;
            fStandaloneQuarters = newUnicodeStringArray(count);
            uprv_arrayCopy( quartersArray,fStandaloneQuarters,count);
            fStandaloneQuartersCount = count;
            break;
        case ABBREVIATED :
            if (fStandaloneShortQuarters)
                delete[] fStandaloneShortQuarters;
            fStandaloneShortQuarters = newUnicodeStringArray(count);
            uprv_arrayCopy( quartersArray,fStandaloneShortQuarters,count);
            fStandaloneShortQuartersCount = count;
            break;
        case NARROW :
        /*
           if (fStandaloneNarrowQuarters)
                delete[] fStandaloneNarrowQuarters;
            fStandaloneNarrowQuarters = newUnicodeStringArray(count);
            uprv_arrayCopy( quartersArray,fStandaloneNarrowQuarters,count);
            fStandaloneNarrowQuartersCount = count;
        */
            break;
        default :
            break;
        }
        break;
    case DT_CONTEXT_COUNT :
        break;
    }
}

void
DateFormatSymbols::setAmPmStrings(const UnicodeString* amPmsArray, int32_t count)
{
    // delete the old list if we own it
    if (fAmPms) delete[] fAmPms;

    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)
    fAmPms = newUnicodeStringArray(count);
    uprv_arrayCopy(amPmsArray,fAmPms,count);
    fAmPmsCount = count;
}

void
DateFormatSymbols::setTimeSeparatorString(const UnicodeString& newTimeSeparator)
{
    fTimeSeparator = newTimeSeparator;
}

const UnicodeString**
DateFormatSymbols::getZoneStrings(int32_t& rowCount, int32_t& columnCount) const
{
    const UnicodeString **result = NULL;
    static UMutex LOCK;

    umtx_lock(&LOCK);
    if (fZoneStrings == NULL) {
        if (fLocaleZoneStrings == NULL) {
            ((DateFormatSymbols*)this)->initZoneStringsArray();
        }
        result = (const UnicodeString**)fLocaleZoneStrings;
    } else {
        result = (const UnicodeString**)fZoneStrings;
    }
    rowCount = fZoneStringsRowCount;
    columnCount = fZoneStringsColCount;
    umtx_unlock(&LOCK);

    return result;
}

// For now, we include all zones
#define ZONE_SET UCAL_ZONE_TYPE_ANY

// This code must be called within a synchronized block
void
DateFormatSymbols::initZoneStringsArray(void) {
    if (fZoneStrings != NULL || fLocaleZoneStrings != NULL) {
        return;
    }

    UErrorCode status = U_ZERO_ERROR;

    StringEnumeration *tzids = NULL;
    UnicodeString ** zarray = NULL;
    TimeZoneNames *tzNames = NULL;
    int32_t rows = 0;

    static const UTimeZoneNameType TYPES[] = {
        UTZNM_LONG_STANDARD, UTZNM_SHORT_STANDARD,
        UTZNM_LONG_DAYLIGHT, UTZNM_SHORT_DAYLIGHT
    };
    static const int32_t NUM_TYPES = 4;

    do { // dummy do-while

        tzids = TimeZone::createTimeZoneIDEnumeration(ZONE_SET, NULL, NULL, status);
        rows = tzids->count(status);
        if (U_FAILURE(status)) {
            break;
        }

        // Allocate array
        int32_t size = rows * sizeof(UnicodeString*);
        zarray = (UnicodeString**)uprv_malloc(size);
        if (zarray == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            break;
        }
        uprv_memset(zarray, 0, size);

        tzNames = TimeZoneNames::createInstance(fZSFLocale, status);
        tzNames->loadAllDisplayNames(status);
        if (U_FAILURE(status)) { break; }

        const UnicodeString *tzid;
        int32_t i = 0;
        UDate now = Calendar::getNow();
        UnicodeString tzDispName;

        while ((tzid = tzids->snext(status)) != 0) {
            if (U_FAILURE(status)) {
                break;
            }

            zarray[i] = new UnicodeString[5];
            if (zarray[i] == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                break;
            }

            zarray[i][0].setTo(*tzid);
            tzNames->getDisplayNames(*tzid, TYPES, NUM_TYPES, now, zarray[i]+1, status);
            i++;
        }

    } while (FALSE);

    if (U_FAILURE(status)) {
        if (zarray) {
            for (int32_t i = 0; i < rows; i++) {
                if (zarray[i]) {
                    delete[] zarray[i];
                }
            }
            uprv_free(zarray);
            zarray = NULL;
        }
    }

    if (tzNames) {
        delete tzNames;
    }
    if (tzids) {
        delete tzids;
    }

    fLocaleZoneStrings = zarray;
    fZoneStringsRowCount = rows;
    fZoneStringsColCount = 1 + NUM_TYPES;
}

void
DateFormatSymbols::setZoneStrings(const UnicodeString* const *strings, int32_t rowCount, int32_t columnCount)
{
    // since deleting a 2-d array is a pain in the butt, we offload that task to
    // a separate function
    disposeZoneStrings();
    // we always own the new list, which we create here (we duplicate rather
    // than adopting the list passed in)
    fZoneStringsRowCount = rowCount;
    fZoneStringsColCount = columnCount;
    createZoneStrings((const UnicodeString**)strings);
}

//------------------------------------------------------

const char16_t * U_EXPORT2
DateFormatSymbols::getPatternUChars(void)
{
    return gPatternChars;
}

UDateFormatField U_EXPORT2
DateFormatSymbols::getPatternCharIndex(UChar c) {
    const UChar *p = u_strchr(gPatternChars, c);
    if (p == NULL) {
        return UDAT_FIELD_COUNT;
    } else {
        return static_cast<UDateFormatField>(p - gPatternChars);
    }
}

static const uint64_t kNumericFieldsAlways =
    ((uint64_t)1 << UDAT_YEAR_FIELD) |                      // y
    ((uint64_t)1 << UDAT_DATE_FIELD) |                      // d
    ((uint64_t)1 << UDAT_HOUR_OF_DAY1_FIELD) |              // k
    ((uint64_t)1 << UDAT_HOUR_OF_DAY0_FIELD) |              // H
    ((uint64_t)1 << UDAT_MINUTE_FIELD) |                    // m
    ((uint64_t)1 << UDAT_SECOND_FIELD) |                    // s
    ((uint64_t)1 << UDAT_FRACTIONAL_SECOND_FIELD) |         // S
    ((uint64_t)1 << UDAT_DAY_OF_YEAR_FIELD) |               // D
    ((uint64_t)1 << UDAT_DAY_OF_WEEK_IN_MONTH_FIELD) |      // F
    ((uint64_t)1 << UDAT_WEEK_OF_YEAR_FIELD) |              // w
    ((uint64_t)1 << UDAT_WEEK_OF_MONTH_FIELD) |             // W
    ((uint64_t)1 << UDAT_HOUR1_FIELD) |                     // h
    ((uint64_t)1 << UDAT_HOUR0_FIELD) |                     // K
    ((uint64_t)1 << UDAT_YEAR_WOY_FIELD) |                  // Y
    ((uint64_t)1 << UDAT_EXTENDED_YEAR_FIELD) |             // u
    ((uint64_t)1 << UDAT_JULIAN_DAY_FIELD) |                // g
    ((uint64_t)1 << UDAT_MILLISECONDS_IN_DAY_FIELD) |       // A
    ((uint64_t)1 << UDAT_RELATED_YEAR_FIELD);               // r

static const uint64_t kNumericFieldsForCount12 =
    ((uint64_t)1 << UDAT_MONTH_FIELD) |                     // M or MM
    ((uint64_t)1 << UDAT_DOW_LOCAL_FIELD) |                 // e or ee
    ((uint64_t)1 << UDAT_STANDALONE_DAY_FIELD) |            // c or cc
    ((uint64_t)1 << UDAT_STANDALONE_MONTH_FIELD) |          // L or LL
    ((uint64_t)1 << UDAT_QUARTER_FIELD) |                   // Q or QQ
    ((uint64_t)1 << UDAT_STANDALONE_QUARTER_FIELD);         // q or qq

UBool U_EXPORT2
DateFormatSymbols::isNumericField(UDateFormatField f, int32_t count) {
    if (f == UDAT_FIELD_COUNT) {
        return FALSE;
    }
    uint64_t flag = ((uint64_t)1 << f);
    return ((kNumericFieldsAlways & flag) != 0 || ((kNumericFieldsForCount12 & flag) != 0 && count < 3));
}

UBool U_EXPORT2
DateFormatSymbols::isNumericPatternChar(UChar c, int32_t count) {
    return isNumericField(getPatternCharIndex(c), count);
}

//------------------------------------------------------

UnicodeString&
DateFormatSymbols::getLocalPatternChars(UnicodeString& result) const
{
    // fastCopyFrom() - see assignArray comments
    return result.fastCopyFrom(fLocalPatternChars);
}

//------------------------------------------------------

void
DateFormatSymbols::setLocalPatternChars(const UnicodeString& newLocalPatternChars)
{
    fLocalPatternChars = newLocalPatternChars;
}

//------------------------------------------------------

namespace {

// Constants declarations
static const UChar kCalendarAliasPrefixUChar[] = {
    dt_impl_SOLIDUS, dt_impl_CAP_L, dt_impl_CAP_O, dt_impl_CAP_C, dt_impl_CAP_A, dt_impl_CAP_L, dt_impl_CAP_E, dt_impl_SOLIDUS,
    dt_impl_LOW_C, dt_impl_LOW_A, dt_impl_LOW_L, dt_impl_LOW_E, dt_impl_LOW_N, dt_impl_LOW_D, dt_impl_LOW_A, dt_impl_LOW_R, dt_impl_SOLIDUS
};
static const UChar kGregorianTagUChar[] = {
    dt_impl_LOW_G, dt_impl_LOW_R, dt_impl_LOW_E, dt_impl_LOW_G, dt_impl_LOW_O, dt_impl_LOW_R, dt_impl_LOW_I, dt_impl_LOW_A, dt_impl_LOW_N
};
static const UChar kVariantTagUChar[] = {
    dt_impl_PERCENT, dt_impl_LOW_V, dt_impl_LOW_A, dt_impl_LOW_R, dt_impl_LOW_I, dt_impl_LOW_A, dt_impl_LOW_N, dt_impl_LOW_T
};
static const UChar kLeapTagUChar[] = {
    dt_impl_LOW_L, dt_impl_LOW_E, dt_impl_LOW_A, dt_impl_LOW_P
};
static const UChar kCyclicNameSetsTagUChar[] = {
    dt_impl_LOW_C, dt_impl_LOW_Y, dt_impl_LOW_C, dt_impl_LOW_L, dt_impl_LOW_I, dt_impl_LOW_C, dt_impl_CAP_N, dt_impl_LOW_A, dt_impl_LOW_M, dt_impl_LOW_E, dt_impl_CAP_S, dt_impl_LOW_E, dt_impl_LOW_T, dt_impl_LOW_S
};
static const UChar kYearsTagUChar[] = {
    dt_impl_SOLIDUS, dt_impl_LOW_Y, dt_impl_LOW_E, dt_impl_LOW_A, dt_impl_LOW_R, dt_impl_LOW_S
};
static const UChar kZodiacsUChar[] = {
    dt_impl_SOLIDUS, dt_impl_LOW_Z, dt_impl_LOW_O, dt_impl_LOW_D, dt_impl_LOW_I, dt_impl_LOW_A, dt_impl_LOW_C, dt_impl_LOW_S
};
static const UChar kDayPartsTagUChar[] = {
    dt_impl_SOLIDUS, dt_impl_LOW_D, dt_impl_LOW_A, dt_impl_LOW_Y, dt_impl_CAP_P, dt_impl_LOW_A, dt_impl_LOW_R, dt_impl_LOW_T, dt_impl_LOW_S
};
static const UChar kFormatTagUChar[] = {
    dt_impl_SOLIDUS, dt_impl_LOW_F, dt_impl_LOW_O, dt_impl_LOW_R, dt_impl_LOW_M, dt_impl_LOW_A, dt_impl_LOW_T
};
static const UChar kAbbrTagUChar[] = {
    dt_impl_SOLIDUS, dt_impl_LOW_A, dt_impl_LOW_B, dt_impl_LOW_B, dt_impl_LOW_R, dt_impl_LOW_E, dt_impl_LOW_V, dt_impl_LOW_I, dt_impl_LOW_A, dt_impl_LOW_T, dt_impl_LOW_E, dt_impl_LOW_D
};

// ResourceSink to enumerate all calendar resources
struct CalendarDataSink : public ResourceSink {

    // Enum which specifies the type of alias received, or no alias
    enum AliasType {
        SAME_CALENDAR,
        DIFFERENT_CALENDAR,
        GREGORIAN,
        NONE
    };

    // Data structures to store resources from the current resource bundle
    Hashtable arrays;
    Hashtable arraySizes;
    Hashtable maps;
    /**
     * Whenever there are aliases, the same object will be added twice to 'map'.
     * To avoid double deletion, 'maps' won't take ownership of the objects. Instead,
     * 'mapRefs' will own them and will delete them when CalendarDataSink is deleted.
     */
    MemoryPool<Hashtable> mapRefs;

    // Paths and the aliases they point to
    UVector aliasPathPairs;

    // Current and next calendar resource table which should be loaded
    UnicodeString currentCalendarType;
    UnicodeString nextCalendarType;

    // Resources to visit when enumerating fallback calendars
    LocalPointer<UVector> resourcesToVisit;

    // Alias' relative path populated whenever an alias is read
    UnicodeString aliasRelativePath;

    // Initializes CalendarDataSink with default values
    CalendarDataSink(UErrorCode& status)
    :   arrays(FALSE, status), arraySizes(FALSE, status), maps(FALSE, status),
        mapRefs(),
        aliasPathPairs(uprv_deleteUObject, uhash_compareUnicodeString, status),
        currentCalendarType(), nextCalendarType(),
        resourcesToVisit(NULL), aliasRelativePath() {
        if (U_FAILURE(status)) { return; }
    }
    virtual ~CalendarDataSink();

    // Configure the CalendarSink to visit all the resources
    void visitAllResources() {
        resourcesToVisit.adoptInstead(NULL);
    }

    // Actions to be done before enumerating
    void preEnumerate(const UnicodeString &calendarType) {
        currentCalendarType = calendarType;
        nextCalendarType.setToBogus();
        aliasPathPairs.removeAllElements();
    }

    virtual void put(const char *key, ResourceValue &value, UBool, UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return; }
        U_ASSERT(!currentCalendarType.isEmpty());

        // Stores the resources to visit on the next calendar.
        LocalPointer<UVector> resourcesToVisitNext(NULL);
        ResourceTable calendarData = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }

        // Enumerate all resources for this calendar
        for (int i = 0; calendarData.getKeyAndValue(i, key, value); i++) {
            UnicodeString keyUString(key, -1, US_INV);

            // == Handle aliases ==
            AliasType aliasType = processAliasFromValue(keyUString, value, errorCode);
            if (U_FAILURE(errorCode)) { return; }
            if (aliasType == GREGORIAN) {
                // Ignore aliases to the gregorian calendar, all of its resources will be loaded anyway.
                continue;

            } else if (aliasType == DIFFERENT_CALENDAR) {
                // Whenever an alias to the next calendar (except gregorian) is encountered, register the
                // calendar type it's pointing to
                if (resourcesToVisitNext.isNull()) {
                    resourcesToVisitNext
                        .adoptInsteadAndCheckErrorCode(new UVector(uprv_deleteUObject, uhash_compareUnicodeString, errorCode),
                                                       errorCode);
                    if (U_FAILURE(errorCode)) { return; }
                }
                LocalPointer<UnicodeString> aliasRelativePathCopy(new UnicodeString(aliasRelativePath), errorCode);
                resourcesToVisitNext->addElement(aliasRelativePathCopy.getAlias(), errorCode);
                if (U_FAILURE(errorCode)) { return; }
                // Only release ownership after resourcesToVisitNext takes it (no error happened):
                aliasRelativePathCopy.orphan();
                continue;

            } else if (aliasType == SAME_CALENDAR) {
                // Register same-calendar alias
                if (arrays.get(aliasRelativePath) == NULL && maps.get(aliasRelativePath) == NULL) {
                    LocalPointer<UnicodeString> aliasRelativePathCopy(new UnicodeString(aliasRelativePath), errorCode);
                    aliasPathPairs.addElement(aliasRelativePathCopy.getAlias(), errorCode);
                    if (U_FAILURE(errorCode)) { return; }
                    // Only release ownership after aliasPathPairs takes it (no error happened):
                    aliasRelativePathCopy.orphan();
                    LocalPointer<UnicodeString> keyUStringCopy(new UnicodeString(keyUString), errorCode);
                    aliasPathPairs.addElement(keyUStringCopy.getAlias(), errorCode);
                    if (U_FAILURE(errorCode)) { return; }
                    // Only release ownership after aliasPathPairs takes it (no error happened):
                    keyUStringCopy.orphan();
                }
                continue;
            }

            // Only visit the resources that were referenced by an alias on the previous calendar
            // (AmPmMarkersAbbr is an exception).
            if (!resourcesToVisit.isNull() && !resourcesToVisit->isEmpty() && !resourcesToVisit->contains(&keyUString)
                && uprv_strcmp(key, dtfmtsym_gAmPmMarkersAbbrTag) != 0) { continue; }

            // == Handle data ==
            if (uprv_strcmp(key, dtfmtsym_gAmPmMarkersTag) == 0
                || uprv_strcmp(key, dtfmtsym_gAmPmMarkersAbbrTag) == 0
                || uprv_strcmp(key, dtfmtsym_gAmPmMarkersNarrowTag) == 0) {
                if (arrays.get(keyUString) == NULL) {
                    ResourceArray resourceArray = value.getArray(errorCode);
                    int32_t arraySize = resourceArray.getSize();
                    LocalArray<UnicodeString> stringArray(new UnicodeString[arraySize], errorCode);
                    value.getStringArray(stringArray.getAlias(), arraySize, errorCode);
                    arrays.put(keyUString, stringArray.orphan(), errorCode);
                    arraySizes.puti(keyUString, arraySize, errorCode);
                    if (U_FAILURE(errorCode)) { return; }
                }
            } else if (uprv_strcmp(key, dtfmtsym_gErasTag) == 0
                       || uprv_strcmp(key, dtfmtsym_gDayNamesTag) == 0
                       || uprv_strcmp(key, dtfmtsym_gMonthNamesTag) == 0
                       || uprv_strcmp(key, dtfmtsym_gQuartersTag) == 0
                       || uprv_strcmp(key, dtfmtsym_gDayPeriodTag) == 0
                       || uprv_strcmp(key, dtfmtsym_gMonthPatternsTag) == 0
                       || uprv_strcmp(key, dtfmtsym_gCyclicNameSetsTag) == 0) {
                processResource(keyUString, key, value, errorCode);
            }
        }

        // Apply same-calendar aliases
        UBool modified;
        do {
            modified = false;
            for (int32_t i = 0; i < aliasPathPairs.size();) {
                UBool mod = false;
                UnicodeString *alias = (UnicodeString*)aliasPathPairs[i];
                UnicodeString *aliasArray;
                Hashtable *aliasMap;
                if ((aliasArray = (UnicodeString*)arrays.get(*alias)) != NULL) {
                    UnicodeString *path = (UnicodeString*)aliasPathPairs[i + 1];
                    if (arrays.get(*path) == NULL) {
                        // Clone the array
                        int32_t aliasArraySize = arraySizes.geti(*alias);
                        LocalArray<UnicodeString> aliasArrayCopy(new UnicodeString[aliasArraySize], errorCode);
                        if (U_FAILURE(errorCode)) { return; }
                        uprv_arrayCopy(aliasArray, aliasArrayCopy.getAlias(), aliasArraySize);
                        // Put the array on the 'arrays' map
                        arrays.put(*path, aliasArrayCopy.orphan(), errorCode);
                        arraySizes.puti(*path, aliasArraySize, errorCode);
                    }
                    if (U_FAILURE(errorCode)) { return; }
                    mod = true;
                } else if ((aliasMap = (Hashtable*)maps.get(*alias)) != NULL) {
                    UnicodeString *path = (UnicodeString*)aliasPathPairs[i + 1];
                    if (maps.get(*path) == NULL) {
                        maps.put(*path, aliasMap, errorCode);
                    }
                    if (U_FAILURE(errorCode)) { return; }
                    mod = true;
                }
                if (mod) {
                    aliasPathPairs.removeElementAt(i + 1);
                    aliasPathPairs.removeElementAt(i);
                    modified = true;
                } else {
                    i += 2;
                }
            }
        } while (modified && !aliasPathPairs.isEmpty());

        // Set the resources to visit on the next calendar
        if (!resourcesToVisitNext.isNull()) {
            resourcesToVisit = std::move(resourcesToVisitNext);
        }
    }

    // Process the nested resource bundle tables
    void processResource(UnicodeString &path, const char *key, ResourceValue &value, UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) return;

        ResourceTable table = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) return;
        Hashtable* stringMap = NULL;

        // Iterate over all the elements of the table and add them to the map
        for (int i = 0; table.getKeyAndValue(i, key, value); i++) {
            UnicodeString keyUString(key, -1, US_INV);

            // Ignore '%variant' keys
            if (keyUString.endsWith(kVariantTagUChar, UPRV_LENGTHOF(kVariantTagUChar))) {
                continue;
            }

            // == Handle String elements ==
            if (value.getType() == URES_STRING) {
                // We are on a leaf, store the map elements into the stringMap
                if (i == 0) {
                    // mapRefs will keep ownership of 'stringMap':
                    stringMap = mapRefs.create(FALSE, errorCode);
                    if (stringMap == NULL) {
                        errorCode = U_MEMORY_ALLOCATION_ERROR;
                        return;
                    }
                    maps.put(path, stringMap, errorCode);
                    if (U_FAILURE(errorCode)) { return; }
                    stringMap->setValueDeleter(uprv_deleteUObject);
                }
                U_ASSERT(stringMap != NULL);
                int32_t valueStringSize;
                const UChar *valueString = value.getString(valueStringSize, errorCode);
                if (U_FAILURE(errorCode)) { return; }
                LocalPointer<UnicodeString> valueUString(new UnicodeString(TRUE, valueString, valueStringSize), errorCode);
                stringMap->put(keyUString, valueUString.orphan(), errorCode);
                if (U_FAILURE(errorCode)) { return; }
                continue;
            }
            U_ASSERT(stringMap == NULL);

            // Store the current path's length and append the current key to the path.
            int32_t pathLength = path.length();
            path.append(dt_impl_SOLIDUS).append(keyUString);

            // In cyclicNameSets ignore everything but years/format/abbreviated
            // and zodiacs/format/abbreviated
            if (path.startsWith(kCyclicNameSetsTagUChar, UPRV_LENGTHOF(kCyclicNameSetsTagUChar))) {
                UBool skip = TRUE;
                int32_t startIndex = UPRV_LENGTHOF(kCyclicNameSetsTagUChar);
                int32_t length = 0;
                if (startIndex == path.length()
                    || path.compare(startIndex, (length = UPRV_LENGTHOF(kZodiacsUChar)), kZodiacsUChar, 0, UPRV_LENGTHOF(kZodiacsUChar)) == 0
                    || path.compare(startIndex, (length = UPRV_LENGTHOF(kYearsTagUChar)), kYearsTagUChar, 0, UPRV_LENGTHOF(kYearsTagUChar)) == 0
                    || path.compare(startIndex, (length = UPRV_LENGTHOF(kDayPartsTagUChar)), kDayPartsTagUChar, 0, UPRV_LENGTHOF(kDayPartsTagUChar)) == 0) {
                    startIndex += length;
                    length = 0;
                    if (startIndex == path.length()
                        || path.compare(startIndex, (length = UPRV_LENGTHOF(kFormatTagUChar)), kFormatTagUChar, 0, UPRV_LENGTHOF(kFormatTagUChar)) == 0) {
                        startIndex += length;
                        length = 0;
                        if (startIndex == path.length()
                            || path.compare(startIndex, (length = UPRV_LENGTHOF(kAbbrTagUChar)), kAbbrTagUChar, 0, UPRV_LENGTHOF(kAbbrTagUChar)) == 0) {
                            skip = FALSE;
                        }
                    }
                }
                if (skip) {
                    // Drop the latest key on the path and continue
                    path.retainBetween(0, pathLength);
                    continue;
                }
            }

            // == Handle aliases ==
            if (arrays.get(path) != NULL || maps.get(path) != NULL) {
                // Drop the latest key on the path and continue
                path.retainBetween(0, pathLength);
                continue;
            }

            AliasType aliasType = processAliasFromValue(path, value, errorCode);
            if (U_FAILURE(errorCode)) { return; }
            if (aliasType == SAME_CALENDAR) {
                // Store the alias path and the current path on aliasPathPairs
                LocalPointer<UnicodeString> aliasRelativePathCopy(new UnicodeString(aliasRelativePath), errorCode);
                aliasPathPairs.addElement(aliasRelativePathCopy.getAlias(), errorCode);
                if (U_FAILURE(errorCode)) { return; }
                // Only release ownership after aliasPathPairs takes it (no error happened):
                aliasRelativePathCopy.orphan();
                LocalPointer<UnicodeString> pathCopy(new UnicodeString(path), errorCode);
                aliasPathPairs.addElement(pathCopy.getAlias(), errorCode);
                if (U_FAILURE(errorCode)) { return; }
                // Only release ownership after aliasPathPairs takes it (no error happened):
                pathCopy.orphan();

                // Drop the latest key on the path and continue
                path.retainBetween(0, pathLength);
                continue;
            }
            U_ASSERT(aliasType == NONE);

            // == Handle data ==
            if (value.getType() == URES_ARRAY) {
                // We are on a leaf, store the array
                ResourceArray rDataArray = value.getArray(errorCode);
                int32_t dataArraySize = rDataArray.getSize();
                LocalArray<UnicodeString> dataArray(new UnicodeString[dataArraySize], errorCode);
                value.getStringArray(dataArray.getAlias(), dataArraySize, errorCode);
                arrays.put(path, dataArray.orphan(), errorCode);
                arraySizes.puti(path, dataArraySize, errorCode);
                if (U_FAILURE(errorCode)) { return; }
            } else if (value.getType() == URES_TABLE) {
                // We are not on a leaf, recursively process the subtable.
                processResource(path, key, value, errorCode);
                if (U_FAILURE(errorCode)) { return; }
            }

            // Drop the latest key on the path
            path.retainBetween(0, pathLength);
        }
    }

    // Populates an AliasIdentifier with the alias information contained on the UResource.Value.
    AliasType processAliasFromValue(UnicodeString &currentRelativePath, ResourceValue &value,
                                    UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return NONE; }

        if (value.getType() == URES_ALIAS) {
            int32_t aliasPathSize;
            const UChar* aliasPathUChar = value.getAliasString(aliasPathSize, errorCode);
            if (U_FAILURE(errorCode)) { return NONE; }
            UnicodeString aliasPath(aliasPathUChar, aliasPathSize);
            const int32_t aliasPrefixLength = UPRV_LENGTHOF(kCalendarAliasPrefixUChar);
            if (aliasPath.startsWith(kCalendarAliasPrefixUChar, aliasPrefixLength)
                && aliasPath.length() > aliasPrefixLength) {
                int32_t typeLimit = aliasPath.indexOf(dt_impl_SOLIDUS, aliasPrefixLength);
                if (typeLimit > aliasPrefixLength) {
                    const UnicodeString aliasCalendarType =
                        aliasPath.tempSubStringBetween(aliasPrefixLength, typeLimit);
                    aliasRelativePath.setTo(aliasPath, typeLimit + 1, aliasPath.length());

                    if (currentCalendarType == aliasCalendarType
                        && currentRelativePath != aliasRelativePath) {
                        // If we have an alias to the same calendar, the path to the resource must be different
                        return SAME_CALENDAR;

                    } else if (currentCalendarType != aliasCalendarType
                               && currentRelativePath == aliasRelativePath) {
                        // If we have an alias to a different calendar, the path to the resource must be the same
                        if (aliasCalendarType.compare(kGregorianTagUChar, UPRV_LENGTHOF(kGregorianTagUChar)) == 0) {
                            return GREGORIAN;
                        } else if (nextCalendarType.isBogus()) {
                            nextCalendarType = aliasCalendarType;
                            return DIFFERENT_CALENDAR;
                        } else if (nextCalendarType == aliasCalendarType) {
                            return DIFFERENT_CALENDAR;
                        }
                    }
                }
            }
            errorCode = U_INTERNAL_PROGRAM_ERROR;
            return NONE;
        }
        return NONE;
    }

    // Deleter function to be used by 'arrays'
    static void U_CALLCONV deleteUnicodeStringArray(void *uArray) {
        delete[] static_cast<UnicodeString *>(uArray);
    }
};
// Virtual destructors have to be defined out of line
CalendarDataSink::~CalendarDataSink() {
    arrays.setValueDeleter(deleteUnicodeStringArray);
}
}

//------------------------------------------------------

static void
initField(UnicodeString **field, int32_t& length, const UChar *data, LastResortSize numStr, LastResortSize strLen, UErrorCode &status) {
    if (U_SUCCESS(status)) {
        length = numStr;
        *field = newUnicodeStringArray((size_t)numStr);
        if (*field) {
            for(int32_t i = 0; i<length; i++) {
                // readonly aliases - all "data" strings are constant
                // -1 as length for variable-length strings (gLastResortDayNames[0] is empty)
                (*(field)+i)->setTo(TRUE, data+(i*((int32_t)strLen)), -1);
            }
        }
        else {
            length = 0;
            status = U_MEMORY_ALLOCATION_ERROR;
        }
    }
}

static void
initField(UnicodeString **field, int32_t& length, CalendarDataSink &sink, CharString &key, UErrorCode &status) {
    if (U_SUCCESS(status)) {
        UnicodeString keyUString(key.data(), -1, US_INV);
        UnicodeString* array = static_cast<UnicodeString*>(sink.arrays.get(keyUString));

        if (array != NULL) {
            length = sink.arraySizes.geti(keyUString);
            *field = array;
            // DateFormatSymbols takes ownership of the array:
            sink.arrays.remove(keyUString);
        } else {
            length = 0;
            status = U_MISSING_RESOURCE_ERROR;
        }
    }
}

static void
initField(UnicodeString **field, int32_t& length, CalendarDataSink &sink, CharString &key, int32_t arrayOffset, UErrorCode &status) {
    if (U_SUCCESS(status)) {
        UnicodeString keyUString(key.data(), -1, US_INV);
        UnicodeString* array = static_cast<UnicodeString*>(sink.arrays.get(keyUString));

        if (array != NULL) {
            int32_t arrayLength = sink.arraySizes.geti(keyUString);
            length = arrayLength + arrayOffset;
            *field = new UnicodeString[length];
            if (*field == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            uprv_arrayCopy(array, 0, *field, arrayOffset, arrayLength);
        } else {
            length = 0;
            status = U_MISSING_RESOURCE_ERROR;
        }
    }
}

static void
initLeapMonthPattern(UnicodeString *field, int32_t index, CalendarDataSink &sink, CharString &path, UErrorCode &status) {
    field[index].remove();
    if (U_SUCCESS(status)) {
        UnicodeString pathUString(path.data(), -1, US_INV);
        Hashtable *leapMonthTable = static_cast<Hashtable*>(sink.maps.get(pathUString));
        if (leapMonthTable != NULL) {
            UnicodeString leapLabel(FALSE, kLeapTagUChar, UPRV_LENGTHOF(kLeapTagUChar));
            UnicodeString *leapMonthPattern = static_cast<UnicodeString*>(leapMonthTable->get(leapLabel));
            if (leapMonthPattern != NULL) {
                field[index].fastCopyFrom(*leapMonthPattern);
            } else {
                field[index].setToBogus();
            }
            return;
        }
        status = U_MISSING_RESOURCE_ERROR;
    }
}

static CharString
&buildResourcePath(CharString &path, const char* segment1, UErrorCode &errorCode) {
    return path.clear().append(segment1, -1, errorCode);
}

static CharString
&buildResourcePath(CharString &path, const char* segment1, const char* segment2,
                   UErrorCode &errorCode) {
    return buildResourcePath(path, segment1, errorCode).append('/', errorCode)
                                                       .append(segment2, -1, errorCode);
}

static CharString
&buildResourcePath(CharString &path, const char* segment1, const char* segment2,
                   const char* segment3, UErrorCode &errorCode) {
    return buildResourcePath(path, segment1, segment2, errorCode).append('/', errorCode)
                                                                 .append(segment3, -1, errorCode);
}

static CharString
&buildResourcePath(CharString &path, const char* segment1, const char* segment2,
                   const char* segment3, const char* segment4, UErrorCode &errorCode) {
    return buildResourcePath(path, segment1, segment2, segment3, errorCode).append('/', errorCode)
                                                                           .append(segment4, -1, errorCode);
}

typedef struct {
    const char * usageTypeName;
    DateFormatSymbols::ECapitalizationContextUsageType usageTypeEnumValue;
} ContextUsageTypeNameToEnumValue;

static const ContextUsageTypeNameToEnumValue contextUsageTypeMap[] = {
   // Entries must be sorted by usageTypeName; entry with NULL name terminates list.
    { "day-format-except-narrow", DateFormatSymbols::kCapContextUsageDayFormat },
    { "day-narrow",     DateFormatSymbols::kCapContextUsageDayNarrow },
    { "day-standalone-except-narrow", DateFormatSymbols::kCapContextUsageDayStandalone },
    { "era-abbr",       DateFormatSymbols::kCapContextUsageEraAbbrev },
    { "era-name",       DateFormatSymbols::kCapContextUsageEraWide },
    { "era-narrow",     DateFormatSymbols::kCapContextUsageEraNarrow },
    { "metazone-long",  DateFormatSymbols::kCapContextUsageMetazoneLong },
    { "metazone-short", DateFormatSymbols::kCapContextUsageMetazoneShort },
    { "month-format-except-narrow", DateFormatSymbols::kCapContextUsageMonthFormat },
    { "month-narrow",   DateFormatSymbols::kCapContextUsageMonthNarrow },
    { "month-standalone-except-narrow", DateFormatSymbols::kCapContextUsageMonthStandalone },
    { "zone-long",      DateFormatSymbols::kCapContextUsageZoneLong },
    { "zone-short",     DateFormatSymbols::kCapContextUsageZoneShort },
    { NULL, (DateFormatSymbols::ECapitalizationContextUsageType)0 },
};

// Resource keys to look up localized strings for day periods.
// The first one must be midnight and the second must be noon, so that their indices coincide
// with the am/pm field. Formatting and parsing code for day periods relies on this coincidence.
static const char *dayPeriodKeys[] = {"midnight", "noon",
                         "morning1", "afternoon1", "evening1", "night1",
                         "morning2", "afternoon2", "evening2", "night2"};

UnicodeString* loadDayPeriodStrings(CalendarDataSink &sink, CharString &path,
                                    int32_t &stringCount,  UErrorCode &status) {
    if (U_FAILURE(status)) { return NULL; }

    UnicodeString pathUString(path.data(), -1, US_INV);
    Hashtable* map = static_cast<Hashtable*>(sink.maps.get(pathUString));

    stringCount = UPRV_LENGTHOF(dayPeriodKeys);
    UnicodeString *strings = new UnicodeString[stringCount];
    if (strings == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    if (map != NULL) {
        for (int32_t i = 0; i < stringCount; ++i) {
            UnicodeString dayPeriodKey(dayPeriodKeys[i], -1, US_INV);
            UnicodeString *dayPeriod = static_cast<UnicodeString*>(map->get(dayPeriodKey));
            if (dayPeriod != NULL) {
                strings[i].fastCopyFrom(*dayPeriod);
            } else {
                strings[i].setToBogus();
            }
        }
    } else {
        for (int32_t i = 0; i < stringCount; i++) {
            strings[i].setToBogus();
        }
    }
    return strings;
}


void
DateFormatSymbols::initializeData(const Locale& locale, const char *type, UErrorCode& status, UBool useLastResortData)
{
    int32_t len = 0;
    /* In case something goes wrong, initialize all of the data to NULL. */
    fEras = NULL;
    fErasCount = 0;
    fEraNames = NULL;
    fEraNamesCount = 0;
    fNarrowEras = NULL;
    fNarrowErasCount = 0;
    fMonths = NULL;
    fMonthsCount=0;
    fShortMonths = NULL;
    fShortMonthsCount=0;
    fNarrowMonths = NULL;
    fNarrowMonthsCount=0;
    fStandaloneMonths = NULL;
    fStandaloneMonthsCount=0;
    fStandaloneShortMonths = NULL;
    fStandaloneShortMonthsCount=0;
    fStandaloneNarrowMonths = NULL;
    fStandaloneNarrowMonthsCount=0;
    fWeekdays = NULL;
    fWeekdaysCount=0;
    fShortWeekdays = NULL;
    fShortWeekdaysCount=0;
    fShorterWeekdays = NULL;
    fShorterWeekdaysCount=0;
    fNarrowWeekdays = NULL;
    fNarrowWeekdaysCount=0;
    fStandaloneWeekdays = NULL;
    fStandaloneWeekdaysCount=0;
    fStandaloneShortWeekdays = NULL;
    fStandaloneShortWeekdaysCount=0;
    fStandaloneShorterWeekdays = NULL;
    fStandaloneShorterWeekdaysCount=0;
    fStandaloneNarrowWeekdays = NULL;
    fStandaloneNarrowWeekdaysCount=0;
    fAmPms = NULL;
    fAmPmsCount=0;
    fNarrowAmPms = NULL;
    fNarrowAmPmsCount=0;
    fTimeSeparator.setToBogus();
    fQuarters = NULL;
    fQuartersCount = 0;
    fShortQuarters = NULL;
    fShortQuartersCount = 0;
    fStandaloneQuarters = NULL;
    fStandaloneQuartersCount = 0;
    fStandaloneShortQuarters = NULL;
    fStandaloneShortQuartersCount = 0;
    fLeapMonthPatterns = NULL;
    fLeapMonthPatternsCount = 0;
    fShortYearNames = NULL;
    fShortYearNamesCount = 0;
    fShortZodiacNames = NULL;
    fShortZodiacNamesCount = 0;
    fZoneStringsRowCount = 0;
    fZoneStringsColCount = 0;
    fZoneStrings = NULL;
    fLocaleZoneStrings = NULL;
    fAbbreviatedDayPeriods = NULL;
    fAbbreviatedDayPeriodsCount = 0;
    fWideDayPeriods = NULL;
    fWideDayPeriodsCount = 0;
    fNarrowDayPeriods = NULL;
    fNarrowDayPeriodsCount = 0;
    fStandaloneAbbreviatedDayPeriods = NULL;
    fStandaloneAbbreviatedDayPeriodsCount = 0;
    fStandaloneWideDayPeriods = NULL;
    fStandaloneWideDayPeriodsCount = 0;
    fStandaloneNarrowDayPeriods = NULL;
    fStandaloneNarrowDayPeriodsCount = 0;
    uprv_memset(fCapitalization, 0, sizeof(fCapitalization));

    // We need to preserve the requested locale for
    // lazy ZoneStringFormat instantiation.  ZoneStringFormat
    // is region sensitive, thus, bundle locale bundle's locale
    // is not sufficient.
    fZSFLocale = locale;

    if (U_FAILURE(status)) return;

    // Create a CalendarDataSink to process this data and the resouce bundles
    CalendarDataSink calendarSink(status);
    UResourceBundle *rb = ures_open(NULL, locale.getBaseName(), &status);
    UResourceBundle *cb = ures_getByKey(rb, dtfmtsym_gCalendarTag, NULL, &status);

    if (U_FAILURE(status)) return;

    // Iterate over the resource bundle data following the fallbacks through different calendar types
    UnicodeString calendarType((type != NULL && *type != '\0')? type : dtfmtsym_gGregorianTag, -1, US_INV);
    while (!calendarType.isBogus()) {
        CharString calendarTypeBuffer;
        calendarTypeBuffer.appendInvariantChars(calendarType, status);
        if (U_FAILURE(status)) { return; }
        const char *calendarTypeCArray = calendarTypeBuffer.data();

        // Enumerate this calendar type. If the calendar is not found fallback to gregorian
        UErrorCode oldStatus = status;
        UResourceBundle *ctb = ures_getByKeyWithFallback(cb, calendarTypeCArray, NULL, &status);
        if (status == U_MISSING_RESOURCE_ERROR) {
            ures_close(ctb);
            if (uprv_strcmp(calendarTypeCArray, dtfmtsym_gGregorianTag) != 0) {
                calendarType.setTo(FALSE, kGregorianTagUChar, UPRV_LENGTHOF(kGregorianTagUChar));
                calendarSink.visitAllResources();
                status = oldStatus;
                continue;
            }
            return;
        }

        calendarSink.preEnumerate(calendarType);
        ures_getAllItemsWithFallback(ctb, "", calendarSink, status);
        ures_close(ctb);
        if (U_FAILURE(status)) break;

        // Stop loading when gregorian was loaded
        if (uprv_strcmp(calendarTypeCArray, dtfmtsym_gGregorianTag) == 0) {
            break;
        }

        // Get the next calendar type to process from the sink
        calendarType = calendarSink.nextCalendarType;

        // Gregorian is always the last fallback
        if (calendarType.isBogus()) {
            calendarType.setTo(FALSE, kGregorianTagUChar, UPRV_LENGTHOF(kGregorianTagUChar));
            calendarSink.visitAllResources();
        }
    }

    // CharString object to build paths
    CharString path;

    // Load Leap Month Patterns
    UErrorCode tempStatus = status;
    fLeapMonthPatterns = newUnicodeStringArray(kMonthPatternsCount);
    if (fLeapMonthPatterns) {
        initLeapMonthPattern(fLeapMonthPatterns, kLeapMonthPatternFormatWide, calendarSink,
                             buildResourcePath(path, dtfmtsym_gMonthPatternsTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesWideTag, tempStatus), tempStatus);
        initLeapMonthPattern(fLeapMonthPatterns, kLeapMonthPatternFormatAbbrev, calendarSink,
                             buildResourcePath(path, dtfmtsym_gMonthPatternsTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesAbbrTag, tempStatus), tempStatus);
        initLeapMonthPattern(fLeapMonthPatterns, kLeapMonthPatternFormatNarrow, calendarSink,
                             buildResourcePath(path, dtfmtsym_gMonthPatternsTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesNarrowTag, tempStatus), tempStatus);
        initLeapMonthPattern(fLeapMonthPatterns, kLeapMonthPatternStandaloneWide, calendarSink,
                             buildResourcePath(path, dtfmtsym_gMonthPatternsTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesWideTag, tempStatus), tempStatus);
        initLeapMonthPattern(fLeapMonthPatterns, kLeapMonthPatternStandaloneAbbrev, calendarSink,
                             buildResourcePath(path, dtfmtsym_gMonthPatternsTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesAbbrTag, tempStatus), tempStatus);
        initLeapMonthPattern(fLeapMonthPatterns, kLeapMonthPatternStandaloneNarrow, calendarSink,
                             buildResourcePath(path, dtfmtsym_gMonthPatternsTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesNarrowTag, tempStatus), tempStatus);
        initLeapMonthPattern(fLeapMonthPatterns, kLeapMonthPatternNumeric, calendarSink,
                             buildResourcePath(path, dtfmtsym_gMonthPatternsTag, dtfmtsym_gNamesNumericTag, dtfmtsym_gNamesAllTag, tempStatus), tempStatus);
        if (U_SUCCESS(tempStatus)) {
            // Hack to fix bad C inheritance for dangi monthPatterns (OK in J); this should be handled by aliases in root, but isn't.
            // The ordering of the following statements is important.
            if (fLeapMonthPatterns[kLeapMonthPatternFormatAbbrev].isEmpty()) {
                fLeapMonthPatterns[kLeapMonthPatternFormatAbbrev].setTo(fLeapMonthPatterns[kLeapMonthPatternFormatWide]);
            }
            if (fLeapMonthPatterns[kLeapMonthPatternFormatNarrow].isEmpty()) {
                fLeapMonthPatterns[kLeapMonthPatternFormatNarrow].setTo(fLeapMonthPatterns[kLeapMonthPatternStandaloneNarrow]);
            }
            if (fLeapMonthPatterns[kLeapMonthPatternStandaloneWide].isEmpty()) {
                fLeapMonthPatterns[kLeapMonthPatternStandaloneWide].setTo(fLeapMonthPatterns[kLeapMonthPatternFormatWide]);
            }
            if (fLeapMonthPatterns[kLeapMonthPatternStandaloneAbbrev].isEmpty()) {
                fLeapMonthPatterns[kLeapMonthPatternStandaloneAbbrev].setTo(fLeapMonthPatterns[kLeapMonthPatternFormatAbbrev]);
            }
            // end of hack
            fLeapMonthPatternsCount = kMonthPatternsCount;
        } else {
            delete[] fLeapMonthPatterns;
            fLeapMonthPatterns = NULL;
        }
    }

    // Load cyclic names sets
    tempStatus = status;
    initField(&fShortYearNames, fShortYearNamesCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gCyclicNameSetsTag, dtfmtsym_gNameSetYearsTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesAbbrTag, tempStatus), tempStatus);
    initField(&fShortZodiacNames, fShortZodiacNamesCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gCyclicNameSetsTag, dtfmtsym_gNameSetZodiacsTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesAbbrTag, tempStatus), tempStatus);

    // Load context transforms and capitalization
    tempStatus = U_ZERO_ERROR;
    UResourceBundle *localeBundle = ures_open(NULL, locale.getName(), &tempStatus);
    if (U_SUCCESS(tempStatus)) {
        UResourceBundle *contextTransforms = ures_getByKeyWithFallback(localeBundle, gContextTransformsTag, NULL, &tempStatus);
        if (U_SUCCESS(tempStatus)) {
            UResourceBundle *contextTransformUsage;
            while ( (contextTransformUsage = ures_getNextResource(contextTransforms, NULL, &tempStatus)) != NULL ) {
                const int32_t * intVector = ures_getIntVector(contextTransformUsage, &len, &status);
                if (U_SUCCESS(tempStatus) && intVector != NULL && len >= 2) {
                    const char* usageType = ures_getKey(contextTransformUsage);
                    if (usageType != NULL) {
                        const ContextUsageTypeNameToEnumValue * typeMapPtr = contextUsageTypeMap;
                        int32_t compResult = 0;
                        // linear search; list is short and we cannot be sure that bsearch is available
                        while ( typeMapPtr->usageTypeName != NULL && (compResult = uprv_strcmp(usageType, typeMapPtr->usageTypeName)) > 0 ) {
                            ++typeMapPtr;
                        }
                        if (typeMapPtr->usageTypeName != NULL && compResult == 0) {
                            fCapitalization[typeMapPtr->usageTypeEnumValue][0] = static_cast<UBool>(intVector[0]);
                            fCapitalization[typeMapPtr->usageTypeEnumValue][1] = static_cast<UBool>(intVector[1]);
                        }
                    }
                }
                tempStatus = U_ZERO_ERROR;
                ures_close(contextTransformUsage);
            }
            ures_close(contextTransforms);
        }

        tempStatus = U_ZERO_ERROR;
        const LocalPointer<NumberingSystem> numberingSystem(
                NumberingSystem::createInstance(locale, tempStatus), tempStatus);
        if (U_SUCCESS(tempStatus)) {
            // These functions all fail gracefully if passed NULL pointers and
            // do nothing unless U_SUCCESS(tempStatus), so it's only necessary
            // to check for errors once after all calls are made.
            const LocalUResourceBundlePointer numberElementsData(ures_getByKeyWithFallback(
                    localeBundle, dtfmtsym_gNumberElementsTag, NULL, &tempStatus));
            const LocalUResourceBundlePointer nsNameData(ures_getByKeyWithFallback(
                    numberElementsData.getAlias(), numberingSystem->getName(), NULL, &tempStatus));
            const LocalUResourceBundlePointer symbolsData(ures_getByKeyWithFallback(
                    nsNameData.getAlias(), dtfmtsym_gSymbolsTag, NULL, &tempStatus));
            fTimeSeparator = ures_getUnicodeStringByKey(
                    symbolsData.getAlias(), dtfmtsym_gTimeSeparatorTag, &tempStatus);
            if (U_FAILURE(tempStatus)) {
                fTimeSeparator.setToBogus();
            }
        }

        ures_close(localeBundle);
    }

    if (fTimeSeparator.isBogus()) {
        fTimeSeparator.setTo(DateFormatSymbols::DEFAULT_TIME_SEPARATOR);
    }

    // Load day periods
    fWideDayPeriods = loadDayPeriodStrings(calendarSink,
                            buildResourcePath(path, dtfmtsym_gDayPeriodTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesWideTag, status),
                            fWideDayPeriodsCount, status);
    fNarrowDayPeriods = loadDayPeriodStrings(calendarSink,
                            buildResourcePath(path, dtfmtsym_gDayPeriodTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesNarrowTag, status),
                            fNarrowDayPeriodsCount, status);
    fAbbreviatedDayPeriods = loadDayPeriodStrings(calendarSink,
                            buildResourcePath(path, dtfmtsym_gDayPeriodTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesAbbrTag, status),
                            fAbbreviatedDayPeriodsCount, status);
    fStandaloneWideDayPeriods = loadDayPeriodStrings(calendarSink,
                            buildResourcePath(path, dtfmtsym_gDayPeriodTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesWideTag, status),
                            fStandaloneWideDayPeriodsCount, status);
    fStandaloneNarrowDayPeriods = loadDayPeriodStrings(calendarSink,
                            buildResourcePath(path, dtfmtsym_gDayPeriodTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesNarrowTag, status),
                            fStandaloneNarrowDayPeriodsCount, status);
    fStandaloneAbbreviatedDayPeriods = loadDayPeriodStrings(calendarSink,
                            buildResourcePath(path, dtfmtsym_gDayPeriodTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesAbbrTag, status),
                            fStandaloneAbbreviatedDayPeriodsCount, status);

    U_LOCALE_BASED(locBased, *this);
    // if we make it to here, the resource data is cool, and we can get everything out
    // of it that we need except for the time-zone and localized-pattern data, which
    // are stored in a separate file
    locBased.setLocaleIDs(ures_getLocaleByType(cb, ULOC_VALID_LOCALE, &status),
                          ures_getLocaleByType(cb, ULOC_ACTUAL_LOCALE, &status));

    // Load eras
    initField(&fEras, fErasCount, calendarSink, buildResourcePath(path, dtfmtsym_gErasTag, dtfmtsym_gNamesAbbrTag, status), status);
    UErrorCode oldStatus = status;
    initField(&fEraNames, fEraNamesCount, calendarSink, buildResourcePath(path, dtfmtsym_gErasTag, dtfmtsym_gNamesWideTag, status), status);
    if (status == U_MISSING_RESOURCE_ERROR) { // Workaround because eras/wide was omitted from CLDR 1.3
        status = oldStatus;
        assignArray(fEraNames, fEraNamesCount, fEras, fErasCount);
    }
    // current ICU4J falls back to abbreviated if narrow eras are missing, so we will too
    oldStatus = status;
    initField(&fNarrowEras, fNarrowErasCount, calendarSink, buildResourcePath(path, dtfmtsym_gErasTag, dtfmtsym_gNamesNarrowTag, status), status);
    if (status == U_MISSING_RESOURCE_ERROR) { // Workaround because eras/wide was omitted from CLDR 1.3
        status = oldStatus;
        assignArray(fNarrowEras, fNarrowErasCount, fEras, fErasCount);
    }

    // Load month names
    initField(&fMonths, fMonthsCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gMonthNamesTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesWideTag, status), status);
    initField(&fShortMonths, fShortMonthsCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gMonthNamesTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesAbbrTag, status), status);
    initField(&fStandaloneMonths, fStandaloneMonthsCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gMonthNamesTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesWideTag, status), status);
    if (status == U_MISSING_RESOURCE_ERROR) { /* If standalone/wide not available, use format/wide */
        status = U_ZERO_ERROR;
        assignArray(fStandaloneMonths, fStandaloneMonthsCount, fMonths, fMonthsCount);
    }
    initField(&fStandaloneShortMonths, fStandaloneShortMonthsCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gMonthNamesTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesAbbrTag, status), status);
    if (status == U_MISSING_RESOURCE_ERROR) { /* If standalone/abbreviated not available, use format/abbreviated */
        status = U_ZERO_ERROR;
        assignArray(fStandaloneShortMonths, fStandaloneShortMonthsCount, fShortMonths, fShortMonthsCount);
    }

    UErrorCode narrowMonthsEC = status;
    UErrorCode standaloneNarrowMonthsEC = status;
    initField(&fNarrowMonths, fNarrowMonthsCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gMonthNamesTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesNarrowTag, narrowMonthsEC), narrowMonthsEC);
    initField(&fStandaloneNarrowMonths, fStandaloneNarrowMonthsCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gMonthNamesTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesNarrowTag, narrowMonthsEC), standaloneNarrowMonthsEC);
    if (narrowMonthsEC == U_MISSING_RESOURCE_ERROR && standaloneNarrowMonthsEC != U_MISSING_RESOURCE_ERROR) {
        // If format/narrow not available, use standalone/narrow
        assignArray(fNarrowMonths, fNarrowMonthsCount, fStandaloneNarrowMonths, fStandaloneNarrowMonthsCount);
    } else if (narrowMonthsEC != U_MISSING_RESOURCE_ERROR && standaloneNarrowMonthsEC == U_MISSING_RESOURCE_ERROR) {
        // If standalone/narrow not availabe, use format/narrow
        assignArray(fStandaloneNarrowMonths, fStandaloneNarrowMonthsCount, fNarrowMonths, fNarrowMonthsCount);
    } else if (narrowMonthsEC == U_MISSING_RESOURCE_ERROR && standaloneNarrowMonthsEC == U_MISSING_RESOURCE_ERROR) {
        // If neither is available, use format/abbreviated
        assignArray(fNarrowMonths, fNarrowMonthsCount, fShortMonths, fShortMonthsCount);
        assignArray(fStandaloneNarrowMonths, fStandaloneNarrowMonthsCount, fShortMonths, fShortMonthsCount);
    }

    // Load AM/PM markers; if wide or narrow not available, use short
    UErrorCode ampmStatus = U_ZERO_ERROR;
    initField(&fAmPms, fAmPmsCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gAmPmMarkersTag, ampmStatus), ampmStatus);
    if (U_FAILURE(ampmStatus)) {
        initField(&fAmPms, fAmPmsCount, calendarSink,
                  buildResourcePath(path, dtfmtsym_gAmPmMarkersAbbrTag, status), status);
    }
    ampmStatus = U_ZERO_ERROR;
    initField(&fNarrowAmPms, fNarrowAmPmsCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gAmPmMarkersNarrowTag, ampmStatus), ampmStatus);
    if (U_FAILURE(ampmStatus)) {
        initField(&fNarrowAmPms, fNarrowAmPmsCount, calendarSink,
                  buildResourcePath(path, dtfmtsym_gAmPmMarkersAbbrTag, status), status);
    }

    // Load quarters
    initField(&fQuarters, fQuartersCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gQuartersTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesWideTag, status), status);
    initField(&fShortQuarters, fShortQuartersCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gQuartersTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesAbbrTag, status), status);

    initField(&fStandaloneQuarters, fStandaloneQuartersCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gQuartersTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesWideTag, status), status);
    if(status == U_MISSING_RESOURCE_ERROR) {
        status = U_ZERO_ERROR;
        assignArray(fStandaloneQuarters, fStandaloneQuartersCount, fQuarters, fQuartersCount);
    }
    initField(&fStandaloneShortQuarters, fStandaloneShortQuartersCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gQuartersTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesAbbrTag, status), status);
    if(status == U_MISSING_RESOURCE_ERROR) {
        status = U_ZERO_ERROR;
        assignArray(fStandaloneShortQuarters, fStandaloneShortQuartersCount, fShortQuarters, fShortQuartersCount);
    }

    // ICU 3.8 or later version no longer uses localized date-time pattern characters by default (ticket#5597)
    /*
    // fastCopyFrom()/setTo() - see assignArray comments
    resStr = ures_getStringByKey(fResourceBundle, gLocalPatternCharsTag, &len, &status);
    fLocalPatternChars.setTo(TRUE, resStr, len);
    // If the locale data does not include new pattern chars, use the defaults
    // TODO: Consider making this an error, since this may add conflicting characters.
    if (len < PATTERN_CHARS_LEN) {
        fLocalPatternChars.append(UnicodeString(TRUE, &gPatternChars[len], PATTERN_CHARS_LEN-len));
    }
    */
    fLocalPatternChars.setTo(TRUE, gPatternChars, PATTERN_CHARS_LEN);

    // Format wide weekdays -> fWeekdays
    // {sfb} fixed to handle 1-based weekdays
    initField(&fWeekdays, fWeekdaysCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gDayNamesTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesWideTag, status), 1, status);

    // Format abbreviated weekdays -> fShortWeekdays
    initField(&fShortWeekdays, fShortWeekdaysCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gDayNamesTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesAbbrTag, status), 1, status);

    // Format short weekdays -> fShorterWeekdays (fall back to abbreviated)
    initField(&fShorterWeekdays, fShorterWeekdaysCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gDayNamesTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesShortTag, status), 1, status);
    if (status == U_MISSING_RESOURCE_ERROR) {
        status = U_ZERO_ERROR;
        assignArray(fShorterWeekdays, fShorterWeekdaysCount, fShortWeekdays, fShortWeekdaysCount);
    }

    // Stand-alone wide weekdays -> fStandaloneWeekdays
    initField(&fStandaloneWeekdays, fStandaloneWeekdaysCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gDayNamesTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesWideTag, status), 1, status);
    if (status == U_MISSING_RESOURCE_ERROR) { /* If standalone/wide is not available, use format/wide */
        status = U_ZERO_ERROR;
        assignArray(fStandaloneWeekdays, fStandaloneWeekdaysCount, fWeekdays, fWeekdaysCount);
    }

    // Stand-alone abbreviated weekdays -> fStandaloneShortWeekdays
    initField(&fStandaloneShortWeekdays, fStandaloneShortWeekdaysCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gDayNamesTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesAbbrTag, status), 1, status);
    if (status == U_MISSING_RESOURCE_ERROR) { /* If standalone/abbreviated is not available, use format/abbreviated */
        status = U_ZERO_ERROR;
        assignArray(fStandaloneShortWeekdays, fStandaloneShortWeekdaysCount, fShortWeekdays, fShortWeekdaysCount);
    }

    // Stand-alone short weekdays -> fStandaloneShorterWeekdays (fall back to format abbreviated)
    initField(&fStandaloneShorterWeekdays, fStandaloneShorterWeekdaysCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gDayNamesTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesShortTag, status), 1, status);
    if (status == U_MISSING_RESOURCE_ERROR) { /* If standalone/short is not available, use format/short */
        status = U_ZERO_ERROR;
        assignArray(fStandaloneShorterWeekdays, fStandaloneShorterWeekdaysCount, fShorterWeekdays, fShorterWeekdaysCount);
    }

    // Format narrow weekdays -> fNarrowWeekdays
    UErrorCode narrowWeeksEC = status;
    initField(&fNarrowWeekdays, fNarrowWeekdaysCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gDayNamesTag, dtfmtsym_gNamesFormatTag, dtfmtsym_gNamesNarrowTag, status), 1, narrowWeeksEC);
    // Stand-alone narrow weekdays -> fStandaloneNarrowWeekdays
    UErrorCode standaloneNarrowWeeksEC = status;
    initField(&fStandaloneNarrowWeekdays, fStandaloneNarrowWeekdaysCount, calendarSink,
              buildResourcePath(path, dtfmtsym_gDayNamesTag, dtfmtsym_gNamesStandaloneTag, dtfmtsym_gNamesNarrowTag, status), 1, standaloneNarrowWeeksEC);

    if (narrowWeeksEC == U_MISSING_RESOURCE_ERROR && standaloneNarrowWeeksEC != U_MISSING_RESOURCE_ERROR) {
        // If format/narrow not available, use standalone/narrow
        assignArray(fNarrowWeekdays, fNarrowWeekdaysCount, fStandaloneNarrowWeekdays, fStandaloneNarrowWeekdaysCount);
    } else if (narrowWeeksEC != U_MISSING_RESOURCE_ERROR && standaloneNarrowWeeksEC == U_MISSING_RESOURCE_ERROR) {
        // If standalone/narrow not available, use format/narrow
        assignArray(fStandaloneNarrowWeekdays, fStandaloneNarrowWeekdaysCount, fNarrowWeekdays, fNarrowWeekdaysCount);
    } else if (narrowWeeksEC == U_MISSING_RESOURCE_ERROR && standaloneNarrowWeeksEC == U_MISSING_RESOURCE_ERROR ) {
        // If neither is available, use format/abbreviated
        assignArray(fNarrowWeekdays, fNarrowWeekdaysCount, fShortWeekdays, fShortWeekdaysCount);
        assignArray(fStandaloneNarrowWeekdays, fStandaloneNarrowWeekdaysCount, fShortWeekdays, fShortWeekdaysCount);
    }

    // Last resort fallback in case previous data wasn't loaded
    if (U_FAILURE(status))
    {
        if (useLastResortData)
        {
            // Handle the case in which there is no resource data present.
            // We don't have to generate usable patterns in this situation;
            // we just need to produce something that will be semi-intelligible
            // in most locales.

            status = U_USING_FALLBACK_WARNING;
            //TODO(fabalbon): make sure we are storing las resort data for all fields in here.
            initField(&fEras, fErasCount, (const UChar *)gLastResortEras, kEraNum, kEraLen, status);
            initField(&fEraNames, fEraNamesCount, (const UChar *)gLastResortEras, kEraNum, kEraLen, status);
            initField(&fNarrowEras, fNarrowErasCount, (const UChar *)gLastResortEras, kEraNum, kEraLen, status);
            initField(&fMonths, fMonthsCount, (const UChar *)gLastResortMonthNames, kMonthNum, kMonthLen,  status);
            initField(&fShortMonths, fShortMonthsCount, (const UChar *)gLastResortMonthNames, kMonthNum, kMonthLen, status);
            initField(&fNarrowMonths, fNarrowMonthsCount, (const UChar *)gLastResortMonthNames, kMonthNum, kMonthLen, status);
            initField(&fStandaloneMonths, fStandaloneMonthsCount, (const UChar *)gLastResortMonthNames, kMonthNum, kMonthLen,  status);
            initField(&fStandaloneShortMonths, fStandaloneShortMonthsCount, (const UChar *)gLastResortMonthNames, kMonthNum, kMonthLen, status);
            initField(&fStandaloneNarrowMonths, fStandaloneNarrowMonthsCount, (const UChar *)gLastResortMonthNames, kMonthNum, kMonthLen, status);
            initField(&fWeekdays, fWeekdaysCount, (const UChar *)gLastResortDayNames, kDayNum, kDayLen, status);
            initField(&fShortWeekdays, fShortWeekdaysCount, (const UChar *)gLastResortDayNames, kDayNum, kDayLen, status);
            initField(&fShorterWeekdays, fShorterWeekdaysCount, (const UChar *)gLastResortDayNames, kDayNum, kDayLen, status);
            initField(&fNarrowWeekdays, fNarrowWeekdaysCount, (const UChar *)gLastResortDayNames, kDayNum, kDayLen, status);
            initField(&fStandaloneWeekdays, fStandaloneWeekdaysCount, (const UChar *)gLastResortDayNames, kDayNum, kDayLen, status);
            initField(&fStandaloneShortWeekdays, fStandaloneShortWeekdaysCount, (const UChar *)gLastResortDayNames, kDayNum, kDayLen, status);
            initField(&fStandaloneShorterWeekdays, fStandaloneShorterWeekdaysCount, (const UChar *)gLastResortDayNames, kDayNum, kDayLen, status);
            initField(&fStandaloneNarrowWeekdays, fStandaloneNarrowWeekdaysCount, (const UChar *)gLastResortDayNames, kDayNum, kDayLen, status);
            initField(&fAmPms, fAmPmsCount, (const UChar *)gLastResortAmPmMarkers, kAmPmNum, kAmPmLen, status);
            initField(&fNarrowAmPms, fNarrowAmPmsCount, (const UChar *)gLastResortAmPmMarkers, kAmPmNum, kAmPmLen, status);
            initField(&fQuarters, fQuartersCount, (const UChar *)gLastResortQuarters, kQuarterNum, kQuarterLen, status);
            initField(&fShortQuarters, fShortQuartersCount, (const UChar *)gLastResortQuarters, kQuarterNum, kQuarterLen, status);
            initField(&fStandaloneQuarters, fStandaloneQuartersCount, (const UChar *)gLastResortQuarters, kQuarterNum, kQuarterLen, status);
            initField(&fStandaloneShortQuarters, fStandaloneShortQuartersCount, (const UChar *)gLastResortQuarters, kQuarterNum, kQuarterLen, status);
            fLocalPatternChars.setTo(TRUE, gPatternChars, PATTERN_CHARS_LEN);
        }
    }

    // Close resources
    ures_close(cb);
    ures_close(rb);
}

Locale
DateFormatSymbols::getLocale(ULocDataLocaleType type, UErrorCode& status) const {
    U_LOCALE_BASED(locBased, *this);
    return locBased.getLocale(type, status);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*******************************************************************************
* Copyright (C) 2008-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File DTITVFMT.CPP
*
*******************************************************************************
*/

  // for 'typeid' to work

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/********************************************************************************
* Copyright (C) 2008-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File DTITVFMT.H
*
*******************************************************************************
*/





#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Format and parse date interval in a language-independent manner.
 */

#if !UCONFIG_NO_FORMATTING




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2008-2016, International Business Machines Corporation and
 * others. All Rights Reserved.
 *******************************************************************************
 *
 * File DTITVINF.H
 *
 *******************************************************************************
 */




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Date/Time interval patterns for formatting date/time interval
 */

#if !UCONFIG_NO_FORMATTING






U_NAMESPACE_BEGIN

/**
 * DateIntervalInfo is a public class for encapsulating localizable
 * date time interval patterns. It is used by DateIntervalFormat.
 *
 * <P>
 * For most users, ordinary use of DateIntervalFormat does not need to create
 * DateIntervalInfo object directly.
 * DateIntervalFormat will take care of it when creating a date interval
 * formatter when user pass in skeleton and locale.
 *
 * <P>
 * For power users, who want to create their own date interval patterns,
 * or want to re-set date interval patterns, they could do so by
 * directly creating DateIntervalInfo and manupulating it.
 *
 * <P>
 * Logically, the interval patterns are mappings
 * from (skeleton, the_largest_different_calendar_field)
 * to (date_interval_pattern).
 *
 * <P>
 * A skeleton
 * <ol>
 * <li>
 * only keeps the field pattern letter and ignores all other parts
 * in a pattern, such as space, punctuations, and string literals.
 * <li>
 * hides the order of fields.
 * <li>
 * might hide a field's pattern letter length.
 *
 * For those non-digit calendar fields, the pattern letter length is
 * important, such as MMM, MMMM, and MMMMM; EEE and EEEE,
 * and the field's pattern letter length is honored.
 *
 * For the digit calendar fields,  such as M or MM, d or dd, yy or yyyy,
 * the field pattern length is ignored and the best match, which is defined
 * in date time patterns, will be returned without honor the field pattern
 * letter length in skeleton.
 * </ol>
 *
 * <P>
 * The calendar fields we support for interval formatting are:
 * year, month, date, day-of-week, am-pm, hour, hour-of-day, and minute.
 * Those calendar fields can be defined in the following order:
 * year >  month > date > am-pm > hour >  minute
 *
 * The largest different calendar fields between 2 calendars is the
 * first different calendar field in above order.
 *
 * For example: the largest different calendar fields between &quot;Jan 10, 2007&quot;
 * and &quot;Feb 20, 2008&quot; is year.
 *
 * <P>
 * There is a set of pre-defined static skeleton strings.
 * There are pre-defined interval patterns for those pre-defined skeletons
 * in locales' resource files.
 * For example, for a skeleton UDAT_YEAR_ABBR_MONTH_DAY, which is  &quot;yMMMd&quot;,
 * in  en_US, if the largest different calendar field between date1 and date2
 * is &quot;year&quot;, the date interval pattern  is &quot;MMM d, yyyy - MMM d, yyyy&quot;,
 * such as &quot;Jan 10, 2007 - Jan 10, 2008&quot;.
 * If the largest different calendar field between date1 and date2 is &quot;month&quot;,
 * the date interval pattern is &quot;MMM d - MMM d, yyyy&quot;,
 * such as &quot;Jan 10 - Feb 10, 2007&quot;.
 * If the largest different calendar field between date1 and date2 is &quot;day&quot;,
 * the date interval pattern is &quot;MMM d-d, yyyy&quot;, such as &quot;Jan 10-20, 2007&quot;.
 *
 * For date skeleton, the interval patterns when year, or month, or date is
 * different are defined in resource files.
 * For time skeleton, the interval patterns when am/pm, or hour, or minute is
 * different are defined in resource files.
 *
 *
 * <P>
 * There are 2 dates in interval pattern. For most locales, the first date
 * in an interval pattern is the earlier date. There might be a locale in which
 * the first date in an interval pattern is the later date.
 * We use fallback format for the default order for the locale.
 * For example, if the fallback format is &quot;{0} - {1}&quot;, it means
 * the first date in the interval pattern for this locale is earlier date.
 * If the fallback format is &quot;{1} - {0}&quot;, it means the first date is the
 * later date.
 * For a particular interval pattern, the default order can be overriden
 * by prefixing &quot;latestFirst:&quot; or &quot;earliestFirst:&quot; to the interval pattern.
 * For example, if the fallback format is &quot;{0}-{1}&quot;,
 * but for skeleton &quot;yMMMd&quot;, the interval pattern when day is different is
 * &quot;latestFirst:d-d MMM yy&quot;, it means by default, the first date in interval
 * pattern is the earlier date. But for skeleton &quot;yMMMd&quot;, when day is different,
 * the first date in &quot;d-d MMM yy&quot; is the later date.
 *
 * <P>
 * The recommended way to create a DateIntervalFormat object is to pass in
 * the locale.
 * By using a Locale parameter, the DateIntervalFormat object is
 * initialized with the pre-defined interval patterns for a given or
 * default locale.
 * <P>
 * Users can also create DateIntervalFormat object
 * by supplying their own interval patterns.
 * It provides flexibility for power users.
 *
 * <P>
 * After a DateIntervalInfo object is created, clients may modify
 * the interval patterns using setIntervalPattern function as so desired.
 * Currently, users can only set interval patterns when the following
 * calendar fields are different: ERA, YEAR, MONTH, DATE,  DAY_OF_MONTH,
 * DAY_OF_WEEK, AM_PM,  HOUR, HOUR_OF_DAY, and MINUTE.
 * Interval patterns when other calendar fields are different is not supported.
 * <P>
 * DateIntervalInfo objects are cloneable.
 * When clients obtain a DateIntervalInfo object,
 * they can feel free to modify it as necessary.
 * <P>
 * DateIntervalInfo are not expected to be subclassed.
 * Data for a calendar is loaded out of resource bundles.
 * Through ICU 4.4, date interval patterns are only supported in the Gregorian
 * calendar; non-Gregorian calendars are supported from ICU 4.4.1.
 * @stable ICU 4.0
**/
class U_I18N_API DateIntervalInfo U_FINAL : public UObject {
public:
    /**
     * Default constructor.
     * It does not initialize any interval patterns except
     * that it initialize default fall-back pattern as "{0} - {1}",
     * which can be reset by setFallbackIntervalPattern().
     * It should be followed by setFallbackIntervalPattern() and
     * setIntervalPattern(),
     * and is recommended to be used only for power users who
     * wants to create their own interval patterns and use them to create
     * date interval formatter.
     * @param status   output param set to success/failure code on exit
     * @internal ICU 4.0
     */
    DateIntervalInfo(UErrorCode& status);


    /**
     * Construct DateIntervalInfo for the given locale,
     * @param locale  the interval patterns are loaded from the appropriate calendar
     *                data (specified calendar or default calendar) in this locale.
     * @param status  output param set to success/failure code on exit
     * @stable ICU 4.0
     */
    DateIntervalInfo(const Locale& locale, UErrorCode& status);


    /**
     * Copy constructor.
     * @stable ICU 4.0
     */
    DateIntervalInfo(const DateIntervalInfo&);

    /**
     * Assignment operator
     * @stable ICU 4.0
     */
    DateIntervalInfo& operator=(const DateIntervalInfo&);

    /**
     * Clone this object polymorphically.
     * The caller owns the result and should delete it when done.
     * @return   a copy of the object
     * @stable ICU 4.0
     */
    virtual DateIntervalInfo* clone() const;

    /**
     * Destructor.
     * It is virtual to be safe, but it is not designed to be subclassed.
     * @stable ICU 4.0
     */
    virtual ~DateIntervalInfo();


    /**
     * Return true if another object is semantically equal to this one.
     *
     * @param other    the DateIntervalInfo object to be compared with.
     * @return         true if other is semantically equal to this.
     * @stable ICU 4.0
     */
    virtual UBool operator==(const DateIntervalInfo& other) const;

    /**
     * Return true if another object is semantically unequal to this one.
     *
     * @param other    the DateIntervalInfo object to be compared with.
     * @return         true if other is semantically unequal to this.
     * @stable ICU 4.0
     */
    UBool operator!=(const DateIntervalInfo& other) const;



    /**
     * Provides a way for client to build interval patterns.
     * User could construct DateIntervalInfo by providing a list of skeletons
     * and their patterns.
     * <P>
     * For example:
     * <pre>
     * UErrorCode status = U_ZERO_ERROR;
     * DateIntervalInfo dIntervalInfo = new DateIntervalInfo();
     * dIntervalInfo->setFallbackIntervalPattern("{0} ~ {1}");
     * dIntervalInfo->setIntervalPattern("yMd", UCAL_YEAR, "'from' yyyy-M-d 'to' yyyy-M-d", status);
     * dIntervalInfo->setIntervalPattern("yMMMd", UCAL_MONTH, "'from' yyyy MMM d 'to' MMM d", status);
     * dIntervalInfo->setIntervalPattern("yMMMd", UCAL_DAY, "yyyy MMM d-d", status, status);
     * </pre>
     *
     * Restriction:
     * Currently, users can only set interval patterns when the following
     * calendar fields are different: ERA, YEAR, MONTH, DATE,  DAY_OF_MONTH,
     * DAY_OF_WEEK, AM_PM,  HOUR, HOUR_OF_DAY, and MINUTE.
     * Interval patterns when other calendar fields are different are
     * not supported.
     *
     * @param skeleton         the skeleton on which interval pattern based
     * @param lrgDiffCalUnit   the largest different calendar unit.
     * @param intervalPattern  the interval pattern on the largest different
     *                         calendar unit.
     *                         For example, if lrgDiffCalUnit is
     *                         "year", the interval pattern for en_US when year
     *                         is different could be "'from' yyyy 'to' yyyy".
     * @param status           output param set to success/failure code on exit
     * @stable ICU 4.0
     */
    void setIntervalPattern(const UnicodeString& skeleton,
                            UCalendarDateFields lrgDiffCalUnit,
                            const UnicodeString& intervalPattern,
                            UErrorCode& status);

    /**
     * Get the interval pattern given skeleton and
     * the largest different calendar field.
     * @param skeleton   the skeleton
     * @param field      the largest different calendar field
     * @param result     output param to receive the pattern
     * @param status     output param set to success/failure code on exit
     * @return a reference to 'result'
     * @stable ICU 4.0
     */
    UnicodeString& getIntervalPattern(const UnicodeString& skeleton,
                                      UCalendarDateFields field,
                                      UnicodeString& result,
                                      UErrorCode& status) const;

    /**
     * Get the fallback interval pattern.
     * @param  result   output param to receive the pattern
     * @return a reference to 'result'
     * @stable ICU 4.0
     */
    UnicodeString& getFallbackIntervalPattern(UnicodeString& result) const;


    /**
     * Re-set the fallback interval pattern.
     *
     * In construction, default fallback pattern is set as "{0} - {1}".
     * And constructor taking locale as parameter will set the
     * fallback pattern as what defined in the locale resource file.
     *
     * This method provides a way for user to replace the fallback pattern.
     *
     * @param fallbackPattern  fall-back interval pattern.
     * @param status           output param set to success/failure code on exit
     * @stable ICU 4.0
     */
    void setFallbackIntervalPattern(const UnicodeString& fallbackPattern,
                                    UErrorCode& status);


    /** Get default order -- whether the first date in pattern is later date
                             or not.
     * return default date ordering in interval pattern. TRUE if the first date
     *        in pattern is later date, FALSE otherwise.
     * @stable ICU 4.0
     */
    UBool getDefaultOrder() const;


    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 4.0
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 4.0
     */
    static UClassID U_EXPORT2 getStaticClassID();


private:
    /**
     * DateIntervalFormat will need access to
     * getBestSkeleton(), parseSkeleton(), enum IntervalPatternIndex,
     * and calendarFieldToPatternIndex().
     *
     * Instead of making above public,
     * make DateIntervalFormat a friend of DateIntervalInfo.
     */
    friend class DateIntervalFormat;

    /**
     * Internal struct used to load resource bundle data.
     */
    struct DateIntervalSink;

    /**
     * Following is for saving the interval patterns.
     * We only support interval patterns on
     * ERA, YEAR, MONTH, DAY, AM_PM, HOUR, and MINUTE
     */
    enum IntervalPatternIndex
    {
        kIPI_ERA,
        kIPI_YEAR,
        kIPI_MONTH,
        kIPI_DATE,
        kIPI_AM_PM,
        kIPI_HOUR,
        kIPI_MINUTE,
        kIPI_SECOND,
        kIPI_MAX_INDEX
    };
public:
#ifndef U_HIDE_INTERNAL_API
    /**
     * Max index for stored interval patterns
     * @internal ICU 4.4
     */
     enum {
         kMaxIntervalPatternIndex = kIPI_MAX_INDEX
     };
#endif  /* U_HIDE_INTERNAL_API */
private:


    /**
     * Initialize the DateIntervalInfo from locale
     * @param locale   the given locale.
     * @param status   output param set to success/failure code on exit
     */
    void initializeData(const Locale& locale, UErrorCode& status);


    /* Set Interval pattern.
     *
     * It sets interval pattern into the hash map.
     *
     * @param skeleton         skeleton on which the interval pattern based
     * @param lrgDiffCalUnit   the largest different calendar unit.
     * @param intervalPattern  the interval pattern on the largest different
     *                         calendar unit.
     * @param status           output param set to success/failure code on exit
     */
    void setIntervalPatternInternally(const UnicodeString& skeleton,
                                      UCalendarDateFields lrgDiffCalUnit,
                                      const UnicodeString& intervalPattern,
                                      UErrorCode& status);


    /**given an input skeleton, get the best match skeleton
     * which has pre-defined interval pattern in resource file.
     * Also return the difference between the input skeleton
     * and the best match skeleton.
     *
     * TODO (xji): set field weight or
     *             isolate the funtionality in DateTimePatternGenerator
     * @param  skeleton               input skeleton
     * @param  bestMatchDistanceInfo  the difference between input skeleton
     *                                and best match skeleton.
     *         0, if there is exact match for input skeleton
     *         1, if there is only field width difference between
     *            the best match and the input skeleton
     *         2, the only field difference is 'v' and 'z'
     *        -1, if there is calendar field difference between
     *            the best match and the input skeleton
     * @return                        best match skeleton
     */
    const UnicodeString* getBestSkeleton(const UnicodeString& skeleton,
                                         int8_t& bestMatchDistanceInfo) const;


    /**
     * Parse skeleton, save each field's width.
     * It is used for looking for best match skeleton,
     * and adjust pattern field width.
     * @param skeleton            skeleton to be parsed
     * @param skeletonFieldWidth  parsed skeleton field width
     */
    static void U_EXPORT2 parseSkeleton(const UnicodeString& skeleton,
                                        int32_t* skeletonFieldWidth);


    /**
     * Check whether one field width is numeric while the other is string.
     *
     * TODO (xji): make it general
     *
     * @param fieldWidth          one field width
     * @param anotherFieldWidth   another field width
     * @param patternLetter       pattern letter char
     * @return true if one field width is numeric and the other is string,
     *         false otherwise.
     */
    static UBool U_EXPORT2 stringNumeric(int32_t fieldWidth,
                                         int32_t anotherFieldWidth,
                                         char patternLetter);


    /**
     * Convert calendar field to the interval pattern index in
     * hash table.
     *
     * Since we only support the following calendar fields:
     * ERA, YEAR, MONTH, DATE,  DAY_OF_MONTH, DAY_OF_WEEK,
     * AM_PM,  HOUR, HOUR_OF_DAY, and MINUTE,
     * We reserve only 4 interval patterns for a skeleton.
     *
     * @param field    calendar field
     * @param status   output param set to success/failure code on exit
     * @return  interval pattern index in hash table
     */
    static IntervalPatternIndex U_EXPORT2 calendarFieldToIntervalIndex(
                                                      UCalendarDateFields field,
                                                      UErrorCode& status);


    /**
     * delete hash table (of type fIntervalPatterns).
     *
     * @param hTable  hash table to be deleted
     */
    void deleteHash(Hashtable* hTable);


    /**
     * initialize hash table (of type fIntervalPatterns).
     *
     * @param status   output param set to success/failure code on exit
     * @return         hash table initialized
     */
    Hashtable* initHash(UErrorCode& status);



    /**
     * copy hash table (of type fIntervalPatterns).
     *
     * @param source   the source to copy from
     * @param target   the target to copy to
     * @param status   output param set to success/failure code on exit
     */
    void copyHash(const Hashtable* source, Hashtable* target, UErrorCode& status);


    // data members
    // fallback interval pattern
    UnicodeString fFallbackIntervalPattern;
    // default order
    UBool fFirstDateInPtnIsLaterDate;

    // HashMap<UnicodeString, UnicodeString[kIPI_MAX_INDEX]>
    // HashMap( skeleton, pattern[largest_different_field] )
    Hashtable* fIntervalPatterns;

};// end class DateIntervalInfo


inline UBool
DateIntervalInfo::operator!=(const DateIntervalInfo& other) const {
    return !operator==(other);
}


U_NAMESPACE_END

#endif

#endif /* U_SHOW_CPLUSPLUS_API */





U_NAMESPACE_BEGIN


class FormattedDateIntervalData;
class DateIntervalFormat;

#ifndef U_HIDE_DRAFT_API
/**
 * An immutable class containing the result of a date interval formatting operation.
 *
 * Instances of this class are immutable and thread-safe.
 *
 * When calling nextPosition():
 * The fields are returned from left to right. The special field category
 * UFIELD_CATEGORY_DATE_INTERVAL_SPAN is used to indicate which datetime
 * primitives came from which arguments: 0 means fromCalendar, and 1 means
 * toCalendar. The span category will always occur before the
 * corresponding fields in UFIELD_CATEGORY_DATE
 * in the nextPosition() iterator.
 *
 * Not intended for public subclassing.
 *
 * @draft ICU 64
 */
class U_I18N_API FormattedDateInterval : public UMemory, public FormattedValue {
  public:
    /**
     * Default constructor; makes an empty FormattedDateInterval.
     * @draft ICU 64
     */
    FormattedDateInterval() : fData(nullptr), fErrorCode(U_INVALID_STATE_ERROR) {}

    /**
     * Move constructor: Leaves the source FormattedDateInterval in an undefined state.
     * @draft ICU 64
     */
    FormattedDateInterval(FormattedDateInterval&& src) U_NOEXCEPT;

    /**
     * Destruct an instance of FormattedDateInterval.
     * @draft ICU 64
     */
    virtual ~FormattedDateInterval() U_OVERRIDE;

    /** Copying not supported; use move constructor instead. */
    FormattedDateInterval(const FormattedDateInterval&) = delete;

    /** Copying not supported; use move assignment instead. */
    FormattedDateInterval& operator=(const FormattedDateInterval&) = delete;

    /**
     * Move assignment: Leaves the source FormattedDateInterval in an undefined state.
     * @draft ICU 64
     */
    FormattedDateInterval& operator=(FormattedDateInterval&& src) U_NOEXCEPT;

    /** @copydoc FormattedValue::toString() */
    UnicodeString toString(UErrorCode& status) const U_OVERRIDE;

    /** @copydoc FormattedValue::toTempString() */
    UnicodeString toTempString(UErrorCode& status) const U_OVERRIDE;

    /** @copydoc FormattedValue::appendTo() */
    Appendable &appendTo(Appendable& appendable, UErrorCode& status) const U_OVERRIDE;

    /** @copydoc FormattedValue::nextPosition() */
    UBool nextPosition(ConstrainedFieldPosition& cfpos, UErrorCode& status) const U_OVERRIDE;

  private:
    FormattedDateIntervalData *fData;
    UErrorCode fErrorCode;
    explicit FormattedDateInterval(FormattedDateIntervalData *results)
        : fData(results), fErrorCode(U_ZERO_ERROR) {}
    explicit FormattedDateInterval(UErrorCode errorCode)
        : fData(nullptr), fErrorCode(errorCode) {}
    friend class DateIntervalFormat;
};
#endif /* U_HIDE_DRAFT_API */


/**
 * DateIntervalFormat is a class for formatting and parsing date
 * intervals in a language-independent manner.
 * Only formatting is supported, parsing is not supported.
 *
 * <P>
 * Date interval means from one date to another date,
 * for example, from "Jan 11, 2008" to "Jan 18, 2008".
 * We introduced class DateInterval to represent it.
 * DateInterval is a pair of UDate, which is
 * the standard milliseconds since 24:00 GMT, Jan 1, 1970.
 *
 * <P>
 * DateIntervalFormat formats a DateInterval into
 * text as compactly as possible.
 * For example, the date interval format from "Jan 11, 2008" to "Jan 18,. 2008"
 * is "Jan 11-18, 2008" for English.
 * And it parses text into DateInterval,
 * although initially, parsing is not supported.
 *
 * <P>
 * There is no structural information in date time patterns.
 * For any punctuations and string literals inside a date time pattern,
 * we do not know whether it is just a separator, or a prefix, or a suffix.
 * Without such information, so, it is difficult to generate a sub-pattern
 * (or super-pattern) by algorithm.
 * So, formatting a DateInterval is pattern-driven. It is very
 * similar to formatting in SimpleDateFormat.
 * We introduce class DateIntervalInfo to save date interval
 * patterns, similar to date time pattern in SimpleDateFormat.
 *
 * <P>
 * Logically, the interval patterns are mappings
 * from (skeleton, the_largest_different_calendar_field)
 * to (date_interval_pattern).
 *
 * <P>
 * A skeleton
 * <ol>
 * <li>
 * only keeps the field pattern letter and ignores all other parts
 * in a pattern, such as space, punctuations, and string literals.
 * </li>
 * <li>
 * hides the order of fields.
 * </li>
 * <li>
 * might hide a field's pattern letter length.
 * </li>
 * </ol>
 *
 * For those non-digit calendar fields, the pattern letter length is
 * important, such as MMM, MMMM, and MMMMM; EEE and EEEE,
 * and the field's pattern letter length is honored.
 *
 * For the digit calendar fields,  such as M or MM, d or dd, yy or yyyy,
 * the field pattern length is ignored and the best match, which is defined
 * in date time patterns, will be returned without honor the field pattern
 * letter length in skeleton.
 *
 * <P>
 * The calendar fields we support for interval formatting are:
 * year, month, date, day-of-week, am-pm, hour, hour-of-day, minute, and second
 * (though we do not currently have specific intervalFormat date for skeletons
 * with seconds).
 * Those calendar fields can be defined in the following order:
 * year >  month > date > hour (in day) >  minute > second
 *
 * The largest different calendar fields between 2 calendars is the
 * first different calendar field in above order.
 *
 * For example: the largest different calendar fields between "Jan 10, 2007"
 * and "Feb 20, 2008" is year.
 *
 * <P>
 * For other calendar fields, the compact interval formatting is not
 * supported. And the interval format will be fall back to fall-back
 * patterns, which is mostly "{date0} - {date1}".
 *
 * <P>
 * There is a set of pre-defined static skeleton strings.
 * There are pre-defined interval patterns for those pre-defined skeletons
 * in locales' resource files.
 * For example, for a skeleton UDAT_YEAR_ABBR_MONTH_DAY, which is  &quot;yMMMd&quot;,
 * in  en_US, if the largest different calendar field between date1 and date2
 * is &quot;year&quot;, the date interval pattern  is &quot;MMM d, yyyy - MMM d, yyyy&quot;,
 * such as &quot;Jan 10, 2007 - Jan 10, 2008&quot;.
 * If the largest different calendar field between date1 and date2 is &quot;month&quot;,
 * the date interval pattern is &quot;MMM d - MMM d, yyyy&quot;,
 * such as &quot;Jan 10 - Feb 10, 2007&quot;.
 * If the largest different calendar field between date1 and date2 is &quot;day&quot;,
 * the date interval pattern is &quot;MMM d-d, yyyy&quot;, such as &quot;Jan 10-20, 2007&quot;.
 *
 * For date skeleton, the interval patterns when year, or month, or date is
 * different are defined in resource files.
 * For time skeleton, the interval patterns when am/pm, or hour, or minute is
 * different are defined in resource files.
 *
 * <P>
 * If a skeleton is not found in a locale's DateIntervalInfo, which means
 * the interval patterns for the skeleton is not defined in resource file,
 * the interval pattern will falls back to the interval "fallback" pattern
 * defined in resource file.
 * If the interval "fallback" pattern is not defined, the default fall-back
 * is "{date0} - {data1}".
 *
 * <P>
 * For the combination of date and time,
 * The rule to generate interval patterns are:
 * <ol>
 * <li>
 *    when the year, month, or day differs, falls back to fall-back
 *    interval pattern, which mostly is the concatenate the two original
 *    expressions with a separator between,
 *    For example, interval pattern from "Jan 10, 2007 10:10 am"
 *    to "Jan 11, 2007 10:10am" is
 *    "Jan 10, 2007 10:10 am - Jan 11, 2007 10:10am"
 * </li>
 * <li>
 *    otherwise, present the date followed by the range expression
 *    for the time.
 *    For example, interval pattern from "Jan 10, 2007 10:10 am"
 *    to "Jan 10, 2007 11:10am" is "Jan 10, 2007 10:10 am - 11:10am"
 * </li>
 * </ol>
 *
 *
 * <P>
 * If two dates are the same, the interval pattern is the single date pattern.
 * For example, interval pattern from "Jan 10, 2007" to "Jan 10, 2007" is
 * "Jan 10, 2007".
 *
 * Or if the presenting fields between 2 dates have the exact same values,
 * the interval pattern is the  single date pattern.
 * For example, if user only requests year and month,
 * the interval pattern from "Jan 10, 2007" to "Jan 20, 2007" is "Jan 2007".
 *
 * <P>
 * DateIntervalFormat needs the following information for correct
 * formatting: time zone, calendar type, pattern, date format symbols,
 * and date interval patterns.
 * It can be instantiated in 2 ways:
 * <ol>
 * <li>
 *    create an instance using default or given locale plus given skeleton.
 *    Users are encouraged to created date interval formatter this way and
 *    to use the pre-defined skeleton macros, such as
 *    UDAT_YEAR_NUM_MONTH, which consists the calendar fields and
 *    the format style.
 * </li>
 * <li>
 *    create an instance using default or given locale plus given skeleton
 *    plus a given DateIntervalInfo.
 *    This factory method is for powerful users who want to provide their own
 *    interval patterns.
 *    Locale provides the timezone, calendar, and format symbols information.
 *    Local plus skeleton provides full pattern information.
 *    DateIntervalInfo provides the date interval patterns.
 * </li>
 * </ol>
 *
 * <P>
 * For the calendar field pattern letter, such as G, y, M, d, a, h, H, m, s etc.
 * DateIntervalFormat uses the same syntax as that of
 * DateTime format.
 *
 * <P>
 * Code Sample: general usage
 * <pre>
 * \code
 *   // the date interval object which the DateIntervalFormat formats on
 *   // and parses into
 *   DateInterval*  dtInterval = new DateInterval(1000*3600*24, 1000*3600*24*2);
 *   UErrorCode status = U_ZERO_ERROR;
 *   DateIntervalFormat* dtIntervalFmt = DateIntervalFormat::createInstance(
 *                           UDAT_YEAR_MONTH_DAY,
 *                           Locale("en", "GB", ""), status);
 *   UnicodeUnicodeString dateIntervalString;
 *   FieldPosition pos = 0;
 *   // formatting
 *   dtIntervalFmt->format(dtInterval, dateIntervalUnicodeString, pos, status);
 *   delete dtIntervalFmt;
 * \endcode
 * </pre>
 */
class U_I18N_API DateIntervalFormat : public Format {
public:

    /**
     * Construct a DateIntervalFormat from skeleton and  the default locale.
     *
     * This is a convenient override of
     * createInstance(const UnicodeString& skeleton, const Locale& locale,
     *                UErrorCode&)
     * with the value of locale as default locale.
     *
     * @param skeleton  the skeleton on which interval format based.
     * @param status    output param set to success/failure code on exit
     * @return          a date time interval formatter which the caller owns.
     * @stable ICU 4.0
     */
    static DateIntervalFormat* U_EXPORT2 createInstance(
                                               const UnicodeString& skeleton,
                                               UErrorCode& status);

    /**
     * Construct a DateIntervalFormat from skeleton and a given locale.
     * <P>
     * In this factory method,
     * the date interval pattern information is load from resource files.
     * Users are encouraged to created date interval formatter this way and
     * to use the pre-defined skeleton macros.
     *
     * <P>
     * There are pre-defined skeletons (defined in udate.h) having predefined
     * interval patterns in resource files.
     * Users are encouraged to use those macros.
     * For example:
     * DateIntervalFormat::createInstance(UDAT_MONTH_DAY, status)
     *
     * The given Locale provides the interval patterns.
     * For example, for en_GB, if skeleton is UDAT_YEAR_ABBR_MONTH_WEEKDAY_DAY,
     * which is "yMMMEEEd",
     * the interval patterns defined in resource file to above skeleton are:
     * "EEE, d MMM, yyyy - EEE, d MMM, yyyy" for year differs,
     * "EEE, d MMM - EEE, d MMM, yyyy" for month differs,
     * "EEE, d - EEE, d MMM, yyyy" for day differs,
     * @param skeleton  the skeleton on which the interval format is based.
     * @param locale    the given locale
     * @param status    output param set to success/failure code on exit
     * @return          a date time interval formatter which the caller owns.
     * @stable ICU 4.0
	 * <p>
	 * <h4>Sample code</h4>
	 * \snippet samples/dtitvfmtsample/dtitvfmtsample.cpp dtitvfmtPreDefined1
	 * \snippet samples/dtitvfmtsample/dtitvfmtsample.cpp dtitvfmtPreDefined
	 * <p>
     */

    static DateIntervalFormat* U_EXPORT2 createInstance(
                                               const UnicodeString& skeleton,
                                               const Locale& locale,
                                               UErrorCode& status);

    /**
     * Construct a DateIntervalFormat from skeleton
     *  DateIntervalInfo, and default locale.
     *
     * This is a convenient override of
     * createInstance(const UnicodeString& skeleton, const Locale& locale,
     *                const DateIntervalInfo& dtitvinf, UErrorCode&)
     * with the locale value as default locale.
     *
     * @param skeleton  the skeleton on which interval format based.
     * @param dtitvinf  the DateIntervalInfo object.
     * @param status    output param set to success/failure code on exit
     * @return          a date time interval formatter which the caller owns.
     * @stable ICU 4.0
     */
    static DateIntervalFormat* U_EXPORT2 createInstance(
                                              const UnicodeString& skeleton,
                                              const DateIntervalInfo& dtitvinf,
                                              UErrorCode& status);

    /**
     * Construct a DateIntervalFormat from skeleton
     * a DateIntervalInfo, and the given locale.
     *
     * <P>
     * In this factory method, user provides its own date interval pattern
     * information, instead of using those pre-defined data in resource file.
     * This factory method is for powerful users who want to provide their own
     * interval patterns.
     * <P>
     * There are pre-defined skeletons (defined in udate.h) having predefined
     * interval patterns in resource files.
     * Users are encouraged to use those macros.
     * For example:
     * DateIntervalFormat::createInstance(UDAT_MONTH_DAY, status)
     *
     * The DateIntervalInfo provides the interval patterns.
     * and the DateIntervalInfo ownership remains to the caller.
     *
     * User are encouraged to set default interval pattern in DateIntervalInfo
     * as well, if they want to set other interval patterns ( instead of
     * reading the interval patterns from resource files).
     * When the corresponding interval pattern for a largest calendar different
     * field is not found ( if user not set it ), interval format fallback to
     * the default interval pattern.
     * If user does not provide default interval pattern, it fallback to
     * "{date0} - {date1}"
     *
     * @param skeleton  the skeleton on which interval format based.
     * @param locale    the given locale
     * @param dtitvinf  the DateIntervalInfo object.
     * @param status    output param set to success/failure code on exit
     * @return          a date time interval formatter which the caller owns.
     * @stable ICU 4.0
	 * <p>
	 * <h4>Sample code</h4>
	 * \snippet samples/dtitvfmtsample/dtitvfmtsample.cpp dtitvfmtPreDefined1
	 * \snippet samples/dtitvfmtsample/dtitvfmtsample.cpp dtitvfmtCustomized
	 * <p>
     */
    static DateIntervalFormat* U_EXPORT2 createInstance(
                                              const UnicodeString& skeleton,
                                              const Locale& locale,
                                              const DateIntervalInfo& dtitvinf,
                                              UErrorCode& status);

    /**
     * Destructor.
     * @stable ICU 4.0
     */
    virtual ~DateIntervalFormat();

    /**
     * Clone this Format object polymorphically. The caller owns the result and
     * should delete it when done.
     * @return    A copy of the object.
     * @stable ICU 4.0
     */
    virtual DateIntervalFormat* clone() const;

    /**
     * Return true if the given Format objects are semantically equal. Objects
     * of different subclasses are considered unequal.
     * @param other    the object to be compared with.
     * @return         true if the given Format objects are semantically equal.
     * @stable ICU 4.0
     */
    virtual UBool operator==(const Format& other) const;

    /**
     * Return true if the given Format objects are not semantically equal.
     * Objects of different subclasses are considered unequal.
     * @param other the object to be compared with.
     * @return      true if the given Format objects are not semantically equal.
     * @stable ICU 4.0
     */
    UBool operator!=(const Format& other) const;


    using Format::format;

    /**
     * Format an object to produce a string. This method handles Formattable
     * objects with a DateInterval type.
     * If a the Formattable object type is not a DateInterval,
     * then it returns a failing UErrorCode.
     *
     * @param obj               The object to format.
     *                          Must be a DateInterval.
     * @param appendTo          Output parameter to receive result.
     *                          Result is appended to existing contents.
     * @param fieldPosition     On input: an alignment field, if desired.
     *                          On output: the offsets of the alignment field.
     *                          There may be multiple instances of a given field type
     *                          in an interval format; in this case the fieldPosition
     *                          offsets refer to the first instance.
     * @param status            Output param filled with success/failure status.
     * @return                  Reference to 'appendTo' parameter.
     * @stable ICU 4.0
     */
    virtual UnicodeString& format(const Formattable& obj,
                                  UnicodeString& appendTo,
                                  FieldPosition& fieldPosition,
                                  UErrorCode& status) const ;



    /**
     * Format a DateInterval to produce a string.
     *
     * @param dtInterval        DateInterval to be formatted.
     * @param appendTo          Output parameter to receive result.
     *                          Result is appended to existing contents.
     * @param fieldPosition     On input: an alignment field, if desired.
     *                          On output: the offsets of the alignment field.
     *                          There may be multiple instances of a given field type
     *                          in an interval format; in this case the fieldPosition
     *                          offsets refer to the first instance.
     * @param status            Output param filled with success/failure status.
     * @return                  Reference to 'appendTo' parameter.
     * @stable ICU 4.0
     */
    UnicodeString& format(const DateInterval* dtInterval,
                          UnicodeString& appendTo,
                          FieldPosition& fieldPosition,
                          UErrorCode& status) const ;

#ifndef U_HIDE_DRAFT_API
    /**
     * Format a DateInterval to produce a FormattedDateInterval.
     *
     * The FormattedDateInterval exposes field information about the formatted string.
     *
     * @param dtInterval        DateInterval to be formatted.
     * @param status            Set if an error occurs.
     * @return                  A FormattedDateInterval containing the format result.
     * @draft ICU 64
     */
    FormattedDateInterval formatToValue(
        const DateInterval& dtInterval,
        UErrorCode& status) const;
#endif /* U_HIDE_DRAFT_API */

    /**
     * Format 2 Calendars to produce a string.
     *
     * Note: "fromCalendar" and "toCalendar" are not const,
     * since calendar is not const in  SimpleDateFormat::format(Calendar&),
     *
     * @param fromCalendar      calendar set to the from date in date interval
     *                          to be formatted into date interval string
     * @param toCalendar        calendar set to the to date in date interval
     *                          to be formatted into date interval string
     * @param appendTo          Output parameter to receive result.
     *                          Result is appended to existing contents.
     * @param fieldPosition     On input: an alignment field, if desired.
     *                          On output: the offsets of the alignment field.
     *                          There may be multiple instances of a given field type
     *                          in an interval format; in this case the fieldPosition
     *                          offsets refer to the first instance.
     * @param status            Output param filled with success/failure status.
     *                          Caller needs to make sure it is SUCCESS
     *                          at the function entrance
     * @return                  Reference to 'appendTo' parameter.
     * @stable ICU 4.0
     */
    UnicodeString& format(Calendar& fromCalendar,
                          Calendar& toCalendar,
                          UnicodeString& appendTo,
                          FieldPosition& fieldPosition,
                          UErrorCode& status) const ;

#ifndef U_HIDE_DRAFT_API
    /**
     * Format 2 Calendars to produce a FormattedDateInterval.
     *
     * The FormattedDateInterval exposes field information about the formatted string.
     *
     * Note: "fromCalendar" and "toCalendar" are not const,
     * since calendar is not const in  SimpleDateFormat::format(Calendar&),
     *
     * @param fromCalendar      calendar set to the from date in date interval
     *                          to be formatted into date interval string
     * @param toCalendar        calendar set to the to date in date interval
     *                          to be formatted into date interval string
     * @param status            Set if an error occurs.
     * @return                  A FormattedDateInterval containing the format result.
     * @draft ICU 64
     */
    FormattedDateInterval formatToValue(
        Calendar& fromCalendar,
        Calendar& toCalendar,
        UErrorCode& status) const;
#endif /* U_HIDE_DRAFT_API */

    /**
     * Date interval parsing is not supported. Please do not use.
     * <P>
     * This method should handle parsing of
     * date time interval strings into Formattable objects with
     * DateInterval type, which is a pair of UDate.
     * <P>
     * Before calling, set parse_pos.index to the offset you want to start
     * parsing at in the source. After calling, parse_pos.index is the end of
     * the text you parsed. If error occurs, index is unchanged.
     * <P>
     * When parsing, leading whitespace is discarded (with a successful parse),
     * while trailing whitespace is left as is.
     * <P>
     * See Format::parseObject() for more.
     *
     * @param source    The string to be parsed into an object.
     * @param result    Formattable to be set to the parse result.
     *                  If parse fails, return contents are undefined.
     * @param parse_pos The position to start parsing at. Since no parsing
     *                  is supported, upon return this param is unchanged.
     * @return          A newly created Formattable* object, or NULL
     *                  on failure.  The caller owns this and should
     *                  delete it when done.
     * @internal ICU 4.0
     */
    virtual void parseObject(const UnicodeString& source,
                             Formattable& result,
                             ParsePosition& parse_pos) const;


    /**
     * Gets the date time interval patterns.
     * @return the date time interval patterns associated with
     * this date interval formatter.
     * @stable ICU 4.0
     */
    const DateIntervalInfo* getDateIntervalInfo(void) const;


    /**
     * Set the date time interval patterns.
     * @param newIntervalPatterns   the given interval patterns to copy.
     * @param status          output param set to success/failure code on exit
     * @stable ICU 4.0
     */
    void setDateIntervalInfo(const DateIntervalInfo& newIntervalPatterns,
                             UErrorCode& status);


    /**
     * Gets the date formatter. The DateIntervalFormat instance continues to own
     * the returned DateFormatter object, and will use and possibly modify it
     * during format operations. In a multi-threaded environment, the returned
     * DateFormat can only be used if it is certain that no other threads are
     * concurrently using this DateIntervalFormatter, even for nominally const
     * functions.
     *
     * @return the date formatter associated with this date interval formatter.
     * @stable ICU 4.0
     */
    const DateFormat* getDateFormat(void) const;

    /**
     * Returns a reference to the TimeZone used by this DateIntervalFormat's calendar.
     * @return the time zone associated with the calendar of DateIntervalFormat.
     * @stable ICU 4.8
     */
    virtual const TimeZone& getTimeZone(void) const;

    /**
     * Sets the time zone for the calendar used by this DateIntervalFormat object. The
     * caller no longer owns the TimeZone object and should not delete it after this call.
     * @param zoneToAdopt the TimeZone to be adopted.
     * @stable ICU 4.8
     */
    virtual void adoptTimeZone(TimeZone* zoneToAdopt);

    /**
     * Sets the time zone for the calendar used by this DateIntervalFormat object.
     * @param zone the new time zone.
     * @stable ICU 4.8
     */
    virtual void setTimeZone(const TimeZone& zone);

    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 4.0
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 4.0
     */
    virtual UClassID getDynamicClassID(void) const;

protected:

    /**
     * Copy constructor.
     * @stable ICU 4.0
     */
    DateIntervalFormat(const DateIntervalFormat&);

    /**
     * Assignment operator.
     * @stable ICU 4.0
     */
    DateIntervalFormat& operator=(const DateIntervalFormat&);

private:

    /*
     * This is for ICU internal use only. Please do not use.
     * Save the interval pattern information.
     * Interval pattern consists of 2 single date patterns and the separator.
     * For example, interval pattern "MMM d - MMM d, yyyy" consists
     * a single date pattern "MMM d", another single date pattern "MMM d, yyyy",
     * and a separator "-".
     * The pattern is divided into 2 parts. For above example,
     * the first part is "MMM d - ", and the second part is "MMM d, yyyy".
     * Also, the first date appears in an interval pattern could be
     * the earlier date or the later date.
     * And such information is saved in the interval pattern as well.
     */
    struct PatternInfo {
        UnicodeString firstPart;
        UnicodeString secondPart;
        /**
         * Whether the first date in interval pattern is later date or not.
         * Fallback format set the default ordering.
         * And for a particular interval pattern, the order can be
         * overriden by prefixing the interval pattern with "latestFirst:" or
         * "earliestFirst:"
         * For example, given 2 date, Jan 10, 2007 to Feb 10, 2007.
         * if the fallback format is "{0} - {1}",
         * and the pattern is "d MMM - d MMM yyyy", the interval format is
         * "10 Jan - 10 Feb, 2007".
         * If the pattern is "latestFirst:d MMM - d MMM yyyy",
         * the interval format is "10 Feb - 10 Jan, 2007"
         */
        UBool         laterDateFirst;
    };


    /**
     * default constructor
     * @internal (private)
     */
    DateIntervalFormat();

    /**
     * Construct a DateIntervalFormat from DateFormat,
     * a DateIntervalInfo, and skeleton.
     * DateFormat provides the timezone, calendar,
     * full pattern, and date format symbols information.
     * It should be a SimpleDateFormat object which
     * has a pattern in it.
     * the DateIntervalInfo provides the interval patterns.
     *
     * Note: the DateIntervalFormat takes ownership of both
     * DateFormat and DateIntervalInfo objects.
     * Caller should not delete them.
     *
     * @param locale    the locale of this date interval formatter.
     * @param dtItvInfo the DateIntervalInfo object to be adopted.
     * @param skeleton  the skeleton of the date formatter
     * @param status    output param set to success/failure code on exit
     */
    DateIntervalFormat(const Locale& locale, DateIntervalInfo* dtItvInfo,
                       const UnicodeString* skeleton, UErrorCode& status);


    /**
     * Construct a DateIntervalFormat from DateFormat
     * and a DateIntervalInfo.
     *
     * It is a wrapper of the constructor.
     *
     * @param locale    the locale of this date interval formatter.
     * @param dtitvinf  the DateIntervalInfo object to be adopted.
     * @param skeleton  the skeleton of this formatter.
     * @param status    Output param set to success/failure code.
     * @return          a date time interval formatter which the caller owns.
     */
    static DateIntervalFormat* U_EXPORT2 create(const Locale& locale,
                                                DateIntervalInfo* dtitvinf,
                                                const UnicodeString* skeleton,
                                                UErrorCode& status);

    /**
     *  Below are for generating interval patterns local to the formatter
     */

    /** Like fallbackFormat, but only formats the range part of the fallback. */
    void fallbackFormatRange(
        Calendar& fromCalendar,
        Calendar& toCalendar,
        UnicodeString& appendTo,
        int8_t& firstIndex,
        FieldPositionHandler& fphandler,
        UErrorCode& status) const;

    /**
     * Format 2 Calendars using fall-back interval pattern
     *
     * The full pattern used in this fall-back format is the
     * full pattern of the date formatter.
     *
     * gFormatterMutex must already be locked when calling this function.
     *
     * @param fromCalendar      calendar set to the from date in date interval
     *                          to be formatted into date interval string
     * @param toCalendar        calendar set to the to date in date interval
     *                          to be formatted into date interval string
     * @param fromToOnSameDay   TRUE iff from and to dates are on the same day
     *                          (any difference is in ampm/hours or below)
     * @param appendTo          Output parameter to receive result.
     *                          Result is appended to existing contents.
     * @param firstIndex        See formatImpl for more information.
     * @param fphandler         See formatImpl for more information.
     * @param status            output param set to success/failure code on exit
     * @return                  Reference to 'appendTo' parameter.
     * @internal (private)
     */
    UnicodeString& fallbackFormat(Calendar& fromCalendar,
                                  Calendar& toCalendar,
                                  UBool fromToOnSameDay,
                                  UnicodeString& appendTo,
                                  int8_t& firstIndex,
                                  FieldPositionHandler& fphandler,
                                  UErrorCode& status) const;



    /**
     * Initialize interval patterns locale to this formatter
     *
     * This code is a bit complicated since
     * 1. the interval patterns saved in resource bundle files are interval
     *    patterns based on date or time only.
     *    It does not have interval patterns based on both date and time.
     *    Interval patterns on both date and time are algorithm generated.
     *
     *    For example, it has interval patterns on skeleton "dMy" and "hm",
     *    but it does not have interval patterns on skeleton "dMyhm".
     *
     *    The rule to generate interval patterns for both date and time skeleton are
     *    1) when the year, month, or day differs, concatenate the two original
     *    expressions with a separator between,
     *    For example, interval pattern from "Jan 10, 2007 10:10 am"
     *    to "Jan 11, 2007 10:10am" is
     *    "Jan 10, 2007 10:10 am - Jan 11, 2007 10:10am"
     *
     *    2) otherwise, present the date followed by the range expression
     *    for the time.
     *    For example, interval pattern from "Jan 10, 2007 10:10 am"
     *    to "Jan 10, 2007 11:10am" is
     *    "Jan 10, 2007 10:10 am - 11:10am"
     *
     * 2. even a pattern does not request a certain calendar field,
     *    the interval pattern needs to include such field if such fields are
     *    different between 2 dates.
     *    For example, a pattern/skeleton is "hm", but the interval pattern
     *    includes year, month, and date when year, month, and date differs.
     *
     *
     * @param status    output param set to success/failure code on exit
     */
    void initializePattern(UErrorCode& status);



    /**
     * Set fall back interval pattern given a calendar field,
     * a skeleton, and a date time pattern generator.
     * @param field      the largest different calendar field
     * @param skeleton   a skeleton
     * @param status     output param set to success/failure code on exit
     */
    void setFallbackPattern(UCalendarDateFields field,
                            const UnicodeString& skeleton,
                            UErrorCode& status);



    /**
     * get separated date and time skeleton from a combined skeleton.
     *
     * The difference between date skeleton and normalizedDateSkeleton are:
     * 1. both 'y' and 'd' are appeared only once in normalizeDateSkeleton
     * 2. 'E' and 'EE' are normalized into 'EEE'
     * 3. 'MM' is normalized into 'M'
     *
     ** the difference between time skeleton and normalizedTimeSkeleton are:
     * 1. both 'H' and 'h' are normalized as 'h' in normalized time skeleton,
     * 2. 'a' is omitted in normalized time skeleton.
     * 3. there is only one appearance for 'h', 'm','v', 'z' in normalized time
     *    skeleton
     *
     *
     *  @param skeleton               given combined skeleton.
     *  @param date                   Output parameter for date only skeleton.
     *  @param normalizedDate         Output parameter for normalized date only
     *
     *  @param time                   Output parameter for time only skeleton.
     *  @param normalizedTime         Output parameter for normalized time only
     *                                skeleton.
     *
     */
    static void  U_EXPORT2 getDateTimeSkeleton(const UnicodeString& skeleton,
                                    UnicodeString& date,
                                    UnicodeString& normalizedDate,
                                    UnicodeString& time,
                                    UnicodeString& normalizedTime);



    /**
     * Generate date or time interval pattern from resource,
     * and set them into the interval pattern locale to this formatter.
     *
     * It needs to handle the following:
     * 1. need to adjust field width.
     *    For example, the interval patterns saved in DateIntervalInfo
     *    includes "dMMMy", but not "dMMMMy".
     *    Need to get interval patterns for dMMMMy from dMMMy.
     *    Another example, the interval patterns saved in DateIntervalInfo
     *    includes "hmv", but not "hmz".
     *    Need to get interval patterns for "hmz' from 'hmv'
     *
     * 2. there might be no pattern for 'y' differ for skeleton "Md",
     *    in order to get interval patterns for 'y' differ,
     *    need to look for it from skeleton 'yMd'
     *
     * @param dateSkeleton   normalized date skeleton
     * @param timeSkeleton   normalized time skeleton
     * @return               whether the resource is found for the skeleton.
     *                       TRUE if interval pattern found for the skeleton,
     *                       FALSE otherwise.
     */
    UBool setSeparateDateTimePtn(const UnicodeString& dateSkeleton,
                                 const UnicodeString& timeSkeleton);




    /**
     * Generate interval pattern from existing resource
     *
     * It not only save the interval patterns,
     * but also return the extended skeleton and its best match skeleton.
     *
     * @param field           largest different calendar field
     * @param skeleton        skeleton
     * @param bestSkeleton    the best match skeleton which has interval pattern
     *                        defined in resource
     * @param differenceInfo  the difference between skeleton and best skeleton
     *         0 means the best matched skeleton is the same as input skeleton
     *         1 means the fields are the same, but field width are different
     *         2 means the only difference between fields are v/z,
     *        -1 means there are other fields difference
     *
     * @param extendedSkeleton      extended skeleton
     * @param extendedBestSkeleton  extended best match skeleton
     * @return                      whether the interval pattern is found
     *                              through extending skeleton or not.
     *                              TRUE if interval pattern is found by
     *                              extending skeleton, FALSE otherwise.
     */
    UBool setIntervalPattern(UCalendarDateFields field,
                             const UnicodeString* skeleton,
                             const UnicodeString* bestSkeleton,
                             int8_t differenceInfo,
                             UnicodeString* extendedSkeleton = NULL,
                             UnicodeString* extendedBestSkeleton = NULL);

    /**
     * Adjust field width in best match interval pattern to match
     * the field width in input skeleton.
     *
     * TODO (xji) make a general solution
     * The adjusting rule can be:
     * 1. always adjust
     * 2. never adjust
     * 3. default adjust, which means adjust according to the following rules
     * 3.1 always adjust string, such as MMM and MMMM
     * 3.2 never adjust between string and numeric, such as MM and MMM
     * 3.3 always adjust year
     * 3.4 do not adjust 'd', 'h', or 'm' if h presents
     * 3.5 do not adjust 'M' if it is numeric(?)
     *
     * Since date interval format is well-formed format,
     * date and time skeletons are normalized previously,
     * till this stage, the adjust here is only "adjust strings, such as MMM
     * and MMMM, EEE and EEEE.
     *
     * @param inputSkeleton            the input skeleton
     * @param bestMatchSkeleton        the best match skeleton
     * @param bestMatchIntervalPattern the best match interval pattern
     * @param differenceInfo           the difference between 2 skeletons
     *                                 1 means only field width differs
     *                                 2 means v/z exchange
     * @param adjustedIntervalPattern  adjusted interval pattern
     */
    static void U_EXPORT2 adjustFieldWidth(
                            const UnicodeString& inputSkeleton,
                            const UnicodeString& bestMatchSkeleton,
                            const UnicodeString& bestMatchIntervalPattern,
                            int8_t differenceInfo,
                            UnicodeString& adjustedIntervalPattern);

    /**
     * Concat a single date pattern with a time interval pattern,
     * set it into the intervalPatterns, while field is time field.
     * This is used to handle time interval patterns on skeleton with
     * both time and date. Present the date followed by
     * the range expression for the time.
     * @param format         date and time format
     * @param datePattern    date pattern
     * @param field          time calendar field: AM_PM, HOUR, MINUTE
     * @param status         output param set to success/failure code on exit
     */
    void concatSingleDate2TimeInterval(UnicodeString& format,
                                       const UnicodeString& datePattern,
                                       UCalendarDateFields field,
                                       UErrorCode& status);

    /**
     * check whether a calendar field present in a skeleton.
     * @param field      calendar field need to check
     * @param skeleton   given skeleton on which to check the calendar field
     * @return           true if field present in a skeleton.
     */
    static UBool U_EXPORT2 fieldExistsInSkeleton(UCalendarDateFields field,
                                                 const UnicodeString& skeleton);


    /**
     * Split interval patterns into 2 part.
     * @param intervalPattern  interval pattern
     * @return the index in interval pattern which split the pattern into 2 part
     */
    static int32_t  U_EXPORT2 splitPatternInto2Part(const UnicodeString& intervalPattern);


    /**
     * Break interval patterns as 2 part and save them into pattern info.
     * @param field            calendar field
     * @param intervalPattern  interval pattern
     */
    void setIntervalPattern(UCalendarDateFields field,
                            const UnicodeString& intervalPattern);


    /**
     * Break interval patterns as 2 part and save them into pattern info.
     * @param field            calendar field
     * @param intervalPattern  interval pattern
     * @param laterDateFirst   whether later date appear first in interval pattern
     */
    void setIntervalPattern(UCalendarDateFields field,
                            const UnicodeString& intervalPattern,
                            UBool laterDateFirst);


    /**
     * Set pattern information.
     *
     * @param field            calendar field
     * @param firstPart        the first part in interval pattern
     * @param secondPart       the second part in interval pattern
     * @param laterDateFirst   whether the first date in intervalPattern
     *                         is earlier date or later date
     */
    void setPatternInfo(UCalendarDateFields field,
                        const UnicodeString* firstPart,
                        const UnicodeString* secondPart,
                        UBool laterDateFirst);

    /**
     * Format 2 Calendars to produce a string.
     * Implementation of the similar public format function.
     * Must be called with gFormatterMutex already locked.
     *
     * Note: "fromCalendar" and "toCalendar" are not const,
     * since calendar is not const in  SimpleDateFormat::format(Calendar&),
     *
     * @param fromCalendar      calendar set to the from date in date interval
     *                          to be formatted into date interval string
     * @param toCalendar        calendar set to the to date in date interval
     *                          to be formatted into date interval string
     * @param appendTo          Output parameter to receive result.
     *                          Result is appended to existing contents.
     * @param firstIndex        0 if the first output date is fromCalendar;
     *                          1 if it corresponds to toCalendar;
     *                          -1 if there is only one date printed.
     * @param fphandler         Handler for field position information.
     *                          The fields will be from the UDateFormatField enum.
     * @param status            Output param filled with success/failure status.
     *                          Caller needs to make sure it is SUCCESS
     *                          at the function entrance
     * @return                  Reference to 'appendTo' parameter.
     * @internal (private)
     */
    UnicodeString& formatImpl(Calendar& fromCalendar,
                              Calendar& toCalendar,
                              UnicodeString& appendTo,
                              int8_t& firstIndex,
                              FieldPositionHandler& fphandler,
                              UErrorCode& status) const ;

    /** Version of formatImpl for DateInterval. */
    UnicodeString& formatIntervalImpl(const DateInterval& dtInterval,
                              UnicodeString& appendTo,
                              int8_t& firstIndex,
                              FieldPositionHandler& fphandler,
                              UErrorCode& status) const;


    // from calendar field to pattern letter
    static const char16_t fgCalendarFieldToPatternLetter[];


    /**
     * The interval patterns for this locale.
     */
    DateIntervalInfo*     fInfo;

    /**
     * The DateFormat object used to format single pattern
     */
    SimpleDateFormat*     fDateFormat;

    /**
     * The 2 calendars with the from and to date.
     * could re-use the calendar in fDateFormat,
     * but keeping 2 calendars make it clear and clean.
     */
    Calendar* fFromCalendar;
    Calendar* fToCalendar;

    Locale fLocale;

    /**
     * Following are interval information relevant (locale) to this formatter.
     */
    UnicodeString fSkeleton;
    PatternInfo fIntervalPatterns[DateIntervalInfo::kIPI_MAX_INDEX];

    /**
     * Patterns for fallback formatting.
     */
    UnicodeString* fDatePattern;
    UnicodeString* fTimePattern;
    UnicodeString* fDateTimeFormat;
};

inline UBool
DateIntervalFormat::operator!=(const Format& other) const  {
    return !operator==(other);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



#if !UCONFIG_NO_FORMATTING

//TODO: put in compilation
//#define DTITVFMT_DEBUG 1







// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File DTITV_IMPL.H
*
*******************************************************************************
*/



/**
 * \file
 * \brief C++ API: Defines macros for interval format implementation
 */

#if !UCONFIG_NO_FORMATTING




#define dtitv_impl_QUOTE             ((UChar)0x0027)
#define dtitv_impl_LOW_LINE          ((UChar)0x005F)
#define dtitv_impl_COLON             ((UChar)0x003A)
#define dtitv_impl_LEFT_CURLY_BRACKET  ((UChar)0x007B)
#define dtitv_impl_RIGHT_CURLY_BRACKET ((UChar)0x007D)
#define dtitv_impl_SPACE             ((UChar)0x0020)
#define dtitv_impl_EN_DASH           ((UChar)0x2013)
#define dtitv_impl_SOLIDUS           ((UChar)0x002F)

#define dtitv_impl_DIGIT_ZERO        ((UChar)0x0030)
#define dtitv_impl_DIGIT_ONE         ((UChar)0x0031)

#define dtitv_impl_LOW_A             ((UChar)0x0061)
#define dtitv_impl_LOW_B             ((UChar)0x0062)
#define dtitv_impl_LOW_C             ((UChar)0x0063)
#define dtitv_impl_LOW_D             ((UChar)0x0064)
#define dtitv_impl_LOW_E             ((UChar)0x0065)
#define dtitv_impl_LOW_F             ((UChar)0x0066)
#define dtitv_impl_LOW_G             ((UChar)0x0067)
#define dtitv_impl_LOW_H             ((UChar)0x0068)
#define dtitv_impl_LOW_I             ((UChar)0x0069)
#define dtitv_impl_LOW_J             ((UChar)0x006a)
#define dtitv_impl_LOW_K             ((UChar)0x006B)
#define dtitv_impl_LOW_L             ((UChar)0x006C)
#define dtitv_impl_LOW_M             ((UChar)0x006D)
#define dtitv_impl_LOW_N             ((UChar)0x006E)
#define dtitv_impl_LOW_O             ((UChar)0x006F)
#define dtitv_impl_LOW_P             ((UChar)0x0070)
#define dtitv_impl_LOW_Q             ((UChar)0x0071)
#define dtitv_impl_LOW_R             ((UChar)0x0072)
#define dtitv_impl_LOW_S             ((UChar)0x0073)
#define dtitv_impl_LOW_T             ((UChar)0x0074)
#define dtitv_impl_LOW_U             ((UChar)0x0075)
#define dtitv_impl_LOW_V             ((UChar)0x0076)
#define dtitv_impl_LOW_W             ((UChar)0x0077)
#define dtitv_impl_LOW_Y             ((UChar)0x0079)
#define dtitv_impl_LOW_Z             ((UChar)0x007A)

#define dtitv_impl_CAP_A             ((UChar)0x0041)
#define dtitv_impl_CAP_C             ((UChar)0x0043)
#define dtitv_impl_CAP_D             ((UChar)0x0044)
#define dtitv_impl_CAP_E             ((UChar)0x0045)
#define dtitv_impl_CAP_F             ((UChar)0x0046)
#define dtitv_impl_CAP_G             ((UChar)0x0047)
#define dtitv_impl_CAP_H             ((UChar)0x0048)
#define dtitv_impl_CAP_K             ((UChar)0x004B)
#define dtitv_impl_CAP_L             ((UChar)0x004C)
#define dtitv_impl_CAP_M             ((UChar)0x004D)
#define dtitv_impl_CAP_O             ((UChar)0x004F)
#define dtitv_impl_CAP_Q             ((UChar)0x0051)
#define dtitv_impl_CAP_S             ((UChar)0x0053)
#define dtitv_impl_CAP_T             ((UChar)0x0054)
#define dtitv_impl_CAP_U             ((UChar)0x0055)
#define dtitv_impl_CAP_V             ((UChar)0x0056)
#define dtitv_impl_CAP_W             ((UChar)0x0057)
#define dtitv_impl_CAP_Y             ((UChar)0x0059)
#define dtitv_impl_CAP_Z             ((UChar)0x005A)

//#define MINIMUM_SUPPORTED_CALENDAR_FIELD    UCAL_MINUTE

#define dtitv_impl_MAX_E_COUNT      5
#define dtitv_impl_MAX_M_COUNT      5
//#define MAX_INTERVAL_INDEX 4
#define dtitv_impl_MAX_POSITIVE_INT  56632


#endif /* #if !UCONFIG_NO_FORMATTING */






#ifdef DTITVFMT_DEBUG
#include <iostream>
#endif

U_NAMESPACE_BEGIN



#ifdef DTITVFMT_DEBUG
#define PRINTMESG(msg) { std::cout << "(" << __FILE__ << ":" << __LINE__ << ") " << msg << "\n"; }
#endif


static const UChar dtitvfmt_gDateFormatSkeleton[][11] = {
//yMMMMEEEEd
{dtitv_impl_LOW_Y, dtitv_impl_CAP_M, dtitv_impl_CAP_M, dtitv_impl_CAP_M, dtitv_impl_CAP_M, dtitv_impl_CAP_E, dtitv_impl_CAP_E, dtitv_impl_CAP_E, dtitv_impl_CAP_E, dtitv_impl_LOW_D, 0},
//yMMMMd
{dtitv_impl_LOW_Y, dtitv_impl_CAP_M, dtitv_impl_CAP_M, dtitv_impl_CAP_M, dtitv_impl_CAP_M, dtitv_impl_LOW_D, 0},
//yMMMd
{dtitv_impl_LOW_Y, dtitv_impl_CAP_M, dtitv_impl_CAP_M, dtitv_impl_CAP_M, dtitv_impl_LOW_D, 0},
//yMd
{dtitv_impl_LOW_Y, dtitv_impl_CAP_M, dtitv_impl_LOW_D, 0} };


static const char dtitvfmt_gCalendarTag[] = "calendar";
static const char dtitvfmt_gGregorianTag[] = "gregorian";
static const char dtitvfmt_gDateTimePatternsTag[] = "DateTimePatterns";


// latestFirst:
static const UChar dtitvfmt_gLaterFirstPrefix[] = {dtitv_impl_LOW_L, dtitv_impl_LOW_A, dtitv_impl_LOW_T, dtitv_impl_LOW_E, dtitv_impl_LOW_S,dtitv_impl_LOW_T, dtitv_impl_CAP_F, dtitv_impl_LOW_I, dtitv_impl_LOW_R, dtitv_impl_LOW_S, dtitv_impl_LOW_T, dtitv_impl_COLON};

// earliestFirst:
static const UChar dtitvfmt_gEarlierFirstPrefix[] = {dtitv_impl_LOW_E, dtitv_impl_LOW_A, dtitv_impl_LOW_R, dtitv_impl_LOW_L, dtitv_impl_LOW_I, dtitv_impl_LOW_E, dtitv_impl_LOW_S, dtitv_impl_LOW_T, dtitv_impl_CAP_F, dtitv_impl_LOW_I, dtitv_impl_LOW_R, dtitv_impl_LOW_S, dtitv_impl_LOW_T, dtitv_impl_COLON};


class FormattedDateIntervalData : public FormattedValueFieldPositionIteratorImpl {
public:
    FormattedDateIntervalData(UErrorCode& status) : FormattedValueFieldPositionIteratorImpl(5, status) {}
    virtual ~FormattedDateIntervalData();
};

FormattedDateIntervalData::~FormattedDateIntervalData() = default;

UPRV_FORMATTED_VALUE_SUBCLASS_AUTO_IMPL(FormattedDateInterval)


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DateIntervalFormat)

// Mutex, protects access to fDateFormat, fFromCalendar and fToCalendar.
//        Needed because these data members are modified by const methods of DateIntervalFormat.

static UMutex gFormatterMutex;

DateIntervalFormat* U_EXPORT2
DateIntervalFormat::createInstance(const UnicodeString& skeleton,
                                   UErrorCode& status) {
    return createInstance(skeleton, Locale::getDefault(), status);
}


DateIntervalFormat* U_EXPORT2
DateIntervalFormat::createInstance(const UnicodeString& skeleton,
                                   const Locale& locale,
                                   UErrorCode& status) {
#ifdef DTITVFMT_DEBUG
    char result[1000];
    char result_1[1000];
    char mesg[2000];
    skeleton.extract(0,  skeleton.length(), result, "UTF-8");
    UnicodeString pat;
    ((SimpleDateFormat*)dtfmt)->toPattern(pat);
    pat.extract(0,  pat.length(), result_1, "UTF-8");
    sprintf(mesg, "skeleton: %s; pattern: %s\n", result, result_1);
    PRINTMESG(mesg)
#endif

    DateIntervalInfo* dtitvinf = new DateIntervalInfo(locale, status);
    return create(locale, dtitvinf, &skeleton, status);
}



DateIntervalFormat* U_EXPORT2
DateIntervalFormat::createInstance(const UnicodeString& skeleton,
                                   const DateIntervalInfo& dtitvinf,
                                   UErrorCode& status) {
    return createInstance(skeleton, Locale::getDefault(), dtitvinf, status);
}


DateIntervalFormat* U_EXPORT2
DateIntervalFormat::createInstance(const UnicodeString& skeleton,
                                   const Locale& locale,
                                   const DateIntervalInfo& dtitvinf,
                                   UErrorCode& status) {
    DateIntervalInfo* ptn = dtitvinf.clone();
    return create(locale, ptn, &skeleton, status);
}


DateIntervalFormat::DateIntervalFormat()
:   fInfo(NULL),
    fDateFormat(NULL),
    fFromCalendar(NULL),
    fToCalendar(NULL),
    fLocale(Locale::getRoot()),
    fDatePattern(NULL),
    fTimePattern(NULL),
    fDateTimeFormat(NULL)
{}


DateIntervalFormat::DateIntervalFormat(const DateIntervalFormat& itvfmt)
:   Format(itvfmt),
    fInfo(NULL),
    fDateFormat(NULL),
    fFromCalendar(NULL),
    fToCalendar(NULL),
    fLocale(itvfmt.fLocale),
    fDatePattern(NULL),
    fTimePattern(NULL),
    fDateTimeFormat(NULL) {
    *this = itvfmt;
}


DateIntervalFormat&
DateIntervalFormat::operator=(const DateIntervalFormat& itvfmt) {
    if ( this != &itvfmt ) {
        delete fDateFormat;
        delete fInfo;
        delete fFromCalendar;
        delete fToCalendar;
        delete fDatePattern;
        delete fTimePattern;
        delete fDateTimeFormat;
        {
            Mutex lock(&gFormatterMutex);
            if ( itvfmt.fDateFormat ) {
                fDateFormat = itvfmt.fDateFormat->clone();
            } else {
                fDateFormat = NULL;
            }
            if ( itvfmt.fFromCalendar ) {
                fFromCalendar = itvfmt.fFromCalendar->clone();
            } else {
                fFromCalendar = NULL;
            }
            if ( itvfmt.fToCalendar ) {
                fToCalendar = itvfmt.fToCalendar->clone();
            } else {
                fToCalendar = NULL;
            }
        }
        if ( itvfmt.fInfo ) {
            fInfo = itvfmt.fInfo->clone();
        } else {
            fInfo = NULL;
        }
        fSkeleton = itvfmt.fSkeleton;
        int8_t i;
        for ( i = 0; i< DateIntervalInfo::kIPI_MAX_INDEX; ++i ) {
            fIntervalPatterns[i] = itvfmt.fIntervalPatterns[i];
        }
        fLocale = itvfmt.fLocale;
        fDatePattern    = (itvfmt.fDatePattern)?    itvfmt.fDatePattern->clone(): NULL;
        fTimePattern    = (itvfmt.fTimePattern)?    itvfmt.fTimePattern->clone(): NULL;
        fDateTimeFormat = (itvfmt.fDateTimeFormat)? itvfmt.fDateTimeFormat->clone(): NULL;
    }
    return *this;
}


DateIntervalFormat::~DateIntervalFormat() {
    delete fInfo;
    delete fDateFormat;
    delete fFromCalendar;
    delete fToCalendar;
    delete fDatePattern;
    delete fTimePattern;
    delete fDateTimeFormat;
}


DateIntervalFormat*
DateIntervalFormat::clone() const {
    return new DateIntervalFormat(*this);
}


UBool
DateIntervalFormat::operator==(const Format& other) const {
    if (typeid(*this) != typeid(other)) {return FALSE;}
    const DateIntervalFormat* fmt = (DateIntervalFormat*)&other;
    if (this == fmt) {return TRUE;}
    if (!Format::operator==(other)) {return FALSE;}
    if ((fInfo != fmt->fInfo) && (fInfo == NULL || fmt->fInfo == NULL)) {return FALSE;}
    if (fInfo && fmt->fInfo && (*fInfo != *fmt->fInfo )) {return FALSE;}
    {
        Mutex lock(&gFormatterMutex);
        if (fDateFormat != fmt->fDateFormat && (fDateFormat == NULL || fmt->fDateFormat == NULL)) {return FALSE;}
        if (fDateFormat && fmt->fDateFormat && (*fDateFormat != *fmt->fDateFormat)) {return FALSE;}
    }
    // note: fFromCalendar and fToCalendar hold no persistent state, and therefore do not participate in operator ==.
    //       fDateFormat has the master calendar for the DateIntervalFormat.
    if (fSkeleton != fmt->fSkeleton) {return FALSE;}
    if (fDatePattern != fmt->fDatePattern && (fDatePattern == NULL || fmt->fDatePattern == NULL)) {return FALSE;}
    if (fDatePattern && fmt->fDatePattern && (*fDatePattern != *fmt->fDatePattern)) {return FALSE;}
    if (fTimePattern != fmt->fTimePattern && (fTimePattern == NULL || fmt->fTimePattern == NULL)) {return FALSE;}
    if (fTimePattern && fmt->fTimePattern && (*fTimePattern != *fmt->fTimePattern)) {return FALSE;}
    if (fDateTimeFormat != fmt->fDateTimeFormat && (fDateTimeFormat == NULL || fmt->fDateTimeFormat == NULL)) {return FALSE;}
    if (fDateTimeFormat && fmt->fDateTimeFormat && (*fDateTimeFormat != *fmt->fDateTimeFormat)) {return FALSE;}
    if (fLocale != fmt->fLocale) {return FALSE;}

    for (int32_t i = 0; i< DateIntervalInfo::kIPI_MAX_INDEX; ++i ) {
        if (fIntervalPatterns[i].firstPart != fmt->fIntervalPatterns[i].firstPart) {return FALSE;}
        if (fIntervalPatterns[i].secondPart != fmt->fIntervalPatterns[i].secondPart ) {return FALSE;}
        if (fIntervalPatterns[i].laterDateFirst != fmt->fIntervalPatterns[i].laterDateFirst) {return FALSE;}
    }
    return TRUE;
}


UnicodeString&
DateIntervalFormat::format(const Formattable& obj,
                           UnicodeString& appendTo,
                           FieldPosition& fieldPosition,
                           UErrorCode& status) const {
    if ( U_FAILURE(status) ) {
        return appendTo;
    }

    if ( obj.getType() == Formattable::kObject ) {
        const UObject* formatObj = obj.getObject();
        const DateInterval* interval = dynamic_cast<const DateInterval*>(formatObj);
        if (interval != NULL) {
            return format(interval, appendTo, fieldPosition, status);
        }
    }
    status = U_ILLEGAL_ARGUMENT_ERROR;
    return appendTo;
}


UnicodeString&
DateIntervalFormat::format(const DateInterval* dtInterval,
                           UnicodeString& appendTo,
                           FieldPosition& fieldPosition,
                           UErrorCode& status) const {
    if ( U_FAILURE(status) ) {
        return appendTo;
    }
    if (fDateFormat == NULL || fInfo == NULL) {
        status = U_INVALID_STATE_ERROR;
        return appendTo;
    }

    FieldPositionOnlyHandler handler(fieldPosition);
    handler.setAcceptFirstOnly(TRUE);
    int8_t ignore;

    Mutex lock(&gFormatterMutex);
    return formatIntervalImpl(*dtInterval, appendTo, ignore, handler, status);
}


FormattedDateInterval DateIntervalFormat::formatToValue(
        const DateInterval& dtInterval,
        UErrorCode& status) const {
    LocalPointer<FormattedDateIntervalData> result(new FormattedDateIntervalData(status), status);
    if (U_FAILURE(status)) {
        return FormattedDateInterval(status);
    }
    UnicodeString string;
    int8_t firstIndex;
    auto handler = result->getHandler(status);
    handler.setCategory(UFIELD_CATEGORY_DATE);
    {
        Mutex lock(&gFormatterMutex);
        formatIntervalImpl(dtInterval, string, firstIndex, handler, status);
    }
    handler.getError(status);
    result->appendString(string, status);
    if (U_FAILURE(status)) {
        return FormattedDateInterval(status);
    }

    // Compute the span fields and sort them into place:
    if (firstIndex != -1) {
        result->addOverlapSpans(UFIELD_CATEGORY_DATE_INTERVAL_SPAN, firstIndex, status);
        if (U_FAILURE(status)) {
            return FormattedDateInterval(status);
        }
        result->sort();
    }

    return FormattedDateInterval(result.orphan());
}


UnicodeString&
DateIntervalFormat::format(Calendar& fromCalendar,
                           Calendar& toCalendar,
                           UnicodeString& appendTo,
                           FieldPosition& pos,
                           UErrorCode& status) const {
    FieldPositionOnlyHandler handler(pos);
    handler.setAcceptFirstOnly(TRUE);
    int8_t ignore;

    Mutex lock(&gFormatterMutex);
    return formatImpl(fromCalendar, toCalendar, appendTo, ignore, handler, status);
}


FormattedDateInterval DateIntervalFormat::formatToValue(
        Calendar& fromCalendar,
        Calendar& toCalendar,
        UErrorCode& status) const {
    LocalPointer<FormattedDateIntervalData> result(new FormattedDateIntervalData(status), status);
    if (U_FAILURE(status)) {
        return FormattedDateInterval(status);
    }
    UnicodeString string;
    int8_t firstIndex;
    auto handler = result->getHandler(status);
    handler.setCategory(UFIELD_CATEGORY_DATE);
    {
        Mutex lock(&gFormatterMutex);
        formatImpl(fromCalendar, toCalendar, string, firstIndex, handler, status);
    }
    handler.getError(status);
    result->appendString(string, status);
    if (U_FAILURE(status)) {
        return FormattedDateInterval(status);
    }

    // Compute the span fields and sort them into place:
    if (firstIndex != -1) {
        result->addOverlapSpans(UFIELD_CATEGORY_DATE_INTERVAL_SPAN, firstIndex, status);
        result->sort();
    }

    return FormattedDateInterval(result.orphan());
}


UnicodeString& DateIntervalFormat::formatIntervalImpl(
        const DateInterval& dtInterval,
        UnicodeString& appendTo,
        int8_t& firstIndex,
        FieldPositionHandler& fphandler,
        UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    if (fFromCalendar == nullptr || fToCalendar == nullptr) {
        status = U_INVALID_STATE_ERROR;
        return appendTo;
    }
    fFromCalendar->setTime(dtInterval.getFromDate(), status);
    fToCalendar->setTime(dtInterval.getToDate(), status);
    return formatImpl(*fFromCalendar, *fToCalendar, appendTo, firstIndex, fphandler, status);
}


UnicodeString&
DateIntervalFormat::formatImpl(Calendar& fromCalendar,
                           Calendar& toCalendar,
                           UnicodeString& appendTo,
                           int8_t& firstIndex,
                           FieldPositionHandler& fphandler,
                           UErrorCode& status) const {
    if ( U_FAILURE(status) ) {
        return appendTo;
    }

    // Initialize firstIndex to -1 (single date, no range)
    firstIndex = -1;

    // not support different calendar types and time zones
    //if ( fromCalendar.getType() != toCalendar.getType() ) {
    if ( !fromCalendar.isEquivalentTo(toCalendar) ) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return appendTo;
    }

    // First, find the largest different calendar field.
    UCalendarDateFields field = UCAL_FIELD_COUNT;

    if ( fromCalendar.get(UCAL_ERA,status) != toCalendar.get(UCAL_ERA,status)) {
        field = UCAL_ERA;
    } else if ( fromCalendar.get(UCAL_YEAR, status) !=
                toCalendar.get(UCAL_YEAR, status) ) {
        field = UCAL_YEAR;
    } else if ( fromCalendar.get(UCAL_MONTH, status) !=
                toCalendar.get(UCAL_MONTH, status) ) {
        field = UCAL_MONTH;
    } else if ( fromCalendar.get(UCAL_DATE, status) !=
                toCalendar.get(UCAL_DATE, status) ) {
        field = UCAL_DATE;
    } else if ( fromCalendar.get(UCAL_AM_PM, status) !=
                toCalendar.get(UCAL_AM_PM, status) ) {
        field = UCAL_AM_PM;
    } else if ( fromCalendar.get(UCAL_HOUR, status) !=
                toCalendar.get(UCAL_HOUR, status) ) {
        field = UCAL_HOUR;
    } else if ( fromCalendar.get(UCAL_MINUTE, status) !=
                toCalendar.get(UCAL_MINUTE, status) ) {
        field = UCAL_MINUTE;
    } else if ( fromCalendar.get(UCAL_SECOND, status) !=
                toCalendar.get(UCAL_SECOND, status) ) {
        field = UCAL_SECOND;
    }

    if ( U_FAILURE(status) ) {
        return appendTo;
    }
    if ( field == UCAL_FIELD_COUNT ) {
        /* ignore the millisecond etc. small fields' difference.
         * use single date when all the above are the same.
         */
        return fDateFormat->_format(fromCalendar, appendTo, fphandler, status);
    }
    UBool fromToOnSameDay = (field==UCAL_AM_PM || field==UCAL_HOUR || field==UCAL_MINUTE || field==UCAL_SECOND);

    // following call should not set wrong status,
    // all the pass-in fields are valid till here
    int32_t itvPtnIndex = DateIntervalInfo::calendarFieldToIntervalIndex(field,
                                                                        status);
    const PatternInfo& intervalPattern = fIntervalPatterns[itvPtnIndex];

    if ( intervalPattern.firstPart.isEmpty() &&
         intervalPattern.secondPart.isEmpty() ) {
        if ( fDateFormat->isFieldUnitIgnored(field) ) {
            /* the largest different calendar field is small than
             * the smallest calendar field in pattern,
             * return single date format.
             */
            return fDateFormat->_format(fromCalendar, appendTo, fphandler, status);
        }
        return fallbackFormat(fromCalendar, toCalendar, fromToOnSameDay, appendTo, firstIndex, fphandler, status);
    }
    // If the first part in interval pattern is empty,
    // the 2nd part of it saves the full-pattern used in fall-back.
    // For a 'real' interval pattern, the first part will never be empty.
    if ( intervalPattern.firstPart.isEmpty() ) {
        // fall back
        UnicodeString originalPattern;
        fDateFormat->toPattern(originalPattern);
        fDateFormat->applyPattern(intervalPattern.secondPart);
        appendTo = fallbackFormat(fromCalendar, toCalendar, fromToOnSameDay, appendTo, firstIndex, fphandler, status);
        fDateFormat->applyPattern(originalPattern);
        return appendTo;
    }
    Calendar* firstCal;
    Calendar* secondCal;
    if ( intervalPattern.laterDateFirst ) {
        firstCal = &toCalendar;
        secondCal = &fromCalendar;
        firstIndex = 1;
    } else {
        firstCal = &fromCalendar;
        secondCal = &toCalendar;
        firstIndex = 0;
    }
    // break the interval pattern into 2 parts,
    // first part should not be empty,
    UnicodeString originalPattern;
    fDateFormat->toPattern(originalPattern);
    fDateFormat->applyPattern(intervalPattern.firstPart);
    fDateFormat->_format(*firstCal, appendTo, fphandler, status);

    if ( !intervalPattern.secondPart.isEmpty() ) {
        fDateFormat->applyPattern(intervalPattern.secondPart);
        fDateFormat->_format(*secondCal, appendTo, fphandler, status);
    }
    fDateFormat->applyPattern(originalPattern);
    return appendTo;
}



void
DateIntervalFormat::parseObject(const UnicodeString& /* source */,
                                Formattable& /* result */,
                                ParsePosition& /* parse_pos */) const {
    // parseObject(const UnicodeString&, Formattable&, UErrorCode&) const
    // will set status as U_INVALID_FORMAT_ERROR if
    // parse_pos is still 0
}




const DateIntervalInfo*
DateIntervalFormat::getDateIntervalInfo() const {
    return fInfo;
}


void
DateIntervalFormat::setDateIntervalInfo(const DateIntervalInfo& newItvPattern,
                                        UErrorCode& status) {
    delete fInfo;
    fInfo = new DateIntervalInfo(newItvPattern);

    // Delete patterns that get reset by initializePattern
    delete fDatePattern;
    fDatePattern = NULL;
    delete fTimePattern;
    fTimePattern = NULL;
    delete fDateTimeFormat;
    fDateTimeFormat = NULL;

    if (fDateFormat) {
        initializePattern(status);
    }
}



const DateFormat*
DateIntervalFormat::getDateFormat() const {
    return fDateFormat;
}


void
DateIntervalFormat::adoptTimeZone(TimeZone* zone)
{
    if (fDateFormat != NULL) {
        fDateFormat->adoptTimeZone(zone);
    }
    // The fDateFormat has the master calendar for the DateIntervalFormat and has
    // ownership of any adopted TimeZone; fFromCalendar and fToCalendar are internal
    // work clones of that calendar (and should not also be given ownership of the
    // adopted TimeZone).
    if (fFromCalendar) {
        fFromCalendar->setTimeZone(*zone);
    }
    if (fToCalendar) {
        fToCalendar->setTimeZone(*zone);
    }
}

void
DateIntervalFormat::setTimeZone(const TimeZone& zone)
{
    if (fDateFormat != NULL) {
        fDateFormat->setTimeZone(zone);
    }
    // The fDateFormat has the master calendar for the DateIntervalFormat;
    // fFromCalendar and fToCalendar are internal work clones of that calendar.
    if (fFromCalendar) {
        fFromCalendar->setTimeZone(zone);
    }
    if (fToCalendar) {
        fToCalendar->setTimeZone(zone);
    }
}

const TimeZone&
DateIntervalFormat::getTimeZone() const
{
    if (fDateFormat != NULL) {
        Mutex lock(&gFormatterMutex);
        return fDateFormat->getTimeZone();
    }
    // If fDateFormat is NULL (unexpected), create default timezone.
    return *(TimeZone::createDefault());
}

DateIntervalFormat::DateIntervalFormat(const Locale& locale,
                                       DateIntervalInfo* dtItvInfo,
                                       const UnicodeString* skeleton,
                                       UErrorCode& status)
:   fInfo(NULL),
    fDateFormat(NULL),
    fFromCalendar(NULL),
    fToCalendar(NULL),
    fLocale(locale),
    fDatePattern(NULL),
    fTimePattern(NULL),
    fDateTimeFormat(NULL)
{
    LocalPointer<DateIntervalInfo> info(dtItvInfo, status);
    LocalPointer<SimpleDateFormat> dtfmt(static_cast<SimpleDateFormat *>(
            DateFormat::createInstanceForSkeleton(*skeleton, locale, status)), status);
    if (U_FAILURE(status)) {
        return;
    }

    if ( skeleton ) {
        fSkeleton = *skeleton;
    }
    fInfo = info.orphan();
    fDateFormat = dtfmt.orphan();
    if ( fDateFormat->getCalendar() ) {
        fFromCalendar = fDateFormat->getCalendar()->clone();
        fToCalendar = fDateFormat->getCalendar()->clone();
    }
    initializePattern(status);
}

DateIntervalFormat* U_EXPORT2
DateIntervalFormat::create(const Locale& locale,
                           DateIntervalInfo* dtitvinf,
                           const UnicodeString* skeleton,
                           UErrorCode& status) {
    DateIntervalFormat* f = new DateIntervalFormat(locale, dtitvinf,
                                                   skeleton, status);
    if ( f == NULL ) {
        status = U_MEMORY_ALLOCATION_ERROR;
        delete dtitvinf;
    } else if ( U_FAILURE(status) ) {
        // safe to delete f, although nothing acutally is saved
        delete f;
        f = 0;
    }
    return f;
}



/**
 * Initialize interval patterns locale to this formatter
 *
 * This code is a bit complicated since
 * 1. the interval patterns saved in resource bundle files are interval
 *    patterns based on date or time only.
 *    It does not have interval patterns based on both date and time.
 *    Interval patterns on both date and time are algorithm generated.
 *
 *    For example, it has interval patterns on skeleton "dMy" and "hm",
 *    but it does not have interval patterns on skeleton "dMyhm".
 *
 *    The rule to genearte interval patterns for both date and time skeleton are
 *    1) when the year, month, or day differs, concatenate the two original
 *    expressions with a separator between,
 *    For example, interval pattern from "Jan 10, 2007 10:10 am"
 *    to "Jan 11, 2007 10:10am" is
 *    "Jan 10, 2007 10:10 am - Jan 11, 2007 10:10am"
 *
 *    2) otherwise, present the date followed by the range expression
 *    for the time.
 *    For example, interval pattern from "Jan 10, 2007 10:10 am"
 *    to "Jan 10, 2007 11:10am" is
 *    "Jan 10, 2007 10:10 am - 11:10am"
 *
 * 2. even a pattern does not request a certion calendar field,
 *    the interval pattern needs to include such field if such fields are
 *    different between 2 dates.
 *    For example, a pattern/skeleton is "hm", but the interval pattern
 *    includes year, month, and date when year, month, and date differs.
 *
 * @param status          output param set to success/failure code on exit
 * @stable ICU 4.0
 */
void
DateIntervalFormat::initializePattern(UErrorCode& status) {
    if ( U_FAILURE(status) ) {
        return;
    }
    const Locale& locale = fDateFormat->getSmpFmtLocale();
    if ( fSkeleton.isEmpty() ) {
        UnicodeString fullPattern;
        fDateFormat->toPattern(fullPattern);
#ifdef DTITVFMT_DEBUG
    char result[1000];
    char result_1[1000];
    char mesg[2000];
    fSkeleton.extract(0,  fSkeleton.length(), result, "UTF-8");
    sprintf(mesg, "in getBestSkeleton: fSkeleton: %s; \n", result);
    PRINTMESG(mesg)
#endif
        // fSkeleton is already set by createDateIntervalInstance()
        // or by createInstance(UnicodeString skeleton, .... )
        fSkeleton = DateTimePatternGenerator::staticGetSkeleton(
                fullPattern, status);
        if ( U_FAILURE(status) ) {
            return;
        }
    }

    // initialize the fIntervalPattern ordering
    int8_t i;
    for ( i = 0; i < DateIntervalInfo::kIPI_MAX_INDEX; ++i ) {
        fIntervalPatterns[i].laterDateFirst = fInfo->getDefaultOrder();
    }

    /* Check whether the skeleton is a combination of date and time.
     * For the complication reason 1 explained above.
     */
    UnicodeString dateSkeleton;
    UnicodeString timeSkeleton;
    UnicodeString normalizedTimeSkeleton;
    UnicodeString normalizedDateSkeleton;


    /* the difference between time skeleton and normalizedTimeSkeleton are:
     * 1. (Formerly, normalized time skeleton folded 'H' to 'h'; no longer true)
     * 2. 'a' is omitted in normalized time skeleton.
     * 3. there is only one appearance for 'h' or 'H', 'm','v', 'z' in normalized
     *    time skeleton
     *
     * The difference between date skeleton and normalizedDateSkeleton are:
     * 1. both 'y' and 'd' appear only once in normalizeDateSkeleton
     * 2. 'E' and 'EE' are normalized into 'EEE'
     * 3. 'MM' is normalized into 'M'
     */
    getDateTimeSkeleton(fSkeleton, dateSkeleton, normalizedDateSkeleton,
                        timeSkeleton, normalizedTimeSkeleton);

#ifdef DTITVFMT_DEBUG
    char result[1000];
    char result_1[1000];
    char mesg[2000];
    fSkeleton.extract(0,  fSkeleton.length(), result, "UTF-8");
    sprintf(mesg, "in getBestSkeleton: fSkeleton: %s; \n", result);
    PRINTMESG(mesg)
#endif

    // move this up here since we need it for fallbacks
    if ( timeSkeleton.length() > 0 && dateSkeleton.length() > 0 ) {
        // Need the Date/Time pattern for concatenation of the date
        // with the time interval.
        // The date/time pattern ( such as {0} {1} ) is saved in
        // calendar, that is why need to get the CalendarData here.
        LocalUResourceBundlePointer dateTimePatternsRes(ures_open(NULL, locale.getBaseName(), &status));
        ures_getByKey(dateTimePatternsRes.getAlias(), dtitvfmt_gCalendarTag,
                      dateTimePatternsRes.getAlias(), &status);
        ures_getByKeyWithFallback(dateTimePatternsRes.getAlias(), dtitvfmt_gGregorianTag,
                                  dateTimePatternsRes.getAlias(), &status);
        ures_getByKeyWithFallback(dateTimePatternsRes.getAlias(), dtitvfmt_gDateTimePatternsTag,
                                  dateTimePatternsRes.getAlias(), &status);

        int32_t dateTimeFormatLength;
        const UChar* dateTimeFormat = ures_getStringByIndex(
                                            dateTimePatternsRes.getAlias(),
                                            (int32_t)DateFormat::kDateTime,
                                            &dateTimeFormatLength, &status);
        if ( U_SUCCESS(status) && dateTimeFormatLength >= 3 ) {
            fDateTimeFormat = new UnicodeString(dateTimeFormat, dateTimeFormatLength);
        }
    }

    UBool found = setSeparateDateTimePtn(normalizedDateSkeleton,
                                         normalizedTimeSkeleton);

    // for skeletons with seconds, found is false and we enter this block
    if ( found == false ) {
        // use fallback
        // TODO: if user asks "m"(minute), but "d"(day) differ
        if ( timeSkeleton.length() != 0 ) {
            if ( dateSkeleton.length() == 0 ) {
                // prefix with yMd
                timeSkeleton.insert(0, dtitvfmt_gDateFormatSkeleton[DateFormat::kShort], -1);
                UnicodeString pattern = DateFormat::getBestPattern(
                        locale, timeSkeleton, status);
                if ( U_FAILURE(status) ) {
                    return;
                }
                // for fall back interval patterns,
                // the first part of the pattern is empty,
                // the second part of the pattern is the full-pattern
                // should be used in fall-back.
                setPatternInfo(UCAL_DATE, NULL, &pattern, fInfo->getDefaultOrder());
                setPatternInfo(UCAL_MONTH, NULL, &pattern, fInfo->getDefaultOrder());
                setPatternInfo(UCAL_YEAR, NULL, &pattern, fInfo->getDefaultOrder());
            } else {
                // TODO: fall back
            }
        } else {
            // TODO: fall back
        }
        return;
    } // end of skeleton not found
    // interval patterns for skeleton are found in resource
    if ( timeSkeleton.length() == 0 ) {
        // done
    } else if ( dateSkeleton.length() == 0 ) {
        // prefix with yMd
        timeSkeleton.insert(0, dtitvfmt_gDateFormatSkeleton[DateFormat::kShort], -1);
        UnicodeString pattern = DateFormat::getBestPattern(
                locale, timeSkeleton, status);
        if ( U_FAILURE(status) ) {
            return;
        }
        // for fall back interval patterns,
        // the first part of the pattern is empty,
        // the second part of the pattern is the full-pattern
        // should be used in fall-back.
        setPatternInfo(UCAL_DATE, NULL, &pattern, fInfo->getDefaultOrder());
        setPatternInfo(UCAL_MONTH, NULL, &pattern, fInfo->getDefaultOrder());
        setPatternInfo(UCAL_YEAR, NULL, &pattern, fInfo->getDefaultOrder());
    } else {
        /* if both present,
         * 1) when the year, month, or day differs,
         * concatenate the two original expressions with a separator between,
         * 2) otherwise, present the date followed by the
         * range expression for the time.
         */
        /*
         * 1) when the year, month, or day differs,
         * concatenate the two original expressions with a separator between,
         */
        // if field exists, use fall back
        UnicodeString skeleton = fSkeleton;
        if ( !fieldExistsInSkeleton(UCAL_DATE, dateSkeleton) ) {
            // prefix skeleton with 'd'
            skeleton.insert(0, dtitv_impl_LOW_D);
            setFallbackPattern(UCAL_DATE, skeleton, status);
        }
        if ( !fieldExistsInSkeleton(UCAL_MONTH, dateSkeleton) ) {
            // then prefix skeleton with 'M'
            skeleton.insert(0, dtitv_impl_CAP_M);
            setFallbackPattern(UCAL_MONTH, skeleton, status);
        }
        if ( !fieldExistsInSkeleton(UCAL_YEAR, dateSkeleton) ) {
            // then prefix skeleton with 'y'
            skeleton.insert(0, dtitv_impl_LOW_Y);
            setFallbackPattern(UCAL_YEAR, skeleton, status);
        }

        /*
         * 2) otherwise, present the date followed by the
         * range expression for the time.
         */

        if ( fDateTimeFormat == NULL ) {
            // earlier failure getting dateTimeFormat
            return;
        }

        UnicodeString datePattern = DateFormat::getBestPattern(
                locale, dateSkeleton, status);

        concatSingleDate2TimeInterval(*fDateTimeFormat, datePattern, UCAL_AM_PM, status);
        concatSingleDate2TimeInterval(*fDateTimeFormat, datePattern, UCAL_HOUR, status);
        concatSingleDate2TimeInterval(*fDateTimeFormat, datePattern, UCAL_MINUTE, status);
    }
}



void  U_EXPORT2
DateIntervalFormat::getDateTimeSkeleton(const UnicodeString& skeleton,
                                        UnicodeString& dateSkeleton,
                                        UnicodeString& normalizedDateSkeleton,
                                        UnicodeString& timeSkeleton,
                                        UnicodeString& normalizedTimeSkeleton) {
    // dateSkeleton follows the sequence of y*M*E*d*
    // timeSkeleton follows the sequence of hm*[v|z]?
    int32_t ECount = 0;
    int32_t dCount = 0;
    int32_t MCount = 0;
    int32_t yCount = 0;
    int32_t hCount = 0;
    int32_t HCount = 0;
    int32_t mCount = 0;
    int32_t vCount = 0;
    int32_t zCount = 0;
    int32_t i;

    for (i = 0; i < skeleton.length(); ++i) {
        UChar ch = skeleton[i];
        switch ( ch ) {
          case dtitv_impl_CAP_E:
            dateSkeleton.append(ch);
            ++ECount;
            break;
          case dtitv_impl_LOW_D:
            dateSkeleton.append(ch);
            ++dCount;
            break;
          case dtitv_impl_CAP_M:
            dateSkeleton.append(ch);
            ++MCount;
            break;
          case dtitv_impl_LOW_Y:
            dateSkeleton.append(ch);
            ++yCount;
            break;
          case dtitv_impl_CAP_G:
          case dtitv_impl_CAP_Y:
          case dtitv_impl_LOW_U:
          case dtitv_impl_CAP_Q:
          case dtitv_impl_LOW_Q:
          case dtitv_impl_CAP_L:
          case dtitv_impl_LOW_L:
          case dtitv_impl_CAP_W:
          case dtitv_impl_LOW_W:
          case dtitv_impl_CAP_D:
          case dtitv_impl_CAP_F:
          case dtitv_impl_LOW_G:
          case dtitv_impl_LOW_E:
          case dtitv_impl_LOW_C:
          case dtitv_impl_CAP_U:
          case dtitv_impl_LOW_R:
            normalizedDateSkeleton.append(ch);
            dateSkeleton.append(ch);
            break;
          case dtitv_impl_LOW_A:
            // 'a' is implicitly handled
            timeSkeleton.append(ch);
            break;
          case dtitv_impl_LOW_H:
            timeSkeleton.append(ch);
            ++hCount;
            break;
          case dtitv_impl_CAP_H:
            timeSkeleton.append(ch);
            ++HCount;
            break;
          case dtitv_impl_LOW_M:
            timeSkeleton.append(ch);
            ++mCount;
            break;
          case dtitv_impl_LOW_Z:
            ++zCount;
            timeSkeleton.append(ch);
            break;
          case dtitv_impl_LOW_V:
            ++vCount;
            timeSkeleton.append(ch);
            break;
          case dtitv_impl_CAP_V:
          case dtitv_impl_CAP_Z:
          case dtitv_impl_LOW_K:
          case dtitv_impl_CAP_K:
          case dtitv_impl_LOW_J:
          case dtitv_impl_LOW_S:
          case dtitv_impl_CAP_S:
          case dtitv_impl_CAP_A:
            timeSkeleton.append(ch);
            normalizedTimeSkeleton.append(ch);
            break;
        }
    }

    /* generate normalized form for date*/
    if ( yCount != 0 ) {
        for (i = 0; i < yCount; ++i) {
            normalizedDateSkeleton.append(dtitv_impl_LOW_Y);
        }
    }
    if ( MCount != 0 ) {
        if ( MCount < 3 ) {
            normalizedDateSkeleton.append(dtitv_impl_CAP_M);
        } else {
            for ( int32_t j = 0; j < MCount && j < dtitv_impl_MAX_M_COUNT; ++j) {
                 normalizedDateSkeleton.append(dtitv_impl_CAP_M);
            }
        }
    }
    if ( ECount != 0 ) {
        if ( ECount <= 3 ) {
            normalizedDateSkeleton.append(dtitv_impl_CAP_E);
        } else {
            for ( int32_t j = 0; j < ECount && j < dtitv_impl_MAX_E_COUNT; ++j ) {
                 normalizedDateSkeleton.append(dtitv_impl_CAP_E);
            }
        }
    }
    if ( dCount != 0 ) {
        normalizedDateSkeleton.append(dtitv_impl_LOW_D);
    }

    /* generate normalized form for time */
    if ( HCount != 0 ) {
        normalizedTimeSkeleton.append(dtitv_impl_CAP_H);
    }
    else if ( hCount != 0 ) {
        normalizedTimeSkeleton.append(dtitv_impl_LOW_H);
    }
    if ( mCount != 0 ) {
        normalizedTimeSkeleton.append(dtitv_impl_LOW_M);
    }
    if ( zCount != 0 ) {
        normalizedTimeSkeleton.append(dtitv_impl_LOW_Z);
    }
    if ( vCount != 0 ) {
        normalizedTimeSkeleton.append(dtitv_impl_LOW_V);
    }
}


/**
 * Generate date or time interval pattern from resource,
 * and set them into the interval pattern locale to this formatter.
 *
 * It needs to handle the following:
 * 1. need to adjust field width.
 *    For example, the interval patterns saved in DateIntervalInfo
 *    includes "dMMMy", but not "dMMMMy".
 *    Need to get interval patterns for dMMMMy from dMMMy.
 *    Another example, the interval patterns saved in DateIntervalInfo
 *    includes "hmv", but not "hmz".
 *    Need to get interval patterns for "hmz' from 'hmv'
 *
 * 2. there might be no pattern for 'y' differ for skeleton "Md",
 *    in order to get interval patterns for 'y' differ,
 *    need to look for it from skeleton 'yMd'
 *
 * @param dateSkeleton   normalized date skeleton
 * @param timeSkeleton   normalized time skeleton
 * @return               whether the resource is found for the skeleton.
 *                       TRUE if interval pattern found for the skeleton,
 *                       FALSE otherwise.
 * @stable ICU 4.0
 */
UBool
DateIntervalFormat::setSeparateDateTimePtn(
                                 const UnicodeString& dateSkeleton,
                                 const UnicodeString& timeSkeleton) {
    const UnicodeString* skeleton;
    // if both date and time skeleton present,
    // the final interval pattern might include time interval patterns
    // ( when, am_pm, hour, minute differ ),
    // but not date interval patterns ( when year, month, day differ ).
    // For year/month/day differ, it falls back to fall-back pattern.
    if ( timeSkeleton.length() != 0  ) {
        skeleton = &timeSkeleton;
    } else {
        skeleton = &dateSkeleton;
    }

    /* interval patterns for skeleton "dMMMy" (but not "dMMMMy")
     * are defined in resource,
     * interval patterns for skeleton "dMMMMy" are calculated by
     * 1. get the best match skeleton for "dMMMMy", which is "dMMMy"
     * 2. get the interval patterns for "dMMMy",
     * 3. extend "MMM" to "MMMM" in above interval patterns for "dMMMMy"
     * getBestSkeleton() is step 1.
     */
    // best skeleton, and the difference information
    int8_t differenceInfo = 0;
    const UnicodeString* bestSkeleton = fInfo->getBestSkeleton(*skeleton,
                                                               differenceInfo);
    /* best skeleton could be NULL.
       For example: in "ca" resource file,
       interval format is defined as following
           intervalFormats{
                fallback{"{0} - {1}"}
            }
       there is no skeletons/interval patterns defined,
       and the best skeleton match could be NULL
     */
    if ( bestSkeleton == NULL ) {
        return false;
    }

    // Set patterns for fallback use, need to do this
    // before returning if differenceInfo == -1
    UErrorCode status;
    if ( dateSkeleton.length() != 0) {
        status = U_ZERO_ERROR;
        fDatePattern = new UnicodeString(DateFormat::getBestPattern(
                fLocale, dateSkeleton, status));
    }
    if ( timeSkeleton.length() != 0) {
        status = U_ZERO_ERROR;
        fTimePattern = new UnicodeString(DateFormat::getBestPattern(
                fLocale, timeSkeleton, status));
    }

    // difference:
    // 0 means the best matched skeleton is the same as input skeleton
    // 1 means the fields are the same, but field width are different
    // 2 means the only difference between fields are v/z,
    // -1 means there are other fields difference
    // (this will happen, for instance, if the supplied skeleton has seconds,
    //  but no skeletons in the intervalFormats data do)
    if ( differenceInfo == -1 ) {
        // skeleton has different fields, not only  v/z difference
        return false;
    }

    if ( timeSkeleton.length() == 0 ) {
        UnicodeString extendedSkeleton;
        UnicodeString extendedBestSkeleton;
        // only has date skeleton
        setIntervalPattern(UCAL_DATE, skeleton, bestSkeleton, differenceInfo,
                           &extendedSkeleton, &extendedBestSkeleton);

        UBool extended = setIntervalPattern(UCAL_MONTH, skeleton, bestSkeleton,
                                     differenceInfo,
                                     &extendedSkeleton, &extendedBestSkeleton);

        if ( extended ) {
            bestSkeleton = &extendedBestSkeleton;
            skeleton = &extendedSkeleton;
        }
        setIntervalPattern(UCAL_YEAR, skeleton, bestSkeleton, differenceInfo,
                           &extendedSkeleton, &extendedBestSkeleton);
        setIntervalPattern(UCAL_ERA, skeleton, bestSkeleton, differenceInfo,
                           &extendedSkeleton, &extendedBestSkeleton);
     } else {
        setIntervalPattern(UCAL_MINUTE, skeleton, bestSkeleton, differenceInfo);
        setIntervalPattern(UCAL_HOUR, skeleton, bestSkeleton, differenceInfo);
        setIntervalPattern(UCAL_AM_PM, skeleton, bestSkeleton, differenceInfo);
    }
    return true;
}



void
DateIntervalFormat::setFallbackPattern(UCalendarDateFields field,
                                       const UnicodeString& skeleton,
                                       UErrorCode& status) {
    if ( U_FAILURE(status) ) {
        return;
    }
    UnicodeString pattern = DateFormat::getBestPattern(
            fLocale, skeleton, status);
    if ( U_FAILURE(status) ) {
        return;
    }
    setPatternInfo(field, NULL, &pattern, fInfo->getDefaultOrder());
}




void
DateIntervalFormat::setPatternInfo(UCalendarDateFields field,
                                   const UnicodeString* firstPart,
                                   const UnicodeString* secondPart,
                                   UBool laterDateFirst) {
    // for fall back interval patterns,
    // the first part of the pattern is empty,
    // the second part of the pattern is the full-pattern
    // should be used in fall-back.
    UErrorCode status = U_ZERO_ERROR;
    // following should not set any wrong status.
    int32_t itvPtnIndex = DateIntervalInfo::calendarFieldToIntervalIndex(field,
                                                                        status);
    if ( U_FAILURE(status) ) {
        return;
    }
    PatternInfo& ptn = fIntervalPatterns[itvPtnIndex];
    if ( firstPart ) {
        ptn.firstPart = *firstPart;
    }
    if ( secondPart ) {
        ptn.secondPart = *secondPart;
    }
    ptn.laterDateFirst = laterDateFirst;
}

void
DateIntervalFormat::setIntervalPattern(UCalendarDateFields field,
                                       const UnicodeString& intervalPattern) {
    UBool order = fInfo->getDefaultOrder();
    setIntervalPattern(field, intervalPattern, order);
}


void
DateIntervalFormat::setIntervalPattern(UCalendarDateFields field,
                                       const UnicodeString& intervalPattern,
                                       UBool laterDateFirst) {
    const UnicodeString* pattern = &intervalPattern;
    UBool order = laterDateFirst;
    // check for "latestFirst:" or "earliestFirst:" prefix
    int8_t prefixLength = UPRV_LENGTHOF(dtitvfmt_gLaterFirstPrefix);
    int8_t earliestFirstLength = UPRV_LENGTHOF(dtitvfmt_gEarlierFirstPrefix);
    UnicodeString realPattern;
    if ( intervalPattern.startsWith(dtitvfmt_gLaterFirstPrefix, prefixLength) ) {
        order = true;
        intervalPattern.extract(prefixLength,
                                intervalPattern.length() - prefixLength,
                                realPattern);
        pattern = &realPattern;
    } else if ( intervalPattern.startsWith(dtitvfmt_gEarlierFirstPrefix,
                                           earliestFirstLength) ) {
        order = false;
        intervalPattern.extract(earliestFirstLength,
                                intervalPattern.length() - earliestFirstLength,
                                realPattern);
        pattern = &realPattern;
    }

    int32_t splitPoint = splitPatternInto2Part(*pattern);

    UnicodeString firstPart;
    UnicodeString secondPart;
    pattern->extract(0, splitPoint, firstPart);
    if ( splitPoint < pattern->length() ) {
        pattern->extract(splitPoint, pattern->length()-splitPoint, secondPart);
    }
    setPatternInfo(field, &firstPart, &secondPart, order);
}




/**
 * Generate interval pattern from existing resource
 *
 * It not only save the interval patterns,
 * but also return the extended skeleton and its best match skeleton.
 *
 * @param field           largest different calendar field
 * @param skeleton        skeleton
 * @param bestSkeleton    the best match skeleton which has interval pattern
 *                        defined in resource
 * @param differenceInfo  the difference between skeleton and best skeleton
 *         0 means the best matched skeleton is the same as input skeleton
 *         1 means the fields are the same, but field width are different
 *         2 means the only difference between fields are v/z,
 *        -1 means there are other fields difference
 *
 * @param extendedSkeleton      extended skeleton
 * @param extendedBestSkeleton  extended best match skeleton
 * @return                      whether the interval pattern is found
 *                              through extending skeleton or not.
 *                              TRUE if interval pattern is found by
 *                              extending skeleton, FALSE otherwise.
 * @stable ICU 4.0
 */
UBool
DateIntervalFormat::setIntervalPattern(UCalendarDateFields field,
                                       const UnicodeString* skeleton,
                                       const UnicodeString* bestSkeleton,
                                       int8_t differenceInfo,
                                       UnicodeString* extendedSkeleton,
                                       UnicodeString* extendedBestSkeleton) {
    UErrorCode status = U_ZERO_ERROR;
    // following getIntervalPattern() should not generate error status
    UnicodeString pattern;
    fInfo->getIntervalPattern(*bestSkeleton, field, pattern, status);
    if ( pattern.isEmpty() ) {
        // single date
        if ( SimpleDateFormat::isFieldUnitIgnored(*bestSkeleton, field) ) {
            // do nothing, format will handle it
            return false;
        }

        // for 24 hour system, interval patterns in resource file
        // might not include pattern when am_pm differ,
        // which should be the same as hour differ.
        // add it here for simplicity
        if ( field == UCAL_AM_PM ) {
            fInfo->getIntervalPattern(*bestSkeleton, UCAL_HOUR, pattern,status);
            if ( !pattern.isEmpty() ) {
                setIntervalPattern(field, pattern);
            }
            return false;
        }
        // else, looking for pattern when 'y' differ for 'dMMMM' skeleton,
        // first, get best match pattern "MMMd",
        // since there is no pattern for 'y' differs for skeleton 'MMMd',
        // need to look for it from skeleton 'yMMMd',
        // if found, adjust field width in interval pattern from
        // "MMM" to "MMMM".
        UChar fieldLetter = fgCalendarFieldToPatternLetter[field];
        if ( extendedSkeleton ) {
            *extendedSkeleton = *skeleton;
            *extendedBestSkeleton = *bestSkeleton;
            extendedSkeleton->insert(0, fieldLetter);
            extendedBestSkeleton->insert(0, fieldLetter);
            // for example, looking for patterns when 'y' differ for
            // skeleton "MMMM".
            fInfo->getIntervalPattern(*extendedBestSkeleton,field,pattern,status);
            if ( pattern.isEmpty() && differenceInfo == 0 ) {
                // if there is no skeleton "yMMMM" defined,
                // look for the best match skeleton, for example: "yMMM"
                const UnicodeString* tmpBest = fInfo->getBestSkeleton(
                                        *extendedBestSkeleton, differenceInfo);
                if ( tmpBest != 0 && differenceInfo != -1 ) {
                    fInfo->getIntervalPattern(*tmpBest, field, pattern, status);
                    bestSkeleton = tmpBest;
                }
            }
        }
    }
    if ( !pattern.isEmpty() ) {
        if ( differenceInfo != 0 ) {
            UnicodeString adjustIntervalPattern;
            adjustFieldWidth(*skeleton, *bestSkeleton, pattern, differenceInfo,
                              adjustIntervalPattern);
            setIntervalPattern(field, adjustIntervalPattern);
        } else {
            setIntervalPattern(field, pattern);
        }
        if ( extendedSkeleton && !extendedSkeleton->isEmpty() ) {
            return TRUE;
        }
    }
    return FALSE;
}



int32_t  U_EXPORT2
DateIntervalFormat::splitPatternInto2Part(const UnicodeString& intervalPattern) {
    UBool inQuote = false;
    UChar prevCh = 0;
    int32_t count = 0;

    /* repeatedPattern used to record whether a pattern has already seen.
       It is a pattern applies to first calendar if it is first time seen,
       otherwise, it is a pattern applies to the second calendar
     */
    UBool patternRepeated[] =
    {
    //       A   B   C   D   E   F   G   H   I   J   K   L   M   N   O
             0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    //   P   Q   R   S   T   U   V   W   X   Y   Z
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 0, 0,  0, 0, 0,
    //       a   b   c   d   e   f   g   h   i   j   k   l   m   n   o
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    //   p   q   r   s   t   u   v   w   x   y   z
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
    };

    int8_t PATTERN_CHAR_BASE = 0x41;

    /* loop through the pattern string character by character looking for
     * the first repeated pattern letter, which breaks the interval pattern
     * into 2 parts.
     */
    int32_t i;
    UBool foundRepetition = false;
    for (i = 0; i < intervalPattern.length(); ++i) {
        UChar ch = intervalPattern.charAt(i);

        if (ch != prevCh && count > 0) {
            // check the repeativeness of pattern letter
            UBool repeated = patternRepeated[(int)(prevCh - PATTERN_CHAR_BASE)];
            if ( repeated == FALSE ) {
                patternRepeated[prevCh - PATTERN_CHAR_BASE] = TRUE;
            } else {
                foundRepetition = true;
                break;
            }
            count = 0;
        }
        if (ch == 0x0027 /*'*/) {
            // Consecutive single quotes are a single quote literal,
            // either outside of quotes or between quotes
            if ((i+1) < intervalPattern.length() &&
                intervalPattern.charAt(i+1) == 0x0027 /*'*/) {
                ++i;
            } else {
                inQuote = ! inQuote;
            }
        }
        else if (!inQuote && ((ch >= 0x0061 /*'a'*/ && ch <= 0x007A /*'z'*/)
                    || (ch >= 0x0041 /*'A'*/ && ch <= 0x005A /*'Z'*/))) {
            // ch is a date-time pattern character
            prevCh = ch;
            ++count;
        }
    }
    // check last pattern char, distinguish
    // "dd MM" ( no repetition ),
    // "d-d"(last char repeated ), and
    // "d-d MM" ( repetition found )
    if ( count > 0 && foundRepetition == FALSE ) {
        if ( patternRepeated[(int)(prevCh - PATTERN_CHAR_BASE)] == FALSE ) {
            count = 0;
        }
    }
    return (i - count);
}

void DateIntervalFormat::fallbackFormatRange(
        Calendar& fromCalendar,
        Calendar& toCalendar,
        UnicodeString& appendTo,
        int8_t& firstIndex,
        FieldPositionHandler& fphandler,
        UErrorCode& status) const {
    UnicodeString fallbackPattern;
    fInfo->getFallbackIntervalPattern(fallbackPattern);
    SimpleFormatter sf(fallbackPattern, 2, 2, status);
    if (U_FAILURE(status)) {
        return;
    }
    int32_t offsets[2];
    UnicodeString patternBody = sf.getTextWithNoArguments(offsets, 2);

    // TODO(ICU-20406): Use SimpleFormatter Iterator interface when available.
    if (offsets[0] < offsets[1]) {
        firstIndex = 0;
        appendTo.append(patternBody.tempSubStringBetween(0, offsets[0]));
        fDateFormat->_format(fromCalendar, appendTo, fphandler, status);
        appendTo.append(patternBody.tempSubStringBetween(offsets[0], offsets[1]));
        fDateFormat->_format(toCalendar, appendTo, fphandler, status);
        appendTo.append(patternBody.tempSubStringBetween(offsets[1]));
    } else {
        firstIndex = 1;
        appendTo.append(patternBody.tempSubStringBetween(0, offsets[1]));
        fDateFormat->_format(toCalendar, appendTo, fphandler, status);
        appendTo.append(patternBody.tempSubStringBetween(offsets[1], offsets[0]));
        fDateFormat->_format(fromCalendar, appendTo, fphandler, status);
        appendTo.append(patternBody.tempSubStringBetween(offsets[0]));
    }
}

UnicodeString&
DateIntervalFormat::fallbackFormat(Calendar& fromCalendar,
                                   Calendar& toCalendar,
                                   UBool fromToOnSameDay, // new
                                   UnicodeString& appendTo,
                                   int8_t& firstIndex,
                                   FieldPositionHandler& fphandler,
                                   UErrorCode& status) const {
    if ( U_FAILURE(status) ) {
        return appendTo;
    }

    UBool formatDatePlusTimeRange = (fromToOnSameDay && fDatePattern && fTimePattern);
    if (formatDatePlusTimeRange) {
        SimpleFormatter sf(*fDateTimeFormat, 2, 2, status);
        if (U_FAILURE(status)) {
            return appendTo;
        }
        int32_t offsets[2];
        UnicodeString patternBody = sf.getTextWithNoArguments(offsets, 2);

        UnicodeString fullPattern; // for saving the pattern in fDateFormat
        fDateFormat->toPattern(fullPattern); // save current pattern, restore later

        // {0} is time range
        // {1} is single date portion
        // TODO(ICU-20406): Use SimpleFormatter Iterator interface when available.
        if (offsets[0] < offsets[1]) {
            appendTo.append(patternBody.tempSubStringBetween(0, offsets[0]));
            fDateFormat->applyPattern(*fTimePattern);
            fallbackFormatRange(fromCalendar, toCalendar, appendTo, firstIndex, fphandler, status);
            appendTo.append(patternBody.tempSubStringBetween(offsets[0], offsets[1]));
            fDateFormat->applyPattern(*fDatePattern);
            fDateFormat->_format(fromCalendar, appendTo, fphandler, status);
            appendTo.append(patternBody.tempSubStringBetween(offsets[1]));
        } else {
            appendTo.append(patternBody.tempSubStringBetween(0, offsets[1]));
            fDateFormat->applyPattern(*fDatePattern);
            fDateFormat->_format(fromCalendar, appendTo, fphandler, status);
            appendTo.append(patternBody.tempSubStringBetween(offsets[1], offsets[0]));
            fDateFormat->applyPattern(*fTimePattern);
            fallbackFormatRange(fromCalendar, toCalendar, appendTo, firstIndex, fphandler, status);
            appendTo.append(patternBody.tempSubStringBetween(offsets[0]));
        }

        // restore full pattern
        fDateFormat->applyPattern(fullPattern);
    } else {
        fallbackFormatRange(fromCalendar, toCalendar, appendTo, firstIndex, fphandler, status);
    }
    return appendTo;
}




UBool  U_EXPORT2
DateIntervalFormat::fieldExistsInSkeleton(UCalendarDateFields field,
                                          const UnicodeString& skeleton)
{
    const UChar fieldChar = fgCalendarFieldToPatternLetter[field];
    return ( (skeleton.indexOf(fieldChar) == -1)?FALSE:TRUE ) ;
}



void  U_EXPORT2
DateIntervalFormat::adjustFieldWidth(const UnicodeString& inputSkeleton,
                 const UnicodeString& bestMatchSkeleton,
                 const UnicodeString& bestIntervalPattern,
                 int8_t differenceInfo,
                 UnicodeString& adjustedPtn) {
    adjustedPtn = bestIntervalPattern;
    int32_t inputSkeletonFieldWidth[] =
    {
    //       A   B   C   D   E   F   G   H   I   J   K   L   M   N   O
             0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    //   P   Q   R   S   T   U   V   W   X   Y   Z
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 0, 0,  0, 0, 0,
    //       a   b   c   d   e   f   g   h   i   j   k   l   m   n   o
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    //   p   q   r   s   t   u   v   w   x   y   z
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
    };

    int32_t bestMatchSkeletonFieldWidth[] =
    {
    //       A   B   C   D   E   F   G   H   I   J   K   L   M   N   O
             0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    //   P   Q   R   S   T   U   V   W   X   Y   Z
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 0, 0,  0, 0, 0,
    //       a   b   c   d   e   f   g   h   i   j   k   l   m   n   o
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    //   p   q   r   s   t   u   v   w   x   y   z
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
    };

    DateIntervalInfo::parseSkeleton(inputSkeleton, inputSkeletonFieldWidth);
    DateIntervalInfo::parseSkeleton(bestMatchSkeleton, bestMatchSkeletonFieldWidth);
    if ( differenceInfo == 2 ) {
        adjustedPtn.findAndReplace(UnicodeString((UChar)0x76 /* v */),
                                   UnicodeString((UChar)0x7a /* z */));
    }

    UBool inQuote = false;
    UChar prevCh = 0;
    int32_t count = 0;

    const int8_t PATTERN_CHAR_BASE = 0x41;

    // loop through the pattern string character by character
    int32_t adjustedPtnLength = adjustedPtn.length();
    int32_t i;
    for (i = 0; i < adjustedPtnLength; ++i) {
        UChar ch = adjustedPtn.charAt(i);
        if (ch != prevCh && count > 0) {
            // check the repeativeness of pattern letter
            UChar skeletonChar = prevCh;
            if ( skeletonChar ==  dtitv_impl_CAP_L ) {
                // there is no "L" (always be "M") in skeleton,
                // but there is "L" in pattern.
                // for skeleton "M+", the pattern might be "...L..."
                skeletonChar = dtitv_impl_CAP_M;
            }
            int32_t fieldCount = bestMatchSkeletonFieldWidth[(int)(skeletonChar - PATTERN_CHAR_BASE)];
            int32_t inputFieldCount = inputSkeletonFieldWidth[(int)(skeletonChar - PATTERN_CHAR_BASE)];
            if ( fieldCount == count && inputFieldCount > fieldCount ) {
                count = inputFieldCount - fieldCount;
                int32_t j;
                for ( j = 0; j < count; ++j ) {
                    adjustedPtn.insert(i, prevCh);
                }
                i += count;
                adjustedPtnLength += count;
            }
            count = 0;
        }
        if (ch == 0x0027 /*'*/) {
            // Consecutive single quotes are a single quote literal,
            // either outside of quotes or between quotes
            if ((i+1) < adjustedPtn.length() && adjustedPtn.charAt(i+1) == 0x0027 /* ' */) {
                ++i;
            } else {
                inQuote = ! inQuote;
            }
        }
        else if ( ! inQuote && ((ch >= 0x0061 /*'a'*/ && ch <= 0x007A /*'z'*/)
                    || (ch >= 0x0041 /*'A'*/ && ch <= 0x005A /*'Z'*/))) {
            // ch is a date-time pattern character
            prevCh = ch;
            ++count;
        }
    }
    if ( count > 0 ) {
        // last item
        // check the repeativeness of pattern letter
        UChar skeletonChar = prevCh;
        if ( skeletonChar == dtitv_impl_CAP_L ) {
            // there is no "L" (always be "M") in skeleton,
            // but there is "L" in pattern.
            // for skeleton "M+", the pattern might be "...L..."
            skeletonChar = dtitv_impl_CAP_M;
        }
        int32_t fieldCount = bestMatchSkeletonFieldWidth[(int)(skeletonChar - PATTERN_CHAR_BASE)];
        int32_t inputFieldCount = inputSkeletonFieldWidth[(int)(skeletonChar - PATTERN_CHAR_BASE)];
        if ( fieldCount == count && inputFieldCount > fieldCount ) {
            count = inputFieldCount - fieldCount;
            int32_t j;
            for ( j = 0; j < count; ++j ) {
                adjustedPtn.append(prevCh);
            }
        }
    }
}



void
DateIntervalFormat::concatSingleDate2TimeInterval(UnicodeString& format,
                                              const UnicodeString& datePattern,
                                              UCalendarDateFields field,
                                              UErrorCode& status) {
    // following should not set wrong status
    int32_t itvPtnIndex = DateIntervalInfo::calendarFieldToIntervalIndex(field,
                                                                        status);
    if ( U_FAILURE(status) ) {
        return;
    }
    PatternInfo&  timeItvPtnInfo = fIntervalPatterns[itvPtnIndex];
    if ( !timeItvPtnInfo.firstPart.isEmpty() ) {
        UnicodeString timeIntervalPattern(timeItvPtnInfo.firstPart);
        timeIntervalPattern.append(timeItvPtnInfo.secondPart);
        UnicodeString combinedPattern;
        SimpleFormatter(format, 2, 2, status).
                format(timeIntervalPattern, datePattern, combinedPattern, status);
        if ( U_FAILURE(status) ) {
            return;
        }
        setIntervalPattern(field, combinedPattern, timeItvPtnInfo.laterDateFirst);
    }
    // else: fall back
    // it should not happen if the interval format defined is valid
}



const UChar
DateIntervalFormat::fgCalendarFieldToPatternLetter[] =
{
    /*GyM*/ dtitv_impl_CAP_G, dtitv_impl_LOW_Y, dtitv_impl_CAP_M,
    /*wWd*/ dtitv_impl_LOW_W, dtitv_impl_CAP_W, dtitv_impl_LOW_D,
    /*DEF*/ dtitv_impl_CAP_D, dtitv_impl_CAP_E, dtitv_impl_CAP_F,
    /*ahH*/ dtitv_impl_LOW_A, dtitv_impl_LOW_H, dtitv_impl_CAP_H,
    /*msS*/ dtitv_impl_LOW_M, dtitv_impl_LOW_S, dtitv_impl_CAP_S, // MINUTE, SECOND, MILLISECOND
    /*z.Y*/ dtitv_impl_LOW_Z, dtitv_impl_SPACE, dtitv_impl_CAP_Y, // ZONE_OFFSET, DST_OFFSET, YEAR_WOY,
    /*eug*/ dtitv_impl_LOW_E, dtitv_impl_LOW_U, dtitv_impl_LOW_G, // DOW_LOCAL, EXTENDED_YEAR, JULIAN_DAY,
    /*A..*/ dtitv_impl_CAP_A, dtitv_impl_SPACE, dtitv_impl_SPACE, // MILLISECONDS_IN_DAY, IS_LEAP_MONTH, FIELD_COUNT
};



U_NAMESPACE_END

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*******************************************************************************
* Copyright (C) 2008-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File DTITVINF.CPP
*
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING

//TODO: define it in compiler time
//#define DTITVINF_DEBUG 1


#ifdef DTITVINF_DEBUG
#include <iostream>
#endif
















U_NAMESPACE_BEGIN


#ifdef DTITVINF_DEBUG
#define PRINTMESG(msg) UPRV_BLOCK_MACRO_BEGIN { \
    std::cout << "(" << __FILE__ << ":" << __LINE__ << ") " << msg << "\n"; \
} UPRV_BLOCK_MACRO_END
#endif

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DateIntervalInfo)

static const char dtitvinf_gCalendarTag[]="calendar";
static const char dtitvinf_gGenericTag[]="generic";
static const char dtitvinf_gGregorianTag[]="gregorian";
static const char dtitvinf_gIntervalDateTimePatternTag[]="intervalFormats";
static const char dtitvinf_gFallbackPatternTag[]="fallback";

// {0}
static const UChar dtitvinf_gFirstPattern[] = {dtitv_impl_LEFT_CURLY_BRACKET, dtitv_impl_DIGIT_ZERO, dtitv_impl_RIGHT_CURLY_BRACKET};
// {1}
static const UChar dtitvinf_gSecondPattern[] = {dtitv_impl_LEFT_CURLY_BRACKET, dtitv_impl_DIGIT_ONE, dtitv_impl_RIGHT_CURLY_BRACKET};

// default fall-back
static const UChar dtitvinf_gDefaultFallbackPattern[] = {dtitv_impl_LEFT_CURLY_BRACKET, dtitv_impl_DIGIT_ZERO, dtitv_impl_RIGHT_CURLY_BRACKET, dtitv_impl_SPACE, dtitv_impl_EN_DASH, dtitv_impl_SPACE, dtitv_impl_LEFT_CURLY_BRACKET, dtitv_impl_DIGIT_ONE, dtitv_impl_RIGHT_CURLY_BRACKET, 0};

DateIntervalInfo::DateIntervalInfo(UErrorCode& status)
:   fFallbackIntervalPattern(dtitvinf_gDefaultFallbackPattern),
    fFirstDateInPtnIsLaterDate(false),
    fIntervalPatterns(NULL)
{
    fIntervalPatterns = initHash(status);
}



DateIntervalInfo::DateIntervalInfo(const Locale& locale, UErrorCode& status)
:   fFallbackIntervalPattern(dtitvinf_gDefaultFallbackPattern),
    fFirstDateInPtnIsLaterDate(false),
    fIntervalPatterns(NULL)
{
    initializeData(locale, status);
}



void
DateIntervalInfo::setIntervalPattern(const UnicodeString& skeleton,
                                     UCalendarDateFields lrgDiffCalUnit,
                                     const UnicodeString& intervalPattern,
                                     UErrorCode& status) {

    if ( lrgDiffCalUnit == UCAL_HOUR_OF_DAY ) {
        setIntervalPatternInternally(skeleton, UCAL_AM_PM, intervalPattern, status);
        setIntervalPatternInternally(skeleton, UCAL_HOUR, intervalPattern, status);
    } else if ( lrgDiffCalUnit == UCAL_DAY_OF_MONTH ||
                lrgDiffCalUnit == UCAL_DAY_OF_WEEK ) {
        setIntervalPatternInternally(skeleton, UCAL_DATE, intervalPattern, status);
    } else {
        setIntervalPatternInternally(skeleton, lrgDiffCalUnit, intervalPattern, status);
    }
}


void
DateIntervalInfo::setFallbackIntervalPattern(
                                    const UnicodeString& fallbackPattern,
                                    UErrorCode& status) {
    if ( U_FAILURE(status) ) {
        return;
    }
    int32_t firstPatternIndex = fallbackPattern.indexOf(dtitvinf_gFirstPattern,
                        UPRV_LENGTHOF(dtitvinf_gFirstPattern), 0);
    int32_t secondPatternIndex = fallbackPattern.indexOf(dtitvinf_gSecondPattern,
                        UPRV_LENGTHOF(dtitvinf_gSecondPattern), 0);
    if ( firstPatternIndex == -1 || secondPatternIndex == -1 ) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if ( firstPatternIndex > secondPatternIndex ) {
        fFirstDateInPtnIsLaterDate = true;
    }
    fFallbackIntervalPattern = fallbackPattern;
}



DateIntervalInfo::DateIntervalInfo(const DateIntervalInfo& dtitvinf)
:   UObject(dtitvinf),
    fIntervalPatterns(NULL)
{
    *this = dtitvinf;
}



DateIntervalInfo&
DateIntervalInfo::operator=(const DateIntervalInfo& dtitvinf) {
    if ( this == &dtitvinf ) {
        return *this;
    }

    UErrorCode status = U_ZERO_ERROR;
    deleteHash(fIntervalPatterns);
    fIntervalPatterns = initHash(status);
    copyHash(dtitvinf.fIntervalPatterns, fIntervalPatterns, status);
    if ( U_FAILURE(status) ) {
        return *this;
    }

    fFallbackIntervalPattern = dtitvinf.fFallbackIntervalPattern;
    fFirstDateInPtnIsLaterDate = dtitvinf.fFirstDateInPtnIsLaterDate;
    return *this;
}


DateIntervalInfo*
DateIntervalInfo::clone() const {
    return new DateIntervalInfo(*this);
}


DateIntervalInfo::~DateIntervalInfo() {
    deleteHash(fIntervalPatterns);
    fIntervalPatterns = NULL;
}


UBool
DateIntervalInfo::operator==(const DateIntervalInfo& other) const {
    UBool equal = (
      fFallbackIntervalPattern == other.fFallbackIntervalPattern &&
      fFirstDateInPtnIsLaterDate == other.fFirstDateInPtnIsLaterDate );

    if ( equal == TRUE ) {
        equal = fIntervalPatterns->equals(*(other.fIntervalPatterns));
    }

    return equal;
}


UnicodeString&
DateIntervalInfo::getIntervalPattern(const UnicodeString& skeleton,
                                     UCalendarDateFields field,
                                     UnicodeString& result,
                                     UErrorCode& status) const {
    if ( U_FAILURE(status) ) {
        return result;
    }

    const UnicodeString* patternsOfOneSkeleton = (UnicodeString*) fIntervalPatterns->get(skeleton);
    if ( patternsOfOneSkeleton != NULL ) {
        IntervalPatternIndex index = calendarFieldToIntervalIndex(field, status);
        if ( U_FAILURE(status) ) {
            return result;
        }
        const UnicodeString& intervalPattern =  patternsOfOneSkeleton[index];
        if ( !intervalPattern.isEmpty() ) {
            result = intervalPattern;
        }
    }
    return result;
}


UBool
DateIntervalInfo::getDefaultOrder() const {
    return fFirstDateInPtnIsLaterDate;
}


UnicodeString&
DateIntervalInfo::getFallbackIntervalPattern(UnicodeString& result) const {
    result = fFallbackIntervalPattern;
    return result;
}

#define ULOC_LOCALE_IDENTIFIER_CAPACITY (ULOC_FULLNAME_CAPACITY + 1 + ULOC_KEYWORD_AND_VALUES_CAPACITY)


static const int32_t PATH_PREFIX_LENGTH = 17;
static const UChar PATH_PREFIX[] = {dtitv_impl_SOLIDUS, dtitv_impl_CAP_L, dtitv_impl_CAP_O, dtitv_impl_CAP_C, dtitv_impl_CAP_A, dtitv_impl_CAP_L, dtitv_impl_CAP_E, dtitv_impl_SOLIDUS,
                                    dtitv_impl_LOW_C, dtitv_impl_LOW_A, dtitv_impl_LOW_L, dtitv_impl_LOW_E, dtitv_impl_LOW_N, dtitv_impl_LOW_D, dtitv_impl_LOW_A, dtitv_impl_LOW_R, dtitv_impl_SOLIDUS};
static const int32_t PATH_SUFFIX_LENGTH = 16;
static const UChar PATH_SUFFIX[] = {dtitv_impl_SOLIDUS, dtitv_impl_LOW_I, dtitv_impl_LOW_N, dtitv_impl_LOW_T, dtitv_impl_LOW_E, dtitv_impl_LOW_R, dtitv_impl_LOW_V, dtitv_impl_LOW_A,
                                    dtitv_impl_LOW_L, dtitv_impl_CAP_F, dtitv_impl_LOW_O, dtitv_impl_LOW_R, dtitv_impl_LOW_M, dtitv_impl_LOW_A, dtitv_impl_LOW_T, dtitv_impl_LOW_S};

/**
 * Sink for enumerating all of the date interval skeletons.
 */
struct DateIntervalInfo::DateIntervalSink : public ResourceSink {

    // Output data
    DateIntervalInfo &dateIntervalInfo;

    // Next calendar type
    UnicodeString nextCalendarType;

    DateIntervalSink(DateIntervalInfo &diInfo, const char *currentCalendarType)
            : dateIntervalInfo(diInfo), nextCalendarType(currentCalendarType, -1, US_INV) { }
    virtual ~DateIntervalSink();

    virtual void put(const char *key, ResourceValue &value, UBool /*noFallback*/, UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return; }

        // Iterate over all the calendar entries and only pick the 'intervalFormats' table.
        ResourceTable dateIntervalData = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }
        for (int32_t i = 0; dateIntervalData.getKeyAndValue(i, key, value); i++) {
            if (uprv_strcmp(key, dtitvinf_gIntervalDateTimePatternTag) != 0) {
                continue;
            }

            // Handle aliases and tables. Ignore the rest.
            if (value.getType() == URES_ALIAS) {
                // Get the calendar type for the alias path.
                const UnicodeString &aliasPath = value.getAliasUnicodeString(errorCode);
                if (U_FAILURE(errorCode)) { return; }

                nextCalendarType.remove();
                getCalendarTypeFromPath(aliasPath, nextCalendarType, errorCode);

                if (U_FAILURE(errorCode)) {
                    resetNextCalendarType();
                }
                break;

            } else if (value.getType() == URES_TABLE) {
                // Iterate over all the skeletons in the 'intervalFormat' table.
                ResourceTable skeletonData = value.getTable(errorCode);
                if (U_FAILURE(errorCode)) { return; }
                for (int32_t j = 0; skeletonData.getKeyAndValue(j, key, value); j++) {
                    if (value.getType() == URES_TABLE) {
                        // Process the skeleton
                        processSkeletonTable(key, value, errorCode);
                        if (U_FAILURE(errorCode)) { return; }
                    }
                }
                break;
            }
        }
    }

    /**
     * Processes the patterns for a skeleton table
     */
    void processSkeletonTable(const char *key, ResourceValue &value, UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return; }

        // Iterate over all the patterns in the current skeleton table
        const char *currentSkeleton = key;
        ResourceTable patternData = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }
        for (int32_t k = 0; patternData.getKeyAndValue(k, key, value); k++) {
            if (value.getType() == URES_STRING) {
                // Process the key
                UCalendarDateFields calendarField = validateAndProcessPatternLetter(key);

                // If the calendar field has a valid value
                if (calendarField < UCAL_FIELD_COUNT) {
                    // Set the interval pattern
                    setIntervalPatternIfAbsent(currentSkeleton, calendarField, value, errorCode);
                    if (U_FAILURE(errorCode)) { return; }
                }
            }
        }
    }

    /**
     * Extracts the calendar type from the path.
     */
    static void getCalendarTypeFromPath(const UnicodeString &path, UnicodeString &calendarType,
                                        UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return; }

        if (!path.startsWith(PATH_PREFIX, PATH_PREFIX_LENGTH) || !path.endsWith(PATH_SUFFIX, PATH_SUFFIX_LENGTH)) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }

        path.extractBetween(PATH_PREFIX_LENGTH, path.length() - PATH_SUFFIX_LENGTH, calendarType);
    }

    /**
     * Validates and processes the pattern letter
     */
    UCalendarDateFields validateAndProcessPatternLetter(const char *patternLetter) {
        // Check that patternLetter is just one letter
        char c0;
        if ((c0 = patternLetter[0]) != 0 && patternLetter[1] == 0) {
            // Check that the pattern letter is accepted
            if (c0 == 'G') {
                return UCAL_ERA;
            } else if (c0 == 'y') {
                return UCAL_YEAR;
            } else if (c0 == 'M') {
                return UCAL_MONTH;
            } else if (c0 == 'd') {
                return UCAL_DATE;
            } else if (c0 == 'a') {
                return UCAL_AM_PM;
            } else if (c0 == 'h' || c0 == 'H') {
                return UCAL_HOUR;
            } else if (c0 == 'm') {
                return UCAL_MINUTE;
            }// TODO(ticket:12190): Why icu4c doesn't accept the calendar field "s" but icu4j does?
        }
        return UCAL_FIELD_COUNT;
    }

    /**
     * Stores the interval pattern for the current skeleton in the internal data structure
     * if it's not present.
     */
    void setIntervalPatternIfAbsent(const char *currentSkeleton, UCalendarDateFields lrgDiffCalUnit,
                                    const ResourceValue &value, UErrorCode &errorCode) {
        // Check if the pattern has already been stored on the data structure
        IntervalPatternIndex index =
            dateIntervalInfo.calendarFieldToIntervalIndex(lrgDiffCalUnit, errorCode);
        if (U_FAILURE(errorCode)) { return; }

        UnicodeString skeleton(currentSkeleton, -1, US_INV);
        UnicodeString* patternsOfOneSkeleton =
            (UnicodeString*)(dateIntervalInfo.fIntervalPatterns->get(skeleton));

        if (patternsOfOneSkeleton == NULL || patternsOfOneSkeleton[index].isEmpty()) {
            UnicodeString pattern = value.getUnicodeString(errorCode);
            dateIntervalInfo.setIntervalPatternInternally(skeleton, lrgDiffCalUnit,
                                                          pattern, errorCode);
        }
    }

    const UnicodeString &getNextCalendarType() {
        return nextCalendarType;
    }

    void resetNextCalendarType() {
        nextCalendarType.setToBogus();
    }
};

// Virtual destructors must be defined out of line.
DateIntervalInfo::DateIntervalSink::~DateIntervalSink() {}



void
DateIntervalInfo::initializeData(const Locale& locale, UErrorCode& status)
{
    fIntervalPatterns = initHash(status);
    if (U_FAILURE(status)) {
      return;
    }
    const char *locName = locale.getName();

    // Get the correct calendar type
    const char * calendarTypeToUse = dtitvinf_gGregorianTag; // initial default
    char         calendarType[ULOC_KEYWORDS_CAPACITY]; // to be filled in with the type to use, if all goes well
    char         localeWithCalendarKey[ULOC_LOCALE_IDENTIFIER_CAPACITY];
    // obtain a locale that always has the calendar key value that should be used
    (void)ures_getFunctionalEquivalent(localeWithCalendarKey, ULOC_LOCALE_IDENTIFIER_CAPACITY, NULL,
                                     "calendar", "calendar", locName, NULL, FALSE, &status);
    localeWithCalendarKey[ULOC_LOCALE_IDENTIFIER_CAPACITY-1] = 0; // ensure null termination
    // now get the calendar key value from that locale
    int32_t calendarTypeLen = uloc_getKeywordValue(localeWithCalendarKey, "calendar", calendarType,
                                                   ULOC_KEYWORDS_CAPACITY, &status);
    if (U_SUCCESS(status) && calendarTypeLen < ULOC_KEYWORDS_CAPACITY) {
        calendarTypeToUse = calendarType;
    }
    status = U_ZERO_ERROR;

    // Instantiate the resource bundles
    UResourceBundle *rb, *calBundle;
    rb = ures_open(NULL, locName, &status);
    if (U_FAILURE(status)) {
        return;
    }
    calBundle = ures_getByKeyWithFallback(rb, dtitvinf_gCalendarTag, NULL, &status);


    if (U_SUCCESS(status)) {
        UResourceBundle *calTypeBundle, *itvDtPtnResource;

        // Get the fallback pattern
        const UChar* resStr = nullptr;
        int32_t resStrLen = 0;
        calTypeBundle = ures_getByKeyWithFallback(calBundle, calendarTypeToUse, NULL, &status);
        itvDtPtnResource = ures_getByKeyWithFallback(calTypeBundle,
                                                     dtitvinf_gIntervalDateTimePatternTag, NULL, &status);
        // TODO(ICU-20400): After the fixing, we should find the "fallback" from
        // the rb directly by the path "calendar/${calendar}/intervalFormats/fallback".
        if ( U_SUCCESS(status) ) {
            resStr = ures_getStringByKeyWithFallback(itvDtPtnResource, dtitvinf_gFallbackPatternTag,
                                                     &resStrLen, &status);
            if ( U_FAILURE(status) ) {
                // Try to find "fallback" from "generic" to work around the bug in
                // ures_getByKeyWithFallback
                UErrorCode localStatus = U_ZERO_ERROR;
                UResourceBundle *genericCalBundle =
                    ures_getByKeyWithFallback(calBundle, dtitvinf_gGenericTag, NULL, &localStatus);
                UResourceBundle *genericItvDtPtnResource =
                    ures_getByKeyWithFallback(
                        genericCalBundle, dtitvinf_gIntervalDateTimePatternTag, NULL, &localStatus);
                resStr = ures_getStringByKeyWithFallback(
                    genericItvDtPtnResource, dtitvinf_gFallbackPatternTag, &resStrLen, &localStatus);
                ures_close(genericItvDtPtnResource);
                ures_close(genericCalBundle);
                if ( U_SUCCESS(localStatus) ) {
                    status = U_USING_FALLBACK_WARNING;;
                }
            }
        }

        if ( U_SUCCESS(status) && (resStr != nullptr)) {
            UnicodeString pattern = UnicodeString(TRUE, resStr, resStrLen);
            setFallbackIntervalPattern(pattern, status);
        }
        ures_close(itvDtPtnResource);
        ures_close(calTypeBundle);


        // Instantiate the sink
        DateIntervalSink sink(*this, calendarTypeToUse);
        const UnicodeString &calendarTypeToUseUString = sink.getNextCalendarType();

        // Already loaded calendar types
        Hashtable loadedCalendarTypes(FALSE, status);

        if (U_SUCCESS(status)) {
            while (!calendarTypeToUseUString.isBogus()) {
                // Set an error when a loop is detected
                if (loadedCalendarTypes.geti(calendarTypeToUseUString) == 1) {
                    status = U_INVALID_FORMAT_ERROR;
                    break;
                }

                // Register the calendar type to avoid loops
                loadedCalendarTypes.puti(calendarTypeToUseUString, 1, status);
                if (U_FAILURE(status)) { break; }

                // Get the calendar string
                CharString calTypeBuffer;
                calTypeBuffer.appendInvariantChars(calendarTypeToUseUString, status);
                if (U_FAILURE(status)) { break; }
                const char *calType = calTypeBuffer.data();

                // Reset the next calendar type to load.
                sink.resetNextCalendarType();

                // Get all resources for this calendar type
                ures_getAllItemsWithFallback(calBundle, calType, sink, status);
            }
        }
    }

    // Close the opened resource bundles
    ures_close(calBundle);
    ures_close(rb);
}

void
DateIntervalInfo::setIntervalPatternInternally(const UnicodeString& skeleton,
                                      UCalendarDateFields lrgDiffCalUnit,
                                      const UnicodeString& intervalPattern,
                                      UErrorCode& status) {
    IntervalPatternIndex index = calendarFieldToIntervalIndex(lrgDiffCalUnit,status);
    if ( U_FAILURE(status) ) {
        return;
    }
    UnicodeString* patternsOfOneSkeleton = (UnicodeString*)(fIntervalPatterns->get(skeleton));
    UBool emptyHash = false;
    if ( patternsOfOneSkeleton == NULL ) {
        patternsOfOneSkeleton = new UnicodeString[kIPI_MAX_INDEX];
        emptyHash = true;
    }

    patternsOfOneSkeleton[index] = intervalPattern;
    if ( emptyHash == TRUE ) {
        fIntervalPatterns->put(skeleton, patternsOfOneSkeleton, status);
    }
}



void
DateIntervalInfo::parseSkeleton(const UnicodeString& skeleton,
                                int32_t* skeletonFieldWidth) {
    const int8_t PATTERN_CHAR_BASE = 0x41;
    int32_t i;
    for ( i = 0; i < skeleton.length(); ++i ) {
        // it is an ASCII char in skeleton
        int8_t ch = (int8_t)skeleton.charAt(i);
        ++skeletonFieldWidth[ch - PATTERN_CHAR_BASE];
    }
}



UBool
DateIntervalInfo::stringNumeric(int32_t fieldWidth, int32_t anotherFieldWidth,
                                char patternLetter) {
    if ( patternLetter == 'M' ) {
        if ( (fieldWidth <= 2 && anotherFieldWidth > 2) ||
             (fieldWidth > 2 && anotherFieldWidth <= 2 )) {
            return true;
        }
    }
    return false;
}



const UnicodeString*
DateIntervalInfo::getBestSkeleton(const UnicodeString& skeleton,
                                  int8_t& bestMatchDistanceInfo) const {
#ifdef DTITVINF_DEBUG
    char result[1000];
    char result_1[1000];
    char mesg[2000];
    skeleton.extract(0,  skeleton.length(), result, "UTF-8");
    sprintf(mesg, "in getBestSkeleton: skeleton: %s; \n", result);
    PRINTMESG(mesg)
#endif


    int32_t inputSkeletonFieldWidth[] =
    {
    //       A   B   C   D   E   F   G   H   I   J   K   L   M   N   O
             0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    //   P   Q   R   S   T   U   V   W   X   Y   Z
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 0, 0,  0, 0, 0,
    //       a   b   c   d   e   f   g   h   i   j   k   l   m   n   o
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    //   p   q   r   s   t   u   v   w   x   y   z
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
    };

    int32_t skeletonFieldWidth[] =
    {
    //       A   B   C   D   E   F   G   H   I   J   K   L   M   N   O
             0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    //   P   Q   R   S   T   U   V   W   X   Y   Z
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 0, 0,  0, 0, 0,
    //       a   b   c   d   e   f   g   h   i   j   k   l   m   n   o
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    //   p   q   r   s   t   u   v   w   x   y   z
         0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
    };

    const int32_t DIFFERENT_FIELD = 0x1000;
    const int32_t STRING_NUMERIC_DIFFERENCE = 0x100;
    const int32_t BASE = 0x41;
    const UChar CHAR_V = 0x0076;
    const UChar CHAR_Z = 0x007A;

    // hack for 'v' and 'z'.
    // resource bundle only have time skeletons ending with 'v',
    // but not for time skeletons ending with 'z'.
    UBool replaceZWithV = false;
    const UnicodeString* inputSkeleton = &skeleton;
    UnicodeString copySkeleton;
    if ( skeleton.indexOf(CHAR_Z) != -1 ) {
        copySkeleton = skeleton;
        copySkeleton.findAndReplace(UnicodeString(CHAR_Z), UnicodeString(CHAR_V));
        inputSkeleton = &copySkeleton;
        replaceZWithV = true;
    }

    parseSkeleton(*inputSkeleton, inputSkeletonFieldWidth);
    int32_t bestDistance = dtitv_impl_MAX_POSITIVE_INT;
    const UnicodeString* bestSkeleton = NULL;

    // 0 means exact the same skeletons;
    // 1 means having the same field, but with different length,
    // 2 means only z/v differs
    // -1 means having different field.
    bestMatchDistanceInfo = 0;
    int8_t fieldLength = UPRV_LENGTHOF(skeletonFieldWidth);

    int32_t pos = UHASH_FIRST;
    const UHashElement* elem = NULL;
    while ( (elem = fIntervalPatterns->nextElement(pos)) != NULL ) {
        const UHashTok keyTok = elem->key;
        UnicodeString* newSkeleton = (UnicodeString*)keyTok.pointer;
#ifdef DTITVINF_DEBUG
    skeleton->extract(0,  skeleton->length(), result, "UTF-8");
    sprintf(mesg, "available skeletons: skeleton: %s; \n", result);
    PRINTMESG(mesg)
#endif

        // clear skeleton field width
        int8_t i;
        for ( i = 0; i < fieldLength; ++i ) {
            skeletonFieldWidth[i] = 0;
        }
        parseSkeleton(*newSkeleton, skeletonFieldWidth);
        // calculate distance
        int32_t distance = 0;
        int8_t fieldDifference = 1;
        for ( i = 0; i < fieldLength; ++i ) {
            int32_t inputFieldWidth = inputSkeletonFieldWidth[i];
            int32_t fieldWidth = skeletonFieldWidth[i];
            if ( inputFieldWidth == fieldWidth ) {
                continue;
            }
            if ( inputFieldWidth == 0 ) {
                fieldDifference = -1;
                distance += DIFFERENT_FIELD;
            } else if ( fieldWidth == 0 ) {
                fieldDifference = -1;
                distance += DIFFERENT_FIELD;
            } else if (stringNumeric(inputFieldWidth, fieldWidth,
                                     (char)(i+BASE) ) ) {
                distance += STRING_NUMERIC_DIFFERENCE;
            } else {
                distance += (inputFieldWidth > fieldWidth) ?
                            (inputFieldWidth - fieldWidth) :
                            (fieldWidth - inputFieldWidth);
            }
        }
        if ( distance < bestDistance ) {
            bestSkeleton = newSkeleton;
            bestDistance = distance;
            bestMatchDistanceInfo = fieldDifference;
        }
        if ( distance == 0 ) {
            bestMatchDistanceInfo = 0;
            break;
        }
    }
    if ( replaceZWithV && bestMatchDistanceInfo != -1 ) {
        bestMatchDistanceInfo = 2;
    }
    return bestSkeleton;
}



DateIntervalInfo::IntervalPatternIndex
DateIntervalInfo::calendarFieldToIntervalIndex(UCalendarDateFields field,
                                               UErrorCode& status) {
    if ( U_FAILURE(status) ) {
        return kIPI_MAX_INDEX;
    }
    IntervalPatternIndex index = kIPI_MAX_INDEX;
    switch ( field ) {
      case UCAL_ERA:
        index = kIPI_ERA;
        break;
      case UCAL_YEAR:
        index = kIPI_YEAR;
        break;
      case UCAL_MONTH:
        index = kIPI_MONTH;
        break;
      case UCAL_DATE:
      case UCAL_DAY_OF_WEEK:
      //case UCAL_DAY_OF_MONTH:
        index = kIPI_DATE;
        break;
      case UCAL_AM_PM:
        index = kIPI_AM_PM;
        break;
      case UCAL_HOUR:
      case UCAL_HOUR_OF_DAY:
        index = kIPI_HOUR;
        break;
      case UCAL_MINUTE:
        index = kIPI_MINUTE;
        break;
      case UCAL_SECOND:
        index = kIPI_SECOND;
        break;
      default:
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return index;
}



void
DateIntervalInfo::deleteHash(Hashtable* hTable)
{
    if ( hTable == NULL ) {
        return;
    }
    int32_t pos = UHASH_FIRST;
    const UHashElement* element = NULL;
    while ( (element = hTable->nextElement(pos)) != NULL ) {
        const UHashTok valueTok = element->value;
        const UnicodeString* value = (UnicodeString*)valueTok.pointer;
        delete[] value;
    }
    delete fIntervalPatterns;
}


U_CDECL_BEGIN

/**
 * set hash table value comparator
 *
 * @param val1  one value in comparison
 * @param val2  the other value in comparison
 * @return      TRUE if 2 values are the same, FALSE otherwise
 */
static UBool U_CALLCONV dtitvinfHashTableValueComparator(UHashTok val1, UHashTok val2);

static UBool
U_CALLCONV dtitvinfHashTableValueComparator(UHashTok val1, UHashTok val2) {
    const UnicodeString* pattern1 = (UnicodeString*)val1.pointer;
    const UnicodeString* pattern2 = (UnicodeString*)val2.pointer;
    UBool ret = TRUE;
    int8_t i;
    for ( i = 0; i < DateIntervalInfo::kMaxIntervalPatternIndex && ret == TRUE; ++i ) {
        ret = (pattern1[i] == pattern2[i]);
    }
    return ret;
}

U_CDECL_END


Hashtable*
DateIntervalInfo::initHash(UErrorCode& status) {
    if ( U_FAILURE(status) ) {
        return NULL;
    }
    Hashtable* hTable;
    if ( (hTable = new Hashtable(FALSE, status)) == NULL ) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    if ( U_FAILURE(status) ) {
        delete hTable;
        return NULL;
    }
    hTable->setValueComparator(dtitvinfHashTableValueComparator);
    return hTable;
}


void
DateIntervalInfo::copyHash(const Hashtable* source,
                           Hashtable* target,
                           UErrorCode& status) {
    if ( U_FAILURE(status) ) {
        return;
    }
    int32_t pos = UHASH_FIRST;
    const UHashElement* element = NULL;
    if ( source ) {
        while ( (element = source->nextElement(pos)) != NULL ) {
            const UHashTok keyTok = element->key;
            const UnicodeString* key = (UnicodeString*)keyTok.pointer;
            const UHashTok valueTok = element->value;
            const UnicodeString* value = (UnicodeString*)valueTok.pointer;
            UnicodeString* copy = new UnicodeString[kIPI_MAX_INDEX];
            int8_t i;
            for ( i = 0; i < kIPI_MAX_INDEX; ++i ) {
                copy[i] = value[i];
            }
            target->put(UnicodeString(*key), copy, status);
            if ( U_FAILURE(status) ) {
                return;
            }
        }
    }
}


U_NAMESPACE_END

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File DTPTNGEN.CPP
*
*******************************************************************************
*/


#if !UCONFIG_NO_FORMATTING















// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2014-2016, International Business Machines Corporation and others.
 * All Rights Reserved.
 *******************************************************************************
 */


/**
 * \file 
 * \brief C++ API: Region classes (territory containment)
 */



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2014, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/





/**
 * \file
 * \brief C API: URegion (territory containment and mapping)
 *
 * URegion objects represent data associated with a particular Unicode Region Code, also known as a 
 * Unicode Region Subtag, which is defined based upon the BCP 47 standard. These include:
 * * Two-letter codes defined by ISO 3166-1, with special LDML treatment of certain private-use or
 *   reserved codes;
 * * A subset of 3-digit numeric codes defined by UN M.49.
 * URegion objects can also provide mappings to and from additional codes. There are different types
 * of regions that are important to distinguish:
 * <p>
 * Macroregion - A code for a "macro geographical (continental) region, geographical sub-region, or 
 * selected economic and other grouping" as defined in UN M.49. These are typically 3-digit codes,
 * but contain some 2-letter codes for LDML extensions, such as "QO" for Outlying Oceania.
 * Macroregions are represented in ICU by one of three region types: WORLD (code 001),
 * CONTINENTS (regions contained directly by WORLD), and SUBCONTINENTS (regions contained directly
 * by a continent ).
 * <p>
 * TERRITORY - A Region that is not a Macroregion. These are typically codes for countries, but also
 * include areas that are not separate countries, such as the code "AQ" for Antarctica or the code 
 * "HK" for Hong Kong (SAR China). Overseas dependencies of countries may or may not have separate 
 * codes. The codes are typically 2-letter codes aligned with ISO 3166, but BCP47 allows for the use
 * of 3-digit codes in the future.
 * <p>
 * UNKNOWN - The code ZZ is defined by Unicode LDML for use in indicating that region is unknown,
 * or that the value supplied as a region was invalid.
 * <p>
 * DEPRECATED - Region codes that have been defined in the past but are no longer in modern usage,
 * usually due to a country splitting into multiple territories or changing its name.
 * <p>
 * GROUPING - A widely understood grouping of territories that has a well defined membership such
 * that a region code has been assigned for it.  Some of these are UN M.49 codes that don't fall into 
 * the world/continent/sub-continent hierarchy, while others are just well-known groupings that have
 * their own region code. Region "EU" (European Union) is one such region code that is a grouping.
 * Groupings will never be returned by the uregion_getContainingRegion, since a different type of region
 * (WORLD, CONTINENT, or SUBCONTINENT) will always be the containing region instead.
 *
 * URegion objects are const/immutable, owned and maintained by ICU itself, so there are not functions
 * to open or close them.
 */

/**
 * URegionType is an enumeration defining the different types of regions.  Current possible
 * values are URGN_WORLD, URGN_CONTINENT, URGN_SUBCONTINENT, URGN_TERRITORY, URGN_GROUPING,
 * URGN_DEPRECATED, and URGN_UNKNOWN.
 * 
 * @stable ICU 51 
 */
typedef enum URegionType {
    /**
     * Type representing the unknown region.
     * @stable ICU 51 
     */
    URGN_UNKNOWN,

    /**
     * Type representing a territory.
     * @stable ICU 51 
     */
    URGN_TERRITORY,

    /**
     * Type representing the whole world.
     * @stable ICU 51 
     */
    URGN_WORLD,

    /**
     * Type representing a continent.
     * @stable ICU 51 
     */
    URGN_CONTINENT,

    /**
     * Type representing a sub-continent.
     * @stable ICU 51 
     */
    URGN_SUBCONTINENT,

    /**
     * Type representing a grouping of territories that is not to be used in
     * the normal WORLD/CONTINENT/SUBCONTINENT/TERRITORY containment tree.
     * @stable ICU 51 
     */
    URGN_GROUPING,

    /**
     * Type representing a region whose code has been deprecated, usually
     * due to a country splitting into multiple territories or changing its name.
     * @stable ICU 51 
     */
    URGN_DEPRECATED,

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal URegionType value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    URGN_LIMIT
#endif  /* U_HIDE_DEPRECATED_API */
} URegionType;

#if !UCONFIG_NO_FORMATTING

/**
 * Opaque URegion object for use in C programs.
 * @stable ICU 52
 */
struct URegion;
typedef struct URegion URegion; /**< @stable ICU 52 */

/**
 * Returns a pointer to a URegion for the specified region code: A 2-letter or 3-letter ISO 3166
 * code, UN M.49 numeric code (superset of ISO 3166 numeric codes), or other valid Unicode Region
 * Code as defined by the LDML specification. The code will be canonicalized internally. If the
 * region code is NULL or not recognized, the appropriate error code will be set
 * (U_ILLEGAL_ARGUMENT_ERROR).
 * @stable ICU 52
 */
U_STABLE const URegion* U_EXPORT2
uregion_getRegionFromCode(const char *regionCode, UErrorCode *status);

/**
 * Returns a pointer to a URegion for the specified numeric region code. If the numeric region
 * code is not recognized, the appropriate error code will be set (U_ILLEGAL_ARGUMENT_ERROR).
 * @stable ICU 52
 */
U_STABLE const URegion* U_EXPORT2
uregion_getRegionFromNumericCode (int32_t code, UErrorCode *status);

/**
 * Returns an enumeration over the canonical codes of all known regions that match the given type.
 * The enumeration must be closed with with uenum_close().
 * @stable ICU 52
 */
U_STABLE UEnumeration* U_EXPORT2
uregion_getAvailable(URegionType type, UErrorCode *status);

/**
 * Returns true if the specified uregion is equal to the specified otherRegion.
 * @stable ICU 52
 */
U_STABLE UBool U_EXPORT2
uregion_areEqual(const URegion* uregion, const URegion* otherRegion);

/**
 * Returns a pointer to the URegion that contains the specified uregion. Returns NULL if the
 * specified uregion is code "001" (World) or "ZZ" (Unknown region). For example, calling
 * this method with region "IT" (Italy) returns the URegion for "039" (Southern Europe).
 * @stable ICU 52
 */
U_STABLE const URegion* U_EXPORT2
uregion_getContainingRegion(const URegion* uregion);

/**
 * Return a pointer to the URegion that geographically contains this uregion and matches the
 * specified type, moving multiple steps up the containment chain if necessary. Returns NULL if no
 * containing region can be found that matches the specified type. Will return NULL if URegionType
 * is URGN_GROUPING, URGN_DEPRECATED, or URGN_UNKNOWN which are not appropriate for this API.
 * For example, calling this method with uregion "IT" (Italy) for type URGN_CONTINENT returns the
 * URegion "150" (Europe).
 * @stable ICU 52
 */
U_STABLE const URegion* U_EXPORT2
uregion_getContainingRegionOfType(const URegion* uregion, URegionType type);

/**
 * Return an enumeration over the canonical codes of all the regions that are immediate children
 * of the specified uregion in the region hierarchy. These returned regions could be either macro
 * regions, territories, or a mixture of the two, depending on the containment data as defined in
 * CLDR. This API returns NULL if this uregion doesn't have any sub-regions. For example, calling
 * this function for uregion "150" (Europe) returns an enumeration containing the various
 * sub-regions of Europe: "039" (Southern Europe), "151" (Eastern Europe), "154" (Northern Europe),
 * and "155" (Western Europe). The enumeration must be closed with with uenum_close().
 * @stable ICU 52
 */
U_STABLE UEnumeration* U_EXPORT2
uregion_getContainedRegions(const URegion* uregion, UErrorCode *status);

/**
 * Returns an enumeration over the canonical codes of all the regions that are children of the
 * specified uregion anywhere in the region hierarchy and match the given type. This API may return
 * an empty enumeration if this uregion doesn't have any sub-regions that match the given type.
 * For example, calling this method with region "150" (Europe) and type URGN_TERRITORY" returns an
 * enumeration containing all the territories in Europe: "FR" (France), "IT" (Italy), "DE" (Germany),
 * etc. The enumeration must be closed with with uenum_close().
 * @stable ICU 52
 */
U_STABLE UEnumeration* U_EXPORT2
uregion_getContainedRegionsOfType(const URegion* uregion, URegionType type, UErrorCode *status);

/**
 * Returns true if the specified uregion contains the specified otherRegion anywhere in the region
 * hierarchy.
 * @stable ICU 52
 */
U_STABLE UBool U_EXPORT2
uregion_contains(const URegion* uregion, const URegion* otherRegion);

/**
 * If the specified uregion is deprecated, returns an enumeration over the canonical codes of the
 * regions that are the preferred replacement regions for the specified uregion. If the specified
 * uregion is not deprecated, returns NULL. For example, calling this method with uregion
 * "SU" (Soviet Union) returns a list of the regions containing "RU" (Russia), "AM" (Armenia),
 * "AZ" (Azerbaijan), etc... The enumeration must be closed with with uenum_close().
 * @stable ICU 52
 */
U_STABLE UEnumeration* U_EXPORT2
uregion_getPreferredValues(const URegion* uregion, UErrorCode *status);

/**
 * Returns the specified uregion's canonical code.
 * @stable ICU 52
 */
U_STABLE const char* U_EXPORT2
uregion_getRegionCode(const URegion* uregion);

/**
 * Returns the specified uregion's numeric code, or a negative value if there is no numeric code
 * for the specified uregion.
 * @stable ICU 52
 */
U_STABLE int32_t U_EXPORT2
uregion_getNumericCode(const URegion* uregion);

/**
 * Returns the URegionType of the specified uregion.
 * @stable ICU 52
 */
U_STABLE URegionType U_EXPORT2
uregion_getType(const URegion* uregion);


#endif /* #if !UCONFIG_NO_FORMATTING */







U_NAMESPACE_BEGIN

/**
 * <code>Region</code> is the class representing a Unicode Region Code, also known as a 
 * Unicode Region Subtag, which is defined based upon the BCP 47 standard. We often think of
 * "regions" as "countries" when defining the characteristics of a locale.  Region codes There are different
 * types of region codes that are important to distinguish.
 * <p>
 *  Macroregion - A code for a "macro geographical (continental) region, geographical sub-region, or 
 *  selected economic and other grouping" as defined in 
 *  UN M.49 (http://unstats.un.org/unsd/methods/m49/m49regin.htm). 
 *  These are typically 3-digit codes, but contain some 2-letter codes, such as the LDML code QO 
 *  added for Outlying Oceania.  Not all UNM.49 codes are defined in LDML, but most of them are.
 *  Macroregions are represented in ICU by one of three region types: WORLD ( region code 001 ),
 *  CONTINENTS ( regions contained directly by WORLD ), and SUBCONTINENTS ( things contained directly
 *  by a continent ).
 *  <p>
 *  TERRITORY - A Region that is not a Macroregion. These are typically codes for countries, but also
 *  include areas that are not separate countries, such as the code "AQ" for Antarctica or the code 
 *  "HK" for Hong Kong (SAR China). Overseas dependencies of countries may or may not have separate 
 *  codes. The codes are typically 2-letter codes aligned with the ISO 3166 standard, but BCP47 allows
 *  for the use of 3-digit codes in the future.
 *  <p>
 *  UNKNOWN - The code ZZ is defined by Unicode LDML for use to indicate that the Region is unknown,
 *  or that the value supplied as a region was invalid.
 *  <p>
 *  DEPRECATED - Region codes that have been defined in the past but are no longer in modern usage,
 *  usually due to a country splitting into multiple territories or changing its name.
 *  <p>
 *  GROUPING - A widely understood grouping of territories that has a well defined membership such
 *  that a region code has been assigned for it.  Some of these are UNM.49 codes that do't fall into 
 *  the world/continent/sub-continent hierarchy, while others are just well known groupings that have
 *  their own region code. Region "EU" (European Union) is one such region code that is a grouping.
 *  Groupings will never be returned by the getContainingRegion() API, since a different type of region
 *  ( WORLD, CONTINENT, or SUBCONTINENT ) will always be the containing region instead.
 *
 * The Region class is not intended for public subclassing.
 *
 * @author       John Emmons
 * @stable ICU 51
 */

class U_I18N_API Region : public UObject {
public:
    /**
     * Destructor.
     * @stable ICU 51
     */
    virtual ~Region();

    /**
     * Returns true if the two regions are equal.
     * @stable ICU 51
     */
    UBool operator==(const Region &that) const;

    /**
     * Returns true if the two regions are NOT equal; that is, if operator ==() returns false.
     * @stable ICU 51
     */
    UBool operator!=(const Region &that) const;
 
    /**
     * Returns a pointer to a Region using the given region code.  The region code can be either 2-letter ISO code,
     * 3-letter ISO code,  UNM.49 numeric code, or other valid Unicode Region Code as defined by the LDML specification.
     * The identifier will be canonicalized internally using the supplemental metadata as defined in the CLDR.
     * If the region code is NULL or not recognized, the appropriate error code will be set ( U_ILLEGAL_ARGUMENT_ERROR )
     * @stable ICU 51 
     */
    static const Region* U_EXPORT2 getInstance(const char *region_code, UErrorCode &status);

    /**
     * Returns a pointer to a Region using the given numeric region code. If the numeric region code is not recognized,
     * the appropriate error code will be set ( U_ILLEGAL_ARGUMENT_ERROR ).
     * @stable ICU 51 
     */
    static const Region* U_EXPORT2 getInstance (int32_t code, UErrorCode &status);

    /**
     * Returns an enumeration over the IDs of all known regions that match the given type.
     * @stable ICU 55
     */
    static StringEnumeration* U_EXPORT2 getAvailable(URegionType type, UErrorCode &status);
   
    /**
     * Returns a pointer to the region that contains this region.  Returns NULL if this region is code "001" (World)
     * or "ZZ" (Unknown region). For example, calling this method with region "IT" (Italy) returns the
     * region "039" (Southern Europe).
     * @stable ICU 51 
     */
    const Region* getContainingRegion() const;

    /**
     * Return a pointer to the region that geographically contains this region and matches the given type,
     * moving multiple steps up the containment chain if necessary.  Returns NULL if no containing region can be found
     * that matches the given type. Note: The URegionTypes = "URGN_GROUPING", "URGN_DEPRECATED", or "URGN_UNKNOWN"
     * are not appropriate for use in this API. NULL will be returned in this case. For example, calling this method
     * with region "IT" (Italy) for type "URGN_CONTINENT" returns the region "150" ( Europe ).
     * @stable ICU 51 
     */
    const Region* getContainingRegion(URegionType type) const;

    /**
     * Return an enumeration over the IDs of all the regions that are immediate children of this region in the
     * region hierarchy. These returned regions could be either macro regions, territories, or a mixture of the two,
     * depending on the containment data as defined in CLDR.  This API may return NULL if this region doesn't have
     * any sub-regions. For example, calling this method with region "150" (Europe) returns an enumeration containing
     * the various sub regions of Europe - "039" (Southern Europe) - "151" (Eastern Europe) - "154" (Northern Europe)
     * and "155" (Western Europe).
     * @stable ICU 55
     */
    StringEnumeration* getContainedRegions(UErrorCode &status) const;

    /**
     * Returns an enumeration over the IDs of all the regions that are children of this region anywhere in the region
     * hierarchy and match the given type.  This API may return an empty enumeration if this region doesn't have any
     * sub-regions that match the given type. For example, calling this method with region "150" (Europe) and type
     * "URGN_TERRITORY" returns a set containing all the territories in Europe ( "FR" (France) - "IT" (Italy) - "DE" (Germany) etc. )
     * @stable ICU 55
     */
    StringEnumeration* getContainedRegions( URegionType type, UErrorCode &status ) const;
 
    /**
     * Returns true if this region contains the supplied other region anywhere in the region hierarchy.
     * @stable ICU 51 
     */
    UBool contains(const Region &other) const;

    /**
     * For deprecated regions, return an enumeration over the IDs of the regions that are the preferred replacement
     * regions for this region.  Returns null for a non-deprecated region.  For example, calling this method with region
     * "SU" (Soviet Union) would return a list of the regions containing "RU" (Russia), "AM" (Armenia), "AZ" (Azerbaijan), etc...
     * @stable ICU 55 
     */
    StringEnumeration* getPreferredValues(UErrorCode &status) const;

    /**
     * Return this region's canonical region code.
     * @stable ICU 51 
     */
    const char* getRegionCode() const;

    /**
     * Return this region's numeric code.
     * Returns a negative value if the given region does not have a numeric code assigned to it.
     * @stable ICU 51 
     */
    int32_t getNumericCode() const;

    /**
     * Returns the region type of this region.
     * @stable ICU 51 
     */
    URegionType getType() const;

#ifndef U_HIDE_INTERNAL_API
    /**
     * Cleans up statically allocated memory.
     * @internal 
     */
    static void cleanupRegionData();
#endif  /* U_HIDE_INTERNAL_API */

private:
    char id[4];
    UnicodeString idStr;
    int32_t code;
    URegionType fType;
    Region *containingRegion;
    UVector *containedRegions;
    UVector *preferredValues;

    /**
     * Default Constructor. Internal - use factory methods only.
     */
    Region();


    /*
     * Initializes the region data from the ICU resource bundles.  The region data
     * contains the basic relationships such as which regions are known, what the numeric
     * codes are, any known aliases, and the territory containment data.
     * 
     * If the region data has already loaded, then this method simply returns without doing
     * anything meaningful.
     */

    static void U_CALLCONV loadRegionData(UErrorCode &status);

};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */











// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File DTPTNGEN.H
*
*******************************************************************************
*/








// TODO(claireho): Split off Builder class.
// TODO(claireho): If splitting off Builder class: As subclass or independent?

#define dtptngen_MAX_PATTERN_ENTRIES 52
#define dtptngen_MAX_CLDR_FIELD_LEN  60
#define dtptngen_MAX_DT_TOKEN        50
#define dtptngen_MAX_RESOURCE_FIELD  12
#define dtptngen_MAX_AVAILABLE_FORMATS  12
#define dtptngen_NONE          0
#define dtptngen_EXTRA_FIELD   0x10000
#define dtptngen_MISSING_FIELD  0x1000
#define dtptngen_MAX_STRING_ENUMERATION  200
#define dtptngen_SINGLE_QUOTE      ((UChar)0x0027)
#define dtptngen_FORWARDSLASH      ((UChar)0x002F)
#define dtptngen_BACKSLASH         ((UChar)0x005C)
#define dtptngen_SPACE             ((UChar)0x0020)
#define dtptngen_QUOTATION_MARK    ((UChar)0x0022)
#define dtptngen_ASTERISK          ((UChar)0x002A)
#define dtptngen_PLUSSITN          ((UChar)0x002B)
#define dtptngen_COMMA             ((UChar)0x002C)
#define dtptngen_HYPHEN            ((UChar)0x002D)
#define dtptngen_DOT               ((UChar)0x002E)
#define dtptngen_COLON             ((UChar)0x003A)
#define dtptngen_CAP_A             ((UChar)0x0041)
#define dtptngen_CAP_B             ((UChar)0x0042)
#define dtptngen_CAP_C             ((UChar)0x0043)
#define dtptngen_CAP_D             ((UChar)0x0044)
#define dtptngen_CAP_E             ((UChar)0x0045)
#define dtptngen_CAP_F             ((UChar)0x0046)
#define dtptngen_CAP_G             ((UChar)0x0047)
#define dtptngen_CAP_H             ((UChar)0x0048)
#define dtptngen_CAP_J             ((UChar)0x004A)
#define dtptngen_CAP_K             ((UChar)0x004B)
#define dtptngen_CAP_L             ((UChar)0x004C)
#define dtptngen_CAP_M             ((UChar)0x004D)
#define dtptngen_CAP_O             ((UChar)0x004F)
#define dtptngen_CAP_Q             ((UChar)0x0051)
#define dtptngen_CAP_S             ((UChar)0x0053)
#define dtptngen_CAP_T             ((UChar)0x0054)
#define dtptngen_CAP_U             ((UChar)0x0055)
#define dtptngen_CAP_V             ((UChar)0x0056)
#define dtptngen_CAP_W             ((UChar)0x0057)
#define dtptngen_CAP_X             ((UChar)0x0058)
#define dtptngen_CAP_Y             ((UChar)0x0059)
#define dtptngen_CAP_Z             ((UChar)0x005A)
#define dtptngen_LOWLINE           ((UChar)0x005F)
#define dtptngen_LOW_A             ((UChar)0x0061)
#define dtptngen_LOW_B             ((UChar)0x0062)
#define dtptngen_LOW_C             ((UChar)0x0063)
#define dtptngen_LOW_D             ((UChar)0x0064)
#define dtptngen_LOW_E             ((UChar)0x0065)
#define dtptngen_LOW_F             ((UChar)0x0066)
#define dtptngen_LOW_G             ((UChar)0x0067)
#define dtptngen_LOW_H             ((UChar)0x0068)
#define dtptngen_LOW_I             ((UChar)0x0069)
#define dtptngen_LOW_J             ((UChar)0x006A)
#define dtptngen_LOW_K             ((UChar)0x006B)
#define dtptngen_LOW_L             ((UChar)0x006C)
#define dtptngen_LOW_M             ((UChar)0x006D)
#define dtptngen_LOW_N             ((UChar)0x006E)
#define dtptngen_LOW_O             ((UChar)0x006F)
#define dtptngen_LOW_P             ((UChar)0x0070)
#define dtptngen_LOW_Q             ((UChar)0x0071)
#define dtptngen_LOW_R             ((UChar)0x0072)
#define dtptngen_LOW_S             ((UChar)0x0073)
#define dtptngen_LOW_T             ((UChar)0x0074)
#define dtptngen_LOW_U             ((UChar)0x0075)
#define dtptngen_LOW_V             ((UChar)0x0076)
#define dtptngen_LOW_W             ((UChar)0x0077)
#define dtptngen_LOW_X             ((UChar)0x0078)
#define dtptngen_LOW_Y             ((UChar)0x0079)
#define dtptngen_LOW_Z             ((UChar)0x007A)
#define dtptngen_DT_NARROW         -0x101
#define dtptngen_DT_SHORTER        -0x102
#define dtptngen_DT_SHORT          -0x103
#define dtptngen_DT_LONG           -0x104
#define dtptngen_DT_NUMERIC         0x100
#define dtptngen_DT_DELTA           0x10

U_NAMESPACE_BEGIN

const int32_t UDATPG_FRACTIONAL_MASK = 1<<UDATPG_FRACTIONAL_SECOND_FIELD;
const int32_t UDATPG_SECOND_AND_FRACTIONAL_MASK = (1<<UDATPG_SECOND_FIELD) | (1<<UDATPG_FRACTIONAL_SECOND_FIELD);

typedef enum dtStrEnum {
    DT_BASESKELETON,
    DT_SKELETON,
    DT_PATTERN
}dtStrEnum;

typedef struct dtTypeElem {
    UChar                  patternChar;
    UDateTimePatternField  field;
    int16_t                type;
    int16_t                minLen;
    int16_t                weight;
} dtTypeElem;

// A compact storage mechanism for skeleton field strings.  Several dozen of these will be created
// for a typical DateTimePatternGenerator instance.
class SkeletonFields : public UMemory {
public:
    SkeletonFields();
    void clear();
    void copyFrom(const SkeletonFields& other);
    void clearField(int32_t field);
    UChar getFieldChar(int32_t field) const;
    int32_t getFieldLength(int32_t field) const;
    void populate(int32_t field, const UnicodeString& value);
    void populate(int32_t field, UChar repeatChar, int32_t repeatCount);
    UBool isFieldEmpty(int32_t field) const;
    UnicodeString& appendTo(UnicodeString& string) const;
    UnicodeString& appendFieldTo(int32_t field, UnicodeString& string) const;
    UChar getFirstChar() const;
    inline UBool operator==(const SkeletonFields& other) const;
    inline UBool operator!=(const SkeletonFields& other) const;

private:
    int8_t chars[UDATPG_FIELD_COUNT];
    int8_t lengths[UDATPG_FIELD_COUNT];
};

inline UBool SkeletonFields::operator==(const SkeletonFields& other) const {
    return (uprv_memcmp(chars, other.chars, sizeof(chars)) == 0
        && uprv_memcmp(lengths, other.lengths, sizeof(lengths)) == 0);
}

inline UBool SkeletonFields::operator!=(const SkeletonFields& other) const {
    return (! operator==(other));
}

class PtnSkeleton : public UMemory {
public:
    int32_t type[UDATPG_FIELD_COUNT];
    SkeletonFields original;
    SkeletonFields baseOriginal;
    UBool addedDefaultDayPeriod;

    PtnSkeleton();
    PtnSkeleton(const PtnSkeleton& other);
    void copyFrom(const PtnSkeleton& other);
    void clear();
    UBool equals(const PtnSkeleton& other) const;
    UnicodeString getSkeleton() const;
    UnicodeString getBaseSkeleton() const;
    UChar getFirstChar() const;

    // TODO: Why is this virtual, as well as the other destructors in this file? We don't want
    // vtables when we don't use class objects polymorphically.
    virtual ~PtnSkeleton();
};

class PtnElem : public UMemory {
public:
    UnicodeString basePattern;
    LocalPointer<PtnSkeleton> skeleton;
    UnicodeString pattern;
    UBool         skeletonWasSpecified; // if specified in availableFormats, not derived
    LocalPointer<PtnElem> next;

    PtnElem(const UnicodeString &basePattern, const UnicodeString &pattern);
    virtual ~PtnElem();
};

class FormatParser : public UMemory {
public:
    UnicodeString items[dtptngen_MAX_DT_TOKEN];
    int32_t itemNumber;

    FormatParser();
    virtual ~FormatParser();
    void set(const UnicodeString& patternString);
    void getQuoteLiteral(UnicodeString& quote, int32_t *itemIndex);
    UBool isPatternSeparator(const UnicodeString& field) const;
    static UBool isQuoteLiteral(const UnicodeString& s);
    static int32_t getCanonicalIndex(const UnicodeString& s) { return getCanonicalIndex(s, TRUE); }
    static int32_t getCanonicalIndex(const UnicodeString& s, UBool strict);

private:
   typedef enum TokenStatus {
       START,
       ADD_TOKEN,
       SYNTAX_ERROR,
       DONE
   } TokenStatus;

   TokenStatus status;
   virtual TokenStatus setTokens(const UnicodeString& pattern, int32_t startPos, int32_t *len);
};

class DistanceInfo : public UMemory {
public:
    int32_t missingFieldMask;
    int32_t extraFieldMask;

    DistanceInfo() {}
    virtual ~DistanceInfo();
    void clear() { missingFieldMask = extraFieldMask = 0; }
    void setTo(const DistanceInfo& other);
    void addMissing(int32_t field) { missingFieldMask |= (1<<field); }
    void addExtra(int32_t field) { extraFieldMask |= (1<<field); }
};

class DateTimeMatcher: public UMemory {
public:
    PtnSkeleton skeleton;

    void getBasePattern(UnicodeString& basePattern);
    UnicodeString getPattern();
    void set(const UnicodeString& pattern, FormatParser* fp);
    void set(const UnicodeString& pattern, FormatParser* fp, PtnSkeleton& skeleton);
    void copyFrom(const PtnSkeleton& skeleton);
    void copyFrom();
    PtnSkeleton* getSkeletonPtr();
    UBool equals(const DateTimeMatcher* other) const;
    int32_t getDistance(const DateTimeMatcher& other, int32_t includeMask, DistanceInfo& distanceInfo) const;
    DateTimeMatcher();
    DateTimeMatcher(const DateTimeMatcher& other);
    virtual ~DateTimeMatcher();
    int32_t getFieldMask() const;
};

class PatternMap : public UMemory {
public:
    PtnElem *boot[dtptngen_MAX_PATTERN_ENTRIES];
    PatternMap();
    virtual  ~PatternMap();
    void  add(const UnicodeString& basePattern, const PtnSkeleton& skeleton, const UnicodeString& value, UBool skeletonWasSpecified, UErrorCode& status);
    const UnicodeString* getPatternFromBasePattern(const UnicodeString& basePattern, UBool& skeletonWasSpecified) const;
    const UnicodeString* getPatternFromSkeleton(const PtnSkeleton& skeleton, const PtnSkeleton** specifiedSkeletonPtr = 0) const;
    void copyFrom(const PatternMap& other, UErrorCode& status);
    PtnElem* getHeader(UChar baseChar) const;
    UBool equals(const PatternMap& other) const;
private:
    UBool isDupAllowed;
    PtnElem*  getDuplicateElem(const UnicodeString& basePattern, const PtnSkeleton& skeleton, PtnElem *baseElem);
}; // end  PatternMap

class PatternMapIterator : public UMemory {
public:
    PatternMapIterator(UErrorCode &status);
    virtual ~PatternMapIterator();
    void set(PatternMap& patternMap);
    PtnSkeleton* getSkeleton() const;
    UBool hasNext() const;
    DateTimeMatcher& next();
private:
    int32_t bootIndex;
    PtnElem *nodePtr;
    LocalPointer<DateTimeMatcher> matcher;
    PatternMap *patternMap;
};

class DTSkeletonEnumeration : public StringEnumeration {
public:
    DTSkeletonEnumeration(PatternMap& patternMap, dtStrEnum type, UErrorCode& status);
    virtual ~DTSkeletonEnumeration();
    static UClassID U_EXPORT2 getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
    virtual const UnicodeString* snext(UErrorCode& status);
    virtual void reset(UErrorCode& status);
    virtual int32_t count(UErrorCode& status) const;
private:
    int32_t pos;
    UBool isCanonicalItem(const UnicodeString& item);
    LocalPointer<UVector> fSkeletons;
};

class DTRedundantEnumeration : public StringEnumeration {
public:
    DTRedundantEnumeration();
    virtual ~DTRedundantEnumeration();
    static UClassID U_EXPORT2 getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
    virtual const UnicodeString* snext(UErrorCode& status);
    virtual void reset(UErrorCode& status);
    virtual int32_t count(UErrorCode& status) const;
    void add(const UnicodeString &pattern, UErrorCode& status);
private:
    int32_t pos;
    UBool isCanonicalItem(const UnicodeString& item) const;
    LocalPointer<UVector> fPatterns;
};

U_NAMESPACE_END






#if U_CHARSET_FAMILY==U_EBCDIC_FAMILY
/**
 * If we are on EBCDIC, use an iterator which will
 * traverse the bundles in ASCII order.
 */
#define U_USE_ASCII_BUNDLE_ITERATOR
#define U_SORT_ASCII_BUNDLE_ITERATOR
#endif

#if defined(U_USE_ASCII_BUNDLE_ITERATOR)




struct UResAEntry {
    UChar *key;
    UResourceBundle *item;
};

struct UResourceBundleAIterator {
    UResourceBundle  *bund;
    UResAEntry *entries;
    int32_t num;
    int32_t cursor;
};

/* Must be C linkage to pass function pointer to the sort function */

U_CDECL_BEGIN

static int32_t U_CALLCONV
ures_a_codepointSort(const void *context, const void *left, const void *right) {
    //CompareContext *cmp=(CompareContext *)context;
    return u_strcmp(((const UResAEntry *)left)->key,
                    ((const UResAEntry *)right)->key);
}

U_CDECL_END

static void ures_a_open(UResourceBundleAIterator *aiter, UResourceBundle *bund, UErrorCode *status) {
    if(U_FAILURE(*status)) {
        return;
    }
    aiter->bund = bund;
    aiter->num = ures_getSize(aiter->bund);
    aiter->cursor = 0;
#if !defined(U_SORT_ASCII_BUNDLE_ITERATOR)
    aiter->entries = nullptr;
#else
    aiter->entries = (UResAEntry*)uprv_malloc(sizeof(UResAEntry)*aiter->num);
    for(int i=0;i<aiter->num;i++) {
        aiter->entries[i].item = ures_getByIndex(aiter->bund, i, nullptr, status);
        const char *akey = ures_getKey(aiter->entries[i].item);
        int32_t len = uprv_strlen(akey)+1;
        aiter->entries[i].key = (UChar*)uprv_malloc(len*sizeof(UChar));
        u_charsToUChars(akey, aiter->entries[i].key, len);
    }
    uprv_sortArray(aiter->entries, aiter->num, sizeof(UResAEntry), ures_a_codepointSort, nullptr, TRUE, status);
#endif
}

static void ures_a_close(UResourceBundleAIterator *aiter) {
#if defined(U_SORT_ASCII_BUNDLE_ITERATOR)
    for(int i=0;i<aiter->num;i++) {
        uprv_free(aiter->entries[i].key);
        ures_close(aiter->entries[i].item);
    }
#endif
}

static const UChar *ures_a_getNextString(UResourceBundleAIterator *aiter, int32_t *len, const char **key, UErrorCode *err) {
#if !defined(U_SORT_ASCII_BUNDLE_ITERATOR)
    return ures_getNextString(aiter->bund, len, key, err);
#else
    if(U_FAILURE(*err)) return nullptr;
    UResourceBundle *item = aiter->entries[aiter->cursor].item;
    const UChar* ret = ures_getString(item, len, err);
    *key = ures_getKey(item);
    aiter->cursor++;
    return ret;
#endif
}


#endif


U_NAMESPACE_BEGIN

// *****************************************************************************
// class DateTimePatternGenerator
// *****************************************************************************
static const UChar Canonical_Items[] = {
    // GyQMwWEDFdaHmsSv
    dtptngen_CAP_G, dtptngen_LOW_Y, dtptngen_CAP_Q, dtptngen_CAP_M, dtptngen_LOW_W, dtptngen_CAP_W, dtptngen_CAP_E,
    dtptngen_CAP_D, dtptngen_CAP_F, dtptngen_LOW_D, dtptngen_LOW_A, // The UDATPG_x_FIELD constants and these fields have a different order than in ICU4J
    dtptngen_CAP_H, dtptngen_LOW_M, dtptngen_LOW_S, dtptngen_CAP_S, dtptngen_LOW_V, 0
};

static const dtTypeElem dtTypes[] = {
    // patternChar, field, type, minLen, weight
    {dtptngen_CAP_G, UDATPG_ERA_FIELD, dtptngen_DT_SHORT, 1, 3,},
    {dtptngen_CAP_G, UDATPG_ERA_FIELD, dtptngen_DT_LONG,  4, 0},
    {dtptngen_CAP_G, UDATPG_ERA_FIELD, dtptngen_DT_NARROW, 5, 0},

    {dtptngen_LOW_Y, UDATPG_YEAR_FIELD, dtptngen_DT_NUMERIC, 1, 20},
    {dtptngen_CAP_Y, UDATPG_YEAR_FIELD, dtptngen_DT_NUMERIC + dtptngen_DT_DELTA, 1, 20},
    {dtptngen_LOW_U, UDATPG_YEAR_FIELD, dtptngen_DT_NUMERIC + 2*dtptngen_DT_DELTA, 1, 20},
    {dtptngen_LOW_R, UDATPG_YEAR_FIELD, dtptngen_DT_NUMERIC + 3*dtptngen_DT_DELTA, 1, 20},
    {dtptngen_CAP_U, UDATPG_YEAR_FIELD, dtptngen_DT_SHORT, 1, 3},
    {dtptngen_CAP_U, UDATPG_YEAR_FIELD, dtptngen_DT_LONG, 4, 0},
    {dtptngen_CAP_U, UDATPG_YEAR_FIELD, dtptngen_DT_NARROW, 5, 0},

    {dtptngen_CAP_Q, UDATPG_QUARTER_FIELD, dtptngen_DT_NUMERIC, 1, 2},
    {dtptngen_CAP_Q, UDATPG_QUARTER_FIELD, dtptngen_DT_SHORT, 3, 0},
    {dtptngen_CAP_Q, UDATPG_QUARTER_FIELD, dtptngen_DT_LONG, 4, 0},
    {dtptngen_CAP_Q, UDATPG_QUARTER_FIELD, dtptngen_DT_NARROW, 5, 0},
    {dtptngen_LOW_Q, UDATPG_QUARTER_FIELD, dtptngen_DT_NUMERIC + dtptngen_DT_DELTA, 1, 2},
    {dtptngen_LOW_Q, UDATPG_QUARTER_FIELD, dtptngen_DT_SHORT - dtptngen_DT_DELTA, 3, 0},
    {dtptngen_LOW_Q, UDATPG_QUARTER_FIELD, dtptngen_DT_LONG - dtptngen_DT_DELTA, 4, 0},
    {dtptngen_LOW_Q, UDATPG_QUARTER_FIELD, dtptngen_DT_NARROW - dtptngen_DT_DELTA, 5, 0},

    {dtptngen_CAP_M, UDATPG_MONTH_FIELD, dtptngen_DT_NUMERIC, 1, 2},
    {dtptngen_CAP_M, UDATPG_MONTH_FIELD, dtptngen_DT_SHORT, 3, 0},
    {dtptngen_CAP_M, UDATPG_MONTH_FIELD, dtptngen_DT_LONG, 4, 0},
    {dtptngen_CAP_M, UDATPG_MONTH_FIELD, dtptngen_DT_NARROW, 5, 0},
    {dtptngen_CAP_L, UDATPG_MONTH_FIELD, dtptngen_DT_NUMERIC + dtptngen_DT_DELTA, 1, 2},
    {dtptngen_CAP_L, UDATPG_MONTH_FIELD, dtptngen_DT_SHORT - dtptngen_DT_DELTA, 3, 0},
    {dtptngen_CAP_L, UDATPG_MONTH_FIELD, dtptngen_DT_LONG - dtptngen_DT_DELTA, 4, 0},
    {dtptngen_CAP_L, UDATPG_MONTH_FIELD, dtptngen_DT_NARROW - dtptngen_DT_DELTA, 5, 0},
    {dtptngen_LOW_L, UDATPG_MONTH_FIELD, dtptngen_DT_NUMERIC + dtptngen_DT_DELTA, 1, 1},

    {dtptngen_LOW_W, UDATPG_WEEK_OF_YEAR_FIELD, dtptngen_DT_NUMERIC, 1, 2},

    {dtptngen_CAP_W, UDATPG_WEEK_OF_MONTH_FIELD, dtptngen_DT_NUMERIC, 1, 0},

    {dtptngen_CAP_E, UDATPG_WEEKDAY_FIELD, dtptngen_DT_SHORT, 1, 3},
    {dtptngen_CAP_E, UDATPG_WEEKDAY_FIELD, dtptngen_DT_LONG, 4, 0},
    {dtptngen_CAP_E, UDATPG_WEEKDAY_FIELD, dtptngen_DT_NARROW, 5, 0},
    {dtptngen_CAP_E, UDATPG_WEEKDAY_FIELD, dtptngen_DT_SHORTER, 6, 0},
    {dtptngen_LOW_C, UDATPG_WEEKDAY_FIELD, dtptngen_DT_NUMERIC + 2*dtptngen_DT_DELTA, 1, 2},
    {dtptngen_LOW_C, UDATPG_WEEKDAY_FIELD, dtptngen_DT_SHORT - 2*dtptngen_DT_DELTA, 3, 0},
    {dtptngen_LOW_C, UDATPG_WEEKDAY_FIELD, dtptngen_DT_LONG - 2*dtptngen_DT_DELTA, 4, 0},
    {dtptngen_LOW_C, UDATPG_WEEKDAY_FIELD, dtptngen_DT_NARROW - 2*dtptngen_DT_DELTA, 5, 0},
    {dtptngen_LOW_C, UDATPG_WEEKDAY_FIELD, dtptngen_DT_SHORTER - 2*dtptngen_DT_DELTA, 6, 0},
    {dtptngen_LOW_E, UDATPG_WEEKDAY_FIELD, dtptngen_DT_NUMERIC + dtptngen_DT_DELTA, 1, 2}, // dtptngen_LOW_E is currently not used in CLDR data, should not be canonical
    {dtptngen_LOW_E, UDATPG_WEEKDAY_FIELD, dtptngen_DT_SHORT - dtptngen_DT_DELTA, 3, 0},
    {dtptngen_LOW_E, UDATPG_WEEKDAY_FIELD, dtptngen_DT_LONG - dtptngen_DT_DELTA, 4, 0},
    {dtptngen_LOW_E, UDATPG_WEEKDAY_FIELD, dtptngen_DT_NARROW - dtptngen_DT_DELTA, 5, 0},
    {dtptngen_LOW_E, UDATPG_WEEKDAY_FIELD, dtptngen_DT_SHORTER - dtptngen_DT_DELTA, 6, 0},

    {dtptngen_LOW_D, UDATPG_DAY_FIELD, dtptngen_DT_NUMERIC, 1, 2},
    {dtptngen_LOW_G, UDATPG_DAY_FIELD, dtptngen_DT_NUMERIC + dtptngen_DT_DELTA, 1, 20}, // really internal use, so we don't care

    {dtptngen_CAP_D, UDATPG_DAY_OF_YEAR_FIELD, dtptngen_DT_NUMERIC, 1, 3},

    {dtptngen_CAP_F, UDATPG_DAY_OF_WEEK_IN_MONTH_FIELD, dtptngen_DT_NUMERIC, 1, 0},

    {dtptngen_LOW_A, UDATPG_DAYPERIOD_FIELD, dtptngen_DT_SHORT, 1, 3},
    {dtptngen_LOW_A, UDATPG_DAYPERIOD_FIELD, dtptngen_DT_LONG, 4, 0},
    {dtptngen_LOW_A, UDATPG_DAYPERIOD_FIELD, dtptngen_DT_NARROW, 5, 0},
    {dtptngen_LOW_B, UDATPG_DAYPERIOD_FIELD, dtptngen_DT_SHORT - dtptngen_DT_DELTA, 1, 3},
    {dtptngen_LOW_B, UDATPG_DAYPERIOD_FIELD, dtptngen_DT_LONG - dtptngen_DT_DELTA, 4, 0},
    {dtptngen_LOW_B, UDATPG_DAYPERIOD_FIELD, dtptngen_DT_NARROW - dtptngen_DT_DELTA, 5, 0},
    // b needs to be closer to a than to B, so we make this 3*dtptngen_DT_DELTA
    {dtptngen_CAP_B, UDATPG_DAYPERIOD_FIELD, dtptngen_DT_SHORT - 3*dtptngen_DT_DELTA, 1, 3},
    {dtptngen_CAP_B, UDATPG_DAYPERIOD_FIELD, dtptngen_DT_LONG - 3*dtptngen_DT_DELTA, 4, 0},
    {dtptngen_CAP_B, UDATPG_DAYPERIOD_FIELD, dtptngen_DT_NARROW - 3*dtptngen_DT_DELTA, 5, 0},

    {dtptngen_CAP_H, UDATPG_HOUR_FIELD, dtptngen_DT_NUMERIC + 10*dtptngen_DT_DELTA, 1, 2}, // 24 hour
    {dtptngen_LOW_K, UDATPG_HOUR_FIELD, dtptngen_DT_NUMERIC + 11*dtptngen_DT_DELTA, 1, 2}, // 24 hour
    {dtptngen_LOW_H, UDATPG_HOUR_FIELD, dtptngen_DT_NUMERIC, 1, 2}, // 12 hour
    {dtptngen_CAP_K, UDATPG_HOUR_FIELD, dtptngen_DT_NUMERIC + dtptngen_DT_DELTA, 1, 2}, // 12 hour
    // The C code has had versions of the following 3, keep & update. Should not need these, but...
    // Without these, certain tests using e.g. staticGetSkeleton fail because j/J in patterns
    // get skipped instead of mapped to the right hour chars, for example in
    //   DateFormatTest::TestPatternFromSkeleton
    //   IntlTestDateTimePatternGeneratorAPI:: testStaticGetSkeleton
    //   DateIntervalFormatTest::testTicket11985
    // Need to investigate better handling of jJC replacement e.g. in staticGetSkeleton.
    {dtptngen_CAP_J, UDATPG_HOUR_FIELD, dtptngen_DT_NUMERIC + 5*dtptngen_DT_DELTA, 1, 2}, // 12/24 hour no AM/PM
    {dtptngen_LOW_J, UDATPG_HOUR_FIELD, dtptngen_DT_NUMERIC + 6*dtptngen_DT_DELTA, 1, 6}, // 12/24 hour
    {dtptngen_CAP_C, UDATPG_HOUR_FIELD, dtptngen_DT_NUMERIC + 7*dtptngen_DT_DELTA, 1, 6}, // 12/24 hour with preferred dayPeriods for 12

    {dtptngen_LOW_M, UDATPG_MINUTE_FIELD, dtptngen_DT_NUMERIC, 1, 2},

    {dtptngen_LOW_S, UDATPG_SECOND_FIELD, dtptngen_DT_NUMERIC, 1, 2},
    {dtptngen_CAP_A, UDATPG_SECOND_FIELD, dtptngen_DT_NUMERIC + dtptngen_DT_DELTA, 1, 1000},

    {dtptngen_CAP_S, UDATPG_FRACTIONAL_SECOND_FIELD, dtptngen_DT_NUMERIC, 1, 1000},

    {dtptngen_LOW_V, UDATPG_ZONE_FIELD, dtptngen_DT_SHORT - 2*dtptngen_DT_DELTA, 1, 0},
    {dtptngen_LOW_V, UDATPG_ZONE_FIELD, dtptngen_DT_LONG - 2*dtptngen_DT_DELTA, 4, 0},
    {dtptngen_LOW_Z, UDATPG_ZONE_FIELD, dtptngen_DT_SHORT, 1, 3},
    {dtptngen_LOW_Z, UDATPG_ZONE_FIELD, dtptngen_DT_LONG, 4, 0},
    {dtptngen_CAP_Z, UDATPG_ZONE_FIELD, dtptngen_DT_NARROW - dtptngen_DT_DELTA, 1, 3},
    {dtptngen_CAP_Z, UDATPG_ZONE_FIELD, dtptngen_DT_LONG - dtptngen_DT_DELTA, 4, 0},
    {dtptngen_CAP_Z, UDATPG_ZONE_FIELD, dtptngen_DT_SHORT - dtptngen_DT_DELTA, 5, 0},
    {dtptngen_CAP_O, UDATPG_ZONE_FIELD, dtptngen_DT_SHORT - dtptngen_DT_DELTA, 1, 0},
    {dtptngen_CAP_O, UDATPG_ZONE_FIELD, dtptngen_DT_LONG - dtptngen_DT_DELTA, 4, 0},
    {dtptngen_CAP_V, UDATPG_ZONE_FIELD, dtptngen_DT_SHORT - dtptngen_DT_DELTA, 1, 0},
    {dtptngen_CAP_V, UDATPG_ZONE_FIELD, dtptngen_DT_LONG - dtptngen_DT_DELTA, 2, 0},
    {dtptngen_CAP_V, UDATPG_ZONE_FIELD, dtptngen_DT_LONG-1 - dtptngen_DT_DELTA, 3, 0},
    {dtptngen_CAP_V, UDATPG_ZONE_FIELD, dtptngen_DT_LONG-2 - dtptngen_DT_DELTA, 4, 0},
    {dtptngen_CAP_X, UDATPG_ZONE_FIELD, dtptngen_DT_NARROW - dtptngen_DT_DELTA, 1, 0},
    {dtptngen_CAP_X, UDATPG_ZONE_FIELD, dtptngen_DT_SHORT - dtptngen_DT_DELTA, 2, 0},
    {dtptngen_CAP_X, UDATPG_ZONE_FIELD, dtptngen_DT_LONG - dtptngen_DT_DELTA, 4, 0},
    {dtptngen_LOW_X, UDATPG_ZONE_FIELD, dtptngen_DT_NARROW - dtptngen_DT_DELTA, 1, 0},
    {dtptngen_LOW_X, UDATPG_ZONE_FIELD, dtptngen_DT_SHORT - dtptngen_DT_DELTA, 2, 0},
    {dtptngen_LOW_X, UDATPG_ZONE_FIELD, dtptngen_DT_LONG - dtptngen_DT_DELTA, 4, 0},

    {0, UDATPG_FIELD_COUNT, 0, 0, 0} , // last row of dtTypes[]
 };

static const char* const CLDR_FIELD_APPEND[] = {
    "Era", "Year", "Quarter", "Month", "Week", "*", "Day-Of-Week",
    "*", "*", "Day", "*", // The UDATPG_x_FIELD constants and these fields have a different order than in ICU4J
    "Hour", "Minute", "Second", "*", "Timezone"
};

static const char* const CLDR_FIELD_NAME[UDATPG_FIELD_COUNT] = {
    "era", "year", "quarter", "month", "week", "weekOfMonth", "weekday",
    "dayOfYear", "weekdayOfMonth", "day", "dayperiod", // The UDATPG_x_FIELD constants and these fields have a different order than in ICU4J
    "hour", "minute", "second", "*", "zone"
};

static const char* const CLDR_FIELD_WIDTH[] = { // [UDATPG_WIDTH_COUNT]
    "", "-short", "-narrow"
};

// TODO(ticket:13619): remove when definition uncommented in dtptngen.h.
static const int32_t UDATPG_WIDTH_COUNT = UDATPG_NARROW + 1;
static constexpr UDateTimePGDisplayWidth UDATPG_WIDTH_APPENDITEM = UDATPG_WIDE;
static constexpr int32_t UDATPG_FIELD_KEY_MAX = 24; // max length of CLDR field tag (type + width)

// For appendItems
static const UChar UDATPG_ItemFormat[]= {0x7B, 0x30, 0x7D, 0x20, 0x251C, 0x7B, 0x32, 0x7D, 0x3A,
    0x20, 0x7B, 0x31, 0x7D, 0x2524, 0};  // {0} \u251C{2}: {1}\u2524

//static const UChar repeatedPatterns[6]={dtptngen_CAP_G, dtptngen_CAP_E, dtptngen_LOW_Z, dtptngen_LOW_V, dtptngen_CAP_Q, 0}; // "GEzvQ"

static const char DT_DateTimePatternsTag[]="DateTimePatterns";
static const char DT_DateTimeCalendarTag[]="calendar";
static const char DT_DateTimeGregorianTag[]="gregorian";
static const char DT_DateTimeAppendItemsTag[]="appendItems";
static const char DT_DateTimeFieldsTag[]="fields";
static const char DT_DateTimeAvailableFormatsTag[]="availableFormats";
//static const UnicodeString repeatedPattern=UnicodeString(repeatedPatterns);

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DateTimePatternGenerator)
UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DTSkeletonEnumeration)
UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DTRedundantEnumeration)

DateTimePatternGenerator*  U_EXPORT2
DateTimePatternGenerator::createInstance(UErrorCode& status) {
    return createInstance(Locale::getDefault(), status);
}

DateTimePatternGenerator* U_EXPORT2
DateTimePatternGenerator::createInstance(const Locale& locale, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    LocalPointer<DateTimePatternGenerator> result(
            new DateTimePatternGenerator(locale, status), status);
    return U_SUCCESS(status) ? result.orphan() : nullptr;
}

DateTimePatternGenerator*  U_EXPORT2
DateTimePatternGenerator::createEmptyInstance(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    LocalPointer<DateTimePatternGenerator> result(
            new DateTimePatternGenerator(status), status);
    return U_SUCCESS(status) ? result.orphan() : nullptr;
}

DateTimePatternGenerator::DateTimePatternGenerator(UErrorCode &status) :
    skipMatcher(nullptr),
    fAvailableFormatKeyHash(nullptr),
    internalErrorCode(U_ZERO_ERROR)
{
    fp = new FormatParser();
    dtMatcher = new DateTimeMatcher();
    distanceInfo = new DistanceInfo();
    patternMap = new PatternMap();
    if (fp == nullptr || dtMatcher == nullptr || distanceInfo == nullptr || patternMap == nullptr) {
        internalErrorCode = status = U_MEMORY_ALLOCATION_ERROR;
    }
}

DateTimePatternGenerator::DateTimePatternGenerator(const Locale& locale, UErrorCode &status) :
    skipMatcher(nullptr),
    fAvailableFormatKeyHash(nullptr),
    internalErrorCode(U_ZERO_ERROR)
{
    fp = new FormatParser();
    dtMatcher = new DateTimeMatcher();
    distanceInfo = new DistanceInfo();
    patternMap = new PatternMap();
    if (fp == nullptr || dtMatcher == nullptr || distanceInfo == nullptr || patternMap == nullptr) {
        internalErrorCode = status = U_MEMORY_ALLOCATION_ERROR;
    }
    else {
        initData(locale, status);
    }
}

DateTimePatternGenerator::DateTimePatternGenerator(const DateTimePatternGenerator& other) :
    UObject(),
    skipMatcher(nullptr),
    fAvailableFormatKeyHash(nullptr),
    internalErrorCode(U_ZERO_ERROR)
{
    fp = new FormatParser();
    dtMatcher = new DateTimeMatcher();
    distanceInfo = new DistanceInfo();
    patternMap = new PatternMap();
    if (fp == nullptr || dtMatcher == nullptr || distanceInfo == nullptr || patternMap == nullptr) {
        internalErrorCode = U_MEMORY_ALLOCATION_ERROR;
    }
    *this=other;
}

DateTimePatternGenerator&
DateTimePatternGenerator::operator=(const DateTimePatternGenerator& other) {
    // reflexive case
    if (&other == this) {
        return *this;
    }
    internalErrorCode = other.internalErrorCode;
    pLocale = other.pLocale;
    fDefaultHourFormatChar = other.fDefaultHourFormatChar;
    *fp = *(other.fp);
    dtMatcher->copyFrom(other.dtMatcher->skeleton);
    *distanceInfo = *(other.distanceInfo);
    dateTimeFormat = other.dateTimeFormat;
    decimal = other.decimal;
    // NUL-terminate for the C API.
    dateTimeFormat.getTerminatedBuffer();
    decimal.getTerminatedBuffer();
    delete skipMatcher;
    if ( other.skipMatcher == nullptr ) {
        skipMatcher = nullptr;
    }
    else {
        skipMatcher = new DateTimeMatcher(*other.skipMatcher);
        if (skipMatcher == nullptr)
        {
            internalErrorCode = U_MEMORY_ALLOCATION_ERROR;
            return *this;
        }
    }
    for (int32_t i=0; i< UDATPG_FIELD_COUNT; ++i ) {
        appendItemFormats[i] = other.appendItemFormats[i];
        appendItemFormats[i].getTerminatedBuffer(); // NUL-terminate for the C API.
        for (int32_t j=0; j< UDATPG_WIDTH_COUNT; ++j ) {
            fieldDisplayNames[i][j] = other.fieldDisplayNames[i][j];
            fieldDisplayNames[i][j].getTerminatedBuffer(); // NUL-terminate for the C API.
        }
    }
    patternMap->copyFrom(*other.patternMap, internalErrorCode);
    copyHashtable(other.fAvailableFormatKeyHash, internalErrorCode);
    return *this;
}


UBool
DateTimePatternGenerator::operator==(const DateTimePatternGenerator& other) const {
    if (this == &other) {
        return TRUE;
    }
    if ((pLocale==other.pLocale) && (patternMap->equals(*other.patternMap)) &&
        (dateTimeFormat==other.dateTimeFormat) && (decimal==other.decimal)) {
        for ( int32_t i=0 ; i<UDATPG_FIELD_COUNT; ++i ) {
            if (appendItemFormats[i] != other.appendItemFormats[i]) {
                return FALSE;
            }
            for (int32_t j=0; j< UDATPG_WIDTH_COUNT; ++j ) {
                if (fieldDisplayNames[i][j] != other.fieldDisplayNames[i][j]) {
                    return FALSE;
                }
            }
        }
        return TRUE;
    }
    else {
        return FALSE;
    }
}

UBool
DateTimePatternGenerator::operator!=(const DateTimePatternGenerator& other) const {
    return  !operator==(other);
}

DateTimePatternGenerator::~DateTimePatternGenerator() {
    if (fAvailableFormatKeyHash!=nullptr) {
        delete fAvailableFormatKeyHash;
    }

    if (fp != nullptr) delete fp;
    if (dtMatcher != nullptr) delete dtMatcher;
    if (distanceInfo != nullptr) delete distanceInfo;
    if (patternMap != nullptr) delete patternMap;
    if (skipMatcher != nullptr) delete skipMatcher;
}

namespace {

UInitOnce dtptngen_initOnce = U_INITONCE_INITIALIZER;
UHashtable *localeToAllowedHourFormatsMap = nullptr;

// Value deleter for hashmap.
U_CFUNC void U_CALLCONV deleteAllowedHourFormats(void *ptr) {
    uprv_free(ptr);
}

// Close hashmap at cleanup.
U_CFUNC UBool U_CALLCONV allowedHourFormatsCleanup() {
    uhash_close(localeToAllowedHourFormatsMap);
    return TRUE;
}

enum AllowedHourFormat{
    ALLOWED_HOUR_FORMAT_UNKNOWN = -1,
    ALLOWED_HOUR_FORMAT_h,
    ALLOWED_HOUR_FORMAT_H,
    ALLOWED_HOUR_FORMAT_K,  // Added ICU-20383, used by JP
    ALLOWED_HOUR_FORMAT_k,  // Added ICU-20383, not currently used
    ALLOWED_HOUR_FORMAT_hb,
    ALLOWED_HOUR_FORMAT_hB,
    ALLOWED_HOUR_FORMAT_Kb, // Added ICU-20383, not currently used
    ALLOWED_HOUR_FORMAT_KB, // Added ICU-20383, not currently used
    // ICU-20383 The following are unlikely and not currently used
    ALLOWED_HOUR_FORMAT_Hb,
    ALLOWED_HOUR_FORMAT_HB
};

}  // namespace

void
DateTimePatternGenerator::initData(const Locale& locale, UErrorCode &status) {
    //const char *baseLangName = locale.getBaseName(); // unused

    skipMatcher = nullptr;
    fAvailableFormatKeyHash=nullptr;
    addCanonicalItems(status);
    addICUPatterns(locale, status);
    addCLDRData(locale, status);
    setDateTimeFromCalendar(locale, status);
    setDecimalSymbols(locale, status);
    umtx_initOnce(dtptngen_initOnce, loadAllowedHourFormatsData, status);
    getAllowedHourFormats(locale, status);
    // If any of the above methods failed then the object is in an invalid state.
    internalErrorCode = status;
} // DateTimePatternGenerator::initData

namespace {

struct AllowedHourFormatsSink : public ResourceSink {
    // Initialize sub-sinks.
    AllowedHourFormatsSink() {}
    virtual ~AllowedHourFormatsSink();

    virtual void put(const char *key, ResourceValue &value, UBool /*noFallback*/,
                     UErrorCode &errorCode) {
        ResourceTable timeData = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }
        for (int32_t i = 0; timeData.getKeyAndValue(i, key, value); ++i) {
            const char *regionOrLocale = key;
            ResourceTable formatList = value.getTable(errorCode);
            if (U_FAILURE(errorCode)) { return; }
            // below we construct a list[] that has an entry for the "preferred" value at [0],
            // followed by 1 or more entries for the "allowed" values, terminated with an
            // entry for ALLOWED_HOUR_FORMAT_UNKNOWN (not included in length below)
            LocalMemory<int32_t> list;
            int32_t length = 0;
            int32_t preferredFormat = ALLOWED_HOUR_FORMAT_UNKNOWN;
            for (int32_t j = 0; formatList.getKeyAndValue(j, key, value); ++j) {
                if (uprv_strcmp(key, "allowed") == 0) {
                    if (value.getType() == URES_STRING) {
                        length = 2; // 1 preferred to add later, 1 allowed to add now
                        if (list.allocateInsteadAndReset(length + 1) == nullptr) {
                            errorCode = U_MEMORY_ALLOCATION_ERROR;
                            return;
                        }
                        list[1] = getHourFormatFromUnicodeString(value.getUnicodeString(errorCode));
                    }
                    else {
                        ResourceArray allowedFormats = value.getArray(errorCode);
                        length = allowedFormats.getSize() + 1; // 1 preferred, getSize allowed
                        if (list.allocateInsteadAndReset(length + 1) == nullptr) {
                            errorCode = U_MEMORY_ALLOCATION_ERROR;
                            return;
                        }
                        for (int32_t k = 1; k < length; ++k) {
                            allowedFormats.getValue(k-1, value);
                            list[k] = getHourFormatFromUnicodeString(value.getUnicodeString(errorCode));
                        }
                    }
                } else if (uprv_strcmp(key, "preferred") == 0) {
                    preferredFormat = getHourFormatFromUnicodeString(value.getUnicodeString(errorCode));
                }
            }
            if (length > 1) {
                list[0] = (preferredFormat!=ALLOWED_HOUR_FORMAT_UNKNOWN)? preferredFormat: list[1];
            } else {
                // fallback handling for missing data
                length = 2; // 1 preferred, 1 allowed
                if (list.allocateInsteadAndReset(length + 1) == nullptr) {
                    errorCode = U_MEMORY_ALLOCATION_ERROR;
                    return;
                }
                list[0] = (preferredFormat!=ALLOWED_HOUR_FORMAT_UNKNOWN)? preferredFormat: ALLOWED_HOUR_FORMAT_H;
                list[1] = list[0];
            }
            list[length] = ALLOWED_HOUR_FORMAT_UNKNOWN;
            // At this point list[] will have at least two non-ALLOWED_HOUR_FORMAT_UNKNOWN entries,
            // followed by ALLOWED_HOUR_FORMAT_UNKNOWN.
            uhash_put(localeToAllowedHourFormatsMap, const_cast<char *>(regionOrLocale), list.orphan(), &errorCode);
            if (U_FAILURE(errorCode)) { return; }
        }
    }

    AllowedHourFormat getHourFormatFromUnicodeString(const UnicodeString &s) {
        if (s.length() == 1) {
            if (s[0] == dtptngen_LOW_H) { return ALLOWED_HOUR_FORMAT_h; }
            if (s[0] == dtptngen_CAP_H) { return ALLOWED_HOUR_FORMAT_H; }
            if (s[0] == dtptngen_CAP_K) { return ALLOWED_HOUR_FORMAT_K; }
            if (s[0] == dtptngen_LOW_K) { return ALLOWED_HOUR_FORMAT_k; }
        } else if (s.length() == 2) {
            if (s[0] == dtptngen_LOW_H && s[1] == dtptngen_LOW_B) { return ALLOWED_HOUR_FORMAT_hb; }
            if (s[0] == dtptngen_LOW_H && s[1] == dtptngen_CAP_B) { return ALLOWED_HOUR_FORMAT_hB; }
            if (s[0] == dtptngen_CAP_K && s[1] == dtptngen_LOW_B) { return ALLOWED_HOUR_FORMAT_Kb; }
            if (s[0] == dtptngen_CAP_K && s[1] == dtptngen_CAP_B) { return ALLOWED_HOUR_FORMAT_KB; }
            if (s[0] == dtptngen_CAP_H && s[1] == dtptngen_LOW_B) { return ALLOWED_HOUR_FORMAT_Hb; }
            if (s[0] == dtptngen_CAP_H && s[1] == dtptngen_CAP_B) { return ALLOWED_HOUR_FORMAT_HB; }
        }

        return ALLOWED_HOUR_FORMAT_UNKNOWN;
    }
};

}  // namespace

AllowedHourFormatsSink::~AllowedHourFormatsSink() {}

U_CFUNC void U_CALLCONV DateTimePatternGenerator::loadAllowedHourFormatsData(UErrorCode &status) {
    if (U_FAILURE(status)) { return; }
    localeToAllowedHourFormatsMap = uhash_open(
        uhash_hashChars, uhash_compareChars, nullptr, &status);
    if (U_FAILURE(status)) { return; }

    uhash_setValueDeleter(localeToAllowedHourFormatsMap, deleteAllowedHourFormats);
    ucln_i18n_registerCleanup(UCLN_I18N_ALLOWED_HOUR_FORMATS, allowedHourFormatsCleanup);

    LocalUResourceBundlePointer rb(ures_openDirect(nullptr, "supplementalData", &status));
    if (U_FAILURE(status)) { return; }

    AllowedHourFormatsSink sink;
    // TODO: Currently in the enumeration each table allocates a new array.
    // Try to reduce the number of memory allocations. Consider storing a
    // UVector32 with the concatenation of all of the sub-arrays, put the start index
    // into the hashmap, store 6 single-value sub-arrays right at the beginning of the
    // vector (at index enum*2) for easy data sharing, copy sub-arrays into runtime
    // object. Remember to clean up the vector, too.
    ures_getAllItemsWithFallback(rb.getAlias(), "timeData", sink, status);
}

static int32_t* getAllowedHourFormatsLangCountry(const char* language, const char* country, UErrorCode& status) {
    CharString langCountry;
    langCountry.append(language, status);
    langCountry.append('_', status);
    langCountry.append(country, status);

    int32_t* allowedFormats;
    allowedFormats = (int32_t *)uhash_get(localeToAllowedHourFormatsMap, langCountry.data());
    if (allowedFormats == nullptr) {
        allowedFormats = (int32_t *)uhash_get(localeToAllowedHourFormatsMap, const_cast<char *>(country));
    }

    return allowedFormats;
}

void DateTimePatternGenerator::getAllowedHourFormats(const Locale &locale, UErrorCode &status) {
    if (U_FAILURE(status)) { return; }

    const char *language = locale.getLanguage();
    const char *country = locale.getCountry();
    Locale maxLocale;  // must be here for correct lifetime
    if (*language == '\0' || *country == '\0') {
        maxLocale = locale;
        UErrorCode localStatus = U_ZERO_ERROR;
        maxLocale.addLikelySubtags(localStatus);
        if (U_SUCCESS(localStatus)) {
            language = maxLocale.getLanguage();
            country = maxLocale.getCountry();
        }
    }
    if (*language == '\0') {
        // Unexpected, but fail gracefully
        language = "und";
    }
    if (*country == '\0') {
        country = "001";
    }

    int32_t* allowedFormats = getAllowedHourFormatsLangCountry(language, country, status);

    // Check if the region has an alias
    if (allowedFormats == nullptr) {
        UErrorCode localStatus = U_ZERO_ERROR;
        const Region* region = Region::getInstance(country, localStatus);
        if (U_SUCCESS(localStatus)) {
            country = region->getRegionCode(); // the real region code
            allowedFormats = getAllowedHourFormatsLangCountry(language, country, status);
        }
    }

    if (allowedFormats != nullptr) {  // Lookup is successful
        // Here allowedFormats points to a list consisting of key for preferredFormat,
        // followed by one or more keys for allowedFormats, then followed by ALLOWED_HOUR_FORMAT_UNKNOWN.
        switch (allowedFormats[0]) {
            case ALLOWED_HOUR_FORMAT_h: fDefaultHourFormatChar = dtptngen_LOW_H; break;
            case ALLOWED_HOUR_FORMAT_H: fDefaultHourFormatChar = dtptngen_CAP_H; break;
            case ALLOWED_HOUR_FORMAT_K: fDefaultHourFormatChar = dtptngen_CAP_K; break;
            case ALLOWED_HOUR_FORMAT_k: fDefaultHourFormatChar = dtptngen_LOW_K; break;
            default: fDefaultHourFormatChar = dtptngen_CAP_H; break;
        }
        for (int32_t i = 0; i < UPRV_LENGTHOF(fAllowedHourFormats); ++i) {
            fAllowedHourFormats[i] = allowedFormats[i + 1];
            if (fAllowedHourFormats[i] == ALLOWED_HOUR_FORMAT_UNKNOWN) {
                break;
            }
        }
    } else {  // Lookup failed, twice
        fDefaultHourFormatChar = dtptngen_CAP_H;
        fAllowedHourFormats[0] = ALLOWED_HOUR_FORMAT_H;
        fAllowedHourFormats[1] = ALLOWED_HOUR_FORMAT_UNKNOWN;
    }
}

UnicodeString
DateTimePatternGenerator::getSkeleton(const UnicodeString& pattern, UErrorCode&
/*status*/) {
    FormatParser fp2;
    DateTimeMatcher matcher;
    PtnSkeleton localSkeleton;
    matcher.set(pattern, &fp2, localSkeleton);
    return localSkeleton.getSkeleton();
}

UnicodeString
DateTimePatternGenerator::staticGetSkeleton(
        const UnicodeString& pattern, UErrorCode& /*status*/) {
    FormatParser fp;
    DateTimeMatcher matcher;
    PtnSkeleton localSkeleton;
    matcher.set(pattern, &fp, localSkeleton);
    return localSkeleton.getSkeleton();
}

UnicodeString
DateTimePatternGenerator::getBaseSkeleton(const UnicodeString& pattern, UErrorCode& /*status*/) {
    FormatParser fp2;
    DateTimeMatcher matcher;
    PtnSkeleton localSkeleton;
    matcher.set(pattern, &fp2, localSkeleton);
    return localSkeleton.getBaseSkeleton();
}

UnicodeString
DateTimePatternGenerator::staticGetBaseSkeleton(
        const UnicodeString& pattern, UErrorCode& /*status*/) {
    FormatParser fp;
    DateTimeMatcher matcher;
    PtnSkeleton localSkeleton;
    matcher.set(pattern, &fp, localSkeleton);
    return localSkeleton.getBaseSkeleton();
}

void
DateTimePatternGenerator::addICUPatterns(const Locale& locale, UErrorCode& status) {
    if (U_FAILURE(status)) { return; }
    UnicodeString dfPattern;
    UnicodeString conflictingString;
    DateFormat* df;

    // Load with ICU patterns
    for (int32_t i=DateFormat::kFull; i<=DateFormat::kShort; i++) {
        DateFormat::EStyle style = (DateFormat::EStyle)i;
        df = DateFormat::createDateInstance(style, locale);
        SimpleDateFormat* sdf;
        if (df != nullptr && (sdf = dynamic_cast<SimpleDateFormat*>(df)) != nullptr) {
            sdf->toPattern(dfPattern);
            addPattern(dfPattern, FALSE, conflictingString, status);
        }
        // TODO Maybe we should return an error when the date format isn't simple.
        delete df;
        if (U_FAILURE(status)) { return; }

        df = DateFormat::createTimeInstance(style, locale);
        if (df != nullptr && (sdf = dynamic_cast<SimpleDateFormat*>(df)) != nullptr) {
            sdf->toPattern(dfPattern);
            addPattern(dfPattern, FALSE, conflictingString, status);

            // TODO: C++ and Java are inconsistent (see #12568).
            // C++ uses MEDIUM, but Java uses SHORT.
            if ( i==DateFormat::kShort && !dfPattern.isEmpty() ) {
                consumeShortTimePattern(dfPattern, status);
            }
        }
        // TODO Maybe we should return an error when the date format isn't simple.
        delete df;
        if (U_FAILURE(status)) { return; }
    }
}

void
DateTimePatternGenerator::hackTimes(const UnicodeString& hackPattern, UErrorCode& status)  {
    UnicodeString conflictingString;

    fp->set(hackPattern);
    UnicodeString mmss;
    UBool gotMm=FALSE;
    for (int32_t i=0; i<fp->itemNumber; ++i) {
        UnicodeString field = fp->items[i];
        if ( fp->isQuoteLiteral(field) ) {
            if ( gotMm ) {
               UnicodeString quoteLiteral;
               fp->getQuoteLiteral(quoteLiteral, &i);
               mmss += quoteLiteral;
            }
        }
        else {
            if (fp->isPatternSeparator(field) && gotMm) {
                mmss+=field;
            }
            else {
                UChar ch=field.charAt(0);
                if (ch==dtptngen_LOW_M) {
                    gotMm=TRUE;
                    mmss+=field;
                }
                else {
                    if (ch==dtptngen_LOW_S) {
                        if (!gotMm) {
                            break;
                        }
                        mmss+= field;
                        addPattern(mmss, FALSE, conflictingString, status);
                        break;
                    }
                    else {
                        if (gotMm || ch==dtptngen_LOW_Z || ch==dtptngen_CAP_Z || ch==dtptngen_LOW_V || ch==dtptngen_CAP_V) {
                            break;
                        }
                    }
                }
            }
        }
    }
}

#define ULOC_LOCALE_IDENTIFIER_CAPACITY (ULOC_FULLNAME_CAPACITY + 1 + ULOC_KEYWORD_AND_VALUES_CAPACITY)

void
DateTimePatternGenerator::getCalendarTypeToUse(const Locale& locale, CharString& destination, UErrorCode& err) {
    destination.clear().append(DT_DateTimeGregorianTag, -1, err); // initial default
    if ( U_SUCCESS(err) ) {
        UErrorCode localStatus = U_ZERO_ERROR;
        char localeWithCalendarKey[ULOC_LOCALE_IDENTIFIER_CAPACITY];
        // obtain a locale that always has the calendar key value that should be used
        ures_getFunctionalEquivalent(
            localeWithCalendarKey,
            ULOC_LOCALE_IDENTIFIER_CAPACITY,
            nullptr,
            "calendar",
            "calendar",
            locale.getName(),
            nullptr,
            FALSE,
            &localStatus);
        localeWithCalendarKey[ULOC_LOCALE_IDENTIFIER_CAPACITY-1] = 0; // ensure null termination
        // now get the calendar key value from that locale
        char calendarType[ULOC_KEYWORDS_CAPACITY];
        int32_t calendarTypeLen = uloc_getKeywordValue(
            localeWithCalendarKey,
            "calendar",
            calendarType,
            ULOC_KEYWORDS_CAPACITY,
            &localStatus);
        // If the input locale was invalid, don't fail with missing resource error, instead
        // continue with default of Gregorian.
        if (U_FAILURE(localStatus) && localStatus != U_MISSING_RESOURCE_ERROR) {
            err = localStatus;
            return;
        }
        if (calendarTypeLen < ULOC_KEYWORDS_CAPACITY) {
            destination.clear().append(calendarType, -1, err);
            if (U_FAILURE(err)) { return; }
        }
    }
}

void
DateTimePatternGenerator::consumeShortTimePattern(const UnicodeString& shortTimePattern,
        UErrorCode& status) {
    if (U_FAILURE(status)) { return; }
    // ICU-20383 No longer set fDefaultHourFormatChar to the hour format character from
    // this pattern; instead it is set from localeToAllowedHourFormatsMap which now
    // includes entries for both preferred and allowed formats.

    // HACK for hh:ss
    hackTimes(shortTimePattern, status);
}

struct DateTimePatternGenerator::AppendItemFormatsSink : public ResourceSink {

    // Destination for data, modified via setters.
    DateTimePatternGenerator& dtpg;

    AppendItemFormatsSink(DateTimePatternGenerator& _dtpg) : dtpg(_dtpg) {}
    virtual ~AppendItemFormatsSink();

    virtual void put(const char *key, ResourceValue &value, UBool /*noFallback*/,
            UErrorCode &errorCode) {
        ResourceTable itemsTable = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }
        for (int32_t i = 0; itemsTable.getKeyAndValue(i, key, value); ++i) {
            UDateTimePatternField field = dtpg.getAppendFormatNumber(key);
            if (field == UDATPG_FIELD_COUNT) { continue; }
            const UnicodeString& valueStr = value.getUnicodeString(errorCode);
            if (dtpg.getAppendItemFormat(field).isEmpty() && !valueStr.isEmpty()) {
                dtpg.setAppendItemFormat(field, valueStr);
            }
        }
    }

    void fillInMissing() {
        UnicodeString defaultItemFormat(TRUE, UDATPG_ItemFormat, UPRV_LENGTHOF(UDATPG_ItemFormat)-1);  // Read-only alias.
        for (int32_t i = 0; i < UDATPG_FIELD_COUNT; i++) {
            UDateTimePatternField field = (UDateTimePatternField)i;
            if (dtpg.getAppendItemFormat(field).isEmpty()) {
                dtpg.setAppendItemFormat(field, defaultItemFormat);
            }
        }
    }
};

struct DateTimePatternGenerator::AppendItemNamesSink : public ResourceSink {

    // Destination for data, modified via setters.
    DateTimePatternGenerator& dtpg;

    AppendItemNamesSink(DateTimePatternGenerator& _dtpg) : dtpg(_dtpg) {}
    virtual ~AppendItemNamesSink();

    virtual void put(const char *key, ResourceValue &value, UBool /*noFallback*/,
            UErrorCode &errorCode) {
        ResourceTable itemsTable = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }
        for (int32_t i = 0; itemsTable.getKeyAndValue(i, key, value); ++i) {
            UDateTimePGDisplayWidth width;
            UDateTimePatternField field = dtpg.getFieldAndWidthIndices(key, &width);
            if (field == UDATPG_FIELD_COUNT) { continue; }
            ResourceTable detailsTable = value.getTable(errorCode);
            if (U_FAILURE(errorCode)) { return; }
            for (int32_t j = 0; detailsTable.getKeyAndValue(j, key, value); ++j) {
                if (uprv_strcmp(key, "dn") != 0) { continue; }
                const UnicodeString& valueStr = value.getUnicodeString(errorCode);
                if (dtpg.getFieldDisplayName(field,width).isEmpty() && !valueStr.isEmpty()) {
                    dtpg.setFieldDisplayName(field,width,valueStr);
                }
                break;
            }
        }
    }

    void fillInMissing() {
        for (int32_t i = 0; i < UDATPG_FIELD_COUNT; i++) {
            UnicodeString& valueStr = dtpg.getMutableFieldDisplayName((UDateTimePatternField)i, UDATPG_WIDE);
            if (valueStr.isEmpty()) {
                valueStr = dtptngen_CAP_F;
                U_ASSERT(i < 20);
                if (i < 10) {
                    // F0, F1, ..., F9
                    valueStr += (UChar)(i+0x30);
                } else {
                    // F10, F11, ...
                    valueStr += (UChar)0x31;
                    valueStr += (UChar)(i-10 + 0x30);
                }
                // NUL-terminate for the C API.
                valueStr.getTerminatedBuffer();
            }
            for (int32_t j = 1; j < UDATPG_WIDTH_COUNT; j++) {
                UnicodeString& valueStr2 = dtpg.getMutableFieldDisplayName((UDateTimePatternField)i, (UDateTimePGDisplayWidth)j);
                if (valueStr2.isEmpty()) {
                    valueStr2 = dtpg.getFieldDisplayName((UDateTimePatternField)i, (UDateTimePGDisplayWidth)(j-1));
                }
            }
        }
    }
};

struct DateTimePatternGenerator::AvailableFormatsSink : public ResourceSink {

    // Destination for data, modified via setters.
    DateTimePatternGenerator& dtpg;

    // Temporary variable, required for calling addPatternWithSkeleton.
    UnicodeString conflictingPattern;

    AvailableFormatsSink(DateTimePatternGenerator& _dtpg) : dtpg(_dtpg) {}
    virtual ~AvailableFormatsSink();

    virtual void put(const char *key, ResourceValue &value, UBool isRoot,
            UErrorCode &errorCode) {
        ResourceTable itemsTable = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }
        for (int32_t i = 0; itemsTable.getKeyAndValue(i, key, value); ++i) {
            const UnicodeString formatKey(key, -1, US_INV);
            if (!dtpg.isAvailableFormatSet(formatKey) ) {
                dtpg.setAvailableFormat(formatKey, errorCode);
                // Add pattern with its associated skeleton. Override any duplicate
                // derived from std patterns, but not a previous availableFormats entry:
                const UnicodeString& formatValue = value.getUnicodeString(errorCode);
                conflictingPattern.remove();
                dtpg.addPatternWithSkeleton(formatValue, &formatKey, !isRoot, conflictingPattern, errorCode);
            }
        }
    }
};

// Virtual destructors must be defined out of line.
DateTimePatternGenerator::AppendItemFormatsSink::~AppendItemFormatsSink() {}
DateTimePatternGenerator::AppendItemNamesSink::~AppendItemNamesSink() {}
DateTimePatternGenerator::AvailableFormatsSink::~AvailableFormatsSink() {}

void
DateTimePatternGenerator::addCLDRData(const Locale& locale, UErrorCode& errorCode) {
    if (U_FAILURE(errorCode)) { return; }
    UnicodeString rbPattern, value, field;
    CharString path;

    LocalUResourceBundlePointer rb(ures_open(nullptr, locale.getName(), &errorCode));
    if (U_FAILURE(errorCode)) { return; }

    CharString calendarTypeToUse; // to be filled in with the type to use, if all goes well
    getCalendarTypeToUse(locale, calendarTypeToUse, errorCode);
    if (U_FAILURE(errorCode)) { return; }

    // Local err to ignore resource not found exceptions
    UErrorCode err = U_ZERO_ERROR;

    // Load append item formats.
    AppendItemFormatsSink appendItemFormatsSink(*this);
    path.clear()
        .append(DT_DateTimeCalendarTag, errorCode)
        .append('/', errorCode)
        .append(calendarTypeToUse, errorCode)
        .append('/', errorCode)
        .append(DT_DateTimeAppendItemsTag, errorCode); // i.e., calendar/xxx/appendItems
    if (U_FAILURE(errorCode)) { return; }
    ures_getAllItemsWithFallback(rb.getAlias(), path.data(), appendItemFormatsSink, err);
    appendItemFormatsSink.fillInMissing();

    // Load CLDR item names.
    err = U_ZERO_ERROR;
    AppendItemNamesSink appendItemNamesSink(*this);
    ures_getAllItemsWithFallback(rb.getAlias(), DT_DateTimeFieldsTag, appendItemNamesSink, err);
    appendItemNamesSink.fillInMissing();

    // Load the available formats from CLDR.
    err = U_ZERO_ERROR;
    initHashtable(errorCode);
    if (U_FAILURE(errorCode)) { return; }
    AvailableFormatsSink availableFormatsSink(*this);
    path.clear()
        .append(DT_DateTimeCalendarTag, errorCode)
        .append('/', errorCode)
        .append(calendarTypeToUse, errorCode)
        .append('/', errorCode)
        .append(DT_DateTimeAvailableFormatsTag, errorCode); // i.e., calendar/xxx/availableFormats
    if (U_FAILURE(errorCode)) { return; }
    ures_getAllItemsWithFallback(rb.getAlias(), path.data(), availableFormatsSink, err);
}

void
DateTimePatternGenerator::initHashtable(UErrorCode& err) {
    if (U_FAILURE(err)) { return; }
    if (fAvailableFormatKeyHash!=nullptr) {
        return;
    }
    LocalPointer<Hashtable> hash(new Hashtable(FALSE, err), err);
    if (U_SUCCESS(err)) {
        fAvailableFormatKeyHash = hash.orphan();
    }
}

void
DateTimePatternGenerator::setAppendItemFormat(UDateTimePatternField field, const UnicodeString& value) {
    appendItemFormats[field] = value;
    // NUL-terminate for the C API.
    appendItemFormats[field].getTerminatedBuffer();
}

const UnicodeString&
DateTimePatternGenerator::getAppendItemFormat(UDateTimePatternField field) const {
    return appendItemFormats[field];
}

void
DateTimePatternGenerator::setAppendItemName(UDateTimePatternField field, const UnicodeString& value) {
    setFieldDisplayName(field, UDATPG_WIDTH_APPENDITEM, value);
}

const UnicodeString&
DateTimePatternGenerator::getAppendItemName(UDateTimePatternField field) const {
    return fieldDisplayNames[field][UDATPG_WIDTH_APPENDITEM];
}

void
DateTimePatternGenerator::setFieldDisplayName(UDateTimePatternField field, UDateTimePGDisplayWidth width, const UnicodeString& value) {
    fieldDisplayNames[field][width] = value;
    // NUL-terminate for the C API.
    fieldDisplayNames[field][width].getTerminatedBuffer();
}

UnicodeString
DateTimePatternGenerator::getFieldDisplayName(UDateTimePatternField field, UDateTimePGDisplayWidth width) const {
    return fieldDisplayNames[field][width];
}

UnicodeString&
DateTimePatternGenerator::getMutableFieldDisplayName(UDateTimePatternField field, UDateTimePGDisplayWidth width) {
    return fieldDisplayNames[field][width];
}

void
DateTimePatternGenerator::getAppendName(UDateTimePatternField field, UnicodeString& value) {
    value = dtptngen_SINGLE_QUOTE;
    value += fieldDisplayNames[field][UDATPG_WIDTH_APPENDITEM];
    value += dtptngen_SINGLE_QUOTE;
}

UnicodeString
DateTimePatternGenerator::getBestPattern(const UnicodeString& patternForm, UErrorCode& status) {
    return getBestPattern(patternForm, UDATPG_MATCH_NO_OPTIONS, status);
}

UnicodeString
DateTimePatternGenerator::getBestPattern(const UnicodeString& patternForm, UDateTimePatternMatchOptions options, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return UnicodeString();
    }
    if (U_FAILURE(internalErrorCode)) {
        status = internalErrorCode;
        return UnicodeString();
    }
    const UnicodeString *bestPattern = nullptr;
    UnicodeString dtFormat;
    UnicodeString resultPattern;
    int32_t flags = kDTPGNoFlags;

    int32_t dateMask=(1<<UDATPG_DAYPERIOD_FIELD) - 1;
    int32_t timeMask=(1<<UDATPG_FIELD_COUNT) - 1 - dateMask;

    // Replace hour metacharacters 'j', 'C' and 'J', set flags as necessary
    UnicodeString patternFormMapped = mapSkeletonMetacharacters(patternForm, &flags, status);
    if (U_FAILURE(status)) {
        return UnicodeString();
    }

    resultPattern.remove();
    dtMatcher->set(patternFormMapped, fp);
    const PtnSkeleton* specifiedSkeleton = nullptr;
    bestPattern=getBestRaw(*dtMatcher, -1, distanceInfo, status, &specifiedSkeleton);
    if (U_FAILURE(status)) {
        return UnicodeString();
    }

    if ( distanceInfo->missingFieldMask==0 && distanceInfo->extraFieldMask==0 ) {
        resultPattern = adjustFieldTypes(*bestPattern, specifiedSkeleton, flags, options);

        return resultPattern;
    }
    int32_t neededFields = dtMatcher->getFieldMask();
    UnicodeString datePattern=getBestAppending(neededFields & dateMask, flags, status, options);
    UnicodeString timePattern=getBestAppending(neededFields & timeMask, flags, status, options);
    if (U_FAILURE(status)) {
        return UnicodeString();
    }
    if (datePattern.length()==0) {
        if (timePattern.length()==0) {
            resultPattern.remove();
        }
        else {
            return timePattern;
        }
    }
    if (timePattern.length()==0) {
        return datePattern;
    }
    resultPattern.remove();
    status = U_ZERO_ERROR;
    dtFormat=getDateTimeFormat();
    SimpleFormatter(dtFormat, 2, 2, status).format(timePattern, datePattern, resultPattern, status);
    return resultPattern;
}

/*
 * Map a skeleton that may have metacharacters jJC to one without, by replacing
 * the metacharacters with locale-appropriate fields of h/H/k/K and of a/b/B
 * (depends on fDefaultHourFormatChar and fAllowedHourFormats being set, which in
 * turn depends on initData having been run). This method also updates the flags
 * as necessary. Returns the updated skeleton.
 */
UnicodeString
DateTimePatternGenerator::mapSkeletonMetacharacters(const UnicodeString& patternForm, int32_t* flags, UErrorCode& status) {
    UnicodeString patternFormMapped;
    patternFormMapped.remove();
    UBool inQuoted = FALSE;
    int32_t patPos, patLen = patternForm.length();
    for (patPos = 0; patPos < patLen; patPos++) {
        UChar patChr = patternForm.charAt(patPos);
        if (patChr == dtptngen_SINGLE_QUOTE) {
            inQuoted = !inQuoted;
        } else if (!inQuoted) {
            // Handle special mappings for 'j' and 'C' in which fields lengths
            // 1,3,5 => hour field length 1
            // 2,4,6 => hour field length 2
            // 1,2 => abbreviated dayPeriod (field length 1..3)
            // 3,4 => long dayPeriod (field length 4)
            // 5,6 => narrow dayPeriod (field length 5)
            if (patChr == dtptngen_LOW_J || patChr == dtptngen_CAP_C) {
                int32_t extraLen = 0; // 1 less than total field length
                while (patPos+1 < patLen && patternForm.charAt(patPos+1)==patChr) {
                    extraLen++;
                    patPos++;
                }
                int32_t hourLen = 1 + (extraLen & 1);
                int32_t dayPeriodLen = (extraLen < 2)? 1: 3 + (extraLen >> 1);
                UChar hourChar = dtptngen_LOW_H;
                UChar dayPeriodChar = dtptngen_LOW_A;
                if (patChr == dtptngen_LOW_J) {
                    hourChar = fDefaultHourFormatChar;
                } else {
                    AllowedHourFormat bestAllowed;
                    if (fAllowedHourFormats[0] != ALLOWED_HOUR_FORMAT_UNKNOWN) {
                        bestAllowed = (AllowedHourFormat)fAllowedHourFormats[0];
                    } else {
                        status = U_INVALID_FORMAT_ERROR;
                        return UnicodeString();
                    }
                    if (bestAllowed == ALLOWED_HOUR_FORMAT_H || bestAllowed == ALLOWED_HOUR_FORMAT_HB || bestAllowed == ALLOWED_HOUR_FORMAT_Hb) {
                        hourChar = dtptngen_CAP_H;
                    } else if (bestAllowed == ALLOWED_HOUR_FORMAT_K || bestAllowed == ALLOWED_HOUR_FORMAT_KB || bestAllowed == ALLOWED_HOUR_FORMAT_Kb) {
                        hourChar = dtptngen_CAP_K;
                    } else if (bestAllowed == ALLOWED_HOUR_FORMAT_k) {
                        hourChar = dtptngen_LOW_K;
                    }
                    // in #13183 just add b/B to skeleton, no longer need to set special flags
                    if (bestAllowed == ALLOWED_HOUR_FORMAT_HB || bestAllowed == ALLOWED_HOUR_FORMAT_hB || bestAllowed == ALLOWED_HOUR_FORMAT_KB) {
                        dayPeriodChar = dtptngen_CAP_B;
                    } else if (bestAllowed == ALLOWED_HOUR_FORMAT_Hb || bestAllowed == ALLOWED_HOUR_FORMAT_hb || bestAllowed == ALLOWED_HOUR_FORMAT_Kb) {
                        dayPeriodChar = dtptngen_LOW_B;
                    }
                }
                if (hourChar==dtptngen_CAP_H || hourChar==dtptngen_LOW_K) {
                    dayPeriodLen = 0;
                }
                while (dayPeriodLen-- > 0) {
                    patternFormMapped.append(dayPeriodChar);
                }
                while (hourLen-- > 0) {
                    patternFormMapped.append(hourChar);
                }
            } else if (patChr == dtptngen_CAP_J) {
                // Get pattern for skeleton with H, then replace H or k
                // with fDefaultHourFormatChar (if different)
                patternFormMapped.append(dtptngen_CAP_H);
                *flags |= kDTPGSkeletonUsesCapJ;
            } else {
                patternFormMapped.append(patChr);
            }
        }
    }
    return patternFormMapped;
}

UnicodeString
DateTimePatternGenerator::replaceFieldTypes(const UnicodeString& pattern,
                                            const UnicodeString& skeleton,
                                            UErrorCode& status) {
    return replaceFieldTypes(pattern, skeleton, UDATPG_MATCH_NO_OPTIONS, status);
}

UnicodeString
DateTimePatternGenerator::replaceFieldTypes(const UnicodeString& pattern,
                                            const UnicodeString& skeleton,
                                            UDateTimePatternMatchOptions options,
                                            UErrorCode& status) {
    if (U_FAILURE(status)) {
        return UnicodeString();
    }
    if (U_FAILURE(internalErrorCode)) {
        status = internalErrorCode;
        return UnicodeString();
    }
    dtMatcher->set(skeleton, fp);
    UnicodeString result = adjustFieldTypes(pattern, nullptr, kDTPGNoFlags, options);
    return result;
}

void
DateTimePatternGenerator::setDecimal(const UnicodeString& newDecimal) {
    this->decimal = newDecimal;
    // NUL-terminate for the C API.
    this->decimal.getTerminatedBuffer();
}

const UnicodeString&
DateTimePatternGenerator::getDecimal() const {
    return decimal;
}

void
DateTimePatternGenerator::addCanonicalItems(UErrorCode& status) {
    if (U_FAILURE(status)) { return; }
    UnicodeString  conflictingPattern;

    for (int32_t i=0; i<UDATPG_FIELD_COUNT; i++) {
        if (Canonical_Items[i] > 0) {
            addPattern(UnicodeString(Canonical_Items[i]), FALSE, conflictingPattern, status);
        }
        if (U_FAILURE(status)) { return; }
    }
}

void
DateTimePatternGenerator::setDateTimeFormat(const UnicodeString& dtFormat) {
    dateTimeFormat = dtFormat;
    // NUL-terminate for the C API.
    dateTimeFormat.getTerminatedBuffer();
}

const UnicodeString&
DateTimePatternGenerator::getDateTimeFormat() const {
    return dateTimeFormat;
}

void
DateTimePatternGenerator::setDateTimeFromCalendar(const Locale& locale, UErrorCode& status) {
    if (U_FAILURE(status)) { return; }

    const UChar *resStr;
    int32_t resStrLen = 0;

    LocalPointer<Calendar> fCalendar(Calendar::createInstance(locale, status), status);
    if (U_FAILURE(status)) { return; }

    LocalUResourceBundlePointer calData(ures_open(nullptr, locale.getBaseName(), &status));
    if (U_FAILURE(status)) { return; }
    ures_getByKey(calData.getAlias(), DT_DateTimeCalendarTag, calData.getAlias(), &status);
    if (U_FAILURE(status)) { return; }

    LocalUResourceBundlePointer dateTimePatterns;
    if (fCalendar->getType() != nullptr && *fCalendar->getType() != '\0'
            && uprv_strcmp(fCalendar->getType(), DT_DateTimeGregorianTag) != 0) {
        dateTimePatterns.adoptInstead(ures_getByKeyWithFallback(calData.getAlias(), fCalendar->getType(),
                                                                nullptr, &status));
        ures_getByKeyWithFallback(dateTimePatterns.getAlias(), DT_DateTimePatternsTag,
                                  dateTimePatterns.getAlias(), &status);
    }

    if (dateTimePatterns.isNull() || status == U_MISSING_RESOURCE_ERROR) {
        status = U_ZERO_ERROR;
        dateTimePatterns.adoptInstead(ures_getByKeyWithFallback(calData.getAlias(), DT_DateTimeGregorianTag,
                                                                dateTimePatterns.orphan(), &status));
        ures_getByKeyWithFallback(dateTimePatterns.getAlias(), DT_DateTimePatternsTag,
                                  dateTimePatterns.getAlias(), &status);
    }
    if (U_FAILURE(status)) { return; }

    if (ures_getSize(dateTimePatterns.getAlias()) <= DateFormat::kDateTime)
    {
        status = U_INVALID_FORMAT_ERROR;
        return;
    }
    resStr = ures_getStringByIndex(dateTimePatterns.getAlias(), (int32_t)DateFormat::kDateTime, &resStrLen, &status);
    setDateTimeFormat(UnicodeString(TRUE, resStr, resStrLen));
}

void
DateTimePatternGenerator::setDecimalSymbols(const Locale& locale, UErrorCode& status) {
    DecimalFormatSymbols dfs = DecimalFormatSymbols(locale, status);
    if(U_SUCCESS(status)) {
        decimal = dfs.getSymbol(DecimalFormatSymbols::kDecimalSeparatorSymbol);
        // NUL-terminate for the C API.
        decimal.getTerminatedBuffer();
    }
}

UDateTimePatternConflict
DateTimePatternGenerator::addPattern(
    const UnicodeString& pattern,
    UBool override,
    UnicodeString &conflictingPattern,
    UErrorCode& status)
{
    if (U_FAILURE(internalErrorCode)) {
        status = internalErrorCode;
        return UDATPG_NO_CONFLICT;
    }

    return addPatternWithSkeleton(pattern, nullptr, override, conflictingPattern, status);
}

// For DateTimePatternGenerator::addPatternWithSkeleton -
// If skeletonToUse is specified, then an availableFormats entry is being added. In this case:
// 1. We pass that skeleton to matcher.set instead of having it derive a skeleton from the pattern.
// 2. If the new entry's skeleton or basePattern does match an existing entry but that entry also had a skeleton specified
// (i.e. it was also from availableFormats), then the new entry does not override it regardless of the value of the override
// parameter. This prevents later availableFormats entries from a parent locale overriding earlier ones from the actual
// specified locale. However, availableFormats entries *should* override entries with matching skeleton whose skeleton was
// derived (i.e. entries derived from the standard date/time patters for the specified locale).
// 3. When adding the pattern (patternMap->add), we set a new boolean to indicate that the added entry had a
// specified skeleton (which sets a new field in the PtnElem in the PatternMap).
UDateTimePatternConflict
DateTimePatternGenerator::addPatternWithSkeleton(
    const UnicodeString& pattern,
    const UnicodeString* skeletonToUse,
    UBool override,
    UnicodeString& conflictingPattern,
    UErrorCode& status)
{
    if (U_FAILURE(internalErrorCode)) {
        status = internalErrorCode;
        return UDATPG_NO_CONFLICT;
    }

    UnicodeString basePattern;
    PtnSkeleton   skeleton;
    UDateTimePatternConflict conflictingStatus = UDATPG_NO_CONFLICT;

    DateTimeMatcher matcher;
    if ( skeletonToUse == nullptr ) {
        matcher.set(pattern, fp, skeleton);
        matcher.getBasePattern(basePattern);
    } else {
        matcher.set(*skeletonToUse, fp, skeleton); // no longer trims skeleton fields to max len 3, per #7930
        matcher.getBasePattern(basePattern); // or perhaps instead: basePattern = *skeletonToUse;
    }
    // We only care about base conflicts - and replacing the pattern associated with a base - if:
    // 1. the conflicting previous base pattern did *not* have an explicit skeleton; in that case the previous
    // base + pattern combination was derived from either (a) a canonical item, (b) a standard format, or
    // (c) a pattern specified programmatically with a previous call to addPattern (which would only happen
    // if we are getting here from a subsequent call to addPattern).
    // 2. a skeleton is specified for the current pattern, but override=false; in that case we are checking
    // availableFormats items from root, which should not override any previous entry with the same base.
    UBool entryHadSpecifiedSkeleton;
    const UnicodeString *duplicatePattern = patternMap->getPatternFromBasePattern(basePattern, entryHadSpecifiedSkeleton);
    if (duplicatePattern != nullptr && (!entryHadSpecifiedSkeleton || (skeletonToUse != nullptr && !override))) {
        conflictingStatus = UDATPG_BASE_CONFLICT;
        conflictingPattern = *duplicatePattern;
        if (!override) {
            return conflictingStatus;
        }
    }
    // The only time we get here with override=true and skeletonToUse!=null is when adding availableFormats
    // items from CLDR data. In that case, we don't want an item from a parent locale to replace an item with
    // same skeleton from the specified locale, so skip the current item if skeletonWasSpecified is true for
    // the previously-specified conflicting item.
    const PtnSkeleton* entrySpecifiedSkeleton = nullptr;
    duplicatePattern = patternMap->getPatternFromSkeleton(skeleton, &entrySpecifiedSkeleton);
    if (duplicatePattern != nullptr ) {
        conflictingStatus = UDATPG_CONFLICT;
        conflictingPattern = *duplicatePattern;
        if (!override || (skeletonToUse != nullptr && entrySpecifiedSkeleton != nullptr)) {
            return conflictingStatus;
        }
    }
    patternMap->add(basePattern, skeleton, pattern, skeletonToUse != nullptr, status);
    if(U_FAILURE(status)) {
        return conflictingStatus;
    }

    return UDATPG_NO_CONFLICT;
}


UDateTimePatternField
DateTimePatternGenerator::getAppendFormatNumber(const char* field) const {
    for (int32_t i=0; i<UDATPG_FIELD_COUNT; ++i ) {
        if (uprv_strcmp(CLDR_FIELD_APPEND[i], field)==0) {
            return (UDateTimePatternField)i;
        }
    }
    return UDATPG_FIELD_COUNT;
}

UDateTimePatternField
DateTimePatternGenerator::getFieldAndWidthIndices(const char* key, UDateTimePGDisplayWidth* widthP) const {
    char cldrFieldKey[UDATPG_FIELD_KEY_MAX + 1];
    uprv_strncpy(cldrFieldKey, key, UDATPG_FIELD_KEY_MAX);
    cldrFieldKey[UDATPG_FIELD_KEY_MAX]=0; // ensure termination
    *widthP = UDATPG_WIDE;
    char* hyphenPtr = uprv_strchr(cldrFieldKey, '-');
    if (hyphenPtr) {
        for (int32_t i=UDATPG_WIDTH_COUNT-1; i>0; --i) {
            if (uprv_strcmp(CLDR_FIELD_WIDTH[i], hyphenPtr)==0) {
                *widthP=(UDateTimePGDisplayWidth)i;
                break;
            }
        }
        *hyphenPtr = 0; // now delete width portion of key
    }
    for (int32_t i=0; i<UDATPG_FIELD_COUNT; ++i ) {
        if (uprv_strcmp(CLDR_FIELD_NAME[i],cldrFieldKey)==0) {
            return (UDateTimePatternField)i;
        }
    }
    return UDATPG_FIELD_COUNT;
}

const UnicodeString*
DateTimePatternGenerator::getBestRaw(DateTimeMatcher& source,
                                     int32_t includeMask,
                                     DistanceInfo* missingFields,
                                     UErrorCode &status,
                                     const PtnSkeleton** specifiedSkeletonPtr) {
    int32_t bestDistance = 0x7fffffff;
    DistanceInfo tempInfo;
    const UnicodeString *bestPattern=nullptr;
    const PtnSkeleton* specifiedSkeleton=nullptr;

    PatternMapIterator it(status);
    if (U_FAILURE(status)) { return nullptr; }

    for (it.set(*patternMap); it.hasNext(); ) {
        DateTimeMatcher trial = it.next();
        if (trial.equals(skipMatcher)) {
            continue;
        }
        int32_t distance=source.getDistance(trial, includeMask, tempInfo);
        if (distance<bestDistance) {
            bestDistance=distance;
            bestPattern=patternMap->getPatternFromSkeleton(*trial.getSkeletonPtr(), &specifiedSkeleton);
            missingFields->setTo(tempInfo);
            if (distance==0) {
                break;
            }
        }
    }

    // If the best raw match had a specified skeleton and that skeleton was requested by the caller,
    // then return it too. This generally happens when the caller needs to pass that skeleton
    // through to adjustFieldTypes so the latter can do a better job.
    if (bestPattern && specifiedSkeletonPtr) {
        *specifiedSkeletonPtr = specifiedSkeleton;
    }
    return bestPattern;
}

UnicodeString
DateTimePatternGenerator::adjustFieldTypes(const UnicodeString& pattern,
                                           const PtnSkeleton* specifiedSkeleton,
                                           int32_t flags,
                                           UDateTimePatternMatchOptions options) {
    UnicodeString newPattern;
    fp->set(pattern);
    for (int32_t i=0; i < fp->itemNumber; i++) {
        UnicodeString field = fp->items[i];
        if ( fp->isQuoteLiteral(field) ) {

            UnicodeString quoteLiteral;
            fp->getQuoteLiteral(quoteLiteral, &i);
            newPattern += quoteLiteral;
        }
        else {
            if (fp->isPatternSeparator(field)) {
                newPattern+=field;
                continue;
            }
            int32_t canonicalIndex = fp->getCanonicalIndex(field);
            if (canonicalIndex < 0) {
                newPattern+=field;
                continue;  // don't adjust
            }
            const dtTypeElem *row = &dtTypes[canonicalIndex];
            int32_t typeValue = row->field;

            // handle day periods - with #13183, no longer need special handling here, integrated with normal types

            if ((flags & kDTPGFixFractionalSeconds) != 0 && typeValue == UDATPG_SECOND_FIELD) {
                field += decimal;
                dtMatcher->skeleton.original.appendFieldTo(UDATPG_FRACTIONAL_SECOND_FIELD, field);
            } else if (dtMatcher->skeleton.type[typeValue]!=0) {
                    // Here:
                    // - "reqField" is the field from the originally requested skeleton, with length
                    // "reqFieldLen".
                    // - "field" is the field from the found pattern.
                    //
                    // The adjusted field should consist of characters from the originally requested
                    // skeleton, except in the case of UDATPG_HOUR_FIELD or UDATPG_MONTH_FIELD or
                    // UDATPG_WEEKDAY_FIELD or UDATPG_YEAR_FIELD, in which case it should consist
                    // of characters from the  found pattern.
                    //
                    // The length of the adjusted field (adjFieldLen) should match that in the originally
                    // requested skeleton, except that in the following cases the length of the adjusted field
                    // should match that in the found pattern (i.e. the length of this pattern field should
                    // not be adjusted):
                    // 1. typeValue is UDATPG_HOUR_FIELD/MINUTE/SECOND and the corresponding bit in options is
                    //    not set (ticket #7180). Note, we may want to implement a similar change for other
                    //    numeric fields (MM, dd, etc.) so the default behavior is to get locale preference for
                    //    field length, but options bits can be used to override this.
                    // 2. There is a specified skeleton for the found pattern and one of the following is true:
                    //    a) The length of the field in the skeleton (skelFieldLen) is equal to reqFieldLen.
                    //    b) The pattern field is numeric and the skeleton field is not, or vice versa.

                    UChar reqFieldChar = dtMatcher->skeleton.original.getFieldChar(typeValue);
                    int32_t reqFieldLen = dtMatcher->skeleton.original.getFieldLength(typeValue);
                    if (reqFieldChar == dtptngen_CAP_E && reqFieldLen < 3)
                        reqFieldLen = 3; // 1-3 for E are equivalent to 3 for c,e
                    int32_t adjFieldLen = reqFieldLen;
                    if ( (typeValue==UDATPG_HOUR_FIELD && (options & UDATPG_MATCH_HOUR_FIELD_LENGTH)==0) ||
                         (typeValue==UDATPG_MINUTE_FIELD && (options & UDATPG_MATCH_MINUTE_FIELD_LENGTH)==0) ||
                         (typeValue==UDATPG_SECOND_FIELD && (options & UDATPG_MATCH_SECOND_FIELD_LENGTH)==0) ) {
                         adjFieldLen = field.length();
                    } else if (specifiedSkeleton) {
                        int32_t skelFieldLen = specifiedSkeleton->original.getFieldLength(typeValue);
                        UBool patFieldIsNumeric = (row->type > 0);
                        UBool skelFieldIsNumeric = (specifiedSkeleton->type[typeValue] > 0);
                        if (skelFieldLen == reqFieldLen || (patFieldIsNumeric && !skelFieldIsNumeric) || (skelFieldIsNumeric && !patFieldIsNumeric)) {
                            // don't adjust the field length in the found pattern
                            adjFieldLen = field.length();
                        }
                    }
                    UChar c = (typeValue!= UDATPG_HOUR_FIELD
                            && typeValue!= UDATPG_MONTH_FIELD
                            && typeValue!= UDATPG_WEEKDAY_FIELD
                            && (typeValue!= UDATPG_YEAR_FIELD || reqFieldChar==dtptngen_CAP_Y))
                            ? reqFieldChar
                            : field.charAt(0);
                    if (typeValue == UDATPG_HOUR_FIELD && (flags & kDTPGSkeletonUsesCapJ) != 0) {
                        c = fDefaultHourFormatChar;
                    }
                    field.remove();
                    for (int32_t j=adjFieldLen; j>0; --j) {
                        field += c;
                    }
            }
            newPattern+=field;
        }
    }
    return newPattern;
}

UnicodeString
DateTimePatternGenerator::getBestAppending(int32_t missingFields, int32_t flags, UErrorCode &status, UDateTimePatternMatchOptions options) {
    if (U_FAILURE(status)) {
        return UnicodeString();
    }
    UnicodeString  resultPattern, tempPattern;
    const UnicodeString* tempPatternPtr;
    int32_t lastMissingFieldMask=0;
    if (missingFields!=0) {
        resultPattern=UnicodeString();
        const PtnSkeleton* specifiedSkeleton=nullptr;
        tempPatternPtr = getBestRaw(*dtMatcher, missingFields, distanceInfo, status, &specifiedSkeleton);
        if (U_FAILURE(status)) {
            return UnicodeString();
        }
        tempPattern = *tempPatternPtr;
        resultPattern = adjustFieldTypes(tempPattern, specifiedSkeleton, flags, options);
        if ( distanceInfo->missingFieldMask==0 ) {
            return resultPattern;
        }
        while (distanceInfo->missingFieldMask!=0) { // precondition: EVERY single field must work!
            if ( lastMissingFieldMask == distanceInfo->missingFieldMask ) {
                break;  // cannot find the proper missing field
            }
            if (((distanceInfo->missingFieldMask & UDATPG_SECOND_AND_FRACTIONAL_MASK)==UDATPG_FRACTIONAL_MASK) &&
                ((missingFields & UDATPG_SECOND_AND_FRACTIONAL_MASK) == UDATPG_SECOND_AND_FRACTIONAL_MASK)) {
                resultPattern = adjustFieldTypes(resultPattern, specifiedSkeleton, flags | kDTPGFixFractionalSeconds, options);
                distanceInfo->missingFieldMask &= ~UDATPG_FRACTIONAL_MASK;
                continue;
            }
            int32_t startingMask = distanceInfo->missingFieldMask;
            tempPatternPtr = getBestRaw(*dtMatcher, distanceInfo->missingFieldMask, distanceInfo, status, &specifiedSkeleton);
            if (U_FAILURE(status)) {
                return UnicodeString();
            }
            tempPattern = *tempPatternPtr;
            tempPattern = adjustFieldTypes(tempPattern, specifiedSkeleton, flags, options);
            int32_t foundMask=startingMask& ~distanceInfo->missingFieldMask;
            int32_t topField=getTopBitNumber(foundMask);

            if (appendItemFormats[topField].length() != 0) {
                UnicodeString appendName;
                getAppendName((UDateTimePatternField)topField, appendName);
                const UnicodeString *values[3] = {
                    &resultPattern,
                    &tempPattern,
                    &appendName
                };
                SimpleFormatter(appendItemFormats[topField], 2, 3, status).
                    formatAndReplace(values, 3, resultPattern, nullptr, 0, status);
            }
            lastMissingFieldMask = distanceInfo->missingFieldMask;
        }
    }
    return resultPattern;
}

int32_t
DateTimePatternGenerator::getTopBitNumber(int32_t foundMask) const {
    if ( foundMask==0 ) {
        return 0;
    }
    int32_t i=0;
    while (foundMask!=0) {
        foundMask >>=1;
        ++i;
    }
    if (i-1 >UDATPG_ZONE_FIELD) {
        return UDATPG_ZONE_FIELD;
    }
    else
        return i-1;
}

void
DateTimePatternGenerator::setAvailableFormat(const UnicodeString &key, UErrorCode& err)
{
    fAvailableFormatKeyHash->puti(key, 1, err);
}

UBool
DateTimePatternGenerator::isAvailableFormatSet(const UnicodeString &key) const {
    return (UBool)(fAvailableFormatKeyHash->geti(key) == 1);
}

void
DateTimePatternGenerator::copyHashtable(Hashtable *other, UErrorCode &status) {
    if (other == nullptr || U_FAILURE(status)) {
        return;
    }
    if (fAvailableFormatKeyHash != nullptr) {
        delete fAvailableFormatKeyHash;
        fAvailableFormatKeyHash = nullptr;
    }
    initHashtable(status);
    if(U_FAILURE(status)){
        return;
    }
    int32_t pos = UHASH_FIRST;
    const UHashElement* elem = nullptr;
    // walk through the hash table and create a deep clone
    while((elem = other->nextElement(pos))!= nullptr){
        const UHashTok otherKeyTok = elem->key;
        UnicodeString* otherKey = (UnicodeString*)otherKeyTok.pointer;
        fAvailableFormatKeyHash->puti(*otherKey, 1, status);
        if(U_FAILURE(status)){
            return;
        }
    }
}

StringEnumeration*
DateTimePatternGenerator::getSkeletons(UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (U_FAILURE(internalErrorCode)) {
        status = internalErrorCode;
        return nullptr;
    }
    LocalPointer<StringEnumeration> skeletonEnumerator(
        new DTSkeletonEnumeration(*patternMap, DT_SKELETON, status), status);

    return U_SUCCESS(status) ? skeletonEnumerator.orphan() : nullptr;
}

const UnicodeString&
DateTimePatternGenerator::getPatternForSkeleton(const UnicodeString& skeleton) const {
    PtnElem *curElem;

    if (skeleton.length() ==0) {
        return emptyString;
    }
    curElem = patternMap->getHeader(skeleton.charAt(0));
    while ( curElem != nullptr ) {
        if ( curElem->skeleton->getSkeleton()==skeleton ) {
            return curElem->pattern;
        }
        curElem = curElem->next.getAlias();
    }
    return emptyString;
}

StringEnumeration*
DateTimePatternGenerator::getBaseSkeletons(UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (U_FAILURE(internalErrorCode)) {
        status = internalErrorCode;
        return nullptr;
    }
    LocalPointer<StringEnumeration> baseSkeletonEnumerator(
        new DTSkeletonEnumeration(*patternMap, DT_BASESKELETON, status), status);

    return U_SUCCESS(status) ? baseSkeletonEnumerator.orphan() : nullptr;
}

StringEnumeration*
DateTimePatternGenerator::getRedundants(UErrorCode& status) {
    if (U_FAILURE(status)) { return nullptr; }
    if (U_FAILURE(internalErrorCode)) {
        status = internalErrorCode;
        return nullptr;
    }
    LocalPointer<StringEnumeration> output(new DTRedundantEnumeration(), status);
    if (U_FAILURE(status)) { return nullptr; }
    const UnicodeString *pattern;
    PatternMapIterator it(status);
    if (U_FAILURE(status)) { return nullptr; }

    for (it.set(*patternMap); it.hasNext(); ) {
        DateTimeMatcher current = it.next();
        pattern = patternMap->getPatternFromSkeleton(*(it.getSkeleton()));
        if ( isCanonicalItem(*pattern) ) {
            continue;
        }
        if ( skipMatcher == nullptr ) {
            skipMatcher = new DateTimeMatcher(current);
            if (skipMatcher == nullptr) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return nullptr;
            }
        }
        else {
            *skipMatcher = current;
        }
        UnicodeString trial = getBestPattern(current.getPattern(), status);
        if (U_FAILURE(status)) { return nullptr; }
        if (trial == *pattern) {
            ((DTRedundantEnumeration *)output.getAlias())->add(*pattern, status);
            if (U_FAILURE(status)) { return nullptr; }
        }
        if (current.equals(skipMatcher)) {
            continue;
        }
    }
    return output.orphan();
}

UBool
DateTimePatternGenerator::isCanonicalItem(const UnicodeString& item) const {
    if ( item.length() != 1 ) {
        return FALSE;
    }
    for (int32_t i=0; i<UDATPG_FIELD_COUNT; ++i) {
        if (item.charAt(0)==Canonical_Items[i]) {
            return TRUE;
        }
    }
    return FALSE;
}


DateTimePatternGenerator*
DateTimePatternGenerator::clone() const {
    return new DateTimePatternGenerator(*this);
}

PatternMap::PatternMap() {
   for (int32_t i=0; i < dtptngen_MAX_PATTERN_ENTRIES; ++i ) {
       boot[i] = nullptr;
   }
   isDupAllowed = TRUE;
}

void
PatternMap::copyFrom(const PatternMap& other, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    this->isDupAllowed = other.isDupAllowed;
    for (int32_t bootIndex = 0; bootIndex < dtptngen_MAX_PATTERN_ENTRIES; ++bootIndex) {
        PtnElem *curElem, *otherElem, *prevElem=nullptr;
        otherElem = other.boot[bootIndex];
        while (otherElem != nullptr) {
            LocalPointer<PtnElem> newElem(new PtnElem(otherElem->basePattern, otherElem->pattern), status);
            if (U_FAILURE(status)) {
                return; // out of memory
            }
            newElem->skeleton.adoptInsteadAndCheckErrorCode(new PtnSkeleton(*(otherElem->skeleton)), status);
            if (U_FAILURE(status)) {
                return; // out of memory
            }
            newElem->skeletonWasSpecified = otherElem->skeletonWasSpecified;

            // Release ownership from the LocalPointer of the PtnElem object.
            // The PtnElem will now be owned by either the boot (for the first entry in the linked-list)
            // or owned by the previous PtnElem object in the linked-list.
            curElem = newElem.orphan();

            if (this->boot[bootIndex] == nullptr) {
                this->boot[bootIndex] = curElem;
            } else {
                if (prevElem != nullptr) {
                    prevElem->next.adoptInstead(curElem);
                } else {
                    UPRV_UNREACHABLE;
                }
            }
            prevElem = curElem;
            otherElem = otherElem->next.getAlias();
        }

    }
}

PtnElem*
PatternMap::getHeader(UChar baseChar) const {
    PtnElem* curElem;

    if ( (baseChar >= dtptngen_CAP_A) && (baseChar <= dtptngen_CAP_Z) ) {
         curElem = boot[baseChar-dtptngen_CAP_A];
    }
    else {
        if ( (baseChar >=dtptngen_LOW_A) && (baseChar <= dtptngen_LOW_Z) ) {
            curElem = boot[26+baseChar-dtptngen_LOW_A];
        }
        else {
            return nullptr;
        }
    }
    return curElem;
}

PatternMap::~PatternMap() {
   for (int32_t i=0; i < dtptngen_MAX_PATTERN_ENTRIES; ++i ) {
       if (boot[i] != nullptr ) {
           delete boot[i];
           boot[i] = nullptr;
       }
   }
}  // PatternMap destructor

void
PatternMap::add(const UnicodeString& basePattern,
                const PtnSkeleton& skeleton,
                const UnicodeString& value,// mapped pattern value
                UBool skeletonWasSpecified,
                UErrorCode &status) {
    UChar baseChar = basePattern.charAt(0);
    PtnElem *curElem, *baseElem;
    status = U_ZERO_ERROR;

    // the baseChar must be A-Z or a-z
    if ((baseChar >= dtptngen_CAP_A) && (baseChar <= dtptngen_CAP_Z)) {
        baseElem = boot[baseChar-dtptngen_CAP_A];
    }
    else {
        if ((baseChar >=dtptngen_LOW_A) && (baseChar <= dtptngen_LOW_Z)) {
            baseElem = boot[26+baseChar-dtptngen_LOW_A];
         }
         else {
             status = U_ILLEGAL_CHARACTER;
             return;
         }
    }

    if (baseElem == nullptr) {
        LocalPointer<PtnElem> newElem(new PtnElem(basePattern, value), status);
        if (U_FAILURE(status)) {
            return; // out of memory
        }
        newElem->skeleton.adoptInsteadAndCheckErrorCode(new PtnSkeleton(skeleton), status);
        if (U_FAILURE(status)) {
            return; // out of memory
        }
        newElem->skeletonWasSpecified = skeletonWasSpecified;
        if (baseChar >= dtptngen_LOW_A) {
            boot[26 + (baseChar - dtptngen_LOW_A)] = newElem.orphan(); // the boot array now owns the PtnElem.
        }
        else {
            boot[baseChar - dtptngen_CAP_A] = newElem.orphan(); // the boot array now owns the PtnElem.
        }
    }
    if ( baseElem != nullptr ) {
        curElem = getDuplicateElem(basePattern, skeleton, baseElem);

        if (curElem == nullptr) {
            // add new element to the list.
            curElem = baseElem;
            while( curElem -> next != nullptr )
            {
                curElem = curElem->next.getAlias();
            }

            LocalPointer<PtnElem> newElem(new PtnElem(basePattern, value), status);
            if (U_FAILURE(status)) {
                return; // out of memory
            }
            newElem->skeleton.adoptInsteadAndCheckErrorCode(new PtnSkeleton(skeleton), status);
            if (U_FAILURE(status)) {
                return; // out of memory
            }
            newElem->skeletonWasSpecified = skeletonWasSpecified;
            curElem->next.adoptInstead(newElem.orphan());
            curElem = curElem->next.getAlias();
        }
        else {
            // Pattern exists in the list already.
            if ( !isDupAllowed ) {
                return;
            }
            // Overwrite the value.
            curElem->pattern = value;
            // It was a bug that we were not doing the following previously,
            // though that bug hid other problems by making things partly work.
            curElem->skeletonWasSpecified = skeletonWasSpecified;
        }
    }
}  // PatternMap::add

// Find the pattern from the given basePattern string.
const UnicodeString *
PatternMap::getPatternFromBasePattern(const UnicodeString& basePattern, UBool& skeletonWasSpecified) const { // key to search for
   PtnElem *curElem;

   if ((curElem=getHeader(basePattern.charAt(0)))==nullptr) {
       return nullptr;  // no match
   }

   do  {
       if ( basePattern.compare(curElem->basePattern)==0 ) {
          skeletonWasSpecified = curElem->skeletonWasSpecified;
          return &(curElem->pattern);
       }
       curElem = curElem->next.getAlias();
   } while (curElem != nullptr);

   return nullptr;
}  // PatternMap::getFromBasePattern


// Find the pattern from the given skeleton.
// At least when this is called from getBestRaw & addPattern (in which case specifiedSkeletonPtr is non-NULL),
// the comparison should be based on skeleton.original (which is unique and tied to the distance measurement in bestRaw)
// and not skeleton.baseOriginal (which is not unique); otherwise we may pick a different skeleton than the one with the
// optimum distance value in getBestRaw. When this is called from public getRedundants (specifiedSkeletonPtr is NULL),
// for now it will continue to compare based on baseOriginal so as not to change the behavior unnecessarily.
const UnicodeString *
PatternMap::getPatternFromSkeleton(const PtnSkeleton& skeleton, const PtnSkeleton** specifiedSkeletonPtr) const { // key to search for
   PtnElem *curElem;

   if (specifiedSkeletonPtr) {
       *specifiedSkeletonPtr = nullptr;
   }

   // find boot entry
   UChar baseChar = skeleton.getFirstChar();
   if ((curElem=getHeader(baseChar))==nullptr) {
       return nullptr;  // no match
   }

   do  {
       UBool equal;
       if (specifiedSkeletonPtr != nullptr) { // called from DateTimePatternGenerator::getBestRaw or addPattern, use original
           equal = curElem->skeleton->original == skeleton.original;
       } else { // called from DateTimePatternGenerator::getRedundants, use baseOriginal
           equal = curElem->skeleton->baseOriginal == skeleton.baseOriginal;
       }
       if (equal) {
           if (specifiedSkeletonPtr && curElem->skeletonWasSpecified) {
               *specifiedSkeletonPtr = curElem->skeleton.getAlias();
           }
           return &(curElem->pattern);
       }
       curElem = curElem->next.getAlias();
   } while (curElem != nullptr);

   return nullptr;
}

UBool
PatternMap::equals(const PatternMap& other) const {
    if ( this==&other ) {
        return TRUE;
    }
    for (int32_t bootIndex = 0; bootIndex < dtptngen_MAX_PATTERN_ENTRIES; ++bootIndex) {
        if (boot[bootIndex] == other.boot[bootIndex]) {
            continue;
        }
        if ((boot[bootIndex] == nullptr) || (other.boot[bootIndex] == nullptr)) {
            return FALSE;
        }
        PtnElem *otherElem = other.boot[bootIndex];
        PtnElem *myElem = boot[bootIndex];
        while ((otherElem != nullptr) || (myElem != nullptr)) {
            if ( myElem == otherElem ) {
                break;
            }
            if ((otherElem == nullptr) || (myElem == nullptr)) {
                return FALSE;
            }
            if ( (myElem->basePattern != otherElem->basePattern) ||
                 (myElem->pattern != otherElem->pattern) ) {
                return FALSE;
            }
            if ((myElem->skeleton.getAlias() != otherElem->skeleton.getAlias()) &&
                !myElem->skeleton->equals(*(otherElem->skeleton))) {
                return FALSE;
            }
            myElem = myElem->next.getAlias();
            otherElem = otherElem->next.getAlias();
        }
    }
    return TRUE;
}

// find any key existing in the mapping table already.
// return TRUE if there is an existing key, otherwise return FALSE.
PtnElem*
PatternMap::getDuplicateElem(
            const UnicodeString &basePattern,
            const PtnSkeleton &skeleton,
            PtnElem *baseElem) {
   PtnElem *curElem;

   if ( baseElem == nullptr ) {
         return nullptr;
   }
   else {
         curElem = baseElem;
   }
   do {
     if ( basePattern.compare(curElem->basePattern)==0 ) {
         UBool isEqual = TRUE;
         for (int32_t i = 0; i < UDATPG_FIELD_COUNT; ++i) {
            if (curElem->skeleton->type[i] != skeleton.type[i] ) {
                isEqual = FALSE;
                break;
            }
        }
        if (isEqual) {
            return curElem;
        }
     }
     curElem = curElem->next.getAlias();
   } while( curElem != nullptr );

   // end of the list
   return nullptr;

}  // PatternMap::getDuplicateElem

DateTimeMatcher::DateTimeMatcher(void) {
}

DateTimeMatcher::~DateTimeMatcher() {}

DateTimeMatcher::DateTimeMatcher(const DateTimeMatcher& other) {
    copyFrom(other.skeleton);
}


void
DateTimeMatcher::set(const UnicodeString& pattern, FormatParser* fp) {
    PtnSkeleton localSkeleton;
    return set(pattern, fp, localSkeleton);
}

void
DateTimeMatcher::set(const UnicodeString& pattern, FormatParser* fp, PtnSkeleton& skeletonResult) {
    int32_t i;
    for (i=0; i<UDATPG_FIELD_COUNT; ++i) {
        skeletonResult.type[i] = dtptngen_NONE;
    }
    skeletonResult.original.clear();
    skeletonResult.baseOriginal.clear();
    skeletonResult.addedDefaultDayPeriod = FALSE;

    fp->set(pattern);
    for (i=0; i < fp->itemNumber; i++) {
        const UnicodeString& value = fp->items[i];
        // don't skip 'a' anymore, dayPeriod handled specially below

        if ( fp->isQuoteLiteral(value) ) {
            UnicodeString quoteLiteral;
            fp->getQuoteLiteral(quoteLiteral, &i);
            continue;
        }
        int32_t canonicalIndex = fp->getCanonicalIndex(value);
        if (canonicalIndex < 0) {
            continue;
        }
        const dtTypeElem *row = &dtTypes[canonicalIndex];
        int32_t field = row->field;
        skeletonResult.original.populate(field, value);
        UChar repeatChar = row->patternChar;
        int32_t repeatCount = row->minLen;
        skeletonResult.baseOriginal.populate(field, repeatChar, repeatCount);
        int16_t subField = row->type;
        if (row->type > 0) {
            U_ASSERT(value.length() < INT16_MAX);
            subField += static_cast<int16_t>(value.length());
        }
        skeletonResult.type[field] = subField;
    }
    // #13183, handle special behavior for day period characters (a, b, B)
    if (!skeletonResult.original.isFieldEmpty(UDATPG_HOUR_FIELD)) {
        if (skeletonResult.original.getFieldChar(UDATPG_HOUR_FIELD)==dtptngen_LOW_H || skeletonResult.original.getFieldChar(UDATPG_HOUR_FIELD)==dtptngen_CAP_K) {
            // We have a skeleton with 12-hour-cycle format
            if (skeletonResult.original.isFieldEmpty(UDATPG_DAYPERIOD_FIELD)) {
                // But we do not have a day period in the skeleton; add the default DAYPERIOD (currently "a")
                for (i = 0; dtTypes[i].patternChar != 0; i++) {
                    if ( dtTypes[i].field == UDATPG_DAYPERIOD_FIELD ) {
                        // first entry for UDATPG_DAYPERIOD_FIELD
                        skeletonResult.original.populate(UDATPG_DAYPERIOD_FIELD, dtTypes[i].patternChar, dtTypes[i].minLen);
                        skeletonResult.baseOriginal.populate(UDATPG_DAYPERIOD_FIELD, dtTypes[i].patternChar, dtTypes[i].minLen);
                        skeletonResult.type[UDATPG_DAYPERIOD_FIELD] = dtTypes[i].type;
                        skeletonResult.addedDefaultDayPeriod = TRUE;
                        break;
                    }
                }
            }
        } else {
            // Skeleton has 24-hour-cycle hour format and has dayPeriod, delete dayPeriod (i.e. ignore it)
            skeletonResult.original.clearField(UDATPG_DAYPERIOD_FIELD);
            skeletonResult.baseOriginal.clearField(UDATPG_DAYPERIOD_FIELD);
            skeletonResult.type[UDATPG_DAYPERIOD_FIELD] = dtptngen_NONE;
        }
    }
    copyFrom(skeletonResult);
}

void
DateTimeMatcher::getBasePattern(UnicodeString &result ) {
    result.remove(); // Reset the result first.
    skeleton.baseOriginal.appendTo(result);
}

UnicodeString
DateTimeMatcher::getPattern() {
    UnicodeString result;
    return skeleton.original.appendTo(result);
}

int32_t
DateTimeMatcher::getDistance(const DateTimeMatcher& other, int32_t includeMask, DistanceInfo& distanceInfo) const {
    int32_t result = 0;
    distanceInfo.clear();
    for (int32_t i=0; i<UDATPG_FIELD_COUNT; ++i ) {
        int32_t myType = (includeMask&(1<<i))==0 ? 0 : skeleton.type[i];
        int32_t otherType = other.skeleton.type[i];
        if (myType==otherType) {
            continue;
        }
        if (myType==0) {// and other is not
            result += dtptngen_EXTRA_FIELD;
            distanceInfo.addExtra(i);
        }
        else {
            if (otherType==0) {
                result += dtptngen_MISSING_FIELD;
                distanceInfo.addMissing(i);
            }
            else {
                result += abs(myType - otherType);
            }
        }

    }
    return result;
}

void
DateTimeMatcher::copyFrom(const PtnSkeleton& newSkeleton) {
    skeleton.copyFrom(newSkeleton);
}

void
DateTimeMatcher::copyFrom() {
    // same as clear
    skeleton.clear();
}

UBool
DateTimeMatcher::equals(const DateTimeMatcher* other) const {
    if (other==nullptr) { return FALSE; }
    return skeleton.original == other->skeleton.original;
}

int32_t
DateTimeMatcher::getFieldMask() const {
    int32_t result = 0;

    for (int32_t i=0; i<UDATPG_FIELD_COUNT; ++i) {
        if (skeleton.type[i]!=0) {
            result |= (1<<i);
        }
    }
    return result;
}

PtnSkeleton*
DateTimeMatcher::getSkeletonPtr() {
    return &skeleton;
}

FormatParser::FormatParser () {
    status = START;
    itemNumber = 0;
}


FormatParser::~FormatParser () {
}


// Find the next token with the starting position and length
// Note: the startPos may
FormatParser::TokenStatus
FormatParser::setTokens(const UnicodeString& pattern, int32_t startPos, int32_t *len) {
    int32_t curLoc = startPos;
    if ( curLoc >= pattern.length()) {
        return DONE;
    }
    // check the current char is between A-Z or a-z
    do {
        UChar c=pattern.charAt(curLoc);
        if ( (c>=dtptngen_CAP_A && c<=dtptngen_CAP_Z) || (c>=dtptngen_LOW_A && c<=dtptngen_LOW_Z) ) {
           curLoc++;
        }
        else {
               startPos = curLoc;
               *len=1;
               return ADD_TOKEN;
        }

        if ( pattern.charAt(curLoc)!= pattern.charAt(startPos) ) {
            break;  // not the same token
        }
    } while(curLoc <= pattern.length());
    *len = curLoc-startPos;
    return ADD_TOKEN;
}

void
FormatParser::set(const UnicodeString& pattern) {
    int32_t startPos = 0;
    TokenStatus result = START;
    int32_t len = 0;
    itemNumber = 0;

    do {
        result = setTokens( pattern, startPos, &len );
        if ( result == ADD_TOKEN )
        {
            items[itemNumber++] = UnicodeString(pattern, startPos, len );
            startPos += len;
        }
        else {
            break;
        }
    } while (result==ADD_TOKEN && itemNumber < dtptngen_MAX_DT_TOKEN);
}

int32_t
FormatParser::getCanonicalIndex(const UnicodeString& s, UBool strict) {
    int32_t len = s.length();
    if (len == 0) {
        return -1;
    }
    UChar ch = s.charAt(0);

    // Verify that all are the same character.
    for (int32_t l = 1; l < len; l++) {
        if (ch != s.charAt(l)) {
            return -1;
        }
    }
    int32_t i = 0;
    int32_t bestRow = -1;
    while (dtTypes[i].patternChar != 0x0000) {
        if ( dtTypes[i].patternChar != ch ) {
            ++i;
            continue;
        }
        bestRow = i;
        if (dtTypes[i].patternChar != dtTypes[i+1].patternChar) {
            return i;
        }
        if (dtTypes[i+1].minLen <= len) {
            ++i;
            continue;
        }
        return i;
    }
    return strict ? -1 : bestRow;
}

UBool
FormatParser::isQuoteLiteral(const UnicodeString& s) {
    return (UBool)(s.charAt(0) == dtptngen_SINGLE_QUOTE);
}

// This function assumes the current itemIndex points to the quote literal.
// Please call isQuoteLiteral prior to this function.
void
FormatParser::getQuoteLiteral(UnicodeString& quote, int32_t *itemIndex) {
    int32_t i = *itemIndex;

    quote.remove();
    if (items[i].charAt(0)==dtptngen_SINGLE_QUOTE) {
        quote += items[i];
        ++i;
    }
    while ( i < itemNumber ) {
        if ( items[i].charAt(0)==dtptngen_SINGLE_QUOTE ) {
            if ( (i+1<itemNumber) && (items[i+1].charAt(0)==dtptngen_SINGLE_QUOTE)) {
                // two single quotes e.g. 'o''clock'
                quote += items[i++];
                quote += items[i++];
                continue;
            }
            else {
                quote += items[i];
                break;
            }
        }
        else {
            quote += items[i];
        }
        ++i;
    }
    *itemIndex=i;
}

UBool
FormatParser::isPatternSeparator(const UnicodeString& field) const {
    for (int32_t i=0; i<field.length(); ++i ) {
        UChar c= field.charAt(i);
        if ( (c==dtptngen_SINGLE_QUOTE) || (c==dtptngen_BACKSLASH) || (c==dtptngen_SPACE) || (c==dtptngen_COLON) ||
             (c==dtptngen_QUOTATION_MARK) || (c==dtptngen_COMMA) || (c==dtptngen_HYPHEN) ||(items[i].charAt(0)==dtptngen_DOT) ) {
            continue;
        }
        else {
            return FALSE;
        }
    }
    return TRUE;
}

DistanceInfo::~DistanceInfo() {}

void
DistanceInfo::setTo(const DistanceInfo& other) {
    missingFieldMask = other.missingFieldMask;
    extraFieldMask= other.extraFieldMask;
}

PatternMapIterator::PatternMapIterator(UErrorCode& status) :
    bootIndex(0), nodePtr(nullptr), matcher(nullptr), patternMap(nullptr)
{
    if (U_FAILURE(status)) { return; }
    matcher.adoptInsteadAndCheckErrorCode(new DateTimeMatcher(), status);
}

PatternMapIterator::~PatternMapIterator() {
}

void
PatternMapIterator::set(PatternMap& newPatternMap) {
    this->patternMap=&newPatternMap;
}

PtnSkeleton*
PatternMapIterator::getSkeleton() const {
    if ( nodePtr == nullptr ) {
        return nullptr;
    }
    else {
        return nodePtr->skeleton.getAlias();
    }
}

UBool
PatternMapIterator::hasNext() const {
    int32_t headIndex = bootIndex;
    PtnElem *curPtr = nodePtr;

    if (patternMap==nullptr) {
        return FALSE;
    }
    while ( headIndex < dtptngen_MAX_PATTERN_ENTRIES ) {
        if ( curPtr != nullptr ) {
            if ( curPtr->next != nullptr ) {
                return TRUE;
            }
            else {
                headIndex++;
                curPtr=nullptr;
                continue;
            }
        }
        else {
            if ( patternMap->boot[headIndex] != nullptr ) {
                return TRUE;
            }
            else {
                headIndex++;
                continue;
            }
        }
    }
    return FALSE;
}

DateTimeMatcher&
PatternMapIterator::next() {
    while ( bootIndex < dtptngen_MAX_PATTERN_ENTRIES ) {
        if ( nodePtr != nullptr ) {
            if ( nodePtr->next != nullptr ) {
                nodePtr = nodePtr->next.getAlias();
                break;
            }
            else {
                bootIndex++;
                nodePtr=nullptr;
                continue;
            }
        }
        else {
            if ( patternMap->boot[bootIndex] != nullptr ) {
                nodePtr = patternMap->boot[bootIndex];
                break;
            }
            else {
                bootIndex++;
                continue;
            }
        }
    }
    if (nodePtr!=nullptr) {
        matcher->copyFrom(*nodePtr->skeleton);
    }
    else {
        matcher->copyFrom();
    }
    return *matcher;
}


SkeletonFields::SkeletonFields() {
    // Set initial values to zero
    clear();
}

void SkeletonFields::clear() {
    uprv_memset(chars, 0, sizeof(chars));
    uprv_memset(lengths, 0, sizeof(lengths));
}

void SkeletonFields::copyFrom(const SkeletonFields& other) {
    uprv_memcpy(chars, other.chars, sizeof(chars));
    uprv_memcpy(lengths, other.lengths, sizeof(lengths));
}

void SkeletonFields::clearField(int32_t field) {
    chars[field] = 0;
    lengths[field] = 0;
}

UChar SkeletonFields::getFieldChar(int32_t field) const {
    return chars[field];
}

int32_t SkeletonFields::getFieldLength(int32_t field) const {
    return lengths[field];
}

void SkeletonFields::populate(int32_t field, const UnicodeString& value) {
    populate(field, value.charAt(0), value.length());
}

void SkeletonFields::populate(int32_t field, UChar ch, int32_t length) {
    chars[field] = (int8_t) ch;
    lengths[field] = (int8_t) length;
}

UBool SkeletonFields::isFieldEmpty(int32_t field) const {
    return lengths[field] == 0;
}

UnicodeString& SkeletonFields::appendTo(UnicodeString& string) const {
    for (int32_t i = 0; i < UDATPG_FIELD_COUNT; ++i) {
        appendFieldTo(i, string);
    }
    return string;
}

UnicodeString& SkeletonFields::appendFieldTo(int32_t field, UnicodeString& string) const {
    UChar ch(chars[field]);
    int32_t length = (int32_t) lengths[field];

    for (int32_t i=0; i<length; i++) {
        string += ch;
    }
    return string;
}

UChar SkeletonFields::getFirstChar() const {
    for (int32_t i = 0; i < UDATPG_FIELD_COUNT; ++i) {
        if (lengths[i] != 0) {
            return chars[i];
        }
    }
    return '\0';
}


PtnSkeleton::PtnSkeleton()
    : addedDefaultDayPeriod(FALSE) {
}

PtnSkeleton::PtnSkeleton(const PtnSkeleton& other) {
    copyFrom(other);
}

void PtnSkeleton::copyFrom(const PtnSkeleton& other) {
    uprv_memcpy(type, other.type, sizeof(type));
    original.copyFrom(other.original);
    baseOriginal.copyFrom(other.baseOriginal);
    addedDefaultDayPeriod = other.addedDefaultDayPeriod;
}

void PtnSkeleton::clear() {
    uprv_memset(type, 0, sizeof(type));
    original.clear();
    baseOriginal.clear();
}

UBool
PtnSkeleton::equals(const PtnSkeleton& other) const  {
    return (original == other.original)
        && (baseOriginal == other.baseOriginal)
        && (uprv_memcmp(type, other.type, sizeof(type)) == 0);
}

UnicodeString
PtnSkeleton::getSkeleton() const {
    UnicodeString result;
    result = original.appendTo(result);
    int32_t pos;
    if (addedDefaultDayPeriod && (pos = result.indexOf(dtptngen_LOW_A)) >= 0) {
        // for backward compatibility: if DateTimeMatcher.set added a single 'a' that
        // was not in the provided skeleton, remove it here before returning skeleton.
        result.remove(pos, 1);
    }
    return result;
}

UnicodeString
PtnSkeleton::getBaseSkeleton() const {
    UnicodeString result;
    result = baseOriginal.appendTo(result);
    int32_t pos;
    if (addedDefaultDayPeriod && (pos = result.indexOf(dtptngen_LOW_A)) >= 0) {
        // for backward compatibility: if DateTimeMatcher.set added a single 'a' that
        // was not in the provided skeleton, remove it here before returning skeleton.
        result.remove(pos, 1);
    }
    return result;
}

UChar
PtnSkeleton::getFirstChar() const {
    return baseOriginal.getFirstChar();
}

PtnSkeleton::~PtnSkeleton() {
}

PtnElem::PtnElem(const UnicodeString &basePat, const UnicodeString &pat) :
    basePattern(basePat), skeleton(nullptr), pattern(pat), next(nullptr)
{
}

PtnElem::~PtnElem() {
}

DTSkeletonEnumeration::DTSkeletonEnumeration(PatternMap& patternMap, dtStrEnum type, UErrorCode& status) : fSkeletons(nullptr) {
    PtnElem  *curElem;
    PtnSkeleton *curSkeleton;
    UnicodeString s;
    int32_t bootIndex;

    pos=0;
    fSkeletons.adoptInsteadAndCheckErrorCode(new UVector(status), status);
    if (U_FAILURE(status)) {
        return;
    }

    for (bootIndex=0; bootIndex<dtptngen_MAX_PATTERN_ENTRIES; ++bootIndex ) {
        curElem = patternMap.boot[bootIndex];
        while (curElem!=nullptr) {
            switch(type) {
                case DT_BASESKELETON:
                    s=curElem->basePattern;
                    break;
                case DT_PATTERN:
                    s=curElem->pattern;
                    break;
                case DT_SKELETON:
                    curSkeleton=curElem->skeleton.getAlias();
                    s=curSkeleton->getSkeleton();
                    break;
            }
            if ( !isCanonicalItem(s) ) {
                LocalPointer<UnicodeString> newElem(new UnicodeString(s), status);
                if (U_FAILURE(status)) {
                    return;
                }
                fSkeletons->addElement(newElem.getAlias(), status);
                if (U_FAILURE(status)) {
                    fSkeletons.adoptInstead(nullptr);
                    return;
                }
                newElem.orphan(); // fSkeletons vector now owns the UnicodeString.
            }
            curElem = curElem->next.getAlias();
        }
    }
    if ((bootIndex==dtptngen_MAX_PATTERN_ENTRIES) && (curElem!=nullptr) ) {
        status = U_BUFFER_OVERFLOW_ERROR;
    }
}

const UnicodeString*
DTSkeletonEnumeration::snext(UErrorCode& status) {
    if (U_SUCCESS(status) && fSkeletons.isValid() && pos < fSkeletons->size()) {
        return (const UnicodeString*)fSkeletons->elementAt(pos++);
    }
    return nullptr;
}

void
DTSkeletonEnumeration::reset(UErrorCode& /*status*/) {
    pos=0;
}

int32_t
DTSkeletonEnumeration::count(UErrorCode& /*status*/) const {
   return (fSkeletons.isNull()) ? 0 : fSkeletons->size();
}

UBool
DTSkeletonEnumeration::isCanonicalItem(const UnicodeString& item) {
    if ( item.length() != 1 ) {
        return FALSE;
    }
    for (int32_t i=0; i<UDATPG_FIELD_COUNT; ++i) {
        if (item.charAt(0)==Canonical_Items[i]) {
            return TRUE;
        }
    }
    return FALSE;
}

DTSkeletonEnumeration::~DTSkeletonEnumeration() {
    UnicodeString *s;
    if (fSkeletons.isValid()) {
        for (int32_t i = 0; i < fSkeletons->size(); ++i) {
            if ((s = (UnicodeString *)fSkeletons->elementAt(i)) != nullptr) {
                delete s;
            }
        }
    }
}

DTRedundantEnumeration::DTRedundantEnumeration() : pos(0), fPatterns(nullptr) {
}

void
DTRedundantEnumeration::add(const UnicodeString& pattern, UErrorCode& status) {
    if (U_FAILURE(status)) { return; }
    if (fPatterns.isNull())  {
        fPatterns.adoptInsteadAndCheckErrorCode(new UVector(status), status);
        if (U_FAILURE(status)) {
            return;
       }
    }
    LocalPointer<UnicodeString> newElem(new UnicodeString(pattern), status);
    if (U_FAILURE(status)) {
        return;
    }
    fPatterns->addElement(newElem.getAlias(), status);
    if (U_FAILURE(status)) {
        fPatterns.adoptInstead(nullptr);
        return;
    }
    newElem.orphan(); // fPatterns now owns the string.
}

const UnicodeString*
DTRedundantEnumeration::snext(UErrorCode& status) {
    if (U_SUCCESS(status) && fPatterns.isValid() && pos < fPatterns->size()) {
        return (const UnicodeString*)fPatterns->elementAt(pos++);
    }
    return nullptr;
}

void
DTRedundantEnumeration::reset(UErrorCode& /*status*/) {
    pos=0;
}

int32_t
DTRedundantEnumeration::count(UErrorCode& /*status*/) const {
    return (fPatterns.isNull()) ? 0 : fPatterns->size();
}

UBool
DTRedundantEnumeration::isCanonicalItem(const UnicodeString& item) const {
    if ( item.length() != 1 ) {
        return FALSE;
    }
    for (int32_t i=0; i<UDATPG_FIELD_COUNT; ++i) {
        if (item.charAt(0)==Canonical_Items[i]) {
            return TRUE;
        }
    }
    return FALSE;
}

DTRedundantEnumeration::~DTRedundantEnumeration() {
    UnicodeString *s;
    if (fPatterns.isValid()) {
        for (int32_t i = 0; i < fPatterns->size(); ++i) {
            if ((s = (UnicodeString *)fPatterns->elementAt(i)) != nullptr) {
                delete s;
            }
        }
    }
}

U_NAMESPACE_END


#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2012, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

  // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(DateTimeRule)

DateTimeRule::DateTimeRule(int32_t month,
                           int32_t dayOfMonth,
                           int32_t millisInDay,
                           TimeRuleType timeType)
: fMonth(month), fDayOfMonth(dayOfMonth), fDayOfWeek(0), fWeekInMonth(0), fMillisInDay(millisInDay),
  fDateRuleType(DateTimeRule::DOM), fTimeRuleType(timeType) {
}

DateTimeRule::DateTimeRule(int32_t month,
                           int32_t weekInMonth,
                           int32_t dayOfWeek,
                           int32_t millisInDay,
                           TimeRuleType timeType)
: fMonth(month), fDayOfMonth(0), fDayOfWeek(dayOfWeek), fWeekInMonth(weekInMonth), fMillisInDay(millisInDay),
  fDateRuleType(DateTimeRule::DOW), fTimeRuleType(timeType) {
}

DateTimeRule::DateTimeRule(int32_t month,
                           int32_t dayOfMonth,
                           int32_t dayOfWeek,
                           UBool after,
                           int32_t millisInDay,
                           TimeRuleType timeType)
: UObject(),
  fMonth(month), fDayOfMonth(dayOfMonth), fDayOfWeek(dayOfWeek), fWeekInMonth(0), fMillisInDay(millisInDay),
  fTimeRuleType(timeType) {
    if (after) {
        fDateRuleType = DateTimeRule::DOW_GEQ_DOM;
    } else {
        fDateRuleType = DateTimeRule::DOW_LEQ_DOM;
    }
}

DateTimeRule::DateTimeRule(const DateTimeRule& source)
: UObject(source),
  fMonth(source.fMonth), fDayOfMonth(source.fDayOfMonth), fDayOfWeek(source.fDayOfWeek),
  fWeekInMonth(source.fWeekInMonth), fMillisInDay(source.fMillisInDay),
  fDateRuleType(source.fDateRuleType), fTimeRuleType(source.fTimeRuleType) {
}

DateTimeRule::~DateTimeRule() {
}

DateTimeRule*
DateTimeRule::clone() const {
    return new DateTimeRule(*this);
}

DateTimeRule&
DateTimeRule::operator=(const DateTimeRule& right) {
    if (this != &right) {
        fMonth = right.fMonth;
        fDayOfMonth = right.fDayOfMonth;
        fDayOfWeek = right.fDayOfWeek;
        fWeekInMonth = right.fWeekInMonth;
        fMillisInDay = right.fMillisInDay;
        fDateRuleType = right.fDateRuleType;
        fTimeRuleType = right.fTimeRuleType;
    }
    return *this;
}

UBool
DateTimeRule::operator==(const DateTimeRule& that) const {
    return ((this == &that) ||
            (typeid(*this) == typeid(that) &&
            fMonth == that.fMonth &&
            fDayOfMonth == that.fDayOfMonth &&
            fDayOfWeek == that.fDayOfWeek &&
            fWeekInMonth == that.fWeekInMonth &&
            fMillisInDay == that.fMillisInDay &&
            fDateRuleType == that.fDateRuleType &&
            fTimeRuleType == that.fTimeRuleType));
}

UBool
DateTimeRule::operator!=(const DateTimeRule& that) const {
    return !operator==(that);
}

DateTimeRule::DateRuleType
DateTimeRule::getDateRuleType(void) const {
    return fDateRuleType;
}

DateTimeRule::TimeRuleType
DateTimeRule::getTimeRuleType(void) const {
    return fTimeRuleType;
}

int32_t
DateTimeRule::getRuleMonth(void) const {
    return fMonth;
}

int32_t
DateTimeRule::getRuleDayOfMonth(void) const {
    return fDayOfMonth;
}

int32_t
DateTimeRule::getRuleDayOfWeek(void) const {
    return fDayOfWeek;
}

int32_t
DateTimeRule::getRuleWeekInMonth(void) const {
    return fWeekInMonth;
}

int32_t
DateTimeRule::getRuleMillisInDay(void) const {
    return fMillisInDay;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html

#include <utility>



#if !UCONFIG_NO_FORMATTING

#include <stdlib.h>






// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html




#if !UCONFIG_NO_FORMATTING





U_NAMESPACE_BEGIN

// Export an explicit template instantiation of LocalMemory used as a data member of EraRules.
// When building DLLs for Windows this is required even though no direct access leaks out of the i18n library.
// See digitlst.h, pluralaffix.h, datefmt.h, and others for similar examples.
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN
#if defined(_MSC_VER)
// Ignore warning 4661 as LocalPointerBase does not use operator== or operator!=
#pragma warning(push)
#pragma warning(disable: 4661)
#endif
template class U_I18N_API LocalPointerBase<int32_t>;
template class U_I18N_API LocalMemory<int32_t>;
#if defined(_MSC_VER)
#pragma warning(pop)
#endif
#endif

class U_I18N_API EraRules : public UMemory {
public:
    ~EraRules();

    static EraRules* createInstance(const char *calType, UBool includeTentativeEra, UErrorCode& status);

    /**
     * Gets number of effective eras
     * @return  number of effective eras
     */
    inline int32_t getNumberOfEras() const {
        return numEras;
    }

    /**
     * Gets start date of an era
     * @param eraIdx    Era index
     * @param fields    Receives date fields. The result includes values of year, month,
     *                  day of month in this order. When an era has no start date, the result
     *                  will be January 1st in year whose value is minimum integer.
     * @param status    Receives status.
     */
    void getStartDate(int32_t eraIdx, int32_t (&fields)[3], UErrorCode& status) const;

    /**
     * Gets start year of an era
     * @param eraIdx    Era index
     * @param status    Receives status.
     * @return The first year of an era. When a era has no start date, minimum int32
     *          value is returned.
     */
    int32_t getStartYear(int32_t eraIdx, UErrorCode& status) const;

    /**
     * Returns era index for the specified year/month/day.
     * @param year  Year
     * @param month Month (1-base)
     * @param day   Day of month
     * @param status    Receives status
     * @return  era index (or 0, when the specified date is before the first era)
     */
    int32_t getEraIndex(int32_t year, int32_t month, int32_t day, UErrorCode& status) const;

    /**
     * Gets the current era index. This is calculated only once for an instance of
     * EraRules. The current era calculation is based on the default time zone at
     * the time of instantiation.
     *
     * @return era index of current era (or 0, when current date is before the first era)
     */
    inline int32_t getCurrentEraIndex() const {
        return currentEra;
    }

private:
    EraRules(LocalMemory<int32_t>& eraStartDates, int32_t numEra);

    void initCurrentEra();

    LocalMemory<int32_t> startDates;
    int32_t numEras;
    int32_t currentEra;
};

U_NAMESPACE_END
#endif /* #if !UCONFIG_NO_FORMATTING */




U_NAMESPACE_BEGIN

static const int32_t MAX_ENCODED_START_YEAR = 32767;
static const int32_t MIN_ENCODED_START_YEAR = -32768;
static const int32_t MIN_ENCODED_START = -2147483391;   // encodeDate(MIN_ENCODED_START_YEAR, 1, 1, ...);

static const int32_t YEAR_MASK = 0xFFFF0000;
static const int32_t MONTH_MASK = 0x0000FF00;
static const int32_t DAY_MASK = 0x000000FF;

static const int32_t MAX_INT32 = 0x7FFFFFFF;
static const int32_t MIN_INT32 = 0xFFFFFFFF;

static const UChar VAL_FALSE[] = {0x66, 0x61, 0x6c, 0x73, 0x65};    // "false"
static const UChar VAL_FALSE_LEN = 5;

static UBool isSet(int startDate) {
    return startDate != 0;
}

static UBool isValidRuleStartDate(int32_t year, int32_t month, int32_t day) {
    return year >= MIN_ENCODED_START_YEAR && year <= MAX_ENCODED_START_YEAR
            && month >= 1 && month <= 12 && day >=1 && day <= 31;
}

/**
 * Encode year/month/date to a single integer.
 * year is high 16 bits (-32768 to 32767), month is
 * next 8 bits and day of month is last 8 bits.
 *
 * @param year  year
 * @param month month (1-base)
 * @param day   day of month
 * @return  an encoded date.
 */
static int32_t encodeDate(int32_t year, int32_t month, int32_t day) {
    return year << 16 | month << 8 | day;
}

static void decodeDate(int32_t encodedDate, int32_t (&fields)[3]) {
    if (encodedDate == MIN_ENCODED_START) {
        fields[0] = MIN_INT32;
        fields[1] = 1;
        fields[2] = 1;
    } else {
        fields[0] = (encodedDate & YEAR_MASK) >> 16;
        fields[1] = (encodedDate & MONTH_MASK) >> 8;
        fields[2] = encodedDate & DAY_MASK;
    }
}

/**
 * Compare an encoded date with another date specified by year/month/day.
 * @param encoded   An encoded date
 * @param year      Year of another date
 * @param month     Month of another date
 * @param day       Day of another date
 * @return -1 when encoded date is earlier, 0 when two dates are same,
 *          and 1 when encoded date is later.
 */
static int32_t compareEncodedDateWithYMD(int encoded, int year, int month, int day) {
    if (year < MIN_ENCODED_START_YEAR) {
        if (encoded == MIN_ENCODED_START) {
            if (year > MIN_INT32 || month > 1 || day > 1) {
                return -1;
            }
            return 0;
        } else {
            return 1;
        }
    } else if (year > MAX_ENCODED_START_YEAR) {
        return -1;
    } else {
        int tmp = encodeDate(year, month, day);
        if (encoded < tmp) {
            return -1;
        } else if (encoded == tmp) {
            return 0;
        } else {
            return 1;
        }
    }
}

EraRules::EraRules(LocalMemory<int32_t>& eraStartDates, int32_t numEras)
    : numEras(numEras) {
    startDates = std::move(eraStartDates);
    initCurrentEra();
}

EraRules::~EraRules() {
}

EraRules* EraRules::createInstance(const char *calType, UBool includeTentativeEra, UErrorCode& status) {
    if(U_FAILURE(status)) {
        return nullptr;
    }
    LocalUResourceBundlePointer rb(ures_openDirect(nullptr, "supplementalData", &status));
    ures_getByKey(rb.getAlias(), "calendarData", rb.getAlias(), &status);
    ures_getByKey(rb.getAlias(), calType, rb.getAlias(), &status);
    ures_getByKey(rb.getAlias(), "eras", rb.getAlias(), &status);

    if (U_FAILURE(status)) {
        return nullptr;
    }

    int32_t numEras = ures_getSize(rb.getAlias());
    int32_t firstTentativeIdx = MAX_INT32;

    LocalMemory<int32_t> startDates(static_cast<int32_t *>(uprv_malloc(numEras * sizeof(int32_t))));
    if (startDates.isNull()) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    uprv_memset(startDates.getAlias(), 0 , numEras * sizeof(int32_t));

    while (ures_hasNext(rb.getAlias())) {
        LocalUResourceBundlePointer eraRuleRes(ures_getNextResource(rb.getAlias(), nullptr, &status));
        if (U_FAILURE(status)) {
            return nullptr;
        }
        const char *eraIdxStr = ures_getKey(eraRuleRes.getAlias());
        char *endp;
        int32_t eraIdx = (int32_t)strtol(eraIdxStr, &endp, 10);
        if ((size_t)(endp - eraIdxStr) != uprv_strlen(eraIdxStr)) {
            status = U_INVALID_FORMAT_ERROR;
            return nullptr;
        }
        if (eraIdx < 0 || eraIdx >= numEras) {
            status = U_INVALID_FORMAT_ERROR;
            return nullptr;
        }
        if (isSet(startDates[eraIdx])) {
            // start date of the index was already set
            status = U_INVALID_FORMAT_ERROR;
            return nullptr;
        }

        UBool hasName = TRUE;
        UBool hasEnd = TRUE;
        int32_t len;
        while (ures_hasNext(eraRuleRes.getAlias())) {
            LocalUResourceBundlePointer res(ures_getNextResource(eraRuleRes.getAlias(), nullptr, &status));
            if (U_FAILURE(status)) {
                return nullptr;
            }
            const char *key = ures_getKey(res.getAlias());
            if (uprv_strcmp(key, "start") == 0) {
                const int32_t *fields = ures_getIntVector(res.getAlias(), &len, &status);
                if (U_FAILURE(status)) {
                    return nullptr;
                }
                if (len != 3 || !isValidRuleStartDate(fields[0], fields[1], fields[2])) {
                    status = U_INVALID_FORMAT_ERROR;
                    return nullptr;
                }
                startDates[eraIdx] = encodeDate(fields[0], fields[1], fields[2]);
            } else if (uprv_strcmp(key, "named") == 0) {
                const UChar *val = ures_getString(res.getAlias(), &len, &status);
                if (u_strncmp(val, VAL_FALSE, VAL_FALSE_LEN) == 0) {
                    hasName = FALSE;
                }
            } else if (uprv_strcmp(key, "end") == 0) {
                hasEnd = TRUE;
            }
        }

        if (isSet(startDates[eraIdx])) {
            if (hasEnd) {
                // This implementation assumes either start or end is available, not both.
                // For now, just ignore the end rule.
            }
        } else {
            if (hasEnd) {
                if (eraIdx != 0) {
                    // This implementation does not support end only rule for eras other than
                    // the first one.
                    status = U_INVALID_FORMAT_ERROR;
                    return nullptr;
                }
                U_ASSERT(eraIdx == 0);
                startDates[eraIdx] = MIN_ENCODED_START;
            } else {
                status = U_INVALID_FORMAT_ERROR;
                return nullptr;
            }
        }

        if (hasName) {
            if (eraIdx >= firstTentativeIdx) {
                status = U_INVALID_FORMAT_ERROR;
                return nullptr;
            }
        } else {
            if (eraIdx < firstTentativeIdx) {
                firstTentativeIdx = eraIdx;
            }
        }
    }

    EraRules *result;
    if (firstTentativeIdx < MAX_INT32 && !includeTentativeEra) {
        result = new EraRules(startDates, firstTentativeIdx);
    } else {
        result = new EraRules(startDates, numEras);
    }

    if (result == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return result;
}

void EraRules::getStartDate(int32_t eraIdx, int32_t (&fields)[3], UErrorCode& status) const {
    if(U_FAILURE(status)) {
        return;
    }
    if (eraIdx < 0 || eraIdx >= numEras) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    decodeDate(startDates[eraIdx], fields);
}

int32_t EraRules::getStartYear(int32_t eraIdx, UErrorCode& status) const {
    int year = MAX_INT32;   // bogus value
    if(U_FAILURE(status)) {
        return year;
    }
    if (eraIdx < 0 || eraIdx >= numEras) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return year;
    }
    int fields[3];
    decodeDate(startDates[eraIdx], fields);
    year = fields[0];

    return year;
}

int32_t EraRules::getEraIndex(int32_t year, int32_t month, int32_t day, UErrorCode& status) const {
    if(U_FAILURE(status)) {
        return -1;
    }

    if (month < 1 || month > 12 || day < 1 || day > 31) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }
    int32_t high = numEras; // last index + 1
    int32_t low;

    // Short circuit for recent years.  Most modern computations will
    // occur in the last few eras.
    if (compareEncodedDateWithYMD(startDates[getCurrentEraIndex()], year, month, day) <= 0) {
        low = getCurrentEraIndex();
    } else {
        low = 0;
    }

    // Do binary search
    while (low < high - 1) {
        int i = (low + high) / 2;
        if (compareEncodedDateWithYMD(startDates[i], year, month, day) <= 0) {
            low = i;
        } else {
            high = i;
        }
    }
    return low;
}

void EraRules::initCurrentEra() {
    // Compute local wall time in millis using ICU's default time zone.
    UErrorCode ec = U_ZERO_ERROR;
    UDate localMillis = ucal_getNow();

    int32_t rawOffset, dstOffset;
    TimeZone* zone = TimeZone::createDefault();
    // If we failed to create the default time zone, we are in a bad state and don't
    // really have many options. Carry on using UTC millis as a fallback.
    if (zone != nullptr) {
        zone->getOffset(localMillis, FALSE, rawOffset, dstOffset, ec);
        delete zone;
        localMillis += (rawOffset + dstOffset);
    }

    int year, month0, dom, dow, doy, mid;
    Grego::timeToFields(localMillis, year, month0, dom, dow, doy, mid);
    int currentEncodedDate = encodeDate(year, month0 + 1 /* changes to 1-base */, dom);
    int eraIdx = numEras - 1;
    while (eraIdx > 0) {
        if (currentEncodedDate >= startDates[eraIdx]) {
            break;
        }
        eraIdx--;
    }
    // Note: current era could be before the first era.
    // In this case, this implementation returns the first era index (0).
    currentEra = eraIdx;
}

U_NAMESPACE_END
#endif /* #if !UCONFIG_NO_FORMATTING */


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2003 - 2013, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING




#include <float.h>

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(EthiopicCalendar)

//static const int32_t JD_EPOCH_OFFSET_AMETE_ALEM = -285019;
static const int32_t JD_EPOCH_OFFSET_AMETE_MIHRET = 1723856;
static const int32_t AMETE_MIHRET_DELTA = 5500; // 5501 - 1 (Amete Alem 5501 = Amete Mihret 1)

//-------------------------------------------------------------------------
// Constructors...
//-------------------------------------------------------------------------

EthiopicCalendar::EthiopicCalendar(const Locale& aLocale,
                                   UErrorCode& success,
                                   EEraType type /*= AMETE_MIHRET_ERA*/)
:   CECalendar(aLocale, success),
    eraType(type)
{
}

EthiopicCalendar::EthiopicCalendar(const EthiopicCalendar& other)
:   CECalendar(other),
    eraType(other.eraType)
{
}

EthiopicCalendar::~EthiopicCalendar()
{
}

EthiopicCalendar*
EthiopicCalendar::clone() const
{
    return new EthiopicCalendar(*this);
}

const char *
EthiopicCalendar::getType() const
{
    if (isAmeteAlemEra()) {
        return "ethiopic-amete-alem";
    }
    return "ethiopic";
}

void
EthiopicCalendar::setAmeteAlemEra(UBool onOff)
{
    eraType = onOff ? AMETE_ALEM_ERA : AMETE_MIHRET_ERA;
}

UBool
EthiopicCalendar::isAmeteAlemEra() const
{
    return (eraType == AMETE_ALEM_ERA);
}

//-------------------------------------------------------------------------
// Calendar framework
//-------------------------------------------------------------------------

int32_t
EthiopicCalendar::handleGetExtendedYear()
{
    // Ethiopic calendar uses EXTENDED_YEAR aligned to
    // Amelete Hihret year always.
    int32_t eyear;
    if (newerField(UCAL_EXTENDED_YEAR, UCAL_YEAR) == UCAL_EXTENDED_YEAR) {
        eyear = internalGet(UCAL_EXTENDED_YEAR, 1); // Default to year 1
    } else if (isAmeteAlemEra()) {
        eyear = internalGet(UCAL_YEAR, 1 + AMETE_MIHRET_DELTA)
            - AMETE_MIHRET_DELTA; // Default to year 1 of Amelete Mihret
    } else {
        // The year defaults to the epoch start, the era to AMETE_MIHRET
        int32_t era = internalGet(UCAL_ERA, AMETE_MIHRET);
        if (era == AMETE_MIHRET) {
            eyear = internalGet(UCAL_YEAR, 1); // Default to year 1
        } else {
            eyear = internalGet(UCAL_YEAR, 1) - AMETE_MIHRET_DELTA;
        }
    }
    return eyear;
}

void
EthiopicCalendar::handleComputeFields(int32_t julianDay, UErrorCode &/*status*/)
{
    int32_t eyear, month, day, era, year;
    jdToCE(julianDay, getJDEpochOffset(), eyear, month, day);

    if (isAmeteAlemEra()) {
        era = AMETE_ALEM;
        year = eyear + AMETE_MIHRET_DELTA;
    } else {
        if (eyear > 0) {
            era = AMETE_MIHRET;
            year = eyear;
        } else {
            era = AMETE_ALEM;
            year = eyear + AMETE_MIHRET_DELTA;
        }
    }

    internalSet(UCAL_EXTENDED_YEAR, eyear);
    internalSet(UCAL_ERA, era);
    internalSet(UCAL_YEAR, year);
    internalSet(UCAL_MONTH, month);
    internalSet(UCAL_DATE, day);
    internalSet(UCAL_DAY_OF_YEAR, (30 * month) + day);
}

int32_t
EthiopicCalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const
{
    if (isAmeteAlemEra() && field == UCAL_ERA) {
        return 0; // Only one era in this mode, era is always 0
    }
    return CECalendar::handleGetLimit(field, limitType);
}

/**
 * The system maintains a static default century start date and Year.  They are
 * initialized the first time they are used.  Once the system default century date
 * and year are set, they do not change.
 */
static UDate           gethpccal_gSystemDefaultCenturyStart       = DBL_MIN;
static int32_t         gethpccal_gSystemDefaultCenturyStartYear   = -1;
static icu::UInitOnce  gethpccal_gSystemDefaultCenturyInit        = U_INITONCE_INITIALIZER;

static void U_CALLCONV ethpccal_initializeSystemDefaultCentury()
{
    UErrorCode status = U_ZERO_ERROR;
    EthiopicCalendar calendar(Locale("@calendar=ethiopic"), status);
    if (U_SUCCESS(status)) {
        calendar.setTime(Calendar::getNow(), status);
        calendar.add(UCAL_YEAR, -80, status);

        gethpccal_gSystemDefaultCenturyStart = calendar.getTime(status);
        gethpccal_gSystemDefaultCenturyStartYear = calendar.get(UCAL_YEAR, status);
    }
    // We have no recourse upon failure unless we want to propagate the failure
    // out.
}

UDate
EthiopicCalendar::defaultCenturyStart() const
{
    // lazy-evaluate systemDefaultCenturyStart
    umtx_initOnce(gethpccal_gSystemDefaultCenturyInit, &ethpccal_initializeSystemDefaultCentury);
    return gethpccal_gSystemDefaultCenturyStart;
}

int32_t
EthiopicCalendar::defaultCenturyStartYear() const
{
    // lazy-evaluate systemDefaultCenturyStartYear
    umtx_initOnce(gethpccal_gSystemDefaultCenturyInit, &ethpccal_initializeSystemDefaultCentury);
    if (isAmeteAlemEra()) {
        return gethpccal_gSystemDefaultCenturyStartYear + AMETE_MIHRET_DELTA;
    }
    return gethpccal_gSystemDefaultCenturyStartYear;
}


int32_t
EthiopicCalendar::getJDEpochOffset() const
{
    return JD_EPOCH_OFFSET_AMETE_MIHRET;
}


#if 0
// We do not want to introduce this API in ICU4C.
// It was accidentally introduced in ICU4J as a public API.

//-------------------------------------------------------------------------
// Calendar system Conversion methods...
//-------------------------------------------------------------------------

int32_t
EthiopicCalendar::ethiopicToJD(int32_t year, int32_t month, int32_t date)
{
    return ceToJD(year, month, date, JD_EPOCH_OFFSET_AMETE_MIHRET);
}
#endif

U_NAMESPACE_END

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File FMTABLE.CPP
*
* Modification History:
*
*   Date        Name        Description
*   03/25/97    clhuang     Initial Implementation.
********************************************************************************
*/



#if !UCONFIG_NO_FORMATTING

#include <cstdlib>
#include <math.h>








// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2014, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

/** 
 * Maximum int64_t value that can be stored in a double without chancing losing precision.
 *   IEEE doubles have 53 bits of mantissa, 10 bits exponent, 1 bit sign.
 *   IBM Mainframes have 56 bits of mantissa, 7 bits of base 16 exponent, 1 bit sign.
 * Define this constant to the smallest value from those for supported platforms.
 * @internal
 */
static const int64_t MAX_INT64_IN_DOUBLE = 0x001FFFFFFFFFFFFFLL;

U_NAMESPACE_END

#endif // #if !UCONFIG_NO_FORMATTING



// *****************************************************************************
// class Formattable
// *****************************************************************************

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(Formattable)

using number::impl::DecimalQuantity;


//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

// NOTE: As of 3.0, there are limitations to the UObject API.  It does
// not (yet) support cloning, operator=, nor operator==.  To
// work around this, I implement some simple inlines here.  Later
// these can be modified or removed.  [alan]

// NOTE: These inlines assume that all fObjects are in fact instances
// of the Measure class, which is true as of 3.0.  [alan]

// Return TRUE if *a == *b.
static inline UBool objectEquals(const UObject* a, const UObject* b) {
    // LATER: return *a == *b;
    return *((const Measure*) a) == *((const Measure*) b);
}

// Return a clone of *a.
static inline UObject* objectClone(const UObject* a) {
    // LATER: return a->clone();
    return ((const Measure*) a)->clone();
}

// Return TRUE if *a is an instance of Measure.
static inline UBool instanceOfMeasure(const UObject* a) {
    return dynamic_cast<const Measure*>(a) != NULL;
}

/**
 * Creates a new Formattable array and copies the values from the specified
 * original.
 * @param array the original array
 * @param count the original array count
 * @return the new Formattable array.
 */
static Formattable* createArrayCopy(const Formattable* array, int32_t count) {
    Formattable *result = new Formattable[count];
    if (result != NULL) {
        for (int32_t i=0; i<count; ++i)
            result[i] = array[i]; // Don't memcpy!
    }
    return result;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

/**
 * Set 'ec' to 'err' only if 'ec' is not already set to a failing UErrorCode.
 */
static void setError(UErrorCode& ec, UErrorCode err) {
    if (U_SUCCESS(ec)) {
        ec = err;
    }
}

//
//  Common initialization code, shared by constructors.
//  Put everything into a known state.
//
void  Formattable::init() {
    fValue.fInt64 = 0;
    fType = kLong;
    fDecimalStr = NULL;
    fDecimalQuantity = NULL;
    fBogus.setToBogus();
}

// -------------------------------------
// default constructor.
// Creates a formattable object with a long value 0.

Formattable::Formattable() {
    init();
}

// -------------------------------------
// Creates a formattable object with a Date instance.

Formattable::Formattable(UDate date, ISDATE /*isDate*/)
{
    init();
    fType = kDate;
    fValue.fDate = date;
}

// -------------------------------------
// Creates a formattable object with a double value.

Formattable::Formattable(double value)
{
    init();
    fType = kDouble;
    fValue.fDouble = value;
}

// -------------------------------------
// Creates a formattable object with an int32_t value.

Formattable::Formattable(int32_t value)
{
    init();
    fValue.fInt64 = value;
}

// -------------------------------------
// Creates a formattable object with an int64_t value.

Formattable::Formattable(int64_t value)
{
    init();
    fType = kInt64;
    fValue.fInt64 = value;
}

// -------------------------------------
// Creates a formattable object with a decimal number value from a string.

Formattable::Formattable(StringPiece number, UErrorCode &status) {
    init();
    setDecimalNumber(number, status);
}


// -------------------------------------
// Creates a formattable object with a UnicodeString instance.

Formattable::Formattable(const UnicodeString& stringToCopy)
{
    init();
    fType = kString;
    fValue.fString = new UnicodeString(stringToCopy);
}

// -------------------------------------
// Creates a formattable object with a UnicodeString* value.
// (adopting symantics)

Formattable::Formattable(UnicodeString* stringToAdopt)
{
    init();
    fType = kString;
    fValue.fString = stringToAdopt;
}

Formattable::Formattable(UObject* objectToAdopt)
{
    init();
    fType = kObject;
    fValue.fObject = objectToAdopt;
}

// -------------------------------------

Formattable::Formattable(const Formattable* arrayToCopy, int32_t count)
    :   UObject(), fType(kArray)
{
    init();
    fType = kArray;
    fValue.fArrayAndCount.fArray = createArrayCopy(arrayToCopy, count);
    fValue.fArrayAndCount.fCount = count;
}

// -------------------------------------
// copy constructor


Formattable::Formattable(const Formattable &source)
     :  UObject(*this)
{
    init();
    *this = source;
}

// -------------------------------------
// assignment operator

Formattable&
Formattable::operator=(const Formattable& source)
{
    if (this != &source)
    {
        // Disposes the current formattable value/setting.
        dispose();

        // Sets the correct data type for this value.
        fType = source.fType;
        switch (fType)
        {
        case kArray:
            // Sets each element in the array one by one and records the array count.
            fValue.fArrayAndCount.fCount = source.fValue.fArrayAndCount.fCount;
            fValue.fArrayAndCount.fArray = createArrayCopy(source.fValue.fArrayAndCount.fArray,
                                                           source.fValue.fArrayAndCount.fCount);
            break;
        case kString:
            // Sets the string value.
            fValue.fString = new UnicodeString(*source.fValue.fString);
            break;
        case kDouble:
            // Sets the double value.
            fValue.fDouble = source.fValue.fDouble;
            break;
        case kLong:
        case kInt64:
            // Sets the long value.
            fValue.fInt64 = source.fValue.fInt64;
            break;
        case kDate:
            // Sets the Date value.
            fValue.fDate = source.fValue.fDate;
            break;
        case kObject:
            fValue.fObject = objectClone(source.fValue.fObject);
            break;
        }

        UErrorCode status = U_ZERO_ERROR;
        if (source.fDecimalQuantity != NULL) {
          fDecimalQuantity = new DecimalQuantity(*source.fDecimalQuantity);
        }
        if (source.fDecimalStr != NULL) {
            fDecimalStr = new CharString(*source.fDecimalStr, status);
            if (U_FAILURE(status)) {
                delete fDecimalStr;
                fDecimalStr = NULL;
            }
        }
    }
    return *this;
}

// -------------------------------------

UBool
Formattable::operator==(const Formattable& that) const
{
    int32_t i;

    if (this == &that) return TRUE;

    // Returns FALSE if the data types are different.
    if (fType != that.fType) return FALSE;

    // Compares the actual data values.
    UBool equal = TRUE;
    switch (fType) {
    case kDate:
        equal = (fValue.fDate == that.fValue.fDate);
        break;
    case kDouble:
        equal = (fValue.fDouble == that.fValue.fDouble);
        break;
    case kLong:
    case kInt64:
        equal = (fValue.fInt64 == that.fValue.fInt64);
        break;
    case kString:
        equal = (*(fValue.fString) == *(that.fValue.fString));
        break;
    case kArray:
        if (fValue.fArrayAndCount.fCount != that.fValue.fArrayAndCount.fCount) {
            equal = FALSE;
            break;
        }
        // Checks each element for equality.
        for (i=0; i<fValue.fArrayAndCount.fCount; ++i) {
            if (fValue.fArrayAndCount.fArray[i] != that.fValue.fArrayAndCount.fArray[i]) {
                equal = FALSE;
                break;
            }
        }
        break;
    case kObject:
        if (fValue.fObject == NULL || that.fValue.fObject == NULL) {
            equal = FALSE;
        } else {
            equal = objectEquals(fValue.fObject, that.fValue.fObject);
        }
        break;
    }

    // TODO:  compare digit lists if numeric.
    return equal;
}

// -------------------------------------

Formattable::~Formattable()
{
    dispose();
}

// -------------------------------------

void Formattable::dispose()
{
    // Deletes the data value if necessary.
    switch (fType) {
    case kString:
        delete fValue.fString;
        break;
    case kArray:
        delete[] fValue.fArrayAndCount.fArray;
        break;
    case kObject:
        delete fValue.fObject;
        break;
    default:
        break;
    }

    fType = kLong;
    fValue.fInt64 = 0;

    delete fDecimalStr;
    fDecimalStr = NULL;

    delete fDecimalQuantity;
    fDecimalQuantity = NULL;
}

Formattable *
Formattable::clone() const {
    return new Formattable(*this);
}

// -------------------------------------
// Gets the data type of this Formattable object.
Formattable::Type
Formattable::getType() const
{
    return fType;
}

UBool
Formattable::isNumeric() const {
    switch (fType) {
    case kDouble:
    case kLong:
    case kInt64:
        return TRUE;
    default:
        return FALSE;
    }
}

// -------------------------------------
int32_t
//Formattable::getLong(UErrorCode* status) const
Formattable::getLong(UErrorCode& status) const
{
    if (U_FAILURE(status)) {
        return 0;
    }

    switch (fType) {
    case Formattable::kLong:
        return (int32_t)fValue.fInt64;
    case Formattable::kInt64:
        if (fValue.fInt64 > INT32_MAX) {
            status = U_INVALID_FORMAT_ERROR;
            return INT32_MAX;
        } else if (fValue.fInt64 < INT32_MIN) {
            status = U_INVALID_FORMAT_ERROR;
            return INT32_MIN;
        } else {
            return (int32_t)fValue.fInt64;
        }
    case Formattable::kDouble:
        if (fValue.fDouble > INT32_MAX) {
            status = U_INVALID_FORMAT_ERROR;
            return INT32_MAX;
        } else if (fValue.fDouble < INT32_MIN) {
            status = U_INVALID_FORMAT_ERROR;
            return INT32_MIN;
        } else {
            return (int32_t)fValue.fDouble; // loses fraction
        }
    case Formattable::kObject:
        if (fValue.fObject == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return 0;
        }
        // TODO Later replace this with instanceof call
        if (instanceOfMeasure(fValue.fObject)) {
            return ((const Measure*) fValue.fObject)->
                getNumber().getLong(status);
        }
        U_FALLTHROUGH;
    default:
        status = U_INVALID_FORMAT_ERROR;
        return 0;
    }
}

// -------------------------------------
// Maximum int that can be represented exactly in a double.  (53 bits)
//    Larger ints may be rounded to a near-by value as not all are representable.
// TODO:  move this constant elsewhere, possibly configure it for different
//        floating point formats, if any non-standard ones are still in use.
static const int64_t U_DOUBLE_MAX_EXACT_INT = 9007199254740992LL;

int64_t
Formattable::getInt64(UErrorCode& status) const
{
    if (U_FAILURE(status)) {
        return 0;
    }

    switch (fType) {
    case Formattable::kLong:
    case Formattable::kInt64:
        return fValue.fInt64;
    case Formattable::kDouble:
        if (fValue.fDouble > (double)U_INT64_MAX) {
            status = U_INVALID_FORMAT_ERROR;
            return U_INT64_MAX;
        } else if (fValue.fDouble < (double)U_INT64_MIN) {
            status = U_INVALID_FORMAT_ERROR;
            return U_INT64_MIN;
        } else if (fabs(fValue.fDouble) > U_DOUBLE_MAX_EXACT_INT && fDecimalQuantity != NULL) {
            if (fDecimalQuantity->fitsInLong(true)) {
                return fDecimalQuantity->toLong();
            } else {
                // Unexpected
                status = U_INVALID_FORMAT_ERROR;
                return fDecimalQuantity->isNegative() ? U_INT64_MIN : U_INT64_MAX;
            }
        } else {
            return (int64_t)fValue.fDouble;
        }
    case Formattable::kObject:
        if (fValue.fObject == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return 0;
        }
        if (instanceOfMeasure(fValue.fObject)) {
            return ((const Measure*) fValue.fObject)->
                getNumber().getInt64(status);
        }
        U_FALLTHROUGH;
    default:
        status = U_INVALID_FORMAT_ERROR;
        return 0;
    }
}

// -------------------------------------
double
Formattable::getDouble(UErrorCode& status) const
{
    if (U_FAILURE(status)) {
        return 0;
    }

    switch (fType) {
    case Formattable::kLong:
    case Formattable::kInt64: // loses precision
        return (double)fValue.fInt64;
    case Formattable::kDouble:
        return fValue.fDouble;
    case Formattable::kObject:
        if (fValue.fObject == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return 0;
        }
        // TODO Later replace this with instanceof call
        if (instanceOfMeasure(fValue.fObject)) {
            return ((const Measure*) fValue.fObject)->
                getNumber().getDouble(status);
        }
        U_FALLTHROUGH;
    default:
        status = U_INVALID_FORMAT_ERROR;
        return 0;
    }
}

const UObject*
Formattable::getObject() const {
    return (fType == kObject) ? fValue.fObject : NULL;
}

// -------------------------------------
// Sets the value to a double value d.

void
Formattable::setDouble(double d)
{
    dispose();
    fType = kDouble;
    fValue.fDouble = d;
}

// -------------------------------------
// Sets the value to a long value l.

void
Formattable::setLong(int32_t l)
{
    dispose();
    fType = kLong;
    fValue.fInt64 = l;
}

// -------------------------------------
// Sets the value to an int64 value ll.

void
Formattable::setInt64(int64_t ll)
{
    dispose();
    fType = kInt64;
    fValue.fInt64 = ll;
}

// -------------------------------------
// Sets the value to a Date instance d.

void
Formattable::setDate(UDate d)
{
    dispose();
    fType = kDate;
    fValue.fDate = d;
}

// -------------------------------------
// Sets the value to a string value stringToCopy.

void
Formattable::setString(const UnicodeString& stringToCopy)
{
    dispose();
    fType = kString;
    fValue.fString = new UnicodeString(stringToCopy);
}

// -------------------------------------
// Sets the value to an array of Formattable objects.

void
Formattable::setArray(const Formattable* array, int32_t count)
{
    dispose();
    fType = kArray;
    fValue.fArrayAndCount.fArray = createArrayCopy(array, count);
    fValue.fArrayAndCount.fCount = count;
}

// -------------------------------------
// Adopts the stringToAdopt value.

void
Formattable::adoptString(UnicodeString* stringToAdopt)
{
    dispose();
    fType = kString;
    fValue.fString = stringToAdopt;
}

// -------------------------------------
// Adopts the array value and its count.

void
Formattable::adoptArray(Formattable* array, int32_t count)
{
    dispose();
    fType = kArray;
    fValue.fArrayAndCount.fArray = array;
    fValue.fArrayAndCount.fCount = count;
}

void
Formattable::adoptObject(UObject* objectToAdopt) {
    dispose();
    fType = kObject;
    fValue.fObject = objectToAdopt;
}

// -------------------------------------
UnicodeString&
Formattable::getString(UnicodeString& result, UErrorCode& status) const
{
    if (fType != kString) {
        setError(status, U_INVALID_FORMAT_ERROR);
        result.setToBogus();
    } else {
        if (fValue.fString == NULL) {
            setError(status, U_MEMORY_ALLOCATION_ERROR);
        } else {
            result = *fValue.fString;
        }
    }
    return result;
}

// -------------------------------------
const UnicodeString&
Formattable::getString(UErrorCode& status) const
{
    if (fType != kString) {
        setError(status, U_INVALID_FORMAT_ERROR);
        return *getBogus();
    }
    if (fValue.fString == NULL) {
        setError(status, U_MEMORY_ALLOCATION_ERROR);
        return *getBogus();
    }
    return *fValue.fString;
}

// -------------------------------------
UnicodeString&
Formattable::getString(UErrorCode& status)
{
    if (fType != kString) {
        setError(status, U_INVALID_FORMAT_ERROR);
        return *getBogus();
    }
    if (fValue.fString == NULL) {
    	setError(status, U_MEMORY_ALLOCATION_ERROR);
    	return *getBogus();
    }
    return *fValue.fString;
}

// -------------------------------------
const Formattable*
Formattable::getArray(int32_t& count, UErrorCode& status) const
{
    if (fType != kArray) {
        setError(status, U_INVALID_FORMAT_ERROR);
        count = 0;
        return NULL;
    }
    count = fValue.fArrayAndCount.fCount;
    return fValue.fArrayAndCount.fArray;
}

// -------------------------------------
// Gets the bogus string, ensures mondo bogosity.

UnicodeString*
Formattable::getBogus() const
{
    return (UnicodeString*)&fBogus; /* cast away const :-( */
}


// --------------------------------------
StringPiece Formattable::getDecimalNumber(UErrorCode &status) {
    if (U_FAILURE(status)) {
        return "";
    }
    if (fDecimalStr != NULL) {
      return fDecimalStr->toStringPiece();
    }

    CharString *decimalStr = internalGetCharString(status);
    if(decimalStr == NULL) {
      return ""; // getDecimalNumber returns "" for error cases
    } else {
      return decimalStr->toStringPiece();
    }
}

CharString *Formattable::internalGetCharString(UErrorCode &status) {
    if(fDecimalStr == NULL) {
      if (fDecimalQuantity == NULL) {
        // No decimal number for the formattable yet.  Which means the value was
        // set directly by the user as an int, int64 or double.  If the value came
        // from parsing, or from the user setting a decimal number, fDecimalNum
        // would already be set.
        //
        LocalPointer<DecimalQuantity> dq(new DecimalQuantity(), status);
        if (U_FAILURE(status)) { return nullptr; }
        populateDecimalQuantity(*dq, status);
        if (U_FAILURE(status)) { return nullptr; }
        fDecimalQuantity = dq.orphan();
      }

      fDecimalStr = new CharString();
      if (fDecimalStr == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
      }
      // Older ICUs called uprv_decNumberToString here, which is not exactly the same as
      // DecimalQuantity::toScientificString(). The biggest difference is that uprv_decNumberToString does
      // not print scientific notation for magnitudes greater than -5 and smaller than some amount (+5?).
      if (fDecimalQuantity->isInfinite()) {
        fDecimalStr->append("Infinity", status);
      } else if (fDecimalQuantity->isNaN()) {
        fDecimalStr->append("NaN", status);
      } else if (fDecimalQuantity->isZeroish()) {
        fDecimalStr->append("0", -1, status);
      } else if (fType==kLong || fType==kInt64 || // use toPlainString for integer types
                  (fDecimalQuantity->getMagnitude() != INT32_MIN && std::abs(fDecimalQuantity->getMagnitude()) < 5)) {
        fDecimalStr->appendInvariantChars(fDecimalQuantity->toPlainString(), status);
      } else {
        fDecimalStr->appendInvariantChars(fDecimalQuantity->toScientificString(), status);
      }
    }
    return fDecimalStr;
}

void
Formattable::populateDecimalQuantity(number::impl::DecimalQuantity& output, UErrorCode& status) const {
    if (fDecimalQuantity != nullptr) {
        output = *fDecimalQuantity;
        return;
    }

    switch (fType) {
        case kDouble:
            output.setToDouble(this->getDouble());
            output.roundToInfinity();
            break;
        case kLong:
            output.setToInt(this->getLong());
            break;
        case kInt64:
            output.setToLong(this->getInt64());
            break;
        default:
            // The formattable's value is not a numeric type.
            status = U_INVALID_STATE_ERROR;
    }
}

// ---------------------------------------
void
Formattable::adoptDecimalQuantity(DecimalQuantity *dq) {
    if (fDecimalQuantity != NULL) {
        delete fDecimalQuantity;
    }
    fDecimalQuantity = dq;
    if (dq == NULL) { // allow adoptDigitList(NULL) to clear
        return;
    }

    // Set the value into the Union of simple type values.
    // Cannot use the set() functions because they would delete the fDecimalNum value.
    if (fDecimalQuantity->fitsInLong()) {
        fValue.fInt64 = fDecimalQuantity->toLong();
        if (fValue.fInt64 <= INT32_MAX && fValue.fInt64 >= INT32_MIN) {
            fType = kLong;
        } else {
            fType = kInt64;
        }
    } else {
        fType = kDouble;
        fValue.fDouble = fDecimalQuantity->toDouble();
    }
}


// ---------------------------------------
void
Formattable::setDecimalNumber(StringPiece numberString, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return;
    }
    dispose();

    auto* dq = new DecimalQuantity();
    dq->setToDecNumber(numberString, status);
    adoptDecimalQuantity(dq);

    // Note that we do not hang on to the caller's input string.
    // If we are asked for the string, we will regenerate one from fDecimalQuantity.
}

U_NAMESPACE_END

/* ---- UFormattable implementation ---- */

U_NAMESPACE_USE

U_DRAFT UFormattable* U_EXPORT2
ufmt_open(UErrorCode *status) {
  if( U_FAILURE(*status) ) {
    return NULL;
  }
  UFormattable *fmt = (new Formattable())->toUFormattable();

  if( fmt == NULL ) {
    *status = U_MEMORY_ALLOCATION_ERROR;
  }
  return fmt;
}

U_DRAFT void U_EXPORT2
ufmt_close(UFormattable *fmt) {
  Formattable *obj = Formattable::fromUFormattable(fmt);

  delete obj;
}

U_INTERNAL UFormattableType U_EXPORT2
ufmt_getType(const UFormattable *fmt, UErrorCode *status) {
  if(U_FAILURE(*status)) {
    return (UFormattableType)UFMT_COUNT;
  }
  const Formattable *obj = Formattable::fromUFormattable(fmt);
  return (UFormattableType)obj->getType();
}


U_INTERNAL UBool U_EXPORT2
ufmt_isNumeric(const UFormattable *fmt) {
  const Formattable *obj = Formattable::fromUFormattable(fmt);
  return obj->isNumeric();
}

U_DRAFT UDate U_EXPORT2
ufmt_getDate(const UFormattable *fmt, UErrorCode *status) {
  const Formattable *obj = Formattable::fromUFormattable(fmt);

  return obj->getDate(*status);
}

U_DRAFT double U_EXPORT2
ufmt_getDouble(UFormattable *fmt, UErrorCode *status) {
  Formattable *obj = Formattable::fromUFormattable(fmt);

  return obj->getDouble(*status);
}

U_DRAFT int32_t U_EXPORT2
ufmt_getLong(UFormattable *fmt, UErrorCode *status) {
  Formattable *obj = Formattable::fromUFormattable(fmt);

  return obj->getLong(*status);
}


U_DRAFT const void *U_EXPORT2
ufmt_getObject(const UFormattable *fmt, UErrorCode *status) {
  const Formattable *obj = Formattable::fromUFormattable(fmt);

  const void *ret = obj->getObject();
  if( ret==NULL &&
      (obj->getType() != Formattable::kObject) &&
      U_SUCCESS( *status )) {
    *status = U_INVALID_FORMAT_ERROR;
  }
  return ret;
}

U_DRAFT const UChar* U_EXPORT2
ufmt_getUChars(UFormattable *fmt, int32_t *len, UErrorCode *status) {
  Formattable *obj = Formattable::fromUFormattable(fmt);

  // avoid bogosity by checking the type first.
  if( obj->getType() != Formattable::kString ) {
    if( U_SUCCESS(*status) ){
      *status = U_INVALID_FORMAT_ERROR;
    }
    return NULL;
  }

  // This should return a valid string
  UnicodeString &str = obj->getString(*status);
  if( U_SUCCESS(*status) && len != NULL ) {
    *len = str.length();
  }
  return str.getTerminatedBuffer();
}

U_DRAFT int32_t U_EXPORT2
ufmt_getArrayLength(const UFormattable* fmt, UErrorCode *status) {
  const Formattable *obj = Formattable::fromUFormattable(fmt);

  int32_t count;
  (void)obj->getArray(count, *status);
  return count;
}

U_DRAFT UFormattable * U_EXPORT2
ufmt_getArrayItemByIndex(UFormattable* fmt, int32_t n, UErrorCode *status) {
  Formattable *obj = Formattable::fromUFormattable(fmt);
  int32_t count;
  (void)obj->getArray(count, *status);
  if(U_FAILURE(*status)) {
    return NULL;
  } else if(n<0 || n>=count) {
    setError(*status, U_INDEX_OUTOFBOUNDS_ERROR);
    return NULL;
  } else {
    return (*obj)[n].toUFormattable(); // returns non-const Formattable
  }
}

U_DRAFT const char * U_EXPORT2
ufmt_getDecNumChars(UFormattable *fmt, int32_t *len, UErrorCode *status) {
  if(U_FAILURE(*status)) {
    return "";
  }
  Formattable *obj = Formattable::fromUFormattable(fmt);
  CharString *charString = obj->internalGetCharString(*status);
  if(U_FAILURE(*status)) {
    return "";
  }
  if(charString == NULL) {
    *status = U_MEMORY_ALLOCATION_ERROR;
    return "";
  } else {
    if(len!=NULL) {
      *len = charString->length();
    }
    return charString->data();
  }
}

U_DRAFT int64_t U_EXPORT2
ufmt_getInt64(UFormattable *fmt, UErrorCode *status) {
  Formattable *obj = Formattable::fromUFormattable(fmt);
  return obj->getInt64(*status);
}

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2010, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* File FMTABLE.CPP
*
* Modification History:
*
*   Date        Name        Description
*   03/25/97    clhuang     Initial Implementation.
********************************************************************************
*/



#if !UCONFIG_NO_FORMATTING && !UCONFIG_NO_CONVERSION



// *****************************************************************************
// class Formattable
// *****************************************************************************

U_NAMESPACE_BEGIN

// -------------------------------------
// Creates a formattable object with a char* string.
// This API is useless. The API that takes a UnicodeString is actually just as good.
// This is just a grandfathered API.

Formattable::Formattable(const char* stringToCopy)
{
    init();
    fType = kString;
    fValue.fString = new UnicodeString(stringToCopy);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING || !UCONFIG_NO_CONVERSION */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2012, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* File FORMAT.CPP
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/17/97    clhuang     Implemented with new APIs.
*   03/27/97    helena      Updated to pass the simple test after code review.
*   07/20/98    stephen        Added explicit init values for Field/ParsePosition
********************************************************************************
*/
// *****************************************************************************
// This file was generated from the java source file Format.java
// *****************************************************************************

  // for 'typeid' to work



/*
 * Dummy code:
 * If all modules in the I18N library are switched off, then there are no
 * library exports and MSVC 6 writes a .dll but not a .lib file.
 * Unless we export _something_ in that case...
 */
#if UCONFIG_NO_COLLATION && UCONFIG_NO_FORMATTING && UCONFIG_NO_TRANSLITERATION
U_CAPI int32_t U_EXPORT2
uprv_icuin_lib_dummy(int32_t i) {
    return -i;
}
#endif

/* Format class implementation ---------------------------------------------- */

#if !UCONFIG_NO_FORMATTING






// *****************************************************************************
// class Format
// *****************************************************************************

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(FieldPosition)

FieldPosition::~FieldPosition() {}

FieldPosition *
FieldPosition::clone() const {
    return new FieldPosition(*this);
}

// -------------------------------------
// default constructor

Format::Format()
    : UObject()
{
    *validLocale = *actualLocale = 0;
}

// -------------------------------------

Format::~Format()
{
}

// -------------------------------------
// copy constructor

Format::Format(const Format &that)
    : UObject(that)
{
    *this = that;
}

// -------------------------------------
// assignment operator

Format&
Format::operator=(const Format& that)
{
    if (this != &that) {
        uprv_strcpy(validLocale, that.validLocale);
        uprv_strcpy(actualLocale, that.actualLocale);
    }
    return *this;
}

// -------------------------------------
// Formats the obj and append the result in the buffer, toAppendTo.
// This calls the actual implementation in the concrete subclasses.

UnicodeString&
Format::format(const Formattable& obj,
               UnicodeString& toAppendTo,
               UErrorCode& status) const
{
    if (U_FAILURE(status)) return toAppendTo;

    FieldPosition pos(FieldPosition::DONT_CARE);

    return format(obj, toAppendTo, pos, status);
}

// -------------------------------------
// Default implementation sets unsupported error; subclasses should
// override.

UnicodeString&
Format::format(const Formattable& /* unused obj */,
               UnicodeString& toAppendTo,
               FieldPositionIterator* /* unused posIter */,
               UErrorCode& status) const
{
    if (!U_FAILURE(status)) {
      status = U_UNSUPPORTED_ERROR;
    }
    return toAppendTo;
}

// -------------------------------------
// Parses the source string and create the corresponding
// result object.  Checks the parse position for errors.

void
Format::parseObject(const UnicodeString& source,
                    Formattable& result,
                    UErrorCode& status) const
{
    if (U_FAILURE(status)) return;

    ParsePosition parsePosition(0);
    parseObject(source, result, parsePosition);
    if (parsePosition.getIndex() == 0) {
        status = U_INVALID_FORMAT_ERROR;
    }
}

// -------------------------------------

UBool
Format::operator==(const Format& that) const
{
    // Subclasses: Call this method and then add more specific checks.
    return typeid(*this) == typeid(that);
}
//---------------------------------------

/**
 * Simple function for initializing a UParseError from a UnicodeString.
 *
 * @param pattern The pattern to copy into the parseError
 * @param pos The position in pattern where the error occured
 * @param parseError The UParseError object to fill in
 * @draft ICU 2.4
 */
void Format::syntaxError(const UnicodeString& pattern,
                         int32_t pos,
                         UParseError& parseError) {
    parseError.offset = pos;
    parseError.line=0;  // we are not using line number

    // for pre-context
    int32_t start = (pos < U_PARSE_CONTEXT_LEN)? 0 : (pos - (U_PARSE_CONTEXT_LEN-1
                                                             /* subtract 1 so that we have room for null*/));
    int32_t stop  = pos;
    pattern.extract(start,stop-start,parseError.preContext,0);
    //null terminate the buffer
    parseError.preContext[stop-start] = 0;

    //for post-context
    start = pos+1;
    stop  = ((pos+U_PARSE_CONTEXT_LEN)<=pattern.length()) ? (pos+(U_PARSE_CONTEXT_LEN-1)) :
        pattern.length();
    pattern.extract(start,stop-start,parseError.postContext,0);
    //null terminate the buffer
    parseError.postContext[stop-start]= 0;
}

Locale
Format::getLocale(ULocDataLocaleType type, UErrorCode& status) const {
    U_LOCALE_BASED(locBased, *this);
    return locBased.getLocale(type, status);
}

const char *
Format::getLocaleID(ULocDataLocaleType type, UErrorCode& status) const {
    U_LOCALE_BASED(locBased, *this);
    return locBased.getLocaleID(type, status);
}

void
Format::setLocaleIDs(const char* valid, const char* actual) {
    U_LOCALE_BASED(locBased, *this);
    locBased.setLocaleIDs(valid, actual);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING





namespace {

// A version of uprv_memcpy that checks for length 0.
// By default, uprv_memcpy requires a length of at least 1.
inline void uprv_memcpy2(void* dest, const void* src, size_t len) {
    if (len > 0) {
        uprv_memcpy(dest, src, len);
    }
}

// A version of uprv_memmove that checks for length 0.
// By default, uprv_memmove requires a length of at least 1.
inline void uprv_memmove2(void* dest, const void* src, size_t len) {
    if (len > 0) {
        uprv_memmove(dest, src, len);
    }
}

} // namespace


U_NAMESPACE_BEGIN

FormattedStringBuilder::FormattedStringBuilder() {
#if U_DEBUG
    // Initializing the memory to non-zero helps catch some bugs that involve
    // reading from an improperly terminated string.
    for (int32_t i=0; i<getCapacity(); i++) {
        getCharPtr()[i] = 1;
    }
#endif
}

FormattedStringBuilder::~FormattedStringBuilder() {
    if (fUsingHeap) {
        uprv_free(fChars.heap.ptr);
        uprv_free(fFields.heap.ptr);
    }
}

FormattedStringBuilder::FormattedStringBuilder(const FormattedStringBuilder &other) {
    *this = other;
}

FormattedStringBuilder &FormattedStringBuilder::operator=(const FormattedStringBuilder &other) {
    // Check for self-assignment
    if (this == &other) {
        return *this;
    }

    // Continue with deallocation and copying
    if (fUsingHeap) {
        uprv_free(fChars.heap.ptr);
        uprv_free(fFields.heap.ptr);
        fUsingHeap = false;
    }

    int32_t capacity = other.getCapacity();
    if (capacity > DEFAULT_CAPACITY) {
        // FIXME: uprv_malloc
        // C++ note: malloc appears in two places: here and in prepareForInsertHelper.
        auto newChars = static_cast<char16_t *> (uprv_malloc(sizeof(char16_t) * capacity));
        auto newFields = static_cast<Field *>(uprv_malloc(sizeof(Field) * capacity));
        if (newChars == nullptr || newFields == nullptr) {
            // UErrorCode is not available; fail silently.
            uprv_free(newChars);
            uprv_free(newFields);
            *this = FormattedStringBuilder();  // can't fail
            return *this;
        }

        fUsingHeap = true;
        fChars.heap.capacity = capacity;
        fChars.heap.ptr = newChars;
        fFields.heap.capacity = capacity;
        fFields.heap.ptr = newFields;
    }

    uprv_memcpy2(getCharPtr(), other.getCharPtr(), sizeof(char16_t) * capacity);
    uprv_memcpy2(getFieldPtr(), other.getFieldPtr(), sizeof(Field) * capacity);

    fZero = other.fZero;
    fLength = other.fLength;
    return *this;
}

int32_t FormattedStringBuilder::length() const {
    return fLength;
}

int32_t FormattedStringBuilder::codePointCount() const {
    return u_countChar32(getCharPtr() + fZero, fLength);
}

UChar32 FormattedStringBuilder::getFirstCodePoint() const {
    if (fLength == 0) {
        return -1;
    }
    UChar32 cp;
    U16_GET(getCharPtr() + fZero, 0, 0, fLength, cp);
    return cp;
}

UChar32 FormattedStringBuilder::getLastCodePoint() const {
    if (fLength == 0) {
        return -1;
    }
    int32_t offset = fLength;
    U16_BACK_1(getCharPtr() + fZero, 0, offset);
    UChar32 cp;
    U16_GET(getCharPtr() + fZero, 0, offset, fLength, cp);
    return cp;
}

UChar32 FormattedStringBuilder::codePointAt(int32_t index) const {
    UChar32 cp;
    U16_GET(getCharPtr() + fZero, 0, index, fLength, cp);
    return cp;
}

UChar32 FormattedStringBuilder::codePointBefore(int32_t index) const {
    int32_t offset = index;
    U16_BACK_1(getCharPtr() + fZero, 0, offset);
    UChar32 cp;
    U16_GET(getCharPtr() + fZero, 0, offset, fLength, cp);
    return cp;
}

FormattedStringBuilder &FormattedStringBuilder::clear() {
    // TODO: Reset the heap here?
    fZero = getCapacity() / 2;
    fLength = 0;
    return *this;
}

int32_t
FormattedStringBuilder::insertCodePoint(int32_t index, UChar32 codePoint, Field field, UErrorCode &status) {
    int32_t count = U16_LENGTH(codePoint);
    int32_t position = prepareForInsert(index, count, status);
    if (U_FAILURE(status)) {
        return count;
    }
    if (count == 1) {
        getCharPtr()[position] = (char16_t) codePoint;
        getFieldPtr()[position] = field;
    } else {
        getCharPtr()[position] = U16_LEAD(codePoint);
        getCharPtr()[position + 1] = U16_TRAIL(codePoint);
        getFieldPtr()[position] = getFieldPtr()[position + 1] = field;
    }
    return count;
}

int32_t FormattedStringBuilder::insert(int32_t index, const UnicodeString &unistr, Field field,
                                    UErrorCode &status) {
    if (unistr.length() == 0) {
        // Nothing to insert.
        return 0;
    } else if (unistr.length() == 1) {
        // Fast path: insert using insertCodePoint.
        return insertCodePoint(index, unistr.charAt(0), field, status);
    } else {
        return insert(index, unistr, 0, unistr.length(), field, status);
    }
}

int32_t
FormattedStringBuilder::insert(int32_t index, const UnicodeString &unistr, int32_t start, int32_t end,
                            Field field, UErrorCode &status) {
    int32_t count = end - start;
    int32_t position = prepareForInsert(index, count, status);
    if (U_FAILURE(status)) {
        return count;
    }
    for (int32_t i = 0; i < count; i++) {
        getCharPtr()[position + i] = unistr.charAt(start + i);
        getFieldPtr()[position + i] = field;
    }
    return count;
}

int32_t
FormattedStringBuilder::splice(int32_t startThis, int32_t endThis,  const UnicodeString &unistr,
                            int32_t startOther, int32_t endOther, Field field, UErrorCode& status) {
    int32_t thisLength = endThis - startThis;
    int32_t otherLength = endOther - startOther;
    int32_t count = otherLength - thisLength;
    int32_t position;
    if (count > 0) {
        // Overall, chars need to be added.
        position = prepareForInsert(startThis, count, status);
    } else {
        // Overall, chars need to be removed or kept the same.
        position = remove(startThis, -count);
    }
    if (U_FAILURE(status)) {
        return count;
    }
    for (int32_t i = 0; i < otherLength; i++) {
        getCharPtr()[position + i] = unistr.charAt(startOther + i);
        getFieldPtr()[position + i] = field;
    }
    return count;
}

int32_t FormattedStringBuilder::append(const FormattedStringBuilder &other, UErrorCode &status) {
    return insert(fLength, other, status);
}

int32_t
FormattedStringBuilder::insert(int32_t index, const FormattedStringBuilder &other, UErrorCode &status) {
    if (this == &other) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    int32_t count = other.fLength;
    if (count == 0) {
        // Nothing to insert.
        return 0;
    }
    int32_t position = prepareForInsert(index, count, status);
    if (U_FAILURE(status)) {
        return count;
    }
    for (int32_t i = 0; i < count; i++) {
        getCharPtr()[position + i] = other.charAt(i);
        getFieldPtr()[position + i] = other.fieldAt(i);
    }
    return count;
}

void FormattedStringBuilder::writeTerminator(UErrorCode& status) {
    int32_t position = prepareForInsert(fLength, 1, status);
    if (U_FAILURE(status)) {
        return;
    }
    getCharPtr()[position] = 0;
    getFieldPtr()[position] = UNUM_FIELD_COUNT;
    fLength--;
}

int32_t FormattedStringBuilder::prepareForInsert(int32_t index, int32_t count, UErrorCode &status) {
    U_ASSERT(index >= 0);
    U_ASSERT(index <= fLength);
    U_ASSERT(count >= 0);
    if (index == 0 && fZero - count >= 0) {
        // Append to start
        fZero -= count;
        fLength += count;
        return fZero;
    } else if (index == fLength && fZero + fLength + count < getCapacity()) {
        // Append to end
        fLength += count;
        return fZero + fLength - count;
    } else {
        // Move chars around and/or allocate more space
        return prepareForInsertHelper(index, count, status);
    }
}

int32_t FormattedStringBuilder::prepareForInsertHelper(int32_t index, int32_t count, UErrorCode &status) {
    int32_t oldCapacity = getCapacity();
    int32_t oldZero = fZero;
    char16_t *oldChars = getCharPtr();
    Field *oldFields = getFieldPtr();
    if (fLength + count > oldCapacity) {
        int32_t newCapacity = (fLength + count) * 2;
        int32_t newZero = newCapacity / 2 - (fLength + count) / 2;

        // C++ note: malloc appears in two places: here and in the assignment operator.
        auto newChars = static_cast<char16_t *> (uprv_malloc(sizeof(char16_t) * newCapacity));
        auto newFields = static_cast<Field *>(uprv_malloc(sizeof(Field) * newCapacity));
        if (newChars == nullptr || newFields == nullptr) {
            uprv_free(newChars);
            uprv_free(newFields);
            status = U_MEMORY_ALLOCATION_ERROR;
            return -1;
        }

        // First copy the prefix and then the suffix, leaving room for the new chars that the
        // caller wants to insert.
        // C++ note: memcpy is OK because the src and dest do not overlap.
        uprv_memcpy2(newChars + newZero, oldChars + oldZero, sizeof(char16_t) * index);
        uprv_memcpy2(newChars + newZero + index + count,
                oldChars + oldZero + index,
                sizeof(char16_t) * (fLength - index));
        uprv_memcpy2(newFields + newZero, oldFields + oldZero, sizeof(Field) * index);
        uprv_memcpy2(newFields + newZero + index + count,
                oldFields + oldZero + index,
                sizeof(Field) * (fLength - index));

        if (fUsingHeap) {
            uprv_free(oldChars);
            uprv_free(oldFields);
        }
        fUsingHeap = true;
        fChars.heap.ptr = newChars;
        fChars.heap.capacity = newCapacity;
        fFields.heap.ptr = newFields;
        fFields.heap.capacity = newCapacity;
        fZero = newZero;
        fLength += count;
    } else {
        int32_t newZero = oldCapacity / 2 - (fLength + count) / 2;

        // C++ note: memmove is required because src and dest may overlap.
        // First copy the entire string to the location of the prefix, and then move the suffix
        // to make room for the new chars that the caller wants to insert.
        uprv_memmove2(oldChars + newZero, oldChars + oldZero, sizeof(char16_t) * fLength);
        uprv_memmove2(oldChars + newZero + index + count,
                oldChars + newZero + index,
                sizeof(char16_t) * (fLength - index));
        uprv_memmove2(oldFields + newZero, oldFields + oldZero, sizeof(Field) * fLength);
        uprv_memmove2(oldFields + newZero + index + count,
                oldFields + newZero + index,
                sizeof(Field) * (fLength - index));

        fZero = newZero;
        fLength += count;
    }
    return fZero + index;
}

int32_t FormattedStringBuilder::remove(int32_t index, int32_t count) {
    // TODO: Reset the heap here?  (If the string after removal can fit on stack?)
    int32_t position = index + fZero;
    uprv_memmove2(getCharPtr() + position,
            getCharPtr() + position + count,
            sizeof(char16_t) * (fLength - index - count));
    uprv_memmove2(getFieldPtr() + position,
            getFieldPtr() + position + count,
            sizeof(Field) * (fLength - index - count));
    fLength -= count;
    return position;
}

UnicodeString FormattedStringBuilder::toUnicodeString() const {
    return UnicodeString(getCharPtr() + fZero, fLength);
}

const UnicodeString FormattedStringBuilder::toTempUnicodeString() const {
    // Readonly-alias constructor:
    return UnicodeString(FALSE, getCharPtr() + fZero, fLength);
}

UnicodeString FormattedStringBuilder::toDebugString() const {
    UnicodeString sb;
    sb.append(u"<FormattedStringBuilder [", -1);
    sb.append(toUnicodeString());
    sb.append(u"] [", -1);
    for (int i = 0; i < fLength; i++) {
        if (fieldAt(i) == UNUM_FIELD_COUNT) {
            sb.append(u'n');
        } else {
            char16_t c;
            switch (fieldAt(i)) {
                case UNUM_SIGN_FIELD:
                    c = u'-';
                    break;
                case UNUM_INTEGER_FIELD:
                    c = u'i';
                    break;
                case UNUM_FRACTION_FIELD:
                    c = u'f';
                    break;
                case UNUM_EXPONENT_FIELD:
                    c = u'e';
                    break;
                case UNUM_EXPONENT_SIGN_FIELD:
                    c = u'+';
                    break;
                case UNUM_EXPONENT_SYMBOL_FIELD:
                    c = u'E';
                    break;
                case UNUM_DECIMAL_SEPARATOR_FIELD:
                    c = u'.';
                    break;
                case UNUM_GROUPING_SEPARATOR_FIELD:
                    c = u',';
                    break;
                case UNUM_PERCENT_FIELD:
                    c = u'%';
                    break;
                case UNUM_PERMILL_FIELD:
                    c = u'\x2030';
                    break;
                case UNUM_CURRENCY_FIELD:
                    c = u'$';
                    break;
                default:
                    c = u'?';
                    break;
            }
            sb.append(c);
        }
    }
    sb.append(u"]>", -1);
    return sb;
}

const char16_t *FormattedStringBuilder::chars() const {
    return getCharPtr() + fZero;
}

bool FormattedStringBuilder::contentEquals(const FormattedStringBuilder &other) const {
    if (fLength != other.fLength) {
        return false;
    }
    for (int32_t i = 0; i < fLength; i++) {
        if (charAt(i) != other.charAt(i) || fieldAt(i) != other.fieldAt(i)) {
            return false;
        }
    }
    return true;
}

bool FormattedStringBuilder::containsField(Field field) const {
    for (int32_t i = 0; i < fLength; i++) {
        if (field == fieldAt(i)) {
            return true;
        }
    }
    return false;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// This file contains one implementation of FormattedValue.
// Other independent implementations should go into their own cpp file for
// better dependency modularization.




U_NAMESPACE_BEGIN


FormattedValueFieldPositionIteratorImpl::FormattedValueFieldPositionIteratorImpl(
        int32_t initialFieldCapacity,
        UErrorCode& status)
        : fFields(initialFieldCapacity * 4, status) {
}

FormattedValueFieldPositionIteratorImpl::~FormattedValueFieldPositionIteratorImpl() = default;

UnicodeString FormattedValueFieldPositionIteratorImpl::toString(
        UErrorCode&) const {
    return fString;
}

UnicodeString FormattedValueFieldPositionIteratorImpl::toTempString(
        UErrorCode&) const {
    // The alias must point to memory owned by this object;
    // fastCopyFrom doesn't do this when using a stack buffer.
    return UnicodeString(TRUE, fString.getBuffer(), fString.length());
}

Appendable& FormattedValueFieldPositionIteratorImpl::appendTo(
        Appendable& appendable,
        UErrorCode&) const {
    appendable.appendString(fString.getBuffer(), fString.length());
    return appendable;
}

UBool FormattedValueFieldPositionIteratorImpl::nextPosition(
        ConstrainedFieldPosition& cfpos,
        UErrorCode&) const {
    U_ASSERT(fFields.size() % 4 == 0);
    int32_t numFields = fFields.size() / 4;
    int32_t i = static_cast<int32_t>(cfpos.getInt64IterationContext());
    for (; i < numFields; i++) {
        UFieldCategory category = static_cast<UFieldCategory>(fFields.elementAti(i * 4));
        int32_t field = fFields.elementAti(i * 4 + 1);
        if (cfpos.matchesField(category, field)) {
            int32_t start = fFields.elementAti(i * 4 + 2);
            int32_t limit = fFields.elementAti(i * 4 + 3);
            cfpos.setState(category, field, start, limit);
            break;
        }
    }
    cfpos.setInt64IterationContext(i == numFields ? i : i + 1);
    return i < numFields;
}


FieldPositionIteratorHandler FormattedValueFieldPositionIteratorImpl::getHandler(
        UErrorCode& status) {
    return FieldPositionIteratorHandler(&fFields, status);
}

void FormattedValueFieldPositionIteratorImpl::appendString(
        UnicodeString string,
        UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    fString.append(string);
    // Make the string NUL-terminated
    if (fString.getTerminatedBuffer() == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
}


void FormattedValueFieldPositionIteratorImpl::addOverlapSpans(
        UFieldCategory spanCategory,
        int8_t firstIndex,
        UErrorCode& status) {
    // In order to avoid fancy data structures, this is an O(N^2) algorithm,
    // which should be fine for all real-life applications of this function.
    int32_t s1a = INT32_MAX;
    int32_t s1b = 0;
    int32_t s2a = INT32_MAX;
    int32_t s2b = 0;
    int32_t numFields = fFields.size() / 4;
    for (int32_t i = 0; i<numFields; i++) {
        int32_t field1 = fFields.elementAti(i * 4 + 1);
        for (int32_t j = i + 1; j<numFields; j++) {
            int32_t field2 = fFields.elementAti(j * 4 + 1);
            if (field1 != field2) {
                continue;
            }
            // Found a duplicate
            s1a = uprv_min(s1a, fFields.elementAti(i * 4 + 2));
            s1b = uprv_max(s1b, fFields.elementAti(i * 4 + 3));
            s2a = uprv_min(s2a, fFields.elementAti(j * 4 + 2));
            s2b = uprv_max(s2b, fFields.elementAti(j * 4 + 3));
            break;
        }
    }
    if (s1a != INT32_MAX) {
        // Success: add the two span fields
        fFields.addElement(spanCategory, status);
        fFields.addElement(firstIndex, status);
        fFields.addElement(s1a, status);
        fFields.addElement(s1b, status);
        fFields.addElement(spanCategory, status);
        fFields.addElement(1 - firstIndex, status);
        fFields.addElement(s2a, status);
        fFields.addElement(s2b, status);
    }
}


void FormattedValueFieldPositionIteratorImpl::sort() {
    // Use bubble sort, O(N^2) but easy and no fancy data structures.
    int32_t numFields = fFields.size() / 4;
    while (true) {
        bool isSorted = true;
        for (int32_t i=0; i<numFields-1; i++) {
            int32_t categ1 = fFields.elementAti(i*4 + 0);
            int32_t field1 = fFields.elementAti(i*4 + 1);
            int32_t start1 = fFields.elementAti(i*4 + 2);
            int32_t limit1 = fFields.elementAti(i*4 + 3);
            int32_t categ2 = fFields.elementAti(i*4 + 4);
            int32_t field2 = fFields.elementAti(i*4 + 5);
            int32_t start2 = fFields.elementAti(i*4 + 6);
            int32_t limit2 = fFields.elementAti(i*4 + 7);
            int64_t comparison = 0;
            if (start1 != start2) {
                // Higher start index -> higher rank
                comparison = start2 - start1;
            } else if (limit1 != limit2) {
                // Higher length (end index) -> lower rank
                comparison = limit1 - limit2;
            } else if (categ1 != categ2) {
                // Higher field category -> lower rank
                comparison = categ1 - categ2;
            } else if (field1 != field2) {
                // Higher field -> higher rank
                comparison = field2 - field1;
            }
            if (comparison < 0) {
                // Perform a swap
                isSorted = false;
                fFields.setElementAt(categ2, i*4 + 0);
                fFields.setElementAt(field2, i*4 + 1);
                fFields.setElementAt(start2, i*4 + 2);
                fFields.setElementAt(limit2, i*4 + 3);
                fFields.setElementAt(categ1, i*4 + 4);
                fFields.setElementAt(field1, i*4 + 5);
                fFields.setElementAt(start1, i*4 + 6);
                fFields.setElementAt(limit1, i*4 + 7);
            }
        }
        if (isSorted) {
            break;
        }
    }
}


U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// This file contains one implementation of FormattedValue.
// Other independent implementations should go into their own cpp file for
// better dependency modularization.








U_NAMESPACE_BEGIN


typedef FormattedStringBuilder::Field Field;


FormattedValueStringBuilderImpl::FormattedValueStringBuilderImpl(Field numericField)
        : fNumericField(numericField) {
}

FormattedValueStringBuilderImpl::~FormattedValueStringBuilderImpl() {
}


UnicodeString FormattedValueStringBuilderImpl::toString(UErrorCode&) const {
    return fString.toUnicodeString();
}

UnicodeString FormattedValueStringBuilderImpl::toTempString(UErrorCode&) const {
    return fString.toTempUnicodeString();
}

Appendable& FormattedValueStringBuilderImpl::appendTo(Appendable& appendable, UErrorCode&) const {
    appendable.appendString(fString.chars(), fString.length());
    return appendable;
}

UBool FormattedValueStringBuilderImpl::nextPosition(ConstrainedFieldPosition& cfpos, UErrorCode& status) const {
    // NOTE: MSVC sometimes complains when implicitly converting between bool and UBool
    return nextPositionImpl(cfpos, fNumericField, status) ? TRUE : FALSE;
}

UBool FormattedValueStringBuilderImpl::nextFieldPosition(FieldPosition& fp, UErrorCode& status) const {
    int32_t rawField = fp.getField();

    if (rawField == FieldPosition::DONT_CARE) {
        return FALSE;
    }

    if (rawField < 0 || rawField >= UNUM_FIELD_COUNT) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }

    ConstrainedFieldPosition cfpos;
    cfpos.constrainField(UFIELD_CATEGORY_NUMBER, rawField);
    cfpos.setState(UFIELD_CATEGORY_NUMBER, rawField, fp.getBeginIndex(), fp.getEndIndex());
    if (nextPositionImpl(cfpos, 0, status)) {
        fp.setBeginIndex(cfpos.getStart());
        fp.setEndIndex(cfpos.getLimit());
        return TRUE;
    }

    // Special case: fraction should start after integer if fraction is not present
    if (rawField == UNUM_FRACTION_FIELD && fp.getEndIndex() == 0) {
        bool inside = false;
        int32_t i = fString.fZero;
        for (; i < fString.fZero + fString.fLength; i++) {
            if (isIntOrGroup(fString.getFieldPtr()[i]) || fString.getFieldPtr()[i] == UNUM_DECIMAL_SEPARATOR_FIELD) {
                inside = true;
            } else if (inside) {
                break;
            }
        }
        fp.setBeginIndex(i - fString.fZero);
        fp.setEndIndex(i - fString.fZero);
    }

    return FALSE;
}

void FormattedValueStringBuilderImpl::getAllFieldPositions(FieldPositionIteratorHandler& fpih,
                                               UErrorCode& status) const {
    ConstrainedFieldPosition cfpos;
    while (nextPositionImpl(cfpos, 0, status)) {
        fpih.addAttribute(cfpos.getField(), cfpos.getStart(), cfpos.getLimit());
    }
}

// Signal the end of the string using a field that doesn't exist and that is
// different from UNUM_FIELD_COUNT, which is used for "null number field".
static constexpr Field kEndField = 0xff;

bool FormattedValueStringBuilderImpl::nextPositionImpl(ConstrainedFieldPosition& cfpos, Field numericField, UErrorCode& /*status*/) const {
    auto numericCAF = StringBuilderFieldUtils::expand(numericField);
    int32_t fieldStart = -1;
    Field currField = UNUM_FIELD_COUNT;
    for (int32_t i = fString.fZero + cfpos.getLimit(); i <= fString.fZero + fString.fLength; i++) {
        Field _field = (i < fString.fZero + fString.fLength) ? fString.getFieldPtr()[i] : kEndField;
        // Case 1: currently scanning a field.
        if (currField != UNUM_FIELD_COUNT) {
            if (currField != _field) {
                int32_t end = i - fString.fZero;
                // Grouping separators can be whitespace; don't throw them out!
                if (currField != UNUM_GROUPING_SEPARATOR_FIELD) {
                    end = trimBack(i - fString.fZero);
                }
                if (end <= fieldStart) {
                    // Entire field position is ignorable; skip.
                    fieldStart = -1;
                    currField = UNUM_FIELD_COUNT;
                    i--;  // look at this index again
                    continue;
                }
                int32_t start = fieldStart;
                if (currField != UNUM_GROUPING_SEPARATOR_FIELD) {
                    start = trimFront(start);
                }
                auto caf = StringBuilderFieldUtils::expand(currField);
                cfpos.setState(caf.category, caf.field, start, end);
                return true;
            }
            continue;
        }
        // Special case: coalesce the INTEGER if we are pointing at the end of the INTEGER.
        if (cfpos.matchesField(UFIELD_CATEGORY_NUMBER, UNUM_INTEGER_FIELD)
                && i > fString.fZero
                // don't return the same field twice in a row:
                && i - fString.fZero > cfpos.getLimit()
                && isIntOrGroup(fString.getFieldPtr()[i - 1])
                && !isIntOrGroup(_field)) {
            int j = i - 1;
            for (; j >= fString.fZero && isIntOrGroup(fString.getFieldPtr()[j]); j--) {}
            cfpos.setState(
                UFIELD_CATEGORY_NUMBER,
                UNUM_INTEGER_FIELD,
                j - fString.fZero + 1,
                i - fString.fZero);
            return true;
        }
        // Special case: coalesce NUMERIC if we are pointing at the end of the NUMERIC.
        if (numericField != 0
                && cfpos.matchesField(numericCAF.category, numericCAF.field)
                && i > fString.fZero
                // don't return the same field twice in a row:
                && (i - fString.fZero > cfpos.getLimit()
                    || cfpos.getCategory() != numericCAF.category
                    || cfpos.getField() != numericCAF.field)
                && isNumericField(fString.getFieldPtr()[i - 1])
                && !isNumericField(_field)) {
            int j = i - 1;
            for (; j >= fString.fZero && isNumericField(fString.getFieldPtr()[j]); j--) {}
            cfpos.setState(
                numericCAF.category,
                numericCAF.field,
                j - fString.fZero + 1,
                i - fString.fZero);
            return true;
        }
        // Special case: skip over INTEGER; will be coalesced later.
        if (_field == UNUM_INTEGER_FIELD) {
            _field = UNUM_FIELD_COUNT;
        }
        // Case 2: no field starting at this position.
        if (_field == UNUM_FIELD_COUNT || _field == kEndField) {
            continue;
        }
        // Case 3: check for field starting at this position
        auto caf = StringBuilderFieldUtils::expand(_field);
        if (cfpos.matchesField(caf.category, caf.field)) {
            fieldStart = i - fString.fZero;
            currField = _field;
        }
    }

    U_ASSERT(currField == UNUM_FIELD_COUNT);
    return false;
}

bool FormattedValueStringBuilderImpl::isIntOrGroup(Field field) {
    return field == UNUM_INTEGER_FIELD
        || field == UNUM_GROUPING_SEPARATOR_FIELD;
}

bool FormattedValueStringBuilderImpl::isNumericField(Field field) {
    return StringBuilderFieldUtils::isNumericField(field);
}

int32_t FormattedValueStringBuilderImpl::trimBack(int32_t limit) const {
    return unisets::get(unisets::UNISET_KEY_DEFAULT_IGNORABLES)->spanBack(
        fString.getCharPtr() + fString.fZero,
        limit,
        USET_SPAN_CONTAINED);
}

int32_t FormattedValueStringBuilderImpl::trimFront(int32_t start) const {
    return start + unisets::get(unisets::UNISET_KEY_DEFAULT_IGNORABLES)->span(
        fString.getCharPtr() + fString.fZero + start,
        fString.fLength - start,
        USET_SPAN_CONTAINED);
}


U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING





U_NAMESPACE_BEGIN


ConstrainedFieldPosition::ConstrainedFieldPosition() {}

ConstrainedFieldPosition::~ConstrainedFieldPosition() {}

void ConstrainedFieldPosition::reset() {
    fContext = 0LL;
    fField = 0;
    fStart = 0;
    fLimit = 0;
    fConstraint = UCFPOS_CONSTRAINT_NONE;
    fCategory = UFIELD_CATEGORY_UNDEFINED;
}

void ConstrainedFieldPosition::constrainCategory(int32_t category) {
    fConstraint = UCFPOS_CONSTRAINT_CATEGORY;
    fCategory = category;
}

void ConstrainedFieldPosition::constrainField(int32_t category, int32_t field) {
    fConstraint = UCFPOS_CONSTRAINT_FIELD;
    fCategory = category;
    fField = field;
}

void ConstrainedFieldPosition::setInt64IterationContext(int64_t context) {
    fContext = context;
}

UBool ConstrainedFieldPosition::matchesField(int32_t category, int32_t field) const {
    switch (fConstraint) {
    case UCFPOS_CONSTRAINT_NONE:
        return TRUE;
    case UCFPOS_CONSTRAINT_CATEGORY:
        return fCategory == category;
    case UCFPOS_CONSTRAINT_FIELD:
        return fCategory == category && fField == field;
    default:
        UPRV_UNREACHABLE;
    }
}

void ConstrainedFieldPosition::setState(
        int32_t category,
        int32_t field,
        int32_t start,
        int32_t limit) {
    fCategory = category;
    fField = field;
    fStart = start;
    fLimit = limit;
}


FormattedValue::~FormattedValue() = default;


///////////////////////
/// C API FUNCTIONS ///
///////////////////////

struct UConstrainedFieldPositionImpl : public UMemory,
        // Magic number as ASCII == "UCF"
        public IcuCApiHelper<UConstrainedFieldPosition, UConstrainedFieldPositionImpl, 0x55434600> {
    ConstrainedFieldPosition fImpl;
};

U_CAPI UConstrainedFieldPosition* U_EXPORT2
ucfpos_open(UErrorCode* ec) {
    auto* impl = new UConstrainedFieldPositionImpl();
    if (impl == nullptr) {
        *ec = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    return impl->exportForC();
}

U_CAPI void U_EXPORT2
ucfpos_reset(UConstrainedFieldPosition* ptr, UErrorCode* ec) {
    auto* impl = UConstrainedFieldPositionImpl::validate(ptr, *ec);
    if (U_FAILURE(*ec)) {
        return;
    }
    impl->fImpl.reset();
}

U_CAPI void U_EXPORT2
ucfpos_constrainCategory(UConstrainedFieldPosition* ptr, int32_t category, UErrorCode* ec) {
    auto* impl = UConstrainedFieldPositionImpl::validate(ptr, *ec);
    if (U_FAILURE(*ec)) {
        return;
    }
    impl->fImpl.constrainCategory(category);
}

U_CAPI void U_EXPORT2
ucfpos_constrainField(UConstrainedFieldPosition* ptr, int32_t category, int32_t field, UErrorCode* ec) {
    auto* impl = UConstrainedFieldPositionImpl::validate(ptr, *ec);
    if (U_FAILURE(*ec)) {
        return;
    }
    impl->fImpl.constrainField(category, field);
}

U_CAPI int32_t U_EXPORT2
ucfpos_getCategory(const UConstrainedFieldPosition* ptr, UErrorCode* ec) {
    const auto* impl = UConstrainedFieldPositionImpl::validate(ptr, *ec);
    if (U_FAILURE(*ec)) {
        return UFIELD_CATEGORY_UNDEFINED;
    }
    return impl->fImpl.getCategory();
}

U_CAPI int32_t U_EXPORT2
ucfpos_getField(const UConstrainedFieldPosition* ptr, UErrorCode* ec) {
    const auto* impl = UConstrainedFieldPositionImpl::validate(ptr, *ec);
    if (U_FAILURE(*ec)) {
        return 0;
    }
    return impl->fImpl.getField();
}

U_CAPI void U_EXPORT2
ucfpos_getIndexes(const UConstrainedFieldPosition* ptr, int32_t* pStart, int32_t* pLimit, UErrorCode* ec) {
    const auto* impl = UConstrainedFieldPositionImpl::validate(ptr, *ec);
    if (U_FAILURE(*ec)) {
        return;
    }
    *pStart = impl->fImpl.getStart();
    *pLimit = impl->fImpl.getLimit();
}

U_CAPI int64_t U_EXPORT2
ucfpos_getInt64IterationContext(const UConstrainedFieldPosition* ptr, UErrorCode* ec) {
    const auto* impl = UConstrainedFieldPositionImpl::validate(ptr, *ec);
    if (U_FAILURE(*ec)) {
        return 0;
    }
    return impl->fImpl.getInt64IterationContext();
}

U_CAPI void U_EXPORT2
ucfpos_setInt64IterationContext(UConstrainedFieldPosition* ptr, int64_t context, UErrorCode* ec) {
    auto* impl = UConstrainedFieldPositionImpl::validate(ptr, *ec);
    if (U_FAILURE(*ec)) {
        return;
    }
    impl->fImpl.setInt64IterationContext(context);
}

U_CAPI UBool U_EXPORT2
ucfpos_matchesField(const UConstrainedFieldPosition* ptr, int32_t category, int32_t field, UErrorCode* ec) {
    const auto* impl = UConstrainedFieldPositionImpl::validate(ptr, *ec);
    if (U_FAILURE(*ec)) {
        return 0;
    }
    return impl->fImpl.matchesField(category, field);
}

U_CAPI void U_EXPORT2
ucfpos_setState(
        UConstrainedFieldPosition* ptr,
        int32_t category,
        int32_t field,
        int32_t start,
        int32_t limit,
        UErrorCode* ec) {
    auto* impl = UConstrainedFieldPositionImpl::validate(ptr, *ec);
    if (U_FAILURE(*ec)) {
        return;
    }
    impl->fImpl.setState(category, field, start, limit);
}

U_CAPI void U_EXPORT2
ucfpos_close(UConstrainedFieldPosition* ptr) {
    UErrorCode localStatus = U_ZERO_ERROR;
    auto* impl = UConstrainedFieldPositionImpl::validate(ptr, localStatus);
    delete impl;
}


U_DRAFT const UChar* U_EXPORT2
ufmtval_getString(
        const UFormattedValue* ufmtval,
        int32_t* pLength,
        UErrorCode* ec) {
    const auto* impl = UFormattedValueApiHelper::validate(ufmtval, *ec);
    if (U_FAILURE(*ec)) {
        return nullptr;
    }
    UnicodeString readOnlyAlias = impl->fFormattedValue->toTempString(*ec);
    if (U_FAILURE(*ec)) {
        return nullptr;
    }
    if (pLength != nullptr) {
        *pLength = readOnlyAlias.length();
    }
    return readOnlyAlias.getBuffer();
}


U_DRAFT UBool U_EXPORT2
ufmtval_nextPosition(
        const UFormattedValue* ufmtval,
        UConstrainedFieldPosition* ucfpos,
        UErrorCode* ec) {
    const auto* fmtval = UFormattedValueApiHelper::validate(ufmtval, *ec);
    auto* cfpos = UConstrainedFieldPositionImpl::validate(ucfpos, *ec);
    if (U_FAILURE(*ec)) {
        return FALSE;
    }
    return fmtval->fFormattedValue->nextPosition(cfpos->fImpl, *ec);
}


U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2015, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

// utility FieldPositionHandler
// base class, null implementation

FieldPositionHandler::~FieldPositionHandler() {
}

void FieldPositionHandler::setShift(int32_t delta) {
  fShift = delta;
}


// utility subclass FieldPositionOnlyHandler

FieldPositionOnlyHandler::FieldPositionOnlyHandler(FieldPosition& _pos)
  : pos(_pos) {
}

FieldPositionOnlyHandler::~FieldPositionOnlyHandler() {
}

void
FieldPositionOnlyHandler::addAttribute(int32_t id, int32_t start, int32_t limit) {
  if (pos.getField() == id && (!acceptFirstOnly || !seenFirst)) {
    seenFirst = TRUE;
    pos.setBeginIndex(start + fShift);
    pos.setEndIndex(limit + fShift);
  }
}

void
FieldPositionOnlyHandler::shiftLast(int32_t delta) {
  if (delta != 0 && pos.getField() != FieldPosition::DONT_CARE && pos.getBeginIndex() != -1) {
    pos.setBeginIndex(delta + pos.getBeginIndex());
    pos.setEndIndex(delta + pos.getEndIndex());
  }
}

UBool
FieldPositionOnlyHandler::isRecording(void) const {
  return pos.getField() != FieldPosition::DONT_CARE;
}

void FieldPositionOnlyHandler::setAcceptFirstOnly(UBool acceptFirstOnly) {
  this->acceptFirstOnly = acceptFirstOnly;
}


// utility subclass FieldPositionIteratorHandler

FieldPositionIteratorHandler::FieldPositionIteratorHandler(FieldPositionIterator* posIter,
                                                           UErrorCode& _status)
    : iter(posIter), vec(NULL), status(_status), fCategory(UFIELD_CATEGORY_UNDEFINED) {
  if (iter && U_SUCCESS(status)) {
    vec = new UVector32(status);
  }
}

FieldPositionIteratorHandler::FieldPositionIteratorHandler(
    UVector32* vec,
    UErrorCode& status)
    : iter(nullptr), vec(vec), status(status), fCategory(UFIELD_CATEGORY_UNDEFINED) {
}

FieldPositionIteratorHandler::~FieldPositionIteratorHandler() {
  // setData adopts the vec regardless of status, so it's safe to null it
  if (iter) {
    iter->setData(vec, status);
  }
  // if iter is null, we never allocated vec, so no need to free it
  vec = NULL;
}

void
FieldPositionIteratorHandler::addAttribute(int32_t id, int32_t start, int32_t limit) {
  if (vec && U_SUCCESS(status) && start < limit) {
    int32_t size = vec->size();
    vec->addElement(fCategory, status);
    vec->addElement(id, status);
    vec->addElement(start + fShift, status);
    vec->addElement(limit + fShift, status);
    if (!U_SUCCESS(status)) {
      vec->setSize(size);
    }
  }
}

void
FieldPositionIteratorHandler::shiftLast(int32_t delta) {
  if (U_SUCCESS(status) && delta != 0) {
    int32_t i = vec->size();
    if (i > 0) {
      --i;
      vec->setElementAt(delta + vec->elementAti(i), i);
      --i;
      vec->setElementAt(delta + vec->elementAti(i), i);
    }
  }
}

UBool
FieldPositionIteratorHandler::isRecording(void) const {
  return U_SUCCESS(status);
}

U_NAMESPACE_END

#endif /* !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2009-2012, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*   Date        Name        Description
*   12/14/09    doug        Creation.
******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING





U_NAMESPACE_BEGIN

FieldPositionIterator::~FieldPositionIterator() {
  delete data;
  data = NULL;
  pos = -1;
}

FieldPositionIterator::FieldPositionIterator()
    : data(NULL), pos(-1) {
}

FieldPositionIterator::FieldPositionIterator(const FieldPositionIterator &rhs)
  : UObject(rhs), data(NULL), pos(rhs.pos) {

  if (rhs.data) {
    UErrorCode status = U_ZERO_ERROR;
    data = new UVector32(status);
    data->assign(*rhs.data, status);
    if (status != U_ZERO_ERROR) {
      delete data;
      data = NULL;
      pos = -1;
    }
  }
}

UBool FieldPositionIterator::operator==(const FieldPositionIterator &rhs) const {
  if (&rhs == this) {
    return TRUE;
  }
  if (pos != rhs.pos) {
    return FALSE;
  }
  if (!data) {
    return rhs.data == NULL;
  }
  return rhs.data ? data->operator==(*rhs.data) : FALSE;
}

void FieldPositionIterator::setData(UVector32 *adopt, UErrorCode& status) {
  // Verify that adopt has valid data, and update status if it doesn't.
  if (U_SUCCESS(status)) {
    if (adopt) {
      if (adopt->size() == 0) {
        delete adopt;
        adopt = NULL;
      } else if ((adopt->size() % 4) != 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
      } else {
        for (int i = 2; i < adopt->size(); i += 4) {
          if (adopt->elementAti(i) >= adopt->elementAti(i+1)) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            break;
          }
        }
      }
    }
  }

  // We own the data, even if status is in error, so we need to delete it now
  // if we're not keeping track of it.
  if (!U_SUCCESS(status)) {
    delete adopt;
    return;
  }

  delete data;
  data = adopt;
  pos = adopt == NULL ? -1 : 0;
}

UBool FieldPositionIterator::next(FieldPosition& fp) {
  if (pos == -1) {
    return FALSE;
  }

  // Ignore the first element of the tetrad: used for field category
  pos++;
  fp.setField(data->elementAti(pos++));
  fp.setBeginIndex(data->elementAti(pos++));
  fp.setEndIndex(data->elementAti(pos++));

  if (pos == data->size()) {
    pos = -1;
  }

  return TRUE;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2008-2013, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
*
* File GENDER.CPP
*
* Modification History:*
*   Date        Name        Description
*
********************************************************************************
*/



#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2008-2013, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
*
* File GENDER.H
*
* Modification History:*
*   Date        Name        Description
*
********************************************************************************
*/


/**
 * \file
 * \brief C++ API: GenderInfo computes the gender of a list.
 */



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2010-2013, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/




#if !UCONFIG_NO_FORMATTING



/**
 * \file
 * \brief C API: The purpose of this API is to compute the gender of a list as a
 * whole given the gender of each element.
 *
 */

/**
 * Genders
 * @stable ICU 50
 */
enum UGender {
    /**
     * Male gender.
     * @stable ICU 50
     */
    UGENDER_MALE,
    /**
     * Female gender.
     * @stable ICU 50
     */
    UGENDER_FEMALE,
    /**
     * Neutral gender.
     * @stable ICU 50
     */
    UGENDER_OTHER
};
/**
 * @stable ICU 50
 */
typedef enum UGender UGender;

struct UGenderInfo;
/**
 * Opaque UGenderInfo object for use in C programs.
 * @stable ICU 50
 */
typedef struct UGenderInfo UGenderInfo;

/**
 * Opens a new UGenderInfo object given locale.
 * @param locale The locale for which the rules are desired.
 * @param status UErrorCode pointer
 * @return A UGenderInfo for the specified locale, or NULL if an error occurred.
 * @stable ICU 50
 */
U_STABLE const UGenderInfo* U_EXPORT2
ugender_getInstance(const char *locale, UErrorCode *status);


/**
 * Given a list, returns the gender of the list as a whole.
 * @param genderInfo pointer that ugender_getInstance returns.
 * @param genders the gender of each element in the list.
 * @param size the size of the list.
 * @param status A pointer to a UErrorCode to receive any errors.
 * @return The gender of the list.
 * @stable ICU 50
 */
U_STABLE UGender U_EXPORT2
ugender_getListGender(const UGenderInfo* genderInfo, const UGender *genders, int32_t size, UErrorCode *status);

#endif /* #if !UCONFIG_NO_FORMATTING */




class GenderInfoTest;

U_NAMESPACE_BEGIN

/** \internal Forward Declaration  */
void U_CALLCONV GenderInfo_initCache(UErrorCode &status);

/**
 * GenderInfo computes the gender of a list as a whole given the gender of
 * each element.
 * @stable ICU 50
 */
class U_I18N_API GenderInfo : public UObject {
public:

    /**
     * Provides access to the predefined GenderInfo object for a given
     * locale.
     *
     * @param locale  The locale for which a <code>GenderInfo</code> object is
     *                returned.
     * @param status  Output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @return        The predefined <code>GenderInfo</code> object pointer for
     *                this locale. The returned object is immutable, so it is
     *                declared as const. Caller does not own the returned
     *                pointer, so it must not attempt to free it.
     * @stable ICU 50
     */
    static const GenderInfo* U_EXPORT2 getInstance(const Locale& locale, UErrorCode& status);

    /**
     * Determines the gender of a list as a whole given the gender of each
     * of the elements.
     *
     * @param genders the gender of each element in the list.
     * @param length the length of gender array.
     * @param status  Output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @return        the gender of the whole list.
     * @stable ICU 50
     */
    UGender getListGender(const UGender* genders, int32_t length, UErrorCode& status) const;

    /**
     * Destructor.
     *
     * @stable ICU 50
     */
    virtual ~GenderInfo();

private:
    int32_t _style;

    /**
     * Copy constructor. One object per locale invariant. Clients
     * must never copy GenderInfo objects.
     */
    GenderInfo(const GenderInfo& other);

    /**
      * Assignment operator. Not applicable to immutable objects.
      */
    GenderInfo& operator=(const GenderInfo&);

    GenderInfo();

    static const GenderInfo* getNeutralInstance();

    static const GenderInfo* getMixedNeutralInstance();

    static const GenderInfo* getMaleTaintsInstance();

    static const GenderInfo* loadInstance(const Locale& locale, UErrorCode& status);

    friend class ::GenderInfoTest;
    friend void U_CALLCONV GenderInfo_initCache(UErrorCode &status);
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */













static UHashtable* gGenderInfoCache = NULL;

static const char* gNeutralStr = "neutral";
static const char* gMailTaintsStr = "maleTaints";
static const char* gMixedNeutralStr = "mixedNeutral";
static icu::GenderInfo* gObjs = NULL;
static icu::UInitOnce gGenderInitOnce = U_INITONCE_INITIALIZER;

enum GenderStyle {
  NEUTRAL,
  MIXED_NEUTRAL,
  MALE_TAINTS,
  GENDER_STYLE_LENGTH
};

U_CDECL_BEGIN

static UBool U_CALLCONV gender_cleanup(void) {
  if (gGenderInfoCache != NULL) {
    uhash_close(gGenderInfoCache);
    gGenderInfoCache = NULL;
    delete [] gObjs;
  }
  gGenderInitOnce.reset();
  return TRUE;
}

U_CDECL_END

U_NAMESPACE_BEGIN

void U_CALLCONV GenderInfo_initCache(UErrorCode &status) {
  ucln_i18n_registerCleanup(UCLN_I18N_GENDERINFO, gender_cleanup);
  U_ASSERT(gGenderInfoCache == NULL);
  if (U_FAILURE(status)) {
      return;
  }
  gObjs = new GenderInfo[GENDER_STYLE_LENGTH];
  if (gObjs == NULL) {
    status = U_MEMORY_ALLOCATION_ERROR;
    return;
  }
  for (int i = 0; i < GENDER_STYLE_LENGTH; i++) {
    gObjs[i]._style = i;
  }
  gGenderInfoCache = uhash_open(uhash_hashChars, uhash_compareChars, NULL, &status);
  if (U_FAILURE(status)) {
    delete [] gObjs;
    return;
  }
  uhash_setKeyDeleter(gGenderInfoCache, uprv_free);
}


GenderInfo::GenderInfo() {
}

GenderInfo::~GenderInfo() {
}

const GenderInfo* GenderInfo::getInstance(const Locale& locale, UErrorCode& status) {
  // Make sure our cache exists.
  umtx_initOnce(gGenderInitOnce, &GenderInfo_initCache, status);
  if (U_FAILURE(status)) {
    return NULL;
  }

  static UMutex gGenderMetaLock;
  const GenderInfo* result = NULL;
  const char* key = locale.getName();
  {
    Mutex lock(&gGenderMetaLock);
    result = (const GenderInfo*) uhash_get(gGenderInfoCache, key);
  }
  if (result) {
    return result;
  }

  // On cache miss, try to create GenderInfo from CLDR data
  result = loadInstance(locale, status);
  if (U_FAILURE(status)) {
    return NULL;
  }

  // Try to put our GenderInfo object in cache. If there is a race condition,
  // favor the GenderInfo object that is already in the cache.
  {
    Mutex lock(&gGenderMetaLock);
    GenderInfo* temp = (GenderInfo*) uhash_get(gGenderInfoCache, key);
    if (temp) {
      result = temp;
    } else {
      uhash_put(gGenderInfoCache, uprv_strdup(key), (void*) result, &status);
      if (U_FAILURE(status)) {
        return NULL;
      }
    }
  }
  return result;
}

const GenderInfo* GenderInfo::loadInstance(const Locale& locale, UErrorCode& status) {
  LocalUResourceBundlePointer rb(
      ures_openDirect(NULL, "genderList", &status));
  if (U_FAILURE(status)) {
    return NULL;
  }
  LocalUResourceBundlePointer locRes(ures_getByKey(rb.getAlias(), "genderList", NULL, &status));
  if (U_FAILURE(status)) {
    return NULL;
  }
  int32_t resLen = 0;
  const char* curLocaleName = locale.getName();
  UErrorCode key_status = U_ZERO_ERROR;
  const UChar* s = ures_getStringByKey(locRes.getAlias(), curLocaleName, &resLen, &key_status);
  if (s == NULL) {
    key_status = U_ZERO_ERROR;
    char parentLocaleName[ULOC_FULLNAME_CAPACITY];
    uprv_strcpy(parentLocaleName, curLocaleName);
    while (s == NULL && uloc_getParent(parentLocaleName, parentLocaleName, ULOC_FULLNAME_CAPACITY, &key_status) > 0) {
      key_status = U_ZERO_ERROR;
      resLen = 0;
      s = ures_getStringByKey(locRes.getAlias(), parentLocaleName, &resLen, &key_status);
      key_status = U_ZERO_ERROR;
    }
  }
  if (s == NULL) {
    return &gObjs[NEUTRAL];
  }
  char type_str[256];
  u_UCharsToChars(s, type_str, resLen + 1);
  if (uprv_strcmp(type_str, gNeutralStr) == 0) {
    return &gObjs[NEUTRAL];
  }
  if (uprv_strcmp(type_str, gMixedNeutralStr) == 0) {
    return &gObjs[MIXED_NEUTRAL]; 
  }
  if (uprv_strcmp(type_str, gMailTaintsStr) == 0) {
    return &gObjs[MALE_TAINTS];
  }
  return &gObjs[NEUTRAL];
}

UGender GenderInfo::getListGender(const UGender* genders, int32_t length, UErrorCode& status) const {
  if (U_FAILURE(status)) {
    return UGENDER_OTHER;
  }
  if (length == 0) {
    return UGENDER_OTHER;
  }
  if (length == 1) {
    return genders[0];
  }
  UBool has_female = FALSE;
  UBool has_male = FALSE;
  switch (_style) {
    case NEUTRAL:
      return UGENDER_OTHER;
    case MIXED_NEUTRAL:
      for (int32_t i = 0; i < length; ++i) {
        switch (genders[i]) {
          case UGENDER_OTHER:
            return UGENDER_OTHER;
            break;
          case UGENDER_FEMALE:
            if (has_male) {
              return UGENDER_OTHER;
            }
            has_female = TRUE;
            break;
          case UGENDER_MALE:
            if (has_female) {
              return UGENDER_OTHER;
            }
            has_male = TRUE;
            break;
          default:
            break;
        }
      }
      return has_male ? UGENDER_MALE : UGENDER_FEMALE;
      break;
    case MALE_TAINTS:
      for (int32_t i = 0; i < length; ++i) {
        if (genders[i] != UGENDER_FEMALE) {
          return UGENDER_MALE;
        }
      }
      return UGENDER_FEMALE;
      break;
    default:
      return UGENDER_OTHER;
      break;
  }
}

const GenderInfo* GenderInfo::getNeutralInstance() {
  return &gObjs[NEUTRAL];
}

const GenderInfo* GenderInfo::getMixedNeutralInstance() {
  return &gObjs[MIXED_NEUTRAL];
}

const GenderInfo* GenderInfo::getMaleTaintsInstance() {
  return &gObjs[MALE_TAINTS];
}

U_NAMESPACE_END

U_CAPI const UGenderInfo* U_EXPORT2
ugender_getInstance(const char* locale, UErrorCode* status) {
  return (const UGenderInfo*) icu::GenderInfo::getInstance(locale, *status);
}

U_CAPI UGender U_EXPORT2
ugender_getListGender(const UGenderInfo* genderInfo, const UGender* genders, int32_t size, UErrorCode* status) {
  return ((const icu::GenderInfo *)genderInfo)->getListGender(genders, size, *status);
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File GREGOCAL.CPP
*
* Modification History:
*
*   Date        Name        Description
*   02/05/97    clhuang     Creation.
*   03/28/97    aliu        Made highly questionable fix to computeFields to
*                           handle DST correctly.
*   04/22/97    aliu        Cleaned up code drastically.  Added monthLength().
*                           Finished unimplemented parts of computeTime() for
*                           week-based date determination.  Removed quetionable
*                           fix and wrote correct fix for computeFields() and
*                           daylight time handling.  Rewrote inDaylightTime()
*                           and computeFields() to handle sensitive Daylight to
*                           Standard time transitions correctly.
*   05/08/97    aliu        Added code review changes.  Fixed isLeapYear() to
*                           not cutover.
*   08/12/97    aliu        Added equivalentTo.  Misc other fixes.  Updated
*                           add() from Java source.
*    07/28/98    stephen        Sync up with JDK 1.2
*    09/14/98    stephen        Changed type of kOneDay, kOneWeek to double.
*                            Fixed bug in roll()
*   10/15/99    aliu        Fixed j31, incorrect WEEK_OF_YEAR computation.
*   10/15/99    aliu        Fixed j32, cannot set date to Feb 29 2000 AD.
*                           {JDK bug 4210209 4209272}
*   11/15/99    weiv        Added YEAR_WOY and DOW_LOCAL computation
*                           to timeToFields method, updated kMinValues, kMaxValues & kLeastMaxValues
*   12/09/99    aliu        Fixed j81, calculation errors and roll bugs
*                           in year of cutover.
*   01/24/2000  aliu        Revised computeJulianDay for YEAR YEAR_WOY WOY.
********************************************************************************
*/


#include <float.h>

#if !UCONFIG_NO_FORMATTING






// *****************************************************************************
// class GregorianCalendar
// *****************************************************************************

/**
* Note that the Julian date used here is not a true Julian date, since
* it is measured from midnight, not noon.  This value is the Julian
* day number of January 1, 1970 (Gregorian calendar) at noon UTC. [LIU]
*/

static const int16_t kNumDays[]
= {0,31,59,90,120,151,181,212,243,273,304,334}; // 0-based, for day-in-year
static const int16_t kLeapNumDays[]
= {0,31,60,91,121,152,182,213,244,274,305,335}; // 0-based, for day-in-year
static const int8_t kMonthLength[]
= {31,28,31,30,31,30,31,31,30,31,30,31}; // 0-based
static const int8_t kLeapMonthLength[]
= {31,29,31,30,31,30,31,31,30,31,30,31}; // 0-based

// setTimeInMillis() limits the Julian day range to +/-7F000000.
// This would seem to limit the year range to:
//  ms=+183882168921600000  jd=7f000000  December 20, 5828963 AD
//  ms=-184303902528000000  jd=81000000  September 20, 5838270 BC
// HOWEVER, CalendarRegressionTest/Test4167060 shows that the actual
// range limit on the year field is smaller (~ +/-140000). [alan 3.0]

static const int32_t kGregorianCalendarLimits[UCAL_FIELD_COUNT][4] = {
    // Minimum  Greatest    Least  Maximum
    //           Minimum  Maximum
    {        0,        0,        1,        1}, // ERA
    {        1,        1,   140742,   144683}, // YEAR
    {        0,        0,       11,       11}, // MONTH
    {        1,        1,       52,       53}, // WEEK_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // WEEK_OF_MONTH
    {        1,        1,       28,       31}, // DAY_OF_MONTH
    {        1,        1,      365,      366}, // DAY_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DAY_OF_WEEK
    {       -1,       -1,        4,        5}, // DAY_OF_WEEK_IN_MONTH
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // AM_PM
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR_OF_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MINUTE
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // SECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // ZONE_OFFSET
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DST_OFFSET
    {  -140742,  -140742,   140742,   144683}, // YEAR_WOY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DOW_LOCAL
    {  -140742,  -140742,   140742,   144683}, // EXTENDED_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // JULIAN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECONDS_IN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // IS_LEAP_MONTH
};

/*
* <pre>
*                            Greatest       Least
* Field name        Minimum   Minimum     Maximum     Maximum
* ----------        -------   -------     -------     -------
* ERA                     0         0           1           1
* YEAR                    1         1      140742      144683
* MONTH                   0         0          11          11
* WEEK_OF_YEAR            1         1          52          53
* WEEK_OF_MONTH           0         0           4           6
* DAY_OF_MONTH            1         1          28          31
* DAY_OF_YEAR             1         1         365         366
* DAY_OF_WEEK             1         1           7           7
* DAY_OF_WEEK_IN_MONTH   -1        -1           4           5
* AM_PM                   0         0           1           1
* HOUR                    0         0          11          11
* HOUR_OF_DAY             0         0          23          23
* MINUTE                  0         0          59          59
* SECOND                  0         0          59          59
* MILLISECOND             0         0         999         999
* ZONE_OFFSET           -12*      -12*         12*         12*
* DST_OFFSET              0         0           1*          1*
* YEAR_WOY                1         1      140742      144683
* DOW_LOCAL               1         1           7           7
* </pre>
* (*) In units of one-hour
*/

#if defined( U_DEBUG_CALSVC ) || defined (U_DEBUG_CAL)
#include <stdio.h>
#endif

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(GregorianCalendar)

// 00:00:00 UTC, October 15, 1582, expressed in ms from the epoch.
// Note that only Italy and other Catholic countries actually
// observed this cutover.  Most other countries followed in
// the next few centuries, some as late as 1928. [LIU]
// in Java, -12219292800000L
//const UDate GregorianCalendar::kPapalCutover = -12219292800000L;
static const uint32_t kCutoverJulianDay = 2299161;
static const UDate kPapalCutover = (2299161.0 - kEpochStartAsJulianDay) * U_MILLIS_PER_DAY;
//static const UDate kPapalCutoverJulian = (2299161.0 - kEpochStartAsJulianDay);

// -------------------------------------

GregorianCalendar::GregorianCalendar(UErrorCode& status)
:   Calendar(status),
fGregorianCutover(kPapalCutover),
fCutoverJulianDay(kCutoverJulianDay), fNormalizedGregorianCutover(fGregorianCutover), fGregorianCutoverYear(1582),
fIsGregorian(TRUE), fInvertGregorian(FALSE)
{
    setTimeInMillis(getNow(), status);
}

// -------------------------------------

GregorianCalendar::GregorianCalendar(TimeZone* zone, UErrorCode& status)
:   Calendar(zone, Locale::getDefault(), status),
fGregorianCutover(kPapalCutover),
fCutoverJulianDay(kCutoverJulianDay), fNormalizedGregorianCutover(fGregorianCutover), fGregorianCutoverYear(1582),
fIsGregorian(TRUE), fInvertGregorian(FALSE)
{
    setTimeInMillis(getNow(), status);
}

// -------------------------------------

GregorianCalendar::GregorianCalendar(const TimeZone& zone, UErrorCode& status)
:   Calendar(zone, Locale::getDefault(), status),
fGregorianCutover(kPapalCutover),
fCutoverJulianDay(kCutoverJulianDay), fNormalizedGregorianCutover(fGregorianCutover), fGregorianCutoverYear(1582),
fIsGregorian(TRUE), fInvertGregorian(FALSE)
{
    setTimeInMillis(getNow(), status);
}

// -------------------------------------

GregorianCalendar::GregorianCalendar(const Locale& aLocale, UErrorCode& status)
:   Calendar(TimeZone::createDefault(), aLocale, status),
fGregorianCutover(kPapalCutover),
fCutoverJulianDay(kCutoverJulianDay), fNormalizedGregorianCutover(fGregorianCutover), fGregorianCutoverYear(1582),
fIsGregorian(TRUE), fInvertGregorian(FALSE)
{
    setTimeInMillis(getNow(), status);
}

// -------------------------------------

GregorianCalendar::GregorianCalendar(TimeZone* zone, const Locale& aLocale,
                                     UErrorCode& status)
                                     :   Calendar(zone, aLocale, status),
                                     fGregorianCutover(kPapalCutover),
                                     fCutoverJulianDay(kCutoverJulianDay), fNormalizedGregorianCutover(fGregorianCutover), fGregorianCutoverYear(1582),
                                     fIsGregorian(TRUE), fInvertGregorian(FALSE)
{
    setTimeInMillis(getNow(), status);
}

// -------------------------------------

GregorianCalendar::GregorianCalendar(const TimeZone& zone, const Locale& aLocale,
                                     UErrorCode& status)
                                     :   Calendar(zone, aLocale, status),
                                     fGregorianCutover(kPapalCutover),
                                     fCutoverJulianDay(kCutoverJulianDay), fNormalizedGregorianCutover(fGregorianCutover), fGregorianCutoverYear(1582),
                                     fIsGregorian(TRUE), fInvertGregorian(FALSE)
{
    setTimeInMillis(getNow(), status);
}

// -------------------------------------

GregorianCalendar::GregorianCalendar(int32_t year, int32_t month, int32_t date,
                                     UErrorCode& status)
                                     :   Calendar(TimeZone::createDefault(), Locale::getDefault(), status),
                                     fGregorianCutover(kPapalCutover),
                                     fCutoverJulianDay(kCutoverJulianDay), fNormalizedGregorianCutover(fGregorianCutover), fGregorianCutoverYear(1582),
                                     fIsGregorian(TRUE), fInvertGregorian(FALSE)
{
    set(UCAL_ERA, AD);
    set(UCAL_YEAR, year);
    set(UCAL_MONTH, month);
    set(UCAL_DATE, date);
}

// -------------------------------------

GregorianCalendar::GregorianCalendar(int32_t year, int32_t month, int32_t date,
                                     int32_t hour, int32_t minute, UErrorCode& status)
                                     :   Calendar(TimeZone::createDefault(), Locale::getDefault(), status),
                                     fGregorianCutover(kPapalCutover),
                                     fCutoverJulianDay(kCutoverJulianDay), fNormalizedGregorianCutover(fGregorianCutover), fGregorianCutoverYear(1582),
                                     fIsGregorian(TRUE), fInvertGregorian(FALSE)
{
    set(UCAL_ERA, AD);
    set(UCAL_YEAR, year);
    set(UCAL_MONTH, month);
    set(UCAL_DATE, date);
    set(UCAL_HOUR_OF_DAY, hour);
    set(UCAL_MINUTE, minute);
}

// -------------------------------------

GregorianCalendar::GregorianCalendar(int32_t year, int32_t month, int32_t date,
                                     int32_t hour, int32_t minute, int32_t second,
                                     UErrorCode& status)
                                     :   Calendar(TimeZone::createDefault(), Locale::getDefault(), status),
                                     fGregorianCutover(kPapalCutover),
                                     fCutoverJulianDay(kCutoverJulianDay), fNormalizedGregorianCutover(fGregorianCutover), fGregorianCutoverYear(1582),
                                     fIsGregorian(TRUE), fInvertGregorian(FALSE)
{
    set(UCAL_ERA, AD);
    set(UCAL_YEAR, year);
    set(UCAL_MONTH, month);
    set(UCAL_DATE, date);
    set(UCAL_HOUR_OF_DAY, hour);
    set(UCAL_MINUTE, minute);
    set(UCAL_SECOND, second);
}

// -------------------------------------

GregorianCalendar::~GregorianCalendar()
{
}

// -------------------------------------

GregorianCalendar::GregorianCalendar(const GregorianCalendar &source)
:   Calendar(source),
fGregorianCutover(source.fGregorianCutover),
fCutoverJulianDay(source.fCutoverJulianDay), fNormalizedGregorianCutover(source.fNormalizedGregorianCutover), fGregorianCutoverYear(source.fGregorianCutoverYear),
fIsGregorian(source.fIsGregorian), fInvertGregorian(source.fInvertGregorian)
{
}

// -------------------------------------

GregorianCalendar* GregorianCalendar::clone() const
{
    return new GregorianCalendar(*this);
}

// -------------------------------------

GregorianCalendar &
GregorianCalendar::operator=(const GregorianCalendar &right)
{
    if (this != &right)
    {
        Calendar::operator=(right);
        fGregorianCutover = right.fGregorianCutover;
        fNormalizedGregorianCutover = right.fNormalizedGregorianCutover;
        fGregorianCutoverYear = right.fGregorianCutoverYear;
        fCutoverJulianDay = right.fCutoverJulianDay;
    }
    return *this;
}

// -------------------------------------

UBool GregorianCalendar::isEquivalentTo(const Calendar& other) const
{
    // Calendar override.
    return Calendar::isEquivalentTo(other) &&
        fGregorianCutover == ((GregorianCalendar*)&other)->fGregorianCutover;
}

// -------------------------------------

void
GregorianCalendar::setGregorianChange(UDate date, UErrorCode& status)
{
    if (U_FAILURE(status))
        return;

    // Precompute two internal variables which we use to do the actual
    // cutover computations.  These are the normalized cutover, which is the
    // midnight at or before the cutover, and the cutover year.  The
    // normalized cutover is in pure date milliseconds; it contains no time
    // of day or timezone component, and it used to compare against other
    // pure date values.
    double cutoverDay = ClockMath::floorDivide(date, (double)kOneDay);

    // Handle the rare case of numeric overflow where the user specifies a time
    // outside of INT32_MIN .. INT32_MAX number of days.

    if (cutoverDay <= INT32_MIN) {
        cutoverDay = INT32_MIN;
        fGregorianCutover = fNormalizedGregorianCutover = cutoverDay * kOneDay;
    } else if (cutoverDay >= INT32_MAX) {
        cutoverDay = INT32_MAX;
        fGregorianCutover = fNormalizedGregorianCutover = cutoverDay * kOneDay;
    } else {
        fNormalizedGregorianCutover = cutoverDay * kOneDay;
        fGregorianCutover = date;
    }

    // Normalize the year so BC values are represented as 0 and negative
    // values.
    GregorianCalendar *cal = new GregorianCalendar(getTimeZone(), status);
    /* test for NULL */
    if (cal == 0) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    if(U_FAILURE(status))
        return;
    cal->setTime(date, status);
    fGregorianCutoverYear = cal->get(UCAL_YEAR, status);
    if (cal->get(UCAL_ERA, status) == BC)
        fGregorianCutoverYear = 1 - fGregorianCutoverYear;
    fCutoverJulianDay = (int32_t)cutoverDay;
    delete cal;
}


void GregorianCalendar::handleComputeFields(int32_t julianDay, UErrorCode& status) {
    int32_t eyear, month, dayOfMonth, dayOfYear, unusedRemainder;


    if(U_FAILURE(status)) {
        return;
    }

#if defined (U_DEBUG_CAL)
    fprintf(stderr, "%s:%d: jd%d- (greg's %d)- [cut=%d]\n",
        __FILE__, __LINE__, julianDay, getGregorianDayOfYear(), fCutoverJulianDay);
#endif


    if (julianDay >= fCutoverJulianDay) {
        month = getGregorianMonth();
        dayOfMonth = getGregorianDayOfMonth();
        dayOfYear = getGregorianDayOfYear();
        eyear = getGregorianYear();
    } else {
        // The Julian epoch day (not the same as Julian Day)
        // is zero on Saturday December 30, 0 (Gregorian).
        int32_t julianEpochDay = julianDay - (kJan1_1JulianDay - 2);
		eyear = (int32_t) ClockMath::floorDivide((4.0*julianEpochDay) + 1464.0, (int32_t) 1461, unusedRemainder);

        // Compute the Julian calendar day number for January 1, eyear
        int32_t january1 = 365*(eyear-1) + ClockMath::floorDivide(eyear-1, (int32_t)4);
        dayOfYear = (julianEpochDay - january1); // 0-based

        // Julian leap years occurred historically every 4 years starting
        // with 8 AD.  Before 8 AD the spacing is irregular; every 3 years
        // from 45 BC to 9 BC, and then none until 8 AD.  However, we don't
        // implement this historical detail; instead, we implement the
        // computatinally cleaner proleptic calendar, which assumes
        // consistent 4-year cycles throughout time.
        UBool isLeap = ((eyear&0x3) == 0); // equiv. to (eyear%4 == 0)

        // Common Julian/Gregorian calculation
        int32_t correction = 0;
        int32_t march1 = isLeap ? 60 : 59; // zero-based DOY for March 1
        if (dayOfYear >= march1) {
            correction = isLeap ? 1 : 2;
        }
        month = (12 * (dayOfYear + correction) + 6) / 367; // zero-based month
        dayOfMonth = dayOfYear - (isLeap?kLeapNumDays[month]:kNumDays[month]) + 1; // one-based DOM
        ++dayOfYear;
#if defined (U_DEBUG_CAL)
        //     fprintf(stderr, "%d - %d[%d] + 1\n", dayOfYear, isLeap?kLeapNumDays[month]:kNumDays[month], month );
        //           fprintf(stderr, "%s:%d:  greg's HCF %d -> %d/%d/%d not %d/%d/%d\n",
        //                   __FILE__, __LINE__,julianDay,
        //          eyear,month,dayOfMonth,
        //          getGregorianYear(), getGregorianMonth(), getGregorianDayOfMonth()  );
        fprintf(stderr, "%s:%d: doy %d (greg's %d)- [cut=%d]\n",
            __FILE__, __LINE__, dayOfYear, getGregorianDayOfYear(), fCutoverJulianDay);
#endif

    }

    // [j81] if we are after the cutover in its year, shift the day of the year
    if((eyear == fGregorianCutoverYear) && (julianDay >= fCutoverJulianDay)) {
        //from handleComputeMonthStart
        int32_t gregShift = Grego::gregorianShift(eyear);
#if defined (U_DEBUG_CAL)
        fprintf(stderr, "%s:%d:  gregorian shift %d :::  doy%d => %d [cut=%d]\n",
            __FILE__, __LINE__,gregShift, dayOfYear, dayOfYear+gregShift, fCutoverJulianDay);
#endif
        dayOfYear += gregShift;
    }

    internalSet(UCAL_MONTH, month);
    internalSet(UCAL_DAY_OF_MONTH, dayOfMonth);
    internalSet(UCAL_DAY_OF_YEAR, dayOfYear);
    internalSet(UCAL_EXTENDED_YEAR, eyear);
    int32_t era = AD;
    if (eyear < 1) {
        era = BC;
        eyear = 1 - eyear;
    }
    internalSet(UCAL_ERA, era);
    internalSet(UCAL_YEAR, eyear);
}


// -------------------------------------

UDate
GregorianCalendar::getGregorianChange() const
{
    return fGregorianCutover;
}

// -------------------------------------

UBool
GregorianCalendar::isLeapYear(int32_t year) const
{
    // MSVC complains bitterly if we try to use Grego::isLeapYear here
    // NOTE: year&0x3 == year%4
    return (year >= fGregorianCutoverYear ?
        (((year&0x3) == 0) && ((year%100 != 0) || (year%400 == 0))) : // Gregorian
    ((year&0x3) == 0)); // Julian
}

// -------------------------------------

int32_t GregorianCalendar::handleComputeJulianDay(UCalendarDateFields bestField)
{
    fInvertGregorian = FALSE;

    int32_t jd = Calendar::handleComputeJulianDay(bestField);

    if((bestField == UCAL_WEEK_OF_YEAR) &&  // if we are doing WOY calculations, we are counting relative to Jan 1 *julian*
        (internalGet(UCAL_EXTENDED_YEAR)==fGregorianCutoverYear) &&
        jd >= fCutoverJulianDay) {
            fInvertGregorian = TRUE;  // So that the Julian Jan 1 will be used in handleComputeMonthStart
            return Calendar::handleComputeJulianDay(bestField);
        }


        // The following check handles portions of the cutover year BEFORE the
        // cutover itself happens.
        //if ((fIsGregorian==TRUE) != (jd >= fCutoverJulianDay)) {  /*  cutoverJulianDay)) { */
        if ((fIsGregorian==TRUE) != (jd >= fCutoverJulianDay)) {  /*  cutoverJulianDay)) { */
#if defined (U_DEBUG_CAL)
            fprintf(stderr, "%s:%d: jd [invert] %d\n",
                __FILE__, __LINE__, jd);
#endif
            fInvertGregorian = TRUE;
            jd = Calendar::handleComputeJulianDay(bestField);
#if defined (U_DEBUG_CAL)
            fprintf(stderr, "%s:%d:  fIsGregorian %s, fInvertGregorian %s - ",
                __FILE__, __LINE__,fIsGregorian?"T":"F", fInvertGregorian?"T":"F");
            fprintf(stderr, " jd NOW %d\n",
                jd);
#endif
        } else {
#if defined (U_DEBUG_CAL)
            fprintf(stderr, "%s:%d: jd [==] %d - %sfIsGregorian %sfInvertGregorian, %d\n",
                __FILE__, __LINE__, jd, fIsGregorian?"T":"F", fInvertGregorian?"T":"F", bestField);
#endif
        }

        if(fIsGregorian && (internalGet(UCAL_EXTENDED_YEAR) == fGregorianCutoverYear)) {
            int32_t gregShift = Grego::gregorianShift(internalGet(UCAL_EXTENDED_YEAR));
            if (bestField == UCAL_DAY_OF_YEAR) {
#if defined (U_DEBUG_CAL)
                fprintf(stderr, "%s:%d: [DOY%d] gregorian shift of JD %d += %d\n",
                    __FILE__, __LINE__, fFields[bestField],jd, gregShift);
#endif
                jd -= gregShift;
            } else if ( bestField == UCAL_WEEK_OF_MONTH ) {
                int32_t weekShift = 14;
#if defined (U_DEBUG_CAL)
                fprintf(stderr, "%s:%d: [WOY/WOM] gregorian week shift of %d += %d\n",
                    __FILE__, __LINE__, jd, weekShift);
#endif
                jd += weekShift; // shift by weeks for week based fields.
            }
        }

        return jd;
}

int32_t GregorianCalendar::handleComputeMonthStart(int32_t eyear, int32_t month,

                                                   UBool /* useMonth */) const
{
    GregorianCalendar *nonConstThis = (GregorianCalendar*)this; // cast away const

    // If the month is out of range, adjust it into range, and
    // modify the extended year value accordingly.
    if (month < 0 || month > 11) {
        eyear += ClockMath::floorDivide(month, 12, month);
    }

    UBool isLeap = eyear%4 == 0;
    int64_t y = (int64_t)eyear-1;
    int64_t julianDay = 365*y + ClockMath::floorDivide(y, (int64_t)4) + (kJan1_1JulianDay - 3);

    nonConstThis->fIsGregorian = (eyear >= fGregorianCutoverYear);
#if defined (U_DEBUG_CAL)
    fprintf(stderr, "%s:%d: (hcms%d/%d) fIsGregorian %s, fInvertGregorian %s\n",
        __FILE__, __LINE__, eyear,month, fIsGregorian?"T":"F", fInvertGregorian?"T":"F");
#endif
    if (fInvertGregorian) {
        nonConstThis->fIsGregorian = !fIsGregorian;
    }
    if (fIsGregorian) {
        isLeap = isLeap && ((eyear%100 != 0) || (eyear%400 == 0));
        // Add 2 because Gregorian calendar starts 2 days after
        // Julian calendar
        int32_t gregShift = Grego::gregorianShift(eyear);
#if defined (U_DEBUG_CAL)
        fprintf(stderr, "%s:%d: (hcms%d/%d) gregorian shift of %d += %d\n",
            __FILE__, __LINE__, eyear, month, julianDay, gregShift);
#endif
        julianDay += gregShift;
    }

    // At this point julianDay indicates the day BEFORE the first
    // day of January 1, <eyear> of either the Julian or Gregorian
    // calendar.

    if (month != 0) {
        julianDay += isLeap?kLeapNumDays[month]:kNumDays[month];
    }

    return static_cast<int32_t>(julianDay);
}

int32_t GregorianCalendar::handleGetMonthLength(int32_t extendedYear, int32_t month)  const
{
    // If the month is out of range, adjust it into range, and
    // modify the extended year value accordingly.
    if (month < 0 || month > 11) {
        extendedYear += ClockMath::floorDivide(month, 12, month);
    }

    return isLeapYear(extendedYear) ? kLeapMonthLength[month] : kMonthLength[month];
}

int32_t GregorianCalendar::handleGetYearLength(int32_t eyear) const {
    return isLeapYear(eyear) ? 366 : 365;
}


int32_t
GregorianCalendar::monthLength(int32_t month) const
{
    int32_t year = internalGet(UCAL_EXTENDED_YEAR);
    return handleGetMonthLength(year, month);
}

// -------------------------------------

int32_t
GregorianCalendar::monthLength(int32_t month, int32_t year) const
{
    return isLeapYear(year) ? kLeapMonthLength[month] : kMonthLength[month];
}

// -------------------------------------

int32_t
GregorianCalendar::yearLength(int32_t year) const
{
    return isLeapYear(year) ? 366 : 365;
}

// -------------------------------------

int32_t
GregorianCalendar::yearLength() const
{
    return isLeapYear(internalGet(UCAL_YEAR)) ? 366 : 365;
}

// -------------------------------------

/**
* After adjustments such as add(MONTH), add(YEAR), we don't want the
* month to jump around.  E.g., we don't want Jan 31 + 1 month to go to Mar
* 3, we want it to go to Feb 28.  Adjustments which might run into this
* problem call this method to retain the proper month.
*/
void
GregorianCalendar::pinDayOfMonth()
{
    int32_t monthLen = monthLength(internalGet(UCAL_MONTH));
    int32_t dom = internalGet(UCAL_DATE);
    if(dom > monthLen)
        set(UCAL_DATE, monthLen);
}

// -------------------------------------


UBool
GregorianCalendar::validateFields() const
{
    for (int32_t field = 0; field < UCAL_FIELD_COUNT; field++) {
        // Ignore DATE and DAY_OF_YEAR which are handled below
        if (field != UCAL_DATE &&
            field != UCAL_DAY_OF_YEAR &&
            isSet((UCalendarDateFields)field) &&
            ! boundsCheck(internalGet((UCalendarDateFields)field), (UCalendarDateFields)field))
            return FALSE;
    }

    // Values differ in Least-Maximum and Maximum should be handled
    // specially.
    if (isSet(UCAL_DATE)) {
        int32_t date = internalGet(UCAL_DATE);
        if (date < getMinimum(UCAL_DATE) ||
            date > monthLength(internalGet(UCAL_MONTH))) {
                return FALSE;
            }
    }

    if (isSet(UCAL_DAY_OF_YEAR)) {
        int32_t days = internalGet(UCAL_DAY_OF_YEAR);
        if (days < 1 || days > yearLength()) {
            return FALSE;
        }
    }

    // Handle DAY_OF_WEEK_IN_MONTH, which must not have the value zero.
    // We've checked against minimum and maximum above already.
    if (isSet(UCAL_DAY_OF_WEEK_IN_MONTH) &&
        0 == internalGet(UCAL_DAY_OF_WEEK_IN_MONTH)) {
            return FALSE;
        }

        return TRUE;
}

// -------------------------------------

UBool
GregorianCalendar::boundsCheck(int32_t value, UCalendarDateFields field) const
{
    return value >= getMinimum(field) && value <= getMaximum(field);
}

// -------------------------------------

UDate
GregorianCalendar::getEpochDay(UErrorCode& status)
{
    complete(status);
    // Divide by 1000 (convert to seconds) in order to prevent overflow when
    // dealing with UDate(Long.MIN_VALUE) and UDate(Long.MAX_VALUE).
    double wallSec = internalGetTime()/1000 + (internalGet(UCAL_ZONE_OFFSET) + internalGet(UCAL_DST_OFFSET))/1000;

    return ClockMath::floorDivide(wallSec, kOneDay/1000.0);
}

// -------------------------------------


// -------------------------------------

/**
* Compute the julian day number of the day BEFORE the first day of
* January 1, year 1 of the given calendar.  If julianDay == 0, it
* specifies (Jan. 1, 1) - 1, in whatever calendar we are using (Julian
* or Gregorian).
*/
double GregorianCalendar::computeJulianDayOfYear(UBool isGregorian,
                                                 int32_t year, UBool& isLeap)
{
    isLeap = year%4 == 0;
    int32_t y = year - 1;
    double julianDay = 365.0*y + ClockMath::floorDivide(y, 4) + (kJan1_1JulianDay - 3);

    if (isGregorian) {
        isLeap = isLeap && ((year%100 != 0) || (year%400 == 0));
        // Add 2 because Gregorian calendar starts 2 days after Julian calendar
        julianDay += Grego::gregorianShift(year);
    }

    return julianDay;
}

// /**
//  * Compute the day of week, relative to the first day of week, from
//  * 0..6, of the current DOW_LOCAL or DAY_OF_WEEK fields.  This is
//  * equivalent to get(DOW_LOCAL) - 1.
//  */
// int32_t GregorianCalendar::computeRelativeDOW() const {
//     int32_t relDow = 0;
//     if (fStamp[UCAL_DOW_LOCAL] > fStamp[UCAL_DAY_OF_WEEK]) {
//         relDow = internalGet(UCAL_DOW_LOCAL) - 1; // 1-based
//     } else if (fStamp[UCAL_DAY_OF_WEEK] != kUnset) {
//         relDow = internalGet(UCAL_DAY_OF_WEEK) - getFirstDayOfWeek();
//         if (relDow < 0) relDow += 7;
//     }
//     return relDow;
// }

// /**
//  * Compute the day of week, relative to the first day of week,
//  * from 0..6 of the given julian day.
//  */
// int32_t GregorianCalendar::computeRelativeDOW(double julianDay) const {
//   int32_t relDow = julianDayToDayOfWeek(julianDay) - getFirstDayOfWeek();
//     if (relDow < 0) {
//         relDow += 7;
//     }
//     return relDow;
// }

// /**
//  * Compute the DOY using the WEEK_OF_YEAR field and the julian day
//  * of the day BEFORE January 1 of a year (a return value from
//  * computeJulianDayOfYear).
//  */
// int32_t GregorianCalendar::computeDOYfromWOY(double julianDayOfYear) const {
//     // Compute DOY from day of week plus week of year

//     // Find the day of the week for the first of this year.  This
//     // is zero-based, with 0 being the locale-specific first day of
//     // the week.  Add 1 to get first day of year.
//     int32_t fdy = computeRelativeDOW(julianDayOfYear + 1);

//     return
//         // Compute doy of first (relative) DOW of WOY 1
//         (((7 - fdy) < getMinimalDaysInFirstWeek())
//          ? (8 - fdy) : (1 - fdy))

//         // Adjust for the week number.
//         + (7 * (internalGet(UCAL_WEEK_OF_YEAR) - 1))

//         // Adjust for the DOW
//         + computeRelativeDOW();
// }

// -------------------------------------

double
GregorianCalendar::millisToJulianDay(UDate millis)
{
    return (double)kEpochStartAsJulianDay + ClockMath::floorDivide(millis, (double)kOneDay);
}

// -------------------------------------

UDate
GregorianCalendar::julianDayToMillis(double julian)
{
    return (UDate) ((julian - kEpochStartAsJulianDay) * (double) kOneDay);
}

// -------------------------------------

int32_t
GregorianCalendar::aggregateStamp(int32_t stamp_a, int32_t stamp_b)
{
    return (((stamp_a != kUnset && stamp_b != kUnset)
        ? uprv_max(stamp_a, stamp_b)
        : (int32_t)kUnset));
}

// -------------------------------------

/**
* Roll a field by a signed amount.
* Note: This will be made public later. [LIU]
*/

void
GregorianCalendar::roll(EDateFields field, int32_t amount, UErrorCode& status) {
    roll((UCalendarDateFields) field, amount, status);
}

void
GregorianCalendar::roll(UCalendarDateFields field, int32_t amount, UErrorCode& status)
{
    if((amount == 0) || U_FAILURE(status)) {
        return;
    }

    // J81 processing. (gregorian cutover)
    UBool inCutoverMonth = FALSE;
    int32_t cMonthLen=0; // 'c' for cutover; in days
    int32_t cDayOfMonth=0; // no discontinuity: [0, cMonthLen)
    double cMonthStart=0.0; // in ms

    // Common code - see if we're in the cutover month of the cutover year
    if(get(UCAL_EXTENDED_YEAR, status) == fGregorianCutoverYear) {
        switch (field) {
        case UCAL_DAY_OF_MONTH:
        case UCAL_WEEK_OF_MONTH:
            {
                int32_t max = monthLength(internalGet(UCAL_MONTH));
                UDate t = internalGetTime();
                // We subtract 1 from the DAY_OF_MONTH to make it zero-based, and an
                // additional 10 if we are after the cutover. Thus the monthStart
                // value will be correct iff we actually are in the cutover month.
                cDayOfMonth = internalGet(UCAL_DAY_OF_MONTH) - ((t >= fGregorianCutover) ? 10 : 0);
                cMonthStart = t - ((cDayOfMonth - 1) * kOneDay);
                // A month containing the cutover is 10 days shorter.
                if ((cMonthStart < fGregorianCutover) &&
                    (cMonthStart + (cMonthLen=(max-10))*kOneDay >= fGregorianCutover)) {
                        inCutoverMonth = TRUE;
                    }
            }
            break;
        default:
            ;
        }
    }

    switch (field) {
    case UCAL_WEEK_OF_YEAR: {
        // Unlike WEEK_OF_MONTH, WEEK_OF_YEAR never shifts the day of the
        // week.  Also, rolling the week of the year can have seemingly
        // strange effects simply because the year of the week of year
        // may be different from the calendar year.  For example, the
        // date Dec 28, 1997 is the first day of week 1 of 1998 (if
        // weeks start on Sunday and the minimal days in first week is
        // <= 3).
        int32_t woy = get(UCAL_WEEK_OF_YEAR, status);
        // Get the ISO year, which matches the week of year.  This
        // may be one year before or after the calendar year.
        int32_t isoYear = get(UCAL_YEAR_WOY, status);
        int32_t isoDoy = internalGet(UCAL_DAY_OF_YEAR);
        if (internalGet(UCAL_MONTH) == UCAL_JANUARY) {
            if (woy >= 52) {
                isoDoy += handleGetYearLength(isoYear);
            }
        } else {
            if (woy == 1) {
                isoDoy -= handleGetYearLength(isoYear - 1);
            }
        }
        woy += amount;
        // Do fast checks to avoid unnecessary computation:
        if (woy < 1 || woy > 52) {
            // Determine the last week of the ISO year.
            // We do this using the standard formula we use
            // everywhere in this file.  If we can see that the
            // days at the end of the year are going to fall into
            // week 1 of the next year, we drop the last week by
            // subtracting 7 from the last day of the year.
            int32_t lastDoy = handleGetYearLength(isoYear);
            int32_t lastRelDow = (lastDoy - isoDoy + internalGet(UCAL_DAY_OF_WEEK) -
                getFirstDayOfWeek()) % 7;
            if (lastRelDow < 0) lastRelDow += 7;
            if ((6 - lastRelDow) >= getMinimalDaysInFirstWeek()) lastDoy -= 7;
            int32_t lastWoy = weekNumber(lastDoy, lastRelDow + 1);
            woy = ((woy + lastWoy - 1) % lastWoy) + 1;
        }
        set(UCAL_WEEK_OF_YEAR, woy);
        set(UCAL_YEAR_WOY,isoYear);
        return;
                            }

    case UCAL_DAY_OF_MONTH:
        if( !inCutoverMonth ) {
            Calendar::roll(field, amount, status);
            return;
        } else {
            // [j81] 1582 special case for DOM
            // The default computation works except when the current month
            // contains the Gregorian cutover.  We handle this special case
            // here.  [j81 - aliu]
            double monthLen = cMonthLen * kOneDay;
            double msIntoMonth = uprv_fmod(internalGetTime() - cMonthStart +
                amount * kOneDay, monthLen);
            if (msIntoMonth < 0) {
                msIntoMonth += monthLen;
            }
#if defined (U_DEBUG_CAL)
            fprintf(stderr, "%s:%d: roll DOM %d  -> %.0lf ms  \n",
                __FILE__, __LINE__,amount, cMonthLen, cMonthStart+msIntoMonth);
#endif
            setTimeInMillis(cMonthStart + msIntoMonth, status);
            return;
        }

    case UCAL_WEEK_OF_MONTH:
        if( !inCutoverMonth ) {
            Calendar::roll(field, amount, status);
            return;
        } else {
#if defined (U_DEBUG_CAL)
            fprintf(stderr, "%s:%d: roll WOM %d ??????????????????? \n",
                __FILE__, __LINE__,amount);
#endif
            // NOTE: following copied from  the old
            //     GregorianCalendar::roll( WEEK_OF_MONTH )  code

            // This is tricky, because during the roll we may have to shift
            // to a different day of the week.  For example:

            //    s  m  t  w  r  f  s
            //          1  2  3  4  5
            //    6  7  8  9 10 11 12

            // When rolling from the 6th or 7th back one week, we go to the
            // 1st (assuming that the first partial week counts).  The same
            // thing happens at the end of the month.

            // The other tricky thing is that we have to figure out whether
            // the first partial week actually counts or not, based on the
            // minimal first days in the week.  And we have to use the
            // correct first day of the week to delineate the week
            // boundaries.

            // Here's our algorithm.  First, we find the real boundaries of
            // the month.  Then we discard the first partial week if it
            // doesn't count in this locale.  Then we fill in the ends with
            // phantom days, so that the first partial week and the last
            // partial week are full weeks.  We then have a nice square
            // block of weeks.  We do the usual rolling within this block,
            // as is done elsewhere in this method.  If we wind up on one of
            // the phantom days that we added, we recognize this and pin to
            // the first or the last day of the month.  Easy, eh?

            // Another wrinkle: To fix jitterbug 81, we have to make all this
            // work in the oddball month containing the Gregorian cutover.
            // This month is 10 days shorter than usual, and also contains
            // a discontinuity in the days; e.g., the default cutover month
            // is Oct 1582, and goes from day of month 4 to day of month 15.

            // Normalize the DAY_OF_WEEK so that 0 is the first day of the week
            // in this locale.  We have dow in 0..6.
            int32_t dow = internalGet(UCAL_DAY_OF_WEEK) - getFirstDayOfWeek();
            if (dow < 0)
                dow += 7;

            // Find the day of month, compensating for cutover discontinuity.
            int32_t dom = cDayOfMonth;

            // Find the day of the week (normalized for locale) for the first
            // of the month.
            int32_t fdm = (dow - dom + 1) % 7;
            if (fdm < 0)
                fdm += 7;

            // Get the first day of the first full week of the month,
            // including phantom days, if any.  Figure out if the first week
            // counts or not; if it counts, then fill in phantom days.  If
            // not, advance to the first real full week (skip the partial week).
            int32_t start;
            if ((7 - fdm) < getMinimalDaysInFirstWeek())
                start = 8 - fdm; // Skip the first partial week
            else
                start = 1 - fdm; // This may be zero or negative

            // Get the day of the week (normalized for locale) for the last
            // day of the month.
            int32_t monthLen = cMonthLen;
            int32_t ldm = (monthLen - dom + dow) % 7;
            // We know monthLen >= DAY_OF_MONTH so we skip the += 7 step here.

            // Get the limit day for the blocked-off rectangular month; that
            // is, the day which is one past the last day of the month,
            // after the month has already been filled in with phantom days
            // to fill out the last week.  This day has a normalized DOW of 0.
            int32_t limit = monthLen + 7 - ldm;

            // Now roll between start and (limit - 1).
            int32_t gap = limit - start;
            int32_t newDom = (dom + amount*7 - start) % gap;
            if (newDom < 0)
                newDom += gap;
            newDom += start;

            // Finally, pin to the real start and end of the month.
            if (newDom < 1)
                newDom = 1;
            if (newDom > monthLen)
                newDom = monthLen;

            // Set the DAY_OF_MONTH.  We rely on the fact that this field
            // takes precedence over everything else (since all other fields
            // are also set at this point).  If this fact changes (if the
            // disambiguation algorithm changes) then we will have to unset
            // the appropriate fields here so that DAY_OF_MONTH is attended
            // to.

            // If we are in the cutover month, manipulate ms directly.  Don't do
            // this in general because it doesn't work across DST boundaries
            // (details, details).  This takes care of the discontinuity.
            setTimeInMillis(cMonthStart + (newDom-1)*kOneDay, status);
            return;
        }

    default:
        Calendar::roll(field, amount, status);
        return;
    }
}

// -------------------------------------


/**
* Return the minimum value that this field could have, given the current date.
* For the Gregorian calendar, this is the same as getMinimum() and getGreatestMinimum().
* @param field    the time field.
* @return         the minimum value that this field could have, given the current date.
* @deprecated ICU 2.6. Use getActualMinimum(UCalendarDateFields field) instead.
*/
int32_t GregorianCalendar::getActualMinimum(EDateFields field) const
{
    return getMinimum((UCalendarDateFields)field);
}

int32_t GregorianCalendar::getActualMinimum(EDateFields field, UErrorCode& /* status */) const
{
    return getMinimum((UCalendarDateFields)field);
}

/**
* Return the minimum value that this field could have, given the current date.
* For the Gregorian calendar, this is the same as getMinimum() and getGreatestMinimum().
* @param field    the time field.
* @return         the minimum value that this field could have, given the current date.
* @draft ICU 2.6.
*/
int32_t GregorianCalendar::getActualMinimum(UCalendarDateFields field, UErrorCode& /* status */) const
{
    return getMinimum(field);
}


// ------------------------------------

/**
* Old year limits were least max 292269054, max 292278994.
*/

/**
* @stable ICU 2.0
*/
int32_t GregorianCalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const {
    return kGregorianCalendarLimits[field][limitType];
}

/**
* Return the maximum value that this field could have, given the current date.
* For example, with the date "Feb 3, 1997" and the DAY_OF_MONTH field, the actual
* maximum would be 28; for "Feb 3, 1996" it s 29.  Similarly for a Hebrew calendar,
* for some years the actual maximum for MONTH is 12, and for others 13.
* @stable ICU 2.0
*/
int32_t GregorianCalendar::getActualMaximum(UCalendarDateFields field, UErrorCode& status) const
{
    /* It is a known limitation that the code here (and in getActualMinimum)
    * won't behave properly at the extreme limits of GregorianCalendar's
    * representable range (except for the code that handles the YEAR
    * field).  That's because the ends of the representable range are at
    * odd spots in the year.  For calendars with the default Gregorian
    * cutover, these limits are Sun Dec 02 16:47:04 GMT 292269055 BC to Sun
    * Aug 17 07:12:55 GMT 292278994 AD, somewhat different for non-GMT
    * zones.  As a result, if the calendar is set to Aug 1 292278994 AD,
    * the actual maximum of DAY_OF_MONTH is 17, not 30.  If the date is Mar
    * 31 in that year, the actual maximum month might be Jul, whereas is
    * the date is Mar 15, the actual maximum might be Aug -- depending on
    * the precise semantics that are desired.  Similar considerations
    * affect all fields.  Nonetheless, this effect is sufficiently arcane
    * that we permit it, rather than complicating the code to handle such
    * intricacies. - liu 8/20/98

    * UPDATE: No longer true, since we have pulled in the limit values on
    * the year. - Liu 11/6/00 */

    switch (field) {

    case UCAL_YEAR:
        /* The year computation is no different, in principle, from the
        * others, however, the range of possible maxima is large.  In
        * addition, the way we know we've exceeded the range is different.
        * For these reasons, we use the special case code below to handle
        * this field.
        *
        * The actual maxima for YEAR depend on the type of calendar:
        *
        *     Gregorian = May 17, 292275056 BC - Aug 17, 292278994 AD
        *     Julian    = Dec  2, 292269055 BC - Jan  3, 292272993 AD
        *     Hybrid    = Dec  2, 292269055 BC - Aug 17, 292278994 AD
        *
        * We know we've exceeded the maximum when either the month, date,
        * time, or era changes in response to setting the year.  We don't
        * check for month, date, and time here because the year and era are
        * sufficient to detect an invalid year setting.  NOTE: If code is
        * added to check the month and date in the future for some reason,
        * Feb 29 must be allowed to shift to Mar 1 when setting the year.
        */
        {
            if(U_FAILURE(status)) return 0;
            Calendar *cal = clone();
            if(!cal) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return 0;
            }

            cal->setLenient(TRUE);

            int32_t era = cal->get(UCAL_ERA, status);
            UDate d = cal->getTime(status);

            /* Perform a binary search, with the invariant that lowGood is a
            * valid year, and highBad is an out of range year.
            */
            int32_t lowGood = kGregorianCalendarLimits[UCAL_YEAR][1];
            int32_t highBad = kGregorianCalendarLimits[UCAL_YEAR][2]+1;
            while ((lowGood + 1) < highBad) {
                int32_t y = (lowGood + highBad) / 2;
                cal->set(UCAL_YEAR, y);
                if (cal->get(UCAL_YEAR, status) == y && cal->get(UCAL_ERA, status) == era) {
                    lowGood = y;
                } else {
                    highBad = y;
                    cal->setTime(d, status); // Restore original fields
                }
            }

            delete cal;
            return lowGood;
        }

    default:
        return Calendar::getActualMaximum(field,status);
    }
}


int32_t GregorianCalendar::handleGetExtendedYear() {
    // the year to return
    int32_t year = kEpochYear;

    // year field to use
    int32_t yearField = UCAL_EXTENDED_YEAR;

    // There are three separate fields which could be used to
    // derive the proper year.  Use the one most recently set.
    if (fStamp[yearField] < fStamp[UCAL_YEAR])
        yearField = UCAL_YEAR;
    if (fStamp[yearField] < fStamp[UCAL_YEAR_WOY])
        yearField = UCAL_YEAR_WOY;

    // based on the "best" year field, get the year
    switch(yearField) {
    case UCAL_EXTENDED_YEAR:
        year = internalGet(UCAL_EXTENDED_YEAR, kEpochYear);
        break;

    case UCAL_YEAR:
        {
            // The year defaults to the epoch start, the era to AD
            int32_t era = internalGet(UCAL_ERA, AD);
            if (era == BC) {
                year = 1 - internalGet(UCAL_YEAR, 1); // Convert to extended year
            } else {
                year = internalGet(UCAL_YEAR, kEpochYear);
            }
        }
        break;

    case UCAL_YEAR_WOY:
        year = handleGetExtendedYearFromWeekFields(internalGet(UCAL_YEAR_WOY), internalGet(UCAL_WEEK_OF_YEAR));
#if defined (U_DEBUG_CAL)
        //    if(internalGet(UCAL_YEAR_WOY) != year) {
        fprintf(stderr, "%s:%d: hGEYFWF[%d,%d] ->  %d\n",
            __FILE__, __LINE__,internalGet(UCAL_YEAR_WOY),internalGet(UCAL_WEEK_OF_YEAR),year);
        //}
#endif
        break;

    default:
        year = kEpochYear;
    }
    return year;
}

int32_t GregorianCalendar::handleGetExtendedYearFromWeekFields(int32_t yearWoy, int32_t woy)
{
    // convert year to extended form
    int32_t era = internalGet(UCAL_ERA, AD);
    if(era == BC) {
        yearWoy = 1 - yearWoy;
    }
    return Calendar::handleGetExtendedYearFromWeekFields(yearWoy, woy);
}


// -------------------------------------

UBool
GregorianCalendar::inDaylightTime(UErrorCode& status) const
{
    if (U_FAILURE(status) || !getTimeZone().useDaylightTime())
        return FALSE;

    // Force an update of the state of the Calendar.
    ((GregorianCalendar*)this)->complete(status); // cast away const

    return (UBool)(U_SUCCESS(status) ? (internalGet(UCAL_DST_OFFSET) != 0) : FALSE);
}

// -------------------------------------

/**
* Return the ERA.  We need a special method for this because the
* default ERA is AD, but a zero (unset) ERA is BC.
*/
int32_t
GregorianCalendar::internalGetEra() const {
    return isSet(UCAL_ERA) ? internalGet(UCAL_ERA) : (int32_t)AD;
}

const char *
GregorianCalendar::getType() const {
    //static const char kGregorianType = "gregorian";

    return "gregorian";
}

/**
 * The system maintains a static default century start date and Year.  They are
 * initialized the first time they are used.  Once the system default century date
 * and year are set, they do not change.
 */
static UDate           gregocal_gSystemDefaultCenturyStart       = DBL_MIN;
static int32_t         gregocal_gSystemDefaultCenturyStartYear   = -1;
static icu::UInitOnce  gregocal_gSystemDefaultCenturyInit        = U_INITONCE_INITIALIZER;


UBool GregorianCalendar::haveDefaultCentury() const
{
    return TRUE;
}

static void U_CALLCONV
gregocal_initializeSystemDefaultCentury()
{
    // initialize systemDefaultCentury and systemDefaultCenturyYear based
    // on the current time.  They'll be set to 80 years before
    // the current time.
    UErrorCode status = U_ZERO_ERROR;
    GregorianCalendar calendar(status);
    if (U_SUCCESS(status)) {
        calendar.setTime(Calendar::getNow(), status);
        calendar.add(UCAL_YEAR, -80, status);

        gregocal_gSystemDefaultCenturyStart = calendar.getTime(status);
        gregocal_gSystemDefaultCenturyStartYear = calendar.get(UCAL_YEAR, status);
    }
    // We have no recourse upon failure unless we want to propagate the failure
    // out.
}

UDate GregorianCalendar::defaultCenturyStart() const {
    // lazy-evaluate systemDefaultCenturyStart
    umtx_initOnce(gregocal_gSystemDefaultCenturyInit, &gregocal_initializeSystemDefaultCentury);
    return gregocal_gSystemDefaultCenturyStart;
}

int32_t GregorianCalendar::defaultCenturyStartYear() const {
    // lazy-evaluate systemDefaultCenturyStartYear
    umtx_initOnce(gregocal_gSystemDefaultCenturyInit, &gregocal_initializeSystemDefaultCentury);
    return gregocal_gSystemDefaultCenturyStartYear;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 **********************************************************************
 * Copyright (c) 2003-2008, International Business Machines
 * Corporation and others.  All Rights Reserved.
 **********************************************************************
 * Author: Alan Liu
 * Created: September 2 2003
 * Since: ICU 2.8
 **********************************************************************
 */



#if !UCONFIG_NO_FORMATTING






U_NAMESPACE_BEGIN

int32_t ClockMath::floorDivide(int32_t numerator, int32_t denominator) {
    return (numerator >= 0) ?
        numerator / denominator : ((numerator + 1) / denominator) - 1;
}

int64_t ClockMath::floorDivide(int64_t numerator, int64_t denominator) {
    return (numerator >= 0) ?
        numerator / denominator : ((numerator + 1) / denominator) - 1;
}

int32_t ClockMath::floorDivide(double numerator, int32_t denominator,
                          int32_t& remainder) {
    double quotient;
    quotient = uprv_floor(numerator / denominator);
    remainder = (int32_t) (numerator - (quotient * denominator));
    return (int32_t) quotient;
}

double ClockMath::floorDivide(double dividend, double divisor,
                         double& remainder) {
    // Only designed to work for positive divisors
    U_ASSERT(divisor > 0);
    double quotient = floorDivide(dividend, divisor);
    remainder = dividend - (quotient * divisor);
    // N.B. For certain large dividends, on certain platforms, there
    // is a bug such that the quotient is off by one.  If you doubt
    // this to be true, set a breakpoint below and run cintltst.
    if (remainder < 0 || remainder >= divisor) {
        // E.g. 6.7317038241449352e+022 / 86400000.0 is wrong on my
        // machine (too high by one).  4.1792057231752762e+024 /
        // 86400000.0 is wrong the other way (too low).
        double q = quotient;
        quotient += (remainder < 0) ? -1 : +1;
        if (q == quotient) {
            // For quotients > ~2^53, we won't be able to add or
            // subtract one, since the LSB of the mantissa will be >
            // 2^0; that is, the exponent (base 2) will be larger than
            // the length, in bits, of the mantissa.  In that case, we
            // can't give a correct answer, so we set the remainder to
            // zero.  This has the desired effect of making extreme
            // values give back an approximate answer rather than
            // crashing.  For example, UDate values above a ~10^25
            // might all have a time of midnight.
            remainder = 0;
        } else {
            remainder = dividend - (quotient * divisor);
        }
    }
    U_ASSERT(0 <= remainder && remainder < divisor);
    return quotient;
}

const int32_t JULIAN_1_CE    = 1721426; // January 1, 1 CE Gregorian
const int32_t JULIAN_1970_CE = 2440588; // January 1, 1970 CE Gregorian

const int16_t Grego::DAYS_BEFORE[24] =
    {0,31,59,90,120,151,181,212,243,273,304,334,
     0,31,60,91,121,152,182,213,244,274,305,335};

const int8_t Grego::MONTH_LENGTH[24] =
    {31,28,31,30,31,30,31,31,30,31,30,31,
     31,29,31,30,31,30,31,31,30,31,30,31};

double Grego::fieldsToDay(int32_t year, int32_t month, int32_t dom) {

    int32_t y = year - 1;

    double julian = 365 * y + ClockMath::floorDivide(y, 4) + (JULIAN_1_CE - 3) + // Julian cal
        ClockMath::floorDivide(y, 400) - ClockMath::floorDivide(y, 100) + 2 + // => Gregorian cal
        DAYS_BEFORE[month + (isLeapYear(year) ? 12 : 0)] + dom; // => month/dom

    return julian - JULIAN_1970_CE; // JD => epoch day
}

void Grego::dayToFields(double day, int32_t& year, int32_t& month,
                        int32_t& dom, int32_t& dow, int32_t& doy) {

    // Convert from 1970 CE epoch to 1 CE epoch (Gregorian calendar)
    day += JULIAN_1970_CE - JULIAN_1_CE;

    // Convert from the day number to the multiple radix
    // representation.  We use 400-year, 100-year, and 4-year cycles.
    // For example, the 4-year cycle has 4 years + 1 leap day; giving
    // 1461 == 365*4 + 1 days.
    int32_t n400 = ClockMath::floorDivide(day, 146097, doy); // 400-year cycle length
    int32_t n100 = ClockMath::floorDivide(doy, 36524, doy); // 100-year cycle length
    int32_t n4   = ClockMath::floorDivide(doy, 1461, doy); // 4-year cycle length
    int32_t n1   = ClockMath::floorDivide(doy, 365, doy);
    year = 400*n400 + 100*n100 + 4*n4 + n1;
    if (n100 == 4 || n1 == 4) {
        doy = 365; // Dec 31 at end of 4- or 400-year cycle
    } else {
        ++year;
    }
    
    UBool isLeap = isLeapYear(year);
    
    // Gregorian day zero is a Monday.
    dow = (int32_t) uprv_fmod(day + 1, 7);
    dow += (dow < 0) ? (UCAL_SUNDAY + 7) : UCAL_SUNDAY;

    // Common Julian/Gregorian calculation
    int32_t correction = 0;
    int32_t march1 = isLeap ? 60 : 59; // zero-based DOY for March 1
    if (doy >= march1) {
        correction = isLeap ? 1 : 2;
    }
    month = (12 * (doy + correction) + 6) / 367; // zero-based month
    dom = doy - DAYS_BEFORE[month + (isLeap ? 12 : 0)] + 1; // one-based DOM
    doy++; // one-based doy
}

void Grego::timeToFields(UDate time, int32_t& year, int32_t& month,
                        int32_t& dom, int32_t& dow, int32_t& doy, int32_t& mid) {
    double millisInDay;
    double day = ClockMath::floorDivide((double)time, (double)U_MILLIS_PER_DAY, millisInDay);
    mid = (int32_t)millisInDay;
    dayToFields(day, year, month, dom, dow, doy);
}

int32_t Grego::dayOfWeek(double day) {
    int32_t dow;
    ClockMath::floorDivide(day + UCAL_THURSDAY, 7, dow);
    return (dow == 0) ? UCAL_SATURDAY : dow;
}

int32_t Grego::dayOfWeekInMonth(int32_t year, int32_t month, int32_t dom) {
    int32_t weekInMonth = (dom + 6)/7;
    if (weekInMonth == 4) {
        if (dom + 7 > monthLength(year, month)) {
            weekInMonth = -1;
        }
    } else if (weekInMonth == 5) {
        weekInMonth = -1;
    }
    return weekInMonth;
}

U_NAMESPACE_END

#endif
//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2003-2016, International Business Machines Corporation
* and others. All Rights Reserved.
******************************************************************************
*
* File HEBRWCAL.CPP
*
* Modification History:
*
*   Date        Name        Description
*   12/03/2003  srl         ported from java HebrewCalendar
*****************************************************************************
*/



#if !UCONFIG_NO_FORMATTING



#include <float.h>
 // Math
 // CalendarAstronomer



// Hebrew Calendar implementation

/**
* The absolute date, in milliseconds since 1/1/1970 AD, Gregorian,
* of the start of the Hebrew calendar.  In order to keep this calendar's
* time of day in sync with that of the Gregorian calendar, we use
* midnight, rather than sunset the day before.
*/
//static const double EPOCH_MILLIS = -180799862400000.; // 1/1/1 HY

static const int32_t HEBRWCAL_LIMITS[UCAL_FIELD_COUNT][4] = {
    // Minimum  Greatest    Least  Maximum
    //           Minimum  Maximum
    {        0,        0,        0,        0}, // ERA
    { -5000000, -5000000,  5000000,  5000000}, // YEAR
    {        0,        0,       12,       12}, // MONTH
    {        1,        1,       51,       56}, // WEEK_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // WEEK_OF_MONTH
    {        1,        1,       29,       30}, // DAY_OF_MONTH
    {        1,        1,      353,      385}, // DAY_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DAY_OF_WEEK
    {       -1,       -1,        5,        5}, // DAY_OF_WEEK_IN_MONTH
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // AM_PM
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR_OF_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MINUTE
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // SECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // ZONE_OFFSET
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DST_OFFSET
    { -5000000, -5000000,  5000000,  5000000}, // YEAR_WOY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DOW_LOCAL
    { -5000000, -5000000,  5000000,  5000000}, // EXTENDED_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // JULIAN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECONDS_IN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // IS_LEAP_MONTH
};

/**
* The lengths of the Hebrew months.  This is complicated, because there
* are three different types of years, or six if you count leap years.
* Due to the rules for postponing the start of the year to avoid having
* certain holidays fall on the sabbath, the year can end up being three
* different lengths, called "deficient", "normal", and "complete".
*/
static const int8_t MONTH_LENGTH[][3] = {
    // Deficient  Normal     Complete
    {   30,         30,         30     },           //Tishri
    {   29,         29,         30     },           //Heshvan
    {   29,         30,         30     },           //Kislev
    {   29,         29,         29     },           //Tevet
    {   30,         30,         30     },           //Shevat
    {   30,         30,         30     },           //Adar I (leap years only)
    {   29,         29,         29     },           //Adar
    {   30,         30,         30     },           //Nisan
    {   29,         29,         29     },           //Iyar
    {   30,         30,         30     },           //Sivan
    {   29,         29,         29     },           //Tammuz
    {   30,         30,         30     },           //Av
    {   29,         29,         29     },           //Elul
};

/**
* The cumulative # of days to the end of each month in a non-leap year
* Although this can be calculated from the MONTH_LENGTH table,
* keeping it around separately makes some calculations a lot faster
*/

static const int16_t MONTH_START[][3] = {
    // Deficient  Normal     Complete
    {    0,          0,          0  },          // (placeholder)
    {   30,         30,         30  },          // Tishri
    {   59,         59,         60  },          // Heshvan
    {   88,         89,         90  },          // Kislev
    {  117,        118,        119  },          // Tevet
    {  147,        148,        149  },          // Shevat
    {  147,        148,        149  },          // (Adar I placeholder)
    {  176,        177,        178  },          // Adar
    {  206,        207,        208  },          // Nisan
    {  235,        236,        237  },          // Iyar
    {  265,        266,        267  },          // Sivan
    {  294,        295,        296  },          // Tammuz
    {  324,        325,        326  },          // Av
    {  353,        354,        355  },          // Elul
};

/**
* The cumulative # of days to the end of each month in a leap year
*/
static const int16_t  LEAP_MONTH_START[][3] = {
    // Deficient  Normal     Complete
    {    0,          0,          0  },          // (placeholder)
    {   30,         30,         30  },          // Tishri
    {   59,         59,         60  },          // Heshvan
    {   88,         89,         90  },          // Kislev
    {  117,        118,        119  },          // Tevet
    {  147,        148,        149  },          // Shevat
    {  177,        178,        179  },          // Adar I
    {  206,        207,        208  },          // Adar II
    {  236,        237,        238  },          // Nisan
    {  265,        266,        267  },          // Iyar
    {  295,        296,        297  },          // Sivan
    {  324,        325,        326  },          // Tammuz
    {  354,        355,        356  },          // Av
    {  383,        384,        385  },          // Elul
};

static icu::CalendarCache *hebrwcal_gCache =  NULL;

U_CDECL_BEGIN
static UBool calendar_hebrew_cleanup(void) {
    delete hebrwcal_gCache;
    hebrwcal_gCache = NULL;
    return TRUE;
}
U_CDECL_END

U_NAMESPACE_BEGIN
//-------------------------------------------------------------------------
// Constructors...
//-------------------------------------------------------------------------

/**
* Constructs a default <code>HebrewCalendar</code> using the current time
* in the default time zone with the default locale.
* @internal
*/
HebrewCalendar::HebrewCalendar(const Locale& aLocale, UErrorCode& success)
:   Calendar(TimeZone::createDefault(), aLocale, success)

{
    setTimeInMillis(getNow(), success); // Call this again now that the vtable is set up properly.
}


HebrewCalendar::~HebrewCalendar() {
}

const char *HebrewCalendar::getType() const {
    return "hebrew";
}

HebrewCalendar* HebrewCalendar::clone() const {
    return new HebrewCalendar(*this);
}

HebrewCalendar::HebrewCalendar(const HebrewCalendar& other) : Calendar(other) {
}


//-------------------------------------------------------------------------
// Rolling and adding functions overridden from Calendar
//
// These methods call through to the default implementation in IBMCalendar
// for most of the fields and only handle the unusual ones themselves.
//-------------------------------------------------------------------------

/**
* Add a signed amount to a specified field, using this calendar's rules.
* For example, to add three days to the current date, you can call
* <code>add(Calendar.DATE, 3)</code>.
* <p>
* When adding to certain fields, the values of other fields may conflict and
* need to be changed.  For example, when adding one to the {@link #MONTH MONTH} field
* for the date "30 Av 5758", the {@link #DAY_OF_MONTH DAY_OF_MONTH} field
* must be adjusted so that the result is "29 Elul 5758" rather than the invalid
* "30 Elul 5758".
* <p>
* This method is able to add to
* all fields except for {@link #ERA ERA}, {@link #DST_OFFSET DST_OFFSET},
* and {@link #ZONE_OFFSET ZONE_OFFSET}.
* <p>
* <b>Note:</b> You should always use {@link #roll roll} and add rather
* than attempting to perform arithmetic operations directly on the fields
* of a <tt>HebrewCalendar</tt>.  Since the {@link #MONTH MONTH} field behaves
* discontinuously in non-leap years, simple arithmetic can give invalid results.
* <p>
* @param field     the time field.
* @param amount    the amount to add to the field.
*
* @exception   IllegalArgumentException if the field is invalid or refers
*              to a field that cannot be handled by this method.
* @internal
*/
void HebrewCalendar::add(UCalendarDateFields field, int32_t amount, UErrorCode& status)
{
    if(U_FAILURE(status)) {
        return;
    }
    switch (field) {
  case UCAL_MONTH:
      {
          // We can't just do a set(MONTH, get(MONTH) + amount).  The
          // reason is ADAR_1.  Suppose amount is +2 and we land in
          // ADAR_1 -- then we have to bump to ADAR_2 aka ADAR.  But
          // if amount is -2 and we land in ADAR_1, then we have to
          // bump the other way -- down to SHEVAT.  - Alan 11/00
          int32_t month = get(UCAL_MONTH, status);
          int32_t year = get(UCAL_YEAR, status);
          UBool acrossAdar1;
          if (amount > 0) {
              acrossAdar1 = (month < ADAR_1); // started before ADAR_1?
              month += amount;
              for (;;) {
                  if (acrossAdar1 && month>=ADAR_1 && !isLeapYear(year)) {
                      ++month;
                  }
                  if (month <= ELUL) {
                      break;
                  }
                  month -= ELUL+1;
                  ++year;
                  acrossAdar1 = TRUE;
              }
          } else {
              acrossAdar1 = (month > ADAR_1); // started after ADAR_1?
              month += amount;
              for (;;) {
                  if (acrossAdar1 && month<=ADAR_1 && !isLeapYear(year)) {
                      --month;
                  }
                  if (month >= 0) {
                      break;
                  }
                  month += ELUL+1;
                  --year;
                  acrossAdar1 = TRUE;
              }
          }
          set(UCAL_MONTH, month);
          set(UCAL_YEAR, year);
          pinField(UCAL_DAY_OF_MONTH, status);
          break;
      }

  default:
      Calendar::add(field, amount, status);
      break;
    }
}

/**
* @deprecated ICU 2.6 use UCalendarDateFields instead of EDateFields
*/
void HebrewCalendar::add(EDateFields field, int32_t amount, UErrorCode& status)
{
    add((UCalendarDateFields)field, amount, status);
}

/**
* Rolls (up/down) a specified amount time on the given field.  For
* example, to roll the current date up by three days, you can call
* <code>roll(Calendar.DATE, 3)</code>.  If the
* field is rolled past its maximum allowable value, it will "wrap" back
* to its minimum and continue rolling.
* For example, calling <code>roll(Calendar.DATE, 10)</code>
* on a Hebrew calendar set to "25 Av 5758" will result in the date "5 Av 5758".
* <p>
* When rolling certain fields, the values of other fields may conflict and
* need to be changed.  For example, when rolling the {@link #MONTH MONTH} field
* upward by one for the date "30 Av 5758", the {@link #DAY_OF_MONTH DAY_OF_MONTH} field
* must be adjusted so that the result is "29 Elul 5758" rather than the invalid
* "30 Elul".
* <p>
* This method is able to roll
* all fields except for {@link #ERA ERA}, {@link #DST_OFFSET DST_OFFSET},
* and {@link #ZONE_OFFSET ZONE_OFFSET}.  Subclasses may, of course, add support for
* additional fields in their overrides of <code>roll</code>.
* <p>
* <b>Note:</b> You should always use roll and {@link #add add} rather
* than attempting to perform arithmetic operations directly on the fields
* of a <tt>HebrewCalendar</tt>.  Since the {@link #MONTH MONTH} field behaves
* discontinuously in non-leap years, simple arithmetic can give invalid results.
* <p>
* @param field     the time field.
* @param amount    the amount by which the field should be rolled.
*
* @exception   IllegalArgumentException if the field is invalid or refers
*              to a field that cannot be handled by this method.
* @internal
*/
void HebrewCalendar::roll(UCalendarDateFields field, int32_t amount, UErrorCode& status)
{
    if(U_FAILURE(status)) {
        return;
    }
    switch (field) {
  case UCAL_MONTH:
      {
          int32_t month = get(UCAL_MONTH, status);
          int32_t year = get(UCAL_YEAR, status);

          UBool leapYear = isLeapYear(year);
          int32_t yearLength = monthsInYear(year);
          int32_t newMonth = month + (amount % yearLength);
          //
          // If it's not a leap year and we're rolling past the missing month
          // of ADAR_1, we need to roll an extra month to make up for it.
          //
          if (!leapYear) {
              if (amount > 0 && month < ADAR_1 && newMonth >= ADAR_1) {
                  newMonth++;
              } else if (amount < 0 && month > ADAR_1 && newMonth <= ADAR_1) {
                  newMonth--;
              }
          }
          set(UCAL_MONTH, (newMonth + 13) % 13);
          pinField(UCAL_DAY_OF_MONTH, status);
          return;
      }
  default:
      Calendar::roll(field, amount, status);
    }
}

void HebrewCalendar::roll(EDateFields field, int32_t amount, UErrorCode& status) {
    roll((UCalendarDateFields)field, amount, status);
}

//-------------------------------------------------------------------------
// Support methods
//-------------------------------------------------------------------------

// Hebrew date calculations are performed in terms of days, hours, and
// "parts" (or halakim), which are 1/1080 of an hour, or 3 1/3 seconds.
static const int32_t HOUR_PARTS = 1080;
static const int32_t DAY_PARTS  = 24*HOUR_PARTS;

// An approximate value for the length of a lunar month.
// It is used to calculate the approximate year and month of a given
// absolute date.
static const int32_t  MONTH_DAYS = 29;
static const int32_t MONTH_FRACT = 12*HOUR_PARTS + 793;
static const int32_t MONTH_PARTS = MONTH_DAYS*DAY_PARTS + MONTH_FRACT;

// The time of the new moon (in parts) on 1 Tishri, year 1 (the epoch)
// counting from noon on the day before.  BAHARAD is an abbreviation of
// Bet (Monday), Hey (5 hours from sunset), Resh-Daled (204).
static const int32_t BAHARAD = 11*HOUR_PARTS + 204;

/**
* Finds the day # of the first day in the given Hebrew year.
* To do this, we want to calculate the time of the Tishri 1 new moon
* in that year.
* <p>
* The algorithm here is similar to ones described in a number of
* references, including:
* <ul>
* <li>"Calendrical Calculations", by Nachum Dershowitz & Edward Reingold,
*     Cambridge University Press, 1997, pages 85-91.
*
* <li>Hebrew Calendar Science and Myths,
*     <a href="http://www.geocities.com/Athens/1584/">
*     http://www.geocities.com/Athens/1584/</a>
*
* <li>The Calendar FAQ,
*      <a href="http://www.faqs.org/faqs/calendars/faq/">
*      http://www.faqs.org/faqs/calendars/faq/</a>
* </ul>
*/
int32_t HebrewCalendar::startOfYear(int32_t year, UErrorCode &status)
{
    ucln_i18n_registerCleanup(UCLN_I18N_HEBREW_CALENDAR, calendar_hebrew_cleanup);
    int32_t day = CalendarCache::get(&hebrwcal_gCache, year, status);

    if (day == 0) {
        int32_t months = (235 * year - 234) / 19;           // # of months before year

        int64_t frac = (int64_t)months * MONTH_FRACT + BAHARAD;  // Fractional part of day #
        day  = months * 29 + (int32_t)(frac / DAY_PARTS);        // Whole # part of calculation
        frac = frac % DAY_PARTS;                        // Time of day

        int32_t wd = (day % 7);                        // Day of week (0 == Monday)

        if (wd == 2 || wd == 4 || wd == 6) {
            // If the 1st is on Sun, Wed, or Fri, postpone to the next day
            day += 1;
            wd = (day % 7);
        }
        if (wd == 1 && frac > 15*HOUR_PARTS+204 && !isLeapYear(year) ) {
            // If the new moon falls after 3:11:20am (15h204p from the previous noon)
            // on a Tuesday and it is not a leap year, postpone by 2 days.
            // This prevents 356-day years.
            day += 2;
        }
        else if (wd == 0 && frac > 21*HOUR_PARTS+589 && isLeapYear(year-1) ) {
            // If the new moon falls after 9:32:43 1/3am (21h589p from yesterday noon)
            // on a Monday and *last* year was a leap year, postpone by 1 day.
            // Prevents 382-day years.
            day += 1;
        }
        CalendarCache::put(&hebrwcal_gCache, year, day, status);
    }
    return day;
}

/**
* Find the day of the week for a given day
*
* @param day   The # of days since the start of the Hebrew calendar,
*              1-based (i.e. 1/1/1 AM is day 1).
*/
int32_t HebrewCalendar::absoluteDayToDayOfWeek(int32_t day)
{
    // We know that 1/1/1 AM is a Monday, which makes the math easy...
    return (day % 7) + 1;
}

/**
* Returns the the type of a given year.
*  0   "Deficient" year with 353 or 383 days
*  1   "Normal"    year with 354 or 384 days
*  2   "Complete"  year with 355 or 385 days
*/
int32_t HebrewCalendar::yearType(int32_t year) const
{
    int32_t yearLength = handleGetYearLength(year);

    if (yearLength > 380) {
        yearLength -= 30;        // Subtract length of leap month.
    }

    int type = 0;

    switch (yearLength) {
  case 353:
      type = 0; break;
  case 354:
      type = 1; break;
  case 355:
      type = 2; break;
  default:
      //throw new RuntimeException("Illegal year length " + yearLength + " in year " + year);
      type = 1;
    }
    return type;
}

/**
* Determine whether a given Hebrew year is a leap year
*
* The rule here is that if (year % 19) == 0, 3, 6, 8, 11, 14, or 17.
* The formula below performs the same test, believe it or not.
*/
UBool HebrewCalendar::isLeapYear(int32_t year) {
    //return (year * 12 + 17) % 19 >= 12;
    int32_t x = (year*12 + 17) % 19;
    return x >= ((x < 0) ? -7 : 12);
}

int32_t HebrewCalendar::monthsInYear(int32_t year) {
    return isLeapYear(year) ? 13 : 12;
}

//-------------------------------------------------------------------------
// Calendar framework
//-------------------------------------------------------------------------

/**
* @internal
*/
int32_t HebrewCalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const {
    return HEBRWCAL_LIMITS[field][limitType];
}

/**
* Returns the length of the given month in the given year
* @internal
*/
int32_t HebrewCalendar::handleGetMonthLength(int32_t extendedYear, int32_t month) const {
    // Resolve out-of-range months.  This is necessary in order to
    // obtain the correct year.  We correct to
    // a 12- or 13-month year (add/subtract 12 or 13, depending
    // on the year) but since we _always_ number from 0..12, and
    // the leap year determines whether or not month 5 (Adar 1)
    // is present, we allow 0..12 in any given year.
    while (month < 0) {
        month += monthsInYear(--extendedYear);
    }
    // Careful: allow 0..12 in all years
    while (month > 12) {
        month -= monthsInYear(extendedYear++);
    }

    switch (month) {
    case HESHVAN:
    case KISLEV:
      // These two month lengths can vary
      return MONTH_LENGTH[month][yearType(extendedYear)];

    default:
      // The rest are a fixed length
      return MONTH_LENGTH[month][0];
    }
}

/**
* Returns the number of days in the given Hebrew year
* @internal
*/
int32_t HebrewCalendar::handleGetYearLength(int32_t eyear) const {
    UErrorCode status = U_ZERO_ERROR;
    return startOfYear(eyear+1, status) - startOfYear(eyear, status);
}

void HebrewCalendar::validateField(UCalendarDateFields field, UErrorCode &status) {
    if (field == UCAL_MONTH && !isLeapYear(handleGetExtendedYear()) && internalGet(UCAL_MONTH) == ADAR_1) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    Calendar::validateField(field, status);
}
//-------------------------------------------------------------------------
// Functions for converting from milliseconds to field values
//-------------------------------------------------------------------------

/**
* Subclasses may override this method to compute several fields
* specific to each calendar system.  These are:
*
* <ul><li>ERA
* <li>YEAR
* <li>MONTH
* <li>DAY_OF_MONTH
* <li>DAY_OF_YEAR
* <li>EXTENDED_YEAR</ul>
*
* Subclasses can refer to the DAY_OF_WEEK and DOW_LOCAL fields,
* which will be set when this method is called.  Subclasses can
* also call the getGregorianXxx() methods to obtain Gregorian
* calendar equivalents for the given Julian day.
*
* <p>In addition, subclasses should compute any subclass-specific
* fields, that is, fields from BASE_FIELD_COUNT to
* getFieldCount() - 1.
* @internal
*/
void HebrewCalendar::handleComputeFields(int32_t julianDay, UErrorCode &status) {
    int32_t d = julianDay - 347997;
    double m = ((d * (double)DAY_PARTS)/ (double) MONTH_PARTS);         // Months (approx)
    int32_t year = (int32_t)( ((19. * m + 234.) / 235.) + 1.);     // Years (approx)
    int32_t ys  = startOfYear(year, status);                   // 1st day of year
    int32_t dayOfYear = (d - ys);

    // Because of the postponement rules, it's possible to guess wrong.  Fix it.
    while (dayOfYear < 1) {
        year--;
        ys  = startOfYear(year, status);
        dayOfYear = (d - ys);
    }

    // Now figure out which month we're in, and the date within that month
    int32_t type = yearType(year);
    UBool isLeap = isLeapYear(year);

    int32_t month = 0;
    int32_t momax = UPRV_LENGTHOF(MONTH_START);
    while (month < momax && dayOfYear > (  isLeap ? LEAP_MONTH_START[month][type] : MONTH_START[month][type] ) ) {
        month++;
    }
    if (month >= momax || month<=0) {
        // TODO: I found dayOfYear could be out of range when
        // a large value is set to julianDay.  I patched startOfYear
        // to reduce the chace, but it could be still reproduced either
        // by startOfYear or other places.  For now, we check
        // the month is in valid range to avoid out of array index
        // access problem here.  However, we need to carefully review
        // the calendar implementation to check the extreme limit of
        // each calendar field and the code works well for any values
        // in the valid value range.  -yoshito
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    month--;
    int dayOfMonth = dayOfYear - (isLeap ? LEAP_MONTH_START[month][type] : MONTH_START[month][type]);

    internalSet(UCAL_ERA, 0);
    internalSet(UCAL_YEAR, year);
    internalSet(UCAL_EXTENDED_YEAR, year);
    internalSet(UCAL_MONTH, month);
    internalSet(UCAL_DAY_OF_MONTH, dayOfMonth);
    internalSet(UCAL_DAY_OF_YEAR, dayOfYear);
}

//-------------------------------------------------------------------------
// Functions for converting from field values to milliseconds
//-------------------------------------------------------------------------

/**
* @internal
*/
int32_t HebrewCalendar::handleGetExtendedYear() {
    int32_t year;
    if (newerField(UCAL_EXTENDED_YEAR, UCAL_YEAR) == UCAL_EXTENDED_YEAR) {
        year = internalGet(UCAL_EXTENDED_YEAR, 1); // Default to year 1
    } else {
        year = internalGet(UCAL_YEAR, 1); // Default to year 1
    }
    return year;
}

/**
* Return JD of start of given month/year.
* @internal
*/
int32_t HebrewCalendar::handleComputeMonthStart(int32_t eyear, int32_t month, UBool /*useMonth*/) const {
    UErrorCode status = U_ZERO_ERROR;
    // Resolve out-of-range months.  This is necessary in order to
    // obtain the correct year.  We correct to
    // a 12- or 13-month year (add/subtract 12 or 13, depending
    // on the year) but since we _always_ number from 0..12, and
    // the leap year determines whether or not month 5 (Adar 1)
    // is present, we allow 0..12 in any given year.
    while (month < 0) {
        month += monthsInYear(--eyear);
    }
    // Careful: allow 0..12 in all years
    while (month > 12) {
        month -= monthsInYear(eyear++);
    }

    int32_t day = startOfYear(eyear, status);

    if(U_FAILURE(status)) {
        return 0;
    }

    if (month != 0) {
        if (isLeapYear(eyear)) {
            day += LEAP_MONTH_START[month][yearType(eyear)];
        } else {
            day += MONTH_START[month][yearType(eyear)];
        }
    }

    return (int) (day + 347997);
}

UBool
HebrewCalendar::inDaylightTime(UErrorCode& status) const
{
    // copied from GregorianCalendar
    if (U_FAILURE(status) || !getTimeZone().useDaylightTime())
        return FALSE;

    // Force an update of the state of the Calendar.
    ((HebrewCalendar*)this)->complete(status); // cast away const

    return (UBool)(U_SUCCESS(status) ? (internalGet(UCAL_DST_OFFSET) != 0) : FALSE);
}

/**
 * The system maintains a static default century start date and Year.  They are
 * initialized the first time they are used.  Once the system default century date
 * and year are set, they do not change.
 */
static UDate           hebrwcal_gSystemDefaultCenturyStart       = DBL_MIN;
static int32_t         hebrwcal_gSystemDefaultCenturyStartYear   = -1;
static icu::UInitOnce  hebrwcal_gSystemDefaultCenturyInit        = U_INITONCE_INITIALIZER;

UBool HebrewCalendar::haveDefaultCentury() const
{
    return TRUE;
}

static void U_CALLCONV hebrwcal_initializeSystemDefaultCentury()
{
    // initialize systemDefaultCentury and systemDefaultCenturyYear based
    // on the current time.  They'll be set to 80 years before
    // the current time.
    UErrorCode status = U_ZERO_ERROR;
    HebrewCalendar calendar(Locale("@calendar=hebrew"),status);
    if (U_SUCCESS(status)) {
        calendar.setTime(Calendar::getNow(), status);
        calendar.add(UCAL_YEAR, -80, status);

        hebrwcal_gSystemDefaultCenturyStart = calendar.getTime(status);
        hebrwcal_gSystemDefaultCenturyStartYear = calendar.get(UCAL_YEAR, status);
    }
    // We have no recourse upon failure unless we want to propagate the failure
    // out.
}


UDate HebrewCalendar::defaultCenturyStart() const {
    // lazy-evaluate systemDefaultCenturyStart
    umtx_initOnce(hebrwcal_gSystemDefaultCenturyInit, &hebrwcal_initializeSystemDefaultCentury);
    return hebrwcal_gSystemDefaultCenturyStart;
}

int32_t HebrewCalendar::defaultCenturyStartYear() const {
    // lazy-evaluate systemDefaultCenturyStartYear
    umtx_initOnce(hebrwcal_gSystemDefaultCenturyInit, &hebrwcal_initializeSystemDefaultCentury);
    return hebrwcal_gSystemDefaultCenturyStartYear;
}


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(HebrewCalendar)

U_NAMESPACE_END

#endif // UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 * Copyright (C) 2003-2014, International Business Machines Corporation
 * and others. All Rights Reserved.
 ******************************************************************************
 *
 * File INDIANCAL.CPP
 *****************************************************************************
 */


#include <stdlib.h>
#if !UCONFIG_NO_FORMATTING


#include <float.h>
 // Math
 // CalendarAstronomer


// Debugging
#ifdef U_DEBUG_INDIANCAL
#include <stdio.h>
#include <stdarg.h>

#endif

U_NAMESPACE_BEGIN

// Implementation of the IndianCalendar class

//-------------------------------------------------------------------------
// Constructors...
//-------------------------------------------------------------------------


IndianCalendar* IndianCalendar::clone() const {
  return new IndianCalendar(*this);
}

IndianCalendar::IndianCalendar(const Locale& aLocale, UErrorCode& success)
  :   Calendar(TimeZone::createDefault(), aLocale, success)
{
  setTimeInMillis(getNow(), success); // Call this again now that the vtable is set up properly.
}

IndianCalendar::IndianCalendar(const IndianCalendar& other) : Calendar(other) {
}

IndianCalendar::~IndianCalendar()
{
}
const char *IndianCalendar::getType() const {
   return "indian";
}

static const int32_t INDIANCAL_LIMITS[UCAL_FIELD_COUNT][4] = {
    // Minimum  Greatest     Least   Maximum
    //           Minimum   Maximum
    {        0,        0,        0,        0}, // ERA
    { -5000000, -5000000,  5000000,  5000000}, // YEAR
    {        0,        0,       11,       11}, // MONTH
    {        1,        1,       52,       53}, // WEEK_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // WEEK_OF_MONTH
    {        1,        1,       30,       31}, // DAY_OF_MONTH
    {        1,        1,      365,      366}, // DAY_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DAY_OF_WEEK
    {       -1,       -1,        5,        5}, // DAY_OF_WEEK_IN_MONTH
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // AM_PM
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR_OF_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MINUTE
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // SECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // ZONE_OFFSET
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DST_OFFSET
    { -5000000, -5000000,  5000000,  5000000}, // YEAR_WOY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DOW_LOCAL
    { -5000000, -5000000,  5000000,  5000000}, // EXTENDED_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // JULIAN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECONDS_IN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // IS_LEAP_MONTH
};

static const double JULIAN_EPOCH = 1721425.5;
static const int32_t INDIAN_ERA_START  = 78;
static const int32_t INDIAN_YEAR_START = 80;

int32_t IndianCalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const {
  return INDIANCAL_LIMITS[field][limitType];
}

/*
 * Determine whether the given gregorian year is a Leap year
 */
static UBool isGregorianLeap(int32_t year)
{
    return ((year % 4) == 0) && (!(((year % 100) == 0) && ((year % 400) != 0)));
}

//----------------------------------------------------------------------
// Calendar framework
//----------------------------------------------------------------------

/*
 * Return the length (in days) of the given month.
 *
 * @param eyear  The year in Saka Era
 * @param month  The month(0-based) in Indian calendar
 */
int32_t IndianCalendar::handleGetMonthLength(int32_t eyear, int32_t month) const {
   if (month < 0 || month > 11) {
      eyear += ClockMath::floorDivide(month, 12, month);
   }

   if (isGregorianLeap(eyear + INDIAN_ERA_START) && month == 0) {
       return 31;
   }

   if (month >= 1 && month <= 5) {
       return 31;
   }

   return 30;
}

/*
 * Return the number of days in the given Indian year
 *
 * @param eyear The year in Saka Era.
 */
int32_t IndianCalendar::handleGetYearLength(int32_t eyear) const {
    return isGregorianLeap(eyear + INDIAN_ERA_START) ? 366 : 365;
}
/*
 * Returns the Julian Day corresponding to gregorian date
 *
 * @param year The Gregorian year
 * @param month The month in Gregorian Year
 * @param date The date in Gregorian day in month
 */
static double gregorianToJD(int32_t year, int32_t month, int32_t date) {
   double julianDay = (JULIAN_EPOCH - 1) +
      (365 * (year - 1)) +
      uprv_floor((year - 1) / 4) +
      (-uprv_floor((year - 1) / 100)) +
      uprv_floor((year - 1) / 400) +
      uprv_floor((((367 * month) - 362) / 12) +
            ((month <= 2) ? 0 :
             (isGregorianLeap(year) ? -1 : -2)
            ) +
            date);

   return julianDay;
}

/*
 * Returns the Gregorian Date corresponding to a given Julian Day
 * @param jd The Julian Day
 */
static int32_t* jdToGregorian(double jd, int32_t gregorianDate[3]) {
   double wjd, depoch, quadricent, dqc, cent, dcent, quad, dquad, yindex, yearday, leapadj;
   int32_t year, month, day;
   wjd = uprv_floor(jd - 0.5) + 0.5;
   depoch = wjd - JULIAN_EPOCH;
   quadricent = uprv_floor(depoch / 146097);
   dqc = (int32_t)uprv_floor(depoch) % 146097;
   cent = uprv_floor(dqc / 36524);
   dcent = (int32_t)uprv_floor(dqc) % 36524;
   quad = uprv_floor(dcent / 1461);
   dquad = (int32_t)uprv_floor(dcent) % 1461;
   yindex = uprv_floor(dquad / 365);
   year = (int32_t)((quadricent * 400) + (cent * 100) + (quad * 4) + yindex);
   if (!((cent == 4) || (yindex == 4))) {
      year++;
   }
   yearday = wjd - gregorianToJD(year, 1, 1);
   leapadj = ((wjd < gregorianToJD(year, 3, 1)) ? 0
         :
         (isGregorianLeap(year) ? 1 : 2)
         );
   month = (int32_t)uprv_floor((((yearday + leapadj) * 12) + 373) / 367);
   day = (int32_t)(wjd - gregorianToJD(year, month, 1)) + 1;

   gregorianDate[0] = year;
   gregorianDate[1] = month;
   gregorianDate[2] = day;

   return gregorianDate;
}


//-------------------------------------------------------------------------
// Functions for converting from field values to milliseconds....
//-------------------------------------------------------------------------
static double IndianToJD(int32_t year, int32_t month, int32_t date) {
   int32_t leapMonth, gyear, m;
   double start, jd;

   gyear = year + INDIAN_ERA_START;


   if(isGregorianLeap(gyear)) {
      leapMonth = 31;
      start = gregorianToJD(gyear, 3, 21);
   }
   else {
      leapMonth = 30;
      start = gregorianToJD(gyear, 3, 22);
   }

   if (month == 1) {
      jd = start + (date - 1);
   } else {
      jd = start + leapMonth;
      m = month - 2;

      //m = Math.min(m, 5);
      if (m > 5) {
          m = 5;
      }

      jd += m * 31;

      if (month >= 8) {
         m = month - 7;
         jd += m * 30;
      }
      jd += date - 1;
   }

   return jd;
}

/*
 * Return JD of start of given month/year of Indian Calendar
 * @param eyear The year in Indian Calendar measured from Saka Era (78 AD).
 * @param month The month in Indian calendar
 */
int32_t IndianCalendar::handleComputeMonthStart(int32_t eyear, int32_t month, UBool /* useMonth */ ) const {

   //month is 0 based; converting it to 1-based
   int32_t imonth;

    // If the month is out of range, adjust it into range, and adjust the extended eyar accordingly
   if (month < 0 || month > 11) {
      eyear += (int32_t)ClockMath::floorDivide(month, 12, month);
   }

   if(month == 12){
       imonth = 1;
   } else {
       imonth = month + 1;
   }

   double jd = IndianToJD(eyear ,imonth, 1);

   return (int32_t)jd;
}

//-------------------------------------------------------------------------
// Functions for converting from milliseconds to field values
//-------------------------------------------------------------------------

int32_t IndianCalendar::handleGetExtendedYear() {
    int32_t year;

    if (newerField(UCAL_EXTENDED_YEAR, UCAL_YEAR) == UCAL_EXTENDED_YEAR) {
        year = internalGet(UCAL_EXTENDED_YEAR, 1); // Default to year 1
    } else {
        year = internalGet(UCAL_YEAR, 1); // Default to year 1
    }

    return year;
}

/*
 * Override Calendar to compute several fields specific to the Indian
 * calendar system.  These are:
 *
 * <ul><li>ERA
 * <li>YEAR
 * <li>MONTH
 * <li>DAY_OF_MONTH
 * <li>EXTENDED_YEAR</ul>
 *
 * The DAY_OF_WEEK and DOW_LOCAL fields are already set when this
 * method is called. The getGregorianXxx() methods return Gregorian
 * calendar equivalents for the given Julian day.
 */
void IndianCalendar::handleComputeFields(int32_t julianDay, UErrorCode&  /* status */) {
    double jdAtStartOfGregYear;
    int32_t leapMonth, IndianYear, yday, IndianMonth, IndianDayOfMonth, mday;
    int32_t gregorianYear;      // Stores gregorian date corresponding to Julian day;
    int32_t gd[3];

    gregorianYear = jdToGregorian(julianDay, gd)[0];          // Gregorian date for Julian day
    IndianYear = gregorianYear - INDIAN_ERA_START;            // Year in Saka era
    jdAtStartOfGregYear = gregorianToJD(gregorianYear, 1, 1); // JD at start of Gregorian year
    yday = (int32_t)(julianDay - jdAtStartOfGregYear);        // Day number in Gregorian year (starting from 0)

    if (yday < INDIAN_YEAR_START) {
        // Day is at the end of the preceding Saka year
        IndianYear -= 1;
        leapMonth = isGregorianLeap(gregorianYear - 1) ? 31 : 30; // Days in leapMonth this year, previous Gregorian year
        yday += leapMonth + (31 * 5) + (30 * 3) + 10;
    } else {
        leapMonth = isGregorianLeap(gregorianYear) ? 31 : 30; // Days in leapMonth this year
        yday -= INDIAN_YEAR_START;
    }

    if (yday < leapMonth) {
        IndianMonth = 0;
        IndianDayOfMonth = yday + 1;
    } else {
        mday = yday - leapMonth;
        if (mday < (31 * 5)) {
            IndianMonth = (int32_t)uprv_floor(mday / 31) + 1;
            IndianDayOfMonth = (mday % 31) + 1;
        } else {
            mday -= 31 * 5;
            IndianMonth = (int32_t)uprv_floor(mday / 30) + 6;
            IndianDayOfMonth = (mday % 30) + 1;
        }
   }

   internalSet(UCAL_ERA, 0);
   internalSet(UCAL_EXTENDED_YEAR, IndianYear);
   internalSet(UCAL_YEAR, IndianYear);
   internalSet(UCAL_MONTH, IndianMonth);
   internalSet(UCAL_DAY_OF_MONTH, IndianDayOfMonth);
   internalSet(UCAL_DAY_OF_YEAR, yday + 1); // yday is 0-based
}

UBool
IndianCalendar::inDaylightTime(UErrorCode& status) const
{
    // copied from GregorianCalendar
    if (U_FAILURE(status) || !getTimeZone().useDaylightTime()) {
        return FALSE;
    }

    // Force an update of the state of the Calendar.
    ((IndianCalendar*)this)->complete(status); // cast away const

    return (UBool)(U_SUCCESS(status) ? (internalGet(UCAL_DST_OFFSET) != 0) : FALSE);
}


/**
 * The system maintains a static default century start date and Year.  They are
 * initialized the first time they are used.  Once the system default century date
 * and year are set, they do not change.
 */
static UDate           indiancal_gSystemDefaultCenturyStart       = DBL_MIN;
static int32_t         indiancal_gSystemDefaultCenturyStartYear   = -1;
static icu::UInitOnce  indiancal_gSystemDefaultCenturyInit        = U_INITONCE_INITIALIZER;


UBool IndianCalendar::haveDefaultCentury() const
{
    return TRUE;
}

static void U_CALLCONV
indiancal_initializeSystemDefaultCentury()
{
    // initialize systemDefaultCentury and systemDefaultCenturyYear based
    // on the current time.  They'll be set to 80 years before
    // the current time.
    UErrorCode status = U_ZERO_ERROR;

    IndianCalendar calendar ( Locale ( "@calendar=Indian" ), status);
    if ( U_SUCCESS ( status ) ) {
        calendar.setTime ( Calendar::getNow(), status );
        calendar.add ( UCAL_YEAR, -80, status );

        UDate    newStart = calendar.getTime ( status );
        int32_t  newYear  = calendar.get ( UCAL_YEAR, status );

        indiancal_gSystemDefaultCenturyStart = newStart;
        indiancal_gSystemDefaultCenturyStartYear = newYear;
    }
    // We have no recourse upon failure.
}


UDate
IndianCalendar::defaultCenturyStart() const
{
    // lazy-evaluate systemDefaultCenturyStart
    umtx_initOnce(indiancal_gSystemDefaultCenturyInit, &indiancal_initializeSystemDefaultCentury);
    return indiancal_gSystemDefaultCenturyStart;
}

int32_t
IndianCalendar::defaultCenturyStartYear() const
{
    // lazy-evaluate systemDefaultCenturyStartYear
    umtx_initOnce(indiancal_gSystemDefaultCenturyInit, &indiancal_initializeSystemDefaultCentury);
    return    indiancal_gSystemDefaultCenturyStartYear;
}


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(IndianCalendar)

U_NAMESPACE_END

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2003-2015, International Business Machines Corporation
* and others. All Rights Reserved.
******************************************************************************
*
* File ISLAMCAL.H
*
* Modification History:
*
*   Date        Name        Description
*   10/14/2003  srl         ported from java IslamicCalendar
*****************************************************************************
*/



#if !UCONFIG_NO_FORMATTING


#include <float.h>
 // Math
 // CalendarAstronomer




static const UDate HIJRA_MILLIS = -42521587200000.0;    // 7/16/622 AD 00:00

// Debugging
#ifdef U_DEBUG_ISLAMCAL
# include <stdio.h>
# include <stdarg.h>
static void debug_islamcal_loc(const char *f, int32_t l)
{
    fprintf(stderr, "%s:%d: ", f, l);
}

static void debug_islamcal_msg(const char *pat, ...)
{
    va_list ap;
    va_start(ap, pat);
    vfprintf(stderr, pat, ap);
    fflush(stderr);
}
// must use double parens, i.e.:  U_DEBUG_ISLAMCAL_MSG(("four is: %d",4));
#define U_DEBUG_ISLAMCAL_MSG(x) {debug_islamcal_loc(__FILE__,__LINE__);debug_islamcal_msg x;}
#else
#define U_DEBUG_ISLAMCAL_MSG(x)
#endif


// --- The cache --
// cache of months
static icu::CalendarCache *gMonthCache = NULL;
static icu::CalendarAstronomer *gIslamicCalendarAstro = NULL;

U_CDECL_BEGIN
static UBool calendar_islamic_cleanup(void) {
    if (gMonthCache) {
        delete gMonthCache;
        gMonthCache = NULL;
    }
    if (gIslamicCalendarAstro) {
        delete gIslamicCalendarAstro;
        gIslamicCalendarAstro = NULL;
    }
    return TRUE;
}
U_CDECL_END

U_NAMESPACE_BEGIN

// Implementation of the IslamicCalendar class

/**
 * Friday EPOC
 */
static const int32_t CIVIL_EPOC = 1948440; // CE 622 July 16 Friday (Julian calendar) / CE 622 July 19 (Gregorian calendar)

/**
  * Thursday EPOC
  */
static const int32_t ASTRONOMICAL_EPOC = 1948439; // CE 622 July 15 Thursday (Julian calendar)


static const int32_t UMALQURA_YEAR_START = 1300;
static const int32_t UMALQURA_YEAR_END = 1600;

static const int UMALQURA_MONTHLENGTH[] = {
    //* 1300 -1302 */ "1010 1010 1010", "1101 0101 0100", "1110 1100 1001",
                            0x0AAA,           0x0D54,           0x0EC9,
    //* 1303 -1307 */ "0110 1101 0100", "0110 1110 1010", "0011 0110 1100", "1010 1010 1101", "0101 0101 0101",
                            0x06D4,           0x06EA,           0x036C,           0x0AAD,           0x0555,
    //* 1308 -1312 */ "0110 1010 1001", "0111 1001 0010", "1011 1010 1001", "0101 1101 0100", "1010 1101 1010",
                            0x06A9,           0x0792,           0x0BA9,           0x05D4,           0x0ADA,
    //* 1313 -1317 */ "0101 0101 1100", "1101 0010 1101", "0110 1001 0101", "0111 0100 1010", "1011 0101 0100",
                            0x055C,           0x0D2D,           0x0695,           0x074A,           0x0B54,
    //* 1318 -1322 */ "1011 0110 1010", "0101 1010 1101", "0100 1010 1110", "1010 0100 1111", "0101 0001 0111",
                            0x0B6A,           0x05AD,           0x04AE,           0x0A4F,           0x0517,
    //* 1323 -1327 */ "0110 1000 1011", "0110 1010 0101", "1010 1101 0101", "0010 1101 0110", "1001 0101 1011",
                            0x068B,           0x06A5,           0x0AD5,           0x02D6,           0x095B,
    //* 1328 -1332 */ "0100 1001 1101", "1010 0100 1101", "1101 0010 0110", "1101 1001 0101", "0101 1010 1100",
                            0x049D,           0x0A4D,           0x0D26,           0x0D95,           0x05AC,
    //* 1333 -1337 */ "1001 1011 0110", "0010 1011 1010", "1010 0101 1011", "0101 0010 1011", "1010 1001 0101",
                            0x09B6,           0x02BA,           0x0A5B,           0x052B,           0x0A95,
    //* 1338 -1342 */ "0110 1100 1010", "1010 1110 1001", "0010 1111 0100", "1001 0111 0110", "0010 1011 0110",
                            0x06CA,           0x0AE9,           0x02F4,           0x0976,           0x02B6,
    //* 1343 -1347 */ "1001 0101 0110", "1010 1100 1010", "1011 1010 0100", "1011 1101 0010", "0101 1101 1001",
                            0x0956,           0x0ACA,           0x0BA4,           0x0BD2,           0x05D9,
    //* 1348 -1352 */ "0010 1101 1100", "1001 0110 1101", "0101 0100 1101", "1010 1010 0101", "1011 0101 0010",
                            0x02DC,           0x096D,           0x054D,           0x0AA5,           0x0B52,
    //* 1353 -1357 */ "1011 1010 0101", "0101 1011 0100", "1001 1011 0110", "0101 0101 0111", "0010 1001 0111",
                            0x0BA5,           0x05B4,           0x09B6,           0x0557,           0x0297,
    //* 1358 -1362 */ "0101 0100 1011", "0110 1010 0011", "0111 0101 0010", "1011 0110 0101", "0101 0110 1010",
                            0x054B,           0x06A3,           0x0752,           0x0B65,           0x056A,
    //* 1363 -1367 */ "1010 1010 1011", "0101 0010 1011", "1100 1001 0101", "1101 0100 1010", "1101 1010 0101",
                            0x0AAB,           0x052B,           0x0C95,           0x0D4A,           0x0DA5,
    //* 1368 -1372 */ "0101 1100 1010", "1010 1101 0110", "1001 0101 0111", "0100 1010 1011", "1001 0100 1011",
                            0x05CA,           0x0AD6,           0x0957,           0x04AB,           0x094B,
    //* 1373 -1377 */ "1010 1010 0101", "1011 0101 0010", "1011 0110 1010", "0101 0111 0101", "0010 0111 0110",
                            0x0AA5,           0x0B52,           0x0B6A,           0x0575,           0x0276,
    //* 1378 -1382 */ "1000 1011 0111", "0100 0101 1011", "0101 0101 0101", "0101 1010 1001", "0101 1011 0100",
                            0x08B7,           0x045B,           0x0555,           0x05A9,           0x05B4,
    //* 1383 -1387 */ "1001 1101 1010", "0100 1101 1101", "0010 0110 1110", "1001 0011 0110", "1010 1010 1010",
                            0x09DA,           0x04DD,           0x026E,           0x0936,           0x0AAA,
    //* 1388 -1392 */ "1101 0101 0100", "1101 1011 0010", "0101 1101 0101", "0010 1101 1010", "1001 0101 1011",
                            0x0D54,           0x0DB2,           0x05D5,           0x02DA,           0x095B,
    //* 1393 -1397 */ "0100 1010 1011", "1010 0101 0101", "1011 0100 1001", "1011 0110 0100", "1011 0111 0001",
                            0x04AB,           0x0A55,           0x0B49,           0x0B64,           0x0B71,
    //* 1398 -1402 */ "0101 1011 0100", "1010 1011 0101", "1010 0101 0101", "1101 0010 0101", "1110 1001 0010",
                            0x05B4,           0x0AB5,           0x0A55,           0x0D25,           0x0E92,
    //* 1403 -1407 */ "1110 1100 1001", "0110 1101 0100", "1010 1110 1001", "1001 0110 1011", "0100 1010 1011",
                            0x0EC9,           0x06D4,           0x0AE9,           0x096B,           0x04AB,
    //* 1408 -1412 */ "1010 1001 0011", "1101 0100 1001", "1101 1010 0100", "1101 1011 0010", "1010 1011 1001",
                            0x0A93,           0x0D49,         0x0DA4,           0x0DB2,           0x0AB9,
    //* 1413 -1417 */ "0100 1011 1010", "1010 0101 1011", "0101 0010 1011", "1010 1001 0101", "1011 0010 1010",
                            0x04BA,           0x0A5B,           0x052B,           0x0A95,           0x0B2A,
    //* 1418 -1422 */ "1011 0101 0101", "0101 0101 1100", "0100 1011 1101", "0010 0011 1101", "1001 0001 1101",
                            0x0B55,           0x055C,           0x04BD,           0x023D,           0x091D,
    //* 1423 -1427 */ "1010 1001 0101", "1011 0100 1010", "1011 0101 1010", "0101 0110 1101", "0010 1011 0110",
                            0x0A95,           0x0B4A,           0x0B5A,           0x056D,           0x02B6,
    //* 1428 -1432 */ "1001 0011 1011", "0100 1001 1011", "0110 0101 0101", "0110 1010 1001", "0111 0101 0100",
                            0x093B,           0x049B,           0x0655,           0x06A9,           0x0754,
    //* 1433 -1437 */ "1011 0110 1010", "0101 0110 1100", "1010 1010 1101", "0101 0101 0101", "1011 0010 1001",
                            0x0B6A,           0x056C,           0x0AAD,           0x0555,           0x0B29,
    //* 1438 -1442 */ "1011 1001 0010", "1011 1010 1001", "0101 1101 0100", "1010 1101 1010", "0101 0101 1010",
                            0x0B92,           0x0BA9,           0x05D4,           0x0ADA,           0x055A,
    //* 1443 -1447 */ "1010 1010 1011", "0101 1001 0101", "0111 0100 1001", "0111 0110 0100", "1011 1010 1010",
                            0x0AAB,           0x0595,           0x0749,           0x0764,           0x0BAA,
    //* 1448 -1452 */ "0101 1011 0101", "0010 1011 0110", "1010 0101 0110", "1110 0100 1101", "1011 0010 0101",
                            0x05B5,           0x02B6,           0x0A56,           0x0E4D,           0x0B25,
    //* 1453 -1457 */ "1011 0101 0010", "1011 0110 1010", "0101 1010 1101", "0010 1010 1110", "1001 0010 1111",
                            0x0B52,           0x0B6A,           0x05AD,           0x02AE,           0x092F,
    //* 1458 -1462 */ "0100 1001 0111", "0110 0100 1011", "0110 1010 0101", "0110 1010 1100", "1010 1101 0110",
                            0x0497,           0x064B,           0x06A5,           0x06AC,           0x0AD6,
    //* 1463 -1467 */ "0101 0101 1101", "0100 1001 1101", "1010 0100 1101", "1101 0001 0110", "1101 1001 0101",
                            0x055D,           0x049D,           0x0A4D,           0x0D16,           0x0D95,
    //* 1468 -1472 */ "0101 1010 1010", "0101 1011 0101", "0010 1101 1010", "1001 0101 1011", "0100 1010 1101",
                            0x05AA,           0x05B5,           0x02DA,           0x095B,           0x04AD,
    //* 1473 -1477 */ "0101 1001 0101", "0110 1100 1010", "0110 1110 0100", "1010 1110 1010", "0100 1111 0101",
                            0x0595,           0x06CA,           0x06E4,           0x0AEA,           0x04F5,
    //* 1478 -1482 */ "0010 1011 0110", "1001 0101 0110", "1010 1010 1010", "1011 0101 0100", "1011 1101 0010",
                            0x02B6,           0x0956,           0x0AAA,           0x0B54,           0x0BD2,
    //* 1483 -1487 */ "0101 1101 1001", "0010 1110 1010", "1001 0110 1101", "0100 1010 1101", "1010 1001 0101",
                            0x05D9,           0x02EA,           0x096D,           0x04AD,           0x0A95,
    //* 1488 -1492 */ "1011 0100 1010", "1011 1010 0101", "0101 1011 0010", "1001 1011 0101", "0100 1101 0110",
                            0x0B4A,           0x0BA5,           0x05B2,           0x09B5,           0x04D6,
    //* 1493 -1497 */ "1010 1001 0111", "0101 0100 0111", "0110 1001 0011", "0111 0100 1001", "1011 0101 0101",
                            0x0A97,           0x0547,           0x0693,           0x0749,           0x0B55,
    //* 1498 -1508 */ "0101 0110 1010", "1010 0110 1011", "0101 0010 1011", "1010 1000 1011", "1101 0100 0110", "1101 1010 0011", "0101 1100 1010", "1010 1101 0110", "0100 1101 1011", "0010 0110 1011", "1001 0100 1011",
                            0x056A,           0x0A6B,           0x052B,           0x0A8B,           0x0D46,           0x0DA3,           0x05CA,           0x0AD6,           0x04DB,           0x026B,           0x094B,
    //* 1509 -1519 */ "1010 1010 0101", "1011 0101 0010", "1011 0110 1001", "0101 0111 0101", "0001 0111 0110", "1000 1011 0111", "0010 0101 1011", "0101 0010 1011", "0101 0110 0101", "0101 1011 0100", "1001 1101 1010",
                            0x0AA5,           0x0B52,           0x0B69,           0x0575,           0x0176,           0x08B7,           0x025B,           0x052B,           0x0565,           0x05B4,           0x09DA,
    //* 1520 -1530 */ "0100 1110 1101", "0001 0110 1101", "1000 1011 0110", "1010 1010 0110", "1101 0101 0010", "1101 1010 1001", "0101 1101 0100", "1010 1101 1010", "1001 0101 1011", "0100 1010 1011", "0110 0101 0011",
                            0x04ED,           0x016D,           0x08B6,           0x0AA6,           0x0D52,           0x0DA9,           0x05D4,           0x0ADA,           0x095B,           0x04AB,           0x0653,
    //* 1531 -1541 */ "0111 0010 1001", "0111 0110 0010", "1011 1010 1001", "0101 1011 0010", "1010 1011 0101", "0101 0101 0101", "1011 0010 0101", "1101 1001 0010", "1110 1100 1001", "0110 1101 0010", "1010 1110 1001",
                            0x0729,           0x0762,           0x0BA9,           0x05B2,           0x0AB5,           0x0555,           0x0B25,           0x0D92,           0x0EC9,           0x06D2,           0x0AE9,
    //* 1542 -1552 */ "0101 0110 1011", "0100 1010 1011", "1010 0101 0101", "1101 0010 1001", "1101 0101 0100", "1101 1010 1010", "1001 1011 0101", "0100 1011 1010", "1010 0011 1011", "0100 1001 1011", "1010 0100 1101",
                            0x056B,           0x04AB,           0x0A55,           0x0D29,           0x0D54,           0x0DAA,           0x09B5,           0x04BA,           0x0A3B,           0x049B,           0x0A4D,
    //* 1553 -1563 */ "1010 1010 1010", "1010 1101 0101", "0010 1101 1010", "1001 0101 1101", "0100 0101 1110", "1010 0010 1110", "1100 1001 1010", "1101 0101 0101", "0110 1011 0010", "0110 1011 1001", "0100 1011 1010",
                            0x0AAA,           0x0AD5,           0x02DA,           0x095D,           0x045E,           0x0A2E,           0x0C9A,           0x0D55,           0x06B2,           0x06B9,           0x04BA,
    //* 1564 -1574 */ "1010 0101 1101", "0101 0010 1101", "1010 1001 0101", "1011 0101 0010", "1011 1010 1000", "1011 1011 0100", "0101 1011 1001", "0010 1101 1010", "1001 0101 1010", "1011 0100 1010", "1101 1010 0100",
                            0x0A5D,           0x052D,           0x0A95,           0x0B52,           0x0BA8,           0x0BB4,           0x05B9,           0x02DA,           0x095A,           0x0B4A,           0x0DA4,
    //* 1575 -1585 */ "1110 1101 0001", "0110 1110 1000", "1011 0110 1010", "0101 0110 1101", "0101 0011 0101", "0110 1001 0101", "1101 0100 1010", "1101 1010 1000", "1101 1101 0100", "0110 1101 1010", "0101 0101 1011",
                            0x0ED1,           0x06E8,           0x0B6A,           0x056D,           0x0535,           0x0695,           0x0D4A,           0x0DA8,           0x0DD4,           0x06DA,           0x055B,
    //* 1586 -1596 */ "0010 1001 1101", "0110 0010 1011", "1011 0001 0101", "1011 0100 1010", "1011 1001 0101", "0101 1010 1010", "1010 1010 1110", "1001 0010 1110", "1100 1000 1111", "0101 0010 0111", "0110 1001 0101",
                            0x029D,           0x062B,           0x0B15,           0x0B4A,           0x0B95,           0x05AA,           0x0AAE,           0x092E,           0x0C8F,           0x0527,           0x0695,
    //* 1597 -1600 */ "0110 1010 1010", "1010 1101 0110", "0101 0101 1101", "0010 1001 1101", };
                            0x06AA,           0x0AD6,           0x055D,           0x029D
};

int32_t getUmalqura_MonthLength(int32_t y, int32_t m) {
    int32_t mask = (int32_t) (0x01 << (11 - m));    // set mask for bit corresponding to month
    if((UMALQURA_MONTHLENGTH[y] & mask) == 0 )
        return 29;
    else
        return 30;

}

//-------------------------------------------------------------------------
// Constructors...
//-------------------------------------------------------------------------

const char *IslamicCalendar::getType() const {
    const char *sType = NULL;

    switch (cType) {
    case CIVIL:
        sType = "islamic-civil";
        break;
    case ASTRONOMICAL:
        sType = "islamic";
        break;
    case TBLA:
        sType = "islamic-tbla";
        break;
    case UMALQURA:
        sType = "islamic-umalqura";
        break;
    default:
        UPRV_UNREACHABLE; // out of range
    }
    return sType;
}

IslamicCalendar* IslamicCalendar::clone() const {
    return new IslamicCalendar(*this);
}

IslamicCalendar::IslamicCalendar(const Locale& aLocale, UErrorCode& success, ECalculationType type)
:   Calendar(TimeZone::createDefault(), aLocale, success),
cType(type)
{
    setTimeInMillis(getNow(), success); // Call this again now that the vtable is set up properly.
}

IslamicCalendar::IslamicCalendar(const IslamicCalendar& other) : Calendar(other), cType(other.cType) {
}

IslamicCalendar::~IslamicCalendar()
{
}

void IslamicCalendar::setCalculationType(ECalculationType type, UErrorCode &status)
{
    if (cType != type) {
        // The fields of the calendar will become invalid, because the calendar
        // rules are different
        UDate m = getTimeInMillis(status);
        cType = type;
        clear();
        setTimeInMillis(m, status);
    }
}

/**
* Returns <code>true</code> if this object is using the fixed-cycle civil
* calendar, or <code>false</code> if using the religious, astronomical
* calendar.
* @draft ICU 2.4
*/
UBool IslamicCalendar::isCivil() {
    return (cType == CIVIL);
}

//-------------------------------------------------------------------------
// Minimum / Maximum access functions
//-------------------------------------------------------------------------

// Note: Current IslamicCalendar implementation does not work
// well with negative years.

// TODO: In some cases the current ICU Islamic calendar implementation shows
// a month as having 31 days. Since date parsing now uses range checks based
// on the table below, we need to change the range for last day of month to
// include 31 as a workaround until the implementation is fixed.
static const int32_t ISLAMCAL_LIMITS[UCAL_FIELD_COUNT][4] = {
    // Minimum  Greatest    Least  Maximum
    //           Minimum  Maximum
    {        0,        0,        0,        0}, // ERA
    {        1,        1,  5000000,  5000000}, // YEAR
    {        0,        0,       11,       11}, // MONTH
    {        1,        1,       50,       51}, // WEEK_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // WEEK_OF_MONTH
    {        1,        1,       29,       31}, // DAY_OF_MONTH - 31 to workaround for cal implementation bug, should be 30
    {        1,        1,      354,      355}, // DAY_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DAY_OF_WEEK
    {       -1,       -1,        5,        5}, // DAY_OF_WEEK_IN_MONTH
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // AM_PM
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR_OF_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MINUTE
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // SECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // ZONE_OFFSET
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DST_OFFSET
    {        1,        1,  5000000,  5000000}, // YEAR_WOY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DOW_LOCAL
    {        1,        1,  5000000,  5000000}, // EXTENDED_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // JULIAN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECONDS_IN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // IS_LEAP_MONTH
};

/**
* @draft ICU 2.4
*/
int32_t IslamicCalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const {
    return ISLAMCAL_LIMITS[field][limitType];
}

//-------------------------------------------------------------------------
// Assorted calculation utilities
//

// we could compress this down more if we need to
static const int8_t umAlQuraYrStartEstimateFix[] = {
     0,  0, -1,  0, -1,  0,  0,  0,  0,  0, // 1300..
    -1,  0,  0,  0,  0,  0,  0,  0, -1,  0, // 1310..
     1,  0,  1,  1,  0,  0,  0,  0,  1,  0, // 1320..
     0,  0,  0,  0,  0,  0,  1,  0,  0,  0, // 1330..
     0,  0,  1,  0,  0, -1, -1,  0,  0,  0, // 1340..
     1,  0,  0, -1,  0,  0,  0,  1,  1,  0, // 1350..
     0,  0,  0,  0,  0,  0,  0, -1,  0,  0, // 1360..
     0,  1,  1,  0,  0, -1,  0,  1,  0,  1, // 1370..
     1,  0,  0, -1,  0,  1,  0,  0,  0, -1, // 1380..
     0,  1,  0,  1,  0,  0,  0, -1,  0,  0, // 1390..
     0,  0, -1, -1,  0, -1,  0,  1,  0,  0, // 1400..
     0, -1,  0,  0,  0,  1,  0,  0,  0,  0, // 1410..
     0,  1,  0,  0, -1, -1,  0,  0,  0,  1, // 1420..
     0,  0, -1, -1,  0, -1,  0,  0, -1, -1, // 1430..
     0, -1,  0, -1,  0,  0, -1, -1,  0,  0, // 1440..
     0,  0,  0,  0, -1,  0,  1,  0,  1,  1, // 1450..
     0,  0, -1,  0,  1,  0,  0,  0,  0,  0, // 1460..
     1,  0,  1,  0,  0,  0, -1,  0,  1,  0, // 1470..
     0, -1, -1,  0,  0,  0,  1,  0,  0,  0, // 1480..
     0,  0,  0,  0,  1,  0,  0,  0,  0,  0, // 1490..
     1,  0,  0, -1,  0,  0,  0,  1,  1,  0, // 1500..
     0, -1,  0,  1,  0,  1,  1,  0,  0,  0, // 1510..
     0,  1,  0,  0,  0, -1,  0,  0,  0,  1, // 1520..
     0,  0,  0, -1,  0,  0,  0,  0,  0, -1, // 1530..
     0, -1,  0,  1,  0,  0,  0, -1,  0,  1, // 1540..
     0,  1,  0,  0,  0,  0,  0,  1,  0,  0, // 1550..
    -1,  0,  0,  0,  0,  1,  0,  0,  0, -1, // 1560..
     0,  0,  0,  0, -1, -1,  0, -1,  0,  1, // 1570..
     0,  0, -1, -1,  0,  0,  1,  1,  0,  0, // 1580..
    -1,  0,  0,  0,  0,  1,  0,  0,  0,  0, // 1590..
     1 // 1600
};

/**
* Determine whether a year is a leap year in the Islamic civil calendar
*/
UBool IslamicCalendar::civilLeapYear(int32_t year)
{
    return (14 + 11 * year) % 30 < 11;
}

/**
* Return the day # on which the given year starts.  Days are counted
* from the Hijri epoch, origin 0.
*/
int32_t IslamicCalendar::yearStart(int32_t year) const{
    if (cType == CIVIL || cType == TBLA ||
        (cType == UMALQURA && (year < UMALQURA_YEAR_START || year > UMALQURA_YEAR_END)))
    {
        return (year-1)*354 + ClockMath::floorDivide((3+11*year),30);
    } else if(cType==ASTRONOMICAL){
        return trueMonthStart(12*(year-1));
    } else {
        year -= UMALQURA_YEAR_START;
        // rounded least-squares fit of the dates previously calculated from UMALQURA_MONTHLENGTH iteration
        int32_t yrStartLinearEstimate = (int32_t)((354.36720 * (double)year) + 460322.05 + 0.5);
        // need a slight correction to some
        return yrStartLinearEstimate + umAlQuraYrStartEstimateFix[year];
    }
}

/**
* Return the day # on which the given month starts.  Days are counted
* from the Hijri epoch, origin 0.
*
* @param year  The hijri year
* @param month The hijri month, 0-based (assumed to be in range 0..11)
*/
int32_t IslamicCalendar::monthStart(int32_t year, int32_t month) const {
    if (cType == CIVIL || cType == TBLA) {
        // This does not handle months out of the range 0..11
        return (int32_t)uprv_ceil(29.5*month)
            + (year-1)*354 + (int32_t)ClockMath::floorDivide((3+11*year),30);
    } else if(cType==ASTRONOMICAL){
        return trueMonthStart(12*(year-1) + month);
    } else {
        int32_t ms = yearStart(year);
        for(int i=0; i< month; i++){
            ms+= handleGetMonthLength(year, i);
        }
        return ms;
    }
}

/**
* Find the day number on which a particular month of the true/lunar
* Islamic calendar starts.
*
* @param month The month in question, origin 0 from the Hijri epoch
*
* @return The day number on which the given month starts.
*/
int32_t IslamicCalendar::trueMonthStart(int32_t month) const
{
    UErrorCode status = U_ZERO_ERROR;
    int32_t start = CalendarCache::get(&gMonthCache, month, status);

    if (start==0) {
        // Make a guess at when the month started, using the average length
        UDate origin = HIJRA_MILLIS
            + uprv_floor(month * CalendarAstronomer::SYNODIC_MONTH) * kOneDay;

        // moonAge will fail due to memory allocation error
        double age = moonAge(origin, status);
        if (U_FAILURE(status)) {
            goto trueMonthStartEnd;
        }

        if (age >= 0) {
            // The month has already started
            do {
                origin -= kOneDay;
                age = moonAge(origin, status);
                if (U_FAILURE(status)) {
                    goto trueMonthStartEnd;
                }
            } while (age >= 0);
        }
        else {
            // Preceding month has not ended yet.
            do {
                origin += kOneDay;
                age = moonAge(origin, status);
                if (U_FAILURE(status)) {
                    goto trueMonthStartEnd;
                }
            } while (age < 0);
        }
        start = (int32_t)ClockMath::floorDivide((origin - HIJRA_MILLIS), (double)kOneDay) + 1;
        CalendarCache::put(&gMonthCache, month, start, status);
    }
trueMonthStartEnd :
    if(U_FAILURE(status)) {
        start = 0;
    }
    return start;
}

/**
* Return the "age" of the moon at the given time; this is the difference
* in ecliptic latitude between the moon and the sun.  This method simply
* calls CalendarAstronomer.moonAge, converts to degrees,
* and adjusts the result to be in the range [-180, 180].
*
* @param time  The time at which the moon's age is desired,
*              in millis since 1/1/1970.
*/
double IslamicCalendar::moonAge(UDate time, UErrorCode &status)
{
    double age = 0;

    static UMutex astroLock;      // pod bay door lock
    umtx_lock(&astroLock);
    if(gIslamicCalendarAstro == NULL) {
        gIslamicCalendarAstro = new CalendarAstronomer();
        if (gIslamicCalendarAstro == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return age;
        }
        ucln_i18n_registerCleanup(UCLN_I18N_ISLAMIC_CALENDAR, calendar_islamic_cleanup);
    }
    gIslamicCalendarAstro->setTime(time);
    age = gIslamicCalendarAstro->getMoonAge();
    umtx_unlock(&astroLock);

    // Convert to degrees and normalize...
    age = age * 180 / CalendarAstronomer::PI;
    if (age > 180) {
        age = age - 360;
    }

    return age;
}

//----------------------------------------------------------------------
// Calendar framework
//----------------------------------------------------------------------

/**
* Return the length (in days) of the given month.
*
* @param year  The hijri year
* @param year  The hijri month, 0-based
* @draft ICU 2.4
*/
int32_t IslamicCalendar::handleGetMonthLength(int32_t extendedYear, int32_t month) const {

    int32_t length = 0;

    if (cType == CIVIL || cType == TBLA ||
        (cType == UMALQURA && (extendedYear<UMALQURA_YEAR_START || extendedYear>UMALQURA_YEAR_END)) ) {
        length = 29 + (month+1) % 2;
        if (month == DHU_AL_HIJJAH && civilLeapYear(extendedYear)) {
            length++;
        }
    } else if(cType == ASTRONOMICAL){
        month = 12*(extendedYear-1) + month;
        length =  trueMonthStart(month+1) - trueMonthStart(month) ;
    } else {
        length = getUmalqura_MonthLength(extendedYear - UMALQURA_YEAR_START, month);
    }
    return length;
}

/**
* Return the number of days in the given Islamic year
* @draft ICU 2.4
*/
int32_t IslamicCalendar::handleGetYearLength(int32_t extendedYear) const {
    if (cType == CIVIL || cType == TBLA ||
        (cType == UMALQURA && (extendedYear<UMALQURA_YEAR_START || extendedYear>UMALQURA_YEAR_END)) ) {
        return 354 + (civilLeapYear(extendedYear) ? 1 : 0);
    } else if(cType == ASTRONOMICAL){
        int32_t month = 12*(extendedYear-1);
        return (trueMonthStart(month + 12) - trueMonthStart(month));
    } else {
        int len = 0;
        for(int i=0; i<12; i++) {
            len += handleGetMonthLength(extendedYear, i);
        }
        return len;
    }
}

//-------------------------------------------------------------------------
// Functions for converting from field values to milliseconds....
//-------------------------------------------------------------------------

// Return JD of start of given month/year
// Calendar says:
// Get the Julian day of the day BEFORE the start of this year.
// If useMonth is true, get the day before the start of the month.
// Hence the -1
/**
* @draft ICU 2.4
*/
int32_t IslamicCalendar::handleComputeMonthStart(int32_t eyear, int32_t month, UBool /* useMonth */) const {
    // This may be called by Calendar::handleComputeJulianDay with months out of the range
    // 0..11. Need to handle that here since monthStart requires months in the range 0.11.
    if (month > 11) {
        eyear += (month / 12);
        month %= 12;
    } else if (month < 0) {
        month++;
        eyear += (month / 12) - 1;
        month = (month % 12) + 11;
    }
    return monthStart(eyear, month) + ((cType == TBLA)? ASTRONOMICAL_EPOC: CIVIL_EPOC) - 1;
}

//-------------------------------------------------------------------------
// Functions for converting from milliseconds to field values
//-------------------------------------------------------------------------

/**
* @draft ICU 2.4
*/
int32_t IslamicCalendar::handleGetExtendedYear() {
    int32_t year;
    if (newerField(UCAL_EXTENDED_YEAR, UCAL_YEAR) == UCAL_EXTENDED_YEAR) {
        year = internalGet(UCAL_EXTENDED_YEAR, 1); // Default to year 1
    } else {
        year = internalGet(UCAL_YEAR, 1); // Default to year 1
    }
    return year;
}

/**
* Override Calendar to compute several fields specific to the Islamic
* calendar system.  These are:
*
* <ul><li>ERA
* <li>YEAR
* <li>MONTH
* <li>DAY_OF_MONTH
* <li>DAY_OF_YEAR
* <li>EXTENDED_YEAR</ul>
*
* The DAY_OF_WEEK and DOW_LOCAL fields are already set when this
* method is called. The getGregorianXxx() methods return Gregorian
* calendar equivalents for the given Julian day.
* @draft ICU 2.4
*/
void IslamicCalendar::handleComputeFields(int32_t julianDay, UErrorCode &status) {
    int32_t year, month, dayOfMonth, dayOfYear;
    int32_t startDate;
    int32_t days = julianDay - CIVIL_EPOC;

    if (cType == CIVIL || cType == TBLA) {
        if(cType == TBLA) {
            days = julianDay - ASTRONOMICAL_EPOC;
        }
        // Use the civil calendar approximation, which is just arithmetic
        year  = (int32_t)ClockMath::floorDivide(30 * (int64_t)days + 10646, (int64_t)10631);
        month = (int32_t)uprv_ceil((days - 29 - yearStart(year)) / 29.5 );
        month = month<11?month:11;
        startDate = monthStart(year, month);
    } else if(cType == ASTRONOMICAL){
        // Guess at the number of elapsed full months since the epoch
        int32_t months = (int32_t)uprv_floor((double)days / CalendarAstronomer::SYNODIC_MONTH);

        startDate = (int32_t)uprv_floor(months * CalendarAstronomer::SYNODIC_MONTH);

        double age = moonAge(internalGetTime(), status);
        if (U_FAILURE(status)) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        if ( days - startDate >= 25 && age > 0) {
            // If we're near the end of the month, assume next month and search backwards
            months++;
        }

        // Find out the last time that the new moon was actually visible at this longitude
        // This returns midnight the night that the moon was visible at sunset.
        while ((startDate = trueMonthStart(months)) > days) {
            // If it was after the date in question, back up a month and try again
            months--;
        }

        year = months / 12 + 1;
        month = months % 12;
    } else if(cType == UMALQURA) {
        int32_t umalquraStartdays = yearStart(UMALQURA_YEAR_START) ;
        if( days < umalquraStartdays){
                //Use Civil calculation
                year  = (int)ClockMath::floorDivide( (double)(30 * days + 10646) , 10631.0 );
                month = (int32_t)uprv_ceil((days - 29 - yearStart(year)) / 29.5 );
                month = month<11?month:11;
                startDate = monthStart(year, month);
            }else{
                int y =UMALQURA_YEAR_START-1, m =0;
                long d = 1;
                while(d > 0){
                    y++;
                    d = days - yearStart(y) +1;
                    if(d == handleGetYearLength(y)){
                        m=11;
                        break;
                    }else if(d < handleGetYearLength(y) ){
                        int monthLen = handleGetMonthLength(y, m);
                        m=0;
                        while(d > monthLen){
                            d -= monthLen;
                            m++;
                            monthLen = handleGetMonthLength(y, m);
                        }
                        break;
                    }
                }
                year = y;
                month = m;
            }
    } else { // invalid 'civil'
      UPRV_UNREACHABLE; // should not get here, out of range
    }

    dayOfMonth = (days - monthStart(year, month)) + 1;

    // Now figure out the day of the year.
    dayOfYear = (days - monthStart(year, 0)) + 1;


    internalSet(UCAL_ERA, 0);
    internalSet(UCAL_YEAR, year);
    internalSet(UCAL_EXTENDED_YEAR, year);
    internalSet(UCAL_MONTH, month);
    internalSet(UCAL_DAY_OF_MONTH, dayOfMonth);
    internalSet(UCAL_DAY_OF_YEAR, dayOfYear);
}

UBool
IslamicCalendar::inDaylightTime(UErrorCode& status) const
{
    // copied from GregorianCalendar
    if (U_FAILURE(status) || !getTimeZone().useDaylightTime())
        return FALSE;

    // Force an update of the state of the Calendar.
    ((IslamicCalendar*)this)->complete(status); // cast away const

    return (UBool)(U_SUCCESS(status) ? (internalGet(UCAL_DST_OFFSET) != 0) : FALSE);
}

/**
 * The system maintains a static default century start date and Year.  They are
 * initialized the first time they are used.  Once the system default century date
 * and year are set, they do not change.
 */
static UDate           islamcal_gSystemDefaultCenturyStart       = DBL_MIN;
static int32_t         islamcal_gSystemDefaultCenturyStartYear   = -1;
static icu::UInitOnce  islamcal_gSystemDefaultCenturyInit        = U_INITONCE_INITIALIZER;


UBool IslamicCalendar::haveDefaultCentury() const
{
    return TRUE;
}

UDate IslamicCalendar::defaultCenturyStart() const
{
    // lazy-evaluate systemDefaultCenturyStart
    umtx_initOnce(islamcal_gSystemDefaultCenturyInit, &initializeSystemDefaultCentury);
    return islamcal_gSystemDefaultCenturyStart;
}

int32_t IslamicCalendar::defaultCenturyStartYear() const
{
    // lazy-evaluate systemDefaultCenturyStartYear
    umtx_initOnce(islamcal_gSystemDefaultCenturyInit, &initializeSystemDefaultCentury);
    return islamcal_gSystemDefaultCenturyStartYear;
}


U_CFUNC void U_CALLCONV
IslamicCalendar::initializeSystemDefaultCentury()
{
    // initialize systemDefaultCentury and systemDefaultCenturyYear based
    // on the current time.  They'll be set to 80 years before
    // the current time.
    UErrorCode status = U_ZERO_ERROR;
    IslamicCalendar calendar(Locale("@calendar=islamic-civil"),status);
    if (U_SUCCESS(status)) {
        calendar.setTime(Calendar::getNow(), status);
        calendar.add(UCAL_YEAR, -80, status);

        islamcal_gSystemDefaultCenturyStart = calendar.getTime(status);
        islamcal_gSystemDefaultCenturyStartYear = calendar.get(UCAL_YEAR, status);
    }
    // We have no recourse upon failure unless we want to propagate the failure
    // out.
}



UOBJECT_DEFINE_RTTI_IMPLEMENTATION(IslamicCalendar)

U_NAMESPACE_END

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2013-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  listformatter.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2012aug27
*   created by: Umesh P. Nair
*/


  // FieldPositionIterator
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2012-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  listformatter.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 20120426
*   created by: Umesh P. Nair
*/




#if U_SHOW_CPLUSPLUS_API





U_NAMESPACE_BEGIN

class FieldPositionIterator;
class FieldPositionHandler;
class FormattedListData;
class ListFormatter;

/** @internal */
class Hashtable;

/** @internal */
struct ListFormatInternal;

/* The following can't be #ifndef U_HIDE_INTERNAL_API, needed for other .h file declarations */
/**
 * @internal
 * \cond
 */
struct ListFormatData : public UMemory {
    UnicodeString twoPattern;
    UnicodeString startPattern;
    UnicodeString middlePattern;
    UnicodeString endPattern;

  ListFormatData(const UnicodeString& two, const UnicodeString& start, const UnicodeString& middle, const UnicodeString& end) :
      twoPattern(two), startPattern(start), middlePattern(middle), endPattern(end) {}
};
/** \endcond */


/**
 * \file
 * \brief C++ API: API for formatting a list.
 */


#if !UCONFIG_NO_FORMATTING
#ifndef U_HIDE_DRAFT_API
/**
 * An immutable class containing the result of a list formatting operation.
 *
 * Instances of this class are immutable and thread-safe.
 *
 * When calling nextPosition():
 * The fields are returned from start to end. The special field category
 * UFIELD_CATEGORY_LIST_SPAN is used to indicate which argument
 * was inserted at the given position. The span category will
 * always occur before the corresponding instance of UFIELD_CATEGORY_LIST
 * in the nextPosition() iterator.
 *
 * Not intended for public subclassing.
 *
 * @draft ICU 64
 */
class U_I18N_API FormattedList : public UMemory, public FormattedValue {
  public:
    /**
     * Default constructor; makes an empty FormattedList.
     * @draft ICU 64
     */
    FormattedList() : fData(nullptr), fErrorCode(U_INVALID_STATE_ERROR) {}

    /**
     * Move constructor: Leaves the source FormattedList in an undefined state.
     * @draft ICU 64
     */
    FormattedList(FormattedList&& src) U_NOEXCEPT;

    /**
     * Destruct an instance of FormattedList.
     * @draft ICU 64
     */
    virtual ~FormattedList() U_OVERRIDE;

    /** Copying not supported; use move constructor instead. */
    FormattedList(const FormattedList&) = delete;

    /** Copying not supported; use move assignment instead. */
    FormattedList& operator=(const FormattedList&) = delete;

    /**
     * Move assignment: Leaves the source FormattedList in an undefined state.
     * @draft ICU 64
     */
    FormattedList& operator=(FormattedList&& src) U_NOEXCEPT;

    /** @copydoc FormattedValue::toString() */
    UnicodeString toString(UErrorCode& status) const U_OVERRIDE;

    /** @copydoc FormattedValue::toTempString() */
    UnicodeString toTempString(UErrorCode& status) const U_OVERRIDE;

    /** @copydoc FormattedValue::appendTo() */
    Appendable &appendTo(Appendable& appendable, UErrorCode& status) const U_OVERRIDE;

    /** @copydoc FormattedValue::nextPosition() */
    UBool nextPosition(ConstrainedFieldPosition& cfpos, UErrorCode& status) const U_OVERRIDE;

  private:
    FormattedListData *fData;
    UErrorCode fErrorCode;
    explicit FormattedList(FormattedListData *results)
        : fData(results), fErrorCode(U_ZERO_ERROR) {}
    explicit FormattedList(UErrorCode errorCode)
        : fData(nullptr), fErrorCode(errorCode) {}
    friend class ListFormatter;
};
#endif /* U_HIDE_DRAFT_API */
#endif // !UCONFIG_NO_FORMATTING


/**
 * An immutable class for formatting a list, using data from CLDR (or supplied
 * separately).
 *
 * Example: Input data ["Alice", "Bob", "Charlie", "Delta"] will be formatted
 * as "Alice, Bob, Charlie and Delta" in English.
 *
 * The ListFormatter class is not intended for public subclassing.
 * @stable ICU 50
 */
class U_I18N_API ListFormatter : public UObject{

  public:

    /**
     * Copy constructor.
     * @stable ICU 52
     */
    ListFormatter(const ListFormatter&);

    /**
     * Assignment operator.
     * @stable ICU 52
     */
    ListFormatter& operator=(const ListFormatter& other);

    /**
     * Creates a ListFormatter appropriate for the default locale.
     *
     * @param errorCode ICU error code, set if no data available for default locale.
     * @return Pointer to a ListFormatter object for the default locale,
     *     created from internal data derived from CLDR data.
     * @stable ICU 50
     */
    static ListFormatter* createInstance(UErrorCode& errorCode);

    /**
     * Creates a ListFormatter appropriate for a locale.
     *
     * @param locale The locale.
     * @param errorCode ICU error code, set if no data available for the given locale.
     * @return A ListFormatter object created from internal data derived from
     *     CLDR data.
     * @stable ICU 50
     */
    static ListFormatter* createInstance(const Locale& locale, UErrorCode& errorCode);

#ifndef U_HIDE_INTERNAL_API
    /**
     * Creates a ListFormatter appropriate for a locale and style.
     *
     * @param locale The locale.
     * @param style the style, either "standard", "or", "unit", "unit-narrow", or "unit-short"
     * @param errorCode ICU error code, set if no data available for the given locale.
     * @return A ListFormatter object created from internal data derived from
     *     CLDR data.
     * @internal
     */
    static ListFormatter* createInstance(const Locale& locale, const char* style, UErrorCode& errorCode);
#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Destructor.
     *
     * @stable ICU 50
     */
    virtual ~ListFormatter();


    /**
     * Formats a list of strings.
     *
     * @param items An array of strings to be combined and formatted.
     * @param n_items Length of the array items.
     * @param appendTo The string to which the result should be appended to.
     * @param errorCode ICU error code, set if there is an error.
     * @return Formatted string combining the elements of items, appended to appendTo.
     * @stable ICU 50
     */
    UnicodeString& format(const UnicodeString items[], int32_t n_items,
        UnicodeString& appendTo, UErrorCode& errorCode) const;

#ifndef U_HIDE_DRAFT_API
    /**
     * Format a list of strings.
     *
     * @param items     An array of strings to be combined and formatted.
     * @param n_items   Length of the array items.
     * @param appendTo  The string to which the formatted result will be
     *                  appended.
     * @param posIter   On return, can be used to iterate over positions of
     *                  fields generated by this format call. Field values are
     *                  defined in UListFormatterField. Can be NULL.
     * @param errorCode ICU error code returned here.
     * @return          Formatted string combining the elements of items,
     *                  appended to appendTo.
     * @draft ICU 63
     */
    UnicodeString& format(const UnicodeString items[], int32_t n_items,
        UnicodeString & appendTo, FieldPositionIterator* posIter,
        UErrorCode& errorCode) const;
#endif // U_HIDE_DRAFT_API

#if !UCONFIG_NO_FORMATTING
#ifndef U_HIDE_DRAFT_API
    /**
     * Formats a list of strings to a FormattedList, which exposes field
     * position information. The FormattedList contains more information than
     * a FieldPositionIterator.
     *
     * @param items     An array of strings to be combined and formatted.
     * @param n_items   Length of the array items.
     * @param errorCode ICU error code returned here.
     * @return          A FormattedList containing field information.
     * @draft ICU 64
     */
    FormattedList formatStringsToValue(
        const UnicodeString items[],
        int32_t n_items,
        UErrorCode& errorCode) const;
#endif  /* U_HIDE_DRAFT_API */
#endif // !UCONFIG_NO_FORMATTING

#ifndef U_HIDE_INTERNAL_API
    /**
      @internal for MeasureFormat
    */
    UnicodeString& format(
            const UnicodeString items[],
            int32_t n_items,
            UnicodeString& appendTo,
            int32_t index,
            int32_t &offset,
            UErrorCode& errorCode) const;
    /**
     * @internal constructor made public for testing.
     */
    ListFormatter(const ListFormatData &data, UErrorCode &errorCode);
    /**
     * @internal constructor made public for testing.
     */
    ListFormatter(const ListFormatInternal* listFormatterInternal);
#endif  /* U_HIDE_INTERNAL_API */

  private:
    static void initializeHash(UErrorCode& errorCode);
    static const ListFormatInternal* getListFormatInternal(const Locale& locale, const char *style, UErrorCode& errorCode);
    struct ListPatternsSink;
    static ListFormatInternal* loadListFormatInternal(const Locale& locale, const char* style, UErrorCode& errorCode);

    UnicodeString& format_(
        const UnicodeString items[], int32_t n_items, UnicodeString& appendTo,
        int32_t index, int32_t &offset, FieldPositionHandler* handler, UErrorCode& errorCode) const;

    ListFormatter();

    ListFormatInternal* owned;
    const ListFormatInternal* data;
};

U_NAMESPACE_END

#endif /* U_SHOW_CPLUSPLUS_API */



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2015-2016, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/




#if !UCONFIG_NO_FORMATTING




/**
 * \file
 * \brief C API: Format a list in a locale-appropriate way.
 *
 * A UListFormatter is used to format a list of items in a locale-appropriate way, 
 * using data from CLDR.
 * Example: Input data ["Alice", "Bob", "Charlie", "Delta"] will be formatted
 * as "Alice, Bob, Charlie, and Delta" in English.
 */

/**
 * Opaque UListFormatter object for use in C
 * @stable ICU 55
 */
struct UListFormatter;
typedef struct UListFormatter UListFormatter;  /**< C typedef for struct UListFormatter. @stable ICU 55 */

#ifndef U_HIDE_DRAFT_API
struct UFormattedList;
/**
 * Opaque struct to contain the results of a UListFormatter operation.
 * @draft ICU 64
 */
typedef struct UFormattedList UFormattedList;
#endif  /* U_HIDE_DRAFT_API */

#ifndef U_HIDE_DRAFT_API
/**
 * FieldPosition and UFieldPosition selectors for format fields
 * defined by ListFormatter.
 * @draft ICU 63
 */
typedef enum UListFormatterField {
    /**
     * The literal text in the result which came from the resources.
     * @draft ICU 63
     */
    ULISTFMT_LITERAL_FIELD,
    /**
     * The element text in the result which came from the input strings.
     * @draft ICU 63
     */
    ULISTFMT_ELEMENT_FIELD
} UListFormatterField;
#endif /* U_HIDE_DRAFT_API */

/**
 * Open a new UListFormatter object using the rules for a given locale.
 * @param locale
 *            The locale whose rules should be used; may be NULL for
 *            default locale.
 * @param status
 *            A pointer to a standard ICU UErrorCode (input/output parameter).
 *            Its input value must pass the U_SUCCESS() test, or else the
 *            function returns immediately. The caller should check its output
 *            value with U_FAILURE(), or use with function chaining (see User
 *            Guide for details).
 * @return
 *            A pointer to a UListFormatter object for the specified locale,
 *            or NULL if an error occurred.
 * @stable ICU 55
 */
U_CAPI UListFormatter* U_EXPORT2
ulistfmt_open(const char*  locale,
              UErrorCode*  status);

/**
 * Close a UListFormatter object. Once closed it may no longer be used.
 * @param listfmt
 *            The UListFormatter object to close.
 * @stable ICU 55
 */
U_CAPI void U_EXPORT2
ulistfmt_close(UListFormatter *listfmt);

#ifndef U_HIDE_DRAFT_API
/**
 * Creates an object to hold the result of a UListFormatter
 * operation. The object can be used repeatedly; it is cleared whenever
 * passed to a format function.
 *
 * @param ec Set if an error occurs.
 * @return A pointer needing ownership.
 * @draft ICU 64
 */
U_CAPI UFormattedList* U_EXPORT2
ulistfmt_openResult(UErrorCode* ec);

/**
 * Returns a representation of a UFormattedList as a UFormattedValue,
 * which can be subsequently passed to any API requiring that type.
 *
 * The returned object is owned by the UFormattedList and is valid
 * only as long as the UFormattedList is present and unchanged in memory.
 *
 * You can think of this method as a cast between types.
 *
 * When calling ufmtval_nextPosition():
 * The fields are returned from start to end. The special field category
 * UFIELD_CATEGORY_LIST_SPAN is used to indicate which argument
 * was inserted at the given position. The span category will
 * always occur before the corresponding instance of UFIELD_CATEGORY_LIST
 * in the ufmtval_nextPosition() iterator.
 *
 * @param uresult The object containing the formatted string.
 * @param ec Set if an error occurs.
 * @return A UFormattedValue owned by the input object.
 * @draft ICU 64
 */
U_CAPI const UFormattedValue* U_EXPORT2
ulistfmt_resultAsValue(const UFormattedList* uresult, UErrorCode* ec);

/**
 * Releases the UFormattedList created by ulistfmt_openResult().
 *
 * @param uresult The object to release.
 * @draft ICU 64
 */
U_CAPI void U_EXPORT2
ulistfmt_closeResult(UFormattedList* uresult);
#endif /* U_HIDE_DRAFT_API */


#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUListFormatterPointer
 * "Smart pointer" class, closes a UListFormatter via ulistfmt_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 55
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUListFormatterPointer, UListFormatter, ulistfmt_close);

#ifndef U_HIDE_DRAFT_API
/**
 * \class LocalUFormattedListPointer
 * "Smart pointer" class, closes a UFormattedList via ulistfmt_closeResult().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @draft ICU 64
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUFormattedListPointer, UFormattedList, ulistfmt_closeResult);
#endif /* U_HIDE_DRAFT_API */

U_NAMESPACE_END

#endif

/**
 * Formats a list of strings using the conventions established for the
 * UListFormatter object.
 * @param listfmt
 *            The UListFormatter object specifying the list conventions.
 * @param strings
 *            An array of pointers to UChar strings; the array length is
 *            specified by stringCount. Must be non-NULL if stringCount > 0.
 * @param stringLengths
 *            An array of string lengths corresponding to the strings[]
 *            parameter; any individual length value may be negative to indicate
 *            that the corresponding strings[] entry is 0-terminated, or
 *            stringLengths itself may be NULL if all of the strings are
 *            0-terminated. If non-NULL, the stringLengths array must have
 *            stringCount entries.
 * @param stringCount
 *            the number of entries in strings[], and the number of entries
 *            in the stringLengths array if it is not NULL. Must be >= 0.
 * @param result
 *            A pointer to a buffer to receive the formatted list.
 * @param resultCapacity
 *            The maximum size of result.
 * @param status
 *            A pointer to a standard ICU UErrorCode (input/output parameter).
 *            Its input value must pass the U_SUCCESS() test, or else the
 *            function returns immediately. The caller should check its output
 *            value with U_FAILURE(), or use with function chaining (see User
 *            Guide for details).
 * @return
 *            The total buffer size needed; if greater than resultLength, the
 *            output was truncated. May be <=0 if unable to determine the
 *            total buffer size needed (e.g. for illegal arguments).
 * @stable ICU 55
 */
U_CAPI int32_t U_EXPORT2
ulistfmt_format(const UListFormatter* listfmt,
                const UChar* const strings[],
                const int32_t *    stringLengths,
                int32_t            stringCount,
                UChar*             result,
                int32_t            resultCapacity,
                UErrorCode*        status);

#ifndef U_HIDE_DRAFT_API
/**
 * Formats a list of strings to a UFormattedList, which exposes more
 * information than the string exported by ulistfmt_format().
 *
 * @param listfmt
 *            The UListFormatter object specifying the list conventions.
 * @param strings
 *            An array of pointers to UChar strings; the array length is
 *            specified by stringCount. Must be non-NULL if stringCount > 0.
 * @param stringLengths
 *            An array of string lengths corresponding to the strings[]
 *            parameter; any individual length value may be negative to indicate
 *            that the corresponding strings[] entry is 0-terminated, or
 *            stringLengths itself may be NULL if all of the strings are
 *            0-terminated. If non-NULL, the stringLengths array must have
 *            stringCount entries.
 * @param stringCount
 *            the number of entries in strings[], and the number of entries
 *            in the stringLengths array if it is not NULL. Must be >= 0.
 * @param uresult
 *            The object in which to store the result of the list formatting
 *            operation. See ulistfmt_openResult().
 * @param status
 *            Error code set if an error occurred during formatting.
 * @draft ICU 64
 */
U_CAPI void U_EXPORT2
ulistfmt_formatStringsToResult(
                const UListFormatter* listfmt,
                const UChar* const strings[],
                const int32_t *    stringLengths,
                int32_t            stringCount,
                UFormattedList*    uresult,
                UErrorCode*        status);
#endif /* U_HIDE_DRAFT_API */

#endif /* #if !UCONFIG_NO_FORMATTING */














U_NAMESPACE_BEGIN

struct ListFormatInternal : public UMemory {
    SimpleFormatter twoPattern;
    SimpleFormatter startPattern;
    SimpleFormatter middlePattern;
    SimpleFormatter endPattern;

ListFormatInternal(
        const UnicodeString& two,
        const UnicodeString& start,
        const UnicodeString& middle,
        const UnicodeString& end,
        UErrorCode &errorCode) :
        twoPattern(two, 2, 2, errorCode),
        startPattern(start, 2, 2, errorCode),
        middlePattern(middle, 2, 2, errorCode),
        endPattern(end, 2, 2, errorCode) {}

ListFormatInternal(const ListFormatData &data, UErrorCode &errorCode) :
        twoPattern(data.twoPattern, errorCode),
        startPattern(data.startPattern, errorCode),
        middlePattern(data.middlePattern, errorCode),
        endPattern(data.endPattern, errorCode) { }

ListFormatInternal(const ListFormatInternal &other) :
    twoPattern(other.twoPattern),
    startPattern(other.startPattern),
    middlePattern(other.middlePattern),
    endPattern(other.endPattern) { }
};


#if !UCONFIG_NO_FORMATTING
class FormattedListData : public FormattedValueFieldPositionIteratorImpl {
public:
    FormattedListData(UErrorCode& status) : FormattedValueFieldPositionIteratorImpl(5, status) {}
    virtual ~FormattedListData();
};

FormattedListData::~FormattedListData() = default;

UPRV_FORMATTED_VALUE_SUBCLASS_AUTO_IMPL(FormattedList)
#endif


static Hashtable* listPatternHash = nullptr;
static const char STANDARD_STYLE[] = "standard";

U_CDECL_BEGIN
static UBool U_CALLCONV uprv_listformatter_cleanup() {
    delete listPatternHash;
    listPatternHash = nullptr;
    return TRUE;
}

static void U_CALLCONV
uprv_deleteListFormatInternal(void *obj) {
    delete static_cast<ListFormatInternal *>(obj);
}

U_CDECL_END

ListFormatter::ListFormatter(const ListFormatter& other) :
        owned(other.owned), data(other.data) {
    if (other.owned != nullptr) {
        owned = new ListFormatInternal(*other.owned);
        data = owned;
    }
}

ListFormatter& ListFormatter::operator=(const ListFormatter& other) {
    if (this == &other) {
        return *this;
    }
    delete owned;
    if (other.owned) {
        owned = new ListFormatInternal(*other.owned);
        data = owned;
    } else {
        owned = nullptr;
        data = other.data;
    }
    return *this;
}

void ListFormatter::initializeHash(UErrorCode& errorCode) {
    if (U_FAILURE(errorCode)) {
        return;
    }

    listPatternHash = new Hashtable();
    if (listPatternHash == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    listPatternHash->setValueDeleter(uprv_deleteListFormatInternal);
    ucln_i18n_registerCleanup(UCLN_I18N_LIST_FORMATTER, uprv_listformatter_cleanup);

}

const ListFormatInternal* ListFormatter::getListFormatInternal(
        const Locale& locale, const char *style, UErrorCode& errorCode) {
    if (U_FAILURE(errorCode)) {
        return nullptr;
    }
    CharString keyBuffer(locale.getName(), errorCode);
    keyBuffer.append(':', errorCode).append(style, errorCode);
    UnicodeString key(keyBuffer.data(), -1, US_INV);
    ListFormatInternal* result = nullptr;
    static UMutex listFormatterMutex;
    {
        Mutex m(&listFormatterMutex);
        if (listPatternHash == nullptr) {
            initializeHash(errorCode);
            if (U_FAILURE(errorCode)) {
                return nullptr;
            }
        }
        result = static_cast<ListFormatInternal*>(listPatternHash->get(key));
    }
    if (result != nullptr) {
        return result;
    }
    result = loadListFormatInternal(locale, style, errorCode);
    if (U_FAILURE(errorCode)) {
        return nullptr;
    }

    {
        Mutex m(&listFormatterMutex);
        ListFormatInternal* temp = static_cast<ListFormatInternal*>(listPatternHash->get(key));
        if (temp != nullptr) {
            delete result;
            result = temp;
        } else {
            listPatternHash->put(key, result, errorCode);
            if (U_FAILURE(errorCode)) {
                return nullptr;
            }
        }
    }
    return result;
}

static const UChar solidus = 0x2F;
static const UChar aliasPrefix[] = { 0x6C,0x69,0x73,0x74,0x50,0x61,0x74,0x74,0x65,0x72,0x6E,0x2F }; // "listPattern/"
enum {
    kAliasPrefixLen = UPRV_LENGTHOF(aliasPrefix),
    kStyleLenMax = 24 // longest currently is 14
};

struct ListFormatter::ListPatternsSink : public ResourceSink {
    UnicodeString two, start, middle, end;
#if ((U_PLATFORM == U_PF_AIX) || (U_PLATFORM == U_PF_OS390)) && (U_CPLUSPLUS_VERSION < 11)
    char aliasedStyle[kStyleLenMax+1];
    ListPatternsSink() {
      uprv_memset(aliasedStyle, 0, kStyleLenMax+1);
    }
#else
    char aliasedStyle[kStyleLenMax+1] = {0};

    ListPatternsSink() {}
#endif
    virtual ~ListPatternsSink();

    void setAliasedStyle(UnicodeString alias) {
        int32_t startIndex = alias.indexOf(aliasPrefix, kAliasPrefixLen, 0);
        if (startIndex < 0) {
            return;
        }
        startIndex += kAliasPrefixLen;
        int32_t endIndex = alias.indexOf(solidus, startIndex);
        if (endIndex < 0) {
            endIndex = alias.length();
        }
        alias.extract(startIndex, endIndex-startIndex, aliasedStyle, kStyleLenMax+1, US_INV);
        aliasedStyle[kStyleLenMax] = 0;
    }

    void handleValueForPattern(ResourceValue &value, UnicodeString &pattern, UErrorCode &errorCode) {
        if (pattern.isEmpty()) {
            if (value.getType() == URES_ALIAS) {
                if (aliasedStyle[0] == 0) {
                    setAliasedStyle(value.getAliasUnicodeString(errorCode));
                }
            } else {
                pattern = value.getUnicodeString(errorCode);
            }
        }
    }

    virtual void put(const char *key, ResourceValue &value, UBool /*noFallback*/,
            UErrorCode &errorCode) {
        aliasedStyle[0] = 0;
        if (value.getType() == URES_ALIAS) {
            setAliasedStyle(value.getAliasUnicodeString(errorCode));
            return;
        }
        ResourceTable listPatterns = value.getTable(errorCode);
        for (int i = 0; U_SUCCESS(errorCode) && listPatterns.getKeyAndValue(i, key, value); ++i) {
            if (uprv_strcmp(key, "2") == 0) {
                handleValueForPattern(value, two, errorCode);
            } else if (uprv_strcmp(key, "end") == 0) {
                handleValueForPattern(value, end, errorCode);
            } else if (uprv_strcmp(key, "middle") == 0) {
                handleValueForPattern(value, middle, errorCode);
            } else if (uprv_strcmp(key, "start") == 0) {
                handleValueForPattern(value, start, errorCode);
            }
        }
    }
};

// Virtual destructors must be defined out of line.
ListFormatter::ListPatternsSink::~ListPatternsSink() {}

ListFormatInternal* ListFormatter::loadListFormatInternal(
        const Locale& locale, const char * style, UErrorCode& errorCode) {
    UResourceBundle* rb = ures_open(nullptr, locale.getName(), &errorCode);
    rb = ures_getByKeyWithFallback(rb, "listPattern", rb, &errorCode);
    if (U_FAILURE(errorCode)) {
        ures_close(rb);
        return nullptr;
    }
    ListFormatter::ListPatternsSink sink;
    char currentStyle[kStyleLenMax+1];
    uprv_strncpy(currentStyle, style, kStyleLenMax);
    currentStyle[kStyleLenMax] = 0;

    for (;;) {
        ures_getAllItemsWithFallback(rb, currentStyle, sink, errorCode);
        if (U_FAILURE(errorCode) || sink.aliasedStyle[0] == 0 || uprv_strcmp(currentStyle, sink.aliasedStyle) == 0) {
            break;
        }
        uprv_strcpy(currentStyle, sink.aliasedStyle);
    }
    ures_close(rb);
    if (U_FAILURE(errorCode)) {
        return nullptr;
    }
    if (sink.two.isEmpty() || sink.start.isEmpty() || sink.middle.isEmpty() || sink.end.isEmpty()) {
        errorCode = U_MISSING_RESOURCE_ERROR;
        return nullptr;
    }
    ListFormatInternal* result = new ListFormatInternal(sink.two, sink.start, sink.middle, sink.end, errorCode);
    if (result == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    if (U_FAILURE(errorCode)) {
        delete result;
        return nullptr;
    }
    return result;
}

ListFormatter* ListFormatter::createInstance(UErrorCode& errorCode) {
    Locale locale;  // The default locale.
    return createInstance(locale, errorCode);
}

ListFormatter* ListFormatter::createInstance(const Locale& locale, UErrorCode& errorCode) {
    return createInstance(locale, STANDARD_STYLE, errorCode);
}

ListFormatter* ListFormatter::createInstance(const Locale& locale, const char *style, UErrorCode& errorCode) {
    const ListFormatInternal* listFormatInternal = getListFormatInternal(locale, style, errorCode);
    if (U_FAILURE(errorCode)) {
        return nullptr;
    }
    ListFormatter* p = new ListFormatter(listFormatInternal);
    if (p == nullptr) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    return p;
}

ListFormatter::ListFormatter(const ListFormatData& listFormatData, UErrorCode &errorCode) {
    owned = new ListFormatInternal(listFormatData, errorCode);
    data = owned;
}

ListFormatter::ListFormatter(const ListFormatInternal* listFormatterInternal) : owned(nullptr), data(listFormatterInternal) {
}

ListFormatter::~ListFormatter() {
    delete owned;
}

/**
 * Joins first and second using the pattern pat.
 * On entry offset is an offset into first or -1 if offset unspecified.
 * On exit offset is offset of second in result if recordOffset was set
 * Otherwise if it was >=0 it is set to point into result where it used
 * to point into first. On exit, result is the join of first and second
 * according to pat. Any previous value of result gets replaced.
 */
static void joinStringsAndReplace(
        const SimpleFormatter& pat,
        const UnicodeString& first,
        const UnicodeString& second,
        UnicodeString &result,
        UBool recordOffset,
        int32_t &offset,
        int32_t *offsetFirst,
        int32_t *offsetSecond,
        UErrorCode& errorCode) {
    if (U_FAILURE(errorCode)) {
        return;
    }
    const UnicodeString *params[2] = {&first, &second};
    int32_t offsets[2];
    pat.formatAndReplace(
            params,
            UPRV_LENGTHOF(params),
            result,
            offsets,
            UPRV_LENGTHOF(offsets),
            errorCode);
    if (U_FAILURE(errorCode)) {
        return;
    }
    if (offsets[0] == -1 || offsets[1] == -1) {
        errorCode = U_INVALID_FORMAT_ERROR;
        return;
    }
    if (recordOffset) {
        offset = offsets[1];
    } else if (offset >= 0) {
        offset += offsets[0];
    }
    if (offsetFirst != nullptr) *offsetFirst = offsets[0];
    if (offsetSecond != nullptr) *offsetSecond = offsets[1];
}

UnicodeString& ListFormatter::format(
        const UnicodeString items[],
        int32_t nItems,
        UnicodeString& appendTo,
        UErrorCode& errorCode) const {
    int32_t offset;
    return format(items, nItems, appendTo, -1, offset, errorCode);
}

#if !UCONFIG_NO_FORMATTING
UnicodeString& ListFormatter::format(
        const UnicodeString items[],
        int32_t nItems,
        UnicodeString & appendTo,
        FieldPositionIterator* posIter,
        UErrorCode& errorCode) const {
  int32_t offset;
  FieldPositionIteratorHandler handler(posIter, errorCode);
  return format_(items, nItems, appendTo, -1, offset, &handler, errorCode);
}
#endif

UnicodeString& ListFormatter::format(
        const UnicodeString items[],
        int32_t nItems,
        UnicodeString& appendTo,
        int32_t index,
        int32_t &offset,
        UErrorCode& errorCode) const {
  return format_(items, nItems, appendTo, index, offset, nullptr, errorCode);
}

#if !UCONFIG_NO_FORMATTING
FormattedList ListFormatter::formatStringsToValue(
        const UnicodeString items[],
        int32_t nItems,
        UErrorCode& errorCode) const {
    LocalPointer<FormattedListData> result(new FormattedListData(errorCode), errorCode);
    if (U_FAILURE(errorCode)) {
        return FormattedList(errorCode);
    }
    UnicodeString string;
    int32_t offset;
    auto handler = result->getHandler(errorCode);
    handler.setCategory(UFIELD_CATEGORY_LIST);
    format_(items, nItems, string, -1, offset, &handler, errorCode);
    handler.getError(errorCode);
    result->appendString(string, errorCode);
    if (U_FAILURE(errorCode)) {
        return FormattedList(errorCode);
    }

    // Add span fields and sort
    ConstrainedFieldPosition cfpos;
    cfpos.constrainField(UFIELD_CATEGORY_LIST, ULISTFMT_ELEMENT_FIELD);
    int32_t i = 0;
    handler.setCategory(UFIELD_CATEGORY_LIST_SPAN);
    while (result->nextPosition(cfpos, errorCode)) {
        handler.addAttribute(i++, cfpos.getStart(), cfpos.getLimit());
    }
    handler.getError(errorCode);
    if (U_FAILURE(errorCode)) {
        return FormattedList(errorCode);
    }
    result->sort();

    return FormattedList(result.orphan());
}
#endif

UnicodeString& ListFormatter::format_(
        const UnicodeString items[],
        int32_t nItems,
        UnicodeString& appendTo,
        int32_t index,
        int32_t &offset,
        FieldPositionHandler* handler,
        UErrorCode& errorCode) const {
#if !UCONFIG_NO_FORMATTING
    offset = -1;
    if (U_FAILURE(errorCode)) {
        return appendTo;
    }
    if (data == nullptr) {
        errorCode = U_INVALID_STATE_ERROR;
        return appendTo;
    }

    if (nItems <= 0) {
        return appendTo;
    }
    if (nItems == 1) {
        if (index == 0) {
            offset = appendTo.length();
        }
        if (handler != nullptr) {
            handler->addAttribute(ULISTFMT_ELEMENT_FIELD,
                                  appendTo.length(),
                                  appendTo.length() + items[0].length());
        }
        appendTo.append(items[0]);
        return appendTo;
    }
    UnicodeString result(items[0]);
    if (index == 0) {
        offset = 0;
    }
    int32_t offsetFirst = 0;
    int32_t offsetSecond = 0;
    int32_t prefixLength = 0;
    // for n items, there are 2 * (n + 1) boundary including 0 and the upper
    // edge.
    MaybeStackArray<int32_t, 10> offsets((handler != nullptr) ? 2 * (nItems + 1): 0);
    joinStringsAndReplace(
            nItems == 2 ? data->twoPattern : data->startPattern,
            result,
            items[1],
            result,
            index == 1,
            offset,
            &offsetFirst,
            &offsetSecond,
            errorCode);
    if (handler != nullptr) {
        offsets[0] = 0;
        prefixLength += offsetFirst;
        offsets[1] = offsetSecond - prefixLength;
    }
    if (nItems > 2) {
        for (int32_t i = 2; i < nItems - 1; ++i) {
             joinStringsAndReplace(
                     data->middlePattern,
                     result,
                     items[i],
                     result,
                     index == i,
                     offset,
                     &offsetFirst,
                     &offsetSecond,
                     errorCode);
            if (handler != nullptr) {
                prefixLength += offsetFirst;
                offsets[i] = offsetSecond - prefixLength;
            }
        }
        joinStringsAndReplace(
                data->endPattern,
                result,
                items[nItems - 1],
                result,
                index == nItems - 1,
                offset,
                &offsetFirst,
                &offsetSecond,
                errorCode);
        if (handler != nullptr) {
            prefixLength += offsetFirst;
            offsets[nItems - 1] = offsetSecond - prefixLength;
        }
    }
    if (handler != nullptr) {
        // If there are already some data in appendTo, we need to adjust the index
        // by shifting that lenght while insert into handler.
        int32_t shift = appendTo.length() + prefixLength;
        // Output the ULISTFMT_ELEMENT_FIELD in the order of the input elements
        for (int32_t i = 0; i < nItems; ++i) {
            offsets[i + nItems] = offsets[i] + items[i].length() + shift;
            offsets[i] += shift;
            handler->addAttribute(
                ULISTFMT_ELEMENT_FIELD,  // id
                offsets[i],  // index
                offsets[i + nItems]);  // limit
        }
        // The locale pattern may reorder the items (such as in ur-IN locale),
        // so we cannot assume the array is in accendning order.
        // To handle the edging case, just insert the two ends into the array
        // and sort. Then we output ULISTFMT_LITERAL_FIELD if the indecies
        // between the even and odd position are not the same in the sorted array.
        offsets[2 * nItems] = shift - prefixLength;
        offsets[2 * nItems + 1] = result.length() + shift - prefixLength;
        uprv_sortArray(offsets.getAlias(), 2 * (nItems + 1), sizeof(int32_t),
               uprv_int32Comparator, nullptr,
               false, &errorCode);
        for (int32_t i = 0; i <= nItems; ++i) {
          if (offsets[i * 2] != offsets[i * 2 + 1]) {
            handler->addAttribute(
                ULISTFMT_LITERAL_FIELD,  // id
                offsets[i * 2],  // index
                offsets[i * 2 + 1]);  // limit
          }
        }
    }
    if (U_SUCCESS(errorCode)) {
        if (offset >= 0) {
            offset += appendTo.length();
        }
        appendTo += result;
    }
#endif  
    return appendTo;
}

U_NAMESPACE_END
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 20, 2004
* Since: ICU 3.0
**********************************************************************
*/
  // for 'typeid' to work


#if !UCONFIG_NO_FORMATTING







// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2014-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* quantityformatter.h
*/





#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

class SimpleFormatter;
class UnicodeString;
class PluralRules;
class NumberFormat;
class Formattable;
class FieldPosition;
class FormattedStringBuilder;

/**
 * A plural aware formatter that is good for expressing a single quantity and
 * a unit.
 * <p>
 * First use the add() methods to add a pattern for each plural variant.
 * There must be a pattern for the "other" variant.
 * Then use the format() method.
 * <p>
 * Concurrent calls only to const methods on a QuantityFormatter object are 
 * safe, but concurrent const and non-const method calls on a QuantityFormatter
 * object are not safe and require synchronization.
 * 
 */
class U_I18N_API QuantityFormatter : public UMemory {
public:
    /**
     * Default constructor.
     */
    QuantityFormatter();

    /**
     * Copy constructor.
     */
    QuantityFormatter(const QuantityFormatter& other);

    /**
     * Assignment operator
     */
    QuantityFormatter &operator=(const QuantityFormatter& other);

    /**
     * Destructor.
     */
    ~QuantityFormatter();

    /**
     * Removes all variants from this object including the "other" variant.
     */
    void reset();

    /**
     * Adds a plural variant if there is none yet for the plural form.
     *
     * @param variant "zero", "one", "two", "few", "many", "other"
     * @param rawPattern the pattern for the variant e.g "{0} meters"
     * @param status any error returned here.
     * @return TRUE on success; FALSE if status was set to a non zero error.
     */
    UBool addIfAbsent(const char *variant, const UnicodeString &rawPattern, UErrorCode &status);

    /**
     * returns TRUE if this object has at least the "other" variant.
     */
    UBool isValid() const;

    /**
     * Gets the pattern formatter that would be used for a particular variant.
     * If isValid() returns TRUE, this method is guaranteed to return a
     * non-NULL value.
     */
    const SimpleFormatter *getByVariant(const char *variant) const;

    /**
     * Formats a number with this object appending the result to appendTo.
     * At least the "other" variant must be added to this object for this
     * method to work.
     * 
     * @param number the single number.
     * @param fmt formats the number
     * @param rules computes the plural variant to use.
     * @param appendTo result appended here.
     * @param status any error returned here.
     * @return appendTo
     */
    UnicodeString &format(
            const Formattable &number,
            const NumberFormat &fmt,
            const PluralRules &rules,
            UnicodeString &appendTo,
            FieldPosition &pos,
            UErrorCode &status) const;

    /**
     * Selects the standard plural form for the number/formatter/rules.
     * TODO(13591): Remove this method.
     */
    static StandardPlural::Form selectPlural(
            const Formattable &number,
            const NumberFormat &fmt,
            const PluralRules &rules,
            UnicodeString &formattedNumber,
            FieldPosition &pos,
            UErrorCode &status);

    /**
     * Formats a quantity and selects its plural form. The output is appended
     * to a FormattedStringBuilder in order to retain field information.
     *
     * @param quantity The number to format.
     * @param fmt The formatter to use to format the number.
     * @param rules The rules to use to select the plural form of the
     *              formatted number.
     * @param output Where to append the result of the format operation.
     * @param pluralForm Output variable populated with the plural form of the
     *                   formatted number.
     * @param status Set if an error occurs.
     */
    static void formatAndSelect(
            double quantity,
            const NumberFormat& fmt,
            const PluralRules& rules,
            FormattedStringBuilder& output,
            StandardPlural::Form& pluralForm,
            UErrorCode& status);

    /**
     * Formats the pattern with the value and adjusts the FieldPosition.
     * TODO: Remove?
     */
    static UnicodeString &format(
            const SimpleFormatter &pattern,
            const UnicodeString &value,
            UnicodeString &appendTo,
            FieldPosition &pos,
            UErrorCode &status);

private:
    SimpleFormatter *formatters[StandardPlural::COUNT];
};

U_NAMESPACE_END

#endif

















// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING





U_NAMESPACE_BEGIN namespace number {
namespace impl {

class LongNameHandler : public MicroPropsGenerator, public ModifierStore, public UMemory {
  public:
    static UnicodeString getUnitDisplayName(
        const Locale& loc,
        const MeasureUnit& unit,
        UNumberUnitWidth width,
        UErrorCode& status);

    static UnicodeString getUnitPattern(
        const Locale& loc,
        const MeasureUnit& unit,
        UNumberUnitWidth width,
        StandardPlural::Form pluralForm,
        UErrorCode& status);

    static LongNameHandler*
    forCurrencyLongNames(const Locale &loc, const CurrencyUnit &currency, const PluralRules *rules,
                         const MicroPropsGenerator *parent, UErrorCode &status);

    static LongNameHandler*
    forMeasureUnit(const Locale &loc, const MeasureUnit &unit, const MeasureUnit &perUnit,
                   const UNumberUnitWidth &width, const PluralRules *rules,
                   const MicroPropsGenerator *parent, UErrorCode &status);

    void
    processQuantity(DecimalQuantity &quantity, MicroProps &micros, UErrorCode &status) const U_OVERRIDE;

    const Modifier* getModifier(Signum signum, StandardPlural::Form plural) const U_OVERRIDE;

  private:
    SimpleModifier fModifiers[StandardPlural::Form::COUNT];
    const PluralRules *rules;
    const MicroPropsGenerator *parent;

    LongNameHandler(const PluralRules *rules, const MicroPropsGenerator *parent)
            : rules(rules), parent(parent) {}

    static LongNameHandler*
    forCompoundUnit(const Locale &loc, const MeasureUnit &unit, const MeasureUnit &perUnit,
                    const UNumberUnitWidth &width, const PluralRules *rules,
                    const MicroPropsGenerator *parent, UErrorCode &status);

    void simpleFormatsToModifiers(const UnicodeString *simpleFormats, Field field, UErrorCode &status);
    void multiSimpleFormatsToModifiers(const UnicodeString *leadFormats, UnicodeString trailFormat,
                                       Field field, UErrorCode &status);
};

}  // namespace impl
}  // namespace number
U_NAMESPACE_END

#endif //__NUMBER_LONGNAMES_H__



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2014, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* sharednumberformat.h
*/





U_NAMESPACE_BEGIN

class NumberFormat;

class U_I18N_API SharedNumberFormat : public SharedObject {
public:
    SharedNumberFormat(NumberFormat *nfToAdopt) : ptr(nfToAdopt) { }
    virtual ~SharedNumberFormat();
    const NumberFormat *get() const { return ptr; }
    const NumberFormat *operator->() const { return ptr; }
    const NumberFormat &operator*() const { return *ptr; }
private:
    NumberFormat *ptr;
    SharedNumberFormat(const SharedNumberFormat &);
    SharedNumberFormat &operator=(const SharedNumberFormat &);
};

U_NAMESPACE_END


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2014, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* sharedpluralrules.h
*/





U_NAMESPACE_BEGIN

class PluralRules;

class U_I18N_API SharedPluralRules : public SharedObject {
public:
    SharedPluralRules(PluralRules *prToAdopt) : ptr(prToAdopt) { }
    virtual ~SharedPluralRules();
    const PluralRules *operator->() const { return ptr; }
    const PluralRules &operator*() const { return *ptr; }
private:
    PluralRules *ptr;
    SharedPluralRules(const SharedPluralRules &);
    SharedPluralRules &operator=(const SharedPluralRules &);
};

U_NAMESPACE_END






U_NAMESPACE_BEGIN

static constexpr int32_t WIDTH_INDEX_COUNT = UMEASFMT_WIDTH_NARROW + 1;

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(MeasureFormat)

// Used to format durations like 5:47 or 21:35:42.
class NumericDateFormatters : public UMemory {
public:
    // Formats like H:mm
    UnicodeString hourMinute;

    // formats like M:ss
    UnicodeString minuteSecond;

    // formats like H:mm:ss
    UnicodeString hourMinuteSecond;

    // Constructor that takes the actual patterns for hour-minute,
    // minute-second, and hour-minute-second respectively.
    NumericDateFormatters(
            const UnicodeString &hm,
            const UnicodeString &ms,
            const UnicodeString &hms) :
            hourMinute(hm),
            minuteSecond(ms),
            hourMinuteSecond(hms) {
    }
private:
    NumericDateFormatters(const NumericDateFormatters &other);
    NumericDateFormatters &operator=(const NumericDateFormatters &other);
};

static UMeasureFormatWidth getRegularWidth(UMeasureFormatWidth width) {
    if (width >= WIDTH_INDEX_COUNT) {
        return UMEASFMT_WIDTH_NARROW;
    }
    return width;
}

static UNumberUnitWidth getUnitWidth(UMeasureFormatWidth width) {
    switch (width) {
    case UMEASFMT_WIDTH_WIDE:
        return UNUM_UNIT_WIDTH_FULL_NAME;
    case UMEASFMT_WIDTH_NARROW:
    case UMEASFMT_WIDTH_NUMERIC:
        return UNUM_UNIT_WIDTH_NARROW;
    case UMEASFMT_WIDTH_SHORT:
    default:
        return UNUM_UNIT_WIDTH_SHORT;
    }
}

/**
 * Instances contain all MeasureFormat specific data for a particular locale.
 * This data is cached. It is never copied, but is shared via shared pointers.
 *
 * Note: We might change the cache data to have an array[WIDTH_INDEX_COUNT] of
 * complete sets of unit & per patterns,
 * to correspond to the resource data and its aliases.
 *
 * TODO: Maybe store more sparsely in general, with pointers rather than potentially-empty objects.
 */
class MeasureFormatCacheData : public SharedObject {
public:

    /**
     * Redirection data from root-bundle, top-level sideways aliases.
     * - UMEASFMT_WIDTH_COUNT: initial value, just fall back to root
     * - UMEASFMT_WIDTH_WIDE/SHORT/NARROW: sideways alias for missing data
     */
    UMeasureFormatWidth widthFallback[WIDTH_INDEX_COUNT];

    MeasureFormatCacheData();
    virtual ~MeasureFormatCacheData();

    void adoptCurrencyFormat(int32_t widthIndex, NumberFormat *nfToAdopt) {
        delete currencyFormats[widthIndex];
        currencyFormats[widthIndex] = nfToAdopt;
    }
    const NumberFormat *getCurrencyFormat(UMeasureFormatWidth width) const {
        return currencyFormats[getRegularWidth(width)];
    }
    void adoptIntegerFormat(NumberFormat *nfToAdopt) {
        delete integerFormat;
        integerFormat = nfToAdopt;
    }
    const NumberFormat *getIntegerFormat() const {
        return integerFormat;
    }
    void adoptNumericDateFormatters(NumericDateFormatters *formattersToAdopt) {
        delete numericDateFormatters;
        numericDateFormatters = formattersToAdopt;
    }
    const NumericDateFormatters *getNumericDateFormatters() const {
        return numericDateFormatters;
    }

private:
    NumberFormat* currencyFormats[WIDTH_INDEX_COUNT];
    NumberFormat* integerFormat;
    NumericDateFormatters* numericDateFormatters;

    MeasureFormatCacheData(const MeasureFormatCacheData &other);
    MeasureFormatCacheData &operator=(const MeasureFormatCacheData &other);
};

MeasureFormatCacheData::MeasureFormatCacheData()
        : integerFormat(nullptr), numericDateFormatters(nullptr) {
    for (int32_t i = 0; i < WIDTH_INDEX_COUNT; ++i) {
        widthFallback[i] = UMEASFMT_WIDTH_COUNT;
    }
    memset(currencyFormats, 0, sizeof(currencyFormats));
}

MeasureFormatCacheData::~MeasureFormatCacheData() {
    for (int32_t i = 0; i < UPRV_LENGTHOF(currencyFormats); ++i) {
        delete currencyFormats[i];
    }
    // Note: the contents of 'dnams' are pointers into the resource bundle
    delete integerFormat;
    delete numericDateFormatters;
}

static UBool isCurrency(const MeasureUnit &unit) {
    return (uprv_strcmp(unit.getType(), "currency") == 0);
}

static UBool getString(
        const UResourceBundle *resource,
        UnicodeString &result,
        UErrorCode &status) {
    int32_t len = 0;
    const UChar *resStr = ures_getString(resource, &len, &status);
    if (U_FAILURE(status)) {
        return FALSE;
    }
    result.setTo(TRUE, resStr, len);
    return TRUE;
}

static UnicodeString loadNumericDateFormatterPattern(
        const UResourceBundle *resource,
        const char *pattern,
        UErrorCode &status) {
    UnicodeString result;
    if (U_FAILURE(status)) {
        return result;
    }
    CharString chs;
    chs.append("durationUnits", status)
            .append("/", status).append(pattern, status);
    LocalUResourceBundlePointer patternBundle(
            ures_getByKeyWithFallback(
                resource,
                chs.data(),
                NULL,
                &status));
    if (U_FAILURE(status)) {
        return result;
    }
    getString(patternBundle.getAlias(), result, status);
    // Replace 'h' with 'H'
    int32_t len = result.length();
    UChar *buffer = result.getBuffer(len);
    for (int32_t i = 0; i < len; ++i) {
        if (buffer[i] == 0x68) { // 'h'
            buffer[i] = 0x48; // 'H'
        }
    }
    result.releaseBuffer(len);
    return result;
}

static NumericDateFormatters *loadNumericDateFormatters(
        const UResourceBundle *resource,
        UErrorCode &status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    NumericDateFormatters *result = new NumericDateFormatters(
        loadNumericDateFormatterPattern(resource, "hm", status),
        loadNumericDateFormatterPattern(resource, "ms", status),
        loadNumericDateFormatterPattern(resource, "hms", status));
    if (U_FAILURE(status)) {
        delete result;
        return NULL;
    }
    return result;
}

template<> U_I18N_API
const MeasureFormatCacheData *LocaleCacheKey<MeasureFormatCacheData>::createObject(
        const void * /*unused*/, UErrorCode &status) const {
    const char *localeId = fLoc.getName();
    LocalUResourceBundlePointer unitsBundle(ures_open(U_ICUDATA_UNIT, localeId, &status));
    static UNumberFormatStyle currencyStyles[] = {
            UNUM_CURRENCY_PLURAL, UNUM_CURRENCY_ISO, UNUM_CURRENCY};
    LocalPointer<MeasureFormatCacheData> result(new MeasureFormatCacheData(), status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    result->adoptNumericDateFormatters(loadNumericDateFormatters(
            unitsBundle.getAlias(), status));
    if (U_FAILURE(status)) {
        return NULL;
    }

    for (int32_t i = 0; i < WIDTH_INDEX_COUNT; ++i) {
        // NumberFormat::createInstance can erase warning codes from status, so pass it
        // a separate status instance
        UErrorCode localStatus = U_ZERO_ERROR;
        result->adoptCurrencyFormat(i, NumberFormat::createInstance(
                localeId, currencyStyles[i], localStatus));
        if (localStatus != U_ZERO_ERROR) {
            status = localStatus;
        }
        if (U_FAILURE(status)) {
            return NULL;
        }
    }
    NumberFormat *inf = NumberFormat::createInstance(
            localeId, UNUM_DECIMAL, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    inf->setMaximumFractionDigits(0);
    DecimalFormat *decfmt = dynamic_cast<DecimalFormat *>(inf);
    if (decfmt != NULL) {
        decfmt->setRoundingMode(DecimalFormat::kRoundDown);
    }
    result->adoptIntegerFormat(inf);
    result->addRef();
    return result.orphan();
}

static UBool isTimeUnit(const MeasureUnit &mu, const char *tu) {
    return uprv_strcmp(mu.getType(), "duration") == 0 &&
            uprv_strcmp(mu.getSubtype(), tu) == 0;
}

// Converts a composite measure into hours-minutes-seconds and stores at hms
// array. [0] is hours; [1] is minutes; [2] is seconds. Returns a bit map of
// units found: 1=hours, 2=minutes, 4=seconds. For example, if measures
// contains hours-minutes, this function would return 3.
//
// If measures cannot be converted into hours, minutes, seconds or if amounts
// are negative, or if hours, minutes, seconds are out of order, returns 0.
static int32_t toHMS(
        const Measure *measures,
        int32_t measureCount,
        Formattable *hms,
        UErrorCode &status) {
    if (U_FAILURE(status)) {
        return 0;
    }
    int32_t result = 0;
    if (U_FAILURE(status)) {
        return 0;
    }
    // We use copy constructor to ensure that both sides of equality operator
    // are instances of MeasureUnit base class and not a subclass. Otherwise,
    // operator== will immediately return false.
    for (int32_t i = 0; i < measureCount; ++i) {
        if (isTimeUnit(measures[i].getUnit(), "hour")) {
            // hour must come first
            if (result >= 1) {
                return 0;
            }
            hms[0] = measures[i].getNumber();
            if (hms[0].getDouble() < 0.0) {
                return 0;
            }
            result |= 1;
        } else if (isTimeUnit(measures[i].getUnit(), "minute")) {
            // minute must come after hour
            if (result >= 2) {
                return 0;
            }
            hms[1] = measures[i].getNumber();
            if (hms[1].getDouble() < 0.0) {
                return 0;
            }
            result |= 2;
        } else if (isTimeUnit(measures[i].getUnit(), "second")) {
            // second must come after hour and minute
            if (result >= 4) {
                return 0;
            }
            hms[2] = measures[i].getNumber();
            if (hms[2].getDouble() < 0.0) {
                return 0;
            }
            result |= 4;
        } else {
            return 0;
        }
    }
    return result;
}


MeasureFormat::MeasureFormat(
        const Locale &locale, UMeasureFormatWidth w, UErrorCode &status)
        : cache(NULL),
          numberFormat(NULL),
          pluralRules(NULL),
          fWidth(w),
          listFormatter(NULL) {
    initMeasureFormat(locale, w, NULL, status);
}

MeasureFormat::MeasureFormat(
        const Locale &locale,
        UMeasureFormatWidth w,
        NumberFormat *nfToAdopt,
        UErrorCode &status) 
        : cache(NULL),
          numberFormat(NULL),
          pluralRules(NULL),
          fWidth(w),
          listFormatter(NULL) {
    initMeasureFormat(locale, w, nfToAdopt, status);
}

MeasureFormat::MeasureFormat(const MeasureFormat &other) :
        Format(other),
        cache(other.cache),
        numberFormat(other.numberFormat),
        pluralRules(other.pluralRules),
        fWidth(other.fWidth),
        listFormatter(NULL) {
    cache->addRef();
    numberFormat->addRef();
    pluralRules->addRef();
    if (other.listFormatter != NULL) {
        listFormatter = new ListFormatter(*other.listFormatter);
    }
}

MeasureFormat &MeasureFormat::operator=(const MeasureFormat &other) {
    if (this == &other) {
        return *this;
    }
    Format::operator=(other);
    SharedObject::copyPtr(other.cache, cache);
    SharedObject::copyPtr(other.numberFormat, numberFormat);
    SharedObject::copyPtr(other.pluralRules, pluralRules);
    fWidth = other.fWidth;
    delete listFormatter;
    if (other.listFormatter != NULL) {
        listFormatter = new ListFormatter(*other.listFormatter);
    } else {
        listFormatter = NULL;
    }
    return *this;
}

MeasureFormat::MeasureFormat() :
        cache(NULL),
        numberFormat(NULL),
        pluralRules(NULL),
        fWidth(UMEASFMT_WIDTH_SHORT),
        listFormatter(NULL) {
}

MeasureFormat::~MeasureFormat() {
    if (cache != NULL) {
        cache->removeRef();
    }
    if (numberFormat != NULL) {
        numberFormat->removeRef();
    }
    if (pluralRules != NULL) {
        pluralRules->removeRef();
    }
    delete listFormatter;
}

UBool MeasureFormat::operator==(const Format &other) const {
    if (this == &other) { // Same object, equal
        return TRUE;
    }
    if (!Format::operator==(other)) {
        return FALSE;
    }
    const MeasureFormat &rhs = static_cast<const MeasureFormat &>(other);

    // Note: Since the ListFormatter depends only on Locale and width, we
    // don't have to check it here.

    // differing widths aren't equivalent
    if (fWidth != rhs.fWidth) {
        return FALSE;
    }
    // Width the same check locales.
    // We don't need to check locales if both objects have same cache.
    if (cache != rhs.cache) {
        UErrorCode status = U_ZERO_ERROR;
        const char *localeId = getLocaleID(status);
        const char *rhsLocaleId = rhs.getLocaleID(status);
        if (U_FAILURE(status)) {
            // On failure, assume not equal
            return FALSE;
        }
        if (uprv_strcmp(localeId, rhsLocaleId) != 0) {
            return FALSE;
        }
    }
    // Locales same, check NumberFormat if shared data differs.
    return (
            numberFormat == rhs.numberFormat ||
            **numberFormat == **rhs.numberFormat);
}

MeasureFormat *MeasureFormat::clone() const {
    return new MeasureFormat(*this);
}

UnicodeString &MeasureFormat::format(
        const Formattable &obj,
        UnicodeString &appendTo,
        FieldPosition &pos,
        UErrorCode &status) const {
    if (U_FAILURE(status)) return appendTo;
    if (obj.getType() == Formattable::kObject) {
        const UObject* formatObj = obj.getObject();
        const Measure* amount = dynamic_cast<const Measure*>(formatObj);
        if (amount != NULL) {
            return formatMeasure(
                    *amount, **numberFormat, appendTo, pos, status);
        }
    }
    status = U_ILLEGAL_ARGUMENT_ERROR;
    return appendTo;
}

void MeasureFormat::parseObject(
        const UnicodeString & /*source*/,
        Formattable & /*result*/,
        ParsePosition& /*pos*/) const {
    return;
}

UnicodeString &MeasureFormat::formatMeasurePerUnit(
        const Measure &measure,
        const MeasureUnit &perUnit,
        UnicodeString &appendTo,
        FieldPosition &pos,
        UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    auto* df = dynamic_cast<const DecimalFormat*>(&getNumberFormatInternal());
    if (df == nullptr) {
        // Don't know how to handle other types of NumberFormat
        status = U_UNSUPPORTED_ERROR;
        return appendTo;
    }
    number::FormattedNumber result;
    if (auto* lnf = df->toNumberFormatter(status)) {
        result = lnf->unit(measure.getUnit())
            .perUnit(perUnit)
            .unitWidth(getUnitWidth(fWidth))
            .formatDouble(measure.getNumber().getDouble(status), status);
    }
    DecimalFormat::fieldPositionHelper(result, pos, appendTo.length(), status);
    appendTo.append(result.toTempString(status));
    return appendTo;
}

UnicodeString &MeasureFormat::formatMeasures(
        const Measure *measures,
        int32_t measureCount,
        UnicodeString &appendTo,
        FieldPosition &pos,
        UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    if (measureCount == 0) {
        return appendTo;
    }
    if (measureCount == 1) {
        return formatMeasure(measures[0], **numberFormat, appendTo, pos, status);
    }
    if (fWidth == UMEASFMT_WIDTH_NUMERIC) {
        Formattable hms[3];
        int32_t bitMap = toHMS(measures, measureCount, hms, status);
        if (bitMap > 0) {
            return formatNumeric(hms, bitMap, appendTo, status);
        }
    }
    if (pos.getField() != FieldPosition::DONT_CARE) {
        return formatMeasuresSlowTrack(
                measures, measureCount, appendTo, pos, status);
    }
    UnicodeString *results = new UnicodeString[measureCount];
    if (results == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return appendTo;
    }
    for (int32_t i = 0; i < measureCount; ++i) {
        const NumberFormat *nf = cache->getIntegerFormat();
        if (i == measureCount - 1) {
            nf = numberFormat->get();
        }
        formatMeasure(
                measures[i],
                *nf,
                results[i],
                pos,
                status);
    }
    listFormatter->format(results, measureCount, appendTo, status);
    delete [] results;
    return appendTo;
}

UnicodeString MeasureFormat::getUnitDisplayName(const MeasureUnit& unit, UErrorCode& status) const {
    return number::impl::LongNameHandler::getUnitDisplayName(
        getLocale(status),
        unit,
        getUnitWidth(fWidth),
        status);
}

void MeasureFormat::initMeasureFormat(
        const Locale &locale,
        UMeasureFormatWidth w,
        NumberFormat *nfToAdopt,
        UErrorCode &status) {
    static const char *listStyles[] = {"unit", "unit-short", "unit-narrow"};
    LocalPointer<NumberFormat> nf(nfToAdopt);
    if (U_FAILURE(status)) {
        return;
    }
    const char *name = locale.getName();
    setLocaleIDs(name, name);

    UnifiedCache::getByLocale(locale, cache, status);
    if (U_FAILURE(status)) {
        return;
    }

    const SharedPluralRules *pr = PluralRules::createSharedInstance(
            locale, UPLURAL_TYPE_CARDINAL, status);
    if (U_FAILURE(status)) {
        return;
    }
    SharedObject::copyPtr(pr, pluralRules);
    pr->removeRef();
    if (nf.isNull()) {
        // TODO: Clean this up
        const SharedNumberFormat *shared = NumberFormat::createSharedInstance(
                locale, UNUM_DECIMAL, status);
        if (U_FAILURE(status)) {
            return;
        }
        SharedObject::copyPtr(shared, numberFormat);
        shared->removeRef();
    } else {
        adoptNumberFormat(nf.orphan(), status);
        if (U_FAILURE(status)) {
            return;
        }
    }
    fWidth = w;
    delete listFormatter;
    listFormatter = ListFormatter::createInstance(
            locale,
            listStyles[getRegularWidth(fWidth)],
            status);
}

void MeasureFormat::adoptNumberFormat(
        NumberFormat *nfToAdopt, UErrorCode &status) {
    LocalPointer<NumberFormat> nf(nfToAdopt);
    if (U_FAILURE(status)) {
        return;
    }
    SharedNumberFormat *shared = new SharedNumberFormat(nf.getAlias());
    if (shared == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    nf.orphan();
    SharedObject::copyPtr(shared, numberFormat);
}

UBool MeasureFormat::setMeasureFormatLocale(const Locale &locale, UErrorCode &status) {
    if (U_FAILURE(status) || locale == getLocale(status)) {
        return FALSE;
    }
    initMeasureFormat(locale, fWidth, NULL, status);
    return U_SUCCESS(status);
} 

const NumberFormat &MeasureFormat::getNumberFormatInternal() const {
    return **numberFormat;
}

const NumberFormat &MeasureFormat::getCurrencyFormatInternal() const {
    return *cache->getCurrencyFormat(UMEASFMT_WIDTH_NARROW);
}

const PluralRules &MeasureFormat::getPluralRules() const {
    return **pluralRules;
}

Locale MeasureFormat::getLocale(UErrorCode &status) const {
    return Format::getLocale(ULOC_VALID_LOCALE, status);
}

const char *MeasureFormat::getLocaleID(UErrorCode &status) const {
    return Format::getLocaleID(ULOC_VALID_LOCALE, status);
}

UnicodeString &MeasureFormat::formatMeasure(
        const Measure &measure,
        const NumberFormat &nf,
        UnicodeString &appendTo,
        FieldPosition &pos,
        UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    const Formattable& amtNumber = measure.getNumber();
    const MeasureUnit& amtUnit = measure.getUnit();
    if (isCurrency(amtUnit)) {
        UChar isoCode[4];
        u_charsToUChars(amtUnit.getSubtype(), isoCode, 4);
        return cache->getCurrencyFormat(fWidth)->format(
                new CurrencyAmount(amtNumber, isoCode, status),
                appendTo,
                pos,
                status);
    }
    auto* df = dynamic_cast<const DecimalFormat*>(&nf);
    if (df == nullptr) {
        // Handle other types of NumberFormat using the ICU 63 code, modified to
        // get the unitPattern from LongNameHandler and handle fallback to OTHER.
        UnicodeString formattedNumber;
        StandardPlural::Form pluralForm = QuantityFormatter::selectPlural(
                amtNumber, nf, **pluralRules, formattedNumber, pos, status);
        UnicodeString pattern = number::impl::LongNameHandler::getUnitPattern(getLocale(status),
                amtUnit, getUnitWidth(fWidth), pluralForm, status);
        // The above  handles fallback from other widths to short, and from other plural forms to OTHER
        if (U_FAILURE(status)) {
            return appendTo;
        }
        SimpleFormatter formatter(pattern, 0, 1, status);
        return QuantityFormatter::format(formatter, formattedNumber, appendTo, pos, status);
    }
    number::FormattedNumber result;
    if (auto* lnf = df->toNumberFormatter(status)) {
        result = lnf->unit(amtUnit)
            .unitWidth(getUnitWidth(fWidth))
            .formatDouble(amtNumber.getDouble(status), status);
    }
    DecimalFormat::fieldPositionHelper(result, pos, appendTo.length(), status);
    appendTo.append(result.toTempString(status));
    return appendTo;
}


// Formats numeric time duration as 5:00:47 or 3:54.
UnicodeString &MeasureFormat::formatNumeric(
        const Formattable *hms,  // always length 3
        int32_t bitMap,   // 1=hour set, 2=minute set, 4=second set
        UnicodeString &appendTo,
        UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }

    UnicodeString pattern;

    double hours = hms[0].getDouble(status);
    double minutes = hms[1].getDouble(status);
    double seconds = hms[2].getDouble(status);
    if (U_FAILURE(status)) {
        return appendTo;
    }

    // All possible combinations: "h", "m", "s", "hm", "hs", "ms", "hms"
    if (bitMap == 5 || bitMap == 7) { // "hms" & "hs" (we add minutes if "hs")
        pattern = cache->getNumericDateFormatters()->hourMinuteSecond;
        hours = uprv_trunc(hours);
        minutes = uprv_trunc(minutes);
    } else if (bitMap == 3) { // "hm"
        pattern = cache->getNumericDateFormatters()->hourMinute;
        hours = uprv_trunc(hours);
    } else if (bitMap == 6) { // "ms"
        pattern = cache->getNumericDateFormatters()->minuteSecond;
        minutes = uprv_trunc(minutes);
    } else { // h m s, handled outside formatNumeric. No value is also an error.
        status = U_INTERNAL_PROGRAM_ERROR;
        return appendTo;
    }

    const DecimalFormat *numberFormatter = dynamic_cast<const DecimalFormat*>(numberFormat->get());
    if (!numberFormatter) {
        status = U_INTERNAL_PROGRAM_ERROR;
        return appendTo;
    }
    number::LocalizedNumberFormatter numberFormatter2;
    if (auto* lnf = numberFormatter->toNumberFormatter(status)) {
        numberFormatter2 = lnf->integerWidth(number::IntegerWidth::zeroFillTo(2));
    } else {
        return appendTo;
    }

    FormattedStringBuilder fsb;

    UBool protect = FALSE;
    const int32_t patternLength = pattern.length();
    for (int32_t i = 0; i < patternLength; i++) {
        char16_t c = pattern[i];

        // Also set the proper field in this switch
        // We don't use DateFormat.Field because this is not a date / time, is a duration.
        double value = 0;
        switch (c) {
            case u'H': value = hours; break;
            case u'm': value = minutes; break;
            case u's': value = seconds; break;
        }

        // For undefined field we use UNUM_FIELD_COUNT, for historical reasons.
        // See cleanup bug: https://unicode-org.atlassian.net/browse/ICU-20665
        // But we give it a clear name, to keep "the ugly part" in one place.
        constexpr UNumberFormatFields undefinedField = UNUM_FIELD_COUNT;

        // There is not enough info to add Field(s) for the unit because all we have are plain
        // text patterns. For example in "21:51" there is no text for something like "hour",
        // while in something like "21h51" there is ("h"). But we can't really tell...
        switch (c) {
            case u'H':
            case u'm':
            case u's':
                if (protect) {
                    fsb.appendChar16(c, undefinedField, status);
                } else {
                    UnicodeString tmp;
                    if ((i + 1 < patternLength) && pattern[i + 1] == c) { // doubled
                        tmp = numberFormatter2.formatDouble(value, status).toString(status);
                        i++;
                    } else {
                        numberFormatter->format(value, tmp, status);
                    }
                    // TODO: Use proper Field
                    fsb.append(tmp, undefinedField, status);
                }
                break;
            case u'\'':
                // '' is escaped apostrophe
                if ((i + 1 < patternLength) && pattern[i + 1] == c) {
                    fsb.appendChar16(c, undefinedField, status);
                    i++;
                } else {
                    protect = !protect;
                }
                break;
            default:
                fsb.appendChar16(c, undefinedField, status);
        }
    }

    appendTo.append(fsb.toTempUnicodeString());

    return appendTo;
}

UnicodeString &MeasureFormat::formatMeasuresSlowTrack(
        const Measure *measures,
        int32_t measureCount,
        UnicodeString& appendTo,
        FieldPosition& pos,
        UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    FieldPosition dontCare(FieldPosition::DONT_CARE);
    FieldPosition fpos(pos.getField());
    LocalArray<UnicodeString> results(new UnicodeString[measureCount], status);
    int32_t fieldPositionFoundIndex = -1;
    for (int32_t i = 0; i < measureCount; ++i) {
        const NumberFormat *nf = cache->getIntegerFormat();
        if (i == measureCount - 1) {
            nf = numberFormat->get();
        }
        if (fieldPositionFoundIndex == -1) {
            formatMeasure(measures[i], *nf, results[i], fpos, status);
            if (U_FAILURE(status)) {
                return appendTo;
            }
            if (fpos.getBeginIndex() != 0 || fpos.getEndIndex() != 0) {
                fieldPositionFoundIndex = i;
            }
        } else {
            formatMeasure(measures[i], *nf, results[i], dontCare, status);
        }
    }
    int32_t offset;
    listFormatter->format(
            results.getAlias(),
            measureCount,
            appendTo,
            fieldPositionFoundIndex,
            offset,
            status);
    if (U_FAILURE(status)) {
        return appendTo;
    }
    // Fix up FieldPosition indexes if our field is found.
    if (offset != -1) {
        pos.setBeginIndex(fpos.getBeginIndex() + offset);
        pos.setEndIndex(fpos.getEndIndex() + offset);
    }
    return appendTo;
}

MeasureFormat* U_EXPORT2 MeasureFormat::createCurrencyFormat(const Locale& locale,
                                                   UErrorCode& ec) {
    if (U_FAILURE(ec)) {
        return nullptr;
    }
    LocalPointer<CurrencyFormat> fmt(new CurrencyFormat(locale, ec), ec);
    return fmt.orphan();
}

MeasureFormat* U_EXPORT2 MeasureFormat::createCurrencyFormat(UErrorCode& ec) {
    if (U_FAILURE(ec)) {
        return nullptr;
    }
    return MeasureFormat::createCurrencyFormat(Locale::getDefault(), ec);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 26, 2004
* Since: ICU 3.0
**********************************************************************
*/
 // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING






U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(MeasureUnit)

// All code between the "Start generated code" comment and
// the "End generated code" comment is auto generated code
// and must not be edited manually. For instructions on how to correctly
// update this code, refer to:
// http://site.icu-project.org/design/formatting/measureformat/updating-measure-unit
//
// Start generated code


static const int32_t gOffsets[] = {
    0,
    2,
    7,
    17,
    25,
    29,
    328,
    339,
    355,
    359,
    368,
    370,
    374,
    381,
    402,
    404,
    418,
    421,
    427,
    437,
    441,
    445,
    447,
    474
};

static const int32_t gIndexes[] = {
    0,
    2,
    7,
    17,
    25,
    29,
    29,
    40,
    56,
    60,
    69,
    71,
    75,
    82,
    103,
    105,
    119,
    122,
    128,
    138,
    142,
    146,
    148,
    175
};

// Must be sorted alphabetically.
static const char * const gTypes[] = {
    "acceleration",
    "angle",
    "area",
    "concentr",
    "consumption",
    "currency",
    "digital",
    "duration",
    "electric",
    "energy",
    "force",
    "frequency",
    "graphics",
    "length",
    "light",
    "mass",
    "none",
    "power",
    "pressure",
    "speed",
    "temperature",
    "torque",
    "volume"
};

// Must be grouped by type and sorted alphabetically within each type.
static const char * const gSubTypes[] = {
    "g-force",
    "meter-per-second-squared",
    "arc-minute",
    "arc-second",
    "degree",
    "radian",
    "revolution",
    "acre",
    "dunam",
    "hectare",
    "square-centimeter",
    "square-foot",
    "square-inch",
    "square-kilometer",
    "square-meter",
    "square-mile",
    "square-yard",
    "karat",
    "milligram-per-deciliter",
    "millimole-per-liter",
    "mole",
    "part-per-million",
    "percent",
    "permille",
    "permyriad",
    "liter-per-100kilometers",
    "liter-per-kilometer",
    "mile-per-gallon",
    "mile-per-gallon-imperial",
    "ADP",
    "AED",
    "AFA",
    "AFN",
    "ALK",
    "ALL",
    "AMD",
    "ANG",
    "AOA",
    "AOK",
    "AON",
    "AOR",
    "ARA",
    "ARP",
    "ARS",
    "ARY",
    "ATS",
    "AUD",
    "AWG",
    "AYM",
    "AZM",
    "AZN",
    "BAD",
    "BAM",
    "BBD",
    "BDT",
    "BEC",
    "BEF",
    "BEL",
    "BGJ",
    "BGK",
    "BGL",
    "BGN",
    "BHD",
    "BIF",
    "BMD",
    "BND",
    "BOB",
    "BOP",
    "BOV",
    "BRB",
    "BRC",
    "BRE",
    "BRL",
    "BRN",
    "BRR",
    "BSD",
    "BTN",
    "BUK",
    "BWP",
    "BYB",
    "BYN",
    "BYR",
    "BZD",
    "CAD",
    "CDF",
    "CHC",
    "CHE",
    "CHF",
    "CHW",
    "CLF",
    "CLP",
    "CNY",
    "COP",
    "COU",
    "CRC",
    "CSD",
    "CSJ",
    "CSK",
    "CUC",
    "CUP",
    "CVE",
    "CYP",
    "CZK",
    "DDM",
    "DEM",
    "DJF",
    "DKK",
    "DOP",
    "DZD",
    "ECS",
    "ECV",
    "EEK",
    "EGP",
    "ERN",
    "ESA",
    "ESB",
    "ESP",
    "ETB",
    "EUR",
    "FIM",
    "FJD",
    "FKP",
    "FRF",
    "GBP",
    "GEK",
    "GEL",
    "GHC",
    "GHP",
    "GHS",
    "GIP",
    "GMD",
    "GNE",
    "GNF",
    "GNS",
    "GQE",
    "GRD",
    "GTQ",
    "GWE",
    "GWP",
    "GYD",
    "HKD",
    "HNL",
    "HRD",
    "HRK",
    "HTG",
    "HUF",
    "IDR",
    "IEP",
    "ILP",
    "ILR",
    "ILS",
    "INR",
    "IQD",
    "IRR",
    "ISJ",
    "ISK",
    "ITL",
    "JMD",
    "JOD",
    "JPY",
    "KES",
    "KGS",
    "KHR",
    "KMF",
    "KPW",
    "KRW",
    "KWD",
    "KYD",
    "KZT",
    "LAJ",
    "LAK",
    "LBP",
    "LKR",
    "LRD",
    "LSL",
    "LSM",
    "LTL",
    "LTT",
    "LUC",
    "LUF",
    "LUL",
    "LVL",
    "LVR",
    "LYD",
    "MAD",
    "MDL",
    "MGA",
    "MGF",
    "MKD",
    "MLF",
    "MMK",
    "MNT",
    "MOP",
    "MRO",
    "MRU",
    "MTL",
    "MTP",
    "MUR",
    "MVQ",
    "MVR",
    "MWK",
    "MXN",
    "MXP",
    "MXV",
    "MYR",
    "MZE",
    "MZM",
    "MZN",
    "NAD",
    "NGN",
    "NIC",
    "NIO",
    "NLG",
    "NOK",
    "NPR",
    "NZD",
    "OMR",
    "PAB",
    "PEH",
    "PEI",
    "PEN",
    "PES",
    "PGK",
    "PHP",
    "PKR",
    "PLN",
    "PLZ",
    "PTE",
    "PYG",
    "QAR",
    "RHD",
    "ROK",
    "ROL",
    "RON",
    "RSD",
    "RUB",
    "RUR",
    "RWF",
    "SAR",
    "SBD",
    "SCR",
    "SDD",
    "SDG",
    "SDP",
    "SEK",
    "SGD",
    "SHP",
    "SIT",
    "SKK",
    "SLL",
    "SOS",
    "SRD",
    "SRG",
    "SSP",
    "STD",
    "STN",
    "SUR",
    "SVC",
    "SYP",
    "SZL",
    "THB",
    "TJR",
    "TJS",
    "TMM",
    "TMT",
    "TND",
    "TOP",
    "TPE",
    "TRL",
    "TRY",
    "TTD",
    "TWD",
    "TZS",
    "UAH",
    "UAK",
    "UGS",
    "UGW",
    "UGX",
    "USD",
    "USN",
    "USS",
    "UYI",
    "UYN",
    "UYP",
    "UYU",
    "UYW",
    "UZS",
    "VEB",
    "VEF",
    "VES",
    "VNC",
    "VND",
    "VUV",
    "WST",
    "XAF",
    "XAG",
    "XAU",
    "XBA",
    "XBB",
    "XBC",
    "XBD",
    "XCD",
    "XDR",
    "XEU",
    "XOF",
    "XPD",
    "XPF",
    "XPT",
    "XSU",
    "XTS",
    "XUA",
    "XXX",
    "YDD",
    "YER",
    "YUD",
    "YUM",
    "YUN",
    "ZAL",
    "ZAR",
    "ZMK",
    "ZMW",
    "ZRN",
    "ZRZ",
    "ZWC",
    "ZWD",
    "ZWL",
    "ZWN",
    "ZWR",
    "bit",
    "byte",
    "gigabit",
    "gigabyte",
    "kilobit",
    "kilobyte",
    "megabit",
    "megabyte",
    "petabyte",
    "terabit",
    "terabyte",
    "century",
    "day",
    "day-person",
    "decade",
    "hour",
    "microsecond",
    "millisecond",
    "minute",
    "month",
    "month-person",
    "nanosecond",
    "second",
    "week",
    "week-person",
    "year",
    "year-person",
    "ampere",
    "milliampere",
    "ohm",
    "volt",
    "british-thermal-unit",
    "calorie",
    "electronvolt",
    "foodcalorie",
    "joule",
    "kilocalorie",
    "kilojoule",
    "kilowatt-hour",
    "therm-us",
    "newton",
    "pound-force",
    "gigahertz",
    "hertz",
    "kilohertz",
    "megahertz",
    "dot-per-centimeter",
    "dot-per-inch",
    "em",
    "megapixel",
    "pixel",
    "pixel-per-centimeter",
    "pixel-per-inch",
    "astronomical-unit",
    "centimeter",
    "decimeter",
    "fathom",
    "foot",
    "furlong",
    "inch",
    "kilometer",
    "light-year",
    "meter",
    "micrometer",
    "mile",
    "mile-scandinavian",
    "millimeter",
    "nanometer",
    "nautical-mile",
    "parsec",
    "picometer",
    "point",
    "solar-radius",
    "yard",
    "lux",
    "solar-luminosity",
    "carat",
    "dalton",
    "earth-mass",
    "gram",
    "kilogram",
    "metric-ton",
    "microgram",
    "milligram",
    "ounce",
    "ounce-troy",
    "pound",
    "solar-mass",
    "stone",
    "ton",
    "base",
    "percent",
    "permille",
    "gigawatt",
    "horsepower",
    "kilowatt",
    "megawatt",
    "milliwatt",
    "watt",
    "atmosphere",
    "bar",
    "hectopascal",
    "inch-hg",
    "kilopascal",
    "megapascal",
    "millibar",
    "millimeter-of-mercury",
    "pascal",
    "pound-per-square-inch",
    "kilometer-per-hour",
    "knot",
    "meter-per-second",
    "mile-per-hour",
    "celsius",
    "fahrenheit",
    "generic",
    "kelvin",
    "newton-meter",
    "pound-foot",
    "acre-foot",
    "barrel",
    "bushel",
    "centiliter",
    "cubic-centimeter",
    "cubic-foot",
    "cubic-inch",
    "cubic-kilometer",
    "cubic-meter",
    "cubic-mile",
    "cubic-yard",
    "cup",
    "cup-metric",
    "deciliter",
    "fluid-ounce",
    "fluid-ounce-imperial",
    "gallon",
    "gallon-imperial",
    "hectoliter",
    "liter",
    "megaliter",
    "milliliter",
    "pint",
    "pint-metric",
    "quart",
    "tablespoon",
    "teaspoon"
};

// Must be sorted by first value and then second value.
static int32_t unitPerUnitToSingleUnit[][4] = {
        {378, 382, 12, 5},
        {378, 387, 12, 6},
        {388, 343, 19, 0},
        {390, 350, 19, 2},
        {392, 343, 19, 3},
        {392, 463, 4, 2},
        {392, 464, 4, 3},
        {411, 460, 3, 1},
        {414, 12, 18, 9},
        {466, 388, 4, 1}
};

// Shortcuts to the base unit in order to make the default constructor fast
static const int32_t kBaseTypeIdx = 16;
static const int32_t kBaseSubTypeIdx = 0;

MeasureUnit *MeasureUnit::createGForce(UErrorCode &status) {
    return MeasureUnit::create(0, 0, status);
}

MeasureUnit MeasureUnit::getGForce() {
    return MeasureUnit(0, 0);
}

MeasureUnit *MeasureUnit::createMeterPerSecondSquared(UErrorCode &status) {
    return MeasureUnit::create(0, 1, status);
}

MeasureUnit MeasureUnit::getMeterPerSecondSquared() {
    return MeasureUnit(0, 1);
}

MeasureUnit *MeasureUnit::createArcMinute(UErrorCode &status) {
    return MeasureUnit::create(1, 0, status);
}

MeasureUnit MeasureUnit::getArcMinute() {
    return MeasureUnit(1, 0);
}

MeasureUnit *MeasureUnit::createArcSecond(UErrorCode &status) {
    return MeasureUnit::create(1, 1, status);
}

MeasureUnit MeasureUnit::getArcSecond() {
    return MeasureUnit(1, 1);
}

MeasureUnit *MeasureUnit::createDegree(UErrorCode &status) {
    return MeasureUnit::create(1, 2, status);
}

MeasureUnit MeasureUnit::getDegree() {
    return MeasureUnit(1, 2);
}

MeasureUnit *MeasureUnit::createRadian(UErrorCode &status) {
    return MeasureUnit::create(1, 3, status);
}

MeasureUnit MeasureUnit::getRadian() {
    return MeasureUnit(1, 3);
}

MeasureUnit *MeasureUnit::createRevolutionAngle(UErrorCode &status) {
    return MeasureUnit::create(1, 4, status);
}

MeasureUnit MeasureUnit::getRevolutionAngle() {
    return MeasureUnit(1, 4);
}

MeasureUnit *MeasureUnit::createAcre(UErrorCode &status) {
    return MeasureUnit::create(2, 0, status);
}

MeasureUnit MeasureUnit::getAcre() {
    return MeasureUnit(2, 0);
}

MeasureUnit *MeasureUnit::createDunam(UErrorCode &status) {
    return MeasureUnit::create(2, 1, status);
}

MeasureUnit MeasureUnit::getDunam() {
    return MeasureUnit(2, 1);
}

MeasureUnit *MeasureUnit::createHectare(UErrorCode &status) {
    return MeasureUnit::create(2, 2, status);
}

MeasureUnit MeasureUnit::getHectare() {
    return MeasureUnit(2, 2);
}

MeasureUnit *MeasureUnit::createSquareCentimeter(UErrorCode &status) {
    return MeasureUnit::create(2, 3, status);
}

MeasureUnit MeasureUnit::getSquareCentimeter() {
    return MeasureUnit(2, 3);
}

MeasureUnit *MeasureUnit::createSquareFoot(UErrorCode &status) {
    return MeasureUnit::create(2, 4, status);
}

MeasureUnit MeasureUnit::getSquareFoot() {
    return MeasureUnit(2, 4);
}

MeasureUnit *MeasureUnit::createSquareInch(UErrorCode &status) {
    return MeasureUnit::create(2, 5, status);
}

MeasureUnit MeasureUnit::getSquareInch() {
    return MeasureUnit(2, 5);
}

MeasureUnit *MeasureUnit::createSquareKilometer(UErrorCode &status) {
    return MeasureUnit::create(2, 6, status);
}

MeasureUnit MeasureUnit::getSquareKilometer() {
    return MeasureUnit(2, 6);
}

MeasureUnit *MeasureUnit::createSquareMeter(UErrorCode &status) {
    return MeasureUnit::create(2, 7, status);
}

MeasureUnit MeasureUnit::getSquareMeter() {
    return MeasureUnit(2, 7);
}

MeasureUnit *MeasureUnit::createSquareMile(UErrorCode &status) {
    return MeasureUnit::create(2, 8, status);
}

MeasureUnit MeasureUnit::getSquareMile() {
    return MeasureUnit(2, 8);
}

MeasureUnit *MeasureUnit::createSquareYard(UErrorCode &status) {
    return MeasureUnit::create(2, 9, status);
}

MeasureUnit MeasureUnit::getSquareYard() {
    return MeasureUnit(2, 9);
}

MeasureUnit *MeasureUnit::createKarat(UErrorCode &status) {
    return MeasureUnit::create(3, 0, status);
}

MeasureUnit MeasureUnit::getKarat() {
    return MeasureUnit(3, 0);
}

MeasureUnit *MeasureUnit::createMilligramPerDeciliter(UErrorCode &status) {
    return MeasureUnit::create(3, 1, status);
}

MeasureUnit MeasureUnit::getMilligramPerDeciliter() {
    return MeasureUnit(3, 1);
}

MeasureUnit *MeasureUnit::createMillimolePerLiter(UErrorCode &status) {
    return MeasureUnit::create(3, 2, status);
}

MeasureUnit MeasureUnit::getMillimolePerLiter() {
    return MeasureUnit(3, 2);
}

MeasureUnit *MeasureUnit::createMole(UErrorCode &status) {
    return MeasureUnit::create(3, 3, status);
}

MeasureUnit MeasureUnit::getMole() {
    return MeasureUnit(3, 3);
}

MeasureUnit *MeasureUnit::createPartPerMillion(UErrorCode &status) {
    return MeasureUnit::create(3, 4, status);
}

MeasureUnit MeasureUnit::getPartPerMillion() {
    return MeasureUnit(3, 4);
}

MeasureUnit *MeasureUnit::createPercent(UErrorCode &status) {
    return MeasureUnit::create(3, 5, status);
}

MeasureUnit MeasureUnit::getPercent() {
    return MeasureUnit(3, 5);
}

MeasureUnit *MeasureUnit::createPermille(UErrorCode &status) {
    return MeasureUnit::create(3, 6, status);
}

MeasureUnit MeasureUnit::getPermille() {
    return MeasureUnit(3, 6);
}

MeasureUnit *MeasureUnit::createPermyriad(UErrorCode &status) {
    return MeasureUnit::create(3, 7, status);
}

MeasureUnit MeasureUnit::getPermyriad() {
    return MeasureUnit(3, 7);
}

MeasureUnit *MeasureUnit::createLiterPer100Kilometers(UErrorCode &status) {
    return MeasureUnit::create(4, 0, status);
}

MeasureUnit MeasureUnit::getLiterPer100Kilometers() {
    return MeasureUnit(4, 0);
}

MeasureUnit *MeasureUnit::createLiterPerKilometer(UErrorCode &status) {
    return MeasureUnit::create(4, 1, status);
}

MeasureUnit MeasureUnit::getLiterPerKilometer() {
    return MeasureUnit(4, 1);
}

MeasureUnit *MeasureUnit::createMilePerGallon(UErrorCode &status) {
    return MeasureUnit::create(4, 2, status);
}

MeasureUnit MeasureUnit::getMilePerGallon() {
    return MeasureUnit(4, 2);
}

MeasureUnit *MeasureUnit::createMilePerGallonImperial(UErrorCode &status) {
    return MeasureUnit::create(4, 3, status);
}

MeasureUnit MeasureUnit::getMilePerGallonImperial() {
    return MeasureUnit(4, 3);
}

MeasureUnit *MeasureUnit::createBit(UErrorCode &status) {
    return MeasureUnit::create(6, 0, status);
}

MeasureUnit MeasureUnit::getBit() {
    return MeasureUnit(6, 0);
}

MeasureUnit *MeasureUnit::createByte(UErrorCode &status) {
    return MeasureUnit::create(6, 1, status);
}

MeasureUnit MeasureUnit::getByte() {
    return MeasureUnit(6, 1);
}

MeasureUnit *MeasureUnit::createGigabit(UErrorCode &status) {
    return MeasureUnit::create(6, 2, status);
}

MeasureUnit MeasureUnit::getGigabit() {
    return MeasureUnit(6, 2);
}

MeasureUnit *MeasureUnit::createGigabyte(UErrorCode &status) {
    return MeasureUnit::create(6, 3, status);
}

MeasureUnit MeasureUnit::getGigabyte() {
    return MeasureUnit(6, 3);
}

MeasureUnit *MeasureUnit::createKilobit(UErrorCode &status) {
    return MeasureUnit::create(6, 4, status);
}

MeasureUnit MeasureUnit::getKilobit() {
    return MeasureUnit(6, 4);
}

MeasureUnit *MeasureUnit::createKilobyte(UErrorCode &status) {
    return MeasureUnit::create(6, 5, status);
}

MeasureUnit MeasureUnit::getKilobyte() {
    return MeasureUnit(6, 5);
}

MeasureUnit *MeasureUnit::createMegabit(UErrorCode &status) {
    return MeasureUnit::create(6, 6, status);
}

MeasureUnit MeasureUnit::getMegabit() {
    return MeasureUnit(6, 6);
}

MeasureUnit *MeasureUnit::createMegabyte(UErrorCode &status) {
    return MeasureUnit::create(6, 7, status);
}

MeasureUnit MeasureUnit::getMegabyte() {
    return MeasureUnit(6, 7);
}

MeasureUnit *MeasureUnit::createPetabyte(UErrorCode &status) {
    return MeasureUnit::create(6, 8, status);
}

MeasureUnit MeasureUnit::getPetabyte() {
    return MeasureUnit(6, 8);
}

MeasureUnit *MeasureUnit::createTerabit(UErrorCode &status) {
    return MeasureUnit::create(6, 9, status);
}

MeasureUnit MeasureUnit::getTerabit() {
    return MeasureUnit(6, 9);
}

MeasureUnit *MeasureUnit::createTerabyte(UErrorCode &status) {
    return MeasureUnit::create(6, 10, status);
}

MeasureUnit MeasureUnit::getTerabyte() {
    return MeasureUnit(6, 10);
}

MeasureUnit *MeasureUnit::createCentury(UErrorCode &status) {
    return MeasureUnit::create(7, 0, status);
}

MeasureUnit MeasureUnit::getCentury() {
    return MeasureUnit(7, 0);
}

MeasureUnit *MeasureUnit::createDay(UErrorCode &status) {
    return MeasureUnit::create(7, 1, status);
}

MeasureUnit MeasureUnit::getDay() {
    return MeasureUnit(7, 1);
}

MeasureUnit *MeasureUnit::createDayPerson(UErrorCode &status) {
    return MeasureUnit::create(7, 2, status);
}

MeasureUnit MeasureUnit::getDayPerson() {
    return MeasureUnit(7, 2);
}

MeasureUnit *MeasureUnit::createDecade(UErrorCode &status) {
    return MeasureUnit::create(7, 3, status);
}

MeasureUnit MeasureUnit::getDecade() {
    return MeasureUnit(7, 3);
}

MeasureUnit *MeasureUnit::createHour(UErrorCode &status) {
    return MeasureUnit::create(7, 4, status);
}

MeasureUnit MeasureUnit::getHour() {
    return MeasureUnit(7, 4);
}

MeasureUnit *MeasureUnit::createMicrosecond(UErrorCode &status) {
    return MeasureUnit::create(7, 5, status);
}

MeasureUnit MeasureUnit::getMicrosecond() {
    return MeasureUnit(7, 5);
}

MeasureUnit *MeasureUnit::createMillisecond(UErrorCode &status) {
    return MeasureUnit::create(7, 6, status);
}

MeasureUnit MeasureUnit::getMillisecond() {
    return MeasureUnit(7, 6);
}

MeasureUnit *MeasureUnit::createMinute(UErrorCode &status) {
    return MeasureUnit::create(7, 7, status);
}

MeasureUnit MeasureUnit::getMinute() {
    return MeasureUnit(7, 7);
}

MeasureUnit *MeasureUnit::createMonth(UErrorCode &status) {
    return MeasureUnit::create(7, 8, status);
}

MeasureUnit MeasureUnit::getMonth() {
    return MeasureUnit(7, 8);
}

MeasureUnit *MeasureUnit::createMonthPerson(UErrorCode &status) {
    return MeasureUnit::create(7, 9, status);
}

MeasureUnit MeasureUnit::getMonthPerson() {
    return MeasureUnit(7, 9);
}

MeasureUnit *MeasureUnit::createNanosecond(UErrorCode &status) {
    return MeasureUnit::create(7, 10, status);
}

MeasureUnit MeasureUnit::getNanosecond() {
    return MeasureUnit(7, 10);
}

MeasureUnit *MeasureUnit::createSecond(UErrorCode &status) {
    return MeasureUnit::create(7, 11, status);
}

MeasureUnit MeasureUnit::getSecond() {
    return MeasureUnit(7, 11);
}

MeasureUnit *MeasureUnit::createWeek(UErrorCode &status) {
    return MeasureUnit::create(7, 12, status);
}

MeasureUnit MeasureUnit::getWeek() {
    return MeasureUnit(7, 12);
}

MeasureUnit *MeasureUnit::createWeekPerson(UErrorCode &status) {
    return MeasureUnit::create(7, 13, status);
}

MeasureUnit MeasureUnit::getWeekPerson() {
    return MeasureUnit(7, 13);
}

MeasureUnit *MeasureUnit::createYear(UErrorCode &status) {
    return MeasureUnit::create(7, 14, status);
}

MeasureUnit MeasureUnit::getYear() {
    return MeasureUnit(7, 14);
}

MeasureUnit *MeasureUnit::createYearPerson(UErrorCode &status) {
    return MeasureUnit::create(7, 15, status);
}

MeasureUnit MeasureUnit::getYearPerson() {
    return MeasureUnit(7, 15);
}

MeasureUnit *MeasureUnit::createAmpere(UErrorCode &status) {
    return MeasureUnit::create(8, 0, status);
}

MeasureUnit MeasureUnit::getAmpere() {
    return MeasureUnit(8, 0);
}

MeasureUnit *MeasureUnit::createMilliampere(UErrorCode &status) {
    return MeasureUnit::create(8, 1, status);
}

MeasureUnit MeasureUnit::getMilliampere() {
    return MeasureUnit(8, 1);
}

MeasureUnit *MeasureUnit::createOhm(UErrorCode &status) {
    return MeasureUnit::create(8, 2, status);
}

MeasureUnit MeasureUnit::getOhm() {
    return MeasureUnit(8, 2);
}

MeasureUnit *MeasureUnit::createVolt(UErrorCode &status) {
    return MeasureUnit::create(8, 3, status);
}

MeasureUnit MeasureUnit::getVolt() {
    return MeasureUnit(8, 3);
}

MeasureUnit *MeasureUnit::createBritishThermalUnit(UErrorCode &status) {
    return MeasureUnit::create(9, 0, status);
}

MeasureUnit MeasureUnit::getBritishThermalUnit() {
    return MeasureUnit(9, 0);
}

MeasureUnit *MeasureUnit::createCalorie(UErrorCode &status) {
    return MeasureUnit::create(9, 1, status);
}

MeasureUnit MeasureUnit::getCalorie() {
    return MeasureUnit(9, 1);
}

MeasureUnit *MeasureUnit::createElectronvolt(UErrorCode &status) {
    return MeasureUnit::create(9, 2, status);
}

MeasureUnit MeasureUnit::getElectronvolt() {
    return MeasureUnit(9, 2);
}

MeasureUnit *MeasureUnit::createFoodcalorie(UErrorCode &status) {
    return MeasureUnit::create(9, 3, status);
}

MeasureUnit MeasureUnit::getFoodcalorie() {
    return MeasureUnit(9, 3);
}

MeasureUnit *MeasureUnit::createJoule(UErrorCode &status) {
    return MeasureUnit::create(9, 4, status);
}

MeasureUnit MeasureUnit::getJoule() {
    return MeasureUnit(9, 4);
}

MeasureUnit *MeasureUnit::createKilocalorie(UErrorCode &status) {
    return MeasureUnit::create(9, 5, status);
}

MeasureUnit MeasureUnit::getKilocalorie() {
    return MeasureUnit(9, 5);
}

MeasureUnit *MeasureUnit::createKilojoule(UErrorCode &status) {
    return MeasureUnit::create(9, 6, status);
}

MeasureUnit MeasureUnit::getKilojoule() {
    return MeasureUnit(9, 6);
}

MeasureUnit *MeasureUnit::createKilowattHour(UErrorCode &status) {
    return MeasureUnit::create(9, 7, status);
}

MeasureUnit MeasureUnit::getKilowattHour() {
    return MeasureUnit(9, 7);
}

MeasureUnit *MeasureUnit::createThermUs(UErrorCode &status) {
    return MeasureUnit::create(9, 8, status);
}

MeasureUnit MeasureUnit::getThermUs() {
    return MeasureUnit(9, 8);
}

MeasureUnit *MeasureUnit::createNewton(UErrorCode &status) {
    return MeasureUnit::create(10, 0, status);
}

MeasureUnit MeasureUnit::getNewton() {
    return MeasureUnit(10, 0);
}

MeasureUnit *MeasureUnit::createPoundForce(UErrorCode &status) {
    return MeasureUnit::create(10, 1, status);
}

MeasureUnit MeasureUnit::getPoundForce() {
    return MeasureUnit(10, 1);
}

MeasureUnit *MeasureUnit::createGigahertz(UErrorCode &status) {
    return MeasureUnit::create(11, 0, status);
}

MeasureUnit MeasureUnit::getGigahertz() {
    return MeasureUnit(11, 0);
}

MeasureUnit *MeasureUnit::createHertz(UErrorCode &status) {
    return MeasureUnit::create(11, 1, status);
}

MeasureUnit MeasureUnit::getHertz() {
    return MeasureUnit(11, 1);
}

MeasureUnit *MeasureUnit::createKilohertz(UErrorCode &status) {
    return MeasureUnit::create(11, 2, status);
}

MeasureUnit MeasureUnit::getKilohertz() {
    return MeasureUnit(11, 2);
}

MeasureUnit *MeasureUnit::createMegahertz(UErrorCode &status) {
    return MeasureUnit::create(11, 3, status);
}

MeasureUnit MeasureUnit::getMegahertz() {
    return MeasureUnit(11, 3);
}

MeasureUnit *MeasureUnit::createDotPerCentimeter(UErrorCode &status) {
    return MeasureUnit::create(12, 0, status);
}

MeasureUnit MeasureUnit::getDotPerCentimeter() {
    return MeasureUnit(12, 0);
}

MeasureUnit *MeasureUnit::createDotPerInch(UErrorCode &status) {
    return MeasureUnit::create(12, 1, status);
}

MeasureUnit MeasureUnit::getDotPerInch() {
    return MeasureUnit(12, 1);
}

MeasureUnit *MeasureUnit::createEm(UErrorCode &status) {
    return MeasureUnit::create(12, 2, status);
}

MeasureUnit MeasureUnit::getEm() {
    return MeasureUnit(12, 2);
}

MeasureUnit *MeasureUnit::createMegapixel(UErrorCode &status) {
    return MeasureUnit::create(12, 3, status);
}

MeasureUnit MeasureUnit::getMegapixel() {
    return MeasureUnit(12, 3);
}

MeasureUnit *MeasureUnit::createPixel(UErrorCode &status) {
    return MeasureUnit::create(12, 4, status);
}

MeasureUnit MeasureUnit::getPixel() {
    return MeasureUnit(12, 4);
}

MeasureUnit *MeasureUnit::createPixelPerCentimeter(UErrorCode &status) {
    return MeasureUnit::create(12, 5, status);
}

MeasureUnit MeasureUnit::getPixelPerCentimeter() {
    return MeasureUnit(12, 5);
}

MeasureUnit *MeasureUnit::createPixelPerInch(UErrorCode &status) {
    return MeasureUnit::create(12, 6, status);
}

MeasureUnit MeasureUnit::getPixelPerInch() {
    return MeasureUnit(12, 6);
}

MeasureUnit *MeasureUnit::createAstronomicalUnit(UErrorCode &status) {
    return MeasureUnit::create(13, 0, status);
}

MeasureUnit MeasureUnit::getAstronomicalUnit() {
    return MeasureUnit(13, 0);
}

MeasureUnit *MeasureUnit::createCentimeter(UErrorCode &status) {
    return MeasureUnit::create(13, 1, status);
}

MeasureUnit MeasureUnit::getCentimeter() {
    return MeasureUnit(13, 1);
}

MeasureUnit *MeasureUnit::createDecimeter(UErrorCode &status) {
    return MeasureUnit::create(13, 2, status);
}

MeasureUnit MeasureUnit::getDecimeter() {
    return MeasureUnit(13, 2);
}

MeasureUnit *MeasureUnit::createFathom(UErrorCode &status) {
    return MeasureUnit::create(13, 3, status);
}

MeasureUnit MeasureUnit::getFathom() {
    return MeasureUnit(13, 3);
}

MeasureUnit *MeasureUnit::createFoot(UErrorCode &status) {
    return MeasureUnit::create(13, 4, status);
}

MeasureUnit MeasureUnit::getFoot() {
    return MeasureUnit(13, 4);
}

MeasureUnit *MeasureUnit::createFurlong(UErrorCode &status) {
    return MeasureUnit::create(13, 5, status);
}

MeasureUnit MeasureUnit::getFurlong() {
    return MeasureUnit(13, 5);
}

MeasureUnit *MeasureUnit::createInch(UErrorCode &status) {
    return MeasureUnit::create(13, 6, status);
}

MeasureUnit MeasureUnit::getInch() {
    return MeasureUnit(13, 6);
}

MeasureUnit *MeasureUnit::createKilometer(UErrorCode &status) {
    return MeasureUnit::create(13, 7, status);
}

MeasureUnit MeasureUnit::getKilometer() {
    return MeasureUnit(13, 7);
}

MeasureUnit *MeasureUnit::createLightYear(UErrorCode &status) {
    return MeasureUnit::create(13, 8, status);
}

MeasureUnit MeasureUnit::getLightYear() {
    return MeasureUnit(13, 8);
}

MeasureUnit *MeasureUnit::createMeter(UErrorCode &status) {
    return MeasureUnit::create(13, 9, status);
}

MeasureUnit MeasureUnit::getMeter() {
    return MeasureUnit(13, 9);
}

MeasureUnit *MeasureUnit::createMicrometer(UErrorCode &status) {
    return MeasureUnit::create(13, 10, status);
}

MeasureUnit MeasureUnit::getMicrometer() {
    return MeasureUnit(13, 10);
}

MeasureUnit *MeasureUnit::createMile(UErrorCode &status) {
    return MeasureUnit::create(13, 11, status);
}

MeasureUnit MeasureUnit::getMile() {
    return MeasureUnit(13, 11);
}

MeasureUnit *MeasureUnit::createMileScandinavian(UErrorCode &status) {
    return MeasureUnit::create(13, 12, status);
}

MeasureUnit MeasureUnit::getMileScandinavian() {
    return MeasureUnit(13, 12);
}

MeasureUnit *MeasureUnit::createMillimeter(UErrorCode &status) {
    return MeasureUnit::create(13, 13, status);
}

MeasureUnit MeasureUnit::getMillimeter() {
    return MeasureUnit(13, 13);
}

MeasureUnit *MeasureUnit::createNanometer(UErrorCode &status) {
    return MeasureUnit::create(13, 14, status);
}

MeasureUnit MeasureUnit::getNanometer() {
    return MeasureUnit(13, 14);
}

MeasureUnit *MeasureUnit::createNauticalMile(UErrorCode &status) {
    return MeasureUnit::create(13, 15, status);
}

MeasureUnit MeasureUnit::getNauticalMile() {
    return MeasureUnit(13, 15);
}

MeasureUnit *MeasureUnit::createParsec(UErrorCode &status) {
    return MeasureUnit::create(13, 16, status);
}

MeasureUnit MeasureUnit::getParsec() {
    return MeasureUnit(13, 16);
}

MeasureUnit *MeasureUnit::createPicometer(UErrorCode &status) {
    return MeasureUnit::create(13, 17, status);
}

MeasureUnit MeasureUnit::getPicometer() {
    return MeasureUnit(13, 17);
}

MeasureUnit *MeasureUnit::createPoint(UErrorCode &status) {
    return MeasureUnit::create(13, 18, status);
}

MeasureUnit MeasureUnit::getPoint() {
    return MeasureUnit(13, 18);
}

MeasureUnit *MeasureUnit::createSolarRadius(UErrorCode &status) {
    return MeasureUnit::create(13, 19, status);
}

MeasureUnit MeasureUnit::getSolarRadius() {
    return MeasureUnit(13, 19);
}

MeasureUnit *MeasureUnit::createYard(UErrorCode &status) {
    return MeasureUnit::create(13, 20, status);
}

MeasureUnit MeasureUnit::getYard() {
    return MeasureUnit(13, 20);
}

MeasureUnit *MeasureUnit::createLux(UErrorCode &status) {
    return MeasureUnit::create(14, 0, status);
}

MeasureUnit MeasureUnit::getLux() {
    return MeasureUnit(14, 0);
}

MeasureUnit *MeasureUnit::createSolarLuminosity(UErrorCode &status) {
    return MeasureUnit::create(14, 1, status);
}

MeasureUnit MeasureUnit::getSolarLuminosity() {
    return MeasureUnit(14, 1);
}

MeasureUnit *MeasureUnit::createCarat(UErrorCode &status) {
    return MeasureUnit::create(15, 0, status);
}

MeasureUnit MeasureUnit::getCarat() {
    return MeasureUnit(15, 0);
}

MeasureUnit *MeasureUnit::createDalton(UErrorCode &status) {
    return MeasureUnit::create(15, 1, status);
}

MeasureUnit MeasureUnit::getDalton() {
    return MeasureUnit(15, 1);
}

MeasureUnit *MeasureUnit::createEarthMass(UErrorCode &status) {
    return MeasureUnit::create(15, 2, status);
}

MeasureUnit MeasureUnit::getEarthMass() {
    return MeasureUnit(15, 2);
}

MeasureUnit *MeasureUnit::createGram(UErrorCode &status) {
    return MeasureUnit::create(15, 3, status);
}

MeasureUnit MeasureUnit::getGram() {
    return MeasureUnit(15, 3);
}

MeasureUnit *MeasureUnit::createKilogram(UErrorCode &status) {
    return MeasureUnit::create(15, 4, status);
}

MeasureUnit MeasureUnit::getKilogram() {
    return MeasureUnit(15, 4);
}

MeasureUnit *MeasureUnit::createMetricTon(UErrorCode &status) {
    return MeasureUnit::create(15, 5, status);
}

MeasureUnit MeasureUnit::getMetricTon() {
    return MeasureUnit(15, 5);
}

MeasureUnit *MeasureUnit::createMicrogram(UErrorCode &status) {
    return MeasureUnit::create(15, 6, status);
}

MeasureUnit MeasureUnit::getMicrogram() {
    return MeasureUnit(15, 6);
}

MeasureUnit *MeasureUnit::createMilligram(UErrorCode &status) {
    return MeasureUnit::create(15, 7, status);
}

MeasureUnit MeasureUnit::getMilligram() {
    return MeasureUnit(15, 7);
}

MeasureUnit *MeasureUnit::createOunce(UErrorCode &status) {
    return MeasureUnit::create(15, 8, status);
}

MeasureUnit MeasureUnit::getOunce() {
    return MeasureUnit(15, 8);
}

MeasureUnit *MeasureUnit::createOunceTroy(UErrorCode &status) {
    return MeasureUnit::create(15, 9, status);
}

MeasureUnit MeasureUnit::getOunceTroy() {
    return MeasureUnit(15, 9);
}

MeasureUnit *MeasureUnit::createPound(UErrorCode &status) {
    return MeasureUnit::create(15, 10, status);
}

MeasureUnit MeasureUnit::getPound() {
    return MeasureUnit(15, 10);
}

MeasureUnit *MeasureUnit::createSolarMass(UErrorCode &status) {
    return MeasureUnit::create(15, 11, status);
}

MeasureUnit MeasureUnit::getSolarMass() {
    return MeasureUnit(15, 11);
}

MeasureUnit *MeasureUnit::createStone(UErrorCode &status) {
    return MeasureUnit::create(15, 12, status);
}

MeasureUnit MeasureUnit::getStone() {
    return MeasureUnit(15, 12);
}

MeasureUnit *MeasureUnit::createTon(UErrorCode &status) {
    return MeasureUnit::create(15, 13, status);
}

MeasureUnit MeasureUnit::getTon() {
    return MeasureUnit(15, 13);
}

MeasureUnit *MeasureUnit::createGigawatt(UErrorCode &status) {
    return MeasureUnit::create(17, 0, status);
}

MeasureUnit MeasureUnit::getGigawatt() {
    return MeasureUnit(17, 0);
}

MeasureUnit *MeasureUnit::createHorsepower(UErrorCode &status) {
    return MeasureUnit::create(17, 1, status);
}

MeasureUnit MeasureUnit::getHorsepower() {
    return MeasureUnit(17, 1);
}

MeasureUnit *MeasureUnit::createKilowatt(UErrorCode &status) {
    return MeasureUnit::create(17, 2, status);
}

MeasureUnit MeasureUnit::getKilowatt() {
    return MeasureUnit(17, 2);
}

MeasureUnit *MeasureUnit::createMegawatt(UErrorCode &status) {
    return MeasureUnit::create(17, 3, status);
}

MeasureUnit MeasureUnit::getMegawatt() {
    return MeasureUnit(17, 3);
}

MeasureUnit *MeasureUnit::createMilliwatt(UErrorCode &status) {
    return MeasureUnit::create(17, 4, status);
}

MeasureUnit MeasureUnit::getMilliwatt() {
    return MeasureUnit(17, 4);
}

MeasureUnit *MeasureUnit::createWatt(UErrorCode &status) {
    return MeasureUnit::create(17, 5, status);
}

MeasureUnit MeasureUnit::getWatt() {
    return MeasureUnit(17, 5);
}

MeasureUnit *MeasureUnit::createAtmosphere(UErrorCode &status) {
    return MeasureUnit::create(18, 0, status);
}

MeasureUnit MeasureUnit::getAtmosphere() {
    return MeasureUnit(18, 0);
}

MeasureUnit *MeasureUnit::createBar(UErrorCode &status) {
    return MeasureUnit::create(18, 1, status);
}

MeasureUnit MeasureUnit::getBar() {
    return MeasureUnit(18, 1);
}

MeasureUnit *MeasureUnit::createHectopascal(UErrorCode &status) {
    return MeasureUnit::create(18, 2, status);
}

MeasureUnit MeasureUnit::getHectopascal() {
    return MeasureUnit(18, 2);
}

MeasureUnit *MeasureUnit::createInchHg(UErrorCode &status) {
    return MeasureUnit::create(18, 3, status);
}

MeasureUnit MeasureUnit::getInchHg() {
    return MeasureUnit(18, 3);
}

MeasureUnit *MeasureUnit::createKilopascal(UErrorCode &status) {
    return MeasureUnit::create(18, 4, status);
}

MeasureUnit MeasureUnit::getKilopascal() {
    return MeasureUnit(18, 4);
}

MeasureUnit *MeasureUnit::createMegapascal(UErrorCode &status) {
    return MeasureUnit::create(18, 5, status);
}

MeasureUnit MeasureUnit::getMegapascal() {
    return MeasureUnit(18, 5);
}

MeasureUnit *MeasureUnit::createMillibar(UErrorCode &status) {
    return MeasureUnit::create(18, 6, status);
}

MeasureUnit MeasureUnit::getMillibar() {
    return MeasureUnit(18, 6);
}

MeasureUnit *MeasureUnit::createMillimeterOfMercury(UErrorCode &status) {
    return MeasureUnit::create(18, 7, status);
}

MeasureUnit MeasureUnit::getMillimeterOfMercury() {
    return MeasureUnit(18, 7);
}

MeasureUnit *MeasureUnit::createPascal(UErrorCode &status) {
    return MeasureUnit::create(18, 8, status);
}

MeasureUnit MeasureUnit::getPascal() {
    return MeasureUnit(18, 8);
}

MeasureUnit *MeasureUnit::createPoundPerSquareInch(UErrorCode &status) {
    return MeasureUnit::create(18, 9, status);
}

MeasureUnit MeasureUnit::getPoundPerSquareInch() {
    return MeasureUnit(18, 9);
}

MeasureUnit *MeasureUnit::createKilometerPerHour(UErrorCode &status) {
    return MeasureUnit::create(19, 0, status);
}

MeasureUnit MeasureUnit::getKilometerPerHour() {
    return MeasureUnit(19, 0);
}

MeasureUnit *MeasureUnit::createKnot(UErrorCode &status) {
    return MeasureUnit::create(19, 1, status);
}

MeasureUnit MeasureUnit::getKnot() {
    return MeasureUnit(19, 1);
}

MeasureUnit *MeasureUnit::createMeterPerSecond(UErrorCode &status) {
    return MeasureUnit::create(19, 2, status);
}

MeasureUnit MeasureUnit::getMeterPerSecond() {
    return MeasureUnit(19, 2);
}

MeasureUnit *MeasureUnit::createMilePerHour(UErrorCode &status) {
    return MeasureUnit::create(19, 3, status);
}

MeasureUnit MeasureUnit::getMilePerHour() {
    return MeasureUnit(19, 3);
}

MeasureUnit *MeasureUnit::createCelsius(UErrorCode &status) {
    return MeasureUnit::create(20, 0, status);
}

MeasureUnit MeasureUnit::getCelsius() {
    return MeasureUnit(20, 0);
}

MeasureUnit *MeasureUnit::createFahrenheit(UErrorCode &status) {
    return MeasureUnit::create(20, 1, status);
}

MeasureUnit MeasureUnit::getFahrenheit() {
    return MeasureUnit(20, 1);
}

MeasureUnit *MeasureUnit::createGenericTemperature(UErrorCode &status) {
    return MeasureUnit::create(20, 2, status);
}

MeasureUnit MeasureUnit::getGenericTemperature() {
    return MeasureUnit(20, 2);
}

MeasureUnit *MeasureUnit::createKelvin(UErrorCode &status) {
    return MeasureUnit::create(20, 3, status);
}

MeasureUnit MeasureUnit::getKelvin() {
    return MeasureUnit(20, 3);
}

MeasureUnit *MeasureUnit::createNewtonMeter(UErrorCode &status) {
    return MeasureUnit::create(21, 0, status);
}

MeasureUnit MeasureUnit::getNewtonMeter() {
    return MeasureUnit(21, 0);
}

MeasureUnit *MeasureUnit::createPoundFoot(UErrorCode &status) {
    return MeasureUnit::create(21, 1, status);
}

MeasureUnit MeasureUnit::getPoundFoot() {
    return MeasureUnit(21, 1);
}

MeasureUnit *MeasureUnit::createAcreFoot(UErrorCode &status) {
    return MeasureUnit::create(22, 0, status);
}

MeasureUnit MeasureUnit::getAcreFoot() {
    return MeasureUnit(22, 0);
}

MeasureUnit *MeasureUnit::createBarrel(UErrorCode &status) {
    return MeasureUnit::create(22, 1, status);
}

MeasureUnit MeasureUnit::getBarrel() {
    return MeasureUnit(22, 1);
}

MeasureUnit *MeasureUnit::createBushel(UErrorCode &status) {
    return MeasureUnit::create(22, 2, status);
}

MeasureUnit MeasureUnit::getBushel() {
    return MeasureUnit(22, 2);
}

MeasureUnit *MeasureUnit::createCentiliter(UErrorCode &status) {
    return MeasureUnit::create(22, 3, status);
}

MeasureUnit MeasureUnit::getCentiliter() {
    return MeasureUnit(22, 3);
}

MeasureUnit *MeasureUnit::createCubicCentimeter(UErrorCode &status) {
    return MeasureUnit::create(22, 4, status);
}

MeasureUnit MeasureUnit::getCubicCentimeter() {
    return MeasureUnit(22, 4);
}

MeasureUnit *MeasureUnit::createCubicFoot(UErrorCode &status) {
    return MeasureUnit::create(22, 5, status);
}

MeasureUnit MeasureUnit::getCubicFoot() {
    return MeasureUnit(22, 5);
}

MeasureUnit *MeasureUnit::createCubicInch(UErrorCode &status) {
    return MeasureUnit::create(22, 6, status);
}

MeasureUnit MeasureUnit::getCubicInch() {
    return MeasureUnit(22, 6);
}

MeasureUnit *MeasureUnit::createCubicKilometer(UErrorCode &status) {
    return MeasureUnit::create(22, 7, status);
}

MeasureUnit MeasureUnit::getCubicKilometer() {
    return MeasureUnit(22, 7);
}

MeasureUnit *MeasureUnit::createCubicMeter(UErrorCode &status) {
    return MeasureUnit::create(22, 8, status);
}

MeasureUnit MeasureUnit::getCubicMeter() {
    return MeasureUnit(22, 8);
}

MeasureUnit *MeasureUnit::createCubicMile(UErrorCode &status) {
    return MeasureUnit::create(22, 9, status);
}

MeasureUnit MeasureUnit::getCubicMile() {
    return MeasureUnit(22, 9);
}

MeasureUnit *MeasureUnit::createCubicYard(UErrorCode &status) {
    return MeasureUnit::create(22, 10, status);
}

MeasureUnit MeasureUnit::getCubicYard() {
    return MeasureUnit(22, 10);
}

MeasureUnit *MeasureUnit::createCup(UErrorCode &status) {
    return MeasureUnit::create(22, 11, status);
}

MeasureUnit MeasureUnit::getCup() {
    return MeasureUnit(22, 11);
}

MeasureUnit *MeasureUnit::createCupMetric(UErrorCode &status) {
    return MeasureUnit::create(22, 12, status);
}

MeasureUnit MeasureUnit::getCupMetric() {
    return MeasureUnit(22, 12);
}

MeasureUnit *MeasureUnit::createDeciliter(UErrorCode &status) {
    return MeasureUnit::create(22, 13, status);
}

MeasureUnit MeasureUnit::getDeciliter() {
    return MeasureUnit(22, 13);
}

MeasureUnit *MeasureUnit::createFluidOunce(UErrorCode &status) {
    return MeasureUnit::create(22, 14, status);
}

MeasureUnit MeasureUnit::getFluidOunce() {
    return MeasureUnit(22, 14);
}

MeasureUnit *MeasureUnit::createFluidOunceImperial(UErrorCode &status) {
    return MeasureUnit::create(22, 15, status);
}

MeasureUnit MeasureUnit::getFluidOunceImperial() {
    return MeasureUnit(22, 15);
}

MeasureUnit *MeasureUnit::createGallon(UErrorCode &status) {
    return MeasureUnit::create(22, 16, status);
}

MeasureUnit MeasureUnit::getGallon() {
    return MeasureUnit(22, 16);
}

MeasureUnit *MeasureUnit::createGallonImperial(UErrorCode &status) {
    return MeasureUnit::create(22, 17, status);
}

MeasureUnit MeasureUnit::getGallonImperial() {
    return MeasureUnit(22, 17);
}

MeasureUnit *MeasureUnit::createHectoliter(UErrorCode &status) {
    return MeasureUnit::create(22, 18, status);
}

MeasureUnit MeasureUnit::getHectoliter() {
    return MeasureUnit(22, 18);
}

MeasureUnit *MeasureUnit::createLiter(UErrorCode &status) {
    return MeasureUnit::create(22, 19, status);
}

MeasureUnit MeasureUnit::getLiter() {
    return MeasureUnit(22, 19);
}

MeasureUnit *MeasureUnit::createMegaliter(UErrorCode &status) {
    return MeasureUnit::create(22, 20, status);
}

MeasureUnit MeasureUnit::getMegaliter() {
    return MeasureUnit(22, 20);
}

MeasureUnit *MeasureUnit::createMilliliter(UErrorCode &status) {
    return MeasureUnit::create(22, 21, status);
}

MeasureUnit MeasureUnit::getMilliliter() {
    return MeasureUnit(22, 21);
}

MeasureUnit *MeasureUnit::createPint(UErrorCode &status) {
    return MeasureUnit::create(22, 22, status);
}

MeasureUnit MeasureUnit::getPint() {
    return MeasureUnit(22, 22);
}

MeasureUnit *MeasureUnit::createPintMetric(UErrorCode &status) {
    return MeasureUnit::create(22, 23, status);
}

MeasureUnit MeasureUnit::getPintMetric() {
    return MeasureUnit(22, 23);
}

MeasureUnit *MeasureUnit::createQuart(UErrorCode &status) {
    return MeasureUnit::create(22, 24, status);
}

MeasureUnit MeasureUnit::getQuart() {
    return MeasureUnit(22, 24);
}

MeasureUnit *MeasureUnit::createTablespoon(UErrorCode &status) {
    return MeasureUnit::create(22, 25, status);
}

MeasureUnit MeasureUnit::getTablespoon() {
    return MeasureUnit(22, 25);
}

MeasureUnit *MeasureUnit::createTeaspoon(UErrorCode &status) {
    return MeasureUnit::create(22, 26, status);
}

MeasureUnit MeasureUnit::getTeaspoon() {
    return MeasureUnit(22, 26);
}

// End generated code

static int32_t binarySearch(
        const char * const * array, int32_t start, int32_t end, const char * key) {
    while (start < end) {
        int32_t mid = (start + end) / 2;
        int32_t cmp = uprv_strcmp(array[mid], key);
        if (cmp < 0) {
            start = mid + 1;
            continue;
        }
        if (cmp == 0) {
            return mid;
        }
        end = mid;
    }
    return -1;
}

MeasureUnit::MeasureUnit() {
    fCurrency[0] = 0;
    fTypeId = kBaseTypeIdx;
    fSubTypeId = kBaseSubTypeIdx;
}

MeasureUnit::MeasureUnit(const MeasureUnit &other)
        : fTypeId(other.fTypeId), fSubTypeId(other.fSubTypeId) {
    uprv_strcpy(fCurrency, other.fCurrency);
}

MeasureUnit &MeasureUnit::operator=(const MeasureUnit &other) {
    if (this == &other) {
        return *this;
    }
    fTypeId = other.fTypeId;
    fSubTypeId = other.fSubTypeId;
    uprv_strcpy(fCurrency, other.fCurrency);
    return *this;
}

MeasureUnit *MeasureUnit::clone() const {
    return new MeasureUnit(*this);
}

MeasureUnit::~MeasureUnit() {
}

const char *MeasureUnit::getType() const {
    return gTypes[fTypeId];
}

const char *MeasureUnit::getSubtype() const {
    return fCurrency[0] == 0 ? gSubTypes[getOffset()] : fCurrency;
}

UBool MeasureUnit::operator==(const UObject& other) const {
    if (this == &other) {  // Same object, equal
        return TRUE;
    }
    if (typeid(*this) != typeid(other)) { // Different types, not equal
        return FALSE;
    }
    const MeasureUnit &rhs = static_cast<const MeasureUnit&>(other);
    return (
            fTypeId == rhs.fTypeId
            && fSubTypeId == rhs.fSubTypeId
            && uprv_strcmp(fCurrency, rhs.fCurrency) == 0);
}

int32_t MeasureUnit::getIndex() const {
    return gIndexes[fTypeId] + fSubTypeId;
}

int32_t MeasureUnit::getAvailable(
        MeasureUnit *dest,
        int32_t destCapacity,
        UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) {
        return 0;
    }
    if (destCapacity < UPRV_LENGTHOF(gSubTypes)) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return UPRV_LENGTHOF(gSubTypes);
    }
    int32_t idx = 0;
    for (int32_t typeIdx = 0; typeIdx < UPRV_LENGTHOF(gTypes); ++typeIdx) {
        int32_t len = gOffsets[typeIdx + 1] - gOffsets[typeIdx];
        for (int32_t subTypeIdx = 0; subTypeIdx < len; ++subTypeIdx) {
            dest[idx].setTo(typeIdx, subTypeIdx);
            ++idx;
        }
    }
    U_ASSERT(idx == UPRV_LENGTHOF(gSubTypes));
    return UPRV_LENGTHOF(gSubTypes);
}

int32_t MeasureUnit::getAvailable(
        const char *type,
        MeasureUnit *dest,
        int32_t destCapacity,
        UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) {
        return 0;
    }
    int32_t typeIdx = binarySearch(gTypes, 0, UPRV_LENGTHOF(gTypes), type);
    if (typeIdx == -1) {
        return 0;
    }
    int32_t len = gOffsets[typeIdx + 1] - gOffsets[typeIdx];
    if (destCapacity < len) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return len;
    }
    for (int subTypeIdx = 0; subTypeIdx < len; ++subTypeIdx) {
        dest[subTypeIdx].setTo(typeIdx, subTypeIdx);
    }
    return len;
}

StringEnumeration* MeasureUnit::getAvailableTypes(UErrorCode &errorCode) {
    UEnumeration *uenum = uenum_openCharStringsEnumeration(
            gTypes, UPRV_LENGTHOF(gTypes), &errorCode);
    if (U_FAILURE(errorCode)) {
        uenum_close(uenum);
        return NULL;
    }
    StringEnumeration *result = new UStringEnumeration(uenum);
    if (result == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        uenum_close(uenum);
        return NULL;
    }
    return result;
}

int32_t MeasureUnit::getIndexCount() {
    return gIndexes[UPRV_LENGTHOF(gIndexes) - 1];
}

int32_t MeasureUnit::internalGetIndexForTypeAndSubtype(const char *type, const char *subtype) {
    int32_t t = binarySearch(gTypes, 0, UPRV_LENGTHOF(gTypes), type);
    if (t < 0) {
        return t;
    }
    int32_t st = binarySearch(gSubTypes, gOffsets[t], gOffsets[t + 1], subtype);
    if (st < 0) {
        return st;
    }
    return gIndexes[t] + st - gOffsets[t];
}

MeasureUnit MeasureUnit::resolveUnitPerUnit(
        const MeasureUnit &unit, const MeasureUnit &perUnit, bool* isResolved) {
    int32_t unitOffset = unit.getOffset();
    int32_t perUnitOffset = perUnit.getOffset();

    // binary search for (unitOffset, perUnitOffset)
    int32_t start = 0;
    int32_t end = UPRV_LENGTHOF(unitPerUnitToSingleUnit);
    while (start < end) {
        int32_t mid = (start + end) / 2;
        int32_t *midRow = unitPerUnitToSingleUnit[mid];
        if (unitOffset < midRow[0]) {
            end = mid;
        } else if (unitOffset > midRow[0]) {
            start = mid + 1;
        } else if (perUnitOffset < midRow[1]) {
            end = mid;
        } else if (perUnitOffset > midRow[1]) {
            start = mid + 1;
        } else {
            // We found a resolution for our unit / per-unit combo
            // return it.
            *isResolved = true;
            return MeasureUnit(midRow[2], midRow[3]);
        }
    }

    *isResolved = false;
    return MeasureUnit();
}

MeasureUnit *MeasureUnit::create(int typeId, int subTypeId, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    MeasureUnit *result = new MeasureUnit(typeId, subTypeId);
    if (result == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return result;
}

void MeasureUnit::initTime(const char *timeId) {
    int32_t result = binarySearch(gTypes, 0, UPRV_LENGTHOF(gTypes), "duration");
    U_ASSERT(result != -1);
    fTypeId = result;
    result = binarySearch(gSubTypes, gOffsets[fTypeId], gOffsets[fTypeId + 1], timeId);
    U_ASSERT(result != -1);
    fSubTypeId = result - gOffsets[fTypeId]; 
}

void MeasureUnit::initCurrency(const char *isoCurrency) {
    int32_t result = binarySearch(gTypes, 0, UPRV_LENGTHOF(gTypes), "currency");
    U_ASSERT(result != -1);
    fTypeId = result;
    result = binarySearch(
            gSubTypes, gOffsets[fTypeId], gOffsets[fTypeId + 1], isoCurrency);
    if (result != -1) {
        fSubTypeId = result - gOffsets[fTypeId];
    } else {
        uprv_strncpy(fCurrency, isoCurrency, UPRV_LENGTHOF(fCurrency));
        fCurrency[3] = 0;
    }
}

void MeasureUnit::initNoUnit(const char *subtype) {
    int32_t result = binarySearch(gTypes, 0, UPRV_LENGTHOF(gTypes), "none");
    U_ASSERT(result != -1);
    fTypeId = result;
    result = binarySearch(gSubTypes, gOffsets[fTypeId], gOffsets[fTypeId + 1], subtype);
    U_ASSERT(result != -1);
    fSubTypeId = result - gOffsets[fTypeId];
}

void MeasureUnit::setTo(int32_t typeId, int32_t subTypeId) {
    fTypeId = typeId;
    fSubTypeId = subTypeId;
    fCurrency[0] = 0;
}

int32_t MeasureUnit::getOffset() const {
    return gOffsets[fTypeId] + fSubTypeId;
}

U_NAMESPACE_END

#endif /* !UNCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2004-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: April 26, 2004
* Since: ICU 3.0
**********************************************************************
*/
  // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(Measure)

Measure::Measure() {}

Measure::Measure(const Formattable& _number, MeasureUnit* adoptedUnit,
                 UErrorCode& ec) :
    number(_number), unit(adoptedUnit) {
    if (U_SUCCESS(ec) &&
        (!number.isNumeric() || adoptedUnit == 0)) {
        ec = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

Measure::Measure(const Measure& other) :
    UObject(other), unit(0) {
    *this = other;
}

Measure& Measure::operator=(const Measure& other) {
    if (this != &other) {
        delete unit;
        number = other.number;
        unit = other.unit->clone();
    }
    return *this;
}

Measure *Measure::clone() const {
    return new Measure(*this);
}

Measure::~Measure() {
    delete unit;
}

UBool Measure::operator==(const UObject& other) const {
    if (this == &other) {  // Same object, equal
        return TRUE;
    }
    if (typeid(*this) != typeid(other)) { // Different types, not equal
        return FALSE;
    }
    const Measure &m = static_cast<const Measure&>(other);
    return number == m.number &&
        ((unit == NULL) == (m.unit == NULL)) &&
        (unit == NULL || *unit == *m.unit);
}

U_NAMESPACE_END

#endif // !UCONFIG_NO_FORMATTING
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/********************************************************************
 * COPYRIGHT:
 * Copyright (c) 1997-2015, International Business Machines Corporation and
 * others. All Rights Reserved.
 ********************************************************************
 *
 * File MSGFMT.CPP
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   02/19/97    aliu        Converted from java.
 *   03/20/97    helena      Finished first cut of implementation.
 *   04/10/97    aliu        Made to work on AIX.  Added stoi to replace wtoi.
 *   06/11/97    helena      Fixed addPattern to take the pattern correctly.
 *   06/17/97    helena      Fixed the getPattern to return the correct pattern.
 *   07/09/97    helena      Made ParsePosition into a class.
 *   02/22/99    stephen     Removed character literals for EBCDIC safety
 *   11/01/09    kirtig      Added SelectFormat
 ********************************************************************/



#if !UCONFIG_NO_FORMATTING









// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2015, International Business Machines Corporation and others.
* All Rights Reserved.
*******************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API

/**
 * \file
 * \brief C++ API: Rule Based Number Format
 */

/**
 * \def U_HAVE_RBNF
 * This will be 0 if RBNF support is not included in ICU
 * and 1 if it is.
 *
 * @stable ICU 2.4
 */
#if UCONFIG_NO_FORMATTING
#define U_HAVE_RBNF 0
#else
#define U_HAVE_RBNF 1







// #include "unicode/brkiter.h"


U_NAMESPACE_BEGIN

class NFRule;
class NFRuleSet;
class LocalizationInfo;
class PluralFormat;
class RuleBasedCollator;

/**
 * Tags for the predefined rulesets.
 *
 * @stable ICU 2.2
 */
enum URBNFRuleSetTag {
    URBNF_SPELLOUT,
    URBNF_ORDINAL,
    URBNF_DURATION,
    URBNF_NUMBERING_SYSTEM,
#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal URBNFRuleSetTag value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    URBNF_COUNT
#endif  // U_HIDE_DEPRECATED_API
};

/**
 * The RuleBasedNumberFormat class formats numbers according to a set of rules. This number formatter is
 * typically used for spelling out numeric values in words (e.g., 25,3476 as
 * &quot;twenty-five thousand three hundred seventy-six&quot; or &quot;vingt-cinq mille trois
 * cents soixante-seize&quot; or
 * &quot;f&uuml;nfundzwanzigtausenddreihundertsechsundsiebzig&quot;), but can also be used for
 * other complicated formatting tasks, such as formatting a number of seconds as hours,
 * minutes and seconds (e.g., 3,730 as &quot;1:02:10&quot;).
 *
 * <p>The resources contain three predefined formatters for each locale: spellout, which
 * spells out a value in words (123 is &quot;one hundred twenty-three&quot;); ordinal, which
 * appends an ordinal suffix to the end of a numeral (123 is &quot;123rd&quot;); and
 * duration, which shows a duration in seconds as hours, minutes, and seconds (123 is
 * &quot;2:03&quot;).&nbsp; The client can also define more specialized <tt>RuleBasedNumberFormat</tt>s
 * by supplying programmer-defined rule sets.</p>
 *
 * <p>The behavior of a <tt>RuleBasedNumberFormat</tt> is specified by a textual description
 * that is either passed to the constructor as a <tt>String</tt> or loaded from a resource
 * bundle. In its simplest form, the description consists of a semicolon-delimited list of <em>rules.</em>
 * Each rule has a string of output text and a value or range of values it is applicable to.
 * In a typical spellout rule set, the first twenty rules are the words for the numbers from
 * 0 to 19:</p>
 *
 * <pre>zero; one; two; three; four; five; six; seven; eight; nine;
 * ten; eleven; twelve; thirteen; fourteen; fifteen; sixteen; seventeen; eighteen; nineteen;</pre>
 *
 * <p>For larger numbers, we can use the preceding set of rules to format the ones place, and
 * we only have to supply the words for the multiples of 10:</p>
 *
 * <pre> 20: twenty[-&gt;&gt;];
 * 30: thirty[-&gt;&gt;];
 * 40: forty[-&gt;&gt;];
 * 50: fifty[-&gt;&gt;];
 * 60: sixty[-&gt;&gt;];
 * 70: seventy[-&gt;&gt;];
 * 80: eighty[-&gt;&gt;];
 * 90: ninety[-&gt;&gt;];</pre>
 *
 * <p>In these rules, the <em>base value</em> is spelled out explicitly and set off from the
 * rule's output text with a colon. The rules are in a sorted list, and a rule is applicable
 * to all numbers from its own base value to one less than the next rule's base value. The
 * &quot;&gt;&gt;&quot; token is called a <em>substitution</em> and tells the fomatter to
 * isolate the number's ones digit, format it using this same set of rules, and place the
 * result at the position of the &quot;&gt;&gt;&quot; token. Text in brackets is omitted if
 * the number being formatted is an even multiple of 10 (the hyphen is a literal hyphen; 24
 * is &quot;twenty-four,&quot; not &quot;twenty four&quot;).</p>
 *
 * <p>For even larger numbers, we can actually look up several parts of the number in the
 * list:</p>
 *
 * <pre>100: &lt;&lt; hundred[ &gt;&gt;];</pre>
 *
 * <p>The &quot;&lt;&lt;&quot; represents a new kind of substitution. The &lt;&lt; isolates
 * the hundreds digit (and any digits to its left), formats it using this same rule set, and
 * places the result where the &quot;&lt;&lt;&quot; was. Notice also that the meaning of
 * &gt;&gt; has changed: it now refers to both the tens and the ones digits. The meaning of
 * both substitutions depends on the rule's base value. The base value determines the rule's <em>divisor,</em>
 * which is the highest power of 10 that is less than or equal to the base value (the user
 * can change this). To fill in the substitutions, the formatter divides the number being
 * formatted by the divisor. The integral quotient is used to fill in the &lt;&lt;
 * substitution, and the remainder is used to fill in the &gt;&gt; substitution. The meaning
 * of the brackets changes similarly: text in brackets is omitted if the value being
 * formatted is an even multiple of the rule's divisor. The rules are applied recursively, so
 * if a substitution is filled in with text that includes another substitution, that
 * substitution is also filled in.</p>
 *
 * <p>This rule covers values up to 999, at which point we add another rule:</p>
 *
 * <pre>1000: &lt;&lt; thousand[ &gt;&gt;];</pre>
 *
 * <p>Again, the meanings of the brackets and substitution tokens shift because the rule's
 * base value is a higher power of 10, changing the rule's divisor. This rule can actually be
 * used all the way up to 999,999. This allows us to finish out the rules as follows:</p>
 *
 * <pre> 1,000,000: &lt;&lt; million[ &gt;&gt;];
 * 1,000,000,000: &lt;&lt; billion[ &gt;&gt;];
 * 1,000,000,000,000: &lt;&lt; trillion[ &gt;&gt;];
 * 1,000,000,000,000,000: OUT OF RANGE!;</pre>
 *
 * <p>Commas, periods, and spaces can be used in the base values to improve legibility and
 * are ignored by the rule parser. The last rule in the list is customarily treated as an
 * &quot;overflow rule,&quot; applying to everything from its base value on up, and often (as
 * in this example) being used to print out an error message or default representation.
 * Notice also that the size of the major groupings in large numbers is controlled by the
 * spacing of the rules: because in English we group numbers by thousand, the higher rules
 * are separated from each other by a factor of 1,000.</p>
 *
 * <p>To see how these rules actually work in practice, consider the following example:
 * Formatting 25,430 with this rule set would work like this:</p>
 *
 * <table border="0" width="100%">
 *   <tr>
 *     <td><strong>&lt;&lt; thousand &gt;&gt;</strong></td>
 *     <td>[the rule whose base value is 1,000 is applicable to 25,340]</td>
 *   </tr>
 *   <tr>
 *     <td><strong>twenty-&gt;&gt;</strong> thousand &gt;&gt;</td>
 *     <td>[25,340 over 1,000 is 25. The rule for 20 applies.]</td>
 *   </tr>
 *   <tr>
 *     <td>twenty-<strong>five</strong> thousand &gt;&gt;</td>
 *     <td>[25 mod 10 is 5. The rule for 5 is &quot;five.&quot;</td>
 *   </tr>
 *   <tr>
 *     <td>twenty-five thousand <strong>&lt;&lt; hundred &gt;&gt;</strong></td>
 *     <td>[25,340 mod 1,000 is 340. The rule for 100 applies.]</td>
 *   </tr>
 *   <tr>
 *     <td>twenty-five thousand <strong>three</strong> hundred &gt;&gt;</td>
 *     <td>[340 over 100 is 3. The rule for 3 is &quot;three.&quot;]</td>
 *   </tr>
 *   <tr>
 *     <td>twenty-five thousand three hundred <strong>forty</strong></td>
 *     <td>[340 mod 100 is 40. The rule for 40 applies. Since 40 divides
 *     evenly by 10, the hyphen and substitution in the brackets are omitted.]</td>
 *   </tr>
 * </table>
 *
 * <p>The above syntax suffices only to format positive integers. To format negative numbers,
 * we add a special rule:</p>
 *
 * <pre>-x: minus &gt;&gt;;</pre>
 *
 * <p>This is called a <em>negative-number rule,</em> and is identified by &quot;-x&quot;
 * where the base value would be. This rule is used to format all negative numbers. the
 * &gt;&gt; token here means &quot;find the number's absolute value, format it with these
 * rules, and put the result here.&quot;</p>
 *
 * <p>We also add a special rule called a <em>fraction rule </em>for numbers with fractional
 * parts:</p>
 *
 * <pre>x.x: &lt;&lt; point &gt;&gt;;</pre>
 *
 * <p>This rule is used for all positive non-integers (negative non-integers pass through the
 * negative-number rule first and then through this rule). Here, the &lt;&lt; token refers to
 * the number's integral part, and the &gt;&gt; to the number's fractional part. The
 * fractional part is formatted as a series of single-digit numbers (e.g., 123.456 would be
 * formatted as &quot;one hundred twenty-three point four five six&quot;).</p>
 *
 * <p>To see how this rule syntax is applied to various languages, examine the resource data.</p>
 *
 * <p>There is actually much more flexibility built into the rule language than the
 * description above shows. A formatter may own multiple rule sets, which can be selected by
 * the caller, and which can use each other to fill in their substitutions. Substitutions can
 * also be filled in with digits, using a DecimalFormat object. There is syntax that can be
 * used to alter a rule's divisor in various ways. And there is provision for much more
 * flexible fraction handling. A complete description of the rule syntax follows:</p>
 *
 * <hr>
 *
 * <p>The description of a <tt>RuleBasedNumberFormat</tt>'s behavior consists of one or more <em>rule
 * sets.</em> Each rule set consists of a name, a colon, and a list of <em>rules.</em> A rule
 * set name must begin with a % sign. Rule sets with names that begin with a single % sign
 * are <em>public:</em> the caller can specify that they be used to format and parse numbers.
 * Rule sets with names that begin with %% are <em>private:</em> they exist only for the use
 * of other rule sets. If a formatter only has one rule set, the name may be omitted.</p>
 *
 * <p>The user can also specify a special &quot;rule set&quot; named <tt>%%lenient-parse</tt>.
 * The body of <tt>%%lenient-parse</tt> isn't a set of number-formatting rules, but a <tt>RuleBasedCollator</tt>
 * description which is used to define equivalences for lenient parsing. For more information
 * on the syntax, see <tt>RuleBasedCollator</tt>. For more information on lenient parsing,
 * see <tt>setLenientParse()</tt>.  <em>Note:</em> symbols that have syntactic meaning
 * in collation rules, such as '&amp;', have no particular meaning when appearing outside
 * of the <tt>lenient-parse</tt> rule set.</p>
 *
 * <p>The body of a rule set consists of an ordered, semicolon-delimited list of <em>rules.</em>
 * Internally, every rule has a base value, a divisor, rule text, and zero, one, or two <em>substitutions.</em>
 * These parameters are controlled by the description syntax, which consists of a <em>rule
 * descriptor,</em> a colon, and a <em>rule body.</em></p>
 *
 * <p>A rule descriptor can take one of the following forms (text in <em>italics</em> is the
 * name of a token):</p>
 *
 * <table border="0" width="100%">
 *   <tr>
 *     <td><em>bv</em>:</td>
 *     <td><em>bv</em> specifies the rule's base value. <em>bv</em> is a decimal
 *     number expressed using ASCII digits. <em>bv</em> may contain spaces, period, and commas,
 *     which are ignored. The rule's divisor is the highest power of 10 less than or equal to
 *     the base value.</td>
 *   </tr>
 *   <tr>
 *     <td><em>bv</em>/<em>rad</em>:</td>
 *     <td><em>bv</em> specifies the rule's base value. The rule's divisor is the
 *     highest power of <em>rad</em> less than or equal to the base value.</td>
 *   </tr>
 *   <tr>
 *     <td><em>bv</em>&gt;:</td>
 *     <td><em>bv</em> specifies the rule's base value. To calculate the divisor,
 *     let the radix be 10, and the exponent be the highest exponent of the radix that yields a
 *     result less than or equal to the base value. Every &gt; character after the base value
 *     decreases the exponent by 1. If the exponent is positive or 0, the divisor is the radix
 *     raised to the power of the exponent; otherwise, the divisor is 1.</td>
 *   </tr>
 *   <tr>
 *     <td><em>bv</em>/<em>rad</em>&gt;:</td>
 *     <td><em>bv</em> specifies the rule's base value. To calculate the divisor,
 *     let the radix be <em>rad</em>, and the exponent be the highest exponent of the radix that
 *     yields a result less than or equal to the base value. Every &gt; character after the radix
 *     decreases the exponent by 1. If the exponent is positive or 0, the divisor is the radix
 *     raised to the power of the exponent; otherwise, the divisor is 1.</td>
 *   </tr>
 *   <tr>
 *     <td>-x:</td>
 *     <td>The rule is a negative-number rule.</td>
 *   </tr>
 *   <tr>
 *     <td>x.x:</td>
 *     <td>The rule is an <em>improper fraction rule</em>. If the full stop in
 *     the middle of the rule name is replaced with the decimal point
 *     that is used in the language or DecimalFormatSymbols, then that rule will
 *     have precedence when formatting and parsing this rule. For example, some
 *     languages use the comma, and can thus be written as x,x instead. For example,
 *     you can use "x.x: &lt;&lt; point &gt;&gt;;x,x: &lt;&lt; comma &gt;&gt;;" to
 *     handle the decimal point that matches the language's natural spelling of
 *     the punctuation of either the full stop or comma.</td>
 *   </tr>
 *   <tr>
 *     <td>0.x:</td>
 *     <td>The rule is a <em>proper fraction rule</em>. If the full stop in
 *     the middle of the rule name is replaced with the decimal point
 *     that is used in the language or DecimalFormatSymbols, then that rule will
 *     have precedence when formatting and parsing this rule. For example, some
 *     languages use the comma, and can thus be written as 0,x instead. For example,
 *     you can use "0.x: point &gt;&gt;;0,x: comma &gt;&gt;;" to
 *     handle the decimal point that matches the language's natural spelling of
 *     the punctuation of either the full stop or comma.</td>
 *   </tr>
 *   <tr>
 *     <td>x.0:</td>
 *     <td>The rule is a <em>master rule</em>. If the full stop in
 *     the middle of the rule name is replaced with the decimal point
 *     that is used in the language or DecimalFormatSymbols, then that rule will
 *     have precedence when formatting and parsing this rule. For example, some
 *     languages use the comma, and can thus be written as x,0 instead. For example,
 *     you can use "x.0: &lt;&lt; point;x,0: &lt;&lt; comma;" to
 *     handle the decimal point that matches the language's natural spelling of
 *     the punctuation of either the full stop or comma.</td>
 *   </tr>
 *   <tr>
 *     <td>Inf:</td>
 *     <td>The rule for infinity.</td>
 *   </tr>
 *   <tr>
 *     <td>NaN:</td>
 *     <td>The rule for an IEEE 754 NaN (not a number).</td>
 *   </tr>
 *   <tr>
 *     <td><em>nothing</em></td>
 *     <td>If the rule's rule descriptor is left out, the base value is one plus the
 *     preceding rule's base value (or zero if this is the first rule in the list) in a normal
 *     rule set.&nbsp; In a fraction rule set, the base value is the same as the preceding rule's
 *     base value.</td>
 *   </tr>
 * </table>
 *
 * <p>A rule set may be either a regular rule set or a <em>fraction rule set,</em> depending
 * on whether it is used to format a number's integral part (or the whole number) or a
 * number's fractional part. Using a rule set to format a rule's fractional part makes it a
 * fraction rule set.</p>
 *
 * <p>Which rule is used to format a number is defined according to one of the following
 * algorithms: If the rule set is a regular rule set, do the following:
 *
 * <ul>
 *   <li>If the rule set includes a master rule (and the number was passed in as a <tt>double</tt>),
 *     use the master rule.&nbsp; (If the number being formatted was passed in as a <tt>long</tt>,
 *     the master rule is ignored.)</li>
 *   <li>If the number is negative, use the negative-number rule.</li>
 *   <li>If the number has a fractional part and is greater than 1, use the improper fraction
 *     rule.</li>
 *   <li>If the number has a fractional part and is between 0 and 1, use the proper fraction
 *     rule.</li>
 *   <li>Binary-search the rule list for the rule with the highest base value less than or equal
 *     to the number. If that rule has two substitutions, its base value is not an even multiple
 *     of its divisor, and the number <em>is</em> an even multiple of the rule's divisor, use the
 *     rule that precedes it in the rule list. Otherwise, use the rule itself.</li>
 * </ul>
 *
 * <p>If the rule set is a fraction rule set, do the following:
 *
 * <ul>
 *   <li>Ignore negative-number and fraction rules.</li>
 *   <li>For each rule in the list, multiply the number being formatted (which will always be
 *     between 0 and 1) by the rule's base value. Keep track of the distance between the result
 *     the nearest integer.</li>
 *   <li>Use the rule that produced the result closest to zero in the above calculation. In the
 *     event of a tie or a direct hit, use the first matching rule encountered. (The idea here is
 *     to try each rule's base value as a possible denominator of a fraction. Whichever
 *     denominator produces the fraction closest in value to the number being formatted wins.) If
 *     the rule following the matching rule has the same base value, use it if the numerator of
 *     the fraction is anything other than 1; if the numerator is 1, use the original matching
 *     rule. (This is to allow singular and plural forms of the rule text without a lot of extra
 *     hassle.)</li>
 * </ul>
 *
 * <p>A rule's body consists of a string of characters terminated by a semicolon. The rule
 * may include zero, one, or two <em>substitution tokens,</em> and a range of text in
 * brackets. The brackets denote optional text (and may also include one or both
 * substitutions). The exact meanings of the substitution tokens, and under what conditions
 * optional text is omitted, depend on the syntax of the substitution token and the context.
 * The rest of the text in a rule body is literal text that is output when the rule matches
 * the number being formatted.</p>
 *
 * <p>A substitution token begins and ends with a <em>token character.</em> The token
 * character and the context together specify a mathematical operation to be performed on the
 * number being formatted. An optional <em>substitution descriptor </em>specifies how the
 * value resulting from that operation is used to fill in the substitution. The position of
 * the substitution token in the rule body specifies the location of the resultant text in
 * the original rule text.</p>
 *
 * <p>The meanings of the substitution token characters are as follows:</p>
 *
 * <table border="0" width="100%">
 *   <tr>
 *     <td>&gt;&gt;</td>
 *     <td>in normal rule</td>
 *     <td>Divide the number by the rule's divisor and format the remainder</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in negative-number rule</td>
 *     <td>Find the absolute value of the number and format the result</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in fraction or master rule</td>
 *     <td>Isolate the number's fractional part and format it.</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in rule in fraction rule set</td>
 *     <td>Not allowed.</td>
 *   </tr>
 *   <tr>
 *     <td>&gt;&gt;&gt;</td>
 *     <td>in normal rule</td>
 *     <td>Divide the number by the rule's divisor and format the remainder,
 *       but bypass the normal rule-selection process and just use the
 *       rule that precedes this one in this rule list.</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in all other rules</td>
 *     <td>Not allowed.</td>
 *   </tr>
 *   <tr>
 *     <td>&lt;&lt;</td>
 *     <td>in normal rule</td>
 *     <td>Divide the number by the rule's divisor and format the quotient</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in negative-number rule</td>
 *     <td>Not allowed.</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in fraction or master rule</td>
 *     <td>Isolate the number's integral part and format it.</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in rule in fraction rule set</td>
 *     <td>Multiply the number by the rule's base value and format the result.</td>
 *   </tr>
 *   <tr>
 *     <td>==</td>
 *     <td>in all rule sets</td>
 *     <td>Format the number unchanged</td>
 *   </tr>
 *   <tr>
 *     <td>[]</td>
 *     <td>in normal rule</td>
 *     <td>Omit the optional text if the number is an even multiple of the rule's divisor</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in negative-number rule</td>
 *     <td>Not allowed.</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in improper-fraction rule</td>
 *     <td>Omit the optional text if the number is between 0 and 1 (same as specifying both an
 *     x.x rule and a 0.x rule)</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in master rule</td>
 *     <td>Omit the optional text if the number is an integer (same as specifying both an x.x
 *     rule and an x.0 rule)</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in proper-fraction rule</td>
 *     <td>Not allowed.</td>
 *   </tr>
 *   <tr>
 *     <td></td>
 *     <td>in rule in fraction rule set</td>
 *     <td>Omit the optional text if multiplying the number by the rule's base value yields 1.</td>
 *   </tr>
 *   <tr>
 *     <td width="37">$(cardinal,<i>plural syntax</i>)$</td>
 *     <td width="23"></td>
 *     <td width="165" valign="top">in all rule sets</td>
 *     <td>This provides the ability to choose a word based on the number divided by the radix to the power of the
 *     exponent of the base value for the specified locale, which is normally equivalent to the &lt;&lt; value.
 *     This uses the cardinal plural rules from PluralFormat. All strings used in the plural format are treated
 *     as the same base value for parsing.</td>
 *   </tr>
 *   <tr>
 *     <td width="37">$(ordinal,<i>plural syntax</i>)$</td>
 *     <td width="23"></td>
 *     <td width="165" valign="top">in all rule sets</td>
 *     <td>This provides the ability to choose a word based on the number divided by the radix to the power of the
 *     exponent of the base value for the specified locale, which is normally equivalent to the &lt;&lt; value.
 *     This uses the ordinal plural rules from PluralFormat. All strings used in the plural format are treated
 *     as the same base value for parsing.</td>
 *   </tr>
 * </table>
 *
 * <p>The substitution descriptor (i.e., the text between the token characters) may take one
 * of three forms:</p>
 *
 * <table border="0" width="100%">
 *   <tr>
 *     <td>a rule set name</td>
 *     <td>Perform the mathematical operation on the number, and format the result using the
 *     named rule set.</td>
 *   </tr>
 *   <tr>
 *     <td>a DecimalFormat pattern</td>
 *     <td>Perform the mathematical operation on the number, and format the result using a
 *     DecimalFormat with the specified pattern.&nbsp; The pattern must begin with 0 or #.</td>
 *   </tr>
 *   <tr>
 *     <td>nothing</td>
 *     <td>Perform the mathematical operation on the number, and format the result using the rule
 *     set containing the current rule, except:
 *     <ul>
 *       <li>You can't have an empty substitution descriptor with a == substitution.</li>
 *       <li>If you omit the substitution descriptor in a &gt;&gt; substitution in a fraction rule,
 *         format the result one digit at a time using the rule set containing the current rule.</li>
 *       <li>If you omit the substitution descriptor in a &lt;&lt; substitution in a rule in a
 *         fraction rule set, format the result using the default rule set for this formatter.</li>
 *     </ul>
 *     </td>
 *   </tr>
 * </table>
 *
 * <p>Whitespace is ignored between a rule set name and a rule set body, between a rule
 * descriptor and a rule body, or between rules. If a rule body begins with an apostrophe,
 * the apostrophe is ignored, but all text after it becomes significant (this is how you can
 * have a rule's rule text begin with whitespace). There is no escape function: the semicolon
 * is not allowed in rule set names or in rule text, and the colon is not allowed in rule set
 * names. The characters beginning a substitution token are always treated as the beginning
 * of a substitution token.</p>
 *
 * <p>See the resource data and the demo program for annotated examples of real rule sets
 * using these features.</p>
 *
 * <p><em>User subclasses are not supported.</em> While clients may write
 * subclasses, such code will not necessarily work and will not be
 * guaranteed to work stably from release to release.
 *
 * <p><b>Localizations</b></p>
 * <p>Constructors are available that allow the specification of localizations for the
 * public rule sets (and also allow more control over what public rule sets are available).
 * Localization data is represented as a textual description.  The description represents
 * an array of arrays of string.  The first element is an array of the public rule set names,
 * each of these must be one of the public rule set names that appear in the rules.  Only
 * names in this array will be treated as public rule set names by the API.  Each subsequent
 * element is an array of localizations of these names.  The first element of one of these
 * subarrays is the locale name, and the remaining elements are localizations of the
 * public rule set names, in the same order as they were listed in the first arrray.</p>
 * <p>In the syntax, angle brackets '<', '>' are used to delimit the arrays, and comma ',' is used
 * to separate elements of an array.  Whitespace is ignored, unless quoted.</p>
 * <p>For example:<pre>
 * < < %foo, %bar, %baz >,
 *   < en, Foo, Bar, Baz >,
 *   < fr, 'le Foo', 'le Bar', 'le Baz' >
 *   < zh, \\u7532, \\u4e59, \\u4e19 > >
 * </pre></p>
 * @author Richard Gillam
 * @see NumberFormat
 * @see DecimalFormat
 * @see PluralFormat
 * @see PluralRules
 * @stable ICU 2.0
 */
class U_I18N_API RuleBasedNumberFormat : public NumberFormat {
public:

  //-----------------------------------------------------------------------
  // constructors
  //-----------------------------------------------------------------------

    /**
     * Creates a RuleBasedNumberFormat that behaves according to the description
     * passed in.  The formatter uses the default locale.
     * @param rules A description of the formatter's desired behavior.
     * See the class documentation for a complete explanation of the description
     * syntax.
     * @param perror The parse error if an error was encountered.
     * @param status The status indicating whether the constructor succeeded.
     * @stable ICU 3.2
     */
    RuleBasedNumberFormat(const UnicodeString& rules, UParseError& perror, UErrorCode& status);

    /**
     * Creates a RuleBasedNumberFormat that behaves according to the description
     * passed in.  The formatter uses the default locale.
     * <p>
     * The localizations data provides information about the public
     * rule sets and their localized display names for different
     * locales. The first element in the list is an array of the names
     * of the public rule sets.  The first element in this array is
     * the initial default ruleset.  The remaining elements in the
     * list are arrays of localizations of the names of the public
     * rule sets.  Each of these is one longer than the initial array,
     * with the first String being the ULocale ID, and the remaining
     * Strings being the localizations of the rule set names, in the
     * same order as the initial array.  Arrays are NULL-terminated.
     * @param rules A description of the formatter's desired behavior.
     * See the class documentation for a complete explanation of the description
     * syntax.
     * @param localizations the localization information.
     * names in the description.  These will be copied by the constructor.
     * @param perror The parse error if an error was encountered.
     * @param status The status indicating whether the constructor succeeded.
     * @stable ICU 3.2
     */
    RuleBasedNumberFormat(const UnicodeString& rules, const UnicodeString& localizations,
                        UParseError& perror, UErrorCode& status);

  /**
   * Creates a RuleBasedNumberFormat that behaves according to the rules
   * passed in.  The formatter uses the specified locale to determine the
   * characters to use when formatting numerals, and to define equivalences
   * for lenient parsing.
   * @param rules The formatter rules.
   * See the class documentation for a complete explanation of the rule
   * syntax.
   * @param locale A locale that governs which characters are used for
   * formatting values in numerals and which characters are equivalent in
   * lenient parsing.
   * @param perror The parse error if an error was encountered.
   * @param status The status indicating whether the constructor succeeded.
   * @stable ICU 2.0
   */
  RuleBasedNumberFormat(const UnicodeString& rules, const Locale& locale,
                        UParseError& perror, UErrorCode& status);

    /**
     * Creates a RuleBasedNumberFormat that behaves according to the description
     * passed in.  The formatter uses the default locale.
     * <p>
     * The localizations data provides information about the public
     * rule sets and their localized display names for different
     * locales. The first element in the list is an array of the names
     * of the public rule sets.  The first element in this array is
     * the initial default ruleset.  The remaining elements in the
     * list are arrays of localizations of the names of the public
     * rule sets.  Each of these is one longer than the initial array,
     * with the first String being the ULocale ID, and the remaining
     * Strings being the localizations of the rule set names, in the
     * same order as the initial array.  Arrays are NULL-terminated.
     * @param rules A description of the formatter's desired behavior.
     * See the class documentation for a complete explanation of the description
     * syntax.
     * @param localizations a list of localizations for the rule set
     * names in the description.  These will be copied by the constructor.
     * @param locale A locale that governs which characters are used for
     * formatting values in numerals and which characters are equivalent in
     * lenient parsing.
     * @param perror The parse error if an error was encountered.
     * @param status The status indicating whether the constructor succeeded.
     * @stable ICU 3.2
     */
    RuleBasedNumberFormat(const UnicodeString& rules, const UnicodeString& localizations,
                        const Locale& locale, UParseError& perror, UErrorCode& status);

  /**
   * Creates a RuleBasedNumberFormat from a predefined ruleset.  The selector
   * code choosed among three possible predefined formats: spellout, ordinal,
   * and duration.
   * @param tag A selector code specifying which kind of formatter to create for that
   * locale.  There are four legal values: URBNF_SPELLOUT, which creates a formatter that
   * spells out a value in words in the desired language, URBNF_ORDINAL, which attaches
   * an ordinal suffix from the desired language to the end of a number (e.g. "123rd"),
   * URBNF_DURATION, which formats a duration in seconds as hours, minutes, and seconds always rounding down,
   * and URBNF_NUMBERING_SYSTEM, which is used to invoke rules for alternate numbering
   * systems such as the Hebrew numbering system, or for Roman Numerals, etc.
   * @param locale The locale for the formatter.
   * @param status The status indicating whether the constructor succeeded.
   * @stable ICU 2.0
   */
  RuleBasedNumberFormat(URBNFRuleSetTag tag, const Locale& locale, UErrorCode& status);

  //-----------------------------------------------------------------------
  // boilerplate
  //-----------------------------------------------------------------------

  /**
   * Copy constructor
   * @param rhs    the object to be copied from.
   * @stable ICU 2.6
   */
  RuleBasedNumberFormat(const RuleBasedNumberFormat& rhs);

  /**
   * Assignment operator
   * @param rhs    the object to be copied from.
   * @stable ICU 2.6
   */
  RuleBasedNumberFormat& operator=(const RuleBasedNumberFormat& rhs);

  /**
   * Release memory allocated for a RuleBasedNumberFormat when you are finished with it.
   * @stable ICU 2.6
   */
  virtual ~RuleBasedNumberFormat();

  /**
   * Clone this object polymorphically.  The caller is responsible
   * for deleting the result when done.
   * @return  A copy of the object.
   * @stable ICU 2.6
   */
  virtual RuleBasedNumberFormat* clone() const;

  /**
   * Return true if the given Format objects are semantically equal.
   * Objects of different subclasses are considered unequal.
   * @param other    the object to be compared with.
   * @return        true if the given Format objects are semantically equal.
   * @stable ICU 2.6
   */
  virtual UBool operator==(const Format& other) const;

//-----------------------------------------------------------------------
// public API functions
//-----------------------------------------------------------------------

  /**
   * return the rules that were provided to the RuleBasedNumberFormat.
   * @return the result String that was passed in
   * @stable ICU 2.0
   */
  virtual UnicodeString getRules() const;

  /**
   * Return the number of public rule set names.
   * @return the number of public rule set names.
   * @stable ICU 2.0
   */
  virtual int32_t getNumberOfRuleSetNames() const;

  /**
   * Return the name of the index'th public ruleSet.  If index is not valid,
   * the function returns null.
   * @param index the index of the ruleset
   * @return the name of the index'th public ruleSet.
   * @stable ICU 2.0
   */
  virtual UnicodeString getRuleSetName(int32_t index) const;

  /**
   * Return the number of locales for which we have localized rule set display names.
   * @return the number of locales for which we have localized rule set display names.
   * @stable ICU 3.2
   */
  virtual int32_t getNumberOfRuleSetDisplayNameLocales(void) const;

  /**
   * Return the index'th display name locale.
   * @param index the index of the locale
   * @param status set to a failure code when this function fails
   * @return the locale
   * @see #getNumberOfRuleSetDisplayNameLocales
   * @stable ICU 3.2
   */
  virtual Locale getRuleSetDisplayNameLocale(int32_t index, UErrorCode& status) const;

    /**
     * Return the rule set display names for the provided locale.  These are in the same order
     * as those returned by getRuleSetName.  The locale is matched against the locales for
     * which there is display name data, using normal fallback rules.  If no locale matches,
     * the default display names are returned.  (These are the internal rule set names minus
     * the leading '%'.)
     * @param index the index of the rule set
     * @param locale the locale (returned by getRuleSetDisplayNameLocales) for which the localized
     * display name is desired
     * @return the display name for the given index, which might be bogus if there is an error
     * @see #getRuleSetName
     * @stable ICU 3.2
     */
  virtual UnicodeString getRuleSetDisplayName(int32_t index,
                          const Locale& locale = Locale::getDefault());

    /**
     * Return the rule set display name for the provided rule set and locale.
     * The locale is matched against the locales for which there is display name data, using
     * normal fallback rules.  If no locale matches, the default display name is returned.
     * @return the display name for the rule set
     * @stable ICU 3.2
     * @see #getRuleSetDisplayName
     */
  virtual UnicodeString getRuleSetDisplayName(const UnicodeString& ruleSetName,
                          const Locale& locale = Locale::getDefault());


  using NumberFormat::format;

  /**
   * Formats the specified 32-bit number using the default ruleset.
   * @param number The number to format.
   * @param toAppendTo the string that will hold the (appended) result
   * @param pos the fieldposition
   * @return A textual representation of the number.
   * @stable ICU 2.0
   */
  virtual UnicodeString& format(int32_t number,
                                UnicodeString& toAppendTo,
                                FieldPosition& pos) const;

  /**
   * Formats the specified 64-bit number using the default ruleset.
   * @param number The number to format.
   * @param toAppendTo the string that will hold the (appended) result
   * @param pos the fieldposition
   * @return A textual representation of the number.
   * @stable ICU 2.1
   */
  virtual UnicodeString& format(int64_t number,
                                UnicodeString& toAppendTo,
                                FieldPosition& pos) const;
  /**
   * Formats the specified number using the default ruleset.
   * @param number The number to format.
   * @param toAppendTo the string that will hold the (appended) result
   * @param pos the fieldposition
   * @return A textual representation of the number.
   * @stable ICU 2.0
   */
  virtual UnicodeString& format(double number,
                                UnicodeString& toAppendTo,
                                FieldPosition& pos) const;

  /**
   * Formats the specified number using the named ruleset.
   * @param number The number to format.
   * @param ruleSetName The name of the rule set to format the number with.
   * This must be the name of a valid public rule set for this formatter.
   * @param toAppendTo the string that will hold the (appended) result
   * @param pos the fieldposition
   * @param status the status
   * @return A textual representation of the number.
   * @stable ICU 2.0
   */
  virtual UnicodeString& format(int32_t number,
                                const UnicodeString& ruleSetName,
                                UnicodeString& toAppendTo,
                                FieldPosition& pos,
                                UErrorCode& status) const;
  /**
   * Formats the specified 64-bit number using the named ruleset.
   * @param number The number to format.
   * @param ruleSetName The name of the rule set to format the number with.
   * This must be the name of a valid public rule set for this formatter.
   * @param toAppendTo the string that will hold the (appended) result
   * @param pos the fieldposition
   * @param status the status
   * @return A textual representation of the number.
   * @stable ICU 2.1
   */
  virtual UnicodeString& format(int64_t number,
                                const UnicodeString& ruleSetName,
                                UnicodeString& toAppendTo,
                                FieldPosition& pos,
                                UErrorCode& status) const;
  /**
   * Formats the specified number using the named ruleset.
   * @param number The number to format.
   * @param ruleSetName The name of the rule set to format the number with.
   * This must be the name of a valid public rule set for this formatter.
   * @param toAppendTo the string that will hold the (appended) result
   * @param pos the fieldposition
   * @param status the status
   * @return A textual representation of the number.
   * @stable ICU 2.0
   */
  virtual UnicodeString& format(double number,
                                const UnicodeString& ruleSetName,
                                UnicodeString& toAppendTo,
                                FieldPosition& pos,
                                UErrorCode& status) const;

protected:
    /**
     * Format a decimal number.
     * The number is a DigitList wrapper onto a floating point decimal number.
     * The default implementation in NumberFormat converts the decimal number
     * to a double and formats that.  Subclasses of NumberFormat that want
     * to specifically handle big decimal numbers must override this method.
     * class DecimalFormat does so.
     *
     * @param number    The number, a DigitList format Decimal Floating Point.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    Output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @internal
     */
    virtual UnicodeString& format(const number::impl::DecimalQuantity &number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos,
                                  UErrorCode& status) const;
public:

  using NumberFormat::parse;

  /**
   * Parses the specfied string, beginning at the specified position, according
   * to this formatter's rules.  This will match the string against all of the
   * formatter's public rule sets and return the value corresponding to the longest
   * parseable substring.  This function's behavior is affected by the lenient
   * parse mode.
   * @param text The string to parse
   * @param result the result of the parse, either a double or a long.
   * @param parsePosition On entry, contains the position of the first character
   * in "text" to examine.  On exit, has been updated to contain the position
   * of the first character in "text" that wasn't consumed by the parse.
   * @see #setLenient
   * @stable ICU 2.0
   */
  virtual void parse(const UnicodeString& text,
                     Formattable& result,
                     ParsePosition& parsePosition) const;

#if !UCONFIG_NO_COLLATION

  /**
   * Turns lenient parse mode on and off.
   *
   * When in lenient parse mode, the formatter uses a Collator for parsing the text.
   * Only primary differences are treated as significant.  This means that case
   * differences, accent differences, alternate spellings of the same letter
   * (e.g., ae and a-umlaut in German), ignorable characters, etc. are ignored in
   * matching the text.  In many cases, numerals will be accepted in place of words
   * or phrases as well.
   *
   * For example, all of the following will correctly parse as 255 in English in
   * lenient-parse mode:
   * <br>"two hundred fifty-five"
   * <br>"two hundred fifty five"
   * <br>"TWO HUNDRED FIFTY-FIVE"
   * <br>"twohundredfiftyfive"
   * <br>"2 hundred fifty-5"
   *
   * The Collator used is determined by the locale that was
   * passed to this object on construction.  The description passed to this object
   * on construction may supply additional collation rules that are appended to the
   * end of the default collator for the locale, enabling additional equivalences
   * (such as adding more ignorable characters or permitting spelled-out version of
   * symbols; see the demo program for examples).
   *
   * It's important to emphasize that even strict parsing is relatively lenient: it
   * will accept some text that it won't produce as output.  In English, for example,
   * it will correctly parse "two hundred zero" and "fifteen hundred".
   *
   * @param enabled If true, turns lenient-parse mode on; if false, turns it off.
   * @see RuleBasedCollator
   * @stable ICU 2.0
   */
  virtual void setLenient(UBool enabled);

  /**
   * Returns true if lenient-parse mode is turned on.  Lenient parsing is off
   * by default.
   * @return true if lenient-parse mode is turned on.
   * @see #setLenient
   * @stable ICU 2.0
   */
  virtual inline UBool isLenient(void) const;

#endif

  /**
   * Override the default rule set to use.  If ruleSetName is null, reset
   * to the initial default rule set.  If the rule set is not a public rule set name,
   * U_ILLEGAL_ARGUMENT_ERROR is returned in status.
   * @param ruleSetName the name of the rule set, or null to reset the initial default.
   * @param status set to failure code when a problem occurs.
   * @stable ICU 2.6
   */
  virtual void setDefaultRuleSet(const UnicodeString& ruleSetName, UErrorCode& status);

  /**
   * Return the name of the current default rule set.  If the current rule set is
   * not public, returns a bogus (and empty) UnicodeString.
   * @return the name of the current default rule set
   * @stable ICU 3.0
   */
  virtual UnicodeString getDefaultRuleSetName() const;

  /**
   * Set a particular UDisplayContext value in the formatter, such as
   * UDISPCTX_CAPITALIZATION_FOR_STANDALONE. Note: For getContext, see
   * NumberFormat.
   * @param value The UDisplayContext value to set.
   * @param status Input/output status. If at entry this indicates a failure
   *               status, the function will do nothing; otherwise this will be
   *               updated with any new status from the function.
   * @stable ICU 53
   */
  virtual void setContext(UDisplayContext value, UErrorCode& status);

    /**
     * Get the rounding mode.
     * @return A rounding mode
     * @stable ICU 60
     */
    virtual ERoundingMode getRoundingMode(void) const;

    /**
     * Set the rounding mode.
     * @param roundingMode A rounding mode
     * @stable ICU 60
     */
    virtual void setRoundingMode(ERoundingMode roundingMode);

public:
    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     *
     * @stable ICU 2.8
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     *
     * @stable ICU 2.8
     */
    virtual UClassID getDynamicClassID(void) const;

    /**
     * Sets the decimal format symbols, which is generally not changed
     * by the programmer or user. The formatter takes ownership of
     * symbolsToAdopt; the client must not delete it.
     *
     * @param symbolsToAdopt DecimalFormatSymbols to be adopted.
     * @stable ICU 49
     */
    virtual void adoptDecimalFormatSymbols(DecimalFormatSymbols* symbolsToAdopt);

    /**
     * Sets the decimal format symbols, which is generally not changed
     * by the programmer or user. A clone of the symbols is created and
     * the symbols is _not_ adopted; the client is still responsible for
     * deleting it.
     *
     * @param symbols DecimalFormatSymbols.
     * @stable ICU 49
     */
    virtual void setDecimalFormatSymbols(const DecimalFormatSymbols& symbols);

private:
    RuleBasedNumberFormat(); // default constructor not implemented

    // this will ref the localizations if they are not NULL
    // caller must deref to get adoption
    RuleBasedNumberFormat(const UnicodeString& description, LocalizationInfo* localizations,
              const Locale& locale, UParseError& perror, UErrorCode& status);

    void init(const UnicodeString& rules, LocalizationInfo* localizations, UParseError& perror, UErrorCode& status);
    void initCapitalizationContextInfo(const Locale& thelocale);
    void dispose();
    void stripWhitespace(UnicodeString& src);
    void initDefaultRuleSet();
    NFRuleSet* findRuleSet(const UnicodeString& name, UErrorCode& status) const;

    /* friend access */
    friend class NFSubstitution;
    friend class NFRule;
    friend class NFRuleSet;
    friend class FractionalPartSubstitution;

    inline NFRuleSet * getDefaultRuleSet() const;
    const RuleBasedCollator * getCollator() const;
    DecimalFormatSymbols * initializeDecimalFormatSymbols(UErrorCode &status);
    const DecimalFormatSymbols * getDecimalFormatSymbols() const;
    NFRule * initializeDefaultInfinityRule(UErrorCode &status);
    const NFRule * getDefaultInfinityRule() const;
    NFRule * initializeDefaultNaNRule(UErrorCode &status);
    const NFRule * getDefaultNaNRule() const;
    PluralFormat *createPluralFormat(UPluralType pluralType, const UnicodeString &pattern, UErrorCode& status) const;
    UnicodeString& adjustForCapitalizationContext(int32_t startPos, UnicodeString& currentResult, UErrorCode& status) const;
    UnicodeString& format(int64_t number, NFRuleSet *ruleSet, UnicodeString& toAppendTo, UErrorCode& status) const;
    void format(double number, NFRuleSet& rs, UnicodeString& toAppendTo, UErrorCode& status) const;

private:
    NFRuleSet **fRuleSets;
    UnicodeString* ruleSetDescriptions;
    int32_t numRuleSets;
    NFRuleSet *defaultRuleSet;
    Locale locale;
    RuleBasedCollator* collator;
    DecimalFormatSymbols* decimalFormatSymbols;
    NFRule *defaultInfinityRule;
    NFRule *defaultNaNRule;
    ERoundingMode fRoundingMode;
    UBool lenient;
    UnicodeString* lenientParseRules;
    LocalizationInfo* localizations;
    UnicodeString originalDescription;
    UBool capitalizationInfoSet;
    UBool capitalizationForUIListMenu;
    UBool capitalizationForStandAlone;
    void* capitalizationBrkIter;
};

// ---------------

#if !UCONFIG_NO_COLLATION

inline UBool
RuleBasedNumberFormat::isLenient(void) const {
    return lenient;
}

#endif

inline NFRuleSet*
RuleBasedNumberFormat::getDefaultRuleSet() const {
    return defaultRuleSet;
}

U_NAMESPACE_END

/* U_HAVE_RBNF */
#endif

#endif /* U_SHOW_CPLUSPLUS_API */

/* RBNF_H */

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/********************************************************************
 * COPYRIGHT:
 * Copyright (c) 1997-2011, International Business Machines Corporation and
 * others. All Rights Reserved.
 * Copyright (C) 2010 , Yahoo! Inc.
 ********************************************************************
 *
 * File SELFMT.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   11/11/09    kirtig      Finished first cut of implementation.
 ********************************************************************/




#if U_SHOW_CPLUSPLUS_API




/**
 * \file
 * \brief C++ API: SelectFormat object
 */

#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

class MessageFormat;

/**
  * <p><code>SelectFormat</code> supports the creation of  internationalized
  * messages by selecting phrases based on keywords. The pattern  specifies
  * how to map keywords to phrases and provides a default phrase. The
  * object provided to the format method is a string that's matched
  * against the keywords. If there is a match, the corresponding phrase
  * is selected; otherwise, the default phrase is used.</p>
  *
  * <h4>Using <code>SelectFormat</code> for Gender Agreement</h4>
  *
  * <p>Note: Typically, select formatting is done via <code>MessageFormat</code>
  * with a <code>select</code> argument type,
  * rather than using a stand-alone <code>SelectFormat</code>.</p>
  *
  * <p>The main use case for the select format is gender based  inflection.
  * When names or nouns are inserted into sentences, their gender can  affect pronouns,
  * verb forms, articles, and adjectives. Special care needs to be
  * taken for the case where the gender cannot be determined.
  * The impact varies between languages:</p>
  * \htmlonly
  * <ul>
  * <li>English has three genders, and unknown gender is handled as a  special
  * case. Names use the gender of the named person (if known), nouns  referring
  * to people use natural gender, and inanimate objects are usually  neutral.
  * The gender only affects pronouns: "he", "she", "it", "they".
  *
  * <li>German differs from English in that the gender of nouns is  rather
  * arbitrary, even for nouns referring to people ("M&#x00E4;dchen", girl, is  neutral).
  * The gender affects pronouns ("er", "sie", "es"), articles ("der",  "die",
  * "das"), and adjective forms ("guter Mann", "gute Frau", "gutes  M&#x00E4;dchen").
  *
  * <li>French has only two genders; as in German the gender of nouns
  * is rather arbitrary - for sun and moon, the genders
  * are the opposite of those in German. The gender affects
  * pronouns ("il", "elle"), articles ("le", "la"),
  * adjective forms ("bon", "bonne"), and sometimes
  * verb forms ("all&#x00E9;", "all&#x00E9;e").
  *
  * <li>Polish distinguishes five genders (or noun classes),
  * human masculine, animate non-human masculine, inanimate masculine,
  * feminine, and neuter.
  * </ul>
  * \endhtmlonly
  * <p>Some other languages have noun classes that are not related to  gender,
  * but similar in grammatical use.
  * Some African languages have around 20 noun classes.</p>
  *
  * <p><b>Note:</b>For the gender of a <i>person</i> in a given sentence,
  * we usually need to distinguish only between female, male and other/unknown.</p>
  *
  * <p>To enable localizers to create sentence patterns that take their
  * language's gender dependencies into consideration, software has to  provide
  * information about the gender associated with a noun or name to
  * <code>MessageFormat</code>.
  * Two main cases can be distinguished:</p>
  *
  * <ul>
  * <li>For people, natural gender information should be maintained  for each person.
  * Keywords like "male", "female", "mixed" (for groups of people)
  * and "unknown" could be used.
  *
  * <li>For nouns, grammatical gender information should be maintained  for
  * each noun and per language, e.g., in resource bundles.
  * The keywords "masculine", "feminine", and "neuter" are commonly  used,
  * but some languages may require other keywords.
  * </ul>
  *
  * <p>The resulting keyword is provided to <code>MessageFormat</code>  as a
  * parameter separate from the name or noun it's associated with. For  example,
  * to generate a message such as "Jean went to Paris", three separate  arguments
  * would be provided: The name of the person as argument 0, the  gender of
  * the person as argument 1, and the name of the city as argument 2.
  * The sentence pattern for English, where the gender of the person has
  * no impact on this simple sentence, would not refer to argument 1  at all:</p>
  *
  * <pre>{0} went to {2}.</pre>
  *
  * <p><b>Note:</b> The entire sentence should be included (and partially repeated)
  * inside each phrase. Otherwise translators would have to be trained on how to
  * move bits of the sentence in and out of the select argument of a message.
  * (The examples below do not follow this recommendation!)</p>
  *
  * <p>The sentence pattern for French, where the gender of the person affects
  * the form of the participle, uses a select format based on argument 1:</p>
  *
  * \htmlonly<pre>{0} est {1, select, female {all&#x00E9;e} other {all&#x00E9;}} &#x00E0; {2}.</pre>\endhtmlonly
  *
  * <p>Patterns can be nested, so that it's possible to handle  interactions of
  * number and gender where necessary. For example, if the above  sentence should
  * allow for the names of several people to be inserted, the  following sentence
  * pattern can be used (with argument 0 the list of people's names,
  * argument 1 the number of people, argument 2 their combined gender, and
  * argument 3 the city name):</p>
  *
  * \htmlonly
  * <pre>{0} {1, plural,
  *                 one {est {2, select, female {all&#x00E9;e} other  {all&#x00E9;}}}
  *                 other {sont {2, select, female {all&#x00E9;es} other {all&#x00E9;s}}}
  *          }&#x00E0; {3}.</pre>
  * \endhtmlonly
  *
  * <h4>Patterns and Their Interpretation</h4>
  *
  * <p>The <code>SelectFormat</code> pattern string defines the phrase output
  * for each user-defined keyword.
  * The pattern is a sequence of (keyword, message) pairs.
  * A keyword is a "pattern identifier": [^[[:Pattern_Syntax:][:Pattern_White_Space:]]]+</p>
  *
  * <p>Each message is a MessageFormat pattern string enclosed in {curly braces}.</p>
  *
  * <p>You always have to define a phrase for the default keyword
  * <code>other</code>; this phrase is returned when the keyword
  * provided to
  * the <code>format</code> method matches no other keyword.
  * If a pattern does not provide a phrase for <code>other</code>, the  method
  * it's provided to returns the error  <code>U_DEFAULT_KEYWORD_MISSING</code>.
  * <br>
  * Pattern_White_Space between keywords and messages is ignored.
  * Pattern_White_Space within a message is preserved and output.</p>
  *
  * <p><pre>Example:
  * \htmlonly
  *
  * UErrorCode status = U_ZERO_ERROR;
  * MessageFormat *msgFmt = new MessageFormat(UnicodeString("{0} est  {1, select, female {all&#x00E9;e} other {all&#x00E9;}} &#x00E0; Paris."), Locale("fr"),  status);
  * if (U_FAILURE(status)) {
  *       return;
  * }
  * FieldPosition ignore(FieldPosition::DONT_CARE);
  * UnicodeString result;
  *
  * char* str1= "Kirti,female";
  * Formattable args1[] = {"Kirti","female"};
  * msgFmt->format(args1, 2, result, ignore, status);
  * cout << "Input is " << str1 << " and result is: " << result << endl;
  * delete msgFmt;
  *
  * \endhtmlonly
  * </pre>
  * </p>
  *
  * Produces the output:<br>
  * \htmlonly
  * <code>Kirti est all&#x00E9;e &#x00E0; Paris.</code>
  * \endhtmlonly
  *
  * @stable ICU 4.4
  */

class U_I18N_API SelectFormat : public Format {
public:

    /**
     * Creates a new <code>SelectFormat</code> for a given pattern string.
     * @param  pattern the pattern for this <code>SelectFormat</code>.
     *                 errors are returned to status if the pattern is invalid.
     * @param status   output param set to success/failure code on exit, which
     *                 must not indicate a failure before the function call.
     * @stable ICU 4.4
     */
    SelectFormat(const UnicodeString& pattern, UErrorCode& status);

    /**
     * copy constructor.
     * @stable ICU 4.4
     */
    SelectFormat(const SelectFormat& other);

    /**
     * Destructor.
     * @stable ICU 4.4
     */
    virtual ~SelectFormat();

    /**
     * Sets the pattern used by this select format.
     * for the keyword rules.
     * Patterns and their interpretation are specified in the class description.
     *
     * @param pattern the pattern for this select format
     *                errors are returned to status if the pattern is invalid.
     * @param status  output param set to success/failure code on exit, which
     *                must not indicate a failure before the function call.
     * @stable ICU 4.4
     */
    void applyPattern(const UnicodeString& pattern, UErrorCode& status);


    using Format::format;

    /**
     * Selects the phrase for  the given keyword
     *
     * @param keyword  The keyword that is used to select an alternative.
     * @param appendTo output parameter to receive result.
     *                 result is appended to existing contents.
     * @param pos      On input: an alignment field, if desired.
     *                 On output: the offsets of the alignment field.
     * @param status  output param set to success/failure code on exit, which
     *                 must not indicate a failure before the function call.
     * @return         Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    UnicodeString& format(const UnicodeString& keyword,
                            UnicodeString& appendTo,
                            FieldPosition& pos,
                            UErrorCode& status) const;

    /**
     * Assignment operator
     *
     * @param other    the SelectFormat object to copy from.
     * @stable ICU 4.4
     */
    SelectFormat& operator=(const SelectFormat& other);

    /**
     * Return true if another object is semantically equal to this one.
     *
     * @param other    the SelectFormat object to be compared with.
     * @return         true if other is semantically equal to this.
     * @stable ICU 4.4
     */
    virtual UBool operator==(const Format& other) const;

    /**
     * Return true if another object is semantically unequal to this one.
     *
     * @param other    the SelectFormat object to be compared with.
     * @return         true if other is semantically unequal to this.
     * @stable ICU 4.4
     */
    virtual UBool operator!=(const Format& other) const;

    /**
     * Clones this Format object polymorphically.  The caller owns the
     * result and should delete it when done.
     * @stable ICU 4.4
     */
    virtual SelectFormat* clone() const;

    /**
     * Format an object to produce a string.
     * This method handles keyword strings.
     * If the Formattable object is not a <code>UnicodeString</code>,
     * then it returns a failing UErrorCode.
     *
     * @param obj       A keyword string that is used to select an alternative.
     * @param appendTo  output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @param status    output param filled with success/failure status.
     * @return          Reference to 'appendTo' parameter.
     * @stable ICU 4.4
     */
    UnicodeString& format(const Formattable& obj,
                         UnicodeString& appendTo,
                         FieldPosition& pos,
                         UErrorCode& status) const;

    /**
     * Returns the pattern from applyPattern() or constructor.
     *
     * @param  appendTo  output parameter to receive result.
     *                  Result is appended to existing contents.
     * @return the UnicodeString with inserted pattern.
     * @stable ICU 4.4
     */
    UnicodeString& toPattern(UnicodeString& appendTo);

    /**
     * This method is not yet supported by <code>SelectFormat</code>.
     * <P>
     * Before calling, set parse_pos.index to the offset you want to start
     * parsing at in the source. After calling, parse_pos.index is the end of
     * the text you parsed. If error occurs, index is unchanged.
     * <P>
     * When parsing, leading whitespace is discarded (with a successful parse),
     * while trailing whitespace is left as is.
     * <P>
     * See Format::parseObject() for more.
     *
     * @param source     The string to be parsed into an object.
     * @param result     Formattable to be set to the parse result.
     *     If parse fails, return contents are undefined.
     * @param parse_pos The position to start parsing at. Upon return
     *     this param is set to the position after the
     *     last character successfully parsed. If the
     *     source is not parsed successfully, this param
     *     will remain unchanged.
     * @stable ICU 4.4
     */
    virtual void parseObject(const UnicodeString& source,
                            Formattable& result,
                            ParsePosition& parse_pos) const;

    /**
     * ICU "poor man's RTTI", returns a UClassID for this class.
     * @stable ICU 4.4
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * ICU "poor man's RTTI", returns a UClassID for the actual class.
     * @stable ICU 4.4
     */
    virtual UClassID getDynamicClassID() const;

private:
    friend class MessageFormat;

    SelectFormat();   // default constructor not implemented.

    /**
     * Finds the SelectFormat sub-message for the given keyword, or the "other" sub-message.
     * @param pattern A MessagePattern.
     * @param partIndex the index of the first SelectFormat argument style part.
     * @param keyword a keyword to be matched to one of the SelectFormat argument's keywords.
     * @param ec Error code.
     * @return the sub-message start part index.
     */
    static int32_t findSubMessage(const MessagePattern& pattern, int32_t partIndex,
                                  const UnicodeString& keyword, UErrorCode& ec);

    MessagePattern msgPattern;
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/********************************************************************
 * COPYRIGHT: 
 * Copyright (c) 1997-2011, International Business Machines Corporation and
 * others. All Rights Reserved.
 * Copyright (C) 2010 , Yahoo! Inc. 
 ********************************************************************
 *
 *   file name:  umsg.h
 *   encoding:   UTF-8
 *   tab size:   8 (not used)
 *   indentation:4
 *
 *   Change history:
 *
 *   08/5/2001  Ram         Added C wrappers for C++ API.
 ********************************************************************/




#if !UCONFIG_NO_FORMATTING




#include <stdarg.h>

/**
 * \file
 * \brief C API: MessageFormat
 *
 * <h2>MessageFormat C API </h2>
 *
 * <p>MessageFormat prepares strings for display to users,
 * with optional arguments (variables/placeholders).
 * The arguments can occur in any order, which is necessary for translation
 * into languages with different grammars.
 *
 * <p>The opaque UMessageFormat type is a thin C wrapper around
 * a C++ MessageFormat. It is constructed from a <em>pattern</em> string
 * with arguments in {curly braces} which will be replaced by formatted values.
 *
 * <p>Currently, the C API supports only numbered arguments.
 *
 * <p>For details about the pattern syntax and behavior,
 * especially about the ASCII apostrophe vs. the
 * real apostrophe (single quote) character \htmlonly&#x2019;\endhtmlonly (U+2019),
 * see the C++ MessageFormat class documentation.
 *
 * <p>Here are some examples of C API usage:
 * Example 1:
 * <pre>
 * \code
 *     UChar *result, *tzID, *str;
 *     UChar pattern[100];
 *     int32_t resultLengthOut, resultlength;
 *     UCalendar *cal;
 *     UDate d1;
 *     UDateFormat *def1;
 *     UErrorCode status = U_ZERO_ERROR;
 *
 *     str=(UChar*)malloc(sizeof(UChar) * (strlen("disturbance in force") +1));
 *     u_uastrcpy(str, "disturbance in force");
 *     tzID=(UChar*)malloc(sizeof(UChar) * 4);
 *     u_uastrcpy(tzID, "PST");
 *     cal=ucal_open(tzID, u_strlen(tzID), "en_US", UCAL_TRADITIONAL, &status);
 *     ucal_setDateTime(cal, 1999, UCAL_MARCH, 18, 0, 0, 0, &status);
 *     d1=ucal_getMillis(cal, &status);
 *     u_uastrcpy(pattern, "On {0, date, long}, there was a {1} on planet {2,number,integer}");
 *     resultlength=0;
 *     resultLengthOut=u_formatMessage( "en_US", pattern, u_strlen(pattern), NULL, resultlength, &status, d1, str, 7);
 *     if(status==U_BUFFER_OVERFLOW_ERROR){
 *         status=U_ZERO_ERROR;
 *         resultlength=resultLengthOut+1;
 *         result=(UChar*)realloc(result, sizeof(UChar) * resultlength);
 *         u_formatMessage( "en_US", pattern, u_strlen(pattern), result, resultlength, &status, d1, str, 7);
 *     }
 *     printf("%s\n", austrdup(result) );//austrdup( a function used to convert UChar* to char*)
 *     //output>: "On March 18, 1999, there was a disturbance in force on planet 7
 * \endcode
 * </pre>
 * Typically, the message format will come from resources, and the
 * arguments will be dynamically set at runtime.
 * <P>
 * Example 2:
 * <pre>
 * \code
 *     UChar* str;
 *     UErrorCode status = U_ZERO_ERROR;
 *     UChar *result;
 *     UChar pattern[100];
 *     int32_t resultlength, resultLengthOut, i;
 *     double testArgs= { 100.0, 1.0, 0.0};
 *
 *     str=(UChar*)malloc(sizeof(UChar) * 10);
 *     u_uastrcpy(str, "MyDisk");
 *     u_uastrcpy(pattern, "The disk {1} contains {0,choice,0#no files|1#one file|1<{0,number,integer} files}");
 *     for(i=0; i<3; i++){
 *       resultlength=0; 
 *       resultLengthOut=u_formatMessage( "en_US", pattern, u_strlen(pattern), NULL, resultlength, &status, testArgs[i], str); 
 *       if(status==U_BUFFER_OVERFLOW_ERROR){
 *         status=U_ZERO_ERROR;
 *         resultlength=resultLengthOut+1;
 *         result=(UChar*)malloc(sizeof(UChar) * resultlength);
 *         u_formatMessage( "en_US", pattern, u_strlen(pattern), result, resultlength, &status, testArgs[i], str);
 *       }
 *       printf("%s\n", austrdup(result) );  //austrdup( a function used to convert UChar* to char*)
 *       free(result);
 *     }
 *     // output, with different testArgs:
 *     // output: The disk "MyDisk" contains 100 files.
 *     // output: The disk "MyDisk" contains one file.
 *     // output: The disk "MyDisk" contains no files.
 * \endcode
 *  </pre>
 *
 *
 * Example 3:
 * <pre>
 * \code
 * UChar* str;
 * UChar* str1;
 * UErrorCode status = U_ZERO_ERROR;
 * UChar *result;
 * UChar pattern[100];
 * UChar expected[100];
 * int32_t resultlength,resultLengthOut;

 * str=(UChar*)malloc(sizeof(UChar) * 25);
 * u_uastrcpy(str, "Kirti");
 * str1=(UChar*)malloc(sizeof(UChar) * 25);
 * u_uastrcpy(str1, "female");
 * log_verbose("Testing message format with Select test #1\n:");
 * u_uastrcpy(pattern, "{0} est {1, select, female {all\\u00E9e} other {all\\u00E9}} \\u00E0 Paris.");
 * u_uastrcpy(expected, "Kirti est all\\u00E9e \\u00E0 Paris.");
 * resultlength=0;
 * resultLengthOut=u_formatMessage( "fr", pattern, u_strlen(pattern), NULL, resultlength, &status, str , str1);
 * if(status==U_BUFFER_OVERFLOW_ERROR)
 *  {
 *      status=U_ZERO_ERROR;
 *      resultlength=resultLengthOut+1;
 *      result=(UChar*)malloc(sizeof(UChar) * resultlength);
 *      u_formatMessage( "fr", pattern, u_strlen(pattern), result, resultlength, &status, str , str1);
 *      if(u_strcmp(result, expected)==0)
 *          log_verbose("PASS: MessagFormat successful on Select test#1\n");
 *      else{
 *          log_err("FAIL: Error in MessageFormat on Select test#1\n GOT %s EXPECTED %s\n", austrdup(result),
 *          austrdup(expected) );
 *      }
 *      free(result);
 * }
 * \endcode
 *  </pre>
 */

/**
 * Format a message for a locale.
 * This function may perform re-ordering of the arguments depending on the
 * locale. For all numeric arguments, double is assumed unless the type is
 * explicitly integer.  All choice format arguments must be of type double.
 * @param locale The locale for which the message will be formatted
 * @param pattern The pattern specifying the message's format
 * @param patternLength The length of pattern
 * @param result A pointer to a buffer to receive the formatted message.
 * @param resultLength The maximum size of result.
 * @param status A pointer to an UErrorCode to receive any errors
 * @param ... A variable-length argument list containing the arguments specified
 * in pattern.
 * @return The total buffer size needed; if greater than resultLength, the
 * output was truncated.
 * @see u_parseMessage
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
u_formatMessage(const char  *locale,
                 const UChar *pattern,
                int32_t     patternLength,
                UChar       *result,
                int32_t     resultLength,
                UErrorCode  *status,
                ...);

/**
 * Format a message for a locale.
 * This function may perform re-ordering of the arguments depending on the
 * locale. For all numeric arguments, double is assumed unless the type is
 * explicitly integer.  All choice format arguments must be of type double.
 * @param locale The locale for which the message will be formatted
 * @param pattern The pattern specifying the message's format
 * @param patternLength The length of pattern
 * @param result A pointer to a buffer to receive the formatted message.
 * @param resultLength The maximum size of result.
 * @param ap A variable-length argument list containing the arguments specified
 * @param status A pointer to an UErrorCode to receive any errors
 * in pattern.
 * @return The total buffer size needed; if greater than resultLength, the
 * output was truncated.
 * @see u_parseMessage
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
u_vformatMessage(   const char  *locale,
                    const UChar *pattern,
                    int32_t     patternLength,
                    UChar       *result,
                    int32_t     resultLength,
                    va_list     ap,
                    UErrorCode  *status);

/**
 * Parse a message.
 * For numeric arguments, this function will always use doubles.  Integer types
 * should not be passed.
 * This function is not able to parse all output from {@link #u_formatMessage }.
 * @param locale The locale for which the message is formatted
 * @param pattern The pattern specifying the message's format
 * @param patternLength The length of pattern
 * @param source The text to parse.
 * @param sourceLength The length of source, or -1 if null-terminated.
 * @param status A pointer to an UErrorCode to receive any errors
 * @param ... A variable-length argument list containing the arguments
 * specified in pattern.
 * @see u_formatMessage
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
u_parseMessage( const char   *locale,
                const UChar  *pattern,
                int32_t      patternLength,
                const UChar  *source,
                int32_t      sourceLength,
                UErrorCode   *status,
                ...);

/**
 * Parse a message.
 * For numeric arguments, this function will always use doubles.  Integer types
 * should not be passed.
 * This function is not able to parse all output from {@link #u_formatMessage }.
 * @param locale The locale for which the message is formatted
 * @param pattern The pattern specifying the message's format
 * @param patternLength The length of pattern
 * @param source The text to parse.
 * @param sourceLength The length of source, or -1 if null-terminated.
 * @param ap A variable-length argument list containing the arguments
 * @param status A pointer to an UErrorCode to receive any errors
 * specified in pattern.
 * @see u_formatMessage
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
u_vparseMessage(const char  *locale,
                const UChar *pattern,
                int32_t     patternLength,
                const UChar *source,
                int32_t     sourceLength,
                va_list     ap,
                UErrorCode  *status);

/**
 * Format a message for a locale.
 * This function may perform re-ordering of the arguments depending on the
 * locale. For all numeric arguments, double is assumed unless the type is
 * explicitly integer.  All choice format arguments must be of type double.
 * @param locale The locale for which the message will be formatted
 * @param pattern The pattern specifying the message's format
 * @param patternLength The length of pattern
 * @param result A pointer to a buffer to receive the formatted message.
 * @param resultLength The maximum size of result.
 * @param status A pointer to an UErrorCode to receive any errors
 * @param ... A variable-length argument list containing the arguments specified
 * in pattern.
 * @param parseError  A pointer to UParseError to receive information about errors
 *                     occurred during parsing.
 * @return The total buffer size needed; if greater than resultLength, the
 * output was truncated.
 * @see u_parseMessage
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
u_formatMessageWithError(   const char    *locale,
                            const UChar   *pattern,
                            int32_t       patternLength,
                            UChar         *result,
                            int32_t       resultLength,
                            UParseError   *parseError,
                            UErrorCode    *status,
                            ...);

/**
 * Format a message for a locale.
 * This function may perform re-ordering of the arguments depending on the
 * locale. For all numeric arguments, double is assumed unless the type is
 * explicitly integer.  All choice format arguments must be of type double.
 * @param locale The locale for which the message will be formatted
 * @param pattern The pattern specifying the message's format
 * @param patternLength The length of pattern
 * @param result A pointer to a buffer to receive the formatted message.
 * @param resultLength The maximum size of result.
 * @param parseError  A pointer to UParseError to receive information about errors
 *                    occurred during parsing.
 * @param ap A variable-length argument list containing the arguments specified
 * @param status A pointer to an UErrorCode to receive any errors
 * in pattern.
 * @return The total buffer size needed; if greater than resultLength, the
 * output was truncated.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
u_vformatMessageWithError(  const char   *locale,
                            const UChar  *pattern,
                            int32_t      patternLength,
                            UChar        *result,
                            int32_t      resultLength,
                            UParseError* parseError,
                            va_list      ap,
                            UErrorCode   *status);

/**
 * Parse a message.
 * For numeric arguments, this function will always use doubles.  Integer types
 * should not be passed.
 * This function is not able to parse all output from {@link #u_formatMessage }.
 * @param locale The locale for which the message is formatted
 * @param pattern The pattern specifying the message's format
 * @param patternLength The length of pattern
 * @param source The text to parse.
 * @param sourceLength The length of source, or -1 if null-terminated.
 * @param parseError  A pointer to UParseError to receive information about errors
 *                     occurred during parsing.
 * @param status A pointer to an UErrorCode to receive any errors
 * @param ... A variable-length argument list containing the arguments
 * specified in pattern.
 * @see u_formatMessage
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
u_parseMessageWithError(const char  *locale,
                        const UChar *pattern,
                        int32_t     patternLength,
                        const UChar *source,
                        int32_t     sourceLength,
                        UParseError *parseError,
                        UErrorCode  *status,
                        ...);

/**
 * Parse a message.
 * For numeric arguments, this function will always use doubles.  Integer types
 * should not be passed.
 * This function is not able to parse all output from {@link #u_formatMessage }.
 * @param locale The locale for which the message is formatted
 * @param pattern The pattern specifying the message's format
 * @param patternLength The length of pattern
 * @param source The text to parse.
 * @param sourceLength The length of source, or -1 if null-terminated.
 * @param ap A variable-length argument list containing the arguments
 * @param parseError  A pointer to UParseError to receive information about errors
 *                     occurred during parsing.
 * @param status A pointer to an UErrorCode to receive any errors
 * specified in pattern.
 * @see u_formatMessage
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
u_vparseMessageWithError(const char  *locale,
                         const UChar *pattern,
                         int32_t     patternLength,
                         const UChar *source,
                         int32_t     sourceLength,
                         va_list     ap,
                         UParseError *parseError,
                         UErrorCode* status);

/*----------------------- New experimental API --------------------------- */
/** 
 * The message format object
 * @stable ICU 2.0
 */
typedef void* UMessageFormat;


/**
 * Open a message formatter with given pattern and for the given locale.
 * @param pattern       A pattern specifying the format to use.
 * @param patternLength Length of the pattern to use
 * @param locale        The locale for which the messages are formatted.
 * @param parseError    A pointer to UParseError struct to receive any errors 
 *                      occured during parsing. Can be NULL.
 * @param status        A pointer to an UErrorCode to receive any errors.
 * @return              A pointer to a UMessageFormat to use for formatting 
 *                      messages, or 0 if an error occurred. 
 * @stable ICU 2.0
 */
U_STABLE UMessageFormat* U_EXPORT2 
umsg_open(  const UChar     *pattern,
            int32_t         patternLength,
            const  char     *locale,
            UParseError     *parseError,
            UErrorCode      *status);

/**
 * Close a UMessageFormat.
 * Once closed, a UMessageFormat may no longer be used.
 * @param format The formatter to close.
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
umsg_close(UMessageFormat* format);

#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUMessageFormatPointer
 * "Smart pointer" class, closes a UMessageFormat via umsg_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.4
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUMessageFormatPointer, UMessageFormat, umsg_close);

U_NAMESPACE_END

#endif

/**
 * Open a copy of a UMessageFormat.
 * This function performs a deep copy.
 * @param fmt The formatter to copy
 * @param status A pointer to an UErrorCode to receive any errors.
 * @return A pointer to a UDateFormat identical to fmt.
 * @stable ICU 2.0
 */
U_STABLE UMessageFormat U_EXPORT2 
umsg_clone(const UMessageFormat *fmt,
           UErrorCode *status);

/**
 * Sets the locale. This locale is used for fetching default number or date
 * format information.
 * @param fmt The formatter to set
 * @param locale The locale the formatter should use.
 * @stable ICU 2.0
 */
U_STABLE void  U_EXPORT2 
umsg_setLocale(UMessageFormat *fmt,
               const char* locale);

/**
 * Gets the locale. This locale is used for fetching default number or date
 * format information.
 * @param fmt The formatter to querry
 * @return the locale.
 * @stable ICU 2.0
 */
U_STABLE const char*  U_EXPORT2 
umsg_getLocale(const UMessageFormat *fmt);

/**
 * Sets the pattern.
 * @param fmt           The formatter to use
 * @param pattern       The pattern to be applied.
 * @param patternLength Length of the pattern to use
 * @param parseError    Struct to receive information on position 
 *                      of error if an error is encountered.Can be NULL.
 * @param status        Output param set to success/failure code on
 *                      exit. If the pattern is invalid, this will be
 *                      set to a failure result.
 * @stable ICU 2.0
 */
U_STABLE void  U_EXPORT2 
umsg_applyPattern( UMessageFormat *fmt,
                   const UChar* pattern,
                   int32_t patternLength,
                   UParseError* parseError,
                   UErrorCode* status);

/**
 * Gets the pattern.
 * @param fmt          The formatter to use
 * @param result       A pointer to a buffer to receive the pattern.
 * @param resultLength The maximum size of result.
 * @param status       Output param set to success/failure code on
 *                     exit. If the pattern is invalid, this will be
 *                     set to a failure result.  
 * @return the pattern of the format
 * @stable ICU 2.0
 */
U_STABLE int32_t  U_EXPORT2 
umsg_toPattern(const UMessageFormat *fmt,
               UChar* result, 
               int32_t resultLength,
               UErrorCode* status);

/**
 * Format a message for a locale.
 * This function may perform re-ordering of the arguments depending on the
 * locale. For all numeric arguments, double is assumed unless the type is
 * explicitly integer.  All choice format arguments must be of type double.
 * @param fmt           The formatter to use
 * @param result        A pointer to a buffer to receive the formatted message.
 * @param resultLength  The maximum size of result.
 * @param status        A pointer to an UErrorCode to receive any errors
 * @param ...           A variable-length argument list containing the arguments 
 *                      specified in pattern.
 * @return              The total buffer size needed; if greater than resultLength, 
 *                      the output was truncated.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
umsg_format(    const UMessageFormat *fmt,
                UChar          *result,
                int32_t        resultLength,
                UErrorCode     *status,
                ...);

/**
 * Format a message for a locale.
 * This function may perform re-ordering of the arguments depending on the
 * locale. For all numeric arguments, double is assumed unless the type is
 * explicitly integer.  All choice format arguments must be of type double.
 * @param fmt          The formatter to use 
 * @param result       A pointer to a buffer to receive the formatted message.
 * @param resultLength The maximum size of result.
 * @param ap           A variable-length argument list containing the arguments 
 * @param status       A pointer to an UErrorCode to receive any errors
 *                     specified in pattern.
 * @return             The total buffer size needed; if greater than resultLength, 
 *                     the output was truncated.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
umsg_vformat(   const UMessageFormat *fmt,
                UChar          *result,
                int32_t        resultLength,
                va_list        ap,
                UErrorCode     *status);

/**
 * Parse a message.
 * For numeric arguments, this function will always use doubles.  Integer types
 * should not be passed.
 * This function is not able to parse all output from {@link #umsg_format }.
 * @param fmt           The formatter to use 
 * @param source        The text to parse.
 * @param sourceLength  The length of source, or -1 if null-terminated.
 * @param count         Output param to receive number of elements returned.
 * @param status        A pointer to an UErrorCode to receive any errors
 * @param ...           A variable-length argument list containing the arguments
 *                      specified in pattern.
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
umsg_parse( const UMessageFormat *fmt,
            const UChar    *source,
            int32_t        sourceLength,
            int32_t        *count,
            UErrorCode     *status,
            ...);

/**
 * Parse a message.
 * For numeric arguments, this function will always use doubles.  Integer types
 * should not be passed.
 * This function is not able to parse all output from {@link #umsg_format }.
 * @param fmt           The formatter to use 
 * @param source        The text to parse.
 * @param sourceLength  The length of source, or -1 if null-terminated.
 * @param count         Output param to receive number of elements returned.
 * @param ap            A variable-length argument list containing the arguments
 * @param status        A pointer to an UErrorCode to receive any errors
 *                      specified in pattern.
 * @see u_formatMessage
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
umsg_vparse(const UMessageFormat *fmt,
            const UChar    *source,
            int32_t        sourceLength,
            int32_t        *count,
            va_list        ap,
            UErrorCode     *status);


/**
 * Convert an 'apostrophe-friendly' pattern into a standard
 * pattern.  Standard patterns treat all apostrophes as
 * quotes, which is problematic in some languages, e.g. 
 * French, where apostrophe is commonly used.  This utility
 * assumes that only an unpaired apostrophe immediately before
 * a brace is a true quote.  Other unpaired apostrophes are paired,
 * and the resulting standard pattern string is returned.
 *
 * <p><b>Note</b> it is not guaranteed that the returned pattern
 * is indeed a valid pattern.  The only effect is to convert
 * between patterns having different quoting semantics.
 *
 * @param pattern the 'apostrophe-friendly' patttern to convert
 * @param patternLength the length of pattern, or -1 if unknown and pattern is null-terminated
 * @param dest the buffer for the result, or NULL if preflight only
 * @param destCapacity the length of the buffer, or 0 if preflighting
 * @param ec the error code
 * @return the length of the resulting text, not including trailing null
 *        if buffer has room for the trailing null, it is provided, otherwise
 *        not
 * @stable ICU 3.4
 */
U_STABLE int32_t U_EXPORT2 
umsg_autoQuoteApostrophe(const UChar* pattern, 
                         int32_t patternLength,
                         UChar* dest,
                         int32_t destCapacity,
                         UErrorCode* ec);

#endif /* #if !UCONFIG_NO_FORMATTING */






// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2008, International Business Machines Corporation and
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* File MSGFMT.H
*
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING
    




U_NAMESPACE_BEGIN

class FormatNameEnumeration : public StringEnumeration {
public:
    FormatNameEnumeration(UVector *fFormatNames, UErrorCode& status);
    virtual ~FormatNameEnumeration();
    static UClassID U_EXPORT2 getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
    virtual const UnicodeString* snext(UErrorCode& status);
    virtual void reset(UErrorCode& status);
    virtual int32_t count(UErrorCode& status) const;
private:
    int32_t pos;
    UVector *fFormatNames;
};

U_NAMESPACE_END

#endif











// *****************************************************************************
// class MessageFormat
// *****************************************************************************

#define MSGFMT_SINGLE_QUOTE      ((UChar)0x0027)
#define MSGFMT_COMMA             ((UChar)0x002C)
#define MSGFMT_LEFT_CURLY_BRACE  ((UChar)0x007B)
#define MSGFMT_RIGHT_CURLY_BRACE ((UChar)0x007D)

//---------------------------------------
// static data

static const UChar ID_NUMBER[]    = {
    0x6E, 0x75, 0x6D, 0x62, 0x65, 0x72, 0  /* "number" */
};
static const UChar ID_DATE[]      = {
    0x64, 0x61, 0x74, 0x65, 0              /* "date" */
};
static const UChar ID_TIME[]      = {
    0x74, 0x69, 0x6D, 0x65, 0              /* "time" */
};
static const UChar ID_SPELLOUT[]  = {
    0x73, 0x70, 0x65, 0x6c, 0x6c, 0x6f, 0x75, 0x74, 0 /* "spellout" */
};
static const UChar ID_ORDINAL[]   = {
    0x6f, 0x72, 0x64, 0x69, 0x6e, 0x61, 0x6c, 0 /* "ordinal" */
};
static const UChar ID_DURATION[]  = {
    0x64, 0x75, 0x72, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0 /* "duration" */
};

// MessageFormat Type List  Number, Date, Time or Choice
static const UChar * const TYPE_IDS[] = {
    ID_NUMBER,
    ID_DATE,
    ID_TIME,
    ID_SPELLOUT,
    ID_ORDINAL,
    ID_DURATION,
    NULL,
};

static const UChar ID_EMPTY[]     = {
    0 /* empty string, used for default so that null can mark end of list */
};
static const UChar ID_CURRENCY[]  = {
    0x63, 0x75, 0x72, 0x72, 0x65, 0x6E, 0x63, 0x79, 0  /* "currency" */
};
static const UChar ID_PERCENT[]   = {
    0x70, 0x65, 0x72, 0x63, 0x65, 0x6E, 0x74, 0        /* "percent" */
};
static const UChar ID_INTEGER[]   = {
    0x69, 0x6E, 0x74, 0x65, 0x67, 0x65, 0x72, 0        /* "integer" */
};

// NumberFormat modifier list, default, currency, percent or integer
static const UChar * const NUMBER_STYLE_IDS[] = {
    ID_EMPTY,
    ID_CURRENCY,
    ID_PERCENT,
    ID_INTEGER,
    NULL,
};

static const UChar ID_SHORT[]     = {
    0x73, 0x68, 0x6F, 0x72, 0x74, 0        /* "short" */
};
static const UChar ID_MEDIUM[]    = {
    0x6D, 0x65, 0x64, 0x69, 0x75, 0x6D, 0  /* "medium" */
};
static const UChar ID_LONG[]      = {
    0x6C, 0x6F, 0x6E, 0x67, 0              /* "long" */
};
static const UChar ID_FULL[]      = {
    0x66, 0x75, 0x6C, 0x6C, 0              /* "full" */
};

// DateFormat modifier list, default, short, medium, long or full
static const UChar * const DATE_STYLE_IDS[] = {
    ID_EMPTY,
    ID_SHORT,
    ID_MEDIUM,
    ID_LONG,
    ID_FULL,
    NULL,
};

static const icu::DateFormat::EStyle DATE_STYLES[] = {
    icu::DateFormat::kDefault,
    icu::DateFormat::kShort,
    icu::DateFormat::kMedium,
    icu::DateFormat::kLong,
    icu::DateFormat::kFull,
};

static const int32_t DEFAULT_INITIAL_CAPACITY = 10;

static const UChar NULL_STRING[] = {
    0x6E, 0x75, 0x6C, 0x6C, 0  // "null"
};

static const UChar OTHER_STRING[] = {
    0x6F, 0x74, 0x68, 0x65, 0x72, 0  // "other"
};

U_CDECL_BEGIN
static UBool U_CALLCONV equalFormatsForHash(const UHashTok key1,
                                            const UHashTok key2) {
    return icu::MessageFormat::equalFormats(key1.pointer, key2.pointer);
}

U_CDECL_END

U_NAMESPACE_BEGIN

// -------------------------------------
UOBJECT_DEFINE_RTTI_IMPLEMENTATION(MessageFormat)
UOBJECT_DEFINE_RTTI_IMPLEMENTATION(FormatNameEnumeration)

//--------------------------------------------------------------------

/**
 * Convert an integer value to a string and append the result to
 * the given UnicodeString.
 */
static UnicodeString& itos(int32_t i, UnicodeString& appendTo) {
    UChar temp[16];
    uprv_itou(temp,16,i,10,0); // 10 == radix
    appendTo.append(temp, -1);
    return appendTo;
}


// AppendableWrapper: encapsulates the result of formatting, keeping track
// of the string and its length.
class AppendableWrapper : public UMemory {
public:
    AppendableWrapper(Appendable& appendable) : app(appendable), len(0) {
    }
    void append(const UnicodeString& s) {
        app.appendString(s.getBuffer(), s.length());
        len += s.length();
    }
    void append(const UChar* s, const int32_t sLength) {
        app.appendString(s, sLength);
        len += sLength;
    }
    void append(const UnicodeString& s, int32_t start, int32_t length) {
        append(s.tempSubString(start, length));
    }
    void formatAndAppend(const Format* formatter, const Formattable& arg, UErrorCode& ec) {
        UnicodeString s;
        formatter->format(arg, s, ec);
        if (U_SUCCESS(ec)) {
            append(s);
        }
    }
    void formatAndAppend(const Format* formatter, const Formattable& arg,
                         const UnicodeString &argString, UErrorCode& ec) {
        if (!argString.isEmpty()) {
            if (U_SUCCESS(ec)) {
                append(argString);
            }
        } else {
            formatAndAppend(formatter, arg, ec);
        }
    }
    int32_t length() {
        return len;
    }
private:
    Appendable& app;
    int32_t len;
};


// -------------------------------------
// Creates a MessageFormat instance based on the pattern.

MessageFormat::MessageFormat(const UnicodeString& pattern,
                             UErrorCode& success)
: fLocale(Locale::getDefault()),  // Uses the default locale
  msgPattern(success),
  formatAliases(NULL),
  formatAliasesCapacity(0),
  argTypes(NULL),
  argTypeCount(0),
  argTypeCapacity(0),
  hasArgTypeConflicts(FALSE),
  defaultNumberFormat(NULL),
  defaultDateFormat(NULL),
  cachedFormatters(NULL),
  customFormatArgStarts(NULL),
  pluralProvider(*this, UPLURAL_TYPE_CARDINAL),
  ordinalProvider(*this, UPLURAL_TYPE_ORDINAL)
{
    setLocaleIDs(fLocale.getName(), fLocale.getName());
    applyPattern(pattern, success);
}

MessageFormat::MessageFormat(const UnicodeString& pattern,
                             const Locale& newLocale,
                             UErrorCode& success)
: fLocale(newLocale),
  msgPattern(success),
  formatAliases(NULL),
  formatAliasesCapacity(0),
  argTypes(NULL),
  argTypeCount(0),
  argTypeCapacity(0),
  hasArgTypeConflicts(FALSE),
  defaultNumberFormat(NULL),
  defaultDateFormat(NULL),
  cachedFormatters(NULL),
  customFormatArgStarts(NULL),
  pluralProvider(*this, UPLURAL_TYPE_CARDINAL),
  ordinalProvider(*this, UPLURAL_TYPE_ORDINAL)
{
    setLocaleIDs(fLocale.getName(), fLocale.getName());
    applyPattern(pattern, success);
}

MessageFormat::MessageFormat(const UnicodeString& pattern,
                             const Locale& newLocale,
                             UParseError& parseError,
                             UErrorCode& success)
: fLocale(newLocale),
  msgPattern(success),
  formatAliases(NULL),
  formatAliasesCapacity(0),
  argTypes(NULL),
  argTypeCount(0),
  argTypeCapacity(0),
  hasArgTypeConflicts(FALSE),
  defaultNumberFormat(NULL),
  defaultDateFormat(NULL),
  cachedFormatters(NULL),
  customFormatArgStarts(NULL),
  pluralProvider(*this, UPLURAL_TYPE_CARDINAL),
  ordinalProvider(*this, UPLURAL_TYPE_ORDINAL)
{
    setLocaleIDs(fLocale.getName(), fLocale.getName());
    applyPattern(pattern, parseError, success);
}

MessageFormat::MessageFormat(const MessageFormat& that)
:
  Format(that),
  fLocale(that.fLocale),
  msgPattern(that.msgPattern),
  formatAliases(NULL),
  formatAliasesCapacity(0),
  argTypes(NULL),
  argTypeCount(0),
  argTypeCapacity(0),
  hasArgTypeConflicts(that.hasArgTypeConflicts),
  defaultNumberFormat(NULL),
  defaultDateFormat(NULL),
  cachedFormatters(NULL),
  customFormatArgStarts(NULL),
  pluralProvider(*this, UPLURAL_TYPE_CARDINAL),
  ordinalProvider(*this, UPLURAL_TYPE_ORDINAL)
{
    // This will take care of creating the hash tables (since they are NULL).
    UErrorCode ec = U_ZERO_ERROR;
    copyObjects(that, ec);
    if (U_FAILURE(ec)) {
        resetPattern();
    }
}

MessageFormat::~MessageFormat()
{
    uhash_close(cachedFormatters);
    uhash_close(customFormatArgStarts);

    uprv_free(argTypes);
    uprv_free(formatAliases);
    delete defaultNumberFormat;
    delete defaultDateFormat;
}

//--------------------------------------------------------------------
// Variable-size array management

/**
 * Allocate argTypes[] to at least the given capacity and return
 * TRUE if successful.  If not, leave argTypes[] unchanged.
 *
 * If argTypes is NULL, allocate it.  If it is not NULL, enlarge it
 * if necessary to be at least as large as specified.
 */
UBool MessageFormat::allocateArgTypes(int32_t capacity, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (argTypeCapacity >= capacity) {
        return TRUE;
    }
    if (capacity < DEFAULT_INITIAL_CAPACITY) {
        capacity = DEFAULT_INITIAL_CAPACITY;
    } else if (capacity < 2*argTypeCapacity) {
        capacity = 2*argTypeCapacity;
    }
    Formattable::Type* a = (Formattable::Type*)
            uprv_realloc(argTypes, sizeof(*argTypes) * capacity);
    if (a == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    argTypes = a;
    argTypeCapacity = capacity;
    return TRUE;
}

// -------------------------------------
// assignment operator

const MessageFormat&
MessageFormat::operator=(const MessageFormat& that)
{
    if (this != &that) {
        // Calls the super class for assignment first.
        Format::operator=(that);

        setLocale(that.fLocale);
        msgPattern = that.msgPattern;
        hasArgTypeConflicts = that.hasArgTypeConflicts;

        UErrorCode ec = U_ZERO_ERROR;
        copyObjects(that, ec);
        if (U_FAILURE(ec)) {
            resetPattern();
        }
    }
    return *this;
}

UBool
MessageFormat::operator==(const Format& rhs) const
{
    if (this == &rhs) return TRUE;

    MessageFormat& that = (MessageFormat&)rhs;

    // Check class ID before checking MessageFormat members
    if (!Format::operator==(rhs) ||
        msgPattern != that.msgPattern ||
        fLocale != that.fLocale) {
        return FALSE;
    }

    // Compare hashtables.
    if ((customFormatArgStarts == NULL) != (that.customFormatArgStarts == NULL)) {
        return FALSE;
    }
    if (customFormatArgStarts == NULL) {
        return TRUE;
    }

    UErrorCode ec = U_ZERO_ERROR;
    const int32_t count = uhash_count(customFormatArgStarts);
    const int32_t rhs_count = uhash_count(that.customFormatArgStarts);
    if (count != rhs_count) {
        return FALSE;
    }
    int32_t idx = 0, rhs_idx = 0, pos = UHASH_FIRST, rhs_pos = UHASH_FIRST;
    for (; idx < count && rhs_idx < rhs_count && U_SUCCESS(ec); ++idx, ++rhs_idx) {
        const UHashElement* cur = uhash_nextElement(customFormatArgStarts, &pos);
        const UHashElement* rhs_cur = uhash_nextElement(that.customFormatArgStarts, &rhs_pos);
        if (cur->key.integer != rhs_cur->key.integer) {
            return FALSE;
        }
        const Format* format = (const Format*)uhash_iget(cachedFormatters, cur->key.integer);
        const Format* rhs_format = (const Format*)uhash_iget(that.cachedFormatters, rhs_cur->key.integer);
        if (*format != *rhs_format) {
            return FALSE;
        }
    }
    return TRUE;
}

// -------------------------------------
// Creates a copy of this MessageFormat, the caller owns the copy.

MessageFormat*
MessageFormat::clone() const
{
    return new MessageFormat(*this);
}

// -------------------------------------
// Sets the locale of this MessageFormat object to theLocale.

void
MessageFormat::setLocale(const Locale& theLocale)
{
    if (fLocale != theLocale) {
        delete defaultNumberFormat;
        defaultNumberFormat = NULL;
        delete defaultDateFormat;
        defaultDateFormat = NULL;
        fLocale = theLocale;
        setLocaleIDs(fLocale.getName(), fLocale.getName());
        pluralProvider.reset();
        ordinalProvider.reset();
    }
}

// -------------------------------------
// Gets the locale of this MessageFormat object.

const Locale&
MessageFormat::getLocale() const
{
    return fLocale;
}

void
MessageFormat::applyPattern(const UnicodeString& newPattern,
                            UErrorCode& status)
{
    UParseError parseError;
    applyPattern(newPattern,parseError,status);
}


// -------------------------------------
// Applies the new pattern and returns an error if the pattern
// is not correct.
void
MessageFormat::applyPattern(const UnicodeString& pattern,
                            UParseError& parseError,
                            UErrorCode& ec)
{
    if(U_FAILURE(ec)) {
        return;
    }
    msgPattern.parse(pattern, &parseError, ec);
    cacheExplicitFormats(ec);

    if (U_FAILURE(ec)) {
        resetPattern();
    }
}

void MessageFormat::resetPattern() {
    msgPattern.clear();
    uhash_close(cachedFormatters);
    cachedFormatters = NULL;
    uhash_close(customFormatArgStarts);
    customFormatArgStarts = NULL;
    argTypeCount = 0;
    hasArgTypeConflicts = FALSE;
}

void
MessageFormat::applyPattern(const UnicodeString& pattern,
                            UMessagePatternApostropheMode aposMode,
                            UParseError* parseError,
                            UErrorCode& status) {
    if (aposMode != msgPattern.getApostropheMode()) {
        msgPattern.clearPatternAndSetApostropheMode(aposMode);
    }
    applyPattern(pattern, *parseError, status);
}

// -------------------------------------
// Converts this MessageFormat instance to a pattern.

UnicodeString&
MessageFormat::toPattern(UnicodeString& appendTo) const {
    if ((customFormatArgStarts != NULL && 0 != uhash_count(customFormatArgStarts)) ||
        0 == msgPattern.countParts()
    ) {
        appendTo.setToBogus();
        return appendTo;
    }
    return appendTo.append(msgPattern.getPatternString());
}

int32_t MessageFormat::nextTopLevelArgStart(int32_t partIndex) const {
    if (partIndex != 0) {
        partIndex = msgPattern.getLimitPartIndex(partIndex);
    }
    for (;;) {
        UMessagePatternPartType type = msgPattern.getPartType(++partIndex);
        if (type == UMSGPAT_PART_TYPE_ARG_START) {
            return partIndex;
        }
        if (type == UMSGPAT_PART_TYPE_MSG_LIMIT) {
            return -1;
        }
    }
}

void MessageFormat::setArgStartFormat(int32_t argStart,
                                      Format* formatter,
                                      UErrorCode& status) {
    if (U_FAILURE(status)) {
        delete formatter;
        return;
    }
    if (cachedFormatters == NULL) {
        cachedFormatters=uhash_open(uhash_hashLong, uhash_compareLong,
                                    equalFormatsForHash, &status);
        if (U_FAILURE(status)) {
            delete formatter;
            return;
        }
        uhash_setValueDeleter(cachedFormatters, uprv_deleteUObject);
    }
    if (formatter == NULL) {
        formatter = new DummyFormat();
    }
    uhash_iput(cachedFormatters, argStart, formatter, &status);
}


UBool MessageFormat::argNameMatches(int32_t partIndex, const UnicodeString& argName, int32_t argNumber) {
    const MessagePattern::Part& part = msgPattern.getPart(partIndex);
    return part.getType() == UMSGPAT_PART_TYPE_ARG_NAME ?
        msgPattern.partSubstringMatches(part, argName) :
        part.getValue() == argNumber;  // ARG_NUMBER
}

// Sets a custom formatter for a MessagePattern ARG_START part index.
// "Custom" formatters are provided by the user via setFormat() or similar APIs.
void MessageFormat::setCustomArgStartFormat(int32_t argStart,
                                            Format* formatter,
                                            UErrorCode& status) {
    setArgStartFormat(argStart, formatter, status);
    if (customFormatArgStarts == NULL) {
        customFormatArgStarts=uhash_open(uhash_hashLong, uhash_compareLong,
                                         NULL, &status);
    }
    uhash_iputi(customFormatArgStarts, argStart, 1, &status);
}

Format* MessageFormat::getCachedFormatter(int32_t argumentNumber) const {
    if (cachedFormatters == NULL) {
        return NULL;
    }
    void* ptr = uhash_iget(cachedFormatters, argumentNumber);
    if (ptr != NULL && dynamic_cast<DummyFormat*>((Format*)ptr) == NULL) {
        return (Format*) ptr;
    } else {
        // Not cached, or a DummyFormat representing setFormat(NULL).
        return NULL;
    }
}

// -------------------------------------
// Adopts the new formats array and updates the array count.
// This MessageFormat instance owns the new formats.
void
MessageFormat::adoptFormats(Format** newFormats,
                            int32_t count) {
    if (newFormats == NULL || count < 0) {
        return;
    }
    // Throw away any cached formatters.
    if (cachedFormatters != NULL) {
        uhash_removeAll(cachedFormatters);
    }
    if (customFormatArgStarts != NULL) {
        uhash_removeAll(customFormatArgStarts);
    }

    int32_t formatNumber = 0;
    UErrorCode status = U_ZERO_ERROR;
    for (int32_t partIndex = 0;
        formatNumber < count && U_SUCCESS(status) &&
            (partIndex = nextTopLevelArgStart(partIndex)) >= 0;) {
        setCustomArgStartFormat(partIndex, newFormats[formatNumber], status);
        ++formatNumber;
    }
    // Delete those that didn't get used (if any).
    for (; formatNumber < count; ++formatNumber) {
        delete newFormats[formatNumber];
    }

}

// -------------------------------------
// Sets the new formats array and updates the array count.
// This MessageFormat instance maks a copy of the new formats.

void
MessageFormat::setFormats(const Format** newFormats,
                          int32_t count) {
    if (newFormats == NULL || count < 0) {
        return;
    }
    // Throw away any cached formatters.
    if (cachedFormatters != NULL) {
        uhash_removeAll(cachedFormatters);
    }
    if (customFormatArgStarts != NULL) {
        uhash_removeAll(customFormatArgStarts);
    }

    UErrorCode status = U_ZERO_ERROR;
    int32_t formatNumber = 0;
    for (int32_t partIndex = 0;
        formatNumber < count && U_SUCCESS(status) && (partIndex = nextTopLevelArgStart(partIndex)) >= 0;) {
      Format* newFormat = NULL;
      if (newFormats[formatNumber] != NULL) {
          newFormat = newFormats[formatNumber]->clone();
          if (newFormat == NULL) {
              status = U_MEMORY_ALLOCATION_ERROR;
          }
      }
      setCustomArgStartFormat(partIndex, newFormat, status);
      ++formatNumber;
    }
    if (U_FAILURE(status)) {
        resetPattern();
    }
}

// -------------------------------------
// Adopt a single format by format number.
// Do nothing if the format number is not less than the array count.

void
MessageFormat::adoptFormat(int32_t n, Format *newFormat) {
    LocalPointer<Format> p(newFormat);
    if (n >= 0) {
        int32_t formatNumber = 0;
        for (int32_t partIndex = 0; (partIndex = nextTopLevelArgStart(partIndex)) >= 0;) {
            if (n == formatNumber) {
                UErrorCode status = U_ZERO_ERROR;
                setCustomArgStartFormat(partIndex, p.orphan(), status);
                return;
            }
            ++formatNumber;
        }
    }
}

// -------------------------------------
// Adopt a single format by format name.
// Do nothing if there is no match of formatName.
void
MessageFormat::adoptFormat(const UnicodeString& formatName,
                           Format* formatToAdopt,
                           UErrorCode& status) {
    LocalPointer<Format> p(formatToAdopt);
    if (U_FAILURE(status)) {
        return;
    }
    int32_t argNumber = MessagePattern::validateArgumentName(formatName);
    if (argNumber < UMSGPAT_ARG_NAME_NOT_NUMBER) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    for (int32_t partIndex = 0;
        (partIndex = nextTopLevelArgStart(partIndex)) >= 0 && U_SUCCESS(status);
    ) {
        if (argNameMatches(partIndex + 1, formatName, argNumber)) {
            Format* f;
            if (p.isValid()) {
                f = p.orphan();
            } else if (formatToAdopt == NULL) {
                f = NULL;
            } else {
                f = formatToAdopt->clone();
                if (f == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                    return;
                }
            }
            setCustomArgStartFormat(partIndex, f, status);
        }
    }
}

// -------------------------------------
// Set a single format.
// Do nothing if the variable is not less than the array count.
void
MessageFormat::setFormat(int32_t n, const Format& newFormat) {

    if (n >= 0) {
        int32_t formatNumber = 0;
        for (int32_t partIndex = 0;
             (partIndex = nextTopLevelArgStart(partIndex)) >= 0;) {
            if (n == formatNumber) {
                Format* new_format = newFormat.clone();
                if (new_format) {
                    UErrorCode status = U_ZERO_ERROR;
                    setCustomArgStartFormat(partIndex, new_format, status);
                }
                return;
            }
            ++formatNumber;
        }
    }
}

// -------------------------------------
// Get a single format by format name.
// Do nothing if the variable is not less than the array count.
Format *
MessageFormat::getFormat(const UnicodeString& formatName, UErrorCode& status) {
    if (U_FAILURE(status) || cachedFormatters == NULL) return NULL;

    int32_t argNumber = MessagePattern::validateArgumentName(formatName);
    if (argNumber < UMSGPAT_ARG_NAME_NOT_NUMBER) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    for (int32_t partIndex = 0; (partIndex = nextTopLevelArgStart(partIndex)) >= 0;) {
        if (argNameMatches(partIndex + 1, formatName, argNumber)) {
            return getCachedFormatter(partIndex);
        }
    }
    return NULL;
}

// -------------------------------------
// Set a single format by format name
// Do nothing if the variable is not less than the array count.
void
MessageFormat::setFormat(const UnicodeString& formatName,
                         const Format& newFormat,
                         UErrorCode& status) {
    if (U_FAILURE(status)) return;

    int32_t argNumber = MessagePattern::validateArgumentName(formatName);
    if (argNumber < UMSGPAT_ARG_NAME_NOT_NUMBER) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    for (int32_t partIndex = 0;
        (partIndex = nextTopLevelArgStart(partIndex)) >= 0 && U_SUCCESS(status);
    ) {
        if (argNameMatches(partIndex + 1, formatName, argNumber)) {
            Format* new_format = newFormat.clone();
            if (new_format == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            setCustomArgStartFormat(partIndex, new_format, status);
        }
    }
}

// -------------------------------------
// Gets the format array.
const Format**
MessageFormat::getFormats(int32_t& cnt) const
{
    // This old API returns an array (which we hold) of Format*
    // pointers.  The array is valid up to the next call to any
    // method on this object.  We construct and resize an array
    // on demand that contains aliases to the subformats[i].format
    // pointers.

    // Get total required capacity first (it's refreshed on each call).
    int32_t totalCapacity = 0;
    for (int32_t partIndex = 0; (partIndex = nextTopLevelArgStart(partIndex)) >= 0; ++totalCapacity) {}

    MessageFormat* t = const_cast<MessageFormat*> (this);
    cnt = 0;
    if (formatAliases == nullptr) {
        t->formatAliasesCapacity = totalCapacity;
        Format** a = (Format**)
            uprv_malloc(sizeof(Format*) * formatAliasesCapacity);
        if (a == nullptr) {
            t->formatAliasesCapacity = 0;
            return nullptr;
        }
        t->formatAliases = a;
    } else if (totalCapacity > formatAliasesCapacity) {
        Format** a = (Format**)
            uprv_realloc(formatAliases, sizeof(Format*) * totalCapacity);
        if (a == nullptr) {
            t->formatAliasesCapacity = 0;
            return nullptr;
        }
        t->formatAliases = a;
        t->formatAliasesCapacity = totalCapacity;
    }

    for (int32_t partIndex = 0; (partIndex = nextTopLevelArgStart(partIndex)) >= 0;) {
        t->formatAliases[cnt++] = getCachedFormatter(partIndex);
    }

    return (const Format**)formatAliases;
}


UnicodeString MessageFormat::getArgName(int32_t partIndex) {
    const MessagePattern::Part& part = msgPattern.getPart(partIndex);
    return msgPattern.getSubstring(part);
}

StringEnumeration*
MessageFormat::getFormatNames(UErrorCode& status) {
    if (U_FAILURE(status))  return NULL;

    UVector *fFormatNames = new UVector(status);
    if (U_FAILURE(status)) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    fFormatNames->setDeleter(uprv_deleteUObject);

    for (int32_t partIndex = 0; (partIndex = nextTopLevelArgStart(partIndex)) >= 0;) {
        fFormatNames->addElement(new UnicodeString(getArgName(partIndex + 1)), status);
    }

    StringEnumeration* nameEnumerator = new FormatNameEnumeration(fFormatNames, status);
    return nameEnumerator;
}

// -------------------------------------
// Formats the source Formattable array and copy into the result buffer.
// Ignore the FieldPosition result for error checking.

UnicodeString&
MessageFormat::format(const Formattable* source,
                      int32_t cnt,
                      UnicodeString& appendTo,
                      FieldPosition& ignore,
                      UErrorCode& success) const
{
    return format(source, NULL, cnt, appendTo, &ignore, success);
}

// -------------------------------------
// Internally creates a MessageFormat instance based on the
// pattern and formats the arguments Formattable array and
// copy into the appendTo buffer.

UnicodeString&
MessageFormat::format(  const UnicodeString& pattern,
                        const Formattable* arguments,
                        int32_t cnt,
                        UnicodeString& appendTo,
                        UErrorCode& success)
{
    MessageFormat temp(pattern, success);
    return temp.format(arguments, NULL, cnt, appendTo, NULL, success);
}

// -------------------------------------
// Formats the source Formattable object and copy into the
// appendTo buffer.  The Formattable object must be an array
// of Formattable instances, returns error otherwise.

UnicodeString&
MessageFormat::format(const Formattable& source,
                      UnicodeString& appendTo,
                      FieldPosition& ignore,
                      UErrorCode& success) const
{
    if (U_FAILURE(success))
        return appendTo;
    if (source.getType() != Formattable::kArray) {
        success = U_ILLEGAL_ARGUMENT_ERROR;
        return appendTo;
    }
    int32_t cnt;
    const Formattable* tmpPtr = source.getArray(cnt);
    return format(tmpPtr, NULL, cnt, appendTo, &ignore, success);
}

UnicodeString&
MessageFormat::format(const UnicodeString* argumentNames,
                      const Formattable* arguments,
                      int32_t count,
                      UnicodeString& appendTo,
                      UErrorCode& success) const {
    return format(arguments, argumentNames, count, appendTo, NULL, success);
}

// Does linear search to find the match for an ArgName.
const Formattable* MessageFormat::getArgFromListByName(const Formattable* arguments,
                                                       const UnicodeString *argumentNames,
                                                       int32_t cnt, UnicodeString& name) const {
    for (int32_t i = 0; i < cnt; ++i) {
        if (0 == argumentNames[i].compare(name)) {
            return arguments + i;
        }
    }
    return NULL;
}


UnicodeString&
MessageFormat::format(const Formattable* arguments,
                      const UnicodeString *argumentNames,
                      int32_t cnt,
                      UnicodeString& appendTo,
                      FieldPosition* pos,
                      UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }

    UnicodeStringAppendable usapp(appendTo);
    AppendableWrapper app(usapp);
    format(0, NULL, arguments, argumentNames, cnt, app, pos, status);
    return appendTo;
}

namespace {

/**
 * Mutable input/output values for the PluralSelectorProvider.
 * Separate so that it is possible to make MessageFormat Freezable.
 */
class PluralSelectorContext {
public:
    PluralSelectorContext(int32_t start, const UnicodeString &name,
                          const Formattable &num, double off, UErrorCode &errorCode)
            : startIndex(start), argName(name), offset(off),
              numberArgIndex(-1), formatter(NULL), forReplaceNumber(FALSE) {
        // number needs to be set even when select() is not called.
        // Keep it as a Number/Formattable:
        // For format() methods, and to preserve information (e.g., BigDecimal).
        if(off == 0) {
            number = num;
        } else {
            number = num.getDouble(errorCode) - off;
        }
    }

    // Input values for plural selection with decimals.
    int32_t startIndex;
    const UnicodeString &argName;
    /** argument number - plural offset */
    Formattable number;
    double offset;
    // Output values for plural selection with decimals.
    /** -1 if REPLACE_NUMBER, 0 arg not found, >0 ARG_START index */
    int32_t numberArgIndex;
    const Format *formatter;
    /** formatted argument number - plural offset */
    UnicodeString numberString;
    /** TRUE if number-offset was formatted with the stock number formatter */
    UBool forReplaceNumber;
};

}  // namespace

// if argumentNames is NULL, this means arguments is a numeric array.
// arguments can not be NULL.
// We use const void *plNumber rather than const PluralSelectorContext *pluralNumber
// so that we need not declare the PluralSelectorContext in the public header file.
void MessageFormat::format(int32_t msgStart, const void *plNumber,
                           const Formattable* arguments,
                           const UnicodeString *argumentNames,
                           int32_t cnt,
                           AppendableWrapper& appendTo,
                           FieldPosition* ignore,
                           UErrorCode& success) const {
    if (U_FAILURE(success)) {
        return;
    }

    const UnicodeString& msgString = msgPattern.getPatternString();
    int32_t prevIndex = msgPattern.getPart(msgStart).getLimit();
    for (int32_t i = msgStart + 1; U_SUCCESS(success) ; ++i) {
        const MessagePattern::Part* part = &msgPattern.getPart(i);
        const UMessagePatternPartType type = part->getType();
        int32_t index = part->getIndex();
        appendTo.append(msgString, prevIndex, index - prevIndex);
        if (type == UMSGPAT_PART_TYPE_MSG_LIMIT) {
            return;
        }
        prevIndex = part->getLimit();
        if (type == UMSGPAT_PART_TYPE_REPLACE_NUMBER) {
            const PluralSelectorContext &pluralNumber =
                *static_cast<const PluralSelectorContext *>(plNumber);
            if(pluralNumber.forReplaceNumber) {
                // number-offset was already formatted.
                appendTo.formatAndAppend(pluralNumber.formatter,
                        pluralNumber.number, pluralNumber.numberString, success);
            } else {
                const NumberFormat* nf = getDefaultNumberFormat(success);
                appendTo.formatAndAppend(nf, pluralNumber.number, success);
            }
            continue;
        }
        if (type != UMSGPAT_PART_TYPE_ARG_START) {
            continue;
        }
        int32_t argLimit = msgPattern.getLimitPartIndex(i);
        UMessagePatternArgType argType = part->getArgType();
        part = &msgPattern.getPart(++i);
        const Formattable* arg;
        UBool noArg = FALSE;
        UnicodeString argName = msgPattern.getSubstring(*part);
        if (argumentNames == NULL) {
            int32_t argNumber = part->getValue();  // ARG_NUMBER
            if (0 <= argNumber && argNumber < cnt) {
                arg = arguments + argNumber;
            } else {
                arg = NULL;
                noArg = TRUE;
            }
        } else {
            arg = getArgFromListByName(arguments, argumentNames, cnt, argName);
            if (arg == NULL) {
                noArg = TRUE;
            }
        }
        ++i;
        int32_t prevDestLength = appendTo.length();
        const Format* formatter = NULL;
        if (noArg) {
            appendTo.append(
                UnicodeString(MSGFMT_LEFT_CURLY_BRACE).append(argName).append(MSGFMT_RIGHT_CURLY_BRACE));
        } else if (arg == NULL) {
            appendTo.append(NULL_STRING, 4);
        } else if(plNumber!=NULL &&
                static_cast<const PluralSelectorContext *>(plNumber)->numberArgIndex==(i-2)) {
            const PluralSelectorContext &pluralNumber =
                *static_cast<const PluralSelectorContext *>(plNumber);
            if(pluralNumber.offset == 0) {
                // The number was already formatted with this formatter.
                appendTo.formatAndAppend(pluralNumber.formatter, pluralNumber.number,
                                         pluralNumber.numberString, success);
            } else {
                // Do not use the formatted (number-offset) string for a named argument
                // that formats the number without subtracting the offset.
                appendTo.formatAndAppend(pluralNumber.formatter, *arg, success);
            }
        } else if ((formatter = getCachedFormatter(i -2)) != 0) {
            // Handles all ArgType.SIMPLE, and formatters from setFormat() and its siblings.
            if (dynamic_cast<const ChoiceFormat*>(formatter) ||
                dynamic_cast<const PluralFormat*>(formatter) ||
                dynamic_cast<const SelectFormat*>(formatter)) {
                // We only handle nested formats here if they were provided via
                // setFormat() or its siblings. Otherwise they are not cached and instead
                // handled below according to argType.
                UnicodeString subMsgString;
                formatter->format(*arg, subMsgString, success);
                if (subMsgString.indexOf(MSGFMT_LEFT_CURLY_BRACE) >= 0 ||
                    (subMsgString.indexOf(MSGFMT_SINGLE_QUOTE) >= 0 && !MessageImpl::jdkAposMode(msgPattern))
                ) {
                    MessageFormat subMsgFormat(subMsgString, fLocale, success);
                    subMsgFormat.format(0, NULL, arguments, argumentNames, cnt, appendTo, ignore, success);
                } else {
                    appendTo.append(subMsgString);
                }
            } else {
                appendTo.formatAndAppend(formatter, *arg, success);
            }
        } else if (argType == UMSGPAT_ARG_TYPE_NONE || (cachedFormatters && uhash_iget(cachedFormatters, i - 2))) {
            // We arrive here if getCachedFormatter returned NULL, but there was actually an element in the hash table.
            // This can only happen if the hash table contained a DummyFormat, so the if statement above is a check
            // for the hash table containind DummyFormat.
            if (arg->isNumeric()) {
                const NumberFormat* nf = getDefaultNumberFormat(success);
                appendTo.formatAndAppend(nf, *arg, success);
            } else if (arg->getType() == Formattable::kDate) {
                const DateFormat* df = getDefaultDateFormat(success);
                appendTo.formatAndAppend(df, *arg, success);
            } else {
                appendTo.append(arg->getString(success));
            }
        } else if (argType == UMSGPAT_ARG_TYPE_CHOICE) {
            if (!arg->isNumeric()) {
                success = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
            // We must use the Formattable::getDouble() variant with the UErrorCode parameter
            // because only this one converts non-double numeric types to double.
            const double number = arg->getDouble(success);
            int32_t subMsgStart = ChoiceFormat::findSubMessage(msgPattern, i, number);
            formatComplexSubMessage(subMsgStart, NULL, arguments, argumentNames,
                                    cnt, appendTo, success);
        } else if (UMSGPAT_ARG_TYPE_HAS_PLURAL_STYLE(argType)) {
            if (!arg->isNumeric()) {
                success = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
            const PluralSelectorProvider &selector =
                argType == UMSGPAT_ARG_TYPE_PLURAL ? pluralProvider : ordinalProvider;
            // We must use the Formattable::getDouble() variant with the UErrorCode parameter
            // because only this one converts non-double numeric types to double.
            double offset = msgPattern.getPluralOffset(i);
            PluralSelectorContext context(i, argName, *arg, offset, success);
            int32_t subMsgStart = PluralFormat::findSubMessage(
                    msgPattern, i, selector, &context, arg->getDouble(success), success);
            formatComplexSubMessage(subMsgStart, &context, arguments, argumentNames,
                                    cnt, appendTo, success);
        } else if (argType == UMSGPAT_ARG_TYPE_SELECT) {
            int32_t subMsgStart = SelectFormat::findSubMessage(msgPattern, i, arg->getString(success), success);
            formatComplexSubMessage(subMsgStart, NULL, arguments, argumentNames,
                                    cnt, appendTo, success);
        } else {
            // This should never happen.
            success = U_INTERNAL_PROGRAM_ERROR;
            return;
        }
        ignore = updateMetaData(appendTo, prevDestLength, ignore, arg);
        prevIndex = msgPattern.getPart(argLimit).getLimit();
        i = argLimit;
    }
}


void MessageFormat::formatComplexSubMessage(int32_t msgStart,
                                            const void *plNumber,
                                            const Formattable* arguments,
                                            const UnicodeString *argumentNames,
                                            int32_t cnt,
                                            AppendableWrapper& appendTo,
                                            UErrorCode& success) const {
    if (U_FAILURE(success)) {
        return;
    }

    if (!MessageImpl::jdkAposMode(msgPattern)) {
        format(msgStart, plNumber, arguments, argumentNames, cnt, appendTo, NULL, success);
        return;
    }

    // JDK compatibility mode: (see JDK MessageFormat.format() API docs)
    // - remove SKIP_SYNTAX; that is, remove half of the apostrophes
    // - if the result string contains an open curly brace '{' then
    //   instantiate a temporary MessageFormat object and format again;
    //   otherwise just append the result string
    const UnicodeString& msgString = msgPattern.getPatternString();
    UnicodeString sb;
    int32_t prevIndex = msgPattern.getPart(msgStart).getLimit();
    for (int32_t i = msgStart;;) {
        const MessagePattern::Part& part = msgPattern.getPart(++i);
        const UMessagePatternPartType type = part.getType();
        int32_t index = part.getIndex();
        if (type == UMSGPAT_PART_TYPE_MSG_LIMIT) {
            sb.append(msgString, prevIndex, index - prevIndex);
            break;
        } else if (type == UMSGPAT_PART_TYPE_REPLACE_NUMBER || type == UMSGPAT_PART_TYPE_SKIP_SYNTAX) {
            sb.append(msgString, prevIndex, index - prevIndex);
            if (type == UMSGPAT_PART_TYPE_REPLACE_NUMBER) {
                const PluralSelectorContext &pluralNumber =
                    *static_cast<const PluralSelectorContext *>(plNumber);
                if(pluralNumber.forReplaceNumber) {
                    // number-offset was already formatted.
                    sb.append(pluralNumber.numberString);
                } else {
                    const NumberFormat* nf = getDefaultNumberFormat(success);
                    sb.append(nf->format(pluralNumber.number, sb, success));
                }
            }
            prevIndex = part.getLimit();
        } else if (type == UMSGPAT_PART_TYPE_ARG_START) {
            sb.append(msgString, prevIndex, index - prevIndex);
            prevIndex = index;
            i = msgPattern.getLimitPartIndex(i);
            index = msgPattern.getPart(i).getLimit();
            MessageImpl::appendReducedApostrophes(msgString, prevIndex, index, sb);
            prevIndex = index;
        }
    }
    if (sb.indexOf(MSGFMT_LEFT_CURLY_BRACE) >= 0) {
        UnicodeString emptyPattern;  // gcc 3.3.3 fails with "UnicodeString()" as the first parameter.
        MessageFormat subMsgFormat(emptyPattern, fLocale, success);
        subMsgFormat.applyPattern(sb, UMSGPAT_APOS_DOUBLE_REQUIRED, NULL, success);
        subMsgFormat.format(0, NULL, arguments, argumentNames, cnt, appendTo, NULL, success);
    } else {
        appendTo.append(sb);
    }
}


UnicodeString MessageFormat::getLiteralStringUntilNextArgument(int32_t from) const {
    const UnicodeString& msgString=msgPattern.getPatternString();
    int32_t prevIndex=msgPattern.getPart(from).getLimit();
    UnicodeString b;
    for (int32_t i = from + 1; ; ++i) {
        const MessagePattern::Part& part = msgPattern.getPart(i);
        const UMessagePatternPartType type=part.getType();
        int32_t index=part.getIndex();
        b.append(msgString, prevIndex, index - prevIndex);
        if(type==UMSGPAT_PART_TYPE_ARG_START || type==UMSGPAT_PART_TYPE_MSG_LIMIT) {
            return b;
        }
        // Unexpected Part "part" in parsed message.
        U_ASSERT(type==UMSGPAT_PART_TYPE_SKIP_SYNTAX || type==UMSGPAT_PART_TYPE_INSERT_CHAR);
        prevIndex=part.getLimit();
    }
}


FieldPosition* MessageFormat::updateMetaData(AppendableWrapper& /*dest*/, int32_t /*prevLength*/,
                             FieldPosition* /*fp*/, const Formattable* /*argId*/) const {
    // Unlike in Java, there are no field attributes defined for MessageFormat. Do nothing.
    return NULL;
    /*
      if (fp != NULL && Field.ARGUMENT.equals(fp.getFieldAttribute())) {
          fp->setBeginIndex(prevLength);
          fp->setEndIndex(dest.get_length());
          return NULL;
      }
      return fp;
    */
}

int32_t
MessageFormat::findOtherSubMessage(int32_t partIndex) const {
    int32_t count=msgPattern.countParts();
    const MessagePattern::Part *part = &msgPattern.getPart(partIndex);
    if(MessagePattern::Part::hasNumericValue(part->getType())) {
        ++partIndex;
    }
    // Iterate over (ARG_SELECTOR [ARG_INT|ARG_DOUBLE] message) tuples
    // until ARG_LIMIT or end of plural-only pattern.
    UnicodeString other(FALSE, OTHER_STRING, 5);
    do {
        part=&msgPattern.getPart(partIndex++);
        UMessagePatternPartType type=part->getType();
        if(type==UMSGPAT_PART_TYPE_ARG_LIMIT) {
            break;
        }
        U_ASSERT(type==UMSGPAT_PART_TYPE_ARG_SELECTOR);
        // part is an ARG_SELECTOR followed by an optional explicit value, and then a message
        if(msgPattern.partSubstringMatches(*part, other)) {
            return partIndex;
        }
        if(MessagePattern::Part::hasNumericValue(msgPattern.getPartType(partIndex))) {
            ++partIndex;  // skip the numeric-value part of "=1" etc.
        }
        partIndex=msgPattern.getLimitPartIndex(partIndex);
    } while(++partIndex<count);
    return 0;
}

int32_t
MessageFormat::findFirstPluralNumberArg(int32_t msgStart, const UnicodeString &argName) const {
    for(int32_t i=msgStart+1;; ++i) {
        const MessagePattern::Part &part=msgPattern.getPart(i);
        UMessagePatternPartType type=part.getType();
        if(type==UMSGPAT_PART_TYPE_MSG_LIMIT) {
            return 0;
        }
        if(type==UMSGPAT_PART_TYPE_REPLACE_NUMBER) {
            return -1;
        }
        if(type==UMSGPAT_PART_TYPE_ARG_START) {
            UMessagePatternArgType argType=part.getArgType();
            if(!argName.isEmpty() && (argType==UMSGPAT_ARG_TYPE_NONE || argType==UMSGPAT_ARG_TYPE_SIMPLE)) {
                // ARG_NUMBER or ARG_NAME
                if(msgPattern.partSubstringMatches(msgPattern.getPart(i+1), argName)) {
                    return i;
                }
            }
            i=msgPattern.getLimitPartIndex(i);
        }
    }
}

void MessageFormat::copyObjects(const MessageFormat& that, UErrorCode& ec) {
    // Deep copy pointer fields.
    // We need not copy the formatAliases because they are re-filled
    // in each getFormats() call.
    // The defaultNumberFormat, defaultDateFormat and pluralProvider.rules
    // also get created on demand.
    argTypeCount = that.argTypeCount;
    if (argTypeCount > 0) {
        if (!allocateArgTypes(argTypeCount, ec)) {
            return;
        }
        uprv_memcpy(argTypes, that.argTypes, argTypeCount * sizeof(argTypes[0]));
    }
    if (cachedFormatters != NULL) {
        uhash_removeAll(cachedFormatters);
    }
    if (customFormatArgStarts != NULL) {
        uhash_removeAll(customFormatArgStarts);
    }
    if (that.cachedFormatters) {
        if (cachedFormatters == NULL) {
            cachedFormatters=uhash_open(uhash_hashLong, uhash_compareLong,
                                        equalFormatsForHash, &ec);
            if (U_FAILURE(ec)) {
                return;
            }
            uhash_setValueDeleter(cachedFormatters, uprv_deleteUObject);
        }

        const int32_t count = uhash_count(that.cachedFormatters);
        int32_t pos, idx;
        for (idx = 0, pos = UHASH_FIRST; idx < count && U_SUCCESS(ec); ++idx) {
            const UHashElement* cur = uhash_nextElement(that.cachedFormatters, &pos);
            Format* newFormat = ((Format*)(cur->value.pointer))->clone();
            if (newFormat) {
                uhash_iput(cachedFormatters, cur->key.integer, newFormat, &ec);
            } else {
                ec = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
        }
    }
    if (that.customFormatArgStarts) {
        if (customFormatArgStarts == NULL) {
            customFormatArgStarts=uhash_open(uhash_hashLong, uhash_compareLong,
                                              NULL, &ec);
        }
        const int32_t count = uhash_count(that.customFormatArgStarts);
        int32_t pos, idx;
        for (idx = 0, pos = UHASH_FIRST; idx < count && U_SUCCESS(ec); ++idx) {
            const UHashElement* cur = uhash_nextElement(that.customFormatArgStarts, &pos);
            uhash_iputi(customFormatArgStarts, cur->key.integer, cur->value.integer, &ec);
        }
    }
}


Formattable*
MessageFormat::parse(int32_t msgStart,
                     const UnicodeString& source,
                     ParsePosition& pos,
                     int32_t& count,
                     UErrorCode& ec) const {
    count = 0;
    if (U_FAILURE(ec)) {
        pos.setErrorIndex(pos.getIndex());
        return NULL;
    }
    // parse() does not work with named arguments.
    if (msgPattern.hasNamedArguments()) {
        ec = U_ARGUMENT_TYPE_MISMATCH;
        pos.setErrorIndex(pos.getIndex());
        return NULL;
    }
    LocalArray<Formattable> resultArray(new Formattable[argTypeCount ? argTypeCount : 1]);
    const UnicodeString& msgString=msgPattern.getPatternString();
    int32_t prevIndex=msgPattern.getPart(msgStart).getLimit();
    int32_t sourceOffset = pos.getIndex();
    ParsePosition tempStatus(0);

    for(int32_t i=msgStart+1; ; ++i) {
        UBool haveArgResult = FALSE;
        const MessagePattern::Part* part=&msgPattern.getPart(i);
        const UMessagePatternPartType type=part->getType();
        int32_t index=part->getIndex();
        // Make sure the literal string matches.
        int32_t len = index - prevIndex;
        if (len == 0 || (0 == msgString.compare(prevIndex, len, source, sourceOffset, len))) {
            sourceOffset += len;
            prevIndex += len;
        } else {
            pos.setErrorIndex(sourceOffset);
            return NULL; // leave index as is to signal error
        }
        if(type==UMSGPAT_PART_TYPE_MSG_LIMIT) {
            // Things went well! Done.
            pos.setIndex(sourceOffset);
            return resultArray.orphan();
        }
        if(type==UMSGPAT_PART_TYPE_SKIP_SYNTAX || type==UMSGPAT_PART_TYPE_INSERT_CHAR) {
            prevIndex=part->getLimit();
            continue;
        }
        // We do not support parsing Plural formats. (No REPLACE_NUMBER here.)
        // Unexpected Part "part" in parsed message.
        U_ASSERT(type==UMSGPAT_PART_TYPE_ARG_START);
        int32_t argLimit=msgPattern.getLimitPartIndex(i);

        UMessagePatternArgType argType=part->getArgType();
        part=&msgPattern.getPart(++i);
        int32_t argNumber = part->getValue();  // ARG_NUMBER
        UnicodeString key;
        ++i;
        const Format* formatter = NULL;
        Formattable& argResult = resultArray[argNumber];

        if(cachedFormatters!=NULL && (formatter = getCachedFormatter(i - 2))!=NULL) {
            // Just parse using the formatter.
            tempStatus.setIndex(sourceOffset);
            formatter->parseObject(source, argResult, tempStatus);
            if (tempStatus.getIndex() == sourceOffset) {
                pos.setErrorIndex(sourceOffset);
                return NULL; // leave index as is to signal error
            }
            sourceOffset = tempStatus.getIndex();
            haveArgResult = TRUE;
        } else if(
            argType==UMSGPAT_ARG_TYPE_NONE || (cachedFormatters && uhash_iget(cachedFormatters, i -2))) {
            // We arrive here if getCachedFormatter returned NULL, but there was actually an element in the hash table.
            // This can only happen if the hash table contained a DummyFormat, so the if statement above is a check
            // for the hash table containind DummyFormat.

            // Match as a string.
            // if at end, use longest possible match
            // otherwise uses first match to intervening string
            // does NOT recursively try all possibilities
            UnicodeString stringAfterArgument = getLiteralStringUntilNextArgument(argLimit);
            int32_t next;
            if (!stringAfterArgument.isEmpty()) {
                next = source.indexOf(stringAfterArgument, sourceOffset);
            } else {
                next = source.length();
            }
            if (next < 0) {
                pos.setErrorIndex(sourceOffset);
                return NULL; // leave index as is to signal error
            } else {
                UnicodeString strValue(source.tempSubString(sourceOffset, next - sourceOffset));
                UnicodeString compValue;
                compValue.append(MSGFMT_LEFT_CURLY_BRACE);
                itos(argNumber, compValue);
                compValue.append(MSGFMT_RIGHT_CURLY_BRACE);
                if (0 != strValue.compare(compValue)) {
                    argResult.setString(strValue);
                    haveArgResult = TRUE;
                }
                sourceOffset = next;
            }
        } else if(argType==UMSGPAT_ARG_TYPE_CHOICE) {
            tempStatus.setIndex(sourceOffset);
            double choiceResult = ChoiceFormat::parseArgument(msgPattern, i, source, tempStatus);
            if (tempStatus.getIndex() == sourceOffset) {
                pos.setErrorIndex(sourceOffset);
                return NULL; // leave index as is to signal error
            }
            argResult.setDouble(choiceResult);
            haveArgResult = TRUE;
            sourceOffset = tempStatus.getIndex();
        } else if(UMSGPAT_ARG_TYPE_HAS_PLURAL_STYLE(argType) || argType==UMSGPAT_ARG_TYPE_SELECT) {
            // Parsing not supported.
            ec = U_UNSUPPORTED_ERROR;
            return NULL;
        } else {
            // This should never happen.
            ec = U_INTERNAL_PROGRAM_ERROR;
            return NULL;
        }
        if (haveArgResult && count <= argNumber) {
            count = argNumber + 1;
        }
        prevIndex=msgPattern.getPart(argLimit).getLimit();
        i=argLimit;
    }
}
// -------------------------------------
// Parses the source pattern and returns the Formattable objects array,
// the array count and the ending parse position.  The caller of this method
// owns the array.

Formattable*
MessageFormat::parse(const UnicodeString& source,
                     ParsePosition& pos,
                     int32_t& count) const {
    UErrorCode ec = U_ZERO_ERROR;
    return parse(0, source, pos, count, ec);
}

// -------------------------------------
// Parses the source string and returns the array of
// Formattable objects and the array count.  The caller
// owns the returned array.

Formattable*
MessageFormat::parse(const UnicodeString& source,
                     int32_t& cnt,
                     UErrorCode& success) const
{
    if (msgPattern.hasNamedArguments()) {
        success = U_ARGUMENT_TYPE_MISMATCH;
        return NULL;
    }
    ParsePosition status(0);
    // Calls the actual implementation method and starts
    // from zero offset of the source text.
    Formattable* result = parse(source, status, cnt);
    if (status.getIndex() == 0) {
        success = U_MESSAGE_PARSE_ERROR;
        delete[] result;
        return NULL;
    }
    return result;
}

// -------------------------------------
// Parses the source text and copy into the result buffer.

void
MessageFormat::parseObject( const UnicodeString& source,
                            Formattable& result,
                            ParsePosition& status) const
{
    int32_t cnt = 0;
    Formattable* tmpResult = parse(source, status, cnt);
    if (tmpResult != NULL)
        result.adoptArray(tmpResult, cnt);
}

UnicodeString
MessageFormat::autoQuoteApostrophe(const UnicodeString& pattern, UErrorCode& status) {
    UnicodeString result;
    if (U_SUCCESS(status)) {
        int32_t plen = pattern.length();
        const UChar* pat = pattern.getBuffer();
        int32_t blen = plen * 2 + 1; // space for null termination, convenience
        UChar* buf = result.getBuffer(blen);
        if (buf == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
        } else {
            int32_t len = umsg_autoQuoteApostrophe(pat, plen, buf, blen, &status);
            result.releaseBuffer(U_SUCCESS(status) ? len : 0);
        }
    }
    if (U_FAILURE(status)) {
        result.setToBogus();
    }
    return result;
}

// -------------------------------------

static Format* makeRBNF(URBNFRuleSetTag tag, const Locale& locale, const UnicodeString& defaultRuleSet, UErrorCode& ec) {
    RuleBasedNumberFormat* fmt = new RuleBasedNumberFormat(tag, locale, ec);
    if (fmt == NULL) {
        ec = U_MEMORY_ALLOCATION_ERROR;
    } else if (U_SUCCESS(ec) && defaultRuleSet.length() > 0) {
        UErrorCode localStatus = U_ZERO_ERROR; // ignore unrecognized default rule set
        fmt->setDefaultRuleSet(defaultRuleSet, localStatus);
    }
    return fmt;
}

void MessageFormat::cacheExplicitFormats(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }

    if (cachedFormatters != NULL) {
        uhash_removeAll(cachedFormatters);
    }
    if (customFormatArgStarts != NULL) {
        uhash_removeAll(customFormatArgStarts);
    }

    // The last two "parts" can at most be ARG_LIMIT and MSG_LIMIT
    // which we need not examine.
    int32_t limit = msgPattern.countParts() - 2;
    argTypeCount = 0;
    // We also need not look at the first two "parts"
    // (at most MSG_START and ARG_START) in this loop.
    // We determine the argTypeCount first so that we can allocateArgTypes
    // so that the next loop can set argTypes[argNumber].
    // (This is for the C API which needs the argTypes to read its va_arg list.)
    for (int32_t i = 2; i < limit && U_SUCCESS(status); ++i) {
        const MessagePattern::Part& part = msgPattern.getPart(i);
        if (part.getType() == UMSGPAT_PART_TYPE_ARG_NUMBER) {
            const int argNumber = part.getValue();
            if (argNumber >= argTypeCount) {
                argTypeCount = argNumber + 1;
            }
        }
    }
    if (!allocateArgTypes(argTypeCount, status)) {
        return;
    }
    // Set all argTypes to kObject, as a "none" value, for lack of any better value.
    // We never use kObject for real arguments.
    // We use it as "no argument yet" for the check for hasArgTypeConflicts.
    for (int32_t i = 0; i < argTypeCount; ++i) {
        argTypes[i] = Formattable::kObject;
    }
    hasArgTypeConflicts = FALSE;

    // This loop starts at part index 1 because we do need to examine
    // ARG_START parts. (But we can ignore the MSG_START.)
    for (int32_t i = 1; i < limit && U_SUCCESS(status); ++i) {
        const MessagePattern::Part* part = &msgPattern.getPart(i);
        if (part->getType() != UMSGPAT_PART_TYPE_ARG_START) {
            continue;
        }
        UMessagePatternArgType argType = part->getArgType();

        int32_t argNumber = -1;
        part = &msgPattern.getPart(i + 1);
        if (part->getType() == UMSGPAT_PART_TYPE_ARG_NUMBER) {
            argNumber = part->getValue();
        }
        Formattable::Type formattableType;

        switch (argType) {
        case UMSGPAT_ARG_TYPE_NONE:
            formattableType = Formattable::kString;
            break;
        case UMSGPAT_ARG_TYPE_SIMPLE: {
            int32_t index = i;
            i += 2;
            UnicodeString explicitType = msgPattern.getSubstring(msgPattern.getPart(i++));
            UnicodeString style;
            if ((part = &msgPattern.getPart(i))->getType() == UMSGPAT_PART_TYPE_ARG_STYLE) {
                style = msgPattern.getSubstring(*part);
                ++i;
            }
            UParseError parseError;
            Format* formatter = createAppropriateFormat(explicitType, style, formattableType, parseError, status);
            setArgStartFormat(index, formatter, status);
            break;
        }
        case UMSGPAT_ARG_TYPE_CHOICE:
        case UMSGPAT_ARG_TYPE_PLURAL:
        case UMSGPAT_ARG_TYPE_SELECTORDINAL:
            formattableType = Formattable::kDouble;
            break;
        case UMSGPAT_ARG_TYPE_SELECT:
            formattableType = Formattable::kString;
            break;
        default:
            status = U_INTERNAL_PROGRAM_ERROR;  // Should be unreachable.
            formattableType = Formattable::kString;
            break;
        }
        if (argNumber != -1) {
            if (argTypes[argNumber] != Formattable::kObject && argTypes[argNumber] != formattableType) {
                hasArgTypeConflicts = TRUE;
            }
            argTypes[argNumber] = formattableType;
        }
    }
}

Format* MessageFormat::createAppropriateFormat(UnicodeString& type, UnicodeString& style,
                                               Formattable::Type& formattableType, UParseError& parseError,
                                               UErrorCode& ec) {
    if (U_FAILURE(ec)) {
        return NULL;
    }
    Format* fmt = NULL;
    int32_t typeID, styleID;
    DateFormat::EStyle date_style;
    int32_t firstNonSpace;

    switch (typeID = findKeyword(type, TYPE_IDS)) {
    case 0: // number
        formattableType = Formattable::kDouble;
        switch (findKeyword(style, NUMBER_STYLE_IDS)) {
        case 0: // default
            fmt = NumberFormat::createInstance(fLocale, ec);
            break;
        case 1: // currency
            fmt = NumberFormat::createCurrencyInstance(fLocale, ec);
            break;
        case 2: // percent
            fmt = NumberFormat::createPercentInstance(fLocale, ec);
            break;
        case 3: // integer
            formattableType = Formattable::kLong;
            fmt = createIntegerFormat(fLocale, ec);
            break;
        default: // pattern or skeleton
            firstNonSpace = PatternProps::skipWhiteSpace(style, 0);
            if (style.compare(firstNonSpace, 2, u"::", 0, 2) == 0) {
                // Skeleton
                UnicodeString skeleton = style.tempSubString(firstNonSpace + 2);
                fmt = number::NumberFormatter::forSkeleton(skeleton, ec).locale(fLocale).toFormat(ec);
            } else {
                // Pattern
                fmt = NumberFormat::createInstance(fLocale, ec);
                if (fmt) {
                    auto* decfmt = dynamic_cast<DecimalFormat*>(fmt);
                    if (decfmt != nullptr) {
                        decfmt->applyPattern(style, parseError, ec);
                    }
                }
            }
            break;
        }
        break;

    case 1: // date
    case 2: // time
        formattableType = Formattable::kDate;
        firstNonSpace = PatternProps::skipWhiteSpace(style, 0);
        if (style.compare(firstNonSpace, 2, u"::", 0, 2) == 0) {
            // Skeleton
            UnicodeString skeleton = style.tempSubString(firstNonSpace + 2);
            fmt = DateFormat::createInstanceForSkeleton(skeleton, fLocale, ec);
        } else {
            // Pattern
            styleID = findKeyword(style, DATE_STYLE_IDS);
            date_style = (styleID >= 0) ? DATE_STYLES[styleID] : DateFormat::kDefault;

            if (typeID == 1) {
                fmt = DateFormat::createDateInstance(date_style, fLocale);
            } else {
                fmt = DateFormat::createTimeInstance(date_style, fLocale);
            }

            if (styleID < 0 && fmt != NULL) {
                SimpleDateFormat* sdtfmt = dynamic_cast<SimpleDateFormat*>(fmt);
                if (sdtfmt != NULL) {
                    sdtfmt->applyPattern(style);
                }
            }
        }
        break;

    case 3: // spellout
        formattableType = Formattable::kDouble;
        fmt = makeRBNF(URBNF_SPELLOUT, fLocale, style, ec);
        break;
    case 4: // ordinal
        formattableType = Formattable::kDouble;
        fmt = makeRBNF(URBNF_ORDINAL, fLocale, style, ec);
        break;
    case 5: // duration
        formattableType = Formattable::kDouble;
        fmt = makeRBNF(URBNF_DURATION, fLocale, style, ec);
        break;
    default:
        formattableType = Formattable::kString;
        ec = U_ILLEGAL_ARGUMENT_ERROR;
        break;
    }

    return fmt;
}


//-------------------------------------
// Finds the string, s, in the string array, list.
int32_t MessageFormat::findKeyword(const UnicodeString& s,
                                   const UChar * const *list)
{
    if (s.isEmpty()) {
        return 0; // default
    }

    int32_t length = s.length();
    const UChar *ps = PatternProps::trimWhiteSpace(s.getBuffer(), length);
    UnicodeString buffer(FALSE, ps, length);
    // Trims the space characters and turns all characters
    // in s to lower case.
    buffer.toLower("");
    for (int32_t i = 0; list[i]; ++i) {
        if (!buffer.compare(list[i], u_strlen(list[i]))) {
            return i;
        }
    }
    return -1;
}

/**
 * Convenience method that ought to be in NumberFormat
 */
NumberFormat*
MessageFormat::createIntegerFormat(const Locale& locale, UErrorCode& status) const {
    NumberFormat *temp = NumberFormat::createInstance(locale, status);
    DecimalFormat *temp2;
    if (temp != NULL && (temp2 = dynamic_cast<DecimalFormat*>(temp)) != NULL) {
        temp2->setMaximumFractionDigits(0);
        temp2->setDecimalSeparatorAlwaysShown(FALSE);
        temp2->setParseIntegerOnly(TRUE);
    }

    return temp;
}

/**
 * Return the default number format.  Used to format a numeric
 * argument when subformats[i].format is NULL.  Returns NULL
 * on failure.
 *
 * Semantically const but may modify *this.
 */
const NumberFormat* MessageFormat::getDefaultNumberFormat(UErrorCode& ec) const {
    if (defaultNumberFormat == NULL) {
        MessageFormat* t = (MessageFormat*) this;
        t->defaultNumberFormat = NumberFormat::createInstance(fLocale, ec);
        if (U_FAILURE(ec)) {
            delete t->defaultNumberFormat;
            t->defaultNumberFormat = NULL;
        } else if (t->defaultNumberFormat == NULL) {
            ec = U_MEMORY_ALLOCATION_ERROR;
        }
    }
    return defaultNumberFormat;
}

/**
 * Return the default date format.  Used to format a date
 * argument when subformats[i].format is NULL.  Returns NULL
 * on failure.
 *
 * Semantically const but may modify *this.
 */
const DateFormat* MessageFormat::getDefaultDateFormat(UErrorCode& ec) const {
    if (defaultDateFormat == NULL) {
        MessageFormat* t = (MessageFormat*) this;
        t->defaultDateFormat = DateFormat::createDateTimeInstance(DateFormat::kShort, DateFormat::kShort, fLocale);
        if (t->defaultDateFormat == NULL) {
            ec = U_MEMORY_ALLOCATION_ERROR;
        }
    }
    return defaultDateFormat;
}

UBool
MessageFormat::usesNamedArguments() const {
    return msgPattern.hasNamedArguments();
}

int32_t
MessageFormat::getArgTypeCount() const {
    return argTypeCount;
}

UBool MessageFormat::equalFormats(const void* left, const void* right) {
    return *(const Format*)left==*(const Format*)right;
}


UBool MessageFormat::DummyFormat::operator==(const Format&) const {
    return TRUE;
}

MessageFormat::DummyFormat* MessageFormat::DummyFormat::clone() const {
    return new DummyFormat();
}

UnicodeString& MessageFormat::DummyFormat::format(const Formattable&,
                          UnicodeString& appendTo,
                          UErrorCode& status) const {
    if (U_SUCCESS(status)) {
        status = U_UNSUPPORTED_ERROR;
    }
    return appendTo;
}

UnicodeString& MessageFormat::DummyFormat::format(const Formattable&,
                          UnicodeString& appendTo,
                          FieldPosition&,
                          UErrorCode& status) const {
    if (U_SUCCESS(status)) {
        status = U_UNSUPPORTED_ERROR;
    }
    return appendTo;
}

UnicodeString& MessageFormat::DummyFormat::format(const Formattable&,
                          UnicodeString& appendTo,
                          FieldPositionIterator*,
                          UErrorCode& status) const {
    if (U_SUCCESS(status)) {
        status = U_UNSUPPORTED_ERROR;
    }
    return appendTo;
}

void MessageFormat::DummyFormat::parseObject(const UnicodeString&,
                                                     Formattable&,
                                                     ParsePosition& ) const {
}


FormatNameEnumeration::FormatNameEnumeration(UVector *fNameList, UErrorCode& /*status*/) {
    pos=0;
    fFormatNames = fNameList;
}

const UnicodeString*
FormatNameEnumeration::snext(UErrorCode& status) {
    if (U_SUCCESS(status) && pos < fFormatNames->size()) {
        return (const UnicodeString*)fFormatNames->elementAt(pos++);
    }
    return NULL;
}

void
FormatNameEnumeration::reset(UErrorCode& /*status*/) {
    pos=0;
}

int32_t
FormatNameEnumeration::count(UErrorCode& /*status*/) const {
    return (fFormatNames==NULL) ? 0 : fFormatNames->size();
}

FormatNameEnumeration::~FormatNameEnumeration() {
    delete fFormatNames;
}

MessageFormat::PluralSelectorProvider::PluralSelectorProvider(const MessageFormat &mf, UPluralType t)
        : msgFormat(mf), rules(NULL), type(t) {
}

MessageFormat::PluralSelectorProvider::~PluralSelectorProvider() {
    delete rules;
}

UnicodeString MessageFormat::PluralSelectorProvider::select(void *ctx, double number,
                                                            UErrorCode& ec) const {
    if (U_FAILURE(ec)) {
        return UnicodeString(FALSE, OTHER_STRING, 5);
    }
    MessageFormat::PluralSelectorProvider* t = const_cast<MessageFormat::PluralSelectorProvider*>(this);
    if(rules == NULL) {
        t->rules = PluralRules::forLocale(msgFormat.fLocale, type, ec);
        if (U_FAILURE(ec)) {
            return UnicodeString(FALSE, OTHER_STRING, 5);
        }
    }
    // Select a sub-message according to how the number is formatted,
    // which is specified in the selected sub-message.
    // We avoid this circle by looking at how
    // the number is formatted in the "other" sub-message
    // which must always be present and usually contains the number.
    // Message authors should be consistent across sub-messages.
    PluralSelectorContext &context = *static_cast<PluralSelectorContext *>(ctx);
    int32_t otherIndex = msgFormat.findOtherSubMessage(context.startIndex);
    context.numberArgIndex = msgFormat.findFirstPluralNumberArg(otherIndex, context.argName);
    if(context.numberArgIndex > 0 && msgFormat.cachedFormatters != NULL) {
        context.formatter =
            (const Format*)uhash_iget(msgFormat.cachedFormatters, context.numberArgIndex);
    }
    if(context.formatter == NULL) {
        context.formatter = msgFormat.getDefaultNumberFormat(ec);
        context.forReplaceNumber = TRUE;
    }
    if (context.number.getDouble(ec) != number) {
        ec = U_INTERNAL_PROGRAM_ERROR;
        return UnicodeString(FALSE, OTHER_STRING, 5);
    }
    context.formatter->format(context.number, context.numberString, ec);
    auto* decFmt = dynamic_cast<const DecimalFormat *>(context.formatter);
    if(decFmt != NULL) {
        number::impl::DecimalQuantity dq;
        decFmt->formatToDecimalQuantity(context.number, dq, ec);
        if (U_FAILURE(ec)) {
            return UnicodeString(FALSE, OTHER_STRING, 5);
        }
        return rules->select(dq);
    } else {
        return rules->select(number);
    }
}

void MessageFormat::PluralSelectorProvider::reset() {
    delete rules;
    rules = NULL;
}


U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 1997-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*   file name:  nfrs.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
* Modification history
* Date        Name      Comments
* 10/11/2001  Doug      Ported from ICU4J
*/

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 1997-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*   file name:  nfrs.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
* Modification history
* Date        Name      Comments
* 10/11/2001  Doug      Ported from ICU4J
*/





#if U_HAVE_RBNF




// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 1997-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*   file name:  nfrlist.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
* Modification history
* Date        Name      Comments
* 10/11/2001  Doug      Ported from ICU4J
*/




#if U_HAVE_RBNF


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2015, International Business Machines
* Corporation and others. All Rights Reserved.
*******************************************************************************
*/




#if U_HAVE_RBNF





U_NAMESPACE_BEGIN

class FieldPosition;
class Formattable;
class NFRuleList;
class NFRuleSet;
class NFSubstitution;
class ParsePosition;
class PluralFormat;
class RuleBasedNumberFormat;
class UnicodeString;

class NFRule : public UMemory {
public:

    enum ERuleType {
        kNoBase = 0,
        kNegativeNumberRule = -1,
        kImproperFractionRule = -2,
        kProperFractionRule = -3,
        kMasterRule = -4,
        kInfinityRule = -5,
        kNaNRule = -6,
        kOtherRule = -7
    };

    static void makeRules(UnicodeString& definition,
                          NFRuleSet* ruleSet, 
                          const NFRule* predecessor, 
                          const RuleBasedNumberFormat* rbnf, 
                          NFRuleList& ruleList,
                          UErrorCode& status);

    NFRule(const RuleBasedNumberFormat* rbnf, const UnicodeString &ruleText, UErrorCode &status);
    ~NFRule();

    UBool operator==(const NFRule& rhs) const;
    UBool operator!=(const NFRule& rhs) const { return !operator==(rhs); }

    ERuleType getType() const { return (ERuleType)(baseValue <= kNoBase ? (ERuleType)baseValue : kOtherRule); }
    void setType(ERuleType ruleType) { baseValue = (int32_t)ruleType; }

    int64_t getBaseValue() const { return baseValue; }
    void setBaseValue(int64_t value, UErrorCode& status);

    UChar getDecimalPoint() const { return decimalPoint; }

    int64_t getDivisor() const;

    void doFormat(int64_t number, UnicodeString& toAppendTo, int32_t pos, int32_t recursionCount, UErrorCode& status) const;
    void doFormat(double  number, UnicodeString& toAppendTo, int32_t pos, int32_t recursionCount, UErrorCode& status) const;

    UBool doParse(const UnicodeString& text, 
                  ParsePosition& pos, 
                  UBool isFractional, 
                  double upperBound,
                  uint32_t nonNumericalExecutedRuleMask,
                  Formattable& result) const;

    UBool shouldRollBack(int64_t number) const;

    void _appendRuleText(UnicodeString& result) const;

    int32_t findTextLenient(const UnicodeString& str, const UnicodeString& key, 
                     int32_t startingAt, int32_t* resultCount) const;

    void setDecimalFormatSymbols(const DecimalFormatSymbols &newSymbols, UErrorCode& status);

private:
    void parseRuleDescriptor(UnicodeString& descriptor, UErrorCode& status);
    void extractSubstitutions(const NFRuleSet* ruleSet, const UnicodeString &ruleText, const NFRule* predecessor, UErrorCode& status);
    NFSubstitution* extractSubstitution(const NFRuleSet* ruleSet, const NFRule* predecessor, UErrorCode& status);
    
    int16_t expectedExponent() const;
    int32_t indexOfAnyRulePrefix() const;
    double matchToDelimiter(const UnicodeString& text, int32_t startPos, double baseValue,
                            const UnicodeString& delimiter, ParsePosition& pp, const NFSubstitution* sub, 
                            uint32_t nonNumericalExecutedRuleMask,
                            double upperBound) const;
    void stripPrefix(UnicodeString& text, const UnicodeString& prefix, ParsePosition& pp) const;

    int32_t prefixLength(const UnicodeString& str, const UnicodeString& prefix, UErrorCode& status) const;
    UBool allIgnorable(const UnicodeString& str, UErrorCode& status) const;
    int32_t findText(const UnicodeString& str, const UnicodeString& key, 
                     int32_t startingAt, int32_t* resultCount) const;

private:
    int64_t baseValue;
    int32_t radix;
    int16_t exponent;
    UChar decimalPoint;
    UnicodeString fRuleText;
    NFSubstitution* sub1;
    NFSubstitution* sub2;
    const RuleBasedNumberFormat* formatter;
    const PluralFormat* rulePatternFormat;

    NFRule(const NFRule &other); // forbid copying of this class
    NFRule &operator=(const NFRule &other); // forbid copying of this class
};

U_NAMESPACE_END

/* U_HAVE_RBNF */
#endif

// NFRULE_H




U_NAMESPACE_BEGIN

// unsafe class for internal use only.  assume memory allocations succeed, indexes are valid.
// should be a template, but we can't use them

class NFRuleList : public UMemory {
protected:
    NFRule** fStuff;
    uint32_t fCount;
    uint32_t fCapacity;
public:
    NFRuleList(uint32_t capacity = 10) 
        : fStuff(capacity ? (NFRule**)uprv_malloc(capacity * sizeof(NFRule*)) : NULL)
        , fCount(0)
        , fCapacity(capacity) {}
    ~NFRuleList() {
        if (fStuff) {
            for(uint32_t i = 0; i < fCount; ++i) {
                delete fStuff[i];
            }
            uprv_free(fStuff);
        }
    }
    NFRule* operator[](uint32_t index) const { return fStuff != NULL ? fStuff[index] : NULL; }
    NFRule* remove(uint32_t index) {
    	if (fStuff == NULL) {
    		return NULL;
    	}
        NFRule* result = fStuff[index];
        fCount -= 1;
        for (uint32_t i = index; i < fCount; ++i) { // assumes small arrays
            fStuff[i] = fStuff[i+1];
        }
        return result;
    }
    void add(NFRule* thing) {
        if (fCount == fCapacity) {
            fCapacity += 10;
            fStuff = (NFRule**)uprv_realloc(fStuff, fCapacity * sizeof(NFRule*)); // assume success
        }
        if (fStuff != NULL) {
        	fStuff[fCount++] = thing;
        } else {
        	fCapacity = 0;
        	fCount = 0;
        }
    }
    uint32_t size() const { return fCount; }
    NFRule* last() const { return (fCount > 0 && fStuff != NULL) ? fStuff[fCount-1] : NULL; }
    NFRule** release() {
        add(NULL); // ensure null termination
        NFRule** result = fStuff;
        fStuff = NULL;
        fCount = 0;
        fCapacity = 0;
        return result;
    }
    void deleteAll() {
        NFRule** tmp = NULL;
        int32_t size = fCount;
        if (size > 0) {
            tmp = release();
            for (int32_t i = 0; i < size; i++) {
                delete tmp[i];
            }
            if (tmp) {
                uprv_free(tmp);
            }
        }
    }

private:
    NFRuleList(const NFRuleList &other); // forbid copying of this class
    NFRuleList &operator=(const NFRuleList &other); // forbid copying of this class
};

U_NAMESPACE_END

/* U_HAVE_RBNF */
#endif

// NFRLIST_H


U_NAMESPACE_BEGIN

class NFRuleSet : public UMemory {
public:
    NFRuleSet(RuleBasedNumberFormat *owner, UnicodeString* descriptions, int32_t index, UErrorCode& status);
    void parseRules(UnicodeString& rules, UErrorCode& status);
    void setNonNumericalRule(NFRule *rule);
    void setBestFractionRule(int32_t originalIndex, NFRule *newRule, UBool rememberRule);
    void makeIntoFractionRuleSet() { fIsFractionRuleSet = TRUE; }

    ~NFRuleSet();

    UBool operator==(const NFRuleSet& rhs) const;
    UBool operator!=(const NFRuleSet& rhs) const { return !operator==(rhs); }

    UBool isPublic() const { return fIsPublic; }

    UBool isParseable() const { return fIsParseable; }

    UBool isFractionRuleSet() const { return fIsFractionRuleSet; }

    void  getName(UnicodeString& result) const { result.setTo(name); }
    UBool isNamed(const UnicodeString& _name) const { return this->name == _name; }

    void  format(int64_t number, UnicodeString& toAppendTo, int32_t pos, int32_t recursionCount, UErrorCode& status) const;
    void  format(double number, UnicodeString& toAppendTo, int32_t pos, int32_t recursionCount, UErrorCode& status) const;

    UBool parse(const UnicodeString& text, ParsePosition& pos, double upperBound, uint32_t nonNumericalExecutedRuleMask, Formattable& result) const;

    void appendRules(UnicodeString& result) const; // toString

    void setDecimalFormatSymbols(const DecimalFormatSymbols &newSymbols, UErrorCode& status);

    const RuleBasedNumberFormat *getOwner() const { return owner; }
private:
    const NFRule * findNormalRule(int64_t number) const;
    const NFRule * findDoubleRule(double number) const;
    const NFRule * findFractionRuleSetRule(double number) const;
    
    friend class NFSubstitution;

private:
    UnicodeString name;
    NFRuleList rules;
    NFRule *nonNumericalRules[6];
    RuleBasedNumberFormat *owner;
    NFRuleList fractionRules;
    UBool fIsFractionRuleSet;
    UBool fIsPublic;
    UBool fIsParseable;

    NFRuleSet(const NFRuleSet &other); // forbid copying of this class
    NFRuleSet &operator=(const NFRuleSet &other); // forbid copying of this class
};

// utilities from old llong.h
// convert mantissa portion of double to int64
int64_t util64_fromDouble(double d);

// raise radix to the power exponent, only non-negative exponents
// Arithmetic is performed in unsigned space since overflow in
// signed space is undefined behavior.
uint64_t util64_pow(uint32_t radix, uint16_t exponent);

// convert n to digit string in buffer, return length of string
uint32_t util64_tou(int64_t n, UChar* buffer, uint32_t buflen, uint32_t radix = 10, UBool raw = FALSE);

#ifdef RBNF_DEBUG
int64_t util64_utoi(const UChar* str, uint32_t radix = 10);
uint32_t util64_toa(int64_t n, char* buffer, uint32_t buflen, uint32_t radix = 10, UBool raw = FALSE);
int64_t util64_atoi(const char* str, uint32_t radix);
#endif


U_NAMESPACE_END

/* U_HAVE_RBNF */
#endif

// NFRS_H


#if U_HAVE_RBNF







#ifdef RBNF_DEBUG

#endif

enum {
    /** -x */
    NEGATIVE_RULE_INDEX = 0,
    /** x.x */
    IMPROPER_FRACTION_RULE_INDEX = 1,
    /** 0.x */
    PROPER_FRACTION_RULE_INDEX = 2,
    /** x.0 */
    MASTER_RULE_INDEX = 3,
    /** Inf */
    INFINITY_RULE_INDEX = 4,
    /** NaN */
    NAN_RULE_INDEX = 5,
    NON_NUMERICAL_RULE_LENGTH = 6
};

U_NAMESPACE_BEGIN

#if 0
// euclid's algorithm works with doubles
// note, doubles only get us up to one quadrillion or so, which
// isn't as much range as we get with longs.  We probably still
// want either 64-bit math, or BigInteger.

static int64_t
util_lcm(int64_t x, int64_t y)
{
    x.abs();
    y.abs();

    if (x == 0 || y == 0) {
        return 0;
    } else {
        do {
            if (x < y) {
                int64_t t = x; x = y; y = t;
            }
            x -= y * (x/y);
        } while (x != 0);

        return y;
    }
}

#else
/**
 * Calculates the least common multiple of x and y.
 */
static int64_t
util_lcm(int64_t x, int64_t y)
{
    // binary gcd algorithm from Knuth, "The Art of Computer Programming,"
    // vol. 2, 1st ed., pp. 298-299
    int64_t x1 = x;
    int64_t y1 = y;

    int p2 = 0;
    while ((x1 & 1) == 0 && (y1 & 1) == 0) {
        ++p2;
        x1 >>= 1;
        y1 >>= 1;
    }

    int64_t t;
    if ((x1 & 1) == 1) {
        t = -y1;
    } else {
        t = x1;
    }

    while (t != 0) {
        while ((t & 1) == 0) {
            t = t >> 1;
        }
        if (t > 0) {
            x1 = t;
        } else {
            y1 = -t;
        }
        t = x1 - y1;
    }

    int64_t gcd = x1 << p2;

    // x * y == gcd(x, y) * lcm(x, y)
    return x / gcd * y;
}
#endif

static const UChar nfrs_gPercent = 0x0025;
static const UChar nfrs_gColon = 0x003a;
static const UChar nfrs_gSemicolon = 0x003b;
static const UChar nfrs_gLineFeed = 0x000a;

static const UChar nfrs_gPercentPercent[] =
{
    0x25, 0x25, 0
}; /* "%%" */

static const UChar nfrs_gNoparse[] =
{
    0x40, 0x6E, 0x6F, 0x70, 0x61, 0x72, 0x73, 0x65, 0
}; /* "@noparse" */

NFRuleSet::NFRuleSet(RuleBasedNumberFormat *_owner, UnicodeString* descriptions, int32_t index, UErrorCode& status)
  : name()
  , rules(0)
  , owner(_owner)
  , fractionRules()
  , fIsFractionRuleSet(FALSE)
  , fIsPublic(FALSE)
  , fIsParseable(TRUE)
{
    for (int32_t i = 0; i < NON_NUMERICAL_RULE_LENGTH; ++i) {
        nonNumericalRules[i] = NULL;
    }

    if (U_FAILURE(status)) {
        return;
    }

    UnicodeString& description = descriptions[index]; // !!! make sure index is valid

    if (description.length() == 0) {
        // throw new IllegalArgumentException("Empty rule set description");
        status = U_PARSE_ERROR;
        return;
    }

    // if the description begins with a rule set name (the rule set
    // name can be omitted in formatter descriptions that consist
    // of only one rule set), copy it out into our "name" member
    // and delete it from the description
    if (description.charAt(0) == nfrs_gPercent) {
        int32_t pos = description.indexOf(nfrs_gColon);
        if (pos == -1) {
            // throw new IllegalArgumentException("Rule set name doesn't end in colon");
            status = U_PARSE_ERROR;
        } else {
            name.setTo(description, 0, pos);
            while (pos < description.length() && PatternProps::isWhiteSpace(description.charAt(++pos))) {
            }
            description.remove(0, pos);
        }
    } else {
        name.setTo(UNICODE_STRING_SIMPLE("%default"));
    }

    if (description.length() == 0) {
        // throw new IllegalArgumentException("Empty rule set description");
        status = U_PARSE_ERROR;
    }

    fIsPublic = name.indexOf(nfrs_gPercentPercent, 2, 0) != 0;

    if ( name.endsWith(nfrs_gNoparse,8) ) {
        fIsParseable = FALSE;
        name.truncate(name.length()-8); // remove the @noparse from the name
    }

    // all of the other members of NFRuleSet are initialized
    // by parseRules()
}

void
NFRuleSet::parseRules(UnicodeString& description, UErrorCode& status)
{
    // start by creating a Vector whose elements are Strings containing
    // the descriptions of the rules (one rule per element).  The rules
    // are separated by semicolons (there's no escape facility: ALL
    // semicolons are rule delimiters)

    if (U_FAILURE(status)) {
        return;
    }

    // ensure we are starting with an empty rule list
    rules.deleteAll();

    // dlf - the original code kept a separate description array for no reason,
    // so I got rid of it.  The loop was too complex so I simplified it.

    UnicodeString currentDescription;
    int32_t oldP = 0;
    while (oldP < description.length()) {
        int32_t p = description.indexOf(nfrs_gSemicolon, oldP);
        if (p == -1) {
            p = description.length();
        }
        currentDescription.setTo(description, oldP, p - oldP);
        NFRule::makeRules(currentDescription, this, rules.last(), owner, rules, status);
        oldP = p + 1;
    }

    // for rules that didn't specify a base value, their base values
    // were initialized to 0.  Make another pass through the list and
    // set all those rules' base values.  We also remove any special
    // rules from the list and put them into their own member variables
    int64_t defaultBaseValue = 0;

    // (this isn't a for loop because we might be deleting items from
    // the vector-- we want to make sure we only increment i when
    // we _didn't_ delete aything from the vector)
    int32_t rulesSize = rules.size();
    for (int32_t i = 0; i < rulesSize; i++) {
        NFRule* rule = rules[i];
        int64_t baseValue = rule->getBaseValue();

        if (baseValue == 0) {
            // if the rule's base value is 0, fill in a default
            // base value (this will be 1 plus the preceding
            // rule's base value for regular rule sets, and the
            // same as the preceding rule's base value in fraction
            // rule sets)
            rule->setBaseValue(defaultBaseValue, status);
        }
        else {
            // if it's a regular rule that already knows its base value,
            // check to make sure the rules are in order, and update
            // the default base value for the next rule
            if (baseValue < defaultBaseValue) {
                // throw new IllegalArgumentException("Rules are not in order");
                status = U_PARSE_ERROR;
                return;
            }
            defaultBaseValue = baseValue;
        }
        if (!fIsFractionRuleSet) {
            ++defaultBaseValue;
        }
    }
}

/**
 * Set one of the non-numerical rules.
 * @param rule The rule to set.
 */
void NFRuleSet::setNonNumericalRule(NFRule *rule) {
    int64_t baseValue = rule->getBaseValue();
    if (baseValue == NFRule::kNegativeNumberRule) {
        delete nonNumericalRules[NEGATIVE_RULE_INDEX];
        nonNumericalRules[NEGATIVE_RULE_INDEX] = rule;
    }
    else if (baseValue == NFRule::kImproperFractionRule) {
        setBestFractionRule(IMPROPER_FRACTION_RULE_INDEX, rule, TRUE);
    }
    else if (baseValue == NFRule::kProperFractionRule) {
        setBestFractionRule(PROPER_FRACTION_RULE_INDEX, rule, TRUE);
    }
    else if (baseValue == NFRule::kMasterRule) {
        setBestFractionRule(MASTER_RULE_INDEX, rule, TRUE);
    }
    else if (baseValue == NFRule::kInfinityRule) {
        delete nonNumericalRules[INFINITY_RULE_INDEX];
        nonNumericalRules[INFINITY_RULE_INDEX] = rule;
    }
    else if (baseValue == NFRule::kNaNRule) {
        delete nonNumericalRules[NAN_RULE_INDEX];
        nonNumericalRules[NAN_RULE_INDEX] = rule;
    }
}

/**
 * Determine the best fraction rule to use. Rules matching the decimal point from
 * DecimalFormatSymbols become the main set of rules to use.
 * @param originalIndex The index into nonNumericalRules
 * @param newRule The new rule to consider
 * @param rememberRule Should the new rule be added to fractionRules.
 */
void NFRuleSet::setBestFractionRule(int32_t originalIndex, NFRule *newRule, UBool rememberRule) {
    if (rememberRule) {
        fractionRules.add(newRule);
    }
    NFRule *bestResult = nonNumericalRules[originalIndex];
    if (bestResult == NULL) {
        nonNumericalRules[originalIndex] = newRule;
    }
    else {
        // We have more than one. Which one is better?
        const DecimalFormatSymbols *decimalFormatSymbols = owner->getDecimalFormatSymbols();
        if (decimalFormatSymbols->getSymbol(DecimalFormatSymbols::kDecimalSeparatorSymbol).charAt(0)
            == newRule->getDecimalPoint())
        {
            nonNumericalRules[originalIndex] = newRule;
        }
        // else leave it alone
    }
}

NFRuleSet::~NFRuleSet()
{
    for (int i = 0; i < NON_NUMERICAL_RULE_LENGTH; i++) {
        if (i != IMPROPER_FRACTION_RULE_INDEX
            && i != PROPER_FRACTION_RULE_INDEX
            && i != MASTER_RULE_INDEX)
        {
            delete nonNumericalRules[i];
        }
        // else it will be deleted via NFRuleList fractionRules
    }
}

static UBool
util_equalRules(const NFRule* rule1, const NFRule* rule2)
{
    if (rule1) {
        if (rule2) {
            return *rule1 == *rule2;
        }
    } else if (!rule2) {
        return TRUE;
    }
    return FALSE;
}

UBool
NFRuleSet::operator==(const NFRuleSet& rhs) const
{
    if (rules.size() == rhs.rules.size() &&
        fIsFractionRuleSet == rhs.fIsFractionRuleSet &&
        name == rhs.name) {

        // ...then compare the non-numerical rule lists...
        for (int i = 0; i < NON_NUMERICAL_RULE_LENGTH; i++) {
            if (!util_equalRules(nonNumericalRules[i], rhs.nonNumericalRules[i])) {
                return FALSE;
            }
        }

        // ...then compare the rule lists...
        for (uint32_t i = 0; i < rules.size(); ++i) {
            if (*rules[i] != *rhs.rules[i]) {
                return FALSE;
            }
        }
        return TRUE;
    }
    return FALSE;
}

void
NFRuleSet::setDecimalFormatSymbols(const DecimalFormatSymbols &newSymbols, UErrorCode& status) {
    for (uint32_t i = 0; i < rules.size(); ++i) {
        rules[i]->setDecimalFormatSymbols(newSymbols, status);
    }
    // Switch the fraction rules to mirror the DecimalFormatSymbols.
    for (int32_t nonNumericalIdx = IMPROPER_FRACTION_RULE_INDEX; nonNumericalIdx <= MASTER_RULE_INDEX; nonNumericalIdx++) {
        if (nonNumericalRules[nonNumericalIdx]) {
            for (uint32_t fIdx = 0; fIdx < fractionRules.size(); fIdx++) {
                NFRule *fractionRule = fractionRules[fIdx];
                if (nonNumericalRules[nonNumericalIdx]->getBaseValue() == fractionRule->getBaseValue()) {
                    setBestFractionRule(nonNumericalIdx, fractionRule, FALSE);
                }
            }
        }
    }

    for (uint32_t nnrIdx = 0; nnrIdx < NON_NUMERICAL_RULE_LENGTH; nnrIdx++) {
        NFRule *rule = nonNumericalRules[nnrIdx];
        if (rule) {
            rule->setDecimalFormatSymbols(newSymbols, status);
        }
    }
}

#define RECURSION_LIMIT 64

void
NFRuleSet::format(int64_t number, UnicodeString& toAppendTo, int32_t pos, int32_t recursionCount, UErrorCode& status) const
{
    if (recursionCount >= RECURSION_LIMIT) {
        // stop recursion
        status = U_INVALID_STATE_ERROR;
        return;
    }
    const NFRule *rule = findNormalRule(number);
    if (rule) { // else error, but can't report it
        rule->doFormat(number, toAppendTo, pos, ++recursionCount, status);
    }
}

void
NFRuleSet::format(double number, UnicodeString& toAppendTo, int32_t pos, int32_t recursionCount, UErrorCode& status) const
{
    if (recursionCount >= RECURSION_LIMIT) {
        // stop recursion
        status = U_INVALID_STATE_ERROR;
        return;
    }
    const NFRule *rule = findDoubleRule(number);
    if (rule) { // else error, but can't report it
        rule->doFormat(number, toAppendTo, pos, ++recursionCount, status);
    }
}

const NFRule*
NFRuleSet::findDoubleRule(double number) const
{
    // if this is a fraction rule set, use findFractionRuleSetRule()
    if (isFractionRuleSet()) {
        return findFractionRuleSetRule(number);
    }

    if (uprv_isNaN(number)) {
        const NFRule *rule = nonNumericalRules[NAN_RULE_INDEX];
        if (!rule) {
            rule = owner->getDefaultNaNRule();
        }
        return rule;
    }

    // if the number is negative, return the negative number rule
    // (if there isn't a negative-number rule, we pretend it's a
    // positive number)
    if (number < 0) {
        if (nonNumericalRules[NEGATIVE_RULE_INDEX]) {
            return  nonNumericalRules[NEGATIVE_RULE_INDEX];
        } else {
            number = -number;
        }
    }

    if (uprv_isInfinite(number)) {
        const NFRule *rule = nonNumericalRules[INFINITY_RULE_INDEX];
        if (!rule) {
            rule = owner->getDefaultInfinityRule();
        }
        return rule;
    }

    // if the number isn't an integer, we use one of the fraction rules...
    if (number != uprv_floor(number)) {
        // if the number is between 0 and 1, return the proper
        // fraction rule
        if (number < 1 && nonNumericalRules[PROPER_FRACTION_RULE_INDEX]) {
            return nonNumericalRules[PROPER_FRACTION_RULE_INDEX];
        }
        // otherwise, return the improper fraction rule
        else if (nonNumericalRules[IMPROPER_FRACTION_RULE_INDEX]) {
            return nonNumericalRules[IMPROPER_FRACTION_RULE_INDEX];
        }
    }

    // if there's a master rule, use it to format the number
    if (nonNumericalRules[MASTER_RULE_INDEX]) {
        return nonNumericalRules[MASTER_RULE_INDEX];
    }

    // and if we haven't yet returned a rule, use findNormalRule()
    // to find the applicable rule
    int64_t r = util64_fromDouble(number + 0.5);
    return findNormalRule(r);
}

const NFRule *
NFRuleSet::findNormalRule(int64_t number) const
{
    // if this is a fraction rule set, use findFractionRuleSetRule()
    // to find the rule (we should only go into this clause if the
    // value is 0)
    if (fIsFractionRuleSet) {
        return findFractionRuleSetRule((double)number);
    }

    // if the number is negative, return the negative-number rule
    // (if there isn't one, pretend the number is positive)
    if (number < 0) {
        if (nonNumericalRules[NEGATIVE_RULE_INDEX]) {
            return nonNumericalRules[NEGATIVE_RULE_INDEX];
        } else {
            number = -number;
        }
    }

    // we have to repeat the preceding two checks, even though we
    // do them in findRule(), because the version of format() that
    // takes a long bypasses findRule() and goes straight to this
    // function.  This function does skip the fraction rules since
    // we know the value is an integer (it also skips the master
    // rule, since it's considered a fraction rule.  Skipping the
    // master rule in this function is also how we avoid infinite
    // recursion)

    // {dlf} unfortunately this fails if there are no rules except
    // special rules.  If there are no rules, use the master rule.

    // binary-search the rule list for the applicable rule
    // (a rule is used for all values from its base value to
    // the next rule's base value)
    int32_t hi = rules.size();
    if (hi > 0) {
        int32_t lo = 0;

        while (lo < hi) {
            int32_t mid = (lo + hi) / 2;
            if (rules[mid]->getBaseValue() == number) {
                return rules[mid];
            }
            else if (rules[mid]->getBaseValue() > number) {
                hi = mid;
            }
            else {
                lo = mid + 1;
            }
        }
        if (hi == 0) { // bad rule set, minimum base > 0
            return NULL; // want to throw exception here
        }

        NFRule *result = rules[hi - 1];

        // use shouldRollBack() to see whether we need to invoke the
        // rollback rule (see shouldRollBack()'s documentation for
        // an explanation of the rollback rule).  If we do, roll back
        // one rule and return that one instead of the one we'd normally
        // return
        if (result->shouldRollBack(number)) {
            if (hi == 1) { // bad rule set, no prior rule to rollback to from this base
                return NULL;
            }
            result = rules[hi - 2];
        }
        return result;
    }
    // else use the master rule
    return nonNumericalRules[MASTER_RULE_INDEX];
}

/**
 * If this rule is a fraction rule set, this function is used by
 * findRule() to select the most appropriate rule for formatting
 * the number.  Basically, the base value of each rule in the rule
 * set is treated as the denominator of a fraction.  Whichever
 * denominator can produce the fraction closest in value to the
 * number passed in is the result.  If there's a tie, the earlier
 * one in the list wins.  (If there are two rules in a row with the
 * same base value, the first one is used when the numerator of the
 * fraction would be 1, and the second rule is used the rest of the
 * time.
 * @param number The number being formatted (which will always be
 * a number between 0 and 1)
 * @return The rule to use to format this number
 */
const NFRule*
NFRuleSet::findFractionRuleSetRule(double number) const
{
    // the obvious way to do this (multiply the value being formatted
    // by each rule's base value until you get an integral result)
    // doesn't work because of rounding error.  This method is more
    // accurate

    // find the least common multiple of the rules' base values
    // and multiply this by the number being formatted.  This is
    // all the precision we need, and we can do all of the rest
    // of the math using integer arithmetic
    int64_t leastCommonMultiple = rules[0]->getBaseValue();
    int64_t numerator;
    {
        for (uint32_t i = 1; i < rules.size(); ++i) {
            leastCommonMultiple = util_lcm(leastCommonMultiple, rules[i]->getBaseValue());
        }
        numerator = util64_fromDouble(number * (double)leastCommonMultiple + 0.5);
    }
    // for each rule, do the following...
    int64_t tempDifference;
    int64_t difference = util64_fromDouble(uprv_maxMantissa());
    int32_t winner = 0;
    for (uint32_t i = 0; i < rules.size(); ++i) {
        // "numerator" is the numerator of the fraction if the
        // denominator is the LCD.  The numerator if the rule's
        // base value is the denominator is "numerator" times the
        // base value divided bythe LCD.  Here we check to see if
        // that's an integer, and if not, how close it is to being
        // an integer.
        tempDifference = numerator * rules[i]->getBaseValue() % leastCommonMultiple;


        // normalize the result of the above calculation: we want
        // the numerator's distance from the CLOSEST multiple
        // of the LCD
        if (leastCommonMultiple - tempDifference < tempDifference) {
            tempDifference = leastCommonMultiple - tempDifference;
        }

        // if this is as close as we've come, keep track of how close
        // that is, and the line number of the rule that did it.  If
        // we've scored a direct hit, we don't have to look at any more
        // rules
        if (tempDifference < difference) {
            difference = tempDifference;
            winner = i;
            if (difference == 0) {
                break;
            }
        }
    }

    // if we have two successive rules that both have the winning base
    // value, then the first one (the one we found above) is used if
    // the numerator of the fraction is 1 and the second one is used if
    // the numerator of the fraction is anything else (this lets us
    // do things like "one third"/"two thirds" without haveing to define
    // a whole bunch of extra rule sets)
    if ((unsigned)(winner + 1) < rules.size() &&
        rules[winner + 1]->getBaseValue() == rules[winner]->getBaseValue()) {
        double n = ((double)rules[winner]->getBaseValue()) * number;
        if (n < 0.5 || n >= 2) {
            ++winner;
        }
    }

    // finally, return the winning rule
    return rules[winner];
}

/**
 * Parses a string.  Matches the string to be parsed against each
 * of its rules (with a base value less than upperBound) and returns
 * the value produced by the rule that matched the most charcters
 * in the source string.
 * @param text The string to parse
 * @param parsePosition The initial position is ignored and assumed
 * to be 0.  On exit, this object has been updated to point to the
 * first character position this rule set didn't consume.
 * @param upperBound Limits the rules that can be allowed to match.
 * Only rules whose base values are strictly less than upperBound
 * are considered.
 * @return The numerical result of parsing this string.  This will
 * be the matching rule's base value, composed appropriately with
 * the results of matching any of its substitutions.  The object
 * will be an instance of Long if it's an integral value; otherwise,
 * it will be an instance of Double.  This function always returns
 * a valid object: If nothing matched the input string at all,
 * this function returns new Long(0), and the parse position is
 * left unchanged.
 */
#ifdef RBNF_DEBUG
#include <stdio.h>

static void dumpUS(FILE* f, const UnicodeString& us) {
  int len = us.length();
  char* buf = (char *)uprv_malloc((len+1)*sizeof(char)); //new char[len+1];
  if (buf != NULL) {
	  us.extract(0, len, buf);
	  buf[len] = 0;
	  fprintf(f, "%s", buf);
	  uprv_free(buf); //delete[] buf;
  }
}
#endif

UBool
NFRuleSet::parse(const UnicodeString& text, ParsePosition& pos, double upperBound, uint32_t nonNumericalExecutedRuleMask, Formattable& result) const
{
    // try matching each rule in the rule set against the text being
    // parsed.  Whichever one matches the most characters is the one
    // that determines the value we return.

    result.setLong(0);

    // dump out if there's no text to parse
    if (text.length() == 0) {
        return 0;
    }

    ParsePosition highWaterMark;
    ParsePosition workingPos = pos;

#ifdef RBNF_DEBUG
    fprintf(stderr, "<nfrs> %x '", this);
    dumpUS(stderr, name);
    fprintf(stderr, "' text '");
    dumpUS(stderr, text);
    fprintf(stderr, "'\n");
    fprintf(stderr, "  parse negative: %d\n", this, negativeNumberRule != 0);
#endif
    // Try each of the negative rules, fraction rules, infinity rules and NaN rules
    for (int i = 0; i < NON_NUMERICAL_RULE_LENGTH; i++) {
        if (nonNumericalRules[i] && ((nonNumericalExecutedRuleMask >> i) & 1) == 0) {
            // Mark this rule as being executed so that we don't try to execute it again.
            nonNumericalExecutedRuleMask |= 1 << i;

            Formattable tempResult;
            UBool success = nonNumericalRules[i]->doParse(text, workingPos, 0, upperBound, nonNumericalExecutedRuleMask, tempResult);
            if (success && (workingPos.getIndex() > highWaterMark.getIndex())) {
                result = tempResult;
                highWaterMark = workingPos;
            }
            workingPos = pos;
        }
    }
#ifdef RBNF_DEBUG
    fprintf(stderr, "<nfrs> continue other with text '");
    dumpUS(stderr, text);
    fprintf(stderr, "' hwm: %d\n", highWaterMark.getIndex());
#endif

    // finally, go through the regular rules one at a time.  We start
    // at the end of the list because we want to try matching the most
    // sigificant rule first (this helps ensure that we parse
    // "five thousand three hundred six" as
    // "(five thousand) (three hundred) (six)" rather than
    // "((five thousand three) hundred) (six)").  Skip rules whose
    // base values are higher than the upper bound (again, this helps
    // limit ambiguity by making sure the rules that match a rule's
    // are less significant than the rule containing the substitutions)/
    {
        int64_t ub = util64_fromDouble(upperBound);
#ifdef RBNF_DEBUG
        {
            char ubstr[64];
            util64_toa(ub, ubstr, 64);
            char ubstrhex[64];
            util64_toa(ub, ubstrhex, 64, 16);
            fprintf(stderr, "ub: %g, i64: %s (%s)\n", upperBound, ubstr, ubstrhex);
        }
#endif
        for (int32_t i = rules.size(); --i >= 0 && highWaterMark.getIndex() < text.length();) {
            if ((!fIsFractionRuleSet) && (rules[i]->getBaseValue() >= ub)) {
                continue;
            }
            Formattable tempResult;
            UBool success = rules[i]->doParse(text, workingPos, fIsFractionRuleSet, upperBound, nonNumericalExecutedRuleMask, tempResult);
            if (success && workingPos.getIndex() > highWaterMark.getIndex()) {
                result = tempResult;
                highWaterMark = workingPos;
            }
            workingPos = pos;
        }
    }
#ifdef RBNF_DEBUG
    fprintf(stderr, "<nfrs> exit\n");
#endif
    // finally, update the parse postion we were passed to point to the
    // first character we didn't use, and return the result that
    // corresponds to that string of characters
    pos = highWaterMark;

    return 1;
}

void
NFRuleSet::appendRules(UnicodeString& result) const
{
    uint32_t i;

    // the rule set name goes first...
    result.append(name);
    result.append(nfrs_gColon);
    result.append(nfrs_gLineFeed);

    // followed by the regular rules...
    for (i = 0; i < rules.size(); i++) {
        rules[i]->_appendRuleText(result);
        result.append(nfrs_gLineFeed);
    }

    // followed by the special rules (if they exist)
    for (i = 0; i < NON_NUMERICAL_RULE_LENGTH; ++i) {
        NFRule *rule = nonNumericalRules[i];
        if (nonNumericalRules[i]) {
            if (rule->getBaseValue() == NFRule::kImproperFractionRule
                || rule->getBaseValue() == NFRule::kProperFractionRule
                || rule->getBaseValue() == NFRule::kMasterRule)
            {
                for (uint32_t fIdx = 0; fIdx < fractionRules.size(); fIdx++) {
                    NFRule *fractionRule = fractionRules[fIdx];
                    if (fractionRule->getBaseValue() == rule->getBaseValue()) {
                        fractionRule->_appendRuleText(result);
                        result.append(nfrs_gLineFeed);
                    }
                }
            }
            else {
                rule->_appendRuleText(result);
                result.append(nfrs_gLineFeed);
            }
        }
    }
}

// utility functions

int64_t util64_fromDouble(double d) {
    int64_t result = 0;
    if (!uprv_isNaN(d)) {
        double mant = uprv_maxMantissa();
        if (d < -mant) {
            d = -mant;
        } else if (d > mant) {
            d = mant;
        }
        UBool neg = d < 0;
        if (neg) {
            d = -d;
        }
        result = (int64_t)uprv_floor(d);
        if (neg) {
            result = -result;
        }
    }
    return result;
}

uint64_t util64_pow(uint32_t base, uint16_t exponent)  {
    if (base == 0) {
        return 0;
    }
    uint64_t result = 1;
    uint64_t pow = base;
    while (true) {
        if ((exponent & 1) == 1) {
            result *= pow;
        }
        exponent >>= 1;
        if (exponent == 0) {
            break;
        }
        pow *= pow;
    }
    return result;
}

static const uint8_t asciiDigits[] = {
    0x30u, 0x31u, 0x32u, 0x33u, 0x34u, 0x35u, 0x36u, 0x37u,
    0x38u, 0x39u, 0x61u, 0x62u, 0x63u, 0x64u, 0x65u, 0x66u,
    0x67u, 0x68u, 0x69u, 0x6au, 0x6bu, 0x6cu, 0x6du, 0x6eu,
    0x6fu, 0x70u, 0x71u, 0x72u, 0x73u, 0x74u, 0x75u, 0x76u,
    0x77u, 0x78u, 0x79u, 0x7au,
};

static const UChar kUMinus = (UChar)0x002d;

#ifdef RBNF_DEBUG
static const char kMinus = '-';

static const uint8_t digitInfo[] = {
        0,     0,     0,     0,     0,     0,     0,     0,
        0,     0,     0,     0,     0,     0,     0,     0,
        0,     0,     0,     0,     0,     0,     0,     0,
        0,     0,     0,     0,     0,     0,     0,     0,
        0,     0,     0,     0,     0,     0,     0,     0,
        0,     0,     0,     0,     0,     0,     0,     0,
    0x80u, 0x81u, 0x82u, 0x83u, 0x84u, 0x85u, 0x86u, 0x87u,
    0x88u, 0x89u,     0,     0,     0,     0,     0,     0,
        0, 0x8au, 0x8bu, 0x8cu, 0x8du, 0x8eu, 0x8fu, 0x90u,
    0x91u, 0x92u, 0x93u, 0x94u, 0x95u, 0x96u, 0x97u, 0x98u,
    0x99u, 0x9au, 0x9bu, 0x9cu, 0x9du, 0x9eu, 0x9fu, 0xa0u,
    0xa1u, 0xa2u, 0xa3u,     0,     0,     0,     0,     0,
        0, 0x8au, 0x8bu, 0x8cu, 0x8du, 0x8eu, 0x8fu, 0x90u,
    0x91u, 0x92u, 0x93u, 0x94u, 0x95u, 0x96u, 0x97u, 0x98u,
    0x99u, 0x9au, 0x9bu, 0x9cu, 0x9du, 0x9eu, 0x9fu, 0xa0u,
    0xa1u, 0xa2u, 0xa3u,     0,     0,     0,     0,     0,
};

int64_t util64_atoi(const char* str, uint32_t radix)
{
    if (radix > 36) {
        radix = 36;
    } else if (radix < 2) {
        radix = 2;
    }
    int64_t lradix = radix;

    int neg = 0;
    if (*str == kMinus) {
        ++str;
        neg = 1;
    }
    int64_t result = 0;
    uint8_t b;
    while ((b = digitInfo[*str++]) && ((b &= 0x7f) < radix)) {
        result *= lradix;
        result += (int32_t)b;
    }
    if (neg) {
        result = -result;
    }
    return result;
}

int64_t util64_utoi(const UChar* str, uint32_t radix)
{
    if (radix > 36) {
        radix = 36;
    } else if (radix < 2) {
        radix = 2;
    }
    int64_t lradix = radix;

    int neg = 0;
    if (*str == kUMinus) {
        ++str;
        neg = 1;
    }
    int64_t result = 0;
    UChar c;
    uint8_t b;
    while (((c = *str++) < 0x0080) && (b = digitInfo[c]) && ((b &= 0x7f) < radix)) {
        result *= lradix;
        result += (int32_t)b;
    }
    if (neg) {
        result = -result;
    }
    return result;
}

uint32_t util64_toa(int64_t w, char* buf, uint32_t len, uint32_t radix, UBool raw)
{
    if (radix > 36) {
        radix = 36;
    } else if (radix < 2) {
        radix = 2;
    }
    int64_t base = radix;

    char* p = buf;
    if (len && (w < 0) && (radix == 10) && !raw) {
        w = -w;
        *p++ = kMinus;
        --len;
    } else if (len && (w == 0)) {
        *p++ = (char)raw ? 0 : asciiDigits[0];
        --len;
    }

    while (len && w != 0) {
        int64_t n = w / base;
        int64_t m = n * base;
        int32_t d = (int32_t)(w-m);
        *p++ = raw ? (char)d : asciiDigits[d];
        w = n;
        --len;
    }
    if (len) {
        *p = 0; // null terminate if room for caller convenience
    }

    len = p - buf;
    if (*buf == kMinus) {
        ++buf;
    }
    while (--p > buf) {
        char c = *p;
        *p = *buf;
        *buf = c;
        ++buf;
    }

    return len;
}
#endif

uint32_t util64_tou(int64_t w, UChar* buf, uint32_t len, uint32_t radix, UBool raw)
{
    if (radix > 36) {
        radix = 36;
    } else if (radix < 2) {
        radix = 2;
    }
    int64_t base = radix;

    UChar* p = buf;
    if (len && (w < 0) && (radix == 10) && !raw) {
        w = -w;
        *p++ = kUMinus;
        --len;
    } else if (len && (w == 0)) {
        *p++ = (UChar)raw ? 0 : asciiDigits[0];
        --len;
    }

    while (len && (w != 0)) {
        int64_t n = w / base;
        int64_t m = n * base;
        int32_t d = (int32_t)(w-m);
        *p++ = (UChar)(raw ? d : asciiDigits[d]);
        w = n;
        --len;
    }
    if (len) {
        *p = 0; // null terminate if room for caller convenience
    }

    len = (uint32_t)(p - buf);
    if (*buf == kUMinus) {
        ++buf;
    }
    while (--p > buf) {
        UChar c = *p;
        *p = *buf;
        *buf = c;
        ++buf;
    }

    return len;
}


U_NAMESPACE_END

/* U_HAVE_RBNF */
#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 1997-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*   file name:  nfrule.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
* Modification history
* Date        Name      Comments
* 10/11/2001  Doug      Ported from ICU4J
*/



#if U_HAVE_RBNF










// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 1997-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*   file name:  nfsubs.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
* Modification history
* Date        Name      Comments
* 10/11/2001  Doug      Ported from ICU4J
*/






#if U_HAVE_RBNF




#include <float.h>

U_NAMESPACE_BEGIN

class NFSubstitution : public UObject {
    int32_t pos;
    const NFRuleSet* ruleSet;
    DecimalFormat* numberFormat;
    
protected:
    NFSubstitution(int32_t pos,
        const NFRuleSet* ruleSet,
        const UnicodeString& description,
        UErrorCode& status);
    
    /**
     * Get the Ruleset of the object.
     * @return the Ruleset of the object.
     */
    const NFRuleSet* getRuleSet() const { return ruleSet; }

    /**
     * get the NumberFormat of this object.
     * @return the numberformat of this object.
     */
    const DecimalFormat* getNumberFormat() const { return numberFormat; }
    
public:
    static NFSubstitution* makeSubstitution(int32_t pos, 
        const NFRule* rule, 
        const NFRule* predecessor,
        const NFRuleSet* ruleSet, 
        const RuleBasedNumberFormat* rbnf, 
        const UnicodeString& description,
        UErrorCode& status);
    
    /**
     * Destructor.
     */
    virtual ~NFSubstitution();
    
    /**
     * Return true if the given Format objects are semantically equal.
     * Objects of different subclasses are considered unequal.
     * @param rhs    the object to be compared with.
     * @return       true if the given Format objects are semantically equal.
     */
    virtual UBool operator==(const NFSubstitution& rhs) const;

    /**
     * Return true if the given Format objects are semantically unequal.
     * Objects of different subclasses are considered unequal.
     * @param rhs    the object to be compared with.
     * @return       true if the given Format objects are semantically unequal.
     */
    UBool operator!=(const NFSubstitution& rhs) const { return !operator==(rhs); }
    
    /**
     * Sets the substitution's divisor.  Used by NFRule.setBaseValue().
     * A no-op for all substitutions except multiplier and modulus
     * substitutions.
     * @param radix The radix of the divisor
     * @param exponent The exponent of the divisor
     */
    virtual void setDivisor(int32_t radix, int16_t exponent, UErrorCode& status);
    
    /**
     * Replaces result with the string describing the substitution.
     * @param result    Output param which will receive the string.
     */
    virtual void toString(UnicodeString& result) const;
    
    void setDecimalFormatSymbols(const DecimalFormatSymbols &newSymbols, UErrorCode& status);

    //-----------------------------------------------------------------------
    // formatting
    //-----------------------------------------------------------------------
    
    /**
     * Performs a mathematical operation on the number, formats it using
     * either ruleSet or decimalFormat, and inserts the result into
     * toInsertInto.
     * @param number The number being formatted.
     * @param toInsertInto The string we insert the result into
     * @param pos The position in toInsertInto where the owning rule's
     * rule text begins (this value is added to this substitution's
     * position to determine exactly where to insert the new text)
     */
    virtual void doSubstitution(int64_t number, UnicodeString& toInsertInto, int32_t pos, int32_t recursionCount, UErrorCode& status) const;

    /**
     * Performs a mathematical operation on the number, formats it using
     * either ruleSet or decimalFormat, and inserts the result into
     * toInsertInto.
     * @param number The number being formatted.
     * @param toInsertInto The string we insert the result into
     * @param pos The position in toInsertInto where the owning rule's
     * rule text begins (this value is added to this substitution's
     * position to determine exactly where to insert the new text)
     */
    virtual void doSubstitution(double number, UnicodeString& toInsertInto, int32_t pos, int32_t recursionCount, UErrorCode& status) const;
    
protected:
    /**
     * Subclasses override this function to perform some kind of
     * mathematical operation on the number.  The result of this operation
     * is formatted using the rule set or DecimalFormat that this
     * substitution refers to, and the result is inserted into the result
     * string.
     * @param The number being formatted
     * @return The result of performing the opreration on the number
     */
    virtual int64_t transformNumber(int64_t number) const = 0;

    /**
     * Subclasses override this function to perform some kind of
     * mathematical operation on the number.  The result of this operation
     * is formatted using the rule set or DecimalFormat that this
     * substitution refers to, and the result is inserted into the result
     * string.
     * @param The number being formatted
     * @return The result of performing the opreration on the number
     */
    virtual double transformNumber(double number) const = 0;
    
public:
    //-----------------------------------------------------------------------
    // parsing
    //-----------------------------------------------------------------------
    
    /**
     * Parses a string using the rule set or DecimalFormat belonging
     * to this substitution.  If there's a match, a mathematical
     * operation (the inverse of the one used in formatting) is
     * performed on the result of the parse and the value passed in
     * and returned as the result.  The parse position is updated to
     * point to the first unmatched character in the string.
     * @param text The string to parse
     * @param parsePosition On entry, ignored, but assumed to be 0.
     * On exit, this is updated to point to the first unmatched
     * character (or 0 if the substitution didn't match)
     * @param baseValue A partial parse result that should be
     * combined with the result of this parse
     * @param upperBound When searching the rule set for a rule
     * matching the string passed in, only rules with base values
     * lower than this are considered
     * @param lenientParse If true and matching against rules fails,
     * the substitution will also try matching the text against
     * numerals using a default-costructed NumberFormat.  If false,
     * no extra work is done.  (This value is false whenever the
     * formatter isn't in lenient-parse mode, but is also false
     * under some conditions even when the formatter _is_ in
     * lenient-parse mode.)
     * @return If there's a match, this is the result of composing
     * baseValue with whatever was returned from matching the
     * characters.  This will be either a Long or a Double.  If there's
     * no match this is new Long(0) (not null), and parsePosition
     * is left unchanged.
     */
    virtual UBool doParse(const UnicodeString& text, 
        ParsePosition& parsePosition, 
        double baseValue,
        double upperBound, 
        UBool lenientParse,
        uint32_t nonNumericalExecutedRuleMask,
        Formattable& result) const;
    
    /**
     * Derives a new value from the two values passed in.  The two values
     * are typically either the base values of two rules (the one containing
     * the substitution and the one matching the substitution) or partial
     * parse results derived in some other way.  The operation is generally
     * the inverse of the operation performed by transformNumber().
     * @param newRuleValue The value produced by matching this substitution
     * @param oldRuleValue The value that was passed to the substitution
     * by the rule that owns it
     * @return A third value derived from the other two, representing a
     * partial parse result
     */
    virtual double composeRuleValue(double newRuleValue, double oldRuleValue) const = 0;
    
    /**
     * Calculates an upper bound when searching for a rule that matches
     * this substitution.  Rules with base values greater than or equal
     * to upperBound are not considered.
     * @param oldUpperBound    The current upper-bound setting.  The new
     *                         upper bound can't be any higher.
     * @return                 the upper bound when searching for a rule that matches
     *                         this substitution.
     */
    virtual double calcUpperBound(double oldUpperBound) const = 0;
    
    //-----------------------------------------------------------------------
    // simple accessors
    //-----------------------------------------------------------------------
    
    /**
     * Returns the substitution's position in the rule that owns it.
     * @return The substitution's position in the rule that owns it.
     */
    int32_t getPos() const { return pos; }
    
    /**
     * Returns the character used in the textual representation of
     * substitutions of this type.  Used by toString().
     * @return This substitution's token character.
     */
    virtual UChar tokenChar() const = 0;
    
    /**
     * Returns true if this is a modulus substitution.  (We didn't do this
     * with instanceof partially because it causes source files to
     * proliferate and partially because we have to port this to C++.)
     * @return true if this object is an instance of ModulusSubstitution
     */
    virtual UBool isModulusSubstitution() const;
    
private:
    NFSubstitution(const NFSubstitution &other); // forbid copying of this class
    NFSubstitution &operator=(const NFSubstitution &other); // forbid copying of this class

public:
    static UClassID getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
};

U_NAMESPACE_END

/* U_HAVE_RBNF */
#endif

// NFSUBS_H




U_NAMESPACE_BEGIN

NFRule::NFRule(const RuleBasedNumberFormat* _rbnf, const UnicodeString &_ruleText, UErrorCode &status)
  : baseValue((int32_t)0)
  , radix(10)
  , exponent(0)
  , decimalPoint(0)
  , fRuleText(_ruleText)
  , sub1(NULL)
  , sub2(NULL)
  , formatter(_rbnf)
  , rulePatternFormat(NULL)
{
    if (!fRuleText.isEmpty()) {
        parseRuleDescriptor(fRuleText, status);
    }
}

NFRule::~NFRule()
{
    if (sub1 != sub2) {
        delete sub2;
        sub2 = NULL;
    }
    delete sub1;
    sub1 = NULL;
    delete rulePatternFormat;
    rulePatternFormat = NULL;
}

static const UChar nfrule_gLeftBracket = 0x005b;
static const UChar nfrule_gRightBracket = 0x005d;
static const UChar nfrule_gColon = 0x003a;
static const UChar nfrule_gZero = 0x0030;
static const UChar nfrule_gNine = 0x0039;
static const UChar nfrule_gSpace = 0x0020;
static const UChar nfrule_gSlash = 0x002f;
static const UChar nfrule_gGreaterThan = 0x003e;
static const UChar nfrule_gLessThan = 0x003c;
static const UChar nfrule_gComma = 0x002c;
static const UChar nfrule_gDot = 0x002e;
static const UChar nfrule_gTick = 0x0027;
//static const UChar gMinus = 0x002d;
static const UChar nfrule_gSemicolon = 0x003b;
static const UChar nfrule_gX = 0x0078;

static const UChar nfrule_gMinusX[] =                  {0x2D, 0x78, 0};    /* "-x" */
static const UChar nfrule_gInf[] =                     {0x49, 0x6E, 0x66, 0}; /* "Inf" */
static const UChar nfrule_gNaN[] =                     {0x4E, 0x61, 0x4E, 0}; /* "NaN" */

static const UChar nfrule_gDollarOpenParenthesis[] =   {0x24, 0x28, 0}; /* "$(" */
static const UChar nfrule_gClosedParenthesisDollar[] = {0x29, 0x24, 0}; /* ")$" */

static const UChar nfrule_gLessLess[] =                {0x3C, 0x3C, 0};    /* "<<" */
static const UChar nfrule_gLessPercent[] =             {0x3C, 0x25, 0};    /* "<%" */
static const UChar nfrule_gLessHash[] =                {0x3C, 0x23, 0};    /* "<#" */
static const UChar nfrule_gLessZero[] =                {0x3C, 0x30, 0};    /* "<0" */
static const UChar nfrule_gGreaterGreater[] =          {0x3E, 0x3E, 0};    /* ">>" */
static const UChar nfrule_gGreaterPercent[] =          {0x3E, 0x25, 0};    /* ">%" */
static const UChar nfrule_gGreaterHash[] =             {0x3E, 0x23, 0};    /* ">#" */
static const UChar nfrule_gGreaterZero[] =             {0x3E, 0x30, 0};    /* ">0" */
static const UChar nfrule_gEqualPercent[] =            {0x3D, 0x25, 0};    /* "=%" */
static const UChar nfrule_gEqualHash[] =               {0x3D, 0x23, 0};    /* "=#" */
static const UChar nfrule_gEqualZero[] =               {0x3D, 0x30, 0};    /* "=0" */
static const UChar nfrule_gGreaterGreaterGreater[] =   {0x3E, 0x3E, 0x3E, 0}; /* ">>>" */

static const UChar * const nfrule_RULE_PREFIXES[] = {
    nfrule_gLessLess, nfrule_gLessPercent, nfrule_gLessHash, nfrule_gLessZero,
    nfrule_gGreaterGreater, nfrule_gGreaterPercent,nfrule_gGreaterHash, nfrule_gGreaterZero,
    nfrule_gEqualPercent, nfrule_gEqualHash, nfrule_gEqualZero, NULL
};

void
NFRule::makeRules(UnicodeString& description,
                  NFRuleSet *owner,
                  const NFRule *predecessor,
                  const RuleBasedNumberFormat *rbnf,
                  NFRuleList& rules,
                  UErrorCode& status)
{
    // we know we're making at least one rule, so go ahead and
    // new it up and initialize its basevalue and divisor
    // (this also strips the rule descriptor, if any, off the
    // descripton string)
    NFRule* rule1 = new NFRule(rbnf, description, status);
    /* test for NULL */
    if (rule1 == 0) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    description = rule1->fRuleText;

    // check the description to see whether there's text enclosed
    // in brackets
    int32_t brack1 = description.indexOf(nfrule_gLeftBracket);
    int32_t brack2 = brack1 < 0 ? -1 : description.indexOf(nfrule_gRightBracket);

    // if the description doesn't contain a matched pair of brackets,
    // or if it's of a type that doesn't recognize bracketed text,
    // then leave the description alone, initialize the rule's
    // rule text and substitutions, and return that rule
    if (brack2 < 0 || brack1 > brack2
        || rule1->getType() == kProperFractionRule
        || rule1->getType() == kNegativeNumberRule
        || rule1->getType() == kInfinityRule
        || rule1->getType() == kNaNRule)
    {
        rule1->extractSubstitutions(owner, description, predecessor, status);
    }
    else {
        // if the description does contain a matched pair of brackets,
        // then it's really shorthand for two rules (with one exception)
        NFRule* rule2 = NULL;
        UnicodeString sbuf;

        // we'll actually only split the rule into two rules if its
        // base value is an even multiple of its divisor (or it's one
        // of the special rules)
        if ((rule1->baseValue > 0
            && (rule1->baseValue % util64_pow(rule1->radix, rule1->exponent)) == 0)
            || rule1->getType() == kImproperFractionRule
            || rule1->getType() == kMasterRule) {

            // if it passes that test, new up the second rule.  If the
            // rule set both rules will belong to is a fraction rule
            // set, they both have the same base value; otherwise,
            // increment the original rule's base value ("rule1" actually
            // goes SECOND in the rule set's rule list)
            rule2 = new NFRule(rbnf, UnicodeString(), status);
            /* test for NULL */
            if (rule2 == 0) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            if (rule1->baseValue >= 0) {
                rule2->baseValue = rule1->baseValue;
                if (!owner->isFractionRuleSet()) {
                    ++rule1->baseValue;
                }
            }

            // if the description began with "x.x" and contains bracketed
            // text, it describes both the improper fraction rule and
            // the proper fraction rule
            else if (rule1->getType() == kImproperFractionRule) {
                rule2->setType(kProperFractionRule);
            }

            // if the description began with "x.0" and contains bracketed
            // text, it describes both the master rule and the
            // improper fraction rule
            else if (rule1->getType() == kMasterRule) {
                rule2->baseValue = rule1->baseValue;
                rule1->setType(kImproperFractionRule);
            }

            // both rules have the same radix and exponent (i.e., the
            // same divisor)
            rule2->radix = rule1->radix;
            rule2->exponent = rule1->exponent;

            // rule2's rule text omits the stuff in brackets: initalize
            // its rule text and substitutions accordingly
            sbuf.append(description, 0, brack1);
            if (brack2 + 1 < description.length()) {
                sbuf.append(description, brack2 + 1, description.length() - brack2 - 1);
            }
            rule2->extractSubstitutions(owner, sbuf, predecessor, status);
        }

        // rule1's text includes the text in the brackets but omits
        // the brackets themselves: initialize _its_ rule text and
        // substitutions accordingly
        sbuf.setTo(description, 0, brack1);
        sbuf.append(description, brack1 + 1, brack2 - brack1 - 1);
        if (brack2 + 1 < description.length()) {
            sbuf.append(description, brack2 + 1, description.length() - brack2 - 1);
        }
        rule1->extractSubstitutions(owner, sbuf, predecessor, status);

        // if we only have one rule, return it; if we have two, return
        // a two-element array containing them (notice that rule2 goes
        // BEFORE rule1 in the list: in all cases, rule2 OMITS the
        // material in the brackets and rule1 INCLUDES the material
        // in the brackets)
        if (rule2 != NULL) {
            if (rule2->baseValue >= kNoBase) {
                rules.add(rule2);
            }
            else {
                owner->setNonNumericalRule(rule2);
            }
        }
    }
    if (rule1->baseValue >= kNoBase) {
        rules.add(rule1);
    }
    else {
        owner->setNonNumericalRule(rule1);
    }
}

/**
 * This function parses the rule's rule descriptor (i.e., the base
 * value and/or other tokens that precede the rule's rule text
 * in the description) and sets the rule's base value, radix, and
 * exponent according to the descriptor.  (If the description doesn't
 * include a rule descriptor, then this function sets everything to
 * default values and the rule set sets the rule's real base value).
 * @param description The rule's description
 * @return If "description" included a rule descriptor, this is
 * "description" with the descriptor and any trailing whitespace
 * stripped off.  Otherwise; it's "descriptor" unchangd.
 */
void
NFRule::parseRuleDescriptor(UnicodeString& description, UErrorCode& status)
{
    // the description consists of a rule descriptor and a rule body,
    // separated by a colon.  The rule descriptor is optional.  If
    // it's omitted, just set the base value to 0.
    int32_t p = description.indexOf(nfrule_gColon);
    if (p != -1) {
        // copy the descriptor out into its own string and strip it,
        // along with any trailing whitespace, out of the original
        // description
        UnicodeString descriptor;
        descriptor.setTo(description, 0, p);

        ++p;
        while (p < description.length() && PatternProps::isWhiteSpace(description.charAt(p))) {
            ++p;
        }
        description.removeBetween(0, p);

        // check first to see if the rule descriptor matches the token
        // for one of the special rules.  If it does, set the base
        // value to the correct identifier value
        int descriptorLength = descriptor.length();
        UChar firstChar = descriptor.charAt(0);
        UChar lastChar = descriptor.charAt(descriptorLength - 1);
        if (firstChar >= nfrule_gZero && firstChar <= nfrule_gNine && lastChar != nfrule_gX) {
            // if the rule descriptor begins with a digit, it's a descriptor
            // for a normal rule
            // since we don't have Long.parseLong, and this isn't much work anyway,
            // just build up the value as we encounter the digits.
            int64_t val = 0;
            p = 0;
            UChar c = nfrule_gSpace;

            // begin parsing the descriptor: copy digits
            // into "tempValue", skip periods, commas, and spaces,
            // stop on a slash or > sign (or at the end of the string),
            // and throw an exception on any other character
            int64_t ll_10 = 10;
            while (p < descriptorLength) {
                c = descriptor.charAt(p);
                if (c >= nfrule_gZero && c <= nfrule_gNine) {
                    val = val * ll_10 + (int32_t)(c - nfrule_gZero);
                }
                else if (c == nfrule_gSlash || c == nfrule_gGreaterThan) {
                    break;
                }
                else if (PatternProps::isWhiteSpace(c) || c == nfrule_gComma || c == nfrule_gDot) {
                }
                else {
                    // throw new IllegalArgumentException("Illegal character in rule descriptor");
                    status = U_PARSE_ERROR;
                    return;
                }
                ++p;
            }

            // we have the base value, so set it
            setBaseValue(val, status);

            // if we stopped the previous loop on a slash, we're
            // now parsing the rule's radix.  Again, accumulate digits
            // in tempValue, skip punctuation, stop on a > mark, and
            // throw an exception on anything else
            if (c == nfrule_gSlash) {
                val = 0;
                ++p;
                ll_10 = 10;
                while (p < descriptorLength) {
                    c = descriptor.charAt(p);
                    if (c >= nfrule_gZero && c <= nfrule_gNine) {
                        val = val * ll_10 + (int32_t)(c - nfrule_gZero);
                    }
                    else if (c == nfrule_gGreaterThan) {
                        break;
                    }
                    else if (PatternProps::isWhiteSpace(c) || c == nfrule_gComma || c == nfrule_gDot) {
                    }
                    else {
                        // throw new IllegalArgumentException("Illegal character is rule descriptor");
                        status = U_PARSE_ERROR;
                        return;
                    }
                    ++p;
                }

                // tempValue now contain's the rule's radix.  Set it
                // accordingly, and recalculate the rule's exponent
                radix = (int32_t)val;
                if (radix == 0) {
                    // throw new IllegalArgumentException("Rule can't have radix of 0");
                    status = U_PARSE_ERROR;
                }

                exponent = expectedExponent();
            }

            // if we stopped the previous loop on a > sign, then continue
            // for as long as we still see > signs.  For each one,
            // decrement the exponent (unless the exponent is already 0).
            // If we see another character before reaching the end of
            // the descriptor, that's also a syntax error.
            if (c == nfrule_gGreaterThan) {
                while (p < descriptor.length()) {
                    c = descriptor.charAt(p);
                    if (c == nfrule_gGreaterThan && exponent > 0) {
                        --exponent;
                    } else {
                        // throw new IllegalArgumentException("Illegal character in rule descriptor");
                        status = U_PARSE_ERROR;
                        return;
                    }
                    ++p;
                }
            }
        }
        else if (0 == descriptor.compare(nfrule_gMinusX, 2)) {
            setType(kNegativeNumberRule);
        }
        else if (descriptorLength == 3) {
            if (firstChar == nfrule_gZero && lastChar == nfrule_gX) {
                setBaseValue(kProperFractionRule, status);
                decimalPoint = descriptor.charAt(1);
            }
            else if (firstChar == nfrule_gX && lastChar == nfrule_gX) {
                setBaseValue(kImproperFractionRule, status);
                decimalPoint = descriptor.charAt(1);
            }
            else if (firstChar == nfrule_gX && lastChar == nfrule_gZero) {
                setBaseValue(kMasterRule, status);
                decimalPoint = descriptor.charAt(1);
            }
            else if (descriptor.compare(nfrule_gNaN, 3) == 0) {
                setBaseValue(kNaNRule, status);
            }
            else if (descriptor.compare(nfrule_gInf, 3) == 0) {
                setBaseValue(kInfinityRule, status);
            }
        }
    }
    // else use the default base value for now.

    // finally, if the rule body begins with an apostrophe, strip it off
    // (this is generally used to put whitespace at the beginning of
    // a rule's rule text)
    if (description.length() > 0 && description.charAt(0) == nfrule_gTick) {
        description.removeBetween(0, 1);
    }

    // return the description with all the stuff we've just waded through
    // stripped off the front.  It now contains just the rule body.
    // return description;
}

/**
* Searches the rule's rule text for the substitution tokens,
* creates the substitutions, and removes the substitution tokens
* from the rule's rule text.
* @param owner The rule set containing this rule
* @param predecessor The rule preseding this one in "owners" rule list
* @param ownersOwner The RuleBasedFormat that owns this rule
*/
void
NFRule::extractSubstitutions(const NFRuleSet* ruleSet,
                             const UnicodeString &ruleText,
                             const NFRule* predecessor,
                             UErrorCode& status)
{
    if (U_FAILURE(status)) {
        return;
    }
    fRuleText = ruleText;
    sub1 = extractSubstitution(ruleSet, predecessor, status);
    if (sub1 == NULL) {
        // Small optimization. There is no need to create a redundant NullSubstitution.
        sub2 = NULL;
    }
    else {
        sub2 = extractSubstitution(ruleSet, predecessor, status);
    }
    int32_t pluralRuleStart = fRuleText.indexOf(nfrule_gDollarOpenParenthesis, -1, 0);
    int32_t pluralRuleEnd = (pluralRuleStart >= 0 ? fRuleText.indexOf(nfrule_gClosedParenthesisDollar, -1, pluralRuleStart) : -1);
    if (pluralRuleEnd >= 0) {
        int32_t endType = fRuleText.indexOf(nfrule_gComma, pluralRuleStart);
        if (endType < 0) {
            status = U_PARSE_ERROR;
            return;
        }
        UnicodeString type(fRuleText.tempSubString(pluralRuleStart + 2, endType - pluralRuleStart - 2));
        UPluralType pluralType;
        if (type.startsWith(UNICODE_STRING_SIMPLE("cardinal"))) {
            pluralType = UPLURAL_TYPE_CARDINAL;
        }
        else if (type.startsWith(UNICODE_STRING_SIMPLE("ordinal"))) {
            pluralType = UPLURAL_TYPE_ORDINAL;
        }
        else {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        rulePatternFormat = formatter->createPluralFormat(pluralType,
                fRuleText.tempSubString(endType + 1, pluralRuleEnd - endType - 1), status);
    }
}

/**
* Searches the rule's rule text for the first substitution token,
* creates a substitution based on it, and removes the token from
* the rule's rule text.
* @param owner The rule set containing this rule
* @param predecessor The rule preceding this one in the rule set's
* rule list
* @param ownersOwner The RuleBasedNumberFormat that owns this rule
* @return The newly-created substitution.  This is never null; if
* the rule text doesn't contain any substitution tokens, this will
* be a NullSubstitution.
*/
NFSubstitution *
NFRule::extractSubstitution(const NFRuleSet* ruleSet,
                            const NFRule* predecessor,
                            UErrorCode& status)
{
    NFSubstitution* result = NULL;

    // search the rule's rule text for the first two characters of
    // a substitution token
    int32_t subStart = indexOfAnyRulePrefix();
    int32_t subEnd = subStart;

    // if we didn't find one, create a null substitution positioned
    // at the end of the rule text
    if (subStart == -1) {
        return NULL;
    }

    // special-case the ">>>" token, since searching for the > at the
    // end will actually find the > in the middle
    if (fRuleText.indexOf(nfrule_gGreaterGreaterGreater, 3, 0) == subStart) {
        subEnd = subStart + 2;

        // otherwise the substitution token ends with the same character
        // it began with
    } else {
        UChar c = fRuleText.charAt(subStart);
        subEnd = fRuleText.indexOf(c, subStart + 1);
        // special case for '<%foo<<'
        if (c == nfrule_gLessThan && subEnd != -1 && subEnd < fRuleText.length() - 1 && fRuleText.charAt(subEnd+1) == c) {
            // ordinals use "=#,##0==%abbrev=" as their rule.  Notice that the '==' in the middle
            // occurs because of the juxtaposition of two different rules.  The check for '<' is a hack
            // to get around this.  Having the duplicate at the front would cause problems with
            // rules like "<<%" to format, say, percents...
            ++subEnd;
        }
   }

    // if we don't find the end of the token (i.e., if we're on a single,
    // unmatched token character), create a null substitution positioned
    // at the end of the rule
    if (subEnd == -1) {
        return NULL;
    }

    // if we get here, we have a real substitution token (or at least
    // some text bounded by substitution token characters).  Use
    // makeSubstitution() to create the right kind of substitution
    UnicodeString subToken;
    subToken.setTo(fRuleText, subStart, subEnd + 1 - subStart);
    result = NFSubstitution::makeSubstitution(subStart, this, predecessor, ruleSet,
        this->formatter, subToken, status);

    // remove the substitution from the rule text
    fRuleText.removeBetween(subStart, subEnd+1);

    return result;
}

/**
 * Sets the rule's base value, and causes the radix and exponent
 * to be recalculated.  This is used during construction when we
 * don't know the rule's base value until after it's been
 * constructed.  It should be used at any other time.
 * @param The new base value for the rule.
 */
void
NFRule::setBaseValue(int64_t newBaseValue, UErrorCode& status)
{
    // set the base value
    baseValue = newBaseValue;
    radix = 10;

    // if this isn't a special rule, recalculate the radix and exponent
    // (the radix always defaults to 10; if it's supposed to be something
    // else, it's cleaned up by the caller and the exponent is
    // recalculated again-- the only function that does this is
    // NFRule.parseRuleDescriptor() )
    if (baseValue >= 1) {
        exponent = expectedExponent();

        // this function gets called on a fully-constructed rule whose
        // description didn't specify a base value.  This means it
        // has substitutions, and some substitutions hold on to copies
        // of the rule's divisor.  Fix their copies of the divisor.
        if (sub1 != NULL) {
            sub1->setDivisor(radix, exponent, status);
        }
        if (sub2 != NULL) {
            sub2->setDivisor(radix, exponent, status);
        }

        // if this is a special rule, its radix and exponent are basically
        // ignored.  Set them to "safe" default values
    } else {
        exponent = 0;
    }
}

/**
* This calculates the rule's exponent based on its radix and base
* value.  This will be the highest power the radix can be raised to
* and still produce a result less than or equal to the base value.
*/
int16_t
NFRule::expectedExponent() const
{
    // since the log of 0, or the log base 0 of something, causes an
    // error, declare the exponent in these cases to be 0 (we also
    // deal with the special-rule identifiers here)
    if (radix == 0 || baseValue < 1) {
        return 0;
    }

    // we get rounding error in some cases-- for example, log 1000 / log 10
    // gives us 1.9999999996 instead of 2.  The extra logic here is to take
    // that into account
    int16_t tempResult = (int16_t)(uprv_log((double)baseValue) / uprv_log((double)radix));
    int64_t temp = util64_pow(radix, tempResult + 1);
    if (temp <= baseValue) {
        tempResult += 1;
    }
    return tempResult;
}

/**
 * Searches the rule's rule text for any of the specified strings.
 * @return The index of the first match in the rule's rule text
 * (i.e., the first substring in the rule's rule text that matches
 * _any_ of the strings in "strings").  If none of the strings in
 * "strings" is found in the rule's rule text, returns -1.
 */
int32_t
NFRule::indexOfAnyRulePrefix() const
{
    int result = -1;
    for (int i = 0; nfrule_RULE_PREFIXES[i]; i++) {
        int32_t pos = fRuleText.indexOf(*nfrule_RULE_PREFIXES[i]);
        if (pos != -1 && (result == -1 || pos < result)) {
            result = pos;
        }
    }
    return result;
}

//-----------------------------------------------------------------------
// boilerplate
//-----------------------------------------------------------------------

static UBool
util_equalSubstitutions(const NFSubstitution* sub1, const NFSubstitution* sub2)
{
    if (sub1) {
        if (sub2) {
            return *sub1 == *sub2;
        }
    } else if (!sub2) {
        return TRUE;
    }
    return FALSE;
}

/**
* Tests two rules for equality.
* @param that The rule to compare this one against
* @return True is the two rules are functionally equivalent
*/
UBool
NFRule::operator==(const NFRule& rhs) const
{
    return baseValue == rhs.baseValue
        && radix == rhs.radix
        && exponent == rhs.exponent
        && fRuleText == rhs.fRuleText
        && util_equalSubstitutions(sub1, rhs.sub1)
        && util_equalSubstitutions(sub2, rhs.sub2);
}

/**
* Returns a textual representation of the rule.  This won't
* necessarily be the same as the description that this rule
* was created with, but it will produce the same result.
* @return A textual description of the rule
*/
static void util_append64(UnicodeString& result, int64_t n)
{
    UChar buffer[256];
    int32_t len = util64_tou(n, buffer, sizeof(buffer));
    UnicodeString temp(buffer, len);
    result.append(temp);
}

void
NFRule::_appendRuleText(UnicodeString& result) const
{
    switch (getType()) {
    case kNegativeNumberRule: result.append(nfrule_gMinusX, 2); break;
    case kImproperFractionRule: result.append(nfrule_gX).append(decimalPoint == 0 ? nfrule_gDot : decimalPoint).append(nfrule_gX); break;
    case kProperFractionRule: result.append(nfrule_gZero).append(decimalPoint == 0 ? nfrule_gDot : decimalPoint).append(nfrule_gX); break;
    case kMasterRule: result.append(nfrule_gX).append(decimalPoint == 0 ? nfrule_gDot : decimalPoint).append(nfrule_gZero); break;
    case kInfinityRule: result.append(nfrule_gInf, 3); break;
    case kNaNRule: result.append(nfrule_gNaN, 3); break;
    default:
        // for a normal rule, write out its base value, and if the radix is
        // something other than 10, write out the radix (with the preceding
        // slash, of course).  Then calculate the expected exponent and if
        // if isn't the same as the actual exponent, write an appropriate
        // number of > signs.  Finally, terminate the whole thing with
        // a colon.
        util_append64(result, baseValue);
        if (radix != 10) {
            result.append(nfrule_gSlash);
            util_append64(result, radix);
        }
        int numCarets = expectedExponent() - exponent;
        for (int i = 0; i < numCarets; i++) {
            result.append(nfrule_gGreaterThan);
        }
        break;
    }
    result.append(nfrule_gColon);
    result.append(nfrule_gSpace);

    // if the rule text begins with a space, write an apostrophe
    // (whitespace after the rule descriptor is ignored; the
    // apostrophe is used to make the whitespace significant)
    if (fRuleText.charAt(0) == nfrule_gSpace && (sub1 == NULL || sub1->getPos() != 0)) {
        result.append(nfrule_gTick);
    }

    // now, write the rule's rule text, inserting appropriate
    // substitution tokens in the appropriate places
    UnicodeString ruleTextCopy;
    ruleTextCopy.setTo(fRuleText);

    UnicodeString temp;
    if (sub2 != NULL) {
        sub2->toString(temp);
        ruleTextCopy.insert(sub2->getPos(), temp);
    }
    if (sub1 != NULL) {
        sub1->toString(temp);
        ruleTextCopy.insert(sub1->getPos(), temp);
    }

    result.append(ruleTextCopy);

    // and finally, top the whole thing off with a semicolon and
    // return the result
    result.append(nfrule_gSemicolon);
}

int64_t NFRule::getDivisor() const
{
    return util64_pow(radix, exponent);
}


//-----------------------------------------------------------------------
// formatting
//-----------------------------------------------------------------------

/**
* Formats the number, and inserts the resulting text into
* toInsertInto.
* @param number The number being formatted
* @param toInsertInto The string where the resultant text should
* be inserted
* @param pos The position in toInsertInto where the resultant text
* should be inserted
*/
void
NFRule::doFormat(int64_t number, UnicodeString& toInsertInto, int32_t pos, int32_t recursionCount, UErrorCode& status) const
{
    // first, insert the rule's rule text into toInsertInto at the
    // specified position, then insert the results of the substitutions
    // into the right places in toInsertInto (notice we do the
    // substitutions in reverse order so that the offsets don't get
    // messed up)
    int32_t pluralRuleStart = fRuleText.length();
    int32_t lengthOffset = 0;
    if (!rulePatternFormat) {
        toInsertInto.insert(pos, fRuleText);
    }
    else {
        pluralRuleStart = fRuleText.indexOf(nfrule_gDollarOpenParenthesis, -1, 0);
        int pluralRuleEnd = fRuleText.indexOf(nfrule_gClosedParenthesisDollar, -1, pluralRuleStart);
        int initialLength = toInsertInto.length();
        if (pluralRuleEnd < fRuleText.length() - 1) {
            toInsertInto.insert(pos, fRuleText.tempSubString(pluralRuleEnd + 2));
        }
        toInsertInto.insert(pos,
            rulePatternFormat->format((int32_t)(number/util64_pow(radix, exponent)), status));
        if (pluralRuleStart > 0) {
            toInsertInto.insert(pos, fRuleText.tempSubString(0, pluralRuleStart));
        }
        lengthOffset = fRuleText.length() - (toInsertInto.length() - initialLength);
    }

    if (sub2 != NULL) {
        sub2->doSubstitution(number, toInsertInto, pos - (sub2->getPos() > pluralRuleStart ? lengthOffset : 0), recursionCount, status);
    }
    if (sub1 != NULL) {
        sub1->doSubstitution(number, toInsertInto, pos - (sub1->getPos() > pluralRuleStart ? lengthOffset : 0), recursionCount, status);
    }
}

/**
* Formats the number, and inserts the resulting text into
* toInsertInto.
* @param number The number being formatted
* @param toInsertInto The string where the resultant text should
* be inserted
* @param pos The position in toInsertInto where the resultant text
* should be inserted
*/
void
NFRule::doFormat(double number, UnicodeString& toInsertInto, int32_t pos, int32_t recursionCount, UErrorCode& status) const
{
    // first, insert the rule's rule text into toInsertInto at the
    // specified position, then insert the results of the substitutions
    // into the right places in toInsertInto
    // [again, we have two copies of this routine that do the same thing
    // so that we don't sacrifice precision in a long by casting it
    // to a double]
    int32_t pluralRuleStart = fRuleText.length();
    int32_t lengthOffset = 0;
    if (!rulePatternFormat) {
        toInsertInto.insert(pos, fRuleText);
    }
    else {
        pluralRuleStart = fRuleText.indexOf(nfrule_gDollarOpenParenthesis, -1, 0);
        int pluralRuleEnd = fRuleText.indexOf(nfrule_gClosedParenthesisDollar, -1, pluralRuleStart);
        int initialLength = toInsertInto.length();
        if (pluralRuleEnd < fRuleText.length() - 1) {
            toInsertInto.insert(pos, fRuleText.tempSubString(pluralRuleEnd + 2));
        }
        double pluralVal = number;
        if (0 <= pluralVal && pluralVal < 1) {
            // We're in a fractional rule, and we have to match the NumeratorSubstitution behavior.
            // 2.3 can become 0.2999999999999998 for the fraction due to rounding errors.
            pluralVal = uprv_round(pluralVal * util64_pow(radix, exponent));
        }
        else {
            pluralVal = pluralVal / util64_pow(radix, exponent);
        }
        toInsertInto.insert(pos, rulePatternFormat->format((int32_t)(pluralVal), status));
        if (pluralRuleStart > 0) {
            toInsertInto.insert(pos, fRuleText.tempSubString(0, pluralRuleStart));
        }
        lengthOffset = fRuleText.length() - (toInsertInto.length() - initialLength);
    }

    if (sub2 != NULL) {
        sub2->doSubstitution(number, toInsertInto, pos - (sub2->getPos() > pluralRuleStart ? lengthOffset : 0), recursionCount, status);
    }
    if (sub1 != NULL) {
        sub1->doSubstitution(number, toInsertInto, pos - (sub1->getPos() > pluralRuleStart ? lengthOffset : 0), recursionCount, status);
    }
}

/**
* Used by the owning rule set to determine whether to invoke the
* rollback rule (i.e., whether this rule or the one that precedes
* it in the rule set's list should be used to format the number)
* @param The number being formatted
* @return True if the rule set should use the rule that precedes
* this one in its list; false if it should use this rule
*/
UBool
NFRule::shouldRollBack(int64_t number) const
{
    // we roll back if the rule contains a modulus substitution,
    // the number being formatted is an even multiple of the rule's
    // divisor, and the rule's base value is NOT an even multiple
    // of its divisor
    // In other words, if the original description had
    //    100: << hundred[ >>];
    // that expands into
    //    100: << hundred;
    //    101: << hundred >>;
    // internally.  But when we're formatting 200, if we use the rule
    // at 101, which would normally apply, we get "two hundred zero".
    // To prevent this, we roll back and use the rule at 100 instead.
    // This is the logic that makes this happen: the rule at 101 has
    // a modulus substitution, its base value isn't an even multiple
    // of 100, and the value we're trying to format _is_ an even
    // multiple of 100.  This is called the "rollback rule."
    if ((sub1 != NULL && sub1->isModulusSubstitution()) || (sub2 != NULL && sub2->isModulusSubstitution())) {
        int64_t re = util64_pow(radix, exponent);
        return (number % re) == 0 && (baseValue % re) != 0;
    }
    return FALSE;
}

//-----------------------------------------------------------------------
// parsing
//-----------------------------------------------------------------------

/**
* Attempts to parse the string with this rule.
* @param text The string being parsed
* @param parsePosition On entry, the value is ignored and assumed to
* be 0. On exit, this has been updated with the position of the first
* character not consumed by matching the text against this rule
* (if this rule doesn't match the text at all, the parse position
* if left unchanged (presumably at 0) and the function returns
* new Long(0)).
* @param isFractionRule True if this rule is contained within a
* fraction rule set.  This is only used if the rule has no
* substitutions.
* @return If this rule matched the text, this is the rule's base value
* combined appropriately with the results of parsing the substitutions.
* If nothing matched, this is new Long(0) and the parse position is
* left unchanged.  The result will be an instance of Long if the
* result is an integer and Double otherwise.  The result is never null.
*/
#ifdef RBNF_DEBUG
#include <stdio.h>

static void dumpUS(FILE* f, const UnicodeString& us) {
  int len = us.length();
  char* buf = (char *)uprv_malloc((len+1)*sizeof(char)); //new char[len+1];
  if (buf != NULL) {
	  us.extract(0, len, buf);
	  buf[len] = 0;
	  fprintf(f, "%s", buf);
	  uprv_free(buf); //delete[] buf;
  }
}
#endif
UBool
NFRule::doParse(const UnicodeString& text,
                ParsePosition& parsePosition,
                UBool isFractionRule,
                double upperBound,
                uint32_t nonNumericalExecutedRuleMask,
                Formattable& resVal) const
{
    // internally we operate on a copy of the string being parsed
    // (because we're going to change it) and use our own ParsePosition
    ParsePosition pp;
    UnicodeString workText(text);

    int32_t sub1Pos = sub1 != NULL ? sub1->getPos() : fRuleText.length();
    int32_t sub2Pos = sub2 != NULL ? sub2->getPos() : fRuleText.length();

    // check to see whether the text before the first substitution
    // matches the text at the beginning of the string being
    // parsed.  If it does, strip that off the front of workText;
    // otherwise, dump out with a mismatch
    UnicodeString prefix;
    prefix.setTo(fRuleText, 0, sub1Pos);

#ifdef RBNF_DEBUG
    fprintf(stderr, "doParse %p ", this);
    {
        UnicodeString rt;
        _appendRuleText(rt);
        dumpUS(stderr, rt);
    }

    fprintf(stderr, " text: '");
    dumpUS(stderr, text);
    fprintf(stderr, "' prefix: '");
    dumpUS(stderr, prefix);
#endif
    stripPrefix(workText, prefix, pp);
    int32_t prefixLength = text.length() - workText.length();

#ifdef RBNF_DEBUG
    fprintf(stderr, "' pl: %d ppi: %d s1p: %d\n", prefixLength, pp.getIndex(), sub1Pos);
#endif

    if (pp.getIndex() == 0 && sub1Pos != 0) {
        // commented out because ParsePosition doesn't have error index in 1.1.x
        // restored for ICU4C port
        parsePosition.setErrorIndex(pp.getErrorIndex());
        resVal.setLong(0);
        return TRUE;
    }
    if (baseValue == kInfinityRule) {
        // If you match this, don't try to perform any calculations on it.
        parsePosition.setIndex(pp.getIndex());
        resVal.setDouble(uprv_getInfinity());
        return TRUE;
    }
    if (baseValue == kNaNRule) {
        // If you match this, don't try to perform any calculations on it.
        parsePosition.setIndex(pp.getIndex());
        resVal.setDouble(uprv_getNaN());
        return TRUE;
    }

    // this is the fun part.  The basic guts of the rule-matching
    // logic is matchToDelimiter(), which is called twice.  The first
    // time it searches the input string for the rule text BETWEEN
    // the substitutions and tries to match the intervening text
    // in the input string with the first substitution.  If that
    // succeeds, it then calls it again, this time to look for the
    // rule text after the second substitution and to match the
    // intervening input text against the second substitution.
    //
    // For example, say we have a rule that looks like this:
    //    first << middle >> last;
    // and input text that looks like this:
    //    first one middle two last
    // First we use stripPrefix() to match "first " in both places and
    // strip it off the front, leaving
    //    one middle two last
    // Then we use matchToDelimiter() to match " middle " and try to
    // match "one" against a substitution.  If it's successful, we now
    // have
    //    two last
    // We use matchToDelimiter() a second time to match " last" and
    // try to match "two" against a substitution.  If "two" matches
    // the substitution, we have a successful parse.
    //
    // Since it's possible in many cases to find multiple instances
    // of each of these pieces of rule text in the input string,
    // we need to try all the possible combinations of these
    // locations.  This prevents us from prematurely declaring a mismatch,
    // and makes sure we match as much input text as we can.
    int highWaterMark = 0;
    double result = 0;
    int start = 0;
    double tempBaseValue = (double)(baseValue <= 0 ? 0 : baseValue);

    UnicodeString temp;
    do {
        // our partial parse result starts out as this rule's base
        // value.  If it finds a successful match, matchToDelimiter()
        // will compose this in some way with what it gets back from
        // the substitution, giving us a new partial parse result
        pp.setIndex(0);

        temp.setTo(fRuleText, sub1Pos, sub2Pos - sub1Pos);
        double partialResult = matchToDelimiter(workText, start, tempBaseValue,
            temp, pp, sub1,
            nonNumericalExecutedRuleMask,
            upperBound);

        // if we got a successful match (or were trying to match a
        // null substitution), pp is now pointing at the first unmatched
        // character.  Take note of that, and try matchToDelimiter()
        // on the input text again
        if (pp.getIndex() != 0 || sub1 == NULL) {
            start = pp.getIndex();

            UnicodeString workText2;
            workText2.setTo(workText, pp.getIndex(), workText.length() - pp.getIndex());
            ParsePosition pp2;

            // the second matchToDelimiter() will compose our previous
            // partial result with whatever it gets back from its
            // substitution if there's a successful match, giving us
            // a real result
            temp.setTo(fRuleText, sub2Pos, fRuleText.length() - sub2Pos);
            partialResult = matchToDelimiter(workText2, 0, partialResult,
                temp, pp2, sub2,
                nonNumericalExecutedRuleMask,
                upperBound);

            // if we got a successful match on this second
            // matchToDelimiter() call, update the high-water mark
            // and result (if necessary)
            if (pp2.getIndex() != 0 || sub2 == NULL) {
                if (prefixLength + pp.getIndex() + pp2.getIndex() > highWaterMark) {
                    highWaterMark = prefixLength + pp.getIndex() + pp2.getIndex();
                    result = partialResult;
                }
            }
            else {
                // commented out because ParsePosition doesn't have error index in 1.1.x
                // restored for ICU4C port
                int32_t i_temp = pp2.getErrorIndex() + sub1Pos + pp.getIndex();
                if (i_temp> parsePosition.getErrorIndex()) {
                    parsePosition.setErrorIndex(i_temp);
                }
            }
        }
        else {
            // commented out because ParsePosition doesn't have error index in 1.1.x
            // restored for ICU4C port
            int32_t i_temp = sub1Pos + pp.getErrorIndex();
            if (i_temp > parsePosition.getErrorIndex()) {
                parsePosition.setErrorIndex(i_temp);
            }
        }
        // keep trying to match things until the outer matchToDelimiter()
        // call fails to make a match (each time, it picks up where it
        // left off the previous time)
    } while (sub1Pos != sub2Pos
        && pp.getIndex() > 0
        && pp.getIndex() < workText.length()
        && pp.getIndex() != start);

    // update the caller's ParsePosition with our high-water mark
    // (i.e., it now points at the first character this function
    // didn't match-- the ParsePosition is therefore unchanged if
    // we didn't match anything)
    parsePosition.setIndex(highWaterMark);
    // commented out because ParsePosition doesn't have error index in 1.1.x
    // restored for ICU4C port
    if (highWaterMark > 0) {
        parsePosition.setErrorIndex(0);
    }

    // this is a hack for one unusual condition: Normally, whether this
    // rule belong to a fraction rule set or not is handled by its
    // substitutions.  But if that rule HAS NO substitutions, then
    // we have to account for it here.  By definition, if the matching
    // rule in a fraction rule set has no substitutions, its numerator
    // is 1, and so the result is the reciprocal of its base value.
    if (isFractionRule && highWaterMark > 0 && sub1 == NULL) {
        result = 1 / result;
    }

    resVal.setDouble(result);
    return TRUE; // ??? do we need to worry if it is a long or a double?
}

/**
* This function is used by parse() to match the text being parsed
* against a possible prefix string.  This function
* matches characters from the beginning of the string being parsed
* to characters from the prospective prefix.  If they match, pp is
* updated to the first character not matched, and the result is
* the unparsed part of the string.  If they don't match, the whole
* string is returned, and pp is left unchanged.
* @param text The string being parsed
* @param prefix The text to match against
* @param pp On entry, ignored and assumed to be 0.  On exit, points
* to the first unmatched character (assuming the whole prefix matched),
* or is unchanged (if the whole prefix didn't match).
* @return If things match, this is the unparsed part of "text";
* if they didn't match, this is "text".
*/
void
NFRule::stripPrefix(UnicodeString& text, const UnicodeString& prefix, ParsePosition& pp) const
{
    // if the prefix text is empty, dump out without doing anything
    if (prefix.length() != 0) {
    	UErrorCode status = U_ZERO_ERROR;
        // use prefixLength() to match the beginning of
        // "text" against "prefix".  This function returns the
        // number of characters from "text" that matched (or 0 if
        // we didn't match the whole prefix)
        int32_t pfl = prefixLength(text, prefix, status);
        if (U_FAILURE(status)) { // Memory allocation error.
        	return;
        }
        if (pfl != 0) {
            // if we got a successful match, update the parse position
            // and strip the prefix off of "text"
            pp.setIndex(pp.getIndex() + pfl);
            text.remove(0, pfl);
        }
    }
}

/**
* Used by parse() to match a substitution and any following text.
* "text" is searched for instances of "delimiter".  For each instance
* of delimiter, the intervening text is tested to see whether it
* matches the substitution.  The longest match wins.
* @param text The string being parsed
* @param startPos The position in "text" where we should start looking
* for "delimiter".
* @param baseValue A partial parse result (often the rule's base value),
* which is combined with the result from matching the substitution
* @param delimiter The string to search "text" for.
* @param pp Ignored and presumed to be 0 on entry.  If there's a match,
* on exit this will point to the first unmatched character.
* @param sub If we find "delimiter" in "text", this substitution is used
* to match the text between the beginning of the string and the
* position of "delimiter."  (If "delimiter" is the empty string, then
* this function just matches against this substitution and updates
* everything accordingly.)
* @param upperBound When matching the substitution, it will only
* consider rules with base values lower than this value.
* @return If there's a match, this is the result of composing
* baseValue with the result of matching the substitution.  Otherwise,
* this is new Long(0).  It's never null.  If the result is an integer,
* this will be an instance of Long; otherwise, it's an instance of
* Double.
*
* !!! note {dlf} in point of fact, in the java code the caller always converts
* the result to a double, so we might as well return one.
*/
double
NFRule::matchToDelimiter(const UnicodeString& text,
                         int32_t startPos,
                         double _baseValue,
                         const UnicodeString& delimiter,
                         ParsePosition& pp,
                         const NFSubstitution* sub,
                         uint32_t nonNumericalExecutedRuleMask,
                         double upperBound) const
{
	UErrorCode status = U_ZERO_ERROR;
    // if "delimiter" contains real (i.e., non-ignorable) text, search
    // it for "delimiter" beginning at "start".  If that succeeds, then
    // use "sub"'s doParse() method to match the text before the
    // instance of "delimiter" we just found.
    if (!allIgnorable(delimiter, status)) {
    	if (U_FAILURE(status)) { //Memory allocation error.
    		return 0;
    	}
        ParsePosition tempPP;
        Formattable result;

        // use findText() to search for "delimiter".  It returns a two-
        // element array: element 0 is the position of the match, and
        // element 1 is the number of characters that matched
        // "delimiter".
        int32_t dLen;
        int32_t dPos = findText(text, delimiter, startPos, &dLen);

        // if findText() succeeded, isolate the text preceding the
        // match, and use "sub" to match that text
        while (dPos >= 0) {
            UnicodeString subText;
            subText.setTo(text, 0, dPos);
            if (subText.length() > 0) {
                UBool success = sub->doParse(subText, tempPP, _baseValue, upperBound,
#if UCONFIG_NO_COLLATION
                    FALSE,
#else
                    formatter->isLenient(),
#endif
                    nonNumericalExecutedRuleMask,
                    result);

                // if the substitution could match all the text up to
                // where we found "delimiter", then this function has
                // a successful match.  Bump the caller's parse position
                // to point to the first character after the text
                // that matches "delimiter", and return the result
                // we got from parsing the substitution.
                if (success && tempPP.getIndex() == dPos) {
                    pp.setIndex(dPos + dLen);
                    return result.getDouble();
                }
                else {
                    // commented out because ParsePosition doesn't have error index in 1.1.x
                    // restored for ICU4C port
                    if (tempPP.getErrorIndex() > 0) {
                        pp.setErrorIndex(tempPP.getErrorIndex());
                    } else {
                        pp.setErrorIndex(tempPP.getIndex());
                    }
                }
            }

            // if we didn't match the substitution, search for another
            // copy of "delimiter" in "text" and repeat the loop if
            // we find it
            tempPP.setIndex(0);
            dPos = findText(text, delimiter, dPos + dLen, &dLen);
        }
        // if we make it here, this was an unsuccessful match, and we
        // leave pp unchanged and return 0
        pp.setIndex(0);
        return 0;

        // if "delimiter" is empty, or consists only of ignorable characters
        // (i.e., is semantically empty), thwe we obviously can't search
        // for "delimiter".  Instead, just use "sub" to parse as much of
        // "text" as possible.
    }
    else if (sub == NULL) {
        return _baseValue;
    }
    else {
        ParsePosition tempPP;
        Formattable result;

        // try to match the whole string against the substitution
        UBool success = sub->doParse(text, tempPP, _baseValue, upperBound,
#if UCONFIG_NO_COLLATION
            FALSE,
#else
            formatter->isLenient(),
#endif
            nonNumericalExecutedRuleMask,
            result);
        if (success && (tempPP.getIndex() != 0)) {
            // if there's a successful match (or it's a null
            // substitution), update pp to point to the first
            // character we didn't match, and pass the result from
            // sub.doParse() on through to the caller
            pp.setIndex(tempPP.getIndex());
            return result.getDouble();
        }
        else {
            // commented out because ParsePosition doesn't have error index in 1.1.x
            // restored for ICU4C port
            pp.setErrorIndex(tempPP.getErrorIndex());
        }

        // and if we get to here, then nothing matched, so we return
        // 0 and leave pp alone
        return 0;
    }
}

/**
* Used by stripPrefix() to match characters.  If lenient parse mode
* is off, this just calls startsWith().  If lenient parse mode is on,
* this function uses CollationElementIterators to match characters in
* the strings (only primary-order differences are significant in
* determining whether there's a match).
* @param str The string being tested
* @param prefix The text we're hoping to see at the beginning
* of "str"
* @return If "prefix" is found at the beginning of "str", this
* is the number of characters in "str" that were matched (this
* isn't necessarily the same as the length of "prefix" when matching
* text with a collator).  If there's no match, this is 0.
*/
int32_t
NFRule::prefixLength(const UnicodeString& str, const UnicodeString& prefix, UErrorCode& status) const
{
    // if we're looking for an empty prefix, it obviously matches
    // zero characters.  Just go ahead and return 0.
    if (prefix.length() == 0) {
        return 0;
    }

#if !UCONFIG_NO_COLLATION
    // go through all this grief if we're in lenient-parse mode
    if (formatter->isLenient()) {
        // get the formatter's collator and use it to create two
        // collation element iterators, one over the target string
        // and another over the prefix (right now, we'll throw an
        // exception if the collator we get back from the formatter
        // isn't a RuleBasedCollator, because RuleBasedCollator defines
        // the CollationElementIterator protocol.  Hopefully, this
        // will change someday.)
        const RuleBasedCollator* collator = formatter->getCollator();
        if (collator == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return 0;
        }
        LocalPointer<CollationElementIterator> strIter(collator->createCollationElementIterator(str));
        LocalPointer<CollationElementIterator> prefixIter(collator->createCollationElementIterator(prefix));
        // Check for memory allocation error.
        if (strIter.isNull() || prefixIter.isNull()) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return 0;
        }

        UErrorCode err = U_ZERO_ERROR;

        // The original code was problematic.  Consider this match:
        // prefix = "fifty-"
        // string = " fifty-7"
        // The intent is to match string up to the '7', by matching 'fifty-' at position 1
        // in the string.  Unfortunately, we were getting a match, and then computing where
        // the match terminated by rematching the string.  The rematch code was using as an
        // initial guess the substring of string between 0 and prefix.length.  Because of
        // the leading space and trailing hyphen (both ignorable) this was succeeding, leaving
        // the position before the hyphen in the string.  Recursing down, we then parsed the
        // remaining string '-7' as numeric.  The resulting number turned out as 43 (50 - 7).
        // This was not pretty, especially since the string "fifty-7" parsed just fine.
        //
        // We have newer APIs now, so we can use calls on the iterator to determine what we
        // matched up to.  If we terminate because we hit the last element in the string,
        // our match terminates at this length.  If we terminate because we hit the last element
        // in the target, our match terminates at one before the element iterator position.

        // match collation elements between the strings
        int32_t oStr = strIter->next(err);
        int32_t oPrefix = prefixIter->next(err);

        while (oPrefix != CollationElementIterator::NULLORDER) {
            // skip over ignorable characters in the target string
            while (CollationElementIterator::primaryOrder(oStr) == 0
                && oStr != CollationElementIterator::NULLORDER) {
                oStr = strIter->next(err);
            }

            // skip over ignorable characters in the prefix
            while (CollationElementIterator::primaryOrder(oPrefix) == 0
                && oPrefix != CollationElementIterator::NULLORDER) {
                oPrefix = prefixIter->next(err);
            }

            // dlf: move this above following test, if we consume the
            // entire target, aren't we ok even if the source was also
            // entirely consumed?

            // if skipping over ignorables brought to the end of
            // the prefix, we DID match: drop out of the loop
            if (oPrefix == CollationElementIterator::NULLORDER) {
                break;
            }

            // if skipping over ignorables brought us to the end
            // of the target string, we didn't match and return 0
            if (oStr == CollationElementIterator::NULLORDER) {
                return 0;
            }

            // match collation elements from the two strings
            // (considering only primary differences).  If we
            // get a mismatch, dump out and return 0
            if (CollationElementIterator::primaryOrder(oStr)
                != CollationElementIterator::primaryOrder(oPrefix)) {
                return 0;

                // otherwise, advance to the next character in each string
                // and loop (we drop out of the loop when we exhaust
                // collation elements in the prefix)
            } else {
                oStr = strIter->next(err);
                oPrefix = prefixIter->next(err);
            }
        }

        int32_t result = strIter->getOffset();
        if (oStr != CollationElementIterator::NULLORDER) {
            --result; // back over character that we don't want to consume;
        }

#ifdef RBNF_DEBUG
        fprintf(stderr, "prefix length: %d\n", result);
#endif
        return result;
#if 0
        //----------------------------------------------------------------
        // JDK 1.2-specific API call
        // return strIter.getOffset();
        //----------------------------------------------------------------
        // JDK 1.1 HACK (take out for 1.2-specific code)

        // if we make it to here, we have a successful match.  Now we
        // have to find out HOW MANY characters from the target string
        // matched the prefix (there isn't necessarily a one-to-one
        // mapping between collation elements and characters).
        // In JDK 1.2, there's a simple getOffset() call we can use.
        // In JDK 1.1, on the other hand, we have to go through some
        // ugly contortions.  First, use the collator to compare the
        // same number of characters from the prefix and target string.
        // If they're equal, we're done.
        collator->setStrength(Collator::PRIMARY);
        if (str.length() >= prefix.length()) {
            UnicodeString temp;
            temp.setTo(str, 0, prefix.length());
            if (collator->equals(temp, prefix)) {
#ifdef RBNF_DEBUG
                fprintf(stderr, "returning: %d\n", prefix.length());
#endif
                return prefix.length();
            }
        }

        // if they're not equal, then we have to compare successively
        // larger and larger substrings of the target string until we
        // get to one that matches the prefix.  At that point, we know
        // how many characters matched the prefix, and we can return.
        int32_t p = 1;
        while (p <= str.length()) {
            UnicodeString temp;
            temp.setTo(str, 0, p);
            if (collator->equals(temp, prefix)) {
                return p;
            } else {
                ++p;
            }
        }

        // SHOULD NEVER GET HERE!!!
        return 0;
        //----------------------------------------------------------------
#endif

        // If lenient parsing is turned off, forget all that crap above.
        // Just use String.startsWith() and be done with it.
  } else
#endif
  {
      if (str.startsWith(prefix)) {
          return prefix.length();
      } else {
          return 0;
      }
  }
}

/**
* Searches a string for another string.  If lenient parsing is off,
* this just calls indexOf().  If lenient parsing is on, this function
* uses CollationElementIterator to match characters, and only
* primary-order differences are significant in determining whether
* there's a match.
* @param str The string to search
* @param key The string to search "str" for
* @param startingAt The index into "str" where the search is to
* begin
* @return A two-element array of ints.  Element 0 is the position
* of the match, or -1 if there was no match.  Element 1 is the
* number of characters in "str" that matched (which isn't necessarily
* the same as the length of "key")
*/
int32_t
NFRule::findText(const UnicodeString& str,
                 const UnicodeString& key,
                 int32_t startingAt,
                 int32_t* length) const
{
    if (rulePatternFormat) {
        Formattable result;
        FieldPosition position(UNUM_INTEGER_FIELD);
        position.setBeginIndex(startingAt);
        rulePatternFormat->parseType(str, this, result, position);
        int start = position.getBeginIndex();
        if (start >= 0) {
            int32_t pluralRuleStart = fRuleText.indexOf(nfrule_gDollarOpenParenthesis, -1, 0);
            int32_t pluralRuleSuffix = fRuleText.indexOf(nfrule_gClosedParenthesisDollar, -1, pluralRuleStart) + 2;
            int32_t matchLen = position.getEndIndex() - start;
            UnicodeString prefix(fRuleText.tempSubString(0, pluralRuleStart));
            UnicodeString suffix(fRuleText.tempSubString(pluralRuleSuffix));
            if (str.compare(start - prefix.length(), prefix.length(), prefix, 0, prefix.length()) == 0
                    && str.compare(start + matchLen, suffix.length(), suffix, 0, suffix.length()) == 0)
            {
                *length = matchLen + prefix.length() + suffix.length();
                return start - prefix.length();
            }
        }
        *length = 0;
        return -1;
    }
    if (!formatter->isLenient()) {
        // if lenient parsing is turned off, this is easy: just call
        // String.indexOf() and we're done
        *length = key.length();
        return str.indexOf(key, startingAt);
    }
    else {
        // but if lenient parsing is turned ON, we've got some work
        // ahead of us
        return findTextLenient(str, key, startingAt, length);
    }
}

int32_t
NFRule::findTextLenient(const UnicodeString& str,
                 const UnicodeString& key,
                 int32_t startingAt,
                 int32_t* length) const
{
    //----------------------------------------------------------------
    // JDK 1.1 HACK (take out of 1.2-specific code)

    // in JDK 1.2, CollationElementIterator provides us with an
    // API to map between character offsets and collation elements
    // and we can do this by marching through the string comparing
    // collation elements.  We can't do that in JDK 1.1.  Insted,
    // we have to go through this horrible slow mess:
    int32_t p = startingAt;
    int32_t keyLen = 0;

    // basically just isolate smaller and smaller substrings of
    // the target string (each running to the end of the string,
    // and with the first one running from startingAt to the end)
    // and then use prefixLength() to see if the search key is at
    // the beginning of each substring.  This is excruciatingly
    // slow, but it will locate the key and tell use how long the
    // matching text was.
    UnicodeString temp;
    UErrorCode status = U_ZERO_ERROR;
    while (p < str.length() && keyLen == 0) {
        temp.setTo(str, p, str.length() - p);
        keyLen = prefixLength(temp, key, status);
        if (U_FAILURE(status)) {
            break;
        }
        if (keyLen != 0) {
            *length = keyLen;
            return p;
        }
        ++p;
    }
    // if we make it to here, we didn't find it.  Return -1 for the
    // location.  The length should be ignored, but set it to 0,
    // which should be "safe"
    *length = 0;
    return -1;
}

/**
* Checks to see whether a string consists entirely of ignorable
* characters.
* @param str The string to test.
* @return true if the string is empty of consists entirely of
* characters that the number formatter's collator says are
* ignorable at the primary-order level.  false otherwise.
*/
UBool
NFRule::allIgnorable(const UnicodeString& str, UErrorCode& status) const
{
    // if the string is empty, we can just return true
    if (str.length() == 0) {
        return TRUE;
    }

#if !UCONFIG_NO_COLLATION
    // if lenient parsing is turned on, walk through the string with
    // a collation element iterator and make sure each collation
    // element is 0 (ignorable) at the primary level
    if (formatter->isLenient()) {
        const RuleBasedCollator* collator = formatter->getCollator();
        if (collator == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return FALSE;
        }
        LocalPointer<CollationElementIterator> iter(collator->createCollationElementIterator(str));

        // Memory allocation error check.
        if (iter.isNull()) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return FALSE;
        }

        UErrorCode err = U_ZERO_ERROR;
        int32_t o = iter->next(err);
        while (o != CollationElementIterator::NULLORDER
            && CollationElementIterator::primaryOrder(o) == 0) {
            o = iter->next(err);
        }

        return o == CollationElementIterator::NULLORDER;
    }
#endif

    // if lenient parsing is turned off, there is no such thing as
    // an ignorable character: return true only if the string is empty
    return FALSE;
}

void
NFRule::setDecimalFormatSymbols(const DecimalFormatSymbols& newSymbols, UErrorCode& status) {
    if (sub1 != NULL) {
        sub1->setDecimalFormatSymbols(newSymbols, status);
    }
    if (sub2 != NULL) {
        sub2->setDecimalFormatSymbols(newSymbols, status);
    }
}

U_NAMESPACE_END

/* U_HAVE_RBNF */
#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 1997-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*   file name:  nfsubs.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
* Modification history
* Date        Name      Comments
* 10/11/2001  Doug      Ported from ICU4J
*/

#include <stdio.h>
  // for 'typeid' to work






#if U_HAVE_RBNF

static const UChar gLessThan = 0x003c;
static const UChar gEquals = 0x003d;
static const UChar gGreaterThan = 0x003e;
static const UChar gPercent = 0x0025;
static const UChar gPound = 0x0023;
static const UChar gZero = 0x0030;
static const UChar gSpace = 0x0020;

static const UChar gEqualsEquals[] =
{
    0x3D, 0x3D, 0
}; /* "==" */
static const UChar gGreaterGreaterGreaterThan[] =
{
    0x3E, 0x3E, 0x3E, 0
}; /* ">>>" */
static const UChar gGreaterGreaterThan[] =
{
    0x3E, 0x3E, 0
}; /* ">>" */

U_NAMESPACE_BEGIN

using number::impl::DecimalQuantity;

class SameValueSubstitution : public NFSubstitution {
public:
    SameValueSubstitution(int32_t pos,
        const NFRuleSet* ruleset,
        const UnicodeString& description,
        UErrorCode& status);
    virtual ~SameValueSubstitution();

    virtual int64_t transformNumber(int64_t number) const { return number; }
    virtual double transformNumber(double number) const { return number; }
    virtual double composeRuleValue(double newRuleValue, double /*oldRuleValue*/) const { return newRuleValue; }
    virtual double calcUpperBound(double oldUpperBound) const { return oldUpperBound; }
    virtual UChar tokenChar() const { return (UChar)0x003d; } // '='

public:
    static UClassID getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
};

SameValueSubstitution::~SameValueSubstitution() {}

class MultiplierSubstitution : public NFSubstitution {
    int64_t divisor;

public:
    MultiplierSubstitution(int32_t _pos,
        const NFRule *rule,
        const NFRuleSet* _ruleSet,
        const UnicodeString& description,
        UErrorCode& status)
        : NFSubstitution(_pos, _ruleSet, description, status), divisor(rule->getDivisor())
    {
        if (divisor == 0) {
            status = U_PARSE_ERROR;
        }
    }
    virtual ~MultiplierSubstitution();

    virtual void setDivisor(int32_t radix, int16_t exponent, UErrorCode& status) {
        divisor = util64_pow(radix, exponent);

        if(divisor == 0) {
            status = U_PARSE_ERROR;
        }
    }

    virtual UBool operator==(const NFSubstitution& rhs) const;

    virtual int64_t transformNumber(int64_t number) const {
        return number / divisor;
    }

    virtual double transformNumber(double number) const {
        if (getRuleSet()) {
            return uprv_floor(number / divisor);
        } else {
            return number / divisor;
        }
    }

    virtual double composeRuleValue(double newRuleValue, double /*oldRuleValue*/) const {
        return newRuleValue * divisor;
    }

    virtual double calcUpperBound(double /*oldUpperBound*/) const { return static_cast<double>(divisor); }

    virtual UChar tokenChar() const { return (UChar)0x003c; } // '<'

public:
    static UClassID getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
};

MultiplierSubstitution::~MultiplierSubstitution() {}

class ModulusSubstitution : public NFSubstitution {
    int64_t  divisor;
    const NFRule* ruleToUse;
public:
    ModulusSubstitution(int32_t pos,
        const NFRule* rule,
        const NFRule* rulePredecessor,
        const NFRuleSet* ruleSet,
        const UnicodeString& description,
        UErrorCode& status);
    virtual ~ModulusSubstitution();

    virtual void setDivisor(int32_t radix, int16_t exponent, UErrorCode& status) {
        divisor = util64_pow(radix, exponent);

        if (divisor == 0) {
            status = U_PARSE_ERROR;
        }
    }

    virtual UBool operator==(const NFSubstitution& rhs) const;

    virtual void doSubstitution(int64_t number, UnicodeString& toInsertInto, int32_t pos, int32_t recursionCount, UErrorCode& status) const;
    virtual void doSubstitution(double number, UnicodeString& toInsertInto, int32_t pos, int32_t recursionCount, UErrorCode& status) const;

    virtual int64_t transformNumber(int64_t number) const { return number % divisor; }
    virtual double transformNumber(double number) const { return uprv_fmod(number, static_cast<double>(divisor)); }

    virtual UBool doParse(const UnicodeString& text,
        ParsePosition& parsePosition,
        double baseValue,
        double upperBound,
        UBool lenientParse,
        uint32_t nonNumericalExecutedRuleMask,
        Formattable& result) const;

    virtual double composeRuleValue(double newRuleValue, double oldRuleValue) const {
        return oldRuleValue - uprv_fmod(oldRuleValue, static_cast<double>(divisor)) + newRuleValue;
    }

    virtual double calcUpperBound(double /*oldUpperBound*/) const { return static_cast<double>(divisor); }

    virtual UBool isModulusSubstitution() const { return TRUE; }

    virtual UChar tokenChar() const { return (UChar)0x003e; } // '>'

	virtual void toString(UnicodeString& result) const;

public:
    static UClassID getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
};

ModulusSubstitution::~ModulusSubstitution() {}

class IntegralPartSubstitution : public NFSubstitution {
public:
    IntegralPartSubstitution(int32_t _pos,
        const NFRuleSet* _ruleSet,
        const UnicodeString& description,
        UErrorCode& status)
        : NFSubstitution(_pos, _ruleSet, description, status) {}
    virtual ~IntegralPartSubstitution();

    virtual int64_t transformNumber(int64_t number) const { return number; }
    virtual double transformNumber(double number) const { return uprv_floor(number); }
    virtual double composeRuleValue(double newRuleValue, double oldRuleValue) const { return newRuleValue + oldRuleValue; }
    virtual double calcUpperBound(double /*oldUpperBound*/) const { return DBL_MAX; }
    virtual UChar tokenChar() const { return (UChar)0x003c; } // '<'

public:
    static UClassID getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
};

IntegralPartSubstitution::~IntegralPartSubstitution() {}

class FractionalPartSubstitution : public NFSubstitution {
    UBool byDigits;
    UBool useSpaces;
    enum { kMaxDecimalDigits = 8 };
public:
    FractionalPartSubstitution(int32_t pos,
        const NFRuleSet* ruleSet,
        const UnicodeString& description,
        UErrorCode& status);
    virtual ~FractionalPartSubstitution();

    virtual UBool operator==(const NFSubstitution& rhs) const;

    virtual void doSubstitution(double number, UnicodeString& toInsertInto, int32_t pos, int32_t recursionCount, UErrorCode& status) const;
    virtual void doSubstitution(int64_t /*number*/, UnicodeString& /*toInsertInto*/, int32_t /*_pos*/, int32_t /*recursionCount*/, UErrorCode& /*status*/) const {}
    virtual int64_t transformNumber(int64_t /*number*/) const { return 0; }
    virtual double transformNumber(double number) const { return number - uprv_floor(number); }

    virtual UBool doParse(const UnicodeString& text,
        ParsePosition& parsePosition,
        double baseValue,
        double upperBound,
        UBool lenientParse,
        uint32_t nonNumericalExecutedRuleMask,
        Formattable& result) const;

    virtual double composeRuleValue(double newRuleValue, double oldRuleValue) const { return newRuleValue + oldRuleValue; }
    virtual double calcUpperBound(double /*oldUpperBound*/) const { return 0.0; }
    virtual UChar tokenChar() const { return (UChar)0x003e; } // '>'

public:
    static UClassID getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
};

FractionalPartSubstitution::~FractionalPartSubstitution() {}

class AbsoluteValueSubstitution : public NFSubstitution {
public:
    AbsoluteValueSubstitution(int32_t _pos,
        const NFRuleSet* _ruleSet,
        const UnicodeString& description,
        UErrorCode& status)
        : NFSubstitution(_pos, _ruleSet, description, status) {}
    virtual ~AbsoluteValueSubstitution();

    virtual int64_t transformNumber(int64_t number) const { return number >= 0 ? number : -number; }
    virtual double transformNumber(double number) const { return uprv_fabs(number); }
    virtual double composeRuleValue(double newRuleValue, double /*oldRuleValue*/) const { return -newRuleValue; }
    virtual double calcUpperBound(double /*oldUpperBound*/) const { return DBL_MAX; }
    virtual UChar tokenChar() const { return (UChar)0x003e; } // '>'

public:
    static UClassID getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
};

AbsoluteValueSubstitution::~AbsoluteValueSubstitution() {}

class NumeratorSubstitution : public NFSubstitution {
    double denominator;
    int64_t ldenominator;
    UBool withZeros;
public:
    static inline UnicodeString fixdesc(const UnicodeString& desc) {
        if (desc.endsWith(LTLT, 2)) {
            UnicodeString result(desc, 0, desc.length()-1);
            return result;
        }
        return desc;
    }
    NumeratorSubstitution(int32_t _pos,
        double _denominator,
        NFRuleSet* _ruleSet,
        const UnicodeString& description,
        UErrorCode& status)
        : NFSubstitution(_pos, _ruleSet, fixdesc(description), status), denominator(_denominator)
    {
        ldenominator = util64_fromDouble(denominator);
        withZeros = description.endsWith(LTLT, 2);
    }
    virtual ~NumeratorSubstitution();

    virtual UBool operator==(const NFSubstitution& rhs) const;

    virtual int64_t transformNumber(int64_t number) const { return number * ldenominator; }
    virtual double transformNumber(double number) const { return uprv_round(number * denominator); }

    virtual void doSubstitution(int64_t /*number*/, UnicodeString& /*toInsertInto*/, int32_t /*_pos*/, int32_t /*recursionCount*/, UErrorCode& /*status*/) const {}
    virtual void doSubstitution(double number, UnicodeString& toInsertInto, int32_t pos, int32_t recursionCount, UErrorCode& status) const;
    virtual UBool doParse(const UnicodeString& text,
        ParsePosition& parsePosition,
        double baseValue,
        double upperBound,
        UBool /*lenientParse*/,
        uint32_t nonNumericalExecutedRuleMask,
        Formattable& result) const;

    virtual double composeRuleValue(double newRuleValue, double oldRuleValue) const { return newRuleValue / oldRuleValue; }
    virtual double calcUpperBound(double /*oldUpperBound*/) const { return denominator; }
    virtual UChar tokenChar() const { return (UChar)0x003c; } // '<'
private:
    static const UChar LTLT[2];

public:
    static UClassID getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
};

NumeratorSubstitution::~NumeratorSubstitution() {}

NFSubstitution*
NFSubstitution::makeSubstitution(int32_t pos,
                                 const NFRule* rule,
                                 const NFRule* predecessor,
                                 const NFRuleSet* ruleSet,
                                 const RuleBasedNumberFormat* formatter,
                                 const UnicodeString& description,
                                 UErrorCode& status)
{
    // if the description is empty, return a NullSubstitution
    if (description.length() == 0) {
        return NULL;
    }

    switch (description.charAt(0)) {
        // if the description begins with '<'...
    case gLessThan:
        // throw an exception if the rule is a negative number
        // rule
        if (rule->getBaseValue() == NFRule::kNegativeNumberRule) {
            // throw new IllegalArgumentException("<< not allowed in negative-number rule");
            status = U_PARSE_ERROR;
            return NULL;
        }

        // if the rule is a fraction rule, return an
        // IntegralPartSubstitution
        else if (rule->getBaseValue() == NFRule::kImproperFractionRule
            || rule->getBaseValue() == NFRule::kProperFractionRule
            || rule->getBaseValue() == NFRule::kMasterRule) {
            return new IntegralPartSubstitution(pos, ruleSet, description, status);
        }

        // if the rule set containing the rule is a fraction
        // rule set, return a NumeratorSubstitution
        else if (ruleSet->isFractionRuleSet()) {
            return new NumeratorSubstitution(pos, (double)rule->getBaseValue(),
                formatter->getDefaultRuleSet(), description, status);
        }

        // otherwise, return a MultiplierSubstitution
        else {
            return new MultiplierSubstitution(pos, rule, ruleSet,
                description, status);
        }

        // if the description begins with '>'...
    case gGreaterThan:
        // if the rule is a negative-number rule, return
        // an AbsoluteValueSubstitution
        if (rule->getBaseValue() == NFRule::kNegativeNumberRule) {
            return new AbsoluteValueSubstitution(pos, ruleSet, description, status);
        }

        // if the rule is a fraction rule, return a
        // FractionalPartSubstitution
        else if (rule->getBaseValue() == NFRule::kImproperFractionRule
            || rule->getBaseValue() == NFRule::kProperFractionRule
            || rule->getBaseValue() == NFRule::kMasterRule) {
            return new FractionalPartSubstitution(pos, ruleSet, description, status);
        }

        // if the rule set owning the rule is a fraction rule set,
        // throw an exception
        else if (ruleSet->isFractionRuleSet()) {
            // throw new IllegalArgumentException(">> not allowed in fraction rule set");
            status = U_PARSE_ERROR;
            return NULL;
        }

        // otherwise, return a ModulusSubstitution
        else {
            return new ModulusSubstitution(pos, rule, predecessor,
                ruleSet, description, status);
        }

        // if the description begins with '=', always return a
        // SameValueSubstitution
    case gEquals:
        return new SameValueSubstitution(pos, ruleSet, description, status);

        // and if it's anything else, throw an exception
    default:
        // throw new IllegalArgumentException("Illegal substitution character");
        status = U_PARSE_ERROR;
    }
    return NULL;
}

NFSubstitution::NFSubstitution(int32_t _pos,
                               const NFRuleSet* _ruleSet,
                               const UnicodeString& description,
                               UErrorCode& status)
                               : pos(_pos), ruleSet(NULL), numberFormat(NULL)
{
    // the description should begin and end with the same character.
    // If it doesn't that's a syntax error.  Otherwise,
    // makeSubstitution() was the only thing that needed to know
    // about these characters, so strip them off
    UnicodeString workingDescription(description);
    if (description.length() >= 2
        && description.charAt(0) == description.charAt(description.length() - 1))
    {
        workingDescription.remove(description.length() - 1, 1);
        workingDescription.remove(0, 1);
    }
    else if (description.length() != 0) {
        // throw new IllegalArgumentException("Illegal substitution syntax");
        status = U_PARSE_ERROR;
        return;
    }

    if (workingDescription.length() == 0) {
        // if the description was just two paired token characters
        // (i.e., "<<" or ">>"), it uses the rule set it belongs to to
        // format its result
        this->ruleSet = _ruleSet;
    }
    else if (workingDescription.charAt(0) == gPercent) {
        // if the description contains a rule set name, that's the rule
        // set we use to format the result: get a reference to the
        // names rule set
        this->ruleSet = _ruleSet->getOwner()->findRuleSet(workingDescription, status);
    }
    else if (workingDescription.charAt(0) == gPound || workingDescription.charAt(0) ==gZero) {
        // if the description begins with 0 or #, treat it as a
        // DecimalFormat pattern, and initialize a DecimalFormat with
        // that pattern (then set it to use the DecimalFormatSymbols
        // belonging to our formatter)
        const DecimalFormatSymbols* sym = _ruleSet->getOwner()->getDecimalFormatSymbols();
        if (!sym) {
            status = U_MISSING_RESOURCE_ERROR;
            return;
        }
        DecimalFormat *tempNumberFormat = new DecimalFormat(workingDescription, *sym, status);
        /* test for NULL */
        if (!tempNumberFormat) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        if (U_FAILURE(status)) {
            delete tempNumberFormat;
            return;
        }
        this->numberFormat = tempNumberFormat;
    }
    else if (workingDescription.charAt(0) == gGreaterThan) {
        // if the description is ">>>", this substitution bypasses the
        // usual rule-search process and always uses the rule that precedes
        // it in its own rule set's rule list (this is used for place-value
        // notations: formats where you want to see a particular part of
        // a number even when it's 0)

        // this causes problems when >>> is used in a frationalPartSubstitution
        // this->ruleSet = NULL;
        this->ruleSet = _ruleSet;
        this->numberFormat = NULL;
    }
    else {
        // and of the description is none of these things, it's a syntax error

        // throw new IllegalArgumentException("Illegal substitution syntax");
        status = U_PARSE_ERROR;
    }
}

NFSubstitution::~NFSubstitution()
{
    delete numberFormat;
    numberFormat = NULL;
}

/**
 * Set's the substitution's divisor.  Used by NFRule.setBaseValue().
 * A no-op for all substitutions except multiplier and modulus
 * substitutions.
 * @param radix The radix of the divisor
 * @param exponent The exponent of the divisor
 */
void
NFSubstitution::setDivisor(int32_t /*radix*/, int16_t /*exponent*/, UErrorCode& /*status*/) {
  // a no-op for all substitutions except multiplier and modulus substitutions
}

void
NFSubstitution::setDecimalFormatSymbols(const DecimalFormatSymbols &newSymbols, UErrorCode& /*status*/) {
    if (numberFormat != NULL) {
        numberFormat->setDecimalFormatSymbols(newSymbols);
    }
}

//-----------------------------------------------------------------------
// boilerplate
//-----------------------------------------------------------------------

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(NFSubstitution)

/**
 * Compares two substitutions for equality
 * @param The substitution to compare this one to
 * @return true if the two substitutions are functionally equivalent
 */
UBool
NFSubstitution::operator==(const NFSubstitution& rhs) const
{
  // compare class and all of the fields all substitutions have
  // in common
  // this should be called by subclasses before their own equality tests
  return typeid(*this) == typeid(rhs)
  && pos == rhs.pos
  && (ruleSet == NULL) == (rhs.ruleSet == NULL)
  // && ruleSet == rhs.ruleSet causes circularity, other checks to make instead?
  && (numberFormat == NULL
      ? (rhs.numberFormat == NULL)
      : (*numberFormat == *rhs.numberFormat));
}

/**
 * Returns a textual description of the substitution
 * @return A textual description of the substitution.  This might
 * not be identical to the description it was created from, but
 * it'll produce the same result.
 */
void
NFSubstitution::toString(UnicodeString& text) const
{
  // use tokenChar() to get the character at the beginning and
  // end of the substitutin token.  In between them will go
  // either the name of the rule set it uses, or the pattern of
  // the DecimalFormat it uses
  text.remove();
  text.append(tokenChar());

  UnicodeString temp;
  if (ruleSet != NULL) {
    ruleSet->getName(temp);
  } else if (numberFormat != NULL) {
    numberFormat->toPattern(temp);
  }
  text.append(temp);
  text.append(tokenChar());
}

//-----------------------------------------------------------------------
// formatting
//-----------------------------------------------------------------------

/**
 * Performs a mathematical operation on the number, formats it using
 * either ruleSet or decimalFormat, and inserts the result into
 * toInsertInto.
 * @param number The number being formatted.
 * @param toInsertInto The string we insert the result into
 * @param pos The position in toInsertInto where the owning rule's
 * rule text begins (this value is added to this substitution's
 * position to determine exactly where to insert the new text)
 */
void
NFSubstitution::doSubstitution(int64_t number, UnicodeString& toInsertInto, int32_t _pos, int32_t recursionCount, UErrorCode& status) const
{
    if (ruleSet != NULL) {
        // Perform a transformation on the number that is dependent
        // on the type of substitution this is, then just call its
        // rule set's format() method to format the result
        ruleSet->format(transformNumber(number), toInsertInto, _pos + this->pos, recursionCount, status);
    } else if (numberFormat != NULL) {
        if (number <= MAX_INT64_IN_DOUBLE) {
            // or perform the transformation on the number (preserving
            // the result's fractional part if the formatter it set
            // to show it), then use that formatter's format() method
            // to format the result
            double numberToFormat = transformNumber((double)number);
            if (numberFormat->getMaximumFractionDigits() == 0) {
                numberToFormat = uprv_floor(numberToFormat);
            }

            UnicodeString temp;
            numberFormat->format(numberToFormat, temp, status);
            toInsertInto.insert(_pos + this->pos, temp);
        }
        else {
            // We have gone beyond double precision. Something has to give.
            // We're favoring accuracy of the large number over potential rules
            // that round like a CompactDecimalFormat, which is not a common use case.
            //
            // Perform a transformation on the number that is dependent
            // on the type of substitution this is, then just call its
            // rule set's format() method to format the result
            int64_t numberToFormat = transformNumber(number);
            UnicodeString temp;
            numberFormat->format(numberToFormat, temp, status);
            toInsertInto.insert(_pos + this->pos, temp);
        }
    }
}

/**
 * Performs a mathematical operation on the number, formats it using
 * either ruleSet or decimalFormat, and inserts the result into
 * toInsertInto.
 * @param number The number being formatted.
 * @param toInsertInto The string we insert the result into
 * @param pos The position in toInsertInto where the owning rule's
 * rule text begins (this value is added to this substitution's
 * position to determine exactly where to insert the new text)
 */
void
NFSubstitution::doSubstitution(double number, UnicodeString& toInsertInto, int32_t _pos, int32_t recursionCount, UErrorCode& status) const {
    // perform a transformation on the number being formatted that
    // is dependent on the type of substitution this is
    double numberToFormat = transformNumber(number);

    if (uprv_isInfinite(numberToFormat)) {
        // This is probably a minus rule. Combine it with an infinite rule.
        const NFRule *infiniteRule = ruleSet->findDoubleRule(uprv_getInfinity());
        infiniteRule->doFormat(numberToFormat, toInsertInto, _pos + this->pos, recursionCount, status);
        return;
    }

    // if the result is an integer, from here on out we work in integer
    // space (saving time and memory and preserving accuracy)
    if (numberToFormat == uprv_floor(numberToFormat) && ruleSet != NULL) {
        ruleSet->format(util64_fromDouble(numberToFormat), toInsertInto, _pos + this->pos, recursionCount, status);

        // if the result isn't an integer, then call either our rule set's
        // format() method or our DecimalFormat's format() method to
        // format the result
    } else {
        if (ruleSet != NULL) {
            ruleSet->format(numberToFormat, toInsertInto, _pos + this->pos, recursionCount, status);
        } else if (numberFormat != NULL) {
            UnicodeString temp;
            numberFormat->format(numberToFormat, temp);
            toInsertInto.insert(_pos + this->pos, temp);
        }
    }
}


    //-----------------------------------------------------------------------
    // parsing
    //-----------------------------------------------------------------------

#ifdef RBNF_DEBUG
#include <stdio.h>
#endif

/**
 * Parses a string using the rule set or DecimalFormat belonging
 * to this substitution.  If there's a match, a mathematical
 * operation (the inverse of the one used in formatting) is
 * performed on the result of the parse and the value passed in
 * and returned as the result.  The parse position is updated to
 * point to the first unmatched character in the string.
 * @param text The string to parse
 * @param parsePosition On entry, ignored, but assumed to be 0.
 * On exit, this is updated to point to the first unmatched
 * character (or 0 if the substitution didn't match)
 * @param baseValue A partial parse result that should be
 * combined with the result of this parse
 * @param upperBound When searching the rule set for a rule
 * matching the string passed in, only rules with base values
 * lower than this are considered
 * @param lenientParse If true and matching against rules fails,
 * the substitution will also try matching the text against
 * numerals using a default-costructed NumberFormat.  If false,
 * no extra work is done.  (This value is false whenever the
 * formatter isn't in lenient-parse mode, but is also false
 * under some conditions even when the formatter _is_ in
 * lenient-parse mode.)
 * @return If there's a match, this is the result of composing
 * baseValue with whatever was returned from matching the
 * characters.  This will be either a Long or a Double.  If there's
 * no match this is new Long(0) (not null), and parsePosition
 * is left unchanged.
 */
UBool
NFSubstitution::doParse(const UnicodeString& text,
                        ParsePosition& parsePosition,
                        double baseValue,
                        double upperBound,
                        UBool lenientParse,
                        uint32_t nonNumericalExecutedRuleMask,
                        Formattable& result) const
{
#ifdef RBNF_DEBUG
    fprintf(stderr, "<nfsubs> %x bv: %g ub: %g\n", this, baseValue, upperBound);
#endif
    // figure out the highest base value a rule can have and match
    // the text being parsed (this varies according to the type of
    // substitutions: multiplier, modulus, and numerator substitutions
    // restrict the search to rules with base values lower than their
    // own; same-value substitutions leave the upper bound wherever
    // it was, and the others allow any rule to match
    upperBound = calcUpperBound(upperBound);

    // use our rule set to parse the text.  If that fails and
    // lenient parsing is enabled (this is always false if the
    // formatter's lenient-parsing mode is off, but it may also
    // be false even when the formatter's lenient-parse mode is
    // on), then also try parsing the text using a default-
    // constructed NumberFormat
    if (ruleSet != NULL) {
        ruleSet->parse(text, parsePosition, upperBound, nonNumericalExecutedRuleMask, result);
        if (lenientParse && !ruleSet->isFractionRuleSet() && parsePosition.getIndex() == 0) {
            UErrorCode status = U_ZERO_ERROR;
            NumberFormat* fmt = NumberFormat::createInstance(status);
            if (U_SUCCESS(status)) {
                fmt->parse(text, result, parsePosition);
            }
            delete fmt;
        }

        // ...or use our DecimalFormat to parse the text
    } else if (numberFormat != NULL) {
        numberFormat->parse(text, result, parsePosition);
    }

    // if the parse was successful, we've already advanced the caller's
    // parse position (this is the one function that doesn't have one
    // of its own).  Derive a parse result and return it as a Long,
    // if possible, or a Double
    if (parsePosition.getIndex() != 0) {
        UErrorCode status = U_ZERO_ERROR;
        double tempResult = result.getDouble(status);

        // composeRuleValue() produces a full parse result from
        // the partial parse result passed to this function from
        // the caller (this is either the owning rule's base value
        // or the partial result obtained from composing the
        // owning rule's base value with its other substitution's
        // parse result) and the partial parse result obtained by
        // matching the substitution (which will be the same value
        // the caller would get by parsing just this part of the
        // text with RuleBasedNumberFormat.parse() ).  How the two
        // values are used to derive the full parse result depends
        // on the types of substitutions: For a regular rule, the
        // ultimate result is its multiplier substitution's result
        // times the rule's divisor (or the rule's base value) plus
        // the modulus substitution's result (which will actually
        // supersede part of the rule's base value).  For a negative-
        // number rule, the result is the negative of its substitution's
        // result.  For a fraction rule, it's the sum of its two
        // substitution results.  For a rule in a fraction rule set,
        // it's the numerator substitution's result divided by
        // the rule's base value.  Results from same-value substitutions
        // propagate back upard, and null substitutions don't affect
        // the result.
        tempResult = composeRuleValue(tempResult, baseValue);
        result.setDouble(tempResult);
        return TRUE;
        // if the parse was UNsuccessful, return 0
    } else {
        result.setLong(0);
        return FALSE;
    }
}

    /**
     * Returns true if this is a modulus substitution.  (We didn't do this
     * with instanceof partially because it causes source files to
     * proliferate and partially because we have to port this to C++.)
     * @return true if this object is an instance of ModulusSubstitution
     */
UBool
NFSubstitution::isModulusSubstitution() const {
    return FALSE;
}

//===================================================================
// SameValueSubstitution
//===================================================================

/**
 * A substitution that passes the value passed to it through unchanged.
 * Represented by == in rule descriptions.
 */
SameValueSubstitution::SameValueSubstitution(int32_t _pos,
                        const NFRuleSet* _ruleSet,
                        const UnicodeString& description,
                        UErrorCode& status)
: NFSubstitution(_pos, _ruleSet, description, status)
{
    if (0 == description.compare(gEqualsEquals, 2)) {
        // throw new IllegalArgumentException("== is not a legal token");
        status = U_PARSE_ERROR;
    }
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(SameValueSubstitution)

//===================================================================
// MultiplierSubstitution
//===================================================================

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(MultiplierSubstitution)

UBool MultiplierSubstitution::operator==(const NFSubstitution& rhs) const
{
    return NFSubstitution::operator==(rhs) &&
        divisor == ((const MultiplierSubstitution*)&rhs)->divisor;
}


//===================================================================
// ModulusSubstitution
//===================================================================

/**
 * A substitution that divides the number being formatted by the its rule's
 * divisor and formats the remainder.  Represented by "&gt;&gt;" in a
 * regular rule.
 */
ModulusSubstitution::ModulusSubstitution(int32_t _pos,
                                         const NFRule* rule,
                                         const NFRule* predecessor,
                                         const NFRuleSet* _ruleSet,
                                         const UnicodeString& description,
                                         UErrorCode& status)
 : NFSubstitution(_pos, _ruleSet, description, status)
 , divisor(rule->getDivisor())
 , ruleToUse(NULL)
{
  // the owning rule's divisor controls the behavior of this
  // substitution: rather than keeping a backpointer to the rule,
  // we keep a copy of the divisor

  if (divisor == 0) {
      status = U_PARSE_ERROR;
  }

  if (0 == description.compare(gGreaterGreaterGreaterThan, 3)) {
    // the >>> token doesn't alter how this substituion calculates the
    // values it uses for formatting and parsing, but it changes
    // what's done with that value after it's obtained: >>> short-
    // circuits the rule-search process and goes straight to the
    // specified rule to format the substitution value
    ruleToUse = predecessor;
  }
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(ModulusSubstitution)

UBool ModulusSubstitution::operator==(const NFSubstitution& rhs) const
{
  return NFSubstitution::operator==(rhs) &&
  divisor == ((const ModulusSubstitution*)&rhs)->divisor &&
  ruleToUse == ((const ModulusSubstitution*)&rhs)->ruleToUse;
}

//-----------------------------------------------------------------------
// formatting
//-----------------------------------------------------------------------


/**
 * If this is a &gt;&gt;&gt; substitution, use ruleToUse to fill in
 * the substitution.  Otherwise, just use the superclass function.
 * @param number The number being formatted
 * @toInsertInto The string to insert the result of this substitution
 * into
 * @param pos The position of the rule text in toInsertInto
 */
void
ModulusSubstitution::doSubstitution(int64_t number, UnicodeString& toInsertInto, int32_t _pos, int32_t recursionCount, UErrorCode& status) const
{
    // if this isn't a >>> substitution, just use the inherited version
    // of this function (which uses either a rule set or a DecimalFormat
    // to format its substitution value)
    if (ruleToUse == NULL) {
        NFSubstitution::doSubstitution(number, toInsertInto, _pos, recursionCount, status);

        // a >>> substitution goes straight to a particular rule to
        // format the substitution value
    } else {
        int64_t numberToFormat = transformNumber(number);
        ruleToUse->doFormat(numberToFormat, toInsertInto, _pos + getPos(), recursionCount, status);
    }
}

/**
* If this is a &gt;&gt;&gt; substitution, use ruleToUse to fill in
* the substitution.  Otherwise, just use the superclass function.
* @param number The number being formatted
* @toInsertInto The string to insert the result of this substitution
* into
* @param pos The position of the rule text in toInsertInto
*/
void
ModulusSubstitution::doSubstitution(double number, UnicodeString& toInsertInto, int32_t _pos, int32_t recursionCount, UErrorCode& status) const
{
    // if this isn't a >>> substitution, just use the inherited version
    // of this function (which uses either a rule set or a DecimalFormat
    // to format its substitution value)
    if (ruleToUse == NULL) {
        NFSubstitution::doSubstitution(number, toInsertInto, _pos, recursionCount, status);

        // a >>> substitution goes straight to a particular rule to
        // format the substitution value
    } else {
        double numberToFormat = transformNumber(number);

        ruleToUse->doFormat(numberToFormat, toInsertInto, _pos + getPos(), recursionCount, status);
    }
}

//-----------------------------------------------------------------------
// parsing
//-----------------------------------------------------------------------

/**
 * If this is a &gt;&gt;&gt; substitution, match only against ruleToUse.
 * Otherwise, use the superclass function.
 * @param text The string to parse
 * @param parsePosition Ignored on entry, updated on exit to point to
 * the first unmatched character.
 * @param baseValue The partial parse result prior to calling this
 * routine.
 */
UBool
ModulusSubstitution::doParse(const UnicodeString& text,
                             ParsePosition& parsePosition,
                             double baseValue,
                             double upperBound,
                             UBool lenientParse,
                             uint32_t nonNumericalExecutedRuleMask,
                             Formattable& result) const
{
    // if this isn't a >>> substitution, we can just use the
    // inherited parse() routine to do the parsing
    if (ruleToUse == NULL) {
        return NFSubstitution::doParse(text, parsePosition, baseValue, upperBound, lenientParse, nonNumericalExecutedRuleMask, result);

        // but if it IS a >>> substitution, we have to do it here: we
        // use the specific rule's doParse() method, and then we have to
        // do some of the other work of NFRuleSet.parse()
    } else {
        ruleToUse->doParse(text, parsePosition, FALSE, upperBound, nonNumericalExecutedRuleMask, result);

        if (parsePosition.getIndex() != 0) {
            UErrorCode status = U_ZERO_ERROR;
            double tempResult = result.getDouble(status);
            tempResult = composeRuleValue(tempResult, baseValue);
            result.setDouble(tempResult);
        }

        return TRUE;
    }
}
/**
 * Returns a textual description of the substitution
 * @return A textual description of the substitution.  This might
 * not be identical to the description it was created from, but
 * it'll produce the same result.
 */
void
ModulusSubstitution::toString(UnicodeString& text) const
{
  // use tokenChar() to get the character at the beginning and
  // end of the substitutin token.  In between them will go
  // either the name of the rule set it uses, or the pattern of
  // the DecimalFormat it uses

  if ( ruleToUse != NULL ) { // Must have been a >>> substitution.
      text.remove();
      text.append(tokenChar());
      text.append(tokenChar());
      text.append(tokenChar());
  } else { // Otherwise just use the super-class function.
	  NFSubstitution::toString(text);
  }
}
//===================================================================
// IntegralPartSubstitution
//===================================================================

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(IntegralPartSubstitution)


//===================================================================
// FractionalPartSubstitution
//===================================================================


    /**
     * Constructs a FractionalPartSubstitution.  This object keeps a flag
     * telling whether it should format by digits or not.  In addition,
     * it marks the rule set it calls (if any) as a fraction rule set.
     */
FractionalPartSubstitution::FractionalPartSubstitution(int32_t _pos,
                             const NFRuleSet* _ruleSet,
                             const UnicodeString& description,
                             UErrorCode& status)
 : NFSubstitution(_pos, _ruleSet, description, status)
 , byDigits(FALSE)
 , useSpaces(TRUE)

{
    // akk, ruleSet can change in superclass constructor
    if (0 == description.compare(gGreaterGreaterThan, 2) ||
        0 == description.compare(gGreaterGreaterGreaterThan, 3) ||
        _ruleSet == getRuleSet()) {
        byDigits = TRUE;
        if (0 == description.compare(gGreaterGreaterGreaterThan, 3)) {
            useSpaces = FALSE;
        }
    } else {
        // cast away const
        ((NFRuleSet*)getRuleSet())->makeIntoFractionRuleSet();
    }
}

//-----------------------------------------------------------------------
// formatting
//-----------------------------------------------------------------------

/**
 * If in "by digits" mode, fills in the substitution one decimal digit
 * at a time using the rule set containing this substitution.
 * Otherwise, uses the superclass function.
 * @param number The number being formatted
 * @param toInsertInto The string to insert the result of formatting
 * the substitution into
 * @param pos The position of the owning rule's rule text in
 * toInsertInto
 */
void
FractionalPartSubstitution::doSubstitution(double number, UnicodeString& toInsertInto,
                                           int32_t _pos, int32_t recursionCount, UErrorCode& status) const
{
  // if we're not in "byDigits" mode, just use the inherited
  // doSubstitution() routine
  if (!byDigits) {
    NFSubstitution::doSubstitution(number, toInsertInto, _pos, recursionCount, status);

    // if we're in "byDigits" mode, transform the value into an integer
    // by moving the decimal point eight places to the right and
    // pulling digits off the right one at a time, formatting each digit
    // as an integer using this substitution's owning rule set
    // (this is slower, but more accurate, than doing it from the
    // other end)
  } else {
    //          int32_t numberToFormat = (int32_t)uprv_round(transformNumber(number) * uprv_pow(10, kMaxDecimalDigits));
    //          // this flag keeps us from formatting trailing zeros.  It starts
    //          // out false because we're pulling from the right, and switches
    //          // to true the first time we encounter a non-zero digit
    //          UBool doZeros = FALSE;
    //          for (int32_t i = 0; i < kMaxDecimalDigits; i++) {
    //              int64_t digit = numberToFormat % 10;
    //              if (digit != 0 || doZeros) {
    //                  if (doZeros && useSpaces) {
    //                      toInsertInto.insert(_pos + getPos(), gSpace);
    //                  }
    //                  doZeros = TRUE;
    //                  getRuleSet()->format(digit, toInsertInto, _pos + getPos());
    //              }
    //              numberToFormat /= 10;
    //          }

    DecimalQuantity dl;
    dl.setToDouble(number);
    dl.roundToMagnitude(-20, UNUM_ROUND_HALFEVEN, status);     // round to 20 fraction digits.

    UBool pad = FALSE;
    for (int32_t didx = dl.getLowerDisplayMagnitude(); didx<0; didx++) {
      // Loop iterates over fraction digits, starting with the LSD.
      //   include both real digits from the number, and zeros
      //   to the left of the MSD but to the right of the decimal point.
      if (pad && useSpaces) {
        toInsertInto.insert(_pos + getPos(), gSpace);
      } else {
        pad = TRUE;
      }
      int64_t digit = dl.getDigit(didx);
      getRuleSet()->format(digit, toInsertInto, _pos + getPos(), recursionCount, status);
    }

    if (!pad) {
      // hack around lack of precision in digitlist. if we would end up with
      // "foo point" make sure we add a " zero" to the end.
      getRuleSet()->format((int64_t)0, toInsertInto, _pos + getPos(), recursionCount, status);
    }
  }
}

//-----------------------------------------------------------------------
// parsing
//-----------------------------------------------------------------------

/**
 * If in "by digits" mode, parses the string as if it were a string
 * of individual digits; otherwise, uses the superclass function.
 * @param text The string to parse
 * @param parsePosition Ignored on entry, but updated on exit to point
 * to the first unmatched character
 * @param baseValue The partial parse result prior to entering this
 * function
 * @param upperBound Only consider rules with base values lower than
 * this when filling in the substitution
 * @param lenientParse If true, try matching the text as numerals if
 * matching as words doesn't work
 * @return If the match was successful, the current partial parse
 * result; otherwise new Long(0).  The result is either a Long or
 * a Double.
 */

UBool
FractionalPartSubstitution::doParse(const UnicodeString& text,
                ParsePosition& parsePosition,
                double baseValue,
                double /*upperBound*/,
                UBool lenientParse,
                uint32_t nonNumericalExecutedRuleMask,
                Formattable& resVal) const
{
    // if we're not in byDigits mode, we can just use the inherited
    // doParse()
    if (!byDigits) {
        return NFSubstitution::doParse(text, parsePosition, baseValue, 0, lenientParse, nonNumericalExecutedRuleMask, resVal);

        // if we ARE in byDigits mode, parse the text one digit at a time
        // using this substitution's owning rule set (we do this by setting
        // upperBound to 10 when calling doParse() ) until we reach
        // nonmatching text
    } else {
        UnicodeString workText(text);
        ParsePosition workPos(1);
        double result = 0;
        int32_t digit;
//          double p10 = 0.1;

        DecimalQuantity dl;
        int32_t totalDigits = 0;
        NumberFormat* fmt = NULL;
        while (workText.length() > 0 && workPos.getIndex() != 0) {
            workPos.setIndex(0);
            Formattable temp;
            getRuleSet()->parse(workText, workPos, 10, nonNumericalExecutedRuleMask, temp);
            UErrorCode status = U_ZERO_ERROR;
            digit = temp.getLong(status);
//            digit = temp.getType() == Formattable::kLong ?
//               temp.getLong() :
//            (int32_t)temp.getDouble();

            if (lenientParse && workPos.getIndex() == 0) {
                if (!fmt) {
                    status = U_ZERO_ERROR;
                    fmt = NumberFormat::createInstance(status);
                    if (U_FAILURE(status)) {
                        delete fmt;
                        fmt = NULL;
                    }
                }
                if (fmt) {
                    fmt->parse(workText, temp, workPos);
                    digit = temp.getLong(status);
                }
            }

            if (workPos.getIndex() != 0) {
                dl.appendDigit(static_cast<int8_t>(digit), 0, true);
                totalDigits++;
//                  result += digit * p10;
//                  p10 /= 10;
                parsePosition.setIndex(parsePosition.getIndex() + workPos.getIndex());
                workText.removeBetween(0, workPos.getIndex());
                while (workText.length() > 0 && workText.charAt(0) == gSpace) {
                    workText.removeBetween(0, 1);
                    parsePosition.setIndex(parsePosition.getIndex() + 1);
                }
            }
        }
        delete fmt;

        dl.adjustMagnitude(-totalDigits);
        result = dl.toDouble();
        result = composeRuleValue(result, baseValue);
        resVal.setDouble(result);
        return TRUE;
    }
}

UBool
FractionalPartSubstitution::operator==(const NFSubstitution& rhs) const
{
  return NFSubstitution::operator==(rhs) &&
  ((const FractionalPartSubstitution*)&rhs)->byDigits == byDigits;
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(FractionalPartSubstitution)


//===================================================================
// AbsoluteValueSubstitution
//===================================================================

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(AbsoluteValueSubstitution)

//===================================================================
// NumeratorSubstitution
//===================================================================

void
NumeratorSubstitution::doSubstitution(double number, UnicodeString& toInsertInto, int32_t apos, int32_t recursionCount, UErrorCode& status) const {
    // perform a transformation on the number being formatted that
    // is dependent on the type of substitution this is

    double numberToFormat = transformNumber(number);
    int64_t longNF = util64_fromDouble(numberToFormat);

    const NFRuleSet* aruleSet = getRuleSet();
    if (withZeros && aruleSet != NULL) {
        // if there are leading zeros in the decimal expansion then emit them
        int64_t nf =longNF;
        int32_t len = toInsertInto.length();
        while ((nf *= 10) < denominator) {
            toInsertInto.insert(apos + getPos(), gSpace);
            aruleSet->format((int64_t)0, toInsertInto, apos + getPos(), recursionCount, status);
        }
        apos += toInsertInto.length() - len;
    }

    // if the result is an integer, from here on out we work in integer
    // space (saving time and memory and preserving accuracy)
    if (numberToFormat == longNF && aruleSet != NULL) {
        aruleSet->format(longNF, toInsertInto, apos + getPos(), recursionCount, status);

        // if the result isn't an integer, then call either our rule set's
        // format() method or our DecimalFormat's format() method to
        // format the result
    } else {
        if (aruleSet != NULL) {
            aruleSet->format(numberToFormat, toInsertInto, apos + getPos(), recursionCount, status);
        } else {
            UnicodeString temp;
            getNumberFormat()->format(numberToFormat, temp, status);
            toInsertInto.insert(apos + getPos(), temp);
        }
    }
}

UBool
NumeratorSubstitution::doParse(const UnicodeString& text,
                               ParsePosition& parsePosition,
                               double baseValue,
                               double upperBound,
                               UBool /*lenientParse*/,
                               uint32_t nonNumericalExecutedRuleMask,
                               Formattable& result) const
{
    // we don't have to do anything special to do the parsing here,
    // but we have to turn lenient parsing off-- if we leave it on,
    // it SERIOUSLY messes up the algorithm

    // if withZeros is true, we need to count the zeros
    // and use that to adjust the parse result
    UErrorCode status = U_ZERO_ERROR;
    int32_t zeroCount = 0;
    UnicodeString workText(text);

    if (withZeros) {
        ParsePosition workPos(1);
        Formattable temp;

        while (workText.length() > 0 && workPos.getIndex() != 0) {
            workPos.setIndex(0);
            getRuleSet()->parse(workText, workPos, 1, nonNumericalExecutedRuleMask, temp); // parse zero or nothing at all
            if (workPos.getIndex() == 0) {
                // we failed, either there were no more zeros, or the number was formatted with digits
                // either way, we're done
                break;
            }

            ++zeroCount;
            parsePosition.setIndex(parsePosition.getIndex() + workPos.getIndex());
            workText.remove(0, workPos.getIndex());
            while (workText.length() > 0 && workText.charAt(0) == gSpace) {
                workText.remove(0, 1);
                parsePosition.setIndex(parsePosition.getIndex() + 1);
            }
        }

        workText = text;
        workText.remove(0, (int32_t)parsePosition.getIndex());
        parsePosition.setIndex(0);
    }

    // we've parsed off the zeros, now let's parse the rest from our current position
    NFSubstitution::doParse(workText, parsePosition, withZeros ? 1 : baseValue, upperBound, FALSE, nonNumericalExecutedRuleMask, result);

    if (withZeros) {
        // any base value will do in this case.  is there a way to
        // force this to not bother trying all the base values?

        // compute the 'effective' base and prescale the value down
        int64_t n = result.getLong(status); // force conversion!
        int64_t d = 1;
        int32_t pow = 0;
        while (d <= n) {
            d *= 10;
            ++pow;
        }
        // now add the zeros
        while (zeroCount > 0) {
            d *= 10;
            --zeroCount;
        }
        // d is now our true denominator
        result.setDouble((double)n/(double)d);
    }

    return TRUE;
}

UBool
NumeratorSubstitution::operator==(const NFSubstitution& rhs) const
{
    return NFSubstitution::operator==(rhs) &&
        denominator == ((const NumeratorSubstitution*)&rhs)->denominator;
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(NumeratorSubstitution)

const UChar NumeratorSubstitution::LTLT[] = { 0x003c, 0x003c };

U_NAMESPACE_END

/* U_HAVE_RBNF */
#endif

// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html




#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(NoUnit)

NoUnit U_EXPORT2 NoUnit::base() {
    return NoUnit("base");
}

NoUnit U_EXPORT2 NoUnit::percent() {
    return NoUnit("percent");
}

NoUnit U_EXPORT2 NoUnit::permille() {
    return NoUnit("permille");
}

NoUnit::NoUnit(const char* subtype) {
    initNoUnit(subtype);
}

NoUnit::NoUnit(const NoUnit& other) : MeasureUnit(other) {
}

NoUnit* NoUnit::clone() const {
    return new NoUnit(*this);
}

NoUnit::~NoUnit() {
}


U_NAMESPACE_END

#endif
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING





using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

TokenConsumer::~TokenConsumer() = default;
SymbolProvider::~SymbolProvider() = default;

int32_t AffixUtils::estimateLength(const UnicodeString &patternString, UErrorCode &status) {
    AffixPatternState state = STATE_BASE;
    int32_t offset = 0;
    int32_t length = 0;
    for (; offset < patternString.length();) {
        UChar32 cp = patternString.char32At(offset);

        switch (state) {
            case STATE_BASE:
                if (cp == u'\'') {
                    // First quote
                    state = STATE_FIRST_QUOTE;
                } else {
                    // Unquoted symbol
                    length++;
                }
                break;
            case STATE_FIRST_QUOTE:
                if (cp == u'\'') {
                    // Repeated quote
                    length++;
                    state = STATE_BASE;
                } else {
                    // Quoted code point
                    length++;
                    state = STATE_INSIDE_QUOTE;
                }
                break;
            case STATE_INSIDE_QUOTE:
                if (cp == u'\'') {
                    // End of quoted sequence
                    state = STATE_AFTER_QUOTE;
                } else {
                    // Quoted code point
                    length++;
                }
                break;
            case STATE_AFTER_QUOTE:
                if (cp == u'\'') {
                    // Double quote inside of quoted sequence
                    length++;
                    state = STATE_INSIDE_QUOTE;
                } else {
                    // Unquoted symbol
                    length++;
                }
                break;
            default:
                UPRV_UNREACHABLE;
        }

        offset += U16_LENGTH(cp);
    }

    switch (state) {
        case STATE_FIRST_QUOTE:
        case STATE_INSIDE_QUOTE:
            status = U_ILLEGAL_ARGUMENT_ERROR;
            break;
        default:
            break;
    }

    return length;
}

UnicodeString AffixUtils::escape(const UnicodeString &input) {
    AffixPatternState state = STATE_BASE;
    int32_t offset = 0;
    UnicodeString output;
    for (; offset < input.length();) {
        UChar32 cp = input.char32At(offset);

        switch (cp) {
            case u'\'':
                output.append(u"''", -1);
                break;

            case u'-':
            case u'+':
            case u'%':
            case u'\x2030':
            case u'\xa4':
                if (state == STATE_BASE) {
                    output.append(u'\'');
                    output.append(cp);
                    state = STATE_INSIDE_QUOTE;
                } else {
                    output.append(cp);
                }
                break;

            default:
                if (state == STATE_INSIDE_QUOTE) {
                    output.append(u'\'');
                    output.append(cp);
                    state = STATE_BASE;
                } else {
                    output.append(cp);
                }
                break;
        }
        offset += U16_LENGTH(cp);
    }

    if (state == STATE_INSIDE_QUOTE) {
        output.append(u'\'');
    }

    return output;
}

Field AffixUtils::getFieldForType(AffixPatternType type) {
    switch (type) {
        case TYPE_MINUS_SIGN:
            return UNUM_SIGN_FIELD;
        case TYPE_PLUS_SIGN:
            return UNUM_SIGN_FIELD;
        case TYPE_PERCENT:
            return UNUM_PERCENT_FIELD;
        case TYPE_PERMILLE:
            return UNUM_PERMILL_FIELD;
        case TYPE_CURRENCY_SINGLE:
            return UNUM_CURRENCY_FIELD;
        case TYPE_CURRENCY_DOUBLE:
            return UNUM_CURRENCY_FIELD;
        case TYPE_CURRENCY_TRIPLE:
            return UNUM_CURRENCY_FIELD;
        case TYPE_CURRENCY_QUAD:
            return UNUM_CURRENCY_FIELD;
        case TYPE_CURRENCY_QUINT:
            return UNUM_CURRENCY_FIELD;
        case TYPE_CURRENCY_OVERFLOW:
            return UNUM_CURRENCY_FIELD;
        default:
            UPRV_UNREACHABLE;
    }
}

int32_t
AffixUtils::unescape(const UnicodeString &affixPattern, FormattedStringBuilder &output, int32_t position,
                     const SymbolProvider &provider, Field field, UErrorCode &status) {
    int32_t length = 0;
    AffixTag tag;
    while (hasNext(tag, affixPattern)) {
        tag = nextToken(tag, affixPattern, status);
        if (U_FAILURE(status)) { return length; }
        if (tag.type == TYPE_CURRENCY_OVERFLOW) {
            // Don't go to the provider for this special case
            length += output.insertCodePoint(position + length, 0xFFFD, UNUM_CURRENCY_FIELD, status);
        } else if (tag.type < 0) {
            length += output.insert(
                    position + length, provider.getSymbol(tag.type), getFieldForType(tag.type), status);
        } else {
            length += output.insertCodePoint(position + length, tag.codePoint, field, status);
        }
    }
    return length;
}

int32_t AffixUtils::unescapedCodePointCount(const UnicodeString &affixPattern,
                                            const SymbolProvider &provider, UErrorCode &status) {
    int32_t length = 0;
    AffixTag tag;
    while (hasNext(tag, affixPattern)) {
        tag = nextToken(tag, affixPattern, status);
        if (U_FAILURE(status)) { return length; }
        if (tag.type == TYPE_CURRENCY_OVERFLOW) {
            length += 1;
        } else if (tag.type < 0) {
            length += provider.getSymbol(tag.type).length();
        } else {
            length += U16_LENGTH(tag.codePoint);
        }
    }
    return length;
}

bool
AffixUtils::containsType(const UnicodeString &affixPattern, AffixPatternType type, UErrorCode &status) {
    if (affixPattern.length() == 0) {
        return false;
    }
    AffixTag tag;
    while (hasNext(tag, affixPattern)) {
        tag = nextToken(tag, affixPattern, status);
        if (U_FAILURE(status)) { return false; }
        if (tag.type == type) {
            return true;
        }
    }
    return false;
}

bool AffixUtils::hasCurrencySymbols(const UnicodeString &affixPattern, UErrorCode &status) {
    if (affixPattern.length() == 0) {
        return false;
    }
    AffixTag tag;
    while (hasNext(tag, affixPattern)) {
        tag = nextToken(tag, affixPattern, status);
        if (U_FAILURE(status)) { return false; }
        if (tag.type < 0 && getFieldForType(tag.type) == UNUM_CURRENCY_FIELD) {
            return true;
        }
    }
    return false;
}

UnicodeString AffixUtils::replaceType(const UnicodeString &affixPattern, AffixPatternType type,
                                      char16_t replacementChar, UErrorCode &status) {
    UnicodeString output(affixPattern); // copy
    if (affixPattern.length() == 0) {
        return output;
    }
    AffixTag tag;
    while (hasNext(tag, affixPattern)) {
        tag = nextToken(tag, affixPattern, status);
        if (U_FAILURE(status)) { return output; }
        if (tag.type == type) {
            output.replace(tag.offset - 1, 1, replacementChar);
        }
    }
    return output;
}

bool AffixUtils::containsOnlySymbolsAndIgnorables(const UnicodeString& affixPattern,
                                                  const UnicodeSet& ignorables, UErrorCode& status) {
    if (affixPattern.length() == 0) {
        return true;
    }
    AffixTag tag;
    while (hasNext(tag, affixPattern)) {
        tag = nextToken(tag, affixPattern, status);
        if (U_FAILURE(status)) { return false; }
        if (tag.type == TYPE_CODEPOINT && !ignorables.contains(tag.codePoint)) {
            return false;
        }
    }
    return true;
}

void AffixUtils::iterateWithConsumer(const UnicodeString& affixPattern, TokenConsumer& consumer,
                                     UErrorCode& status) {
    if (affixPattern.length() == 0) {
        return;
    }
    AffixTag tag;
    while (hasNext(tag, affixPattern)) {
        tag = nextToken(tag, affixPattern, status);
        if (U_FAILURE(status)) { return; }
        consumer.consumeToken(tag.type, tag.codePoint, status);
        if (U_FAILURE(status)) { return; }
    }
}

AffixTag AffixUtils::nextToken(AffixTag tag, const UnicodeString &patternString, UErrorCode &status) {
    int32_t offset = tag.offset;
    int32_t state = tag.state;
    for (; offset < patternString.length();) {
        UChar32 cp = patternString.char32At(offset);
        int32_t count = U16_LENGTH(cp);

        switch (state) {
            case STATE_BASE:
                switch (cp) {
                    case u'\'':
                        state = STATE_FIRST_QUOTE;
                        offset += count;
                        // continue to the next code point
                        break;
                    case u'-':
                        return makeTag(offset + count, TYPE_MINUS_SIGN, STATE_BASE, 0);
                    case u'+':
                        return makeTag(offset + count, TYPE_PLUS_SIGN, STATE_BASE, 0);
                    case u'%':
                        return makeTag(offset + count, TYPE_PERCENT, STATE_BASE, 0);
                    case u'\x2030':
                        return makeTag(offset + count, TYPE_PERMILLE, STATE_BASE, 0);
                    case u'\xa4':
                        state = STATE_FIRST_CURR;
                        offset += count;
                        // continue to the next code point
                        break;
                    default:
                        return makeTag(offset + count, TYPE_CODEPOINT, STATE_BASE, cp);
                }
                break;
            case STATE_FIRST_QUOTE:
                if (cp == u'\'') {
                    return makeTag(offset + count, TYPE_CODEPOINT, STATE_BASE, cp);
                } else {
                    return makeTag(offset + count, TYPE_CODEPOINT, STATE_INSIDE_QUOTE, cp);
                }
            case STATE_INSIDE_QUOTE:
                if (cp == u'\'') {
                    state = STATE_AFTER_QUOTE;
                    offset += count;
                    // continue to the next code point
                    break;
                } else {
                    return makeTag(offset + count, TYPE_CODEPOINT, STATE_INSIDE_QUOTE, cp);
                }
            case STATE_AFTER_QUOTE:
                if (cp == u'\'') {
                    return makeTag(offset + count, TYPE_CODEPOINT, STATE_INSIDE_QUOTE, cp);
                } else {
                    state = STATE_BASE;
                    // re-evaluate this code point
                    break;
                }
            case STATE_FIRST_CURR:
                if (cp == u'\xa4') {
                    state = STATE_SECOND_CURR;
                    offset += count;
                    // continue to the next code point
                    break;
                } else {
                    return makeTag(offset, TYPE_CURRENCY_SINGLE, STATE_BASE, 0);
                }
            case STATE_SECOND_CURR:
                if (cp == u'\xa4') {
                    state = STATE_THIRD_CURR;
                    offset += count;
                    // continue to the next code point
                    break;
                } else {
                    return makeTag(offset, TYPE_CURRENCY_DOUBLE, STATE_BASE, 0);
                }
            case STATE_THIRD_CURR:
                if (cp == u'\xa4') {
                    state = STATE_FOURTH_CURR;
                    offset += count;
                    // continue to the next code point
                    break;
                } else {
                    return makeTag(offset, TYPE_CURRENCY_TRIPLE, STATE_BASE, 0);
                }
            case STATE_FOURTH_CURR:
                if (cp == u'\xa4') {
                    state = STATE_FIFTH_CURR;
                    offset += count;
                    // continue to the next code point
                    break;
                } else {
                    return makeTag(offset, TYPE_CURRENCY_QUAD, STATE_BASE, 0);
                }
            case STATE_FIFTH_CURR:
                if (cp == u'\xa4') {
                    state = STATE_OVERFLOW_CURR;
                    offset += count;
                    // continue to the next code point
                    break;
                } else {
                    return makeTag(offset, TYPE_CURRENCY_QUINT, STATE_BASE, 0);
                }
            case STATE_OVERFLOW_CURR:
                if (cp == u'\xa4') {
                    offset += count;
                    // continue to the next code point and loop back to this state
                    break;
                } else {
                    return makeTag(offset, TYPE_CURRENCY_OVERFLOW, STATE_BASE, 0);
                }
            default:
                UPRV_UNREACHABLE;
        }
    }
    // End of string
    switch (state) {
        case STATE_BASE:
            // No more tokens in string.
            return {-1};
        case STATE_FIRST_QUOTE:
        case STATE_INSIDE_QUOTE:
            // For consistent behavior with the JDK and ICU 58, set an error here.
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return {-1};
        case STATE_AFTER_QUOTE:
            // No more tokens in string.
            return {-1};
        case STATE_FIRST_CURR:
            return makeTag(offset, TYPE_CURRENCY_SINGLE, STATE_BASE, 0);
        case STATE_SECOND_CURR:
            return makeTag(offset, TYPE_CURRENCY_DOUBLE, STATE_BASE, 0);
        case STATE_THIRD_CURR:
            return makeTag(offset, TYPE_CURRENCY_TRIPLE, STATE_BASE, 0);
        case STATE_FOURTH_CURR:
            return makeTag(offset, TYPE_CURRENCY_QUAD, STATE_BASE, 0);
        case STATE_FIFTH_CURR:
            return makeTag(offset, TYPE_CURRENCY_QUINT, STATE_BASE, 0);
        case STATE_OVERFLOW_CURR:
            return makeTag(offset, TYPE_CURRENCY_OVERFLOW, STATE_BASE, 0);
        default:
            UPRV_UNREACHABLE;
    }
}

bool AffixUtils::hasNext(const AffixTag &tag, const UnicodeString &string) {
    // First check for the {-1} and default initializer syntax.
    if (tag.offset < 0) {
        return false;
    } else if (tag.offset == 0) {
        return string.length() > 0;
    }
    // The rest of the fields are safe to use now.
    // Special case: the last character in string is an end quote.
    if (tag.state == STATE_INSIDE_QUOTE && tag.offset == string.length() - 1 &&
        string.charAt(tag.offset) == u'\'') {
        return false;
    } else if (tag.state != STATE_BASE) {
        return true;
    } else {
        return tag.offset < string.length();
    }
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif


#include <stdlib.h>
#include <cmath>
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING












U_NAMESPACE_BEGIN namespace number {
namespace impl {

/**
 * A wrapper around LocalizedNumberFormatter implementing the Format interface, enabling improved
 * compatibility with other APIs.
 *
 * @draft ICU 62
 * @see NumberFormatter
 */
class U_I18N_API LocalizedNumberFormatterAsFormat : public Format {
  public:
    LocalizedNumberFormatterAsFormat(const LocalizedNumberFormatter& formatter, const Locale& locale);

    /**
     * Destructor.
     */
    ~LocalizedNumberFormatterAsFormat() U_OVERRIDE;

    /**
     * Equals operator.
     */
    UBool operator==(const Format& other) const U_OVERRIDE;

    /**
     * Creates a copy of this object.
     */
    LocalizedNumberFormatterAsFormat* clone() const U_OVERRIDE;

    /**
     * Formats a Number using the wrapped LocalizedNumberFormatter. The provided formattable must be a
     * number type.
     */
    UnicodeString& format(const Formattable& obj, UnicodeString& appendTo, FieldPosition& pos,
                          UErrorCode& status) const U_OVERRIDE;

    /**
     * Formats a Number using the wrapped LocalizedNumberFormatter. The provided formattable must be a
     * number type.
     */
    UnicodeString& format(const Formattable& obj, UnicodeString& appendTo, FieldPositionIterator* posIter,
                          UErrorCode& status) const U_OVERRIDE;

    /**
     * Not supported: sets an error index and returns.
     */
    void parseObject(const UnicodeString& source, Formattable& result,
                     ParsePosition& parse_pos) const U_OVERRIDE;

    /**
     * Gets the LocalizedNumberFormatter that this wrapper class uses to format numbers.
     *
     * For maximum efficiency, this function returns by const reference. You must copy the return value
     * into a local variable if you want to use it beyond the lifetime of the current object:
     *
     * <pre>
     * LocalizedNumberFormatter localFormatter = fmt->getNumberFormatter();
     * </pre>
     *
     * You can however use the return value directly when chaining:
     *
     * <pre>
     * FormattedNumber result = fmt->getNumberFormatter().formatDouble(514.23, status);
     * </pre>
     *
     * @return The unwrapped LocalizedNumberFormatter.
     */
    const LocalizedNumberFormatter& getNumberFormatter() const;

    UClassID getDynamicClassID() const U_OVERRIDE;
    static UClassID U_EXPORT2 getStaticClassID();

  private:
    LocalizedNumberFormatter fFormatter;

    // Even though the locale is inside the LocalizedNumberFormatter, we have to keep it here, too, because
    // LocalizedNumberFormatter doesn't have a getLocale() method, and ICU-TC didn't want to add one.
    Locale fLocale;
};

} // namespace impl
} // namespace number
U_NAMESPACE_END

#endif // __NUMBER_ASFORMAT_H__







using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(LocalizedNumberFormatterAsFormat)

LocalizedNumberFormatterAsFormat::LocalizedNumberFormatterAsFormat(
        const LocalizedNumberFormatter& formatter, const Locale& locale)
        : fFormatter(formatter), fLocale(locale) {
    const char* localeName = locale.getName();
    setLocaleIDs(localeName, localeName);
}

LocalizedNumberFormatterAsFormat::~LocalizedNumberFormatterAsFormat() = default;

UBool LocalizedNumberFormatterAsFormat::operator==(const Format& other) const {
    auto* _other = dynamic_cast<const LocalizedNumberFormatterAsFormat*>(&other);
    if (_other == nullptr) {
        return false;
    }
    // TODO: Change this to use LocalizedNumberFormatter::operator== if it is ever proposed.
    // This implementation is fine, but not particularly efficient.
    UErrorCode localStatus = U_ZERO_ERROR;
    return fFormatter.toSkeleton(localStatus) == _other->fFormatter.toSkeleton(localStatus);
}

LocalizedNumberFormatterAsFormat* LocalizedNumberFormatterAsFormat::clone() const {
    return new LocalizedNumberFormatterAsFormat(*this);
}

UnicodeString& LocalizedNumberFormatterAsFormat::format(const Formattable& obj, UnicodeString& appendTo,
                                                        FieldPosition& pos, UErrorCode& status) const {
    if (U_FAILURE(status)) { return appendTo; }
    UFormattedNumberData data;
    obj.populateDecimalQuantity(data.quantity, status);
    if (U_FAILURE(status)) {
        return appendTo;
    }
    fFormatter.formatImpl(&data, status);
    if (U_FAILURE(status)) {
        return appendTo;
    }
    // always return first occurrence:
    pos.setBeginIndex(0);
    pos.setEndIndex(0);
    bool found = data.nextFieldPosition(pos, status);
    if (found && appendTo.length() != 0) {
        pos.setBeginIndex(pos.getBeginIndex() + appendTo.length());
        pos.setEndIndex(pos.getEndIndex() + appendTo.length());
    }
    appendTo.append(data.toTempString(status));
    return appendTo;
}

UnicodeString& LocalizedNumberFormatterAsFormat::format(const Formattable& obj, UnicodeString& appendTo,
                                                        FieldPositionIterator* posIter,
                                                        UErrorCode& status) const {
    if (U_FAILURE(status)) { return appendTo; }
    UFormattedNumberData data;
    obj.populateDecimalQuantity(data.quantity, status);
    if (U_FAILURE(status)) {
        return appendTo;
    }
    fFormatter.formatImpl(&data, status);
    if (U_FAILURE(status)) {
        return appendTo;
    }
    appendTo.append(data.toTempString(status));
    if (posIter != nullptr) {
        FieldPositionIteratorHandler fpih(posIter, status);
        data.getAllFieldPositions(fpih, status);
    }
    return appendTo;
}

void LocalizedNumberFormatterAsFormat::parseObject(const UnicodeString&, Formattable&,
                                                   ParsePosition& parse_pos) const {
    // Not supported.
    parse_pos.setErrorIndex(0);
}

const LocalizedNumberFormatter& LocalizedNumberFormatterAsFormat::getNumberFormatter() const {
    return fFormatter;
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif









using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;


U_NAMESPACE_BEGIN
namespace number {
namespace impl {

/**
 * Implementation class for UNumberFormatter. Wraps a LocalizedNumberFormatter.
 */
struct UNumberFormatterData : public UMemory,
        // Magic number as ASCII == "NFR" (NumberFormatteR)
        public IcuCApiHelper<UNumberFormatter, UNumberFormatterData, 0x4E465200> {
    LocalizedNumberFormatter fFormatter;
};

struct UFormattedNumberImpl;

// Magic number as ASCII == "FDN" (FormatteDNumber)
typedef IcuCApiHelper<UFormattedNumber, UFormattedNumberImpl, 0x46444E00> UFormattedNumberApiHelper;

struct UFormattedNumberImpl : public UFormattedValueImpl, public UFormattedNumberApiHelper {
    UFormattedNumberImpl();
    ~UFormattedNumberImpl();

    FormattedNumber fImpl;
    UFormattedNumberData fData;
};

UFormattedNumberImpl::UFormattedNumberImpl()
        : fImpl(&fData) {
    fFormattedValue = &fImpl;
}

UFormattedNumberImpl::~UFormattedNumberImpl() {
    // Disown the data from fImpl so it doesn't get deleted twice
    fImpl.fData = nullptr;
}

}
}
U_NAMESPACE_END


UPRV_FORMATTED_VALUE_CAPI_NO_IMPLTYPE_AUTO_IMPL(
    UFormattedNumber,
    UFormattedNumberImpl,
    UFormattedNumberApiHelper,
    unumf)


const DecimalQuantity* icu::number::impl::validateUFormattedNumberToDecimalQuantity(
        const UFormattedNumber* uresult, UErrorCode& status) {
    auto* result = UFormattedNumberApiHelper::validate(uresult, status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    return &result->fData.quantity;
}



U_CAPI UNumberFormatter* U_EXPORT2
unumf_openForSkeletonAndLocale(const UChar* skeleton, int32_t skeletonLen, const char* locale,
                               UErrorCode* ec) {
    auto* impl = new UNumberFormatterData();
    if (impl == nullptr) {
        *ec = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    // Readonly-alias constructor (first argument is whether we are NUL-terminated)
    UnicodeString skeletonString(skeletonLen == -1, skeleton, skeletonLen);
    impl->fFormatter = NumberFormatter::forSkeleton(skeletonString, *ec).locale(locale);
    return impl->exportForC();
}

U_CAPI UNumberFormatter* U_EXPORT2
unumf_openForSkeletonAndLocaleWithError(const UChar* skeleton, int32_t skeletonLen, const char* locale,
                                         UParseError* perror, UErrorCode* ec) {
    auto* impl = new UNumberFormatterData();
    if (impl == nullptr) {
        *ec = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    // Readonly-alias constructor (first argument is whether we are NUL-terminated)
    UnicodeString skeletonString(skeletonLen == -1, skeleton, skeletonLen);
    impl->fFormatter = NumberFormatter::forSkeleton(skeletonString, *perror, *ec).locale(locale);
    return impl->exportForC();
}

U_CAPI void U_EXPORT2
unumf_formatInt(const UNumberFormatter* uformatter, int64_t value, UFormattedNumber* uresult,
                UErrorCode* ec) {
    const UNumberFormatterData* formatter = UNumberFormatterData::validate(uformatter, *ec);
    auto* result = UFormattedNumberApiHelper::validate(uresult, *ec);
    if (U_FAILURE(*ec)) { return; }

    result->fData.getStringRef().clear();
    result->fData.quantity.setToLong(value);
    formatter->fFormatter.formatImpl(&result->fData, *ec);
}

U_CAPI void U_EXPORT2
unumf_formatDouble(const UNumberFormatter* uformatter, double value, UFormattedNumber* uresult,
                   UErrorCode* ec) {
    const UNumberFormatterData* formatter = UNumberFormatterData::validate(uformatter, *ec);
    auto* result = UFormattedNumberApiHelper::validate(uresult, *ec);
    if (U_FAILURE(*ec)) { return; }

    result->fData.getStringRef().clear();
    result->fData.quantity.setToDouble(value);
    formatter->fFormatter.formatImpl(&result->fData, *ec);
}

U_CAPI void U_EXPORT2
unumf_formatDecimal(const UNumberFormatter* uformatter, const char* value, int32_t valueLen,
                    UFormattedNumber* uresult, UErrorCode* ec) {
    const UNumberFormatterData* formatter = UNumberFormatterData::validate(uformatter, *ec);
    auto* result = UFormattedNumberApiHelper::validate(uresult, *ec);
    if (U_FAILURE(*ec)) { return; }

    result->fData.getStringRef().clear();
    result->fData.quantity.setToDecNumber({value, valueLen}, *ec);
    if (U_FAILURE(*ec)) { return; }
    formatter->fFormatter.formatImpl(&result->fData, *ec);
}

U_CAPI int32_t U_EXPORT2
unumf_resultToString(const UFormattedNumber* uresult, UChar* buffer, int32_t bufferCapacity,
                     UErrorCode* ec) {
    const auto* result = UFormattedNumberApiHelper::validate(uresult, *ec);
    if (U_FAILURE(*ec)) { return 0; }

    if (buffer == nullptr ? bufferCapacity != 0 : bufferCapacity < 0) {
        *ec = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    return result->fImpl.toTempString(*ec).extract(buffer, bufferCapacity, *ec);
}

U_CAPI UBool U_EXPORT2
unumf_resultNextFieldPosition(const UFormattedNumber* uresult, UFieldPosition* ufpos, UErrorCode* ec) {
    const auto* result = UFormattedNumberApiHelper::validate(uresult, *ec);
    if (U_FAILURE(*ec)) { return FALSE; }

    if (ufpos == nullptr) {
        *ec = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }

    FieldPosition fp;
    fp.setField(ufpos->field);
    fp.setBeginIndex(ufpos->beginIndex);
    fp.setEndIndex(ufpos->endIndex);
    bool retval = result->fImpl.nextFieldPosition(fp, *ec);
    ufpos->beginIndex = fp.getBeginIndex();
    ufpos->endIndex = fp.getEndIndex();
    // NOTE: MSVC sometimes complains when implicitly converting between bool and UBool
    return retval ? TRUE : FALSE;
}

U_CAPI void U_EXPORT2
unumf_resultGetAllFieldPositions(const UFormattedNumber* uresult, UFieldPositionIterator* ufpositer,
                                 UErrorCode* ec) {
    const auto* result = UFormattedNumberApiHelper::validate(uresult, *ec);
    if (U_FAILURE(*ec)) { return; }

    if (ufpositer == nullptr) {
        *ec = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    auto* fpi = reinterpret_cast<FieldPositionIterator*>(ufpositer);
    result->fImpl.getAllFieldPositions(*fpi, *ec);
}

U_CAPI void U_EXPORT2
unumf_close(UNumberFormatter* f) {
    UErrorCode localStatus = U_ZERO_ERROR;
    const UNumberFormatterData* impl = UNumberFormatterData::validate(f, localStatus);
    delete impl;
}


#endif /* #if !UCONFIG_NO_FORMATTING */



























// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING






// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING






// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING









U_NAMESPACE_BEGIN

// Export an explicit template instantiation of the LocalPointer that is used as a
// data member of AdoptingModifierStore.
// (When building DLLs for Windows this is required.)
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN
#if defined(_MSC_VER)
// Ignore warning 4661 as LocalPointerBase does not use operator== or operator!=
#pragma warning(push)
#pragma warning(disable : 4661)
#endif
template class U_I18N_API LocalPointerBase<number::impl::AdoptingModifierStore>;
template class U_I18N_API LocalPointer<number::impl::AdoptingModifierStore>;
#if defined(_MSC_VER)
#pragma warning(pop)
#endif
#endif

namespace number {
namespace impl {

// Forward declaration
class MutablePatternModifier;

// Exported as U_I18N_API because it is needed for the unit test PatternModifierTest
class U_I18N_API ImmutablePatternModifier : public MicroPropsGenerator, public UMemory {
  public:
    ~ImmutablePatternModifier() U_OVERRIDE = default;

    void processQuantity(DecimalQuantity&, MicroProps& micros, UErrorCode& status) const U_OVERRIDE;

    void applyToMicros(MicroProps& micros, const DecimalQuantity& quantity, UErrorCode& status) const;

    const Modifier* getModifier(Signum signum, StandardPlural::Form plural) const;

  private:
    ImmutablePatternModifier(AdoptingModifierStore* pm, const PluralRules* rules,
                             const MicroPropsGenerator* parent);

    const LocalPointer<AdoptingModifierStore> pm;
    const PluralRules* rules;
    const MicroPropsGenerator* parent;

    friend class MutablePatternModifier;
};

/**
 * This class is a {@link Modifier} that wraps a decimal format pattern. It applies the pattern's affixes in
 * {@link Modifier#apply}.
 *
 * <p>
 * In addition to being a Modifier, this class contains the business logic for substituting the correct locale symbols
 * into the affixes of the decimal format pattern.
 *
 * <p>
 * In order to use this class, create a new instance and call the following four setters: {@link #setPatternInfo},
 * {@link #setPatternAttributes}, {@link #setSymbols}, and {@link #setNumberProperties}. After calling these four
 * setters, the instance will be ready for use as a Modifier.
 *
 * <p>
 * This is a MUTABLE, NON-THREAD-SAFE class designed for performance. Do NOT save references to this or attempt to use
 * it from multiple threads! Instead, you can obtain a safe, immutable decimal format pattern modifier by calling
 * {@link MutablePatternModifier#createImmutable}, in effect treating this instance as a builder for the immutable
 * variant.
 */
class U_I18N_API MutablePatternModifier
        : public MicroPropsGenerator,
          public Modifier,
          public SymbolProvider,
          public UMemory {
  public:

    ~MutablePatternModifier() U_OVERRIDE = default;

    /**
     * @param isStrong
     *            Whether the modifier should be considered strong. For more information, see
     *            {@link Modifier#isStrong()}. Most of the time, decimal format pattern modifiers should be considered
     *            as non-strong.
     */
    explicit MutablePatternModifier(bool isStrong);

    /**
     * Sets a reference to the parsed decimal format pattern, usually obtained from
     * {@link PatternStringParser#parseToPatternInfo(String)}, but any implementation of {@link AffixPatternProvider} is
     * accepted.
     *
     * @param field
     *            Which field to use for literal characters in the pattern.
     */
    void setPatternInfo(const AffixPatternProvider *patternInfo, Field field);

    /**
     * Sets attributes that imply changes to the literal interpretation of the pattern string affixes.
     *
     * @param signDisplay
     *            Whether to force a plus sign on positive numbers.
     * @param perMille
     *            Whether to substitute the percent sign in the pattern with a permille sign.
     */
    void setPatternAttributes(UNumberSignDisplay signDisplay, bool perMille);

    /**
     * Sets locale-specific details that affect the symbols substituted into the pattern string affixes.
     *
     * @param symbols
     *            The desired instance of DecimalFormatSymbols.
     * @param currencySymbols
     *            The currency symbols to be used when substituting currency values into the affixes.
     * @param unitWidth
     *            The width used to render currencies.
     * @param rules
     *            Required if the triple currency sign, "¤¤¤", appears in the pattern, which can be determined from the
     *            convenience method {@link #needsPlurals()}.
     */
    void setSymbols(const DecimalFormatSymbols* symbols, const CurrencySymbols* currencySymbols,
                    UNumberUnitWidth unitWidth, const PluralRules* rules);

    /**
     * Sets attributes of the current number being processed.
     *
     * @param signum
     *            -1 if negative; +1 if positive; or 0 if zero.
     * @param plural
     *            The plural form of the number, required only if the pattern contains the triple
     *            currency sign, "¤¤¤" (and as indicated by {@link #needsPlurals()}).
     */
    void setNumberProperties(Signum signum, StandardPlural::Form plural);

    /**
     * Returns true if the pattern represented by this MurkyModifier requires a plural keyword in order to localize.
     * This is currently true only if there is a currency long name placeholder in the pattern ("¤¤¤").
     */
    bool needsPlurals() const;

    /**
     * Creates a new quantity-dependent Modifier that behaves the same as the current instance, but which is immutable
     * and can be saved for future use. The number properties in the current instance are mutated; all other properties
     * are left untouched.
     *
     * <p>
     * The resulting modifier cannot be used in a QuantityChain.
     *
     * <p>
     * CREATES A NEW HEAP OBJECT; THE CALLER GETS OWNERSHIP.
     *
     * @return An immutable that supports both positive and negative numbers.
     */
    ImmutablePatternModifier *createImmutable(UErrorCode &status);

    /**
     * Creates a new quantity-dependent Modifier that behaves the same as the current instance, but which is immutable
     * and can be saved for future use. The number properties in the current instance are mutated; all other properties
     * are left untouched.
     *
     * <p>
     * CREATES A NEW HEAP OBJECT; THE CALLER GETS OWNERSHIP.
     *
     * @param parent
     *            The QuantityChain to which to chain this immutable.
     * @return An immutable that supports both positive and negative numbers.
     */
    ImmutablePatternModifier *
    createImmutableAndChain(const MicroPropsGenerator *parent, UErrorCode &status);

    MicroPropsGenerator &addToChain(const MicroPropsGenerator *parent);

    void processQuantity(DecimalQuantity &, MicroProps &micros, UErrorCode &status) const U_OVERRIDE;

    int32_t apply(FormattedStringBuilder &output, int32_t leftIndex, int32_t rightIndex,
                  UErrorCode &status) const U_OVERRIDE;

    int32_t getPrefixLength() const U_OVERRIDE;

    int32_t getCodePointCount() const U_OVERRIDE;

    bool isStrong() const U_OVERRIDE;

    bool containsField(UNumberFormatFields field) const U_OVERRIDE;

    void getParameters(Parameters& output) const U_OVERRIDE;

    bool semanticallyEquivalent(const Modifier& other) const U_OVERRIDE;

    /**
     * Returns the string that substitutes a given symbol type in a pattern.
     */
    UnicodeString getSymbol(AffixPatternType type) const U_OVERRIDE;

    UnicodeString toUnicodeString() const;

  private:
    // Modifier details (initialized in constructor)
    const bool fStrong;

    // Pattern details (initialized in setPatternInfo and setPatternAttributes)
    const AffixPatternProvider *fPatternInfo;
    Field fField;
    UNumberSignDisplay fSignDisplay;
    bool fPerMilleReplacesPercent;

    // Symbol details (initialized in setSymbols)
    const DecimalFormatSymbols *fSymbols;
    UNumberUnitWidth fUnitWidth;
    const CurrencySymbols *fCurrencySymbols;
    const PluralRules *fRules;

    // Number details (initialized in setNumberProperties)
    Signum fSignum;
    StandardPlural::Form fPlural;

    // QuantityChain details (initialized in addToChain)
    const MicroPropsGenerator *fParent;

    // Transient fields for rendering
    UnicodeString currentAffix;

    /**
     * Uses the current properties to create a single {@link ConstantMultiFieldModifier} with currency spacing support
     * if required.
     *
     * <p>
     * CREATES A NEW HEAP OBJECT; THE CALLER GETS OWNERSHIP.
     *
     * @param a
     *            A working FormattedStringBuilder object; passed from the outside to prevent the need to create many new
     *            instances if this method is called in a loop.
     * @param b
     *            Another working FormattedStringBuilder object.
     * @return The constant modifier object.
     */
    ConstantMultiFieldModifier *createConstantModifier(UErrorCode &status);

    int32_t insertPrefix(FormattedStringBuilder &sb, int position, UErrorCode &status);

    int32_t insertSuffix(FormattedStringBuilder &sb, int position, UErrorCode &status);

    void prepareAffix(bool isPrefix);
};


}  // namespace impl
}  // namespace number
U_NAMESPACE_END

#endif //__NUMBER_PATTERNMODIFIER_H__



U_NAMESPACE_BEGIN namespace number {
namespace impl {

static const int32_t COMPACT_MAX_DIGITS = 15;

class CompactData : public MultiplierProducer {
  public:
    CompactData();

    void populate(const Locale &locale, const char *nsName, CompactStyle compactStyle,
                  CompactType compactType, UErrorCode &status);

    int32_t getMultiplier(int32_t magnitude) const U_OVERRIDE;

    const UChar *getPattern(int32_t magnitude, StandardPlural::Form plural) const;

    void getUniquePatterns(UVector &output, UErrorCode &status) const;

  private:
    const UChar *patterns[(COMPACT_MAX_DIGITS + 1) * StandardPlural::COUNT];
    int8_t multipliers[COMPACT_MAX_DIGITS + 1];
    int8_t largestMagnitude;
    UBool isEmpty;

    class CompactDataSink : public ResourceSink {
      public:
        explicit CompactDataSink(CompactData &data) : data(data) {}

        void put(const char *key, ResourceValue &value, UBool /*noFallback*/, UErrorCode &status) U_OVERRIDE;

      private:
        CompactData &data;
    };
};

struct CompactModInfo {
    const ImmutablePatternModifier *mod;
    const UChar* patternString;
};

class CompactHandler : public MicroPropsGenerator, public UMemory {
  public:
    CompactHandler(CompactStyle compactStyle, const Locale &locale, const char *nsName,
                   CompactType compactType, const PluralRules *rules,
                   MutablePatternModifier *buildReference, const MicroPropsGenerator *parent,
                   UErrorCode &status);

    ~CompactHandler() U_OVERRIDE;

    void
    processQuantity(DecimalQuantity &quantity, MicroProps &micros, UErrorCode &status) const U_OVERRIDE;

  private:
    const PluralRules *rules;
    const MicroPropsGenerator *parent;
    // Initial capacity of 12 for 0K, 00K, 000K, ...M, ...B, and ...T
    MaybeStackArray<CompactModInfo, 12> precomputedMods;
    int32_t precomputedModsLength = 0;
    CompactData data;
    ParsedPatternInfo unsafePatternInfo;
    UBool safe;

    /** Used by the safe code path */
    void precomputeAllModifiers(MutablePatternModifier &buildReference, UErrorCode &status);
};


} // namespace impl
} // namespace number
U_NAMESPACE_END

#endif //__NUMBER_COMPACT_H__


// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// TODO: minimize includes











U_NAMESPACE_BEGIN namespace number {
namespace impl {

struct MicroProps : public MicroPropsGenerator {

    // NOTE: All of these fields are properly initialized in NumberFormatterImpl.
    RoundingImpl rounder;
    Grouper grouping;
    Padder padding;
    IntegerWidth integerWidth;
    UNumberSignDisplay sign;
    UNumberDecimalSeparatorDisplay decimal;
    bool useCurrency;
    char nsName[9];

    // Note: This struct has no direct ownership of the following pointers.
    const DecimalFormatSymbols* symbols;
    const Modifier* modOuter;
    const Modifier* modMiddle;
    const Modifier* modInner;

    // The following "helper" fields may optionally be used during the MicroPropsGenerator.
    // They live here to retain memory.
    struct {
        ScientificModifier scientificModifier;
        EmptyModifier emptyWeakModifier{false};
        EmptyModifier emptyStrongModifier{true};
        MultiplierFormatHandler multiplier;
    } helpers;


    MicroProps() = default;

    MicroProps(const MicroProps& other) = default;

    MicroProps& operator=(const MicroProps& other) = default;

    void processQuantity(DecimalQuantity&, MicroProps& micros, UErrorCode& status) const U_OVERRIDE {
        (void) status;
        if (this == &micros) {
            // Unsafe path: no need to perform a copy.
            U_ASSERT(!exhausted);
            micros.exhausted = true;
            U_ASSERT(exhausted);
        } else {
            // Safe path: copy self into the output micros.
            micros = *this;
        }
    }

  private:
    // Internal fields:
    bool exhausted = false;
};

} // namespace impl
} // namespace number
U_NAMESPACE_END

#endif // __NUMBER_MICROPROPS_H__




using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

namespace {

// A dummy object used when a "0" compact decimal entry is encountered. This is necessary
// in order to prevent falling back to root. Object equality ("==") is intended.
const UChar *USE_FALLBACK = u"<USE FALLBACK>";

/** Produces a string like "NumberElements/latn/patternsShort/decimalFormat". */
void getResourceBundleKey(const char *nsName, CompactStyle compactStyle, CompactType compactType,
                                 CharString &sb, UErrorCode &status) {
    sb.clear();
    sb.append("NumberElements/", status);
    sb.append(nsName, status);
    sb.append(compactStyle == CompactStyle::UNUM_SHORT ? "/patternsShort" : "/patternsLong", status);
    sb.append(compactType == CompactType::TYPE_DECIMAL ? "/decimalFormat" : "/currencyFormat", status);
}

int32_t number_compact_getIndex(int32_t magnitude, StandardPlural::Form plural) {
    return magnitude * StandardPlural::COUNT + plural;
}

int32_t countZeros(const UChar *patternString, int32_t patternLength) {
    // NOTE: This strategy for computing the number of zeros is a hack for efficiency.
    // It could break if there are any 0s that aren't part of the main pattern.
    int32_t numZeros = 0;
    for (int32_t i = 0; i < patternLength; i++) {
        if (patternString[i] == u'0') {
            numZeros++;
        } else if (numZeros > 0) {
            break; // zeros should always be contiguous
        }
    }
    return numZeros;
}

} // namespace

// NOTE: patterns and multipliers both get zero-initialized.
CompactData::CompactData() : patterns(), multipliers(), largestMagnitude(0), isEmpty(TRUE) {
}

void CompactData::populate(const Locale &locale, const char *nsName, CompactStyle compactStyle,
                           CompactType compactType, UErrorCode &status) {
    CompactDataSink sink(*this);
    LocalUResourceBundlePointer rb(ures_open(nullptr, locale.getName(), &status));
    if (U_FAILURE(status)) { return; }

    bool nsIsLatn = strcmp(nsName, "latn") == 0;
    bool compactIsShort = compactStyle == CompactStyle::UNUM_SHORT;

    // Fall back to latn numbering system and/or short compact style.
    CharString resourceKey;
    getResourceBundleKey(nsName, compactStyle, compactType, resourceKey, status);
    UErrorCode localStatus = U_ZERO_ERROR;
    ures_getAllItemsWithFallback(rb.getAlias(), resourceKey.data(), sink, localStatus);
    if (isEmpty && !nsIsLatn) {
        getResourceBundleKey("latn", compactStyle, compactType, resourceKey, status);
        localStatus = U_ZERO_ERROR;
        ures_getAllItemsWithFallback(rb.getAlias(), resourceKey.data(), sink, localStatus);
    }
    if (isEmpty && !compactIsShort) {
        getResourceBundleKey(nsName, CompactStyle::UNUM_SHORT, compactType, resourceKey, status);
        localStatus = U_ZERO_ERROR;
        ures_getAllItemsWithFallback(rb.getAlias(), resourceKey.data(), sink, localStatus);
    }
    if (isEmpty && !nsIsLatn && !compactIsShort) {
        getResourceBundleKey("latn", CompactStyle::UNUM_SHORT, compactType, resourceKey, status);
        localStatus = U_ZERO_ERROR;
        ures_getAllItemsWithFallback(rb.getAlias(), resourceKey.data(), sink, localStatus);
    }

    // The last fallback should be guaranteed to return data.
    if (isEmpty) {
        status = U_INTERNAL_PROGRAM_ERROR;
    }
}

int32_t CompactData::getMultiplier(int32_t magnitude) const {
    if (magnitude < 0) {
        return 0;
    }
    if (magnitude > largestMagnitude) {
        magnitude = largestMagnitude;
    }
    return multipliers[magnitude];
}

const UChar *CompactData::getPattern(int32_t magnitude, StandardPlural::Form plural) const {
    if (magnitude < 0) {
        return nullptr;
    }
    if (magnitude > largestMagnitude) {
        magnitude = largestMagnitude;
    }
    const UChar *patternString = patterns[number_compact_getIndex(magnitude, plural)];
    if (patternString == nullptr && plural != StandardPlural::OTHER) {
        // Fall back to "other" plural variant
        patternString = patterns[number_compact_getIndex(magnitude, StandardPlural::OTHER)];
    }
    if (patternString == USE_FALLBACK) { // == is intended
        // Return null if USE_FALLBACK is present
        patternString = nullptr;
    }
    return patternString;
}

void CompactData::getUniquePatterns(UVector &output, UErrorCode &status) const {
    U_ASSERT(output.isEmpty());
    // NOTE: In C++, this is done more manually with a UVector.
    // In Java, we can take advantage of JDK HashSet.
    for (auto pattern : patterns) {
        if (pattern == nullptr || pattern == USE_FALLBACK) {
            continue;
        }

        // Insert pattern into the UVector if the UVector does not already contain the pattern.
        // Search the UVector from the end since identical patterns are likely to be adjacent.
        for (int32_t i = output.size() - 1; i >= 0; i--) {
            if (u_strcmp(pattern, static_cast<const UChar *>(output[i])) == 0) {
                goto continue_outer;
            }
        }

        // The string was not found; add it to the UVector.
        // ANDY: This requires a const_cast.  Why?
        output.addElement(const_cast<UChar *>(pattern), status);

        continue_outer:
        continue;
    }
}

void CompactData::CompactDataSink::put(const char *key, ResourceValue &value, UBool /*noFallback*/,
                                       UErrorCode &status) {
    // traverse into the table of powers of ten
    ResourceTable powersOfTenTable = value.getTable(status);
    if (U_FAILURE(status)) { return; }
    for (int i3 = 0; powersOfTenTable.getKeyAndValue(i3, key, value); ++i3) {

        // Assumes that the keys are always of the form "10000" where the magnitude is the
        // length of the key minus one.  We expect magnitudes to be less than MAX_DIGITS.
        auto magnitude = static_cast<int8_t> (strlen(key) - 1);
        int8_t multiplier = data.multipliers[magnitude];
        U_ASSERT(magnitude < COMPACT_MAX_DIGITS);

        // Iterate over the plural variants ("one", "other", etc)
        ResourceTable pluralVariantsTable = value.getTable(status);
        if (U_FAILURE(status)) { return; }
        for (int i4 = 0; pluralVariantsTable.getKeyAndValue(i4, key, value); ++i4) {

            // Skip this magnitude/plural if we already have it from a child locale.
            // Note: This also skips USE_FALLBACK entries.
            StandardPlural::Form plural = StandardPlural::fromString(key, status);
            if (U_FAILURE(status)) { return; }
            if (data.patterns[number_compact_getIndex(magnitude, plural)] != nullptr) {
                continue;
            }

            // The value "0" means that we need to use the default pattern and not fall back
            // to parent locales. Example locale where this is relevant: 'it'.
            int32_t patternLength;
            const UChar *patternString = value.getString(patternLength, status);
            if (U_FAILURE(status)) { return; }
            if (u_strcmp(patternString, u"0") == 0) {
                patternString = USE_FALLBACK;
                patternLength = 0;
            }

            // Save the pattern string. We will parse it lazily.
            data.patterns[number_compact_getIndex(magnitude, plural)] = patternString;

            // If necessary, compute the multiplier: the difference between the magnitude
            // and the number of zeros in the pattern.
            if (multiplier == 0) {
                int32_t numZeros = countZeros(patternString, patternLength);
                if (numZeros > 0) { // numZeros==0 in certain cases, like Somali "Kun"
                    multiplier = static_cast<int8_t> (numZeros - magnitude - 1);
                }
            }
        }

        // Save the multiplier.
        if (data.multipliers[magnitude] == 0) {
            data.multipliers[magnitude] = multiplier;
            if (magnitude > data.largestMagnitude) {
                data.largestMagnitude = magnitude;
            }
            data.isEmpty = false;
        } else {
            U_ASSERT(data.multipliers[magnitude] == multiplier);
        }
    }
}

///////////////////////////////////////////////////////////
/// END OF CompactData.java; BEGIN CompactNotation.java ///
///////////////////////////////////////////////////////////

CompactHandler::CompactHandler(CompactStyle compactStyle, const Locale &locale, const char *nsName,
                               CompactType compactType, const PluralRules *rules,
                               MutablePatternModifier *buildReference, const MicroPropsGenerator *parent,
                               UErrorCode &status)
        : rules(rules), parent(parent) {
    data.populate(locale, nsName, compactStyle, compactType, status);
    if (buildReference != nullptr) {
        // Safe code path
        precomputeAllModifiers(*buildReference, status);
        safe = TRUE;
    } else {
        // Unsafe code path
        safe = FALSE;
    }
}

CompactHandler::~CompactHandler() {
    for (int32_t i = 0; i < precomputedModsLength; i++) {
        delete precomputedMods[i].mod;
    }
}

void CompactHandler::precomputeAllModifiers(MutablePatternModifier &buildReference, UErrorCode &status) {
    if (U_FAILURE(status)) { return; }

    // Initial capacity of 12 for 0K, 00K, 000K, ...M, ...B, and ...T
    UVector allPatterns(12, status);
    if (U_FAILURE(status)) { return; }
    data.getUniquePatterns(allPatterns, status);
    if (U_FAILURE(status)) { return; }

    // C++ only: ensure that precomputedMods has room.
    precomputedModsLength = allPatterns.size();
    if (precomputedMods.getCapacity() < precomputedModsLength) {
        precomputedMods.resize(allPatterns.size(), status);
        if (U_FAILURE(status)) { return; }
    }

    for (int32_t i = 0; i < precomputedModsLength; i++) {
        auto patternString = static_cast<const UChar *>(allPatterns[i]);
        UnicodeString hello(patternString);
        CompactModInfo &info = precomputedMods[i];
        ParsedPatternInfo patternInfo;
        PatternParser::parseToPatternInfo(UnicodeString(patternString), patternInfo, status);
        if (U_FAILURE(status)) { return; }
        buildReference.setPatternInfo(&patternInfo, UNUM_COMPACT_FIELD);
        info.mod = buildReference.createImmutable(status);
        if (U_FAILURE(status)) { return; }
        info.patternString = patternString;
    }
}

void CompactHandler::processQuantity(DecimalQuantity &quantity, MicroProps &micros,
                                     UErrorCode &status) const {
    parent->processQuantity(quantity, micros, status);
    if (U_FAILURE(status)) { return; }

    // Treat zero, NaN, and infinity as if they had magnitude 0
    int32_t magnitude;
    if (quantity.isZeroish()) {
        magnitude = 0;
        micros.rounder.apply(quantity, status);
    } else {
        // TODO: Revisit chooseMultiplierAndApply
        int32_t multiplier = micros.rounder.chooseMultiplierAndApply(quantity, data, status);
        magnitude = quantity.isZeroish() ? 0 : quantity.getMagnitude();
        magnitude -= multiplier;
    }

    StandardPlural::Form plural = utils::getStandardPlural(rules, quantity);
    const UChar *patternString = data.getPattern(magnitude, plural);
    if (patternString == nullptr) {
        // Use the default (non-compact) modifier.
        // No need to take any action.
    } else if (safe) {
        // Safe code path.
        // Java uses a hash set here for O(1) lookup.  C++ uses a linear search.
        // TODO: Benchmark this and maybe change to a binary search or hash table.
        int32_t i = 0;
        for (; i < precomputedModsLength; i++) {
            const CompactModInfo &info = precomputedMods[i];
            if (u_strcmp(patternString, info.patternString) == 0) {
                info.mod->applyToMicros(micros, quantity, status);
                break;
            }
        }
        // It should be guaranteed that we found the entry.
        U_ASSERT(i < precomputedModsLength);
    } else {
        // Unsafe code path.
        // Overwrite the PatternInfo in the existing modMiddle.
        // C++ Note: Use unsafePatternInfo for proper lifecycle.
        ParsedPatternInfo &patternInfo = const_cast<CompactHandler *>(this)->unsafePatternInfo;
        PatternParser::parseToPatternInfo(UnicodeString(patternString), patternInfo, status);
        static_cast<MutablePatternModifier*>(const_cast<Modifier*>(micros.modMiddle))
            ->setPatternInfo(&patternInfo, UNUM_COMPACT_FIELD);
    }

    // We already performed rounding. Do not perform it again.
    micros.rounder = RoundingImpl::passThrough();
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif





using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;


CurrencySymbols::CurrencySymbols(CurrencyUnit currency, const Locale& locale, UErrorCode& status)
        : fCurrency(currency), fLocaleName(locale.getName(), status) {
    fCurrencySymbol.setToBogus();
    fIntlCurrencySymbol.setToBogus();
}

CurrencySymbols::CurrencySymbols(CurrencyUnit currency, const Locale& locale,
                                 const DecimalFormatSymbols& symbols, UErrorCode& status)
        : CurrencySymbols(currency, locale, status) {
    // If either of the overrides is present, save it in the local UnicodeString.
    if (symbols.isCustomCurrencySymbol()) {
        fCurrencySymbol = symbols.getConstSymbol(DecimalFormatSymbols::kCurrencySymbol);
    }
    if (symbols.isCustomIntlCurrencySymbol()) {
        fIntlCurrencySymbol = symbols.getConstSymbol(DecimalFormatSymbols::kIntlCurrencySymbol);
    }
}

const char16_t* CurrencySymbols::getIsoCode() const {
    return fCurrency.getISOCurrency();
}

UnicodeString CurrencySymbols::getNarrowCurrencySymbol(UErrorCode& status) const {
    // Note: currently no override is available for narrow currency symbol
    return loadSymbol(UCURR_NARROW_SYMBOL_NAME, status);
}

UnicodeString CurrencySymbols::getCurrencySymbol(UErrorCode& status) const {
    if (!fCurrencySymbol.isBogus()) {
        return fCurrencySymbol;
    }
    return loadSymbol(UCURR_SYMBOL_NAME, status);
}

UnicodeString CurrencySymbols::loadSymbol(UCurrNameStyle selector, UErrorCode& status) const {
    const char16_t* isoCode = fCurrency.getISOCurrency();
    int32_t symbolLen = 0;
    const char16_t* symbol = ucurr_getName(
            isoCode,
            fLocaleName.data(),
            selector,
            nullptr /* isChoiceFormat */,
            &symbolLen,
            &status);
    // If given an unknown currency, ucurr_getName returns the input string, which we can't alias safely!
    // Otherwise, symbol points to a resource bundle, and we can use readonly-aliasing constructor.
    if (symbol == isoCode) {
        return UnicodeString(isoCode, 3);
    } else {
        return UnicodeString(TRUE, symbol, symbolLen);
    }
}

UnicodeString CurrencySymbols::getIntlCurrencySymbol(UErrorCode&) const {
    if (!fIntlCurrencySymbol.isBogus()) {
        return fIntlCurrencySymbol;
    }
    // Note: Not safe to use readonly-aliasing constructor here because the buffer belongs to this object,
    // which could be destructed or moved during the lifetime of the return value.
    return UnicodeString(fCurrency.getISOCurrency(), 3);
}

UnicodeString CurrencySymbols::getPluralName(StandardPlural::Form plural, UErrorCode& status) const {
    const char16_t* isoCode = fCurrency.getISOCurrency();
    int32_t symbolLen = 0;
    const char16_t* symbol = ucurr_getPluralName(
            isoCode,
            fLocaleName.data(),
            nullptr /* isChoiceFormat */,
            StandardPlural::getKeyword(plural),
            &symbolLen,
            &status);
    // If given an unknown currency, ucurr_getName returns the input string, which we can't alias safely!
    // Otherwise, symbol points to a resource bundle, and we can use readonly-aliasing constructor.
    if (symbol == isoCode) {
        return UnicodeString(isoCode, 3);
    } else {
        return UnicodeString(TRUE, symbol, symbolLen);
    }
}


CurrencyUnit
icu::number::impl::resolveCurrency(const DecimalFormatProperties& properties, const Locale& locale,
                                   UErrorCode& status) {
    if (!properties.currency.isNull()) {
        return properties.currency.getNoError();
    } else {
        UErrorCode localStatus = U_ZERO_ERROR;
        char16_t buf[4] = {};
        ucurr_forLocale(locale.getName(), buf, 4, &localStatus);
        if (U_SUCCESS(localStatus)) {
            return CurrencyUnit(buf, status);
        } else {
            // Default currency (XXX)
            return CurrencyUnit();
        }
    }
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

#include <cstdlib>
#include <cmath>
#include <limits>
#include <stdlib.h>



// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

#define DECNUM_INITIAL_CAPACITY 34

// Export an explicit template instantiation of the MaybeStackHeaderAndArray that is used as a data member of DecNum.
// When building DLLs for Windows this is required even though no direct access to the MaybeStackHeaderAndArray leaks out of the i18n library.
// (See digitlst.h, pluralaffix.h, datefmt.h, and others for similar examples.)
#if U_PF_WINDOWS <= U_PLATFORM && U_PLATFORM <= U_PF_CYGWIN
template class U_I18N_API MaybeStackHeaderAndArray<decNumber, char, DECNUM_INITIAL_CAPACITY>;
#endif

namespace number {
namespace impl {

/** A very thin C++ wrapper around decNumber.h */
// Exported as U_I18N_API for tests
class U_I18N_API DecNum : public UMemory {
  public:
    DecNum();  // leaves object in valid but undefined state

    // Copy-like constructor; use the default move operators.
    DecNum(const DecNum& other, UErrorCode& status);

    /** Sets the decNumber to the StringPiece. */
    void setTo(StringPiece str, UErrorCode& status);

    /** Sets the decNumber to the NUL-terminated char string. */
    void setTo(const char* str, UErrorCode& status);

    /** Uses double_conversion to set this decNumber to the given double. */
    void setTo(double d, UErrorCode& status);

    /** Sets the decNumber to the BCD representation. */
    void setTo(const uint8_t* bcd, int32_t length, int32_t scale, bool isNegative, UErrorCode& status);

    void normalize();

    void multiplyBy(const DecNum& rhs, UErrorCode& status);

    void divideBy(const DecNum& rhs, UErrorCode& status);

    bool isNegative() const;

    bool isZero() const;

    void toString(ByteSink& output, UErrorCode& status) const;

    inline const decNumber* getRawDecNumber() const {
        return fData.getAlias();
    }

  private:
    static constexpr int32_t kDefaultDigits = DECNUM_INITIAL_CAPACITY;
    MaybeStackHeaderAndArray<decNumber, char, kDefaultDigits> fData;
    decContext fContext;

    void _setTo(const char* str, int32_t maxDigits, UErrorCode& status);
};

} // namespace impl
} // namespace number

U_NAMESPACE_END

#endif // __NUMBER_DECNUM_H__





// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
//
// From the double-conversion library. Original license:
//
// Copyright 2012 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ICU PATCH: ifdef around UCONFIG_NO_FORMATTING

#if !UCONFIG_NO_FORMATTING


// ICU PATCH: Customize header file paths for ICU.




#endif  // DOUBLE_CONVERSION_DOUBLE_CONVERSION_H_





using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

using icu::double_conversion::DoubleToStringConverter;
using icu::double_conversion::StringToDoubleConverter;

namespace {

int8_t NEGATIVE_FLAG = 1;
int8_t INFINITY_FLAG = 2;
int8_t NAN_FLAG = 4;

/** Helper function for safe subtraction (no overflow). */
inline int32_t safeSubtract(int32_t a, int32_t b) {
    // Note: In C++, signed integer subtraction is undefined behavior.
    int32_t diff = static_cast<int32_t>(static_cast<uint32_t>(a) - static_cast<uint32_t>(b));
    if (b < 0 && diff < a) { return INT32_MAX; }
    if (b > 0 && diff > a) { return INT32_MIN; }
    return diff;
}

static double DOUBLE_MULTIPLIERS[] = {
        1e0,
        1e1,
        1e2,
        1e3,
        1e4,
        1e5,
        1e6,
        1e7,
        1e8,
        1e9,
        1e10,
        1e11,
        1e12,
        1e13,
        1e14,
        1e15,
        1e16,
        1e17,
        1e18,
        1e19,
        1e20,
        1e21};

}  // namespace

icu::IFixedDecimal::~IFixedDecimal() = default;

DecimalQuantity::DecimalQuantity() {
    setBcdToZero();
    flags = 0;
}

DecimalQuantity::~DecimalQuantity() {
    if (usingBytes) {
        uprv_free(fBCD.bcdBytes.ptr);
        fBCD.bcdBytes.ptr = nullptr;
        usingBytes = false;
    }
}

DecimalQuantity::DecimalQuantity(const DecimalQuantity &other) {
    *this = other;
}

DecimalQuantity::DecimalQuantity(DecimalQuantity&& src) U_NOEXCEPT {
    *this = std::move(src);
}

DecimalQuantity &DecimalQuantity::operator=(const DecimalQuantity &other) {
    if (this == &other) {
        return *this;
    }
    copyBcdFrom(other);
    copyFieldsFrom(other);
    return *this;
}

DecimalQuantity& DecimalQuantity::operator=(DecimalQuantity&& src) U_NOEXCEPT {
    if (this == &src) {
        return *this;
    }
    moveBcdFrom(src);
    copyFieldsFrom(src);
    return *this;
}

void DecimalQuantity::copyFieldsFrom(const DecimalQuantity& other) {
    bogus = other.bogus;
    lReqPos = other.lReqPos;
    rReqPos = other.rReqPos;
    scale = other.scale;
    precision = other.precision;
    flags = other.flags;
    origDouble = other.origDouble;
    origDelta = other.origDelta;
    isApproximate = other.isApproximate;
}

void DecimalQuantity::clear() {
    lReqPos = 0;
    rReqPos = 0;
    flags = 0;
    setBcdToZero(); // sets scale, precision, hasDouble, origDouble, origDelta, and BCD data
}

void DecimalQuantity::setMinInteger(int32_t minInt) {
    // Validation should happen outside of DecimalQuantity, e.g., in the Precision class.
    U_ASSERT(minInt >= 0);

    // Special behavior: do not set minInt to be less than what is already set.
    // This is so significant digits rounding can set the integer length.
    if (minInt < lReqPos) {
        minInt = lReqPos;
    }

    // Save values into internal state
    lReqPos = minInt;
}

void DecimalQuantity::setMinFraction(int32_t minFrac) {
    // Validation should happen outside of DecimalQuantity, e.g., in the Precision class.
    U_ASSERT(minFrac >= 0);

    // Save values into internal state
    // Negation is safe for minFrac/maxFrac because -Integer.MAX_VALUE > Integer.MIN_VALUE
    rReqPos = -minFrac;
}

void DecimalQuantity::applyMaxInteger(int32_t maxInt) {
    // Validation should happen outside of DecimalQuantity, e.g., in the Precision class.
    U_ASSERT(maxInt >= 0);

    if (precision == 0) {
        return;
    }

    if (maxInt <= scale) {
        setBcdToZero();
        return;
    }

    int32_t magnitude = getMagnitude();
    if (maxInt <= magnitude) {
        popFromLeft(magnitude - maxInt + 1);
        compact();
    }
}

uint64_t DecimalQuantity::getPositionFingerprint() const {
    uint64_t fingerprint = 0;
    fingerprint ^= (lReqPos << 16);
    fingerprint ^= (static_cast<uint64_t>(rReqPos) << 32);
    return fingerprint;
}

void DecimalQuantity::roundToIncrement(double roundingIncrement, RoundingMode roundingMode,
                                       UErrorCode& status) {
    // Do not call this method with an increment having only a 1 or a 5 digit!
    // Use a more efficient call to either roundToMagnitude() or roundToNickel().
    // Check a few popular rounding increments; a more thorough check is in Java.
    U_ASSERT(roundingIncrement != 0.01);
    U_ASSERT(roundingIncrement != 0.05);
    U_ASSERT(roundingIncrement != 0.1);
    U_ASSERT(roundingIncrement != 0.5);
    U_ASSERT(roundingIncrement != 1);
    U_ASSERT(roundingIncrement != 5);

    DecNum incrementDN;
    incrementDN.setTo(roundingIncrement, status);
    if (U_FAILURE(status)) { return; }

    // Divide this DecimalQuantity by the increment, round, then multiply back.
    divideBy(incrementDN, status);
    if (U_FAILURE(status)) { return; }
    roundToMagnitude(0, roundingMode, status);
    if (U_FAILURE(status)) { return; }
    multiplyBy(incrementDN, status);
    if (U_FAILURE(status)) { return; }
}

void DecimalQuantity::multiplyBy(const DecNum& multiplicand, UErrorCode& status) {
    if (isZeroish()) {
        return;
    }
    // Convert to DecNum, multiply, and convert back.
    DecNum decnum;
    toDecNum(decnum, status);
    if (U_FAILURE(status)) { return; }
    decnum.multiplyBy(multiplicand, status);
    if (U_FAILURE(status)) { return; }
    setToDecNum(decnum, status);
}

void DecimalQuantity::divideBy(const DecNum& divisor, UErrorCode& status) {
    if (isZeroish()) {
        return;
    }
    // Convert to DecNum, multiply, and convert back.
    DecNum decnum;
    toDecNum(decnum, status);
    if (U_FAILURE(status)) { return; }
    decnum.divideBy(divisor, status);
    if (U_FAILURE(status)) { return; }
    setToDecNum(decnum, status);
}

void DecimalQuantity::negate() {
    flags ^= NEGATIVE_FLAG;
}

int32_t DecimalQuantity::getMagnitude() const {
    U_ASSERT(precision != 0);
    return scale + precision - 1;
}

bool DecimalQuantity::adjustMagnitude(int32_t delta) {
    if (precision != 0) {
        // i.e., scale += delta; origDelta += delta
        bool overflow = uprv_add32_overflow(scale, delta, &scale);
        overflow = uprv_add32_overflow(origDelta, delta, &origDelta) || overflow;
        // Make sure that precision + scale won't overflow, either
        int32_t dummy;
        overflow = overflow || uprv_add32_overflow(scale, precision, &dummy);
        return overflow;
    }
    return false;
}

double DecimalQuantity::getPluralOperand(PluralOperand operand) const {
    // If this assertion fails, you need to call roundToInfinity() or some other rounding method.
    // See the comment at the top of this file explaining the "isApproximate" field.
    U_ASSERT(!isApproximate);

    switch (operand) {
        case PLURAL_OPERAND_I:
            // Invert the negative sign if necessary
            return static_cast<double>(isNegative() ? -toLong(true) : toLong(true));
        case PLURAL_OPERAND_F:
            return static_cast<double>(toFractionLong(true));
        case PLURAL_OPERAND_T:
            return static_cast<double>(toFractionLong(false));
        case PLURAL_OPERAND_V:
            return fractionCount();
        case PLURAL_OPERAND_W:
            return fractionCountWithoutTrailingZeros();
        default:
            return std::abs(toDouble());
    }
}

bool DecimalQuantity::hasIntegerValue() const {
    return scale >= 0;
}

int32_t DecimalQuantity::getUpperDisplayMagnitude() const {
    // If this assertion fails, you need to call roundToInfinity() or some other rounding method.
    // See the comment in the header file explaining the "isApproximate" field.
    U_ASSERT(!isApproximate);

    int32_t magnitude = scale + precision;
    int32_t result = (lReqPos > magnitude) ? lReqPos : magnitude;
    return result - 1;
}

int32_t DecimalQuantity::getLowerDisplayMagnitude() const {
    // If this assertion fails, you need to call roundToInfinity() or some other rounding method.
    // See the comment in the header file explaining the "isApproximate" field.
    U_ASSERT(!isApproximate);

    int32_t magnitude = scale;
    int32_t result = (rReqPos < magnitude) ? rReqPos : magnitude;
    return result;
}

int8_t DecimalQuantity::getDigit(int32_t magnitude) const {
    // If this assertion fails, you need to call roundToInfinity() or some other rounding method.
    // See the comment at the top of this file explaining the "isApproximate" field.
    U_ASSERT(!isApproximate);

    return getDigitPos(magnitude - scale);
}

int32_t DecimalQuantity::fractionCount() const {
    return -getLowerDisplayMagnitude();
}

int32_t DecimalQuantity::fractionCountWithoutTrailingZeros() const {
    return -scale > 0 ? -scale : 0;  // max(-scale, 0)
}

bool DecimalQuantity::isNegative() const {
    return (flags & NEGATIVE_FLAG) != 0;
}

Signum DecimalQuantity::signum() const {
    if (isNegative()) {
        return SIGNUM_NEG;
    } else if (isZeroish() && !isInfinite()) {
        return SIGNUM_ZERO;
    } else {
        return SIGNUM_POS;
    }
}

bool DecimalQuantity::isInfinite() const {
    return (flags & INFINITY_FLAG) != 0;
}

bool DecimalQuantity::isNaN() const {
    return (flags & NAN_FLAG) != 0;
}

bool DecimalQuantity::isZeroish() const {
    return precision == 0;
}

DecimalQuantity &DecimalQuantity::setToInt(int32_t n) {
    setBcdToZero();
    flags = 0;
    if (n == INT32_MIN) {
        flags |= NEGATIVE_FLAG;
        // leave as INT32_MIN; handled below in _setToInt()
    } else if (n < 0) {
        flags |= NEGATIVE_FLAG;
        n = -n;
    }
    if (n != 0) {
        _setToInt(n);
        compact();
    }
    return *this;
}

void DecimalQuantity::_setToInt(int32_t n) {
    if (n == INT32_MIN) {
        readLongToBcd(-static_cast<int64_t>(n));
    } else {
        readIntToBcd(n);
    }
}

DecimalQuantity &DecimalQuantity::setToLong(int64_t n) {
    setBcdToZero();
    flags = 0;
    if (n < 0 && n > INT64_MIN) {
        flags |= NEGATIVE_FLAG;
        n = -n;
    }
    if (n != 0) {
        _setToLong(n);
        compact();
    }
    return *this;
}

void DecimalQuantity::_setToLong(int64_t n) {
    if (n == INT64_MIN) {
        DecNum decnum;
        UErrorCode localStatus = U_ZERO_ERROR;
        decnum.setTo("9.223372036854775808E+18", localStatus);
        if (U_FAILURE(localStatus)) { return; } // unexpected
        flags |= NEGATIVE_FLAG;
        readDecNumberToBcd(decnum);
    } else if (n <= INT32_MAX) {
        readIntToBcd(static_cast<int32_t>(n));
    } else {
        readLongToBcd(n);
    }
}

DecimalQuantity &DecimalQuantity::setToDouble(double n) {
    setBcdToZero();
    flags = 0;
    // signbit() from <math.h> handles +0.0 vs -0.0
    if (std::signbit(n)) {
        flags |= NEGATIVE_FLAG;
        n = -n;
    }
    if (std::isnan(n) != 0) {
        flags |= NAN_FLAG;
    } else if (std::isfinite(n) == 0) {
        flags |= INFINITY_FLAG;
    } else if (n != 0) {
        _setToDoubleFast(n);
        compact();
    }
    return *this;
}

void DecimalQuantity::_setToDoubleFast(double n) {
    isApproximate = true;
    origDouble = n;
    origDelta = 0;

    // Make sure the double is an IEEE 754 double.  If not, fall back to the slow path right now.
    // TODO: Make a fast path for other types of doubles.
    if (!std::numeric_limits<double>::is_iec559) {
        convertToAccurateDouble();
        // Turn off the approximate double flag, since the value is now exact.
        isApproximate = false;
        origDouble = 0.0;
        return;
    }

    // To get the bits from the double, use memcpy, which takes care of endianness.
    uint64_t ieeeBits;
    uprv_memcpy(&ieeeBits, &n, sizeof(n));
    int32_t exponent = static_cast<int32_t>((ieeeBits & 0x7ff0000000000000L) >> 52) - 0x3ff;

    // Not all integers can be represented exactly for exponent > 52
    if (exponent <= 52 && static_cast<int64_t>(n) == n) {
        _setToLong(static_cast<int64_t>(n));
        return;
    }

    // 3.3219... is log2(10)
    auto fracLength = static_cast<int32_t> ((52 - exponent) / 3.32192809489);
    if (fracLength >= 0) {
        int32_t i = fracLength;
        // 1e22 is the largest exact double.
        for (; i >= 22; i -= 22) n *= 1e22;
        n *= DOUBLE_MULTIPLIERS[i];
    } else {
        int32_t i = fracLength;
        // 1e22 is the largest exact double.
        for (; i <= -22; i += 22) n /= 1e22;
        n /= DOUBLE_MULTIPLIERS[-i];
    }
    auto result = static_cast<int64_t>(std::round(n));
    if (result != 0) {
        _setToLong(result);
        scale -= fracLength;
    }
}

void DecimalQuantity::convertToAccurateDouble() {
    U_ASSERT(origDouble != 0);
    int32_t delta = origDelta;

    // Call the slow oracle function (Double.toString in Java, DoubleToAscii in C++).
    char buffer[DoubleToStringConverter::kBase10MaximalLength + 1];
    bool sign; // unused; always positive
    int32_t length;
    int32_t point;
    DoubleToStringConverter::DoubleToAscii(
        origDouble,
        DoubleToStringConverter::DtoaMode::SHORTEST,
        0,
        buffer,
        sizeof(buffer),
        &sign,
        &length,
        &point
    );

    setBcdToZero();
    readDoubleConversionToBcd(buffer, length, point);
    scale += delta;
    explicitExactDouble = true;
}

DecimalQuantity &DecimalQuantity::setToDecNumber(StringPiece n, UErrorCode& status) {
    setBcdToZero();
    flags = 0;

    // Compute the decNumber representation
    DecNum decnum;
    decnum.setTo(n, status);

    _setToDecNum(decnum, status);
    return *this;
}

DecimalQuantity& DecimalQuantity::setToDecNum(const DecNum& decnum, UErrorCode& status) {
    setBcdToZero();
    flags = 0;

    _setToDecNum(decnum, status);
    return *this;
}

void DecimalQuantity::_setToDecNum(const DecNum& decnum, UErrorCode& status) {
    if (U_FAILURE(status)) { return; }
    if (decnum.isNegative()) {
        flags |= NEGATIVE_FLAG;
    }
    if (!decnum.isZero()) {
        readDecNumberToBcd(decnum);
        compact();
    }
}

int64_t DecimalQuantity::toLong(bool truncateIfOverflow) const {
    // NOTE: Call sites should be guarded by fitsInLong(), like this:
    // if (dq.fitsInLong()) { /* use dq.toLong() */ } else { /* use some fallback */ }
    // Fallback behavior upon truncateIfOverflow is to truncate at 17 digits.
    uint64_t result = 0L;
    int32_t upperMagnitude = scale + precision - 1;
    if (truncateIfOverflow) {
        upperMagnitude = std::min(upperMagnitude, 17);
    }
    for (int32_t magnitude = upperMagnitude; magnitude >= 0; magnitude--) {
        result = result * 10 + getDigitPos(magnitude - scale);
    }
    if (isNegative()) {
        return static_cast<int64_t>(0LL - result); // i.e., -result
    }
    return static_cast<int64_t>(result);
}

uint64_t DecimalQuantity::toFractionLong(bool includeTrailingZeros) const {
    uint64_t result = 0L;
    int32_t magnitude = -1;
    int32_t lowerMagnitude = scale;
    if (includeTrailingZeros) {
        lowerMagnitude = std::min(lowerMagnitude, rReqPos);
    }
    for (; magnitude >= lowerMagnitude && result <= 1e18L; magnitude--) {
        result = result * 10 + getDigitPos(magnitude - scale);
    }
    // Remove trailing zeros; this can happen during integer overflow cases.
    if (!includeTrailingZeros) {
        while (result > 0 && (result % 10) == 0) {
            result /= 10;
        }
    }
    return result;
}

bool DecimalQuantity::fitsInLong(bool ignoreFraction) const {
    if (isInfinite() || isNaN()) {
        return false;
    }
    if (isZeroish()) {
        return true;
    }
    if (scale < 0 && !ignoreFraction) {
        return false;
    }
    int magnitude = getMagnitude();
    if (magnitude < 18) {
        return true;
    }
    if (magnitude > 18) {
        return false;
    }
    // Hard case: the magnitude is 10^18.
    // The largest int64 is: 9,223,372,036,854,775,807
    for (int p = 0; p < precision; p++) {
        int8_t digit = getDigit(18 - p);
        static int8_t INT64_BCD[] = { 9, 2, 2, 3, 3, 7, 2, 0, 3, 6, 8, 5, 4, 7, 7, 5, 8, 0, 8 };
        if (digit < INT64_BCD[p]) {
            return true;
        } else if (digit > INT64_BCD[p]) {
            return false;
        }
    }
    // Exactly equal to max long plus one.
    return isNegative();
}

double DecimalQuantity::toDouble() const {
    // If this assertion fails, you need to call roundToInfinity() or some other rounding method.
    // See the comment in the header file explaining the "isApproximate" field.
    U_ASSERT(!isApproximate);

    if (isNaN()) {
        return NAN;
    } else if (isInfinite()) {
        return isNegative() ? -INFINITY : INFINITY;
    }

    // We are processing well-formed input, so we don't need any special options to StringToDoubleConverter.
    StringToDoubleConverter converter(0, 0, 0, "", "");
    UnicodeString numberString = this->toScientificString();
    int32_t count;
    return converter.StringToDouble(
            reinterpret_cast<const uint16_t*>(numberString.getBuffer()),
            numberString.length(),
            &count);
}

void DecimalQuantity::toDecNum(DecNum& output, UErrorCode& status) const {
    // Special handling for zero
    if (precision == 0) {
        output.setTo("0", status);
    }

    // Use the BCD constructor. We need to do a little bit of work to convert, though.
    // The decNumber constructor expects most-significant first, but we store least-significant first.
    MaybeStackArray<uint8_t, 20> ubcd(precision);
    for (int32_t m = 0; m < precision; m++) {
        ubcd[precision - m - 1] = static_cast<uint8_t>(getDigitPos(m));
    }
    output.setTo(ubcd.getAlias(), precision, scale, isNegative(), status);
}

void DecimalQuantity::truncate() {
    if (scale < 0) {
        shiftRight(-scale);
        scale = 0;
        compact();
    }
}

void DecimalQuantity::roundToNickel(int32_t magnitude, RoundingMode roundingMode, UErrorCode& status) {
    roundToMagnitude(magnitude, roundingMode, true, status);
}

void DecimalQuantity::roundToMagnitude(int32_t magnitude, RoundingMode roundingMode, UErrorCode& status) {
    roundToMagnitude(magnitude, roundingMode, false, status);
}

void DecimalQuantity::roundToMagnitude(int32_t magnitude, RoundingMode roundingMode, bool nickel, UErrorCode& status) {
    // The position in the BCD at which rounding will be performed; digits to the right of position
    // will be rounded away.
    int position = safeSubtract(magnitude, scale);

    // "trailing" = least significant digit to the left of rounding
    int8_t trailingDigit = getDigitPos(position);

    if (position <= 0 && !isApproximate && (!nickel || trailingDigit == 0 || trailingDigit == 5)) {
        // All digits are to the left of the rounding magnitude.
    } else if (precision == 0) {
        // No rounding for zero.
    } else {
        // Perform rounding logic.
        // "leading" = most significant digit to the right of rounding
        int8_t leadingDigit = getDigitPos(safeSubtract(position, 1));

        // Compute which section of the number we are in.
        // EDGE means we are at the bottom or top edge, like 1.000 or 1.999 (used by doubles)
        // LOWER means we are between the bottom edge and the midpoint, like 1.391
        // MIDPOINT means we are exactly in the middle, like 1.500
        // UPPER means we are between the midpoint and the top edge, like 1.916
        roundingutils::Section section;
        if (!isApproximate) {
            if (nickel && trailingDigit != 2 && trailingDigit != 7) {
                // Nickel rounding, and not at .02x or .07x
                if (trailingDigit < 2) {
                    // .00, .01 => down to .00
                    section = roundingutils::SECTION_LOWER;
                } else if (trailingDigit < 5) {
                    // .03, .04 => up to .05
                    section = roundingutils::SECTION_UPPER;
                } else if (trailingDigit < 7) {
                    // .05, .06 => down to .05
                    section = roundingutils::SECTION_LOWER;
                } else {
                    // .08, .09 => up to .10
                    section = roundingutils::SECTION_UPPER;
                }
            } else if (leadingDigit < 5) {
                // Includes nickel rounding .020-.024 and .070-.074
                section = roundingutils::SECTION_LOWER;
            } else if (leadingDigit > 5) {
                // Includes nickel rounding .026-.029 and .076-.079
                section = roundingutils::SECTION_UPPER;
            } else {
                // Includes nickel rounding .025 and .075
                section = roundingutils::SECTION_MIDPOINT;
                for (int p = safeSubtract(position, 2); p >= 0; p--) {
                    if (getDigitPos(p) != 0) {
                        section = roundingutils::SECTION_UPPER;
                        break;
                    }
                }
            }
        } else {
            int32_t p = safeSubtract(position, 2);
            int32_t minP = uprv_max(0, precision - 14);
            if (leadingDigit == 0 && (!nickel || trailingDigit == 0 || trailingDigit == 5)) {
                section = roundingutils::SECTION_LOWER_EDGE;
                for (; p >= minP; p--) {
                    if (getDigitPos(p) != 0) {
                        section = roundingutils::SECTION_LOWER;
                        break;
                    }
                }
            } else if (leadingDigit == 4 && (!nickel || trailingDigit == 2 || trailingDigit == 7)) {
                section = roundingutils::SECTION_MIDPOINT;
                for (; p >= minP; p--) {
                    if (getDigitPos(p) != 9) {
                        section = roundingutils::SECTION_LOWER;
                        break;
                    }
                }
            } else if (leadingDigit == 5 && (!nickel || trailingDigit == 2 || trailingDigit == 7)) {
                section = roundingutils::SECTION_MIDPOINT;
                for (; p >= minP; p--) {
                    if (getDigitPos(p) != 0) {
                        section = roundingutils::SECTION_UPPER;
                        break;
                    }
                }
            } else if (leadingDigit == 9 && (!nickel || trailingDigit == 4 || trailingDigit == 9)) {
                section = roundingutils::SECTION_UPPER_EDGE;
                for (; p >= minP; p--) {
                    if (getDigitPos(p) != 9) {
                        section = roundingutils::SECTION_UPPER;
                        break;
                    }
                }
            } else if (nickel && trailingDigit != 2 && trailingDigit != 7) {
                // Nickel rounding, and not at .02x or .07x
                if (trailingDigit < 2) {
                    // .00, .01 => down to .00
                    section = roundingutils::SECTION_LOWER;
                } else if (trailingDigit < 5) {
                    // .03, .04 => up to .05
                    section = roundingutils::SECTION_UPPER;
                } else if (trailingDigit < 7) {
                    // .05, .06 => down to .05
                    section = roundingutils::SECTION_LOWER;
                } else {
                    // .08, .09 => up to .10
                    section = roundingutils::SECTION_UPPER;
                }
            } else if (leadingDigit < 5) {
                // Includes nickel rounding .020-.024 and .070-.074
                section = roundingutils::SECTION_LOWER;
            } else {
                // Includes nickel rounding .026-.029 and .076-.079
                section = roundingutils::SECTION_UPPER;
            }

            bool roundsAtMidpoint = roundingutils::roundsAtMidpoint(roundingMode);
            if (safeSubtract(position, 1) < precision - 14 ||
                (roundsAtMidpoint && section == roundingutils::SECTION_MIDPOINT) ||
                (!roundsAtMidpoint && section < 0 /* i.e. at upper or lower edge */)) {
                // Oops! This means that we have to get the exact representation of the double,
                // because the zone of uncertainty is along the rounding boundary.
                convertToAccurateDouble();
                roundToMagnitude(magnitude, roundingMode, nickel, status); // start over
                return;
            }

            // Turn off the approximate double flag, since the value is now confirmed to be exact.
            isApproximate = false;
            origDouble = 0.0;
            origDelta = 0;

            if (position <= 0 && (!nickel || trailingDigit == 0 || trailingDigit == 5)) {
                // All digits are to the left of the rounding magnitude.
                return;
            }

            // Good to continue rounding.
            if (section == -1) { section = roundingutils::SECTION_LOWER; }
            if (section == -2) { section = roundingutils::SECTION_UPPER; }
        }

        // Nickel rounding "half even" goes to the nearest whole (away from the 5).
        bool isEven = nickel
                ? (trailingDigit < 2 || trailingDigit > 7
                        || (trailingDigit == 2 && section != roundingutils::SECTION_UPPER)
                        || (trailingDigit == 7 && section == roundingutils::SECTION_UPPER))
                : (trailingDigit % 2) == 0;

        bool roundDown = roundingutils::getRoundingDirection(isEven,
                isNegative(),
                section,
                roundingMode,
                status);
        if (U_FAILURE(status)) {
            return;
        }

        // Perform truncation
        if (position >= precision) {
            setBcdToZero();
            scale = magnitude;
        } else {
            shiftRight(position);
        }

        if (nickel) {
            if (trailingDigit < 5 && roundDown) {
                setDigitPos(0, 0);
                compact();
                return;
            } else if (trailingDigit >= 5 && !roundDown) {
                setDigitPos(0, 9);
                trailingDigit = 9;
                // do not return: use the bubbling logic below
            } else {
                setDigitPos(0, 5);
                // compact not necessary: digit at position 0 is nonzero
                return;
            }
        }

        // Bubble the result to the higher digits
        if (!roundDown) {
            if (trailingDigit == 9) {
                int bubblePos = 0;
                // Note: in the long implementation, the most digits BCD can have at this point is
                // 15, so bubblePos <= 15 and getDigitPos(bubblePos) is safe.
                for (; getDigitPos(bubblePos) == 9; bubblePos++) {}
                shiftRight(bubblePos); // shift off the trailing 9s
            }
            int8_t digit0 = getDigitPos(0);
            U_ASSERT(digit0 != 9);
            setDigitPos(0, static_cast<int8_t>(digit0 + 1));
            precision += 1; // in case an extra digit got added
        }

        compact();
    }
}

void DecimalQuantity::roundToInfinity() {
    if (isApproximate) {
        convertToAccurateDouble();
    }
}

void DecimalQuantity::appendDigit(int8_t value, int32_t leadingZeros, bool appendAsInteger) {
    U_ASSERT(leadingZeros >= 0);

    // Zero requires special handling to maintain the invariant that the least-significant digit
    // in the BCD is nonzero.
    if (value == 0) {
        if (appendAsInteger && precision != 0) {
            scale += leadingZeros + 1;
        }
        return;
    }

    // Deal with trailing zeros
    if (scale > 0) {
        leadingZeros += scale;
        if (appendAsInteger) {
            scale = 0;
        }
    }

    // Append digit
    shiftLeft(leadingZeros + 1);
    setDigitPos(0, value);

    // Fix scale if in integer mode
    if (appendAsInteger) {
        scale += leadingZeros + 1;
    }
}

UnicodeString DecimalQuantity::toPlainString() const {
    U_ASSERT(!isApproximate);
    UnicodeString sb;
    if (isNegative()) {
        sb.append(u'-');
    }
    if (precision == 0 || getMagnitude() < 0) {
        sb.append(u'0');
    }
    for (int m = getUpperDisplayMagnitude(); m >= getLowerDisplayMagnitude(); m--) {
        if (m == -1) { sb.append(u'.'); }
        sb.append(getDigit(m) + u'0');
    }
    return sb;
}

UnicodeString DecimalQuantity::toScientificString() const {
    U_ASSERT(!isApproximate);
    UnicodeString result;
    if (isNegative()) {
        result.append(u'-');
    }
    if (precision == 0) {
        result.append(u"0E+0", -1);
        return result;
    }
    int32_t upperPos = precision - 1;
    int32_t lowerPos = 0;
    int32_t p = upperPos;
    result.append(u'0' + getDigitPos(p));
    if ((--p) >= lowerPos) {
        result.append(u'.');
        for (; p >= lowerPos; p--) {
            result.append(u'0' + getDigitPos(p));
        }
    }
    result.append(u'E');
    int32_t _scale = upperPos + scale;
    if (_scale == INT32_MIN) {
        result.append({u"-2147483648", -1});
        return result;
    } else if (_scale < 0) {
        _scale *= -1;
        result.append(u'-');
    } else {
        result.append(u'+');
    }
    if (_scale == 0) {
        result.append(u'0');
    }
    int32_t insertIndex = result.length();
    while (_scale > 0) {
        std::div_t res = std::div(_scale, 10);
        result.insert(insertIndex, u'0' + res.rem);
        _scale = res.quot;
    }
    return result;
}

////////////////////////////////////////////////////
/// End of DecimalQuantity_AbstractBCD.java      ///
/// Start of DecimalQuantity_DualStorageBCD.java ///
////////////////////////////////////////////////////

int8_t DecimalQuantity::getDigitPos(int32_t position) const {
    if (usingBytes) {
        if (position < 0 || position >= precision) { return 0; }
        return fBCD.bcdBytes.ptr[position];
    } else {
        if (position < 0 || position >= 16) { return 0; }
        return (int8_t) ((fBCD.bcdLong >> (position * 4)) & 0xf);
    }
}

void DecimalQuantity::setDigitPos(int32_t position, int8_t value) {
    U_ASSERT(position >= 0);
    if (usingBytes) {
        ensureCapacity(position + 1);
        fBCD.bcdBytes.ptr[position] = value;
    } else if (position >= 16) {
        switchStorage();
        ensureCapacity(position + 1);
        fBCD.bcdBytes.ptr[position] = value;
    } else {
        int shift = position * 4;
        fBCD.bcdLong = (fBCD.bcdLong & ~(0xfL << shift)) | ((long) value << shift);
    }
}

void DecimalQuantity::shiftLeft(int32_t numDigits) {
    if (!usingBytes && precision + numDigits > 16) {
        switchStorage();
    }
    if (usingBytes) {
        ensureCapacity(precision + numDigits);
        int i = precision + numDigits - 1;
        for (; i >= numDigits; i--) {
            fBCD.bcdBytes.ptr[i] = fBCD.bcdBytes.ptr[i - numDigits];
        }
        for (; i >= 0; i--) {
            fBCD.bcdBytes.ptr[i] = 0;
        }
    } else {
        fBCD.bcdLong <<= (numDigits * 4);
    }
    scale -= numDigits;
    precision += numDigits;
}

void DecimalQuantity::shiftRight(int32_t numDigits) {
    if (usingBytes) {
        int i = 0;
        for (; i < precision - numDigits; i++) {
            fBCD.bcdBytes.ptr[i] = fBCD.bcdBytes.ptr[i + numDigits];
        }
        for (; i < precision; i++) {
            fBCD.bcdBytes.ptr[i] = 0;
        }
    } else {
        fBCD.bcdLong >>= (numDigits * 4);
    }
    scale += numDigits;
    precision -= numDigits;
}

void DecimalQuantity::popFromLeft(int32_t numDigits) {
    U_ASSERT(numDigits <= precision);
    if (usingBytes) {
        int i = precision - 1;
        for (; i >= precision - numDigits; i--) {
            fBCD.bcdBytes.ptr[i] = 0;
        }
    } else {
        fBCD.bcdLong &= (static_cast<uint64_t>(1) << ((precision - numDigits) * 4)) - 1;
    }
    precision -= numDigits;
}

void DecimalQuantity::setBcdToZero() {
    if (usingBytes) {
        uprv_free(fBCD.bcdBytes.ptr);
        fBCD.bcdBytes.ptr = nullptr;
        usingBytes = false;
    }
    fBCD.bcdLong = 0L;
    scale = 0;
    precision = 0;
    isApproximate = false;
    origDouble = 0;
    origDelta = 0;
}

void DecimalQuantity::readIntToBcd(int32_t n) {
    U_ASSERT(n != 0);
    // ints always fit inside the long implementation.
    uint64_t result = 0L;
    int i = 16;
    for (; n != 0; n /= 10, i--) {
        result = (result >> 4) + ((static_cast<uint64_t>(n) % 10) << 60);
    }
    U_ASSERT(!usingBytes);
    fBCD.bcdLong = result >> (i * 4);
    scale = 0;
    precision = 16 - i;
}

void DecimalQuantity::readLongToBcd(int64_t n) {
    U_ASSERT(n != 0);
    if (n >= 10000000000000000L) {
        ensureCapacity();
        int i = 0;
        for (; n != 0L; n /= 10L, i++) {
            fBCD.bcdBytes.ptr[i] = static_cast<int8_t>(n % 10);
        }
        U_ASSERT(usingBytes);
        scale = 0;
        precision = i;
    } else {
        uint64_t result = 0L;
        int i = 16;
        for (; n != 0L; n /= 10L, i--) {
            result = (result >> 4) + ((n % 10) << 60);
        }
        U_ASSERT(i >= 0);
        U_ASSERT(!usingBytes);
        fBCD.bcdLong = result >> (i * 4);
        scale = 0;
        precision = 16 - i;
    }
}

void DecimalQuantity::readDecNumberToBcd(const DecNum& decnum) {
    const decNumber* dn = decnum.getRawDecNumber();
    if (dn->digits > 16) {
        ensureCapacity(dn->digits);
        for (int32_t i = 0; i < dn->digits; i++) {
            fBCD.bcdBytes.ptr[i] = dn->lsu[i];
        }
    } else {
        uint64_t result = 0L;
        for (int32_t i = 0; i < dn->digits; i++) {
            result |= static_cast<uint64_t>(dn->lsu[i]) << (4 * i);
        }
        fBCD.bcdLong = result;
    }
    scale = dn->exponent;
    precision = dn->digits;
}

void DecimalQuantity::readDoubleConversionToBcd(
        const char* buffer, int32_t length, int32_t point) {
    // NOTE: Despite the fact that double-conversion's API is called
    // "DoubleToAscii", they actually use '0' (as opposed to u8'0').
    if (length > 16) {
        ensureCapacity(length);
        for (int32_t i = 0; i < length; i++) {
            fBCD.bcdBytes.ptr[i] = buffer[length-i-1] - '0';
        }
    } else {
        uint64_t result = 0L;
        for (int32_t i = 0; i < length; i++) {
            result |= static_cast<uint64_t>(buffer[length-i-1] - '0') << (4 * i);
        }
        fBCD.bcdLong = result;
    }
    scale = point - length;
    precision = length;
}

void DecimalQuantity::compact() {
    if (usingBytes) {
        int32_t delta = 0;
        for (; delta < precision && fBCD.bcdBytes.ptr[delta] == 0; delta++);
        if (delta == precision) {
            // Number is zero
            setBcdToZero();
            return;
        } else {
            // Remove trailing zeros
            shiftRight(delta);
        }

        // Compute precision
        int32_t leading = precision - 1;
        for (; leading >= 0 && fBCD.bcdBytes.ptr[leading] == 0; leading--);
        precision = leading + 1;

        // Switch storage mechanism if possible
        if (precision <= 16) {
            switchStorage();
        }

    } else {
        if (fBCD.bcdLong == 0L) {
            // Number is zero
            setBcdToZero();
            return;
        }

        // Compact the number (remove trailing zeros)
        // TODO: Use a more efficient algorithm here and below. There is a logarithmic one.
        int32_t delta = 0;
        for (; delta < precision && getDigitPos(delta) == 0; delta++);
        fBCD.bcdLong >>= delta * 4;
        scale += delta;

        // Compute precision
        int32_t leading = precision - 1;
        for (; leading >= 0 && getDigitPos(leading) == 0; leading--);
        precision = leading + 1;
    }
}

void DecimalQuantity::ensureCapacity() {
    ensureCapacity(40);
}

void DecimalQuantity::ensureCapacity(int32_t capacity) {
    if (capacity == 0) { return; }
    int32_t oldCapacity = usingBytes ? fBCD.bcdBytes.len : 0;
    if (!usingBytes) {
        // TODO: There is nothing being done to check for memory allocation failures.
        // TODO: Consider indexing by nybbles instead of bytes in C++, so that we can
        // make these arrays half the size.
        fBCD.bcdBytes.ptr = static_cast<int8_t*>(uprv_malloc(capacity * sizeof(int8_t)));
        fBCD.bcdBytes.len = capacity;
        // Initialize the byte array to zeros (this is done automatically in Java)
        uprv_memset(fBCD.bcdBytes.ptr, 0, capacity * sizeof(int8_t));
    } else if (oldCapacity < capacity) {
        auto bcd1 = static_cast<int8_t*>(uprv_malloc(capacity * 2 * sizeof(int8_t)));
        uprv_memcpy(bcd1, fBCD.bcdBytes.ptr, oldCapacity * sizeof(int8_t));
        // Initialize the rest of the byte array to zeros (this is done automatically in Java)
        uprv_memset(bcd1 + oldCapacity, 0, (capacity - oldCapacity) * sizeof(int8_t));
        uprv_free(fBCD.bcdBytes.ptr);
        fBCD.bcdBytes.ptr = bcd1;
        fBCD.bcdBytes.len = capacity * 2;
    }
    usingBytes = true;
}

void DecimalQuantity::switchStorage() {
    if (usingBytes) {
        // Change from bytes to long
        uint64_t bcdLong = 0L;
        for (int i = precision - 1; i >= 0; i--) {
            bcdLong <<= 4;
            bcdLong |= fBCD.bcdBytes.ptr[i];
        }
        uprv_free(fBCD.bcdBytes.ptr);
        fBCD.bcdBytes.ptr = nullptr;
        fBCD.bcdLong = bcdLong;
        usingBytes = false;
    } else {
        // Change from long to bytes
        // Copy the long into a local variable since it will get munged when we allocate the bytes
        uint64_t bcdLong = fBCD.bcdLong;
        ensureCapacity();
        for (int i = 0; i < precision; i++) {
            fBCD.bcdBytes.ptr[i] = static_cast<int8_t>(bcdLong & 0xf);
            bcdLong >>= 4;
        }
        U_ASSERT(usingBytes);
    }
}

void DecimalQuantity::copyBcdFrom(const DecimalQuantity &other) {
    setBcdToZero();
    if (other.usingBytes) {
        ensureCapacity(other.precision);
        uprv_memcpy(fBCD.bcdBytes.ptr, other.fBCD.bcdBytes.ptr, other.precision * sizeof(int8_t));
    } else {
        fBCD.bcdLong = other.fBCD.bcdLong;
    }
}

void DecimalQuantity::moveBcdFrom(DecimalQuantity &other) {
    setBcdToZero();
    if (other.usingBytes) {
        usingBytes = true;
        fBCD.bcdBytes.ptr = other.fBCD.bcdBytes.ptr;
        fBCD.bcdBytes.len = other.fBCD.bcdBytes.len;
        // Take ownership away from the old instance:
        other.fBCD.bcdBytes.ptr = nullptr;
        other.usingBytes = false;
    } else {
        fBCD.bcdLong = other.fBCD.bcdLong;
    }
}

const char16_t* DecimalQuantity::checkHealth() const {
    if (usingBytes) {
        if (precision == 0) { return u"Zero precision but we are in byte mode"; }
        int32_t capacity = fBCD.bcdBytes.len;
        if (precision > capacity) { return u"Precision exceeds length of byte array"; }
        if (getDigitPos(precision - 1) == 0) { return u"Most significant digit is zero in byte mode"; }
        if (getDigitPos(0) == 0) { return u"Least significant digit is zero in long mode"; }
        for (int i = 0; i < precision; i++) {
            if (getDigitPos(i) >= 10) { return u"Digit exceeding 10 in byte array"; }
            if (getDigitPos(i) < 0) { return u"Digit below 0 in byte array"; }
        }
        for (int i = precision; i < capacity; i++) {
            if (getDigitPos(i) != 0) { return u"Nonzero digits outside of range in byte array"; }
        }
    } else {
        if (precision == 0 && fBCD.bcdLong != 0) {
            return u"Value in bcdLong even though precision is zero";
        }
        if (precision > 16) { return u"Precision exceeds length of long"; }
        if (precision != 0 && getDigitPos(precision - 1) == 0) {
            return u"Most significant digit is zero in long mode";
        }
        if (precision != 0 && getDigitPos(0) == 0) {
            return u"Least significant digit is zero in long mode";
        }
        for (int i = 0; i < precision; i++) {
            if (getDigitPos(i) >= 10) { return u"Digit exceeding 10 in long"; }
            if (getDigitPos(i) < 0) { return u"Digit below 0 in long (?!)"; }
        }
        for (int i = precision; i < 16; i++) {
            if (getDigitPos(i) != 0) { return u"Nonzero digits outside of range in long"; }
        }
    }

    // No error
    return nullptr;
}

bool DecimalQuantity::operator==(const DecimalQuantity& other) const {
    bool basicEquals =
            scale == other.scale
            && precision == other.precision
            && flags == other.flags
            && lReqPos == other.lReqPos
            && rReqPos == other.rReqPos
            && isApproximate == other.isApproximate;
    if (!basicEquals) {
        return false;
    }

    if (precision == 0) {
        return true;
    } else if (isApproximate) {
        return origDouble == other.origDouble && origDelta == other.origDelta;
    } else {
        for (int m = getUpperDisplayMagnitude(); m >= getLowerDisplayMagnitude(); m--) {
            if (getDigit(m) != other.getDigit(m)) {
                return false;
            }
        }
        return true;
    }
}

UnicodeString DecimalQuantity::toString() const {
    MaybeStackArray<char, 30> digits(precision + 1);
    for (int32_t i = 0; i < precision; i++) {
        digits[i] = getDigitPos(precision - i - 1) + '0';
    }
    digits[precision] = 0; // terminate buffer
    char buffer8[100];
    snprintf(
            buffer8,
            sizeof(buffer8),
            "<DecimalQuantity %d:%d %s %s%s%s%d>",
            lReqPos,
            rReqPos,
            (usingBytes ? "bytes" : "long"),
            (isNegative() ? "-" : ""),
            (precision == 0 ? "0" : digits.getAlias()),
            "E",
            scale);
    return UnicodeString(buffer8, -1, US_INV);
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING




using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;


namespace {

alignas(DecimalFormatProperties)
char kRawDefaultProperties[sizeof(DecimalFormatProperties)];

icu::UInitOnce gDefaultPropertiesInitOnce = U_INITONCE_INITIALIZER;

void U_CALLCONV initDefaultProperties(UErrorCode&) {
    // can't fail, uses placement new into staticly allocated space.
    new(kRawDefaultProperties) DecimalFormatProperties(); // set to the default instance
}

}


DecimalFormatProperties::DecimalFormatProperties() {
    clear();
}

void DecimalFormatProperties::clear() {
    compactStyle.nullify();
    currency.nullify();
    currencyPluralInfo.fPtr.adoptInstead(nullptr);
    currencyUsage.nullify();
    decimalPatternMatchRequired = false;
    decimalSeparatorAlwaysShown = false;
    exponentSignAlwaysShown = false;
    formatFailIfMoreThanMaxDigits = false;
    formatWidth = -1;
    groupingSize = -1;
    groupingUsed = true;
    magnitudeMultiplier = 0;
    maximumFractionDigits = -1;
    maximumIntegerDigits = -1;
    maximumSignificantDigits = -1;
    minimumExponentDigits = -1;
    minimumFractionDigits = -1;
    minimumGroupingDigits = -1;
    minimumIntegerDigits = -1;
    minimumSignificantDigits = -1;
    multiplier = 1;
    multiplierScale = 0;
    negativePrefix.setToBogus();
    negativePrefixPattern.setToBogus();
    negativeSuffix.setToBogus();
    negativeSuffixPattern.setToBogus();
    padPosition.nullify();
    padString.setToBogus();
    parseCaseSensitive = false;
    parseIntegerOnly = false;
    parseMode.nullify();
    parseNoExponent = false;
    parseToBigDecimal = false;
    parseAllInput = UNUM_MAYBE;
    positivePrefix.setToBogus();
    positivePrefixPattern.setToBogus();
    positiveSuffix.setToBogus();
    positiveSuffixPattern.setToBogus();
    roundingIncrement = 0.0;
    roundingMode.nullify();
    secondaryGroupingSize = -1;
    signAlwaysShown = false;
}

bool
DecimalFormatProperties::_equals(const DecimalFormatProperties& other, bool ignoreForFastFormat) const {
    bool eq = true;

    // Properties that must be equal both normally and for fast-path formatting
    eq = eq && compactStyle == other.compactStyle;
    eq = eq && currency == other.currency;
    eq = eq && currencyPluralInfo.fPtr.getAlias() == other.currencyPluralInfo.fPtr.getAlias();
    eq = eq && currencyUsage == other.currencyUsage;
    eq = eq && decimalSeparatorAlwaysShown == other.decimalSeparatorAlwaysShown;
    eq = eq && exponentSignAlwaysShown == other.exponentSignAlwaysShown;
    eq = eq && formatFailIfMoreThanMaxDigits == other.formatFailIfMoreThanMaxDigits;
    eq = eq && formatWidth == other.formatWidth;
    eq = eq && magnitudeMultiplier == other.magnitudeMultiplier;
    eq = eq && maximumSignificantDigits == other.maximumSignificantDigits;
    eq = eq && minimumExponentDigits == other.minimumExponentDigits;
    eq = eq && minimumGroupingDigits == other.minimumGroupingDigits;
    eq = eq && minimumSignificantDigits == other.minimumSignificantDigits;
    eq = eq && multiplier == other.multiplier;
    eq = eq && multiplierScale == other.multiplierScale;
    eq = eq && negativePrefix == other.negativePrefix;
    eq = eq && negativeSuffix == other.negativeSuffix;
    eq = eq && padPosition == other.padPosition;
    eq = eq && padString == other.padString;
    eq = eq && positivePrefix == other.positivePrefix;
    eq = eq && positiveSuffix == other.positiveSuffix;
    eq = eq && roundingIncrement == other.roundingIncrement;
    eq = eq && roundingMode == other.roundingMode;
    eq = eq && secondaryGroupingSize == other.secondaryGroupingSize;
    eq = eq && signAlwaysShown == other.signAlwaysShown;

    if (ignoreForFastFormat) {
        return eq;
    }

    // Properties ignored by fast-path formatting
    // Formatting (special handling required):
    eq = eq && groupingSize == other.groupingSize;
    eq = eq && groupingUsed == other.groupingUsed;
    eq = eq && minimumFractionDigits == other.minimumFractionDigits;
    eq = eq && maximumFractionDigits == other.maximumFractionDigits;
    eq = eq && maximumIntegerDigits == other.maximumIntegerDigits;
    eq = eq && minimumIntegerDigits == other.minimumIntegerDigits;
    eq = eq && negativePrefixPattern == other.negativePrefixPattern;
    eq = eq && negativeSuffixPattern == other.negativeSuffixPattern;
    eq = eq && positivePrefixPattern == other.positivePrefixPattern;
    eq = eq && positiveSuffixPattern == other.positiveSuffixPattern;

    // Parsing (always safe to ignore):
    eq = eq && decimalPatternMatchRequired == other.decimalPatternMatchRequired;
    eq = eq && parseCaseSensitive == other.parseCaseSensitive;
    eq = eq && parseIntegerOnly == other.parseIntegerOnly;
    eq = eq && parseMode == other.parseMode;
    eq = eq && parseNoExponent == other.parseNoExponent;
    eq = eq && parseToBigDecimal == other.parseToBigDecimal;
    eq = eq && parseAllInput == other.parseAllInput;

    return eq;
}

bool DecimalFormatProperties::equalsDefaultExceptFastFormat() const {
    UErrorCode localStatus = U_ZERO_ERROR;
    umtx_initOnce(gDefaultPropertiesInitOnce, &initDefaultProperties, localStatus);
    return _equals(*reinterpret_cast<DecimalFormatProperties*>(kRawDefaultProperties), true);
}

const DecimalFormatProperties& DecimalFormatProperties::getDefault() {
    UErrorCode localStatus = U_ZERO_ERROR;
    umtx_initOnce(gDefaultPropertiesInitOnce, &initDefaultProperties, localStatus);
    return *reinterpret_cast<const DecimalFormatProperties*>(kRawDefaultProperties);
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING




// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING










U_NAMESPACE_BEGIN namespace number {
namespace impl {

/**
 * This is the "brain" of the number formatting pipeline. It ties all the pieces together, taking in a MacroProps and a
 * DecimalQuantity and outputting a properly formatted number string.
 */
class NumberFormatterImpl : public UMemory {
  public:
    /**
     * Builds a "safe" MicroPropsGenerator, which is thread-safe and can be used repeatedly.
     * The caller owns the returned NumberFormatterImpl.
     */
    NumberFormatterImpl(const MacroProps &macros, UErrorCode &status);

    /**
     * Builds and evaluates an "unsafe" MicroPropsGenerator, which is cheaper but can be used only once.
     */
    static int32_t
    formatStatic(const MacroProps &macros, DecimalQuantity &inValue, FormattedStringBuilder &outString,
                 UErrorCode &status);

    /**
     * Prints only the prefix and suffix; used for DecimalFormat getters.
     *
     * @return The index into the output at which the prefix ends and the suffix starts; in other words,
     *         the prefix length.
     */
    static int32_t getPrefixSuffixStatic(const MacroProps& macros, Signum signum,
                                         StandardPlural::Form plural, FormattedStringBuilder& outString,
                                         UErrorCode& status);

    /**
     * Evaluates the "safe" MicroPropsGenerator created by "fromMacros".
     */
    int32_t format(DecimalQuantity& inValue, FormattedStringBuilder& outString, UErrorCode& status) const;

    /**
     * Like format(), but saves the result into an output MicroProps without additional processing.
     */
    void preProcess(DecimalQuantity& inValue, MicroProps& microsOut, UErrorCode& status) const;

    /**
     * Like getPrefixSuffixStatic() but uses the safe compiled object.
     */
    int32_t getPrefixSuffix(Signum signum, StandardPlural::Form plural, FormattedStringBuilder& outString,
                            UErrorCode& status) const;

    const MicroProps& getRawMicroProps() const {
        return fMicros;
    }

    /**
     * Synthesizes the output string from a MicroProps and DecimalQuantity.
     * This method formats only the main number, not affixes.
     */
    static int32_t writeNumber(const MicroProps& micros, DecimalQuantity& quantity,
                               FormattedStringBuilder& string, int32_t index, UErrorCode& status);

    /**
     * Adds the affixes.  Intended to be called immediately after formatNumber.
     */
    static int32_t writeAffixes(const MicroProps& micros, FormattedStringBuilder& string, int32_t start,
                                int32_t end, UErrorCode& status);

  private:
    // Head of the MicroPropsGenerator linked list:
    const MicroPropsGenerator *fMicroPropsGenerator = nullptr;

    // Tail of the list:
    MicroProps fMicros;

    // Other fields possibly used by the number formatting pipeline:
    // TODO: Convert more of these LocalPointers to value objects to reduce the number of news?
    LocalPointer<const DecimalFormatSymbols> fSymbols;
    LocalPointer<const PluralRules> fRules;
    LocalPointer<const ParsedPatternInfo> fPatternInfo;
    LocalPointer<const ScientificHandler> fScientificHandler;
    LocalPointer<MutablePatternModifier> fPatternModifier;
    LocalPointer<const ImmutablePatternModifier> fImmutablePatternModifier;
    LocalPointer<const LongNameHandler> fLongNameHandler;
    LocalPointer<const CompactHandler> fCompactHandler;

    // Value objects possibly used by the number formatting pipeline:
    struct Warehouse {
        CurrencySymbols fCurrencySymbols;
    } fWarehouse;


    NumberFormatterImpl(const MacroProps &macros, bool safe, UErrorCode &status);

    MicroProps& preProcessUnsafe(DecimalQuantity &inValue, UErrorCode &status);

    int32_t getPrefixSuffixUnsafe(Signum signum, StandardPlural::Form plural,
                                  FormattedStringBuilder& outString, UErrorCode& status);

    /**
     * If rulesPtr is non-null, return it.  Otherwise, return a PluralRules owned by this object for the
     * specified locale, creating it if necessary.
     */
    const PluralRules *
    resolvePluralRules(const PluralRules *rulesPtr, const Locale &locale, UErrorCode &status);

    /**
     * Synthesizes the MacroProps into a MicroPropsGenerator. All information, including the locale, is encoded into the
     * MicroPropsGenerator, except for the quantity itself, which is left abstract and must be provided to the returned
     * MicroPropsGenerator instance.
     *
     * @see MicroPropsGenerator
     * @param macros
     *            The {@link MacroProps} to consume. This method does not mutate the MacroProps instance.
     * @param safe
     *            If true, the returned MicroPropsGenerator will be thread-safe. If false, the returned value will
     *            <em>not</em> be thread-safe, intended for a single "one-shot" use only. Building the thread-safe
     *            object is more expensive.
     */
    const MicroPropsGenerator *
    macrosToMicroGenerator(const MacroProps &macros, bool safe, UErrorCode &status);

    static int32_t
    writeIntegerDigits(const MicroProps &micros, DecimalQuantity &quantity, FormattedStringBuilder &string,
                       int32_t index, UErrorCode &status);

    static int32_t
    writeFractionDigits(const MicroProps &micros, DecimalQuantity &quantity, FormattedStringBuilder &string,
                        int32_t index, UErrorCode &status);
};

}  // namespace impl
}  // namespace number
U_NAMESPACE_END


#endif //__NUMBER_FORMATIMPL_H__




// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING






U_NAMESPACE_BEGIN
namespace number {
namespace impl {

// Forward-declaration
struct SeenMacroProps;

// namespace for enums and entrypoint functions
namespace skeleton {

///////////////////////////////////////////////////////////////////////////////////////
// NOTE: For an example of how to add a new stem to the number skeleton parser, see: //
// http://bugs.icu-project.org/trac/changeset/41193                                  //
///////////////////////////////////////////////////////////////////////////////////////

/**
 * While parsing a skeleton, this enum records what type of option we expect to find next.
 */
enum ParseState {

    // Section 0: We expect whitespace or a stem, but not an option:

    STATE_NULL,

    // Section 1: We might accept an option, but it is not required:

    STATE_SCIENTIFIC,
    STATE_FRACTION_PRECISION,

    // Section 2: An option is required:

    STATE_INCREMENT_PRECISION,
    STATE_MEASURE_UNIT,
    STATE_PER_MEASURE_UNIT,
    STATE_CURRENCY_UNIT,
    STATE_INTEGER_WIDTH,
    STATE_NUMBERING_SYSTEM,
    STATE_SCALE,
};

/**
 * All possible stem literals have an entry in the StemEnum. The enum name is the kebab case stem
 * string literal written in upper snake case.
 *
 * @see StemToObject
 * @see #SERIALIZED_STEM_TRIE
 */
enum StemEnum {

    // Section 1: Stems that do not require an option:

    STEM_COMPACT_SHORT,
    STEM_COMPACT_LONG,
    STEM_SCIENTIFIC,
    STEM_ENGINEERING,
    STEM_NOTATION_SIMPLE,
    STEM_BASE_UNIT,
    STEM_PERCENT,
    STEM_PERMILLE,
    STEM_PRECISION_INTEGER,
    STEM_PRECISION_UNLIMITED,
    STEM_PRECISION_CURRENCY_STANDARD,
    STEM_PRECISION_CURRENCY_CASH,
    STEM_ROUNDING_MODE_CEILING,
    STEM_ROUNDING_MODE_FLOOR,
    STEM_ROUNDING_MODE_DOWN,
    STEM_ROUNDING_MODE_UP,
    STEM_ROUNDING_MODE_HALF_EVEN,
    STEM_ROUNDING_MODE_HALF_DOWN,
    STEM_ROUNDING_MODE_HALF_UP,
    STEM_ROUNDING_MODE_UNNECESSARY,
    STEM_GROUP_OFF,
    STEM_GROUP_MIN2,
    STEM_GROUP_AUTO,
    STEM_GROUP_ON_ALIGNED,
    STEM_GROUP_THOUSANDS,
    STEM_LATIN,
    STEM_UNIT_WIDTH_NARROW,
    STEM_UNIT_WIDTH_SHORT,
    STEM_UNIT_WIDTH_FULL_NAME,
    STEM_UNIT_WIDTH_ISO_CODE,
    STEM_UNIT_WIDTH_HIDDEN,
    STEM_SIGN_AUTO,
    STEM_SIGN_ALWAYS,
    STEM_SIGN_NEVER,
    STEM_SIGN_ACCOUNTING,
    STEM_SIGN_ACCOUNTING_ALWAYS,
    STEM_SIGN_EXCEPT_ZERO,
    STEM_SIGN_ACCOUNTING_EXCEPT_ZERO,
    STEM_DECIMAL_AUTO,
    STEM_DECIMAL_ALWAYS,

    // Section 2: Stems that DO require an option:

    STEM_PRECISION_INCREMENT,
    STEM_MEASURE_UNIT,
    STEM_PER_MEASURE_UNIT,
    STEM_CURRENCY,
    STEM_INTEGER_WIDTH,
    STEM_NUMBERING_SYSTEM,
    STEM_SCALE,
};

/**
 * Creates a NumberFormatter corresponding to the given skeleton string.
 *
 * @param skeletonString
 *            A number skeleton string, possibly not in its shortest form.
 * @return An UnlocalizedNumberFormatter with behavior defined by the given skeleton string.
 */
UnlocalizedNumberFormatter create(
    const UnicodeString& skeletonString, UParseError* perror, UErrorCode& status);

/**
 * Create a skeleton string corresponding to the given NumberFormatter.
 *
 * @param macros
 *            The NumberFormatter options object.
 * @return A skeleton string in normalized form.
 */
UnicodeString generate(const MacroProps& macros, UErrorCode& status);

/**
 * Converts from a skeleton string to a MacroProps. This method contains the primary parse loop.
 *
 * Internal: use the create() endpoint instead of this function.
 */
MacroProps parseSkeleton(const UnicodeString& skeletonString, int32_t& errOffset, UErrorCode& status);

/**
 * Given that the current segment represents a stem, parse it and save the result.
 *
 * @return The next state after parsing this stem, corresponding to what subset of options to expect.
 */
ParseState parseStem(const StringSegment& segment, const UCharsTrie& stemTrie, SeenMacroProps& seen,
                     MacroProps& macros, UErrorCode& status);

/**
 * Given that the current segment represents an option, parse it and save the result.
 *
 * @return The next state after parsing this option, corresponding to what subset of options to
 *         expect next.
 */
ParseState
parseOption(ParseState stem, const StringSegment& segment, MacroProps& macros, UErrorCode& status);

} // namespace skeleton


/**
 * Namespace for utility methods that convert from StemEnum to corresponding objects or enums. This
 * applies to only the "Section 1" stems, those that are well-defined without an option.
 */
namespace stem_to_object {

Notation notation(skeleton::StemEnum stem);

MeasureUnit unit(skeleton::StemEnum stem);

Precision precision(skeleton::StemEnum stem);

UNumberFormatRoundingMode roundingMode(skeleton::StemEnum stem);

UNumberGroupingStrategy groupingStrategy(skeleton::StemEnum stem);

UNumberUnitWidth unitWidth(skeleton::StemEnum stem);

UNumberSignDisplay signDisplay(skeleton::StemEnum stem);

UNumberDecimalSeparatorDisplay decimalSeparatorDisplay(skeleton::StemEnum stem);

} // namespace stem_to_object

/**
 * Namespace for utility methods that convert from enums to stem strings. More complex object conversions
 * take place in the object_to_stem_string namespace.
 */
namespace enum_to_stem_string {

void roundingMode(UNumberFormatRoundingMode value, UnicodeString& sb);

void groupingStrategy(UNumberGroupingStrategy value, UnicodeString& sb);

void unitWidth(UNumberUnitWidth value, UnicodeString& sb);

void signDisplay(UNumberSignDisplay value, UnicodeString& sb);

void decimalSeparatorDisplay(UNumberDecimalSeparatorDisplay value, UnicodeString& sb);

} // namespace enum_to_stem_string

/**
 * Namespace for utility methods for processing stems and options that cannot be interpreted literally.
 */
namespace blueprint_helpers {

/** @return Whether we successfully found and parsed an exponent width option. */
bool parseExponentWidthOption(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void generateExponentWidthOption(int32_t minExponentDigits, UnicodeString& sb, UErrorCode& status);

/** @return Whether we successfully found and parsed an exponent sign option. */
bool parseExponentSignOption(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void parseCurrencyOption(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void generateCurrencyOption(const CurrencyUnit& currency, UnicodeString& sb, UErrorCode& status);

void parseMeasureUnitOption(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void generateMeasureUnitOption(const MeasureUnit& measureUnit, UnicodeString& sb, UErrorCode& status);

void parseMeasurePerUnitOption(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void parseFractionStem(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void generateFractionStem(int32_t minFrac, int32_t maxFrac, UnicodeString& sb, UErrorCode& status);

void parseDigitsStem(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void generateDigitsStem(int32_t minSig, int32_t maxSig, UnicodeString& sb, UErrorCode& status);

/** @return Whether we successfully found and parsed a frac-sig option. */
bool parseFracSigOption(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void parseIncrementOption(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void
generateIncrementOption(double increment, int32_t trailingZeros, UnicodeString& sb, UErrorCode& status);

void parseIntegerWidthOption(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void generateIntegerWidthOption(int32_t minInt, int32_t maxInt, UnicodeString& sb, UErrorCode& status);

void parseNumberingSystemOption(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void generateNumberingSystemOption(const NumberingSystem& ns, UnicodeString& sb, UErrorCode& status);

void parseScaleOption(const StringSegment& segment, MacroProps& macros, UErrorCode& status);

void generateScaleOption(int32_t magnitude, const DecNum* arbitrary, UnicodeString& sb,
                              UErrorCode& status);

} // namespace blueprint_helpers

/**
 * Class for utility methods for generating a token corresponding to each macro-prop. Each method
 * returns whether or not a token was written to the string builder.
 *
 * This needs to be a class, not a namespace, so it can be friended.
 */
class GeneratorHelpers {
  public:
    /**
     * Main skeleton generator function. Appends the normalized skeleton for the MacroProps to the given
     * StringBuilder.
     *
     * Internal: use the create() endpoint instead of this function.
     */
    static void generateSkeleton(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

  private:
    static bool notation(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool unit(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool perUnit(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool precision(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool roundingMode(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool grouping(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool integerWidth(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool symbols(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool unitWidth(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool sign(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool decimal(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

    static bool scale(const MacroProps& macros, UnicodeString& sb, UErrorCode& status);

};

/**
 * Struct for null-checking.
 * In Java, we can just check the object reference. In C++, we need a different method.
 */
struct SeenMacroProps {
    bool notation = false;
    bool unit = false;
    bool perUnit = false;
    bool precision = false;
    bool roundingMode = false;
    bool grouper = false;
    bool padder = false;
    bool integerWidth = false;
    bool symbols = false;
    bool unitWidth = false;
    bool sign = false;
    bool decimal = false;
    bool scale = false;
};

} // namespace impl
} // namespace number
U_NAMESPACE_END

#endif //__SOURCE_NUMBER_SKELETONS_H__






using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

template<typename Derived>
Derived NumberFormatterSettings<Derived>::notation(const Notation& notation) const& {
    Derived copy(*this);
    // NOTE: Slicing is OK.
    copy.fMacros.notation = notation;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::notation(const Notation& notation)&& {
    Derived move(std::move(*this));
    // NOTE: Slicing is OK.
    move.fMacros.notation = notation;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::unit(const icu::MeasureUnit& unit) const& {
    Derived copy(*this);
    // NOTE: Slicing occurs here. However, CurrencyUnit can be restored from MeasureUnit.
    // TimeUnit may be affected, but TimeUnit is not as relevant to number formatting.
    copy.fMacros.unit = unit;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::unit(const icu::MeasureUnit& unit)&& {
    Derived move(std::move(*this));
    // See comments above about slicing.
    move.fMacros.unit = unit;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::adoptUnit(icu::MeasureUnit* unit) const& {
    Derived copy(*this);
    // Just move the unit into the MacroProps by value, and delete it since we have ownership.
    // NOTE: Slicing occurs here. However, CurrencyUnit can be restored from MeasureUnit.
    // TimeUnit may be affected, but TimeUnit is not as relevant to number formatting.
    if (unit != nullptr) {
        // TODO: On nullptr, reset to default value?
        copy.fMacros.unit = std::move(*unit);
        delete unit;
    }
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::adoptUnit(icu::MeasureUnit* unit)&& {
    Derived move(std::move(*this));
    // See comments above about slicing and ownership.
    if (unit != nullptr) {
        // TODO: On nullptr, reset to default value?
        move.fMacros.unit = std::move(*unit);
        delete unit;
    }
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::perUnit(const icu::MeasureUnit& perUnit) const& {
    Derived copy(*this);
    // See comments above about slicing.
    copy.fMacros.perUnit = perUnit;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::perUnit(const icu::MeasureUnit& perUnit)&& {
    Derived move(std::move(*this));
    // See comments above about slicing.
    move.fMacros.perUnit = perUnit;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::adoptPerUnit(icu::MeasureUnit* perUnit) const& {
    Derived copy(*this);
    // See comments above about slicing and ownership.
    if (perUnit != nullptr) {
        // TODO: On nullptr, reset to default value?
        copy.fMacros.perUnit = std::move(*perUnit);
        delete perUnit;
    }
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::adoptPerUnit(icu::MeasureUnit* perUnit)&& {
    Derived move(std::move(*this));
    // See comments above about slicing and ownership.
    if (perUnit != nullptr) {
        // TODO: On nullptr, reset to default value?
        move.fMacros.perUnit = std::move(*perUnit);
        delete perUnit;
    }
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::precision(const Precision& precision) const& {
    Derived copy(*this);
    // NOTE: Slicing is OK.
    copy.fMacros.precision = precision;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::precision(const Precision& precision)&& {
    Derived move(std::move(*this));
    // NOTE: Slicing is OK.
    move.fMacros.precision = precision;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::roundingMode(UNumberFormatRoundingMode roundingMode) const& {
    Derived copy(*this);
    copy.fMacros.roundingMode = roundingMode;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::roundingMode(UNumberFormatRoundingMode roundingMode)&& {
    Derived move(std::move(*this));
    move.fMacros.roundingMode = roundingMode;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::grouping(UNumberGroupingStrategy strategy) const& {
    Derived copy(*this);
    // NOTE: This is slightly different than how the setting is stored in Java
    // because we want to put it on the stack.
    copy.fMacros.grouper = Grouper::forStrategy(strategy);
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::grouping(UNumberGroupingStrategy strategy)&& {
    Derived move(std::move(*this));
    move.fMacros.grouper = Grouper::forStrategy(strategy);
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::integerWidth(const IntegerWidth& style) const& {
    Derived copy(*this);
    copy.fMacros.integerWidth = style;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::integerWidth(const IntegerWidth& style)&& {
    Derived move(std::move(*this));
    move.fMacros.integerWidth = style;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::symbols(const DecimalFormatSymbols& symbols) const& {
    Derived copy(*this);
    copy.fMacros.symbols.setTo(symbols);
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::symbols(const DecimalFormatSymbols& symbols)&& {
    Derived move(std::move(*this));
    move.fMacros.symbols.setTo(symbols);
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::adoptSymbols(NumberingSystem* ns) const& {
    Derived copy(*this);
    copy.fMacros.symbols.setTo(ns);
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::adoptSymbols(NumberingSystem* ns)&& {
    Derived move(std::move(*this));
    move.fMacros.symbols.setTo(ns);
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::unitWidth(UNumberUnitWidth width) const& {
    Derived copy(*this);
    copy.fMacros.unitWidth = width;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::unitWidth(UNumberUnitWidth width)&& {
    Derived move(std::move(*this));
    move.fMacros.unitWidth = width;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::sign(UNumberSignDisplay style) const& {
    Derived copy(*this);
    copy.fMacros.sign = style;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::sign(UNumberSignDisplay style)&& {
    Derived move(std::move(*this));
    move.fMacros.sign = style;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::decimal(UNumberDecimalSeparatorDisplay style) const& {
    Derived copy(*this);
    copy.fMacros.decimal = style;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::decimal(UNumberDecimalSeparatorDisplay style)&& {
    Derived move(std::move(*this));
    move.fMacros.decimal = style;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::scale(const Scale& scale) const& {
    Derived copy(*this);
    copy.fMacros.scale = scale;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::scale(const Scale& scale)&& {
    Derived move(std::move(*this));
    move.fMacros.scale = scale;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::padding(const Padder& padder) const& {
    Derived copy(*this);
    copy.fMacros.padder = padder;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::padding(const Padder& padder)&& {
    Derived move(std::move(*this));
    move.fMacros.padder = padder;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::threshold(int32_t threshold) const& {
    Derived copy(*this);
    copy.fMacros.threshold = threshold;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::threshold(int32_t threshold)&& {
    Derived move(std::move(*this));
    move.fMacros.threshold = threshold;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::macros(const icu::number::impl::MacroProps& macros) const& {
    Derived copy(*this);
    copy.fMacros = macros;
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::macros(const icu::number::impl::MacroProps& macros)&& {
    Derived move(std::move(*this));
    move.fMacros = macros;
    return move;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::macros(icu::number::impl::MacroProps&& macros) const& {
    Derived copy(*this);
    copy.fMacros = std::move(macros);
    return copy;
}

template<typename Derived>
Derived NumberFormatterSettings<Derived>::macros(icu::number::impl::MacroProps&& macros)&& {
    Derived move(std::move(*this));
    move.fMacros = std::move(macros);
    return move;
}

template<typename Derived>
UnicodeString NumberFormatterSettings<Derived>::toSkeleton(UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return ICU_Utility::makeBogusString();
    }
    if (fMacros.copyErrorTo(status)) {
        return ICU_Utility::makeBogusString();
    }
    return skeleton::generate(fMacros, status);
}

template<typename Derived>
LocalPointer<Derived> NumberFormatterSettings<Derived>::clone() const & {
    return LocalPointer<Derived>(new Derived(*this));
}

template<typename Derived>
LocalPointer<Derived> NumberFormatterSettings<Derived>::clone() && {
    return LocalPointer<Derived>(new Derived(std::move(*this)));
}

// Declare all classes that implement NumberFormatterSettings
// See https://stackoverflow.com/a/495056/1407170
template
class icu::number::NumberFormatterSettings<icu::number::UnlocalizedNumberFormatter>;
template
class icu::number::NumberFormatterSettings<icu::number::LocalizedNumberFormatter>;


UnlocalizedNumberFormatter NumberFormatter::with() {
    UnlocalizedNumberFormatter result;
    return result;
}

LocalizedNumberFormatter NumberFormatter::withLocale(const Locale& locale) {
    return with().locale(locale);
}

UnlocalizedNumberFormatter
NumberFormatter::forSkeleton(const UnicodeString& skeleton, UErrorCode& status) {
    return skeleton::create(skeleton, nullptr, status);
}

UnlocalizedNumberFormatter
NumberFormatter::forSkeleton(const UnicodeString& skeleton, UParseError& perror, UErrorCode& status) {
    return skeleton::create(skeleton, &perror, status);
}


template<typename T> using fluent_NFS = NumberFormatterSettings<T>;
using fluent_LNF = LocalizedNumberFormatter;
using fluent_UNF = UnlocalizedNumberFormatter;

UnlocalizedNumberFormatter::UnlocalizedNumberFormatter(const fluent_UNF& other)
        : fluent_UNF(static_cast<const fluent_NFS<fluent_UNF>&>(other)) {}

UnlocalizedNumberFormatter::UnlocalizedNumberFormatter(const fluent_NFS<fluent_UNF>& other)
        : fluent_NFS<fluent_UNF>(other) {
    // No additional fields to assign
}

// Make default copy constructor call the NumberFormatterSettings copy constructor.
UnlocalizedNumberFormatter::UnlocalizedNumberFormatter(fluent_UNF&& src) U_NOEXCEPT
        : fluent_UNF(static_cast<fluent_NFS<fluent_UNF>&&>(src)) {}

UnlocalizedNumberFormatter::UnlocalizedNumberFormatter(fluent_NFS<fluent_UNF>&& src) U_NOEXCEPT
        : fluent_NFS<fluent_UNF>(std::move(src)) {
    // No additional fields to assign
}

UnlocalizedNumberFormatter& UnlocalizedNumberFormatter::operator=(const fluent_UNF& other) {
    fluent_NFS<fluent_UNF>::operator=(static_cast<const fluent_NFS<fluent_UNF>&>(other));
    // No additional fields to assign
    return *this;
}

UnlocalizedNumberFormatter& UnlocalizedNumberFormatter::operator=(fluent_UNF&& src) U_NOEXCEPT {
    fluent_NFS<fluent_UNF>::operator=(static_cast<fluent_NFS<fluent_UNF>&&>(src));
    // No additional fields to assign
    return *this;
}

// Make default copy constructor call the NumberFormatterSettings copy constructor.
LocalizedNumberFormatter::LocalizedNumberFormatter(const fluent_LNF& other)
        : fluent_LNF(static_cast<const fluent_NFS<fluent_LNF>&>(other)) {}

LocalizedNumberFormatter::LocalizedNumberFormatter(const fluent_NFS<fluent_LNF>& other)
        : fluent_NFS<fluent_LNF>(other) {
    // No additional fields to assign (let call count and compiled formatter reset to defaults)
}

LocalizedNumberFormatter::LocalizedNumberFormatter(LocalizedNumberFormatter&& src) U_NOEXCEPT
        : fluent_LNF(static_cast<fluent_NFS<fluent_LNF>&&>(src)) {}

LocalizedNumberFormatter::LocalizedNumberFormatter(fluent_NFS<fluent_LNF>&& src) U_NOEXCEPT
        : fluent_NFS<fluent_LNF>(std::move(src)) {
    // For the move operators, copy over the compiled formatter.
    // Note: if the formatter is not compiled, call count information is lost.
    if (static_cast<fluent_LNF&&>(src).fCompiled != nullptr) {
        lnfMoveHelper(static_cast<fluent_LNF&&>(src));
    }
}

LocalizedNumberFormatter& LocalizedNumberFormatter::operator=(const fluent_LNF& other) {
    fluent_NFS<fluent_LNF>::operator=(static_cast<const fluent_NFS<fluent_LNF>&>(other));
    // Reset to default values.
    clear();
    return *this;
}

LocalizedNumberFormatter& LocalizedNumberFormatter::operator=(fluent_LNF&& src) U_NOEXCEPT {
    fluent_NFS<fluent_LNF>::operator=(static_cast<fluent_NFS<fluent_LNF>&&>(src));
    // For the move operators, copy over the compiled formatter.
    // Note: if the formatter is not compiled, call count information is lost.
    if (static_cast<fluent_LNF&&>(src).fCompiled != nullptr) {
        // Formatter is compiled
        lnfMoveHelper(static_cast<fluent_LNF&&>(src));
    } else {
        clear();
    }
    return *this;
}

void LocalizedNumberFormatter::clear() {
    // Reset to default values.
    auto* callCount = reinterpret_cast<u_atomic_int32_t*>(fUnsafeCallCount);
    umtx_storeRelease(*callCount, 0);
    delete fCompiled;
    fCompiled = nullptr;
}

void LocalizedNumberFormatter::lnfMoveHelper(fluent_LNF&& src) {
    // Copy over the compiled formatter and set call count to INT32_MIN as in computeCompiled().
    // Don't copy the call count directly because doing so requires a loadAcquire/storeRelease.
    // The bits themselves appear to be platform-dependent, so copying them might not be safe.
    auto* callCount = reinterpret_cast<u_atomic_int32_t*>(fUnsafeCallCount);
    umtx_storeRelease(*callCount, INT32_MIN);
    delete fCompiled;
    fCompiled = src.fCompiled;
    // Reset the source object to leave it in a safe state.
    auto* srcCallCount = reinterpret_cast<u_atomic_int32_t*>(src.fUnsafeCallCount);
    umtx_storeRelease(*srcCallCount, 0);
    src.fCompiled = nullptr;
}


LocalizedNumberFormatter::~LocalizedNumberFormatter() {
    delete fCompiled;
}

LocalizedNumberFormatter::LocalizedNumberFormatter(const MacroProps& macros, const Locale& locale) {
    fMacros = macros;
    fMacros.locale = locale;
}

LocalizedNumberFormatter::LocalizedNumberFormatter(MacroProps&& macros, const Locale& locale) {
    fMacros = std::move(macros);
    fMacros.locale = locale;
}

LocalizedNumberFormatter UnlocalizedNumberFormatter::locale(const Locale& locale) const& {
    return LocalizedNumberFormatter(fMacros, locale);
}

LocalizedNumberFormatter UnlocalizedNumberFormatter::locale(const Locale& locale)&& {
    return LocalizedNumberFormatter(std::move(fMacros), locale);
}

SymbolsWrapper::SymbolsWrapper(const SymbolsWrapper& other) {
    doCopyFrom(other);
}

SymbolsWrapper::SymbolsWrapper(SymbolsWrapper&& src) U_NOEXCEPT {
    doMoveFrom(std::move(src));
}

SymbolsWrapper& SymbolsWrapper::operator=(const SymbolsWrapper& other) {
    if (this == &other) {
        return *this;
    }
    doCleanup();
    doCopyFrom(other);
    return *this;
}

SymbolsWrapper& SymbolsWrapper::operator=(SymbolsWrapper&& src) U_NOEXCEPT {
    if (this == &src) {
        return *this;
    }
    doCleanup();
    doMoveFrom(std::move(src));
    return *this;
}

SymbolsWrapper::~SymbolsWrapper() {
    doCleanup();
}

void SymbolsWrapper::setTo(const DecimalFormatSymbols& dfs) {
    doCleanup();
    fType = SYMPTR_DFS;
    fPtr.dfs = new DecimalFormatSymbols(dfs);
}

void SymbolsWrapper::setTo(const NumberingSystem* ns) {
    doCleanup();
    fType = SYMPTR_NS;
    fPtr.ns = ns;
}

void SymbolsWrapper::doCopyFrom(const SymbolsWrapper& other) {
    fType = other.fType;
    switch (fType) {
        case SYMPTR_NONE:
            // No action necessary
            break;
        case SYMPTR_DFS:
            // Memory allocation failures are exposed in copyErrorTo()
            if (other.fPtr.dfs != nullptr) {
                fPtr.dfs = new DecimalFormatSymbols(*other.fPtr.dfs);
            } else {
                fPtr.dfs = nullptr;
            }
            break;
        case SYMPTR_NS:
            // Memory allocation failures are exposed in copyErrorTo()
            if (other.fPtr.ns != nullptr) {
                fPtr.ns = new NumberingSystem(*other.fPtr.ns);
            } else {
                fPtr.ns = nullptr;
            }
            break;
    }
}

void SymbolsWrapper::doMoveFrom(SymbolsWrapper&& src) {
    fType = src.fType;
    switch (fType) {
        case SYMPTR_NONE:
            // No action necessary
            break;
        case SYMPTR_DFS:
            fPtr.dfs = src.fPtr.dfs;
            src.fPtr.dfs = nullptr;
            break;
        case SYMPTR_NS:
            fPtr.ns = src.fPtr.ns;
            src.fPtr.ns = nullptr;
            break;
    }
}

void SymbolsWrapper::doCleanup() {
    switch (fType) {
        case SYMPTR_NONE:
            // No action necessary
            break;
        case SYMPTR_DFS:
            delete fPtr.dfs;
            break;
        case SYMPTR_NS:
            delete fPtr.ns;
            break;
    }
}

bool SymbolsWrapper::isDecimalFormatSymbols() const {
    return fType == SYMPTR_DFS;
}

bool SymbolsWrapper::isNumberingSystem() const {
    return fType == SYMPTR_NS;
}

const DecimalFormatSymbols* SymbolsWrapper::getDecimalFormatSymbols() const {
    U_ASSERT(fType == SYMPTR_DFS);
    return fPtr.dfs;
}

const NumberingSystem* SymbolsWrapper::getNumberingSystem() const {
    U_ASSERT(fType == SYMPTR_NS);
    return fPtr.ns;
}


FormattedNumber LocalizedNumberFormatter::formatInt(int64_t value, UErrorCode& status) const {
    if (U_FAILURE(status)) { return FormattedNumber(U_ILLEGAL_ARGUMENT_ERROR); }
    auto results = new UFormattedNumberData();
    if (results == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return FormattedNumber(status);
    }
    results->quantity.setToLong(value);
    formatImpl(results, status);

    // Do not save the results object if we encountered a failure.
    if (U_SUCCESS(status)) {
        return FormattedNumber(results);
    } else {
        delete results;
        return FormattedNumber(status);
    }
}

FormattedNumber LocalizedNumberFormatter::formatDouble(double value, UErrorCode& status) const {
    if (U_FAILURE(status)) { return FormattedNumber(U_ILLEGAL_ARGUMENT_ERROR); }
    auto results = new UFormattedNumberData();
    if (results == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return FormattedNumber(status);
    }
    results->quantity.setToDouble(value);
    formatImpl(results, status);

    // Do not save the results object if we encountered a failure.
    if (U_SUCCESS(status)) {
        return FormattedNumber(results);
    } else {
        delete results;
        return FormattedNumber(status);
    }
}

FormattedNumber LocalizedNumberFormatter::formatDecimal(StringPiece value, UErrorCode& status) const {
    if (U_FAILURE(status)) { return FormattedNumber(U_ILLEGAL_ARGUMENT_ERROR); }
    auto results = new UFormattedNumberData();
    if (results == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return FormattedNumber(status);
    }
    results->quantity.setToDecNumber(value, status);
    formatImpl(results, status);

    // Do not save the results object if we encountered a failure.
    if (U_SUCCESS(status)) {
        return FormattedNumber(results);
    } else {
        delete results;
        return FormattedNumber(status);
    }
}

FormattedNumber
LocalizedNumberFormatter::formatDecimalQuantity(const DecimalQuantity& dq, UErrorCode& status) const {
    if (U_FAILURE(status)) { return FormattedNumber(U_ILLEGAL_ARGUMENT_ERROR); }
    auto results = new UFormattedNumberData();
    if (results == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return FormattedNumber(status);
    }
    results->quantity = dq;
    formatImpl(results, status);

    // Do not save the results object if we encountered a failure.
    if (U_SUCCESS(status)) {
        return FormattedNumber(results);
    } else {
        delete results;
        return FormattedNumber(status);
    }
}

void LocalizedNumberFormatter::formatImpl(icu::number::impl::UFormattedNumberData* results, UErrorCode& status) const {
    if (computeCompiled(status)) {
        fCompiled->format(results->quantity, results->getStringRef(), status);
    } else {
        NumberFormatterImpl::formatStatic(fMacros, results->quantity, results->getStringRef(), status);
    }
    if (U_FAILURE(status)) {
        return;
    }
    results->getStringRef().writeTerminator(status);
}

void LocalizedNumberFormatter::getAffixImpl(bool isPrefix, bool isNegative, UnicodeString& result,
                                            UErrorCode& status) const {
    FormattedStringBuilder string;
    auto signum = static_cast<Signum>(isNegative ? SIGNUM_NEG : SIGNUM_POS);
    // Always return affixes for plural form OTHER.
    static const StandardPlural::Form plural = StandardPlural::OTHER;
    int32_t prefixLength;
    if (computeCompiled(status)) {
        prefixLength = fCompiled->getPrefixSuffix(signum, plural, string, status);
    } else {
        prefixLength = NumberFormatterImpl::getPrefixSuffixStatic(fMacros, signum, plural, string, status);
    }
    result.remove();
    if (isPrefix) {
        result.append(string.toTempUnicodeString().tempSubStringBetween(0, prefixLength));
    } else {
        result.append(string.toTempUnicodeString().tempSubStringBetween(prefixLength, string.length()));
    }
}

bool LocalizedNumberFormatter::computeCompiled(UErrorCode& status) const {
    // fUnsafeCallCount contains memory to be interpreted as an atomic int, most commonly
    // std::atomic<int32_t>.  Since the type of atomic int is platform-dependent, we cast the
    // bytes in fUnsafeCallCount to u_atomic_int32_t, a typedef for the platform-dependent
    // atomic int type defined in umutex.h.
    static_assert(
            sizeof(u_atomic_int32_t) <= sizeof(fUnsafeCallCount),
            "Atomic integer size on this platform exceeds the size allocated by fUnsafeCallCount");
    auto* callCount = reinterpret_cast<u_atomic_int32_t*>(
            const_cast<LocalizedNumberFormatter*>(this)->fUnsafeCallCount);

    // A positive value in the atomic int indicates that the data structure is not yet ready;
    // a negative value indicates that it is ready. If, after the increment, the atomic int
    // is exactly threshold, then it is the current thread's job to build the data structure.
    // Note: We set the callCount to INT32_MIN so that if another thread proceeds to increment
    // the atomic int, the value remains below zero.
    int32_t currentCount = umtx_loadAcquire(*callCount);
    if (0 <= currentCount && currentCount <= fMacros.threshold && fMacros.threshold > 0) {
        currentCount = umtx_atomic_inc(callCount);
    }

    if (currentCount == fMacros.threshold && fMacros.threshold > 0) {
        // Build the data structure and then use it (slow to fast path).
        const NumberFormatterImpl* compiled = new NumberFormatterImpl(fMacros, status);
        if (compiled == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return false;
        }
        U_ASSERT(fCompiled == nullptr);
        const_cast<LocalizedNumberFormatter*>(this)->fCompiled = compiled;
        umtx_storeRelease(*callCount, INT32_MIN);
        return true;
    } else if (currentCount < 0) {
        // The data structure is already built; use it (fast path).
        U_ASSERT(fCompiled != nullptr);
        return true;
    } else {
        // Format the number without building the data structure (slow path).
        return false;
    }
}

const icu::number::impl::NumberFormatterImpl* LocalizedNumberFormatter::getCompiled() const {
    return fCompiled;
}

int32_t LocalizedNumberFormatter::getCallCount() const {
    auto* callCount = reinterpret_cast<u_atomic_int32_t*>(
            const_cast<LocalizedNumberFormatter*>(this)->fUnsafeCallCount);
    return umtx_loadAcquire(*callCount);
}

Format* LocalizedNumberFormatter::toFormat(UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    LocalPointer<LocalizedNumberFormatterAsFormat> retval(
            new LocalizedNumberFormatterAsFormat(*this, fMacros.locale), status);
    return retval.orphan();
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING
















using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

namespace {

struct CurrencyFormatInfoResult {
    bool exists;
    const char16_t* pattern;
    const char16_t* decimalSeparator;
    const char16_t* groupingSeparator;
};

CurrencyFormatInfoResult
getCurrencyFormatInfo(const Locale& locale, const char* isoCode, UErrorCode& status) {
    // TODO: Load this data in a centralized location like ICU4J?
    // TODO: Move this into the CurrencySymbols class?
    // TODO: Parts of this same data are loaded in dcfmtsym.cpp; should clean up.
    CurrencyFormatInfoResult result = {false, nullptr, nullptr, nullptr};
    if (U_FAILURE(status)) { return result; }
    CharString key;
    key.append("Currencies/", status);
    key.append(isoCode, status);
    UErrorCode localStatus = status;
    LocalUResourceBundlePointer bundle(ures_open(U_ICUDATA_CURR, locale.getName(), &localStatus));
    ures_getByKeyWithFallback(bundle.getAlias(), key.data(), bundle.getAlias(), &localStatus);
    if (U_SUCCESS(localStatus) &&
        ures_getSize(bundle.getAlias()) > 2) { // the length is 3 if more data is present
        ures_getByIndex(bundle.getAlias(), 2, bundle.getAlias(), &localStatus);
        int32_t dummy;
        result.exists = true;
        result.pattern = ures_getStringByIndex(bundle.getAlias(), 0, &dummy, &localStatus);
        result.decimalSeparator = ures_getStringByIndex(bundle.getAlias(), 1, &dummy, &localStatus);
        result.groupingSeparator = ures_getStringByIndex(bundle.getAlias(), 2, &dummy, &localStatus);
        status = localStatus;
    } else if (localStatus != U_MISSING_RESOURCE_ERROR) {
        status = localStatus;
    }
    return result;
}

}  // namespace


MicroPropsGenerator::~MicroPropsGenerator() = default;


NumberFormatterImpl::NumberFormatterImpl(const MacroProps& macros, UErrorCode& status)
    : NumberFormatterImpl(macros, true, status) {
}

int32_t NumberFormatterImpl::formatStatic(const MacroProps& macros, DecimalQuantity& inValue,
                                       FormattedStringBuilder& outString, UErrorCode& status) {
    NumberFormatterImpl impl(macros, false, status);
    MicroProps& micros = impl.preProcessUnsafe(inValue, status);
    if (U_FAILURE(status)) { return 0; }
    int32_t length = writeNumber(micros, inValue, outString, 0, status);
    length += writeAffixes(micros, outString, 0, length, status);
    return length;
}

int32_t NumberFormatterImpl::getPrefixSuffixStatic(const MacroProps& macros, Signum signum,
                                                   StandardPlural::Form plural,
                                                   FormattedStringBuilder& outString, UErrorCode& status) {
    NumberFormatterImpl impl(macros, false, status);
    return impl.getPrefixSuffixUnsafe(signum, plural, outString, status);
}

// NOTE: C++ SPECIFIC DIFFERENCE FROM JAVA:
// The "safe" apply method uses a new MicroProps. In the MicroPropsGenerator, fMicros is copied into the new instance.
// The "unsafe" method simply re-uses fMicros, eliminating the extra copy operation.
// See MicroProps::processQuantity() for details.

int32_t NumberFormatterImpl::format(DecimalQuantity& inValue, FormattedStringBuilder& outString,
                                UErrorCode& status) const {
    MicroProps micros;
    preProcess(inValue, micros, status);
    if (U_FAILURE(status)) { return 0; }
    int32_t length = writeNumber(micros, inValue, outString, 0, status);
    length += writeAffixes(micros, outString, 0, length, status);
    return length;
}

void NumberFormatterImpl::preProcess(DecimalQuantity& inValue, MicroProps& microsOut,
                                     UErrorCode& status) const {
    if (U_FAILURE(status)) { return; }
    if (fMicroPropsGenerator == nullptr) {
        status = U_INTERNAL_PROGRAM_ERROR;
        return;
    }
    fMicroPropsGenerator->processQuantity(inValue, microsOut, status);
    microsOut.rounder.apply(inValue, status);
    microsOut.integerWidth.apply(inValue, status);
}

MicroProps& NumberFormatterImpl::preProcessUnsafe(DecimalQuantity& inValue, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return fMicros; // must always return a value
    }
    if (fMicroPropsGenerator == nullptr) {
        status = U_INTERNAL_PROGRAM_ERROR;
        return fMicros; // must always return a value
    }
    fMicroPropsGenerator->processQuantity(inValue, fMicros, status);
    fMicros.rounder.apply(inValue, status);
    fMicros.integerWidth.apply(inValue, status);
    return fMicros;
}

int32_t NumberFormatterImpl::getPrefixSuffix(Signum signum, StandardPlural::Form plural,
                                             FormattedStringBuilder& outString, UErrorCode& status) const {
    if (U_FAILURE(status)) { return 0; }
    // #13453: DecimalFormat wants the affixes from the pattern only (modMiddle, aka pattern modifier).
    // Safe path: use fImmutablePatternModifier.
    const Modifier* modifier = fImmutablePatternModifier->getModifier(signum, plural);
    modifier->apply(outString, 0, 0, status);
    if (U_FAILURE(status)) { return 0; }
    return modifier->getPrefixLength();
}

int32_t NumberFormatterImpl::getPrefixSuffixUnsafe(Signum signum, StandardPlural::Form plural,
                                                   FormattedStringBuilder& outString, UErrorCode& status) {
    if (U_FAILURE(status)) { return 0; }
    // #13453: DecimalFormat wants the affixes from the pattern only (modMiddle, aka pattern modifier).
    // Unsafe path: use fPatternModifier.
    fPatternModifier->setNumberProperties(signum, plural);
    fPatternModifier->apply(outString, 0, 0, status);
    if (U_FAILURE(status)) { return 0; }
    return fPatternModifier->getPrefixLength();
}

NumberFormatterImpl::NumberFormatterImpl(const MacroProps& macros, bool safe, UErrorCode& status) {
    fMicroPropsGenerator = macrosToMicroGenerator(macros, safe, status);
}

//////////

const MicroPropsGenerator*
NumberFormatterImpl::macrosToMicroGenerator(const MacroProps& macros, bool safe, UErrorCode& status) {
    if (U_FAILURE(status)) { return nullptr; }
    const MicroPropsGenerator* chain = &fMicros;

    // Check that macros is error-free before continuing.
    if (macros.copyErrorTo(status)) {
        return nullptr;
    }

    // TODO: Accept currency symbols from DecimalFormatSymbols?

    // Pre-compute a few values for efficiency.
    bool isCurrency = utils::unitIsCurrency(macros.unit);
    bool isNoUnit = utils::unitIsNoUnit(macros.unit);
    bool isPercent = utils::unitIsPercent(macros.unit);
    bool isPermille = utils::unitIsPermille(macros.unit);
    bool isAccounting =
            macros.sign == UNUM_SIGN_ACCOUNTING || macros.sign == UNUM_SIGN_ACCOUNTING_ALWAYS ||
            macros.sign == UNUM_SIGN_ACCOUNTING_EXCEPT_ZERO;
    CurrencyUnit currency(u"", status);
    if (isCurrency) {
        currency = CurrencyUnit(macros.unit, status); // Restore CurrencyUnit from MeasureUnit
    }
    const CurrencySymbols* currencySymbols;
    if (macros.currencySymbols != nullptr) {
        // Used by the DecimalFormat code path
        currencySymbols = macros.currencySymbols;
    } else {
        fWarehouse.fCurrencySymbols = {currency, macros.locale, status};
        currencySymbols = &fWarehouse.fCurrencySymbols;
    }
    UNumberUnitWidth unitWidth = UNUM_UNIT_WIDTH_SHORT;
    if (macros.unitWidth != UNUM_UNIT_WIDTH_COUNT) {
        unitWidth = macros.unitWidth;
    }
    bool isCldrUnit = !isCurrency && !isNoUnit &&
        (unitWidth == UNUM_UNIT_WIDTH_FULL_NAME || !(isPercent || isPermille));

    // Select the numbering system.
    LocalPointer<const NumberingSystem> nsLocal;
    const NumberingSystem* ns;
    if (macros.symbols.isNumberingSystem()) {
        ns = macros.symbols.getNumberingSystem();
    } else {
        // TODO: Is there a way to avoid creating the NumberingSystem object?
        ns = NumberingSystem::createInstance(macros.locale, status);
        // Give ownership to the function scope.
        nsLocal.adoptInstead(ns);
    }
    const char* nsName = U_SUCCESS(status) ? ns->getName() : "latn";
    uprv_strncpy(fMicros.nsName, nsName, 8);
    fMicros.nsName[8] = 0; // guarantee NUL-terminated

    // Resolve the symbols. Do this here because currency may need to customize them.
    if (macros.symbols.isDecimalFormatSymbols()) {
        fMicros.symbols = macros.symbols.getDecimalFormatSymbols();
    } else {
        auto newSymbols = new DecimalFormatSymbols(macros.locale, *ns, status);
        if (newSymbols == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return nullptr;
        }
        fMicros.symbols = newSymbols;
        // Give ownership to the NumberFormatterImpl.
        fSymbols.adoptInstead(fMicros.symbols);
    }

    // Load and parse the pattern string. It is used for grouping sizes and affixes only.
    // If we are formatting currency, check for a currency-specific pattern.
    const char16_t* pattern = nullptr;
    if (isCurrency) {
        CurrencyFormatInfoResult info = getCurrencyFormatInfo(
                macros.locale, currency.getSubtype(), status);
        if (info.exists) {
            pattern = info.pattern;
            // It's clunky to clone an object here, but this code is not frequently executed.
            auto symbols = new DecimalFormatSymbols(*fMicros.symbols);
            if (symbols == nullptr) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return nullptr;
            }
            fMicros.symbols = symbols;
            fSymbols.adoptInstead(symbols);
            symbols->setSymbol(
                    DecimalFormatSymbols::ENumberFormatSymbol::kMonetarySeparatorSymbol,
                    UnicodeString(info.decimalSeparator),
                    FALSE);
            symbols->setSymbol(
                    DecimalFormatSymbols::ENumberFormatSymbol::kMonetaryGroupingSeparatorSymbol,
                    UnicodeString(info.groupingSeparator),
                    FALSE);
        }
    }
    if (pattern == nullptr) {
        CldrPatternStyle patternStyle;
        if (isCldrUnit) {
            patternStyle = CLDR_PATTERN_STYLE_DECIMAL;
        } else if (isPercent || isPermille) {
            patternStyle = CLDR_PATTERN_STYLE_PERCENT;
        } else if (!isCurrency || unitWidth == UNUM_UNIT_WIDTH_FULL_NAME) {
            patternStyle = CLDR_PATTERN_STYLE_DECIMAL;
        } else if (isAccounting) {
            // NOTE: Although ACCOUNTING and ACCOUNTING_ALWAYS are only supported in currencies right now,
            // the API contract allows us to add support to other units in the future.
            patternStyle = CLDR_PATTERN_STYLE_ACCOUNTING;
        } else {
            patternStyle = CLDR_PATTERN_STYLE_CURRENCY;
        }
        pattern = utils::getPatternForStyle(macros.locale, nsName, patternStyle, status);
    }
    auto patternInfo = new ParsedPatternInfo();
    if (patternInfo == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    fPatternInfo.adoptInstead(patternInfo);
    PatternParser::parseToPatternInfo(UnicodeString(pattern), *patternInfo, status);

    /////////////////////////////////////////////////////////////////////////////////////
    /// START POPULATING THE DEFAULT MICROPROPS AND BUILDING THE MICROPROPS GENERATOR ///
    /////////////////////////////////////////////////////////////////////////////////////

    // Multiplier
    if (macros.scale.isValid()) {
        fMicros.helpers.multiplier.setAndChain(macros.scale, chain);
        chain = &fMicros.helpers.multiplier;
    }

    // Rounding strategy
    Precision precision;
    if (!macros.precision.isBogus()) {
        precision = macros.precision;
    } else if (macros.notation.fType == Notation::NTN_COMPACT) {
        precision = Precision::integer().withMinDigits(2);
    } else if (isCurrency) {
        precision = Precision::currency(UCURR_USAGE_STANDARD);
    } else {
        precision = Precision::maxFraction(6);
    }
    UNumberFormatRoundingMode roundingMode;
    if (macros.roundingMode != kDefaultMode) {
        roundingMode = macros.roundingMode;
    } else {
        // Temporary until ICU 64
        roundingMode = precision.fRoundingMode;
    }
    fMicros.rounder = {precision, roundingMode, currency, status};

    // Grouping strategy
    if (!macros.grouper.isBogus()) {
        fMicros.grouping = macros.grouper;
    } else if (macros.notation.fType == Notation::NTN_COMPACT) {
        // Compact notation uses minGrouping by default since ICU 59
        fMicros.grouping = Grouper::forStrategy(UNUM_GROUPING_MIN2);
    } else {
        fMicros.grouping = Grouper::forStrategy(UNUM_GROUPING_AUTO);
    }
    fMicros.grouping.setLocaleData(*fPatternInfo, macros.locale);

    // Padding strategy
    if (!macros.padder.isBogus()) {
        fMicros.padding = macros.padder;
    } else {
        fMicros.padding = Padder::none();
    }

    // Integer width
    if (!macros.integerWidth.isBogus()) {
        fMicros.integerWidth = macros.integerWidth;
    } else {
        fMicros.integerWidth = IntegerWidth::standard();
    }

    // Sign display
    if (macros.sign != UNUM_SIGN_COUNT) {
        fMicros.sign = macros.sign;
    } else {
        fMicros.sign = UNUM_SIGN_AUTO;
    }

    // Decimal mark display
    if (macros.decimal != UNUM_DECIMAL_SEPARATOR_COUNT) {
        fMicros.decimal = macros.decimal;
    } else {
        fMicros.decimal = UNUM_DECIMAL_SEPARATOR_AUTO;
    }

    // Use monetary separator symbols
    fMicros.useCurrency = isCurrency;

    // Inner modifier (scientific notation)
    if (macros.notation.fType == Notation::NTN_SCIENTIFIC) {
        auto newScientificHandler = new ScientificHandler(&macros.notation, fMicros.symbols, chain);
        if (newScientificHandler == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return nullptr;
        }
        fScientificHandler.adoptInstead(newScientificHandler);
        chain = fScientificHandler.getAlias();
    } else {
        // No inner modifier required
        fMicros.modInner = &fMicros.helpers.emptyStrongModifier;
    }

    // Middle modifier (patterns, positive/negative, currency symbols, percent)
    auto patternModifier = new MutablePatternModifier(false);
    if (patternModifier == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    fPatternModifier.adoptInstead(patternModifier);
    patternModifier->setPatternInfo(
            macros.affixProvider != nullptr ? macros.affixProvider
                                            : static_cast<const AffixPatternProvider*>(fPatternInfo.getAlias()),
            UNUM_FIELD_COUNT);
    patternModifier->setPatternAttributes(fMicros.sign, isPermille);
    if (patternModifier->needsPlurals()) {
        patternModifier->setSymbols(
                fMicros.symbols,
                currencySymbols,
                unitWidth,
                resolvePluralRules(macros.rules, macros.locale, status));
    } else {
        patternModifier->setSymbols(fMicros.symbols, currencySymbols, unitWidth, nullptr);
    }
    if (safe) {
        fImmutablePatternModifier.adoptInstead(patternModifier->createImmutableAndChain(chain, status));
        chain = fImmutablePatternModifier.getAlias();
    } else {
        patternModifier->addToChain(chain);
        chain = patternModifier;
    }

    // Outer modifier (CLDR units and currency long names)
    if (isCldrUnit) {
        fLongNameHandler.adoptInstead(
                LongNameHandler::forMeasureUnit(
                        macros.locale,
                        macros.unit,
                        macros.perUnit,
                        unitWidth,
                        resolvePluralRules(macros.rules, macros.locale, status),
                        chain,
                        status));
        chain = fLongNameHandler.getAlias();
    } else if (isCurrency && unitWidth == UNUM_UNIT_WIDTH_FULL_NAME) {
        fLongNameHandler.adoptInstead(
                LongNameHandler::forCurrencyLongNames(
                        macros.locale,
                        currency,
                        resolvePluralRules(macros.rules, macros.locale, status),
                        chain,
                        status));
        chain = fLongNameHandler.getAlias();
    } else {
        // No outer modifier required
        fMicros.modOuter = &fMicros.helpers.emptyWeakModifier;
    }

    // Compact notation
    // NOTE: Compact notation can (but might not) override the middle modifier and rounding.
    // It therefore needs to go at the end of the chain.
    if (macros.notation.fType == Notation::NTN_COMPACT) {
        CompactType compactType = (isCurrency && unitWidth != UNUM_UNIT_WIDTH_FULL_NAME)
                                  ? CompactType::TYPE_CURRENCY : CompactType::TYPE_DECIMAL;
        auto newCompactHandler = new CompactHandler(
            macros.notation.fUnion.compactStyle,
            macros.locale,
            nsName,
            compactType,
            resolvePluralRules(macros.rules, macros.locale, status),
            safe ? patternModifier : nullptr,
            chain,
            status);
        if (newCompactHandler == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return nullptr;
        }
        fCompactHandler.adoptInstead(newCompactHandler);
        chain = fCompactHandler.getAlias();
    }

    return chain;
}

const PluralRules*
NumberFormatterImpl::resolvePluralRules(const PluralRules* rulesPtr, const Locale& locale,
                                        UErrorCode& status) {
    if (rulesPtr != nullptr) {
        return rulesPtr;
    }
    // Lazily create PluralRules
    if (fRules.isNull()) {
        fRules.adoptInstead(PluralRules::forLocale(locale, status));
    }
    return fRules.getAlias();
}

int32_t NumberFormatterImpl::writeAffixes(const MicroProps& micros, FormattedStringBuilder& string,
                                          int32_t start, int32_t end, UErrorCode& status) {
    // Always apply the inner modifier (which is "strong").
    int32_t length = micros.modInner->apply(string, start, end, status);
    if (micros.padding.isValid()) {
        length += micros.padding
                .padAndApply(*micros.modMiddle, *micros.modOuter, string, start, length + end, status);
    } else {
        length += micros.modMiddle->apply(string, start, length + end, status);
        length += micros.modOuter->apply(string, start, length + end, status);
    }
    return length;
}

int32_t NumberFormatterImpl::writeNumber(const MicroProps& micros, DecimalQuantity& quantity,
                                         FormattedStringBuilder& string, int32_t index,
                                         UErrorCode& status) {
    int32_t length = 0;
    if (quantity.isInfinite()) {
        length += string.insert(
                length + index,
                micros.symbols->getSymbol(DecimalFormatSymbols::ENumberFormatSymbol::kInfinitySymbol),
                UNUM_INTEGER_FIELD,
                status);

    } else if (quantity.isNaN()) {
        length += string.insert(
                length + index,
                micros.symbols->getSymbol(DecimalFormatSymbols::ENumberFormatSymbol::kNaNSymbol),
                UNUM_INTEGER_FIELD,
                status);

    } else {
        // Add the integer digits
        length += writeIntegerDigits(micros, quantity, string, length + index, status);

        // Add the decimal point
        if (quantity.getLowerDisplayMagnitude() < 0 || micros.decimal == UNUM_DECIMAL_SEPARATOR_ALWAYS) {
            length += string.insert(
                    length + index,
                    micros.useCurrency ? micros.symbols->getSymbol(
                            DecimalFormatSymbols::ENumberFormatSymbol::kMonetarySeparatorSymbol) : micros
                            .symbols
                            ->getSymbol(
                                    DecimalFormatSymbols::ENumberFormatSymbol::kDecimalSeparatorSymbol),
                    UNUM_DECIMAL_SEPARATOR_FIELD,
                    status);
        }

        // Add the fraction digits
        length += writeFractionDigits(micros, quantity, string, length + index, status);
    }

    return length;
}

int32_t NumberFormatterImpl::writeIntegerDigits(const MicroProps& micros, DecimalQuantity& quantity,
                                                FormattedStringBuilder& string, int32_t index,
                                                UErrorCode& status) {
    int length = 0;
    int integerCount = quantity.getUpperDisplayMagnitude() + 1;
    for (int i = 0; i < integerCount; i++) {
        // Add grouping separator
        if (micros.grouping.groupAtPosition(i, quantity)) {
            length += string.insert(
                    index,
                    micros.useCurrency ? micros.symbols->getSymbol(
                            DecimalFormatSymbols::ENumberFormatSymbol::kMonetaryGroupingSeparatorSymbol)
                                       : micros.symbols->getSymbol(
                            DecimalFormatSymbols::ENumberFormatSymbol::kGroupingSeparatorSymbol),
                    UNUM_GROUPING_SEPARATOR_FIELD,
                    status);
        }

        // Get and append the next digit value
        int8_t nextDigit = quantity.getDigit(i);
        length += utils::insertDigitFromSymbols(
                string, index, nextDigit, *micros.symbols, UNUM_INTEGER_FIELD, status);
    }
    return length;
}

int32_t NumberFormatterImpl::writeFractionDigits(const MicroProps& micros, DecimalQuantity& quantity,
                                                 FormattedStringBuilder& string, int32_t index,
                                                 UErrorCode& status) {
    int length = 0;
    int fractionCount = -quantity.getLowerDisplayMagnitude();
    for (int i = 0; i < fractionCount; i++) {
        // Get and append the next digit value
        int8_t nextDigit = quantity.getDigit(-i - 1);
        length += utils::insertDigitFromSymbols(
                string, length + index, nextDigit, *micros.symbols, UNUM_FRACTION_FIELD, status);
    }
    return length;
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING





using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

namespace {

int16_t getMinGroupingForLocale(const Locale& locale) {
    // TODO: Cache this?
    UErrorCode localStatus = U_ZERO_ERROR;
    LocalUResourceBundlePointer bundle(ures_open(NULL, locale.getName(), &localStatus));
    int32_t resultLen = 0;
    const char16_t* result = ures_getStringByKeyWithFallback(
        bundle.getAlias(),
        "NumberElements/minimumGroupingDigits",
        &resultLen,
        &localStatus);
    // TODO: Is it safe to assume resultLen == 1? Would locales set minGrouping >= 10?
    if (U_FAILURE(localStatus) || resultLen != 1) {
        return 1;
    }
    return result[0] - u'0';
}

}

Grouper Grouper::forStrategy(UNumberGroupingStrategy grouping) {
    switch (grouping) {
    case UNUM_GROUPING_OFF:
        return {-1, -1, -2, grouping};
    case UNUM_GROUPING_AUTO:
        return {-2, -2, -2, grouping};
    case UNUM_GROUPING_MIN2:
        return {-2, -2, -3, grouping};
    case UNUM_GROUPING_ON_ALIGNED:
        return {-4, -4, 1, grouping};
    case UNUM_GROUPING_THOUSANDS:
        return {3, 3, 1, grouping};
    default:
        UPRV_UNREACHABLE;
    }
}

Grouper Grouper::forProperties(const DecimalFormatProperties& properties) {
    if (!properties.groupingUsed) {
        return forStrategy(UNUM_GROUPING_OFF);
    }
    auto grouping1 = static_cast<int16_t>(properties.groupingSize);
    auto grouping2 = static_cast<int16_t>(properties.secondaryGroupingSize);
    auto minGrouping = static_cast<int16_t>(properties.minimumGroupingDigits);
    grouping1 = grouping1 > 0 ? grouping1 : grouping2 > 0 ? grouping2 : grouping1;
    grouping2 = grouping2 > 0 ? grouping2 : grouping1;
    return {grouping1, grouping2, minGrouping, UNUM_GROUPING_COUNT};
}

void Grouper::setLocaleData(const number::impl::ParsedPatternInfo &patternInfo, const Locale& locale) {
    if (fGrouping1 != -2 && fGrouping2 != -4) {
        return;
    }
    auto grouping1 = static_cast<int16_t> (patternInfo.positive.groupingSizes & 0xffff);
    auto grouping2 = static_cast<int16_t> ((patternInfo.positive.groupingSizes >> 16) & 0xffff);
    auto grouping3 = static_cast<int16_t> ((patternInfo.positive.groupingSizes >> 32) & 0xffff);
    if (grouping2 == -1) {
        grouping1 = fGrouping1 == -4 ? (short) 3 : (short) -1;
    }
    if (grouping3 == -1) {
        grouping2 = grouping1;
    }
    if (fMinGrouping == -2) {
        fMinGrouping = getMinGroupingForLocale(locale);
    } else if (fMinGrouping == -3) {
        fMinGrouping = static_cast<int16_t>(uprv_max(2, getMinGroupingForLocale(locale)));
    } else {
        // leave fMinGrouping alone
    }
    fGrouping1 = grouping1;
    fGrouping2 = grouping2;
}

bool Grouper::groupAtPosition(int32_t position, const number::impl::DecimalQuantity &value) const {
    U_ASSERT(fGrouping1 > -2);
    if (fGrouping1 == -1 || fGrouping1 == 0) {
        // Either -1 or 0 means "no grouping"
        return false;
    }
    position -= fGrouping1;
    return position >= 0 && (position % fGrouping2) == 0
           && value.getUpperDisplayMagnitude() - fGrouping1 + 1 >= fMinGrouping;
}

int16_t Grouper::getPrimary() const {
    return fGrouping1;
}

int16_t Grouper::getSecondary() const {
    return fGrouping2;
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING





using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

IntegerWidth::IntegerWidth(digits_t minInt, digits_t maxInt, bool formatFailIfMoreThanMaxDigits) {
    fUnion.minMaxInt.fMinInt = minInt;
    fUnion.minMaxInt.fMaxInt = maxInt;
    fUnion.minMaxInt.fFormatFailIfMoreThanMaxDigits = formatFailIfMoreThanMaxDigits;
}

IntegerWidth IntegerWidth::zeroFillTo(int32_t minInt) {
    if (minInt >= 0 && minInt <= kMaxIntFracSig) {
        return {static_cast<digits_t>(minInt), -1, false};
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

IntegerWidth IntegerWidth::truncateAt(int32_t maxInt) {
    if (fHasError) { return *this; }  // No-op on error
    digits_t minInt = fUnion.minMaxInt.fMinInt;
    if (maxInt >= 0 && maxInt <= kMaxIntFracSig && minInt <= maxInt) {
        return {minInt, static_cast<digits_t>(maxInt), false};
    } else if (maxInt == -1) {
        return {minInt, -1, false};
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

void IntegerWidth::apply(number::impl::DecimalQuantity& quantity, UErrorCode& status) const {
    if (fHasError) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
    } else if (fUnion.minMaxInt.fMaxInt == -1) {
        quantity.setMinInteger(fUnion.minMaxInt.fMinInt);
    } else {
        // Enforce the backwards-compatibility feature "FormatFailIfMoreThanMaxDigits"
        if (fUnion.minMaxInt.fFormatFailIfMoreThanMaxDigits &&
            fUnion.minMaxInt.fMaxInt < quantity.getMagnitude()) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
        }
        quantity.setMinInteger(fUnion.minMaxInt.fMinInt);
        quantity.applyMaxInteger(fUnion.minMaxInt.fMaxInt);
    }
}

bool IntegerWidth::operator==(const IntegerWidth& other) const {
    // Private operator==; do error and bogus checking first!
    U_ASSERT(!fHasError);
    U_ASSERT(!other.fHasError);
    U_ASSERT(!isBogus());
    U_ASSERT(!other.isBogus());
    return fUnion.minMaxInt.fMinInt == other.fUnion.minMaxInt.fMinInt &&
           fUnion.minMaxInt.fMaxInt == other.fUnion.minMaxInt.fMaxInt;
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING








#include <algorithm>



using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

namespace {

constexpr int32_t number_longnames_DNAM_INDEX = StandardPlural::Form::COUNT;
constexpr int32_t number_longnames_PER_INDEX = StandardPlural::Form::COUNT + 1;
constexpr int32_t number_longnames_ARRAY_LENGTH = StandardPlural::Form::COUNT + 2;

static int32_t number_longnames_getIndex(const char* pluralKeyword, UErrorCode& status) {
    // pluralKeyword can also be "dnam" or "per"
    if (uprv_strcmp(pluralKeyword, "dnam") == 0) {
        return number_longnames_DNAM_INDEX;
    } else if (uprv_strcmp(pluralKeyword, "per") == 0) {
        return number_longnames_PER_INDEX;
    } else {
        StandardPlural::Form plural = StandardPlural::fromString(pluralKeyword, status);
        return plural;
    }
}

static UnicodeString getWithPlural(
        const UnicodeString* strings,
        StandardPlural::Form plural,
        UErrorCode& status) {
    UnicodeString result = strings[plural];
    if (result.isBogus()) {
        result = strings[StandardPlural::Form::OTHER];
    }
    if (result.isBogus()) {
        // There should always be data in the "other" plural variant.
        status = U_INTERNAL_PROGRAM_ERROR;
    }
    return result;
}


//////////////////////////
/// BEGIN DATA LOADING ///
//////////////////////////

class PluralTableSink : public ResourceSink {
  public:
    explicit PluralTableSink(UnicodeString *outArray) : outArray(outArray) {
        // Initialize the array to bogus strings.
        for (int32_t i = 0; i < number_longnames_ARRAY_LENGTH; i++) {
            outArray[i].setToBogus();
        }
    }

    void put(const char *key, ResourceValue &value, UBool /*noFallback*/, UErrorCode &status) U_OVERRIDE {
        ResourceTable pluralsTable = value.getTable(status);
        if (U_FAILURE(status)) { return; }
        for (int32_t i = 0; pluralsTable.getKeyAndValue(i, key, value); ++i) {
            int32_t index = number_longnames_getIndex(key, status);
            if (U_FAILURE(status)) { return; }
            if (!outArray[index].isBogus()) {
                continue;
            }
            outArray[index] = value.getUnicodeString(status);
            if (U_FAILURE(status)) { return; }
        }
    }

  private:
    UnicodeString *outArray;
};

// NOTE: outArray MUST have room for all StandardPlural values.  No bounds checking is performed.

void getMeasureData(const Locale &locale, const MeasureUnit &unit, const UNumberUnitWidth &width,
                    UnicodeString *outArray, UErrorCode &status) {
    PluralTableSink sink(outArray);
    LocalUResourceBundlePointer unitsBundle(ures_open(U_ICUDATA_UNIT, locale.getName(), &status));
    if (U_FAILURE(status)) { return; }

    // Map duration-year-person, duration-week-person, etc. to duration-year, duration-week, ...
    // TODO(ICU-20400): Get duration-*-person data properly with aliases.
    StringPiece subtypeForResource;
    int32_t subtypeLen = static_cast<int32_t>(uprv_strlen(unit.getSubtype()));
    if (subtypeLen > 7 && uprv_strcmp(unit.getSubtype() + subtypeLen - 7, "-person") == 0) {
        subtypeForResource = {unit.getSubtype(), subtypeLen - 7};
    } else {
        subtypeForResource = unit.getSubtype();
    }

    CharString key;
    key.append("units", status);
    if (width == UNUM_UNIT_WIDTH_NARROW) {
        key.append("Narrow", status);
    } else if (width == UNUM_UNIT_WIDTH_SHORT) {
        key.append("Short", status);
    }
    key.append("/", status);
    key.append(unit.getType(), status);
    key.append("/", status);
    key.append(subtypeForResource, status);

    UErrorCode localStatus = U_ZERO_ERROR;
    ures_getAllItemsWithFallback(unitsBundle.getAlias(), key.data(), sink, localStatus);
    if (width == UNUM_UNIT_WIDTH_SHORT) {
        if (U_FAILURE(localStatus)) {
            status = localStatus;
        }
        return;
    }

    // TODO(ICU-13353): The fallback to short does not work in ICU4C.
    // Manually fall back to short (this is done automatically in Java).
    key.clear();
    key.append("unitsShort/", status);
    key.append(unit.getType(), status);
    key.append("/", status);
    key.append(subtypeForResource, status);
    ures_getAllItemsWithFallback(unitsBundle.getAlias(), key.data(), sink, status);
}

void getCurrencyLongNameData(const Locale &locale, const CurrencyUnit &currency, UnicodeString *outArray,
                             UErrorCode &status) {
    // In ICU4J, this method gets a CurrencyData from CurrencyData.provider.
    // TODO(ICU4J): Implement this without going through CurrencyData, like in ICU4C?
    PluralTableSink sink(outArray);
    LocalUResourceBundlePointer unitsBundle(ures_open(U_ICUDATA_CURR, locale.getName(), &status));
    if (U_FAILURE(status)) { return; }
    ures_getAllItemsWithFallback(unitsBundle.getAlias(), "CurrencyUnitPatterns", sink, status);
    if (U_FAILURE(status)) { return; }
    for (int32_t i = 0; i < StandardPlural::Form::COUNT; i++) {
        UnicodeString &pattern = outArray[i];
        if (pattern.isBogus()) {
            continue;
        }
        int32_t longNameLen = 0;
        const char16_t *longName = ucurr_getPluralName(
                currency.getISOCurrency(),
                locale.getName(),
                nullptr /* isChoiceFormat */,
                StandardPlural::getKeyword(static_cast<StandardPlural::Form>(i)),
                &longNameLen,
                &status);
        // Example pattern from data: "{0} {1}"
        // Example output after find-and-replace: "{0} US dollars"
        pattern.findAndReplace(UnicodeString(u"{1}"), UnicodeString(longName, longNameLen));
    }
}

UnicodeString getPerUnitFormat(const Locale& locale, const UNumberUnitWidth &width, UErrorCode& status) {
    LocalUResourceBundlePointer unitsBundle(ures_open(U_ICUDATA_UNIT, locale.getName(), &status));
    if (U_FAILURE(status)) { return {}; }
    CharString key;
    key.append("units", status);
    if (width == UNUM_UNIT_WIDTH_NARROW) {
        key.append("Narrow", status);
    } else if (width == UNUM_UNIT_WIDTH_SHORT) {
        key.append("Short", status);
    }
    key.append("/compound/per", status);
    int32_t len = 0;
    const UChar* ptr = ures_getStringByKeyWithFallback(unitsBundle.getAlias(), key.data(), &len, &status);
    return UnicodeString(ptr, len);
}

////////////////////////
/// END DATA LOADING ///
////////////////////////

} // namespace

LongNameHandler*
LongNameHandler::forMeasureUnit(const Locale &loc, const MeasureUnit &unitRef, const MeasureUnit &perUnit,
                                const UNumberUnitWidth &width, const PluralRules *rules,
                                const MicroPropsGenerator *parent, UErrorCode &status) {
    MeasureUnit unit = unitRef;
    if (uprv_strcmp(perUnit.getType(), "none") != 0) {
        // Compound unit: first try to simplify (e.g., meters per second is its own unit).
        bool isResolved = false;
        MeasureUnit resolved = MeasureUnit::resolveUnitPerUnit(unit, perUnit, &isResolved);
        if (isResolved) {
            unit = resolved;
        } else {
            // No simplified form is available.
            return forCompoundUnit(loc, unit, perUnit, width, rules, parent, status);
        }
    }

    auto* result = new LongNameHandler(rules, parent);
    if (result == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    UnicodeString simpleFormats[number_longnames_ARRAY_LENGTH];
    getMeasureData(loc, unit, width, simpleFormats, status);
    if (U_FAILURE(status)) { return result; }
    result->simpleFormatsToModifiers(simpleFormats, UNUM_MEASURE_UNIT_FIELD, status);
    return result;
}

LongNameHandler*
LongNameHandler::forCompoundUnit(const Locale &loc, const MeasureUnit &unit, const MeasureUnit &perUnit,
                                 const UNumberUnitWidth &width, const PluralRules *rules,
                                 const MicroPropsGenerator *parent, UErrorCode &status) {
    auto* result = new LongNameHandler(rules, parent);
    if (result == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    UnicodeString primaryData[number_longnames_ARRAY_LENGTH];
    getMeasureData(loc, unit, width, primaryData, status);
    if (U_FAILURE(status)) { return result; }
    UnicodeString secondaryData[number_longnames_ARRAY_LENGTH];
    getMeasureData(loc, perUnit, width, secondaryData, status);
    if (U_FAILURE(status)) { return result; }

    UnicodeString perUnitFormat;
    if (!secondaryData[number_longnames_PER_INDEX].isBogus()) {
        perUnitFormat = secondaryData[number_longnames_PER_INDEX];
    } else {
        UnicodeString rawPerUnitFormat = getPerUnitFormat(loc, width, status);
        if (U_FAILURE(status)) { return result; }
        // rawPerUnitFormat is something like "{0}/{1}"; we need to substitute in the secondary unit.
        SimpleFormatter compiled(rawPerUnitFormat, 2, 2, status);
        if (U_FAILURE(status)) { return result; }
        UnicodeString secondaryFormat = getWithPlural(secondaryData, StandardPlural::Form::ONE, status);
        if (U_FAILURE(status)) { return result; }
        SimpleFormatter secondaryCompiled(secondaryFormat, 1, 1, status);
        if (U_FAILURE(status)) { return result; }
        UnicodeString secondaryString = secondaryCompiled.getTextWithNoArguments().trim();
        // TODO: Why does UnicodeString need to be explicit in the following line?
        compiled.format(UnicodeString(u"{0}"), secondaryString, perUnitFormat, status);
        if (U_FAILURE(status)) { return result; }
    }
    result->multiSimpleFormatsToModifiers(primaryData, perUnitFormat, UNUM_MEASURE_UNIT_FIELD, status);
    return result;
}

UnicodeString LongNameHandler::getUnitDisplayName(
        const Locale& loc,
        const MeasureUnit& unit,
        UNumberUnitWidth width,
        UErrorCode& status) {
    if (U_FAILURE(status)) {
        return ICU_Utility::makeBogusString();
    }
    UnicodeString simpleFormats[number_longnames_ARRAY_LENGTH];
    getMeasureData(loc, unit, width, simpleFormats, status);
    return simpleFormats[number_longnames_DNAM_INDEX];
}

UnicodeString LongNameHandler::getUnitPattern(
        const Locale& loc,
        const MeasureUnit& unit,
        UNumberUnitWidth width,
        StandardPlural::Form pluralForm,
        UErrorCode& status) {
    if (U_FAILURE(status)) {
        return ICU_Utility::makeBogusString();
    }
    UnicodeString simpleFormats[number_longnames_ARRAY_LENGTH];
    getMeasureData(loc, unit, width, simpleFormats, status);
    // The above already handles fallback from other widths to short
    if (U_FAILURE(status)) {
        return ICU_Utility::makeBogusString();
    }
    // Now handle fallback from other plural forms to OTHER
    return (!(simpleFormats[pluralForm]).isBogus())? simpleFormats[pluralForm]:
            simpleFormats[StandardPlural::Form::OTHER];
}

LongNameHandler* LongNameHandler::forCurrencyLongNames(const Locale &loc, const CurrencyUnit &currency,
                                                      const PluralRules *rules,
                                                      const MicroPropsGenerator *parent,
                                                      UErrorCode &status) {
    auto* result = new LongNameHandler(rules, parent);
    if (result == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    UnicodeString simpleFormats[number_longnames_ARRAY_LENGTH];
    getCurrencyLongNameData(loc, currency, simpleFormats, status);
    if (U_FAILURE(status)) { return nullptr; }
    result->simpleFormatsToModifiers(simpleFormats, UNUM_CURRENCY_FIELD, status);
    return result;
}

void LongNameHandler::simpleFormatsToModifiers(const UnicodeString *simpleFormats, Field field,
                                               UErrorCode &status) {
    for (int32_t i = 0; i < StandardPlural::Form::COUNT; i++) {
        StandardPlural::Form plural = static_cast<StandardPlural::Form>(i);
        UnicodeString simpleFormat = getWithPlural(simpleFormats, plural, status);
        if (U_FAILURE(status)) { return; }
        SimpleFormatter compiledFormatter(simpleFormat, 0, 1, status);
        if (U_FAILURE(status)) { return; }
        fModifiers[i] = SimpleModifier(compiledFormatter, field, false, {this, SIGNUM_ZERO, plural});
    }
}

void LongNameHandler::multiSimpleFormatsToModifiers(const UnicodeString *leadFormats, UnicodeString trailFormat,
                                                    Field field, UErrorCode &status) {
    SimpleFormatter trailCompiled(trailFormat, 1, 1, status);
    if (U_FAILURE(status)) { return; }
    for (int32_t i = 0; i < StandardPlural::Form::COUNT; i++) {
        StandardPlural::Form plural = static_cast<StandardPlural::Form>(i);
        UnicodeString leadFormat = getWithPlural(leadFormats, plural, status);
        if (U_FAILURE(status)) { return; }
        UnicodeString compoundFormat;
        trailCompiled.format(leadFormat, compoundFormat, status);
        if (U_FAILURE(status)) { return; }
        SimpleFormatter compoundCompiled(compoundFormat, 0, 1, status);
        if (U_FAILURE(status)) { return; }
        fModifiers[i] = SimpleModifier(compoundCompiled, field, false, {this, SIGNUM_ZERO, plural});
    }
}

void LongNameHandler::processQuantity(DecimalQuantity &quantity, MicroProps &micros,
                                      UErrorCode &status) const {
    parent->processQuantity(quantity, micros, status);
    StandardPlural::Form pluralForm = utils::getPluralSafe(micros.rounder, rules, quantity, status);
    micros.modOuter = &fModifiers[pluralForm];
}

const Modifier* LongNameHandler::getModifier(Signum /*signum*/, StandardPlural::Form plural) const {
    return &fModifiers[plural];
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif








using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;


UnlocalizedNumberFormatter NumberPropertyMapper::create(const DecimalFormatProperties& properties,
                                                        const DecimalFormatSymbols& symbols,
                                                        DecimalFormatWarehouse& warehouse,
                                                        UErrorCode& status) {
    return NumberFormatter::with().macros(oldToNew(properties, symbols, warehouse, nullptr, status));
}

UnlocalizedNumberFormatter NumberPropertyMapper::create(const DecimalFormatProperties& properties,
                                                        const DecimalFormatSymbols& symbols,
                                                        DecimalFormatWarehouse& warehouse,
                                                        DecimalFormatProperties& exportedProperties,
                                                        UErrorCode& status) {
    return NumberFormatter::with().macros(
            oldToNew(
                    properties, symbols, warehouse, &exportedProperties, status));
}

MacroProps NumberPropertyMapper::oldToNew(const DecimalFormatProperties& properties,
                                          const DecimalFormatSymbols& symbols,
                                          DecimalFormatWarehouse& warehouse,
                                          DecimalFormatProperties* exportedProperties,
                                          UErrorCode& status) {
    MacroProps macros;
    Locale locale = symbols.getLocale();

    /////////////
    // SYMBOLS //
    /////////////

    macros.symbols.setTo(symbols);

    //////////////////
    // PLURAL RULES //
    //////////////////

    if (!properties.currencyPluralInfo.fPtr.isNull()) {
        macros.rules = properties.currencyPluralInfo.fPtr->getPluralRules();
    }

    /////////////
    // AFFIXES //
    /////////////

    AffixPatternProvider* affixProvider;
    if (properties.currencyPluralInfo.fPtr.isNull()) {
        warehouse.currencyPluralInfoAPP.setToBogus();
        warehouse.propertiesAPP.setTo(properties, status);
        affixProvider = &warehouse.propertiesAPP;
    } else {
        warehouse.currencyPluralInfoAPP.setTo(*properties.currencyPluralInfo.fPtr, properties, status);
        warehouse.propertiesAPP.setToBogus();
        affixProvider = &warehouse.currencyPluralInfoAPP;
    }
    macros.affixProvider = affixProvider;

    ///////////
    // UNITS //
    ///////////

    bool useCurrency = (
            !properties.currency.isNull() ||
            !properties.currencyPluralInfo.fPtr.isNull() ||
            !properties.currencyUsage.isNull() ||
            affixProvider->hasCurrencySign());
    CurrencyUnit currency = resolveCurrency(properties, locale, status);
    UCurrencyUsage currencyUsage = properties.currencyUsage.getOrDefault(UCURR_USAGE_STANDARD);
    if (useCurrency) {
        // NOTE: Slicing is OK.
        macros.unit = currency; // NOLINT
    }
    warehouse.currencySymbols = {currency, locale, symbols, status};
    macros.currencySymbols = &warehouse.currencySymbols;

    ///////////////////////
    // ROUNDING STRATEGY //
    ///////////////////////

    int32_t maxInt = properties.maximumIntegerDigits;
    int32_t minInt = properties.minimumIntegerDigits;
    int32_t maxFrac = properties.maximumFractionDigits;
    int32_t minFrac = properties.minimumFractionDigits;
    int32_t minSig = properties.minimumSignificantDigits;
    int32_t maxSig = properties.maximumSignificantDigits;
    double roundingIncrement = properties.roundingIncrement;
    RoundingMode roundingMode = properties.roundingMode.getOrDefault(UNUM_ROUND_HALFEVEN);
    bool explicitMinMaxFrac = minFrac != -1 || maxFrac != -1;
    bool explicitMinMaxSig = minSig != -1 || maxSig != -1;
    // Resolve min/max frac for currencies, required for the validation logic and for when minFrac or
    // maxFrac was
    // set (but not both) on a currency instance.
    // NOTE: Increments are handled in "Precision.constructCurrency()".
    if (useCurrency && (minFrac == -1 || maxFrac == -1)) {
        int32_t digits = ucurr_getDefaultFractionDigitsForUsage(
                currency.getISOCurrency(), currencyUsage, &status);
        if (minFrac == -1 && maxFrac == -1) {
            minFrac = digits;
            maxFrac = digits;
        } else if (minFrac == -1) {
            minFrac = std::min(maxFrac, digits);
        } else /* if (maxFrac == -1) */ {
            maxFrac = std::max(minFrac, digits);
        }
    }
    // Validate min/max int/frac.
    // For backwards compatibility, minimum overrides maximum if the two conflict.
    // The following logic ensures that there is always a minimum of at least one digit.
    if (minInt == 0 && maxFrac != 0) {
        // Force a digit after the decimal point.
        minFrac = minFrac <= 0 ? 1 : minFrac;
        maxFrac = maxFrac < 0 ? -1 : maxFrac < minFrac ? minFrac : maxFrac;
        minInt = 0;
        maxInt = maxInt < 0 ? -1 : maxInt > kMaxIntFracSig ? -1 : maxInt;
    } else {
        // Force a digit before the decimal point.
        minFrac = minFrac < 0 ? 0 : minFrac;
        maxFrac = maxFrac < 0 ? -1 : maxFrac < minFrac ? minFrac : maxFrac;
        minInt = minInt <= 0 ? 1 : minInt > kMaxIntFracSig ? 1 : minInt;
        maxInt = maxInt < 0 ? -1 : maxInt < minInt ? minInt : maxInt > kMaxIntFracSig ? -1 : maxInt;
    }
    Precision precision;
    if (!properties.currencyUsage.isNull()) {
        precision = Precision::constructCurrency(currencyUsage).withCurrency(currency);
    } else if (roundingIncrement != 0.0) {
        if (PatternStringUtils::ignoreRoundingIncrement(roundingIncrement, maxFrac)) {
            precision = Precision::constructFraction(minFrac, maxFrac);
        } else {
            precision = Precision::constructIncrement(roundingIncrement, minFrac);
        }
    } else if (explicitMinMaxSig) {
        minSig = minSig < 1 ? 1 : minSig > kMaxIntFracSig ? kMaxIntFracSig : minSig;
        maxSig = maxSig < 0 ? kMaxIntFracSig : maxSig < minSig ? minSig : maxSig > kMaxIntFracSig
                                                                          ? kMaxIntFracSig : maxSig;
        precision = Precision::constructSignificant(minSig, maxSig);
    } else if (explicitMinMaxFrac) {
        precision = Precision::constructFraction(minFrac, maxFrac);
    } else if (useCurrency) {
        precision = Precision::constructCurrency(currencyUsage);
    }
    if (!precision.isBogus()) {
        precision.fRoundingMode = roundingMode;
        macros.precision = precision;
    }

    ///////////////////
    // INTEGER WIDTH //
    ///////////////////

    macros.integerWidth = IntegerWidth(
            static_cast<digits_t>(minInt),
            static_cast<digits_t>(maxInt),
            properties.formatFailIfMoreThanMaxDigits);

    ///////////////////////
    // GROUPING STRATEGY //
    ///////////////////////

    macros.grouper = Grouper::forProperties(properties);

    /////////////
    // PADDING //
    /////////////

    if (properties.formatWidth > 0) {
        macros.padder = Padder::forProperties(properties);
    }

    ///////////////////////////////
    // DECIMAL MARK ALWAYS SHOWN //
    ///////////////////////////////

    macros.decimal = properties.decimalSeparatorAlwaysShown ? UNUM_DECIMAL_SEPARATOR_ALWAYS
                                                            : UNUM_DECIMAL_SEPARATOR_AUTO;

    ///////////////////////
    // SIGN ALWAYS SHOWN //
    ///////////////////////

    macros.sign = properties.signAlwaysShown ? UNUM_SIGN_ALWAYS : UNUM_SIGN_AUTO;

    /////////////////////////
    // SCIENTIFIC NOTATION //
    /////////////////////////

    if (properties.minimumExponentDigits != -1) {
        // Scientific notation is required.
        // This whole section feels like a hack, but it is needed for regression tests.
        // The mapping from property bag to scientific notation is nontrivial due to LDML rules.
        if (maxInt > 8) {
            // But #13110: The maximum of 8 digits has unknown origins and is not in the spec.
            // If maxInt is greater than 8, it is set to minInt, even if minInt is greater than 8.
            maxInt = minInt;
            macros.integerWidth = IntegerWidth::zeroFillTo(minInt).truncateAt(maxInt);
        } else if (maxInt > minInt && minInt > 1) {
            // Bug #13289: if maxInt > minInt > 1, then minInt should be 1.
            minInt = 1;
            macros.integerWidth = IntegerWidth::zeroFillTo(minInt).truncateAt(maxInt);
        }
        int engineering = maxInt < 0 ? -1 : maxInt;
        macros.notation = ScientificNotation(
                // Engineering interval:
                static_cast<int8_t>(engineering),
                // Enforce minimum integer digits (for patterns like "000.00E0"):
                (engineering == minInt),
                // Minimum exponent digits:
                static_cast<digits_t>(properties.minimumExponentDigits),
                // Exponent sign always shown:
                properties.exponentSignAlwaysShown ? UNUM_SIGN_ALWAYS : UNUM_SIGN_AUTO);
        // Scientific notation also involves overriding the rounding mode.
        // TODO: Overriding here is a bit of a hack. Should this logic go earlier?
        if (macros.precision.fType == Precision::PrecisionType::RND_FRACTION) {
            // For the purposes of rounding, get the original min/max int/frac, since the local
            // variables have been manipulated for display purposes.
            int maxInt_ = properties.maximumIntegerDigits;
            int minInt_ = properties.minimumIntegerDigits;
            int minFrac_ = properties.minimumFractionDigits;
            int maxFrac_ = properties.maximumFractionDigits;
            if (minInt_ == 0 && maxFrac_ == 0) {
                // Patterns like "#E0" and "##E0", which mean no rounding!
                macros.precision = Precision::unlimited();
            } else if (minInt_ == 0 && minFrac_ == 0) {
                // Patterns like "#.##E0" (no zeros in the mantissa), which mean round to maxFrac+1
                macros.precision = Precision::constructSignificant(1, maxFrac_ + 1);
            } else {
                int maxSig_ = minInt_ + maxFrac_;
                // Bug #20058: if maxInt_ > minInt_ > 1, then minInt_ should be 1.
                if (maxInt_ > minInt_ && minInt_ > 1) {
                    minInt_ = 1;
                }
                int minSig_ = minInt_ + minFrac_;
                // To avoid regression, maxSig is not reset when minInt_ set to 1.
                // TODO: Reset maxSig_ = 1 + minFrac_ to follow the spec.
                macros.precision = Precision::constructSignificant(minSig_, maxSig_);
            }
            macros.precision.fRoundingMode = roundingMode;
        }
    }

    //////////////////////
    // COMPACT NOTATION //
    //////////////////////

    if (!properties.compactStyle.isNull()) {
        if (properties.compactStyle.getNoError() == UNumberCompactStyle::UNUM_LONG) {
            macros.notation = Notation::compactLong();
        } else {
            macros.notation = Notation::compactShort();
        }
        // Do not forward the affix provider.
        macros.affixProvider = nullptr;
    }

    /////////////////
    // MULTIPLIERS //
    /////////////////

    macros.scale = scaleFromProperties(properties);

    //////////////////////
    // PROPERTY EXPORTS //
    //////////////////////

    if (exportedProperties != nullptr) {

        exportedProperties->currency = currency;
        exportedProperties->roundingMode = roundingMode;
        exportedProperties->minimumIntegerDigits = minInt;
        exportedProperties->maximumIntegerDigits = maxInt == -1 ? INT32_MAX : maxInt;

        Precision rounding_;
        if (precision.fType == Precision::PrecisionType::RND_CURRENCY) {
            rounding_ = precision.withCurrency(currency, status);
        } else {
            rounding_ = precision;
        }
        int minFrac_ = minFrac;
        int maxFrac_ = maxFrac;
        int minSig_ = minSig;
        int maxSig_ = maxSig;
        double increment_ = 0.0;
        if (rounding_.fType == Precision::PrecisionType::RND_FRACTION) {
            minFrac_ = rounding_.fUnion.fracSig.fMinFrac;
            maxFrac_ = rounding_.fUnion.fracSig.fMaxFrac;
        } else if (rounding_.fType == Precision::PrecisionType::RND_INCREMENT
                || rounding_.fType == Precision::PrecisionType::RND_INCREMENT_ONE
                || rounding_.fType == Precision::PrecisionType::RND_INCREMENT_FIVE) {
            increment_ = rounding_.fUnion.increment.fIncrement;
            minFrac_ = rounding_.fUnion.increment.fMinFrac;
            maxFrac_ = rounding_.fUnion.increment.fMinFrac;
        } else if (rounding_.fType == Precision::PrecisionType::RND_SIGNIFICANT) {
            minSig_ = rounding_.fUnion.fracSig.fMinSig;
            maxSig_ = rounding_.fUnion.fracSig.fMaxSig;
        }

        exportedProperties->minimumFractionDigits = minFrac_;
        exportedProperties->maximumFractionDigits = maxFrac_;
        exportedProperties->minimumSignificantDigits = minSig_;
        exportedProperties->maximumSignificantDigits = maxSig_;
        exportedProperties->roundingIncrement = increment_;
    }

    return macros;
}


void PropertiesAffixPatternProvider::setTo(const DecimalFormatProperties& properties, UErrorCode& status) {
    fBogus = false;

    // There are two ways to set affixes in DecimalFormat: via the pattern string (applyPattern), and via the
    // explicit setters (setPositivePrefix and friends).  The way to resolve the settings is as follows:
    //
    // 1) If the explicit setting is present for the field, use it.
    // 2) Otherwise, follows UTS 35 rules based on the pattern string.
    //
    // Importantly, the explicit setters affect only the one field they override.  If you set the positive
    // prefix, that should not affect the negative prefix.

    // Convenience: Extract the properties into local variables.
    // Variables are named with three chars: [p/n][p/s][o/p]
    // [p/n] => p for positive, n for negative
    // [p/s] => p for prefix, s for suffix
    // [o/p] => o for escaped custom override string, p for pattern string
    UnicodeString ppo = AffixUtils::escape(properties.positivePrefix);
    UnicodeString pso = AffixUtils::escape(properties.positiveSuffix);
    UnicodeString npo = AffixUtils::escape(properties.negativePrefix);
    UnicodeString nso = AffixUtils::escape(properties.negativeSuffix);
    const UnicodeString& ppp = properties.positivePrefixPattern;
    const UnicodeString& psp = properties.positiveSuffixPattern;
    const UnicodeString& npp = properties.negativePrefixPattern;
    const UnicodeString& nsp = properties.negativeSuffixPattern;

    if (!properties.positivePrefix.isBogus()) {
        posPrefix = ppo;
    } else if (!ppp.isBogus()) {
        posPrefix = ppp;
    } else {
        // UTS 35: Default positive prefix is empty string.
        posPrefix = u"";
    }

    if (!properties.positiveSuffix.isBogus()) {
        posSuffix = pso;
    } else if (!psp.isBogus()) {
        posSuffix = psp;
    } else {
        // UTS 35: Default positive suffix is empty string.
        posSuffix = u"";
    }

    if (!properties.negativePrefix.isBogus()) {
        negPrefix = npo;
    } else if (!npp.isBogus()) {
        negPrefix = npp;
    } else {
        // UTS 35: Default negative prefix is "-" with positive prefix.
        // Important: We prepend the "-" to the pattern, not the override!
        negPrefix = ppp.isBogus() ? u"-" : u"-" + ppp;
    }

    if (!properties.negativeSuffix.isBogus()) {
        negSuffix = nso;
    } else if (!nsp.isBogus()) {
        negSuffix = nsp;
    } else {
        // UTS 35: Default negative prefix is the positive prefix.
        negSuffix = psp.isBogus() ? u"" : psp;
    }

    // For declaring if this is a currency pattern, we need to look at the
    // original pattern, not at any user-specified overrides.
    isCurrencyPattern = (
        AffixUtils::hasCurrencySymbols(ppp, status) ||
        AffixUtils::hasCurrencySymbols(psp, status) ||
        AffixUtils::hasCurrencySymbols(npp, status) ||
        AffixUtils::hasCurrencySymbols(nsp, status));
}

char16_t PropertiesAffixPatternProvider::charAt(int flags, int i) const {
    return getStringInternal(flags).charAt(i);
}

int PropertiesAffixPatternProvider::length(int flags) const {
    return getStringInternal(flags).length();
}

UnicodeString PropertiesAffixPatternProvider::getString(int32_t flags) const {
    return getStringInternal(flags);
}

const UnicodeString& PropertiesAffixPatternProvider::getStringInternal(int32_t flags) const {
    bool prefix = (flags & AFFIX_PREFIX) != 0;
    bool negative = (flags & AFFIX_NEGATIVE_SUBPATTERN) != 0;
    if (prefix && negative) {
        return negPrefix;
    } else if (prefix) {
        return posPrefix;
    } else if (negative) {
        return negSuffix;
    } else {
        return posSuffix;
    }
}

bool PropertiesAffixPatternProvider::positiveHasPlusSign() const {
    // TODO: Change the internal APIs to propagate out the error?
    ErrorCode localStatus;
    return AffixUtils::containsType(posPrefix, TYPE_PLUS_SIGN, localStatus) ||
           AffixUtils::containsType(posSuffix, TYPE_PLUS_SIGN, localStatus);
}

bool PropertiesAffixPatternProvider::hasNegativeSubpattern() const {
    return (
        (negSuffix != posSuffix) ||
        negPrefix.tempSubString(1) != posPrefix ||
        negPrefix.charAt(0) != u'-'
    );
}

bool PropertiesAffixPatternProvider::negativeHasMinusSign() const {
    ErrorCode localStatus;
    return AffixUtils::containsType(negPrefix, TYPE_MINUS_SIGN, localStatus) ||
           AffixUtils::containsType(negSuffix, TYPE_MINUS_SIGN, localStatus);
}

bool PropertiesAffixPatternProvider::hasCurrencySign() const {
    return isCurrencyPattern;
}

bool PropertiesAffixPatternProvider::containsSymbolType(AffixPatternType type, UErrorCode& status) const {
    return AffixUtils::containsType(posPrefix, type, status) ||
           AffixUtils::containsType(posSuffix, type, status) ||
           AffixUtils::containsType(negPrefix, type, status) ||
           AffixUtils::containsType(negSuffix, type, status);
}

bool PropertiesAffixPatternProvider::hasBody() const {
    return true;
}


void CurrencyPluralInfoAffixProvider::setTo(const CurrencyPluralInfo& cpi,
                                            const DecimalFormatProperties& properties,
                                            UErrorCode& status) {
    // We need to use a PropertiesAffixPatternProvider, not the simpler version ParsedPatternInfo,
    // because user-specified affix overrides still need to work.
    fBogus = false;
    DecimalFormatProperties pluralProperties(properties);
    for (int32_t plural = 0; plural < StandardPlural::COUNT; plural++) {
        const char* keyword = StandardPlural::getKeyword(static_cast<StandardPlural::Form>(plural));
		UnicodeString kwstr = UnicodeString::fromUTF8(StringPiece(keyword));
        UnicodeString patternString;
        patternString = cpi.getCurrencyPluralPattern(kwstr, patternString);
        PatternParser::parseToExistingProperties(
                patternString,
                pluralProperties,
                IGNORE_ROUNDING_NEVER,
                status);
        affixesByPlural[plural].setTo(pluralProperties, status);
    }
}

char16_t CurrencyPluralInfoAffixProvider::charAt(int32_t flags, int32_t i) const {
    int32_t pluralOrdinal = (flags & AFFIX_PLURAL_MASK);
    return affixesByPlural[pluralOrdinal].charAt(flags, i);
}

int32_t CurrencyPluralInfoAffixProvider::length(int32_t flags) const {
    int32_t pluralOrdinal = (flags & AFFIX_PLURAL_MASK);
    return affixesByPlural[pluralOrdinal].length(flags);
}

UnicodeString CurrencyPluralInfoAffixProvider::getString(int32_t flags) const {
    int32_t pluralOrdinal = (flags & AFFIX_PLURAL_MASK);
    return affixesByPlural[pluralOrdinal].getString(flags);
}

bool CurrencyPluralInfoAffixProvider::positiveHasPlusSign() const {
    return affixesByPlural[StandardPlural::OTHER].positiveHasPlusSign();
}

bool CurrencyPluralInfoAffixProvider::hasNegativeSubpattern() const {
    return affixesByPlural[StandardPlural::OTHER].hasNegativeSubpattern();
}

bool CurrencyPluralInfoAffixProvider::negativeHasMinusSign() const {
    return affixesByPlural[StandardPlural::OTHER].negativeHasMinusSign();
}

bool CurrencyPluralInfoAffixProvider::hasCurrencySign() const {
    return affixesByPlural[StandardPlural::OTHER].hasCurrencySign();
}

bool CurrencyPluralInfoAffixProvider::containsSymbolType(AffixPatternType type, UErrorCode& status) const {
    return affixesByPlural[StandardPlural::OTHER].containsSymbolType(type, status);
}

bool CurrencyPluralInfoAffixProvider::hasBody() const {
    return affixesByPlural[StandardPlural::OTHER].hasBody();
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING






using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

namespace {

// TODO: This is copied from simpleformatter.cpp
const int32_t NUMBER_MODIFIERS_ARG_NUM_LIMIT = 0x100;

// These are the default currency spacing UnicodeSets in CLDR.
// Pre-compute them for performance.
// The Java unit test testCurrencySpacingPatternStability() will start failing if these change in CLDR.
icu::UInitOnce gDefaultCurrencySpacingInitOnce = U_INITONCE_INITIALIZER;

UnicodeSet *UNISET_DIGIT = nullptr;
UnicodeSet *UNISET_NOTS = nullptr;

UBool U_CALLCONV cleanupDefaultCurrencySpacing() {
    delete UNISET_DIGIT;
    UNISET_DIGIT = nullptr;
    delete UNISET_NOTS;
    UNISET_NOTS = nullptr;
    gDefaultCurrencySpacingInitOnce.reset();
    return TRUE;
}

void U_CALLCONV initDefaultCurrencySpacing(UErrorCode &status) {
    ucln_i18n_registerCleanup(UCLN_I18N_CURRENCY_SPACING, cleanupDefaultCurrencySpacing);
    UNISET_DIGIT = new UnicodeSet(UnicodeString(u"[:digit:]"), status);
    UNISET_NOTS = new UnicodeSet(UnicodeString(u"[:^S:]"), status);
    if (UNISET_DIGIT == nullptr || UNISET_NOTS == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    UNISET_DIGIT->freeze();
    UNISET_NOTS->freeze();
}

}  // namespace


Modifier::~Modifier() = default;

Modifier::Parameters::Parameters()
        : obj(nullptr) {}

Modifier::Parameters::Parameters(
    const ModifierStore* _obj, Signum _signum, StandardPlural::Form _plural)
        : obj(_obj), signum(_signum), plural(_plural) {}

ModifierStore::~ModifierStore() = default;

AdoptingModifierStore::~AdoptingModifierStore()  {
    for (const Modifier *mod : mods) {
        delete mod;
    }
}


int32_t ConstantAffixModifier::apply(FormattedStringBuilder &output, int leftIndex, int rightIndex,
                                     UErrorCode &status) const {
    // Insert the suffix first since inserting the prefix will change the rightIndex
    int length = output.insert(rightIndex, fSuffix, fField, status);
    length += output.insert(leftIndex, fPrefix, fField, status);
    return length;
}

int32_t ConstantAffixModifier::getPrefixLength() const {
    return fPrefix.length();
}

int32_t ConstantAffixModifier::getCodePointCount() const {
    return fPrefix.countChar32() + fSuffix.countChar32();
}

bool ConstantAffixModifier::isStrong() const {
    return fStrong;
}

bool ConstantAffixModifier::containsField(UNumberFormatFields field) const {
    (void)field;
    // This method is not currently used.
    UPRV_UNREACHABLE;
}

void ConstantAffixModifier::getParameters(Parameters& output) const {
    (void)output;
    // This method is not currently used.
    UPRV_UNREACHABLE;
}

bool ConstantAffixModifier::semanticallyEquivalent(const Modifier& other) const {
    auto* _other = dynamic_cast<const ConstantAffixModifier*>(&other);
    if (_other == nullptr) {
        return false;
    }
    return fPrefix == _other->fPrefix
        && fSuffix == _other->fSuffix
        && fField == _other->fField
        && fStrong == _other->fStrong;
}


SimpleModifier::SimpleModifier(const SimpleFormatter &simpleFormatter, Field field, bool strong)
        : SimpleModifier(simpleFormatter, field, strong, {}) {}

SimpleModifier::SimpleModifier(const SimpleFormatter &simpleFormatter, Field field, bool strong,
                               const Modifier::Parameters parameters)
        : fCompiledPattern(simpleFormatter.compiledPattern), fField(field), fStrong(strong),
          fParameters(parameters) {
    int32_t argLimit = SimpleFormatter::getArgumentLimit(
            fCompiledPattern.getBuffer(), fCompiledPattern.length());
    if (argLimit == 0) {
        // No arguments in compiled pattern
        fPrefixLength = fCompiledPattern.charAt(1) - NUMBER_MODIFIERS_ARG_NUM_LIMIT;
        U_ASSERT(2 + fPrefixLength == fCompiledPattern.length());
        // Set suffixOffset = -1 to indicate no arguments in compiled pattern.
        fSuffixOffset = -1;
        fSuffixLength = 0;
    } else {
        U_ASSERT(argLimit == 1);
        if (fCompiledPattern.charAt(1) != 0) {
            // Found prefix
            fPrefixLength = fCompiledPattern.charAt(1) - NUMBER_MODIFIERS_ARG_NUM_LIMIT;
            fSuffixOffset = 3 + fPrefixLength;
        } else {
            // No prefix
            fPrefixLength = 0;
            fSuffixOffset = 2;
        }
        if (3 + fPrefixLength < fCompiledPattern.length()) {
            // Found suffix
            fSuffixLength = fCompiledPattern.charAt(fSuffixOffset) - NUMBER_MODIFIERS_ARG_NUM_LIMIT;
        } else {
            // No suffix
            fSuffixLength = 0;
        }
    }
}

SimpleModifier::SimpleModifier()
        : fField(UNUM_FIELD_COUNT), fStrong(false), fPrefixLength(0), fSuffixLength(0) {
}

int32_t SimpleModifier::apply(FormattedStringBuilder &output, int leftIndex, int rightIndex,
                              UErrorCode &status) const {
    return formatAsPrefixSuffix(output, leftIndex, rightIndex, status);
}

int32_t SimpleModifier::getPrefixLength() const {
    return fPrefixLength;
}

int32_t SimpleModifier::getCodePointCount() const {
    int32_t count = 0;
    if (fPrefixLength > 0) {
        count += fCompiledPattern.countChar32(2, fPrefixLength);
    }
    if (fSuffixLength > 0) {
        count += fCompiledPattern.countChar32(1 + fSuffixOffset, fSuffixLength);
    }
    return count;
}

bool SimpleModifier::isStrong() const {
    return fStrong;
}

bool SimpleModifier::containsField(UNumberFormatFields field) const {
    (void)field;
    // This method is not currently used.
    UPRV_UNREACHABLE;
}

void SimpleModifier::getParameters(Parameters& output) const {
    output = fParameters;
}

bool SimpleModifier::semanticallyEquivalent(const Modifier& other) const {
    auto* _other = dynamic_cast<const SimpleModifier*>(&other);
    if (_other == nullptr) {
        return false;
    }
    if (fParameters.obj != nullptr) {
        return fParameters.obj == _other->fParameters.obj;
    }
    return fCompiledPattern == _other->fCompiledPattern
        && fField == _other->fField
        && fStrong == _other->fStrong;
}


int32_t
SimpleModifier::formatAsPrefixSuffix(FormattedStringBuilder &result, int32_t startIndex, int32_t endIndex,
                                     UErrorCode &status) const {
    if (fSuffixOffset == -1 && fPrefixLength + fSuffixLength > 0) {
        // There is no argument for the inner number; overwrite the entire segment with our string.
        return result.splice(startIndex, endIndex, fCompiledPattern, 2, 2 + fPrefixLength, fField, status);
    } else {
        if (fPrefixLength > 0) {
            result.insert(startIndex, fCompiledPattern, 2, 2 + fPrefixLength, fField, status);
        }
        if (fSuffixLength > 0) {
            result.insert(
                    endIndex + fPrefixLength,
                    fCompiledPattern,
                    1 + fSuffixOffset,
                    1 + fSuffixOffset + fSuffixLength,
                    fField,
                    status);
        }
        return fPrefixLength + fSuffixLength;
    }
}


int32_t
SimpleModifier::formatTwoArgPattern(const SimpleFormatter& compiled, FormattedStringBuilder& result,
                                    int32_t index, int32_t* outPrefixLength, int32_t* outSuffixLength,
                                    Field field, UErrorCode& status) {
    const UnicodeString& compiledPattern = compiled.compiledPattern;
    int32_t argLimit = SimpleFormatter::getArgumentLimit(
            compiledPattern.getBuffer(), compiledPattern.length());
    if (argLimit != 2) {
        status = U_INTERNAL_PROGRAM_ERROR;
        return 0;
    }
    int32_t offset = 1; // offset into compiledPattern
    int32_t length = 0; // chars added to result

    int32_t prefixLength = compiledPattern.charAt(offset);
    offset++;
    if (prefixLength < NUMBER_MODIFIERS_ARG_NUM_LIMIT) {
        // No prefix
        prefixLength = 0;
    } else {
        prefixLength -= NUMBER_MODIFIERS_ARG_NUM_LIMIT;
        result.insert(index + length, compiledPattern, offset, offset + prefixLength, field, status);
        offset += prefixLength;
        length += prefixLength;
        offset++;
    }

    int32_t infixLength = compiledPattern.charAt(offset);
    offset++;
    if (infixLength < NUMBER_MODIFIERS_ARG_NUM_LIMIT) {
        // No infix
        infixLength = 0;
    } else {
        infixLength -= NUMBER_MODIFIERS_ARG_NUM_LIMIT;
        result.insert(index + length, compiledPattern, offset, offset + infixLength, field, status);
        offset += infixLength;
        length += infixLength;
        offset++;
    }

    int32_t suffixLength;
    if (offset == compiledPattern.length()) {
        // No suffix
        suffixLength = 0;
    } else {
        suffixLength = compiledPattern.charAt(offset) -  NUMBER_MODIFIERS_ARG_NUM_LIMIT;
        offset++;
        result.insert(index + length, compiledPattern, offset, offset + suffixLength, field, status);
        length += suffixLength;
    }

    *outPrefixLength = prefixLength;
    *outSuffixLength = suffixLength;

    return length;
}


int32_t ConstantMultiFieldModifier::apply(FormattedStringBuilder &output, int leftIndex, int rightIndex,
                                          UErrorCode &status) const {
    int32_t length = output.insert(leftIndex, fPrefix, status);
    if (fOverwrite) {
        length += output.splice(
            leftIndex + length,
            rightIndex + length,
            UnicodeString(), 0, 0,
            UNUM_FIELD_COUNT, status);
    }
    length += output.insert(rightIndex + length, fSuffix, status);
    return length;
}

int32_t ConstantMultiFieldModifier::getPrefixLength() const {
    return fPrefix.length();
}

int32_t ConstantMultiFieldModifier::getCodePointCount() const {
    return fPrefix.codePointCount() + fSuffix.codePointCount();
}

bool ConstantMultiFieldModifier::isStrong() const {
    return fStrong;
}

bool ConstantMultiFieldModifier::containsField(UNumberFormatFields field) const {
    return fPrefix.containsField(field) || fSuffix.containsField(field);
}

void ConstantMultiFieldModifier::getParameters(Parameters& output) const {
    output = fParameters;
}

bool ConstantMultiFieldModifier::semanticallyEquivalent(const Modifier& other) const {
    auto* _other = dynamic_cast<const ConstantMultiFieldModifier*>(&other);
    if (_other == nullptr) {
        return false;
    }
    if (fParameters.obj != nullptr) {
        return fParameters.obj == _other->fParameters.obj;
    }
    return fPrefix.contentEquals(_other->fPrefix)
        && fSuffix.contentEquals(_other->fSuffix)
        && fOverwrite == _other->fOverwrite
        && fStrong == _other->fStrong;
}


CurrencySpacingEnabledModifier::CurrencySpacingEnabledModifier(const FormattedStringBuilder &prefix,
                                                               const FormattedStringBuilder &suffix,
                                                               bool overwrite,
                                                               bool strong,
                                                               const DecimalFormatSymbols &symbols,
                                                               UErrorCode &status)
        : ConstantMultiFieldModifier(prefix, suffix, overwrite, strong) {
    // Check for currency spacing. Do not build the UnicodeSets unless there is
    // a currency code point at a boundary.
    if (prefix.length() > 0 && prefix.fieldAt(prefix.length() - 1) == UNUM_CURRENCY_FIELD) {
        int prefixCp = prefix.getLastCodePoint();
        UnicodeSet prefixUnicodeSet = getUnicodeSet(symbols, IN_CURRENCY, PREFIX, status);
        if (prefixUnicodeSet.contains(prefixCp)) {
            fAfterPrefixUnicodeSet = getUnicodeSet(symbols, IN_NUMBER, PREFIX, status);
            fAfterPrefixUnicodeSet.freeze();
            fAfterPrefixInsert = getInsertString(symbols, PREFIX, status);
        } else {
            fAfterPrefixUnicodeSet.setToBogus();
            fAfterPrefixInsert.setToBogus();
        }
    } else {
        fAfterPrefixUnicodeSet.setToBogus();
        fAfterPrefixInsert.setToBogus();
    }
    if (suffix.length() > 0 && suffix.fieldAt(0) == UNUM_CURRENCY_FIELD) {
        int suffixCp = suffix.getLastCodePoint();
        UnicodeSet suffixUnicodeSet = getUnicodeSet(symbols, IN_CURRENCY, SUFFIX, status);
        if (suffixUnicodeSet.contains(suffixCp)) {
            fBeforeSuffixUnicodeSet = getUnicodeSet(symbols, IN_NUMBER, SUFFIX, status);
            fBeforeSuffixUnicodeSet.freeze();
            fBeforeSuffixInsert = getInsertString(symbols, SUFFIX, status);
        } else {
            fBeforeSuffixUnicodeSet.setToBogus();
            fBeforeSuffixInsert.setToBogus();
        }
    } else {
        fBeforeSuffixUnicodeSet.setToBogus();
        fBeforeSuffixInsert.setToBogus();
    }
}

int32_t CurrencySpacingEnabledModifier::apply(FormattedStringBuilder &output, int leftIndex, int rightIndex,
                                              UErrorCode &status) const {
    // Currency spacing logic
    int length = 0;
    if (rightIndex - leftIndex > 0 && !fAfterPrefixUnicodeSet.isBogus() &&
        fAfterPrefixUnicodeSet.contains(output.codePointAt(leftIndex))) {
        // TODO: Should we use the CURRENCY field here?
        length += output.insert(leftIndex, fAfterPrefixInsert, UNUM_FIELD_COUNT, status);
    }
    if (rightIndex - leftIndex > 0 && !fBeforeSuffixUnicodeSet.isBogus() &&
        fBeforeSuffixUnicodeSet.contains(output.codePointBefore(rightIndex))) {
        // TODO: Should we use the CURRENCY field here?
        length += output.insert(rightIndex + length, fBeforeSuffixInsert, UNUM_FIELD_COUNT, status);
    }

    // Call super for the remaining logic
    length += ConstantMultiFieldModifier::apply(output, leftIndex, rightIndex + length, status);
    return length;
}

int32_t
CurrencySpacingEnabledModifier::applyCurrencySpacing(FormattedStringBuilder &output, int32_t prefixStart,
                                                     int32_t prefixLen, int32_t suffixStart,
                                                     int32_t suffixLen,
                                                     const DecimalFormatSymbols &symbols,
                                                     UErrorCode &status) {
    int length = 0;
    bool hasPrefix = (prefixLen > 0);
    bool hasSuffix = (suffixLen > 0);
    bool hasNumber = (suffixStart - prefixStart - prefixLen > 0); // could be empty string
    if (hasPrefix && hasNumber) {
        length += applyCurrencySpacingAffix(output, prefixStart + prefixLen, PREFIX, symbols, status);
    }
    if (hasSuffix && hasNumber) {
        length += applyCurrencySpacingAffix(output, suffixStart + length, SUFFIX, symbols, status);
    }
    return length;
}

int32_t
CurrencySpacingEnabledModifier::applyCurrencySpacingAffix(FormattedStringBuilder &output, int32_t index,
                                                          EAffix affix,
                                                          const DecimalFormatSymbols &symbols,
                                                          UErrorCode &status) {
    // NOTE: For prefix, output.fieldAt(index-1) gets the last field type in the prefix.
    // This works even if the last code point in the prefix is 2 code units because the
    // field value gets populated to both indices in the field array.
    Field affixField = (affix == PREFIX) ? output.fieldAt(index - 1) : output.fieldAt(index);
    if (affixField != UNUM_CURRENCY_FIELD) {
        return 0;
    }
    int affixCp = (affix == PREFIX) ? output.codePointBefore(index) : output.codePointAt(index);
    UnicodeSet affixUniset = getUnicodeSet(symbols, IN_CURRENCY, affix, status);
    if (!affixUniset.contains(affixCp)) {
        return 0;
    }
    int numberCp = (affix == PREFIX) ? output.codePointAt(index) : output.codePointBefore(index);
    UnicodeSet numberUniset = getUnicodeSet(symbols, IN_NUMBER, affix, status);
    if (!numberUniset.contains(numberCp)) {
        return 0;
    }
    UnicodeString spacingString = getInsertString(symbols, affix, status);

    // NOTE: This next line *inserts* the spacing string, triggering an arraycopy.
    // It would be more efficient if this could be done before affixes were attached,
    // so that it could be prepended/appended instead of inserted.
    // However, the build code path is more efficient, and this is the most natural
    // place to put currency spacing in the non-build code path.
    // TODO: Should we use the CURRENCY field here?
    return output.insert(index, spacingString, UNUM_FIELD_COUNT, status);
}

UnicodeSet
CurrencySpacingEnabledModifier::getUnicodeSet(const DecimalFormatSymbols &symbols, EPosition position,
                                              EAffix affix, UErrorCode &status) {
    // Ensure the static defaults are initialized:
    umtx_initOnce(gDefaultCurrencySpacingInitOnce, &initDefaultCurrencySpacing, status);
    if (U_FAILURE(status)) {
        return UnicodeSet();
    }

    const UnicodeString& pattern = symbols.getPatternForCurrencySpacing(
            position == IN_CURRENCY ? UNUM_CURRENCY_MATCH : UNUM_CURRENCY_SURROUNDING_MATCH,
            affix == SUFFIX,
            status);
    if (pattern.compare(u"[:digit:]", -1) == 0) {
        return *UNISET_DIGIT;
    } else if (pattern.compare(u"[:^S:]", -1) == 0) {
        return *UNISET_NOTS;
    } else {
        return UnicodeSet(pattern, status);
    }
}

UnicodeString
CurrencySpacingEnabledModifier::getInsertString(const DecimalFormatSymbols &symbols, EAffix affix,
                                                UErrorCode &status) {
    return symbols.getPatternForCurrencySpacing(UNUM_CURRENCY_INSERT, affix == SUFFIX, status);
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif









using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;
using namespace icu::numparse::impl;


Scale::Scale(int32_t magnitude, DecNum* arbitraryToAdopt)
        : fMagnitude(magnitude), fArbitrary(arbitraryToAdopt), fError(U_ZERO_ERROR) {
    if (fArbitrary != nullptr) {
        // Attempt to convert the DecNum to a magnitude multiplier.
        fArbitrary->normalize();
        if (fArbitrary->getRawDecNumber()->digits == 1 && fArbitrary->getRawDecNumber()->lsu[0] == 1 &&
            !fArbitrary->isNegative()) {
            // Success!
            fMagnitude += fArbitrary->getRawDecNumber()->exponent;
            delete fArbitrary;
            fArbitrary = nullptr;
        }
    }
}

Scale::Scale(const Scale& other)
        : fMagnitude(other.fMagnitude), fArbitrary(nullptr), fError(other.fError) {
    if (other.fArbitrary != nullptr) {
        UErrorCode localStatus = U_ZERO_ERROR;
        fArbitrary = new DecNum(*other.fArbitrary, localStatus);
    }
}

Scale& Scale::operator=(const Scale& other) {
    fMagnitude = other.fMagnitude;
    if (other.fArbitrary != nullptr) {
        UErrorCode localStatus = U_ZERO_ERROR;
        fArbitrary = new DecNum(*other.fArbitrary, localStatus);
    } else {
        fArbitrary = nullptr;
    }
    fError = other.fError;
    return *this;
}

Scale::Scale(Scale&& src) U_NOEXCEPT
        : fMagnitude(src.fMagnitude), fArbitrary(src.fArbitrary), fError(src.fError) {
    // Take ownership away from src if necessary
    src.fArbitrary = nullptr;
}

Scale& Scale::operator=(Scale&& src) U_NOEXCEPT {
    fMagnitude = src.fMagnitude;
    if (fArbitrary != nullptr) {
        delete fArbitrary;
    }
    fArbitrary = src.fArbitrary;
    fError = src.fError;
    // Take ownership away from src if necessary
    src.fArbitrary = nullptr;
    return *this;
}

Scale::~Scale() {
    delete fArbitrary;
}


Scale Scale::none() {
    return {0, nullptr};
}

Scale Scale::powerOfTen(int32_t power) {
    return {power, nullptr};
}

Scale Scale::byDecimal(StringPiece multiplicand) {
    UErrorCode localError = U_ZERO_ERROR;
    LocalPointer<DecNum> decnum(new DecNum(), localError);
    if (U_FAILURE(localError)) {
        return {localError};
    }
    decnum->setTo(multiplicand, localError);
    if (U_FAILURE(localError)) {
        return {localError};
    }
    return {0, decnum.orphan()};
}

Scale Scale::byDouble(double multiplicand) {
    UErrorCode localError = U_ZERO_ERROR;
    LocalPointer<DecNum> decnum(new DecNum(), localError);
    if (U_FAILURE(localError)) {
        return {localError};
    }
    decnum->setTo(multiplicand, localError);
    if (U_FAILURE(localError)) {
        return {localError};
    }
    return {0, decnum.orphan()};
}

Scale Scale::byDoubleAndPowerOfTen(double multiplicand, int32_t power) {
    UErrorCode localError = U_ZERO_ERROR;
    LocalPointer<DecNum> decnum(new DecNum(), localError);
    if (U_FAILURE(localError)) {
        return {localError};
    }
    decnum->setTo(multiplicand, localError);
    if (U_FAILURE(localError)) {
        return {localError};
    }
    return {power, decnum.orphan()};
}

void Scale::applyTo(number::impl::DecimalQuantity& quantity) const {
    quantity.adjustMagnitude(fMagnitude);
    if (fArbitrary != nullptr) {
        UErrorCode localStatus = U_ZERO_ERROR;
        quantity.multiplyBy(*fArbitrary, localStatus);
    }
}

void Scale::applyReciprocalTo(number::impl::DecimalQuantity& quantity) const {
    quantity.adjustMagnitude(-fMagnitude);
    if (fArbitrary != nullptr) {
        UErrorCode localStatus = U_ZERO_ERROR;
        quantity.divideBy(*fArbitrary, localStatus);
    }
}


void
MultiplierFormatHandler::setAndChain(const Scale& multiplier, const MicroPropsGenerator* parent) {
    fMultiplier = multiplier;
    fParent = parent;
}

void MultiplierFormatHandler::processQuantity(DecimalQuantity& quantity, MicroProps& micros,
                                              UErrorCode& status) const {
    fParent->processQuantity(quantity, micros, status);
    fMultiplier.applyTo(quantity);
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING




using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;


ScientificNotation Notation::scientific() {
    // NOTE: ISO C++ does not allow C99 designated initializers.
    ScientificSettings settings;
    settings.fEngineeringInterval = 1;
    settings.fRequireMinInt = false;
    settings.fMinExponentDigits = 1;
    settings.fExponentSignDisplay = UNUM_SIGN_AUTO;
    NotationUnion union_;
    union_.scientific = settings;
    return {NTN_SCIENTIFIC, union_};
}

ScientificNotation Notation::engineering() {
    ScientificSettings settings;
    settings.fEngineeringInterval = 3;
    settings.fRequireMinInt = false;
    settings.fMinExponentDigits = 1;
    settings.fExponentSignDisplay = UNUM_SIGN_AUTO;
    NotationUnion union_;
    union_.scientific = settings;
    return {NTN_SCIENTIFIC, union_};
}

ScientificNotation::ScientificNotation(int8_t fEngineeringInterval, bool fRequireMinInt,
                                       number::impl::digits_t fMinExponentDigits,
                                       UNumberSignDisplay fExponentSignDisplay) {
    ScientificSettings settings;
    settings.fEngineeringInterval = fEngineeringInterval;
    settings.fRequireMinInt = fRequireMinInt;
    settings.fMinExponentDigits = fMinExponentDigits;
    settings.fExponentSignDisplay = fExponentSignDisplay;
    NotationUnion union_;
    union_.scientific = settings;
    *this = {NTN_SCIENTIFIC, union_};
}

Notation Notation::compactShort() {
    NotationUnion union_;
    union_.compactStyle = CompactStyle::UNUM_SHORT;
    return {NTN_COMPACT, union_};
}

Notation Notation::compactLong() {
    NotationUnion union_;
    union_.compactStyle = CompactStyle::UNUM_LONG;
    return {NTN_COMPACT, union_};
}

Notation Notation::simple() {
    return {};
}

ScientificNotation
ScientificNotation::withMinExponentDigits(int32_t minExponentDigits) const {
    if (minExponentDigits >= 1 && minExponentDigits <= kMaxIntFracSig) {
        ScientificSettings settings = fUnion.scientific;
        settings.fMinExponentDigits = static_cast<digits_t>(minExponentDigits);
        NotationUnion union_ = {settings};
        return {NTN_SCIENTIFIC, union_};
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

ScientificNotation
ScientificNotation::withExponentSignDisplay(UNumberSignDisplay exponentSignDisplay) const {
    ScientificSettings settings = fUnion.scientific;
    settings.fExponentSignDisplay = exponentSignDisplay;
    NotationUnion union_ = {settings};
    return {NTN_SCIENTIFIC, union_};
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2019 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING







U_NAMESPACE_BEGIN
namespace number {


UPRV_FORMATTED_VALUE_SUBCLASS_AUTO_IMPL(FormattedNumber)

#define UPRV_NOARG

UBool FormattedNumber::nextFieldPosition(FieldPosition& fieldPosition, UErrorCode& status) const {
    UPRV_FORMATTED_VALUE_METHOD_GUARD(FALSE)
    return fData->nextFieldPosition(fieldPosition, status);
}

void FormattedNumber::getAllFieldPositions(FieldPositionIterator& iterator, UErrorCode& status) const {
    FieldPositionIteratorHandler fpih(&iterator, status);
    getAllFieldPositionsImpl(fpih, status);
}

void FormattedNumber::toDecimalNumber(ByteSink& sink, UErrorCode& status) const {
    UPRV_FORMATTED_VALUE_METHOD_GUARD(UPRV_NOARG)
    number::impl::DecNum decnum;
    fData->quantity.toDecNum(decnum, status);
    decnum.toString(sink, status);
}

void FormattedNumber::getAllFieldPositionsImpl(FieldPositionIteratorHandler& fpih,
                                               UErrorCode& status) const {
    UPRV_FORMATTED_VALUE_METHOD_GUARD(UPRV_NOARG)
    fData->getAllFieldPositions(fpih, status);
}

void FormattedNumber::getDecimalQuantity(number::impl::DecimalQuantity& output, UErrorCode& status) const {
    UPRV_FORMATTED_VALUE_METHOD_GUARD(UPRV_NOARG)
    output = fData->quantity;
}


impl::UFormattedNumberData::~UFormattedNumberData() = default;


} // namespace number
U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING






using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

namespace {

int32_t
addPaddingHelper(UChar32 paddingCp, int32_t requiredPadding, FormattedStringBuilder &string, int32_t index,
                 UErrorCode &status) {
    for (int32_t i = 0; i < requiredPadding; i++) {
        // TODO: If appending to the end, this will cause actual insertion operations. Improve.
        string.insertCodePoint(index, paddingCp, UNUM_FIELD_COUNT, status);
    }
    return U16_LENGTH(paddingCp) * requiredPadding;
}

}

Padder::Padder(UChar32 cp, int32_t width, UNumberFormatPadPosition position) : fWidth(width) {
    // TODO(13034): Consider making this a string instead of code point.
    fUnion.padding.fCp = cp;
    fUnion.padding.fPosition = position;
}

Padder::Padder(int32_t width) : fWidth(width) {}

Padder Padder::none() {
    return {-1};
}

Padder Padder::codePoints(UChar32 cp, int32_t targetWidth, UNumberFormatPadPosition position) {
    // TODO: Validate the code point?
    if (targetWidth >= 0) {
        return {cp, targetWidth, position};
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

Padder Padder::forProperties(const DecimalFormatProperties& properties) {
    UChar32 padCp;
    if (properties.padString.length() > 0) {
        padCp = properties.padString.char32At(0);
    } else {
        padCp = kFallbackPaddingString[0];
    }
    return {padCp, properties.formatWidth, properties.padPosition.getOrDefault(UNUM_PAD_BEFORE_PREFIX)};
}

int32_t Padder::padAndApply(const Modifier &mod1, const Modifier &mod2,
                            FormattedStringBuilder &string, int32_t leftIndex, int32_t rightIndex,
                            UErrorCode &status) const {
    int32_t modLength = mod1.getCodePointCount() + mod2.getCodePointCount();
    int32_t requiredPadding = fWidth - modLength - string.codePointCount();
    U_ASSERT(leftIndex == 0 &&
             rightIndex == string.length()); // fix the previous line to remove this assertion

    int length = 0;
    if (requiredPadding <= 0) {
        // Padding is not required.
        length += mod1.apply(string, leftIndex, rightIndex, status);
        length += mod2.apply(string, leftIndex, rightIndex + length, status);
        return length;
    }

    PadPosition position = fUnion.padding.fPosition;
    UChar32 paddingCp = fUnion.padding.fCp;
    if (position == UNUM_PAD_AFTER_PREFIX) {
        length += addPaddingHelper(paddingCp, requiredPadding, string, leftIndex, status);
    } else if (position == UNUM_PAD_BEFORE_SUFFIX) {
        length += addPaddingHelper(paddingCp, requiredPadding, string, rightIndex + length, status);
    }
    length += mod1.apply(string, leftIndex, rightIndex + length, status);
    length += mod2.apply(string, leftIndex, rightIndex + length, status);
    if (position == UNUM_PAD_BEFORE_PREFIX) {
        length += addPaddingHelper(paddingCp, requiredPadding, string, leftIndex, status);
    } else if (position == UNUM_PAD_AFTER_SUFFIX) {
        length += addPaddingHelper(paddingCp, requiredPadding, string, rightIndex + length, status);
    }

    return length;
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING








using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;


AffixPatternProvider::~AffixPatternProvider() = default;


MutablePatternModifier::MutablePatternModifier(bool isStrong)
        : fStrong(isStrong) {}

void MutablePatternModifier::setPatternInfo(const AffixPatternProvider* patternInfo, Field field) {
    fPatternInfo = patternInfo;
    fField = field;
}

void MutablePatternModifier::setPatternAttributes(UNumberSignDisplay signDisplay, bool perMille) {
    fSignDisplay = signDisplay;
    fPerMilleReplacesPercent = perMille;
}

void MutablePatternModifier::setSymbols(const DecimalFormatSymbols* symbols,
                                        const CurrencySymbols* currencySymbols,
                                        const UNumberUnitWidth unitWidth, const PluralRules* rules) {
    U_ASSERT((rules != nullptr) == needsPlurals());
    fSymbols = symbols;
    fCurrencySymbols = currencySymbols;
    fUnitWidth = unitWidth;
    fRules = rules;
}

void MutablePatternModifier::setNumberProperties(Signum signum, StandardPlural::Form plural) {
    fSignum = signum;
    fPlural = plural;
}

bool MutablePatternModifier::needsPlurals() const {
    UErrorCode statusLocal = U_ZERO_ERROR;
    return fPatternInfo->containsSymbolType(AffixPatternType::TYPE_CURRENCY_TRIPLE, statusLocal);
    // Silently ignore any error codes.
}

ImmutablePatternModifier* MutablePatternModifier::createImmutable(UErrorCode& status) {
    return createImmutableAndChain(nullptr, status);
}

ImmutablePatternModifier*
MutablePatternModifier::createImmutableAndChain(const MicroPropsGenerator* parent, UErrorCode& status) {

    // TODO: Move StandardPlural VALUES to standardplural.h
    static const StandardPlural::Form STANDARD_PLURAL_VALUES[] = {
            StandardPlural::Form::ZERO,
            StandardPlural::Form::ONE,
            StandardPlural::Form::TWO,
            StandardPlural::Form::FEW,
            StandardPlural::Form::MANY,
            StandardPlural::Form::OTHER};

    auto pm = new AdoptingModifierStore();
    if (pm == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }

    if (needsPlurals()) {
        // Slower path when we require the plural keyword.
        for (StandardPlural::Form plural : STANDARD_PLURAL_VALUES) {
            setNumberProperties(SIGNUM_POS, plural);
            pm->adoptModifier(SIGNUM_POS, plural, createConstantModifier(status));
            setNumberProperties(SIGNUM_ZERO, plural);
            pm->adoptModifier(SIGNUM_ZERO, plural, createConstantModifier(status));
            setNumberProperties(SIGNUM_NEG, plural);
            pm->adoptModifier(SIGNUM_NEG, plural, createConstantModifier(status));
        }
        if (U_FAILURE(status)) {
            delete pm;
            return nullptr;
        }
        return new ImmutablePatternModifier(pm, fRules, parent);  // adopts pm
    } else {
        // Faster path when plural keyword is not needed.
        setNumberProperties(SIGNUM_POS, StandardPlural::Form::COUNT);
        pm->adoptModifierWithoutPlural(SIGNUM_POS, createConstantModifier(status));
        setNumberProperties(SIGNUM_ZERO, StandardPlural::Form::COUNT);
        pm->adoptModifierWithoutPlural(SIGNUM_ZERO, createConstantModifier(status));
        setNumberProperties(SIGNUM_NEG, StandardPlural::Form::COUNT);
        pm->adoptModifierWithoutPlural(SIGNUM_NEG, createConstantModifier(status));
        if (U_FAILURE(status)) {
            delete pm;
            return nullptr;
        }
        return new ImmutablePatternModifier(pm, nullptr, parent);  // adopts pm
    }
}

ConstantMultiFieldModifier* MutablePatternModifier::createConstantModifier(UErrorCode& status) {
    FormattedStringBuilder a;
    FormattedStringBuilder b;
    insertPrefix(a, 0, status);
    insertSuffix(b, 0, status);
    if (fPatternInfo->hasCurrencySign()) {
        return new CurrencySpacingEnabledModifier(
                a, b, !fPatternInfo->hasBody(), fStrong, *fSymbols, status);
    } else {
        return new ConstantMultiFieldModifier(a, b, !fPatternInfo->hasBody(), fStrong);
    }
}

ImmutablePatternModifier::ImmutablePatternModifier(AdoptingModifierStore* pm, const PluralRules* rules,
                                                   const MicroPropsGenerator* parent)
        : pm(pm), rules(rules), parent(parent) {}

void ImmutablePatternModifier::processQuantity(DecimalQuantity& quantity, MicroProps& micros,
                                               UErrorCode& status) const {
    parent->processQuantity(quantity, micros, status);
    applyToMicros(micros, quantity, status);
}

void ImmutablePatternModifier::applyToMicros(
        MicroProps& micros, const DecimalQuantity& quantity, UErrorCode& status) const {
    if (rules == nullptr) {
        micros.modMiddle = pm->getModifierWithoutPlural(quantity.signum());
    } else {
        StandardPlural::Form pluralForm = utils::getPluralSafe(micros.rounder, rules, quantity, status);
        micros.modMiddle = pm->getModifier(quantity.signum(), pluralForm);
    }
}

const Modifier* ImmutablePatternModifier::getModifier(Signum signum, StandardPlural::Form plural) const {
    if (rules == nullptr) {
        return pm->getModifierWithoutPlural(signum);
    } else {
        return pm->getModifier(signum, plural);
    }
}


/** Used by the unsafe code path. */
MicroPropsGenerator& MutablePatternModifier::addToChain(const MicroPropsGenerator* parent) {
    fParent = parent;
    return *this;
}

void MutablePatternModifier::processQuantity(DecimalQuantity& fq, MicroProps& micros,
                                             UErrorCode& status) const {
    fParent->processQuantity(fq, micros, status);
    // The unsafe code path performs self-mutation, so we need a const_cast.
    // This method needs to be const because it overrides a const method in the parent class.
    auto nonConstThis = const_cast<MutablePatternModifier*>(this);
    if (needsPlurals()) {
        StandardPlural::Form pluralForm = utils::getPluralSafe(micros.rounder, fRules, fq, status);
        nonConstThis->setNumberProperties(fq.signum(), pluralForm);
    } else {
        nonConstThis->setNumberProperties(fq.signum(), StandardPlural::Form::COUNT);
    }
    micros.modMiddle = this;
}

int32_t MutablePatternModifier::apply(FormattedStringBuilder& output, int32_t leftIndex, int32_t rightIndex,
                                      UErrorCode& status) const {
    // The unsafe code path performs self-mutation, so we need a const_cast.
    // This method needs to be const because it overrides a const method in the parent class.
    auto nonConstThis = const_cast<MutablePatternModifier*>(this);
    int32_t prefixLen = nonConstThis->insertPrefix(output, leftIndex, status);
    int32_t suffixLen = nonConstThis->insertSuffix(output, rightIndex + prefixLen, status);
    // If the pattern had no decimal stem body (like #,##0.00), overwrite the value.
    int32_t overwriteLen = 0;
    if (!fPatternInfo->hasBody()) {
        overwriteLen = output.splice(
                leftIndex + prefixLen,
                rightIndex + prefixLen,
                UnicodeString(),
                0,
                0,
                UNUM_FIELD_COUNT,
                status);
    }
    CurrencySpacingEnabledModifier::applyCurrencySpacing(
            output,
            leftIndex,
            prefixLen,
            rightIndex + overwriteLen + prefixLen,
            suffixLen,
            *fSymbols,
            status);
    return prefixLen + overwriteLen + suffixLen;
}

int32_t MutablePatternModifier::getPrefixLength() const {
    // The unsafe code path performs self-mutation, so we need a const_cast.
    // This method needs to be const because it overrides a const method in the parent class.
    auto nonConstThis = const_cast<MutablePatternModifier*>(this);

    // Enter and exit CharSequence Mode to get the length.
    UErrorCode status = U_ZERO_ERROR; // status fails only with an iilegal argument exception
    nonConstThis->prepareAffix(true);
    int result = AffixUtils::unescapedCodePointCount(currentAffix, *this, status);  // prefix length
    return result;
}

int32_t MutablePatternModifier::getCodePointCount() const {
    // The unsafe code path performs self-mutation, so we need a const_cast.
    // This method needs to be const because it overrides a const method in the parent class.
    auto nonConstThis = const_cast<MutablePatternModifier*>(this);

    // Render the affixes to get the length
    UErrorCode status = U_ZERO_ERROR; // status fails only with an iilegal argument exception
    nonConstThis->prepareAffix(true);
    int result = AffixUtils::unescapedCodePointCount(currentAffix, *this, status);  // prefix length
    nonConstThis->prepareAffix(false);
    result += AffixUtils::unescapedCodePointCount(currentAffix, *this, status);  // suffix length
    return result;
}

bool MutablePatternModifier::isStrong() const {
    return fStrong;
}

bool MutablePatternModifier::containsField(UNumberFormatFields field) const {
    (void)field;
    // This method is not currently used.
    UPRV_UNREACHABLE;
}

void MutablePatternModifier::getParameters(Parameters& output) const {
    (void)output;
    // This method is not currently used.
    UPRV_UNREACHABLE;
}

bool MutablePatternModifier::semanticallyEquivalent(const Modifier& other) const {
    (void)other;
    // This method is not currently used.
    UPRV_UNREACHABLE;
}

int32_t MutablePatternModifier::insertPrefix(FormattedStringBuilder& sb, int position, UErrorCode& status) {
    prepareAffix(true);
    int32_t length = AffixUtils::unescape(currentAffix, sb, position, *this, fField, status);
    return length;
}

int32_t MutablePatternModifier::insertSuffix(FormattedStringBuilder& sb, int position, UErrorCode& status) {
    prepareAffix(false);
    int32_t length = AffixUtils::unescape(currentAffix, sb, position, *this, fField, status);
    return length;
}

/** This method contains the heart of the logic for rendering LDML affix strings. */
void MutablePatternModifier::prepareAffix(bool isPrefix) {
    PatternStringUtils::patternInfoToStringBuilder(
            *fPatternInfo, isPrefix, fSignum, fSignDisplay, fPlural, fPerMilleReplacesPercent, currentAffix);
}

UnicodeString MutablePatternModifier::getSymbol(AffixPatternType type) const {
    UErrorCode localStatus = U_ZERO_ERROR;
    switch (type) {
        case AffixPatternType::TYPE_MINUS_SIGN:
            return fSymbols->getSymbol(DecimalFormatSymbols::ENumberFormatSymbol::kMinusSignSymbol);
        case AffixPatternType::TYPE_PLUS_SIGN:
            return fSymbols->getSymbol(DecimalFormatSymbols::ENumberFormatSymbol::kPlusSignSymbol);
        case AffixPatternType::TYPE_PERCENT:
            return fSymbols->getSymbol(DecimalFormatSymbols::ENumberFormatSymbol::kPercentSymbol);
        case AffixPatternType::TYPE_PERMILLE:
            return fSymbols->getSymbol(DecimalFormatSymbols::ENumberFormatSymbol::kPerMillSymbol);
        case AffixPatternType::TYPE_CURRENCY_SINGLE: {
            // UnitWidth ISO and HIDDEN overrides the singular currency symbol.
            if (fUnitWidth == UNumberUnitWidth::UNUM_UNIT_WIDTH_ISO_CODE) {
                return fCurrencySymbols->getIntlCurrencySymbol(localStatus);
            } else if (fUnitWidth == UNumberUnitWidth::UNUM_UNIT_WIDTH_HIDDEN) {
                return UnicodeString();
            } else if (fUnitWidth == UNumberUnitWidth::UNUM_UNIT_WIDTH_NARROW) {
                return fCurrencySymbols->getNarrowCurrencySymbol(localStatus);
            } else {
                return fCurrencySymbols->getCurrencySymbol(localStatus);
            }
        }
        case AffixPatternType::TYPE_CURRENCY_DOUBLE:
            return fCurrencySymbols->getIntlCurrencySymbol(localStatus);
        case AffixPatternType::TYPE_CURRENCY_TRIPLE:
            // NOTE: This is the code path only for patterns containing "¤¤¤".
            // Plural currencies set via the API are formatted in LongNameHandler.
            // This code path is used by DecimalFormat via CurrencyPluralInfo.
            U_ASSERT(fPlural != StandardPlural::Form::COUNT);
            return fCurrencySymbols->getPluralName(fPlural, localStatus);
        case AffixPatternType::TYPE_CURRENCY_QUAD:
            return UnicodeString(u"\uFFFD");
        case AffixPatternType::TYPE_CURRENCY_QUINT:
            return UnicodeString(u"\uFFFD");
        default:
            UPRV_UNREACHABLE;
    }
}

UnicodeString MutablePatternModifier::toUnicodeString() const {
    // Never called by AffixUtils
    UPRV_UNREACHABLE;
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif

#ifndef UNISTR_FROM_CHAR_EXPLICIT
#define UNISTR_FROM_CHAR_EXPLICIT
#endif








using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;


void PatternParser::parseToPatternInfo(const UnicodeString& patternString, ParsedPatternInfo& patternInfo,
                                       UErrorCode& status) {
    patternInfo.consumePattern(patternString, status);
}

DecimalFormatProperties
PatternParser::parseToProperties(const UnicodeString& pattern, IgnoreRounding ignoreRounding,
                                 UErrorCode& status) {
    DecimalFormatProperties properties;
    parseToExistingPropertiesImpl(pattern, properties, ignoreRounding, status);
    return properties;
}

DecimalFormatProperties PatternParser::parseToProperties(const UnicodeString& pattern,
                                                         UErrorCode& status) {
    return parseToProperties(pattern, IGNORE_ROUNDING_NEVER, status);
}

void
PatternParser::parseToExistingProperties(const UnicodeString& pattern, DecimalFormatProperties& properties,
                                         IgnoreRounding ignoreRounding, UErrorCode& status) {
    parseToExistingPropertiesImpl(pattern, properties, ignoreRounding, status);
}


char16_t ParsedPatternInfo::charAt(int32_t flags, int32_t index) const {
    const Endpoints& endpoints = getEndpoints(flags);
    if (index < 0 || index >= endpoints.end - endpoints.start) {
        UPRV_UNREACHABLE;
    }
    return pattern.charAt(endpoints.start + index);
}

int32_t ParsedPatternInfo::length(int32_t flags) const {
    return getLengthFromEndpoints(getEndpoints(flags));
}

int32_t ParsedPatternInfo::getLengthFromEndpoints(const Endpoints& endpoints) {
    return endpoints.end - endpoints.start;
}

UnicodeString ParsedPatternInfo::getString(int32_t flags) const {
    const Endpoints& endpoints = getEndpoints(flags);
    if (endpoints.start == endpoints.end) {
        return UnicodeString();
    }
    // Create a new UnicodeString
    return UnicodeString(pattern, endpoints.start, endpoints.end - endpoints.start);
}

const Endpoints& ParsedPatternInfo::getEndpoints(int32_t flags) const {
    bool prefix = (flags & AFFIX_PREFIX) != 0;
    bool isNegative = (flags & AFFIX_NEGATIVE_SUBPATTERN) != 0;
    bool padding = (flags & AFFIX_PADDING) != 0;
    if (isNegative && padding) {
        return negative.paddingEndpoints;
    } else if (padding) {
        return positive.paddingEndpoints;
    } else if (prefix && isNegative) {
        return negative.prefixEndpoints;
    } else if (prefix) {
        return positive.prefixEndpoints;
    } else if (isNegative) {
        return negative.suffixEndpoints;
    } else {
        return positive.suffixEndpoints;
    }
}

bool ParsedPatternInfo::positiveHasPlusSign() const {
    return positive.hasPlusSign;
}

bool ParsedPatternInfo::hasNegativeSubpattern() const {
    return fHasNegativeSubpattern;
}

bool ParsedPatternInfo::negativeHasMinusSign() const {
    return negative.hasMinusSign;
}

bool ParsedPatternInfo::hasCurrencySign() const {
    return positive.hasCurrencySign || (fHasNegativeSubpattern && negative.hasCurrencySign);
}

bool ParsedPatternInfo::containsSymbolType(AffixPatternType type, UErrorCode& status) const {
    return AffixUtils::containsType(pattern, type, status);
}

bool ParsedPatternInfo::hasBody() const {
    return positive.integerTotal > 0;
}

/////////////////////////////////////////////////////
/// BEGIN RECURSIVE DESCENT PARSER IMPLEMENTATION ///
/////////////////////////////////////////////////////

UChar32 ParsedPatternInfo::ParserState::peek() {
    if (offset == pattern.length()) {
        return -1;
    } else {
        return pattern.char32At(offset);
    }
}

UChar32 ParsedPatternInfo::ParserState::next() {
    int codePoint = peek();
    offset += U16_LENGTH(codePoint);
    return codePoint;
}

void ParsedPatternInfo::consumePattern(const UnicodeString& patternString, UErrorCode& status) {
    if (U_FAILURE(status)) { return; }
    this->pattern = patternString;

    // This class is not intended for writing twice!
    // Use move assignment to overwrite instead.
    U_ASSERT(state.offset == 0);

    // pattern := subpattern (';' subpattern)?
    currentSubpattern = &positive;
    consumeSubpattern(status);
    if (U_FAILURE(status)) { return; }
    if (state.peek() == u';') {
        state.next(); // consume the ';'
        // Don't consume the negative subpattern if it is empty (trailing ';')
        if (state.peek() != -1) {
            fHasNegativeSubpattern = true;
            currentSubpattern = &negative;
            consumeSubpattern(status);
            if (U_FAILURE(status)) { return; }
        }
    }
    if (state.peek() != -1) {
        state.toParseException(u"Found unquoted special character");
        status = U_UNQUOTED_SPECIAL;
    }
}

void ParsedPatternInfo::consumeSubpattern(UErrorCode& status) {
    // subpattern := literals? number exponent? literals?
    consumePadding(PadPosition::UNUM_PAD_BEFORE_PREFIX, status);
    if (U_FAILURE(status)) { return; }
    consumeAffix(currentSubpattern->prefixEndpoints, status);
    if (U_FAILURE(status)) { return; }
    consumePadding(PadPosition::UNUM_PAD_AFTER_PREFIX, status);
    if (U_FAILURE(status)) { return; }
    consumeFormat(status);
    if (U_FAILURE(status)) { return; }
    consumeExponent(status);
    if (U_FAILURE(status)) { return; }
    consumePadding(PadPosition::UNUM_PAD_BEFORE_SUFFIX, status);
    if (U_FAILURE(status)) { return; }
    consumeAffix(currentSubpattern->suffixEndpoints, status);
    if (U_FAILURE(status)) { return; }
    consumePadding(PadPosition::UNUM_PAD_AFTER_SUFFIX, status);
    if (U_FAILURE(status)) { return; }
}

void ParsedPatternInfo::consumePadding(PadPosition paddingLocation, UErrorCode& status) {
    if (state.peek() != u'*') {
        return;
    }
    if (currentSubpattern->hasPadding) {
        state.toParseException(u"Cannot have multiple pad specifiers");
        status = U_MULTIPLE_PAD_SPECIFIERS;
        return;
    }
    currentSubpattern->paddingLocation = paddingLocation;
    currentSubpattern->hasPadding = true;
    state.next(); // consume the '*'
    currentSubpattern->paddingEndpoints.start = state.offset;
    consumeLiteral(status);
    currentSubpattern->paddingEndpoints.end = state.offset;
}

void ParsedPatternInfo::consumeAffix(Endpoints& endpoints, UErrorCode& status) {
    // literals := { literal }
    endpoints.start = state.offset;
    while (true) {
        switch (state.peek()) {
            case u'#':
            case u'@':
            case u';':
            case u'*':
            case u'.':
            case u',':
            case u'0':
            case u'1':
            case u'2':
            case u'3':
            case u'4':
            case u'5':
            case u'6':
            case u'7':
            case u'8':
            case u'9':
            case -1:
                // Characters that cannot appear unquoted in a literal
                // break outer;
                goto after_outer;

            case u'%':
                currentSubpattern->hasPercentSign = true;
                break;

            case u'\x2030':
                currentSubpattern->hasPerMilleSign = true;
                break;

            case u'\xa4':
                currentSubpattern->hasCurrencySign = true;
                break;

            case u'-':
                currentSubpattern->hasMinusSign = true;
                break;

            case u'+':
                currentSubpattern->hasPlusSign = true;
                break;

            default:
                break;
        }
        consumeLiteral(status);
        if (U_FAILURE(status)) { return; }
    }
    after_outer:
    endpoints.end = state.offset;
}

void ParsedPatternInfo::consumeLiteral(UErrorCode& status) {
    if (state.peek() == -1) {
        state.toParseException(u"Expected unquoted literal but found EOL");
        status = U_PATTERN_SYNTAX_ERROR;
        return;
    } else if (state.peek() == u'\'') {
        state.next(); // consume the starting quote
        while (state.peek() != u'\'') {
            if (state.peek() == -1) {
                state.toParseException(u"Expected quoted literal but found EOL");
                status = U_PATTERN_SYNTAX_ERROR;
                return;
            } else {
                state.next(); // consume a quoted character
            }
        }
        state.next(); // consume the ending quote
    } else {
        // consume a non-quoted literal character
        state.next();
    }
}

void ParsedPatternInfo::consumeFormat(UErrorCode& status) {
    consumeIntegerFormat(status);
    if (U_FAILURE(status)) { return; }
    if (state.peek() == u'.') {
        state.next(); // consume the decimal point
        currentSubpattern->hasDecimal = true;
        currentSubpattern->widthExceptAffixes += 1;
        consumeFractionFormat(status);
        if (U_FAILURE(status)) { return; }
    }
}

void ParsedPatternInfo::consumeIntegerFormat(UErrorCode& status) {
    // Convenience reference:
    ParsedSubpatternInfo& result = *currentSubpattern;

    while (true) {
        switch (state.peek()) {
            case u',':
                result.widthExceptAffixes += 1;
                result.groupingSizes <<= 16;
                break;

            case u'#':
                if (result.integerNumerals > 0) {
                    state.toParseException(u"# cannot follow 0 before decimal point");
                    status = U_UNEXPECTED_TOKEN;
                    return;
                }
                result.widthExceptAffixes += 1;
                result.groupingSizes += 1;
                if (result.integerAtSigns > 0) {
                    result.integerTrailingHashSigns += 1;
                } else {
                    result.integerLeadingHashSigns += 1;
                }
                result.integerTotal += 1;
                break;

            case u'@':
                if (result.integerNumerals > 0) {
                    state.toParseException(u"Cannot mix 0 and @");
                    status = U_UNEXPECTED_TOKEN;
                    return;
                }
                if (result.integerTrailingHashSigns > 0) {
                    state.toParseException(u"Cannot nest # inside of a run of @");
                    status = U_UNEXPECTED_TOKEN;
                    return;
                }
                result.widthExceptAffixes += 1;
                result.groupingSizes += 1;
                result.integerAtSigns += 1;
                result.integerTotal += 1;
                break;

            case u'0':
            case u'1':
            case u'2':
            case u'3':
            case u'4':
            case u'5':
            case u'6':
            case u'7':
            case u'8':
            case u'9':
                if (result.integerAtSigns > 0) {
                    state.toParseException(u"Cannot mix @ and 0");
                    status = U_UNEXPECTED_TOKEN;
                    return;
                }
                result.widthExceptAffixes += 1;
                result.groupingSizes += 1;
                result.integerNumerals += 1;
                result.integerTotal += 1;
                if (!result.rounding.isZeroish() || state.peek() != u'0') {
                    result.rounding.appendDigit(static_cast<int8_t>(state.peek() - u'0'), 0, true);
                }
                break;

            default:
                goto after_outer;
        }
        state.next(); // consume the symbol
    }

    after_outer:
    // Disallow patterns with a trailing ',' or with two ',' next to each other
    auto grouping1 = static_cast<int16_t> (result.groupingSizes & 0xffff);
    auto grouping2 = static_cast<int16_t> ((result.groupingSizes >> 16) & 0xffff);
    auto grouping3 = static_cast<int16_t> ((result.groupingSizes >> 32) & 0xffff);
    if (grouping1 == 0 && grouping2 != -1) {
        state.toParseException(u"Trailing grouping separator is invalid");
        status = U_UNEXPECTED_TOKEN;
        return;
    }
    if (grouping2 == 0 && grouping3 != -1) {
        state.toParseException(u"Grouping width of zero is invalid");
        status = U_PATTERN_SYNTAX_ERROR;
        return;
    }
}

void ParsedPatternInfo::consumeFractionFormat(UErrorCode& status) {
    // Convenience reference:
    ParsedSubpatternInfo& result = *currentSubpattern;

    int32_t zeroCounter = 0;
    while (true) {
        switch (state.peek()) {
            case u'#':
                result.widthExceptAffixes += 1;
                result.fractionHashSigns += 1;
                result.fractionTotal += 1;
                zeroCounter++;
                break;

            case u'0':
            case u'1':
            case u'2':
            case u'3':
            case u'4':
            case u'5':
            case u'6':
            case u'7':
            case u'8':
            case u'9':
                if (result.fractionHashSigns > 0) {
                    state.toParseException(u"0 cannot follow # after decimal point");
                    status = U_UNEXPECTED_TOKEN;
                    return;
                }
                result.widthExceptAffixes += 1;
                result.fractionNumerals += 1;
                result.fractionTotal += 1;
                if (state.peek() == u'0') {
                    zeroCounter++;
                } else {
                    result.rounding
                            .appendDigit(static_cast<int8_t>(state.peek() - u'0'), zeroCounter, false);
                    zeroCounter = 0;
                }
                break;

            default:
                return;
        }
        state.next(); // consume the symbol
    }
}

void ParsedPatternInfo::consumeExponent(UErrorCode& status) {
    // Convenience reference:
    ParsedSubpatternInfo& result = *currentSubpattern;

    if (state.peek() != u'E') {
        return;
    }
    if ((result.groupingSizes & 0xffff0000L) != 0xffff0000L) {
        state.toParseException(u"Cannot have grouping separator in scientific notation");
        status = U_MALFORMED_EXPONENTIAL_PATTERN;
        return;
    }
    state.next(); // consume the E
    result.widthExceptAffixes++;
    if (state.peek() == u'+') {
        state.next(); // consume the +
        result.exponentHasPlusSign = true;
        result.widthExceptAffixes++;
    }
    while (state.peek() == u'0') {
        state.next(); // consume the 0
        result.exponentZeros += 1;
        result.widthExceptAffixes++;
    }
}

///////////////////////////////////////////////////
/// END RECURSIVE DESCENT PARSER IMPLEMENTATION ///
///////////////////////////////////////////////////

void PatternParser::parseToExistingPropertiesImpl(const UnicodeString& pattern,
                                                  DecimalFormatProperties& properties,
                                                  IgnoreRounding ignoreRounding, UErrorCode& status) {
    if (pattern.length() == 0) {
        // Backwards compatibility requires that we reset to the default values.
        // TODO: Only overwrite the properties that "saveToProperties" normally touches?
        properties.clear();
        return;
    }

    ParsedPatternInfo patternInfo;
    parseToPatternInfo(pattern, patternInfo, status);
    if (U_FAILURE(status)) { return; }
    patternInfoToProperties(properties, patternInfo, ignoreRounding, status);
}

void
PatternParser::patternInfoToProperties(DecimalFormatProperties& properties, ParsedPatternInfo& patternInfo,
                                       IgnoreRounding _ignoreRounding, UErrorCode& status) {
    // Translate from PatternParseResult to Properties.
    // Note that most data from "negative" is ignored per the specification of DecimalFormat.

    const ParsedSubpatternInfo& positive = patternInfo.positive;

    bool ignoreRounding;
    if (_ignoreRounding == IGNORE_ROUNDING_NEVER) {
        ignoreRounding = false;
    } else if (_ignoreRounding == IGNORE_ROUNDING_IF_CURRENCY) {
        ignoreRounding = positive.hasCurrencySign;
    } else {
        U_ASSERT(_ignoreRounding == IGNORE_ROUNDING_ALWAYS);
        ignoreRounding = true;
    }

    // Grouping settings
    auto grouping1 = static_cast<int16_t> (positive.groupingSizes & 0xffff);
    auto grouping2 = static_cast<int16_t> ((positive.groupingSizes >> 16) & 0xffff);
    auto grouping3 = static_cast<int16_t> ((positive.groupingSizes >> 32) & 0xffff);
    if (grouping2 != -1) {
        properties.groupingSize = grouping1;
        properties.groupingUsed = true;
    } else {
        properties.groupingSize = -1;
        properties.groupingUsed = false;
    }
    if (grouping3 != -1) {
        properties.secondaryGroupingSize = grouping2;
    } else {
        properties.secondaryGroupingSize = -1;
    }

    // For backwards compatibility, require that the pattern emit at least one min digit.
    int minInt, minFrac;
    if (positive.integerTotal == 0 && positive.fractionTotal > 0) {
        // patterns like ".##"
        minInt = 0;
        minFrac = uprv_max(1, positive.fractionNumerals);
    } else if (positive.integerNumerals == 0 && positive.fractionNumerals == 0) {
        // patterns like "#.##"
        minInt = 1;
        minFrac = 0;
    } else {
        minInt = positive.integerNumerals;
        minFrac = positive.fractionNumerals;
    }

    // Rounding settings
    // Don't set basic rounding when there is a currency sign; defer to CurrencyUsage
    if (positive.integerAtSigns > 0) {
        properties.minimumFractionDigits = -1;
        properties.maximumFractionDigits = -1;
        properties.roundingIncrement = 0.0;
        properties.minimumSignificantDigits = positive.integerAtSigns;
        properties.maximumSignificantDigits = positive.integerAtSigns + positive.integerTrailingHashSigns;
    } else if (!positive.rounding.isZeroish()) {
        if (!ignoreRounding) {
            properties.minimumFractionDigits = minFrac;
            properties.maximumFractionDigits = positive.fractionTotal;
            properties.roundingIncrement = positive.rounding.toDouble();
        } else {
            properties.minimumFractionDigits = -1;
            properties.maximumFractionDigits = -1;
            properties.roundingIncrement = 0.0;
        }
        properties.minimumSignificantDigits = -1;
        properties.maximumSignificantDigits = -1;
    } else {
        if (!ignoreRounding) {
            properties.minimumFractionDigits = minFrac;
            properties.maximumFractionDigits = positive.fractionTotal;
            properties.roundingIncrement = 0.0;
        } else {
            properties.minimumFractionDigits = -1;
            properties.maximumFractionDigits = -1;
            properties.roundingIncrement = 0.0;
        }
        properties.minimumSignificantDigits = -1;
        properties.maximumSignificantDigits = -1;
    }

    // If the pattern ends with a '.' then force the decimal point.
    if (positive.hasDecimal && positive.fractionTotal == 0) {
        properties.decimalSeparatorAlwaysShown = true;
    } else {
        properties.decimalSeparatorAlwaysShown = false;
    }

    // Scientific notation settings
    if (positive.exponentZeros > 0) {
        properties.exponentSignAlwaysShown = positive.exponentHasPlusSign;
        properties.minimumExponentDigits = positive.exponentZeros;
        if (positive.integerAtSigns == 0) {
            // patterns without '@' can define max integer digits, used for engineering notation
            properties.minimumIntegerDigits = positive.integerNumerals;
            properties.maximumIntegerDigits = positive.integerTotal;
        } else {
            // patterns with '@' cannot define max integer digits
            properties.minimumIntegerDigits = 1;
            properties.maximumIntegerDigits = -1;
        }
    } else {
        properties.exponentSignAlwaysShown = false;
        properties.minimumExponentDigits = -1;
        properties.minimumIntegerDigits = minInt;
        properties.maximumIntegerDigits = -1;
    }

    // Compute the affix patterns (required for both padding and affixes)
    UnicodeString posPrefix = patternInfo.getString(AffixPatternProvider::AFFIX_PREFIX);
    UnicodeString posSuffix = patternInfo.getString(0);

    // Padding settings
    if (positive.hasPadding) {
        // The width of the positive prefix and suffix templates are included in the padding
        int paddingWidth = positive.widthExceptAffixes +
                           AffixUtils::estimateLength(posPrefix, status) +
                           AffixUtils::estimateLength(posSuffix, status);
        properties.formatWidth = paddingWidth;
        UnicodeString rawPaddingString = patternInfo.getString(AffixPatternProvider::AFFIX_PADDING);
        if (rawPaddingString.length() == 1) {
            properties.padString = rawPaddingString;
        } else if (rawPaddingString.length() == 2) {
            if (rawPaddingString.charAt(0) == u'\'') {
                properties.padString.setTo(u"'", -1);
            } else {
                properties.padString = rawPaddingString;
            }
        } else {
            properties.padString = UnicodeString(rawPaddingString, 1, rawPaddingString.length() - 2);
        }
        properties.padPosition = positive.paddingLocation;
    } else {
        properties.formatWidth = -1;
        properties.padString.setToBogus();
        properties.padPosition.nullify();
    }

    // Set the affixes
    // Always call the setter, even if the prefixes are empty, especially in the case of the
    // negative prefix pattern, to prevent default values from overriding the pattern.
    properties.positivePrefixPattern = posPrefix;
    properties.positiveSuffixPattern = posSuffix;
    if (patternInfo.fHasNegativeSubpattern) {
        properties.negativePrefixPattern = patternInfo.getString(
                AffixPatternProvider::AFFIX_NEGATIVE_SUBPATTERN | AffixPatternProvider::AFFIX_PREFIX);
        properties.negativeSuffixPattern = patternInfo.getString(
                AffixPatternProvider::AFFIX_NEGATIVE_SUBPATTERN);
    } else {
        properties.negativePrefixPattern.setToBogus();
        properties.negativeSuffixPattern.setToBogus();
    }

    // Set the magnitude multiplier
    if (positive.hasPercentSign) {
        properties.magnitudeMultiplier = 2;
    } else if (positive.hasPerMilleSign) {
        properties.magnitudeMultiplier = 3;
    } else {
        properties.magnitudeMultiplier = 0;
    }
}

///////////////////////////////////////////////////////////////////
/// End PatternStringParser.java; begin PatternStringUtils.java ///
///////////////////////////////////////////////////////////////////

// Determine whether a given roundingIncrement should be ignored for formatting
// based on the current maxFrac value (maximum fraction digits). For example a
// roundingIncrement of 0.01 should be ignored if maxFrac is 1, but not if maxFrac
// is 2 or more. Note that roundingIncrements are rounded in significance, so
// a roundingIncrement of 0.006 is treated like 0.01 for this determination, i.e.
// it should not be ignored if maxFrac is 2 or more (but a roundingIncrement of
// 0.005 is treated like 0.001 for significance). This is the reason for the
// initial doubling below.
// roundIncr must be non-zero.
bool PatternStringUtils::ignoreRoundingIncrement(double roundIncr, int32_t maxFrac) {
    if (maxFrac < 0) {
        return false;
    }
    int32_t frac = 0;
    roundIncr *= 2.0;
    for (frac = 0; frac <= maxFrac && roundIncr <= 1.0; frac++, roundIncr *= 10.0);
    return (frac > maxFrac);
}

UnicodeString PatternStringUtils::propertiesToPatternString(const DecimalFormatProperties& properties,
                                                            UErrorCode& status) {
    UnicodeString sb;

    // Convenience references
    // The uprv_min() calls prevent DoS
    int32_t dosMax = 100;
    int32_t grouping1 = uprv_max(0, uprv_min(properties.groupingSize, dosMax));
    int32_t grouping2 = uprv_max(0, uprv_min(properties.secondaryGroupingSize, dosMax));
    bool useGrouping = properties.groupingUsed;
    int32_t paddingWidth = uprv_min(properties.formatWidth, dosMax);
    NullableValue<PadPosition> paddingLocation = properties.padPosition;
    UnicodeString paddingString = properties.padString;
    int32_t minInt = uprv_max(0, uprv_min(properties.minimumIntegerDigits, dosMax));
    int32_t maxInt = uprv_min(properties.maximumIntegerDigits, dosMax);
    int32_t minFrac = uprv_max(0, uprv_min(properties.minimumFractionDigits, dosMax));
    int32_t maxFrac = uprv_min(properties.maximumFractionDigits, dosMax);
    int32_t minSig = uprv_min(properties.minimumSignificantDigits, dosMax);
    int32_t maxSig = uprv_min(properties.maximumSignificantDigits, dosMax);
    bool alwaysShowDecimal = properties.decimalSeparatorAlwaysShown;
    int32_t exponentDigits = uprv_min(properties.minimumExponentDigits, dosMax);
    bool exponentShowPlusSign = properties.exponentSignAlwaysShown;

    PropertiesAffixPatternProvider affixes(properties, status);

    // Prefixes
    sb.append(affixes.getString(AffixPatternProvider::AFFIX_POS_PREFIX));
    int32_t afterPrefixPos = sb.length();

    // Figure out the grouping sizes.
    if (!useGrouping) {
        grouping1 = 0;
        grouping2 = 0;
    } else if (grouping1 == grouping2) {
        grouping1 = 0;
    }
    int32_t groupingLength = grouping1 + grouping2 + 1;

    // Figure out the digits we need to put in the pattern.
    double roundingInterval = properties.roundingIncrement;
    UnicodeString digitsString;
    int32_t digitsStringScale = 0;
    if (maxSig != uprv_min(dosMax, -1)) {
        // Significant Digits.
        while (digitsString.length() < minSig) {
            digitsString.append(u'@');
        }
        while (digitsString.length() < maxSig) {
            digitsString.append(u'#');
        }
    } else if (roundingInterval != 0.0 && !ignoreRoundingIncrement(roundingInterval,maxFrac)) {
        // Rounding Interval.
        digitsStringScale = -roundingutils::doubleFractionLength(roundingInterval, nullptr);
        // TODO: Check for DoS here?
        DecimalQuantity incrementQuantity;
        incrementQuantity.setToDouble(roundingInterval);
        incrementQuantity.adjustMagnitude(-digitsStringScale);
        incrementQuantity.roundToMagnitude(0, kDefaultMode, status);
        UnicodeString str = incrementQuantity.toPlainString();
        if (str.charAt(0) == u'-') {
            // TODO: Unsupported operation exception or fail silently?
            digitsString.append(str, 1, str.length() - 1);
        } else {
            digitsString.append(str);
        }
    }
    while (digitsString.length() + digitsStringScale < minInt) {
        digitsString.insert(0, u'0');
    }
    while (-digitsStringScale < minFrac) {
        digitsString.append(u'0');
        digitsStringScale--;
    }

    // Write the digits to the string builder
    int32_t m0 = uprv_max(groupingLength, digitsString.length() + digitsStringScale);
    m0 = (maxInt != dosMax) ? uprv_max(maxInt, m0) - 1 : m0 - 1;
    int32_t mN = (maxFrac != dosMax) ? uprv_min(-maxFrac, digitsStringScale) : digitsStringScale;
    for (int32_t magnitude = m0; magnitude >= mN; magnitude--) {
        int32_t di = digitsString.length() + digitsStringScale - magnitude - 1;
        if (di < 0 || di >= digitsString.length()) {
            sb.append(u'#');
        } else {
            sb.append(digitsString.charAt(di));
        }
        // Decimal separator
        if (magnitude == 0 && (alwaysShowDecimal || mN < 0)) {
            sb.append(u'.');
        }
        if (!useGrouping) {
            continue;
        }
        // Least-significant grouping separator
        if (magnitude > 0 && magnitude == grouping1) {
            sb.append(u',');
        }
        // All other grouping separators
        if (magnitude > grouping1 && grouping2 > 0 && (magnitude - grouping1) % grouping2 == 0) {
            sb.append(u',');
        }
    }

    // Exponential notation
    if (exponentDigits != uprv_min(dosMax, -1)) {
        sb.append(u'E');
        if (exponentShowPlusSign) {
            sb.append(u'+');
        }
        for (int32_t i = 0; i < exponentDigits; i++) {
            sb.append(u'0');
        }
    }

    // Suffixes
    int32_t beforeSuffixPos = sb.length();
    sb.append(affixes.getString(AffixPatternProvider::AFFIX_POS_SUFFIX));

    // Resolve Padding
    if (paddingWidth > 0 && !paddingLocation.isNull()) {
        while (paddingWidth - sb.length() > 0) {
            sb.insert(afterPrefixPos, u'#');
            beforeSuffixPos++;
        }
        int32_t addedLength;
        switch (paddingLocation.get(status)) {
            case PadPosition::UNUM_PAD_BEFORE_PREFIX:
                addedLength = escapePaddingString(paddingString, sb, 0, status);
                sb.insert(0, u'*');
                afterPrefixPos += addedLength + 1;
                beforeSuffixPos += addedLength + 1;
                break;
            case PadPosition::UNUM_PAD_AFTER_PREFIX:
                addedLength = escapePaddingString(paddingString, sb, afterPrefixPos, status);
                sb.insert(afterPrefixPos, u'*');
                afterPrefixPos += addedLength + 1;
                beforeSuffixPos += addedLength + 1;
                break;
            case PadPosition::UNUM_PAD_BEFORE_SUFFIX:
                escapePaddingString(paddingString, sb, beforeSuffixPos, status);
                sb.insert(beforeSuffixPos, u'*');
                break;
            case PadPosition::UNUM_PAD_AFTER_SUFFIX:
                sb.append(u'*');
                escapePaddingString(paddingString, sb, sb.length(), status);
                break;
        }
        if (U_FAILURE(status)) { return sb; }
    }

    // Negative affixes
    // Ignore if the negative prefix pattern is "-" and the negative suffix is empty
    if (affixes.hasNegativeSubpattern()) {
        sb.append(u';');
        sb.append(affixes.getString(AffixPatternProvider::AFFIX_NEG_PREFIX));
        // Copy the positive digit format into the negative.
        // This is optional; the pattern is the same as if '#' were appended here instead.
        // NOTE: It is not safe to append the UnicodeString to itself, so we need to copy.
        // See http://bugs.icu-project.org/trac/ticket/13707
        UnicodeString copy(sb);
        sb.append(copy, afterPrefixPos, beforeSuffixPos - afterPrefixPos);
        sb.append(affixes.getString(AffixPatternProvider::AFFIX_NEG_SUFFIX));
    }

    return sb;
}

int PatternStringUtils::escapePaddingString(UnicodeString input, UnicodeString& output, int startIndex,
                                            UErrorCode& status) {
    (void) status;
    if (input.length() == 0) {
        input.setTo(kFallbackPaddingString, -1);
    }
    int startLength = output.length();
    if (input.length() == 1) {
        if (input.compare(u"'", -1) == 0) {
            output.insert(startIndex, u"''", -1);
        } else {
            output.insert(startIndex, input);
        }
    } else {
        output.insert(startIndex, u'\'');
        int offset = 1;
        for (int i = 0; i < input.length(); i++) {
            // it's okay to deal in chars here because the quote mark is the only interesting thing.
            char16_t ch = input.charAt(i);
            if (ch == u'\'') {
                output.insert(startIndex + offset, u"''", -1);
                offset += 2;
            } else {
                output.insert(startIndex + offset, ch);
                offset += 1;
            }
        }
        output.insert(startIndex + offset, u'\'');
    }
    return output.length() - startLength;
}

UnicodeString
PatternStringUtils::convertLocalized(const UnicodeString& input, const DecimalFormatSymbols& symbols,
                                     bool toLocalized, UErrorCode& status) {
    // Construct a table of strings to be converted between localized and standard.
    static constexpr int32_t LEN = 21;
    UnicodeString table[LEN][2];
    int standIdx = toLocalized ? 0 : 1;
    int localIdx = toLocalized ? 1 : 0;
    table[0][standIdx] = u"%";
    table[0][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kPercentSymbol);
    table[1][standIdx] = u"‰";
    table[1][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kPerMillSymbol);
    table[2][standIdx] = u".";
    table[2][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kDecimalSeparatorSymbol);
    table[3][standIdx] = u",";
    table[3][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kGroupingSeparatorSymbol);
    table[4][standIdx] = u"-";
    table[4][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kMinusSignSymbol);
    table[5][standIdx] = u"+";
    table[5][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kPlusSignSymbol);
    table[6][standIdx] = u";";
    table[6][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kPatternSeparatorSymbol);
    table[7][standIdx] = u"@";
    table[7][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kSignificantDigitSymbol);
    table[8][standIdx] = u"E";
    table[8][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kExponentialSymbol);
    table[9][standIdx] = u"*";
    table[9][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kPadEscapeSymbol);
    table[10][standIdx] = u"#";
    table[10][localIdx] = symbols.getConstSymbol(DecimalFormatSymbols::kDigitSymbol);
    for (int i = 0; i < 10; i++) {
        table[11 + i][standIdx] = u'0' + i;
        table[11 + i][localIdx] = symbols.getConstDigitSymbol(i);
    }

    // Special case: quotes are NOT allowed to be in any localIdx strings.
    // Substitute them with '’' instead.
    for (int32_t i = 0; i < LEN; i++) {
        UnicodeString oldText(u'\'');
        UnicodeString newText(u'\x2019');
        table[i][localIdx].findAndReplace(oldText, newText);
    }

    // Iterate through the string and convert.
    // State table:
    // 0 => base state
    // 1 => first char inside a quoted sequence in input and output string
    // 2 => inside a quoted sequence in input and output string
    // 3 => first char after a close quote in input string;
    // close quote still needs to be written to output string
    // 4 => base state in input string; inside quoted sequence in output string
    // 5 => first char inside a quoted sequence in input string;
    // inside quoted sequence in output string
    UnicodeString result;
    int state = 0;
    for (int offset = 0; offset < input.length(); offset++) {
        UChar ch = input.charAt(offset);

        // Handle a quote character (state shift)
        if (ch == u'\'') {
            if (state == 0) {
                result.append(u'\'');
                state = 1;
                continue;
            } else if (state == 1) {
                result.append(u'\'');
                state = 0;
                continue;
            } else if (state == 2) {
                state = 3;
                continue;
            } else if (state == 3) {
                result.append(u'\'');
                result.append(u'\'');
                state = 1;
                continue;
            } else if (state == 4) {
                state = 5;
                continue;
            } else {
                U_ASSERT(state == 5);
                result.append(u'\'');
                result.append(u'\'');
                state = 4;
                continue;
            }
        }

        if (state == 0 || state == 3 || state == 4) {
            for (auto& pair : table) {
                // Perform a greedy match on this symbol string
                UnicodeString temp = input.tempSubString(offset, pair[0].length());
                if (temp == pair[0]) {
                    // Skip ahead past this region for the next iteration
                    offset += pair[0].length() - 1;
                    if (state == 3 || state == 4) {
                        result.append(u'\'');
                        state = 0;
                    }
                    result.append(pair[1]);
                    goto continue_outer;
                }
            }
            // No replacement found. Check if a special quote is necessary
            for (auto& pair : table) {
                UnicodeString temp = input.tempSubString(offset, pair[1].length());
                if (temp == pair[1]) {
                    if (state == 0) {
                        result.append(u'\'');
                        state = 4;
                    }
                    result.append(ch);
                    goto continue_outer;
                }
            }
            // Still nothing. Copy the char verbatim. (Add a close quote if necessary)
            if (state == 3 || state == 4) {
                result.append(u'\'');
                state = 0;
            }
            result.append(ch);
        } else {
            U_ASSERT(state == 1 || state == 2 || state == 5);
            result.append(ch);
            state = 2;
        }
        continue_outer:;
    }
    // Resolve final quotes
    if (state == 3 || state == 4) {
        result.append(u'\'');
        state = 0;
    }
    if (state != 0) {
        // Malformed localized pattern: unterminated quote
        status = U_PATTERN_SYNTAX_ERROR;
    }
    return result;
}

void PatternStringUtils::patternInfoToStringBuilder(const AffixPatternProvider& patternInfo, bool isPrefix,
                                                    Signum signum, UNumberSignDisplay signDisplay,
                                                    StandardPlural::Form plural,
                                                    bool perMilleReplacesPercent, UnicodeString& output) {

    // Should the output render '+' where '-' would normally appear in the pattern?
    bool plusReplacesMinusSign = signum != -1 && (
            signDisplay == UNUM_SIGN_ALWAYS || signDisplay == UNUM_SIGN_ACCOUNTING_ALWAYS || (
                    signum == 1 && (
                            signDisplay == UNUM_SIGN_EXCEPT_ZERO ||
                            signDisplay == UNUM_SIGN_ACCOUNTING_EXCEPT_ZERO))) &&
                                 patternInfo.positiveHasPlusSign() == false;

    // Should we use the affix from the negative subpattern? (If not, we will use the positive
    // subpattern.)
    // TODO: Deal with signum
    bool useNegativeAffixPattern = patternInfo.hasNegativeSubpattern() && (
            signum == -1 || (patternInfo.negativeHasMinusSign() && plusReplacesMinusSign));

    // Resolve the flags for the affix pattern.
    int flags = 0;
    if (useNegativeAffixPattern) {
        flags |= AffixPatternProvider::AFFIX_NEGATIVE_SUBPATTERN;
    }
    if (isPrefix) {
        flags |= AffixPatternProvider::AFFIX_PREFIX;
    }
    if (plural != StandardPlural::Form::COUNT) {
        U_ASSERT(plural == (AffixPatternProvider::AFFIX_PLURAL_MASK & plural));
        flags |= plural;
    }

    // Should we prepend a sign to the pattern?
    bool prependSign;
    if (!isPrefix || useNegativeAffixPattern) {
        prependSign = false;
    } else if (signum == -1) {
        prependSign = signDisplay != UNUM_SIGN_NEVER;
    } else {
        prependSign = plusReplacesMinusSign;
    }

    // Compute the length of the affix pattern.
    int length = patternInfo.length(flags) + (prependSign ? 1 : 0);

    // Finally, set the result into the StringBuilder.
    output.remove();
    for (int index = 0; index < length; index++) {
        char16_t candidate;
        if (prependSign && index == 0) {
            candidate = u'-';
        } else if (prependSign) {
            candidate = patternInfo.charAt(flags, index - 1);
        } else {
            candidate = patternInfo.charAt(flags, index);
        }
        if (plusReplacesMinusSign && candidate == u'-') {
            candidate = u'+';
        }
        if (perMilleReplacesPercent && candidate == u'%') {
            candidate = u'\x2030';
        }
        output.append(candidate);
    }
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING









using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;


using double_conversion::DoubleToStringConverter;

namespace {

int32_t getRoundingMagnitudeFraction(int maxFrac) {
    if (maxFrac == -1) {
        return INT32_MIN;
    }
    return -maxFrac;
}

int32_t getRoundingMagnitudeSignificant(const DecimalQuantity &value, int maxSig) {
    if (maxSig == -1) {
        return INT32_MIN;
    }
    int magnitude = value.isZeroish() ? 0 : value.getMagnitude();
    return magnitude - maxSig + 1;
}

int32_t getDisplayMagnitudeFraction(int minFrac) {
    if (minFrac == 0) {
        return INT32_MAX;
    }
    return -minFrac;
}

int32_t getDisplayMagnitudeSignificant(const DecimalQuantity &value, int minSig) {
    int magnitude = value.isZeroish() ? 0 : value.getMagnitude();
    return magnitude - minSig + 1;
}

}


MultiplierProducer::~MultiplierProducer() = default;


digits_t roundingutils::doubleFractionLength(double input, int8_t* singleDigit) {
    char buffer[DoubleToStringConverter::kBase10MaximalLength + 1];
    bool sign; // unused; always positive
    int32_t length;
    int32_t point;
    DoubleToStringConverter::DoubleToAscii(
            input,
            DoubleToStringConverter::DtoaMode::SHORTEST,
            0,
            buffer,
            sizeof(buffer),
            &sign,
            &length,
            &point
    );

    if (singleDigit == nullptr) {
        // no-op
    } else if (length == 1) {
        *singleDigit = buffer[0] - '0';
    } else {
        *singleDigit = -1;
    }

    return static_cast<digits_t>(length - point);
}


Precision Precision::unlimited() {
    return Precision(RND_NONE, {}, kDefaultMode);
}

FractionPrecision Precision::integer() {
    return constructFraction(0, 0);
}

FractionPrecision Precision::fixedFraction(int32_t minMaxFractionPlaces) {
    if (minMaxFractionPlaces >= 0 && minMaxFractionPlaces <= kMaxIntFracSig) {
        return constructFraction(minMaxFractionPlaces, minMaxFractionPlaces);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

FractionPrecision Precision::minFraction(int32_t minFractionPlaces) {
    if (minFractionPlaces >= 0 && minFractionPlaces <= kMaxIntFracSig) {
        return constructFraction(minFractionPlaces, -1);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

FractionPrecision Precision::maxFraction(int32_t maxFractionPlaces) {
    if (maxFractionPlaces >= 0 && maxFractionPlaces <= kMaxIntFracSig) {
        return constructFraction(0, maxFractionPlaces);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

FractionPrecision Precision::minMaxFraction(int32_t minFractionPlaces, int32_t maxFractionPlaces) {
    if (minFractionPlaces >= 0 && maxFractionPlaces <= kMaxIntFracSig &&
        minFractionPlaces <= maxFractionPlaces) {
        return constructFraction(minFractionPlaces, maxFractionPlaces);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

Precision Precision::fixedSignificantDigits(int32_t minMaxSignificantDigits) {
    if (minMaxSignificantDigits >= 1 && minMaxSignificantDigits <= kMaxIntFracSig) {
        return constructSignificant(minMaxSignificantDigits, minMaxSignificantDigits);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

Precision Precision::minSignificantDigits(int32_t minSignificantDigits) {
    if (minSignificantDigits >= 1 && minSignificantDigits <= kMaxIntFracSig) {
        return constructSignificant(minSignificantDigits, -1);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

Precision Precision::maxSignificantDigits(int32_t maxSignificantDigits) {
    if (maxSignificantDigits >= 1 && maxSignificantDigits <= kMaxIntFracSig) {
        return constructSignificant(1, maxSignificantDigits);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

Precision Precision::minMaxSignificantDigits(int32_t minSignificantDigits, int32_t maxSignificantDigits) {
    if (minSignificantDigits >= 1 && maxSignificantDigits <= kMaxIntFracSig &&
        minSignificantDigits <= maxSignificantDigits) {
        return constructSignificant(minSignificantDigits, maxSignificantDigits);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

IncrementPrecision Precision::increment(double roundingIncrement) {
    if (roundingIncrement > 0.0) {
        return constructIncrement(roundingIncrement, 0);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

CurrencyPrecision Precision::currency(UCurrencyUsage currencyUsage) {
    return constructCurrency(currencyUsage);
}

Precision FractionPrecision::withMinDigits(int32_t minSignificantDigits) const {
    if (fType == RND_ERROR) { return *this; } // no-op in error state
    if (minSignificantDigits >= 1 && minSignificantDigits <= kMaxIntFracSig) {
        return constructFractionSignificant(*this, minSignificantDigits, -1);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

Precision FractionPrecision::withMaxDigits(int32_t maxSignificantDigits) const {
    if (fType == RND_ERROR) { return *this; } // no-op in error state
    if (maxSignificantDigits >= 1 && maxSignificantDigits <= kMaxIntFracSig) {
        return constructFractionSignificant(*this, -1, maxSignificantDigits);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

// Private method on base class
Precision Precision::withCurrency(const CurrencyUnit &currency, UErrorCode &status) const {
    if (fType == RND_ERROR) { return *this; } // no-op in error state
    U_ASSERT(fType == RND_CURRENCY);
    const char16_t *isoCode = currency.getISOCurrency();
    double increment = ucurr_getRoundingIncrementForUsage(isoCode, fUnion.currencyUsage, &status);
    int32_t minMaxFrac = ucurr_getDefaultFractionDigitsForUsage(
            isoCode, fUnion.currencyUsage, &status);
    if (increment != 0.0) {
        return constructIncrement(increment, minMaxFrac);
    } else {
        return constructFraction(minMaxFrac, minMaxFrac);
    }
}

// Public method on CurrencyPrecision subclass
Precision CurrencyPrecision::withCurrency(const CurrencyUnit &currency) const {
    UErrorCode localStatus = U_ZERO_ERROR;
    Precision result = Precision::withCurrency(currency, localStatus);
    if (U_FAILURE(localStatus)) {
        return {localStatus};
    }
    return result;
}

Precision IncrementPrecision::withMinFraction(int32_t minFrac) const {
    if (fType == RND_ERROR) { return *this; } // no-op in error state
    if (minFrac >= 0 && minFrac <= kMaxIntFracSig) {
        return constructIncrement(fUnion.increment.fIncrement, minFrac);
    } else {
        return {U_NUMBER_ARG_OUTOFBOUNDS_ERROR};
    }
}

FractionPrecision Precision::constructFraction(int32_t minFrac, int32_t maxFrac) {
    FractionSignificantSettings settings;
    settings.fMinFrac = static_cast<digits_t>(minFrac);
    settings.fMaxFrac = static_cast<digits_t>(maxFrac);
    settings.fMinSig = -1;
    settings.fMaxSig = -1;
    PrecisionUnion union_;
    union_.fracSig = settings;
    return {RND_FRACTION, union_, kDefaultMode};
}

Precision Precision::constructSignificant(int32_t minSig, int32_t maxSig) {
    FractionSignificantSettings settings;
    settings.fMinFrac = -1;
    settings.fMaxFrac = -1;
    settings.fMinSig = static_cast<digits_t>(minSig);
    settings.fMaxSig = static_cast<digits_t>(maxSig);
    PrecisionUnion union_;
    union_.fracSig = settings;
    return {RND_SIGNIFICANT, union_, kDefaultMode};
}

Precision
Precision::constructFractionSignificant(const FractionPrecision &base, int32_t minSig, int32_t maxSig) {
    FractionSignificantSettings settings = base.fUnion.fracSig;
    settings.fMinSig = static_cast<digits_t>(minSig);
    settings.fMaxSig = static_cast<digits_t>(maxSig);
    PrecisionUnion union_;
    union_.fracSig = settings;
    return {RND_FRACTION_SIGNIFICANT, union_, kDefaultMode};
}

IncrementPrecision Precision::constructIncrement(double increment, int32_t minFrac) {
    IncrementSettings settings;
    // Note: For number formatting, fIncrement is used for RND_INCREMENT but not
    // RND_INCREMENT_ONE or RND_INCREMENT_FIVE. However, fIncrement is used in all
    // three when constructing a skeleton.
    settings.fIncrement = increment;
    settings.fMinFrac = static_cast<digits_t>(minFrac);
    // One of the few pre-computed quantities:
    // Note: it is possible for minFrac to be more than maxFrac... (misleading)
    int8_t singleDigit;
    settings.fMaxFrac = roundingutils::doubleFractionLength(increment, &singleDigit);
    PrecisionUnion union_;
    union_.increment = settings;
    if (singleDigit == 1) {
        // NOTE: In C++, we must return the correct value type with the correct union.
        // It would be invalid to return a RND_FRACTION here because the methods on the
        // IncrementPrecision type assume that the union is backed by increment data.
        return {RND_INCREMENT_ONE, union_, kDefaultMode};
    } else if (singleDigit == 5) {
        return {RND_INCREMENT_FIVE, union_, kDefaultMode};
    } else {
        return {RND_INCREMENT, union_, kDefaultMode};
    }
}

CurrencyPrecision Precision::constructCurrency(UCurrencyUsage usage) {
    PrecisionUnion union_;
    union_.currencyUsage = usage;
    return {RND_CURRENCY, union_, kDefaultMode};
}


RoundingImpl::RoundingImpl(const Precision& precision, UNumberFormatRoundingMode roundingMode,
                           const CurrencyUnit& currency, UErrorCode& status)
        : fPrecision(precision), fRoundingMode(roundingMode), fPassThrough(false) {
    if (precision.fType == Precision::RND_CURRENCY) {
        fPrecision = precision.withCurrency(currency, status);
    }
}

RoundingImpl RoundingImpl::passThrough() {
    RoundingImpl retval;
    retval.fPassThrough = true;
    return retval;
}

bool RoundingImpl::isSignificantDigits() const {
    return fPrecision.fType == Precision::RND_SIGNIFICANT;
}

int32_t
RoundingImpl::chooseMultiplierAndApply(number::impl::DecimalQuantity &input, const number::impl::MultiplierProducer &producer,
                                  UErrorCode &status) {
    // Do not call this method with zero, NaN, or infinity.
    U_ASSERT(!input.isZeroish());

    // Perform the first attempt at rounding.
    int magnitude = input.getMagnitude();
    int multiplier = producer.getMultiplier(magnitude);
    input.adjustMagnitude(multiplier);
    apply(input, status);

    // If the number rounded to zero, exit.
    if (input.isZeroish() || U_FAILURE(status)) {
        return multiplier;
    }

    // If the new magnitude after rounding is the same as it was before rounding, then we are done.
    // This case applies to most numbers.
    if (input.getMagnitude() == magnitude + multiplier) {
        return multiplier;
    }

    // If the above case DIDN'T apply, then we have a case like 99.9 -> 100 or 999.9 -> 1000:
    // The number rounded up to the next magnitude. Check if the multiplier changes; if it doesn't,
    // we do not need to make any more adjustments.
    int _multiplier = producer.getMultiplier(magnitude + 1);
    if (multiplier == _multiplier) {
        return multiplier;
    }

    // We have a case like 999.9 -> 1000, where the correct output is "1K", not "1000".
    // Fix the magnitude and re-apply the rounding strategy.
    input.adjustMagnitude(_multiplier - multiplier);
    apply(input, status);
    return _multiplier;
}

/** This is the method that contains the actual rounding logic. */
void RoundingImpl::apply(number::impl::DecimalQuantity &value, UErrorCode& status) const {
    if (fPassThrough) {
        return;
    }
    switch (fPrecision.fType) {
        case Precision::RND_BOGUS:
        case Precision::RND_ERROR:
            // Errors should be caught before the apply() method is called
            status = U_INTERNAL_PROGRAM_ERROR;
            break;

        case Precision::RND_NONE:
            value.roundToInfinity();
            break;

        case Precision::RND_FRACTION:
            value.roundToMagnitude(
                    getRoundingMagnitudeFraction(fPrecision.fUnion.fracSig.fMaxFrac),
                    fRoundingMode,
                    status);
            value.setMinFraction(
                    uprv_max(0, -getDisplayMagnitudeFraction(fPrecision.fUnion.fracSig.fMinFrac)));
            break;

        case Precision::RND_SIGNIFICANT:
            value.roundToMagnitude(
                    getRoundingMagnitudeSignificant(value, fPrecision.fUnion.fracSig.fMaxSig),
                    fRoundingMode,
                    status);
            value.setMinFraction(
                    uprv_max(0, -getDisplayMagnitudeSignificant(value, fPrecision.fUnion.fracSig.fMinSig)));
            // Make sure that digits are displayed on zero.
            if (value.isZeroish() && fPrecision.fUnion.fracSig.fMinSig > 0) {
                value.setMinInteger(1);
            }
            break;

        case Precision::RND_FRACTION_SIGNIFICANT: {
            int32_t displayMag = getDisplayMagnitudeFraction(fPrecision.fUnion.fracSig.fMinFrac);
            int32_t roundingMag = getRoundingMagnitudeFraction(fPrecision.fUnion.fracSig.fMaxFrac);
            if (fPrecision.fUnion.fracSig.fMinSig == -1) {
                // Max Sig override
                int32_t candidate = getRoundingMagnitudeSignificant(
                        value,
                        fPrecision.fUnion.fracSig.fMaxSig);
                roundingMag = uprv_max(roundingMag, candidate);
            } else {
                // Min Sig override
                int32_t candidate = getDisplayMagnitudeSignificant(
                        value,
                        fPrecision.fUnion.fracSig.fMinSig);
                roundingMag = uprv_min(roundingMag, candidate);
            }
            value.roundToMagnitude(roundingMag, fRoundingMode, status);
            value.setMinFraction(uprv_max(0, -displayMag));
            break;
        }

        case Precision::RND_INCREMENT:
            value.roundToIncrement(
                    fPrecision.fUnion.increment.fIncrement,
                    fRoundingMode,
                    status);
            value.setMinFraction(fPrecision.fUnion.increment.fMinFrac);
            break;

        case Precision::RND_INCREMENT_ONE:
            value.roundToMagnitude(
                    -fPrecision.fUnion.increment.fMaxFrac,
                    fRoundingMode,
                    status);
            value.setMinFraction(fPrecision.fUnion.increment.fMinFrac);
            break;

        case Precision::RND_INCREMENT_FIVE:
            value.roundToNickel(
                    -fPrecision.fUnion.increment.fMaxFrac,
                    fRoundingMode,
                    status);
            value.setMinFraction(fPrecision.fUnion.increment.fMinFrac);
            break;

        case Precision::RND_CURRENCY:
            // Call .withCurrency() before .apply()!
            UPRV_UNREACHABLE;

        default:
            UPRV_UNREACHABLE;
    }
}

void RoundingImpl::apply(number::impl::DecimalQuantity &value, int32_t minInt, UErrorCode /*status*/) {
    // This method is intended for the one specific purpose of helping print "00.000E0".
    U_ASSERT(isSignificantDigits());
    U_ASSERT(value.isZeroish());
    value.setMinFraction(fPrecision.fUnion.fracSig.fMinSig - minInt);
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2017 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

#include <cstdlib>






using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

// NOTE: The object lifecycle of ScientificModifier and ScientificHandler differ greatly in Java and C++.
//
// During formatting, we need to provide an object with state (the exponent) as the inner modifier.
//
// In Java, where the priority is put on reducing object creations, the unsafe code path re-uses the
// ScientificHandler as a ScientificModifier, and the safe code path pre-computes 25 ScientificModifier
// instances.  This scheme reduces the number of object creations by 1 in both safe and unsafe.
//
// In C++, MicroProps provides a pre-allocated ScientificModifier, and ScientificHandler simply populates
// the state (the exponent) into that ScientificModifier. There is no difference between safe and unsafe.

ScientificModifier::ScientificModifier() : fExponent(0), fHandler(nullptr) {}

void ScientificModifier::set(int32_t exponent, const ScientificHandler *handler) {
    // ScientificModifier should be set only once.
    U_ASSERT(fHandler == nullptr);
    fExponent = exponent;
    fHandler = handler;
}

int32_t ScientificModifier::apply(FormattedStringBuilder &output, int32_t /*leftIndex*/, int32_t rightIndex,
                                  UErrorCode &status) const {
    // FIXME: Localized exponent separator location.
    int i = rightIndex;
    // Append the exponent separator and sign
    i += output.insert(
            i,
            fHandler->fSymbols->getSymbol(DecimalFormatSymbols::ENumberFormatSymbol::kExponentialSymbol),
            UNUM_EXPONENT_SYMBOL_FIELD,
            status);
    if (fExponent < 0 && fHandler->fSettings.fExponentSignDisplay != UNUM_SIGN_NEVER) {
        i += output.insert(
                i,
                fHandler->fSymbols
                        ->getSymbol(DecimalFormatSymbols::ENumberFormatSymbol::kMinusSignSymbol),
                UNUM_EXPONENT_SIGN_FIELD,
                status);
    } else if (fExponent >= 0 && fHandler->fSettings.fExponentSignDisplay == UNUM_SIGN_ALWAYS) {
        i += output.insert(
                i,
                fHandler->fSymbols
                        ->getSymbol(DecimalFormatSymbols::ENumberFormatSymbol::kPlusSignSymbol),
                UNUM_EXPONENT_SIGN_FIELD,
                status);
    }
    // Append the exponent digits (using a simple inline algorithm)
    int32_t disp = std::abs(fExponent);
    for (int j = 0; j < fHandler->fSettings.fMinExponentDigits || disp > 0; j++, disp /= 10) {
        auto d = static_cast<int8_t>(disp % 10);
        i += utils::insertDigitFromSymbols(
                output,
                i - j,
                d,
                *fHandler->fSymbols,
                UNUM_EXPONENT_FIELD,
                status);
    }
    return i - rightIndex;
}

int32_t ScientificModifier::getPrefixLength() const {
    // TODO: Localized exponent separator location.
    return 0;
}

int32_t ScientificModifier::getCodePointCount() const {
    // NOTE: This method is only called one place, NumberRangeFormatterImpl.
    // The call site only cares about != 0 and != 1.
    // Return a very large value so that if this method is used elsewhere, we should notice.
    return 999;
}

bool ScientificModifier::isStrong() const {
    // Scientific is always strong
    return true;
}

bool ScientificModifier::containsField(UNumberFormatFields field) const {
    (void)field;
    // This method is not used for inner modifiers.
    UPRV_UNREACHABLE;
}

void ScientificModifier::getParameters(Parameters& output) const {
    // Not part of any plural sets
    output.obj = nullptr;
}

bool ScientificModifier::semanticallyEquivalent(const Modifier& other) const {
    auto* _other = dynamic_cast<const ScientificModifier*>(&other);
    if (_other == nullptr) {
        return false;
    }
    // TODO: Check for locale symbols and settings as well? Could be less efficient.
    return fExponent == _other->fExponent;
}

// Note: Visual Studio does not compile this function without full name space. Why?
icu::number::impl::ScientificHandler::ScientificHandler(const Notation *notation, const DecimalFormatSymbols *symbols,
	const MicroPropsGenerator *parent) : 
	fSettings(notation->fUnion.scientific), fSymbols(symbols), fParent(parent) {}

void ScientificHandler::processQuantity(DecimalQuantity &quantity, MicroProps &micros,
                                        UErrorCode &status) const {
    fParent->processQuantity(quantity, micros, status);
    if (U_FAILURE(status)) { return; }

    // Do not apply scientific notation to special doubles
    if (quantity.isInfinite() || quantity.isNaN()) {
        micros.modInner = &micros.helpers.emptyStrongModifier;
        return;
    }

    // Treat zero as if it had magnitude 0
    int32_t exponent;
    if (quantity.isZeroish()) {
        if (fSettings.fRequireMinInt && micros.rounder.isSignificantDigits()) {
            // Show "00.000E0" on pattern "00.000E0"
            micros.rounder.apply(quantity, fSettings.fEngineeringInterval, status);
            exponent = 0;
        } else {
            micros.rounder.apply(quantity, status);
            exponent = 0;
        }
    } else {
        exponent = -micros.rounder.chooseMultiplierAndApply(quantity, *this, status);
    }

    // Use MicroProps's helper ScientificModifier and save it as the modInner.
    ScientificModifier &mod = micros.helpers.scientificModifier;
    mod.set(exponent, this);
    micros.modInner = &mod;

    // We already performed rounding. Do not perform it again.
    micros.rounder = RoundingImpl::passThrough();
}

int32_t ScientificHandler::getMultiplier(int32_t magnitude) const {
    int32_t interval = fSettings.fEngineeringInterval;
    int32_t digitsShown;
    if (fSettings.fRequireMinInt) {
        // For patterns like "000.00E0" and ".00E0"
        digitsShown = interval;
    } else if (interval <= 1) {
        // For patterns like "0.00E0" and "@@@E0"
        digitsShown = 1;
    } else {
        // For patterns like "##0.00"
        digitsShown = ((magnitude % interval + interval) % interval) + 1;
    }
    return digitsShown - magnitude - 1;
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif














using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;
using namespace icu::number::impl::skeleton;

namespace {

icu::UInitOnce gNumberSkeletonsInitOnce = U_INITONCE_INITIALIZER;

char16_t* kSerializedStemTrie = nullptr;

UBool U_CALLCONV cleanupNumberSkeletons() {
    uprv_free(kSerializedStemTrie);
    kSerializedStemTrie = nullptr;
    gNumberSkeletonsInitOnce.reset();
    return TRUE;
}

void U_CALLCONV initNumberSkeletons(UErrorCode& status) {
    ucln_i18n_registerCleanup(UCLN_I18N_NUMBER_SKELETONS, cleanupNumberSkeletons);

    UCharsTrieBuilder b(status);
    if (U_FAILURE(status)) { return; }

    // Section 1:
    b.add(u"compact-short", STEM_COMPACT_SHORT, status);
    b.add(u"compact-long", STEM_COMPACT_LONG, status);
    b.add(u"scientific", STEM_SCIENTIFIC, status);
    b.add(u"engineering", STEM_ENGINEERING, status);
    b.add(u"notation-simple", STEM_NOTATION_SIMPLE, status);
    b.add(u"base-unit", STEM_BASE_UNIT, status);
    b.add(u"percent", STEM_PERCENT, status);
    b.add(u"permille", STEM_PERMILLE, status);
    b.add(u"precision-integer", STEM_PRECISION_INTEGER, status);
    b.add(u"precision-unlimited", STEM_PRECISION_UNLIMITED, status);
    b.add(u"precision-currency-standard", STEM_PRECISION_CURRENCY_STANDARD, status);
    b.add(u"precision-currency-cash", STEM_PRECISION_CURRENCY_CASH, status);
    b.add(u"rounding-mode-ceiling", STEM_ROUNDING_MODE_CEILING, status);
    b.add(u"rounding-mode-floor", STEM_ROUNDING_MODE_FLOOR, status);
    b.add(u"rounding-mode-down", STEM_ROUNDING_MODE_DOWN, status);
    b.add(u"rounding-mode-up", STEM_ROUNDING_MODE_UP, status);
    b.add(u"rounding-mode-half-even", STEM_ROUNDING_MODE_HALF_EVEN, status);
    b.add(u"rounding-mode-half-down", STEM_ROUNDING_MODE_HALF_DOWN, status);
    b.add(u"rounding-mode-half-up", STEM_ROUNDING_MODE_HALF_UP, status);
    b.add(u"rounding-mode-unnecessary", STEM_ROUNDING_MODE_UNNECESSARY, status);
    b.add(u"group-off", STEM_GROUP_OFF, status);
    b.add(u"group-min2", STEM_GROUP_MIN2, status);
    b.add(u"group-auto", STEM_GROUP_AUTO, status);
    b.add(u"group-on-aligned", STEM_GROUP_ON_ALIGNED, status);
    b.add(u"group-thousands", STEM_GROUP_THOUSANDS, status);
    b.add(u"latin", STEM_LATIN, status);
    b.add(u"unit-width-narrow", STEM_UNIT_WIDTH_NARROW, status);
    b.add(u"unit-width-short", STEM_UNIT_WIDTH_SHORT, status);
    b.add(u"unit-width-full-name", STEM_UNIT_WIDTH_FULL_NAME, status);
    b.add(u"unit-width-iso-code", STEM_UNIT_WIDTH_ISO_CODE, status);
    b.add(u"unit-width-hidden", STEM_UNIT_WIDTH_HIDDEN, status);
    b.add(u"sign-auto", STEM_SIGN_AUTO, status);
    b.add(u"sign-always", STEM_SIGN_ALWAYS, status);
    b.add(u"sign-never", STEM_SIGN_NEVER, status);
    b.add(u"sign-accounting", STEM_SIGN_ACCOUNTING, status);
    b.add(u"sign-accounting-always", STEM_SIGN_ACCOUNTING_ALWAYS, status);
    b.add(u"sign-except-zero", STEM_SIGN_EXCEPT_ZERO, status);
    b.add(u"sign-accounting-except-zero", STEM_SIGN_ACCOUNTING_EXCEPT_ZERO, status);
    b.add(u"decimal-auto", STEM_DECIMAL_AUTO, status);
    b.add(u"decimal-always", STEM_DECIMAL_ALWAYS, status);
    if (U_FAILURE(status)) { return; }

    // Section 2:
    b.add(u"precision-increment", STEM_PRECISION_INCREMENT, status);
    b.add(u"measure-unit", STEM_MEASURE_UNIT, status);
    b.add(u"per-measure-unit", STEM_PER_MEASURE_UNIT, status);
    b.add(u"currency", STEM_CURRENCY, status);
    b.add(u"integer-width", STEM_INTEGER_WIDTH, status);
    b.add(u"numbering-system", STEM_NUMBERING_SYSTEM, status);
    b.add(u"scale", STEM_SCALE, status);
    if (U_FAILURE(status)) { return; }

    // Build the CharsTrie
    // TODO: Use SLOW or FAST here?
    UnicodeString result;
    b.buildUnicodeString(USTRINGTRIE_BUILD_FAST, result, status);
    if (U_FAILURE(status)) { return; }

    // Copy the result into the global constant pointer
    size_t numBytes = result.length() * sizeof(char16_t);
    kSerializedStemTrie = static_cast<char16_t*>(uprv_malloc(numBytes));
    uprv_memcpy(kSerializedStemTrie, result.getBuffer(), numBytes);
}


inline void appendMultiple(UnicodeString& sb, UChar32 cp, int32_t count) {
    for (int i = 0; i < count; i++) {
        sb.append(cp);
    }
}


#define CHECK_NULL(seen, field, status) (void)(seen); /* for auto-format line wrapping */ \
UPRV_BLOCK_MACRO_BEGIN { \
    if ((seen).field) { \
        (status) = U_NUMBER_SKELETON_SYNTAX_ERROR; \
        return STATE_NULL; \
    } \
    (seen).field = true; \
} UPRV_BLOCK_MACRO_END


#define SKELETON_UCHAR_TO_CHAR(dest, src, start, end, status) (void)(dest); \
UPRV_BLOCK_MACRO_BEGIN { \
    UErrorCode conversionStatus = U_ZERO_ERROR; \
    (dest).appendInvariantChars({FALSE, (src).getBuffer() + (start), (end) - (start)}, conversionStatus); \
    if (conversionStatus == U_INVARIANT_CONVERSION_ERROR) { \
        /* Don't propagate the invariant conversion error; it is a skeleton syntax error */ \
        (status) = U_NUMBER_SKELETON_SYNTAX_ERROR; \
        return; \
    } else if (U_FAILURE(conversionStatus)) { \
        (status) = conversionStatus; \
        return; \
    } \
} UPRV_BLOCK_MACRO_END


} // anonymous namespace


Notation stem_to_object::notation(skeleton::StemEnum stem) {
    switch (stem) {
        case STEM_COMPACT_SHORT:
            return Notation::compactShort();
        case STEM_COMPACT_LONG:
            return Notation::compactLong();
        case STEM_SCIENTIFIC:
            return Notation::scientific();
        case STEM_ENGINEERING:
            return Notation::engineering();
        case STEM_NOTATION_SIMPLE:
            return Notation::simple();
        default:
            UPRV_UNREACHABLE;
    }
}

MeasureUnit stem_to_object::unit(skeleton::StemEnum stem) {
    switch (stem) {
        case STEM_BASE_UNIT:
            // Slicing is okay
            return NoUnit::base(); // NOLINT
        case STEM_PERCENT:
            // Slicing is okay
            return NoUnit::percent(); // NOLINT
        case STEM_PERMILLE:
            // Slicing is okay
            return NoUnit::permille(); // NOLINT
        default:
            UPRV_UNREACHABLE;
    }
}

Precision stem_to_object::precision(skeleton::StemEnum stem) {
    switch (stem) {
        case STEM_PRECISION_INTEGER:
            return Precision::integer();
        case STEM_PRECISION_UNLIMITED:
            return Precision::unlimited();
        case STEM_PRECISION_CURRENCY_STANDARD:
            return Precision::currency(UCURR_USAGE_STANDARD);
        case STEM_PRECISION_CURRENCY_CASH:
            return Precision::currency(UCURR_USAGE_CASH);
        default:
            UPRV_UNREACHABLE;
    }
}

UNumberFormatRoundingMode stem_to_object::roundingMode(skeleton::StemEnum stem) {
    switch (stem) {
        case STEM_ROUNDING_MODE_CEILING:
            return UNUM_ROUND_CEILING;
        case STEM_ROUNDING_MODE_FLOOR:
            return UNUM_ROUND_FLOOR;
        case STEM_ROUNDING_MODE_DOWN:
            return UNUM_ROUND_DOWN;
        case STEM_ROUNDING_MODE_UP:
            return UNUM_ROUND_UP;
        case STEM_ROUNDING_MODE_HALF_EVEN:
            return UNUM_ROUND_HALFEVEN;
        case STEM_ROUNDING_MODE_HALF_DOWN:
            return UNUM_ROUND_HALFDOWN;
        case STEM_ROUNDING_MODE_HALF_UP:
            return UNUM_ROUND_HALFUP;
        case STEM_ROUNDING_MODE_UNNECESSARY:
            return UNUM_ROUND_UNNECESSARY;
        default:
            UPRV_UNREACHABLE;
    }
}

UNumberGroupingStrategy stem_to_object::groupingStrategy(skeleton::StemEnum stem) {
    switch (stem) {
        case STEM_GROUP_OFF:
            return UNUM_GROUPING_OFF;
        case STEM_GROUP_MIN2:
            return UNUM_GROUPING_MIN2;
        case STEM_GROUP_AUTO:
            return UNUM_GROUPING_AUTO;
        case STEM_GROUP_ON_ALIGNED:
            return UNUM_GROUPING_ON_ALIGNED;
        case STEM_GROUP_THOUSANDS:
            return UNUM_GROUPING_THOUSANDS;
        default:
            return UNUM_GROUPING_COUNT; // for objects, throw; for enums, return COUNT
    }
}

UNumberUnitWidth stem_to_object::unitWidth(skeleton::StemEnum stem) {
    switch (stem) {
        case STEM_UNIT_WIDTH_NARROW:
            return UNUM_UNIT_WIDTH_NARROW;
        case STEM_UNIT_WIDTH_SHORT:
            return UNUM_UNIT_WIDTH_SHORT;
        case STEM_UNIT_WIDTH_FULL_NAME:
            return UNUM_UNIT_WIDTH_FULL_NAME;
        case STEM_UNIT_WIDTH_ISO_CODE:
            return UNUM_UNIT_WIDTH_ISO_CODE;
        case STEM_UNIT_WIDTH_HIDDEN:
            return UNUM_UNIT_WIDTH_HIDDEN;
        default:
            return UNUM_UNIT_WIDTH_COUNT; // for objects, throw; for enums, return COUNT
    }
}

UNumberSignDisplay stem_to_object::signDisplay(skeleton::StemEnum stem) {
    switch (stem) {
        case STEM_SIGN_AUTO:
            return UNUM_SIGN_AUTO;
        case STEM_SIGN_ALWAYS:
            return UNUM_SIGN_ALWAYS;
        case STEM_SIGN_NEVER:
            return UNUM_SIGN_NEVER;
        case STEM_SIGN_ACCOUNTING:
            return UNUM_SIGN_ACCOUNTING;
        case STEM_SIGN_ACCOUNTING_ALWAYS:
            return UNUM_SIGN_ACCOUNTING_ALWAYS;
        case STEM_SIGN_EXCEPT_ZERO:
            return UNUM_SIGN_EXCEPT_ZERO;
        case STEM_SIGN_ACCOUNTING_EXCEPT_ZERO:
            return UNUM_SIGN_ACCOUNTING_EXCEPT_ZERO;
        default:
            return UNUM_SIGN_COUNT; // for objects, throw; for enums, return COUNT
    }
}

UNumberDecimalSeparatorDisplay stem_to_object::decimalSeparatorDisplay(skeleton::StemEnum stem) {
    switch (stem) {
        case STEM_DECIMAL_AUTO:
            return UNUM_DECIMAL_SEPARATOR_AUTO;
        case STEM_DECIMAL_ALWAYS:
            return UNUM_DECIMAL_SEPARATOR_ALWAYS;
        default:
            return UNUM_DECIMAL_SEPARATOR_COUNT; // for objects, throw; for enums, return COUNT
    }
}


void enum_to_stem_string::roundingMode(UNumberFormatRoundingMode value, UnicodeString& sb) {
    switch (value) {
        case UNUM_ROUND_CEILING:
            sb.append(u"rounding-mode-ceiling", -1);
            break;
        case UNUM_ROUND_FLOOR:
            sb.append(u"rounding-mode-floor", -1);
            break;
        case UNUM_ROUND_DOWN:
            sb.append(u"rounding-mode-down", -1);
            break;
        case UNUM_ROUND_UP:
            sb.append(u"rounding-mode-up", -1);
            break;
        case UNUM_ROUND_HALFEVEN:
            sb.append(u"rounding-mode-half-even", -1);
            break;
        case UNUM_ROUND_HALFDOWN:
            sb.append(u"rounding-mode-half-down", -1);
            break;
        case UNUM_ROUND_HALFUP:
            sb.append(u"rounding-mode-half-up", -1);
            break;
        case UNUM_ROUND_UNNECESSARY:
            sb.append(u"rounding-mode-unnecessary", -1);
            break;
        default:
            UPRV_UNREACHABLE;
    }
}

void enum_to_stem_string::groupingStrategy(UNumberGroupingStrategy value, UnicodeString& sb) {
    switch (value) {
        case UNUM_GROUPING_OFF:
            sb.append(u"group-off", -1);
            break;
        case UNUM_GROUPING_MIN2:
            sb.append(u"group-min2", -1);
            break;
        case UNUM_GROUPING_AUTO:
            sb.append(u"group-auto", -1);
            break;
        case UNUM_GROUPING_ON_ALIGNED:
            sb.append(u"group-on-aligned", -1);
            break;
        case UNUM_GROUPING_THOUSANDS:
            sb.append(u"group-thousands", -1);
            break;
        default:
            UPRV_UNREACHABLE;
    }
}

void enum_to_stem_string::unitWidth(UNumberUnitWidth value, UnicodeString& sb) {
    switch (value) {
        case UNUM_UNIT_WIDTH_NARROW:
            sb.append(u"unit-width-narrow", -1);
            break;
        case UNUM_UNIT_WIDTH_SHORT:
            sb.append(u"unit-width-short", -1);
            break;
        case UNUM_UNIT_WIDTH_FULL_NAME:
            sb.append(u"unit-width-full-name", -1);
            break;
        case UNUM_UNIT_WIDTH_ISO_CODE:
            sb.append(u"unit-width-iso-code", -1);
            break;
        case UNUM_UNIT_WIDTH_HIDDEN:
            sb.append(u"unit-width-hidden", -1);
            break;
        default:
            UPRV_UNREACHABLE;
    }
}

void enum_to_stem_string::signDisplay(UNumberSignDisplay value, UnicodeString& sb) {
    switch (value) {
        case UNUM_SIGN_AUTO:
            sb.append(u"sign-auto", -1);
            break;
        case UNUM_SIGN_ALWAYS:
            sb.append(u"sign-always", -1);
            break;
        case UNUM_SIGN_NEVER:
            sb.append(u"sign-never", -1);
            break;
        case UNUM_SIGN_ACCOUNTING:
            sb.append(u"sign-accounting", -1);
            break;
        case UNUM_SIGN_ACCOUNTING_ALWAYS:
            sb.append(u"sign-accounting-always", -1);
            break;
        case UNUM_SIGN_EXCEPT_ZERO:
            sb.append(u"sign-except-zero", -1);
            break;
        case UNUM_SIGN_ACCOUNTING_EXCEPT_ZERO:
            sb.append(u"sign-accounting-except-zero", -1);
            break;
        default:
            UPRV_UNREACHABLE;
    }
}

void
enum_to_stem_string::decimalSeparatorDisplay(UNumberDecimalSeparatorDisplay value, UnicodeString& sb) {
    switch (value) {
        case UNUM_DECIMAL_SEPARATOR_AUTO:
            sb.append(u"decimal-auto", -1);
            break;
        case UNUM_DECIMAL_SEPARATOR_ALWAYS:
            sb.append(u"decimal-always", -1);
            break;
        default:
            UPRV_UNREACHABLE;
    }
}


UnlocalizedNumberFormatter skeleton::create(
        const UnicodeString& skeletonString, UParseError* perror, UErrorCode& status) {

    // Initialize perror
    if (perror != nullptr) {
        perror->line = 0;
        perror->offset = -1;
        perror->preContext[0] = 0;
        perror->postContext[0] = 0;
    }

    umtx_initOnce(gNumberSkeletonsInitOnce, &initNumberSkeletons, status);
    if (U_FAILURE(status)) {
        return {};
    }

    int32_t errOffset;
    MacroProps macros = parseSkeleton(skeletonString, errOffset, status);
    if (U_SUCCESS(status)) {
        return NumberFormatter::with().macros(macros);
    }

    if (perror == nullptr) {
        return {};
    }

    // Populate the UParseError with the error location
    perror->offset = errOffset;
    int32_t contextStart = uprv_max(0, errOffset - U_PARSE_CONTEXT_LEN + 1);
    int32_t contextEnd = uprv_min(skeletonString.length(), errOffset + U_PARSE_CONTEXT_LEN - 1);
    skeletonString.extract(contextStart, errOffset - contextStart, perror->preContext, 0);
    perror->preContext[errOffset - contextStart] = 0;
    skeletonString.extract(errOffset, contextEnd - errOffset, perror->postContext, 0);
    perror->postContext[contextEnd - errOffset] = 0;
    return {};
}

UnicodeString skeleton::generate(const MacroProps& macros, UErrorCode& status) {
    umtx_initOnce(gNumberSkeletonsInitOnce, &initNumberSkeletons, status);
    UnicodeString sb;
    GeneratorHelpers::generateSkeleton(macros, sb, status);
    return sb;
}

MacroProps skeleton::parseSkeleton(
        const UnicodeString& skeletonString, int32_t& errOffset, UErrorCode& status) {
    U_ASSERT(U_SUCCESS(status));

    // Add a trailing whitespace to the end of the skeleton string to make code cleaner.
    UnicodeString tempSkeletonString(skeletonString);
    tempSkeletonString.append(u' ');

    SeenMacroProps seen;
    MacroProps macros;
    StringSegment segment(tempSkeletonString, false);
    UCharsTrie stemTrie(kSerializedStemTrie);
    ParseState stem = STATE_NULL;
    int32_t offset = 0;

    // Primary skeleton parse loop:
    while (offset < segment.length()) {
        UChar32 cp = segment.codePointAt(offset);
        bool isTokenSeparator = PatternProps::isWhiteSpace(cp);
        bool isOptionSeparator = (cp == u'/');

        if (!isTokenSeparator && !isOptionSeparator) {
            // Non-separator token; consume it.
            offset += U16_LENGTH(cp);
            if (stem == STATE_NULL) {
                // We are currently consuming a stem.
                // Go to the next state in the stem trie.
                stemTrie.nextForCodePoint(cp);
            }
            continue;
        }

        // We are looking at a token or option separator.
        // If the segment is nonempty, parse it and reset the segment.
        // Otherwise, make sure it is a valid repeating separator.
        if (offset != 0) {
            segment.setLength(offset);
            if (stem == STATE_NULL) {
                // The first separator after the start of a token. Parse it as a stem.
                stem = parseStem(segment, stemTrie, seen, macros, status);
                stemTrie.reset();
            } else {
                // A separator after the first separator of a token. Parse it as an option.
                stem = parseOption(stem, segment, macros, status);
            }
            segment.resetLength();
            if (U_FAILURE(status)) {
                errOffset = segment.getOffset();
                return macros;
            }

            // Consume the segment:
            segment.adjustOffset(offset);
            offset = 0;

        } else if (stem != STATE_NULL) {
            // A separator ('/' or whitespace) following an option separator ('/')
            // segment.setLength(U16_LENGTH(cp)); // for error message
            // throw new SkeletonSyntaxException("Unexpected separator character", segment);
            status = U_NUMBER_SKELETON_SYNTAX_ERROR;
            errOffset = segment.getOffset();
            return macros;

        } else {
            // Two spaces in a row; this is OK.
        }

        // Does the current stem forbid options?
        if (isOptionSeparator && stem == STATE_NULL) {
            // segment.setLength(U16_LENGTH(cp)); // for error message
            // throw new SkeletonSyntaxException("Unexpected option separator", segment);
            status = U_NUMBER_SKELETON_SYNTAX_ERROR;
            errOffset = segment.getOffset();
            return macros;
        }

        // Does the current stem require an option?
        if (isTokenSeparator && stem != STATE_NULL) {
            switch (stem) {
                case STATE_INCREMENT_PRECISION:
                case STATE_MEASURE_UNIT:
                case STATE_PER_MEASURE_UNIT:
                case STATE_CURRENCY_UNIT:
                case STATE_INTEGER_WIDTH:
                case STATE_NUMBERING_SYSTEM:
                case STATE_SCALE:
                    // segment.setLength(U16_LENGTH(cp)); // for error message
                    // throw new SkeletonSyntaxException("Stem requires an option", segment);
                    status = U_NUMBER_SKELETON_SYNTAX_ERROR;
                    errOffset = segment.getOffset();
                    return macros;
                default:
                    break;
            }
            stem = STATE_NULL;
        }

        // Consume the separator:
        segment.adjustOffset(U16_LENGTH(cp));
    }
    U_ASSERT(stem == STATE_NULL);
    return macros;
}

ParseState
skeleton::parseStem(const StringSegment& segment, const UCharsTrie& stemTrie, SeenMacroProps& seen,
                    MacroProps& macros, UErrorCode& status) {
    // First check for "blueprint" stems, which start with a "signal char"
    switch (segment.charAt(0)) {
        case u'.':
        CHECK_NULL(seen, precision, status);
            blueprint_helpers::parseFractionStem(segment, macros, status);
            return STATE_FRACTION_PRECISION;
        case u'@':
        CHECK_NULL(seen, precision, status);
            blueprint_helpers::parseDigitsStem(segment, macros, status);
            return STATE_NULL;
        default:
            break;
    }

    // Now look at the stemsTrie, which is already be pointing at our stem.
    UStringTrieResult stemResult = stemTrie.current();

    if (stemResult != USTRINGTRIE_INTERMEDIATE_VALUE && stemResult != USTRINGTRIE_FINAL_VALUE) {
        // throw new SkeletonSyntaxException("Unknown stem", segment);
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return STATE_NULL;
    }

    auto stem = static_cast<StemEnum>(stemTrie.getValue());
    switch (stem) {

        // Stems with meaning on their own, not requiring an option:

        case STEM_COMPACT_SHORT:
        case STEM_COMPACT_LONG:
        case STEM_SCIENTIFIC:
        case STEM_ENGINEERING:
        case STEM_NOTATION_SIMPLE:
        CHECK_NULL(seen, notation, status);
            macros.notation = stem_to_object::notation(stem);
            switch (stem) {
                case STEM_SCIENTIFIC:
                case STEM_ENGINEERING:
                    return STATE_SCIENTIFIC; // allows for scientific options
                default:
                    return STATE_NULL;
            }

        case STEM_BASE_UNIT:
        case STEM_PERCENT:
        case STEM_PERMILLE:
        CHECK_NULL(seen, unit, status);
            macros.unit = stem_to_object::unit(stem);
            return STATE_NULL;

        case STEM_PRECISION_INTEGER:
        case STEM_PRECISION_UNLIMITED:
        case STEM_PRECISION_CURRENCY_STANDARD:
        case STEM_PRECISION_CURRENCY_CASH:
        CHECK_NULL(seen, precision, status);
            macros.precision = stem_to_object::precision(stem);
            switch (stem) {
                case STEM_PRECISION_INTEGER:
                    return STATE_FRACTION_PRECISION; // allows for "precision-integer/@##"
                default:
                    return STATE_NULL;
            }

        case STEM_ROUNDING_MODE_CEILING:
        case STEM_ROUNDING_MODE_FLOOR:
        case STEM_ROUNDING_MODE_DOWN:
        case STEM_ROUNDING_MODE_UP:
        case STEM_ROUNDING_MODE_HALF_EVEN:
        case STEM_ROUNDING_MODE_HALF_DOWN:
        case STEM_ROUNDING_MODE_HALF_UP:
        case STEM_ROUNDING_MODE_UNNECESSARY:
        CHECK_NULL(seen, roundingMode, status);
            macros.roundingMode = stem_to_object::roundingMode(stem);
            return STATE_NULL;

        case STEM_GROUP_OFF:
        case STEM_GROUP_MIN2:
        case STEM_GROUP_AUTO:
        case STEM_GROUP_ON_ALIGNED:
        case STEM_GROUP_THOUSANDS:
        CHECK_NULL(seen, grouper, status);
            macros.grouper = Grouper::forStrategy(stem_to_object::groupingStrategy(stem));
            return STATE_NULL;

        case STEM_LATIN:
        CHECK_NULL(seen, symbols, status);
            macros.symbols.setTo(NumberingSystem::createInstanceByName("latn", status));
            return STATE_NULL;

        case STEM_UNIT_WIDTH_NARROW:
        case STEM_UNIT_WIDTH_SHORT:
        case STEM_UNIT_WIDTH_FULL_NAME:
        case STEM_UNIT_WIDTH_ISO_CODE:
        case STEM_UNIT_WIDTH_HIDDEN:
        CHECK_NULL(seen, unitWidth, status);
            macros.unitWidth = stem_to_object::unitWidth(stem);
            return STATE_NULL;

        case STEM_SIGN_AUTO:
        case STEM_SIGN_ALWAYS:
        case STEM_SIGN_NEVER:
        case STEM_SIGN_ACCOUNTING:
        case STEM_SIGN_ACCOUNTING_ALWAYS:
        case STEM_SIGN_EXCEPT_ZERO:
        case STEM_SIGN_ACCOUNTING_EXCEPT_ZERO:
        CHECK_NULL(seen, sign, status);
            macros.sign = stem_to_object::signDisplay(stem);
            return STATE_NULL;

        case STEM_DECIMAL_AUTO:
        case STEM_DECIMAL_ALWAYS:
        CHECK_NULL(seen, decimal, status);
            macros.decimal = stem_to_object::decimalSeparatorDisplay(stem);
            return STATE_NULL;

            // Stems requiring an option:

        case STEM_PRECISION_INCREMENT:
        CHECK_NULL(seen, precision, status);
            return STATE_INCREMENT_PRECISION;

        case STEM_MEASURE_UNIT:
        CHECK_NULL(seen, unit, status);
            return STATE_MEASURE_UNIT;

        case STEM_PER_MEASURE_UNIT:
        CHECK_NULL(seen, perUnit, status);
            return STATE_PER_MEASURE_UNIT;

        case STEM_CURRENCY:
        CHECK_NULL(seen, unit, status);
            return STATE_CURRENCY_UNIT;

        case STEM_INTEGER_WIDTH:
        CHECK_NULL(seen, integerWidth, status);
            return STATE_INTEGER_WIDTH;

        case STEM_NUMBERING_SYSTEM:
        CHECK_NULL(seen, symbols, status);
            return STATE_NUMBERING_SYSTEM;

        case STEM_SCALE:
        CHECK_NULL(seen, scale, status);
            return STATE_SCALE;

        default:
            UPRV_UNREACHABLE;
    }
}

ParseState skeleton::parseOption(ParseState stem, const StringSegment& segment, MacroProps& macros,
                                 UErrorCode& status) {

    ///// Required options: /////

    switch (stem) {
        case STATE_CURRENCY_UNIT:
            blueprint_helpers::parseCurrencyOption(segment, macros, status);
            return STATE_NULL;
        case STATE_MEASURE_UNIT:
            blueprint_helpers::parseMeasureUnitOption(segment, macros, status);
            return STATE_NULL;
        case STATE_PER_MEASURE_UNIT:
            blueprint_helpers::parseMeasurePerUnitOption(segment, macros, status);
            return STATE_NULL;
        case STATE_INCREMENT_PRECISION:
            blueprint_helpers::parseIncrementOption(segment, macros, status);
            return STATE_NULL;
        case STATE_INTEGER_WIDTH:
            blueprint_helpers::parseIntegerWidthOption(segment, macros, status);
            return STATE_NULL;
        case STATE_NUMBERING_SYSTEM:
            blueprint_helpers::parseNumberingSystemOption(segment, macros, status);
            return STATE_NULL;
        case STATE_SCALE:
            blueprint_helpers::parseScaleOption(segment, macros, status);
            return STATE_NULL;
        default:
            break;
    }

    ///// Non-required options: /////

    // Scientific options
    switch (stem) {
        case STATE_SCIENTIFIC:
            if (blueprint_helpers::parseExponentWidthOption(segment, macros, status)) {
                return STATE_SCIENTIFIC;
            }
            if (U_FAILURE(status)) {
                return {};
            }
            if (blueprint_helpers::parseExponentSignOption(segment, macros, status)) {
                return STATE_SCIENTIFIC;
            }
            if (U_FAILURE(status)) {
                return {};
            }
            break;
        default:
            break;
    }

    // Frac-sig option
    switch (stem) {
        case STATE_FRACTION_PRECISION:
            if (blueprint_helpers::parseFracSigOption(segment, macros, status)) {
                return STATE_NULL;
            }
            if (U_FAILURE(status)) {
                return {};
            }
            break;
        default:
            break;
    }

    // Unknown option
    // throw new SkeletonSyntaxException("Invalid option", segment);
    status = U_NUMBER_SKELETON_SYNTAX_ERROR;
    return STATE_NULL;
}

void GeneratorHelpers::generateSkeleton(const MacroProps& macros, UnicodeString& sb, UErrorCode& status) {
    if (U_FAILURE(status)) { return; }

    // Supported options
    if (GeneratorHelpers::notation(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::unit(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::perUnit(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::precision(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::roundingMode(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::grouping(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::integerWidth(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::symbols(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::unitWidth(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::sign(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::decimal(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }
    if (GeneratorHelpers::scale(macros, sb, status)) {
        sb.append(u' ');
    }
    if (U_FAILURE(status)) { return; }

    // Unsupported options
    if (!macros.padder.isBogus()) {
        status = U_UNSUPPORTED_ERROR;
        return;
    }
    if (macros.affixProvider != nullptr) {
        status = U_UNSUPPORTED_ERROR;
        return;
    }
    if (macros.rules != nullptr) {
        status = U_UNSUPPORTED_ERROR;
        return;
    }
    if (macros.currencySymbols != nullptr) {
        status = U_UNSUPPORTED_ERROR;
        return;
    }

    // Remove the trailing space
    if (sb.length() > 0) {
        sb.truncate(sb.length() - 1);
    }
}


bool blueprint_helpers::parseExponentWidthOption(const StringSegment& segment, MacroProps& macros,
                                                 UErrorCode&) {
    if (segment.charAt(0) != u'+') {
        return false;
    }
    int32_t offset = 1;
    int32_t minExp = 0;
    for (; offset < segment.length(); offset++) {
        if (segment.charAt(offset) == u'e') {
            minExp++;
        } else {
            break;
        }
    }
    if (offset < segment.length()) {
        return false;
    }
    // Use the public APIs to enforce bounds checking
    macros.notation = static_cast<ScientificNotation&>(macros.notation).withMinExponentDigits(minExp);
    return true;
}

void
blueprint_helpers::generateExponentWidthOption(int32_t minExponentDigits, UnicodeString& sb, UErrorCode&) {
    sb.append(u'+');
    appendMultiple(sb, u'e', minExponentDigits);
}

bool
blueprint_helpers::parseExponentSignOption(const StringSegment& segment, MacroProps& macros, UErrorCode&) {
    // Get the sign display type out of the CharsTrie data structure.
    UCharsTrie tempStemTrie(kSerializedStemTrie);
    UStringTrieResult result = tempStemTrie.next(
            segment.toTempUnicodeString().getBuffer(),
            segment.length());
    if (result != USTRINGTRIE_INTERMEDIATE_VALUE && result != USTRINGTRIE_FINAL_VALUE) {
        return false;
    }
    auto sign = stem_to_object::signDisplay(static_cast<StemEnum>(tempStemTrie.getValue()));
    if (sign == UNUM_SIGN_COUNT) {
        return false;
    }
    macros.notation = static_cast<ScientificNotation&>(macros.notation).withExponentSignDisplay(sign);
    return true;
}

void blueprint_helpers::parseCurrencyOption(const StringSegment& segment, MacroProps& macros,
                                            UErrorCode& status) {
    // Unlike ICU4J, have to check length manually because ICU4C CurrencyUnit does not check it for us
    if (segment.length() != 3) {
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return;
    }
    const UChar* currencyCode = segment.toTempUnicodeString().getBuffer();
    UErrorCode localStatus = U_ZERO_ERROR;
    CurrencyUnit currency(currencyCode, localStatus);
    if (U_FAILURE(localStatus)) {
        // Not 3 ascii chars
        // throw new SkeletonSyntaxException("Invalid currency", segment);
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return;
    }
    // Slicing is OK
    macros.unit = currency; // NOLINT
}

void
blueprint_helpers::generateCurrencyOption(const CurrencyUnit& currency, UnicodeString& sb, UErrorCode&) {
    sb.append(currency.getISOCurrency(), -1);
}

void blueprint_helpers::parseMeasureUnitOption(const StringSegment& segment, MacroProps& macros,
                                               UErrorCode& status) {
    const UnicodeString stemString = segment.toTempUnicodeString();

    // NOTE: The category (type) of the unit is guaranteed to be a valid subtag (alphanumeric)
    // http://unicode.org/reports/tr35/#Validity_Data
    int firstHyphen = 0;
    while (firstHyphen < stemString.length() && stemString.charAt(firstHyphen) != '-') {
        firstHyphen++;
    }
    if (firstHyphen == stemString.length()) {
        // throw new SkeletonSyntaxException("Invalid measure unit option", segment);
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return;
    }

    // Need to do char <-> UChar conversion...
    U_ASSERT(U_SUCCESS(status));
    CharString type;
    SKELETON_UCHAR_TO_CHAR(type, stemString, 0, firstHyphen, status);
    CharString subType;
    SKELETON_UCHAR_TO_CHAR(subType, stemString, firstHyphen + 1, stemString.length(), status);

    // Note: the largest type as of this writing (March 2018) is "volume", which has 24 units.
    static constexpr int32_t CAPACITY = 30;
    MeasureUnit units[CAPACITY];
    UErrorCode localStatus = U_ZERO_ERROR;
    int32_t numUnits = MeasureUnit::getAvailable(type.data(), units, CAPACITY, localStatus);
    if (U_FAILURE(localStatus)) {
        // More than 30 units in this type?
        status = U_INTERNAL_PROGRAM_ERROR;
        return;
    }
    for (int32_t i = 0; i < numUnits; i++) {
        auto& unit = units[i];
        if (uprv_strcmp(subType.data(), unit.getSubtype()) == 0) {
            macros.unit = unit;
            return;
        }
    }

    // throw new SkeletonSyntaxException("Unknown measure unit", segment);
    status = U_NUMBER_SKELETON_SYNTAX_ERROR;
}

void blueprint_helpers::generateMeasureUnitOption(const MeasureUnit& measureUnit, UnicodeString& sb,
                                                  UErrorCode&) {
    // Need to do char <-> UChar conversion...
    sb.append(UnicodeString(measureUnit.getType(), -1, US_INV));
    sb.append(u'-');
    sb.append(UnicodeString(measureUnit.getSubtype(), -1, US_INV));
}

void blueprint_helpers::parseMeasurePerUnitOption(const StringSegment& segment, MacroProps& macros,
                                                  UErrorCode& status) {
    // A little bit of a hack: safe the current unit (numerator), call the main measure unit
    // parsing code, put back the numerator unit, and put the new unit into per-unit.
    MeasureUnit numerator = macros.unit;
    parseMeasureUnitOption(segment, macros, status);
    if (U_FAILURE(status)) { return; }
    macros.perUnit = macros.unit;
    macros.unit = numerator;
}

void blueprint_helpers::parseFractionStem(const StringSegment& segment, MacroProps& macros,
                                          UErrorCode& status) {
    U_ASSERT(segment.charAt(0) == u'.');
    int32_t offset = 1;
    int32_t minFrac = 0;
    int32_t maxFrac;
    for (; offset < segment.length(); offset++) {
        if (segment.charAt(offset) == u'0') {
            minFrac++;
        } else {
            break;
        }
    }
    if (offset < segment.length()) {
        if (segment.charAt(offset) == u'+') {
            maxFrac = -1;
            offset++;
        } else {
            maxFrac = minFrac;
            for (; offset < segment.length(); offset++) {
                if (segment.charAt(offset) == u'#') {
                    maxFrac++;
                } else {
                    break;
                }
            }
        }
    } else {
        maxFrac = minFrac;
    }
    if (offset < segment.length()) {
        // throw new SkeletonSyntaxException("Invalid fraction stem", segment);
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return;
    }
    // Use the public APIs to enforce bounds checking
    if (maxFrac == -1) {
        macros.precision = Precision::minFraction(minFrac);
    } else {
        macros.precision = Precision::minMaxFraction(minFrac, maxFrac);
    }
}

void
blueprint_helpers::generateFractionStem(int32_t minFrac, int32_t maxFrac, UnicodeString& sb, UErrorCode&) {
    if (minFrac == 0 && maxFrac == 0) {
        sb.append(u"precision-integer", -1);
        return;
    }
    sb.append(u'.');
    appendMultiple(sb, u'0', minFrac);
    if (maxFrac == -1) {
        sb.append(u'+');
    } else {
        appendMultiple(sb, u'#', maxFrac - minFrac);
    }
}

void
blueprint_helpers::parseDigitsStem(const StringSegment& segment, MacroProps& macros, UErrorCode& status) {
    U_ASSERT(segment.charAt(0) == u'@');
    int offset = 0;
    int minSig = 0;
    int maxSig;
    for (; offset < segment.length(); offset++) {
        if (segment.charAt(offset) == u'@') {
            minSig++;
        } else {
            break;
        }
    }
    if (offset < segment.length()) {
        if (segment.charAt(offset) == u'+') {
            maxSig = -1;
            offset++;
        } else {
            maxSig = minSig;
            for (; offset < segment.length(); offset++) {
                if (segment.charAt(offset) == u'#') {
                    maxSig++;
                } else {
                    break;
                }
            }
        }
    } else {
        maxSig = minSig;
    }
    if (offset < segment.length()) {
        // throw new SkeletonSyntaxException("Invalid significant digits stem", segment);
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return;
    }
    // Use the public APIs to enforce bounds checking
    if (maxSig == -1) {
        macros.precision = Precision::minSignificantDigits(minSig);
    } else {
        macros.precision = Precision::minMaxSignificantDigits(minSig, maxSig);
    }
}

void
blueprint_helpers::generateDigitsStem(int32_t minSig, int32_t maxSig, UnicodeString& sb, UErrorCode&) {
    appendMultiple(sb, u'@', minSig);
    if (maxSig == -1) {
        sb.append(u'+');
    } else {
        appendMultiple(sb, u'#', maxSig - minSig);
    }
}

bool blueprint_helpers::parseFracSigOption(const StringSegment& segment, MacroProps& macros,
                                           UErrorCode& status) {
    if (segment.charAt(0) != u'@') {
        return false;
    }
    int offset = 0;
    int minSig = 0;
    int maxSig;
    for (; offset < segment.length(); offset++) {
        if (segment.charAt(offset) == u'@') {
            minSig++;
        } else {
            break;
        }
    }
    // For the frac-sig option, there must be minSig or maxSig but not both.
    // Valid: @+, @@+, @@@+
    // Valid: @#, @##, @###
    // Invalid: @, @@, @@@
    // Invalid: @@#, @@##, @@@#
    if (offset < segment.length()) {
        if (segment.charAt(offset) == u'+') {
            maxSig = -1;
            offset++;
        } else if (minSig > 1) {
            // @@#, @@##, @@@#
            // throw new SkeletonSyntaxException("Invalid digits option for fraction rounder", segment);
            status = U_NUMBER_SKELETON_SYNTAX_ERROR;
            return false;
        } else {
            maxSig = minSig;
            for (; offset < segment.length(); offset++) {
                if (segment.charAt(offset) == u'#') {
                    maxSig++;
                } else {
                    break;
                }
            }
        }
    } else {
        // @, @@, @@@
        // throw new SkeletonSyntaxException("Invalid digits option for fraction rounder", segment);
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return false;
    }
    if (offset < segment.length()) {
        // throw new SkeletonSyntaxException("Invalid digits option for fraction rounder", segment);
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return false;
    }

    auto& oldPrecision = static_cast<const FractionPrecision&>(macros.precision);
    if (maxSig == -1) {
        macros.precision = oldPrecision.withMinDigits(minSig);
    } else {
        macros.precision = oldPrecision.withMaxDigits(maxSig);
    }
    return true;
}

void blueprint_helpers::parseIncrementOption(const StringSegment& segment, MacroProps& macros,
                                             UErrorCode& status) {
    // Need to do char <-> UChar conversion...
    U_ASSERT(U_SUCCESS(status));
    CharString buffer;
    SKELETON_UCHAR_TO_CHAR(buffer, segment.toTempUnicodeString(), 0, segment.length(), status);

    // Utilize DecimalQuantity/decNumber to parse this for us.
    DecimalQuantity dq;
    UErrorCode localStatus = U_ZERO_ERROR;
    dq.setToDecNumber({buffer.data(), buffer.length()}, localStatus);
    if (U_FAILURE(localStatus)) {
        // throw new SkeletonSyntaxException("Invalid rounding increment", segment, e);
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return;
    }
    double increment = dq.toDouble();

    // We also need to figure out how many digits. Do a brute force string operation.
    int decimalOffset = 0;
    while (decimalOffset < segment.length() && segment.charAt(decimalOffset) != '.') {
        decimalOffset++;
    }
    if (decimalOffset == segment.length()) {
        macros.precision = Precision::increment(increment);
    } else {
        int32_t fractionLength = segment.length() - decimalOffset - 1;
        macros.precision = Precision::increment(increment).withMinFraction(fractionLength);
    }
}

void blueprint_helpers::generateIncrementOption(double increment, int32_t trailingZeros, UnicodeString& sb,
                                                UErrorCode&) {
    // Utilize DecimalQuantity/double_conversion to format this for us.
    DecimalQuantity dq;
    dq.setToDouble(increment);
    dq.roundToInfinity();
    sb.append(dq.toPlainString());

    // We might need to append extra trailing zeros for min fraction...
    if (trailingZeros > 0) {
        appendMultiple(sb, u'0', trailingZeros);
    }
}

void blueprint_helpers::parseIntegerWidthOption(const StringSegment& segment, MacroProps& macros,
                                                UErrorCode& status) {
    int32_t offset = 0;
    int32_t minInt = 0;
    int32_t maxInt;
    if (segment.charAt(0) == u'+') {
        maxInt = -1;
        offset++;
    } else {
        maxInt = 0;
    }
    for (; offset < segment.length(); offset++) {
        if (maxInt != -1 && segment.charAt(offset) == u'#') {
            maxInt++;
        } else {
            break;
        }
    }
    if (offset < segment.length()) {
        for (; offset < segment.length(); offset++) {
            if (segment.charAt(offset) == u'0') {
                minInt++;
            } else {
                break;
            }
        }
    }
    if (maxInt != -1) {
        maxInt += minInt;
    }
    if (offset < segment.length()) {
        // throw new SkeletonSyntaxException("Invalid integer width stem", segment);
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return;
    }
    // Use the public APIs to enforce bounds checking
    if (maxInt == -1) {
        macros.integerWidth = IntegerWidth::zeroFillTo(minInt);
    } else {
        macros.integerWidth = IntegerWidth::zeroFillTo(minInt).truncateAt(maxInt);
    }
}

void blueprint_helpers::generateIntegerWidthOption(int32_t minInt, int32_t maxInt, UnicodeString& sb,
                                                   UErrorCode&) {
    if (maxInt == -1) {
        sb.append(u'+');
    } else {
        appendMultiple(sb, u'#', maxInt - minInt);
    }
    appendMultiple(sb, u'0', minInt);
}

void blueprint_helpers::parseNumberingSystemOption(const StringSegment& segment, MacroProps& macros,
                                                   UErrorCode& status) {
    // Need to do char <-> UChar conversion...
    U_ASSERT(U_SUCCESS(status));
    CharString buffer;
    SKELETON_UCHAR_TO_CHAR(buffer, segment.toTempUnicodeString(), 0, segment.length(), status);

    NumberingSystem* ns = NumberingSystem::createInstanceByName(buffer.data(), status);
    if (ns == nullptr || U_FAILURE(status)) {
        // This is a skeleton syntax error; don't bubble up the low-level NumberingSystem error
        // throw new SkeletonSyntaxException("Unknown numbering system", segment);
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return;
    }
    macros.symbols.setTo(ns);
}

void blueprint_helpers::generateNumberingSystemOption(const NumberingSystem& ns, UnicodeString& sb,
                                                      UErrorCode&) {
    // Need to do char <-> UChar conversion...
    sb.append(UnicodeString(ns.getName(), -1, US_INV));
}

void blueprint_helpers::parseScaleOption(const StringSegment& segment, MacroProps& macros,
                                              UErrorCode& status) {
    // Need to do char <-> UChar conversion...
    U_ASSERT(U_SUCCESS(status));
    CharString buffer;
    SKELETON_UCHAR_TO_CHAR(buffer, segment.toTempUnicodeString(), 0, segment.length(), status);

    LocalPointer<DecNum> decnum(new DecNum(), status);
    if (U_FAILURE(status)) { return; }
    decnum->setTo({buffer.data(), buffer.length()}, status);
    if (U_FAILURE(status)) {
        // This is a skeleton syntax error; don't let the low-level decnum error bubble up
        status = U_NUMBER_SKELETON_SYNTAX_ERROR;
        return;
    }

    // NOTE: The constructor will optimize the decnum for us if possible.
    macros.scale = {0, decnum.orphan()};
}

void blueprint_helpers::generateScaleOption(int32_t magnitude, const DecNum* arbitrary, UnicodeString& sb,
                                            UErrorCode& status) {
    // Utilize DecimalQuantity/double_conversion to format this for us.
    DecimalQuantity dq;
    if (arbitrary != nullptr) {
        dq.setToDecNum(*arbitrary, status);
        if (U_FAILURE(status)) { return; }
    } else {
        dq.setToInt(1);
    }
    dq.adjustMagnitude(magnitude);
    dq.roundToInfinity();
    sb.append(dq.toPlainString());
}


bool GeneratorHelpers::notation(const MacroProps& macros, UnicodeString& sb, UErrorCode& status) {
    if (macros.notation.fType == Notation::NTN_COMPACT) {
        UNumberCompactStyle style = macros.notation.fUnion.compactStyle;
        if (style == UNumberCompactStyle::UNUM_LONG) {
            sb.append(u"compact-long", -1);
            return true;
        } else if (style == UNumberCompactStyle::UNUM_SHORT) {
            sb.append(u"compact-short", -1);
            return true;
        } else {
            // Compact notation generated from custom data (not supported in skeleton)
            // The other compact notations are literals
            status = U_UNSUPPORTED_ERROR;
            return false;
        }
    } else if (macros.notation.fType == Notation::NTN_SCIENTIFIC) {
        const Notation::ScientificSettings& impl = macros.notation.fUnion.scientific;
        if (impl.fEngineeringInterval == 3) {
            sb.append(u"engineering", -1);
        } else {
            sb.append(u"scientific", -1);
        }
        if (impl.fMinExponentDigits > 1) {
            sb.append(u'/');
            blueprint_helpers::generateExponentWidthOption(impl.fMinExponentDigits, sb, status);
            if (U_FAILURE(status)) {
                return false;
            }
        }
        if (impl.fExponentSignDisplay != UNUM_SIGN_AUTO) {
            sb.append(u'/');
            enum_to_stem_string::signDisplay(impl.fExponentSignDisplay, sb);
        }
        return true;
    } else {
        // Default value is not shown in normalized form
        return false;
    }
}

bool GeneratorHelpers::unit(const MacroProps& macros, UnicodeString& sb, UErrorCode& status) {
    if (utils::unitIsCurrency(macros.unit)) {
        sb.append(u"currency/", -1);
        CurrencyUnit currency(macros.unit, status);
        if (U_FAILURE(status)) {
            return false;
        }
        blueprint_helpers::generateCurrencyOption(currency, sb, status);
        return true;
    } else if (utils::unitIsNoUnit(macros.unit)) {
        if (utils::unitIsPercent(macros.unit)) {
            sb.append(u"percent", -1);
            return true;
        } else if (utils::unitIsPermille(macros.unit)) {
            sb.append(u"permille", -1);
            return true;
        } else {
            // Default value is not shown in normalized form
            return false;
        }
    } else {
        sb.append(u"measure-unit/", -1);
        blueprint_helpers::generateMeasureUnitOption(macros.unit, sb, status);
        return true;
    }
}

bool GeneratorHelpers::perUnit(const MacroProps& macros, UnicodeString& sb, UErrorCode& status) {
    // Per-units are currently expected to be only MeasureUnits.
    if (utils::unitIsNoUnit(macros.perUnit)) {
        if (utils::unitIsPercent(macros.perUnit) || utils::unitIsPermille(macros.perUnit)) {
            status = U_UNSUPPORTED_ERROR;
            return false;
        } else {
            // Default value: ok to ignore
            return false;
        }
    } else if (utils::unitIsCurrency(macros.perUnit)) {
        status = U_UNSUPPORTED_ERROR;
        return false;
    } else {
        sb.append(u"per-measure-unit/", -1);
        blueprint_helpers::generateMeasureUnitOption(macros.perUnit, sb, status);
        return true;
    }
}

bool GeneratorHelpers::precision(const MacroProps& macros, UnicodeString& sb, UErrorCode& status) {
    if (macros.precision.fType == Precision::RND_NONE) {
        sb.append(u"precision-unlimited", -1);
    } else if (macros.precision.fType == Precision::RND_FRACTION) {
        const Precision::FractionSignificantSettings& impl = macros.precision.fUnion.fracSig;
        blueprint_helpers::generateFractionStem(impl.fMinFrac, impl.fMaxFrac, sb, status);
    } else if (macros.precision.fType == Precision::RND_SIGNIFICANT) {
        const Precision::FractionSignificantSettings& impl = macros.precision.fUnion.fracSig;
        blueprint_helpers::generateDigitsStem(impl.fMinSig, impl.fMaxSig, sb, status);
    } else if (macros.precision.fType == Precision::RND_FRACTION_SIGNIFICANT) {
        const Precision::FractionSignificantSettings& impl = macros.precision.fUnion.fracSig;
        blueprint_helpers::generateFractionStem(impl.fMinFrac, impl.fMaxFrac, sb, status);
        sb.append(u'/');
        if (impl.fMinSig == -1) {
            blueprint_helpers::generateDigitsStem(1, impl.fMaxSig, sb, status);
        } else {
            blueprint_helpers::generateDigitsStem(impl.fMinSig, -1, sb, status);
        }
    } else if (macros.precision.fType == Precision::RND_INCREMENT
            || macros.precision.fType == Precision::RND_INCREMENT_ONE
            || macros.precision.fType == Precision::RND_INCREMENT_FIVE) {
        const Precision::IncrementSettings& impl = macros.precision.fUnion.increment;
        sb.append(u"precision-increment/", -1);
        blueprint_helpers::generateIncrementOption(
                impl.fIncrement,
                impl.fMinFrac - impl.fMaxFrac,
                sb,
                status);
    } else if (macros.precision.fType == Precision::RND_CURRENCY) {
        UCurrencyUsage usage = macros.precision.fUnion.currencyUsage;
        if (usage == UCURR_USAGE_STANDARD) {
            sb.append(u"precision-currency-standard", -1);
        } else {
            sb.append(u"precision-currency-cash", -1);
        }
    } else {
        // Bogus or Error
        return false;
    }

    // NOTE: Always return true for rounding because the default value depends on other options.
    return true;
}

bool GeneratorHelpers::roundingMode(const MacroProps& macros, UnicodeString& sb, UErrorCode&) {
    if (macros.roundingMode == kDefaultMode) {
        return false; // Default
    }
    enum_to_stem_string::roundingMode(macros.roundingMode, sb);
    return true;
}

bool GeneratorHelpers::grouping(const MacroProps& macros, UnicodeString& sb, UErrorCode& status) {
    if (macros.grouper.isBogus()) {
        return false; // No value
    } else if (macros.grouper.fStrategy == UNUM_GROUPING_COUNT) {
        status = U_UNSUPPORTED_ERROR;
        return false;
    } else if (macros.grouper.fStrategy == UNUM_GROUPING_AUTO) {
        return false; // Default value
    } else {
        enum_to_stem_string::groupingStrategy(macros.grouper.fStrategy, sb);
        return true;
    }
}

bool GeneratorHelpers::integerWidth(const MacroProps& macros, UnicodeString& sb, UErrorCode& status) {
    if (macros.integerWidth.fHasError || macros.integerWidth.isBogus() ||
        macros.integerWidth == IntegerWidth::standard()) {
        // Error or Default
        return false;
    }
    sb.append(u"integer-width/", -1);
    blueprint_helpers::generateIntegerWidthOption(
            macros.integerWidth.fUnion.minMaxInt.fMinInt,
            macros.integerWidth.fUnion.minMaxInt.fMaxInt,
            sb,
            status);
    return true;
}

bool GeneratorHelpers::symbols(const MacroProps& macros, UnicodeString& sb, UErrorCode& status) {
    if (macros.symbols.isNumberingSystem()) {
        const NumberingSystem& ns = *macros.symbols.getNumberingSystem();
        if (uprv_strcmp(ns.getName(), "latn") == 0) {
            sb.append(u"latin", -1);
        } else {
            sb.append(u"numbering-system/", -1);
            blueprint_helpers::generateNumberingSystemOption(ns, sb, status);
        }
        return true;
    } else if (macros.symbols.isDecimalFormatSymbols()) {
        status = U_UNSUPPORTED_ERROR;
        return false;
    } else {
        // No custom symbols
        return false;
    }
}

bool GeneratorHelpers::unitWidth(const MacroProps& macros, UnicodeString& sb, UErrorCode&) {
    if (macros.unitWidth == UNUM_UNIT_WIDTH_SHORT || macros.unitWidth == UNUM_UNIT_WIDTH_COUNT) {
        return false; // Default or Bogus
    }
    enum_to_stem_string::unitWidth(macros.unitWidth, sb);
    return true;
}

bool GeneratorHelpers::sign(const MacroProps& macros, UnicodeString& sb, UErrorCode&) {
    if (macros.sign == UNUM_SIGN_AUTO || macros.sign == UNUM_SIGN_COUNT) {
        return false; // Default or Bogus
    }
    enum_to_stem_string::signDisplay(macros.sign, sb);
    return true;
}

bool GeneratorHelpers::decimal(const MacroProps& macros, UnicodeString& sb, UErrorCode&) {
    if (macros.decimal == UNUM_DECIMAL_SEPARATOR_AUTO || macros.decimal == UNUM_DECIMAL_SEPARATOR_COUNT) {
        return false; // Default or Bogus
    }
    enum_to_stem_string::decimalSeparatorDisplay(macros.decimal, sb);
    return true;
}

bool GeneratorHelpers::scale(const MacroProps& macros, UnicodeString& sb, UErrorCode& status) {
    if (!macros.scale.isValid()) {
        return false; // Default or Bogus
    }
    sb.append(u"scale/", -1);
    blueprint_helpers::generateScaleOption(
            macros.scale.fMagnitude,
            macros.scale.fArbitrary,
            sb,
            status);
    return true;
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif

#include <stdlib.h>
#include <cmath>











using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

using icu::double_conversion::DoubleToStringConverter;


namespace {

const char16_t*
doGetPattern(UResourceBundle* res, const char* nsName, const char* patternKey, UErrorCode& publicStatus,
             UErrorCode& localStatus) {
    // Construct the path into the resource bundle
    CharString key;
    key.append("NumberElements/", publicStatus);
    key.append(nsName, publicStatus);
    key.append("/patterns/", publicStatus);
    key.append(patternKey, publicStatus);
    if (U_FAILURE(publicStatus)) {
        return u"";
    }
    return ures_getStringByKeyWithFallback(res, key.data(), nullptr, &localStatus);
}

}


const char16_t* utils::getPatternForStyle(const Locale& locale, const char* nsName, CldrPatternStyle style,
                                          UErrorCode& status) {
    const char* patternKey;
    switch (style) {
        case CLDR_PATTERN_STYLE_DECIMAL:
            patternKey = "decimalFormat";
            break;
        case CLDR_PATTERN_STYLE_CURRENCY:
            patternKey = "currencyFormat";
            break;
        case CLDR_PATTERN_STYLE_ACCOUNTING:
            patternKey = "accountingFormat";
            break;
        case CLDR_PATTERN_STYLE_PERCENT:
            patternKey = "percentFormat";
            break;
        case CLDR_PATTERN_STYLE_SCIENTIFIC:
            patternKey = "scientificFormat";
            break;
        default:
            patternKey = "decimalFormat"; // silence compiler error
            UPRV_UNREACHABLE;
    }
    LocalUResourceBundlePointer res(ures_open(nullptr, locale.getName(), &status));
    if (U_FAILURE(status)) { return u""; }

    // Attempt to get the pattern with the native numbering system.
    UErrorCode localStatus = U_ZERO_ERROR;
    const char16_t* pattern;
    pattern = doGetPattern(res.getAlias(), nsName, patternKey, status, localStatus);
    if (U_FAILURE(status)) { return u""; }

    // Fall back to latn if native numbering system does not have the right pattern
    if (U_FAILURE(localStatus) && uprv_strcmp("latn", nsName) != 0) {
        localStatus = U_ZERO_ERROR;
        pattern = doGetPattern(res.getAlias(), "latn", patternKey, status, localStatus);
        if (U_FAILURE(status)) { return u""; }
    }

    return pattern;
}


DecNum::DecNum() {
    uprv_decContextDefault(&fContext, DEC_INIT_BASE);
    uprv_decContextSetRounding(&fContext, DEC_ROUND_HALF_EVEN);
    fContext.traps = 0; // no traps, thank you (what does this even mean?)
}

DecNum::DecNum(const DecNum& other, UErrorCode& status)
        : fContext(other.fContext) {
    // Allocate memory for the new DecNum.
    U_ASSERT(fContext.digits == other.fData.getCapacity());
    if (fContext.digits > kDefaultDigits) {
        void* p = fData.resize(fContext.digits, 0);
        if (p == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
    }

    // Copy the data from the old DecNum to the new one.
    uprv_memcpy(fData.getAlias(), other.fData.getAlias(), sizeof(decNumber));
    uprv_memcpy(fData.getArrayStart(),
            other.fData.getArrayStart(),
            other.fData.getArrayLimit() - other.fData.getArrayStart());
}

void DecNum::setTo(StringPiece str, UErrorCode& status) {
    // We need NUL-terminated for decNumber; CharString guarantees this, but not StringPiece.
    CharString cstr(str, status);
    if (U_FAILURE(status)) { return; }
    _setTo(cstr.data(), str.length(), status);
}

void DecNum::setTo(const char* str, UErrorCode& status) {
    _setTo(str, static_cast<int32_t>(uprv_strlen(str)), status);
}

void DecNum::setTo(double d, UErrorCode& status) {
    // Need to check for NaN and Infinity before going into DoubleToStringConverter
    if (std::isnan(d) != 0 || std::isfinite(d) == 0) {
        status = U_UNSUPPORTED_ERROR;
        return;
    }

    // First convert from double to string, then string to DecNum.
    // Allocate enough room for: all digits, "E-324", and NUL-terminator.
    char buffer[DoubleToStringConverter::kBase10MaximalLength + 6];
    bool sign; // unused; always positive
    int32_t length;
    int32_t point;
    DoubleToStringConverter::DoubleToAscii(
            d,
            DoubleToStringConverter::DtoaMode::SHORTEST,
            0,
            buffer,
            sizeof(buffer),
            &sign,
            &length,
            &point
    );

    // Read initial result as a string.
    _setTo(buffer, length, status);

    // Set exponent and bitmask. Note that DoubleToStringConverter does not do negatives.
    fData.getAlias()->exponent += point - length;
    fData.getAlias()->bits |= static_cast<uint8_t>(std::signbit(d) ? DECNEG : 0);
}

void DecNum::_setTo(const char* str, int32_t maxDigits, UErrorCode& status) {
    if (maxDigits > kDefaultDigits) {
        fData.resize(maxDigits, 0);
        fContext.digits = maxDigits;
    } else {
        fContext.digits = kDefaultDigits;
    }

    static_assert(DECDPUN == 1, "Assumes that DECDPUN is set to 1");
    uprv_decNumberFromString(fData.getAlias(), str, &fContext);

    // Check for invalid syntax and set the corresponding error code.
    if ((fContext.status & DEC_Conversion_syntax) != 0) {
        status = U_DECIMAL_NUMBER_SYNTAX_ERROR;
        return;
    } else if (fContext.status != 0) {
        // Not a syntax error, but some other error, like an exponent that is too large.
        status = U_UNSUPPORTED_ERROR;
        return;
    }

    // For consistency with Java BigDecimal, no support for DecNum that is NaN or Infinity!
    if (decNumberIsSpecial(fData.getAlias())) {
        status = U_UNSUPPORTED_ERROR;
        return;
    }
}

void
DecNum::setTo(const uint8_t* bcd, int32_t length, int32_t scale, bool isNegative, UErrorCode& status) {
    if (length > kDefaultDigits) {
        fData.resize(length, 0);
        fContext.digits = length;
    } else {
        fContext.digits = kDefaultDigits;
    }

    // "digits is of type int32_t, and must have a value in the range 1 through 999,999,999."
    if (length < 1 || length > 999999999) {
        // Too large for decNumber
        status = U_UNSUPPORTED_ERROR;
        return;
    }
    // "The exponent field holds the exponent of the number. Its range is limited by the requirement that
    // "the range of the adjusted exponent of the number be balanced and fit within a whole number of
    // "decimal digits (in this implementation, be –999,999,999 through +999,999,999). The adjusted
    // "exponent is the exponent that would result if the number were expressed with a single digit before
    // "the decimal point, and is therefore given by exponent+digits-1."
    if (scale > 999999999 - length + 1 || scale < -999999999 - length + 1) {
        // Too large for decNumber
        status = U_UNSUPPORTED_ERROR;
        return;
    }

    fData.getAlias()->digits = length;
    fData.getAlias()->exponent = scale;
    fData.getAlias()->bits = static_cast<uint8_t>(isNegative ? DECNEG : 0);
    uprv_decNumberSetBCD(fData, bcd, static_cast<uint32_t>(length));
    if (fContext.status != 0) {
        // Some error occurred while constructing the decNumber.
        status = U_INTERNAL_PROGRAM_ERROR;
    }
}

void DecNum::normalize() {
    uprv_decNumberReduce(fData, fData, &fContext);
}

void DecNum::multiplyBy(const DecNum& rhs, UErrorCode& status) {
    uprv_decNumberMultiply(fData, fData, rhs.fData, &fContext);
    if (fContext.status != 0) {
        status = U_INTERNAL_PROGRAM_ERROR;
    }
}

void DecNum::divideBy(const DecNum& rhs, UErrorCode& status) {
    uprv_decNumberDivide(fData, fData, rhs.fData, &fContext);
    if ((fContext.status & DEC_Inexact) != 0) {
        // Ignore.
    } else if (fContext.status != 0) {
        status = U_INTERNAL_PROGRAM_ERROR;
    }
}

bool DecNum::isNegative() const {
    return decNumberIsNegative(fData.getAlias());
}

bool DecNum::isZero() const {
    return decNumberIsZero(fData.getAlias());
}

void DecNum::toString(ByteSink& output, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    // "string must be at least dn->digits+14 characters long"
    int32_t minCapacity = fData.getAlias()->digits + 14;
    MaybeStackArray<char, 30> buffer(minCapacity);
    uprv_decNumberToString(fData, buffer.getAlias());
    output.Append(buffer.getAlias(), static_cast<int32_t>(uprv_strlen(buffer.getAlias())));
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2015, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File NUMFMT.CPP
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/18/97    clhuang     Implemented with C++ APIs.
*   04/17/97    aliu        Enlarged MAX_INTEGER_DIGITS to fully accomodate the
*                           largest double, by default.
*                           Changed DigitCount to int per code review.
*    07/20/98    stephen        Changed operator== to check for grouping
*                            Changed setMaxIntegerDigits per Java implementation.
*                            Changed setMinIntegerDigits per Java implementation.
*                            Changed setMinFractionDigits per Java implementation.
*                            Changed setMaxFractionDigits per Java implementation.
********************************************************************************
*/



#if !UCONFIG_NO_FORMATTING













// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 2005-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File WINNMFMT.H
*
********************************************************************************
*/




#if U_PLATFORM_USES_ONLY_WIN32_API







#if !UCONFIG_NO_FORMATTING

/**
 * \file
 * \brief C++ API: Format numbers using Windows API.
 */

U_NAMESPACE_BEGIN

union FormatInfo;

class Win32NumberFormat : public NumberFormat
{
public:
    Win32NumberFormat(const Locale &locale, UBool currency, UErrorCode &status);

    Win32NumberFormat(const Win32NumberFormat &other);

    virtual ~Win32NumberFormat();

    virtual Win32NumberFormat *clone() const;

    Win32NumberFormat &operator=(const Win32NumberFormat &other);

    /**
     * Format a double number. Concrete subclasses must implement
     * these pure virtual methods.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
     */
    virtual UnicodeString& format(double number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos) const;
    /**
     * Format a long number. Concrete subclasses must implement
     * these pure virtual methods.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
    */
    virtual UnicodeString& format(int32_t number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos) const;

    /**
     * Format an int64 number.
     *
     * @param number    The value to be formatted.
     * @param appendTo  Output parameter to receive result.
     *                  Result is appended to existing contents.
     * @param pos       On input: an alignment field, if desired.
     *                  On output: the offsets of the alignment field.
     * @return          Reference to 'appendTo' parameter.
    */
    virtual UnicodeString& format(int64_t number,
                                  UnicodeString& appendTo,
                                  FieldPosition& pos) const;

    using NumberFormat::format;

// Use the default behavior for the following.
//    virtual UnicodeString &format(double number, UnicodeString &appendTo) const;
//    virtual UnicodeString &format(int32_t number, UnicodeString &appendTo) const;
//    virtual UnicodeString &format(int64_t number, UnicodeString &appendTo) const;

    virtual void parse(const UnicodeString& text, Formattable& result, ParsePosition& parsePosition) const;

    /**
     * Sets the maximum number of digits allowed in the fraction portion of a
     * number. maximumFractionDigits must be >= minimumFractionDigits.  If the
     * new value for maximumFractionDigits is less than the current value
     * of minimumFractionDigits, then minimumFractionDigits will also be set to
     * the new value.
     * @param newValue    the new value to be set.
     * @see getMaximumFractionDigits
     */
    virtual void setMaximumFractionDigits(int32_t newValue);

    /**
     * Sets the minimum number of digits allowed in the fraction portion of a
     * number. minimumFractionDigits must be &lt;= maximumFractionDigits.   If the
     * new value for minimumFractionDigits exceeds the current value
     * of maximumFractionDigits, then maximumIntegerDigits will also be set to
     * the new value
     * @param newValue    the new value to be set.
     * @see getMinimumFractionDigits
     */
    virtual void setMinimumFractionDigits(int32_t newValue);

    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       derived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     */
    U_I18N_API static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     */
    virtual UClassID getDynamicClassID(void) const;

private:
    UnicodeString &format(int32_t numDigits, UnicodeString &appendTo, const wchar_t *format, ...) const;

    UBool fCurrency;
    Locale fLocale;
    int32_t fLCID;
    FormatInfo *fFormatInfo;
    UBool fFractionDigitsSet;

    UnicodeString* fWindowsLocaleName; // Stores the equivalent Windows locale name.
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif // U_PLATFORM_USES_ONLY_WIN32_API












#include <float.h>





//#define FMT_DEBUG

#ifdef FMT_DEBUG
#include <stdio.h>
static inline void debugout(UnicodeString s) {
    char buf[2000];
    s.extract((int32_t) 0, s.length(), buf);
    printf("%s", buf);
}
#define debug(x) printf("%s", x);
#else
#define debugout(x)
#define debug(x)
#endif

// If no number pattern can be located for a locale, this is the last
// resort. The patterns are same as the ones in root locale.
static const UChar gLastResortDecimalPat[] = {
    0x23, 0x2C, 0x23, 0x23, 0x30, 0x2E, 0x23, 0x23, 0x23, 0 /* "#,##0.###" */
};
static const UChar gLastResortCurrencyPat[] = {
    0xA4, 0xA0, 0x23, 0x2C, 0x23, 0x23, 0x30, 0x2E, 0x30, 0x30, 0 /* "\u00A4\u00A0#,##0.00" */
};
static const UChar gLastResortPercentPat[] = {
    0x23, 0x2C, 0x23, 0x23, 0x30, 0x25, 0 /* "#,##0%" */
};
static const UChar gLastResortScientificPat[] = {
    0x23, 0x45, 0x30, 0 /* "#E0" */
};
static const UChar gLastResortIsoCurrencyPat[] = {
    0xA4, 0xA4, 0xA0, 0x23, 0x2C, 0x23, 0x23, 0x30, 0x2E, 0x30, 0x30, 0  /* "\u00A4\u00A4\u00A0#,##0.00" */
};
static const UChar gLastResortPluralCurrencyPat[] = {
    0x23, 0x2C, 0x23, 0x23, 0x30, 0x2E, 0x23, 0x23, 0x23, 0x20, 0xA4, 0xA4, 0xA4, 0 /* "#,##0.### \u00A4\u00A4\u00A4*/
};
static const UChar gLastResortAccountingCurrencyPat[] =  {
    0xA4, 0xA0, 0x23, 0x2C, 0x23, 0x23, 0x30, 0x2E, 0x30, 0x30, 0 /* "\u00A4\u00A0#,##0.00" */
};

static const UChar gSingleCurrencySign[] = {0xA4, 0};
static const UChar gDoubleCurrencySign[] = {0xA4, 0xA4, 0};

static const UChar gSlash = 0x2f;

// If the maximum base 10 exponent were 4, then the largest number would
// be 99,999 which has 5 digits.
// On IEEE754 systems gMaxIntegerDigits is 308 + possible denormalized 15 digits + rounding digit
// With big decimal, the max exponent is 999,999,999 and the max number of digits is the same, 999,999,999
const int32_t icu::NumberFormat::gDefaultMaxIntegerDigits = 2000000000;
const int32_t icu::NumberFormat::gDefaultMinIntegerDigits = 127;

static const UChar * const gLastResortNumberPatterns[UNUM_FORMAT_STYLE_COUNT] = {
    NULL,  // UNUM_PATTERN_DECIMAL
    gLastResortDecimalPat,  // UNUM_DECIMAL
    gLastResortCurrencyPat,  // UNUM_CURRENCY
    gLastResortPercentPat,  // UNUM_PERCENT
    gLastResortScientificPat,  // UNUM_SCIENTIFIC
    NULL,  // UNUM_SPELLOUT
    NULL,  // UNUM_ORDINAL
    NULL,  // UNUM_DURATION
    NULL,  // UNUM_NUMBERING_SYSTEM
    NULL,  // UNUM_PATTERN_RULEBASED
    gLastResortIsoCurrencyPat,  // UNUM_CURRENCY_ISO
    gLastResortPluralCurrencyPat,  // UNUM_CURRENCY_PLURAL
    gLastResortAccountingCurrencyPat, // UNUM_CURRENCY_ACCOUNTING
    gLastResortCurrencyPat,  // UNUM_CASH_CURRENCY 
    NULL,  // UNUM_DECIMAL_COMPACT_SHORT
    NULL,  // UNUM_DECIMAL_COMPACT_LONG
    gLastResortCurrencyPat,  // UNUM_CURRENCY_STANDARD
};

// Keys used for accessing resource bundles

static const icu::number::impl::CldrPatternStyle gFormatCldrStyles[UNUM_FORMAT_STYLE_COUNT] = {
    /* NULL */ icu::number::impl::CLDR_PATTERN_STYLE_COUNT,  // UNUM_PATTERN_DECIMAL
    icu::number::impl::CLDR_PATTERN_STYLE_DECIMAL,  // UNUM_DECIMAL
    icu::number::impl::CLDR_PATTERN_STYLE_CURRENCY,  // UNUM_CURRENCY
    icu::number::impl::CLDR_PATTERN_STYLE_PERCENT,  // UNUM_PERCENT
    icu::number::impl::CLDR_PATTERN_STYLE_SCIENTIFIC,  // UNUM_SCIENTIFIC
    /* NULL */ icu::number::impl::CLDR_PATTERN_STYLE_COUNT,  // UNUM_SPELLOUT
    /* NULL */ icu::number::impl::CLDR_PATTERN_STYLE_COUNT,  // UNUM_ORDINAL
    /* NULL */ icu::number::impl::CLDR_PATTERN_STYLE_COUNT,  // UNUM_DURATION
    /* NULL */ icu::number::impl::CLDR_PATTERN_STYLE_COUNT,  // UNUM_NUMBERING_SYSTEM
    /* NULL */ icu::number::impl::CLDR_PATTERN_STYLE_COUNT,  // UNUM_PATTERN_RULEBASED
    // For UNUM_CURRENCY_ISO and UNUM_CURRENCY_PLURAL,
    // the pattern is the same as the pattern of UNUM_CURRENCY
    // except for replacing the single currency sign with
    // double currency sign or triple currency sign.
    icu::number::impl::CLDR_PATTERN_STYLE_CURRENCY,  // UNUM_CURRENCY_ISO
    icu::number::impl::CLDR_PATTERN_STYLE_CURRENCY,  // UNUM_CURRENCY_PLURAL
    icu::number::impl::CLDR_PATTERN_STYLE_ACCOUNTING,  // UNUM_CURRENCY_ACCOUNTING
    icu::number::impl::CLDR_PATTERN_STYLE_CURRENCY,  // UNUM_CASH_CURRENCY
    /* NULL */ icu::number::impl::CLDR_PATTERN_STYLE_COUNT,  // UNUM_DECIMAL_COMPACT_SHORT
    /* NULL */ icu::number::impl::CLDR_PATTERN_STYLE_COUNT,  // UNUM_DECIMAL_COMPACT_LONG
    icu::number::impl::CLDR_PATTERN_STYLE_CURRENCY,  // UNUM_CURRENCY_STANDARD
};

// Static hashtable cache of NumberingSystem objects used by NumberFormat
static UHashtable * NumberingSystem_cache = NULL;
static icu::UInitOnce gNSCacheInitOnce = U_INITONCE_INITIALIZER;

#if !UCONFIG_NO_SERVICE
static icu::ICULocaleService* gService = NULL;
static icu::UInitOnce gServiceInitOnce = U_INITONCE_INITIALIZER;
#endif

/**
 * Release all static memory held by Number Format.
 */
U_CDECL_BEGIN
static void U_CALLCONV
deleteNumberingSystem(void *obj) {
    delete (icu::NumberingSystem *)obj;
}

static UBool U_CALLCONV numfmt_cleanup(void) {
#if !UCONFIG_NO_SERVICE
    gServiceInitOnce.reset();
    if (gService) {
        delete gService;
        gService = NULL;
    }
#endif
    gNSCacheInitOnce.reset();
    if (NumberingSystem_cache) {
        // delete NumberingSystem_cache;
        uhash_close(NumberingSystem_cache);
        NumberingSystem_cache = NULL;
    }
    return TRUE;
}
U_CDECL_END

// *****************************************************************************
// class NumberFormat
// *****************************************************************************

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_ABSTRACT_RTTI_IMPLEMENTATION(NumberFormat)

#if !UCONFIG_NO_SERVICE
// -------------------------------------
// SimpleNumberFormatFactory implementation
NumberFormatFactory::~NumberFormatFactory() {}
SimpleNumberFormatFactory::SimpleNumberFormatFactory(const Locale& locale, UBool visible)
    : _visible(visible)
{
    LocaleUtility::initNameFromLocale(locale, _id);
}

SimpleNumberFormatFactory::~SimpleNumberFormatFactory() {}

UBool SimpleNumberFormatFactory::visible(void) const {
    return _visible;
}

const UnicodeString *
SimpleNumberFormatFactory::getSupportedIDs(int32_t &count, UErrorCode& status) const
{
    if (U_SUCCESS(status)) {
        count = 1;
        return &_id;
    }
    count = 0;
    return NULL;
}
#endif /* #if !UCONFIG_NO_SERVICE */

// -------------------------------------
// default constructor
NumberFormat::NumberFormat()
:   fGroupingUsed(TRUE),
    fMaxIntegerDigits(gDefaultMaxIntegerDigits),
    fMinIntegerDigits(1),
    fMaxFractionDigits(3), // invariant, >= minFractionDigits
    fMinFractionDigits(0),
    fParseIntegerOnly(FALSE),
    fLenient(FALSE),
    fCapitalizationContext(UDISPCTX_CAPITALIZATION_NONE)
{
    fCurrency[0] = 0;
}

// -------------------------------------

NumberFormat::~NumberFormat()
{
}

SharedNumberFormat::~SharedNumberFormat() {
    delete ptr;
}

// -------------------------------------
// copy constructor

NumberFormat::NumberFormat(const NumberFormat &source)
:   Format(source)
{
    *this = source;
}

// -------------------------------------
// assignment operator

NumberFormat&
NumberFormat::operator=(const NumberFormat& rhs)
{
    if (this != &rhs)
    {
        Format::operator=(rhs);
        fGroupingUsed = rhs.fGroupingUsed;
        fMaxIntegerDigits = rhs.fMaxIntegerDigits;
        fMinIntegerDigits = rhs.fMinIntegerDigits;
        fMaxFractionDigits = rhs.fMaxFractionDigits;
        fMinFractionDigits = rhs.fMinFractionDigits;
        fParseIntegerOnly = rhs.fParseIntegerOnly;
        u_strncpy(fCurrency, rhs.fCurrency, 3);
        fCurrency[3] = 0;
        fLenient = rhs.fLenient;
        fCapitalizationContext = rhs.fCapitalizationContext;
    }
    return *this;
}

// -------------------------------------

UBool
NumberFormat::operator==(const Format& that) const
{
    // Format::operator== guarantees this cast is safe
    NumberFormat* other = (NumberFormat*)&that;

#ifdef FMT_DEBUG
    // This code makes it easy to determine why two format objects that should
    // be equal aren't.
    UBool first = TRUE;
    if (!Format::operator==(that)) {
        if (first) { printf("[ "); first = FALSE; } else { printf(", "); }
        debug("Format::!=");
    }
    if (!(fMaxIntegerDigits == other->fMaxIntegerDigits &&
          fMinIntegerDigits == other->fMinIntegerDigits)) {
        if (first) { printf("[ "); first = FALSE; } else { printf(", "); }
        debug("Integer digits !=");
    }
    if (!(fMaxFractionDigits == other->fMaxFractionDigits &&
          fMinFractionDigits == other->fMinFractionDigits)) {
        if (first) { printf("[ "); first = FALSE; } else { printf(", "); }
        debug("Fraction digits !=");
    }
    if (!(fGroupingUsed == other->fGroupingUsed)) {
        if (first) { printf("[ "); first = FALSE; } else { printf(", "); }
        debug("fGroupingUsed != ");
    }
    if (!(fParseIntegerOnly == other->fParseIntegerOnly)) {
        if (first) { printf("[ "); first = FALSE; } else { printf(", "); }
        debug("fParseIntegerOnly != ");
    }
    if (!(u_strcmp(fCurrency, other->fCurrency) == 0)) {
        if (first) { printf("[ "); first = FALSE; } else { printf(", "); }
        debug("fCurrency !=");
    }
    if (!(fLenient == other->fLenient)) {
        if (first) { printf("[ "); first = FALSE; } else { printf(", "); }
        debug("fLenient != ");
    }
    if (!(fCapitalizationContext == other->fCapitalizationContext)) {
        if (first) { printf("[ "); first = FALSE; } else { printf(", "); }
        debug("fCapitalizationContext != ");
    }
    if (!first) { printf(" ]"); }
#endif

    return ((this == &that) ||
            ((Format::operator==(that) &&
              fMaxIntegerDigits == other->fMaxIntegerDigits &&
              fMinIntegerDigits == other->fMinIntegerDigits &&
              fMaxFractionDigits == other->fMaxFractionDigits &&
              fMinFractionDigits == other->fMinFractionDigits &&
              fGroupingUsed == other->fGroupingUsed &&
              fParseIntegerOnly == other->fParseIntegerOnly &&
              u_strcmp(fCurrency, other->fCurrency) == 0 &&
              fLenient == other->fLenient &&
              fCapitalizationContext == other->fCapitalizationContext)));
}

// -------------------------------------
// Default implementation sets unsupported error; subclasses should
// override.

UnicodeString&
NumberFormat::format(double /* unused number */,
                     UnicodeString& toAppendTo,
                     FieldPositionIterator* /* unused posIter */,
                     UErrorCode& status) const
{
    if (!U_FAILURE(status)) {
        status = U_UNSUPPORTED_ERROR;
    }
    return toAppendTo;
}

// -------------------------------------
// Default implementation sets unsupported error; subclasses should
// override.

UnicodeString&
NumberFormat::format(int32_t /* unused number */,
                     UnicodeString& toAppendTo,
                     FieldPositionIterator* /* unused posIter */,
                     UErrorCode& status) const
{
    if (!U_FAILURE(status)) {
        status = U_UNSUPPORTED_ERROR;
    }
    return toAppendTo;
}

// -------------------------------------
// Default implementation sets unsupported error; subclasses should
// override.

UnicodeString&
NumberFormat::format(int64_t /* unused number */,
                     UnicodeString& toAppendTo,
                     FieldPositionIterator* /* unused posIter */,
                     UErrorCode& status) const
{
    if (!U_FAILURE(status)) {
        status = U_UNSUPPORTED_ERROR;
    }
    return toAppendTo;
}

// ------------------------------------------
// These functions add the status code, just fall back to the non-status versions
UnicodeString&
NumberFormat::format(double number,
                     UnicodeString& appendTo,
                     FieldPosition& pos,
                     UErrorCode &status) const {
    if(U_SUCCESS(status)) {
        return format(number,appendTo,pos);
    } else {
        return appendTo;
    }
}

UnicodeString&
NumberFormat::format(int32_t number,
                     UnicodeString& appendTo,
                     FieldPosition& pos,
                     UErrorCode &status) const {
    if(U_SUCCESS(status)) {
        return format(number,appendTo,pos);
    } else {
        return appendTo;
    }
}

UnicodeString&
NumberFormat::format(int64_t number,
                     UnicodeString& appendTo,
                     FieldPosition& pos,
                     UErrorCode &status) const {
    if(U_SUCCESS(status)) {
        return format(number,appendTo,pos);
    } else {
        return appendTo;
    }
}



// -------------------------------------
// Decimal Number format() default implementation 
// Subclasses do not normally override this function, but rather the DigitList
// formatting functions..
//   The expected call chain from here is
//      this function ->
//      NumberFormat::format(Formattable  ->
//      DecimalFormat::format(DigitList    
//
//   Or, for subclasses of Formattable that do not know about DigitList,
//       this Function ->
//       NumberFormat::format(Formattable  ->
//       NumberFormat::format(DigitList  ->
//       XXXFormat::format(double

UnicodeString&
NumberFormat::format(StringPiece decimalNum,
                     UnicodeString& toAppendTo,
                     FieldPositionIterator* fpi,
                     UErrorCode& status) const
{
    Formattable f;
    f.setDecimalNumber(decimalNum, status);
    format(f, toAppendTo, fpi, status);
    return toAppendTo;
}

/**
 *
// Formats the number object and save the format
// result in the toAppendTo string buffer.

// utility to save/restore state, used in two overloads
// of format(const Formattable&...) below.
*
* Old purpose of ArgExtractor was to avoid const. Not thread safe!
*
* keeping it around as a shim.
*/
class ArgExtractor {
  const Formattable* num;
  UChar save[4];
  UBool fWasCurrency;

 public:
  ArgExtractor(const NumberFormat& nf, const Formattable& obj, UErrorCode& status);
  ~ArgExtractor();

  const Formattable* number(void) const;
  const UChar *iso(void) const;
  UBool wasCurrency(void) const;
};

inline const Formattable*
ArgExtractor::number(void) const {
  return num;
}

inline UBool
ArgExtractor::wasCurrency(void) const {
  return fWasCurrency;
}

inline const UChar *
ArgExtractor::iso(void) const {
  return save;
}

ArgExtractor::ArgExtractor(const NumberFormat& /*nf*/, const Formattable& obj, UErrorCode& /*status*/)
  : num(&obj), fWasCurrency(FALSE) {

    const UObject* o = obj.getObject(); // most commonly o==NULL
    const CurrencyAmount* amt;
    if (o != NULL && (amt = dynamic_cast<const CurrencyAmount*>(o)) != NULL) {
        // getISOCurrency() returns a pointer to internal storage, so we
        // copy it to retain it across the call to setCurrency().
        //const UChar* curr = amt->getISOCurrency();
        u_strcpy(save, amt->getISOCurrency());
        num = &amt->getNumber();
        fWasCurrency=TRUE;
    } else {
      save[0]=0;
    }
}

ArgExtractor::~ArgExtractor() {
}

UnicodeString& NumberFormat::format(const number::impl::DecimalQuantity &number,
                      UnicodeString& appendTo,
                      FieldPositionIterator* posIter,
                      UErrorCode& status) const {
    // DecimalFormat overrides this function, and handles DigitList based big decimals.
    // Other subclasses (ChoiceFormat) do not (yet) handle DigitLists,
    // so this default implementation falls back to formatting decimal numbers as doubles.
    if (U_FAILURE(status)) {
        return appendTo;
    }
    double dnum = number.toDouble();
    format(dnum, appendTo, posIter, status);
    return appendTo;
}



UnicodeString&
NumberFormat::format(const number::impl::DecimalQuantity &number,
                     UnicodeString& appendTo,
                     FieldPosition& pos,
                     UErrorCode &status) const {
    // DecimalFormat overrides this function, and handles DigitList based big decimals.
    // Other subclasses (ChoiceFormat) do not (yet) handle DigitLists,
    // so this default implementation falls back to formatting decimal numbers as doubles.
    if (U_FAILURE(status)) {
        return appendTo;
    }
    double dnum = number.toDouble();
    format(dnum, appendTo, pos, status);
    return appendTo;
}

UnicodeString&
NumberFormat::format(const Formattable& obj,
                        UnicodeString& appendTo,
                        FieldPosition& pos,
                        UErrorCode& status) const
{
    if (U_FAILURE(status)) return appendTo;

    ArgExtractor arg(*this, obj, status);
    const Formattable *n = arg.number();
    const UChar *iso = arg.iso();

    if(arg.wasCurrency() && u_strcmp(iso, getCurrency())) {
      // trying to format a different currency.
      // Right now, we clone.
      LocalPointer<NumberFormat> cloneFmt(this->clone());
      cloneFmt->setCurrency(iso, status);
      // next line should NOT recurse, because n is numeric whereas obj was a wrapper around currency amount.
      return cloneFmt->format(*n, appendTo, pos, status);
    }

    if (n->isNumeric() && n->getDecimalQuantity() != NULL) {
        // Decimal Number.  We will have a DigitList available if the value was
        //   set to a decimal number, or if the value originated with a parse.
        //
        // The default implementation for formatting a DigitList converts it
        // to a double, and formats that, allowing formatting classes that don't
        // know about DigitList to continue to operate as they had.
        //
        // DecimalFormat overrides the DigitList formatting functions.
        format(*n->getDecimalQuantity(), appendTo, pos, status);
    } else {
        switch (n->getType()) {
        case Formattable::kDouble:
            format(n->getDouble(), appendTo, pos, status);
            break;
        case Formattable::kLong:
            format(n->getLong(), appendTo, pos, status);
            break;
        case Formattable::kInt64:
            format(n->getInt64(), appendTo, pos, status);
            break;
        default:
            status = U_INVALID_FORMAT_ERROR;
            break;
        }
    }

    return appendTo;
}

// -------------------------------------x
// Formats the number object and save the format
// result in the toAppendTo string buffer.

UnicodeString&
NumberFormat::format(const Formattable& obj,
                        UnicodeString& appendTo,
                        FieldPositionIterator* posIter,
                        UErrorCode& status) const
{
    if (U_FAILURE(status)) return appendTo;

    ArgExtractor arg(*this, obj, status);
    const Formattable *n = arg.number();
    const UChar *iso = arg.iso();

    if(arg.wasCurrency() && u_strcmp(iso, getCurrency())) {
      // trying to format a different currency.
      // Right now, we clone.
      LocalPointer<NumberFormat> cloneFmt(this->clone());
      cloneFmt->setCurrency(iso, status);
      // next line should NOT recurse, because n is numeric whereas obj was a wrapper around currency amount.
      return cloneFmt->format(*n, appendTo, posIter, status);
    }

    if (n->isNumeric() && n->getDecimalQuantity() != NULL) {
        // Decimal Number
        format(*n->getDecimalQuantity(), appendTo, posIter, status);
    } else {
        switch (n->getType()) {
        case Formattable::kDouble:
            format(n->getDouble(), appendTo, posIter, status);
            break;
        case Formattable::kLong:
            format(n->getLong(), appendTo, posIter, status);
            break;
        case Formattable::kInt64:
            format(n->getInt64(), appendTo, posIter, status);
            break;
        default:
            status = U_INVALID_FORMAT_ERROR;
            break;
        }
    }

    return appendTo;
}

// -------------------------------------

UnicodeString&
NumberFormat::format(int64_t number,
                     UnicodeString& appendTo,
                     FieldPosition& pos) const
{
    // default so we don't introduce a new abstract method
    return format((int32_t)number, appendTo, pos);
}

// -------------------------------------
// Parses the string and save the result object as well
// as the final parsed position.

void
NumberFormat::parseObject(const UnicodeString& source,
                             Formattable& result,
                             ParsePosition& parse_pos) const
{
    parse(source, result, parse_pos);
}

// -------------------------------------
// Formats a double number and save the result in a string.

UnicodeString&
NumberFormat::format(double number, UnicodeString& appendTo) const
{
    FieldPosition pos(FieldPosition::DONT_CARE);
    return format(number, appendTo, pos);
}

// -------------------------------------
// Formats a long number and save the result in a string.

UnicodeString&
NumberFormat::format(int32_t number, UnicodeString& appendTo) const
{
    FieldPosition pos(FieldPosition::DONT_CARE);
    return format(number, appendTo, pos);
}

// -------------------------------------
// Formats a long number and save the result in a string.

UnicodeString&
NumberFormat::format(int64_t number, UnicodeString& appendTo) const
{
    FieldPosition pos(FieldPosition::DONT_CARE);
    return format(number, appendTo, pos);
}

// -------------------------------------
// Parses the text and save the result object.  If the returned
// parse position is 0, that means the parsing failed, the status
// code needs to be set to failure.  Ignores the returned parse
// position, otherwise.

void
NumberFormat::parse(const UnicodeString& text,
                        Formattable& result,
                        UErrorCode& status) const
{
    if (U_FAILURE(status)) return;

    ParsePosition parsePosition(0);
    parse(text, result, parsePosition);
    if (parsePosition.getIndex() == 0) {
        status = U_INVALID_FORMAT_ERROR;
    }
}

CurrencyAmount* NumberFormat::parseCurrency(const UnicodeString& text,
                                            ParsePosition& pos) const {
    // Default implementation only -- subclasses should override
    Formattable parseResult;
    int32_t start = pos.getIndex();
    parse(text, parseResult, pos);
    if (pos.getIndex() != start) {
        UChar curr[4];
        UErrorCode ec = U_ZERO_ERROR;
        getEffectiveCurrency(curr, ec);
        if (U_SUCCESS(ec)) {
            LocalPointer<CurrencyAmount> currAmt(new CurrencyAmount(parseResult, curr, ec), ec);
            if (U_FAILURE(ec)) {
                pos.setIndex(start); // indicate failure
            } else {
                return currAmt.orphan();
            }
        }
    }
    return NULL;
}

// -------------------------------------
// Sets to only parse integers.

void
NumberFormat::setParseIntegerOnly(UBool value)
{
    fParseIntegerOnly = value;
}

// -------------------------------------
// Sets whether lenient parse is enabled.

void
NumberFormat::setLenient(UBool enable)
{
    fLenient = enable;
}

// -------------------------------------
// Create a number style NumberFormat instance with the default locale.

NumberFormat* U_EXPORT2
NumberFormat::createInstance(UErrorCode& status)
{
    return createInstance(Locale::getDefault(), UNUM_DECIMAL, status);
}

// -------------------------------------
// Create a number style NumberFormat instance with the inLocale locale.

NumberFormat* U_EXPORT2
NumberFormat::createInstance(const Locale& inLocale, UErrorCode& status)
{
    return createInstance(inLocale, UNUM_DECIMAL, status);
}

// -------------------------------------
// Create a currency style NumberFormat instance with the default locale.

NumberFormat* U_EXPORT2
NumberFormat::createCurrencyInstance(UErrorCode& status)
{
    return createCurrencyInstance(Locale::getDefault(),  status);
}

// -------------------------------------
// Create a currency style NumberFormat instance with the inLocale locale.

NumberFormat* U_EXPORT2
NumberFormat::createCurrencyInstance(const Locale& inLocale, UErrorCode& status)
{
    return createInstance(inLocale, UNUM_CURRENCY, status);
}

// -------------------------------------
// Create a percent style NumberFormat instance with the default locale.

NumberFormat* U_EXPORT2
NumberFormat::createPercentInstance(UErrorCode& status)
{
    return createInstance(Locale::getDefault(), UNUM_PERCENT, status);
}

// -------------------------------------
// Create a percent style NumberFormat instance with the inLocale locale.

NumberFormat* U_EXPORT2
NumberFormat::createPercentInstance(const Locale& inLocale, UErrorCode& status)
{
    return createInstance(inLocale, UNUM_PERCENT, status);
}

// -------------------------------------
// Create a scientific style NumberFormat instance with the default locale.

NumberFormat* U_EXPORT2
NumberFormat::createScientificInstance(UErrorCode& status)
{
    return createInstance(Locale::getDefault(), UNUM_SCIENTIFIC, status);
}

// -------------------------------------
// Create a scientific style NumberFormat instance with the inLocale locale.

NumberFormat* U_EXPORT2
NumberFormat::createScientificInstance(const Locale& inLocale, UErrorCode& status)
{
    return createInstance(inLocale, UNUM_SCIENTIFIC, status);
}

// -------------------------------------

const Locale* U_EXPORT2
NumberFormat::getAvailableLocales(int32_t& count)
{
    return Locale::getAvailableLocales(count);
}

// ------------------------------------------
//
// Registration
//
//-------------------------------------------

#if !UCONFIG_NO_SERVICE

// -------------------------------------

class ICUNumberFormatFactory : public ICUResourceBundleFactory {
public:
    virtual ~ICUNumberFormatFactory();
protected:
    virtual UObject* handleCreate(const Locale& loc, int32_t kind, const ICUService* /* service */, UErrorCode& status) const {
        return NumberFormat::makeInstance(loc, (UNumberFormatStyle)kind, status);
    }
};

ICUNumberFormatFactory::~ICUNumberFormatFactory() {}

// -------------------------------------

class NFFactory : public LocaleKeyFactory {
private:
    NumberFormatFactory* _delegate;
    Hashtable* _ids;

public:
    NFFactory(NumberFormatFactory* delegate)
        : LocaleKeyFactory(delegate->visible() ? VISIBLE : INVISIBLE)
        , _delegate(delegate)
        , _ids(NULL)
    {
    }

    virtual ~NFFactory();

    virtual UObject* create(const ICUServiceKey& key, const ICUService* service, UErrorCode& status) const
    {
        if (handlesKey(key, status)) {
            const LocaleKey& lkey = (const LocaleKey&)key;
            Locale loc;
            lkey.canonicalLocale(loc);
            int32_t kind = lkey.kind();

            UObject* result = _delegate->createFormat(loc, (UNumberFormatStyle)kind);
            if (result == NULL) {
                result = service->getKey((ICUServiceKey&)key /* cast away const */, NULL, this, status);
            }
            return result;
        }
        return NULL;
    }

protected:
    /**
     * Return the set of ids that this factory supports (visible or
     * otherwise).  This can be called often and might need to be
     * cached if it is expensive to create.
     */
    virtual const Hashtable* getSupportedIDs(UErrorCode& status) const
    {
        if (U_SUCCESS(status)) {
            if (!_ids) {
                int32_t count = 0;
                const UnicodeString * const idlist = _delegate->getSupportedIDs(count, status);
                ((NFFactory*)this)->_ids = new Hashtable(status); /* cast away const */
                if (_ids) {
                    for (int i = 0; i < count; ++i) {
                        _ids->put(idlist[i], (void*)this, status);
                    }
                }
            }
            return _ids;
        }
        return NULL;
    }
};

NFFactory::~NFFactory()
{
    delete _delegate;
    delete _ids;
}

class ICUNumberFormatService : public ICULocaleService {
public:
    ICUNumberFormatService()
        : ICULocaleService(UNICODE_STRING_SIMPLE("Number Format"))
    {
        UErrorCode status = U_ZERO_ERROR;
        registerFactory(new ICUNumberFormatFactory(), status);
    }

    virtual ~ICUNumberFormatService();

    virtual UObject* cloneInstance(UObject* instance) const {
        return ((NumberFormat*)instance)->clone();
    }

    virtual UObject* handleDefault(const ICUServiceKey& key, UnicodeString* /* actualID */, UErrorCode& status) const {
        LocaleKey& lkey = (LocaleKey&)key;
        int32_t kind = lkey.kind();
        Locale loc;
        lkey.currentLocale(loc);
        return NumberFormat::makeInstance(loc, (UNumberFormatStyle)kind, status);
    }

    virtual UBool isDefault() const {
        return countFactories() == 1;
    }
};

ICUNumberFormatService::~ICUNumberFormatService() {}

// -------------------------------------

static void U_CALLCONV initNumberFormatService() {
    U_ASSERT(gService == NULL);
    ucln_i18n_registerCleanup(UCLN_I18N_NUMFMT, numfmt_cleanup);
    gService = new ICUNumberFormatService();
}

static ICULocaleService*
getNumberFormatService(void)
{
    umtx_initOnce(gServiceInitOnce, &initNumberFormatService);
    return gService;
}

static UBool haveService() {
    return !gServiceInitOnce.isReset() && (getNumberFormatService() != NULL);
}

// -------------------------------------

URegistryKey U_EXPORT2
NumberFormat::registerFactory(NumberFormatFactory* toAdopt, UErrorCode& status)
{
    if (U_FAILURE(status)) {
        delete toAdopt;
        return nullptr;
    }
    ICULocaleService *service = getNumberFormatService();
    if (service) {
        NFFactory *tempnnf = new NFFactory(toAdopt);
        if (tempnnf != NULL) {
            return service->registerFactory(tempnnf, status);
        }
    }
    status = U_MEMORY_ALLOCATION_ERROR;
    return NULL;
}

// -------------------------------------

UBool U_EXPORT2
NumberFormat::unregister(URegistryKey key, UErrorCode& status)
{
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (haveService()) {
        return gService->unregister(key, status);
    } else {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return FALSE;
    }
}

// -------------------------------------
StringEnumeration* U_EXPORT2
NumberFormat::getAvailableLocales(void)
{
  ICULocaleService *service = getNumberFormatService();
  if (service) {
      return service->getAvailableLocales();
  }
  return NULL; // no way to return error condition
}
#endif /* UCONFIG_NO_SERVICE */
// -------------------------------------

enum { kKeyValueLenMax = 32 };

NumberFormat*
NumberFormat::internalCreateInstance(const Locale& loc, UNumberFormatStyle kind, UErrorCode& status) {
    if (kind == UNUM_CURRENCY) {
        char cfKeyValue[kKeyValueLenMax] = {0};
        UErrorCode kvStatus = U_ZERO_ERROR;
        int32_t kLen = loc.getKeywordValue("cf", cfKeyValue, kKeyValueLenMax, kvStatus);
        if (U_SUCCESS(kvStatus) && kLen > 0 && uprv_strcmp(cfKeyValue,"account")==0) {
            kind = UNUM_CURRENCY_ACCOUNTING;
        }
    }
#if !UCONFIG_NO_SERVICE
    if (haveService()) {
        return (NumberFormat*)gService->get(loc, kind, status);
    }
#endif
    return makeInstance(loc, kind, status);
}

NumberFormat* U_EXPORT2
NumberFormat::createInstance(const Locale& loc, UNumberFormatStyle kind, UErrorCode& status) {
    if (kind != UNUM_DECIMAL) {
        return internalCreateInstance(loc, kind, status);
    }
    const SharedNumberFormat *shared = createSharedInstance(loc, kind, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    NumberFormat *result = (*shared)->clone();
    shared->removeRef();
    if (result == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return result;
}
    

// -------------------------------------
// Checks if the thousand/10 thousand grouping is used in the
// NumberFormat instance.

UBool
NumberFormat::isGroupingUsed() const
{
    return fGroupingUsed;
}

// -------------------------------------
// Sets to use the thousand/10 thousand grouping in the
// NumberFormat instance.

void
NumberFormat::setGroupingUsed(UBool newValue)
{
    fGroupingUsed = newValue;
}

// -------------------------------------
// Gets the maximum number of digits for the integral part for
// this NumberFormat instance.

int32_t NumberFormat::getMaximumIntegerDigits() const
{
    return fMaxIntegerDigits;
}

// -------------------------------------
// Sets the maximum number of digits for the integral part for
// this NumberFormat instance.

void
NumberFormat::setMaximumIntegerDigits(int32_t newValue)
{
    fMaxIntegerDigits = uprv_max(0, uprv_min(newValue, gDefaultMaxIntegerDigits));
    if(fMinIntegerDigits > fMaxIntegerDigits)
        fMinIntegerDigits = fMaxIntegerDigits;
}

// -------------------------------------
// Gets the minimum number of digits for the integral part for
// this NumberFormat instance.

int32_t
NumberFormat::getMinimumIntegerDigits() const
{
    return fMinIntegerDigits;
}

// -------------------------------------
// Sets the minimum number of digits for the integral part for
// this NumberFormat instance.

void
NumberFormat::setMinimumIntegerDigits(int32_t newValue)
{
    fMinIntegerDigits = uprv_max(0, uprv_min(newValue, gDefaultMinIntegerDigits));
    if(fMinIntegerDigits > fMaxIntegerDigits)
        fMaxIntegerDigits = fMinIntegerDigits;
}

// -------------------------------------
// Gets the maximum number of digits for the fractional part for
// this NumberFormat instance.

int32_t
NumberFormat::getMaximumFractionDigits() const
{
    return fMaxFractionDigits;
}

// -------------------------------------
// Sets the maximum number of digits for the fractional part for
// this NumberFormat instance.

void
NumberFormat::setMaximumFractionDigits(int32_t newValue)
{
    fMaxFractionDigits = uprv_max(0, uprv_min(newValue, gDefaultMaxIntegerDigits));
    if(fMaxFractionDigits < fMinFractionDigits)
        fMinFractionDigits = fMaxFractionDigits;
}

// -------------------------------------
// Gets the minimum number of digits for the fractional part for
// this NumberFormat instance.

int32_t
NumberFormat::getMinimumFractionDigits() const
{
    return fMinFractionDigits;
}

// -------------------------------------
// Sets the minimum number of digits for the fractional part for
// this NumberFormat instance.

void
NumberFormat::setMinimumFractionDigits(int32_t newValue)
{
    fMinFractionDigits = uprv_max(0, uprv_min(newValue, gDefaultMinIntegerDigits));
    if (fMaxFractionDigits < fMinFractionDigits)
        fMaxFractionDigits = fMinFractionDigits;
}

// -------------------------------------

void NumberFormat::setCurrency(const UChar* theCurrency, UErrorCode& ec) {
    if (U_FAILURE(ec)) {
        return;
    }
    if (theCurrency) {
        u_strncpy(fCurrency, theCurrency, 3);
        fCurrency[3] = 0;
    } else {
        fCurrency[0] = 0;
    }
}

const char16_t* NumberFormat::getCurrency() const {
    return fCurrency;
}

void NumberFormat::getEffectiveCurrency(UChar* result, UErrorCode& ec) const {
    const UChar* c = getCurrency();
    if (*c != 0) {
        u_strncpy(result, c, 3);
        result[3] = 0;
    } else {
        const char* loc = getLocaleID(ULOC_VALID_LOCALE, ec);
        if (loc == NULL) {
            loc = uloc_getDefault();
        }
        ucurr_forLocale(loc, result, 4, &ec);
    }
}

//----------------------------------------------------------------------


void NumberFormat::setContext(UDisplayContext value, UErrorCode& status)
{
    if (U_FAILURE(status))
        return;
    if ( (UDisplayContextType)((uint32_t)value >> 8) == UDISPCTX_TYPE_CAPITALIZATION ) {
        fCapitalizationContext = value;
    } else {
        status = U_ILLEGAL_ARGUMENT_ERROR;
   }
}


UDisplayContext NumberFormat::getContext(UDisplayContextType type, UErrorCode& status) const
{
    if (U_FAILURE(status))
        return (UDisplayContext)0;
    if (type != UDISPCTX_TYPE_CAPITALIZATION) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return (UDisplayContext)0;
    }
    return fCapitalizationContext;
}


// -------------------------------------
// Creates the NumberFormat instance of the specified style (number, currency,
// or percent) for the desired locale.

static void U_CALLCONV nscacheInit() {
    U_ASSERT(NumberingSystem_cache == NULL);
    ucln_i18n_registerCleanup(UCLN_I18N_NUMFMT, numfmt_cleanup);
    UErrorCode status = U_ZERO_ERROR;
    NumberingSystem_cache = uhash_open(uhash_hashLong,
                                       uhash_compareLong,
                                       NULL,
                                       &status);
    if (U_FAILURE(status)) {
        // Number Format code will run with no cache if creation fails.
        NumberingSystem_cache = NULL;
        return;
    }
    uhash_setValueDeleter(NumberingSystem_cache, deleteNumberingSystem);
}

template<> U_I18N_API
const SharedNumberFormat *LocaleCacheKey<SharedNumberFormat>::createObject(
        const void * /*unused*/, UErrorCode &status) const {
    const char *localeId = fLoc.getName();
    NumberFormat *nf = NumberFormat::internalCreateInstance(
            localeId, UNUM_DECIMAL, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    SharedNumberFormat *result = new SharedNumberFormat(nf);
    if (result == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        delete nf;
        return NULL;
    }
    result->addRef();
    return result;
}

const SharedNumberFormat* U_EXPORT2
NumberFormat::createSharedInstance(const Locale& loc, UNumberFormatStyle kind, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    if (kind != UNUM_DECIMAL) {
        status = U_UNSUPPORTED_ERROR;
        return NULL;
    }
    const SharedNumberFormat *result = NULL;
    UnifiedCache::getByLocale(loc, result, status);
    return result;
}

UBool
NumberFormat::isStyleSupported(UNumberFormatStyle style) {
    return gLastResortNumberPatterns[style] != NULL;
}

NumberFormat*
NumberFormat::makeInstance(const Locale& desiredLocale,
                           UNumberFormatStyle style,
                           UErrorCode& status) {
  return makeInstance(desiredLocale, style, false, status);
}

NumberFormat*
NumberFormat::makeInstance(const Locale& desiredLocale,
                           UNumberFormatStyle style,
                           UBool mustBeDecimalFormat,
                           UErrorCode& status) {
    if (U_FAILURE(status)) return NULL;

    if (style < 0 || style >= UNUM_FORMAT_STYLE_COUNT) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    // Some styles are not supported. This is a result of merging
    // the @draft ICU 4.2 NumberFormat::EStyles into the long-existing UNumberFormatStyle.
    // Ticket #8503 is for reviewing/fixing/merging the two relevant implementations:
    // this one and unum_open().
    // The UNUM_PATTERN_ styles are not supported here
    // because this method does not take a pattern string.
    if (!isStyleSupported(style)) {
        status = U_UNSUPPORTED_ERROR;
        return NULL;
    }

#if U_PLATFORM_USES_ONLY_WIN32_API
    if (!mustBeDecimalFormat) {
        char buffer[8];
        int32_t count = desiredLocale.getKeywordValue("compat", buffer, sizeof(buffer), status);

        // if the locale has "@compat=host", create a host-specific NumberFormat
        if (U_SUCCESS(status) && count > 0 && uprv_strcmp(buffer, "host") == 0) {
            UBool curr = TRUE;

            switch (style) {
            case UNUM_DECIMAL:
                curr = FALSE;
                // fall-through
                U_FALLTHROUGH;

            case UNUM_CURRENCY:
            case UNUM_CURRENCY_ISO: // do not support plural formatting here
            case UNUM_CURRENCY_PLURAL:
            case UNUM_CURRENCY_ACCOUNTING:
            case UNUM_CASH_CURRENCY:
            case UNUM_CURRENCY_STANDARD:
            {
                LocalPointer<Win32NumberFormat> f(new Win32NumberFormat(desiredLocale, curr, status), status);
                if (U_SUCCESS(status)) {
                    return f.orphan();
                }
            }
            break;
            default:
                break;
            }
        }
    }
#endif
    // Use numbering system cache hashtable
    umtx_initOnce(gNSCacheInitOnce, &nscacheInit);

    // Get cached numbering system
    LocalPointer<NumberingSystem> ownedNs;
    NumberingSystem *ns = NULL;
    if (NumberingSystem_cache != NULL) {
        // TODO: Bad hash key usage, see ticket #8504.
        int32_t hashKey = desiredLocale.hashCode();

        static UMutex nscacheMutex;
        Mutex lock(&nscacheMutex);
        ns = (NumberingSystem *)uhash_iget(NumberingSystem_cache, hashKey);
        if (ns == NULL) {
            ns = NumberingSystem::createInstance(desiredLocale,status);
            uhash_iput(NumberingSystem_cache, hashKey, (void*)ns, &status);
        }
    } else {
        ownedNs.adoptInstead(NumberingSystem::createInstance(desiredLocale,status));
        ns = ownedNs.getAlias();
    }

    // check results of getting a numbering system
    if (U_FAILURE(status)) {
        return NULL;
    }

    if (mustBeDecimalFormat && ns->isAlgorithmic()) {
        status = U_UNSUPPORTED_ERROR;
        return NULL;
    }

    LocalPointer<DecimalFormatSymbols> symbolsToAdopt;
    UnicodeString pattern;
    LocalUResourceBundlePointer ownedResource(ures_open(NULL, desiredLocale.getName(), &status));
    if (U_FAILURE(status)) {
        return NULL;
    }
    else {
        // Loads the decimal symbols of the desired locale.
        symbolsToAdopt.adoptInsteadAndCheckErrorCode(new DecimalFormatSymbols(desiredLocale, status), status);
        if (U_FAILURE(status)) {
            return NULL;
        }

        // Load the pattern from data using the common library function
        const char16_t* patternPtr = number::impl::utils::getPatternForStyle(
                desiredLocale,
                ns->getName(),
                gFormatCldrStyles[style],
                status);
        pattern = UnicodeString(TRUE, patternPtr, -1);
    }
    if (U_FAILURE(status)) {
        return NULL;
    }
    if(style==UNUM_CURRENCY || style == UNUM_CURRENCY_ISO || style == UNUM_CURRENCY_ACCOUNTING 
        || style == UNUM_CASH_CURRENCY || style == UNUM_CURRENCY_STANDARD){
        const UChar* currPattern = symbolsToAdopt->getCurrencyPattern();
        if(currPattern!=NULL){
            pattern.setTo(currPattern, u_strlen(currPattern));
        }
    }

    LocalPointer<NumberFormat> f;
    if (ns->isAlgorithmic()) {
        UnicodeString nsDesc;
        UnicodeString nsRuleSetGroup;
        UnicodeString nsRuleSetName;
        Locale nsLoc;
        URBNFRuleSetTag desiredRulesType = URBNF_NUMBERING_SYSTEM;

        nsDesc.setTo(ns->getDescription());
        int32_t firstSlash = nsDesc.indexOf(gSlash);
        int32_t lastSlash = nsDesc.lastIndexOf(gSlash);
        if ( lastSlash > firstSlash ) {
            CharString nsLocID;

            nsLocID.appendInvariantChars(nsDesc.tempSubString(0, firstSlash), status);
            nsRuleSetGroup.setTo(nsDesc,firstSlash+1,lastSlash-firstSlash-1);
            nsRuleSetName.setTo(nsDesc,lastSlash+1);

            nsLoc = Locale::createFromName(nsLocID.data());

            UnicodeString SpelloutRules = UNICODE_STRING_SIMPLE("SpelloutRules");
            if ( nsRuleSetGroup.compare(SpelloutRules) == 0 ) {
                desiredRulesType = URBNF_SPELLOUT;
            }
        } else {
            nsLoc = desiredLocale;
            nsRuleSetName.setTo(nsDesc);
        }

        RuleBasedNumberFormat *r = new RuleBasedNumberFormat(desiredRulesType,nsLoc,status);
        if (r == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        r->setDefaultRuleSet(nsRuleSetName,status);
        f.adoptInstead(r);
    } else {
        // replace single currency sign in the pattern with double currency sign
        // if the style is UNUM_CURRENCY_ISO
        if (style == UNUM_CURRENCY_ISO) {
            pattern.findAndReplace(UnicodeString(TRUE, gSingleCurrencySign, 1),
                                   UnicodeString(TRUE, gDoubleCurrencySign, 2));
        }

        // "new DecimalFormat()" does not adopt the symbols argument if its memory allocation fails.
        // So we can't use adoptInsteadAndCheckErrorCode as we need to know if the 'new' failed.
        DecimalFormatSymbols *syms = symbolsToAdopt.getAlias();
        LocalPointer<DecimalFormat> df(new DecimalFormat(pattern, syms, style, status));

        if (df.isValid()) {
            // if the DecimalFormat object was successfully new'ed, then it will own symbolsToAdopt, even if the status is a failure.
            symbolsToAdopt.orphan();
        }
        else {
            status = U_MEMORY_ALLOCATION_ERROR;
        }

        if (U_FAILURE(status)) {
            return nullptr;
        }

        // if it is cash currency style, setCurrencyUsage with usage
        if (style == UNUM_CASH_CURRENCY){
            df->setCurrencyUsage(UCURR_USAGE_CASH, &status);
        }

        if (U_FAILURE(status)) {
            return nullptr;
        }

        f.adoptInstead(df.orphan());
    }

    f->setLocaleIDs(ures_getLocaleByType(ownedResource.getAlias(), ULOC_VALID_LOCALE, &status),
                    ures_getLocaleByType(ownedResource.getAlias(), ULOC_ACTUAL_LOCALE, &status));
    if (U_FAILURE(status)) {
        return NULL;
    }
    return f.orphan();
}

/**
 * Get the rounding mode.
 * @return A rounding mode
 */
NumberFormat::ERoundingMode NumberFormat::getRoundingMode() const {
    // Default value. ICU4J throws an exception and we can't change this API.
    return NumberFormat::ERoundingMode::kRoundUnnecessary;
}

/**
 * Set the rounding mode.  This has no effect unless the rounding
 * increment is greater than zero.
 * @param roundingMode A rounding mode
 */
void NumberFormat::setRoundingMode(NumberFormat::ERoundingMode /*roundingMode*/) {
    // No-op ICU4J throws an exception, and we can't change this API.
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif



// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN namespace numparse {
namespace impl {
namespace utils {


// inline static void putLeadCodePoints(const UnicodeSet* input, UnicodeSet* output) {
//     for (int32_t i = 0; i < input->getRangeCount(); i++) {
//         output->add(input->getRangeStart(i), input->getRangeEnd(i));
//     }
//     // TODO: ANDY: How to iterate over the strings in ICU4C UnicodeSet?
// }

// inline static void putLeadCodePoint(const UnicodeString& input, UnicodeSet* output) {
//     if (!input.isEmpty()) {
//         output->add(input.char32At(0));
//     }
// }

inline static void copyCurrencyCode(UChar* dest, const UChar* src) {
    uprv_memcpy(dest, src, sizeof(UChar) * 3);
    dest[3] = 0;
}


} // namespace utils
} // namespace impl
} // namespace numparse
U_NAMESPACE_END

#endif //__NUMPARSE_UTILS_H__




using namespace icu;
using namespace icu::numparse;
using namespace icu::numparse::impl;
using namespace icu::number;
using namespace icu::number::impl;


namespace {

/**
 * Helper method to return whether the given AffixPatternMatcher equals the given pattern string.
 * Either both arguments must be null or the pattern string inside the AffixPatternMatcher must equal
 * the given pattern string.
 */
static bool matched(const AffixPatternMatcher* affix, const UnicodeString& patternString) {
    return (affix == nullptr && patternString.isBogus()) ||
           (affix != nullptr && affix->getPattern() == patternString);
}

/**
 * Helper method to return the length of the given AffixPatternMatcher. Returns 0 for null.
 */
static int32_t length(const AffixPatternMatcher* matcher) {
    return matcher == nullptr ? 0 : matcher->getPattern().length();
}

/**
 * Helper method to return whether (1) both lhs and rhs are null/invalid, or (2) if they are both
 * valid, whether they are equal according to operator==.  Similar to Java Objects.equals()
 */
static bool equals(const AffixPatternMatcher* lhs, const AffixPatternMatcher* rhs) {
    if (lhs == nullptr && rhs == nullptr) {
        return true;
    }
    if (lhs == nullptr || rhs == nullptr) {
        return false;
    }
    return *lhs == *rhs;
}

}


AffixPatternMatcherBuilder::AffixPatternMatcherBuilder(const UnicodeString& pattern,
                                                       AffixTokenMatcherWarehouse& warehouse,
                                                       IgnorablesMatcher* ignorables)
        : fMatchersLen(0),
          fLastTypeOrCp(0),
          fPattern(pattern),
          fWarehouse(warehouse),
          fIgnorables(ignorables) {}

void AffixPatternMatcherBuilder::consumeToken(AffixPatternType type, UChar32 cp, UErrorCode& status) {
    // This is called by AffixUtils.iterateWithConsumer() for each token.

    // Add an ignorables matcher between tokens except between two literals, and don't put two
    // ignorables matchers in a row.
    if (fIgnorables != nullptr && fMatchersLen > 0 &&
        (fLastTypeOrCp < 0 || !fIgnorables->getSet()->contains(fLastTypeOrCp))) {
        addMatcher(*fIgnorables);
    }

    if (type != TYPE_CODEPOINT) {
        // Case 1: the token is a symbol.
        switch (type) {
            case TYPE_MINUS_SIGN:
                addMatcher(fWarehouse.minusSign());
                break;
            case TYPE_PLUS_SIGN:
                addMatcher(fWarehouse.plusSign());
                break;
            case TYPE_PERCENT:
                addMatcher(fWarehouse.percent());
                break;
            case TYPE_PERMILLE:
                addMatcher(fWarehouse.permille());
                break;
            case TYPE_CURRENCY_SINGLE:
            case TYPE_CURRENCY_DOUBLE:
            case TYPE_CURRENCY_TRIPLE:
            case TYPE_CURRENCY_QUAD:
            case TYPE_CURRENCY_QUINT:
                // All currency symbols use the same matcher
                addMatcher(fWarehouse.currency(status));
                break;
            default:
                UPRV_UNREACHABLE;
        }

    } else if (fIgnorables != nullptr && fIgnorables->getSet()->contains(cp)) {
        // Case 2: the token is an ignorable literal.
        // No action necessary: the ignorables matcher has already been added.

    } else {
        // Case 3: the token is a non-ignorable literal.
        if (auto* ptr = fWarehouse.nextCodePointMatcher(cp, status)) {
            addMatcher(*ptr);
        } else {
            // OOM; unwind the stack
            return;
        }
    }
    fLastTypeOrCp = type != TYPE_CODEPOINT ? type : cp;
}

void AffixPatternMatcherBuilder::addMatcher(NumberParseMatcher& matcher) {
    if (fMatchersLen >= fMatchers.getCapacity()) {
        fMatchers.resize(fMatchersLen * 2, fMatchersLen);
    }
    fMatchers[fMatchersLen++] = &matcher;
}

AffixPatternMatcher AffixPatternMatcherBuilder::build() {
    return AffixPatternMatcher(fMatchers, fMatchersLen, fPattern);
}

AffixTokenMatcherWarehouse::AffixTokenMatcherWarehouse(const AffixTokenMatcherSetupData* setupData)
        : fSetupData(setupData) {}

NumberParseMatcher& AffixTokenMatcherWarehouse::minusSign() {
    return fMinusSign = {fSetupData->dfs, true};
}

NumberParseMatcher& AffixTokenMatcherWarehouse::plusSign() {
    return fPlusSign = {fSetupData->dfs, true};
}

NumberParseMatcher& AffixTokenMatcherWarehouse::percent() {
    return fPercent = {fSetupData->dfs};
}

NumberParseMatcher& AffixTokenMatcherWarehouse::permille() {
    return fPermille = {fSetupData->dfs};
}

NumberParseMatcher& AffixTokenMatcherWarehouse::currency(UErrorCode& status) {
    return fCurrency = {fSetupData->currencySymbols, fSetupData->dfs, fSetupData->parseFlags, status};
}

IgnorablesMatcher& AffixTokenMatcherWarehouse::ignorables() {
    return fSetupData->ignorables;
}

NumberParseMatcher* AffixTokenMatcherWarehouse::nextCodePointMatcher(UChar32 cp, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    auto* result = fCodePoints.create(cp);
    if (result == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return result;
}


CodePointMatcher::CodePointMatcher(UChar32 cp)
        : fCp(cp) {}

bool CodePointMatcher::match(StringSegment& segment, ParsedNumber& result, UErrorCode&) const {
    if (segment.startsWith(fCp)) {
        segment.adjustOffsetByCodePoint();
        result.setCharsConsumed(segment);
    }
    return false;
}

bool CodePointMatcher::smokeTest(const StringSegment& segment) const {
    return segment.startsWith(fCp);
}

UnicodeString CodePointMatcher::toString() const {
    return u"<CodePoint>";
}


AffixPatternMatcher AffixPatternMatcher::fromAffixPattern(const UnicodeString& affixPattern,
                                                          AffixTokenMatcherWarehouse& tokenWarehouse,
                                                          parse_flags_t parseFlags, bool* success,
                                                          UErrorCode& status) {
    if (affixPattern.isEmpty()) {
        *success = false;
        return {};
    }
    *success = true;

    IgnorablesMatcher* ignorables;
    if (0 != (parseFlags & PARSE_FLAG_EXACT_AFFIX)) {
        ignorables = nullptr;
    } else {
        ignorables = &tokenWarehouse.ignorables();
    }

    AffixPatternMatcherBuilder builder(affixPattern, tokenWarehouse, ignorables);
    AffixUtils::iterateWithConsumer(affixPattern, builder, status);
    return builder.build();
}

AffixPatternMatcher::AffixPatternMatcher(MatcherArray& matchers, int32_t matchersLen,
                                         const UnicodeString& pattern)
        : ArraySeriesMatcher(matchers, matchersLen), fPattern(pattern) {}

UnicodeString AffixPatternMatcher::getPattern() const {
    return fPattern.toAliasedUnicodeString();
}

bool AffixPatternMatcher::operator==(const AffixPatternMatcher& other) const {
    return fPattern == other.fPattern;
}


AffixMatcherWarehouse::AffixMatcherWarehouse(AffixTokenMatcherWarehouse* tokenWarehouse)
        : fTokenWarehouse(tokenWarehouse) {
}

bool AffixMatcherWarehouse::isInteresting(const AffixPatternProvider& patternInfo,
                                          const IgnorablesMatcher& ignorables, parse_flags_t parseFlags,
                                          UErrorCode& status) {
    UnicodeString posPrefixString = patternInfo.getString(AffixPatternProvider::AFFIX_POS_PREFIX);
    UnicodeString posSuffixString = patternInfo.getString(AffixPatternProvider::AFFIX_POS_SUFFIX);
    UnicodeString negPrefixString;
    UnicodeString negSuffixString;
    if (patternInfo.hasNegativeSubpattern()) {
        negPrefixString = patternInfo.getString(AffixPatternProvider::AFFIX_NEG_PREFIX);
        negSuffixString = patternInfo.getString(AffixPatternProvider::AFFIX_NEG_SUFFIX);
    }

    if (0 == (parseFlags & PARSE_FLAG_USE_FULL_AFFIXES) &&
        AffixUtils::containsOnlySymbolsAndIgnorables(posPrefixString, *ignorables.getSet(), status) &&
        AffixUtils::containsOnlySymbolsAndIgnorables(posSuffixString, *ignorables.getSet(), status) &&
        AffixUtils::containsOnlySymbolsAndIgnorables(negPrefixString, *ignorables.getSet(), status) &&
        AffixUtils::containsOnlySymbolsAndIgnorables(negSuffixString, *ignorables.getSet(), status)
        // HACK: Plus and minus sign are a special case: we accept them trailing only if they are
        // trailing in the pattern string.
        && !AffixUtils::containsType(posSuffixString, TYPE_PLUS_SIGN, status) &&
        !AffixUtils::containsType(posSuffixString, TYPE_MINUS_SIGN, status) &&
        !AffixUtils::containsType(negSuffixString, TYPE_PLUS_SIGN, status) &&
        !AffixUtils::containsType(negSuffixString, TYPE_MINUS_SIGN, status)) {
        // The affixes contain only symbols and ignorables.
        // No need to generate affix matchers.
        return false;
    }
    return true;
}

void AffixMatcherWarehouse::createAffixMatchers(const AffixPatternProvider& patternInfo,
                                                MutableMatcherCollection& output,
                                                const IgnorablesMatcher& ignorables,
                                                parse_flags_t parseFlags, UErrorCode& status) {
    if (!isInteresting(patternInfo, ignorables, parseFlags, status)) {
        return;
    }

    // The affixes have interesting characters, or we are in strict mode.
    // Use initial capacity of 6, the highest possible number of AffixMatchers.
    UnicodeString sb;
    bool includeUnpaired = 0 != (parseFlags & PARSE_FLAG_INCLUDE_UNPAIRED_AFFIXES);
    UNumberSignDisplay signDisplay = (0 != (parseFlags & PARSE_FLAG_PLUS_SIGN_ALLOWED)) ? UNUM_SIGN_ALWAYS
                                                                                        : UNUM_SIGN_AUTO;

    int32_t numAffixMatchers = 0;
    int32_t numAffixPatternMatchers = 0;

    AffixPatternMatcher* posPrefix = nullptr;
    AffixPatternMatcher* posSuffix = nullptr;

    // Pre-process the affix strings to resolve LDML rules like sign display.
    for (int8_t signumInt = 1; signumInt >= -1; signumInt--) {
        auto signum = static_cast<Signum>(signumInt);

        // Generate Prefix
        bool hasPrefix = false;
        PatternStringUtils::patternInfoToStringBuilder(
                patternInfo, true, signum, signDisplay, StandardPlural::OTHER, false, sb);
        fAffixPatternMatchers[numAffixPatternMatchers] = AffixPatternMatcher::fromAffixPattern(
                sb, *fTokenWarehouse, parseFlags, &hasPrefix, status);
        AffixPatternMatcher* prefix = hasPrefix ? &fAffixPatternMatchers[numAffixPatternMatchers++]
                                                : nullptr;

        // Generate Suffix
        bool hasSuffix = false;
        PatternStringUtils::patternInfoToStringBuilder(
                patternInfo, false, signum, signDisplay, StandardPlural::OTHER, false, sb);
        fAffixPatternMatchers[numAffixPatternMatchers] = AffixPatternMatcher::fromAffixPattern(
                sb, *fTokenWarehouse, parseFlags, &hasSuffix, status);
        AffixPatternMatcher* suffix = hasSuffix ? &fAffixPatternMatchers[numAffixPatternMatchers++]
                                                : nullptr;

        if (signum == 1) {
            posPrefix = prefix;
            posSuffix = suffix;
        } else if (equals(prefix, posPrefix) && equals(suffix, posSuffix)) {
            // Skip adding these matchers (we already have equivalents)
            continue;
        }

        // Flags for setting in the ParsedNumber; the token matchers may add more.
        int flags = (signum == -1) ? FLAG_NEGATIVE : 0;

        // Note: it is indeed possible for posPrefix and posSuffix to both be null.
        // We still need to add that matcher for strict mode to work.
        fAffixMatchers[numAffixMatchers++] = {prefix, suffix, flags};
        if (includeUnpaired && prefix != nullptr && suffix != nullptr) {
            // The following if statements are designed to prevent adding two identical matchers.
            if (signum == 1 || !equals(prefix, posPrefix)) {
                fAffixMatchers[numAffixMatchers++] = {prefix, nullptr, flags};
            }
            if (signum == 1 || !equals(suffix, posSuffix)) {
                fAffixMatchers[numAffixMatchers++] = {nullptr, suffix, flags};
            }
        }
    }

    // Put the AffixMatchers in order, and then add them to the output.
    // Since there are at most 9 elements, do a simple-to-implement bubble sort.
    bool madeChanges;
    do {
        madeChanges = false;
        for (int32_t i = 1; i < numAffixMatchers; i++) {
            if (fAffixMatchers[i - 1].compareTo(fAffixMatchers[i]) > 0) {
                madeChanges = true;
                AffixMatcher temp = std::move(fAffixMatchers[i - 1]);
                fAffixMatchers[i - 1] = std::move(fAffixMatchers[i]);
                fAffixMatchers[i] = std::move(temp);
            }
        }
    } while (madeChanges);

    for (int32_t i = 0; i < numAffixMatchers; i++) {
        // Enable the following line to debug affixes
        //std::cout << "Adding affix matcher: " << CStr(fAffixMatchers[i].toString())() << std::endl;
        output.addMatcher(fAffixMatchers[i]);
    }
}


AffixMatcher::AffixMatcher(AffixPatternMatcher* prefix, AffixPatternMatcher* suffix, result_flags_t flags)
        : fPrefix(prefix), fSuffix(suffix), fFlags(flags) {}

bool AffixMatcher::match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const {
    if (!result.seenNumber()) {
        // Prefix
        // Do not match if:
        // 1. We have already seen a prefix (result.prefix != null)
        // 2. The prefix in this AffixMatcher is empty (prefix == null)
        if (!result.prefix.isBogus() || fPrefix == nullptr) {
            return false;
        }

        // Attempt to match the prefix.
        int initialOffset = segment.getOffset();
        bool maybeMore = fPrefix->match(segment, result, status);
        if (initialOffset != segment.getOffset()) {
            result.prefix = fPrefix->getPattern();
        }
        return maybeMore;

    } else {
        // Suffix
        // Do not match if:
        // 1. We have already seen a suffix (result.suffix != null)
        // 2. The suffix in this AffixMatcher is empty (suffix == null)
        // 3. The matched prefix does not equal this AffixMatcher's prefix
        if (!result.suffix.isBogus() || fSuffix == nullptr || !matched(fPrefix, result.prefix)) {
            return false;
        }

        // Attempt to match the suffix.
        int initialOffset = segment.getOffset();
        bool maybeMore = fSuffix->match(segment, result, status);
        if (initialOffset != segment.getOffset()) {
            result.suffix = fSuffix->getPattern();
        }
        return maybeMore;
    }
}

bool AffixMatcher::smokeTest(const StringSegment& segment) const {
    return (fPrefix != nullptr && fPrefix->smokeTest(segment)) ||
           (fSuffix != nullptr && fSuffix->smokeTest(segment));
}

void AffixMatcher::postProcess(ParsedNumber& result) const {
    // Check to see if our affix is the one that was matched. If so, set the flags in the result.
    if (matched(fPrefix, result.prefix) && matched(fSuffix, result.suffix)) {
        // Fill in the result prefix and suffix with non-null values (empty string).
        // Used by strict mode to determine whether an entire affix pair was matched.
        if (result.prefix.isBogus()) {
            result.prefix = UnicodeString();
        }
        if (result.suffix.isBogus()) {
            result.suffix = UnicodeString();
        }
        result.flags |= fFlags;
        if (fPrefix != nullptr) {
            fPrefix->postProcess(result);
        }
        if (fSuffix != nullptr) {
            fSuffix->postProcess(result);
        }
    }
}

int8_t AffixMatcher::compareTo(const AffixMatcher& rhs) const {
    const AffixMatcher& lhs = *this;
    if (length(lhs.fPrefix) != length(rhs.fPrefix)) {
        return length(lhs.fPrefix) > length(rhs.fPrefix) ? -1 : 1;
    } else if (length(lhs.fSuffix) != length(rhs.fSuffix)) {
        return length(lhs.fSuffix) > length(rhs.fSuffix) ? -1 : 1;
    } else {
        return 0;
    }
}

UnicodeString AffixMatcher::toString() const {
    bool isNegative = 0 != (fFlags & FLAG_NEGATIVE);
    return UnicodeString(u"<Affix") + (isNegative ? u":negative " : u" ") +
           (fPrefix ? fPrefix->getPattern() : u"null") + u"#" +
           (fSuffix ? fSuffix->getPattern() : u"null") + u">";

}


#endif /* #if !UCONFIG_NO_FORMATTING */

























// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif






using namespace icu;
using namespace icu::numparse;
using namespace icu::numparse::impl;


bool SeriesMatcher::match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const {
    ParsedNumber backup(result);

    int32_t initialOffset = segment.getOffset();
    bool maybeMore = true;
    for (auto* it = begin(); it < end();) {
        const NumberParseMatcher* matcher = *it;
        int matcherOffset = segment.getOffset();
        if (segment.length() != 0) {
            maybeMore = matcher->match(segment, result, status);
        } else {
            // Nothing for this matcher to match; ask for more.
            maybeMore = true;
        }

        bool success = (segment.getOffset() != matcherOffset);
        bool isFlexible = matcher->isFlexible();
        if (success && isFlexible) {
            // Match succeeded, and this is a flexible matcher. Re-run it.
        } else if (success) {
            // Match succeeded, and this is NOT a flexible matcher. Proceed to the next matcher.
            it++;
            // Small hack: if there is another matcher coming, do not accept trailing weak chars.
            // Needed for proper handling of currency spacing.
            if (it < end() && segment.getOffset() != result.charEnd && result.charEnd > matcherOffset) {
                segment.setOffset(result.charEnd);
            }
        } else if (isFlexible) {
            // Match failed, and this is a flexible matcher. Try again with the next matcher.
            it++;
        } else {
            // Match failed, and this is NOT a flexible matcher. Exit.
            segment.setOffset(initialOffset);
            result = backup;
            return maybeMore;
        }
    }

    // All matchers in the series succeeded.
    return maybeMore;
}

bool SeriesMatcher::smokeTest(const StringSegment& segment) const {
    // NOTE: The range-based for loop calls the virtual begin() and end() methods.
    // NOTE: We only want the first element. Use the for loop for boundary checking.
    for (auto& matcher : *this) {
        // SeriesMatchers are never allowed to start with a Flexible matcher.
        U_ASSERT(!matcher->isFlexible());
        return matcher->smokeTest(segment);
    }
    return false;
}

void SeriesMatcher::postProcess(ParsedNumber& result) const {
    // NOTE: The range-based for loop calls the virtual begin() and end() methods.
    for (auto* matcher : *this) {
        matcher->postProcess(result);
    }
}


ArraySeriesMatcher::ArraySeriesMatcher()
        : fMatchersLen(0) {
}

ArraySeriesMatcher::ArraySeriesMatcher(MatcherArray& matchers, int32_t matchersLen)
        : fMatchers(std::move(matchers)), fMatchersLen(matchersLen) {
}

int32_t ArraySeriesMatcher::length() const {
    return fMatchersLen;
}

const NumberParseMatcher* const* ArraySeriesMatcher::begin() const {
    return fMatchers.getAlias();
}

const NumberParseMatcher* const* ArraySeriesMatcher::end() const {
    return fMatchers.getAlias() + fMatchersLen;
}

UnicodeString ArraySeriesMatcher::toString() const {
    return u"<ArraySeries>";
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif








using namespace icu;
using namespace icu::numparse;
using namespace icu::numparse::impl;


CombinedCurrencyMatcher::CombinedCurrencyMatcher(const CurrencySymbols& currencySymbols, const DecimalFormatSymbols& dfs,
                                                 parse_flags_t parseFlags, UErrorCode& status)
        : fCurrency1(currencySymbols.getCurrencySymbol(status)),
          fCurrency2(currencySymbols.getIntlCurrencySymbol(status)),
          fUseFullCurrencyData(0 == (parseFlags & PARSE_FLAG_NO_FOREIGN_CURRENCY)),
          afterPrefixInsert(dfs.getPatternForCurrencySpacing(UNUM_CURRENCY_INSERT, false, status)),
          beforeSuffixInsert(dfs.getPatternForCurrencySpacing(UNUM_CURRENCY_INSERT, true, status)),
          fLocaleName(dfs.getLocale().getName(), -1, status) {
    utils::copyCurrencyCode(fCurrencyCode, currencySymbols.getIsoCode());

    // Pre-load the long names for the current locale and currency
    // if we are parsing without the full currency data.
    if (!fUseFullCurrencyData) {
        for (int32_t i=0; i<StandardPlural::COUNT; i++) {
            auto plural = static_cast<StandardPlural::Form>(i);
            fLocalLongNames[i] = currencySymbols.getPluralName(plural, status);
        }
    }

    // TODO: Figure out how to make this faster and re-enable.
    // Computing the "lead code points" set for fastpathing is too slow to use in production.
    // See http://bugs.icu-project.org/trac/ticket/13584
//    // Compute the full set of characters that could be the first in a currency to allow for
//    // efficient smoke test.
//    fLeadCodePoints.add(fCurrency1.char32At(0));
//    fLeadCodePoints.add(fCurrency2.char32At(0));
//    fLeadCodePoints.add(beforeSuffixInsert.char32At(0));
//    uprv_currencyLeads(fLocaleName.data(), fLeadCodePoints, status);
//    // Always apply case mapping closure for currencies
//    fLeadCodePoints.closeOver(USET_ADD_CASE_MAPPINGS);
//    fLeadCodePoints.freeze();
}

bool
CombinedCurrencyMatcher::match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const {
    if (result.currencyCode[0] != 0) {
        return false;
    }

    // Try to match a currency spacing separator.
    int32_t initialOffset = segment.getOffset();
    bool maybeMore = false;
    if (result.seenNumber() && !beforeSuffixInsert.isEmpty()) {
        int32_t overlap = segment.getCommonPrefixLength(beforeSuffixInsert);
        if (overlap == beforeSuffixInsert.length()) {
            segment.adjustOffset(overlap);
            // Note: let currency spacing be a weak match. Don't update chars consumed.
        }
        maybeMore = maybeMore || overlap == segment.length();
    }

    // Match the currency string, and reset if we didn't find one.
    maybeMore = maybeMore || matchCurrency(segment, result, status);
    if (result.currencyCode[0] == 0) {
        segment.setOffset(initialOffset);
        return maybeMore;
    }

    // Try to match a currency spacing separator.
    if (!result.seenNumber() && !afterPrefixInsert.isEmpty()) {
        int32_t overlap = segment.getCommonPrefixLength(afterPrefixInsert);
        if (overlap == afterPrefixInsert.length()) {
            segment.adjustOffset(overlap);
            // Note: let currency spacing be a weak match. Don't update chars consumed.
        }
        maybeMore = maybeMore || overlap == segment.length();
    }

    return maybeMore;
}

bool CombinedCurrencyMatcher::matchCurrency(StringSegment& segment, ParsedNumber& result,
                                            UErrorCode& status) const {
    bool maybeMore = false;

    int32_t overlap1;
    if (!fCurrency1.isEmpty()) {
        overlap1 = segment.getCaseSensitivePrefixLength(fCurrency1);
    } else {
        overlap1 = -1;
    }
    maybeMore = maybeMore || overlap1 == segment.length();
    if (overlap1 == fCurrency1.length()) {
        utils::copyCurrencyCode(result.currencyCode, fCurrencyCode);
        segment.adjustOffset(overlap1);
        result.setCharsConsumed(segment);
        return maybeMore;
    }

    int32_t overlap2;
    if (!fCurrency2.isEmpty()) {
        // ISO codes should be accepted case-insensitive.
        // https://unicode-org.atlassian.net/browse/ICU-13696
        overlap2 = segment.getCommonPrefixLength(fCurrency2);
    } else {
        overlap2 = -1;
    }
    maybeMore = maybeMore || overlap2 == segment.length();
    if (overlap2 == fCurrency2.length()) {
        utils::copyCurrencyCode(result.currencyCode, fCurrencyCode);
        segment.adjustOffset(overlap2);
        result.setCharsConsumed(segment);
        return maybeMore;
    }

    if (fUseFullCurrencyData) {
        // Use the full currency data.
        // NOTE: This call site should be improved with #13584.
        const UnicodeString segmentString = segment.toTempUnicodeString();

        // Try to parse the currency
        ParsePosition ppos(0);
        int32_t partialMatchLen = 0;
        uprv_parseCurrency(
                fLocaleName.data(),
                segmentString,
                ppos,
                UCURR_SYMBOL_NAME, // checks for both UCURR_SYMBOL_NAME and UCURR_LONG_NAME
                &partialMatchLen,
                result.currencyCode,
                status);
        maybeMore = maybeMore || partialMatchLen == segment.length();

        if (U_SUCCESS(status) && ppos.getIndex() != 0) {
            // Complete match.
            // NOTE: The currency code should already be saved in the ParsedNumber.
            segment.adjustOffset(ppos.getIndex());
            result.setCharsConsumed(segment);
            return maybeMore;
        }

    } else {
        // Use the locale long names.
        int32_t longestFullMatch = 0;
        for (int32_t i=0; i<StandardPlural::COUNT; i++) {
            const UnicodeString& name = fLocalLongNames[i];
            int32_t overlap = segment.getCommonPrefixLength(name);
            if (overlap == name.length() && name.length() > longestFullMatch) {
                longestFullMatch = name.length();
            }
            maybeMore = maybeMore || overlap > 0;
        }
        if (longestFullMatch > 0) {
            utils::copyCurrencyCode(result.currencyCode, fCurrencyCode);
            segment.adjustOffset(longestFullMatch);
            result.setCharsConsumed(segment);
            return maybeMore;
        }
    }

    // No match found.
    return maybeMore;
}

bool CombinedCurrencyMatcher::smokeTest(const StringSegment&) const {
    // TODO: See constructor
    return true;
    //return segment.startsWith(fLeadCodePoints);
}

UnicodeString CombinedCurrencyMatcher::toString() const {
    return u"<CombinedCurrencyMatcher>";
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif









using namespace icu;
using namespace icu::numparse;
using namespace icu::numparse::impl;


DecimalMatcher::DecimalMatcher(const DecimalFormatSymbols& symbols, const Grouper& grouper,
                               parse_flags_t parseFlags) {
    if (0 != (parseFlags & PARSE_FLAG_MONETARY_SEPARATORS)) {
        groupingSeparator = symbols.getConstSymbol(DecimalFormatSymbols::kMonetaryGroupingSeparatorSymbol);
        decimalSeparator = symbols.getConstSymbol(DecimalFormatSymbols::kMonetarySeparatorSymbol);
    } else {
        groupingSeparator = symbols.getConstSymbol(DecimalFormatSymbols::kGroupingSeparatorSymbol);
        decimalSeparator = symbols.getConstSymbol(DecimalFormatSymbols::kDecimalSeparatorSymbol);
    }
    bool strictSeparators = 0 != (parseFlags & PARSE_FLAG_STRICT_SEPARATORS);
    unisets::Key groupingKey = strictSeparators ? unisets::UNISET_KEY_STRICT_ALL_SEPARATORS
                                                : unisets::UNISET_KEY_ALL_SEPARATORS;

    // Attempt to find separators in the static cache

    groupingUniSet = unisets::get(groupingKey);
    unisets::Key decimalKey = unisets::chooseFrom(
            decimalSeparator,
            strictSeparators ? unisets::UNISET_KEY_STRICT_COMMA : unisets::UNISET_KEY_COMMA,
            strictSeparators ? unisets::UNISET_KEY_STRICT_PERIOD : unisets::UNISET_KEY_PERIOD);
    if (decimalKey >= 0) {
        decimalUniSet = unisets::get(decimalKey);
    } else if (!decimalSeparator.isEmpty()) {
        auto* set = new UnicodeSet();
        set->add(decimalSeparator.char32At(0));
        set->freeze();
        decimalUniSet = set;
        fLocalDecimalUniSet.adoptInstead(set);
    } else {
        decimalUniSet = unisets::get(unisets::UNISET_KEY_EMPTY);
    }

    if (groupingKey >= 0 && decimalKey >= 0) {
        // Everything is available in the static cache
        separatorSet = groupingUniSet;
        leadSet = unisets::get(
                strictSeparators ? unisets::UNISET_KEY_DIGITS_OR_ALL_SEPARATORS
                                 : unisets::UNISET_KEY_DIGITS_OR_STRICT_ALL_SEPARATORS);
    } else {
        auto* set = new UnicodeSet();
        set->addAll(*groupingUniSet);
        set->addAll(*decimalUniSet);
        set->freeze();
        separatorSet = set;
        fLocalSeparatorSet.adoptInstead(set);
        leadSet = nullptr;
    }

    UChar32 cpZero = symbols.getCodePointZero();
    if (cpZero == -1 || !u_isdigit(cpZero) || u_digit(cpZero, 10) != 0) {
        // Uncommon case: okay to allocate.
        auto digitStrings = new UnicodeString[10];
        fLocalDigitStrings.adoptInstead(digitStrings);
        for (int32_t i = 0; i <= 9; i++) {
            digitStrings[i] = symbols.getConstDigitSymbol(i);
        }
    }

    requireGroupingMatch = 0 != (parseFlags & PARSE_FLAG_STRICT_GROUPING_SIZE);
    groupingDisabled = 0 != (parseFlags & PARSE_FLAG_GROUPING_DISABLED);
    integerOnly = 0 != (parseFlags & PARSE_FLAG_INTEGER_ONLY);
    grouping1 = grouper.getPrimary();
    grouping2 = grouper.getSecondary();

    // Fraction grouping parsing is disabled for now but could be enabled later.
    // See http://bugs.icu-project.org/trac/ticket/10794
    // fractionGrouping = 0 != (parseFlags & PARSE_FLAG_FRACTION_GROUPING_ENABLED);
}

bool DecimalMatcher::match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const {
    return match(segment, result, 0, status);
}

bool DecimalMatcher::match(StringSegment& segment, ParsedNumber& result, int8_t exponentSign,
                           UErrorCode&) const {
    if (result.seenNumber() && exponentSign == 0) {
        // A number has already been consumed.
        return false;
    } else if (exponentSign != 0) {
        // scientific notation always comes after the number
        U_ASSERT(!result.quantity.bogus);
    }

    // Initial offset before any character consumption.
    int32_t initialOffset = segment.getOffset();

    // Return value: whether to ask for more characters.
    bool maybeMore = false;

    // All digits consumed so far.
    number::impl::DecimalQuantity digitsConsumed;
    digitsConsumed.bogus = true;

    // The total number of digits after the decimal place, used for scaling the result.
    int32_t digitsAfterDecimalPlace = 0;

    // The actual grouping and decimal separators used in the string.
    // If non-null, we have seen that token.
    UnicodeString actualGroupingString;
    UnicodeString actualDecimalString;
    actualGroupingString.setToBogus();
    actualDecimalString.setToBogus();

    // Information for two groups: the previous group and the current group.
    //
    // Each group has three pieces of information:
    //
    // Offset: the string position of the beginning of the group, including a leading separator
    // if there was a leading separator. This is needed in case we need to rewind the parse to
    // that position.
    //
    // Separator type:
    // 0 => beginning of string
    // 1 => lead separator is a grouping separator
    // 2 => lead separator is a decimal separator
    //
    // Count: the number of digits in the group. If -1, the group has been validated.
    int32_t currGroupOffset = 0;
    int32_t currGroupSepType = 0;
    int32_t currGroupCount = 0;
    int32_t prevGroupOffset = -1;
    int32_t prevGroupSepType = -1;
    int32_t prevGroupCount = -1;

    while (segment.length() > 0) {
        maybeMore = false;

        // Attempt to match a digit.
        int8_t digit = -1;

        // Try by code point digit value.
        UChar32 cp = segment.getCodePoint();
        if (u_isdigit(cp)) {
            segment.adjustOffset(U16_LENGTH(cp));
            digit = static_cast<int8_t>(u_digit(cp, 10));
        }

        // Try by digit string.
        if (digit == -1 && !fLocalDigitStrings.isNull()) {
            for (int32_t i = 0; i < 10; i++) {
                const UnicodeString& str = fLocalDigitStrings[i];
                if (str.isEmpty()) {
                    continue;
                }
                int32_t overlap = segment.getCommonPrefixLength(str);
                if (overlap == str.length()) {
                    segment.adjustOffset(overlap);
                    digit = static_cast<int8_t>(i);
                    break;
                }
                maybeMore = maybeMore || (overlap == segment.length());
            }
        }

        if (digit >= 0) {
            // Digit was found.
            if (digitsConsumed.bogus) {
                digitsConsumed.bogus = false;
                digitsConsumed.clear();
            }
            digitsConsumed.appendDigit(digit, 0, true);
            currGroupCount++;
            if (!actualDecimalString.isBogus()) {
                digitsAfterDecimalPlace++;
            }
            continue;
        }

        // Attempt to match a literal grouping or decimal separator.
        bool isDecimal = false;
        bool isGrouping = false;

        // 1) Attempt the decimal separator string literal.
        // if (we have not seen a decimal separator yet) { ... }
        if (actualDecimalString.isBogus() && !decimalSeparator.isEmpty()) {
            int32_t overlap = segment.getCommonPrefixLength(decimalSeparator);
            maybeMore = maybeMore || (overlap == segment.length());
            if (overlap == decimalSeparator.length()) {
                isDecimal = true;
                actualDecimalString = decimalSeparator;
            }
        }

        // 2) Attempt to match the actual grouping string literal.
        if (!actualGroupingString.isBogus()) {
            int32_t overlap = segment.getCommonPrefixLength(actualGroupingString);
            maybeMore = maybeMore || (overlap == segment.length());
            if (overlap == actualGroupingString.length()) {
                isGrouping = true;
            }
        }

        // 2.5) Attempt to match a new the grouping separator string literal.
        // if (we have not seen a grouping or decimal separator yet) { ... }
        if (!groupingDisabled && actualGroupingString.isBogus() && actualDecimalString.isBogus() &&
            !groupingSeparator.isEmpty()) {
            int32_t overlap = segment.getCommonPrefixLength(groupingSeparator);
            maybeMore = maybeMore || (overlap == segment.length());
            if (overlap == groupingSeparator.length()) {
                isGrouping = true;
                actualGroupingString = groupingSeparator;
            }
        }

        // 3) Attempt to match a decimal separator from the equivalence set.
        // if (we have not seen a decimal separator yet) { ... }
        // The !isGrouping is to confirm that we haven't yet matched the current character.
        if (!isGrouping && actualDecimalString.isBogus()) {
            if (decimalUniSet->contains(cp)) {
                isDecimal = true;
                actualDecimalString = UnicodeString(cp);
            }
        }

        // 4) Attempt to match a grouping separator from the equivalence set.
        // if (we have not seen a grouping or decimal separator yet) { ... }
        if (!groupingDisabled && actualGroupingString.isBogus() && actualDecimalString.isBogus()) {
            if (groupingUniSet->contains(cp)) {
                isGrouping = true;
                actualGroupingString = UnicodeString(cp);
            }
        }

        // Leave if we failed to match this as a separator.
        if (!isDecimal && !isGrouping) {
            break;
        }

        // Check for conditions when we don't want to accept the separator.
        if (isDecimal && integerOnly) {
            break;
        } else if (currGroupSepType == 2 && isGrouping) {
            // Fraction grouping
            break;
        }

        // Validate intermediate grouping sizes.
        bool prevValidSecondary = validateGroup(prevGroupSepType, prevGroupCount, false);
        bool currValidPrimary = validateGroup(currGroupSepType, currGroupCount, true);
        if (!prevValidSecondary || (isDecimal && !currValidPrimary)) {
            // Invalid grouping sizes.
            if (isGrouping && currGroupCount == 0) {
                // Trailing grouping separators: these are taken care of below
                U_ASSERT(currGroupSepType == 1);
            } else if (requireGroupingMatch) {
                // Strict mode: reject the parse
                digitsConsumed.clear();
                digitsConsumed.bogus = true;
            }
            break;
        } else if (requireGroupingMatch && currGroupCount == 0 && currGroupSepType == 1) {
            break;
        } else {
            // Grouping sizes OK so far.
            prevGroupOffset = currGroupOffset;
            prevGroupCount = currGroupCount;
            if (isDecimal) {
                // Do not validate this group any more.
                prevGroupSepType = -1;
            } else {
                prevGroupSepType = currGroupSepType;
            }
        }

        // OK to accept the separator.
        // Special case: don't update currGroup if it is empty; this allows two grouping
        // separators in a row in lenient mode.
        if (currGroupCount != 0) {
            currGroupOffset = segment.getOffset();
        }
        currGroupSepType = isGrouping ? 1 : 2;
        currGroupCount = 0;
        if (isGrouping) {
            segment.adjustOffset(actualGroupingString.length());
        } else {
            segment.adjustOffset(actualDecimalString.length());
        }
    }

    // End of main loop.
    // Back up if there was a trailing grouping separator.
    // Shift prev -> curr so we can check it as a final group.
    if (currGroupSepType != 2 && currGroupCount == 0) {
        maybeMore = true;
        segment.setOffset(currGroupOffset);
        currGroupOffset = prevGroupOffset;
        currGroupSepType = prevGroupSepType;
        currGroupCount = prevGroupCount;
        prevGroupOffset = -1;
        prevGroupSepType = 0;
        prevGroupCount = 1;
    }

    // Validate final grouping sizes.
    bool prevValidSecondary = validateGroup(prevGroupSepType, prevGroupCount, false);
    bool currValidPrimary = validateGroup(currGroupSepType, currGroupCount, true);
    if (!requireGroupingMatch) {
        // The cases we need to handle here are lone digits.
        // Examples: "1,1"  "1,1,"  "1,1,1"  "1,1,1,"  ",1" (all parse as 1)
        // See more examples in numberformattestspecification.txt
        int32_t digitsToRemove = 0;
        if (!prevValidSecondary) {
            segment.setOffset(prevGroupOffset);
            digitsToRemove += prevGroupCount;
            digitsToRemove += currGroupCount;
        } else if (!currValidPrimary && (prevGroupSepType != 0 || prevGroupCount != 0)) {
            maybeMore = true;
            segment.setOffset(currGroupOffset);
            digitsToRemove += currGroupCount;
        }
        if (digitsToRemove != 0) {
            digitsConsumed.adjustMagnitude(-digitsToRemove);
            digitsConsumed.truncate();
        }
        prevValidSecondary = true;
        currValidPrimary = true;
    }
    if (currGroupSepType != 2 && (!prevValidSecondary || !currValidPrimary)) {
        // Grouping failure.
        digitsConsumed.bogus = true;
    }

    // Strings that start with a separator but have no digits,
    // or strings that failed a grouping size check.
    if (digitsConsumed.bogus) {
        maybeMore = maybeMore || (segment.length() == 0);
        segment.setOffset(initialOffset);
        return maybeMore;
    }

    // We passed all inspections. Start post-processing.

    // Adjust for fraction part.
    digitsConsumed.adjustMagnitude(-digitsAfterDecimalPlace);

    // Set the digits, either normal or exponent.
    if (exponentSign != 0 && segment.getOffset() != initialOffset) {
        bool overflow = false;
        if (digitsConsumed.fitsInLong()) {
            int64_t exponentLong = digitsConsumed.toLong(false);
            U_ASSERT(exponentLong >= 0);
            if (exponentLong <= INT32_MAX) {
                auto exponentInt = static_cast<int32_t>(exponentLong);
                if (result.quantity.adjustMagnitude(exponentSign * exponentInt)) {
                    overflow = true;
                }
            } else {
                overflow = true;
            }
        } else {
            overflow = true;
        }
        if (overflow) {
            if (exponentSign == -1) {
                // Set to zero
                result.quantity.clear();
            } else {
                // Set to infinity
                result.quantity.bogus = true;
                result.flags |= FLAG_INFINITY;
            }
        }
    } else {
        result.quantity = digitsConsumed;
    }

    // Set other information into the result and return.
    if (!actualDecimalString.isBogus()) {
        result.flags |= FLAG_HAS_DECIMAL_SEPARATOR;
    }
    result.setCharsConsumed(segment);
    return segment.length() == 0 || maybeMore;
}

bool DecimalMatcher::validateGroup(int32_t sepType, int32_t count, bool isPrimary) const {
    if (requireGroupingMatch) {
        if (sepType == -1) {
            // No such group (prevGroup before first shift).
            return true;
        } else if (sepType == 0) {
            // First group.
            if (isPrimary) {
                // No grouping separators is OK.
                return true;
            } else {
                return count != 0 && count <= grouping2;
            }
        } else if (sepType == 1) {
            // Middle group.
            if (isPrimary) {
                return count == grouping1;
            } else {
                return count == grouping2;
            }
        } else {
            U_ASSERT(sepType == 2);
            // After the decimal separator.
            return true;
        }
    } else {
        if (sepType == 1) {
            // #11230: don't accept middle groups with only 1 digit.
            return count != 1;
        } else {
            return true;
        }
    }
}

bool DecimalMatcher::smokeTest(const StringSegment& segment) const {
    // The common case uses a static leadSet for efficiency.
    if (fLocalDigitStrings.isNull() && leadSet != nullptr) {
        return segment.startsWith(*leadSet);
    }
    if (segment.startsWith(*separatorSet) || u_isdigit(segment.getCodePoint())) {
        return true;
    }
    if (fLocalDigitStrings.isNull()) {
        return false;
    }
    for (int32_t i = 0; i < 10; i++) {
        if (segment.startsWith(fLocalDigitStrings[i])) {
            return true;
        }
    }
    return false;
}

UnicodeString DecimalMatcher::toString() const {
    return u"<Decimal>";
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif

#include <typeinfo>
#include <array>











using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;
using namespace icu::numparse;
using namespace icu::numparse::impl;


NumberParseMatcher::~NumberParseMatcher() = default;


NumberParserImpl*
NumberParserImpl::createSimpleParser(const Locale& locale, const UnicodeString& patternString,
                                     parse_flags_t parseFlags, UErrorCode& status) {

    LocalPointer<NumberParserImpl> parser(new NumberParserImpl(parseFlags));
    DecimalFormatSymbols symbols(locale, status);

    parser->fLocalMatchers.ignorables = {parseFlags};
    IgnorablesMatcher& ignorables = parser->fLocalMatchers.ignorables;

    DecimalFormatSymbols dfs(locale, status);
    dfs.setSymbol(DecimalFormatSymbols::kCurrencySymbol, u"IU$");
    dfs.setSymbol(DecimalFormatSymbols::kIntlCurrencySymbol, u"ICU");
    CurrencySymbols currencySymbols({u"ICU", status}, locale, dfs, status);

    ParsedPatternInfo patternInfo;
    PatternParser::parseToPatternInfo(patternString, patternInfo, status);

    // The following statements set up the affix matchers.
    AffixTokenMatcherSetupData affixSetupData = {
            currencySymbols, symbols, ignorables, locale, parseFlags};
    parser->fLocalMatchers.affixTokenMatcherWarehouse = {&affixSetupData};
    parser->fLocalMatchers.affixMatcherWarehouse = {&parser->fLocalMatchers.affixTokenMatcherWarehouse};
    parser->fLocalMatchers.affixMatcherWarehouse.createAffixMatchers(
            patternInfo, *parser, ignorables, parseFlags, status);

    Grouper grouper = Grouper::forStrategy(UNUM_GROUPING_AUTO);
    grouper.setLocaleData(patternInfo, locale);

    parser->addMatcher(parser->fLocalMatchers.ignorables);
    parser->addMatcher(parser->fLocalMatchers.decimal = {symbols, grouper, parseFlags});
    parser->addMatcher(parser->fLocalMatchers.minusSign = {symbols, false});
    parser->addMatcher(parser->fLocalMatchers.plusSign = {symbols, false});
    parser->addMatcher(parser->fLocalMatchers.percent = {symbols});
    parser->addMatcher(parser->fLocalMatchers.permille = {symbols});
    parser->addMatcher(parser->fLocalMatchers.nan = {symbols});
    parser->addMatcher(parser->fLocalMatchers.infinity = {symbols});
    parser->addMatcher(parser->fLocalMatchers.padding = {u"@"});
    parser->addMatcher(parser->fLocalMatchers.scientific = {symbols, grouper});
    parser->addMatcher(parser->fLocalMatchers.currency = {currencySymbols, symbols, parseFlags, status});
    parser->addMatcher(parser->fLocalValidators.number = {});

    parser->freeze();
    return parser.orphan();
}

NumberParserImpl*
NumberParserImpl::createParserFromProperties(const number::impl::DecimalFormatProperties& properties,
                                             const DecimalFormatSymbols& symbols, bool parseCurrency,
                                             UErrorCode& status) {
    Locale locale = symbols.getLocale();
    PropertiesAffixPatternProvider localPAPP;
    CurrencyPluralInfoAffixProvider localCPIAP;
    AffixPatternProvider* affixProvider;
    if (properties.currencyPluralInfo.fPtr.isNull()) {
        localPAPP.setTo(properties, status);
        affixProvider = &localPAPP;
    } else {
        localCPIAP.setTo(*properties.currencyPluralInfo.fPtr, properties, status);
        affixProvider = &localCPIAP;
    }
    if (affixProvider == nullptr || U_FAILURE(status)) { return nullptr; }
    CurrencyUnit currency = resolveCurrency(properties, locale, status);
    CurrencySymbols currencySymbols(currency, locale, symbols, status);
    bool isStrict = properties.parseMode.getOrDefault(PARSE_MODE_STRICT) == PARSE_MODE_STRICT;
    Grouper grouper = Grouper::forProperties(properties);
    int parseFlags = 0;
    if (affixProvider == nullptr || U_FAILURE(status)) { return nullptr; }
    if (!properties.parseCaseSensitive) {
        parseFlags |= PARSE_FLAG_IGNORE_CASE;
    }
    if (properties.parseIntegerOnly) {
        parseFlags |= PARSE_FLAG_INTEGER_ONLY;
    }
    if (properties.signAlwaysShown) {
        parseFlags |= PARSE_FLAG_PLUS_SIGN_ALLOWED;
    }
    if (isStrict) {
        parseFlags |= PARSE_FLAG_STRICT_GROUPING_SIZE;
        parseFlags |= PARSE_FLAG_STRICT_SEPARATORS;
        parseFlags |= PARSE_FLAG_USE_FULL_AFFIXES;
        parseFlags |= PARSE_FLAG_EXACT_AFFIX;
        parseFlags |= PARSE_FLAG_STRICT_IGNORABLES;
    } else {
        parseFlags |= PARSE_FLAG_INCLUDE_UNPAIRED_AFFIXES;
    }
    if (grouper.getPrimary() <= 0) {
        parseFlags |= PARSE_FLAG_GROUPING_DISABLED;
    }
    if (parseCurrency || affixProvider->hasCurrencySign()) {
        parseFlags |= PARSE_FLAG_MONETARY_SEPARATORS;
    }
    if (!parseCurrency) {
        parseFlags |= PARSE_FLAG_NO_FOREIGN_CURRENCY;
    }

    LocalPointer<NumberParserImpl> parser(new NumberParserImpl(parseFlags));

    parser->fLocalMatchers.ignorables = {parseFlags};
    IgnorablesMatcher& ignorables = parser->fLocalMatchers.ignorables;

    //////////////////////
    /// AFFIX MATCHERS ///
    //////////////////////

    // The following statements set up the affix matchers.
    AffixTokenMatcherSetupData affixSetupData = {
            currencySymbols, symbols, ignorables, locale, parseFlags};
    parser->fLocalMatchers.affixTokenMatcherWarehouse = {&affixSetupData};
    parser->fLocalMatchers.affixMatcherWarehouse = {&parser->fLocalMatchers.affixTokenMatcherWarehouse};
    parser->fLocalMatchers.affixMatcherWarehouse.createAffixMatchers(
            *affixProvider, *parser, ignorables, parseFlags, status);

    ////////////////////////
    /// CURRENCY MATCHER ///
    ////////////////////////

    if (parseCurrency || affixProvider->hasCurrencySign()) {
        parser->addMatcher(parser->fLocalMatchers.currency = {currencySymbols, symbols, parseFlags, status});
    }

    ///////////////
    /// PERCENT ///
    ///////////////

    // ICU-TC meeting, April 11, 2018: accept percent/permille only if it is in the pattern,
    // and to maintain regressive behavior, divide by 100 even if no percent sign is present.
    if (!isStrict && affixProvider->containsSymbolType(AffixPatternType::TYPE_PERCENT, status)) {
        parser->addMatcher(parser->fLocalMatchers.percent = {symbols});
    }
    if (!isStrict && affixProvider->containsSymbolType(AffixPatternType::TYPE_PERMILLE, status)) {
        parser->addMatcher(parser->fLocalMatchers.permille = {symbols});
    }

    ///////////////////////////////
    /// OTHER STANDARD MATCHERS ///
    ///////////////////////////////

    if (!isStrict) {
        parser->addMatcher(parser->fLocalMatchers.plusSign = {symbols, false});
        parser->addMatcher(parser->fLocalMatchers.minusSign = {symbols, false});
    }
    parser->addMatcher(parser->fLocalMatchers.nan = {symbols});
    parser->addMatcher(parser->fLocalMatchers.infinity = {symbols});
    UnicodeString padString = properties.padString;
    if (!padString.isBogus() && !ignorables.getSet()->contains(padString)) {
        parser->addMatcher(parser->fLocalMatchers.padding = {padString});
    }
    parser->addMatcher(parser->fLocalMatchers.ignorables);
    parser->addMatcher(parser->fLocalMatchers.decimal = {symbols, grouper, parseFlags});
    // NOTE: parseNoExponent doesn't disable scientific parsing if we have a scientific formatter
    if (!properties.parseNoExponent || properties.minimumExponentDigits > 0) {
        parser->addMatcher(parser->fLocalMatchers.scientific = {symbols, grouper});
    }

    //////////////////
    /// VALIDATORS ///
    //////////////////

    parser->addMatcher(parser->fLocalValidators.number = {});
    if (isStrict) {
        parser->addMatcher(parser->fLocalValidators.affix = {});
    }
    if (parseCurrency) {
        parser->addMatcher(parser->fLocalValidators.currency = {});
    }
    if (properties.decimalPatternMatchRequired) {
        bool patternHasDecimalSeparator =
                properties.decimalSeparatorAlwaysShown || properties.maximumFractionDigits != 0;
        parser->addMatcher(parser->fLocalValidators.decimalSeparator = {patternHasDecimalSeparator});
    }
    // The multiplier takes care of scaling percentages.
    Scale multiplier = scaleFromProperties(properties);
    if (multiplier.isValid()) {
        parser->addMatcher(parser->fLocalValidators.multiplier = {multiplier});
    }

    parser->freeze();
    return parser.orphan();
}

NumberParserImpl::NumberParserImpl(parse_flags_t parseFlags)
        : fParseFlags(parseFlags) {
}

NumberParserImpl::~NumberParserImpl() {
    fNumMatchers = 0;
}

void NumberParserImpl::addMatcher(NumberParseMatcher& matcher) {
    if (fNumMatchers + 1 > fMatchers.getCapacity()) {
        fMatchers.resize(fNumMatchers * 2, fNumMatchers);
    }
    fMatchers[fNumMatchers] = &matcher;
    fNumMatchers++;
}

void NumberParserImpl::freeze() {
    fFrozen = true;
}

parse_flags_t NumberParserImpl::getParseFlags() const {
    return fParseFlags;
}

void NumberParserImpl::parse(const UnicodeString& input, bool greedy, ParsedNumber& result,
                             UErrorCode& status) const {
    return parse(input, 0, greedy, result, status);
}

void NumberParserImpl::parse(const UnicodeString& input, int32_t start, bool greedy, ParsedNumber& result,
                             UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    U_ASSERT(fFrozen);
    // TODO: Check start >= 0 and start < input.length()
    StringSegment segment(input, 0 != (fParseFlags & PARSE_FLAG_IGNORE_CASE));
    segment.adjustOffset(start);
    if (greedy) {
        parseGreedy(segment, result, status);
    } else if (0 != (fParseFlags & PARSE_FLAG_ALLOW_INFINITE_RECURSION)) {
        // Start at 1 so that recursionLevels never gets to 0
        parseLongestRecursive(segment, result, 1, status);
    } else {
        // Arbitrary recursion safety limit: 100 levels.
        parseLongestRecursive(segment, result, -100, status);
    }
    for (int32_t i = 0; i < fNumMatchers; i++) {
        fMatchers[i]->postProcess(result);
    }
    result.postProcess();
}

void NumberParserImpl::parseGreedy(StringSegment& segment, ParsedNumber& result,
                                            UErrorCode& status) const {
    // Note: this method is not recursive in order to avoid stack overflow.
    for (int i = 0; i <fNumMatchers;) {
        // Base Case
        if (segment.length() == 0) {
            return;
        }
        const NumberParseMatcher* matcher = fMatchers[i];
        if (!matcher->smokeTest(segment)) {
            // Matcher failed smoke test: try the next one
            i++;
            continue;
        }
        int32_t initialOffset = segment.getOffset();
        matcher->match(segment, result, status);
        if (U_FAILURE(status)) {
            return;
        }
        if (segment.getOffset() != initialOffset) {
            // Greedy heuristic: accept the match and loop back
            i = 0;
            continue;
        } else {
            // Matcher did not match: try the next one
            i++;
            continue;
        }
        UPRV_UNREACHABLE;
    }

    // NOTE: If we get here, the greedy parse completed without consuming the entire string.
}

void NumberParserImpl::parseLongestRecursive(StringSegment& segment, ParsedNumber& result,
                                             int32_t recursionLevels,
                                             UErrorCode& status) const {
    // Base Case
    if (segment.length() == 0) {
        return;
    }

    // Safety against stack overflow
    if (recursionLevels == 0) {
        return;
    }

    // TODO: Give a nice way for the matcher to reset the ParsedNumber?
    ParsedNumber initial(result);
    ParsedNumber candidate;

    int initialOffset = segment.getOffset();
    for (int32_t i = 0; i < fNumMatchers; i++) {
        const NumberParseMatcher* matcher = fMatchers[i];
        if (!matcher->smokeTest(segment)) {
            continue;
        }

        // In a non-greedy parse, we attempt all possible matches and pick the best.
        for (int32_t charsToConsume = 0; charsToConsume < segment.length();) {
            charsToConsume += U16_LENGTH(segment.codePointAt(charsToConsume));

            // Run the matcher on a segment of the current length.
            candidate = initial;
            segment.setLength(charsToConsume);
            bool maybeMore = matcher->match(segment, candidate, status);
            segment.resetLength();
            if (U_FAILURE(status)) {
                return;
            }

            // If the entire segment was consumed, recurse.
            if (segment.getOffset() - initialOffset == charsToConsume) {
                parseLongestRecursive(segment, candidate, recursionLevels + 1, status);
                if (U_FAILURE(status)) {
                    return;
                }
                if (candidate.isBetterThan(result)) {
                    result = candidate;
                }
            }

            // Since the segment can be re-used, reset the offset.
            // This does not have an effect if the matcher did not consume any chars.
            segment.setOffset(initialOffset);

            // Unless the matcher wants to see the next char, continue to the next matcher.
            if (!maybeMore) {
                break;
            }
        }
    }
}

UnicodeString NumberParserImpl::toString() const {
    UnicodeString result(u"<NumberParserImpl matchers:[");
    for (int32_t i = 0; i < fNumMatchers; i++) {
        result.append(u' ');
        result.append(fMatchers[i]->toString());
    }
    result.append(u" ]>", -1);
    return result;
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif





#include <cmath>

using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;
using namespace icu::numparse;
using namespace icu::numparse::impl;


ParsedNumber::ParsedNumber() {
    clear();
}

void ParsedNumber::clear() {
    quantity.bogus = true;
    charEnd = 0;
    flags = 0;
    prefix.setToBogus();
    suffix.setToBogus();
    currencyCode[0] = 0;
}

void ParsedNumber::setCharsConsumed(const StringSegment& segment) {
    charEnd = segment.getOffset();
}

void ParsedNumber::postProcess() {
    if (!quantity.bogus && 0 != (flags & FLAG_NEGATIVE)) {
        quantity.negate();
    }
}

bool ParsedNumber::success() const {
    return charEnd > 0 && 0 == (flags & FLAG_FAIL);
}

bool ParsedNumber::seenNumber() const {
    return !quantity.bogus || 0 != (flags & FLAG_NAN) || 0 != (flags & FLAG_INFINITY);
}

double ParsedNumber::getDouble(UErrorCode& status) const {
    bool sawNaN = 0 != (flags & FLAG_NAN);
    bool sawInfinity = 0 != (flags & FLAG_INFINITY);

    // Check for NaN, infinity, and -0.0
    if (sawNaN) {
        // Can't use NAN or std::nan because the byte pattern is platform-dependent;
        // MSVC sets the sign bit, but Clang and GCC do not
        return uprv_getNaN();
    }
    if (sawInfinity) {
        if (0 != (flags & FLAG_NEGATIVE)) {
            return -INFINITY;
        } else {
            return INFINITY;
        }
    }
    if (quantity.bogus) {
        status = U_INVALID_STATE_ERROR;
        return 0.0;
    }
    if (quantity.isZeroish() && quantity.isNegative()) {
        return -0.0;
    }

    if (quantity.fitsInLong()) {
        return static_cast<double>(quantity.toLong());
    } else {
        return quantity.toDouble();
    }
}

void ParsedNumber::populateFormattable(Formattable& output, parse_flags_t parseFlags) const {
    bool sawNaN = 0 != (flags & FLAG_NAN);
    bool sawInfinity = 0 != (flags & FLAG_INFINITY);
    bool integerOnly = 0 != (parseFlags & PARSE_FLAG_INTEGER_ONLY);

    // Check for NaN, infinity, and -0.0
    if (sawNaN) {
        // Can't use NAN or std::nan because the byte pattern is platform-dependent;
        // MSVC sets the sign bit, but Clang and GCC do not
        output.setDouble(uprv_getNaN());
        return;
    }
    if (sawInfinity) {
        if (0 != (flags & FLAG_NEGATIVE)) {
            output.setDouble(-INFINITY);
            return;
        } else {
            output.setDouble(INFINITY);
            return;
        }
    }
    U_ASSERT(!quantity.bogus);
    if (quantity.isZeroish() && quantity.isNegative() && !integerOnly) {
        output.setDouble(-0.0);
        return;
    }

    // All other numbers
    output.adoptDecimalQuantity(new DecimalQuantity(quantity));
}

bool ParsedNumber::isBetterThan(const ParsedNumber& other) {
    // Favor results with strictly more characters consumed.
    return charEnd > other.charEnd;
}



#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif






using namespace icu;
using namespace icu::numparse;
using namespace icu::numparse::impl;


namespace {

inline const UnicodeSet& minusSignSet() {
    return *unisets::get(unisets::UNISET_KEY_MINUS_SIGN);
}

inline const UnicodeSet& plusSignSet() {
    return *unisets::get(unisets::UNISET_KEY_PLUS_SIGN);
}

} // namespace


ScientificMatcher::ScientificMatcher(const DecimalFormatSymbols& dfs, const Grouper& grouper)
        : fExponentSeparatorString(dfs.getConstSymbol(DecimalFormatSymbols::kExponentialSymbol)),
          fExponentMatcher(dfs, grouper, PARSE_FLAG_INTEGER_ONLY | PARSE_FLAG_GROUPING_DISABLED),
          fIgnorablesMatcher(PARSE_FLAG_STRICT_IGNORABLES) {

    const UnicodeString& minusSign = dfs.getConstSymbol(DecimalFormatSymbols::kMinusSignSymbol);
    if (minusSignSet().contains(minusSign)) {
        fCustomMinusSign.setToBogus();
    } else {
        fCustomMinusSign = minusSign;
    }

    const UnicodeString& plusSign = dfs.getConstSymbol(DecimalFormatSymbols::kPlusSignSymbol);
    if (plusSignSet().contains(plusSign)) {
        fCustomPlusSign.setToBogus();
    } else {
        fCustomPlusSign = plusSign;
    }
}

bool ScientificMatcher::match(StringSegment& segment, ParsedNumber& result, UErrorCode& status) const {
    // Only accept scientific notation after the mantissa.
    if (!result.seenNumber()) {
        return false;
    }

    // Only accept one exponent per string.
    if (0 != (result.flags & FLAG_HAS_EXPONENT)) {
        return false;
    }

    // First match the scientific separator, and then match another number after it.
    // NOTE: This is guarded by the smoke test; no need to check fExponentSeparatorString length again.
    int32_t initialOffset = segment.getOffset();
    int32_t overlap = segment.getCommonPrefixLength(fExponentSeparatorString);
    if (overlap == fExponentSeparatorString.length()) {
        // Full exponent separator match.

        // First attempt to get a code point, returning true if we can't get one.
        if (segment.length() == overlap) {
            return true;
        }
        segment.adjustOffset(overlap);

        // Allow ignorables before the sign.
        // Note: call site is guarded by the segment.length() check above.
        // Note: the ignorables matcher should not touch the result.
        fIgnorablesMatcher.match(segment, result, status);
        if (segment.length() == 0) {
            segment.setOffset(initialOffset);
            return true;
        }

        // Allow a sign, and then try to match digits.
        int8_t exponentSign = 1;
        if (segment.startsWith(minusSignSet())) {
            exponentSign = -1;
            segment.adjustOffsetByCodePoint();
        } else if (segment.startsWith(plusSignSet())) {
            segment.adjustOffsetByCodePoint();
        } else if (segment.startsWith(fCustomMinusSign)) {
            overlap = segment.getCommonPrefixLength(fCustomMinusSign);
            if (overlap != fCustomMinusSign.length()) {
                // Partial custom sign match
                segment.setOffset(initialOffset);
                return true;
            }
            exponentSign = -1;
            segment.adjustOffset(overlap);
        } else if (segment.startsWith(fCustomPlusSign)) {
            overlap = segment.getCommonPrefixLength(fCustomPlusSign);
            if (overlap != fCustomPlusSign.length()) {
                // Partial custom sign match
                segment.setOffset(initialOffset);
                return true;
            }
            segment.adjustOffset(overlap);
        }

        // Return true if the segment is empty.
        if (segment.length() == 0) {
            segment.setOffset(initialOffset);
            return true;
        }

        // Allow ignorables after the sign.
        // Note: call site is guarded by the segment.length() check above.
        // Note: the ignorables matcher should not touch the result.
        fIgnorablesMatcher.match(segment, result, status);
        if (segment.length() == 0) {
            segment.setOffset(initialOffset);
            return true;
        }

        // We are supposed to accept E0 after NaN, so we need to make sure result.quantity is available.
        bool wasBogus = result.quantity.bogus;
        result.quantity.bogus = false;
        int digitsOffset = segment.getOffset();
        bool digitsReturnValue = fExponentMatcher.match(segment, result, exponentSign, status);
        result.quantity.bogus = wasBogus;

        if (segment.getOffset() != digitsOffset) {
            // At least one exponent digit was matched.
            result.flags |= FLAG_HAS_EXPONENT;
        } else {
            // No exponent digits were matched
            segment.setOffset(initialOffset);
        }
        return digitsReturnValue;

    } else if (overlap == segment.length()) {
        // Partial exponent separator match
        return true;
    }

    // No match
    return false;
}

bool ScientificMatcher::smokeTest(const StringSegment& segment) const {
    return segment.startsWith(fExponentSeparatorString);
}

UnicodeString ScientificMatcher::toString() const {
    return u"<Scientific>";
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif






using namespace icu;
using namespace icu::numparse;
using namespace icu::numparse::impl;


SymbolMatcher::SymbolMatcher(const UnicodeString& symbolString, unisets::Key key) {
    fUniSet = unisets::get(key);
    if (fUniSet->contains(symbolString)) {
        fString.setToBogus();
    } else {
        fString = symbolString;
    }
}

const UnicodeSet* SymbolMatcher::getSet() const {
    return fUniSet;
}

bool SymbolMatcher::match(StringSegment& segment, ParsedNumber& result, UErrorCode&) const {
    // Smoke test first; this matcher might be disabled.
    if (isDisabled(result)) {
        return false;
    }

    // Test the string first in order to consume trailing chars greedily.
    int overlap = 0;
    if (!fString.isEmpty()) {
        overlap = segment.getCommonPrefixLength(fString);
        if (overlap == fString.length()) {
            segment.adjustOffset(fString.length());
            accept(segment, result);
            return false;
        }
    }

    int cp = segment.getCodePoint();
    if (cp != -1 && fUniSet->contains(cp)) {
        segment.adjustOffset(U16_LENGTH(cp));
        accept(segment, result);
        return false;
    }

    return overlap == segment.length();
}

bool SymbolMatcher::smokeTest(const StringSegment& segment) const {
    return segment.startsWith(*fUniSet) || segment.startsWith(fString);
}

UnicodeString SymbolMatcher::toString() const {
    // TODO: Customize output for each symbol
    return u"<Symbol>";
}


IgnorablesMatcher::IgnorablesMatcher(parse_flags_t parseFlags) :
        SymbolMatcher(
            {},
            (0 != (parseFlags & PARSE_FLAG_STRICT_IGNORABLES)) ?
                unisets::UNISET_KEY_STRICT_IGNORABLES :
                unisets::UNISET_KEY_DEFAULT_IGNORABLES) {
}

bool IgnorablesMatcher::isFlexible() const {
    return true;
}

UnicodeString IgnorablesMatcher::toString() const {
    return u"<Ignorables>";
}

bool IgnorablesMatcher::isDisabled(const ParsedNumber&) const {
    return false;
}

void IgnorablesMatcher::accept(StringSegment&, ParsedNumber&) const {
    // No-op
}


InfinityMatcher::InfinityMatcher(const DecimalFormatSymbols& dfs)
        : SymbolMatcher(dfs.getConstSymbol(DecimalFormatSymbols::kInfinitySymbol), unisets::UNISET_KEY_INFINITY_SIGN) {
}

bool InfinityMatcher::isDisabled(const ParsedNumber& result) const {
    return 0 != (result.flags & FLAG_INFINITY);
}

void InfinityMatcher::accept(StringSegment& segment, ParsedNumber& result) const {
    result.flags |= FLAG_INFINITY;
    result.setCharsConsumed(segment);
}


MinusSignMatcher::MinusSignMatcher(const DecimalFormatSymbols& dfs, bool allowTrailing)
        : SymbolMatcher(dfs.getConstSymbol(DecimalFormatSymbols::kMinusSignSymbol), unisets::UNISET_KEY_MINUS_SIGN),
          fAllowTrailing(allowTrailing) {
}

bool MinusSignMatcher::isDisabled(const ParsedNumber& result) const {
    return !fAllowTrailing && result.seenNumber();
}

void MinusSignMatcher::accept(StringSegment& segment, ParsedNumber& result) const {
    result.flags |= FLAG_NEGATIVE;
    result.setCharsConsumed(segment);
}


NanMatcher::NanMatcher(const DecimalFormatSymbols& dfs)
        : SymbolMatcher(dfs.getConstSymbol(DecimalFormatSymbols::kNaNSymbol), unisets::UNISET_KEY_EMPTY) {
}

bool NanMatcher::isDisabled(const ParsedNumber& result) const {
    return result.seenNumber();
}

void NanMatcher::accept(StringSegment& segment, ParsedNumber& result) const {
    result.flags |= FLAG_NAN;
    result.setCharsConsumed(segment);
}


PaddingMatcher::PaddingMatcher(const UnicodeString& padString)
        : SymbolMatcher(padString, unisets::UNISET_KEY_EMPTY) {}

bool PaddingMatcher::isFlexible() const {
    return true;
}

bool PaddingMatcher::isDisabled(const ParsedNumber&) const {
    return false;
}

void PaddingMatcher::accept(StringSegment&, ParsedNumber&) const {
    // No-op
}


PercentMatcher::PercentMatcher(const DecimalFormatSymbols& dfs)
        : SymbolMatcher(dfs.getConstSymbol(DecimalFormatSymbols::kPercentSymbol), unisets::UNISET_KEY_PERCENT_SIGN) {
}

bool PercentMatcher::isDisabled(const ParsedNumber& result) const {
    return 0 != (result.flags & FLAG_PERCENT);
}

void PercentMatcher::accept(StringSegment& segment, ParsedNumber& result) const {
    result.flags |= FLAG_PERCENT;
    result.setCharsConsumed(segment);
}


PermilleMatcher::PermilleMatcher(const DecimalFormatSymbols& dfs)
        : SymbolMatcher(dfs.getConstSymbol(DecimalFormatSymbols::kPerMillSymbol), unisets::UNISET_KEY_PERMILLE_SIGN) {
}

bool PermilleMatcher::isDisabled(const ParsedNumber& result) const {
    return 0 != (result.flags & FLAG_PERMILLE);
}

void PermilleMatcher::accept(StringSegment& segment, ParsedNumber& result) const {
    result.flags |= FLAG_PERMILLE;
    result.setCharsConsumed(segment);
}


PlusSignMatcher::PlusSignMatcher(const DecimalFormatSymbols& dfs, bool allowTrailing)
        : SymbolMatcher(dfs.getConstSymbol(DecimalFormatSymbols::kPlusSignSymbol), unisets::UNISET_KEY_PLUS_SIGN),
          fAllowTrailing(allowTrailing) {
}

bool PlusSignMatcher::isDisabled(const ParsedNumber& result) const {
    return !fAllowTrailing && result.seenNumber();
}

void PlusSignMatcher::accept(StringSegment& segment, ParsedNumber& result) const {
    result.setCharsConsumed(segment);
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif





using namespace icu;
using namespace icu::numparse;
using namespace icu::numparse::impl;


void RequireAffixValidator::postProcess(ParsedNumber& result) const {
    if (result.prefix.isBogus() || result.suffix.isBogus()) {
        // We saw a prefix or a suffix but not both. Fail the parse.
        result.flags |= FLAG_FAIL;
    }
}

UnicodeString RequireAffixValidator::toString() const {
    return u"<ReqAffix>";
}


void RequireCurrencyValidator::postProcess(ParsedNumber& result) const {
    if (result.currencyCode[0] == 0) {
        result.flags |= FLAG_FAIL;
    }
}

UnicodeString RequireCurrencyValidator::toString() const {
    return u"<ReqCurrency>";
}


RequireDecimalSeparatorValidator::RequireDecimalSeparatorValidator(bool patternHasDecimalSeparator)
        : fPatternHasDecimalSeparator(patternHasDecimalSeparator) {
}

void RequireDecimalSeparatorValidator::postProcess(ParsedNumber& result) const {
    bool parseHasDecimalSeparator = 0 != (result.flags & FLAG_HAS_DECIMAL_SEPARATOR);
    if (parseHasDecimalSeparator != fPatternHasDecimalSeparator) {
        result.flags |= FLAG_FAIL;
    }
}

UnicodeString RequireDecimalSeparatorValidator::toString() const {
    return u"<ReqDecimal>";
}


void RequireNumberValidator::postProcess(ParsedNumber& result) const {
    // Require that a number is matched.
    if (!result.seenNumber()) {
        result.flags |= FLAG_FAIL;
    }
}

UnicodeString RequireNumberValidator::toString() const {
    return u"<ReqNumber>";
}

MultiplierParseHandler::MultiplierParseHandler(::icu::number::Scale multiplier)
        : fMultiplier(std::move(multiplier)) {}

void MultiplierParseHandler::postProcess(ParsedNumber& result) const {
    if (!result.quantity.bogus) {
        fMultiplier.applyReciprocalTo(result.quantity);
        // NOTE: It is okay if the multiplier was negative.
    }
}

UnicodeString MultiplierParseHandler::toString() const {
    return u"<Scale>";
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif

// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING


// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html




#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING

#include <atomic>






/**
 * \file
 * \brief C++ API: Library for localized formatting of number, currency, and unit ranges.
 *
 * The main entrypoint to the formatting of ranges of numbers, including currencies and other units of measurement.
 * <p>
 * Usage example:
 * <p>
 * <pre>
 * NumberRangeFormatter::with()
 *     .identityFallback(UNUM_IDENTITY_FALLBACK_APPROXIMATELY_OR_SINGLE_VALUE)
 *     .numberFormatterFirst(NumberFormatter::with().adoptUnit(MeasureUnit::createMeter()))
 *     .numberFormatterSecond(NumberFormatter::with().adoptUnit(MeasureUnit::createKilometer()))
 *     .locale("en-GB")
 *     .formatRange(750, 1.2, status)
 *     .toString(status);
 * // => "750 m - 1.2 km"
 * </pre>
 * <p>
 * Like NumberFormatter, NumberRangeFormatter instances (i.e., LocalizedNumberRangeFormatter
 * and UnlocalizedNumberRangeFormatter) are immutable and thread-safe. This API is based on the
 * <em>fluent</em> design pattern popularized by libraries such as Google's Guava.
 *
 * @author Shane Carr
 */


/**
 * Defines how to merge fields that are identical across the range sign.
 *
 * @stable ICU 63
 */
typedef enum UNumberRangeCollapse {
    /**
     * Use locale data and heuristics to determine how much of the string to collapse. Could end up collapsing none,
     * some, or all repeated pieces in a locale-sensitive way.
     *
     * The heuristics used for this option are subject to change over time.
     *
     * @stable ICU 63
     */
    UNUM_RANGE_COLLAPSE_AUTO,

    /**
     * Do not collapse any part of the number. Example: "3.2 thousand kilograms – 5.3 thousand kilograms"
     *
     * @stable ICU 63
     */
    UNUM_RANGE_COLLAPSE_NONE,

    /**
     * Collapse the unit part of the number, but not the notation, if present. Example: "3.2 thousand – 5.3 thousand
     * kilograms"
     *
     * @stable ICU 63
     */
    UNUM_RANGE_COLLAPSE_UNIT,

    /**
     * Collapse any field that is equal across the range sign. May introduce ambiguity on the magnitude of the
     * number. Example: "3.2 – 5.3 thousand kilograms"
     *
     * @stable ICU 63
     */
    UNUM_RANGE_COLLAPSE_ALL
} UNumberRangeCollapse;

/**
 * Defines the behavior when the two numbers in the range are identical after rounding. To programmatically detect
 * when the identity fallback is used, compare the lower and upper BigDecimals via FormattedNumber.
 *
 * @stable ICU 63
 * @see NumberRangeFormatter
 */
typedef enum UNumberRangeIdentityFallback {
    /**
     * Show the number as a single value rather than a range. Example: "$5"
     *
     * @stable ICU 63
     */
    UNUM_IDENTITY_FALLBACK_SINGLE_VALUE,

    /**
     * Show the number using a locale-sensitive approximation pattern. If the numbers were the same before rounding,
     * show the single value. Example: "~$5" or "$5"
     *
     * @stable ICU 63
     */
    UNUM_IDENTITY_FALLBACK_APPROXIMATELY_OR_SINGLE_VALUE,

    /**
     * Show the number using a locale-sensitive approximation pattern. Use the range pattern always, even if the
     * inputs are the same. Example: "~$5"
     *
     * @stable ICU 63
     */
    UNUM_IDENTITY_FALLBACK_APPROXIMATELY,

    /**
     * Show the number as the range of two equal values. Use the range pattern always, even if the inputs are the
     * same. Example (with RangeCollapse.NONE): "$5 – $5"
     *
     * @stable ICU 63
     */
    UNUM_IDENTITY_FALLBACK_RANGE
} UNumberRangeIdentityFallback;

/**
 * Used in the result class FormattedNumberRange to indicate to the user whether the numbers formatted in the range
 * were equal or not, and whether or not the identity fallback was applied.
 *
 * @stable ICU 63
 * @see NumberRangeFormatter
 */
typedef enum UNumberRangeIdentityResult {
    /**
     * Used to indicate that the two numbers in the range were equal, even before any rounding rules were applied.
     *
     * @stable ICU 63
     * @see NumberRangeFormatter
     */
    UNUM_IDENTITY_RESULT_EQUAL_BEFORE_ROUNDING,

    /**
     * Used to indicate that the two numbers in the range were equal, but only after rounding rules were applied.
     *
     * @stable ICU 63
     * @see NumberRangeFormatter
     */
    UNUM_IDENTITY_RESULT_EQUAL_AFTER_ROUNDING,

    /**
     * Used to indicate that the two numbers in the range were not equal, even after rounding rules were applied.
     *
     * @stable ICU 63
     * @see NumberRangeFormatter
     */
    UNUM_IDENTITY_RESULT_NOT_EQUAL,

#ifndef U_HIDE_INTERNAL_API
    /**
     * The number of entries in this enum.
     * @internal
     */
    UNUM_IDENTITY_RESULT_COUNT
#endif

} UNumberRangeIdentityResult;

U_NAMESPACE_BEGIN

namespace number {  // icu::number

// Forward declarations:
class UnlocalizedNumberRangeFormatter;
class LocalizedNumberRangeFormatter;
class FormattedNumberRange;

namespace impl {

// Forward declarations:
struct RangeMacroProps;
class DecimalQuantity;
class UFormattedNumberRangeData;
class NumberRangeFormatterImpl;

} // namespace impl

/**
 * \cond
 * Export an explicit template instantiation. See datefmt.h
 * (When building DLLs for Windows this is required.)
 */
#if U_PLATFORM == U_PF_WINDOWS && !defined(U_IN_DOXYGEN)
} // namespace icu::number
U_NAMESPACE_END

template struct U_I18N_API std::atomic< U_NAMESPACE_QUALIFIER number::impl::NumberRangeFormatterImpl*>;

U_NAMESPACE_BEGIN
namespace number {  // icu::number
#endif
/** \endcond */

// Other helper classes would go here, but there are none.

namespace impl {  // icu::number::impl

// Do not enclose entire MacroProps with #ifndef U_HIDE_INTERNAL_API, needed for a protected field
/** @internal */
struct U_I18N_API RangeMacroProps : public UMemory {
    /** @internal */
    UnlocalizedNumberFormatter formatter1; // = NumberFormatter::with();

    /** @internal */
    UnlocalizedNumberFormatter formatter2; // = NumberFormatter::with();

    /** @internal */
    bool singleFormatter = true;

    /** @internal */
    UNumberRangeCollapse collapse = UNUM_RANGE_COLLAPSE_AUTO;

    /** @internal */
    UNumberRangeIdentityFallback identityFallback = UNUM_IDENTITY_FALLBACK_APPROXIMATELY;

    /** @internal */
    Locale locale;

    // NOTE: Uses default copy and move constructors.

    /**
     * Check all members for errors.
     * @internal
     */
    bool copyErrorTo(UErrorCode &status) const {
        return formatter1.copyErrorTo(status) || formatter2.copyErrorTo(status);
    }
};

} // namespace impl

/**
 * An abstract base class for specifying settings related to number formatting. This class is implemented by
 * {@link UnlocalizedNumberRangeFormatter} and {@link LocalizedNumberRangeFormatter}. This class is not intended for
 * public subclassing.
 */
template<typename Derived>
class U_I18N_API NumberRangeFormatterSettings {
  public:
    /**
     * Sets the NumberFormatter instance to use for the numbers in the range. The same formatter is applied to both
     * sides of the range.
     * <p>
     * The NumberFormatter instances must not have a locale applied yet; the locale specified on the
     * NumberRangeFormatter will be used.
     *
     * @param formatter
     *            The formatter to use for both numbers in the range.
     * @return The fluent chain.
     * @stable ICU 63
     */
    Derived numberFormatterBoth(const UnlocalizedNumberFormatter &formatter) const &;

    /**
     * Overload of numberFormatterBoth() for use on an rvalue reference.
     *
     * @param formatter
     *            The formatter to use for both numbers in the range.
     * @return The fluent chain.
     * @see #numberFormatterBoth
     * @stable ICU 63
     */
    Derived numberFormatterBoth(const UnlocalizedNumberFormatter &formatter) &&;

    /**
     * Overload of numberFormatterBoth() for use on an rvalue reference.
     *
     * @param formatter
     *            The formatter to use for both numbers in the range.
     * @return The fluent chain.
     * @see #numberFormatterBoth
     * @stable ICU 63
     */
    Derived numberFormatterBoth(UnlocalizedNumberFormatter &&formatter) const &;

    /**
     * Overload of numberFormatterBoth() for use on an rvalue reference.
     *
     * @param formatter
     *            The formatter to use for both numbers in the range.
     * @return The fluent chain.
     * @see #numberFormatterBoth
     * @stable ICU 63
     */
    Derived numberFormatterBoth(UnlocalizedNumberFormatter &&formatter) &&;

    /**
     * Sets the NumberFormatter instance to use for the first number in the range.
     * <p>
     * The NumberFormatter instances must not have a locale applied yet; the locale specified on the
     * NumberRangeFormatter will be used.
     *
     * @param formatterFirst
     *            The formatter to use for the first number in the range.
     * @return The fluent chain.
     * @stable ICU 63
     */
    Derived numberFormatterFirst(const UnlocalizedNumberFormatter &formatterFirst) const &;

    /**
     * Overload of numberFormatterFirst() for use on an rvalue reference.
     *
     * @param formatterFirst
     *            The formatter to use for the first number in the range.
     * @return The fluent chain.
     * @see #numberFormatterFirst
     * @stable ICU 63
     */
    Derived numberFormatterFirst(const UnlocalizedNumberFormatter &formatterFirst) &&;

    /**
     * Overload of numberFormatterFirst() for use on an rvalue reference.
     *
     * @param formatterFirst
     *            The formatter to use for the first number in the range.
     * @return The fluent chain.
     * @see #numberFormatterFirst
     * @stable ICU 63
     */
    Derived numberFormatterFirst(UnlocalizedNumberFormatter &&formatterFirst) const &;

    /**
     * Overload of numberFormatterFirst() for use on an rvalue reference.
     *
     * @param formatterFirst
     *            The formatter to use for the first number in the range.
     * @return The fluent chain.
     * @see #numberFormatterFirst
     * @stable ICU 63
     */
    Derived numberFormatterFirst(UnlocalizedNumberFormatter &&formatterFirst) &&;

    /**
     * Sets the NumberFormatter instance to use for the second number in the range.
     * <p>
     * The NumberFormatter instances must not have a locale applied yet; the locale specified on the
     * NumberRangeFormatter will be used.
     *
     * @param formatterSecond
     *            The formatter to use for the second number in the range.
     * @return The fluent chain.
     * @stable ICU 63
     */
    Derived numberFormatterSecond(const UnlocalizedNumberFormatter &formatterSecond) const &;

    /**
     * Overload of numberFormatterSecond() for use on an rvalue reference.
     *
     * @param formatterSecond
     *            The formatter to use for the second number in the range.
     * @return The fluent chain.
     * @see #numberFormatterSecond
     * @stable ICU 63
     */
    Derived numberFormatterSecond(const UnlocalizedNumberFormatter &formatterSecond) &&;

    /**
     * Overload of numberFormatterSecond() for use on an rvalue reference.
     *
     * @param formatterSecond
     *            The formatter to use for the second number in the range.
     * @return The fluent chain.
     * @see #numberFormatterSecond
     * @stable ICU 63
     */
    Derived numberFormatterSecond(UnlocalizedNumberFormatter &&formatterSecond) const &;

    /**
     * Overload of numberFormatterSecond() for use on an rvalue reference.
     *
     * @param formatterSecond
     *            The formatter to use for the second number in the range.
     * @return The fluent chain.
     * @see #numberFormatterSecond
     * @stable ICU 63
     */
    Derived numberFormatterSecond(UnlocalizedNumberFormatter &&formatterSecond) &&;

    /**
     * Sets the aggressiveness of "collapsing" fields across the range separator. Possible values:
     * <p>
     * <ul>
     * <li>ALL: "3-5K miles"</li>
     * <li>UNIT: "3K - 5K miles"</li>
     * <li>NONE: "3K miles - 5K miles"</li>
     * <li>AUTO: usually UNIT or NONE, depending on the locale and formatter settings</li>
     * </ul>
     * <p>
     * The default value is AUTO.
     *
     * @param collapse
     *            The collapsing strategy to use for this range.
     * @return The fluent chain.
     * @stable ICU 63
     */
    Derived collapse(UNumberRangeCollapse collapse) const &;

    /**
     * Overload of collapse() for use on an rvalue reference.
     *
     * @param collapse
     *            The collapsing strategy to use for this range.
     * @return The fluent chain.
     * @see #collapse
     * @stable ICU 63
     */
    Derived collapse(UNumberRangeCollapse collapse) &&;

    /**
     * Sets the behavior when the two sides of the range are the same. This could happen if the same two numbers are
     * passed to the formatRange function, or if different numbers are passed to the function but they become the same
     * after rounding rules are applied. Possible values:
     * <p>
     * <ul>
     * <li>SINGLE_VALUE: "5 miles"</li>
     * <li>APPROXIMATELY_OR_SINGLE_VALUE: "~5 miles" or "5 miles", depending on whether the number was the same before
     * rounding was applied</li>
     * <li>APPROXIMATELY: "~5 miles"</li>
     * <li>RANGE: "5-5 miles" (with collapse=UNIT)</li>
     * </ul>
     * <p>
     * The default value is APPROXIMATELY.
     *
     * @param identityFallback
     *            The strategy to use when formatting two numbers that end up being the same.
     * @return The fluent chain.
     * @stable ICU 63
     */
    Derived identityFallback(UNumberRangeIdentityFallback identityFallback) const &;

    /**
     * Overload of identityFallback() for use on an rvalue reference.
     *
     * @param identityFallback
     *            The strategy to use when formatting two numbers that end up being the same.
     * @return The fluent chain.
     * @see #identityFallback
     * @stable ICU 63
     */
    Derived identityFallback(UNumberRangeIdentityFallback identityFallback) &&;

#ifndef U_HIDE_DRAFT_API
    /**
     * Returns the current (Un)LocalizedNumberRangeFormatter as a LocalPointer
     * wrapping a heap-allocated copy of the current object.
     *
     * This is equivalent to new-ing the move constructor with a value object
     * as the argument.
     *
     * @return A wrapped (Un)LocalizedNumberRangeFormatter pointer, or a wrapped
     *         nullptr on failure.
     * @draft ICU 64
     */
    LocalPointer<Derived> clone() const &;

    /**
     * Overload of clone for use on an rvalue reference.
     *
     * @return A wrapped (Un)LocalizedNumberRangeFormatter pointer, or a wrapped
     *         nullptr on failure.
     * @draft ICU 64
     */
    LocalPointer<Derived> clone() &&;
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Sets the UErrorCode if an error occurred in the fluent chain.
     * Preserves older error codes in the outErrorCode.
     * @return TRUE if U_FAILURE(outErrorCode)
     * @stable ICU 63
     */
    UBool copyErrorTo(UErrorCode &outErrorCode) const {
        if (U_FAILURE(outErrorCode)) {
            // Do not overwrite the older error code
            return TRUE;
        }
        fMacros.copyErrorTo(outErrorCode);
        return U_FAILURE(outErrorCode);
    }

    // NOTE: Uses default copy and move constructors.

  private:
    number::impl::RangeMacroProps fMacros;

    // Don't construct me directly!  Use (Un)LocalizedNumberFormatter.
    NumberRangeFormatterSettings() = default;

    friend class LocalizedNumberRangeFormatter;
    friend class UnlocalizedNumberRangeFormatter;
};

/**
 * A NumberRangeFormatter that does not yet have a locale. In order to format, a locale must be specified.
 *
 * Instances of this class are immutable and thread-safe.
 *
 * @see NumberRangeFormatter
 * @stable ICU 63
 */
class U_I18N_API UnlocalizedNumberRangeFormatter
        : public NumberRangeFormatterSettings<UnlocalizedNumberRangeFormatter>, public UMemory {

  public:
    /**
     * Associate the given locale with the number range formatter. The locale is used for picking the
     * appropriate symbols, formats, and other data for number display.
     *
     * @param locale
     *            The locale to use when loading data for number formatting.
     * @return The fluent chain.
     * @stable ICU 63
     */
    LocalizedNumberRangeFormatter locale(const icu::Locale &locale) const &;

    /**
     * Overload of locale() for use on an rvalue reference.
     *
     * @param locale
     *            The locale to use when loading data for number formatting.
     * @return The fluent chain.
     * @see #locale
     * @stable ICU 63
     */
    LocalizedNumberRangeFormatter locale(const icu::Locale &locale) &&;

    /**
     * Default constructor: puts the formatter into a valid but undefined state.
     *
     * @stable ICU 63
     */
    UnlocalizedNumberRangeFormatter() = default;

    /**
     * Returns a copy of this UnlocalizedNumberRangeFormatter.
     * @stable ICU 63
     */
    UnlocalizedNumberRangeFormatter(const UnlocalizedNumberRangeFormatter &other);

    /**
     * Move constructor:
     * The source UnlocalizedNumberRangeFormatter will be left in a valid but undefined state.
     * @stable ICU 63
     */
    UnlocalizedNumberRangeFormatter(UnlocalizedNumberRangeFormatter&& src) U_NOEXCEPT;

    /**
     * Copy assignment operator.
     * @stable ICU 63
     */
    UnlocalizedNumberRangeFormatter& operator=(const UnlocalizedNumberRangeFormatter& other);

    /**
     * Move assignment operator:
     * The source UnlocalizedNumberRangeFormatter will be left in a valid but undefined state.
     * @stable ICU 63
     */
    UnlocalizedNumberRangeFormatter& operator=(UnlocalizedNumberRangeFormatter&& src) U_NOEXCEPT;

  private:
    explicit UnlocalizedNumberRangeFormatter(
            const NumberRangeFormatterSettings<UnlocalizedNumberRangeFormatter>& other);

    explicit UnlocalizedNumberRangeFormatter(
            NumberRangeFormatterSettings<UnlocalizedNumberRangeFormatter>&& src) U_NOEXCEPT;

    // To give the fluent setters access to this class's constructor:
    friend class NumberRangeFormatterSettings<UnlocalizedNumberRangeFormatter>;

    // To give NumberRangeFormatter::with() access to this class's constructor:
    friend class NumberRangeFormatter;
};

/**
 * A NumberRangeFormatter that has a locale associated with it; this means .formatRange() methods are available.
 *
 * Instances of this class are immutable and thread-safe.
 *
 * @see NumberFormatter
 * @stable ICU 63
 */
class U_I18N_API LocalizedNumberRangeFormatter
        : public NumberRangeFormatterSettings<LocalizedNumberRangeFormatter>, public UMemory {
  public:
    /**
     * Format the given Formattables to a string using the settings specified in the NumberRangeFormatter fluent setting
     * chain.
     *
     * @param first
     *            The first number in the range, usually to the left in LTR locales.
     * @param second
     *            The second number in the range, usually to the right in LTR locales.
     * @param status
     *            Set if an error occurs while formatting.
     * @return A FormattedNumberRange object; call .toString() to get the string.
     * @stable ICU 63
     */
    FormattedNumberRange formatFormattableRange(
        const Formattable& first, const Formattable& second, UErrorCode& status) const;

    /**
     * Default constructor: puts the formatter into a valid but undefined state.
     *
     * @stable ICU 63
     */
    LocalizedNumberRangeFormatter() = default;

    /**
     * Returns a copy of this LocalizedNumberRangeFormatter.
     * @stable ICU 63
     */
    LocalizedNumberRangeFormatter(const LocalizedNumberRangeFormatter &other);

    /**
     * Move constructor:
     * The source LocalizedNumberRangeFormatter will be left in a valid but undefined state.
     * @stable ICU 63
     */
    LocalizedNumberRangeFormatter(LocalizedNumberRangeFormatter&& src) U_NOEXCEPT;

    /**
     * Copy assignment operator.
     * @stable ICU 63
     */
    LocalizedNumberRangeFormatter& operator=(const LocalizedNumberRangeFormatter& other);

    /**
     * Move assignment operator:
     * The source LocalizedNumberRangeFormatter will be left in a valid but undefined state.
     * @stable ICU 63
     */
    LocalizedNumberRangeFormatter& operator=(LocalizedNumberRangeFormatter&& src) U_NOEXCEPT;

#ifndef U_HIDE_INTERNAL_API

    /**
     * @param results
     *            The results object. This method will mutate it to save the results.
     * @param equalBeforeRounding
     *            Whether the number was equal before copying it into a DecimalQuantity.
     *            Used for determining the identity fallback behavior.
     * @param status
     *            Set if an error occurs while formatting.
     * @internal
     */
    void formatImpl(impl::UFormattedNumberRangeData& results, bool equalBeforeRounding,
                    UErrorCode& status) const;

#endif  /* U_HIDE_INTERNAL_API */

    /**
     * Destruct this LocalizedNumberRangeFormatter, cleaning up any memory it might own.
     * @stable ICU 63
     */
    ~LocalizedNumberRangeFormatter();

  private:
    std::atomic<impl::NumberRangeFormatterImpl*> fAtomicFormatter = {};

    const number::impl::NumberRangeFormatterImpl* getFormatter(UErrorCode& stauts) const;

    explicit LocalizedNumberRangeFormatter(
        const NumberRangeFormatterSettings<LocalizedNumberRangeFormatter>& other);

    explicit LocalizedNumberRangeFormatter(
        NumberRangeFormatterSettings<LocalizedNumberRangeFormatter>&& src) U_NOEXCEPT;

    LocalizedNumberRangeFormatter(const number::impl::RangeMacroProps &macros, const Locale &locale);

    LocalizedNumberRangeFormatter(impl::RangeMacroProps &&macros, const Locale &locale);

    void clear();

    // To give the fluent setters access to this class's constructor:
    friend class NumberRangeFormatterSettings<UnlocalizedNumberRangeFormatter>;
    friend class NumberRangeFormatterSettings<LocalizedNumberRangeFormatter>;

    // To give UnlocalizedNumberRangeFormatter::locale() access to this class's constructor:
    friend class UnlocalizedNumberRangeFormatter;
};

/**
 * The result of a number range formatting operation. This class allows the result to be exported in several data types,
 * including a UnicodeString and a FieldPositionIterator.
 *
 * Instances of this class are immutable and thread-safe.
 *
 * @stable ICU 63
 */
class U_I18N_API FormattedNumberRange : public UMemory, public FormattedValue {
  public:
    // Copybrief: this method is older than the parent method
    /**
     * @copybrief FormattedValue::toString()
     *
     * For more information, see FormattedValue::toString()
     *
     * @stable ICU 63
     */
    UnicodeString toString(UErrorCode& status) const U_OVERRIDE;

    // Copydoc: this method is new in ICU 64
    /** @copydoc FormattedValue::toTempString() */
    UnicodeString toTempString(UErrorCode& status) const U_OVERRIDE;

    // Copybrief: this method is older than the parent method
    /**
     * @copybrief FormattedValue::appendTo()
     *
     * For more information, see FormattedValue::appendTo()
     *
     * @stable ICU 63
     */
    Appendable &appendTo(Appendable &appendable, UErrorCode& status) const U_OVERRIDE;

    // Copydoc: this method is new in ICU 64
    /** @copydoc FormattedValue::nextPosition() */
    UBool nextPosition(ConstrainedFieldPosition& cfpos, UErrorCode& status) const U_OVERRIDE;

#ifndef U_HIDE_DRAFT_API
    /**
     * Determines the start (inclusive) and end (exclusive) indices of the next occurrence of the given
     * <em>field</em> in the output string. This allows you to determine the locations of, for example,
     * the integer part, fraction part, or symbols.
     *
     * If both sides of the range have the same field, the field will occur twice, once before the
     * range separator and once after the range separator, if applicable.
     *
     * If a field occurs just once, calling this method will find that occurrence and return it. If a
     * field occurs multiple times, this method may be called repeatedly with the following pattern:
     *
     * <pre>
     * FieldPosition fpos(UNUM_INTEGER_FIELD);
     * while (formattedNumberRange.nextFieldPosition(fpos, status)) {
     *   // do something with fpos.
     * }
     * </pre>
     *
     * This method is useful if you know which field to query. If you want all available field position
     * information, use #getAllFieldPositions().
     *
     * @param fieldPosition
     *            Input+output variable. See {@link FormattedNumber#nextFieldPosition}.
     * @param status
     *            Set if an error occurs while populating the FieldPosition.
     * @return TRUE if a new occurrence of the field was found; FALSE otherwise.
     * @draft ICU 63
     * @see UNumberFormatFields
     */
    UBool nextFieldPosition(FieldPosition& fieldPosition, UErrorCode& status) const;

    /**
     * Export the formatted number range to a FieldPositionIterator. This allows you to determine which characters in
     * the output string correspond to which <em>fields</em>, such as the integer part, fraction part, and sign.
     *
     * If information on only one field is needed, use #nextFieldPosition() instead.
     *
     * @param iterator
     *            The FieldPositionIterator to populate with all of the fields present in the formatted number.
     * @param status
     *            Set if an error occurs while populating the FieldPositionIterator.
     * @draft ICU 63
     * @see UNumberFormatFields
     */
    void getAllFieldPositions(FieldPositionIterator &iterator, UErrorCode &status) const;

    /**
     * Export the first formatted number as a decimal number. This endpoint
     * is useful for obtaining the exact number being printed after scaling
     * and rounding have been applied by the number range formatting pipeline.
     *
     * The syntax of the unformatted number is a "numeric string"
     * as defined in the Decimal Arithmetic Specification, available at
     * http://speleotrove.com/decimal
     *
     * @return A decimal representation of the first formatted number.
     * @draft ICU 63
     * @see NumberRangeFormatter
     * @see #getSecondDecimal
     */
    UnicodeString getFirstDecimal(UErrorCode& status) const;

    /**
     * Export the second formatted number as a decimal number. This endpoint
     * is useful for obtaining the exact number being printed after scaling
     * and rounding have been applied by the number range formatting pipeline.
     *
     * The syntax of the unformatted number is a "numeric string"
     * as defined in the Decimal Arithmetic Specification, available at
     * http://speleotrove.com/decimal
     *
     * @return A decimal representation of the second formatted number.
     * @draft ICU 63
     * @see NumberRangeFormatter
     * @see #getFirstDecimal
     */
    UnicodeString getSecondDecimal(UErrorCode& status) const;
#endif // U_HIDE_DRAFT_API

    /**
     * Returns whether the pair of numbers was successfully formatted as a range or whether an identity fallback was
     * used. For example, if the first and second number were the same either before or after rounding occurred, an
     * identity fallback was used.
     *
     * @return An indication the resulting identity situation in the formatted number range.
     * @stable ICU 63
     * @see UNumberRangeIdentityFallback
     */
    UNumberRangeIdentityResult getIdentityResult(UErrorCode& status) const;

    /**
     * Copying not supported; use move constructor instead.
     */
    FormattedNumberRange(const FormattedNumberRange&) = delete;

    /**
     * Copying not supported; use move assignment instead.
     */
    FormattedNumberRange& operator=(const FormattedNumberRange&) = delete;

    /**
     * Move constructor:
     * Leaves the source FormattedNumberRange in an undefined state.
     * @stable ICU 63
     */
    FormattedNumberRange(FormattedNumberRange&& src) U_NOEXCEPT;

    /**
     * Move assignment:
     * Leaves the source FormattedNumberRange in an undefined state.
     * @stable ICU 63
     */
    FormattedNumberRange& operator=(FormattedNumberRange&& src) U_NOEXCEPT;

    /**
     * Destruct an instance of FormattedNumberRange, cleaning up any memory it might own.
     * @stable ICU 63
     */
    ~FormattedNumberRange();

  private:
    // Can't use LocalPointer because UFormattedNumberRangeData is forward-declared
    const number::impl::UFormattedNumberRangeData *fData;

    // Error code for the terminal methods
    UErrorCode fErrorCode;

    /**
     * Internal constructor from data type. Adopts the data pointer.
     */
    explicit FormattedNumberRange(impl::UFormattedNumberRangeData *results)
        : fData(results), fErrorCode(U_ZERO_ERROR) {}

    explicit FormattedNumberRange(UErrorCode errorCode)
        : fData(nullptr), fErrorCode(errorCode) {}

    void getAllFieldPositionsImpl(FieldPositionIteratorHandler& fpih, UErrorCode& status) const;

    // To give LocalizedNumberRangeFormatter format methods access to this class's constructor:
    friend class LocalizedNumberRangeFormatter;
};

/**
 * See the main description in numberrangeformatter.h for documentation and examples.
 *
 * @stable ICU 63
 */
class U_I18N_API NumberRangeFormatter final {
  public:
    /**
     * Call this method at the beginning of a NumberRangeFormatter fluent chain in which the locale is not currently
     * known at the call site.
     *
     * @return An {@link UnlocalizedNumberRangeFormatter}, to be used for chaining.
     * @stable ICU 63
     */
    static UnlocalizedNumberRangeFormatter with();

    /**
     * Call this method at the beginning of a NumberRangeFormatter fluent chain in which the locale is known at the call
     * site.
     *
     * @param locale
     *            The locale from which to load formats and symbols for number range formatting.
     * @return A {@link LocalizedNumberRangeFormatter}, to be used for chaining.
     * @stable ICU 63
     */
    static LocalizedNumberRangeFormatter withLocale(const Locale &locale);

    /**
     * Use factory methods instead of the constructor to create a NumberFormatter.
     */
    NumberRangeFormatter() = delete;
};

}  // namespace number
U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */









U_NAMESPACE_BEGIN namespace number {
namespace impl {


/**
 * Class similar to UFormattedNumberData.
 *
 * Has incomplete magic number logic that will need to be finished
 * if this is to be exposed as C API in the future.
 *
 * Possible magic number: 0x46445200
 * Reads in ASCII as "FDR" (FormatteDnumberRange with room at the end)
 */
class UFormattedNumberRangeData : public FormattedValueStringBuilderImpl {
public:
    UFormattedNumberRangeData() : FormattedValueStringBuilderImpl(0) {}
    virtual ~UFormattedNumberRangeData();

    DecimalQuantity quantity1;
    DecimalQuantity quantity2;
    UNumberRangeIdentityResult identityResult = UNUM_IDENTITY_RESULT_COUNT;
};


class StandardPluralRanges : public UMemory {
  public:
    void initialize(const Locale& locale, UErrorCode& status);
    StandardPlural::Form resolve(StandardPlural::Form first, StandardPlural::Form second) const;

    /** Used for data loading. */
    void addPluralRange(
        StandardPlural::Form first,
        StandardPlural::Form second,
        StandardPlural::Form result);

    /** Used for data loading. */
    void setCapacity(int32_t length);

  private:
    struct StandardPluralRangeTriple {
        StandardPlural::Form first;
        StandardPlural::Form second;
        StandardPlural::Form result;
    };

    // TODO: An array is simple here, but it results in linear lookup time.
    // Certain locales have 20-30 entries in this list.
    // Consider changing to a smarter data structure.
    typedef MaybeStackArray<StandardPluralRangeTriple, 3> PluralRangeTriples;
    PluralRangeTriples fTriples;
    int32_t fTriplesLen = 0;
};


class NumberRangeFormatterImpl : public UMemory {
  public:
    NumberRangeFormatterImpl(const RangeMacroProps& macros, UErrorCode& status);

    void format(UFormattedNumberRangeData& data, bool equalBeforeRounding, UErrorCode& status) const;

  private:
    NumberFormatterImpl formatterImpl1;
    NumberFormatterImpl formatterImpl2;
    bool fSameFormatters;

    UNumberRangeCollapse fCollapse;
    UNumberRangeIdentityFallback fIdentityFallback;

    SimpleFormatter fRangeFormatter;
    SimpleModifier fApproximatelyModifier;

    StandardPluralRanges fPluralRanges;

    void formatSingleValue(UFormattedNumberRangeData& data,
                           MicroProps& micros1, MicroProps& micros2,
                           UErrorCode& status) const;

    void formatApproximately(UFormattedNumberRangeData& data,
                             MicroProps& micros1, MicroProps& micros2,
                             UErrorCode& status) const;

    void formatRange(UFormattedNumberRangeData& data,
                     MicroProps& micros1, MicroProps& micros2,
                     UErrorCode& status) const;

    const Modifier& resolveModifierPlurals(const Modifier& first, const Modifier& second) const;
};


} // namespace impl
} // namespace number
U_NAMESPACE_END

#endif //__SOURCE_NUMRANGE_TYPES_H__




using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;


// This function needs to be declared in this namespace so it can be friended.
// NOTE: In Java, this logic is handled in the resolve() function.
void icu::number::impl::touchRangeLocales(RangeMacroProps& macros) {
    macros.formatter1.fMacros.locale = macros.locale;
    macros.formatter2.fMacros.locale = macros.locale;
}


template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterBoth(const UnlocalizedNumberFormatter& formatter) const& {
    Derived copy(*this);
    copy.fMacros.formatter1 = formatter;
    copy.fMacros.singleFormatter = true;
    touchRangeLocales(copy.fMacros);
    return copy;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterBoth(const UnlocalizedNumberFormatter& formatter) && {
    Derived move(std::move(*this));
    move.fMacros.formatter1 = formatter;
    move.fMacros.singleFormatter = true;
    touchRangeLocales(move.fMacros);
    return move;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterBoth(UnlocalizedNumberFormatter&& formatter) const& {
    Derived copy(*this);
    copy.fMacros.formatter1 = std::move(formatter);
    copy.fMacros.singleFormatter = true;
    touchRangeLocales(copy.fMacros);
    return copy;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterBoth(UnlocalizedNumberFormatter&& formatter) && {
    Derived move(std::move(*this));
    move.fMacros.formatter1 = std::move(formatter);
    move.fMacros.singleFormatter = true;
    touchRangeLocales(move.fMacros);
    return move;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterFirst(const UnlocalizedNumberFormatter& formatter) const& {
    Derived copy(*this);
    copy.fMacros.formatter1 = formatter;
    copy.fMacros.singleFormatter = false;
    touchRangeLocales(copy.fMacros);
    return copy;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterFirst(const UnlocalizedNumberFormatter& formatter) && {
    Derived move(std::move(*this));
    move.fMacros.formatter1 = formatter;
    move.fMacros.singleFormatter = false;
    touchRangeLocales(move.fMacros);
    return move;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterFirst(UnlocalizedNumberFormatter&& formatter) const& {
    Derived copy(*this);
    copy.fMacros.formatter1 = std::move(formatter);
    copy.fMacros.singleFormatter = false;
    touchRangeLocales(copy.fMacros);
    return copy;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterFirst(UnlocalizedNumberFormatter&& formatter) && {
    Derived move(std::move(*this));
    move.fMacros.formatter1 = std::move(formatter);
    move.fMacros.singleFormatter = false;
    touchRangeLocales(move.fMacros);
    return move;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterSecond(const UnlocalizedNumberFormatter& formatter) const& {
    Derived copy(*this);
    copy.fMacros.formatter2 = formatter;
    copy.fMacros.singleFormatter = false;
    touchRangeLocales(copy.fMacros);
    return copy;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterSecond(const UnlocalizedNumberFormatter& formatter) && {
    Derived move(std::move(*this));
    move.fMacros.formatter2 = formatter;
    move.fMacros.singleFormatter = false;
    touchRangeLocales(move.fMacros);
    return move;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterSecond(UnlocalizedNumberFormatter&& formatter) const& {
    Derived copy(*this);
    copy.fMacros.formatter2 = std::move(formatter);
    copy.fMacros.singleFormatter = false;
    touchRangeLocales(copy.fMacros);
    return copy;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::numberFormatterSecond(UnlocalizedNumberFormatter&& formatter) && {
    Derived move(std::move(*this));
    move.fMacros.formatter2 = std::move(formatter);
    move.fMacros.singleFormatter = false;
    touchRangeLocales(move.fMacros);
    return move;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::collapse(UNumberRangeCollapse collapse) const& {
    Derived copy(*this);
    copy.fMacros.collapse = collapse;
    return copy;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::collapse(UNumberRangeCollapse collapse) && {
    Derived move(std::move(*this));
    move.fMacros.collapse = collapse;
    return move;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::identityFallback(UNumberRangeIdentityFallback identityFallback) const& {
    Derived copy(*this);
    copy.fMacros.identityFallback = identityFallback;
    return copy;
}

template<typename Derived>
Derived NumberRangeFormatterSettings<Derived>::identityFallback(UNumberRangeIdentityFallback identityFallback) && {
    Derived move(std::move(*this));
    move.fMacros.identityFallback = identityFallback;
    return move;
}

template<typename Derived>
LocalPointer<Derived> NumberRangeFormatterSettings<Derived>::clone() const & {
    return LocalPointer<Derived>(new Derived(*this));
}

template<typename Derived>
LocalPointer<Derived> NumberRangeFormatterSettings<Derived>::clone() && {
    return LocalPointer<Derived>(new Derived(std::move(*this)));
}

// Declare all classes that implement NumberRangeFormatterSettings
// See https://stackoverflow.com/a/495056/1407170
template
class icu::number::NumberRangeFormatterSettings<icu::number::UnlocalizedNumberRangeFormatter>;
template
class icu::number::NumberRangeFormatterSettings<icu::number::LocalizedNumberRangeFormatter>;


UnlocalizedNumberRangeFormatter NumberRangeFormatter::with() {
    UnlocalizedNumberRangeFormatter result;
    return result;
}

LocalizedNumberRangeFormatter NumberRangeFormatter::withLocale(const Locale& locale) {
    return with().locale(locale);
}


template<typename T> using numrange_fluent_NFS = NumberRangeFormatterSettings<T>;
using numrange_fluent_LNF = LocalizedNumberRangeFormatter;
using numrange_fluent_UNF = UnlocalizedNumberRangeFormatter;

UnlocalizedNumberRangeFormatter::UnlocalizedNumberRangeFormatter(const numrange_fluent_UNF& other)
        : numrange_fluent_UNF(static_cast<const numrange_fluent_NFS<numrange_fluent_UNF>&>(other)) {}

UnlocalizedNumberRangeFormatter::UnlocalizedNumberRangeFormatter(const numrange_fluent_NFS<numrange_fluent_UNF>& other)
        : numrange_fluent_NFS<numrange_fluent_UNF>(other) {
    // No additional fields to assign
}

// Make default copy constructor call the NumberRangeFormatterSettings copy constructor.
UnlocalizedNumberRangeFormatter::UnlocalizedNumberRangeFormatter(numrange_fluent_UNF&& src) U_NOEXCEPT
        : numrange_fluent_UNF(static_cast<numrange_fluent_NFS<numrange_fluent_UNF>&&>(src)) {}

UnlocalizedNumberRangeFormatter::UnlocalizedNumberRangeFormatter(numrange_fluent_NFS<numrange_fluent_UNF>&& src) U_NOEXCEPT
        : numrange_fluent_NFS<numrange_fluent_UNF>(std::move(src)) {
    // No additional fields to assign
}

UnlocalizedNumberRangeFormatter& UnlocalizedNumberRangeFormatter::operator=(const numrange_fluent_UNF& other) {
    numrange_fluent_NFS<numrange_fluent_UNF>::operator=(static_cast<const numrange_fluent_NFS<numrange_fluent_UNF>&>(other));
    // No additional fields to assign
    return *this;
}

UnlocalizedNumberRangeFormatter& UnlocalizedNumberRangeFormatter::operator=(numrange_fluent_UNF&& src) U_NOEXCEPT {
    numrange_fluent_NFS<numrange_fluent_UNF>::operator=(static_cast<numrange_fluent_NFS<numrange_fluent_UNF>&&>(src));
    // No additional fields to assign
    return *this;
}

// Make default copy constructor call the NumberRangeFormatterSettings copy constructor.
LocalizedNumberRangeFormatter::LocalizedNumberRangeFormatter(const numrange_fluent_LNF& other)
        : numrange_fluent_LNF(static_cast<const numrange_fluent_NFS<numrange_fluent_LNF>&>(other)) {}

LocalizedNumberRangeFormatter::LocalizedNumberRangeFormatter(const numrange_fluent_NFS<numrange_fluent_LNF>& other)
        : numrange_fluent_NFS<numrange_fluent_LNF>(other) {
    // No additional fields to assign
}

LocalizedNumberRangeFormatter::LocalizedNumberRangeFormatter(LocalizedNumberRangeFormatter&& src) U_NOEXCEPT
        : numrange_fluent_LNF(static_cast<numrange_fluent_NFS<numrange_fluent_LNF>&&>(src)) {}

LocalizedNumberRangeFormatter::LocalizedNumberRangeFormatter(numrange_fluent_NFS<numrange_fluent_LNF>&& src) U_NOEXCEPT
        : numrange_fluent_NFS<numrange_fluent_LNF>(std::move(src)) {
    // Steal the compiled formatter
    numrange_fluent_LNF&& _src = static_cast<numrange_fluent_LNF&&>(src);
    auto* stolen = _src.fAtomicFormatter.exchange(nullptr);
    delete fAtomicFormatter.exchange(stolen);
}

LocalizedNumberRangeFormatter& LocalizedNumberRangeFormatter::operator=(const numrange_fluent_LNF& other) {
    numrange_fluent_NFS<numrange_fluent_LNF>::operator=(static_cast<const numrange_fluent_NFS<numrange_fluent_LNF>&>(other));
    // Do not steal; just clear
    delete fAtomicFormatter.exchange(nullptr);
    return *this;
}

LocalizedNumberRangeFormatter& LocalizedNumberRangeFormatter::operator=(numrange_fluent_LNF&& src) U_NOEXCEPT {
    numrange_fluent_NFS<numrange_fluent_LNF>::operator=(static_cast<numrange_fluent_NFS<numrange_fluent_LNF>&&>(src));
    // Steal the compiled formatter
    auto* stolen = src.fAtomicFormatter.exchange(nullptr);
    delete fAtomicFormatter.exchange(stolen);
    return *this;
}


LocalizedNumberRangeFormatter::~LocalizedNumberRangeFormatter() {
    delete fAtomicFormatter.exchange(nullptr);
}

LocalizedNumberRangeFormatter::LocalizedNumberRangeFormatter(const RangeMacroProps& macros, const Locale& locale) {
    fMacros = macros;
    fMacros.locale = locale;
    touchRangeLocales(fMacros);
}

LocalizedNumberRangeFormatter::LocalizedNumberRangeFormatter(RangeMacroProps&& macros, const Locale& locale) {
    fMacros = std::move(macros);
    fMacros.locale = locale;
    touchRangeLocales(fMacros);
}

LocalizedNumberRangeFormatter UnlocalizedNumberRangeFormatter::locale(const Locale& locale) const& {
    return LocalizedNumberRangeFormatter(fMacros, locale);
}

LocalizedNumberRangeFormatter UnlocalizedNumberRangeFormatter::locale(const Locale& locale)&& {
    return LocalizedNumberRangeFormatter(std::move(fMacros), locale);
}


FormattedNumberRange LocalizedNumberRangeFormatter::formatFormattableRange(
        const Formattable& first, const Formattable& second, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return FormattedNumberRange(U_ILLEGAL_ARGUMENT_ERROR);
    }

    auto results = new UFormattedNumberRangeData();
    if (results == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return FormattedNumberRange(status);
    }

    first.populateDecimalQuantity(results->quantity1, status);
    if (U_FAILURE(status)) {
        return FormattedNumberRange(status);
    }

    second.populateDecimalQuantity(results->quantity2, status);
    if (U_FAILURE(status)) {
        return FormattedNumberRange(status);
    }

    formatImpl(*results, first == second, status);

    // Do not save the results object if we encountered a failure.
    if (U_SUCCESS(status)) {
        return FormattedNumberRange(results);
    } else {
        delete results;
        return FormattedNumberRange(status);
    }
}

void LocalizedNumberRangeFormatter::formatImpl(
        UFormattedNumberRangeData& results, bool equalBeforeRounding, UErrorCode& status) const {
    auto* impl = getFormatter(status);
    if (U_FAILURE(status)) {
        return;
    }
    if (impl == nullptr) {
        status = U_INTERNAL_PROGRAM_ERROR;
        return;
    }
    impl->format(results, equalBeforeRounding, status);
    if (U_FAILURE(status)) {
        return;
    }
    results.getStringRef().writeTerminator(status);
}

const icu::number::impl::NumberRangeFormatterImpl*
LocalizedNumberRangeFormatter::getFormatter(UErrorCode& status) const {
    // TODO: Move this into umutex.h? (similar logic also in decimfmt.cpp)
    // See ICU-20146

    if (U_FAILURE(status)) {
        return nullptr;
    }

    // First try to get the pre-computed formatter
    auto* ptr = fAtomicFormatter.load();
    if (ptr != nullptr) {
        return ptr;
    }

    // Try computing the formatter on our own
    auto* temp = new NumberRangeFormatterImpl(fMacros, status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (temp == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }

    // Note: ptr starts as nullptr; during compare_exchange,
    // it is set to what is actually stored in the atomic
    // if another thread beat us to computing the formatter object.
    auto* nonConstThis = const_cast<LocalizedNumberRangeFormatter*>(this);
    if (!nonConstThis->fAtomicFormatter.compare_exchange_strong(ptr, temp)) {
        // Another thread beat us to computing the formatter
        delete temp;
        return ptr;
    } else {
        // Our copy of the formatter got stored in the atomic
        return temp;
    }

}


UPRV_FORMATTED_VALUE_SUBCLASS_AUTO_IMPL(FormattedNumberRange)

#define UPRV_NOARG

UBool FormattedNumberRange::nextFieldPosition(FieldPosition& fieldPosition, UErrorCode& status) const {
    UPRV_FORMATTED_VALUE_METHOD_GUARD(FALSE)
    // NOTE: MSVC sometimes complains when implicitly converting between bool and UBool
    return fData->nextFieldPosition(fieldPosition, status);
}

void FormattedNumberRange::getAllFieldPositions(FieldPositionIterator& iterator, UErrorCode& status) const {
    FieldPositionIteratorHandler fpih(&iterator, status);
    getAllFieldPositionsImpl(fpih, status);
}

void FormattedNumberRange::getAllFieldPositionsImpl(
        FieldPositionIteratorHandler& fpih, UErrorCode& status) const {
    UPRV_FORMATTED_VALUE_METHOD_GUARD(UPRV_NOARG)
    fData->getAllFieldPositions(fpih, status);
}

UnicodeString FormattedNumberRange::getFirstDecimal(UErrorCode& status) const {
    UPRV_FORMATTED_VALUE_METHOD_GUARD(ICU_Utility::makeBogusString())
    return fData->quantity1.toScientificString();
}

UnicodeString FormattedNumberRange::getSecondDecimal(UErrorCode& status) const {
    UPRV_FORMATTED_VALUE_METHOD_GUARD(ICU_Utility::makeBogusString())
    return fData->quantity2.toScientificString();
}

UNumberRangeIdentityResult FormattedNumberRange::getIdentityResult(UErrorCode& status) const {
    UPRV_FORMATTED_VALUE_METHOD_GUARD(UNUM_IDENTITY_RESULT_NOT_EQUAL)
    return fData->identityResult;
}


UFormattedNumberRangeData::~UFormattedNumberRangeData() = default;



#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif







using namespace icu;
using namespace icu::number;
using namespace icu::number::impl;

namespace {

// Helper function for 2-dimensional switch statement
constexpr int8_t identity2d(UNumberRangeIdentityFallback a, UNumberRangeIdentityResult b) {
    return static_cast<int8_t>(a) | (static_cast<int8_t>(b) << 4);
}


struct NumberRangeData {
    SimpleFormatter rangePattern;
    SimpleFormatter approximatelyPattern;
};

class NumberRangeDataSink : public ResourceSink {
  public:
    NumberRangeDataSink(NumberRangeData& data) : fData(data) {}

    void put(const char* key, ResourceValue& value, UBool /*noFallback*/, UErrorCode& status) U_OVERRIDE {
        ResourceTable miscTable = value.getTable(status);
        if (U_FAILURE(status)) { return; }
        for (int i = 0; miscTable.getKeyAndValue(i, key, value); i++) {
            if (uprv_strcmp(key, "range") == 0) {
                if (hasRangeData()) {
                    continue; // have already seen this pattern
                }
                fData.rangePattern = {value.getUnicodeString(status), status};
            } else if (uprv_strcmp(key, "approximately") == 0) {
                if (hasApproxData()) {
                    continue; // have already seen this pattern
                }
                fData.approximatelyPattern = {value.getUnicodeString(status), status};
            }
        }
    }

    bool hasRangeData() {
        return fData.rangePattern.getArgumentLimit() != 0;
    }

    bool hasApproxData() {
        return fData.approximatelyPattern.getArgumentLimit() != 0;
    }

    bool isComplete() {
        return hasRangeData() && hasApproxData();
    }

    void fillInDefaults(UErrorCode& status) {
        if (!hasRangeData()) {
            fData.rangePattern = {u"{0}–{1}", status};
        }
        if (!hasApproxData()) {
            fData.approximatelyPattern = {u"~{0}", status};
        }
    }

  private:
    NumberRangeData& fData;
};

void getNumberRangeData(const char* localeName, const char* nsName, NumberRangeData& data, UErrorCode& status) {
    if (U_FAILURE(status)) { return; }
    LocalUResourceBundlePointer rb(ures_open(NULL, localeName, &status));
    if (U_FAILURE(status)) { return; }
    NumberRangeDataSink sink(data);

    CharString dataPath;
    dataPath.append("NumberElements/", -1, status);
    dataPath.append(nsName, -1, status);
    dataPath.append("/miscPatterns", -1, status);
    if (U_FAILURE(status)) { return; }

    UErrorCode localStatus = U_ZERO_ERROR;
    ures_getAllItemsWithFallback(rb.getAlias(), dataPath.data(), sink, localStatus);
    if (U_FAILURE(localStatus) && localStatus != U_MISSING_RESOURCE_ERROR) {
        status = localStatus;
        return;
    }

    // Fall back to latn if necessary
    if (!sink.isComplete()) {
        ures_getAllItemsWithFallback(rb.getAlias(), "NumberElements/latn/miscPatterns", sink, status);
    }

    sink.fillInDefaults(status);
}

class PluralRangesDataSink : public ResourceSink {
  public:
    PluralRangesDataSink(StandardPluralRanges& output) : fOutput(output) {}

    void put(const char* /*key*/, ResourceValue& value, UBool /*noFallback*/, UErrorCode& status) U_OVERRIDE {
        ResourceArray entriesArray = value.getArray(status);
        if (U_FAILURE(status)) { return; }
        fOutput.setCapacity(entriesArray.getSize());
        for (int i = 0; entriesArray.getValue(i, value); i++) {
            ResourceArray pluralFormsArray = value.getArray(status);
            if (U_FAILURE(status)) { return; }
            pluralFormsArray.getValue(0, value);
            StandardPlural::Form first = StandardPlural::fromString(value.getUnicodeString(status), status);
            if (U_FAILURE(status)) { return; }
            pluralFormsArray.getValue(1, value);
            StandardPlural::Form second = StandardPlural::fromString(value.getUnicodeString(status), status);
            if (U_FAILURE(status)) { return; }
            pluralFormsArray.getValue(2, value);
            StandardPlural::Form result = StandardPlural::fromString(value.getUnicodeString(status), status);
            if (U_FAILURE(status)) { return; }
            fOutput.addPluralRange(first, second, result);
        }
    }

  private:
    StandardPluralRanges& fOutput;
};

void getPluralRangesData(const Locale& locale, StandardPluralRanges& output, UErrorCode& status) {
    if (U_FAILURE(status)) { return; }
    LocalUResourceBundlePointer rb(ures_openDirect(nullptr, "pluralRanges", &status));
    if (U_FAILURE(status)) { return; }

    CharString dataPath;
    dataPath.append("locales/", -1, status);
    dataPath.append(locale.getLanguage(), -1, status);
    if (U_FAILURE(status)) { return; }
    int32_t setLen;
    // Not all languages are covered: fail gracefully
    UErrorCode internalStatus = U_ZERO_ERROR;
    const UChar* set = ures_getStringByKeyWithFallback(rb.getAlias(), dataPath.data(), &setLen, &internalStatus);
    if (U_FAILURE(internalStatus)) { return; }

    dataPath.clear();
    dataPath.append("rules/", -1, status);
    dataPath.appendInvariantChars(set, setLen, status);
    if (U_FAILURE(status)) { return; }
    PluralRangesDataSink sink(output);
    ures_getAllItemsWithFallback(rb.getAlias(), dataPath.data(), sink, status);
    if (U_FAILURE(status)) { return; }
}

} // namespace


void StandardPluralRanges::initialize(const Locale& locale, UErrorCode& status) {
    getPluralRangesData(locale, *this, status);
}

void StandardPluralRanges::addPluralRange(
        StandardPlural::Form first,
        StandardPlural::Form second,
        StandardPlural::Form result) {
    U_ASSERT(fTriplesLen < fTriples.getCapacity());
    fTriples[fTriplesLen] = {first, second, result};
    fTriplesLen++;
}

void StandardPluralRanges::setCapacity(int32_t length) {
    if (length > fTriples.getCapacity()) {
        fTriples.resize(length, 0);
    }
}

StandardPlural::Form
StandardPluralRanges::resolve(StandardPlural::Form first, StandardPlural::Form second) const {
    for (int32_t i=0; i<fTriplesLen; i++) {
        const auto& triple = fTriples[i];
        if (triple.first == first && triple.second == second) {
            return triple.result;
        }
    }
    // Default fallback
    return StandardPlural::OTHER;
}


NumberRangeFormatterImpl::NumberRangeFormatterImpl(const RangeMacroProps& macros, UErrorCode& status)
    : formatterImpl1(macros.formatter1.fMacros, status),
      formatterImpl2(macros.formatter2.fMacros, status),
      fSameFormatters(macros.singleFormatter),
      fCollapse(macros.collapse),
      fIdentityFallback(macros.identityFallback) {

    const char* nsName = formatterImpl1.getRawMicroProps().nsName;
    if (uprv_strcmp(nsName, formatterImpl2.getRawMicroProps().nsName) != 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    NumberRangeData data;
    getNumberRangeData(macros.locale.getName(), nsName, data, status);
    if (U_FAILURE(status)) { return; }
    fRangeFormatter = data.rangePattern;
    fApproximatelyModifier = {data.approximatelyPattern, UNUM_FIELD_COUNT, false};

    // TODO: Get locale from PluralRules instead?
    fPluralRanges.initialize(macros.locale, status);
    if (U_FAILURE(status)) { return; }
}

void NumberRangeFormatterImpl::format(UFormattedNumberRangeData& data, bool equalBeforeRounding, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }

    MicroProps micros1;
    MicroProps micros2;
    formatterImpl1.preProcess(data.quantity1, micros1, status);
    if (fSameFormatters) {
        formatterImpl1.preProcess(data.quantity2, micros2, status);
    } else {
        formatterImpl2.preProcess(data.quantity2, micros2, status);
    }
    if (U_FAILURE(status)) {
        return;
    }

    // If any of the affixes are different, an identity is not possible
    // and we must use formatRange().
    // TODO: Write this as MicroProps operator==() ?
    // TODO: Avoid the redundancy of these equality operations with the
    // ones in formatRange?
    if (!micros1.modInner->semanticallyEquivalent(*micros2.modInner)
            || !micros1.modMiddle->semanticallyEquivalent(*micros2.modMiddle)
            || !micros1.modOuter->semanticallyEquivalent(*micros2.modOuter)) {
        formatRange(data, micros1, micros2, status);
        data.identityResult = UNUM_IDENTITY_RESULT_NOT_EQUAL;
        return;
    }

    // Check for identity
    if (equalBeforeRounding) {
        data.identityResult = UNUM_IDENTITY_RESULT_EQUAL_BEFORE_ROUNDING;
    } else if (data.quantity1 == data.quantity2) {
        data.identityResult = UNUM_IDENTITY_RESULT_EQUAL_AFTER_ROUNDING;
    } else {
        data.identityResult = UNUM_IDENTITY_RESULT_NOT_EQUAL;
    }

    switch (identity2d(fIdentityFallback, data.identityResult)) {
        case identity2d(UNUM_IDENTITY_FALLBACK_RANGE,
                        UNUM_IDENTITY_RESULT_NOT_EQUAL):
        case identity2d(UNUM_IDENTITY_FALLBACK_RANGE,
                        UNUM_IDENTITY_RESULT_EQUAL_AFTER_ROUNDING):
        case identity2d(UNUM_IDENTITY_FALLBACK_RANGE,
                        UNUM_IDENTITY_RESULT_EQUAL_BEFORE_ROUNDING):
        case identity2d(UNUM_IDENTITY_FALLBACK_APPROXIMATELY,
                        UNUM_IDENTITY_RESULT_NOT_EQUAL):
        case identity2d(UNUM_IDENTITY_FALLBACK_APPROXIMATELY_OR_SINGLE_VALUE,
                        UNUM_IDENTITY_RESULT_NOT_EQUAL):
        case identity2d(UNUM_IDENTITY_FALLBACK_SINGLE_VALUE,
                        UNUM_IDENTITY_RESULT_NOT_EQUAL):
            formatRange(data, micros1, micros2, status);
            break;

        case identity2d(UNUM_IDENTITY_FALLBACK_APPROXIMATELY,
                        UNUM_IDENTITY_RESULT_EQUAL_AFTER_ROUNDING):
        case identity2d(UNUM_IDENTITY_FALLBACK_APPROXIMATELY,
                        UNUM_IDENTITY_RESULT_EQUAL_BEFORE_ROUNDING):
        case identity2d(UNUM_IDENTITY_FALLBACK_APPROXIMATELY_OR_SINGLE_VALUE,
                        UNUM_IDENTITY_RESULT_EQUAL_AFTER_ROUNDING):
            formatApproximately(data, micros1, micros2, status);
            break;

        case identity2d(UNUM_IDENTITY_FALLBACK_APPROXIMATELY_OR_SINGLE_VALUE,
                        UNUM_IDENTITY_RESULT_EQUAL_BEFORE_ROUNDING):
        case identity2d(UNUM_IDENTITY_FALLBACK_SINGLE_VALUE,
                        UNUM_IDENTITY_RESULT_EQUAL_AFTER_ROUNDING):
        case identity2d(UNUM_IDENTITY_FALLBACK_SINGLE_VALUE,
                        UNUM_IDENTITY_RESULT_EQUAL_BEFORE_ROUNDING):
            formatSingleValue(data, micros1, micros2, status);
            break;

        default:
            UPRV_UNREACHABLE;
    }
}


void NumberRangeFormatterImpl::formatSingleValue(UFormattedNumberRangeData& data,
                                                 MicroProps& micros1, MicroProps& micros2,
                                                 UErrorCode& status) const {
    if (U_FAILURE(status)) { return; }
    if (fSameFormatters) {
        int32_t length = NumberFormatterImpl::writeNumber(micros1, data.quantity1, data.getStringRef(), 0, status);
        NumberFormatterImpl::writeAffixes(micros1, data.getStringRef(), 0, length, status);
    } else {
        formatRange(data, micros1, micros2, status);
    }
}


void NumberRangeFormatterImpl::formatApproximately (UFormattedNumberRangeData& data,
                                                    MicroProps& micros1, MicroProps& micros2,
                                                    UErrorCode& status) const {
    if (U_FAILURE(status)) { return; }
    if (fSameFormatters) {
        int32_t length = NumberFormatterImpl::writeNumber(micros1, data.quantity1, data.getStringRef(), 0, status);
        // HEURISTIC: Desired modifier order: inner, middle, approximately, outer.
        length += micros1.modInner->apply(data.getStringRef(), 0, length, status);
        length += micros1.modMiddle->apply(data.getStringRef(), 0, length, status);
        length += fApproximatelyModifier.apply(data.getStringRef(), 0, length, status);
        micros1.modOuter->apply(data.getStringRef(), 0, length, status);
    } else {
        formatRange(data, micros1, micros2, status);
    }
}


void NumberRangeFormatterImpl::formatRange(UFormattedNumberRangeData& data,
                                           MicroProps& micros1, MicroProps& micros2,
                                           UErrorCode& status) const {
    if (U_FAILURE(status)) { return; }

    // modInner is always notation (scientific); collapsable in ALL.
    // modOuter is always units; collapsable in ALL, AUTO, and UNIT.
    // modMiddle could be either; collapsable in ALL and sometimes AUTO and UNIT.
    // Never collapse an outer mod but not an inner mod.
    bool collapseOuter, collapseMiddle, collapseInner;
    switch (fCollapse) {
        case UNUM_RANGE_COLLAPSE_ALL:
        case UNUM_RANGE_COLLAPSE_AUTO:
        case UNUM_RANGE_COLLAPSE_UNIT:
        {
            // OUTER MODIFIER
            collapseOuter = micros1.modOuter->semanticallyEquivalent(*micros2.modOuter);

            if (!collapseOuter) {
                // Never collapse inner mods if outer mods are not collapsable
                collapseMiddle = false;
                collapseInner = false;
                break;
            }

            // MIDDLE MODIFIER
            collapseMiddle = micros1.modMiddle->semanticallyEquivalent(*micros2.modMiddle);

            if (!collapseMiddle) {
                // Never collapse inner mods if outer mods are not collapsable
                collapseInner = false;
                break;
            }

            // MIDDLE MODIFIER HEURISTICS
            // (could disable collapsing of the middle modifier)
            // The modifiers are equal by this point, so we can look at just one of them.
            const Modifier* mm = micros1.modMiddle;
            if (fCollapse == UNUM_RANGE_COLLAPSE_UNIT) {
                // Only collapse if the modifier is a unit.
                // TODO: Make a better way to check for a unit?
                // TODO: Handle case where the modifier has both notation and unit (compact currency)?
                if (!mm->containsField(UNUM_CURRENCY_FIELD) && !mm->containsField(UNUM_PERCENT_FIELD)) {
                    collapseMiddle = false;
                }
            } else if (fCollapse == UNUM_RANGE_COLLAPSE_AUTO) {
                // Heuristic as of ICU 63: collapse only if the modifier is more than one code point.
                if (mm->getCodePointCount() <= 1) {
                    collapseMiddle = false;
                }
            }

            if (!collapseMiddle || fCollapse != UNUM_RANGE_COLLAPSE_ALL) {
                collapseInner = false;
                break;
            }

            // INNER MODIFIER
            collapseInner = micros1.modInner->semanticallyEquivalent(*micros2.modInner);

            // All done checking for collapsability.
            break;
        }

        default:
            collapseOuter = false;
            collapseMiddle = false;
            collapseInner = false;
            break;
    }

    FormattedStringBuilder& string = data.getStringRef();
    int32_t lengthPrefix = 0;
    int32_t length1 = 0;
    int32_t lengthInfix = 0;
    int32_t length2 = 0;
    int32_t lengthSuffix = 0;

    // Use #define so that these are evaluated at the call site.
    #define UPRV_INDEX_0 (lengthPrefix)
    #define UPRV_INDEX_1 (lengthPrefix + length1)
    #define UPRV_INDEX_2 (lengthPrefix + length1 + lengthInfix)
    #define UPRV_INDEX_3 (lengthPrefix + length1 + lengthInfix + length2)

    int32_t lengthRange = SimpleModifier::formatTwoArgPattern(
        fRangeFormatter,
        string,
        0,
        &lengthPrefix,
        &lengthSuffix,
        UNUM_FIELD_COUNT,
        status);
    if (U_FAILURE(status)) { return; }
    lengthInfix = lengthRange - lengthPrefix - lengthSuffix;
    U_ASSERT(lengthInfix > 0);

    // SPACING HEURISTIC
    // Add spacing unless all modifiers are collapsed.
    // TODO: add API to control this?
    // TODO: Use a data-driven heuristic like currency spacing?
    // TODO: Use Unicode [:whitespace:] instead of PatternProps whitespace? (consider speed implications)
    {
        bool repeatInner = !collapseInner && micros1.modInner->getCodePointCount() > 0;
        bool repeatMiddle = !collapseMiddle && micros1.modMiddle->getCodePointCount() > 0;
        bool repeatOuter = !collapseOuter && micros1.modOuter->getCodePointCount() > 0;
        if (repeatInner || repeatMiddle || repeatOuter) {
            // Add spacing if there is not already spacing
            if (!PatternProps::isWhiteSpace(string.charAt(UPRV_INDEX_1))) {
                lengthInfix += string.insertCodePoint(UPRV_INDEX_1, u'\u0020', UNUM_FIELD_COUNT, status);
            }
            if (!PatternProps::isWhiteSpace(string.charAt(UPRV_INDEX_2 - 1))) {
                lengthInfix += string.insertCodePoint(UPRV_INDEX_2, u'\u0020', UNUM_FIELD_COUNT, status);
            }
        }
    }

    length1 += NumberFormatterImpl::writeNumber(micros1, data.quantity1, string, UPRV_INDEX_0, status);
    length2 += NumberFormatterImpl::writeNumber(micros2, data.quantity2, string, UPRV_INDEX_2, status);

    // TODO: Support padding?

    if (collapseInner) {
        // Note: this is actually a mix of prefix and suffix, but adding to infix length works
        const Modifier& mod = resolveModifierPlurals(*micros1.modInner, *micros2.modInner);
        lengthInfix += mod.apply(string, UPRV_INDEX_0, UPRV_INDEX_3, status);
    } else {
        length1 += micros1.modInner->apply(string, UPRV_INDEX_0, UPRV_INDEX_1, status);
        length2 += micros2.modInner->apply(string, UPRV_INDEX_2, UPRV_INDEX_3, status);
    }

    if (collapseMiddle) {
        // Note: this is actually a mix of prefix and suffix, but adding to infix length works
        const Modifier& mod = resolveModifierPlurals(*micros1.modMiddle, *micros2.modMiddle);
        lengthInfix += mod.apply(string, UPRV_INDEX_0, UPRV_INDEX_3, status);
    } else {
        length1 += micros1.modMiddle->apply(string, UPRV_INDEX_0, UPRV_INDEX_1, status);
        length2 += micros2.modMiddle->apply(string, UPRV_INDEX_2, UPRV_INDEX_3, status);
    }

    if (collapseOuter) {
        // Note: this is actually a mix of prefix and suffix, but adding to infix length works
        const Modifier& mod = resolveModifierPlurals(*micros1.modOuter, *micros2.modOuter);
        lengthInfix += mod.apply(string, UPRV_INDEX_0, UPRV_INDEX_3, status);
    } else {
        length1 += micros1.modOuter->apply(string, UPRV_INDEX_0, UPRV_INDEX_1, status);
        length2 += micros2.modOuter->apply(string, UPRV_INDEX_2, UPRV_INDEX_3, status);
    }
}


const Modifier&
NumberRangeFormatterImpl::resolveModifierPlurals(const Modifier& first, const Modifier& second) const {
    Modifier::Parameters parameters;
    first.getParameters(parameters);
    if (parameters.obj == nullptr) {
        // No plural form; return a fallback (e.g., the first)
        return first;
    }
    StandardPlural::Form firstPlural = parameters.plural;

    second.getParameters(parameters);
    if (parameters.obj == nullptr) {
        // No plural form; return a fallback (e.g., the first)
        return first;
    }
    StandardPlural::Form secondPlural = parameters.plural;

    // Get the required plural form from data
    StandardPlural::Form resultPlural = fPluralRanges.resolve(firstPlural, secondPlural);

    // Get and return the new Modifier
    const Modifier* mod = parameters.obj->getModifier(parameters.signum, resultPlural);
    U_ASSERT(mod != nullptr);
    return *mod;
}



#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2015, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
*
* File NUMSYS.CPP
*
* Modification History:*
*   Date        Name        Description
*
********************************************************************************
*/















// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2015, International Business Machines Corporation and
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* File NUMSYS_IMPL.H
*
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING
    




U_NAMESPACE_BEGIN

class NumsysNameEnumeration : public StringEnumeration {
public:
    NumsysNameEnumeration(UErrorCode& status);

    virtual ~NumsysNameEnumeration();
    static UClassID U_EXPORT2 getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const override;
    virtual const UnicodeString* snext(UErrorCode& status) override;
    virtual void reset(UErrorCode& status) override;
    virtual int32_t count(UErrorCode& status) const override;
private:
    int32_t pos;
};

U_NAMESPACE_END

#endif



#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

// Useful constants

#define DEFAULT_DIGITS UNICODE_STRING_SIMPLE("0123456789")
static const char numsys_gNumberingSystems[] = "numberingSystems";
static const char numsys_gNumberElements[] = "NumberElements";
static const char numsys_gDefault[] = "default";
static const char numsys_gNative[] = "native";
static const char numsys_gTraditional[] = "traditional";
static const char numsys_gFinance[] = "finance";
static const char numsys_gDesc[] = "desc";
static const char numsys_gRadix[] = "radix";
static const char numsys_gAlgorithmic[] = "algorithmic";
static const char numsys_gLatn[] = "latn";


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(NumberingSystem)
UOBJECT_DEFINE_RTTI_IMPLEMENTATION(NumsysNameEnumeration)

    /**
     * Default Constructor.
     *
     * @draft ICU 4.2
     */

NumberingSystem::NumberingSystem() {
     radix = 10;
     algorithmic = FALSE;
     UnicodeString defaultDigits = DEFAULT_DIGITS;
     desc.setTo(defaultDigits);
     uprv_strcpy(name,numsys_gLatn);
}

    /**
     * Copy constructor.
     * @draft ICU 4.2
     */

NumberingSystem::NumberingSystem(const NumberingSystem& other)
:  UObject(other) {
    *this=other;
}

NumberingSystem* U_EXPORT2
NumberingSystem::createInstance(int32_t radix_in, UBool isAlgorithmic_in, const UnicodeString & desc_in, UErrorCode &status) {

    if (U_FAILURE(status)) {
        return nullptr;
    }

    if ( radix_in < 2 ) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }

    if ( !isAlgorithmic_in ) {
       if ( desc_in.countChar32() != radix_in ) {
           status = U_ILLEGAL_ARGUMENT_ERROR;
           return nullptr;
       }
    }

    LocalPointer<NumberingSystem> ns(new NumberingSystem(), status);
    if (U_FAILURE(status)) {
        return nullptr;
    }

    ns->setRadix(radix_in);
    ns->setDesc(desc_in);
    ns->setAlgorithmic(isAlgorithmic_in);
    ns->setName(nullptr);

    return ns.orphan();
}

NumberingSystem* U_EXPORT2
NumberingSystem::createInstance(const Locale & inLocale, UErrorCode& status) {

    if (U_FAILURE(status)) {
        return nullptr;
    }

    UBool nsResolved = TRUE;
    UBool usingFallback = FALSE;
    char buffer[ULOC_KEYWORDS_CAPACITY];
    int32_t count = inLocale.getKeywordValue("numbers", buffer, sizeof(buffer), status);
    if (U_FAILURE(status) || status == U_STRING_NOT_TERMINATED_WARNING) {
        // the "numbers" keyword exceeds ULOC_KEYWORDS_CAPACITY; ignore and use default.
        count = 0;
        status = U_ZERO_ERROR;
    }
    if ( count > 0 ) { // @numbers keyword was specified in the locale
        U_ASSERT(count < ULOC_KEYWORDS_CAPACITY);
        buffer[count] = '\0'; // Make sure it is null terminated.
        if ( !uprv_strcmp(buffer,numsys_gDefault) || !uprv_strcmp(buffer,numsys_gNative) ||
             !uprv_strcmp(buffer,numsys_gTraditional) || !uprv_strcmp(buffer,numsys_gFinance)) {
            nsResolved = FALSE;
        }
    } else {
        uprv_strcpy(buffer, numsys_gDefault);
        nsResolved = FALSE;
    }

    if (!nsResolved) { // Resolve the numbering system ( default, native, traditional or finance ) into a "real" numbering system
        UErrorCode localStatus = U_ZERO_ERROR;
        LocalUResourceBundlePointer resource(ures_open(nullptr, inLocale.getName(), &localStatus));
        LocalUResourceBundlePointer numberElementsRes(ures_getByKey(resource.getAlias(), numsys_gNumberElements, nullptr, &localStatus));
        // Don't stomp on the catastrophic failure of OOM.
        if (localStatus == U_MEMORY_ALLOCATION_ERROR) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return nullptr;
        }
        while (!nsResolved) {
            localStatus = U_ZERO_ERROR;
            count = 0;
            const UChar *nsName = ures_getStringByKeyWithFallback(numberElementsRes.getAlias(), buffer, &count, &localStatus);
            // Don't stomp on the catastrophic failure of OOM.
            if (localStatus == U_MEMORY_ALLOCATION_ERROR) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return nullptr;
            }
            if ( count > 0 && count < ULOC_KEYWORDS_CAPACITY ) { // numbering system found
                u_UCharsToChars(nsName, buffer, count);
                buffer[count] = '\0'; // Make sure it is null terminated.
                nsResolved = TRUE;
            }

            if (!nsResolved) { // Fallback behavior per TR35 - traditional falls back to native, finance and native fall back to default
                if (!uprv_strcmp(buffer,numsys_gNative) || !uprv_strcmp(buffer,numsys_gFinance)) {
                    uprv_strcpy(buffer,numsys_gDefault);
                } else if (!uprv_strcmp(buffer,numsys_gTraditional)) {
                    uprv_strcpy(buffer,numsys_gNative);
                } else { // If we get here we couldn't find even the default numbering system
                    usingFallback = TRUE;
                    nsResolved = TRUE;
                }
            }
        }
    }

    if (usingFallback) {
        status = U_USING_FALLBACK_WARNING;
        NumberingSystem *ns = new NumberingSystem();
        if (ns == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
        }
        return ns;
    } else {
        return NumberingSystem::createInstanceByName(buffer, status);
    }
 }

NumberingSystem* U_EXPORT2
NumberingSystem::createInstance(UErrorCode& status) {
    return NumberingSystem::createInstance(Locale::getDefault(), status);
}

NumberingSystem* U_EXPORT2
NumberingSystem::createInstanceByName(const char *name, UErrorCode& status) {
    int32_t radix = 10;
    int32_t algorithmic = 0;

    LocalUResourceBundlePointer numberingSystemsInfo(ures_openDirect(nullptr, numsys_gNumberingSystems, &status));
    LocalUResourceBundlePointer nsCurrent(ures_getByKey(numberingSystemsInfo.getAlias(), numsys_gNumberingSystems, nullptr, &status));
    LocalUResourceBundlePointer nsTop(ures_getByKey(nsCurrent.getAlias(), name, nullptr, &status));

    UnicodeString nsd = ures_getUnicodeStringByKey(nsTop.getAlias(), numsys_gDesc, &status);

    ures_getByKey(nsTop.getAlias(), numsys_gRadix, nsCurrent.getAlias(), &status);
    radix = ures_getInt(nsCurrent.getAlias(), &status);

    ures_getByKey(nsTop.getAlias(), numsys_gAlgorithmic, nsCurrent.getAlias(), &status);
    algorithmic = ures_getInt(nsCurrent.getAlias(), &status);

    UBool isAlgorithmic = ( algorithmic == 1 );

    if (U_FAILURE(status)) {
        // Don't stomp on the catastrophic failure of OOM.
        if (status != U_MEMORY_ALLOCATION_ERROR) {
            status = U_UNSUPPORTED_ERROR;
        }
        return nullptr;
    }

    LocalPointer<NumberingSystem> ns(NumberingSystem::createInstance(radix, isAlgorithmic, nsd, status), status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    ns->setName(name);
    return ns.orphan();
}

    /**
     * Destructor.
     * @draft ICU 4.2
     */
NumberingSystem::~NumberingSystem() {
}

int32_t NumberingSystem::getRadix() const {
    return radix;
}

UnicodeString NumberingSystem::getDescription() const {
    return desc;
}

const char * NumberingSystem::getName() const {
    return name;
}

void NumberingSystem::setRadix(int32_t r) {
    radix = r;
}

void NumberingSystem::setAlgorithmic(UBool c) {
    algorithmic = c;
}

void NumberingSystem::setDesc(const UnicodeString &d) {
    desc.setTo(d);
}
void NumberingSystem::setName(const char *n) {
    if ( n == nullptr ) {
        name[0] = (char) 0;
    } else {
        uprv_strncpy(name,n,kInternalNumSysNameCapacity);
        name[kInternalNumSysNameCapacity] = '\0'; // Make sure it is null terminated.
    }
}
UBool NumberingSystem::isAlgorithmic() const {
    return ( algorithmic );
}

namespace {

UVector* gNumsysNames = nullptr;
UInitOnce gNumSysInitOnce = U_INITONCE_INITIALIZER;

U_CFUNC UBool U_CALLCONV numSysCleanup() {
    delete gNumsysNames;
    gNumsysNames = nullptr;
    gNumSysInitOnce.reset();
    return true;
}

U_CFUNC void initNumsysNames(UErrorCode &status) {
    U_ASSERT(gNumsysNames == nullptr);
    ucln_i18n_registerCleanup(UCLN_I18N_NUMSYS, numSysCleanup);

    // TODO: Simple array of UnicodeString objects, based on length of table resource?
    LocalPointer<UVector> numsysNames(new UVector(uprv_deleteUObject, nullptr, status), status);
    if (U_FAILURE(status)) {
        return;
    }

    UErrorCode rbstatus = U_ZERO_ERROR;
    UResourceBundle *numberingSystemsInfo = ures_openDirect(nullptr, "numberingSystems", &rbstatus);
    numberingSystemsInfo =
            ures_getByKey(numberingSystemsInfo, "numberingSystems", numberingSystemsInfo, &rbstatus);
    if (U_FAILURE(rbstatus)) {
        // Don't stomp on the catastrophic failure of OOM.
        if (rbstatus == U_MEMORY_ALLOCATION_ERROR) {
            status = rbstatus;
        } else {
            status = U_MISSING_RESOURCE_ERROR;
        }
        ures_close(numberingSystemsInfo);
        return;
    }

    while ( ures_hasNext(numberingSystemsInfo) && U_SUCCESS(status) ) {
        LocalUResourceBundlePointer nsCurrent(ures_getNextResource(numberingSystemsInfo, nullptr, &rbstatus));
        if (rbstatus == U_MEMORY_ALLOCATION_ERROR) {
            status = rbstatus; // we want to report OOM failure back to the caller.
            break;
        }
        const char *nsName = ures_getKey(nsCurrent.getAlias());
        LocalPointer<UnicodeString> newElem(new UnicodeString(nsName, -1, US_INV), status);
        if (U_SUCCESS(status)) {
            numsysNames->addElement(newElem.getAlias(), status);
            if (U_SUCCESS(status)) {
                newElem.orphan(); // on success, the numsysNames vector owns newElem.
            }
        }
    }

    ures_close(numberingSystemsInfo);
    if (U_SUCCESS(status)) {
        gNumsysNames = numsysNames.orphan();
    }
    return;
}

}   // end anonymous namespace

StringEnumeration* NumberingSystem::getAvailableNames(UErrorCode &status) {
    umtx_initOnce(gNumSysInitOnce, &initNumsysNames, status);
    LocalPointer<StringEnumeration> result(new NumsysNameEnumeration(status), status);
    return result.orphan();
}

NumsysNameEnumeration::NumsysNameEnumeration(UErrorCode& status) : pos(0) {
    (void)status;
}

const UnicodeString*
NumsysNameEnumeration::snext(UErrorCode& status) {
    if (U_SUCCESS(status) && (gNumsysNames != nullptr) && (pos < gNumsysNames->size())) {
        return (const UnicodeString*)gNumsysNames->elementAt(pos++);
    }
    return nullptr;
}

void
NumsysNameEnumeration::reset(UErrorCode& /*status*/) {
    pos=0;
}

int32_t
NumsysNameEnumeration::count(UErrorCode& /*status*/) const {
    return (gNumsysNames==nullptr) ? 0 : gNumsysNames->size();
}

NumsysNameEnumeration::~NumsysNameEnumeration() {
}
U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2003-2013, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
* Author: Alan Liu
* Created: July 21 2003
* Since: ICU 2.8
**********************************************************************
*/

  // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING








#include <float.h> // DBL_MAX

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2013, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

typedef struct OlsonToMetaMappingEntry {
    const UChar *mzid; // const because it's a reference to a resource bundle string.
    UDate from;
    UDate to;
} OlsonToMetaMappingEntry;

class UVector;
class TimeZone;

class U_I18N_API ZoneMeta {
public:
    /**
     * Return the canonical id for this tzid defined by CLDR, which might be the id itself.
     * If the given system tzid is not known, U_ILLEGAL_ARGUMENT_ERROR is set in the status.
     *
     * Note: this internal API supports all known system IDs and "Etc/Unknown" (which is
     * NOT a system ID).
     */
    static UnicodeString& U_EXPORT2 getCanonicalCLDRID(const UnicodeString &tzid, UnicodeString &systemID, UErrorCode& status);

    /**
     * Return the canonical id for this tzid defined by CLDR, which might be the id itself.
     * This overload method returns a persistent const UChar*, which is guranteed to persist
     * (a pointer to a resource). If the given system tzid is not known, U_ILLEGAL_ARGUMENT_ERROR
     * is set in the status.
     * @param tzid Zone ID
     * @param status Receives the status
     * @return The canonical ID for the input time zone ID
     */
    static const UChar* U_EXPORT2 getCanonicalCLDRID(const UnicodeString &tzid, UErrorCode& status);

    /*
     * Conveninent method returning CLDR canonical ID for the given time zone
     */
    static const UChar* U_EXPORT2 getCanonicalCLDRID(const TimeZone& tz);

    /**
     * Return the canonical country code for this tzid.  If we have none, or if the time zone
     * is not associated with a country, return bogus string.
     * @param tzid Zone ID
     * @param country [output] Country code
     * @param isPrimary [output] TRUE if the zone is the primary zone for the country
     * @return A reference to the result country
     */
    static UnicodeString& U_EXPORT2 getCanonicalCountry(const UnicodeString &tzid, UnicodeString &country, UBool *isPrimary = NULL);

    /**
     * Returns a CLDR metazone ID for the given Olson tzid and time.
     */
    static UnicodeString& U_EXPORT2 getMetazoneID(const UnicodeString &tzid, UDate date, UnicodeString &result);
    /**
     * Returns an Olson ID for the ginve metazone and region
     */
    static UnicodeString& U_EXPORT2 getZoneIdByMetazone(const UnicodeString &mzid, const UnicodeString &region, UnicodeString &result);

    static const UVector* U_EXPORT2 getMetazoneMappings(const UnicodeString &tzid);

    static const UVector* U_EXPORT2 getAvailableMetazoneIDs();

    /**
     * Returns the pointer to the persistent time zone ID string, or NULL if the given tzid is not in the
     * tz database. This method is useful when you maintain persistent zone IDs without duplication.
     */
    static const UChar* U_EXPORT2 findTimeZoneID(const UnicodeString& tzid);

    /**
     * Returns the pointer to the persistent meta zone ID string, or NULL if the given mzid is not available.
     * This method is useful when you maintain persistent meta zone IDs without duplication.
     */
    static const UChar* U_EXPORT2 findMetaZoneID(const UnicodeString& mzid);

    /**
     * Creates a custom zone for the offset
     * @param offset GMT offset in milliseconds
     * @return A custom TimeZone for the offset with normalized time zone id
     */
    static TimeZone* createCustomTimeZone(int32_t offset);

    /**
     * Returns the time zone's short ID (null terminated) for the zone.
     * For example, "uslax" for zone "America/Los_Angeles".
     * @param tz the time zone
     * @return the short ID of the time zone, or null if the short ID is not available.
     */
    static const UChar* U_EXPORT2 getShortID(const TimeZone& tz);

    /**
     * Returns the time zone's short ID (null terminated) for the zone ID.
     * For example, "uslax" for zone ID "America/Los_Angeles".
     * @param tz the time zone ID
     * @return the short ID of the time zone ID, or null if the short ID is not available.
     */
    static const UChar* U_EXPORT2 getShortID(const UnicodeString& id);

private:
    ZoneMeta(); // Prevent construction.
    static UVector* createMetazoneMappings(const UnicodeString &tzid);
    static UnicodeString& formatCustomID(uint8_t hour, uint8_t min, uint8_t sec, UBool negative, UnicodeString& id);
    static const UChar* getShortIDFromCanonical(const UChar* canonicalID);
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */



#ifdef U_DEBUG_TZ
# include <stdio.h>
 // for debugging

static void debug_tz_loc(const char *f, int32_t l)
{
  fprintf(stderr, "%s:%d: ", f, l);
}

static void debug_tz_msg(const char *pat, ...)
{
  va_list ap;
  va_start(ap, pat);
  vfprintf(stderr, pat, ap);
  fflush(stderr);
}
// must use double parens, i.e.:  U_DEBUG_TZ_MSG(("four is: %d",4));
#define U_DEBUG_TZ_MSG(x) {debug_tz_loc(__FILE__,__LINE__);debug_tz_msg x;}
#else
#define U_DEBUG_TZ_MSG(x)
#endif

static UBool arrayEqual(const void *a1, const void *a2, int32_t size) {
    if (a1 == NULL && a2 == NULL) {
        return TRUE;
    }
    if ((a1 != NULL && a2 == NULL) || (a1 == NULL && a2 != NULL)) {
        return FALSE;
    }
    if (a1 == a2) {
        return TRUE;
    }

    return (uprv_memcmp(a1, a2, size) == 0);
}

U_NAMESPACE_BEGIN

#define kTRANS          "trans"
#define kTRANSPRE32     "transPre32"
#define kTRANSPOST32    "transPost32"
#define kTYPEOFFSETS    "typeOffsets"
#define kTYPEMAP        "typeMap"
#define kLINKS          "links"
#define kFINALRULE      "finalRule"
#define kFINALRAW       "finalRaw"
#define kFINALYEAR      "finalYear"

#define SECONDS_PER_DAY (24*60*60)

static const int32_t ZEROS[] = {0,0};

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(OlsonTimeZone)

/**
 * Default constructor.  Creates a time zone with an empty ID and
 * a fixed GMT offset of zero.
 */
/*OlsonTimeZone::OlsonTimeZone() : finalYear(INT32_MAX), finalMillis(DBL_MAX), finalZone(0), transitionRulesInitialized(FALSE) {
    clearTransitionRules();
    constructEmpty();
}*/

/**
 * Construct a GMT+0 zone with no transitions.  This is done when a
 * constructor fails so the resultant object is well-behaved.
 */
void OlsonTimeZone::constructEmpty() {
    canonicalID = NULL;

    transitionCountPre32 = transitionCount32 = transitionCountPost32 = 0;
    transitionTimesPre32 = transitionTimes32 = transitionTimesPost32 = NULL;

    typeMapData = NULL;

    typeCount = 1;
    typeOffsets = ZEROS;

    finalZone = NULL;
}

/**
 * Construct from a resource bundle
 * @param top the top-level zoneinfo resource bundle.  This is used
 * to lookup the rule that `res' may refer to, if there is one.
 * @param res the resource bundle of the zone to be constructed
 * @param ec input-output error code
 */
OlsonTimeZone::OlsonTimeZone(const UResourceBundle* top,
                             const UResourceBundle* res,
                             const UnicodeString& tzid,
                             UErrorCode& ec) :
  BasicTimeZone(tzid), finalZone(NULL)
{
    clearTransitionRules();
    U_DEBUG_TZ_MSG(("OlsonTimeZone(%s)\n", ures_getKey((UResourceBundle*)res)));
    if ((top == NULL || res == NULL) && U_SUCCESS(ec)) {
        ec = U_ILLEGAL_ARGUMENT_ERROR;
    }
    if (U_SUCCESS(ec)) {
        // TODO -- clean up -- Doesn't work if res points to an alias
        //        // TODO remove nonconst casts below when ures_* API is fixed
        //        setID(ures_getKey((UResourceBundle*) res)); // cast away const

        int32_t len;
        StackUResourceBundle r;

        // Pre-32bit second transitions
        ures_getByKey(res, kTRANSPRE32, r.getAlias(), &ec);
        transitionTimesPre32 = ures_getIntVector(r.getAlias(), &len, &ec);
        transitionCountPre32 = static_cast<int16_t>(len >> 1);
        if (ec == U_MISSING_RESOURCE_ERROR) {
            // No pre-32bit transitions
            transitionTimesPre32 = NULL;
            transitionCountPre32 = 0;
            ec = U_ZERO_ERROR;
        } else if (U_SUCCESS(ec) && (len < 0 || len > 0x7FFF || (len & 1) != 0) /* len must be even */) {
            ec = U_INVALID_FORMAT_ERROR;
        }

        // 32bit second transitions
        ures_getByKey(res, kTRANS, r.getAlias(), &ec);
        transitionTimes32 = ures_getIntVector(r.getAlias(), &len, &ec);
        transitionCount32 = static_cast<int16_t>(len);
        if (ec == U_MISSING_RESOURCE_ERROR) {
            // No 32bit transitions
            transitionTimes32 = NULL;
            transitionCount32 = 0;
            ec = U_ZERO_ERROR;
        } else if (U_SUCCESS(ec) && (len < 0 || len > 0x7FFF)) {
            ec = U_INVALID_FORMAT_ERROR;
        }

        // Post-32bit second transitions
        ures_getByKey(res, kTRANSPOST32, r.getAlias(), &ec);
        transitionTimesPost32 = ures_getIntVector(r.getAlias(), &len, &ec);
        transitionCountPost32 = static_cast<int16_t>(len >> 1);
        if (ec == U_MISSING_RESOURCE_ERROR) {
            // No pre-32bit transitions
            transitionTimesPost32 = NULL;
            transitionCountPost32 = 0;
            ec = U_ZERO_ERROR;
        } else if (U_SUCCESS(ec) && (len < 0 || len > 0x7FFF || (len & 1) != 0) /* len must be even */) {
            ec = U_INVALID_FORMAT_ERROR;
        }

        // Type offsets list must be of even size, with size >= 2
        ures_getByKey(res, kTYPEOFFSETS, r.getAlias(), &ec);
        typeOffsets = ures_getIntVector(r.getAlias(), &len, &ec);
        if (U_SUCCESS(ec) && (len < 2 || len > 0x7FFE || (len & 1) != 0)) {
            ec = U_INVALID_FORMAT_ERROR;
        }
        typeCount = (int16_t) len >> 1;

        // Type map data must be of the same size as the transition count
        typeMapData =  NULL;
        if (transitionCount() > 0) {
            ures_getByKey(res, kTYPEMAP, r.getAlias(), &ec);
            typeMapData = ures_getBinary(r.getAlias(), &len, &ec);
            if (ec == U_MISSING_RESOURCE_ERROR) {
                // no type mapping data
                ec = U_INVALID_FORMAT_ERROR;
            } else if (U_SUCCESS(ec) && len != transitionCount()) {
                ec = U_INVALID_FORMAT_ERROR;
            }
        }

        // Process final rule and data, if any
        const UChar *ruleIdUStr = ures_getStringByKey(res, kFINALRULE, &len, &ec);
        ures_getByKey(res, kFINALRAW, r.getAlias(), &ec);
        int32_t ruleRaw = ures_getInt(r.getAlias(), &ec);
        ures_getByKey(res, kFINALYEAR, r.getAlias(), &ec);
        int32_t ruleYear = ures_getInt(r.getAlias(), &ec);
        if (U_SUCCESS(ec)) {
            UnicodeString ruleID(TRUE, ruleIdUStr, len);
            UResourceBundle *rule = TimeZone::loadRule(top, ruleID, NULL, ec);
            const int32_t *ruleData = ures_getIntVector(rule, &len, &ec); 
            if (U_SUCCESS(ec) && len == 11) {
                UnicodeString emptyStr;
                finalZone = new SimpleTimeZone(
                    ruleRaw * U_MILLIS_PER_SECOND,
                    emptyStr,
                    (int8_t)ruleData[0], (int8_t)ruleData[1], (int8_t)ruleData[2],
                    ruleData[3] * U_MILLIS_PER_SECOND,
                    (SimpleTimeZone::TimeMode) ruleData[4],
                    (int8_t)ruleData[5], (int8_t)ruleData[6], (int8_t)ruleData[7],
                    ruleData[8] * U_MILLIS_PER_SECOND,
                    (SimpleTimeZone::TimeMode) ruleData[9],
                    ruleData[10] * U_MILLIS_PER_SECOND, ec);
                if (finalZone == NULL) {
                    ec = U_MEMORY_ALLOCATION_ERROR;
                } else {
                    finalStartYear = ruleYear;

                    // Note: Setting finalStartYear to the finalZone is problematic.  When a date is around
                    // year boundary, SimpleTimeZone may return false result when DST is observed at the 
                    // beginning of year.  We could apply safe margin (day or two), but when one of recurrent
                    // rules falls around year boundary, it could return false result.  Without setting the
                    // start year, finalZone works fine around the year boundary of the start year.

                    // finalZone->setStartYear(finalStartYear);


                    // Compute the millis for Jan 1, 0:00 GMT of the finalYear

                    // Note: finalStartMillis is used for detecting either if
                    // historic transition data or finalZone to be used.  In an
                    // extreme edge case - for example, two transitions fall into
                    // small windows of time around the year boundary, this may
                    // result incorrect offset computation.  But I think it will
                    // never happen practically.  Yoshito - Feb 20, 2010
                    finalStartMillis = Grego::fieldsToDay(finalStartYear, 0, 1) * U_MILLIS_PER_DAY;
                }
            } else {
                ec = U_INVALID_FORMAT_ERROR;
            }
            ures_close(rule);
        } else if (ec == U_MISSING_RESOURCE_ERROR) {
            // No final zone
            ec = U_ZERO_ERROR;
        }

        // initialize canonical ID
        canonicalID = ZoneMeta::getCanonicalCLDRID(tzid, ec);
    }

    if (U_FAILURE(ec)) {
        constructEmpty();
    }
}

/**
 * Copy constructor
 */
OlsonTimeZone::OlsonTimeZone(const OlsonTimeZone& other) :
    BasicTimeZone(other), finalZone(0) {
    *this = other;
}

/**
 * Assignment operator
 */
OlsonTimeZone& OlsonTimeZone::operator=(const OlsonTimeZone& other) {
    canonicalID = other.canonicalID;

    transitionTimesPre32 = other.transitionTimesPre32;
    transitionTimes32 = other.transitionTimes32;
    transitionTimesPost32 = other.transitionTimesPost32;

    transitionCountPre32 = other.transitionCountPre32;
    transitionCount32 = other.transitionCount32;
    transitionCountPost32 = other.transitionCountPost32;

    typeCount = other.typeCount;
    typeOffsets = other.typeOffsets;
    typeMapData = other.typeMapData;

    delete finalZone;
    finalZone = (other.finalZone != 0) ? other.finalZone->clone() : 0;

    finalStartYear = other.finalStartYear;
    finalStartMillis = other.finalStartMillis;

    clearTransitionRules();

    return *this;
}

/**
 * Destructor
 */
OlsonTimeZone::~OlsonTimeZone() {
    deleteTransitionRules();
    delete finalZone;
}

/**
 * Returns true if the two TimeZone objects are equal.
 */
UBool OlsonTimeZone::operator==(const TimeZone& other) const {
    return ((this == &other) ||
            (typeid(*this) == typeid(other) &&
            TimeZone::operator==(other) &&
            hasSameRules(other)));
}

/**
 * TimeZone API.
 */
OlsonTimeZone* OlsonTimeZone::clone() const {
    return new OlsonTimeZone(*this);
}

/**
 * TimeZone API.
 */
int32_t OlsonTimeZone::getOffset(uint8_t era, int32_t year, int32_t month,
                                 int32_t dom, uint8_t dow,
                                 int32_t millis, UErrorCode& ec) const {
    if (month < UCAL_JANUARY || month > UCAL_DECEMBER) {
        if (U_SUCCESS(ec)) {
            ec = U_ILLEGAL_ARGUMENT_ERROR;
        }
        return 0;
    } else {
        return getOffset(era, year, month, dom, dow, millis,
                         Grego::monthLength(year, month),
                         ec);
    }
}

/**
 * TimeZone API.
 */
int32_t OlsonTimeZone::getOffset(uint8_t era, int32_t year, int32_t month,
                                 int32_t dom, uint8_t dow,
                                 int32_t millis, int32_t monthLength,
                                 UErrorCode& ec) const {
    if (U_FAILURE(ec)) {
        return 0;
    }

    if ((era != GregorianCalendar::AD && era != GregorianCalendar::BC)
        || month < UCAL_JANUARY
        || month > UCAL_DECEMBER
        || dom < 1
        || dom > monthLength
        || dow < UCAL_SUNDAY
        || dow > UCAL_SATURDAY
        || millis < 0
        || millis >= U_MILLIS_PER_DAY
        || monthLength < 28
        || monthLength > 31) {
        ec = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if (era == GregorianCalendar::BC) {
        year = -year;
    }

    if (finalZone != NULL && year >= finalStartYear) {
        return finalZone->getOffset(era, year, month, dom, dow,
                                    millis, monthLength, ec);
    }

    // Compute local epoch millis from input fields
    UDate date = (UDate)(Grego::fieldsToDay(year, month, dom) * U_MILLIS_PER_DAY + millis);
    int32_t rawoff, dstoff;
    getHistoricalOffset(date, TRUE, kDaylight, kStandard, rawoff, dstoff);
    return rawoff + dstoff;
}

/**
 * TimeZone API.
 */
void OlsonTimeZone::getOffset(UDate date, UBool local, int32_t& rawoff,
                              int32_t& dstoff, UErrorCode& ec) const {
    if (U_FAILURE(ec)) {
        return;
    }
    if (finalZone != NULL && date >= finalStartMillis) {
        finalZone->getOffset(date, local, rawoff, dstoff, ec);
    } else {
        getHistoricalOffset(date, local, kFormer, kLatter, rawoff, dstoff);
    }
}

void
OlsonTimeZone::getOffsetFromLocal(UDate date, int32_t nonExistingTimeOpt, int32_t duplicatedTimeOpt,
                                  int32_t& rawoff, int32_t& dstoff, UErrorCode& ec) const {
    if (U_FAILURE(ec)) {
        return;
    }
    if (finalZone != NULL && date >= finalStartMillis) {
        finalZone->getOffsetFromLocal(date, nonExistingTimeOpt, duplicatedTimeOpt, rawoff, dstoff, ec);
    } else {
        getHistoricalOffset(date, TRUE, nonExistingTimeOpt, duplicatedTimeOpt, rawoff, dstoff);
    }
}


/**
 * TimeZone API.
 */
void OlsonTimeZone::setRawOffset(int32_t /*offsetMillis*/) {
    // We don't support this operation, since OlsonTimeZones are
    // immutable (except for the ID, which is in the base class).

    // Nothing to do!
}

/**
 * TimeZone API.
 */
int32_t OlsonTimeZone::getRawOffset() const {
    UErrorCode ec = U_ZERO_ERROR;
    int32_t raw, dst;
    getOffset((double) uprv_getUTCtime() * U_MILLIS_PER_SECOND,
              FALSE, raw, dst, ec);
    return raw;
}

#if defined U_DEBUG_TZ
void printTime(double ms) {
            int32_t year, month, dom, dow;
            double millis=0;
            double days = ClockMath::floorDivide(((double)ms), (double)U_MILLIS_PER_DAY, millis);
            
            Grego::dayToFields(days, year, month, dom, dow);
            U_DEBUG_TZ_MSG(("   getHistoricalOffset:  time %.1f (%04d.%02d.%02d+%.1fh)\n", ms,
                            year, month+1, dom, (millis/kOneHour)));
    }
#endif

int64_t
OlsonTimeZone::transitionTimeInSeconds(int16_t transIdx) const {
    U_ASSERT(transIdx >= 0 && transIdx < transitionCount()); 

    if (transIdx < transitionCountPre32) {
        return (((int64_t)((uint32_t)transitionTimesPre32[transIdx << 1])) << 32)
            | ((int64_t)((uint32_t)transitionTimesPre32[(transIdx << 1) + 1]));
    }

    transIdx -= transitionCountPre32;
    if (transIdx < transitionCount32) {
        return (int64_t)transitionTimes32[transIdx];
    }

    transIdx -= transitionCount32;
    return (((int64_t)((uint32_t)transitionTimesPost32[transIdx << 1])) << 32)
        | ((int64_t)((uint32_t)transitionTimesPost32[(transIdx << 1) + 1]));
}

// Maximum absolute offset in seconds (86400 seconds = 1 day)
// getHistoricalOffset uses this constant as safety margin of
// quick zone transition checking.
#define MAX_OFFSET_SECONDS 86400

void
OlsonTimeZone::getHistoricalOffset(UDate date, UBool local,
                                   int32_t NonExistingTimeOpt, int32_t DuplicatedTimeOpt,
                                   int32_t& rawoff, int32_t& dstoff) const {
    U_DEBUG_TZ_MSG(("getHistoricalOffset(%.1f, %s, %d, %d, raw, dst)\n",
        date, local?"T":"F", NonExistingTimeOpt, DuplicatedTimeOpt));
#if defined U_DEBUG_TZ
        printTime(date*1000.0);
#endif
    int16_t transCount = transitionCount();

    if (transCount > 0) {
        double sec = uprv_floor(date / U_MILLIS_PER_SECOND);
        if (!local && sec < transitionTimeInSeconds(0)) {
            // Before the first transition time
            rawoff = initialRawOffset() * U_MILLIS_PER_SECOND;
            dstoff = initialDstOffset() * U_MILLIS_PER_SECOND;
        } else {
            // Linear search from the end is the fastest approach, since
            // most lookups will happen at/near the end.
            int16_t transIdx;
            for (transIdx = transCount - 1; transIdx >= 0; transIdx--) {
                int64_t transition = transitionTimeInSeconds(transIdx);

                if (local && (sec >= (transition - MAX_OFFSET_SECONDS))) {
                    int32_t offsetBefore = zoneOffsetAt(transIdx - 1);
                    UBool dstBefore = dstOffsetAt(transIdx - 1) != 0;

                    int32_t offsetAfter = zoneOffsetAt(transIdx);
                    UBool dstAfter = dstOffsetAt(transIdx) != 0;

                    UBool dstToStd = dstBefore && !dstAfter;
                    UBool stdToDst = !dstBefore && dstAfter;
                    
                    if (offsetAfter - offsetBefore >= 0) {
                        // Positive transition, which makes a non-existing local time range
                        if (((NonExistingTimeOpt & kStdDstMask) == kStandard && dstToStd)
                                || ((NonExistingTimeOpt & kStdDstMask) == kDaylight && stdToDst)) {
                            transition += offsetBefore;
                        } else if (((NonExistingTimeOpt & kStdDstMask) == kStandard && stdToDst)
                                || ((NonExistingTimeOpt & kStdDstMask) == kDaylight && dstToStd)) {
                            transition += offsetAfter;
                        } else if ((NonExistingTimeOpt & kFormerLatterMask) == kLatter) {
                            transition += offsetBefore;
                        } else {
                            // Interprets the time with rule before the transition,
                            // default for non-existing time range
                            transition += offsetAfter;
                        }
                    } else {
                        // Negative transition, which makes a duplicated local time range
                        if (((DuplicatedTimeOpt & kStdDstMask) == kStandard && dstToStd)
                                || ((DuplicatedTimeOpt & kStdDstMask) == kDaylight && stdToDst)) {
                            transition += offsetAfter;
                        } else if (((DuplicatedTimeOpt & kStdDstMask) == kStandard && stdToDst)
                                || ((DuplicatedTimeOpt & kStdDstMask) == kDaylight && dstToStd)) {
                            transition += offsetBefore;
                        } else if ((DuplicatedTimeOpt & kFormerLatterMask) == kFormer) {
                            transition += offsetBefore;
                        } else {
                            // Interprets the time with rule after the transition,
                            // default for duplicated local time range
                            transition += offsetAfter;
                        }
                    }
                }
                if (sec >= transition) {
                    break;
                }
            }
            // transIdx could be -1 when local=true
            rawoff = rawOffsetAt(transIdx) * U_MILLIS_PER_SECOND;
            dstoff = dstOffsetAt(transIdx) * U_MILLIS_PER_SECOND;
        }
    } else {
        // No transitions, single pair of offsets only
        rawoff = initialRawOffset() * U_MILLIS_PER_SECOND;
        dstoff = initialDstOffset() * U_MILLIS_PER_SECOND;
    }
    U_DEBUG_TZ_MSG(("getHistoricalOffset(%.1f, %s, %d, %d, raw, dst) - raw=%d, dst=%d\n",
        date, local?"T":"F", NonExistingTimeOpt, DuplicatedTimeOpt, rawoff, dstoff));
}

/**
 * TimeZone API.
 */
UBool OlsonTimeZone::useDaylightTime() const {
    // If DST was observed in 1942 (for example) but has never been
    // observed from 1943 to the present, most clients will expect
    // this method to return FALSE.  This method determines whether
    // DST is in use in the current year (at any point in the year)
    // and returns TRUE if so.

    UDate current = uprv_getUTCtime();
    if (finalZone != NULL && current >= finalStartMillis) {
        return finalZone->useDaylightTime();
    }

    int32_t year, month, dom, dow, doy, mid;
    Grego::timeToFields(current, year, month, dom, dow, doy, mid);

    // Find start of this year, and start of next year
    double start = Grego::fieldsToDay(year, 0, 1) * SECONDS_PER_DAY;
    double limit = Grego::fieldsToDay(year+1, 0, 1) * SECONDS_PER_DAY;

    // Return TRUE if DST is observed at any time during the current
    // year.
    for (int16_t i = 0; i < transitionCount(); ++i) {
        double transition = (double)transitionTimeInSeconds(i);
        if (transition >= limit) {
            break;
        }
        if ((transition >= start && dstOffsetAt(i) != 0)
                || (transition > start && dstOffsetAt(i - 1) != 0)) {
            return TRUE;
        }
    }
    return FALSE;
}
int32_t 
OlsonTimeZone::getDSTSavings() const{
    if (finalZone != NULL){
        return finalZone->getDSTSavings();
    }
    return TimeZone::getDSTSavings();
}
/**
 * TimeZone API.
 */
UBool OlsonTimeZone::inDaylightTime(UDate date, UErrorCode& ec) const {
    int32_t raw, dst;
    getOffset(date, FALSE, raw, dst, ec);
    return dst != 0;
}

UBool
OlsonTimeZone::hasSameRules(const TimeZone &other) const {
    if (this == &other) {
        return TRUE;
    }
    const OlsonTimeZone* z = dynamic_cast<const OlsonTimeZone*>(&other);
    if (z == NULL) {
        return FALSE;
    }

    // [sic] pointer comparison: typeMapData points into
    // memory-mapped or DLL space, so if two zones have the same
    // pointer, they are equal.
    if (typeMapData == z->typeMapData) {
        return TRUE;
    }
    
    // If the pointers are not equal, the zones may still
    // be equal if their rules and transitions are equal
    if ((finalZone == NULL && z->finalZone != NULL)
        || (finalZone != NULL && z->finalZone == NULL)
        || (finalZone != NULL && z->finalZone != NULL && *finalZone != *z->finalZone)) {
        return FALSE;
    }

    if (finalZone != NULL) {
        if (finalStartYear != z->finalStartYear || finalStartMillis != z->finalStartMillis) {
            return FALSE;
        }
    }
    if (typeCount != z->typeCount
        || transitionCountPre32 != z->transitionCountPre32
        || transitionCount32 != z->transitionCount32
        || transitionCountPost32 != z->transitionCountPost32) {
        return FALSE;
    }

    return
        arrayEqual(transitionTimesPre32, z->transitionTimesPre32, sizeof(transitionTimesPre32[0]) * transitionCountPre32 << 1)
        && arrayEqual(transitionTimes32, z->transitionTimes32, sizeof(transitionTimes32[0]) * transitionCount32)
        && arrayEqual(transitionTimesPost32, z->transitionTimesPost32, sizeof(transitionTimesPost32[0]) * transitionCountPost32 << 1)
        && arrayEqual(typeOffsets, z->typeOffsets, sizeof(typeOffsets[0]) * typeCount << 1)
        && arrayEqual(typeMapData, z->typeMapData, sizeof(typeMapData[0]) * transitionCount());
}

void
OlsonTimeZone::clearTransitionRules(void) {
    initialRule = NULL;
    firstTZTransition = NULL;
    firstFinalTZTransition = NULL;
    historicRules = NULL;
    historicRuleCount = 0;
    finalZoneWithStartYear = NULL;
    firstTZTransitionIdx = 0;
    transitionRulesInitOnce.reset();
}

void
OlsonTimeZone::deleteTransitionRules(void) {
    if (initialRule != NULL) {
        delete initialRule;
    }
    if (firstTZTransition != NULL) {
        delete firstTZTransition;
    }
    if (firstFinalTZTransition != NULL) {
        delete firstFinalTZTransition;
    }
    if (finalZoneWithStartYear != NULL) {
        delete finalZoneWithStartYear;
    }
    if (historicRules != NULL) {
        for (int i = 0; i < historicRuleCount; i++) {
            if (historicRules[i] != NULL) {
                delete historicRules[i];
            }
        }
        uprv_free(historicRules);
    }
    clearTransitionRules();
}

/*
 * Lazy transition rules initializer
 */

static void U_CALLCONV initRules(OlsonTimeZone *This, UErrorCode &status) {
    This->initTransitionRules(status);
}
    
void
OlsonTimeZone::checkTransitionRules(UErrorCode& status) const {
    OlsonTimeZone *ncThis = const_cast<OlsonTimeZone *>(this);
    umtx_initOnce(ncThis->transitionRulesInitOnce, &initRules, ncThis, status);
}

void
OlsonTimeZone::initTransitionRules(UErrorCode& status) {
    if(U_FAILURE(status)) {
        return;
    }
    deleteTransitionRules();
    UnicodeString tzid;
    getID(tzid);

    UnicodeString stdName = tzid + UNICODE_STRING_SIMPLE("(STD)");
    UnicodeString dstName = tzid + UNICODE_STRING_SIMPLE("(DST)");

    int32_t raw, dst;

    // Create initial rule
    raw = initialRawOffset() * U_MILLIS_PER_SECOND;
    dst = initialDstOffset() * U_MILLIS_PER_SECOND;
    initialRule = new InitialTimeZoneRule((dst == 0 ? stdName : dstName), raw, dst);
    // Check to make sure initialRule was created
    if (initialRule == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        deleteTransitionRules();
        return;
    }

    int32_t transCount = transitionCount();
    if (transCount > 0) {
        int16_t transitionIdx, typeIdx;

        // We probably no longer need to check the first "real" transition
        // here, because the new tzcode remove such transitions already.
        // For now, keeping this code for just in case. Feb 19, 2010 Yoshito
        firstTZTransitionIdx = 0;
        for (transitionIdx = 0; transitionIdx < transCount; transitionIdx++) {
            if (typeMapData[transitionIdx] != 0) { // type 0 is the initial type
                break;
            }
            firstTZTransitionIdx++;
        }
        if (transitionIdx == transCount) {
            // Actually no transitions...
        } else {
            // Build historic rule array
            UDate* times = (UDate*)uprv_malloc(sizeof(UDate)*transCount); /* large enough to store all transition times */
            if (times == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                deleteTransitionRules();
                return;
            }
            for (typeIdx = 0; typeIdx < typeCount; typeIdx++) {
                // Gather all start times for each pair of offsets
                int32_t nTimes = 0;
                for (transitionIdx = firstTZTransitionIdx; transitionIdx < transCount; transitionIdx++) {
                    if (typeIdx == (int16_t)typeMapData[transitionIdx]) {
                        UDate tt = (UDate)transitionTime(transitionIdx);
                        if (finalZone == NULL || tt <= finalStartMillis) {
                            // Exclude transitions after finalMillis
                            times[nTimes++] = tt;
                        }
                    }
                }
                if (nTimes > 0) {
                    // Create a TimeArrayTimeZoneRule
                    raw = typeOffsets[typeIdx << 1] * U_MILLIS_PER_SECOND;
                    dst = typeOffsets[(typeIdx << 1) + 1] * U_MILLIS_PER_SECOND;
                    if (historicRules == NULL) {
                        historicRuleCount = typeCount;
                        historicRules = (TimeArrayTimeZoneRule**)uprv_malloc(sizeof(TimeArrayTimeZoneRule*)*historicRuleCount);
                        if (historicRules == NULL) {
                            status = U_MEMORY_ALLOCATION_ERROR;
                            deleteTransitionRules();
                            uprv_free(times);
                            return;
                        }
                        for (int i = 0; i < historicRuleCount; i++) {
                            // Initialize TimeArrayTimeZoneRule pointers as NULL
                            historicRules[i] = NULL;
                        }
                    }
                    historicRules[typeIdx] = new TimeArrayTimeZoneRule((dst == 0 ? stdName : dstName),
                        raw, dst, times, nTimes, DateTimeRule::UTC_TIME);
                    // Check for memory allocation error
                    if (historicRules[typeIdx] == NULL) {
                        status = U_MEMORY_ALLOCATION_ERROR;
                        deleteTransitionRules();
                        return;
                    }
                }
            }
            uprv_free(times);

            // Create initial transition
            typeIdx = (int16_t)typeMapData[firstTZTransitionIdx];
            firstTZTransition = new TimeZoneTransition((UDate)transitionTime(firstTZTransitionIdx),
                    *initialRule, *historicRules[typeIdx]);
            // Check to make sure firstTZTransition was created.
            if (firstTZTransition == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                deleteTransitionRules();
                return;
            }
        }
    }
    if (finalZone != NULL) {
        // Get the first occurence of final rule starts
        UDate startTime = (UDate)finalStartMillis;
        TimeZoneRule *firstFinalRule = NULL;

        if (finalZone->useDaylightTime()) {
            /*
             * Note: When an OlsonTimeZone is constructed, we should set the final year
             * as the start year of finalZone.  However, the bounday condition used for
             * getting offset from finalZone has some problems.
             * For now, we do not set the valid start year when the construction time
             * and create a clone and set the start year when extracting rules.
             */
            finalZoneWithStartYear = finalZone->clone();
            // Check to make sure finalZone was actually cloned.
            if (finalZoneWithStartYear == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                deleteTransitionRules();
                return;
            }
            finalZoneWithStartYear->setStartYear(finalStartYear);

            TimeZoneTransition tzt;
            finalZoneWithStartYear->getNextTransition(startTime, false, tzt);
            firstFinalRule  = tzt.getTo()->clone();
            // Check to make sure firstFinalRule received proper clone.
            if (firstFinalRule == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                deleteTransitionRules();
                return;
            }
            startTime = tzt.getTime();
        } else {
            // final rule with no transitions
            finalZoneWithStartYear = finalZone->clone();
            // Check to make sure finalZone was actually cloned.
            if (finalZoneWithStartYear == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                deleteTransitionRules();
                return;
            }
            finalZone->getID(tzid);
            firstFinalRule = new TimeArrayTimeZoneRule(tzid,
                finalZone->getRawOffset(), 0, &startTime, 1, DateTimeRule::UTC_TIME);
            // Check firstFinalRule was properly created.
            if (firstFinalRule == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                deleteTransitionRules();
                return;
            }
        }
        TimeZoneRule *prevRule = NULL;
        if (transCount > 0) {
            prevRule = historicRules[typeMapData[transCount - 1]];
        }
        if (prevRule == NULL) {
            // No historic transitions, but only finalZone available
            prevRule = initialRule;
        }
        firstFinalTZTransition = new TimeZoneTransition();
        // Check to make sure firstFinalTZTransition was created before dereferencing
        if (firstFinalTZTransition == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            deleteTransitionRules();
            return;
        }
        firstFinalTZTransition->setTime(startTime);
        firstFinalTZTransition->adoptFrom(prevRule->clone());
        firstFinalTZTransition->adoptTo(firstFinalRule);
    }
}

UBool
OlsonTimeZone::getNextTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const {
    UErrorCode status = U_ZERO_ERROR;
    checkTransitionRules(status);
    if (U_FAILURE(status)) {
        return FALSE;
    }

    if (finalZone != NULL) {
        if (inclusive && base == firstFinalTZTransition->getTime()) {
            result = *firstFinalTZTransition;
            return TRUE;
        } else if (base >= firstFinalTZTransition->getTime()) {
            if (finalZone->useDaylightTime()) {
                //return finalZone->getNextTransition(base, inclusive, result);
                return finalZoneWithStartYear->getNextTransition(base, inclusive, result);
            } else {
                // No more transitions
                return FALSE;
            }
        }
    }
    if (historicRules != NULL) {
        // Find a historical transition
        int16_t transCount = transitionCount();
        int16_t ttidx = transCount - 1;
        for (; ttidx >= firstTZTransitionIdx; ttidx--) {
            UDate t = (UDate)transitionTime(ttidx);
            if (base > t || (!inclusive && base == t)) {
                break;
            }
        }
        if (ttidx == transCount - 1)  {
            if (firstFinalTZTransition != NULL) {
                result = *firstFinalTZTransition;
                return TRUE;
            } else {
                return FALSE;
            }
        } else if (ttidx < firstTZTransitionIdx) {
            result = *firstTZTransition;
            return TRUE;
        } else {
            // Create a TimeZoneTransition
            TimeZoneRule *to = historicRules[typeMapData[ttidx + 1]];
            TimeZoneRule *from = historicRules[typeMapData[ttidx]];
            UDate startTime = (UDate)transitionTime(ttidx+1);

            // The transitions loaded from zoneinfo.res may contain non-transition data
            UnicodeString fromName, toName;
            from->getName(fromName);
            to->getName(toName);
            if (fromName == toName && from->getRawOffset() == to->getRawOffset()
                    && from->getDSTSavings() == to->getDSTSavings()) {
                return getNextTransition(startTime, false, result);
            }
            result.setTime(startTime);
            result.adoptFrom(from->clone());
            result.adoptTo(to->clone());
            return TRUE;
        }
    }
    return FALSE;
}

UBool
OlsonTimeZone::getPreviousTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const {
    UErrorCode status = U_ZERO_ERROR;
    checkTransitionRules(status);
    if (U_FAILURE(status)) {
        return FALSE;
    }

    if (finalZone != NULL) {
        if (inclusive && base == firstFinalTZTransition->getTime()) {
            result = *firstFinalTZTransition;
            return TRUE;
        } else if (base > firstFinalTZTransition->getTime()) {
            if (finalZone->useDaylightTime()) {
                //return finalZone->getPreviousTransition(base, inclusive, result);
                return finalZoneWithStartYear->getPreviousTransition(base, inclusive, result);
            } else {
                result = *firstFinalTZTransition;
                return TRUE;
            }
        }
    }

    if (historicRules != NULL) {
        // Find a historical transition
        int16_t ttidx = transitionCount() - 1;
        for (; ttidx >= firstTZTransitionIdx; ttidx--) {
            UDate t = (UDate)transitionTime(ttidx);
            if (base > t || (inclusive && base == t)) {
                break;
            }
        }
        if (ttidx < firstTZTransitionIdx) {
            // No more transitions
            return FALSE;
        } else if (ttidx == firstTZTransitionIdx) {
            result = *firstTZTransition;
            return TRUE;
        } else {
            // Create a TimeZoneTransition
            TimeZoneRule *to = historicRules[typeMapData[ttidx]];
            TimeZoneRule *from = historicRules[typeMapData[ttidx-1]];
            UDate startTime = (UDate)transitionTime(ttidx);

            // The transitions loaded from zoneinfo.res may contain non-transition data
            UnicodeString fromName, toName;
            from->getName(fromName);
            to->getName(toName);
            if (fromName == toName && from->getRawOffset() == to->getRawOffset()
                    && from->getDSTSavings() == to->getDSTSavings()) {
                return getPreviousTransition(startTime, false, result);
            }
            result.setTime(startTime);
            result.adoptFrom(from->clone());
            result.adoptTo(to->clone());
            return TRUE;
        }
    }
    return FALSE;
}

int32_t
OlsonTimeZone::countTransitionRules(UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return 0;
    }
    checkTransitionRules(status);
    if (U_FAILURE(status)) {
        return 0;
    }

    int32_t count = 0;
    if (historicRules != NULL) {
        // historicRules may contain null entries when original zoneinfo data
        // includes non transition data.
        for (int32_t i = 0; i < historicRuleCount; i++) {
            if (historicRules[i] != NULL) {
                count++;
            }
        }
    }
    if (finalZone != NULL) {
        if (finalZone->useDaylightTime()) {
            count += 2;
        } else {
            count++;
        }
    }
    return count;
}

void
OlsonTimeZone::getTimeZoneRules(const InitialTimeZoneRule*& initial,
                                const TimeZoneRule* trsrules[],
                                int32_t& trscount,
                                UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    checkTransitionRules(status);
    if (U_FAILURE(status)) {
        return;
    }

    // Initial rule
    initial = initialRule;

    // Transition rules
    int32_t cnt = 0;
    if (historicRules != NULL && trscount > cnt) {
        // historicRules may contain null entries when original zoneinfo data
        // includes non transition data.
        for (int32_t i = 0; i < historicRuleCount; i++) {
            if (historicRules[i] != NULL) {
                trsrules[cnt++] = historicRules[i];
                if (cnt >= trscount) {
                    break;
                }
            }
        }
    }
    if (finalZoneWithStartYear != NULL && trscount > cnt) {
        const InitialTimeZoneRule *tmpini;
        int32_t tmpcnt = trscount - cnt;
        finalZoneWithStartYear->getTimeZoneRules(tmpini, &trsrules[cnt], tmpcnt, status);
        if (U_FAILURE(status)) {
            return;
        }
        cnt += tmpcnt;
    }
    // Set the result length
    trscount = cnt;
}

U_NAMESPACE_END

#endif // !UCONFIG_NO_FORMATTING

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 ******************************************************************************
 * Copyright (C) 2003-2013, International Business Machines Corporation
 * and others. All Rights Reserved.
 ******************************************************************************
 *
 * File PERSNCAL.CPP
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   9/23/2003   mehran      posted to icu-design
 *   10/1/2012   roozbeh     Fixed algorithm and heavily refactored and rewrote
 *                           based on the implementation of Gregorian
 *****************************************************************************
 */



#if !UCONFIG_NO_FORMATTING


 // Math
#include <float.h>

static const int16_t kPersianNumDays[]
= {0,31,62,93,124,155,186,216,246,276,306,336}; // 0-based, for day-in-year
static const int8_t kPersianMonthLength[]
= {31,31,31,31,31,31,30,30,30,30,30,29}; // 0-based
static const int8_t kPersianLeapMonthLength[]
= {31,31,31,31,31,31,30,30,30,30,30,30}; // 0-based

static const int32_t kPersianCalendarLimits[UCAL_FIELD_COUNT][4] = {
    // Minimum  Greatest     Least   Maximum
    //           Minimum   Maximum
    {        0,        0,        0,        0}, // ERA
    { -5000000, -5000000,  5000000,  5000000}, // YEAR
    {        0,        0,       11,       11}, // MONTH
    {        1,        1,       52,       53}, // WEEK_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // WEEK_OF_MONTH
    {        1,       1,        29,       31}, // DAY_OF_MONTH
    {        1,       1,       365,      366}, // DAY_OF_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DAY_OF_WEEK
    {        1,       1,         5,        5}, // DAY_OF_WEEK_IN_MONTH
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // AM_PM
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // HOUR_OF_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MINUTE
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // SECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECOND
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // ZONE_OFFSET
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DST_OFFSET
    { -5000000, -5000000,  5000000,  5000000}, // YEAR_WOY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // DOW_LOCAL
    { -5000000, -5000000,  5000000,  5000000}, // EXTENDED_YEAR
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // JULIAN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // MILLISECONDS_IN_DAY
    {/*N/A*/-1,/*N/A*/-1,/*N/A*/-1,/*N/A*/-1}, // IS_LEAP_MONTH
};

U_NAMESPACE_BEGIN

static const int32_t PERSIAN_EPOCH = 1948320;

// Implementation of the PersianCalendar class

//-------------------------------------------------------------------------
// Constructors...
//-------------------------------------------------------------------------

const char *PersianCalendar::getType() const {
    return "persian";
}

PersianCalendar* PersianCalendar::clone() const {
    return new PersianCalendar(*this);
}

PersianCalendar::PersianCalendar(const Locale& aLocale, UErrorCode& success)
  :   Calendar(TimeZone::createDefault(), aLocale, success)
{
    setTimeInMillis(getNow(), success); // Call this again now that the vtable is set up properly.
}

PersianCalendar::PersianCalendar(const PersianCalendar& other) : Calendar(other) {
}

PersianCalendar::~PersianCalendar()
{
}

//-------------------------------------------------------------------------
// Minimum / Maximum access functions
//-------------------------------------------------------------------------


int32_t PersianCalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const {
    return kPersianCalendarLimits[field][limitType];
}

//-------------------------------------------------------------------------
// Assorted calculation utilities
//

/**
 * Determine whether a year is a leap year in the Persian calendar
 */
UBool PersianCalendar::isLeapYear(int32_t year)
{
    int32_t remainder;
    ClockMath::floorDivide(25 * year + 11, 33, remainder);
    return (remainder < 8);
}

/**
 * Return the day # on which the given year starts.  Days are counted
 * from the Persian epoch, origin 0.
 */
int32_t PersianCalendar::yearStart(int32_t year) {
    return handleComputeMonthStart(year,0,FALSE);
}

/**
 * Return the day # on which the given month starts.  Days are counted
 * from the Persian epoch, origin 0.
 *
 * @param year  The Persian year
 * @param year  The Persian month, 0-based
 */
int32_t PersianCalendar::monthStart(int32_t year, int32_t month) const {
    return handleComputeMonthStart(year,month,TRUE);
}

//----------------------------------------------------------------------
// Calendar framework
//----------------------------------------------------------------------

/**
 * Return the length (in days) of the given month.
 *
 * @param year  The Persian year
 * @param year  The Persian month, 0-based
 */
int32_t PersianCalendar::handleGetMonthLength(int32_t extendedYear, int32_t month) const {
    // If the month is out of range, adjust it into range, and
    // modify the extended year value accordingly.
    if (month < 0 || month > 11) {
        extendedYear += ClockMath::floorDivide(month, 12, month);
    }

    return isLeapYear(extendedYear) ? kPersianLeapMonthLength[month] : kPersianMonthLength[month];
}

/**
 * Return the number of days in the given Persian year
 */
int32_t PersianCalendar::handleGetYearLength(int32_t extendedYear) const {
    return isLeapYear(extendedYear) ? 366 : 365;
}

//-------------------------------------------------------------------------
// Functions for converting from field values to milliseconds....
//-------------------------------------------------------------------------

// Return JD of start of given month/year
int32_t PersianCalendar::handleComputeMonthStart(int32_t eyear, int32_t month, UBool /*useMonth*/) const {
    // If the month is out of range, adjust it into range, and
    // modify the extended year value accordingly.
    if (month < 0 || month > 11) {
        eyear += ClockMath::floorDivide(month, 12, month);
    }

    int32_t julianDay = PERSIAN_EPOCH - 1 + 365 * (eyear - 1) + ClockMath::floorDivide(8 * eyear + 21, 33);

    if (month != 0) {
        julianDay += kPersianNumDays[month];
    }

    return julianDay;
}

//-------------------------------------------------------------------------
// Functions for converting from milliseconds to field values
//-------------------------------------------------------------------------

int32_t PersianCalendar::handleGetExtendedYear() {
    int32_t year;
    if (newerField(UCAL_EXTENDED_YEAR, UCAL_YEAR) == UCAL_EXTENDED_YEAR) {
        year = internalGet(UCAL_EXTENDED_YEAR, 1); // Default to year 1
    } else {
        year = internalGet(UCAL_YEAR, 1); // Default to year 1
    }
    return year;
}

/**
 * Override Calendar to compute several fields specific to the Persian
 * calendar system.  These are:
 *
 * <ul><li>ERA
 * <li>YEAR
 * <li>MONTH
 * <li>DAY_OF_MONTH
 * <li>DAY_OF_YEAR
 * <li>EXTENDED_YEAR</ul>
 *
 * The DAY_OF_WEEK and DOW_LOCAL fields are already set when this
 * method is called.
 */
void PersianCalendar::handleComputeFields(int32_t julianDay, UErrorCode &/*status*/) {
    int32_t year, month, dayOfMonth, dayOfYear;

    int32_t daysSinceEpoch = julianDay - PERSIAN_EPOCH;
    year = 1 + (int32_t)ClockMath::floorDivide(33 * (int64_t)daysSinceEpoch + 3, (int64_t)12053);

    int32_t farvardin1 = 365 * (year - 1) + ClockMath::floorDivide(8 * year + 21, 33);
    dayOfYear = (daysSinceEpoch - farvardin1); // 0-based
    if (dayOfYear < 216) { // Compute 0-based month
        month = dayOfYear / 31;
    } else {
        month = (dayOfYear - 6) / 30;
    }
    dayOfMonth = dayOfYear - kPersianNumDays[month] + 1;
    ++dayOfYear; // Make it 1-based now

    internalSet(UCAL_ERA, 0);
    internalSet(UCAL_YEAR, year);
    internalSet(UCAL_EXTENDED_YEAR, year);
    internalSet(UCAL_MONTH, month);
    internalSet(UCAL_DAY_OF_MONTH, dayOfMonth);
    internalSet(UCAL_DAY_OF_YEAR, dayOfYear);
}

UBool
PersianCalendar::inDaylightTime(UErrorCode& status) const
{
    // copied from GregorianCalendar
    if (U_FAILURE(status) || !getTimeZone().useDaylightTime())
        return FALSE;

    // Force an update of the state of the Calendar.
    ((PersianCalendar*)this)->complete(status); // cast away const

    return (UBool)(U_SUCCESS(status) ? (internalGet(UCAL_DST_OFFSET) != 0) : FALSE);
}

// default century

static UDate           persncal_gSystemDefaultCenturyStart       = DBL_MIN;
static int32_t         persncal_gSystemDefaultCenturyStartYear   = -1;
static icu::UInitOnce  persncal_gSystemDefaultCenturyInit        = U_INITONCE_INITIALIZER;

UBool PersianCalendar::haveDefaultCentury() const
{
    return TRUE;
}

static void U_CALLCONV persncal_initializeSystemDefaultCentury() {
    // initialize systemDefaultCentury and systemDefaultCenturyYear based
    // on the current time.  They'll be set to 80 years before
    // the current time.
    UErrorCode status = U_ZERO_ERROR;
    PersianCalendar calendar(Locale("@calendar=persian"),status);
    if (U_SUCCESS(status))
    {
        calendar.setTime(Calendar::getNow(), status);
        calendar.add(UCAL_YEAR, -80, status);

        persncal_gSystemDefaultCenturyStart = calendar.getTime(status);
        persncal_gSystemDefaultCenturyStartYear = calendar.get(UCAL_YEAR, status);
    }
    // We have no recourse upon failure unless we want to propagate the failure
    // out.
}

UDate PersianCalendar::defaultCenturyStart() const {
    // lazy-evaluate systemDefaultCenturyStart
    umtx_initOnce(persncal_gSystemDefaultCenturyInit, &persncal_initializeSystemDefaultCentury);
    return persncal_gSystemDefaultCenturyStart;
}

int32_t PersianCalendar::defaultCenturyStartYear() const {
    // lazy-evaluate systemDefaultCenturyStartYear
    umtx_initOnce(persncal_gSystemDefaultCenturyInit, &persncal_initializeSystemDefaultCentury);
    return persncal_gSystemDefaultCenturyStartYear;
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(PersianCalendar)

U_NAMESPACE_END

#endif

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2015, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File PLURFMT.CPP
*******************************************************************************
*/
















#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

using number::impl::DecimalQuantity;

static const UChar OTHER_STRING[] = {
    0x6F, 0x74, 0x68, 0x65, 0x72, 0  // "other"
};

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(PluralFormat)

PluralFormat::PluralFormat(UErrorCode& status)
        : locale(Locale::getDefault()),
          msgPattern(status),
          numberFormat(NULL),
          offset(0) {
    init(NULL, UPLURAL_TYPE_CARDINAL, status);
}

PluralFormat::PluralFormat(const Locale& loc, UErrorCode& status)
        : locale(loc),
          msgPattern(status),
          numberFormat(NULL),
          offset(0) {
    init(NULL, UPLURAL_TYPE_CARDINAL, status);
}

PluralFormat::PluralFormat(const PluralRules& rules, UErrorCode& status)
        : locale(Locale::getDefault()),
          msgPattern(status),
          numberFormat(NULL),
          offset(0) {
    init(&rules, UPLURAL_TYPE_COUNT, status);
}

PluralFormat::PluralFormat(const Locale& loc,
                           const PluralRules& rules,
                           UErrorCode& status)
        : locale(loc),
          msgPattern(status),
          numberFormat(NULL),
          offset(0) {
    init(&rules, UPLURAL_TYPE_COUNT, status);
}

PluralFormat::PluralFormat(const Locale& loc,
                           UPluralType type,
                           UErrorCode& status)
        : locale(loc),
          msgPattern(status),
          numberFormat(NULL),
          offset(0) {
    init(NULL, type, status);
}

PluralFormat::PluralFormat(const UnicodeString& pat,
                           UErrorCode& status)
        : locale(Locale::getDefault()),
          msgPattern(status),
          numberFormat(NULL),
          offset(0) {
    init(NULL, UPLURAL_TYPE_CARDINAL, status);
    applyPattern(pat, status);
}

PluralFormat::PluralFormat(const Locale& loc,
                           const UnicodeString& pat,
                           UErrorCode& status)
        : locale(loc),
          msgPattern(status),
          numberFormat(NULL),
          offset(0) {
    init(NULL, UPLURAL_TYPE_CARDINAL, status);
    applyPattern(pat, status);
}

PluralFormat::PluralFormat(const PluralRules& rules,
                           const UnicodeString& pat,
                           UErrorCode& status)
        : locale(Locale::getDefault()),
          msgPattern(status),
          numberFormat(NULL),
          offset(0) {
    init(&rules, UPLURAL_TYPE_COUNT, status);
    applyPattern(pat, status);
}

PluralFormat::PluralFormat(const Locale& loc,
                           const PluralRules& rules,
                           const UnicodeString& pat,
                           UErrorCode& status)
        : locale(loc),
          msgPattern(status),
          numberFormat(NULL),
          offset(0) {
    init(&rules, UPLURAL_TYPE_COUNT, status);
    applyPattern(pat, status);
}

PluralFormat::PluralFormat(const Locale& loc,
                           UPluralType type,
                           const UnicodeString& pat,
                           UErrorCode& status)
        : locale(loc),
          msgPattern(status),
          numberFormat(NULL),
          offset(0) {
    init(NULL, type, status);
    applyPattern(pat, status);
}

PluralFormat::PluralFormat(const PluralFormat& other)
        : Format(other),
          locale(other.locale),
          msgPattern(other.msgPattern),
          numberFormat(NULL),
          offset(other.offset) {
    copyObjects(other);
}

void
PluralFormat::copyObjects(const PluralFormat& other) {
    UErrorCode status = U_ZERO_ERROR;
    if (numberFormat != NULL) {
        delete numberFormat;
    }
    if (pluralRulesWrapper.pluralRules != NULL) {
        delete pluralRulesWrapper.pluralRules;
    }

    if (other.numberFormat == NULL) {
        numberFormat = NumberFormat::createInstance(locale, status);
    } else {
        numberFormat = other.numberFormat->clone();
    }
    if (other.pluralRulesWrapper.pluralRules == NULL) {
        pluralRulesWrapper.pluralRules = PluralRules::forLocale(locale, status);
    } else {
        pluralRulesWrapper.pluralRules = other.pluralRulesWrapper.pluralRules->clone();
    }
}


PluralFormat::~PluralFormat() {
    delete numberFormat;
}

void
PluralFormat::init(const PluralRules* rules, UPluralType type, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }

    if (rules==NULL) {
        pluralRulesWrapper.pluralRules = PluralRules::forLocale(locale, type, status);
    } else {
        pluralRulesWrapper.pluralRules = rules->clone();
        if (pluralRulesWrapper.pluralRules == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
    }

    numberFormat= NumberFormat::createInstance(locale, status);
}

void
PluralFormat::applyPattern(const UnicodeString& newPattern, UErrorCode& status) {
    msgPattern.parsePluralStyle(newPattern, NULL, status);
    if (U_FAILURE(status)) {
        msgPattern.clear();
        offset = 0;
        return;
    }
    offset = msgPattern.getPluralOffset(0);
}

UnicodeString&
PluralFormat::format(const Formattable& obj,
                   UnicodeString& appendTo,
                   FieldPosition& pos,
                   UErrorCode& status) const
{
    if (U_FAILURE(status)) return appendTo;

    if (obj.isNumeric()) {
        return format(obj, obj.getDouble(), appendTo, pos, status);
    } else {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return appendTo;
    }
}

UnicodeString
PluralFormat::format(int32_t number, UErrorCode& status) const {
    FieldPosition fpos(FieldPosition::DONT_CARE);
    UnicodeString result;
    return format(Formattable(number), number, result, fpos, status);
}

UnicodeString
PluralFormat::format(double number, UErrorCode& status) const {
    FieldPosition fpos(FieldPosition::DONT_CARE);
    UnicodeString result;
    return format(Formattable(number), number, result, fpos, status);
}


UnicodeString&
PluralFormat::format(int32_t number,
                     UnicodeString& appendTo,
                     FieldPosition& pos,
                     UErrorCode& status) const {
    return format(Formattable(number), (double)number, appendTo, pos, status);
}

UnicodeString&
PluralFormat::format(double number,
                     UnicodeString& appendTo,
                     FieldPosition& pos,
                     UErrorCode& status) const {
    return format(Formattable(number), (double)number, appendTo, pos, status);
}

UnicodeString&
PluralFormat::format(const Formattable& numberObject, double number,
                     UnicodeString& appendTo,
                     FieldPosition& pos,
                     UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    if (msgPattern.countParts() == 0) {
        return numberFormat->format(numberObject, appendTo, pos, status);
    }

    // Get the appropriate sub-message.
    // Select it based on the formatted number-offset.
    double numberMinusOffset = number - offset;
    // Call NumberFormatter to get both the DecimalQuantity and the string.
    // This call site needs to use more internal APIs than the Java equivalent.
    number::impl::UFormattedNumberData data;
    if (offset == 0) {
        // could be BigDecimal etc.
        numberObject.populateDecimalQuantity(data.quantity, status);
    } else {
        data.quantity.setToDouble(numberMinusOffset);
    }
    UnicodeString numberString;
    auto *decFmt = dynamic_cast<DecimalFormat *>(numberFormat);
    if(decFmt != nullptr) {
        const number::LocalizedNumberFormatter* lnf = decFmt->toNumberFormatter(status);
        if (U_FAILURE(status)) {
            return appendTo;
        }
        lnf->formatImpl(&data, status); // mutates &data
        if (U_FAILURE(status)) {
            return appendTo;
        }
        numberString = data.getStringRef().toUnicodeString();
    } else {
        if (offset == 0) {
            numberFormat->format(numberObject, numberString, status);
        } else {
            numberFormat->format(numberMinusOffset, numberString, status);
        }
    }

    int32_t partIndex = findSubMessage(msgPattern, 0, pluralRulesWrapper, &data.quantity, number, status);
    if (U_FAILURE(status)) { return appendTo; }
    // Replace syntactic # signs in the top level of this sub-message
    // (not in nested arguments) with the formatted number-offset.
    const UnicodeString& pattern = msgPattern.getPatternString();
    int32_t prevIndex = msgPattern.getPart(partIndex).getLimit();
    for (;;) {
        const MessagePattern::Part& part = msgPattern.getPart(++partIndex);
        const UMessagePatternPartType type = part.getType();
        int32_t index = part.getIndex();
        if (type == UMSGPAT_PART_TYPE_MSG_LIMIT) {
            return appendTo.append(pattern, prevIndex, index - prevIndex);
        } else if ((type == UMSGPAT_PART_TYPE_REPLACE_NUMBER) ||
            (type == UMSGPAT_PART_TYPE_SKIP_SYNTAX && MessageImpl::jdkAposMode(msgPattern))) {
            appendTo.append(pattern, prevIndex, index - prevIndex);
            if (type == UMSGPAT_PART_TYPE_REPLACE_NUMBER) {
                appendTo.append(numberString);
            }
            prevIndex = part.getLimit();
        } else if (type == UMSGPAT_PART_TYPE_ARG_START) {
            appendTo.append(pattern, prevIndex, index - prevIndex);
            prevIndex = index;
            partIndex = msgPattern.getLimitPartIndex(partIndex);
            index = msgPattern.getPart(partIndex).getLimit();
            MessageImpl::appendReducedApostrophes(pattern, prevIndex, index, appendTo);
            prevIndex = index;
        }
    }
}

UnicodeString&
PluralFormat::toPattern(UnicodeString& appendTo) {
    if (0 == msgPattern.countParts()) {
        appendTo.setToBogus();
    } else {
        appendTo.append(msgPattern.getPatternString());
    }
    return appendTo;
}

void
PluralFormat::setLocale(const Locale& loc, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    locale = loc;
    msgPattern.clear();
    delete numberFormat;
    offset = 0;
    numberFormat = NULL;
    pluralRulesWrapper.reset();
    init(NULL, UPLURAL_TYPE_CARDINAL, status);
}

void
PluralFormat::setNumberFormat(const NumberFormat* format, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    NumberFormat* nf = format->clone();
    if (nf != NULL) {
        delete numberFormat;
        numberFormat = nf;
    } else {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
}

PluralFormat*
PluralFormat::clone() const
{
    return new PluralFormat(*this);
}


PluralFormat&
PluralFormat::operator=(const PluralFormat& other) {
    if (this != &other) {
        locale = other.locale;
        msgPattern = other.msgPattern;
        offset = other.offset;
        copyObjects(other);
    }

    return *this;
}

UBool
PluralFormat::operator==(const Format& other) const {
    if (this == &other) {
        return TRUE;
    }
    if (!Format::operator==(other)) {
        return FALSE;
    }
    const PluralFormat& o = (const PluralFormat&)other;
    return
        locale == o.locale &&
        msgPattern == o.msgPattern &&  // implies same offset
        (numberFormat == NULL) == (o.numberFormat == NULL) &&
        (numberFormat == NULL || *numberFormat == *o.numberFormat) &&
        (pluralRulesWrapper.pluralRules == NULL) == (o.pluralRulesWrapper.pluralRules == NULL) &&
        (pluralRulesWrapper.pluralRules == NULL ||
            *pluralRulesWrapper.pluralRules == *o.pluralRulesWrapper.pluralRules);
}

UBool
PluralFormat::operator!=(const Format& other) const {
    return  !operator==(other);
}

void
PluralFormat::parseObject(const UnicodeString& /*source*/,
                        Formattable& /*result*/,
                        ParsePosition& pos) const
{
    // Parsing not supported.
    pos.setErrorIndex(pos.getIndex());
}

int32_t PluralFormat::findSubMessage(const MessagePattern& pattern, int32_t partIndex,
                                     const PluralSelector& selector, void *context,
                                     double number, UErrorCode& ec) {
    if (U_FAILURE(ec)) {
        return 0;
    }
    int32_t count=pattern.countParts();
    double offset;
    const MessagePattern::Part* part=&pattern.getPart(partIndex);
    if (MessagePattern::Part::hasNumericValue(part->getType())) {
        offset=pattern.getNumericValue(*part);
        ++partIndex;
    } else {
        offset=0;
    }
    // The keyword is empty until we need to match against a non-explicit, not-"other" value.
    // Then we get the keyword from the selector.
    // (In other words, we never call the selector if we match against an explicit value,
    // or if the only non-explicit keyword is "other".)
    UnicodeString keyword;
    UnicodeString other(FALSE, OTHER_STRING, 5);
    // When we find a match, we set msgStart>0 and also set this boolean to true
    // to avoid matching the keyword again (duplicates are allowed)
    // while we continue to look for an explicit-value match.
    UBool haveKeywordMatch=FALSE;
    // msgStart is 0 until we find any appropriate sub-message.
    // We remember the first "other" sub-message if we have not seen any
    // appropriate sub-message before.
    // We remember the first matching-keyword sub-message if we have not seen
    // one of those before.
    // (The parser allows [does not check for] duplicate keywords.
    // We just have to make sure to take the first one.)
    // We avoid matching the keyword twice by also setting haveKeywordMatch=true
    // at the first keyword match.
    // We keep going until we find an explicit-value match or reach the end of the plural style.
    int32_t msgStart=0;
    // Iterate over (ARG_SELECTOR [ARG_INT|ARG_DOUBLE] message) tuples
    // until ARG_LIMIT or end of plural-only pattern.
    do {
        part=&pattern.getPart(partIndex++);
        const UMessagePatternPartType type = part->getType();
        if(type==UMSGPAT_PART_TYPE_ARG_LIMIT) {
            break;
        }
        U_ASSERT (type==UMSGPAT_PART_TYPE_ARG_SELECTOR);
        // part is an ARG_SELECTOR followed by an optional explicit value, and then a message
        if(MessagePattern::Part::hasNumericValue(pattern.getPartType(partIndex))) {
            // explicit value like "=2"
            part=&pattern.getPart(partIndex++);
            if(number==pattern.getNumericValue(*part)) {
                // matches explicit value
                return partIndex;
            }
        } else if(!haveKeywordMatch) {
            // plural keyword like "few" or "other"
            // Compare "other" first and call the selector if this is not "other".
            if(pattern.partSubstringMatches(*part, other)) {
                if(msgStart==0) {
                    msgStart=partIndex;
                    if(0 == keyword.compare(other)) {
                        // This is the first "other" sub-message,
                        // and the selected keyword is also "other".
                        // Do not match "other" again.
                        haveKeywordMatch=TRUE;
                    }
                }
            } else {
                if(keyword.isEmpty()) {
                    keyword=selector.select(context, number-offset, ec);
                    if(msgStart!=0 && (0 == keyword.compare(other))) {
                        // We have already seen an "other" sub-message.
                        // Do not match "other" again.
                        haveKeywordMatch=TRUE;
                        // Skip keyword matching but do getLimitPartIndex().
                    }
                }
                if(!haveKeywordMatch && pattern.partSubstringMatches(*part, keyword)) {
                    // keyword matches
                    msgStart=partIndex;
                    // Do not match this keyword again.
                    haveKeywordMatch=TRUE;
                }
            }
        }
        partIndex=pattern.getLimitPartIndex(partIndex);
    } while(++partIndex<count);
    return msgStart;
}

void PluralFormat::parseType(const UnicodeString& source, const NFRule *rbnfLenientScanner, Formattable& result, FieldPosition& pos) const {
    // If no pattern was applied, return null.
    if (msgPattern.countParts() == 0) {
        pos.setBeginIndex(-1);
        pos.setEndIndex(-1);
        return;
    }
    int partIndex = 0;
    int currMatchIndex;
    int count=msgPattern.countParts();
    int startingAt = pos.getBeginIndex();
    if (startingAt < 0) {
        startingAt = 0;
    }

    // The keyword is null until we need to match against a non-explicit, not-"other" value.
    // Then we get the keyword from the selector.
    // (In other words, we never call the selector if we match against an explicit value,
    // or if the only non-explicit keyword is "other".)
    UnicodeString keyword;
    UnicodeString matchedWord;
    const UnicodeString& pattern = msgPattern.getPatternString();
    int matchedIndex = -1;
    // Iterate over (ARG_SELECTOR ARG_START message ARG_LIMIT) tuples
    // until the end of the plural-only pattern.
    while (partIndex < count) {
        const MessagePattern::Part* partSelector = &msgPattern.getPart(partIndex++);
        if (partSelector->getType() != UMSGPAT_PART_TYPE_ARG_SELECTOR) {
            // Bad format
            continue;
        }

        const MessagePattern::Part* partStart = &msgPattern.getPart(partIndex++);
        if (partStart->getType() != UMSGPAT_PART_TYPE_MSG_START) {
            // Bad format
            continue;
        }

        const MessagePattern::Part* partLimit = &msgPattern.getPart(partIndex++);
        if (partLimit->getType() != UMSGPAT_PART_TYPE_MSG_LIMIT) {
            // Bad format
            continue;
        }

        UnicodeString currArg = pattern.tempSubString(partStart->getLimit(), partLimit->getIndex() - partStart->getLimit());
        if (rbnfLenientScanner != NULL) {
            // If lenient parsing is turned ON, we've got some time consuming parsing ahead of us.
            int32_t length = -1;
            currMatchIndex = rbnfLenientScanner->findTextLenient(source, currArg, startingAt, &length);
        }
        else {
            currMatchIndex = source.indexOf(currArg, startingAt);
        }
        if (currMatchIndex >= 0 && currMatchIndex >= matchedIndex && currArg.length() > matchedWord.length()) {
            matchedIndex = currMatchIndex;
            matchedWord = currArg;
            keyword = pattern.tempSubString(partStart->getLimit(), partLimit->getIndex() - partStart->getLimit());
        }
    }
    if (matchedIndex >= 0) {
        pos.setBeginIndex(matchedIndex);
        pos.setEndIndex(matchedIndex + matchedWord.length());
        result.setString(keyword);
        return;
    }

    // Not found!
    pos.setBeginIndex(-1);
    pos.setEndIndex(-1);
}

PluralFormat::PluralSelector::~PluralSelector() {}

PluralFormat::PluralSelectorAdapter::~PluralSelectorAdapter() {
    delete pluralRules;
}

UnicodeString PluralFormat::PluralSelectorAdapter::select(void *context, double number,
                                                          UErrorCode& /*ec*/) const {
    (void)number;  // unused except in the assertion
    IFixedDecimal *dec=static_cast<IFixedDecimal *>(context);
    return pluralRules->select(*dec);
}

void PluralFormat::PluralSelectorAdapter::reset() {
    delete pluralRules;
    pluralRules = NULL;
}


U_NAMESPACE_END


#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File plurrule.cpp
*/

#include <math.h>
#include <stdio.h>


























#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

using icu::number::impl::DecimalQuantity;

static const UChar PLURAL_KEYWORD_OTHER[]={PLURRULE_LOW_O,PLURRULE_LOW_T,PLURRULE_LOW_H,PLURRULE_LOW_E,PLURRULE_LOW_R,0};
static const UChar PLURAL_DEFAULT_RULE[]={PLURRULE_LOW_O,PLURRULE_LOW_T,PLURRULE_LOW_H,PLURRULE_LOW_E,PLURRULE_LOW_R,PLURRULE_COLON,PLURRULE_SPACE,PLURRULE_LOW_N,0};
static const UChar PK_IN[]={PLURRULE_LOW_I,PLURRULE_LOW_N,0};
static const UChar PK_NOT[]={PLURRULE_LOW_N,PLURRULE_LOW_O,PLURRULE_LOW_T,0};
static const UChar PK_IS[]={PLURRULE_LOW_I,PLURRULE_LOW_S,0};
static const UChar PK_MOD[]={PLURRULE_LOW_M,PLURRULE_LOW_O,PLURRULE_LOW_D,0};
static const UChar PK_AND[]={PLURRULE_LOW_A,PLURRULE_LOW_N,PLURRULE_LOW_D,0};
static const UChar PK_OR[]={PLURRULE_LOW_O,PLURRULE_LOW_R,0};
static const UChar PK_VAR_N[]={PLURRULE_LOW_N,0};
static const UChar PK_VAR_I[]={PLURRULE_LOW_I,0};
static const UChar PK_VAR_F[]={PLURRULE_LOW_F,0};
static const UChar PK_VAR_T[]={PLURRULE_LOW_T,0};
static const UChar PK_VAR_V[]={PLURRULE_LOW_V,0};
static const UChar PK_WITHIN[]={PLURRULE_LOW_W,PLURRULE_LOW_I,PLURRULE_LOW_T,PLURRULE_LOW_H,PLURRULE_LOW_I,PLURRULE_LOW_N,0};
static const UChar PK_DECIMAL[]={PLURRULE_LOW_D,PLURRULE_LOW_E,PLURRULE_LOW_C,PLURRULE_LOW_I,PLURRULE_LOW_M,PLURRULE_LOW_A,PLURRULE_LOW_L,0};
static const UChar PK_INTEGER[]={PLURRULE_LOW_I,PLURRULE_LOW_N,PLURRULE_LOW_T,PLURRULE_LOW_E,PLURRULE_LOW_G,PLURRULE_LOW_E,PLURRULE_LOW_R,0};

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(PluralRules)
UOBJECT_DEFINE_RTTI_IMPLEMENTATION(PluralKeywordEnumeration)

PluralRules::PluralRules(UErrorCode& /*status*/)
:   UObject(),
    mRules(nullptr),
    mInternalStatus(U_ZERO_ERROR)
{
}

PluralRules::PluralRules(const PluralRules& other)
: UObject(other),
    mRules(nullptr),
    mInternalStatus(U_ZERO_ERROR)
{
    *this=other;
}

PluralRules::~PluralRules() {
    delete mRules;
}

SharedPluralRules::~SharedPluralRules() {
    delete ptr;
}

PluralRules*
PluralRules::clone() const {
    PluralRules* newObj = new PluralRules(*this);
    // Since clone doesn't have a 'status' parameter, the best we can do is return nullptr if
    // the newly created object was not fully constructed properly (an error occurred).
    if (newObj != nullptr && U_FAILURE(newObj->mInternalStatus)) {
        delete newObj;
        newObj = nullptr;
    }
    return newObj;
}

PluralRules&
PluralRules::operator=(const PluralRules& other) {
    if (this != &other) {
        delete mRules;
        mRules = nullptr;
        mInternalStatus = other.mInternalStatus;
        if (U_FAILURE(mInternalStatus)) {
            // bail out early if the object we were copying from was already 'invalid'.
            return *this;
        }
        if (other.mRules != nullptr) {
            mRules = new RuleChain(*other.mRules);
            if (mRules == nullptr) {
                mInternalStatus = U_MEMORY_ALLOCATION_ERROR;
            }
            else if (U_FAILURE(mRules->fInternalStatus)) {
                // If the RuleChain wasn't fully copied, then set our status to failure as well.
                mInternalStatus = mRules->fInternalStatus;
            }
        }
    }
    return *this;
}

StringEnumeration* PluralRules::getAvailableLocales(UErrorCode &status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    LocalPointer<StringEnumeration> result(new PluralAvailableLocalesEnumeration(status), status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    return result.orphan();
}


PluralRules* U_EXPORT2
PluralRules::createRules(const UnicodeString& description, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    PluralRuleParser parser;
    LocalPointer<PluralRules> newRules(new PluralRules(status), status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    parser.parse(description, newRules.getAlias(), status);
    if (U_FAILURE(status)) {
        newRules.adoptInstead(nullptr);
    }
    return newRules.orphan();
}


PluralRules* U_EXPORT2
PluralRules::createDefaultRules(UErrorCode& status) {
    return createRules(UnicodeString(TRUE, PLURAL_DEFAULT_RULE, -1), status);
}

/******************************************************************************/
/* Create PluralRules cache */

template<> U_I18N_API
const SharedPluralRules *LocaleCacheKey<SharedPluralRules>::createObject(
        const void * /*unused*/, UErrorCode &status) const {
    const char *localeId = fLoc.getName();
    LocalPointer<PluralRules> pr(PluralRules::internalForLocale(localeId, UPLURAL_TYPE_CARDINAL, status), status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    LocalPointer<SharedPluralRules> result(new SharedPluralRules(pr.getAlias()), status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    pr.orphan(); // result was successfully created so it nows pr.
    result->addRef();
    return result.orphan();
}

/* end plural rules cache */
/******************************************************************************/

const SharedPluralRules* U_EXPORT2
PluralRules::createSharedInstance(
        const Locale& locale, UPluralType type, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (type != UPLURAL_TYPE_CARDINAL) {
        status = U_UNSUPPORTED_ERROR;
        return nullptr;
    }
    const SharedPluralRules *result = nullptr;
    UnifiedCache::getByLocale(locale, result, status);
    return result;
}

PluralRules* U_EXPORT2
PluralRules::forLocale(const Locale& locale, UErrorCode& status) {
    return forLocale(locale, UPLURAL_TYPE_CARDINAL, status);
}

PluralRules* U_EXPORT2
PluralRules::forLocale(const Locale& locale, UPluralType type, UErrorCode& status) {
    if (type != UPLURAL_TYPE_CARDINAL) {
        return internalForLocale(locale, type, status);
    }
    const SharedPluralRules *shared = createSharedInstance(
            locale, type, status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    PluralRules *result = (*shared)->clone();
    shared->removeRef();
    if (result == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return result;
}

PluralRules* U_EXPORT2
PluralRules::internalForLocale(const Locale& locale, UPluralType type, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (type >= UPLURAL_TYPE_COUNT) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    LocalPointer<PluralRules> newObj(new PluralRules(status), status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    UnicodeString locRule = newObj->getRuleFromResource(locale, type, status);
    // TODO: which other errors, if any, should be returned?
    if (locRule.length() == 0) {
        // If an out-of-memory error occurred, then stop and report the failure.
        if (status == U_MEMORY_ALLOCATION_ERROR) {
            return nullptr;
        }
        // Locales with no specific rules (all numbers have the "other" category
        //   will return a U_MISSING_RESOURCE_ERROR at this point. This is not
        //   an error.
        locRule =  UnicodeString(PLURAL_DEFAULT_RULE);
        status = U_ZERO_ERROR;
    }
    PluralRuleParser parser;
    parser.parse(locRule, newObj.getAlias(), status);
        //  TODO: should rule parse errors be returned, or
        //        should we silently use default rules?
        //        Original impl used default rules.
        //        Ask the question to ICU Core.

    return newObj.orphan();
}

UnicodeString
PluralRules::select(int32_t number) const {
    return select(FixedDecimal(number));
}

UnicodeString
PluralRules::select(double number) const {
    return select(FixedDecimal(number));
}

UnicodeString
PluralRules::select(const number::FormattedNumber& number, UErrorCode& status) const {
    DecimalQuantity dq;
    number.getDecimalQuantity(dq, status);
    if (U_FAILURE(status)) {
        return ICU_Utility::makeBogusString();
    }
    return select(dq);
}

UnicodeString
PluralRules::select(const IFixedDecimal &number) const {
    if (mRules == nullptr) {
        return UnicodeString(TRUE, PLURAL_DEFAULT_RULE, -1);
    }
    else {
        return mRules->select(number);
    }
}



StringEnumeration*
PluralRules::getKeywords(UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (U_FAILURE(mInternalStatus)) {
        status = mInternalStatus;
        return nullptr;
    }
    LocalPointer<StringEnumeration> nameEnumerator(new PluralKeywordEnumeration(mRules, status), status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    return nameEnumerator.orphan();
}

double
PluralRules::getUniqueKeywordValue(const UnicodeString& /* keyword */) {
  // Not Implemented.
  return UPLRULES_NO_UNIQUE_VALUE;
}

int32_t
PluralRules::getAllKeywordValues(const UnicodeString & /* keyword */, double * /* dest */,
                                 int32_t /* destCapacity */, UErrorCode& error) {
    error = U_UNSUPPORTED_ERROR;
    return 0;
}


static double scaleForInt(double d) {
    double scale = 1.0;
    while (d != floor(d)) {
        d = d * 10.0;
        scale = scale * 10.0;
    }
    return scale;
}

static int32_t
getSamplesFromString(const UnicodeString &samples, double *dest,
                        int32_t destCapacity, UErrorCode& status) {
    int32_t sampleCount = 0;
    int32_t sampleStartIdx = 0;
    int32_t sampleEndIdx = 0;

    //std::string ss;  // TODO: debugging.
    // std::cout << "PluralRules::getSamples(), samples = \"" << samples.toUTF8String(ss) << "\"\n";
    for (sampleCount = 0; sampleCount < destCapacity && sampleStartIdx < samples.length(); ) {
        sampleEndIdx = samples.indexOf(PLURRULE_COMMA, sampleStartIdx);
        if (sampleEndIdx == -1) {
            sampleEndIdx = samples.length();
        }
        const UnicodeString &sampleRange = samples.tempSubStringBetween(sampleStartIdx, sampleEndIdx);
        // ss.erase();
        // std::cout << "PluralRules::getSamples(), samplesRange = \"" << sampleRange.toUTF8String(ss) << "\"\n";
        int32_t tildeIndex = sampleRange.indexOf(PLURRULE_TILDE);
        if (tildeIndex < 0) {
            FixedDecimal fixed(sampleRange, status);
            double sampleValue = fixed.source;
            if (fixed.visibleDecimalDigitCount == 0 || sampleValue != floor(sampleValue)) {
                dest[sampleCount++] = sampleValue;
            }
        } else {

            FixedDecimal fixedLo(sampleRange.tempSubStringBetween(0, tildeIndex), status);
            FixedDecimal fixedHi(sampleRange.tempSubStringBetween(tildeIndex+1), status);
            double rangeLo = fixedLo.source;
            double rangeHi = fixedHi.source;
            if (U_FAILURE(status)) {
                break;
            }
            if (rangeHi < rangeLo) {
                status = U_INVALID_FORMAT_ERROR;
                break;
            }

            // For ranges of samples with fraction decimal digits, scale the number up so that we
            //   are adding one in the units place. Avoids roundoffs from repetitive adds of tenths.

            double scale = scaleForInt(rangeLo);
            double t = scaleForInt(rangeHi);
            if (t > scale) {
                scale = t;
            }
            rangeLo *= scale;
            rangeHi *= scale;
            for (double n=rangeLo; n<=rangeHi; n+=1) {
                // Hack Alert: don't return any decimal samples with integer values that
                //    originated from a format with trailing decimals.
                //    This API is returning doubles, which can't distinguish having displayed
                //    zeros to the right of the decimal.
                //    This results in test failures with values mapping back to a different keyword.
                double sampleValue = n/scale;
                if (!(sampleValue == floor(sampleValue) && fixedLo.visibleDecimalDigitCount > 0)) {
                    dest[sampleCount++] = sampleValue;
                }
                if (sampleCount >= destCapacity) {
                    break;
                }
            }
        }
        sampleStartIdx = sampleEndIdx + 1;
    }
    return sampleCount;
}


int32_t
PluralRules::getSamples(const UnicodeString &keyword, double *dest,
                        int32_t destCapacity, UErrorCode& status) {
    if (destCapacity == 0 || U_FAILURE(status)) {
        return 0;
    }
    if (U_FAILURE(mInternalStatus)) {
        status = mInternalStatus;
        return 0;
    }
    RuleChain *rc = rulesForKeyword(keyword);
    if (rc == nullptr) {
        return 0;
    }
    int32_t numSamples = getSamplesFromString(rc->fIntegerSamples, dest, destCapacity, status);
    if (numSamples == 0) {
        numSamples = getSamplesFromString(rc->fDecimalSamples, dest, destCapacity, status);
    }
    return numSamples;
}


RuleChain *PluralRules::rulesForKeyword(const UnicodeString &keyword) const {
    RuleChain *rc;
    for (rc = mRules; rc != nullptr; rc = rc->fNext) {
        if (rc->fKeyword == keyword) {
            break;
        }
    }
    return rc;
}


UBool
PluralRules::isKeyword(const UnicodeString& keyword) const {
    if (0 == keyword.compare(PLURAL_KEYWORD_OTHER, 5)) {
        return true;
    }
    return rulesForKeyword(keyword) != nullptr;
}

UnicodeString
PluralRules::getKeywordOther() const {
    return UnicodeString(TRUE, PLURAL_KEYWORD_OTHER, 5);
}

UBool
PluralRules::operator==(const PluralRules& other) const  {
    const UnicodeString *ptrKeyword;
    UErrorCode status= U_ZERO_ERROR;

    if ( this == &other ) {
        return TRUE;
    }
    LocalPointer<StringEnumeration> myKeywordList(getKeywords(status));
    LocalPointer<StringEnumeration> otherKeywordList(other.getKeywords(status));
    if (U_FAILURE(status)) {
        return FALSE;
    }

    if (myKeywordList->count(status)!=otherKeywordList->count(status)) {
        return FALSE;
    }
    myKeywordList->reset(status);
    while ((ptrKeyword=myKeywordList->snext(status))!=nullptr) {
        if (!other.isKeyword(*ptrKeyword)) {
            return FALSE;
        }
    }
    otherKeywordList->reset(status);
    while ((ptrKeyword=otherKeywordList->snext(status))!=nullptr) {
        if (!this->isKeyword(*ptrKeyword)) {
            return FALSE;
        }
    }
    if (U_FAILURE(status)) {
        return FALSE;
    }

    return TRUE;
}


void
PluralRuleParser::parse(const UnicodeString& ruleData, PluralRules *prules, UErrorCode &status)
{
    if (U_FAILURE(status)) {
        return;
    }
    U_ASSERT(ruleIndex == 0);    // Parsers are good for a single use only!
    ruleSrc = &ruleData;

    while (ruleIndex< ruleSrc->length()) {
        getNextToken(status);
        if (U_FAILURE(status)) {
            return;
        }
        checkSyntax(status);
        if (U_FAILURE(status)) {
            return;
        }
        switch (type) {
        case plurrule_token_tAnd:
            U_ASSERT(curAndConstraint != nullptr);
            curAndConstraint = curAndConstraint->add(status);
            break;
        case plurrule_token_tOr:
            {
                U_ASSERT(currentChain != nullptr);
                OrConstraint *orNode=currentChain->ruleHeader;
                while (orNode->next != nullptr) {
                    orNode = orNode->next;
                }
                orNode->next= new OrConstraint();
                if (orNode->next == nullptr) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                    break;
                }
                orNode=orNode->next;
                orNode->next=nullptr;
                curAndConstraint = orNode->add(status);
            }
            break;
        case plurrule_token_tIs:
            U_ASSERT(curAndConstraint != nullptr);
            U_ASSERT(curAndConstraint->value == -1);
            U_ASSERT(curAndConstraint->rangeList == nullptr);
            break;
        case plurrule_token_tNot:
            U_ASSERT(curAndConstraint != nullptr);
            curAndConstraint->negated=TRUE;
            break;

        case plurrule_token_tNotEqual:
            curAndConstraint->negated=TRUE;
            U_FALLTHROUGH;
        case plurrule_token_tIn:
        case plurrule_token_tWithin:
        case plurrule_token_tEqual:
            {
                U_ASSERT(curAndConstraint != nullptr);
                LocalPointer<UVector32> newRangeList(new UVector32(status), status);
                if (U_FAILURE(status)) {
                    break;
                }
                curAndConstraint->rangeList = newRangeList.orphan();
                curAndConstraint->rangeList->addElement(-1, status);  // range Low
                curAndConstraint->rangeList->addElement(-1, status);  // range Hi
                rangeLowIdx = 0;
                rangeHiIdx  = 1;
                curAndConstraint->value=PLURAL_RANGE_HIGH;
                curAndConstraint->integerOnly = (type != plurrule_token_tWithin);
            }
            break;
        case plurrule_token_tNumber:
            U_ASSERT(curAndConstraint != nullptr);
            if ( (curAndConstraint->op==AndConstraint::MOD)&&
                 (curAndConstraint->opNum == -1 ) ) {
                curAndConstraint->opNum=getNumberValue(token);
            }
            else {
                if (curAndConstraint->rangeList == nullptr) {
                    // this is for an 'is' rule
                    curAndConstraint->value = getNumberValue(token);
                } else {
                    // this is for an 'in' or 'within' rule
                    if (curAndConstraint->rangeList->elementAti(rangeLowIdx) == -1) {
                        curAndConstraint->rangeList->setElementAt(getNumberValue(token), rangeLowIdx);
                        curAndConstraint->rangeList->setElementAt(getNumberValue(token), rangeHiIdx);
                    }
                    else {
                        curAndConstraint->rangeList->setElementAt(getNumberValue(token), rangeHiIdx);
                        if (curAndConstraint->rangeList->elementAti(rangeLowIdx) >
                                curAndConstraint->rangeList->elementAti(rangeHiIdx)) {
                            // Range Lower bound > Range Upper bound.
                            // U_UNEXPECTED_TOKEN seems a little funny, but it is consistently
                            // used for all plural rule parse errors.
                            status = U_UNEXPECTED_TOKEN;
                            break;
                        }
                    }
                }
            }
            break;
        case plurrule_token_tComma:
            // TODO: rule syntax checking is inadequate, can happen with badly formed rules.
            //       Catch cases like "n mod 10, is 1" here instead.
            if (curAndConstraint == nullptr || curAndConstraint->rangeList == nullptr) {
                status = U_UNEXPECTED_TOKEN;
                break;
            }
            U_ASSERT(curAndConstraint->rangeList->size() >= 2);
            rangeLowIdx = curAndConstraint->rangeList->size();
            curAndConstraint->rangeList->addElement(-1, status);  // range Low
            rangeHiIdx = curAndConstraint->rangeList->size();
            curAndConstraint->rangeList->addElement(-1, status);  // range Hi
            break;
        case plurrule_token_tMod:
            U_ASSERT(curAndConstraint != nullptr);
            curAndConstraint->op=AndConstraint::MOD;
            break;
        case plurrule_token_tVariableN:
        case plurrule_token_tVariableI:
        case plurrule_token_tVariableF:
        case plurrule_token_tVariableT:
        case plurrule_token_tVariableV:
            U_ASSERT(curAndConstraint != nullptr);
            curAndConstraint->digitsType = type;
            break;
        case plurrule_token_tKeyword:
            {
            RuleChain *newChain = new RuleChain;
            if (newChain == nullptr) {
                status = U_MEMORY_ALLOCATION_ERROR;
                break;
            }
            newChain->fKeyword = token;
            if (prules->mRules == nullptr) {
                prules->mRules = newChain;
            } else {
                // The new rule chain goes at the end of the linked list of rule chains,
                //   unless there is an "other" keyword & chain. "other" must remain last.
                RuleChain *insertAfter = prules->mRules;
                while (insertAfter->fNext!=nullptr &&
                       insertAfter->fNext->fKeyword.compare(PLURAL_KEYWORD_OTHER, 5) != 0 ){
                    insertAfter=insertAfter->fNext;
                }
                newChain->fNext = insertAfter->fNext;
                insertAfter->fNext = newChain;
            }
            OrConstraint *orNode = new OrConstraint();
            if (orNode == nullptr) {
                status = U_MEMORY_ALLOCATION_ERROR;
                break;
            }
            newChain->ruleHeader = orNode;
            curAndConstraint = orNode->add(status);
            currentChain = newChain;
            }
            break;

        case plurrule_token_tInteger:
            for (;;) {
                getNextToken(status);
                if (U_FAILURE(status) || type == plurrule_token_tSemiColon || type == plurrule_token_tEOF || type == plurrule_token_tAt) {
                    break;
                }
                if (type == plurrule_token_tEllipsis) {
                    currentChain->fIntegerSamplesUnbounded = TRUE;
                    continue;
                }
                currentChain->fIntegerSamples.append(token);
            }
            break;

        case plurrule_token_tDecimal:
            for (;;) {
                getNextToken(status);
                if (U_FAILURE(status) || type == plurrule_token_tSemiColon || type == plurrule_token_tEOF || type == plurrule_token_tAt) {
                    break;
                }
                if (type == plurrule_token_tEllipsis) {
                    currentChain->fDecimalSamplesUnbounded = TRUE;
                    continue;
                }
                currentChain->fDecimalSamples.append(token);
            }
            break;

        default:
            break;
        }
        prevType=type;
        if (U_FAILURE(status)) {
            break;
        }
    }
}

UnicodeString
PluralRules::getRuleFromResource(const Locale& locale, UPluralType type, UErrorCode& errCode) {
    UnicodeString emptyStr;

    if (U_FAILURE(errCode)) {
        return emptyStr;
    }
    LocalUResourceBundlePointer rb(ures_openDirect(nullptr, "plurals", &errCode));
    if(U_FAILURE(errCode)) {
        return emptyStr;
    }
    const char *typeKey;
    switch (type) {
    case UPLURAL_TYPE_CARDINAL:
        typeKey = "locales";
        break;
    case UPLURAL_TYPE_ORDINAL:
        typeKey = "locales_ordinals";
        break;
    default:
        // Must not occur: The caller should have checked for valid types.
        errCode = U_ILLEGAL_ARGUMENT_ERROR;
        return emptyStr;
    }
    LocalUResourceBundlePointer locRes(ures_getByKey(rb.getAlias(), typeKey, nullptr, &errCode));
    if(U_FAILURE(errCode)) {
        return emptyStr;
    }
    int32_t resLen=0;
    const char *curLocaleName=locale.getBaseName();
    const UChar* s = ures_getStringByKey(locRes.getAlias(), curLocaleName, &resLen, &errCode);

    if (s == nullptr) {
        // Check parent locales.
        UErrorCode status = U_ZERO_ERROR;
        char parentLocaleName[ULOC_FULLNAME_CAPACITY];
        const char *curLocaleName2=locale.getBaseName();
        uprv_strcpy(parentLocaleName, curLocaleName2);

        while (uloc_getParent(parentLocaleName, parentLocaleName,
                                       ULOC_FULLNAME_CAPACITY, &status) > 0) {
            resLen=0;
            s = ures_getStringByKey(locRes.getAlias(), parentLocaleName, &resLen, &status);
            if (s != nullptr) {
                errCode = U_ZERO_ERROR;
                break;
            }
            status = U_ZERO_ERROR;
        }
    }
    if (s==nullptr) {
        return emptyStr;
    }

    char setKey[256];
    u_UCharsToChars(s, setKey, resLen + 1);
    // printf("\n PluralRule: %s\n", setKey);

    LocalUResourceBundlePointer ruleRes(ures_getByKey(rb.getAlias(), "rules", nullptr, &errCode));
    if(U_FAILURE(errCode)) {
        return emptyStr;
    }
    LocalUResourceBundlePointer setRes(ures_getByKey(ruleRes.getAlias(), setKey, nullptr, &errCode));
    if (U_FAILURE(errCode)) {
        return emptyStr;
    }

    int32_t numberKeys = ures_getSize(setRes.getAlias());
    UnicodeString result;
    const char *key=nullptr;
    for(int32_t i=0; i<numberKeys; ++i) {   // Keys are zero, one, few, ...
        UnicodeString rules = ures_getNextUnicodeString(setRes.getAlias(), &key, &errCode);
        UnicodeString uKey(key, -1, US_INV);
        result.append(uKey);
        result.append(PLURRULE_COLON);
        result.append(rules);
        result.append(PLURRULE_SEMI_COLON);
    }
    return result;
}


UnicodeString
PluralRules::getRules() const {
    UnicodeString rules;
    if (mRules != nullptr) {
        mRules->dumpRules(rules);
    }
    return rules;
}

AndConstraint::AndConstraint(const AndConstraint& other) {
    this->fInternalStatus = other.fInternalStatus;
    if (U_FAILURE(fInternalStatus)) {
        return; // stop early if the object we are copying from is invalid.
    }
    this->op = other.op;
    this->opNum=other.opNum;
    this->value=other.value;
    if (other.rangeList != nullptr) {
        LocalPointer<UVector32> newRangeList(new UVector32(fInternalStatus), fInternalStatus);
        if (U_FAILURE(fInternalStatus)) {
            return;
        }
        this->rangeList = newRangeList.orphan();
        this->rangeList->assign(*other.rangeList, fInternalStatus);
    }
    this->integerOnly=other.integerOnly;
    this->negated=other.negated;
    this->digitsType = other.digitsType;
    if (other.next != nullptr) {
        this->next = new AndConstraint(*other.next);
        if (this->next == nullptr) {
            fInternalStatus = U_MEMORY_ALLOCATION_ERROR;
        }
    }
}

AndConstraint::~AndConstraint() {
    delete rangeList;
    rangeList = nullptr;
    delete next;
    next = nullptr;
}

UBool
AndConstraint::isFulfilled(const IFixedDecimal &number) {
    UBool result = TRUE;
    if (digitsType == plurrule_token_none) {
        // An empty AndConstraint, created by a rule with a keyword but no following expression.
        return TRUE;
    }

    PluralOperand operand = tokenTypeToPluralOperand(digitsType);
    double n = number.getPluralOperand(operand);     // pulls n | i | v | f value for the number.
                                                     // Will always be positive.
                                                     // May be non-integer (n option only)
    do {
        if (integerOnly && n != uprv_floor(n)) {
            result = FALSE;
            break;
        }

        if (op == MOD) {
            n = fmod(n, opNum);
        }
        if (rangeList == nullptr) {
            result = value == -1 ||    // empty rule
                     n == value;       //  'is' rule
            break;
        }
        result = FALSE;                // 'in' or 'within' rule
        for (int32_t r=0; r<rangeList->size(); r+=2) {
            if (rangeList->elementAti(r) <= n && n <= rangeList->elementAti(r+1)) {
                result = TRUE;
                break;
            }
        }
    } while (FALSE);

    if (negated) {
        result = !result;
    }
    return result;
}

AndConstraint*
AndConstraint::add(UErrorCode& status) {
    if (U_FAILURE(fInternalStatus)) {
        status = fInternalStatus;
        return nullptr;
    }
    this->next = new AndConstraint();
    if (this->next == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return this->next;
}


OrConstraint::OrConstraint(const OrConstraint& other) {
    this->fInternalStatus = other.fInternalStatus;
    if (U_FAILURE(fInternalStatus)) {
        return; // stop early if the object we are copying from is invalid.
    }
    if ( other.childNode != nullptr ) {
        this->childNode = new AndConstraint(*(other.childNode));
        if (this->childNode == nullptr) {
            fInternalStatus = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
    }
    if (other.next != nullptr ) {
        this->next = new OrConstraint(*(other.next));
        if (this->next == nullptr) {
            fInternalStatus = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        if (U_FAILURE(this->next->fInternalStatus)) {
            this->fInternalStatus = this->next->fInternalStatus;
        }
    }
}

OrConstraint::~OrConstraint() {
    delete childNode;
    childNode = nullptr;
    delete next;
    next = nullptr;
}

AndConstraint*
OrConstraint::add(UErrorCode& status) {
    if (U_FAILURE(fInternalStatus)) {
        status = fInternalStatus;
        return nullptr;
    }
    OrConstraint *curOrConstraint=this;
    {
        while (curOrConstraint->next!=nullptr) {
            curOrConstraint = curOrConstraint->next;
        }
        U_ASSERT(curOrConstraint->childNode == nullptr);
        curOrConstraint->childNode = new AndConstraint();
        if (curOrConstraint->childNode == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
        }
    }
    return curOrConstraint->childNode;
}

UBool
OrConstraint::isFulfilled(const IFixedDecimal &number) {
    OrConstraint* orRule=this;
    UBool result=FALSE;

    while (orRule!=nullptr && !result) {
        result=TRUE;
        AndConstraint* andRule = orRule->childNode;
        while (andRule!=nullptr && result) {
            result = andRule->isFulfilled(number);
            andRule=andRule->next;
        }
        orRule = orRule->next;
    }

    return result;
}


RuleChain::RuleChain(const RuleChain& other) :
        fKeyword(other.fKeyword), fDecimalSamples(other.fDecimalSamples),
        fIntegerSamples(other.fIntegerSamples), fDecimalSamplesUnbounded(other.fDecimalSamplesUnbounded),
        fIntegerSamplesUnbounded(other.fIntegerSamplesUnbounded), fInternalStatus(other.fInternalStatus) {
    if (U_FAILURE(this->fInternalStatus)) {
        return; // stop early if the object we are copying from is invalid.
    }
    if (other.ruleHeader != nullptr) {
        this->ruleHeader = new OrConstraint(*(other.ruleHeader));
        if (this->ruleHeader == nullptr) {
            this->fInternalStatus = U_MEMORY_ALLOCATION_ERROR;
        }
        else if (U_FAILURE(this->ruleHeader->fInternalStatus)) {
            // If the OrConstraint wasn't fully copied, then set our status to failure as well.
            this->fInternalStatus = this->ruleHeader->fInternalStatus;
            return; // exit early.
        }
    }
    if (other.fNext != nullptr ) {
        this->fNext = new RuleChain(*other.fNext);
        if (this->fNext == nullptr) {
            this->fInternalStatus = U_MEMORY_ALLOCATION_ERROR;
        }
        else if (U_FAILURE(this->fNext->fInternalStatus)) {
            // If the RuleChain wasn't fully copied, then set our status to failure as well.
            this->fInternalStatus = this->fNext->fInternalStatus;
        }
    }
}

RuleChain::~RuleChain() {
    delete fNext;
    delete ruleHeader;
}

UnicodeString
RuleChain::select(const IFixedDecimal &number) const {
    if (!number.isNaN() && !number.isInfinite()) {
        for (const RuleChain *rules = this; rules != nullptr; rules = rules->fNext) {
             if (rules->ruleHeader->isFulfilled(number)) {
                 return rules->fKeyword;
             }
        }
    }
    return UnicodeString(TRUE, PLURAL_KEYWORD_OTHER, 5);
}

static UnicodeString tokenString(tokenType tok) {
    UnicodeString s;
    switch (tok) {
      case plurrule_token_tVariableN:
        s.append(PLURRULE_LOW_N); break;
      case plurrule_token_tVariableI:
        s.append(PLURRULE_LOW_I); break;
      case plurrule_token_tVariableF:
        s.append(PLURRULE_LOW_F); break;
      case plurrule_token_tVariableV:
        s.append(PLURRULE_LOW_V); break;
      case plurrule_token_tVariableT:
        s.append(PLURRULE_LOW_T); break;
      default:
        s.append(PLURRULE_TILDE);
    }
    return s;
}

void
RuleChain::dumpRules(UnicodeString& result) {
    UChar digitString[16];

    if ( ruleHeader != nullptr ) {
        result +=  fKeyword;
        result += PLURRULE_COLON;
        result += PLURRULE_SPACE;
        OrConstraint* orRule=ruleHeader;
        while ( orRule != nullptr ) {
            AndConstraint* andRule=orRule->childNode;
            while ( andRule != nullptr ) {
                if ((andRule->op==AndConstraint::NONE) &&  (andRule->rangeList==nullptr) && (andRule->value == -1)) {
                    // Empty Rules.
                } else if ( (andRule->op==AndConstraint::NONE) && (andRule->rangeList==nullptr) ) {
                    result += tokenString(andRule->digitsType);
                    result += UNICODE_STRING_SIMPLE(" is ");
                    if (andRule->negated) {
                        result += UNICODE_STRING_SIMPLE("not ");
                    }
                    uprv_itou(digitString,16, andRule->value,10,0);
                    result += UnicodeString(digitString);
                }
                else {
                    result += tokenString(andRule->digitsType);
                    result += PLURRULE_SPACE;
                    if (andRule->op==AndConstraint::MOD) {
                        result += UNICODE_STRING_SIMPLE("mod ");
                        uprv_itou(digitString,16, andRule->opNum,10,0);
                        result += UnicodeString(digitString);
                    }
                    if (andRule->rangeList==nullptr) {
                        if (andRule->negated) {
                            result += UNICODE_STRING_SIMPLE(" is not ");
                            uprv_itou(digitString,16, andRule->value,10,0);
                            result += UnicodeString(digitString);
                        }
                        else {
                            result += UNICODE_STRING_SIMPLE(" is ");
                            uprv_itou(digitString,16, andRule->value,10,0);
                            result += UnicodeString(digitString);
                        }
                    }
                    else {
                        if (andRule->negated) {
                            if ( andRule->integerOnly ) {
                                result += UNICODE_STRING_SIMPLE(" not in ");
                            }
                            else {
                                result += UNICODE_STRING_SIMPLE(" not within ");
                            }
                        }
                        else {
                            if ( andRule->integerOnly ) {
                                result += UNICODE_STRING_SIMPLE(" in ");
                            }
                            else {
                                result += UNICODE_STRING_SIMPLE(" within ");
                            }
                        }
                        for (int32_t r=0; r<andRule->rangeList->size(); r+=2) {
                            int32_t rangeLo = andRule->rangeList->elementAti(r);
                            int32_t rangeHi = andRule->rangeList->elementAti(r+1);
                            uprv_itou(digitString,16, rangeLo, 10, 0);
                            result += UnicodeString(digitString);
                            result += UNICODE_STRING_SIMPLE("..");
                            uprv_itou(digitString,16, rangeHi, 10,0);
                            result += UnicodeString(digitString);
                            if (r+2 < andRule->rangeList->size()) {
                                result += UNICODE_STRING_SIMPLE(", ");
                            }
                        }
                    }
                }
                if ( (andRule=andRule->next) != nullptr) {
                    result += UNICODE_STRING_SIMPLE(" and ");
                }
            }
            if ( (orRule = orRule->next) != nullptr ) {
                result += UNICODE_STRING_SIMPLE(" or ");
            }
        }
    }
    if ( fNext != nullptr ) {
        result += UNICODE_STRING_SIMPLE("; ");
        fNext->dumpRules(result);
    }
}


UErrorCode
RuleChain::getKeywords(int32_t capacityOfKeywords, UnicodeString* keywords, int32_t& arraySize) const {
    if (U_FAILURE(fInternalStatus)) {
        return fInternalStatus;
    }
    if ( arraySize < capacityOfKeywords-1 ) {
        keywords[arraySize++]=fKeyword;
    }
    else {
        return U_BUFFER_OVERFLOW_ERROR;
    }

    if ( fNext != nullptr ) {
        return fNext->getKeywords(capacityOfKeywords, keywords, arraySize);
    }
    else {
        return U_ZERO_ERROR;
    }
}

UBool
RuleChain::isKeyword(const UnicodeString& keywordParam) const {
    if ( fKeyword == keywordParam ) {
        return TRUE;
    }

    if ( fNext != nullptr ) {
        return fNext->isKeyword(keywordParam);
    }
    else {
        return FALSE;
    }
}


PluralRuleParser::PluralRuleParser() :
        ruleIndex(0), token(), type(plurrule_token_none), prevType(plurrule_token_none),
        curAndConstraint(nullptr), currentChain(nullptr), rangeLowIdx(-1), rangeHiIdx(-1)
{
}

PluralRuleParser::~PluralRuleParser() {
}


int32_t
PluralRuleParser::getNumberValue(const UnicodeString& token) {
    int32_t i;
    char digits[128];

    i = token.extract(0, token.length(), digits, UPRV_LENGTHOF(digits), US_INV);
    digits[i]='\0';

    return((int32_t)atoi(digits));
}


void
PluralRuleParser::checkSyntax(UErrorCode &status)
{
    if (U_FAILURE(status)) {
        return;
    }
    if (!(prevType==plurrule_token_none || prevType==plurrule_token_tSemiColon)) {
        type = getKeyType(token, type);  // Switch token type from plurrule_token_tKeyword if we scanned a reserved word,
                                               //   and we are not at the start of a rule, where a
                                               //   keyword is expected.
    }

    switch(prevType) {
    case plurrule_token_none:
    case plurrule_token_tSemiColon:
        if (type!=plurrule_token_tKeyword && type != plurrule_token_tEOF) {
            status = U_UNEXPECTED_TOKEN;
        }
        break;
    case plurrule_token_tVariableN:
    case plurrule_token_tVariableI:
    case plurrule_token_tVariableF:
    case plurrule_token_tVariableT:
    case plurrule_token_tVariableV:
        if (type != plurrule_token_tIs && type != plurrule_token_tMod && type != plurrule_token_tIn &&
            type != plurrule_token_tNot && type != plurrule_token_tWithin && type != plurrule_token_tEqual && type != plurrule_token_tNotEqual) {
            status = U_UNEXPECTED_TOKEN;
        }
        break;
    case plurrule_token_tKeyword:
        if (type != plurrule_token_tColon) {
            status = U_UNEXPECTED_TOKEN;
        }
        break;
    case plurrule_token_tColon:
        if (!(type == plurrule_token_tVariableN ||
              type == plurrule_token_tVariableI ||
              type == plurrule_token_tVariableF ||
              type == plurrule_token_tVariableT ||
              type == plurrule_token_tVariableV ||
              type == plurrule_token_tAt)) {
            status = U_UNEXPECTED_TOKEN;
        }
        break;
    case plurrule_token_tIs:
        if ( type != plurrule_token_tNumber && type != plurrule_token_tNot) {
            status = U_UNEXPECTED_TOKEN;
        }
        break;
    case plurrule_token_tNot:
        if (type != plurrule_token_tNumber && type != plurrule_token_tIn && type != plurrule_token_tWithin) {
            status = U_UNEXPECTED_TOKEN;
        }
        break;
    case plurrule_token_tMod:
    case plurrule_token_tDot2:
    case plurrule_token_tIn:
    case plurrule_token_tWithin:
    case plurrule_token_tEqual:
    case plurrule_token_tNotEqual:
        if (type != plurrule_token_tNumber) {
            status = U_UNEXPECTED_TOKEN;
        }
        break;
    case plurrule_token_tAnd:
    case plurrule_token_tOr:
        if ( type != plurrule_token_tVariableN &&
             type != plurrule_token_tVariableI &&
             type != plurrule_token_tVariableF &&
             type != plurrule_token_tVariableT &&
             type != plurrule_token_tVariableV) {
            status = U_UNEXPECTED_TOKEN;
        }
        break;
    case plurrule_token_tComma:
        if (type != plurrule_token_tNumber) {
            status = U_UNEXPECTED_TOKEN;
        }
        break;
    case plurrule_token_tNumber:
        if (type != plurrule_token_tDot2  && type != plurrule_token_tSemiColon && type != plurrule_token_tIs       && type != plurrule_token_tNot    &&
            type != plurrule_token_tIn    && type != plurrule_token_tEqual     && type != plurrule_token_tNotEqual && type != plurrule_token_tWithin &&
            type != plurrule_token_tAnd   && type != plurrule_token_tOr        && type != plurrule_token_tComma    && type != plurrule_token_tAt     &&
            type != plurrule_token_tEOF)
        {
            status = U_UNEXPECTED_TOKEN;
        }
        // TODO: a comma following a number that is not part of a range will be allowed.
        //       It's not the only case of this sort of thing. Parser needs a re-write.
        break;
    case plurrule_token_tAt:
        if (type != plurrule_token_tDecimal && type != plurrule_token_tInteger) {
            status = U_UNEXPECTED_TOKEN;
        }
        break;
    default:
        status = U_UNEXPECTED_TOKEN;
        break;
    }
}


/*
 *  Scan the next token from the input rules.
 *     rules and returned token type are in the parser state variables.
 */
void
PluralRuleParser::getNextToken(UErrorCode &status)
{
    if (U_FAILURE(status)) {
        return;
    }

    UChar ch;
    while (ruleIndex < ruleSrc->length()) {
        ch = ruleSrc->charAt(ruleIndex);
        type = charType(ch);
        if (type != plurrule_token_tSpace) {
            break;
        }
        ++(ruleIndex);
    }
    if (ruleIndex >= ruleSrc->length()) {
        type = plurrule_token_tEOF;
        return;
    }
    int32_t curIndex= ruleIndex;

    switch (type) {
      case plurrule_token_tColon:
      case plurrule_token_tSemiColon:
      case plurrule_token_tComma:
      case plurrule_token_tEllipsis:
      case plurrule_token_tTilde:   // scanned '~'
      case plurrule_token_tAt:      // scanned '@'
      case plurrule_token_tEqual:   // scanned '='
      case plurrule_token_tMod:     // scanned '%'
        // Single character tokens.
        ++curIndex;
        break;

      case plurrule_token_tNotEqual:  // scanned '!'
        if (ruleSrc->charAt(curIndex+1) == PLURRULE_EQUALS) {
            curIndex += 2;
        } else {
            type = plurrule_token_none;
            curIndex += 1;
        }
        break;

      case plurrule_token_tKeyword:
         while (type == plurrule_token_tKeyword && ++curIndex < ruleSrc->length()) {
             ch = ruleSrc->charAt(curIndex);
             type = charType(ch);
         }
         type = plurrule_token_tKeyword;
         break;

      case plurrule_token_tNumber:
         while (type == plurrule_token_tNumber && ++curIndex < ruleSrc->length()) {
             ch = ruleSrc->charAt(curIndex);
             type = charType(ch);
         }
         type = plurrule_token_tNumber;
         break;

       case plurrule_token_tDot:
         // We could be looking at either ".." in a range, or "..." at the end of a sample.
         if (curIndex+1 >= ruleSrc->length() || ruleSrc->charAt(curIndex+1) != PLURRULE_DOT) {
             ++curIndex;
             break; // Single dot
         }
         if (curIndex+2 >= ruleSrc->length() || ruleSrc->charAt(curIndex+2) != PLURRULE_DOT) {
             curIndex += 2;
             type = plurrule_token_tDot2;
             break; // double dot
         }
         type = plurrule_token_tEllipsis;
         curIndex += 3;
         break;     // triple dot

       default:
         status = U_UNEXPECTED_TOKEN;
         ++curIndex;
         break;
    }

    U_ASSERT(ruleIndex <= ruleSrc->length());
    U_ASSERT(curIndex <= ruleSrc->length());
    token=UnicodeString(*ruleSrc, ruleIndex, curIndex-ruleIndex);
    ruleIndex = curIndex;
}

tokenType
PluralRuleParser::charType(UChar ch) {
    if ((ch>=PLURRULE_U_ZERO) && (ch<=PLURRULE_U_NINE)) {
        return plurrule_token_tNumber;
    }
    if (ch>=PLURRULE_LOW_A && ch<=PLURRULE_LOW_Z) {
        return plurrule_token_tKeyword;
    }
    switch (ch) {
    case PLURRULE_COLON:
        return plurrule_token_tColon;
    case PLURRULE_SPACE:
        return plurrule_token_tSpace;
    case PLURRULE_SEMI_COLON:
        return plurrule_token_tSemiColon;
    case PLURRULE_DOT:
        return plurrule_token_tDot;
    case PLURRULE_COMMA:
        return plurrule_token_tComma;
    case PLURRULE_EXCLAMATION:
        return plurrule_token_tNotEqual;
    case PLURRULE_EQUALS:
        return plurrule_token_tEqual;
    case PLURRULE_PERCENT_SIGN:
        return plurrule_token_tMod;
    case PLURRULE_AT:
        return plurrule_token_tAt;
    case PLURRULE_ELLIPSIS:
        return plurrule_token_tEllipsis;
    case PLURRULE_TILDE:
        return plurrule_token_tTilde;
    default :
        return plurrule_token_none;
    }
}


//  Set token type for reserved words in the Plural Rule syntax.

tokenType
PluralRuleParser::getKeyType(const UnicodeString &token, tokenType keyType)
{
    if (keyType != plurrule_token_tKeyword) {
        return keyType;
    }

    if (0 == token.compare(PK_VAR_N, 1)) {
        keyType = plurrule_token_tVariableN;
    } else if (0 == token.compare(PK_VAR_I, 1)) {
        keyType = plurrule_token_tVariableI;
    } else if (0 == token.compare(PK_VAR_F, 1)) {
        keyType = plurrule_token_tVariableF;
    } else if (0 == token.compare(PK_VAR_T, 1)) {
        keyType = plurrule_token_tVariableT;
    } else if (0 == token.compare(PK_VAR_V, 1)) {
        keyType = plurrule_token_tVariableV;
    } else if (0 == token.compare(PK_IS, 2)) {
        keyType = plurrule_token_tIs;
    } else if (0 == token.compare(PK_AND, 3)) {
        keyType = plurrule_token_tAnd;
    } else if (0 == token.compare(PK_IN, 2)) {
        keyType = plurrule_token_tIn;
    } else if (0 == token.compare(PK_WITHIN, 6)) {
        keyType = plurrule_token_tWithin;
    } else if (0 == token.compare(PK_NOT, 3)) {
        keyType = plurrule_token_tNot;
    } else if (0 == token.compare(PK_MOD, 3)) {
        keyType = plurrule_token_tMod;
    } else if (0 == token.compare(PK_OR, 2)) {
        keyType = plurrule_token_tOr;
    } else if (0 == token.compare(PK_DECIMAL, 7)) {
        keyType = plurrule_token_tDecimal;
    } else if (0 == token.compare(PK_INTEGER, 7)) {
        keyType = plurrule_token_tInteger;
    }
    return keyType;
}


PluralKeywordEnumeration::PluralKeywordEnumeration(RuleChain *header, UErrorCode& status)
        : pos(0), fKeywordNames(status) {
    if (U_FAILURE(status)) {
        return;
    }
    fKeywordNames.setDeleter(uprv_deleteUObject);
    UBool  addKeywordOther = TRUE;
    RuleChain *node = header;
    while (node != nullptr) {
        auto newElem = new UnicodeString(node->fKeyword);
        if (newElem == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        fKeywordNames.addElement(newElem, status);
        if (U_FAILURE(status)) {
            delete newElem;
            return;
        }
        if (0 == node->fKeyword.compare(PLURAL_KEYWORD_OTHER, 5)) {
            addKeywordOther = FALSE;
        }
        node = node->fNext;
    }

    if (addKeywordOther) {
        auto newElem = new UnicodeString(PLURAL_KEYWORD_OTHER);
        if (newElem == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        fKeywordNames.addElement(newElem, status);
        if (U_FAILURE(status)) {
            delete newElem;
            return;
        }
    }
}

const UnicodeString*
PluralKeywordEnumeration::snext(UErrorCode& status) {
    if (U_SUCCESS(status) && pos < fKeywordNames.size()) {
        return (const UnicodeString*)fKeywordNames.elementAt(pos++);
    }
    return nullptr;
}

void
PluralKeywordEnumeration::reset(UErrorCode& /*status*/) {
    pos=0;
}

int32_t
PluralKeywordEnumeration::count(UErrorCode& /*status*/) const {
    return fKeywordNames.size();
}

PluralKeywordEnumeration::~PluralKeywordEnumeration() {
}

PluralOperand tokenTypeToPluralOperand(tokenType tt) {
    switch(tt) {
    case plurrule_token_tVariableN:
        return PLURAL_OPERAND_N;
    case plurrule_token_tVariableI:
        return PLURAL_OPERAND_I;
    case plurrule_token_tVariableF:
        return PLURAL_OPERAND_F;
    case plurrule_token_tVariableV:
        return PLURAL_OPERAND_V;
    case plurrule_token_tVariableT:
        return PLURAL_OPERAND_T;
    default:
        UPRV_UNREACHABLE;  // unexpected.
    }
}

FixedDecimal::FixedDecimal(double n, int32_t v, int64_t f) {
    init(n, v, f);
    // check values. TODO make into unit test.
    //
    //            long visiblePower = (int) Math.pow(10, v);
    //            if (decimalDigits > visiblePower) {
    //                throw new IllegalArgumentException();
    //            }
    //            double fraction = intValue + (decimalDigits / (double) visiblePower);
    //            if (fraction != source) {
    //                double diff = Math.abs(fraction - source)/(Math.abs(fraction) + Math.abs(source));
    //                if (diff > 0.00000001d) {
    //                    throw new IllegalArgumentException();
    //                }
    //            }
}

FixedDecimal::FixedDecimal(double n, int32_t v) {
    // Ugly, but for samples we don't care.
    init(n, v, getFractionalDigits(n, v));
}

FixedDecimal::FixedDecimal(double n) {
    init(n);
}

FixedDecimal::FixedDecimal() {
    init(0, 0, 0);
}


// Create a FixedDecimal from a UnicodeString containing a number.
//    Inefficient, but only used for samples, so simplicity trumps efficiency.

FixedDecimal::FixedDecimal(const UnicodeString &num, UErrorCode &status) {
    CharString cs;
    cs.appendInvariantChars(num, status);
    DecimalQuantity dl;
    dl.setToDecNumber(cs.toStringPiece(), status);
    if (U_FAILURE(status)) {
        init(0, 0, 0);
        return;
    }
    int32_t decimalPoint = num.indexOf(PLURRULE_DOT);
    double n = dl.toDouble();
    if (decimalPoint == -1) {
        init(n, 0, 0);
    } else {
        int32_t v = num.length() - decimalPoint - 1;
        init(n, v, getFractionalDigits(n, v));
    }
}


FixedDecimal::FixedDecimal(const FixedDecimal &other) {
    source = other.source;
    visibleDecimalDigitCount = other.visibleDecimalDigitCount;
    decimalDigits = other.decimalDigits;
    decimalDigitsWithoutTrailingZeros = other.decimalDigitsWithoutTrailingZeros;
    intValue = other.intValue;
    _hasIntegerValue = other._hasIntegerValue;
    isNegative = other.isNegative;
    _isNaN = other._isNaN;
    _isInfinite = other._isInfinite;
}

FixedDecimal::~FixedDecimal() = default;


void FixedDecimal::init(double n) {
    int32_t numFractionDigits = decimals(n);
    init(n, numFractionDigits, getFractionalDigits(n, numFractionDigits));
}


void FixedDecimal::init(double n, int32_t v, int64_t f) {
    isNegative = n < 0.0;
    source = fabs(n);
    _isNaN = uprv_isNaN(source);
    _isInfinite = uprv_isInfinite(source);
    if (_isNaN || _isInfinite) {
        v = 0;
        f = 0;
        intValue = 0;
        _hasIntegerValue = FALSE;
    } else {
        intValue = (int64_t)source;
        _hasIntegerValue = (source == intValue);
    }

    visibleDecimalDigitCount = v;
    decimalDigits = f;
    if (f == 0) {
         decimalDigitsWithoutTrailingZeros = 0;
    } else {
        int64_t fdwtz = f;
        while ((fdwtz%10) == 0) {
            fdwtz /= 10;
        }
        decimalDigitsWithoutTrailingZeros = fdwtz;
    }
}


//  Fast path only exact initialization. Return true if successful.
//     Note: Do not multiply by 10 each time through loop, rounding cruft can build
//           up that makes the check for an integer result fail.
//           A single multiply of the original number works more reliably.
static int32_t p10[] = {1, 10, 100, 1000, 10000};
UBool FixedDecimal::quickInit(double n) {
    UBool success = FALSE;
    n = fabs(n);
    int32_t numFractionDigits;
    for (numFractionDigits = 0; numFractionDigits <= 3; numFractionDigits++) {
        double scaledN = n * p10[numFractionDigits];
        if (scaledN == floor(scaledN)) {
            success = TRUE;
            break;
        }
    }
    if (success) {
        init(n, numFractionDigits, getFractionalDigits(n, numFractionDigits));
    }
    return success;
}



int32_t FixedDecimal::decimals(double n) {
    // Count the number of decimal digits in the fraction part of the number, excluding trailing zeros.
    // fastpath the common cases, integers or fractions with 3 or fewer digits
    n = fabs(n);
    for (int ndigits=0; ndigits<=3; ndigits++) {
        double scaledN = n * p10[ndigits];
        if (scaledN == floor(scaledN)) {
            return ndigits;
        }
    }

    // Slow path, convert with sprintf, parse converted output.
    char  buf[30] = {0};
    sprintf(buf, "%1.15e", n);
    // formatted number looks like this: 1.234567890123457e-01
    int exponent = atoi(buf+18);
    int numFractionDigits = 15;
    for (int i=16; ; --i) {
        if (buf[i] != '0') {
            break;
        }
        --numFractionDigits;
    }
    numFractionDigits -= exponent;   // Fraction part of fixed point representation.
    return numFractionDigits;
}


// Get the fraction digits of a double, represented as an integer.
//    v is the number of visible fraction digits in the displayed form of the number.
//       Example: n = 1001.234, v = 6, result = 234000
//    TODO: need to think through how this is used in the plural rule context.
//          This function can easily encounter integer overflow,
//          and can easily return noise digits when the precision of a double is exceeded.

int64_t FixedDecimal::getFractionalDigits(double n, int32_t v) {
    if (v == 0 || n == floor(n) || uprv_isNaN(n) || uprv_isPositiveInfinity(n)) {
        return 0;
    }
    n = fabs(n);
    double fract = n - floor(n);
    switch (v) {
      case 1: return (int64_t)(fract*10.0 + 0.5);
      case 2: return (int64_t)(fract*100.0 + 0.5);
      case 3: return (int64_t)(fract*1000.0 + 0.5);
      default:
          double scaled = floor(fract * pow(10.0, (double)v) + 0.5);
          if (scaled > U_INT64_MAX) {
              return U_INT64_MAX;
          } else {
              return (int64_t)scaled;
          }
      }
}


void FixedDecimal::adjustForMinFractionDigits(int32_t minFractionDigits) {
    int32_t numTrailingFractionZeros = minFractionDigits - visibleDecimalDigitCount;
    if (numTrailingFractionZeros > 0) {
        for (int32_t i=0; i<numTrailingFractionZeros; i++) {
            // Do not let the decimalDigits value overflow if there are many trailing zeros.
            // Limit the value to 18 digits, the most that a 64 bit int can fully represent.
            if (decimalDigits >= 100000000000000000LL) {
                break;
            }
            decimalDigits *= 10;
        }
        visibleDecimalDigitCount += numTrailingFractionZeros;
    }
}


double FixedDecimal::getPluralOperand(PluralOperand operand) const {
    switch(operand) {
        case PLURAL_OPERAND_N: return source;
        case PLURAL_OPERAND_I: return static_cast<double>(intValue);
        case PLURAL_OPERAND_F: return static_cast<double>(decimalDigits);
        case PLURAL_OPERAND_T: return static_cast<double>(decimalDigitsWithoutTrailingZeros);
        case PLURAL_OPERAND_V: return visibleDecimalDigitCount;
        default:
             UPRV_UNREACHABLE;  // unexpected.
    }
}

bool FixedDecimal::isNaN() const {
    return _isNaN;
}

bool FixedDecimal::isInfinite() const {
    return _isInfinite;
}

bool FixedDecimal::hasIntegerValue() const {
    return _hasIntegerValue;
}

bool FixedDecimal::isNanOrInfinity() const {
    return _isNaN || _isInfinite;
}

int32_t FixedDecimal::getVisibleFractionDigitCount() const {
    return visibleDecimalDigitCount;
}



PluralAvailableLocalesEnumeration::PluralAvailableLocalesEnumeration(UErrorCode &status) {
    fOpenStatus = status;
    if (U_FAILURE(status)) {
        return;
    }
    fOpenStatus = U_ZERO_ERROR; // clear any warnings.
    LocalUResourceBundlePointer rb(ures_openDirect(nullptr, "plurals", &fOpenStatus));
    fLocales = ures_getByKey(rb.getAlias(), "locales", nullptr, &fOpenStatus);
}

PluralAvailableLocalesEnumeration::~PluralAvailableLocalesEnumeration() {
    ures_close(fLocales);
    ures_close(fRes);
    fLocales = nullptr;
    fRes = nullptr;
}

const char *PluralAvailableLocalesEnumeration::next(int32_t *resultLength, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (U_FAILURE(fOpenStatus)) {
        status = fOpenStatus;
        return nullptr;
    }
    fRes = ures_getNextResource(fLocales, fRes, &status);
    if (fRes == nullptr || U_FAILURE(status)) {
        if (status == U_INDEX_OUTOFBOUNDS_ERROR) {
            status = U_ZERO_ERROR;
        }
        return nullptr;
    }
    const char *result = ures_getKey(fRes);
    if (resultLength != nullptr) {
        *resultLength = static_cast<int32_t>(uprv_strlen(result));
    }
    return result;
}


void PluralAvailableLocalesEnumeration::reset(UErrorCode &status) {
    if (U_FAILURE(status)) {
       return;
    }
    if (U_FAILURE(fOpenStatus)) {
        status = fOpenStatus;
        return;
    }
    ures_resetIterator(fLocales);
}

int32_t PluralAvailableLocalesEnumeration::count(UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return 0;
    }
    if (U_FAILURE(fOpenStatus)) {
        status = fOpenStatus;
        return 0;
    }
    return ures_getSize(fLocales);
}

U_NAMESPACE_END


#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2014-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* quantityformatter.cpp
*/



#if !UCONFIG_NO_FORMATTING

















U_NAMESPACE_BEGIN

QuantityFormatter::QuantityFormatter() {
    for (int32_t i = 0; i < UPRV_LENGTHOF(formatters); ++i) {
        formatters[i] = NULL;
    }
}

QuantityFormatter::QuantityFormatter(const QuantityFormatter &other) {
    for (int32_t i = 0; i < UPRV_LENGTHOF(formatters); ++i) {
        if (other.formatters[i] == NULL) {
            formatters[i] = NULL;
        } else {
            formatters[i] = new SimpleFormatter(*other.formatters[i]);
        }
    }
}

QuantityFormatter &QuantityFormatter::operator=(
        const QuantityFormatter& other) {
    if (this == &other) {
        return *this;
    }
    for (int32_t i = 0; i < UPRV_LENGTHOF(formatters); ++i) {
        delete formatters[i];
        if (other.formatters[i] == NULL) {
            formatters[i] = NULL;
        } else {
            formatters[i] = new SimpleFormatter(*other.formatters[i]);
        }
    }
    return *this;
}

QuantityFormatter::~QuantityFormatter() {
    for (int32_t i = 0; i < UPRV_LENGTHOF(formatters); ++i) {
        delete formatters[i];
    }
}

void QuantityFormatter::reset() {
    for (int32_t i = 0; i < UPRV_LENGTHOF(formatters); ++i) {
        delete formatters[i];
        formatters[i] = NULL;
    }
}

UBool QuantityFormatter::addIfAbsent(
        const char *variant,
        const UnicodeString &rawPattern,
        UErrorCode &status) {
    int32_t pluralIndex = StandardPlural::indexFromString(variant, status);
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (formatters[pluralIndex] != NULL) {
        return TRUE;
    }
    SimpleFormatter *newFmt = new SimpleFormatter(rawPattern, 0, 1, status);
    if (newFmt == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    if (U_FAILURE(status)) {
        delete newFmt;
        return FALSE;
    }
    formatters[pluralIndex] = newFmt;
    return TRUE;
}

UBool QuantityFormatter::isValid() const {
    return formatters[StandardPlural::OTHER] != NULL;
}

const SimpleFormatter *QuantityFormatter::getByVariant(
        const char *variant) const {
    U_ASSERT(isValid());
    int32_t pluralIndex = StandardPlural::indexOrOtherIndexFromString(variant);
    const SimpleFormatter *pattern = formatters[pluralIndex];
    if (pattern == NULL) {
        pattern = formatters[StandardPlural::OTHER];
    }
    return pattern;
}

UnicodeString &QuantityFormatter::format(
            const Formattable &number,
            const NumberFormat &fmt,
            const PluralRules &rules,
            UnicodeString &appendTo,
            FieldPosition &pos,
            UErrorCode &status) const {
    UnicodeString formattedNumber;
    StandardPlural::Form p = selectPlural(number, fmt, rules, formattedNumber, pos, status);
    if (U_FAILURE(status)) {
        return appendTo;
    }
    const SimpleFormatter *pattern = formatters[p];
    if (pattern == NULL) {
        pattern = formatters[StandardPlural::OTHER];
        if (pattern == NULL) {
            status = U_INVALID_STATE_ERROR;
            return appendTo;
        }
    }
    return format(*pattern, formattedNumber, appendTo, pos, status);
}

// The following methods live here so that class PluralRules does not depend on number formatting,
// and the SimpleFormatter does not depend on FieldPosition.

StandardPlural::Form QuantityFormatter::selectPlural(
            const Formattable &number,
            const NumberFormat &fmt,
            const PluralRules &rules,
            UnicodeString &formattedNumber,
            FieldPosition &pos,
            UErrorCode &status) {
    if (U_FAILURE(status)) {
        return StandardPlural::OTHER;
    }
    UnicodeString pluralKeyword;
    const DecimalFormat *decFmt = dynamic_cast<const DecimalFormat *>(&fmt);
    if (decFmt != NULL) {
        number::impl::DecimalQuantity dq;
        decFmt->formatToDecimalQuantity(number, dq, status);
        if (U_FAILURE(status)) {
            return StandardPlural::OTHER;
        }
        pluralKeyword = rules.select(dq);
        decFmt->format(number, formattedNumber, pos, status);
    } else {
        if (number.getType() == Formattable::kDouble) {
            pluralKeyword = rules.select(number.getDouble());
        } else if (number.getType() == Formattable::kLong) {
            pluralKeyword = rules.select(number.getLong());
        } else if (number.getType() == Formattable::kInt64) {
            pluralKeyword = rules.select((double) number.getInt64());
        } else {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return StandardPlural::OTHER;
        }
        fmt.format(number, formattedNumber, pos, status);
    }
    return StandardPlural::orOtherFromString(pluralKeyword);
}

void QuantityFormatter::formatAndSelect(
        double quantity,
        const NumberFormat& fmt,
        const PluralRules& rules,
        FormattedStringBuilder& output,
        StandardPlural::Form& pluralForm,
        UErrorCode& status) {
    UnicodeString pluralKeyword;
    const DecimalFormat* df = dynamic_cast<const DecimalFormat*>(&fmt);
    if (df != nullptr) {
        number::impl::UFormattedNumberData fn;
        fn.quantity.setToDouble(quantity);
        const number::LocalizedNumberFormatter* lnf = df->toNumberFormatter(status);
        if (U_FAILURE(status)) {
            return;
        }
        lnf->formatImpl(&fn, status);
        if (U_FAILURE(status)) {
            return;
        }
        output = std::move(fn.getStringRef());
        pluralKeyword = rules.select(fn.quantity);
    } else {
        UnicodeString result;
        fmt.format(quantity, result, status);
        if (U_FAILURE(status)) {
            return;
        }
        output.append(result, UNUM_FIELD_COUNT, status);
        if (U_FAILURE(status)) {
            return;
        }
        pluralKeyword = rules.select(quantity);
    }
    pluralForm = StandardPlural::orOtherFromString(pluralKeyword);
}

UnicodeString &QuantityFormatter::format(
            const SimpleFormatter &pattern,
            const UnicodeString &value,
            UnicodeString &appendTo,
            FieldPosition &pos,
            UErrorCode &status) {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    const UnicodeString *param = &value;
    int32_t offset;
    pattern.formatAndAppend(&param, 1, appendTo, &offset, 1, status);
    if (pos.getBeginIndex() != 0 || pos.getEndIndex() != 0) {
        if (offset >= 0) {
            pos.setBeginIndex(pos.getBeginIndex() + offset);
            pos.setEndIndex(pos.getEndIndex() + offset);
        } else {
            pos.setBeginIndex(0);
            pos.setEndIndex(0);
        }
    }
    return appendTo;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2015, International Business Machines Corporation
* and others. All Rights Reserved.
*******************************************************************************
*/


  // for 'typeid' to work



#if U_HAVE_RBNF













// #include "unicode/brkiter.h"









// debugging
// #define RBNF_DEBUG

#ifdef RBNF_DEBUG
#include <stdio.h>
#endif

#define U_ICUDATA_RBNF U_ICUDATA_NAME U_TREE_SEPARATOR_STRING "rbnf"

static const UChar rbnf_gPercentPercent[] =
{
    0x25, 0x25, 0
}; /* "%%" */

// All urbnf objects are created through openRules, so we init all of the
// Unicode string constants required by rbnf, nfrs, or nfr here.
static const UChar rbnf_gLenientParse[] =
{
    0x25, 0x25, 0x6C, 0x65, 0x6E, 0x69, 0x65, 0x6E, 0x74, 0x2D, 0x70, 0x61, 0x72, 0x73, 0x65, 0x3A, 0
}; /* "%%lenient-parse:" */
static const UChar rbnf_gSemiColon = 0x003B;
static const UChar rbnf_gSemiPercent[] =
{
    0x3B, 0x25, 0
}; /* ";%" */

#define kSomeNumberOfBitsDiv2 22
#define kHalfMaxDouble (double)(1 << kSomeNumberOfBitsDiv2)
#define kMaxDouble (kHalfMaxDouble * kHalfMaxDouble)

U_NAMESPACE_BEGIN

using number::impl::DecimalQuantity;

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(RuleBasedNumberFormat)

/*
This is a utility class. It does not use ICU's RTTI.
If ICU's RTTI is needed again, you can uncomment the RTTI code and derive from UObject.
Please make sure that intltest passes on Windows in Release mode,
since the string pooling per compilation unit will mess up how RTTI works.
The RTTI code was also removed due to lack of code coverage.
*/
class LocalizationInfo : public UMemory {
protected:
    virtual ~LocalizationInfo();
    uint32_t refcount;

public:
    LocalizationInfo() : refcount(0) {}

    LocalizationInfo* ref(void) {
        ++refcount;
        return this;
    }

    LocalizationInfo* unref(void) {
        if (refcount && --refcount == 0) {
            delete this;
        }
        return NULL;
    }

    virtual UBool operator==(const LocalizationInfo* rhs) const;
    inline  UBool operator!=(const LocalizationInfo* rhs) const { return !operator==(rhs); }

    virtual int32_t getNumberOfRuleSets(void) const = 0;
    virtual const UChar* getRuleSetName(int32_t index) const = 0;
    virtual int32_t getNumberOfDisplayLocales(void) const = 0;
    virtual const UChar* getLocaleName(int32_t index) const = 0;
    virtual const UChar* getDisplayName(int32_t localeIndex, int32_t ruleIndex) const = 0;

    virtual int32_t indexForLocale(const UChar* locale) const;
    virtual int32_t indexForRuleSet(const UChar* ruleset) const;

//    virtual UClassID getDynamicClassID() const = 0;
//    static UClassID getStaticClassID(void);
};

LocalizationInfo::~LocalizationInfo() {}

//UOBJECT_DEFINE_ABSTRACT_RTTI_IMPLEMENTATION(LocalizationInfo)

// if both strings are NULL, this returns TRUE
static UBool
streq(const UChar* lhs, const UChar* rhs) {
    if (rhs == lhs) {
        return TRUE;
    }
    if (lhs && rhs) {
        return u_strcmp(lhs, rhs) == 0;
    }
    return FALSE;
}

UBool
LocalizationInfo::operator==(const LocalizationInfo* rhs) const {
    if (rhs) {
        if (this == rhs) {
            return TRUE;
        }

        int32_t rsc = getNumberOfRuleSets();
        if (rsc == rhs->getNumberOfRuleSets()) {
            for (int i = 0; i < rsc; ++i) {
                if (!streq(getRuleSetName(i), rhs->getRuleSetName(i))) {
                    return FALSE;
                }
            }
            int32_t dlc = getNumberOfDisplayLocales();
            if (dlc == rhs->getNumberOfDisplayLocales()) {
                for (int i = 0; i < dlc; ++i) {
                    const UChar* locale = getLocaleName(i);
                    int32_t ix = rhs->indexForLocale(locale);
                    // if no locale, ix is -1, getLocaleName returns null, so streq returns false
                    if (!streq(locale, rhs->getLocaleName(ix))) {
                        return FALSE;
                    }
                    for (int j = 0; j < rsc; ++j) {
                        if (!streq(getDisplayName(i, j), rhs->getDisplayName(ix, j))) {
                            return FALSE;
                        }
                    }
                }
                return TRUE;
            }
        }
    }
    return FALSE;
}

int32_t
LocalizationInfo::indexForLocale(const UChar* locale) const {
    for (int i = 0; i < getNumberOfDisplayLocales(); ++i) {
        if (streq(locale, getLocaleName(i))) {
            return i;
        }
    }
    return -1;
}

int32_t
LocalizationInfo::indexForRuleSet(const UChar* ruleset) const {
    if (ruleset) {
        for (int i = 0; i < getNumberOfRuleSets(); ++i) {
            if (streq(ruleset, getRuleSetName(i))) {
                return i;
            }
        }
    }
    return -1;
}


typedef void (*Fn_Deleter)(void*);

class VArray {
    void** buf;
    int32_t cap;
    int32_t size;
    Fn_Deleter deleter;
public:
    VArray() : buf(NULL), cap(0), size(0), deleter(NULL) {}

    VArray(Fn_Deleter del) : buf(NULL), cap(0), size(0), deleter(del) {}

    ~VArray() {
        if (deleter) {
            for (int i = 0; i < size; ++i) {
                (*deleter)(buf[i]);
            }
        }
        uprv_free(buf);
    }

    int32_t length() {
        return size;
    }

    void add(void* elem, UErrorCode& status) {
        if (U_SUCCESS(status)) {
            if (size == cap) {
                if (cap == 0) {
                    cap = 1;
                } else if (cap < 256) {
                    cap *= 2;
                } else {
                    cap += 256;
                }
                if (buf == NULL) {
                    buf = (void**)uprv_malloc(cap * sizeof(void*));
                } else {
                    buf = (void**)uprv_realloc(buf, cap * sizeof(void*));
                }
                if (buf == NULL) {
                    // if we couldn't realloc, we leak the memory we've already allocated, but we're in deep trouble anyway
                    status = U_MEMORY_ALLOCATION_ERROR;
                    return;
                }
                void* start = &buf[size];
                size_t count = (cap - size) * sizeof(void*);
                uprv_memset(start, 0, count); // fill with nulls, just because
            }
            buf[size++] = elem;
        }
    }

    void** release(void) {
        void** result = buf;
        buf = NULL;
        cap = 0;
        size = 0;
        return result;
    }
};

class LocDataParser;

class StringLocalizationInfo : public LocalizationInfo {
    UChar* info;
    UChar*** data;
    int32_t numRuleSets;
    int32_t numLocales;

friend class LocDataParser;

    StringLocalizationInfo(UChar* i, UChar*** d, int32_t numRS, int32_t numLocs)
        : info(i), data(d), numRuleSets(numRS), numLocales(numLocs)
    {
    }

public:
    static StringLocalizationInfo* create(const UnicodeString& info, UParseError& perror, UErrorCode& status);

    virtual ~StringLocalizationInfo();
    virtual int32_t getNumberOfRuleSets(void) const { return numRuleSets; }
    virtual const UChar* getRuleSetName(int32_t index) const;
    virtual int32_t getNumberOfDisplayLocales(void) const { return numLocales; }
    virtual const UChar* getLocaleName(int32_t index) const;
    virtual const UChar* getDisplayName(int32_t localeIndex, int32_t ruleIndex) const;

//    virtual UClassID getDynamicClassID() const;
//    static UClassID getStaticClassID(void);

private:
    void init(UErrorCode& status) const;
};


enum {
    rbnf_OPEN_ANGLE = 0x003c, /* '<' */
    rbnf_CLOSE_ANGLE = 0x003e, /* '>' */
    rbnf_COMMA = 0x002c,
    rbnf_TICK = 0x0027,
    rbnf_QUOTE = 0x0022,
    rbnf_SPACE = 0x0020
};

/**
 * Utility for parsing a localization string and returning a StringLocalizationInfo*.
 */
class LocDataParser {
    UChar* data;
    const UChar* e;
    UChar* p;
    UChar ch;
    UParseError& pe;
    UErrorCode& ec;

public:
    LocDataParser(UParseError& parseError, UErrorCode& status)
        : data(NULL), e(NULL), p(NULL), ch(0xffff), pe(parseError), ec(status) {}
    ~LocDataParser() {}

    /*
    * On a successful parse, return a StringLocalizationInfo*, otherwise delete locData, set perror and status,
    * and return NULL.  The StringLocalizationInfo will adopt locData if it is created.
    */
    StringLocalizationInfo* parse(UChar* data, int32_t len);

private:

    inline void inc(void) {
        ++p;
        ch = 0xffff;
    }
    inline UBool checkInc(UChar c) {
        if (p < e && (ch == c || *p == c)) {
            inc();
            return TRUE;
        }
        return FALSE;
    }
    inline UBool check(UChar c) {
        return p < e && (ch == c || *p == c);
    }
    inline void skipWhitespace(void) {
        while (p < e && PatternProps::isWhiteSpace(ch != 0xffff ? ch : *p)) {
            inc();
        }
    }
    inline UBool inList(UChar c, const UChar* list) const {
        if (*list == rbnf_SPACE && PatternProps::isWhiteSpace(c)) {
            return TRUE;
        }
        while (*list && *list != c) {
            ++list;
        }
        return *list == c;
    }
    void parseError(const char* msg);

    StringLocalizationInfo* doParse(void);

    UChar** nextArray(int32_t& requiredLength);
    UChar*  nextString(void);
};

#ifdef RBNF_DEBUG
#define rbnf_ERROR(msg) UPRV_BLOCK_MACRO_BEGIN { \
    parseError(msg); \
    return NULL; \
} UPRV_BLOCK_MACRO_END
#define EXPLANATION_ARG explanationArg
#else
#define rbnf_ERROR(msg) UPRV_BLOCK_MACRO_BEGIN { \
    parseError(NULL); \
    return NULL; \
} UPRV_BLOCK_MACRO_END
#define EXPLANATION_ARG
#endif


static const UChar Drbnf_QUOTE_STOPLIST[] = {
    rbnf_QUOTE, 0
};

static const UChar Srbnf_QUOTE_STOPLIST[] = {
    rbnf_TICK, 0
};

static const UChar NOrbnf_QUOTE_STOPLIST[] = {
    rbnf_SPACE, rbnf_COMMA, rbnf_CLOSE_ANGLE, rbnf_OPEN_ANGLE, rbnf_TICK, rbnf_QUOTE, 0
};

static void
DeleteFn(void* p) {
  uprv_free(p);
}

StringLocalizationInfo*
LocDataParser::parse(UChar* _data, int32_t len) {
    if (U_FAILURE(ec)) {
        if (_data) uprv_free(_data);
        return NULL;
    }

    pe.line = 0;
    pe.offset = -1;
    pe.postContext[0] = 0;
    pe.preContext[0] = 0;

    if (_data == NULL) {
        ec = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if (len <= 0) {
        ec = U_ILLEGAL_ARGUMENT_ERROR;
        uprv_free(_data);
        return NULL;
    }

    data = _data;
    e = data + len;
    p = _data;
    ch = 0xffff;

    return doParse();
}


StringLocalizationInfo*
LocDataParser::doParse(void) {
    skipWhitespace();
    if (!checkInc(rbnf_OPEN_ANGLE)) {
        rbnf_ERROR("Missing open angle");
    } else {
        VArray array(DeleteFn);
        UBool mightHaveNext = TRUE;
        int32_t requiredLength = -1;
        while (mightHaveNext) {
            mightHaveNext = FALSE;
            UChar** elem = nextArray(requiredLength);
            skipWhitespace();
            UBool haveComma = check(rbnf_COMMA);
            if (elem) {
                array.add(elem, ec);
                if (haveComma) {
                    inc();
                    mightHaveNext = TRUE;
                }
            } else if (haveComma) {
                rbnf_ERROR("Unexpected character");
            }
        }

        skipWhitespace();
        if (!checkInc(rbnf_CLOSE_ANGLE)) {
            if (check(rbnf_OPEN_ANGLE)) {
                rbnf_ERROR("Missing comma in outer array");
            } else {
                rbnf_ERROR("Missing close angle bracket in outer array");
            }
        }

        skipWhitespace();
        if (p != e) {
            rbnf_ERROR("Extra text after close of localization data");
        }

        array.add(NULL, ec);
        if (U_SUCCESS(ec)) {
            int32_t numLocs = array.length() - 2; // subtract first, NULL
            UChar*** result = (UChar***)array.release();

            return new StringLocalizationInfo(data, result, requiredLength-2, numLocs); // subtract first, NULL
        }
    }

    rbnf_ERROR("Unknown error");
}

UChar**
LocDataParser::nextArray(int32_t& requiredLength) {
    if (U_FAILURE(ec)) {
        return NULL;
    }

    skipWhitespace();
    if (!checkInc(rbnf_OPEN_ANGLE)) {
        rbnf_ERROR("Missing open angle");
    }

    VArray array;
    UBool mightHaveNext = TRUE;
    while (mightHaveNext) {
        mightHaveNext = FALSE;
        UChar* elem = nextString();
        skipWhitespace();
        UBool haveComma = check(rbnf_COMMA);
        if (elem) {
            array.add(elem, ec);
            if (haveComma) {
                inc();
                mightHaveNext = TRUE;
            }
        } else if (haveComma) {
            rbnf_ERROR("Unexpected comma");
        }
    }
    skipWhitespace();
    if (!checkInc(rbnf_CLOSE_ANGLE)) {
        if (check(rbnf_OPEN_ANGLE)) {
            rbnf_ERROR("Missing close angle bracket in inner array");
        } else {
            rbnf_ERROR("Missing comma in inner array");
        }
    }

    array.add(NULL, ec);
    if (U_SUCCESS(ec)) {
        if (requiredLength == -1) {
            requiredLength = array.length() + 1;
        } else if (array.length() != requiredLength) {
            ec = U_ILLEGAL_ARGUMENT_ERROR;
            rbnf_ERROR("Array not of required length");
        }

        return (UChar**)array.release();
    }
    rbnf_ERROR("Unknown Error");
}

UChar*
LocDataParser::nextString() {
    UChar* result = NULL;

    skipWhitespace();
    if (p < e) {
        const UChar* terminators;
        UChar c = *p;
        UBool haveQuote = c == rbnf_QUOTE || c == rbnf_TICK;
        if (haveQuote) {
            inc();
            terminators = c == rbnf_QUOTE ? Drbnf_QUOTE_STOPLIST : Srbnf_QUOTE_STOPLIST;
        } else {
            terminators = NOrbnf_QUOTE_STOPLIST;
        }
        UChar* start = p;
        while (p < e && !inList(*p, terminators)) ++p;
        if (p == e) {
            rbnf_ERROR("Unexpected end of data");
        }

        UChar x = *p;
        if (p > start) {
            ch = x;
            *p = 0x0; // terminate by writing to data
            result = start; // just point into data
        }
        if (haveQuote) {
            if (x != c) {
                rbnf_ERROR("Missing matching quote");
            } else if (p == start) {
                rbnf_ERROR("Empty string");
            }
            inc();
        } else if (x == rbnf_OPEN_ANGLE || x == rbnf_TICK || x == rbnf_QUOTE) {
            rbnf_ERROR("Unexpected character in string");
        }
    }

    // ok for there to be no next string
    return result;
}

void LocDataParser::parseError(const char* EXPLANATION_ARG)
{
    if (!data) {
        return;
    }

    const UChar* start = p - U_PARSE_CONTEXT_LEN - 1;
    if (start < data) {
        start = data;
    }
    for (UChar* x = p; --x >= start;) {
        if (!*x) {
            start = x+1;
            break;
        }
    }
    const UChar* limit = p + U_PARSE_CONTEXT_LEN - 1;
    if (limit > e) {
        limit = e;
    }
    u_strncpy(pe.preContext, start, (int32_t)(p-start));
    pe.preContext[p-start] = 0;
    u_strncpy(pe.postContext, p, (int32_t)(limit-p));
    pe.postContext[limit-p] = 0;
    pe.offset = (int32_t)(p - data);

#ifdef RBNF_DEBUG
    fprintf(stderr, "%s at or near character %ld: ", EXPLANATION_ARG, p-data);

    UnicodeString msg;
    msg.append(start, p - start);
    msg.append((UChar)0x002f); /* SOLIDUS/SLASH */
    msg.append(p, limit-p);
    msg.append(UNICODE_STRING_SIMPLE("'"));

    char buf[128];
    int32_t len = msg.extract(0, msg.length(), buf, 128);
    if (len >= 128) {
        buf[127] = 0;
    } else {
        buf[len] = 0;
    }
    fprintf(stderr, "%s\n", buf);
    fflush(stderr);
#endif

    uprv_free(data);
    data = NULL;
    p = NULL;
    e = NULL;

    if (U_SUCCESS(ec)) {
        ec = U_PARSE_ERROR;
    }
}

//UOBJECT_DEFINE_RTTI_IMPLEMENTATION(StringLocalizationInfo)

StringLocalizationInfo*
StringLocalizationInfo::create(const UnicodeString& info, UParseError& perror, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }

    int32_t len = info.length();
    if (len == 0) {
        return NULL; // no error;
    }

    UChar* p = (UChar*)uprv_malloc(len * sizeof(UChar));
    if (!p) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    info.extract(p, len, status);
    if (!U_FAILURE(status)) {
        status = U_ZERO_ERROR; // clear warning about non-termination
    }

    LocDataParser parser(perror, status);
    return parser.parse(p, len);
}

StringLocalizationInfo::~StringLocalizationInfo() {
    for (UChar*** p = (UChar***)data; *p; ++p) {
        // remaining data is simply pointer into our unicode string data.
        if (*p) uprv_free(*p);
    }
    if (data) uprv_free(data);
    if (info) uprv_free(info);
}


const UChar*
StringLocalizationInfo::getRuleSetName(int32_t index) const {
    if (index >= 0 && index < getNumberOfRuleSets()) {
        return data[0][index];
    }
    return NULL;
}

const UChar*
StringLocalizationInfo::getLocaleName(int32_t index) const {
    if (index >= 0 && index < getNumberOfDisplayLocales()) {
        return data[index+1][0];
    }
    return NULL;
}

const UChar*
StringLocalizationInfo::getDisplayName(int32_t localeIndex, int32_t ruleIndex) const {
    if (localeIndex >= 0 && localeIndex < getNumberOfDisplayLocales() &&
        ruleIndex >= 0 && ruleIndex < getNumberOfRuleSets()) {
        return data[localeIndex+1][ruleIndex+1];
    }
    return NULL;
}

// ----------

RuleBasedNumberFormat::RuleBasedNumberFormat(const UnicodeString& description,
                                             const UnicodeString& locs,
                                             const Locale& alocale, UParseError& perror, UErrorCode& status)
  : fRuleSets(NULL)
  , ruleSetDescriptions(NULL)
  , numRuleSets(0)
  , defaultRuleSet(NULL)
  , locale(alocale)
  , collator(NULL)
  , decimalFormatSymbols(NULL)
  , defaultInfinityRule(NULL)
  , defaultNaNRule(NULL)
  , fRoundingMode(DecimalFormat::ERoundingMode::kRoundUnnecessary)
  , lenient(FALSE)
  , lenientParseRules(NULL)
  , localizations(NULL)
  , capitalizationInfoSet(FALSE)
  , capitalizationForUIListMenu(FALSE)
  , capitalizationForStandAlone(FALSE)
  , capitalizationBrkIter(NULL)
{
  LocalizationInfo* locinfo = StringLocalizationInfo::create(locs, perror, status);
  init(description, locinfo, perror, status);
}

RuleBasedNumberFormat::RuleBasedNumberFormat(const UnicodeString& description,
                                             const UnicodeString& locs,
                                             UParseError& perror, UErrorCode& status)
  : fRuleSets(NULL)
  , ruleSetDescriptions(NULL)
  , numRuleSets(0)
  , defaultRuleSet(NULL)
  , locale(Locale::getDefault())
  , collator(NULL)
  , decimalFormatSymbols(NULL)
  , defaultInfinityRule(NULL)
  , defaultNaNRule(NULL)
  , fRoundingMode(DecimalFormat::ERoundingMode::kRoundUnnecessary)
  , lenient(FALSE)
  , lenientParseRules(NULL)
  , localizations(NULL)
  , capitalizationInfoSet(FALSE)
  , capitalizationForUIListMenu(FALSE)
  , capitalizationForStandAlone(FALSE)
  , capitalizationBrkIter(NULL)
{
  LocalizationInfo* locinfo = StringLocalizationInfo::create(locs, perror, status);
  init(description, locinfo, perror, status);
}

RuleBasedNumberFormat::RuleBasedNumberFormat(const UnicodeString& description,
                                             LocalizationInfo* info,
                                             const Locale& alocale, UParseError& perror, UErrorCode& status)
  : fRuleSets(NULL)
  , ruleSetDescriptions(NULL)
  , numRuleSets(0)
  , defaultRuleSet(NULL)
  , locale(alocale)
  , collator(NULL)
  , decimalFormatSymbols(NULL)
  , defaultInfinityRule(NULL)
  , defaultNaNRule(NULL)
  , fRoundingMode(DecimalFormat::ERoundingMode::kRoundUnnecessary)
  , lenient(FALSE)
  , lenientParseRules(NULL)
  , localizations(NULL)
  , capitalizationInfoSet(FALSE)
  , capitalizationForUIListMenu(FALSE)
  , capitalizationForStandAlone(FALSE)
  , capitalizationBrkIter(NULL)
{
  init(description, info, perror, status);
}

RuleBasedNumberFormat::RuleBasedNumberFormat(const UnicodeString& description,
                         UParseError& perror,
                         UErrorCode& status)
  : fRuleSets(NULL)
  , ruleSetDescriptions(NULL)
  , numRuleSets(0)
  , defaultRuleSet(NULL)
  , locale(Locale::getDefault())
  , collator(NULL)
  , decimalFormatSymbols(NULL)
  , defaultInfinityRule(NULL)
  , defaultNaNRule(NULL)
  , fRoundingMode(DecimalFormat::ERoundingMode::kRoundUnnecessary)
  , lenient(FALSE)
  , lenientParseRules(NULL)
  , localizations(NULL)
  , capitalizationInfoSet(FALSE)
  , capitalizationForUIListMenu(FALSE)
  , capitalizationForStandAlone(FALSE)
  , capitalizationBrkIter(NULL)
{
    init(description, NULL, perror, status);
}

RuleBasedNumberFormat::RuleBasedNumberFormat(const UnicodeString& description,
                         const Locale& aLocale,
                         UParseError& perror,
                         UErrorCode& status)
  : fRuleSets(NULL)
  , ruleSetDescriptions(NULL)
  , numRuleSets(0)
  , defaultRuleSet(NULL)
  , locale(aLocale)
  , collator(NULL)
  , decimalFormatSymbols(NULL)
  , defaultInfinityRule(NULL)
  , defaultNaNRule(NULL)
  , fRoundingMode(DecimalFormat::ERoundingMode::kRoundUnnecessary)
  , lenient(FALSE)
  , lenientParseRules(NULL)
  , localizations(NULL)
  , capitalizationInfoSet(FALSE)
  , capitalizationForUIListMenu(FALSE)
  , capitalizationForStandAlone(FALSE)
  , capitalizationBrkIter(NULL)
{
    init(description, NULL, perror, status);
}

RuleBasedNumberFormat::RuleBasedNumberFormat(URBNFRuleSetTag tag, const Locale& alocale, UErrorCode& status)
  : fRuleSets(NULL)
  , ruleSetDescriptions(NULL)
  , numRuleSets(0)
  , defaultRuleSet(NULL)
  , locale(alocale)
  , collator(NULL)
  , decimalFormatSymbols(NULL)
  , defaultInfinityRule(NULL)
  , defaultNaNRule(NULL)
  , fRoundingMode(DecimalFormat::ERoundingMode::kRoundUnnecessary)
  , lenient(FALSE)
  , lenientParseRules(NULL)
  , localizations(NULL)
  , capitalizationInfoSet(FALSE)
  , capitalizationForUIListMenu(FALSE)
  , capitalizationForStandAlone(FALSE)
  , capitalizationBrkIter(NULL)
{
    if (U_FAILURE(status)) {
        return;
    }

    const char* rules_tag = "RBNFRules";
    const char* fmt_tag = "";
    switch (tag) {
    case URBNF_SPELLOUT: fmt_tag = "SpelloutRules"; break;
    case URBNF_ORDINAL: fmt_tag = "OrdinalRules"; break;
    case URBNF_DURATION: fmt_tag = "DurationRules"; break;
    case URBNF_NUMBERING_SYSTEM: fmt_tag = "NumberingSystemRules"; break;
    default: status = U_ILLEGAL_ARGUMENT_ERROR; return;
    }

    // TODO: read localization info from resource
    LocalizationInfo* locinfo = NULL;

    UResourceBundle* nfrb = ures_open(U_ICUDATA_RBNF, locale.getName(), &status);
    if (U_SUCCESS(status)) {
        setLocaleIDs(ures_getLocaleByType(nfrb, ULOC_VALID_LOCALE, &status),
                     ures_getLocaleByType(nfrb, ULOC_ACTUAL_LOCALE, &status));

        UResourceBundle* rbnfRules = ures_getByKeyWithFallback(nfrb, rules_tag, NULL, &status);
        if (U_FAILURE(status)) {
            ures_close(nfrb);
        }
        UResourceBundle* ruleSets = ures_getByKeyWithFallback(rbnfRules, fmt_tag, NULL, &status);
        if (U_FAILURE(status)) {
            ures_close(rbnfRules);
            ures_close(nfrb);
            return;
        }

        UnicodeString desc;
        while (ures_hasNext(ruleSets)) {
           desc.append(ures_getNextUnicodeString(ruleSets,NULL,&status));
        }
        UParseError perror;

        init(desc, locinfo, perror, status);

        ures_close(ruleSets);
        ures_close(rbnfRules);
    }
    ures_close(nfrb);
}

RuleBasedNumberFormat::RuleBasedNumberFormat(const RuleBasedNumberFormat& rhs)
  : NumberFormat(rhs)
  , fRuleSets(NULL)
  , ruleSetDescriptions(NULL)
  , numRuleSets(0)
  , defaultRuleSet(NULL)
  , locale(rhs.locale)
  , collator(NULL)
  , decimalFormatSymbols(NULL)
  , defaultInfinityRule(NULL)
  , defaultNaNRule(NULL)
  , fRoundingMode(DecimalFormat::ERoundingMode::kRoundUnnecessary)
  , lenient(FALSE)
  , lenientParseRules(NULL)
  , localizations(NULL)
  , capitalizationInfoSet(FALSE)
  , capitalizationForUIListMenu(FALSE)
  , capitalizationForStandAlone(FALSE)
  , capitalizationBrkIter(NULL)
{
    this->operator=(rhs);
}

// --------

RuleBasedNumberFormat&
RuleBasedNumberFormat::operator=(const RuleBasedNumberFormat& rhs)
{
    if (this == &rhs) {
        return *this;
    }
    NumberFormat::operator=(rhs);
    UErrorCode status = U_ZERO_ERROR;
    dispose();
    locale = rhs.locale;
    lenient = rhs.lenient;

    UParseError perror;
    setDecimalFormatSymbols(*rhs.getDecimalFormatSymbols());
    init(rhs.originalDescription, rhs.localizations ? rhs.localizations->ref() : NULL, perror, status);
    setDefaultRuleSet(rhs.getDefaultRuleSetName(), status);
    setRoundingMode(rhs.getRoundingMode());

    capitalizationInfoSet = rhs.capitalizationInfoSet;
    capitalizationForUIListMenu = rhs.capitalizationForUIListMenu;
    capitalizationForStandAlone = rhs.capitalizationForStandAlone;
#if !UCONFIG_NO_BREAK_ITERATION
    capitalizationBrkIter = (rhs.capitalizationBrkIter!=NULL)? rhs.capitalizationBrkIter->clone(): NULL;
#endif

    return *this;
}

RuleBasedNumberFormat::~RuleBasedNumberFormat()
{
    dispose();
}

RuleBasedNumberFormat*
RuleBasedNumberFormat::clone() const
{
    return new RuleBasedNumberFormat(*this);
}

UBool
RuleBasedNumberFormat::operator==(const Format& other) const
{
    if (this == &other) {
        return TRUE;
    }

    if (typeid(*this) == typeid(other)) {
        const RuleBasedNumberFormat& rhs = (const RuleBasedNumberFormat&)other;
        // test for capitalization info equality is adequately handled
        // by the NumberFormat test for fCapitalizationContext equality;
        // the info here is just derived from that.
        if (locale == rhs.locale &&
            lenient == rhs.lenient &&
            (localizations == NULL
                ? rhs.localizations == NULL
                : (rhs.localizations == NULL
                    ? FALSE
                    : *localizations == rhs.localizations))) {

            NFRuleSet** p = fRuleSets;
            NFRuleSet** q = rhs.fRuleSets;
            if (p == NULL) {
                return q == NULL;
            } else if (q == NULL) {
                return FALSE;
            }
            while (*p && *q && (**p == **q)) {
                ++p;
                ++q;
            }
            return *q == NULL && *p == NULL;
        }
    }

    return FALSE;
}

UnicodeString
RuleBasedNumberFormat::getRules() const
{
    UnicodeString result;
    if (fRuleSets != NULL) {
        for (NFRuleSet** p = fRuleSets; *p; ++p) {
            (*p)->appendRules(result);
        }
    }
    return result;
}

UnicodeString
RuleBasedNumberFormat::getRuleSetName(int32_t index) const
{
    if (localizations) {
        UnicodeString string(TRUE, localizations->getRuleSetName(index), (int32_t)-1);
        return string;
    }
    else if (fRuleSets) {
        UnicodeString result;
        for (NFRuleSet** p = fRuleSets; *p; ++p) {
            NFRuleSet* rs = *p;
            if (rs->isPublic()) {
                if (--index == -1) {
                    rs->getName(result);
                    return result;
                }
            }
        }
    }
    UnicodeString empty;
    return empty;
}

int32_t
RuleBasedNumberFormat::getNumberOfRuleSetNames() const
{
    int32_t result = 0;
    if (localizations) {
        result = localizations->getNumberOfRuleSets();
    }
    else if (fRuleSets) {
        for (NFRuleSet** p = fRuleSets; *p; ++p) {
            if ((**p).isPublic()) {
                ++result;
            }
        }
    }
    return result;
}

int32_t
RuleBasedNumberFormat::getNumberOfRuleSetDisplayNameLocales(void) const {
    if (localizations) {
        return localizations->getNumberOfDisplayLocales();
    }
    return 0;
}

Locale
RuleBasedNumberFormat::getRuleSetDisplayNameLocale(int32_t index, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return Locale("");
    }
    if (localizations && index >= 0 && index < localizations->getNumberOfDisplayLocales()) {
        UnicodeString name(TRUE, localizations->getLocaleName(index), -1);
        char buffer[64];
        int32_t cap = name.length() + 1;
        char* bp = buffer;
        if (cap > 64) {
            bp = (char *)uprv_malloc(cap);
            if (bp == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return Locale("");
            }
        }
        name.extract(0, name.length(), bp, cap, UnicodeString::kInvariant);
        Locale retLocale(bp);
        if (bp != buffer) {
            uprv_free(bp);
        }
        return retLocale;
    }
    status = U_ILLEGAL_ARGUMENT_ERROR;
    Locale retLocale;
    return retLocale;
}

UnicodeString
RuleBasedNumberFormat::getRuleSetDisplayName(int32_t index, const Locale& localeParam) {
    if (localizations && index >= 0 && index < localizations->getNumberOfRuleSets()) {
        UnicodeString localeName(localeParam.getBaseName(), -1, UnicodeString::kInvariant);
        int32_t len = localeName.length();
        UChar* localeStr = localeName.getBuffer(len + 1);
        while (len >= 0) {
            localeStr[len] = 0;
            int32_t ix = localizations->indexForLocale(localeStr);
            if (ix >= 0) {
                UnicodeString name(TRUE, localizations->getDisplayName(ix, index), -1);
                return name;
            }

            // trim trailing portion, skipping over ommitted sections
            do { --len;} while (len > 0 && localeStr[len] != 0x005f); // underscore
            while (len > 0 && localeStr[len-1] == 0x005F) --len;
        }
        UnicodeString name(TRUE, localizations->getRuleSetName(index), -1);
        return name;
    }
    UnicodeString bogus;
    bogus.setToBogus();
    return bogus;
}

UnicodeString
RuleBasedNumberFormat::getRuleSetDisplayName(const UnicodeString& ruleSetName, const Locale& localeParam) {
    if (localizations) {
        UnicodeString rsn(ruleSetName);
        int32_t ix = localizations->indexForRuleSet(rsn.getTerminatedBuffer());
        return getRuleSetDisplayName(ix, localeParam);
    }
    UnicodeString bogus;
    bogus.setToBogus();
    return bogus;
}

NFRuleSet*
RuleBasedNumberFormat::findRuleSet(const UnicodeString& name, UErrorCode& status) const
{
    if (U_SUCCESS(status) && fRuleSets) {
        for (NFRuleSet** p = fRuleSets; *p; ++p) {
            NFRuleSet* rs = *p;
            if (rs->isNamed(name)) {
                return rs;
            }
        }
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return NULL;
}

UnicodeString&
RuleBasedNumberFormat::format(const DecimalQuantity &number,
                     UnicodeString& appendTo,
                     FieldPosition& pos,
                     UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    DecimalQuantity copy(number);
    if (copy.fitsInLong()) {
        format(number.toLong(), appendTo, pos, status);
    }
    else {
        copy.roundToMagnitude(0, number::impl::RoundingMode::UNUM_ROUND_HALFEVEN, status);
        if (copy.fitsInLong()) {
            format(number.toDouble(), appendTo, pos, status);
        }
        else {
            // We're outside of our normal range that this framework can handle.
            // The DecimalFormat will provide more accurate results.

            // TODO this section should probably be optimized. The DecimalFormat is shared in ICU4J.
            LocalPointer<NumberFormat> decimalFormat(NumberFormat::createInstance(locale, UNUM_DECIMAL, status), status);
            if (decimalFormat.isNull()) {
                return appendTo;
            }
            Formattable f;
            LocalPointer<DecimalQuantity> decimalQuantity(new DecimalQuantity(number), status);
            if (decimalQuantity.isNull()) {
                return appendTo;
            }
            f.adoptDecimalQuantity(decimalQuantity.orphan()); // f now owns decimalQuantity.
            decimalFormat->format(f, appendTo, pos, status);
        }
    }
    return appendTo;
}

UnicodeString&
RuleBasedNumberFormat::format(int32_t number,
                              UnicodeString& toAppendTo,
                              FieldPosition& pos) const
{
    return format((int64_t)number, toAppendTo, pos);
}


UnicodeString&
RuleBasedNumberFormat::format(int64_t number,
                              UnicodeString& toAppendTo,
                              FieldPosition& /* pos */) const
{
    if (defaultRuleSet) {
        UErrorCode status = U_ZERO_ERROR;
        format(number, defaultRuleSet, toAppendTo, status);
    }
    return toAppendTo;
}


UnicodeString&
RuleBasedNumberFormat::format(double number,
                              UnicodeString& toAppendTo,
                              FieldPosition& /* pos */) const
{
    UErrorCode status = U_ZERO_ERROR;
    if (defaultRuleSet) {
        format(number, *defaultRuleSet, toAppendTo, status);
    }
    return toAppendTo;
}


UnicodeString&
RuleBasedNumberFormat::format(int32_t number,
                              const UnicodeString& ruleSetName,
                              UnicodeString& toAppendTo,
                              FieldPosition& pos,
                              UErrorCode& status) const
{
    return format((int64_t)number, ruleSetName, toAppendTo, pos, status);
}


UnicodeString&
RuleBasedNumberFormat::format(int64_t number,
                              const UnicodeString& ruleSetName,
                              UnicodeString& toAppendTo,
                              FieldPosition& /* pos */,
                              UErrorCode& status) const
{
    if (U_SUCCESS(status)) {
        if (ruleSetName.indexOf(rbnf_gPercentPercent, 2, 0) == 0) {
            // throw new IllegalArgumentException("Can't use internal rule set");
            status = U_ILLEGAL_ARGUMENT_ERROR;
        } else {
            NFRuleSet *rs = findRuleSet(ruleSetName, status);
            if (rs) {
                format(number, rs, toAppendTo, status);
            }
        }
    }
    return toAppendTo;
}


UnicodeString&
RuleBasedNumberFormat::format(double number,
                              const UnicodeString& ruleSetName,
                              UnicodeString& toAppendTo,
                              FieldPosition& /* pos */,
                              UErrorCode& status) const
{
    if (U_SUCCESS(status)) {
        if (ruleSetName.indexOf(rbnf_gPercentPercent, 2, 0) == 0) {
            // throw new IllegalArgumentException("Can't use internal rule set");
            status = U_ILLEGAL_ARGUMENT_ERROR;
        } else {
            NFRuleSet *rs = findRuleSet(ruleSetName, status);
            if (rs) {
                format(number, *rs, toAppendTo, status);
            }
        }
    }
    return toAppendTo;
}

void
RuleBasedNumberFormat::format(double number,
                              NFRuleSet& rs,
                              UnicodeString& toAppendTo,
                              UErrorCode& status) const
{
    int32_t startPos = toAppendTo.length();
    if (getRoundingMode() != DecimalFormat::ERoundingMode::kRoundUnnecessary && !uprv_isNaN(number) && !uprv_isInfinite(number)) {
        DecimalQuantity digitList;
        digitList.setToDouble(number);
        digitList.roundToMagnitude(
                -getMaximumFractionDigits(),
                static_cast<UNumberFormatRoundingMode>(getRoundingMode()),
                status);
        number = digitList.toDouble();
    }
    rs.format(number, toAppendTo, toAppendTo.length(), 0, status);
    adjustForCapitalizationContext(startPos, toAppendTo, status);
}

/**
 * Bottleneck through which all the public format() methods
 * that take a long pass. By the time we get here, we know
 * which rule set we're using to do the formatting.
 * @param number The number to format
 * @param ruleSet The rule set to use to format the number
 * @return The text that resulted from formatting the number
 */
UnicodeString&
RuleBasedNumberFormat::format(int64_t number, NFRuleSet *ruleSet, UnicodeString& toAppendTo, UErrorCode& status) const
{
    // all API format() routines that take a double vector through
    // here.  We have these two identical functions-- one taking a
    // double and one taking a long-- the couple digits of precision
    // that long has but double doesn't (both types are 8 bytes long,
    // but double has to borrow some of the mantissa bits to hold
    // the exponent).
    // Create an empty string buffer where the result will
    // be built, and pass it to the rule set (along with an insertion
    // position of 0 and the number being formatted) to the rule set
    // for formatting

    if (U_SUCCESS(status)) {
        if (number == U_INT64_MIN) {
            // We can't handle this value right now. Provide an accurate default value.

            // TODO this section should probably be optimized. The DecimalFormat is shared in ICU4J.
            NumberFormat *decimalFormat = NumberFormat::createInstance(locale, UNUM_DECIMAL, status);
            if (decimalFormat == nullptr) {
                return toAppendTo;
            }
            Formattable f;
            FieldPosition pos(FieldPosition::DONT_CARE);
            DecimalQuantity *decimalQuantity = new DecimalQuantity();
            if (decimalQuantity == nullptr) {
                status = U_MEMORY_ALLOCATION_ERROR;
                delete decimalFormat;
                return toAppendTo;
            }
            decimalQuantity->setToLong(number);
            f.adoptDecimalQuantity(decimalQuantity); // f now owns decimalQuantity.
            decimalFormat->format(f, toAppendTo, pos, status);
            delete decimalFormat;
        }
        else {
            int32_t startPos = toAppendTo.length();
            ruleSet->format(number, toAppendTo, toAppendTo.length(), 0, status);
            adjustForCapitalizationContext(startPos, toAppendTo, status);
        }
    }
    return toAppendTo;
}

UnicodeString&
RuleBasedNumberFormat::adjustForCapitalizationContext(int32_t startPos,
                                                      UnicodeString& currentResult,
                                                      UErrorCode& status) const
{
#if !UCONFIG_NO_BREAK_ITERATION
    UDisplayContext capitalizationContext = getContext(UDISPCTX_TYPE_CAPITALIZATION, status);
    if (capitalizationContext != UDISPCTX_CAPITALIZATION_NONE && startPos == 0 && currentResult.length() > 0) {
        // capitalize currentResult according to context
        UChar32 ch = currentResult.char32At(0);
        if (u_islower(ch) && U_SUCCESS(status) && capitalizationBrkIter != NULL &&
              ( capitalizationContext == UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE ||
                (capitalizationContext == UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU && capitalizationForUIListMenu) ||
                (capitalizationContext == UDISPCTX_CAPITALIZATION_FOR_STANDALONE && capitalizationForStandAlone)) ) {
            // titlecase first word of currentResult, here use sentence iterator unlike current implementations
            // in LocaleDisplayNamesImpl::adjustForUsageAndContext and RelativeDateFormat::format
            currentResult.toTitle(capitalizationBrkIter, locale, U_TITLECASE_NO_LOWERCASE | U_TITLECASE_NO_BREAK_ADJUSTMENT);
        }
    }
#endif
    return currentResult;
}


void
RuleBasedNumberFormat::parse(const UnicodeString& text,
                             Formattable& result,
                             ParsePosition& parsePosition) const
{
    if (!fRuleSets) {
        parsePosition.setErrorIndex(0);
        return;
    }

    UnicodeString workingText(text, parsePosition.getIndex());
    ParsePosition workingPos(0);

    ParsePosition high_pp(0);
    Formattable high_result;

    for (NFRuleSet** p = fRuleSets; *p; ++p) {
        NFRuleSet *rp = *p;
        if (rp->isPublic() && rp->isParseable()) {
            ParsePosition working_pp(0);
            Formattable working_result;

            rp->parse(workingText, working_pp, kMaxDouble, 0, working_result);
            if (working_pp.getIndex() > high_pp.getIndex()) {
                high_pp = working_pp;
                high_result = working_result;

                if (high_pp.getIndex() == workingText.length()) {
                    break;
                }
            }
        }
    }

    int32_t startIndex = parsePosition.getIndex();
    parsePosition.setIndex(startIndex + high_pp.getIndex());
    if (high_pp.getIndex() > 0) {
        parsePosition.setErrorIndex(-1);
    } else {
        int32_t errorIndex = (high_pp.getErrorIndex()>0)? high_pp.getErrorIndex(): 0;
        parsePosition.setErrorIndex(startIndex + errorIndex);
    }
    result = high_result;
    if (result.getType() == Formattable::kDouble) {
        double d = result.getDouble();
        if (!uprv_isNaN(d) && d == uprv_trunc(d) && INT32_MIN <= d && d <= INT32_MAX) {
            // Note: casting a double to an int when the double is too large or small
            //       to fit the destination is undefined behavior. The explicit range checks,
            //       above, are required. Just casting and checking the result value is undefined.
            result.setLong(static_cast<int32_t>(d));
        }
    }
}

#if !UCONFIG_NO_COLLATION

void
RuleBasedNumberFormat::setLenient(UBool enabled)
{
    lenient = enabled;
    if (!enabled && collator) {
        delete collator;
        collator = NULL;
    }
}

#endif

void
RuleBasedNumberFormat::setDefaultRuleSet(const UnicodeString& ruleSetName, UErrorCode& status) {
    if (U_SUCCESS(status)) {
        if (ruleSetName.isEmpty()) {
          if (localizations) {
              UnicodeString name(TRUE, localizations->getRuleSetName(0), -1);
              defaultRuleSet = findRuleSet(name, status);
          } else {
            initDefaultRuleSet();
          }
        } else if (ruleSetName.startsWith(UNICODE_STRING_SIMPLE("%%"))) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
        } else {
            NFRuleSet* result = findRuleSet(ruleSetName, status);
            if (result != NULL) {
                defaultRuleSet = result;
            }
        }
    }
}

UnicodeString
RuleBasedNumberFormat::getDefaultRuleSetName() const {
    UnicodeString result;
    if (defaultRuleSet && defaultRuleSet->isPublic()) {
        defaultRuleSet->getName(result);
    } else {
        result.setToBogus();
    }
    return result;
}

void
RuleBasedNumberFormat::initDefaultRuleSet()
{
    defaultRuleSet = NULL;
    if (!fRuleSets) {
        return;
    }

    const UnicodeString spellout(UNICODE_STRING_SIMPLE("%spellout-numbering"));
    const UnicodeString ordinal(UNICODE_STRING_SIMPLE("%digits-ordinal"));
    const UnicodeString duration(UNICODE_STRING_SIMPLE("%duration"));

    NFRuleSet**p = &fRuleSets[0];
    while (*p) {
        if ((*p)->isNamed(spellout) || (*p)->isNamed(ordinal) || (*p)->isNamed(duration)) {
            defaultRuleSet = *p;
            return;
        } else {
            ++p;
        }
    }

    defaultRuleSet = *--p;
    if (!defaultRuleSet->isPublic()) {
        while (p != fRuleSets) {
            if ((*--p)->isPublic()) {
                defaultRuleSet = *p;
                break;
            }
        }
    }
}


void
RuleBasedNumberFormat::init(const UnicodeString& rules, LocalizationInfo* localizationInfos,
                            UParseError& pErr, UErrorCode& status)
{
    // TODO: implement UParseError
    uprv_memset(&pErr, 0, sizeof(UParseError));
    // Note: this can leave ruleSets == NULL, so remaining code should check
    if (U_FAILURE(status)) {
        return;
    }

    initializeDecimalFormatSymbols(status);
    initializeDefaultInfinityRule(status);
    initializeDefaultNaNRule(status);
    if (U_FAILURE(status)) {
        return;
    }

    this->localizations = localizationInfos == NULL ? NULL : localizationInfos->ref();

    UnicodeString description(rules);
    if (!description.length()) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    // start by stripping the trailing whitespace from all the rules
    // (this is all the whitespace follwing each semicolon in the
    // description).  This allows us to look for rule-set boundaries
    // by searching for ";%" without having to worry about whitespace
    // between the ; and the %
    stripWhitespace(description);

    // check to see if there's a set of lenient-parse rules.  If there
    // is, pull them out into our temporary holding place for them,
    // and delete them from the description before the real desciption-
    // parsing code sees them
    int32_t lp = description.indexOf(rbnf_gLenientParse, -1, 0);
    if (lp != -1) {
        // we've got to make sure we're not in the middle of a rule
        // (where "%%lenient-parse" would actually get treated as
        // rule text)
        if (lp == 0 || description.charAt(lp - 1) == rbnf_gSemiColon) {
            // locate the beginning and end of the actual collation
            // rules (there may be whitespace between the name and
            // the first token in the description)
            int lpEnd = description.indexOf(rbnf_gSemiPercent, 2, lp);

            if (lpEnd == -1) {
                lpEnd = description.length() - 1;
            }
            int lpStart = lp + u_strlen(rbnf_gLenientParse);
            while (PatternProps::isWhiteSpace(description.charAt(lpStart))) {
                ++lpStart;
            }

            // copy out the lenient-parse rules and delete them
            // from the description
            lenientParseRules = new UnicodeString();
            /* test for NULL */
            if (lenientParseRules == nullptr) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            lenientParseRules->setTo(description, lpStart, lpEnd - lpStart);

            description.remove(lp, lpEnd + 1 - lp);
        }
    }

    // pre-flight parsing the description and count the number of
    // rule sets (";%" marks the end of one rule set and the beginning
    // of the next)
    numRuleSets = 0;
    for (int32_t p = description.indexOf(rbnf_gSemiPercent, 2, 0); p != -1; p = description.indexOf(rbnf_gSemiPercent, 2, p)) {
        ++numRuleSets;
        ++p;
    }
    ++numRuleSets;

    // our rule list is an array of the appropriate size
    fRuleSets = (NFRuleSet **)uprv_malloc((numRuleSets + 1) * sizeof(NFRuleSet *));
    /* test for NULL */
    if (fRuleSets == 0) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    for (int i = 0; i <= numRuleSets; ++i) {
        fRuleSets[i] = NULL;
    }

    // divide up the descriptions into individual rule-set descriptions
    // and store them in a temporary array.  At each step, we also
    // new up a rule set, but all this does is initialize its name
    // and remove it from its description.  We can't actually parse
    // the rest of the descriptions and finish initializing everything
    // because we have to know the names and locations of all the rule
    // sets before we can actually set everything up
    if(!numRuleSets) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    ruleSetDescriptions = new UnicodeString[numRuleSets];
    if (ruleSetDescriptions == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    {
        int curRuleSet = 0;
        int32_t start = 0;
        for (int32_t p = description.indexOf(rbnf_gSemiPercent, 2, 0); p != -1; p = description.indexOf(rbnf_gSemiPercent, 2, start)) {
            ruleSetDescriptions[curRuleSet].setTo(description, start, p + 1 - start);
            fRuleSets[curRuleSet] = new NFRuleSet(this, ruleSetDescriptions, curRuleSet, status);
            if (fRuleSets[curRuleSet] == nullptr) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            ++curRuleSet;
            start = p + 1;
        }
        ruleSetDescriptions[curRuleSet].setTo(description, start, description.length() - start);
        fRuleSets[curRuleSet] = new NFRuleSet(this, ruleSetDescriptions, curRuleSet, status);
        if (fRuleSets[curRuleSet] == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
    }

    // now we can take note of the formatter's default rule set, which
    // is the last public rule set in the description (it's the last
    // rather than the first so that a user can create a new formatter
    // from an existing formatter and change its default behavior just
    // by appending more rule sets to the end)

    // {dlf} Initialization of a fraction rule set requires the default rule
    // set to be known.  For purposes of initialization, this is always the
    // last public rule set, no matter what the localization data says.
    initDefaultRuleSet();

    // finally, we can go back through the temporary descriptions
    // list and finish setting up the substructure (and we throw
    // away the temporary descriptions as we go)
    {
        for (int i = 0; i < numRuleSets; i++) {
            fRuleSets[i]->parseRules(ruleSetDescriptions[i], status);
        }
    }

    // Now that the rules are initialized, the 'real' default rule
    // set can be adjusted by the localization data.

    // The C code keeps the localization array as is, rather than building
    // a separate array of the public rule set names, so we have less work
    // to do here-- but we still need to check the names.

    if (localizationInfos) {
        // confirm the names, if any aren't in the rules, that's an error
        // it is ok if the rules contain public rule sets that are not in this list
        for (int32_t i = 0; i < localizationInfos->getNumberOfRuleSets(); ++i) {
            UnicodeString name(TRUE, localizationInfos->getRuleSetName(i), -1);
            NFRuleSet* rs = findRuleSet(name, status);
            if (rs == NULL) {
                break; // error
            }
            if (i == 0) {
                defaultRuleSet = rs;
            }
        }
    } else {
        defaultRuleSet = getDefaultRuleSet();
    }
    originalDescription = rules;
}

// override the NumberFormat implementation in order to
// lazily initialize relevant items
void
RuleBasedNumberFormat::setContext(UDisplayContext value, UErrorCode& status)
{
    NumberFormat::setContext(value, status);
    if (U_SUCCESS(status)) {
    	if (!capitalizationInfoSet &&
    	        (value==UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU || value==UDISPCTX_CAPITALIZATION_FOR_STANDALONE)) {
    	    initCapitalizationContextInfo(locale);
    	    capitalizationInfoSet = TRUE;
        }
#if !UCONFIG_NO_BREAK_ITERATION
        if ( capitalizationBrkIter == NULL && (value==UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE ||
                (value==UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU && capitalizationForUIListMenu) ||
                (value==UDISPCTX_CAPITALIZATION_FOR_STANDALONE && capitalizationForStandAlone)) ) {
            status = U_ZERO_ERROR;
            capitalizationBrkIter = BreakIterator::createSentenceInstance(locale, status);
            if (U_FAILURE(status)) {
                delete capitalizationBrkIter;
                capitalizationBrkIter = NULL;
            }
        }
#endif
    }
}

void
RuleBasedNumberFormat::initCapitalizationContextInfo(const Locale& thelocale)
{
#if !UCONFIG_NO_BREAK_ITERATION
    const char * localeID = (thelocale != NULL)? thelocale.getBaseName(): NULL;
    UErrorCode status = U_ZERO_ERROR;
    UResourceBundle *rb = ures_open(NULL, localeID, &status);
    rb = ures_getByKeyWithFallback(rb, "contextTransforms", rb, &status);
    rb = ures_getByKeyWithFallback(rb, "number-spellout", rb, &status);
    if (U_SUCCESS(status) && rb != NULL) {
        int32_t len = 0;
        const int32_t * intVector = ures_getIntVector(rb, &len, &status);
        if (U_SUCCESS(status) && intVector != NULL && len >= 2) {
            capitalizationForUIListMenu = static_cast<UBool>(intVector[0]);
            capitalizationForStandAlone = static_cast<UBool>(intVector[1]);
        }
    }
    ures_close(rb);
#endif
}

void
RuleBasedNumberFormat::stripWhitespace(UnicodeString& description)
{
    // iterate through the characters...
    UnicodeString result;

    int start = 0;
    while (start != -1 && start < description.length()) {
        // seek to the first non-whitespace character...
        while (start < description.length()
            && PatternProps::isWhiteSpace(description.charAt(start))) {
            ++start;
        }

        // locate the next semicolon in the text and copy the text from
        // our current position up to that semicolon into the result
        int32_t p = description.indexOf(rbnf_gSemiColon, start);
        if (p == -1) {
            // or if we don't find a semicolon, just copy the rest of
            // the string into the result
            result.append(description, start, description.length() - start);
            start = -1;
        }
        else if (p < description.length()) {
            result.append(description, start, p + 1 - start);
            start = p + 1;
        }

        // when we get here, we've seeked off the end of the string, and
        // we terminate the loop (we continue until *start* is -1 rather
        // than until *p* is -1, because otherwise we'd miss the last
        // rule in the description)
        else {
            start = -1;
        }
    }

    description.setTo(result);
}


void
RuleBasedNumberFormat::dispose()
{
    if (fRuleSets) {
        for (NFRuleSet** p = fRuleSets; *p; ++p) {
            delete *p;
        }
        uprv_free(fRuleSets);
        fRuleSets = NULL;
    }

    if (ruleSetDescriptions) {
        delete [] ruleSetDescriptions;
        ruleSetDescriptions = NULL;
    }

#if !UCONFIG_NO_COLLATION
    delete collator;
#endif
    collator = NULL;

    delete decimalFormatSymbols;
    decimalFormatSymbols = NULL;

    delete defaultInfinityRule;
    defaultInfinityRule = NULL;

    delete defaultNaNRule;
    defaultNaNRule = NULL;

    delete lenientParseRules;
    lenientParseRules = NULL;

#if !UCONFIG_NO_BREAK_ITERATION
    delete capitalizationBrkIter;
    capitalizationBrkIter = NULL;
#endif

    if (localizations) {
        localizations = localizations->unref();
    }
}


//-----------------------------------------------------------------------
// package-internal API
//-----------------------------------------------------------------------

/**
 * Returns the collator to use for lenient parsing.  The collator is lazily created:
 * this function creates it the first time it's called.
 * @return The collator to use for lenient parsing, or null if lenient parsing
 * is turned off.
*/
const RuleBasedCollator*
RuleBasedNumberFormat::getCollator() const
{
#if !UCONFIG_NO_COLLATION
    if (!fRuleSets) {
        return NULL;
    }

    // lazy-evaluate the collator
    if (collator == NULL && lenient) {
        // create a default collator based on the formatter's locale,
        // then pull out that collator's rules, append any additional
        // rules specified in the description, and create a _new_
        // collator based on the combination of those rules

        UErrorCode status = U_ZERO_ERROR;

        Collator* temp = Collator::createInstance(locale, status);
        RuleBasedCollator* newCollator;
        if (U_SUCCESS(status) && (newCollator = dynamic_cast<RuleBasedCollator*>(temp)) != NULL) {
            if (lenientParseRules) {
                UnicodeString rules(newCollator->getRules());
                rules.append(*lenientParseRules);

                newCollator = new RuleBasedCollator(rules, status);
                // Exit if newCollator could not be created.
                if (newCollator == NULL) {
                    return NULL;
                }
            } else {
                temp = NULL;
            }
            if (U_SUCCESS(status)) {
                newCollator->setAttribute(UCOL_DECOMPOSITION_MODE, UCOL_ON, status);
                // cast away const
                ((RuleBasedNumberFormat*)this)->collator = newCollator;
            } else {
                delete newCollator;
            }
        }
        delete temp;
    }
#endif

    // if lenient-parse mode is off, this will be null
    // (see setLenientParseMode())
    return collator;
}


DecimalFormatSymbols*
RuleBasedNumberFormat::initializeDecimalFormatSymbols(UErrorCode &status)
{
    // lazy-evaluate the DecimalFormatSymbols object.  This object
    // is shared by all DecimalFormat instances belonging to this
    // formatter
    if (decimalFormatSymbols == nullptr) {
        LocalPointer<DecimalFormatSymbols> temp(new DecimalFormatSymbols(locale, status), status);
        if (U_SUCCESS(status)) {
            decimalFormatSymbols = temp.orphan();
        }
    }
    return decimalFormatSymbols;
}

/**
 * Returns the DecimalFormatSymbols object that should be used by all DecimalFormat
 * instances owned by this formatter.
*/
const DecimalFormatSymbols*
RuleBasedNumberFormat::getDecimalFormatSymbols() const
{
    return decimalFormatSymbols;
}

NFRule*
RuleBasedNumberFormat::initializeDefaultInfinityRule(UErrorCode &status)
{
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (defaultInfinityRule == NULL) {
        UnicodeString rule(UNICODE_STRING_SIMPLE("Inf: "));
        rule.append(getDecimalFormatSymbols()->getSymbol(DecimalFormatSymbols::kInfinitySymbol));
        LocalPointer<NFRule> temp(new NFRule(this, rule, status), status);
        if (U_SUCCESS(status)) {
            defaultInfinityRule = temp.orphan();
        }
    }
    return defaultInfinityRule;
}

const NFRule*
RuleBasedNumberFormat::getDefaultInfinityRule() const
{
    return defaultInfinityRule;
}

NFRule*
RuleBasedNumberFormat::initializeDefaultNaNRule(UErrorCode &status)
{
    if (U_FAILURE(status)) {
        return nullptr;
    }
    if (defaultNaNRule == nullptr) {
        UnicodeString rule(UNICODE_STRING_SIMPLE("NaN: "));
        rule.append(getDecimalFormatSymbols()->getSymbol(DecimalFormatSymbols::kNaNSymbol));
        LocalPointer<NFRule> temp(new NFRule(this, rule, status), status);
        if (U_SUCCESS(status)) {
            defaultNaNRule = temp.orphan();
        }
    }
    return defaultNaNRule;
}

const NFRule*
RuleBasedNumberFormat::getDefaultNaNRule() const
{
    return defaultNaNRule;
}

// De-owning the current localized symbols and adopt the new symbols.
void
RuleBasedNumberFormat::adoptDecimalFormatSymbols(DecimalFormatSymbols* symbolsToAdopt)
{
    if (symbolsToAdopt == NULL) {
        return; // do not allow caller to set decimalFormatSymbols to NULL
    }

    if (decimalFormatSymbols != NULL) {
        delete decimalFormatSymbols;
    }

    decimalFormatSymbols = symbolsToAdopt;

    {
        // Apply the new decimalFormatSymbols by reparsing the rulesets
        UErrorCode status = U_ZERO_ERROR;

        delete defaultInfinityRule;
        defaultInfinityRule = NULL;
        initializeDefaultInfinityRule(status); // Reset with the new DecimalFormatSymbols

        delete defaultNaNRule;
        defaultNaNRule = NULL;
        initializeDefaultNaNRule(status); // Reset with the new DecimalFormatSymbols

        if (fRuleSets) {
            for (int32_t i = 0; i < numRuleSets; i++) {
                fRuleSets[i]->setDecimalFormatSymbols(*symbolsToAdopt, status);
            }
        }
    }
}

// Setting the symbols is equivalent to adopting a newly created localized symbols.
void
RuleBasedNumberFormat::setDecimalFormatSymbols(const DecimalFormatSymbols& symbols)
{
    adoptDecimalFormatSymbols(new DecimalFormatSymbols(symbols));
}

PluralFormat *
RuleBasedNumberFormat::createPluralFormat(UPluralType pluralType,
                                          const UnicodeString &pattern,
                                          UErrorCode& status) const
{
    auto *pf = new PluralFormat(locale, pluralType, pattern, status);
    if (pf == nullptr) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return pf;
}

/**
 * Get the rounding mode.
 * @return A rounding mode
 */
DecimalFormat::ERoundingMode RuleBasedNumberFormat::getRoundingMode() const {
    return fRoundingMode;
}

/**
 * Set the rounding mode.  This has no effect unless the rounding
 * increment is greater than zero.
 * @param roundingMode A rounding mode
 */
void RuleBasedNumberFormat::setRoundingMode(DecimalFormat::ERoundingMode roundingMode) {
    fRoundingMode = roundingMode;
}

U_NAMESPACE_END

/* U_HAVE_RBNF */
#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2013, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

  // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING








U_NAMESPACE_BEGIN

/**
 * A struct representing a time zone transition
 */
struct Transition {
    UDate time;
    TimeZoneRule* from;
    TimeZoneRule* to;
};

static UBool compareRules(UVector* rules1, UVector* rules2) {
    if (rules1 == NULL && rules2 == NULL) {
        return TRUE;
    } else if (rules1 == NULL || rules2 == NULL) {
        return FALSE;
    }
    int32_t size = rules1->size();
    if (size != rules2->size()) {
        return FALSE;
    }
    for (int32_t i = 0; i < size; i++) {
        TimeZoneRule *r1 = (TimeZoneRule*)rules1->elementAt(i);
        TimeZoneRule *r2 = (TimeZoneRule*)rules2->elementAt(i);
        if (*r1 != *r2) {
            return FALSE;
        }
    }
    return TRUE;
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(RuleBasedTimeZone)

RuleBasedTimeZone::RuleBasedTimeZone(const UnicodeString& id, InitialTimeZoneRule* initialRule)
: BasicTimeZone(id), fInitialRule(initialRule), fHistoricRules(NULL), fFinalRules(NULL),
  fHistoricTransitions(NULL), fUpToDate(FALSE) {
}

RuleBasedTimeZone::RuleBasedTimeZone(const RuleBasedTimeZone& source)
: BasicTimeZone(source), fInitialRule(source.fInitialRule->clone()),
  fHistoricTransitions(NULL), fUpToDate(FALSE) {
    fHistoricRules = copyRules(source.fHistoricRules);
    fFinalRules = copyRules(source.fFinalRules);
    if (source.fUpToDate) {
        UErrorCode status = U_ZERO_ERROR;
        complete(status);
    }
}

RuleBasedTimeZone::~RuleBasedTimeZone() {
    deleteTransitions();
    deleteRules();
}

RuleBasedTimeZone&
RuleBasedTimeZone::operator=(const RuleBasedTimeZone& right) {
    if (*this != right) {
        BasicTimeZone::operator=(right);
        deleteRules();
        fInitialRule = right.fInitialRule->clone();
        fHistoricRules = copyRules(right.fHistoricRules);
        fFinalRules = copyRules(right.fFinalRules);
        deleteTransitions();
        fUpToDate = FALSE;
    }
    return *this;
}

UBool
RuleBasedTimeZone::operator==(const TimeZone& that) const {
    if (this == &that) {
        return TRUE;
    }
    if (typeid(*this) != typeid(that)
        || BasicTimeZone::operator==(that) == FALSE) {
        return FALSE;
    }
    RuleBasedTimeZone *rbtz = (RuleBasedTimeZone*)&that;
    if (*fInitialRule != *(rbtz->fInitialRule)) {
        return FALSE;
    }
    if (compareRules(fHistoricRules, rbtz->fHistoricRules)
        && compareRules(fFinalRules, rbtz->fFinalRules)) {
        return TRUE;
    }
    return FALSE;
}

UBool
RuleBasedTimeZone::operator!=(const TimeZone& that) const {
    return !operator==(that);
}

void
RuleBasedTimeZone::addTransitionRule(TimeZoneRule* rule, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    AnnualTimeZoneRule* atzrule = dynamic_cast<AnnualTimeZoneRule*>(rule);
    if (atzrule != NULL && atzrule->getEndYear() == AnnualTimeZoneRule::MAX_YEAR) {
        // A final rule
        if (fFinalRules == NULL) {
            fFinalRules = new UVector(status);
            if (U_FAILURE(status)) {
                return;
            }
        } else if (fFinalRules->size() >= 2) {
            // Cannot handle more than two final rules
            status = U_INVALID_STATE_ERROR;
            return;
        }
        fFinalRules->addElement((void*)rule, status);
    } else {
        // Non-final rule
        if (fHistoricRules == NULL) {
            fHistoricRules = new UVector(status);
            if (U_FAILURE(status)) {
                return;
            }
        }
        fHistoricRules->addElement((void*)rule, status);
    }
    // Mark dirty, so transitions are recalculated at next complete() call
    fUpToDate = FALSE;
}


void
RuleBasedTimeZone::completeConst(UErrorCode& status) const {
    static UMutex gLock;
    if (U_FAILURE(status)) {
        return;
    }
    umtx_lock(&gLock);
    if (!fUpToDate) {
        RuleBasedTimeZone *ncThis = const_cast<RuleBasedTimeZone*>(this);
        ncThis->complete(status);
    }
    umtx_unlock(&gLock);
}

void
RuleBasedTimeZone::complete(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    if (fUpToDate) {
        return;
    }
    // Make sure either no final rules or a pair of AnnualTimeZoneRules
    // are available.
    if (fFinalRules != NULL && fFinalRules->size() != 2) {
        status = U_INVALID_STATE_ERROR;
        return;
    }

    UBool *done = NULL;
    // Create a TimezoneTransition and add to the list
    if (fHistoricRules != NULL || fFinalRules != NULL) {
        TimeZoneRule *curRule = fInitialRule;
        UDate lastTransitionTime = MIN_MILLIS;

        // Build the transition array which represents historical time zone
        // transitions.
        if (fHistoricRules != NULL && fHistoricRules->size() > 0) {
            int32_t i;
            int32_t historicCount = fHistoricRules->size();
            done = (UBool*)uprv_malloc(sizeof(UBool) * historicCount);
            if (done == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                goto cleanup;
            }
            for (i = 0; i < historicCount; i++) {
                done[i] = FALSE;
            }
            while (TRUE) {
                int32_t curStdOffset = curRule->getRawOffset();
                int32_t curDstSavings = curRule->getDSTSavings();
                UDate nextTransitionTime = MAX_MILLIS;
                TimeZoneRule *nextRule = NULL;
                TimeZoneRule *r = NULL;
                UBool avail;
                UDate tt;
                UnicodeString curName, name;
                curRule->getName(curName);

                for (i = 0; i < historicCount; i++) {
                    if (done[i]) {
                        continue;
                    }
                    r = (TimeZoneRule*)fHistoricRules->elementAt(i);
                    avail = r->getNextStart(lastTransitionTime, curStdOffset, curDstSavings, false, tt);
                    if (!avail) {
                        // No more transitions from this rule - skip this rule next time
                        done[i] = TRUE;
                    } else {
                        r->getName(name);
                        if (*r == *curRule ||
                            (name == curName && r->getRawOffset() == curRule->getRawOffset()
                            && r->getDSTSavings() == curRule->getDSTSavings())) {
                            continue;
                        }
                        if (tt < nextTransitionTime) {
                            nextTransitionTime = tt;
                            nextRule = r;
                        }
                    }
                }

                if (nextRule ==  NULL) {
                    // Check if all historic rules are done
                    UBool bDoneAll = TRUE;
                    for (int32_t j = 0; j < historicCount; j++) {
                        if (!done[j]) {
                            bDoneAll = FALSE;
                            break;
                        }
                    }
                    if (bDoneAll) {
                        break;
                    }
                }

                if (fFinalRules != NULL) {
                    // Check if one of final rules has earlier transition date
                    for (i = 0; i < 2 /* fFinalRules->size() */; i++) {
                        TimeZoneRule *fr = (TimeZoneRule*)fFinalRules->elementAt(i);
                        if (*fr == *curRule) {
                            continue;
                        }
                        r = (TimeZoneRule*)fFinalRules->elementAt(i);
                        avail = r->getNextStart(lastTransitionTime, curStdOffset, curDstSavings, false, tt);
                        if (avail) {
                            if (tt < nextTransitionTime) {
                                nextTransitionTime = tt;
                                nextRule = r;
                            }
                        }
                    }
                }

                if (nextRule == NULL) {
                    // Nothing more
                    break;
                }

                if (fHistoricTransitions == NULL) {
                    fHistoricTransitions = new UVector(status);
                    if (U_FAILURE(status)) {
                        goto cleanup;
                    }
                }
                Transition *trst = (Transition*)uprv_malloc(sizeof(Transition));
                if (trst == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                    goto cleanup;
                }
                trst->time = nextTransitionTime;
                trst->from = curRule;
                trst->to = nextRule;
                fHistoricTransitions->addElement(trst, status);
                if (U_FAILURE(status)) {
                    goto cleanup;
                }
                lastTransitionTime = nextTransitionTime;
                curRule = nextRule;
            }
        }
        if (fFinalRules != NULL) {
            if (fHistoricTransitions == NULL) {
                fHistoricTransitions = new UVector(status);
                if (U_FAILURE(status)) {
                    goto cleanup;
                }
            }
            // Append the first transition for each
            TimeZoneRule *rule0 = (TimeZoneRule*)fFinalRules->elementAt(0);
            TimeZoneRule *rule1 = (TimeZoneRule*)fFinalRules->elementAt(1);
            UDate tt0, tt1;
            UBool avail0 = rule0->getNextStart(lastTransitionTime, curRule->getRawOffset(), curRule->getDSTSavings(), false, tt0);
            UBool avail1 = rule1->getNextStart(lastTransitionTime, curRule->getRawOffset(), curRule->getDSTSavings(), false, tt1);
            if (!avail0 || !avail1) {
                // Should not happen, because both rules are permanent
                status = U_INVALID_STATE_ERROR;
                goto cleanup;
            }
            Transition *final0 = (Transition*)uprv_malloc(sizeof(Transition));
            if (final0 == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                goto cleanup;
            }
            Transition *final1 = (Transition*)uprv_malloc(sizeof(Transition));
            if (final1 == NULL) {
                uprv_free(final0);
                status = U_MEMORY_ALLOCATION_ERROR;
                goto cleanup;
            }
            if (tt0 < tt1) {
                final0->time = tt0;
                final0->from = curRule;
                final0->to = rule0;
                rule1->getNextStart(tt0, rule0->getRawOffset(), rule0->getDSTSavings(), false, final1->time);
                final1->from = rule0;
                final1->to = rule1;
            } else {
                final0->time = tt1;
                final0->from = curRule;
                final0->to = rule1;
                rule0->getNextStart(tt1, rule1->getRawOffset(), rule1->getDSTSavings(), false, final1->time);
                final1->from = rule1;
                final1->to = rule0;
            }
            fHistoricTransitions->addElement(final0, status);
            if (U_FAILURE(status)) {
                goto cleanup;
            }
            fHistoricTransitions->addElement(final1, status);
            if (U_FAILURE(status)) {
                goto cleanup;
            }
        }
    }
    fUpToDate = TRUE;
    if (done != NULL) {
        uprv_free(done);
    }
    return;

cleanup:
    deleteTransitions();
    if (done != NULL) {
        uprv_free(done);
    }
    fUpToDate = FALSE;
}

RuleBasedTimeZone*
RuleBasedTimeZone::clone() const {
    return new RuleBasedTimeZone(*this);
}

int32_t
RuleBasedTimeZone::getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                             uint8_t dayOfWeek, int32_t millis, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return 0;
    }
    if (month < UCAL_JANUARY || month > UCAL_DECEMBER) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    } else {
        return getOffset(era, year, month, day, dayOfWeek, millis,
                         Grego::monthLength(year, month), status);
    }
}

int32_t
RuleBasedTimeZone::getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                             uint8_t /*dayOfWeek*/, int32_t millis,
                             int32_t /*monthLength*/, UErrorCode& status) const {
    // dayOfWeek and monthLength are unused
    if (U_FAILURE(status)) {
        return 0;
    }
    if (era == GregorianCalendar::BC) {
        // Convert to extended year
        year = 1 - year;
    }
    int32_t rawOffset, dstOffset;
    UDate time = (UDate)Grego::fieldsToDay(year, month, day) * U_MILLIS_PER_DAY + millis;
    getOffsetInternal(time, TRUE, kDaylight, kStandard, rawOffset, dstOffset, status);
    if (U_FAILURE(status)) {
        return 0;
    }
    return (rawOffset + dstOffset);
}

void
RuleBasedTimeZone::getOffset(UDate date, UBool local, int32_t& rawOffset,
                             int32_t& dstOffset, UErrorCode& status) const {
    getOffsetInternal(date, local, kFormer, kLatter, rawOffset, dstOffset, status);
}

void
RuleBasedTimeZone::getOffsetFromLocal(UDate date, int32_t nonExistingTimeOpt, int32_t duplicatedTimeOpt,
                                      int32_t& rawOffset, int32_t& dstOffset, UErrorCode& status) const {
    getOffsetInternal(date, TRUE, nonExistingTimeOpt, duplicatedTimeOpt, rawOffset, dstOffset, status);
}


/*
 * The internal getOffset implementation
 */
void
RuleBasedTimeZone::getOffsetInternal(UDate date, UBool local,
                                     int32_t NonExistingTimeOpt, int32_t DuplicatedTimeOpt,
                                     int32_t& rawOffset, int32_t& dstOffset,
                                     UErrorCode& status) const {
    rawOffset = 0;
    dstOffset = 0;

    if (U_FAILURE(status)) {
        return;
    }
    if (!fUpToDate) {
        // Transitions are not yet resolved.  We cannot do it here
        // because this method is const.  Thus, do nothing and return
        // error status.
        status = U_INVALID_STATE_ERROR;
        return;
    }
    const TimeZoneRule *rule = NULL;
    if (fHistoricTransitions == NULL) {
        rule = fInitialRule;
    } else {
        UDate tstart = getTransitionTime((Transition*)fHistoricTransitions->elementAt(0),
            local, NonExistingTimeOpt, DuplicatedTimeOpt);
        if (date < tstart) {
            rule = fInitialRule;
        } else {
            int32_t idx = fHistoricTransitions->size() - 1;
            UDate tend = getTransitionTime((Transition*)fHistoricTransitions->elementAt(idx),
                local, NonExistingTimeOpt, DuplicatedTimeOpt);
            if (date > tend) {
                if (fFinalRules != NULL) {
                    rule = findRuleInFinal(date, local, NonExistingTimeOpt, DuplicatedTimeOpt);
                }
                if (rule == NULL) {
                    // no final rules or the given time is before the first transition
                    // specified by the final rules -> use the last rule 
                    rule = ((Transition*)fHistoricTransitions->elementAt(idx))->to;
                }
            } else {
                // Find a historical transition
                while (idx >= 0) {
                    if (date >= getTransitionTime((Transition*)fHistoricTransitions->elementAt(idx),
                        local, NonExistingTimeOpt, DuplicatedTimeOpt)) {
                        break;
                    }
                    idx--;
                }
                rule = ((Transition*)fHistoricTransitions->elementAt(idx))->to;
            }
        }
    }
    if (rule != NULL) {
        rawOffset = rule->getRawOffset();
        dstOffset = rule->getDSTSavings();
    }
}

void
RuleBasedTimeZone::setRawOffset(int32_t /*offsetMillis*/) {
    // We don't support this operation at this moment.
    // Nothing to do!
}

int32_t
RuleBasedTimeZone::getRawOffset(void) const {
    // Note: This implementation returns standard GMT offset
    // as of current time.
    UErrorCode status = U_ZERO_ERROR;
    int32_t raw, dst;
    getOffset(uprv_getUTCtime() * U_MILLIS_PER_SECOND,
        FALSE, raw, dst, status);
    return raw;
}

UBool
RuleBasedTimeZone::useDaylightTime(void) const {
    // Note: This implementation returns true when
    // daylight saving time is used as of now or
    // after the next transition.
    UErrorCode status = U_ZERO_ERROR;
    UDate now = uprv_getUTCtime() * U_MILLIS_PER_SECOND;
    int32_t raw, dst;
    getOffset(now, FALSE, raw, dst, status);
    if (dst != 0) {
        return TRUE;
    }
    // If DST is not used now, check if DST is used after the next transition
    UDate time;
    TimeZoneRule *from, *to;
    UBool avail = findNext(now, FALSE, time, from, to);
    if (avail && to->getDSTSavings() != 0) {
        return TRUE;
    }
    return FALSE;
}

UBool
RuleBasedTimeZone::inDaylightTime(UDate date, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    int32_t raw, dst;
    getOffset(date, FALSE, raw, dst, status);
    if (dst != 0) {
        return TRUE;
    }
    return FALSE;
}

UBool
RuleBasedTimeZone::hasSameRules(const TimeZone& other) const {
    if (this == &other) {
        return TRUE;
    }
    if (typeid(*this) != typeid(other)) {
        return FALSE;
    }
    const RuleBasedTimeZone& that = (const RuleBasedTimeZone&)other;
    if (*fInitialRule != *(that.fInitialRule)) {
        return FALSE;
    }
    if (compareRules(fHistoricRules, that.fHistoricRules)
        && compareRules(fFinalRules, that.fFinalRules)) {
        return TRUE;
    }
    return FALSE;
}

UBool
RuleBasedTimeZone::getNextTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const {
    UErrorCode status = U_ZERO_ERROR;
    completeConst(status);
    if (U_FAILURE(status)) {
        return FALSE;
    }
    UDate transitionTime;
    TimeZoneRule *fromRule, *toRule;
    UBool found = findNext(base, inclusive, transitionTime, fromRule, toRule);
    if (found) {
        result.setTime(transitionTime);
        result.setFrom((const TimeZoneRule&)*fromRule);
        result.setTo((const TimeZoneRule&)*toRule);
        return TRUE;
    }
    return FALSE;
}

UBool
RuleBasedTimeZone::getPreviousTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const {
    UErrorCode status = U_ZERO_ERROR;
    completeConst(status);
    if (U_FAILURE(status)) {
        return FALSE;
    }
    UDate transitionTime;
    TimeZoneRule *fromRule, *toRule;
    UBool found = findPrev(base, inclusive, transitionTime, fromRule, toRule);
    if (found) {
        result.setTime(transitionTime);
        result.setFrom((const TimeZoneRule&)*fromRule);
        result.setTo((const TimeZoneRule&)*toRule);
        return TRUE;
    }
    return FALSE;
}

int32_t
RuleBasedTimeZone::countTransitionRules(UErrorCode& /*status*/) const {
    int32_t count = 0;
    if (fHistoricRules != NULL) {
        count += fHistoricRules->size();
    }
    if (fFinalRules != NULL) {
        count += fFinalRules->size();
    }
    return count;
}

void
RuleBasedTimeZone::getTimeZoneRules(const InitialTimeZoneRule*& initial,
                                    const TimeZoneRule* trsrules[],
                                    int32_t& trscount,
                                    UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    // Initial rule
    initial = fInitialRule;

    // Transition rules
    int32_t cnt = 0;
    int32_t idx;
    if (fHistoricRules != NULL && cnt < trscount) {
        int32_t historicCount = fHistoricRules->size();
        idx = 0;
        while (cnt < trscount && idx < historicCount) {
            trsrules[cnt++] = (const TimeZoneRule*)fHistoricRules->elementAt(idx++);
        }
    }
    if (fFinalRules != NULL && cnt < trscount) {
        int32_t finalCount = fFinalRules->size();
        idx = 0;
        while (cnt < trscount && idx < finalCount) {
            trsrules[cnt++] = (const TimeZoneRule*)fFinalRules->elementAt(idx++);
        }
    }
    // Set the result length
    trscount = cnt;
}

void
RuleBasedTimeZone::deleteRules(void) {
    delete fInitialRule;
    fInitialRule = NULL;
    if (fHistoricRules != NULL) {
        while (!fHistoricRules->isEmpty()) {
            delete (TimeZoneRule*)(fHistoricRules->orphanElementAt(0));
        }
        delete fHistoricRules;
        fHistoricRules = NULL;
    }
    if (fFinalRules != NULL) {
        while (!fFinalRules->isEmpty()) {
            delete (AnnualTimeZoneRule*)(fFinalRules->orphanElementAt(0));
        }
        delete fFinalRules;
        fFinalRules = NULL;
    }
}

void
RuleBasedTimeZone::deleteTransitions(void) {
    if (fHistoricTransitions != NULL) {
        while (!fHistoricTransitions->isEmpty()) {
            Transition *trs = (Transition*)fHistoricTransitions->orphanElementAt(0);
            uprv_free(trs);
        }
        delete fHistoricTransitions;
    }
    fHistoricTransitions = NULL;
}

UVector*
RuleBasedTimeZone::copyRules(UVector* source) {
    if (source == NULL) {
        return NULL;
    }
    UErrorCode ec = U_ZERO_ERROR;
    int32_t size = source->size();
    UVector *rules = new UVector(size, ec);
    if (U_FAILURE(ec)) {
        return NULL;
    }
    int32_t i;
    for (i = 0; i < size; i++) {
        rules->addElement(((TimeZoneRule*)source->elementAt(i))->clone(), ec);
        if (U_FAILURE(ec)) {
            break;
        }
    }
    if (U_FAILURE(ec)) {
        // In case of error, clean up
        for (i = 0; i < rules->size(); i++) {
            TimeZoneRule *rule = (TimeZoneRule*)rules->orphanElementAt(i);
            delete rule;
        }
        delete rules;
        return NULL;
    }
    return rules;
}

TimeZoneRule*
RuleBasedTimeZone::findRuleInFinal(UDate date, UBool local,
                                   int32_t NonExistingTimeOpt, int32_t DuplicatedTimeOpt) const {
    if (fFinalRules == NULL) {
        return NULL;
    }

    AnnualTimeZoneRule* fr0 = (AnnualTimeZoneRule*)fFinalRules->elementAt(0);
    AnnualTimeZoneRule* fr1 = (AnnualTimeZoneRule*)fFinalRules->elementAt(1);
    if (fr0 == NULL || fr1 == NULL) {
        return NULL;
    }

    UDate start0, start1;
    UDate base;
    int32_t localDelta;

    base = date;
    if (local) {
        localDelta = getLocalDelta(fr1->getRawOffset(), fr1->getDSTSavings(),
                                   fr0->getRawOffset(), fr0->getDSTSavings(),
                                   NonExistingTimeOpt, DuplicatedTimeOpt);
        base -= localDelta;
    }
    UBool avail0 = fr0->getPreviousStart(base, fr1->getRawOffset(), fr1->getDSTSavings(), TRUE, start0);

    base = date;
    if (local) {
        localDelta = getLocalDelta(fr0->getRawOffset(), fr0->getDSTSavings(),
                                   fr1->getRawOffset(), fr1->getDSTSavings(),
                                   NonExistingTimeOpt, DuplicatedTimeOpt);
        base -= localDelta;
    }
    UBool avail1 = fr1->getPreviousStart(base, fr0->getRawOffset(), fr0->getDSTSavings(), TRUE, start1);

    if (!avail0 || !avail1) {
        if (avail0) {
            return fr0;
        } else if (avail1) {
            return fr1;
        }
        // Both rules take effect after the given time
        return NULL;
    }

    return (start0 > start1) ? fr0 : fr1;
}

UBool
RuleBasedTimeZone::findNext(UDate base, UBool inclusive, UDate& transitionTime,
                            TimeZoneRule*& fromRule, TimeZoneRule*& toRule) const {
    if (fHistoricTransitions == NULL) {
        return FALSE;
    }
    UBool isFinal = FALSE;
    UBool found = FALSE;
    Transition result;
    Transition *tzt = (Transition*)fHistoricTransitions->elementAt(0);
    UDate tt = tzt->time;
    if (tt > base || (inclusive && tt == base)) {
        result = *tzt;
        found = TRUE;
    } else {
        int32_t idx = fHistoricTransitions->size() - 1;        
        tzt = (Transition*)fHistoricTransitions->elementAt(idx);
        tt = tzt->time;
        if (inclusive && tt == base) {
            result = *tzt;
            found = TRUE;
        } else if (tt <= base) {
            if (fFinalRules != NULL) {
                // Find a transion time with finalRules
                TimeZoneRule *r0 = (TimeZoneRule*)fFinalRules->elementAt(0);
                TimeZoneRule *r1 = (TimeZoneRule*)fFinalRules->elementAt(1);
                UDate start0, start1;
                UBool avail0 = r0->getNextStart(base, r1->getRawOffset(), r1->getDSTSavings(), inclusive, start0);
                UBool avail1 = r1->getNextStart(base, r0->getRawOffset(), r0->getDSTSavings(), inclusive, start1);
                //  avail0/avail1 should be always TRUE
                if (!avail0 && !avail1) {
                    return FALSE;
                }
                if (!avail1 || start0 < start1) {
                    result.time = start0;
                    result.from = r1;
                    result.to = r0;
                } else {
                    result.time = start1;
                    result.from = r0;
                    result.to = r1;
                }
                isFinal = TRUE;
                found = TRUE;
            }
        } else {
            // Find a transition within the historic transitions
            idx--;
            Transition *prev = tzt;
            while (idx > 0) {
                tzt = (Transition*)fHistoricTransitions->elementAt(idx);
                tt = tzt->time;
                if (tt < base || (!inclusive && tt == base)) {
                    break;
                }
                idx--;
                prev = tzt;
            }
            result.time = prev->time;
            result.from = prev->from;
            result.to = prev->to;
            found = TRUE;
        }
    }
    if (found) {
        // For now, this implementation ignore transitions with only zone name changes.
        if (result.from->getRawOffset() == result.to->getRawOffset()
            && result.from->getDSTSavings() == result.to->getDSTSavings()) {
            if (isFinal) {
                return FALSE;
            } else {
                // No offset changes.  Try next one if not final
                return findNext(result.time, FALSE /* always exclusive */,
                    transitionTime, fromRule, toRule);
            }
        }
        transitionTime = result.time;
        fromRule = result.from;
        toRule = result.to;
        return TRUE;
    }
    return FALSE;
}

UBool
RuleBasedTimeZone::findPrev(UDate base, UBool inclusive, UDate& transitionTime,
                            TimeZoneRule*& fromRule, TimeZoneRule*& toRule) const {
    if (fHistoricTransitions == NULL) {
        return FALSE;
    }
    UBool found = FALSE;
    Transition result;
    Transition *tzt = (Transition*)fHistoricTransitions->elementAt(0);
    UDate tt = tzt->time;
    if (inclusive && tt == base) {
        result = *tzt;
        found = TRUE;
    } else if (tt < base) {
        int32_t idx = fHistoricTransitions->size() - 1;        
        tzt = (Transition*)fHistoricTransitions->elementAt(idx);
        tt = tzt->time;
        if (inclusive && tt == base) {
            result = *tzt;
            found = TRUE;
        } else if (tt < base) {
            if (fFinalRules != NULL) {
                // Find a transion time with finalRules
                TimeZoneRule *r0 = (TimeZoneRule*)fFinalRules->elementAt(0);
                TimeZoneRule *r1 = (TimeZoneRule*)fFinalRules->elementAt(1);
                UDate start0, start1;
                UBool avail0 = r0->getPreviousStart(base, r1->getRawOffset(), r1->getDSTSavings(), inclusive, start0);
                UBool avail1 = r1->getPreviousStart(base, r0->getRawOffset(), r0->getDSTSavings(), inclusive, start1);
                //  avail0/avail1 should be always TRUE
                if (!avail0 && !avail1) {
                    return FALSE;
                }
                if (!avail1 || start0 > start1) {
                    result.time = start0;
                    result.from = r1;
                    result.to = r0;
                } else {
                    result.time = start1;
                    result.from = r0;
                    result.to = r1;
                }
            } else {
                result = *tzt;
            }
            found = TRUE;
        } else {
            // Find a transition within the historic transitions
            idx--;
            while (idx >= 0) {
                tzt = (Transition*)fHistoricTransitions->elementAt(idx);
                tt = tzt->time;
                if (tt < base || (inclusive && tt == base)) {
                    break;
                }
                idx--;
            }
            result = *tzt;
            found = TRUE;
        }
    }
    if (found) {
        // For now, this implementation ignore transitions with only zone name changes.
        if (result.from->getRawOffset() == result.to->getRawOffset()
            && result.from->getDSTSavings() == result.to->getDSTSavings()) {
            // No offset changes.  Try next one if not final
            return findPrev(result.time, FALSE /* always exclusive */,
                transitionTime, fromRule, toRule);
        }
        transitionTime = result.time;
        fromRule = result.from;
        toRule = result.to;
        return TRUE;
    }
    return FALSE;
}

UDate
RuleBasedTimeZone::getTransitionTime(Transition* transition, UBool local,
                                     int32_t NonExistingTimeOpt, int32_t DuplicatedTimeOpt) const {
    UDate time = transition->time;
    if (local) {
        time += getLocalDelta(transition->from->getRawOffset(), transition->from->getDSTSavings(),
                              transition->to->getRawOffset(), transition->to->getDSTSavings(),
                              NonExistingTimeOpt, DuplicatedTimeOpt);
    }
    return time;
}

int32_t
RuleBasedTimeZone::getLocalDelta(int32_t rawBefore, int32_t dstBefore, int32_t rawAfter, int32_t dstAfter,
                             int32_t NonExistingTimeOpt, int32_t DuplicatedTimeOpt) const {
    int32_t delta = 0;

    int32_t offsetBefore = rawBefore + dstBefore;
    int32_t offsetAfter = rawAfter + dstAfter;

    UBool dstToStd = (dstBefore != 0) && (dstAfter == 0);
    UBool stdToDst = (dstBefore == 0) && (dstAfter != 0);

    if (offsetAfter - offsetBefore >= 0) {
        // Positive transition, which makes a non-existing local time range
        if (((NonExistingTimeOpt & kStdDstMask) == kStandard && dstToStd)
                || ((NonExistingTimeOpt & kStdDstMask) == kDaylight && stdToDst)) {
            delta = offsetBefore;
        } else if (((NonExistingTimeOpt & kStdDstMask) == kStandard && stdToDst)
                || ((NonExistingTimeOpt & kStdDstMask) == kDaylight && dstToStd)) {
            delta = offsetAfter;
        } else if ((NonExistingTimeOpt & kFormerLatterMask) == kLatter) {
            delta = offsetBefore;
        } else {
            // Interprets the time with rule before the transition,
            // default for non-existing time range
            delta = offsetAfter;
        }
    } else {
        // Negative transition, which makes a duplicated local time range
        if (((DuplicatedTimeOpt & kStdDstMask) == kStandard && dstToStd)
                || ((DuplicatedTimeOpt & kStdDstMask) == kDaylight && stdToDst)) {
            delta = offsetAfter;
        } else if (((DuplicatedTimeOpt & kStdDstMask) == kStandard && stdToDst)
                || ((DuplicatedTimeOpt & kStdDstMask) == kDaylight && dstToStd)) {
            delta = offsetBefore;
        } else if ((DuplicatedTimeOpt & kFormerLatterMask) == kFormer) {
            delta = offsetBefore;
        } else {
            // Interprets the time with rule after the transition,
            // default for duplicated local time range
            delta = offsetAfter;
        }
    }
    return delta;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2014-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
*
* File REGION.CPP
*
* Modification History:*
*   Date        Name        Description
* 01/15/13      Emmons      Original Port from ICU4J
********************************************************************************
*/

/**
 * \file
 * \brief C++ API: Region classes (territory containment)
 */












// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2013, International Business Machines Corporation and         *
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* File REGION_IMPL.H
*
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING
    



U_NAMESPACE_BEGIN


class RegionNameEnumeration : public StringEnumeration {
public:
    RegionNameEnumeration(UVector *fNameList, UErrorCode& status);
    virtual ~RegionNameEnumeration();
    static UClassID U_EXPORT2 getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
    virtual const UnicodeString* snext(UErrorCode& status);
    virtual void reset(UErrorCode& status);
    virtual int32_t count(UErrorCode& status) const;
private:
    int32_t pos;
    UVector *fRegionNames;
};

U_NAMESPACE_END

#endif




#if !UCONFIG_NO_FORMATTING


U_CDECL_BEGIN

static void U_CALLCONV
deleteRegion(void *obj) {
    delete (icu::Region *)obj;
}

/**
 * Cleanup callback func
 */
static UBool U_CALLCONV region_cleanup(void)
{
    icu::Region::cleanupRegionData();

    return TRUE;
}

U_CDECL_END

U_NAMESPACE_BEGIN

static UInitOnce gRegionDataInitOnce = U_INITONCE_INITIALIZER;
static UVector* availableRegions[URGN_LIMIT];

static UHashtable *regionAliases = NULL;
static UHashtable *regionIDMap = NULL;
static UHashtable *numericCodeMap = NULL;
static UVector *allRegions = NULL;

static const UChar UNKNOWN_REGION_ID [] = { 0x5A, 0x5A, 0 };  /* "ZZ" */
static const UChar OUTLYING_OCEANIA_REGION_ID [] = { 0x51, 0x4F, 0 };  /* "QO" */
static const UChar WORLD_ID [] = { 0x30, 0x30, 0x31, 0 };  /* "001" */
static const UChar RANGE_MARKER = 0x7E; /* '~' */

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(RegionNameEnumeration)

/*
 * Initializes the region data from the ICU resource bundles.  The region data
 * contains the basic relationships such as which regions are known, what the numeric
 * codes are, any known aliases, and the territory containment data.
 *
 * If the region data has already loaded, then this method simply returns without doing
 * anything meaningful.
 */
void U_CALLCONV Region::loadRegionData(UErrorCode &status) {

    // Construct service objs first
    LocalUHashtablePointer newRegionIDMap(uhash_open(uhash_hashUnicodeString, uhash_compareUnicodeString, NULL, &status));
    LocalUHashtablePointer newNumericCodeMap(uhash_open(uhash_hashLong,uhash_compareLong,NULL,&status));
    LocalUHashtablePointer newRegionAliases(uhash_open(uhash_hashUnicodeString,uhash_compareUnicodeString,NULL,&status));

    LocalPointer<UVector> continents(new UVector(uprv_deleteUObject, uhash_compareUnicodeString, status), status);
    LocalPointer<UVector> groupings(new UVector(uprv_deleteUObject, uhash_compareUnicodeString, status), status);
    allRegions = new UVector(uprv_deleteUObject, uhash_compareUnicodeString, status);

    LocalUResourceBundlePointer metadata(ures_openDirect(NULL,"metadata",&status));
    LocalUResourceBundlePointer metadataAlias(ures_getByKey(metadata.getAlias(),"alias",NULL,&status));
    LocalUResourceBundlePointer territoryAlias(ures_getByKey(metadataAlias.getAlias(),"territory",NULL,&status));

    LocalUResourceBundlePointer supplementalData(ures_openDirect(NULL,"supplementalData",&status));
    LocalUResourceBundlePointer codeMappings(ures_getByKey(supplementalData.getAlias(),"codeMappings",NULL,&status));

    LocalUResourceBundlePointer idValidity(ures_getByKey(supplementalData.getAlias(),"idValidity",NULL,&status));
    LocalUResourceBundlePointer regionList(ures_getByKey(idValidity.getAlias(),"region",NULL,&status));
    LocalUResourceBundlePointer regionRegular(ures_getByKey(regionList.getAlias(),"regular",NULL,&status));
    LocalUResourceBundlePointer regionMacro(ures_getByKey(regionList.getAlias(),"macroregion",NULL,&status));
    LocalUResourceBundlePointer regionUnknown(ures_getByKey(regionList.getAlias(),"unknown",NULL,&status));

    LocalUResourceBundlePointer territoryContainment(ures_getByKey(supplementalData.getAlias(),"territoryContainment",NULL,&status));
    LocalUResourceBundlePointer worldContainment(ures_getByKey(territoryContainment.getAlias(),"001",NULL,&status));
    LocalUResourceBundlePointer groupingContainment(ures_getByKey(territoryContainment.getAlias(),"grouping",NULL,&status));

    if (U_FAILURE(status)) {
        return;
    }

    // now, initialize
    uhash_setValueDeleter(newRegionIDMap.getAlias(), deleteRegion);  // regionIDMap owns objs
    uhash_setKeyDeleter(newRegionAliases.getAlias(), uprv_deleteUObject); // regionAliases owns the string keys


    while ( ures_hasNext(regionRegular.getAlias()) ) {
        UnicodeString regionName = ures_getNextUnicodeString(regionRegular.getAlias(),NULL,&status);
        int32_t rangeMarkerLocation = regionName.indexOf(RANGE_MARKER);
        UChar buf[6];
        regionName.extract(buf,6,status);
        if ( rangeMarkerLocation > 0 ) {
            UChar endRange = regionName.charAt(rangeMarkerLocation+1);
            buf[rangeMarkerLocation] = 0;
            while ( buf[rangeMarkerLocation-1] <= endRange ) {
                LocalPointer<UnicodeString> newRegion(new UnicodeString(buf), status);
                allRegions->addElement(newRegion.orphan(),status);
                buf[rangeMarkerLocation-1]++;
            }
        } else {
            LocalPointer<UnicodeString> newRegion(new UnicodeString(regionName), status);
            allRegions->addElement(newRegion.orphan(),status);
        }
    }

    while ( ures_hasNext(regionMacro.getAlias()) ) {
        UnicodeString regionName = ures_getNextUnicodeString(regionMacro.getAlias(),NULL,&status);
        int32_t rangeMarkerLocation = regionName.indexOf(RANGE_MARKER);
        UChar buf[6];
        regionName.extract(buf,6,status);
        if ( rangeMarkerLocation > 0 ) {
            UChar endRange = regionName.charAt(rangeMarkerLocation+1);
            buf[rangeMarkerLocation] = 0;
            while ( buf[rangeMarkerLocation-1] <= endRange ) {
                LocalPointer<UnicodeString> newRegion(new UnicodeString(buf), status);
                allRegions->addElement(newRegion.orphan(),status);
                buf[rangeMarkerLocation-1]++;
            }
        } else {
            LocalPointer<UnicodeString> newRegion(new UnicodeString(regionName), status);
            allRegions->addElement(newRegion.orphan(),status);
        }
    }

    while ( ures_hasNext(regionUnknown.getAlias()) ) {
        LocalPointer<UnicodeString> regionName (new UnicodeString(ures_getNextUnicodeString(regionUnknown.getAlias(),NULL,&status),status));
        allRegions->addElement(regionName.orphan(),status);
    }

    while ( ures_hasNext(worldContainment.getAlias()) ) {
        UnicodeString *continentName = new UnicodeString(ures_getNextUnicodeString(worldContainment.getAlias(),NULL,&status));
        continents->addElement(continentName,status);
    }

    UResourceBundle *groupingBundle = nullptr;
    while ( ures_hasNext(groupingContainment.getAlias()) ) {
        groupingBundle = ures_getNextResource(groupingContainment.getAlias(), groupingBundle, &status);
        if (U_FAILURE(status)) {
            break;
        }
        UnicodeString *groupingName = new UnicodeString(ures_getKey(groupingBundle), -1, US_INV);
        if (groupingName) {
            groupings->addElement(groupingName,status);
        }
    }
    ures_close(groupingBundle);

    for ( int32_t i = 0 ; i < allRegions->size() ; i++ ) {
        LocalPointer<Region> r(new Region(), status);
        if ( U_FAILURE(status) ) {
           return;
        }
        UnicodeString *regionName = (UnicodeString *)allRegions->elementAt(i);
        r->idStr = *regionName;

        r->idStr.extract(0,r->idStr.length(),r->id,sizeof(r->id),US_INV);
        r->fType = URGN_TERRITORY; // Only temporary - figure out the real type later once the aliases are known.

        int32_t pos = 0;
        int32_t result = ICU_Utility::parseAsciiInteger(r->idStr, pos);
        if (pos > 0) {
            r->code = result; // Convert string to number
            uhash_iput(newNumericCodeMap.getAlias(),r->code,(void *)(r.getAlias()),&status);
            r->fType = URGN_SUBCONTINENT;
        } else {
            r->code = -1;
        }
        void* idStrAlias = (void*)&(r->idStr); // about to orphan 'r'. Save this off.
        uhash_put(newRegionIDMap.getAlias(),idStrAlias,(void *)(r.orphan()),&status); // regionIDMap takes ownership
    }

    // Process the territory aliases
    while ( ures_hasNext(territoryAlias.getAlias()) ) {
        LocalUResourceBundlePointer res(ures_getNextResource(territoryAlias.getAlias(),NULL,&status));
        const char *aliasFrom = ures_getKey(res.getAlias());
        LocalPointer<UnicodeString> aliasFromStr(new UnicodeString(aliasFrom, -1, US_INV), status);
        UnicodeString aliasTo = ures_getUnicodeStringByKey(res.getAlias(),"replacement",&status);
        res.adoptInstead(NULL);

        const Region *aliasToRegion = (Region *) uhash_get(newRegionIDMap.getAlias(),&aliasTo);
        Region *aliasFromRegion = (Region *)uhash_get(newRegionIDMap.getAlias(),aliasFromStr.getAlias());

        if ( aliasToRegion != NULL && aliasFromRegion == NULL ) { // This is just an alias from some string to a region
            uhash_put(newRegionAliases.getAlias(),(void *)aliasFromStr.orphan(), (void *)aliasToRegion,&status);
        } else {
            if ( aliasFromRegion == NULL ) { // Deprecated region code not in the master codes list - so need to create a deprecated region for it.
                LocalPointer<Region> newRgn(new Region, status); 
                if ( U_SUCCESS(status) ) {
                    aliasFromRegion = newRgn.orphan();
                } else {
                    return; // error out
                }
                aliasFromRegion->idStr.setTo(*aliasFromStr);
                aliasFromRegion->idStr.extract(0,aliasFromRegion->idStr.length(),aliasFromRegion->id,sizeof(aliasFromRegion->id),US_INV);
                uhash_put(newRegionIDMap.getAlias(),(void *)&(aliasFromRegion->idStr),(void *)aliasFromRegion,&status);
                int32_t pos = 0;
                int32_t result = ICU_Utility::parseAsciiInteger(aliasFromRegion->idStr, pos);
                if ( pos > 0 ) {
                    aliasFromRegion->code = result; // Convert string to number
                    uhash_iput(newNumericCodeMap.getAlias(),aliasFromRegion->code,(void *)aliasFromRegion,&status);
                } else {
                    aliasFromRegion->code = -1;
                }
                aliasFromRegion->fType = URGN_DEPRECATED;
            } else {
                aliasFromRegion->fType = URGN_DEPRECATED;
            }

            {
                LocalPointer<UVector> newPreferredValues(new UVector(uprv_deleteUObject, uhash_compareUnicodeString, status), status);
                aliasFromRegion->preferredValues = newPreferredValues.orphan();
            }
            if( U_FAILURE(status)) {
                return;
            }
            UnicodeString currentRegion;
            //currentRegion.remove();   TODO: was already 0 length?
            for (int32_t i = 0 ; i < aliasTo.length() ; i++ ) {
                if ( aliasTo.charAt(i) != 0x0020 ) {
                    currentRegion.append(aliasTo.charAt(i));
                }
                if ( aliasTo.charAt(i) == 0x0020 || i+1 == aliasTo.length() ) {
                    Region *target = (Region *)uhash_get(newRegionIDMap.getAlias(),(void *)&currentRegion);
                    if (target) {
                        LocalPointer<UnicodeString> preferredValue(new UnicodeString(target->idStr), status);
                        aliasFromRegion->preferredValues->addElement((void *)preferredValue.orphan(),status);  // may add null if err
                    }
                    currentRegion.remove();
                }
            }
        }
    }

    // Process the code mappings - This will allow us to assign numeric codes to most of the territories.
    while ( ures_hasNext(codeMappings.getAlias()) ) {
        UResourceBundle *mapping = ures_getNextResource(codeMappings.getAlias(),NULL,&status);
        if ( ures_getType(mapping) == URES_ARRAY && ures_getSize(mapping) == 3) {
            UnicodeString codeMappingID = ures_getUnicodeStringByIndex(mapping,0,&status);
            UnicodeString codeMappingNumber = ures_getUnicodeStringByIndex(mapping,1,&status);
            UnicodeString codeMapping3Letter = ures_getUnicodeStringByIndex(mapping,2,&status);

            Region *r = (Region *)uhash_get(newRegionIDMap.getAlias(),(void *)&codeMappingID);
            if ( r ) {
                int32_t pos = 0;
                int32_t result = ICU_Utility::parseAsciiInteger(codeMappingNumber, pos);
                if ( pos > 0 ) {
                    r->code = result; // Convert string to number
                    uhash_iput(newNumericCodeMap.getAlias(),r->code,(void *)r,&status);
                }
                LocalPointer<UnicodeString> code3(new UnicodeString(codeMapping3Letter), status);
                uhash_put(newRegionAliases.getAlias(),(void *)code3.orphan(), (void *)r,&status);
            }
        }
        ures_close(mapping);
    }

    // Now fill in the special cases for WORLD, UNKNOWN, CONTINENTS, and GROUPINGS
    Region *r;
    UnicodeString WORLD_ID_STRING(WORLD_ID);
    r = (Region *) uhash_get(newRegionIDMap.getAlias(),(void *)&WORLD_ID_STRING);
    if ( r ) {
        r->fType = URGN_WORLD;
    }

    UnicodeString UNKNOWN_REGION_ID_STRING(UNKNOWN_REGION_ID);
    r = (Region *) uhash_get(newRegionIDMap.getAlias(),(void *)&UNKNOWN_REGION_ID_STRING);
    if ( r ) {
        r->fType = URGN_UNKNOWN;
    }

    for ( int32_t i = 0 ; i < continents->size() ; i++ ) {
        r = (Region *) uhash_get(newRegionIDMap.getAlias(),(void *)continents->elementAt(i));
        if ( r ) {
            r->fType = URGN_CONTINENT;
        }
    }

    for ( int32_t i = 0 ; i < groupings->size() ; i++ ) {
        r = (Region *) uhash_get(newRegionIDMap.getAlias(),(void *)groupings->elementAt(i));
        if ( r ) {
            r->fType = URGN_GROUPING;
        }
    }

    // Special case: The region code "QO" (Outlying Oceania) is a subcontinent code added by CLDR
    // even though it looks like a territory code.  Need to handle it here.

    UnicodeString OUTLYING_OCEANIA_REGION_ID_STRING(OUTLYING_OCEANIA_REGION_ID);
    r = (Region *) uhash_get(newRegionIDMap.getAlias(),(void *)&OUTLYING_OCEANIA_REGION_ID_STRING);
    if ( r ) {
        r->fType = URGN_SUBCONTINENT;
    }

    // Load territory containment info from the supplemental data.
    while ( ures_hasNext(territoryContainment.getAlias()) ) {
        LocalUResourceBundlePointer mapping(ures_getNextResource(territoryContainment.getAlias(),NULL,&status));
        if( U_FAILURE(status) ) {
            return;  // error out
        }
        const char *parent = ures_getKey(mapping.getAlias());
        if (uprv_strcmp(parent, "containedGroupings") == 0 || uprv_strcmp(parent, "deprecated") == 0) {
            continue; // handle new pseudo-parent types added in ICU data per cldrbug 7808; for now just skip.
            // #11232 is to do something useful with these.
        }
        UnicodeString parentStr = UnicodeString(parent, -1 , US_INV);
        Region *parentRegion = (Region *) uhash_get(newRegionIDMap.getAlias(),(void *)&parentStr);

        for ( int j = 0 ; j < ures_getSize(mapping.getAlias()); j++ ) {
            UnicodeString child = ures_getUnicodeStringByIndex(mapping.getAlias(),j,&status);
            Region *childRegion = (Region *) uhash_get(newRegionIDMap.getAlias(),(void *)&child);
            if ( parentRegion != NULL && childRegion != NULL ) {

                // Add the child region to the set of regions contained by the parent
                if (parentRegion->containedRegions == NULL) {
                    parentRegion->containedRegions = new UVector(uprv_deleteUObject, uhash_compareUnicodeString, status);
                }

                LocalPointer<UnicodeString> childStr(new UnicodeString(), status);
                if( U_FAILURE(status) ) {
                    return;  // error out
                }
                childStr->fastCopyFrom(childRegion->idStr);
                parentRegion->containedRegions->addElement((void *)childStr.orphan(),status);

                // Set the parent region to be the containing region of the child.
                // Regions of type GROUPING can't be set as the parent, since another region
                // such as a SUBCONTINENT, CONTINENT, or WORLD must always be the parent.
                if ( parentRegion->fType != URGN_GROUPING) {
                    childRegion->containingRegion = parentRegion;
                }
            }
        }
    }

    // Create the availableRegions lists
    int32_t pos = UHASH_FIRST;
    while ( const UHashElement* element = uhash_nextElement(newRegionIDMap.getAlias(),&pos)) {
        Region *ar = (Region *)element->value.pointer;
        if ( availableRegions[ar->fType] == NULL ) {
            LocalPointer<UVector> newAr(new UVector(uprv_deleteUObject, uhash_compareUnicodeString, status), status);
            availableRegions[ar->fType] = newAr.orphan();
        }
        LocalPointer<UnicodeString> arString(new UnicodeString(ar->idStr), status);
        if( U_FAILURE(status) ) {
            return;  // error out
        }
        availableRegions[ar->fType]->addElement((void *)arString.orphan(),status);
    }
    
    ucln_i18n_registerCleanup(UCLN_I18N_REGION, region_cleanup);
    // copy hashtables
    numericCodeMap = newNumericCodeMap.orphan();
    regionIDMap = newRegionIDMap.orphan();
    regionAliases = newRegionAliases.orphan();
}

void Region::cleanupRegionData() {
    for (int32_t i = 0 ; i < URGN_LIMIT ; i++ ) {
        if ( availableRegions[i] ) {
            delete availableRegions[i];
        }
    }

    if (regionAliases) {
        uhash_close(regionAliases);
    }

    if (numericCodeMap) {
        uhash_close(numericCodeMap);
    }

    if (regionIDMap) {
        uhash_close(regionIDMap);
    }
    if (allRegions) {
        allRegions->removeAllElements(); // Don't need the temporary list anymore.
        delete allRegions;
        allRegions = NULL;
    }

    regionAliases = numericCodeMap = regionIDMap = NULL;

    gRegionDataInitOnce.reset();
}

Region::Region ()
        : code(-1),
          fType(URGN_UNKNOWN),
          containingRegion(NULL),
          containedRegions(NULL),
          preferredValues(NULL) {
    id[0] = 0;
}

Region::~Region () {
        if (containedRegions) {
            delete containedRegions;
        }
        if (preferredValues) {
            delete preferredValues;
        }
}

/**
 * Returns true if the two regions are equal.
 * Per PMC, just use pointer compare, since we have at most one instance of each Region.
 */
UBool
Region::operator==(const Region &that) const {
    return (idStr == that.idStr);
}

/**
 * Returns true if the two regions are NOT equal; that is, if operator ==() returns false.
 * Per PMC, just use pointer compare, since we have at most one instance of each Region.
 */
UBool
Region::operator!=(const Region &that) const {
        return (idStr != that.idStr);
}

/**
 * Returns a pointer to a Region using the given region code.  The region code can be either 2-letter ISO code,
 * 3-letter ISO code,  UNM.49 numeric code, or other valid Unicode Region Code as defined by the LDML specification.
 * The identifier will be canonicalized internally using the supplemental metadata as defined in the CLDR.
 * If the region code is NULL or not recognized, the appropriate error code will be set ( U_ILLEGAL_ARGUMENT_ERROR )
 */
const Region* U_EXPORT2
Region::getInstance(const char *region_code, UErrorCode &status) {

    umtx_initOnce(gRegionDataInitOnce, &loadRegionData, status);
    if (U_FAILURE(status)) {
        return NULL;
    }

    if ( !region_code ) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    UnicodeString regionCodeString = UnicodeString(region_code, -1, US_INV);
    Region *r = (Region *)uhash_get(regionIDMap,(void *)&regionCodeString);

    if ( !r ) {
        r = (Region *)uhash_get(regionAliases,(void *)&regionCodeString);
    }

    if ( !r ) { // Unknown region code
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if ( r->fType == URGN_DEPRECATED && r->preferredValues->size() == 1) {
        StringEnumeration *pv = r->getPreferredValues(status);
        pv->reset(status);
        const UnicodeString *ustr = pv->snext(status);
        r = (Region *)uhash_get(regionIDMap,(void *)ustr);
        delete pv;
    }

    return r;

}

/**
 * Returns a pointer to a Region using the given numeric region code. If the numeric region code is not recognized,
 * the appropriate error code will be set ( U_ILLEGAL_ARGUMENT_ERROR ).
 */
const Region* U_EXPORT2
Region::getInstance (int32_t code, UErrorCode &status) {

    umtx_initOnce(gRegionDataInitOnce, &loadRegionData, status);
    if (U_FAILURE(status)) {
        return NULL;
    }

    Region *r = (Region *)uhash_iget(numericCodeMap,code);

    if ( !r ) { // Just in case there's an alias that's numeric, try to find it.
        UnicodeString id;
        ICU_Utility::appendNumber(id, code, 10, 1);
        r = (Region *)uhash_get(regionAliases,&id);
    }

    if( U_FAILURE(status) ) {
        return NULL;
    }

    if ( !r ) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    if ( r->fType == URGN_DEPRECATED && r->preferredValues->size() == 1) {
        StringEnumeration *pv = r->getPreferredValues(status);
        pv->reset(status);
        const UnicodeString *ustr = pv->snext(status);
        r = (Region *)uhash_get(regionIDMap,(void *)ustr);
        delete pv;
    }

    return r;
}


/**
 * Returns an enumeration over the IDs of all known regions that match the given type.
 */
StringEnumeration* U_EXPORT2
Region::getAvailable(URegionType type, UErrorCode &status) {
    umtx_initOnce(gRegionDataInitOnce, &loadRegionData, status); // returns immediately if U_FAILURE(status)
    if (U_FAILURE(status)) {
        return NULL;
    }
    return new RegionNameEnumeration(availableRegions[type],status);
}

/**
 * Returns a pointer to the region that contains this region.  Returns NULL if this region is code "001" (World)
 * or "ZZ" (Unknown region). For example, calling this method with region "IT" (Italy) returns the
 * region "039" (Southern Europe).
 */
const Region*
Region::getContainingRegion() const {
    UErrorCode status = U_ZERO_ERROR;
    umtx_initOnce(gRegionDataInitOnce, &loadRegionData, status);
    return containingRegion;
}

/**
 * Return a pointer to the region that geographically contains this region and matches the given type,
 * moving multiple steps up the containment chain if necessary.  Returns NULL if no containing region can be found
 * that matches the given type. Note: The URegionTypes = "URGN_GROUPING", "URGN_DEPRECATED", or "URGN_UNKNOWN"
 * are not appropriate for use in this API. NULL will be returned in this case. For example, calling this method
 * with region "IT" (Italy) for type "URGN_CONTINENT" returns the region "150" ( Europe ).
 */
const Region*
Region::getContainingRegion(URegionType type) const {
    UErrorCode status = U_ZERO_ERROR;
    umtx_initOnce(gRegionDataInitOnce, &loadRegionData, status);
    if ( containingRegion == NULL ) {
        return NULL;
    }

    return ( containingRegion->fType == type)? containingRegion: containingRegion->getContainingRegion(type);
}

/**
 * Return an enumeration over the IDs of all the regions that are immediate children of this region in the
 * region hierarchy. These returned regions could be either macro regions, territories, or a mixture of the two,
 * depending on the containment data as defined in CLDR.  This API may return NULL if this region doesn't have
 * any sub-regions. For example, calling this method with region "150" (Europe) returns an enumeration containing
 * the various sub regions of Europe - "039" (Southern Europe) - "151" (Eastern Europe) - "154" (Northern Europe)
 * and "155" (Western Europe).
 */
StringEnumeration*
Region::getContainedRegions(UErrorCode &status) const {
    umtx_initOnce(gRegionDataInitOnce, &loadRegionData, status); // returns immediately if U_FAILURE(status)
    if (U_FAILURE(status)) {
        return NULL;
    }
    return new RegionNameEnumeration(containedRegions,status);
}

/**
 * Returns an enumeration over the IDs of all the regions that are children of this region anywhere in the region
 * hierarchy and match the given type.  This API may return an empty enumeration if this region doesn't have any
 * sub-regions that match the given type. For example, calling this method with region "150" (Europe) and type
 * "URGN_TERRITORY" returns a set containing all the territories in Europe ( "FR" (France) - "IT" (Italy) - "DE" (Germany) etc. )
 */
StringEnumeration*
Region::getContainedRegions( URegionType type, UErrorCode &status ) const {
    umtx_initOnce(gRegionDataInitOnce, &loadRegionData, status); // returns immediately if U_FAILURE(status)
    if (U_FAILURE(status)) {
        return NULL;
    }

    UVector *result = new UVector(NULL, uhash_compareChars, status);

    StringEnumeration *cr = getContainedRegions(status);

    for ( int32_t i = 0 ; i < cr->count(status) ; i++ ) {
        const char *regionId = cr->next(NULL,status);
        const Region *r = Region::getInstance(regionId,status);
        if ( r->getType() == type) {
            result->addElement((void *)&r->idStr,status);
        } else {
            StringEnumeration *children = r->getContainedRegions(type, status);
            for ( int32_t j = 0 ; j < children->count(status) ; j++ ) {
                const char *id2 = children->next(NULL,status);
                const Region *r2 = Region::getInstance(id2,status);
                result->addElement((void *)&r2->idStr,status);
            }
            delete children;
        }
    }
    delete cr;
    StringEnumeration* resultEnumeration = new RegionNameEnumeration(result,status);
    delete result;
    return resultEnumeration;
}

/**
 * Returns true if this region contains the supplied other region anywhere in the region hierarchy.
 */
UBool
Region::contains(const Region &other) const {
    UErrorCode status = U_ZERO_ERROR;
    umtx_initOnce(gRegionDataInitOnce, &loadRegionData, status);

    if (!containedRegions) {
          return FALSE;
    }
    if (containedRegions->contains((void *)&other.idStr)) {
        return TRUE;
    } else {
        for ( int32_t i = 0 ; i < containedRegions->size() ; i++ ) {
            UnicodeString *crStr = (UnicodeString *)containedRegions->elementAt(i);
            Region *cr = (Region *) uhash_get(regionIDMap,(void *)crStr);
            if ( cr && cr->contains(other) ) {
                return TRUE;
            }
        }
    }

    return FALSE;
}

/**
 * For deprecated regions, return an enumeration over the IDs of the regions that are the preferred replacement
 * regions for this region.  Returns NULL for a non-deprecated region.  For example, calling this method with region
 * "SU" (Soviet Union) would return a list of the regions containing "RU" (Russia), "AM" (Armenia), "AZ" (Azerbaijan), etc...
 */
StringEnumeration*
Region::getPreferredValues(UErrorCode &status) const {
    umtx_initOnce(gRegionDataInitOnce, &loadRegionData, status); // returns immediately if U_FAILURE(status)
    if (U_FAILURE(status) || fType != URGN_DEPRECATED) {
        return NULL;
    }
    return new RegionNameEnumeration(preferredValues,status);
}


/**
 * Return this region's canonical region code.
 */
const char*
Region::getRegionCode() const {
    return id;
}

int32_t
Region::getNumericCode() const {
    return code;
}

/**
 * Returns the region type of this region.
 */
URegionType
Region::getType() const {
    return fType;
}

RegionNameEnumeration::RegionNameEnumeration(UVector *fNameList, UErrorCode& status) {
    pos=0;
    if (fNameList && U_SUCCESS(status)) {
        fRegionNames = new UVector(uprv_deleteUObject, uhash_compareUnicodeString, fNameList->size(),status);
        for ( int32_t i = 0 ; i < fNameList->size() ; i++ ) {
            UnicodeString* this_region_name = (UnicodeString *)fNameList->elementAt(i);
            UnicodeString* new_region_name = new UnicodeString(*this_region_name);
            fRegionNames->addElement((void *)new_region_name,status);
        }
    }
    else {
        fRegionNames = NULL;
    }
}

const UnicodeString*
RegionNameEnumeration::snext(UErrorCode& status) {
  if (U_FAILURE(status) || (fRegionNames==NULL)) {
    return NULL;
  }
  const UnicodeString* nextStr = (const UnicodeString *)fRegionNames->elementAt(pos);
  if (nextStr!=NULL) {
    pos++;
  }
  return nextStr;
}

void
RegionNameEnumeration::reset(UErrorCode& /*status*/) {
    pos=0;
}

int32_t
RegionNameEnumeration::count(UErrorCode& /*status*/) const {
    return (fRegionNames==NULL) ? 0 : fRegionNames->size();
}

RegionNameEnumeration::~RegionNameEnumeration() {
    delete fRegionNames;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2014-2016, International Business Machines Corporation and
* others. All Rights Reserved.
******************************************************************************
*
* File reldatefmt.cpp
******************************************************************************
*/

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************
* Copyright (C) 2014-2016, International Business Machines Corporation and
* others.
* All Rights Reserved.
*****************************************************************************
*
* File RELDATEFMT.H
*****************************************************************************
*/




#if U_SHOW_CPLUSPLUS_API



// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// *****************************************************************************************
// * Copyright (C) 2016, International Business Machines
// * Corporation and others. All Rights Reserved.
// *****************************************************************************************
// */

// #ifndef URELDATEFMT_H
// #define URELDATEFMT_H

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_FORMATTING && !UCONFIG_NO_BREAK_ITERATION

// 
// 
// 
// #include "unicode/uformattedvalue.h"

// /**
//  * \file
//  * \brief C API: URelativeDateTimeFormatter, relative date formatting of unit + numeric offset.
//  *
//  * Provides simple formatting of relative dates, in two ways
//  * <ul>
//  *   <li>relative dates with a quantity e.g "in 5 days"</li>
//  *   <li>relative dates without a quantity e.g "next Tuesday"</li>
//  * </ul>
//  * <p>
//  * This does not provide compound formatting for multiple units,
//  * other than the ability to combine a time string with a relative date,
//  * as in "next Tuesday at 3:45 PM". It also does not provide support
//  * for determining which unit to use, such as deciding between "in 7 days"
//  * and "in 1 week".
//  *
//  * @stable ICU 57
//  */

// /**
//  * The formatting style
//  * @stable ICU 54
//  */
// typedef enum UDateRelativeDateTimeFormatterStyle {
//   /**
//    * Everything spelled out.
//    * @stable ICU 54
//    */
//   UDAT_STYLE_LONG,

//   /**
//    * Abbreviations used when possible.
//    * @stable ICU 54
//    */
//   UDAT_STYLE_SHORT,

//   /**
//    * Use the shortest possible form.
//    * @stable ICU 54
//    */
//   UDAT_STYLE_NARROW,

// #ifndef U_HIDE_DEPRECATED_API
//     /**
//      * One more than the highest normal UDateRelativeDateTimeFormatterStyle value.
//      * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
//      */
//     UDAT_STYLE_COUNT
// #endif  /* U_HIDE_DEPRECATED_API */
// } UDateRelativeDateTimeFormatterStyle;

// /**
//  * Represents the unit for formatting a relative date. e.g "in 5 days"
//  * or "next year"
//  * @stable ICU 57
//  */
// typedef enum URelativeDateTimeUnit {
//     /**
//      * Specifies that relative unit is year, e.g. "last year",
//      * "in 5 years".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_YEAR,
//     /**
//      * Specifies that relative unit is quarter, e.g. "last quarter",
//      * "in 5 quarters".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_QUARTER,
//     /**
//      * Specifies that relative unit is month, e.g. "last month",
//      * "in 5 months".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_MONTH,
//     /**
//      * Specifies that relative unit is week, e.g. "last week",
//      * "in 5 weeks".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_WEEK,
//     /**
//      * Specifies that relative unit is day, e.g. "yesterday",
//      * "in 5 days".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_DAY,
//     /**
//      * Specifies that relative unit is hour, e.g. "1 hour ago",
//      * "in 5 hours".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_HOUR,
//     /**
//      * Specifies that relative unit is minute, e.g. "1 minute ago",
//      * "in 5 minutes".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_MINUTE,
//     /**
//      * Specifies that relative unit is second, e.g. "1 second ago",
//      * "in 5 seconds".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_SECOND,
//     /**
//      * Specifies that relative unit is Sunday, e.g. "last Sunday",
//      * "this Sunday", "next Sunday", "in 5 Sundays".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_SUNDAY,
//     /**
//      * Specifies that relative unit is Monday, e.g. "last Monday",
//      * "this Monday", "next Monday", "in 5 Mondays".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_MONDAY,
//     /**
//      * Specifies that relative unit is Tuesday, e.g. "last Tuesday",
//      * "this Tuesday", "next Tuesday", "in 5 Tuesdays".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_TUESDAY,
//     /**
//      * Specifies that relative unit is Wednesday, e.g. "last Wednesday",
//      * "this Wednesday", "next Wednesday", "in 5 Wednesdays".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_WEDNESDAY,
//     /**
//      * Specifies that relative unit is Thursday, e.g. "last Thursday",
//      * "this Thursday", "next Thursday", "in 5 Thursdays".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_THURSDAY,
//     /**
//      * Specifies that relative unit is Friday, e.g. "last Friday",
//      * "this Friday", "next Friday", "in 5 Fridays".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_FRIDAY,
//     /**
//      * Specifies that relative unit is Saturday, e.g. "last Saturday",
//      * "this Saturday", "next Saturday", "in 5 Saturdays".
//      * @stable ICU 57
//      */
//     UDAT_REL_UNIT_SATURDAY,
// #ifndef U_HIDE_DEPRECATED_API
//     /**
//      * One more than the highest normal URelativeDateTimeUnit value.
//      * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
//      */
//     UDAT_REL_UNIT_COUNT
// #endif  /* U_HIDE_DEPRECATED_API */
// } URelativeDateTimeUnit;

// #ifndef U_HIDE_DRAFT_API
// /**
//  * FieldPosition and UFieldPosition selectors for format fields
//  * defined by RelativeDateTimeFormatter.
//  * @draft ICU 64
//  */
// typedef enum URelativeDateTimeFormatterField {
//     /**
//      * Represents a literal text string, like "tomorrow" or "days ago".
//      * @draft ICU 64
//      */
//     UDAT_REL_LITERAL_FIELD,
//     /**
//      * Represents a number quantity, like "3" in "3 days ago".
//      * @draft ICU 64
//      */
//     UDAT_REL_NUMERIC_FIELD,
// } URelativeDateTimeFormatterField;
// #endif // U_HIDE_DRAFT_API


// /**
//  * Opaque URelativeDateTimeFormatter object for use in C programs.
//  * @stable ICU 57
//  */
// struct URelativeDateTimeFormatter;
// typedef struct URelativeDateTimeFormatter URelativeDateTimeFormatter;  /**< C typedef for struct URelativeDateTimeFormatter. @stable ICU 57 */


// /**
//  * Open a new URelativeDateTimeFormatter object for a given locale using the
//  * specified width and capitalizationContext, along with a number formatter
//  * (if desired) to override the default formatter that would be used for
//  * display of numeric field offsets. The default formatter typically rounds
//  * toward 0 and has a minimum of 0 fraction digits and a maximum of 3
//  * fraction digits (i.e. it will show as many decimal places as necessary
//  * up to 3, without showing trailing 0s).
//  *
//  * @param locale
//  *          The locale
//  * @param nfToAdopt
//  *          A number formatter to set for this URelativeDateTimeFormatter
//  *          object (instead of the default decimal formatter). Ownership of
//  *          this UNumberFormat object will pass to the URelativeDateTimeFormatter
//  *          object (the URelativeDateTimeFormatter adopts the UNumberFormat),
//  *          which becomes responsible for closing it. If the caller wishes to
//  *          retain ownership of the UNumberFormat object, the caller must clone
//  *          it (with unum_clone) and pass the clone to ureldatefmt_open. May be
//  *          NULL to use the default decimal formatter.
//  * @param width
//  *          The width - wide, short, narrow, etc.
//  * @param capitalizationContext
//  *          A value from UDisplayContext that pertains to capitalization, e.g.
//  *          UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE.
//  * @param status
//  *          A pointer to a UErrorCode to receive any errors.
//  * @return
//  *          A pointer to a URelativeDateTimeFormatter object for the specified locale,
//  *          or NULL if an error occurred.
//  * @stable ICU 57
//  */
// U_STABLE URelativeDateTimeFormatter* U_EXPORT2
// ureldatefmt_open( const char*          locale,
//                   UNumberFormat*       nfToAdopt,
//                   UDateRelativeDateTimeFormatterStyle width,
//                   UDisplayContext      capitalizationContext,
//                   UErrorCode*          status );

// /**
//  * Close a URelativeDateTimeFormatter object. Once closed it may no longer be used.
//  * @param reldatefmt
//  *            The URelativeDateTimeFormatter object to close.
//  * @stable ICU 57
//  */
// U_STABLE void U_EXPORT2
// ureldatefmt_close(URelativeDateTimeFormatter *reldatefmt);

// #ifndef U_HIDE_DRAFT_API
// struct UFormattedRelativeDateTime;
// /**
//  * Opaque struct to contain the results of a URelativeDateTimeFormatter operation.
//  * @draft ICU 64
//  */
// typedef struct UFormattedRelativeDateTime UFormattedRelativeDateTime;

// /**
//  * Creates an object to hold the result of a URelativeDateTimeFormatter
//  * operation. The object can be used repeatedly; it is cleared whenever
//  * passed to a format function.
//  *
//  * @param ec Set if an error occurs.
//  * @return A pointer needing ownership.
//  * @draft ICU 64
//  */
// U_DRAFT UFormattedRelativeDateTime* U_EXPORT2
// ureldatefmt_openResult(UErrorCode* ec);

// /**
//  * Returns a representation of a UFormattedRelativeDateTime as a UFormattedValue,
//  * which can be subsequently passed to any API requiring that type.
//  *
//  * The returned object is owned by the UFormattedRelativeDateTime and is valid
//  * only as long as the UFormattedRelativeDateTime is present and unchanged in memory.
//  *
//  * You can think of this method as a cast between types.
//  *
//  * @param ufrdt The object containing the formatted string.
//  * @param ec Set if an error occurs.
//  * @return A UFormattedValue owned by the input object.
//  * @draft ICU 64
//  */
// U_DRAFT const UFormattedValue* U_EXPORT2
// ureldatefmt_resultAsValue(const UFormattedRelativeDateTime* ufrdt, UErrorCode* ec);

// /**
//  * Releases the UFormattedRelativeDateTime created by ureldatefmt_openResult.
//  *
//  * @param ufrdt The object to release.
//  * @draft ICU 64
//  */
// U_DRAFT void U_EXPORT2
// ureldatefmt_closeResult(UFormattedRelativeDateTime* ufrdt);
// #endif  /* U_HIDE_DRAFT_API */


// #if U_SHOW_CPLUSPLUS_API

// U_NAMESPACE_BEGIN

// /**
//  * \class LocalURelativeDateTimeFormatterPointer
//  * "Smart pointer" class, closes a URelativeDateTimeFormatter via ureldatefmt_close().
//  * For most methods see the LocalPointerBase base class.
//  *
//  * @see LocalPointerBase
//  * @see LocalPointer
//  * @stable ICU 57
//  */
// U_DEFINE_LOCAL_OPEN_POINTER(LocalURelativeDateTimeFormatterPointer, URelativeDateTimeFormatter, ureldatefmt_close);

// #ifndef U_HIDE_DRAFT_API
// /**
//  * \class LocalUFormattedRelativeDateTimePointer
//  * "Smart pointer" class, closes a UFormattedRelativeDateTime via ureldatefmt_closeResult().
//  * For most methods see the LocalPointerBase base class.
//  *
//  * @see LocalPointerBase
//  * @see LocalPointer
//  * @draft ICU 64
//  */
// U_DEFINE_LOCAL_OPEN_POINTER(LocalUFormattedRelativeDateTimePointer, UFormattedRelativeDateTime, ureldatefmt_closeResult);
// #endif  /* U_HIDE_DRAFT_API */

// U_NAMESPACE_END

// #endif

// /**
//  * Format a combination of URelativeDateTimeUnit and numeric
//  * offset using a numeric style, e.g. "1 week ago", "in 1 week",
//  * "5 weeks ago", "in 5 weeks".
//  *
//  * @param reldatefmt
//  *          The URelativeDateTimeFormatter object specifying the
//  *          format conventions.
//  * @param offset
//  *          The signed offset for the specified unit. This will
//  *          be formatted according to this object's UNumberFormat
//  *          object.
//  * @param unit
//  *          The unit to use when formatting the relative
//  *          date, e.g. UDAT_REL_UNIT_WEEK, UDAT_REL_UNIT_FRIDAY.
//  * @param result
//  *          A pointer to a buffer to receive the formatted result.
//  * @param resultCapacity
//  *          The maximum size of result.
//  * @param status
//  *          A pointer to a UErrorCode to receive any errors. In
//  *          case of error status, the contents of result are
//  *          undefined.
//  * @return
//  *          The length of the formatted result; may be greater
//  *          than resultCapacity, in which case an error is returned.
//  * @stable ICU 57
//  */
// U_STABLE int32_t U_EXPORT2
// ureldatefmt_formatNumeric( const URelativeDateTimeFormatter* reldatefmt,
//                     double                offset,
//                     URelativeDateTimeUnit unit,
//                     UChar*                result,
//                     int32_t               resultCapacity,
//                     UErrorCode*           status);

// #ifndef U_HIDE_DRAFT_API
// /**
//  * Format a combination of URelativeDateTimeUnit and numeric
//  * offset using a numeric style, e.g. "1 week ago", "in 1 week",
//  * "5 weeks ago", "in 5 weeks".
//  *
//  * @param reldatefmt
//  *          The URelativeDateTimeFormatter object specifying the
//  *          format conventions.
//  * @param offset
//  *          The signed offset for the specified unit. This will
//  *          be formatted according to this object's UNumberFormat
//  *          object.
//  * @param unit
//  *          The unit to use when formatting the relative
//  *          date, e.g. UDAT_REL_UNIT_WEEK, UDAT_REL_UNIT_FRIDAY.
//  * @param result
//  *          A pointer to a UFormattedRelativeDateTime to populate.
//  * @param status
//  *          A pointer to a UErrorCode to receive any errors. In
//  *          case of error status, the contents of result are
//  *          undefined.
//  * @draft ICU 64
//  */
// U_DRAFT void U_EXPORT2
// ureldatefmt_formatNumericToResult(
//     const URelativeDateTimeFormatter* reldatefmt,
//     double                            offset,
//     URelativeDateTimeUnit             unit,
//     UFormattedRelativeDateTime*       result,
//     UErrorCode*                       status);
// #endif  /* U_HIDE_DRAFT_API */

// /**
//  * Format a combination of URelativeDateTimeUnit and numeric offset
//  * using a text style if possible, e.g. "last week", "this week",
//  * "next week", "yesterday", "tomorrow". Falls back to numeric
//  * style if no appropriate text term is available for the specified
//  * offset in the object's locale.
//  *
//  * @param reldatefmt
//  *          The URelativeDateTimeFormatter object specifying the
//  *          format conventions.
//  * @param offset
//  *          The signed offset for the specified unit.
//  * @param unit
//  *          The unit to use when formatting the relative
//  *          date, e.g. UDAT_REL_UNIT_WEEK, UDAT_REL_UNIT_FRIDAY.
//  * @param result
//  *          A pointer to a buffer to receive the formatted result.
//  * @param resultCapacity
//  *          The maximum size of result.
//  * @param status
//  *          A pointer to a UErrorCode to receive any errors. In
//  *          case of error status, the contents of result are
//  *          undefined.
//  * @return
//  *          The length of the formatted result; may be greater
//  *          than resultCapacity, in which case an error is returned.
//  * @stable ICU 57
//  */
// U_STABLE int32_t U_EXPORT2
// ureldatefmt_format( const URelativeDateTimeFormatter* reldatefmt,
//                     double                offset,
//                     URelativeDateTimeUnit unit,
//                     UChar*                result,
//                     int32_t               resultCapacity,
//                     UErrorCode*           status);

// #ifndef U_HIDE_DRAFT_API
// /**
//  * Format a combination of URelativeDateTimeUnit and numeric offset
//  * using a text style if possible, e.g. "last week", "this week",
//  * "next week", "yesterday", "tomorrow". Falls back to numeric
//  * style if no appropriate text term is available for the specified
//  * offset in the object's locale.
//  *
//  * This method populates a UFormattedRelativeDateTime, which exposes more
//  * information than the string populated by format().
//  *
//  * @param reldatefmt
//  *          The URelativeDateTimeFormatter object specifying the
//  *          format conventions.
//  * @param offset
//  *          The signed offset for the specified unit.
//  * @param unit
//  *          The unit to use when formatting the relative
//  *          date, e.g. UDAT_REL_UNIT_WEEK, UDAT_REL_UNIT_FRIDAY.
//  * @param result
//  *          A pointer to a UFormattedRelativeDateTime to populate.
//  * @param status
//  *          A pointer to a UErrorCode to receive any errors. In
//  *          case of error status, the contents of result are
//  *          undefined.
//  * @draft ICU 64
//  */
// U_DRAFT void U_EXPORT2
// ureldatefmt_formatToResult(
//     const URelativeDateTimeFormatter* reldatefmt,
//     double                            offset,
//     URelativeDateTimeUnit             unit,
//     UFormattedRelativeDateTime*       result,
//     UErrorCode*                       status);
// #endif  /* U_HIDE_DRAFT_API */

// /**
//  * Combines a relative date string and a time string in this object's
//  * locale. This is done with the same date-time separator used for the
//  * default calendar in this locale to produce a result such as
//  * "yesterday at 3:45 PM".
//  *
//  * @param reldatefmt
//  *          The URelativeDateTimeFormatter object specifying the format conventions.
//  * @param relativeDateString
//  *          The relative date string.
//  * @param relativeDateStringLen
//  *          The length of relativeDateString; may be -1 if relativeDateString
//  *          is zero-terminated.
//  * @param timeString
//  *          The time string.
//  * @param timeStringLen
//  *          The length of timeString; may be -1 if timeString is zero-terminated.
//  * @param result
//  *          A pointer to a buffer to receive the formatted result.
//  * @param resultCapacity
//  *          The maximum size of result.
//  * @param status
//  *          A pointer to a UErrorCode to receive any errors. In case of error status,
//  *          the contents of result are undefined.
//  * @return
//  *          The length of the formatted result; may be greater than resultCapacity,
//  *          in which case an error is returned.
//  * @stable ICU 57
//  */
// U_STABLE int32_t U_EXPORT2
// ureldatefmt_combineDateAndTime( const URelativeDateTimeFormatter* reldatefmt,
//                     const UChar *     relativeDateString,
//                     int32_t           relativeDateStringLen,
//                     const UChar *     timeString,
//                     int32_t           timeStringLen,
//                     UChar*            result,
//                     int32_t           resultCapacity,
//                     UErrorCode*       status );

// #endif /* !UCONFIG_NO_FORMATTING && !UCONFIG_NO_BREAK_ITERATION */

// #endif




/**
 * \file
 * \brief C++ API: Formats relative dates such as "1 day ago" or "tomorrow"
 */

#if !UCONFIG_NO_FORMATTING

/**
 * Represents the unit for formatting a relative date. e.g "in 5 days"
 * or "in 3 months"
 * @stable ICU 53
 */
typedef enum UDateRelativeUnit {

    /**
     * Seconds
     * @stable ICU 53
     */
    UDAT_RELATIVE_SECONDS,

    /**
     * Minutes
     * @stable ICU 53
     */
    UDAT_RELATIVE_MINUTES,

    /**
     * Hours
     * @stable ICU 53
     */
    UDAT_RELATIVE_HOURS,

    /**
     * Days
     * @stable ICU 53
     */
    UDAT_RELATIVE_DAYS,

    /**
     * Weeks
     * @stable ICU 53
     */
    UDAT_RELATIVE_WEEKS,

    /**
     * Months
     * @stable ICU 53
     */
    UDAT_RELATIVE_MONTHS,

    /**
     * Years
     * @stable ICU 53
     */
    UDAT_RELATIVE_YEARS,

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UDateRelativeUnit value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UDAT_RELATIVE_UNIT_COUNT
#endif  // U_HIDE_DEPRECATED_API
} UDateRelativeUnit;

/**
 * Represents an absolute unit.
 * @stable ICU 53
 */
typedef enum UDateAbsoluteUnit {

    // Days of week have to remain together and in order from Sunday to
    // Saturday.
    /**
     * Sunday
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_SUNDAY,

    /**
     * Monday
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_MONDAY,

    /**
     * Tuesday
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_TUESDAY,

    /**
     * Wednesday
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_WEDNESDAY,

    /**
     * Thursday
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_THURSDAY,

    /**
     * Friday
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_FRIDAY,

    /**
     * Saturday
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_SATURDAY,

    /**
     * Day
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_DAY,

    /**
     * Week
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_WEEK,

    /**
     * Month
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_MONTH,

    /**
     * Year
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_YEAR,

    /**
     * Now
     * @stable ICU 53
     */
    UDAT_ABSOLUTE_NOW,

    /**
     * Quarter
     * @stable ICU 63
     */
    UDAT_ABSOLUTE_QUARTER,

#ifndef U_HIDE_DRAFT_API
    /**
     * Hour
     * @draft ICU 65
     */
    UDAT_ABSOLUTE_HOUR,

    /**
     * Minute
     * @draft ICU 65
     */
    UDAT_ABSOLUTE_MINUTE,
#endif // U_HIDE_DRAFT_API

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UDateAbsoluteUnit value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UDAT_ABSOLUTE_UNIT_COUNT = UDAT_ABSOLUTE_NOW + 4
#endif  // U_HIDE_DEPRECATED_API
} UDateAbsoluteUnit;

/**
 * Represents a direction for an absolute unit e.g "Next Tuesday"
 * or "Last Tuesday"
 * @stable ICU 53
 */
typedef enum UDateDirection {

    /**
     * Two before. Not fully supported in every locale.
     * @stable ICU 53
     */
    UDAT_DIRECTION_LAST_2,

    /**
     * Last
     * @stable ICU 53
     */
    UDAT_DIRECTION_LAST,

    /**
     * This
     * @stable ICU 53
     */
    UDAT_DIRECTION_THIS,

    /**
     * Next
     * @stable ICU 53
     */
    UDAT_DIRECTION_NEXT,

    /**
     * Two after. Not fully supported in every locale.
     * @stable ICU 53
     */
    UDAT_DIRECTION_NEXT_2,

    /**
     * Plain, which means the absence of a qualifier.
     * @stable ICU 53
     */
    UDAT_DIRECTION_PLAIN,

#ifndef U_HIDE_DEPRECATED_API
    /**
     * One more than the highest normal UDateDirection value.
     * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
     */
    UDAT_DIRECTION_COUNT
#endif  // U_HIDE_DEPRECATED_API
} UDateDirection;

#if !UCONFIG_NO_BREAK_ITERATION

U_NAMESPACE_BEGIN

class BreakIterator;
class RelativeDateTimeCacheData;
class SharedNumberFormat;
class SharedPluralRules;
class SharedBreakIterator;
class NumberFormat;
class UnicodeString;
class FormattedRelativeDateTime;
class FormattedRelativeDateTimeData;

#ifndef U_HIDE_DRAFT_API
/**
 * An immutable class containing the result of a relative datetime formatting operation.
 *
 * Instances of this class are immutable and thread-safe.
 *
 * Not intended for public subclassing.
 *
 * @draft ICU 64
 */
class U_I18N_API FormattedRelativeDateTime : public UMemory, public FormattedValue {
  public:
    /**
     * Default constructor; makes an empty FormattedRelativeDateTime.
     * @draft ICU 64
     */
    FormattedRelativeDateTime() : fData(nullptr), fErrorCode(U_INVALID_STATE_ERROR) {}

    /**
     * Move constructor: Leaves the source FormattedRelativeDateTime in an undefined state.
     * @draft ICU 64
     */
    FormattedRelativeDateTime(FormattedRelativeDateTime&& src) U_NOEXCEPT;

    /**
     * Destruct an instance of FormattedRelativeDateTime.
     * @draft ICU 64
     */
    virtual ~FormattedRelativeDateTime() U_OVERRIDE;

    /** Copying not supported; use move constructor instead. */
    FormattedRelativeDateTime(const FormattedRelativeDateTime&) = delete;

    /** Copying not supported; use move assignment instead. */
    FormattedRelativeDateTime& operator=(const FormattedRelativeDateTime&) = delete;

    /**
     * Move assignment: Leaves the source FormattedRelativeDateTime in an undefined state.
     * @draft ICU 64
     */
    FormattedRelativeDateTime& operator=(FormattedRelativeDateTime&& src) U_NOEXCEPT;

    /** @copydoc FormattedValue::toString() */
    UnicodeString toString(UErrorCode& status) const U_OVERRIDE;

    /** @copydoc FormattedValue::toTempString() */
    UnicodeString toTempString(UErrorCode& status) const U_OVERRIDE;

    /** @copydoc FormattedValue::appendTo() */
    Appendable &appendTo(Appendable& appendable, UErrorCode& status) const U_OVERRIDE;

    /** @copydoc FormattedValue::nextPosition() */
    UBool nextPosition(ConstrainedFieldPosition& cfpos, UErrorCode& status) const U_OVERRIDE;

  private:
    FormattedRelativeDateTimeData *fData;
    UErrorCode fErrorCode;
    explicit FormattedRelativeDateTime(FormattedRelativeDateTimeData *results)
        : fData(results), fErrorCode(U_ZERO_ERROR) {}
    explicit FormattedRelativeDateTime(UErrorCode errorCode)
        : fData(nullptr), fErrorCode(errorCode) {}
    friend class RelativeDateTimeFormatter;
};
#endif  /* U_HIDE_DRAFT_API */

/**
 * Formats simple relative dates. There are two types of relative dates that
 * it handles:
 * <ul>
 *   <li>relative dates with a quantity e.g "in 5 days"</li>
 *   <li>relative dates without a quantity e.g "next Tuesday"</li>
 * </ul>
 * <p>
 * This API is very basic and is intended to be a building block for more
 * fancy APIs. The caller tells it exactly what to display in a locale
 * independent way. While this class automatically provides the correct plural
 * forms, the grammatical form is otherwise as neutral as possible. It is the
 * caller's responsibility to handle cut-off logic such as deciding between
 * displaying "in 7 days" or "in 1 week." This API supports relative dates
 * involving one single unit. This API does not support relative dates
 * involving compound units,
 * e.g "in 5 days and 4 hours" nor does it support parsing.
 * <p>
 * This class is mostly thread safe and immutable with the following caveats:
 * 1. The assignment operator violates Immutability. It must not be used
 *    concurrently with other operations.
 * 2. Caller must not hold onto adopted pointers.
 * <p>
 * This class is not intended for public subclassing.
 * <p>
 * Here are some examples of use:
 * <blockquote>
 * <pre>
 * UErrorCode status = U_ZERO_ERROR;
 * UnicodeString appendTo;
 * RelativeDateTimeFormatter fmt(status);
 * // Appends "in 1 day"
 * fmt.format(
 *     1, UDAT_DIRECTION_NEXT, UDAT_RELATIVE_DAYS, appendTo, status);
 * // Appends "in 3 days"
 * fmt.format(
 *     3, UDAT_DIRECTION_NEXT, UDAT_RELATIVE_DAYS, appendTo, status);
 * // Appends "3.2 years ago"
 * fmt.format(
 *     3.2, UDAT_DIRECTION_LAST, UDAT_RELATIVE_YEARS, appendTo, status);
 * // Appends "last Sunday"
 * fmt.format(UDAT_DIRECTION_LAST, UDAT_ABSOLUTE_SUNDAY, appendTo, status);
 * // Appends "this Sunday"
 * fmt.format(UDAT_DIRECTION_THIS, UDAT_ABSOLUTE_SUNDAY, appendTo, status);
 * // Appends "next Sunday"
 * fmt.format(UDAT_DIRECTION_NEXT, UDAT_ABSOLUTE_SUNDAY, appendTo, status);
 * // Appends "Sunday"
 * fmt.format(UDAT_DIRECTION_PLAIN, UDAT_ABSOLUTE_SUNDAY, appendTo, status);
 *
 * // Appends "yesterday"
 * fmt.format(UDAT_DIRECTION_LAST, UDAT_ABSOLUTE_DAY, appendTo, status);
 * // Appends "today"
 * fmt.format(UDAT_DIRECTION_THIS, UDAT_ABSOLUTE_DAY, appendTo, status);
 * // Appends "tomorrow"
 * fmt.format(UDAT_DIRECTION_NEXT, UDAT_ABSOLUTE_DAY, appendTo, status);
 * // Appends "now"
 * fmt.format(UDAT_DIRECTION_PLAIN, UDAT_ABSOLUTE_NOW, appendTo, status);
 *
 * </pre>
 * </blockquote>
 * <p>
 * In the future, we may add more forms, such as abbreviated/short forms
 * (3 secs ago), and relative day periods ("yesterday afternoon"), etc.
 *
 * The RelativeDateTimeFormatter class is not intended for public subclassing.
 *
 * @stable ICU 53
 */
class U_I18N_API RelativeDateTimeFormatter : public UObject {
public:

    /**
     * Create RelativeDateTimeFormatter with default locale.
     * @stable ICU 53
     */
    RelativeDateTimeFormatter(UErrorCode& status);

    /**
     * Create RelativeDateTimeFormatter with given locale.
     * @stable ICU 53
     */
    RelativeDateTimeFormatter(const Locale& locale, UErrorCode& status);

    /**
     * Create RelativeDateTimeFormatter with given locale and NumberFormat.
     *
     * @param locale the locale
     * @param nfToAdopt Constructed object takes ownership of this pointer.
     *   It is an error for caller to delete this pointer or change its
     *   contents after calling this constructor.
     * @param status Any error is returned here.
     * @stable ICU 53
     */
    RelativeDateTimeFormatter(
        const Locale& locale, NumberFormat *nfToAdopt, UErrorCode& status);

    /**
     * Create RelativeDateTimeFormatter with given locale, NumberFormat,
     * and capitalization context.
     *
     * @param locale the locale
     * @param nfToAdopt Constructed object takes ownership of this pointer.
     *   It is an error for caller to delete this pointer or change its
     *   contents after calling this constructor. Caller may pass NULL for
     *   this argument if they want default number format behavior.
     * @param style the format style. The UDAT_RELATIVE bit field has no effect.
     * @param capitalizationContext A value from UDisplayContext that pertains to
     * capitalization.
     * @param status Any error is returned here.
     * @stable ICU 54
     */
    RelativeDateTimeFormatter(
            const Locale& locale,
            NumberFormat *nfToAdopt,
            UDateRelativeDateTimeFormatterStyle style,
            UDisplayContext capitalizationContext,
            UErrorCode& status);

    /**
     * Copy constructor.
     * @stable ICU 53
     */
    RelativeDateTimeFormatter(const RelativeDateTimeFormatter& other);

    /**
     * Assignment operator.
     * @stable ICU 53
     */
    RelativeDateTimeFormatter& operator=(
            const RelativeDateTimeFormatter& other);

    /**
     * Destructor.
     * @stable ICU 53
     */
    virtual ~RelativeDateTimeFormatter();

    /**
     * Formats a relative date with a quantity such as "in 5 days" or
     * "3 months ago"
     *
     * This method returns a String. To get more information about the
     * formatting result, use formatToValue().
     *
     * @param quantity The numerical amount e.g 5. This value is formatted
     * according to this object's NumberFormat object.
     * @param direction NEXT means a future relative date; LAST means a past
     * relative date. If direction is anything else, this method sets
     * status to U_ILLEGAL_ARGUMENT_ERROR.
     * @param unit the unit e.g day? month? year?
     * @param appendTo The string to which the formatted result will be
     *  appended
     * @param status ICU error code returned here.
     * @return appendTo
     * @stable ICU 53
     */
    UnicodeString& format(
            double quantity,
            UDateDirection direction,
            UDateRelativeUnit unit,
            UnicodeString& appendTo,
            UErrorCode& status) const;

#ifndef U_HIDE_DRAFT_API
    /**
     * Formats a relative date with a quantity such as "in 5 days" or
     * "3 months ago"
     *
     * This method returns a FormattedRelativeDateTime, which exposes more
     * information than the String returned by format().
     *
     * @param quantity The numerical amount e.g 5. This value is formatted
     * according to this object's NumberFormat object.
     * @param direction NEXT means a future relative date; LAST means a past
     * relative date. If direction is anything else, this method sets
     * status to U_ILLEGAL_ARGUMENT_ERROR.
     * @param unit the unit e.g day? month? year?
     * @param status ICU error code returned here.
     * @return The formatted relative datetime
     * @draft ICU 64
     */
    FormattedRelativeDateTime formatToValue(
            double quantity,
            UDateDirection direction,
            UDateRelativeUnit unit,
            UErrorCode& status) const;
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Formats a relative date without a quantity.
     *
     * This method returns a String. To get more information about the
     * formatting result, use formatToValue().
     *
     * @param direction NEXT, LAST, THIS, etc.
     * @param unit e.g SATURDAY, DAY, MONTH
     * @param appendTo The string to which the formatted result will be
     *  appended. If the value of direction is documented as not being fully
     *  supported in all locales then this method leaves appendTo unchanged if
     *  no format string is available.
     * @param status ICU error code returned here.
     * @return appendTo
     * @stable ICU 53
     */
    UnicodeString& format(
            UDateDirection direction,
            UDateAbsoluteUnit unit,
            UnicodeString& appendTo,
            UErrorCode& status) const;

#ifndef U_HIDE_DRAFT_API
    /**
     * Formats a relative date without a quantity.
     *
     * This method returns a FormattedRelativeDateTime, which exposes more
     * information than the String returned by format().
     *
     * If the string is not available in the requested locale, the return
     * value will be empty (calling toString will give an empty string).
     *
     * @param direction NEXT, LAST, THIS, etc.
     * @param unit e.g SATURDAY, DAY, MONTH
     * @param status ICU error code returned here.
     * @return The formatted relative datetime
     * @draft ICU 64
     */
    FormattedRelativeDateTime formatToValue(
            UDateDirection direction,
            UDateAbsoluteUnit unit,
            UErrorCode& status) const;
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Format a combination of URelativeDateTimeUnit and numeric offset
     * using a numeric style, e.g. "1 week ago", "in 1 week",
     * "5 weeks ago", "in 5 weeks".
     *
     * This method returns a String. To get more information about the
     * formatting result, use formatNumericToValue().
     * 
     * @param offset    The signed offset for the specified unit. This
     *                  will be formatted according to this object's
     *                  NumberFormat object.
     * @param unit      The unit to use when formatting the relative
     *                  date, e.g. UDAT_REL_UNIT_WEEK,
     *                  UDAT_REL_UNIT_FRIDAY.
     * @param appendTo  The string to which the formatted result will be
     *                  appended.
     * @param status    ICU error code returned here.
     * @return          appendTo
     * @stable ICU 57
     */
    UnicodeString& formatNumeric(
            double offset,
            URelativeDateTimeUnit unit,
            UnicodeString& appendTo,
            UErrorCode& status) const;

#ifndef U_HIDE_DRAFT_API
    /**
     * Format a combination of URelativeDateTimeUnit and numeric offset
     * using a numeric style, e.g. "1 week ago", "in 1 week",
     * "5 weeks ago", "in 5 weeks".
     *
     * This method returns a FormattedRelativeDateTime, which exposes more
     * information than the String returned by formatNumeric().
     * 
     * @param offset    The signed offset for the specified unit. This
     *                  will be formatted according to this object's
     *                  NumberFormat object.
     * @param unit      The unit to use when formatting the relative
     *                  date, e.g. UDAT_REL_UNIT_WEEK,
     *                  UDAT_REL_UNIT_FRIDAY.
     * @param status    ICU error code returned here.
     * @return          The formatted relative datetime
     * @draft ICU 64
     */
    FormattedRelativeDateTime formatNumericToValue(
            double offset,
            URelativeDateTimeUnit unit,
            UErrorCode& status) const;
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Format a combination of URelativeDateTimeUnit and numeric offset
     * using a text style if possible, e.g. "last week", "this week",
     * "next week", "yesterday", "tomorrow". Falls back to numeric
     * style if no appropriate text term is available for the specified
     * offset in the object's locale.
     *
     * This method returns a String. To get more information about the
     * formatting result, use formatToValue().
     *
     * @param offset    The signed offset for the specified unit.
     * @param unit      The unit to use when formatting the relative
     *                  date, e.g. UDAT_REL_UNIT_WEEK,
     *                  UDAT_REL_UNIT_FRIDAY.
     * @param appendTo  The string to which the formatted result will be
     *                  appended.
     * @param status    ICU error code returned here.
     * @return          appendTo
     * @stable ICU 57
     */
    UnicodeString& format(
            double offset,
            URelativeDateTimeUnit unit,
            UnicodeString& appendTo,
            UErrorCode& status) const;

#ifndef U_HIDE_DRAFT_API
    /**
     * Format a combination of URelativeDateTimeUnit and numeric offset
     * using a text style if possible, e.g. "last week", "this week",
     * "next week", "yesterday", "tomorrow". Falls back to numeric
     * style if no appropriate text term is available for the specified
     * offset in the object's locale.
     *
     * This method returns a FormattedRelativeDateTime, which exposes more
     * information than the String returned by format().
     *
     * @param offset    The signed offset for the specified unit.
     * @param unit      The unit to use when formatting the relative
     *                  date, e.g. UDAT_REL_UNIT_WEEK,
     *                  UDAT_REL_UNIT_FRIDAY.
     * @param status    ICU error code returned here.
     * @return          The formatted relative datetime
     * @draft ICU 64
     */
    FormattedRelativeDateTime formatToValue(
            double offset,
            URelativeDateTimeUnit unit,
            UErrorCode& status) const;
#endif  /* U_HIDE_DRAFT_API */

    /**
     * Combines a relative date string and a time string in this object's
     * locale. This is done with the same date-time separator used for the
     * default calendar in this locale.
     *
     * @param relativeDateString the relative date, e.g 'yesterday'
     * @param timeString the time e.g '3:45'
     * @param appendTo concatenated date and time appended here
     * @param status ICU error code returned here.
     * @return appendTo
     * @stable ICU 53
     */
    UnicodeString& combineDateAndTime(
            const UnicodeString& relativeDateString,
            const UnicodeString& timeString,
            UnicodeString& appendTo,
            UErrorCode& status) const;

    /**
     * Returns the NumberFormat this object is using.
     *
     * @stable ICU 53
     */
    const NumberFormat& getNumberFormat() const;

    /**
     * Returns the capitalization context.
     *
     * @stable ICU 54
     */
    UDisplayContext getCapitalizationContext() const;

    /**
     * Returns the format style.
     *
     * @stable ICU 54
     */
    UDateRelativeDateTimeFormatterStyle getFormatStyle() const;

private:
    const RelativeDateTimeCacheData* fCache;
    const SharedNumberFormat *fNumberFormat;
    const SharedPluralRules *fPluralRules;
    UDateRelativeDateTimeFormatterStyle fStyle;
    UDisplayContext fContext;
    const SharedBreakIterator *fOptBreakIterator;
    Locale fLocale;
    void init(
            NumberFormat *nfToAdopt,
            BreakIterator *brkIter,
            UErrorCode &status);
    UnicodeString& adjustForContext(UnicodeString &) const;
    UBool checkNoAdjustForContext(UErrorCode& status) const;

    template<typename F, typename... Args>
    UnicodeString& doFormat(
            F callback,
            UnicodeString& appendTo,
            UErrorCode& status,
            Args... args) const;

#ifndef U_HIDE_DRAFT_API  // for FormattedRelativeDateTime
    template<typename F, typename... Args>
    FormattedRelativeDateTime doFormatToValue(
            F callback,
            UErrorCode& status,
            Args... args) const;
#endif  // U_HIDE_DRAFT_API

    void formatImpl(
            double quantity,
            UDateDirection direction,
            UDateRelativeUnit unit,
            FormattedRelativeDateTimeData& output,
            UErrorCode& status) const;
    void formatAbsoluteImpl(
            UDateDirection direction,
            UDateAbsoluteUnit unit,
            FormattedRelativeDateTimeData& output,
            UErrorCode& status) const;
    void formatNumericImpl(
            double offset,
            URelativeDateTimeUnit unit,
            FormattedRelativeDateTimeData& output,
            UErrorCode& status) const;
    void formatRelativeImpl(
            double offset,
            URelativeDateTimeUnit unit,
            FormattedRelativeDateTimeData& output,
            UErrorCode& status) const;
};

U_NAMESPACE_END

#endif /* !UCONFIG_NO_BREAK_ITERATION */
#endif /* !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



#if !UCONFIG_NO_FORMATTING && !UCONFIG_NO_BREAK_ITERATION

#include <cmath>
#include <functional>










// #include "unicode/brkiter.h"










// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
* Copyright (C) 2014, International Business Machines
* Corporation and others.  All Rights Reserved.
******************************************************************************
* sharedbreakiterator.h
*/





#if !UCONFIG_NO_BREAK_ITERATION

U_NAMESPACE_BEGIN

class BreakIterator;

// SharedBreakIterator encapsulates a shared BreakIterator. Because
// BreakIterator has mutable semantics, clients must ensure that all uses
// of a particular shared BreakIterator is protected by the same mutex
// ensuring that only one thread at a time gets access to that shared
// BreakIterator. Clients can accomplish this by creating a mutex for all
// uses of break iterator within a particular class. Then objects of that
// class may then freely share break iterators among themselves. However,
// these shared break iterators must never be exposed outside of that class.
class U_I18N_API SharedBreakIterator : public SharedObject {
public:
    SharedBreakIterator(BreakIterator *biToAdopt);
    virtual ~SharedBreakIterator();

    BreakIterator *get() const { return ptr; }
    BreakIterator *operator->() const { return ptr; }
    BreakIterator &operator*() const { return *ptr; }
private:
    BreakIterator *ptr;
    SharedBreakIterator(const SharedBreakIterator &);
    SharedBreakIterator &operator=(const SharedBreakIterator &);
};

U_NAMESPACE_END

#endif













// Copied from uscript_props.cpp

U_NAMESPACE_BEGIN

// RelativeDateTimeFormatter specific data for a single locale
class RelativeDateTimeCacheData: public SharedObject {
public:
    RelativeDateTimeCacheData() : combinedDateAndTime(nullptr) {
        // Initialize the cache arrays
        for (int32_t style = 0; style < UDAT_STYLE_COUNT; ++style) {
            for (int32_t relUnit = 0; relUnit < UDAT_REL_UNIT_COUNT; ++relUnit) {
                for (int32_t pl = 0; pl < StandardPlural::COUNT; ++pl) {
                    relativeUnitsFormatters[style][relUnit][0][pl] = nullptr;
                    relativeUnitsFormatters[style][relUnit][1][pl] = nullptr;
                }
            }
        }
        for (int32_t i = 0; i < UDAT_STYLE_COUNT; ++i) {
          fallBackCache[i] = -1;
        }
    }
    virtual ~RelativeDateTimeCacheData();

    // no numbers: e.g Next Tuesday; Yesterday; etc.
    UnicodeString absoluteUnits[UDAT_STYLE_COUNT][UDAT_ABSOLUTE_UNIT_COUNT][UDAT_DIRECTION_COUNT];

    // SimpleFormatter pointers for relative unit format,
    // e.g., Next Tuesday; Yesterday; etc. For third index, 0
    // means past, e.g., 5 days ago; 1 means future, e.g., in 5 days.
    SimpleFormatter *relativeUnitsFormatters[UDAT_STYLE_COUNT]
        [UDAT_REL_UNIT_COUNT][2][StandardPlural::COUNT];

    const UnicodeString& getAbsoluteUnitString(int32_t fStyle,
                                               UDateAbsoluteUnit unit,
                                               UDateDirection direction) const;
    const SimpleFormatter* getRelativeUnitFormatter(int32_t fStyle,
                                                    UDateRelativeUnit unit,
                                                    int32_t pastFutureIndex,
                                                    int32_t pluralUnit) const;
    const SimpleFormatter* getRelativeDateTimeUnitFormatter(int32_t fStyle,
                                                    URelativeDateTimeUnit unit,
                                                    int32_t pastFutureIndex,
                                                    int32_t pluralUnit) const;

    const UnicodeString emptyString;

    // Mappping from source to target styles for alias fallback.
    int32_t fallBackCache[UDAT_STYLE_COUNT];

    void adoptCombinedDateAndTime(SimpleFormatter *fmtToAdopt) {
        delete combinedDateAndTime;
        combinedDateAndTime = fmtToAdopt;
    }
    const SimpleFormatter *getCombinedDateAndTime() const {
        return combinedDateAndTime;
    }

private:
    SimpleFormatter *combinedDateAndTime;
    RelativeDateTimeCacheData(const RelativeDateTimeCacheData &other);
    RelativeDateTimeCacheData& operator=(
            const RelativeDateTimeCacheData &other);
};

RelativeDateTimeCacheData::~RelativeDateTimeCacheData() {
    // clear out the cache arrays
    for (int32_t style = 0; style < UDAT_STYLE_COUNT; ++style) {
        for (int32_t relUnit = 0; relUnit < UDAT_REL_UNIT_COUNT; ++relUnit) {
            for (int32_t pl = 0; pl < StandardPlural::COUNT; ++pl) {
                delete relativeUnitsFormatters[style][relUnit][0][pl];
                delete relativeUnitsFormatters[style][relUnit][1][pl];
            }
        }
    }
    delete combinedDateAndTime;
}


// Use fallback cache for absolute units.
const UnicodeString& RelativeDateTimeCacheData::getAbsoluteUnitString(
        int32_t fStyle, UDateAbsoluteUnit unit, UDateDirection direction) const {
    int32_t style = fStyle;
    do {
        if (!absoluteUnits[style][unit][direction].isEmpty()) {
            return absoluteUnits[style][unit][direction];
        }
        style = fallBackCache[style];
    } while (style != -1);
    return emptyString;
}

 const SimpleFormatter* RelativeDateTimeCacheData::getRelativeUnitFormatter(
        int32_t fStyle,
        UDateRelativeUnit unit,
        int32_t pastFutureIndex,
        int32_t pluralUnit) const {
   URelativeDateTimeUnit rdtunit = UDAT_REL_UNIT_COUNT;
   switch (unit) {
       case UDAT_RELATIVE_YEARS:   rdtunit = UDAT_REL_UNIT_YEAR; break;
       case UDAT_RELATIVE_MONTHS:  rdtunit = UDAT_REL_UNIT_MONTH; break;
       case UDAT_RELATIVE_WEEKS:   rdtunit = UDAT_REL_UNIT_WEEK; break;
       case UDAT_RELATIVE_DAYS:    rdtunit = UDAT_REL_UNIT_DAY; break;
       case UDAT_RELATIVE_HOURS:   rdtunit = UDAT_REL_UNIT_HOUR; break;
       case UDAT_RELATIVE_MINUTES: rdtunit = UDAT_REL_UNIT_MINUTE; break;
       case UDAT_RELATIVE_SECONDS: rdtunit = UDAT_REL_UNIT_SECOND; break;
       default: // a unit that the above method does not handle
            return nullptr;
   }

   return getRelativeDateTimeUnitFormatter(fStyle, rdtunit, pastFutureIndex, pluralUnit);
 }

 // Use fallback cache for SimpleFormatter relativeUnits.
 const SimpleFormatter* RelativeDateTimeCacheData::getRelativeDateTimeUnitFormatter(
        int32_t fStyle,
        URelativeDateTimeUnit unit,
        int32_t pastFutureIndex,
        int32_t pluralUnit) const {
    while (true) {
        int32_t style = fStyle;
        do {
            if (relativeUnitsFormatters[style][unit][pastFutureIndex][pluralUnit] != nullptr) {
                return relativeUnitsFormatters[style][unit][pastFutureIndex][pluralUnit];
            }
            style = fallBackCache[style];
        } while (style != -1);

        if (pluralUnit == StandardPlural::OTHER) {
            break;
        }
        pluralUnit = StandardPlural::OTHER;
    }
    return nullptr;  // No formatter found.
 }

static UBool getStringWithFallback(
        const UResourceBundle *resource,
        const char *key,
        UnicodeString &result,
        UErrorCode &status) {
    int32_t len = 0;
    const UChar *resStr = ures_getStringByKeyWithFallback(
        resource, key, &len, &status);
    if (U_FAILURE(status)) {
        return FALSE;
    }
    result.setTo(TRUE, resStr, len);
    return TRUE;
}


static UBool getStringByIndex(
        const UResourceBundle *resource,
        int32_t idx,
        UnicodeString &result,
        UErrorCode &status) {
    int32_t len = 0;
    const UChar *resStr = ures_getStringByIndex(
            resource, idx, &len, &status);
    if (U_FAILURE(status)) {
        return FALSE;
    }
    result.setTo(TRUE, resStr, len);
    return TRUE;
}

namespace {

/**
 * Sink for enumerating all of the measurement unit display names.
 *
 * More specific bundles (en_GB) are enumerated before their parents (en_001, en, root):
 * Only store a value if it is still missing, that is, it has not been overridden.
 */
struct RelDateTimeFmtDataSink : public ResourceSink {

    /**
     * Sink for patterns for relative dates and times. For example,
     * fields/relative/...
     */

    // Generic unit enum for storing Unit info.
    typedef enum RelAbsUnit {
        INVALID_UNIT = -1,
        SECOND,
        MINUTE,
        HOUR,
        DAY,
        WEEK,
        MONTH,
        QUARTER,
        YEAR,
        SUNDAY,
        MONDAY,
        TUESDAY,
        WEDNESDAY,
        THURSDAY,
        FRIDAY,
        SATURDAY
    } RelAbsUnit;

    static int32_t relUnitFromGeneric(RelAbsUnit genUnit) {
        // Converts the generic units to UDAT_RELATIVE version.
        switch (genUnit) {
            case SECOND:
                return UDAT_REL_UNIT_SECOND;
            case MINUTE:
                return UDAT_REL_UNIT_MINUTE;
            case HOUR:
                return UDAT_REL_UNIT_HOUR;
            case DAY:
                return UDAT_REL_UNIT_DAY;
            case WEEK:
                return UDAT_REL_UNIT_WEEK;
            case MONTH:
                return UDAT_REL_UNIT_MONTH;
            case QUARTER:
                return UDAT_REL_UNIT_QUARTER;
            case YEAR:
                return UDAT_REL_UNIT_YEAR;
            case SUNDAY:
                return UDAT_REL_UNIT_SUNDAY;
            case MONDAY:
                return UDAT_REL_UNIT_MONDAY;
            case TUESDAY:
                return UDAT_REL_UNIT_TUESDAY;
            case WEDNESDAY:
                return UDAT_REL_UNIT_WEDNESDAY;
            case THURSDAY:
                return UDAT_REL_UNIT_THURSDAY;
            case FRIDAY:
                return UDAT_REL_UNIT_FRIDAY;
            case SATURDAY:
                return UDAT_REL_UNIT_SATURDAY;
            default:
                return -1;
        }
    }

    static int32_t absUnitFromGeneric(RelAbsUnit genUnit) {
        // Converts the generic units to UDAT_RELATIVE version.
        switch (genUnit) {
            case DAY:
                return UDAT_ABSOLUTE_DAY;
            case WEEK:
                return UDAT_ABSOLUTE_WEEK;
            case MONTH:
                return UDAT_ABSOLUTE_MONTH;
            case QUARTER:
                return UDAT_ABSOLUTE_QUARTER;
            case YEAR:
                return UDAT_ABSOLUTE_YEAR;
            case SUNDAY:
                return UDAT_ABSOLUTE_SUNDAY;
            case MONDAY:
                return UDAT_ABSOLUTE_MONDAY;
            case TUESDAY:
                return UDAT_ABSOLUTE_TUESDAY;
            case WEDNESDAY:
                return UDAT_ABSOLUTE_WEDNESDAY;
            case THURSDAY:
                return UDAT_ABSOLUTE_THURSDAY;
            case FRIDAY:
                return UDAT_ABSOLUTE_FRIDAY;
            case SATURDAY:
                return UDAT_ABSOLUTE_SATURDAY;
            case HOUR:
                return UDAT_ABSOLUTE_HOUR;
            case MINUTE:
                return UDAT_ABSOLUTE_MINUTE;
            default:
                return -1;
        }
    }

    static int32_t keyToDirection(const char* key) {
        if (uprv_strcmp(key, "-2") == 0) {
            return UDAT_DIRECTION_LAST_2;
        }
        if (uprv_strcmp(key, "-1") == 0) {
            return UDAT_DIRECTION_LAST;
        }
        if (uprv_strcmp(key, "0") == 0) {
            return UDAT_DIRECTION_THIS;
        }
        if (uprv_strcmp(key, "1") == 0) {
            return UDAT_DIRECTION_NEXT;
        }
        if (uprv_strcmp(key, "2") == 0) {
            return UDAT_DIRECTION_NEXT_2;
        }
        return -1;
    }

    // Values kept between levels of parsing the CLDR data.
    int32_t pastFutureIndex;  // 0 == past or 1 ==  future
    UDateRelativeDateTimeFormatterStyle style;  // {LONG, SHORT, NARROW}
    RelAbsUnit genericUnit;

    RelativeDateTimeCacheData &outputData;

    // Constructor
    RelDateTimeFmtDataSink(RelativeDateTimeCacheData& cacheData)
        : outputData(cacheData) {
        // Clear cacheData.fallBackCache
        cacheData.fallBackCache[UDAT_STYLE_LONG] = -1;
        cacheData.fallBackCache[UDAT_STYLE_SHORT] = -1;
        cacheData.fallBackCache[UDAT_STYLE_NARROW] = -1;
    }

    ~RelDateTimeFmtDataSink();

    // Utility functions
    static UDateRelativeDateTimeFormatterStyle styleFromString(const char *s) {
        int32_t len = static_cast<int32_t>(uprv_strlen(s));
        if (len >= 7 && uprv_strcmp(s + len - 7, "-narrow") == 0) {
            return UDAT_STYLE_NARROW;
        }
        if (len >= 6 && uprv_strcmp(s + len - 6, "-short") == 0) {
            return UDAT_STYLE_SHORT;
        }
        return UDAT_STYLE_LONG;
    }

    static int32_t styleSuffixLength(UDateRelativeDateTimeFormatterStyle style) {
        switch (style) {
            case UDAT_STYLE_NARROW:
                return 7;
            case UDAT_STYLE_SHORT:
                return 6;
            default:
                return 0;
        }
    }

    // Utility functions
    static UDateRelativeDateTimeFormatterStyle styleFromAliasUnicodeString(UnicodeString s) {
        static const UChar narrow[7] = {0x002D, 0x006E, 0x0061, 0x0072, 0x0072, 0x006F, 0x0077};
        static const UChar sshort[6] = {0x002D, 0x0073, 0x0068, 0x006F, 0x0072, 0x0074,};
        if (s.endsWith(narrow, 7)) {
            return UDAT_STYLE_NARROW;
        }
        if (s.endsWith(sshort, 6)) {
            return UDAT_STYLE_SHORT;
        }
        return UDAT_STYLE_LONG;
    }

    static RelAbsUnit unitOrNegativeFromString(const char* keyword, int32_t length) {
        // Quick check from string to enum.
        switch (length) {
            case 3:
                if (uprv_strncmp(keyword, "day", length) == 0) {
                    return DAY;
                } else if (uprv_strncmp(keyword, "sun", length) == 0) {
                    return SUNDAY;
                } else if (uprv_strncmp(keyword, "mon", length) == 0) {
                    return MONDAY;
                } else if (uprv_strncmp(keyword, "tue", length) == 0) {
                    return TUESDAY;
                } else if (uprv_strncmp(keyword, "wed", length) == 0) {
                    return WEDNESDAY;
                } else if (uprv_strncmp(keyword, "thu", length) == 0) {
                    return THURSDAY;
                } else if (uprv_strncmp(keyword, "fri", length) == 0) {
                    return FRIDAY;
                } else if (uprv_strncmp(keyword, "sat", length) == 0) {
                    return SATURDAY;
                }
                break;
            case 4:
                if (uprv_strncmp(keyword, "hour", length) == 0) {
                    return HOUR;
                } else if (uprv_strncmp(keyword, "week", length) == 0) {
                    return WEEK;
                } else if (uprv_strncmp(keyword, "year", length) == 0) {
                    return YEAR;
                }
                break;
            case 5:
                if (uprv_strncmp(keyword, "month", length) == 0) {
                    return MONTH;
                }
                break;
            case 6:
                if (uprv_strncmp(keyword, "minute", length) == 0) {
                    return MINUTE;
                } else if (uprv_strncmp(keyword, "second", length) == 0) {
                    return SECOND;
                }
                break;
            case 7:
                if (uprv_strncmp(keyword, "quarter", length) == 0) {
                    return QUARTER;  // TODO: Check @provisional
                  }
                break;
            default:
                break;
        }
        return INVALID_UNIT;
    }

    void handlePlainDirection(ResourceValue &value, UErrorCode &errorCode) {
        // Handle Display Name for PLAIN direction for some units.
        if (U_FAILURE(errorCode)) { return; }

        int32_t absUnit = absUnitFromGeneric(genericUnit);
        if (absUnit < 0) {
          return;  // Not interesting.
        }

        // Store displayname if not set.
        if (outputData.absoluteUnits[style]
            [absUnit][UDAT_DIRECTION_PLAIN].isEmpty()) {
            outputData.absoluteUnits[style]
                [absUnit][UDAT_DIRECTION_PLAIN].fastCopyFrom(value.getUnicodeString(errorCode));
            return;
        }
    }

    void consumeTableRelative(const char *key, ResourceValue &value, UErrorCode &errorCode) {
        ResourceTable unitTypesTable = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }

        for (int32_t i = 0; unitTypesTable.getKeyAndValue(i, key, value); ++i) {
            if (value.getType() == URES_STRING) {
                int32_t direction = keyToDirection(key);
                if (direction < 0) {
                  continue;
                }

                int32_t relUnitIndex = relUnitFromGeneric(genericUnit);
                if (relUnitIndex == UDAT_REL_UNIT_SECOND && uprv_strcmp(key, "0") == 0 &&
                    outputData.absoluteUnits[style][UDAT_ABSOLUTE_NOW][UDAT_DIRECTION_PLAIN].isEmpty()) {
                    // Handle "NOW"
                    outputData.absoluteUnits[style][UDAT_ABSOLUTE_NOW]
                        [UDAT_DIRECTION_PLAIN].fastCopyFrom(value.getUnicodeString(errorCode));
                }

                int32_t absUnitIndex = absUnitFromGeneric(genericUnit);
                if (absUnitIndex < 0) {
                    continue;
                }
                // Only reset if slot is empty.
                if (outputData.absoluteUnits[style][absUnitIndex][direction].isEmpty()) {
                    outputData.absoluteUnits[style][absUnitIndex]
                        [direction].fastCopyFrom(value.getUnicodeString(errorCode));
                }
            }
        }
    }

    void consumeTimeDetail(int32_t relUnitIndex,
                           const char *key, ResourceValue &value, UErrorCode &errorCode) {
        ResourceTable unitTypesTable = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }

          for (int32_t i = 0; unitTypesTable.getKeyAndValue(i, key, value); ++i) {
            if (value.getType() == URES_STRING) {
                int32_t pluralIndex = StandardPlural::indexOrNegativeFromString(key);
                if (pluralIndex >= 0) {
                    SimpleFormatter **patterns =
                        outputData.relativeUnitsFormatters[style][relUnitIndex]
                        [pastFutureIndex];
                    // Only set if not already established.
                    if (patterns[pluralIndex] == nullptr) {
                        patterns[pluralIndex] = new SimpleFormatter(
                            value.getUnicodeString(errorCode), 0, 1, errorCode);
                        if (patterns[pluralIndex] == nullptr) {
                            errorCode = U_MEMORY_ALLOCATION_ERROR;
                        }
                    }
                }
            }
        }
    }

    void consumeTableRelativeTime(const char *key, ResourceValue &value, UErrorCode &errorCode) {
        ResourceTable relativeTimeTable = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }

        int32_t relUnitIndex = relUnitFromGeneric(genericUnit);
        if (relUnitIndex < 0) {
            return;
        }
        for (int32_t i = 0; relativeTimeTable.getKeyAndValue(i, key, value); ++i) {
            if (uprv_strcmp(key, "past") == 0) {
                pastFutureIndex = 0;
            } else if (uprv_strcmp(key, "future") == 0) {
                pastFutureIndex = 1;
            } else {
                // Unknown key.
                continue;
            }
            consumeTimeDetail(relUnitIndex, key, value, errorCode);
        }
    }

    void consumeAlias(const char *key, const ResourceValue &value, UErrorCode &errorCode) {

        UDateRelativeDateTimeFormatterStyle sourceStyle = styleFromString(key);
        const UnicodeString valueStr = value.getAliasUnicodeString(errorCode);
        if (U_FAILURE(errorCode)) { return; }

        UDateRelativeDateTimeFormatterStyle targetStyle =
            styleFromAliasUnicodeString(valueStr);

        if (sourceStyle == targetStyle) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        if (outputData.fallBackCache[sourceStyle] != -1 &&
            outputData.fallBackCache[sourceStyle] != targetStyle) {
            errorCode = U_INVALID_FORMAT_ERROR;
            return;
        }
        outputData.fallBackCache[sourceStyle] = targetStyle;
    }

    void consumeTimeUnit(const char *key, ResourceValue &value, UErrorCode &errorCode) {
        ResourceTable unitTypesTable = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }

        for (int32_t i = 0; unitTypesTable.getKeyAndValue(i, key, value); ++i) {
            // Handle display name.
            if (uprv_strcmp(key, "dn") == 0 && value.getType() == URES_STRING) {
                handlePlainDirection(value, errorCode);
            }
            if (value.getType() == URES_TABLE) {
                if (uprv_strcmp(key, "relative") == 0) {
                    consumeTableRelative(key, value, errorCode);
                } else if (uprv_strcmp(key, "relativeTime") == 0) {
                    consumeTableRelativeTime(key, value, errorCode);
                }
            }
        }
    }

    virtual void put(const char *key, ResourceValue &value,
                     UBool /*noFallback*/, UErrorCode &errorCode) {
        // Main entry point to sink
        ResourceTable table = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }
        for (int32_t i = 0; table.getKeyAndValue(i, key, value); ++i) {
            if (value.getType() == URES_ALIAS) {
                consumeAlias(key, value, errorCode);
            } else {
                style = styleFromString(key);
                int32_t unitSize = static_cast<int32_t>(uprv_strlen(key)) - styleSuffixLength(style);
                genericUnit = unitOrNegativeFromString(key, unitSize);
                if (style >= 0 && genericUnit != INVALID_UNIT) {
                    consumeTimeUnit(key, value, errorCode);
                }
            }
        }
    }

};

// Virtual destructors must be defined out of line.
RelDateTimeFmtDataSink::~RelDateTimeFmtDataSink() {}
} // namespace

static const DateFormatSymbols::DtWidthType styleToDateFormatSymbolWidth[UDAT_STYLE_COUNT] = {
  DateFormatSymbols::WIDE, DateFormatSymbols::SHORT, DateFormatSymbols::NARROW
};

// Get days of weeks from the DateFormatSymbols class.
static void loadWeekdayNames(UnicodeString absoluteUnits[UDAT_STYLE_COUNT]
                                 [UDAT_ABSOLUTE_UNIT_COUNT][UDAT_DIRECTION_COUNT],
                             const char* localeId,
                             UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    Locale locale(localeId);
    DateFormatSymbols dfSym(locale, status);
    if (U_FAILURE(status)) {
        return;
    }
    for (int32_t style = 0; style < UDAT_STYLE_COUNT; ++style) {
        DateFormatSymbols::DtWidthType dtfmtWidth = styleToDateFormatSymbolWidth[style];
        int32_t count;
        const UnicodeString* weekdayNames =
            dfSym.getWeekdays(count, DateFormatSymbols::STANDALONE, dtfmtWidth);
        for (int32_t dayIndex = UDAT_ABSOLUTE_SUNDAY;
                dayIndex <= UDAT_ABSOLUTE_SATURDAY; ++ dayIndex) {
            int32_t dateSymbolIndex = (dayIndex - UDAT_ABSOLUTE_SUNDAY) + UCAL_SUNDAY;
            absoluteUnits[style][dayIndex][UDAT_DIRECTION_PLAIN].fastCopyFrom(
                weekdayNames[dateSymbolIndex]);
        }
    }
}

static UBool loadUnitData(
        const UResourceBundle *resource,
        RelativeDateTimeCacheData &cacheData,
        const char* localeId,
        UErrorCode &status) {

    RelDateTimeFmtDataSink sink(cacheData);

    ures_getAllItemsWithFallback(resource, "fields", sink, status);
    if (U_FAILURE(status)) {
        return false;
    }

    // Get the weekday names from DateFormatSymbols.
    loadWeekdayNames(cacheData.absoluteUnits, localeId, status);
    return U_SUCCESS(status);
}

static UBool getDateTimePattern(
        const UResourceBundle *resource,
        UnicodeString &result,
        UErrorCode &status) {
    UnicodeString defaultCalendarName;
    if (!getStringWithFallback(
            resource,
            "calendar/default",
            defaultCalendarName,
            status)) {
        return FALSE;
    }
    CharString pathBuffer;
    pathBuffer.append("calendar/", status)
            .appendInvariantChars(defaultCalendarName, status)
            .append("/DateTimePatterns", status);
    LocalUResourceBundlePointer topLevel(
            ures_getByKeyWithFallback(
                    resource, pathBuffer.data(), nullptr, &status));
    if (U_FAILURE(status)) {
        return FALSE;
    }
    int32_t size = ures_getSize(topLevel.getAlias());
    if (size <= 8) {
        // Oops, size is too small to access the index that we want, fallback
        // to a hard-coded value.
        result = UNICODE_STRING_SIMPLE("{1} {0}");
        return TRUE;
    }
    return getStringByIndex(topLevel.getAlias(), 8, result, status);
}

template<> U_I18N_API
const RelativeDateTimeCacheData *LocaleCacheKey<RelativeDateTimeCacheData>::createObject(const void * /*unused*/, UErrorCode &status) const {
    const char *localeId = fLoc.getName();
    LocalUResourceBundlePointer topLevel(ures_open(nullptr, localeId, &status));
    if (U_FAILURE(status)) {
        return nullptr;
    }
    LocalPointer<RelativeDateTimeCacheData> result(
            new RelativeDateTimeCacheData());
    if (result.isNull()) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    if (!loadUnitData(
            topLevel.getAlias(),
            *result,
            localeId,
            status)) {
        return nullptr;
    }
    UnicodeString dateTimePattern;
    if (!getDateTimePattern(topLevel.getAlias(), dateTimePattern, status)) {
        return nullptr;
    }
    result->adoptCombinedDateAndTime(
            new SimpleFormatter(dateTimePattern, 2, 2, status));
    if (U_FAILURE(status)) {
        return nullptr;
    }
    result->addRef();
    return result.orphan();
}



static constexpr number::impl::Field kRDTNumericField
    = StringBuilderFieldUtils::compress<UFIELD_CATEGORY_RELATIVE_DATETIME, UDAT_REL_NUMERIC_FIELD>();

static constexpr number::impl::Field kRDTLiteralField
    = StringBuilderFieldUtils::compress<UFIELD_CATEGORY_RELATIVE_DATETIME, UDAT_REL_LITERAL_FIELD>();

class FormattedRelativeDateTimeData : public FormattedValueStringBuilderImpl {
public:
    FormattedRelativeDateTimeData() : FormattedValueStringBuilderImpl(kRDTNumericField) {}
    virtual ~FormattedRelativeDateTimeData();
};

FormattedRelativeDateTimeData::~FormattedRelativeDateTimeData() = default;


UPRV_FORMATTED_VALUE_SUBCLASS_AUTO_IMPL(FormattedRelativeDateTime)


RelativeDateTimeFormatter::RelativeDateTimeFormatter(UErrorCode& status) :
        fCache(nullptr),
        fNumberFormat(nullptr),
        fPluralRules(nullptr),
        fStyle(UDAT_STYLE_LONG),
        fContext(UDISPCTX_CAPITALIZATION_NONE),
        fOptBreakIterator(nullptr) {
    init(nullptr, nullptr, status);
}

RelativeDateTimeFormatter::RelativeDateTimeFormatter(
        const Locale& locale, UErrorCode& status) :
        fCache(nullptr),
        fNumberFormat(nullptr),
        fPluralRules(nullptr),
        fStyle(UDAT_STYLE_LONG),
        fContext(UDISPCTX_CAPITALIZATION_NONE),
        fOptBreakIterator(nullptr),
        fLocale(locale) {
    init(nullptr, nullptr, status);
}

RelativeDateTimeFormatter::RelativeDateTimeFormatter(
        const Locale& locale, NumberFormat *nfToAdopt, UErrorCode& status) :
        fCache(nullptr),
        fNumberFormat(nullptr),
        fPluralRules(nullptr),
        fStyle(UDAT_STYLE_LONG),
        fContext(UDISPCTX_CAPITALIZATION_NONE),
        fOptBreakIterator(nullptr),
        fLocale(locale) {
    init(nfToAdopt, nullptr, status);
}

RelativeDateTimeFormatter::RelativeDateTimeFormatter(
        const Locale& locale,
        NumberFormat *nfToAdopt,
        UDateRelativeDateTimeFormatterStyle styl,
        UDisplayContext capitalizationContext,
        UErrorCode& status) :
        fCache(nullptr),
        fNumberFormat(nullptr),
        fPluralRules(nullptr),
        fStyle(styl),
        fContext(capitalizationContext),
        fOptBreakIterator(nullptr),
        fLocale(locale) {
    if (U_FAILURE(status)) {
        return;
    }
    if ((capitalizationContext >> 8) != UDISPCTX_TYPE_CAPITALIZATION) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if (capitalizationContext == UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE) {
        BreakIterator *bi = BreakIterator::createSentenceInstance(locale, status);
        if (U_FAILURE(status)) {
            return;
        }
        init(nfToAdopt, bi, status);
    } else {
        init(nfToAdopt, nullptr, status);
    }
}

RelativeDateTimeFormatter::RelativeDateTimeFormatter(
        const RelativeDateTimeFormatter& other)
        : UObject(other),
          fCache(other.fCache),
          fNumberFormat(other.fNumberFormat),
          fPluralRules(other.fPluralRules),
          fStyle(other.fStyle),
          fContext(other.fContext),
          fOptBreakIterator(other.fOptBreakIterator),
          fLocale(other.fLocale) {
    fCache->addRef();
    fNumberFormat->addRef();
    fPluralRules->addRef();
    if (fOptBreakIterator != nullptr) {
      fOptBreakIterator->addRef();
    }
}

RelativeDateTimeFormatter& RelativeDateTimeFormatter::operator=(
        const RelativeDateTimeFormatter& other) {
    if (this != &other) {
        SharedObject::copyPtr(other.fCache, fCache);
        SharedObject::copyPtr(other.fNumberFormat, fNumberFormat);
        SharedObject::copyPtr(other.fPluralRules, fPluralRules);
        SharedObject::copyPtr(other.fOptBreakIterator, fOptBreakIterator);
        fStyle = other.fStyle;
        fContext = other.fContext;
        fLocale = other.fLocale;
    }
    return *this;
}

RelativeDateTimeFormatter::~RelativeDateTimeFormatter() {
    if (fCache != nullptr) {
        fCache->removeRef();
    }
    if (fNumberFormat != nullptr) {
        fNumberFormat->removeRef();
    }
    if (fPluralRules != nullptr) {
        fPluralRules->removeRef();
    }
    if (fOptBreakIterator != nullptr) {
        fOptBreakIterator->removeRef();
    }
}

const NumberFormat& RelativeDateTimeFormatter::getNumberFormat() const {
    return **fNumberFormat;
}

UDisplayContext RelativeDateTimeFormatter::getCapitalizationContext() const {
    return fContext;
}

UDateRelativeDateTimeFormatterStyle RelativeDateTimeFormatter::getFormatStyle() const {
    return fStyle;
}


// To reduce boilerplate code, we use a helper function that forwards variadic
// arguments to the formatImpl function.

template<typename F, typename... Args>
UnicodeString& RelativeDateTimeFormatter::doFormat(
        F callback,
        UnicodeString& appendTo,
        UErrorCode& status,
        Args... args) const {
    FormattedRelativeDateTimeData output;
    (this->*callback)(std::forward<Args>(args)..., output, status);
    if (U_FAILURE(status)) {
        return appendTo;
    }
    UnicodeString result = output.getStringRef().toUnicodeString();
    return appendTo.append(adjustForContext(result));
}

template<typename F, typename... Args>
FormattedRelativeDateTime RelativeDateTimeFormatter::doFormatToValue(
        F callback,
        UErrorCode& status,
        Args... args) const {
    if (!checkNoAdjustForContext(status)) {
        return FormattedRelativeDateTime(status);
    }
    LocalPointer<FormattedRelativeDateTimeData> output(
        new FormattedRelativeDateTimeData(), status);
    if (U_FAILURE(status)) {
        return FormattedRelativeDateTime(status);
    }
    (this->*callback)(std::forward<Args>(args)..., *output, status);
    output->getStringRef().writeTerminator(status);
    return FormattedRelativeDateTime(output.orphan());
}

UnicodeString& RelativeDateTimeFormatter::format(
        double quantity,
        UDateDirection direction,
        UDateRelativeUnit unit,
        UnicodeString& appendTo,
        UErrorCode& status) const {
    return doFormat(
        &RelativeDateTimeFormatter::formatImpl,
        appendTo,
        status,
        quantity,
        direction,
        unit);
}

FormattedRelativeDateTime RelativeDateTimeFormatter::formatToValue(
        double quantity,
        UDateDirection direction,
        UDateRelativeUnit unit,
        UErrorCode& status) const {
    return doFormatToValue(
        &RelativeDateTimeFormatter::formatImpl,
        status,
        quantity,
        direction,
        unit);
}

void RelativeDateTimeFormatter::formatImpl(
        double quantity,
        UDateDirection direction,
        UDateRelativeUnit unit,
        FormattedRelativeDateTimeData& output,
        UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    if (direction != UDAT_DIRECTION_LAST && direction != UDAT_DIRECTION_NEXT) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    int32_t bFuture = direction == UDAT_DIRECTION_NEXT ? 1 : 0;

    StandardPlural::Form pluralForm;
    QuantityFormatter::formatAndSelect(
        quantity,
        **fNumberFormat,
        **fPluralRules,
        output.getStringRef(),
        pluralForm,
        status);
    if (U_FAILURE(status)) {
        return;
    }

    const SimpleFormatter* formatter =
        fCache->getRelativeUnitFormatter(fStyle, unit, bFuture, pluralForm);
    if (formatter == nullptr) {
        // TODO: WARN - look at quantity formatter's action with an error.
        status = U_INVALID_FORMAT_ERROR;
        return;
    }

    number::impl::SimpleModifier modifier(*formatter, kRDTLiteralField, false);
    modifier.formatAsPrefixSuffix(
        output.getStringRef(), 0, output.getStringRef().length(), status);
}

UnicodeString& RelativeDateTimeFormatter::formatNumeric(
        double offset,
        URelativeDateTimeUnit unit,
        UnicodeString& appendTo,
        UErrorCode& status) const {
    return doFormat(
        &RelativeDateTimeFormatter::formatNumericImpl,
        appendTo,
        status,
        offset,
        unit);
}

FormattedRelativeDateTime RelativeDateTimeFormatter::formatNumericToValue(
        double offset,
        URelativeDateTimeUnit unit,
        UErrorCode& status) const {
    return doFormatToValue(
        &RelativeDateTimeFormatter::formatNumericImpl,
        status,
        offset,
        unit);
}

void RelativeDateTimeFormatter::formatNumericImpl(
        double offset,
        URelativeDateTimeUnit unit,
        FormattedRelativeDateTimeData& output,
        UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    UDateDirection direction = UDAT_DIRECTION_NEXT;
    if (std::signbit(offset)) { // needed to handle -0.0
        direction = UDAT_DIRECTION_LAST;
        offset = -offset;
    }
    if (direction != UDAT_DIRECTION_LAST && direction != UDAT_DIRECTION_NEXT) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    int32_t bFuture = direction == UDAT_DIRECTION_NEXT ? 1 : 0;

    StandardPlural::Form pluralForm;
    QuantityFormatter::formatAndSelect(
        offset,
        **fNumberFormat,
        **fPluralRules,
        output.getStringRef(),
        pluralForm,
        status);
    if (U_FAILURE(status)) {
        return;
    }

    const SimpleFormatter* formatter =
        fCache->getRelativeDateTimeUnitFormatter(fStyle, unit, bFuture, pluralForm);
    if (formatter == nullptr) {
        // TODO: WARN - look at quantity formatter's action with an error.
        status = U_INVALID_FORMAT_ERROR;
        return;
    }

    number::impl::SimpleModifier modifier(*formatter, kRDTLiteralField, false);
    modifier.formatAsPrefixSuffix(
        output.getStringRef(), 0, output.getStringRef().length(), status);
}

UnicodeString& RelativeDateTimeFormatter::format(
        UDateDirection direction,
        UDateAbsoluteUnit unit,
        UnicodeString& appendTo,
        UErrorCode& status) const {
    return doFormat(
        &RelativeDateTimeFormatter::formatAbsoluteImpl,
        appendTo,
        status,
        direction,
        unit);
}

FormattedRelativeDateTime RelativeDateTimeFormatter::formatToValue(
        UDateDirection direction,
        UDateAbsoluteUnit unit,
        UErrorCode& status) const {
    return doFormatToValue(
        &RelativeDateTimeFormatter::formatAbsoluteImpl,
        status,
        direction,
        unit);
}

void RelativeDateTimeFormatter::formatAbsoluteImpl(
        UDateDirection direction,
        UDateAbsoluteUnit unit,
        FormattedRelativeDateTimeData& output,
        UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    if (unit == UDAT_ABSOLUTE_NOW && direction != UDAT_DIRECTION_PLAIN) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }

    // Get string using fallback.
    output.getStringRef().append(
        fCache->getAbsoluteUnitString(fStyle, unit, direction),
        kRDTLiteralField,
        status);
}

UnicodeString& RelativeDateTimeFormatter::format(
        double offset,
        URelativeDateTimeUnit unit,
        UnicodeString& appendTo,
        UErrorCode& status) const {
    return doFormat(
        &RelativeDateTimeFormatter::formatRelativeImpl,
        appendTo,
        status,
        offset,
        unit);
}

FormattedRelativeDateTime RelativeDateTimeFormatter::formatToValue(
        double offset,
        URelativeDateTimeUnit unit,
        UErrorCode& status) const {
    return doFormatToValue(
        &RelativeDateTimeFormatter::formatRelativeImpl,
        status,
        offset,
        unit);
}

void RelativeDateTimeFormatter::formatRelativeImpl(
        double offset,
        URelativeDateTimeUnit unit,
        FormattedRelativeDateTimeData& output,
        UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    // TODO:
    // The full implementation of this depends on CLDR data that is not yet available,
    // see: http://unicode.org/cldr/trac/ticket/9165 Add more relative field data.
    // In the meantime do a quick bring-up by calling the old format method; this
    // leaves some holes (even for data that is currently available, such as quarter).
    // When the new CLDR data is available, update the data storage accordingly,
    // rewrite this to use it directly, and rewrite the old format method to call this
    // new one; that is covered by http://bugs.icu-project.org/trac/ticket/12171.
    UDateDirection direction = UDAT_DIRECTION_COUNT;
    if (offset > -2.1 && offset < 2.1) {
        // Allow a 1% epsilon, so offsets in -1.01..-0.99 map to LAST
        double offsetx100 = offset * 100.0;
        int32_t intoffset = (offsetx100 < 0)? (int32_t)(offsetx100-0.5) : (int32_t)(offsetx100+0.5);
        switch (intoffset) {
            case -200/*-2*/: direction = UDAT_DIRECTION_LAST_2; break;
            case -100/*-1*/: direction = UDAT_DIRECTION_LAST; break;
            case    0/* 0*/: direction = UDAT_DIRECTION_THIS; break;
            case  100/* 1*/: direction = UDAT_DIRECTION_NEXT; break;
            case  200/* 2*/: direction = UDAT_DIRECTION_NEXT_2; break;
            default: break;
    	}
    }
    UDateAbsoluteUnit absunit = UDAT_ABSOLUTE_UNIT_COUNT;
    switch (unit) {
        case UDAT_REL_UNIT_YEAR:    absunit = UDAT_ABSOLUTE_YEAR; break;
        case UDAT_REL_UNIT_QUARTER: absunit = UDAT_ABSOLUTE_QUARTER; break;
        case UDAT_REL_UNIT_MONTH:   absunit = UDAT_ABSOLUTE_MONTH; break;
        case UDAT_REL_UNIT_WEEK:    absunit = UDAT_ABSOLUTE_WEEK; break;
        case UDAT_REL_UNIT_DAY:     absunit = UDAT_ABSOLUTE_DAY; break;
        case UDAT_REL_UNIT_SECOND:
            if (direction == UDAT_DIRECTION_THIS) {
                absunit = UDAT_ABSOLUTE_NOW;
                direction = UDAT_DIRECTION_PLAIN;
            }
            break;
        case UDAT_REL_UNIT_SUNDAY:  absunit = UDAT_ABSOLUTE_SUNDAY; break;
        case UDAT_REL_UNIT_MONDAY:  absunit = UDAT_ABSOLUTE_MONDAY; break;
        case UDAT_REL_UNIT_TUESDAY:  absunit = UDAT_ABSOLUTE_TUESDAY; break;
        case UDAT_REL_UNIT_WEDNESDAY:  absunit = UDAT_ABSOLUTE_WEDNESDAY; break;
        case UDAT_REL_UNIT_THURSDAY:  absunit = UDAT_ABSOLUTE_THURSDAY; break;
        case UDAT_REL_UNIT_FRIDAY:  absunit = UDAT_ABSOLUTE_FRIDAY; break;
        case UDAT_REL_UNIT_SATURDAY:  absunit = UDAT_ABSOLUTE_SATURDAY; break;
        case UDAT_REL_UNIT_HOUR:  absunit = UDAT_ABSOLUTE_HOUR; break;
        case UDAT_REL_UNIT_MINUTE:  absunit = UDAT_ABSOLUTE_MINUTE; break;
        default: break;
    }
    if (direction != UDAT_DIRECTION_COUNT && absunit != UDAT_ABSOLUTE_UNIT_COUNT) {
        formatAbsoluteImpl(direction, absunit, output, status);
        if (output.getStringRef().length() != 0) {
            return;
        }
    }
    // otherwise fallback to formatNumeric
    formatNumericImpl(offset, unit, output, status);
}

UnicodeString& RelativeDateTimeFormatter::combineDateAndTime(
        const UnicodeString& relativeDateString, const UnicodeString& timeString,
        UnicodeString& appendTo, UErrorCode& status) const {
    return fCache->getCombinedDateAndTime()->format(
            timeString, relativeDateString, appendTo, status);
}

UnicodeString& RelativeDateTimeFormatter::adjustForContext(UnicodeString &str) const {
    if (fOptBreakIterator == nullptr
        || str.length() == 0 || !u_islower(str.char32At(0))) {
        return str;
    }

    // Must guarantee that one thread at a time accesses the shared break
    // iterator.
    static UMutex gBrkIterMutex;
    Mutex lock(&gBrkIterMutex);
    str.toTitle(
            fOptBreakIterator->get(),
            fLocale,
            U_TITLECASE_NO_LOWERCASE | U_TITLECASE_NO_BREAK_ADJUSTMENT);
    return str;
}

UBool RelativeDateTimeFormatter::checkNoAdjustForContext(UErrorCode& status) const {
    // This is unsupported because it's hard to keep fields in sync with title
    // casing. The code could be written and tested if there is demand.
    if (fOptBreakIterator != nullptr) {
        status = U_UNSUPPORTED_ERROR;
        return FALSE;
    }
    return TRUE;
}

void RelativeDateTimeFormatter::init(
        NumberFormat *nfToAdopt,
        BreakIterator *biToAdopt,
        UErrorCode &status) {
    LocalPointer<NumberFormat> nf(nfToAdopt);
    LocalPointer<BreakIterator> bi(biToAdopt);
    UnifiedCache::getByLocale(fLocale, fCache, status);
    if (U_FAILURE(status)) {
        return;
    }
    const SharedPluralRules *pr = PluralRules::createSharedInstance(
            fLocale, UPLURAL_TYPE_CARDINAL, status);
    if (U_FAILURE(status)) {
        return;
    }
    SharedObject::copyPtr(pr, fPluralRules);
    pr->removeRef();
    if (nf.isNull()) {
       const SharedNumberFormat *shared = NumberFormat::createSharedInstance(
               fLocale, UNUM_DECIMAL, status);
        if (U_FAILURE(status)) {
            return;
        }
        SharedObject::copyPtr(shared, fNumberFormat);
        shared->removeRef();
    } else {
        SharedNumberFormat *shared = new SharedNumberFormat(nf.getAlias());
        if (shared == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        nf.orphan();
        SharedObject::copyPtr(shared, fNumberFormat);
    }
    if (bi.isNull()) {
        SharedObject::clearPtr(fOptBreakIterator);
    } else {
        SharedBreakIterator *shared = new SharedBreakIterator(bi.getAlias());
        if (shared == nullptr) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        bi.orphan();
        SharedObject::copyPtr(shared, fOptBreakIterator);
    }
}

U_NAMESPACE_END

// Plain C API

U_NAMESPACE_USE


// Magic number: "FRDT" (FormattedRelativeDateTime) in ASCII
UPRV_FORMATTED_VALUE_CAPI_AUTO_IMPL(
    FormattedRelativeDateTime,
    UFormattedRelativeDateTime,
    UFormattedRelativeDateTimeImpl,
    UFormattedRelativeDateTimeApiHelper,
    ureldatefmt,
    0x46524454)


U_CAPI URelativeDateTimeFormatter* U_EXPORT2
ureldatefmt_open( const char*          locale,
                  UNumberFormat*       nfToAdopt,
                  UDateRelativeDateTimeFormatterStyle width,
                  UDisplayContext      capitalizationContext,
                  UErrorCode*          status )
{
    if (U_FAILURE(*status)) {
        return nullptr;
    }
    LocalPointer<RelativeDateTimeFormatter> formatter(new RelativeDateTimeFormatter(Locale(locale),
                                                              (NumberFormat*)nfToAdopt, width,
                                                              capitalizationContext, *status), *status);
    if (U_FAILURE(*status)) {
        return nullptr;
    }
    return (URelativeDateTimeFormatter*)formatter.orphan();
}

U_CAPI void U_EXPORT2
ureldatefmt_close(URelativeDateTimeFormatter *reldatefmt)
{
    delete (RelativeDateTimeFormatter*)reldatefmt;
}

U_CAPI int32_t U_EXPORT2
ureldatefmt_formatNumeric( const URelativeDateTimeFormatter* reldatefmt,
                    double                offset,
                    URelativeDateTimeUnit unit,
                    UChar*                result,
                    int32_t               resultCapacity,
                    UErrorCode*           status)
{
    if (U_FAILURE(*status)) {
        return 0;
    }
    if (result == nullptr ? resultCapacity != 0 : resultCapacity < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString res;
    if (result != nullptr) {
        // nullptr destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer (copied from udat_format)
        res.setTo(result, 0, resultCapacity);
    }
    ((RelativeDateTimeFormatter*)reldatefmt)->formatNumeric(offset, unit, res, *status);
    if (U_FAILURE(*status)) {
        return 0;
    }
    return res.extract(result, resultCapacity, *status);
}

U_STABLE void U_EXPORT2
ureldatefmt_formatNumericToResult(
        const URelativeDateTimeFormatter* reldatefmt,
        double                            offset,
        URelativeDateTimeUnit             unit,
        UFormattedRelativeDateTime*       result,
        UErrorCode*                       status) {
    if (U_FAILURE(*status)) {
        return;
    }
    auto* fmt = reinterpret_cast<const RelativeDateTimeFormatter*>(reldatefmt);
    auto* resultImpl = UFormattedRelativeDateTimeApiHelper::validate(result, *status);
    resultImpl->fImpl = fmt->formatNumericToValue(offset, unit, *status);
}

U_CAPI int32_t U_EXPORT2
ureldatefmt_format( const URelativeDateTimeFormatter* reldatefmt,
                    double                offset,
                    URelativeDateTimeUnit unit,
                    UChar*                result,
                    int32_t               resultCapacity,
                    UErrorCode*           status)
{
    if (U_FAILURE(*status)) {
        return 0;
    }
    if (result == nullptr ? resultCapacity != 0 : resultCapacity < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString res;
    if (result != nullptr) {
        // nullptr destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer (copied from udat_format)
        res.setTo(result, 0, resultCapacity);
    }
    ((RelativeDateTimeFormatter*)reldatefmt)->format(offset, unit, res, *status);
    if (U_FAILURE(*status)) {
        return 0;
    }
    return res.extract(result, resultCapacity, *status);
}

U_DRAFT void U_EXPORT2
ureldatefmt_formatToResult(
        const URelativeDateTimeFormatter* reldatefmt,
        double                            offset,
        URelativeDateTimeUnit             unit,
        UFormattedRelativeDateTime*       result,
        UErrorCode*                       status) {
    if (U_FAILURE(*status)) {
        return;
    }
    auto* fmt = reinterpret_cast<const RelativeDateTimeFormatter*>(reldatefmt);
    auto* resultImpl = UFormattedRelativeDateTimeApiHelper::validate(result, *status);
    resultImpl->fImpl = fmt->formatToValue(offset, unit, *status);
}

U_CAPI int32_t U_EXPORT2
ureldatefmt_combineDateAndTime( const URelativeDateTimeFormatter* reldatefmt,
                    const UChar *     relativeDateString,
                    int32_t           relativeDateStringLen,
                    const UChar *     timeString,
                    int32_t           timeStringLen,
                    UChar*            result,
                    int32_t           resultCapacity,
                    UErrorCode*       status )
{
    if (U_FAILURE(*status)) {
        return 0;
    }
    if (result == nullptr ? resultCapacity != 0 : resultCapacity < 0 ||
            (relativeDateString == nullptr ? relativeDateStringLen != 0 : relativeDateStringLen < -1) ||
            (timeString == nullptr ? timeStringLen != 0 : timeStringLen < -1)) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString relDateStr((UBool)(relativeDateStringLen == -1), relativeDateString, relativeDateStringLen);
    UnicodeString timeStr((UBool)(timeStringLen == -1), timeString, timeStringLen);
    UnicodeString res(result, 0, resultCapacity);
    ((RelativeDateTimeFormatter*)reldatefmt)->combineDateAndTime(relDateStr, timeStr, res, *status);
    if (U_FAILURE(*status)) {
        return 0;
    }
    return res.extract(result, resultCapacity, *status);
}

#endif /* !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING

#include <stdlib.h>







// #include "unicode/brkiter.h"





U_NAMESPACE_BEGIN


/**
 * An array of URelativeString structs is used to store the resource data loaded out of the bundle.
 */
struct URelativeString {
    int32_t offset;         /** offset of this item, such as, the relative date **/
    int32_t len;            /** length of the string **/
    const UChar* string;    /** string, or NULL if not set **/
};

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(RelativeDateFormat)

RelativeDateFormat::RelativeDateFormat(const RelativeDateFormat& other) :
 DateFormat(other), fDateTimeFormatter(NULL), fDatePattern(other.fDatePattern),
 fTimePattern(other.fTimePattern), fCombinedFormat(NULL),
 fDateStyle(other.fDateStyle), fLocale(other.fLocale),
 fDatesLen(other.fDatesLen), fDates(NULL),
 fCombinedHasDateAtStart(other.fCombinedHasDateAtStart),
 fCapitalizationInfoSet(other.fCapitalizationInfoSet),
 fCapitalizationOfRelativeUnitsForUIListMenu(other.fCapitalizationOfRelativeUnitsForUIListMenu),
 fCapitalizationOfRelativeUnitsForStandAlone(other.fCapitalizationOfRelativeUnitsForStandAlone),
 fCapitalizationBrkIter(NULL)
{
    if(other.fDateTimeFormatter != NULL) {
        fDateTimeFormatter = other.fDateTimeFormatter->clone();
    }
    if(other.fCombinedFormat != NULL) {
        fCombinedFormat = new SimpleFormatter(*other.fCombinedFormat);
    }
    if (fDatesLen > 0) {
        fDates = (URelativeString*) uprv_malloc(sizeof(fDates[0])*(size_t)fDatesLen);
        uprv_memcpy(fDates, other.fDates, sizeof(fDates[0])*(size_t)fDatesLen);
    }
#if !UCONFIG_NO_BREAK_ITERATION
    if (other.fCapitalizationBrkIter != NULL) {
        fCapitalizationBrkIter = (other.fCapitalizationBrkIter)->clone();
    }
#endif
}

RelativeDateFormat::RelativeDateFormat( UDateFormatStyle timeStyle, UDateFormatStyle dateStyle,
                                        const Locale& locale, UErrorCode& status) :
 DateFormat(), fDateTimeFormatter(NULL), fDatePattern(), fTimePattern(), fCombinedFormat(NULL),
 fDateStyle(dateStyle), fLocale(locale), fDatesLen(0), fDates(NULL),
 fCombinedHasDateAtStart(FALSE), fCapitalizationInfoSet(FALSE),
 fCapitalizationOfRelativeUnitsForUIListMenu(FALSE), fCapitalizationOfRelativeUnitsForStandAlone(FALSE),
 fCapitalizationBrkIter(NULL)
{
    if(U_FAILURE(status) ) {
        return;
    }

    if (timeStyle < UDAT_NONE || timeStyle > UDAT_SHORT) {
        // don't support other time styles (e.g. relative styles), for now
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    UDateFormatStyle baseDateStyle = (dateStyle > UDAT_SHORT)? (UDateFormatStyle)(dateStyle & ~UDAT_RELATIVE): dateStyle;
    DateFormat * df;
    // Get fDateTimeFormatter from either date or time style (does not matter, we will override the pattern).
    // We do need to get separate patterns for the date & time styles.
    if (baseDateStyle != UDAT_NONE) {
        df = createDateInstance((EStyle)baseDateStyle, locale);
        fDateTimeFormatter=dynamic_cast<SimpleDateFormat *>(df);
        if (fDateTimeFormatter == NULL) {
            status = U_UNSUPPORTED_ERROR;
             return;
        }
        fDateTimeFormatter->toPattern(fDatePattern);
        if (timeStyle != UDAT_NONE) {
            df = createTimeInstance((EStyle)timeStyle, locale);
            SimpleDateFormat *sdf = dynamic_cast<SimpleDateFormat *>(df);
            if (sdf != NULL) {
                sdf->toPattern(fTimePattern);
                delete sdf;
            }
        }
    } else {
        // does not matter whether timeStyle is UDAT_NONE, we need something for fDateTimeFormatter
        df = createTimeInstance((EStyle)timeStyle, locale);
        fDateTimeFormatter=dynamic_cast<SimpleDateFormat *>(df);
        if (fDateTimeFormatter == NULL) {
            status = U_UNSUPPORTED_ERROR;
            delete df;
            return;
        }
        fDateTimeFormatter->toPattern(fTimePattern);
    }

    // Initialize the parent fCalendar, so that parse() works correctly.
    initializeCalendar(NULL, locale, status);
    loadDates(status);
}

RelativeDateFormat::~RelativeDateFormat() {
    delete fDateTimeFormatter;
    delete fCombinedFormat;
    uprv_free(fDates);
#if !UCONFIG_NO_BREAK_ITERATION
    delete fCapitalizationBrkIter;
#endif
}


RelativeDateFormat* RelativeDateFormat::clone() const {
    return new RelativeDateFormat(*this);
}

UBool RelativeDateFormat::operator==(const Format& other) const {
    if(DateFormat::operator==(other)) {
        // The DateFormat::operator== check for fCapitalizationContext equality above
        //   is sufficient to check equality of all derived context-related data.
        // DateFormat::operator== guarantees following cast is safe
        RelativeDateFormat* that = (RelativeDateFormat*)&other;
        return (fDateStyle==that->fDateStyle   &&
                fDatePattern==that->fDatePattern   &&
                fTimePattern==that->fTimePattern   &&
                fLocale==that->fLocale );
    }
    return FALSE;
}

static const UChar APOSTROPHE = (UChar)0x0027;

UnicodeString& RelativeDateFormat::format(  Calendar& cal,
                                UnicodeString& appendTo,
                                FieldPosition& pos) const {

    UErrorCode status = U_ZERO_ERROR;
    UnicodeString relativeDayString;
    UDisplayContext capitalizationContext = getContext(UDISPCTX_TYPE_CAPITALIZATION, status);

    // calculate the difference, in days, between 'cal' and now.
    int dayDiff = dayDifference(cal, status);

    // look up string
    int32_t len = 0;
    const UChar *theString = getStringForDay(dayDiff, len, status);
    if(U_SUCCESS(status) && (theString!=NULL)) {
        // found a relative string
        relativeDayString.setTo(theString, len);
    }

    if ( relativeDayString.length() > 0 && !fDatePattern.isEmpty() &&
         (fTimePattern.isEmpty() || fCombinedFormat == NULL || fCombinedHasDateAtStart)) {
#if !UCONFIG_NO_BREAK_ITERATION
        // capitalize relativeDayString according to context for relative, set formatter no context
        if ( u_islower(relativeDayString.char32At(0)) && fCapitalizationBrkIter!= NULL &&
             ( capitalizationContext==UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE ||
               (capitalizationContext==UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU && fCapitalizationOfRelativeUnitsForUIListMenu) ||
               (capitalizationContext==UDISPCTX_CAPITALIZATION_FOR_STANDALONE && fCapitalizationOfRelativeUnitsForStandAlone) ) ) {
            // titlecase first word of relativeDayString
            relativeDayString.toTitle(fCapitalizationBrkIter, fLocale, U_TITLECASE_NO_LOWERCASE | U_TITLECASE_NO_BREAK_ADJUSTMENT);
        }
#endif
        fDateTimeFormatter->setContext(UDISPCTX_CAPITALIZATION_NONE, status);
    } else {
        // set our context for the formatter
        fDateTimeFormatter->setContext(capitalizationContext, status);
    }

    if (fDatePattern.isEmpty()) {
        fDateTimeFormatter->applyPattern(fTimePattern);
        fDateTimeFormatter->format(cal,appendTo,pos);
    } else if (fTimePattern.isEmpty() || fCombinedFormat == NULL) {
        if (relativeDayString.length() > 0) {
            appendTo.append(relativeDayString);
        } else {
            fDateTimeFormatter->applyPattern(fDatePattern);
            fDateTimeFormatter->format(cal,appendTo,pos);
        }
    } else {
        UnicodeString datePattern;
        if (relativeDayString.length() > 0) {
            // Need to quote the relativeDayString to make it a legal date pattern
            relativeDayString.findAndReplace(UNICODE_STRING("'", 1), UNICODE_STRING("''", 2)); // double any existing APOSTROPHE
            relativeDayString.insert(0, APOSTROPHE); // add APOSTROPHE at beginning...
            relativeDayString.append(APOSTROPHE); // and at end
            datePattern.setTo(relativeDayString);
        } else {
            datePattern.setTo(fDatePattern);
        }
        UnicodeString combinedPattern;
        fCombinedFormat->format(fTimePattern, datePattern, combinedPattern, status);
        fDateTimeFormatter->applyPattern(combinedPattern);
        fDateTimeFormatter->format(cal,appendTo,pos);
    }

    return appendTo;
}



UnicodeString&
RelativeDateFormat::format(const Formattable& obj,
                         UnicodeString& appendTo,
                         FieldPosition& pos,
                         UErrorCode& status) const
{
    // this is just here to get around the hiding problem
    // (the previous format() override would hide the version of
    // format() on DateFormat that this function correspond to, so we
    // have to redefine it here)
    return DateFormat::format(obj, appendTo, pos, status);
}


void RelativeDateFormat::parse( const UnicodeString& text,
                    Calendar& cal,
                    ParsePosition& pos) const {

    int32_t startIndex = pos.getIndex();
    if (fDatePattern.isEmpty()) {
        // no date pattern, try parsing as time
        fDateTimeFormatter->applyPattern(fTimePattern);
        fDateTimeFormatter->parse(text,cal,pos);
    } else if (fTimePattern.isEmpty() || fCombinedFormat == NULL) {
        // no time pattern or way to combine, try parsing as date
        // first check whether text matches a relativeDayString
        UBool matchedRelative = FALSE;
        for (int n=0; n < fDatesLen && !matchedRelative; n++) {
            if (fDates[n].string != NULL &&
                    text.compare(startIndex, fDates[n].len, fDates[n].string) == 0) {
                // it matched, handle the relative day string
                UErrorCode status = U_ZERO_ERROR;
                matchedRelative = TRUE;

                // Set the calendar to now+offset
                cal.setTime(Calendar::getNow(),status);
                cal.add(UCAL_DATE,fDates[n].offset, status);

                if(U_FAILURE(status)) {
                    // failure in setting calendar field, set offset to beginning of rel day string
                    pos.setErrorIndex(startIndex);
                } else {
                    pos.setIndex(startIndex + fDates[n].len);
                }
            }
        }
        if (!matchedRelative) {
            // just parse as normal date
            fDateTimeFormatter->applyPattern(fDatePattern);
            fDateTimeFormatter->parse(text,cal,pos);
        }
    } else {
        // Here we replace any relativeDayString in text with the equivalent date
        // formatted per fDatePattern, then parse text normally using the combined pattern.
        UnicodeString modifiedText(text);
        FieldPosition fPos;
        int32_t dateStart = 0, origDateLen = 0, modDateLen = 0;
        UErrorCode status = U_ZERO_ERROR;
        for (int n=0; n < fDatesLen; n++) {
            int32_t relativeStringOffset;
            if (fDates[n].string != NULL &&
                    (relativeStringOffset = modifiedText.indexOf(fDates[n].string, fDates[n].len, startIndex)) >= startIndex) {
                // it matched, replace the relative date with a real one for parsing
                UnicodeString dateString;
                Calendar * tempCal = cal.clone();

                // Set the calendar to now+offset
                tempCal->setTime(Calendar::getNow(),status);
                tempCal->add(UCAL_DATE,fDates[n].offset, status);
                if(U_FAILURE(status)) {
                    pos.setErrorIndex(startIndex);
                    delete tempCal;
                    return;
                }

                fDateTimeFormatter->applyPattern(fDatePattern);
                fDateTimeFormatter->format(*tempCal, dateString, fPos);
                dateStart = relativeStringOffset;
                origDateLen = fDates[n].len;
                modDateLen = dateString.length();
                modifiedText.replace(dateStart, origDateLen, dateString);
                delete tempCal;
                break;
            }
        }
        UnicodeString combinedPattern;
        fCombinedFormat->format(fTimePattern, fDatePattern, combinedPattern, status);
        fDateTimeFormatter->applyPattern(combinedPattern);
        fDateTimeFormatter->parse(modifiedText,cal,pos);

        // Adjust offsets
        UBool noError = (pos.getErrorIndex() < 0);
        int32_t offset = (noError)? pos.getIndex(): pos.getErrorIndex();
        if (offset >= dateStart + modDateLen) {
            // offset at or after the end of the replaced text,
            // correct by the difference between original and replacement
            offset -= (modDateLen - origDateLen);
        } else if (offset >= dateStart) {
            // offset in the replaced text, set it to the beginning of that text
            // (i.e. the beginning of the relative day string)
            offset = dateStart;
        }
        if (noError) {
            pos.setIndex(offset);
        } else {
            pos.setErrorIndex(offset);
        }
    }
}

UDate
RelativeDateFormat::parse( const UnicodeString& text,
                         ParsePosition& pos) const {
    // redefined here because the other parse() function hides this function's
    // cunterpart on DateFormat
    return DateFormat::parse(text, pos);
}

UDate
RelativeDateFormat::parse(const UnicodeString& text, UErrorCode& status) const
{
    // redefined here because the other parse() function hides this function's
    // counterpart on DateFormat
    return DateFormat::parse(text, status);
}


const UChar *RelativeDateFormat::getStringForDay(int32_t day, int32_t &len, UErrorCode &status) const {
    if(U_FAILURE(status)) {
        return NULL;
    }

    // Is it inside the resource bundle's range?
    int n = day + UDAT_DIRECTION_THIS;
    if (n >= 0 && n < fDatesLen) {
        if (fDates[n].offset == day && fDates[n].string != NULL) {
            len = fDates[n].len;
            return fDates[n].string;
        }
    }
    return NULL;  // not found.
}

UnicodeString&
RelativeDateFormat::toPattern(UnicodeString& result, UErrorCode& status) const
{
    if (!U_FAILURE(status)) {
        result.remove();
        if (fDatePattern.isEmpty()) {
            result.setTo(fTimePattern);
        } else if (fTimePattern.isEmpty() || fCombinedFormat == NULL) {
            result.setTo(fDatePattern);
        } else {
            fCombinedFormat->format(fTimePattern, fDatePattern, result, status);
        }
    }
    return result;
}

UnicodeString&
RelativeDateFormat::toPatternDate(UnicodeString& result, UErrorCode& status) const
{
    if (!U_FAILURE(status)) {
        result.remove();
        result.setTo(fDatePattern);
    }
    return result;
}

UnicodeString&
RelativeDateFormat::toPatternTime(UnicodeString& result, UErrorCode& status) const
{
    if (!U_FAILURE(status)) {
        result.remove();
        result.setTo(fTimePattern);
    }
    return result;
}

void
RelativeDateFormat::applyPatterns(const UnicodeString& datePattern, const UnicodeString& timePattern, UErrorCode &status)
{
    if (!U_FAILURE(status)) {
        fDatePattern.setTo(datePattern);
        fTimePattern.setTo(timePattern);
    }
}

const DateFormatSymbols*
RelativeDateFormat::getDateFormatSymbols() const
{
    return fDateTimeFormatter->getDateFormatSymbols();
}

// override the DateFormat implementation in order to
// lazily initialize relevant items
void
RelativeDateFormat::setContext(UDisplayContext value, UErrorCode& status)
{
    DateFormat::setContext(value, status);
    if (U_SUCCESS(status)) {
        if (!fCapitalizationInfoSet &&
                (value==UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU || value==UDISPCTX_CAPITALIZATION_FOR_STANDALONE)) {
            initCapitalizationContextInfo(fLocale);
            fCapitalizationInfoSet = TRUE;
        }
#if !UCONFIG_NO_BREAK_ITERATION
        if ( fCapitalizationBrkIter == NULL && (value==UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE ||
                (value==UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU && fCapitalizationOfRelativeUnitsForUIListMenu) ||
                (value==UDISPCTX_CAPITALIZATION_FOR_STANDALONE && fCapitalizationOfRelativeUnitsForStandAlone)) ) {
            status = U_ZERO_ERROR;
            fCapitalizationBrkIter = BreakIterator::createSentenceInstance(fLocale, status);
            if (U_FAILURE(status)) {
                delete fCapitalizationBrkIter;
                fCapitalizationBrkIter = NULL;
            }
        }
#endif
    }
}

void
RelativeDateFormat::initCapitalizationContextInfo(const Locale& thelocale)
{
#if !UCONFIG_NO_BREAK_ITERATION
    const char * localeID = (thelocale != NULL)? thelocale.getBaseName(): NULL;
    UErrorCode status = U_ZERO_ERROR;
    LocalUResourceBundlePointer rb(ures_open(NULL, localeID, &status));
    ures_getByKeyWithFallback(rb.getAlias(),
                              "contextTransforms/relative",
                               rb.getAlias(), &status);
    if (U_SUCCESS(status) && rb != NULL) {
        int32_t len = 0;
        const int32_t * intVector = ures_getIntVector(rb.getAlias(),
                                                      &len, &status);
        if (U_SUCCESS(status) && intVector != NULL && len >= 2) {
            fCapitalizationOfRelativeUnitsForUIListMenu = static_cast<UBool>(intVector[0]);
            fCapitalizationOfRelativeUnitsForStandAlone = static_cast<UBool>(intVector[1]);
        }
    }
#endif
}

namespace {

/**
 * Sink for getting data from fields/day/relative data.
 * For loading relative day names, e.g., "yesterday", "today".
 */

struct RelDateFmtDataSink : public ResourceSink {
  URelativeString *fDatesPtr;
  int32_t fDatesLen;

  RelDateFmtDataSink(URelativeString* fDates, int32_t len) : fDatesPtr(fDates), fDatesLen(len) {
    for (int32_t i = 0; i < fDatesLen; ++i) {
      fDatesPtr[i].offset = 0;
      fDatesPtr[i].string = NULL;
      fDatesPtr[i].len = -1;
    }
  }

  virtual ~RelDateFmtDataSink();

  virtual void put(const char *key, ResourceValue &value,
                   UBool /*noFallback*/, UErrorCode &errorCode) {
      ResourceTable relDayTable = value.getTable(errorCode);
      int32_t n = 0;
      int32_t len = 0;
      for (int32_t i = 0; relDayTable.getKeyAndValue(i, key, value); ++i) {
        // Find the relative offset.
        int32_t offset = atoi(key);

        // Put in the proper spot, but don't override existing data.
        n = offset + UDAT_DIRECTION_THIS; // Converts to index in UDAT_R
        if (n < fDatesLen && fDatesPtr[n].string == NULL) {
          // Not found and n is an empty slot.
          fDatesPtr[n].offset = offset;
          fDatesPtr[n].string = value.getString(len, errorCode);
          fDatesPtr[n].len = len;
        }
      }
  }
};


// Virtual destructors must be defined out of line.
RelDateFmtDataSink::~RelDateFmtDataSink() {}

}  // Namespace


static const UChar patItem1[] = {0x7B,0x31,0x7D}; // "{1}"
static const int32_t patItem1Len = 3;

void RelativeDateFormat::loadDates(UErrorCode &status) {
    UResourceBundle *rb = ures_open(NULL, fLocale.getBaseName(), &status);
    LocalUResourceBundlePointer dateTimePatterns(
        ures_getByKeyWithFallback(rb,
                                  "calendar/gregorian/DateTimePatterns",
                                  (UResourceBundle*)NULL, &status));
    if(U_SUCCESS(status)) {
        int32_t patternsSize = ures_getSize(dateTimePatterns.getAlias());
        if (patternsSize > kDateTime) {
            int32_t resStrLen = 0;
            int32_t glueIndex = kDateTime;
            if (patternsSize >= (kDateTimeOffset + kShort + 1)) {
                int32_t offsetIncrement = (fDateStyle & ~kRelative); // Remove relative bit.
                if (offsetIncrement >= (int32_t)kFull &&
                    offsetIncrement <= (int32_t)kShortRelative) {
                    glueIndex = kDateTimeOffset + offsetIncrement;
                }
            }

            const UChar *resStr = ures_getStringByIndex(dateTimePatterns.getAlias(), glueIndex, &resStrLen, &status);
            if (U_SUCCESS(status) && resStrLen >= patItem1Len && u_strncmp(resStr,patItem1,patItem1Len)==0) {
                fCombinedHasDateAtStart = TRUE;
            }
            fCombinedFormat = new SimpleFormatter(UnicodeString(TRUE, resStr, resStrLen), 2, 2, status);
        }
    }

    // Data loading for relative names, e.g., "yesterday", "today", "tomorrow".
    fDatesLen = UDAT_DIRECTION_COUNT; // Maximum defined by data.
    fDates = (URelativeString*) uprv_malloc(sizeof(fDates[0])*fDatesLen);

    RelDateFmtDataSink sink(fDates, fDatesLen);
    ures_getAllItemsWithFallback(rb, "fields/day/relative", sink, status);

    ures_close(rb);

    if(U_FAILURE(status)) {
        fDatesLen=0;
        return;
    }
}

//----------------------------------------------------------------------

// this should to be in DateFormat, instead it was copied from SimpleDateFormat.

Calendar*
RelativeDateFormat::initializeCalendar(TimeZone* adoptZone, const Locale& locale, UErrorCode& status)
{
    if(!U_FAILURE(status)) {
        fCalendar = Calendar::createInstance(adoptZone?adoptZone:TimeZone::createDefault(), locale, status);
    }
    if (U_SUCCESS(status) && fCalendar == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return fCalendar;
}

int32_t RelativeDateFormat::dayDifference(Calendar &cal, UErrorCode &status) {
    if(U_FAILURE(status)) {
        return 0;
    }
    // TODO: Cache the nowCal to avoid heap allocs? Would be difficult, don't know the calendar type
    Calendar *nowCal = cal.clone();
    nowCal->setTime(Calendar::getNow(), status);

    // For the day difference, we are interested in the difference in the (modified) julian day number
    // which is midnight to midnight.  Using fieldDifference() is NOT correct here, because
    // 6pm Jan 4th  to 10am Jan 5th should be considered "tomorrow".
    int32_t dayDiff = cal.get(UCAL_JULIAN_DAY, status) - nowCal->get(UCAL_JULIAN_DAY, status);

    delete nowCal;
    return dayDiff;
}

U_NAMESPACE_END

#endif  /* !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1996-2015, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* rulebasedcollator.cpp
*
* (replaced the former tblcoll.cpp)
*
* created on: 2012feb14 with new and old collation code
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION

































// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* uitercollationiterator.h
*
* created on: 2012sep23 (from utf16collationiterator.h)
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION








U_NAMESPACE_BEGIN

/**
 * UCharIterator-based collation element and character iterator.
 * Handles normalized text inline, with length or NUL-terminated.
 * Unnormalized text is handled by a subclass.
 */
class U_I18N_API UIterCollationIterator : public CollationIterator {
public:
    UIterCollationIterator(const CollationData *d, UBool numeric, UCharIterator &ui)
            : CollationIterator(d, numeric), iter(ui) {}

    virtual ~UIterCollationIterator();

    virtual void resetToOffset(int32_t newOffset);

    virtual int32_t getOffset() const;

    virtual UChar32 nextCodePoint(UErrorCode &errorCode);

    virtual UChar32 previousCodePoint(UErrorCode &errorCode);

protected:
    virtual uint32_t handleNextCE32(UChar32 &c, UErrorCode &errorCode);

    virtual UChar handleGetTrailSurrogate();

    virtual void forwardNumCodePoints(int32_t num, UErrorCode &errorCode);

    virtual void backwardNumCodePoints(int32_t num, UErrorCode &errorCode);

    UCharIterator &iter;
};

/**
 * Incrementally checks the input text for FCD and normalizes where necessary.
 */
class U_I18N_API FCDUIterCollationIterator : public UIterCollationIterator {
public:
    FCDUIterCollationIterator(const CollationData *data, UBool numeric, UCharIterator &ui, int32_t startIndex)
            : UIterCollationIterator(data, numeric, ui),
              state(ITER_CHECK_FWD), start(startIndex),
              nfcImpl(data->nfcImpl) {}

    virtual ~FCDUIterCollationIterator();

    virtual void resetToOffset(int32_t newOffset);

    virtual int32_t getOffset() const;

    virtual UChar32 nextCodePoint(UErrorCode &errorCode);

    virtual UChar32 previousCodePoint(UErrorCode &errorCode);

protected:
    virtual uint32_t handleNextCE32(UChar32 &c, UErrorCode &errorCode);

    virtual UChar handleGetTrailSurrogate();

    virtual void forwardNumCodePoints(int32_t num, UErrorCode &errorCode);

    virtual void backwardNumCodePoints(int32_t num, UErrorCode &errorCode);

private:
    /**
     * Switches to forward checking if possible.
     */
    void switchToForward();

    /**
     * Extends the FCD text segment forward or normalizes around pos.
     * @return TRUE if success
     */
    UBool nextSegment(UErrorCode &errorCode);

    /**
     * Switches to backward checking.
     */
    void switchToBackward();

    /**
     * Extends the FCD text segment backward or normalizes around pos.
     * @return TRUE if success
     */
    UBool previousSegment(UErrorCode &errorCode);

    UBool normalize(const UnicodeString &s, UErrorCode &errorCode);

    enum State {
        /**
         * The input text [start..(iter index)[ passes the FCD check.
         * Moving forward checks incrementally.
         * pos & limit are undefined.
         */
        ITER_CHECK_FWD,
        /**
         * The input text [(iter index)..limit[ passes the FCD check.
         * Moving backward checks incrementally.
         * start & pos are undefined.
         */
        ITER_CHECK_BWD,
        /**
         * The input text [start..limit[ passes the FCD check.
         * pos tracks the current text index.
         */
        ITER_IN_FCD_SEGMENT,
        /**
         * The input text [start..limit[ failed the FCD check and was normalized.
         * pos tracks the current index in the normalized string.
         * The text iterator is at the limit index.
         */
        IN_NORM_ITER_AT_LIMIT,
        /**
         * The input text [start..limit[ failed the FCD check and was normalized.
         * pos tracks the current index in the normalized string.
         * The text iterator is at the start index.
         */
        IN_NORM_ITER_AT_START
    };

    State state;

    int32_t start;
    int32_t pos;
    int32_t limit;

    const Normalizer2Impl &nfcImpl;
    UnicodeString normalized;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION



// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* utf8collationiterator.h
*
* created on: 2012nov12 (from utf16collationiterator.h & uitercollationiterator.h)
* created by: Markus W. Scherer
*/




#if !UCONFIG_NO_COLLATION







U_NAMESPACE_BEGIN

/**
 * UTF-8 collation element and character iterator.
 * Handles normalized UTF-8 text inline, with length or NUL-terminated.
 * Unnormalized text is handled by a subclass.
 */
class U_I18N_API UTF8CollationIterator : public CollationIterator {
public:
    UTF8CollationIterator(const CollationData *d, UBool numeric,
                          const uint8_t *s, int32_t p, int32_t len)
            : CollationIterator(d, numeric),
              u8(s), pos(p), length(len) {}

    virtual ~UTF8CollationIterator();

    virtual void resetToOffset(int32_t newOffset);

    virtual int32_t getOffset() const;

    virtual UChar32 nextCodePoint(UErrorCode &errorCode);

    virtual UChar32 previousCodePoint(UErrorCode &errorCode);

protected:
    /**
     * For byte sequences that are illegal in UTF-8, an error value may be returned
     * together with a bogus code point. The caller will ignore that code point.
     *
     * Special values may be returned for surrogate code points, which are also illegal in UTF-8,
     * but the caller will treat them like U+FFFD because forbidSurrogateCodePoints() returns TRUE.
     *
     * Valid lead surrogates are returned from inside a normalized text segment,
     * where handleGetTrailSurrogate() will return the matching trail surrogate.
     */
    virtual uint32_t handleNextCE32(UChar32 &c, UErrorCode &errorCode);

    virtual UBool foundNULTerminator();

    virtual UBool forbidSurrogateCodePoints() const;

    virtual void forwardNumCodePoints(int32_t num, UErrorCode &errorCode);

    virtual void backwardNumCodePoints(int32_t num, UErrorCode &errorCode);

    const uint8_t *u8;
    int32_t pos;
    int32_t length;  // <0 for NUL-terminated strings
};

/**
 * Incrementally checks the input text for FCD and normalizes where necessary.
 */
class U_I18N_API FCDUTF8CollationIterator : public UTF8CollationIterator {
public:
    FCDUTF8CollationIterator(const CollationData *data, UBool numeric,
                             const uint8_t *s, int32_t p, int32_t len)
            : UTF8CollationIterator(data, numeric, s, p, len),
              state(CHECK_FWD), start(p),
              nfcImpl(data->nfcImpl) {}

    virtual ~FCDUTF8CollationIterator();

    virtual void resetToOffset(int32_t newOffset);

    virtual int32_t getOffset() const;

    virtual UChar32 nextCodePoint(UErrorCode &errorCode);

    virtual UChar32 previousCodePoint(UErrorCode &errorCode);

protected:
    virtual uint32_t handleNextCE32(UChar32 &c, UErrorCode &errorCode);

    virtual UChar handleGetTrailSurrogate();

    virtual UBool foundNULTerminator();

    virtual void forwardNumCodePoints(int32_t num, UErrorCode &errorCode);

    virtual void backwardNumCodePoints(int32_t num, UErrorCode &errorCode);

private:
    UBool nextHasLccc() const;
    UBool previousHasTccc() const;

    /**
     * Switches to forward checking if possible.
     */
    void switchToForward();

    /**
     * Extends the FCD text segment forward or normalizes around pos.
     * @return TRUE if success
     */
    UBool nextSegment(UErrorCode &errorCode);

    /**
     * Switches to backward checking.
     */
    void switchToBackward();

    /**
     * Extends the FCD text segment backward or normalizes around pos.
     * @return TRUE if success
     */
    UBool previousSegment(UErrorCode &errorCode);

    UBool normalize(const UnicodeString &s, UErrorCode &errorCode);

    enum State {
        /**
         * The input text [start..pos[ passes the FCD check.
         * Moving forward checks incrementally.
         * limit is undefined.
         */
        CHECK_FWD,
        /**
         * The input text [pos..limit[ passes the FCD check.
         * Moving backward checks incrementally.
         * start is undefined.
         */
        CHECK_BWD,
        /**
         * The input text [start..limit[ passes the FCD check.
         * pos tracks the current text index.
         */
        IN_FCD_SEGMENT,
        /**
         * The input text [start..limit[ failed the FCD check and was normalized.
         * pos tracks the current index in the normalized string.
         */
        IN_NORMALIZED
    };

    State state;

    int32_t start;
    int32_t limit;

    const Normalizer2Impl &nfcImpl;
    UnicodeString normalized;
};

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION



U_NAMESPACE_BEGIN

namespace {

class FixedSortKeyByteSink : public SortKeyByteSink {
public:
    FixedSortKeyByteSink(char *dest, int32_t destCapacity)
            : SortKeyByteSink(dest, destCapacity) {}
    virtual ~FixedSortKeyByteSink();

private:
    virtual void AppendBeyondCapacity(const char *bytes, int32_t n, int32_t length);
    virtual UBool Resize(int32_t appendCapacity, int32_t length);
};

FixedSortKeyByteSink::~FixedSortKeyByteSink() {}

void
FixedSortKeyByteSink::AppendBeyondCapacity(const char *bytes, int32_t /*n*/, int32_t length) {
    // buffer_ != NULL && bytes != NULL && n > 0 && appended_ > capacity_
    // Fill the buffer completely.
    int32_t available = capacity_ - length;
    if (available > 0) {
        uprv_memcpy(buffer_ + length, bytes, available);
    }
}

UBool
FixedSortKeyByteSink::Resize(int32_t /*appendCapacity*/, int32_t /*length*/) {
    return FALSE;
}

}  // namespace

// Not in an anonymous namespace, so that it can be a friend of CollationKey.
class CollationKeyByteSink : public SortKeyByteSink {
public:
    CollationKeyByteSink(CollationKey &key)
            : SortKeyByteSink(reinterpret_cast<char *>(key.getBytes()), key.getCapacity()),
              key_(key) {}
    virtual ~CollationKeyByteSink();

private:
    virtual void AppendBeyondCapacity(const char *bytes, int32_t n, int32_t length);
    virtual UBool Resize(int32_t appendCapacity, int32_t length);

    CollationKey &key_;
};

CollationKeyByteSink::~CollationKeyByteSink() {}

void
CollationKeyByteSink::AppendBeyondCapacity(const char *bytes, int32_t n, int32_t length) {
    // buffer_ != NULL && bytes != NULL && n > 0 && appended_ > capacity_
    if (Resize(n, length)) {
        uprv_memcpy(buffer_ + length, bytes, n);
    }
}

UBool
CollationKeyByteSink::Resize(int32_t appendCapacity, int32_t length) {
    if (buffer_ == NULL) {
        return FALSE;  // allocation failed before already
    }
    int32_t newCapacity = 2 * capacity_;
    int32_t altCapacity = length + 2 * appendCapacity;
    if (newCapacity < altCapacity) {
        newCapacity = altCapacity;
    }
    if (newCapacity < 200) {
        newCapacity = 200;
    }
    uint8_t *newBuffer = key_.reallocate(newCapacity, length);
    if (newBuffer == NULL) {
        SetNotOk();
        return FALSE;
    }
    buffer_ = reinterpret_cast<char *>(newBuffer);
    capacity_ = newCapacity;
    return TRUE;
}

RuleBasedCollator::RuleBasedCollator(const RuleBasedCollator &other)
        : Collator(other),
          data(other.data),
          settings(other.settings),
          tailoring(other.tailoring),
          cacheEntry(other.cacheEntry),
          validLocale(other.validLocale),
          explicitlySetAttributes(other.explicitlySetAttributes),
          actualLocaleIsSameAsValid(other.actualLocaleIsSameAsValid) {
    settings->addRef();
    cacheEntry->addRef();
}

RuleBasedCollator::RuleBasedCollator(const uint8_t *bin, int32_t length,
                                     const RuleBasedCollator *base, UErrorCode &errorCode)
        : data(NULL),
          settings(NULL),
          tailoring(NULL),
          cacheEntry(NULL),
          validLocale(""),
          explicitlySetAttributes(0),
          actualLocaleIsSameAsValid(FALSE) {
    if(U_FAILURE(errorCode)) { return; }
    if(bin == NULL || length == 0 || base == NULL) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    const CollationTailoring *root = CollationRoot::getRoot(errorCode);
    if(U_FAILURE(errorCode)) { return; }
    if(base->tailoring != root) {
        errorCode = U_UNSUPPORTED_ERROR;
        return;
    }
    LocalPointer<CollationTailoring> t(new CollationTailoring(base->tailoring->settings));
    if(t.isNull() || t->isBogus()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    CollationDataReader::read(base->tailoring, bin, length, *t, errorCode);
    if(U_FAILURE(errorCode)) { return; }
    t->actualLocale.setToBogus();
    adoptTailoring(t.orphan(), errorCode);
}

RuleBasedCollator::RuleBasedCollator(const CollationCacheEntry *entry)
        : data(entry->tailoring->data),
          settings(entry->tailoring->settings),
          tailoring(entry->tailoring),
          cacheEntry(entry),
          validLocale(entry->validLocale),
          explicitlySetAttributes(0),
          actualLocaleIsSameAsValid(FALSE) {
    settings->addRef();
    cacheEntry->addRef();
}

RuleBasedCollator::~RuleBasedCollator() {
    SharedObject::clearPtr(settings);
    SharedObject::clearPtr(cacheEntry);
}

void
RuleBasedCollator::adoptTailoring(CollationTailoring *t, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) {
        t->deleteIfZeroRefCount();
        return;
    }
    U_ASSERT(settings == NULL && data == NULL && tailoring == NULL && cacheEntry == NULL);
    cacheEntry = new CollationCacheEntry(t->actualLocale, t);
    if(cacheEntry == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        t->deleteIfZeroRefCount();
        return;
    }
    data = t->data;
    settings = t->settings;
    settings->addRef();
    tailoring = t;
    cacheEntry->addRef();
    validLocale = t->actualLocale;
    actualLocaleIsSameAsValid = FALSE;
}

RuleBasedCollator *
RuleBasedCollator::clone() const {
    return new RuleBasedCollator(*this);
}

RuleBasedCollator &RuleBasedCollator::operator=(const RuleBasedCollator &other) {
    if(this == &other) { return *this; }
    SharedObject::copyPtr(other.settings, settings);
    tailoring = other.tailoring;
    SharedObject::copyPtr(other.cacheEntry, cacheEntry);
    data = tailoring->data;
    validLocale = other.validLocale;
    explicitlySetAttributes = other.explicitlySetAttributes;
    actualLocaleIsSameAsValid = other.actualLocaleIsSameAsValid;
    return *this;
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(RuleBasedCollator)

UBool
RuleBasedCollator::operator==(const Collator& other) const {
    if(this == &other) { return TRUE; }
    if(!Collator::operator==(other)) { return FALSE; }
    const RuleBasedCollator &o = static_cast<const RuleBasedCollator &>(other);
    if(*settings != *o.settings) { return FALSE; }
    if(data == o.data) { return TRUE; }
    UBool thisIsRoot = data->base == NULL;
    UBool otherIsRoot = o.data->base == NULL;
    U_ASSERT(!thisIsRoot || !otherIsRoot);  // otherwise their data pointers should be ==
    if(thisIsRoot != otherIsRoot) { return FALSE; }
    if((thisIsRoot || !tailoring->rules.isEmpty()) &&
            (otherIsRoot || !o.tailoring->rules.isEmpty())) {
        // Shortcut: If both collators have valid rule strings, then compare those.
        if(tailoring->rules == o.tailoring->rules) { return TRUE; }
    }
    // Different rule strings can result in the same or equivalent tailoring.
    // The rule strings are optional in ICU resource bundles, although included by default.
    // cloneBinary() drops the rule string.
    UErrorCode errorCode = U_ZERO_ERROR;
    LocalPointer<UnicodeSet> thisTailored(getTailoredSet(errorCode));
    LocalPointer<UnicodeSet> otherTailored(o.getTailoredSet(errorCode));
    if(U_FAILURE(errorCode)) { return FALSE; }
    if(*thisTailored != *otherTailored) { return FALSE; }
    // For completeness, we should compare all of the mappings;
    // or we should create a list of strings, sort it with one collator,
    // and check if both collators compare adjacent strings the same
    // (order & strength, down to quaternary); or similar.
    // Testing equality of collators seems unusual.
    return TRUE;
}

int32_t
RuleBasedCollator::hashCode() const {
    int32_t h = settings->hashCode();
    if(data->base == NULL) { return h; }  // root collator
    // Do not rely on the rule string, see comments in operator==().
    UErrorCode errorCode = U_ZERO_ERROR;
    LocalPointer<UnicodeSet> set(getTailoredSet(errorCode));
    if(U_FAILURE(errorCode)) { return 0; }
    UnicodeSetIterator iter(*set);
    while(iter.next() && !iter.isString()) {
        h ^= data->getCE32(iter.getCodepoint());
    }
    return h;
}

void
RuleBasedCollator::setLocales(const Locale &requested, const Locale &valid,
                              const Locale &actual) {
    if(actual == tailoring->actualLocale) {
        actualLocaleIsSameAsValid = FALSE;
    } else {
        U_ASSERT(actual == valid);
        actualLocaleIsSameAsValid = TRUE;
    }
    // Do not modify tailoring.actualLocale:
    // We cannot be sure that that would be thread-safe.
    validLocale = valid;
    (void)requested;  // Ignore, see also ticket #10477.
}

Locale
RuleBasedCollator::getLocale(ULocDataLocaleType type, UErrorCode& errorCode) const {
    if(U_FAILURE(errorCode)) {
        return Locale::getRoot();
    }
    switch(type) {
    case ULOC_ACTUAL_LOCALE:
        return actualLocaleIsSameAsValid ? validLocale : tailoring->actualLocale;
    case ULOC_VALID_LOCALE:
        return validLocale;
    case ULOC_REQUESTED_LOCALE:
    default:
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return Locale::getRoot();
    }
}

const char *
RuleBasedCollator::internalGetLocaleID(ULocDataLocaleType type, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return NULL;
    }
    const Locale *result;
    switch(type) {
    case ULOC_ACTUAL_LOCALE:
        result = actualLocaleIsSameAsValid ? &validLocale : &tailoring->actualLocale;
        break;
    case ULOC_VALID_LOCALE:
        result = &validLocale;
        break;
    case ULOC_REQUESTED_LOCALE:
    default:
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    if(result->isBogus()) { return NULL; }
    const char *id = result->getName();
    return id[0] == 0 ? "root" : id;
}

const UnicodeString&
RuleBasedCollator::getRules() const {
    return tailoring->rules;
}

void
RuleBasedCollator::getRules(UColRuleOption delta, UnicodeString &buffer) const {
    if(delta == UCOL_TAILORING_ONLY) {
        buffer = tailoring->rules;
        return;
    }
    // UCOL_FULL_RULES
    buffer.remove();
    CollationLoader::appendRootRules(buffer);
    buffer.append(tailoring->rules).getTerminatedBuffer();
}

void
RuleBasedCollator::getVersion(UVersionInfo version) const {
    uprv_memcpy(version, tailoring->version, U_MAX_VERSION_LENGTH);
    version[0] += (UCOL_RUNTIME_VERSION << 4) + (UCOL_RUNTIME_VERSION >> 4);
}

UnicodeSet *
RuleBasedCollator::getTailoredSet(UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return NULL; }
    UnicodeSet *tailored = new UnicodeSet();
    if(tailored == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    if(data->base != NULL) {
        TailoredSet(tailored).forData(data, errorCode);
        if(U_FAILURE(errorCode)) {
            delete tailored;
            return NULL;
        }
    }
    return tailored;
}

void
RuleBasedCollator::internalGetContractionsAndExpansions(
        UnicodeSet *contractions, UnicodeSet *expansions,
        UBool addPrefixes, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return; }
    if(contractions != NULL) {
        contractions->clear();
    }
    if(expansions != NULL) {
        expansions->clear();
    }
    ContractionsAndExpansions(contractions, expansions, NULL, addPrefixes).forData(data, errorCode);
}

void
RuleBasedCollator::internalAddContractions(UChar32 c, UnicodeSet &set, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return; }
    ContractionsAndExpansions(&set, NULL, NULL, FALSE).forCodePoint(data, c, errorCode);
}

const CollationSettings &
RuleBasedCollator::getDefaultSettings() const {
    return *tailoring->settings;
}

UColAttributeValue
RuleBasedCollator::getAttribute(UColAttribute attr, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return UCOL_DEFAULT; }
    int32_t option;
    switch(attr) {
    case UCOL_FRENCH_COLLATION:
        option = CollationSettings::BACKWARD_SECONDARY;
        break;
    case UCOL_ALTERNATE_HANDLING:
        return settings->getAlternateHandling();
    case UCOL_CASE_FIRST:
        return settings->getCaseFirst();
    case UCOL_CASE_LEVEL:
        option = CollationSettings::CASE_LEVEL;
        break;
    case UCOL_NORMALIZATION_MODE:
        option = CollationSettings::CHECK_FCD;
        break;
    case UCOL_STRENGTH:
        return (UColAttributeValue)settings->getStrength();
    case UCOL_HIRAGANA_QUATERNARY_MODE:
        // Deprecated attribute, unsettable.
        return UCOL_OFF;
    case UCOL_NUMERIC_COLLATION:
        option = CollationSettings::NUMERIC;
        break;
    default:
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return UCOL_DEFAULT;
    }
    return ((settings->options & option) == 0) ? UCOL_OFF : UCOL_ON;
}

void
RuleBasedCollator::setAttribute(UColAttribute attr, UColAttributeValue value,
                                UErrorCode &errorCode) {
    UColAttributeValue oldValue = getAttribute(attr, errorCode);
    if(U_FAILURE(errorCode)) { return; }
    if(value == oldValue) {
        setAttributeExplicitly(attr);
        return;
    }
    const CollationSettings &defaultSettings = getDefaultSettings();
    if(settings == &defaultSettings) {
        if(value == UCOL_DEFAULT) {
            setAttributeDefault(attr);
            return;
        }
    }
    CollationSettings *ownedSettings = SharedObject::copyOnWrite(settings);
    if(ownedSettings == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    switch(attr) {
    case UCOL_FRENCH_COLLATION:
        ownedSettings->setFlag(CollationSettings::BACKWARD_SECONDARY, value,
                               defaultSettings.options, errorCode);
        break;
    case UCOL_ALTERNATE_HANDLING:
        ownedSettings->setAlternateHandling(value, defaultSettings.options, errorCode);
        break;
    case UCOL_CASE_FIRST:
        ownedSettings->setCaseFirst(value, defaultSettings.options, errorCode);
        break;
    case UCOL_CASE_LEVEL:
        ownedSettings->setFlag(CollationSettings::CASE_LEVEL, value,
                               defaultSettings.options, errorCode);
        break;
    case UCOL_NORMALIZATION_MODE:
        ownedSettings->setFlag(CollationSettings::CHECK_FCD, value,
                               defaultSettings.options, errorCode);
        break;
    case UCOL_STRENGTH:
        ownedSettings->setStrength(value, defaultSettings.options, errorCode);
        break;
    case UCOL_HIRAGANA_QUATERNARY_MODE:
        // Deprecated attribute. Check for valid values but do not change anything.
        if(value != UCOL_OFF && value != UCOL_ON && value != UCOL_DEFAULT) {
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        }
        break;
    case UCOL_NUMERIC_COLLATION:
        ownedSettings->setFlag(CollationSettings::NUMERIC, value, defaultSettings.options, errorCode);
        break;
    default:
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        break;
    }
    if(U_FAILURE(errorCode)) { return; }
    setFastLatinOptions(*ownedSettings);
    if(value == UCOL_DEFAULT) {
        setAttributeDefault(attr);
    } else {
        setAttributeExplicitly(attr);
    }
}

Collator &
RuleBasedCollator::setMaxVariable(UColReorderCode group, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return *this; }
    // Convert the reorder code into a MaxVariable number, or UCOL_DEFAULT=-1.
    int32_t value;
    if(group == UCOL_REORDER_CODE_DEFAULT) {
        value = UCOL_DEFAULT;
    } else if(UCOL_REORDER_CODE_FIRST <= group && group <= UCOL_REORDER_CODE_CURRENCY) {
        value = group - UCOL_REORDER_CODE_FIRST;
    } else {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return *this;
    }
    CollationSettings::MaxVariable oldValue = settings->getMaxVariable();
    if(value == oldValue) {
        setAttributeExplicitly(ATTR_VARIABLE_TOP);
        return *this;
    }
    const CollationSettings &defaultSettings = getDefaultSettings();
    if(settings == &defaultSettings) {
        if(value == UCOL_DEFAULT) {
            setAttributeDefault(ATTR_VARIABLE_TOP);
            return *this;
        }
    }
    CollationSettings *ownedSettings = SharedObject::copyOnWrite(settings);
    if(ownedSettings == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return *this;
    }

    if(group == UCOL_REORDER_CODE_DEFAULT) {
        group = (UColReorderCode)(UCOL_REORDER_CODE_FIRST + defaultSettings.getMaxVariable());
    }
    uint32_t varTop = data->getLastPrimaryForGroup(group);
    U_ASSERT(varTop != 0);
    ownedSettings->setMaxVariable(value, defaultSettings.options, errorCode);
    if(U_FAILURE(errorCode)) { return *this; }
    ownedSettings->variableTop = varTop;
    setFastLatinOptions(*ownedSettings);
    if(value == UCOL_DEFAULT) {
        setAttributeDefault(ATTR_VARIABLE_TOP);
    } else {
        setAttributeExplicitly(ATTR_VARIABLE_TOP);
    }
    return *this;
}

UColReorderCode
RuleBasedCollator::getMaxVariable() const {
    return (UColReorderCode)(UCOL_REORDER_CODE_FIRST + settings->getMaxVariable());
}

uint32_t
RuleBasedCollator::getVariableTop(UErrorCode & /*errorCode*/) const {
    return settings->variableTop;
}

uint32_t
RuleBasedCollator::setVariableTop(const UChar *varTop, int32_t len, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return 0; }
    if(varTop == NULL && len !=0) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    if(len < 0) { len = u_strlen(varTop); }
    if(len == 0) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UBool numeric = settings->isNumeric();
    int64_t ce1, ce2;
    if(settings->dontCheckFCD()) {
        UTF16CollationIterator ci(data, numeric, varTop, varTop, varTop + len);
        ce1 = ci.nextCE(errorCode);
        ce2 = ci.nextCE(errorCode);
    } else {
        FCDUTF16CollationIterator ci(data, numeric, varTop, varTop, varTop + len);
        ce1 = ci.nextCE(errorCode);
        ce2 = ci.nextCE(errorCode);
    }
    if(ce1 == Collation::NO_CE || ce2 != Collation::NO_CE) {
        errorCode = U_CE_NOT_FOUND_ERROR;
        return 0;
    }
    setVariableTop((uint32_t)(ce1 >> 32), errorCode);
    return settings->variableTop;
}

uint32_t
RuleBasedCollator::setVariableTop(const UnicodeString &varTop, UErrorCode &errorCode) {
    return setVariableTop(varTop.getBuffer(), varTop.length(), errorCode);
}

void
RuleBasedCollator::setVariableTop(uint32_t varTop, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(varTop != settings->variableTop) {
        // Pin the variable top to the end of the reordering group which contains it.
        // Only a few special groups are supported.
        int32_t group = data->getGroupForPrimary(varTop);
        if(group < UCOL_REORDER_CODE_FIRST || UCOL_REORDER_CODE_CURRENCY < group) {
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        uint32_t v = data->getLastPrimaryForGroup(group);
        U_ASSERT(v != 0 && v >= varTop);
        varTop = v;
        if(varTop != settings->variableTop) {
            CollationSettings *ownedSettings = SharedObject::copyOnWrite(settings);
            if(ownedSettings == NULL) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            ownedSettings->setMaxVariable(group - UCOL_REORDER_CODE_FIRST,
                                          getDefaultSettings().options, errorCode);
            if(U_FAILURE(errorCode)) { return; }
            ownedSettings->variableTop = varTop;
            setFastLatinOptions(*ownedSettings);
        }
    }
    if(varTop == getDefaultSettings().variableTop) {
        setAttributeDefault(ATTR_VARIABLE_TOP);
    } else {
        setAttributeExplicitly(ATTR_VARIABLE_TOP);
    }
}

int32_t
RuleBasedCollator::getReorderCodes(int32_t *dest, int32_t capacity,
                                   UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return 0; }
    if(capacity < 0 || (dest == NULL && capacity > 0)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    int32_t length = settings->reorderCodesLength;
    if(length == 0) { return 0; }
    if(length > capacity) {
        errorCode = U_BUFFER_OVERFLOW_ERROR;
        return length;
    }
    uprv_memcpy(dest, settings->reorderCodes, length * 4);
    return length;
}

void
RuleBasedCollator::setReorderCodes(const int32_t *reorderCodes, int32_t length,
                                   UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(length < 0 || (reorderCodes == NULL && length > 0)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if(length == 1 && reorderCodes[0] == UCOL_REORDER_CODE_NONE) {
        length = 0;
    }
    if(length == settings->reorderCodesLength &&
            uprv_memcmp(reorderCodes, settings->reorderCodes, length * 4) == 0) {
        return;
    }
    const CollationSettings &defaultSettings = getDefaultSettings();
    if(length == 1 && reorderCodes[0] == UCOL_REORDER_CODE_DEFAULT) {
        if(settings != &defaultSettings) {
            CollationSettings *ownedSettings = SharedObject::copyOnWrite(settings);
            if(ownedSettings == NULL) {
                errorCode = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            ownedSettings->copyReorderingFrom(defaultSettings, errorCode);
            setFastLatinOptions(*ownedSettings);
        }
        return;
    }
    CollationSettings *ownedSettings = SharedObject::copyOnWrite(settings);
    if(ownedSettings == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    ownedSettings->setReordering(*data, reorderCodes, length, errorCode);
    setFastLatinOptions(*ownedSettings);
}

void
RuleBasedCollator::setFastLatinOptions(CollationSettings &ownedSettings) const {
    ownedSettings.fastLatinOptions = CollationFastLatin::getOptions(
            data, ownedSettings,
            ownedSettings.fastLatinPrimaries, UPRV_LENGTHOF(ownedSettings.fastLatinPrimaries));
}

UCollationResult
RuleBasedCollator::compare(const UnicodeString &left, const UnicodeString &right,
                           UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return UCOL_EQUAL; }
    return doCompare(left.getBuffer(), left.length(),
                     right.getBuffer(), right.length(), errorCode);
}

UCollationResult
RuleBasedCollator::compare(const UnicodeString &left, const UnicodeString &right,
                           int32_t length, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode) || length == 0) { return UCOL_EQUAL; }
    if(length < 0) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return UCOL_EQUAL;
    }
    int32_t leftLength = left.length();
    int32_t rightLength = right.length();
    if(leftLength > length) { leftLength = length; }
    if(rightLength > length) { rightLength = length; }
    return doCompare(left.getBuffer(), leftLength,
                     right.getBuffer(), rightLength, errorCode);
}

UCollationResult
RuleBasedCollator::compare(const UChar *left, int32_t leftLength,
                           const UChar *right, int32_t rightLength,
                           UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return UCOL_EQUAL; }
    if((left == NULL && leftLength != 0) || (right == NULL && rightLength != 0)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return UCOL_EQUAL;
    }
    // Make sure both or neither strings have a known length.
    // We do not optimize for mixed length/termination.
    if(leftLength >= 0) {
        if(rightLength < 0) { rightLength = u_strlen(right); }
    } else {
        if(rightLength >= 0) { leftLength = u_strlen(left); }
    }
    return doCompare(left, leftLength, right, rightLength, errorCode);
}

UCollationResult
RuleBasedCollator::compareUTF8(const StringPiece &left, const StringPiece &right,
                               UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return UCOL_EQUAL; }
    const uint8_t *leftBytes = reinterpret_cast<const uint8_t *>(left.data());
    const uint8_t *rightBytes = reinterpret_cast<const uint8_t *>(right.data());
    if((leftBytes == NULL && !left.empty()) || (rightBytes == NULL && !right.empty())) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return UCOL_EQUAL;
    }
    return doCompare(leftBytes, left.length(), rightBytes, right.length(), errorCode);
}

UCollationResult
RuleBasedCollator::internalCompareUTF8(const char *left, int32_t leftLength,
                                       const char *right, int32_t rightLength,
                                       UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return UCOL_EQUAL; }
    if((left == NULL && leftLength != 0) || (right == NULL && rightLength != 0)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return UCOL_EQUAL;
    }
    // Make sure both or neither strings have a known length.
    // We do not optimize for mixed length/termination.
    if(leftLength >= 0) {
        if(rightLength < 0) { rightLength = static_cast<int32_t>(uprv_strlen(right)); }
    } else {
        if(rightLength >= 0) { leftLength = static_cast<int32_t>(uprv_strlen(left)); }
    }
    return doCompare(reinterpret_cast<const uint8_t *>(left), leftLength,
                     reinterpret_cast<const uint8_t *>(right), rightLength, errorCode);
}

namespace {

/**
 * Abstract iterator for identical-level string comparisons.
 * Returns FCD code points and handles temporary switching to NFD.
 */
class NFDIterator : public UObject {
public:
    NFDIterator() : index(-1), length(0) {}
    virtual ~NFDIterator() {}
    /**
     * Returns the next code point from the internal normalization buffer,
     * or else the next text code point.
     * Returns -1 at the end of the text.
     */
    UChar32 nextCodePoint() {
        if(index >= 0) {
            if(index == length) {
                index = -1;
            } else {
                UChar32 c;
                U16_NEXT_UNSAFE(decomp, index, c);
                return c;
            }
        }
        return nextRawCodePoint();
    }
    /**
     * @param nfcImpl
     * @param c the last code point returned by nextCodePoint() or nextDecomposedCodePoint()
     * @return the first code point in c's decomposition,
     *         or c itself if it was decomposed already or if it does not decompose
     */
    UChar32 nextDecomposedCodePoint(const Normalizer2Impl &nfcImpl, UChar32 c) {
        if(index >= 0) { return c; }
        decomp = nfcImpl.getDecomposition(c, buffer, length);
        if(decomp == NULL) { return c; }
        index = 0;
        U16_NEXT_UNSAFE(decomp, index, c);
        return c;
    }
protected:
    /**
     * Returns the next text code point in FCD order.
     * Returns -1 at the end of the text.
     */
    virtual UChar32 nextRawCodePoint() = 0;
private:
    const UChar *decomp;
    UChar buffer[4];
    int32_t index;
    int32_t length;
};

class UTF16NFDIterator : public NFDIterator {
public:
    UTF16NFDIterator(const UChar *text, const UChar *textLimit) : s(text), limit(textLimit) {}
protected:
    virtual UChar32 nextRawCodePoint() {
        if(s == limit) { return U_SENTINEL; }
        UChar32 c = *s++;
        if(limit == NULL && c == 0) {
            s = NULL;
            return U_SENTINEL;
        }
        UChar trail;
        if(U16_IS_LEAD(c) && s != limit && U16_IS_TRAIL(trail = *s)) {
            ++s;
            c = U16_GET_SUPPLEMENTARY(c, trail);
        }
        return c;
    }

    const UChar *s;
    const UChar *limit;
};

class FCDUTF16NFDIterator : public UTF16NFDIterator {
public:
    FCDUTF16NFDIterator(const Normalizer2Impl &nfcImpl, const UChar *text, const UChar *textLimit)
            : UTF16NFDIterator(NULL, NULL) {
        UErrorCode errorCode = U_ZERO_ERROR;
        const UChar *spanLimit = nfcImpl.makeFCD(text, textLimit, NULL, errorCode);
        if(U_FAILURE(errorCode)) { return; }
        if(spanLimit == textLimit || (textLimit == NULL && *spanLimit == 0)) {
            s = text;
            limit = spanLimit;
        } else {
            str.setTo(text, (int32_t)(spanLimit - text));
            {
                ReorderingBuffer r_buffer(nfcImpl, str);
                if(r_buffer.init(str.length(), errorCode)) {
                    nfcImpl.makeFCD(spanLimit, textLimit, &r_buffer, errorCode);
                }
            }
            if(U_SUCCESS(errorCode)) {
                s = str.getBuffer();
                limit = s + str.length();
            }
        }
    }
private:
    UnicodeString str;
};

class UTF8NFDIterator : public NFDIterator {
public:
    UTF8NFDIterator(const uint8_t *text, int32_t textLength)
        : s(text), pos(0), length(textLength) {}
protected:
    virtual UChar32 nextRawCodePoint() {
        if(pos == length || (s[pos] == 0 && length < 0)) { return U_SENTINEL; }
        UChar32 c;
        U8_NEXT_OR_FFFD(s, pos, length, c);
        return c;
    }

    const uint8_t *s;
    int32_t pos;
    int32_t length;
};

class FCDUTF8NFDIterator : public NFDIterator {
public:
    FCDUTF8NFDIterator(const CollationData *data, const uint8_t *text, int32_t textLength)
            : u8ci(data, FALSE, text, 0, textLength) {}
protected:
    virtual UChar32 nextRawCodePoint() {
        UErrorCode errorCode = U_ZERO_ERROR;
        return u8ci.nextCodePoint(errorCode);
    }
private:
    FCDUTF8CollationIterator u8ci;
};

class UIterNFDIterator : public NFDIterator {
public:
    UIterNFDIterator(UCharIterator &it) : iter(it) {}
protected:
    virtual UChar32 nextRawCodePoint() {
        return uiter_next32(&iter);
    }
private:
    UCharIterator &iter;
};

class FCDUIterNFDIterator : public NFDIterator {
public:
    FCDUIterNFDIterator(const CollationData *data, UCharIterator &it, int32_t startIndex)
            : uici(data, FALSE, it, startIndex) {}
protected:
    virtual UChar32 nextRawCodePoint() {
        UErrorCode errorCode = U_ZERO_ERROR;
        return uici.nextCodePoint(errorCode);
    }
private:
    FCDUIterCollationIterator uici;
};

UCollationResult compareNFDIter(const Normalizer2Impl &nfcImpl,
                                NFDIterator &left, NFDIterator &right) {
    for(;;) {
        // Fetch the next FCD code point from each string.
        UChar32 leftCp = left.nextCodePoint();
        UChar32 rightCp = right.nextCodePoint();
        if(leftCp == rightCp) {
            if(leftCp < 0) { break; }
            continue;
        }
        // If they are different, then decompose each and compare again.
        if(leftCp < 0) {
            leftCp = -2;  // end of string
        } else if(leftCp == 0xfffe) {
            leftCp = -1;  // U+FFFE: merge separator
        } else {
            leftCp = left.nextDecomposedCodePoint(nfcImpl, leftCp);
        }
        if(rightCp < 0) {
            rightCp = -2;  // end of string
        } else if(rightCp == 0xfffe) {
            rightCp = -1;  // U+FFFE: merge separator
        } else {
            rightCp = right.nextDecomposedCodePoint(nfcImpl, rightCp);
        }
        if(leftCp < rightCp) { return UCOL_LESS; }
        if(leftCp > rightCp) { return UCOL_GREATER; }
    }
    return UCOL_EQUAL;
}

}  // namespace

UCollationResult
RuleBasedCollator::doCompare(const UChar *left, int32_t leftLength,
                             const UChar *right, int32_t rightLength,
                             UErrorCode &errorCode) const {
    // U_FAILURE(errorCode) checked by caller.
    if(left == right && leftLength == rightLength) {
        return UCOL_EQUAL;
    }

    // Identical-prefix test.
    const UChar *leftLimit;
    const UChar *rightLimit;
    int32_t equalPrefixLength = 0;
    if(leftLength < 0) {
        leftLimit = NULL;
        rightLimit = NULL;
        UChar c;
        while((c = left[equalPrefixLength]) == right[equalPrefixLength]) {
            if(c == 0) { return UCOL_EQUAL; }
            ++equalPrefixLength;
        }
    } else {
        leftLimit = left + leftLength;
        rightLimit = right + rightLength;
        for(;;) {
            if(equalPrefixLength == leftLength) {
                if(equalPrefixLength == rightLength) { return UCOL_EQUAL; }
                break;
            } else if(equalPrefixLength == rightLength ||
                      left[equalPrefixLength] != right[equalPrefixLength]) {
                break;
            }
            ++equalPrefixLength;
        }
    }

    UBool numeric = settings->isNumeric();
    if(equalPrefixLength > 0) {
        if((equalPrefixLength != leftLength &&
                    data->isUnsafeBackward(left[equalPrefixLength], numeric)) ||
                (equalPrefixLength != rightLength &&
                    data->isUnsafeBackward(right[equalPrefixLength], numeric))) {
            // Identical prefix: Back up to the start of a contraction or reordering sequence.
            while(--equalPrefixLength > 0 &&
                    data->isUnsafeBackward(left[equalPrefixLength], numeric)) {}
        }
        // Notes:
        // - A longer string can compare equal to a prefix of it if only ignorables follow.
        // - With a backward level, a longer string can compare less-than a prefix of it.

        // Pass the actual start of each string into the CollationIterators,
        // plus the equalPrefixLength position,
        // so that prefix matches back into the equal prefix work.
    }

    int32_t result;
    int32_t fastLatinOptions = settings->fastLatinOptions;
    if(fastLatinOptions >= 0 &&
            (equalPrefixLength == leftLength ||
                left[equalPrefixLength] <= CollationFastLatin::LATIN_MAX) &&
            (equalPrefixLength == rightLength ||
                right[equalPrefixLength] <= CollationFastLatin::LATIN_MAX)) {
        if(leftLength >= 0) {
            result = CollationFastLatin::compareUTF16(data->fastLatinTable,
                                                      settings->fastLatinPrimaries,
                                                      fastLatinOptions,
                                                      left + equalPrefixLength,
                                                      leftLength - equalPrefixLength,
                                                      right + equalPrefixLength,
                                                      rightLength - equalPrefixLength);
        } else {
            result = CollationFastLatin::compareUTF16(data->fastLatinTable,
                                                      settings->fastLatinPrimaries,
                                                      fastLatinOptions,
                                                      left + equalPrefixLength, -1,
                                                      right + equalPrefixLength, -1);
        }
    } else {
        result = CollationFastLatin::BAIL_OUT_RESULT;
    }

    if(result == CollationFastLatin::BAIL_OUT_RESULT) {
        if(settings->dontCheckFCD()) {
            UTF16CollationIterator leftIter(data, numeric,
                                            left, left + equalPrefixLength, leftLimit);
            UTF16CollationIterator rightIter(data, numeric,
                                            right, right + equalPrefixLength, rightLimit);
            result = CollationCompare::compareUpToQuaternary(leftIter, rightIter, *settings, errorCode);
        } else {
            FCDUTF16CollationIterator leftIter(data, numeric,
                                              left, left + equalPrefixLength, leftLimit);
            FCDUTF16CollationIterator rightIter(data, numeric,
                                                right, right + equalPrefixLength, rightLimit);
            result = CollationCompare::compareUpToQuaternary(leftIter, rightIter, *settings, errorCode);
        }
    }
    if(result != UCOL_EQUAL || settings->getStrength() < UCOL_IDENTICAL || U_FAILURE(errorCode)) {
        return (UCollationResult)result;
    }

    // Note: If NUL-terminated, we could get the actual limits from the iterators now.
    // That would complicate the iterators a bit, NUL-terminated strings are only a C convenience,
    // and the benefit seems unlikely to be measurable.

    // Compare identical level.
    const Normalizer2Impl &nfcImpl = data->nfcImpl;
    left += equalPrefixLength;
    right += equalPrefixLength;
    if(settings->dontCheckFCD()) {
        UTF16NFDIterator leftIter(left, leftLimit);
        UTF16NFDIterator rightIter(right, rightLimit);
        return compareNFDIter(nfcImpl, leftIter, rightIter);
    } else {
        FCDUTF16NFDIterator leftIter(nfcImpl, left, leftLimit);
        FCDUTF16NFDIterator rightIter(nfcImpl, right, rightLimit);
        return compareNFDIter(nfcImpl, leftIter, rightIter);
    }
}

UCollationResult
RuleBasedCollator::doCompare(const uint8_t *left, int32_t leftLength,
                             const uint8_t *right, int32_t rightLength,
                             UErrorCode &errorCode) const {
    // U_FAILURE(errorCode) checked by caller.
    if(left == right && leftLength == rightLength) {
        return UCOL_EQUAL;
    }

    // Identical-prefix test.
    int32_t equalPrefixLength = 0;
    if(leftLength < 0) {
        uint8_t c;
        while((c = left[equalPrefixLength]) == right[equalPrefixLength]) {
            if(c == 0) { return UCOL_EQUAL; }
            ++equalPrefixLength;
        }
    } else {
        for(;;) {
            if(equalPrefixLength == leftLength) {
                if(equalPrefixLength == rightLength) { return UCOL_EQUAL; }
                break;
            } else if(equalPrefixLength == rightLength ||
                      left[equalPrefixLength] != right[equalPrefixLength]) {
                break;
            }
            ++equalPrefixLength;
        }
    }
    // Back up to the start of a partially-equal code point.
    if(equalPrefixLength > 0 &&
            ((equalPrefixLength != leftLength && U8_IS_TRAIL(left[equalPrefixLength])) ||
            (equalPrefixLength != rightLength && U8_IS_TRAIL(right[equalPrefixLength])))) {
        while(--equalPrefixLength > 0 && U8_IS_TRAIL(left[equalPrefixLength])) {}
    }

    UBool numeric = settings->isNumeric();
    if(equalPrefixLength > 0) {
        UBool unsafe = FALSE;
        if(equalPrefixLength != leftLength) {
            int32_t i = equalPrefixLength;
            UChar32 c;
            U8_NEXT_OR_FFFD(left, i, leftLength, c);
            unsafe = data->isUnsafeBackward(c, numeric);
        }
        if(!unsafe && equalPrefixLength != rightLength) {
            int32_t i = equalPrefixLength;
            UChar32 c;
            U8_NEXT_OR_FFFD(right, i, rightLength, c);
            unsafe = data->isUnsafeBackward(c, numeric);
        }
        if(unsafe) {
            // Identical prefix: Back up to the start of a contraction or reordering sequence.
            UChar32 c;
            do {
                U8_PREV_OR_FFFD(left, 0, equalPrefixLength, c);
            } while(equalPrefixLength > 0 && data->isUnsafeBackward(c, numeric));
        }
        // See the notes in the UTF-16 version.

        // Pass the actual start of each string into the CollationIterators,
        // plus the equalPrefixLength position,
        // so that prefix matches back into the equal prefix work.
    }

    int32_t result;
    int32_t fastLatinOptions = settings->fastLatinOptions;
    if(fastLatinOptions >= 0 &&
            (equalPrefixLength == leftLength ||
                left[equalPrefixLength] <= CollationFastLatin::LATIN_MAX_UTF8_LEAD) &&
            (equalPrefixLength == rightLength ||
                right[equalPrefixLength] <= CollationFastLatin::LATIN_MAX_UTF8_LEAD)) {
        if(leftLength >= 0) {
            result = CollationFastLatin::compareUTF8(data->fastLatinTable,
                                                     settings->fastLatinPrimaries,
                                                     fastLatinOptions,
                                                     left + equalPrefixLength,
                                                     leftLength - equalPrefixLength,
                                                     right + equalPrefixLength,
                                                     rightLength - equalPrefixLength);
        } else {
            result = CollationFastLatin::compareUTF8(data->fastLatinTable,
                                                     settings->fastLatinPrimaries,
                                                     fastLatinOptions,
                                                     left + equalPrefixLength, -1,
                                                     right + equalPrefixLength, -1);
        }
    } else {
        result = CollationFastLatin::BAIL_OUT_RESULT;
    }

    if(result == CollationFastLatin::BAIL_OUT_RESULT) {
        if(settings->dontCheckFCD()) {
            UTF8CollationIterator leftIter(data, numeric, left, equalPrefixLength, leftLength);
            UTF8CollationIterator rightIter(data, numeric, right, equalPrefixLength, rightLength);
            result = CollationCompare::compareUpToQuaternary(leftIter, rightIter, *settings, errorCode);
        } else {
            FCDUTF8CollationIterator leftIter(data, numeric, left, equalPrefixLength, leftLength);
            FCDUTF8CollationIterator rightIter(data, numeric, right, equalPrefixLength, rightLength);
            result = CollationCompare::compareUpToQuaternary(leftIter, rightIter, *settings, errorCode);
        }
    }
    if(result != UCOL_EQUAL || settings->getStrength() < UCOL_IDENTICAL || U_FAILURE(errorCode)) {
        return (UCollationResult)result;
    }

    // Note: If NUL-terminated, we could get the actual limits from the iterators now.
    // That would complicate the iterators a bit, NUL-terminated strings are only a C convenience,
    // and the benefit seems unlikely to be measurable.

    // Compare identical level.
    const Normalizer2Impl &nfcImpl = data->nfcImpl;
    left += equalPrefixLength;
    right += equalPrefixLength;
    if(leftLength > 0) {
        leftLength -= equalPrefixLength;
        rightLength -= equalPrefixLength;
    }
    if(settings->dontCheckFCD()) {
        UTF8NFDIterator leftIter(left, leftLength);
        UTF8NFDIterator rightIter(right, rightLength);
        return compareNFDIter(nfcImpl, leftIter, rightIter);
    } else {
        FCDUTF8NFDIterator leftIter(data, left, leftLength);
        FCDUTF8NFDIterator rightIter(data, right, rightLength);
        return compareNFDIter(nfcImpl, leftIter, rightIter);
    }
}

UCollationResult
RuleBasedCollator::compare(UCharIterator &left, UCharIterator &right,
                           UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode) || &left == &right) { return UCOL_EQUAL; }
    UBool numeric = settings->isNumeric();

    // Identical-prefix test.
    int32_t equalPrefixLength = 0;
    {
        UChar32 leftUnit;
        UChar32 rightUnit;
        while((leftUnit = left.next(&left)) == (rightUnit = right.next(&right))) {
            if(leftUnit < 0) { return UCOL_EQUAL; }
            ++equalPrefixLength;
        }

        // Back out the code units that differed, for the real collation comparison.
        if(leftUnit >= 0) { left.previous(&left); }
        if(rightUnit >= 0) { right.previous(&right); }

        if(equalPrefixLength > 0) {
            if((leftUnit >= 0 && data->isUnsafeBackward(leftUnit, numeric)) ||
                    (rightUnit >= 0 && data->isUnsafeBackward(rightUnit, numeric))) {
                // Identical prefix: Back up to the start of a contraction or reordering sequence.
                do {
                    --equalPrefixLength;
                    leftUnit = left.previous(&left);
                    right.previous(&right);
                } while(equalPrefixLength > 0 && data->isUnsafeBackward(leftUnit, numeric));
            }
            // See the notes in the UTF-16 version.
        }
    }

    UCollationResult result;
    if(settings->dontCheckFCD()) {
        UIterCollationIterator leftIter(data, numeric, left);
        UIterCollationIterator rightIter(data, numeric, right);
        result = CollationCompare::compareUpToQuaternary(leftIter, rightIter, *settings, errorCode);
    } else {
        FCDUIterCollationIterator leftIter(data, numeric, left, equalPrefixLength);
        FCDUIterCollationIterator rightIter(data, numeric, right, equalPrefixLength);
        result = CollationCompare::compareUpToQuaternary(leftIter, rightIter, *settings, errorCode);
    }
    if(result != UCOL_EQUAL || settings->getStrength() < UCOL_IDENTICAL || U_FAILURE(errorCode)) {
        return result;
    }

    // Compare identical level.
    left.move(&left, equalPrefixLength, UITER_ZERO);
    right.move(&right, equalPrefixLength, UITER_ZERO);
    const Normalizer2Impl &nfcImpl = data->nfcImpl;
    if(settings->dontCheckFCD()) {
        UIterNFDIterator leftIter(left);
        UIterNFDIterator rightIter(right);
        return compareNFDIter(nfcImpl, leftIter, rightIter);
    } else {
        FCDUIterNFDIterator leftIter(data, left, equalPrefixLength);
        FCDUIterNFDIterator rightIter(data, right, equalPrefixLength);
        return compareNFDIter(nfcImpl, leftIter, rightIter);
    }
}

CollationKey &
RuleBasedCollator::getCollationKey(const UnicodeString &s, CollationKey &key,
                                   UErrorCode &errorCode) const {
    return getCollationKey(s.getBuffer(), s.length(), key, errorCode);
}

CollationKey &
RuleBasedCollator::getCollationKey(const UChar *s, int32_t length, CollationKey& key,
                                   UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) {
        return key.setToBogus();
    }
    if(s == NULL && length != 0) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return key.setToBogus();
    }
    key.reset();  // resets the "bogus" state
    CollationKeyByteSink sink(key);
    writeSortKey(s, length, sink, errorCode);
    if(U_FAILURE(errorCode)) {
        key.setToBogus();
    } else if(key.isBogus()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
    } else {
        key.setLength(sink.NumberOfBytesAppended());
    }
    return key;
}

int32_t
RuleBasedCollator::getSortKey(const UnicodeString &s,
                              uint8_t *dest, int32_t capacity) const {
    return getSortKey(s.getBuffer(), s.length(), dest, capacity);
}

int32_t
RuleBasedCollator::getSortKey(const UChar *s, int32_t length,
                              uint8_t *dest, int32_t capacity) const {
    if((s == NULL && length != 0) || capacity < 0 || (dest == NULL && capacity > 0)) {
        return 0;
    }
    uint8_t noDest[1] = { 0 };
    if(dest == NULL) {
        // Distinguish pure preflighting from an allocation error.
        dest = noDest;
        capacity = 0;
    }
    FixedSortKeyByteSink sink(reinterpret_cast<char *>(dest), capacity);
    UErrorCode errorCode = U_ZERO_ERROR;
    writeSortKey(s, length, sink, errorCode);
    return U_SUCCESS(errorCode) ? sink.NumberOfBytesAppended() : 0;
}

void
RuleBasedCollator::writeSortKey(const UChar *s, int32_t length,
                                SortKeyByteSink &sink, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return; }
    const UChar *limit = (length >= 0) ? s + length : NULL;
    UBool numeric = settings->isNumeric();
    CollationKeys::LevelCallback callback;
    if(settings->dontCheckFCD()) {
        UTF16CollationIterator iter(data, numeric, s, s, limit);
        CollationKeys::writeSortKeyUpToQuaternary(iter, data->compressibleBytes, *settings,
                                                  sink, Collation::PRIMARY_LEVEL,
                                                  callback, TRUE, errorCode);
    } else {
        FCDUTF16CollationIterator iter(data, numeric, s, s, limit);
        CollationKeys::writeSortKeyUpToQuaternary(iter, data->compressibleBytes, *settings,
                                                  sink, Collation::PRIMARY_LEVEL,
                                                  callback, TRUE, errorCode);
    }
    if(settings->getStrength() == UCOL_IDENTICAL) {
        writeIdenticalLevel(s, limit, sink, errorCode);
    }
    static const char terminator = 0;  // TERMINATOR_BYTE
    sink.Append(&terminator, 1);
}

void
RuleBasedCollator::writeIdenticalLevel(const UChar *s, const UChar *limit,
                                       SortKeyByteSink &sink, UErrorCode &errorCode) const {
    // NFD quick check
    const UChar *nfdQCYesLimit = data->nfcImpl.decompose(s, limit, NULL, errorCode);
    if(U_FAILURE(errorCode)) { return; }
    sink.Append(Collation::LEVEL_SEPARATOR_BYTE);
    UChar32 prev = 0;
    if(nfdQCYesLimit != s) {
        prev = u_writeIdenticalLevelRun(prev, s, (int32_t)(nfdQCYesLimit - s), sink);
    }
    // Is there non-NFD text?
    int32_t destLengthEstimate;
    if(limit != NULL) {
        if(nfdQCYesLimit == limit) { return; }
        destLengthEstimate = (int32_t)(limit - nfdQCYesLimit);
    } else {
        // s is NUL-terminated
        if(*nfdQCYesLimit == 0) { return; }
        destLengthEstimate = -1;
    }
    UnicodeString nfd;
    data->nfcImpl.decompose(nfdQCYesLimit, limit, nfd, destLengthEstimate, errorCode);
    u_writeIdenticalLevelRun(prev, nfd.getBuffer(), nfd.length(), sink);
}

namespace {

/**
 * internalNextSortKeyPart() calls CollationKeys::writeSortKeyUpToQuaternary()
 * with an instance of this callback class.
 * When another level is about to be written, the callback
 * records the level and the number of bytes that will be written until
 * the sink (which is actually a FixedSortKeyByteSink) fills up.
 *
 * When internalNextSortKeyPart() is called again, it restarts with the last level
 * and ignores as many bytes as were written previously for that level.
 */
class PartLevelCallback : public CollationKeys::LevelCallback {
public:
    PartLevelCallback(const SortKeyByteSink &s)
            : sink(s), level(Collation::PRIMARY_LEVEL) {
        levelCapacity = sink.GetRemainingCapacity();
    }
    virtual ~PartLevelCallback() {}
    virtual UBool needToWrite(Collation::Level l) {
        if(!sink.Overflowed()) {
            // Remember a level that will be at least partially written.
            level = l;
            levelCapacity = sink.GetRemainingCapacity();
            return TRUE;
        } else {
            return FALSE;
        }
    }
    Collation::Level getLevel() const { return level; }
    int32_t getLevelCapacity() const { return levelCapacity; }

private:
    const SortKeyByteSink &sink;
    Collation::Level level;
    int32_t levelCapacity;
};

}  // namespace

int32_t
RuleBasedCollator::internalNextSortKeyPart(UCharIterator *iter, uint32_t state[2],
                                           uint8_t *dest, int32_t count, UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return 0; }
    if(iter == NULL || state == NULL || count < 0 || (count > 0 && dest == NULL)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    if(count == 0) { return 0; }

    FixedSortKeyByteSink sink(reinterpret_cast<char *>(dest), count);
    sink.IgnoreBytes((int32_t)state[1]);
    iter->move(iter, 0, UITER_START);

    Collation::Level level = (Collation::Level)state[0];
    if(level <= Collation::QUATERNARY_LEVEL) {
        UBool numeric = settings->isNumeric();
        PartLevelCallback callback(sink);
        if(settings->dontCheckFCD()) {
            UIterCollationIterator ci(data, numeric, *iter);
            CollationKeys::writeSortKeyUpToQuaternary(ci, data->compressibleBytes, *settings,
                                                      sink, level, callback, FALSE, errorCode);
        } else {
            FCDUIterCollationIterator ci(data, numeric, *iter, 0);
            CollationKeys::writeSortKeyUpToQuaternary(ci, data->compressibleBytes, *settings,
                                                      sink, level, callback, FALSE, errorCode);
        }
        if(U_FAILURE(errorCode)) { return 0; }
        if(sink.NumberOfBytesAppended() > count) {
            state[0] = (uint32_t)callback.getLevel();
            state[1] = (uint32_t)callback.getLevelCapacity();
            return count;
        }
        // All of the normal levels are done.
        if(settings->getStrength() == UCOL_IDENTICAL) {
            level = Collation::IDENTICAL_LEVEL;
            iter->move(iter, 0, UITER_START);
        }
        // else fall through to setting ZERO_LEVEL
    }

    if(level == Collation::IDENTICAL_LEVEL) {
        int32_t levelCapacity = sink.GetRemainingCapacity();
        UnicodeString s;
        for(;;) {
            UChar32 c = iter->next(iter);
            if(c < 0) { break; }
            s.append((UChar)c);
        }
        const UChar *sArray = s.getBuffer();
        writeIdenticalLevel(sArray, sArray + s.length(), sink, errorCode);
        if(U_FAILURE(errorCode)) { return 0; }
        if(sink.NumberOfBytesAppended() > count) {
            state[0] = (uint32_t)level;
            state[1] = (uint32_t)levelCapacity;
            return count;
        }
    }

    // ZERO_LEVEL: Fill the remainder of dest with 00 bytes.
    state[0] = (uint32_t)Collation::ZERO_LEVEL;
    state[1] = 0;
    int32_t length = sink.NumberOfBytesAppended();
    int32_t i = length;
    while(i < count) { dest[i++] = 0; }
    return length;
}

void
RuleBasedCollator::internalGetCEs(const UnicodeString &str, UVector64 &ces,
                                  UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return; }
    const UChar *s = str.getBuffer();
    const UChar *limit = s + str.length();
    UBool numeric = settings->isNumeric();
    if(settings->dontCheckFCD()) {
        UTF16CollationIterator iter(data, numeric, s, s, limit);
        int64_t ce;
        while((ce = iter.nextCE(errorCode)) != Collation::NO_CE) {
            ces.addElement(ce, errorCode);
        }
    } else {
        FCDUTF16CollationIterator iter(data, numeric, s, s, limit);
        int64_t ce;
        while((ce = iter.nextCE(errorCode)) != Collation::NO_CE) {
            ces.addElement(ce, errorCode);
        }
    }
}

namespace {

void appendSubtag(CharString &s, char letter, const char *subtag, int32_t length,
                  UErrorCode &errorCode) {
    if(U_FAILURE(errorCode) || length == 0) { return; }
    if(!s.isEmpty()) {
        s.append('_', errorCode);
    }
    s.append(letter, errorCode);
    for(int32_t i = 0; i < length; ++i) {
        s.append(uprv_toupper(subtag[i]), errorCode);
    }
}

void appendAttribute(CharString &s, char letter, UColAttributeValue value,
                     UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    if(!s.isEmpty()) {
        s.append('_', errorCode);
    }
    static const char *valueChars = "1234...........IXO..SN..LU......";
    s.append(letter, errorCode);
    s.append(valueChars[value], errorCode);
}

}  // namespace

int32_t
RuleBasedCollator::internalGetShortDefinitionString(const char *locale,
                                                    char *buffer, int32_t capacity,
                                                    UErrorCode &errorCode) const {
    if(U_FAILURE(errorCode)) { return 0; }
    if(buffer == NULL ? capacity != 0 : capacity < 0) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    if(locale == NULL) {
        locale = internalGetLocaleID(ULOC_VALID_LOCALE, errorCode);
    }

    char resultLocale[ULOC_FULLNAME_CAPACITY + 1];
    int32_t length = ucol_getFunctionalEquivalent(resultLocale, ULOC_FULLNAME_CAPACITY,
                                                  "collation", locale,
                                                  NULL, &errorCode);
    if(U_FAILURE(errorCode)) { return 0; }
    resultLocale[length] = 0;

    // Append items in alphabetic order of their short definition letters.
    CharString result;
    char subtag[ULOC_KEYWORD_AND_VALUES_CAPACITY];

    if(attributeHasBeenSetExplicitly(UCOL_ALTERNATE_HANDLING)) {
        appendAttribute(result, 'A', getAttribute(UCOL_ALTERNATE_HANDLING, errorCode), errorCode);
    }
    // ATTR_VARIABLE_TOP not supported because 'B' was broken.
    // See ICU tickets #10372 and #10386.
    if(attributeHasBeenSetExplicitly(UCOL_CASE_FIRST)) {
        appendAttribute(result, 'C', getAttribute(UCOL_CASE_FIRST, errorCode), errorCode);
    }
    if(attributeHasBeenSetExplicitly(UCOL_NUMERIC_COLLATION)) {
        appendAttribute(result, 'D', getAttribute(UCOL_NUMERIC_COLLATION, errorCode), errorCode);
    }
    if(attributeHasBeenSetExplicitly(UCOL_CASE_LEVEL)) {
        appendAttribute(result, 'E', getAttribute(UCOL_CASE_LEVEL, errorCode), errorCode);
    }
    if(attributeHasBeenSetExplicitly(UCOL_FRENCH_COLLATION)) {
        appendAttribute(result, 'F', getAttribute(UCOL_FRENCH_COLLATION, errorCode), errorCode);
    }
    // Note: UCOL_HIRAGANA_QUATERNARY_MODE is deprecated and never changes away from default.
    length = uloc_getKeywordValue(resultLocale, "collation", subtag, UPRV_LENGTHOF(subtag), &errorCode);
    appendSubtag(result, 'K', subtag, length, errorCode);
    length = uloc_getLanguage(resultLocale, subtag, UPRV_LENGTHOF(subtag), &errorCode);
    if (length == 0) {
        appendSubtag(result, 'L', "root", 4, errorCode);
    } else {
        appendSubtag(result, 'L', subtag, length, errorCode);
    }
    if(attributeHasBeenSetExplicitly(UCOL_NORMALIZATION_MODE)) {
        appendAttribute(result, 'N', getAttribute(UCOL_NORMALIZATION_MODE, errorCode), errorCode);
    }
    length = uloc_getCountry(resultLocale, subtag, UPRV_LENGTHOF(subtag), &errorCode);
    appendSubtag(result, 'R', subtag, length, errorCode);
    if(attributeHasBeenSetExplicitly(UCOL_STRENGTH)) {
        appendAttribute(result, 'S', getAttribute(UCOL_STRENGTH, errorCode), errorCode);
    }
    length = uloc_getVariant(resultLocale, subtag, UPRV_LENGTHOF(subtag), &errorCode);
    appendSubtag(result, 'V', subtag, length, errorCode);
    length = uloc_getScript(resultLocale, subtag, UPRV_LENGTHOF(subtag), &errorCode);
    appendSubtag(result, 'Z', subtag, length, errorCode);

    if(U_FAILURE(errorCode)) { return 0; }
    if(result.length() <= capacity) {
        uprv_memcpy(buffer, result.data(), result.length());
    }
    return u_terminateChars(buffer, capacity, result.length(), &errorCode);
}

UBool
RuleBasedCollator::isUnsafe(UChar32 c) const {
    return data->isUnsafeBackward(c, settings->isNumeric());
}

void U_CALLCONV
RuleBasedCollator::computeMaxExpansions(const CollationTailoring *t, UErrorCode &errorCode) {
    t->maxExpansions = CollationElementIterator::computeMaxExpansions(t->data, errorCode);
}

UBool
RuleBasedCollator::initMaxExpansions(UErrorCode &errorCode) const {
    umtx_initOnce(tailoring->maxExpansionsInitOnce, computeMaxExpansions, tailoring, errorCode);
    return U_SUCCESS(errorCode);
}

CollationElementIterator *
RuleBasedCollator::createCollationElementIterator(const UnicodeString& source) const {
    UErrorCode errorCode = U_ZERO_ERROR;
    if(!initMaxExpansions(errorCode)) { return NULL; }
    CollationElementIterator *cei = new CollationElementIterator(source, this, errorCode);
    if(U_FAILURE(errorCode)) {
        delete cei;
        return NULL;
    }
    return cei;
}

CollationElementIterator *
RuleBasedCollator::createCollationElementIterator(const CharacterIterator& source) const {
    UErrorCode errorCode = U_ZERO_ERROR;
    if(!initMaxExpansions(errorCode)) { return NULL; }
    CollationElementIterator *cei = new CollationElementIterator(source, this, errorCode);
    if(U_FAILURE(errorCode)) {
        delete cei;
        return NULL;
    }
    return cei;
}

int32_t
RuleBasedCollator::getMaxExpansion(int32_t order) const {
    UErrorCode errorCode = U_ZERO_ERROR;
    (void)initMaxExpansions(errorCode);
    return CollationElementIterator::getMaxExpansion(tailoring->maxExpansions, order);
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2014, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
*/


#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
* Copyright (c) 2014-2016, International Business Machines
* Corporation and others.  All Rights Reserved.
**********************************************************************
*/



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING




/**
 * \file 
 * \brief C++ API: Formats in scientific notation.
 */

U_NAMESPACE_BEGIN

class FieldPositionIterator;
class DecimalFormatSymbols;
class DecimalFormat;
class Formattable;

/**
 * A formatter that formats numbers in user-friendly scientific notation.
 *
 * Sample code:
 * <pre>
 * UErrorCode status = U_ZERO_ERROR;
 * LocalPointer<ScientificNumberFormatter> fmt(
 *         ScientificNumberFormatter::createMarkupInstance(
 *                 "en", "<sup>", "</sup>", status));
 * if (U_FAILURE(status)) {
 *     return;
 * }
 * UnicodeString appendTo;
 * // appendTo = "1.23456x10<sup>-78</sup>"
 * fmt->format(1.23456e-78, appendTo, status);
 * </pre>
 *
 * @stable ICU 55
 */
class U_I18N_API ScientificNumberFormatter : public UObject {
public:

    /**
     * Creates a ScientificNumberFormatter instance that uses
     * superscript characters for exponents.
     * @param fmtToAdopt The DecimalFormat which must be configured for
     *   scientific notation.
     * @param status error returned here.
     * @return The new ScientificNumberFormatter instance.
     *
     * @stable ICU 55
     */
    static ScientificNumberFormatter *createSuperscriptInstance(
            DecimalFormat *fmtToAdopt, UErrorCode &status);

    /**
     * Creates a ScientificNumberFormatter instance that uses
     * superscript characters for exponents for this locale.
     * @param locale The locale
     * @param status error returned here.
     * @return The ScientificNumberFormatter instance.
     *
     * @stable ICU 55
     */
    static ScientificNumberFormatter *createSuperscriptInstance(
            const Locale &locale, UErrorCode &status);


    /**
     * Creates a ScientificNumberFormatter instance that uses
     * markup for exponents.
     * @param fmtToAdopt The DecimalFormat which must be configured for
     *   scientific notation.
     * @param beginMarkup the markup to start superscript.
     * @param endMarkup the markup to end superscript.
     * @param status error returned here.
     * @return The new ScientificNumberFormatter instance.
     *
     * @stable ICU 55
     */
    static ScientificNumberFormatter *createMarkupInstance(
            DecimalFormat *fmtToAdopt,
            const UnicodeString &beginMarkup,
            const UnicodeString &endMarkup,
            UErrorCode &status);

    /**
     * Creates a ScientificNumberFormatter instance that uses
     * markup for exponents for this locale.
     * @param locale The locale
     * @param beginMarkup the markup to start superscript.
     * @param endMarkup the markup to end superscript.
     * @param status error returned here.
     * @return The ScientificNumberFormatter instance.
     *
     * @stable ICU 55
     */
    static ScientificNumberFormatter *createMarkupInstance(
            const Locale &locale,
            const UnicodeString &beginMarkup,
            const UnicodeString &endMarkup,
            UErrorCode &status);


    /**
     * Returns a copy of this object. Caller must free returned copy.
     * @stable ICU 55
     */
    ScientificNumberFormatter *clone() const {
        return new ScientificNumberFormatter(*this);
    }

    /**
     * Destructor.
     * @stable ICU 55
     */
    virtual ~ScientificNumberFormatter();

    /**
     * Formats a number into user friendly scientific notation.
     *
     * @param number the number to format.
     * @param appendTo formatted string appended here.
     * @param status any error returned here.
     * @return appendTo
     *
     * @stable ICU 55
     */
    UnicodeString &format(
            const Formattable &number,
            UnicodeString &appendTo,
            UErrorCode &status) const;
 private:
    class U_I18N_API Style : public UObject {
    public:
        virtual Style *clone() const = 0;
    protected:
        virtual UnicodeString &format(
                const UnicodeString &original,
                FieldPositionIterator &fpi,
                const UnicodeString &preExponent,
                UnicodeString &appendTo,
                UErrorCode &status) const = 0;
    private:
        friend class ScientificNumberFormatter;
    };

    class U_I18N_API SuperscriptStyle : public Style {
    public:
        virtual SuperscriptStyle *clone() const;
    protected:
        virtual UnicodeString &format(
                const UnicodeString &original,
                FieldPositionIterator &fpi,
                const UnicodeString &preExponent,
                UnicodeString &appendTo,
                UErrorCode &status) const;
    };

    class U_I18N_API MarkupStyle : public Style {
    public:
        MarkupStyle(
                const UnicodeString &beginMarkup,
                const UnicodeString &endMarkup)
                : Style(),
                  fBeginMarkup(beginMarkup),
                  fEndMarkup(endMarkup) { }
        virtual MarkupStyle *clone() const;
    protected:
        virtual UnicodeString &format(
                const UnicodeString &original,
                FieldPositionIterator &fpi,
                const UnicodeString &preExponent,
                UnicodeString &appendTo,
                UErrorCode &status) const;
    private:
        UnicodeString fBeginMarkup;
        UnicodeString fEndMarkup;
    };

    ScientificNumberFormatter(
            DecimalFormat *fmtToAdopt,
            Style *styleToAdopt,
            UErrorCode &status);

    ScientificNumberFormatter(const ScientificNumberFormatter &other);
    ScientificNumberFormatter &operator=(const ScientificNumberFormatter &);

    static void getPreExponent(
            const DecimalFormatSymbols &dfs, UnicodeString &preExponent);

    static ScientificNumberFormatter *createInstance(
            DecimalFormat *fmtToAdopt,
            Style *styleToAdopt,
            UErrorCode &status);

    UnicodeString fPreExponent;
    DecimalFormat *fDecimalFormat;
    Style *fStyle;

};

U_NAMESPACE_END


#endif /* !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */









U_NAMESPACE_BEGIN

static const UChar kSuperscriptDigits[] = {
        0x2070,
        0xB9,
        0xB2,
        0xB3,
        0x2074,
        0x2075,
        0x2076,
        0x2077,
        0x2078,
        0x2079};

static const UChar kSuperscriptPlusSign = 0x207A;
static const UChar kSuperscriptMinusSign = 0x207B;

static UBool copyAsSuperscript(
        const UnicodeString &s,
        int32_t beginIndex,
        int32_t endIndex,
        UnicodeString &result,
        UErrorCode &status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    for (int32_t i = beginIndex; i < endIndex;) {
        UChar32 c = s.char32At(i);
        int32_t digit = u_charDigitValue(c);
        if (digit < 0) {
            status = U_INVALID_CHAR_FOUND;
            return FALSE;
        }
        result.append(kSuperscriptDigits[digit]);
        i += U16_LENGTH(c);
    }
    return TRUE;
}

ScientificNumberFormatter *ScientificNumberFormatter::createSuperscriptInstance(
            DecimalFormat *fmtToAdopt, UErrorCode &status) {
    return createInstance(fmtToAdopt, new SuperscriptStyle(), status);
}

ScientificNumberFormatter *ScientificNumberFormatter::createSuperscriptInstance(
            const Locale &locale, UErrorCode &status) {
    return createInstance(
            static_cast<DecimalFormat *>(
                    DecimalFormat::createScientificInstance(locale, status)),
            new SuperscriptStyle(),
            status);
}

ScientificNumberFormatter *ScientificNumberFormatter::createMarkupInstance(
        DecimalFormat *fmtToAdopt,
        const UnicodeString &beginMarkup,
        const UnicodeString &endMarkup,
        UErrorCode &status) {
    return createInstance(
            fmtToAdopt,
            new MarkupStyle(beginMarkup, endMarkup),
            status);
}

ScientificNumberFormatter *ScientificNumberFormatter::createMarkupInstance(
        const Locale &locale,
        const UnicodeString &beginMarkup,
        const UnicodeString &endMarkup,
        UErrorCode &status) {
    return createInstance(
            static_cast<DecimalFormat *>(
                    DecimalFormat::createScientificInstance(locale, status)),
            new MarkupStyle(beginMarkup, endMarkup),
            status);
}

ScientificNumberFormatter *ScientificNumberFormatter::createInstance(
            DecimalFormat *fmtToAdopt,
            Style *styleToAdopt,
            UErrorCode &status) {
    LocalPointer<DecimalFormat> fmt(fmtToAdopt);
    LocalPointer<Style> style(styleToAdopt);
    if (U_FAILURE(status)) {
        return NULL;
    }
    ScientificNumberFormatter *result =
            new ScientificNumberFormatter(
                    fmt.getAlias(),
                    style.getAlias(),
                    status);
    if (result == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    fmt.orphan();
    style.orphan();
    if (U_FAILURE(status)) {
        delete result;
        return NULL;
    }
    return result;
}

ScientificNumberFormatter::SuperscriptStyle *ScientificNumberFormatter::SuperscriptStyle::clone() const {
    return new ScientificNumberFormatter::SuperscriptStyle(*this);
}

UnicodeString &ScientificNumberFormatter::SuperscriptStyle::format(
        const UnicodeString &original,
        FieldPositionIterator &fpi,
        const UnicodeString &preExponent,
        UnicodeString &appendTo,
        UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    FieldPosition fp;
    int32_t copyFromOffset = 0;
    while (fpi.next(fp)) {
        switch (fp.getField()) {
        case UNUM_EXPONENT_SYMBOL_FIELD:
            appendTo.append(
                    original,
                    copyFromOffset,
                    fp.getBeginIndex() - copyFromOffset);
            copyFromOffset = fp.getEndIndex();
            appendTo.append(preExponent);
            break;
        case UNUM_EXPONENT_SIGN_FIELD:
            {
                using namespace icu::numparse::impl;
                int32_t beginIndex = fp.getBeginIndex();
                int32_t endIndex = fp.getEndIndex();
                UChar32 aChar = original.char32At(beginIndex);
                if (unisets::get(unisets::UNISET_KEY_MINUS_SIGN)->contains(aChar)) {
                    appendTo.append(
                            original,
                            copyFromOffset,
                            beginIndex - copyFromOffset);
                    appendTo.append(kSuperscriptMinusSign);
                } else if (unisets::get(unisets::UNISET_KEY_PLUS_SIGN)->contains(aChar)) {
                    appendTo.append(
                           original,
                           copyFromOffset,
                           beginIndex - copyFromOffset);
                    appendTo.append(kSuperscriptPlusSign);
                } else {
                    status = U_INVALID_CHAR_FOUND;
                    return appendTo;
                }
                copyFromOffset = endIndex;
            }
            break;
        case UNUM_EXPONENT_FIELD:
            appendTo.append(
                    original,
                    copyFromOffset,
                    fp.getBeginIndex() - copyFromOffset);
            if (!copyAsSuperscript(
                    original,
                    fp.getBeginIndex(),
                    fp.getEndIndex(),
                    appendTo,
                    status)) {
              return appendTo;
            }
            copyFromOffset = fp.getEndIndex();
            break;
        default:
            break;
        }
    }
    appendTo.append(
            original, copyFromOffset, original.length() - copyFromOffset);
    return appendTo;
}

ScientificNumberFormatter::MarkupStyle *ScientificNumberFormatter::MarkupStyle::clone() const {
    return new ScientificNumberFormatter::MarkupStyle(*this);
}

UnicodeString &ScientificNumberFormatter::MarkupStyle::format(
        const UnicodeString &original,
        FieldPositionIterator &fpi,
        const UnicodeString &preExponent,
        UnicodeString &appendTo,
        UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    FieldPosition fp;
    int32_t copyFromOffset = 0;
    while (fpi.next(fp)) {
        switch (fp.getField()) {
        case UNUM_EXPONENT_SYMBOL_FIELD:
            appendTo.append(
                    original,
                    copyFromOffset,
                    fp.getBeginIndex() - copyFromOffset);
            copyFromOffset = fp.getEndIndex();
            appendTo.append(preExponent);
            appendTo.append(fBeginMarkup);
            break;
        case UNUM_EXPONENT_FIELD:
            appendTo.append(
                    original,
                    copyFromOffset,
                    fp.getEndIndex() - copyFromOffset);
            copyFromOffset = fp.getEndIndex();
            appendTo.append(fEndMarkup);
            break;
        default:
            break;
        }
    }
    appendTo.append(
            original, copyFromOffset, original.length() - copyFromOffset);
    return appendTo;
}

ScientificNumberFormatter::ScientificNumberFormatter(
        DecimalFormat *fmtToAdopt, Style *styleToAdopt, UErrorCode &status)
        : fPreExponent(),
          fDecimalFormat(fmtToAdopt),
          fStyle(styleToAdopt) {
    if (U_FAILURE(status)) {
        return;
    }
    if (fDecimalFormat == NULL || fStyle == NULL) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    const DecimalFormatSymbols *sym = fDecimalFormat->getDecimalFormatSymbols();
    if (sym == NULL) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    getPreExponent(*sym, fPreExponent);
}

ScientificNumberFormatter::ScientificNumberFormatter(
        const ScientificNumberFormatter &other)
        : UObject(other),
          fPreExponent(other.fPreExponent),
          fDecimalFormat(NULL),
          fStyle(NULL) {
    fDecimalFormat = static_cast<DecimalFormat *>(
            other.fDecimalFormat->clone());
    fStyle = other.fStyle->clone();
}

ScientificNumberFormatter::~ScientificNumberFormatter() {
    delete fDecimalFormat;
    delete fStyle;
}

UnicodeString &ScientificNumberFormatter::format(
        const Formattable &number,
        UnicodeString &appendTo,
        UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    UnicodeString original;
    FieldPositionIterator fpi;
    fDecimalFormat->format(number, original, &fpi, status);
    return fStyle->format(
            original,
            fpi,
            fPreExponent,
            appendTo,
            status);
}

void ScientificNumberFormatter::getPreExponent(
        const DecimalFormatSymbols &dfs, UnicodeString &preExponent) {
    preExponent.append(dfs.getConstSymbol(
            DecimalFormatSymbols::kExponentMultiplicationSymbol));
    preExponent.append(dfs.getConstSymbol(DecimalFormatSymbols::kOneDigitSymbol));
    preExponent.append(dfs.getConstSymbol(DecimalFormatSymbols::kZeroDigitSymbol));
}

U_NAMESPACE_END

#endif /* !UCONFIG_NO_FORMATTING */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 2001-2008,2010 IBM and others. All rights reserved.
// **********************************************************************
// *   Date        Name        Description
// *  03/22/2000   helena      Creation.
// **********************************************************************
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_COLLATION && !UCONFIG_NO_BREAK_ITERATION

// // #include "unicode/brkiter.h"
// #include "unicode/schriter.h"
// #include "unicode/search.h"
// #include "usrchimp.h"
// #include "cmemory.h"

// // public constructors and destructors -----------------------------------
// U_NAMESPACE_BEGIN

// SearchIterator::SearchIterator(const SearchIterator &other)
//     : UObject(other)
// {
//     m_breakiterator_            = other.m_breakiterator_;
//     m_text_                     = other.m_text_;
//     m_search_                   = (USearch *)uprv_malloc(sizeof(USearch));
//     m_search_->breakIter        = other.m_search_->breakIter;
//     m_search_->isCanonicalMatch = other.m_search_->isCanonicalMatch;
//     m_search_->isOverlap        = other.m_search_->isOverlap;
//     m_search_->elementComparisonType = other.m_search_->elementComparisonType;
//     m_search_->matchedIndex     = other.m_search_->matchedIndex;
//     m_search_->matchedLength    = other.m_search_->matchedLength;
//     m_search_->text             = other.m_search_->text;
//     m_search_->textLength       = other.m_search_->textLength;
// }

// SearchIterator::~SearchIterator()
// {
//     if (m_search_ != NULL) {
//         uprv_free(m_search_);
//     }
// }

// // public get and set methods ----------------------------------------

// void SearchIterator::setAttribute(USearchAttribute       attribute,
//                                   USearchAttributeValue  value,
//                                   UErrorCode            &status)
// {
//     if (U_SUCCESS(status)) {
//         switch (attribute)
//         {
//         case USEARCH_OVERLAP :
//             m_search_->isOverlap = (value == USEARCH_ON ? TRUE : FALSE);
//             break;
//         case USEARCH_CANONICAL_MATCH :
//             m_search_->isCanonicalMatch = (value == USEARCH_ON ? TRUE : FALSE);
//             break;
//         case USEARCH_ELEMENT_COMPARISON :
//             if (value == USEARCH_PATTERN_BASE_WEIGHT_IS_WILDCARD || value == USEARCH_ANY_BASE_WEIGHT_IS_WILDCARD) {
//                 m_search_->elementComparisonType = (int16_t)value;
//             } else {
//                 m_search_->elementComparisonType = 0;
//             }
//             break;
//         default:
//             status = U_ILLEGAL_ARGUMENT_ERROR;
//         }
//     }
//     if (value == USEARCH_ATTRIBUTE_VALUE_COUNT) {
//         status = U_ILLEGAL_ARGUMENT_ERROR;
//     }
// }

// USearchAttributeValue SearchIterator::getAttribute(
//                                           USearchAttribute  attribute) const
// {
//     switch (attribute) {
//     case USEARCH_OVERLAP :
//         return (m_search_->isOverlap == TRUE ? USEARCH_ON : USEARCH_OFF);
//     case USEARCH_CANONICAL_MATCH :
//         return (m_search_->isCanonicalMatch == TRUE ? USEARCH_ON :
//                                                                 USEARCH_OFF);
//     case USEARCH_ELEMENT_COMPARISON :
//         {
//             int16_t value = m_search_->elementComparisonType;
//             if (value == USEARCH_PATTERN_BASE_WEIGHT_IS_WILDCARD || value == USEARCH_ANY_BASE_WEIGHT_IS_WILDCARD) {
//                 return (USearchAttributeValue)value;
//             } else {
//                 return USEARCH_STANDARD_ELEMENT_COMPARISON;
//             }
//         }
//     default :
//         return USEARCH_DEFAULT;
//     }
// }

// int32_t SearchIterator::getMatchedStart() const
// {
//     return m_search_->matchedIndex;
// }

// int32_t SearchIterator::getMatchedLength() const
// {
//     return m_search_->matchedLength;
// }

// void SearchIterator::getMatchedText(UnicodeString &result) const
// {
//     int32_t matchedindex  = m_search_->matchedIndex;
//     int32_t     matchedlength = m_search_->matchedLength;
//     if (matchedindex != USEARCH_DONE && matchedlength != 0) {
//         result.setTo(m_search_->text + matchedindex, matchedlength);
//     }
//     else {
//         result.remove();
//     }
// }

// void SearchIterator::setBreakIterator(BreakIterator *breakiter,
//                                       UErrorCode &status)
// {
//     if (U_SUCCESS(status)) {
// #if 0
//         m_search_->breakIter = NULL;
//         // the c++ breakiterator may not make use of ubreakiterator.
//         // so we'll have to keep track of it ourselves.
// #else
//         // Well, gee... the Constructors that take a BreakIterator
//         // all cast the BreakIterator to a UBreakIterator and
//         // pass it to the corresponding usearch_openFromXXX
//         // routine, so there's no reason not to do this.
//         //
//         // Besides, a UBreakIterator is a BreakIterator, so
//         // any subclass of BreakIterator should work fine here...
//         m_search_->breakIter = (UBreakIterator *) breakiter;
// #endif

//         m_breakiterator_ = breakiter;
//     }
// }

// const BreakIterator * SearchIterator::getBreakIterator(void) const
// {
//     return m_breakiterator_;
// }

// void SearchIterator::setText(const UnicodeString &text, UErrorCode &status)
// {
//     if (U_SUCCESS(status)) {
//         if (text.length() == 0) {
//             status = U_ILLEGAL_ARGUMENT_ERROR;
//         }
//         else {
//             m_text_        = text;
//             m_search_->text = m_text_.getBuffer();
//             m_search_->textLength = m_text_.length();
//         }
//     }
// }

// void SearchIterator::setText(CharacterIterator &text, UErrorCode &status)
// {
//     if (U_SUCCESS(status)) {
//         text.getText(m_text_);
//         setText(m_text_, status);
//     }
// }

// const UnicodeString & SearchIterator::getText(void) const
// {
//     return m_text_;
// }

// // operator overloading ----------------------------------------------

// UBool SearchIterator::operator==(const SearchIterator &that) const
// {
//     if (this == &that) {
//         return TRUE;
//     }
//     return (m_breakiterator_            == that.m_breakiterator_ &&
//             m_search_->isCanonicalMatch == that.m_search_->isCanonicalMatch &&
//             m_search_->isOverlap        == that.m_search_->isOverlap &&
//             m_search_->elementComparisonType == that.m_search_->elementComparisonType &&
//             m_search_->matchedIndex     == that.m_search_->matchedIndex &&
//             m_search_->matchedLength    == that.m_search_->matchedLength &&
//             m_search_->textLength       == that.m_search_->textLength &&
//             getOffset() == that.getOffset() &&
//             (uprv_memcmp(m_search_->text, that.m_search_->text,
//                               m_search_->textLength * sizeof(UChar)) == 0));
// }

// // public methods ----------------------------------------------------

// int32_t SearchIterator::first(UErrorCode &status)
// {
//     if (U_FAILURE(status)) {
//         return USEARCH_DONE;
//     }
//     setOffset(0, status);
//     return handleNext(0, status);
// }

// int32_t SearchIterator::following(int32_t position,
//                                       UErrorCode &status)
// {
//     if (U_FAILURE(status)) {
//         return USEARCH_DONE;
//     }
//     setOffset(position, status);
//     return handleNext(position, status);
// }

// int32_t SearchIterator::last(UErrorCode &status)
// {
//     if (U_FAILURE(status)) {
//         return USEARCH_DONE;
//     }
//     setOffset(m_search_->textLength, status);
//     return handlePrev(m_search_->textLength, status);
// }

// int32_t SearchIterator::preceding(int32_t position,
//                                       UErrorCode &status)
// {
//     if (U_FAILURE(status)) {
//         return USEARCH_DONE;
//     }
//     setOffset(position, status);
//     return handlePrev(position, status);
// }

// int32_t SearchIterator::next(UErrorCode &status)
// {
//     if (U_SUCCESS(status)) {
//         int32_t offset = getOffset();
//         int32_t matchindex  = m_search_->matchedIndex;
//         int32_t     matchlength = m_search_->matchedLength;
//         m_search_->reset = FALSE;
//         if (m_search_->isForwardSearching == TRUE) {
//             int32_t textlength = m_search_->textLength;
//             if (offset == textlength || matchindex == textlength ||
//                 (matchindex != USEARCH_DONE &&
//                 matchindex + matchlength >= textlength)) {
//                 // not enough characters to match
//                 setMatchNotFound();
//                 return USEARCH_DONE;
//             }
//         }
//         else {
//             // switching direction.
//             // if matchedIndex == USEARCH_DONE, it means that either a
//             // setOffset has been called or that previous ran off the text
//             // string. the iterator would have been set to offset 0 if a
//             // match is not found.
//             m_search_->isForwardSearching = TRUE;
//             if (m_search_->matchedIndex != USEARCH_DONE) {
//                 // there's no need to set the collation element iterator
//                 // the next call to next will set the offset.
//                 return matchindex;
//             }
//         }

//         if (matchlength > 0) {
//             // if matchlength is 0 we are at the start of the iteration
//             if (m_search_->isOverlap) {
//                 offset ++;
//             }
//             else {
//                 offset += matchlength;
//             }
//         }
//         return handleNext(offset, status);
//     }
//     return USEARCH_DONE;
// }

// int32_t SearchIterator::previous(UErrorCode &status)
// {
//     if (U_SUCCESS(status)) {
//         int32_t offset;
//         if (m_search_->reset) {
//             offset                       = m_search_->textLength;
//             m_search_->isForwardSearching = FALSE;
//             m_search_->reset              = FALSE;
//             setOffset(offset, status);
//         }
//         else {
//             offset = getOffset();
//         }

//         int32_t matchindex = m_search_->matchedIndex;
//         if (m_search_->isForwardSearching == TRUE) {
//             // switching direction.
//             // if matchedIndex == USEARCH_DONE, it means that either a
//             // setOffset has been called or that next ran off the text
//             // string. the iterator would have been set to offset textLength if
//             // a match is not found.
//             m_search_->isForwardSearching = FALSE;
//             if (matchindex != USEARCH_DONE) {
//                 return matchindex;
//             }
//         }
//         else {
//             if (offset == 0 || matchindex == 0) {
//                 // not enough characters to match
//                 setMatchNotFound();
//                 return USEARCH_DONE;
//             }
//         }

//         if (matchindex != USEARCH_DONE) {
//             if (m_search_->isOverlap) {
//                 matchindex += m_search_->matchedLength - 2;
//             }

//             return handlePrev(matchindex, status);
//         }

//         return handlePrev(offset, status);
//     }

//     return USEARCH_DONE;
// }

// void SearchIterator::reset()
// {
//     UErrorCode status = U_ZERO_ERROR;
//     setMatchNotFound();
//     setOffset(0, status);
//     m_search_->isOverlap          = FALSE;
//     m_search_->isCanonicalMatch   = FALSE;
//     m_search_->elementComparisonType = 0;
//     m_search_->isForwardSearching = TRUE;
//     m_search_->reset              = TRUE;
// }

// // protected constructors and destructors -----------------------------

// SearchIterator::SearchIterator()
// {
//     m_search_                     = (USearch *)uprv_malloc(sizeof(USearch));
//     m_search_->breakIter          = NULL;
//     m_search_->isOverlap          = FALSE;
//     m_search_->isCanonicalMatch   = FALSE;
//     m_search_->elementComparisonType = 0;
//     m_search_->isForwardSearching = TRUE;
//     m_search_->reset              = TRUE;
//     m_search_->matchedIndex       = USEARCH_DONE;
//     m_search_->matchedLength      = 0;
//     m_search_->text               = NULL;
//     m_search_->textLength         = 0;
//     m_breakiterator_              = NULL;
// }

// SearchIterator::SearchIterator(const UnicodeString &text,
//                                      BreakIterator *breakiter) :
//                                      m_breakiterator_(breakiter),
//                                      m_text_(text)
// {
//     m_search_                     = (USearch *)uprv_malloc(sizeof(USearch));
//     m_search_->breakIter          = NULL;
//     m_search_->isOverlap          = FALSE;
//     m_search_->isCanonicalMatch   = FALSE;
//     m_search_->elementComparisonType = 0;
//     m_search_->isForwardSearching = TRUE;
//     m_search_->reset              = TRUE;
//     m_search_->matchedIndex       = USEARCH_DONE;
//     m_search_->matchedLength      = 0;
//     m_search_->text               = m_text_.getBuffer();
//     m_search_->textLength         = text.length();
// }

// SearchIterator::SearchIterator(CharacterIterator &text,
//                                BreakIterator     *breakiter) :
//                                m_breakiterator_(breakiter)
// {
//     m_search_                     = (USearch *)uprv_malloc(sizeof(USearch));
//     m_search_->breakIter          = NULL;
//     m_search_->isOverlap          = FALSE;
//     m_search_->isCanonicalMatch   = FALSE;
//     m_search_->elementComparisonType = 0;
//     m_search_->isForwardSearching = TRUE;
//     m_search_->reset              = TRUE;
//     m_search_->matchedIndex       = USEARCH_DONE;
//     m_search_->matchedLength      = 0;
//     text.getText(m_text_);
//     m_search_->text               = m_text_.getBuffer();
//     m_search_->textLength         = m_text_.length();
//     m_breakiterator_             = breakiter;
// }

// // protected methods ------------------------------------------------------

// SearchIterator & SearchIterator::operator=(const SearchIterator &that)
// {
//     if (this != &that) {
//         m_breakiterator_            = that.m_breakiterator_;
//         m_text_                     = that.m_text_;
//         m_search_->breakIter        = that.m_search_->breakIter;
//         m_search_->isCanonicalMatch = that.m_search_->isCanonicalMatch;
//         m_search_->isOverlap        = that.m_search_->isOverlap;
//         m_search_->elementComparisonType = that.m_search_->elementComparisonType;
//         m_search_->matchedIndex     = that.m_search_->matchedIndex;
//         m_search_->matchedLength    = that.m_search_->matchedLength;
//         m_search_->text             = that.m_search_->text;
//         m_search_->textLength       = that.m_search_->textLength;
//     }
//     return *this;
// }

// void SearchIterator::setMatchLength(int32_t length)
// {
//     m_search_->matchedLength = length;
// }

// void SearchIterator::setMatchStart(int32_t position)
// {
//     m_search_->matchedIndex = position;
// }

// void SearchIterator::setMatchNotFound()
// {
//     setMatchStart(USEARCH_DONE);
//     setMatchLength(0);
//     UErrorCode status = U_ZERO_ERROR;
//     // by default no errors should be returned here since offsets are within
//     // range.
//     if (m_search_->isForwardSearching) {
//         setOffset(m_search_->textLength, status);
//     }
//     else {
//         setOffset(0, status);
//     }
// }


// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_COLLATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/********************************************************************
 * COPYRIGHT:
 * Copyright (c) 1997-2012, International Business Machines Corporation and
 * others. All Rights Reserved.
 * Copyright (C) 2010 , Yahoo! Inc.
 ********************************************************************
 *
 * File SELFMT.CPP
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   11/11/09    kirtig      Finished first cut of implementation.
 *   11/16/09    kirtig      Improved version
 ********************************************************************/

  // for 'typeid' to work





// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 1999-2009, International Business Machines
// *   Corporation and others.  All Rights Reserved.
// **********************************************************************
//  *
//  *
//  *   ucnv_err.h:
//  */

// /**
//  * \file
//  * \brief C UConverter predefined error callbacks
//  *
//  *  <h2>Error Behaviour Functions</h2>
//  *  Defines some error behaviour functions called by ucnv_{from,to}Unicode
//  *  These are provided as part of ICU and many are stable, but they
//  *  can also be considered only as an example of what can be done with
//  *  callbacks.  You may of course write your own.
//  *
//  *  If you want to write your own, you may also find the functions from
//  *  ucnv_cb.h useful when writing your own callbacks.
//  *
//  *  These functions, although public, should NEVER be called directly.
//  *  They should be used as parameters to the ucnv_setFromUCallback
//  *  and ucnv_setToUCallback functions, to set the behaviour of a converter
//  *  when it encounters ILLEGAL/UNMAPPED/INVALID sequences.
//  *
//  *  usage example:  'STOP' doesn't need any context, but newContext
//  *    could be set to something other than 'NULL' if needed. The available
//  *    contexts in this header can modify the default behavior of the callback.
//  *
//  *  \code
//  *  UErrorCode err = U_ZERO_ERROR;
//  *  UConverter *myConverter = ucnv_open("ibm-949", &err);
//  *  const void *oldContext;
//  *  UConverterFromUCallback oldAction;
//  *
//  *
//  *  if (U_SUCCESS(err))
//  *  {
//  *      ucnv_setFromUCallBack(myConverter,
//  *                       UCNV_FROM_U_CALLBACK_STOP,
//  *                       NULL,
//  *                       &oldAction,
//  *                       &oldContext,
//  *                       &status);
//  *  }
//  *  \endcode
//  *
//  *  The code above tells "myConverter" to stop when it encounters an
//  *  ILLEGAL/TRUNCATED/INVALID sequences when it is used to convert from
//  *  Unicode -> Codepage. The behavior from Codepage to Unicode is not changed,
//  *  and ucnv_setToUCallBack would need to be called in order to change
//  *  that behavior too.
//  *
//  *  Here is an example with a context:
//  *
//  *  \code
//  *  UErrorCode err = U_ZERO_ERROR;
//  *  UConverter *myConverter = ucnv_open("ibm-949", &err);
//  *  const void *oldContext;
//  *  UConverterFromUCallback oldAction;
//  *
//  *
//  *  if (U_SUCCESS(err))
//  *  {
//  *      ucnv_setToUCallBack(myConverter,
//  *                       UCNV_TO_U_CALLBACK_SUBSTITUTE,
//  *                       UCNV_SUB_STOP_ON_ILLEGAL,
//  *                       &oldAction,
//  *                       &oldContext,
//  *                       &status);
//  *  }
//  *  \endcode
//  *
//  *  The code above tells "myConverter" to stop when it encounters an
//  *  ILLEGAL/TRUNCATED/INVALID sequences when it is used to convert from
//  *  Codepage -> Unicode. Any unmapped and legal characters will be
//  *  substituted to be the default substitution character.
//  */

// #ifndef UCNV_ERR_H
// #define UCNV_ERR_H

// 

// #if !UCONFIG_NO_CONVERSION

// /** Forward declaring the UConverter structure. @stable ICU 2.0 */
// struct UConverter;

// /** @stable ICU 2.0 */
// typedef struct UConverter UConverter;

// /**
//  * FROM_U, TO_U context options for sub callback
//  * @stable ICU 2.0
//  */
// #define UCNV_SUB_STOP_ON_ILLEGAL "i"

// /**
//  * FROM_U, TO_U context options for skip callback
//  * @stable ICU 2.0
//  */
// #define UCNV_SKIP_STOP_ON_ILLEGAL "i"

// /**
//  * FROM_U_CALLBACK_ESCAPE context option to escape the code unit according to ICU (%UXXXX)
//  * @stable ICU 2.0
//  */
// #define UCNV_ESCAPE_ICU       NULL
// /**
//  * FROM_U_CALLBACK_ESCAPE context option to escape the code unit according to JAVA (\\uXXXX)
//  * @stable ICU 2.0
//  */
// #define UCNV_ESCAPE_JAVA      "J"
// /**
//  * FROM_U_CALLBACK_ESCAPE context option to escape the code unit according to C (\\uXXXX \\UXXXXXXXX)
//  * TO_U_CALLBACK_ESCAPE option to escape the character value according to C (\\xXXXX)
//  * @stable ICU 2.0
//  */
// #define UCNV_ESCAPE_C         "C"
// /**
//  * FROM_U_CALLBACK_ESCAPE context option to escape the code unit according to XML Decimal escape \htmlonly(&amp;#DDDD;)\endhtmlonly
//  * TO_U_CALLBACK_ESCAPE context option to escape the character value according to XML Decimal escape \htmlonly(&amp;#DDDD;)\endhtmlonly
//  * @stable ICU 2.0
//  */
// #define UCNV_ESCAPE_XML_DEC   "D"
// /**
//  * FROM_U_CALLBACK_ESCAPE context option to escape the code unit according to XML Hex escape \htmlonly(&amp;#xXXXX;)\endhtmlonly
//  * TO_U_CALLBACK_ESCAPE context option to escape the character value according to XML Hex escape \htmlonly(&amp;#xXXXX;)\endhtmlonly
//  * @stable ICU 2.0
//  */
// #define UCNV_ESCAPE_XML_HEX   "X"
// /**
//  * FROM_U_CALLBACK_ESCAPE context option to escape the code unit according to Unicode (U+XXXXX)
//  * @stable ICU 2.0
//  */
// #define UCNV_ESCAPE_UNICODE   "U"

// /**
//  * FROM_U_CALLBACK_ESCAPE context option to escape the code unit according to CSS2 conventions (\\HH..H<space>, that is,
//  * a backslash, 1..6 hex digits, and a space)
//  * @stable ICU 4.0
//  */
// #define UCNV_ESCAPE_CSS2   "S"

// /**
//  * The process condition code to be used with the callbacks.
//  * Codes which are greater than UCNV_IRREGULAR should be
//  * passed on to any chained callbacks.
//  * @stable ICU 2.0
//  */
// typedef enum {
//     UCNV_UNASSIGNED = 0,  /**< The code point is unassigned.
//                              The error code U_INVALID_CHAR_FOUND will be set. */
//     UCNV_ILLEGAL = 1,     /**< The code point is illegal. For example,
//                              \\x81\\x2E is illegal in SJIS because \\x2E
//                              is not a valid trail byte for the \\x81
//                              lead byte.
//                              Also, starting with Unicode 3.0.1, non-shortest byte sequences
//                              in UTF-8 (like \\xC1\\xA1 instead of \\x61 for U+0061)
//                              are also illegal, not just irregular.
//                              The error code U_ILLEGAL_CHAR_FOUND will be set. */
//     UCNV_IRREGULAR = 2,   /**< The codepoint is not a regular sequence in
//                              the encoding. For example, \\xED\\xA0\\x80..\\xED\\xBF\\xBF
//                              are irregular UTF-8 byte sequences for single surrogate
//                              code points.
//                              The error code U_INVALID_CHAR_FOUND will be set. */
//     UCNV_RESET = 3,       /**< The callback is called with this reason when a
//                              'reset' has occurred. Callback should reset all
//                              state. */
//     UCNV_CLOSE = 4,        /**< Called when the converter is closed. The
//                              callback should release any allocated memory.*/
//     UCNV_CLONE = 5         /**< Called when ucnv_safeClone() is called on the
//                               converter. the pointer available as the
//                               'context' is an alias to the original converters'
//                               context pointer. If the context must be owned
//                               by the new converter, the callback must clone
//                               the data and call ucnv_setFromUCallback
//                               (or setToUCallback) with the correct pointer.
//                               @stable ICU 2.2
//                            */
// } UConverterCallbackReason;


// /**
//  * The structure for the fromUnicode callback function parameter.
//  * @stable ICU 2.0
//  */
// typedef struct {
//     uint16_t size;              /**< The size of this struct. @stable ICU 2.0 */
//     UBool flush;                /**< The internal state of converter will be reset and data flushed if set to TRUE. @stable ICU 2.0    */
//     UConverter *converter;      /**< Pointer to the converter that is opened and to which this struct is passed as an argument. @stable ICU 2.0  */
//     const UChar *source;        /**< Pointer to the source source buffer. @stable ICU 2.0    */
//     const UChar *sourceLimit;   /**< Pointer to the limit (end + 1) of source buffer. @stable ICU 2.0    */
//     char *target;               /**< Pointer to the target buffer. @stable ICU 2.0    */
//     const char *targetLimit;    /**< Pointer to the limit (end + 1) of target buffer. @stable ICU 2.0     */
//     int32_t *offsets;           /**< Pointer to the buffer that receives the offsets. *offset = blah ; offset++;. @stable ICU 2.0  */
// } UConverterFromUnicodeArgs;


// /**
//  * The structure for the toUnicode callback function parameter.
//  * @stable ICU 2.0
//  */
// typedef struct {
//     uint16_t size;              /**< The size of this struct   @stable ICU 2.0 */
//     UBool flush;                /**< The internal state of converter will be reset and data flushed if set to TRUE. @stable ICU 2.0   */
//     UConverter *converter;      /**< Pointer to the converter that is opened and to which this struct is passed as an argument. @stable ICU 2.0 */
//     const char *source;         /**< Pointer to the source source buffer. @stable ICU 2.0    */
//     const char *sourceLimit;    /**< Pointer to the limit (end + 1) of source buffer. @stable ICU 2.0    */
//     UChar *target;              /**< Pointer to the target buffer. @stable ICU 2.0    */
//     const UChar *targetLimit;   /**< Pointer to the limit (end + 1) of target buffer. @stable ICU 2.0     */
//     int32_t *offsets;           /**< Pointer to the buffer that receives the offsets. *offset = blah ; offset++;. @stable ICU 2.0  */
// } UConverterToUnicodeArgs;


// /**
//  * DO NOT CALL THIS FUNCTION DIRECTLY!
//  * This From Unicode callback STOPS at the ILLEGAL_SEQUENCE,
//  * returning the error code back to the caller immediately.
//  *
//  * @param context Pointer to the callback's private data
//  * @param fromUArgs Information about the conversion in progress
//  * @param codeUnits Points to 'length' UChars of the concerned Unicode sequence
//  * @param length Size (in bytes) of the concerned codepage sequence
//  * @param codePoint Single UChar32 (UTF-32) containing the concerend Unicode codepoint.
//  * @param reason Defines the reason the callback was invoked
//  * @param err This should always be set to a failure status prior to calling.
//  * @stable ICU 2.0
//  */
// U_STABLE void U_EXPORT2 UCNV_FROM_U_CALLBACK_STOP (
//                   const void *context,
//                   UConverterFromUnicodeArgs *fromUArgs,
//                   const UChar* codeUnits,
//                   int32_t length,
//                   UChar32 codePoint,
//                   UConverterCallbackReason reason,
//                   UErrorCode * err);



// /**
//  * DO NOT CALL THIS FUNCTION DIRECTLY!
//  * This To Unicode callback STOPS at the ILLEGAL_SEQUENCE,
//  * returning the error code back to the caller immediately.
//  *
//  * @param context Pointer to the callback's private data
//  * @param toUArgs Information about the conversion in progress
//  * @param codeUnits Points to 'length' bytes of the concerned codepage sequence
//  * @param length Size (in bytes) of the concerned codepage sequence
//  * @param reason Defines the reason the callback was invoked
//  * @param err This should always be set to a failure status prior to calling.
//  * @stable ICU 2.0
//  */
// U_STABLE void U_EXPORT2 UCNV_TO_U_CALLBACK_STOP (
//                   const void *context,
//                   UConverterToUnicodeArgs *toUArgs,
//                   const char* codeUnits,
//                   int32_t length,
//                   UConverterCallbackReason reason,
//                   UErrorCode * err);

// /**
//  * DO NOT CALL THIS FUNCTION DIRECTLY!
//  * This From Unicode callback skips any ILLEGAL_SEQUENCE, or
//  * skips only UNASSINGED_SEQUENCE depending on the context parameter
//  * simply ignoring those characters.
//  *
//  * @param context  The function currently recognizes the callback options:
//  *                 UCNV_SKIP_STOP_ON_ILLEGAL: STOPS at the ILLEGAL_SEQUENCE,
//  *                      returning the error code back to the caller immediately.
//  *                 NULL: Skips any ILLEGAL_SEQUENCE
//  * @param fromUArgs Information about the conversion in progress
//  * @param codeUnits Points to 'length' UChars of the concerned Unicode sequence
//  * @param length Size (in bytes) of the concerned codepage sequence
//  * @param codePoint Single UChar32 (UTF-32) containing the concerend Unicode codepoint.
//  * @param reason Defines the reason the callback was invoked
//  * @param err Return value will be set to success if the callback was handled,
//  *      otherwise this value will be set to a failure status.
//  * @stable ICU 2.0
//  */
// U_STABLE void U_EXPORT2 UCNV_FROM_U_CALLBACK_SKIP (
//                   const void *context,
//                   UConverterFromUnicodeArgs *fromUArgs,
//                   const UChar* codeUnits,
//                   int32_t length,
//                   UChar32 codePoint,
//                   UConverterCallbackReason reason,
//                   UErrorCode * err);

// /**
//  * DO NOT CALL THIS FUNCTION DIRECTLY!
//  * This From Unicode callback will Substitute the ILLEGAL SEQUENCE, or
//  * UNASSIGNED_SEQUENCE depending on context parameter, with the
//  * current substitution string for the converter. This is the default
//  * callback.
//  *
//  * @param context The function currently recognizes the callback options:
//  *                 UCNV_SUB_STOP_ON_ILLEGAL: STOPS at the ILLEGAL_SEQUENCE,
//  *                      returning the error code back to the caller immediately.
//  *                 NULL: Substitutes any ILLEGAL_SEQUENCE
//  * @param fromUArgs Information about the conversion in progress
//  * @param codeUnits Points to 'length' UChars of the concerned Unicode sequence
//  * @param length Size (in bytes) of the concerned codepage sequence
//  * @param codePoint Single UChar32 (UTF-32) containing the concerend Unicode codepoint.
//  * @param reason Defines the reason the callback was invoked
//  * @param err Return value will be set to success if the callback was handled,
//  *      otherwise this value will be set to a failure status.
//  * @see ucnv_setSubstChars
//  * @stable ICU 2.0
//  */
// U_STABLE void U_EXPORT2 UCNV_FROM_U_CALLBACK_SUBSTITUTE (
//                   const void *context,
//                   UConverterFromUnicodeArgs *fromUArgs,
//                   const UChar* codeUnits,
//                   int32_t length,
//                   UChar32 codePoint,
//                   UConverterCallbackReason reason,
//                   UErrorCode * err);

// /**
//  * DO NOT CALL THIS FUNCTION DIRECTLY!
//  * This From Unicode callback will Substitute the ILLEGAL SEQUENCE with the
//  * hexadecimal representation of the illegal codepoints
//  *
//  * @param context The function currently recognizes the callback options:
//  *        <ul>
//  *        <li>UCNV_ESCAPE_ICU: Substitues the  ILLEGAL SEQUENCE with the hexadecimal
//  *          representation in the format  %UXXXX, e.g. "%uFFFE%u00AC%uC8FE").
//  *          In the Event the converter doesn't support the characters {%,U}[A-F][0-9],
//  *          it will  substitute  the illegal sequence with the substitution characters.
//  *          Note that  codeUnit(32bit int eg: unit of a surrogate pair) is represented as
//  *          %UD84D%UDC56</li>
//  *        <li>UCNV_ESCAPE_JAVA: Substitues the  ILLEGAL SEQUENCE with the hexadecimal
//  *          representation in the format  \\uXXXX, e.g. "\\uFFFE\\u00AC\\uC8FE").
//  *          In the Event the converter doesn't support the characters {\,u}[A-F][0-9],
//  *          it will  substitute  the illegal sequence with the substitution characters.
//  *          Note that  codeUnit(32bit int eg: unit of a surrogate pair) is represented as
//  *          \\uD84D\\uDC56</li>
//  *        <li>UCNV_ESCAPE_C: Substitues the  ILLEGAL SEQUENCE with the hexadecimal
//  *          representation in the format  \\uXXXX, e.g. "\\uFFFE\\u00AC\\uC8FE").
//  *          In the Event the converter doesn't support the characters {\,u,U}[A-F][0-9],
//  *          it will  substitute  the illegal sequence with the substitution characters.
//  *          Note that  codeUnit(32bit int eg: unit of a surrogate pair) is represented as
//  *          \\U00023456</li>
//  *        <li>UCNV_ESCAPE_XML_DEC: Substitues the  ILLEGAL SEQUENCE with the decimal
//  *          representation in the format \htmlonly&amp;#DDDDDDDD;, e.g. "&amp;#65534;&amp;#172;&amp;#51454;")\endhtmlonly.
//  *          In the Event the converter doesn't support the characters {&amp;,#}[0-9],
//  *          it will  substitute  the illegal sequence with the substitution characters.
//  *          Note that  codeUnit(32bit int eg: unit of a surrogate pair) is represented as
//  *          &amp;#144470; and Zero padding is ignored.</li>
//  *        <li>UCNV_ESCAPE_XML_HEX:Substitues the  ILLEGAL SEQUENCE with the decimal
//  *          representation in the format \htmlonly&amp;#xXXXX; e.g. "&amp;#xFFFE;&amp;#x00AC;&amp;#xC8FE;")\endhtmlonly.
//  *          In the Event the converter doesn't support the characters {&,#,x}[0-9],
//  *          it will  substitute  the illegal sequence with the substitution characters.
//  *          Note that  codeUnit(32bit int eg: unit of a surrogate pair) is represented as
//  *          \htmlonly&amp;#x23456;\endhtmlonly</li>
//  *        </ul>
//  * @param fromUArgs Information about the conversion in progress
//  * @param codeUnits Points to 'length' UChars of the concerned Unicode sequence
//  * @param length Size (in bytes) of the concerned codepage sequence
//  * @param codePoint Single UChar32 (UTF-32) containing the concerend Unicode codepoint.
//  * @param reason Defines the reason the callback was invoked
//  * @param err Return value will be set to success if the callback was handled,
//  *      otherwise this value will be set to a failure status.
//  * @stable ICU 2.0
//  */
// U_STABLE void U_EXPORT2 UCNV_FROM_U_CALLBACK_ESCAPE (
//                   const void *context,
//                   UConverterFromUnicodeArgs *fromUArgs,
//                   const UChar* codeUnits,
//                   int32_t length,
//                   UChar32 codePoint,
//                   UConverterCallbackReason reason,
//                   UErrorCode * err);


// /**
//  * DO NOT CALL THIS FUNCTION DIRECTLY!
//  * This To Unicode callback skips any ILLEGAL_SEQUENCE, or
//  * skips only UNASSINGED_SEQUENCE depending on the context parameter
//  * simply ignoring those characters.
//  *
//  * @param context  The function currently recognizes the callback options:
//  *                 UCNV_SKIP_STOP_ON_ILLEGAL: STOPS at the ILLEGAL_SEQUENCE,
//  *                      returning the error code back to the caller immediately.
//  *                 NULL: Skips any ILLEGAL_SEQUENCE
//  * @param toUArgs Information about the conversion in progress
//  * @param codeUnits Points to 'length' bytes of the concerned codepage sequence
//  * @param length Size (in bytes) of the concerned codepage sequence
//  * @param reason Defines the reason the callback was invoked
//  * @param err Return value will be set to success if the callback was handled,
//  *      otherwise this value will be set to a failure status.
//  * @stable ICU 2.0
//  */
// U_STABLE void U_EXPORT2 UCNV_TO_U_CALLBACK_SKIP (
//                   const void *context,
//                   UConverterToUnicodeArgs *toUArgs,
//                   const char* codeUnits,
//                   int32_t length,
//                   UConverterCallbackReason reason,
//                   UErrorCode * err);

// /**
//  * DO NOT CALL THIS FUNCTION DIRECTLY!
//  * This To Unicode callback will Substitute the ILLEGAL SEQUENCE,or
//  * UNASSIGNED_SEQUENCE depending on context parameter,  with the
//  * Unicode substitution character, U+FFFD.
//  *
//  * @param context  The function currently recognizes the callback options:
//  *                 UCNV_SUB_STOP_ON_ILLEGAL: STOPS at the ILLEGAL_SEQUENCE,
//  *                      returning the error code back to the caller immediately.
//  *                 NULL: Substitutes any ILLEGAL_SEQUENCE
//  * @param toUArgs Information about the conversion in progress
//  * @param codeUnits Points to 'length' bytes of the concerned codepage sequence
//  * @param length Size (in bytes) of the concerned codepage sequence
//  * @param reason Defines the reason the callback was invoked
//  * @param err Return value will be set to success if the callback was handled,
//  *      otherwise this value will be set to a failure status.
//  * @stable ICU 2.0
//  */
// U_STABLE void U_EXPORT2 UCNV_TO_U_CALLBACK_SUBSTITUTE (
//                   const void *context,
//                   UConverterToUnicodeArgs *toUArgs,
//                   const char* codeUnits,
//                   int32_t length,
//                   UConverterCallbackReason reason,
//                   UErrorCode * err);

// /**
//  * DO NOT CALL THIS FUNCTION DIRECTLY!
//  * This To Unicode callback will Substitute the ILLEGAL SEQUENCE with the
//  * hexadecimal representation of the illegal bytes
//  *  (in the format  %XNN, e.g. "%XFF%X0A%XC8%X03").
//  *
//  * @param context This function currently recognizes the callback options:
//  *      UCNV_ESCAPE_ICU, UCNV_ESCAPE_JAVA, UCNV_ESCAPE_C, UCNV_ESCAPE_XML_DEC,
//  *      UCNV_ESCAPE_XML_HEX and UCNV_ESCAPE_UNICODE.
//  * @param toUArgs Information about the conversion in progress
//  * @param codeUnits Points to 'length' bytes of the concerned codepage sequence
//  * @param length Size (in bytes) of the concerned codepage sequence
//  * @param reason Defines the reason the callback was invoked
//  * @param err Return value will be set to success if the callback was handled,
//  *      otherwise this value will be set to a failure status.
//  * @stable ICU 2.0
//  */

// U_STABLE void U_EXPORT2 UCNV_TO_U_CALLBACK_ESCAPE (
//                   const void *context,
//                   UConverterToUnicodeArgs *toUArgs,
//                   const char* codeUnits,
//                   int32_t length,
//                   UConverterCallbackReason reason,
//                   UErrorCode * err);

// #endif

// #endif

// /*UCNV_ERR_H*/







// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/********************************************************************
 * COPYRIGHT:
 * Copyright (c) 1997-2011, International Business Machines Corporation and
 * others. All Rights Reserved.
 * Copyright (C) 2010 , Yahoo! Inc.
 ********************************************************************
 * File selectfmtimpl.h
 *
 *   Date        Name        Description
 *   11/11/09    kirtig      Finished first cut of implementation.
 *********************************************************************/



#if !UCONFIG_NO_FORMATTING








U_NAMESPACE_BEGIN

#define selfmtimpl_DOT               ((UChar)0x002E)
#define selfmtimpl_SINGLE_QUOTE      ((UChar)0x0027)
#define selfmtimpl_SLASH             ((UChar)0x002F)
#define selfmtimpl_BACKSLASH         ((UChar)0x005C)
#define selfmtimpl_SPACE             ((UChar)0x0020)
#define selfmtimpl_TAB               ((UChar)0x0009)
#define selfmtimpl_QUOTATION_MARK    ((UChar)0x0022)
#define selfmtimpl_ASTERISK          ((UChar)0x002A)
#define selfmtimpl_COMMA             ((UChar)0x002C)
#define selfmtimpl_HYPHEN            ((UChar)0x002D)
#define selfmtimpl_U_ZERO            ((UChar)0x0030)
#define selfmtimpl_U_ONE             ((UChar)0x0031)
#define selfmtimpl_U_TWO             ((UChar)0x0032)
#define selfmtimpl_U_THREE           ((UChar)0x0033)
#define selfmtimpl_U_FOUR            ((UChar)0x0034)
#define selfmtimpl_U_FIVE            ((UChar)0x0035)
#define selfmtimpl_U_SIX             ((UChar)0x0036)
#define selfmtimpl_U_SEVEN           ((UChar)0x0037)
#define selfmtimpl_U_EIGHT           ((UChar)0x0038)
#define selfmtimpl_U_NINE            ((UChar)0x0039)
#define selfmtimpl_COLON             ((UChar)0x003A)
#define selfmtimpl_SEMI_COLON        ((UChar)0x003B)
#define selfmtimpl_CAP_A             ((UChar)0x0041)
#define selfmtimpl_CAP_B             ((UChar)0x0042)
#define selfmtimpl_CAP_R             ((UChar)0x0052)
#define selfmtimpl_CAP_Z             ((UChar)0x005A)
#define selfmtimpl_LOWLINE           ((UChar)0x005F)
#define selfmtimpl_LEFTBRACE         ((UChar)0x007B)
#define selfmtimpl_RIGHTBRACE        ((UChar)0x007D)

#define selfmtimpl_LOW_A             ((UChar)0x0061)
#define selfmtimpl_LOW_B             ((UChar)0x0062)
#define selfmtimpl_LOW_C             ((UChar)0x0063)
#define selfmtimpl_LOW_D             ((UChar)0x0064)
#define selfmtimpl_LOW_E             ((UChar)0x0065)
#define selfmtimpl_LOW_F             ((UChar)0x0066)
#define selfmtimpl_LOW_G             ((UChar)0x0067)
#define selfmtimpl_LOW_H             ((UChar)0x0068)
#define selfmtimpl_LOW_I             ((UChar)0x0069)
#define selfmtimpl_LOW_J             ((UChar)0x006a)
#define selfmtimpl_LOW_K             ((UChar)0x006B)
#define selfmtimpl_LOW_L             ((UChar)0x006C)
#define selfmtimpl_LOW_M             ((UChar)0x006D)
#define selfmtimpl_LOW_N             ((UChar)0x006E)
#define selfmtimpl_LOW_O             ((UChar)0x006F)
#define selfmtimpl_LOW_P             ((UChar)0x0070)
#define selfmtimpl_LOW_Q             ((UChar)0x0071)
#define selfmtimpl_LOW_R             ((UChar)0x0072)
#define selfmtimpl_LOW_S             ((UChar)0x0073)
#define selfmtimpl_LOW_T             ((UChar)0x0074)
#define selfmtimpl_LOW_U             ((UChar)0x0075)
#define selfmtimpl_LOW_V             ((UChar)0x0076)
#define selfmtimpl_LOW_W             ((UChar)0x0077)
#define selfmtimpl_LOW_X             ((UChar)0x0078)
#define selfmtimpl_LOW_Y             ((UChar)0x0079)
#define selfmtimpl_LOW_Z             ((UChar)0x007A)

class UnicodeSet;

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */







#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(SelectFormat)

static const UChar SELECT_KEYWORD_OTHER[] = {selfmtimpl_LOW_O, selfmtimpl_LOW_T, selfmtimpl_LOW_H, selfmtimpl_LOW_E, selfmtimpl_LOW_R, 0};

SelectFormat::SelectFormat(const UnicodeString& pat,
                           UErrorCode& status) : msgPattern(status) {
   applyPattern(pat, status);
}

SelectFormat::SelectFormat(const SelectFormat& other) : Format(other),
                                                        msgPattern(other.msgPattern) {
}

SelectFormat::~SelectFormat() {
}

void
SelectFormat::applyPattern(const UnicodeString& newPattern, UErrorCode& status) {
    if (U_FAILURE(status)) {
      return;
    }

    msgPattern.parseSelectStyle(newPattern, NULL, status);
    if (U_FAILURE(status)) {
        msgPattern.clear();
    }
}

UnicodeString&
SelectFormat::format(const Formattable& obj,
                   UnicodeString& appendTo,
                   FieldPosition& pos,
                   UErrorCode& status) const
{
    if (U_FAILURE(status)) {
        return appendTo;
    }
    if (obj.getType() == Formattable::kString) {
        return format(obj.getString(status), appendTo, pos, status);
    } else {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return appendTo;
    }
}

UnicodeString&
SelectFormat::format(const UnicodeString& keyword,
                     UnicodeString& appendTo,
                     FieldPosition& /*pos */,
                     UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    // Check for the validity of the keyword
    if (!PatternProps::isIdentifier(keyword.getBuffer(), keyword.length())) {
        status = U_ILLEGAL_ARGUMENT_ERROR;  // Invalid formatting argument.
    }
    if (msgPattern.countParts() == 0) {
        status = U_INVALID_STATE_ERROR;
        return appendTo;
    }
    int32_t msgStart = findSubMessage(msgPattern, 0, keyword, status);
    if (!MessageImpl::jdkAposMode(msgPattern)) {
        int32_t patternStart = msgPattern.getPart(msgStart).getLimit();
        int32_t msgLimit = msgPattern.getLimitPartIndex(msgStart);
        appendTo.append(msgPattern.getPatternString(),
                        patternStart,
                        msgPattern.getPatternIndex(msgLimit) - patternStart);
        return appendTo;
    }
    // JDK compatibility mode: Remove SKIP_SYNTAX.
    return MessageImpl::appendSubMessageWithoutSkipSyntax(msgPattern, msgStart, appendTo);
}

UnicodeString&
SelectFormat::toPattern(UnicodeString& appendTo) {
    if (0 == msgPattern.countParts()) {
        appendTo.setToBogus();
    } else {
        appendTo.append(msgPattern.getPatternString());
    }
    return appendTo;
}


int32_t SelectFormat::findSubMessage(const MessagePattern& pattern, int32_t partIndex,
                                     const UnicodeString& keyword, UErrorCode& ec) {
    if (U_FAILURE(ec)) {
        return 0;
    }
    UnicodeString other(FALSE, SELECT_KEYWORD_OTHER, 5);
    int32_t count = pattern.countParts();
    int32_t msgStart=0;
    // Iterate over (ARG_SELECTOR, message) pairs until ARG_LIMIT or end of select-only pattern.
    do {
        const MessagePattern::Part& part=pattern.getPart(partIndex++);
        const UMessagePatternPartType type=part.getType();
        if(type==UMSGPAT_PART_TYPE_ARG_LIMIT) {
            break;
        }
        // part is an ARG_SELECTOR followed by a message
        if(pattern.partSubstringMatches(part, keyword)) {
            // keyword matches
            return partIndex;
        } else if(msgStart==0 && pattern.partSubstringMatches(part, other)) {
            msgStart=partIndex;
        }
        partIndex=pattern.getLimitPartIndex(partIndex);
    } while(++partIndex<count);
    return msgStart;
}

SelectFormat* SelectFormat::clone() const
{
    return new SelectFormat(*this);
}

SelectFormat&
SelectFormat::operator=(const SelectFormat& other) {
    if (this != &other) {
        msgPattern = other.msgPattern;
    }
    return *this;
}

UBool
SelectFormat::operator==(const Format& other) const {
    if (this == &other) {
        return TRUE;
    }
    if (!Format::operator==(other)) {
        return FALSE;
    }
    const SelectFormat& o = (const SelectFormat&)other;
    return msgPattern == o.msgPattern;
}

UBool
SelectFormat::operator!=(const Format& other) const {
    return  !operator==(other);
}

void
SelectFormat::parseObject(const UnicodeString& /*source*/,
                        Formattable& /*result*/,
                        ParsePosition& pos) const
{
    // Parsing not supported.
    pos.setErrorIndex(pos.getIndex());
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 1997-2013, International Business Machines Corporation and
 * others. All Rights Reserved.
 *******************************************************************************
 *
 * File SIMPLETZ.H
 *
 * Modification History:
 *
 *   Date        Name        Description
 *   12/05/96    clhuang     Creation.
 *   04/21/97    aliu        Fixed miscellaneous bugs found by inspection and
 *                           testing.
 *   07/29/97    aliu        Ported source bodies back from Java version with
 *                           numerous feature enhancements and bug fixes.
 *   08/10/98    stephen     JDK 1.2 sync.
 *   09/17/98    stephen     Fixed getOffset() for last hour of year and DST
 *   12/02/99    aliu        Added TimeMode and constructor and setStart/EndRule
 *                           methods that take TimeMode. Whitespace cleanup.
 ********************************************************************************
 */

  // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING









U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(SimpleTimeZone)

// Use only for decodeStartRule() and decodeEndRule() where the year is not
// available. Set February to 29 days to accomodate rules with that date
// and day-of-week-on-or-before-that-date mode (DOW_LE_DOM_MODE).
// The compareToRule() method adjusts to February 28 in non-leap years.
//
// For actual getOffset() calculations, use Grego::monthLength() and
// Grego::previousMonthLength() which take leap years into account.
// We handle leap years assuming always
// Gregorian, since we know they didn't have daylight time when
// Gregorian calendar started.
const int8_t SimpleTimeZone::STATICMONTHLENGTH[] = {31,29,31,30,31,30,31,31,30,31,30,31};

static const UChar DST_STR[] = {0x0028,0x0044,0x0053,0x0054,0x0029,0}; // "(DST)"
static const UChar STD_STR[] = {0x0028,0x0053,0x0054,0x0044,0x0029,0}; // "(STD)"


// *****************************************************************************
// class SimpleTimeZone
// *****************************************************************************


SimpleTimeZone::SimpleTimeZone(int32_t rawOffsetGMT, const UnicodeString& ID)
:   BasicTimeZone(ID),
    startMonth(0),
    startDay(0),
    startDayOfWeek(0),
    startTime(0),
    startTimeMode(WALL_TIME),
    endTimeMode(WALL_TIME),
    endMonth(0),
    endDay(0),
    endDayOfWeek(0),
    endTime(0),
    startYear(0),
    rawOffset(rawOffsetGMT),
    useDaylight(FALSE),
    startMode(DOM_MODE),
    endMode(DOM_MODE),
    dstSavings(U_MILLIS_PER_HOUR)
{
    clearTransitionRules();
}

// -------------------------------------

SimpleTimeZone::SimpleTimeZone(int32_t rawOffsetGMT, const UnicodeString& ID,
    int8_t savingsStartMonth, int8_t savingsStartDay,
    int8_t savingsStartDayOfWeek, int32_t savingsStartTime,
    int8_t savingsEndMonth, int8_t savingsEndDay,
    int8_t savingsEndDayOfWeek, int32_t savingsEndTime,
    UErrorCode& status)
:   BasicTimeZone(ID)
{
    clearTransitionRules();
    construct(rawOffsetGMT,
              savingsStartMonth, savingsStartDay, savingsStartDayOfWeek,
              savingsStartTime, WALL_TIME,
              savingsEndMonth, savingsEndDay, savingsEndDayOfWeek,
              savingsEndTime, WALL_TIME,
              U_MILLIS_PER_HOUR, status);
}

// -------------------------------------

SimpleTimeZone::SimpleTimeZone(int32_t rawOffsetGMT, const UnicodeString& ID,
    int8_t savingsStartMonth, int8_t savingsStartDay,
    int8_t savingsStartDayOfWeek, int32_t savingsStartTime,
    int8_t savingsEndMonth, int8_t savingsEndDay,
    int8_t savingsEndDayOfWeek, int32_t savingsEndTime,
    int32_t savingsDST, UErrorCode& status)
:   BasicTimeZone(ID)
{
    clearTransitionRules();
    construct(rawOffsetGMT,
              savingsStartMonth, savingsStartDay, savingsStartDayOfWeek,
              savingsStartTime, WALL_TIME,
              savingsEndMonth, savingsEndDay, savingsEndDayOfWeek,
              savingsEndTime, WALL_TIME,
              savingsDST, status);
}

// -------------------------------------

SimpleTimeZone::SimpleTimeZone(int32_t rawOffsetGMT, const UnicodeString& ID,
    int8_t savingsStartMonth, int8_t savingsStartDay,
    int8_t savingsStartDayOfWeek, int32_t savingsStartTime,
    TimeMode savingsStartTimeMode,
    int8_t savingsEndMonth, int8_t savingsEndDay,
    int8_t savingsEndDayOfWeek, int32_t savingsEndTime,
    TimeMode savingsEndTimeMode,
    int32_t savingsDST, UErrorCode& status)
:   BasicTimeZone(ID)
{
    clearTransitionRules();
    construct(rawOffsetGMT,
              savingsStartMonth, savingsStartDay, savingsStartDayOfWeek,
              savingsStartTime, savingsStartTimeMode,
              savingsEndMonth, savingsEndDay, savingsEndDayOfWeek,
              savingsEndTime, savingsEndTimeMode,
              savingsDST, status);
}

/**
 * Internal construction method.
 */
void SimpleTimeZone::construct(int32_t rawOffsetGMT,
                               int8_t savingsStartMonth,
                               int8_t savingsStartDay,
                               int8_t savingsStartDayOfWeek,
                               int32_t savingsStartTime,
                               TimeMode savingsStartTimeMode,
                               int8_t savingsEndMonth,
                               int8_t savingsEndDay,
                               int8_t savingsEndDayOfWeek,
                               int32_t savingsEndTime,
                               TimeMode savingsEndTimeMode,
                               int32_t savingsDST,
                               UErrorCode& status)
{
    this->rawOffset      = rawOffsetGMT;
    this->startMonth     = savingsStartMonth;
    this->startDay       = savingsStartDay;
    this->startDayOfWeek = savingsStartDayOfWeek;
    this->startTime      = savingsStartTime;
    this->startTimeMode  = savingsStartTimeMode;
    this->endMonth       = savingsEndMonth;
    this->endDay         = savingsEndDay;
    this->endDayOfWeek   = savingsEndDayOfWeek;
    this->endTime        = savingsEndTime;
    this->endTimeMode    = savingsEndTimeMode;
    this->dstSavings     = savingsDST;
    this->startYear      = 0;
    this->startMode      = DOM_MODE;
    this->endMode        = DOM_MODE;

    decodeRules(status);

    if (savingsDST == 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

// -------------------------------------

SimpleTimeZone::~SimpleTimeZone()
{
    deleteTransitionRules();
}

// -------------------------------------

// Called by TimeZone::createDefault(), then clone() inside a Mutex - be careful.
SimpleTimeZone::SimpleTimeZone(const SimpleTimeZone &source)
:   BasicTimeZone(source)
{
    *this = source;
}

// -------------------------------------

// Called by TimeZone::createDefault(), then clone() inside a Mutex - be careful.
SimpleTimeZone &
SimpleTimeZone::operator=(const SimpleTimeZone &right)
{
    if (this != &right)
    {
        TimeZone::operator=(right);
        rawOffset      = right.rawOffset;
        startMonth     = right.startMonth;
        startDay       = right.startDay;
        startDayOfWeek = right.startDayOfWeek;
        startTime      = right.startTime;
        startTimeMode  = right.startTimeMode;
        startMode      = right.startMode;
        endMonth       = right.endMonth;
        endDay         = right.endDay;
        endDayOfWeek   = right.endDayOfWeek;
        endTime        = right.endTime;
        endTimeMode    = right.endTimeMode;
        endMode        = right.endMode;
        startYear      = right.startYear;
        dstSavings     = right.dstSavings;
        useDaylight    = right.useDaylight;
        clearTransitionRules();
    }
    return *this;
}

// -------------------------------------

UBool
SimpleTimeZone::operator==(const TimeZone& that) const
{
    return ((this == &that) ||
            (typeid(*this) == typeid(that) &&
            TimeZone::operator==(that) &&
            hasSameRules(that)));
}

// -------------------------------------

// Called by TimeZone::createDefault() inside a Mutex - be careful.
SimpleTimeZone*
SimpleTimeZone::clone() const
{
    return new SimpleTimeZone(*this);
}

// -------------------------------------

/**
 * Sets the daylight savings starting year, that is, the year this time zone began
 * observing its specified daylight savings time rules.  The time zone is considered
 * not to observe daylight savings time prior to that year; SimpleTimeZone doesn't
 * support historical daylight-savings-time rules.
 * @param year the daylight savings starting year.
 */
void
SimpleTimeZone::setStartYear(int32_t year)
{
    startYear = year;
    transitionRulesInitialized = FALSE;
}

// -------------------------------------

/**
 * Sets the daylight savings starting rule. For example, in the U.S., Daylight Savings
 * Time starts at the first Sunday in April, at 2 AM in standard time.
 * Therefore, you can set the start rule by calling:
 * setStartRule(TimeFields.APRIL, 1, TimeFields.SUNDAY, 2*60*60*1000);
 * The dayOfWeekInMonth and dayOfWeek parameters together specify how to calculate
 * the exact starting date.  Their exact meaning depend on their respective signs,
 * allowing various types of rules to be constructed, as follows:<ul>
 *   <li>If both dayOfWeekInMonth and dayOfWeek are positive, they specify the
 *       day of week in the month (e.g., (2, WEDNESDAY) is the second Wednesday
 *       of the month).
 *   <li>If dayOfWeek is positive and dayOfWeekInMonth is negative, they specify
 *       the day of week in the month counting backward from the end of the month.
 *       (e.g., (-1, MONDAY) is the last Monday in the month)
 *   <li>If dayOfWeek is zero and dayOfWeekInMonth is positive, dayOfWeekInMonth
 *       specifies the day of the month, regardless of what day of the week it is.
 *       (e.g., (10, 0) is the tenth day of the month)
 *   <li>If dayOfWeek is zero and dayOfWeekInMonth is negative, dayOfWeekInMonth
 *       specifies the day of the month counting backward from the end of the
 *       month, regardless of what day of the week it is (e.g., (-2, 0) is the
 *       next-to-last day of the month).
 *   <li>If dayOfWeek is negative and dayOfWeekInMonth is positive, they specify the
 *       first specified day of the week on or after the specfied day of the month.
 *       (e.g., (15, -SUNDAY) is the first Sunday after the 15th of the month
 *       [or the 15th itself if the 15th is a Sunday].)
 *   <li>If dayOfWeek and DayOfWeekInMonth are both negative, they specify the
 *       last specified day of the week on or before the specified day of the month.
 *       (e.g., (-20, -TUESDAY) is the last Tuesday before the 20th of the month
 *       [or the 20th itself if the 20th is a Tuesday].)</ul>
 * @param month the daylight savings starting month. Month is 0-based.
 * eg, 0 for January.
 * @param dayOfWeekInMonth the daylight savings starting
 * day-of-week-in-month. Please see the member description for an example.
 * @param dayOfWeek the daylight savings starting day-of-week. Please see
 * the member description for an example.
 * @param time the daylight savings starting time. Please see the member
 * description for an example.
 */
 
void
SimpleTimeZone::setStartRule(int32_t month, int32_t dayOfWeekInMonth, int32_t dayOfWeek,
                             int32_t time, TimeMode mode, UErrorCode& status)
{
    startMonth     = (int8_t)month;
    startDay       = (int8_t)dayOfWeekInMonth;
    startDayOfWeek = (int8_t)dayOfWeek;
    startTime      = time;
    startTimeMode  = mode;
    decodeStartRule(status);
    transitionRulesInitialized = FALSE;
}

// -------------------------------------

void 
SimpleTimeZone::setStartRule(int32_t month, int32_t dayOfMonth, 
                             int32_t time, TimeMode mode, UErrorCode& status) 
{
    setStartRule(month, dayOfMonth, 0, time, mode, status);
}

// -------------------------------------

void 
SimpleTimeZone::setStartRule(int32_t month, int32_t dayOfMonth, int32_t dayOfWeek, 
                             int32_t time, TimeMode mode, UBool after, UErrorCode& status)
{
    setStartRule(month, after ? dayOfMonth : -dayOfMonth,
                 -dayOfWeek, time, mode, status);
}

// -------------------------------------

/**
 * Sets the daylight savings ending rule. For example, in the U.S., Daylight
 * Savings Time ends at the last (-1) Sunday in October, at 2 AM in standard time.
 * Therefore, you can set the end rule by calling:
 * setEndRule(TimeFields.OCTOBER, -1, TimeFields.SUNDAY, 2*60*60*1000);
 * Various other types of rules can be specified by manipulating the dayOfWeek
 * and dayOfWeekInMonth parameters.  For complete details, see the documentation
 * for setStartRule().
 * @param month the daylight savings ending month. Month is 0-based.
 * eg, 0 for January.
 * @param dayOfWeekInMonth the daylight savings ending
 * day-of-week-in-month. See setStartRule() for a complete explanation.
 * @param dayOfWeek the daylight savings ending day-of-week. See setStartRule()
 * for a complete explanation.
 * @param time the daylight savings ending time. Please see the member
 * description for an example.
 */

void
SimpleTimeZone::setEndRule(int32_t month, int32_t dayOfWeekInMonth, int32_t dayOfWeek,
                           int32_t time, TimeMode mode, UErrorCode& status)
{
    endMonth     = (int8_t)month;
    endDay       = (int8_t)dayOfWeekInMonth;
    endDayOfWeek = (int8_t)dayOfWeek;
    endTime      = time;
    endTimeMode  = mode;
    decodeEndRule(status);
    transitionRulesInitialized = FALSE;
}

// -------------------------------------

void 
SimpleTimeZone::setEndRule(int32_t month, int32_t dayOfMonth, 
                           int32_t time, TimeMode mode, UErrorCode& status)
{
    setEndRule(month, dayOfMonth, 0, time, mode, status);
}

// -------------------------------------

void 
SimpleTimeZone::setEndRule(int32_t month, int32_t dayOfMonth, int32_t dayOfWeek, 
                           int32_t time, TimeMode mode, UBool after, UErrorCode& status)
{
    setEndRule(month, after ? dayOfMonth : -dayOfMonth,
               -dayOfWeek, time, mode, status);
}

// -------------------------------------

int32_t
SimpleTimeZone::getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                          uint8_t dayOfWeek, int32_t millis, UErrorCode& status) const
{
    // Check the month before calling Grego::monthLength(). This
    // duplicates the test that occurs in the 7-argument getOffset(),
    // however, this is unavoidable. We don't mind because this method, in
    // fact, should not be called; internal code should always call the
    // 7-argument getOffset(), and outside code should use Calendar.get(int
    // field) with fields ZONE_OFFSET and DST_OFFSET. We can't get rid of
    // this method because it's public API. - liu 8/10/98
    if(month < UCAL_JANUARY || month > UCAL_DECEMBER) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    return getOffset(era, year, month, day, dayOfWeek, millis, Grego::monthLength(year, month), status);
}

int32_t 
SimpleTimeZone::getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                          uint8_t dayOfWeek, int32_t millis, 
                          int32_t /*monthLength*/, UErrorCode& status) const
{
    // Check the month before calling Grego::monthLength(). This
    // duplicates a test that occurs in the 9-argument getOffset(),
    // however, this is unavoidable. We don't mind because this method, in
    // fact, should not be called; internal code should always call the
    // 9-argument getOffset(), and outside code should use Calendar.get(int
    // field) with fields ZONE_OFFSET and DST_OFFSET. We can't get rid of
    // this method because it's public API. - liu 8/10/98
    if (month < UCAL_JANUARY
        || month > UCAL_DECEMBER) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    // We ignore monthLength because it can be derived from year and month.
    // This is so that February in leap years is calculated correctly.
    // We keep this argument in this function for backwards compatibility.
    return getOffset(era, year, month, day, dayOfWeek, millis,
                     Grego::monthLength(year, month),
                     Grego::previousMonthLength(year, month),
                     status);
}

int32_t 
SimpleTimeZone::getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                          uint8_t dayOfWeek, int32_t millis, 
                          int32_t monthLength, int32_t prevMonthLength,
                          UErrorCode& status) const
{
    if(U_FAILURE(status)) return 0;

    if ((era != GregorianCalendar::AD && era != GregorianCalendar::BC)
        || month < UCAL_JANUARY
        || month > UCAL_DECEMBER
        || day < 1
        || day > monthLength
        || dayOfWeek < UCAL_SUNDAY
        || dayOfWeek > UCAL_SATURDAY
        || millis < 0
        || millis >= U_MILLIS_PER_DAY
        || monthLength < 28
        || monthLength > 31
        || prevMonthLength < 28
        || prevMonthLength > 31) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    int32_t result = rawOffset;

    // Bail out if we are before the onset of daylight savings time
    if(!useDaylight || year < startYear || era != GregorianCalendar::AD) 
        return result;

    // Check for southern hemisphere.  We assume that the start and end
    // month are different.
    UBool southern = (startMonth > endMonth);

    // Compare the date to the starting and ending rules.+1 = date>rule, -1
    // = date<rule, 0 = date==rule.
    int32_t startCompare = compareToRule((int8_t)month, (int8_t)monthLength, (int8_t)prevMonthLength,
                                         (int8_t)day, (int8_t)dayOfWeek, millis,
                                         startTimeMode == UTC_TIME ? -rawOffset : 0,
                                         startMode, (int8_t)startMonth, (int8_t)startDayOfWeek,
                                         (int8_t)startDay, startTime);
    int32_t endCompare = 0;

    /* We don't always have to compute endCompare.  For many instances,
     * startCompare is enough to determine if we are in DST or not.  In the
     * northern hemisphere, if we are before the start rule, we can't have
     * DST.  In the southern hemisphere, if we are after the start rule, we
     * must have DST.  This is reflected in the way the next if statement
     * (not the one immediately following) short circuits. */
    if(southern != (startCompare >= 0)) {
        endCompare = compareToRule((int8_t)month, (int8_t)monthLength, (int8_t)prevMonthLength,
                                   (int8_t)day, (int8_t)dayOfWeek, millis,
                                   endTimeMode == WALL_TIME ? dstSavings :
                                    (endTimeMode == UTC_TIME ? -rawOffset : 0),
                                   endMode, (int8_t)endMonth, (int8_t)endDayOfWeek,
                                   (int8_t)endDay, endTime);
    }

    // Check for both the northern and southern hemisphere cases.  We
    // assume that in the northern hemisphere, the start rule is before the
    // end rule within the calendar year, and vice versa for the southern
    // hemisphere.
    if ((!southern && (startCompare >= 0 && endCompare < 0)) ||
        (southern && (startCompare >= 0 || endCompare < 0)))
        result += dstSavings;

    return result;
}

void
SimpleTimeZone::getOffsetFromLocal(UDate date, int32_t nonExistingTimeOpt, int32_t duplicatedTimeOpt,
                                   int32_t& rawOffsetGMT, int32_t& savingsDST, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }

    rawOffsetGMT = getRawOffset();
    int32_t year, month, dom, dow;
    double day = uprv_floor(date / U_MILLIS_PER_DAY);
    int32_t millis = (int32_t) (date - day * U_MILLIS_PER_DAY);

    Grego::dayToFields(day, year, month, dom, dow);

    savingsDST = getOffset(GregorianCalendar::AD, year, month, dom,
                          (uint8_t) dow, millis,
                          Grego::monthLength(year, month),
                          status) - rawOffsetGMT;
    if (U_FAILURE(status)) {
        return;
    }

    UBool recalc = FALSE;

    // Now we need some adjustment
    if (savingsDST > 0) {
        if ((nonExistingTimeOpt & kStdDstMask) == kStandard
            || ((nonExistingTimeOpt & kStdDstMask) != kDaylight && (nonExistingTimeOpt & kFormerLatterMask) != kLatter)) {
            date -= getDSTSavings();
            recalc = TRUE;
        }
    } else {
        if ((duplicatedTimeOpt & kStdDstMask) == kDaylight
                || ((duplicatedTimeOpt & kStdDstMask) != kStandard && (duplicatedTimeOpt & kFormerLatterMask) == kFormer)) {
            date -= getDSTSavings();
            recalc = TRUE;
        }
    }
    if (recalc) {
        day = uprv_floor(date / U_MILLIS_PER_DAY);
        millis = (int32_t) (date - day * U_MILLIS_PER_DAY);
        Grego::dayToFields(day, year, month, dom, dow);
        savingsDST = getOffset(GregorianCalendar::AD, year, month, dom,
                          (uint8_t) dow, millis,
                          Grego::monthLength(year, month),
                          status) - rawOffsetGMT;
    }
}

// -------------------------------------

/**
 * Compare a given date in the year to a rule. Return 1, 0, or -1, depending
 * on whether the date is after, equal to, or before the rule date. The
 * millis are compared directly against the ruleMillis, so any
 * standard-daylight adjustments must be handled by the caller.
 *
 * @return  1 if the date is after the rule date, -1 if the date is before
 *          the rule date, or 0 if the date is equal to the rule date.
 */
int32_t 
SimpleTimeZone::compareToRule(int8_t month, int8_t monthLen, int8_t prevMonthLen,
                              int8_t dayOfMonth,
                              int8_t dayOfWeek, int32_t millis, int32_t millisDelta,
                              EMode ruleMode, int8_t ruleMonth, int8_t ruleDayOfWeek,
                              int8_t ruleDay, int32_t ruleMillis)
{
    // Make adjustments for startTimeMode and endTimeMode
    millis += millisDelta;
    while (millis >= U_MILLIS_PER_DAY) {
        millis -= U_MILLIS_PER_DAY;
        ++dayOfMonth;
        dayOfWeek = (int8_t)(1 + (dayOfWeek % 7)); // dayOfWeek is one-based
        if (dayOfMonth > monthLen) {
            dayOfMonth = 1;
            /* When incrementing the month, it is desirible to overflow
             * from DECEMBER to DECEMBER+1, since we use the result to
             * compare against a real month. Wraparound of the value
             * leads to bug 4173604. */
            ++month;
        }
    }
    while (millis < 0) {
        millis += U_MILLIS_PER_DAY;
        --dayOfMonth;
        dayOfWeek = (int8_t)(1 + ((dayOfWeek+5) % 7)); // dayOfWeek is one-based
        if (dayOfMonth < 1) {
            dayOfMonth = prevMonthLen;
            --month;
        }
    }

    // first compare months.  If they're different, we don't have to worry about days
    // and times
    if (month < ruleMonth) return -1;
    else if (month > ruleMonth) return 1;

    // calculate the actual day of month for the rule
    int32_t ruleDayOfMonth = 0;

    // Adjust the ruleDay to the monthLen, for non-leap year February 29 rule days.
    if (ruleDay > monthLen) {
        ruleDay = monthLen;
    }

    switch (ruleMode)
    {
    // if the mode is day-of-month, the day of month is given
    case DOM_MODE:
        ruleDayOfMonth = ruleDay;
        break;

    // if the mode is day-of-week-in-month, calculate the day-of-month from it
    case DOW_IN_MONTH_MODE:
        // In this case ruleDay is the day-of-week-in-month (this code is using
        // the dayOfWeek and dayOfMonth parameters to figure out the day-of-week
        // of the first day of the month, so it's trusting that they're really
        // consistent with each other)
        if (ruleDay > 0)
            ruleDayOfMonth = 1 + (ruleDay - 1) * 7 +
                (7 + ruleDayOfWeek - (dayOfWeek - dayOfMonth + 1)) % 7;
        
        // if ruleDay is negative (we assume it's not zero here), we have to do
        // the same calculation figuring backward from the last day of the month.
        else
        {
            // (again, this code is trusting that dayOfWeek and dayOfMonth are
            // consistent with each other here, since we're using them to figure
            // the day of week of the first of the month)
            ruleDayOfMonth = monthLen + (ruleDay + 1) * 7 -
                (7 + (dayOfWeek + monthLen - dayOfMonth) - ruleDayOfWeek) % 7;
        }
        break;

    case DOW_GE_DOM_MODE:
        ruleDayOfMonth = ruleDay +
            (49 + ruleDayOfWeek - ruleDay - dayOfWeek + dayOfMonth) % 7;
        break;

    case DOW_LE_DOM_MODE:
        ruleDayOfMonth = ruleDay -
            (49 - ruleDayOfWeek + ruleDay + dayOfWeek - dayOfMonth) % 7;
        // Note at this point ruleDayOfMonth may be <1, although it will
        // be >=1 for well-formed rules.
        break;
    }

    // now that we have a real day-in-month for the rule, we can compare days...
    if (dayOfMonth < ruleDayOfMonth) return -1;
    else if (dayOfMonth > ruleDayOfMonth) return 1;

    // ...and if they're equal, we compare times
    if (millis < ruleMillis) return -1;
    else if (millis > ruleMillis) return 1;
    else return 0;
}

// -------------------------------------

int32_t
SimpleTimeZone::getRawOffset() const
{
    return rawOffset;
}

// -------------------------------------

void
SimpleTimeZone::setRawOffset(int32_t offsetMillis)
{
    rawOffset = offsetMillis;
    transitionRulesInitialized = FALSE;
}

// -------------------------------------

void 
SimpleTimeZone::setDSTSavings(int32_t millisSavedDuringDST, UErrorCode& status) 
{
    if (millisSavedDuringDST == 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    else {
        dstSavings = millisSavedDuringDST;
    }
    transitionRulesInitialized = FALSE;
}

// -------------------------------------

int32_t 
SimpleTimeZone::getDSTSavings() const
{
    return dstSavings;
}

// -------------------------------------

UBool
SimpleTimeZone::useDaylightTime() const
{
    return useDaylight;
}

// -------------------------------------

/**
 * Overrides TimeZone
 * Queries if the given date is in Daylight Savings Time.
 */
UBool SimpleTimeZone::inDaylightTime(UDate date, UErrorCode& status) const
{
    // This method is wasteful since it creates a new GregorianCalendar and
    // deletes it each time it is called.  However, this is a deprecated method
    // and provided only for Java compatibility as of 8/6/97 [LIU].
    if (U_FAILURE(status)) return FALSE;
    GregorianCalendar *gc = new GregorianCalendar(*this, status);
    /* test for NULL */
    if (gc == 0) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return FALSE;
    }
    gc->setTime(date, status);
    UBool result = gc->inDaylightTime(status);
    delete gc;
    return result;
}

// -------------------------------------

/**
 * Return true if this zone has the same rules and offset as another zone.
 * @param other the TimeZone object to be compared with
 * @return true if the given zone has the same rules and offset as this one
 */
UBool 
SimpleTimeZone::hasSameRules(const TimeZone& other) const
{
    if (this == &other) return TRUE;
    if (typeid(*this) != typeid(other)) return FALSE;
    SimpleTimeZone *that = (SimpleTimeZone*)&other;
    return rawOffset     == that->rawOffset &&
        useDaylight     == that->useDaylight &&
        (!useDaylight
         // Only check rules if using DST
         || (dstSavings     == that->dstSavings &&
             startMode      == that->startMode &&
             startMonth     == that->startMonth &&
             startDay       == that->startDay &&
             startDayOfWeek == that->startDayOfWeek &&
             startTime      == that->startTime &&
             startTimeMode  == that->startTimeMode &&
             endMode        == that->endMode &&
             endMonth       == that->endMonth &&
             endDay         == that->endDay &&
             endDayOfWeek   == that->endDayOfWeek &&
             endTime        == that->endTime &&
             endTimeMode    == that->endTimeMode &&
             startYear      == that->startYear));
}

// -------------------------------------

//----------------------------------------------------------------------
// Rule representation
//
// We represent the following flavors of rules:
//       5        the fifth of the month
//       lastSun  the last Sunday in the month
//       lastMon  the last Monday in the month
//       Sun>=8   first Sunday on or after the eighth
//       Sun<=25  last Sunday on or before the 25th
// This is further complicated by the fact that we need to remain
// backward compatible with the 1.1 FCS.  Finally, we need to minimize
// API changes.  In order to satisfy these requirements, we support
// three representation systems, and we translate between them.
//
// INTERNAL REPRESENTATION
// This is the format SimpleTimeZone objects take after construction or
// streaming in is complete.  Rules are represented directly, using an
// unencoded format.  We will discuss the start rule only below; the end
// rule is analogous.
//   startMode      Takes on enumerated values DAY_OF_MONTH,
//                  DOW_IN_MONTH, DOW_AFTER_DOM, or DOW_BEFORE_DOM.
//   startDay       The day of the month, or for DOW_IN_MONTH mode, a
//                  value indicating which DOW, such as +1 for first,
//                  +2 for second, -1 for last, etc.
//   startDayOfWeek The day of the week.  Ignored for DAY_OF_MONTH.
//
// ENCODED REPRESENTATION
// This is the format accepted by the constructor and by setStartRule()
// and setEndRule().  It uses various combinations of positive, negative,
// and zero values to encode the different rules.  This representation
// allows us to specify all the different rule flavors without altering
// the API.
//   MODE              startMonth    startDay    startDayOfWeek
//   DOW_IN_MONTH_MODE >=0           !=0         >0
//   DOM_MODE          >=0           >0          ==0
//   DOW_GE_DOM_MODE   >=0           >0          <0
//   DOW_LE_DOM_MODE   >=0           <0          <0
//   (no DST)          don't care    ==0         don't care
//
// STREAMED REPRESENTATION
// We must retain binary compatibility with the 1.1 FCS.  The 1.1 code only
// handles DOW_IN_MONTH_MODE and non-DST mode, the latter indicated by the
// flag useDaylight.  When we stream an object out, we translate into an
// approximate DOW_IN_MONTH_MODE representation so the object can be parsed
// and used by 1.1 code.  Following that, we write out the full
// representation separately so that contemporary code can recognize and
// parse it.  The full representation is written in a "packed" format,
// consisting of a version number, a length, and an array of bytes.  Future
// versions of this class may specify different versions.  If they wish to
// include additional data, they should do so by storing them after the
// packed representation below.
//----------------------------------------------------------------------

/**
 * Given a set of encoded rules in startDay and startDayOfMonth, decode
 * them and set the startMode appropriately.  Do the same for endDay and
 * endDayOfMonth.  Upon entry, the day of week variables may be zero or
 * negative, in order to indicate special modes.  The day of month
 * variables may also be negative.  Upon exit, the mode variables will be
 * set, and the day of week and day of month variables will be positive.
 * This method also recognizes a startDay or endDay of zero as indicating
 * no DST.
 */
void 
SimpleTimeZone::decodeRules(UErrorCode& status)
{
    decodeStartRule(status);
    decodeEndRule(status);
}

/**
 * Decode the start rule and validate the parameters.  The parameters are
 * expected to be in encoded form, which represents the various rule modes
 * by negating or zeroing certain values.  Representation formats are:
 * <p>
 * <pre>
 *            DOW_IN_MONTH  DOM    DOW>=DOM  DOW<=DOM  no DST
 *            ------------  -----  --------  --------  ----------
 * month       0..11        same    same      same     don't care
 * day        -5..5         1..31   1..31    -1..-31   0
 * dayOfWeek   1..7         0      -1..-7    -1..-7    don't care
 * time        0..ONEDAY    same    same      same     don't care
 * </pre>
 * The range for month does not include UNDECIMBER since this class is
 * really specific to GregorianCalendar, which does not use that month.
 * The range for time includes ONEDAY (vs. ending at ONEDAY-1) because the
 * end rule is an exclusive limit point.  That is, the range of times that
 * are in DST include those >= the start and < the end.  For this reason,
 * it should be possible to specify an end of ONEDAY in order to include the
 * entire day.  Although this is equivalent to time 0 of the following day,
 * it's not always possible to specify that, for example, on December 31.
 * While arguably the start range should still be 0..ONEDAY-1, we keep
 * the start and end ranges the same for consistency.
 */
void 
SimpleTimeZone::decodeStartRule(UErrorCode& status) 
{
    if(U_FAILURE(status)) return;

    useDaylight = (UBool)((startDay != 0) && (endDay != 0) ? TRUE : FALSE);
    if (useDaylight && dstSavings == 0) {
        dstSavings = U_MILLIS_PER_HOUR;
    }
    if (startDay != 0) {
        if (startMonth < UCAL_JANUARY || startMonth > UCAL_DECEMBER) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        if (startTime < 0 || startTime > U_MILLIS_PER_DAY ||
            startTimeMode < WALL_TIME || startTimeMode > UTC_TIME) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        if (startDayOfWeek == 0) {
            startMode = DOM_MODE;
        } else {
            if (startDayOfWeek > 0) {
                startMode = DOW_IN_MONTH_MODE;
            } else {
                startDayOfWeek = (int8_t)-startDayOfWeek;
                if (startDay > 0) {
                    startMode = DOW_GE_DOM_MODE;
                } else {
                    startDay = (int8_t)-startDay;
                    startMode = DOW_LE_DOM_MODE;
                }
            }
            if (startDayOfWeek > UCAL_SATURDAY) {
                status = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
        }
        if (startMode == DOW_IN_MONTH_MODE) {
            if (startDay < -5 || startDay > 5) {
                status = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
        } else if (startDay<1 || startDay > STATICMONTHLENGTH[startMonth]) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
    }
}

/**
 * Decode the end rule and validate the parameters.  This method is exactly
 * analogous to decodeStartRule().
 * @see decodeStartRule
 */
void 
SimpleTimeZone::decodeEndRule(UErrorCode& status) 
{
    if(U_FAILURE(status)) return;

    useDaylight = (UBool)((startDay != 0) && (endDay != 0) ? TRUE : FALSE);
    if (useDaylight && dstSavings == 0) {
        dstSavings = U_MILLIS_PER_HOUR;
    }
    if (endDay != 0) {
        if (endMonth < UCAL_JANUARY || endMonth > UCAL_DECEMBER) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        if (endTime < 0 || endTime > U_MILLIS_PER_DAY ||
            endTimeMode < WALL_TIME || endTimeMode > UTC_TIME) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        if (endDayOfWeek == 0) {
            endMode = DOM_MODE;
        } else {
            if (endDayOfWeek > 0) {
                endMode = DOW_IN_MONTH_MODE;
            } else {
                endDayOfWeek = (int8_t)-endDayOfWeek;
                if (endDay > 0) {
                    endMode = DOW_GE_DOM_MODE;
                } else {
                    endDay = (int8_t)-endDay;
                    endMode = DOW_LE_DOM_MODE;
                }
            }
            if (endDayOfWeek > UCAL_SATURDAY) {
                status = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
        }
        if (endMode == DOW_IN_MONTH_MODE) {
            if (endDay < -5 || endDay > 5) {
                status = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
        } else if (endDay<1 || endDay > STATICMONTHLENGTH[endMonth]) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
    }
}

UBool
SimpleTimeZone::getNextTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const {
    if (!useDaylight) {
        return FALSE;
    }

    UErrorCode status = U_ZERO_ERROR;
    checkTransitionRules(status);
    if (U_FAILURE(status)) {
        return FALSE;
    }

    UDate firstTransitionTime = firstTransition->getTime();
    if (base < firstTransitionTime || (inclusive && base == firstTransitionTime)) {
        result = *firstTransition;
    }
    UDate stdDate, dstDate;
    UBool stdAvail = stdRule->getNextStart(base, dstRule->getRawOffset(), dstRule->getDSTSavings(), inclusive, stdDate);
    UBool dstAvail = dstRule->getNextStart(base, stdRule->getRawOffset(), stdRule->getDSTSavings(), inclusive, dstDate);
    if (stdAvail && (!dstAvail || stdDate < dstDate)) {
        result.setTime(stdDate);
        result.setFrom((const TimeZoneRule&)*dstRule);
        result.setTo((const TimeZoneRule&)*stdRule);
        return TRUE;
    }
    if (dstAvail && (!stdAvail || dstDate < stdDate)) {
        result.setTime(dstDate);
        result.setFrom((const TimeZoneRule&)*stdRule);
        result.setTo((const TimeZoneRule&)*dstRule);
        return TRUE;
    }
    return FALSE;
}

UBool
SimpleTimeZone::getPreviousTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const {
    if (!useDaylight) {
        return FALSE;
    }

    UErrorCode status = U_ZERO_ERROR;
    checkTransitionRules(status);
    if (U_FAILURE(status)) {
        return FALSE;
    }

    UDate firstTransitionTime = firstTransition->getTime();
    if (base < firstTransitionTime || (!inclusive && base == firstTransitionTime)) {
        return FALSE;
    }
    UDate stdDate, dstDate;
    UBool stdAvail = stdRule->getPreviousStart(base, dstRule->getRawOffset(), dstRule->getDSTSavings(), inclusive, stdDate);
    UBool dstAvail = dstRule->getPreviousStart(base, stdRule->getRawOffset(), stdRule->getDSTSavings(), inclusive, dstDate);
    if (stdAvail && (!dstAvail || stdDate > dstDate)) {
        result.setTime(stdDate);
        result.setFrom((const TimeZoneRule&)*dstRule);
        result.setTo((const TimeZoneRule&)*stdRule);
        return TRUE;
    }
    if (dstAvail && (!stdAvail || dstDate > stdDate)) {
        result.setTime(dstDate);
        result.setFrom((const TimeZoneRule&)*stdRule);
        result.setTo((const TimeZoneRule&)*dstRule);
        return TRUE;
    }
    return FALSE;
}

void
SimpleTimeZone::clearTransitionRules(void) {
    initialRule = NULL;
    firstTransition = NULL;
    stdRule = NULL;
    dstRule = NULL;
    transitionRulesInitialized = FALSE;
}

void
SimpleTimeZone::deleteTransitionRules(void) {
    if (initialRule != NULL) {
        delete initialRule;
    }
    if (firstTransition != NULL) {
        delete firstTransition;
    }
    if (stdRule != NULL) {
        delete stdRule;
    }
    if (dstRule != NULL) {
        delete dstRule;
    }
    clearTransitionRules();
 }

/*
 * Lazy transition rules initializer
 *
 *    Note On the removal of UMTX_CHECK from checkTransitionRules():
 *
 *         It would be faster to have a UInitOnce as part of a SimpleTimeZone object,
 *         which would avoid needing to lock a mutex to check the initialization state.
 *         But we can't easily because simpletz.h is a public header, and including
 *         a UInitOnce as a member of SimpleTimeZone would publicly expose internal ICU headers.
 *
 *         Alternatively we could have a pointer to a UInitOnce in the SimpleTimeZone object,
 *         allocate it in the constructors. This would be a more intrusive change, but doable
 *         if performance turns out to be an issue.
 */

void
SimpleTimeZone::checkTransitionRules(UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    static UMutex gLock;
    umtx_lock(&gLock);
    if (!transitionRulesInitialized) {
        SimpleTimeZone *ncThis = const_cast<SimpleTimeZone*>(this);
        ncThis->initTransitionRules(status);
    }
    umtx_unlock(&gLock);
}

void
SimpleTimeZone::initTransitionRules(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    if (transitionRulesInitialized) {
        return;
    }
    deleteTransitionRules();
    UnicodeString tzid;
    getID(tzid);

    if (useDaylight) {
        DateTimeRule* dtRule;
        DateTimeRule::TimeRuleType timeRuleType;
        UDate firstStdStart, firstDstStart;

        // Create a TimeZoneRule for daylight saving time
        timeRuleType = (startTimeMode == STANDARD_TIME) ? DateTimeRule::STANDARD_TIME :
            ((startTimeMode == UTC_TIME) ? DateTimeRule::UTC_TIME : DateTimeRule::WALL_TIME);
        switch (startMode) {
        case DOM_MODE:
            dtRule = new DateTimeRule(startMonth, startDay, startTime, timeRuleType);
            break;
        case DOW_IN_MONTH_MODE:
            dtRule = new DateTimeRule(startMonth, startDay, startDayOfWeek, startTime, timeRuleType);
            break;
        case DOW_GE_DOM_MODE:
            dtRule = new DateTimeRule(startMonth, startDay, startDayOfWeek, true, startTime, timeRuleType);
            break;
        case DOW_LE_DOM_MODE:
            dtRule = new DateTimeRule(startMonth, startDay, startDayOfWeek, false, startTime, timeRuleType);
            break;
        default:
            status = U_INVALID_STATE_ERROR;
            return;
        }
        // Check for Null pointer
        if (dtRule == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        // For now, use ID + "(DST)" as the name
        dstRule = new AnnualTimeZoneRule(tzid+UnicodeString(DST_STR), getRawOffset(), getDSTSavings(),
            dtRule, startYear, AnnualTimeZoneRule::MAX_YEAR);
        
        // Check for Null pointer
        if (dstRule == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            deleteTransitionRules();
            return;
        }
 
        // Calculate the first DST start time
        dstRule->getFirstStart(getRawOffset(), 0, firstDstStart);

        // Create a TimeZoneRule for standard time
        timeRuleType = (endTimeMode == STANDARD_TIME) ? DateTimeRule::STANDARD_TIME :
            ((endTimeMode == UTC_TIME) ? DateTimeRule::UTC_TIME : DateTimeRule::WALL_TIME);
        switch (endMode) {
        case DOM_MODE:
            dtRule = new DateTimeRule(endMonth, endDay, endTime, timeRuleType);
            break;
        case DOW_IN_MONTH_MODE:
            dtRule = new DateTimeRule(endMonth, endDay, endDayOfWeek, endTime, timeRuleType);
            break;
        case DOW_GE_DOM_MODE:
            dtRule = new DateTimeRule(endMonth, endDay, endDayOfWeek, true, endTime, timeRuleType);
            break;
        case DOW_LE_DOM_MODE:
            dtRule = new DateTimeRule(endMonth, endDay, endDayOfWeek, false, endTime, timeRuleType);
            break;
        }
        
        // Check for Null pointer
        if (dtRule == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            deleteTransitionRules();
            return;
        }
        // For now, use ID + "(STD)" as the name
        stdRule = new AnnualTimeZoneRule(tzid+UnicodeString(STD_STR), getRawOffset(), 0,
            dtRule, startYear, AnnualTimeZoneRule::MAX_YEAR);
        
        //Check for Null pointer
        if (stdRule == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            deleteTransitionRules();
            return;
        }

        // Calculate the first STD start time
        stdRule->getFirstStart(getRawOffset(), dstRule->getDSTSavings(), firstStdStart);

        // Create a TimeZoneRule for initial time
        if (firstStdStart < firstDstStart) {
            initialRule = new InitialTimeZoneRule(tzid+UnicodeString(DST_STR), getRawOffset(), dstRule->getDSTSavings());
            if (initialRule == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                deleteTransitionRules();
                return;
            }
            firstTransition = new TimeZoneTransition(firstStdStart, *initialRule, *stdRule);
        } else {
            initialRule = new InitialTimeZoneRule(tzid+UnicodeString(STD_STR), getRawOffset(), 0);
            if (initialRule == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                deleteTransitionRules();
                return;
            }
            firstTransition = new TimeZoneTransition(firstDstStart, *initialRule, *dstRule);
        }
        if (firstTransition == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            deleteTransitionRules();
            return;
        }
        
    } else {
        // Create a TimeZoneRule for initial time
        initialRule = new InitialTimeZoneRule(tzid, getRawOffset(), 0);
        // Check for null pointer.
        if (initialRule == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            deleteTransitionRules();
            return;
        }
    }

    transitionRulesInitialized = TRUE;
}

int32_t
SimpleTimeZone::countTransitionRules(UErrorCode& /*status*/) const {
    return (useDaylight) ? 2 : 0;
}

void
SimpleTimeZone::getTimeZoneRules(const InitialTimeZoneRule*& initial,
                                 const TimeZoneRule* trsrules[],
                                 int32_t& trscount,
                                 UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    checkTransitionRules(status);
    if (U_FAILURE(status)) {
        return;
    }
    initial = initialRule;
    int32_t cnt = 0;
    if (stdRule != NULL) {
        if (cnt < trscount) {
            trsrules[cnt++] = stdRule;
        }
        if (cnt < trscount) {
            trsrules[cnt++] = dstRule;
        }
    }
    trscount = cnt;
}


U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2016, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* File SMPDTFMT.CPP
*
* Modification History:
*
*   Date        Name        Description
*   02/19/97    aliu        Converted from java.
*   03/31/97    aliu        Modified extensively to work with 50 locales.
*   04/01/97    aliu        Added support for centuries.
*   07/09/97    helena      Made ParsePosition into a class.
*   07/21/98    stephen     Added initializeDefaultCentury.
*                             Removed getZoneIndex (added in DateFormatSymbols)
*                             Removed subParseLong
*                             Removed chk
*   02/22/99    stephen     Removed character literals for EBCDIC safety
*   10/14/99    aliu        Updated 2-digit year parsing so that only "00" thru
*                           "99" are recognized. {j28 4182066}
*   11/15/99    weiv        Added support for week of year/day of week format
********************************************************************************
*/

#define ZID_KEY_MAX 128



#if !UCONFIG_NO_FORMATTING





















// #include "unicode/brkiter.h"










#include <float.h>
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2013, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* This file contains declarations for the class SimpleDateFormatStaticSets
*
* SimpleDateFormatStaticSets holds the UnicodeSets that are needed for lenient
* parsing of literal characters in date/time strings.
********************************************************************************
*/





#if !UCONFIG_NO_FORMATTING



U_NAMESPACE_BEGIN

class  UnicodeSet;


class SimpleDateFormatStaticSets : public UMemory
{
public:
    SimpleDateFormatStaticSets(UErrorCode &status);
    ~SimpleDateFormatStaticSets();
    
    static void    initSets(UErrorCode *status);
    static UBool   cleanup();
    
    static UnicodeSet *getIgnorables(UDateFormatField fieldIndex);
    
private:
    UnicodeSet *fDateIgnorables;
    UnicodeSet *fTimeIgnorables;
    UnicodeSet *fOtherIgnorables;
};


U_NAMESPACE_END

#endif   // #if !UCONFIG_NO_FORMATTING








// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2011-2016, International Business Machines Corporation and
 * others. All Rights Reserved.
 *******************************************************************************
 */



/**
 * \file
 * \brief C++ API: TimeZoneNames object
 */



#if !UCONFIG_NO_FORMATTING








// Some zone display names involving supplementary characters can be over 50 chars, 100 UTF-16 code units, 200 UTF-8 bytes
#define ZONE_NAME_U16_MAX 128

U_NAMESPACE_BEGIN

/*
 * ZNStringPool    Pool of (UChar *) strings.  Provides for sharing of repeated
 *                 zone strings.
 */
struct ZNStringPoolChunk;
class U_I18N_API ZNStringPool: public UMemory {
  public:
    ZNStringPool(UErrorCode &status);
    ~ZNStringPool();

    /* Get the pooled string that is equal to the supplied string s.
     * Copy the string into the pool if it is not already present.
     *
     * Life time of the returned string is that of the pool.
     */
    const UChar *get(const UChar *s, UErrorCode &status);

    /* Get the pooled string that is equal to the supplied string s.
     * Copy the string into the pool if it is not already present.
     */
    const UChar *get(const UnicodeString &s, UErrorCode &status);

    /* Adopt a string into the pool, without copying it.
     * Used for strings from resource bundles, which will persist without copying.
     */
    const UChar *adopt(const UChar *s, UErrorCode &status);

    /* Freeze the string pool.  Discards the hash table that is used
     * for looking up a string.  All pointers to pooled strings remain valid.
     */
    void freeze();

  private:
    ZNStringPoolChunk   *fChunks;
    UHashtable           *fHash;
};

/*
 * Character node used by TextTrieMap
 */
struct CharacterNode {
    // No constructor or destructor.
    // We malloc and free an uninitalized array of CharacterNode objects
    // and clear and delete them ourselves.

    void clear();
    void deleteValues(UObjectDeleter *valueDeleter);

    void addValue(void *value, UObjectDeleter *valueDeleter, UErrorCode &status);
    inline UBool hasValues() const;
    inline int32_t countValues() const;
    inline const void *getValue(int32_t index) const;

    void     *fValues;      // Union of one single value vs. UVector of values.
    UChar    fCharacter;    // UTF-16 code unit.
    uint16_t fFirstChild;   // 0 if no children.
    uint16_t fNextSibling;  // 0 terminates the list.
    UBool    fHasValuesVector;
    UBool    fPadding;

    // No value:   fValues == NULL               and  fHasValuesVector == FALSE
    // One value:  fValues == value              and  fHasValuesVector == FALSE
    // >=2 values: fValues == UVector of values  and  fHasValuesVector == TRUE
};

inline UBool CharacterNode::hasValues() const {
    return (UBool)(fValues != NULL);
}

inline int32_t CharacterNode::countValues() const {
    return
        fValues == NULL ? 0 :
        !fHasValuesVector ? 1 :
        ((const UVector *)fValues)->size();
}

inline const void *CharacterNode::getValue(int32_t index) const {
    if (!fHasValuesVector) {
        return fValues;  // Assume index == 0.
    } else {
        return ((const UVector *)fValues)->elementAt(index);
    }
}

/*
 * Search result handler callback interface used by TextTrieMap search.
 */
class TextTrieMapSearchResultHandler : public UMemory {
public:
    virtual UBool handleMatch(int32_t matchLength,
                              const CharacterNode *node, UErrorCode& status) = 0;
    virtual ~TextTrieMapSearchResultHandler(); //added to avoid warning
};

/**
 * TextTrieMap is a trie implementation for supporting
 * fast prefix match for the string key.
 */
class U_I18N_API TextTrieMap : public UMemory {
public:
    TextTrieMap(UBool ignoreCase, UObjectDeleter *valeDeleter);
    virtual ~TextTrieMap();

    void put(const UnicodeString &key, void *value, ZNStringPool &sp, UErrorCode &status);
    void put(const UChar*, void *value, UErrorCode &status);
    void search(const UnicodeString &text, int32_t start,
        TextTrieMapSearchResultHandler *handler, UErrorCode& status) const;
    int32_t isEmpty() const;

private:
    UBool           fIgnoreCase;
    CharacterNode   *fNodes;
    int32_t         fNodesCapacity;
    int32_t         fNodesCount;

    UVector         *fLazyContents;
    UBool           fIsEmpty;
    UObjectDeleter  *fValueDeleter;

    UBool growNodes();
    CharacterNode* addChildNode(CharacterNode *parent, UChar c, UErrorCode &status);
    CharacterNode* getChildNode(CharacterNode *parent, UChar c) const;

    void putImpl(const UnicodeString &key, void *value, UErrorCode &status);
    void buildTrie(UErrorCode &status);
    void search(CharacterNode *node, const UnicodeString &text, int32_t start,
        int32_t index, TextTrieMapSearchResultHandler *handler, UErrorCode &status) const;
};



class ZNames;
class TextTrieMap;
class ZNameSearchHandler;

class TimeZoneNamesImpl : public TimeZoneNames {
public:
    TimeZoneNamesImpl(const Locale& locale, UErrorCode& status);

    virtual ~TimeZoneNamesImpl();

    virtual UBool operator==(const TimeZoneNames& other) const;
    virtual TimeZoneNamesImpl* clone() const;

    StringEnumeration* getAvailableMetaZoneIDs(UErrorCode& status) const;
    StringEnumeration* getAvailableMetaZoneIDs(const UnicodeString& tzID, UErrorCode& status) const;

    UnicodeString& getMetaZoneID(const UnicodeString& tzID, UDate date, UnicodeString& mzID) const;
    UnicodeString& getReferenceZoneID(const UnicodeString& mzID, const char* region, UnicodeString& tzID) const;

    UnicodeString& getMetaZoneDisplayName(const UnicodeString& mzID, UTimeZoneNameType type, UnicodeString& name) const;
    UnicodeString& getTimeZoneDisplayName(const UnicodeString& tzID, UTimeZoneNameType type, UnicodeString& name) const;

    UnicodeString& getExemplarLocationName(const UnicodeString& tzID, UnicodeString& name) const;

    TimeZoneNames::MatchInfoCollection* find(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const;

    void loadAllDisplayNames(UErrorCode& status);
    void getDisplayNames(const UnicodeString& tzID, const UTimeZoneNameType types[], int32_t numTypes, UDate date, UnicodeString dest[], UErrorCode& status) const;

    static UnicodeString& getDefaultExemplarLocationName(const UnicodeString& tzID, UnicodeString& name);

    static StringEnumeration* _getAvailableMetaZoneIDs(UErrorCode& status);
    static StringEnumeration* _getAvailableMetaZoneIDs(const UnicodeString& tzID, UErrorCode& status);
    static UnicodeString& _getMetaZoneID(const UnicodeString& tzID, UDate date, UnicodeString& mzID);
    static UnicodeString& _getReferenceZoneID(const UnicodeString& mzID, const char* region, UnicodeString& tzID);

private:

    Locale fLocale;

    UResourceBundle* fZoneStrings;

    UHashtable* fTZNamesMap;
    UHashtable* fMZNamesMap;

    UBool fNamesTrieFullyLoaded;
    UBool fNamesFullyLoaded;
    TextTrieMap fNamesTrie;

    void initialize(const Locale& locale, UErrorCode& status);
    void cleanup();

    void loadStrings(const UnicodeString& tzCanonicalID, UErrorCode& status);

    ZNames* loadMetaZoneNames(const UnicodeString& mzId, UErrorCode& status);
    ZNames* loadTimeZoneNames(const UnicodeString& mzId, UErrorCode& status);
    TimeZoneNames::MatchInfoCollection* doFind(ZNameSearchHandler& handler,
        const UnicodeString& text, int32_t start, UErrorCode& status) const;
    void addAllNamesIntoTrie(UErrorCode& errorCode);

    void internalLoadAllDisplayNames(UErrorCode& status);

    struct ZoneStringsLoader;
};

class TZDBNames;

class TZDBTimeZoneNames : public TimeZoneNames {
public:
    TZDBTimeZoneNames(const Locale& locale);
    virtual ~TZDBTimeZoneNames();

    virtual UBool operator==(const TimeZoneNames& other) const;
    virtual TZDBTimeZoneNames* clone() const;

    StringEnumeration* getAvailableMetaZoneIDs(UErrorCode& status) const;
    StringEnumeration* getAvailableMetaZoneIDs(const UnicodeString& tzID, UErrorCode& status) const;

    UnicodeString& getMetaZoneID(const UnicodeString& tzID, UDate date, UnicodeString& mzID) const;
    UnicodeString& getReferenceZoneID(const UnicodeString& mzID, const char* region, UnicodeString& tzID) const;

    UnicodeString& getMetaZoneDisplayName(const UnicodeString& mzID, UTimeZoneNameType type, UnicodeString& name) const;
    UnicodeString& getTimeZoneDisplayName(const UnicodeString& tzID, UTimeZoneNameType type, UnicodeString& name) const;

    TimeZoneNames::MatchInfoCollection* find(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const;

    // When TZDBNames for the metazone is not available, this method returns NULL,
    // but does NOT set U_MISSING_RESOURCE_ERROR to status.
    static const TZDBNames* getMetaZoneNames(const UnicodeString& mzId, UErrorCode& status);

private:
    Locale fLocale;
    char fRegion[ULOC_COUNTRY_CAPACITY];
};

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

   // ZONE_NAME_U16_MAX


#if defined( U_DEBUG_CALSVC ) || defined (U_DEBUG_CAL)
#include <stdio.h>
#endif

// *****************************************************************************
// class SimpleDateFormat
// *****************************************************************************

U_NAMESPACE_BEGIN

/**
 * Last-resort string to use for "GMT" when constructing time zone strings.
 */
// For time zones that have no names, use strings GMT+minutes and
// GMT-minutes. For instance, in France the time zone is GMT+60.
// Also accepted are GMT+H:MM or GMT-H:MM.
// Currently not being used
//static const UChar gGmt[]      = {0x0047, 0x004D, 0x0054, 0x0000};         // "GMT"
//static const UChar gGmtPlus[]  = {0x0047, 0x004D, 0x0054, 0x002B, 0x0000}; // "GMT+"
//static const UChar gGmtMinus[] = {0x0047, 0x004D, 0x0054, 0x002D, 0x0000}; // "GMT-"
//static const UChar gDefGmtPat[]       = {0x0047, 0x004D, 0x0054, 0x007B, 0x0030, 0x007D, 0x0000}; /* GMT{0} */
//static const UChar gDefGmtNegHmsPat[] = {0x002D, 0x0048, 0x0048, 0x003A, 0x006D, 0x006D, 0x003A, 0x0073, 0x0073, 0x0000}; /* -HH:mm:ss */
//static const UChar gDefGmtNegHmPat[]  = {0x002D, 0x0048, 0x0048, 0x003A, 0x006D, 0x006D, 0x0000}; /* -HH:mm */
//static const UChar gDefGmtPosHmsPat[] = {0x002B, 0x0048, 0x0048, 0x003A, 0x006D, 0x006D, 0x003A, 0x0073, 0x0073, 0x0000}; /* +HH:mm:ss */
//static const UChar gDefGmtPosHmPat[]  = {0x002B, 0x0048, 0x0048, 0x003A, 0x006D, 0x006D, 0x0000}; /* +HH:mm */
//static const UChar gUt[]       = {0x0055, 0x0054, 0x0000};  // "UT"
//static const UChar gUtc[]      = {0x0055, 0x0054, 0x0043, 0x0000};  // "UT"

typedef enum GmtPatSize {
    kGmtLen = 3,
    kGmtPatLen = 6,
    kNegHmsLen = 9,
    kNegHmLen = 6,
    kPosHmsLen = 9,
    kPosHmLen = 6,
    kUtLen = 2,
    kUtcLen = 3
} GmtPatSize;

// Stuff needed for numbering system overrides

typedef enum OvrStrType {
    kOvrStrDate = 0,
    kOvrStrTime = 1,
    kOvrStrBoth = 2
} OvrStrType;

static const UDateFormatField kDateFields[] = {
    UDAT_YEAR_FIELD,
    UDAT_MONTH_FIELD,
    UDAT_DATE_FIELD,
    UDAT_DAY_OF_YEAR_FIELD,
    UDAT_DAY_OF_WEEK_IN_MONTH_FIELD,
    UDAT_WEEK_OF_YEAR_FIELD,
    UDAT_WEEK_OF_MONTH_FIELD,
    UDAT_YEAR_WOY_FIELD,
    UDAT_EXTENDED_YEAR_FIELD,
    UDAT_JULIAN_DAY_FIELD,
    UDAT_STANDALONE_DAY_FIELD,
    UDAT_STANDALONE_MONTH_FIELD,
    UDAT_QUARTER_FIELD,
    UDAT_STANDALONE_QUARTER_FIELD,
    UDAT_YEAR_NAME_FIELD,
    UDAT_RELATED_YEAR_FIELD };
static const int8_t kDateFieldsCount = 16;

static const UDateFormatField kTimeFields[] = {
    UDAT_HOUR_OF_DAY1_FIELD,
    UDAT_HOUR_OF_DAY0_FIELD,
    UDAT_MINUTE_FIELD,
    UDAT_SECOND_FIELD,
    UDAT_FRACTIONAL_SECOND_FIELD,
    UDAT_HOUR1_FIELD,
    UDAT_HOUR0_FIELD,
    UDAT_MILLISECONDS_IN_DAY_FIELD,
    UDAT_TIMEZONE_RFC_FIELD,
    UDAT_TIMEZONE_LOCALIZED_GMT_OFFSET_FIELD };
static const int8_t kTimeFieldsCount = 10;


// This is a pattern-of-last-resort used when we can't load a usable pattern out
// of a resource.
static const UChar gDefaultPattern[] =
{
    0x79, 0x79, 0x79, 0x79, 0x4D, 0x4D, 0x64, 0x64, 0x20, 0x68, 0x68, 0x3A, 0x6D, 0x6D, 0x20, 0x61, 0
};  /* "yyyyMMdd hh:mm a" */

// This prefix is designed to NEVER MATCH real text, in order to
// suppress the parsing of negative numbers.  Adjust as needed (if
// this becomes valid Unicode).
static const UChar SUPPRESS_NEGATIVE_PREFIX[] = {0xAB00, 0};

/**
 * These are the tags we expect to see in normal resource bundle files associated
 * with a locale.
 */
static const UChar QUOTE = 0x27; // Single quote

/*
 * The field range check bias for each UDateFormatField.
 * The bias is added to the minimum and maximum values
 * before they are compared to the parsed number.
 * For example, the calendar stores zero-based month numbers
 * but the parsed month numbers start at 1, so the bias is 1.
 *
 * A value of -1 means that the value is not checked.
 */
static const int32_t gFieldRangeBias[] = {
    -1,  // 'G' - UDAT_ERA_FIELD
    -1,  // 'y' - UDAT_YEAR_FIELD
     1,  // 'M' - UDAT_MONTH_FIELD
     0,  // 'd' - UDAT_DATE_FIELD
    -1,  // 'k' - UDAT_HOUR_OF_DAY1_FIELD
    -1,  // 'H' - UDAT_HOUR_OF_DAY0_FIELD
     0,  // 'm' - UDAT_MINUTE_FIELD
     0,  // 's' - UDAT_SECOND_FIELD
    -1,  // 'S' - UDAT_FRACTIONAL_SECOND_FIELD (0-999?)
    -1,  // 'E' - UDAT_DAY_OF_WEEK_FIELD (1-7?)
    -1,  // 'D' - UDAT_DAY_OF_YEAR_FIELD (1 - 366?)
    -1,  // 'F' - UDAT_DAY_OF_WEEK_IN_MONTH_FIELD (1-5?)
    -1,  // 'w' - UDAT_WEEK_OF_YEAR_FIELD (1-52?)
    -1,  // 'W' - UDAT_WEEK_OF_MONTH_FIELD (1-5?)
    -1,  // 'a' - UDAT_AM_PM_FIELD
    -1,  // 'h' - UDAT_HOUR1_FIELD
    -1,  // 'K' - UDAT_HOUR0_FIELD
    -1,  // 'z' - UDAT_TIMEZONE_FIELD
    -1,  // 'Y' - UDAT_YEAR_WOY_FIELD
    -1,  // 'e' - UDAT_DOW_LOCAL_FIELD
    -1,  // 'u' - UDAT_EXTENDED_YEAR_FIELD
    -1,  // 'g' - UDAT_JULIAN_DAY_FIELD
    -1,  // 'A' - UDAT_MILLISECONDS_IN_DAY_FIELD
    -1,  // 'Z' - UDAT_TIMEZONE_RFC_FIELD
    -1,  // 'v' - UDAT_TIMEZONE_GENERIC_FIELD
     0,  // 'c' - UDAT_STANDALONE_DAY_FIELD
     1,  // 'L' - UDAT_STANDALONE_MONTH_FIELD
    -1,  // 'Q' - UDAT_QUARTER_FIELD (1-4?)
    -1,  // 'q' - UDAT_STANDALONE_QUARTER_FIELD
    -1,  // 'V' - UDAT_TIMEZONE_SPECIAL_FIELD
    -1,  // 'U' - UDAT_YEAR_NAME_FIELD
    -1,  // 'O' - UDAT_TIMEZONE_LOCALIZED_GMT_OFFSET_FIELD
    -1,  // 'X' - UDAT_TIMEZONE_ISO_FIELD
    -1,  // 'x' - UDAT_TIMEZONE_ISO_LOCAL_FIELD
    -1,  // 'r' - UDAT_RELATED_YEAR_FIELD
#if UDAT_HAS_PATTERN_CHAR_FOR_TIME_SEPARATOR
    -1,  // ':' - UDAT_TIME_SEPARATOR_FIELD
#else
    -1,  // (no pattern character currently) - UDAT_TIME_SEPARATOR_FIELD
#endif
};

// When calendar uses hebr numbering (i.e. he@calendar=hebrew),
// offset the years within the current millenium down to 1-999
static const int32_t HEBREW_CAL_CUR_MILLENIUM_START_YEAR = 5000;
static const int32_t HEBREW_CAL_CUR_MILLENIUM_END_YEAR = 6000;

static UMutex LOCK;

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(SimpleDateFormat)

SimpleDateFormat::NSOverride::~NSOverride() {
    if (snf != NULL) {
        snf->removeRef();
    }
}


void SimpleDateFormat::NSOverride::free() {
    NSOverride *cur = this;
    while (cur) {
        NSOverride *next_temp = cur->next;
        delete cur;
        cur = next_temp;
    }
}

// no matter what the locale's default number format looked like, we want
// to modify it so that it doesn't use thousands separators, doesn't always
// show the decimal point, and recognizes integers only when parsing
static void fixNumberFormatForDates(NumberFormat &nf) {
    nf.setGroupingUsed(FALSE);
    DecimalFormat* decfmt = dynamic_cast<DecimalFormat*>(&nf);
    if (decfmt != NULL) {
        decfmt->setDecimalSeparatorAlwaysShown(FALSE);
    }
    nf.setParseIntegerOnly(TRUE);
    nf.setMinimumFractionDigits(0); // To prevent "Jan 1.00, 1997.00"
}

static const SharedNumberFormat *createSharedNumberFormat(
        NumberFormat *nfToAdopt) {
    fixNumberFormatForDates(*nfToAdopt);
    const SharedNumberFormat *result = new SharedNumberFormat(nfToAdopt);
    if (result == NULL) {
        delete nfToAdopt;
    }
    return result;
}

static const SharedNumberFormat *createSharedNumberFormat(
        const Locale &loc, UErrorCode &status) {
    NumberFormat *nf = NumberFormat::createInstance(loc, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    const SharedNumberFormat *result = createSharedNumberFormat(nf);
    if (result == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    return result;
}

static const SharedNumberFormat **allocSharedNumberFormatters() {
    const SharedNumberFormat **result = (const SharedNumberFormat**)
            uprv_malloc(UDAT_FIELD_COUNT * sizeof(const SharedNumberFormat*));
    if (result == NULL) {
        return NULL;
    }
    for (int32_t i = 0; i < UDAT_FIELD_COUNT; ++i) {
        result[i] = NULL;
    }
    return result;
}

static void freeSharedNumberFormatters(const SharedNumberFormat ** list) {
    for (int32_t i = 0; i < UDAT_FIELD_COUNT; ++i) {
        SharedObject::clearPtr(list[i]);
    }
    uprv_free(list);
}

const NumberFormat *SimpleDateFormat::getNumberFormatByIndex(
        UDateFormatField index) const {
    if (fSharedNumberFormatters == NULL ||
        fSharedNumberFormatters[index] == NULL) {
        return fNumberFormat;
    }
    return &(**fSharedNumberFormatters[index]);
}

//----------------------------------------------------------------------

SimpleDateFormat::~SimpleDateFormat()
{
    delete fSymbols;
    if (fSharedNumberFormatters) {
        freeSharedNumberFormatters(fSharedNumberFormatters);
    }
    if (fTimeZoneFormat) {
        delete fTimeZoneFormat;
    }
    freeFastNumberFormatters();

#if !UCONFIG_NO_BREAK_ITERATION
    delete fCapitalizationBrkIter;
#endif
}

//----------------------------------------------------------------------

SimpleDateFormat::SimpleDateFormat(UErrorCode& status)
  :   fLocale(Locale::getDefault()),
      fSymbols(NULL),
      fTimeZoneFormat(NULL),
      fSharedNumberFormatters(NULL),
      fCapitalizationBrkIter(NULL)
{
    initializeBooleanAttributes();
    construct(kShort, (EStyle) (kShort + kDateOffset), fLocale, status);
    initializeDefaultCentury();
}

//----------------------------------------------------------------------

SimpleDateFormat::SimpleDateFormat(const UnicodeString& pattern,
                                   UErrorCode &status)
:   fPattern(pattern),
    fLocale(Locale::getDefault()),
    fSymbols(NULL),
    fTimeZoneFormat(NULL),
    fSharedNumberFormatters(NULL),
    fCapitalizationBrkIter(NULL)
{
    fDateOverride.setToBogus();
    fTimeOverride.setToBogus();
    initializeBooleanAttributes();
    initializeCalendar(NULL,fLocale,status);
    fSymbols = DateFormatSymbols::createForLocale(fLocale, status);
    initialize(fLocale, status);
    initializeDefaultCentury();

}
//----------------------------------------------------------------------

SimpleDateFormat::SimpleDateFormat(const UnicodeString& pattern,
                                   const UnicodeString& override,
                                   UErrorCode &status)
:   fPattern(pattern),
    fLocale(Locale::getDefault()),
    fSymbols(NULL),
    fTimeZoneFormat(NULL),
    fSharedNumberFormatters(NULL),
    fCapitalizationBrkIter(NULL)
{
    fDateOverride.setTo(override);
    fTimeOverride.setToBogus();
    initializeBooleanAttributes();
    initializeCalendar(NULL,fLocale,status);
    fSymbols = DateFormatSymbols::createForLocale(fLocale, status);
    initialize(fLocale, status);
    initializeDefaultCentury();

    processOverrideString(fLocale,override,kOvrStrBoth,status);

}

//----------------------------------------------------------------------

SimpleDateFormat::SimpleDateFormat(const UnicodeString& pattern,
                                   const Locale& locale,
                                   UErrorCode& status)
:   fPattern(pattern),
    fLocale(locale),
    fTimeZoneFormat(NULL),
    fSharedNumberFormatters(NULL),
    fCapitalizationBrkIter(NULL)
{

    fDateOverride.setToBogus();
    fTimeOverride.setToBogus();
    initializeBooleanAttributes();

    initializeCalendar(NULL,fLocale,status);
    fSymbols = DateFormatSymbols::createForLocale(fLocale, status);
    initialize(fLocale, status);
    initializeDefaultCentury();
}

//----------------------------------------------------------------------

SimpleDateFormat::SimpleDateFormat(const UnicodeString& pattern,
                                   const UnicodeString& override,
                                   const Locale& locale,
                                   UErrorCode& status)
:   fPattern(pattern),
    fLocale(locale),
    fTimeZoneFormat(NULL),
    fSharedNumberFormatters(NULL),
    fCapitalizationBrkIter(NULL)
{

    fDateOverride.setTo(override);
    fTimeOverride.setToBogus();
    initializeBooleanAttributes();

    initializeCalendar(NULL,fLocale,status);
    fSymbols = DateFormatSymbols::createForLocale(fLocale, status);
    initialize(fLocale, status);
    initializeDefaultCentury();

    processOverrideString(locale,override,kOvrStrBoth,status);

}

//----------------------------------------------------------------------

SimpleDateFormat::SimpleDateFormat(const UnicodeString& pattern,
                                   DateFormatSymbols* symbolsToAdopt,
                                   UErrorCode& status)
:   fPattern(pattern),
    fLocale(Locale::getDefault()),
    fSymbols(symbolsToAdopt),
    fTimeZoneFormat(NULL),
    fSharedNumberFormatters(NULL),
    fCapitalizationBrkIter(NULL)
{

    fDateOverride.setToBogus();
    fTimeOverride.setToBogus();
    initializeBooleanAttributes();

    initializeCalendar(NULL,fLocale,status);
    initialize(fLocale, status);
    initializeDefaultCentury();
}

//----------------------------------------------------------------------

SimpleDateFormat::SimpleDateFormat(const UnicodeString& pattern,
                                   const DateFormatSymbols& symbols,
                                   UErrorCode& status)
:   fPattern(pattern),
    fLocale(Locale::getDefault()),
    fSymbols(new DateFormatSymbols(symbols)),
    fTimeZoneFormat(NULL),
    fSharedNumberFormatters(NULL),
    fCapitalizationBrkIter(NULL)
{

    fDateOverride.setToBogus();
    fTimeOverride.setToBogus();
    initializeBooleanAttributes();

    initializeCalendar(NULL, fLocale, status);
    initialize(fLocale, status);
    initializeDefaultCentury();
}

//----------------------------------------------------------------------

// Not for public consumption; used by DateFormat
SimpleDateFormat::SimpleDateFormat(EStyle timeStyle,
                                   EStyle dateStyle,
                                   const Locale& locale,
                                   UErrorCode& status)
:   fLocale(locale),
    fSymbols(NULL),
    fTimeZoneFormat(NULL),
    fSharedNumberFormatters(NULL),
    fCapitalizationBrkIter(NULL)
{
    initializeBooleanAttributes();
    construct(timeStyle, dateStyle, fLocale, status);
    if(U_SUCCESS(status)) {
      initializeDefaultCentury();
    }
}

//----------------------------------------------------------------------

/**
 * Not for public consumption; used by DateFormat.  This constructor
 * never fails.  If the resource data is not available, it uses the
 * the last resort symbols.
 */
SimpleDateFormat::SimpleDateFormat(const Locale& locale,
                                   UErrorCode& status)
:   fPattern(gDefaultPattern),
    fLocale(locale),
    fSymbols(NULL),
    fTimeZoneFormat(NULL),
    fSharedNumberFormatters(NULL),
    fCapitalizationBrkIter(NULL)
{
    if (U_FAILURE(status)) return;
    initializeBooleanAttributes();
    initializeCalendar(NULL, fLocale, status);
    fSymbols = DateFormatSymbols::createForLocale(fLocale, status);
    if (U_FAILURE(status))
    {
        status = U_ZERO_ERROR;
        delete fSymbols;
        // This constructor doesn't fail; it uses last resort data
        fSymbols = new DateFormatSymbols(status);
        /* test for NULL */
        if (fSymbols == 0) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
    }

    fDateOverride.setToBogus();
    fTimeOverride.setToBogus();

    initialize(fLocale, status);
    if(U_SUCCESS(status)) {
      initializeDefaultCentury();
    }
}

//----------------------------------------------------------------------

SimpleDateFormat::SimpleDateFormat(const SimpleDateFormat& other)
:   DateFormat(other),
    fLocale(other.fLocale),
    fSymbols(NULL),
    fTimeZoneFormat(NULL),
    fSharedNumberFormatters(NULL),
    fCapitalizationBrkIter(NULL)
{
    initializeBooleanAttributes();
    *this = other;
}

//----------------------------------------------------------------------

SimpleDateFormat& SimpleDateFormat::operator=(const SimpleDateFormat& other)
{
    if (this == &other) {
        return *this;
    }
    DateFormat::operator=(other);
    fDateOverride = other.fDateOverride;
    fTimeOverride = other.fTimeOverride;

    delete fSymbols;
    fSymbols = NULL;

    if (other.fSymbols)
        fSymbols = new DateFormatSymbols(*other.fSymbols);

    fDefaultCenturyStart         = other.fDefaultCenturyStart;
    fDefaultCenturyStartYear     = other.fDefaultCenturyStartYear;
    fHaveDefaultCentury          = other.fHaveDefaultCentury;

    fPattern = other.fPattern;
    fHasMinute = other.fHasMinute;
    fHasSecond = other.fHasSecond;

    // TimeZoneFormat in ICU4C only depends on a locale for now
    if (fLocale != other.fLocale) {
        delete fTimeZoneFormat;
        fTimeZoneFormat = NULL; // forces lazy instantiation with the other locale
        fLocale = other.fLocale;
    }

#if !UCONFIG_NO_BREAK_ITERATION
    if (other.fCapitalizationBrkIter != NULL) {
        fCapitalizationBrkIter = (other.fCapitalizationBrkIter)->clone();
    }
#endif

    if (fSharedNumberFormatters != NULL) {
        freeSharedNumberFormatters(fSharedNumberFormatters);
        fSharedNumberFormatters = NULL;
    }
    if (other.fSharedNumberFormatters != NULL) {
        fSharedNumberFormatters = allocSharedNumberFormatters();
        if (fSharedNumberFormatters) {
            for (int32_t i = 0; i < UDAT_FIELD_COUNT; ++i) {
                SharedObject::copyPtr(
                        other.fSharedNumberFormatters[i],
                        fSharedNumberFormatters[i]);
            }
        }
    }

    UErrorCode localStatus = U_ZERO_ERROR;
    freeFastNumberFormatters();
    initFastNumberFormatters(localStatus);

    return *this;
}

//----------------------------------------------------------------------

SimpleDateFormat*
SimpleDateFormat::clone() const
{
    return new SimpleDateFormat(*this);
}

//----------------------------------------------------------------------

UBool
SimpleDateFormat::operator==(const Format& other) const
{
    if (DateFormat::operator==(other)) {
        // The DateFormat::operator== check for fCapitalizationContext equality above
        //   is sufficient to check equality of all derived context-related data.
        // DateFormat::operator== guarantees following cast is safe
        SimpleDateFormat* that = (SimpleDateFormat*)&other;
        return (fPattern             == that->fPattern &&
                fSymbols             != NULL && // Check for pathological object
                that->fSymbols       != NULL && // Check for pathological object
                *fSymbols            == *that->fSymbols &&
                fHaveDefaultCentury  == that->fHaveDefaultCentury &&
                fDefaultCenturyStart == that->fDefaultCenturyStart);
    }
    return FALSE;
}

//----------------------------------------------------------------------

void SimpleDateFormat::construct(EStyle timeStyle,
                                 EStyle dateStyle,
                                 const Locale& locale,
                                 UErrorCode& status)
{
    // called by several constructors to load pattern data from the resources
    if (U_FAILURE(status)) return;

    // We will need the calendar to know what type of symbols to load.
    initializeCalendar(NULL, locale, status);
    if (U_FAILURE(status)) return;

    // Load date time patterns directly from resources.
    const char* cType = fCalendar ? fCalendar->getType() : NULL;
    LocalUResourceBundlePointer bundle(ures_open(NULL, locale.getBaseName(), &status));
    if (U_FAILURE(status)) return;

    UBool cTypeIsGregorian = TRUE;
    LocalUResourceBundlePointer dateTimePatterns;
    if (cType != NULL && uprv_strcmp(cType, "gregorian") != 0) {
        CharString resourcePath("calendar/", status);
        resourcePath.append(cType, status).append("/DateTimePatterns", status);
        dateTimePatterns.adoptInstead(
            ures_getByKeyWithFallback(bundle.getAlias(), resourcePath.data(),
                                      (UResourceBundle*)NULL, &status));
        cTypeIsGregorian = FALSE;
    }

    // Check for "gregorian" fallback.
    if (cTypeIsGregorian || status == U_MISSING_RESOURCE_ERROR) {
        status = U_ZERO_ERROR;
        dateTimePatterns.adoptInstead(
            ures_getByKeyWithFallback(bundle.getAlias(),
                                      "calendar/gregorian/DateTimePatterns",
                                      (UResourceBundle*)NULL, &status));
    }
    if (U_FAILURE(status)) return;

    LocalUResourceBundlePointer currentBundle;

    if (ures_getSize(dateTimePatterns.getAlias()) <= kDateTime)
    {
        status = U_INVALID_FORMAT_ERROR;
        return;
    }

    setLocaleIDs(ures_getLocaleByType(dateTimePatterns.getAlias(), ULOC_VALID_LOCALE, &status),
                 ures_getLocaleByType(dateTimePatterns.getAlias(), ULOC_ACTUAL_LOCALE, &status));

    // create a symbols object from the locale
    fSymbols = DateFormatSymbols::createForLocale(locale, status);
    if (U_FAILURE(status)) return;
    /* test for NULL */
    if (fSymbols == 0) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    const UChar *resStr,*ovrStr;
    int32_t resStrLen,ovrStrLen = 0;
    fDateOverride.setToBogus();
    fTimeOverride.setToBogus();

    // if the pattern should include both date and time information, use the date/time
    // pattern string as a guide to tell use how to glue together the appropriate date
    // and time pattern strings.
    if ((timeStyle != kNone) && (dateStyle != kNone))
    {
        currentBundle.adoptInstead(
                ures_getByIndex(dateTimePatterns.getAlias(), (int32_t)timeStyle, NULL, &status));
        if (U_FAILURE(status)) {
           status = U_INVALID_FORMAT_ERROR;
           return;
        }
        switch (ures_getType(currentBundle.getAlias())) {
            case URES_STRING: {
               resStr = ures_getString(currentBundle.getAlias(), &resStrLen, &status);
               break;
            }
            case URES_ARRAY: {
               resStr = ures_getStringByIndex(currentBundle.getAlias(), 0, &resStrLen, &status);
               ovrStr = ures_getStringByIndex(currentBundle.getAlias(), 1, &ovrStrLen, &status);
               fTimeOverride.setTo(TRUE, ovrStr, ovrStrLen);
               break;
            }
            default: {
               status = U_INVALID_FORMAT_ERROR;
               return;
            }
        }

        UnicodeString tempus1(TRUE, resStr, resStrLen);

        currentBundle.adoptInstead(
                ures_getByIndex(dateTimePatterns.getAlias(), (int32_t)dateStyle, NULL, &status));
        if (U_FAILURE(status)) {
           status = U_INVALID_FORMAT_ERROR;
           return;
        }
        switch (ures_getType(currentBundle.getAlias())) {
            case URES_STRING: {
               resStr = ures_getString(currentBundle.getAlias(), &resStrLen, &status);
               break;
            }
            case URES_ARRAY: {
               resStr = ures_getStringByIndex(currentBundle.getAlias(), 0, &resStrLen, &status);
               ovrStr = ures_getStringByIndex(currentBundle.getAlias(), 1, &ovrStrLen, &status);
               fDateOverride.setTo(TRUE, ovrStr, ovrStrLen);
               break;
            }
            default: {
               status = U_INVALID_FORMAT_ERROR;
               return;
            }
        }

        UnicodeString tempus2(TRUE, resStr, resStrLen);

        int32_t glueIndex = kDateTime;
        int32_t patternsSize = ures_getSize(dateTimePatterns.getAlias());
        if (patternsSize >= (kDateTimeOffset + kShort + 1)) {
            // Get proper date time format
            glueIndex = (int32_t)(kDateTimeOffset + (dateStyle - kDateOffset));
        }

        resStr = ures_getStringByIndex(dateTimePatterns.getAlias(), glueIndex, &resStrLen, &status);
        SimpleFormatter(UnicodeString(TRUE, resStr, resStrLen), 2, 2, status).
                format(tempus1, tempus2, fPattern, status);
    }
    // if the pattern includes just time data or just date date, load the appropriate
    // pattern string from the resources
    // setTo() - see DateFormatSymbols::assignArray comments
    else if (timeStyle != kNone) {
        currentBundle.adoptInstead(
                ures_getByIndex(dateTimePatterns.getAlias(), (int32_t)timeStyle, NULL, &status));
        if (U_FAILURE(status)) {
           status = U_INVALID_FORMAT_ERROR;
           return;
        }
        switch (ures_getType(currentBundle.getAlias())) {
            case URES_STRING: {
               resStr = ures_getString(currentBundle.getAlias(), &resStrLen, &status);
               break;
            }
            case URES_ARRAY: {
               resStr = ures_getStringByIndex(currentBundle.getAlias(), 0, &resStrLen, &status);
               ovrStr = ures_getStringByIndex(currentBundle.getAlias(), 1, &ovrStrLen, &status);
               fDateOverride.setTo(TRUE, ovrStr, ovrStrLen);
               break;
            }
            default: {
               status = U_INVALID_FORMAT_ERROR;
               return;
            }
        }
        fPattern.setTo(TRUE, resStr, resStrLen);
    }
    else if (dateStyle != kNone) {
        currentBundle.adoptInstead(
                ures_getByIndex(dateTimePatterns.getAlias(), (int32_t)dateStyle, NULL, &status));
        if (U_FAILURE(status)) {
           status = U_INVALID_FORMAT_ERROR;
           return;
        }
        switch (ures_getType(currentBundle.getAlias())) {
            case URES_STRING: {
               resStr = ures_getString(currentBundle.getAlias(), &resStrLen, &status);
               break;
            }
            case URES_ARRAY: {
               resStr = ures_getStringByIndex(currentBundle.getAlias(), 0, &resStrLen, &status);
               ovrStr = ures_getStringByIndex(currentBundle.getAlias(), 1, &ovrStrLen, &status);
               fDateOverride.setTo(TRUE, ovrStr, ovrStrLen);
               break;
            }
            default: {
               status = U_INVALID_FORMAT_ERROR;
               return;
            }
        }
        fPattern.setTo(TRUE, resStr, resStrLen);
    }

    // and if it includes _neither_, that's an error
    else
        status = U_INVALID_FORMAT_ERROR;

    // finally, finish initializing by creating a Calendar and a NumberFormat
    initialize(locale, status);
}

//----------------------------------------------------------------------

Calendar*
SimpleDateFormat::initializeCalendar(TimeZone* adoptZone, const Locale& locale, UErrorCode& status)
{
    if(!U_FAILURE(status)) {
        fCalendar = Calendar::createInstance(adoptZone?adoptZone:TimeZone::createDefault(), locale, status);
    }
    return fCalendar;
}

void
SimpleDateFormat::initialize(const Locale& locale,
                             UErrorCode& status)
{
    if (U_FAILURE(status)) return;

    parsePattern(); // Need this before initNumberFormatters(), to set fHasHanYearChar

    // Simple-minded hack to force Gannen year numbering for ja@calendar=japanese
    // if format is non-numeric (includes 年) and fDateOverride is not already specified.
    // Now this does get updated if applyPattern subsequently changes the pattern type.
    if (fDateOverride.isBogus() && fHasHanYearChar &&
            fCalendar != nullptr && uprv_strcmp(fCalendar->getType(),"japanese") == 0 &&
            uprv_strcmp(fLocale.getLanguage(),"ja") == 0) {
        fDateOverride.setTo(u"y=jpanyear", -1);
    }

    // We don't need to check that the row count is >= 1, since all 2d arrays have at
    // least one row
    fNumberFormat = NumberFormat::createInstance(locale, status);
    if (fNumberFormat != NULL && U_SUCCESS(status))
    {
        fixNumberFormatForDates(*fNumberFormat);
        //fNumberFormat->setLenient(TRUE); // Java uses a custom DateNumberFormat to format/parse

        initNumberFormatters(locale, status);
        initFastNumberFormatters(status);

    }
    else if (U_SUCCESS(status))
    {
        status = U_MISSING_RESOURCE_ERROR;
    }
}

/* Initialize the fields we use to disambiguate ambiguous years. Separate
 * so we can call it from readObject().
 */
void SimpleDateFormat::initializeDefaultCentury()
{
  if(fCalendar) {
    fHaveDefaultCentury = fCalendar->haveDefaultCentury();
    if(fHaveDefaultCentury) {
      fDefaultCenturyStart = fCalendar->defaultCenturyStart();
      fDefaultCenturyStartYear = fCalendar->defaultCenturyStartYear();
    } else {
      fDefaultCenturyStart = DBL_MIN;
      fDefaultCenturyStartYear = -1;
    }
  }
}

/*
 * Initialize the boolean attributes. Separate so we can call it from all constructors.
 */
void SimpleDateFormat::initializeBooleanAttributes()
{
    UErrorCode status = U_ZERO_ERROR;

    setBooleanAttribute(UDAT_PARSE_ALLOW_WHITESPACE, true, status);
    setBooleanAttribute(UDAT_PARSE_ALLOW_NUMERIC, true, status);
    setBooleanAttribute(UDAT_PARSE_PARTIAL_LITERAL_MATCH, true, status);
    setBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, true, status);
}

/* Define one-century window into which to disambiguate dates using
 * two-digit years. Make public in JDK 1.2.
 */
void SimpleDateFormat::parseAmbiguousDatesAsAfter(UDate startDate, UErrorCode& status)
{
    if(U_FAILURE(status)) {
        return;
    }
    if(!fCalendar) {
      status = U_ILLEGAL_ARGUMENT_ERROR;
      return;
    }

    fCalendar->setTime(startDate, status);
    if(U_SUCCESS(status)) {
        fHaveDefaultCentury = TRUE;
        fDefaultCenturyStart = startDate;
        fDefaultCenturyStartYear = fCalendar->get(UCAL_YEAR, status);
    }
}

//----------------------------------------------------------------------

UnicodeString&
SimpleDateFormat::format(Calendar& cal, UnicodeString& appendTo, FieldPosition& pos) const
{
  UErrorCode status = U_ZERO_ERROR;
  FieldPositionOnlyHandler handler(pos);
  return _format(cal, appendTo, handler, status);
}

//----------------------------------------------------------------------

UnicodeString&
SimpleDateFormat::format(Calendar& cal, UnicodeString& appendTo,
                         FieldPositionIterator* posIter, UErrorCode& status) const
{
  FieldPositionIteratorHandler handler(posIter, status);
  return _format(cal, appendTo, handler, status);
}

//----------------------------------------------------------------------

UnicodeString&
SimpleDateFormat::_format(Calendar& cal, UnicodeString& appendTo,
                            FieldPositionHandler& handler, UErrorCode& status) const
{
    if ( U_FAILURE(status) ) {
       return appendTo;
    }
    Calendar* workCal = &cal;
    Calendar* calClone = NULL;
    if (&cal != fCalendar && uprv_strcmp(cal.getType(), fCalendar->getType()) != 0) {
        // Different calendar type
        // We use the time and time zone from the input calendar, but
        // do not use the input calendar for field calculation.
        calClone = fCalendar->clone();
        if (calClone != NULL) {
            UDate t = cal.getTime(status);
            calClone->setTime(t, status);
            calClone->setTimeZone(cal.getTimeZone());
            workCal = calClone;
        } else {
            status = U_MEMORY_ALLOCATION_ERROR;
            return appendTo;
        }
    }

    UBool inQuote = FALSE;
    UChar prevCh = 0;
    int32_t count = 0;
    int32_t fieldNum = 0;
    UDisplayContext capitalizationContext = getContext(UDISPCTX_TYPE_CAPITALIZATION, status);

    // loop through the pattern string character by character
    for (int32_t i = 0; i < fPattern.length() && U_SUCCESS(status); ++i) {
        UChar ch = fPattern[i];

        // Use subFormat() to format a repeated pattern character
        // when a different pattern or non-pattern character is seen
        if (ch != prevCh && count > 0) {
            subFormat(appendTo, prevCh, count, capitalizationContext, fieldNum++, handler, *workCal, status);
            count = 0;
        }
        if (ch == QUOTE) {
            // Consecutive single quotes are a single quote literal,
            // either outside of quotes or between quotes
            if ((i+1) < fPattern.length() && fPattern[i+1] == QUOTE) {
                appendTo += (UChar)QUOTE;
                ++i;
            } else {
                inQuote = ! inQuote;
            }
        }
        else if (!inQuote && isSyntaxChar(ch)) {
            // ch is a date-time pattern character to be interpreted
            // by subFormat(); count the number of times it is repeated
            prevCh = ch;
            ++count;
        }
        else {
            // Append quoted characters and unquoted non-pattern characters
            appendTo += ch;
        }
    }

    // Format the last item in the pattern, if any
    if (count > 0) {
        subFormat(appendTo, prevCh, count, capitalizationContext, fieldNum++, handler, *workCal, status);
    }

    if (calClone != NULL) {
        delete calClone;
    }

    return appendTo;
}

//----------------------------------------------------------------------

/* Map calendar field into calendar field level.
 * the larger the level, the smaller the field unit.
 * For example, UCAL_ERA level is 0, UCAL_YEAR level is 10,
 * UCAL_MONTH level is 20.
 * NOTE: if new fields adds in, the table needs to update.
 */
const int32_t
SimpleDateFormat::fgCalendarFieldToLevel[] =
{
    /*GyM*/ 0, 10, 20,
    /*wW*/ 20, 30,
    /*dDEF*/ 30, 20, 30, 30,
    /*ahHm*/ 40, 50, 50, 60,
    /*sS*/ 70, 80,
    /*z?Y*/ 0, 0, 10,
    /*eug*/ 30, 10, 0,
    /*A?.*/ 40, 0, 0
};

int32_t SimpleDateFormat::getLevelFromChar(UChar ch) {
    // Map date field LETTER into calendar field level.
    // the larger the level, the smaller the field unit.
    // NOTE: if new fields adds in, the table needs to update.
    static const int32_t mapCharToLevel[] = {
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        //
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        //       !   "   #   $   %   &   '   (   )   *   +   ,   -   .   /
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
#if UDAT_HAS_PATTERN_CHAR_FOR_TIME_SEPARATOR
        //   0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  0, -1, -1, -1, -1, -1,
#else
        //   0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
#endif
        //   @   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O
            -1, 40, -1, -1, 20, 30, 30,  0, 50, -1, -1, 50, 20, 20, -1,  0,
        //   P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _
            -1, 20, -1, 80, -1, 10,  0, 30,  0, 10,  0, -1, -1, -1, -1, -1,
        //   `   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o
            -1, 40, -1, 30, 30, 30, -1,  0, 50, -1, -1, 50,  0, 60, -1, -1,
        //   p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~
            -1, 20, 10, 70, -1, 10,  0, 20,  0, 10,  0, -1, -1, -1, -1, -1
    };

    return ch < UPRV_LENGTHOF(mapCharToLevel) ? mapCharToLevel[ch] : -1;
}

UBool SimpleDateFormat::isSyntaxChar(UChar ch) {
    static const UBool mapCharToIsSyntax[] = {
        //
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
        //
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
        //
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
        //
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
        //         !      "      #      $      %      &      '
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
        //  (      )      *      +      ,      -      .      /
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
        //  0      1      2      3      4      5      6      7
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
#if UDAT_HAS_PATTERN_CHAR_FOR_TIME_SEPARATOR
        //  8      9      :      ;      <      =      >      ?
        FALSE, FALSE,  TRUE, FALSE, FALSE, FALSE, FALSE, FALSE,
#else
        //  8      9      :      ;      <      =      >      ?
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
#endif
        //  @      A      B      C      D      E      F      G
        FALSE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,
        //  H      I      J      K      L      M      N      O
         TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,
        //  P      Q      R      S      T      U      V      W
         TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,
        //  X      Y      Z      [      \      ]      ^      _
         TRUE,  TRUE,  TRUE, FALSE, FALSE, FALSE, FALSE, FALSE,
        //  `      a      b      c      d      e      f      g
        FALSE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,
        //  h      i      j      k      l      m      n      o
         TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,
        //  p      q      r      s      t      u      v      w
         TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,
        //  x      y      z      {      |      }      ~
         TRUE,  TRUE,  TRUE, FALSE, FALSE, FALSE, FALSE, FALSE
    };

    return ch < UPRV_LENGTHOF(mapCharToIsSyntax) ? mapCharToIsSyntax[ch] : FALSE;
}

// Map index into pattern character string to Calendar field number.
const UCalendarDateFields
SimpleDateFormat::fgPatternIndexToCalendarField[] =
{
    /*GyM*/ UCAL_ERA, UCAL_YEAR, UCAL_MONTH,
    /*dkH*/ UCAL_DATE, UCAL_HOUR_OF_DAY, UCAL_HOUR_OF_DAY,
    /*msS*/ UCAL_MINUTE, UCAL_SECOND, UCAL_MILLISECOND,
    /*EDF*/ UCAL_DAY_OF_WEEK, UCAL_DAY_OF_YEAR, UCAL_DAY_OF_WEEK_IN_MONTH,
    /*wWa*/ UCAL_WEEK_OF_YEAR, UCAL_WEEK_OF_MONTH, UCAL_AM_PM,
    /*hKz*/ UCAL_HOUR, UCAL_HOUR, UCAL_ZONE_OFFSET,
    /*Yeu*/ UCAL_YEAR_WOY, UCAL_DOW_LOCAL, UCAL_EXTENDED_YEAR,
    /*gAZ*/ UCAL_JULIAN_DAY, UCAL_MILLISECONDS_IN_DAY, UCAL_ZONE_OFFSET,
    /*v*/   UCAL_ZONE_OFFSET,
    /*c*/   UCAL_DOW_LOCAL,
    /*L*/   UCAL_MONTH,
    /*Q*/   UCAL_MONTH,
    /*q*/   UCAL_MONTH,
    /*V*/   UCAL_ZONE_OFFSET,
    /*U*/   UCAL_YEAR,
    /*O*/   UCAL_ZONE_OFFSET,
    /*Xx*/  UCAL_ZONE_OFFSET, UCAL_ZONE_OFFSET,
    /*r*/   UCAL_EXTENDED_YEAR,
    /*bB*/   UCAL_FIELD_COUNT, UCAL_FIELD_COUNT,  // no mappings to calendar fields
#if UDAT_HAS_PATTERN_CHAR_FOR_TIME_SEPARATOR
    /*:*/   UCAL_FIELD_COUNT, /* => no useful mapping to any calendar field */
#else
    /*no pattern char for UDAT_TIME_SEPARATOR_FIELD*/   UCAL_FIELD_COUNT, /* => no useful mapping to any calendar field */
#endif
};

// Map index into pattern character string to DateFormat field number
const UDateFormatField
SimpleDateFormat::fgPatternIndexToDateFormatField[] = {
    /*GyM*/ UDAT_ERA_FIELD, UDAT_YEAR_FIELD, UDAT_MONTH_FIELD,
    /*dkH*/ UDAT_DATE_FIELD, UDAT_HOUR_OF_DAY1_FIELD, UDAT_HOUR_OF_DAY0_FIELD,
    /*msS*/ UDAT_MINUTE_FIELD, UDAT_SECOND_FIELD, UDAT_FRACTIONAL_SECOND_FIELD,
    /*EDF*/ UDAT_DAY_OF_WEEK_FIELD, UDAT_DAY_OF_YEAR_FIELD, UDAT_DAY_OF_WEEK_IN_MONTH_FIELD,
    /*wWa*/ UDAT_WEEK_OF_YEAR_FIELD, UDAT_WEEK_OF_MONTH_FIELD, UDAT_AM_PM_FIELD,
    /*hKz*/ UDAT_HOUR1_FIELD, UDAT_HOUR0_FIELD, UDAT_TIMEZONE_FIELD,
    /*Yeu*/ UDAT_YEAR_WOY_FIELD, UDAT_DOW_LOCAL_FIELD, UDAT_EXTENDED_YEAR_FIELD,
    /*gAZ*/ UDAT_JULIAN_DAY_FIELD, UDAT_MILLISECONDS_IN_DAY_FIELD, UDAT_TIMEZONE_RFC_FIELD,
    /*v*/   UDAT_TIMEZONE_GENERIC_FIELD,
    /*c*/   UDAT_STANDALONE_DAY_FIELD,
    /*L*/   UDAT_STANDALONE_MONTH_FIELD,
    /*Q*/   UDAT_QUARTER_FIELD,
    /*q*/   UDAT_STANDALONE_QUARTER_FIELD,
    /*V*/   UDAT_TIMEZONE_SPECIAL_FIELD,
    /*U*/   UDAT_YEAR_NAME_FIELD,
    /*O*/   UDAT_TIMEZONE_LOCALIZED_GMT_OFFSET_FIELD,
    /*Xx*/  UDAT_TIMEZONE_ISO_FIELD, UDAT_TIMEZONE_ISO_LOCAL_FIELD,
    /*r*/   UDAT_RELATED_YEAR_FIELD,
    /*bB*/  UDAT_AM_PM_MIDNIGHT_NOON_FIELD, UDAT_FLEXIBLE_DAY_PERIOD_FIELD,
#if UDAT_HAS_PATTERN_CHAR_FOR_TIME_SEPARATOR
    /*:*/   UDAT_TIME_SEPARATOR_FIELD,
#else
    /*no pattern char for UDAT_TIME_SEPARATOR_FIELD*/   UDAT_TIME_SEPARATOR_FIELD,
#endif
};

//----------------------------------------------------------------------

/**
 * Append symbols[value] to dst.  Make sure the array index is not out
 * of bounds.
 */
static inline void
_appendSymbol(UnicodeString& dst,
              int32_t value,
              const UnicodeString* symbols,
              int32_t symbolsCount) {
    U_ASSERT(0 <= value && value < symbolsCount);
    if (0 <= value && value < symbolsCount) {
        dst += symbols[value];
    }
}

static inline void
_appendSymbolWithMonthPattern(UnicodeString& dst, int32_t value, const UnicodeString* symbols, int32_t symbolsCount,
              const UnicodeString* monthPattern, UErrorCode& status) {
    U_ASSERT(0 <= value && value < symbolsCount);
    if (0 <= value && value < symbolsCount) {
        if (monthPattern == NULL) {
            dst += symbols[value];
        } else {
            SimpleFormatter(*monthPattern, 1, 1, status).format(symbols[value], dst, status);
        }
    }
}

//----------------------------------------------------------------------

static number::LocalizedNumberFormatter*
createFastFormatter(const DecimalFormat* df, int32_t minInt, int32_t maxInt, UErrorCode& status) {
    const number::LocalizedNumberFormatter* lnfBase = df->toNumberFormatter(status);
    if (U_FAILURE(status)) {
        return nullptr;
    }
    return lnfBase->integerWidth(
        number::IntegerWidth::zeroFillTo(minInt).truncateAt(maxInt)
    ).clone().orphan();
}

void SimpleDateFormat::initFastNumberFormatters(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    auto* df = dynamic_cast<const DecimalFormat*>(fNumberFormat);
    if (df == nullptr) {
        return;
    }
    fFastNumberFormatters[SMPDTFMT_NF_1x10] = createFastFormatter(df, 1, 10, status);
    fFastNumberFormatters[SMPDTFMT_NF_2x10] = createFastFormatter(df, 2, 10, status);
    fFastNumberFormatters[SMPDTFMT_NF_3x10] = createFastFormatter(df, 3, 10, status);
    fFastNumberFormatters[SMPDTFMT_NF_4x10] = createFastFormatter(df, 4, 10, status);
    fFastNumberFormatters[SMPDTFMT_NF_2x2] = createFastFormatter(df, 2, 2, status);
}

void SimpleDateFormat::freeFastNumberFormatters() {
    delete fFastNumberFormatters[SMPDTFMT_NF_1x10];
    delete fFastNumberFormatters[SMPDTFMT_NF_2x10];
    delete fFastNumberFormatters[SMPDTFMT_NF_3x10];
    delete fFastNumberFormatters[SMPDTFMT_NF_4x10];
    delete fFastNumberFormatters[SMPDTFMT_NF_2x2];
    fFastNumberFormatters[SMPDTFMT_NF_1x10] = nullptr;
    fFastNumberFormatters[SMPDTFMT_NF_2x10] = nullptr;
    fFastNumberFormatters[SMPDTFMT_NF_3x10] = nullptr;
    fFastNumberFormatters[SMPDTFMT_NF_4x10] = nullptr;
    fFastNumberFormatters[SMPDTFMT_NF_2x2] = nullptr;
}


void
SimpleDateFormat::initNumberFormatters(const Locale &locale,UErrorCode &status) {
    if (U_FAILURE(status)) {
        return;
    }
    if ( fDateOverride.isBogus() && fTimeOverride.isBogus() ) {
        return;
    }
    umtx_lock(&LOCK);
    if (fSharedNumberFormatters == NULL) {
        fSharedNumberFormatters = allocSharedNumberFormatters();
        if (fSharedNumberFormatters == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
        }
    }
    umtx_unlock(&LOCK);

    if (U_FAILURE(status)) {
        return;
    }

    processOverrideString(locale,fDateOverride,kOvrStrDate,status);
    processOverrideString(locale,fTimeOverride,kOvrStrTime,status);
}

void
SimpleDateFormat::processOverrideString(const Locale &locale, const UnicodeString &str, int8_t type, UErrorCode &status) {
    if (str.isBogus() || U_FAILURE(status)) {
        return;
    }

    int32_t start = 0;
    int32_t len;
    UnicodeString nsName;
    UnicodeString ovrField;
    UBool moreToProcess = TRUE;
    NSOverride *overrideList = NULL;

    while (moreToProcess) {
        int32_t delimiterPosition = str.indexOf((UChar)ULOC_KEYWORD_ITEM_SEPARATOR_UNICODE,start);
        if (delimiterPosition == -1) {
            moreToProcess = FALSE;
            len = str.length() - start;
        } else {
            len = delimiterPosition - start;
        }
        UnicodeString currentString(str,start,len);
        int32_t equalSignPosition = currentString.indexOf((UChar)ULOC_KEYWORD_ASSIGN_UNICODE,0);
        if (equalSignPosition == -1) { // Simple override string such as "hebrew"
            nsName.setTo(currentString);
            ovrField.setToBogus();
        } else { // Field specific override string such as "y=hebrew"
            nsName.setTo(currentString,equalSignPosition+1);
            ovrField.setTo(currentString,0,1); // We just need the first character.
        }

        int32_t nsNameHash = nsName.hashCode();
        // See if the numbering system is in the override list, if not, then add it.
        NSOverride *curr = overrideList;
        const SharedNumberFormat *snf = NULL;
        UBool found = FALSE;
        while ( curr && !found ) {
            if ( curr->hash == nsNameHash ) {
                snf = curr->snf;
                found = TRUE;
            }
            curr = curr->next;
        }

        if (!found) {
           LocalPointer<NSOverride> cur(new NSOverride);
           if (!cur.isNull()) {
               char kw[ULOC_KEYWORD_AND_VALUES_CAPACITY];
               uprv_strcpy(kw,"numbers=");
               nsName.extract(0,len,kw+8,ULOC_KEYWORD_AND_VALUES_CAPACITY-8,US_INV);

               Locale ovrLoc(locale.getLanguage(),locale.getCountry(),locale.getVariant(),kw);
               cur->hash = nsNameHash;
               cur->next = overrideList;
               SharedObject::copyPtr(
                       createSharedNumberFormat(ovrLoc, status), cur->snf);
               if (U_FAILURE(status)) {
                   if (overrideList) {
                       overrideList->free();
                   }
                   return;
               }
               snf = cur->snf;
               overrideList = cur.orphan();
           } else {
               status = U_MEMORY_ALLOCATION_ERROR;
               if (overrideList) {
                   overrideList->free();
               }
               return;
           }
        }

        // Now that we have an appropriate number formatter, fill in the appropriate spaces in the
        // number formatters table.
        if (ovrField.isBogus()) {
            switch (type) {
                case kOvrStrDate:
                case kOvrStrBoth: {
                    for ( int8_t i=0 ; i<kDateFieldsCount; i++ ) {
                        SharedObject::copyPtr(snf, fSharedNumberFormatters[kDateFields[i]]);
                    }
                    if (type==kOvrStrDate) {
                        break;
                    }
                    U_FALLTHROUGH;
                }
                case kOvrStrTime : {
                    for ( int8_t i=0 ; i<kTimeFieldsCount; i++ ) {
                        SharedObject::copyPtr(snf, fSharedNumberFormatters[kTimeFields[i]]);
                    }
                    break;
                }
            }
        } else {
           // if the pattern character is unrecognized, signal an error and bail out
           UDateFormatField patternCharIndex =
              DateFormatSymbols::getPatternCharIndex(ovrField.charAt(0));
           if (patternCharIndex == UDAT_FIELD_COUNT) {
               status = U_INVALID_FORMAT_ERROR;
               if (overrideList) {
                   overrideList->free();
               }
               return;
           }
           SharedObject::copyPtr(snf, fSharedNumberFormatters[patternCharIndex]);
        }

        start = delimiterPosition + 1;
    }
    if (overrideList) {
        overrideList->free();
    }
}

//---------------------------------------------------------------------
void
SimpleDateFormat::subFormat(UnicodeString &appendTo,
                            UChar ch,
                            int32_t count,
                            UDisplayContext capitalizationContext,
                            int32_t fieldNum,
                            FieldPositionHandler& handler,
                            Calendar& cal,
                            UErrorCode& status) const
{
    if (U_FAILURE(status)) {
        return;
    }

    // this function gets called by format() to produce the appropriate substitution
    // text for an individual pattern symbol (e.g., "HH" or "yyyy")

    UDateFormatField patternCharIndex = DateFormatSymbols::getPatternCharIndex(ch);
    const int32_t maxIntCount = 10;
    int32_t beginOffset = appendTo.length();
    const NumberFormat *currentNumberFormat;
    DateFormatSymbols::ECapitalizationContextUsageType capContextUsageType = DateFormatSymbols::kCapContextUsageOther;

    UBool isHebrewCalendar = (uprv_strcmp(cal.getType(),"hebrew") == 0);
    UBool isChineseCalendar = (uprv_strcmp(cal.getType(),"chinese") == 0 || uprv_strcmp(cal.getType(),"dangi") == 0);

    // if the pattern character is unrecognized, signal an error and dump out
    if (patternCharIndex == UDAT_FIELD_COUNT)
    {
        if (ch != 0x6C) { // pattern char 'l' (SMALL LETTER L) just gets ignored
            status = U_INVALID_FORMAT_ERROR;
        }
        return;
    }

    UCalendarDateFields field = fgPatternIndexToCalendarField[patternCharIndex];
    int32_t value = 0;
    // Don't get value unless it is useful
    if (field < UCAL_FIELD_COUNT) {
        value = (patternCharIndex != UDAT_RELATED_YEAR_FIELD)? cal.get(field, status): cal.getRelatedYear(status);
    }
    if (U_FAILURE(status)) {
        return;
    }

    currentNumberFormat = getNumberFormatByIndex(patternCharIndex);
    if (currentNumberFormat == NULL) {
        status = U_INTERNAL_PROGRAM_ERROR;
        return;
    }
    UnicodeString hebr("hebr", 4, US_INV);

    switch (patternCharIndex) {

    // for any "G" symbol, write out the appropriate era string
    // "GGGG" is wide era name, "GGGGG" is narrow era name, anything else is abbreviated name
    case UDAT_ERA_FIELD:
        if (isChineseCalendar) {
            zeroPaddingNumber(currentNumberFormat,appendTo, value, 1, 9); // as in ICU4J
        } else {
            if (count == 5) {
                _appendSymbol(appendTo, value, fSymbols->fNarrowEras, fSymbols->fNarrowErasCount);
                capContextUsageType = DateFormatSymbols::kCapContextUsageEraNarrow;
            } else if (count == 4) {
                _appendSymbol(appendTo, value, fSymbols->fEraNames, fSymbols->fEraNamesCount);
                capContextUsageType = DateFormatSymbols::kCapContextUsageEraWide;
            } else {
                _appendSymbol(appendTo, value, fSymbols->fEras, fSymbols->fErasCount);
                capContextUsageType = DateFormatSymbols::kCapContextUsageEraAbbrev;
            }
        }
        break;

     case UDAT_YEAR_NAME_FIELD:
        if (fSymbols->fShortYearNames != NULL && value <= fSymbols->fShortYearNamesCount) {
            // the Calendar YEAR field runs 1 through 60 for cyclic years
            _appendSymbol(appendTo, value - 1, fSymbols->fShortYearNames, fSymbols->fShortYearNamesCount);
            break;
        }
        // else fall through to numeric year handling, do not break here
        U_FALLTHROUGH;

   // OLD: for "yyyy", write out the whole year; for "yy", write out the last 2 digits
    // NEW: UTS#35:
//Year         y     yy     yyy     yyyy     yyyyy
//AD 1         1     01     001     0001     00001
//AD 12       12     12     012     0012     00012
//AD 123     123     23     123     0123     00123
//AD 1234   1234     34    1234     1234     01234
//AD 12345 12345     45   12345    12345     12345
    case UDAT_YEAR_FIELD:
    case UDAT_YEAR_WOY_FIELD:
        if (fDateOverride.compare(hebr)==0 && value>HEBREW_CAL_CUR_MILLENIUM_START_YEAR && value<HEBREW_CAL_CUR_MILLENIUM_END_YEAR) {
            value-=HEBREW_CAL_CUR_MILLENIUM_START_YEAR;
        }
        if(count == 2)
            zeroPaddingNumber(currentNumberFormat, appendTo, value, 2, 2);
        else
            zeroPaddingNumber(currentNumberFormat, appendTo, value, count, maxIntCount);
        break;

    // for "MMMM"/"LLLL", write out the whole month name, for "MMM"/"LLL", write out the month
    // abbreviation, for "M"/"L" or "MM"/"LL", write out the month as a number with the
    // appropriate number of digits
    // for "MMMMM"/"LLLLL", use the narrow form
    case UDAT_MONTH_FIELD:
    case UDAT_STANDALONE_MONTH_FIELD:
        if ( isHebrewCalendar ) {
           HebrewCalendar *hc = (HebrewCalendar*)&cal;
           if (hc->isLeapYear(hc->get(UCAL_YEAR,status)) && value == 6 && count >= 3 )
               value = 13; // Show alternate form for Adar II in leap years in Hebrew calendar.
           if (!hc->isLeapYear(hc->get(UCAL_YEAR,status)) && value >= 6 && count < 3 )
               value--; // Adjust the month number down 1 in Hebrew non-leap years, i.e. Adar is 6, not 7.
        }
        {
            int32_t isLeapMonth = (fSymbols->fLeapMonthPatterns != NULL && fSymbols->fLeapMonthPatternsCount >= DateFormatSymbols::kMonthPatternsCount)?
                        cal.get(UCAL_IS_LEAP_MONTH, status): 0;
            // should consolidate the next section by using arrays of pointers & counts for the right symbols...
            if (count == 5) {
                if (patternCharIndex == UDAT_MONTH_FIELD) {
                    _appendSymbolWithMonthPattern(appendTo, value, fSymbols->fNarrowMonths, fSymbols->fNarrowMonthsCount,
                            (isLeapMonth!=0)? &(fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternFormatNarrow]): NULL, status);
                } else {
                    _appendSymbolWithMonthPattern(appendTo, value, fSymbols->fStandaloneNarrowMonths, fSymbols->fStandaloneNarrowMonthsCount,
                            (isLeapMonth!=0)? &(fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternStandaloneNarrow]): NULL, status);
                }
                capContextUsageType = DateFormatSymbols::kCapContextUsageMonthNarrow;
            } else if (count == 4) {
                if (patternCharIndex == UDAT_MONTH_FIELD) {
                    _appendSymbolWithMonthPattern(appendTo, value, fSymbols->fMonths, fSymbols->fMonthsCount,
                            (isLeapMonth!=0)? &(fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternFormatWide]): NULL, status);
                    capContextUsageType = DateFormatSymbols::kCapContextUsageMonthFormat;
                } else {
                    _appendSymbolWithMonthPattern(appendTo, value, fSymbols->fStandaloneMonths, fSymbols->fStandaloneMonthsCount,
                            (isLeapMonth!=0)? &(fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternStandaloneWide]): NULL, status);
                    capContextUsageType = DateFormatSymbols::kCapContextUsageMonthStandalone;
                }
            } else if (count == 3) {
                if (patternCharIndex == UDAT_MONTH_FIELD) {
                    _appendSymbolWithMonthPattern(appendTo, value, fSymbols->fShortMonths, fSymbols->fShortMonthsCount,
                            (isLeapMonth!=0)? &(fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternFormatAbbrev]): NULL, status);
                    capContextUsageType = DateFormatSymbols::kCapContextUsageMonthFormat;
                } else {
                    _appendSymbolWithMonthPattern(appendTo, value, fSymbols->fStandaloneShortMonths, fSymbols->fStandaloneShortMonthsCount,
                            (isLeapMonth!=0)? &(fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternStandaloneAbbrev]): NULL, status);
                    capContextUsageType = DateFormatSymbols::kCapContextUsageMonthStandalone;
                }
            } else {
                UnicodeString monthNumber;
                zeroPaddingNumber(currentNumberFormat,monthNumber, value + 1, count, maxIntCount);
                _appendSymbolWithMonthPattern(appendTo, 0, &monthNumber, 1,
                        (isLeapMonth!=0)? &(fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternNumeric]): NULL, status);
            }
        }
        break;

    // for "k" and "kk", write out the hour, adjusting midnight to appear as "24"
    case UDAT_HOUR_OF_DAY1_FIELD:
        if (value == 0)
            zeroPaddingNumber(currentNumberFormat,appendTo, cal.getMaximum(UCAL_HOUR_OF_DAY) + 1, count, maxIntCount);
        else
            zeroPaddingNumber(currentNumberFormat,appendTo, value, count, maxIntCount);
        break;

    case UDAT_FRACTIONAL_SECOND_FIELD:
        // Fractional seconds left-justify
        {
            int32_t minDigits = (count > 3) ? 3 : count;
            if (count == 1) {
                value /= 100;
            } else if (count == 2) {
                value /= 10;
            }
            zeroPaddingNumber(currentNumberFormat, appendTo, value, minDigits, maxIntCount);
            if (count > 3) {
                zeroPaddingNumber(currentNumberFormat, appendTo, 0, count - 3, maxIntCount);
            }
        }
        break;

    // for "ee" or "e", use local numeric day-of-the-week
    // for "EEEEEE" or "eeeeee", write out the short day-of-the-week name
    // for "EEEEE" or "eeeee", write out the narrow day-of-the-week name
    // for "EEEE" or "eeee", write out the wide day-of-the-week name
    // for "EEE" or "EE" or "E" or "eee", write out the abbreviated day-of-the-week name
    case UDAT_DOW_LOCAL_FIELD:
        if ( count < 3 ) {
            zeroPaddingNumber(currentNumberFormat,appendTo, value, count, maxIntCount);
            break;
        }
        // fall through to EEEEE-EEE handling, but for that we don't want local day-of-week,
        // we want standard day-of-week, so first fix value to work for EEEEE-EEE.
        value = cal.get(UCAL_DAY_OF_WEEK, status);
        if (U_FAILURE(status)) {
            return;
        }
        // fall through, do not break here
        U_FALLTHROUGH;
    case UDAT_DAY_OF_WEEK_FIELD:
        if (count == 5) {
            _appendSymbol(appendTo, value, fSymbols->fNarrowWeekdays,
                          fSymbols->fNarrowWeekdaysCount);
            capContextUsageType = DateFormatSymbols::kCapContextUsageDayNarrow;
        } else if (count == 4) {
            _appendSymbol(appendTo, value, fSymbols->fWeekdays,
                          fSymbols->fWeekdaysCount);
            capContextUsageType = DateFormatSymbols::kCapContextUsageDayFormat;
        } else if (count == 6) {
            _appendSymbol(appendTo, value, fSymbols->fShorterWeekdays,
                          fSymbols->fShorterWeekdaysCount);
            capContextUsageType = DateFormatSymbols::kCapContextUsageDayFormat;
        } else {
            _appendSymbol(appendTo, value, fSymbols->fShortWeekdays,
                          fSymbols->fShortWeekdaysCount);
            capContextUsageType = DateFormatSymbols::kCapContextUsageDayFormat;
        }
        break;

    // for "ccc", write out the abbreviated day-of-the-week name
    // for "cccc", write out the wide day-of-the-week name
    // for "ccccc", use the narrow day-of-the-week name
    // for "ccccc", use the short day-of-the-week name
    case UDAT_STANDALONE_DAY_FIELD:
        if ( count < 3 ) {
            zeroPaddingNumber(currentNumberFormat,appendTo, value, 1, maxIntCount);
            break;
        }
        // fall through to alpha DOW handling, but for that we don't want local day-of-week,
        // we want standard day-of-week, so first fix value.
        value = cal.get(UCAL_DAY_OF_WEEK, status);
        if (U_FAILURE(status)) {
            return;
        }
        if (count == 5) {
            _appendSymbol(appendTo, value, fSymbols->fStandaloneNarrowWeekdays,
                          fSymbols->fStandaloneNarrowWeekdaysCount);
            capContextUsageType = DateFormatSymbols::kCapContextUsageDayNarrow;
        } else if (count == 4) {
            _appendSymbol(appendTo, value, fSymbols->fStandaloneWeekdays,
                          fSymbols->fStandaloneWeekdaysCount);
            capContextUsageType = DateFormatSymbols::kCapContextUsageDayStandalone;
        } else if (count == 6) {
            _appendSymbol(appendTo, value, fSymbols->fStandaloneShorterWeekdays,
                          fSymbols->fStandaloneShorterWeekdaysCount);
            capContextUsageType = DateFormatSymbols::kCapContextUsageDayStandalone;
        } else { // count == 3
            _appendSymbol(appendTo, value, fSymbols->fStandaloneShortWeekdays,
                          fSymbols->fStandaloneShortWeekdaysCount);
            capContextUsageType = DateFormatSymbols::kCapContextUsageDayStandalone;
        }
        break;

    // for "a" symbol, write out the whole AM/PM string
    case UDAT_AM_PM_FIELD:
        if (count < 5) {
            _appendSymbol(appendTo, value, fSymbols->fAmPms,
                          fSymbols->fAmPmsCount);
        } else {
            _appendSymbol(appendTo, value, fSymbols->fNarrowAmPms,
                          fSymbols->fNarrowAmPmsCount);
        }
        break;

    // if we see pattern character for UDAT_TIME_SEPARATOR_FIELD (none currently defined),
    // write out the time separator string. Leave support in for future definition.
    case UDAT_TIME_SEPARATOR_FIELD:
        {
            UnicodeString separator;
            appendTo += fSymbols->getTimeSeparatorString(separator);
        }
        break;

    // for "h" and "hh", write out the hour, adjusting noon and midnight to show up
    // as "12"
    case UDAT_HOUR1_FIELD:
        if (value == 0)
            zeroPaddingNumber(currentNumberFormat,appendTo, cal.getLeastMaximum(UCAL_HOUR) + 1, count, maxIntCount);
        else
            zeroPaddingNumber(currentNumberFormat,appendTo, value, count, maxIntCount);
        break;

    case UDAT_TIMEZONE_FIELD: // 'z'
    case UDAT_TIMEZONE_RFC_FIELD: // 'Z'
    case UDAT_TIMEZONE_GENERIC_FIELD: // 'v'
    case UDAT_TIMEZONE_SPECIAL_FIELD: // 'V'
    case UDAT_TIMEZONE_LOCALIZED_GMT_OFFSET_FIELD: // 'O'
    case UDAT_TIMEZONE_ISO_FIELD: // 'X'
    case UDAT_TIMEZONE_ISO_LOCAL_FIELD: // 'x'
        {
            UChar zsbuf[ZONE_NAME_U16_MAX];
            UnicodeString zoneString(zsbuf, 0, UPRV_LENGTHOF(zsbuf));
            const TimeZone& tz = cal.getTimeZone();
            UDate date = cal.getTime(status);
            const TimeZoneFormat *tzfmt = tzFormat(status);
            if (U_SUCCESS(status)) {
                if (patternCharIndex == UDAT_TIMEZONE_FIELD) {
                    if (count < 4) {
                        // "z", "zz", "zzz"
                        tzfmt->format(UTZFMT_STYLE_SPECIFIC_SHORT, tz, date, zoneString);
                        capContextUsageType = DateFormatSymbols::kCapContextUsageMetazoneShort;
                    } else {
                        // "zzzz" or longer
                        tzfmt->format(UTZFMT_STYLE_SPECIFIC_LONG, tz, date, zoneString);
                        capContextUsageType = DateFormatSymbols::kCapContextUsageMetazoneLong;
                    }
                }
                else if (patternCharIndex == UDAT_TIMEZONE_RFC_FIELD) {
                    if (count < 4) {
                        // "Z"
                        tzfmt->format(UTZFMT_STYLE_ISO_BASIC_LOCAL_FULL, tz, date, zoneString);
                    } else if (count == 5) {
                        // "ZZZZZ"
                        tzfmt->format(UTZFMT_STYLE_ISO_EXTENDED_FULL, tz, date, zoneString);
                    } else {
                        // "ZZ", "ZZZ", "ZZZZ"
                        tzfmt->format(UTZFMT_STYLE_LOCALIZED_GMT, tz, date, zoneString);
                    }
                }
                else if (patternCharIndex == UDAT_TIMEZONE_GENERIC_FIELD) {
                    if (count == 1) {
                        // "v"
                        tzfmt->format(UTZFMT_STYLE_GENERIC_SHORT, tz, date, zoneString);
                        capContextUsageType = DateFormatSymbols::kCapContextUsageMetazoneShort;
                    } else if (count == 4) {
                        // "vvvv"
                        tzfmt->format(UTZFMT_STYLE_GENERIC_LONG, tz, date, zoneString);
                        capContextUsageType = DateFormatSymbols::kCapContextUsageMetazoneLong;
                    }
                }
                else if (patternCharIndex == UDAT_TIMEZONE_SPECIAL_FIELD) {
                    if (count == 1) {
                        // "V"
                        tzfmt->format(UTZFMT_STYLE_ZONE_ID_SHORT, tz, date, zoneString);
                    } else if (count == 2) {
                        // "VV"
                        tzfmt->format(UTZFMT_STYLE_ZONE_ID, tz, date, zoneString);
                    } else if (count == 3) {
                        // "VVV"
                        tzfmt->format(UTZFMT_STYLE_EXEMPLAR_LOCATION, tz, date, zoneString);
                    } else if (count == 4) {
                        // "VVVV"
                        tzfmt->format(UTZFMT_STYLE_GENERIC_LOCATION, tz, date, zoneString);
                        capContextUsageType = DateFormatSymbols::kCapContextUsageZoneLong;
                    }
                }
                else if (patternCharIndex == UDAT_TIMEZONE_LOCALIZED_GMT_OFFSET_FIELD) {
                    if (count == 1) {
                        // "O"
                        tzfmt->format(UTZFMT_STYLE_LOCALIZED_GMT_SHORT, tz, date, zoneString);
                    } else if (count == 4) {
                        // "OOOO"
                        tzfmt->format(UTZFMT_STYLE_LOCALIZED_GMT, tz, date, zoneString);
                    }
                }
                else if (patternCharIndex == UDAT_TIMEZONE_ISO_FIELD) {
                    if (count == 1) {
                        // "X"
                        tzfmt->format(UTZFMT_STYLE_ISO_BASIC_SHORT, tz, date, zoneString);
                    } else if (count == 2) {
                        // "XX"
                        tzfmt->format(UTZFMT_STYLE_ISO_BASIC_FIXED, tz, date, zoneString);
                    } else if (count == 3) {
                        // "XXX"
                        tzfmt->format(UTZFMT_STYLE_ISO_EXTENDED_FIXED, tz, date, zoneString);
                    } else if (count == 4) {
                        // "XXXX"
                        tzfmt->format(UTZFMT_STYLE_ISO_BASIC_FULL, tz, date, zoneString);
                    } else if (count == 5) {
                        // "XXXXX"
                        tzfmt->format(UTZFMT_STYLE_ISO_EXTENDED_FULL, tz, date, zoneString);
                    }
                }
                else if (patternCharIndex == UDAT_TIMEZONE_ISO_LOCAL_FIELD) {
                    if (count == 1) {
                        // "x"
                        tzfmt->format(UTZFMT_STYLE_ISO_BASIC_LOCAL_SHORT, tz, date, zoneString);
                    } else if (count == 2) {
                        // "xx"
                        tzfmt->format(UTZFMT_STYLE_ISO_BASIC_LOCAL_FIXED, tz, date, zoneString);
                    } else if (count == 3) {
                        // "xxx"
                        tzfmt->format(UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FIXED, tz, date, zoneString);
                    } else if (count == 4) {
                        // "xxxx"
                        tzfmt->format(UTZFMT_STYLE_ISO_BASIC_LOCAL_FULL, tz, date, zoneString);
                    } else if (count == 5) {
                        // "xxxxx"
                        tzfmt->format(UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FULL, tz, date, zoneString);
                    }
                }
                else {
                    UPRV_UNREACHABLE;
                }
            }
            appendTo += zoneString;
        }
        break;

    case UDAT_QUARTER_FIELD:
        if (count >= 4)
            _appendSymbol(appendTo, value/3, fSymbols->fQuarters,
                          fSymbols->fQuartersCount);
        else if (count == 3)
            _appendSymbol(appendTo, value/3, fSymbols->fShortQuarters,
                          fSymbols->fShortQuartersCount);
        else
            zeroPaddingNumber(currentNumberFormat,appendTo, (value/3) + 1, count, maxIntCount);
        break;

    case UDAT_STANDALONE_QUARTER_FIELD:
        if (count >= 4)
            _appendSymbol(appendTo, value/3, fSymbols->fStandaloneQuarters,
                          fSymbols->fStandaloneQuartersCount);
        else if (count == 3)
            _appendSymbol(appendTo, value/3, fSymbols->fStandaloneShortQuarters,
                          fSymbols->fStandaloneShortQuartersCount);
        else
            zeroPaddingNumber(currentNumberFormat,appendTo, (value/3) + 1, count, maxIntCount);
        break;

    case UDAT_AM_PM_MIDNIGHT_NOON_FIELD:
    {
        const UnicodeString *toAppend = NULL;
        int32_t hour = cal.get(UCAL_HOUR_OF_DAY, status);

        // Note: "midnight" can be ambiguous as to whether it refers to beginning of day or end of day.
        // For ICU 57 output of "midnight" is temporarily suppressed.

        // For "midnight" and "noon":
        // Time, as displayed, must be exactly noon or midnight.
        // This means minutes and seconds, if present, must be zero.
        if ((/*hour == 0 ||*/ hour == 12) &&
                (!fHasMinute || cal.get(UCAL_MINUTE, status) == 0) &&
                (!fHasSecond || cal.get(UCAL_SECOND, status) == 0)) {
            // Stealing am/pm value to use as our array index.
            // It works out: am/midnight are both 0, pm/noon are both 1,
            // 12 am is 12 midnight, and 12 pm is 12 noon.
            int32_t val = cal.get(UCAL_AM_PM, status);

            if (count <= 3) {
                toAppend = &fSymbols->fAbbreviatedDayPeriods[val];
            } else if (count == 4 || count > 5) {
                toAppend = &fSymbols->fWideDayPeriods[val];
            } else { // count == 5
                toAppend = &fSymbols->fNarrowDayPeriods[val];
            }
        }

        // toAppend is NULL if time isn't exactly midnight or noon (as displayed).
        // toAppend is bogus if time is midnight or noon, but no localized string exists.
        // In either case, fall back to am/pm.
        if (toAppend == NULL || toAppend->isBogus()) {
            // Reformat with identical arguments except ch, now changed to 'a'.
            subFormat(appendTo, 0x61, count, capitalizationContext, fieldNum,
                      handler, cal, status);
        } else {
            appendTo += *toAppend;
        }

        break;
    }

    case UDAT_FLEXIBLE_DAY_PERIOD_FIELD:
    {
        // TODO: Maybe fetch the DayperiodRules during initialization (instead of at the first
        // loading of an instance) if a relevant pattern character (b or B) is used.
        const DayPeriodRules *ruleSet = DayPeriodRules::getInstance(this->getSmpFmtLocale(), status);
        if (U_FAILURE(status)) {
            // Data doesn't conform to spec, therefore loading failed.
            break;
        }
        if (ruleSet == NULL) {
            // Data doesn't exist for the locale we're looking for.
            // Falling back to am/pm.
            subFormat(appendTo, 0x61, count, capitalizationContext, fieldNum,
                      handler, cal, status);
            break;
        }

        // Get current display time.
        int32_t hour = cal.get(UCAL_HOUR_OF_DAY, status);
        int32_t minute = 0;
        if (fHasMinute) {
            minute = cal.get(UCAL_MINUTE, status);
        }
        int32_t second = 0;
        if (fHasSecond) {
            second = cal.get(UCAL_SECOND, status);
        }

        // Determine day period.
        DayPeriodRules::DayPeriod periodType;
        if (hour == 0 && minute == 0 && second == 0 && ruleSet->hasMidnight()) {
            periodType = DayPeriodRules::DAYPERIOD_MIDNIGHT;
        } else if (hour == 12 && minute == 0 && second == 0 && ruleSet->hasNoon()) {
            periodType = DayPeriodRules::DAYPERIOD_NOON;
        } else {
            periodType = ruleSet->getDayPeriodForHour(hour);
        }

        // Rule set exists, therefore periodType can't be UNKNOWN.
        // Get localized string.
        U_ASSERT(periodType != DayPeriodRules::DAYPERIOD_UNKNOWN);
        UnicodeString *toAppend = NULL;
        int32_t index;

        // Note: "midnight" can be ambiguous as to whether it refers to beginning of day or end of day.
        // For ICU 57 output of "midnight" is temporarily suppressed.

        if (periodType != DayPeriodRules::DAYPERIOD_AM &&
                periodType != DayPeriodRules::DAYPERIOD_PM &&
                periodType != DayPeriodRules::DAYPERIOD_MIDNIGHT) {
            index = (int32_t)periodType;
            if (count <= 3) {
                toAppend = &fSymbols->fAbbreviatedDayPeriods[index];  // i.e. short
            } else if (count == 4 || count > 5) {
                toAppend = &fSymbols->fWideDayPeriods[index];
            } else {  // count == 5
                toAppend = &fSymbols->fNarrowDayPeriods[index];
            }
        }

        // Fallback schedule:
        // Midnight/Noon -> General Periods -> AM/PM.

        // Midnight/Noon -> General Periods.
        if ((toAppend == NULL || toAppend->isBogus()) &&
                (periodType == DayPeriodRules::DAYPERIOD_MIDNIGHT ||
                 periodType == DayPeriodRules::DAYPERIOD_NOON)) {
            periodType = ruleSet->getDayPeriodForHour(hour);
            index = (int32_t)periodType;

            if (count <= 3) {
                toAppend = &fSymbols->fAbbreviatedDayPeriods[index];  // i.e. short
            } else if (count == 4 || count > 5) {
                toAppend = &fSymbols->fWideDayPeriods[index];
            } else {  // count == 5
                toAppend = &fSymbols->fNarrowDayPeriods[index];
            }
        }

        // General Periods -> AM/PM.
        if (periodType == DayPeriodRules::DAYPERIOD_AM ||
            periodType == DayPeriodRules::DAYPERIOD_PM ||
            toAppend->isBogus()) {
            subFormat(appendTo, 0x61, count, capitalizationContext, fieldNum,
                      handler, cal, status);
        }
        else {
            appendTo += *toAppend;
        }

        break;
    }

    // all of the other pattern symbols can be formatted as simple numbers with
    // appropriate zero padding
    default:
        zeroPaddingNumber(currentNumberFormat,appendTo, value, count, maxIntCount);
        break;
    }
#if !UCONFIG_NO_BREAK_ITERATION
    // if first field, check to see whether we need to and are able to titlecase it
    if (fieldNum == 0 && fCapitalizationBrkIter != NULL && appendTo.length() > beginOffset &&
            u_islower(appendTo.char32At(beginOffset))) {
        UBool titlecase = FALSE;
        switch (capitalizationContext) {
            case UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE:
                titlecase = TRUE;
                break;
            case UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU:
                titlecase = fSymbols->fCapitalization[capContextUsageType][0];
                break;
            case UDISPCTX_CAPITALIZATION_FOR_STANDALONE:
                titlecase = fSymbols->fCapitalization[capContextUsageType][1];
                break;
            default:
                // titlecase = FALSE;
                break;
        }
        if (titlecase) {
            BreakIterator* const mutableCapitalizationBrkIter = fCapitalizationBrkIter->clone();
            UnicodeString firstField(appendTo, beginOffset);
            firstField.toTitle(mutableCapitalizationBrkIter, fLocale, U_TITLECASE_NO_LOWERCASE | U_TITLECASE_NO_BREAK_ADJUSTMENT);
            appendTo.replaceBetween(beginOffset, appendTo.length(), firstField);
            delete mutableCapitalizationBrkIter;
        }
    }
#else
    (void) capContextUsageType;
#endif

    handler.addAttribute(fgPatternIndexToDateFormatField[patternCharIndex], beginOffset, appendTo.length());
}

//----------------------------------------------------------------------

void SimpleDateFormat::adoptNumberFormat(NumberFormat *formatToAdopt) {
    fixNumberFormatForDates(*formatToAdopt);
    delete fNumberFormat;
    fNumberFormat = formatToAdopt;

    // We successfully set the default number format. Now delete the overrides
    // (can't fail).
    if (fSharedNumberFormatters) {
        freeSharedNumberFormatters(fSharedNumberFormatters);
        fSharedNumberFormatters = NULL;
    }

    // Also re-compute the fast formatters.
    UErrorCode localStatus = U_ZERO_ERROR;
    freeFastNumberFormatters();
    initFastNumberFormatters(localStatus);
}

void SimpleDateFormat::adoptNumberFormat(const UnicodeString& fields, NumberFormat *formatToAdopt, UErrorCode &status){
    fixNumberFormatForDates(*formatToAdopt);
    LocalPointer<NumberFormat> fmt(formatToAdopt);
    if (U_FAILURE(status)) {
        return;
    }

    // We must ensure fSharedNumberFormatters is allocated.
    if (fSharedNumberFormatters == NULL) {
        fSharedNumberFormatters = allocSharedNumberFormatters();
        if (fSharedNumberFormatters == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
    }
    const SharedNumberFormat *newFormat = createSharedNumberFormat(fmt.orphan());
    if (newFormat == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    for (int i=0; i<fields.length(); i++) {
        UChar field = fields.charAt(i);
        // if the pattern character is unrecognized, signal an error and bail out
        UDateFormatField patternCharIndex = DateFormatSymbols::getPatternCharIndex(field);
        if (patternCharIndex == UDAT_FIELD_COUNT) {
            status = U_INVALID_FORMAT_ERROR;
            newFormat->deleteIfZeroRefCount();
            return;
        }

        // Set the number formatter in the table
        SharedObject::copyPtr(
                newFormat, fSharedNumberFormatters[patternCharIndex]);
    }
    newFormat->deleteIfZeroRefCount();
}

const NumberFormat *
SimpleDateFormat::getNumberFormatForField(UChar field) const {
    UDateFormatField index = DateFormatSymbols::getPatternCharIndex(field);
    if (index == UDAT_FIELD_COUNT) {
        return NULL;
    }
    return getNumberFormatByIndex(index);
}

//----------------------------------------------------------------------
void
SimpleDateFormat::zeroPaddingNumber(
        const NumberFormat *currentNumberFormat,
        UnicodeString &appendTo,
        int32_t value, int32_t minDigits, int32_t maxDigits) const
{
    const number::LocalizedNumberFormatter* fastFormatter = nullptr;
    // NOTE: This uses the heuristic that these five min/max int settings account for the vast majority
    // of SimpleDateFormat number formatting cases at the time of writing (ICU 62).
    if (currentNumberFormat == fNumberFormat) {
        if (maxDigits == 10) {
            if (minDigits == 1) {
                fastFormatter = fFastNumberFormatters[SMPDTFMT_NF_1x10];
            } else if (minDigits == 2) {
                fastFormatter = fFastNumberFormatters[SMPDTFMT_NF_2x10];
            } else if (minDigits == 3) {
                fastFormatter = fFastNumberFormatters[SMPDTFMT_NF_3x10];
            } else if (minDigits == 4) {
                fastFormatter = fFastNumberFormatters[SMPDTFMT_NF_4x10];
            }
        } else if (maxDigits == 2) {
            if (minDigits == 2) {
                fastFormatter = fFastNumberFormatters[SMPDTFMT_NF_2x2];
            }
        }
    }
    if (fastFormatter != nullptr) {
        // Can use fast path
        number::impl::UFormattedNumberData result;
        result.quantity.setToInt(value);
        UErrorCode localStatus = U_ZERO_ERROR;
        fastFormatter->formatImpl(&result, localStatus);
        if (U_FAILURE(localStatus)) {
            return;
        }
        appendTo.append(result.getStringRef().toTempUnicodeString());
        return;
    }

    // Check for RBNF (no clone necessary)
    auto* rbnf = dynamic_cast<const RuleBasedNumberFormat*>(currentNumberFormat);
    if (rbnf != nullptr) {
        FieldPosition pos(FieldPosition::DONT_CARE);
        rbnf->format(value, appendTo, pos);  // 3rd arg is there to speed up processing
        return;
    }

    // Fall back to slow path (clone and mutate the NumberFormat)
    if (currentNumberFormat != nullptr) {
        FieldPosition pos(FieldPosition::DONT_CARE);
        LocalPointer<NumberFormat> nf(currentNumberFormat->clone());
        nf->setMinimumIntegerDigits(minDigits);
        nf->setMaximumIntegerDigits(maxDigits);
        nf->format(value, appendTo, pos);  // 3rd arg is there to speed up processing
    }
}

//----------------------------------------------------------------------

/**
 * Return true if the given format character, occuring count
 * times, represents a numeric field.
 */
UBool SimpleDateFormat::isNumeric(UChar formatChar, int32_t count) {
    return DateFormatSymbols::isNumericPatternChar(formatChar, count);
}

UBool
SimpleDateFormat::isAtNumericField(const UnicodeString &pattern, int32_t patternOffset) {
    if (patternOffset >= pattern.length()) {
        // not at any field
        return FALSE;
    }
    UChar ch = pattern.charAt(patternOffset);
    UDateFormatField f = DateFormatSymbols::getPatternCharIndex(ch);
    if (f == UDAT_FIELD_COUNT) {
        // not at any field
        return FALSE;
    }
    int32_t i = patternOffset;
    while (pattern.charAt(++i) == ch) {}
    return DateFormatSymbols::isNumericField(f, i - patternOffset);
}

UBool
SimpleDateFormat::isAfterNonNumericField(const UnicodeString &pattern, int32_t patternOffset) {
    if (patternOffset <= 0) {
        // not after any field
        return FALSE;
    }
    UChar ch = pattern.charAt(--patternOffset);
    UDateFormatField f = DateFormatSymbols::getPatternCharIndex(ch);
    if (f == UDAT_FIELD_COUNT) {
        // not after any field
        return FALSE;
    }
    int32_t i = patternOffset;
    while (pattern.charAt(--i) == ch) {}
    return !DateFormatSymbols::isNumericField(f, patternOffset - i);
}

void
SimpleDateFormat::parse(const UnicodeString& text, Calendar& cal, ParsePosition& parsePos) const
{
    UErrorCode status = U_ZERO_ERROR;
    int32_t pos = parsePos.getIndex();
    if(parsePos.getIndex() < 0) {
        parsePos.setErrorIndex(0);
        return;
    }
    int32_t start = pos;

    // Hold the day period until everything else is parsed, because we need
    // the hour to interpret time correctly.
    int32_t dayPeriodInt = -1;

    UBool ambiguousYear[] = { FALSE };
    int32_t saveHebrewMonth = -1;
    int32_t count = 0;
    UTimeZoneFormatTimeType tzTimeType = UTZFMT_TIME_TYPE_UNKNOWN;

    // For parsing abutting numeric fields. 'abutPat' is the
    // offset into 'pattern' of the first of 2 or more abutting
    // numeric fields.  'abutStart' is the offset into 'text'
    // where parsing the fields begins. 'abutPass' starts off as 0
    // and increments each time we try to parse the fields.
    int32_t abutPat = -1; // If >=0, we are in a run of abutting numeric fields
    int32_t abutStart = 0;
    int32_t abutPass = 0;
    UBool inQuote = FALSE;

    MessageFormat * numericLeapMonthFormatter = NULL;

    Calendar* calClone = NULL;
    Calendar *workCal = &cal;
    if (&cal != fCalendar && uprv_strcmp(cal.getType(), fCalendar->getType()) != 0) {
        // Different calendar type
        // We use the time/zone from the input calendar, but
        // do not use the input calendar for field calculation.
        calClone = fCalendar->clone();
        if (calClone != NULL) {
            calClone->setTime(cal.getTime(status),status);
            if (U_FAILURE(status)) {
                goto ExitParse;
            }
            calClone->setTimeZone(cal.getTimeZone());
            workCal = calClone;
        } else {
            status = U_MEMORY_ALLOCATION_ERROR;
            goto ExitParse;
        }
    }

    if (fSymbols->fLeapMonthPatterns != NULL && fSymbols->fLeapMonthPatternsCount >= DateFormatSymbols::kMonthPatternsCount) {
        numericLeapMonthFormatter = new MessageFormat(fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternNumeric], fLocale, status);
        if (numericLeapMonthFormatter == NULL) {
             status = U_MEMORY_ALLOCATION_ERROR;
             goto ExitParse;
        } else if (U_FAILURE(status)) {
             goto ExitParse; // this will delete numericLeapMonthFormatter
        }
    }

    for (int32_t i=0; i<fPattern.length(); ++i) {
        UChar ch = fPattern.charAt(i);

        // Handle alphabetic field characters.
        if (!inQuote && isSyntaxChar(ch)) {
            int32_t fieldPat = i;

            // Count the length of this field specifier
            count = 1;
            while ((i+1)<fPattern.length() &&
                   fPattern.charAt(i+1) == ch) {
                ++count;
                ++i;
            }

            if (isNumeric(ch, count)) {
                if (abutPat < 0) {
                    // Determine if there is an abutting numeric field.
                    // Record the start of a set of abutting numeric fields.
                    if (isAtNumericField(fPattern, i + 1)) {
                        abutPat = fieldPat;
                        abutStart = pos;
                        abutPass = 0;
                    }
                }
            } else {
                abutPat = -1; // End of any abutting fields
            }

            // Handle fields within a run of abutting numeric fields.  Take
            // the pattern "HHmmss" as an example. We will try to parse
            // 2/2/2 characters of the input text, then if that fails,
            // 1/2/2.  We only adjust the width of the leftmost field; the
            // others remain fixed.  This allows "123456" => 12:34:56, but
            // "12345" => 1:23:45.  Likewise, for the pattern "yyyyMMdd" we
            // try 4/2/2, 3/2/2, 2/2/2, and finally 1/2/2.
            if (abutPat >= 0) {
                // If we are at the start of a run of abutting fields, then
                // shorten this field in each pass.  If we can't shorten
                // this field any more, then the parse of this set of
                // abutting numeric fields has failed.
                if (fieldPat == abutPat) {
                    count -= abutPass++;
                    if (count == 0) {
                        status = U_PARSE_ERROR;
                        goto ExitParse;
                    }
                }

                pos = subParse(text, pos, ch, count,
                               TRUE, FALSE, ambiguousYear, saveHebrewMonth, *workCal, i, numericLeapMonthFormatter, &tzTimeType);

                // If the parse fails anywhere in the run, back up to the
                // start of the run and retry.
                if (pos < 0) {
                    i = abutPat - 1;
                    pos = abutStart;
                    continue;
                }
            }

            // Handle non-numeric fields and non-abutting numeric
            // fields.
            else if (ch != 0x6C) { // pattern char 'l' (SMALL LETTER L) just gets ignored
                int32_t s = subParse(text, pos, ch, count,
                               FALSE, TRUE, ambiguousYear, saveHebrewMonth, *workCal, i, numericLeapMonthFormatter, &tzTimeType, &dayPeriodInt);

                if (s == -pos-1) {
                    // era not present, in special cases allow this to continue
                    // from the position where the era was expected
                    s = pos;

                    if (i+1 < fPattern.length()) {
                        // move to next pattern character
                        UChar c = fPattern.charAt(i+1);

                        // check for whitespace
                        if (PatternProps::isWhiteSpace(c)) {
                            i++;
                            // Advance over run in pattern
                            while ((i+1)<fPattern.length() &&
                                   PatternProps::isWhiteSpace(fPattern.charAt(i+1))) {
                                ++i;
                            }
                        }
                    }
                }
                else if (s <= 0) {
                    status = U_PARSE_ERROR;
                    goto ExitParse;
                }
                pos = s;
            }
        }

        // Handle literal pattern characters.  These are any
        // quoted characters and non-alphabetic unquoted
        // characters.
        else {

            abutPat = -1; // End of any abutting fields

            if (! matchLiterals(fPattern, i, text, pos, getBooleanAttribute(UDAT_PARSE_ALLOW_WHITESPACE, status), getBooleanAttribute(UDAT_PARSE_PARTIAL_LITERAL_MATCH, status), isLenient())) {
                status = U_PARSE_ERROR;
                goto ExitParse;
            }
        }
    }

    // Special hack for trailing "." after non-numeric field.
    if (text.charAt(pos) == 0x2e && getBooleanAttribute(UDAT_PARSE_ALLOW_WHITESPACE, status)) {
        // only do if the last field is not numeric
        if (isAfterNonNumericField(fPattern, fPattern.length())) {
            pos++; // skip the extra "."
        }
    }

    // If dayPeriod is set, use it in conjunction with hour-of-day to determine am/pm.
    if (dayPeriodInt >= 0) {
        DayPeriodRules::DayPeriod dayPeriod = (DayPeriodRules::DayPeriod)dayPeriodInt;
        const DayPeriodRules *ruleSet = DayPeriodRules::getInstance(this->getSmpFmtLocale(), status);

        if (!cal.isSet(UCAL_HOUR) && !cal.isSet(UCAL_HOUR_OF_DAY)) {
            // If hour is not set, set time to the midpoint of current day period, overwriting
            // minutes if it's set.
            double midPoint = ruleSet->getMidPointForDayPeriod(dayPeriod, status);

            // If we can't get midPoint we do nothing.
            if (U_SUCCESS(status)) {
                // Truncate midPoint toward zero to get the hour.
                // Any leftover means it was a half-hour.
                int32_t midPointHour = (int32_t) midPoint;
                int32_t midPointMinute = (midPoint - midPointHour) > 0 ? 30 : 0;

                // No need to set am/pm because hour-of-day is set last therefore takes precedence.
                cal.set(UCAL_HOUR_OF_DAY, midPointHour);
                cal.set(UCAL_MINUTE, midPointMinute);
            }
        } else {
            int hourOfDay;

            if (cal.isSet(UCAL_HOUR_OF_DAY)) {  // Hour is parsed in 24-hour format.
                hourOfDay = cal.get(UCAL_HOUR_OF_DAY, status);
            } else {  // Hour is parsed in 12-hour format.
                hourOfDay = cal.get(UCAL_HOUR, status);
                // cal.get() turns 12 to 0 for 12-hour time; change 0 to 12
                // so 0 unambiguously means a 24-hour time from above.
                if (hourOfDay == 0) { hourOfDay = 12; }
            }
            U_ASSERT(0 <= hourOfDay && hourOfDay <= 23);


            // If hour-of-day is 0 or 13 thru 23 then input time in unambiguously in 24-hour format.
            if (hourOfDay == 0 || (13 <= hourOfDay && hourOfDay <= 23)) {
                // Make hour-of-day take precedence over (hour + am/pm) by setting it again.
                cal.set(UCAL_HOUR_OF_DAY, hourOfDay);
            } else {
                // We have a 12-hour time and need to choose between am and pm.
                // Behave as if dayPeriod spanned 6 hours each way from its center point.
                // This will parse correctly for consistent time + period (e.g. 10 at night) as
                // well as provide a reasonable recovery for inconsistent time + period (e.g.
                // 9 in the afternoon).

                // Assume current time is in the AM.
                // - Change 12 back to 0 for easier handling of 12am.
                // - Append minutes as fractional hours because e.g. 8:15 and 8:45 could be parsed
                // into different half-days if center of dayPeriod is at 14:30.
                // - cal.get(MINUTE) will return 0 if MINUTE is unset, which works.
                if (hourOfDay == 12) { hourOfDay = 0; }
                double currentHour = hourOfDay + (cal.get(UCAL_MINUTE, status)) / 60.0;
                double midPointHour = ruleSet->getMidPointForDayPeriod(dayPeriod, status);

                if (U_SUCCESS(status)) {
                    double hoursAheadMidPoint = currentHour - midPointHour;

                    // Assume current time is in the AM.
                    if (-6 <= hoursAheadMidPoint && hoursAheadMidPoint < 6) {
                        // Assumption holds; set time as such.
                        cal.set(UCAL_AM_PM, 0);
                    } else {
                        cal.set(UCAL_AM_PM, 1);
                    }
                }
            }
        }
    }

    // At this point the fields of Calendar have been set.  Calendar
    // will fill in default values for missing fields when the time
    // is computed.

    parsePos.setIndex(pos);

    // This part is a problem:  When we call parsedDate.after, we compute the time.
    // Take the date April 3 2004 at 2:30 am.  When this is first set up, the year
    // will be wrong if we're parsing a 2-digit year pattern.  It will be 1904.
    // April 3 1904 is a Sunday (unlike 2004) so it is the DST onset day.  2:30 am
    // is therefore an "impossible" time, since the time goes from 1:59 to 3:00 am
    // on that day.  It is therefore parsed out to fields as 3:30 am.  Then we
    // add 100 years, and get April 3 2004 at 3:30 am.  Note that April 3 2004 is
    // a Saturday, so it can have a 2:30 am -- and it should. [LIU]
    /*
        UDate parsedDate = calendar.getTime();
        if( ambiguousYear[0] && !parsedDate.after(fDefaultCenturyStart) ) {
            calendar.add(Calendar.YEAR, 100);
            parsedDate = calendar.getTime();
        }
    */
    // Because of the above condition, save off the fields in case we need to readjust.
    // The procedure we use here is not particularly efficient, but there is no other
    // way to do this given the API restrictions present in Calendar.  We minimize
    // inefficiency by only performing this computation when it might apply, that is,
    // when the two-digit year is equal to the start year, and thus might fall at the
    // front or the back of the default century.  This only works because we adjust
    // the year correctly to start with in other cases -- see subParse().
    if (ambiguousYear[0] || tzTimeType != UTZFMT_TIME_TYPE_UNKNOWN) // If this is true then the two-digit year == the default start year
    {
        // We need a copy of the fields, and we need to avoid triggering a call to
        // complete(), which will recalculate the fields.  Since we can't access
        // the fields[] array in Calendar, we clone the entire object.  This will
        // stop working if Calendar.clone() is ever rewritten to call complete().
        Calendar *copy;
        if (ambiguousYear[0]) {
            copy = cal.clone();
            // Check for failed cloning.
            if (copy == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                goto ExitParse;
            }
            UDate parsedDate = copy->getTime(status);
            // {sfb} check internalGetDefaultCenturyStart
            if (fHaveDefaultCentury && (parsedDate < fDefaultCenturyStart)) {
                // We can't use add here because that does a complete() first.
                cal.set(UCAL_YEAR, fDefaultCenturyStartYear + 100);
            }
            delete copy;
        }

        if (tzTimeType != UTZFMT_TIME_TYPE_UNKNOWN) {
            copy = cal.clone();
            // Check for failed cloning.
            if (copy == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                goto ExitParse;
            }
            const TimeZone & tz = cal.getTimeZone();
            BasicTimeZone *btz = NULL;

            if (dynamic_cast<const OlsonTimeZone *>(&tz) != NULL
                || dynamic_cast<const SimpleTimeZone *>(&tz) != NULL
                || dynamic_cast<const RuleBasedTimeZone *>(&tz) != NULL
                || dynamic_cast<const VTimeZone *>(&tz) != NULL) {
                btz = (BasicTimeZone*)&tz;
            }

            // Get local millis
            copy->set(UCAL_ZONE_OFFSET, 0);
            copy->set(UCAL_DST_OFFSET, 0);
            UDate localMillis = copy->getTime(status);

            // Make sure parsed time zone type (Standard or Daylight)
            // matches the rule used by the parsed time zone.
            int32_t raw, dst;
            if (btz != NULL) {
                if (tzTimeType == UTZFMT_TIME_TYPE_STANDARD) {
                    btz->getOffsetFromLocal(localMillis,
                        BasicTimeZone::kStandard, BasicTimeZone::kStandard, raw, dst, status);
                } else {
                    btz->getOffsetFromLocal(localMillis,
                        BasicTimeZone::kDaylight, BasicTimeZone::kDaylight, raw, dst, status);
                }
            } else {
                // No good way to resolve ambiguous time at transition,
                // but following code work in most case.
                tz.getOffset(localMillis, TRUE, raw, dst, status);
            }

            // Now, compare the results with parsed type, either standard or daylight saving time
            int32_t resolvedSavings = dst;
            if (tzTimeType == UTZFMT_TIME_TYPE_STANDARD) {
                if (dst != 0) {
                    // Override DST_OFFSET = 0 in the result calendar
                    resolvedSavings = 0;
                }
            } else { // tztype == TZTYPE_DST
                if (dst == 0) {
                    if (btz != NULL) {
                        UDate time = localMillis + raw;
                        // We use the nearest daylight saving time rule.
                        TimeZoneTransition beforeTrs, afterTrs;
                        UDate beforeT = time, afterT = time;
                        int32_t beforeSav = 0, afterSav = 0;
                        UBool beforeTrsAvail, afterTrsAvail;

                        // Search for DST rule before or on the time
                        while (TRUE) {
                            beforeTrsAvail = btz->getPreviousTransition(beforeT, TRUE, beforeTrs);
                            if (!beforeTrsAvail) {
                                break;
                            }
                            beforeT = beforeTrs.getTime() - 1;
                            beforeSav = beforeTrs.getFrom()->getDSTSavings();
                            if (beforeSav != 0) {
                                break;
                            }
                        }

                        // Search for DST rule after the time
                        while (TRUE) {
                            afterTrsAvail = btz->getNextTransition(afterT, FALSE, afterTrs);
                            if (!afterTrsAvail) {
                                break;
                            }
                            afterT = afterTrs.getTime();
                            afterSav = afterTrs.getTo()->getDSTSavings();
                            if (afterSav != 0) {
                                break;
                            }
                        }

                        if (beforeTrsAvail && afterTrsAvail) {
                            if (time - beforeT > afterT - time) {
                                resolvedSavings = afterSav;
                            } else {
                                resolvedSavings = beforeSav;
                            }
                        } else if (beforeTrsAvail && beforeSav != 0) {
                            resolvedSavings = beforeSav;
                        } else if (afterTrsAvail && afterSav != 0) {
                            resolvedSavings = afterSav;
                        } else {
                            resolvedSavings = btz->getDSTSavings();
                        }
                    } else {
                        resolvedSavings = tz.getDSTSavings();
                    }
                    if (resolvedSavings == 0) {
                        // final fallback
                        resolvedSavings = U_MILLIS_PER_HOUR;
                    }
                }
            }
            cal.set(UCAL_ZONE_OFFSET, raw);
            cal.set(UCAL_DST_OFFSET, resolvedSavings);
            delete copy;
        }
    }
ExitParse:
    // Set the parsed result if local calendar is used
    // instead of the input calendar
    if (U_SUCCESS(status) && workCal != &cal) {
        cal.setTimeZone(workCal->getTimeZone());
        cal.setTime(workCal->getTime(status), status);
    }

    if (numericLeapMonthFormatter != NULL) {
        delete numericLeapMonthFormatter;
    }
    if (calClone != NULL) {
        delete calClone;
    }

    // If any Calendar calls failed, we pretend that we
    // couldn't parse the string, when in reality this isn't quite accurate--
    // we did parse it; the Calendar calls just failed.
    if (U_FAILURE(status)) {
        parsePos.setErrorIndex(pos);
        parsePos.setIndex(start);
    }
}

//----------------------------------------------------------------------

static int32_t
matchStringWithOptionalDot(const UnicodeString &text,
                            int32_t index,
                            const UnicodeString &data);

int32_t SimpleDateFormat::matchQuarterString(const UnicodeString& text,
                              int32_t start,
                              UCalendarDateFields field,
                              const UnicodeString* data,
                              int32_t dataCount,
                              Calendar& cal) const
{
    int32_t i = 0;
    int32_t count = dataCount;

    // There may be multiple strings in the data[] array which begin with
    // the same prefix (e.g., Cerven and Cervenec (June and July) in Czech).
    // We keep track of the longest match, and return that.  Note that this
    // unfortunately requires us to test all array elements.
    int32_t bestMatchLength = 0, bestMatch = -1;
    UnicodeString bestMatchName;

    for (; i < count; ++i) {
        int32_t matchLength = 0;
        if ((matchLength = matchStringWithOptionalDot(text, start, data[i])) > bestMatchLength) {
            bestMatchLength = matchLength;
            bestMatch = i;
        }
    }

    if (bestMatch >= 0) {
        cal.set(field, bestMatch * 3);
        return start + bestMatchLength;
    }

    return -start;
}

int32_t SimpleDateFormat::matchDayPeriodStrings(const UnicodeString& text, int32_t start,
                              const UnicodeString* data, int32_t dataCount,
                              int32_t &dayPeriod) const
{

    int32_t bestMatchLength = 0, bestMatch = -1;

    for (int32_t i = 0; i < dataCount; ++i) {
        int32_t matchLength = 0;
        if ((matchLength = matchStringWithOptionalDot(text, start, data[i])) > bestMatchLength) {
            bestMatchLength = matchLength;
            bestMatch = i;
        }
    }

    if (bestMatch >= 0) {
        dayPeriod = bestMatch;
        return start + bestMatchLength;
    }

    return -start;
}

//----------------------------------------------------------------------
UBool SimpleDateFormat::matchLiterals(const UnicodeString &pattern,
                                      int32_t &patternOffset,
                                      const UnicodeString &text,
                                      int32_t &textOffset,
                                      UBool whitespaceLenient,
                                      UBool partialMatchLenient,
                                      UBool oldLeniency)
{
    UBool inQuote = FALSE;
    UnicodeString literal;
    int32_t i = patternOffset;

    // scan pattern looking for contiguous literal characters
    for ( ; i < pattern.length(); i += 1) {
        UChar ch = pattern.charAt(i);

        if (!inQuote && isSyntaxChar(ch)) {
            break;
        }

        if (ch == QUOTE) {
            // Match a quote literal ('') inside OR outside of quotes
            if ((i + 1) < pattern.length() && pattern.charAt(i + 1) == QUOTE) {
                i += 1;
            } else {
                inQuote = !inQuote;
                continue;
            }
        }

        literal += ch;
    }

    // at this point, literal contains the literal text
    // and i is the index of the next non-literal pattern character.
    int32_t p;
    int32_t t = textOffset;

    if (whitespaceLenient) {
        // trim leading, trailing whitespace from
        // the literal text
        literal.trim();

        // ignore any leading whitespace in the text
        while (t < text.length() && u_isWhitespace(text.charAt(t))) {
            t += 1;
        }
    }

    for (p = 0; p < literal.length() && t < text.length();) {
        UBool needWhitespace = FALSE;

        while (p < literal.length() && PatternProps::isWhiteSpace(literal.charAt(p))) {
            needWhitespace = TRUE;
            p += 1;
        }

        if (needWhitespace) {
            int32_t tStart = t;

            while (t < text.length()) {
                UChar tch = text.charAt(t);

                if (!u_isUWhiteSpace(tch) && !PatternProps::isWhiteSpace(tch)) {
                    break;
                }

                t += 1;
            }

            // TODO: should we require internal spaces
            // in lenient mode? (There won't be any
            // leading or trailing spaces)
            if (!whitespaceLenient && t == tStart) {
                // didn't find matching whitespace:
                // an error in strict mode
                return FALSE;
            }

            // In strict mode, this run of whitespace
            // may have been at the end.
            if (p >= literal.length()) {
                break;
            }
        }
        if (t >= text.length() || literal.charAt(p) != text.charAt(t)) {
            // Ran out of text, or found a non-matching character:
            // OK in lenient mode, an error in strict mode.
            if (whitespaceLenient) {
                if (t == textOffset && text.charAt(t) == 0x2e &&
                        isAfterNonNumericField(pattern, patternOffset)) {
                    // Lenient mode and the literal input text begins with a "." and
                    // we are after a non-numeric field: We skip the "."
                    ++t;
                    continue;  // Do not update p.
                }
                // if it is actual whitespace and we're whitespace lenient it's OK

                UChar wsc = text.charAt(t);
                if(PatternProps::isWhiteSpace(wsc)) {
                    // Lenient mode and it's just whitespace we skip it
                    ++t;
                    continue;  // Do not update p.
                }
            }
            // hack around oldleniency being a bit of a catch-all bucket and we're just adding support specifically for paritial matches
            if(partialMatchLenient && oldLeniency) {
                break;
            }

            return FALSE;
        }
        ++p;
        ++t;
    }

    // At this point if we're in strict mode we have a complete match.
    // If we're in lenient mode we may have a partial match, or no
    // match at all.
    if (p <= 0) {
        // no match. Pretend it matched a run of whitespace
        // and ignorables in the text.
        const  UnicodeSet *ignorables = NULL;
        UDateFormatField patternCharIndex = DateFormatSymbols::getPatternCharIndex(pattern.charAt(i));
        if (patternCharIndex != UDAT_FIELD_COUNT) {
            ignorables = SimpleDateFormatStaticSets::getIgnorables(patternCharIndex);
        }

        for (t = textOffset; t < text.length(); t += 1) {
            UChar ch = text.charAt(t);

            if (ignorables == NULL || !ignorables->contains(ch)) {
                break;
            }
        }
    }

    // if we get here, we've got a complete match.
    patternOffset = i - 1;
    textOffset = t;

    return TRUE;
}

//----------------------------------------------------------------------

int32_t SimpleDateFormat::matchString(const UnicodeString& text,
                              int32_t start,
                              UCalendarDateFields field,
                              const UnicodeString* data,
                              int32_t dataCount,
                              const UnicodeString* monthPattern,
                              Calendar& cal) const
{
    int32_t i = 0;
    int32_t count = dataCount;

    if (field == UCAL_DAY_OF_WEEK) i = 1;

    // There may be multiple strings in the data[] array which begin with
    // the same prefix (e.g., Cerven and Cervenec (June and July) in Czech).
    // We keep track of the longest match, and return that.  Note that this
    // unfortunately requires us to test all array elements.
    int32_t bestMatchLength = 0, bestMatch = -1;
    UnicodeString bestMatchName;
    int32_t isLeapMonth = 0;

    for (; i < count; ++i) {
        int32_t matchLen = 0;
        if ((matchLen = matchStringWithOptionalDot(text, start, data[i])) > bestMatchLength) {
            bestMatch = i;
            bestMatchLength = matchLen;
        }

        if (monthPattern != NULL) {
            UErrorCode status = U_ZERO_ERROR;
            UnicodeString leapMonthName;
            SimpleFormatter(*monthPattern, 1, 1, status).format(data[i], leapMonthName, status);
            if (U_SUCCESS(status)) {
                if ((matchLen = matchStringWithOptionalDot(text, start, leapMonthName)) > bestMatchLength) {
                    bestMatch = i;
                    bestMatchLength = matchLen;
                    isLeapMonth = 1;
                }
            }
        }
    }

    if (bestMatch >= 0) {
        if (field < UCAL_FIELD_COUNT) {
            // Adjustment for Hebrew Calendar month Adar II
            if (!strcmp(cal.getType(),"hebrew") && field==UCAL_MONTH && bestMatch==13) {
                cal.set(field,6);
            } else {
                if (field == UCAL_YEAR) {
                    bestMatch++; // only get here for cyclic year names, which match 1-based years 1-60
                }
                cal.set(field, bestMatch);
            }
            if (monthPattern != NULL) {
                cal.set(UCAL_IS_LEAP_MONTH, isLeapMonth);
            }
        }

        return start + bestMatchLength;
    }

    return -start;
}

static int32_t
matchStringWithOptionalDot(const UnicodeString &text,
                            int32_t index,
                            const UnicodeString &data) {
    UErrorCode sts = U_ZERO_ERROR;
    int32_t matchLenText = 0;
    int32_t matchLenData = 0;

    u_caseInsensitivePrefixMatch(text.getBuffer() + index, text.length() - index,
                                 data.getBuffer(), data.length(),
                                 0 /* default case option */,
                                 &matchLenText, &matchLenData,
                                 &sts);
    U_ASSERT (U_SUCCESS(sts));

    if (matchLenData == data.length() /* normal match */
        || (data.charAt(data.length() - 1) == 0x2e
            && matchLenData == data.length() - 1 /* match without trailing dot */)) {
        return matchLenText;
    }

    return 0;
}

//----------------------------------------------------------------------

void
SimpleDateFormat::set2DigitYearStart(UDate d, UErrorCode& status)
{
    parseAmbiguousDatesAsAfter(d, status);
}

/**
 * Private member function that converts the parsed date strings into
 * timeFields. Returns -start (for ParsePosition) if failed.
 */
int32_t SimpleDateFormat::subParse(const UnicodeString& text, int32_t& start, UChar ch, int32_t count,
                           UBool obeyCount, UBool allowNegative, UBool ambiguousYear[], int32_t& saveHebrewMonth, Calendar& cal,
                           int32_t patLoc, MessageFormat * numericLeapMonthFormatter, UTimeZoneFormatTimeType *tzTimeType,
                           int32_t *dayPeriod) const
{
    Formattable number;
    int32_t value = 0;
    int32_t i;
    int32_t ps = 0;
    UErrorCode status = U_ZERO_ERROR;
    ParsePosition pos(0);
    UDateFormatField patternCharIndex = DateFormatSymbols::getPatternCharIndex(ch);
    const NumberFormat *currentNumberFormat;
    UnicodeString temp;
    UBool gotNumber = FALSE;

#if defined (U_DEBUG_CAL)
    //fprintf(stderr, "%s:%d - [%c]  st=%d \n", __FILE__, __LINE__, (char) ch, start);
#endif

    if (patternCharIndex == UDAT_FIELD_COUNT) {
        return -start;
    }

    currentNumberFormat = getNumberFormatByIndex(patternCharIndex);
    if (currentNumberFormat == NULL) {
        return -start;
    }
    UCalendarDateFields field = fgPatternIndexToCalendarField[patternCharIndex]; // UCAL_FIELD_COUNT if irrelevant
    UnicodeString hebr("hebr", 4, US_INV);

    if (numericLeapMonthFormatter != NULL) {
        numericLeapMonthFormatter->setFormats((const Format **)&currentNumberFormat, 1);
    }
    UBool isChineseCalendar = (uprv_strcmp(cal.getType(),"chinese") == 0 || uprv_strcmp(cal.getType(),"dangi") == 0);

    // If there are any spaces here, skip over them.  If we hit the end
    // of the string, then fail.
    for (;;) {
        if (start >= text.length()) {
            return -start;
        }
        UChar32 c = text.char32At(start);
        if (!u_isUWhiteSpace(c) /*||*/ && !PatternProps::isWhiteSpace(c)) {
            break;
        }
        start += U16_LENGTH(c);
    }
    pos.setIndex(start);

    // We handle a few special cases here where we need to parse
    // a number value.  We handle further, more generic cases below.  We need
    // to handle some of them here because some fields require extra processing on
    // the parsed value.
    if (patternCharIndex == UDAT_HOUR_OF_DAY1_FIELD ||                       // k
        patternCharIndex == UDAT_HOUR_OF_DAY0_FIELD ||                       // H
        patternCharIndex == UDAT_HOUR1_FIELD ||                              // h
        patternCharIndex == UDAT_HOUR0_FIELD ||                              // K
        (patternCharIndex == UDAT_DOW_LOCAL_FIELD && count <= 2) ||          // e
        (patternCharIndex == UDAT_STANDALONE_DAY_FIELD && count <= 2) ||     // c
        (patternCharIndex == UDAT_MONTH_FIELD && count <= 2) ||              // M
        (patternCharIndex == UDAT_STANDALONE_MONTH_FIELD && count <= 2) ||   // L
        (patternCharIndex == UDAT_QUARTER_FIELD && count <= 2) ||            // Q
        (patternCharIndex == UDAT_STANDALONE_QUARTER_FIELD && count <= 2) || // q
        patternCharIndex == UDAT_YEAR_FIELD ||                               // y
        patternCharIndex == UDAT_YEAR_WOY_FIELD ||                           // Y
        patternCharIndex == UDAT_YEAR_NAME_FIELD ||                          // U (falls back to numeric)
        (patternCharIndex == UDAT_ERA_FIELD && isChineseCalendar) ||         // G
        patternCharIndex == UDAT_FRACTIONAL_SECOND_FIELD)                    // S
    {
        int32_t parseStart = pos.getIndex();
        // It would be good to unify this with the obeyCount logic below,
        // but that's going to be difficult.
        const UnicodeString* src;

        UBool parsedNumericLeapMonth = FALSE;
        if (numericLeapMonthFormatter != NULL && (patternCharIndex == UDAT_MONTH_FIELD || patternCharIndex == UDAT_STANDALONE_MONTH_FIELD)) {
            int32_t argCount;
            Formattable * args = numericLeapMonthFormatter->parse(text, pos, argCount);
            if (args != NULL && argCount == 1 && pos.getIndex() > parseStart && args[0].isNumeric()) {
                parsedNumericLeapMonth = TRUE;
                number.setLong(args[0].getLong());
                cal.set(UCAL_IS_LEAP_MONTH, 1);
                delete[] args;
            } else {
                pos.setIndex(parseStart);
                cal.set(UCAL_IS_LEAP_MONTH, 0);
            }
        }

        if (!parsedNumericLeapMonth) {
            if (obeyCount) {
                if ((start+count) > text.length()) {
                    return -start;
                }

                text.extractBetween(0, start + count, temp);
                src = &temp;
            } else {
                src = &text;
            }

            parseInt(*src, number, pos, allowNegative,currentNumberFormat);
        }

        int32_t txtLoc = pos.getIndex();

        if (txtLoc > parseStart) {
            value = number.getLong();
            gotNumber = TRUE;

            // suffix processing
            if (value < 0 ) {
                txtLoc = checkIntSuffix(text, txtLoc, patLoc+1, TRUE);
                if (txtLoc != pos.getIndex()) {
                    value *= -1;
                }
            }
            else {
                txtLoc = checkIntSuffix(text, txtLoc, patLoc+1, FALSE);
            }

            if (!getBooleanAttribute(UDAT_PARSE_ALLOW_WHITESPACE, status)) {
                // Check the range of the value
                int32_t bias = gFieldRangeBias[patternCharIndex];
                if (bias >= 0 && (value > cal.getMaximum(field) + bias || value < cal.getMinimum(field) + bias)) {
                    return -start;
                }
            }

            pos.setIndex(txtLoc);
        }
    }

    // Make sure that we got a number if
    // we want one, and didn't get one
    // if we don't want one.
    switch (patternCharIndex) {
        case UDAT_HOUR_OF_DAY1_FIELD:
        case UDAT_HOUR_OF_DAY0_FIELD:
        case UDAT_HOUR1_FIELD:
        case UDAT_HOUR0_FIELD:
            // special range check for hours:
            if (value < 0 || value > 24) {
                return -start;
            }

            // fall through to gotNumber check
            U_FALLTHROUGH;
        case UDAT_YEAR_FIELD:
        case UDAT_YEAR_WOY_FIELD:
        case UDAT_FRACTIONAL_SECOND_FIELD:
            // these must be a number
            if (! gotNumber) {
                return -start;
            }

            break;

        default:
            // we check the rest of the fields below.
            break;
    }

    switch (patternCharIndex) {
    case UDAT_ERA_FIELD:
        if (isChineseCalendar) {
            if (!gotNumber) {
                return -start;
            }
            cal.set(UCAL_ERA, value);
            return pos.getIndex();
        }
        if (count == 5) {
            ps = matchString(text, start, UCAL_ERA, fSymbols->fNarrowEras, fSymbols->fNarrowErasCount, NULL, cal);
        } else if (count == 4) {
            ps = matchString(text, start, UCAL_ERA, fSymbols->fEraNames, fSymbols->fEraNamesCount, NULL, cal);
        } else {
            ps = matchString(text, start, UCAL_ERA, fSymbols->fEras, fSymbols->fErasCount, NULL, cal);
        }

        // check return position, if it equals -start, then matchString error
        // special case the return code so we don't necessarily fail out until we
        // verify no year information also
        if (ps == -start)
            ps--;

        return ps;

    case UDAT_YEAR_FIELD:
        // If there are 3 or more YEAR pattern characters, this indicates
        // that the year value is to be treated literally, without any
        // two-digit year adjustments (e.g., from "01" to 2001).  Otherwise
        // we made adjustments to place the 2-digit year in the proper
        // century, for parsed strings from "00" to "99".  Any other string
        // is treated literally:  "2250", "-1", "1", "002".
        if (fDateOverride.compare(hebr)==0 && value < 1000) {
            value += HEBREW_CAL_CUR_MILLENIUM_START_YEAR;
        } else if (text.moveIndex32(start, 2) == pos.getIndex() && !isChineseCalendar
            && u_isdigit(text.char32At(start))
            && u_isdigit(text.char32At(text.moveIndex32(start, 1))))
        {
            // only adjust year for patterns less than 3.
            if(count < 3) {
                // Assume for example that the defaultCenturyStart is 6/18/1903.
                // This means that two-digit years will be forced into the range
                // 6/18/1903 to 6/17/2003.  As a result, years 00, 01, and 02
                // correspond to 2000, 2001, and 2002.  Years 04, 05, etc. correspond
                // to 1904, 1905, etc.  If the year is 03, then it is 2003 if the
                // other fields specify a date before 6/18, or 1903 if they specify a
                // date afterwards.  As a result, 03 is an ambiguous year.  All other
                // two-digit years are unambiguous.
                if(fHaveDefaultCentury) { // check if this formatter even has a pivot year
                    int32_t ambiguousTwoDigitYear = fDefaultCenturyStartYear % 100;
                    ambiguousYear[0] = (value == ambiguousTwoDigitYear);
                    value += (fDefaultCenturyStartYear/100)*100 +
                            (value < ambiguousTwoDigitYear ? 100 : 0);
                }
            }
        }
        cal.set(UCAL_YEAR, value);

        // Delayed checking for adjustment of Hebrew month numbers in non-leap years.
        if (saveHebrewMonth >= 0) {
            HebrewCalendar *hc = (HebrewCalendar*)&cal;
            if (!hc->isLeapYear(value) && saveHebrewMonth >= 6) {
               cal.set(UCAL_MONTH,saveHebrewMonth);
            } else {
               cal.set(UCAL_MONTH,saveHebrewMonth-1);
            }
            saveHebrewMonth = -1;
        }
        return pos.getIndex();

    case UDAT_YEAR_WOY_FIELD:
        // Comment is the same as for UDAT_Year_FIELDs - look above
        if (fDateOverride.compare(hebr)==0 && value < 1000) {
            value += HEBREW_CAL_CUR_MILLENIUM_START_YEAR;
        } else if (text.moveIndex32(start, 2) == pos.getIndex()
            && u_isdigit(text.char32At(start))
            && u_isdigit(text.char32At(text.moveIndex32(start, 1)))
            && fHaveDefaultCentury )
        {
            int32_t ambiguousTwoDigitYear = fDefaultCenturyStartYear % 100;
            ambiguousYear[0] = (value == ambiguousTwoDigitYear);
            value += (fDefaultCenturyStartYear/100)*100 +
                (value < ambiguousTwoDigitYear ? 100 : 0);
        }
        cal.set(UCAL_YEAR_WOY, value);
        return pos.getIndex();

    case UDAT_YEAR_NAME_FIELD:
        if (fSymbols->fShortYearNames != NULL) {
            int32_t newStart = matchString(text, start, UCAL_YEAR, fSymbols->fShortYearNames, fSymbols->fShortYearNamesCount, NULL, cal);
            if (newStart > 0) {
                return newStart;
            }
        }
        if (gotNumber && (getBooleanAttribute(UDAT_PARSE_ALLOW_NUMERIC,status) || value > fSymbols->fShortYearNamesCount)) {
            cal.set(UCAL_YEAR, value);
            return pos.getIndex();
        }
        return -start;

    case UDAT_MONTH_FIELD:
    case UDAT_STANDALONE_MONTH_FIELD:
        if (gotNumber) // i.e., M or MM.
        {
            // When parsing month numbers from the Hebrew Calendar, we might need to adjust the month depending on whether
            // or not it was a leap year.  We may or may not yet know what year it is, so might have to delay checking until
            // the year is parsed.
            if (!strcmp(cal.getType(),"hebrew")) {
                HebrewCalendar *hc = (HebrewCalendar*)&cal;
                if (cal.isSet(UCAL_YEAR)) {
                   UErrorCode monthStatus = U_ZERO_ERROR;
                   if (!hc->isLeapYear(hc->get(UCAL_YEAR, monthStatus)) && value >= 6) {
                       cal.set(UCAL_MONTH, value);
                   } else {
                       cal.set(UCAL_MONTH, value - 1);
                   }
                } else {
                    saveHebrewMonth = value;
                }
            } else {
                // Don't want to parse the month if it is a string
                // while pattern uses numeric style: M/MM, L/LL
                // [We computed 'value' above.]
                cal.set(UCAL_MONTH, value - 1);
            }
            return pos.getIndex();
        } else {
            // count >= 3 // i.e., MMM/MMMM, LLL/LLLL
            // Want to be able to parse both short and long forms.
            // Try count == 4 first:
            UnicodeString * wideMonthPat = NULL;
            UnicodeString * shortMonthPat = NULL;
            if (fSymbols->fLeapMonthPatterns != NULL && fSymbols->fLeapMonthPatternsCount >= DateFormatSymbols::kMonthPatternsCount) {
                if (patternCharIndex==UDAT_MONTH_FIELD) {
                    wideMonthPat = &fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternFormatWide];
                    shortMonthPat = &fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternFormatAbbrev];
                } else {
                    wideMonthPat = &fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternStandaloneWide];
                    shortMonthPat = &fSymbols->fLeapMonthPatterns[DateFormatSymbols::kLeapMonthPatternStandaloneAbbrev];
                }
            }
            int32_t newStart = 0;
            if (patternCharIndex==UDAT_MONTH_FIELD) {
                if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 4) {
                    newStart = matchString(text, start, UCAL_MONTH, fSymbols->fMonths, fSymbols->fMonthsCount, wideMonthPat, cal); // try MMMM
                    if (newStart > 0) {
                        return newStart;
                    }
                }
                if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 3) {
                    newStart = matchString(text, start, UCAL_MONTH, fSymbols->fShortMonths, fSymbols->fShortMonthsCount, shortMonthPat, cal); // try MMM
                }
            } else {
                if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 4) {
                    newStart = matchString(text, start, UCAL_MONTH, fSymbols->fStandaloneMonths, fSymbols->fStandaloneMonthsCount, wideMonthPat, cal); // try LLLL
                    if (newStart > 0) {
                        return newStart;
                    }
                }
                if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 3) {
                    newStart = matchString(text, start, UCAL_MONTH, fSymbols->fStandaloneShortMonths, fSymbols->fStandaloneShortMonthsCount, shortMonthPat, cal); // try LLL
                }
            }
            if (newStart > 0 || !getBooleanAttribute(UDAT_PARSE_ALLOW_NUMERIC, status))  // currently we do not try to parse MMMMM/LLLLL: #8860
                return newStart;
            // else we allowing parsing as number, below
        }
        break;

    case UDAT_HOUR_OF_DAY1_FIELD:
        // [We computed 'value' above.]
        if (value == cal.getMaximum(UCAL_HOUR_OF_DAY) + 1)
            value = 0;

        // fall through to set field
        U_FALLTHROUGH;
    case UDAT_HOUR_OF_DAY0_FIELD:
        cal.set(UCAL_HOUR_OF_DAY, value);
        return pos.getIndex();

    case UDAT_FRACTIONAL_SECOND_FIELD:
        // Fractional seconds left-justify
        i = countDigits(text, start, pos.getIndex());
        if (i < 3) {
            while (i < 3) {
                value *= 10;
                i++;
            }
        } else {
            int32_t a = 1;
            while (i > 3) {
                a *= 10;
                i--;
            }
            value /= a;
        }
        cal.set(UCAL_MILLISECOND, value);
        return pos.getIndex();

    case UDAT_DOW_LOCAL_FIELD:
        if (gotNumber) // i.e., e or ee
        {
            // [We computed 'value' above.]
            cal.set(UCAL_DOW_LOCAL, value);
            return pos.getIndex();
        }
        // else for eee-eeeee fall through to handling of EEE-EEEEE
        // fall through, do not break here
        U_FALLTHROUGH;
    case UDAT_DAY_OF_WEEK_FIELD:
        {
            // Want to be able to parse both short and long forms.
            // Try count == 4 (EEEE) wide first:
            int32_t newStart = 0;
            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 4) {
                if ((newStart = matchString(text, start, UCAL_DAY_OF_WEEK,
                                          fSymbols->fWeekdays, fSymbols->fWeekdaysCount, NULL, cal)) > 0)
                    return newStart;
            }
            // EEEE wide failed, now try EEE abbreviated
            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 3) {
                if ((newStart = matchString(text, start, UCAL_DAY_OF_WEEK,
                                       fSymbols->fShortWeekdays, fSymbols->fShortWeekdaysCount, NULL, cal)) > 0)
                    return newStart;
            }
            // EEE abbreviated failed, now try EEEEEE short
            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 6) {
                if ((newStart = matchString(text, start, UCAL_DAY_OF_WEEK,
                                       fSymbols->fShorterWeekdays, fSymbols->fShorterWeekdaysCount, NULL, cal)) > 0)
                    return newStart;
            }
            // EEEEEE short failed, now try EEEEE narrow
            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 5) {
                if ((newStart = matchString(text, start, UCAL_DAY_OF_WEEK,
                                       fSymbols->fNarrowWeekdays, fSymbols->fNarrowWeekdaysCount, NULL, cal)) > 0)
                    return newStart;
            }
            if (!getBooleanAttribute(UDAT_PARSE_ALLOW_NUMERIC, status) || patternCharIndex == UDAT_DAY_OF_WEEK_FIELD)
                return newStart;
            // else we allowing parsing as number, below
        }
        break;

    case UDAT_STANDALONE_DAY_FIELD:
        {
            if (gotNumber) // c or cc
            {
                // [We computed 'value' above.]
                cal.set(UCAL_DOW_LOCAL, value);
                return pos.getIndex();
            }
            // Want to be able to parse both short and long forms.
            // Try count == 4 (cccc) first:
            int32_t newStart = 0;
            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 4) {
                if ((newStart = matchString(text, start, UCAL_DAY_OF_WEEK,
                                      fSymbols->fStandaloneWeekdays, fSymbols->fStandaloneWeekdaysCount, NULL, cal)) > 0)
                    return newStart;
            }
            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 3) {
                if ((newStart = matchString(text, start, UCAL_DAY_OF_WEEK,
                                          fSymbols->fStandaloneShortWeekdays, fSymbols->fStandaloneShortWeekdaysCount, NULL, cal)) > 0)
                    return newStart;
            }
            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 6) {
                if ((newStart = matchString(text, start, UCAL_DAY_OF_WEEK,
                                          fSymbols->fStandaloneShorterWeekdays, fSymbols->fStandaloneShorterWeekdaysCount, NULL, cal)) > 0)
                    return newStart;
            }
            if (!getBooleanAttribute(UDAT_PARSE_ALLOW_NUMERIC, status))
                return newStart;
            // else we allowing parsing as number, below
        }
        break;

    case UDAT_AM_PM_FIELD:
        {
            // optionally try both wide/abbrev and narrow forms
            int32_t newStart = 0;
            // try wide/abbrev
            if( getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count < 5 ) {
                if ((newStart = matchString(text, start, UCAL_AM_PM, fSymbols->fAmPms, fSymbols->fAmPmsCount, NULL, cal)) > 0) {
                    return newStart;
                }
            }
            // try narrow
            if( getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count >= 5 ) {
                if ((newStart = matchString(text, start, UCAL_AM_PM, fSymbols->fNarrowAmPms, fSymbols->fNarrowAmPmsCount, NULL, cal)) > 0) {
                    return newStart;
                }
            }
            // no matches for given options
            return -start;
        }

    case UDAT_HOUR1_FIELD:
        // [We computed 'value' above.]
        if (value == cal.getLeastMaximum(UCAL_HOUR)+1)
            value = 0;

        // fall through to set field
        U_FALLTHROUGH;
    case UDAT_HOUR0_FIELD:
        cal.set(UCAL_HOUR, value);
        return pos.getIndex();

    case UDAT_QUARTER_FIELD:
        if (gotNumber) // i.e., Q or QQ.
        {
            // Don't want to parse the month if it is a string
            // while pattern uses numeric style: Q or QQ.
            // [We computed 'value' above.]
            cal.set(UCAL_MONTH, (value - 1) * 3);
            return pos.getIndex();
        } else {
            // count >= 3 // i.e., QQQ or QQQQ
            // Want to be able to parse both short and long forms.
            // Try count == 4 first:
            int32_t newStart = 0;

            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 4) {
                if ((newStart = matchQuarterString(text, start, UCAL_MONTH,
                                      fSymbols->fQuarters, fSymbols->fQuartersCount, cal)) > 0)
                    return newStart;
            }
            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 3) {
                if ((newStart = matchQuarterString(text, start, UCAL_MONTH,
                                          fSymbols->fShortQuarters, fSymbols->fShortQuartersCount, cal)) > 0)
                    return newStart;
            }
            if (!getBooleanAttribute(UDAT_PARSE_ALLOW_NUMERIC, status))
                return newStart;
            // else we allowing parsing as number, below
            if(!getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status))
                return -start;
        }
        break;

    case UDAT_STANDALONE_QUARTER_FIELD:
        if (gotNumber) // i.e., q or qq.
        {
            // Don't want to parse the month if it is a string
            // while pattern uses numeric style: q or q.
            // [We computed 'value' above.]
            cal.set(UCAL_MONTH, (value - 1) * 3);
            return pos.getIndex();
        } else {
            // count >= 3 // i.e., qqq or qqqq
            // Want to be able to parse both short and long forms.
            // Try count == 4 first:
            int32_t newStart = 0;

            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 4) {
                if ((newStart = matchQuarterString(text, start, UCAL_MONTH,
                                      fSymbols->fStandaloneQuarters, fSymbols->fStandaloneQuartersCount, cal)) > 0)
                    return newStart;
            }
            if(getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 3) {
                if ((newStart = matchQuarterString(text, start, UCAL_MONTH,
                                          fSymbols->fStandaloneShortQuarters, fSymbols->fStandaloneShortQuartersCount, cal)) > 0)
                    return newStart;
            }
            if (!getBooleanAttribute(UDAT_PARSE_ALLOW_NUMERIC, status))
                return newStart;
            // else we allowing parsing as number, below
            if(!getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status))
                return -start;
        }
        break;

    case UDAT_TIMEZONE_FIELD: // 'z'
        {
            UTimeZoneFormatStyle style = (count < 4) ? UTZFMT_STYLE_SPECIFIC_SHORT : UTZFMT_STYLE_SPECIFIC_LONG;
            const TimeZoneFormat *tzfmt = tzFormat(status);
            if (U_SUCCESS(status)) {
                TimeZone *tz = tzfmt->parse(style, text, pos, tzTimeType);
                if (tz != NULL) {
                    cal.adoptTimeZone(tz);
                    return pos.getIndex();
                }
            }
            return -start;
    }
        break;
    case UDAT_TIMEZONE_RFC_FIELD: // 'Z'
        {
            UTimeZoneFormatStyle style = (count < 4) ?
                UTZFMT_STYLE_ISO_BASIC_LOCAL_FULL : ((count == 5) ? UTZFMT_STYLE_ISO_EXTENDED_FULL: UTZFMT_STYLE_LOCALIZED_GMT);
            const TimeZoneFormat *tzfmt = tzFormat(status);
            if (U_SUCCESS(status)) {
                TimeZone *tz = tzfmt->parse(style, text, pos, tzTimeType);
                if (tz != NULL) {
                    cal.adoptTimeZone(tz);
                    return pos.getIndex();
                }
            }
            return -start;
        }
    case UDAT_TIMEZONE_GENERIC_FIELD: // 'v'
        {
            UTimeZoneFormatStyle style = (count < 4) ? UTZFMT_STYLE_GENERIC_SHORT : UTZFMT_STYLE_GENERIC_LONG;
            const TimeZoneFormat *tzfmt = tzFormat(status);
            if (U_SUCCESS(status)) {
                TimeZone *tz = tzfmt->parse(style, text, pos, tzTimeType);
                if (tz != NULL) {
                    cal.adoptTimeZone(tz);
                    return pos.getIndex();
                }
            }
            return -start;
        }
    case UDAT_TIMEZONE_SPECIAL_FIELD: // 'V'
        {
            UTimeZoneFormatStyle style;
            switch (count) {
            case 1:
                style = UTZFMT_STYLE_ZONE_ID_SHORT;
                break;
            case 2:
                style = UTZFMT_STYLE_ZONE_ID;
                break;
            case 3:
                style = UTZFMT_STYLE_EXEMPLAR_LOCATION;
                break;
            default:
                style = UTZFMT_STYLE_GENERIC_LOCATION;
                break;
            }
            const TimeZoneFormat *tzfmt = tzFormat(status);
            if (U_SUCCESS(status)) {
                TimeZone *tz = tzfmt->parse(style, text, pos, tzTimeType);
                if (tz != NULL) {
                    cal.adoptTimeZone(tz);
                    return pos.getIndex();
                }
            }
            return -start;
        }
    case UDAT_TIMEZONE_LOCALIZED_GMT_OFFSET_FIELD: // 'O'
        {
            UTimeZoneFormatStyle style = (count < 4) ? UTZFMT_STYLE_LOCALIZED_GMT_SHORT : UTZFMT_STYLE_LOCALIZED_GMT;
            const TimeZoneFormat *tzfmt = tzFormat(status);
            if (U_SUCCESS(status)) {
                TimeZone *tz = tzfmt->parse(style, text, pos, tzTimeType);
                if (tz != NULL) {
                    cal.adoptTimeZone(tz);
                    return pos.getIndex();
                }
            }
            return -start;
        }
    case UDAT_TIMEZONE_ISO_FIELD: // 'X'
        {
            UTimeZoneFormatStyle style;
            switch (count) {
            case 1:
                style = UTZFMT_STYLE_ISO_BASIC_SHORT;
                break;
            case 2:
                style = UTZFMT_STYLE_ISO_BASIC_FIXED;
                break;
            case 3:
                style = UTZFMT_STYLE_ISO_EXTENDED_FIXED;
                break;
            case 4:
                style = UTZFMT_STYLE_ISO_BASIC_FULL;
                break;
            default:
                style = UTZFMT_STYLE_ISO_EXTENDED_FULL;
                break;
            }
            const TimeZoneFormat *tzfmt = tzFormat(status);
            if (U_SUCCESS(status)) {
                TimeZone *tz = tzfmt->parse(style, text, pos, tzTimeType);
                if (tz != NULL) {
                    cal.adoptTimeZone(tz);
                    return pos.getIndex();
                }
            }
            return -start;
        }
    case UDAT_TIMEZONE_ISO_LOCAL_FIELD: // 'x'
        {
            UTimeZoneFormatStyle style;
            switch (count) {
            case 1:
                style = UTZFMT_STYLE_ISO_BASIC_LOCAL_SHORT;
                break;
            case 2:
                style = UTZFMT_STYLE_ISO_BASIC_LOCAL_FIXED;
                break;
            case 3:
                style = UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FIXED;
                break;
            case 4:
                style = UTZFMT_STYLE_ISO_BASIC_LOCAL_FULL;
                break;
            default:
                style = UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FULL;
                break;
            }
            const TimeZoneFormat *tzfmt = tzFormat(status);
            if (U_SUCCESS(status)) {
                TimeZone *tz = tzfmt->parse(style, text, pos, tzTimeType);
                if (tz != NULL) {
                    cal.adoptTimeZone(tz);
                    return pos.getIndex();
                }
            }
            return -start;
        }
    // currently no pattern character is defined for UDAT_TIME_SEPARATOR_FIELD
    // so we should not get here. Leave support in for future definition.
    case UDAT_TIME_SEPARATOR_FIELD:
        {
            static const UChar def_sep = DateFormatSymbols::DEFAULT_TIME_SEPARATOR;
            static const UChar alt_sep = DateFormatSymbols::ALTERNATE_TIME_SEPARATOR;

            // Try matching a time separator.
            int32_t count_sep = 1;
            UnicodeString data[3];
            fSymbols->getTimeSeparatorString(data[0]);

            // Add the default, if different from the locale.
            if (data[0].compare(&def_sep, 1) != 0) {
                data[count_sep++].setTo(def_sep);
            }

            // If lenient, add also the alternate, if different from the locale.
            if (isLenient() && data[0].compare(&alt_sep, 1) != 0) {
                data[count_sep++].setTo(alt_sep);
            }

            return matchString(text, start, UCAL_FIELD_COUNT /* => nothing to set */, data, count_sep, NULL, cal);
        }

    case UDAT_AM_PM_MIDNIGHT_NOON_FIELD:
    {
        U_ASSERT(dayPeriod != NULL);
        int32_t ampmStart = subParse(text, start, 0x61, count,
                           obeyCount, allowNegative, ambiguousYear, saveHebrewMonth, cal,
                           patLoc, numericLeapMonthFormatter, tzTimeType);

        if (ampmStart > 0) {
            return ampmStart;
        } else {
            int32_t newStart = 0;

            // Only match the first two strings from the day period strings array.
            if (getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 3) {
                if ((newStart = matchDayPeriodStrings(text, start, fSymbols->fAbbreviatedDayPeriods,
                                                        2, *dayPeriod)) > 0) {
                    return newStart;
                }
            }
            if (getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 5) {
                if ((newStart = matchDayPeriodStrings(text, start, fSymbols->fNarrowDayPeriods,
                                                        2, *dayPeriod)) > 0) {
                    return newStart;
                }
            }
            // count == 4, but allow other counts
            if (getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status)) {
                if ((newStart = matchDayPeriodStrings(text, start, fSymbols->fWideDayPeriods,
                                                        2, *dayPeriod)) > 0) {
                    return newStart;
                }
            }

            return -start;
        }
    }

    case UDAT_FLEXIBLE_DAY_PERIOD_FIELD:
    {
        U_ASSERT(dayPeriod != NULL);
        int32_t newStart = 0;

        if (getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 3) {
            if ((newStart = matchDayPeriodStrings(text, start, fSymbols->fAbbreviatedDayPeriods,
                                fSymbols->fAbbreviatedDayPeriodsCount, *dayPeriod)) > 0) {
                return newStart;
            }
        }
        if (getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 5) {
            if ((newStart = matchDayPeriodStrings(text, start, fSymbols->fNarrowDayPeriods,
                                fSymbols->fNarrowDayPeriodsCount, *dayPeriod)) > 0) {
                return newStart;
            }
        }
        if (getBooleanAttribute(UDAT_PARSE_MULTIPLE_PATTERNS_FOR_MATCH, status) || count == 4) {
            if ((newStart = matchDayPeriodStrings(text, start, fSymbols->fWideDayPeriods,
                                fSymbols->fWideDayPeriodsCount, *dayPeriod)) > 0) {
                return newStart;
            }
        }

        return -start;
    }

    default:
        // Handle "generic" fields
        // this is now handled below, outside the switch block
        break;
    }
    // Handle "generic" fields:
    // switch default case now handled here (outside switch block) to allow
    // parsing of some string fields as digits for lenient case

    int32_t parseStart = pos.getIndex();
    const UnicodeString* src;
    if (obeyCount) {
        if ((start+count) > text.length()) {
            return -start;
        }
        text.extractBetween(0, start + count, temp);
        src = &temp;
    } else {
        src = &text;
    }
    parseInt(*src, number, pos, allowNegative,currentNumberFormat);
    if (pos.getIndex() != parseStart) {
        int32_t val = number.getLong();

        // Don't need suffix processing here (as in number processing at the beginning of the function);
        // the new fields being handled as numeric values (month, weekdays, quarters) should not have suffixes.

        if (!getBooleanAttribute(UDAT_PARSE_ALLOW_NUMERIC, status)) {
            // Check the range of the value
            int32_t bias = gFieldRangeBias[patternCharIndex];
            if (bias >= 0 && (val > cal.getMaximum(field) + bias || val < cal.getMinimum(field) + bias)) {
                return -start;
            }
        }

        // For the following, need to repeat some of the "if (gotNumber)" code above:
        // UDAT_[STANDALONE_]MONTH_FIELD, UDAT_DOW_LOCAL_FIELD, UDAT_STANDALONE_DAY_FIELD,
        // UDAT_[STANDALONE_]QUARTER_FIELD
        switch (patternCharIndex) {
        case UDAT_MONTH_FIELD:
            // See notes under UDAT_MONTH_FIELD case above
            if (!strcmp(cal.getType(),"hebrew")) {
                HebrewCalendar *hc = (HebrewCalendar*)&cal;
                if (cal.isSet(UCAL_YEAR)) {
                   UErrorCode monthStatus = U_ZERO_ERROR;
                   if (!hc->isLeapYear(hc->get(UCAL_YEAR, monthStatus)) && val >= 6) {
                       cal.set(UCAL_MONTH, val);
                   } else {
                       cal.set(UCAL_MONTH, val - 1);
                   }
                } else {
                    saveHebrewMonth = val;
                }
            } else {
                cal.set(UCAL_MONTH, val - 1);
            }
            break;
        case UDAT_STANDALONE_MONTH_FIELD:
            cal.set(UCAL_MONTH, val - 1);
            break;
        case UDAT_DOW_LOCAL_FIELD:
        case UDAT_STANDALONE_DAY_FIELD:
            cal.set(UCAL_DOW_LOCAL, val);
            break;
        case UDAT_QUARTER_FIELD:
        case UDAT_STANDALONE_QUARTER_FIELD:
             cal.set(UCAL_MONTH, (val - 1) * 3);
             break;
        case UDAT_RELATED_YEAR_FIELD:
            cal.setRelatedYear(val);
            break;
        default:
            cal.set(field, val);
            break;
        }
        return pos.getIndex();
    }
    return -start;
}

/**
 * Parse an integer using fNumberFormat.  This method is semantically
 * const, but actually may modify fNumberFormat.
 */
void SimpleDateFormat::parseInt(const UnicodeString& text,
                                Formattable& number,
                                ParsePosition& pos,
                                UBool allowNegative,
                                const NumberFormat *fmt) const {
    parseInt(text, number, -1, pos, allowNegative,fmt);
}

/**
 * Parse an integer using fNumberFormat up to maxDigits.
 */
void SimpleDateFormat::parseInt(const UnicodeString& text,
                                Formattable& number,
                                int32_t maxDigits,
                                ParsePosition& pos,
                                UBool allowNegative,
                                const NumberFormat *fmt) const {
    UnicodeString oldPrefix;
    auto* fmtAsDF = dynamic_cast<const DecimalFormat*>(fmt);
    LocalPointer<DecimalFormat> df;
    if (!allowNegative && fmtAsDF != nullptr) {
        df.adoptInstead(fmtAsDF->clone());
        if (df.isNull()) {
            // Memory allocation error
            return;
        }
        df->setNegativePrefix(UnicodeString(TRUE, SUPPRESS_NEGATIVE_PREFIX, -1));
        fmt = df.getAlias();
    }
    int32_t oldPos = pos.getIndex();
    fmt->parse(text, number, pos);

    if (maxDigits > 0) {
        // adjust the result to fit into
        // the maxDigits and move the position back
        int32_t nDigits = pos.getIndex() - oldPos;
        if (nDigits > maxDigits) {
            int32_t val = number.getLong();
            nDigits -= maxDigits;
            while (nDigits > 0) {
                val /= 10;
                nDigits--;
            }
            pos.setIndex(oldPos + maxDigits);
            number.setLong(val);
        }
    }
}

int32_t SimpleDateFormat::countDigits(const UnicodeString& text, int32_t start, int32_t end) const {
    int32_t numDigits = 0;
    int32_t idx = start;
    while (idx < end) {
        UChar32 cp = text.char32At(idx);
        if (u_isdigit(cp)) {
            numDigits++;
        }
        idx += U16_LENGTH(cp);
    }
    return numDigits;
}

//----------------------------------------------------------------------

void SimpleDateFormat::translatePattern(const UnicodeString& originalPattern,
                                        UnicodeString& translatedPattern,
                                        const UnicodeString& from,
                                        const UnicodeString& to,
                                        UErrorCode& status)
{
    // run through the pattern and convert any pattern symbols from the version
    // in "from" to the corresponding character in "to".  This code takes
    // quoted strings into account (it doesn't try to translate them), and it signals
    // an error if a particular "pattern character" doesn't appear in "from".
    // Depending on the values of "from" and "to" this can convert from generic
    // to localized patterns or localized to generic.
    if (U_FAILURE(status)) {
        return;
    }

    translatedPattern.remove();
    UBool inQuote = FALSE;
    for (int32_t i = 0; i < originalPattern.length(); ++i) {
        UChar c = originalPattern[i];
        if (inQuote) {
            if (c == QUOTE) {
                inQuote = FALSE;
            }
        } else {
            if (c == QUOTE) {
                inQuote = TRUE;
            } else if (isSyntaxChar(c)) {
                int32_t ci = from.indexOf(c);
                if (ci == -1) {
                    status = U_INVALID_FORMAT_ERROR;
                    return;
                }
                c = to[ci];
            }
        }
        translatedPattern += c;
    }
    if (inQuote) {
        status = U_INVALID_FORMAT_ERROR;
        return;
    }
}

//----------------------------------------------------------------------

UnicodeString&
SimpleDateFormat::toPattern(UnicodeString& result) const
{
    result = fPattern;
    return result;
}

//----------------------------------------------------------------------

UnicodeString&
SimpleDateFormat::toLocalizedPattern(UnicodeString& result,
                                     UErrorCode& status) const
{
    translatePattern(fPattern, result,
                     UnicodeString(DateFormatSymbols::getPatternUChars()),
                     fSymbols->fLocalPatternChars, status);
    return result;
}

//----------------------------------------------------------------------

void
SimpleDateFormat::applyPattern(const UnicodeString& pattern)
{
    fPattern = pattern;
    parsePattern();

    // Hack to update use of Gannen year numbering for ja@calendar=japanese -
    // use only if format is non-numeric (includes 年) and no other fDateOverride.
    if (fCalendar != nullptr && uprv_strcmp(fCalendar->getType(),"japanese") == 0 &&
            uprv_strcmp(fLocale.getLanguage(),"ja") == 0) {
        if (fDateOverride==UnicodeString(u"y=jpanyear") && !fHasHanYearChar) {
            // Gannen numbering is set but new pattern should not use it, unset;
            // use procedure from adoptNumberFormat to clear overrides
            if (fSharedNumberFormatters) {
                freeSharedNumberFormatters(fSharedNumberFormatters);
                fSharedNumberFormatters = NULL;
            }
            fDateOverride.setToBogus(); // record status
        } else if (fDateOverride.isBogus() && fHasHanYearChar) {
            // No current override (=> no Gannen numbering) but new pattern needs it;
            // use procedures from initNUmberFormatters / adoptNumberFormat
            umtx_lock(&LOCK);
            if (fSharedNumberFormatters == NULL) {
                fSharedNumberFormatters = allocSharedNumberFormatters();
            }
            umtx_unlock(&LOCK);
            if (fSharedNumberFormatters != NULL) {
                Locale ovrLoc(fLocale.getLanguage(),fLocale.getCountry(),fLocale.getVariant(),"numbers=jpanyear");
                UErrorCode status = U_ZERO_ERROR;
                const SharedNumberFormat *snf = createSharedNumberFormat(ovrLoc, status);
                if (U_SUCCESS(status)) {
                    // Now that we have an appropriate number formatter, fill in the
                    // appropriate slot in the number formatters table.
                    UDateFormatField patternCharIndex = DateFormatSymbols::getPatternCharIndex(u'y');
                    SharedObject::copyPtr(snf, fSharedNumberFormatters[patternCharIndex]);
                    snf->deleteIfZeroRefCount();
                    fDateOverride.setTo(u"y=jpanyear", -1); // record status
                }
            }
        }
    }
}

//----------------------------------------------------------------------

void
SimpleDateFormat::applyLocalizedPattern(const UnicodeString& pattern,
                                        UErrorCode &status)
{
    translatePattern(pattern, fPattern,
                     fSymbols->fLocalPatternChars,
                     UnicodeString(DateFormatSymbols::getPatternUChars()), status);
}

//----------------------------------------------------------------------

const DateFormatSymbols*
SimpleDateFormat::getDateFormatSymbols() const
{
    return fSymbols;
}

//----------------------------------------------------------------------

void
SimpleDateFormat::adoptDateFormatSymbols(DateFormatSymbols* newFormatSymbols)
{
    delete fSymbols;
    fSymbols = newFormatSymbols;
}

//----------------------------------------------------------------------
void
SimpleDateFormat::setDateFormatSymbols(const DateFormatSymbols& newFormatSymbols)
{
    delete fSymbols;
    fSymbols = new DateFormatSymbols(newFormatSymbols);
}

//----------------------------------------------------------------------
const TimeZoneFormat*
SimpleDateFormat::getTimeZoneFormat(void) const {
    // TimeZoneFormat initialization might fail when out of memory.
    // If we always initialize TimeZoneFormat instance, we can return
    // such status there. For now, this implementation lazily instantiates
    // a TimeZoneFormat for performance optimization reasons, but cannot
    // propagate such error (probably just out of memory case) to the caller.
    UErrorCode status = U_ZERO_ERROR;
    return (const TimeZoneFormat*)tzFormat(status);
}

//----------------------------------------------------------------------
void
SimpleDateFormat::adoptTimeZoneFormat(TimeZoneFormat* timeZoneFormatToAdopt)
{
    delete fTimeZoneFormat;
    fTimeZoneFormat = timeZoneFormatToAdopt;
}

//----------------------------------------------------------------------
void
SimpleDateFormat::setTimeZoneFormat(const TimeZoneFormat& newTimeZoneFormat)
{
    delete fTimeZoneFormat;
    fTimeZoneFormat = new TimeZoneFormat(newTimeZoneFormat);
}

//----------------------------------------------------------------------


void SimpleDateFormat::adoptCalendar(Calendar* calendarToAdopt)
{
  UErrorCode status = U_ZERO_ERROR;
  Locale calLocale(fLocale);
  calLocale.setKeywordValue("calendar", calendarToAdopt->getType(), status);
  DateFormatSymbols *newSymbols =
          DateFormatSymbols::createForLocale(calLocale, status);
  if (U_FAILURE(status)) {
      delete calendarToAdopt;
      return;
  }
  DateFormat::adoptCalendar(calendarToAdopt);
  delete fSymbols;
  fSymbols = newSymbols;
  initializeDefaultCentury();  // we need a new century (possibly)
}


//----------------------------------------------------------------------


// override the DateFormat implementation in order to
// lazily initialize fCapitalizationBrkIter
void
SimpleDateFormat::setContext(UDisplayContext value, UErrorCode& status)
{
    DateFormat::setContext(value, status);
#if !UCONFIG_NO_BREAK_ITERATION
    if (U_SUCCESS(status)) {
        if ( fCapitalizationBrkIter == NULL && (value==UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE ||
                value==UDISPCTX_CAPITALIZATION_FOR_UI_LIST_OR_MENU || value==UDISPCTX_CAPITALIZATION_FOR_STANDALONE) ) {
            status = U_ZERO_ERROR;
            fCapitalizationBrkIter = BreakIterator::createSentenceInstance(fLocale, status);
            if (U_FAILURE(status)) {
                delete fCapitalizationBrkIter;
                fCapitalizationBrkIter = NULL;
            }
        }
    }
#endif
}


//----------------------------------------------------------------------


UBool
SimpleDateFormat::isFieldUnitIgnored(UCalendarDateFields field) const {
    return isFieldUnitIgnored(fPattern, field);
}


UBool
SimpleDateFormat::isFieldUnitIgnored(const UnicodeString& pattern,
                                     UCalendarDateFields field) {
    int32_t fieldLevel = fgCalendarFieldToLevel[field];
    int32_t level;
    UChar ch;
    UBool inQuote = FALSE;
    UChar prevCh = 0;
    int32_t count = 0;

    for (int32_t i = 0; i < pattern.length(); ++i) {
        ch = pattern[i];
        if (ch != prevCh && count > 0) {
            level = getLevelFromChar(prevCh);
            // the larger the level, the smaller the field unit.
            if (fieldLevel <= level) {
                return FALSE;
            }
            count = 0;
        }
        if (ch == QUOTE) {
            if ((i+1) < pattern.length() && pattern[i+1] == QUOTE) {
                ++i;
            } else {
                inQuote = ! inQuote;
            }
        }
        else if (!inQuote && isSyntaxChar(ch)) {
            prevCh = ch;
            ++count;
        }
    }
    if (count > 0) {
        // last item
        level = getLevelFromChar(prevCh);
        if (fieldLevel <= level) {
            return FALSE;
        }
    }
    return TRUE;
}

//----------------------------------------------------------------------

const Locale&
SimpleDateFormat::getSmpFmtLocale(void) const {
    return fLocale;
}

//----------------------------------------------------------------------

int32_t
SimpleDateFormat::checkIntSuffix(const UnicodeString& text, int32_t start,
                                 int32_t patLoc, UBool isNegative) const {
    // local variables
    UnicodeString suf;
    int32_t patternMatch;
    int32_t textPreMatch;
    int32_t textPostMatch;

    // check that we are still in range
    if ( (start > text.length()) ||
         (start < 0) ||
         (patLoc < 0) ||
         (patLoc > fPattern.length())) {
        // out of range, don't advance location in text
        return start;
    }

    // get the suffix
    DecimalFormat* decfmt = dynamic_cast<DecimalFormat*>(fNumberFormat);
    if (decfmt != NULL) {
        if (isNegative) {
            suf = decfmt->getNegativeSuffix(suf);
        }
        else {
            suf = decfmt->getPositiveSuffix(suf);
        }
    }

    // check for suffix
    if (suf.length() <= 0) {
        return start;
    }

    // check suffix will be encountered in the pattern
    patternMatch = compareSimpleAffix(suf,fPattern,patLoc);

    // check if a suffix will be encountered in the text
    textPreMatch = compareSimpleAffix(suf,text,start);

    // check if a suffix was encountered in the text
    textPostMatch = compareSimpleAffix(suf,text,start-suf.length());

    // check for suffix match
    if ((textPreMatch >= 0) && (patternMatch >= 0) && (textPreMatch == patternMatch)) {
        return start;
    }
    else if ((textPostMatch >= 0) && (patternMatch >= 0) && (textPostMatch == patternMatch)) {
        return  start - suf.length();
    }

    // should not get here
    return start;
}

//----------------------------------------------------------------------

int32_t
SimpleDateFormat::compareSimpleAffix(const UnicodeString& affix,
                   const UnicodeString& input,
                   int32_t pos) const {
    int32_t start = pos;
    for (int32_t i=0; i<affix.length(); ) {
        UChar32 c = affix.char32At(i);
        int32_t len = U16_LENGTH(c);
        if (PatternProps::isWhiteSpace(c)) {
            // We may have a pattern like: \u200F \u0020
            //        and input text like: \u200F \u0020
            // Note that U+200F and U+0020 are Pattern_White_Space but only
            // U+0020 is UWhiteSpace.  So we have to first do a direct
            // match of the run of Pattern_White_Space in the pattern,
            // then match any extra characters.
            UBool literalMatch = FALSE;
            while (pos < input.length() &&
                   input.char32At(pos) == c) {
                literalMatch = TRUE;
                i += len;
                pos += len;
                if (i == affix.length()) {
                    break;
                }
                c = affix.char32At(i);
                len = U16_LENGTH(c);
                if (!PatternProps::isWhiteSpace(c)) {
                    break;
                }
            }

            // Advance over run in pattern
            i = skipPatternWhiteSpace(affix, i);

            // Advance over run in input text
            // Must see at least one white space char in input,
            // unless we've already matched some characters literally.
            int32_t s = pos;
            pos = skipUWhiteSpace(input, pos);
            if (pos == s && !literalMatch) {
                return -1;
            }

            // If we skip UWhiteSpace in the input text, we need to skip it in the pattern.
            // Otherwise, the previous lines may have skipped over text (such as U+00A0) that
            // is also in the affix.
            i = skipUWhiteSpace(affix, i);
        } else {
            if (pos < input.length() &&
                input.char32At(pos) == c) {
                i += len;
                pos += len;
            } else {
                return -1;
            }
        }
    }
    return pos - start;
}

//----------------------------------------------------------------------

int32_t
SimpleDateFormat::skipPatternWhiteSpace(const UnicodeString& text, int32_t pos) const {
    const UChar* s = text.getBuffer();
    return (int32_t)(PatternProps::skipWhiteSpace(s + pos, text.length() - pos) - s);
}

//----------------------------------------------------------------------

int32_t
SimpleDateFormat::skipUWhiteSpace(const UnicodeString& text, int32_t pos) const {
    while (pos < text.length()) {
        UChar32 c = text.char32At(pos);
        if (!u_isUWhiteSpace(c)) {
            break;
        }
        pos += U16_LENGTH(c);
    }
    return pos;
}

//----------------------------------------------------------------------

// Lazy TimeZoneFormat instantiation, semantically const.
TimeZoneFormat *
SimpleDateFormat::tzFormat(UErrorCode &status) const {
    if (fTimeZoneFormat == NULL) {
        umtx_lock(&LOCK);
        {
            if (fTimeZoneFormat == NULL) {
                TimeZoneFormat *tzfmt = TimeZoneFormat::createInstance(fLocale, status);
                if (U_FAILURE(status)) {
                    return NULL;
                }

                const_cast<SimpleDateFormat *>(this)->fTimeZoneFormat = tzfmt;
            }
        }
        umtx_unlock(&LOCK);
    }
    return fTimeZoneFormat;
}

void SimpleDateFormat::parsePattern() {
    fHasMinute = FALSE;
    fHasSecond = FALSE;
    fHasHanYearChar = FALSE;

    int len = fPattern.length();
    UBool inQuote = FALSE;
    for (int32_t i = 0; i < len; ++i) {
        UChar ch = fPattern[i];
        if (ch == QUOTE) {
            inQuote = !inQuote;
        }
        if (ch == 0x5E74) { // don't care whether this is inside quotes
            fHasHanYearChar = TRUE;
        }
        if (!inQuote) {
            if (ch == 0x6D) {  // 0x6D == 'm'
                fHasMinute = TRUE;
            }
            if (ch == 0x73) {  // 0x73 == 's'
                fHasSecond = TRUE;
            }
        }
    }
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2013, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*
* This file contains the class SimpleDateFormatStaticSets
*
* SimpleDateFormatStaticSets holds the UnicodeSets that are needed for lenient
* parsing of literal characters in date/time strings.
********************************************************************************
*/



#if !UCONFIG_NO_FORMATTING











U_NAMESPACE_BEGIN

SimpleDateFormatStaticSets *gStaticSets = NULL;
UInitOnce gSimpleDateFormatStaticSetsInitOnce = U_INITONCE_INITIALIZER;

SimpleDateFormatStaticSets::SimpleDateFormatStaticSets(UErrorCode &status)
: fDateIgnorables(NULL),
  fTimeIgnorables(NULL),
  fOtherIgnorables(NULL)
{
    fDateIgnorables  = new UnicodeSet(UNICODE_STRING("[-,./[:whitespace:]]", 20), status);
    fTimeIgnorables  = new UnicodeSet(UNICODE_STRING("[-.:[:whitespace:]]", 19),  status);
    fOtherIgnorables = new UnicodeSet(UNICODE_STRING("[:whitespace:]", 14),       status);

    // Check for null pointers
    if (fDateIgnorables == NULL || fTimeIgnorables == NULL || fOtherIgnorables == NULL) {
        goto ExitConstrDeleteAll;
    }

    // Freeze all the sets
    fDateIgnorables->freeze();
    fTimeIgnorables->freeze();
    fOtherIgnorables->freeze();

    return; // If we reached this point, everything is fine so just exit

ExitConstrDeleteAll: // Remove all sets and return error
    delete fDateIgnorables;  fDateIgnorables = NULL;
    delete fTimeIgnorables;  fTimeIgnorables = NULL;
    delete fOtherIgnorables; fOtherIgnorables = NULL;

    status = U_MEMORY_ALLOCATION_ERROR;
}


SimpleDateFormatStaticSets::~SimpleDateFormatStaticSets() {
    delete fDateIgnorables;  fDateIgnorables = NULL;
    delete fTimeIgnorables;  fTimeIgnorables = NULL;
    delete fOtherIgnorables; fOtherIgnorables = NULL;
}


//------------------------------------------------------------------------------
//
//   smpdtfmt_cleanup     Memory cleanup function, free/delete all
//                      cached memory.  Called by ICU's u_cleanup() function.
//
//------------------------------------------------------------------------------
UBool
SimpleDateFormatStaticSets::cleanup(void)
{
    delete gStaticSets;
    gStaticSets = NULL;
    gSimpleDateFormatStaticSetsInitOnce.reset();
    return TRUE;
}

U_CDECL_BEGIN
static UBool U_CALLCONV
smpdtfmt_cleanup(void)
{
    return SimpleDateFormatStaticSets::cleanup();
}

static void U_CALLCONV smpdtfmt_initSets(UErrorCode &status) {
    ucln_i18n_registerCleanup(UCLN_I18N_SMPDTFMT, smpdtfmt_cleanup);
    U_ASSERT(gStaticSets == NULL);
    gStaticSets = new SimpleDateFormatStaticSets(status);
    if (gStaticSets == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
}

U_CDECL_END

UnicodeSet *SimpleDateFormatStaticSets::getIgnorables(UDateFormatField fieldIndex)
{
    UErrorCode status = U_ZERO_ERROR;
    umtx_initOnce(gSimpleDateFormatStaticSetsInitOnce, &smpdtfmt_initSets, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    
    switch (fieldIndex) {
        case UDAT_YEAR_FIELD:
        case UDAT_MONTH_FIELD:
        case UDAT_DATE_FIELD:
        case UDAT_STANDALONE_DAY_FIELD:
        case UDAT_STANDALONE_MONTH_FIELD:
            return gStaticSets->fDateIgnorables;
            
        case UDAT_HOUR_OF_DAY1_FIELD:
        case UDAT_HOUR_OF_DAY0_FIELD:
        case UDAT_MINUTE_FIELD:
        case UDAT_SECOND_FIELD:
        case UDAT_HOUR1_FIELD:
        case UDAT_HOUR0_FIELD:
            return gStaticSets->fTimeIgnorables;
            
        default:
            return gStaticSets->fOtherIgnorables;
    }
}

U_NAMESPACE_END

#endif // #if !UCONFIG_NO_FORMATTING
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1996-2012, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/
//===============================================================================
//
// File sortkey.cpp
//
//
//
// Created by: Helena Shih
//
// Modification History:
//
//  Date         Name          Description
//
//  6/20/97      helena        Java class name change.
//  6/23/97      helena        Added comments to make code more readable.
//  6/26/98      erm           Canged to use byte arrays instead of UnicodeString
//  7/31/98      erm           hashCode: minimum inc should be 2 not 1,
//                             Cleaned up operator=
// 07/12/99      helena        HPUX 11 CC port.
// 03/06/01      synwee        Modified compareTo, to handle the result of
//                             2 string similar in contents, but one is longer
//                             than the other
//===============================================================================



#if !UCONFIG_NO_COLLATION






U_NAMESPACE_BEGIN

// A hash code of kInvalidHashCode indicates that the hash code needs
// to be computed. A hash code of kEmptyHashCode is used for empty keys
// and for any key whose computed hash code is kInvalidHashCode.
static const int32_t kInvalidHashCode = 0;
static const int32_t kEmptyHashCode = 1;
// The "bogus hash code" replaces a separate fBogus flag.
static const int32_t kBogusHashCode = 2;

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(CollationKey)

CollationKey::CollationKey()
    : UObject(), fFlagAndLength(0),
      fHashCode(kEmptyHashCode)
{
}

// Create a collation key from a bit array.
CollationKey::CollationKey(const uint8_t* newValues, int32_t count)
    : UObject(), fFlagAndLength(count),
      fHashCode(kInvalidHashCode)
{
    if (count < 0 || (newValues == NULL && count != 0) ||
            (count > getCapacity() && reallocate(count, 0) == NULL)) {
        setToBogus();
        return;
    }

    if (count > 0) {
        uprv_memcpy(getBytes(), newValues, count);
    }
}

CollationKey::CollationKey(const CollationKey& other)
    : UObject(other), fFlagAndLength(other.getLength()),
      fHashCode(other.fHashCode)
{
    if (other.isBogus())
    {
        setToBogus();
        return;
    }

    int32_t length = fFlagAndLength;
    if (length > getCapacity() && reallocate(length, 0) == NULL) {
        setToBogus();
        return;
    }

    if (length > 0) {
        uprv_memcpy(getBytes(), other.getBytes(), length);
    }
}

CollationKey::~CollationKey()
{
    if(fFlagAndLength < 0) { uprv_free(fUnion.fFields.fBytes); }
}

uint8_t *CollationKey::reallocate(int32_t newCapacity, int32_t length) {
    uint8_t *newBytes = static_cast<uint8_t *>(uprv_malloc(newCapacity));
    if(newBytes == NULL) { return NULL; }
    if(length > 0) {
        uprv_memcpy(newBytes, getBytes(), length);
    }
    if(fFlagAndLength < 0) { uprv_free(fUnion.fFields.fBytes); }
    fUnion.fFields.fBytes = newBytes;
    fUnion.fFields.fCapacity = newCapacity;
    fFlagAndLength |= 0x80000000;
    return newBytes;
}

void CollationKey::setLength(int32_t newLength) {
    // U_ASSERT(newLength >= 0 && newLength <= getCapacity());
    fFlagAndLength = (fFlagAndLength & 0x80000000) | newLength;
    fHashCode = kInvalidHashCode;
}

// set the key to an empty state
CollationKey&
CollationKey::reset()
{
    fFlagAndLength &= 0x80000000;
    fHashCode = kEmptyHashCode;

    return *this;
}

// set the key to a "bogus" or invalid state
CollationKey&
CollationKey::setToBogus()
{
    fFlagAndLength &= 0x80000000;
    fHashCode = kBogusHashCode;

    return *this;
}

UBool
CollationKey::operator==(const CollationKey& source) const
{
    return getLength() == source.getLength() &&
            (this == &source ||
             uprv_memcmp(getBytes(), source.getBytes(), getLength()) == 0);
}

const CollationKey&
CollationKey::operator=(const CollationKey& other)
{
    if (this != &other)
    {
        if (other.isBogus())
        {
            return setToBogus();
        }

        int32_t length = other.getLength();
        if (length > getCapacity() && reallocate(length, 0) == NULL) {
            return setToBogus();
        }
        if (length > 0) {
            uprv_memcpy(getBytes(), other.getBytes(), length);
        }
        fFlagAndLength = (fFlagAndLength & 0x80000000) | length;
        fHashCode = other.fHashCode;
    }

    return *this;
}

// Bitwise comparison for the collation keys.
Collator::EComparisonResult
CollationKey::compareTo(const CollationKey& target) const
{
    UErrorCode errorCode = U_ZERO_ERROR;
    return static_cast<Collator::EComparisonResult>(compareTo(target, errorCode));
}

// Bitwise comparison for the collation keys.
UCollationResult
CollationKey::compareTo(const CollationKey& target, UErrorCode &status) const
{
  if(U_SUCCESS(status)) {
    const uint8_t *src = getBytes();
    const uint8_t *tgt = target.getBytes();

    // are we comparing the same string
    if (src == tgt)
        return  UCOL_EQUAL;

    UCollationResult result;

    // are we comparing different lengths?
    int32_t minLength = getLength();
    int32_t targetLength = target.getLength();
    if (minLength < targetLength) {
        result = UCOL_LESS;
    } else if (minLength == targetLength) {
        result = UCOL_EQUAL;
    } else {
        minLength = targetLength;
        result = UCOL_GREATER;
    }

    if (minLength > 0) {
        int diff = uprv_memcmp(src, tgt, minLength);
        if (diff > 0) {
            return UCOL_GREATER;
        }
        else
            if (diff < 0) {
                return UCOL_LESS;
            }
    }

    return result;
  } else {
    return UCOL_EQUAL;
  }
}

#ifdef U_USE_COLLATION_KEY_DEPRECATES
// Create a copy of the byte array.
uint8_t*
CollationKey::toByteArray(int32_t& count) const
{
    uint8_t *result = (uint8_t*) uprv_malloc( sizeof(uint8_t) * fCount );

    if (result == NULL)
    {
        count = 0;
    }
    else
    {
        count = fCount;
        if (count > 0) {
            uprv_memcpy(result, fBytes, fCount);
        }
    }

    return result;
}
#endif

static int32_t
computeHashCode(const uint8_t *key, int32_t  length) {
    const char *s = reinterpret_cast<const char *>(key);
    int32_t hash;
    if (s == NULL || length == 0) {
        hash = kEmptyHashCode;
    } else {
        hash = ustr_hashCharsN(s, length);
        if (hash == kInvalidHashCode || hash == kBogusHashCode) {
            hash = kEmptyHashCode;
        }
    }
    return hash;
}

int32_t
CollationKey::hashCode() const
{
    // (Cribbed from UnicodeString)
    // We cache the hashCode; when it becomes invalid, due to any change to the
    // string, we note this by setting it to kInvalidHashCode. [LIU]

    // Note: This method is semantically const, but physically non-const.

    if (fHashCode == kInvalidHashCode)
    {
        fHashCode = computeHashCode(getBytes(), getLength());
    }

    return fHashCode;
}

U_NAMESPACE_END

U_CAPI int32_t U_EXPORT2
ucol_keyHashCode(const uint8_t *key, 
                       int32_t  length)
{
    return icu::computeHashCode(key, length);
}

#endif /* #if !UCONFIG_NO_COLLATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2015, International Business Machines Corporation
 * and others. All Rights Reserved.
 *******************************************************************************
 * standardplural.cpp
 *
 * created on: 2015dec14
 * created by: Markus W. Scherer
 */



#if !UCONFIG_NO_FORMATTING






U_NAMESPACE_BEGIN

static const char *gKeywords[StandardPlural::COUNT] = {
    "zero", "one", "two", "few", "many", "other"
};

const char *StandardPlural::getKeyword(Form p) {
    U_ASSERT(ZERO <= p && p < COUNT);
    return gKeywords[p];
}

int32_t StandardPlural::indexOrNegativeFromString(const char *keyword) {
    switch (*keyword++) {
    case 'f':
        if (uprv_strcmp(keyword, "ew") == 0) {
            return FEW;
        }
        break;
    case 'm':
        if (uprv_strcmp(keyword, "any") == 0) {
            return MANY;
        }
        break;
    case 'o':
        if (uprv_strcmp(keyword, "ther") == 0) {
            return OTHER;
        } else if (uprv_strcmp(keyword, "ne") == 0) {
            return ONE;
        }
        break;
    case 't':
        if (uprv_strcmp(keyword, "wo") == 0) {
            return TWO;
        }
        break;
    case 'z':
        if (uprv_strcmp(keyword, "ero") == 0) {
            return ZERO;
        }
        break;
    default:
        break;
    }
    return -1;
}

static const UChar gZero[] = { 0x7A, 0x65, 0x72, 0x6F };
static const UChar gOne[] = { 0x6F, 0x6E, 0x65 };
static const UChar gTwo[] = { 0x74, 0x77, 0x6F };
static const UChar gFew[] = { 0x66, 0x65, 0x77 };
static const UChar gMany[] = { 0x6D, 0x61, 0x6E, 0x79 };
static const UChar gOther[] = { 0x6F, 0x74, 0x68, 0x65, 0x72 };

int32_t StandardPlural::indexOrNegativeFromString(const UnicodeString &keyword) {
    switch (keyword.length()) {
    case 3:
        if (keyword.compare(gOne, 3) == 0) {
            return ONE;
        } else if (keyword.compare(gTwo, 3) == 0) {
            return TWO;
        } else if (keyword.compare(gFew, 3) == 0) {
            return FEW;
        }
        break;
    case 4:
        if (keyword.compare(gMany, 4) == 0) {
            return MANY;
        } else if (keyword.compare(gZero, 4) == 0) {
            return ZERO;
        }
        break;
    case 5:
        if (keyword.compare(gOther, 5) == 0) {
            return OTHER;
        }
        break;
    default:
        break;
    }
    return -1;
}

int32_t StandardPlural::indexFromString(const char *keyword, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return OTHER; }
    int32_t i = indexOrNegativeFromString(keyword);
    if (i >= 0) {
        return i;
    } else {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return OTHER;
    }
}

int32_t StandardPlural::indexFromString(const UnicodeString &keyword, UErrorCode &errorCode) {
    if (U_FAILURE(errorCode)) { return OTHER; }
    int32_t i = indexOrNegativeFromString(keyword);
    if (i >= 0) {
        return i;
    } else {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return OTHER;
    }
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_FORMATTING
// © 2018 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html



#if !UCONFIG_NO_FORMATTING

// Allow implicit conversion from char16_t* to UnicodeString for this file:
// Helpful in toString methods and elsewhere.
#ifndef UNISTR_FROM_STRING_EXPLICIT
#define UNISTR_FROM_STRING_EXPLICIT
#endif







U_NAMESPACE_BEGIN


StringSegment::StringSegment(const UnicodeString& str, bool ignoreCase)
        : fStr(str), fStart(0), fEnd(str.length()),
          fFoldCase(ignoreCase) {}

int32_t StringSegment::getOffset() const {
    return fStart;
}

void StringSegment::setOffset(int32_t start) {
    fStart = start;
}

void StringSegment::adjustOffset(int32_t delta) {
    fStart += delta;
}

void StringSegment::adjustOffsetByCodePoint() {
    fStart += U16_LENGTH(getCodePoint());
}

void StringSegment::setLength(int32_t length) {
    fEnd = fStart + length;
}

void StringSegment::resetLength() {
    fEnd = fStr.length();
}

int32_t StringSegment::length() const {
    return fEnd - fStart;
}

char16_t StringSegment::charAt(int32_t index) const {
    return fStr.charAt(index + fStart);
}

UChar32 StringSegment::codePointAt(int32_t index) const {
    return fStr.char32At(index + fStart);
}

UnicodeString StringSegment::toUnicodeString() const {
    return UnicodeString(fStr.getBuffer() + fStart, fEnd - fStart);
}

const UnicodeString StringSegment::toTempUnicodeString() const {
    // Use the readonly-aliasing constructor for efficiency.
    return UnicodeString(FALSE, fStr.getBuffer() + fStart, fEnd - fStart);
}

UChar32 StringSegment::getCodePoint() const {
    char16_t lead = fStr.charAt(fStart);
    if (U16_IS_LEAD(lead) && fStart + 1 < fEnd) {
        return fStr.char32At(fStart);
    } else if (U16_IS_SURROGATE(lead)) {
        return -1;
    } else {
        return lead;
    }
}

bool StringSegment::startsWith(UChar32 otherCp) const {
    return codePointsEqual(getCodePoint(), otherCp, fFoldCase);
}

bool StringSegment::startsWith(const UnicodeSet& uniset) const {
    // TODO: Move UnicodeSet case-folding logic here.
    // TODO: Handle string matches here instead of separately.
    UChar32 cp = getCodePoint();
    if (cp == -1) {
        return false;
    }
    return uniset.contains(cp);
}

bool StringSegment::startsWith(const UnicodeString& other) const {
    if (other.isBogus() || other.length() == 0 || length() == 0) {
        return false;
    }
    int cp1 = getCodePoint();
    int cp2 = other.char32At(0);
    return codePointsEqual(cp1, cp2, fFoldCase);
}

int32_t StringSegment::getCommonPrefixLength(const UnicodeString& other) {
    return getPrefixLengthInternal(other, fFoldCase);
}

int32_t StringSegment::getCaseSensitivePrefixLength(const UnicodeString& other) {
    return getPrefixLengthInternal(other, false);
}

int32_t StringSegment::getPrefixLengthInternal(const UnicodeString& other, bool foldCase) {
    U_ASSERT(other.length() > 0);
    int32_t offset = 0;
    for (; offset < uprv_min(length(), other.length());) {
        // TODO: case-fold code points, not chars
        char16_t c1 = charAt(offset);
        char16_t c2 = other.charAt(offset);
        if (!codePointsEqual(c1, c2, foldCase)) {
            break;
        }
        offset++;
    }
    return offset;
}

bool StringSegment::codePointsEqual(UChar32 cp1, UChar32 cp2, bool foldCase) {
    if (cp1 == cp2) {
        return true;
    }
    if (!foldCase) {
        return false;
    }
    cp1 = u_foldCase(cp1, TRUE);
    cp2 = u_foldCase(cp2, TRUE);
    return cp1 == cp2;
}

bool StringSegment::operator==(const UnicodeString& other) const {
    return toTempUnicodeString() == other;
}


U_NAMESPACE_END
#endif /* #if !UCONFIG_NO_FORMATTING */
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 2001-2014 IBM and others. All rights reserved.
// **********************************************************************
// *   Date        Name        Description
// *  03/22/2000   helena      Creation.
// **********************************************************************
// */

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_COLLATION && !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/stsearch.h"
// #include "usrchimp.h"
// #include "cmemory.h"

// U_NAMESPACE_BEGIN

// UOBJECT_DEFINE_RTTI_IMPLEMENTATION(StringSearch)

// // public constructors and destructors -----------------------------------

// StringSearch::StringSearch(const UnicodeString &pattern,
//                            const UnicodeString &text,
//                            const Locale        &locale,
//                                  BreakIterator *breakiter,
//                                  UErrorCode    &status) :
//                            SearchIterator(text, breakiter),
//                            m_pattern_(pattern)
// {
//     if (U_FAILURE(status)) {
//         m_strsrch_ = NULL;
//         return;
//     }

//     m_strsrch_ = usearch_open(m_pattern_.getBuffer(), m_pattern_.length(),
//                               m_text_.getBuffer(), m_text_.length(),
//                               locale.getName(), (UBreakIterator *)breakiter,
//                               &status);
//     uprv_free(m_search_);
//     m_search_ = NULL;

//     if (U_SUCCESS(status)) {
//         // m_search_ has been created by the base SearchIterator class
//         m_search_        = m_strsrch_->search;
//     }
// }

// StringSearch::StringSearch(const UnicodeString     &pattern,
//                            const UnicodeString     &text,
//                                  RuleBasedCollator *coll,
//                                  BreakIterator     *breakiter,
//                                  UErrorCode        &status) :
//                            SearchIterator(text, breakiter),
//                            m_pattern_(pattern)
// {
//     if (U_FAILURE(status)) {
//         m_strsrch_ = NULL;
//         return;
//     }
//     if (coll == NULL) {
//         status     = U_ILLEGAL_ARGUMENT_ERROR;
//         m_strsrch_ = NULL;
//         return;
//     }
//     m_strsrch_ = usearch_openFromCollator(m_pattern_.getBuffer(),
//                                           m_pattern_.length(),
//                                           m_text_.getBuffer(),
//                                           m_text_.length(), coll->toUCollator(),
//                                           (UBreakIterator *)breakiter,
//                                           &status);
//     uprv_free(m_search_);
//     m_search_ = NULL;

//     if (U_SUCCESS(status)) {
//         // m_search_ has been created by the base SearchIterator class
//         m_search_ = m_strsrch_->search;
//     }
// }

// StringSearch::StringSearch(const UnicodeString     &pattern,
//                                  CharacterIterator &text,
//                            const Locale            &locale,
//                                  BreakIterator     *breakiter,
//                                  UErrorCode        &status) :
//                            SearchIterator(text, breakiter),
//                            m_pattern_(pattern)
// {
//     if (U_FAILURE(status)) {
//         m_strsrch_ = NULL;
//         return;
//     }
//     m_strsrch_ = usearch_open(m_pattern_.getBuffer(), m_pattern_.length(),
//                               m_text_.getBuffer(), m_text_.length(),
//                               locale.getName(), (UBreakIterator *)breakiter,
//                               &status);
//     uprv_free(m_search_);
//     m_search_ = NULL;

//     if (U_SUCCESS(status)) {
//         // m_search_ has been created by the base SearchIterator class
//         m_search_ = m_strsrch_->search;
//     }
// }

// StringSearch::StringSearch(const UnicodeString     &pattern,
//                                  CharacterIterator &text,
//                                  RuleBasedCollator *coll,
//                                  BreakIterator     *breakiter,
//                                  UErrorCode        &status) :
//                            SearchIterator(text, breakiter),
//                            m_pattern_(pattern)
// {
//     if (U_FAILURE(status)) {
//         m_strsrch_ = NULL;
//         return;
//     }
//     if (coll == NULL) {
//         status     = U_ILLEGAL_ARGUMENT_ERROR;
//         m_strsrch_ = NULL;
//         return;
//     }
//     m_strsrch_ = usearch_openFromCollator(m_pattern_.getBuffer(),
//                                           m_pattern_.length(),
//                                           m_text_.getBuffer(),
//                                           m_text_.length(), coll->toUCollator(),
//                                           (UBreakIterator *)breakiter,
//                                           &status);
//     uprv_free(m_search_);
//     m_search_ = NULL;

//     if (U_SUCCESS(status)) {
//         // m_search_ has been created by the base SearchIterator class
//         m_search_ = m_strsrch_->search;
//     }
// }

// StringSearch::StringSearch(const StringSearch &that) :
//                        SearchIterator(that.m_text_, that.m_breakiterator_),
//                        m_pattern_(that.m_pattern_)
// {
//     UErrorCode status = U_ZERO_ERROR;

//     // Free m_search_ from the superclass
//     uprv_free(m_search_);
//     m_search_ = NULL;

//     if (that.m_strsrch_ == NULL) {
//         // This was not a good copy
//         m_strsrch_ = NULL;
//     }
//     else {
//         // Make a deep copy
//         m_strsrch_ = usearch_openFromCollator(m_pattern_.getBuffer(),
//                                               m_pattern_.length(),
//                                               m_text_.getBuffer(),
//                                               m_text_.length(),
//                                               that.m_strsrch_->collator,
//                                              (UBreakIterator *)that.m_breakiterator_,
//                                               &status);
//         if (U_SUCCESS(status)) {
//             // m_search_ has been created by the base SearchIterator class
//             m_search_        = m_strsrch_->search;
//         }
//     }
// }

// StringSearch::~StringSearch()
// {
//     if (m_strsrch_ != NULL) {
//         usearch_close(m_strsrch_);
//         m_search_ = NULL;
//     }
// }

// StringSearch *
// StringSearch::clone() const {
//     return new StringSearch(*this);
// }

// // operator overloading ---------------------------------------------
// StringSearch & StringSearch::operator=(const StringSearch &that)
// {
//     if ((*this) != that) {
//         UErrorCode status = U_ZERO_ERROR;
//         m_text_          = that.m_text_;
//         m_breakiterator_ = that.m_breakiterator_;
//         m_pattern_       = that.m_pattern_;
//         // all m_search_ in the parent class is linked up with m_strsrch_
//         usearch_close(m_strsrch_);
//         m_strsrch_ = usearch_openFromCollator(m_pattern_.getBuffer(),
//                                               m_pattern_.length(),
//                                               m_text_.getBuffer(),
//                                               m_text_.length(),
//                                               that.m_strsrch_->collator,
//                                               NULL, &status);
//         // Check null pointer
//         if (m_strsrch_ != NULL) {
//             m_search_ = m_strsrch_->search;
//         }
//     }
//     return *this;
// }

// UBool StringSearch::operator==(const SearchIterator &that) const
// {
//     if (this == &that) {
//         return TRUE;
//     }
//     if (SearchIterator::operator ==(that)) {
//         StringSearch &thatsrch = (StringSearch &)that;
//         return (this->m_pattern_ == thatsrch.m_pattern_ &&
//                 this->m_strsrch_->collator == thatsrch.m_strsrch_->collator);
//     }
//     return FALSE;
// }

// // public get and set methods ----------------------------------------

// void StringSearch::setOffset(int32_t position, UErrorCode &status)
// {
//     // status checked in usearch_setOffset
//     usearch_setOffset(m_strsrch_, position, &status);
// }

// int32_t StringSearch::getOffset(void) const
// {
//     return usearch_getOffset(m_strsrch_);
// }

// void StringSearch::setText(const UnicodeString &text, UErrorCode &status)
// {
//     if (U_SUCCESS(status)) {
//         m_text_ = text;
//         usearch_setText(m_strsrch_, text.getBuffer(), text.length(), &status);
//     }
// }

// void StringSearch::setText(CharacterIterator &text, UErrorCode &status)
// {
//     if (U_SUCCESS(status)) {
//         text.getText(m_text_);
//         usearch_setText(m_strsrch_, m_text_.getBuffer(), m_text_.length(), &status);
//     }
// }

// RuleBasedCollator * StringSearch::getCollator() const
// {
//     // Note the const_cast. It would be cleaner if this const method returned a const collator.
//     return RuleBasedCollator::rbcFromUCollator(const_cast<UCollator *>(m_strsrch_->collator));
// }

// void StringSearch::setCollator(RuleBasedCollator *coll, UErrorCode &status)
// {
//     if (U_SUCCESS(status)) {
//         usearch_setCollator(m_strsrch_, coll->toUCollator(), &status);
//     }
// }

// void StringSearch::setPattern(const UnicodeString &pattern,
//                                     UErrorCode    &status)
// {
//     if (U_SUCCESS(status)) {
//         m_pattern_ = pattern;
//         usearch_setPattern(m_strsrch_, m_pattern_.getBuffer(), m_pattern_.length(),
//                            &status);
//     }
// }

// const UnicodeString & StringSearch::getPattern() const
// {
//     return m_pattern_;
// }

// // public methods ----------------------------------------------------

// void StringSearch::reset()
// {
//     usearch_reset(m_strsrch_);
// }

// StringSearch * StringSearch::safeClone() const
// {
//     UErrorCode status = U_ZERO_ERROR;
//     StringSearch *result = new StringSearch(m_pattern_, m_text_,
//                                             getCollator(),
//                                             m_breakiterator_,
//                                             status);
//     /* test for NULL */
//     if (result == 0) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//         return 0;
//     }
//     result->setOffset(getOffset(), status);
//     result->setMatchStart(m_strsrch_->search->matchedIndex);
//     result->setMatchLength(m_strsrch_->search->matchedLength);
//     if (U_FAILURE(status)) {
//         return NULL;
//     }
//     return result;
// }

// // protected method -------------------------------------------------

// int32_t StringSearch::handleNext(int32_t position, UErrorCode &status)
// {
//     // values passed here are already in the pre-shift position
//     if (U_SUCCESS(status)) {
//         if (m_strsrch_->pattern.cesLength == 0) {
//             m_search_->matchedIndex =
//                                     m_search_->matchedIndex == USEARCH_DONE ?
//                                     getOffset() : m_search_->matchedIndex + 1;
//             m_search_->matchedLength = 0;
//             ucol_setOffset(m_strsrch_->textIter, m_search_->matchedIndex,
//                            &status);
//             if (m_search_->matchedIndex == m_search_->textLength) {
//                 m_search_->matchedIndex = USEARCH_DONE;
//             }
//         }
//         else {
//             // looking at usearch.cpp, this part is shifted out to
//             // StringSearch instead of SearchIterator because m_strsrch_ is
//             // not accessible in SearchIterator
// #if 0
//             if (position + m_strsrch_->pattern.defaultShiftSize
//                 > m_search_->textLength) {
//                 setMatchNotFound();
//                 return USEARCH_DONE;
//             }
// #endif
//             if (m_search_->matchedLength <= 0) {
//                 // the flipping direction issue has already been handled
//                 // in next()
//                 // for boundary check purposes. this will ensure that the
//                 // next match will not preceed the current offset
//                 // note search->matchedIndex will always be set to something
//                 // in the code
//                 m_search_->matchedIndex = position - 1;
//             }

//             ucol_setOffset(m_strsrch_->textIter, position, &status);

// #if 0
//             for (;;) {
//                 if (m_search_->isCanonicalMatch) {
//                     // can't use exact here since extra accents are allowed.
//                     usearch_handleNextCanonical(m_strsrch_, &status);
//                 }
//                 else {
//                     usearch_handleNextExact(m_strsrch_, &status);
//                 }
//                 if (U_FAILURE(status)) {
//                     return USEARCH_DONE;
//                 }
//                 if (m_breakiterator_ == NULL
// #if !UCONFIG_NO_BREAK_ITERATION
//                     ||
//                     m_search_->matchedIndex == USEARCH_DONE ||
//                     (m_breakiterator_->isBoundary(m_search_->matchedIndex) &&
//                      m_breakiterator_->isBoundary(m_search_->matchedIndex +
//                                                   m_search_->matchedLength))
// #endif
//                 ) {
//                     if (m_search_->matchedIndex == USEARCH_DONE) {
//                         ucol_setOffset(m_strsrch_->textIter,
//                                        m_search_->textLength, &status);
//                     }
//                     else {
//                         ucol_setOffset(m_strsrch_->textIter,
//                                        m_search_->matchedIndex, &status);
//                     }
//                     return m_search_->matchedIndex;
//                 }
//             }
// #else
//             // if m_strsrch_->breakIter is always the same as m_breakiterator_
//             // then we don't need to check the match boundaries here because
//             // usearch_handleNextXXX will already have done it.
//             if (m_search_->isCanonicalMatch) {
//             	// *could* actually use exact here 'cause no extra accents allowed...
//             	usearch_handleNextCanonical(m_strsrch_, &status);
//             } else {
//             	usearch_handleNextExact(m_strsrch_, &status);
//             }

//             if (U_FAILURE(status)) {
//             	return USEARCH_DONE;
//             }

//             if (m_search_->matchedIndex == USEARCH_DONE) {
//             	ucol_setOffset(m_strsrch_->textIter, m_search_->textLength, &status);
//             } else {
//             	ucol_setOffset(m_strsrch_->textIter, m_search_->matchedIndex, &status);
//             }

//             return m_search_->matchedIndex;
// #endif
//         }
//     }
//     return USEARCH_DONE;
// }

// int32_t StringSearch::handlePrev(int32_t position, UErrorCode &status)
// {
//     // values passed here are already in the pre-shift position
//     if (U_SUCCESS(status)) {
//         if (m_strsrch_->pattern.cesLength == 0) {
//             m_search_->matchedIndex =
//                   (m_search_->matchedIndex == USEARCH_DONE ? getOffset() :
//                    m_search_->matchedIndex);
//             if (m_search_->matchedIndex == 0) {
//                 setMatchNotFound();
//             }
//             else {
//                 m_search_->matchedIndex --;
//                 ucol_setOffset(m_strsrch_->textIter, m_search_->matchedIndex,
//                                &status);
//                 m_search_->matchedLength = 0;
//             }
//         }
//         else {
//             // looking at usearch.cpp, this part is shifted out to
//             // StringSearch instead of SearchIterator because m_strsrch_ is
//             // not accessible in SearchIterator
// #if 0
//             if (!m_search_->isOverlap &&
//                 position - m_strsrch_->pattern.defaultShiftSize < 0) {
//                 setMatchNotFound();
//                 return USEARCH_DONE;
//             }

//             for (;;) {
//                 if (m_search_->isCanonicalMatch) {
//                     // can't use exact here since extra accents are allowed.
//                     usearch_handlePreviousCanonical(m_strsrch_, &status);
//                 }
//                 else {
//                     usearch_handlePreviousExact(m_strsrch_, &status);
//                 }
//                 if (U_FAILURE(status)) {
//                     return USEARCH_DONE;
//                 }
//                 if (m_breakiterator_ == NULL
// #if !UCONFIG_NO_BREAK_ITERATION
//                     ||
//                     m_search_->matchedIndex == USEARCH_DONE ||
//                     (m_breakiterator_->isBoundary(m_search_->matchedIndex) &&
//                      m_breakiterator_->isBoundary(m_search_->matchedIndex +
//                                                   m_search_->matchedLength))
// #endif
//                 ) {
//                     return m_search_->matchedIndex;
//                 }
//             }
// #else
//             ucol_setOffset(m_strsrch_->textIter, position, &status);

//             if (m_search_->isCanonicalMatch) {
//             	// *could* use exact match here since extra accents *not* allowed!
//             	usearch_handlePreviousCanonical(m_strsrch_, &status);
//             } else {
//             	usearch_handlePreviousExact(m_strsrch_, &status);
//             }

//             if (U_FAILURE(status)) {
//             	return USEARCH_DONE;
//             }

//             return m_search_->matchedIndex;
// #endif
//         }

//         return m_search_->matchedIndex;
//     }
//     return USEARCH_DONE;
// }

// U_NAMESPACE_END

// #endif /* #if !UCONFIG_NO_COLLATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2003-2013, International Business Machines Corporation and
 * others. All Rights Reserved.
 *******************************************************************************
 *
 * File TAIWNCAL.CPP
 *
 * Modification History:
 *  05/13/2003    srl     copied from gregocal.cpp
 *  06/29/2007    srl     copied from buddhcal.cpp
 *  05/12/2008    jce     modified to use calendar=roc per CLDR
 *
 */



#if !UCONFIG_NO_FORMATTING




#include <float.h>

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(TaiwanCalendar)

static const int32_t kTaiwanEraStart = 1911;  // 1911 (Gregorian)

static const int32_t taiwncal_kGregorianEpoch = 1970;

TaiwanCalendar::TaiwanCalendar(const Locale& aLocale, UErrorCode& success)
:   GregorianCalendar(aLocale, success)
{
    setTimeInMillis(getNow(), success); // Call this again now that the vtable is set up properly.
}

TaiwanCalendar::~TaiwanCalendar()
{
}

TaiwanCalendar::TaiwanCalendar(const TaiwanCalendar& source)
: GregorianCalendar(source)
{
}

TaiwanCalendar& TaiwanCalendar::operator= ( const TaiwanCalendar& right)
{
    GregorianCalendar::operator=(right);
    return *this;
}

TaiwanCalendar* TaiwanCalendar::clone() const
{
    return new TaiwanCalendar(*this);
}

const char *TaiwanCalendar::getType() const
{
    return "roc";
}

int32_t TaiwanCalendar::handleGetExtendedYear()
{
    // EXTENDED_YEAR in TaiwanCalendar is a Gregorian year
    // The default value of EXTENDED_YEAR is 1970 (Minguo 59)
    int32_t year = taiwncal_kGregorianEpoch;

    if (newerField(UCAL_EXTENDED_YEAR, UCAL_YEAR) == UCAL_EXTENDED_YEAR
        && newerField(UCAL_EXTENDED_YEAR, UCAL_ERA) == UCAL_EXTENDED_YEAR) {
        year = internalGet(UCAL_EXTENDED_YEAR, taiwncal_kGregorianEpoch);
    } else {
        int32_t era = internalGet(UCAL_ERA, MINGUO);
        if(era == MINGUO) {
            year =     internalGet(UCAL_YEAR, 1) + kTaiwanEraStart;
        } else if(era == BEFORE_MINGUO) {
            year = 1 - internalGet(UCAL_YEAR, 1) + kTaiwanEraStart;
        }
    }
    return year;
}

void TaiwanCalendar::handleComputeFields(int32_t julianDay, UErrorCode& status)
{
    GregorianCalendar::handleComputeFields(julianDay, status);
    int32_t y = internalGet(UCAL_EXTENDED_YEAR) - kTaiwanEraStart;
    if(y>0) {
        internalSet(UCAL_ERA, MINGUO);
        internalSet(UCAL_YEAR, y);
    } else {
        internalSet(UCAL_ERA, BEFORE_MINGUO);
        internalSet(UCAL_YEAR, 1-y);
    }
}

int32_t TaiwanCalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const
{
    if(field == UCAL_ERA) {
        if(limitType == UCAL_LIMIT_MINIMUM || limitType == UCAL_LIMIT_GREATEST_MINIMUM) {
            return BEFORE_MINGUO;
        } else {
            return MINGUO;
        }
    } else {
        return GregorianCalendar::handleGetLimit(field,limitType);
    }
}

#if 0
void TaiwanCalendar::timeToFields(UDate theTime, UBool quick, UErrorCode& status)
{
    //Calendar::timeToFields(theTime, quick, status);

    int32_t era = internalGet(UCAL_ERA);
    int32_t year = internalGet(UCAL_YEAR);

    if(era == GregorianCalendar::BC) {
        year = 1-year;
        era = TaiwanCalendar::MINGUO;
    } else if(era == GregorianCalendar::AD) {
        era = TaiwanCalendar::MINGUO;
    } else {
        status = U_INTERNAL_PROGRAM_ERROR;
    }

    year = year - kTaiwanEraStart;

    internalSet(UCAL_ERA, era);
    internalSet(UCAL_YEAR, year);
}
#endif

/**
 * The system maintains a static default century start date and Year.  They are
 * initialized the first time they are used.  Once the system default century date
 * and year are set, they do not change.
 */
static UDate           taiwncal_gSystemDefaultCenturyStart       = DBL_MIN;
static int32_t         taiwncal_gSystemDefaultCenturyStartYear   = -1;
static icu::UInitOnce  taiwncal_gSystemDefaultCenturyInit        = U_INITONCE_INITIALIZER;

UBool TaiwanCalendar::haveDefaultCentury() const
{
    return TRUE;
}

static void U_CALLCONV taiwncal_initializeSystemDefaultCentury()
{
    // initialize systemDefaultCentury and systemDefaultCenturyYear based
    // on the current time.  They'll be set to 80 years before
    // the current time.
    UErrorCode status = U_ZERO_ERROR;
    TaiwanCalendar calendar(Locale("@calendar=roc"),status);
    if (U_SUCCESS(status))
    {
        calendar.setTime(Calendar::getNow(), status);
        calendar.add(UCAL_YEAR, -80, status);

        taiwncal_gSystemDefaultCenturyStart = calendar.getTime(status);
        taiwncal_gSystemDefaultCenturyStartYear = calendar.get(UCAL_YEAR, status);
    }
    // We have no recourse upon failure unless we want to propagate the failure
    // out.
}

UDate TaiwanCalendar::defaultCenturyStart() const {
    // lazy-evaluate systemDefaultCenturyStart
    umtx_initOnce(taiwncal_gSystemDefaultCenturyInit, &taiwncal_initializeSystemDefaultCentury);
    return taiwncal_gSystemDefaultCenturyStart;
}

int32_t TaiwanCalendar::defaultCenturyStartYear() const {
    // lazy-evaluate systemDefaultCenturyStartYear
    umtx_initOnce(taiwncal_gSystemDefaultCenturyInit, &taiwncal_initializeSystemDefaultCentury);
    return taiwncal_gSystemDefaultCenturyStartYear;
}

U_NAMESPACE_END

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 1997-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File TIMEZONE.CPP
*
* Modification History:
*
*   Date        Name        Description
*   12/05/96    clhuang     Creation.
*   04/21/97    aliu        General clean-up and bug fixing.
*   05/08/97    aliu        Fixed Hashtable code per code review.
*   07/09/97    helena      Changed createInstance to createDefault.
*   07/29/97    aliu        Updated with all-new list of 96 UNIX-derived
*                           TimeZones.  Changed mechanism to load from static
*                           array rather than resource bundle.
*   07/07/1998  srl         Bugfixes from the Java side: UTC GMT CAT NST
*                           Added getDisplayName API
*                           going to add custom parsing.
*
*                           ISSUES:
*                               - should getDisplayName cache something?
*                               - should custom time zones be cached? [probably]
*  08/10/98     stephen     Brought getDisplayName() API in-line w/ conventions
*  08/19/98     stephen     Changed createTimeZone() to never return 0
*  09/02/98     stephen     Added getOffset(monthLen) and hasSameRules()
*  09/15/98     stephen     Added getStaticClassID()
*  02/22/99     stephen     Removed character literals for EBCDIC safety
*  05/04/99     stephen     Changed initDefault() for Mutex issues
*  07/12/99     helena      HPUX 11 CC Port.
*  12/03/99     aliu        Moved data out of static table into icudata.dll.
*                           Substantial rewrite of zone lookup, default zone, and
*                           available IDs code.  Misc. cleanup.
*********************************************************************************/

  // for 'typeid' to work






#ifdef U_DEBUG_TZ
# include <stdio.h>
 // for debugging

static void debug_tz_loc(const char *f, int32_t l)
{
  fprintf(stderr, "%s:%d: ", f, l);
}

static void debug_tz_msg(const char *pat, ...)
{
  va_list ap;
  va_start(ap, pat);
  vfprintf(stderr, pat, ap);
  fflush(stderr);
}
static char gStrBuf[256];
#define U_DEBUG_TZ_STR(x) u_austrncpy(gStrBuf,x,sizeof(gStrBuf)-1)
// must use double parens, i.e.:  U_DEBUG_TZ_MSG(("four is: %d",4));
#define U_DEBUG_TZ_MSG(x) {debug_tz_loc(__FILE__,__LINE__);debug_tz_msg x;}
#else
#define U_DEBUG_TZ_MSG(x)
#endif

#if !UCONFIG_NO_FORMATTING








 // struct UResourceBundle










#define timezone_kZONEINFO "zoneinfo64"
#define timezone_kREGIONS  "Regions"
#define timezone_kZONES    "Zones"
#define timezone_kRULES    "Rules"
#define timezone_kNAMES    "Names"
#define timezone_kTZVERSION  "TZVersion"
#define timezone_kLINKS    "links"
#define timezone_kMAX_CUSTOM_HOUR    23
#define timezone_kMAX_CUSTOM_MIN     59
#define timezone_kMAX_CUSTOM_SEC     59
#define timezone_MINUS 0x002D
#define timezone_PLUS 0x002B
#define timezone_ZERO_DIGIT 0x0030
#define timezone_COLON 0x003A

// Static data and constants

static const UChar         timezone_WORLD[] = {0x30, 0x30, 0x31, 0x00}; /* "001" */

static const UChar         timezone_GMT_ID[] = {0x47, 0x4D, 0x54, 0x00}; /* "GMT" */
static const UChar         timezone_UNKNOWN_ZONE_ID[] = {0x45, 0x74, 0x63, 0x2F, 0x55, 0x6E, 0x6B, 0x6E, 0x6F, 0x77, 0x6E, 0x00}; /* "Etc/Unknown" */
static const int32_t       timezone_GMT_ID_LENGTH = 3;
static const int32_t       timezone_UNKNOWN_ZONE_ID_LENGTH = 11;

static icu::TimeZone* timezone_DEFAULT_ZONE = NULL;
static icu::UInitOnce gDefaultZoneInitOnce = U_INITONCE_INITIALIZER;

alignas(icu::SimpleTimeZone)
static char gRawGMT[sizeof(icu::SimpleTimeZone)];

alignas(icu::SimpleTimeZone)
static char gRawUNKNOWN[sizeof(icu::SimpleTimeZone)];

static icu::UInitOnce gStaticZonesInitOnce = U_INITONCE_INITIALIZER;
static UBool gStaticZonesInitialized = FALSE; // Whether the static zones are initialized and ready to use.

static char TZDATA_VERSION[16];
static icu::UInitOnce gTZDataVersionInitOnce = U_INITONCE_INITIALIZER;

static int32_t* MAP_SYSTEM_ZONES = NULL;
static int32_t* MAP_CANONICAL_SYSTEM_ZONES = NULL;
static int32_t* MAP_CANONICAL_SYSTEM_LOCATION_ZONES = NULL;

static int32_t LEN_SYSTEM_ZONES = 0;
static int32_t LEN_CANONICAL_SYSTEM_ZONES = 0;
static int32_t LEN_CANONICAL_SYSTEM_LOCATION_ZONES = 0;

static icu::UInitOnce gSystemZonesInitOnce = U_INITONCE_INITIALIZER;
static icu::UInitOnce gCanonicalZonesInitOnce = U_INITONCE_INITIALIZER;
static icu::UInitOnce gCanonicalLocationZonesInitOnce = U_INITONCE_INITIALIZER;

U_CDECL_BEGIN
static UBool U_CALLCONV timeZone_cleanup(void)
{
    U_NAMESPACE_USE
    delete timezone_DEFAULT_ZONE;
    timezone_DEFAULT_ZONE = NULL;
    gDefaultZoneInitOnce.reset();

    if (gStaticZonesInitialized) {
        reinterpret_cast<SimpleTimeZone*>(gRawGMT)->~SimpleTimeZone();
        reinterpret_cast<SimpleTimeZone*>(gRawUNKNOWN)->~SimpleTimeZone();
        gStaticZonesInitialized = FALSE;
        gStaticZonesInitOnce.reset();
    }

    uprv_memset(TZDATA_VERSION, 0, sizeof(TZDATA_VERSION));
    gTZDataVersionInitOnce.reset();

    LEN_SYSTEM_ZONES = 0;
    uprv_free(MAP_SYSTEM_ZONES);
    MAP_SYSTEM_ZONES = 0;
    gSystemZonesInitOnce.reset();

    LEN_CANONICAL_SYSTEM_ZONES = 0;
    uprv_free(MAP_CANONICAL_SYSTEM_ZONES);
    MAP_CANONICAL_SYSTEM_ZONES = 0;
    gCanonicalZonesInitOnce.reset();

    LEN_CANONICAL_SYSTEM_LOCATION_ZONES = 0;
    uprv_free(MAP_CANONICAL_SYSTEM_LOCATION_ZONES);
    MAP_CANONICAL_SYSTEM_LOCATION_ZONES = 0;
    gCanonicalLocationZonesInitOnce.reset();

    return TRUE;
}
U_CDECL_END

U_NAMESPACE_BEGIN

static int32_t findInStringArray(UResourceBundle* array, const UnicodeString& id, UErrorCode &status)
{
    UnicodeString copy;
    const UChar *u;
    int32_t len;

    int32_t start = 0;
    int32_t limit = ures_getSize(array);
    int32_t mid;
    int32_t lastMid = INT32_MAX;
    if(U_FAILURE(status) || (limit < 1)) {
        return -1;
    }
    U_DEBUG_TZ_MSG(("fisa: Looking for %s, between %d and %d\n", U_DEBUG_TZ_STR(UnicodeString(id).getTerminatedBuffer()), start, limit));

    for (;;) {
        mid = (int32_t)((start + limit) / 2);
        if (lastMid == mid) {   /* Have we moved? */
            break;  /* We haven't moved, and it wasn't found. */
        }
        lastMid = mid;
        u = ures_getStringByIndex(array, mid, &len, &status);
        if (U_FAILURE(status)) {
            break;
        }
        U_DEBUG_TZ_MSG(("tz: compare to %s, %d .. [%d] .. %d\n", U_DEBUG_TZ_STR(u), start, mid, limit));
        copy.setTo(TRUE, u, len);
        int r = id.compare(copy);
        if(r==0) {
            U_DEBUG_TZ_MSG(("fisa: found at %d\n", mid));
            return mid;
        } else if(r<0) {
            limit = mid;
        } else {
            start = mid;
        }
    }
    U_DEBUG_TZ_MSG(("fisa: not found\n"));
    return -1;
}

/**
 * Fetch a specific zone by name.  Replaces the getByKey call.
 * @param top Top timezone resource
 * @param id Time zone ID
 * @param oldbundle Bundle for reuse (or NULL).   see 'ures_open()'
 * @return the zone's bundle if found, or undefined if error.  Reuses oldbundle.
 */
static UResourceBundle* getZoneByName(const UResourceBundle* top, const UnicodeString& id, UResourceBundle *oldbundle, UErrorCode& status) {
    // load the Rules object
    UResourceBundle *tmp = ures_getByKey(top, timezone_kNAMES, NULL, &status);

    // search for the string
    int32_t idx = findInStringArray(tmp, id, status);

    if((idx == -1) && U_SUCCESS(status)) {
        // not found
        status = U_MISSING_RESOURCE_ERROR;
        //ures_close(oldbundle);
        //oldbundle = NULL;
    } else {
        U_DEBUG_TZ_MSG(("gzbn: oldbundle= size %d, type %d, %s\n", ures_getSize(tmp), ures_getType(tmp), u_errorName(status)));
        tmp = ures_getByKey(top, timezone_kZONES, tmp, &status); // get Zones object from top
        U_DEBUG_TZ_MSG(("gzbn: loaded ZONES, size %d, type %d, path %s %s\n", ures_getSize(tmp), ures_getType(tmp), ures_getPath(tmp), u_errorName(status)));
        oldbundle = ures_getByIndex(tmp, idx, oldbundle, &status); // get nth Zone object
        U_DEBUG_TZ_MSG(("gzbn: loaded z#%d, size %d, type %d, path %s, %s\n", idx, ures_getSize(oldbundle), ures_getType(oldbundle), ures_getPath(oldbundle),  u_errorName(status)));
    }
    ures_close(tmp);
    if(U_FAILURE(status)) {
        //ures_close(oldbundle);
        return NULL;
    } else {
        return oldbundle;
    }
}


UResourceBundle* TimeZone::loadRule(const UResourceBundle* top, const UnicodeString& ruleid, UResourceBundle* oldbundle, UErrorCode& status) {
    char key[64];
    ruleid.extract(0, sizeof(key)-1, key, (int32_t)sizeof(key)-1, US_INV);
    U_DEBUG_TZ_MSG(("loadRule(%s)\n", key));
    UResourceBundle *r = ures_getByKey(top, timezone_kRULES, oldbundle, &status);
    U_DEBUG_TZ_MSG(("loadRule(%s) -> timezone_kRULES [%s]\n", key, u_errorName(status)));
    r = ures_getByKey(r, key, r, &status);
    U_DEBUG_TZ_MSG(("loadRule(%s) -> item [%s]\n", key, u_errorName(status)));
    return r;
}

/**
 * Given an ID, open the appropriate resource for the given time zone.
 * Dereference aliases if necessary.
 * @param id zone id
 * @param res resource, which must be ready for use (initialized but not open)
 * @param ec input-output error code
 * @return top-level resource bundle
 */
static UResourceBundle* openOlsonResource(const UnicodeString& id,
                                          UResourceBundle& res,
                                          UErrorCode& ec)
{
#ifdef U_DEBUG_TZ
    char buf[128];
    id.extract(0, sizeof(buf)-1, buf, sizeof(buf), "");
#endif
    UResourceBundle *top = ures_openDirect(0, timezone_kZONEINFO, &ec);
    U_DEBUG_TZ_MSG(("pre: res sz=%d\n", ures_getSize(&res)));
    /* &res = */ getZoneByName(top, id, &res, ec);
    // Dereference if this is an alias.  Docs say result should be 1
    // but it is 0 in 2.8 (?).
    U_DEBUG_TZ_MSG(("Loading zone '%s' (%s, size %d) - %s\n", buf, ures_getKey((UResourceBundle*)&res), ures_getSize(&res), u_errorName(ec)));
    if (ures_getType(&res) == URES_INT) {
        int32_t deref = ures_getInt(&res, &ec) + 0;
        U_DEBUG_TZ_MSG(("getInt: %s - type is %d\n", u_errorName(ec), ures_getType(&res)));
        UResourceBundle *ares = ures_getByKey(top, timezone_kZONES, NULL, &ec); // dereference Zones section
        ures_getByIndex(ares, deref, &res, &ec);
        ures_close(ares);
        U_DEBUG_TZ_MSG(("alias to #%d (%s) - %s\n", deref, "??", u_errorName(ec)));
    } else {
        U_DEBUG_TZ_MSG(("not an alias - size %d\n", ures_getSize(&res)));
    }
    U_DEBUG_TZ_MSG(("%s - final status is %s\n", buf, u_errorName(ec)));
    return top;
}

// -------------------------------------

namespace {

void U_CALLCONV initStaticTimeZones() {
    // Initialize _GMT independently of other static data; it should
    // be valid even if we can't load the time zone UDataMemory.
    ucln_i18n_registerCleanup(UCLN_I18N_TIMEZONE, timeZone_cleanup);

    // new can't fail below, as we use placement new into staticly allocated space.
    new(gRawGMT) SimpleTimeZone(0, UnicodeString(TRUE, timezone_GMT_ID, timezone_GMT_ID_LENGTH));
    new(gRawUNKNOWN) SimpleTimeZone(0, UnicodeString(TRUE, timezone_UNKNOWN_ZONE_ID, timezone_UNKNOWN_ZONE_ID_LENGTH));

    gStaticZonesInitialized = TRUE;
}

}  // anonymous namespace

const TimeZone& U_EXPORT2
TimeZone::getUnknown()
{
    umtx_initOnce(gStaticZonesInitOnce, &initStaticTimeZones);
    return *reinterpret_cast<SimpleTimeZone*>(gRawUNKNOWN);
}

const TimeZone* U_EXPORT2
TimeZone::getGMT(void)
{
    umtx_initOnce(gStaticZonesInitOnce, &initStaticTimeZones);
    return reinterpret_cast<SimpleTimeZone*>(gRawGMT);
}

// *****************************************************************************
// class TimeZone
// *****************************************************************************

UOBJECT_DEFINE_ABSTRACT_RTTI_IMPLEMENTATION(TimeZone)

TimeZone::TimeZone()
    :   UObject(), fID()
{
}

// -------------------------------------

TimeZone::TimeZone(const UnicodeString &id)
    :   UObject(), fID(id)
{
}

// -------------------------------------

TimeZone::~TimeZone()
{
}

// -------------------------------------

TimeZone::TimeZone(const TimeZone &source)
    :   UObject(source), fID(source.fID)
{
}

// -------------------------------------

TimeZone &
TimeZone::operator=(const TimeZone &right)
{
    if (this != &right) fID = right.fID;
    return *this;
}

// -------------------------------------

UBool
TimeZone::operator==(const TimeZone& that) const
{
    return typeid(*this) == typeid(that) &&
        fID == that.fID;
}

// -------------------------------------

namespace {
TimeZone*
createSystemTimeZone(const UnicodeString& id, UErrorCode& ec) {
    if (U_FAILURE(ec)) {
        return NULL;
    }
    TimeZone* z = 0;
    StackUResourceBundle res;
    U_DEBUG_TZ_MSG(("pre-err=%s\n", u_errorName(ec)));
    UResourceBundle *top = openOlsonResource(id, res.ref(), ec);
    U_DEBUG_TZ_MSG(("post-err=%s\n", u_errorName(ec)));
    if (U_SUCCESS(ec)) {
        z = new OlsonTimeZone(top, res.getAlias(), id, ec);
        if (z == NULL) {
            ec = U_MEMORY_ALLOCATION_ERROR;
            U_DEBUG_TZ_MSG(("cstz: olson time zone failed to initialize - err %s\n", u_errorName(ec)));
        }
    }
    ures_close(top);
    if (U_FAILURE(ec)) {
        U_DEBUG_TZ_MSG(("cstz: failed to create, err %s\n", u_errorName(ec)));
        delete z;
        z = NULL;
    }
    return z;
}

/**
 * Lookup the given name in our system zone table.  If found,
 * instantiate a new zone of that name and return it.  If not
 * found, return 0.
 */
TimeZone*
createSystemTimeZone(const UnicodeString& id) {
    UErrorCode ec = U_ZERO_ERROR;
    return createSystemTimeZone(id, ec);
}

}

TimeZone* U_EXPORT2
TimeZone::createTimeZone(const UnicodeString& ID)
{
    /* We first try to lookup the zone ID in our system list.  If this
     * fails, we try to parse it as a custom string GMT[+-]hh:mm.  If
     * all else fails, we return GMT, which is probably not what the
     * user wants, but at least is a functioning TimeZone object.
     *
     * We cannot return NULL, because that would break compatibility
     * with the JDK.
     */
    TimeZone* result = createSystemTimeZone(ID);

    if (result == NULL) {
        U_DEBUG_TZ_MSG(("failed to load system time zone with id - falling to custom"));
        result = createCustomTimeZone(ID);
    }
    if (result == NULL) {
        U_DEBUG_TZ_MSG(("failed to load time zone with id - falling to Etc/Unknown(GMT)"));
        const TimeZone& unknown = getUnknown();
        // Unknown zone uses staticly allocated memory, so creation of it can never fail due to OOM.
        result = unknown.clone();
    }
    return result;
}

// -------------------------------------

TimeZone* U_EXPORT2
TimeZone::detectHostTimeZone()
{
    // We access system timezone data through uprv_tzset(), uprv_tzname(), and others,
    // which have platform specific implementations in putil.cpp
    int32_t rawOffset = 0;
    const char *hostID;
    UBool hostDetectionSucceeded = TRUE;

    // First, try to create a system timezone, based
    // on the string ID in tzname[0].

    uprv_tzset(); // Initialize tz... system data

    uprv_tzname_clear_cache();

    // Get the timezone ID from the host.  This function should do
    // any required host-specific remapping; e.g., on Windows this
    // function maps the Windows Time Zone name to an ICU timezone ID.
    hostID = uprv_tzname(0);

    // Invert sign because UNIX semantics are backwards
    rawOffset = uprv_timezone() * -U_MILLIS_PER_SECOND;

    TimeZone* hostZone = NULL;

    UnicodeString hostStrID(hostID, -1, US_INV);

    if (hostStrID.length() == 0) {
        // The host time zone detection (or remapping) above has failed and
        // we have no name at all. Fallback to using the Unknown zone.
        hostStrID = UnicodeString(TRUE, timezone_UNKNOWN_ZONE_ID, timezone_UNKNOWN_ZONE_ID_LENGTH);
        hostDetectionSucceeded = FALSE;
    }

    hostZone = createSystemTimeZone(hostStrID);

#if U_PLATFORM_USES_ONLY_WIN32_API
    // hostID points to a heap-allocated location on Windows.
    uprv_free(const_cast<char *>(hostID));
#endif

    int32_t hostIDLen = hostStrID.length();
    if (hostZone != NULL && rawOffset != hostZone->getRawOffset()
        && (3 <= hostIDLen && hostIDLen <= 4))
    {
        // Uh oh. This probably wasn't a good id.
        // It was probably an ambiguous abbreviation
        delete hostZone;
        hostZone = NULL;
    }

    // Construct a fixed standard zone with the host's ID
    // and raw offset.
    if (hostZone == NULL && hostDetectionSucceeded) {
        hostZone = new SimpleTimeZone(rawOffset, hostStrID);
    }

    // If we _still_ don't have a time zone, use the Unknown zone.
    //
    // Note: This is extremely unlikely situation. If
    // new SimpleTimeZone(...) above fails, the following
    // code may also fail.
    if (hostZone == NULL) {
        // Unknown zone uses static allocated memory, so it must always exist.
        // However, clone() allocates memory and can fail.
        hostZone = TimeZone::getUnknown().clone();
    }

    return hostZone;
}

// -------------------------------------

static UMutex gDefaultZoneMutex;

/**
 * Initialize timezone_DEFAULT_ZONE from the system default time zone.
 * Upon return, timezone_DEFAULT_ZONE will not be NULL, unless operator new()
 * returns NULL.
 */
static void U_CALLCONV initDefault()
{
    ucln_i18n_registerCleanup(UCLN_I18N_TIMEZONE, timeZone_cleanup);

    Mutex lock(&gDefaultZoneMutex);
    // If setDefault() has already been called we can skip getting the
    // default zone information from the system.
    if (timezone_DEFAULT_ZONE != NULL) {
        return;
    }

    // NOTE:  this code is safely single threaded, being only
    // run via umtx_initOnce().
    //
    // Some of the locale/timezone OS functions may not be thread safe,
    //
    // The operating system might actually use ICU to implement timezones.
    // So we may have ICU calling ICU here, like on AIX.
    // There shouldn't be a problem with this; initOnce does not hold a mutex
    // while the init function is being run.

    // The code detecting the host time zone was separated from this
    // and implemented as TimeZone::detectHostTimeZone()

    TimeZone *default_zone = TimeZone::detectHostTimeZone();

    U_ASSERT(timezone_DEFAULT_ZONE == NULL);

    timezone_DEFAULT_ZONE = default_zone;
}

// -------------------------------------

TimeZone* U_EXPORT2
TimeZone::createDefault()
{
    umtx_initOnce(gDefaultZoneInitOnce, initDefault);
    {
        Mutex lock(&gDefaultZoneMutex);
        return (timezone_DEFAULT_ZONE != NULL) ? timezone_DEFAULT_ZONE->clone() : NULL;
    }
}

// -------------------------------------

void U_EXPORT2
TimeZone::adoptDefault(TimeZone* zone)
{
    if (zone != NULL)
    {
        {
            Mutex lock(&gDefaultZoneMutex);
            TimeZone *old = timezone_DEFAULT_ZONE;
            timezone_DEFAULT_ZONE = zone;
            delete old;
        }
        ucln_i18n_registerCleanup(UCLN_I18N_TIMEZONE, timeZone_cleanup);
    }
}
// -------------------------------------

void U_EXPORT2
TimeZone::setDefault(const TimeZone& zone)
{
    adoptDefault(zone.clone());
}

//----------------------------------------------------------------------


static void U_CALLCONV initMap(USystemTimeZoneType type, UErrorCode& ec) {
    ucln_i18n_registerCleanup(UCLN_I18N_TIMEZONE, timeZone_cleanup);

    UResourceBundle *res = ures_openDirect(0, timezone_kZONEINFO, &ec);
    res = ures_getByKey(res, timezone_kNAMES, res, &ec); // dereference Zones section
    if (U_SUCCESS(ec)) {
        int32_t size = ures_getSize(res);
        int32_t *m = (int32_t *)uprv_malloc(size * sizeof(int32_t));
        if (m == NULL) {
            ec = U_MEMORY_ALLOCATION_ERROR;
        } else {
            int32_t numEntries = 0;
            for (int32_t i = 0; i < size; i++) {
                UnicodeString id = ures_getUnicodeStringByIndex(res, i, &ec);
                if (U_FAILURE(ec)) {
                    break;
                }
                if (0 == id.compare(timezone_UNKNOWN_ZONE_ID, timezone_UNKNOWN_ZONE_ID_LENGTH)) {
                    // exclude Etc/Unknown
                    continue;
                }
                if (type == UCAL_ZONE_TYPE_CANONICAL || type == UCAL_ZONE_TYPE_CANONICAL_LOCATION) {
                    UnicodeString canonicalID;
                    ZoneMeta::getCanonicalCLDRID(id, canonicalID, ec);
                    if (U_FAILURE(ec)) {
                        break;
                    }
                    if (canonicalID != id) {
                        // exclude aliases
                        continue;
                    }
                }
                if (type == UCAL_ZONE_TYPE_CANONICAL_LOCATION) {
                    const UChar *region = TimeZone::getRegion(id, ec);
                    if (U_FAILURE(ec)) {
                        break;
                    }
                    if (u_strcmp(region, timezone_WORLD) == 0) {
                       // exclude non-location ("001")
                        continue;
                    }
                }
                m[numEntries++] = i;
            }
            if (U_SUCCESS(ec)) {
                int32_t *tmp = m;
                m = (int32_t *)uprv_realloc(tmp, numEntries * sizeof(int32_t));
                if (m == NULL) {
                    // realloc failed.. use the original one even it has unused
                    // area at the end
                    m = tmp;
                }

                switch(type) {
                case UCAL_ZONE_TYPE_ANY:
                    U_ASSERT(MAP_SYSTEM_ZONES == NULL);
                    MAP_SYSTEM_ZONES = m;
                    LEN_SYSTEM_ZONES = numEntries;
                    break;
                case UCAL_ZONE_TYPE_CANONICAL:
                    U_ASSERT(MAP_CANONICAL_SYSTEM_ZONES == NULL);
                    MAP_CANONICAL_SYSTEM_ZONES = m;
                    LEN_CANONICAL_SYSTEM_ZONES = numEntries;
                    break;
                case UCAL_ZONE_TYPE_CANONICAL_LOCATION:
                    U_ASSERT(MAP_CANONICAL_SYSTEM_LOCATION_ZONES == NULL);
                    MAP_CANONICAL_SYSTEM_LOCATION_ZONES = m;
                    LEN_CANONICAL_SYSTEM_LOCATION_ZONES = numEntries;
                    break;
                }
            }
        }
    }
    ures_close(res);
}


/**
 * This is the default implementation for subclasses that do not
 * override this method.  This implementation calls through to the
 * 8-argument getOffset() method after suitable computations, and
 * correctly adjusts GMT millis to local millis when necessary.
 */
void TimeZone::getOffset(UDate date, UBool local, int32_t& rawOffset,
                         int32_t& dstOffset, UErrorCode& ec) const {
    if (U_FAILURE(ec)) {
        return;
    }

    rawOffset = getRawOffset();
    if (!local) {
        date += rawOffset; // now in local standard millis
    }

    // When local == TRUE, date might not be in local standard
    // millis.  getOffset taking 7 parameters used here assume
    // the given time in day is local standard time.
    // At STD->DST transition, there is a range of time which
    // does not exist.  When 'date' is in this time range
    // (and local == TRUE), this method interprets the specified
    // local time as DST.  At DST->STD transition, there is a
    // range of time which occurs twice.  In this case, this
    // method interprets the specified local time as STD.
    // To support the behavior above, we need to call getOffset
    // (with 7 args) twice when local == true and DST is
    // detected in the initial call.
    for (int32_t pass=0; ; ++pass) {
        int32_t year, month, dom, dow;
        double day = uprv_floor(date / U_MILLIS_PER_DAY);
        int32_t millis = (int32_t) (date - day * U_MILLIS_PER_DAY);

        Grego::dayToFields(day, year, month, dom, dow);

        dstOffset = getOffset(GregorianCalendar::AD, year, month, dom,
                              (uint8_t) dow, millis,
                              Grego::monthLength(year, month),
                              ec) - rawOffset;

        // Recompute if local==TRUE, dstOffset!=0.
        if (pass!=0 || !local || dstOffset == 0) {
            break;
        }
        // adjust to local standard millis
        date -= dstOffset;
    }
}

// -------------------------------------

// New available IDs API as of ICU 2.4.  Uses StringEnumeration API.

class TZEnumeration : public StringEnumeration {
private:

    // Map into to zones.  Our results are zone[map[i]] for
    // i=0..len-1, where zone[i] is the i-th Olson zone.  If map==NULL
    // then our results are zone[i] for i=0..len-1.  Len will be zero
    // if the zone data could not be loaded.
    int32_t* map;
    int32_t* localMap;
    int32_t  len;
    int32_t  pos;

    TZEnumeration(int32_t* mapData, int32_t mapLen, UBool adoptMapData) : pos(0) {
        map = mapData;
        localMap = adoptMapData ? mapData : NULL;
        len = mapLen;
    }

    UBool getID(int32_t i, UErrorCode& ec) {
        int32_t idLen = 0;
        const UChar* id = NULL;
        UResourceBundle *top = ures_openDirect(0, timezone_kZONEINFO, &ec);
        top = ures_getByKey(top, timezone_kNAMES, top, &ec); // dereference Zones section
        id = ures_getStringByIndex(top, i, &idLen, &ec);
        if(U_FAILURE(ec)) {
            unistr.truncate(0);
        }
        else {
            unistr.fastCopyFrom(UnicodeString(TRUE, id, idLen));
        }
        ures_close(top);
        return U_SUCCESS(ec);
    }

    static int32_t* getMap(USystemTimeZoneType type, int32_t& len, UErrorCode& ec) {
        len = 0;
        if (U_FAILURE(ec)) {
            return NULL;
        }
        int32_t* m = NULL;
        switch (type) {
        case UCAL_ZONE_TYPE_ANY:
            umtx_initOnce(gSystemZonesInitOnce, &initMap, type, ec);
            m = MAP_SYSTEM_ZONES;
            len = LEN_SYSTEM_ZONES;
            break;
        case UCAL_ZONE_TYPE_CANONICAL:
            umtx_initOnce(gCanonicalZonesInitOnce, &initMap, type, ec);
            m = MAP_CANONICAL_SYSTEM_ZONES;
            len = LEN_CANONICAL_SYSTEM_ZONES;
            break;
        case UCAL_ZONE_TYPE_CANONICAL_LOCATION:
            umtx_initOnce(gCanonicalLocationZonesInitOnce, &initMap, type, ec);
            m = MAP_CANONICAL_SYSTEM_LOCATION_ZONES;
            len = LEN_CANONICAL_SYSTEM_LOCATION_ZONES;
            break;
        default:
            ec = U_ILLEGAL_ARGUMENT_ERROR;
            m = NULL;
            len = 0;
            break;
        }
        return m;
    }

public:

#define DEFAULT_FILTERED_MAP_SIZE 8
#define MAP_INCREMENT_SIZE 8

    static TZEnumeration* create(USystemTimeZoneType type, const char* region, const int32_t* rawOffset, UErrorCode& ec) {
        if (U_FAILURE(ec)) {
            return NULL;
        }

        int32_t baseLen;
        int32_t *baseMap = getMap(type, baseLen, ec);

        if (U_FAILURE(ec)) {
            return NULL;
        }

        // If any additional conditions are available,
        // create instance local map filtered by the conditions.

        int32_t *filteredMap = NULL;
        int32_t numEntries = 0;

        if (region != NULL || rawOffset != NULL) {
            int32_t filteredMapSize = DEFAULT_FILTERED_MAP_SIZE;
            filteredMap = (int32_t *)uprv_malloc(filteredMapSize * sizeof(int32_t));
            if (filteredMap == NULL) {
                ec = U_MEMORY_ALLOCATION_ERROR;
                return NULL;
            }

            // Walk through the base map
            UResourceBundle *res = ures_openDirect(0, timezone_kZONEINFO, &ec);
            res = ures_getByKey(res, timezone_kNAMES, res, &ec); // dereference Zones section
            for (int32_t i = 0; i < baseLen; i++) {
                int32_t zidx = baseMap[i];
                UnicodeString id = ures_getUnicodeStringByIndex(res, zidx, &ec);
                if (U_FAILURE(ec)) {
                    break;
                }
                if (region != NULL) {
                    // Filter by region
                    char tzregion[4]; // max 3 letters + null term
                    TimeZone::getRegion(id, tzregion, sizeof(tzregion), ec);
                    if (U_FAILURE(ec)) {
                        break;
                    }
                    if (uprv_stricmp(tzregion, region) != 0) {
                        // region does not match
                        continue;
                    }
                }
                if (rawOffset != NULL) {
                    // Filter by raw offset
                    // Note: This is VERY inefficient
                    TimeZone *z = createSystemTimeZone(id, ec);
                    if (U_FAILURE(ec)) {
                        break;
                    }
                    int32_t tzoffset = z->getRawOffset();
                    delete z;

                    if (tzoffset != *rawOffset) {
                        continue;
                    }
                }

                if (filteredMapSize <= numEntries) {
                    filteredMapSize += MAP_INCREMENT_SIZE;
                    int32_t *tmp = (int32_t *)uprv_realloc(filteredMap, filteredMapSize * sizeof(int32_t));
                    if (tmp == NULL) {
                        ec = U_MEMORY_ALLOCATION_ERROR;
                        break;
                    } else {
                        filteredMap = tmp;
                    }
                }

                filteredMap[numEntries++] = zidx;
            }

            if (U_FAILURE(ec)) {
                uprv_free(filteredMap);
                filteredMap = NULL;
            }

            ures_close(res);
        }

        TZEnumeration *result = NULL;
        if (U_SUCCESS(ec)) {
            // Finally, create a new enumeration instance
            if (filteredMap == NULL) {
                result = new TZEnumeration(baseMap, baseLen, FALSE);
            } else {
                result = new TZEnumeration(filteredMap, numEntries, TRUE);
                filteredMap = NULL;
            }
            if (result == NULL) {
                ec = U_MEMORY_ALLOCATION_ERROR;
            }
        }

        if (filteredMap != NULL) {
            uprv_free(filteredMap);
        }

        return result;
    }

    TZEnumeration(const TZEnumeration &other) : StringEnumeration(), map(NULL), localMap(NULL), len(0), pos(0) {
        if (other.localMap != NULL) {
            localMap = (int32_t *)uprv_malloc(other.len * sizeof(int32_t));
            if (localMap != NULL) {
                len = other.len;
                uprv_memcpy(localMap, other.localMap, len * sizeof(int32_t));
                pos = other.pos;
                map = localMap;
            } else {
                len = 0;
                pos = 0;
                map = NULL;
            }
        } else {
            map = other.map;
            localMap = NULL;
            len = other.len;
            pos = other.pos;
        }
    }

    virtual ~TZEnumeration();

    virtual StringEnumeration *clone() const {
        return new TZEnumeration(*this);
    }

    virtual int32_t count(UErrorCode& status) const {
        return U_FAILURE(status) ? 0 : len;
    }

    virtual const UnicodeString* snext(UErrorCode& status) {
        if (U_SUCCESS(status) && map != NULL && pos < len) {
            getID(map[pos], status);
            ++pos;
            return &unistr;
        }
        return 0;
    }

    virtual void reset(UErrorCode& /*status*/) {
        pos = 0;
    }

public:
    static UClassID U_EXPORT2 getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
};

TZEnumeration::~TZEnumeration() {
    if (localMap != NULL) {
        uprv_free(localMap);
    }
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(TZEnumeration)

StringEnumeration* U_EXPORT2
TimeZone::createTimeZoneIDEnumeration(
            USystemTimeZoneType zoneType,
            const char* region,
            const int32_t* rawOffset,
            UErrorCode& ec) {
    return TZEnumeration::create(zoneType, region, rawOffset, ec);
}

StringEnumeration* U_EXPORT2
TimeZone::createEnumeration() {
    UErrorCode ec = U_ZERO_ERROR;
    return TZEnumeration::create(UCAL_ZONE_TYPE_ANY, NULL, NULL, ec);
}

StringEnumeration* U_EXPORT2
TimeZone::createEnumeration(int32_t rawOffset) {
    UErrorCode ec = U_ZERO_ERROR;
    return TZEnumeration::create(UCAL_ZONE_TYPE_ANY, NULL, &rawOffset, ec);
}

StringEnumeration* U_EXPORT2
TimeZone::createEnumeration(const char* country) {
    UErrorCode ec = U_ZERO_ERROR;
    return TZEnumeration::create(UCAL_ZONE_TYPE_ANY, country, NULL, ec);
}

// ---------------------------------------

int32_t U_EXPORT2
TimeZone::countEquivalentIDs(const UnicodeString& id) {
    int32_t result = 0;
    UErrorCode ec = U_ZERO_ERROR;
    StackUResourceBundle res;
    U_DEBUG_TZ_MSG(("countEquivalentIDs..\n"));
    UResourceBundle *top = openOlsonResource(id, res.ref(), ec);
    if (U_SUCCESS(ec)) {
        StackUResourceBundle r;
        ures_getByKey(res.getAlias(), timezone_kLINKS, r.getAlias(), &ec);
        ures_getIntVector(r.getAlias(), &result, &ec);
    }
    ures_close(top);
    return result;
}

// ---------------------------------------

const UnicodeString U_EXPORT2
TimeZone::getEquivalentID(const UnicodeString& id, int32_t index) {
    U_DEBUG_TZ_MSG(("gEI(%d)\n", index));
    UnicodeString result;
    UErrorCode ec = U_ZERO_ERROR;
    StackUResourceBundle res;
    UResourceBundle *top = openOlsonResource(id, res.ref(), ec);
    int32_t zone = -1;
    if (U_SUCCESS(ec)) {
        StackUResourceBundle r;
        int32_t size;
        ures_getByKey(res.getAlias(), timezone_kLINKS, r.getAlias(), &ec);
        const int32_t *v = ures_getIntVector(r.getAlias(), &size, &ec);
        if (U_SUCCESS(ec)) {
            if (index >= 0 && index < size) {
                zone = v[index];
            }
        }
    }
    if (zone >= 0) {
        UResourceBundle *ares = ures_getByKey(top, timezone_kNAMES, NULL, &ec); // dereference Zones section
        if (U_SUCCESS(ec)) {
            int32_t idLen = 0;
            const UChar* id2 = ures_getStringByIndex(ares, zone, &idLen, &ec);
            result.fastCopyFrom(UnicodeString(TRUE, id2, idLen));
            U_DEBUG_TZ_MSG(("gei(%d) -> %d, len%d, %s\n", index, zone, result.length(), u_errorName(ec)));
        }
        ures_close(ares);
    }
    ures_close(top);
#if defined(U_DEBUG_TZ)
    if(result.length() ==0) {
      U_DEBUG_TZ_MSG(("equiv [__, #%d] -> 0 (%s)\n", index, u_errorName(ec)));
    }
#endif
    return result;
}

// ---------------------------------------

// These methods are used by ZoneMeta class only.

const UChar*
TimeZone::findID(const UnicodeString& id) {
    const UChar *result = NULL;
    UErrorCode ec = U_ZERO_ERROR;
    UResourceBundle *rb = ures_openDirect(NULL, timezone_kZONEINFO, &ec);

    // resolve zone index by name
    UResourceBundle *names = ures_getByKey(rb, timezone_kNAMES, NULL, &ec);
    int32_t idx = findInStringArray(names, id, ec);
    result = ures_getStringByIndex(names, idx, NULL, &ec);
    if (U_FAILURE(ec)) {
        result = NULL;
    }
    ures_close(names);
    ures_close(rb);
    return result;
}


const UChar*
TimeZone::dereferOlsonLink(const UnicodeString& id) {
    const UChar *result = NULL;
    UErrorCode ec = U_ZERO_ERROR;
    UResourceBundle *rb = ures_openDirect(NULL, timezone_kZONEINFO, &ec);

    // resolve zone index by name
    UResourceBundle *names = ures_getByKey(rb, timezone_kNAMES, NULL, &ec);
    int32_t idx = findInStringArray(names, id, ec);
    result = ures_getStringByIndex(names, idx, NULL, &ec);

    // open the zone bundle by index
    ures_getByKey(rb, timezone_kZONES, rb, &ec);
    ures_getByIndex(rb, idx, rb, &ec);

    if (U_SUCCESS(ec)) {
        if (ures_getType(rb) == URES_INT) {
            // this is a link - dereference the link
            int32_t deref = ures_getInt(rb, &ec);
            const UChar* tmp = ures_getStringByIndex(names, deref, NULL, &ec);
            if (U_SUCCESS(ec)) {
                result = tmp;
            }
        }
    }

    ures_close(names);
    ures_close(rb);

    return result;
}

const UChar*
TimeZone::getRegion(const UnicodeString& id) {
    UErrorCode status = U_ZERO_ERROR;
    return getRegion(id, status);
}

const UChar*
TimeZone::getRegion(const UnicodeString& id, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    const UChar *result = NULL;
    UResourceBundle *rb = ures_openDirect(NULL, timezone_kZONEINFO, &status);

    // resolve zone index by name
    UResourceBundle *res = ures_getByKey(rb, timezone_kNAMES, NULL, &status);
    int32_t idx = findInStringArray(res, id, status);

    // get region mapping
    ures_getByKey(rb, timezone_kREGIONS, res, &status);
    const UChar *tmp = ures_getStringByIndex(res, idx, NULL, &status);
    if (U_SUCCESS(status)) {
        result = tmp;
    }

    ures_close(res);
    ures_close(rb);

    return result;
}


// ---------------------------------------
int32_t
TimeZone::getRegion(const UnicodeString& id, char *region, int32_t capacity, UErrorCode& status)
{
    int32_t resultLen = 0;
    *region = 0;
    if (U_FAILURE(status)) {
        return 0;
    }

    const UChar *uregion = NULL;
    // "Etc/Unknown" is not a system zone ID,
    // but in the zone data
    if (id.compare(timezone_UNKNOWN_ZONE_ID, timezone_UNKNOWN_ZONE_ID_LENGTH) != 0) {
        uregion = getRegion(id);
    }
    if (uregion == NULL) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    resultLen = u_strlen(uregion);
    // A region code is represented by invariant characters
    u_UCharsToChars(uregion, region, uprv_min(resultLen, capacity));

    if (capacity < resultLen) {
        status = U_BUFFER_OVERFLOW_ERROR;
        return resultLen;
    }

    return u_terminateChars(region, capacity, resultLen, &status);
}

// ---------------------------------------


UnicodeString&
TimeZone::getDisplayName(UnicodeString& result) const
{
    return getDisplayName(FALSE,LONG,Locale::getDefault(), result);
}

UnicodeString&
TimeZone::getDisplayName(const Locale& locale, UnicodeString& result) const
{
    return getDisplayName(FALSE, LONG, locale, result);
}

UnicodeString&
TimeZone::getDisplayName(UBool inDaylight, EDisplayType style, UnicodeString& result)  const
{
    return getDisplayName(inDaylight,style, Locale::getDefault(), result);
}
//--------------------------------------
int32_t
TimeZone::getDSTSavings()const {
    if (useDaylightTime()) {
        return 3600000;
    }
    return 0;
}
//---------------------------------------
UnicodeString&
TimeZone::getDisplayName(UBool inDaylight, EDisplayType style, const Locale& locale, UnicodeString& result) const
{
    UErrorCode status = U_ZERO_ERROR;
    UDate date = Calendar::getNow();
    UTimeZoneFormatTimeType timeType = UTZFMT_TIME_TYPE_UNKNOWN;
    int32_t offset;

    if (style == GENERIC_LOCATION || style == LONG_GENERIC || style == SHORT_GENERIC) {
        LocalPointer<TimeZoneFormat> tzfmt(TimeZoneFormat::createInstance(locale, status));
        if (U_FAILURE(status)) {
            result.remove();
            return result;
        }
        // Generic format
        switch (style) {
        case GENERIC_LOCATION:
            tzfmt->format(UTZFMT_STYLE_GENERIC_LOCATION, *this, date, result, &timeType);
            break;
        case LONG_GENERIC:
            tzfmt->format(UTZFMT_STYLE_GENERIC_LONG, *this, date, result, &timeType);
            break;
        case SHORT_GENERIC:
            tzfmt->format(UTZFMT_STYLE_GENERIC_SHORT, *this, date, result, &timeType);
            break;
        default:
            UPRV_UNREACHABLE;
        }
        // Generic format many use Localized GMT as the final fallback.
        // When Localized GMT format is used, the result might not be
        // appropriate for the requested daylight value.
        if ((inDaylight && timeType == UTZFMT_TIME_TYPE_STANDARD) || (!inDaylight && timeType == UTZFMT_TIME_TYPE_DAYLIGHT)) {
            offset = inDaylight ? getRawOffset() + getDSTSavings() : getRawOffset();
            if (style == SHORT_GENERIC) {
                tzfmt->formatOffsetShortLocalizedGMT(offset, result, status);
            } else {
                tzfmt->formatOffsetLocalizedGMT(offset, result, status);
            }
        }
    } else if (style == LONG_GMT || style == SHORT_GMT) {
        LocalPointer<TimeZoneFormat> tzfmt(TimeZoneFormat::createInstance(locale, status));
        if (U_FAILURE(status)) {
            result.remove();
            return result;
        }
        offset = inDaylight && useDaylightTime() ? getRawOffset() + getDSTSavings() : getRawOffset();
        switch (style) {
        case LONG_GMT:
            tzfmt->formatOffsetLocalizedGMT(offset, result, status);
            break;
        case SHORT_GMT:
            tzfmt->formatOffsetISO8601Basic(offset, FALSE, FALSE, FALSE, result, status);
            break;
        default:
            UPRV_UNREACHABLE;
        }

    } else {
        U_ASSERT(style == LONG || style == SHORT || style == SHORT_COMMONLY_USED);
        UTimeZoneNameType nameType = UTZNM_UNKNOWN;
        switch (style) {
        case LONG:
            nameType = inDaylight ? UTZNM_LONG_DAYLIGHT : UTZNM_LONG_STANDARD;
            break;
        case SHORT:
        case SHORT_COMMONLY_USED:
            nameType = inDaylight ? UTZNM_SHORT_DAYLIGHT : UTZNM_SHORT_STANDARD;
            break;
        default:
            UPRV_UNREACHABLE;
        }
        LocalPointer<TimeZoneNames> tznames(TimeZoneNames::createInstance(locale, status));
        if (U_FAILURE(status)) {
            result.remove();
            return result;
        }
        UnicodeString canonicalID(ZoneMeta::getCanonicalCLDRID(*this));
        tznames->getDisplayName(canonicalID, nameType, date, result);
        if (result.isEmpty()) {
            // Fallback to localized GMT
            LocalPointer<TimeZoneFormat> tzfmt(TimeZoneFormat::createInstance(locale, status));
            offset = inDaylight && useDaylightTime() ? getRawOffset() + getDSTSavings() : getRawOffset();
            if (style == LONG) {
                tzfmt->formatOffsetLocalizedGMT(offset, result, status);
            } else {
                tzfmt->formatOffsetShortLocalizedGMT(offset, result, status);
            }
        }
    }
    if (U_FAILURE(status)) {
        result.remove();
    }
    return  result;
}

/**
 * Parse a custom time zone identifier and return a corresponding zone.
 * @param id a string of the form GMT[+-]hh:mm, GMT[+-]hhmm, or
 * GMT[+-]hh.
 * @return a newly created SimpleTimeZone with the given offset and
 * no Daylight Savings Time, or null if the id cannot be parsed.
*/
TimeZone*
TimeZone::createCustomTimeZone(const UnicodeString& id)
{
    int32_t sign, hour, min, sec;
    if (parseCustomID(id, sign, hour, min, sec)) {
        UnicodeString customID;
        formatCustomID(hour, min, sec, (sign < 0), customID);
        int32_t offset = sign * ((hour * 60 + min) * 60 + sec) * 1000;
        return new SimpleTimeZone(offset, customID);
    }
    return NULL;
}

UnicodeString&
TimeZone::getCustomID(const UnicodeString& id, UnicodeString& normalized, UErrorCode& status) {
    normalized.remove();
    if (U_FAILURE(status)) {
        return normalized;
    }
    int32_t sign, hour, min, sec;
    if (parseCustomID(id, sign, hour, min, sec)) {
        formatCustomID(hour, min, sec, (sign < 0), normalized);
    } else {
        status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return normalized;
}

UBool
TimeZone::parseCustomID(const UnicodeString& id, int32_t& sign,
                        int32_t& hour, int32_t& min, int32_t& sec) {
    static const int32_t         kParseFailed = -99999;

    NumberFormat* numberFormat = 0;
    UnicodeString idUppercase = id;
    idUppercase.toUpper("");

    if (id.length() > timezone_GMT_ID_LENGTH &&
        idUppercase.startsWith(timezone_GMT_ID, timezone_GMT_ID_LENGTH))
    {
        ParsePosition pos(timezone_GMT_ID_LENGTH);
        sign = 1;
        hour = 0;
        min = 0;
        sec = 0;

        if (id[pos.getIndex()] == timezone_MINUS /*'-'*/) {
            sign = -1;
        } else if (id[pos.getIndex()] != timezone_PLUS /*'+'*/) {
            return FALSE;
        }
        pos.setIndex(pos.getIndex() + 1);

        UErrorCode success = U_ZERO_ERROR;
        numberFormat = NumberFormat::createInstance(success);
        if(U_FAILURE(success)){
            return FALSE;
        }
        numberFormat->setParseIntegerOnly(TRUE);
        //numberFormat->setLenient(TRUE); // TODO: May need to set this, depends on latest timezone parsing

        // Look for either hh:mm, hhmm, or hh
        int32_t start = pos.getIndex();
        Formattable n(kParseFailed);
        numberFormat->parse(id, n, pos);
        if (pos.getIndex() == start) {
            delete numberFormat;
            return FALSE;
        }
        hour = n.getLong();

        if (pos.getIndex() < id.length()) {
            if (pos.getIndex() - start > 2
                || id[pos.getIndex()] != timezone_COLON) {
                delete numberFormat;
                return FALSE;
            }
            // hh:mm
            pos.setIndex(pos.getIndex() + 1);
            int32_t oldPos = pos.getIndex();
            n.setLong(kParseFailed);
            numberFormat->parse(id, n, pos);
            if ((pos.getIndex() - oldPos) != 2) {
                // must be 2 digits
                delete numberFormat;
                return FALSE;
            }
            min = n.getLong();
            if (pos.getIndex() < id.length()) {
                if (id[pos.getIndex()] != timezone_COLON) {
                    delete numberFormat;
                    return FALSE;
                }
                // [:ss]
                pos.setIndex(pos.getIndex() + 1);
                oldPos = pos.getIndex();
                n.setLong(kParseFailed);
                numberFormat->parse(id, n, pos);
                if (pos.getIndex() != id.length()
                        || (pos.getIndex() - oldPos) != 2) {
                    delete numberFormat;
                    return FALSE;
                }
                sec = n.getLong();
            }
        } else {
            // Supported formats are below -
            //
            // HHmmss
            // Hmmss
            // HHmm
            // Hmm
            // HH
            // H

            int32_t length = pos.getIndex() - start;
            if (length <= 0 || 6 < length) {
                // invalid length
                delete numberFormat;
                return FALSE;
            }
            switch (length) {
                case 1:
                case 2:
                    // already set to hour
                    break;
                case 3:
                case 4:
                    min = hour % 100;
                    hour /= 100;
                    break;
                case 5:
                case 6:
                    sec = hour % 100;
                    min = (hour/100) % 100;
                    hour /= 10000;
                    break;
            }
        }

        delete numberFormat;

        if (hour > timezone_kMAX_CUSTOM_HOUR || min > timezone_kMAX_CUSTOM_MIN || sec > timezone_kMAX_CUSTOM_SEC) {
            return FALSE;
        }
        return TRUE;
    }
    return FALSE;
}

UnicodeString&
TimeZone::formatCustomID(int32_t hour, int32_t min, int32_t sec,
                         UBool negative, UnicodeString& id) {
    // Create time zone ID - GMT[+|-]hhmm[ss]
    id.setTo(timezone_GMT_ID, timezone_GMT_ID_LENGTH);
    if (hour | min | sec) {
        if (negative) {
            id += (UChar)timezone_MINUS;
        } else {
            id += (UChar)timezone_PLUS;
        }

        if (hour < 10) {
            id += (UChar)timezone_ZERO_DIGIT;
        } else {
            id += (UChar)(timezone_ZERO_DIGIT + hour/10);
        }
        id += (UChar)(timezone_ZERO_DIGIT + hour%10);
        id += (UChar)timezone_COLON;
        if (min < 10) {
            id += (UChar)timezone_ZERO_DIGIT;
        } else {
            id += (UChar)(timezone_ZERO_DIGIT + min/10);
        }
        id += (UChar)(timezone_ZERO_DIGIT + min%10);

        if (sec) {
            id += (UChar)timezone_COLON;
            if (sec < 10) {
                id += (UChar)timezone_ZERO_DIGIT;
            } else {
                id += (UChar)(timezone_ZERO_DIGIT + sec/10);
            }
            id += (UChar)(timezone_ZERO_DIGIT + sec%10);
        }
    }
    return id;
}


UBool
TimeZone::hasSameRules(const TimeZone& other) const
{
    return (getRawOffset() == other.getRawOffset() &&
            useDaylightTime() == other.useDaylightTime());
}

static void U_CALLCONV initTZDataVersion(UErrorCode &status) {
    ucln_i18n_registerCleanup(UCLN_I18N_TIMEZONE, timeZone_cleanup);
    int32_t len = 0;
    StackUResourceBundle bundle;
    ures_openDirectFillIn(bundle.getAlias(), NULL, timezone_kZONEINFO, &status);
    const UChar *tzver = ures_getStringByKey(bundle.getAlias(), timezone_kTZVERSION, &len, &status);

    if (U_SUCCESS(status)) {
        if (len >= (int32_t)sizeof(TZDATA_VERSION)) {
            // Ensure that there is always space for a trailing nul in TZDATA_VERSION
            len = sizeof(TZDATA_VERSION) - 1;
        }
        u_UCharsToChars(tzver, TZDATA_VERSION, len);
    }
}

const char*
TimeZone::getTZDataVersion(UErrorCode& status)
{
    umtx_initOnce(gTZDataVersionInitOnce, &initTZDataVersion, status);
    return (const char*)TZDATA_VERSION;
}

UnicodeString&
TimeZone::getCanonicalID(const UnicodeString& id, UnicodeString& canonicalID, UErrorCode& status)
{
    UBool isSystemID = FALSE;
    return getCanonicalID(id, canonicalID, isSystemID, status);
}

UnicodeString&
TimeZone::getCanonicalID(const UnicodeString& id, UnicodeString& canonicalID, UBool& isSystemID,
                         UErrorCode& status)
{
    canonicalID.remove();
    isSystemID = FALSE;
    if (U_FAILURE(status)) {
        return canonicalID;
    }
    if (id.compare(timezone_UNKNOWN_ZONE_ID, timezone_UNKNOWN_ZONE_ID_LENGTH) == 0) {
        // special case - Etc/Unknown is a canonical ID, but not system ID
        canonicalID.fastCopyFrom(id);
        isSystemID = FALSE;
    } else {
        ZoneMeta::getCanonicalCLDRID(id, canonicalID, status);
        if (U_SUCCESS(status)) {
            isSystemID = TRUE;
        } else {
            // Not a system ID
            status = U_ZERO_ERROR;
            getCustomID(id, canonicalID, status);
        }
    }
    return canonicalID;
}

UnicodeString&
TimeZone::getWindowsID(const UnicodeString& id, UnicodeString& winid, UErrorCode& status) {
    winid.remove();
    if (U_FAILURE(status)) {
        return winid;
    }

    // canonicalize the input ID
    UnicodeString canonicalID;
    UBool isSystemID = FALSE;

    getCanonicalID(id, canonicalID, isSystemID, status);
    if (U_FAILURE(status) || !isSystemID) {
        // mapping data is only applicable to tz database IDs
        if (status == U_ILLEGAL_ARGUMENT_ERROR) {
            // getWindowsID() sets an empty string where
            // getCanonicalID() sets a U_ILLEGAL_ARGUMENT_ERROR.
            status = U_ZERO_ERROR;
        }
        return winid;
    }

    UResourceBundle *mapTimezones = ures_openDirect(NULL, "windowsZones", &status);
    ures_getByKey(mapTimezones, "mapTimezones", mapTimezones, &status);

    if (U_FAILURE(status)) {
        return winid;
    }

    UResourceBundle *winzone = NULL;
    UBool found = FALSE;
    while (ures_hasNext(mapTimezones) && !found) {
        winzone = ures_getNextResource(mapTimezones, winzone, &status);
        if (U_FAILURE(status)) {
            break;
        }
        if (ures_getType(winzone) != URES_TABLE) {
            continue;
        }
        UResourceBundle *regionalData = NULL;
        while (ures_hasNext(winzone) && !found) {
            regionalData = ures_getNextResource(winzone, regionalData, &status);
            if (U_FAILURE(status)) {
                break;
            }
            if (ures_getType(regionalData) != URES_STRING) {
                continue;
            }
            int32_t len;
            const UChar *tzids = ures_getString(regionalData, &len, &status);
            if (U_FAILURE(status)) {
                break;
            }

            const UChar *start = tzids;
            UBool hasNext = TRUE;
            while (hasNext) {
                const UChar *end = u_strchr(start, (UChar)0x20);
                if (end == NULL) {
                    end = tzids + len;
                    hasNext = FALSE;
                }
                if (canonicalID.compare(start, static_cast<int32_t>(end - start)) == 0) {
                    winid = UnicodeString(ures_getKey(winzone), -1 , US_INV);
                    found = TRUE;
                    break;
                }
                start = end + 1;
            }
        }
        ures_close(regionalData);
    }
    ures_close(winzone);
    ures_close(mapTimezones);

    return winid;
}

#define MAX_WINDOWS_ID_SIZE 128

UnicodeString&
TimeZone::getIDForWindowsID(const UnicodeString& winid, const char* region, UnicodeString& id, UErrorCode& status) {
    id.remove();
    if (U_FAILURE(status)) {
        return id;
    }

    UResourceBundle *zones = ures_openDirect(NULL, "windowsZones", &status);
    ures_getByKey(zones, "mapTimezones", zones, &status);
    if (U_FAILURE(status)) {
        ures_close(zones);
        return id;
    }

    UErrorCode tmperr = U_ZERO_ERROR;
    char winidKey[MAX_WINDOWS_ID_SIZE];
    int32_t winKeyLen = winid.extract(0, winid.length(), winidKey, sizeof(winidKey) - 1, US_INV);

    if (winKeyLen == 0 || winKeyLen >= (int32_t)sizeof(winidKey)) {
        ures_close(zones);
        return id;
    }
    winidKey[winKeyLen] = 0;

    ures_getByKey(zones, winidKey, zones, &tmperr); // use tmperr, because windows mapping might not
                                                    // be avaiable by design
    if (U_FAILURE(tmperr)) {
        ures_close(zones);
        return id;
    }

    const UChar *tzid = NULL;
    int32_t len = 0;
    UBool gotID = FALSE;
    if (region) {
        const UChar *tzids = ures_getStringByKey(zones, region, &len, &tmperr); // use tmperr, because
                                                                                // regional mapping is optional
        if (U_SUCCESS(tmperr)) {
            // first ID delimited by space is the defasult one
            const UChar *end = u_strchr(tzids, (UChar)0x20);
            if (end == NULL) {
                id.setTo(tzids, -1);
            } else {
                id.setTo(tzids, static_cast<int32_t>(end - tzids));
            }
            gotID = TRUE;
        }
    }

    if (!gotID) {
        tzid = ures_getStringByKey(zones, "001", &len, &status);    // using status, because "001" must be
                                                                // available at this point
        if (U_SUCCESS(status)) {
            id.setTo(tzid, len);
        }
    }

    ures_close(zones);
    return id;
}


U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2008-2014, Google, International Business Machines Corporation and
 * others. All Rights Reserved.
 *******************************************************************************
 */

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2009-2016, International Business Machines Corporation,       *
 * Google, and others. All Rights Reserved.                                    *
 *******************************************************************************
 */



/**
 * \file
 * \brief C++ API: time unit object
 */



#if U_SHOW_CPLUSPLUS_API



#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

/**
 * Measurement unit for time units.
 * @see TimeUnitAmount
 * @see TimeUnit
 * @stable ICU 4.2
 */
class U_I18N_API TimeUnit: public MeasureUnit {
public:
    /**
     * Constants for all the time units we supported.
     * @stable ICU 4.2
     */
    enum UTimeUnitFields {
        UTIMEUNIT_YEAR,
        UTIMEUNIT_MONTH,
        UTIMEUNIT_DAY,
        UTIMEUNIT_WEEK,
        UTIMEUNIT_HOUR,
        UTIMEUNIT_MINUTE,
        UTIMEUNIT_SECOND,
#ifndef U_HIDE_DEPRECATED_API
        /**
         * One more than the highest normal UTimeUnitFields value.
         * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
         */
        UTIMEUNIT_FIELD_COUNT
#endif  // U_HIDE_DEPRECATED_API
    };

    /**
     * Create Instance.
     * @param timeUnitField  time unit field based on which the instance
     *                       is created.
     * @param status         input-output error code.
     *                       If the timeUnitField is invalid,
     *                       then this will be set to U_ILLEGAL_ARGUMENT_ERROR.
     * @return               a TimeUnit instance
     * @stable ICU 4.2
     */
    static TimeUnit* U_EXPORT2 createInstance(UTimeUnitFields timeUnitField,
                                              UErrorCode& status);


    /**
     * Override clone.
     * @stable ICU 4.2
     */
    virtual TimeUnit* clone() const;

    /**
     * Copy operator.
     * @stable ICU 4.2
     */
    TimeUnit(const TimeUnit& other);

    /**
     * Assignment operator.
     * @stable ICU 4.2
     */
    TimeUnit& operator=(const TimeUnit& other);

    /**
     * Returns a unique class ID for this object POLYMORPHICALLY.
     * This method implements a simple form of RTTI used by ICU.
     * @return The class ID for this object. All objects of a given
     * class have the same class ID.  Objects of other classes have
     * different class IDs.
     * @stable ICU 4.2
     */
    virtual UClassID getDynamicClassID() const;

    /**
     * Returns the class ID for this class. This is used to compare to
     * the return value of getDynamicClassID().
     * @return The class ID for all objects of this class.
     * @stable ICU 4.2
     */
    static UClassID U_EXPORT2 getStaticClassID();


    /**
     * Get time unit field.
     * @return time unit field.
     * @stable ICU 4.2
     */
    UTimeUnitFields getTimeUnitField() const;

    /**
     * Destructor.
     * @stable ICU 4.2
     */
    virtual ~TimeUnit();

private:
    UTimeUnitFields fTimeUnitField;

    /**
     * Constructor
     * @internal (private)
     */
    TimeUnit(UTimeUnitFields timeUnitField);

};


U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */




#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(TimeUnit)


/*
 * There are only 7 time units.
 * So, TimeUnit could be made as singleton 
 * (similar to uniset_props.cpp, or unorm.cpp,
 * in which a static TimeUnit* array is created, and 
 * the creatInstance() returns a const TimeUnit*).
 * But the constraint is TimeUnit is a data member of Measure.
 * But Measure (which is an existing API) does not expect it's "unit" member
 * as singleton. Meaure takes ownership of the "unit" member.
 * In its constructor, it does not take a const "unit" pointer.
 * Also, Measure can clone and destruct the "unit" pointer.
 * In order to preserve the old behavior and let Measure handle singleton "unit",  
 * 1. a flag need to be added in Measure; 
 * 2. a new constructor which takes const "unit" as parameter need to be added,
 *    and this new constructor will set the flag on.
 * 3. clone and destructor need to check upon this flag to distinguish on how
 *    to handle the "unit". 
 * 
 * Since TimeUnit is such a light weight object, comparing with the heavy weight
 * format operation, we decided to avoid the above complication.
 * 
 * So, both TimeUnit and CurrencyUnit (the 2 subclasses of MeasureUnit) are
 * immutable and non-singleton.
 *
 * Currently, TimeUnitAmount and CurrencyAmount are immutable.
 * If an application needs to create a long list of TimeUnitAmount on the same
 * time unit but different number, for example,
 * 1 hour, 2 hour, 3 hour, ................. 10,000 hour,
 * there might be performance hit because 10,000 TimeUnit object, 
 * although all are the same time unit, will be created in heap and deleted.
 *
 * To address this performance issue, if there is any in the future,
 * we should and need to change TimeUnitAmount and CurrencyAmount to be 
 * immutable by allowing a setter on the number.
 * Or we need to add 2 parallel mutable classes in order to 
 * preserve the existing API.
 * Or we can use freezable.
 */
TimeUnit* U_EXPORT2 
TimeUnit::createInstance(TimeUnit::UTimeUnitFields timeUnitField, 
                         UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    if (timeUnitField < 0 || timeUnitField >= UTIMEUNIT_FIELD_COUNT) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    return new TimeUnit(timeUnitField);
}


TimeUnit::TimeUnit(TimeUnit::UTimeUnitFields timeUnitField) {
    fTimeUnitField = timeUnitField;
    switch (fTimeUnitField) {
    case UTIMEUNIT_YEAR:
        initTime("year");
        break;
    case UTIMEUNIT_MONTH:
        initTime("month");
        break;
    case UTIMEUNIT_DAY:
        initTime("day");
        break;
    case UTIMEUNIT_WEEK:
        initTime("week");
        break;
    case UTIMEUNIT_HOUR:
        initTime("hour");
        break;
    case UTIMEUNIT_MINUTE:
        initTime("minute");
        break;
    case UTIMEUNIT_SECOND:
        initTime("second");
        break;
    default:
        UPRV_UNREACHABLE;
    }
}

TimeUnit::TimeUnit(const TimeUnit& other) 
:   MeasureUnit(other), fTimeUnitField(other.fTimeUnitField) {
}

TimeUnit* 
TimeUnit::clone() const {
    return new TimeUnit(*this);
}

TimeUnit&
TimeUnit::operator=(const TimeUnit& other) {
    if (this == &other) {
        return *this;
    }
    MeasureUnit::operator=(other);
    fTimeUnitField = other.fTimeUnitField;
    return *this;
}

TimeUnit::UTimeUnitFields
TimeUnit::getTimeUnitField() const {
    return fTimeUnitField;
}

TimeUnit::~TimeUnit() {
}


U_NAMESPACE_END

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2008, Google, International Business Machines Corporation and *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 */ 

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2009-2010, Google, International Business Machines Corporation and *
 * others. All Rights Reserved.                                                *
 *******************************************************************************
 */ 



/**
 * \file
 * \brief C++ API: time unit amount object.
 */



#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN


/**
 * Express a duration as a time unit and number. Patterned after Currency.
 * @see TimeUnitAmount
 * @see TimeUnitFormat
 * @stable ICU 4.2
 */
class U_I18N_API TimeUnitAmount: public Measure {
public:
    /**
     * Construct TimeUnitAmount object with the given number and the
     * given time unit. 
     * @param number        a numeric object; number.isNumeric() must be TRUE
     * @param timeUnitField the time unit field of a time unit
     * @param status        the input-output error code. 
     *                      If the number is not numeric or the timeUnitField
     *                      is not valid,
     *                      then this will be set to a failing value:
     *                      U_ILLEGAL_ARGUMENT_ERROR.
     * @stable ICU 4.2
     */
    TimeUnitAmount(const Formattable& number, 
                   TimeUnit::UTimeUnitFields timeUnitField,
                   UErrorCode& status);

    /**
     * Construct TimeUnitAmount object with the given numeric amount and the
     * given time unit. 
     * @param amount        a numeric amount.
     * @param timeUnitField the time unit field on which a time unit amount
     *                      object will be created.
     * @param status        the input-output error code. 
     *                      If the timeUnitField is not valid,
     *                      then this will be set to a failing value:
     *                      U_ILLEGAL_ARGUMENT_ERROR.
     * @stable ICU 4.2
     */
    TimeUnitAmount(double amount, TimeUnit::UTimeUnitFields timeUnitField,
                   UErrorCode& status);


    /**
     * Copy constructor 
     * @stable ICU 4.2
     */
    TimeUnitAmount(const TimeUnitAmount& other);


    /**
     * Assignment operator
     * @stable ICU 4.2
     */
    TimeUnitAmount& operator=(const TimeUnitAmount& other);


    /**
     * Clone. 
     * @return a polymorphic clone of this object. The result will have the same               class as returned by getDynamicClassID().
     * @stable ICU 4.2
     */
    virtual TimeUnitAmount* clone() const;

    
    /**
     * Destructor
     * @stable ICU 4.2
     */
    virtual ~TimeUnitAmount();

    
    /** 
     * Equality operator.  
     * @param other  the object to compare to.
     * @return       true if this object is equal to the given object.
     * @stable ICU 4.2
     */
    virtual UBool operator==(const UObject& other) const;


    /** 
     * Not-equality operator.  
     * @param other  the object to compare to.
     * @return       true if this object is not equal to the given object.
     * @stable ICU 4.2
     */
    UBool operator!=(const UObject& other) const;


    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @stable ICU 4.2
     */
    static UClassID U_EXPORT2 getStaticClassID(void);


    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @stable ICU 4.2
     */
    virtual UClassID getDynamicClassID(void) const;


    /**
     * Get the time unit.
     * @return time unit object.
     * @stable ICU 4.2
     */
    const TimeUnit& getTimeUnit() const;

    /**
     * Get the time unit field value.
     * @return time unit field value.
     * @stable ICU 4.2
     */
    TimeUnit::UTimeUnitFields getTimeUnitField() const;
};



inline UBool 
TimeUnitAmount::operator!=(const UObject& other) const {
    return !operator==(other);
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



#if !UCONFIG_NO_FORMATTING

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(TimeUnitAmount)


TimeUnitAmount::TimeUnitAmount(const Formattable& number, 
                               TimeUnit::UTimeUnitFields timeUnitField,
                               UErrorCode& status)
:    Measure(number, TimeUnit::createInstance(timeUnitField, status), status) {
}


TimeUnitAmount::TimeUnitAmount(double amount, 
                               TimeUnit::UTimeUnitFields timeUnitField,
                               UErrorCode& status)
:   Measure(Formattable(amount), 
            TimeUnit::createInstance(timeUnitField, status),
            status) {
}


TimeUnitAmount::TimeUnitAmount(const TimeUnitAmount& other)
:   Measure(other)
{
}


TimeUnitAmount& 
TimeUnitAmount::operator=(const TimeUnitAmount& other) {
    Measure::operator=(other);
    return *this;
}


UBool
TimeUnitAmount::operator==(const UObject& other) const {
    return Measure::operator==(other);
}

TimeUnitAmount* 
TimeUnitAmount::clone() const {
    return new TimeUnitAmount(*this);
}

    
TimeUnitAmount::~TimeUnitAmount() {
}



const TimeUnit&
TimeUnitAmount::getTimeUnit() const {
    return (const TimeUnit&) getUnit();
}


TimeUnit::UTimeUnitFields
TimeUnitAmount::getTimeUnitField() const {
    return getTimeUnit().getTimeUnitField();
}
    

U_NAMESPACE_END

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2008-2015, Google, International Business Machines Corporation
 * and others. All Rights Reserved.
 *******************************************************************************
 */

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 *******************************************************************************
 * Copyright (C) 2008-2014, Google, International Business Machines Corporation
 * and others. All Rights Reserved.
 *******************************************************************************
 */




/**
 * \file
 * \brief C++ API: Format and parse duration in single time unit
 */


#if U_SHOW_CPLUSPLUS_API

#if !UCONFIG_NO_FORMATTING








#ifndef U_HIDE_DEPRECATED_API

/**
 * Constants for various styles.
 * There are 2 styles: full name and abbreviated name.
 * For example, for English, the full name for hour duration is "3 hours",
 * and the abbreviated name is "3 hrs".
 * @deprecated ICU 53 Use MeasureFormat and UMeasureFormatWidth instead.
 */
enum UTimeUnitFormatStyle {
    /** @deprecated ICU 53 */
    UTMUTFMT_FULL_STYLE,
    /** @deprecated ICU 53 */
    UTMUTFMT_ABBREVIATED_STYLE,
    /** @deprecated ICU 53 */
    UTMUTFMT_FORMAT_STYLE_COUNT
};
typedef enum UTimeUnitFormatStyle UTimeUnitFormatStyle; /**< @deprecated ICU 53 */


U_NAMESPACE_BEGIN

class Hashtable;
class UVector;

struct TimeUnitFormatReadSink;

/**
 * Format or parse a TimeUnitAmount, using plural rules for the units where available.
 *
 * <P>
 * Code Sample:
 * <pre>
 *   // create time unit amount instance - a combination of Number and time unit
 *   UErrorCode status = U_ZERO_ERROR;
 *   TimeUnitAmount* source = new TimeUnitAmount(2, TimeUnit::UTIMEUNIT_YEAR, status);
 *   // create time unit format instance
 *   TimeUnitFormat* format = new TimeUnitFormat(Locale("en"), status);
 *   // format a time unit amount
 *   UnicodeString formatted;
 *   Formattable formattable;
 *   if (U_SUCCESS(status)) {
 *       formattable.adoptObject(source);
 *       formatted = ((Format*)format)->format(formattable, formatted, status);
 *       Formattable result;
 *       ((Format*)format)->parseObject(formatted, result, status);
 *       if (U_SUCCESS(status)) {
 *           assert (result == formattable);
 *       }
 *   }
 * </pre>
 *
 * <P>
 * @see TimeUnitAmount
 * @see TimeUnitFormat
 * @deprecated ICU 53 Use the MeasureFormat class instead.
 */
class U_I18N_API TimeUnitFormat: public MeasureFormat {
public:

    /**
     * Create TimeUnitFormat with default locale, and full name style.
     * Use setLocale and/or setFormat to modify.
     * @deprecated ICU 53
     */
    TimeUnitFormat(UErrorCode& status);

    /**
     * Create TimeUnitFormat given locale, and full name style.
     * @deprecated ICU 53
     */
    TimeUnitFormat(const Locale& locale, UErrorCode& status);

    /**
     * Create TimeUnitFormat given locale and style.
     * @deprecated ICU 53
     */
    TimeUnitFormat(const Locale& locale, UTimeUnitFormatStyle style, UErrorCode& status);

    /**
     * Copy constructor.
     * @deprecated ICU 53
     */
    TimeUnitFormat(const TimeUnitFormat&);

    /**
     * deconstructor
     * @deprecated ICU 53
     */
    virtual ~TimeUnitFormat();

    /**
     * Clone this Format object polymorphically. The caller owns the result and
     * should delete it when done.
     * @return    A copy of the object.
     * @deprecated ICU 53
     */
    virtual TimeUnitFormat* clone() const;

    /**
     * Assignment operator
     * @deprecated ICU 53
     */
    TimeUnitFormat& operator=(const TimeUnitFormat& other);

    /**
     * Return true if the given Format objects are not semantically equal.
     * Objects of different subclasses are considered unequal.
     * @param other the object to be compared with.
     * @return      true if the given Format objects are not semantically equal.
     * @deprecated ICU 53
     */
    UBool operator!=(const Format& other) const;

    /**
     * Set the locale used for formatting or parsing.
     * @param locale  the locale to be set
     * @param status  output param set to success/failure code on exit
     * @deprecated ICU 53
     */
    void setLocale(const Locale& locale, UErrorCode& status);


    /**
     * Set the number format used for formatting or parsing.
     * @param format  the number formatter to be set
     * @param status  output param set to success/failure code on exit
     * @deprecated ICU 53
     */
    void setNumberFormat(const NumberFormat& format, UErrorCode& status);

    /**
     * Parse a TimeUnitAmount.
     * @see Format#parseObject(const UnicodeString&, Formattable&, ParsePosition&) const;
     * @deprecated ICU 53
     */
    virtual void parseObject(const UnicodeString& source,
                             Formattable& result,
                             ParsePosition& pos) const;

    /**
     * Return the class ID for this class. This is useful only for comparing to
     * a return value from getDynamicClassID(). For example:
     * <pre>
     * .   Base* polymorphic_pointer = createPolymorphicObject();
     * .   if (polymorphic_pointer->getDynamicClassID() ==
     * .       erived::getStaticClassID()) ...
     * </pre>
     * @return          The class ID for all objects of this class.
     * @deprecated ICU 53
     */
    static UClassID U_EXPORT2 getStaticClassID(void);

    /**
     * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
     * method is to implement a simple version of RTTI, since not all C++
     * compilers support genuine RTTI. Polymorphic operator==() and clone()
     * methods call this method.
     *
     * @return          The class ID for this object. All objects of a
     *                  given class have the same class ID.  Objects of
     *                  other classes have different class IDs.
     * @deprecated ICU 53
     */
    virtual UClassID getDynamicClassID(void) const;

private:
    Hashtable*    fTimeUnitToCountToPatterns[TimeUnit::UTIMEUNIT_FIELD_COUNT];
    UTimeUnitFormatStyle fStyle;

    void create(UTimeUnitFormatStyle style, UErrorCode& status);

    // it might actually be simpler to make them Decimal Formats later.
    // initialize all private data members
    void setup(UErrorCode& status);

    // initialize data member without fill in data for fTimeUnitToCountToPattern
    void initDataMembers(UErrorCode& status);

    // initialize fTimeUnitToCountToPatterns from current locale's resource.
    void readFromCurrentLocale(UTimeUnitFormatStyle style, const char* key, const UVector& pluralCounts,
                               UErrorCode& status);

    // check completeness of fTimeUnitToCountToPatterns against all time units,
    // and all plural rules, fill in fallback as necessary.
    void checkConsistency(UTimeUnitFormatStyle style, const char* key, UErrorCode& status);

    // fill in fTimeUnitToCountToPatterns from locale fall-back chain
    void searchInLocaleChain(UTimeUnitFormatStyle style, const char* key, const char* localeName,
                             TimeUnit::UTimeUnitFields field, const UnicodeString&,
                             const char*, Hashtable*, UErrorCode&);

    // initialize hash table
    Hashtable* initHash(UErrorCode& status);

    // delete hash table
    void deleteHash(Hashtable* htable);

    // copy hash table
    void copyHash(const Hashtable* source, Hashtable* target, UErrorCode& status);
    // get time unit name, such as "year", from time unit field enum, such as
    // UTIMEUNIT_YEAR.
    static const char* getTimeUnitName(TimeUnit::UTimeUnitFields field, UErrorCode& status);

    friend struct TimeUnitFormatReadSink;
};

inline UBool
TimeUnitFormat::operator!=(const Format& other) const  {
    return !operator==(other);
}

U_NAMESPACE_END

#endif /* U_HIDE_DEPRECATED_API */
#endif /* #if !UCONFIG_NO_FORMATTING */

#endif /* U_SHOW_CPLUSPLUS_API */



#if !UCONFIG_NO_FORMATTING














#define tmutfmt_LEFT_CURLY_BRACKET  ((UChar)0x007B)
#define tmutfmt_RIGHT_CURLY_BRACKET ((UChar)0x007D)
#define tmutfmt_SPACE             ((UChar)0x0020)
#define tmutfmt_DIGIT_ZERO        ((UChar)0x0030)
#define tmutfmt_LOW_S             ((UChar)0x0073)
#define tmutfmt_LOW_M             ((UChar)0x006D)
#define tmutfmt_LOW_I             ((UChar)0x0069)
#define tmutfmt_LOW_N             ((UChar)0x006E)
#define tmutfmt_LOW_H             ((UChar)0x0068)
#define tmutfmt_LOW_W             ((UChar)0x0077)
#define tmutfmt_LOW_D             ((UChar)0x0064)
#define tmutfmt_LOW_Y             ((UChar)0x0079)
#define tmutfmt_LOW_Z             ((UChar)0x007A)
#define tmutfmt_LOW_E             ((UChar)0x0065)
#define tmutfmt_LOW_R             ((UChar)0x0072)
#define tmutfmt_LOW_O             ((UChar)0x006F)
#define tmutfmt_LOW_N             ((UChar)0x006E)
#define tmutfmt_LOW_T             ((UChar)0x0074)


//TODO: define in compile time
//#define TMUTFMT_DEBUG 1

#ifdef TMUTFMT_DEBUG
#include <iostream>
#endif

U_NAMESPACE_BEGIN



UOBJECT_DEFINE_RTTI_IMPLEMENTATION(TimeUnitFormat)

static const char gUnitsTag[] = "units";
static const char gShortUnitsTag[] = "unitsShort";
static const char gTimeUnitYear[] = "year";
static const char gTimeUnitMonth[] = "month";
static const char gTimeUnitDay[] = "day";
static const char gTimeUnitWeek[] = "week";
static const char gTimeUnitHour[] = "hour";
static const char gTimeUnitMinute[] = "minute";
static const char gTimeUnitSecond[] = "second";
static const char gPluralCountOther[] = "other";

static const UChar DEFAULT_PATTERN_FOR_SECOND[] = {tmutfmt_LEFT_CURLY_BRACKET, tmutfmt_DIGIT_ZERO, tmutfmt_RIGHT_CURLY_BRACKET, tmutfmt_SPACE, tmutfmt_LOW_S, 0};
static const UChar DEFAULT_PATTERN_FOR_MINUTE[] = {tmutfmt_LEFT_CURLY_BRACKET, tmutfmt_DIGIT_ZERO, tmutfmt_RIGHT_CURLY_BRACKET, tmutfmt_SPACE, tmutfmt_LOW_M, tmutfmt_LOW_I, tmutfmt_LOW_N, 0};
static const UChar DEFAULT_PATTERN_FOR_HOUR[] = {tmutfmt_LEFT_CURLY_BRACKET, tmutfmt_DIGIT_ZERO, tmutfmt_RIGHT_CURLY_BRACKET, tmutfmt_SPACE, tmutfmt_LOW_H, 0};
static const UChar DEFAULT_PATTERN_FOR_WEEK[] = {tmutfmt_LEFT_CURLY_BRACKET, tmutfmt_DIGIT_ZERO, tmutfmt_RIGHT_CURLY_BRACKET, tmutfmt_SPACE, tmutfmt_LOW_W, 0};
static const UChar DEFAULT_PATTERN_FOR_DAY[] = {tmutfmt_LEFT_CURLY_BRACKET, tmutfmt_DIGIT_ZERO, tmutfmt_RIGHT_CURLY_BRACKET, tmutfmt_SPACE, tmutfmt_LOW_D, 0};
static const UChar DEFAULT_PATTERN_FOR_MONTH[] = {tmutfmt_LEFT_CURLY_BRACKET, tmutfmt_DIGIT_ZERO, tmutfmt_RIGHT_CURLY_BRACKET, tmutfmt_SPACE, tmutfmt_LOW_M, 0};
static const UChar DEFAULT_PATTERN_FOR_YEAR[] = {tmutfmt_LEFT_CURLY_BRACKET, tmutfmt_DIGIT_ZERO, tmutfmt_RIGHT_CURLY_BRACKET, tmutfmt_SPACE, tmutfmt_LOW_Y, 0};

static const UChar PLURAL_COUNT_ZERO[] = {tmutfmt_LOW_Z, tmutfmt_LOW_E, tmutfmt_LOW_R, tmutfmt_LOW_O, 0};
static const UChar PLURAL_COUNT_ONE[] = {tmutfmt_LOW_O, tmutfmt_LOW_N, tmutfmt_LOW_E, 0};
static const UChar PLURAL_COUNT_TWO[] = {tmutfmt_LOW_T, tmutfmt_LOW_W, tmutfmt_LOW_O, 0};

TimeUnitFormat::TimeUnitFormat(UErrorCode& status) {
    initMeasureFormat(Locale::getDefault(), UMEASFMT_WIDTH_WIDE, NULL, status);
    create(UTMUTFMT_FULL_STYLE, status);
}


TimeUnitFormat::TimeUnitFormat(const Locale& locale, UErrorCode& status) {
    initMeasureFormat(locale, UMEASFMT_WIDTH_WIDE, NULL, status);
    create(UTMUTFMT_FULL_STYLE, status);
}


TimeUnitFormat::TimeUnitFormat(const Locale& locale, UTimeUnitFormatStyle style, UErrorCode& status) {
    switch (style) {
    case UTMUTFMT_FULL_STYLE:
        initMeasureFormat(locale, UMEASFMT_WIDTH_WIDE, NULL, status);
        break;
    case UTMUTFMT_ABBREVIATED_STYLE:
        initMeasureFormat(locale, UMEASFMT_WIDTH_SHORT, NULL, status);
        break;
    default:
        initMeasureFormat(locale, UMEASFMT_WIDTH_WIDE, NULL, status);
        break;
    }
    create(style, status);
}

TimeUnitFormat::TimeUnitFormat(const TimeUnitFormat& other)
:   MeasureFormat(other),
    fStyle(other.fStyle)
{
    for (TimeUnit::UTimeUnitFields i = TimeUnit::UTIMEUNIT_YEAR;
         i < TimeUnit::UTIMEUNIT_FIELD_COUNT;
         i = (TimeUnit::UTimeUnitFields)(i+1)) {
        UErrorCode status = U_ZERO_ERROR;
        fTimeUnitToCountToPatterns[i] = initHash(status);
        if (U_SUCCESS(status)) {
            copyHash(other.fTimeUnitToCountToPatterns[i], fTimeUnitToCountToPatterns[i], status);
        } else {
            delete fTimeUnitToCountToPatterns[i];
            fTimeUnitToCountToPatterns[i] = NULL;
        }
    }
}


TimeUnitFormat::~TimeUnitFormat() {
    for (TimeUnit::UTimeUnitFields i = TimeUnit::UTIMEUNIT_YEAR;
         i < TimeUnit::UTIMEUNIT_FIELD_COUNT;
         i = (TimeUnit::UTimeUnitFields)(i+1)) {
        deleteHash(fTimeUnitToCountToPatterns[i]);
        fTimeUnitToCountToPatterns[i] = NULL;
    }
}


TimeUnitFormat*
TimeUnitFormat::clone() const {
    return new TimeUnitFormat(*this);
}


TimeUnitFormat&
TimeUnitFormat::operator=(const TimeUnitFormat& other) {
    if (this == &other) {
        return *this;
    }
    MeasureFormat::operator=(other);
    for (TimeUnit::UTimeUnitFields i = TimeUnit::UTIMEUNIT_YEAR;
         i < TimeUnit::UTIMEUNIT_FIELD_COUNT;
         i = (TimeUnit::UTimeUnitFields)(i+1)) {
        deleteHash(fTimeUnitToCountToPatterns[i]);
        fTimeUnitToCountToPatterns[i] = NULL;
    }
    for (TimeUnit::UTimeUnitFields i = TimeUnit::UTIMEUNIT_YEAR;
         i < TimeUnit::UTIMEUNIT_FIELD_COUNT;
         i = (TimeUnit::UTimeUnitFields)(i+1)) {
        UErrorCode status = U_ZERO_ERROR;
        fTimeUnitToCountToPatterns[i] = initHash(status);
        if (U_SUCCESS(status)) {
            copyHash(other.fTimeUnitToCountToPatterns[i], fTimeUnitToCountToPatterns[i], status);
        } else {
            delete fTimeUnitToCountToPatterns[i];
            fTimeUnitToCountToPatterns[i] = NULL;
        }
    }
    fStyle = other.fStyle;
    return *this;
}

void
TimeUnitFormat::parseObject(const UnicodeString& source,
                            Formattable& result,
                            ParsePosition& pos) const {
    Formattable resultNumber(0.0);
    UBool withNumberFormat = false;
    TimeUnit::UTimeUnitFields resultTimeUnit = TimeUnit::UTIMEUNIT_FIELD_COUNT;
    int32_t oldPos = pos.getIndex();
    int32_t newPos = -1;
    int32_t longestParseDistance = 0;
    UnicodeString* countOfLongestMatch = NULL;
#ifdef TMUTFMT_DEBUG
    char res[1000];
    source.extract(0, source.length(), res, "UTF-8");
    std::cout << "parse source: " << res << "\n";
#endif
    // parse by iterating through all available patterns
    // and looking for the longest match.
    for (TimeUnit::UTimeUnitFields i = TimeUnit::UTIMEUNIT_YEAR;
         i < TimeUnit::UTIMEUNIT_FIELD_COUNT;
         i = (TimeUnit::UTimeUnitFields)(i+1)) {
        Hashtable* countToPatterns = fTimeUnitToCountToPatterns[i];
        int32_t elemPos = UHASH_FIRST;
        const UHashElement* elem = NULL;
        while ((elem = countToPatterns->nextElement(elemPos)) != NULL){
            const UHashTok keyTok = elem->key;
            UnicodeString* count = (UnicodeString*)keyTok.pointer;
#ifdef TMUTFMT_DEBUG
            count->extract(0, count->length(), res, "UTF-8");
            std::cout << "parse plural count: " << res << "\n";
#endif
            const UHashTok valueTok = elem->value;
            // the value is a pair of MessageFormat*
            MessageFormat** patterns = (MessageFormat**)valueTok.pointer;
            for (UTimeUnitFormatStyle style = UTMUTFMT_FULL_STYLE; style < UTMUTFMT_FORMAT_STYLE_COUNT;
                 style = (UTimeUnitFormatStyle)(style + 1)) {
                MessageFormat* pattern = patterns[style];
                pos.setErrorIndex(-1);
                pos.setIndex(oldPos);
                // see if we can parse
                Formattable parsed;
                pattern->parseObject(source, parsed, pos);
                if (pos.getErrorIndex() != -1 || pos.getIndex() == oldPos) {
                    continue;
                }
    #ifdef TMUTFMT_DEBUG
                std::cout << "parsed.getType: " << parsed.getType() << "\n";
    #endif
                Formattable tmpNumber(0.0);
                if (pattern->getArgTypeCount() != 0) {
                    Formattable& temp = parsed[0];
                    if (temp.getType() == Formattable::kString) {
                        UnicodeString tmpString;
                        UErrorCode pStatus = U_ZERO_ERROR;
                        getNumberFormatInternal().parse(temp.getString(tmpString), tmpNumber, pStatus);
                        if (U_FAILURE(pStatus)) {
                            continue;
                        }
                    } else if (temp.isNumeric()) {
                        tmpNumber = temp;
                    } else {
                        continue;
                    }
                }
                int32_t parseDistance = pos.getIndex() - oldPos;
                if (parseDistance > longestParseDistance) {
                    if (pattern->getArgTypeCount() != 0) {
                        resultNumber = tmpNumber;
                        withNumberFormat = true;
                    } else {
                        withNumberFormat = false;
                    }
                    resultTimeUnit = i;
                    newPos = pos.getIndex();
                    longestParseDistance = parseDistance;
                    countOfLongestMatch = count;
                }
            }
        }
    }
    /* After find the longest match, parse the number.
     * Result number could be null for the pattern without number pattern.
     * such as unit pattern in Arabic.
     * When result number is null, use plural rule to set the number.
     */
    if (withNumberFormat == false && longestParseDistance != 0) {
        // set the number using plurrual count
        if (0 == countOfLongestMatch->compare(PLURAL_COUNT_ZERO, 4)) {
            resultNumber = Formattable(0.0);
        } else if (0 == countOfLongestMatch->compare(PLURAL_COUNT_ONE, 3)) {
            resultNumber = Formattable(1.0);
        } else if (0 == countOfLongestMatch->compare(PLURAL_COUNT_TWO, 3)) {
            resultNumber = Formattable(2.0);
        } else {
            // should not happen.
            // TODO: how to handle?
            resultNumber = Formattable(3.0);
        }
    }
    if (longestParseDistance == 0) {
        pos.setIndex(oldPos);
        pos.setErrorIndex(0);
    } else {
        UErrorCode status = U_ZERO_ERROR;
        LocalPointer<TimeUnitAmount> tmutamt(new TimeUnitAmount(resultNumber, resultTimeUnit, status), status);
        if (U_SUCCESS(status)) {
            result.adoptObject(tmutamt.orphan());
            pos.setIndex(newPos);
            pos.setErrorIndex(-1);
        } else {
            pos.setIndex(oldPos);
            pos.setErrorIndex(0);
        }
    }
}

void
TimeUnitFormat::create(UTimeUnitFormatStyle style, UErrorCode& status) {
    // fTimeUnitToCountToPatterns[] must have its elements initialized to NULL first
    // before checking for failure status.
    for (TimeUnit::UTimeUnitFields i = TimeUnit::UTIMEUNIT_YEAR;
         i < TimeUnit::UTIMEUNIT_FIELD_COUNT;
         i = (TimeUnit::UTimeUnitFields)(i+1)) {
        fTimeUnitToCountToPatterns[i] = NULL;
    }

    if (U_FAILURE(status)) {
        return;
    }
    if (style < UTMUTFMT_FULL_STYLE || style >= UTMUTFMT_FORMAT_STYLE_COUNT) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    fStyle = style;

    //TODO: format() and parseObj() are const member functions,
    //so, can not do lazy initialization in C++.
    //setup has to be done in constructors.
    //and here, the behavior is not consistent with Java.
    //In Java, create an empty instance does not setup locale as
    //default locale. If it followed by setNumberFormat(),
    //in format(), the locale will set up as the locale in fNumberFormat.
    //But in C++, this sets the locale as the default locale.
    setup(status);
}

void
TimeUnitFormat::setup(UErrorCode& err) {
    initDataMembers(err);

    UVector pluralCounts(0, uhash_compareUnicodeString, 6, err);
    LocalPointer<StringEnumeration> keywords(getPluralRules().getKeywords(err), err);
    if (U_FAILURE(err)) {
        return;
    }
    UnicodeString* pluralCount;
    while ((pluralCount = const_cast<UnicodeString*>(keywords->snext(err))) != NULL) {
      pluralCounts.addElement(pluralCount, err);
    }
    readFromCurrentLocale(UTMUTFMT_FULL_STYLE, gUnitsTag, pluralCounts, err);
    checkConsistency(UTMUTFMT_FULL_STYLE, gUnitsTag, err);
    readFromCurrentLocale(UTMUTFMT_ABBREVIATED_STYLE, gShortUnitsTag, pluralCounts, err);
    checkConsistency(UTMUTFMT_ABBREVIATED_STYLE, gShortUnitsTag, err);
}


void
TimeUnitFormat::initDataMembers(UErrorCode& err){
    if (U_FAILURE(err)) {
        return;
    }
    for (TimeUnit::UTimeUnitFields i = TimeUnit::UTIMEUNIT_YEAR;
         i < TimeUnit::UTIMEUNIT_FIELD_COUNT;
         i = (TimeUnit::UTimeUnitFields)(i+1)) {
        deleteHash(fTimeUnitToCountToPatterns[i]);
        fTimeUnitToCountToPatterns[i] = NULL;
    }
}

struct TimeUnitFormatReadSink : public ResourceSink {
    TimeUnitFormat *timeUnitFormatObj;
    const UVector &pluralCounts;
    UTimeUnitFormatStyle style;
    UBool beenHere;

    TimeUnitFormatReadSink(TimeUnitFormat *timeUnitFormatObj,
            const UVector &pluralCounts, UTimeUnitFormatStyle style) :
            timeUnitFormatObj(timeUnitFormatObj), pluralCounts(pluralCounts),
            style(style), beenHere(FALSE){}

    virtual ~TimeUnitFormatReadSink();

    virtual void put(const char *key, ResourceValue &value, UBool, UErrorCode &errorCode) {
        // Skip all put() calls except the first one -- discard all fallback data.
        if (beenHere) {
            return;
        } else {
            beenHere = TRUE;
        }

        ResourceTable units = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }

        for (int32_t i = 0; units.getKeyAndValue(i, key, value); ++i) {
            const char* timeUnitName = key;
            if (timeUnitName == NULL) {
                continue;
            }

            TimeUnit::UTimeUnitFields timeUnitField = TimeUnit::UTIMEUNIT_FIELD_COUNT;
            if ( uprv_strcmp(timeUnitName, gTimeUnitYear) == 0 ) {
                timeUnitField = TimeUnit::UTIMEUNIT_YEAR;
            } else if ( uprv_strcmp(timeUnitName, gTimeUnitMonth) == 0 ) {
                timeUnitField = TimeUnit::UTIMEUNIT_MONTH;
            } else if ( uprv_strcmp(timeUnitName, gTimeUnitDay) == 0 ) {
                timeUnitField = TimeUnit::UTIMEUNIT_DAY;
            } else if ( uprv_strcmp(timeUnitName, gTimeUnitHour) == 0 ) {
                timeUnitField = TimeUnit::UTIMEUNIT_HOUR;
            } else if ( uprv_strcmp(timeUnitName, gTimeUnitMinute) == 0 ) {
                timeUnitField = TimeUnit::UTIMEUNIT_MINUTE;
            } else if ( uprv_strcmp(timeUnitName, gTimeUnitSecond) == 0 ) {
                timeUnitField = TimeUnit::UTIMEUNIT_SECOND;
            } else if ( uprv_strcmp(timeUnitName, gTimeUnitWeek) == 0 ) {
                timeUnitField = TimeUnit::UTIMEUNIT_WEEK;
            } else {
                continue;
            }
            LocalPointer<Hashtable> localCountToPatterns;
            Hashtable *countToPatterns =
                timeUnitFormatObj->fTimeUnitToCountToPatterns[timeUnitField];
            if (countToPatterns == NULL) {
                localCountToPatterns.adoptInsteadAndCheckErrorCode(
                    timeUnitFormatObj->initHash(errorCode), errorCode);
                countToPatterns = localCountToPatterns.getAlias();
                if (U_FAILURE(errorCode)) {
                    return;
                }
            }

            ResourceTable countsToPatternTable = value.getTable(errorCode);
            if (U_FAILURE(errorCode)) {
                continue;
            }
            for (int32_t j = 0; countsToPatternTable.getKeyAndValue(j, key, value); ++j) {
                errorCode = U_ZERO_ERROR;
                UnicodeString pattern = value.getUnicodeString(errorCode);
                if (U_FAILURE(errorCode)) {
                    continue;
                }
                UnicodeString pluralCountUniStr(key, -1, US_INV);
                if (!pluralCounts.contains(&pluralCountUniStr)) {
                    continue;
                }
                LocalPointer<MessageFormat> messageFormat(new MessageFormat(
                    pattern, timeUnitFormatObj->getLocale(errorCode), errorCode), errorCode);
                if (U_FAILURE(errorCode)) {
                    return;
                }
                MessageFormat** formatters =
                    (MessageFormat**)countToPatterns->get(pluralCountUniStr);
                if (formatters == NULL) {
                    LocalMemory<MessageFormat *> localFormatters(
                        (MessageFormat **)uprv_malloc(UTMUTFMT_FORMAT_STYLE_COUNT*sizeof(MessageFormat*)));
                    if (localFormatters.isNull()) {
                        errorCode = U_MEMORY_ALLOCATION_ERROR;
                        return;
                    }
                    localFormatters[UTMUTFMT_FULL_STYLE] = NULL;
                    localFormatters[UTMUTFMT_ABBREVIATED_STYLE] = NULL;
                    countToPatterns->put(pluralCountUniStr, localFormatters.getAlias(), errorCode);
                    if (U_FAILURE(errorCode)) {
                        return;
                    }
                    formatters = localFormatters.orphan();
                }
                formatters[style] = messageFormat.orphan();
            }

            if (timeUnitFormatObj->fTimeUnitToCountToPatterns[timeUnitField] == NULL) {
                timeUnitFormatObj->fTimeUnitToCountToPatterns[timeUnitField] = localCountToPatterns.orphan();
            }
        }
    }

};

TimeUnitFormatReadSink::~TimeUnitFormatReadSink() {}

void
TimeUnitFormat::readFromCurrentLocale(UTimeUnitFormatStyle style, const char* key,
                                      const UVector& pluralCounts, UErrorCode& err) {
    if (U_FAILURE(err)) {
        return;
    }
    // fill timeUnitToCountToPatterns from resource file
    // err is used to indicate wrong status except missing resource.
    // status is an error code used in resource lookup.
    // status does not affect "err".
    UErrorCode status = U_ZERO_ERROR;
    LocalUResourceBundlePointer rb(ures_open(U_ICUDATA_UNIT, getLocaleID(status), &status));

    LocalUResourceBundlePointer unitsRes(ures_getByKey(rb.getAlias(), key, NULL, &status));
    ures_getByKey(unitsRes.getAlias(), "duration", unitsRes.getAlias(), &status);
    if (U_FAILURE(status)) {
        return;
    }

    TimeUnitFormatReadSink sink(this, pluralCounts, style);
    ures_getAllItemsWithFallback(unitsRes.getAlias(), "", sink, status);
}

void
TimeUnitFormat::checkConsistency(UTimeUnitFormatStyle style, const char* key, UErrorCode& err) {
    if (U_FAILURE(err)) {
        return;
    }
    // there should be patterns for each plural rule in each time unit.
    // For each time unit,
    //     for each plural rule, following is unit pattern fall-back rule:
    //         ( for example: "one" hour )
    //         look for its unit pattern in its locale tree.
    //         if pattern is not found in its own locale, such as de_DE,
    //         look for the pattern in its parent, such as de,
    //         keep looking till found or till root.
    //         if the pattern is not found in root either,
    //         fallback to plural count "other",
    //         look for the pattern of "other" in the locale tree:
    //         "de_DE" to "de" to "root".
    //         If not found, fall back to value of
    //         static variable DEFAULT_PATTERN_FOR_xxx, such as "{0} h".
    //
    // Following is consistency check to create pattern for each
    // plural rule in each time unit using above fall-back rule.
    //
    LocalPointer<StringEnumeration> keywords(
            getPluralRules().getKeywords(err), err);
    const UnicodeString* pluralCount;
    while (U_SUCCESS(err) && (pluralCount = keywords->snext(err)) != NULL) {
        for (int32_t i = 0; i < TimeUnit::UTIMEUNIT_FIELD_COUNT; ++i) {
            // for each time unit,
            // get all the patterns for each plural rule in this locale.
            Hashtable* countToPatterns = fTimeUnitToCountToPatterns[i];
            if ( countToPatterns == NULL ) {
                fTimeUnitToCountToPatterns[i] = countToPatterns = initHash(err);
                if (U_FAILURE(err)) {
                    return;
                }
            }
            MessageFormat** formatters = (MessageFormat**)countToPatterns->get(*pluralCount);
            if( formatters == NULL || formatters[style] == NULL ) {
                // look through parents
                const char* localeName = getLocaleID(err);
                CharString pluralCountChars;
                pluralCountChars.appendInvariantChars(*pluralCount, err);
                searchInLocaleChain(style, key, localeName,
                                    (TimeUnit::UTimeUnitFields)i,
                                    *pluralCount, pluralCountChars.data(),
                                    countToPatterns, err);
            }
            // TODO: what to do with U_FAILURE(err) at this point.
            //       As is, the outer loop continues to run, but does nothing.
        }
    }
}



// srcPluralCount is the original plural count on which the pattern is
// searched for.
// searchPluralCount is the fallback plural count.
// For example, to search for pattern for ""one" hour",
// "one" is the srcPluralCount,
// if the pattern is not found even in root, fallback to
// using patterns of plural count "other",
// then, "other" is the searchPluralCount.
void
TimeUnitFormat::searchInLocaleChain(UTimeUnitFormatStyle style, const char* key, const char* localeName,
                                TimeUnit::UTimeUnitFields srcTimeUnitField,
                                const UnicodeString& srcPluralCount,
                                const char* searchPluralCount,
                                Hashtable* countToPatterns,
                                UErrorCode& err) {
    if (U_FAILURE(err)) {
        return;
    }
    UErrorCode status = U_ZERO_ERROR;
    char parentLocale[ULOC_FULLNAME_CAPACITY];
    uprv_strcpy(parentLocale, localeName);
    int32_t locNameLen;
    U_ASSERT(countToPatterns != NULL);
    while ((locNameLen = uloc_getParent(parentLocale, parentLocale,
                                        ULOC_FULLNAME_CAPACITY, &status)) >= 0){
        // look for pattern for srcPluralCount in locale tree
        LocalUResourceBundlePointer rb(ures_open(U_ICUDATA_UNIT, parentLocale, &status));
        LocalUResourceBundlePointer unitsRes(ures_getByKey(rb.getAlias(), key, NULL, &status));
        const char* timeUnitName = getTimeUnitName(srcTimeUnitField, status);
        LocalUResourceBundlePointer countsToPatternRB(ures_getByKey(unitsRes.getAlias(), timeUnitName, NULL, &status));
        const UChar* pattern;
        int32_t      ptLength;
        pattern = ures_getStringByKeyWithFallback(countsToPatternRB.getAlias(), searchPluralCount, &ptLength, &status);
        if (U_SUCCESS(status)) {
            //found
            LocalPointer<MessageFormat> messageFormat(
                new MessageFormat(UnicodeString(TRUE, pattern, ptLength), getLocale(err), err), err);
            if (U_FAILURE(err)) {
                return;
            }
            MessageFormat** formatters = (MessageFormat**)countToPatterns->get(srcPluralCount);
            if (formatters == NULL) {
                LocalMemory<MessageFormat *> localFormatters(
                        (MessageFormat**)uprv_malloc(UTMUTFMT_FORMAT_STYLE_COUNT*sizeof(MessageFormat*)));
                formatters = localFormatters.getAlias();
                localFormatters[UTMUTFMT_FULL_STYLE] = NULL;
                localFormatters[UTMUTFMT_ABBREVIATED_STYLE] = NULL;
                countToPatterns->put(srcPluralCount, localFormatters.orphan(), err);
                if (U_FAILURE(err)) {
                    return;
                }
            }
            //delete formatters[style];
            formatters[style] = messageFormat.orphan();
            return;
        }
        status = U_ZERO_ERROR;
        if (locNameLen == 0) {
            break;
        }
    }

    // if no unitsShort resource was found even after fallback to root locale
    // then search the units resource fallback from the current level to root
    if ( locNameLen == 0 && uprv_strcmp(key, gShortUnitsTag) == 0) {
#ifdef TMUTFMT_DEBUG
        std::cout << "loop into searchInLocaleChain since Short-Long-Alternative \n";
#endif
        CharString pLocale(localeName, -1, err);
        // Add an underscore at the tail of locale name,
        // so that searchInLocaleChain will check the current locale before falling back
        pLocale.append('_', err);
        searchInLocaleChain(style, gUnitsTag, pLocale.data(), srcTimeUnitField, srcPluralCount,
                             searchPluralCount, countToPatterns, err);
        if (U_FAILURE(err)) {
            return;
        }
        MessageFormat** formatters = (MessageFormat**)countToPatterns->get(srcPluralCount);
        if (formatters != NULL && formatters[style] != NULL) {
            return;
        }
    }

    // if not found the pattern for this plural count at all,
    // fall-back to plural count "other"
    if ( uprv_strcmp(searchPluralCount, gPluralCountOther) == 0 ) {
        // set default fall back the same as the resource in root
        LocalPointer<MessageFormat> messageFormat;
        const UChar *pattern = NULL;
        if ( srcTimeUnitField == TimeUnit::UTIMEUNIT_SECOND ) {
            pattern = DEFAULT_PATTERN_FOR_SECOND;
        } else if ( srcTimeUnitField == TimeUnit::UTIMEUNIT_MINUTE ) {
            pattern = DEFAULT_PATTERN_FOR_MINUTE;
        } else if ( srcTimeUnitField == TimeUnit::UTIMEUNIT_HOUR ) {
            pattern = DEFAULT_PATTERN_FOR_HOUR;
        } else if ( srcTimeUnitField == TimeUnit::UTIMEUNIT_WEEK ) {
            pattern = DEFAULT_PATTERN_FOR_WEEK;
        } else if ( srcTimeUnitField == TimeUnit::UTIMEUNIT_DAY ) {
            pattern = DEFAULT_PATTERN_FOR_DAY;
        } else if ( srcTimeUnitField == TimeUnit::UTIMEUNIT_MONTH ) {
            pattern = DEFAULT_PATTERN_FOR_MONTH;
        } else if ( srcTimeUnitField == TimeUnit::UTIMEUNIT_YEAR ) {
            pattern = DEFAULT_PATTERN_FOR_YEAR;
        }
        if (pattern != NULL) {
            messageFormat.adoptInsteadAndCheckErrorCode(
                     new MessageFormat(UnicodeString(TRUE, pattern, -1), getLocale(err), err), err);
        }
        if (U_FAILURE(err)) {
            return;
        }
        MessageFormat** formatters = (MessageFormat**)countToPatterns->get(srcPluralCount);
        if (formatters == NULL) {
            LocalMemory<MessageFormat *> localFormatters (
                    (MessageFormat**)uprv_malloc(UTMUTFMT_FORMAT_STYLE_COUNT*sizeof(MessageFormat*)));
            if (localFormatters.isNull()) {
                err = U_MEMORY_ALLOCATION_ERROR;
                return;
            }
            formatters = localFormatters.getAlias();
            formatters[UTMUTFMT_FULL_STYLE] = NULL;
            formatters[UTMUTFMT_ABBREVIATED_STYLE] = NULL;
            countToPatterns->put(srcPluralCount, localFormatters.orphan(), err);
        }
        if (U_SUCCESS(err)) {
            //delete formatters[style];
            formatters[style] = messageFormat.orphan();
        }
    } else {
        // fall back to rule "other", and search in parents
        searchInLocaleChain(style, key, localeName, srcTimeUnitField, srcPluralCount,
                            gPluralCountOther, countToPatterns, err);
    }
}

void
TimeUnitFormat::setLocale(const Locale& locale, UErrorCode& status) {
    if (setMeasureFormatLocale(locale, status)) {
        setup(status);
    }
}


void
TimeUnitFormat::setNumberFormat(const NumberFormat& format, UErrorCode& status){
    if (U_FAILURE(status)) {
        return;
    }
    adoptNumberFormat(format.clone(), status);
}


void
TimeUnitFormat::deleteHash(Hashtable* htable) {
    int32_t pos = UHASH_FIRST;
    const UHashElement* element = NULL;
    if ( htable ) {
        while ( (element = htable->nextElement(pos)) != NULL ) {
            const UHashTok valueTok = element->value;
            const MessageFormat** value = (const MessageFormat**)valueTok.pointer;
            delete value[UTMUTFMT_FULL_STYLE];
            delete value[UTMUTFMT_ABBREVIATED_STYLE];
            //delete[] value;
            uprv_free(value);
        }
    }
    delete htable;
}


void
TimeUnitFormat::copyHash(const Hashtable* source, Hashtable* target, UErrorCode& status) {
    if ( U_FAILURE(status) ) {
        return;
    }
    int32_t pos = UHASH_FIRST;
    const UHashElement* element = NULL;
    if ( source ) {
        while ( (element = source->nextElement(pos)) != NULL ) {
            const UHashTok keyTok = element->key;
            const UnicodeString* key = (UnicodeString*)keyTok.pointer;
            const UHashTok valueTok = element->value;
            const MessageFormat** value = (const MessageFormat**)valueTok.pointer;
            MessageFormat** newVal = (MessageFormat**)uprv_malloc(UTMUTFMT_FORMAT_STYLE_COUNT*sizeof(MessageFormat*));
            newVal[0] = value[0]->clone();
            newVal[1] = value[1]->clone();
            target->put(UnicodeString(*key), newVal, status);
            if ( U_FAILURE(status) ) {
                delete newVal[0];
                delete newVal[1];
                uprv_free(newVal);
                return;
            }
        }
    }
}


U_CDECL_BEGIN

/**
 * set hash table value comparator
 *
 * @param val1  one value in comparison
 * @param val2  the other value in comparison
 * @return      TRUE if 2 values are the same, FALSE otherwise
 */
static UBool U_CALLCONV tmutfmtHashTableValueComparator(UHashTok val1, UHashTok val2);

static UBool
U_CALLCONV tmutfmtHashTableValueComparator(UHashTok val1, UHashTok val2) {
    const MessageFormat** pattern1 = (const MessageFormat**)val1.pointer;
    const MessageFormat** pattern2 = (const MessageFormat**)val2.pointer;
    return *pattern1[0] == *pattern2[0] && *pattern1[1] == *pattern2[1];
}

U_CDECL_END

Hashtable*
TimeUnitFormat::initHash(UErrorCode& status) {
    if ( U_FAILURE(status) ) {
        return NULL;
    }
    Hashtable* hTable;
    if ( (hTable = new Hashtable(TRUE, status)) == NULL ) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    if ( U_FAILURE(status) ) {
        delete hTable;
        return NULL;
    }
    hTable->setValueComparator(tmutfmtHashTableValueComparator);
    return hTable;
}


const char*
TimeUnitFormat::getTimeUnitName(TimeUnit::UTimeUnitFields unitField,
                                UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    switch (unitField) {
      case TimeUnit::UTIMEUNIT_YEAR:
        return gTimeUnitYear;
      case TimeUnit::UTIMEUNIT_MONTH:
        return gTimeUnitMonth;
      case TimeUnit::UTIMEUNIT_DAY:
        return gTimeUnitDay;
      case TimeUnit::UTIMEUNIT_WEEK:
        return gTimeUnitWeek;
      case TimeUnit::UTIMEUNIT_HOUR:
        return gTimeUnitHour;
      case TimeUnit::UTIMEUNIT_MINUTE:
        return gTimeUnitMinute;
      case TimeUnit::UTIMEUNIT_SECOND:
        return gTimeUnitSecond;
      default:
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
}

U_NAMESPACE_END

#endif
// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
//  **********************************************************************
//  *   Copyright (C) 1999-2016, International Business Machines
//  *   Corporation and others.  All Rights Reserved.
//  **********************************************************************
//  *   Date        Name        Description
//  *   11/17/99    aliu        Creation.
//  **********************************************************************
//  */

// #include "utypeinfo.h"  // for 'typeid' to work

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_TRANSLITERATION

// #include "unicode/putil.h"
// #include "unicode/translit.h"
// #include "unicode/locid.h"
// #include "unicode/msgfmt.h"
// #include "unicode/rep.h"
// #include "unicode/resbund.h"
// #include "unicode/unifilt.h"
// #include "unicode/uniset.h"
// #include "unicode/uscript.h"
// #include "unicode/strenum.h"
// #include "unicode/utf16.h"
// #include "cpdtrans.h"
// #include "nultrans.h"
// #include "rbt_data.h"
// #include "rbt_pars.h"
// #include "rbt.h"
// #include "transreg.h"
// #include "name2uni.h"
// #include "nortrans.h"
// #include "remtrans.h"
// #include "titletrn.h"
// #include "tolowtrn.h"
// #include "toupptrn.h"
// #include "uni2name.h"
// #include "brktrans.h"
// #include "esctrn.h"
// #include "unesctrn.h"
// #include "tridpars.h"
// #include "anytrans.h"
// #include "util.h"
// #include "hash.h"
// #include "mutex.h"
// #include "ucln_in.h"
// #include "uassert.h"
// #include "cmemory.h"
// #include "cstring.h"
// #include "uinvchar.h"

// static const UChar TARGET_SEP  = 0x002D; /*-*/
// static const UChar ID_DELIM    = 0x003B; /*;*/
// static const UChar VARIANT_SEP = 0x002F; // '/'

// /**
//  * Prefix for resource bundle key for the display name for a
//  * transliterator.  The ID is appended to this to form the key.
//  * The resource bundle value should be a String.
//  */
// static const char RB_DISPLAY_NAME_PREFIX[] = "%Translit%%";

// /**
//  * Prefix for resource bundle key for the display name for a
//  * transliterator SCRIPT.  The ID is appended to this to form the key.
//  * The resource bundle value should be a String.
//  */
// static const char RB_SCRIPT_DISPLAY_NAME_PREFIX[] = "%Translit%";

// /**
//  * Resource bundle key for display name pattern.
//  * The resource bundle value should be a String forming a
//  * MessageFormat pattern, e.g.:
//  * "{0,choice,0#|1#{1} Transliterator|2#{1} to {2} Transliterator}".
//  */
// static const char RB_DISPLAY_NAME_PATTERN[] = "TransliteratorNamePattern";

// /**
//  * Resource bundle key for the list of RuleBasedTransliterator IDs.
//  * The resource bundle value should be a String[] with each element
//  * being a valid ID.  The ID will be appended to RB_RULE_BASED_PREFIX
//  * to obtain the class name in which the RB_RULE key will be sought.
//  */
// static const char RB_RULE_BASED_IDS[] = "RuleBasedTransliteratorIDs";

// /**
//  * The mutex controlling access to registry object.
//  */
// static icu::UMutex registryMutex;

// /**
//  * System transliterator registry; non-null when initialized.
//  */
// static icu::TransliteratorRegistry* registry = 0;

// // Macro to check/initialize the registry. ONLY USE WITHIN
// // MUTEX. Avoids function call when registry is initialized.
// #define HAVE_REGISTRY(status) (registry!=0 || initializeRegistry(status))

// U_NAMESPACE_BEGIN

// UOBJECT_DEFINE_ABSTRACT_RTTI_IMPLEMENTATION(Transliterator)

// /**
//  * Return TRUE if the given UTransPosition is valid for text of
//  * the given length.
//  */
// static inline UBool positionIsValid(UTransPosition& index, int32_t len) {
//     return !(index.contextStart < 0 ||
//              index.start < index.contextStart ||
//              index.limit < index.start ||
//              index.contextLimit < index.limit ||
//              len < index.contextLimit);
// }

// /**
//  * Default constructor.
//  * @param theID the string identifier for this transliterator
//  * @param theFilter the filter.  Any character for which
//  * <tt>filter.contains()</tt> returns <tt>FALSE</tt> will not be
//  * altered by this transliterator.  If <tt>filter</tt> is
//  * <tt>null</tt> then no filtering is applied.
//  */
// Transliterator::Transliterator(const UnicodeString& theID,
//                                UnicodeFilter* adoptedFilter) :
//     UObject(), ID(theID), filter(adoptedFilter),
//     maximumContextLength(0)
// {
//     // NUL-terminate the ID string, which is a non-aliased copy.
//     ID.append((UChar)0);
//     ID.truncate(ID.length()-1);
// }

// /**
//  * Destructor.
//  */
// Transliterator::~Transliterator() {
//     if (filter) {
//         delete filter;
//     }
// }

// /**
//  * Copy constructor.
//  */
// Transliterator::Transliterator(const Transliterator& other) :
//     UObject(other), ID(other.ID), filter(0),
//     maximumContextLength(other.maximumContextLength)
// {
//     // NUL-terminate the ID string, which is a non-aliased copy.
//     ID.append((UChar)0);
//     ID.truncate(ID.length()-1);

//     if (other.filter != 0) {
//         // We own the filter, so we must have our own copy
//         filter = other.filter->clone();
//     }
// }

// Transliterator* Transliterator::clone() const {
//     return NULL;
// }

// /**
//  * Assignment operator.
//  */
// Transliterator& Transliterator::operator=(const Transliterator& other) {
//     ID = other.ID;
//     // NUL-terminate the ID string
//     ID.getTerminatedBuffer();

//     maximumContextLength = other.maximumContextLength;
//     adoptFilter((other.filter == 0) ? 0 : other.filter->clone());
//     return *this;
// }

// /**
//  * Transliterates a segment of a string.  <code>Transliterator</code> API.
//  * @param text the string to be transliterated
//  * @param start the beginning index, inclusive; <code>0 <= start
//  * <= limit</code>.
//  * @param limit the ending index, exclusive; <code>start <= limit
//  * <= text.length()</code>.
//  * @return the new limit index, or -1
//  */
// int32_t Transliterator::transliterate(Replaceable& text,
//                                       int32_t start, int32_t limit) const {
//     if (start < 0 ||
//         limit < start ||
//         text.length() < limit) {
//         return -1;
//     }

//     UTransPosition offsets;
//     offsets.contextStart= start;
//     offsets.contextLimit = limit;
//     offsets.start = start;
//     offsets.limit = limit;
//     filteredTransliterate(text, offsets, FALSE, TRUE);
//     return offsets.limit;
// }

// /**
//  * Transliterates an entire string in place. Convenience method.
//  * @param text the string to be transliterated
//  */
// void Transliterator::transliterate(Replaceable& text) const {
//     transliterate(text, 0, text.length());
// }

// /**
//  * Transliterates the portion of the text buffer that can be
//  * transliterated unambiguosly after new text has been inserted,
//  * typically as a result of a keyboard event.  The new text in
//  * <code>insertion</code> will be inserted into <code>text</code>
//  * at <code>index.contextLimit</code>, advancing
//  * <code>index.contextLimit</code> by <code>insertion.length()</code>.
//  * Then the transliterator will try to transliterate characters of
//  * <code>text</code> between <code>index.start</code> and
//  * <code>index.contextLimit</code>.  Characters before
//  * <code>index.start</code> will not be changed.
//  *
//  * <p>Upon return, values in <code>index</code> will be updated.
//  * <code>index.contextStart</code> will be advanced to the first
//  * character that future calls to this method will read.
//  * <code>index.start</code> and <code>index.contextLimit</code> will
//  * be adjusted to delimit the range of text that future calls to
//  * this method may change.
//  *
//  * <p>Typical usage of this method begins with an initial call
//  * with <code>index.contextStart</code> and <code>index.contextLimit</code>
//  * set to indicate the portion of <code>text</code> to be
//  * transliterated, and <code>index.start == index.contextStart</code>.
//  * Thereafter, <code>index</code> can be used without
//  * modification in future calls, provided that all changes to
//  * <code>text</code> are made via this method.
//  *
//  * <p>This method assumes that future calls may be made that will
//  * insert new text into the buffer.  As a result, it only performs
//  * unambiguous transliterations.  After the last call to this
//  * method, there may be untransliterated text that is waiting for
//  * more input to resolve an ambiguity.  In order to perform these
//  * pending transliterations, clients should call {@link
//  * #finishKeyboardTransliteration} after the last call to this
//  * method has been made.
//  *
//  * @param text the buffer holding transliterated and untransliterated text
//  * @param index an array of three integers.
//  *
//  * <ul><li><code>index.contextStart</code>: the beginning index,
//  * inclusive; <code>0 <= index.contextStart <= index.contextLimit</code>.
//  *
//  * <li><code>index.contextLimit</code>: the ending index, exclusive;
//  * <code>index.contextStart <= index.contextLimit <= text.length()</code>.
//  * <code>insertion</code> is inserted at
//  * <code>index.contextLimit</code>.
//  *
//  * <li><code>index.start</code>: the next character to be
//  * considered for transliteration; <code>index.contextStart <=
//  * index.start <= index.contextLimit</code>.  Characters before
//  * <code>index.start</code> will not be changed by future calls
//  * to this method.</ul>
//  *
//  * @param insertion text to be inserted and possibly
//  * transliterated into the translation buffer at
//  * <code>index.contextLimit</code>.  If <code>null</code> then no text
//  * is inserted.
//  * @see #START
//  * @see #LIMIT
//  * @see #CURSOR
//  * @see #handleTransliterate
//  * @exception IllegalArgumentException if <code>index</code>
//  * is invalid
//  */
// void Transliterator::transliterate(Replaceable& text,
//                                    UTransPosition& index,
//                                    const UnicodeString& insertion,
//                                    UErrorCode &status) const {
//     _transliterate(text, index, &insertion, status);
// }

// /**
//  * Transliterates the portion of the text buffer that can be
//  * transliterated unambiguosly after a new character has been
//  * inserted, typically as a result of a keyboard event.  This is a
//  * convenience method; see {@link
//  * #transliterate(Replaceable, int[], String)} for details.
//  * @param text the buffer holding transliterated and
//  * untransliterated text
//  * @param index an array of three integers.  See {@link
//  * #transliterate(Replaceable, int[], String)}.
//  * @param insertion text to be inserted and possibly
//  * transliterated into the translation buffer at
//  * <code>index.contextLimit</code>.
//  * @see #transliterate(Replaceable, int[], String)
//  */
// void Transliterator::transliterate(Replaceable& text,
//                                    UTransPosition& index,
//                                    UChar32 insertion,
//                                    UErrorCode& status) const {
//     UnicodeString str(insertion);
//     _transliterate(text, index, &str, status);
// }

// /**
//  * Transliterates the portion of the text buffer that can be
//  * transliterated unambiguosly.  This is a convenience method; see
//  * {@link #transliterate(Replaceable, int[], String)} for
//  * details.
//  * @param text the buffer holding transliterated and
//  * untransliterated text
//  * @param index an array of three integers.  See {@link
//  * #transliterate(Replaceable, int[], String)}.
//  * @see #transliterate(Replaceable, int[], String)
//  */
// void Transliterator::transliterate(Replaceable& text,
//                                    UTransPosition& index,
//                                    UErrorCode& status) const {
//     _transliterate(text, index, 0, status);
// }

// /**
//  * Finishes any pending transliterations that were waiting for
//  * more characters.  Clients should call this method as the last
//  * call after a sequence of one or more calls to
//  * <code>transliterate()</code>.
//  * @param text the buffer holding transliterated and
//  * untransliterated text.
//  * @param index the array of indices previously passed to {@link
//  * #transliterate}
//  */
// void Transliterator::finishTransliteration(Replaceable& text,
//                                            UTransPosition& index) const {
//     if (!positionIsValid(index, text.length())) {
//         return;
//     }

//     filteredTransliterate(text, index, FALSE, TRUE);
// }

// /**
//  * This internal method does keyboard transliteration.  If the
//  * 'insertion' is non-null then we append it to 'text' before
//  * proceeding.  This method calls through to the pure virtual
//  * framework method handleTransliterate() to do the actual
//  * work.
//  */
// void Transliterator::_transliterate(Replaceable& text,
//                                     UTransPosition& index,
//                                     const UnicodeString* insertion,
//                                     UErrorCode &status) const {
//     if (U_FAILURE(status)) {
//         return;
//     }

//     if (!positionIsValid(index, text.length())) {
//         status = U_ILLEGAL_ARGUMENT_ERROR;
//         return;
//     }

// //    int32_t originalStart = index.contextStart;
//     if (insertion != 0) {
//         text.handleReplaceBetween(index.limit, index.limit, *insertion);
//         index.limit += insertion->length();
//         index.contextLimit += insertion->length();
//     }

//     if (index.limit > 0 &&
//         U16_IS_LEAD(text.charAt(index.limit - 1))) {
//         // Oops, there is a dangling lead surrogate in the buffer.
//         // This will break most transliterators, since they will
//         // assume it is part of a pair.  Don't transliterate until
//         // more text comes in.
//         return;
//     }

//     filteredTransliterate(text, index, TRUE, TRUE);

// #if 0
//     // TODO
//     // I CAN'T DO what I'm attempting below now that the Kleene star
//     // operator is supported.  For example, in the rule

//     //   ([:Lu:]+) { x } > $1;

//     // what is the maximum context length?  getMaximumContextLength()
//     // will return 1, but this is just the length of the ante context
//     // part of the pattern string -- 1 character, which is a standin
//     // for a Quantifier, which contains a StringMatcher, which
//     // contains a UnicodeSet.

//     // There is a complicated way to make this work again, and that's
//     // to add a "maximum left context" protocol into the
//     // UnicodeMatcher hierarchy.  At present I'm not convinced this is
//     // worth it.

//     // ---

//     // The purpose of the code below is to keep the context small
//     // while doing incremental transliteration.  When part of the left
//     // context (between contextStart and start) is no longer needed,
//     // we try to advance contextStart past that portion.  We use the
//     // maximum context length to do so.
//     int32_t newCS = index.start;
//     int32_t n = getMaximumContextLength();
//     while (newCS > originalStart && n-- > 0) {
//         --newCS;
//         newCS -= U16_LENGTH(text.char32At(newCS)) - 1;
//     }
//     index.contextStart = uprv_max(newCS, originalStart);
// #endif
// }

// /**
//  * This method breaks up the input text into runs of unfiltered
//  * characters.  It passes each such run to
//  * <subclass>.handleTransliterate().  Subclasses that can handle the
//  * filter logic more efficiently themselves may override this method.
//  *
//  * All transliteration calls in this class go through this method.
//  */
// void Transliterator::filteredTransliterate(Replaceable& text,
//                                            UTransPosition& index,
//                                            UBool incremental,
//                                            UBool rollback) const {
//     // Short circuit path for transliterators with no filter in
//     // non-incremental mode.
//     if (filter == 0 && !rollback) {
//         handleTransliterate(text, index, incremental);
//         return;
//     }

//     //----------------------------------------------------------------------
//     // This method processes text in two groupings:
//     //
//     // RUNS -- A run is a contiguous group of characters which are contained
//     // in the filter for this transliterator (filter.contains(ch) == TRUE).
//     // Text outside of runs may appear as context but it is not modified.
//     // The start and limit Position values are narrowed to each run.
//     //
//     // PASSES (incremental only) -- To make incremental mode work correctly,
//     // each run is broken up into n passes, where n is the length (in code
//     // points) of the run.  Each pass contains the first n characters.  If a
//     // pass is completely transliterated, it is committed, and further passes
//     // include characters after the committed text.  If a pass is blocked,
//     // and does not transliterate completely, then this method rolls back
//     // the changes made during the pass, extends the pass by one code point,
//     // and tries again.
//     //----------------------------------------------------------------------

//     // globalLimit is the limit value for the entire operation.  We
//     // set index.limit to the end of each unfiltered run before
//     // calling handleTransliterate(), so we need to maintain the real
//     // value of index.limit here.  After each transliteration, we
//     // update globalLimit for insertions or deletions that have
//     // happened.
//     int32_t globalLimit = index.limit;

//     // If there is a non-null filter, then break the input text up.  Say the
//     // input text has the form:
//     //   xxxabcxxdefxx
//     // where 'x' represents a filtered character (filter.contains('x') ==
//     // false).  Then we break this up into:
//     //   xxxabc xxdef xx
//     // Each pass through the loop consumes a run of filtered
//     // characters (which are ignored) and a subsequent run of
//     // unfiltered characters (which are transliterated).

//     for (;;) {

//         if (filter != NULL) {
//             // Narrow the range to be transliterated to the first segment
//             // of unfiltered characters at or after index.start.

//             // Advance past filtered chars
//             UChar32 c;
//             while (index.start < globalLimit &&
//                    !filter->contains(c=text.char32At(index.start))) {
//                 index.start += U16_LENGTH(c);
//             }

//             // Find the end of this run of unfiltered chars
//             index.limit = index.start;
//             while (index.limit < globalLimit &&
//                    filter->contains(c=text.char32At(index.limit))) {
//                 index.limit += U16_LENGTH(c);
//             }
//         }

//         // Check to see if the unfiltered run is empty.  This only
//         // happens at the end of the string when all the remaining
//         // characters are filtered.
//         if (index.limit == index.start) {
//             // assert(index.start == globalLimit);
//             break;
//         }

//         // Is this run incremental?  If there is additional
//         // filtered text (if limit < globalLimit) then we pass in
//         // an incremental value of FALSE to force the subclass to
//         // complete the transliteration for this run.
//         UBool isIncrementalRun =
//             (index.limit < globalLimit ? FALSE : incremental);

//         int32_t delta;

//         // Implement rollback.  To understand the need for rollback,
//         // consider the following transliterator:
//         //
//         //  "t" is "a > A;"
//         //  "u" is "A > b;"
//         //  "v" is a compound of "t; NFD; u" with a filter [:Ll:]
//         //
//         // Now apply "c" to the input text "a".  The result is "b".  But if
//         // the transliteration is done incrementally, then the NFD holds
//         // things up after "t" has already transformed "a" to "A".  When
//         // finishTransliterate() is called, "A" is _not_ processed because
//         // it gets excluded by the [:Ll:] filter, and the end result is "A"
//         // -- incorrect.  The problem is that the filter is applied to a
//         // partially-transliterated result, when we only want it to apply to
//         // input text.  Although this example hinges on a compound
//         // transliterator containing NFD and a specific filter, it can
//         // actually happen with any transliterator which may do a partial
//         // transformation in incremental mode into characters outside its
//         // filter.
//         //
//         // To handle this, when in incremental mode we supply characters to
//         // handleTransliterate() in several passes.  Each pass adds one more
//         // input character to the input text.  That is, for input "ABCD", we
//         // first try "A", then "AB", then "ABC", and finally "ABCD".  If at
//         // any point we block (upon return, start < limit) then we roll
//         // back.  If at any point we complete the run (upon return start ==
//         // limit) then we commit that run.

//         if (rollback && isIncrementalRun) {

//             int32_t runStart = index.start;
//             int32_t runLimit = index.limit;
//             int32_t runLength =  runLimit - runStart;

//             // Make a rollback copy at the end of the string
//             int32_t rollbackOrigin = text.length();
//             text.copy(runStart, runLimit, rollbackOrigin);

//             // Variables reflecting the commitment of completely
//             // transliterated text.  passStart is the runStart, advanced
//             // past committed text.  rollbackStart is the rollbackOrigin,
//             // advanced past rollback text that corresponds to committed
//             // text.
//             int32_t passStart = runStart;
//             int32_t rollbackStart = rollbackOrigin;

//             // The limit for each pass; we advance by one code point with
//             // each iteration.
//             int32_t passLimit = index.start;

//             // Total length, in 16-bit code units, of uncommitted text.
//             // This is the length to be rolled back.
//             int32_t uncommittedLength = 0;

//             // Total delta (change in length) for all passes
//             int32_t totalDelta = 0;

//             // PASS MAIN LOOP -- Start with a single character, and extend
//             // the text by one character at a time.  Roll back partial
//             // transliterations and commit complete transliterations.
//             for (;;) {
//                 // Length of additional code point, either one or two
//                 int32_t charLength = U16_LENGTH(text.char32At(passLimit));
//                 passLimit += charLength;
//                 if (passLimit > runLimit) {
//                     break;
//                 }
//                 uncommittedLength += charLength;

//                 index.limit = passLimit;

//                 // Delegate to subclass for actual transliteration.  Upon
//                 // return, start will be updated to point after the
//                 // transliterated text, and limit and contextLimit will be
//                 // adjusted for length changes.
//                 handleTransliterate(text, index, TRUE);

//                 delta = index.limit - passLimit; // change in length

//                 // We failed to completely transliterate this pass.
//                 // Roll back the text.  Indices remain unchanged; reset
//                 // them where necessary.
//                 if (index.start != index.limit) {
//                     // Find the rollbackStart, adjusted for length changes
//                     // and the deletion of partially transliterated text.
//                     int32_t rs = rollbackStart + delta - (index.limit - passStart);

//                     // Delete the partially transliterated text
//                     text.handleReplaceBetween(passStart, index.limit, UnicodeString());

//                     // Copy the rollback text back
//                     text.copy(rs, rs + uncommittedLength, passStart);

//                     // Restore indices to their original values
//                     index.start = passStart;
//                     index.limit = passLimit;
//                     index.contextLimit -= delta;
//                 }

//                 // We did completely transliterate this pass.  Update the
//                 // commit indices to record how far we got.  Adjust indices
//                 // for length change.
//                 else {
//                     // Move the pass indices past the committed text.
//                     passStart = passLimit = index.start;

//                     // Adjust the rollbackStart for length changes and move
//                     // it past the committed text.  All characters we've
//                     // processed to this point are committed now, so zero
//                     // out the uncommittedLength.
//                     rollbackStart += delta + uncommittedLength;
//                     uncommittedLength = 0;

//                     // Adjust indices for length changes.
//                     runLimit += delta;
//                     totalDelta += delta;
//                 }
//             }

//             // Adjust overall limit and rollbackOrigin for insertions and
//             // deletions.  Don't need to worry about contextLimit because
//             // handleTransliterate() maintains that.
//             rollbackOrigin += totalDelta;
//             globalLimit += totalDelta;

//             // Delete the rollback copy
//             text.handleReplaceBetween(rollbackOrigin, rollbackOrigin + runLength, UnicodeString());

//             // Move start past committed text
//             index.start = passStart;
//         }

//         else {
//             // Delegate to subclass for actual transliteration.
//             int32_t limit = index.limit;
//             handleTransliterate(text, index, isIncrementalRun);
//             delta = index.limit - limit; // change in length

//             // In a properly written transliterator, start == limit after
//             // handleTransliterate() returns when incremental is false.
//             // Catch cases where the subclass doesn't do this, and throw
//             // an exception.  (Just pinning start to limit is a bad idea,
//             // because what's probably happening is that the subclass
//             // isn't transliterating all the way to the end, and it should
//             // in non-incremental mode.)
//             if (!incremental && index.start != index.limit) {
//                 // We can't throw an exception, so just fudge things
//                 index.start = index.limit;
//             }

//             // Adjust overall limit for insertions/deletions.  Don't need
//             // to worry about contextLimit because handleTransliterate()
//             // maintains that.
//             globalLimit += delta;
//         }

//         if (filter == NULL || isIncrementalRun) {
//             break;
//         }

//         // If we did completely transliterate this
//         // run, then repeat with the next unfiltered run.
//     }

//     // Start is valid where it is.  Limit needs to be put back where
//     // it was, modulo adjustments for deletions/insertions.
//     index.limit = globalLimit;
// }

// void Transliterator::filteredTransliterate(Replaceable& text,
//                                            UTransPosition& index,
//                                            UBool incremental) const {
//     filteredTransliterate(text, index, incremental, FALSE);
// }

// /**
//  * Method for subclasses to use to set the maximum context length.
//  * @see #getMaximumContextLength
//  */
// void Transliterator::setMaximumContextLength(int32_t maxContextLength) {
//     maximumContextLength = maxContextLength;
// }

// /**
//  * Returns a programmatic identifier for this transliterator.
//  * If this identifier is passed to <code>getInstance()</code>, it
//  * will return this object, if it has been registered.
//  * @see #registerInstance
//  * @see #getAvailableIDs
//  */
// const UnicodeString& Transliterator::getID(void) const {
//     return ID;
// }

// /**
//  * Returns a name for this transliterator that is appropriate for
//  * display to the user in the default locale.  See {@link
//  * #getDisplayName(Locale)} for details.
//  */
// UnicodeString& U_EXPORT2 Transliterator::getDisplayName(const UnicodeString& ID,
//                                               UnicodeString& result) {
//     return getDisplayName(ID, Locale::getDefault(), result);
// }

// /**
//  * Returns a name for this transliterator that is appropriate for
//  * display to the user in the given locale.  This name is taken
//  * from the locale resource data in the standard manner of the
//  * <code>java.text</code> package.
//  *
//  * <p>If no localized names exist in the system resource bundles,
//  * a name is synthesized using a localized
//  * <code>MessageFormat</code> pattern from the resource data.  The
//  * arguments to this pattern are an integer followed by one or two
//  * strings.  The integer is the number of strings, either 1 or 2.
//  * The strings are formed by splitting the ID for this
//  * transliterator at the first TARGET_SEP.  If there is no TARGET_SEP, then the
//  * entire ID forms the only string.
//  * @param inLocale the Locale in which the display name should be
//  * localized.
//  * @see java.text.MessageFormat
//  */
// UnicodeString& U_EXPORT2 Transliterator::getDisplayName(const UnicodeString& id,
//                                               const Locale& inLocale,
//                                               UnicodeString& result) {
//     UErrorCode status = U_ZERO_ERROR;

//     ResourceBundle bundle(U_ICUDATA_TRANSLIT, inLocale, status);

//     // Suspend checking status until later...

//     result.truncate(0);

//     // Normalize the ID
//     UnicodeString source, target, variant;
//     UBool sawSource;
//     TransliteratorIDParser::IDtoSTV(id, source, target, variant, sawSource);
//     if (target.length() < 1) {
//         // No target; malformed id
//         return result;
//     }
//     if (variant.length() > 0) { // Change "Foo" to "/Foo"
//         variant.insert(0, VARIANT_SEP);
//     }
//     UnicodeString ID(source);
//     ID.append(TARGET_SEP).append(target).append(variant);

//     // build the char* key
//     if (uprv_isInvariantUString(ID.getBuffer(), ID.length())) {
//         char key[200];
//         uprv_strcpy(key, RB_DISPLAY_NAME_PREFIX);
//         int32_t length=(int32_t)uprv_strlen(RB_DISPLAY_NAME_PREFIX);
//         ID.extract(0, (int32_t)(sizeof(key)-length), key+length, (int32_t)(sizeof(key)-length), US_INV);

//         // Try to retrieve a UnicodeString from the bundle.
//         UnicodeString resString = bundle.getStringEx(key, status);

//         if (U_SUCCESS(status) && resString.length() != 0) {
//             return result = resString; // [sic] assign & return
//         }

// #if !UCONFIG_NO_FORMATTING
//         // We have failed to get a name from the locale data.  This is
//         // typical, since most transliterators will not have localized
//         // name data.  The next step is to retrieve the MessageFormat
//         // pattern from the locale data and to use it to synthesize the
//         // name from the ID.

//         status = U_ZERO_ERROR;
//         resString = bundle.getStringEx(RB_DISPLAY_NAME_PATTERN, status);

//         if (U_SUCCESS(status) && resString.length() != 0) {
//             MessageFormat msg(resString, inLocale, status);
//             // Suspend checking status until later...

//             // We pass either 2 or 3 Formattable objects to msg.
//             Formattable args[3];
//             int32_t nargs;
//             args[0].setLong(2); // # of args to follow
//             args[1].setString(source);
//             args[2].setString(target);
//             nargs = 3;

//             // Use display names for the scripts, if they exist
//             UnicodeString s;
//             length=(int32_t)uprv_strlen(RB_SCRIPT_DISPLAY_NAME_PREFIX);
//             for (int j=1; j<=2; ++j) {
//                 status = U_ZERO_ERROR;
//                 uprv_strcpy(key, RB_SCRIPT_DISPLAY_NAME_PREFIX);
//                 args[j].getString(s);
//                 if (uprv_isInvariantUString(s.getBuffer(), s.length())) {
//                     s.extract(0, sizeof(key)-length-1, key+length, (int32_t)sizeof(key)-length-1, US_INV);

//                     resString = bundle.getStringEx(key, status);

//                     if (U_SUCCESS(status)) {
//                         args[j] = resString;
//                     }
//                 }
//             }

//             status = U_ZERO_ERROR;
//             FieldPosition pos; // ignored by msg
//             msg.format(args, nargs, result, pos, status);
//             if (U_SUCCESS(status)) {
//                 result.append(variant);
//                 return result;
//             }
//         }
// #endif
//     }

//     // We should not reach this point unless there is something
//     // wrong with the build or the RB_DISPLAY_NAME_PATTERN has
//     // been deleted from the root RB_LOCALE_ELEMENTS resource.
//     result = ID;
//     return result;
// }

// /**
//  * Returns the filter used by this transliterator, or <tt>null</tt>
//  * if this transliterator uses no filter.  Caller musn't delete
//  * the result!
//  */
// const UnicodeFilter* Transliterator::getFilter(void) const {
//     return filter;
// }

// /**
//  * Returns the filter used by this transliterator, or
//  * <tt>NULL</tt> if this transliterator uses no filter.  The
//  * caller must eventually delete the result.  After this call,
//  * this transliterator's filter is set to <tt>NULL</tt>.
//  */
// UnicodeFilter* Transliterator::orphanFilter(void) {
//     UnicodeFilter *result = filter;
//     filter = NULL;
//     return result;
// }

// /**
//  * Changes the filter used by this transliterator.  If the filter
//  * is set to <tt>null</tt> then no filtering will occur.
//  *
//  * <p>Callers must take care if a transliterator is in use by
//  * multiple threads.  The filter should not be changed by one
//  * thread while another thread may be transliterating.
//  */
// void Transliterator::adoptFilter(UnicodeFilter* filterToAdopt) {
//     delete filter;
//     filter = filterToAdopt;
// }

// /**
//  * Returns this transliterator's inverse.  See the class
//  * documentation for details.  This implementation simply inverts
//  * the two entities in the ID and attempts to retrieve the
//  * resulting transliterator.  That is, if <code>getID()</code>
//  * returns "A-B", then this method will return the result of
//  * <code>getInstance("B-A")</code>, or <code>null</code> if that
//  * call fails.
//  *
//  * <p>This method does not take filtering into account.  The
//  * returned transliterator will have no filter.
//  *
//  * <p>Subclasses with knowledge of their inverse may wish to
//  * override this method.
//  *
//  * @return a transliterator that is an inverse, not necessarily
//  * exact, of this transliterator, or <code>null</code> if no such
//  * transliterator is registered.
//  * @see #registerInstance
//  */
// Transliterator* Transliterator::createInverse(UErrorCode& status) const {
//     UParseError parseError;
//     return Transliterator::createInstance(ID, UTRANS_REVERSE,parseError,status);
// }

// Transliterator* U_EXPORT2
// Transliterator::createInstance(const UnicodeString& ID,
//                                 UTransDirection dir,
//                                 UErrorCode& status)
// {
//     UParseError parseError;
//     return createInstance(ID, dir, parseError, status);
// }

// /**
//  * Returns a <code>Transliterator</code> object given its ID.
//  * The ID must be either a system transliterator ID or a ID registered
//  * using <code>registerInstance()</code>.
//  *
//  * @param ID a valid ID, as enumerated by <code>getAvailableIDs()</code>
//  * @return A <code>Transliterator</code> object with the given ID
//  * @see #registerInstance
//  * @see #getAvailableIDs
//  * @see #getID
//  */
// Transliterator* U_EXPORT2
// Transliterator::createInstance(const UnicodeString& ID,
//                                 UTransDirection dir,
//                                 UParseError& parseError,
//                                 UErrorCode& status)
// {
//     if (U_FAILURE(status)) {
//         return 0;
//     }

//     UnicodeString canonID;
//     UVector list(status);
//     if (U_FAILURE(status)) {
//         return NULL;
//     }

//     UnicodeSet* globalFilter = nullptr;
//     // TODO add code for parseError...currently unused, but
//     // later may be used by parsing code...
//     if (!TransliteratorIDParser::parseCompoundID(ID, dir, canonID, list, globalFilter)) {
//         status = U_INVALID_ID;
//         delete globalFilter;
//         return NULL;
//     }
//     LocalPointer<UnicodeSet> lpGlobalFilter(globalFilter);

//     TransliteratorIDParser::instantiateList(list, status);
//     if (U_FAILURE(status)) {
//         return NULL;
//     }

//     U_ASSERT(list.size() > 0);
//     Transliterator* t = NULL;

//     if (list.size() > 1 || canonID.indexOf(ID_DELIM) >= 0) {
//         // [NOTE: If it's a compoundID, we instantiate a CompoundTransliterator even if it only
//         // has one child transliterator.  This is so that toRules() will return the right thing
//         // (without any inactive ID), but our main ID still comes out correct.  That is, if we
//         // instantiate "(Lower);Latin-Greek;", we want the rules to come out as "::Latin-Greek;"
//         // even though the ID is "(Lower);Latin-Greek;".
//         t = new CompoundTransliterator(list, parseError, status);
//     }
//     else {
//         t = (Transliterator*)list.elementAt(0);
//     }
//     // Check null pointer
//     if (t != NULL) {
//         t->setID(canonID);
//         if (lpGlobalFilter.isValid()) {
//             t->adoptFilter(lpGlobalFilter.orphan());
//         }
//     }
//     else if (U_SUCCESS(status)) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//     }
//     return t;
// }

// /**
//  * Create a transliterator from a basic ID.  This is an ID
//  * containing only the forward direction source, target, and
//  * variant.
//  * @param id a basic ID of the form S-T or S-T/V.
//  * @return a newly created Transliterator or null if the ID is
//  * invalid.
//  */
// Transliterator* Transliterator::createBasicInstance(const UnicodeString& id,
//                                                     const UnicodeString* canon) {
//     UParseError pe;
//     UErrorCode ec = U_ZERO_ERROR;
//     TransliteratorAlias* alias = 0;
//     Transliterator* t = 0;

//     umtx_lock(&registryMutex);
//     if (HAVE_REGISTRY(ec)) {
//         t = registry->get(id, alias, ec);
//     }
//     umtx_unlock(&registryMutex);

//     if (U_FAILURE(ec)) {
//         delete t;
//         delete alias;
//         return 0;
//     }

//     // We may have not gotten a transliterator:  Because we can't
//     // instantiate a transliterator from inside TransliteratorRegistry::
//     // get() (that would deadlock), we sometimes pass back an alias.  This
//     // contains the data we need to finish the instantiation outside the
//     // registry mutex.  The alias may, in turn, generate another alias, so
//     // we handle aliases in a loop.  The max times through the loop is two.
//     // [alan]
//     while (alias != 0) {
//         U_ASSERT(t==0);
//         // Rule-based aliases are handled with TransliteratorAlias::
//         // parse(), followed by TransliteratorRegistry::reget().
//         // Other aliases are handled with TransliteratorAlias::create().
//         if (alias->isRuleBased()) {
//             // Step 1. parse
//             TransliteratorParser parser(ec);
//             alias->parse(parser, pe, ec);
//             delete alias;
//             alias = 0;

//             // Step 2. reget
//             umtx_lock(&registryMutex);
//             if (HAVE_REGISTRY(ec)) {
//                 t = registry->reget(id, parser, alias, ec);
//             }
//             umtx_unlock(&registryMutex);

//             // Step 3. Loop back around!
//         } else {
//             t = alias->create(pe, ec);
//             delete alias;
//             alias = 0;
//             break;
//         }
//         if (U_FAILURE(ec)) {
//             delete t;
//             delete alias;
//             t = NULL;
//             break;
//         }
//     }

//     if (t != NULL && canon != NULL) {
//         t->setID(*canon);
//     }

//     return t;
// }

// /**
//  * Returns a <code>Transliterator</code> object constructed from
//  * the given rule string.  This will be a RuleBasedTransliterator,
//  * if the rule string contains only rules, or a
//  * CompoundTransliterator, if it contains ID blocks, or a
//  * NullTransliterator, if it contains ID blocks which parse as
//  * empty for the given direction.
//  */
// Transliterator* U_EXPORT2
// Transliterator::createFromRules(const UnicodeString& ID,
//                                 const UnicodeString& rules,
//                                 UTransDirection dir,
//                                 UParseError& parseError,
//                                 UErrorCode& status)
// {
//     Transliterator* t = NULL;

//     TransliteratorParser parser(status);
//     parser.parse(rules, dir, parseError, status);

//     if (U_FAILURE(status)) {
//         return 0;
//     }

//     // NOTE: The logic here matches that in TransliteratorRegistry.
//     if (parser.idBlockVector.size() == 0 && parser.dataVector.size() == 0) {
//         t = new NullTransliterator();
//     }
//     else if (parser.idBlockVector.size() == 0 && parser.dataVector.size() == 1) {
//         t = new RuleBasedTransliterator(ID, (TransliterationRuleData*)parser.dataVector.orphanElementAt(0), TRUE);
//     }
//     else if (parser.idBlockVector.size() == 1 && parser.dataVector.size() == 0) {
//         // idBlock, no data -- this is an alias.  The ID has
//         // been munged from reverse into forward mode, if
//         // necessary, so instantiate the ID in the forward
//         // direction.
//         if (parser.compoundFilter != NULL) {
//             UnicodeString filterPattern;
//             parser.compoundFilter->toPattern(filterPattern, FALSE);
//             t = createInstance(filterPattern + UnicodeString(ID_DELIM)
//                     + *((UnicodeString*)parser.idBlockVector.elementAt(0)), UTRANS_FORWARD, parseError, status);
//         }
//         else
//             t = createInstance(*((UnicodeString*)parser.idBlockVector.elementAt(0)), UTRANS_FORWARD, parseError, status);


//         if (t != NULL) {
//             t->setID(ID);
//         }
//     }
//     else {
//         UVector transliterators(status);
//         int32_t passNumber = 1;

//         int32_t limit = parser.idBlockVector.size();
//         if (parser.dataVector.size() > limit)
//             limit = parser.dataVector.size();

//         for (int32_t i = 0; i < limit; i++) {
//             if (i < parser.idBlockVector.size()) {
//                 UnicodeString* idBlock = (UnicodeString*)parser.idBlockVector.elementAt(i);
//                 if (!idBlock->isEmpty()) {
//                     Transliterator* temp = createInstance(*idBlock, UTRANS_FORWARD, parseError, status);
//                     if (U_FAILURE(status)) {
//                         delete temp;
//                         return nullptr;
//                     }
//                     if (temp != NULL && typeid(*temp) != typeid(NullTransliterator))
//                         transliterators.addElement(temp, status);
//                     else
//                         delete temp;
//                 }
//             }
//             if (!parser.dataVector.isEmpty()) {
//                 TransliterationRuleData* data = (TransliterationRuleData*)parser.dataVector.orphanElementAt(0);
//                 // TODO: Should passNumber be turned into a decimal-string representation (1 -> "1")?
//                 RuleBasedTransliterator* temprbt = new RuleBasedTransliterator(UnicodeString(CompoundTransliterator::PASS_STRING) + UnicodeString(passNumber++),
//                         data, TRUE);
//                 // Check if NULL before adding it to transliterators to avoid future usage of NULL pointer.
//                 if (temprbt == NULL) {
//                     if (U_SUCCESS(status)) {
//                         status = U_MEMORY_ALLOCATION_ERROR;
//                     }
//                     return t;
//                 }
//                 transliterators.addElement(temprbt, status);
//             }
//         }

//         t = new CompoundTransliterator(transliterators, passNumber - 1, parseError, status);
//         // Null pointer check
//         if (t != NULL) {
//             t->setID(ID);
//             t->adoptFilter(parser.orphanCompoundFilter());
//         }
//     }
//     if (U_SUCCESS(status) && t == NULL) {
//         status = U_MEMORY_ALLOCATION_ERROR;
//     }
//     return t;
// }

// UnicodeString& Transliterator::toRules(UnicodeString& rulesSource,
//                                        UBool escapeUnprintable) const {
//     // The base class implementation of toRules munges the ID into
//     // the correct format.  That is: foo => ::foo
//     if (escapeUnprintable) {
//         rulesSource.truncate(0);
//         UnicodeString id = getID();
//         for (int32_t i=0; i<id.length();) {
//             UChar32 c = id.char32At(i);
//             if (!ICU_Utility::escapeUnprintable(rulesSource, c)) {
//                 rulesSource.append(c);
//             }
//             i += U16_LENGTH(c);
//         }
//     } else {
//         rulesSource = getID();
//     }
//     // KEEP in sync with rbt_pars
//     rulesSource.insert(0, UNICODE_STRING_SIMPLE("::"));
//     rulesSource.append(ID_DELIM);
//     return rulesSource;
// }

// int32_t Transliterator::countElements() const {
//     const CompoundTransliterator* ct = dynamic_cast<const CompoundTransliterator*>(this);
//     return ct != NULL ? ct->getCount() : 0;
// }

// const Transliterator& Transliterator::getElement(int32_t index, UErrorCode& ec) const {
//     if (U_FAILURE(ec)) {
//         return *this;
//     }
//     const CompoundTransliterator* cpd = dynamic_cast<const CompoundTransliterator*>(this);
//     int32_t n = (cpd == NULL) ? 1 : cpd->getCount();
//     if (index < 0 || index >= n) {
//         ec = U_INDEX_OUTOFBOUNDS_ERROR;
//         return *this;
//     } else {
//         return (n == 1) ? *this : cpd->getTransliterator(index);
//     }
// }

// UnicodeSet& Transliterator::getSourceSet(UnicodeSet& result) const {
//     handleGetSourceSet(result);
//     if (filter != NULL) {
//         UnicodeSet* filterSet = dynamic_cast<UnicodeSet*>(filter);
//         UBool deleteFilterSet = FALSE;
//         // Most, but not all filters will be UnicodeSets.  Optimize for
//         // the high-runner case.
//         if (filterSet == NULL) {
//             filterSet = new UnicodeSet();
//             // Check null pointer
//             if (filterSet == NULL) {
//                 return result;
//             }
//             deleteFilterSet = TRUE;
//             filter->addMatchSetTo(*filterSet);
//         }
//         result.retainAll(*filterSet);
//         if (deleteFilterSet) {
//             delete filterSet;
//         }
//     }
//     return result;
// }

// void Transliterator::handleGetSourceSet(UnicodeSet& result) const {
//     result.clear();
// }

// UnicodeSet& Transliterator::getTargetSet(UnicodeSet& result) const {
//     return result.clear();
// }

// // For public consumption
// void U_EXPORT2 Transliterator::registerFactory(const UnicodeString& id,
//                                      Transliterator::Factory factory,
//                                      Transliterator::Token context) {
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     if (HAVE_REGISTRY(ec)) {
//         _registerFactory(id, factory, context);
//     }
// }

// // To be called only by Transliterator subclasses that are called
// // to register themselves by initializeRegistry().
// void Transliterator::_registerFactory(const UnicodeString& id,
//                                       Transliterator::Factory factory,
//                                       Transliterator::Token context) {
//     UErrorCode ec = U_ZERO_ERROR;
//     registry->put(id, factory, context, TRUE, ec);
// }

// // To be called only by Transliterator subclasses that are called
// // to register themselves by initializeRegistry().
// void Transliterator::_registerSpecialInverse(const UnicodeString& target,
//                                              const UnicodeString& inverseTarget,
//                                              UBool bidirectional) {
//     UErrorCode status = U_ZERO_ERROR;
//     TransliteratorIDParser::registerSpecialInverse(target, inverseTarget, bidirectional, status);
// }

// /**
//  * Registers a instance <tt>obj</tt> of a subclass of
//  * <code>Transliterator</code> with the system.  This object must
//  * implement the <tt>clone()</tt> method.  When
//  * <tt>getInstance()</tt> is called with an ID string that is
//  * equal to <tt>obj.getID()</tt>, then <tt>obj.clone()</tt> is
//  * returned.
//  *
//  * @param obj an instance of subclass of
//  * <code>Transliterator</code> that defines <tt>clone()</tt>
//  * @see #getInstance
//  * @see #unregister
//  */
// void U_EXPORT2 Transliterator::registerInstance(Transliterator* adoptedPrototype) {
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     if (HAVE_REGISTRY(ec)) {
//         _registerInstance(adoptedPrototype);
//     }
// }

// void Transliterator::_registerInstance(Transliterator* adoptedPrototype) {
//     UErrorCode ec = U_ZERO_ERROR;
//     registry->put(adoptedPrototype, TRUE, ec);
// }

// void U_EXPORT2 Transliterator::registerAlias(const UnicodeString& aliasID,
//                                              const UnicodeString& realID) {
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     if (HAVE_REGISTRY(ec)) {
//         _registerAlias(aliasID, realID);
//     }
// }

// void Transliterator::_registerAlias(const UnicodeString& aliasID,
//                                     const UnicodeString& realID) {
//     UErrorCode ec = U_ZERO_ERROR;
//     registry->put(aliasID, realID, FALSE, TRUE, ec);
// }

// /**
//  * Unregisters a transliterator or class.  This may be either
//  * a system transliterator or a user transliterator or class.
//  *
//  * @param ID the ID of the transliterator or class
//  * @see #registerInstance

//  */
// void U_EXPORT2 Transliterator::unregister(const UnicodeString& ID) {
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     if (HAVE_REGISTRY(ec)) {
//         registry->remove(ID);
//     }
// }

// /**
//  * == OBSOLETE - remove in ICU 3.4 ==
//  * Return the number of IDs currently registered with the system.
//  * To retrieve the actual IDs, call getAvailableID(i) with
//  * i from 0 to countAvailableIDs() - 1.
//  */
// int32_t U_EXPORT2 Transliterator::countAvailableIDs(void) {
//     int32_t retVal = 0;
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     if (HAVE_REGISTRY(ec)) {
//         retVal = registry->countAvailableIDs();
//     }
//     return retVal;
// }

// /**
//  * == OBSOLETE - remove in ICU 3.4 ==
//  * Return the index-th available ID.  index must be between 0
//  * and countAvailableIDs() - 1, inclusive.  If index is out of
//  * range, the result of getAvailableID(0) is returned.
//  */
// const UnicodeString& U_EXPORT2 Transliterator::getAvailableID(int32_t index) {
//     const UnicodeString* result = NULL;
//     umtx_lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     if (HAVE_REGISTRY(ec)) {
//         result = &registry->getAvailableID(index);
//     }
//     umtx_unlock(&registryMutex);
//     U_ASSERT(result != NULL); // fail if no registry
//     return *result;
// }

// StringEnumeration* U_EXPORT2 Transliterator::getAvailableIDs(UErrorCode& ec) {
//     if (U_FAILURE(ec)) return NULL;
//     StringEnumeration* result = NULL;
//     umtx_lock(&registryMutex);
//     if (HAVE_REGISTRY(ec)) {
//         result = registry->getAvailableIDs();
//     }
//     umtx_unlock(&registryMutex);
//     if (result == NULL) {
//         ec = U_INTERNAL_TRANSLITERATOR_ERROR;
//     }
//     return result;
// }

// int32_t U_EXPORT2 Transliterator::countAvailableSources(void) {
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     return HAVE_REGISTRY(ec) ? _countAvailableSources() : 0;
// }

// UnicodeString& U_EXPORT2 Transliterator::getAvailableSource(int32_t index,
//                                                   UnicodeString& result) {
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     if (HAVE_REGISTRY(ec)) {
//         _getAvailableSource(index, result);
//     }
//     return result;
// }

// int32_t U_EXPORT2 Transliterator::countAvailableTargets(const UnicodeString& source) {
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     return HAVE_REGISTRY(ec) ? _countAvailableTargets(source) : 0;
// }

// UnicodeString& U_EXPORT2 Transliterator::getAvailableTarget(int32_t index,
//                                                   const UnicodeString& source,
//                                                   UnicodeString& result) {
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     if (HAVE_REGISTRY(ec)) {
//         _getAvailableTarget(index, source, result);
//     }
//     return result;
// }

// int32_t U_EXPORT2 Transliterator::countAvailableVariants(const UnicodeString& source,
//                                                const UnicodeString& target) {
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     return HAVE_REGISTRY(ec) ? _countAvailableVariants(source, target) : 0;
// }

// UnicodeString& U_EXPORT2 Transliterator::getAvailableVariant(int32_t index,
//                                                    const UnicodeString& source,
//                                                    const UnicodeString& target,
//                                                    UnicodeString& result) {
//     Mutex lock(&registryMutex);
//     UErrorCode ec = U_ZERO_ERROR;
//     if (HAVE_REGISTRY(ec)) {
//         _getAvailableVariant(index, source, target, result);
//     }
//     return result;
// }

// int32_t Transliterator::_countAvailableSources(void) {
//     return registry->countAvailableSources();
// }

// UnicodeString& Transliterator::_getAvailableSource(int32_t index,
//                                                   UnicodeString& result) {
//     return registry->getAvailableSource(index, result);
// }

// int32_t Transliterator::_countAvailableTargets(const UnicodeString& source) {
//     return registry->countAvailableTargets(source);
// }

// UnicodeString& Transliterator::_getAvailableTarget(int32_t index,
//                                                   const UnicodeString& source,
//                                                   UnicodeString& result) {
//     return registry->getAvailableTarget(index, source, result);
// }

// int32_t Transliterator::_countAvailableVariants(const UnicodeString& source,
//                                                const UnicodeString& target) {
//     return registry->countAvailableVariants(source, target);
// }

// UnicodeString& Transliterator::_getAvailableVariant(int32_t index,
//                                                    const UnicodeString& source,
//                                                    const UnicodeString& target,
//                                                    UnicodeString& result) {
//     return registry->getAvailableVariant(index, source, target, result);
// }

// #ifdef U_USE_DEPRECATED_TRANSLITERATOR_API

// /**
//  * Method for subclasses to use to obtain a character in the given
//  * string, with filtering.
//  * @deprecated the new architecture provides filtering at the top
//  * level.  This method will be removed Dec 31 2001.
//  */
// UChar Transliterator::filteredCharAt(const Replaceable& text, int32_t i) const {
//     UChar c;
//     const UnicodeFilter* localFilter = getFilter();
//     return (localFilter == 0) ? text.charAt(i) :
//         (localFilter->contains(c = text.charAt(i)) ? c : (UChar)0xFFFE);
// }

// #endif

// /**
//  * If the registry is initialized, return TRUE.  If not, initialize it
//  * and return TRUE.  If the registry cannot be initialized, return
//  * FALSE (rare).
//  *
//  * IMPORTANT: Upon entry, registryMutex must be LOCKED.  The entire
//  * initialization is done with the lock held.  There is NO REASON to
//  * unlock, since no other thread that is waiting on the registryMutex
//  * cannot itself proceed until the registry is initialized.
//  */
// UBool Transliterator::initializeRegistry(UErrorCode &status) {
//     if (registry != 0) {
//         return TRUE;
//     }

//     registry = new TransliteratorRegistry(status);
//     if (registry == 0 || U_FAILURE(status)) {
//         delete registry;
//         registry = 0;
//         return FALSE; // can't create registry, no recovery
//     }

//     /* The following code parses the index table located in
//      * icu/data/translit/root.txt.  The index is an n x 4 table
//      * that follows this format:
//      *  <id>{
//      *      file{
//      *          resource{"<resource>"}
//      *          direction{"<direction>"}
//      *      }
//      *  }
//      *  <id>{
//      *      internal{
//      *          resource{"<resource>"}
//      *          direction{"<direction"}
//      *       }
//      *  }
//      *  <id>{
//      *      alias{"<getInstanceArg"}
//      *  }
//      * <id> is the ID of the system transliterator being defined.  These
//      * are public IDs enumerated by Transliterator.getAvailableIDs(),
//      * unless the second field is "internal".
//      *
//      * <resource> is a ResourceReader resource name.  Currently these refer
//      * to file names under com/ibm/text/resources.  This string is passed
//      * directly to ResourceReader, together with <encoding>.
//      *
//      * <direction> is either "FORWARD" or "REVERSE".
//      *
//      * <getInstanceArg> is a string to be passed directly to
//      * Transliterator.getInstance().  The returned Transliterator object
//      * then has its ID changed to <id> and is returned.
//      *
//      * The extra blank field on "alias" lines is to make the array square.
//      */
//     //static const char translit_index[] = "translit_index";

//     UResourceBundle *bundle, *transIDs, *colBund;
//     bundle = ures_open(U_ICUDATA_TRANSLIT, NULL/*open default locale*/, &status);
//     transIDs = ures_getByKey(bundle, RB_RULE_BASED_IDS, 0, &status);
//     const UnicodeString T_PART = UNICODE_STRING_SIMPLE("-t-");

//     int32_t row, maxRows;
//     if (U_SUCCESS(status)) {
//         maxRows = ures_getSize(transIDs);
//         for (row = 0; row < maxRows; row++) {
//             colBund = ures_getByIndex(transIDs, row, 0, &status);
//             if (U_SUCCESS(status)) {
//                 UnicodeString id(ures_getKey(colBund), -1, US_INV);
//                 if(id.indexOf(T_PART) != -1) {
//                     ures_close(colBund);
//                     continue;
//                 }
//                 UResourceBundle* res = ures_getNextResource(colBund, NULL, &status);
//                 const char* typeStr = ures_getKey(res);
//                 UChar type;
//                 u_charsToUChars(typeStr, &type, 1);

//                 if (U_SUCCESS(status)) {
//                     int32_t len = 0;
//                     const UChar *resString;
//                     switch (type) {
//                     case 0x66: // 'f'
//                     case 0x69: // 'i'
//                         // 'file' or 'internal';
//                         // row[2]=resource, row[3]=direction
//                         {

//                             resString = ures_getStringByKey(res, "resource", &len, &status);
//                             UBool visible = (type == 0x0066 /*f*/);
//                             UTransDirection dir =
//                                 (ures_getUnicodeStringByKey(res, "direction", &status).charAt(0) ==
//                                  0x0046 /*F*/) ?
//                                 UTRANS_FORWARD : UTRANS_REVERSE;
//                             registry->put(id, UnicodeString(TRUE, resString, len), dir, TRUE, visible, status);
//                         }
//                         break;
//                     case 0x61: // 'a'
//                         // 'alias'; row[2]=createInstance argument
//                         resString = ures_getString(res, &len, &status);
//                         registry->put(id, UnicodeString(TRUE, resString, len), TRUE, TRUE, status);
//                         break;
//                     }
//                 }
//                 ures_close(res);
//             }
//             ures_close(colBund);
//         }
//     }

//     ures_close(transIDs);
//     ures_close(bundle);

//     // Manually add prototypes that the system knows about to the
//     // cache.  This is how new non-rule-based transliterators are
//     // added to the system.

//     // This is to allow for null pointer check
//     NullTransliterator* tempNullTranslit = new NullTransliterator();
//     LowercaseTransliterator* tempLowercaseTranslit = new LowercaseTransliterator();
//     UppercaseTransliterator* tempUppercaseTranslit = new UppercaseTransliterator();
//     TitlecaseTransliterator* tempTitlecaseTranslit = new TitlecaseTransliterator();
//     UnicodeNameTransliterator* tempUnicodeTranslit = new UnicodeNameTransliterator();
//     NameUnicodeTransliterator* tempNameUnicodeTranslit = new NameUnicodeTransliterator();
// #if !UCONFIG_NO_BREAK_ITERATION
//      // TODO: could or should these transliterators be referenced polymorphically once constructed?
//      BreakTransliterator* tempBreakTranslit         = new BreakTransliterator();
// #endif
//     // Check for null pointers
//     if (tempNullTranslit == NULL || tempLowercaseTranslit == NULL || tempUppercaseTranslit == NULL ||
//         tempTitlecaseTranslit == NULL || tempUnicodeTranslit == NULL ||
// #if !UCONFIG_NO_BREAK_ITERATION
//         tempBreakTranslit == NULL ||
// #endif
//         tempNameUnicodeTranslit == NULL )
//     {
//         delete tempNullTranslit;
//         delete tempLowercaseTranslit;
//         delete tempUppercaseTranslit;
//         delete tempTitlecaseTranslit;
//         delete tempUnicodeTranslit;
//         delete tempNameUnicodeTranslit;
// #if !UCONFIG_NO_BREAK_ITERATION
//         delete tempBreakTranslit;
// #endif
//         // Since there was an error, remove registry
//         delete registry;
//         registry = NULL;

//         status = U_MEMORY_ALLOCATION_ERROR;
//         return 0;
//     }

//     registry->put(tempNullTranslit, TRUE, status);
//     registry->put(tempLowercaseTranslit, TRUE, status);
//     registry->put(tempUppercaseTranslit, TRUE, status);
//     registry->put(tempTitlecaseTranslit, TRUE, status);
//     registry->put(tempUnicodeTranslit, TRUE, status);
//     registry->put(tempNameUnicodeTranslit, TRUE, status);
// #if !UCONFIG_NO_BREAK_ITERATION
//     registry->put(tempBreakTranslit, FALSE, status);   // FALSE means invisible.
// #endif

//     RemoveTransliterator::registerIDs(); // Must be within mutex
//     EscapeTransliterator::registerIDs();
//     UnescapeTransliterator::registerIDs();
//     NormalizationTransliterator::registerIDs();
//     AnyTransliterator::registerIDs();

//     _registerSpecialInverse(UNICODE_STRING_SIMPLE("Null"),
//                             UNICODE_STRING_SIMPLE("Null"), FALSE);
//     _registerSpecialInverse(UNICODE_STRING_SIMPLE("Upper"),
//                             UNICODE_STRING_SIMPLE("Lower"), TRUE);
//     _registerSpecialInverse(UNICODE_STRING_SIMPLE("Title"),
//                             UNICODE_STRING_SIMPLE("Lower"), FALSE);

//     ucln_i18n_registerCleanup(UCLN_I18N_TRANSLITERATOR, utrans_transliterator_cleanup);

//     return TRUE;
// }

// U_NAMESPACE_END

// // Defined in transreg.h:

// /**
//  * Release all static memory held by transliterator.  This will
//  * necessarily invalidate any rule-based transliterators held by the
//  * user, because RBTs hold pointers to common data objects.
//  */
// U_CFUNC UBool utrans_transliterator_cleanup(void) {
//     U_NAMESPACE_USE
//     TransliteratorIDParser::cleanup();
//     if (registry) {
//         delete registry;
//         registry = NULL;
//     }
//     return TRUE;
// }

// #endif /* #if !UCONFIG_NO_TRANSLITERATION */

// //eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2011-2015, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING











// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2011-2012, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/

/**
 * \file 
 * \brief C API: Time zone generic names classe
 */



#if !UCONFIG_NO_FORMATTING






U_CDECL_BEGIN

typedef enum UTimeZoneGenericNameType {
    UTZGNM_UNKNOWN      = 0x00,
    UTZGNM_LOCATION     = 0x01,
    UTZGNM_LONG         = 0x02,
    UTZGNM_SHORT        = 0x04
} UTimeZoneGenericNameType;

U_CDECL_END

U_NAMESPACE_BEGIN

class TimeZone;
struct TZGNCoreRef;

class U_I18N_API TimeZoneGenericNames : public UMemory {
public:
    virtual ~TimeZoneGenericNames();

    static TimeZoneGenericNames* createInstance(const Locale& locale, UErrorCode& status);

    virtual UBool operator==(const TimeZoneGenericNames& other) const;
    virtual UBool operator!=(const TimeZoneGenericNames& other) const {return !operator==(other);}
    virtual TimeZoneGenericNames* clone() const;

    UnicodeString& getDisplayName(const TimeZone& tz, UTimeZoneGenericNameType type,
                        UDate date, UnicodeString& name) const;

    UnicodeString& getGenericLocationName(const UnicodeString& tzCanonicalID, UnicodeString& name) const;

    int32_t findBestMatch(const UnicodeString& text, int32_t start, uint32_t types,
        UnicodeString& tzID, UTimeZoneFormatTimeType& timeType, UErrorCode& status) const;

private:
    TimeZoneGenericNames();
    TZGNCoreRef* fRef;
};

U_NAMESPACE_END
#endif












   // TextTrieMap


U_NAMESPACE_BEGIN

// Bit flags used by the parse method.
// The order must match UTimeZoneFormatStyle enum.
#define ISO_Z_STYLE_FLAG 0x0080
#define ISO_LOCAL_STYLE_FLAG 0x0100
static const int16_t STYLE_PARSE_FLAGS[] = {
    0x0001, // UTZFMT_STYLE_GENERIC_LOCATION,
    0x0002, // UTZFMT_STYLE_GENERIC_LONG,
    0x0004, // UTZFMT_STYLE_GENERIC_SHORT,
    0x0008, // UTZFMT_STYLE_SPECIFIC_LONG,
    0x0010, // UTZFMT_STYLE_SPECIFIC_SHORT,
    0x0020, // UTZFMT_STYLE_LOCALIZED_GMT,
    0x0040, // UTZFMT_STYLE_LOCALIZED_GMT_SHORT,
    ISO_Z_STYLE_FLAG,       // UTZFMT_STYLE_ISO_BASIC_SHORT,
    ISO_LOCAL_STYLE_FLAG,   // UTZFMT_STYLE_ISO_BASIC_LOCAL_SHORT,
    ISO_Z_STYLE_FLAG,       // UTZFMT_STYLE_ISO_BASIC_FIXED,
    ISO_LOCAL_STYLE_FLAG,   // UTZFMT_STYLE_ISO_BASIC_LOCAL_FIXED,
    ISO_Z_STYLE_FLAG,       // UTZFMT_STYLE_ISO_BASIC_FULL,
    ISO_LOCAL_STYLE_FLAG,   // UTZFMT_STYLE_ISO_BASIC_LOCAL_FULL,
    ISO_Z_STYLE_FLAG,       // UTZFMT_STYLE_ISO_EXTENDED_FIXED,
    ISO_LOCAL_STYLE_FLAG,   // UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FIXED,
    ISO_Z_STYLE_FLAG,       // UTZFMT_STYLE_ISO_EXTENDED_FULL,
    ISO_LOCAL_STYLE_FLAG,   // UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FULL,
    0x0200, // UTZFMT_STYLE_ZONE_ID,
    0x0400, // UTZFMT_STYLE_ZONE_ID_SHORT,
    0x0800  // UTZFMT_STYLE_EXEMPLAR_LOCATION
};

static const char gZoneStringsTag[] = "zoneStrings";
static const char gGmtFormatTag[]= "gmtFormat";
static const char gGmtZeroFormatTag[] = "gmtZeroFormat";
static const char gHourFormatTag[]= "hourFormat";

static const UChar TZID_GMT[] = {0x0045, 0x0074, 0x0063, 0x002F, 0x0047, 0x004D, 0x0054, 0};    // Etc/GMT
static const UChar UNKNOWN_ZONE_ID[] = {
    0x0045, 0x0074, 0x0063, 0x002F, 0x0055, 0x006E, 0x006B, 0x006E, 0x006F, 0x0077, 0x006E, 0}; // Etc/Unknown
static const UChar UNKNOWN_SHORT_ZONE_ID[] = {0x0075, 0x006E, 0x006B, 0};   // unk
static const UChar UNKNOWN_LOCATION[] = {0x0055, 0x006E, 0x006B, 0x006E, 0x006F, 0x0077, 0x006E, 0};    // Unknown

static const UChar DEFAULT_GMT_PATTERN[] = {0x0047, 0x004D, 0x0054, 0x007B, 0x0030, 0x007D, 0}; // GMT{0}
//static const UChar DEFAULT_GMT_ZERO[] = {0x0047, 0x004D, 0x0054, 0}; // GMT
static const UChar DEFAULT_GMT_POSITIVE_HM[] = {0x002B, 0x0048, 0x003A, 0x006D, 0x006D, 0}; // +H:mm
static const UChar DEFAULT_GMT_POSITIVE_HMS[] = {0x002B, 0x0048, 0x003A, 0x006D, 0x006D, 0x003A, 0x0073, 0x0073, 0}; // +H:mm:ss
static const UChar DEFAULT_GMT_NEGATIVE_HM[] = {0x002D, 0x0048, 0x003A, 0x006D, 0x006D, 0}; // -H:mm
static const UChar DEFAULT_GMT_NEGATIVE_HMS[] = {0x002D, 0x0048, 0x003A, 0x006D, 0x006D, 0x003A, 0x0073, 0x0073, 0}; // -H:mm:ss
static const UChar DEFAULT_GMT_POSITIVE_H[] = {0x002B, 0x0048, 0}; // +H
static const UChar DEFAULT_GMT_NEGATIVE_H[] = {0x002D, 0x0048, 0}; // -H

static const UChar32 DEFAULT_GMT_DIGITS[] = {
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034,
    0x0035, 0x0036, 0x0037, 0x0038, 0x0039
};

static const UChar DEFAULT_GMT_OFFSET_SEP = 0x003A; // ':'

static const UChar ARG0[] = {0x007B, 0x0030, 0x007D};   // "{0}"
static const int32_t ARG0_LEN = 3;

static const UChar DEFAULT_GMT_OFFSET_MINUTE_PATTERN[] = {0x006D, 0x006D, 0};   // "mm"
static const UChar DEFAULT_GMT_OFFSET_SECOND_PATTERN[] = {0x0073, 0x0073, 0};   // "ss"

static const UChar ALT_GMT_STRINGS[][4] = {
    {0x0047, 0x004D, 0x0054, 0},    // GMT
    {0x0055, 0x0054, 0x0043, 0},    // UTC
    {0x0055, 0x0054, 0, 0},         // UT
    {0, 0, 0, 0}
};

// Order of GMT offset pattern parsing, *_HMS must be evaluated first
// because *_HM is most likely a substring of *_HMS
static const int32_t PARSE_GMT_OFFSET_TYPES[] = {
    UTZFMT_PAT_POSITIVE_HMS,
    UTZFMT_PAT_NEGATIVE_HMS,
    UTZFMT_PAT_POSITIVE_HM,
    UTZFMT_PAT_NEGATIVE_HM,
    UTZFMT_PAT_POSITIVE_H,
    UTZFMT_PAT_NEGATIVE_H,
    -1
};

static const UChar SINGLEQUOTE  = 0x0027;
static const UChar PLUS         = 0x002B;
static const UChar MINUS        = 0x002D;
static const UChar ISO8601_UTC  = 0x005A;   // 'Z'
static const UChar ISO8601_SEP  = 0x003A;   // ':'

static const int32_t MILLIS_PER_HOUR = 60 * 60 * 1000;
static const int32_t MILLIS_PER_MINUTE = 60 * 1000;
static const int32_t MILLIS_PER_SECOND = 1000;

// Maximum offset (exclusive) in millisecond supported by offset formats
static int32_t MAX_OFFSET = 24 * MILLIS_PER_HOUR;

// Maximum values for GMT offset fields
static const int32_t MAX_OFFSET_HOUR = 23;
static const int32_t MAX_OFFSET_MINUTE = 59;
static const int32_t MAX_OFFSET_SECOND = 59;

static const int32_t UNKNOWN_OFFSET = 0x7FFFFFFF;

static const int32_t ALL_SIMPLE_NAME_TYPES = UTZNM_LONG_STANDARD | UTZNM_LONG_DAYLIGHT | UTZNM_SHORT_STANDARD | UTZNM_SHORT_DAYLIGHT | UTZNM_EXEMPLAR_LOCATION;
static const int32_t ALL_GENERIC_NAME_TYPES = UTZGNM_LOCATION | UTZGNM_LONG | UTZGNM_SHORT;

#define DIGIT_VAL(c) (0x0030 <= (c) && (c) <= 0x0039 ? (c) - 0x0030 : -1)
#define MAX_OFFSET_DIGITS 6

// Time Zone ID/Short ID trie
static TextTrieMap *gZoneIdTrie = NULL;
static icu::UInitOnce gZoneIdTrieInitOnce = U_INITONCE_INITIALIZER;

static TextTrieMap *gShortZoneIdTrie = NULL;
static icu::UInitOnce gShortZoneIdTrieInitOnce = U_INITONCE_INITIALIZER;

static UMutex tzfmt_gLock;

U_CDECL_BEGIN
/**
 * Cleanup callback func
 */
static UBool U_CALLCONV tzfmt_cleanup(void)
{
    if (gZoneIdTrie != NULL) {
        delete gZoneIdTrie;
    }
    gZoneIdTrie = NULL;
    gZoneIdTrieInitOnce.reset();

    if (gShortZoneIdTrie != NULL) {
        delete gShortZoneIdTrie;
    }
    gShortZoneIdTrie = NULL;
    gShortZoneIdTrieInitOnce.reset();

    return TRUE;
}
U_CDECL_END

// ------------------------------------------------------------------
// GMTOffsetField
//
// This class represents a localized GMT offset pattern
// item and used by TimeZoneFormat
// ------------------------------------------------------------------
class GMTOffsetField : public UMemory {
public:
    enum FieldType {
        TEXT = 0,
        HOUR = 1,
        MINUTE = 2,
        SECOND = 4
    };

    virtual ~GMTOffsetField();

    static GMTOffsetField* createText(const UnicodeString& text, UErrorCode& status);
    static GMTOffsetField* createTimeField(FieldType type, uint8_t width, UErrorCode& status);
    static UBool isValid(FieldType type, int32_t width);
    static FieldType getTypeByLetter(UChar ch);

    FieldType getType() const;
    uint8_t getWidth() const;
    const UChar* getPatternText(void) const;

private:
    UChar* fText;
    FieldType fType;
    uint8_t fWidth;

    GMTOffsetField();
};

GMTOffsetField::GMTOffsetField()
: fText(NULL), fType(TEXT), fWidth(0) {
}

GMTOffsetField::~GMTOffsetField() {
    if (fText) {
        uprv_free(fText);
    }
}

GMTOffsetField*
GMTOffsetField::createText(const UnicodeString& text, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    GMTOffsetField* result = new GMTOffsetField();
    if (result == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    int32_t len = text.length();
    result->fText = (UChar*)uprv_malloc((len + 1) * sizeof(UChar));
    if (result->fText == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        delete result;
        return NULL;
    }
    u_strncpy(result->fText, text.getBuffer(), len);
    result->fText[len] = 0;
    result->fType = TEXT;

    return result;
}

GMTOffsetField*
GMTOffsetField::createTimeField(FieldType type, uint8_t width, UErrorCode& status) {
    U_ASSERT(type != TEXT);
    if (U_FAILURE(status)) {
        return NULL;
    }
    GMTOffsetField* result = new GMTOffsetField();
    if (result == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    result->fType = type;
    result->fWidth = width;

    return result;
}

UBool
GMTOffsetField::isValid(FieldType type, int32_t width) {
    switch (type) {
    case HOUR:
        return (width == 1 || width == 2);
    case MINUTE:
    case SECOND:
        return (width == 2);
    default:
        UPRV_UNREACHABLE;
    }
    return (width > 0);
}

GMTOffsetField::FieldType
GMTOffsetField::getTypeByLetter(UChar ch) {
    if (ch == 0x0048 /* H */) {
        return HOUR;
    } else if (ch == 0x006D /* m */) {
        return MINUTE;
    } else if (ch == 0x0073 /* s */) {
        return SECOND;
    }
    return TEXT;
}

inline GMTOffsetField::FieldType
GMTOffsetField::getType() const {
     return fType;
 }

inline uint8_t
GMTOffsetField::getWidth() const {
    return fWidth;
}

inline const UChar*
GMTOffsetField::getPatternText(void) const {
    return fText;
}


U_CDECL_BEGIN
static void U_CALLCONV
deleteGMTOffsetField(void *obj) {
    delete static_cast<GMTOffsetField *>(obj);
}
U_CDECL_END


// ------------------------------------------------------------------
// TimeZoneFormat
// ------------------------------------------------------------------
UOBJECT_DEFINE_RTTI_IMPLEMENTATION(TimeZoneFormat)

TimeZoneFormat::TimeZoneFormat(const Locale& locale, UErrorCode& status)
: fLocale(locale), fTimeZoneNames(NULL), fTimeZoneGenericNames(NULL),
  fDefParseOptionFlags(0), fTZDBTimeZoneNames(NULL) {

    for (int32_t i = 0; i < UTZFMT_PAT_COUNT; i++) {
        fGMTOffsetPatternItems[i] = NULL;
    }

    const char* region = fLocale.getCountry();
    int32_t regionLen = static_cast<int32_t>(uprv_strlen(region));
    if (regionLen == 0) {
        CharString loc;
        {
            CharStringByteSink sink(&loc);
            ulocimp_addLikelySubtags(fLocale.getName(), sink, &status);
        }

        regionLen = uloc_getCountry(loc.data(), fTargetRegion, sizeof(fTargetRegion), &status);
        if (U_SUCCESS(status)) {
            fTargetRegion[regionLen] = 0;
        } else {
            return;
        }
    } else if (regionLen < (int32_t)sizeof(fTargetRegion)) {
        uprv_strcpy(fTargetRegion, region);
    } else {
        fTargetRegion[0] = 0;
    }

    fTimeZoneNames = TimeZoneNames::createInstance(locale, status);
    // fTimeZoneGenericNames is lazily instantiated
    if (U_FAILURE(status)) {
        return;
    }

    const UChar* gmtPattern = NULL;
    const UChar* hourFormats = NULL;

    UResourceBundle *zoneBundle = ures_open(U_ICUDATA_ZONE, locale.getName(), &status);
    UResourceBundle *zoneStringsArray = ures_getByKeyWithFallback(zoneBundle, gZoneStringsTag, NULL, &status);
    if (U_SUCCESS(status)) {
        const UChar* resStr;
        int32_t len;
        resStr = ures_getStringByKeyWithFallback(zoneStringsArray, gGmtFormatTag, &len, &status);
        if (len > 0) {
            gmtPattern = resStr;
        }
        resStr = ures_getStringByKeyWithFallback(zoneStringsArray, gGmtZeroFormatTag, &len, &status);
        if (len > 0) {
            fGMTZeroFormat.setTo(TRUE, resStr, len);
        }
        resStr = ures_getStringByKeyWithFallback(zoneStringsArray, gHourFormatTag, &len, &status);
        if (len > 0) {
            hourFormats = resStr;
        }
        ures_close(zoneStringsArray);
        ures_close(zoneBundle);
    }

    if (gmtPattern == NULL) {
        gmtPattern = DEFAULT_GMT_PATTERN;
    }
    initGMTPattern(UnicodeString(TRUE, gmtPattern, -1), status);

    UBool useDefaultOffsetPatterns = TRUE;
    if (hourFormats) {
        UChar *sep = u_strchr(hourFormats, (UChar)0x003B /* ';' */);
        if (sep != NULL) {
            UErrorCode tmpStatus = U_ZERO_ERROR;
            fGMTOffsetPatterns[UTZFMT_PAT_POSITIVE_HM].setTo(FALSE, hourFormats, (int32_t)(sep - hourFormats));
            fGMTOffsetPatterns[UTZFMT_PAT_NEGATIVE_HM].setTo(TRUE, sep + 1, -1);
            expandOffsetPattern(fGMTOffsetPatterns[UTZFMT_PAT_POSITIVE_HM], fGMTOffsetPatterns[UTZFMT_PAT_POSITIVE_HMS], tmpStatus);
            expandOffsetPattern(fGMTOffsetPatterns[UTZFMT_PAT_NEGATIVE_HM], fGMTOffsetPatterns[UTZFMT_PAT_NEGATIVE_HMS], tmpStatus);
            truncateOffsetPattern(fGMTOffsetPatterns[UTZFMT_PAT_POSITIVE_HM], fGMTOffsetPatterns[UTZFMT_PAT_POSITIVE_H], tmpStatus);
            truncateOffsetPattern(fGMTOffsetPatterns[UTZFMT_PAT_NEGATIVE_HM], fGMTOffsetPatterns[UTZFMT_PAT_NEGATIVE_H], tmpStatus);
            if (U_SUCCESS(tmpStatus)) {
                useDefaultOffsetPatterns = FALSE;
            }
        }
    }
    if (useDefaultOffsetPatterns) {
        fGMTOffsetPatterns[UTZFMT_PAT_POSITIVE_H].setTo(TRUE, DEFAULT_GMT_POSITIVE_H, -1);
        fGMTOffsetPatterns[UTZFMT_PAT_POSITIVE_HM].setTo(TRUE, DEFAULT_GMT_POSITIVE_HM, -1);
        fGMTOffsetPatterns[UTZFMT_PAT_POSITIVE_HMS].setTo(TRUE, DEFAULT_GMT_POSITIVE_HMS, -1);
        fGMTOffsetPatterns[UTZFMT_PAT_NEGATIVE_H].setTo(TRUE, DEFAULT_GMT_NEGATIVE_H, -1);
        fGMTOffsetPatterns[UTZFMT_PAT_NEGATIVE_HM].setTo(TRUE, DEFAULT_GMT_NEGATIVE_HM, -1);
        fGMTOffsetPatterns[UTZFMT_PAT_NEGATIVE_HMS].setTo(TRUE, DEFAULT_GMT_NEGATIVE_HMS, -1);
    }
    initGMTOffsetPatterns(status);

    NumberingSystem* ns = NumberingSystem::createInstance(locale, status);
    UBool useDefDigits = TRUE;
    if (ns && !ns->isAlgorithmic()) {
        UnicodeString digits = ns->getDescription();
        useDefDigits = !toCodePoints(digits, fGMTOffsetDigits, 10);
    }
    if (useDefDigits) {
        uprv_memcpy(fGMTOffsetDigits, DEFAULT_GMT_DIGITS, sizeof(UChar32) * 10);
    }
    delete ns;
}

TimeZoneFormat::TimeZoneFormat(const TimeZoneFormat& other)
: Format(other), fTimeZoneNames(NULL), fTimeZoneGenericNames(NULL),
  fTZDBTimeZoneNames(NULL) {

    for (int32_t i = 0; i < UTZFMT_PAT_COUNT; i++) {
        fGMTOffsetPatternItems[i] = NULL;
    }
    *this = other;
}


TimeZoneFormat::~TimeZoneFormat() {
    delete fTimeZoneNames;
    delete fTimeZoneGenericNames;
    delete fTZDBTimeZoneNames;
    for (int32_t i = 0; i < UTZFMT_PAT_COUNT; i++) {
        delete fGMTOffsetPatternItems[i];
    }
}

TimeZoneFormat&
TimeZoneFormat::operator=(const TimeZoneFormat& other) {
    if (this == &other) {
        return *this;
    }

    delete fTimeZoneNames;
    delete fTimeZoneGenericNames;
    fTimeZoneGenericNames = NULL;
    delete fTZDBTimeZoneNames;
    fTZDBTimeZoneNames = NULL;

    fLocale = other.fLocale;
    uprv_memcpy(fTargetRegion, other.fTargetRegion, sizeof(fTargetRegion));

    fTimeZoneNames = other.fTimeZoneNames->clone();
    if (other.fTimeZoneGenericNames) {
        // TODO: this test has dubious thread safety.
        fTimeZoneGenericNames = other.fTimeZoneGenericNames->clone();
    }

    fGMTPattern = other.fGMTPattern;
    fGMTPatternPrefix = other.fGMTPatternPrefix;
    fGMTPatternSuffix = other.fGMTPatternSuffix;

    UErrorCode status = U_ZERO_ERROR;
    for (int32_t i = 0; i < UTZFMT_PAT_COUNT; i++) {
        fGMTOffsetPatterns[i] = other.fGMTOffsetPatterns[i];
        delete fGMTOffsetPatternItems[i];
        fGMTOffsetPatternItems[i] = NULL;
    }
    initGMTOffsetPatterns(status);
    U_ASSERT(U_SUCCESS(status));

    fGMTZeroFormat = other.fGMTZeroFormat;

    uprv_memcpy(fGMTOffsetDigits, other.fGMTOffsetDigits, sizeof(fGMTOffsetDigits));

    fDefParseOptionFlags = other.fDefParseOptionFlags;

    return *this;
}


UBool
TimeZoneFormat::operator==(const Format& other) const {
    TimeZoneFormat* tzfmt = (TimeZoneFormat*)&other;

    UBool isEqual =
            fLocale == tzfmt->fLocale
            && fGMTPattern == tzfmt->fGMTPattern
            && fGMTZeroFormat == tzfmt->fGMTZeroFormat
            && *fTimeZoneNames == *tzfmt->fTimeZoneNames;

    for (int32_t i = 0; i < UTZFMT_PAT_COUNT && isEqual; i++) {
        isEqual = fGMTOffsetPatterns[i] == tzfmt->fGMTOffsetPatterns[i];
    }
    for (int32_t i = 0; i < 10 && isEqual; i++) {
        isEqual = fGMTOffsetDigits[i] == tzfmt->fGMTOffsetDigits[i];
    }
    // TODO
    // Check fTimeZoneGenericNames. For now,
    // if fTimeZoneNames is same, fTimeZoneGenericNames should
    // be also equivalent.
    return isEqual;
}

TimeZoneFormat*
TimeZoneFormat::clone() const {
    return new TimeZoneFormat(*this);
}

TimeZoneFormat* U_EXPORT2
TimeZoneFormat::createInstance(const Locale& locale, UErrorCode& status) {
    TimeZoneFormat* tzfmt = new TimeZoneFormat(locale, status);
    if (U_SUCCESS(status)) {
        return tzfmt;
    }
    delete tzfmt;
    return NULL;
}

// ------------------------------------------------------------------
// Setter and Getter

const TimeZoneNames*
TimeZoneFormat::getTimeZoneNames() const {
    return (const TimeZoneNames*)fTimeZoneNames;
}

void
TimeZoneFormat::adoptTimeZoneNames(TimeZoneNames *tznames) {
    delete fTimeZoneNames;
    fTimeZoneNames = tznames;

    // TODO - We should also update fTimeZoneGenericNames
}

void
TimeZoneFormat::setTimeZoneNames(const TimeZoneNames &tznames) {
    delete fTimeZoneNames;
    fTimeZoneNames = tznames.clone();

    // TODO - We should also update fTimeZoneGenericNames
}

void
TimeZoneFormat::setDefaultParseOptions(uint32_t flags) {
    fDefParseOptionFlags = flags;
}

uint32_t
TimeZoneFormat::getDefaultParseOptions(void) const {
    return fDefParseOptionFlags;
}


UnicodeString&
TimeZoneFormat::getGMTPattern(UnicodeString& pattern) const {
    return pattern.setTo(fGMTPattern);
}

void
TimeZoneFormat::setGMTPattern(const UnicodeString& pattern, UErrorCode& status) {
    initGMTPattern(pattern, status);
}

UnicodeString&
TimeZoneFormat::getGMTOffsetPattern(UTimeZoneFormatGMTOffsetPatternType type, UnicodeString& pattern) const {
    return pattern.setTo(fGMTOffsetPatterns[type]);
}

void
TimeZoneFormat::setGMTOffsetPattern(UTimeZoneFormatGMTOffsetPatternType type, const UnicodeString& pattern, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    if (pattern == fGMTOffsetPatterns[type]) {
        // No need to reset
        return;
    }

    OffsetFields required = FIELDS_HM;
    switch (type) {
    case UTZFMT_PAT_POSITIVE_H:
    case UTZFMT_PAT_NEGATIVE_H:
        required = FIELDS_H;
        break;
    case UTZFMT_PAT_POSITIVE_HM:
    case UTZFMT_PAT_NEGATIVE_HM:
        required = FIELDS_HM;
        break;
    case UTZFMT_PAT_POSITIVE_HMS:
    case UTZFMT_PAT_NEGATIVE_HMS:
        required = FIELDS_HMS;
        break;
    default:
        UPRV_UNREACHABLE;
    }

    UVector* patternItems = parseOffsetPattern(pattern, required, status);
    if (patternItems == NULL) {
        return;
    }

    fGMTOffsetPatterns[type].setTo(pattern);
    delete fGMTOffsetPatternItems[type];
    fGMTOffsetPatternItems[type] = patternItems;
    checkAbuttingHoursAndMinutes();
}

UnicodeString&
TimeZoneFormat::getGMTOffsetDigits(UnicodeString& digits) const {
    digits.remove();
    for (int32_t i = 0; i < 10; i++) {
        digits.append(fGMTOffsetDigits[i]);
    }
    return digits;
}

void
TimeZoneFormat::setGMTOffsetDigits(const UnicodeString& digits, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    UChar32 digitArray[10];
    if (!toCodePoints(digits, digitArray, 10)) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    uprv_memcpy(fGMTOffsetDigits, digitArray, sizeof(UChar32)*10);
}

UnicodeString&
TimeZoneFormat::getGMTZeroFormat(UnicodeString& gmtZeroFormat) const {
    return gmtZeroFormat.setTo(fGMTZeroFormat);
}

void
TimeZoneFormat::setGMTZeroFormat(const UnicodeString& gmtZeroFormat, UErrorCode& status) {
    if (U_SUCCESS(status)) {
        if (gmtZeroFormat.isEmpty()) {
            status = U_ILLEGAL_ARGUMENT_ERROR;
        } else if (gmtZeroFormat != fGMTZeroFormat) {
            fGMTZeroFormat.setTo(gmtZeroFormat);
        }
    }
}

// ------------------------------------------------------------------
// Format and Parse

UnicodeString&
TimeZoneFormat::format(UTimeZoneFormatStyle style, const TimeZone& tz, UDate date,
        UnicodeString& name, UTimeZoneFormatTimeType* timeType /* = NULL */) const {
    if (timeType) {
        *timeType = UTZFMT_TIME_TYPE_UNKNOWN;
    }

    UBool noOffsetFormatFallback = FALSE;

    switch (style) {
    case UTZFMT_STYLE_GENERIC_LOCATION:
        formatGeneric(tz, UTZGNM_LOCATION, date, name);
        break;
    case UTZFMT_STYLE_GENERIC_LONG:
        formatGeneric(tz, UTZGNM_LONG, date, name);
        break;
    case UTZFMT_STYLE_GENERIC_SHORT:
        formatGeneric(tz, UTZGNM_SHORT, date, name);
        break;
    case UTZFMT_STYLE_SPECIFIC_LONG:
        formatSpecific(tz, UTZNM_LONG_STANDARD, UTZNM_LONG_DAYLIGHT, date, name, timeType);
        break;
    case UTZFMT_STYLE_SPECIFIC_SHORT:
        formatSpecific(tz, UTZNM_SHORT_STANDARD, UTZNM_SHORT_DAYLIGHT, date, name, timeType);
        break;

    case UTZFMT_STYLE_ZONE_ID:
        tz.getID(name);
        noOffsetFormatFallback = TRUE;
        break;
    case UTZFMT_STYLE_ZONE_ID_SHORT:
        {
            const UChar* shortID = ZoneMeta::getShortID(tz);
            if (shortID == NULL) {
                shortID = UNKNOWN_SHORT_ZONE_ID;
            }
            name.setTo(shortID, -1);
        }
        noOffsetFormatFallback = TRUE;
        break;

    case UTZFMT_STYLE_EXEMPLAR_LOCATION:
        formatExemplarLocation(tz, name);
        noOffsetFormatFallback = TRUE;
        break;

    default:
        // will be handled below
        break;
    }

    if (name.isEmpty() && !noOffsetFormatFallback) {
        UErrorCode status = U_ZERO_ERROR;
        int32_t rawOffset, dstOffset;
        tz.getOffset(date, FALSE, rawOffset, dstOffset, status);
        int32_t offset = rawOffset + dstOffset;
        if (U_SUCCESS(status)) {
            switch (style) {
            case UTZFMT_STYLE_GENERIC_LOCATION:
            case UTZFMT_STYLE_GENERIC_LONG:
            case UTZFMT_STYLE_SPECIFIC_LONG:
            case UTZFMT_STYLE_LOCALIZED_GMT:
                formatOffsetLocalizedGMT(offset, name, status);
                break;

            case UTZFMT_STYLE_GENERIC_SHORT:
            case UTZFMT_STYLE_SPECIFIC_SHORT:
            case UTZFMT_STYLE_LOCALIZED_GMT_SHORT:
                formatOffsetShortLocalizedGMT(offset, name, status);
                break;

            case UTZFMT_STYLE_ISO_BASIC_SHORT:
                formatOffsetISO8601Basic(offset, TRUE, TRUE, TRUE, name, status);
                break;

            case UTZFMT_STYLE_ISO_BASIC_LOCAL_SHORT:
                formatOffsetISO8601Basic(offset, FALSE, TRUE, TRUE, name, status);
                break;

            case UTZFMT_STYLE_ISO_BASIC_FIXED:
                formatOffsetISO8601Basic(offset, TRUE, FALSE, TRUE, name, status);
                break;

            case UTZFMT_STYLE_ISO_BASIC_LOCAL_FIXED:
                formatOffsetISO8601Basic(offset, FALSE, FALSE, TRUE, name, status);
                break;

            case UTZFMT_STYLE_ISO_EXTENDED_FIXED:
                formatOffsetISO8601Extended(offset, TRUE, FALSE, TRUE, name, status);
                break;

            case UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FIXED:
                formatOffsetISO8601Extended(offset, FALSE, FALSE, TRUE, name, status);
                break;

            case UTZFMT_STYLE_ISO_BASIC_FULL:
                formatOffsetISO8601Basic(offset, TRUE, FALSE, FALSE, name, status);
                break;

            case UTZFMT_STYLE_ISO_BASIC_LOCAL_FULL:
                formatOffsetISO8601Basic(offset, FALSE, FALSE, FALSE, name, status);
                break;

            case UTZFMT_STYLE_ISO_EXTENDED_FULL:
                formatOffsetISO8601Extended(offset, TRUE, FALSE, FALSE, name, status);
                break;

            case UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FULL:
                formatOffsetISO8601Extended(offset, FALSE, FALSE, FALSE, name, status);
                break;

            default:
              // UTZFMT_STYLE_ZONE_ID, UTZFMT_STYLE_ZONE_ID_SHORT, UTZFMT_STYLE_EXEMPLAR_LOCATION
              break;
            }

            if (timeType) {
                *timeType = (dstOffset != 0) ? UTZFMT_TIME_TYPE_DAYLIGHT : UTZFMT_TIME_TYPE_STANDARD;
            }
        }
    }

    return name;
}

UnicodeString&
TimeZoneFormat::format(const Formattable& obj, UnicodeString& appendTo,
        FieldPosition& pos, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return appendTo;
    }
    UDate date = Calendar::getNow();
    if (obj.getType() == Formattable::kObject) {
        const UObject* formatObj = obj.getObject();
        const TimeZone* tz = dynamic_cast<const TimeZone*>(formatObj);
        if (tz == NULL) {
            const Calendar* cal = dynamic_cast<const Calendar*>(formatObj);
            if (cal != NULL) {
                tz = &cal->getTimeZone();
                date = cal->getTime(status);
            }
        }
        if (tz != NULL) {
            int32_t rawOffset, dstOffset;
            tz->getOffset(date, FALSE, rawOffset, dstOffset, status);
            UChar buf[ZONE_NAME_U16_MAX];
            UnicodeString result(buf, 0, UPRV_LENGTHOF(buf));
            formatOffsetLocalizedGMT(rawOffset + dstOffset, result, status);
            if (U_SUCCESS(status)) {
                appendTo.append(result);
                if (pos.getField() == UDAT_TIMEZONE_FIELD) {
                    pos.setBeginIndex(0);
                    pos.setEndIndex(result.length());
                }
            }
        }
    }
    return appendTo;
}

TimeZone*
TimeZoneFormat::parse(UTimeZoneFormatStyle style, const UnicodeString& text, ParsePosition& pos,
        UTimeZoneFormatTimeType* timeType /*= NULL*/) const {
    return parse(style, text, pos, getDefaultParseOptions(), timeType);
}

TimeZone*
TimeZoneFormat::parse(UTimeZoneFormatStyle style, const UnicodeString& text, ParsePosition& pos,
        int32_t parseOptions, UTimeZoneFormatTimeType* timeType /* = NULL */) const {
    if (timeType) {
        *timeType = UTZFMT_TIME_TYPE_UNKNOWN;
    }

    int32_t startIdx = pos.getIndex();
    int32_t maxPos = text.length();
    int32_t offset;

    // Styles using localized GMT format as fallback
    UBool fallbackLocalizedGMT =
        (style == UTZFMT_STYLE_SPECIFIC_LONG || style == UTZFMT_STYLE_GENERIC_LONG || style == UTZFMT_STYLE_GENERIC_LOCATION);
    UBool fallbackShortLocalizedGMT =
        (style == UTZFMT_STYLE_SPECIFIC_SHORT || style == UTZFMT_STYLE_GENERIC_SHORT);

    int32_t evaluated = 0;  // bit flags representing already evaluated styles
    ParsePosition tmpPos(startIdx);

    int32_t parsedOffset = UNKNOWN_OFFSET;  // stores successfully parsed offset for later use
    int32_t parsedPos = -1;                 // stores successfully parsed offset position for later use

    // Try localized GMT format first if necessary
    if (fallbackLocalizedGMT || fallbackShortLocalizedGMT) {
        UBool hasDigitOffset = FALSE;
        offset = parseOffsetLocalizedGMT(text, tmpPos, fallbackShortLocalizedGMT, &hasDigitOffset);
        if (tmpPos.getErrorIndex() == -1) {
            // Even when the input text was successfully parsed as a localized GMT format text,
            // we may still need to evaluate the specified style if -
            //   1) GMT zero format was used, and
            //   2) The input text was not completely processed
            if (tmpPos.getIndex() == maxPos || hasDigitOffset) {
                pos.setIndex(tmpPos.getIndex());
                return createTimeZoneForOffset(offset);
            }
            parsedOffset = offset;
            parsedPos = tmpPos.getIndex();
        }
        // Note: For now, no distinction between long/short localized GMT format in the parser.
        // This might be changed in future.
        // evaluated |= (fallbackLocalizedGMT ? STYLE_PARSE_FLAGS[UTZFMT_STYLE_LOCALIZED_GMT] : STYLE_PARSE_FLAGS[UTZFMT_STYLE_LOCALIZED_GMT_SHORT]);
        evaluated |= STYLE_PARSE_FLAGS[UTZFMT_STYLE_LOCALIZED_GMT] | STYLE_PARSE_FLAGS[UTZFMT_STYLE_LOCALIZED_GMT_SHORT];
    }

    UErrorCode status = U_ZERO_ERROR;
    UChar tzIDBuf[32];
    UnicodeString tzID(tzIDBuf, 0, UPRV_LENGTHOF(tzIDBuf));

    UBool parseTZDBAbbrev = ((parseOptions & UTZFMT_PARSE_OPTION_TZ_DATABASE_ABBREVIATIONS) != 0);

    // Try the specified style
    switch (style) {
    case UTZFMT_STYLE_LOCALIZED_GMT:
        {
            tmpPos.setIndex(startIdx);
            tmpPos.setErrorIndex(-1);

            offset = parseOffsetLocalizedGMT(text, tmpPos);
            if (tmpPos.getErrorIndex() == -1) {
                pos.setIndex(tmpPos.getIndex());
                return createTimeZoneForOffset(offset);
            }

            // Note: For now, no distinction between long/short localized GMT format in the parser.
            // This might be changed in future.
            evaluated |= STYLE_PARSE_FLAGS[UTZFMT_STYLE_LOCALIZED_GMT_SHORT];

            break;
        }
    case UTZFMT_STYLE_LOCALIZED_GMT_SHORT:
        {
            tmpPos.setIndex(startIdx);
            tmpPos.setErrorIndex(-1);

            offset = parseOffsetShortLocalizedGMT(text, tmpPos);
            if (tmpPos.getErrorIndex() == -1) {
                pos.setIndex(tmpPos.getIndex());
                return createTimeZoneForOffset(offset);
            }

            // Note: For now, no distinction between long/short localized GMT format in the parser.
            // This might be changed in future.
            evaluated |= STYLE_PARSE_FLAGS[UTZFMT_STYLE_LOCALIZED_GMT];

            break;
        }
    case UTZFMT_STYLE_ISO_BASIC_SHORT:
    case UTZFMT_STYLE_ISO_BASIC_FIXED:
    case UTZFMT_STYLE_ISO_BASIC_FULL:
    case UTZFMT_STYLE_ISO_EXTENDED_FIXED:
    case UTZFMT_STYLE_ISO_EXTENDED_FULL:
        {
            tmpPos.setIndex(startIdx);
            tmpPos.setErrorIndex(-1);

            offset = parseOffsetISO8601(text, tmpPos);
            if (tmpPos.getErrorIndex() == -1) {
                pos.setIndex(tmpPos.getIndex());
                return createTimeZoneForOffset(offset);
            }

            break;
        }

    case UTZFMT_STYLE_ISO_BASIC_LOCAL_SHORT:
    case UTZFMT_STYLE_ISO_BASIC_LOCAL_FIXED:
    case UTZFMT_STYLE_ISO_BASIC_LOCAL_FULL:
    case UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FIXED:
    case UTZFMT_STYLE_ISO_EXTENDED_LOCAL_FULL:
        {
            tmpPos.setIndex(startIdx);
            tmpPos.setErrorIndex(-1);

            // Exclude the case of UTC Indicator "Z" here
            UBool hasDigitOffset = FALSE;
            offset = parseOffsetISO8601(text, tmpPos, FALSE, &hasDigitOffset);
            if (tmpPos.getErrorIndex() == -1 && hasDigitOffset) {
                pos.setIndex(tmpPos.getIndex());
                return createTimeZoneForOffset(offset);
            }

            break;
        }

    case UTZFMT_STYLE_SPECIFIC_LONG:
    case UTZFMT_STYLE_SPECIFIC_SHORT:
        {
            // Specific styles
            int32_t nameTypes = 0;
            if (style == UTZFMT_STYLE_SPECIFIC_LONG) {
                nameTypes = (UTZNM_LONG_STANDARD | UTZNM_LONG_DAYLIGHT);
            } else {
                U_ASSERT(style == UTZFMT_STYLE_SPECIFIC_SHORT);
                nameTypes = (UTZNM_SHORT_STANDARD | UTZNM_SHORT_DAYLIGHT);
            }
            LocalPointer<TimeZoneNames::MatchInfoCollection> specificMatches(fTimeZoneNames->find(text, startIdx, nameTypes, status));
            if (U_FAILURE(status)) {
                pos.setErrorIndex(startIdx);
                return NULL;
            }
            if (!specificMatches.isNull()) {
                int32_t matchIdx = -1;
                int32_t matchPos = -1;
                for (int32_t i = 0; i < specificMatches->size(); i++) {
                    matchPos  = startIdx + specificMatches->getMatchLengthAt(i);
                    if (matchPos > parsedPos) {
                        matchIdx = i;
                        parsedPos = matchPos;
                    }
                }
                if (matchIdx >= 0) {
                    if (timeType) {
                        *timeType = getTimeType(specificMatches->getNameTypeAt(matchIdx));
                    }
                    pos.setIndex(matchPos);
                    getTimeZoneID(specificMatches.getAlias(), matchIdx, tzID);
                    U_ASSERT(!tzID.isEmpty());
                    return TimeZone::createTimeZone(tzID);
                }
            }

            if (parseTZDBAbbrev && style == UTZFMT_STYLE_SPECIFIC_SHORT) {
                U_ASSERT((nameTypes & UTZNM_SHORT_STANDARD) != 0);
                U_ASSERT((nameTypes & UTZNM_SHORT_DAYLIGHT) != 0);

                const TZDBTimeZoneNames *tzdbTimeZoneNames = getTZDBTimeZoneNames(status);
                if (U_SUCCESS(status)) {
                    LocalPointer<TimeZoneNames::MatchInfoCollection> tzdbNameMatches(
                        tzdbTimeZoneNames->find(text, startIdx, nameTypes, status));
                    if (U_FAILURE(status)) {
                        pos.setErrorIndex(startIdx);
                        return NULL;
                    }
                    if (!tzdbNameMatches.isNull()) {
                        int32_t matchIdx = -1;
                        int32_t matchPos = -1;
                        for (int32_t i = 0; i < tzdbNameMatches->size(); i++) {
                            matchPos = startIdx + tzdbNameMatches->getMatchLengthAt(i);
                            if (matchPos > parsedPos) {
                                matchIdx = i;
                                parsedPos = matchPos;
                            }
                        }
                        if (matchIdx >= 0) {
                            if (timeType) {
                                *timeType = getTimeType(tzdbNameMatches->getNameTypeAt(matchIdx));
                            }
                            pos.setIndex(matchPos);
                            getTimeZoneID(tzdbNameMatches.getAlias(), matchIdx, tzID);
                            U_ASSERT(!tzID.isEmpty());
                            return TimeZone::createTimeZone(tzID);
                        }
                    }
                }
            }
            break;
        }
    case UTZFMT_STYLE_GENERIC_LONG:
    case UTZFMT_STYLE_GENERIC_SHORT:
    case UTZFMT_STYLE_GENERIC_LOCATION:
        {
            int32_t genericNameTypes = 0;
            switch (style) {
            case UTZFMT_STYLE_GENERIC_LOCATION:
                genericNameTypes = UTZGNM_LOCATION;
                break;

            case UTZFMT_STYLE_GENERIC_LONG:
                genericNameTypes = UTZGNM_LONG | UTZGNM_LOCATION;
                break;

            case UTZFMT_STYLE_GENERIC_SHORT:
                genericNameTypes = UTZGNM_SHORT | UTZGNM_LOCATION;
                break;

            default:
                UPRV_UNREACHABLE;
            }

            int32_t len = 0;
            UTimeZoneFormatTimeType tt = UTZFMT_TIME_TYPE_UNKNOWN;
            const TimeZoneGenericNames *gnames = getTimeZoneGenericNames(status);
            if (U_SUCCESS(status)) {
                len = gnames->findBestMatch(text, startIdx, genericNameTypes, tzID, tt, status);
            }
            if (U_FAILURE(status)) {
                pos.setErrorIndex(startIdx);
                return NULL;
            }
            if (len > 0) {
                // Found a match
                if (timeType) {
                    *timeType = tt;
                }
                pos.setIndex(startIdx + len);
                U_ASSERT(!tzID.isEmpty());
                return TimeZone::createTimeZone(tzID);
            }

            break;
        }
    case UTZFMT_STYLE_ZONE_ID:
        {
            tmpPos.setIndex(startIdx);
            tmpPos.setErrorIndex(-1);

            parseZoneID(text, tmpPos, tzID);
            if (tmpPos.getErrorIndex() == -1) {
                pos.setIndex(tmpPos.getIndex());
                return TimeZone::createTimeZone(tzID);
            }
            break;
        }
    case UTZFMT_STYLE_ZONE_ID_SHORT:
        {
            tmpPos.setIndex(startIdx);
            tmpPos.setErrorIndex(-1);

            parseShortZoneID(text, tmpPos, tzID);
            if (tmpPos.getErrorIndex() == -1) {
                pos.setIndex(tmpPos.getIndex());
                return TimeZone::createTimeZone(tzID);
            }
            break;
        }
    case UTZFMT_STYLE_EXEMPLAR_LOCATION:
        {
            tmpPos.setIndex(startIdx);
            tmpPos.setErrorIndex(-1);

            parseExemplarLocation(text, tmpPos, tzID);
            if (tmpPos.getErrorIndex() == -1) {
                pos.setIndex(tmpPos.getIndex());
                return TimeZone::createTimeZone(tzID);
            }
            break;
        }
    }
    evaluated |= STYLE_PARSE_FLAGS[style];


    if (parsedPos > startIdx) {
        // When the specified style is one of SPECIFIC_XXX or GENERIC_XXX, we tried to parse the input
        // as localized GMT format earlier. If parsedOffset is positive, it means it was successfully
        // parsed as localized GMT format, but offset digits were not detected (more specifically, GMT
        // zero format). Then, it tried to find a match within the set of display names, but could not
        // find a match. At this point, we can safely assume the input text contains the localized
        // GMT format.
        U_ASSERT(parsedOffset != UNKNOWN_OFFSET);
        pos.setIndex(parsedPos);
        return createTimeZoneForOffset(parsedOffset);
    }

    // Failed to parse the input text as the time zone format in the specified style.
    // Check the longest match among other styles below.
    UChar parsedIDBuf[32];
    UnicodeString parsedID(parsedIDBuf, 0, UPRV_LENGTHOF(parsedIDBuf));
    UTimeZoneFormatTimeType parsedTimeType = UTZFMT_TIME_TYPE_UNKNOWN;

    U_ASSERT(parsedPos < 0);
    U_ASSERT(parsedOffset == UNKNOWN_OFFSET);

    // ISO 8601
    if (parsedPos < maxPos &&
        ((evaluated & ISO_Z_STYLE_FLAG) == 0 || (evaluated & ISO_LOCAL_STYLE_FLAG) == 0)) {
        tmpPos.setIndex(startIdx);
        tmpPos.setErrorIndex(-1);

        UBool hasDigitOffset = FALSE;
        offset = parseOffsetISO8601(text, tmpPos, FALSE, &hasDigitOffset);
        if (tmpPos.getErrorIndex() == -1) {
            if (tmpPos.getIndex() == maxPos || hasDigitOffset) {
                pos.setIndex(tmpPos.getIndex());
                return createTimeZoneForOffset(offset);
            }
            // Note: When ISO 8601 format contains offset digits, it should not
            // collide with other formats. However, ISO 8601 UTC format "Z" (single letter)
            // may collide with other names. In this case, we need to evaluate other names.
            if (parsedPos < tmpPos.getIndex()) {
                parsedOffset = offset;
                parsedID.setToBogus();
                parsedTimeType = UTZFMT_TIME_TYPE_UNKNOWN;
                parsedPos = tmpPos.getIndex();
                U_ASSERT(parsedPos == startIdx + 1);    // only when "Z" is used
            }
        }
    }

    // Localized GMT format
    if (parsedPos < maxPos &&
        (evaluated & STYLE_PARSE_FLAGS[UTZFMT_STYLE_LOCALIZED_GMT]) == 0) {
        tmpPos.setIndex(startIdx);
        tmpPos.setErrorIndex(-1);

        UBool hasDigitOffset = FALSE;
        offset = parseOffsetLocalizedGMT(text, tmpPos, FALSE, &hasDigitOffset);
        if (tmpPos.getErrorIndex() == -1) {
            if (tmpPos.getIndex() == maxPos || hasDigitOffset) {
                pos.setIndex(tmpPos.getIndex());
                return createTimeZoneForOffset(offset);
            }
            // Evaluate other names - see the comment earlier in this method.
            if (parsedPos < tmpPos.getIndex()) {
                parsedOffset = offset;
                parsedID.setToBogus();
                parsedTimeType = UTZFMT_TIME_TYPE_UNKNOWN;
                parsedPos = tmpPos.getIndex();
            }
        }
    }

    if (parsedPos < maxPos &&
        (evaluated & STYLE_PARSE_FLAGS[UTZFMT_STYLE_LOCALIZED_GMT_SHORT]) == 0) {
        tmpPos.setIndex(startIdx);
        tmpPos.setErrorIndex(-1);

        UBool hasDigitOffset = FALSE;
        offset = parseOffsetLocalizedGMT(text, tmpPos, TRUE, &hasDigitOffset);
        if (tmpPos.getErrorIndex() == -1) {
            if (tmpPos.getIndex() == maxPos || hasDigitOffset) {
                pos.setIndex(tmpPos.getIndex());
                return createTimeZoneForOffset(offset);
            }
            // Evaluate other names - see the comment earlier in this method.
            if (parsedPos < tmpPos.getIndex()) {
                parsedOffset = offset;
                parsedID.setToBogus();
                parsedTimeType = UTZFMT_TIME_TYPE_UNKNOWN;
                parsedPos = tmpPos.getIndex();
            }
        }
    }

    // When ParseOption.ALL_STYLES is available, we also try to look all possible display names and IDs.
    // For example, when style is GENERIC_LONG, "EST" (SPECIFIC_SHORT) is never
    // used for America/New_York. With parseAllStyles true, this code parses "EST"
    // as America/New_York.

    // Note: Adding all possible names into the trie used by the implementation is quite heavy operation,
    // which we want to avoid normally (note that we cache the trie, so this is applicable to the
    // first time only as long as the cache does not expire).

    if (parseOptions & UTZFMT_PARSE_OPTION_ALL_STYLES) {
        // Try all specific names and exemplar location names
        if (parsedPos < maxPos) {
            LocalPointer<TimeZoneNames::MatchInfoCollection> specificMatches(fTimeZoneNames->find(text, startIdx, ALL_SIMPLE_NAME_TYPES, status));
            if (U_FAILURE(status)) {
                pos.setErrorIndex(startIdx);
                return NULL;
            }
            int32_t specificMatchIdx = -1;
            int32_t matchPos = -1;
            if (!specificMatches.isNull()) {
                for (int32_t i = 0; i < specificMatches->size(); i++) {
                    if (startIdx + specificMatches->getMatchLengthAt(i) > matchPos) {
                        specificMatchIdx = i;
                        matchPos = startIdx + specificMatches->getMatchLengthAt(i);
                    }
                }
            }
            if (parsedPos < matchPos) {
                U_ASSERT(specificMatchIdx >= 0);
                parsedPos = matchPos;
                getTimeZoneID(specificMatches.getAlias(), specificMatchIdx, parsedID);
                parsedTimeType = getTimeType(specificMatches->getNameTypeAt(specificMatchIdx));
                parsedOffset = UNKNOWN_OFFSET;
            }
        }
        if (parseTZDBAbbrev && parsedPos < maxPos && (evaluated & STYLE_PARSE_FLAGS[UTZFMT_STYLE_SPECIFIC_SHORT]) == 0) {
            const TZDBTimeZoneNames *tzdbTimeZoneNames = getTZDBTimeZoneNames(status);
            if (U_SUCCESS(status)) {
                LocalPointer<TimeZoneNames::MatchInfoCollection> tzdbNameMatches(
                    tzdbTimeZoneNames->find(text, startIdx, ALL_SIMPLE_NAME_TYPES, status));
                if (U_FAILURE(status)) {
                    pos.setErrorIndex(startIdx);
                    return NULL;
                }
                int32_t tzdbNameMatchIdx = -1;
                int32_t matchPos = -1;
                if (!tzdbNameMatches.isNull()) {
                    for (int32_t i = 0; i < tzdbNameMatches->size(); i++) {
                        if (startIdx + tzdbNameMatches->getMatchLengthAt(i) > matchPos) {
                            tzdbNameMatchIdx = i;
                            matchPos = startIdx + tzdbNameMatches->getMatchLengthAt(i);
                        }
                    }
                }
                if (parsedPos < matchPos) {
                    U_ASSERT(tzdbNameMatchIdx >= 0);
                    parsedPos = matchPos;
                    getTimeZoneID(tzdbNameMatches.getAlias(), tzdbNameMatchIdx, parsedID);
                    parsedTimeType = getTimeType(tzdbNameMatches->getNameTypeAt(tzdbNameMatchIdx));
                    parsedOffset = UNKNOWN_OFFSET;
                }
            }
        }
        // Try generic names
        if (parsedPos < maxPos) {
            int32_t genMatchLen = -1;
            UTimeZoneFormatTimeType tt = UTZFMT_TIME_TYPE_UNKNOWN;

            const TimeZoneGenericNames *gnames = getTimeZoneGenericNames(status);
            if (U_SUCCESS(status)) {
                genMatchLen = gnames->findBestMatch(text, startIdx, ALL_GENERIC_NAME_TYPES, tzID, tt, status);
            }
            if (U_FAILURE(status)) {
                pos.setErrorIndex(startIdx);
                return NULL;
            }

            if (genMatchLen > 0 && parsedPos < startIdx + genMatchLen) {
                parsedPos = startIdx + genMatchLen;
                parsedID.setTo(tzID);
                parsedTimeType = tt;
                parsedOffset = UNKNOWN_OFFSET;
            }
        }

        // Try time zone ID
        if (parsedPos < maxPos && (evaluated & STYLE_PARSE_FLAGS[UTZFMT_STYLE_ZONE_ID]) == 0) {
            tmpPos.setIndex(startIdx);
            tmpPos.setErrorIndex(-1);

            parseZoneID(text, tmpPos, tzID);
            if (tmpPos.getErrorIndex() == -1 && parsedPos < tmpPos.getIndex()) {
                parsedPos = tmpPos.getIndex();
                parsedID.setTo(tzID);
                parsedTimeType = UTZFMT_TIME_TYPE_UNKNOWN;
                parsedOffset = UNKNOWN_OFFSET;
            }
        }
        // Try short time zone ID
        if (parsedPos < maxPos && (evaluated & STYLE_PARSE_FLAGS[UTZFMT_STYLE_ZONE_ID]) == 0) {
            tmpPos.setIndex(startIdx);
            tmpPos.setErrorIndex(-1);

            parseShortZoneID(text, tmpPos, tzID);
            if (tmpPos.getErrorIndex() == -1 && parsedPos < tmpPos.getIndex()) {
                parsedPos = tmpPos.getIndex();
                parsedID.setTo(tzID);
                parsedTimeType = UTZFMT_TIME_TYPE_UNKNOWN;
                parsedOffset = UNKNOWN_OFFSET;
            }
        }
    }

    if (parsedPos > startIdx) {
        // Parsed successfully
        TimeZone* parsedTZ;
        if (parsedID.length() > 0) {
            parsedTZ = TimeZone::createTimeZone(parsedID);
        } else {
            U_ASSERT(parsedOffset != UNKNOWN_OFFSET);
            parsedTZ = createTimeZoneForOffset(parsedOffset);
        }
        if (timeType) {
            *timeType = parsedTimeType;
        }
        pos.setIndex(parsedPos);
        return parsedTZ;
    }

    pos.setErrorIndex(startIdx);
    return NULL;
}

void
TimeZoneFormat::parseObject(const UnicodeString& source, Formattable& result,
        ParsePosition& parse_pos) const {
    result.adoptObject(parse(UTZFMT_STYLE_GENERIC_LOCATION, source, parse_pos, UTZFMT_PARSE_OPTION_ALL_STYLES));
}


// ------------------------------------------------------------------
// Private zone name format/parse implementation

UnicodeString&
TimeZoneFormat::formatGeneric(const TimeZone& tz, int32_t genType, UDate date, UnicodeString& name) const {
    UErrorCode status = U_ZERO_ERROR;
    const TimeZoneGenericNames* gnames = getTimeZoneGenericNames(status);
    if (U_FAILURE(status)) {
        name.setToBogus();
        return name;
    }

    if (genType == UTZGNM_LOCATION) {
        const UChar* canonicalID = ZoneMeta::getCanonicalCLDRID(tz);
        if (canonicalID == NULL) {
            name.setToBogus();
            return name;
        }
        return gnames->getGenericLocationName(UnicodeString(TRUE, canonicalID, -1), name);
    }
    return gnames->getDisplayName(tz, (UTimeZoneGenericNameType)genType, date, name);
}

UnicodeString&
TimeZoneFormat::formatSpecific(const TimeZone& tz, UTimeZoneNameType stdType, UTimeZoneNameType dstType,
        UDate date, UnicodeString& name, UTimeZoneFormatTimeType *timeType) const {
    if (fTimeZoneNames == NULL) {
        name.setToBogus();
        return name;
    }

    UErrorCode status = U_ZERO_ERROR;
    UBool isDaylight = tz.inDaylightTime(date, status);
    const UChar* canonicalID = ZoneMeta::getCanonicalCLDRID(tz);

    if (U_FAILURE(status) || canonicalID == NULL) {
        name.setToBogus();
        return name;
    }

    if (isDaylight) {
        fTimeZoneNames->getDisplayName(UnicodeString(TRUE, canonicalID, -1), dstType, date, name);
    } else {
        fTimeZoneNames->getDisplayName(UnicodeString(TRUE, canonicalID, -1), stdType, date, name);
    }

    if (timeType && !name.isEmpty()) {
        *timeType = isDaylight ? UTZFMT_TIME_TYPE_DAYLIGHT : UTZFMT_TIME_TYPE_STANDARD;
    }
    return name;
}

const TimeZoneGenericNames*
TimeZoneFormat::getTimeZoneGenericNames(UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return NULL;
    }

    umtx_lock(&tzfmt_gLock);
    if (fTimeZoneGenericNames == NULL) {
        TimeZoneFormat *nonConstThis = const_cast<TimeZoneFormat *>(this);
        nonConstThis->fTimeZoneGenericNames = TimeZoneGenericNames::createInstance(fLocale, status);
    }
    umtx_unlock(&tzfmt_gLock);

    return fTimeZoneGenericNames;
}

const TZDBTimeZoneNames*
TimeZoneFormat::getTZDBTimeZoneNames(UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return NULL;
    }

    umtx_lock(&tzfmt_gLock);
    if (fTZDBTimeZoneNames == NULL) {
        TZDBTimeZoneNames *tzdbNames = new TZDBTimeZoneNames(fLocale);
        if (tzdbNames == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
        } else {
            TimeZoneFormat *nonConstThis = const_cast<TimeZoneFormat *>(this);
            nonConstThis->fTZDBTimeZoneNames = tzdbNames;
        }
    }
    umtx_unlock(&tzfmt_gLock);

    return fTZDBTimeZoneNames;
}

UnicodeString&
TimeZoneFormat::formatExemplarLocation(const TimeZone& tz, UnicodeString& name) const {
    UChar locationBuf[ZONE_NAME_U16_MAX];
    UnicodeString location(locationBuf, 0, UPRV_LENGTHOF(locationBuf));
    const UChar* canonicalID = ZoneMeta::getCanonicalCLDRID(tz);

    if (canonicalID) {
        fTimeZoneNames->getExemplarLocationName(UnicodeString(TRUE, canonicalID, -1), location);
    }
    if (location.length() > 0) {
        name.setTo(location);
    } else {
        // Use "unknown" location
        fTimeZoneNames->getExemplarLocationName(UnicodeString(TRUE, UNKNOWN_ZONE_ID, -1), location);
        if (location.length() > 0) {
            name.setTo(location);
        } else {
            // last resort
            name.setTo(UNKNOWN_LOCATION, -1);
        }
    }
    return name;
}


// ------------------------------------------------------------------
// Zone offset format and parse

UnicodeString&
TimeZoneFormat::formatOffsetISO8601Basic(int32_t offset, UBool useUtcIndicator, UBool isShort, UBool ignoreSeconds,
        UnicodeString& result, UErrorCode& status) const {
    return formatOffsetISO8601(offset, TRUE, useUtcIndicator, isShort, ignoreSeconds, result, status);
}

UnicodeString&
TimeZoneFormat::formatOffsetISO8601Extended(int32_t offset, UBool useUtcIndicator, UBool isShort, UBool ignoreSeconds,
        UnicodeString& result, UErrorCode& status) const {
    return formatOffsetISO8601(offset, FALSE, useUtcIndicator, isShort, ignoreSeconds, result, status);
}

UnicodeString&
TimeZoneFormat::formatOffsetLocalizedGMT(int32_t offset, UnicodeString& result, UErrorCode& status) const {
    return formatOffsetLocalizedGMT(offset, FALSE, result, status);
}

UnicodeString&
TimeZoneFormat::formatOffsetShortLocalizedGMT(int32_t offset, UnicodeString& result, UErrorCode& status) const {
    return formatOffsetLocalizedGMT(offset, TRUE, result, status);
}

int32_t
TimeZoneFormat::parseOffsetISO8601(const UnicodeString& text, ParsePosition& pos) const {
    return parseOffsetISO8601(text, pos, FALSE);
}

int32_t
TimeZoneFormat::parseOffsetLocalizedGMT(const UnicodeString& text, ParsePosition& pos) const {
    return parseOffsetLocalizedGMT(text, pos, FALSE, NULL);
}

int32_t
TimeZoneFormat::parseOffsetShortLocalizedGMT(const UnicodeString& text, ParsePosition& pos) const {
    return parseOffsetLocalizedGMT(text, pos, TRUE, NULL);
}

// ------------------------------------------------------------------
// Private zone offset format/parse implementation

UnicodeString&
TimeZoneFormat::formatOffsetISO8601(int32_t offset, UBool isBasic, UBool useUtcIndicator,
        UBool isShort, UBool ignoreSeconds, UnicodeString& result, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        result.setToBogus();
        return result;
    }
    int32_t absOffset = offset < 0 ? -offset : offset;
    if (useUtcIndicator && (absOffset < MILLIS_PER_SECOND || (ignoreSeconds && absOffset < MILLIS_PER_MINUTE))) {
        result.setTo(ISO8601_UTC);
        return result;
    }

    OffsetFields minFields = isShort ? FIELDS_H : FIELDS_HM;
    OffsetFields maxFields = ignoreSeconds ? FIELDS_HM : FIELDS_HMS;
    UChar sep = isBasic ? 0 : ISO8601_SEP;

    // Note: FIELDS_HMS as maxFields is a CLDR/ICU extension. ISO 8601 specification does
    // not support seconds field.

    if (absOffset >= MAX_OFFSET) {
        result.setToBogus();
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return result;
    }

    int fields[3];
    fields[0] = absOffset / MILLIS_PER_HOUR;
    absOffset = absOffset % MILLIS_PER_HOUR;
    fields[1] = absOffset / MILLIS_PER_MINUTE;
    absOffset = absOffset % MILLIS_PER_MINUTE;
    fields[2] = absOffset / MILLIS_PER_SECOND;

    U_ASSERT(fields[0] >= 0 && fields[0] <= MAX_OFFSET_HOUR);
    U_ASSERT(fields[1] >= 0 && fields[1] <= MAX_OFFSET_MINUTE);
    U_ASSERT(fields[2] >= 0 && fields[2] <= MAX_OFFSET_SECOND);

    int32_t lastIdx = maxFields;
    while (lastIdx > minFields) {
        if (fields[lastIdx] != 0) {
            break;
        }
        lastIdx--;
    }

    UChar sign = PLUS;
    if (offset < 0) {
        // if all output fields are 0s, do not use negative sign
        for (int32_t idx = 0; idx <= lastIdx; idx++) {
            if (fields[idx] != 0) {
                sign = MINUS;
                break;
            }
        }
    }
    result.setTo(sign);

    for (int32_t idx = 0; idx <= lastIdx; idx++) {
        if (sep && idx != 0) {
            result.append(sep);
        }
        result.append((UChar)(0x0030 + fields[idx]/10));
        result.append((UChar)(0x0030 + fields[idx]%10));
    }

    return result;
}

UnicodeString&
TimeZoneFormat::formatOffsetLocalizedGMT(int32_t offset, UBool isShort, UnicodeString& result, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        result.setToBogus();
        return result;
    }
    if (offset <= -MAX_OFFSET || offset >= MAX_OFFSET) {
        result.setToBogus();
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return result;
    }

    if (offset == 0) {
        result.setTo(fGMTZeroFormat);
        return result;
    }

    UBool positive = TRUE;
    if (offset < 0) {
        offset = -offset;
        positive = FALSE;
    }

    int32_t offsetH = offset / MILLIS_PER_HOUR;
    offset = offset % MILLIS_PER_HOUR;
    int32_t offsetM = offset / MILLIS_PER_MINUTE;
    offset = offset % MILLIS_PER_MINUTE;
    int32_t offsetS = offset / MILLIS_PER_SECOND;

    U_ASSERT(offsetH <= MAX_OFFSET_HOUR && offsetM <= MAX_OFFSET_MINUTE && offsetS <= MAX_OFFSET_SECOND);

    const UVector* offsetPatternItems = NULL;
    if (positive) {
        if (offsetS != 0) {
            offsetPatternItems = fGMTOffsetPatternItems[UTZFMT_PAT_POSITIVE_HMS];
        } else if (offsetM != 0 || !isShort) {
            offsetPatternItems = fGMTOffsetPatternItems[UTZFMT_PAT_POSITIVE_HM];
        } else {
            offsetPatternItems = fGMTOffsetPatternItems[UTZFMT_PAT_POSITIVE_H];
        }
    } else {
        if (offsetS != 0) {
            offsetPatternItems = fGMTOffsetPatternItems[UTZFMT_PAT_NEGATIVE_HMS];
        } else if (offsetM != 0 || !isShort) {
            offsetPatternItems = fGMTOffsetPatternItems[UTZFMT_PAT_NEGATIVE_HM];
        } else {
            offsetPatternItems = fGMTOffsetPatternItems[UTZFMT_PAT_NEGATIVE_H];
        }
    }

    U_ASSERT(offsetPatternItems != NULL);

    // Building the GMT format string
    result.setTo(fGMTPatternPrefix);

    for (int32_t i = 0; i < offsetPatternItems->size(); i++) {
        const GMTOffsetField* item = (GMTOffsetField*)offsetPatternItems->elementAt(i);
        GMTOffsetField::FieldType type = item->getType();

        switch (type) {
        case GMTOffsetField::TEXT:
            result.append(item->getPatternText(), -1);
            break;

        case GMTOffsetField::HOUR:
            appendOffsetDigits(result, offsetH, (isShort ? 1 : 2));
            break;

        case GMTOffsetField::MINUTE:
            appendOffsetDigits(result, offsetM, 2);
            break;

        case GMTOffsetField::SECOND:
            appendOffsetDigits(result, offsetS, 2);
            break;
        }
    }

    result.append(fGMTPatternSuffix);
    return result;
}

int32_t
TimeZoneFormat::parseOffsetISO8601(const UnicodeString& text, ParsePosition& pos, UBool extendedOnly, UBool* hasDigitOffset /* = NULL */) const {
    if (hasDigitOffset) {
        *hasDigitOffset = FALSE;
    }
    int32_t start = pos.getIndex();
    if (start >= text.length()) {
        pos.setErrorIndex(start);
        return 0;
    }

    UChar firstChar = text.charAt(start);
    if (firstChar == ISO8601_UTC || firstChar == (UChar)(ISO8601_UTC + 0x20)) {
        // "Z" (or "z") - indicates UTC
        pos.setIndex(start + 1);
        return 0;
    }

    int32_t sign = 1;
    if (firstChar == PLUS) {
        sign = 1;
    } else if (firstChar == MINUS) {
        sign = -1;
    } else {
        // Not an ISO 8601 offset string
        pos.setErrorIndex(start);
        return 0;
    }
    ParsePosition posOffset(start + 1);
    int32_t offset = parseAsciiOffsetFields(text, posOffset, ISO8601_SEP, FIELDS_H, FIELDS_HMS);
    if (posOffset.getErrorIndex() == -1 && !extendedOnly && (posOffset.getIndex() - start <= 3)) {
        // If the text is successfully parsed as extended format with the options above, it can be also parsed
        // as basic format. For example, "0230" can be parsed as offset 2:00 (only first digits are valid for
        // extended format), but it can be parsed as offset 2:30 with basic format. We use longer result.
        ParsePosition posBasic(start + 1);
        int32_t tmpOffset = parseAbuttingAsciiOffsetFields(text, posBasic, FIELDS_H, FIELDS_HMS, FALSE);
        if (posBasic.getErrorIndex() == -1 && posBasic.getIndex() > posOffset.getIndex()) {
            offset = tmpOffset;
            posOffset.setIndex(posBasic.getIndex());
        }
    }

    if (posOffset.getErrorIndex() != -1) {
        pos.setErrorIndex(start);
        return 0;
    }

    pos.setIndex(posOffset.getIndex());
    if (hasDigitOffset) {
        *hasDigitOffset = TRUE;
    }
    return sign * offset;
}

int32_t
TimeZoneFormat::parseOffsetLocalizedGMT(const UnicodeString& text, ParsePosition& pos, UBool isShort, UBool* hasDigitOffset) const {
    int32_t start = pos.getIndex();
    int32_t offset = 0;
    int32_t parsedLength = 0;

    if (hasDigitOffset) {
        *hasDigitOffset = FALSE;
    }

    offset = parseOffsetLocalizedGMTPattern(text, start, isShort, parsedLength);

    // For now, parseOffsetLocalizedGMTPattern handles both long and short
    // formats, no matter isShort is true or false. This might be changed in future
    // when strict parsing is necessary, or different set of patterns are used for
    // short/long formats.
#if 0
    if (parsedLength == 0) {
        offset = parseOffsetLocalizedGMTPattern(text, start, !isShort, parsedLength);
    }
#endif

    if (parsedLength > 0) {
        if (hasDigitOffset) {
            *hasDigitOffset = TRUE;
        }
        pos.setIndex(start + parsedLength);
        return offset;
    }

    // Try the default patterns
    offset = parseOffsetDefaultLocalizedGMT(text, start, parsedLength);
    if (parsedLength > 0) {
        if (hasDigitOffset) {
            *hasDigitOffset = TRUE;
        }
        pos.setIndex(start + parsedLength);
        return offset;
    }

    // Check if this is a GMT zero format
    if (text.caseCompare(start, fGMTZeroFormat.length(), fGMTZeroFormat, 0) == 0) {
        pos.setIndex(start + fGMTZeroFormat.length());
        return 0;
    }

    // Check if this is a default GMT zero format
    for (int32_t i = 0; ALT_GMT_STRINGS[i][0] != 0; i++) {
        const UChar* defGMTZero = ALT_GMT_STRINGS[i];
        int32_t defGMTZeroLen = u_strlen(defGMTZero);
        if (text.caseCompare(start, defGMTZeroLen, defGMTZero, 0) == 0) {
            pos.setIndex(start + defGMTZeroLen);
            return 0;
        }
    }

    // Nothing matched
    pos.setErrorIndex(start);
    return 0;
}

int32_t
TimeZoneFormat::parseOffsetLocalizedGMTPattern(const UnicodeString& text, int32_t start, UBool /*isShort*/, int32_t& parsedLen) const {
    int32_t idx = start;
    int32_t offset = 0;
    UBool parsed = FALSE;

    do {
        // Prefix part
        int32_t len = fGMTPatternPrefix.length();
        if (len > 0 && text.caseCompare(idx, len, fGMTPatternPrefix, 0) != 0) {
            // prefix match failed
            break;
        }
        idx += len;

        // Offset part
        offset = parseOffsetFields(text, idx, FALSE, len);
        if (len == 0) {
            // offset field match failed
            break;
        }
        idx += len;

        len = fGMTPatternSuffix.length();
        if (len > 0 && text.caseCompare(idx, len, fGMTPatternSuffix, 0) != 0) {
            // no suffix match
            break;
        }
        idx += len;
        parsed = TRUE;
    } while (FALSE);

    parsedLen = parsed ? idx - start : 0;
    return offset;
}

int32_t
TimeZoneFormat::parseOffsetFields(const UnicodeString& text, int32_t start, UBool /*isShort*/, int32_t& parsedLen) const {
    int32_t outLen = 0;
    int32_t offset = 0;
    int32_t sign = 1;

    parsedLen = 0;

    int32_t offsetH, offsetM, offsetS;
    offsetH = offsetM = offsetS = 0;

    for (int32_t patidx = 0; PARSE_GMT_OFFSET_TYPES[patidx] >= 0; patidx++) {
        int32_t gmtPatType = PARSE_GMT_OFFSET_TYPES[patidx];
        UVector* items = fGMTOffsetPatternItems[gmtPatType];
        U_ASSERT(items != NULL);

        outLen = parseOffsetFieldsWithPattern(text, start, items, FALSE, offsetH, offsetM, offsetS);
        if (outLen > 0) {
            sign = (gmtPatType == UTZFMT_PAT_POSITIVE_H || gmtPatType == UTZFMT_PAT_POSITIVE_HM || gmtPatType == UTZFMT_PAT_POSITIVE_HMS) ?
                1 : -1;
            break;
        }
    }

    if (outLen > 0 && fAbuttingOffsetHoursAndMinutes) {
        // When hours field is sabutting minutes field,
        // the parse result above may not be appropriate.
        // For example, "01020" is parsed as 01:02: above,
        // but it should be parsed as 00:10:20.
        int32_t tmpLen = 0;
        int32_t tmpSign = 1;
        int32_t tmpH = 0;
        int32_t tmpM = 0;
        int32_t tmpS = 0;

        for (int32_t patidx = 0; PARSE_GMT_OFFSET_TYPES[patidx] >= 0; patidx++) {
            int32_t gmtPatType = PARSE_GMT_OFFSET_TYPES[patidx];
            UVector* items = fGMTOffsetPatternItems[gmtPatType];
            U_ASSERT(items != NULL);

            // forcing parse to use single hour digit
            tmpLen = parseOffsetFieldsWithPattern(text, start, items, TRUE, tmpH, tmpM, tmpS);
            if (tmpLen > 0) {
                tmpSign = (gmtPatType == UTZFMT_PAT_POSITIVE_H || gmtPatType == UTZFMT_PAT_POSITIVE_HM || gmtPatType == UTZFMT_PAT_POSITIVE_HMS) ?
                    1 : -1;
                break;
            }
        }
        if (tmpLen > outLen) {
            // Better parse result with single hour digit
            outLen = tmpLen;
            sign = tmpSign;
            offsetH = tmpH;
            offsetM = tmpM;
            offsetS = tmpS;
        }
    }

    if (outLen > 0) {
        offset = ((((offsetH * 60) + offsetM) * 60) + offsetS) * 1000 * sign;
        parsedLen = outLen;
    }

    return offset;
}

int32_t
TimeZoneFormat::parseOffsetFieldsWithPattern(const UnicodeString& text, int32_t start,
        UVector* patternItems, UBool forceSingleHourDigit, int32_t& hour, int32_t& min, int32_t& sec) const {
    UBool failed = FALSE;
    int32_t offsetH, offsetM, offsetS;
    offsetH = offsetM = offsetS = 0;
    int32_t idx = start;

    for (int32_t i = 0; i < patternItems->size(); i++) {
        int32_t len = 0;
        const GMTOffsetField* field = (const GMTOffsetField*)patternItems->elementAt(i);
        GMTOffsetField::FieldType fieldType = field->getType();
        if (fieldType == GMTOffsetField::TEXT) {
            const UChar* patStr = field->getPatternText();
            len = u_strlen(patStr);
            if (i == 0) {
                // When TimeZoneFormat parse() is called from SimpleDateFormat,
                // leading space characters might be truncated. If the first pattern text
                // starts with such character (e.g. Bidi control), then we need to
                // skip the leading space charcters.
                if (idx < text.length() && !PatternProps::isWhiteSpace(text.char32At(idx))) {
                    while (len > 0) {
                        UChar32 ch;
                        int32_t chLen;
                        U16_GET(patStr, 0, 0, len, ch);
                        if (PatternProps::isWhiteSpace(ch)) {
                            chLen = U16_LENGTH(ch);
                            len -= chLen;
                            patStr += chLen;
                        }
                        else {
                            break;
                        }
                    }
                }
            }
            if (text.caseCompare(idx, len, patStr, 0) != 0) {
                failed = TRUE;
                break;
            }
            idx += len;
        } else {
            if (fieldType == GMTOffsetField::HOUR) {
                uint8_t maxDigits = forceSingleHourDigit ? 1 : 2;
                offsetH = parseOffsetFieldWithLocalizedDigits(text, idx, 1, maxDigits, 0, MAX_OFFSET_HOUR, len);
            } else if (fieldType == GMTOffsetField::MINUTE) {
                offsetM = parseOffsetFieldWithLocalizedDigits(text, idx, 2, 2, 0, MAX_OFFSET_MINUTE, len);
            } else if (fieldType == GMTOffsetField::SECOND) {
                offsetS = parseOffsetFieldWithLocalizedDigits(text, idx, 2, 2, 0, MAX_OFFSET_SECOND, len);
            }

            if (len == 0) {
                failed = TRUE;
                break;
            }
            idx += len;
        }
    }

    if (failed) {
        hour = min = sec = 0;
        return 0;
    }

    hour = offsetH;
    min = offsetM;
    sec = offsetS;

    return idx - start;
}

int32_t
TimeZoneFormat::parseAbuttingOffsetFields(const UnicodeString& text, int32_t start, int32_t& parsedLen) const {
    int32_t digits[MAX_OFFSET_DIGITS];
    int32_t parsed[MAX_OFFSET_DIGITS];  // accumulative offsets

    // Parse digits into int[]
    int32_t idx = start;
    int32_t len = 0;
    int32_t numDigits = 0;
    for (int32_t i = 0; i < MAX_OFFSET_DIGITS; i++) {
        digits[i] = parseSingleLocalizedDigit(text, idx, len);
        if (digits[i] < 0) {
            break;
        }
        idx += len;
        parsed[i] = idx - start;
        numDigits++;
    }

    if (numDigits == 0) {
        parsedLen = 0;
        return 0;
    }

    int32_t offset = 0;
    while (numDigits > 0) {
        int32_t hour = 0;
        int32_t min = 0;
        int32_t sec = 0;

        U_ASSERT(numDigits > 0 && numDigits <= MAX_OFFSET_DIGITS);
        switch (numDigits) {
        case 1: // H
            hour = digits[0];
            break;
        case 2: // HH
            hour = digits[0] * 10 + digits[1];
            break;
        case 3: // Hmm
            hour = digits[0];
            min = digits[1] * 10 + digits[2];
            break;
        case 4: // HHmm
            hour = digits[0] * 10 + digits[1];
            min = digits[2] * 10 + digits[3];
            break;
        case 5: // Hmmss
            hour = digits[0];
            min = digits[1] * 10 + digits[2];
            sec = digits[3] * 10 + digits[4];
            break;
        case 6: // HHmmss
            hour = digits[0] * 10 + digits[1];
            min = digits[2] * 10 + digits[3];
            sec = digits[4] * 10 + digits[5];
            break;
        }
        if (hour <= MAX_OFFSET_HOUR && min <= MAX_OFFSET_MINUTE && sec <= MAX_OFFSET_SECOND) {
            // found a valid combination
            offset = hour * MILLIS_PER_HOUR + min * MILLIS_PER_MINUTE + sec * MILLIS_PER_SECOND;
            parsedLen = parsed[numDigits - 1];
            break;
        }
        numDigits--;
    }
    return offset;
}

int32_t
TimeZoneFormat::parseOffsetDefaultLocalizedGMT(const UnicodeString& text, int start, int32_t& parsedLen) const {
    int32_t idx = start;
    int32_t offset = 0;
    int32_t parsed = 0;

    do {
        // check global default GMT alternatives
        int32_t gmtLen = 0;

        for (int32_t i = 0; ALT_GMT_STRINGS[i][0] != 0; i++) {
            const UChar* gmt = ALT_GMT_STRINGS[i];
            int32_t len = u_strlen(gmt);
            if (text.caseCompare(start, len, gmt, 0) == 0) {
                gmtLen = len;
                break;
            }
        }
        if (gmtLen == 0) {
            break;
        }
        idx += gmtLen;

        // offset needs a sign char and a digit at minimum
        if (idx + 1 >= text.length()) {
            break;
        }

        // parse sign
        int32_t sign = 1;
        UChar c = text.charAt(idx);
        if (c == PLUS) {
            sign = 1;
        } else if (c == MINUS) {
            sign = -1;
        } else {
            break;
        }
        idx++;

        // offset part
        // try the default pattern with the separator first
        int32_t lenWithSep = 0;
        int32_t offsetWithSep = parseDefaultOffsetFields(text, idx, DEFAULT_GMT_OFFSET_SEP, lenWithSep);
        if (lenWithSep == text.length() - idx) {
            // maximum match
            offset = offsetWithSep * sign;
            idx += lenWithSep;
        } else {
            // try abutting field pattern
            int32_t lenAbut = 0;
            int32_t offsetAbut = parseAbuttingOffsetFields(text, idx, lenAbut);

            if (lenWithSep > lenAbut) {
                offset = offsetWithSep * sign;
                idx += lenWithSep;
            } else {
                offset = offsetAbut * sign;
                idx += lenAbut;
            }
        }
        parsed = idx - start;
    } while (false);

    parsedLen = parsed;
    return offset;
}

int32_t
TimeZoneFormat::parseDefaultOffsetFields(const UnicodeString& text, int32_t start, UChar separator, int32_t& parsedLen) const {
    int32_t max = text.length();
    int32_t idx = start;
    int32_t len = 0;
    int32_t hour = 0, min = 0, sec = 0;

    parsedLen = 0;

    do {
        hour = parseOffsetFieldWithLocalizedDigits(text, idx, 1, 2, 0, MAX_OFFSET_HOUR, len);
        if (len == 0) {
            break;
        }
        idx += len;

        if (idx + 1 < max && text.charAt(idx) == separator) {
            min = parseOffsetFieldWithLocalizedDigits(text, idx + 1, 2, 2, 0, MAX_OFFSET_MINUTE, len);
            if (len == 0) {
                break;
            }
            idx += (1 + len);

            if (idx + 1 < max && text.charAt(idx) == separator) {
                sec = parseOffsetFieldWithLocalizedDigits(text, idx + 1, 2, 2, 0, MAX_OFFSET_SECOND, len);
                if (len == 0) {
                    break;
                }
                idx += (1 + len);
            }
        }
    } while (FALSE);

    if (idx == start) {
        return 0;
    }

    parsedLen = idx - start;
    return hour * MILLIS_PER_HOUR + min * MILLIS_PER_MINUTE + sec * MILLIS_PER_SECOND;
}

int32_t
TimeZoneFormat::parseOffsetFieldWithLocalizedDigits(const UnicodeString& text, int32_t start, uint8_t minDigits, uint8_t maxDigits, uint16_t minVal, uint16_t maxVal, int32_t& parsedLen) const {
    parsedLen = 0;

    int32_t decVal = 0;
    int32_t numDigits = 0;
    int32_t idx = start;
    int32_t digitLen = 0;

    while (idx < text.length() && numDigits < maxDigits) {
        int32_t digit = parseSingleLocalizedDigit(text, idx, digitLen);
        if (digit < 0) {
            break;
        }
        int32_t tmpVal = decVal * 10 + digit;
        if (tmpVal > maxVal) {
            break;
        }
        decVal = tmpVal;
        numDigits++;
        idx += digitLen;
    }

    // Note: maxVal is checked in the while loop
    if (numDigits < minDigits || decVal < minVal) {
        decVal = -1;
        numDigits = 0;
    } else {
        parsedLen = idx - start;
    }

    return decVal;
}

int32_t
TimeZoneFormat::parseSingleLocalizedDigit(const UnicodeString& text, int32_t start, int32_t& len) const {
    int32_t digit = -1;
    len = 0;
    if (start < text.length()) {
        UChar32 cp = text.char32At(start);

        // First, try digits configured for this instance
        for (int32_t i = 0; i < 10; i++) {
            if (cp == fGMTOffsetDigits[i]) {
                digit = i;
                break;
            }
        }
        // If failed, check if this is a Unicode digit
        if (digit < 0) {
            int32_t tmp = u_charDigitValue(cp);
            digit = (tmp >= 0 && tmp <= 9) ? tmp : -1;
        }

        if (digit >= 0) {
            int32_t next = text.moveIndex32(start, 1);
            len = next - start;
        }
    }
    return digit;
}

UnicodeString&
TimeZoneFormat::formatOffsetWithAsciiDigits(int32_t offset, UChar sep, OffsetFields minFields, OffsetFields maxFields, UnicodeString& result) {
    U_ASSERT(maxFields >= minFields);
    U_ASSERT(offset > -MAX_OFFSET && offset < MAX_OFFSET);

    UChar sign = PLUS;
    if (offset < 0) {
        sign = MINUS;
        offset = -offset;
    }
    result.setTo(sign);

    int fields[3];
    fields[0] = offset / MILLIS_PER_HOUR;
    offset = offset % MILLIS_PER_HOUR;
    fields[1] = offset / MILLIS_PER_MINUTE;
    offset = offset % MILLIS_PER_MINUTE;
    fields[2] = offset / MILLIS_PER_SECOND;

    U_ASSERT(fields[0] >= 0 && fields[0] <= MAX_OFFSET_HOUR);
    U_ASSERT(fields[1] >= 0 && fields[1] <= MAX_OFFSET_MINUTE);
    U_ASSERT(fields[2] >= 0 && fields[2] <= MAX_OFFSET_SECOND);

    int32_t lastIdx = maxFields;
    while (lastIdx > minFields) {
        if (fields[lastIdx] != 0) {
            break;
        }
        lastIdx--;
    }

    for (int32_t idx = 0; idx <= lastIdx; idx++) {
        if (sep && idx != 0) {
            result.append(sep);
        }
        result.append((UChar)(0x0030 + fields[idx]/10));
        result.append((UChar)(0x0030 + fields[idx]%10));
    }

    return result;
}

int32_t
TimeZoneFormat::parseAbuttingAsciiOffsetFields(const UnicodeString& text, ParsePosition& pos, OffsetFields minFields, OffsetFields maxFields, UBool fixedHourWidth) {
    int32_t start = pos.getIndex();

    int32_t minDigits = 2 * (minFields + 1) - (fixedHourWidth ? 0 : 1);
    int32_t maxDigits = 2 * (maxFields + 1);

    U_ASSERT(maxDigits <= MAX_OFFSET_DIGITS);

    int32_t digits[MAX_OFFSET_DIGITS] = {};
    int32_t numDigits = 0;
    int32_t idx = start;
    while (numDigits < maxDigits && idx < text.length()) {
        UChar uch = text.charAt(idx);
        int32_t digit = DIGIT_VAL(uch);
        if (digit < 0) {
            break;
        }
        digits[numDigits] = digit;
        numDigits++;
        idx++;
    }

    if (fixedHourWidth && (numDigits & 1)) {
        // Fixed digits, so the number of digits must be even number. Truncating.
        numDigits--;
    }

    if (numDigits < minDigits) {
        pos.setErrorIndex(start);
        return 0;
    }

    int32_t hour = 0, min = 0, sec = 0;
    UBool bParsed = FALSE;
    while (numDigits >= minDigits) {
        switch (numDigits) {
        case 1: //H
            hour = digits[0];
            break;
        case 2: //HH
            hour = digits[0] * 10 + digits[1];
            break;
        case 3: //Hmm
            hour = digits[0];
            min = digits[1] * 10 + digits[2];
            break;
        case 4: //HHmm
            hour = digits[0] * 10 + digits[1];
            min = digits[2] * 10 + digits[3];
            break;
        case 5: //Hmmss
            hour = digits[0];
            min = digits[1] * 10 + digits[2];
            sec = digits[3] * 10 + digits[4];
            break;
        case 6: //HHmmss
            hour = digits[0] * 10 + digits[1];
            min = digits[2] * 10 + digits[3];
            sec = digits[4] * 10 + digits[5];
            break;
        }

        if (hour <= MAX_OFFSET_HOUR && min <= MAX_OFFSET_MINUTE && sec <= MAX_OFFSET_SECOND) {
            // Successfully parsed
            bParsed = true;
            break;
        }

        // Truncating
        numDigits -= (fixedHourWidth ? 2 : 1);
        hour = min = sec = 0;
    }

    if (!bParsed) {
        pos.setErrorIndex(start);
        return 0;
    }
    pos.setIndex(start + numDigits);
    return ((((hour * 60) + min) * 60) + sec) * 1000;
}

int32_t
TimeZoneFormat::parseAsciiOffsetFields(const UnicodeString& text, ParsePosition& pos, UChar sep, OffsetFields minFields, OffsetFields maxFields) {
    int32_t start = pos.getIndex();
    int32_t fieldVal[] = {0, 0, 0};
    int32_t fieldLen[] = {0, -1, -1};
    for (int32_t idx = start, fieldIdx = 0; idx < text.length() && fieldIdx <= maxFields; idx++) {
        UChar c = text.charAt(idx);
        if (c == sep) {
            if (fieldIdx == 0) {
                if (fieldLen[0] == 0) {
                    // no hours field
                    break;
                }
                // 1 digit hour, move to next field
            } else {
                if (fieldLen[fieldIdx] != -1) {
                    // premature minute or seconds field
                    break;
                }
                fieldLen[fieldIdx] = 0;
            }
            continue;
        } else if (fieldLen[fieldIdx] == -1) {
            // no separator after 2 digit field
            break;
        }
        int32_t digit = DIGIT_VAL(c);
        if (digit < 0) {
            // not a digit
            break;
        }
        fieldVal[fieldIdx] = fieldVal[fieldIdx] * 10 + digit;
        fieldLen[fieldIdx]++;
        if (fieldLen[fieldIdx] >= 2) {
            // parsed 2 digits, move to next field
            fieldIdx++;
        }
    }

    int32_t offset = 0;
    int32_t parsedLen = 0;
    int32_t parsedFields = -1;
    do {
        // hour
        if (fieldLen[0] == 0) {
            break;
        }
        if (fieldVal[0] > MAX_OFFSET_HOUR) {
            offset = (fieldVal[0] / 10) * MILLIS_PER_HOUR;
            parsedFields = FIELDS_H;
            parsedLen = 1;
            break;
        }
        offset = fieldVal[0] * MILLIS_PER_HOUR;
        parsedLen = fieldLen[0];
        parsedFields = FIELDS_H;

        // minute
        if (fieldLen[1] != 2 || fieldVal[1] > MAX_OFFSET_MINUTE) {
            break;
        }
        offset += fieldVal[1] * MILLIS_PER_MINUTE;
        parsedLen += (1 + fieldLen[1]);
        parsedFields = FIELDS_HM;

        // second
        if (fieldLen[2] != 2 || fieldVal[2] > MAX_OFFSET_SECOND) {
            break;
        }
        offset += fieldVal[2] * MILLIS_PER_SECOND;
        parsedLen += (1 + fieldLen[2]);
        parsedFields = FIELDS_HMS;
    } while (false);

    if (parsedFields < minFields) {
        pos.setErrorIndex(start);
        return 0;
    }

    pos.setIndex(start + parsedLen);
    return offset;
}

void
TimeZoneFormat::appendOffsetDigits(UnicodeString& buf, int32_t n, uint8_t minDigits) const {
    U_ASSERT(n >= 0 && n < 60);
    int32_t numDigits = n >= 10 ? 2 : 1;
    for (int32_t i = 0; i < minDigits - numDigits; i++) {
        buf.append(fGMTOffsetDigits[0]);
    }
    if (numDigits == 2) {
        buf.append(fGMTOffsetDigits[n / 10]);
    }
    buf.append(fGMTOffsetDigits[n % 10]);
}

// ------------------------------------------------------------------
// Private misc
void
TimeZoneFormat::initGMTPattern(const UnicodeString& gmtPattern, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    // This implementation not perfect, but sufficient practically.
    int32_t idx = gmtPattern.indexOf(ARG0, ARG0_LEN, 0);
    if (idx < 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    fGMTPattern.setTo(gmtPattern);
    unquote(gmtPattern.tempSubString(0, idx), fGMTPatternPrefix);
    unquote(gmtPattern.tempSubString(idx + ARG0_LEN), fGMTPatternSuffix);
}

UnicodeString&
TimeZoneFormat::unquote(const UnicodeString& pattern, UnicodeString& result) {
    if (pattern.indexOf(SINGLEQUOTE) < 0) {
        result.setTo(pattern);
        return result;
    }
    result.remove();
    UBool isPrevQuote = FALSE;
    UBool inQuote = FALSE;
    for (int32_t i = 0; i < pattern.length(); i++) {
        UChar c = pattern.charAt(i);
        if (c == SINGLEQUOTE) {
            if (isPrevQuote) {
                result.append(c);
                isPrevQuote = FALSE;
            } else {
                isPrevQuote = TRUE;
            }
            inQuote = !inQuote;
        } else {
            isPrevQuote = FALSE;
            result.append(c);
        }
    }
    return result;
}

UVector*
TimeZoneFormat::parseOffsetPattern(const UnicodeString& pattern, OffsetFields required, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    UVector* result = new UVector(deleteGMTOffsetField, NULL, status);
    if (result == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    int32_t checkBits = 0;
    UBool isPrevQuote = FALSE;
    UBool inQuote = FALSE;
    UChar textBuf[32];
    UnicodeString text(textBuf, 0, UPRV_LENGTHOF(textBuf));
    GMTOffsetField::FieldType itemType = GMTOffsetField::TEXT;
    int32_t itemLength = 1;

    for (int32_t i = 0; i < pattern.length(); i++) {
        UChar ch = pattern.charAt(i);
        if (ch == SINGLEQUOTE) {
            if (isPrevQuote) {
                text.append(SINGLEQUOTE);
                isPrevQuote = FALSE;
            } else {
                isPrevQuote = TRUE;
                if (itemType != GMTOffsetField::TEXT) {
                    if (GMTOffsetField::isValid(itemType, itemLength)) {
                        GMTOffsetField* fld = GMTOffsetField::createTimeField(itemType, static_cast<uint8_t>(itemLength), status);
                        result->addElement(fld, status);
                        if (U_FAILURE(status)) {
                            break;
                        }
                    } else {
                        status = U_ILLEGAL_ARGUMENT_ERROR;
                        break;
                    }
                    itemType = GMTOffsetField::TEXT;
                }
            }
            inQuote = !inQuote;
        } else {
            isPrevQuote = FALSE;
            if (inQuote) {
                text.append(ch);
            } else {
                GMTOffsetField::FieldType tmpType = GMTOffsetField::getTypeByLetter(ch);
                if (tmpType != GMTOffsetField::TEXT) {
                    // an offset time pattern character
                    if (tmpType == itemType) {
                        itemLength++;
                    } else {
                        if (itemType == GMTOffsetField::TEXT) {
                            if (text.length() > 0) {
                                GMTOffsetField* textfld = GMTOffsetField::createText(text, status);
                                result->addElement(textfld, status);
                                if (U_FAILURE(status)) {
                                    break;
                                }
                                text.remove();
                            }
                        } else {
                            if (GMTOffsetField::isValid(itemType, itemLength)) {
                                GMTOffsetField* fld = GMTOffsetField::createTimeField(itemType, static_cast<uint8_t>(itemLength), status);
                                result->addElement(fld, status);
                                if (U_FAILURE(status)) {
                                    break;
                                }
                            } else {
                                status = U_ILLEGAL_ARGUMENT_ERROR;
                                break;
                            }
                        }
                        itemType = tmpType;
                        itemLength = 1;
                        checkBits |= tmpType;
                    }
                } else {
                    // a string literal
                    if (itemType != GMTOffsetField::TEXT) {
                        if (GMTOffsetField::isValid(itemType, itemLength)) {
                            GMTOffsetField* fld = GMTOffsetField::createTimeField(itemType, static_cast<uint8_t>(itemLength), status);
                            result->addElement(fld, status);
                            if (U_FAILURE(status)) {
                                break;
                            }
                        } else {
                            status = U_ILLEGAL_ARGUMENT_ERROR;
                            break;
                        }
                        itemType = GMTOffsetField::TEXT;
                    }
                    text.append(ch);
                }
            }
        }
    }
    // handle last item
    if (U_SUCCESS(status)) {
        if (itemType == GMTOffsetField::TEXT) {
            if (text.length() > 0) {
                GMTOffsetField* tfld = GMTOffsetField::createText(text, status);
                result->addElement(tfld, status);
            }
        } else {
            if (GMTOffsetField::isValid(itemType, itemLength)) {
                GMTOffsetField* fld = GMTOffsetField::createTimeField(itemType, static_cast<uint8_t>(itemLength), status);
                result->addElement(fld, status);
            } else {
                status = U_ILLEGAL_ARGUMENT_ERROR;
            }
        }

        // Check all required fields are set
        if (U_SUCCESS(status)) {
            int32_t reqBits = 0;
            switch (required) {
            case FIELDS_H:
                reqBits = GMTOffsetField::HOUR;
                break;
            case FIELDS_HM:
                reqBits = GMTOffsetField::HOUR | GMTOffsetField::MINUTE;
                break;
            case FIELDS_HMS:
                reqBits = GMTOffsetField::HOUR | GMTOffsetField::MINUTE | GMTOffsetField::SECOND;
                break;
            }
            if (checkBits == reqBits) {
                // all required fields are set, no extra fields
                return result;
            }
        }
    }

    // error
    delete result;
    return NULL;
}

UnicodeString&
TimeZoneFormat::expandOffsetPattern(const UnicodeString& offsetHM, UnicodeString& result, UErrorCode& status) {
    result.setToBogus();
    if (U_FAILURE(status)) {
        return result;
    }
    U_ASSERT(u_strlen(DEFAULT_GMT_OFFSET_MINUTE_PATTERN) == 2);

    int32_t idx_mm = offsetHM.indexOf(DEFAULT_GMT_OFFSET_MINUTE_PATTERN, 2, 0);
    if (idx_mm < 0) {
        // Bad time zone hour pattern data
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return result;
    }

    UnicodeString sep;
    int32_t idx_H = offsetHM.tempSubString(0, idx_mm).lastIndexOf((UChar)0x0048 /* H */);
    if (idx_H >= 0) {
        sep = offsetHM.tempSubString(idx_H + 1, idx_mm - (idx_H + 1));
    }
    result.setTo(offsetHM.tempSubString(0, idx_mm + 2));
    result.append(sep);
    result.append(DEFAULT_GMT_OFFSET_SECOND_PATTERN, -1);
    result.append(offsetHM.tempSubString(idx_mm + 2));
    return result;
}

UnicodeString&
TimeZoneFormat::truncateOffsetPattern(const UnicodeString& offsetHM, UnicodeString& result, UErrorCode& status) {
    result.setToBogus();
    if (U_FAILURE(status)) {
        return result;
    }
    U_ASSERT(u_strlen(DEFAULT_GMT_OFFSET_MINUTE_PATTERN) == 2);

    int32_t idx_mm = offsetHM.indexOf(DEFAULT_GMT_OFFSET_MINUTE_PATTERN, 2, 0);
    if (idx_mm < 0) {
        // Bad time zone hour pattern data
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return result;
    }
    UChar HH[] = {0x0048, 0x0048};
    int32_t idx_HH = offsetHM.tempSubString(0, idx_mm).lastIndexOf(HH, 2, 0);
    if (idx_HH >= 0) {
        return result.setTo(offsetHM.tempSubString(0, idx_HH + 2));
    }
    int32_t idx_H = offsetHM.tempSubString(0, idx_mm).lastIndexOf((UChar)0x0048, 0);
    if (idx_H >= 0) {
        return result.setTo(offsetHM.tempSubString(0, idx_H + 1));
    }
    // Bad time zone hour pattern data
    status = U_ILLEGAL_ARGUMENT_ERROR;
    return result;
}

void
TimeZoneFormat::initGMTOffsetPatterns(UErrorCode& status) {
    for (int32_t type = 0; type < UTZFMT_PAT_COUNT; type++) {
        switch (type) {
        case UTZFMT_PAT_POSITIVE_H:
        case UTZFMT_PAT_NEGATIVE_H:
            fGMTOffsetPatternItems[type] = parseOffsetPattern(fGMTOffsetPatterns[type], FIELDS_H, status);
            break;
        case UTZFMT_PAT_POSITIVE_HM:
        case UTZFMT_PAT_NEGATIVE_HM:
            fGMTOffsetPatternItems[type] = parseOffsetPattern(fGMTOffsetPatterns[type], FIELDS_HM, status);
            break;
        case UTZFMT_PAT_POSITIVE_HMS:
        case UTZFMT_PAT_NEGATIVE_HMS:
            fGMTOffsetPatternItems[type] = parseOffsetPattern(fGMTOffsetPatterns[type], FIELDS_HMS, status);
            break;
        }
    }
    checkAbuttingHoursAndMinutes();
}

void
TimeZoneFormat::checkAbuttingHoursAndMinutes() {
    fAbuttingOffsetHoursAndMinutes= FALSE;
    for (int32_t type = 0; type < UTZFMT_PAT_COUNT; type++) {
        UBool afterH = FALSE;
        UVector *items = fGMTOffsetPatternItems[type];
        for (int32_t i = 0; i < items->size(); i++) {
            const GMTOffsetField* item = (GMTOffsetField*)items->elementAt(i);
            GMTOffsetField::FieldType fieldType = item->getType();
            if (fieldType != GMTOffsetField::TEXT) {
                if (afterH) {
                    fAbuttingOffsetHoursAndMinutes = TRUE;
                    break;
                } else if (fieldType == GMTOffsetField::HOUR) {
                    afterH = TRUE;
                }
            } else if (afterH) {
                break;
            }
        }
        if (fAbuttingOffsetHoursAndMinutes) {
            break;
        }
    }
}

UBool
TimeZoneFormat::toCodePoints(const UnicodeString& str, UChar32* codeArray, int32_t size) {
    int32_t count = str.countChar32();
    if (count != size) {
        return FALSE;
    }

    for (int32_t idx = 0, start = 0; idx < size; idx++) {
        codeArray[idx] = str.char32At(start);
        start = str.moveIndex32(start, 1);
    }

    return TRUE;
}

TimeZone*
TimeZoneFormat::createTimeZoneForOffset(int32_t offset) const {
    if (offset == 0) {
        // when offset is 0, we should use "Etc/GMT"
        return TimeZone::createTimeZone(UnicodeString(TRUE, TZID_GMT, -1));
    }
    return ZoneMeta::createCustomTimeZone(offset);
}

UTimeZoneFormatTimeType
TimeZoneFormat::getTimeType(UTimeZoneNameType nameType) {
    switch (nameType) {
    case UTZNM_LONG_STANDARD:
    case UTZNM_SHORT_STANDARD:
        return UTZFMT_TIME_TYPE_STANDARD;

    case UTZNM_LONG_DAYLIGHT:
    case UTZNM_SHORT_DAYLIGHT:
        return UTZFMT_TIME_TYPE_DAYLIGHT;

    default:
        return UTZFMT_TIME_TYPE_UNKNOWN;
    }
}

UnicodeString&
TimeZoneFormat::getTimeZoneID(const TimeZoneNames::MatchInfoCollection* matches, int32_t idx, UnicodeString& tzID) const {
    if (!matches->getTimeZoneIDAt(idx, tzID)) {
        UChar mzIDBuf[32];
        UnicodeString mzID(mzIDBuf, 0, UPRV_LENGTHOF(mzIDBuf));
        if (matches->getMetaZoneIDAt(idx, mzID)) {
            fTimeZoneNames->getReferenceZoneID(mzID, fTargetRegion, tzID);
        }
    }
    return tzID;
}


class ZoneIdMatchHandler : public TextTrieMapSearchResultHandler {
public:
    ZoneIdMatchHandler();
    virtual ~ZoneIdMatchHandler();

    UBool handleMatch(int32_t matchLength, const CharacterNode *node, UErrorCode &status);
    const UChar* getID();
    int32_t getMatchLen();
private:
    int32_t fLen;
    const UChar* fID;
};

ZoneIdMatchHandler::ZoneIdMatchHandler()
: fLen(0), fID(NULL) {
}

ZoneIdMatchHandler::~ZoneIdMatchHandler() {
}

UBool
ZoneIdMatchHandler::handleMatch(int32_t matchLength, const CharacterNode *node, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (node->hasValues()) {
        const UChar* id = (const UChar*)node->getValue(0);
        if (id != NULL) {
            if (fLen < matchLength) {
                fID = id;
                fLen = matchLength;
            }
        }
    }
    return TRUE;
}

const UChar*
ZoneIdMatchHandler::getID() {
    return fID;
}

int32_t
ZoneIdMatchHandler::getMatchLen() {
    return fLen;
}


static void U_CALLCONV initZoneIdTrie(UErrorCode &status) {
    U_ASSERT(gZoneIdTrie == NULL);
    ucln_i18n_registerCleanup(UCLN_I18N_TIMEZONEFORMAT, tzfmt_cleanup);
    gZoneIdTrie = new TextTrieMap(TRUE, NULL);    // No deleter, because values are pooled by ZoneMeta
    if (gZoneIdTrie == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    StringEnumeration *tzenum = TimeZone::createEnumeration();
    const UnicodeString *id;
    while ((id = tzenum->snext(status)) != NULL) {
        const UChar* uid = ZoneMeta::findTimeZoneID(*id);
        if (uid) {
            gZoneIdTrie->put(uid, const_cast<UChar *>(uid), status);
        }
    }
    delete tzenum;
}


UnicodeString&
TimeZoneFormat::parseZoneID(const UnicodeString& text, ParsePosition& pos, UnicodeString& tzID) const {
    UErrorCode status = U_ZERO_ERROR;
    umtx_initOnce(gZoneIdTrieInitOnce, &initZoneIdTrie, status);

    int32_t start = pos.getIndex();
    int32_t len = 0;
    tzID.setToBogus();

    if (U_SUCCESS(status)) {
        LocalPointer<ZoneIdMatchHandler> handler(new ZoneIdMatchHandler());
        gZoneIdTrie->search(text, start, handler.getAlias(), status);
        len = handler->getMatchLen();
        if (len > 0) {
            tzID.setTo(handler->getID(), -1);
        }
    }

    if (len > 0) {
        pos.setIndex(start + len);
    } else {
        pos.setErrorIndex(start);
    }

    return tzID;
}

static void U_CALLCONV initShortZoneIdTrie(UErrorCode &status) {
    U_ASSERT(gShortZoneIdTrie == NULL);
    ucln_i18n_registerCleanup(UCLN_I18N_TIMEZONEFORMAT, tzfmt_cleanup);
    StringEnumeration *tzenum = TimeZone::createTimeZoneIDEnumeration(UCAL_ZONE_TYPE_CANONICAL, NULL, NULL, status);
    if (U_SUCCESS(status)) {
        gShortZoneIdTrie = new TextTrieMap(TRUE, NULL);    // No deleter, because values are pooled by ZoneMeta
        if (gShortZoneIdTrie == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
        } else {
            const UnicodeString *id;
            while ((id = tzenum->snext(status)) != NULL) {
                const UChar* uID = ZoneMeta::findTimeZoneID(*id);
                const UChar* shortID = ZoneMeta::getShortID(*id);
                if (shortID && uID) {
                    gShortZoneIdTrie->put(shortID, const_cast<UChar *>(uID), status);
                }
            }
        }
    }
    delete tzenum;
}


UnicodeString&
TimeZoneFormat::parseShortZoneID(const UnicodeString& text, ParsePosition& pos, UnicodeString& tzID) const {
    UErrorCode status = U_ZERO_ERROR;
    umtx_initOnce(gShortZoneIdTrieInitOnce, &initShortZoneIdTrie, status);

    int32_t start = pos.getIndex();
    int32_t len = 0;
    tzID.setToBogus();

    if (U_SUCCESS(status)) {
        LocalPointer<ZoneIdMatchHandler> handler(new ZoneIdMatchHandler());
        gShortZoneIdTrie->search(text, start, handler.getAlias(), status);
        len = handler->getMatchLen();
        if (len > 0) {
            tzID.setTo(handler->getID(), -1);
        }
    }

    if (len > 0) {
        pos.setIndex(start + len);
    } else {
        pos.setErrorIndex(start);
    }

    return tzID;
}


UnicodeString&
TimeZoneFormat::parseExemplarLocation(const UnicodeString& text, ParsePosition& pos, UnicodeString& tzID) const {
    int32_t startIdx = pos.getIndex();
    int32_t parsedPos = -1;
    tzID.setToBogus();

    UErrorCode status = U_ZERO_ERROR;
    LocalPointer<TimeZoneNames::MatchInfoCollection> exemplarMatches(fTimeZoneNames->find(text, startIdx, UTZNM_EXEMPLAR_LOCATION, status));
    if (U_FAILURE(status)) {
        pos.setErrorIndex(startIdx);
        return tzID;
    }
    int32_t matchIdx = -1;
    if (!exemplarMatches.isNull()) {
        for (int32_t i = 0; i < exemplarMatches->size(); i++) {
            if (startIdx + exemplarMatches->getMatchLengthAt(i) > parsedPos) {
                matchIdx = i;
                parsedPos = startIdx + exemplarMatches->getMatchLengthAt(i);
            }
        }
        if (parsedPos > 0) {
            pos.setIndex(parsedPos);
            getTimeZoneID(exemplarMatches.getAlias(), matchIdx, tzID);
        }
    }

    if (tzID.length() == 0) {
        pos.setErrorIndex(startIdx);
    }

    return tzID;
}

U_NAMESPACE_END

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2011-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING



























U_NAMESPACE_BEGIN

#define ZID_KEY_MAX  128

static const char tzgnames_gZoneStrings[]                = "zoneStrings";

static const char tzgnames_gRegionFormatTag[]            = "regionFormat";
static const char tzgnames_gFallbackFormatTag[]          = "fallbackFormat";

static const UChar tzgnames_gEmpty[]                     = {0x00};

static const UChar tzgnames_gDefRegionPattern[]          = {0x7B, 0x30, 0x7D, 0x00}; // "{0}"
static const UChar tzgnames_gDefFallbackPattern[]        = {0x7B, 0x31, 0x7D, 0x20, 0x28, 0x7B, 0x30, 0x7D, 0x29, 0x00}; // "{1} ({0})"

static const double tzgnames_kDstCheckRange      = (double)184*U_MILLIS_PER_DAY;



U_CDECL_BEGIN

typedef struct PartialLocationKey {
    const UChar* tzID;
    const UChar* mzID;
    UBool isLong;
} PartialLocationKey;

/**
 * Hash function for partial location name hash key
 */
static int32_t U_CALLCONV
hashPartialLocationKey(const UHashTok key) {
    // <tzID>&<mzID>#[L|S]
    PartialLocationKey *p = (PartialLocationKey *)key.pointer;
    UnicodeString str(p->tzID);
    str.append((UChar)0x26)
        .append(p->mzID, -1)
        .append((UChar)0x23)
        .append((UChar)(p->isLong ? 0x4C : 0x53));
    return str.hashCode();
}

/**
 * Comparer for partial location name hash key
 */
static UBool U_CALLCONV
comparePartialLocationKey(const UHashTok key1, const UHashTok key2) {
    PartialLocationKey *p1 = (PartialLocationKey *)key1.pointer;
    PartialLocationKey *p2 = (PartialLocationKey *)key2.pointer;

    if (p1 == p2) {
        return TRUE;
    }
    if (p1 == NULL || p2 == NULL) {
        return FALSE;
    }
    // We just check identity of tzID/mzID
    return (p1->tzID == p2->tzID && p1->mzID == p2->mzID && p1->isLong == p2->isLong);
}

/**
 * Deleter for GNameInfo
 */
static void U_CALLCONV
deleteGNameInfo(void *obj) {
    uprv_free(obj);
}

/**
 * GNameInfo stores zone name information in the local trie
 */
typedef struct GNameInfo {
    UTimeZoneGenericNameType    type;
    const UChar*                tzID;
} tzgnames_ZNameInfo;

/**
 * GMatchInfo stores zone name match information used by find method
 */
typedef struct GMatchInfo {
    const GNameInfo*    gnameInfo;
    int32_t             matchLength;
    UTimeZoneFormatTimeType   timeType;
} tzgnames_ZMatchInfo;

U_CDECL_END

// ---------------------------------------------------
// The class stores time zone generic name match information
// ---------------------------------------------------
class TimeZoneGenericNameMatchInfo : public UMemory {
public:
    TimeZoneGenericNameMatchInfo(UVector* matches);
    ~TimeZoneGenericNameMatchInfo();

    int32_t size() const;
    UTimeZoneGenericNameType getGenericNameType(int32_t index) const;
    int32_t getMatchLength(int32_t index) const;
    UnicodeString& getTimeZoneID(int32_t index, UnicodeString& tzID) const;

private:
    UVector* fMatches;  // vector of MatchEntry
};

TimeZoneGenericNameMatchInfo::TimeZoneGenericNameMatchInfo(UVector* matches)
: fMatches(matches) {
}

TimeZoneGenericNameMatchInfo::~TimeZoneGenericNameMatchInfo() {
    if (fMatches != NULL) {
        delete fMatches;
    }
}

int32_t
TimeZoneGenericNameMatchInfo::size() const {
    if (fMatches == NULL) {
        return 0;
    }
    return fMatches->size();
}

UTimeZoneGenericNameType
TimeZoneGenericNameMatchInfo::getGenericNameType(int32_t index) const {
    GMatchInfo *minfo = (GMatchInfo *)fMatches->elementAt(index);
    if (minfo != NULL) {
        return static_cast<UTimeZoneGenericNameType>(minfo->gnameInfo->type);
    }
    return UTZGNM_UNKNOWN;
}

int32_t
TimeZoneGenericNameMatchInfo::getMatchLength(int32_t index) const {
    tzgnames_ZMatchInfo *minfo = (tzgnames_ZMatchInfo *)fMatches->elementAt(index);
    if (minfo != NULL) {
        return minfo->matchLength;
    }
    return -1;
}

UnicodeString&
TimeZoneGenericNameMatchInfo::getTimeZoneID(int32_t index, UnicodeString& tzID) const {
    GMatchInfo *minfo = (GMatchInfo *)fMatches->elementAt(index);
    if (minfo != NULL && minfo->gnameInfo->tzID != NULL) {
        tzID.setTo(TRUE, minfo->gnameInfo->tzID, -1);
    } else {
        tzID.setToBogus();
    }
    return tzID;
}

// ---------------------------------------------------
// GNameSearchHandler
// ---------------------------------------------------
class GNameSearchHandler : public TextTrieMapSearchResultHandler {
public:
    GNameSearchHandler(uint32_t types);
    virtual ~GNameSearchHandler();

    UBool handleMatch(int32_t matchLength, const CharacterNode *node, UErrorCode &status);
    UVector* getMatches(int32_t& maxMatchLen);

private:
    uint32_t fTypes;
    UVector* fResults;
    int32_t fMaxMatchLen;
};

GNameSearchHandler::GNameSearchHandler(uint32_t types)
: fTypes(types), fResults(NULL), fMaxMatchLen(0) {
}

GNameSearchHandler::~GNameSearchHandler() {
    if (fResults != NULL) {
        delete fResults;
    }
}

UBool
GNameSearchHandler::handleMatch(int32_t matchLength, const CharacterNode *node, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (node->hasValues()) {
        int32_t valuesCount = node->countValues();
        for (int32_t i = 0; i < valuesCount; i++) {
            GNameInfo *nameinfo = (tzgnames_ZNameInfo *)node->getValue(i);
            if (nameinfo == NULL) {
                break;
            }
            if ((nameinfo->type & fTypes) != 0) {
                // matches a requested type
                if (fResults == NULL) {
                    fResults = new UVector(uprv_free, NULL, status);
                    if (fResults == NULL) {
                        status = U_MEMORY_ALLOCATION_ERROR;
                    }
                }
                if (U_SUCCESS(status)) {
                    U_ASSERT(fResults != NULL);
                    GMatchInfo *gmatch = (GMatchInfo *)uprv_malloc(sizeof(GMatchInfo));
                    if (gmatch == NULL) {
                        status = U_MEMORY_ALLOCATION_ERROR;
                    } else {
                        // add the match to the vector
                        gmatch->gnameInfo = nameinfo;
                        gmatch->matchLength = matchLength;
                        gmatch->timeType = UTZFMT_TIME_TYPE_UNKNOWN;
                        fResults->addElement(gmatch, status);
                        if (U_FAILURE(status)) {
                            uprv_free(gmatch);
                        } else {
                            if (matchLength > fMaxMatchLen) {
                                fMaxMatchLen = matchLength;
                            }
                        }
                    }
                }
            }
        }
    }
    return TRUE;
}

UVector*
GNameSearchHandler::getMatches(int32_t& maxMatchLen) {
    // give the ownership to the caller
    UVector *results = fResults;
    maxMatchLen = fMaxMatchLen;

    // reset
    fResults = NULL;
    fMaxMatchLen = 0;
    return results;
}

static UMutex tzgnames_gLock;

class TZGNCore : public UMemory {
public:
    TZGNCore(const Locale& locale, UErrorCode& status);
    virtual ~TZGNCore();

    UnicodeString& getDisplayName(const TimeZone& tz, UTimeZoneGenericNameType type,
                        UDate date, UnicodeString& name) const;

    UnicodeString& getGenericLocationName(const UnicodeString& tzCanonicalID, UnicodeString& name) const;

    int32_t findBestMatch(const UnicodeString& text, int32_t start, uint32_t types,
        UnicodeString& tzID, UTimeZoneFormatTimeType& timeType, UErrorCode& status) const;

private:
    Locale fLocale;
    const TimeZoneNames* fTimeZoneNames;
    UHashtable* fLocationNamesMap;
    UHashtable* fPartialLocationNamesMap;

    SimpleFormatter fRegionFormat;
    SimpleFormatter fFallbackFormat;

    LocaleDisplayNames* fLocaleDisplayNames;
    ZNStringPool fStringPool;

    TextTrieMap fGNamesTrie;
    UBool fGNamesTrieFullyLoaded;

    char fTargetRegion[ULOC_COUNTRY_CAPACITY];

    void initialize(const Locale& locale, UErrorCode& status);
    void cleanup();

    void loadStrings(const UnicodeString& tzCanonicalID);

    const UChar* getGenericLocationName(const UnicodeString& tzCanonicalID);

    UnicodeString& formatGenericNonLocationName(const TimeZone& tz, UTimeZoneGenericNameType type,
                        UDate date, UnicodeString& name) const;

    UnicodeString& getPartialLocationName(const UnicodeString& tzCanonicalID,
                        const UnicodeString& mzID, UBool isLong, const UnicodeString& mzDisplayName,
                        UnicodeString& name) const;

    const UChar* getPartialLocationName(const UnicodeString& tzCanonicalID,
                        const UnicodeString& mzID, UBool isLong, const UnicodeString& mzDisplayName);

    TimeZoneGenericNameMatchInfo* findLocal(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const;

    TimeZoneNames::MatchInfoCollection* findTimeZoneNames(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const;
};


// ---------------------------------------------------
// TZGNCore - core implmentation of TimeZoneGenericNames
//
// TimeZoneGenericNames is parallel to TimeZoneNames,
// but handles run-time generated time zone names.
// This is the main part of this module.
// ---------------------------------------------------
TZGNCore::TZGNCore(const Locale& locale, UErrorCode& status)
: fLocale(locale),
  fTimeZoneNames(NULL),
  fLocationNamesMap(NULL),
  fPartialLocationNamesMap(NULL),
  fLocaleDisplayNames(NULL),
  fStringPool(status),
  fGNamesTrie(TRUE, deleteGNameInfo),
  fGNamesTrieFullyLoaded(FALSE) {
    initialize(locale, status);
}

TZGNCore::~TZGNCore() {
    cleanup();
}

void
TZGNCore::initialize(const Locale& locale, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }

    // TimeZoneNames
    fTimeZoneNames = TimeZoneNames::createInstance(locale, status);
    if (U_FAILURE(status)) {
        return;
    }

    // Initialize format patterns
    UnicodeString rpat(TRUE, tzgnames_gDefRegionPattern, -1);
    UnicodeString fpat(TRUE, tzgnames_gDefFallbackPattern, -1);

    UErrorCode tmpsts = U_ZERO_ERROR;   // OK with fallback warning..
    UResourceBundle *zoneStrings = ures_open(U_ICUDATA_ZONE, locale.getName(), &tmpsts);
    zoneStrings = ures_getByKeyWithFallback(zoneStrings, tzgnames_gZoneStrings, zoneStrings, &tmpsts);

    if (U_SUCCESS(tmpsts)) {
        const UChar *regionPattern = ures_getStringByKeyWithFallback(zoneStrings, tzgnames_gRegionFormatTag, NULL, &tmpsts);
        if (U_SUCCESS(tmpsts) && u_strlen(regionPattern) > 0) {
            rpat.setTo(regionPattern, -1);
        }
        tmpsts = U_ZERO_ERROR;
        const UChar *fallbackPattern = ures_getStringByKeyWithFallback(zoneStrings, tzgnames_gFallbackFormatTag, NULL, &tmpsts);
        if (U_SUCCESS(tmpsts) && u_strlen(fallbackPattern) > 0) {
            fpat.setTo(fallbackPattern, -1);
        }
    }
    ures_close(zoneStrings);

    fRegionFormat.applyPatternMinMaxArguments(rpat, 1, 1, status);
    fFallbackFormat.applyPatternMinMaxArguments(fpat, 2, 2, status);
    if (U_FAILURE(status)) {
        cleanup();
        return;
    }

    // locale display names
    fLocaleDisplayNames = LocaleDisplayNames::createInstance(locale);

    // hash table for names - no key/value deleters
    fLocationNamesMap = uhash_open(uhash_hashUChars, uhash_compareUChars, NULL, &status);
    if (U_FAILURE(status)) {
        cleanup();
        return;
    }

    fPartialLocationNamesMap = uhash_open(hashPartialLocationKey, comparePartialLocationKey, NULL, &status);
    if (U_FAILURE(status)) {
        cleanup();
        return;
    }
    uhash_setKeyDeleter(fPartialLocationNamesMap, uprv_free);
    // no value deleter

    // target region
    const char* region = fLocale.getCountry();
    int32_t regionLen = static_cast<int32_t>(uprv_strlen(region));
    if (regionLen == 0) {
        CharString loc;
        {
            CharStringByteSink sink(&loc);
            ulocimp_addLikelySubtags(fLocale.getName(), sink, &status);
        }

        regionLen = uloc_getCountry(loc.data(), fTargetRegion, sizeof(fTargetRegion), &status);
        if (U_SUCCESS(status)) {
            fTargetRegion[regionLen] = 0;
        } else {
            cleanup();
            return;
        }
    } else if (regionLen < (int32_t)sizeof(fTargetRegion)) {
        uprv_strcpy(fTargetRegion, region);
    } else {
        fTargetRegion[0] = 0;
    }

    // preload generic names for the default zone
    TimeZone *tz = TimeZone::createDefault();
    const UChar *tzID = ZoneMeta::getCanonicalCLDRID(*tz);
    if (tzID != NULL) {
        loadStrings(UnicodeString(TRUE, tzID, -1));
    }
    delete tz;
}

void
TZGNCore::cleanup() {
    if (fLocaleDisplayNames != NULL) {
        delete fLocaleDisplayNames;
    }
    if (fTimeZoneNames != NULL) {
        delete fTimeZoneNames;
    }

    uhash_close(fLocationNamesMap);
    uhash_close(fPartialLocationNamesMap);
}


UnicodeString&
TZGNCore::getDisplayName(const TimeZone& tz, UTimeZoneGenericNameType type, UDate date, UnicodeString& name) const {
    name.setToBogus();
    switch (type) {
    case UTZGNM_LOCATION:
        {
            const UChar* tzCanonicalID = ZoneMeta::getCanonicalCLDRID(tz);
            if (tzCanonicalID != NULL) {
                getGenericLocationName(UnicodeString(TRUE, tzCanonicalID, -1), name);
            }
        }
        break;
    case UTZGNM_LONG:
    case UTZGNM_SHORT:
        formatGenericNonLocationName(tz, type, date, name);
        if (name.isEmpty()) {
            const UChar* tzCanonicalID = ZoneMeta::getCanonicalCLDRID(tz);
            if (tzCanonicalID != NULL) {
                getGenericLocationName(UnicodeString(TRUE, tzCanonicalID, -1), name);
            }
        }
        break;
    default:
        break;
    }
    return name;
}

UnicodeString&
TZGNCore::getGenericLocationName(const UnicodeString& tzCanonicalID, UnicodeString& name) const {
    if (tzCanonicalID.isEmpty()) {
        name.setToBogus();
        return name;
    }

    const UChar *locname = NULL;
    TZGNCore *nonConstThis = const_cast<TZGNCore *>(this);
    umtx_lock(&tzgnames_gLock);
    {
        locname = nonConstThis->getGenericLocationName(tzCanonicalID);
    }
    umtx_unlock(&tzgnames_gLock);

    if (locname == NULL) {
        name.setToBogus();
    } else {
        name.setTo(locname, u_strlen(locname));
    }

    return name;
}

/*
 * This method updates the cache and must be called with a lock
 */
const UChar*
TZGNCore::getGenericLocationName(const UnicodeString& tzCanonicalID) {
    U_ASSERT(!tzCanonicalID.isEmpty());
    if (tzCanonicalID.length() > ZID_KEY_MAX) {
        return NULL;
    }

    UErrorCode status = U_ZERO_ERROR;
    UChar tzIDKey[ZID_KEY_MAX + 1];
    int32_t tzIDKeyLen = tzCanonicalID.extract(tzIDKey, ZID_KEY_MAX + 1, status);
    U_ASSERT(status == U_ZERO_ERROR);   // already checked length above
    tzIDKey[tzIDKeyLen] = 0;

    const UChar *locname = (const UChar *)uhash_get(fLocationNamesMap, tzIDKey);

    if (locname != NULL) {
        // tzgnames_gEmpty indicate the name is not available
        if (locname == tzgnames_gEmpty) {
            return NULL;
        }
        return locname;
    }

    // Construct location name
    UnicodeString name;
    UnicodeString usCountryCode;
    UBool isPrimary = FALSE;

    ZoneMeta::getCanonicalCountry(tzCanonicalID, usCountryCode, &isPrimary);

    if (!usCountryCode.isEmpty()) {
        if (isPrimary) {
            // If this is the primary zone in the country, use the country name.
            char countryCode[ULOC_COUNTRY_CAPACITY];
            U_ASSERT(usCountryCode.length() < ULOC_COUNTRY_CAPACITY);
            int32_t ccLen = usCountryCode.extract(0, usCountryCode.length(), countryCode, sizeof(countryCode), US_INV);
            countryCode[ccLen] = 0;

            UnicodeString country;
            fLocaleDisplayNames->regionDisplayName(countryCode, country);
            fRegionFormat.format(country, name, status);
        } else {
            // If this is not the primary zone in the country,
            // use the exemplar city name.

            // getExemplarLocationName should retur non-empty string
            // if the time zone is associated with a region

            UnicodeString city;
            fTimeZoneNames->getExemplarLocationName(tzCanonicalID, city);
            fRegionFormat.format(city, name, status);
        }
        if (U_FAILURE(status)) {
            return NULL;
        }
    }

    locname = name.isEmpty() ? NULL : fStringPool.get(name, status);
    if (U_SUCCESS(status)) {
        // Cache the result
        const UChar* cacheID = ZoneMeta::findTimeZoneID(tzCanonicalID);
        U_ASSERT(cacheID != NULL);
        if (locname == NULL) {
            // tzgnames_gEmpty to indicate - no location name available
            uhash_put(fLocationNamesMap, (void *)cacheID, (void *)tzgnames_gEmpty, &status);
        } else {
            uhash_put(fLocationNamesMap, (void *)cacheID, (void *)locname, &status);
            if (U_FAILURE(status)) {
                locname = NULL;
            } else {
                // put the name info into the trie
                GNameInfo *nameinfo = (tzgnames_ZNameInfo *)uprv_malloc(sizeof(GNameInfo));
                if (nameinfo != NULL) {
                    nameinfo->type = UTZGNM_LOCATION;
                    nameinfo->tzID = cacheID;
                    fGNamesTrie.put(locname, nameinfo, status);
                }
            }
        }
    }

    return locname;
}

UnicodeString&
TZGNCore::formatGenericNonLocationName(const TimeZone& tz, UTimeZoneGenericNameType type, UDate date, UnicodeString& name) const {
    U_ASSERT(type == UTZGNM_LONG || type == UTZGNM_SHORT);
    name.setToBogus();

    const UChar* uID = ZoneMeta::getCanonicalCLDRID(tz);
    if (uID == NULL) {
        return name;
    }

    UnicodeString tzID(TRUE, uID, -1);

    // Try to get a name from time zone first
    UTimeZoneNameType nameType = (type == UTZGNM_LONG) ? UTZNM_LONG_GENERIC : UTZNM_SHORT_GENERIC;
    fTimeZoneNames->getTimeZoneDisplayName(tzID, nameType, name);

    if (!name.isEmpty()) {
        return name;
    }

    // Try meta zone
    UChar mzIDBuf[32];
    UnicodeString mzID(mzIDBuf, 0, UPRV_LENGTHOF(mzIDBuf));
    fTimeZoneNames->getMetaZoneID(tzID, date, mzID);
    if (!mzID.isEmpty()) {
        UErrorCode status = U_ZERO_ERROR;
        UBool useStandard = FALSE;
        int32_t raw, sav;
        UChar tmpNameBuf[ZONE_NAME_U16_MAX];

        tz.getOffset(date, FALSE, raw, sav, status);
        if (U_FAILURE(status)) {
            return name;
        }

        if (sav == 0) {
            useStandard = TRUE;

            TimeZone *tmptz = tz.clone();
            // Check if the zone actually uses daylight saving time around the time
            BasicTimeZone *btz = NULL;
            if (dynamic_cast<OlsonTimeZone *>(tmptz) != NULL
                || dynamic_cast<SimpleTimeZone *>(tmptz) != NULL
                || dynamic_cast<RuleBasedTimeZone *>(tmptz) != NULL
                || dynamic_cast<VTimeZone *>(tmptz) != NULL) {
                btz = (BasicTimeZone*)tmptz;
            }

            if (btz != NULL) {
                TimeZoneTransition before;
                UBool beforTrs = btz->getPreviousTransition(date, TRUE, before);
                if (beforTrs
                        && (date - before.getTime() < tzgnames_kDstCheckRange)
                        && before.getFrom()->getDSTSavings() != 0) {
                    useStandard = FALSE;
                } else {
                    TimeZoneTransition after;
                    UBool afterTrs = btz->getNextTransition(date, FALSE, after);
                    if (afterTrs
                            && (after.getTime() - date < tzgnames_kDstCheckRange)
                            && after.getTo()->getDSTSavings() != 0) {
                        useStandard = FALSE;
                    }
                }
            } else {
                // If not BasicTimeZone... only if the instance is not an ICU's implementation.
                // We may get a wrong answer in edge case, but it should practically work OK.
                tmptz->getOffset(date - tzgnames_kDstCheckRange, FALSE, raw, sav, status);
                if (sav != 0) {
                    useStandard = FALSE;
                } else {
                    tmptz->getOffset(date + tzgnames_kDstCheckRange, FALSE, raw, sav, status);
                    if (sav != 0){
                        useStandard = FALSE;
                    }
                }
                if (U_FAILURE(status)) {
                    delete tmptz;
                    return name;
                }
            }
            delete tmptz;
        }
        if (useStandard) {
            UTimeZoneNameType stdNameType = (nameType == UTZNM_LONG_GENERIC)
                ? UTZNM_LONG_STANDARD : UTZNM_SHORT_STANDARD;
            UnicodeString stdName(tmpNameBuf, 0, UPRV_LENGTHOF(tmpNameBuf));
            fTimeZoneNames->getDisplayName(tzID, stdNameType, date, stdName);
            if (!stdName.isEmpty()) {
                name.setTo(stdName);

                // TODO: revisit this issue later
                // In CLDR, a same display name is used for both generic and standard
                // for some meta zones in some locales.  This looks like a data bugs.
                // For now, we check if the standard name is different from its generic
                // name below.
                UChar genNameBuf[ZONE_NAME_U16_MAX];
                UnicodeString mzGenericName(genNameBuf, 0, UPRV_LENGTHOF(genNameBuf));
                fTimeZoneNames->getMetaZoneDisplayName(mzID, nameType, mzGenericName);
                if (stdName.caseCompare(mzGenericName, 0) == 0) {
                    name.setToBogus();
                }
            }
        }
        if (name.isEmpty()) {
            // Get a name from meta zone
            UnicodeString mzName(tmpNameBuf, 0, UPRV_LENGTHOF(tmpNameBuf));
            fTimeZoneNames->getMetaZoneDisplayName(mzID, nameType, mzName);
            if (!mzName.isEmpty()) {
                // Check if we need to use a partial location format.
                // This check is done by comparing offset with the meta zone's
                // golden zone at the given date.
                UChar idBuf[32];
                UnicodeString goldenID(idBuf, 0, UPRV_LENGTHOF(idBuf));
                fTimeZoneNames->getReferenceZoneID(mzID, fTargetRegion, goldenID);
                if (!goldenID.isEmpty() && goldenID != tzID) {
                    TimeZone *goldenZone = TimeZone::createTimeZone(goldenID);
                    int32_t raw1, sav1;

                    // Check offset in the golden zone with wall time.
                    // With getOffset(date, false, offsets1),
                    // you may get incorrect results because of time overlap at DST->STD
                    // transition.
                    goldenZone->getOffset(date + raw + sav, TRUE, raw1, sav1, status);
                    delete goldenZone;
                    if (U_SUCCESS(status)) {
                        if (raw != raw1 || sav != sav1) {
                            // Now we need to use a partial location format
                            getPartialLocationName(tzID, mzID, (nameType == UTZNM_LONG_GENERIC), mzName, name);
                        } else {
                            name.setTo(mzName);
                        }
                    }
                } else {
                    name.setTo(mzName);
                }
            }
        }
    }
    return name;
}

UnicodeString&
TZGNCore::getPartialLocationName(const UnicodeString& tzCanonicalID,
                        const UnicodeString& mzID, UBool isLong, const UnicodeString& mzDisplayName,
                        UnicodeString& name) const {
    name.setToBogus();
    if (tzCanonicalID.isEmpty() || mzID.isEmpty() || mzDisplayName.isEmpty()) {
        return name;
    }

    const UChar *uplname = NULL;
    TZGNCore *nonConstThis = const_cast<TZGNCore *>(this);
    umtx_lock(&tzgnames_gLock);
    {
        uplname = nonConstThis->getPartialLocationName(tzCanonicalID, mzID, isLong, mzDisplayName);
    }
    umtx_unlock(&tzgnames_gLock);

    if (uplname == NULL) {
        name.setToBogus();
    } else {
        name.setTo(TRUE, uplname, -1);
    }
    return name;
}

/*
 * This method updates the cache and must be called with a lock
 */
const UChar*
TZGNCore::getPartialLocationName(const UnicodeString& tzCanonicalID,
                        const UnicodeString& mzID, UBool isLong, const UnicodeString& mzDisplayName) {
    U_ASSERT(!tzCanonicalID.isEmpty());
    U_ASSERT(!mzID.isEmpty());
    U_ASSERT(!mzDisplayName.isEmpty());

    PartialLocationKey key;
    key.tzID = ZoneMeta::findTimeZoneID(tzCanonicalID);
    key.mzID = ZoneMeta::findMetaZoneID(mzID);
    key.isLong = isLong;
    U_ASSERT(key.tzID != NULL && key.mzID != NULL);

    const UChar* uplname = (const UChar*)uhash_get(fPartialLocationNamesMap, (void *)&key);
    if (uplname != NULL) {
        return uplname;
    }

    UnicodeString location;
    UnicodeString usCountryCode;
    ZoneMeta::getCanonicalCountry(tzCanonicalID, usCountryCode);
    if (!usCountryCode.isEmpty()) {
        char countryCode[ULOC_COUNTRY_CAPACITY];
        U_ASSERT(usCountryCode.length() < ULOC_COUNTRY_CAPACITY);
        int32_t ccLen = usCountryCode.extract(0, usCountryCode.length(), countryCode, sizeof(countryCode), US_INV);
        countryCode[ccLen] = 0;

        UnicodeString regionalGolden;
        fTimeZoneNames->getReferenceZoneID(mzID, countryCode, regionalGolden);
        if (tzCanonicalID == regionalGolden) {
            // Use country name
            fLocaleDisplayNames->regionDisplayName(countryCode, location);
        } else {
            // Otherwise, use exemplar city name
            fTimeZoneNames->getExemplarLocationName(tzCanonicalID, location);
        }
    } else {
        fTimeZoneNames->getExemplarLocationName(tzCanonicalID, location);
        if (location.isEmpty()) {
            // This could happen when the time zone is not associated with a country,
            // and its ID is not hierarchical, for example, CST6CDT.
            // We use the canonical ID itself as the location for this case.
            location.setTo(tzCanonicalID);
        }
    }

    UErrorCode status = U_ZERO_ERROR;
    UnicodeString name;
    fFallbackFormat.format(location, mzDisplayName, name, status);
    if (U_FAILURE(status)) {
        return NULL;
    }

    uplname = fStringPool.get(name, status);
    if (U_SUCCESS(status)) {
        // Add the name to cache
        PartialLocationKey* cacheKey = (PartialLocationKey *)uprv_malloc(sizeof(PartialLocationKey));
        if (cacheKey != NULL) {
            cacheKey->tzID = key.tzID;
            cacheKey->mzID = key.mzID;
            cacheKey->isLong = key.isLong;
            uhash_put(fPartialLocationNamesMap, (void *)cacheKey, (void *)uplname, &status);
            if (U_FAILURE(status)) {
                uprv_free(cacheKey);
            } else {
                // put the name to the local trie as well
                GNameInfo *nameinfo = (tzgnames_ZNameInfo *)uprv_malloc(sizeof(GNameInfo));
                if (nameinfo != NULL) {
                    nameinfo->type = isLong ? UTZGNM_LONG : UTZGNM_SHORT;
                    nameinfo->tzID = key.tzID;
                    fGNamesTrie.put(uplname, nameinfo, status);
                }
            }
        }
    }
    return uplname;
}

/*
 * This method updates the cache and must be called with a lock,
 * except initializer.
 */
void
TZGNCore::loadStrings(const UnicodeString& tzCanonicalID) {
    // load the generic location name
    getGenericLocationName(tzCanonicalID);

    // partial location names
    UErrorCode status = U_ZERO_ERROR;

    const UnicodeString *mzID;
    UnicodeString goldenID;
    UnicodeString mzGenName;
    UTimeZoneNameType genNonLocTypes[] = {
        UTZNM_LONG_GENERIC, UTZNM_SHORT_GENERIC,
        UTZNM_UNKNOWN /*terminator*/
    };

    StringEnumeration *mzIDs = fTimeZoneNames->getAvailableMetaZoneIDs(tzCanonicalID, status);
    while ((mzID = mzIDs->snext(status)) != NULL) {
        if (U_FAILURE(status)) {
            break;
        }
        // if this time zone is not the golden zone of the meta zone,
        // partial location name (such as "PT (Los Angeles)") might be
        // available.
        fTimeZoneNames->getReferenceZoneID(*mzID, fTargetRegion, goldenID);
        if (tzCanonicalID != goldenID) {
            for (int32_t i = 0; genNonLocTypes[i] != UTZNM_UNKNOWN; i++) {
                fTimeZoneNames->getMetaZoneDisplayName(*mzID, genNonLocTypes[i], mzGenName);
                if (!mzGenName.isEmpty()) {
                    // getPartialLocationName formats a name and put it into the trie
                    getPartialLocationName(tzCanonicalID, *mzID,
                        (genNonLocTypes[i] == UTZNM_LONG_GENERIC), mzGenName);
                }
            }
        }
    }
    if (mzIDs != NULL) {
        delete mzIDs;
    }
}

int32_t
TZGNCore::findBestMatch(const UnicodeString& text, int32_t start, uint32_t types,
        UnicodeString& tzID, UTimeZoneFormatTimeType& timeType, UErrorCode& status) const {
    timeType = UTZFMT_TIME_TYPE_UNKNOWN;
    tzID.setToBogus();

    if (U_FAILURE(status)) {
        return 0;
    }

    // Find matches in the TimeZoneNames first
    TimeZoneNames::MatchInfoCollection *tznamesMatches = findTimeZoneNames(text, start, types, status);
    if (U_FAILURE(status)) {
        return 0;
    }

    int32_t bestMatchLen = 0;
    UTimeZoneFormatTimeType bestMatchTimeType = UTZFMT_TIME_TYPE_UNKNOWN;
    UnicodeString bestMatchTzID;
    // UBool isLongStandard = FALSE;   // workaround - see the comments below
    UBool isStandard = FALSE;       // TODO: Temporary hack (on hack) for short standard name/location name conflict (found in zh_Hant), should be removed after CLDR 21m1 integration

    if (tznamesMatches != NULL) {
        UnicodeString mzID;
        for (int32_t i = 0; i < tznamesMatches->size(); i++) {
            int32_t len = tznamesMatches->getMatchLengthAt(i);
            if (len > bestMatchLen) {
                bestMatchLen = len;
                if (!tznamesMatches->getTimeZoneIDAt(i, bestMatchTzID)) {
                    // name for a meta zone
                    if (tznamesMatches->getMetaZoneIDAt(i, mzID)) {
                        fTimeZoneNames->getReferenceZoneID(mzID, fTargetRegion, bestMatchTzID);
                    }
                }
                UTimeZoneNameType nameType = tznamesMatches->getNameTypeAt(i);
                if (U_FAILURE(status)) {
                    break;
                }
                switch (nameType) {
                case UTZNM_LONG_STANDARD:
                    // isLongStandard = TRUE;
                case UTZNM_SHORT_STANDARD:  // this one is never used for generic, but just in case
                    isStandard = TRUE;      // TODO: Remove this later, see the comments above.
                    bestMatchTimeType = UTZFMT_TIME_TYPE_STANDARD;
                    break;
                case UTZNM_LONG_DAYLIGHT:
                case UTZNM_SHORT_DAYLIGHT: // this one is never used for generic, but just in case
                    bestMatchTimeType = UTZFMT_TIME_TYPE_DAYLIGHT;
                    break;
                default:
                    bestMatchTimeType = UTZFMT_TIME_TYPE_UNKNOWN;
                }
            }
        }
        delete tznamesMatches;
        if (U_FAILURE(status)) {
            return 0;
        }

        if (bestMatchLen == (text.length() - start)) {
            // Full match

            //tzID.setTo(bestMatchTzID);
            //timeType = bestMatchTimeType;
            //return bestMatchLen;

            // TODO Some time zone uses a same name for the long standard name
            // and the location name. When the match is a long standard name,
            // then we need to check if the name is same with the location name.
            // This is probably a data error or a design bug.
/*
            if (!isLongStandard) {
                tzID.setTo(bestMatchTzID);
                timeType = bestMatchTimeType;
                return bestMatchLen;
            }
*/
            // TODO The deprecation of commonlyUsed flag introduced the name
            // conflict not only for long standard names, but short standard names too.
            // These short names (found in zh_Hant) should be gone once we clean
            // up CLDR time zone display name data. Once the short name conflict
            // problem (with location name) is resolved, we should change the condition
            // below back to the original one above. -Yoshito (2011-09-14)
            if (!isStandard) {
                tzID.setTo(bestMatchTzID);
                timeType = bestMatchTimeType;
                return bestMatchLen;
            }
        }
    }

    // Find matches in the local trie
    TimeZoneGenericNameMatchInfo *localMatches = findLocal(text, start, types, status);
    if (U_FAILURE(status)) {
        return 0;
    }
    if (localMatches != NULL) {
        for (int32_t i = 0; i < localMatches->size(); i++) {
            int32_t len = localMatches->getMatchLength(i);

            // TODO See the above TODO. We use len >= bestMatchLen
            // because of the long standard/location name collision
            // problem. If it is also a location name, carrying
            // timeType = UTZFMT_TIME_TYPE_STANDARD will cause a
            // problem in SimpleDateFormat
            if (len >= bestMatchLen) {
                bestMatchLen = localMatches->getMatchLength(i);
                bestMatchTimeType = UTZFMT_TIME_TYPE_UNKNOWN;   // because generic
                localMatches->getTimeZoneID(i, bestMatchTzID);
            }
        }
        delete localMatches;
    }

    if (bestMatchLen > 0) {
        timeType = bestMatchTimeType;
        tzID.setTo(bestMatchTzID);
    }
    return bestMatchLen;
}

TimeZoneGenericNameMatchInfo*
TZGNCore::findLocal(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const {
    GNameSearchHandler handler(types);

    TZGNCore *nonConstThis = const_cast<TZGNCore *>(this);

    umtx_lock(&tzgnames_gLock);
    {
        fGNamesTrie.search(text, start, (TextTrieMapSearchResultHandler *)&handler, status);
    }
    umtx_unlock(&tzgnames_gLock);

    if (U_FAILURE(status)) {
        return NULL;
    }

    TimeZoneGenericNameMatchInfo *gmatchInfo = NULL;

    int32_t maxLen = 0;
    UVector *results = handler.getMatches(maxLen);
    if (results != NULL && ((maxLen == (text.length() - start)) || fGNamesTrieFullyLoaded)) {
        // perfect match
        gmatchInfo = new TimeZoneGenericNameMatchInfo(results);
        if (gmatchInfo == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            delete results;
            return NULL;
        }
        return gmatchInfo;
    }

    if (results != NULL) {
        delete results;
    }

    // All names are not yet loaded into the local trie.
    // Load all available names into the trie. This could be very heavy.
    umtx_lock(&tzgnames_gLock);
    {
        if (!fGNamesTrieFullyLoaded) {
            StringEnumeration *tzIDs = TimeZone::createTimeZoneIDEnumeration(UCAL_ZONE_TYPE_CANONICAL, NULL, NULL, status);
            if (U_SUCCESS(status)) {
                const UnicodeString *tzID;
                while ((tzID = tzIDs->snext(status)) != NULL) {
                    if (U_FAILURE(status)) {
                        break;
                    }
                    nonConstThis->loadStrings(*tzID);
                }
            }
            if (tzIDs != NULL) {
                delete tzIDs;
            }

            if (U_SUCCESS(status)) {
                nonConstThis->fGNamesTrieFullyLoaded = TRUE;
            }
        }
    }
    umtx_unlock(&tzgnames_gLock);

    if (U_FAILURE(status)) {
        return NULL;
    }

    umtx_lock(&tzgnames_gLock);
    {
        // now try it again
        fGNamesTrie.search(text, start, (TextTrieMapSearchResultHandler *)&handler, status);
    }
    umtx_unlock(&tzgnames_gLock);

    results = handler.getMatches(maxLen);
    if (results != NULL && maxLen > 0) {
        gmatchInfo = new TimeZoneGenericNameMatchInfo(results);
        if (gmatchInfo == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            delete results;
            return NULL;
        }
    }

    return gmatchInfo;
}

TimeZoneNames::MatchInfoCollection*
TZGNCore::findTimeZoneNames(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const {
    // Check if the target name typs is really in the TimeZoneNames
    uint32_t nameTypes = 0;
    if (types & UTZGNM_LONG) {
        nameTypes |= (UTZNM_LONG_GENERIC | UTZNM_LONG_STANDARD);
    }
    if (types & UTZGNM_SHORT) {
        nameTypes |= (UTZNM_SHORT_GENERIC | UTZNM_SHORT_STANDARD);
    }

    if (types) {
        // Find matches in the TimeZoneNames
        return fTimeZoneNames->find(text, start, nameTypes, status);
    }

    return NULL;
}

typedef struct TZGNCoreRef {
    TZGNCore*       obj;
    int32_t         refCount;
    double          lastAccess;
} TZGNCoreRef;

// TZGNCore object cache handling
static UMutex gTZGNLock;
static UHashtable *gTZGNCoreCache = NULL;
static UBool gTZGNCoreCacheInitialized = FALSE;

// Access count - incremented every time up to tzgnames_SWEEP_INTERVAL,
// then reset to 0
static int32_t tzgnames_gAccessCount = 0;

// Interval for calling the cache sweep function - every 100 times
#define tzgnames_SWEEP_INTERVAL 100

// Cache expiration in millisecond. When a cached entry is no
// longer referenced and exceeding this threshold since last
// access time, then the cache entry will be deleted by the sweep
// function. For now, 3 minutes.
#define tzgnames_CACHE_EXPIRATION 180000.0

U_CDECL_BEGIN
/**
 * Cleanup callback func
 */
static UBool U_CALLCONV tzgnCore_cleanup(void)
{
    if (gTZGNCoreCache != NULL) {
        uhash_close(gTZGNCoreCache);
        gTZGNCoreCache = NULL;
    }
    gTZGNCoreCacheInitialized = FALSE;
    return TRUE;
}

/**
 * Deleter for TZGNCoreRef
 */
static void U_CALLCONV
deleteTZGNCoreRef(void *obj) {
    icu::TZGNCoreRef *entry = (icu::TZGNCoreRef*)obj;
    delete (icu::TZGNCore*) entry->obj;
    uprv_free(entry);
}
U_CDECL_END

/**
 * Function used for removing unreferrenced cache entries exceeding
 * the expiration time. This function must be called with in the mutex
 * block.
 */
static void tzgnames_sweepCache() {
    int32_t pos = UHASH_FIRST;
    const UHashElement* elem;
    double now = (double)uprv_getUTCtime();

    while ((elem = uhash_nextElement(gTZGNCoreCache, &pos)) != NULL) {
        TZGNCoreRef *entry = (TZGNCoreRef *)elem->value.pointer;
        if (entry->refCount <= 0 && (now - entry->lastAccess) > tzgnames_CACHE_EXPIRATION) {
            // delete this entry
            uhash_removeElement(gTZGNCoreCache, elem);
        }
    }
}

TimeZoneGenericNames::TimeZoneGenericNames()
: fRef(0) {
}

TimeZoneGenericNames::~TimeZoneGenericNames() {
    umtx_lock(&gTZGNLock);
    {
        U_ASSERT(fRef->refCount > 0);
        // Just decrement the reference count
        fRef->refCount--;
    }
    umtx_unlock(&gTZGNLock);
}

TimeZoneGenericNames*
TimeZoneGenericNames::createInstance(const Locale& locale, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    TimeZoneGenericNames* instance = new TimeZoneGenericNames();
    if (instance == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    TZGNCoreRef *cacheEntry = NULL;
    {
        Mutex lock(&gTZGNLock);

        if (!gTZGNCoreCacheInitialized) {
            // Create empty hashtable
            gTZGNCoreCache = uhash_open(uhash_hashChars, uhash_compareChars, NULL, &status);
            if (U_SUCCESS(status)) {
                uhash_setKeyDeleter(gTZGNCoreCache, uprv_free);
                uhash_setValueDeleter(gTZGNCoreCache, deleteTZGNCoreRef);
                gTZGNCoreCacheInitialized = TRUE;
                ucln_i18n_registerCleanup(UCLN_I18N_TIMEZONEGENERICNAMES, tzgnCore_cleanup);
            }
        }
        if (U_FAILURE(status)) {
            return NULL;
        }

        // Check the cache, if not available, create new one and cache
        const char *key = locale.getName();
        cacheEntry = (TZGNCoreRef *)uhash_get(gTZGNCoreCache, key);
        if (cacheEntry == NULL) {
            TZGNCore *tzgnCore = NULL;
            char *newKey = NULL;

            tzgnCore = new TZGNCore(locale, status);
            if (tzgnCore == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
            }
            if (U_SUCCESS(status)) {
                newKey = (char *)uprv_malloc(uprv_strlen(key) + 1);
                if (newKey == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                } else {
                    uprv_strcpy(newKey, key);
                }
            }
            if (U_SUCCESS(status)) {
                cacheEntry = (TZGNCoreRef *)uprv_malloc(sizeof(TZGNCoreRef));
                if (cacheEntry == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                } else {
                    cacheEntry->obj = tzgnCore;
                    cacheEntry->refCount = 1;
                    cacheEntry->lastAccess = (double)uprv_getUTCtime();

                    uhash_put(gTZGNCoreCache, newKey, cacheEntry, &status);
                }
            }
            if (U_FAILURE(status)) {
                if (tzgnCore != NULL) {
                    delete tzgnCore;
                }
                if (newKey != NULL) {
                    uprv_free(newKey);
                }
                if (cacheEntry != NULL) {
                    uprv_free(cacheEntry);
                }
                cacheEntry = NULL;
            }
        } else {
            // Update the reference count
            cacheEntry->refCount++;
            cacheEntry->lastAccess = (double)uprv_getUTCtime();
        }
        tzgnames_gAccessCount++;
        if (tzgnames_gAccessCount >= tzgnames_SWEEP_INTERVAL) {
            // sweep
            tzgnames_sweepCache();
            tzgnames_gAccessCount = 0;
        }
    }  // End of mutex locked block

    if (cacheEntry == NULL) {
        delete instance;
        return NULL;
    }

    instance->fRef = cacheEntry;
    return instance;
}

UBool
TimeZoneGenericNames::operator==(const TimeZoneGenericNames& other) const {
    // Just compare if the other object also use the same
    // ref entry
    return fRef == other.fRef;
}

TimeZoneGenericNames*
TimeZoneGenericNames::clone() const {
    TimeZoneGenericNames* other = new TimeZoneGenericNames();
    if (other) {
        umtx_lock(&gTZGNLock);
        {
            // Just increments the reference count
            fRef->refCount++;
            other->fRef = fRef;
        }
        umtx_unlock(&gTZGNLock);
    }
    return other;
}

UnicodeString&
TimeZoneGenericNames::getDisplayName(const TimeZone& tz, UTimeZoneGenericNameType type,
                        UDate date, UnicodeString& name) const {
    return fRef->obj->getDisplayName(tz, type, date, name);
}

UnicodeString&
TimeZoneGenericNames::getGenericLocationName(const UnicodeString& tzCanonicalID, UnicodeString& name) const {
    return fRef->obj->getGenericLocationName(tzCanonicalID, name);
}

int32_t
TimeZoneGenericNames::findBestMatch(const UnicodeString& text, int32_t start, uint32_t types,
        UnicodeString& tzID, UTimeZoneFormatTimeType& timeType, UErrorCode& status) const {
    return fRef->obj->findBestMatch(text, start, types, tzID, timeType, status);
}

U_NAMESPACE_END
#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2011-2015, International Business Machines Corporation and    *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING
















U_NAMESPACE_BEGIN

// TimeZoneNames object cache handling
static UMutex gTimeZoneNamesLock;
static UHashtable *gTimeZoneNamesCache = NULL;
static UBool gTimeZoneNamesCacheInitialized = FALSE;

// Access count - incremented every time up to tznames_SWEEP_INTERVAL,
// then reset to 0
static int32_t tznames_gAccessCount = 0;

// Interval for calling the cache sweep function - every 100 times
#define tznames_SWEEP_INTERVAL 100

// Cache expiration in millisecond. When a cached entry is no
// longer referenced and exceeding this threshold since last
// access time, then the cache entry will be deleted by the sweep
// function. For now, 3 minutes.
#define tznames_CACHE_EXPIRATION 180000.0

typedef struct TimeZoneNamesCacheEntry {
    TimeZoneNames*  names;
    int32_t         refCount;
    double          lastAccess;
} TimeZoneNamesCacheEntry;

U_CDECL_BEGIN
/**
 * Cleanup callback func
 */
static UBool U_CALLCONV timeZoneNames_cleanup(void)
{
    if (gTimeZoneNamesCache != NULL) {
        uhash_close(gTimeZoneNamesCache);
        gTimeZoneNamesCache = NULL;
    }
    gTimeZoneNamesCacheInitialized = FALSE;
    return TRUE;
}

/**
 * Deleter for TimeZoneNamesCacheEntry
 */
static void U_CALLCONV
deleteTimeZoneNamesCacheEntry(void *obj) {
    icu::TimeZoneNamesCacheEntry *entry = (icu::TimeZoneNamesCacheEntry*)obj;
    delete (icu::TimeZoneNamesImpl*) entry->names;
    uprv_free(entry);
}
U_CDECL_END

/**
 * Function used for removing unreferrenced cache entries exceeding
 * the expiration time. This function must be called with in the mutex
 * block.
 */
static void tznames_sweepCache() {
    int32_t pos = UHASH_FIRST;
    const UHashElement* elem;
    double now = (double)uprv_getUTCtime();

    while ((elem = uhash_nextElement(gTimeZoneNamesCache, &pos)) != 0) {
        TimeZoneNamesCacheEntry *entry = (TimeZoneNamesCacheEntry *)elem->value.pointer;
        if (entry->refCount <= 0 && (now - entry->lastAccess) > tznames_CACHE_EXPIRATION) {
            // delete this entry
            uhash_removeElement(gTimeZoneNamesCache, elem);
        }
    }
}

// ---------------------------------------------------
// TimeZoneNamesDelegate
// ---------------------------------------------------
class TimeZoneNamesDelegate : public TimeZoneNames {
public:
    TimeZoneNamesDelegate(const Locale& locale, UErrorCode& status);
    virtual ~TimeZoneNamesDelegate();

    virtual UBool operator==(const TimeZoneNames& other) const;
    virtual UBool operator!=(const TimeZoneNames& other) const {return !operator==(other);}
    virtual TimeZoneNamesDelegate* clone() const;

    StringEnumeration* getAvailableMetaZoneIDs(UErrorCode& status) const;
    StringEnumeration* getAvailableMetaZoneIDs(const UnicodeString& tzID, UErrorCode& status) const;
    UnicodeString& getMetaZoneID(const UnicodeString& tzID, UDate date, UnicodeString& mzID) const;
    UnicodeString& getReferenceZoneID(const UnicodeString& mzID, const char* region, UnicodeString& tzID) const;

    UnicodeString& getMetaZoneDisplayName(const UnicodeString& mzID, UTimeZoneNameType type, UnicodeString& name) const;
    UnicodeString& getTimeZoneDisplayName(const UnicodeString& tzID, UTimeZoneNameType type, UnicodeString& name) const;

    UnicodeString& getExemplarLocationName(const UnicodeString& tzID, UnicodeString& name) const;

    void loadAllDisplayNames(UErrorCode& status);
    void getDisplayNames(const UnicodeString& tzID, const UTimeZoneNameType types[], int32_t numTypes, UDate date, UnicodeString dest[], UErrorCode& status) const;

    MatchInfoCollection* find(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const;
private:
    TimeZoneNamesDelegate();
    TimeZoneNamesCacheEntry*    fTZnamesCacheEntry;
};

TimeZoneNamesDelegate::TimeZoneNamesDelegate()
: fTZnamesCacheEntry(0) {
}

TimeZoneNamesDelegate::TimeZoneNamesDelegate(const Locale& locale, UErrorCode& status) {
    Mutex lock(&gTimeZoneNamesLock);
    if (!gTimeZoneNamesCacheInitialized) {
        // Create empty hashtable if it is not already initialized.
        gTimeZoneNamesCache = uhash_open(uhash_hashChars, uhash_compareChars, NULL, &status);
        if (U_SUCCESS(status)) {
            uhash_setKeyDeleter(gTimeZoneNamesCache, uprv_free);
            uhash_setValueDeleter(gTimeZoneNamesCache, deleteTimeZoneNamesCacheEntry);
            gTimeZoneNamesCacheInitialized = TRUE;
            ucln_i18n_registerCleanup(UCLN_I18N_TIMEZONENAMES, timeZoneNames_cleanup);
        }
    }

    if (U_FAILURE(status)) {
        return;
    }

    // Check the cache, if not available, create new one and cache
    TimeZoneNamesCacheEntry *cacheEntry = NULL;

    const char *key = locale.getName();
    cacheEntry = (TimeZoneNamesCacheEntry *)uhash_get(gTimeZoneNamesCache, key);
    if (cacheEntry == NULL) {
        TimeZoneNames *tznames = NULL;
        char *newKey = NULL;

        tznames = new TimeZoneNamesImpl(locale, status);
        if (tznames == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
        }
        if (U_SUCCESS(status)) {
            newKey = (char *)uprv_malloc(uprv_strlen(key) + 1);
            if (newKey == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
            } else {
                uprv_strcpy(newKey, key);
            }
        }
        if (U_SUCCESS(status)) {
            cacheEntry = (TimeZoneNamesCacheEntry *)uprv_malloc(sizeof(TimeZoneNamesCacheEntry));
            if (cacheEntry == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
            } else {
                cacheEntry->names = tznames;
                cacheEntry->refCount = 1;
                cacheEntry->lastAccess = (double)uprv_getUTCtime();

                uhash_put(gTimeZoneNamesCache, newKey, cacheEntry, &status);
            }
        }
        if (U_FAILURE(status)) {
            if (tznames != NULL) {
                delete tznames;
            }
            if (newKey != NULL) {
                uprv_free(newKey);
            }
            if (cacheEntry != NULL) {
                uprv_free(cacheEntry);
            }
            cacheEntry = NULL;
        }
    } else {
        // Update the reference count
        cacheEntry->refCount++;
        cacheEntry->lastAccess = (double)uprv_getUTCtime();
    }
    tznames_gAccessCount++;
    if (tznames_gAccessCount >= tznames_SWEEP_INTERVAL) {
        // sweep
        tznames_sweepCache();
        tznames_gAccessCount = 0;
    }
    fTZnamesCacheEntry = cacheEntry;
}

TimeZoneNamesDelegate::~TimeZoneNamesDelegate() {
    umtx_lock(&gTimeZoneNamesLock);
    {
        if (fTZnamesCacheEntry) {
            U_ASSERT(fTZnamesCacheEntry->refCount > 0);
            // Just decrement the reference count
            fTZnamesCacheEntry->refCount--;
        }
    }
    umtx_unlock(&gTimeZoneNamesLock);
}

UBool
TimeZoneNamesDelegate::operator==(const TimeZoneNames& other) const {
    if (this == &other) {
        return TRUE;
    }
    // Just compare if the other object also use the same
    // cache entry
    const TimeZoneNamesDelegate* rhs = dynamic_cast<const TimeZoneNamesDelegate*>(&other);
    if (rhs) {
        return fTZnamesCacheEntry == rhs->fTZnamesCacheEntry;
    }
    return FALSE;
}

TimeZoneNamesDelegate*
TimeZoneNamesDelegate::clone() const {
    TimeZoneNamesDelegate* other = new TimeZoneNamesDelegate();
    if (other != NULL) {
        umtx_lock(&gTimeZoneNamesLock);
        {
            // Just increment the reference count
            fTZnamesCacheEntry->refCount++;
            other->fTZnamesCacheEntry = fTZnamesCacheEntry;
        }
        umtx_unlock(&gTimeZoneNamesLock);
    }
    return other;
}

StringEnumeration*
TimeZoneNamesDelegate::getAvailableMetaZoneIDs(UErrorCode& status) const {
    return fTZnamesCacheEntry->names->getAvailableMetaZoneIDs(status);
}

StringEnumeration*
TimeZoneNamesDelegate::getAvailableMetaZoneIDs(const UnicodeString& tzID, UErrorCode& status) const {
    return fTZnamesCacheEntry->names->getAvailableMetaZoneIDs(tzID, status);
}

UnicodeString&
TimeZoneNamesDelegate::getMetaZoneID(const UnicodeString& tzID, UDate date, UnicodeString& mzID) const {
    return fTZnamesCacheEntry->names->getMetaZoneID(tzID, date, mzID);
}

UnicodeString&
TimeZoneNamesDelegate::getReferenceZoneID(const UnicodeString& mzID, const char* region, UnicodeString& tzID) const {
    return fTZnamesCacheEntry->names->getReferenceZoneID(mzID, region, tzID);
}

UnicodeString&
TimeZoneNamesDelegate::getMetaZoneDisplayName(const UnicodeString& mzID, UTimeZoneNameType type, UnicodeString& name) const {
    return fTZnamesCacheEntry->names->getMetaZoneDisplayName(mzID, type, name);
}

UnicodeString&
TimeZoneNamesDelegate::getTimeZoneDisplayName(const UnicodeString& tzID, UTimeZoneNameType type, UnicodeString& name) const {
    return fTZnamesCacheEntry->names->getTimeZoneDisplayName(tzID, type, name);
}

UnicodeString&
TimeZoneNamesDelegate::getExemplarLocationName(const UnicodeString& tzID, UnicodeString& name) const {
    return fTZnamesCacheEntry->names->getExemplarLocationName(tzID, name);
}

void
TimeZoneNamesDelegate::loadAllDisplayNames(UErrorCode& status) {
    fTZnamesCacheEntry->names->loadAllDisplayNames(status);
}

void
TimeZoneNamesDelegate::getDisplayNames(const UnicodeString& tzID, const UTimeZoneNameType types[], int32_t numTypes, UDate date, UnicodeString dest[], UErrorCode& status) const {
    fTZnamesCacheEntry->names->getDisplayNames(tzID, types, numTypes, date, dest, status);
}

TimeZoneNames::MatchInfoCollection*
TimeZoneNamesDelegate::find(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const {
    return fTZnamesCacheEntry->names->find(text, start, types, status);
}

// ---------------------------------------------------
// TimeZoneNames base class
// ---------------------------------------------------
TimeZoneNames::~TimeZoneNames() {
}

TimeZoneNames*
TimeZoneNames::createInstance(const Locale& locale, UErrorCode& status) {
    TimeZoneNames *instance = NULL;
    if (U_SUCCESS(status)) {
        instance = new TimeZoneNamesDelegate(locale, status);
        if (instance == NULL && U_SUCCESS(status)) {
            status = U_MEMORY_ALLOCATION_ERROR;
        }
    }
    return instance;
}

TimeZoneNames*
TimeZoneNames::createTZDBInstance(const Locale& locale, UErrorCode& status) {
    TimeZoneNames *instance = NULL;
    if (U_SUCCESS(status)) {
        instance = new TZDBTimeZoneNames(locale);
        if (instance == NULL && U_SUCCESS(status)) {
            status = U_MEMORY_ALLOCATION_ERROR;
        }
    }
    return instance;
}

UnicodeString&
TimeZoneNames::getExemplarLocationName(const UnicodeString& tzID, UnicodeString& name) const {
    return TimeZoneNamesImpl::getDefaultExemplarLocationName(tzID, name);
}

UnicodeString&
TimeZoneNames::getDisplayName(const UnicodeString& tzID, UTimeZoneNameType type, UDate date, UnicodeString& name) const {
    getTimeZoneDisplayName(tzID, type, name);
    if (name.isEmpty()) {
        UChar mzIDBuf[32];
        UnicodeString mzID(mzIDBuf, 0, UPRV_LENGTHOF(mzIDBuf));
        getMetaZoneID(tzID, date, mzID);
        getMetaZoneDisplayName(mzID, type, name);
    }
    return name;
}

// Empty default implementation, to be overriden in tznames_impl.cpp.
void
TimeZoneNames::loadAllDisplayNames(UErrorCode& /*status*/) {
}

// A default, lightweight implementation of getDisplayNames.
// Overridden in tznames_impl.cpp.
void
TimeZoneNames::getDisplayNames(const UnicodeString& tzID, const UTimeZoneNameType types[], int32_t numTypes, UDate date, UnicodeString dest[], UErrorCode& status) const {
    if (U_FAILURE(status)) { return; }
    if (tzID.isEmpty()) { return; }
    UnicodeString mzID;
    for (int i = 0; i < numTypes; i++) {
        getTimeZoneDisplayName(tzID, types[i], dest[i]);
        if (dest[i].isEmpty()) {
            if (mzID.isEmpty()) {
                getMetaZoneID(tzID, date, mzID);
            }
            getMetaZoneDisplayName(mzID, types[i], dest[i]);
        }
    }
}


struct MatchInfo : UMemory {
    UTimeZoneNameType nameType;
    UnicodeString id;
    int32_t matchLength;
    UBool isTZID;

    MatchInfo(UTimeZoneNameType nameType, int32_t matchLength, const UnicodeString* tzID, const UnicodeString* mzID) {
        this->nameType = nameType;
        this->matchLength = matchLength;
        if (tzID != NULL) {
            this->id.setTo(*tzID);
            this->isTZID = TRUE;
        } else {
            this->id.setTo(*mzID);
            this->isTZID = FALSE;
        }
    }
};

U_CDECL_BEGIN
static void U_CALLCONV
deleteMatchInfo(void *obj) {
    delete static_cast<MatchInfo *>(obj);
}
U_CDECL_END

// ---------------------------------------------------
// MatchInfoCollection class
// ---------------------------------------------------
TimeZoneNames::MatchInfoCollection::MatchInfoCollection()
: fMatches(NULL) {
}

TimeZoneNames::MatchInfoCollection::~MatchInfoCollection() {
    if (fMatches != NULL) {
        delete fMatches;
    }
}

void
TimeZoneNames::MatchInfoCollection::addZone(UTimeZoneNameType nameType, int32_t matchLength,
            const UnicodeString& tzID, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    MatchInfo* matchInfo = new MatchInfo(nameType, matchLength, &tzID, NULL);
    if (matchInfo == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    matches(status)->addElement(matchInfo, status);
    if (U_FAILURE(status)) {
        delete matchInfo;
    }
}

void
TimeZoneNames::MatchInfoCollection::addMetaZone(UTimeZoneNameType nameType, int32_t matchLength,
            const UnicodeString& mzID, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    MatchInfo* matchInfo = new MatchInfo(nameType, matchLength, NULL, &mzID);
    if (matchInfo == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }
    matches(status)->addElement(matchInfo, status);
    if (U_FAILURE(status)) {
        delete matchInfo;
    }
}

int32_t
TimeZoneNames::MatchInfoCollection::size() const {
    if (fMatches == NULL) {
        return 0;
    }
    return fMatches->size();
}

UTimeZoneNameType
TimeZoneNames::MatchInfoCollection::getNameTypeAt(int32_t idx) const {
    const MatchInfo* match = (const MatchInfo*)fMatches->elementAt(idx);
    if (match) {
        return match->nameType;
    }
    return UTZNM_UNKNOWN;
}

int32_t
TimeZoneNames::MatchInfoCollection::getMatchLengthAt(int32_t idx) const {
    const MatchInfo* match = (const MatchInfo*)fMatches->elementAt(idx);
    if (match) {
        return match->matchLength;
    }
    return 0;
}

UBool
TimeZoneNames::MatchInfoCollection::getTimeZoneIDAt(int32_t idx, UnicodeString& tzID) const {
    tzID.remove();
    const MatchInfo* match = (const MatchInfo*)fMatches->elementAt(idx);
    if (match && match->isTZID) {
        tzID.setTo(match->id);
        return TRUE;
    }
    return FALSE;
}

UBool
TimeZoneNames::MatchInfoCollection::getMetaZoneIDAt(int32_t idx, UnicodeString& mzID) const {
    mzID.remove();
    const MatchInfo* match = (const MatchInfo*)fMatches->elementAt(idx);
    if (match && !match->isTZID) {
        mzID.setTo(match->id);
        return TRUE;
    }
    return FALSE;
}

UVector*
TimeZoneNames::MatchInfoCollection::matches(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    if (fMatches != NULL) {
        return fMatches;
    }
    fMatches = new UVector(deleteMatchInfo, NULL, status);
    if (fMatches == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    } else if (U_FAILURE(status)) {
        delete fMatches;
        fMatches = NULL;
    }
    return fMatches;
}


U_NAMESPACE_END
#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2011-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File TZNAMES_IMPL.CPP
*
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING






















U_NAMESPACE_BEGIN

#define ZID_KEY_MAX  128
#define MZ_PREFIX_LEN 5

static const char tzgnames_impl_gZoneStrings[]        = "zoneStrings";
static const char tzgnames_impl_gMZPrefix[]           = "meta:";

static const char tzgnames_impl_EMPTY[]               = "<empty>";   // place holder for empty ZNames
static const char tzgnames_impl_DUMMY_LOADER[]        = "<dummy>";   // place holder for dummy ZNamesLoader
static const UChar tzgnames_impl_NO_NAME[]            = { 0 };   // for empty no-fallback time zone names

// stuff for TZDBTimeZoneNames
static const char* tzgnames_impl_TZDBNAMES_KEYS[]               = {"ss", "sd"};
static const int32_t tzgnames_impl_TZDBNAMES_KEYS_SIZE = UPRV_LENGTHOF(tzgnames_impl_TZDBNAMES_KEYS);

static UMutex tzgnames_impl_gDataMutex;

static UHashtable* tzgnames_impl_gTZDBNamesMap = NULL;
static icu::UInitOnce tzgnames_impl_gTZDBNamesMapInitOnce = U_INITONCE_INITIALIZER;

static TextTrieMap* tzgnames_impl_gTZDBNamesTrie = NULL;
static icu::UInitOnce tzgnames_impl_gTZDBNamesTrieInitOnce = U_INITONCE_INITIALIZER;

// The order in which strings are stored may be different than the order in the public enum.
enum UTimeZoneNameTypeIndex {
    UTZNM_INDEX_UNKNOWN = -1,
    UTZNM_INDEX_EXEMPLAR_LOCATION,
    UTZNM_INDEX_LONG_GENERIC,
    UTZNM_INDEX_LONG_STANDARD,
    UTZNM_INDEX_LONG_DAYLIGHT,
    UTZNM_INDEX_SHORT_GENERIC,
    UTZNM_INDEX_SHORT_STANDARD,
    UTZNM_INDEX_SHORT_DAYLIGHT,
    UTZNM_INDEX_COUNT
};
static const UChar* const tzgnames_impl_EMPTY_NAMES[UTZNM_INDEX_COUNT] = {0,0,0,0,0,0,0};

U_CDECL_BEGIN
static UBool U_CALLCONV tzdbTimeZoneNames_cleanup(void) {
    if (tzgnames_impl_gTZDBNamesMap != NULL) {
        uhash_close(tzgnames_impl_gTZDBNamesMap);
        tzgnames_impl_gTZDBNamesMap = NULL;
    }
    tzgnames_impl_gTZDBNamesMapInitOnce.reset();

    if (tzgnames_impl_gTZDBNamesTrie != NULL) {
        delete tzgnames_impl_gTZDBNamesTrie;
        tzgnames_impl_gTZDBNamesTrie = NULL;
    }
    tzgnames_impl_gTZDBNamesTrieInitOnce.reset();

    return TRUE;
}
U_CDECL_END

/**
 * tznames_ZNameInfo stores zone name information in the trie
 */
struct tznames_ZNameInfo {
    UTimeZoneNameType   type;
    const UChar*        tzID;
    const UChar*        mzID;
};

/**
 * tznames_ZMatchInfo stores zone name match information used by find method
 */
struct tznames_ZMatchInfo {
    const tznames_ZNameInfo*    znameInfo;
    int32_t             matchLength;
};

// Helper functions
static void mergeTimeZoneKey(const UnicodeString& mzID, char* result);

#define DEFAULT_CHARACTERNODE_CAPACITY 1

// ---------------------------------------------------
// CharacterNode class implementation
// ---------------------------------------------------
void CharacterNode::clear() {
    uprv_memset(this, 0, sizeof(*this));
}

void CharacterNode::deleteValues(UObjectDeleter *valueDeleter) {
    if (fValues == NULL) {
        // Do nothing.
    } else if (!fHasValuesVector) {
        if (valueDeleter) {
            valueDeleter(fValues);
        }
    } else {
        delete (UVector *)fValues;
    }
}

void
CharacterNode::addValue(void *value, UObjectDeleter *valueDeleter, UErrorCode &status) {
    if (U_FAILURE(status)) {
        if (valueDeleter) {
            valueDeleter(value);
        }
        return;
    }
    if (fValues == NULL) {
        fValues = value;
    } else {
        // At least one value already.
        if (!fHasValuesVector) {
            // There is only one value so far, and not in a vector yet.
            // Create a vector and add the old value.
            UVector *values = new UVector(valueDeleter, NULL, DEFAULT_CHARACTERNODE_CAPACITY, status);
            if (U_FAILURE(status)) {
                if (valueDeleter) {
                    valueDeleter(value);
                }
                return;
            }
            values->addElement(fValues, status);
            fValues = values;
            fHasValuesVector = TRUE;
        }
        // Add the new value.
        ((UVector *)fValues)->addElement(value, status);
    }
}

// ---------------------------------------------------
// TextTrieMapSearchResultHandler class implementation
// ---------------------------------------------------
TextTrieMapSearchResultHandler::~TextTrieMapSearchResultHandler(){
}

// ---------------------------------------------------
// TextTrieMap class implementation
// ---------------------------------------------------
TextTrieMap::TextTrieMap(UBool ignoreCase, UObjectDeleter *valueDeleter)
: fIgnoreCase(ignoreCase), fNodes(NULL), fNodesCapacity(0), fNodesCount(0),
  fLazyContents(NULL), fIsEmpty(TRUE), fValueDeleter(valueDeleter) {
}

TextTrieMap::~TextTrieMap() {
    int32_t index;
    for (index = 0; index < fNodesCount; ++index) {
        fNodes[index].deleteValues(fValueDeleter);
    }
    uprv_free(fNodes);
    if (fLazyContents != NULL) {
        for (int32_t i=0; i<fLazyContents->size(); i+=2) {
            if (fValueDeleter) {
                fValueDeleter(fLazyContents->elementAt(i+1));
            }
        }
        delete fLazyContents;
    }
}

int32_t TextTrieMap::isEmpty() const {
    // Use a separate field for fIsEmpty because it will remain unchanged once the
    //   Trie is built, while fNodes and fLazyContents change with the lazy init
    //   of the nodes structure.  Trying to test the changing fields has
    //   thread safety complications.
    return fIsEmpty;
}


//  We defer actually building the TextTrieMap node structure until the first time a
//     search is performed.  put() simply saves the parameters in case we do
//     eventually need to build it.
//
void
TextTrieMap::put(const UnicodeString &key, void *value, ZNStringPool &sp, UErrorCode &status) {
    const UChar *s = sp.get(key, status);
    put(s, value, status);
}

// This method is designed for a persistent key, such as string key stored in
// resource bundle.
void
TextTrieMap::put(const UChar *key, void *value, UErrorCode &status) {
    fIsEmpty = FALSE;
    if (fLazyContents == NULL) {
        fLazyContents = new UVector(status);
        if (fLazyContents == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
        }
    }
    if (U_FAILURE(status)) {
        if (fValueDeleter) {
            fValueDeleter((void*) key);
        }
        return;
    }
    U_ASSERT(fLazyContents != NULL);

    UChar *s = const_cast<UChar *>(key);
    fLazyContents->addElement(s, status);
    if (U_FAILURE(status)) {
        if (fValueDeleter) {
            fValueDeleter((void*) key);
        }
        return;
    }

    fLazyContents->addElement(value, status);
}

void
TextTrieMap::putImpl(const UnicodeString &key, void *value, UErrorCode &status) {
    if (fNodes == NULL) {
        fNodesCapacity = 512;
        fNodes = (CharacterNode *)uprv_malloc(fNodesCapacity * sizeof(CharacterNode));
        if (fNodes == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return;
        }
        fNodes[0].clear();  // Init root node.
        fNodesCount = 1;
    }

    UnicodeString foldedKey;
    const UChar *keyBuffer;
    int32_t keyLength;
    if (fIgnoreCase) {
        // Ok to use fastCopyFrom() because we discard the copy when we return.
        foldedKey.fastCopyFrom(key).foldCase();
        keyBuffer = foldedKey.getBuffer();
        keyLength = foldedKey.length();
    } else {
        keyBuffer = key.getBuffer();
        keyLength = key.length();
    }

    CharacterNode *node = fNodes;
    int32_t index;
    for (index = 0; index < keyLength; ++index) {
        node = addChildNode(node, keyBuffer[index], status);
    }
    node->addValue(value, fValueDeleter, status);
}

UBool
TextTrieMap::growNodes() {
    if (fNodesCapacity == 0xffff) {
        return FALSE;  // We use 16-bit node indexes.
    }
    int32_t newCapacity = fNodesCapacity + 1000;
    if (newCapacity > 0xffff) {
        newCapacity = 0xffff;
    }
    CharacterNode *newNodes = (CharacterNode *)uprv_malloc(newCapacity * sizeof(CharacterNode));
    if (newNodes == NULL) {
        return FALSE;
    }
    uprv_memcpy(newNodes, fNodes, fNodesCount * sizeof(CharacterNode));
    uprv_free(fNodes);
    fNodes = newNodes;
    fNodesCapacity = newCapacity;
    return TRUE;
}

CharacterNode*
TextTrieMap::addChildNode(CharacterNode *parent, UChar c, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    // Linear search of the sorted list of children.
    uint16_t prevIndex = 0;
    uint16_t nodeIndex = parent->fFirstChild;
    while (nodeIndex > 0) {
        CharacterNode *current = fNodes + nodeIndex;
        UChar childCharacter = current->fCharacter;
        if (childCharacter == c) {
            return current;
        } else if (childCharacter > c) {
            break;
        }
        prevIndex = nodeIndex;
        nodeIndex = current->fNextSibling;
    }

    // Ensure capacity. Grow fNodes[] if needed.
    if (fNodesCount == fNodesCapacity) {
        int32_t parentIndex = (int32_t)(parent - fNodes);
        if (!growNodes()) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        parent = fNodes + parentIndex;
    }

    // Insert a new child node with c in sorted order.
    CharacterNode *node = fNodes + fNodesCount;
    node->clear();
    node->fCharacter = c;
    node->fNextSibling = nodeIndex;
    if (prevIndex == 0) {
        parent->fFirstChild = (uint16_t)fNodesCount;
    } else {
        fNodes[prevIndex].fNextSibling = (uint16_t)fNodesCount;
    }
    ++fNodesCount;
    return node;
}

CharacterNode*
TextTrieMap::getChildNode(CharacterNode *parent, UChar c) const {
    // Linear search of the sorted list of children.
    uint16_t nodeIndex = parent->fFirstChild;
    while (nodeIndex > 0) {
        CharacterNode *current = fNodes + nodeIndex;
        UChar childCharacter = current->fCharacter;
        if (childCharacter == c) {
            return current;
        } else if (childCharacter > c) {
            break;
        }
        nodeIndex = current->fNextSibling;
    }
    return NULL;
}


// buildTrie() - The Trie node structure is needed.  Create it from the data that was
//               saved at the time the ZoneStringFormatter was created.  The Trie is only
//               needed for parsing operations, which are less common than formatting,
//               and the Trie is big, which is why its creation is deferred until first use.
void TextTrieMap::buildTrie(UErrorCode &status) {
    if (fLazyContents != NULL) {
        for (int32_t i=0; i<fLazyContents->size(); i+=2) {
            const UChar *key = (UChar *)fLazyContents->elementAt(i);
            void  *val = fLazyContents->elementAt(i+1);
            UnicodeString keyString(TRUE, key, -1);  // Aliasing UnicodeString constructor.
            putImpl(keyString, val, status);
        }
        delete fLazyContents;
        fLazyContents = NULL;
    }
}

void
TextTrieMap::search(const UnicodeString &text, int32_t start,
                  TextTrieMapSearchResultHandler *handler, UErrorCode &status) const {
    {
        // TODO: if locking the mutex for each check proves to be a performance problem,
        //       add a flag of type atomic_int32_t to class TextTrieMap, and use only
        //       the ICU atomic safe functions for assigning and testing.
        //       Don't test the pointer fLazyContents.
        //       Don't do unless it's really required.

        // Mutex for protecting the lazy creation of the Trie node structure on the first call to search().
        static UMutex TextTrieMutex;

        Mutex lock(&TextTrieMutex);
        if (fLazyContents != NULL) {
            TextTrieMap *nonConstThis = const_cast<TextTrieMap *>(this);
            nonConstThis->buildTrie(status);
        }
    }
    if (fNodes == NULL) {
        return;
    }
    search(fNodes, text, start, start, handler, status);
}

void
TextTrieMap::search(CharacterNode *node, const UnicodeString &text, int32_t start,
                  int32_t index, TextTrieMapSearchResultHandler *handler, UErrorCode &status) const {
    if (U_FAILURE(status)) {
        return;
    }
    if (node->hasValues()) {
        if (!handler->handleMatch(index - start, node, status)) {
            return;
        }
        if (U_FAILURE(status)) {
            return;
        }
    }
    if (fIgnoreCase) {
        // for folding we need to get a complete code point.
        // size of character may grow after fold operation;
        // then we need to get result as UTF16 code units.
        UChar32 c32 = text.char32At(index);
        index += U16_LENGTH(c32);
        UnicodeString tmp(c32);
        tmp.foldCase();
        int32_t tmpidx = 0;
        while (tmpidx < tmp.length()) {
            UChar c = tmp.charAt(tmpidx++);
            node = getChildNode(node, c);
            if (node == NULL) {
                break;
            }
        }
    } else {
        // here we just get the next UTF16 code unit
        UChar c = text.charAt(index++);
        node = getChildNode(node, c);
    }
    if (node != NULL) {
        search(node, text, start, index, handler, status);
    }
}

// ---------------------------------------------------
// ZNStringPool class implementation
// ---------------------------------------------------
static const int32_t POOL_CHUNK_SIZE = 2000;
struct ZNStringPoolChunk: public UMemory {
    ZNStringPoolChunk    *fNext;                       // Ptr to next pool chunk
    int32_t               fLimit;                       // Index to start of unused area at end of fStrings
    UChar                 fStrings[POOL_CHUNK_SIZE];    //  Strings array
    ZNStringPoolChunk();
};

ZNStringPoolChunk::ZNStringPoolChunk() {
    fNext = NULL;
    fLimit = 0;
}

ZNStringPool::ZNStringPool(UErrorCode &status) {
    fChunks = NULL;
    fHash   = NULL;
    if (U_FAILURE(status)) {
        return;
    }
    fChunks = new ZNStringPoolChunk;
    if (fChunks == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    fHash   = uhash_open(uhash_hashUChars      /* keyHash */,
                         uhash_compareUChars   /* keyComp */,
                         uhash_compareUChars   /* valueComp */,
                         &status);
    if (U_FAILURE(status)) {
        return;
    }
}

ZNStringPool::~ZNStringPool() {
    if (fHash != NULL) {
        uhash_close(fHash);
        fHash = NULL;
    }

    while (fChunks != NULL) {
        ZNStringPoolChunk *nextChunk = fChunks->fNext;
        delete fChunks;
        fChunks = nextChunk;
    }
}

static const UChar EmptyString = 0;

const UChar *ZNStringPool::get(const UChar *s, UErrorCode &status) {
    const UChar *pooledString;
    if (U_FAILURE(status)) {
        return &EmptyString;
    }

    pooledString = static_cast<UChar *>(uhash_get(fHash, s));
    if (pooledString != NULL) {
        return pooledString;
    }

    int32_t length = u_strlen(s);
    int32_t remainingLength = POOL_CHUNK_SIZE - fChunks->fLimit;
    if (remainingLength <= length) {
        U_ASSERT(length < POOL_CHUNK_SIZE);
        if (length >= POOL_CHUNK_SIZE) {
            status = U_INTERNAL_PROGRAM_ERROR;
            return &EmptyString;
        }
        ZNStringPoolChunk *oldChunk = fChunks;
        fChunks = new ZNStringPoolChunk;
        if (fChunks == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return &EmptyString;
        }
        fChunks->fNext = oldChunk;
    }

    UChar *destString = &fChunks->fStrings[fChunks->fLimit];
    u_strcpy(destString, s);
    fChunks->fLimit += (length + 1);
    uhash_put(fHash, destString, destString, &status);
    return destString;
}


//
//  ZNStringPool::adopt()    Put a string into the hash, but do not copy the string data
//                           into the pool's storage.  Used for strings from resource bundles,
//                           which will perisist for the life of the zone string formatter, and
//                           therefore can be used directly without copying.
const UChar *ZNStringPool::adopt(const UChar * s, UErrorCode &status) {
    const UChar *pooledString;
    if (U_FAILURE(status)) {
        return &EmptyString;
    }
    if (s != NULL) {
        pooledString = static_cast<UChar *>(uhash_get(fHash, s));
        if (pooledString == NULL) {
            UChar *ncs = const_cast<UChar *>(s);
            uhash_put(fHash, ncs, ncs, &status);
        }
    }
    return s;
}


const UChar *ZNStringPool::get(const UnicodeString &s, UErrorCode &status) {
    UnicodeString &nonConstStr = const_cast<UnicodeString &>(s);
    return this->get(nonConstStr.getTerminatedBuffer(), status);
}

/*
 * freeze().   Close the hash table that maps to the pooled strings.
 *             After freezing, the pool can not be searched or added to,
 *             but all existing references to pooled strings remain valid.
 *
 *             The main purpose is to recover the storage used for the hash.
 */
void ZNStringPool::freeze() {
    uhash_close(fHash);
    fHash = NULL;
}


/**
 * This class stores name data for a meta zone or time zone.
 */
class ZNames : public UMemory {
private:
    friend class TimeZoneNamesImpl;

    static UTimeZoneNameTypeIndex getTZNameTypeIndex(UTimeZoneNameType type) {
        switch(type) {
        case UTZNM_EXEMPLAR_LOCATION: return UTZNM_INDEX_EXEMPLAR_LOCATION;
        case UTZNM_LONG_GENERIC: return UTZNM_INDEX_LONG_GENERIC;
        case UTZNM_LONG_STANDARD: return UTZNM_INDEX_LONG_STANDARD;
        case UTZNM_LONG_DAYLIGHT: return UTZNM_INDEX_LONG_DAYLIGHT;
        case UTZNM_SHORT_GENERIC: return UTZNM_INDEX_SHORT_GENERIC;
        case UTZNM_SHORT_STANDARD: return UTZNM_INDEX_SHORT_STANDARD;
        case UTZNM_SHORT_DAYLIGHT: return UTZNM_INDEX_SHORT_DAYLIGHT;
        default: return UTZNM_INDEX_UNKNOWN;
        }
    }
    static UTimeZoneNameType getTZNameType(UTimeZoneNameTypeIndex index) {
        switch(index) {
        case UTZNM_INDEX_EXEMPLAR_LOCATION: return UTZNM_EXEMPLAR_LOCATION;
        case UTZNM_INDEX_LONG_GENERIC: return UTZNM_LONG_GENERIC;
        case UTZNM_INDEX_LONG_STANDARD: return UTZNM_LONG_STANDARD;
        case UTZNM_INDEX_LONG_DAYLIGHT: return UTZNM_LONG_DAYLIGHT;
        case UTZNM_INDEX_SHORT_GENERIC: return UTZNM_SHORT_GENERIC;
        case UTZNM_INDEX_SHORT_STANDARD: return UTZNM_SHORT_STANDARD;
        case UTZNM_INDEX_SHORT_DAYLIGHT: return UTZNM_SHORT_DAYLIGHT;
        default: return UTZNM_UNKNOWN;
        }
    }

    const UChar* fNames[UTZNM_INDEX_COUNT];
    UBool fDidAddIntoTrie;

    // Whether we own the location string, if computed rather than loaded from a bundle.
    // A meta zone names instance never has an exemplar location string.
    UBool fOwnsLocationName;

    ZNames(const UChar* names[], const UChar* locationName)
            : fDidAddIntoTrie(FALSE) {
        uprv_memcpy(fNames, names, sizeof(fNames));
        if (locationName != NULL) {
            fOwnsLocationName = TRUE;
            fNames[UTZNM_INDEX_EXEMPLAR_LOCATION] = locationName;
        } else {
            fOwnsLocationName = FALSE;
        }
    }

public:
    ~ZNames() {
        if (fOwnsLocationName) {
            const UChar* locationName = fNames[UTZNM_INDEX_EXEMPLAR_LOCATION];
            U_ASSERT(locationName != NULL);
            uprv_free((void*) locationName);
        }
    }

private:
    static void* createMetaZoneAndPutInCache(UHashtable* cache, const UChar* names[],
            const UnicodeString& mzID, UErrorCode& status) {
        if (U_FAILURE(status)) { return NULL; }
        U_ASSERT(names != NULL);

        // Use the persistent ID as the resource key, so we can
        // avoid duplications.
        // TODO: Is there a more efficient way, like intern() in Java?
        void* key = (void*) ZoneMeta::findMetaZoneID(mzID);
        void* value;
        if (uprv_memcmp(names, tzgnames_impl_EMPTY_NAMES, sizeof(tzgnames_impl_EMPTY_NAMES)) == 0) {
            value = (void*) tzgnames_impl_EMPTY;
        } else {
            value = (void*) (new ZNames(names, NULL));
            if (value == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                return NULL;
            }
        }
        uhash_put(cache, key, value, &status);
        return value;
    }

    static void* createTimeZoneAndPutInCache(UHashtable* cache, const UChar* names[],
            const UnicodeString& tzID, UErrorCode& status) {
        if (U_FAILURE(status)) { return NULL; }
        U_ASSERT(names != NULL);

        // If necessary, compute the location name from the time zone name.
        UChar* locationName = NULL;
        if (names[UTZNM_INDEX_EXEMPLAR_LOCATION] == NULL) {
            UnicodeString locationNameUniStr;
            TimeZoneNamesImpl::getDefaultExemplarLocationName(tzID, locationNameUniStr);

            // Copy the computed location name to the heap
            if (locationNameUniStr.length() > 0) {
                const UChar* buff = locationNameUniStr.getTerminatedBuffer();
                int32_t len = sizeof(UChar) * (locationNameUniStr.length() + 1);
                locationName = (UChar*) uprv_malloc(len);
                if (locationName == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                    return NULL;
                }
                uprv_memcpy(locationName, buff, len);
            }
        }

        // Use the persistent ID as the resource key, so we can
        // avoid duplications.
        // TODO: Is there a more efficient way, like intern() in Java?
        void* key = (void*) ZoneMeta::findTimeZoneID(tzID);
        void* value = (void*) (new ZNames(names, locationName));
        if (value == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        uhash_put(cache, key, value, &status);
        return value;
    }

    const UChar* getName(UTimeZoneNameType type) const {
        UTimeZoneNameTypeIndex index = getTZNameTypeIndex(type);
        return index >= 0 ? fNames[index] : NULL;
    }

    void addAsMetaZoneIntoTrie(const UChar* mzID, TextTrieMap& trie, UErrorCode& status) {
        addNamesIntoTrie(mzID, NULL, trie, status);
    }
    void addAsTimeZoneIntoTrie(const UChar* tzID, TextTrieMap& trie, UErrorCode& status) {
        addNamesIntoTrie(NULL, tzID, trie, status);
    }

    void addNamesIntoTrie(const UChar* mzID, const UChar* tzID, TextTrieMap& trie,
            UErrorCode& status) {
        if (U_FAILURE(status)) { return; }
        if (fDidAddIntoTrie) { return; }
        fDidAddIntoTrie = TRUE;

        for (int32_t i = 0; i < UTZNM_INDEX_COUNT; i++) {
            const UChar* name = fNames[i];
            if (name != NULL) {
                tznames_ZNameInfo *nameinfo = (tznames_ZNameInfo *)uprv_malloc(sizeof(tznames_ZNameInfo));
                if (nameinfo == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                    return;
                }
                nameinfo->mzID = mzID;
                nameinfo->tzID = tzID;
                nameinfo->type = getTZNameType((UTimeZoneNameTypeIndex)i);
                trie.put(name, nameinfo, status); // trie.put() takes ownership of the key
                if (U_FAILURE(status)) {
                    return;
                }
            }
        }
    }

public:
    struct ZNamesLoader;
};

struct ZNames::ZNamesLoader : public ResourceSink {
    const UChar *names[UTZNM_INDEX_COUNT];

    ZNamesLoader() {
        clear();
    }
    virtual ~ZNamesLoader();

    /** Reset for loading another set of names. */
    void clear() {
        uprv_memcpy(names, tzgnames_impl_EMPTY_NAMES, sizeof(names));
    }

    void loadMetaZone(const UResourceBundle* zoneStrings, const UnicodeString& mzID, UErrorCode& errorCode) {
        if (U_FAILURE(errorCode)) { return; }

        char key[ZID_KEY_MAX + 1];
        mergeTimeZoneKey(mzID, key);

        loadNames(zoneStrings, key, errorCode);
    }

    void loadTimeZone(const UResourceBundle* zoneStrings, const UnicodeString& tzID, UErrorCode& errorCode) {
        // Replace "/" with ":".
        UnicodeString uKey(tzID);
        for (int32_t i = 0; i < uKey.length(); i++) {
            if (uKey.charAt(i) == (UChar)0x2F) {
                uKey.setCharAt(i, (UChar)0x3A);
            }
        }

        char key[ZID_KEY_MAX + 1];
        uKey.extract(0, uKey.length(), key, sizeof(key), US_INV);

        loadNames(zoneStrings, key, errorCode);
    }

    void loadNames(const UResourceBundle* zoneStrings, const char* key, UErrorCode& errorCode) {
        U_ASSERT(zoneStrings != NULL);
        U_ASSERT(key != NULL);
        U_ASSERT(key[0] != '\0');

        UErrorCode localStatus = U_ZERO_ERROR;
        clear();
        ures_getAllItemsWithFallback(zoneStrings, key, *this, localStatus);

        // Ignore errors, but propogate possible warnings.
        if (U_SUCCESS(localStatus)) {
            errorCode = localStatus;
        }
    }

    void setNameIfEmpty(const char* key, const ResourceValue* value, UErrorCode& errorCode) {
        UTimeZoneNameTypeIndex type = nameTypeFromKey(key);
        if (type == UTZNM_INDEX_UNKNOWN) { return; }
        if (names[type] == NULL) {
            int32_t length;
            // 'tzgnames_impl_NO_NAME' indicates internally that this field should remain empty.  It will be
            // replaced by 'NULL' in getNames()
            names[type] = (value == NULL) ? tzgnames_impl_NO_NAME : value->getString(length, errorCode);
        }
    }

    virtual void put(const char* key, ResourceValue& value, UBool /*noFallback*/,
            UErrorCode &errorCode) {
        ResourceTable namesTable = value.getTable(errorCode);
        if (U_FAILURE(errorCode)) { return; }
        for (int32_t i = 0; namesTable.getKeyAndValue(i, key, value); ++i) {
            if (value.isNoInheritanceMarker()) {
                setNameIfEmpty(key, NULL, errorCode);
            } else {
                setNameIfEmpty(key, &value, errorCode);
            }
        }
    }

    static UTimeZoneNameTypeIndex nameTypeFromKey(const char *key) {
        char c0, c1;
        if ((c0 = key[0]) == 0 || (c1 = key[1]) == 0 || key[2] != 0) {
            return UTZNM_INDEX_UNKNOWN;
        }
        if (c0 == 'l') {
            return c1 == 'g' ? UTZNM_INDEX_LONG_GENERIC :
                    c1 == 's' ? UTZNM_INDEX_LONG_STANDARD :
                        c1 == 'd' ? UTZNM_INDEX_LONG_DAYLIGHT : UTZNM_INDEX_UNKNOWN;
        } else if (c0 == 's') {
            return c1 == 'g' ? UTZNM_INDEX_SHORT_GENERIC :
                    c1 == 's' ? UTZNM_INDEX_SHORT_STANDARD :
                        c1 == 'd' ? UTZNM_INDEX_SHORT_DAYLIGHT : UTZNM_INDEX_UNKNOWN;
        } else if (c0 == 'e' && c1 == 'c') {
            return UTZNM_INDEX_EXEMPLAR_LOCATION;
        }
        return UTZNM_INDEX_UNKNOWN;
    }

    /**
    * Returns an array of names.  It is the caller's responsibility to copy the data into a
    * permanent location, as the returned array is owned by the loader instance and may be
    * cleared or leave scope.
    *
    * This is different than Java, where the array will no longer be modified and null
    * may be returned.
    */
    const UChar** getNames() {
        // Remove 'tzgnames_impl_NO_NAME' references in the array and replace with 'NULL'
        for (int32_t i = 0; i < UTZNM_INDEX_COUNT; ++i) {
            if (names[i] == tzgnames_impl_NO_NAME) {
                names[i] = NULL;
            }
        }
        return names;
    }
};

ZNames::ZNamesLoader::~ZNamesLoader() {}


// ---------------------------------------------------
// The meta zone ID enumeration class
// ---------------------------------------------------
class MetaZoneIDsEnumeration : public StringEnumeration {
public:
    MetaZoneIDsEnumeration();
    MetaZoneIDsEnumeration(const UVector& mzIDs);
    MetaZoneIDsEnumeration(UVector* mzIDs);
    virtual ~MetaZoneIDsEnumeration();
    static UClassID U_EXPORT2 getStaticClassID(void);
    virtual UClassID getDynamicClassID(void) const;
    virtual const UnicodeString* snext(UErrorCode& status);
    virtual void reset(UErrorCode& status);
    virtual int32_t count(UErrorCode& status) const;
private:
    int32_t fLen;
    int32_t fPos;
    const UVector* fMetaZoneIDs;
    UVector *fLocalVector;
};

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(MetaZoneIDsEnumeration)

MetaZoneIDsEnumeration::MetaZoneIDsEnumeration()
: fLen(0), fPos(0), fMetaZoneIDs(NULL), fLocalVector(NULL) {
}

MetaZoneIDsEnumeration::MetaZoneIDsEnumeration(const UVector& mzIDs)
: fPos(0), fMetaZoneIDs(&mzIDs), fLocalVector(NULL) {
    fLen = fMetaZoneIDs->size();
}

MetaZoneIDsEnumeration::MetaZoneIDsEnumeration(UVector *mzIDs)
: fLen(0), fPos(0), fMetaZoneIDs(mzIDs), fLocalVector(mzIDs) {
    if (fMetaZoneIDs) {
        fLen = fMetaZoneIDs->size();
    }
}

const UnicodeString*
MetaZoneIDsEnumeration::snext(UErrorCode& status) {
    if (U_SUCCESS(status) && fMetaZoneIDs != NULL && fPos < fLen) {
        unistr.setTo((const UChar*)fMetaZoneIDs->elementAt(fPos++), -1);
        return &unistr;
    }
    return NULL;
}

void
MetaZoneIDsEnumeration::reset(UErrorCode& /*status*/) {
    fPos = 0;
}

int32_t
MetaZoneIDsEnumeration::count(UErrorCode& /*status*/) const {
    return fLen;
}

MetaZoneIDsEnumeration::~MetaZoneIDsEnumeration() {
    if (fLocalVector) {
        delete fLocalVector;
    }
}


// ---------------------------------------------------
// ZNameSearchHandler
// ---------------------------------------------------
class ZNameSearchHandler : public TextTrieMapSearchResultHandler {
public:
    ZNameSearchHandler(uint32_t types);
    virtual ~ZNameSearchHandler();

    UBool handleMatch(int32_t matchLength, const CharacterNode *node, UErrorCode &status);
    TimeZoneNames::MatchInfoCollection* getMatches(int32_t& maxMatchLen);

private:
    uint32_t fTypes;
    int32_t fMaxMatchLen;
    TimeZoneNames::MatchInfoCollection* fResults;
};

ZNameSearchHandler::ZNameSearchHandler(uint32_t types)
: fTypes(types), fMaxMatchLen(0), fResults(NULL) {
}

ZNameSearchHandler::~ZNameSearchHandler() {
    if (fResults != NULL) {
        delete fResults;
    }
}

UBool
ZNameSearchHandler::handleMatch(int32_t matchLength, const CharacterNode *node, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }
    if (node->hasValues()) {
        int32_t valuesCount = node->countValues();
        for (int32_t i = 0; i < valuesCount; i++) {
            tznames_ZNameInfo *nameinfo = (tznames_ZNameInfo *)node->getValue(i);
            if (nameinfo == NULL) {
                continue;
            }
            if ((nameinfo->type & fTypes) != 0) {
                // matches a requested type
                if (fResults == NULL) {
                    fResults = new TimeZoneNames::MatchInfoCollection();
                    if (fResults == NULL) {
                        status = U_MEMORY_ALLOCATION_ERROR;
                    }
                }
                if (U_SUCCESS(status)) {
                    U_ASSERT(fResults != NULL);
                    if (nameinfo->tzID) {
                        fResults->addZone(nameinfo->type, matchLength, UnicodeString(nameinfo->tzID, -1), status);
                    } else {
                        U_ASSERT(nameinfo->mzID);
                        fResults->addMetaZone(nameinfo->type, matchLength, UnicodeString(nameinfo->mzID, -1), status);
                    }
                    if (U_SUCCESS(status) && matchLength > fMaxMatchLen) {
                        fMaxMatchLen = matchLength;
                    }
                }
            }
        }
    }
    return TRUE;
}

TimeZoneNames::MatchInfoCollection*
ZNameSearchHandler::getMatches(int32_t& maxMatchLen) {
    // give the ownership to the caller
    TimeZoneNames::MatchInfoCollection* results = fResults;
    maxMatchLen = fMaxMatchLen;

    // reset
    fResults = NULL;
    fMaxMatchLen = 0;
    return results;
}

// ---------------------------------------------------
// TimeZoneNamesImpl
//
// TimeZoneNames implementation class. This is the main
// part of this module.
// ---------------------------------------------------

U_CDECL_BEGIN
/**
 * Deleter for ZNames
 */
static void U_CALLCONV
deleteZNames(void *obj) {
    if (obj != tzgnames_impl_EMPTY) {
        delete (ZNames*) obj;
    }
}

/**
 * Deleter for tznames_ZNameInfo
 */
static void U_CALLCONV
deletetznames_ZNameInfo(void *obj) {
    uprv_free(obj);
}

U_CDECL_END

TimeZoneNamesImpl::TimeZoneNamesImpl(const Locale& locale, UErrorCode& status)
: fLocale(locale),
  fZoneStrings(NULL),
  fTZNamesMap(NULL),
  fMZNamesMap(NULL),
  fNamesTrieFullyLoaded(FALSE),
  fNamesFullyLoaded(FALSE),
  fNamesTrie(TRUE, deletetznames_ZNameInfo) {
    initialize(locale, status);
}

void
TimeZoneNamesImpl::initialize(const Locale& locale, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }

    // Load zoneStrings bundle
    UErrorCode tmpsts = U_ZERO_ERROR;   // OK with fallback warning..
    fZoneStrings = ures_open(U_ICUDATA_ZONE, locale.getName(), &tmpsts);
    fZoneStrings = ures_getByKeyWithFallback(fZoneStrings, tzgnames_impl_gZoneStrings, fZoneStrings, &tmpsts);
    if (U_FAILURE(tmpsts)) {
        status = tmpsts;
        cleanup();
        return;
    }

    // Initialize hashtables holding time zone/meta zone names
    fMZNamesMap = uhash_open(uhash_hashUChars, uhash_compareUChars, NULL, &status);
    fTZNamesMap = uhash_open(uhash_hashUChars, uhash_compareUChars, NULL, &status);
    if (U_FAILURE(status)) {
        cleanup();
        return;
    }

    uhash_setValueDeleter(fMZNamesMap, deleteZNames);
    uhash_setValueDeleter(fTZNamesMap, deleteZNames);
    // no key deleters for name maps

    // preload zone strings for the default zone
    TimeZone *tz = TimeZone::createDefault();
    const UChar *tzID = ZoneMeta::getCanonicalCLDRID(*tz);
    if (tzID != NULL) {
        loadStrings(UnicodeString(tzID), status);
    }
    delete tz;

    return;
}

/*
 * This method updates the cache and must be called with a lock,
 * except initializer.
 */
void
TimeZoneNamesImpl::loadStrings(const UnicodeString& tzCanonicalID, UErrorCode& status) {
    loadTimeZoneNames(tzCanonicalID, status);
    LocalPointer<StringEnumeration> mzIDs(getAvailableMetaZoneIDs(tzCanonicalID, status));
    if (U_FAILURE(status)) { return; }
    U_ASSERT(!mzIDs.isNull());

    const UnicodeString *mzID;
    while (((mzID = mzIDs->snext(status)) != NULL) && U_SUCCESS(status)) {
        loadMetaZoneNames(*mzID, status);
    }
}

TimeZoneNamesImpl::~TimeZoneNamesImpl() {
    cleanup();
}

void
TimeZoneNamesImpl::cleanup() {
    if (fZoneStrings != NULL) {
        ures_close(fZoneStrings);
        fZoneStrings = NULL;
    }
    if (fMZNamesMap != NULL) {
        uhash_close(fMZNamesMap);
        fMZNamesMap = NULL;
    }
    if (fTZNamesMap != NULL) {
        uhash_close(fTZNamesMap);
        fTZNamesMap = NULL;
    }
}

UBool
TimeZoneNamesImpl::operator==(const TimeZoneNames& other) const {
    if (this == &other) {
        return TRUE;
    }
    // No implementation for now
    return FALSE;
}

TimeZoneNamesImpl*
TimeZoneNamesImpl::clone() const {
    UErrorCode status = U_ZERO_ERROR;
    return new TimeZoneNamesImpl(fLocale, status);
}

StringEnumeration*
TimeZoneNamesImpl::getAvailableMetaZoneIDs(UErrorCode& status) const {
    return TimeZoneNamesImpl::_getAvailableMetaZoneIDs(status);
}

// static implementation of getAvailableMetaZoneIDs(UErrorCode&)
StringEnumeration*
TimeZoneNamesImpl::_getAvailableMetaZoneIDs(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    const UVector* mzIDs = ZoneMeta::getAvailableMetazoneIDs();
    if (mzIDs == NULL) {
        return new MetaZoneIDsEnumeration();
    }
    return new MetaZoneIDsEnumeration(*mzIDs);
}

StringEnumeration*
TimeZoneNamesImpl::getAvailableMetaZoneIDs(const UnicodeString& tzID, UErrorCode& status) const {
    return TimeZoneNamesImpl::_getAvailableMetaZoneIDs(tzID, status);
}

// static implementation of getAvailableMetaZoneIDs(const UnicodeString&, UErrorCode&)
StringEnumeration*
TimeZoneNamesImpl::_getAvailableMetaZoneIDs(const UnicodeString& tzID, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    const UVector* mappings = ZoneMeta::getMetazoneMappings(tzID);
    if (mappings == NULL) {
        return new MetaZoneIDsEnumeration();
    }

    MetaZoneIDsEnumeration *senum = NULL;
    UVector* mzIDs = new UVector(NULL, uhash_compareUChars, status);
    if (mzIDs == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    if (U_SUCCESS(status)) {
        U_ASSERT(mzIDs != NULL);
        for (int32_t i = 0; U_SUCCESS(status) && i < mappings->size(); i++) {

            OlsonToMetaMappingEntry *map = (OlsonToMetaMappingEntry *)mappings->elementAt(i);
            const UChar *mzID = map->mzid;
            if (!mzIDs->contains((void *)mzID)) {
                mzIDs->addElement((void *)mzID, status);
            }
        }
        if (U_SUCCESS(status)) {
            senum = new MetaZoneIDsEnumeration(mzIDs);
        } else {
            delete mzIDs;
        }
    }
    return senum;
}

UnicodeString&
TimeZoneNamesImpl::getMetaZoneID(const UnicodeString& tzID, UDate date, UnicodeString& mzID) const {
    return TimeZoneNamesImpl::_getMetaZoneID(tzID, date, mzID);
}

// static implementation of getMetaZoneID
UnicodeString&
TimeZoneNamesImpl::_getMetaZoneID(const UnicodeString& tzID, UDate date, UnicodeString& mzID) {
    ZoneMeta::getMetazoneID(tzID, date, mzID);
    return mzID;
}

UnicodeString&
TimeZoneNamesImpl::getReferenceZoneID(const UnicodeString& mzID, const char* region, UnicodeString& tzID) const {
    return TimeZoneNamesImpl::_getReferenceZoneID(mzID, region, tzID);
}

// static implementaion of getReferenceZoneID
UnicodeString&
TimeZoneNamesImpl::_getReferenceZoneID(const UnicodeString& mzID, const char* region, UnicodeString& tzID) {
    ZoneMeta::getZoneIdByMetazone(mzID, UnicodeString(region, -1, US_INV), tzID);
    return tzID;
}

UnicodeString&
TimeZoneNamesImpl::getMetaZoneDisplayName(const UnicodeString& mzID,
                                          UTimeZoneNameType type,
                                          UnicodeString& name) const {
    name.setToBogus();  // cleanup result.
    if (mzID.isEmpty()) {
        return name;
    }

    ZNames *znames = NULL;
    TimeZoneNamesImpl *nonConstThis = const_cast<TimeZoneNamesImpl *>(this);

    {
        Mutex lock(&tzgnames_impl_gDataMutex);
        UErrorCode status = U_ZERO_ERROR;
        znames = nonConstThis->loadMetaZoneNames(mzID, status);
        if (U_FAILURE(status)) { return name; }
    }

    if (znames != NULL) {
        const UChar* s = znames->getName(type);
        if (s != NULL) {
            name.setTo(TRUE, s, -1);
        }
    }
    return name;
}

UnicodeString&
TimeZoneNamesImpl::getTimeZoneDisplayName(const UnicodeString& tzID, UTimeZoneNameType type, UnicodeString& name) const {
    name.setToBogus();  // cleanup result.
    if (tzID.isEmpty()) {
        return name;
    }

    ZNames *tznames = NULL;
    TimeZoneNamesImpl *nonConstThis = const_cast<TimeZoneNamesImpl *>(this);

    {
        Mutex lock(&tzgnames_impl_gDataMutex);
        UErrorCode status = U_ZERO_ERROR;
        tznames = nonConstThis->loadTimeZoneNames(tzID, status);
        if (U_FAILURE(status)) { return name; }
    }

    if (tznames != NULL) {
        const UChar *s = tznames->getName(type);
        if (s != NULL) {
            name.setTo(TRUE, s, -1);
        }
    }
    return name;
}

UnicodeString&
TimeZoneNamesImpl::getExemplarLocationName(const UnicodeString& tzID, UnicodeString& name) const {
    name.setToBogus();  // cleanup result.
    const UChar* locName = NULL;
    ZNames *tznames = NULL;
    TimeZoneNamesImpl *nonConstThis = const_cast<TimeZoneNamesImpl *>(this);

    {
        Mutex lock(&tzgnames_impl_gDataMutex);
        UErrorCode status = U_ZERO_ERROR;
        tznames = nonConstThis->loadTimeZoneNames(tzID, status);
        if (U_FAILURE(status)) { return name; }
    }

    if (tznames != NULL) {
        locName = tznames->getName(UTZNM_EXEMPLAR_LOCATION);
    }
    if (locName != NULL) {
        name.setTo(TRUE, locName, -1);
    }

    return name;
}


// Merge the MZ_PREFIX and mzId
static void mergeTimeZoneKey(const UnicodeString& mzID, char* result) {
    if (mzID.isEmpty()) {
        result[0] = '\0';
        return;
    }

    char mzIdChar[ZID_KEY_MAX + 1];
    int32_t keyLen;
    int32_t prefixLen = static_cast<int32_t>(uprv_strlen(tzgnames_impl_gMZPrefix));
    keyLen = mzID.extract(0, mzID.length(), mzIdChar, ZID_KEY_MAX + 1, US_INV);
    uprv_memcpy((void *)result, (void *)tzgnames_impl_gMZPrefix, prefixLen);
    uprv_memcpy((void *)(result + prefixLen), (void *)mzIdChar, keyLen);
    result[keyLen + prefixLen] = '\0';
}

/*
 * This method updates the cache and must be called with a lock
 */
ZNames*
TimeZoneNamesImpl::loadMetaZoneNames(const UnicodeString& mzID, UErrorCode& status) {
    if (U_FAILURE(status)) { return NULL; }
    U_ASSERT(mzID.length() <= ZID_KEY_MAX - MZ_PREFIX_LEN);

    UChar mzIDKey[ZID_KEY_MAX + 1];
    mzID.extract(mzIDKey, ZID_KEY_MAX + 1, status);
    U_ASSERT(U_SUCCESS(status));   // already checked length above
    mzIDKey[mzID.length()] = 0;

    void* mznames = uhash_get(fMZNamesMap, mzIDKey);
    if (mznames == NULL) {
        ZNames::ZNamesLoader loader;
        loader.loadMetaZone(fZoneStrings, mzID, status);
        mznames = ZNames::createMetaZoneAndPutInCache(fMZNamesMap, loader.getNames(), mzID, status);
        if (U_FAILURE(status)) { return NULL; }
    }

    if (mznames != tzgnames_impl_EMPTY) {
        return (ZNames*)mznames;
    } else {
        return NULL;
    }
}

/*
 * This method updates the cache and must be called with a lock
 */
ZNames*
TimeZoneNamesImpl::loadTimeZoneNames(const UnicodeString& tzID, UErrorCode& status) {
    if (U_FAILURE(status)) { return NULL; }
    U_ASSERT(tzID.length() <= ZID_KEY_MAX);

    UChar tzIDKey[ZID_KEY_MAX + 1];
    int32_t tzIDKeyLen = tzID.extract(tzIDKey, ZID_KEY_MAX + 1, status);
    U_ASSERT(U_SUCCESS(status));   // already checked length above
    tzIDKey[tzIDKeyLen] = 0;

    void *tznames = uhash_get(fTZNamesMap, tzIDKey);
    if (tznames == NULL) {
        ZNames::ZNamesLoader loader;
        loader.loadTimeZone(fZoneStrings, tzID, status);
        tznames = ZNames::createTimeZoneAndPutInCache(fTZNamesMap, loader.getNames(), tzID, status);
        if (U_FAILURE(status)) { return NULL; }
    }

    // tznames is never tzgnames_impl_EMPTY
    return (ZNames*)tznames;
}

TimeZoneNames::MatchInfoCollection*
TimeZoneNamesImpl::find(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const {
    ZNameSearchHandler handler(types);
    TimeZoneNames::MatchInfoCollection* matches;
    TimeZoneNamesImpl* nonConstThis = const_cast<TimeZoneNamesImpl*>(this);

    // Synchronize so that data is not loaded multiple times.
    // TODO: Consider more fine-grained synchronization.
    {
        Mutex lock(&tzgnames_impl_gDataMutex);

        // First try of lookup.
        matches = doFind(handler, text, start, status);
        if (U_FAILURE(status)) { return NULL; }
        if (matches != NULL) {
            return matches;
        }

        // All names are not yet loaded into the trie.
        // We may have loaded names for formatting several time zones,
        // and might be parsing one of those.
        // Populate the parsing trie from all of the already-loaded names.
        nonConstThis->addAllNamesIntoTrie(status);

        // Second try of lookup.
        matches = doFind(handler, text, start, status);
        if (U_FAILURE(status)) { return NULL; }
        if (matches != NULL) {
            return matches;
        }

        // There are still some names we haven't loaded into the trie yet.
        // Load everything now.
        nonConstThis->internalLoadAllDisplayNames(status);
        nonConstThis->addAllNamesIntoTrie(status);
        nonConstThis->fNamesTrieFullyLoaded = TRUE;
        if (U_FAILURE(status)) { return NULL; }

        // Third try: we must return this one.
        return doFind(handler, text, start, status);
    }
}

TimeZoneNames::MatchInfoCollection*
TimeZoneNamesImpl::doFind(ZNameSearchHandler& handler,
        const UnicodeString& text, int32_t start, UErrorCode& status) const {

    fNamesTrie.search(text, start, (TextTrieMapSearchResultHandler *)&handler, status);
    if (U_FAILURE(status)) { return NULL; }

    int32_t maxLen = 0;
    TimeZoneNames::MatchInfoCollection* matches = handler.getMatches(maxLen);
    if (matches != NULL && ((maxLen == (text.length() - start)) || fNamesTrieFullyLoaded)) {
        // perfect match, or no more names available
        return matches;
    }
    delete matches;
    return NULL;
}

// Caller must synchronize.
void TimeZoneNamesImpl::addAllNamesIntoTrie(UErrorCode& status) {
    if (U_FAILURE(status)) return;
    int32_t pos;
    const UHashElement* element;

    pos = UHASH_FIRST;
    while ((element = uhash_nextElement(fMZNamesMap, &pos)) != NULL) {
        if (element->value.pointer == tzgnames_impl_EMPTY) { continue; }
        UChar* mzID = (UChar*) element->key.pointer;
        ZNames* znames = (ZNames*) element->value.pointer;
        znames->addAsMetaZoneIntoTrie(mzID, fNamesTrie, status);
        if (U_FAILURE(status)) { return; }
    }

    pos = UHASH_FIRST;
    while ((element = uhash_nextElement(fTZNamesMap, &pos)) != NULL) {
        if (element->value.pointer == tzgnames_impl_EMPTY) { continue; }
        UChar* tzID = (UChar*) element->key.pointer;
        ZNames* znames = (ZNames*) element->value.pointer;
        znames->addAsTimeZoneIntoTrie(tzID, fNamesTrie, status);
        if (U_FAILURE(status)) { return; }
    }
}

U_CDECL_BEGIN
static void U_CALLCONV
deleteZNamesLoader(void* obj) {
    if (obj == tzgnames_impl_DUMMY_LOADER) { return; }
    const ZNames::ZNamesLoader* loader = (const ZNames::ZNamesLoader*) obj;
    delete loader;
}
U_CDECL_END

struct TimeZoneNamesImpl::ZoneStringsLoader : public ResourceSink {
    TimeZoneNamesImpl& tzn;
    UHashtable* keyToLoader;

    ZoneStringsLoader(TimeZoneNamesImpl& _tzn, UErrorCode& status)
            : tzn(_tzn) {
        keyToLoader = uhash_open(uhash_hashChars, uhash_compareChars, NULL, &status);
        if (U_FAILURE(status)) { return; }
        uhash_setKeyDeleter(keyToLoader, uprv_free);
        uhash_setValueDeleter(keyToLoader, deleteZNamesLoader);
    }
    virtual ~ZoneStringsLoader();

    void* createKey(const char* key, UErrorCode& status) {
        int32_t len = sizeof(char) * (static_cast<int32_t>(uprv_strlen(key)) + 1);
        char* newKey = (char*) uprv_malloc(len);
        if (newKey == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        uprv_memcpy(newKey, key, len);
        newKey[len-1] = '\0';
        return (void*) newKey;
    }

    UBool isMetaZone(const char* key) {
        return (uprv_strlen(key) >= MZ_PREFIX_LEN && uprv_memcmp(key, tzgnames_impl_gMZPrefix, MZ_PREFIX_LEN) == 0);
    }

    UnicodeString mzIDFromKey(const char* key) {
        return UnicodeString(key + MZ_PREFIX_LEN, static_cast<int32_t>(uprv_strlen(key)) - MZ_PREFIX_LEN, US_INV);
    }

    UnicodeString tzIDFromKey(const char* key) {
        UnicodeString tzID(key, -1, US_INV);
        // Replace all colons ':' with slashes '/'
        for (int i=0; i<tzID.length(); i++) {
            if (tzID.charAt(i) == 0x003A) {
                tzID.setCharAt(i, 0x002F);
            }
        }
        return tzID;
    }

    void load(UErrorCode& status) {
        ures_getAllItemsWithFallback(tzn.fZoneStrings, "", *this, status);
        if (U_FAILURE(status)) { return; }

        int32_t pos = UHASH_FIRST;
        const UHashElement* element;
        while ((element = uhash_nextElement(keyToLoader, &pos)) != NULL) {
            if (element->value.pointer == tzgnames_impl_DUMMY_LOADER) { continue; }
            ZNames::ZNamesLoader* loader = (ZNames::ZNamesLoader*) element->value.pointer;
            char* key = (char*) element->key.pointer;

            if (isMetaZone(key)) {
                UnicodeString mzID = mzIDFromKey(key);
                ZNames::createMetaZoneAndPutInCache(tzn.fMZNamesMap, loader->getNames(), mzID, status);
            } else {
                UnicodeString tzID = tzIDFromKey(key);
                ZNames::createTimeZoneAndPutInCache(tzn.fTZNamesMap, loader->getNames(), tzID, status);
            }
            if (U_FAILURE(status)) { return; }
        }
    }

    void consumeNamesTable(const char *key, ResourceValue &value, UBool noFallback,
            UErrorCode &status) {
        if (U_FAILURE(status)) { return; }

        void* loader = uhash_get(keyToLoader, key);
        if (loader == NULL) {
            if (isMetaZone(key)) {
                UnicodeString mzID = mzIDFromKey(key);
                void* cacheVal = uhash_get(tzn.fMZNamesMap, mzID.getTerminatedBuffer());
                if (cacheVal != NULL) {
                    // We have already loaded the names for this meta zone.
                    loader = (void*) tzgnames_impl_DUMMY_LOADER;
                } else {
                    loader = (void*) new ZNames::ZNamesLoader();
                    if (loader == NULL) {
                        status = U_MEMORY_ALLOCATION_ERROR;
                        return;
                    }
                }
            } else {
                UnicodeString tzID = tzIDFromKey(key);
                void* cacheVal = uhash_get(tzn.fTZNamesMap, tzID.getTerminatedBuffer());
                if (cacheVal != NULL) {
                    // We have already loaded the names for this time zone.
                    loader = (void*) tzgnames_impl_DUMMY_LOADER;
                } else {
                    loader = (void*) new ZNames::ZNamesLoader();
                    if (loader == NULL) {
                        status = U_MEMORY_ALLOCATION_ERROR;
                        return;
                    }
                }
            }

            void* newKey = createKey(key, status);
            if (U_FAILURE(status)) {
                deleteZNamesLoader(loader);
                return;
            }

            uhash_put(keyToLoader, newKey, loader, &status);
            if (U_FAILURE(status)) { return; }
        }

        if (loader != tzgnames_impl_DUMMY_LOADER) {
            // Let the ZNamesLoader consume the names table.
            ((ZNames::ZNamesLoader*)loader)->put(key, value, noFallback, status);
        }
    }

    virtual void put(const char *key, ResourceValue &value, UBool noFallback,
            UErrorCode &status) {
        ResourceTable timeZonesTable = value.getTable(status);
        if (U_FAILURE(status)) { return; }
        for (int32_t i = 0; timeZonesTable.getKeyAndValue(i, key, value); ++i) {
            U_ASSERT(!value.isNoInheritanceMarker());
            if (value.getType() == URES_TABLE) {
                consumeNamesTable(key, value, noFallback, status);
            } else {
                // Ignore fields that aren't tables (e.g., fallbackFormat and regionFormatStandard).
                // All time zone fields are tables.
            }
            if (U_FAILURE(status)) { return; }
        }
    }
};

// Virtual destructors must be defined out of line.
TimeZoneNamesImpl::ZoneStringsLoader::~ZoneStringsLoader() {
    uhash_close(keyToLoader);
}

void TimeZoneNamesImpl::loadAllDisplayNames(UErrorCode& status) {
    if (U_FAILURE(status)) return;

    {
        Mutex lock(&tzgnames_impl_gDataMutex);
        internalLoadAllDisplayNames(status);
    }
}

void TimeZoneNamesImpl::getDisplayNames(const UnicodeString& tzID,
        const UTimeZoneNameType types[], int32_t numTypes,
        UDate date, UnicodeString dest[], UErrorCode& status) const {
    if (U_FAILURE(status)) return;

    if (tzID.isEmpty()) { return; }
    void* tznames = NULL;
    void* mznames = NULL;
    TimeZoneNamesImpl *nonConstThis = const_cast<TimeZoneNamesImpl*>(this);

    // Load the time zone strings
    {
        Mutex lock(&tzgnames_impl_gDataMutex);
        tznames = (void*) nonConstThis->loadTimeZoneNames(tzID, status);
        if (U_FAILURE(status)) { return; }
    }
    U_ASSERT(tznames != NULL);

    // Load the values into the dest array
    for (int i = 0; i < numTypes; i++) {
        UTimeZoneNameType type = types[i];
        const UChar* name = ((ZNames*)tznames)->getName(type);
        if (name == NULL) {
            if (mznames == NULL) {
                // Load the meta zone name
                UnicodeString mzID;
                getMetaZoneID(tzID, date, mzID);
                if (mzID.isEmpty()) {
                    mznames = (void*) tzgnames_impl_EMPTY;
                } else {
                    // Load the meta zone strings
                    // Mutex is scoped to the "else" statement
                    Mutex lock(&tzgnames_impl_gDataMutex);
                    mznames = (void*) nonConstThis->loadMetaZoneNames(mzID, status);
                    if (U_FAILURE(status)) { return; }
                    // Note: when the metazone doesn't exist, in Java, loadMetaZoneNames returns
                    // a dummy object instead of NULL.
                    if (mznames == NULL) {
                        mznames = (void*) tzgnames_impl_EMPTY;
                    }
                }
            }
            U_ASSERT(mznames != NULL);
            if (mznames != tzgnames_impl_EMPTY) {
                name = ((ZNames*)mznames)->getName(type);
            }
        }
        if (name != NULL) {
            dest[i].setTo(TRUE, name, -1);
        } else {
            dest[i].setToBogus();
        }
    }
}

// Caller must synchronize.
void TimeZoneNamesImpl::internalLoadAllDisplayNames(UErrorCode& status) {
    if (!fNamesFullyLoaded) {
        fNamesFullyLoaded = TRUE;

        ZoneStringsLoader loader(*this, status);
        loader.load(status);
        if (U_FAILURE(status)) { return; }

        const UnicodeString *id;

        // load strings for all zones
        StringEnumeration *tzIDs = TimeZone::createTimeZoneIDEnumeration(
            UCAL_ZONE_TYPE_CANONICAL, NULL, NULL, status);
        if (U_SUCCESS(status)) {
            while ((id = tzIDs->snext(status)) != NULL) {
                if (U_FAILURE(status)) {
                    break;
                }
                UnicodeString copy(*id);
                void* value = uhash_get(fTZNamesMap, copy.getTerminatedBuffer());
                if (value == NULL) {
                    // loadStrings also loads related metazone strings
                    loadStrings(*id, status);
                }
            }
        }
        if (tzIDs != NULL) {
            delete tzIDs;
        }
    }
}



static const UChar gEtcPrefix[]         = { 0x45, 0x74, 0x63, 0x2F }; // "Etc/"
static const int32_t gEtcPrefixLen      = 4;
static const UChar gSystemVPrefix[]     = { 0x53, 0x79, 0x73, 0x74, 0x65, 0x6D, 0x56, 0x2F }; // "SystemV/
static const int32_t gSystemVPrefixLen  = 8;
static const UChar gRiyadh8[]           = { 0x52, 0x69, 0x79, 0x61, 0x64, 0x68, 0x38 }; // "Riyadh8"
static const int32_t gRiyadh8Len       = 7;

UnicodeString& U_EXPORT2
TimeZoneNamesImpl::getDefaultExemplarLocationName(const UnicodeString& tzID, UnicodeString& name) {
    if (tzID.isEmpty() || tzID.startsWith(gEtcPrefix, gEtcPrefixLen)
        || tzID.startsWith(gSystemVPrefix, gSystemVPrefixLen) || tzID.indexOf(gRiyadh8, gRiyadh8Len, 0) > 0) {
        name.setToBogus();
        return name;
    }

    int32_t sep = tzID.lastIndexOf((UChar)0x2F /* '/' */);
    if (sep > 0 && sep + 1 < tzID.length()) {
        name.setTo(tzID, sep + 1);
        name.findAndReplace(UnicodeString((UChar)0x5f /* _ */),
                            UnicodeString((UChar)0x20 /* space */));
    } else {
        name.setToBogus();
    }
    return name;
}

// ---------------------------------------------------
// TZDBTimeZoneNames and its supporting classes
//
// TZDBTimeZoneNames is an implementation class of
// TimeZoneNames holding the IANA tz database abbreviations.
// ---------------------------------------------------

class TZDBNames : public UMemory {
public:
    virtual ~TZDBNames();

    static TZDBNames* createInstance(UResourceBundle* rb, const char* key);
    const UChar* getName(UTimeZoneNameType type) const;
    const char** getParseRegions(int32_t& numRegions) const;

protected:
    TZDBNames(const UChar** names, char** regions, int32_t numRegions);

private:
    const UChar** fNames;
    char** fRegions;
    int32_t fNumRegions;
};

TZDBNames::TZDBNames(const UChar** names, char** regions, int32_t numRegions)
    :   fNames(names),
        fRegions(regions),
        fNumRegions(numRegions) {
}

TZDBNames::~TZDBNames() {
    if (fNames != NULL) {
        uprv_free(fNames);
    }
    if (fRegions != NULL) {
        char **p = fRegions;
        for (int32_t i = 0; i < fNumRegions; p++, i++) {
            uprv_free(*p);
        }
        uprv_free(fRegions);
    }
}

TZDBNames*
TZDBNames::createInstance(UResourceBundle* rb, const char* key) {
    if (rb == NULL || key == NULL || *key == 0) {
        return NULL;
    }

    UErrorCode status = U_ZERO_ERROR;

    const UChar **names = NULL;
    char** regions = NULL;
    int32_t numRegions = 0;

    int32_t len = 0;

    UResourceBundle* rbTable = NULL;
    rbTable = ures_getByKey(rb, key, rbTable, &status);
    if (U_FAILURE(status)) {
        return NULL;
    }

    names = (const UChar **)uprv_malloc(sizeof(const UChar*) * tzgnames_impl_TZDBNAMES_KEYS_SIZE);
    UBool isEmpty = TRUE;
    if (names != NULL) {
        for (int32_t i = 0; i < tzgnames_impl_TZDBNAMES_KEYS_SIZE; i++) {
            status = U_ZERO_ERROR;
            const UChar *value = ures_getStringByKey(rbTable, tzgnames_impl_TZDBNAMES_KEYS[i], &len, &status);
            if (U_FAILURE(status) || len == 0) {
                names[i] = NULL;
            } else {
                names[i] = value;
                isEmpty = FALSE;
            }
        }
    }

    if (isEmpty) {
        if (names != NULL) {
            uprv_free(names);
        }
        return NULL;
    }

    UResourceBundle *regionsRes = ures_getByKey(rbTable, "parseRegions", NULL, &status);
    UBool regionError = FALSE;
    if (U_SUCCESS(status)) {
        numRegions = ures_getSize(regionsRes);
        if (numRegions > 0) {
            regions = (char**)uprv_malloc(sizeof(char*) * numRegions);
            if (regions != NULL) {
                char **pRegion = regions;
                for (int32_t i = 0; i < numRegions; i++, pRegion++) {
                    *pRegion = NULL;
                }
                // filling regions
                pRegion = regions;
                for (int32_t i = 0; i < numRegions; i++, pRegion++) {
                    status = U_ZERO_ERROR;
                    const UChar *uregion = ures_getStringByIndex(regionsRes, i, &len, &status);
                    if (U_FAILURE(status)) {
                        regionError = TRUE;
                        break;
                    }
                    *pRegion = (char*)uprv_malloc(sizeof(char) * (len + 1));
                    if (*pRegion == NULL) {
                        regionError = TRUE;
                        break;
                    }
                    u_UCharsToChars(uregion, *pRegion, len);
                    (*pRegion)[len] = 0;
                }
            }
        }
    }
    ures_close(regionsRes);
    ures_close(rbTable);

    if (regionError) {
        if (names != NULL) {
            uprv_free(names);
        }
        if (regions != NULL) {
            char **p = regions;
            for (int32_t i = 0; i < numRegions; p++, i++) {
                uprv_free(*p);
            }
            uprv_free(regions);
        }
        return NULL;
    }

    return new TZDBNames(names, regions, numRegions);
}

const UChar*
TZDBNames::getName(UTimeZoneNameType type) const {
    if (fNames == NULL) {
        return NULL;
    }
    const UChar *name = NULL;
    switch(type) {
    case UTZNM_SHORT_STANDARD:
        name = fNames[0];
        break;
    case UTZNM_SHORT_DAYLIGHT:
        name = fNames[1];
        break;
    default:
        name = NULL;
    }
    return name;
}

const char**
TZDBNames::getParseRegions(int32_t& numRegions) const {
    if (fRegions == NULL) {
        numRegions = 0;
    } else {
        numRegions = fNumRegions;
    }
    return (const char**)fRegions;
}

U_CDECL_BEGIN
/**
 * TZDBNameInfo stores metazone name information for the IANA abbreviations
 * in the trie
 */
typedef struct TZDBNameInfo {
    const UChar*        mzID;
    UTimeZoneNameType   type;
    UBool               ambiguousType;
    const char**        parseRegions;
    int32_t             nRegions;
} TZDBNameInfo;
U_CDECL_END


class TZDBNameSearchHandler : public TextTrieMapSearchResultHandler {
public:
    TZDBNameSearchHandler(uint32_t types, const char* region);
    virtual ~TZDBNameSearchHandler();

    UBool handleMatch(int32_t matchLength, const CharacterNode *node, UErrorCode &status);
    TimeZoneNames::MatchInfoCollection* getMatches(int32_t& maxMatchLen);

private:
    uint32_t fTypes;
    int32_t fMaxMatchLen;
    TimeZoneNames::MatchInfoCollection* fResults;
    const char* fRegion;
};

TZDBNameSearchHandler::TZDBNameSearchHandler(uint32_t types, const char* region)
: fTypes(types), fMaxMatchLen(0), fResults(NULL), fRegion(region) {
}

TZDBNameSearchHandler::~TZDBNameSearchHandler() {
    if (fResults != NULL) {
        delete fResults;
    }
}

UBool
TZDBNameSearchHandler::handleMatch(int32_t matchLength, const CharacterNode *node, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return FALSE;
    }

    TZDBNameInfo *match = NULL;
    TZDBNameInfo *defaultRegionMatch = NULL;

    if (node->hasValues()) {
        int32_t valuesCount = node->countValues();
        for (int32_t i = 0; i < valuesCount; i++) {
            TZDBNameInfo *ninfo = (TZDBNameInfo *)node->getValue(i);
            if (ninfo == NULL) {
                continue;
            }
            if ((ninfo->type & fTypes) != 0) {
                // Some tz database abbreviations are ambiguous. For example,
                // CST means either Central Standard Time or China Standard Time.
                // Unlike CLDR time zone display names, this implementation
                // does not use unique names. And TimeZoneFormat does not expect
                // multiple results returned for the same time zone type.
                // For this reason, this implementation resolve one among same
                // zone type with a same name at this level.
                if (ninfo->parseRegions == NULL) {
                    // parseRegions == null means this is the default metazone
                    // mapping for the abbreviation.
                    if (defaultRegionMatch == NULL) {
                        match = defaultRegionMatch = ninfo;
                    }
                } else {
                    UBool matchRegion = FALSE;
                    // non-default metazone mapping for an abbreviation
                    // comes with applicable regions. For example, the default
                    // metazone mapping for "CST" is America_Central,
                    // but if region is one of CN/MO/TW, "CST" is parsed
                    // as metazone China (China Standard Time).
                    for (int32_t j = 0; j < ninfo->nRegions; j++) {
                        const char *region = ninfo->parseRegions[j];
                        if (uprv_strcmp(fRegion, region) == 0) {
                            match = ninfo;
                            matchRegion = TRUE;
                            break;
                        }
                    }
                    if (matchRegion) {
                        break;
                    }
                    if (match == NULL) {
                        match = ninfo;
                    }
                }
            }
        }

        if (match != NULL) {
            UTimeZoneNameType ntype = match->type;
            // Note: Workaround for duplicated standard/daylight names
            // The tz database contains a few zones sharing a
            // same name for both standard time and daylight saving
            // time. For example, Australia/Sydney observes DST,
            // but "EST" is used for both standard and daylight.
            // When both SHORT_STANDARD and SHORT_DAYLIGHT are included
            // in the find operation, we cannot tell which one was
            // actually matched.
            // TimeZoneFormat#parse returns a matched name type (standard
            // or daylight) and DateFormat implementation uses the info to
            // to adjust actual time. To avoid false type information,
            // this implementation replaces the name type with SHORT_GENERIC.
            if (match->ambiguousType
                    && (ntype == UTZNM_SHORT_STANDARD || ntype == UTZNM_SHORT_DAYLIGHT)
                    && (fTypes & UTZNM_SHORT_STANDARD) != 0
                    && (fTypes & UTZNM_SHORT_DAYLIGHT) != 0) {
                ntype = UTZNM_SHORT_GENERIC;
            }

            if (fResults == NULL) {
                fResults = new TimeZoneNames::MatchInfoCollection();
                if (fResults == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                }
            }
            if (U_SUCCESS(status)) {
                U_ASSERT(fResults != NULL);
                U_ASSERT(match->mzID != NULL);
                fResults->addMetaZone(ntype, matchLength, UnicodeString(match->mzID, -1), status);
                if (U_SUCCESS(status) && matchLength > fMaxMatchLen) {
                    fMaxMatchLen = matchLength;
                }
            }
        }
    }
    return TRUE;
}

TimeZoneNames::MatchInfoCollection*
TZDBNameSearchHandler::getMatches(int32_t& maxMatchLen) {
    // give the ownership to the caller
    TimeZoneNames::MatchInfoCollection* results = fResults;
    maxMatchLen = fMaxMatchLen;

    // reset
    fResults = NULL;
    fMaxMatchLen = 0;
    return results;
}

U_CDECL_BEGIN
/**
 * Deleter for TZDBNames
 */
static void U_CALLCONV
deleteTZDBNames(void *obj) {
    if (obj != tzgnames_impl_EMPTY) {
        delete (TZDBNames *)obj;
    }
}

static void U_CALLCONV initTZDBNamesMap(UErrorCode &status) {
    tzgnames_impl_gTZDBNamesMap = uhash_open(uhash_hashUChars, uhash_compareUChars, NULL, &status);
    if (U_FAILURE(status)) {
        tzgnames_impl_gTZDBNamesMap = NULL;
        return;
    }
    // no key deleters for tzdb name maps
    uhash_setValueDeleter(tzgnames_impl_gTZDBNamesMap, deleteTZDBNames);
    ucln_i18n_registerCleanup(UCLN_I18N_TZDBTIMEZONENAMES, tzdbTimeZoneNames_cleanup);
}

/**
 * Deleter for TZDBNameInfo
 */
static void U_CALLCONV
deleteTZDBNameInfo(void *obj) {
    if (obj != NULL) {
        uprv_free(obj);
    }
}

static void U_CALLCONV prepareFind(UErrorCode &status) {
    if (U_FAILURE(status)) {
        return;
    }
    tzgnames_impl_gTZDBNamesTrie = new TextTrieMap(TRUE, deleteTZDBNameInfo);
    if (tzgnames_impl_gTZDBNamesTrie == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    const UnicodeString *mzID;
    StringEnumeration *mzIDs = TimeZoneNamesImpl::_getAvailableMetaZoneIDs(status);
    if (U_SUCCESS(status)) {
        while ((mzID = mzIDs->snext(status)) != 0 && U_SUCCESS(status)) {
            const TZDBNames *names = TZDBTimeZoneNames::getMetaZoneNames(*mzID, status);
            if (U_FAILURE(status)) {
                break;
            }
            if (names == NULL) {
                continue;
            }
            const UChar *std = names->getName(UTZNM_SHORT_STANDARD);
            const UChar *dst = names->getName(UTZNM_SHORT_DAYLIGHT);
            if (std == NULL && dst == NULL) {
                continue;
            }
            int32_t numRegions = 0;
            const char **parseRegions = names->getParseRegions(numRegions);

            // The tz database contains a few zones sharing a
            // same name for both standard time and daylight saving
            // time. For example, Australia/Sydney observes DST,
            // but "EST" is used for both standard and daylight.
            // we need to store the information for later processing.
            UBool ambiguousType = (std != NULL && dst != NULL && u_strcmp(std, dst) == 0);

            const UChar *uMzID = ZoneMeta::findMetaZoneID(*mzID);
            if (std != NULL) {
                TZDBNameInfo *stdInf = (TZDBNameInfo *)uprv_malloc(sizeof(TZDBNameInfo));
                if (stdInf == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                    break;
                }
                stdInf->mzID = uMzID;
                stdInf->type = UTZNM_SHORT_STANDARD;
                stdInf->ambiguousType = ambiguousType;
                stdInf->parseRegions = parseRegions;
                stdInf->nRegions = numRegions;
                tzgnames_impl_gTZDBNamesTrie->put(std, stdInf, status);
            }
            if (U_SUCCESS(status) && dst != NULL) {
                TZDBNameInfo *dstInf = (TZDBNameInfo *)uprv_malloc(sizeof(TZDBNameInfo));
                if (dstInf == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                    break;
                }
                dstInf->mzID = uMzID;
                dstInf->type = UTZNM_SHORT_DAYLIGHT;
                dstInf->ambiguousType = ambiguousType;
                dstInf->parseRegions = parseRegions;
                dstInf->nRegions = numRegions;
                tzgnames_impl_gTZDBNamesTrie->put(dst, dstInf, status);
            }
        }
    }
    delete mzIDs;

    if (U_FAILURE(status)) {
        delete tzgnames_impl_gTZDBNamesTrie;
        tzgnames_impl_gTZDBNamesTrie = NULL;
        return;
    }

    ucln_i18n_registerCleanup(UCLN_I18N_TZDBTIMEZONENAMES, tzdbTimeZoneNames_cleanup);
}

U_CDECL_END

TZDBTimeZoneNames::TZDBTimeZoneNames(const Locale& locale)
: fLocale(locale) {
    UBool useWorld = TRUE;
    const char* region = fLocale.getCountry();
    int32_t regionLen = static_cast<int32_t>(uprv_strlen(region));
    if (regionLen == 0) {
        UErrorCode status = U_ZERO_ERROR;
        CharString loc;
        {
            CharStringByteSink sink(&loc);
            ulocimp_addLikelySubtags(fLocale.getName(), sink, &status);
        }
        regionLen = uloc_getCountry(loc.data(), fRegion, sizeof(fRegion), &status);
        if (U_SUCCESS(status) && regionLen < (int32_t)sizeof(fRegion)) {
            useWorld = FALSE;
        }
    } else if (regionLen < (int32_t)sizeof(fRegion)) {
        uprv_strcpy(fRegion, region);
        useWorld = FALSE;
    }
    if (useWorld) {
        uprv_strcpy(fRegion, "001");
    }
}

TZDBTimeZoneNames::~TZDBTimeZoneNames() {
}

UBool
TZDBTimeZoneNames::operator==(const TimeZoneNames& other) const {
    if (this == &other) {
        return TRUE;
    }
    // No implementation for now
    return FALSE;
}

TZDBTimeZoneNames*
TZDBTimeZoneNames::clone() const {
    return new TZDBTimeZoneNames(fLocale);
}

StringEnumeration*
TZDBTimeZoneNames::getAvailableMetaZoneIDs(UErrorCode& status) const {
    return TimeZoneNamesImpl::_getAvailableMetaZoneIDs(status);
}

StringEnumeration*
TZDBTimeZoneNames::getAvailableMetaZoneIDs(const UnicodeString& tzID, UErrorCode& status) const {
    return TimeZoneNamesImpl::_getAvailableMetaZoneIDs(tzID, status);
}

UnicodeString&
TZDBTimeZoneNames::getMetaZoneID(const UnicodeString& tzID, UDate date, UnicodeString& mzID) const {
    return TimeZoneNamesImpl::_getMetaZoneID(tzID, date, mzID);
}

UnicodeString&
TZDBTimeZoneNames::getReferenceZoneID(const UnicodeString& mzID, const char* region, UnicodeString& tzID) const {
    return TimeZoneNamesImpl::_getReferenceZoneID(mzID, region, tzID);
}

UnicodeString&
TZDBTimeZoneNames::getMetaZoneDisplayName(const UnicodeString& mzID,
                                          UTimeZoneNameType type,
                                          UnicodeString& name) const {
    name.setToBogus();
    if (mzID.isEmpty()) {
        return name;
    }

    UErrorCode status = U_ZERO_ERROR;
    const TZDBNames *tzdbNames = TZDBTimeZoneNames::getMetaZoneNames(mzID, status);
    if (U_SUCCESS(status)) {
        if (tzdbNames != NULL) {
            const UChar *s = tzdbNames->getName(type);
            if (s != NULL) {
                name.setTo(TRUE, s, -1);
            }
        }
    }

    return name;
}

UnicodeString&
TZDBTimeZoneNames::getTimeZoneDisplayName(const UnicodeString& /* tzID */, UTimeZoneNameType /* type */, UnicodeString& name) const {
    // No abbreviations associated a zone directly for now.
    name.setToBogus();
    return name;
}

TZDBTimeZoneNames::MatchInfoCollection*
TZDBTimeZoneNames::find(const UnicodeString& text, int32_t start, uint32_t types, UErrorCode& status) const {
    umtx_initOnce(tzgnames_impl_gTZDBNamesTrieInitOnce, &prepareFind, status);
    if (U_FAILURE(status)) {
        return NULL;
    }

    TZDBNameSearchHandler handler(types, fRegion);
    tzgnames_impl_gTZDBNamesTrie->search(text, start, (TextTrieMapSearchResultHandler *)&handler, status);
    if (U_FAILURE(status)) {
        return NULL;
    }
    int32_t maxLen = 0;
    return handler.getMatches(maxLen);
}

const TZDBNames*
TZDBTimeZoneNames::getMetaZoneNames(const UnicodeString& mzID, UErrorCode& status) {
    umtx_initOnce(tzgnames_impl_gTZDBNamesMapInitOnce, &initTZDBNamesMap, status);
    if (U_FAILURE(status)) {
        return NULL;
    }

    TZDBNames* tzdbNames = NULL;

    UChar mzIDKey[ZID_KEY_MAX + 1];
    mzID.extract(mzIDKey, ZID_KEY_MAX + 1, status);
    U_ASSERT(status == U_ZERO_ERROR);   // already checked length above
    mzIDKey[mzID.length()] = 0;

    static UMutex tzgnames_impl_gTZDBNamesMapLock;
    umtx_lock(&tzgnames_impl_gTZDBNamesMapLock);
    {
        void *cacheVal = uhash_get(tzgnames_impl_gTZDBNamesMap, mzIDKey);
        if (cacheVal == NULL) {
            UResourceBundle *zoneStringsRes = ures_openDirect(U_ICUDATA_ZONE, "tzdbNames", &status);
            zoneStringsRes = ures_getByKey(zoneStringsRes, tzgnames_impl_gZoneStrings, zoneStringsRes, &status);
            if (U_SUCCESS(status)) {
                char key[ZID_KEY_MAX + 1];
                mergeTimeZoneKey(mzID, key);
                tzdbNames = TZDBNames::createInstance(zoneStringsRes, key);

                if (tzdbNames == NULL) {
                    cacheVal = (void *)tzgnames_impl_EMPTY;
                } else {
                    cacheVal = tzdbNames;
                }
                // Use the persistent ID as the resource key, so we can
                // avoid duplications.
                // TODO: Is there a more efficient way, like intern() in Java?
                void* newKey = (void*) ZoneMeta::findMetaZoneID(mzID);
                if (newKey != NULL) {
                    uhash_put(tzgnames_impl_gTZDBNamesMap, newKey, cacheVal, &status);
                    if (U_FAILURE(status)) {
                        if (tzdbNames != NULL) {
                            delete tzdbNames;
                            tzdbNames = NULL;
                        }
                    }
                } else {
                    // Should never happen with a valid input
                    if (tzdbNames != NULL) {
                        // It's not possible that we get a valid tzdbNames with unknown ID.
                        // But just in case..
                        delete tzdbNames;
                        tzdbNames = NULL;
                    }
                }
            }
            ures_close(zoneStringsRes);
        } else if (cacheVal != tzgnames_impl_EMPTY) {
            tzdbNames = (TZDBNames *)cacheVal;
        }
    }
    umtx_unlock(&tzgnames_impl_gTZDBNamesMapLock);

    return tzdbNames;
}

U_NAMESPACE_END


#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2012, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

  // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING







U_CDECL_BEGIN
// UComparator function for sorting start times
static int32_t U_CALLCONV
compareDates(const void * /*context*/, const void *left, const void *right) {
    UDate l = *((UDate*)left);
    UDate r = *((UDate*)right);
    int32_t res = l < r ? -1 : (l == r ? 0 : 1);
    return res;
}
U_CDECL_END

U_NAMESPACE_BEGIN

TimeZoneRule::TimeZoneRule(const UnicodeString& name, int32_t rawOffset, int32_t dstSavings)
: UObject(), fName(name), fRawOffset(rawOffset), fDSTSavings(dstSavings) {
}

TimeZoneRule::TimeZoneRule(const TimeZoneRule& source)
: UObject(source), fName(source.fName), fRawOffset(source.fRawOffset), fDSTSavings(source.fDSTSavings) {
}

TimeZoneRule::~TimeZoneRule() {
}

TimeZoneRule&
TimeZoneRule::operator=(const TimeZoneRule& right) {
    if (this != &right) {
        fName = right.fName;
        fRawOffset = right.fRawOffset;
        fDSTSavings = right.fDSTSavings;
    }
    return *this;
}

UBool
TimeZoneRule::operator==(const TimeZoneRule& that) const {
    return ((this == &that) ||
            (typeid(*this) == typeid(that) &&
            fName == that.fName &&
            fRawOffset == that.fRawOffset &&
            fDSTSavings == that.fDSTSavings));
}

UBool
TimeZoneRule::operator!=(const TimeZoneRule& that) const {
    return !operator==(that);
}

UnicodeString&
TimeZoneRule::getName(UnicodeString& name) const {
    name = fName;
    return name;
}

int32_t
TimeZoneRule::getRawOffset(void) const {
    return fRawOffset;
}

int32_t
TimeZoneRule::getDSTSavings(void) const {
    return fDSTSavings;
}

UBool
TimeZoneRule::isEquivalentTo(const TimeZoneRule& other) const {
    return ((this == &other) ||
            (typeid(*this) == typeid(other) &&
            fRawOffset == other.fRawOffset &&
            fDSTSavings == other.fDSTSavings));
}


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(InitialTimeZoneRule)

InitialTimeZoneRule::InitialTimeZoneRule(const UnicodeString& name,
                                         int32_t rawOffset,
                                         int32_t dstSavings)
: TimeZoneRule(name, rawOffset, dstSavings) {
}

InitialTimeZoneRule::InitialTimeZoneRule(const InitialTimeZoneRule& source)
: TimeZoneRule(source) {
}

InitialTimeZoneRule::~InitialTimeZoneRule() {
}

InitialTimeZoneRule*
InitialTimeZoneRule::clone() const {
    return new InitialTimeZoneRule(*this);
}

InitialTimeZoneRule&
InitialTimeZoneRule::operator=(const InitialTimeZoneRule& right) {
    if (this != &right) {
        TimeZoneRule::operator=(right);
    }
    return *this;
}

UBool
InitialTimeZoneRule::operator==(const TimeZoneRule& that) const {
    return ((this == &that) ||
            (typeid(*this) == typeid(that) &&
            TimeZoneRule::operator==(that)));
}

UBool
InitialTimeZoneRule::operator!=(const TimeZoneRule& that) const {
    return !operator==(that);
}

UBool
InitialTimeZoneRule::isEquivalentTo(const TimeZoneRule& other) const {
    if (this == &other) {
        return TRUE;
    }
    if (typeid(*this) != typeid(other) || TimeZoneRule::isEquivalentTo(other) == FALSE) {
        return FALSE;
    }
    return TRUE;
}

UBool
InitialTimeZoneRule::getFirstStart(int32_t /*prevRawOffset*/,
                                  int32_t /*prevDSTSavings*/,
                                  UDate& /*result*/) const {
    return FALSE;
}

UBool
InitialTimeZoneRule::getFinalStart(int32_t /*prevRawOffset*/,
                                  int32_t /*prevDSTSavings*/,
                                  UDate& /*result*/) const {
    return FALSE;
}

UBool
InitialTimeZoneRule::getNextStart(UDate /*base*/,
                                 int32_t /*prevRawOffset*/,
                                 int32_t /*prevDSTSavings*/,
                                 UBool /*inclusive*/,
                                 UDate& /*result*/) const {
    return FALSE;
}

UBool
InitialTimeZoneRule::getPreviousStart(UDate /*base*/,
                                     int32_t /*prevRawOffset*/,
                                     int32_t /*prevDSTSavings*/,
                                     UBool /*inclusive*/,
                                     UDate& /*result*/) const {
    return FALSE;
}


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(AnnualTimeZoneRule)

const int32_t AnnualTimeZoneRule::MAX_YEAR = 0x7FFFFFFF; /* max signed int32 */

AnnualTimeZoneRule::AnnualTimeZoneRule(const UnicodeString& name,
                                       int32_t rawOffset,
                                       int32_t dstSavings, 
                                       const DateTimeRule& dateTimeRule,
                                       int32_t startYear,
                                       int32_t endYear)
: TimeZoneRule(name, rawOffset, dstSavings), fDateTimeRule(new DateTimeRule(dateTimeRule)),
  fStartYear(startYear), fEndYear(endYear) {
}

AnnualTimeZoneRule::AnnualTimeZoneRule(const UnicodeString& name,
                                       int32_t rawOffset,
                                       int32_t dstSavings, 
                                       DateTimeRule* dateTimeRule,
                                       int32_t startYear,
                                       int32_t endYear)
: TimeZoneRule(name, rawOffset, dstSavings), fDateTimeRule(dateTimeRule),
  fStartYear(startYear), fEndYear(endYear) {
}

AnnualTimeZoneRule::AnnualTimeZoneRule(const AnnualTimeZoneRule& source)
: TimeZoneRule(source), fDateTimeRule(new DateTimeRule(*(source.fDateTimeRule))),
  fStartYear(source.fStartYear), fEndYear(source.fEndYear) {
}

AnnualTimeZoneRule::~AnnualTimeZoneRule() {
    delete fDateTimeRule;
}

AnnualTimeZoneRule*
AnnualTimeZoneRule::clone(void) const {
    return new AnnualTimeZoneRule(*this);
}

AnnualTimeZoneRule&
AnnualTimeZoneRule::operator=(const AnnualTimeZoneRule& right) {
    if (this != &right) {
        TimeZoneRule::operator=(right);
        delete fDateTimeRule;
        fDateTimeRule = right.fDateTimeRule->clone();
        fStartYear = right.fStartYear;
        fEndYear = right.fEndYear;
    }
    return *this;
}

UBool
AnnualTimeZoneRule::operator==(const TimeZoneRule& that) const {
    if (this == &that) {
        return TRUE;
    }
    if (typeid(*this) != typeid(that)) {
        return FALSE;
    }
    AnnualTimeZoneRule *atzr = (AnnualTimeZoneRule*)&that;
    return (*fDateTimeRule == *(atzr->fDateTimeRule) &&
            fStartYear == atzr->fStartYear &&
            fEndYear == atzr->fEndYear);
}

UBool
AnnualTimeZoneRule::operator!=(const TimeZoneRule& that) const {
    return !operator==(that);
}

const DateTimeRule*
AnnualTimeZoneRule::getRule() const {
    return fDateTimeRule;
}

int32_t
AnnualTimeZoneRule::getStartYear() const {
    return fStartYear;
}

int32_t
AnnualTimeZoneRule::getEndYear() const {
    return fEndYear;
}

UBool
AnnualTimeZoneRule::getStartInYear(int32_t year,
                                   int32_t prevRawOffset,
                                   int32_t prevDSTSavings,
                                   UDate &result) const {
    if (year < fStartYear || year > fEndYear) {
        return FALSE;
    }
    double ruleDay;
    DateTimeRule::DateRuleType type = fDateTimeRule->getDateRuleType();
    if (type == DateTimeRule::DOM) {
        ruleDay = Grego::fieldsToDay(year, fDateTimeRule->getRuleMonth(), fDateTimeRule->getRuleDayOfMonth());
    } else {
        UBool after = TRUE;
        if (type == DateTimeRule::DOW) {
            // Normalize DOW rule into DOW_GEQ_DOM or DOW_LEQ_DOM
            int32_t weeks = fDateTimeRule->getRuleWeekInMonth();
            if (weeks > 0) {
                ruleDay = Grego::fieldsToDay(year, fDateTimeRule->getRuleMonth(), 1);
                ruleDay += 7 * (weeks - 1);
            } else {
                after = FALSE;
                ruleDay = Grego::fieldsToDay(year, fDateTimeRule->getRuleMonth(),
                    Grego::monthLength(year, fDateTimeRule->getRuleMonth()));
                ruleDay += 7 * (weeks + 1);
           }
        } else {
            int32_t month = fDateTimeRule->getRuleMonth();
            int32_t dom = fDateTimeRule->getRuleDayOfMonth();
            if (type == DateTimeRule::DOW_LEQ_DOM) {
                after = FALSE;
                // Handle Feb <=29
                if (month == UCAL_FEBRUARY && dom == 29 && !Grego::isLeapYear(year)) {
                    dom--;
                }
            }
            ruleDay = Grego::fieldsToDay(year, month, dom);
        }
        int32_t dow = Grego::dayOfWeek(ruleDay);
        int32_t delta = fDateTimeRule->getRuleDayOfWeek() - dow;
        if (after) {
            delta = delta < 0 ? delta + 7 : delta;
        } else {
            delta = delta > 0 ? delta - 7 : delta;
        }
        ruleDay += delta;
    }

    result = ruleDay*U_MILLIS_PER_DAY + fDateTimeRule->getRuleMillisInDay();
    if (fDateTimeRule->getTimeRuleType() != DateTimeRule::UTC_TIME) {
        result -= prevRawOffset;
    }
    if (fDateTimeRule->getTimeRuleType() == DateTimeRule::WALL_TIME) {
        result -= prevDSTSavings;
    }
    return TRUE;
}

UBool
AnnualTimeZoneRule::isEquivalentTo(const TimeZoneRule& other) const {
    if (this == &other) {
        return TRUE;
    }
    if (typeid(*this) != typeid(other) || TimeZoneRule::isEquivalentTo(other) == FALSE) {
        return FALSE;
    }
    AnnualTimeZoneRule* that = (AnnualTimeZoneRule*)&other;
    return (*fDateTimeRule == *(that->fDateTimeRule) &&
            fStartYear == that->fStartYear &&
            fEndYear == that->fEndYear);
}

UBool
AnnualTimeZoneRule::getFirstStart(int32_t prevRawOffset,
                                  int32_t prevDSTSavings,
                                  UDate& result) const {
    return getStartInYear(fStartYear, prevRawOffset, prevDSTSavings, result);
}

UBool
AnnualTimeZoneRule::getFinalStart(int32_t prevRawOffset,
                                  int32_t prevDSTSavings,
                                  UDate& result) const {
    if (fEndYear == MAX_YEAR) {
        return FALSE;
    }
    return getStartInYear(fEndYear, prevRawOffset, prevDSTSavings, result);
}

UBool
AnnualTimeZoneRule::getNextStart(UDate base,
                                 int32_t prevRawOffset,
                                 int32_t prevDSTSavings,
                                 UBool inclusive,
                                 UDate& result) const {
    int32_t year, month, dom, dow, doy, mid;
    Grego::timeToFields(base, year, month, dom, dow, doy, mid);
    if (year < fStartYear) {
        return getFirstStart(prevRawOffset, prevDSTSavings, result);
    }
    UDate tmp;
    if (getStartInYear(year, prevRawOffset, prevDSTSavings, tmp)) {
        if (tmp < base || (!inclusive && (tmp == base))) {
            // Return the next one
            return getStartInYear(year + 1, prevRawOffset, prevDSTSavings, result);
        } else {
            result = tmp;
            return TRUE;
        }
    }
    return FALSE;
}

UBool
AnnualTimeZoneRule::getPreviousStart(UDate base,
                                     int32_t prevRawOffset,
                                     int32_t prevDSTSavings,
                                     UBool inclusive,
                                     UDate& result) const {
    int32_t year, month, dom, dow, doy, mid;
    Grego::timeToFields(base, year, month, dom, dow, doy, mid);
    if (year > fEndYear) {
        return getFinalStart(prevRawOffset, prevDSTSavings, result);
    }
    UDate tmp;
    if (getStartInYear(year, prevRawOffset, prevDSTSavings, tmp)) {
        if (tmp > base || (!inclusive && (tmp == base))) {
            // Return the previous one
            return getStartInYear(year - 1, prevRawOffset, prevDSTSavings, result);
        } else {
            result = tmp;
            return TRUE;
        }
    }
    return FALSE;
}

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(TimeArrayTimeZoneRule)

TimeArrayTimeZoneRule::TimeArrayTimeZoneRule(const UnicodeString& name,
                                             int32_t rawOffset,
                                             int32_t dstSavings,
                                             const UDate* startTimes,
                                             int32_t numStartTimes,
                                             DateTimeRule::TimeRuleType timeRuleType)
: TimeZoneRule(name, rawOffset, dstSavings), fTimeRuleType(timeRuleType),
  fStartTimes(NULL) {
    UErrorCode status = U_ZERO_ERROR;
    initStartTimes(startTimes, numStartTimes, status);
    //TODO - status?
}


TimeArrayTimeZoneRule::TimeArrayTimeZoneRule(const TimeArrayTimeZoneRule& source)
: TimeZoneRule(source), fTimeRuleType(source.fTimeRuleType), fStartTimes(NULL) {
    UErrorCode status = U_ZERO_ERROR;
    initStartTimes(source.fStartTimes, source.fNumStartTimes, status);
    //TODO - status?
}


TimeArrayTimeZoneRule::~TimeArrayTimeZoneRule() {
    if (fStartTimes != NULL && fStartTimes != fLocalStartTimes) {
        uprv_free(fStartTimes);
    }
}

TimeArrayTimeZoneRule*
TimeArrayTimeZoneRule::clone(void) const {
    return new TimeArrayTimeZoneRule(*this);
}


TimeArrayTimeZoneRule&
TimeArrayTimeZoneRule::operator=(const TimeArrayTimeZoneRule& right) {
    if (this != &right) {
        TimeZoneRule::operator=(right);
        UErrorCode status = U_ZERO_ERROR;
        initStartTimes(right.fStartTimes, right.fNumStartTimes, status);
        //TODO - status?
        fTimeRuleType = right.fTimeRuleType;        
    }
    return *this;
}

UBool
TimeArrayTimeZoneRule::operator==(const TimeZoneRule& that) const {
    if (this == &that) {
        return TRUE;
    }
    if (typeid(*this) != typeid(that) || TimeZoneRule::operator==(that) == FALSE) {
        return FALSE;
    }
    TimeArrayTimeZoneRule *tatzr = (TimeArrayTimeZoneRule*)&that;
    if (fTimeRuleType != tatzr->fTimeRuleType ||
        fNumStartTimes != tatzr->fNumStartTimes) {
        return FALSE;
    }
    // Compare start times
    UBool res = TRUE;
    for (int32_t i = 0; i < fNumStartTimes; i++) {
        if (fStartTimes[i] != tatzr->fStartTimes[i]) {
            res = FALSE;
            break;
        }
    }
    return res;
}

UBool
TimeArrayTimeZoneRule::operator!=(const TimeZoneRule& that) const {
    return !operator==(that);
}

DateTimeRule::TimeRuleType
TimeArrayTimeZoneRule::getTimeType(void) const {
    return fTimeRuleType;
}

UBool
TimeArrayTimeZoneRule::getStartTimeAt(int32_t index, UDate& result) const {
    if (index >= fNumStartTimes || index < 0) {
        return FALSE;
    }
    result = fStartTimes[index];
    return TRUE;
}

int32_t
TimeArrayTimeZoneRule::countStartTimes(void) const {
    return fNumStartTimes;
}

UBool
TimeArrayTimeZoneRule::isEquivalentTo(const TimeZoneRule& other) const {
    if (this == &other) {
        return TRUE;
    }
    if (typeid(*this) != typeid(other) || TimeZoneRule::isEquivalentTo(other) == FALSE) {
        return FALSE;
    }
    TimeArrayTimeZoneRule* that = (TimeArrayTimeZoneRule*)&other;
    if (fTimeRuleType != that->fTimeRuleType ||
        fNumStartTimes != that->fNumStartTimes) {
        return FALSE;
    }
    // Compare start times
    UBool res = TRUE;
    for (int32_t i = 0; i < fNumStartTimes; i++) {
        if (fStartTimes[i] != that->fStartTimes[i]) {
            res = FALSE;
            break;
        }
    }
    return res;
}

UBool
TimeArrayTimeZoneRule::getFirstStart(int32_t prevRawOffset,
                                             int32_t prevDSTSavings,
                                             UDate& result) const {
    if (fNumStartTimes <= 0 || fStartTimes == NULL) {
        return FALSE;
    }
    result = getUTC(fStartTimes[0], prevRawOffset, prevDSTSavings);
    return TRUE;
}

UBool
TimeArrayTimeZoneRule::getFinalStart(int32_t prevRawOffset,
                                     int32_t prevDSTSavings,
                                     UDate& result) const {
    if (fNumStartTimes <= 0 || fStartTimes == NULL) {
        return FALSE;
    }
    result = getUTC(fStartTimes[fNumStartTimes - 1], prevRawOffset, prevDSTSavings);
    return TRUE;
}

UBool
TimeArrayTimeZoneRule::getNextStart(UDate base,
                                    int32_t prevRawOffset,
                                    int32_t prevDSTSavings,
                                    UBool inclusive,
                                    UDate& result) const {
    int32_t i = fNumStartTimes - 1;
    for (; i >= 0; i--) {
        UDate time = getUTC(fStartTimes[i], prevRawOffset, prevDSTSavings);
        if (time < base || (!inclusive && time == base)) {
            break;
        }
        result = time;
    }
    if (i == fNumStartTimes - 1) {
        return FALSE;
    }
    return TRUE;
}

UBool
TimeArrayTimeZoneRule::getPreviousStart(UDate base,
                                        int32_t prevRawOffset,
                                        int32_t prevDSTSavings,
                                        UBool inclusive,
                                        UDate& result) const {
    int32_t i = fNumStartTimes - 1;
    for (; i >= 0; i--) {
        UDate time = getUTC(fStartTimes[i], prevRawOffset, prevDSTSavings);
        if (time < base || (inclusive && time == base)) {
            result = time;
            return TRUE;
        }
    }
    return FALSE;
}


// ---- private methods ------

UBool
TimeArrayTimeZoneRule::initStartTimes(const UDate source[], int32_t size, UErrorCode& status) {
    // Free old array
    if (fStartTimes != NULL && fStartTimes != fLocalStartTimes) {
        uprv_free(fStartTimes);
    }
    // Allocate new one if needed
    if (size > TIMEARRAY_STACK_BUFFER_SIZE) {
        fStartTimes = (UDate*)uprv_malloc(sizeof(UDate)*size);
        if (fStartTimes == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            fNumStartTimes = 0;
            return FALSE;
        }
    } else {
        fStartTimes = (UDate*)fLocalStartTimes;
    }
    uprv_memcpy(fStartTimes, source, sizeof(UDate)*size);
    fNumStartTimes = size;
    // Sort dates
    uprv_sortArray(fStartTimes, fNumStartTimes, (int32_t)sizeof(UDate), compareDates, NULL, TRUE, &status);
    if (U_FAILURE(status)) {
        if (fStartTimes != NULL && fStartTimes != fLocalStartTimes) {
            uprv_free(fStartTimes);
        }
        fNumStartTimes = 0;
        return FALSE;
    }
    return TRUE;
}

UDate
TimeArrayTimeZoneRule::getUTC(UDate time, int32_t raw, int32_t dst) const {
    if (fTimeRuleType != DateTimeRule::UTC_TIME) {
        time -= raw;
    }
    if (fTimeRuleType == DateTimeRule::WALL_TIME) {
        time -= dst;
    }
    return time;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2012, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

  // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING




U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(TimeZoneTransition)

TimeZoneTransition::TimeZoneTransition(UDate time, const TimeZoneRule& from, const TimeZoneRule& to)
: UObject(), fTime(time), fFrom(from.clone()), fTo(to.clone()) {
}

TimeZoneTransition::TimeZoneTransition()
: UObject(), fTime(0), fFrom(NULL), fTo(NULL) {
}

TimeZoneTransition::TimeZoneTransition(const TimeZoneTransition& source)
: UObject(), fTime(source.fTime), fFrom(NULL), fTo(NULL) {
      if (source.fFrom != NULL) {
          fFrom = source.fFrom->clone();
      }

      if (source.fTo != NULL) {
          fTo = source.fTo->clone();
      }
}

TimeZoneTransition::~TimeZoneTransition() {
    if (fFrom != NULL) {
        delete fFrom;
    }
    if (fTo != NULL) {
        delete fTo;
    }
}

TimeZoneTransition*
TimeZoneTransition::clone(void) const {
    return new TimeZoneTransition(*this);
}

TimeZoneTransition&
TimeZoneTransition::operator=(const TimeZoneTransition& right) {
    if (this != &right) {
        fTime = right.fTime;
        setFrom(*right.fFrom);
        setTo(*right.fTo);
    }
    return *this;
}

UBool
TimeZoneTransition::operator==(const TimeZoneTransition& that) const {
    if (this == &that) {
        return TRUE;
    }
    if (typeid(*this) != typeid(that)) {
        return FALSE;
    }
    if (fTime != that.fTime) {
        return FALSE;
    }
    if ((fFrom == NULL && that.fFrom == NULL)
        || (fFrom != NULL && that.fFrom != NULL && *fFrom == *(that.fFrom))) {
        if ((fTo == NULL && that.fTo == NULL)
            || (fTo != NULL && that.fTo != NULL && *fTo == *(that.fTo))) {
            return TRUE;
        }
    }
    return FALSE;
}

UBool
TimeZoneTransition::operator!=(const TimeZoneTransition& that) const {
    return !operator==(that);
}

void
TimeZoneTransition::setTime(UDate time) {
    fTime = time;
}

void
TimeZoneTransition::setFrom(const TimeZoneRule& from) {
    if (fFrom != NULL) {
        delete fFrom;
    }
    fFrom = from.clone();
}

void
TimeZoneTransition::adoptFrom(TimeZoneRule* from) {
    if (fFrom != NULL) {
        delete fFrom;
    }
    fFrom = from;
}

void
TimeZoneTransition::setTo(const TimeZoneRule& to) {
    if (fTo != NULL) {
        delete fTo;
    }
    fTo = to.clone();
}

void
TimeZoneTransition::adoptTo(TimeZoneRule* to) {
    if (fTo != NULL) {
        delete fTo;
    }
    fTo = to;
}

UDate
TimeZoneTransition::getTime(void) const {
    return fTime;
}

const TimeZoneRule*
TimeZoneTransition::getTo(void) const {
    return fTo;
}

const TimeZoneRule*
TimeZoneTransition::getFrom(void) const {
    return fFrom;
}

U_NAMESPACE_END

#endif

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 1996-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*/

  // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING

















U_NAMESPACE_USE

static TimeZone*
_createTimeZone(const UChar* zoneID, int32_t len, UErrorCode* ec) {
    TimeZone* zone = NULL;
    if (ec != NULL && U_SUCCESS(*ec)) {
        // Note that if zoneID is invalid, we get back GMT. This odd
        // behavior is by design and goes back to the JDK. The only
        // failure we will see is a memory allocation failure.
        int32_t l = (len<0 ? u_strlen(zoneID) : len);
        UnicodeString zoneStrID;
        zoneStrID.setTo((UBool)(len < 0), zoneID, l); /* temporary read-only alias */
        zone = TimeZone::createTimeZone(zoneStrID);
        if (zone == NULL) {
            *ec = U_MEMORY_ALLOCATION_ERROR;
        }
    }
    return zone;
}

U_CAPI UEnumeration* U_EXPORT2
ucal_openTimeZoneIDEnumeration(USystemTimeZoneType zoneType, const char* region,
                                const int32_t* rawOffset, UErrorCode* ec) {
    return uenum_openFromStringEnumeration(TimeZone::createTimeZoneIDEnumeration(
        zoneType, region, rawOffset, *ec), ec);
}

U_CAPI UEnumeration* U_EXPORT2
ucal_openTimeZones(UErrorCode* ec) {
    return uenum_openFromStringEnumeration(TimeZone::createEnumeration(), ec);
}

U_CAPI UEnumeration* U_EXPORT2
ucal_openCountryTimeZones(const char* country, UErrorCode* ec) {
    return uenum_openFromStringEnumeration(TimeZone::createEnumeration(country), ec);
}

U_CAPI int32_t U_EXPORT2
ucal_getDefaultTimeZone(UChar* result, int32_t resultCapacity, UErrorCode* ec) {
    int32_t len = 0;
    if (ec != NULL && U_SUCCESS(*ec)) {
        TimeZone* zone = TimeZone::createDefault();
        if (zone == NULL) {
            *ec = U_MEMORY_ALLOCATION_ERROR;
        } else {
            UnicodeString id;
            zone->getID(id);
            delete zone;
            len = id.extract(result, resultCapacity, *ec);
        }
    }
    return len;
}

U_CAPI void U_EXPORT2
ucal_setDefaultTimeZone(const UChar* zoneID, UErrorCode* ec) {
    TimeZone* zone = _createTimeZone(zoneID, -1, ec);
    if (zone != NULL) {
        TimeZone::adoptDefault(zone);
    }
}

U_DRAFT int32_t U_EXPORT2
ucal_getHostTimeZone(UChar* result, int32_t resultCapacity, UErrorCode* ec) {
    int32_t len = 0;
    if (ec != NULL && U_SUCCESS(*ec)) {
        TimeZone *zone = TimeZone::detectHostTimeZone();
        if (zone == NULL) {
            *ec = U_MEMORY_ALLOCATION_ERROR;
        } else {
            UnicodeString id;
            zone->getID(id);
            delete zone;
            len = id.extract(result, resultCapacity, *ec);
        }
    }
    return len;
}

U_CAPI int32_t U_EXPORT2
ucal_getDSTSavings(const UChar* zoneID, UErrorCode* ec) {
    int32_t result = 0;
    TimeZone* zone = _createTimeZone(zoneID, -1, ec);
    if (U_SUCCESS(*ec)) {
        SimpleTimeZone* stz = dynamic_cast<SimpleTimeZone*>(zone);
        if (stz != NULL) {
            result = stz->getDSTSavings();
        } else {
            // Since there is no getDSTSavings on TimeZone, we use a
            // heuristic: Starting with the current time, march
            // forwards for one year, looking for DST savings.
            // Stepping by weeks is sufficient.
            UDate d = Calendar::getNow();
            for (int32_t i=0; i<53; ++i, d+=U_MILLIS_PER_DAY*7.0) {
                int32_t raw, dst;
                zone->getOffset(d, FALSE, raw, dst, *ec);
                if (U_FAILURE(*ec)) {
                    break;
                } else if (dst != 0) {
                    result = dst;
                    break;
                }
            }
        }
    }
    delete zone;
    return result;
}

U_CAPI UDate  U_EXPORT2
ucal_getNow()
{

  return Calendar::getNow();
}

#define ULOC_LOCALE_IDENTIFIER_CAPACITY (ULOC_FULLNAME_CAPACITY + 1 + ULOC_KEYWORD_AND_VALUES_CAPACITY)

U_CAPI UCalendar*  U_EXPORT2
ucal_open(  const UChar*  zoneID,
            int32_t       len,
            const char*   locale,
            UCalendarType caltype,
            UErrorCode*   status)
{

  if(U_FAILURE(*status)) return 0;

  LocalPointer<TimeZone> zone( (zoneID==NULL) ? TimeZone::createDefault()
      : _createTimeZone(zoneID, len, status), *status);

  if (U_FAILURE(*status)) {
      return NULL;
  }

  if ( caltype == UCAL_GREGORIAN ) {
      char  localeBuf[ULOC_LOCALE_IDENTIFIER_CAPACITY];
      if ( locale == NULL ) {
          locale = uloc_getDefault();
      }
      uprv_strncpy(localeBuf, locale, ULOC_LOCALE_IDENTIFIER_CAPACITY);
      uloc_setKeywordValue("calendar", "gregorian", localeBuf, ULOC_LOCALE_IDENTIFIER_CAPACITY, status);
      if (U_FAILURE(*status)) {
          return NULL;
      }
      return (UCalendar*)Calendar::createInstance(zone.orphan(), Locale(localeBuf), *status);
  }
  return (UCalendar*)Calendar::createInstance(zone.orphan(), Locale(locale), *status);
}

U_CAPI void U_EXPORT2
ucal_close(UCalendar *cal)
{

  delete (Calendar*) cal;
}

U_CAPI UCalendar* U_EXPORT2
ucal_clone(const UCalendar* cal,
           UErrorCode*      status)
{
  if(U_FAILURE(*status)) return 0;

  Calendar* res = ((Calendar*)cal)->clone();

  if(res == 0) {
    *status = U_MEMORY_ALLOCATION_ERROR;
    return 0;
  }

  return (UCalendar*) res;
}

U_CAPI void  U_EXPORT2
ucal_setTimeZone(    UCalendar*      cal,
            const    UChar*            zoneID,
            int32_t        len,
            UErrorCode *status)
{

  if(U_FAILURE(*status))
    return;

  TimeZone* zone = (zoneID==NULL) ? TimeZone::createDefault()
      : _createTimeZone(zoneID, len, status);

  if (zone != NULL) {
      ((Calendar*)cal)->adoptTimeZone(zone);
  }
}

U_CAPI int32_t U_EXPORT2
ucal_getTimeZoneID(const UCalendar *cal,
                   UChar *result,
                   int32_t resultLength,
                   UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return 0;
    }
    const TimeZone& tz = ((Calendar*)cal)->getTimeZone();
    UnicodeString id;
    tz.getID(id);
    return id.extract(result, resultLength, *status);
}

U_CAPI int32_t U_EXPORT2
ucal_getTimeZoneDisplayName(const     UCalendar*                 cal,
                    UCalendarDisplayNameType     type,
                    const char             *locale,
                    UChar*                  result,
                    int32_t                 resultLength,
                    UErrorCode*             status)
{

    if(U_FAILURE(*status)) return -1;

    const TimeZone& tz = ((Calendar*)cal)->getTimeZone();
    UnicodeString id;
    if(!(result==NULL && resultLength==0)) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        id.setTo(result, 0, resultLength);
    }

    switch(type) {
  case UCAL_STANDARD:
      tz.getDisplayName(FALSE, TimeZone::LONG, Locale(locale), id);
      break;

  case UCAL_SHORT_STANDARD:
      tz.getDisplayName(FALSE, TimeZone::SHORT, Locale(locale), id);
      break;

  case UCAL_DST:
      tz.getDisplayName(TRUE, TimeZone::LONG, Locale(locale), id);
      break;

  case UCAL_SHORT_DST:
      tz.getDisplayName(TRUE, TimeZone::SHORT, Locale(locale), id);
      break;
    }

    return id.extract(result, resultLength, *status);
}

U_CAPI UBool  U_EXPORT2
ucal_inDaylightTime(    const    UCalendar*      cal,
                    UErrorCode*     status )
{

    if(U_FAILURE(*status)) return (UBool) -1;
    return ((Calendar*)cal)->inDaylightTime(*status);
}

U_CAPI void U_EXPORT2
ucal_setGregorianChange(UCalendar *cal, UDate date, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return;
    }
    Calendar *cpp_cal = (Calendar *)cal;
    GregorianCalendar *gregocal = dynamic_cast<GregorianCalendar *>(cpp_cal);
    // Not if(gregocal == NULL) {
    // because we really want to work only with a GregorianCalendar, not with
    // its subclasses like BuddhistCalendar.
    if (cpp_cal == NULL) {
        // We normally don't check "this" pointers for NULL, but this here avoids
        // compiler-generated exception-throwing code in case cal == NULL.
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if(typeid(*cpp_cal) != typeid(GregorianCalendar)) {
        *pErrorCode = U_UNSUPPORTED_ERROR;
        return;
    }
    gregocal->setGregorianChange(date, *pErrorCode);
}

U_CAPI UDate U_EXPORT2
ucal_getGregorianChange(const UCalendar *cal, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return (UDate)0;
    }
    const Calendar *cpp_cal = (const Calendar *)cal;
    const GregorianCalendar *gregocal = dynamic_cast<const GregorianCalendar *>(cpp_cal);
    // Not if(gregocal == NULL) {
    // see comments in ucal_setGregorianChange().
    if (cpp_cal == NULL) {
        // We normally don't check "this" pointers for NULL, but this here avoids
        // compiler-generated exception-throwing code in case cal == NULL.
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return (UDate)0;
    }
    if(typeid(*cpp_cal) != typeid(GregorianCalendar)) {
        *pErrorCode = U_UNSUPPORTED_ERROR;
        return (UDate)0;
    }
    return gregocal->getGregorianChange();
}

U_CAPI int32_t U_EXPORT2
ucal_getAttribute(    const    UCalendar*              cal,
                  UCalendarAttribute      attr)
{

    switch(attr) {
  case UCAL_LENIENT:
      return ((Calendar*)cal)->isLenient();

  case UCAL_FIRST_DAY_OF_WEEK:
      return ((Calendar*)cal)->getFirstDayOfWeek();

  case UCAL_MINIMAL_DAYS_IN_FIRST_WEEK:
      return ((Calendar*)cal)->getMinimalDaysInFirstWeek();

  case UCAL_REPEATED_WALL_TIME:
      return ((Calendar*)cal)->getRepeatedWallTimeOption();

  case UCAL_SKIPPED_WALL_TIME:
      return ((Calendar*)cal)->getSkippedWallTimeOption();

  default:
      break;
    }
    return -1;
}

U_CAPI void U_EXPORT2
ucal_setAttribute(      UCalendar*              cal,
                  UCalendarAttribute      attr,
                  int32_t                 newValue)
{

    switch(attr) {
  case UCAL_LENIENT:
      ((Calendar*)cal)->setLenient((UBool)newValue);
      break;

  case UCAL_FIRST_DAY_OF_WEEK:
      ((Calendar*)cal)->setFirstDayOfWeek((UCalendarDaysOfWeek)newValue);
      break;

  case UCAL_MINIMAL_DAYS_IN_FIRST_WEEK:
      ((Calendar*)cal)->setMinimalDaysInFirstWeek((uint8_t)newValue);
      break;

  case UCAL_REPEATED_WALL_TIME:
      ((Calendar*)cal)->setRepeatedWallTimeOption((UCalendarWallTimeOption)newValue);
      break;

  case UCAL_SKIPPED_WALL_TIME:
      ((Calendar*)cal)->setSkippedWallTimeOption((UCalendarWallTimeOption)newValue);
      break;
    }
}

U_CAPI const char* U_EXPORT2
ucal_getAvailable(int32_t index)
{

    return uloc_getAvailable(index);
}

U_CAPI int32_t U_EXPORT2
ucal_countAvailable()
{

    return uloc_countAvailable();
}

U_CAPI UDate  U_EXPORT2
ucal_getMillis(    const    UCalendar*      cal,
               UErrorCode*     status)
{

    if(U_FAILURE(*status)) return (UDate) 0;

    return ((Calendar*)cal)->getTime(*status);
}

U_CAPI void  U_EXPORT2
ucal_setMillis(        UCalendar*      cal,
               UDate           dateTime,
               UErrorCode*     status )
{
    if(U_FAILURE(*status)) return;

    ((Calendar*)cal)->setTime(dateTime, *status);
}

// TBD: why does this take an UErrorCode?
U_CAPI void  U_EXPORT2
ucal_setDate(        UCalendar*        cal,
             int32_t            year,
             int32_t            month,
             int32_t            date,
             UErrorCode        *status)
{

    if(U_FAILURE(*status)) return;

    ((Calendar*)cal)->set(year, month, date);
}

// TBD: why does this take an UErrorCode?
U_CAPI void  U_EXPORT2
ucal_setDateTime(    UCalendar*        cal,
                 int32_t            year,
                 int32_t            month,
                 int32_t            date,
                 int32_t            hour,
                 int32_t            minute,
                 int32_t            second,
                 UErrorCode        *status)
{
    if(U_FAILURE(*status)) return;

    ((Calendar*)cal)->set(year, month, date, hour, minute, second);
}

U_CAPI UBool  U_EXPORT2
ucal_equivalentTo(    const UCalendar*      cal1,
                  const UCalendar*      cal2)
{

    return ((Calendar*)cal1)->isEquivalentTo(*((Calendar*)cal2));
}

U_CAPI void  U_EXPORT2
ucal_add(    UCalendar*                cal,
         UCalendarDateFields        field,
         int32_t                    amount,
         UErrorCode*                status)
{

    if(U_FAILURE(*status)) return;

    ((Calendar*)cal)->add(field, amount, *status);
}

U_CAPI void  U_EXPORT2
ucal_roll(        UCalendar*            cal,
          UCalendarDateFields field,
          int32_t                amount,
          UErrorCode*            status)
{

    if(U_FAILURE(*status)) return;

    ((Calendar*)cal)->roll(field, amount, *status);
}

U_CAPI int32_t  U_EXPORT2
ucal_get(    const    UCalendar*                cal,
         UCalendarDateFields        field,
         UErrorCode*                status )
{

    if(U_FAILURE(*status)) return -1;

    return ((Calendar*)cal)->get(field, *status);
}

U_CAPI void  U_EXPORT2
ucal_set(    UCalendar*                cal,
         UCalendarDateFields        field,
         int32_t                    value)
{

    ((Calendar*)cal)->set(field, value);
}

U_CAPI UBool  U_EXPORT2
ucal_isSet(    const    UCalendar*                cal,
           UCalendarDateFields        field)
{

    return ((Calendar*)cal)->isSet(field);
}

U_CAPI void  U_EXPORT2
ucal_clearField(    UCalendar*            cal,
                UCalendarDateFields field)
{

    ((Calendar*)cal)->clear(field);
}

U_CAPI void  U_EXPORT2
ucal_clear(UCalendar* calendar)
{

    ((Calendar*)calendar)->clear();
}

U_CAPI int32_t  U_EXPORT2
ucal_getLimit(    const    UCalendar*              cal,
              UCalendarDateFields     field,
              UCalendarLimitType      type,
              UErrorCode        *status)
{

    if(status==0 || U_FAILURE(*status)) {
        return -1;
    }

    switch(type) {
  case UCAL_MINIMUM:
      return ((Calendar*)cal)->getMinimum(field);

  case UCAL_MAXIMUM:
      return ((Calendar*)cal)->getMaximum(field);

  case UCAL_GREATEST_MINIMUM:
      return ((Calendar*)cal)->getGreatestMinimum(field);

  case UCAL_LEAST_MAXIMUM:
      return ((Calendar*)cal)->getLeastMaximum(field);

  case UCAL_ACTUAL_MINIMUM:
      return ((Calendar*)cal)->getActualMinimum(field,
          *status);

  case UCAL_ACTUAL_MAXIMUM:
      return ((Calendar*)cal)->getActualMaximum(field,
          *status);

  default:
      break;
    }
    return -1;
}

U_CAPI const char * U_EXPORT2
ucal_getLocaleByType(const UCalendar *cal, ULocDataLocaleType type, UErrorCode* status)
{
    if (cal == NULL) {
        if (U_SUCCESS(*status)) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
        }
        return NULL;
    }
    return ((Calendar*)cal)->getLocaleID(type, *status);
}

U_CAPI const char * U_EXPORT2
ucal_getTZDataVersion(UErrorCode* status)
{
    return TimeZone::getTZDataVersion(*status);
}

U_CAPI int32_t U_EXPORT2
ucal_getCanonicalTimeZoneID(const UChar* id, int32_t len,
                            UChar* result, int32_t resultCapacity, UBool *isSystemID, UErrorCode* status) {
    if(status == 0 || U_FAILURE(*status)) {
        return 0;
    }
    if (isSystemID) {
        *isSystemID = FALSE;
    }
    if (id == 0 || len == 0 || result == 0 || resultCapacity <= 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    int32_t reslen = 0;
    UnicodeString canonical;
    UBool systemID = FALSE;
    TimeZone::getCanonicalID(UnicodeString(id, len), canonical, systemID, *status);
    if (U_SUCCESS(*status)) {
        if (isSystemID) {
            *isSystemID = systemID;
        }
        reslen = canonical.extract(result, resultCapacity, *status);
    }
    return reslen;
}

U_CAPI const char * U_EXPORT2
ucal_getType(const UCalendar *cal, UErrorCode* status)
{
    if (U_FAILURE(*status)) {
        return NULL;
    }
    return ((Calendar*)cal)->getType();
}

U_CAPI UCalendarWeekdayType U_EXPORT2
ucal_getDayOfWeekType(const UCalendar *cal, UCalendarDaysOfWeek dayOfWeek, UErrorCode* status)
{
    if (U_FAILURE(*status)) {
        return UCAL_WEEKDAY;
    }
    return ((Calendar*)cal)->getDayOfWeekType(dayOfWeek, *status);
}

U_CAPI int32_t U_EXPORT2
ucal_getWeekendTransition(const UCalendar *cal, UCalendarDaysOfWeek dayOfWeek, UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return 0;
    }
    return ((Calendar*)cal)->getWeekendTransition(dayOfWeek, *status);
}

U_CAPI UBool U_EXPORT2
ucal_isWeekend(const UCalendar *cal, UDate date, UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return FALSE;
    }
    return ((Calendar*)cal)->isWeekend(date, *status);
}

U_CAPI int32_t  U_EXPORT2
ucal_getFieldDifference(UCalendar* cal, UDate target,
                        UCalendarDateFields field,
                        UErrorCode* status )
{
    if (U_FAILURE(*status)) {
        return 0;
    }
    return ((Calendar*)cal)->fieldDifference(target, field, *status);
}


static const UEnumeration ucal_defaultKeywordValues = {
    NULL,
    NULL,
    ulist_close_keyword_values_iterator,
    ulist_count_keyword_values,
    uenum_unextDefault,
    ulist_next_keyword_value,
    ulist_reset_keyword_values_iterator
};

static const char * const CAL_TYPES[] = {
        "gregorian",
        "japanese",
        "buddhist",
        "roc",
        "persian",
        "islamic-civil",
        "islamic",
        "hebrew",
        "chinese",
        "indian",
        "coptic",
        "ethiopic",
        "ethiopic-amete-alem",
        "iso8601",
        "dangi",
        "islamic-umalqura",
        "islamic-tbla",
        "islamic-rgsa",
        NULL
};

U_CAPI UEnumeration* U_EXPORT2
ucal_getKeywordValuesForLocale(const char * /* key */, const char* locale, UBool commonlyUsed, UErrorCode *status) {
    // Resolve region
    char prefRegion[ULOC_COUNTRY_CAPACITY];
    (void)ulocimp_getRegionForSupplementalData(locale, TRUE, prefRegion, sizeof(prefRegion), status);

    // Read preferred calendar values from supplementalData calendarPreference
    UResourceBundle *rb = ures_openDirect(NULL, "supplementalData", status);
    ures_getByKey(rb, "calendarPreferenceData", rb, status);
    UResourceBundle *order = ures_getByKey(rb, prefRegion, NULL, status);
    if (*status == U_MISSING_RESOURCE_ERROR && rb != NULL) {
        *status = U_ZERO_ERROR;
        order = ures_getByKey(rb, "001", NULL, status);
    }

    // Create a list of calendar type strings
    UList *values = NULL;
    if (U_SUCCESS(*status)) {
        values = ulist_createEmptyList(status);
        if (U_SUCCESS(*status)) {
            for (int i = 0; i < ures_getSize(order); i++) {
                int32_t len;
                const UChar *type = ures_getStringByIndex(order, i, &len, status);
                char *caltype = (char*)uprv_malloc(len + 1);
                if (caltype == NULL) {
                    *status = U_MEMORY_ALLOCATION_ERROR;
                    break;
                }
                u_UCharsToChars(type, caltype, len);
                *(caltype + len) = 0;

                ulist_addItemEndList(values, caltype, TRUE, status);
                if (U_FAILURE(*status)) {
                    break;
                }
            }

            if (U_SUCCESS(*status) && !commonlyUsed) {
                // If not commonlyUsed, add other available values
                for (int32_t i = 0; CAL_TYPES[i] != NULL; i++) {
                    if (!ulist_containsString(values, CAL_TYPES[i], (int32_t)uprv_strlen(CAL_TYPES[i]))) {
                        ulist_addItemEndList(values, CAL_TYPES[i], FALSE, status);
                        if (U_FAILURE(*status)) {
                            break;
                        }
                    }
                }
            }
            if (U_FAILURE(*status)) {
                ulist_deleteList(values);
                values = NULL;
            }
        }
    }

    ures_close(order);
    ures_close(rb);

    if (U_FAILURE(*status) || values == NULL) {
        return NULL;
    }

    // Create string enumeration
    UEnumeration *en = (UEnumeration*)uprv_malloc(sizeof(UEnumeration));
    if (en == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        ulist_deleteList(values);
        return NULL;
    }
    ulist_resetList(values);
    memcpy(en, &ucal_defaultKeywordValues, sizeof(UEnumeration));
    en->context = values;
    return en;
}

U_CAPI UBool U_EXPORT2
ucal_getTimeZoneTransitionDate(const UCalendar* cal, UTimeZoneTransitionType type,
                               UDate* transition, UErrorCode* status)
{
    if (U_FAILURE(*status)) {
        return FALSE;
    }
    UDate base = ((Calendar*)cal)->getTime(*status);
    const TimeZone& tz = ((Calendar*)cal)->getTimeZone();
    const BasicTimeZone * btz = dynamic_cast<const BasicTimeZone *>(&tz);
    if (btz != NULL && U_SUCCESS(*status)) {
        TimeZoneTransition tzt;
        UBool inclusive = (type == UCAL_TZ_TRANSITION_NEXT_INCLUSIVE || type == UCAL_TZ_TRANSITION_PREVIOUS_INCLUSIVE);
        UBool result = (type == UCAL_TZ_TRANSITION_NEXT || type == UCAL_TZ_TRANSITION_NEXT_INCLUSIVE)?
                        btz->getNextTransition(base, inclusive, tzt):
                        btz->getPreviousTransition(base, inclusive, tzt);
        if (result) {
            *transition = tzt.getTime();
            return TRUE;
        }
    }
    return FALSE;
}

U_CAPI int32_t U_EXPORT2
ucal_getWindowsTimeZoneID(const UChar* id, int32_t len, UChar* winid, int32_t winidCapacity, UErrorCode* status) {
    if (U_FAILURE(*status)) {
        return 0;
    }

    int32_t resultLen = 0;
    UnicodeString resultWinID;

    TimeZone::getWindowsID(UnicodeString(id, len), resultWinID, *status);
    if (U_SUCCESS(*status) && resultWinID.length() > 0) {
        resultLen = resultWinID.length();
        resultWinID.extract(winid, winidCapacity, *status);
    }

    return resultLen;
}

U_CAPI int32_t U_EXPORT2
ucal_getTimeZoneIDForWindowsID(const UChar* winid, int32_t len, const char* region, UChar* id, int32_t idCapacity, UErrorCode* status) {
    if (U_FAILURE(*status)) {
        return 0;
    }

    int32_t resultLen = 0;
    UnicodeString resultID;

    TimeZone::getIDForWindowsID(UnicodeString(winid, len), region, resultID, *status);
    if (U_SUCCESS(*status) && resultID.length() > 0) {
        resultLen = resultID.length();
        resultID.extract(id, idCapacity, *status);
    }

    return resultLen;
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*                                                                            *
* Copyright (C) 2001-2014, International Business Machines                   *
*                Corporation and others. All Rights Reserved.                *
*                                                                            *
******************************************************************************
*   file name:  ucln_in.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2001July05
*   created by: George Rhoten
*/






/**  Auto-client for UCLN_I18N **/
#define UCLN_TYPE UCLN_I18N


/* Leave this copyright notice here! It needs to go somewhere in this library. */
// static const char copyright[] = U_COPYRIGHT_STRING;

static cleanupFunc *gCleanupFunctions[UCLN_I18N_COUNT];

static UBool U_CALLCONV i18n_cleanup(void)
{
    int32_t libType = UCLN_I18N_START;
    // (void)copyright;   /* Suppress unused variable warning with clang. */

    while (++libType<UCLN_I18N_COUNT) {
        if (gCleanupFunctions[libType])
        {
            gCleanupFunctions[libType]();
            gCleanupFunctions[libType] = NULL;
        }
    }
#if !UCLN_NO_AUTO_CLEANUP && (defined(UCLN_AUTO_ATEXIT) || defined(UCLN_AUTO_LOCAL))
    ucln_unRegisterAutomaticCleanup();
#endif
    return TRUE;
}

void ucln_i18n_registerCleanup(ECleanupI18NType type,
                               cleanupFunc *func) {
    U_ASSERT(UCLN_I18N_START < type && type < UCLN_I18N_COUNT);
    {
        icu::Mutex m;   // See ticket 10295 for discussion.
        ucln_registerCleanup(UCLN_I18N, i18n_cleanup);
        if (UCLN_I18N_START < type && type < UCLN_I18N_COUNT) {
            gCleanupFunctions[type] = func;
        }
    }
#if !UCLN_NO_AUTO_CLEANUP && (defined(UCLN_AUTO_ATEXIT) || defined(UCLN_AUTO_LOCAL))
    ucln_registerAutomaticCleanup();
#endif
}

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 1996-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  ucol.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
* Modification history
* Date        Name      Comments
* 1996-1999   various members of ICU team maintained C API for collation framework
* 02/16/2001  synwee    Added internal method getPrevSpecialCE
* 03/01/2001  synwee    Added maxexpansion functionality.
* 03/16/2001  weiv      Collation framework is rewritten in C and made UCA compliant
* 2012-2014   markus    Rewritten in C++ again.
*/



#if !UCONFIG_NO_COLLATION





// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2001-2014, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*
* File ucoleitr.h
*
* Modification History:
*
* Date        Name        Description
* 02/15/2001  synwee      Modified all methods to process its own function 
*                         instead of calling the equivalent c++ api (coleitr.h)
*******************************************************************************/




#if !UCONFIG_NO_COLLATION

/**  
 * This indicates an error has occured during processing or if no more CEs is 
 * to be returned.
 * @stable ICU 2.0
 */
#define UCOL_NULLORDER        ((int32_t)0xFFFFFFFF)



/** 
 * The UCollationElements struct.
 * For usage in C programs.
 * @stable ICU 2.0
 */
typedef struct UCollationElements UCollationElements;

/**
 * \file
 * \brief C API: UCollationElements
 *
 * The UCollationElements API is used as an iterator to walk through each 
 * character of an international string. Use the iterator to return the
 * ordering priority of the positioned character. The ordering priority of a 
 * character, which we refer to as a key, defines how a character is collated 
 * in the given collation object.
 * For example, consider the following in Slovak and in traditional Spanish collation:
 * <pre>
 * .       "ca" -> the first key is key('c') and second key is key('a').
 * .       "cha" -> the first key is key('ch') and second key is key('a').
 * </pre>
 * And in German phonebook collation,
 * <pre>
 * .       "<ae ligature>b"-> the first key is key('a'), the second key is key('e'), and
 * .       the third key is key('b').
 * </pre>
 * <p>Example of the iterator usage: (without error checking)
 * <pre>
 * .  void CollationElementIterator_Example()
 * .  {
 * .      UChar *s;
 * .      t_int32 order, primaryOrder;
 * .      UCollationElements *c;
 * .      UCollatorOld *coll;
 * .      UErrorCode success = U_ZERO_ERROR;
 * .      s=(UChar*)malloc(sizeof(UChar) * (strlen("This is a test")+1) );
 * .      u_uastrcpy(s, "This is a test");
 * .      coll = ucol_open(NULL, &success);
 * .      c = ucol_openElements(coll, str, u_strlen(str), &status);
 * .      order = ucol_next(c, &success);
 * .      ucol_reset(c);
 * .      order = ucol_prev(c, &success);
 * .      free(s);
 * .      ucol_close(coll);
 * .      ucol_closeElements(c);
 * .  }
 * </pre>
 * <p>
 * ucol_next() returns the collation order of the next.
 * ucol_prev() returns the collation order of the previous character.
 * The Collation Element Iterator moves only in one direction between calls to
 * ucol_reset. That is, ucol_next() and ucol_prev can not be inter-used. 
 * Whenever ucol_prev is to be called after ucol_next() or vice versa, 
 * ucol_reset has to be called first to reset the status, shifting pointers to 
 * either the end or the start of the string. Hence at the next call of 
 * ucol_prev or ucol_next, the first or last collation order will be returned. 
 * If a change of direction is done without a ucol_reset, the result is 
 * undefined.
 * The result of a forward iterate (ucol_next) and reversed result of the  
 * backward iterate (ucol_prev) on the same string are equivalent, if 
 * collation orders with the value 0 are ignored.
 * Character based on the comparison level of the collator.  A collation order 
 * consists of primary order, secondary order and tertiary order.  The data 
 * type of the collation order is <strong>int32_t</strong>. 
 *
 * @see UCollator
 */

/**
 * Open the collation elements for a string.
 *
 * @param coll The collator containing the desired collation rules.
 * @param text The text to iterate over.
 * @param textLength The number of characters in text, or -1 if null-terminated
 * @param status A pointer to a UErrorCode to receive any errors.
 * @return a struct containing collation element information
 * @stable ICU 2.0
 */
U_STABLE UCollationElements* U_EXPORT2 
ucol_openElements(const UCollator  *coll,
                  const UChar      *text,
                        int32_t    textLength,
                        UErrorCode *status);


/**
 * get a hash code for a key... Not very useful!
 * @param key    the given key.
 * @param length the size of the key array.
 * @return       the hash code.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
ucol_keyHashCode(const uint8_t* key, int32_t length);

/**
 * Close a UCollationElements.
 * Once closed, a UCollationElements may no longer be used.
 * @param elems The UCollationElements to close.
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
ucol_closeElements(UCollationElements *elems);

/**
 * Reset the collation elements to their initial state.
 * This will move the 'cursor' to the beginning of the text.
 * Property settings for collation will be reset to the current status.
 * @param elems The UCollationElements to reset.
 * @see ucol_next
 * @see ucol_previous
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
ucol_reset(UCollationElements *elems);

/**
 * Get the ordering priority of the next collation element in the text.
 * A single character may contain more than one collation element.
 * @param elems The UCollationElements containing the text.
 * @param status A pointer to a UErrorCode to receive any errors.
 * @return The next collation elements ordering, otherwise returns UCOL_NULLORDER 
 *         if an error has occured or if the end of string has been reached
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
ucol_next(UCollationElements *elems, UErrorCode *status);

/**
 * Get the ordering priority of the previous collation element in the text.
 * A single character may contain more than one collation element.
 * Note that internally a stack is used to store buffered collation elements. 
 * @param elems The UCollationElements containing the text.
 * @param status A pointer to a UErrorCode to receive any errors. Noteably 
 *               a U_BUFFER_OVERFLOW_ERROR is returned if the internal stack
 *               buffer has been exhausted.
 * @return The previous collation elements ordering, otherwise returns 
 *         UCOL_NULLORDER if an error has occured or if the start of string has 
 *         been reached.
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
ucol_previous(UCollationElements *elems, UErrorCode *status);

/**
 * Get the maximum length of any expansion sequences that end with the 
 * specified comparison order.
 * This is useful for .... ?
 * @param elems The UCollationElements containing the text.
 * @param order A collation order returned by previous or next.
 * @return maximum size of the expansion sequences ending with the collation 
 *         element or 1 if collation element does not occur at the end of any 
 *         expansion sequence
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
ucol_getMaxExpansion(const UCollationElements *elems, int32_t order);

/**
 * Set the text containing the collation elements.
 * Property settings for collation will remain the same.
 * In order to reset the iterator to the current collation property settings,
 * the API reset() has to be called.
 * @param elems The UCollationElements to set.
 * @param text The source text containing the collation elements.
 * @param textLength The length of text, or -1 if null-terminated.
 * @param status A pointer to a UErrorCode to receive any errors.
 * @see ucol_getText
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
ucol_setText(      UCollationElements *elems, 
             const UChar              *text,
                   int32_t            textLength,
                   UErrorCode         *status);

/**
 * Get the offset of the current source character.
 * This is an offset into the text of the character containing the current
 * collation elements.
 * @param elems The UCollationElements to query.
 * @return The offset of the current source character.
 * @see ucol_setOffset
 * @stable ICU 2.0
 */
U_STABLE int32_t U_EXPORT2 
ucol_getOffset(const UCollationElements *elems);

/**
 * Set the offset of the current source character.
 * This is an offset into the text of the character to be processed.
 * Property settings for collation will remain the same.
 * In order to reset the iterator to the current collation property settings,
 * the API reset() has to be called.
 * @param elems The UCollationElements to set.
 * @param offset The desired character offset.
 * @param status A pointer to a UErrorCode to receive any errors.
 * @see ucol_getOffset
 * @stable ICU 2.0
 */
U_STABLE void U_EXPORT2 
ucol_setOffset(UCollationElements *elems,
               int32_t        offset,
               UErrorCode         *status);

/**
* Get the primary order of a collation order.
* @param order the collation order
* @return the primary order of a collation order.
* @stable ICU 2.6
*/
U_STABLE int32_t U_EXPORT2
ucol_primaryOrder (int32_t order); 

/**
* Get the secondary order of a collation order.
* @param order the collation order
* @return the secondary order of a collation order.
* @stable ICU 2.6
*/
U_STABLE int32_t U_EXPORT2
ucol_secondaryOrder (int32_t order); 

/**
* Get the tertiary order of a collation order.
* @param order the collation order
* @return the tertiary order of a collation order.
* @stable ICU 2.6
*/
U_STABLE int32_t U_EXPORT2
ucol_tertiaryOrder (int32_t order); 

#endif /* #if !UCONFIG_NO_COLLATION */










U_NAMESPACE_USE

U_CAPI UCollator* U_EXPORT2
ucol_openBinary(const uint8_t *bin, int32_t length,
                const UCollator *base,
                UErrorCode *status)
{
    if(U_FAILURE(*status)) { return NULL; }
    RuleBasedCollator *coll = new RuleBasedCollator(
            bin, length,
            RuleBasedCollator::rbcFromUCollator(base),
            *status);
    if(coll == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    if(U_FAILURE(*status)) {
        delete coll;
        return NULL;
    }
    return coll->toUCollator();
}

U_CAPI int32_t U_EXPORT2
ucol_cloneBinary(const UCollator *coll,
                 uint8_t *buffer, int32_t capacity,
                 UErrorCode *status)
{
    if(U_FAILURE(*status)) {
        return 0;
    }
    const RuleBasedCollator *rbc = RuleBasedCollator::rbcFromUCollator(coll);
    if(rbc == NULL && coll != NULL) {
        *status = U_UNSUPPORTED_ERROR;
        return 0;
    }
    return rbc->cloneBinary(buffer, capacity, *status);
}

U_CAPI UCollator* U_EXPORT2
ucol_safeClone(const UCollator *coll, void * /*stackBuffer*/, int32_t * pBufferSize, UErrorCode *status)
{
    if (status == NULL || U_FAILURE(*status)){
        return NULL;
    }
    if (coll == NULL) {
       *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    if (pBufferSize != NULL) {
        int32_t inputSize = *pBufferSize;
        *pBufferSize = 1;
        if (inputSize == 0) {
            return NULL;  // preflighting for deprecated functionality
        }
    }
    Collator *newColl = Collator::fromUCollator(coll)->clone();
    if (newColl == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    } else {
        *status = U_SAFECLONE_ALLOCATED_WARNING;
    }
    return newColl->toUCollator();
}

U_CAPI void U_EXPORT2
ucol_close(UCollator *coll)
{
    UTRACE_ENTRY_OC(UTRACE_UCOL_CLOSE);
    UTRACE_DATA1(UTRACE_INFO, "coll = %p", coll);
    if(coll != NULL) {
        delete Collator::fromUCollator(coll);
    }
    UTRACE_EXIT();
}

U_CAPI int32_t U_EXPORT2
ucol_mergeSortkeys(const uint8_t *src1, int32_t src1Length,
                   const uint8_t *src2, int32_t src2Length,
                   uint8_t *dest, int32_t destCapacity) {
    /* check arguments */
    if( src1==NULL || src1Length<-1 || src1Length==0 || (src1Length>0 && src1[src1Length-1]!=0) ||
        src2==NULL || src2Length<-1 || src2Length==0 || (src2Length>0 && src2[src2Length-1]!=0) ||
        destCapacity<0 || (destCapacity>0 && dest==NULL)
    ) {
        /* error, attempt to write a zero byte and return 0 */
        if(dest!=NULL && destCapacity>0) {
            *dest=0;
        }
        return 0;
    }

    /* check lengths and capacity */
    if(src1Length<0) {
        src1Length=(int32_t)uprv_strlen((const char *)src1)+1;
    }
    if(src2Length<0) {
        src2Length=(int32_t)uprv_strlen((const char *)src2)+1;
    }

    int32_t destLength=src1Length+src2Length;
    if(destLength>destCapacity) {
        /* the merged sort key does not fit into the destination */
        return destLength;
    }

    /* merge the sort keys with the same number of levels */
    uint8_t *p=dest;
    for(;;) {
        /* copy level from src1 not including 00 or 01 */
        uint8_t b;
        while((b=*src1)>=2) {
            ++src1;
            *p++=b;
        }

        /* add a 02 merge separator */
        *p++=2;

        /* copy level from src2 not including 00 or 01 */
        while((b=*src2)>=2) {
            ++src2;
            *p++=b;
        }

        /* if both sort keys have another level, then add a 01 level separator and continue */
        if(*src1==1 && *src2==1) {
            ++src1;
            ++src2;
            *p++=1;
        } else {
            break;
        }
    }

    /*
     * here, at least one sort key is finished now, but the other one
     * might have some contents left from containing more levels;
     * that contents is just appended to the result
     */
    if(*src1!=0) {
        /* src1 is not finished, therefore *src2==0, and src1 is appended */
        src2=src1;
    }
    /* append src2, "the other, unfinished sort key" */
    while((*p++=*src2++)!=0) {}

    /* the actual length might be less than destLength if either sort key contained illegally embedded zero bytes */
    return (int32_t)(p-dest);
}

U_CAPI int32_t U_EXPORT2
ucol_getSortKey(const    UCollator    *coll,
        const    UChar        *source,
        int32_t        sourceLength,
        uint8_t        *result,
        int32_t        resultLength)
{
    UTRACE_ENTRY(UTRACE_UCOL_GET_SORTKEY);
    if (UTRACE_LEVEL(UTRACE_VERBOSE)) {
        UTRACE_DATA3(UTRACE_VERBOSE, "coll=%p, source string = %vh ", coll, source,
            ((sourceLength==-1 && source!=NULL) ? u_strlen(source) : sourceLength));
    }

    int32_t keySize = Collator::fromUCollator(coll)->
            getSortKey(source, sourceLength, result, resultLength);

    UTRACE_DATA2(UTRACE_VERBOSE, "Sort Key = %vb", result, keySize);
    UTRACE_EXIT_VALUE(keySize);
    return keySize;
}

U_CAPI int32_t U_EXPORT2
ucol_nextSortKeyPart(const UCollator *coll,
                     UCharIterator *iter,
                     uint32_t state[2],
                     uint8_t *dest, int32_t count,
                     UErrorCode *status)
{
    /* error checking */
    if(status==NULL || U_FAILURE(*status)) {
        return 0;
    }
    UTRACE_ENTRY(UTRACE_UCOL_NEXTSORTKEYPART);
    UTRACE_DATA6(UTRACE_VERBOSE, "coll=%p, iter=%p, state=%d %d, dest=%p, count=%d",
                  coll, iter, state[0], state[1], dest, count);

    int32_t i = Collator::fromUCollator(coll)->
            internalNextSortKeyPart(iter, state, dest, count, *status);

    // Return number of meaningful sortkey bytes.
    UTRACE_DATA4(UTRACE_VERBOSE, "dest = %vb, state=%d %d",
                  dest,i, state[0], state[1]);
    UTRACE_EXIT_VALUE_STATUS(i, *status);
    return i;
}

/**
 * Produce a bound for a given sortkey and a number of levels.
 */
U_CAPI int32_t U_EXPORT2
ucol_getBound(const uint8_t       *source,
        int32_t             sourceLength,
        UColBoundMode       boundType,
        uint32_t            noOfLevels,
        uint8_t             *result,
        int32_t             resultLength,
        UErrorCode          *status)
{
    // consistency checks
    if(status == NULL || U_FAILURE(*status)) {
        return 0;
    }
    if(source == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    int32_t sourceIndex = 0;
    // Scan the string until we skip enough of the key OR reach the end of the key
    do {
        sourceIndex++;
        if(source[sourceIndex] == Collation::LEVEL_SEPARATOR_BYTE) {
            noOfLevels--;
        }
    } while (noOfLevels > 0
        && (source[sourceIndex] != 0 || sourceIndex < sourceLength));

    if((source[sourceIndex] == 0 || sourceIndex == sourceLength)
        && noOfLevels > 0) {
            *status = U_SORT_KEY_TOO_SHORT_WARNING;
    }


    // READ ME: this code assumes that the values for boundType
    // enum will not changes. They are set so that the enum value
    // corresponds to the number of extra bytes each bound type
    // needs.
    if(result != NULL && resultLength >= sourceIndex+boundType) {
        uprv_memcpy(result, source, sourceIndex);
        switch(boundType) {
            // Lower bound just gets terminated. No extra bytes
        case UCOL_BOUND_LOWER: // = 0
            break;
            // Upper bound needs one extra byte
        case UCOL_BOUND_UPPER: // = 1
            result[sourceIndex++] = 2;
            break;
            // Upper long bound needs two extra bytes
        case UCOL_BOUND_UPPER_LONG: // = 2
            result[sourceIndex++] = 0xFF;
            result[sourceIndex++] = 0xFF;
            break;
        default:
            *status = U_ILLEGAL_ARGUMENT_ERROR;
            return 0;
        }
        result[sourceIndex++] = 0;

        return sourceIndex;
    } else {
        return sourceIndex+boundType+1;
    }
}

U_CAPI void U_EXPORT2
ucol_setMaxVariable(UCollator *coll, UColReorderCode group, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) { return; }
    Collator::fromUCollator(coll)->setMaxVariable(group, *pErrorCode);
}

U_CAPI UColReorderCode U_EXPORT2
ucol_getMaxVariable(const UCollator *coll) {
    return Collator::fromUCollator(coll)->getMaxVariable();
}

U_CAPI uint32_t  U_EXPORT2
ucol_setVariableTop(UCollator *coll, const UChar *varTop, int32_t len, UErrorCode *status) {
    if(U_FAILURE(*status) || coll == NULL) {
        return 0;
    }
    return Collator::fromUCollator(coll)->setVariableTop(varTop, len, *status);
}

U_CAPI uint32_t U_EXPORT2 ucol_getVariableTop(const UCollator *coll, UErrorCode *status) {
    if(U_FAILURE(*status) || coll == NULL) {
        return 0;
    }
    return Collator::fromUCollator(coll)->getVariableTop(*status);
}

U_CAPI void  U_EXPORT2
ucol_restoreVariableTop(UCollator *coll, const uint32_t varTop, UErrorCode *status) {
    if(U_FAILURE(*status) || coll == NULL) {
        return;
    }
    Collator::fromUCollator(coll)->setVariableTop(varTop, *status);
}

U_CAPI void  U_EXPORT2
ucol_setAttribute(UCollator *coll, UColAttribute attr, UColAttributeValue value, UErrorCode *status) {
    if(U_FAILURE(*status) || coll == NULL) {
      return;
    }

    Collator::fromUCollator(coll)->setAttribute(attr, value, *status);
}

U_CAPI UColAttributeValue  U_EXPORT2
ucol_getAttribute(const UCollator *coll, UColAttribute attr, UErrorCode *status) {
    if(U_FAILURE(*status) || coll == NULL) {
      return UCOL_DEFAULT;
    }

    return Collator::fromUCollator(coll)->getAttribute(attr, *status);
}

U_CAPI void U_EXPORT2
ucol_setStrength(    UCollator                *coll,
            UCollationStrength        strength)
{
    UErrorCode status = U_ZERO_ERROR;
    ucol_setAttribute(coll, UCOL_STRENGTH, strength, &status);
}

U_CAPI UCollationStrength U_EXPORT2
ucol_getStrength(const UCollator *coll)
{
    UErrorCode status = U_ZERO_ERROR;
    return ucol_getAttribute(coll, UCOL_STRENGTH, &status);
}

U_CAPI int32_t U_EXPORT2
ucol_getReorderCodes(const UCollator *coll,
                    int32_t *dest,
                    int32_t destCapacity,
                    UErrorCode *status) {
    if (U_FAILURE(*status)) {
        return 0;
    }

    return Collator::fromUCollator(coll)->getReorderCodes(dest, destCapacity, *status);
}

U_CAPI void U_EXPORT2
ucol_setReorderCodes(UCollator* coll,
                    const int32_t* reorderCodes,
                    int32_t reorderCodesLength,
                    UErrorCode *status) {
    if (U_FAILURE(*status)) {
        return;
    }

    Collator::fromUCollator(coll)->setReorderCodes(reorderCodes, reorderCodesLength, *status);
}

U_CAPI int32_t U_EXPORT2
ucol_getEquivalentReorderCodes(int32_t reorderCode,
                    int32_t* dest,
                    int32_t destCapacity,
                    UErrorCode *pErrorCode) {
    return Collator::getEquivalentReorderCodes(reorderCode, dest, destCapacity, *pErrorCode);
}

U_CAPI void U_EXPORT2
ucol_getVersion(const UCollator* coll,
                UVersionInfo versionInfo)
{
    Collator::fromUCollator(coll)->getVersion(versionInfo);
}

U_CAPI UCollationResult U_EXPORT2
ucol_strcollIter( const UCollator    *coll,
                 UCharIterator *sIter,
                 UCharIterator *tIter,
                 UErrorCode         *status)
{
    if(!status || U_FAILURE(*status)) {
        return UCOL_EQUAL;
    }

    UTRACE_ENTRY(UTRACE_UCOL_STRCOLLITER);
    UTRACE_DATA3(UTRACE_VERBOSE, "coll=%p, sIter=%p, tIter=%p", coll, sIter, tIter);

    if(sIter == NULL || tIter == NULL || coll == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        UTRACE_EXIT_VALUE_STATUS(UCOL_EQUAL, *status);
        return UCOL_EQUAL;
    }

    UCollationResult result = Collator::fromUCollator(coll)->compare(*sIter, *tIter, *status);

    UTRACE_EXIT_VALUE_STATUS(result, *status);
    return result;
}


/*                                                                      */
/* ucol_strcoll     Main public API string comparison function          */
/*                                                                      */
U_CAPI UCollationResult U_EXPORT2
ucol_strcoll( const UCollator    *coll,
              const UChar        *source,
              int32_t            sourceLength,
              const UChar        *target,
              int32_t            targetLength)
{
    UTRACE_ENTRY(UTRACE_UCOL_STRCOLL);
    if (UTRACE_LEVEL(UTRACE_VERBOSE)) {
        UTRACE_DATA3(UTRACE_VERBOSE, "coll=%p, source=%p, target=%p", coll, source, target);
        UTRACE_DATA2(UTRACE_VERBOSE, "source string = %vh ", source, sourceLength);
        UTRACE_DATA2(UTRACE_VERBOSE, "target string = %vh ", target, targetLength);
    }

    UErrorCode status = U_ZERO_ERROR;
    UCollationResult returnVal = Collator::fromUCollator(coll)->
            compare(source, sourceLength, target, targetLength, status);
    UTRACE_EXIT_VALUE_STATUS(returnVal, status);
    return returnVal;
}

U_CAPI UCollationResult U_EXPORT2
ucol_strcollUTF8(
        const UCollator *coll,
        const char      *source,
        int32_t         sourceLength,
        const char      *target,
        int32_t         targetLength,
        UErrorCode      *status)
{
    UTRACE_ENTRY(UTRACE_UCOL_STRCOLLUTF8);
    if (UTRACE_LEVEL(UTRACE_VERBOSE)) {
        UTRACE_DATA3(UTRACE_VERBOSE, "coll=%p, source=%p, target=%p", coll, source, target);
        UTRACE_DATA2(UTRACE_VERBOSE, "source string = %vb ", source, sourceLength);
        UTRACE_DATA2(UTRACE_VERBOSE, "target string = %vb ", target, targetLength);
    }

    if (U_FAILURE(*status)) {
        /* do nothing */
        UTRACE_EXIT_VALUE_STATUS(UCOL_EQUAL, *status);
        return UCOL_EQUAL;
    }

    UCollationResult returnVal = Collator::fromUCollator(coll)->internalCompareUTF8(
            source, sourceLength, target, targetLength, *status);
    UTRACE_EXIT_VALUE_STATUS(returnVal, *status);
    return returnVal;
}


/* convenience function for comparing strings */
U_CAPI UBool U_EXPORT2
ucol_greater(    const    UCollator        *coll,
        const    UChar            *source,
        int32_t            sourceLength,
        const    UChar            *target,
        int32_t            targetLength)
{
    return (ucol_strcoll(coll, source, sourceLength, target, targetLength)
        == UCOL_GREATER);
}

/* convenience function for comparing strings */
U_CAPI UBool U_EXPORT2
ucol_greaterOrEqual(    const    UCollator    *coll,
            const    UChar        *source,
            int32_t        sourceLength,
            const    UChar        *target,
            int32_t        targetLength)
{
    return (ucol_strcoll(coll, source, sourceLength, target, targetLength)
        != UCOL_LESS);
}

/* convenience function for comparing strings */
U_CAPI UBool U_EXPORT2
ucol_equal(        const    UCollator        *coll,
            const    UChar            *source,
            int32_t            sourceLength,
            const    UChar            *target,
            int32_t            targetLength)
{
    return (ucol_strcoll(coll, source, sourceLength, target, targetLength)
        == UCOL_EQUAL);
}

U_CAPI void U_EXPORT2
ucol_getUCAVersion(const UCollator* coll, UVersionInfo info) {
    const Collator *c = Collator::fromUCollator(coll);
    if(c != NULL) {
        UVersionInfo v;
        c->getVersion(v);
        // Note: This is tied to how the current implementation encodes the UCA version
        // in the overall getVersion().
        // Alternatively, we could load the root collator and get at lower-level data from there.
        // Either way, it will reflect the input collator's UCA version only
        // if it is a known implementation.
        // It would be cleaner to make this a virtual Collator method.
        info[0] = v[1] >> 3;
        info[1] = v[1] & 7;
        info[2] = v[2] >> 6;
        info[3] = 0;
    }
}

U_CAPI const UChar * U_EXPORT2
ucol_getRules(const UCollator *coll, int32_t *length) {
    const RuleBasedCollator *rbc = RuleBasedCollator::rbcFromUCollator(coll);
    // OK to crash if coll==NULL: We do not want to check "this" pointers.
    if(rbc != NULL || coll == NULL) {
        const UnicodeString &rules = rbc->getRules();
        U_ASSERT(rules.getBuffer()[rules.length()] == 0);
        *length = rules.length();
        return rules.getBuffer();
    }
    static const UChar _NUL = 0;
    *length = 0;
    return &_NUL;
}

U_CAPI int32_t U_EXPORT2
ucol_getRulesEx(const UCollator *coll, UColRuleOption delta, UChar *buffer, int32_t bufferLen) {
    UnicodeString rules;
    const RuleBasedCollator *rbc = RuleBasedCollator::rbcFromUCollator(coll);
    if(rbc != NULL || coll == NULL) {
        rbc->getRules(delta, rules);
    }
    if(buffer != NULL && bufferLen > 0) {
        UErrorCode errorCode = U_ZERO_ERROR;
        return rules.extract(buffer, bufferLen, errorCode);
    } else {
        return rules.length();
    }
}

U_CAPI const char * U_EXPORT2
ucol_getLocale(const UCollator *coll, ULocDataLocaleType type, UErrorCode *status) {
    return ucol_getLocaleByType(coll, type, status);
}

U_CAPI const char * U_EXPORT2
ucol_getLocaleByType(const UCollator *coll, ULocDataLocaleType type, UErrorCode *status) {
    if(U_FAILURE(*status)) {
        return NULL;
    }
    UTRACE_ENTRY(UTRACE_UCOL_GETLOCALE);
    UTRACE_DATA1(UTRACE_INFO, "coll=%p", coll);

    const char *result;
    const RuleBasedCollator *rbc = RuleBasedCollator::rbcFromUCollator(coll);
    if(rbc == NULL && coll != NULL) {
        *status = U_UNSUPPORTED_ERROR;
        result = NULL;
    } else {
        result = rbc->internalGetLocaleID(type, *status);
    }

    UTRACE_DATA1(UTRACE_INFO, "result = %s", result);
    UTRACE_EXIT_STATUS(*status);
    return result;
}

U_CAPI USet * U_EXPORT2
ucol_getTailoredSet(const UCollator *coll, UErrorCode *status) {
    if(U_FAILURE(*status)) {
        return NULL;
    }
    UnicodeSet *set = Collator::fromUCollator(coll)->getTailoredSet(*status);
    if(U_FAILURE(*status)) {
        delete set;
        return NULL;
    }
    return set->toUSet();
}

U_CAPI UBool U_EXPORT2
ucol_equals(const UCollator *source, const UCollator *target) {
    return source == target ||
        (*Collator::fromUCollator(source)) == (*Collator::fromUCollator(target));
}

#endif /* #if !UCONFIG_NO_COLLATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 1996-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  ucol_res.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
* Description:
* This file contains dependencies that the collation run-time doesn't normally
* need. This mainly contains resource bundle usage and collation meta information
*
* Modification history
* Date        Name      Comments
* 1996-1999   various members of ICU team maintained C API for collation framework
* 02/16/2001  synwee    Added internal method getPrevSpecialCE
* 03/01/2001  synwee    Added maxexpansion functionality.
* 03/16/2001  weiv      Collation framework is rewritten in C and made UCA compliant
* 12/08/2004  grhoten   Split part of ucol.cpp into ucol_res.cpp
* 2012-2014   markus    Rewritten in C++ again.
*/



#if !UCONFIG_NO_COLLATION




























U_NAMESPACE_BEGIN

namespace {

static const UChar *ucol_res_rootRules = NULL;
static int32_t ucol_res_rootRulesLength = 0;
static UResourceBundle *ucol_res_rootBundle = NULL;
static UInitOnce gInitOnceUcolRes = U_INITONCE_INITIALIZER;

}  // namespace

U_CDECL_BEGIN

static UBool U_CALLCONV
ucol_res_cleanup() {
    ucol_res_rootRules = NULL;
    ucol_res_rootRulesLength = 0;
    ures_close(ucol_res_rootBundle);
    ucol_res_rootBundle = NULL;
    gInitOnceUcolRes.reset();
    return TRUE;
}

void U_CALLCONV
CollationLoader::loadRootRules(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    ucol_res_rootBundle = ures_open(U_ICUDATA_COLL, kRootLocaleName, &errorCode);
    if(U_FAILURE(errorCode)) { return; }
    ucol_res_rootRules = ures_getStringByKey(ucol_res_rootBundle, "UCARules", &ucol_res_rootRulesLength, &errorCode);
    if(U_FAILURE(errorCode)) {
        ures_close(ucol_res_rootBundle);
        ucol_res_rootBundle = NULL;
        return;
    }
    ucln_i18n_registerCleanup(UCLN_I18N_UCOL_RES, ucol_res_cleanup);
}

U_CDECL_END

void
CollationLoader::appendRootRules(UnicodeString &s) {
    UErrorCode errorCode = U_ZERO_ERROR;
    umtx_initOnce(gInitOnceUcolRes, CollationLoader::loadRootRules, errorCode);
    if(U_SUCCESS(errorCode)) {
        s.append(ucol_res_rootRules, ucol_res_rootRulesLength);
    }
}

void
CollationLoader::loadRules(const char *localeID, const char *collationType,
                           UnicodeString &rules, UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return; }
    U_ASSERT(collationType != NULL && *collationType != 0);
    // Copy the type for lowercasing.
    char type[16];
    int32_t typeLength = static_cast<int32_t>(uprv_strlen(collationType));
    if(typeLength >= UPRV_LENGTHOF(type)) {
        errorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    uprv_memcpy(type, collationType, typeLength + 1);
    T_CString_toLowerCase(type);

    LocalUResourceBundlePointer bundle(ures_open(U_ICUDATA_COLL, localeID, &errorCode));
    LocalUResourceBundlePointer collations(
            ures_getByKey(bundle.getAlias(), "collations", NULL, &errorCode));
    LocalUResourceBundlePointer data(
            ures_getByKeyWithFallback(collations.getAlias(), type, NULL, &errorCode));
    int32_t length;
    const UChar *s =  ures_getStringByKey(data.getAlias(), "Sequence", &length, &errorCode);
    if(U_FAILURE(errorCode)) { return; }

    // No string pointer aliasing so that we need not hold onto the resource bundle.
    rules.setTo(s, length);
    if(rules.isBogus()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
    }
}

template<> U_I18N_API
const CollationCacheEntry *
LocaleCacheKey<CollationCacheEntry>::createObject(const void *creationContext,
                                                  UErrorCode &errorCode) const {
    CollationLoader *loader =
            reinterpret_cast<CollationLoader *>(
                    const_cast<void *>(creationContext));
    return loader->createCacheEntry(errorCode);
}

const CollationCacheEntry *
CollationLoader::loadTailoring(const Locale &locale, UErrorCode &errorCode) {
    const CollationCacheEntry *rootEntry = CollationRoot::getRootCacheEntry(errorCode);
    if(U_FAILURE(errorCode)) { return NULL; }
    const char *name = locale.getName();
    if(*name == 0 || uprv_strcmp(name, "root") == 0) {

        // Have to add a ref.
        rootEntry->addRef();
        return rootEntry;
    }

    // Clear warning codes before loading where they get cached.
    errorCode = U_ZERO_ERROR;
    CollationLoader loader(rootEntry, locale, errorCode);

    // getCacheEntry adds a ref for us.
    return loader.getCacheEntry(errorCode);
}

CollationLoader::CollationLoader(const CollationCacheEntry *re, const Locale &requested,
                                 UErrorCode &errorCode)
        : cache(UnifiedCache::getInstance(errorCode)), rootEntry(re),
          validLocale(re->validLocale), locale(requested),
          typesTried(0), typeFallback(FALSE),
          bundle(NULL), collations(NULL), data(NULL) {
    type[0] = 0;
    defaultType[0] = 0;
    if(U_FAILURE(errorCode)) { return; }

    // Canonicalize the locale ID: Ignore all irrelevant keywords.
    const char *baseName = locale.getBaseName();
    if(uprv_strcmp(locale.getName(), baseName) != 0) {
        locale = Locale(baseName);

        // Fetch the collation type from the locale ID.
        int32_t typeLength = requested.getKeywordValue("collation",
                type, UPRV_LENGTHOF(type) - 1, errorCode);
        if(U_FAILURE(errorCode)) {
            errorCode = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }
        type[typeLength] = 0;  // in case of U_NOT_TERMINATED_WARNING
        if(typeLength == 0) {
            // No collation type.
        } else if(uprv_stricmp(type, "default") == 0) {
            // Ignore "default" (case-insensitive).
            type[0] = 0;
        } else {
            // Copy the collation type.
            T_CString_toLowerCase(type);
            locale.setKeywordValue("collation", type, errorCode);
        }
    }
}

CollationLoader::~CollationLoader() {
    ures_close(data);
    ures_close(collations);
    ures_close(bundle);
}

const CollationCacheEntry *
CollationLoader::createCacheEntry(UErrorCode &errorCode) {
    // This is a linear lookup and fallback flow turned into a state machine.
    // Most local variables have been turned into instance fields.
    // In a cache miss, cache.get() calls CacheKey::createObject(),
    // which means that we progress via recursion.
    // loadFromCollations() will recurse to itself as well for collation type fallback.
    if(bundle == NULL) {
        return loadFromLocale(errorCode);
    } else if(collations == NULL) {
        return loadFromBundle(errorCode);
    } else if(data == NULL) {
        return loadFromCollations(errorCode);
    } else {
        return loadFromData(errorCode);
    }
}

const CollationCacheEntry *
CollationLoader::loadFromLocale(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return NULL; }
    U_ASSERT(bundle == NULL);
    bundle = ures_openNoDefault(U_ICUDATA_COLL, locale.getBaseName(), &errorCode);
    if(errorCode == U_MISSING_RESOURCE_ERROR) {
        errorCode = U_USING_DEFAULT_WARNING;

        // Have to add that ref that we promise.
        rootEntry->addRef();
        return rootEntry;
    }
    Locale requestedLocale(locale);
    const char *vLocale = ures_getLocaleByType(bundle, ULOC_ACTUAL_LOCALE, &errorCode);
    if(U_FAILURE(errorCode)) { return NULL; }
    locale = validLocale = Locale(vLocale);  // no type until loadFromCollations()
    if(type[0] != 0) {
        locale.setKeywordValue("collation", type, errorCode);
    }
    if(locale != requestedLocale) {
        return getCacheEntry(errorCode);
    } else {
        return loadFromBundle(errorCode);
    }
}

const CollationCacheEntry *
CollationLoader::loadFromBundle(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return NULL; }
    U_ASSERT(collations == NULL);
    // There are zero or more tailorings in the collations table.
    collations = ures_getByKey(bundle, "collations", NULL, &errorCode);
    if(errorCode == U_MISSING_RESOURCE_ERROR) {
        errorCode = U_USING_DEFAULT_WARNING;
        // Return the root tailoring with the validLocale, without collation type.
        return makeCacheEntryFromRoot(validLocale, errorCode);
    }
    if(U_FAILURE(errorCode)) { return NULL; }

    // Fetch the default type from the data.
    {
        UErrorCode internalErrorCode = U_ZERO_ERROR;
        LocalUResourceBundlePointer def(
                ures_getByKeyWithFallback(collations, "default", NULL, &internalErrorCode));
        int32_t length;
        const UChar *s = ures_getString(def.getAlias(), &length, &internalErrorCode);
        if(U_SUCCESS(internalErrorCode) && 0 < length && length < UPRV_LENGTHOF(defaultType)) {
            u_UCharsToChars(s, defaultType, length + 1);
        } else {
            uprv_strcpy(defaultType, "standard");
        }
    }

    // Record which collation types we have looked for already,
    // so that we do not deadlock in the cache.
    //
    // If there is no explicit type, then we look in the cache
    // for the entry with the default type.
    // If the explicit type is the default type, then we do not look in the cache
    // for the entry with an empty type.
    // Otherwise, two concurrent requests with opposite fallbacks would deadlock each other.
    // Also, it is easier to always enter the next method with a non-empty type.
    if(type[0] == 0) {
        uprv_strcpy(type, defaultType);
        typesTried |= TRIED_DEFAULT;
        if(uprv_strcmp(type, "search") == 0) {
            typesTried |= TRIED_SEARCH;
        }
        if(uprv_strcmp(type, "standard") == 0) {
            typesTried |= TRIED_STANDARD;
        }
        locale.setKeywordValue("collation", type, errorCode);
        return getCacheEntry(errorCode);
    } else {
        if(uprv_strcmp(type, defaultType) == 0) {
            typesTried |= TRIED_DEFAULT;
        }
        if(uprv_strcmp(type, "search") == 0) {
            typesTried |= TRIED_SEARCH;
        }
        if(uprv_strcmp(type, "standard") == 0) {
            typesTried |= TRIED_STANDARD;
        }
        return loadFromCollations(errorCode);
    }
}

const CollationCacheEntry *
CollationLoader::loadFromCollations(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return NULL; }
    U_ASSERT(data == NULL);
    // Load the collations/type tailoring, with type fallback.
    LocalUResourceBundlePointer localData(
            ures_getByKeyWithFallback(collations, type, NULL, &errorCode));
    int32_t typeLength = static_cast<int32_t>(uprv_strlen(type));
    if(errorCode == U_MISSING_RESOURCE_ERROR) {
        errorCode = U_USING_DEFAULT_WARNING;
        typeFallback = TRUE;
        if((typesTried & TRIED_SEARCH) == 0 &&
                typeLength > 6 && uprv_strncmp(type, "search", 6) == 0) {
            // fall back from something like "searchjl" to "search"
            typesTried |= TRIED_SEARCH;
            type[6] = 0;
        } else if((typesTried & TRIED_DEFAULT) == 0) {
            // fall back to the default type
            typesTried |= TRIED_DEFAULT;
            uprv_strcpy(type, defaultType);
        } else if((typesTried & TRIED_STANDARD) == 0) {
            // fall back to the "standard" type
            typesTried |= TRIED_STANDARD;
            uprv_strcpy(type, "standard");
        } else {
            // Return the root tailoring with the validLocale, without collation type.
            return makeCacheEntryFromRoot(validLocale, errorCode);
        }
        locale.setKeywordValue("collation", type, errorCode);
        return getCacheEntry(errorCode);
    }
    if(U_FAILURE(errorCode)) { return NULL; }

    data = localData.orphan();
    const char *actualLocale = ures_getLocaleByType(data, ULOC_ACTUAL_LOCALE, &errorCode);
    if(U_FAILURE(errorCode)) { return NULL; }
    const char *vLocale = validLocale.getBaseName();
    UBool actualAndValidLocalesAreDifferent = Locale(actualLocale) != Locale(vLocale);

    // Set the collation types on the informational locales,
    // except when they match the default types (for brevity and backwards compatibility).
    // For the valid locale, suppress the default type.
    if(uprv_strcmp(type, defaultType) != 0) {
        validLocale.setKeywordValue("collation", type, errorCode);
        if(U_FAILURE(errorCode)) { return NULL; }
    }

    // Is this the same as the root collator? If so, then use that instead.
    if((*actualLocale == 0 || uprv_strcmp(actualLocale, "root") == 0) &&
            uprv_strcmp(type, "standard") == 0) {
        if(typeFallback) {
            errorCode = U_USING_DEFAULT_WARNING;
        }
        return makeCacheEntryFromRoot(validLocale, errorCode);
    }

    locale = Locale(actualLocale);
    if(actualAndValidLocalesAreDifferent) {
        locale.setKeywordValue("collation", type, errorCode);
        const CollationCacheEntry *entry = getCacheEntry(errorCode);
        return makeCacheEntry(validLocale, entry, errorCode);
    } else {
        return loadFromData(errorCode);
    }
}

const CollationCacheEntry *
CollationLoader::loadFromData(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return NULL; }
    LocalPointer<CollationTailoring> t(new CollationTailoring(rootEntry->tailoring->settings));
    if(t.isNull() || t->isBogus()) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    // deserialize
    LocalUResourceBundlePointer binary(ures_getByKey(data, "%%CollationBin", NULL, &errorCode));
    // Note: U_MISSING_RESOURCE_ERROR --> The old code built from rules if available
    // but that created undesirable dependencies.
    int32_t length;
    const uint8_t *inBytes = ures_getBinary(binary.getAlias(), &length, &errorCode);
    CollationDataReader::read(rootEntry->tailoring, inBytes, length, *t, errorCode);
    // Note: U_COLLATOR_VERSION_MISMATCH --> The old code built from rules if available
    // but that created undesirable dependencies.
    if(U_FAILURE(errorCode)) { return NULL; }

    // Try to fetch the optional rules string.
    {
        UErrorCode internalErrorCode = U_ZERO_ERROR;
        int32_t len;
        const UChar *s = ures_getStringByKey(data, "Sequence", &len,
                                             &internalErrorCode);
        if(U_SUCCESS(internalErrorCode)) {
            t->rules.setTo(TRUE, s, len);
        }
    }

    const char *actualLocale = locale.getBaseName();  // without type
    const char *vLocale = validLocale.getBaseName();
    UBool actualAndValidLocalesAreDifferent = Locale(actualLocale) != Locale(vLocale);

    // For the actual locale, suppress the default type *according to the actual locale*.
    // For example, zh has default=pinyin and contains all of the Chinese tailorings.
    // zh_Hant has default=stroke but has no other data.
    // For the valid locale "zh_Hant" we need to suppress stroke.
    // For the actual locale "zh" we need to suppress pinyin instead.
    if(actualAndValidLocalesAreDifferent) {
        // Opening a bundle for the actual locale should always succeed.
        LocalUResourceBundlePointer actualBundle(
                ures_open(U_ICUDATA_COLL, actualLocale, &errorCode));
        if(U_FAILURE(errorCode)) { return NULL; }
        UErrorCode internalErrorCode = U_ZERO_ERROR;
        LocalUResourceBundlePointer def(
                ures_getByKeyWithFallback(actualBundle.getAlias(), "collations/default", NULL,
                                          &internalErrorCode));
        int32_t len;
        const UChar *s = ures_getString(def.getAlias(), &len, &internalErrorCode);
        if(U_SUCCESS(internalErrorCode) && len < UPRV_LENGTHOF(defaultType)) {
            u_UCharsToChars(s, defaultType, len + 1);
        } else {
            uprv_strcpy(defaultType, "standard");
        }
    }
    t->actualLocale = locale;
    if(uprv_strcmp(type, defaultType) != 0) {
        t->actualLocale.setKeywordValue("collation", type, errorCode);
    } else if(uprv_strcmp(locale.getName(), locale.getBaseName()) != 0) {
        // Remove the collation keyword if it was set.
        t->actualLocale.setKeywordValue("collation", NULL, errorCode);
    }
    if(U_FAILURE(errorCode)) { return NULL; }

    if(typeFallback) {
        errorCode = U_USING_DEFAULT_WARNING;
    }
    t->bundle = bundle;
    bundle = NULL;
    const CollationCacheEntry *entry = new CollationCacheEntry(validLocale, t.getAlias());
    if(entry == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    } else {
        t.orphan();
    }
    // Have to add that reference that we promise.
    entry->addRef();
    return entry;
}

const CollationCacheEntry *
CollationLoader::getCacheEntry(UErrorCode &errorCode) {
    LocaleCacheKey<CollationCacheEntry> key(locale);
    const CollationCacheEntry *entry = NULL;
    cache->get(key, this, entry, errorCode);
    return entry;
}

const CollationCacheEntry *
CollationLoader::makeCacheEntryFromRoot(
        const Locale &/*loc*/,
        UErrorCode &errorCode) const {
    if (U_FAILURE(errorCode)) {
        return NULL;
    }
    rootEntry->addRef();
    return makeCacheEntry(validLocale, rootEntry, errorCode);
}

const CollationCacheEntry *
CollationLoader::makeCacheEntry(
        const Locale &loc,
        const CollationCacheEntry *entryFromCache,
        UErrorCode &errorCode) {
    if(U_FAILURE(errorCode) || loc == entryFromCache->validLocale) {
        return entryFromCache;
    }
    CollationCacheEntry *entry = new CollationCacheEntry(loc, entryFromCache->tailoring);
    if(entry == NULL) {
        errorCode = U_MEMORY_ALLOCATION_ERROR;
        entryFromCache->removeRef();
        return NULL;
    }
    entry->addRef();
    entryFromCache->removeRef();
    return entry;
}

U_NAMESPACE_END

U_NAMESPACE_USE

U_CAPI UCollator*
ucol_open(const char *loc,
          UErrorCode *status)
{
    UTRACE_ENTRY_OC(UTRACE_UCOL_OPEN);
    UTRACE_DATA1(UTRACE_INFO, "locale = \"%s\"", loc);
    UCollator *result = NULL;

    Collator *coll = Collator::createInstance(loc, *status);
    if(U_SUCCESS(*status)) {
        result = coll->toUCollator();
    }
    UTRACE_EXIT_PTR_STATUS(result, *status);
    return result;
}


U_CAPI int32_t U_EXPORT2
ucol_getDisplayName(    const    char        *objLoc,
                    const    char        *dispLoc,
                    UChar             *result,
                    int32_t         resultLength,
                    UErrorCode        *status)
{
    if(U_FAILURE(*status)) return -1;
    UnicodeString dst;
    if(!(result==NULL && resultLength==0)) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        dst.setTo(result, 0, resultLength);
    }
    Collator::getDisplayName(Locale(objLoc), Locale(dispLoc), dst);
    return dst.extract(result, resultLength, *status);
}

U_CAPI const char* U_EXPORT2
ucol_getAvailable(int32_t index)
{
    int32_t count = 0;
    const Locale *loc = Collator::getAvailableLocales(count);
    if (loc != NULL && index < count) {
        return loc[index].getName();
    }
    return NULL;
}

U_CAPI int32_t U_EXPORT2
ucol_countAvailable()
{
    int32_t count = 0;
    Collator::getAvailableLocales(count);
    return count;
}

#if !UCONFIG_NO_SERVICE
U_CAPI UEnumeration* U_EXPORT2
ucol_openAvailableLocales(UErrorCode *status) {
    // This is a wrapper over Collator::getAvailableLocales()
    if (U_FAILURE(*status)) {
        return NULL;
    }
    StringEnumeration *s = icu::Collator::getAvailableLocales();
    if (s == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    return uenum_openFromStringEnumeration(s, status);
}
#endif

// Note: KEYWORDS[0] != RESOURCE_NAME - alan

static const char RESOURCE_NAME[] = "collations";

static const char* const KEYWORDS[] = { "collation" };

#define KEYWORD_COUNT UPRV_LENGTHOF(KEYWORDS)

U_CAPI UEnumeration* U_EXPORT2
ucol_getKeywords(UErrorCode *status) {
    UEnumeration *result = NULL;
    if (U_SUCCESS(*status)) {
        return uenum_openCharStringsEnumeration(KEYWORDS, KEYWORD_COUNT, status);
    }
    return result;
}

U_CAPI UEnumeration* U_EXPORT2
ucol_getKeywordValues(const char *keyword, UErrorCode *status) {
    if (U_FAILURE(*status)) {
        return NULL;
    }
    // hard-coded to accept exactly one collation keyword
    // modify if additional collation keyword is added later
    if (keyword==NULL || uprv_strcmp(keyword, KEYWORDS[0])!=0)
    {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    return ures_getKeywordValues(U_ICUDATA_COLL, RESOURCE_NAME, status);
}

static const UEnumeration ucol_res_defaultKeywordValues = {
    NULL,
    NULL,
    ulist_close_keyword_values_iterator,
    ulist_count_keyword_values,
    uenum_unextDefault,
    ulist_next_keyword_value,
    ulist_reset_keyword_values_iterator
};

namespace {

struct KeywordsSink : public ResourceSink {
public:
    KeywordsSink(UErrorCode &errorCode) :
            values(ulist_createEmptyList(&errorCode)), hasDefault(FALSE) {}
    virtual ~KeywordsSink();

    virtual void put(const char *key, ResourceValue &value, UBool /*noFallback*/,
                     UErrorCode &errorCode) {
        if (U_FAILURE(errorCode)) { return; }
        ResourceTable collations = value.getTable(errorCode);
        for (int32_t i = 0; collations.getKeyAndValue(i, key, value); ++i) {
            UResType type = value.getType();
            if (type == URES_STRING) {
                if (!hasDefault && uprv_strcmp(key, "default") == 0) {
                    CharString defcoll;
                    defcoll.appendInvariantChars(value.getUnicodeString(errorCode), errorCode);
                    if (U_SUCCESS(errorCode) && !defcoll.isEmpty()) {
                        char *ownedDefault = uprv_strdup(defcoll.data());
                        if (ownedDefault == NULL) {
                            errorCode = U_MEMORY_ALLOCATION_ERROR;
                            return;
                        }
                        ulist_removeString(values, defcoll.data());
                        ulist_addItemBeginList(values, ownedDefault, TRUE, &errorCode);
                        hasDefault = TRUE;
                    }
                }
            } else if (type == URES_TABLE && uprv_strncmp(key, "private-", 8) != 0) {
                if (!ulist_containsString(values, key, (int32_t)uprv_strlen(key))) {
                    ulist_addItemEndList(values, key, FALSE, &errorCode);
                }
            }
            if (U_FAILURE(errorCode)) { return; }
        }
    }

    UList *values;
    UBool hasDefault;
};

KeywordsSink::~KeywordsSink() {
    ulist_deleteList(values);
}

}  // namespace

U_CAPI UEnumeration* U_EXPORT2
ucol_getKeywordValuesForLocale(const char* /*key*/, const char* locale,
                               UBool /*commonlyUsed*/, UErrorCode* status) {
    // Note: The parameter commonlyUsed is not used.
    // The switch is in the method signature for consistency
    // with other locale services.

    // Read available collation values from collation bundles.
    LocalUResourceBundlePointer bundle(ures_open(U_ICUDATA_COLL, locale, status));
    KeywordsSink sink(*status);
    ures_getAllItemsWithFallback(bundle.getAlias(), RESOURCE_NAME, sink, *status);
    if (U_FAILURE(*status)) { return NULL; }

    UEnumeration *en = (UEnumeration *)uprv_malloc(sizeof(UEnumeration));
    if (en == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    memcpy(en, &ucol_res_defaultKeywordValues, sizeof(UEnumeration));
    ulist_resetList(sink.values);  // Initialize the iterator.
    en->context = sink.values;
    sink.values = NULL;  // Avoid deletion in the sink destructor.
    return en;
}

U_CAPI int32_t U_EXPORT2
ucol_getFunctionalEquivalent(char* result, int32_t resultCapacity,
                             const char* keyword, const char* locale,
                             UBool* isAvailable, UErrorCode* status)
{
    // N.B.: Resource name is "collations" but keyword is "collation"
    return ures_getFunctionalEquivalent(result, resultCapacity, U_ICUDATA_COLL,
        "collations", keyword, locale,
        isAvailable, TRUE, status);
}

#endif /* #if !UCONFIG_NO_COLLATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2004-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  ucol_sit.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
* Modification history
* Date        Name      Comments
* 03/12/2004  weiv      Creation
*/











#ifdef UCOL_TRACE_SIT
# include <stdio.h>
#endif

#if !UCONFIG_NO_COLLATION



enum OptionsList {
    UCOL_SIT_LANGUAGE = 0,
    UCOL_SIT_SCRIPT   = 1,
    UCOL_SIT_REGION   = 2,
    UCOL_SIT_VARIANT  = 3,
    UCOL_SIT_KEYWORD  = 4,
    UCOL_SIT_PROVIDER = 5,
    UCOL_SIT_LOCELEMENT_MAX = UCOL_SIT_PROVIDER, /* the last element that's part of LocElements */

    UCOL_SIT_BCP47,
    UCOL_SIT_STRENGTH,
    UCOL_SIT_CASE_LEVEL,
    UCOL_SIT_CASE_FIRST,
    UCOL_SIT_NUMERIC_COLLATION,
    UCOL_SIT_ALTERNATE_HANDLING,
    UCOL_SIT_NORMALIZATION_MODE,
    UCOL_SIT_FRENCH_COLLATION,
    UCOL_SIT_HIRAGANA_QUATERNARY,
    UCOL_SIT_VARIABLE_TOP,
    UCOL_SIT_VARIABLE_TOP_VALUE,
    UCOL_SIT_ITEMS_COUNT
};

/* option starters chars. */
static const char alternateHArg     = 'A';
static const char variableTopValArg = 'B';
static const char caseFirstArg      = 'C';
static const char numericCollArg    = 'D';
static const char caseLevelArg      = 'E';
static const char frenchCollArg     = 'F';
static const char hiraganaQArg      = 'H';
static const char keywordArg        = 'K';
static const char languageArg       = 'L';
static const char normArg           = 'N';
static const char providerArg       = 'P';
static const char regionArg         = 'R';
static const char strengthArg       = 'S';
static const char variableTopArg    = 'T';
static const char variantArg        = 'V';
static const char RFC3066Arg        = 'X';
static const char scriptArg         = 'Z';

static const char collationKeyword[]  = "@collation=";
static const char providerKeyword[]  = "@sp=";


static const int32_t locElementCount = UCOL_SIT_LOCELEMENT_MAX+1;
static const int32_t locElementCapacity = 32;
static const int32_t loc3066Capacity = 256;
static const int32_t locProviderCapacity = 10;
static const int32_t ucol_sit_internalBufferSize = 512;

/* structure containing specification of a collator. Initialized
 * from a short string. Also used to construct a short string from a
 * collator instance
 */
struct CollatorSpec {
    char locElements[locElementCount][locElementCapacity];
    char locale[loc3066Capacity];
    char provider[locProviderCapacity];
    UColAttributeValue options[UCOL_ATTRIBUTE_COUNT];
    uint32_t variableTopValue;
    UChar variableTopString[locElementCapacity];
    int32_t variableTopStringLen;
    UBool variableTopSet;
    struct {
        const char *start;
        int32_t len;
    } entries[UCOL_SIT_ITEMS_COUNT];
};


/* structure for converting between character attribute
 * representation and real collation attribute value.
 */
struct AttributeConversion {
    char letter;
    UColAttributeValue value;
};

static const AttributeConversion conversions[12] = {
    { '1', UCOL_PRIMARY },
    { '2', UCOL_SECONDARY },
    { '3', UCOL_TERTIARY },
    { '4', UCOL_QUATERNARY },
    { 'D', UCOL_DEFAULT },
    { 'I', UCOL_IDENTICAL },
    { 'L', UCOL_LOWER_FIRST },
    { 'N', UCOL_NON_IGNORABLE },
    { 'O', UCOL_ON },
    { 'S', UCOL_SHIFTED },
    { 'U', UCOL_UPPER_FIRST },
    { 'X', UCOL_OFF }
};


static UColAttributeValue
ucol_sit_letterToAttributeValue(char letter, UErrorCode *status) {
    uint32_t i = 0;
    for(i = 0; i < UPRV_LENGTHOF(conversions); i++) {
        if(conversions[i].letter == letter) {
            return conversions[i].value;
        }
    }
    *status = U_ILLEGAL_ARGUMENT_ERROR;
#ifdef UCOL_TRACE_SIT
    fprintf(stderr, "%s:%d: unknown letter %c: %s\n", __FILE__, __LINE__, letter, u_errorName(*status));
#endif
    return UCOL_DEFAULT;
}

/* function prototype for functions used to parse a short string */
U_CDECL_BEGIN
typedef const char* U_CALLCONV
ActionFunction(CollatorSpec *spec, uint32_t value1, const char* string,
               UErrorCode *status);
U_CDECL_END

U_CDECL_BEGIN
static const char* U_CALLCONV
_processLocaleElement(CollatorSpec *spec, uint32_t value, const char* string,
                      UErrorCode *status)
{
    int32_t len = 0;
    do {
        if(value == UCOL_SIT_LANGUAGE || value == UCOL_SIT_KEYWORD || value == UCOL_SIT_PROVIDER) {
            spec->locElements[value][len++] = uprv_tolower(*string);
        } else {
            spec->locElements[value][len++] = *string;
        }
    } while(*(++string) != '_' && *string && len < locElementCapacity);
    if(len >= locElementCapacity) {
        *status = U_BUFFER_OVERFLOW_ERROR;
        return string;
    }
    // don't skip the underscore at the end
    return string;
}
U_CDECL_END

U_CDECL_BEGIN
static const char* U_CALLCONV
_processRFC3066Locale(CollatorSpec *spec, uint32_t, const char* string,
                      UErrorCode *status)
{
    char terminator = *string;
    string++;
    const char *end = uprv_strchr(string+1, terminator);
    if(end == NULL || end - string >= loc3066Capacity) {
        *status = U_BUFFER_OVERFLOW_ERROR;
        return string;
    } else {
        uprv_strncpy(spec->locale, string, end-string);
        return end+1;
    }
}

U_CDECL_END

U_CDECL_BEGIN
static const char* U_CALLCONV
_processCollatorOption(CollatorSpec *spec, uint32_t option, const char* string,
                       UErrorCode *status)
{
    spec->options[option] = ucol_sit_letterToAttributeValue(*string, status);
    if((*(++string) != '_' && *string) || U_FAILURE(*status)) {
#ifdef UCOL_TRACE_SIT
    fprintf(stderr, "%s:%d: unknown collator option at '%s': %s\n", __FILE__, __LINE__, string, u_errorName(*status));
#endif
        *status = U_ILLEGAL_ARGUMENT_ERROR;
    }
    return string;
}
U_CDECL_END


static UChar
readHexCodeUnit(const char **string, UErrorCode *status)
{
    UChar result = 0;
    int32_t value = 0;
    char c;
    int32_t noDigits = 0;
    while((c = **string) != 0 && noDigits < 4) {
        if( c >= '0' && c <= '9') {
            value = c - '0';
        } else if ( c >= 'a' && c <= 'f') {
            value = c - 'a' + 10;
        } else if ( c >= 'A' && c <= 'F') {
            value = c - 'A' + 10;
        } else {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
#ifdef UCOL_TRACE_SIT
            fprintf(stderr, "%s:%d: Bad hex char at '%s': %s\n", __FILE__, __LINE__, *string, u_errorName(*status));
#endif
            return 0;
        }
        result = (result << 4) | (UChar)value;
        noDigits++;
        (*string)++;
    }
    // if the string was terminated before we read 4 digits, set an error
    if(noDigits < 4) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
#ifdef UCOL_TRACE_SIT
        fprintf(stderr, "%s:%d: Short (only %d digits, wanted 4) at '%s': %s\n", __FILE__, __LINE__, noDigits,*string, u_errorName(*status));
#endif
    }
    return result;
}

U_CDECL_BEGIN
static const char* U_CALLCONV
_processVariableTop(CollatorSpec *spec, uint32_t value1, const char* string, UErrorCode *status)
{
    // get four digits
    int32_t i = 0;
    if(!value1) {
        while(U_SUCCESS(*status) && i < locElementCapacity && *string != 0 && *string != '_') {
            spec->variableTopString[i++] = readHexCodeUnit(&string, status);
        }
        spec->variableTopStringLen = i;
        if(i == locElementCapacity && *string != 0 && *string != '_') {
            *status = U_BUFFER_OVERFLOW_ERROR;
        }
    } else {
        spec->variableTopValue = readHexCodeUnit(&string, status);
    }
    if(U_SUCCESS(*status)) {
        spec->variableTopSet = TRUE;
    }
    return string;
}
U_CDECL_END


/* Table for parsing short strings */
struct ShortStringOptions {
    char optionStart;
    ActionFunction *action;
    uint32_t attr;
};

static const ShortStringOptions options[UCOL_SIT_ITEMS_COUNT] =
{
/* 10 ALTERNATE_HANDLING */   {alternateHArg,     _processCollatorOption, UCOL_ALTERNATE_HANDLING }, // alternate  N, S, D
/* 15 VARIABLE_TOP_VALUE */   {variableTopValArg, _processVariableTop,    1 },
/* 08 CASE_FIRST */           {caseFirstArg,      _processCollatorOption, UCOL_CASE_FIRST }, // case first L, U, X, D
/* 09 NUMERIC_COLLATION */    {numericCollArg,    _processCollatorOption, UCOL_NUMERIC_COLLATION }, // codan      O, X, D
/* 07 CASE_LEVEL */           {caseLevelArg,      _processCollatorOption, UCOL_CASE_LEVEL }, // case level O, X, D
/* 12 FRENCH_COLLATION */     {frenchCollArg,     _processCollatorOption, UCOL_FRENCH_COLLATION }, // french     O, X, D
/* 13 HIRAGANA_QUATERNARY] */ {hiraganaQArg,      _processCollatorOption, UCOL_HIRAGANA_QUATERNARY_MODE }, // hiragana   O, X, D
/* 04 KEYWORD */              {keywordArg,        _processLocaleElement,  UCOL_SIT_KEYWORD }, // keyword
/* 00 LANGUAGE */             {languageArg,       _processLocaleElement,  UCOL_SIT_LANGUAGE }, // language
/* 11 NORMALIZATION_MODE */   {normArg,           _processCollatorOption, UCOL_NORMALIZATION_MODE }, // norm       O, X, D
/* 02 REGION */               {regionArg,         _processLocaleElement,  UCOL_SIT_REGION }, // region
/* 06 STRENGTH */             {strengthArg,       _processCollatorOption, UCOL_STRENGTH }, // strength   1, 2, 3, 4, I, D
/* 14 VARIABLE_TOP */         {variableTopArg,    _processVariableTop,    0 },
/* 03 VARIANT */              {variantArg,        _processLocaleElement,  UCOL_SIT_VARIANT }, // variant
/* 05 RFC3066BIS */           {RFC3066Arg,        _processRFC3066Locale,  0 }, // rfc3066bis locale name
/* 01 SCRIPT */               {scriptArg,         _processLocaleElement,  UCOL_SIT_SCRIPT },  // script
/*    PROVIDER */             {providerArg,       _processLocaleElement, UCOL_SIT_PROVIDER }
};


static
const char* ucol_sit_readOption(const char *start, CollatorSpec *spec,
                            UErrorCode *status)
{
  int32_t i = 0;

  for(i = 0; i < UCOL_SIT_ITEMS_COUNT; i++) {
      if(*start == options[i].optionStart) {
          spec->entries[i].start = start;
          const char* end = options[i].action(spec, options[i].attr, start+1, status);
          spec->entries[i].len = (int32_t)(end - start);
          return end;
      }
  }
  *status = U_ILLEGAL_ARGUMENT_ERROR;
#ifdef UCOL_TRACE_SIT
  fprintf(stderr, "%s:%d: Unknown option at '%s': %s\n", __FILE__, __LINE__, start, u_errorName(*status));
#endif
  return start;
}

static
void ucol_sit_initCollatorSpecs(CollatorSpec *spec)
{
    // reset everything
    uprv_memset(spec, 0, sizeof(CollatorSpec));
    // set collation options to default
    int32_t i = 0;
    for(i = 0; i < UCOL_ATTRIBUTE_COUNT; i++) {
        spec->options[i] = UCOL_DEFAULT;
    }
}

static const char*
ucol_sit_readSpecs(CollatorSpec *s, const char *string,
                        UParseError *parseError, UErrorCode *status)
{
    const char *definition = string;
    while(U_SUCCESS(*status) && *string) {
        string = ucol_sit_readOption(string, s, status);
        // advance over '_'
        while(*string && *string == '_') {
            string++;
        }
    }
    if(U_FAILURE(*status)) {
        parseError->offset = (int32_t)(string - definition);
    }
    return string;
}

static
int32_t ucol_sit_dumpSpecs(CollatorSpec *s, char *destination, int32_t capacity, UErrorCode *status)
{
    int32_t i = 0, j = 0;
    int32_t len = 0;
    char optName;
    if(U_SUCCESS(*status)) {
        for(i = 0; i < UCOL_SIT_ITEMS_COUNT; i++) {
            if(s->entries[i].start) {
                if(len) {
                    if(len < capacity) {
                        uprv_strcat(destination, "_");
                    }
                    len++;
                }
                optName = *(s->entries[i].start);
                if(optName == languageArg || optName == regionArg || optName == variantArg || optName == keywordArg) {
                    for(j = 0; j < s->entries[i].len; j++) {
                        if(len + j < capacity) {
                            destination[len+j] = uprv_toupper(*(s->entries[i].start+j));
                        }
                    }
                    len += s->entries[i].len;
                } else {
                    len += s->entries[i].len;
                    if(len < capacity) {
                        uprv_strncat(destination,s->entries[i].start, s->entries[i].len);
                    }
                }
            }
        }
        return len;
    } else {
        return 0;
    }
}

static void
ucol_sit_calculateWholeLocale(CollatorSpec *s) {
    // put the locale together, unless we have a done
    // locale
    if(s->locale[0] == 0) {
        // first the language
        uprv_strcat(s->locale, s->locElements[UCOL_SIT_LANGUAGE]);
        // then the script, if present
        if(*(s->locElements[UCOL_SIT_SCRIPT])) {
            uprv_strcat(s->locale, "_");
            uprv_strcat(s->locale, s->locElements[UCOL_SIT_SCRIPT]);
        }
        // then the region, if present
        if(*(s->locElements[UCOL_SIT_REGION])) {
            uprv_strcat(s->locale, "_");
            uprv_strcat(s->locale, s->locElements[UCOL_SIT_REGION]);
        } else if(*(s->locElements[UCOL_SIT_VARIANT])) { // if there is a variant, we need an underscore
            uprv_strcat(s->locale, "_");
        }
        // add variant, if there
        if(*(s->locElements[UCOL_SIT_VARIANT])) {
            uprv_strcat(s->locale, "_");
            uprv_strcat(s->locale, s->locElements[UCOL_SIT_VARIANT]);
        }

        // if there is a collation keyword, add that too
        if(*(s->locElements[UCOL_SIT_KEYWORD])) {
            uprv_strcat(s->locale, collationKeyword);
            uprv_strcat(s->locale, s->locElements[UCOL_SIT_KEYWORD]);
        }

        // if there is a provider keyword, add that too
        if(*(s->locElements[UCOL_SIT_PROVIDER])) {
            uprv_strcat(s->locale, providerKeyword);
            uprv_strcat(s->locale, s->locElements[UCOL_SIT_PROVIDER]);
        }
    }
}


U_CAPI void U_EXPORT2
ucol_prepareShortStringOpen( const char *definition,
                          UBool,
                          UParseError *parseError,
                          UErrorCode *status)
{
    if(U_FAILURE(*status)) return;

    UParseError internalParseError;

    if(!parseError) {
        parseError = &internalParseError;
    }
    parseError->line = 0;
    parseError->offset = 0;
    parseError->preContext[0] = 0;
    parseError->postContext[0] = 0;


    // first we want to pick stuff out of short string.
    // we'll end up with an UCA version, locale and a bunch of
    // settings

    // analyse the string in order to get everything we need.
    CollatorSpec s;
    ucol_sit_initCollatorSpecs(&s);
    ucol_sit_readSpecs(&s, definition, parseError, status);
    ucol_sit_calculateWholeLocale(&s);

    char buffer[ucol_sit_internalBufferSize];
    uprv_memset(buffer, 0, ucol_sit_internalBufferSize);
    uloc_canonicalize(s.locale, buffer, ucol_sit_internalBufferSize, status);

    UResourceBundle *b = ures_open(U_ICUDATA_COLL, buffer, status);
    /* we try to find stuff from keyword */
    UResourceBundle *collations = ures_getByKey(b, "collations", NULL, status);
    UResourceBundle *collElem = NULL;
    char keyBuffer[256];
    // if there is a keyword, we pick it up and try to get elements
    int32_t keyLen = uloc_getKeywordValue(buffer, "collation", keyBuffer, sizeof(keyBuffer), status);
    // Treat too long a value as no keyword.
    if(keyLen >= (int32_t)sizeof(keyBuffer)) {
      keyLen = 0;
      *status = U_ZERO_ERROR;
    }
    if(keyLen == 0) {
      // no keyword
      // we try to find the default setting, which will give us the keyword value
      UResourceBundle *defaultColl = ures_getByKeyWithFallback(collations, "default", NULL, status);
      if(U_SUCCESS(*status)) {
        int32_t defaultKeyLen = 0;
        const UChar *defaultKey = ures_getString(defaultColl, &defaultKeyLen, status);
        u_UCharsToChars(defaultKey, keyBuffer, defaultKeyLen);
        keyBuffer[defaultKeyLen] = 0;
      } else {
        *status = U_INTERNAL_PROGRAM_ERROR;
        return;
      }
      ures_close(defaultColl);
    }
    collElem = ures_getByKeyWithFallback(collations, keyBuffer, collElem, status);
    ures_close(collElem);
    ures_close(collations);
    ures_close(b);
}


U_CAPI UCollator* U_EXPORT2
ucol_openFromShortString( const char *definition,
                          UBool forceDefaults,
                          UParseError *parseError,
                          UErrorCode *status)
{
    UTRACE_ENTRY_OC(UTRACE_UCOL_OPEN_FROM_SHORT_STRING);
    UTRACE_DATA1(UTRACE_INFO, "short string = \"%s\"", definition);

    if(U_FAILURE(*status)) return 0;

    UParseError internalParseError;

    if(!parseError) {
        parseError = &internalParseError;
    }
    parseError->line = 0;
    parseError->offset = 0;
    parseError->preContext[0] = 0;
    parseError->postContext[0] = 0;


    // first we want to pick stuff out of short string.
    // we'll end up with an UCA version, locale and a bunch of
    // settings

    // analyse the string in order to get everything we need.
    const char *string = definition;
    CollatorSpec s;
    ucol_sit_initCollatorSpecs(&s);
    string = ucol_sit_readSpecs(&s, definition, parseError, status);
    ucol_sit_calculateWholeLocale(&s);

    char buffer[ucol_sit_internalBufferSize];
    uprv_memset(buffer, 0, ucol_sit_internalBufferSize);
    uloc_canonicalize(s.locale, buffer, ucol_sit_internalBufferSize, status);

    UCollator *result = ucol_open(buffer, status);
    int32_t i = 0;

    for(i = 0; i < UCOL_ATTRIBUTE_COUNT; i++) {
        if(s.options[i] != UCOL_DEFAULT) {
            if(forceDefaults || ucol_getAttribute(result, (UColAttribute)i, status) != s.options[i]) {
                ucol_setAttribute(result, (UColAttribute)i, s.options[i], status);
            }

            if(U_FAILURE(*status)) {
                parseError->offset = (int32_t)(string - definition);
                ucol_close(result);
                return NULL;
            }

        }
    }
    if(s.variableTopSet) {
        if(s.variableTopString[0]) {
            ucol_setVariableTop(result, s.variableTopString, s.variableTopStringLen, status);
        } else { // we set by value, using 'B'
            ucol_restoreVariableTop(result, s.variableTopValue, status);
        }
    }


    if(U_FAILURE(*status)) { // here it can only be a bogus value
        ucol_close(result);
        result = NULL;
    }

    UTRACE_EXIT_PTR_STATUS(result, *status);
    return result;
}


U_CAPI int32_t U_EXPORT2
ucol_getShortDefinitionString(const UCollator *coll,
                              const char *locale,
                              char *dst,
                              int32_t capacity,
                              UErrorCode *status)
{
    if(U_FAILURE(*status)) return 0;
    if(coll == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    return ((icu::Collator*)coll)->internalGetShortDefinitionString(locale,dst,capacity,*status);
}

U_CAPI int32_t U_EXPORT2
ucol_normalizeShortDefinitionString(const char *definition,
                                    char *destination,
                                    int32_t capacity,
                                    UParseError *parseError,
                                    UErrorCode *status)
{

    if(U_FAILURE(*status)) {
        return 0;
    }

    if(destination) {
        uprv_memset(destination, 0, capacity*sizeof(char));
    }

    UParseError pe;
    if(!parseError) {
        parseError = &pe;
    }

    // validate
    CollatorSpec s;
    ucol_sit_initCollatorSpecs(&s);
    ucol_sit_readSpecs(&s, definition, parseError, status);
    return ucol_sit_dumpSpecs(&s, destination, capacity, status);
}

/**
 * Get a set containing the contractions defined by the collator. The set includes
 * both the UCA contractions and the contractions defined by the collator
 * @param coll collator
 * @param conts the set to hold the result
 * @param status to hold the error code
 * @return the size of the contraction set
 */
U_CAPI int32_t U_EXPORT2
ucol_getContractions( const UCollator *coll,
                  USet *contractions,
                  UErrorCode *status)
{
  ucol_getContractionsAndExpansions(coll, contractions, NULL, FALSE, status);
  return uset_getItemCount(contractions);
}

/**
 * Get a set containing the expansions defined by the collator. The set includes
 * both the UCA expansions and the expansions defined by the tailoring
 * @param coll collator
 * @param conts the set to hold the result
 * @param addPrefixes add the prefix contextual elements to contractions
 * @param status to hold the error code
 *
 * @draft ICU 3.4
 */
U_CAPI void U_EXPORT2
ucol_getContractionsAndExpansions( const UCollator *coll,
                  USet *contractions,
                  USet *expansions,
                  UBool addPrefixes,
                  UErrorCode *status)
{
    if(U_FAILURE(*status)) {
        return;
    }
    if(coll == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    const icu::RuleBasedCollator *rbc = icu::RuleBasedCollator::rbcFromUCollator(coll);
    if(rbc == NULL) {
        *status = U_UNSUPPORTED_ERROR;
        return;
    }
    rbc->internalGetContractionsAndExpansions(
            icu::UnicodeSet::fromUSet(contractions),
            icu::UnicodeSet::fromUSet(expansions),
            addPrefixes, *status);
}
#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*   Copyright (C) 2001-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
******************************************************************************
*
* File ucoleitr.cpp
*
* Modification History:
*
* Date        Name        Description
* 02/15/2001  synwee      Modified all methods to process its own function
*                         instead of calling the equivalent c++ api (coleitr.h)
* 2012-2014   markus      Rewritten in C++ again.
******************************************************************************/



#if !UCONFIG_NO_COLLATION








// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2001-2015 IBM and others. All rights reserved.
**********************************************************************
*   Date        Name        Description
*  08/13/2001   synwee      Creation.
**********************************************************************
*/



#if !UCONFIG_NO_COLLATION






/* mask off anything but primary order */
#define UCOL_PRIMARYORDERMASK 0xffff0000
/* mask off anything but secondary order */
#define UCOL_SECONDARYORDERMASK 0x0000ff00
/* mask off anything but tertiary order */
#define UCOL_TERTIARYORDERMASK 0x000000ff
/* primary order shift */
#define UCOL_PRIMARYORDERSHIFT 16
/* secondary order shift */
#define UCOL_SECONDARYORDERSHIFT 8

#define UCOL_IGNORABLE 0

/* get weights from a CE */
#define UCOL_PRIMARYORDER(order) (((order) >> 16) & 0xffff)
#define UCOL_SECONDARYORDER(order) (((order) & UCOL_SECONDARYORDERMASK)>> UCOL_SECONDARYORDERSHIFT)
#define UCOL_TERTIARYORDER(order) ((order) & UCOL_TERTIARYORDERMASK)

#define UCOL_CONTINUATION_MARKER 0xC0

#define isContinuation(CE) (((CE) & UCOL_CONTINUATION_MARKER) == UCOL_CONTINUATION_MARKER)

/**
 * This indicates an error has occured during processing or there are no more CEs
 * to be returned.
 */
#define UCOL_PROCESSED_NULLORDER        ((int64_t)U_INT64_MAX)

U_NAMESPACE_BEGIN

class CollationElementIterator;
class Collator;

struct PCEI
{
    uint64_t ce;
    int32_t  low;
    int32_t  high;
};

struct PCEBuffer
{
    PCEI    defaultBuffer[16];
    PCEI   *buffer;
    int32_t bufferIndex;
    int32_t bufferSize;

    PCEBuffer();
    ~PCEBuffer();

    void  reset();
    UBool isEmpty() const;
    void  put(uint64_t ce, int32_t ixLow, int32_t ixHigh, UErrorCode &errorCode);
    const PCEI *get();
};

class UCollationPCE : public UMemory {
private:
    PCEBuffer          pceBuffer;
    CollationElementIterator *cei;
    UCollationStrength strength;
    UBool              toShift;
    UBool              isShifted;
    uint32_t           variableTop;

public:
    UCollationPCE(UCollationElements *elems);
    UCollationPCE(CollationElementIterator *iter);
    ~UCollationPCE();

    void init(UCollationElements *elems);
    void init(CollationElementIterator *iter);

    /**
     * Get the processed ordering priority of the next collation element in the text.
     * A single character may contain more than one collation element.
     *
     * @param ixLow a pointer to an int32_t to receive the iterator index before fetching the CE.
     * @param ixHigh a pointer to an int32_t to receive the iterator index after fetching the CE.
     * @param status A pointer to an UErrorCode to receive any errors.
     * @return The next collation elements ordering, otherwise returns UCOL_PROCESSED_NULLORDER
     *         if an error has occured or if the end of string has been reached
     */
    int64_t nextProcessed(int32_t *ixLow, int32_t *ixHigh, UErrorCode *status);
    /**
     * Get the processed ordering priority of the previous collation element in the text.
     * A single character may contain more than one collation element.
     *
     * @param ixLow A pointer to an int32_t to receive the iterator index after fetching the CE
     * @param ixHigh A pointer to an int32_t to receiver the iterator index before fetching the CE
     * @param status A pointer to an UErrorCode to receive any errors. Noteably
     *               a U_BUFFER_OVERFLOW_ERROR is returned if the internal stack
     *               buffer has been exhausted.
     * @return The previous collation elements ordering, otherwise returns
     *         UCOL_PROCESSED_NULLORDER if an error has occured or if the start of
     *         string has been reached.
     */
    int64_t previousProcessed(int32_t *ixLow, int32_t *ixHigh, UErrorCode *status);

private:
    void init(const Collator &coll);
    uint64_t processCE(uint32_t ce);
};

U_NAMESPACE_END

#define INITIAL_ARRAY_SIZE_       256
#define MAX_TABLE_SIZE_           257

struct USearch {
    // required since collation element iterator does not have a getText API
    const UChar              *text;
          int32_t             textLength; // exact length
          UBool               isOverlap;
          UBool               isCanonicalMatch;
          int16_t             elementComparisonType;
          UBreakIterator     *internalBreakIter;  //internal character breakiterator
          UBreakIterator     *breakIter;
    // value USEARCH_DONE is the default value
    // if we are not at the start of the text or the end of the text,
    // depending on the iteration direction and matchedIndex is USEARCH_DONE
    // it means that we can't find any more matches in that particular direction
          int32_t             matchedIndex;
          int32_t             matchedLength;
          UBool               isForwardSearching;
          UBool               reset;
};

struct UPattern {
    const UChar              *text;
          int32_t             textLength; // exact length
          // length required for backwards ce comparison
          int32_t             cesLength;
          int32_t            *ces;
          int32_t             cesBuffer[INITIAL_ARRAY_SIZE_];
          int32_t             pcesLength;
          int64_t            *pces;
          int64_t             pcesBuffer[INITIAL_ARRAY_SIZE_];
          UBool               hasPrefixAccents;
          UBool               hasSuffixAccents;
          int16_t             defaultShiftSize;
          int16_t             shift[MAX_TABLE_SIZE_];
          int16_t             backShift[MAX_TABLE_SIZE_];
};

struct UStringSearch {
    struct USearch            *search;
    struct UPattern            pattern;
    const  UCollator          *collator;
    const  icu::Normalizer2   *nfd;
    // positions within the collation element iterator is used to determine
    // if we are at the start of the text.
           UCollationElements *textIter;
           icu::UCollationPCE *textProcessedIter;
    // utility collation element, used throughout program for temporary
    // iteration.
           UCollationElements *utilIter;
           UBool               ownCollator;
           UCollationStrength  strength;
           uint32_t            ceMask;
           uint32_t            variableTop;
           UBool               toShift;
           UChar               canonicalPrefixAccents[INITIAL_ARRAY_SIZE_];
           UChar               canonicalSuffixAccents[INITIAL_ARRAY_SIZE_];
};

/**
* Exact matches without checking for the ends for extra accents.
* The match after the position within the collation element iterator is to be
* found.
* After a match is found the offset in the collation element iterator will be
* shifted to the start of the match.
* Implementation note:
* For tertiary we can't use the collator->tertiaryMask, that is a
* preprocessed mask that takes into account case options. since we are only
* concerned with exact matches, we don't need that.
* Alternate handling - since only the 16 most significant digits is only used,
* we can safely do a compare without masking if the ce is a variable, we mask
* and get only the primary values no shifting to quartenary is required since
* all primary values less than variabletop will need to be masked off anyway.
* If the end character is composite and the pattern ce does not match the text
* ce, we skip it until we find a match in the end composite character or when
* it has passed the character. This is so that we can match pattern "a" with
* the text "\u00e6"
* @param strsrch string search data
* @param status error status if any
* @return TRUE if an exact match is found, FALSE otherwise
*/
U_CFUNC
UBool usearch_handleNextExact(UStringSearch *strsrch, UErrorCode *status);

/**
* Canonical matches.
* According to the definition, matches found here will include the whole span
* of beginning and ending accents if it overlaps that region.
* @param strsrch string search data
* @param status error status if any
* @return TRUE if a canonical match is found, FALSE otherwise
*/
U_CFUNC
UBool usearch_handleNextCanonical(UStringSearch *strsrch, UErrorCode *status);

/**
* Gets the previous match.
* Comments follows from handleNextExact
* @param strsrch string search data
* @param status error status if any
* @return True if a exact math is found, FALSE otherwise.
*/
U_CFUNC
UBool usearch_handlePreviousExact(UStringSearch *strsrch, UErrorCode *status);

/**
* Canonical matches.
* According to the definition, matches found here will include the whole span
* of beginning and ending accents if it overlaps that region.
* @param strsrch string search data
* @param status error status if any
* @return TRUE if a canonical match is found, FALSE otherwise
*/
U_CFUNC
UBool usearch_handlePreviousCanonical(UStringSearch *strsrch,
                                      UErrorCode    *status);

#endif /* #if !UCONFIG_NO_COLLATION */



U_NAMESPACE_USE

#define ucoleitr_BUFFER_LENGTH             100

#define ucoleitr_DEFAULT_BUFFER_SIZE 16
#define ucoleitr_BUFFER_GROW 8

#define ucoleitr_ARRAY_COPY(dst, src, count) uprv_memcpy((void *) (dst), (void *) (src), (size_t)(count) * sizeof (src)[0])

#define ucoleitr_NEW_ARRAY(type, count) (type *) uprv_malloc((size_t)(count) * sizeof(type))

#define ucoleitr_DELETE_ARRAY(array) uprv_free((void *) (array))

struct RCEI
{
    uint32_t ce;
    int32_t  low;
    int32_t  high;
};

U_NAMESPACE_BEGIN

struct RCEBuffer
{
    RCEI    defaultBuffer[ucoleitr_DEFAULT_BUFFER_SIZE];
    RCEI   *buffer;
    int32_t bufferIndex;
    int32_t bufferSize;

    RCEBuffer();
    ~RCEBuffer();

    UBool isEmpty() const;
    void  put(uint32_t ce, int32_t ixLow, int32_t ixHigh, UErrorCode &errorCode);
    const RCEI *get();
};

RCEBuffer::RCEBuffer()
{
    buffer = defaultBuffer;
    bufferIndex = 0;
    bufferSize = UPRV_LENGTHOF(defaultBuffer);
}

RCEBuffer::~RCEBuffer()
{
    if (buffer != defaultBuffer) {
        ucoleitr_DELETE_ARRAY(buffer);
    }
}

UBool RCEBuffer::isEmpty() const
{
    return bufferIndex <= 0;
}

void RCEBuffer::put(uint32_t ce, int32_t ixLow, int32_t ixHigh, UErrorCode &errorCode)
{
    if (U_FAILURE(errorCode)) {
        return;
    }
    if (bufferIndex >= bufferSize) {
        RCEI *newBuffer = ucoleitr_NEW_ARRAY(RCEI, bufferSize + ucoleitr_BUFFER_GROW);
        if (newBuffer == NULL) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
        }

        ucoleitr_ARRAY_COPY(newBuffer, buffer, bufferSize);

        if (buffer != defaultBuffer) {
            ucoleitr_DELETE_ARRAY(buffer);
        }

        buffer = newBuffer;
        bufferSize += ucoleitr_BUFFER_GROW;
    }

    buffer[bufferIndex].ce   = ce;
    buffer[bufferIndex].low  = ixLow;
    buffer[bufferIndex].high = ixHigh;

    bufferIndex += 1;
}

const RCEI *RCEBuffer::get()
{
    if (bufferIndex > 0) {
     return &buffer[--bufferIndex];
    }

    return NULL;
}

PCEBuffer::PCEBuffer()
{
    buffer = defaultBuffer;
    bufferIndex = 0;
    bufferSize = UPRV_LENGTHOF(defaultBuffer);
}

PCEBuffer::~PCEBuffer()
{
    if (buffer != defaultBuffer) {
        ucoleitr_DELETE_ARRAY(buffer);
    }
}

void PCEBuffer::reset()
{
    bufferIndex = 0;
}

UBool PCEBuffer::isEmpty() const
{
    return bufferIndex <= 0;
}

void PCEBuffer::put(uint64_t ce, int32_t ixLow, int32_t ixHigh, UErrorCode &errorCode)
{
    if (U_FAILURE(errorCode)) {
        return;
    }
    if (bufferIndex >= bufferSize) {
        PCEI *newBuffer = ucoleitr_NEW_ARRAY(PCEI, bufferSize + ucoleitr_BUFFER_GROW);
        if (newBuffer == NULL) {
            errorCode = U_MEMORY_ALLOCATION_ERROR;
            return;
        }

        ucoleitr_ARRAY_COPY(newBuffer, buffer, bufferSize);

        if (buffer != defaultBuffer) {
            ucoleitr_DELETE_ARRAY(buffer);
        }

        buffer = newBuffer;
        bufferSize += ucoleitr_BUFFER_GROW;
    }

    buffer[bufferIndex].ce   = ce;
    buffer[bufferIndex].low  = ixLow;
    buffer[bufferIndex].high = ixHigh;

    bufferIndex += 1;
}

const PCEI *PCEBuffer::get()
{
    if (bufferIndex > 0) {
     return &buffer[--bufferIndex];
    }

    return NULL;
}

UCollationPCE::UCollationPCE(UCollationElements *elems) { init(elems); }

UCollationPCE::UCollationPCE(CollationElementIterator *iter) { init(iter); }

void UCollationPCE::init(UCollationElements *elems) {
    init(CollationElementIterator::fromUCollationElements(elems));
}

void UCollationPCE::init(CollationElementIterator *iter)
{
    cei = iter;
    init(*iter->rbc_);
}

void UCollationPCE::init(const Collator &coll)
{
    UErrorCode status = U_ZERO_ERROR;

    strength    = coll.getAttribute(UCOL_STRENGTH, status);
    toShift     = coll.getAttribute(UCOL_ALTERNATE_HANDLING, status) == UCOL_SHIFTED;
    isShifted   = FALSE;
    variableTop = coll.getVariableTop(status);
}

UCollationPCE::~UCollationPCE()
{
    // nothing to do
}

uint64_t UCollationPCE::processCE(uint32_t ce)
{
    uint64_t primary = 0, secondary = 0, tertiary = 0, quaternary = 0;

    // This is clean, but somewhat slow...
    // We could apply the mask to ce and then
    // just get all three orders...
    switch(strength) {
    default:
        tertiary = ucol_tertiaryOrder(ce);
        U_FALLTHROUGH;

    case UCOL_SECONDARY:
        secondary = ucol_secondaryOrder(ce);
        U_FALLTHROUGH;

    case UCOL_PRIMARY:
        primary = ucol_primaryOrder(ce);
    }

    // **** This should probably handle continuations too.  ****
    // **** That means that we need 24 bits for the primary ****
    // **** instead of the 16 that we're currently using.   ****
    // **** So we can lay out the 64 bits as: 24.12.12.16.  ****
    // **** Another complication with continuations is that ****
    // **** the *second* CE is marked as a continuation, so ****
    // **** we always have to peek ahead to know how long   ****
    // **** the primary is...                               ****
    if ((toShift && variableTop > ce && primary != 0)
                || (isShifted && primary == 0)) {

        if (primary == 0) {
            return UCOL_IGNORABLE;
        }

        if (strength >= UCOL_QUATERNARY) {
            quaternary = primary;
        }

        primary = secondary = tertiary = 0;
        isShifted = TRUE;
    } else {
        if (strength >= UCOL_QUATERNARY) {
            quaternary = 0xFFFF;
        }

        isShifted = FALSE;
    }

    return primary << 48 | secondary << 32 | tertiary << 16 | quaternary;
}

U_NAMESPACE_END

/* public methods ---------------------------------------------------- */

U_CAPI UCollationElements* U_EXPORT2
ucol_openElements(const UCollator  *coll,
                  const UChar      *text,
                        int32_t    textLength,
                        UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return NULL;
    }
    if (coll == NULL || (text == NULL && textLength != 0)) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    const RuleBasedCollator *rbc = RuleBasedCollator::rbcFromUCollator(coll);
    if (rbc == NULL) {
        *status = U_UNSUPPORTED_ERROR;  // coll is a Collator but not a RuleBasedCollator
        return NULL;
    }

    UnicodeString s((UBool)(textLength < 0), text, textLength);
    CollationElementIterator *cei = rbc->createCollationElementIterator(s);
    if (cei == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }

    return cei->toUCollationElements();
}


U_CAPI void U_EXPORT2
ucol_closeElements(UCollationElements *elems)
{
    delete CollationElementIterator::fromUCollationElements(elems);
}

U_CAPI void U_EXPORT2
ucol_reset(UCollationElements *elems)
{
    CollationElementIterator::fromUCollationElements(elems)->reset();
}

U_CAPI int32_t U_EXPORT2
ucol_next(UCollationElements *elems,
          UErrorCode         *status)
{
    if (U_FAILURE(*status)) {
        return UCOL_NULLORDER;
    }

    return CollationElementIterator::fromUCollationElements(elems)->next(*status);
}

U_NAMESPACE_BEGIN

int64_t
UCollationPCE::nextProcessed(
                   int32_t            *ixLow,
                   int32_t            *ixHigh,
                   UErrorCode         *status)
{
    int64_t result = UCOL_IGNORABLE;
    uint32_t low = 0, high = 0;

    if (U_FAILURE(*status)) {
        return UCOL_PROCESSED_NULLORDER;
    }

    pceBuffer.reset();

    do {
        low = cei->getOffset();
        int32_t ce = cei->next(*status);
        high = cei->getOffset();

        if (ce == UCOL_NULLORDER) {
             result = UCOL_PROCESSED_NULLORDER;
             break;
        }

        result = processCE((uint32_t)ce);
    } while (result == UCOL_IGNORABLE);

    if (ixLow != NULL) {
        *ixLow = low;
    }

    if (ixHigh != NULL) {
        *ixHigh = high;
    }

    return result;
}

U_NAMESPACE_END

U_CAPI int32_t U_EXPORT2
ucol_previous(UCollationElements *elems,
              UErrorCode         *status)
{
    if(U_FAILURE(*status)) {
        return UCOL_NULLORDER;
    }
    return CollationElementIterator::fromUCollationElements(elems)->previous(*status);
}

U_NAMESPACE_BEGIN

int64_t
UCollationPCE::previousProcessed(
                   int32_t            *ixLow,
                   int32_t            *ixHigh,
                   UErrorCode         *status)
{
    int64_t result = UCOL_IGNORABLE;
    int32_t  low = 0, high = 0;

    if (U_FAILURE(*status)) {
        return UCOL_PROCESSED_NULLORDER;
    }

    // pceBuffer.reset();

    while (pceBuffer.isEmpty()) {
        // buffer raw CEs up to non-ignorable primary
        RCEBuffer rceb;
        int32_t ce;

        // **** do we need to reset rceb, or will it always be empty at this point ****
        do {
            high = cei->getOffset();
            ce   = cei->previous(*status);
            low  = cei->getOffset();

            if (ce == UCOL_NULLORDER) {
                if (!rceb.isEmpty()) {
                    break;
                }

                goto finish;
            }

            rceb.put((uint32_t)ce, low, high, *status);
        } while (U_SUCCESS(*status) && ((ce & UCOL_PRIMARYORDERMASK) == 0 || isContinuation(ce)));

        // process the raw CEs
        while (U_SUCCESS(*status) && !rceb.isEmpty()) {
            const RCEI *rcei = rceb.get();

            result = processCE(rcei->ce);

            if (result != UCOL_IGNORABLE) {
                pceBuffer.put(result, rcei->low, rcei->high, *status);
            }
        }
        if (U_FAILURE(*status)) {
            return UCOL_PROCESSED_NULLORDER;
        }
    }

finish:
    if (pceBuffer.isEmpty()) {
        // **** Is -1 the right value for ixLow, ixHigh? ****
    	if (ixLow != NULL) {
    		*ixLow = -1;
    	}

    	if (ixHigh != NULL) {
    		*ixHigh = -1
    		;
    	}
        return UCOL_PROCESSED_NULLORDER;
    }

    const PCEI *pcei = pceBuffer.get();

    if (ixLow != NULL) {
        *ixLow = pcei->low;
    }

    if (ixHigh != NULL) {
        *ixHigh = pcei->high;
    }

    return pcei->ce;
}

U_NAMESPACE_END

U_CAPI int32_t U_EXPORT2
ucol_getMaxExpansion(const UCollationElements *elems,
                           int32_t            order)
{
    return CollationElementIterator::fromUCollationElements(elems)->getMaxExpansion(order);

    // TODO: The old code masked the order according to strength and then did a binary search.
    // However this was probably at least partially broken because of the following comment.
    // Still, it might have found a match when this version may not.

    // FIXME: with a masked search, there might be more than one hit,
    // so we need to look forward and backward from the match to find all
    // of the hits...
}

U_CAPI void U_EXPORT2
ucol_setText(      UCollationElements *elems,
             const UChar              *text,
                   int32_t            textLength,
                   UErrorCode         *status)
{
    if (U_FAILURE(*status)) {
        return;
    }

    if ((text == NULL && textLength != 0)) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    UnicodeString s((UBool)(textLength < 0), text, textLength);
    return CollationElementIterator::fromUCollationElements(elems)->setText(s, *status);
}

U_CAPI int32_t U_EXPORT2
ucol_getOffset(const UCollationElements *elems)
{
    return CollationElementIterator::fromUCollationElements(elems)->getOffset();
}

U_CAPI void U_EXPORT2
ucol_setOffset(UCollationElements    *elems,
               int32_t           offset,
               UErrorCode            *status)
{
    if (U_FAILURE(*status)) {
        return;
    }

    CollationElementIterator::fromUCollationElements(elems)->setOffset(offset, *status);
}

U_CAPI int32_t U_EXPORT2
ucol_primaryOrder (int32_t order)
{
    return (order >> 16) & 0xffff;
}

U_CAPI int32_t U_EXPORT2
ucol_secondaryOrder (int32_t order)
{
    return (order >> 8) & 0xff;
}

U_CAPI int32_t U_EXPORT2
ucol_tertiaryOrder (int32_t order)
{
    return order & 0xff;
}

#endif /* #if !UCONFIG_NO_COLLATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 1996-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING



















U_NAMESPACE_USE

/**
 * Verify that fmt is a SimpleDateFormat. Invalid error if not.
 * @param fmt the UDateFormat, definitely a DateFormat, maybe something else
 * @param status error code, will be set to failure if there is a familure or the fmt is NULL.
 */
static void verifyIsSimpleDateFormat(const UDateFormat* fmt, UErrorCode *status) {
   if(U_SUCCESS(*status) &&
       dynamic_cast<const SimpleDateFormat*>(reinterpret_cast<const DateFormat*>(fmt))==NULL) {
       *status = U_ILLEGAL_ARGUMENT_ERROR;
   }
}

// This mirrors the correspondence between the
// SimpleDateFormat::fgPatternIndexToDateFormatField and
// SimpleDateFormat::fgPatternIndexToCalendarField arrays.
static UCalendarDateFields gDateFieldMapping[] = {
    UCAL_ERA,                  // UDAT_ERA_FIELD = 0
    UCAL_YEAR,                 // UDAT_YEAR_FIELD = 1
    UCAL_MONTH,                // UDAT_MONTH_FIELD = 2
    UCAL_DATE,                 // UDAT_DATE_FIELD = 3
    UCAL_HOUR_OF_DAY,          // UDAT_HOUR_OF_DAY1_FIELD = 4
    UCAL_HOUR_OF_DAY,          // UDAT_HOUR_OF_DAY0_FIELD = 5
    UCAL_MINUTE,               // UDAT_MINUTE_FIELD = 6
    UCAL_SECOND,               // UDAT_SECOND_FIELD = 7
    UCAL_MILLISECOND,          // UDAT_FRACTIONAL_SECOND_FIELD = 8
    UCAL_DAY_OF_WEEK,          // UDAT_DAY_OF_WEEK_FIELD = 9
    UCAL_DAY_OF_YEAR,          // UDAT_DAY_OF_YEAR_FIELD = 10
    UCAL_DAY_OF_WEEK_IN_MONTH, // UDAT_DAY_OF_WEEK_IN_MONTH_FIELD = 11
    UCAL_WEEK_OF_YEAR,         // UDAT_WEEK_OF_YEAR_FIELD = 12
    UCAL_WEEK_OF_MONTH,        // UDAT_WEEK_OF_MONTH_FIELD = 13
    UCAL_AM_PM,                // UDAT_AM_PM_FIELD = 14
    UCAL_HOUR,                 // UDAT_HOUR1_FIELD = 15
    UCAL_HOUR,                 // UDAT_HOUR0_FIELD = 16
    UCAL_ZONE_OFFSET,          // UDAT_TIMEZONE_FIELD = 17
    UCAL_YEAR_WOY,             // UDAT_YEAR_WOY_FIELD = 18
    UCAL_DOW_LOCAL,            // UDAT_DOW_LOCAL_FIELD = 19
    UCAL_EXTENDED_YEAR,        // UDAT_EXTENDED_YEAR_FIELD = 20
    UCAL_JULIAN_DAY,           // UDAT_JULIAN_DAY_FIELD = 21
    UCAL_MILLISECONDS_IN_DAY,  // UDAT_MILLISECONDS_IN_DAY_FIELD = 22
    UCAL_ZONE_OFFSET,          // UDAT_TIMEZONE_RFC_FIELD = 23 (also UCAL_DST_OFFSET)
    UCAL_ZONE_OFFSET,          // UDAT_TIMEZONE_GENERIC_FIELD = 24 (also UCAL_DST_OFFSET)
    UCAL_DOW_LOCAL,            // UDAT_STANDALONE_DAY_FIELD = 25
    UCAL_MONTH,                // UDAT_STANDALONE_MONTH_FIELD = 26
    UCAL_MONTH,                // UDAT_QUARTER_FIELD = 27
    UCAL_MONTH,                // UDAT_STANDALONE_QUARTER_FIELD = 28
    UCAL_ZONE_OFFSET,          // UDAT_TIMEZONE_SPECIAL_FIELD = 29 (also UCAL_DST_OFFSET)
    UCAL_YEAR,                 // UDAT_YEAR_NAME_FIELD = 30
    UCAL_ZONE_OFFSET,          // UDAT_TIMEZONE_LOCALIZED_GMT_OFFSET_FIELD = 31 (also UCAL_DST_OFFSET)
    UCAL_ZONE_OFFSET,          // UDAT_TIMEZONE_ISO_FIELD = 32 (also UCAL_DST_OFFSET)
    UCAL_ZONE_OFFSET,          // UDAT_TIMEZONE_ISO_LOCAL_FIELD = 33 (also UCAL_DST_OFFSET)
    UCAL_EXTENDED_YEAR,        // UDAT_RELATED_YEAR_FIELD = 34 (not an exact match)
    UCAL_FIELD_COUNT,          // UDAT_FIELD_COUNT = 35
    // UCAL_IS_LEAP_MONTH is not the target of a mapping
};

U_CAPI UCalendarDateFields U_EXPORT2
udat_toCalendarDateField(UDateFormatField field) {
  return gDateFieldMapping[field];
}

/* For now- one opener. */
static UDateFormatOpener gOpener = NULL;

U_INTERNAL void U_EXPORT2
udat_registerOpener(UDateFormatOpener opener, UErrorCode *status)
{
  if(U_FAILURE(*status)) return;
  umtx_lock(NULL);
  if(gOpener==NULL) {
    gOpener = opener;
  } else {
    *status = U_ILLEGAL_ARGUMENT_ERROR;
  }
  umtx_unlock(NULL);
}

U_INTERNAL UDateFormatOpener U_EXPORT2
udat_unregisterOpener(UDateFormatOpener opener, UErrorCode *status)
{
  if(U_FAILURE(*status)) return NULL;
  UDateFormatOpener oldOpener = NULL;
  umtx_lock(NULL);
  if(gOpener==NULL || gOpener!=opener) {
    *status = U_ILLEGAL_ARGUMENT_ERROR;
  } else {
    oldOpener=gOpener;
    gOpener=NULL;
  }
  umtx_unlock(NULL);
  return oldOpener;
}



U_CAPI UDateFormat* U_EXPORT2
udat_open(UDateFormatStyle  timeStyle,
          UDateFormatStyle  dateStyle,
          const char        *locale,
          const UChar       *tzID,
          int32_t           tzIDLength,
          const UChar       *pattern,
          int32_t           patternLength,
          UErrorCode        *status)
{
    DateFormat *fmt;
    if(U_FAILURE(*status)) {
        return 0;
    }
    if(gOpener!=NULL) { // if it's registered
      fmt = (DateFormat*) (*gOpener)(timeStyle,dateStyle,locale,tzID,tzIDLength,pattern,patternLength,status);
      if(fmt!=NULL) {
        return (UDateFormat*)fmt;
      } // else fall through.
    }
    if(timeStyle != UDAT_PATTERN) {
        if(locale == 0) {
            fmt = DateFormat::createDateTimeInstance((DateFormat::EStyle)dateStyle,
                (DateFormat::EStyle)timeStyle);
        }
        else {
            fmt = DateFormat::createDateTimeInstance((DateFormat::EStyle)dateStyle,
                (DateFormat::EStyle)timeStyle,
                Locale(locale));
        }
    }
    else {
        UnicodeString pat((UBool)(patternLength == -1), pattern, patternLength);

        if(locale == 0) {
            fmt = new SimpleDateFormat(pat, *status);
        }
        else {
            fmt = new SimpleDateFormat(pat, Locale(locale), *status);
        }
    }

    if(fmt == nullptr) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return nullptr;
    }
    if (U_FAILURE(*status)) {
        delete fmt;
        return nullptr;
    }

    if(tzID != 0) {
        TimeZone *zone = TimeZone::createTimeZone(UnicodeString((UBool)(tzIDLength == -1), tzID, tzIDLength));
        if(zone == 0) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            delete fmt;
            return 0;
        }
        fmt->adoptTimeZone(zone);
    }

    return (UDateFormat*)fmt;
}


U_CAPI void U_EXPORT2
udat_close(UDateFormat* format)
{
    delete (DateFormat*)format;
}

U_CAPI UDateFormat* U_EXPORT2
udat_clone(const UDateFormat *fmt,
       UErrorCode *status)
{
    if(U_FAILURE(*status)) return 0;

    Format *res = ((DateFormat*)fmt)->clone();

    if(res == 0) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }

    return (UDateFormat*) res;
}

U_CAPI int32_t U_EXPORT2
udat_format(    const    UDateFormat*    format,
        UDate           dateToFormat,
        UChar*          result,
        int32_t         resultLength,
        UFieldPosition* position,
        UErrorCode*     status)
{
    if(U_FAILURE(*status)) {
        return -1;
    }
    if (result == NULL ? resultLength != 0 : resultLength < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    UnicodeString res;
    if (result != NULL) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        res.setTo(result, 0, resultLength);
    }

    FieldPosition fp;

    if(position != 0)
        fp.setField(position->field);

    ((DateFormat*)format)->format(dateToFormat, res, fp);

    if(position != 0) {
        position->beginIndex = fp.getBeginIndex();
        position->endIndex = fp.getEndIndex();
    }

    return res.extract(result, resultLength, *status);
}

U_CAPI int32_t U_EXPORT2
udat_formatCalendar(const UDateFormat*  format,
        UCalendar*      calendar,
        UChar*          result,
        int32_t         resultLength,
        UFieldPosition* position,
        UErrorCode*     status)
{
    if(U_FAILURE(*status)) {
        return -1;
    }
    if (result == NULL ? resultLength != 0 : resultLength < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    UnicodeString res;
    if (result != NULL) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        res.setTo(result, 0, resultLength);
    }

    FieldPosition fp;

    if(position != 0)
        fp.setField(position->field);

    ((DateFormat*)format)->format(*(Calendar*)calendar, res, fp);

    if(position != 0) {
        position->beginIndex = fp.getBeginIndex();
        position->endIndex = fp.getEndIndex();
    }

    return res.extract(result, resultLength, *status);
}

U_CAPI int32_t U_EXPORT2
udat_formatForFields(    const    UDateFormat*    format,
        UDate           dateToFormat,
        UChar*          result,
        int32_t         resultLength,
        UFieldPositionIterator* fpositer,
        UErrorCode*     status)
{
    if(U_FAILURE(*status)) {
        return -1;
    }
    if (result == NULL ? resultLength != 0 : resultLength < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    UnicodeString res;
    if (result != NULL) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        res.setTo(result, 0, resultLength);
    }

    ((DateFormat*)format)->format(dateToFormat, res, (FieldPositionIterator*)fpositer, *status);

    return res.extract(result, resultLength, *status);
}

U_CAPI int32_t U_EXPORT2
udat_formatCalendarForFields(const UDateFormat*  format,
        UCalendar*      calendar,
        UChar*          result,
        int32_t         resultLength,
        UFieldPositionIterator* fpositer,
        UErrorCode*     status)
{
    if(U_FAILURE(*status)) {
        return -1;
    }
    if (result == NULL ? resultLength != 0 : resultLength < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    UnicodeString res;
    if (result != NULL) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        res.setTo(result, 0, resultLength);
    }

    ((DateFormat*)format)->format(*(Calendar*)calendar, res, (FieldPositionIterator*)fpositer, *status);

    return res.extract(result, resultLength, *status);
}

U_CAPI UDate U_EXPORT2
udat_parse(    const    UDateFormat*        format,
        const    UChar*          text,
        int32_t         textLength,
        int32_t         *parsePos,
        UErrorCode      *status)
{
    if(U_FAILURE(*status)) return (UDate)0;

    const UnicodeString src((UBool)(textLength == -1), text, textLength);
    ParsePosition pp;
    int32_t stackParsePos = 0;
    UDate res;

    if(parsePos == NULL) {
        parsePos = &stackParsePos;
    }

    pp.setIndex(*parsePos);

    res = ((DateFormat*)format)->parse(src, pp);

    if(pp.getErrorIndex() == -1)
        *parsePos = pp.getIndex();
    else {
        *parsePos = pp.getErrorIndex();
        *status = U_PARSE_ERROR;
    }

    return res;
}

U_CAPI void U_EXPORT2
udat_parseCalendar(const    UDateFormat*    format,
                            UCalendar*      calendar,
                   const    UChar*          text,
                            int32_t         textLength,
                            int32_t         *parsePos,
                            UErrorCode      *status)
{
    if(U_FAILURE(*status)) return;

    const UnicodeString src((UBool)(textLength == -1), text, textLength);
    ParsePosition pp;
    int32_t stackParsePos = 0;

    if(parsePos == NULL) {
        parsePos = &stackParsePos;
    }

    pp.setIndex(*parsePos);

    ((DateFormat*)format)->parse(src, *(Calendar*)calendar, pp);

    if(pp.getErrorIndex() == -1)
        *parsePos = pp.getIndex();
    else {
        *parsePos = pp.getErrorIndex();
        *status = U_PARSE_ERROR;
    }
}

U_CAPI UBool U_EXPORT2
udat_isLenient(const UDateFormat* fmt)
{
    return ((DateFormat*)fmt)->isLenient();
}

U_CAPI void U_EXPORT2
udat_setLenient(    UDateFormat*    fmt,
            UBool          isLenient)
{
    ((DateFormat*)fmt)->setLenient(isLenient);
}

U_DRAFT UBool U_EXPORT2
udat_getBooleanAttribute(const UDateFormat* fmt, 
                         UDateFormatBooleanAttribute attr, 
                         UErrorCode* status)
{
    if(U_FAILURE(*status)) return FALSE;
    return ((DateFormat*)fmt)->getBooleanAttribute(attr, *status);
    //return FALSE;
}

U_DRAFT void U_EXPORT2
udat_setBooleanAttribute(UDateFormat *fmt, 
                         UDateFormatBooleanAttribute attr, 
                         UBool newValue, 
                         UErrorCode* status)
{
    if(U_FAILURE(*status)) return;
    ((DateFormat*)fmt)->setBooleanAttribute(attr, newValue, *status);
}

U_CAPI const UCalendar* U_EXPORT2
udat_getCalendar(const UDateFormat* fmt)
{
    return (const UCalendar*) ((DateFormat*)fmt)->getCalendar();
}

U_CAPI void U_EXPORT2
udat_setCalendar(UDateFormat*    fmt,
                 const   UCalendar*      calendarToSet)
{
    ((DateFormat*)fmt)->setCalendar(*((Calendar*)calendarToSet));
}

U_DRAFT const UNumberFormat* U_EXPORT2 
udat_getNumberFormatForField(const UDateFormat* fmt, UChar field)
{
    UErrorCode status = U_ZERO_ERROR;
    verifyIsSimpleDateFormat(fmt, &status);
    if (U_FAILURE(status)) return (const UNumberFormat*) ((DateFormat*)fmt)->getNumberFormat();
    return (const UNumberFormat*) ((SimpleDateFormat*)fmt)->getNumberFormatForField(field);
}

U_CAPI const UNumberFormat* U_EXPORT2
udat_getNumberFormat(const UDateFormat* fmt)
{
    return (const UNumberFormat*) ((DateFormat*)fmt)->getNumberFormat();
}

U_DRAFT void U_EXPORT2 
udat_adoptNumberFormatForFields(           UDateFormat*    fmt,
                                    const  UChar*          fields,
                                           UNumberFormat*  numberFormatToSet,
                                           UErrorCode*     status)
{
    verifyIsSimpleDateFormat(fmt, status);
    if (U_FAILURE(*status)) return;
    
    if (fields!=NULL) {
        UnicodeString overrideFields(fields);
        ((SimpleDateFormat*)fmt)->adoptNumberFormat(overrideFields, (NumberFormat*)numberFormatToSet, *status);
    }
}

U_CAPI void U_EXPORT2
udat_setNumberFormat(UDateFormat*    fmt,
                     const   UNumberFormat*  numberFormatToSet)
{
    ((DateFormat*)fmt)->setNumberFormat(*((NumberFormat*)numberFormatToSet));
}

U_DRAFT void U_EXPORT2
udat_adoptNumberFormat(      UDateFormat*    fmt,
                             UNumberFormat*  numberFormatToAdopt)
{
    ((DateFormat*)fmt)->adoptNumberFormat((NumberFormat*)numberFormatToAdopt);
}

U_CAPI const char* U_EXPORT2
udat_getAvailable(int32_t index)
{
    return uloc_getAvailable(index);
}

U_CAPI int32_t U_EXPORT2
udat_countAvailable()
{
    return uloc_countAvailable();
}

U_CAPI UDate U_EXPORT2
udat_get2DigitYearStart(    const   UDateFormat     *fmt,
                        UErrorCode      *status)
{
    verifyIsSimpleDateFormat(fmt, status);
    if(U_FAILURE(*status)) return (UDate)0;
    return ((SimpleDateFormat*)fmt)->get2DigitYearStart(*status);
}

U_CAPI void U_EXPORT2
udat_set2DigitYearStart(    UDateFormat     *fmt,
                        UDate           d,
                        UErrorCode      *status)
{
    verifyIsSimpleDateFormat(fmt, status);
    if(U_FAILURE(*status)) return;
    ((SimpleDateFormat*)fmt)->set2DigitYearStart(d, *status);
}

U_CAPI int32_t U_EXPORT2
udat_toPattern(    const   UDateFormat     *fmt,
        UBool          localized,
        UChar           *result,
        int32_t         resultLength,
        UErrorCode      *status)
{
    if(U_FAILURE(*status)) {
        return -1;
    }
    if (result == NULL ? resultLength != 0 : resultLength < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    UnicodeString res;
    if (result != NULL) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        res.setTo(result, 0, resultLength);
    }

    const DateFormat *df=reinterpret_cast<const DateFormat *>(fmt);
    const SimpleDateFormat *sdtfmt=dynamic_cast<const SimpleDateFormat *>(df);
    const RelativeDateFormat *reldtfmt;
    if (sdtfmt!=NULL) {
        if(localized)
            sdtfmt->toLocalizedPattern(res, *status);
        else
            sdtfmt->toPattern(res);
    } else if (!localized && (reldtfmt=dynamic_cast<const RelativeDateFormat *>(df))!=NULL) {
        reldtfmt->toPattern(res, *status);
    } else {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    return res.extract(result, resultLength, *status);
}

// TODO: should this take an UErrorCode?
// A: Yes. Of course.
U_CAPI void U_EXPORT2
udat_applyPattern(  UDateFormat     *format,
                    UBool          localized,
                    const   UChar           *pattern,
                    int32_t         patternLength)
{
    const UnicodeString pat((UBool)(patternLength == -1), pattern, patternLength);
    UErrorCode status = U_ZERO_ERROR;

    verifyIsSimpleDateFormat(format, &status);
    if(U_FAILURE(status)) {
        return;
    }
    
    if(localized)
        ((SimpleDateFormat*)format)->applyLocalizedPattern(pat, status);
    else
        ((SimpleDateFormat*)format)->applyPattern(pat);
}

U_CAPI int32_t U_EXPORT2
udat_getSymbols(const   UDateFormat     *fmt,
                UDateFormatSymbolType   type,
                int32_t                 index,
                UChar                   *result,
                int32_t                 resultLength,
                UErrorCode              *status)
{
    const DateFormatSymbols *syms;
    const SimpleDateFormat* sdtfmt;
    const RelativeDateFormat* rdtfmt;
    if ((sdtfmt = dynamic_cast<const SimpleDateFormat*>(reinterpret_cast<const DateFormat*>(fmt))) != NULL) {
        syms = sdtfmt->getDateFormatSymbols();
    } else if ((rdtfmt = dynamic_cast<const RelativeDateFormat*>(reinterpret_cast<const DateFormat*>(fmt))) != NULL) {
        syms = rdtfmt->getDateFormatSymbols();
    } else {
        return -1;
    }
    int32_t count = 0;
    const UnicodeString *res = NULL;

    switch(type) {
    case UDAT_ERAS:
        res = syms->getEras(count);
        break;

    case UDAT_ERA_NAMES:
        res = syms->getEraNames(count);
        break;

    case UDAT_MONTHS:
        res = syms->getMonths(count);
        break;

    case UDAT_SHORT_MONTHS:
        res = syms->getShortMonths(count);
        break;

    case UDAT_WEEKDAYS:
        res = syms->getWeekdays(count);
        break;

    case UDAT_SHORT_WEEKDAYS:
        res = syms->getShortWeekdays(count);
        break;

    case UDAT_AM_PMS:
        res = syms->getAmPmStrings(count);
        break;

    case UDAT_LOCALIZED_CHARS:
        {
            UnicodeString res1;
            if(!(result==NULL && resultLength==0)) {
                // NULL destination for pure preflighting: empty dummy string
                // otherwise, alias the destination buffer
                res1.setTo(result, 0, resultLength);
            }
            syms->getLocalPatternChars(res1);
            return res1.extract(result, resultLength, *status);
        }

    case UDAT_NARROW_MONTHS:
        res = syms->getMonths(count, DateFormatSymbols::FORMAT, DateFormatSymbols::NARROW);
        break;

    case UDAT_SHORTER_WEEKDAYS:
        res = syms->getWeekdays(count, DateFormatSymbols::FORMAT, DateFormatSymbols::SHORT);
        break;

    case UDAT_NARROW_WEEKDAYS:
        res = syms->getWeekdays(count, DateFormatSymbols::FORMAT, DateFormatSymbols::NARROW);
        break;

    case UDAT_STANDALONE_MONTHS:
        res = syms->getMonths(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::WIDE);
        break;

    case UDAT_STANDALONE_SHORT_MONTHS:
        res = syms->getMonths(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_STANDALONE_NARROW_MONTHS:
        res = syms->getMonths(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::NARROW);
        break;

    case UDAT_STANDALONE_WEEKDAYS:
        res = syms->getWeekdays(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::WIDE);
        break;

    case UDAT_STANDALONE_SHORT_WEEKDAYS:
        res = syms->getWeekdays(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_STANDALONE_SHORTER_WEEKDAYS:
        res = syms->getWeekdays(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::SHORT);
        break;

    case UDAT_STANDALONE_NARROW_WEEKDAYS:
        res = syms->getWeekdays(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::NARROW);
        break;

    case UDAT_QUARTERS:
        res = syms->getQuarters(count, DateFormatSymbols::FORMAT, DateFormatSymbols::WIDE);
        break;

    case UDAT_SHORT_QUARTERS:
        res = syms->getQuarters(count, DateFormatSymbols::FORMAT, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_STANDALONE_QUARTERS:
        res = syms->getQuarters(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::WIDE);
        break;

    case UDAT_STANDALONE_SHORT_QUARTERS:
        res = syms->getQuarters(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_CYCLIC_YEARS_WIDE:
        res = syms->getYearNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::WIDE);
        break;

    case UDAT_CYCLIC_YEARS_ABBREVIATED:
        res = syms->getYearNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_CYCLIC_YEARS_NARROW:
        res = syms->getYearNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::NARROW);
        break;

    case UDAT_ZODIAC_NAMES_WIDE:
        res = syms->getZodiacNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::WIDE);
        break;

    case UDAT_ZODIAC_NAMES_ABBREVIATED:
        res = syms->getZodiacNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_ZODIAC_NAMES_NARROW:
        res = syms->getZodiacNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::NARROW);
        break;

    }

    if(index < count) {
        return res[index].extract(result, resultLength, *status);
    }
    return 0;
}

// TODO: also needs an errorCode.
U_CAPI int32_t U_EXPORT2
udat_countSymbols(    const    UDateFormat                *fmt,
            UDateFormatSymbolType    type)
{
    const DateFormatSymbols *syms;
    const SimpleDateFormat* sdtfmt;
    const RelativeDateFormat* rdtfmt;
    if ((sdtfmt = dynamic_cast<const SimpleDateFormat*>(reinterpret_cast<const DateFormat*>(fmt))) != NULL) {
        syms = sdtfmt->getDateFormatSymbols();
    } else if ((rdtfmt = dynamic_cast<const RelativeDateFormat*>(reinterpret_cast<const DateFormat*>(fmt))) != NULL) {
        syms = rdtfmt->getDateFormatSymbols();
    } else {
        return 0;
    }
    int32_t count = 0;

    switch(type) {
    case UDAT_ERAS:
        syms->getEras(count);
        break;

    case UDAT_MONTHS:
        syms->getMonths(count);
        break;

    case UDAT_SHORT_MONTHS:
        syms->getShortMonths(count);
        break;

    case UDAT_WEEKDAYS:
        syms->getWeekdays(count);
        break;

    case UDAT_SHORT_WEEKDAYS:
        syms->getShortWeekdays(count);
        break;

    case UDAT_AM_PMS:
        syms->getAmPmStrings(count);
        break;

    case UDAT_LOCALIZED_CHARS:
        count = 1;
        break;

    case UDAT_ERA_NAMES:
        syms->getEraNames(count);
        break;

    case UDAT_NARROW_MONTHS:
        syms->getMonths(count, DateFormatSymbols::FORMAT, DateFormatSymbols::NARROW);
        break;

    case UDAT_SHORTER_WEEKDAYS:
        syms->getWeekdays(count, DateFormatSymbols::FORMAT, DateFormatSymbols::SHORT);
        break;

    case UDAT_NARROW_WEEKDAYS:
        syms->getWeekdays(count, DateFormatSymbols::FORMAT, DateFormatSymbols::NARROW);
        break;

    case UDAT_STANDALONE_MONTHS:
        syms->getMonths(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::WIDE);
        break;

    case UDAT_STANDALONE_SHORT_MONTHS:
        syms->getMonths(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_STANDALONE_NARROW_MONTHS:
        syms->getMonths(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::NARROW);
        break;

    case UDAT_STANDALONE_WEEKDAYS:
        syms->getWeekdays(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::WIDE);
        break;

    case UDAT_STANDALONE_SHORT_WEEKDAYS:
        syms->getWeekdays(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_STANDALONE_SHORTER_WEEKDAYS:
        syms->getWeekdays(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::SHORT);
        break;

    case UDAT_STANDALONE_NARROW_WEEKDAYS:
        syms->getWeekdays(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::NARROW);
        break;

    case UDAT_QUARTERS:
        syms->getQuarters(count, DateFormatSymbols::FORMAT, DateFormatSymbols::WIDE);
        break;

    case UDAT_SHORT_QUARTERS:
        syms->getQuarters(count, DateFormatSymbols::FORMAT, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_STANDALONE_QUARTERS:
        syms->getQuarters(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::WIDE);
        break;

    case UDAT_STANDALONE_SHORT_QUARTERS:
        syms->getQuarters(count, DateFormatSymbols::STANDALONE, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_CYCLIC_YEARS_WIDE:
        syms->getYearNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::WIDE);
        break;

    case UDAT_CYCLIC_YEARS_ABBREVIATED:
        syms->getYearNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_CYCLIC_YEARS_NARROW:
        syms->getYearNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::NARROW);
        break;

    case UDAT_ZODIAC_NAMES_WIDE:
        syms->getZodiacNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::WIDE);
        break;

    case UDAT_ZODIAC_NAMES_ABBREVIATED:
        syms->getZodiacNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::ABBREVIATED);
        break;

    case UDAT_ZODIAC_NAMES_NARROW:
        syms->getZodiacNames(count, DateFormatSymbols::FORMAT, DateFormatSymbols::NARROW);
        break;

    }

    return count;
}

U_NAMESPACE_BEGIN

/*
 * This DateFormatSymbolsSingleSetter class is a friend of DateFormatSymbols
 * solely for the purpose of avoiding to clone the array of strings
 * just to modify one of them and then setting all of them back.
 * For example, the old code looked like this:
 *  case UDAT_MONTHS:
 *    res = syms->getMonths(count);
 *    array = new UnicodeString[count];
 *    if(array == 0) {
 *      *status = U_MEMORY_ALLOCATION_ERROR;
 *      return;
 *    }
 *    uprv_arrayCopy(res, array, count);
 *    if(index < count)
 *      array[index] = val;
 *    syms->setMonths(array, count);
 *    break;
 *
 * Even worse, the old code actually cloned the entire DateFormatSymbols object,
 * cloned one value array, changed one value, and then made the SimpleDateFormat
 * replace its DateFormatSymbols object with the new one.
 *
 * markus 2002-oct-14
 */
class DateFormatSymbolsSingleSetter /* not : public UObject because all methods are static */ {
public:
    static void
        setSymbol(UnicodeString *array, int32_t count, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        if(array!=NULL) {
            if(index>=count) {
                errorCode=U_INDEX_OUTOFBOUNDS_ERROR;
            } else if(value==NULL) {
                errorCode=U_ILLEGAL_ARGUMENT_ERROR;
            } else {
                array[index].setTo(value, valueLength);
            }
        }
    }

    static void
        setEra(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fEras, syms->fErasCount, index, value, valueLength, errorCode);
    }

    static void
        setEraName(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fEraNames, syms->fEraNamesCount, index, value, valueLength, errorCode);
    }

    static void
        setMonth(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fMonths, syms->fMonthsCount, index, value, valueLength, errorCode);
    }

    static void
        setShortMonth(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fShortMonths, syms->fShortMonthsCount, index, value, valueLength, errorCode);
    }

    static void
        setNarrowMonth(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fNarrowMonths, syms->fNarrowMonthsCount, index, value, valueLength, errorCode);
    }

    static void
        setStandaloneMonth(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fStandaloneMonths, syms->fStandaloneMonthsCount, index, value, valueLength, errorCode);
    }

    static void
        setStandaloneShortMonth(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fStandaloneShortMonths, syms->fStandaloneShortMonthsCount, index, value, valueLength, errorCode);
    }

    static void
        setStandaloneNarrowMonth(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fStandaloneNarrowMonths, syms->fStandaloneNarrowMonthsCount, index, value, valueLength, errorCode);
    }

    static void
        setWeekday(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fWeekdays, syms->fWeekdaysCount, index, value, valueLength, errorCode);
    }

    static void
        setShortWeekday(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fShortWeekdays, syms->fShortWeekdaysCount, index, value, valueLength, errorCode);
    }

    static void
        setShorterWeekday(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fShorterWeekdays, syms->fShorterWeekdaysCount, index, value, valueLength, errorCode);
    }

    static void
        setNarrowWeekday(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fNarrowWeekdays, syms->fNarrowWeekdaysCount, index, value, valueLength, errorCode);
    }

    static void
        setStandaloneWeekday(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fStandaloneWeekdays, syms->fStandaloneWeekdaysCount, index, value, valueLength, errorCode);
    }

    static void
        setStandaloneShortWeekday(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fStandaloneShortWeekdays, syms->fStandaloneShortWeekdaysCount, index, value, valueLength, errorCode);
    }

    static void
        setStandaloneShorterWeekday(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fStandaloneShorterWeekdays, syms->fStandaloneShorterWeekdaysCount, index, value, valueLength, errorCode);
    }

    static void
        setStandaloneNarrowWeekday(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fStandaloneNarrowWeekdays, syms->fStandaloneNarrowWeekdaysCount, index, value, valueLength, errorCode);
    }

    static void
        setQuarter(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fQuarters, syms->fQuartersCount, index, value, valueLength, errorCode);
    }

    static void
        setShortQuarter(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fShortQuarters, syms->fShortQuartersCount, index, value, valueLength, errorCode);
    }

    static void
        setStandaloneQuarter(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fStandaloneQuarters, syms->fStandaloneQuartersCount, index, value, valueLength, errorCode);
    }

    static void
        setStandaloneShortQuarter(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fStandaloneShortQuarters, syms->fStandaloneShortQuartersCount, index, value, valueLength, errorCode);
    }

    static void
        setShortYearNames(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fShortYearNames, syms->fShortYearNamesCount, index, value, valueLength, errorCode);
    }

    static void
        setShortZodiacNames(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fShortZodiacNames, syms->fShortZodiacNamesCount, index, value, valueLength, errorCode);
    }

    static void
        setAmPm(DateFormatSymbols *syms, int32_t index,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(syms->fAmPms, syms->fAmPmsCount, index, value, valueLength, errorCode);
    }

    static void
        setLocalPatternChars(DateFormatSymbols *syms,
        const UChar *value, int32_t valueLength, UErrorCode &errorCode)
    {
        setSymbol(&syms->fLocalPatternChars, 1, 0, value, valueLength, errorCode);
    }
};

U_NAMESPACE_END

U_CAPI void U_EXPORT2
udat_setSymbols(    UDateFormat             *format,
            UDateFormatSymbolType   type,
            int32_t                 index,
            UChar                   *value,
            int32_t                 valueLength,
            UErrorCode              *status)
{
    verifyIsSimpleDateFormat(format, status);
    if(U_FAILURE(*status)) return;

    DateFormatSymbols *syms = (DateFormatSymbols *)((SimpleDateFormat *)format)->getDateFormatSymbols();

    switch(type) {
    case UDAT_ERAS:
        DateFormatSymbolsSingleSetter::setEra(syms, index, value, valueLength, *status);
        break;

    case UDAT_ERA_NAMES:
        DateFormatSymbolsSingleSetter::setEraName(syms, index, value, valueLength, *status);
        break;

    case UDAT_MONTHS:
        DateFormatSymbolsSingleSetter::setMonth(syms, index, value, valueLength, *status);
        break;

    case UDAT_SHORT_MONTHS:
        DateFormatSymbolsSingleSetter::setShortMonth(syms, index, value, valueLength, *status);
        break;

    case UDAT_NARROW_MONTHS:
        DateFormatSymbolsSingleSetter::setNarrowMonth(syms, index, value, valueLength, *status);
        break;

    case UDAT_STANDALONE_MONTHS:
        DateFormatSymbolsSingleSetter::setStandaloneMonth(syms, index, value, valueLength, *status);
        break;

    case UDAT_STANDALONE_SHORT_MONTHS:
        DateFormatSymbolsSingleSetter::setStandaloneShortMonth(syms, index, value, valueLength, *status);
        break;

    case UDAT_STANDALONE_NARROW_MONTHS:
        DateFormatSymbolsSingleSetter::setStandaloneNarrowMonth(syms, index, value, valueLength, *status);
        break;

    case UDAT_WEEKDAYS:
        DateFormatSymbolsSingleSetter::setWeekday(syms, index, value, valueLength, *status);
        break;

    case UDAT_SHORT_WEEKDAYS:
        DateFormatSymbolsSingleSetter::setShortWeekday(syms, index, value, valueLength, *status);
        break;

    case UDAT_SHORTER_WEEKDAYS:
        DateFormatSymbolsSingleSetter::setShorterWeekday(syms, index, value, valueLength, *status);
        break;

    case UDAT_NARROW_WEEKDAYS:
        DateFormatSymbolsSingleSetter::setNarrowWeekday(syms, index, value, valueLength, *status);
        break;

    case UDAT_STANDALONE_WEEKDAYS:
        DateFormatSymbolsSingleSetter::setStandaloneWeekday(syms, index, value, valueLength, *status);
        break;

    case UDAT_STANDALONE_SHORT_WEEKDAYS:
        DateFormatSymbolsSingleSetter::setStandaloneShortWeekday(syms, index, value, valueLength, *status);
        break;

    case UDAT_STANDALONE_SHORTER_WEEKDAYS:
        DateFormatSymbolsSingleSetter::setStandaloneShorterWeekday(syms, index, value, valueLength, *status);
        break;

    case UDAT_STANDALONE_NARROW_WEEKDAYS:
        DateFormatSymbolsSingleSetter::setStandaloneNarrowWeekday(syms, index, value, valueLength, *status);
        break;

    case UDAT_QUARTERS:
        DateFormatSymbolsSingleSetter::setQuarter(syms, index, value, valueLength, *status);
        break;

    case UDAT_SHORT_QUARTERS:
        DateFormatSymbolsSingleSetter::setShortQuarter(syms, index, value, valueLength, *status);
        break;

    case UDAT_STANDALONE_QUARTERS:
        DateFormatSymbolsSingleSetter::setStandaloneQuarter(syms, index, value, valueLength, *status);
        break;

    case UDAT_STANDALONE_SHORT_QUARTERS:
        DateFormatSymbolsSingleSetter::setStandaloneShortQuarter(syms, index, value, valueLength, *status);
        break;

    case UDAT_CYCLIC_YEARS_ABBREVIATED:
        DateFormatSymbolsSingleSetter::setShortYearNames(syms, index, value, valueLength, *status);
        break;

    case UDAT_ZODIAC_NAMES_ABBREVIATED:
        DateFormatSymbolsSingleSetter::setShortZodiacNames(syms, index, value, valueLength, *status);
        break;

    case UDAT_AM_PMS:
        DateFormatSymbolsSingleSetter::setAmPm(syms, index, value, valueLength, *status);
        break;

    case UDAT_LOCALIZED_CHARS:
        DateFormatSymbolsSingleSetter::setLocalPatternChars(syms, value, valueLength, *status);
        break;

    default:
        *status = U_UNSUPPORTED_ERROR;
        break;
        
    }
}

U_CAPI const char* U_EXPORT2
udat_getLocaleByType(const UDateFormat *fmt,
                     ULocDataLocaleType type,
                     UErrorCode* status)
{
    if (fmt == NULL) {
        if (U_SUCCESS(*status)) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
        }
        return NULL;
    }
    return ((Format*)fmt)->getLocaleID(type, *status);
}

U_CAPI void U_EXPORT2
udat_setContext(UDateFormat* fmt, UDisplayContext value, UErrorCode* status)
{
    if (U_FAILURE(*status)) {
        return;
    }
    ((DateFormat*)fmt)->setContext(value, *status);
    return;
}

U_CAPI UDisplayContext U_EXPORT2
udat_getContext(const UDateFormat* fmt, UDisplayContextType type, UErrorCode* status)
{
    if (U_FAILURE(*status)) {
        return (UDisplayContext)0;
    }
    return ((const DateFormat*)fmt)->getContext(type, *status);
}


/**
 * Verify that fmt is a RelativeDateFormat. Invalid error if not.
 * @param fmt the UDateFormat, definitely a DateFormat, maybe something else
 * @param status error code, will be set to failure if there is a familure or the fmt is NULL.
 */
static void verifyIsRelativeDateFormat(const UDateFormat* fmt, UErrorCode *status) {
   if(U_SUCCESS(*status) &&
       dynamic_cast<const RelativeDateFormat*>(reinterpret_cast<const DateFormat*>(fmt))==NULL) {
       *status = U_ILLEGAL_ARGUMENT_ERROR;
   }
}


U_CAPI int32_t U_EXPORT2 
udat_toPatternRelativeDate(const UDateFormat *fmt,
                           UChar             *result,
                           int32_t           resultLength,
                           UErrorCode        *status)
{
    verifyIsRelativeDateFormat(fmt, status);
    if(U_FAILURE(*status)) {
        return -1;
    }
    if (result == NULL ? resultLength != 0 : resultLength < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    UnicodeString datePattern;
    if (result != NULL) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        datePattern.setTo(result, 0, resultLength);
    }
    ((RelativeDateFormat*)fmt)->toPatternDate(datePattern, *status);
    return datePattern.extract(result, resultLength, *status);
}

U_CAPI int32_t U_EXPORT2 
udat_toPatternRelativeTime(const UDateFormat *fmt,
                           UChar             *result,
                           int32_t           resultLength,
                           UErrorCode        *status)
{
    verifyIsRelativeDateFormat(fmt, status);
    if(U_FAILURE(*status)) {
        return -1;
    }
    if (result == NULL ? resultLength != 0 : resultLength < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    UnicodeString timePattern;
    if (result != NULL) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        timePattern.setTo(result, 0, resultLength);
    }
    ((RelativeDateFormat*)fmt)->toPatternTime(timePattern, *status);
    return timePattern.extract(result, resultLength, *status);
}

U_CAPI void U_EXPORT2 
udat_applyPatternRelative(UDateFormat *format,
                          const UChar *datePattern,
                          int32_t     datePatternLength,
                          const UChar *timePattern,
                          int32_t     timePatternLength,
                          UErrorCode  *status)
{
    verifyIsRelativeDateFormat(format, status);
    if(U_FAILURE(*status)) return;
    const UnicodeString datePat((UBool)(datePatternLength == -1), datePattern, datePatternLength);
    const UnicodeString timePat((UBool)(timePatternLength == -1), timePattern, timePatternLength);
    ((RelativeDateFormat*)format)->applyPatterns(datePat, timePat, *status);
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2010-2011, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/



#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2010-2012,2015 International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/




#if !UCONFIG_NO_FORMATTING





/**
 * \file
 * \brief C API: Format a date interval.
 *
 * A UDateIntervalFormat is used to format the range between two UDate values
 * in a locale-sensitive way, using a skeleton that specifies the precision and
 * completeness of the information to show. If the range smaller than the resolution
 * specified by the skeleton, a single date format will be produced. If the range
 * is larger than the format specified by the skeleton, a locale-specific fallback
 * will be used to format the items missing from the skeleton.
 *
 * For example, if the range is 2010-03-04 07:56 - 2010-03-04 19:56 (12 hours)
 * - The skeleton jm will produce
 *   for en_US, "7:56 AM - 7:56 PM"
 *   for en_GB, "7:56 - 19:56"
 * - The skeleton MMMd will produce
 *   for en_US, "Mar 4"
 *   for en_GB, "4 Mar"
 * If the range is 2010-03-04 07:56 - 2010-03-08 16:11 (4 days, 8 hours, 15 minutes)
 * - The skeleton jm will produce
 *   for en_US, "3/4/2010 7:56 AM - 3/8/2010 4:11 PM"
 *   for en_GB, "4/3/2010 7:56 - 8/3/2010 16:11"
 * - The skeleton MMMd will produce
 *   for en_US, "Mar 4-8"
 *   for en_GB, "4-8 Mar"
 * 
 * Note:  the "-" characters in the above sample output will actually be
 * Unicode 2013, EN_DASH, in all but the last example.
 *
 * Note, in ICU 4.4 the standard skeletons for which date interval format data
 * is usually available are as follows; best results will be obtained by using
 * skeletons from this set, or those formed by combining these standard skeletons
 * (note that for these skeletons, the length of digit field such as d, y, or
 * M vs MM is irrelevant (but for non-digit fields such as MMM vs MMMM it is
 * relevant). Note that a skeleton involving h or H generally explicitly requests
 * that time style (12- or 24-hour time respectively). For a skeleton that
 * requests the locale's default time style (h or H), use 'j' instead of h or H.
 *   h, H, hm, Hm,
 *   hv, Hv, hmv, Hmv,
 *   d,
 *   M, MMM, MMMM,
 *   Md, MMMd,
 *   MEd, MMMEd,
 *   y,
 *   yM, yMMM, yMMMM,
 *   yMd, yMMMd,
 *   yMEd, yMMMEd
 *
 * Locales for which ICU 4.4 seems to have a reasonable amount of this data
 * include:
 *   af, am, ar, be, bg, bn, ca, cs, da, de (_AT), el, en (_AU,_CA,_GB,_IE,_IN...),
 *   eo, es (_AR,_CL,_CO,...,_US) et, fa, fi, fo, fr (_BE,_CH,_CA), fur, gsw, he,
 *   hr, hu, hy, is, it (_CH), ja, kk, km, ko, lt, lv, mk, ml, mt, nb, nl )_BE),
 *   nn, pl, pt (_PT), rm, ro, ru (_UA), sk, sl, so, sq, sr, sr_Latn, sv, th, to,
 *   tr, uk, ur, vi, zh (_SG), zh_Hant (_HK,_MO)
 */

/**
 * Opaque UDateIntervalFormat object for use in C programs.
 * @stable ICU 4.8
 */
struct UDateIntervalFormat;
typedef struct UDateIntervalFormat UDateIntervalFormat;  /**< C typedef for struct UDateIntervalFormat. @stable ICU 4.8 */

#ifndef U_HIDE_DRAFT_API
struct UFormattedDateInterval;
/**
 * Opaque struct to contain the results of a UDateIntervalFormat operation.
 * @draft ICU 64
 */
typedef struct UFormattedDateInterval UFormattedDateInterval;
#endif /* U_HIDE_DRAFT_API */

/**
 * Open a new UDateIntervalFormat object using the predefined rules for a
 * given locale plus a specified skeleton.
 * @param locale
 *            The locale for whose rules should be used; may be NULL for
 *            default locale.
 * @param skeleton
 *            A pattern containing only the fields desired for the interval
 *            format, for example "Hm", "yMMMd", or "yMMMEdHm".
 * @param skeletonLength
 *            The length of skeleton; may be -1 if the skeleton is zero-terminated.
 * @param tzID
 *            A timezone ID specifying the timezone to use. If 0, use the default
 *            timezone.
 * @param tzIDLength
 *            The length of tzID, or -1 if null-terminated. If 0, use the default
 *            timezone.
 * @param status
 *            A pointer to a UErrorCode to receive any errors.
 * @return
 *            A pointer to a UDateIntervalFormat object for the specified locale,
 *            or NULL if an error occurred.
 * @stable ICU 4.8
 */
U_STABLE UDateIntervalFormat* U_EXPORT2
udtitvfmt_open(const char*  locale,
              const UChar* skeleton,
              int32_t      skeletonLength,
              const UChar* tzID,
              int32_t      tzIDLength,
              UErrorCode*  status);

/**
 * Close a UDateIntervalFormat object. Once closed it may no longer be used.
 * @param formatter
 *            The UDateIntervalFormat object to close.
 * @stable ICU 4.8
 */
U_STABLE void U_EXPORT2
udtitvfmt_close(UDateIntervalFormat *formatter);


#ifndef U_HIDE_DRAFT_API
/**
 * Creates an object to hold the result of a UDateIntervalFormat
 * operation. The object can be used repeatedly; it is cleared whenever
 * passed to a format function.
 *
 * @param ec Set if an error occurs.
 * @return A pointer needing ownership.
 * @draft ICU 64
 */
U_CAPI UFormattedDateInterval* U_EXPORT2
udtitvfmt_openResult(UErrorCode* ec);

/**
 * Returns a representation of a UFormattedDateInterval as a UFormattedValue,
 * which can be subsequently passed to any API requiring that type.
 *
 * The returned object is owned by the UFormattedDateInterval and is valid
 * only as long as the UFormattedDateInterval is present and unchanged in memory.
 *
 * You can think of this method as a cast between types.
 *
 * When calling ufmtval_nextPosition():
 * The fields are returned from left to right. The special field category
 * UFIELD_CATEGORY_DATE_INTERVAL_SPAN is used to indicate which datetime
 * primitives came from which arguments: 0 means fromCalendar, and 1 means
 * toCalendar. The span category will always occur before the
 * corresponding fields in UFIELD_CATEGORY_DATE
 * in the ufmtval_nextPosition() iterator.
 *
 * @param uresult The object containing the formatted string.
 * @param ec Set if an error occurs.
 * @return A UFormattedValue owned by the input object.
 * @draft ICU 64
 */
U_CAPI const UFormattedValue* U_EXPORT2
udtitvfmt_resultAsValue(const UFormattedDateInterval* uresult, UErrorCode* ec);

/**
 * Releases the UFormattedDateInterval created by udtitvfmt_openResult().
 *
 * @param uresult The object to release.
 * @draft ICU 64
 */
U_CAPI void U_EXPORT2
udtitvfmt_closeResult(UFormattedDateInterval* uresult);
#endif /* U_HIDE_DRAFT_API */


#if U_SHOW_CPLUSPLUS_API

U_NAMESPACE_BEGIN

/**
 * \class LocalUDateIntervalFormatPointer
 * "Smart pointer" class, closes a UDateIntervalFormat via udtitvfmt_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 4.8
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUDateIntervalFormatPointer, UDateIntervalFormat, udtitvfmt_close);

#ifndef U_HIDE_DRAFT_API
/**
 * \class LocalUFormattedDateIntervalPointer
 * "Smart pointer" class, closes a UFormattedDateInterval via udtitvfmt_close().
 * For most methods see the LocalPointerBase base class.
 *
 * @see LocalPointerBase
 * @see LocalPointer
 * @draft ICU 64
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUFormattedDateIntervalPointer, UFormattedDateInterval, udtitvfmt_closeResult);
#endif /* U_HIDE_DRAFT_API */

U_NAMESPACE_END

#endif


/**
 * Formats a date/time range using the conventions established for the
 * UDateIntervalFormat object.
 * @param formatter
 *            The UDateIntervalFormat object specifying the format conventions.
 * @param fromDate
 *            The starting point of the range.
 * @param toDate
 *            The ending point of the range.
 * @param result
 *            A pointer to a buffer to receive the formatted range.
 * @param resultCapacity
 *            The maximum size of result.
 * @param position
 *            A pointer to a UFieldPosition. On input, position->field is read.
 *            On output, position->beginIndex and position->endIndex indicate
 *            the beginning and ending indices of field number position->field,
 *            if such a field exists. This parameter may be NULL, in which case
 *            no field position data is returned.
 *            There may be multiple instances of a given field type in an
 *            interval format; in this case the position indices refer to the
 *            first instance.
 * @param status
 *            A pointer to a UErrorCode to receive any errors.
 * @return
 *            The total buffer size needed; if greater than resultLength, the
 *            output was truncated.
 * @stable ICU 4.8
 */
U_STABLE int32_t U_EXPORT2
udtitvfmt_format(const UDateIntervalFormat* formatter,
                UDate           fromDate,
                UDate           toDate,
                UChar*          result,
                int32_t         resultCapacity,
                UFieldPosition* position,
                UErrorCode*     status);


#ifndef U_HIDE_DRAFT_API
/**
 * Formats a date/time range using the conventions established for the
 * UDateIntervalFormat object.
 * @param formatter
 *            The UDateIntervalFormat object specifying the format conventions.
 * @param result
 *            The UFormattedDateInterval to contain the result of the
 *            formatting operation.
 * @param fromDate
 *            The starting point of the range.
 * @param toDate
 *            The ending point of the range.
 * @param status
 *            A pointer to a UErrorCode to receive any errors.
 * @draft ICU 64
 */
U_DRAFT void U_EXPORT2
udtitvfmt_formatToResult(
                const UDateIntervalFormat* formatter,
                UFormattedDateInterval* result,
                UDate           fromDate,
                UDate           toDate,
                UErrorCode*     status);
#endif /* U_HIDE_DRAFT_API */


#endif /* #if !UCONFIG_NO_FORMATTING */










U_NAMESPACE_USE


// Magic number: FDIV in ASCII
UPRV_FORMATTED_VALUE_CAPI_AUTO_IMPL(
    FormattedDateInterval,
    UFormattedDateInterval,
    UFormattedDateIntervalImpl,
    UFormattedDateIntervalApiHelper,
    udtitvfmt,
    0x46444956)


U_CAPI UDateIntervalFormat* U_EXPORT2
udtitvfmt_open(const char*  locale,
               const UChar* skeleton,
               int32_t      skeletonLength,
               const UChar* tzID,
               int32_t      tzIDLength,
               UErrorCode*  status)
{
    if (U_FAILURE(*status)) {
        return NULL;
    }
    if ((skeleton == NULL ? skeletonLength != 0 : skeletonLength < -1) ||
        (tzID == NULL ? tzIDLength != 0 : tzIDLength < -1)
    ) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    UnicodeString skel((UBool)(skeletonLength == -1), skeleton, skeletonLength);
    LocalPointer<DateIntervalFormat> formatter(
            DateIntervalFormat::createInstance(skel, Locale(locale), *status));
    if (U_FAILURE(*status)) {
        return NULL;
    }
    if(tzID != 0) {
        TimeZone *zone = TimeZone::createTimeZone(UnicodeString((UBool)(tzIDLength == -1), tzID, tzIDLength));
        if(zone == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        formatter->adoptTimeZone(zone);
    }
    return (UDateIntervalFormat*)formatter.orphan();
}


U_CAPI void U_EXPORT2
udtitvfmt_close(UDateIntervalFormat *formatter)
{
    delete (DateIntervalFormat*)formatter;
}


U_CAPI int32_t U_EXPORT2
udtitvfmt_format(const UDateIntervalFormat* formatter,
                 UDate           fromDate,
                 UDate           toDate,
                 UChar*          result,
                 int32_t         resultCapacity,
                 UFieldPosition* position,
                 UErrorCode*     status)
{
    if (U_FAILURE(*status)) {
        return -1;
    }
    if (result == NULL ? resultCapacity != 0 : resultCapacity < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString res;
    if (result != NULL) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer (copied from udat_format)
        res.setTo(result, 0, resultCapacity);
    }
    FieldPosition fp;
    if (position != 0) {
        fp.setField(position->field);
    }

    DateInterval interval = DateInterval(fromDate,toDate);
    ((const DateIntervalFormat*)formatter)->format( &interval, res, fp, *status );
    if (U_FAILURE(*status)) {
        return -1;
    }
    if (position != 0) {
        position->beginIndex = fp.getBeginIndex();
        position->endIndex = fp.getEndIndex();
    }

    return res.extract(result, resultCapacity, *status);
}


U_DRAFT void U_EXPORT2
udtitvfmt_formatToResult(
                const UDateIntervalFormat* formatter,
                UFormattedDateInterval* result,
                UDate           fromDate,
                UDate           toDate,
                UErrorCode*     status) {
    if (U_FAILURE(*status)) {
        return;
    }
    auto* resultImpl = UFormattedDateIntervalApiHelper::validate(result, *status);
    DateInterval interval = DateInterval(fromDate,toDate);
    resultImpl->fImpl = reinterpret_cast<const DateIntervalFormat*>(formatter)
        ->formatToValue(interval, *status);
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 2009-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  udatpg.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2007jul30
*   created by: Markus W. Scherer
*/



#if !UCONFIG_NO_FORMATTING







U_NAMESPACE_USE

U_CAPI UDateTimePatternGenerator * U_EXPORT2
udatpg_open(const char *locale, UErrorCode *pErrorCode) {
    if(locale==NULL) {
        return (UDateTimePatternGenerator *)DateTimePatternGenerator::createInstance(*pErrorCode);
    } else {
        return (UDateTimePatternGenerator *)DateTimePatternGenerator::createInstance(Locale(locale), *pErrorCode);
    }
}

U_CAPI UDateTimePatternGenerator * U_EXPORT2
udatpg_openEmpty(UErrorCode *pErrorCode) {
    return (UDateTimePatternGenerator *)DateTimePatternGenerator::createEmptyInstance(*pErrorCode);
}

U_CAPI void U_EXPORT2
udatpg_close(UDateTimePatternGenerator *dtpg) {
    delete (DateTimePatternGenerator *)dtpg;
}

U_CAPI UDateTimePatternGenerator * U_EXPORT2
udatpg_clone(const UDateTimePatternGenerator *dtpg, UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return NULL;
    }
    return (UDateTimePatternGenerator *)(((const DateTimePatternGenerator *)dtpg)->clone());
}

U_CAPI int32_t U_EXPORT2
udatpg_getBestPattern(UDateTimePatternGenerator *dtpg,
                      const UChar *skeleton, int32_t length,
                      UChar *bestPattern, int32_t capacity,
                      UErrorCode *pErrorCode) {
    return udatpg_getBestPatternWithOptions(dtpg, skeleton, length,
                                            UDATPG_MATCH_NO_OPTIONS,
                                            bestPattern, capacity, pErrorCode);
}

U_CAPI int32_t U_EXPORT2
udatpg_getBestPatternWithOptions(UDateTimePatternGenerator *dtpg,
                                 const UChar *skeleton, int32_t length,
                                 UDateTimePatternMatchOptions options,
                                 UChar *bestPattern, int32_t capacity,
                                 UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(skeleton==NULL && length!=0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString skeletonString((UBool)(length<0), skeleton, length);
    UnicodeString result=((DateTimePatternGenerator *)dtpg)->getBestPattern(skeletonString, options, *pErrorCode);
    return result.extract(bestPattern, capacity, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
udatpg_getSkeleton(UDateTimePatternGenerator * /* dtpg */,
                   const UChar *pattern, int32_t length,
                   UChar *skeleton, int32_t capacity,
                   UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(pattern==NULL && length!=0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString patternString((UBool)(length<0), pattern, length);
    UnicodeString result=DateTimePatternGenerator::staticGetSkeleton(
            patternString, *pErrorCode);
    return result.extract(skeleton, capacity, *pErrorCode);
}

U_CAPI int32_t U_EXPORT2
udatpg_getBaseSkeleton(UDateTimePatternGenerator * /* dtpg */,
                       const UChar *pattern, int32_t length,
                       UChar *skeleton, int32_t capacity,
                       UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if(pattern==NULL && length!=0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString patternString((UBool)(length<0), pattern, length);
    UnicodeString result=DateTimePatternGenerator::staticGetBaseSkeleton(
            patternString, *pErrorCode);
    return result.extract(skeleton, capacity, *pErrorCode);
}

U_CAPI UDateTimePatternConflict U_EXPORT2
udatpg_addPattern(UDateTimePatternGenerator *dtpg,
                  const UChar *pattern, int32_t patternLength,
                  UBool override,
                  UChar *conflictingPattern, int32_t capacity, int32_t *pLength,
                  UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return UDATPG_NO_CONFLICT;
    }
    if(pattern==NULL && patternLength!=0) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return UDATPG_NO_CONFLICT;
    }
    UnicodeString patternString((UBool)(patternLength<0), pattern, patternLength);
    UnicodeString conflictingPatternString;
    UDateTimePatternConflict result=((DateTimePatternGenerator *)dtpg)->
            addPattern(patternString, override, conflictingPatternString, *pErrorCode);
    int32_t length=conflictingPatternString.extract(conflictingPattern, capacity, *pErrorCode);
    if(pLength!=NULL) {
        *pLength=length;
    }
    return result;
}

U_CAPI void U_EXPORT2
udatpg_setAppendItemFormat(UDateTimePatternGenerator *dtpg,
                           UDateTimePatternField field,
                           const UChar *value, int32_t length) {
    UnicodeString valueString((UBool)(length<0), value, length);
    ((DateTimePatternGenerator *)dtpg)->setAppendItemFormat(field, valueString);
}

U_CAPI const UChar * U_EXPORT2
udatpg_getAppendItemFormat(const UDateTimePatternGenerator *dtpg,
                           UDateTimePatternField field,
                           int32_t *pLength) {
    const UnicodeString &result=((const DateTimePatternGenerator *)dtpg)->getAppendItemFormat(field);
    if(pLength!=NULL) {
        *pLength=result.length();
    }
    return result.getBuffer();
}

U_CAPI void U_EXPORT2
udatpg_setAppendItemName(UDateTimePatternGenerator *dtpg,
                         UDateTimePatternField field,
                         const UChar *value, int32_t length) {
    UnicodeString valueString((UBool)(length<0), value, length);
    ((DateTimePatternGenerator *)dtpg)->setAppendItemName(field, valueString);
}

U_CAPI const UChar * U_EXPORT2
udatpg_getAppendItemName(const UDateTimePatternGenerator *dtpg,
                         UDateTimePatternField field,
                         int32_t *pLength) {
    const UnicodeString &result=((const DateTimePatternGenerator *)dtpg)->getAppendItemName(field);
    if(pLength!=NULL) {
        *pLength=result.length();
    }
    return result.getBuffer();
}

U_CAPI int32_t U_EXPORT2
udatpg_getFieldDisplayName(const UDateTimePatternGenerator *dtpg,
                           UDateTimePatternField field,
                           UDateTimePGDisplayWidth width,
                           UChar *fieldName, int32_t capacity,
                           UErrorCode *pErrorCode) {
    if (U_FAILURE(*pErrorCode))
        return -1;
    if (fieldName == NULL ? capacity != 0 : capacity < 0) {
        *pErrorCode = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }
    UnicodeString result = ((const DateTimePatternGenerator *)dtpg)->getFieldDisplayName(field,width);
    if (fieldName == NULL) {
        return result.length();
    }
    return result.extract(fieldName, capacity, *pErrorCode);
}

U_CAPI void U_EXPORT2
udatpg_setDateTimeFormat(const UDateTimePatternGenerator *dtpg,
                         const UChar *dtFormat, int32_t length) {
    UnicodeString dtFormatString((UBool)(length<0), dtFormat, length);
    ((DateTimePatternGenerator *)dtpg)->setDateTimeFormat(dtFormatString);
}

U_CAPI const UChar * U_EXPORT2
udatpg_getDateTimeFormat(const UDateTimePatternGenerator *dtpg,
                         int32_t *pLength) {
    const UnicodeString &result=((const DateTimePatternGenerator *)dtpg)->getDateTimeFormat();
    if(pLength!=NULL) {
        *pLength=result.length();
    }
    return result.getBuffer();
}

U_CAPI void U_EXPORT2
udatpg_setDecimal(UDateTimePatternGenerator *dtpg,
                  const UChar *decimal, int32_t length) {
    UnicodeString decimalString((UBool)(length<0), decimal, length);
    ((DateTimePatternGenerator *)dtpg)->setDecimal(decimalString);
}

U_CAPI const UChar * U_EXPORT2
udatpg_getDecimal(const UDateTimePatternGenerator *dtpg,
                  int32_t *pLength) {
    const UnicodeString &result=((const DateTimePatternGenerator *)dtpg)->getDecimal();
    if(pLength!=NULL) {
        *pLength=result.length();
    }
    return result.getBuffer();
}

U_CAPI int32_t U_EXPORT2
udatpg_replaceFieldTypes(UDateTimePatternGenerator *dtpg,
                         const UChar *pattern, int32_t patternLength,
                         const UChar *skeleton, int32_t skeletonLength,
                         UChar *dest, int32_t destCapacity,
                         UErrorCode *pErrorCode) {
    return udatpg_replaceFieldTypesWithOptions(dtpg, pattern, patternLength, skeleton, skeletonLength,
                                               UDATPG_MATCH_NO_OPTIONS,
                                               dest, destCapacity, pErrorCode);
}

U_CAPI int32_t U_EXPORT2
udatpg_replaceFieldTypesWithOptions(UDateTimePatternGenerator *dtpg,
                                    const UChar *pattern, int32_t patternLength,
                                    const UChar *skeleton, int32_t skeletonLength,
                                    UDateTimePatternMatchOptions options,
                                    UChar *dest, int32_t destCapacity,
                                    UErrorCode *pErrorCode) {
    if(U_FAILURE(*pErrorCode)) {
        return 0;
    }
    if((pattern==NULL && patternLength!=0) || (skeleton==NULL && skeletonLength!=0)) {
        *pErrorCode=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString patternString((UBool)(patternLength<0), pattern, patternLength);
    UnicodeString skeletonString((UBool)(skeletonLength<0), skeleton, skeletonLength);
    UnicodeString result=((DateTimePatternGenerator *)dtpg)->replaceFieldTypes(patternString, skeletonString, options, *pErrorCode);
    return result.extract(dest, destCapacity, *pErrorCode);
}

U_CAPI UEnumeration * U_EXPORT2
udatpg_openSkeletons(const UDateTimePatternGenerator *dtpg, UErrorCode *pErrorCode) {
    return uenum_openFromStringEnumeration(
                ((DateTimePatternGenerator *)dtpg)->getSkeletons(*pErrorCode),
                pErrorCode);
}

U_CAPI UEnumeration * U_EXPORT2
udatpg_openBaseSkeletons(const UDateTimePatternGenerator *dtpg, UErrorCode *pErrorCode) {
    return uenum_openFromStringEnumeration(
                ((DateTimePatternGenerator *)dtpg)->getBaseSkeletons(*pErrorCode),
                pErrorCode);
}

U_CAPI const UChar * U_EXPORT2
udatpg_getPatternForSkeleton(const UDateTimePatternGenerator *dtpg,
                             const UChar *skeleton, int32_t skeletonLength,
                             int32_t *pLength) {
    UnicodeString skeletonString((UBool)(skeletonLength<0), skeleton, skeletonLength);
    const UnicodeString &result=((const DateTimePatternGenerator *)dtpg)->getPatternForSkeleton(skeletonString);
    if(pLength!=NULL) {
        *pLength=result.length();
    }
    return result.getBuffer();
}

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2015, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/



#if !UCONFIG_NO_FORMATTING





U_NAMESPACE_USE


U_CAPI UFieldPositionIterator* U_EXPORT2
ufieldpositer_open(UErrorCode* status)
{
    if (U_FAILURE(*status)) {
        return NULL;
    }
    FieldPositionIterator* fpositer = new FieldPositionIterator();
    if (fpositer == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
    }
    return (UFieldPositionIterator*)fpositer;
}


U_CAPI void U_EXPORT2
ufieldpositer_close(UFieldPositionIterator *fpositer)
{
    delete (FieldPositionIterator*)fpositer;
}


U_CAPI int32_t U_EXPORT2
ufieldpositer_next(UFieldPositionIterator *fpositer,
                   int32_t *beginIndex, int32_t *endIndex)
{
    FieldPosition fp;
    int32_t field = -1;
    if (((FieldPositionIterator*)fpositer)->next(fp)) {
        field = fp.getField();
        if (beginIndex) {
            *beginIndex = fp.getBeginIndex();
        }
        if (endIndex) {
            *endIndex = fp.getEndIndex();
        }
    }
    return field;
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* uitercollationiterator.cpp
*
* created on: 2012sep23 (from utf16collationiterator.cpp)
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION












U_NAMESPACE_BEGIN

UIterCollationIterator::~UIterCollationIterator() {}

void
UIterCollationIterator::resetToOffset(int32_t newOffset) {
    reset();
    iter.move(&iter, newOffset, UITER_START);
}

int32_t
UIterCollationIterator::getOffset() const {
    return iter.getIndex(&iter, UITER_CURRENT);
}

uint32_t
UIterCollationIterator::handleNextCE32(UChar32 &c, UErrorCode & /*errorCode*/) {
    c = iter.next(&iter);
    if(c < 0) {
        return Collation::FALLBACK_CE32;
    }
    return UTRIE2_GET32_FROM_U16_SINGLE_LEAD(trie, c);
}

UChar
UIterCollationIterator::handleGetTrailSurrogate() {
    UChar32 trail = iter.next(&iter);
    if(!U16_IS_TRAIL(trail) && trail >= 0) { iter.previous(&iter); }
    return (UChar)trail;
}

UChar32
UIterCollationIterator::nextCodePoint(UErrorCode & /*errorCode*/) {
    return uiter_next32(&iter);
}

UChar32
UIterCollationIterator::previousCodePoint(UErrorCode & /*errorCode*/) {
    return uiter_previous32(&iter);
}

void
UIterCollationIterator::forwardNumCodePoints(int32_t num, UErrorCode & /*errorCode*/) {
    while(num > 0 && (uiter_next32(&iter)) >= 0) {
        --num;
    }
}

void
UIterCollationIterator::backwardNumCodePoints(int32_t num, UErrorCode & /*errorCode*/) {
    while(num > 0 && (uiter_previous32(&iter)) >= 0) {
        --num;
    }
}

// FCDUIterCollationIterator ----------------------------------------------- ***

FCDUIterCollationIterator::~FCDUIterCollationIterator() {}

void
FCDUIterCollationIterator::resetToOffset(int32_t newOffset) {
    UIterCollationIterator::resetToOffset(newOffset);
    start = newOffset;
    state = ITER_CHECK_FWD;
}

int32_t
FCDUIterCollationIterator::getOffset() const {
    if(state <= ITER_CHECK_BWD) {
        return iter.getIndex(&iter, UITER_CURRENT);
    } else if(state == ITER_IN_FCD_SEGMENT) {
        return pos;
    } else if(pos == 0) {
        return start;
    } else {
        return limit;
    }
}

uint32_t
FCDUIterCollationIterator::handleNextCE32(UChar32 &c, UErrorCode &errorCode) {
    for(;;) {
        if(state == ITER_CHECK_FWD) {
            c = iter.next(&iter);
            if(c < 0) {
                return Collation::FALLBACK_CE32;
            }
            if(CollationFCD::hasTccc(c)) {
                if(CollationFCD::maybeTibetanCompositeVowel(c) ||
                        CollationFCD::hasLccc(iter.current(&iter))) {
                    iter.previous(&iter);
                    if(!nextSegment(errorCode)) {
                        c = U_SENTINEL;
                        return Collation::FALLBACK_CE32;
                    }
                    continue;
                }
            }
            break;
        } else if(state == ITER_IN_FCD_SEGMENT && pos != limit) {
            c = iter.next(&iter);
            ++pos;
            U_ASSERT(c >= 0);
            break;
        } else if(state >= IN_NORM_ITER_AT_LIMIT && pos != normalized.length()) {
            c = normalized[pos++];
            break;
        } else {
            switchToForward();
        }
    }
    return UTRIE2_GET32_FROM_U16_SINGLE_LEAD(trie, c);
}

UChar
FCDUIterCollationIterator::handleGetTrailSurrogate() {
    if(state <= ITER_IN_FCD_SEGMENT) {
        UChar32 trail = iter.next(&iter);
        if(U16_IS_TRAIL(trail)) {
            if(state == ITER_IN_FCD_SEGMENT) { ++pos; }
        } else if(trail >= 0) {
            iter.previous(&iter);
        }
        return (UChar)trail;
    } else {
        U_ASSERT(pos < normalized.length());
        UChar trail;
        if(U16_IS_TRAIL(trail = normalized[pos])) { ++pos; }
        return trail;
    }
}

UChar32
FCDUIterCollationIterator::nextCodePoint(UErrorCode &errorCode) {
    UChar32 c;
    for(;;) {
        if(state == ITER_CHECK_FWD) {
            c = iter.next(&iter);
            if(c < 0) {
                return c;
            }
            if(CollationFCD::hasTccc(c)) {
                if(CollationFCD::maybeTibetanCompositeVowel(c) ||
                        CollationFCD::hasLccc(iter.current(&iter))) {
                    iter.previous(&iter);
                    if(!nextSegment(errorCode)) {
                        return U_SENTINEL;
                    }
                    continue;
                }
            }
            if(U16_IS_LEAD(c)) {
                UChar32 trail = iter.next(&iter);
                if(U16_IS_TRAIL(trail)) {
                    return U16_GET_SUPPLEMENTARY(c, trail);
                } else if(trail >= 0) {
                    iter.previous(&iter);
                }
            }
            return c;
        } else if(state == ITER_IN_FCD_SEGMENT && pos != limit) {
            c = uiter_next32(&iter);
            pos += U16_LENGTH(c);
            U_ASSERT(c >= 0);
            return c;
        } else if(state >= IN_NORM_ITER_AT_LIMIT && pos != normalized.length()) {
            c = normalized.char32At(pos);
            pos += U16_LENGTH(c);
            return c;
        } else {
            switchToForward();
        }
    }
}

UChar32
FCDUIterCollationIterator::previousCodePoint(UErrorCode &errorCode) {
    UChar32 c;
    for(;;) {
        if(state == ITER_CHECK_BWD) {
            c = iter.previous(&iter);
            if(c < 0) {
                start = pos = 0;
                state = ITER_IN_FCD_SEGMENT;
                return U_SENTINEL;
            }
            if(CollationFCD::hasLccc(c)) {
                UChar32 prev = U_SENTINEL;
                if(CollationFCD::maybeTibetanCompositeVowel(c) ||
                        CollationFCD::hasTccc(prev = iter.previous(&iter))) {
                    iter.next(&iter);
                    if(prev >= 0) {
                        iter.next(&iter);
                    }
                    if(!previousSegment(errorCode)) {
                        return U_SENTINEL;
                    }
                    continue;
                }
                // hasLccc(trail)=true for all trail surrogates
                if(U16_IS_TRAIL(c)) {
                    if(prev < 0) {
                        prev = iter.previous(&iter);
                    }
                    if(U16_IS_LEAD(prev)) {
                        return U16_GET_SUPPLEMENTARY(prev, c);
                    }
                }
                if(prev >= 0) {
                    iter.next(&iter);
                }
            }
            return c;
        } else if(state == ITER_IN_FCD_SEGMENT && pos != start) {
            c = uiter_previous32(&iter);
            pos -= U16_LENGTH(c);
            U_ASSERT(c >= 0);
            return c;
        } else if(state >= IN_NORM_ITER_AT_LIMIT && pos != 0) {
            c = normalized.char32At(pos - 1);
            pos -= U16_LENGTH(c);
            return c;
        } else {
            switchToBackward();
        }
    }
}

void
FCDUIterCollationIterator::forwardNumCodePoints(int32_t num, UErrorCode &errorCode) {
    // Specify the class to avoid a virtual-function indirection.
    // In Java, we would declare this class final.
    while(num > 0 && FCDUIterCollationIterator::nextCodePoint(errorCode) >= 0) {
        --num;
    }
}

void
FCDUIterCollationIterator::backwardNumCodePoints(int32_t num, UErrorCode &errorCode) {
    // Specify the class to avoid a virtual-function indirection.
    // In Java, we would declare this class final.
    while(num > 0 && FCDUIterCollationIterator::previousCodePoint(errorCode) >= 0) {
        --num;
    }
}

void
FCDUIterCollationIterator::switchToForward() {
    U_ASSERT(state == ITER_CHECK_BWD ||
             (state == ITER_IN_FCD_SEGMENT && pos == limit) ||
             (state >= IN_NORM_ITER_AT_LIMIT && pos == normalized.length()));
    if(state == ITER_CHECK_BWD) {
        // Turn around from backward checking.
        start = pos = iter.getIndex(&iter, UITER_CURRENT);
        if(pos == limit) {
            state = ITER_CHECK_FWD;  // Check forward.
        } else {  // pos < limit
            state = ITER_IN_FCD_SEGMENT;  // Stay in FCD segment.
        }
    } else {
        // Reached the end of the FCD segment.
        if(state == ITER_IN_FCD_SEGMENT) {
            // The input text segment is FCD, extend it forward.
        } else {
            // The input text segment needed to be normalized.
            // Switch to checking forward from it.
            if(state == IN_NORM_ITER_AT_START) {
                iter.move(&iter, limit - start, UITER_CURRENT);
            }
            start = limit;
        }
        state = ITER_CHECK_FWD;
    }
}

UBool
FCDUIterCollationIterator::nextSegment(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    U_ASSERT(state == ITER_CHECK_FWD);
    // The input text [start..(iter index)[ passes the FCD check.
    pos = iter.getIndex(&iter, UITER_CURRENT);
    // Collect the characters being checked, in case they need to be normalized.
    UnicodeString s;
    uint8_t prevCC = 0;
    for(;;) {
        // Fetch the next character and its fcd16 value.
        UChar32 c = uiter_next32(&iter);
        if(c < 0) { break; }
        uint16_t fcd16 = nfcImpl.getFCD16(c);
        uint8_t leadCC = (uint8_t)(fcd16 >> 8);
        if(leadCC == 0 && !s.isEmpty()) {
            // FCD boundary before this character.
            uiter_previous32(&iter);
            break;
        }
        s.append(c);
        if(leadCC != 0 && (prevCC > leadCC || CollationFCD::isFCD16OfTibetanCompositeVowel(fcd16))) {
            // Fails FCD check. Find the next FCD boundary and normalize.
            for(;;) {
                c = uiter_next32(&iter);
                if(c < 0) { break; }
                if(nfcImpl.getFCD16(c) <= 0xff) {
                    uiter_previous32(&iter);
                    break;
                }
                s.append(c);
            }
            if(!normalize(s, errorCode)) { return FALSE; }
            start = pos;
            limit = pos + s.length();
            state = IN_NORM_ITER_AT_LIMIT;
            pos = 0;
            return TRUE;
        }
        prevCC = (uint8_t)fcd16;
        if(prevCC == 0) {
            // FCD boundary after the last character.
            break;
        }
    }
    limit = pos + s.length();
    U_ASSERT(pos != limit);
    iter.move(&iter, -s.length(), UITER_CURRENT);
    state = ITER_IN_FCD_SEGMENT;
    return TRUE;
}

void
FCDUIterCollationIterator::switchToBackward() {
    U_ASSERT(state == ITER_CHECK_FWD ||
             (state == ITER_IN_FCD_SEGMENT && pos == start) ||
             (state >= IN_NORM_ITER_AT_LIMIT && pos == 0));
    if(state == ITER_CHECK_FWD) {
        // Turn around from forward checking.
        limit = pos = iter.getIndex(&iter, UITER_CURRENT);
        if(pos == start) {
            state = ITER_CHECK_BWD;  // Check backward.
        } else {  // pos > start
            state = ITER_IN_FCD_SEGMENT;  // Stay in FCD segment.
        }
    } else {
        // Reached the start of the FCD segment.
        if(state == ITER_IN_FCD_SEGMENT) {
            // The input text segment is FCD, extend it backward.
        } else {
            // The input text segment needed to be normalized.
            // Switch to checking backward from it.
            if(state == IN_NORM_ITER_AT_LIMIT) {
                iter.move(&iter, start - limit, UITER_CURRENT);
            }
            limit = start;
        }
        state = ITER_CHECK_BWD;
    }
}

UBool
FCDUIterCollationIterator::previousSegment(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    U_ASSERT(state == ITER_CHECK_BWD);
    // The input text [(iter index)..limit[ passes the FCD check.
    pos = iter.getIndex(&iter, UITER_CURRENT);
    // Collect the characters being checked, in case they need to be normalized.
    UnicodeString s;
    uint8_t nextCC = 0;
    for(;;) {
        // Fetch the previous character and its fcd16 value.
        UChar32 c = uiter_previous32(&iter);
        if(c < 0) { break; }
        uint16_t fcd16 = nfcImpl.getFCD16(c);
        uint8_t trailCC = (uint8_t)fcd16;
        if(trailCC == 0 && !s.isEmpty()) {
            // FCD boundary after this character.
            uiter_next32(&iter);
            break;
        }
        s.append(c);
        if(trailCC != 0 && ((nextCC != 0 && trailCC > nextCC) ||
                            CollationFCD::isFCD16OfTibetanCompositeVowel(fcd16))) {
            // Fails FCD check. Find the previous FCD boundary and normalize.
            while(fcd16 > 0xff) {
                c = uiter_previous32(&iter);
                if(c < 0) { break; }
                fcd16 = nfcImpl.getFCD16(c);
                if(fcd16 == 0) {
                    (void)uiter_next32(&iter);
                    break;
                }
                s.append(c);
            }
            s.reverse();
            if(!normalize(s, errorCode)) { return FALSE; }
            limit = pos;
            start = pos - s.length();
            state = IN_NORM_ITER_AT_START;
            pos = normalized.length();
            return TRUE;
        }
        nextCC = (uint8_t)(fcd16 >> 8);
        if(nextCC == 0) {
            // FCD boundary before the following character.
            break;
        }
    }
    start = pos - s.length();
    U_ASSERT(pos != start);
    iter.move(&iter, s.length(), UITER_CURRENT);
    state = ITER_IN_FCD_SEGMENT;
    return TRUE;
}

UBool
FCDUIterCollationIterator::normalize(const UnicodeString &s, UErrorCode &errorCode) {
    // NFD without argument checking.
    U_ASSERT(U_SUCCESS(errorCode));
    nfcImpl.decompose(s, normalized, errorCode);
    return U_SUCCESS(errorCode);
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2015, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/



#if !UCONFIG_NO_FORMATTING







U_NAMESPACE_USE

U_CAPI UListFormatter* U_EXPORT2
ulistfmt_open(const char*  locale,
              UErrorCode*  status)
{
    if (U_FAILURE(*status)) {
        return NULL;
    }
    LocalPointer<ListFormatter> listfmt(ListFormatter::createInstance(Locale(locale), *status));
    if (U_FAILURE(*status)) {
        return NULL;
    }
    return (UListFormatter*)listfmt.orphan();
}


U_CAPI void U_EXPORT2
ulistfmt_close(UListFormatter *listfmt)
{
    delete (ListFormatter*)listfmt;
}


// Magic number: FLST in ASCII
UPRV_FORMATTED_VALUE_CAPI_AUTO_IMPL(
    FormattedList,
    UFormattedList,
    UFormattedListImpl,
    UFormattedListApiHelper,
    ulistfmt,
    0x464C5354)


static UnicodeString* getUnicodeStrings(
        const UChar* const strings[],
        const int32_t* stringLengths,
        int32_t stringCount,
        UnicodeString* length4StackBuffer,
        LocalArray<UnicodeString>& maybeOwner,
        UErrorCode& status) {
    U_ASSERT(U_SUCCESS(status));
    if (stringCount < 0 || (strings == NULL && stringCount > 0)) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return nullptr;
    }
    UnicodeString* ustrings = length4StackBuffer;
    if (stringCount > 4) {
        maybeOwner.adoptInsteadAndCheckErrorCode(new UnicodeString[stringCount], status);
        if (U_FAILURE(status)) {
            return nullptr;
        }
        ustrings = maybeOwner.getAlias();
    }
    if (stringLengths == NULL) {
        for (int32_t stringIndex = 0; stringIndex < stringCount; stringIndex++) {
            ustrings[stringIndex].setTo(TRUE, strings[stringIndex], -1);
        }
    } else {
        for (int32_t stringIndex = 0; stringIndex < stringCount; stringIndex++) {
            ustrings[stringIndex].setTo(stringLengths[stringIndex] < 0, strings[stringIndex], stringLengths[stringIndex]);
        }
    }
    return ustrings;
}


U_CAPI int32_t U_EXPORT2
ulistfmt_format(const UListFormatter* listfmt,
                const UChar* const strings[],
                const int32_t *    stringLengths,
                int32_t            stringCount,
                UChar*             result,
                int32_t            resultCapacity,
                UErrorCode*        status)
{
    if (U_FAILURE(*status)) {
        return -1;
    }
    if ((result == NULL) ? resultCapacity != 0 : resultCapacity < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }
    UnicodeString length4StackBuffer[4];
    LocalArray<UnicodeString> maybeOwner;
    UnicodeString* ustrings = getUnicodeStrings(
        strings, stringLengths, stringCount, length4StackBuffer, maybeOwner, *status);
    if (U_FAILURE(*status)) {
        return -1;
    }
    UnicodeString res;
    if (result != NULL) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer (copied from udat_format)
        res.setTo(result, 0, resultCapacity);
    }
    reinterpret_cast<const ListFormatter*>(listfmt)->format( ustrings, stringCount, res, *status );
    return res.extract(result, resultCapacity, *status);
}


U_CAPI void U_EXPORT2
ulistfmt_formatStringsToResult(
                const UListFormatter* listfmt,
                const UChar* const strings[],
                const int32_t *    stringLengths,
                int32_t            stringCount,
                UFormattedList*    uresult,
                UErrorCode*        status) {
    auto* result = UFormattedListApiHelper::validate(uresult, *status);
    if (U_FAILURE(*status)) {
        return;
    }
    UnicodeString length4StackBuffer[4];
    LocalArray<UnicodeString> maybeOwner;
    UnicodeString* ustrings = getUnicodeStrings(
        strings, stringLengths, stringCount, length4StackBuffer, maybeOwner, *status);
    if (U_FAILURE(*status)) {
        return;
    }
    result->fImpl = reinterpret_cast<const ListFormatter*>(listfmt)
        ->formatStringsToValue(ustrings, stringCount, *status);
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*                                                                            *
* Copyright (C) 2003-2016, International Business Machines                   *
*                Corporation and others. All Rights Reserved.                *
*                                                                            *
******************************************************************************
*   file name:  ulocdata.c
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2003Oct21
*   created by: Ram Viswanadha,John Emmons
*/










#define MEASUREMENT_SYSTEM  "MeasurementSystem"
#define PAPER_SIZE          "PaperSize"

/** A locale data object.
 *  For usage in C programs.
 *  @draft ICU 3.4
 */
struct ULocaleData {
    /**
     * Controls the "No Substitute" behavior of this locale data object
     */
    UBool noSubstitute;

    /**
     * Pointer to the resource bundle associated with this locale data object
     */
    UResourceBundle *bundle;

    /**
     * Pointer to the lang resource bundle associated with this locale data object
     */
    UResourceBundle *langBundle;
};

U_CAPI ULocaleData* U_EXPORT2
ulocdata_open(const char *localeID, UErrorCode *status)
{
   ULocaleData *uld;

   if (U_FAILURE(*status)) {
       return NULL;
   }

   uld = (ULocaleData *)uprv_malloc(sizeof(ULocaleData));
   if (uld == NULL) {
      *status = U_MEMORY_ALLOCATION_ERROR;
      return(NULL);
   }

   uld->langBundle = NULL;

   uld->noSubstitute = FALSE;
   uld->bundle = ures_open(NULL, localeID, status);
   uld->langBundle = ures_open(U_ICUDATA_LANG, localeID, status);

   if (U_FAILURE(*status)) {
      uprv_free(uld);
      return NULL;
   }

   return uld;
}

U_CAPI void U_EXPORT2
ulocdata_close(ULocaleData *uld)
{
    if ( uld != NULL ) {
       ures_close(uld->langBundle);
       ures_close(uld->bundle);
       uprv_free(uld);
    }
}

U_CAPI void U_EXPORT2
ulocdata_setNoSubstitute(ULocaleData *uld, UBool setting)
{
   uld->noSubstitute = setting;
}

U_CAPI UBool U_EXPORT2
ulocdata_getNoSubstitute(ULocaleData *uld)
{
   return uld->noSubstitute;
}

U_CAPI USet* U_EXPORT2
ulocdata_getExemplarSet(ULocaleData *uld, USet *fillIn,
                        uint32_t options, ULocaleDataExemplarSetType extype, UErrorCode *status){

    static const char* const exemplarSetTypes[] = { "ExemplarCharacters", 
                                                    "AuxExemplarCharacters", 
                                                    "ExemplarCharactersIndex",
                                                    "ExemplarCharactersPunctuation"};
    const UChar *exemplarChars = NULL;
    int32_t len = 0;
    UErrorCode localStatus = U_ZERO_ERROR;

    if (U_FAILURE(*status))
        return NULL;

    exemplarChars = ures_getStringByKey(uld->bundle, exemplarSetTypes[extype], &len, &localStatus);
    if ( (localStatus == U_USING_DEFAULT_WARNING) && uld->noSubstitute ) {
        localStatus = U_MISSING_RESOURCE_ERROR;
    }

    if (localStatus != U_ZERO_ERROR) {
        *status = localStatus;
    }

    if (U_FAILURE(*status))
        return NULL;

    if(fillIn != NULL)
        uset_applyPattern(fillIn, exemplarChars, len,
                          USET_IGNORE_SPACE | options, status);
    else
        fillIn = uset_openPatternOptions(exemplarChars, len,
                                         USET_IGNORE_SPACE | options, status);

    return fillIn;

}

U_CAPI int32_t U_EXPORT2
ulocdata_getDelimiter(ULocaleData *uld, ULocaleDataDelimiterType type,
                      UChar *result, int32_t resultLength, UErrorCode *status){

    static const char* const delimiterKeys[] =  {
        "quotationStart",
        "quotationEnd",
        "alternateQuotationStart",
        "alternateQuotationEnd"
    };

    UResourceBundle *delimiterBundle;
    int32_t len = 0;
    const UChar *delimiter = NULL;
    UErrorCode localStatus = U_ZERO_ERROR;

    if (U_FAILURE(*status))
        return 0;

    delimiterBundle = ures_getByKey(uld->bundle, "delimiters", NULL, &localStatus);

    if ( (localStatus == U_USING_DEFAULT_WARNING) && uld->noSubstitute ) {
        localStatus = U_MISSING_RESOURCE_ERROR;
    }

    if (localStatus != U_ZERO_ERROR) {
        *status = localStatus;
    }

    if (U_FAILURE(*status)){
        ures_close(delimiterBundle);
        return 0;
    }

    delimiter = ures_getStringByKey(delimiterBundle, delimiterKeys[type], &len, &localStatus);
    ures_close(delimiterBundle);

    if ( (localStatus == U_USING_DEFAULT_WARNING) && uld->noSubstitute ) {
        localStatus = U_MISSING_RESOURCE_ERROR;
    }

    if (localStatus != U_ZERO_ERROR) {
        *status = localStatus;
    }

    if (U_FAILURE(*status)){
        return 0;
    }

    u_strncpy(result,delimiter, resultLength);
    return len;
}

static UResourceBundle * measurementTypeBundleForLocale(const char *localeID, const char *measurementType, UErrorCode *status){
    char region[ULOC_COUNTRY_CAPACITY];
    UResourceBundle *rb;
    UResourceBundle *measTypeBundle = NULL;
    
    ulocimp_getRegionForSupplementalData(localeID, TRUE, region, ULOC_COUNTRY_CAPACITY, status);
    
    rb = ures_openDirect(NULL, "supplementalData", status);
    ures_getByKey(rb, "measurementData", rb, status);
    if (rb != NULL) {
        UResourceBundle *measDataBundle = ures_getByKey(rb, region, NULL, status);
        if (U_SUCCESS(*status)) {
        	measTypeBundle = ures_getByKey(measDataBundle, measurementType, NULL, status);
        }
        if (*status == U_MISSING_RESOURCE_ERROR) {
            *status = U_ZERO_ERROR;
            if (measDataBundle != NULL) {
                ures_close(measDataBundle);
            }
            measDataBundle = ures_getByKey(rb, "001", NULL, status);
            measTypeBundle = ures_getByKey(measDataBundle, measurementType, NULL, status);
        }
        ures_close(measDataBundle);
    }
    ures_close(rb);
    return measTypeBundle;
}

U_CAPI UMeasurementSystem U_EXPORT2
ulocdata_getMeasurementSystem(const char *localeID, UErrorCode *status){

    UResourceBundle* measurement=NULL;
    UMeasurementSystem system = UMS_LIMIT;

    if(status == NULL || U_FAILURE(*status)){
        return system;
    }

    measurement = measurementTypeBundleForLocale(localeID, MEASUREMENT_SYSTEM, status);
    system = (UMeasurementSystem) ures_getInt(measurement, status);

    ures_close(measurement);

    return system;

}

U_CAPI void U_EXPORT2
ulocdata_getPaperSize(const char* localeID, int32_t *height, int32_t *width, UErrorCode *status){
    UResourceBundle* paperSizeBundle = NULL;
    const int32_t* paperSize=NULL;
    int32_t len = 0;

    if(status == NULL || U_FAILURE(*status)){
        return;
    }

    paperSizeBundle = measurementTypeBundleForLocale(localeID, PAPER_SIZE, status);
    paperSize = ures_getIntVector(paperSizeBundle, &len,  status);

    if(U_SUCCESS(*status)){
        if(len < 2){
            *status = U_INTERNAL_PROGRAM_ERROR;
        }else{
            *height = paperSize[0];
            *width  = paperSize[1];
        }
    }

    ures_close(paperSizeBundle);

}

U_CAPI void U_EXPORT2
ulocdata_getCLDRVersion(UVersionInfo versionArray, UErrorCode *status) {
    UResourceBundle *rb = NULL;
    rb = ures_openDirect(NULL, "supplementalData", status);
    ures_getVersionByKey(rb, "cldrVersion", versionArray, status);
    ures_close(rb);
}

U_CAPI int32_t U_EXPORT2
ulocdata_getLocaleDisplayPattern(ULocaleData *uld,
                                 UChar *result,
                                 int32_t resultCapacity,
                                 UErrorCode *status) {
    UResourceBundle *patternBundle;
    int32_t len = 0;
    const UChar *pattern = NULL;
    UErrorCode localStatus = U_ZERO_ERROR;

    if (U_FAILURE(*status))
        return 0;

    patternBundle = ures_getByKey(uld->langBundle, "localeDisplayPattern", NULL, &localStatus);

    if ( (localStatus == U_USING_DEFAULT_WARNING) && uld->noSubstitute ) {
        localStatus = U_MISSING_RESOURCE_ERROR;
    }

    if (localStatus != U_ZERO_ERROR) {
        *status = localStatus;
    }

    if (U_FAILURE(*status)){
        ures_close(patternBundle);
        return 0;
    }

    pattern = ures_getStringByKey(patternBundle, "pattern", &len, &localStatus);
    ures_close(patternBundle);

    if ( (localStatus == U_USING_DEFAULT_WARNING) && uld->noSubstitute ) {
        localStatus = U_MISSING_RESOURCE_ERROR;
    }

    if (localStatus != U_ZERO_ERROR) {
        *status = localStatus;
    }

    if (U_FAILURE(*status)){
        return 0;
    }

    u_strncpy(result, pattern, resultCapacity);
    return len;
}


U_CAPI int32_t U_EXPORT2
ulocdata_getLocaleSeparator(ULocaleData *uld,
                            UChar *result,
                            int32_t resultCapacity,
                            UErrorCode *status)  {
    UResourceBundle *separatorBundle;
    int32_t len = 0;
    const UChar *separator = NULL;
    UErrorCode localStatus = U_ZERO_ERROR;
    UChar *p0, *p1;
    static const UChar sub0[4] = { 0x007b, 0x0030, 0x007d , 0x0000 }; /* {0} */
    static const UChar sub1[4] = { 0x007b, 0x0031, 0x007d , 0x0000 }; /* {1} */
    static const int32_t subLen = 3;

    if (U_FAILURE(*status))
        return 0;

    separatorBundle = ures_getByKey(uld->langBundle, "localeDisplayPattern", NULL, &localStatus);

    if ( (localStatus == U_USING_DEFAULT_WARNING) && uld->noSubstitute ) {
        localStatus = U_MISSING_RESOURCE_ERROR;
    }

    if (localStatus != U_ZERO_ERROR) {
        *status = localStatus;
    }

    if (U_FAILURE(*status)){
        ures_close(separatorBundle);
        return 0;
    }

    separator = ures_getStringByKey(separatorBundle, "separator", &len, &localStatus);
    ures_close(separatorBundle);

    if ( (localStatus == U_USING_DEFAULT_WARNING) && uld->noSubstitute ) {
        localStatus = U_MISSING_RESOURCE_ERROR;
    }

    if (localStatus != U_ZERO_ERROR) {
        *status = localStatus;
    }

    if (U_FAILURE(*status)){
        return 0;
    }

    /* For backwards compatibility, if we have a pattern, return the portion between {0} and {1} */
    p0=u_strstr(separator, sub0);
    p1=u_strstr(separator, sub1);
    if (p0!=NULL && p1!=NULL && p0<=p1) {
        separator = (const UChar *)p0 + subLen;
        len = static_cast<int32_t>(p1 - separator);
        /* Desired separator is no longer zero-terminated; handle that if necessary */
        if (len < resultCapacity) {
            u_strncpy(result, separator, len);
            result[len] = 0;
            return len;
        }
    }

    u_strncpy(result, separator, resultCapacity);
    return len;
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*
*   Copyright (C) 1999-2012, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
*******************************************************************************
*   file name:  umsg.cpp
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
* This is a C wrapper to MessageFormat C++ API.
*
*   Change history:
*
*   08/5/2001  Ram         Added C wrappers for C++ API. Changed implementation of old API's
*                          Removed pattern parser.
* 
*/



#if !UCONFIG_NO_FORMATTING










U_NAMESPACE_BEGIN
/**
 * This class isolates our access to private internal methods of
 * MessageFormat.  It is never instantiated; it exists only for C++
 * access management.
 */
class MessageFormatAdapter {
public:
    static const Formattable::Type* getArgTypeList(const MessageFormat& m,
                                                   int32_t& count);
    static UBool hasArgTypeConflicts(const MessageFormat& m) {
        return m.hasArgTypeConflicts;
    }
};
const Formattable::Type*
MessageFormatAdapter::getArgTypeList(const MessageFormat& m,
                                     int32_t& count) {
    return m.getArgTypeList(count);
}
U_NAMESPACE_END

U_NAMESPACE_USE

U_CAPI int32_t
u_formatMessage(const char  *locale,
                const UChar *pattern,
                int32_t     patternLength,
                UChar       *result,
                int32_t     resultLength,
                UErrorCode  *status,
                ...)
{
    va_list    ap;
    int32_t actLen;        
    //argument checking defered to subsequent method calls
    // start vararg processing
    va_start(ap, status);

    actLen = u_vformatMessage(locale,pattern,patternLength,result,resultLength,ap,status);
    // end vararg processing
    va_end(ap);

    return actLen;
}

U_CAPI int32_t U_EXPORT2
u_vformatMessage(   const char  *locale,
                    const UChar *pattern,
                    int32_t     patternLength,
                    UChar       *result,
                    int32_t     resultLength,
                    va_list     ap,
                    UErrorCode  *status)

{
    //argument checking defered to subsequent method calls
    UMessageFormat *fmt = umsg_open(pattern,patternLength,locale,NULL,status);
    int32_t retVal = umsg_vformat(fmt,result,resultLength,ap,status);
    umsg_close(fmt);
    return retVal;
}

U_CAPI int32_t
u_formatMessageWithError(const char *locale,
                        const UChar *pattern,
                        int32_t     patternLength,
                        UChar       *result,
                        int32_t     resultLength,
                        UParseError *parseError,
                        UErrorCode  *status,
                        ...)
{
    va_list    ap;
    int32_t actLen;
    //argument checking defered to subsequent method calls
    // start vararg processing
    va_start(ap, status);

    actLen = u_vformatMessageWithError(locale,pattern,patternLength,result,resultLength,parseError,ap,status);

    // end vararg processing
    va_end(ap);
    return actLen;
}

U_CAPI int32_t U_EXPORT2
u_vformatMessageWithError(  const char  *locale,
                            const UChar *pattern,
                            int32_t     patternLength,
                            UChar       *result,
                            int32_t     resultLength,
                            UParseError *parseError,
                            va_list     ap,
                            UErrorCode  *status)

{
    //argument checking defered to subsequent method calls
    UMessageFormat *fmt = umsg_open(pattern,patternLength,locale,parseError,status);
    int32_t retVal = umsg_vformat(fmt,result,resultLength,ap,status);
    umsg_close(fmt);
    return retVal;
}


// For parse, do the reverse of format:
//  1. Call through to the C++ APIs
//  2. Just assume the user passed in enough arguments.
//  3. Iterate through each formattable returned, and assign to the arguments
U_CAPI void
u_parseMessage( const char   *locale,
                const UChar  *pattern,
                int32_t      patternLength,
                const UChar  *source,
                int32_t      sourceLength,
                UErrorCode   *status,
                ...)
{
    va_list    ap;
    //argument checking defered to subsequent method calls

    // start vararg processing
    va_start(ap, status);

    u_vparseMessage(locale,pattern,patternLength,source,sourceLength,ap,status);
    // end vararg processing
    va_end(ap);
}

U_CAPI void U_EXPORT2
u_vparseMessage(const char  *locale,
                const UChar *pattern,
                int32_t     patternLength,
                const UChar *source,
                int32_t     sourceLength,
                va_list     ap,
                UErrorCode  *status)
{
    //argument checking defered to subsequent method calls
    UMessageFormat *fmt = umsg_open(pattern,patternLength,locale,NULL,status);
    int32_t count = 0;
    umsg_vparse(fmt,source,sourceLength,&count,ap,status);
    umsg_close(fmt);
}

U_CAPI void
u_parseMessageWithError(const char  *locale,
                        const UChar *pattern,
                        int32_t     patternLength,
                        const UChar *source,
                        int32_t     sourceLength,
                        UParseError *error,
                        UErrorCode  *status,
                        ...)
{
    va_list    ap;

    //argument checking defered to subsequent method calls

    // start vararg processing
    va_start(ap, status);

    u_vparseMessageWithError(locale,pattern,patternLength,source,sourceLength,ap,error,status);
    // end vararg processing
    va_end(ap);
}
U_CAPI void U_EXPORT2
u_vparseMessageWithError(const char  *locale,
                         const UChar *pattern,
                         int32_t     patternLength,
                         const UChar *source,
                         int32_t     sourceLength,
                         va_list     ap,
                         UParseError *error,
                         UErrorCode* status)
{
    //argument checking defered to subsequent method calls
    UMessageFormat *fmt = umsg_open(pattern,patternLength,locale,error,status);
    int32_t count = 0;
    umsg_vparse(fmt,source,sourceLength,&count,ap,status);
    umsg_close(fmt);
}
//////////////////////////////////////////////////////////////////////////////////
//
//  Message format C API
//
/////////////////////////////////////////////////////////////////////////////////


U_CAPI UMessageFormat* U_EXPORT2
umsg_open(  const UChar     *pattern,
            int32_t         patternLength,
            const  char     *locale,
            UParseError     *parseError,
            UErrorCode      *status)
{
    //check arguments
    if(status==NULL || U_FAILURE(*status))
    {
      return 0;
    }
    if(pattern==NULL||patternLength<-1){
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    UParseError tErr;
    if(parseError==NULL)
    {
        parseError = &tErr;
    }

    int32_t len = (patternLength == -1 ? u_strlen(pattern) : patternLength);
    UnicodeString patString(patternLength == -1, pattern, len);

    MessageFormat* retVal = new MessageFormat(patString,Locale(locale),*parseError,*status);
    if(retVal == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    if (U_SUCCESS(*status) && MessageFormatAdapter::hasArgTypeConflicts(*retVal)) {
        *status = U_ARGUMENT_TYPE_MISMATCH;
    }
    return (UMessageFormat*)retVal;
}

U_CAPI void U_EXPORT2
umsg_close(UMessageFormat* format)
{
    //check arguments
    if(format==NULL){
        return;
    }
    delete (MessageFormat*) format;
}

U_CAPI UMessageFormat U_EXPORT2
umsg_clone(const UMessageFormat *fmt,
           UErrorCode *status)
{
    //check arguments
    if(status==NULL || U_FAILURE(*status)){
        return NULL;
    }
    if(fmt==NULL){
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    UMessageFormat retVal = (UMessageFormat)((MessageFormat*)fmt)->clone();
    if(retVal == 0) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    return retVal;    
}

U_CAPI void  U_EXPORT2
umsg_setLocale(UMessageFormat *fmt, const char* locale)
{
    //check arguments
    if(fmt==NULL){
        return;
    }
    ((MessageFormat*)fmt)->setLocale(Locale(locale));   
}

U_CAPI const char*  U_EXPORT2
umsg_getLocale(const UMessageFormat *fmt)
{
    //check arguments
    if(fmt==NULL){
        return "";
    }
    return ((const MessageFormat*)fmt)->getLocale().getName();
}

U_CAPI void  U_EXPORT2
umsg_applyPattern(UMessageFormat *fmt,
                           const UChar* pattern,
                           int32_t patternLength,
                           UParseError* parseError,
                           UErrorCode* status)
{
    //check arguments
    UParseError tErr;
    if(status ==NULL||U_FAILURE(*status)){
        return ;
    }
    if(fmt==NULL || (pattern==NULL && patternLength!=0) || patternLength<-1) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return ;
    }

    if(parseError==NULL){
      parseError = &tErr;
    }

    // UnicodeString(pattern, -1) calls u_strlen().
    ((MessageFormat*)fmt)->applyPattern(UnicodeString(pattern,patternLength),*parseError,*status);  
}

U_CAPI int32_t  U_EXPORT2
umsg_toPattern(const UMessageFormat *fmt,
               UChar* result, 
               int32_t resultLength,
               UErrorCode* status)
{
    //check arguments
    if(status ==NULL||U_FAILURE(*status)){
        return -1;
    }
    if(fmt==NULL||resultLength<0 || (resultLength>0 && result==0)){
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }


    UnicodeString res;
    if(!(result==NULL && resultLength==0)) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        res.setTo(result, 0, resultLength);
    }
    ((const MessageFormat*)fmt)->toPattern(res);
    return res.extract(result, resultLength, *status);
}

U_CAPI int32_t
umsg_format(    const UMessageFormat *fmt,
                UChar          *result,
                int32_t        resultLength,
                UErrorCode     *status,
                ...)
{
    va_list    ap;
    int32_t actLen;  
    //argument checking defered to last method call umsg_vformat which
    //saves time when arguments are valid and we dont care when arguments are not
    //since we return an error anyway

    
    // start vararg processing
    va_start(ap, status);

    actLen = umsg_vformat(fmt,result,resultLength,ap,status);

    // end vararg processing
    va_end(ap);

    return actLen;
}

U_CAPI int32_t U_EXPORT2
umsg_vformat(   const UMessageFormat *fmt,
                UChar          *result,
                int32_t        resultLength,
                va_list        ap,
                UErrorCode     *status)
{
    //check arguments
    if(status==0 || U_FAILURE(*status))
    {
        return -1;
    }
    if(fmt==NULL||resultLength<0 || (resultLength>0 && result==0)) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    int32_t count =0;
    const Formattable::Type* argTypes =
        MessageFormatAdapter::getArgTypeList(*(const MessageFormat*)fmt, count);
    // Allocate at least one element.  Allocating an array of length
    // zero causes problems on some platforms (e.g. Win32).
    Formattable* args = new Formattable[count ? count : 1];

    // iterate through the vararg list, and get the arguments out
    for(int32_t i = 0; i < count; ++i) {
        
        UChar *stringVal;
        double tDouble=0;
        int32_t tInt =0;
        int64_t tInt64 = 0;
        UDate tempDate = 0;
        switch(argTypes[i]) {
        case Formattable::kDate:
            tempDate = va_arg(ap, UDate);
            args[i].setDate(tempDate);
            break;
            
        case Formattable::kDouble:
            tDouble =va_arg(ap, double);
            args[i].setDouble(tDouble);
            break;
            
        case Formattable::kLong:
            tInt = va_arg(ap, int32_t);
            args[i].setLong(tInt);
            break;

        case Formattable::kInt64:
            tInt64 = va_arg(ap, int64_t);
            args[i].setInt64(tInt64);
            break;
            
        case Formattable::kString:
            // For some reason, a temporary is needed
            stringVal = va_arg(ap, UChar*);
            if(stringVal){
                args[i].setString(UnicodeString(stringVal));
            }else{
                *status=U_ILLEGAL_ARGUMENT_ERROR;
            }
            break;
            
        case Formattable::kArray:
            // throw away this argument
            // this is highly platform-dependent, and probably won't work
            // so, if you try to skip arguments in the list (and not use them)
            // you'll probably crash
            va_arg(ap, int);
            break;

        case Formattable::kObject:
            // Unused argument number. Read and ignore a pointer argument.
            va_arg(ap, void*);
            break;

        default:
            // Unknown/unsupported argument type.
            UPRV_UNREACHABLE;
        }
    }
    UnicodeString resultStr;
    FieldPosition fieldPosition(FieldPosition::DONT_CARE);
    
    /* format the message */
    ((const MessageFormat*)fmt)->format(args,count,resultStr,fieldPosition,*status);

    delete[] args;

    if(U_FAILURE(*status)){
        return -1;
    }

    return resultStr.extract(result, resultLength, *status);
}

U_CAPI void
umsg_parse( const UMessageFormat *fmt,
            const UChar    *source,
            int32_t        sourceLength,
            int32_t        *count,
            UErrorCode     *status,
            ...)
{
    va_list    ap;
    //argument checking defered to last method call umsg_vparse which
    //saves time when arguments are valid and we dont care when arguments are not
    //since we return an error anyway

    // start vararg processing
    va_start(ap, status);

    umsg_vparse(fmt,source,sourceLength,count,ap,status);

    // end vararg processing
    va_end(ap);
}

U_CAPI void U_EXPORT2
umsg_vparse(const UMessageFormat *fmt,
            const UChar    *source,
            int32_t        sourceLength,
            int32_t        *count,
            va_list        ap,
            UErrorCode     *status)
{
    //check arguments
    if(status==NULL||U_FAILURE(*status))
    {
        return;
    }
    if(fmt==NULL||source==NULL || sourceLength<-1 || count==NULL){
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    if(sourceLength==-1){
        sourceLength=u_strlen(source);
    }

    UnicodeString srcString(source,sourceLength);
    Formattable *args = ((const MessageFormat*)fmt)->parse(srcString,*count,*status);
    UDate *aDate;
    double *aDouble;
    UChar *aString;
    int32_t* aInt;
    int64_t* aInt64;
    UnicodeString temp;
    int len =0;
    // assign formattables to varargs
    for(int32_t i = 0; i < *count; i++) {
        switch(args[i].getType()) {

        case Formattable::kDate:
            aDate = va_arg(ap, UDate*);
            if(aDate){
                *aDate = args[i].getDate();
            }else{
                *status=U_ILLEGAL_ARGUMENT_ERROR;
            }
            break;

        case Formattable::kDouble:
            aDouble = va_arg(ap, double*);
            if(aDouble){
                *aDouble = args[i].getDouble();
            }else{
                *status=U_ILLEGAL_ARGUMENT_ERROR;
            }
            break;

        case Formattable::kLong:
            aInt = va_arg(ap, int32_t*);
            if(aInt){
                *aInt = (int32_t) args[i].getLong();
            }else{
                *status=U_ILLEGAL_ARGUMENT_ERROR;
            }
            break;

        case Formattable::kInt64:
            aInt64 = va_arg(ap, int64_t*);
            if(aInt64){
                *aInt64 = args[i].getInt64();
            }else{
                *status=U_ILLEGAL_ARGUMENT_ERROR;
            }
            break;

        case Formattable::kString:
            aString = va_arg(ap, UChar*);
            if(aString){
                args[i].getString(temp);
                len = temp.length();
                temp.extract(0,len,aString);
                aString[len]=0;
            }else{
                *status= U_ILLEGAL_ARGUMENT_ERROR;
            }
            break;

        case Formattable::kObject:
            // This will never happen because MessageFormat doesn't
            // support kObject.  When MessageFormat is changed to
            // understand MeasureFormats, modify this code to do the
            // right thing. [alan]
            UPRV_UNREACHABLE;

        // better not happen!
        case Formattable::kArray:
            UPRV_UNREACHABLE;
        }
    }

    // clean up
    delete [] args;
}

#define SINGLE_QUOTE      ((UChar)0x0027)
#define CURLY_BRACE_LEFT  ((UChar)0x007B)
#define CURLY_BRACE_RIGHT ((UChar)0x007D)

#define STATE_INITIAL 0
#define STATE_SINGLE_QUOTE 1
#define STATE_IN_QUOTE 2
#define STATE_MSG_ELEMENT 3

#define MAppend(c) if (len < destCapacity) dest[len++] = c; else len++

int32_t umsg_autoQuoteApostrophe(const UChar* pattern, 
                 int32_t patternLength,
                 UChar* dest,
                 int32_t destCapacity,
                 UErrorCode* ec)
{
    int32_t state = STATE_INITIAL;
    int32_t braceCount = 0;
    int32_t len = 0;

    if (ec == NULL || U_FAILURE(*ec)) {
        return -1;
    }

    if (pattern == NULL || patternLength < -1 || (dest == NULL && destCapacity > 0)) {
        *ec = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }
    U_ASSERT(destCapacity >= 0);

    if (patternLength == -1) {
        patternLength = u_strlen(pattern);
    }

    for (int i = 0; i < patternLength; ++i) {
        UChar c = pattern[i];
        switch (state) {
        case STATE_INITIAL:
            switch (c) {
            case SINGLE_QUOTE:
                state = STATE_SINGLE_QUOTE;
                break;
            case CURLY_BRACE_LEFT:
                state = STATE_MSG_ELEMENT;
                ++braceCount;
                break;
            }
            break;

        case STATE_SINGLE_QUOTE:
            switch (c) {
            case SINGLE_QUOTE:
                state = STATE_INITIAL;
                break;
            case CURLY_BRACE_LEFT:
            case CURLY_BRACE_RIGHT:
                state = STATE_IN_QUOTE;
                break;
            default:
                MAppend(SINGLE_QUOTE);
                state = STATE_INITIAL;
                break;
            }
        break;

        case STATE_IN_QUOTE:
            switch (c) {
            case SINGLE_QUOTE:
                state = STATE_INITIAL;
                break;
            }
            break;

        case STATE_MSG_ELEMENT:
            switch (c) {
            case CURLY_BRACE_LEFT:
                ++braceCount;
                break;
            case CURLY_BRACE_RIGHT:
                if (--braceCount == 0) {
                    state = STATE_INITIAL;
                }
                break;
            }
            break;

        default: // Never happens.
            break;
        }

        U_ASSERT(len >= 0);
        MAppend(c);
    }

    // End of scan
    if (state == STATE_SINGLE_QUOTE || state == STATE_IN_QUOTE) {
        MAppend(SINGLE_QUOTE);
    }

    return u_terminateUChars(dest, destCapacity, len, ec);
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 1996-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
* Modification History:
*
*   Date        Name        Description
*   06/24/99    helena      Integrated Alan's NF enhancements and Java2 bug fixes
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING



















U_NAMESPACE_USE


U_CAPI UNumberFormat* U_EXPORT2
unum_open(  UNumberFormatStyle    style,  
            const    UChar*    pattern,
            int32_t            patternLength,
            const    char*     locale,
            UParseError*       parseErr,
            UErrorCode*        status) {
    if(U_FAILURE(*status)) {
        return NULL;
    }

    NumberFormat *retVal = NULL;

    switch(style) {
    case UNUM_DECIMAL:
    case UNUM_CURRENCY:
    case UNUM_PERCENT:
    case UNUM_SCIENTIFIC:
    case UNUM_CURRENCY_ISO:
    case UNUM_CURRENCY_PLURAL:
    case UNUM_CURRENCY_ACCOUNTING:
    case UNUM_CASH_CURRENCY:
    case UNUM_CURRENCY_STANDARD:
        retVal = NumberFormat::createInstance(Locale(locale), style, *status);
        break;

    case UNUM_PATTERN_DECIMAL: {
        UParseError tErr;
        /* UnicodeString can handle the case when patternLength = -1. */
        const UnicodeString pat(pattern, patternLength);

        if(parseErr==NULL){
            parseErr = &tErr;
        }

        DecimalFormatSymbols *syms = new DecimalFormatSymbols(Locale(locale), *status);
        if(syms == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        if (U_FAILURE(*status)) {
            delete syms;
            return NULL;
        }

        retVal = new DecimalFormat(pat, syms, *parseErr, *status);
        if(retVal == NULL) {
            delete syms;
        }
    } break;

#if U_HAVE_RBNF
    case UNUM_PATTERN_RULEBASED: {
        UParseError tErr;
        /* UnicodeString can handle the case when patternLength = -1. */
        const UnicodeString pat(pattern, patternLength);
        
        if(parseErr==NULL){
            parseErr = &tErr;
        }
        
        retVal = new RuleBasedNumberFormat(pat, Locale(locale), *parseErr, *status);
    } break;

    case UNUM_SPELLOUT:
        retVal = new RuleBasedNumberFormat(URBNF_SPELLOUT, Locale(locale), *status);
        break;

    case UNUM_ORDINAL:
        retVal = new RuleBasedNumberFormat(URBNF_ORDINAL, Locale(locale), *status);
        break;

    case UNUM_DURATION:
        retVal = new RuleBasedNumberFormat(URBNF_DURATION, Locale(locale), *status);
        break;

    case UNUM_NUMBERING_SYSTEM:
        retVal = new RuleBasedNumberFormat(URBNF_NUMBERING_SYSTEM, Locale(locale), *status);
        break;
#endif

    case UNUM_DECIMAL_COMPACT_SHORT:
        retVal = CompactDecimalFormat::createInstance(Locale(locale), UNUM_SHORT, *status);
        break;

    case UNUM_DECIMAL_COMPACT_LONG:
        retVal = CompactDecimalFormat::createInstance(Locale(locale), UNUM_LONG, *status);
        break;

    default:
        *status = U_UNSUPPORTED_ERROR;
        return NULL;
    }

    if(retVal == NULL && U_SUCCESS(*status)) {
        *status = U_MEMORY_ALLOCATION_ERROR;
    }

    if (U_FAILURE(*status) && retVal != NULL) {
        delete retVal;
        retVal = NULL;
    }

    return reinterpret_cast<UNumberFormat *>(retVal);
}

U_CAPI void U_EXPORT2
unum_close(UNumberFormat* fmt)
{
    delete (NumberFormat*) fmt;
}

U_CAPI UNumberFormat* U_EXPORT2
unum_clone(const UNumberFormat *fmt,
       UErrorCode *status)
{
    if(U_FAILURE(*status))
        return 0;
    
    Format *res = 0;
    const NumberFormat* nf = reinterpret_cast<const NumberFormat*>(fmt);
    const DecimalFormat* df = dynamic_cast<const DecimalFormat*>(nf);
    if (df != NULL) {
        res = df->clone();
    } else {
        const RuleBasedNumberFormat* rbnf = dynamic_cast<const RuleBasedNumberFormat*>(nf);
        U_ASSERT(rbnf != NULL);
        res = rbnf->clone();
    }

    if(res == 0) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return 0;
    }
    
    return (UNumberFormat*) res;
}

U_CAPI int32_t U_EXPORT2
unum_format(    const    UNumberFormat*    fmt,
        int32_t           number,
        UChar*            result,
        int32_t           resultLength,
        UFieldPosition    *pos,
        UErrorCode*       status)
{
        return unum_formatInt64(fmt, number, result, resultLength, pos, status);
}

U_CAPI int32_t U_EXPORT2
unum_formatInt64(const UNumberFormat* fmt,
        int64_t         number,
        UChar*          result,
        int32_t         resultLength,
        UFieldPosition *pos,
        UErrorCode*     status)
{
    if(U_FAILURE(*status))
        return -1;
    
    UnicodeString res;
    if(!(result==NULL && resultLength==0)) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        res.setTo(result, 0, resultLength);
    }
    
    FieldPosition fp;
    
    if(pos != 0)
        fp.setField(pos->field);
    
    ((const NumberFormat*)fmt)->format(number, res, fp, *status);

    if(pos != 0) {
        pos->beginIndex = fp.getBeginIndex();
        pos->endIndex = fp.getEndIndex();
    }
    
    return res.extract(result, resultLength, *status);
}

U_CAPI int32_t U_EXPORT2
unum_formatDouble(    const    UNumberFormat*  fmt,
            double          number,
            UChar*          result,
            int32_t         resultLength,
            UFieldPosition  *pos, /* 0 if ignore */
            UErrorCode*     status)
{
 
  if(U_FAILURE(*status)) return -1;

  UnicodeString res;
  if(!(result==NULL && resultLength==0)) {
    // NULL destination for pure preflighting: empty dummy string
    // otherwise, alias the destination buffer
    res.setTo(result, 0, resultLength);
  }

  FieldPosition fp;
  
  if(pos != 0)
    fp.setField(pos->field);
  
  ((const NumberFormat*)fmt)->format(number, res, fp, *status);
  
  if(pos != 0) {
    pos->beginIndex = fp.getBeginIndex();
    pos->endIndex = fp.getEndIndex();
  }
  
  return res.extract(result, resultLength, *status);
}

U_CAPI int32_t U_EXPORT2
unum_formatDoubleForFields(const UNumberFormat* format,
                           double number,
                           UChar* result,
                           int32_t resultLength,
                           UFieldPositionIterator* fpositer,
                           UErrorCode* status)
{
    if (U_FAILURE(*status))
        return -1;

    if (result == NULL ? resultLength != 0 : resultLength < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    UnicodeString res;
    if (result != NULL) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        res.setTo(result, 0, resultLength);
    }

    ((const NumberFormat*)format)->format(number, res, (FieldPositionIterator*)fpositer, *status);

    return res.extract(result, resultLength, *status);
}

U_CAPI int32_t U_EXPORT2 
unum_formatDecimal(const    UNumberFormat*  fmt,
            const char *    number,
            int32_t         length,
            UChar*          result,
            int32_t         resultLength,
            UFieldPosition  *pos, /* 0 if ignore */
            UErrorCode*     status) {

    if(U_FAILURE(*status)) {
        return -1;
    }
    if ((result == NULL && resultLength != 0) || resultLength < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }

    FieldPosition fp;
    if(pos != 0) {
        fp.setField(pos->field);
    }

    if (length < 0) {
        length = static_cast<int32_t>(uprv_strlen(number));
    }
    StringPiece numSP(number, length);
    Formattable numFmtbl(numSP, *status);

    UnicodeString resultStr;
    if (resultLength > 0) {
        // Alias the destination buffer.
        resultStr.setTo(result, 0, resultLength);
    }
    ((const NumberFormat*)fmt)->format(numFmtbl, resultStr, fp, *status);
    if(pos != 0) {
        pos->beginIndex = fp.getBeginIndex();
        pos->endIndex = fp.getEndIndex();
    }
    return resultStr.extract(result, resultLength, *status);
}




U_CAPI int32_t U_EXPORT2 
unum_formatDoubleCurrency(const UNumberFormat* fmt,
                          double number,
                          UChar* currency,
                          UChar* result,
                          int32_t resultLength,
                          UFieldPosition* pos, /* ignored if 0 */
                          UErrorCode* status) {
    if (U_FAILURE(*status)) return -1;

    UnicodeString res;
    if (!(result==NULL && resultLength==0)) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        res.setTo(result, 0, resultLength);
    }
    
    FieldPosition fp;
    if (pos != 0) {
        fp.setField(pos->field);
    }
    CurrencyAmount *tempCurrAmnt = new CurrencyAmount(number, currency, *status);
    // Check for null pointer.
    if (tempCurrAmnt == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return -1;
    }
    Formattable n(tempCurrAmnt);
    ((const NumberFormat*)fmt)->format(n, res, fp, *status);
    
    if (pos != 0) {
        pos->beginIndex = fp.getBeginIndex();
        pos->endIndex = fp.getEndIndex();
    }
  
    return res.extract(result, resultLength, *status);
}

static void
parseRes(Formattable& res,
         const   UNumberFormat*  fmt,
         const   UChar*          text,
         int32_t         textLength,
         int32_t         *parsePos /* 0 = start */,
         UErrorCode      *status)
{
    if(U_FAILURE(*status))
        return;
    
    const UnicodeString src((UBool)(textLength == -1), text, textLength);
    ParsePosition pp;
    
    if(parsePos != 0)
        pp.setIndex(*parsePos);
    
    ((const NumberFormat*)fmt)->parse(src, res, pp);
    
    if(pp.getErrorIndex() != -1) {
        *status = U_PARSE_ERROR;
        if(parsePos != 0) {
            *parsePos = pp.getErrorIndex();
        }
    } else if(parsePos != 0) {
        *parsePos = pp.getIndex();
    }
}

U_CAPI int32_t U_EXPORT2
unum_parse(    const   UNumberFormat*  fmt,
        const   UChar*          text,
        int32_t         textLength,
        int32_t         *parsePos /* 0 = start */,
        UErrorCode      *status)
{
    Formattable res;
    parseRes(res, fmt, text, textLength, parsePos, status);
    return res.getLong(*status);
}

U_CAPI int64_t U_EXPORT2
unum_parseInt64(    const   UNumberFormat*  fmt,
        const   UChar*          text,
        int32_t         textLength,
        int32_t         *parsePos /* 0 = start */,
        UErrorCode      *status)
{
    Formattable res;
    parseRes(res, fmt, text, textLength, parsePos, status);
    return res.getInt64(*status);
}

U_CAPI double U_EXPORT2
unum_parseDouble(    const   UNumberFormat*  fmt,
            const   UChar*          text,
            int32_t         textLength,
            int32_t         *parsePos /* 0 = start */,
            UErrorCode      *status)
{
    Formattable res;
    parseRes(res, fmt, text, textLength, parsePos, status);
    return res.getDouble(*status);
}

U_CAPI int32_t U_EXPORT2
unum_parseDecimal(const UNumberFormat*  fmt,
            const UChar*    text,
            int32_t         textLength,
            int32_t         *parsePos /* 0 = start */,
            char            *outBuf,
            int32_t         outBufLength,
            UErrorCode      *status)
{
    if (U_FAILURE(*status)) {
        return -1;
    }
    if ((outBuf == NULL && outBufLength != 0) || outBufLength < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return -1;
    }
    Formattable res;
    parseRes(res, fmt, text, textLength, parsePos, status);
    StringPiece sp = res.getDecimalNumber(*status);
    if (U_FAILURE(*status)) {
       return -1;
    } else if (sp.size() > outBufLength) {
        *status = U_BUFFER_OVERFLOW_ERROR;
    } else if (sp.size() == outBufLength) {
        uprv_strncpy(outBuf, sp.data(), sp.size());
        *status = U_STRING_NOT_TERMINATED_WARNING;
    } else {
        U_ASSERT(outBufLength > 0);
        uprv_strcpy(outBuf, sp.data());
    }
    return sp.size();
}

U_CAPI double U_EXPORT2
unum_parseDoubleCurrency(const UNumberFormat* fmt,
                         const UChar* text,
                         int32_t textLength,
                         int32_t* parsePos, /* 0 = start */
                         UChar* currency,
                         UErrorCode* status) {
    double doubleVal = 0.0;
    currency[0] = 0;
    if (U_FAILURE(*status)) {
        return doubleVal;
    }
    const UnicodeString src((UBool)(textLength == -1), text, textLength);
    ParsePosition pp;
    if (parsePos != NULL) {
        pp.setIndex(*parsePos);
    }
    *status = U_PARSE_ERROR; // assume failure, reset if succeed
    LocalPointer<CurrencyAmount> currAmt(((const NumberFormat*)fmt)->parseCurrency(src, pp));
    if (pp.getErrorIndex() != -1) {
        if (parsePos != NULL) {
            *parsePos = pp.getErrorIndex();
        }
    } else {
        if (parsePos != NULL) {
            *parsePos = pp.getIndex();
        }
        if (pp.getIndex() > 0) {
            *status = U_ZERO_ERROR;
            u_strcpy(currency, currAmt->getISOCurrency());
            doubleVal = currAmt->getNumber().getDouble(*status);
        }
    }
    return doubleVal;
}

U_CAPI const char* U_EXPORT2
unum_getAvailable(int32_t index)
{
    return uloc_getAvailable(index);
}

U_CAPI int32_t U_EXPORT2
unum_countAvailable()
{
    return uloc_countAvailable();
}

U_CAPI int32_t U_EXPORT2
unum_getAttribute(const UNumberFormat*          fmt,
          UNumberFormatAttribute  attr)
{
    const NumberFormat* nf = reinterpret_cast<const NumberFormat*>(fmt);
    if (attr == UNUM_LENIENT_PARSE) {
        // Supported for all subclasses
        return nf->isLenient();
    }
    else if (attr == UNUM_MAX_INTEGER_DIGITS) {
        return nf->getMaximumIntegerDigits();
    }
    else if (attr == UNUM_MIN_INTEGER_DIGITS) {
        return nf->getMinimumIntegerDigits();
    }
    else if (attr == UNUM_INTEGER_DIGITS) {
        // TODO: what should this return?
        return nf->getMinimumIntegerDigits();
    }
    else if (attr == UNUM_MAX_FRACTION_DIGITS) {
        return nf->getMaximumFractionDigits();
    }
    else if (attr == UNUM_MIN_FRACTION_DIGITS) {
        return nf->getMinimumFractionDigits();
    }
    else if (attr == UNUM_FRACTION_DIGITS) {
        // TODO: what should this return?
        return nf->getMinimumFractionDigits();
    }
    else if (attr == UNUM_ROUNDING_MODE) {
        return nf->getRoundingMode();
    }

    // The remaining attributes are only supported for DecimalFormat
    const DecimalFormat* df = dynamic_cast<const DecimalFormat*>(nf);
    if (df != NULL) {
        UErrorCode ignoredStatus = U_ZERO_ERROR;
        return df->getAttribute(attr, ignoredStatus);
    }

    return -1;
}

U_CAPI void U_EXPORT2
unum_setAttribute(    UNumberFormat*          fmt,
            UNumberFormatAttribute  attr,
            int32_t                 newValue)
{
    NumberFormat* nf = reinterpret_cast<NumberFormat*>(fmt);
    if (attr == UNUM_LENIENT_PARSE) {
        // Supported for all subclasses
        // keep this here as the class may not be a DecimalFormat
        return nf->setLenient(newValue != 0);
    }
    else if (attr == UNUM_MAX_INTEGER_DIGITS) {
        return nf->setMaximumIntegerDigits(newValue);
    }
    else if (attr == UNUM_MIN_INTEGER_DIGITS) {
        return nf->setMinimumIntegerDigits(newValue);
    }
    else if (attr == UNUM_INTEGER_DIGITS) {
        nf->setMinimumIntegerDigits(newValue);
        return nf->setMaximumIntegerDigits(newValue);
    }
    else if (attr == UNUM_MAX_FRACTION_DIGITS) {
        return nf->setMaximumFractionDigits(newValue);
    }
    else if (attr == UNUM_MIN_FRACTION_DIGITS) {
        return nf->setMinimumFractionDigits(newValue);
    }
    else if (attr == UNUM_FRACTION_DIGITS) {
        nf->setMinimumFractionDigits(newValue);
        return nf->setMaximumFractionDigits(newValue);
    }
    else if (attr == UNUM_ROUNDING_MODE) {
        return nf->setRoundingMode((NumberFormat::ERoundingMode)newValue);
    }

    // The remaining attributes are only supported for DecimalFormat
    DecimalFormat* df = dynamic_cast<DecimalFormat*>(nf);
    if (df != NULL) {
        UErrorCode ignoredStatus = U_ZERO_ERROR;
        df->setAttribute(attr, newValue, ignoredStatus);
    }
}

U_CAPI double U_EXPORT2
unum_getDoubleAttribute(const UNumberFormat*          fmt,
          UNumberFormatAttribute  attr)
{
    const NumberFormat* nf = reinterpret_cast<const NumberFormat*>(fmt);
    const DecimalFormat* df = dynamic_cast<const DecimalFormat*>(nf);
    if (df != NULL &&  attr == UNUM_ROUNDING_INCREMENT) {
        return df->getRoundingIncrement();
    } else {
        return -1.0;
    }
}

U_CAPI void U_EXPORT2
unum_setDoubleAttribute(    UNumberFormat*          fmt,
            UNumberFormatAttribute  attr,
            double                 newValue)
{
    NumberFormat* nf = reinterpret_cast<NumberFormat*>(fmt);
    DecimalFormat* df = dynamic_cast<DecimalFormat*>(nf);
    if (df != NULL && attr == UNUM_ROUNDING_INCREMENT) {   
        df->setRoundingIncrement(newValue);
    }
}

U_CAPI int32_t U_EXPORT2
unum_getTextAttribute(const UNumberFormat*  fmt,
            UNumberFormatTextAttribute      tag,
            UChar*                          result,
            int32_t                         resultLength,
            UErrorCode*                     status)
{
    if(U_FAILURE(*status))
        return -1;

    UnicodeString res;
    if(!(result==NULL && resultLength==0)) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        res.setTo(result, 0, resultLength);
    }

    const NumberFormat* nf = reinterpret_cast<const NumberFormat*>(fmt);
    const DecimalFormat* df = dynamic_cast<const DecimalFormat*>(nf);
    if (df != NULL) {
        switch(tag) {
        case UNUM_POSITIVE_PREFIX:
            df->getPositivePrefix(res);
            break;

        case UNUM_POSITIVE_SUFFIX:
            df->getPositiveSuffix(res);
            break;

        case UNUM_NEGATIVE_PREFIX:
            df->getNegativePrefix(res);
            break;

        case UNUM_NEGATIVE_SUFFIX:
            df->getNegativeSuffix(res);
            break;

        case UNUM_PADDING_CHARACTER:
            res = df->getPadCharacterString();
            break;

        case UNUM_CURRENCY_CODE:
            res = UnicodeString(df->getCurrency());
            break;

        default:
            *status = U_UNSUPPORTED_ERROR;
            return -1;
        }
    } else {
        const RuleBasedNumberFormat* rbnf = dynamic_cast<const RuleBasedNumberFormat*>(nf);
        U_ASSERT(rbnf != NULL);
        if (tag == UNUM_DEFAULT_RULESET) {
            res = rbnf->getDefaultRuleSetName();
        } else if (tag == UNUM_PUBLIC_RULESETS) {
            int32_t count = rbnf->getNumberOfRuleSetNames();
            for (int i = 0; i < count; ++i) {
                res += rbnf->getRuleSetName(i);
                res += (UChar)0x003b; // semicolon
            }
        } else {
            *status = U_UNSUPPORTED_ERROR;
            return -1;
        }
    }

    return res.extract(result, resultLength, *status);
}

U_CAPI void U_EXPORT2
unum_setTextAttribute(    UNumberFormat*                    fmt,
            UNumberFormatTextAttribute      tag,
            const    UChar*                            newValue,
            int32_t                            newValueLength,
            UErrorCode                        *status)
{
    if(U_FAILURE(*status))
        return;

    UnicodeString val(newValue, newValueLength);
    NumberFormat* nf = reinterpret_cast<NumberFormat*>(fmt);
    DecimalFormat* df = dynamic_cast<DecimalFormat*>(nf);
    if (df != NULL) {
      switch(tag) {
      case UNUM_POSITIVE_PREFIX:
        df->setPositivePrefix(val);
        break;
        
      case UNUM_POSITIVE_SUFFIX:
        df->setPositiveSuffix(val);
        break;
        
      case UNUM_NEGATIVE_PREFIX:
        df->setNegativePrefix(val);
        break;
        
      case UNUM_NEGATIVE_SUFFIX:
        df->setNegativeSuffix(val);
        break;
        
      case UNUM_PADDING_CHARACTER:
        df->setPadCharacter(val);
        break;
        
      case UNUM_CURRENCY_CODE:
        df->setCurrency(val.getTerminatedBuffer(), *status);
        break;
        
      default:
        *status = U_UNSUPPORTED_ERROR;
        break;
      }
    } else {
      RuleBasedNumberFormat* rbnf = dynamic_cast<RuleBasedNumberFormat*>(nf);
      U_ASSERT(rbnf != NULL);
      if (tag == UNUM_DEFAULT_RULESET) {
        rbnf->setDefaultRuleSet(val, *status);
      } else {
        *status = U_UNSUPPORTED_ERROR;
      }
    }
}

U_CAPI int32_t U_EXPORT2
unum_toPattern(    const    UNumberFormat*          fmt,
        UBool                  isPatternLocalized,
        UChar*                  result,
        int32_t                 resultLength,
        UErrorCode*             status)
{
    if(U_FAILURE(*status))
        return -1;
    
    UnicodeString pat;
    if(!(result==NULL && resultLength==0)) {
        // NULL destination for pure preflighting: empty dummy string
        // otherwise, alias the destination buffer
        pat.setTo(result, 0, resultLength);
    }

    const NumberFormat* nf = reinterpret_cast<const NumberFormat*>(fmt);
    const DecimalFormat* df = dynamic_cast<const DecimalFormat*>(nf);
    if (df != NULL) {
      if(isPatternLocalized)
        df->toLocalizedPattern(pat);
      else
        df->toPattern(pat);
    } else {
      const RuleBasedNumberFormat* rbnf = dynamic_cast<const RuleBasedNumberFormat*>(nf);
      U_ASSERT(rbnf != NULL);
      pat = rbnf->getRules();
    }
    return pat.extract(result, resultLength, *status);
}

U_CAPI int32_t U_EXPORT2
unum_getSymbol(const UNumberFormat *fmt,
               UNumberFormatSymbol symbol,
               UChar *buffer,
               int32_t size,
               UErrorCode *status)
{
    if(status==NULL || U_FAILURE(*status)) {
        return 0;
    }
    if(fmt==NULL || symbol< 0 || symbol>=UNUM_FORMAT_SYMBOL_COUNT) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    const NumberFormat *nf = reinterpret_cast<const NumberFormat *>(fmt);
    const DecimalFormat *dcf = dynamic_cast<const DecimalFormat *>(nf);
    if (dcf == NULL) {
      *status = U_UNSUPPORTED_ERROR;
      return 0;
    }

    return dcf->
      getDecimalFormatSymbols()->
        getConstSymbol((DecimalFormatSymbols::ENumberFormatSymbol)symbol).
          extract(buffer, size, *status);
}

U_CAPI void U_EXPORT2
unum_setSymbol(UNumberFormat *fmt,
               UNumberFormatSymbol symbol,
               const UChar *value,
               int32_t length,
               UErrorCode *status)
{
    if(status==NULL || U_FAILURE(*status)) {
        return;
    }
    if(fmt==NULL || symbol< 0 || symbol>=UNUM_FORMAT_SYMBOL_COUNT || value==NULL || length<-1) {
        *status=U_ILLEGAL_ARGUMENT_ERROR;
        return;
    }
    NumberFormat *nf = reinterpret_cast<NumberFormat *>(fmt);
    DecimalFormat *dcf = dynamic_cast<DecimalFormat *>(nf);
    if (dcf == NULL) {
      *status = U_UNSUPPORTED_ERROR;
      return;
    }

    DecimalFormatSymbols symbols(*dcf->getDecimalFormatSymbols());
    symbols.setSymbol((DecimalFormatSymbols::ENumberFormatSymbol)symbol,
        UnicodeString(value, length));  /* UnicodeString can handle the case when length = -1. */
    dcf->setDecimalFormatSymbols(symbols);
}

U_CAPI void U_EXPORT2
unum_applyPattern(  UNumberFormat  *fmt,
                    UBool          localized,
                    const UChar    *pattern,
                    int32_t        patternLength,
                    UParseError    *parseError,
                    UErrorCode*    status)
{
    UErrorCode tStatus = U_ZERO_ERROR;
    UParseError tParseError;
    
    if(parseError == NULL){
        parseError = &tParseError;
    }
    
    if(status==NULL){
        status = &tStatus;
    }
    
    int32_t len = (patternLength == -1 ? u_strlen(pattern) : patternLength);
    const UnicodeString pat((UChar*)pattern, len, len);

    // Verify if the object passed is a DecimalFormat object
    NumberFormat* nf = reinterpret_cast<NumberFormat*>(fmt);
    DecimalFormat* df = dynamic_cast<DecimalFormat*>(nf);
    if (df != NULL) {
      if(localized) {
        df->applyLocalizedPattern(pat,*parseError, *status);
      } else {
        df->applyPattern(pat,*parseError, *status);
      }
    } else {
      *status = U_UNSUPPORTED_ERROR;
      return;
    }
}

U_CAPI const char* U_EXPORT2
unum_getLocaleByType(const UNumberFormat *fmt,
                     ULocDataLocaleType type,
                     UErrorCode* status)
{
    if (fmt == NULL) {
        if (U_SUCCESS(*status)) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
        }
        return NULL;
    }
    return ((const Format*)fmt)->getLocaleID(type, *status);
}

U_CAPI void U_EXPORT2
unum_setContext(UNumberFormat* fmt, UDisplayContext value, UErrorCode* status)
{
    if (U_FAILURE(*status)) {
        return;
    }
    ((NumberFormat*)fmt)->setContext(value, *status);
    return;
}

U_CAPI UDisplayContext U_EXPORT2
unum_getContext(const UNumberFormat *fmt, UDisplayContextType type, UErrorCode* status)
{
    if (U_FAILURE(*status)) {
        return (UDisplayContext)0;
    }
    return ((const NumberFormat*)fmt)->getContext(type, *status);
}

U_INTERNAL UFormattable * U_EXPORT2
unum_parseToUFormattable(const UNumberFormat* fmt,
                         UFormattable *result,
                         const UChar* text,
                         int32_t textLength,
                         int32_t* parsePos, /* 0 = start */
                         UErrorCode* status) {
  UFormattable *newFormattable = NULL;
  if (U_FAILURE(*status)) return result;
  if (fmt == NULL || (text==NULL && textLength!=0)) {
    *status = U_ILLEGAL_ARGUMENT_ERROR;
    return result;
  }
  if (result == NULL) { // allocate if not allocated.
    newFormattable = result = ufmt_open(status);
  }
  parseRes(*(Formattable::fromUFormattable(result)), fmt, text, textLength, parsePos, status);
  if (U_FAILURE(*status) && newFormattable != NULL) {
    ufmt_close(newFormattable);
    result = NULL; // deallocate if there was a parse error
  }
  return result;
}

U_INTERNAL int32_t U_EXPORT2
unum_formatUFormattable(const UNumberFormat* fmt,
                        const UFormattable *number,
                        UChar *result,
                        int32_t resultLength,
                        UFieldPosition *pos, /* ignored if 0 */
                        UErrorCode *status) {
    if (U_FAILURE(*status)) {
      return 0;
    }
    if (fmt == NULL || number==NULL ||
        (result==NULL ? resultLength!=0 : resultLength<0)) {
      *status = U_ILLEGAL_ARGUMENT_ERROR;
      return 0;
    }
    UnicodeString res(result, 0, resultLength);

    FieldPosition fp;

    if(pos != 0)
        fp.setField(pos->field);

    ((const NumberFormat*)fmt)->format(*(Formattable::fromUFormattable(number)), res, fp, *status);

    if(pos != 0) {
        pos->beginIndex = fp.getBeginIndex();
        pos->endIndex = fp.getEndIndex();
    }

    return res.extract(result, resultLength, *status);
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2013, International Business Machines Corporation and others.
* All Rights Reserved.
*****************************************************************************************
*/



#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2013-2014, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/




#if !UCONFIG_NO_FORMATTING




/**
 * \file
 * \brief C API: UNumberingSystem, information about numbering systems
 *
 * Defines numbering systems. A numbering system describes the scheme by which 
 * numbers are to be presented to the end user. In its simplest form, a numbering
 * system describes the set of digit characters that are to be used to display
 * numbers, such as Western digits, Thai digits, Arabic-Indic digits, etc., in a
 * positional numbering system with a specified radix (typically 10).
 * More complicated numbering systems are algorithmic in nature, and require use
 * of an RBNF formatter (rule based number formatter), in order to calculate
 * the characters to be displayed for a given number. Examples of algorithmic
 * numbering systems include Roman numerals, Chinese numerals, and Hebrew numerals.
 * Formatting rules for many commonly used numbering systems are included in
 * the ICU package, based on the numbering system rules defined in CLDR.
 * Alternate numbering systems can be specified to a locale by using the
 * numbers locale keyword.
 */

/**
 * Opaque UNumberingSystem object for use in C programs.
 * @stable ICU 52
 */
struct UNumberingSystem;
typedef struct UNumberingSystem UNumberingSystem;  /**< C typedef for struct UNumberingSystem. @stable ICU 52 */

/**
 * Opens a UNumberingSystem object using the default numbering system for the specified
 * locale.
 * @param locale    The locale for which the default numbering system should be opened.
 * @param status    A pointer to a UErrorCode to receive any errors. For example, this
 *                  may be U_UNSUPPORTED_ERROR for a locale such as "en@numbers=xyz" that
 *                  specifies a numbering system unknown to ICU.
 * @return          A UNumberingSystem for the specified locale, or NULL if an error
 *                  occurred.
 * @stable ICU 52
 */
U_STABLE UNumberingSystem * U_EXPORT2
unumsys_open(const char *locale, UErrorCode *status);

/**
 * Opens a UNumberingSystem object using the name of one of the predefined numbering
 * systems specified by CLDR and known to ICU, such as "latn", "arabext", or "hanidec";
 * the full list is returned by unumsys_openAvailableNames. Note that some of the names
 * listed at http://unicode.org/repos/cldr/tags/latest/common/bcp47/number.xml - e.g.
 * default, native, traditional, finance - do not identify specific numbering systems,
 * but rather key values that may only be used as part of a locale, which in turn
 * defines how they are mapped to a specific numbering system such as "latn" or "hant".
 *
 * @param name      The name of the numbering system for which a UNumberingSystem object
 *                  should be opened.
 * @param status    A pointer to a UErrorCode to receive any errors. For example, this
 *                  may be U_UNSUPPORTED_ERROR for a numbering system such as "xyz" that
 *                  is unknown to ICU.
 * @return          A UNumberingSystem for the specified name, or NULL if an error
 *                  occurred.
 * @stable ICU 52
 */
U_STABLE UNumberingSystem * U_EXPORT2
unumsys_openByName(const char *name, UErrorCode *status);

/**
 * Close a UNumberingSystem object. Once closed it may no longer be used.
 * @param unumsys   The UNumberingSystem object to close.
 * @stable ICU 52
 */
U_STABLE void U_EXPORT2
unumsys_close(UNumberingSystem *unumsys);

#if U_SHOW_CPLUSPLUS_API
U_NAMESPACE_BEGIN

/**
 * \class LocalUNumberingSystemPointer
 * "Smart pointer" class, closes a UNumberingSystem via unumsys_close().
 * For most methods see the LocalPointerBase base class.
 * @see LocalPointerBase
 * @see LocalPointer
 * @stable ICU 52
 */
U_DEFINE_LOCAL_OPEN_POINTER(LocalUNumberingSystemPointer, UNumberingSystem, unumsys_close);

U_NAMESPACE_END
#endif

/**
 * Returns an enumeration over the names of all of the predefined numbering systems known
 * to ICU.
 * The numbering system names will be in alphabetical (invariant) order.
 * @param status    A pointer to a UErrorCode to receive any errors.
 * @return          A pointer to a UEnumeration that must be closed with uenum_close(),
 *                  or NULL if an error occurred.
 * @stable ICU 52
 */
U_STABLE UEnumeration * U_EXPORT2
unumsys_openAvailableNames(UErrorCode *status);

/**
 * Returns the name of the specified UNumberingSystem object (if it is one of the
 * predefined names known to ICU).
 * @param unumsys   The UNumberingSystem whose name is desired.
 * @return          A pointer to the name of the specified UNumberingSystem object, or
 *                  NULL if the name is not one of the ICU predefined names. The pointer
 *                  is only valid for the lifetime of the UNumberingSystem object.
 * @stable ICU 52
 */
U_STABLE const char * U_EXPORT2
unumsys_getName(const UNumberingSystem *unumsys);

/**
 * Returns whether the given UNumberingSystem object is for an algorithmic (not purely
 * positional) system.
 * @param unumsys   The UNumberingSystem whose algorithmic status is desired.
 * @return          TRUE if the specified UNumberingSystem object is for an algorithmic
 *                  system.
 * @stable ICU 52
 */
U_STABLE UBool U_EXPORT2
unumsys_isAlgorithmic(const UNumberingSystem *unumsys);

/**
 * Returns the radix of the specified UNumberingSystem object. Simple positional
 * numbering systems typically have radix 10, but might have a radix of e.g. 16 for
 * hexadecimal. The radix is less well-defined for non-positional algorithmic systems.
 * @param unumsys   The UNumberingSystem whose radix is desired.
 * @return          The radix of the specified UNumberingSystem object.
 * @stable ICU 52
 */
U_STABLE int32_t U_EXPORT2
unumsys_getRadix(const UNumberingSystem *unumsys);

/**
 * Get the description string of the specified UNumberingSystem object. For simple
 * positional systems this is the ordered string of digits (with length matching
 * the radix), e.g. "\u3007\u4E00\u4E8C\u4E09\u56DB\u4E94\u516D\u4E03\u516B\u4E5D"
 * for "hanidec"; it would be "0123456789ABCDEF" for hexadecimal. For
 * algorithmic systems this is the name of the RBNF ruleset used for formatting,
 * e.g. "zh/SpelloutRules/%spellout-cardinal" for "hans" or "%greek-upper" for
 * "grek".
 * @param unumsys   The UNumberingSystem whose description string is desired.
 * @param result    A pointer to a buffer to receive the description string.
 * @param resultLength  The maximum size of result.
 * @param status    A pointer to a UErrorCode to receive any errors.
 * @return          The total buffer size needed; if greater than resultLength, the
 *                  output was truncated.
 * @stable ICU 52
 */
U_STABLE int32_t U_EXPORT2
unumsys_getDescription(const UNumberingSystem *unumsys, UChar *result,
                       int32_t resultLength, UErrorCode *status);

#endif /* #if !UCONFIG_NO_FORMATTING */





U_NAMESPACE_USE


U_CAPI UNumberingSystem* U_EXPORT2
unumsys_open(const char *locale, UErrorCode *status)
{
    // createInstance returns immediately if status indicates error
    return (UNumberingSystem*)NumberingSystem::createInstance(Locale(locale), *status);
}


U_CAPI UNumberingSystem* U_EXPORT2
unumsys_openByName(const char *name, UErrorCode *status)
{
    // createInstanceByName does NOT return immediately if status indicates error
    if (U_FAILURE(*status)) {
        return NULL;
    }
    return (UNumberingSystem*)NumberingSystem::createInstanceByName(name, *status);
}


U_CAPI void U_EXPORT2
unumsys_close(UNumberingSystem *unumsys)
{
    delete ((NumberingSystem*)unumsys);
}


U_CAPI UEnumeration* U_EXPORT2
unumsys_openAvailableNames(UErrorCode *status)
{
    // getAvailableNames returns immediately if status indicates error
    return uenum_openFromStringEnumeration(NumberingSystem::getAvailableNames(*status), status);
}


U_CAPI const char * U_EXPORT2
unumsys_getName(const UNumberingSystem *unumsys)
{
    return ((NumberingSystem*)unumsys)->getName();
}


U_CAPI int32_t U_EXPORT2
unumsys_getRadix(const UNumberingSystem *unumsys)
{
    return ((NumberingSystem*)unumsys)->getRadix();
}


U_CAPI UBool U_EXPORT2
unumsys_isAlgorithmic(const UNumberingSystem *unumsys)
{
    return ((NumberingSystem*)unumsys)->isAlgorithmic();
}

U_CAPI int32_t U_EXPORT2
unumsys_getDescription(const UNumberingSystem *unumsys, UChar *result,
                       int32_t resultLength, UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return -1;
    }
    // implement
    UnicodeString descrip = ((NumberingSystem*)unumsys)->getDescription();
    return descrip.extract(result, resultLength, *status);
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2010-2012, International Business Machines
* Corporation and others. All Rights Reserved.
*****************************************************************************************
*/



#if !UCONFIG_NO_FORMATTING











U_NAMESPACE_USE

namespace {

/**
 * Given a number and a format, returns the keyword of the first applicable
 * rule for the PluralRules object.
 * @param rules The plural rules.
 * @param obj The numeric object for which the rule should be determined.
 * @param fmt The NumberFormat specifying how the number will be formatted
 *        (this can affect the plural form, e.g. "1 dollar" vs "1.0 dollars").
 * @param status  Input/output parameter. If at entry this indicates a
 *                failure status, the method returns immediately; otherwise
 *                this is set to indicate the outcome of the call.
 * @return The keyword of the selected rule. Undefined in the case of an error.
 */
UnicodeString select(const PluralRules &rules, const Formattable& obj, const NumberFormat& fmt, UErrorCode& status) {
    if (U_SUCCESS(status)) {
        const DecimalFormat *decFmt = dynamic_cast<const DecimalFormat *>(&fmt);
        if (decFmt != NULL) {
            number::impl::DecimalQuantity dq;
            decFmt->formatToDecimalQuantity(obj, dq, status);
            if (U_SUCCESS(status)) {
                return rules.select(dq);
            }
        } else {
            double number = obj.getDouble(status);
            if (U_SUCCESS(status)) {
                return rules.select(number);
            }
        }
    }
    return UnicodeString();
}

}  // namespace

U_CAPI UPluralRules* U_EXPORT2
uplrules_open(const char *locale, UErrorCode *status)
{
    return uplrules_openForType(locale, UPLURAL_TYPE_CARDINAL, status);
}

U_CAPI UPluralRules* U_EXPORT2
uplrules_openForType(const char *locale, UPluralType type, UErrorCode *status)
{
    return (UPluralRules*)PluralRules::forLocale(Locale(locale), type, *status);
}

U_CAPI void U_EXPORT2
uplrules_close(UPluralRules *uplrules)
{
    delete (PluralRules*)uplrules;
}

U_CAPI int32_t U_EXPORT2
uplrules_select(const UPluralRules *uplrules,
                double number,
                UChar *keyword, int32_t capacity,
                UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return 0;
    }
    if (keyword == NULL ? capacity != 0 : capacity < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    UnicodeString result = ((PluralRules*)uplrules)->select(number);
    return result.extract(keyword, capacity, *status);
}

U_CAPI int32_t U_EXPORT2
uplrules_selectFormatted(const UPluralRules *uplrules,
                const UFormattedNumber* number,
                UChar *keyword, int32_t capacity,
                UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return 0;
    }
    if (keyword == NULL ? capacity != 0 : capacity < 0) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    const number::impl::DecimalQuantity* dq =
        number::impl::validateUFormattedNumberToDecimalQuantity(number, *status);
    if (U_FAILURE(*status)) {
        return 0;
    }
    UnicodeString result = ((PluralRules*)uplrules)->select(*dq);
    return result.extract(keyword, capacity, *status);
}

U_CAPI int32_t U_EXPORT2
uplrules_selectWithFormat(const UPluralRules *uplrules,
                          double number,
                          const UNumberFormat *fmt,
                          UChar *keyword, int32_t capacity,
                          UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return 0;
    }
    const PluralRules* plrules = reinterpret_cast<const PluralRules*>(uplrules);
    const NumberFormat* nf = reinterpret_cast<const NumberFormat*>(fmt);
    if (plrules == NULL || nf == NULL || ((keyword == NULL)? capacity != 0 : capacity < 0)) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }
    Formattable obj(number);
    UnicodeString result = select(*plrules, obj, *nf, *status);
    return result.extract(keyword, capacity, *status);
}

U_CAPI UEnumeration* U_EXPORT2
uplrules_getKeywords(const UPluralRules *uplrules,
                     UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        return NULL;
    }
    const PluralRules* plrules = reinterpret_cast<const PluralRules*>(uplrules);
    if (plrules == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }
    StringEnumeration *senum = plrules->getKeywords(*status);
    if (U_FAILURE(*status)) {
        return NULL;
    }
    if (senum == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    return uenum_openFromStringEnumeration(senum, status);
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*****************************************************************************************
* Copyright (C) 2013-2015, International Business Machines Corporation and others.
* All Rights Reserved.
*****************************************************************************************
*/



#if !UCONFIG_NO_FORMATTING






U_NAMESPACE_USE


U_CAPI const URegion* U_EXPORT2
uregion_getRegionFromCode(const char *regionCode, UErrorCode *status)
{
    return (URegion*)Region::getInstance(regionCode, *status);
}


U_CAPI const URegion* U_EXPORT2
uregion_getRegionFromNumericCode (int32_t code, UErrorCode *status)
{
    return (URegion*)Region::getInstance(code, *status);
}


U_CAPI UEnumeration* U_EXPORT2
uregion_getAvailable(URegionType type, UErrorCode *status)
{
    StringEnumeration* strenum = Region::getAvailable(type, *status);
    return uenum_openFromStringEnumeration( strenum, status );
}


U_CAPI UBool U_EXPORT2
uregion_areEqual(const URegion* uregion, const URegion* otherRegion)
{
    return ( (void*)uregion == (void*)otherRegion );
}


U_CAPI const URegion* U_EXPORT2
uregion_getContainingRegion(const URegion* uregion)
{
    return (URegion*)((Region*)uregion)->getContainingRegion();
}


U_CAPI const URegion* U_EXPORT2
uregion_getContainingRegionOfType(const URegion* uregion, URegionType type)
{
    return (URegion*)((Region*)uregion)->getContainingRegion(type);
}


U_CAPI UEnumeration* U_EXPORT2
uregion_getContainedRegions(const URegion* uregion, UErrorCode *status)
{
    StringEnumeration* strenum = ((Region*)uregion)->getContainedRegions(*status);
    return uenum_openFromStringEnumeration( strenum, status);
}


U_CAPI UEnumeration* U_EXPORT2
uregion_getContainedRegionsOfType(const URegion* uregion, URegionType type, UErrorCode *status)
{
    StringEnumeration* strenum = ((Region*)uregion)->getContainedRegions(type, *status);
    return uenum_openFromStringEnumeration( strenum, status);
}


U_CAPI UBool U_EXPORT2
uregion_contains(const URegion* uregion, const URegion* otherRegion)
{
    return ((Region*)uregion)->contains(*((Region*)otherRegion));
}


U_CAPI UEnumeration* U_EXPORT2
uregion_getPreferredValues(const URegion* uregion, UErrorCode *status)
{
    StringEnumeration* strenum = ((Region*)uregion)->getPreferredValues(*status);
    return uenum_openFromStringEnumeration( strenum, status);
}


U_CAPI const char* U_EXPORT2
uregion_getRegionCode(const URegion* uregion)
{
    return ((Region*)uregion)->getRegionCode();
}


U_CAPI int32_t U_EXPORT2
uregion_getNumericCode(const URegion* uregion)
{
    return ((Region*)uregion)->getNumericCode();
}


U_CAPI URegionType U_EXPORT2
uregion_getType(const URegion* uregion)
{
    return ((Region*)uregion)->getType();
}


#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
**********************************************************************
*   Copyright (C) 2001-2015 IBM and others. All rights reserved.
**********************************************************************
*   Date        Name        Description
*  07/02/2001   synwee      Creation.
**********************************************************************
*/



#if !UCONFIG_NO_COLLATION && !UCONFIG_NO_BREAK_ITERATION

// // © 2016 and later: Unicode, Inc. and others.
// // License & terms of use: http://www.unicode.org/copyright.html
// /*
// **********************************************************************
// *   Copyright (C) 2001-2011,2014 IBM and others. All rights reserved.
// **********************************************************************
// *   Date        Name        Description
// *  06/28/2001   synwee      Creation.
// **********************************************************************
// */
// #ifndef USEARCH_H
// #define USEARCH_H

// #include "unicode/utypes.h"

// #if !UCONFIG_NO_COLLATION && !UCONFIG_NO_BREAK_ITERATION

// #include "unicode/localpointer.h"
// #include "unicode/ucol.h"
// #include "unicode/ucoleitr.h"
// #include "unicode/ubrk.h"

// /**
//  * \file
//  * \brief C API: StringSearch
//  *
//  * C APIs for an engine that provides language-sensitive text searching based
//  * on the comparison rules defined in a <tt>UCollator</tt> data struct,
//  * see <tt>ucol.h</tt>. This ensures that language eccentricity can be
//  * handled, e.g. for the German collator, characters &szlig; and SS will be matched
//  * if case is chosen to be ignored.
//  * See the <a href="http://source.icu-project.org/repos/icu/icuhtml/trunk/design/collation/ICU_collation_design.htm">
//  * "ICU Collation Design Document"</a> for more information.
//  * <p>
//  * The implementation may use a linear search or a modified form of the Boyer-Moore
//  * search; for more information on the latter see
//  * <a href="http://icu-project.org/docs/papers/efficient_text_searching_in_java.html">
//  * "Efficient Text Searching in Java"</a>, published in <i>Java Report</i>
//  * in February, 1999.
//  * <p>
//  * There are 2 match options for selection:<br>
//  * Let S' be the sub-string of a text string S between the offsets start and
//  * end <start, end>.
//  * <br>
//  * A pattern string P matches a text string S at the offsets <start, end>
//  * if
//  * <pre>
//  * option 1. Some canonical equivalent of P matches some canonical equivalent
//  *           of S'
//  * option 2. P matches S' and if P starts or ends with a combining mark,
//  *           there exists no non-ignorable combining mark before or after S'
//  *           in S respectively.
//  * </pre>
//  * Option 2. will be the default.
//  * <p>
//  * This search has APIs similar to that of other text iteration mechanisms
//  * such as the break iterators in <tt>ubrk.h</tt>. Using these
//  * APIs, it is easy to scan through text looking for all occurrences of
//  * a given pattern. This search iterator allows changing of direction by
//  * calling a <tt>reset</tt> followed by a <tt>next</tt> or <tt>previous</tt>.
//  * Though a direction change can occur without calling <tt>reset</tt> first,
//  * this operation comes with some speed penalty.
//  * Generally, match results in the forward direction will match the result
//  * matches in the backwards direction in the reverse order
//  * <p>
//  * <tt>usearch.h</tt> provides APIs to specify the starting position
//  * within the text string to be searched, e.g. <tt>usearch_setOffset</tt>,
//  * <tt>usearch_preceding</tt> and <tt>usearch_following</tt>. Since the
//  * starting position will be set as it is specified, please take note that
//  * there are some dangerous positions which the search may render incorrect
//  * results:
//  * <ul>
//  * <li> The midst of a substring that requires normalization.
//  * <li> If the following match is to be found, the position should not be the
//  *      second character which requires to be swapped with the preceding
//  *      character. Vice versa, if the preceding match is to be found,
//  *      position to search from should not be the first character which
//  *      requires to be swapped with the next character. E.g certain Thai and
//  *      Lao characters require swapping.
//  * <li> If a following pattern match is to be found, any position within a
//  *      contracting sequence except the first will fail. Vice versa if a
//  *      preceding pattern match is to be found, a invalid starting point
//  *      would be any character within a contracting sequence except the last.
//  * </ul>
//  * <p>
//  * A breakiterator can be used if only matches at logical breaks are desired.
//  * Using a breakiterator will only give you results that exactly matches the
//  * boundaries given by the breakiterator. For instance the pattern "e" will
//  * not be found in the string "\u00e9" if a character break iterator is used.
//  * <p>
//  * Options are provided to handle overlapping matches.
//  * E.g. In English, overlapping matches produces the result 0 and 2
//  * for the pattern "abab" in the text "ababab", where else mutually
//  * exclusive matches only produce the result of 0.
//  * <p>
//  * Options are also provided to implement "asymmetric search" as described in
//  * <a href="http://www.unicode.org/reports/tr10/#Asymmetric_Search">
//  * UTS #10 Unicode Collation Algorithm</a>, specifically the USearchAttribute
//  * USEARCH_ELEMENT_COMPARISON and its values.
//  * <p>
//  * Though collator attributes will be taken into consideration while
//  * performing matches, there are no APIs here for setting and getting the
//  * attributes. These attributes can be set by getting the collator
//  * from <tt>usearch_getCollator</tt> and using the APIs in <tt>ucol.h</tt>.
//  * Lastly to update String Search to the new collator attributes,
//  * usearch_reset() has to be called.
//  * <p>
//  * Restriction: <br>
//  * Currently there are no composite characters that consists of a
//  * character with combining class > 0 before a character with combining
//  * class == 0. However, if such a character exists in the future, the
//  * search mechanism does not guarantee the results for option 1.
//  *
//  * <p>
//  * Example of use:<br>
//  * <pre><code>
//  * char *tgtstr = "The quick brown fox jumped over the lazy fox";
//  * char *patstr = "fox";
//  * UChar target[64];
//  * UChar pattern[16];
//  * UErrorCode status = U_ZERO_ERROR;
//  * u_uastrcpy(target, tgtstr);
//  * u_uastrcpy(pattern, patstr);
//  *
//  * UStringSearch *search = usearch_open(pattern, -1, target, -1, "en_US",
//  *                                  NULL, &status);
//  * if (U_SUCCESS(status)) {
//  *     for (int pos = usearch_first(search, &status);
//  *          pos != USEARCH_DONE;
//  *          pos = usearch_next(search, &status))
//  *     {
//  *         printf("Found match at %d pos, length is %d\n", pos,
//  *                                        usearch_getMatchedLength(search));
//  *     }
//  * }
//  *
//  * usearch_close(search);
//  * </code></pre>
//  * @stable ICU 2.4
//  */

// /**
// * DONE is returned by previous() and next() after all valid matches have
// * been returned, and by first() and last() if there are no matches at all.
// * @stable ICU 2.4
// */
// #define USEARCH_DONE -1

// /**
// * Data structure for searching
// * @stable ICU 2.4
// */
// struct UStringSearch;
// /**
// * Data structure for searching
// * @stable ICU 2.4
// */
// typedef struct UStringSearch UStringSearch;

// /**
// * @stable ICU 2.4
// */
// typedef enum {
//     /**
//      * Option for overlapping matches
//      * @stable ICU 2.4
//      */
//     USEARCH_OVERLAP = 0,
// #ifndef U_HIDE_DEPRECATED_API
//     /**
//      * Option for canonical matches; option 1 in header documentation.
//      * The default value will be USEARCH_OFF.
//      * Note: Setting this option to USEARCH_ON currently has no effect on
//      * search behavior, and this option is deprecated. Instead, to control
//      * canonical match behavior, you must set UCOL_NORMALIZATION_MODE
//      * appropriately (to UCOL_OFF or UCOL_ON) in the UCollator used by
//      * the UStringSearch object.
//      * @see usearch_openFromCollator
//      * @see usearch_getCollator
//      * @see usearch_setCollator
//      * @see ucol_getAttribute
//      * @deprecated ICU 53
//      */
//     USEARCH_CANONICAL_MATCH = 1,
// #endif  /* U_HIDE_DEPRECATED_API */
//     /**
//      * Option to control how collation elements are compared.
//      * The default value will be USEARCH_STANDARD_ELEMENT_COMPARISON.
//      * @stable ICU 4.4
//      */
//     USEARCH_ELEMENT_COMPARISON = 2,

// #ifndef U_HIDE_DEPRECATED_API
//     /**
//      * One more than the highest normal USearchAttribute value.
//      * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
//      */
//     USEARCH_ATTRIBUTE_COUNT = 3
// #endif  /* U_HIDE_DEPRECATED_API */
// } USearchAttribute;

// /**
// * @stable ICU 2.4
// */
// typedef enum {
//     /**
//      * Default value for any USearchAttribute
//      * @stable ICU 2.4
//      */
//     USEARCH_DEFAULT = -1,
//     /**
//      * Value for USEARCH_OVERLAP and USEARCH_CANONICAL_MATCH
//      * @stable ICU 2.4
//      */
//     USEARCH_OFF,
//     /**
//      * Value for USEARCH_OVERLAP and USEARCH_CANONICAL_MATCH
//      * @stable ICU 2.4
//      */
//     USEARCH_ON,
//     /**
//      * Value (default) for USEARCH_ELEMENT_COMPARISON;
//      * standard collation element comparison at the specified collator
//      * strength.
//      * @stable ICU 4.4
//      */
//     USEARCH_STANDARD_ELEMENT_COMPARISON,
//     /**
//      * Value for USEARCH_ELEMENT_COMPARISON;
//      * collation element comparison is modified to effectively provide
//      * behavior between the specified strength and strength - 1. Collation
//      * elements in the pattern that have the base weight for the specified
//      * strength are treated as "wildcards" that match an element with any
//      * other weight at that collation level in the searched text. For
//      * example, with a secondary-strength English collator, a plain 'e' in
//      * the pattern will match a plain e or an e with any diacritic in the
//      * searched text, but an e with diacritic in the pattern will only
//      * match an e with the same diacritic in the searched text.
//      *
//      * This supports "asymmetric search" as described in
//      * <a href="http://www.unicode.org/reports/tr10/#Asymmetric_Search">
//      * UTS #10 Unicode Collation Algorithm</a>.
//      *
//      * @stable ICU 4.4
//      */
//     USEARCH_PATTERN_BASE_WEIGHT_IS_WILDCARD,
//     /**
//      * Value for USEARCH_ELEMENT_COMPARISON.
//      * collation element comparison is modified to effectively provide
//      * behavior between the specified strength and strength - 1. Collation
//      * elements in either the pattern or the searched text that have the
//      * base weight for the specified strength are treated as "wildcards"
//      * that match an element with any other weight at that collation level.
//      * For example, with a secondary-strength English collator, a plain 'e'
//      * in the pattern will match a plain e or an e with any diacritic in the
//      * searched text, but an e with diacritic in the pattern will only
//      * match an e with the same diacritic or a plain e in the searched text.
//      *
//      * This option is similar to "asymmetric search" as described in
//      * [UTS #10 Unicode Collation Algorithm](http://www.unicode.org/reports/tr10/#Asymmetric_Search),
//      * but also allows unmarked characters in the searched text to match
//      * marked or unmarked versions of that character in the pattern.
//      *
//      * @stable ICU 4.4
//      */
//     USEARCH_ANY_BASE_WEIGHT_IS_WILDCARD,

// #ifndef U_HIDE_DEPRECATED_API
//     /**
//      * One more than the highest normal USearchAttributeValue value.
//      * @deprecated ICU 58 The numeric value may change over time, see ICU ticket #12420.
//      */
//     USEARCH_ATTRIBUTE_VALUE_COUNT
// #endif  /* U_HIDE_DEPRECATED_API */
// } USearchAttributeValue;

// /* open and close ------------------------------------------------------ */

// /**
// * Creating a search iterator data struct using the argument locale language
// * rule set. A collator will be created in the process, which will be owned by
// * this search and will be deleted in <tt>usearch_close</tt>.
// * @param pattern for matching
// * @param patternlength length of the pattern, -1 for null-termination
// * @param text text string
// * @param textlength length of the text string, -1 for null-termination
// * @param locale name of locale for the rules to be used
// * @param breakiter A BreakIterator that will be used to restrict the points
// *                  at which matches are detected. If a match is found, but
// *                  the match's start or end index is not a boundary as
// *                  determined by the <tt>BreakIterator</tt>, the match will
// *                  be rejected and another will be searched for.
// *                  If this parameter is <tt>NULL</tt>, no break detection is
// *                  attempted.
// * @param status for errors if it occurs. If pattern or text is NULL, or if
// *               patternlength or textlength is 0 then an
// *               U_ILLEGAL_ARGUMENT_ERROR is returned.
// * @return search iterator data structure, or NULL if there is an error.
// * @stable ICU 2.4
// */
// U_STABLE UStringSearch * U_EXPORT2 usearch_open(const UChar          *pattern,
//                                               int32_t         patternlength,
//                                         const UChar          *text,
//                                               int32_t         textlength,
//                                         const char           *locale,
//                                               UBreakIterator *breakiter,
//                                               UErrorCode     *status);

// /**
// * Creating a search iterator data struct using the argument collator language
// * rule set. Note, user retains the ownership of this collator, thus the
// * responsibility of deletion lies with the user.
// * NOTE: string search cannot be instantiated from a collator that has
// * collate digits as numbers (CODAN) turned on.
// * @param pattern for matching
// * @param patternlength length of the pattern, -1 for null-termination
// * @param text text string
// * @param textlength length of the text string, -1 for null-termination
// * @param collator used for the language rules
// * @param breakiter A BreakIterator that will be used to restrict the points
// *                  at which matches are detected. If a match is found, but
// *                  the match's start or end index is not a boundary as
// *                  determined by the <tt>BreakIterator</tt>, the match will
// *                  be rejected and another will be searched for.
// *                  If this parameter is <tt>NULL</tt>, no break detection is
// *                  attempted.
// * @param status for errors if it occurs. If collator, pattern or text is NULL,
// *               or if patternlength or textlength is 0 then an
// *               U_ILLEGAL_ARGUMENT_ERROR is returned.
// * @return search iterator data structure, or NULL if there is an error.
// * @stable ICU 2.4
// */
// U_STABLE UStringSearch * U_EXPORT2 usearch_openFromCollator(
//                                          const UChar *pattern,
//                                                int32_t         patternlength,
//                                          const UChar          *text,
//                                                int32_t         textlength,
//                                          const UCollator      *collator,
//                                                UBreakIterator *breakiter,
//                                                UErrorCode     *status);

// /**
// * Destroying and cleaning up the search iterator data struct.
// * If a collator is created in <tt>usearch_open</tt>, it will be destroyed here.
// * @param searchiter data struct to clean up
// * @stable ICU 2.4
// */
// U_STABLE void U_EXPORT2 usearch_close(UStringSearch *searchiter);

// #if U_SHOW_CPLUSPLUS_API

// U_NAMESPACE_BEGIN

// /**
//  * \class LocalUStringSearchPointer
//  * "Smart pointer" class, closes a UStringSearch via usearch_close().
//  * For most methods see the LocalPointerBase base class.
//  *
//  * @see LocalPointerBase
//  * @see LocalPointer
//  * @stable ICU 4.4
//  */
// U_DEFINE_LOCAL_OPEN_POINTER(LocalUStringSearchPointer, UStringSearch, usearch_close);

// U_NAMESPACE_END

// #endif

// /* get and set methods -------------------------------------------------- */

// /**
// * Sets the current position in the text string which the next search will
// * start from. Clears previous states.
// * This method takes the argument index and sets the position in the text
// * string accordingly without checking if the index is pointing to a
// * valid starting point to begin searching.
// * Search positions that may render incorrect results are highlighted in the
// * header comments
// * @param strsrch search iterator data struct
// * @param position position to start next search from. If position is less
// *          than or greater than the text range for searching,
// *          an U_INDEX_OUTOFBOUNDS_ERROR will be returned
// * @param status error status if any.
// * @stable ICU 2.4
// */
// U_STABLE void U_EXPORT2 usearch_setOffset(UStringSearch *strsrch,
//                                         int32_t    position,
//                                         UErrorCode    *status);

// /**
// * Return the current index in the string text being searched.
// * If the iteration has gone past the end of the text (or past the beginning
// * for a backwards search), <tt>USEARCH_DONE</tt> is returned.
// * @param strsrch search iterator data struct
// * @see #USEARCH_DONE
// * @stable ICU 2.4
// */
// U_STABLE int32_t U_EXPORT2 usearch_getOffset(const UStringSearch *strsrch);

// /**
// * Sets the text searching attributes located in the enum USearchAttribute
// * with values from the enum USearchAttributeValue.
// * <tt>USEARCH_DEFAULT</tt> can be used for all attributes for resetting.
// * @param strsrch search iterator data struct
// * @param attribute text attribute to be set
// * @param value text attribute value
// * @param status for errors if it occurs
// * @see #usearch_getAttribute
// * @stable ICU 2.4
// */
// U_STABLE void U_EXPORT2 usearch_setAttribute(UStringSearch         *strsrch,
//                                            USearchAttribute       attribute,
//                                            USearchAttributeValue  value,
//                                            UErrorCode            *status);

// /**
// * Gets the text searching attributes.
// * @param strsrch search iterator data struct
// * @param attribute text attribute to be retrieve
// * @return text attribute value
// * @see #usearch_setAttribute
// * @stable ICU 2.4
// */
// U_STABLE USearchAttributeValue U_EXPORT2 usearch_getAttribute(
//                                          const UStringSearch    *strsrch,
//                                                USearchAttribute  attribute);

// /**
// * Returns the index to the match in the text string that was searched.
// * This call returns a valid result only after a successful call to
// * <tt>usearch_first</tt>, <tt>usearch_next</tt>, <tt>usearch_previous</tt>,
// * or <tt>usearch_last</tt>.
// * Just after construction, or after a searching method returns
// * <tt>USEARCH_DONE</tt>, this method will return <tt>USEARCH_DONE</tt>.
// * <p>
// * Use <tt>usearch_getMatchedLength</tt> to get the matched string length.
// * @param strsrch search iterator data struct
// * @return index to a substring within the text string that is being
// *         searched.
// * @see #usearch_first
// * @see #usearch_next
// * @see #usearch_previous
// * @see #usearch_last
// * @see #USEARCH_DONE
// * @stable ICU 2.4
// */
// U_STABLE int32_t U_EXPORT2 usearch_getMatchedStart(
//                                                const UStringSearch *strsrch);

// /**
// * Returns the length of text in the string which matches the search pattern.
// * This call returns a valid result only after a successful call to
// * <tt>usearch_first</tt>, <tt>usearch_next</tt>, <tt>usearch_previous</tt>,
// * or <tt>usearch_last</tt>.
// * Just after construction, or after a searching method returns
// * <tt>USEARCH_DONE</tt>, this method will return 0.
// * @param strsrch search iterator data struct
// * @return The length of the match in the string text, or 0 if there is no
// *         match currently.
// * @see #usearch_first
// * @see #usearch_next
// * @see #usearch_previous
// * @see #usearch_last
// * @see #USEARCH_DONE
// * @stable ICU 2.4
// */
// U_STABLE int32_t U_EXPORT2 usearch_getMatchedLength(
//                                                const UStringSearch *strsrch);

// /**
// * Returns the text that was matched by the most recent call to
// * <tt>usearch_first</tt>, <tt>usearch_next</tt>, <tt>usearch_previous</tt>,
// * or <tt>usearch_last</tt>.
// * If the iterator is not pointing at a valid match (e.g. just after
// * construction or after <tt>USEARCH_DONE</tt> has been returned, returns
// * an empty string. If result is not large enough to store the matched text,
// * result will be filled with the partial text and an U_BUFFER_OVERFLOW_ERROR
// * will be returned in status. result will be null-terminated whenever
// * possible. If the buffer fits the matched text exactly, a null-termination
// * is not possible, then a U_STRING_NOT_TERMINATED_ERROR set in status.
// * Pre-flighting can be either done with length = 0 or the API
// * <tt>usearch_getMatchedLength</tt>.
// * @param strsrch search iterator data struct
// * @param result UChar buffer to store the matched string
// * @param resultCapacity length of the result buffer
// * @param status error returned if result is not large enough
// * @return exact length of the matched text, not counting the null-termination
// * @see #usearch_first
// * @see #usearch_next
// * @see #usearch_previous
// * @see #usearch_last
// * @see #USEARCH_DONE
// * @stable ICU 2.4
// */
// U_STABLE int32_t U_EXPORT2 usearch_getMatchedText(const UStringSearch *strsrch,
//                                             UChar         *result,
//                                             int32_t        resultCapacity,
//                                             UErrorCode    *status);

// #if !UCONFIG_NO_BREAK_ITERATION

// /**
// * Set the BreakIterator that will be used to restrict the points at which
// * matches are detected.
// * @param strsrch search iterator data struct
// * @param breakiter A BreakIterator that will be used to restrict the points
// *                  at which matches are detected. If a match is found, but
// *                  the match's start or end index is not a boundary as
// *                  determined by the <tt>BreakIterator</tt>, the match will
// *                  be rejected and another will be searched for.
// *                  If this parameter is <tt>NULL</tt>, no break detection is
// *                  attempted.
// * @param status for errors if it occurs
// * @see #usearch_getBreakIterator
// * @stable ICU 2.4
// */
// U_STABLE void U_EXPORT2 usearch_setBreakIterator(UStringSearch  *strsrch,
//                                                UBreakIterator *breakiter,
//                                                UErrorCode     *status);

// /**
// * Returns the BreakIterator that is used to restrict the points at which
// * matches are detected. This will be the same object that was passed to the
// * constructor or to <tt>usearch_setBreakIterator</tt>. Note that
// * <tt>NULL</tt>
// * is a legal value; it means that break detection should not be attempted.
// * @param strsrch search iterator data struct
// * @return break iterator used
// * @see #usearch_setBreakIterator
// * @stable ICU 2.4
// */
// U_STABLE const UBreakIterator * U_EXPORT2 usearch_getBreakIterator(
//                                               const UStringSearch *strsrch);

// #endif

// /**
// * Set the string text to be searched. Text iteration will hence begin at the
// * start of the text string. This method is useful if you want to re-use an
// * iterator to search for the same pattern within a different body of text.
// * @param strsrch search iterator data struct
// * @param text new string to look for match
// * @param textlength length of the new string, -1 for null-termination
// * @param status for errors if it occurs. If text is NULL, or textlength is 0
// *               then an U_ILLEGAL_ARGUMENT_ERROR is returned with no change
// *               done to strsrch.
// * @see #usearch_getText
// * @stable ICU 2.4
// */
// U_STABLE void U_EXPORT2 usearch_setText(      UStringSearch *strsrch,
//                                       const UChar         *text,
//                                             int32_t        textlength,
//                                             UErrorCode    *status);

// /**
// * Return the string text to be searched.
// * @param strsrch search iterator data struct
// * @param length returned string text length
// * @return string text
// * @see #usearch_setText
// * @stable ICU 2.4
// */
// U_STABLE const UChar * U_EXPORT2 usearch_getText(const UStringSearch *strsrch,
//                                                int32_t       *length);

// /**
// * Gets the collator used for the language rules.
// * <p>
// * Deleting the returned <tt>UCollator</tt> before calling
// * <tt>usearch_close</tt> would cause the string search to fail.
// * <tt>usearch_close</tt> will delete the collator if this search owns it.
// * @param strsrch search iterator data struct
// * @return collator
// * @stable ICU 2.4
// */
// U_STABLE UCollator * U_EXPORT2 usearch_getCollator(
//                                                const UStringSearch *strsrch);

// /**
// * Sets the collator used for the language rules. User retains the ownership
// * of this collator, thus the responsibility of deletion lies with the user.
// * This method causes internal data such as Boyer-Moore shift tables to
// * be recalculated, but the iterator's position is unchanged.
// * @param strsrch search iterator data struct
// * @param collator to be used
// * @param status for errors if it occurs
// * @stable ICU 2.4
// */
// U_STABLE void U_EXPORT2 usearch_setCollator(      UStringSearch *strsrch,
//                                           const UCollator     *collator,
//                                                 UErrorCode    *status);

// /**
// * Sets the pattern used for matching.
// * Internal data like the Boyer Moore table will be recalculated, but the
// * iterator's position is unchanged.
// * @param strsrch search iterator data struct
// * @param pattern string
// * @param patternlength pattern length, -1 for null-terminated string
// * @param status for errors if it occurs. If text is NULL, or textlength is 0
// *               then an U_ILLEGAL_ARGUMENT_ERROR is returned with no change
// *               done to strsrch.
// * @stable ICU 2.4
// */
// U_STABLE void U_EXPORT2 usearch_setPattern(      UStringSearch *strsrch,
//                                          const UChar         *pattern,
//                                                int32_t        patternlength,
//                                                UErrorCode    *status);

// /**
// * Gets the search pattern
// * @param strsrch search iterator data struct
// * @param length return length of the pattern, -1 indicates that the pattern
// *               is null-terminated
// * @return pattern string
// * @stable ICU 2.4
// */
// U_STABLE const UChar * U_EXPORT2 usearch_getPattern(
//                                                const UStringSearch *strsrch,
//                                                      int32_t       *length);

// /* methods ------------------------------------------------------------- */

// /**
// * Returns the first index at which the string text matches the search
// * pattern.
// * The iterator is adjusted so that its current index (as returned by
// * <tt>usearch_getOffset</tt>) is the match position if one was found.
// * If a match is not found, <tt>USEARCH_DONE</tt> will be returned and
// * the iterator will be adjusted to the index <tt>USEARCH_DONE</tt>.
// * @param strsrch search iterator data struct
// * @param status for errors if it occurs
// * @return The character index of the first match, or
// * <tt>USEARCH_DONE</tt> if there are no matches.
// * @see #usearch_getOffset
// * @see #USEARCH_DONE
// * @stable ICU 2.4
// */
// U_STABLE int32_t U_EXPORT2 usearch_first(UStringSearch *strsrch,
//                                            UErrorCode    *status);

// /**
// * Returns the first index equal or greater than <tt>position</tt> at which
// * the string text
// * matches the search pattern. The iterator is adjusted so that its current
// * index (as returned by <tt>usearch_getOffset</tt>) is the match position if
// * one was found.
// * If a match is not found, <tt>USEARCH_DONE</tt> will be returned and
// * the iterator will be adjusted to the index <tt>USEARCH_DONE</tt>
// * <p>
// * Search positions that may render incorrect results are highlighted in the
// * header comments. If position is less than or greater than the text range
// * for searching, an U_INDEX_OUTOFBOUNDS_ERROR will be returned
// * @param strsrch search iterator data struct
// * @param position to start the search at
// * @param status for errors if it occurs
// * @return The character index of the first match following <tt>pos</tt>,
// *         or <tt>USEARCH_DONE</tt> if there are no matches.
// * @see #usearch_getOffset
// * @see #USEARCH_DONE
// * @stable ICU 2.4
// */
// U_STABLE int32_t U_EXPORT2 usearch_following(UStringSearch *strsrch,
//                                                int32_t    position,
//                                                UErrorCode    *status);

// /**
// * Returns the last index in the target text at which it matches the search
// * pattern. The iterator is adjusted so that its current
// * index (as returned by <tt>usearch_getOffset</tt>) is the match position if
// * one was found.
// * If a match is not found, <tt>USEARCH_DONE</tt> will be returned and
// * the iterator will be adjusted to the index <tt>USEARCH_DONE</tt>.
// * @param strsrch search iterator data struct
// * @param status for errors if it occurs
// * @return The index of the first match, or <tt>USEARCH_DONE</tt> if there
// *         are no matches.
// * @see #usearch_getOffset
// * @see #USEARCH_DONE
// * @stable ICU 2.4
// */
// U_STABLE int32_t U_EXPORT2 usearch_last(UStringSearch *strsrch,
//                                           UErrorCode    *status);

// /**
// * Returns the first index less than <tt>position</tt> at which the string text
// * matches the search pattern. The iterator is adjusted so that its current
// * index (as returned by <tt>usearch_getOffset</tt>) is the match position if
// * one was found.
// * If a match is not found, <tt>USEARCH_DONE</tt> will be returned and
// * the iterator will be adjusted to the index <tt>USEARCH_DONE</tt>
// * <p>
// * Search positions that may render incorrect results are highlighted in the
// * header comments. If position is less than or greater than the text range
// * for searching, an U_INDEX_OUTOFBOUNDS_ERROR will be returned.
// * <p>
// * When <tt>USEARCH_OVERLAP</tt> option is off, the last index of the
// * result match is always less than <tt>position</tt>.
// * When <tt>USERARCH_OVERLAP</tt> is on, the result match may span across
// * <tt>position</tt>.
// * @param strsrch search iterator data struct
// * @param position index position the search is to begin at
// * @param status for errors if it occurs
// * @return The character index of the first match preceding <tt>pos</tt>,
// *         or <tt>USEARCH_DONE</tt> if there are no matches.
// * @see #usearch_getOffset
// * @see #USEARCH_DONE
// * @stable ICU 2.4
// */
// U_STABLE int32_t U_EXPORT2 usearch_preceding(UStringSearch *strsrch,
//                                                int32_t    position,
//                                                UErrorCode    *status);

// /**
// * Returns the index of the next point at which the string text matches the
// * search pattern, starting from the current position.
// * The iterator is adjusted so that its current
// * index (as returned by <tt>usearch_getOffset</tt>) is the match position if
// * one was found.
// * If a match is not found, <tt>USEARCH_DONE</tt> will be returned and
// * the iterator will be adjusted to the index <tt>USEARCH_DONE</tt>
// * @param strsrch search iterator data struct
// * @param status for errors if it occurs
// * @return The index of the next match after the current position, or
// *         <tt>USEARCH_DONE</tt> if there are no more matches.
// * @see #usearch_first
// * @see #usearch_getOffset
// * @see #USEARCH_DONE
// * @stable ICU 2.4
// */
// U_STABLE int32_t U_EXPORT2 usearch_next(UStringSearch *strsrch,
//                                           UErrorCode    *status);

// /**
// * Returns the index of the previous point at which the string text matches
// * the search pattern, starting at the current position.
// * The iterator is adjusted so that its current
// * index (as returned by <tt>usearch_getOffset</tt>) is the match position if
// * one was found.
// * If a match is not found, <tt>USEARCH_DONE</tt> will be returned and
// * the iterator will be adjusted to the index <tt>USEARCH_DONE</tt>
// * @param strsrch search iterator data struct
// * @param status for errors if it occurs
// * @return The index of the previous match before the current position,
// *         or <tt>USEARCH_DONE</tt> if there are no more matches.
// * @see #usearch_last
// * @see #usearch_getOffset
// * @see #USEARCH_DONE
// * @stable ICU 2.4
// */
// U_STABLE int32_t U_EXPORT2 usearch_previous(UStringSearch *strsrch,
//                                               UErrorCode    *status);

// /**
// * Reset the iteration.
// * Search will begin at the start of the text string if a forward iteration
// * is initiated before a backwards iteration. Otherwise if a backwards
// * iteration is initiated before a forwards iteration, the search will begin
// * at the end of the text string.
// * @param strsrch search iterator data struct
// * @see #usearch_first
// * @stable ICU 2.4
// */
// U_STABLE void U_EXPORT2 usearch_reset(UStringSearch *strsrch);

// #ifndef U_HIDE_INTERNAL_API
// /**
//   *  Simple forward search for the pattern, starting at a specified index,
//   *     and using a default set search options.
//   *
//   *  This is an experimental function, and is not an official part of the
//   *      ICU API.
//   *
//   *  The collator options, such as UCOL_STRENGTH and UCOL_NORMALIZTION, are honored.
//   *
//   *  The UStringSearch options USEARCH_CANONICAL_MATCH, USEARCH_OVERLAP and
//   *  any Break Iterator are ignored.
//   *
//   *  Matches obey the following constraints:
//   *
//   *      Characters at the start or end positions of a match that are ignorable
//   *      for collation are not included as part of the match, unless they
//   *      are part of a combining sequence, as described below.
//   *
//   *      A match will not include a partial combining sequence.  Combining
//   *      character sequences  are considered to be  inseparable units,
//   *      and either match the pattern completely, or are considered to not match
//   *      at all.  Thus, for example, an A followed a combining accent mark will
//   *      not be found when searching for a plain (unaccented) A.   (unless
//   *      the collation strength has been set to ignore all accents).
//   *
//   *      When beginning a search, the initial starting position, startIdx,
//   *      is assumed to be an acceptable match boundary with respect to
//   *      combining characters.  A combining sequence that spans across the
//   *      starting point will not suppress a match beginning at startIdx.
//   *
//   *      Characters that expand to multiple collation elements
//   *      (German sharp-S becoming 'ss', or the composed forms of accented
//   *      characters, for example) also must match completely.
//   *      Searching for a single 's' in a string containing only a sharp-s will
//   *      find no match.
//   *
//   *
//   *  @param strsrch    the UStringSearch struct, which references both
//   *                    the text to be searched  and the pattern being sought.
//   *  @param startIdx   The index into the text to begin the search.
//   *  @param matchStart An out parameter, the starting index of the matched text.
//   *                    This parameter may be NULL.
//   *                    A value of -1 will be returned if no match was found.
//   *  @param matchLimit Out parameter, the index of the first position following the matched text.
//   *                    The matchLimit will be at a suitable position for beginning a subsequent search
//   *                    in the input text.
//   *                    This parameter may be NULL.
//   *                    A value of -1 will be returned if no match was found.
//   *
//   *  @param status     Report any errors.  Note that no match found is not an error.
//   *  @return           TRUE if a match was found, FALSE otherwise.
//   *
//   *  @internal
//   */
// U_INTERNAL UBool U_EXPORT2 usearch_search(UStringSearch *strsrch,
//                                           int32_t        startIdx,
//                                           int32_t        *matchStart,
//                                           int32_t        *matchLimit,
//                                           UErrorCode     *status);

// /**
//   *  Simple backwards search for the pattern, starting at a specified index,
//   *     and using using a default set search options.
//   *
//   *  This is an experimental function, and is not an official part of the
//   *      ICU API.
//   *
//   *  The collator options, such as UCOL_STRENGTH and UCOL_NORMALIZTION, are honored.
//   *
//   *  The UStringSearch options USEARCH_CANONICAL_MATCH, USEARCH_OVERLAP and
//   *  any Break Iterator are ignored.
//   *
//   *  Matches obey the following constraints:
//   *
//   *      Characters at the start or end positions of a match that are ignorable
//   *      for collation are not included as part of the match, unless they
//   *      are part of a combining sequence, as described below.
//   *
//   *      A match will not include a partial combining sequence.  Combining
//   *      character sequences  are considered to be  inseparable units,
//   *      and either match the pattern completely, or are considered to not match
//   *      at all.  Thus, for example, an A followed a combining accent mark will
//   *      not be found when searching for a plain (unaccented) A.   (unless
//   *      the collation strength has been set to ignore all accents).
//   *
//   *      When beginning a search, the initial starting position, startIdx,
//   *      is assumed to be an acceptable match boundary with respect to
//   *      combining characters.  A combining sequence that spans across the
//   *      starting point will not suppress a match beginning at startIdx.
//   *
//   *      Characters that expand to multiple collation elements
//   *      (German sharp-S becoming 'ss', or the composed forms of accented
//   *      characters, for example) also must match completely.
//   *      Searching for a single 's' in a string containing only a sharp-s will
//   *      find no match.
//   *
//   *
//   *  @param strsrch    the UStringSearch struct, which references both
//   *                    the text to be searched  and the pattern being sought.
//   *  @param startIdx   The index into the text to begin the search.
//   *  @param matchStart An out parameter, the starting index of the matched text.
//   *                    This parameter may be NULL.
//   *                    A value of -1 will be returned if no match was found.
//   *  @param matchLimit Out parameter, the index of the first position following the matched text.
//   *                    The matchLimit will be at a suitable position for beginning a subsequent search
//   *                    in the input text.
//   *                    This parameter may be NULL.
//   *                    A value of -1 will be returned if no match was found.
//   *
//   *  @param status     Report any errors.  Note that no match found is not an error.
//   *  @return           TRUE if a match was found, FALSE otherwise.
//   *
//   *  @internal
//   */
// U_INTERNAL UBool U_EXPORT2 usearch_searchBackwards(UStringSearch *strsrch,
//                                                    int32_t        startIdx,
//                                                    int32_t        *matchStart,
//                                                    int32_t        *matchLimit,
//                                                    UErrorCode     *status);
// #endif  /* U_HIDE_INTERNAL_API */

// #endif /* #if !UCONFIG_NO_COLLATION  && !UCONFIG_NO_BREAK_ITERATION */

// #endif











U_NAMESPACE_USE

// don't use Boyer-Moore
// (and if we decide to turn this on again there are several new TODOs that will need to be addressed)
#define BOYER_MOORE 0

// internal definition ---------------------------------------------------

#define LAST_BYTE_MASK_          0xFF
#define SECOND_LAST_BYTE_SHIFT_  8
#define SUPPLEMENTARY_MIN_VALUE_ 0x10000

static const Normalizer2Impl *g_nfcImpl = NULL;

// internal methods -------------------------------------------------

/**
* Fast collation element iterator setOffset.
* This function does not check for bounds.
* @param coleiter collation element iterator
* @param offset to set
*/
static
inline void setColEIterOffset(UCollationElements *elems,
                      int32_t             offset)
{
    // Note: Not "fast" any more after the 2013 collation rewrite.
    // We do not want to expose more internals than necessary.
    UErrorCode status = U_ZERO_ERROR;
    ucol_setOffset(elems, offset, &status);
}

/**
* Getting the mask for collation strength
* @param strength collation strength
* @return collation element mask
*/
static
inline uint32_t getMask(UCollationStrength strength)
{
    switch (strength)
    {
    case UCOL_PRIMARY:
        return UCOL_PRIMARYORDERMASK;
    case UCOL_SECONDARY:
        return UCOL_SECONDARYORDERMASK | UCOL_PRIMARYORDERMASK;
    default:
        return UCOL_TERTIARYORDERMASK | UCOL_SECONDARYORDERMASK |
               UCOL_PRIMARYORDERMASK;
    }
}

/**
* @param ce 32-bit collation element
* @return hash code
*/
static
inline int hashFromCE32(uint32_t ce)
{
    int hc = (int)(
            ((((((ce >> 24) * 37) +
            (ce >> 16)) * 37) +
            (ce >> 8)) * 37) +
            ce);
    hc %= MAX_TABLE_SIZE_;
    if (hc < 0) {
        hc += MAX_TABLE_SIZE_;
    }
    return hc;
}

U_CDECL_BEGIN
static UBool U_CALLCONV
usearch_cleanup(void) {
    g_nfcImpl = NULL;
    return TRUE;
}
U_CDECL_END

/**
* Initializing the fcd tables.
* Internal method, status assumed to be a success.
* @param status output error if any, caller to check status before calling
*               method, status assumed to be success when passed in.
*/
static
inline void initializeFCD(UErrorCode *status)
{
    if (g_nfcImpl == NULL) {
        g_nfcImpl = Normalizer2Factory::getNFCImpl(*status);
        ucln_i18n_registerCleanup(UCLN_I18N_USEARCH, usearch_cleanup);
    }
}

/**
* Gets the fcd value for a character at the argument index.
* This method takes into accounts of the supplementary characters.
* @param str UTF16 string where character for fcd retrieval resides
* @param offset position of the character whose fcd is to be retrieved, to be
*               overwritten with the next character position, taking
*               surrogate characters into consideration.
* @param strlength length of the argument string
* @return fcd value
*/
static
uint16_t getFCD(const UChar   *str, int32_t *offset,
                             int32_t  strlength)
{
    const UChar *temp = str + *offset;
    uint16_t    result = g_nfcImpl->nextFCD16(temp, str + strlength);
    *offset = (int32_t)(temp - str);
    return result;
}

/**
* Getting the modified collation elements taking into account the collation
* attributes
* @param strsrch string search data
* @param sourcece
* @return the modified collation element
*/
static
inline int32_t getCE(const UStringSearch *strsrch, uint32_t sourcece)
{
    // note for tertiary we can't use the collator->tertiaryMask, that
    // is a preprocessed mask that takes into account case options. since
    // we are only concerned with exact matches, we don't need that.
    sourcece &= strsrch->ceMask;

    if (strsrch->toShift) {
        // alternate handling here, since only the 16 most significant digits
        // is only used, we can safely do a compare without masking
        // if the ce is a variable, we mask and get only the primary values
        // no shifting to quartenary is required since all primary values
        // less than variabletop will need to be masked off anyway.
        if (strsrch->variableTop > sourcece) {
            if (strsrch->strength >= UCOL_QUATERNARY) {
                sourcece &= UCOL_PRIMARYORDERMASK;
            }
            else {
                sourcece = UCOL_IGNORABLE;
            }
        }
    } else if (strsrch->strength >= UCOL_QUATERNARY && sourcece == UCOL_IGNORABLE) {
        sourcece = 0xFFFF;
    }

    return sourcece;
}

/**
* Allocate a memory and returns NULL if it failed.
* Internal method, status assumed to be a success.
* @param size to allocate
* @param status output error if any, caller to check status before calling
*               method, status assumed to be success when passed in.
* @return newly allocated array, NULL otherwise
*/
static
inline void * allocateMemory(uint32_t size, UErrorCode *status)
{
    uint32_t *result = (uint32_t *)uprv_malloc(size);
    if (result == NULL) {
        *status = U_MEMORY_ALLOCATION_ERROR;
    }
    return result;
}

/**
* Adds a uint32_t value to a destination array.
* Creates a new array if we run out of space. The caller will have to
* manually deallocate the newly allocated array.
* Internal method, status assumed to be success, caller has to check status
* before calling this method. destination not to be NULL and has at least
* size destinationlength.
* @param destination target array
* @param offset destination offset to add value
* @param destinationlength target array size, return value for the new size
* @param value to be added
* @param increments incremental size expected
* @param status output error if any, caller to check status before calling
*               method, status assumed to be success when passed in.
* @return new destination array, destination if there was no new allocation
*/
static
inline int32_t * addTouint32_tArray(int32_t    *destination,
                                    uint32_t    offset,
                                    uint32_t   *destinationlength,
                                    uint32_t    value,
                                    uint32_t    increments,
                                    UErrorCode *status)
{
    uint32_t newlength = *destinationlength;
    if (offset + 1 == newlength) {
        newlength += increments;
        int32_t *temp = (int32_t *)allocateMemory(
                                         sizeof(int32_t) * newlength, status);
        if (U_FAILURE(*status)) {
            return NULL;
        }
        uprv_memcpy(temp, destination, sizeof(int32_t) * (size_t)offset);
        *destinationlength = newlength;
        destination        = temp;
    }
    destination[offset] = value;
    return destination;
}

/**
* Adds a uint64_t value to a destination array.
* Creates a new array if we run out of space. The caller will have to
* manually deallocate the newly allocated array.
* Internal method, status assumed to be success, caller has to check status
* before calling this method. destination not to be NULL and has at least
* size destinationlength.
* @param destination target array
* @param offset destination offset to add value
* @param destinationlength target array size, return value for the new size
* @param value to be added
* @param increments incremental size expected
* @param status output error if any, caller to check status before calling
*               method, status assumed to be success when passed in.
* @return new destination array, destination if there was no new allocation
*/
static
inline int64_t * addTouint64_tArray(int64_t    *destination,
                                    uint32_t    offset,
                                    uint32_t   *destinationlength,
                                    uint64_t    value,
                                    uint32_t    increments,
                                    UErrorCode *status)
{
    uint32_t newlength = *destinationlength;
    if (offset + 1 == newlength) {
        newlength += increments;
        int64_t *temp = (int64_t *)allocateMemory(
                                         sizeof(int64_t) * newlength, status);

        if (U_FAILURE(*status)) {
            return NULL;
        }

        uprv_memcpy(temp, destination, sizeof(int64_t) * (size_t)offset);
        *destinationlength = newlength;
        destination        = temp;
    }

    destination[offset] = value;

    return destination;
}

/**
* Initializing the ce table for a pattern.
* Stores non-ignorable collation keys.
* Table size will be estimated by the size of the pattern text. Table
* expansion will be perform as we go along. Adding 1 to ensure that the table
* size definitely increases.
* Internal method, status assumed to be a success.
* @param strsrch string search data
* @param status output error if any, caller to check status before calling
*               method, status assumed to be success when passed in.
* @return total number of expansions
*/
static
inline uint16_t initializePatternCETable(UStringSearch *strsrch,
                                         UErrorCode    *status)
{
    UPattern *pattern            = &(strsrch->pattern);
    uint32_t  cetablesize        = INITIAL_ARRAY_SIZE_;
    int32_t  *cetable            = pattern->cesBuffer;
    uint32_t  patternlength      = pattern->textLength;
    UCollationElements *coleiter = strsrch->utilIter;

    if (coleiter == NULL) {
        coleiter = ucol_openElements(strsrch->collator, pattern->text,
                                     patternlength, status);
        // status will be checked in ucol_next(..) later and if it is an
        // error UCOL_NULLORDER the result of ucol_next(..) and 0 will be
        // returned.
        strsrch->utilIter = coleiter;
    }
    else {
        ucol_setText(coleiter, pattern->text, pattern->textLength, status);
    }
    if(U_FAILURE(*status)) {
        return 0;
    }

    if (pattern->ces != cetable && pattern->ces) {
        uprv_free(pattern->ces);
    }

    uint32_t  offset      = 0;
    uint16_t  result      = 0;
    int32_t   ce;

    while ((ce = ucol_next(coleiter, status)) != UCOL_NULLORDER &&
           U_SUCCESS(*status)) {
        uint32_t newce = getCE(strsrch, ce);
        if (newce) {
            int32_t *temp = addTouint32_tArray(cetable, offset, &cetablesize,
                                  newce,
                                  patternlength - ucol_getOffset(coleiter) + 1,
                                  status);
            if (U_FAILURE(*status)) {
                return 0;
            }
            offset ++;
            if (cetable != temp && cetable != pattern->cesBuffer) {
                uprv_free(cetable);
            }
            cetable = temp;
        }
        result += (uint16_t)(ucol_getMaxExpansion(coleiter, ce) - 1);
    }

    cetable[offset]   = 0;
    pattern->ces       = cetable;
    pattern->cesLength = offset;

    return result;
}

/**
* Initializing the pce table for a pattern.
* Stores non-ignorable collation keys.
* Table size will be estimated by the size of the pattern text. Table
* expansion will be perform as we go along. Adding 1 to ensure that the table
* size definitely increases.
* Internal method, status assumed to be a success.
* @param strsrch string search data
* @param status output error if any, caller to check status before calling
*               method, status assumed to be success when passed in.
* @return total number of expansions
*/
static
inline uint16_t initializePatternPCETable(UStringSearch *strsrch,
                                          UErrorCode    *status)
{
    UPattern *pattern            = &(strsrch->pattern);
    uint32_t  pcetablesize       = INITIAL_ARRAY_SIZE_;
    int64_t  *pcetable           = pattern->pcesBuffer;
    uint32_t  patternlength      = pattern->textLength;
    UCollationElements *coleiter = strsrch->utilIter;

    if (coleiter == NULL) {
        coleiter = ucol_openElements(strsrch->collator, pattern->text,
                                     patternlength, status);
        // status will be checked in ucol_next(..) later and if it is an
        // error UCOL_NULLORDER the result of ucol_next(..) and 0 will be
        // returned.
        strsrch->utilIter = coleiter;
    } else {
        ucol_setText(coleiter, pattern->text, pattern->textLength, status);
    }
    if(U_FAILURE(*status)) {
        return 0;
    }

    if (pattern->pces != pcetable && pattern->pces != NULL) {
        uprv_free(pattern->pces);
    }

    uint32_t  offset = 0;
    uint16_t  result = 0;
    int64_t   pce;

    icu::UCollationPCE iter(coleiter);

    // ** Should processed CEs be signed or unsigned?
    // ** (the rest of the code in this file seems to play fast-and-loose with
    // **  whether a CE is signed or unsigned. For example, look at routine above this one.)
    while ((pce = iter.nextProcessed(NULL, NULL, status)) != UCOL_PROCESSED_NULLORDER &&
           U_SUCCESS(*status)) {
        int64_t *temp = addTouint64_tArray(pcetable, offset, &pcetablesize,
                              pce,
                              patternlength - ucol_getOffset(coleiter) + 1,
                              status);

        if (U_FAILURE(*status)) {
            return 0;
        }

        offset += 1;

        if (pcetable != temp && pcetable != pattern->pcesBuffer) {
            uprv_free(pcetable);
        }

        pcetable = temp;
        //result += (uint16_t)(ucol_getMaxExpansion(coleiter, ce) - 1);
    }

    pcetable[offset]   = 0;
    pattern->pces       = pcetable;
    pattern->pcesLength = offset;

    return result;
}

/**
* Initializes the pattern struct.
* Internal method, status assumed to be success.
* @param strsrch UStringSearch data storage
* @param status output error if any, caller to check status before calling
*               method, status assumed to be success when passed in.
* @return expansionsize the total expansion size of the pattern
*/
static
inline int16_t initializePattern(UStringSearch *strsrch, UErrorCode *status)
{
    if (U_FAILURE(*status)) { return 0; }
          UPattern   *pattern     = &(strsrch->pattern);
    const UChar      *patterntext = pattern->text;
          int32_t     length      = pattern->textLength;
          int32_t index       = 0;

    // Since the strength is primary, accents are ignored in the pattern.
    if (strsrch->strength == UCOL_PRIMARY) {
        pattern->hasPrefixAccents = 0;
        pattern->hasSuffixAccents = 0;
    } else {
        pattern->hasPrefixAccents = getFCD(patterntext, &index, length) >>
                                                         SECOND_LAST_BYTE_SHIFT_;
        index = length;
        U16_BACK_1(patterntext, 0, index);
        pattern->hasSuffixAccents = getFCD(patterntext, &index, length) &
                                                                 LAST_BYTE_MASK_;
    }

    // ** HACK **
    if (strsrch->pattern.pces != NULL) {
        if (strsrch->pattern.pces != strsrch->pattern.pcesBuffer) {
            uprv_free(strsrch->pattern.pces);
        }

        strsrch->pattern.pces = NULL;
    }

    // since intializePattern is an internal method status is a success.
    return initializePatternCETable(strsrch, status);
}

/**
* Initializing shift tables, with the default values.
* If a corresponding default value is 0, the shift table is not set.
* @param shift table for forwards shift
* @param backshift table for backwards shift
* @param cetable table containing pattern ce
* @param cesize size of the pattern ces
* @param expansionsize total size of the expansions
* @param defaultforward the default forward value
* @param defaultbackward the default backward value
*/
static
inline void setShiftTable(int16_t   shift[], int16_t backshift[],
                          int32_t  *cetable, int32_t cesize,
                          int16_t   expansionsize,
                          int16_t   defaultforward,
                          int16_t   defaultbackward)
{
    // estimate the value to shift. to do that we estimate the smallest
    // number of characters to give the relevant ces, ie approximately
    // the number of ces minus their expansion, since expansions can come
    // from a character.
    int32_t count;
    for (count = 0; count < MAX_TABLE_SIZE_; count ++) {
        shift[count] = defaultforward;
    }
    cesize --; // down to the last index
    for (count = 0; count < cesize; count ++) {
        // number of ces from right of array to the count
        int temp = defaultforward - count - 1;
        shift[hashFromCE32(cetable[count])] = temp > 1 ? static_cast<int16_t>(temp) : 1;
    }
    shift[hashFromCE32(cetable[cesize])] = 1;
    // for ignorables we just shift by one. see test examples.
    shift[hashFromCE32(0)] = 1;

    for (count = 0; count < MAX_TABLE_SIZE_; count ++) {
        backshift[count] = defaultbackward;
    }
    for (count = cesize; count > 0; count --) {
        // the original value count does not seem to work
        backshift[hashFromCE32(cetable[count])] = count > expansionsize ?
                                          (int16_t)(count - expansionsize) : 1;
    }
    backshift[hashFromCE32(cetable[0])] = 1;
    backshift[hashFromCE32(0)] = 1;
}

/**
* Building of the pattern collation element list and the boyer moore strsrch
* table.
* The canonical match will only be performed after the default match fails.
* For both cases we need to remember the size of the composed and decomposed
* versions of the string. Since the Boyer-Moore shift calculations shifts by
* a number of characters in the text and tries to match the pattern from that
* offset, the shift value can not be too large in case we miss some
* characters. To choose a right shift size, we estimate the NFC form of the
* and use its size as a shift guide. The NFC form should be the small
* possible representation of the pattern. Anyways, we'll err on the smaller
* shift size. Hence the calculation for minlength.
* Canonical match will be performed slightly differently. We'll split the
* pattern into 3 parts, the prefix accents (PA), the middle string bounded by
* the first and last base character (MS), the ending accents (EA). Matches
* will be done on MS first, and only when we match MS then some processing
* will be required for the prefix and end accents in order to determine if
* they match PA and EA. Hence the default shift values
* for the canonical match will take the size of either end's accent into
* consideration. Forwards search will take the end accents into consideration
* for the default shift values and the backwards search will take the prefix
* accents into consideration.
* If pattern has no non-ignorable ce, we return a illegal argument error.
* Internal method, status assumed to be success.
* @param strsrch UStringSearch data storage
* @param status  for output errors if it occurs, status is assumed to be a
*                success when it is passed in.
*/
static
inline void initialize(UStringSearch *strsrch, UErrorCode *status)
{
    int16_t expandlength  = initializePattern(strsrch, status);
    if (U_SUCCESS(*status) && strsrch->pattern.cesLength > 0) {
        UPattern *pattern = &strsrch->pattern;
        int32_t   cesize  = pattern->cesLength;

        int16_t minlength = cesize > expandlength
                            ? (int16_t)cesize - expandlength : 1;
        pattern->defaultShiftSize    = minlength;
        setShiftTable(pattern->shift, pattern->backShift, pattern->ces,
                      cesize, expandlength, minlength, minlength);
        return;
    }
    strsrch->pattern.defaultShiftSize = 0;
}

#if BOYER_MOORE
/**
* Check to make sure that the match length is at the end of the character by
* using the breakiterator.
* @param strsrch string search data
* @param start target text start offset
* @param end target text end offset
*/
static
void checkBreakBoundary(const UStringSearch *strsrch, int32_t * /*start*/,
                               int32_t *end)
{
#if !UCONFIG_NO_BREAK_ITERATION
    UBreakIterator *breakiterator = strsrch->search->internalBreakIter;
    if (breakiterator) {
        int32_t matchend = *end;
        //int32_t matchstart = *start;

        if (!ubrk_isBoundary(breakiterator, matchend)) {
            *end = ubrk_following(breakiterator, matchend);
        }

        /* Check the start of the matched text to make sure it doesn't have any accents
         * before it.  This code may not be necessary and so it is commented out */
        /*if (!ubrk_isBoundary(breakiterator, matchstart) && !ubrk_isBoundary(breakiterator, matchstart-1)) {
            *start = ubrk_preceding(breakiterator, matchstart);
        }*/
    }
#endif
}

/**
* Determine whether the target text in UStringSearch bounded by the offset
* start and end is one or more whole units of text as
* determined by the breakiterator in UStringSearch.
* @param strsrch string search data
* @param start target text start offset
* @param end target text end offset
*/
static
UBool isBreakUnit(const UStringSearch *strsrch, int32_t start,
                               int32_t    end)
{
#if !UCONFIG_NO_BREAK_ITERATION
    UBreakIterator *breakiterator = strsrch->search->breakIter;
    //TODO: Add here.
    if (breakiterator) {
        int32_t startindex = ubrk_first(breakiterator);
        int32_t endindex   = ubrk_last(breakiterator);

        // out-of-range indexes are never boundary positions
        if (start < startindex || start > endindex ||
            end < startindex || end > endindex) {
            return FALSE;
        }
        // otherwise, we can use following() on the position before the
        // specified one and return true of the position we get back is the
        // one the user specified
        UBool result = (start == startindex ||
                ubrk_following(breakiterator, start - 1) == start) &&
               (end == endindex ||
                ubrk_following(breakiterator, end - 1) == end);
        if (result) {
            // iterates the individual ces
                  UCollationElements *coleiter  = strsrch->utilIter;
            const UChar              *text      = strsrch->search->text +
                                                                      start;
                  UErrorCode          status    = U_ZERO_ERROR;
            ucol_setText(coleiter, text, end - start, &status);
            for (int32_t count = 0; count < strsrch->pattern.cesLength;
                 count ++) {
                int32_t ce = getCE(strsrch, ucol_next(coleiter, &status));
                if (ce == UCOL_IGNORABLE) {
                    count --;
                    continue;
                }
                if (U_FAILURE(status) || ce != strsrch->pattern.ces[count]) {
                    return FALSE;
                }
            }
            int32_t nextce = ucol_next(coleiter, &status);
            while (ucol_getOffset(coleiter) == (end - start)
                   && getCE(strsrch, nextce) == UCOL_IGNORABLE) {
                nextce = ucol_next(coleiter, &status);
            }
            if (ucol_getOffset(coleiter) == (end - start)
                && nextce != UCOL_NULLORDER) {
                // extra collation elements at the end of the match
                return FALSE;
            }
        }
        return result;
    }
#endif
    return TRUE;
}

/**
* Getting the next base character offset if current offset is an accent,
* or the current offset if the current character contains a base character.
* accents the following base character will be returned
* @param text string
* @param textoffset current offset
* @param textlength length of text string
* @return the next base character or the current offset
*         if the current character is contains a base character.
*/
static
inline int32_t getNextBaseOffset(const UChar       *text,
                                           int32_t  textoffset,
                                           int32_t      textlength)
{
    if (textoffset < textlength) {
        int32_t temp = textoffset;
        if (getFCD(text, &temp, textlength) >> SECOND_LAST_BYTE_SHIFT_) {
            while (temp < textlength) {
                int32_t result = temp;
                if ((getFCD(text, &temp, textlength) >>
                     SECOND_LAST_BYTE_SHIFT_) == 0) {
                    return result;
                }
            }
            return textlength;
        }
    }
    return textoffset;
}

/**
* Gets the next base character offset depending on the string search pattern
* data
* @param strsrch string search data
* @param textoffset current offset, one offset away from the last character
*                   to search for.
* @return start index of the next base character or the current offset
*         if the current character is contains a base character.
*/
static
inline int32_t getNextUStringSearchBaseOffset(UStringSearch *strsrch,
                                                  int32_t    textoffset)
{
    int32_t textlength = strsrch->search->textLength;
    if (strsrch->pattern.hasSuffixAccents &&
        textoffset < textlength) {
              int32_t  temp       = textoffset;
        const UChar       *text       = strsrch->search->text;
        U16_BACK_1(text, 0, temp);
        if (getFCD(text, &temp, textlength) & LAST_BYTE_MASK_) {
            return getNextBaseOffset(text, textoffset, textlength);
        }
    }
    return textoffset;
}

/**
* Shifting the collation element iterator position forward to prepare for
* a following match. If the last character is a unsafe character, we'll only
* shift by 1 to capture contractions, normalization etc.
* Internal method, status assumed to be success.
* @param text strsrch string search data
* @param textoffset start text position to do search
* @param ce the text ce which failed the match.
* @param patternceindex index of the ce within the pattern ce buffer which
*        failed the match
* @return final offset
*/
static
inline int32_t shiftForward(UStringSearch *strsrch,
                                int32_t    textoffset,
                                int32_t       ce,
                                int32_t        patternceindex)
{
    UPattern *pattern = &(strsrch->pattern);
    if (ce != UCOL_NULLORDER) {
        int32_t shift = pattern->shift[hashFromCE32(ce)];
        // this is to adjust for characters in the middle of the
        // substring for matching that failed.
        int32_t adjust = pattern->cesLength - patternceindex;
        if (adjust > 1 && shift >= adjust) {
            shift -= adjust - 1;
        }
        textoffset += shift;
    }
    else {
        textoffset += pattern->defaultShiftSize;
    }

    textoffset = getNextUStringSearchBaseOffset(strsrch, textoffset);
    // check for unsafe characters
    // * if it is the start or middle of a contraction: to be done after
    //   a initial match is found
    // * thai or lao base consonant character: similar to contraction
    // * high surrogate character: similar to contraction
    // * next character is a accent: shift to the next base character
    return textoffset;
}
#endif // #if BOYER_MOORE

/**
* sets match not found
* @param strsrch string search data
*/
static
inline void setMatchNotFound(UStringSearch *strsrch)
{
    // this method resets the match result regardless of the error status.
    strsrch->search->matchedIndex = USEARCH_DONE;
    strsrch->search->matchedLength = 0;
    if (strsrch->search->isForwardSearching) {
        setColEIterOffset(strsrch->textIter, strsrch->search->textLength);
    }
    else {
        setColEIterOffset(strsrch->textIter, 0);
    }
}

#if BOYER_MOORE
/**
* Gets the offset to the next safe point in text.
* ie. not the middle of a contraction, swappable characters or supplementary
* characters.
* @param collator collation sata
* @param text string to work with
* @param textoffset offset in string
* @param textlength length of text string
* @return offset to the next safe character
*/
static
inline int32_t getNextSafeOffset(const UCollator   *collator,
                                     const UChar       *text,
                                           int32_t  textoffset,
                                           int32_t      textlength)
{
    int32_t result = textoffset; // first contraction character
    while (result != textlength && ucol_unsafeCP(text[result], collator)) {
        result ++;
    }
    return result;
}

/**
* This checks for accents in the potential match started with a .
* composite character.
* This is really painful... we have to check that composite character do not
* have any extra accents. We have to normalize the potential match and find
* the immediate decomposed character before the match.
* The first composite character would have been taken care of by the fcd
* checks in checkForwardExactMatch.
* This is the slow path after the fcd of the first character and
* the last character has been checked by checkForwardExactMatch and we
* determine that the potential match has extra non-ignorable preceding
* ces.
* E.g. looking for \u0301 acute in \u01FA A ring above and acute,
* checkExtraMatchAccent should fail since there is a middle ring in \u01FA
* Note here that accents checking are slow and cautioned in the API docs.
* Internal method, status assumed to be a success, caller should check status
* before calling this method
* @param strsrch string search data
* @param start index of the potential unfriendly composite character
* @param end index of the potential unfriendly composite character
* @param status output error status if any.
* @return TRUE if there is non-ignorable accents before at the beginning
*              of the match, FALSE otherwise.
*/

static
UBool checkExtraMatchAccents(const UStringSearch *strsrch, int32_t start,
                                   int32_t    end,
                                   UErrorCode    *status)
{
    UBool result = FALSE;
    if (strsrch->pattern.hasPrefixAccents) {
              int32_t  length = end - start;
              int32_t  offset = 0;
        const UChar       *text   = strsrch->search->text + start;

        U16_FWD_1(text, offset, length);
        // we are only concerned with the first composite character
        if (unorm_quickCheck(text, offset, UNORM_NFD, status) == UNORM_NO) {
            int32_t safeoffset = getNextSafeOffset(strsrch->collator,
                                                       text, 0, length);
            if (safeoffset != length) {
                safeoffset ++;
            }
            UChar   *norm = NULL;
            UChar    buffer[INITIAL_ARRAY_SIZE_];
            int32_t  size = unorm_normalize(text, safeoffset, UNORM_NFD, 0,
                                            buffer, INITIAL_ARRAY_SIZE_,
                                            status);
            if (U_FAILURE(*status)) {
                return FALSE;
            }
            if (size >= INITIAL_ARRAY_SIZE_) {
                norm = (UChar *)allocateMemory((size + 1) * sizeof(UChar),
                                               status);
                // if allocation failed, status will be set to
                // U_MEMORY_ALLOCATION_ERROR and unorm_normalize internally
                // checks for it.
                size = unorm_normalize(text, safeoffset, UNORM_NFD, 0, norm,
                                       size, status);
                if (U_FAILURE(*status) && norm != NULL) {
                    uprv_free(norm);
                    return FALSE;
                }
            }
            else {
                norm = buffer;
            }

            UCollationElements *coleiter  = strsrch->utilIter;
            ucol_setText(coleiter, norm, size, status);
            uint32_t            firstce   = strsrch->pattern.ces[0];
            UBool               ignorable = TRUE;
            uint32_t            ce        = UCOL_IGNORABLE;
            while (U_SUCCESS(*status) && ce != firstce && ce != (uint32_t)UCOL_NULLORDER) {
                offset = ucol_getOffset(coleiter);
                if (ce != firstce && ce != UCOL_IGNORABLE) {
                    ignorable = FALSE;
                }
                ce = ucol_next(coleiter, status);
            }
            UChar32 codepoint;
            U16_PREV(norm, 0, offset, codepoint);
            result = !ignorable && (u_getCombiningClass(codepoint) != 0);

            if (norm != buffer) {
                uprv_free(norm);
            }
        }
    }

    return result;
}

/**
* Used by exact matches, checks if there are accents before the match.
* This is really painful... we have to check that composite characters at
* the start of the matches have to not have any extra accents.
* We check the FCD of the character first, if it starts with an accent and
* the first pattern ce does not match the first ce of the character, we bail.
* Otherwise we try normalizing the first composite
* character and find the immediate decomposed character before the match to
* see if it is an non-ignorable accent.
* Now normalizing the first composite character is enough because we ensure
* that when the match is passed in here with extra beginning ces, the
* first or last ce that match has to occur within the first character.
* E.g. looking for \u0301 acute in \u01FA A ring above and acute,
* checkExtraMatchAccent should fail since there is a middle ring in \u01FA
* Note here that accents checking are slow and cautioned in the API docs.
* @param strsrch string search data
* @param start offset
* @param end offset
* @return TRUE if there are accents on either side of the match,
*         FALSE otherwise
*/
static
UBool hasAccentsBeforeMatch(const UStringSearch *strsrch, int32_t start,
                                  int32_t    end)
{
    if (strsrch->pattern.hasPrefixAccents) {
        UCollationElements *coleiter  = strsrch->textIter;
        UErrorCode          status    = U_ZERO_ERROR;
        // we have been iterating forwards previously
        uint32_t            ignorable = TRUE;
        int32_t             firstce   = strsrch->pattern.ces[0];

        setColEIterOffset(coleiter, start);
        int32_t ce  = getCE(strsrch, ucol_next(coleiter, &status));
        if (U_FAILURE(status)) {
            return TRUE;
        }
        while (ce != firstce) {
            if (ce != UCOL_IGNORABLE) {
                ignorable = FALSE;
            }
            ce = getCE(strsrch, ucol_next(coleiter, &status));
            if (U_FAILURE(status) || ce == UCOL_NULLORDER) {
                return TRUE;
            }
        }
        if (!ignorable && inNormBuf(coleiter)) {
            // within normalization buffer, discontiguous handled here
            return TRUE;
        }

        // within text
        int32_t temp = start;
        // original code
        // accent = (getFCD(strsrch->search->text, &temp,
        //                  strsrch->search->textLength)
        //            >> SECOND_LAST_BYTE_SHIFT_);
        // however this code does not work well with VC7 .net in release mode.
        // maybe the inlines for getFCD combined with shifting has bugs in
        // VC7. anyways this is a work around.
        UBool accent = getFCD(strsrch->search->text, &temp,
                              strsrch->search->textLength) > 0xFF;
        if (!accent) {
            return checkExtraMatchAccents(strsrch, start, end, &status);
        }
        if (!ignorable) {
            return TRUE;
        }
        if (start > 0) {
            temp = start;
            U16_BACK_1(strsrch->search->text, 0, temp);
            if (getFCD(strsrch->search->text, &temp,
                       strsrch->search->textLength) & LAST_BYTE_MASK_) {
                setColEIterOffset(coleiter, start);
                ce = ucol_previous(coleiter, &status);
                if (U_FAILURE(status) ||
                    (ce != UCOL_NULLORDER && ce != UCOL_IGNORABLE)) {
                    return TRUE;
                }
            }
        }
    }

    return FALSE;
}

/**
* Used by exact matches, checks if there are accents bounding the match.
* Note this is the initial boundary check. If the potential match
* starts or ends with composite characters, the accents in those
* characters will be determined later.
* Not doing backwards iteration here, since discontiguos contraction for
* backwards collation element iterator, use up too many characters.
* E.g. looking for \u030A ring in \u01FA A ring above and acute,
* should fail since there is a acute at the end of \u01FA
* Note here that accents checking are slow and cautioned in the API docs.
* @param strsrch string search data
* @param start offset of match
* @param end end offset of the match
* @return TRUE if there are accents on either side of the match,
*         FALSE otherwise
*/
static
UBool hasAccentsAfterMatch(const UStringSearch *strsrch, int32_t start,
                                 int32_t    end)
{
    if (strsrch->pattern.hasSuffixAccents) {
        const UChar       *text       = strsrch->search->text;
              int32_t  temp       = end;
              int32_t      textlength = strsrch->search->textLength;
        U16_BACK_1(text, 0, temp);
        if (getFCD(text, &temp, textlength) & LAST_BYTE_MASK_) {
            int32_t             firstce  = strsrch->pattern.ces[0];
            UCollationElements *coleiter = strsrch->textIter;
            UErrorCode          status   = U_ZERO_ERROR;
            int32_t ce;
            setColEIterOffset(coleiter, start);
            while ((ce = getCE(strsrch, ucol_next(coleiter, &status))) != firstce) {
                if (U_FAILURE(status) || ce == UCOL_NULLORDER) {
                    return TRUE;
                }
            }
            int32_t count = 1;
            while (count < strsrch->pattern.cesLength) {
                if (getCE(strsrch, ucol_next(coleiter, &status))
                    == UCOL_IGNORABLE) {
                    // Thai can give an ignorable here.
                    count --;
                }
                if (U_FAILURE(status)) {
                    return TRUE;
                }
                count ++;
            }

            ce = ucol_next(coleiter, &status);
            if (U_FAILURE(status)) {
                return TRUE;
            }
            if (ce != UCOL_NULLORDER && ce != UCOL_IGNORABLE) {
                ce = getCE(strsrch, ce);
            }
            if (ce != UCOL_NULLORDER && ce != UCOL_IGNORABLE) {
                if (ucol_getOffset(coleiter) <= end) {
                    return TRUE;
                }
                if (getFCD(text, &end, textlength) >> SECOND_LAST_BYTE_SHIFT_) {
                    return TRUE;
                }
            }
        }
    }
    return FALSE;
}
#endif // #if BOYER_MOORE

/**
* Checks if the offset runs out of the text string
* @param offset
* @param textlength of the text string
* @return TRUE if offset is out of bounds, FALSE otherwise
*/
static
inline UBool isOutOfBounds(int32_t textlength, int32_t offset)
{
    return offset < 0 || offset > textlength;
}

/**
* Checks for identical match
* @param strsrch string search data
* @param start offset of possible match
* @param end offset of possible match
* @return TRUE if identical match is found
*/
static
inline UBool checkIdentical(const UStringSearch *strsrch, int32_t start,
                                  int32_t    end)
{
    if (strsrch->strength != UCOL_IDENTICAL) {
        return TRUE;
    }

    // Note: We could use Normalizer::compare() or similar, but for short strings
    // which may not be in FCD it might be faster to just NFD them.
    UErrorCode status = U_ZERO_ERROR;
    UnicodeString t2, p2;
    strsrch->nfd->normalize(
        UnicodeString(FALSE, strsrch->search->text + start, end - start), t2, status);
    strsrch->nfd->normalize(
        UnicodeString(FALSE, strsrch->pattern.text, strsrch->pattern.textLength), p2, status);
    // return FALSE if NFD failed
    return U_SUCCESS(status) && t2 == p2;
}

#if BOYER_MOORE
/**
* Checks to see if the match is repeated
* @param strsrch string search data
* @param start new match start index
* @param end new match end index
* @return TRUE if the the match is repeated, FALSE otherwise
*/
static
inline UBool checkRepeatedMatch(UStringSearch *strsrch,
                                int32_t    start,
                                int32_t    end)
{
    int32_t lastmatchindex = strsrch->search->matchedIndex;
    UBool       result;
    if (lastmatchindex == USEARCH_DONE) {
        return FALSE;
    }
    if (strsrch->search->isForwardSearching) {
        result = start <= lastmatchindex;
    }
    else {
        result = start >= lastmatchindex;
    }
    if (!result && !strsrch->search->isOverlap) {
        if (strsrch->search->isForwardSearching) {
            result = start < lastmatchindex + strsrch->search->matchedLength;
        }
        else {
            result = end > lastmatchindex;
        }
    }
    return result;
}

/**
* Gets the collation element iterator's current offset.
* @param coleiter collation element iterator
* @param forwards flag TRUE if we are moving in th forwards direction
* @return current offset
*/
static
inline int32_t getColElemIterOffset(const UCollationElements *coleiter,
                                              UBool               forwards)
{
    int32_t result = ucol_getOffset(coleiter);
    // intricacies of the the backwards collation element iterator
    if (FALSE && !forwards && inNormBuf(coleiter) && !isFCDPointerNull(coleiter)) {
        result ++;
    }
    return result;
}

/**
* Checks match for contraction.
* If the match ends with a partial contraction we fail.
* If the match starts too far off (because of backwards iteration) we try to
* chip off the extra characters depending on whether a breakiterator has
* been used.
* Internal method, error assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param start offset of potential match, to be modified if necessary
* @param end offset of potential match, to be modified if necessary
* @param status output error status if any
* @return TRUE if match passes the contraction test, FALSE otherwise
*/

static
UBool checkNextExactContractionMatch(UStringSearch *strsrch,
                                     int32_t   *start,
                                     int32_t   *end, UErrorCode  *status)
{
          UCollationElements *coleiter   = strsrch->textIter;
          int32_t             textlength = strsrch->search->textLength;
          int32_t             temp       = *start;
    const UCollator          *collator   = strsrch->collator;
    const UChar              *text       = strsrch->search->text;
    // This part checks if either ends of the match contains potential
    // contraction. If so we'll have to iterate through them
    // The start contraction needs to be checked since ucol_previous dumps
    // all characters till the first safe character into the buffer.
    // *start + 1 is used to test for the unsafe characters instead of *start
    // because ucol_prev takes all unsafe characters till the first safe
    // character ie *start. so by testing *start + 1, we can estimate if
    // excess prefix characters has been included in the potential search
    // results.
    if ((*end < textlength && ucol_unsafeCP(text[*end], collator)) ||
        (*start + 1 < textlength
         && ucol_unsafeCP(text[*start + 1], collator))) {
        int32_t expansion  = getExpansionPrefix(coleiter);
        UBool   expandflag = expansion > 0;
        setColEIterOffset(coleiter, *start);
        while (expansion > 0) {
            // getting rid of the redundant ce, caused by setOffset.
            // since backward contraction/expansion may have extra ces if we
            // are in the normalization buffer, hasAccentsBeforeMatch would
            // have taken care of it.
            // E.g. the character \u01FA will have an expansion of 3, but if
            // we are only looking for acute and ring \u030A and \u0301, we'll
            // have to skip the first ce in the expansion buffer.
            ucol_next(coleiter, status);
            if (U_FAILURE(*status)) {
                return FALSE;
            }
            if (ucol_getOffset(coleiter) != temp) {
                *start = temp;
                temp  = ucol_getOffset(coleiter);
            }
            expansion --;
        }

        int32_t  *patternce       = strsrch->pattern.ces;
        int32_t   patterncelength = strsrch->pattern.cesLength;
        int32_t   count           = 0;
        while (count < patterncelength) {
            int32_t ce = getCE(strsrch, ucol_next(coleiter, status));
            if (ce == UCOL_IGNORABLE) {
                continue;
            }
            if (expandflag && count == 0 && ucol_getOffset(coleiter) != temp) {
                *start = temp;
                temp   = ucol_getOffset(coleiter);
            }
            if (U_FAILURE(*status) || ce != patternce[count]) {
                (*end) ++;
                *end = getNextUStringSearchBaseOffset(strsrch, *end);
                return FALSE;
            }
            count ++;
        }
    }
    return TRUE;
}

/**
* Checks and sets the match information if found.
* Checks
* <ul>
* <li> the potential match does not repeat the previous match
* <li> boundaries are correct
* <li> exact matches has no extra accents
* <li> identical matchesb
* <li> potential match does not end in the middle of a contraction
* <\ul>
* Otherwise the offset will be shifted to the next character.
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param textoffset offset in the collation element text. the returned value
*        will be the truncated end offset of the match or the new start
*        search offset.
* @param status output error status if any
* @return TRUE if the match is valid, FALSE otherwise
*/
static
inline UBool checkNextExactMatch(UStringSearch *strsrch,
                                 int32_t   *textoffset, UErrorCode *status)
{
    UCollationElements *coleiter = strsrch->textIter;
    int32_t         start    = getColElemIterOffset(coleiter, FALSE);

    if (!checkNextExactContractionMatch(strsrch, &start, textoffset, status)) {
        return FALSE;
    }

    // this totally matches, however we need to check if it is repeating
    if (!isBreakUnit(strsrch, start, *textoffset) ||
        checkRepeatedMatch(strsrch, start, *textoffset) ||
        hasAccentsBeforeMatch(strsrch, start, *textoffset) ||
        !checkIdentical(strsrch, start, *textoffset) ||
        hasAccentsAfterMatch(strsrch, start, *textoffset)) {

        (*textoffset) ++;
        *textoffset = getNextUStringSearchBaseOffset(strsrch, *textoffset);
        return FALSE;
    }

    //Add breakiterator boundary check for primary strength search.
    if (!strsrch->search->breakIter && strsrch->strength == UCOL_PRIMARY) {
        checkBreakBoundary(strsrch, &start, textoffset);
    }

    // totally match, we will get rid of the ending ignorables.
    strsrch->search->matchedIndex  = start;
    strsrch->search->matchedLength = *textoffset - start;
    return TRUE;
}

/**
* Getting the previous base character offset, or the current offset if the
* current character is a base character
* @param text string
* @param textoffset one offset after the current character
* @return the offset of the next character after the base character or the first
*         composed character with accents
*/
static
inline int32_t getPreviousBaseOffset(const UChar       *text,
                                               int32_t  textoffset)
{
    if (textoffset > 0) {
        for (;;) {
            int32_t result = textoffset;
            U16_BACK_1(text, 0, textoffset);
            int32_t temp = textoffset;
            uint16_t fcd = getFCD(text, &temp, result);
            if ((fcd >> SECOND_LAST_BYTE_SHIFT_) == 0) {
                if (fcd & LAST_BYTE_MASK_) {
                    return textoffset;
                }
                return result;
            }
            if (textoffset == 0) {
                return 0;
            }
        }
    }
    return textoffset;
}

/**
* Getting the indexes of the accents that are not blocked in the argument
* accent array
* @param accents array of accents in nfd terminated by a 0.
* @param accentsindex array of indexes of the accents that are not blocked
*/
static
inline int getUnblockedAccentIndex(UChar *accents, int32_t *accentsindex)
{
    int32_t index     = 0;
    int32_t     length    = u_strlen(accents);
    UChar32     codepoint = 0;
    int         cclass    = 0;
    int         result    = 0;
    int32_t temp;
    while (index < length) {
        temp = index;
        U16_NEXT(accents, index, length, codepoint);
        if (u_getCombiningClass(codepoint) != cclass) {
            cclass        = u_getCombiningClass(codepoint);
            accentsindex[result] = temp;
            result ++;
        }
    }
    accentsindex[result] = length;
    return result;
}

/**
* Appends 3 UChar arrays to a destination array.
* Creates a new array if we run out of space. The caller will have to
* manually deallocate the newly allocated array.
* Internal method, status assumed to be success, caller has to check status
* before calling this method. destination not to be NULL and has at least
* size destinationlength.
* @param destination target array
* @param destinationlength target array size, returning the appended length
* @param source1 null-terminated first array
* @param source2 second array
* @param source2length length of second array
* @param source3 null-terminated third array
* @param status error status if any
* @return new destination array, destination if there was no new allocation
*/
static
inline UChar * addToUCharArray(      UChar      *destination,
                                     int32_t    *destinationlength,
                               const UChar      *source1,
                               const UChar      *source2,
                                     int32_t     source2length,
                               const UChar      *source3,
                                     UErrorCode *status)
{
    int32_t source1length = source1 ? u_strlen(source1) : 0;
    int32_t source3length = source3 ? u_strlen(source3) : 0;
    if (*destinationlength < source1length + source2length + source3length +
                                                                           1)
    {
        destination = (UChar *)allocateMemory(
          (source1length + source2length + source3length + 1) * sizeof(UChar),
          status);
        // if error allocating memory, status will be
        // U_MEMORY_ALLOCATION_ERROR
        if (U_FAILURE(*status)) {
            *destinationlength = 0;
            return NULL;
        }
    }
    if (source1length != 0) {
        u_memcpy(destination, source1, source1length);
    }
    if (source2length != 0) {
        uprv_memcpy(destination + source1length, source2,
                    sizeof(UChar) * source2length);
    }
    if (source3length != 0) {
        uprv_memcpy(destination + source1length + source2length, source3,
                    sizeof(UChar) * source3length);
    }
    *destinationlength = source1length + source2length + source3length;
    return destination;
}

/**
* Running through a collation element iterator to see if the contents matches
* pattern in string search data
* @param strsrch string search data
* @param coleiter collation element iterator
* @return TRUE if a match if found, FALSE otherwise
*/
static
inline UBool checkCollationMatch(const UStringSearch      *strsrch,
                                       UCollationElements *coleiter)
{
    int         patternceindex = strsrch->pattern.cesLength;
    int32_t    *patternce      = strsrch->pattern.ces;
    UErrorCode  status = U_ZERO_ERROR;
    while (patternceindex > 0) {
        int32_t ce = getCE(strsrch, ucol_next(coleiter, &status));
        if (ce == UCOL_IGNORABLE) {
            continue;
        }
        if (U_FAILURE(status) || ce != *patternce) {
            return FALSE;
        }
        patternce ++;
        patternceindex --;
    }
    return TRUE;
}

/**
* Rearranges the front accents to try matching.
* Prefix accents in the text will be grouped according to their combining
* class and the groups will be mixed and matched to try find the perfect
* match with the pattern.
* So for instance looking for "\u0301" in "\u030A\u0301\u0325"
* step 1: split "\u030A\u0301" into 6 other type of potential accent substrings
*         "\u030A", "\u0301", "\u0325", "\u030A\u0301", "\u030A\u0325",
*         "\u0301\u0325".
* step 2: check if any of the generated substrings matches the pattern.
* Internal method, status is assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search match
* @param start first offset of the accents to start searching
* @param end start of the last accent set
* @param status output error status if any
* @return USEARCH_DONE if a match is not found, otherwise return the starting
*         offset of the match. Note this start includes all preceding accents.
*/
static
int32_t doNextCanonicalPrefixMatch(UStringSearch *strsrch,
                                       int32_t    start,
                                       int32_t    end,
                                       UErrorCode    *status)
{
    const UChar       *text       = strsrch->search->text;
          int32_t      textlength = strsrch->search->textLength;
          int32_t  tempstart  = start;

    if ((getFCD(text, &tempstart, textlength) & LAST_BYTE_MASK_) == 0) {
        // die... failed at a base character
        return USEARCH_DONE;
    }

    int32_t offset = getNextBaseOffset(text, tempstart, textlength);
    start = getPreviousBaseOffset(text, tempstart);

    UChar       accents[INITIAL_ARRAY_SIZE_];
    // normalizing the offensive string
    unorm_normalize(text + start, offset - start, UNORM_NFD, 0, accents,
                    INITIAL_ARRAY_SIZE_, status);
    if (U_FAILURE(*status)) {
        return USEARCH_DONE;
    }

    int32_t         accentsindex[INITIAL_ARRAY_SIZE_];
    int32_t         accentsize = getUnblockedAccentIndex(accents,
                                                                 accentsindex);
    int32_t         count      = (2 << (accentsize - 1)) - 1;
    UChar               buffer[INITIAL_ARRAY_SIZE_];
    UCollationElements *coleiter   = strsrch->utilIter;
    while (U_SUCCESS(*status) && count > 0) {
        UChar *rearrange = strsrch->canonicalPrefixAccents;
        // copy the base characters
        for (int k = 0; k < accentsindex[0]; k ++) {
            *rearrange ++ = accents[k];
        }
        // forming all possible canonical rearrangement by dropping
        // sets of accents
        for (int i = 0; i <= accentsize - 1; i ++) {
            int32_t mask = 1 << (accentsize - i - 1);
            if (count & mask) {
                for (int j = accentsindex[i]; j < accentsindex[i + 1]; j ++) {
                    *rearrange ++ = accents[j];
                }
            }
        }
        *rearrange = 0;
        int32_t  matchsize = INITIAL_ARRAY_SIZE_;
        UChar   *match     = addToUCharArray(buffer, &matchsize,
                                           strsrch->canonicalPrefixAccents,
                                           strsrch->search->text + offset,
                                           end - offset,
                                           strsrch->canonicalSuffixAccents,
                                           status);

        // if status is a failure, ucol_setText does nothing.
        // run the collator iterator through this match
        ucol_setText(coleiter, match, matchsize, status);
        if (U_SUCCESS(*status)) {
            if (checkCollationMatch(strsrch, coleiter)) {
                if (match != buffer) {
                    uprv_free(match);
                }
                return start;
            }
        }
        count --;
    }
    return USEARCH_DONE;
}

/**
* Gets the offset to the safe point in text before textoffset.
* ie. not the middle of a contraction, swappable characters or supplementary
* characters.
* @param collator collation sata
* @param text string to work with
* @param textoffset offset in string
* @param textlength length of text string
* @return offset to the previous safe character
*/
static
inline uint32_t getPreviousSafeOffset(const UCollator   *collator,
                                      const UChar       *text,
                                            int32_t  textoffset)
{
    int32_t result = textoffset; // first contraction character
    while (result != 0 && ucol_unsafeCP(text[result - 1], collator)) {
        result --;
    }
    if (result != 0) {
        // the first contraction character is consider unsafe here
        result --;
    }
    return result;
}

/**
* Cleaning up after we passed the safe zone
* @param strsrch string search data
* @param safetext safe text array
* @param safebuffer safe text buffer
* @param coleiter collation element iterator for safe text
*/
static
inline void cleanUpSafeText(const UStringSearch *strsrch, UChar *safetext,
                                  UChar         *safebuffer)
{
    if (safetext != safebuffer && safetext != strsrch->canonicalSuffixAccents)
    {
       uprv_free(safetext);
    }
}

/**
* Take the rearranged end accents and tries matching. If match failed at
* a separate preceding set of accents (separated from the rearranged on by
* at least a base character) then we rearrange the preceding accents and
* tries matching again.
* We allow skipping of the ends of the accent set if the ces do not match.
* However if the failure is found before the accent set, it fails.
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param textoffset of the start of the rearranged accent
* @param status output error status if any
* @return USEARCH_DONE if a match is not found, otherwise return the starting
*         offset of the match. Note this start includes all preceding accents.
*/
static
int32_t doNextCanonicalSuffixMatch(UStringSearch *strsrch,
                                       int32_t    textoffset,
                                       UErrorCode    *status)
{
    const UChar              *text           = strsrch->search->text;
    const UCollator          *collator       = strsrch->collator;
          int32_t             safelength     = 0;
          UChar              *safetext;
          int32_t             safetextlength;
          UChar               safebuffer[INITIAL_ARRAY_SIZE_];
          UCollationElements *coleiter       = strsrch->utilIter;
          int32_t         safeoffset     = textoffset;

    if (textoffset != 0 && ucol_unsafeCP(strsrch->canonicalSuffixAccents[0],
                                         collator)) {
        safeoffset     = getPreviousSafeOffset(collator, text, textoffset);
        safelength     = textoffset - safeoffset;
        safetextlength = INITIAL_ARRAY_SIZE_;
        safetext       = addToUCharArray(safebuffer, &safetextlength, NULL,
                                         text + safeoffset, safelength,
                                         strsrch->canonicalSuffixAccents,
                                         status);
    }
    else {
        safetextlength = u_strlen(strsrch->canonicalSuffixAccents);
        safetext       = strsrch->canonicalSuffixAccents;
    }

    // if status is a failure, ucol_setText does nothing
    ucol_setText(coleiter, safetext, safetextlength, status);
    // status checked in loop below

    int32_t  *ce        = strsrch->pattern.ces;
    int32_t   celength  = strsrch->pattern.cesLength;
    int       ceindex   = celength - 1;
    UBool     isSafe    = TRUE; // indication flag for position in safe zone

    while (ceindex >= 0) {
        int32_t textce = ucol_previous(coleiter, status);
        if (U_FAILURE(*status)) {
            if (isSafe) {
                cleanUpSafeText(strsrch, safetext, safebuffer);
            }
            return USEARCH_DONE;
        }
        if (textce == UCOL_NULLORDER) {
            // check if we have passed the safe buffer
            if (coleiter == strsrch->textIter) {
                cleanUpSafeText(strsrch, safetext, safebuffer);
                return USEARCH_DONE;
            }
            cleanUpSafeText(strsrch, safetext, safebuffer);
            safetext = safebuffer;
            coleiter = strsrch->textIter;
            setColEIterOffset(coleiter, safeoffset);
            // status checked at the start of the loop
            isSafe = FALSE;
            continue;
        }
        textce = getCE(strsrch, textce);
        if (textce != UCOL_IGNORABLE && textce != ce[ceindex]) {
            // do the beginning stuff
            int32_t failedoffset = getColElemIterOffset(coleiter, FALSE);
            if (isSafe && failedoffset >= safelength) {
                // alas... no hope. failed at rearranged accent set
                cleanUpSafeText(strsrch, safetext, safebuffer);
                return USEARCH_DONE;
            }
            else {
                if (isSafe) {
                    failedoffset += safeoffset;
                    cleanUpSafeText(strsrch, safetext, safebuffer);
                }

                // try rearranging the front accents
                int32_t result = doNextCanonicalPrefixMatch(strsrch,
                                        failedoffset, textoffset, status);
                if (result != USEARCH_DONE) {
                    // if status is a failure, ucol_setOffset does nothing
                    setColEIterOffset(strsrch->textIter, result);
                }
                if (U_FAILURE(*status)) {
                    return USEARCH_DONE;
                }
                return result;
            }
        }
        if (textce == ce[ceindex]) {
            ceindex --;
        }
    }
    // set offset here
    if (isSafe) {
        int32_t result     = getColElemIterOffset(coleiter, FALSE);
        // sets the text iterator here with the correct expansion and offset
        int32_t    leftoverces = getExpansionPrefix(coleiter);
        cleanUpSafeText(strsrch, safetext, safebuffer);
        if (result >= safelength) {
            result = textoffset;
        }
        else {
            result += safeoffset;
        }
        setColEIterOffset(strsrch->textIter, result);
        strsrch->textIter->iteratordata_.toReturn =
                       setExpansionPrefix(strsrch->textIter, leftoverces);
        return result;
    }

    return ucol_getOffset(coleiter);
}

/**
* Trying out the substring and sees if it can be a canonical match.
* This will try normalizing the end accents and arranging them into canonical
* equivalents and check their corresponding ces with the pattern ce.
* Suffix accents in the text will be grouped according to their combining
* class and the groups will be mixed and matched to try find the perfect
* match with the pattern.
* So for instance looking for "\u0301" in "\u030A\u0301\u0325"
* step 1: split "\u030A\u0301" into 6 other type of potential accent substrings
*         "\u030A", "\u0301", "\u0325", "\u030A\u0301", "\u030A\u0325",
*         "\u0301\u0325".
* step 2: check if any of the generated substrings matches the pattern.
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param textoffset end offset in the collation element text that ends with
*                   the accents to be rearranged
* @param status error status if any
* @return TRUE if the match is valid, FALSE otherwise
*/
static
UBool doNextCanonicalMatch(UStringSearch *strsrch,
                           int32_t    textoffset,
                           UErrorCode    *status)
{
    const UChar       *text = strsrch->search->text;
          int32_t  temp = textoffset;
    U16_BACK_1(text, 0, temp);
    if ((getFCD(text, &temp, textoffset) & LAST_BYTE_MASK_) == 0) {
        UCollationElements *coleiter = strsrch->textIter;
        int32_t         offset   = getColElemIterOffset(coleiter, FALSE);
        if (strsrch->pattern.hasPrefixAccents) {
            offset = doNextCanonicalPrefixMatch(strsrch, offset, textoffset,
                                                status);
            if (U_SUCCESS(*status) && offset != USEARCH_DONE) {
                setColEIterOffset(coleiter, offset);
                return TRUE;
            }
        }
        return FALSE;
    }

    if (!strsrch->pattern.hasSuffixAccents) {
        return FALSE;
    }

    UChar       accents[INITIAL_ARRAY_SIZE_];
    // offset to the last base character in substring to search
    int32_t baseoffset = getPreviousBaseOffset(text, textoffset);
    // normalizing the offensive string
    unorm_normalize(text + baseoffset, textoffset - baseoffset, UNORM_NFD,
                               0, accents, INITIAL_ARRAY_SIZE_, status);
    // status checked in loop below

    int32_t accentsindex[INITIAL_ARRAY_SIZE_];
    int32_t size = getUnblockedAccentIndex(accents, accentsindex);

    // 2 power n - 1 plus the full set of accents
    int32_t  count = (2 << (size - 1)) - 1;
    while (U_SUCCESS(*status) && count > 0) {
        UChar *rearrange = strsrch->canonicalSuffixAccents;
        // copy the base characters
        for (int k = 0; k < accentsindex[0]; k ++) {
            *rearrange ++ = accents[k];
        }
        // forming all possible canonical rearrangement by dropping
        // sets of accents
        for (int i = 0; i <= size - 1; i ++) {
            int32_t mask = 1 << (size - i - 1);
            if (count & mask) {
                for (int j = accentsindex[i]; j < accentsindex[i + 1]; j ++) {
                    *rearrange ++ = accents[j];
                }
            }
        }
        *rearrange = 0;
        int32_t offset = doNextCanonicalSuffixMatch(strsrch, baseoffset,
                                                        status);
        if (offset != USEARCH_DONE) {
            return TRUE; // match found
        }
        count --;
    }
    return FALSE;
}

/**
* Gets the previous base character offset depending on the string search
* pattern data
* @param strsrch string search data
* @param textoffset current offset, current character
* @return the offset of the next character after this base character or itself
*         if it is a composed character with accents
*/
static
inline int32_t getPreviousUStringSearchBaseOffset(UStringSearch *strsrch,
                                                      int32_t textoffset)
{
    if (strsrch->pattern.hasPrefixAccents && textoffset > 0) {
        const UChar       *text = strsrch->search->text;
              int32_t  offset = textoffset;
        if (getFCD(text, &offset, strsrch->search->textLength) >>
                                                   SECOND_LAST_BYTE_SHIFT_) {
            return getPreviousBaseOffset(text, textoffset);
        }
    }
    return textoffset;
}

/**
* Checks match for contraction.
* If the match ends with a partial contraction we fail.
* If the match starts too far off (because of backwards iteration) we try to
* chip off the extra characters
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param start offset of potential match, to be modified if necessary
* @param end offset of potential match, to be modified if necessary
* @param status output error status if any
* @return TRUE if match passes the contraction test, FALSE otherwise
*/
static
UBool checkNextCanonicalContractionMatch(UStringSearch *strsrch,
                                         int32_t   *start,
                                         int32_t   *end,
                                         UErrorCode    *status)
{
          UCollationElements *coleiter   = strsrch->textIter;
          int32_t             textlength = strsrch->search->textLength;
          int32_t         temp       = *start;
    const UCollator          *collator   = strsrch->collator;
    const UChar              *text       = strsrch->search->text;
    // This part checks if either ends of the match contains potential
    // contraction. If so we'll have to iterate through them
    if ((*end < textlength && ucol_unsafeCP(text[*end], collator)) ||
        (*start + 1 < textlength
         && ucol_unsafeCP(text[*start + 1], collator))) {
        int32_t expansion  = getExpansionPrefix(coleiter);
        UBool   expandflag = expansion > 0;
        setColEIterOffset(coleiter, *start);
        while (expansion > 0) {
            // getting rid of the redundant ce, caused by setOffset.
            // since backward contraction/expansion may have extra ces if we
            // are in the normalization buffer, hasAccentsBeforeMatch would
            // have taken care of it.
            // E.g. the character \u01FA will have an expansion of 3, but if
            // we are only looking for acute and ring \u030A and \u0301, we'll
            // have to skip the first ce in the expansion buffer.
            ucol_next(coleiter, status);
            if (U_FAILURE(*status)) {
                return FALSE;
            }
            if (ucol_getOffset(coleiter) != temp) {
                *start = temp;
                temp  = ucol_getOffset(coleiter);
            }
            expansion --;
        }

        int32_t  *patternce       = strsrch->pattern.ces;
        int32_t   patterncelength = strsrch->pattern.cesLength;
        int32_t   count           = 0;
        int32_t   textlength      = strsrch->search->textLength;
        while (count < patterncelength) {
            int32_t ce = getCE(strsrch, ucol_next(coleiter, status));
            // status checked below, note that if status is a failure
            // ucol_next returns UCOL_NULLORDER
            if (ce == UCOL_IGNORABLE) {
                continue;
            }
            if (expandflag && count == 0 && ucol_getOffset(coleiter) != temp) {
                *start = temp;
                temp   = ucol_getOffset(coleiter);
            }

            if (count == 0 && ce != patternce[0]) {
                // accents may have extra starting ces, this occurs when a
                // pure accent pattern is matched without rearrangement
                // text \u0325\u0300 and looking for \u0300
                int32_t expected = patternce[0];
                if (getFCD(text, start, textlength) & LAST_BYTE_MASK_) {
                    ce = getCE(strsrch, ucol_next(coleiter, status));
                    while (U_SUCCESS(*status) && ce != expected &&
                           ce != UCOL_NULLORDER &&
                           ucol_getOffset(coleiter) <= *end) {
                        ce = getCE(strsrch, ucol_next(coleiter, status));
                    }
                }
            }
            if (U_FAILURE(*status) || ce != patternce[count]) {
                (*end) ++;
                *end = getNextUStringSearchBaseOffset(strsrch, *end);
                return FALSE;
            }
            count ++;
        }
    }
    return TRUE;
}

/**
* Checks and sets the match information if found.
* Checks
* <ul>
* <li> the potential match does not repeat the previous match
* <li> boundaries are correct
* <li> potential match does not end in the middle of a contraction
* <li> identical matches
* <\ul>
* Otherwise the offset will be shifted to the next character.
* Internal method, status assumed to be success, caller has to check the
* status before calling this method.
* @param strsrch string search data
* @param textoffset offset in the collation element text. the returned value
*        will be the truncated end offset of the match or the new start
*        search offset.
* @param status output error status if any
* @return TRUE if the match is valid, FALSE otherwise
*/
static
inline UBool checkNextCanonicalMatch(UStringSearch *strsrch,
                                     int32_t   *textoffset,
                                     UErrorCode    *status)
{
    // to ensure that the start and ends are not composite characters
    UCollationElements *coleiter = strsrch->textIter;
    // if we have a canonical accent match
    if ((strsrch->pattern.hasSuffixAccents &&
        strsrch->canonicalSuffixAccents[0]) ||
        (strsrch->pattern.hasPrefixAccents &&
        strsrch->canonicalPrefixAccents[0])) {
        strsrch->search->matchedIndex  = getPreviousUStringSearchBaseOffset(
                                                    strsrch,
                                                    ucol_getOffset(coleiter));
        strsrch->search->matchedLength = *textoffset -
                                                strsrch->search->matchedIndex;
        return TRUE;
    }

    int32_t start = getColElemIterOffset(coleiter, FALSE);
    if (!checkNextCanonicalContractionMatch(strsrch, &start, textoffset,
                                            status) || U_FAILURE(*status)) {
        return FALSE;
    }

    start = getPreviousUStringSearchBaseOffset(strsrch, start);
    // this totally matches, however we need to check if it is repeating
    if (checkRepeatedMatch(strsrch, start, *textoffset) ||
        !isBreakUnit(strsrch, start, *textoffset) ||
        !checkIdentical(strsrch, start, *textoffset)) {
        (*textoffset) ++;
        *textoffset = getNextBaseOffset(strsrch->search->text, *textoffset,
                                        strsrch->search->textLength);
        return FALSE;
    }

    strsrch->search->matchedIndex  = start;
    strsrch->search->matchedLength = *textoffset - start;
    return TRUE;
}

/**
* Shifting the collation element iterator position forward to prepare for
* a preceding match. If the first character is a unsafe character, we'll only
* shift by 1 to capture contractions, normalization etc.
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param text strsrch string search data
* @param textoffset start text position to do search
* @param ce the text ce which failed the match.
* @param patternceindex index of the ce within the pattern ce buffer which
*        failed the match
* @return final offset
*/
static
inline int32_t reverseShift(UStringSearch *strsrch,
                                int32_t    textoffset,
                                int32_t       ce,
                                int32_t        patternceindex)
{
    if (strsrch->search->isOverlap) {
        if (textoffset != strsrch->search->textLength) {
            textoffset --;
        }
        else {
            textoffset -= strsrch->pattern.defaultShiftSize;
        }
    }
    else {
        if (ce != UCOL_NULLORDER) {
            int32_t shift = strsrch->pattern.backShift[hashFromCE32(ce)];

            // this is to adjust for characters in the middle of the substring
            // for matching that failed.
            int32_t adjust = patternceindex;
            if (adjust > 1 && shift > adjust) {
                shift -= adjust - 1;
            }
            textoffset -= shift;
        }
        else {
            textoffset -= strsrch->pattern.defaultShiftSize;
        }
    }
    textoffset = getPreviousUStringSearchBaseOffset(strsrch, textoffset);
    return textoffset;
}

/**
* Checks match for contraction.
* If the match starts with a partial contraction we fail.
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param start offset of potential match, to be modified if necessary
* @param end offset of potential match, to be modified if necessary
* @param status output error status if any
* @return TRUE if match passes the contraction test, FALSE otherwise
*/
static
UBool checkPreviousExactContractionMatch(UStringSearch *strsrch,
                                     int32_t   *start,
                                     int32_t   *end, UErrorCode  *status)
{
          UCollationElements *coleiter   = strsrch->textIter;
          int32_t             textlength = strsrch->search->textLength;
          int32_t             temp       = *end;
    const UCollator          *collator   = strsrch->collator;
    const UChar              *text       = strsrch->search->text;
    // This part checks if either if the start of the match contains potential
    // contraction. If so we'll have to iterate through them
    // Since we used ucol_next while previously looking for the potential
    // match, this guarantees that our end will not be a partial contraction,
    // or a partial supplementary character.
    if (*start < textlength && ucol_unsafeCP(text[*start], collator)) {
        int32_t expansion  = getExpansionSuffix(coleiter);
        UBool   expandflag = expansion > 0;
        setColEIterOffset(coleiter, *end);
        while (U_SUCCESS(*status) && expansion > 0) {
            // getting rid of the redundant ce
            // since forward contraction/expansion may have extra ces
            // if we are in the normalization buffer, hasAccentsBeforeMatch
            // would have taken care of it.
            // E.g. the character \u01FA will have an expansion of 3, but if
            // we are only looking for A ring A\u030A, we'll have to skip the
            // last ce in the expansion buffer
            ucol_previous(coleiter, status);
            if (U_FAILURE(*status)) {
                return FALSE;
            }
            if (ucol_getOffset(coleiter) != temp) {
                *end = temp;
                temp  = ucol_getOffset(coleiter);
            }
            expansion --;
        }

        int32_t  *patternce       = strsrch->pattern.ces;
        int32_t   patterncelength = strsrch->pattern.cesLength;
        int32_t   count           = patterncelength;
        while (count > 0) {
            int32_t ce = getCE(strsrch, ucol_previous(coleiter, status));
            // status checked below, note that if status is a failure
            // ucol_previous returns UCOL_NULLORDER
            if (ce == UCOL_IGNORABLE) {
                continue;
            }
            if (expandflag && count == 0 &&
                getColElemIterOffset(coleiter, FALSE) != temp) {
                *end = temp;
                temp  = ucol_getOffset(coleiter);
            }
            if (U_FAILURE(*status) || ce != patternce[count - 1]) {
                (*start) --;
                *start = getPreviousBaseOffset(text, *start);
                return FALSE;
            }
            count --;
        }
    }
    return TRUE;
}

/**
* Checks and sets the match information if found.
* Checks
* <ul>
* <li> the current match does not repeat the last match
* <li> boundaries are correct
* <li> exact matches has no extra accents
* <li> identical matches
* <\ul>
* Otherwise the offset will be shifted to the preceding character.
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param collator
* @param coleiter collation element iterator
* @param text string
* @param textoffset offset in the collation element text. the returned value
*        will be the truncated start offset of the match or the new start
*        search offset.
* @param status output error status if any
* @return TRUE if the match is valid, FALSE otherwise
*/
static
inline UBool checkPreviousExactMatch(UStringSearch *strsrch,
                                     int32_t   *textoffset,
                                     UErrorCode    *status)
{
    // to ensure that the start and ends are not composite characters
    int32_t end = ucol_getOffset(strsrch->textIter);
    if (!checkPreviousExactContractionMatch(strsrch, textoffset, &end, status)
        || U_FAILURE(*status)) {
            return FALSE;
    }

    // this totally matches, however we need to check if it is repeating
    // the old match
    if (checkRepeatedMatch(strsrch, *textoffset, end) ||
        !isBreakUnit(strsrch, *textoffset, end) ||
        hasAccentsBeforeMatch(strsrch, *textoffset, end) ||
        !checkIdentical(strsrch, *textoffset, end) ||
        hasAccentsAfterMatch(strsrch, *textoffset, end)) {
        (*textoffset) --;
        *textoffset = getPreviousBaseOffset(strsrch->search->text,
                                            *textoffset);
        return FALSE;
    }

    //Add breakiterator boundary check for primary strength search.
    if (!strsrch->search->breakIter && strsrch->strength == UCOL_PRIMARY) {
        checkBreakBoundary(strsrch, textoffset, &end);
    }

    strsrch->search->matchedIndex = *textoffset;
    strsrch->search->matchedLength = end - *textoffset;
    return TRUE;
}

/**
* Rearranges the end accents to try matching.
* Suffix accents in the text will be grouped according to their combining
* class and the groups will be mixed and matched to try find the perfect
* match with the pattern.
* So for instance looking for "\u0301" in "\u030A\u0301\u0325"
* step 1: split "\u030A\u0301" into 6 other type of potential accent substrings
*         "\u030A", "\u0301", "\u0325", "\u030A\u0301", "\u030A\u0325",
*         "\u0301\u0325".
* step 2: check if any of the generated substrings matches the pattern.
* Internal method, status assumed to be success, user has to check status
* before calling this method.
* @param strsrch string search match
* @param start offset of the first base character
* @param end start of the last accent set
* @param status only error status if any
* @return USEARCH_DONE if a match is not found, otherwise return the ending
*         offset of the match. Note this start includes all following accents.
*/
static
int32_t doPreviousCanonicalSuffixMatch(UStringSearch *strsrch,
                                           int32_t    start,
                                           int32_t    end,
                                           UErrorCode    *status)
{
    const UChar       *text       = strsrch->search->text;
          int32_t  tempend    = end;

    U16_BACK_1(text, 0, tempend);
    if (!(getFCD(text, &tempend, strsrch->search->textLength) &
                                                           LAST_BYTE_MASK_)) {
        // die... failed at a base character
        return USEARCH_DONE;
    }
    end = getNextBaseOffset(text, end, strsrch->search->textLength);

    if (U_SUCCESS(*status)) {
        UChar       accents[INITIAL_ARRAY_SIZE_];
        int32_t offset = getPreviousBaseOffset(text, end);
        // normalizing the offensive string
        unorm_normalize(text + offset, end - offset, UNORM_NFD, 0, accents,
                        INITIAL_ARRAY_SIZE_, status);

        int32_t         accentsindex[INITIAL_ARRAY_SIZE_];
        int32_t         accentsize = getUnblockedAccentIndex(accents,
                                                         accentsindex);
        int32_t         count      = (2 << (accentsize - 1)) - 1;
        UChar               buffer[INITIAL_ARRAY_SIZE_];
        UCollationElements *coleiter = strsrch->utilIter;
        while (U_SUCCESS(*status) && count > 0) {
            UChar *rearrange = strsrch->canonicalSuffixAccents;
            // copy the base characters
            for (int k = 0; k < accentsindex[0]; k ++) {
                *rearrange ++ = accents[k];
            }
            // forming all possible canonical rearrangement by dropping
            // sets of accents
            for (int i = 0; i <= accentsize - 1; i ++) {
                int32_t mask = 1 << (accentsize - i - 1);
                if (count & mask) {
                    for (int j = accentsindex[i]; j < accentsindex[i + 1]; j ++) {
                        *rearrange ++ = accents[j];
                    }
                }
            }
            *rearrange = 0;
            int32_t  matchsize = INITIAL_ARRAY_SIZE_;
            UChar   *match     = addToUCharArray(buffer, &matchsize,
                                           strsrch->canonicalPrefixAccents,
                                           strsrch->search->text + start,
                                           offset - start,
                                           strsrch->canonicalSuffixAccents,
                                           status);

            // run the collator iterator through this match
            // if status is a failure ucol_setText does nothing
            ucol_setText(coleiter, match, matchsize, status);
            if (U_SUCCESS(*status)) {
                if (checkCollationMatch(strsrch, coleiter)) {
                    if (match != buffer) {
                        uprv_free(match);
                    }
                    return end;
                }
            }
            count --;
        }
    }
    return USEARCH_DONE;
}

/**
* Take the rearranged start accents and tries matching. If match failed at
* a separate following set of accents (separated from the rearranged on by
* at least a base character) then we rearrange the preceding accents and
* tries matching again.
* We allow skipping of the ends of the accent set if the ces do not match.
* However if the failure is found before the accent set, it fails.
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param textoffset of the ends of the rearranged accent
* @param status output error status if any
* @return USEARCH_DONE if a match is not found, otherwise return the ending
*         offset of the match. Note this start includes all following accents.
*/
static
int32_t doPreviousCanonicalPrefixMatch(UStringSearch *strsrch,
                                           int32_t    textoffset,
                                           UErrorCode    *status)
{
    const UChar       *text       = strsrch->search->text;
    const UCollator   *collator   = strsrch->collator;
          int32_t      safelength = 0;
          UChar       *safetext;
          int32_t      safetextlength;
          UChar        safebuffer[INITIAL_ARRAY_SIZE_];
          int32_t  safeoffset = textoffset;

    if (textoffset &&
        ucol_unsafeCP(strsrch->canonicalPrefixAccents[
                                 u_strlen(strsrch->canonicalPrefixAccents) - 1
                                         ], collator)) {
        safeoffset     = getNextSafeOffset(collator, text, textoffset,
                                           strsrch->search->textLength);
        safelength     = safeoffset - textoffset;
        safetextlength = INITIAL_ARRAY_SIZE_;
        safetext       = addToUCharArray(safebuffer, &safetextlength,
                                         strsrch->canonicalPrefixAccents,
                                         text + textoffset, safelength,
                                         NULL, status);
    }
    else {
        safetextlength = u_strlen(strsrch->canonicalPrefixAccents);
        safetext       = strsrch->canonicalPrefixAccents;
    }

    UCollationElements *coleiter = strsrch->utilIter;
     // if status is a failure, ucol_setText does nothing
    ucol_setText(coleiter, safetext, safetextlength, status);
    // status checked in loop below

    int32_t  *ce           = strsrch->pattern.ces;
    int32_t   celength     = strsrch->pattern.cesLength;
    int       ceindex      = 0;
    UBool     isSafe       = TRUE; // safe zone indication flag for position
    int32_t   prefixlength = u_strlen(strsrch->canonicalPrefixAccents);

    while (ceindex < celength) {
        int32_t textce = ucol_next(coleiter, status);
        if (U_FAILURE(*status)) {
            if (isSafe) {
                cleanUpSafeText(strsrch, safetext, safebuffer);
            }
            return USEARCH_DONE;
        }
        if (textce == UCOL_NULLORDER) {
            // check if we have passed the safe buffer
            if (coleiter == strsrch->textIter) {
                cleanUpSafeText(strsrch, safetext, safebuffer);
                return USEARCH_DONE;
            }
            cleanUpSafeText(strsrch, safetext, safebuffer);
            safetext = safebuffer;
            coleiter = strsrch->textIter;
            setColEIterOffset(coleiter, safeoffset);
            // status checked at the start of the loop
            isSafe = FALSE;
            continue;
        }
        textce = getCE(strsrch, textce);
        if (textce != UCOL_IGNORABLE && textce != ce[ceindex]) {
            // do the beginning stuff
            int32_t failedoffset = ucol_getOffset(coleiter);
            if (isSafe && failedoffset <= prefixlength) {
                // alas... no hope. failed at rearranged accent set
                cleanUpSafeText(strsrch, safetext, safebuffer);
                return USEARCH_DONE;
            }
            else {
                if (isSafe) {
                    failedoffset = safeoffset - failedoffset;
                    cleanUpSafeText(strsrch, safetext, safebuffer);
                }

                // try rearranging the end accents
                int32_t result = doPreviousCanonicalSuffixMatch(strsrch,
                                        textoffset, failedoffset, status);
                if (result != USEARCH_DONE) {
                    // if status is a failure, ucol_setOffset does nothing
                    setColEIterOffset(strsrch->textIter, result);
                }
                if (U_FAILURE(*status)) {
                    return USEARCH_DONE;
                }
                return result;
            }
        }
        if (textce == ce[ceindex]) {
            ceindex ++;
        }
    }
    // set offset here
    if (isSafe) {
        int32_t result      = ucol_getOffset(coleiter);
        // sets the text iterator here with the correct expansion and offset
        int32_t     leftoverces = getExpansionSuffix(coleiter);
        cleanUpSafeText(strsrch, safetext, safebuffer);
        if (result <= prefixlength) {
            result = textoffset;
        }
        else {
            result = textoffset + (safeoffset - result);
        }
        setColEIterOffset(strsrch->textIter, result);
        setExpansionSuffix(strsrch->textIter, leftoverces);
        return result;
    }

    return ucol_getOffset(coleiter);
}

/**
* Trying out the substring and sees if it can be a canonical match.
* This will try normalizing the starting accents and arranging them into
* canonical equivalents and check their corresponding ces with the pattern ce.
* Prefix accents in the text will be grouped according to their combining
* class and the groups will be mixed and matched to try find the perfect
* match with the pattern.
* So for instance looking for "\u0301" in "\u030A\u0301\u0325"
* step 1: split "\u030A\u0301" into 6 other type of potential accent substrings
*         "\u030A", "\u0301", "\u0325", "\u030A\u0301", "\u030A\u0325",
*         "\u0301\u0325".
* step 2: check if any of the generated substrings matches the pattern.
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param textoffset start offset in the collation element text that starts
*                   with the accents to be rearranged
* @param status output error status if any
* @return TRUE if the match is valid, FALSE otherwise
*/
static
UBool doPreviousCanonicalMatch(UStringSearch *strsrch,
                               int32_t    textoffset,
                               UErrorCode    *status)
{
    const UChar       *text       = strsrch->search->text;
          int32_t  temp       = textoffset;
          int32_t      textlength = strsrch->search->textLength;
    if ((getFCD(text, &temp, textlength) >> SECOND_LAST_BYTE_SHIFT_) == 0) {
        UCollationElements *coleiter = strsrch->textIter;
        int32_t         offset   = ucol_getOffset(coleiter);
        if (strsrch->pattern.hasSuffixAccents) {
            offset = doPreviousCanonicalSuffixMatch(strsrch, textoffset,
                                                    offset, status);
            if (U_SUCCESS(*status) && offset != USEARCH_DONE) {
                setColEIterOffset(coleiter, offset);
                return TRUE;
            }
        }
        return FALSE;
    }

    if (!strsrch->pattern.hasPrefixAccents) {
        return FALSE;
    }

    UChar       accents[INITIAL_ARRAY_SIZE_];
    // offset to the last base character in substring to search
    int32_t baseoffset = getNextBaseOffset(text, textoffset, textlength);
    // normalizing the offensive string
    unorm_normalize(text + textoffset, baseoffset - textoffset, UNORM_NFD,
                               0, accents, INITIAL_ARRAY_SIZE_, status);
    // status checked in loop

    int32_t accentsindex[INITIAL_ARRAY_SIZE_];
    int32_t size = getUnblockedAccentIndex(accents, accentsindex);

    // 2 power n - 1 plus the full set of accents
    int32_t  count = (2 << (size - 1)) - 1;
    while (U_SUCCESS(*status) && count > 0) {
        UChar *rearrange = strsrch->canonicalPrefixAccents;
        // copy the base characters
        for (int k = 0; k < accentsindex[0]; k ++) {
            *rearrange ++ = accents[k];
        }
        // forming all possible canonical rearrangement by dropping
        // sets of accents
        for (int i = 0; i <= size - 1; i ++) {
            int32_t mask = 1 << (size - i - 1);
            if (count & mask) {
                for (int j = accentsindex[i]; j < accentsindex[i + 1]; j ++) {
                    *rearrange ++ = accents[j];
                }
            }
        }
        *rearrange = 0;
        int32_t offset = doPreviousCanonicalPrefixMatch(strsrch,
                                                          baseoffset, status);
        if (offset != USEARCH_DONE) {
            return TRUE; // match found
        }
        count --;
    }
    return FALSE;
}

/**
* Checks match for contraction.
* If the match starts with a partial contraction we fail.
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param start offset of potential match, to be modified if necessary
* @param end offset of potential match, to be modified if necessary
* @param status only error status if any
* @return TRUE if match passes the contraction test, FALSE otherwise
*/
static
UBool checkPreviousCanonicalContractionMatch(UStringSearch *strsrch,
                                     int32_t   *start,
                                     int32_t   *end, UErrorCode  *status)
{
          UCollationElements *coleiter   = strsrch->textIter;
          int32_t             textlength = strsrch->search->textLength;
          int32_t         temp       = *end;
    const UCollator          *collator   = strsrch->collator;
    const UChar              *text       = strsrch->search->text;
    // This part checks if either if the start of the match contains potential
    // contraction. If so we'll have to iterate through them
    // Since we used ucol_next while previously looking for the potential
    // match, this guarantees that our end will not be a partial contraction,
    // or a partial supplementary character.
    if (*start < textlength && ucol_unsafeCP(text[*start], collator)) {
        int32_t expansion  = getExpansionSuffix(coleiter);
        UBool   expandflag = expansion > 0;
        setColEIterOffset(coleiter, *end);
        while (expansion > 0) {
            // getting rid of the redundant ce
            // since forward contraction/expansion may have extra ces
            // if we are in the normalization buffer, hasAccentsBeforeMatch
            // would have taken care of it.
            // E.g. the character \u01FA will have an expansion of 3, but if
            // we are only looking for A ring A\u030A, we'll have to skip the
            // last ce in the expansion buffer
            ucol_previous(coleiter, status);
            if (U_FAILURE(*status)) {
                return FALSE;
            }
            if (ucol_getOffset(coleiter) != temp) {
                *end = temp;
                temp  = ucol_getOffset(coleiter);
            }
            expansion --;
        }

        int32_t  *patternce       = strsrch->pattern.ces;
        int32_t   patterncelength = strsrch->pattern.cesLength;
        int32_t   count           = patterncelength;
        while (count > 0) {
            int32_t ce = getCE(strsrch, ucol_previous(coleiter, status));
            // status checked below, note that if status is a failure
            // ucol_previous returns UCOL_NULLORDER
            if (ce == UCOL_IGNORABLE) {
                continue;
            }
            if (expandflag && count == 0 &&
                getColElemIterOffset(coleiter, FALSE) != temp) {
                *end = temp;
                temp  = ucol_getOffset(coleiter);
            }
            if (count == patterncelength &&
                ce != patternce[patterncelength - 1]) {
                // accents may have extra starting ces, this occurs when a
                // pure accent pattern is matched without rearrangement
                int32_t    expected = patternce[patterncelength - 1];
                U16_BACK_1(text, 0, *end);
                if (getFCD(text, end, textlength) & LAST_BYTE_MASK_) {
                    ce = getCE(strsrch, ucol_previous(coleiter, status));
                    while (U_SUCCESS(*status) && ce != expected &&
                           ce != UCOL_NULLORDER &&
                           ucol_getOffset(coleiter) <= *start) {
                        ce = getCE(strsrch, ucol_previous(coleiter, status));
                    }
                }
            }
            if (U_FAILURE(*status) || ce != patternce[count - 1]) {
                (*start) --;
                *start = getPreviousBaseOffset(text, *start);
                return FALSE;
            }
            count --;
        }
    }
    return TRUE;
}

/**
* Checks and sets the match information if found.
* Checks
* <ul>
* <li> the potential match does not repeat the previous match
* <li> boundaries are correct
* <li> potential match does not end in the middle of a contraction
* <li> identical matches
* <\ul>
* Otherwise the offset will be shifted to the next character.
* Internal method, status assumed to be success, caller has to check status
* before calling this method.
* @param strsrch string search data
* @param textoffset offset in the collation element text. the returned value
*        will be the truncated start offset of the match or the new start
*        search offset.
* @param status only error status if any
* @return TRUE if the match is valid, FALSE otherwise
*/
static
inline UBool checkPreviousCanonicalMatch(UStringSearch *strsrch,
                                         int32_t   *textoffset,
                                         UErrorCode    *status)
{
    // to ensure that the start and ends are not composite characters
    UCollationElements *coleiter = strsrch->textIter;
    // if we have a canonical accent match
    if ((strsrch->pattern.hasSuffixAccents &&
        strsrch->canonicalSuffixAccents[0]) ||
        (strsrch->pattern.hasPrefixAccents &&
        strsrch->canonicalPrefixAccents[0])) {
        strsrch->search->matchedIndex  = *textoffset;
        strsrch->search->matchedLength =
            getNextUStringSearchBaseOffset(strsrch,
                                      getColElemIterOffset(coleiter, FALSE))
            - *textoffset;
        return TRUE;
    }

    int32_t end = ucol_getOffset(coleiter);
    if (!checkPreviousCanonicalContractionMatch(strsrch, textoffset, &end,
                                                status) ||
         U_FAILURE(*status)) {
        return FALSE;
    }

    end = getNextUStringSearchBaseOffset(strsrch, end);
    // this totally matches, however we need to check if it is repeating
    if (checkRepeatedMatch(strsrch, *textoffset, end) ||
        !isBreakUnit(strsrch, *textoffset, end) ||
        !checkIdentical(strsrch, *textoffset, end)) {
        (*textoffset) --;
        *textoffset = getPreviousBaseOffset(strsrch->search->text,
                                            *textoffset);
        return FALSE;
    }

    strsrch->search->matchedIndex  = *textoffset;
    strsrch->search->matchedLength = end - *textoffset;
    return TRUE;
}
#endif // #if BOYER_MOORE

// constructors and destructor -------------------------------------------

U_CAPI UStringSearch * U_EXPORT2 usearch_open(const UChar *pattern,
                                          int32_t         patternlength,
                                    const UChar          *text,
                                          int32_t         textlength,
                                    const char           *locale,
                                          UBreakIterator *breakiter,
                                          UErrorCode     *status)
{
    if (U_FAILURE(*status)) {
        return NULL;
    }
#if UCONFIG_NO_BREAK_ITERATION
    if (breakiter != NULL) {
        *status = U_UNSUPPORTED_ERROR;
        return NULL;
    }
#endif
    if (locale) {
        // ucol_open internally checks for status
        UCollator     *collator = ucol_open(locale, status);
        // pattern, text checks are done in usearch_openFromCollator
        UStringSearch *result   = usearch_openFromCollator(pattern,
                                              patternlength, text, textlength,
                                              collator, breakiter, status);

        if (result == NULL || U_FAILURE(*status)) {
            if (collator) {
                ucol_close(collator);
            }
            return NULL;
        }
        else {
            result->ownCollator = TRUE;
        }
        return result;
    }
    *status = U_ILLEGAL_ARGUMENT_ERROR;
    return NULL;
}

U_CAPI UStringSearch * U_EXPORT2 usearch_openFromCollator(
                                  const UChar          *pattern,
                                        int32_t         patternlength,
                                  const UChar          *text,
                                        int32_t         textlength,
                                  const UCollator      *collator,
                                        UBreakIterator *breakiter,
                                        UErrorCode     *status)
{
    if (U_FAILURE(*status)) {
        return NULL;
    }
#if UCONFIG_NO_BREAK_ITERATION
    if (breakiter != NULL) {
        *status = U_UNSUPPORTED_ERROR;
        return NULL;
    }
#endif
    if (pattern == NULL || text == NULL || collator == NULL) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    // string search does not really work when numeric collation is turned on
    if(ucol_getAttribute(collator, UCOL_NUMERIC_COLLATION, status) == UCOL_ON) {
        *status = U_UNSUPPORTED_ERROR;
        return NULL;
    }

    if (U_SUCCESS(*status)) {
        initializeFCD(status);
        if (U_FAILURE(*status)) {
            return NULL;
        }

        UStringSearch *result;
        if (textlength == -1) {
            textlength = u_strlen(text);
        }
        if (patternlength == -1) {
            patternlength = u_strlen(pattern);
        }
        if (textlength <= 0 || patternlength <= 0) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
            return NULL;
        }

        result = (UStringSearch *)uprv_malloc(sizeof(UStringSearch));
        if (result == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }

        result->collator    = collator;
        result->strength    = ucol_getStrength(collator);
        result->ceMask      = getMask(result->strength);
        result->toShift     =
             ucol_getAttribute(collator, UCOL_ALTERNATE_HANDLING, status) ==
                                                            UCOL_SHIFTED;
        result->variableTop = ucol_getVariableTop(collator, status);

        result->nfd         = Normalizer2::getNFDInstance(*status);

        if (U_FAILURE(*status)) {
            uprv_free(result);
            return NULL;
        }

        result->search             = (USearch *)uprv_malloc(sizeof(USearch));
        if (result->search == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            uprv_free(result);
            return NULL;
        }

        result->search->text       = text;
        result->search->textLength = textlength;

        result->pattern.text       = pattern;
        result->pattern.textLength = patternlength;
        result->pattern.ces         = NULL;
        result->pattern.pces        = NULL;

        result->search->breakIter  = breakiter;
#if !UCONFIG_NO_BREAK_ITERATION
        result->search->internalBreakIter = ubrk_open(UBRK_CHARACTER, ucol_getLocaleByType(result->collator, ULOC_VALID_LOCALE, status), text, textlength, status);
        if (breakiter) {
            ubrk_setText(breakiter, text, textlength, status);
        }
#endif

        result->ownCollator           = FALSE;
        result->search->matchedLength = 0;
        result->search->matchedIndex  = USEARCH_DONE;
        result->utilIter              = NULL;
        result->textIter              = ucol_openElements(collator, text,
                                                          textlength, status);
        result->textProcessedIter     = NULL;
        if (U_FAILURE(*status)) {
            usearch_close(result);
            return NULL;
        }

        result->search->isOverlap          = FALSE;
        result->search->isCanonicalMatch   = FALSE;
        result->search->elementComparisonType = 0;
        result->search->isForwardSearching = TRUE;
        result->search->reset              = TRUE;

        initialize(result, status);

        if (U_FAILURE(*status)) {
            usearch_close(result);
            return NULL;
        }

        return result;
    }
    return NULL;
}

U_CAPI void U_EXPORT2 usearch_close(UStringSearch *strsrch)
{
    if (strsrch) {
        if (strsrch->pattern.ces != strsrch->pattern.cesBuffer &&
            strsrch->pattern.ces) {
            uprv_free(strsrch->pattern.ces);
        }

        if (strsrch->pattern.pces != NULL &&
            strsrch->pattern.pces != strsrch->pattern.pcesBuffer) {
            uprv_free(strsrch->pattern.pces);
        }

        delete strsrch->textProcessedIter;
        ucol_closeElements(strsrch->textIter);
        ucol_closeElements(strsrch->utilIter);

        if (strsrch->ownCollator && strsrch->collator) {
            ucol_close((UCollator *)strsrch->collator);
        }

#if !UCONFIG_NO_BREAK_ITERATION
        if (strsrch->search->internalBreakIter) {
            ubrk_close(strsrch->search->internalBreakIter);
        }
#endif

        uprv_free(strsrch->search);
        uprv_free(strsrch);
    }
}

namespace {

UBool initTextProcessedIter(UStringSearch *strsrch, UErrorCode *status) {
    if (U_FAILURE(*status)) { return FALSE; }
    if (strsrch->textProcessedIter == NULL) {
        strsrch->textProcessedIter = new icu::UCollationPCE(strsrch->textIter);
        if (strsrch->textProcessedIter == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
            return FALSE;
        }
    } else {
        strsrch->textProcessedIter->init(strsrch->textIter);
    }
    return TRUE;
}

}

// set and get methods --------------------------------------------------

U_CAPI void U_EXPORT2 usearch_setOffset(UStringSearch *strsrch,
                                        int32_t    position,
                                        UErrorCode    *status)
{
    if (U_SUCCESS(*status) && strsrch) {
        if (isOutOfBounds(strsrch->search->textLength, position)) {
            *status = U_INDEX_OUTOFBOUNDS_ERROR;
        }
        else {
            setColEIterOffset(strsrch->textIter, position);
        }
        strsrch->search->matchedIndex  = USEARCH_DONE;
        strsrch->search->matchedLength = 0;
        strsrch->search->reset         = FALSE;
    }
}

U_CAPI int32_t U_EXPORT2 usearch_getOffset(const UStringSearch *strsrch)
{
    if (strsrch) {
        int32_t result = ucol_getOffset(strsrch->textIter);
        if (isOutOfBounds(strsrch->search->textLength, result)) {
            return USEARCH_DONE;
        }
        return result;
    }
    return USEARCH_DONE;
}

U_CAPI void U_EXPORT2 usearch_setAttribute(UStringSearch *strsrch,
                                 USearchAttribute attribute,
                                 USearchAttributeValue value,
                                 UErrorCode *status)
{
    if (U_SUCCESS(*status) && strsrch) {
        switch (attribute)
        {
        case USEARCH_OVERLAP :
            strsrch->search->isOverlap = (value == USEARCH_ON ? TRUE : FALSE);
            break;
        case USEARCH_CANONICAL_MATCH :
            strsrch->search->isCanonicalMatch = (value == USEARCH_ON ? TRUE :
                                                                      FALSE);
            break;
        case USEARCH_ELEMENT_COMPARISON :
            if (value == USEARCH_PATTERN_BASE_WEIGHT_IS_WILDCARD || value == USEARCH_ANY_BASE_WEIGHT_IS_WILDCARD) {
                strsrch->search->elementComparisonType = (int16_t)value;
            } else {
                strsrch->search->elementComparisonType = 0;
            }
            break;
        case USEARCH_ATTRIBUTE_COUNT :
        default:
            *status = U_ILLEGAL_ARGUMENT_ERROR;
        }
    }
    if (value == USEARCH_ATTRIBUTE_VALUE_COUNT) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
    }
}

U_CAPI USearchAttributeValue U_EXPORT2 usearch_getAttribute(
                                                const UStringSearch *strsrch,
                                                USearchAttribute attribute)
{
    if (strsrch) {
        switch (attribute) {
        case USEARCH_OVERLAP :
            return (strsrch->search->isOverlap == TRUE ? USEARCH_ON :
                                                        USEARCH_OFF);
        case USEARCH_CANONICAL_MATCH :
            return (strsrch->search->isCanonicalMatch == TRUE ? USEARCH_ON :
                                                               USEARCH_OFF);
        case USEARCH_ELEMENT_COMPARISON :
            {
                int16_t value = strsrch->search->elementComparisonType;
                if (value == USEARCH_PATTERN_BASE_WEIGHT_IS_WILDCARD || value == USEARCH_ANY_BASE_WEIGHT_IS_WILDCARD) {
                    return (USearchAttributeValue)value;
                } else {
                    return USEARCH_STANDARD_ELEMENT_COMPARISON;
                }
            }
        case USEARCH_ATTRIBUTE_COUNT :
            return USEARCH_DEFAULT;
        }
    }
    return USEARCH_DEFAULT;
}

U_CAPI int32_t U_EXPORT2 usearch_getMatchedStart(
                                                const UStringSearch *strsrch)
{
    if (strsrch == NULL) {
        return USEARCH_DONE;
    }
    return strsrch->search->matchedIndex;
}


U_CAPI int32_t U_EXPORT2 usearch_getMatchedText(const UStringSearch *strsrch,
                                            UChar         *result,
                                            int32_t        resultCapacity,
                                            UErrorCode    *status)
{
    if (U_FAILURE(*status)) {
        return USEARCH_DONE;
    }
    if (strsrch == NULL || resultCapacity < 0 || (resultCapacity > 0 &&
        result == NULL)) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return USEARCH_DONE;
    }

    int32_t     copylength = strsrch->search->matchedLength;
    int32_t copyindex  = strsrch->search->matchedIndex;
    if (copyindex == USEARCH_DONE) {
        u_terminateUChars(result, resultCapacity, 0, status);
        return USEARCH_DONE;
    }

    if (resultCapacity < copylength) {
        copylength = resultCapacity;
    }
    if (copylength > 0) {
        uprv_memcpy(result, strsrch->search->text + copyindex,
                    copylength * sizeof(UChar));
    }
    return u_terminateUChars(result, resultCapacity,
                             strsrch->search->matchedLength, status);
}

U_CAPI int32_t U_EXPORT2 usearch_getMatchedLength(
                                              const UStringSearch *strsrch)
{
    if (strsrch) {
        return strsrch->search->matchedLength;
    }
    return USEARCH_DONE;
}

#if !UCONFIG_NO_BREAK_ITERATION

U_CAPI void U_EXPORT2 usearch_setBreakIterator(UStringSearch  *strsrch,
                                               UBreakIterator *breakiter,
                                               UErrorCode     *status)
{
    if (U_SUCCESS(*status) && strsrch) {
        strsrch->search->breakIter = breakiter;
        if (breakiter) {
            ubrk_setText(breakiter, strsrch->search->text,
                         strsrch->search->textLength, status);
        }
    }
}

U_CAPI const UBreakIterator* U_EXPORT2
usearch_getBreakIterator(const UStringSearch *strsrch)
{
    if (strsrch) {
        return strsrch->search->breakIter;
    }
    return NULL;
}

#endif

U_CAPI void U_EXPORT2 usearch_setText(      UStringSearch *strsrch,
                                      const UChar         *text,
                                            int32_t        textlength,
                                            UErrorCode    *status)
{
    if (U_SUCCESS(*status)) {
        if (strsrch == NULL || text == NULL || textlength < -1 ||
            textlength == 0) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
        }
        else {
            if (textlength == -1) {
                textlength = u_strlen(text);
            }
            strsrch->search->text       = text;
            strsrch->search->textLength = textlength;
            ucol_setText(strsrch->textIter, text, textlength, status);
            strsrch->search->matchedIndex  = USEARCH_DONE;
            strsrch->search->matchedLength = 0;
            strsrch->search->reset         = TRUE;
#if !UCONFIG_NO_BREAK_ITERATION
            if (strsrch->search->breakIter != NULL) {
                ubrk_setText(strsrch->search->breakIter, text,
                             textlength, status);
            }
            ubrk_setText(strsrch->search->internalBreakIter, text, textlength, status);
#endif
        }
    }
}

U_CAPI const UChar * U_EXPORT2 usearch_getText(const UStringSearch *strsrch,
                                                     int32_t       *length)
{
    if (strsrch) {
        *length = strsrch->search->textLength;
        return strsrch->search->text;
    }
    return NULL;
}

U_CAPI void U_EXPORT2 usearch_setCollator(      UStringSearch *strsrch,
                                          const UCollator     *collator,
                                                UErrorCode    *status)
{
    if (U_SUCCESS(*status)) {
        if (collator == NULL) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
            return;
        }

        if (strsrch) {
            delete strsrch->textProcessedIter;
            strsrch->textProcessedIter = NULL;
            ucol_closeElements(strsrch->textIter);
            ucol_closeElements(strsrch->utilIter);
            strsrch->textIter = strsrch->utilIter = NULL;
            if (strsrch->ownCollator && (strsrch->collator != collator)) {
                ucol_close((UCollator *)strsrch->collator);
                strsrch->ownCollator = FALSE;
            }
            strsrch->collator    = collator;
            strsrch->strength    = ucol_getStrength(collator);
            strsrch->ceMask      = getMask(strsrch->strength);
#if !UCONFIG_NO_BREAK_ITERATION
            ubrk_close(strsrch->search->internalBreakIter);
            strsrch->search->internalBreakIter = ubrk_open(UBRK_CHARACTER, ucol_getLocaleByType(collator, ULOC_VALID_LOCALE, status),
                                                     strsrch->search->text, strsrch->search->textLength, status);
#endif
            // if status is a failure, ucol_getAttribute returns UCOL_DEFAULT
            strsrch->toShift     =
               ucol_getAttribute(collator, UCOL_ALTERNATE_HANDLING, status) ==
                                                                UCOL_SHIFTED;
            // if status is a failure, ucol_getVariableTop returns 0
            strsrch->variableTop = ucol_getVariableTop(collator, status);
            strsrch->textIter = ucol_openElements(collator,
                                      strsrch->search->text,
                                      strsrch->search->textLength,
                                      status);
            strsrch->utilIter = ucol_openElements(
                    collator, strsrch->pattern.text, strsrch->pattern.textLength, status);
            // initialize() _after_ setting the iterators for the new collator.
            initialize(strsrch, status);
        }

        // **** are these calls needed?
        // **** we call uprv_init_pce in initializePatternPCETable
        // **** and the CEIBuffer constructor...
#if 0
        uprv_init_pce(strsrch->textIter);
        uprv_init_pce(strsrch->utilIter);
#endif
    }
}

U_CAPI UCollator * U_EXPORT2 usearch_getCollator(const UStringSearch *strsrch)
{
    if (strsrch) {
        return (UCollator *)strsrch->collator;
    }
    return NULL;
}

U_CAPI void U_EXPORT2 usearch_setPattern(      UStringSearch *strsrch,
                                         const UChar         *pattern,
                                               int32_t        patternlength,
                                               UErrorCode    *status)
{
    if (U_SUCCESS(*status)) {
        if (strsrch == NULL || pattern == NULL) {
            *status = U_ILLEGAL_ARGUMENT_ERROR;
        }
        else {
            if (patternlength == -1) {
                patternlength = u_strlen(pattern);
            }
            if (patternlength == 0) {
                *status = U_ILLEGAL_ARGUMENT_ERROR;
                return;
            }
            strsrch->pattern.text       = pattern;
            strsrch->pattern.textLength = patternlength;
            initialize(strsrch, status);
        }
    }
}

U_CAPI const UChar* U_EXPORT2
usearch_getPattern(const UStringSearch *strsrch,
                   int32_t       *length)
{
    if (strsrch) {
        *length = strsrch->pattern.textLength;
        return strsrch->pattern.text;
    }
    return NULL;
}

// miscellanous methods --------------------------------------------------

U_CAPI int32_t U_EXPORT2 usearch_first(UStringSearch *strsrch,
                                           UErrorCode    *status)
{
    if (strsrch && U_SUCCESS(*status)) {
        strsrch->search->isForwardSearching = TRUE;
        usearch_setOffset(strsrch, 0, status);
        if (U_SUCCESS(*status)) {
            return usearch_next(strsrch, status);
        }
    }
    return USEARCH_DONE;
}

U_CAPI int32_t U_EXPORT2 usearch_following(UStringSearch *strsrch,
                                               int32_t    position,
                                               UErrorCode    *status)
{
    if (strsrch && U_SUCCESS(*status)) {
        strsrch->search->isForwardSearching = TRUE;
        // position checked in usearch_setOffset
        usearch_setOffset(strsrch, position, status);
        if (U_SUCCESS(*status)) {
            return usearch_next(strsrch, status);
        }
    }
    return USEARCH_DONE;
}

U_CAPI int32_t U_EXPORT2 usearch_last(UStringSearch *strsrch,
                                          UErrorCode    *status)
{
    if (strsrch && U_SUCCESS(*status)) {
        strsrch->search->isForwardSearching = FALSE;
        usearch_setOffset(strsrch, strsrch->search->textLength, status);
        if (U_SUCCESS(*status)) {
            return usearch_previous(strsrch, status);
        }
    }
    return USEARCH_DONE;
}

U_CAPI int32_t U_EXPORT2 usearch_preceding(UStringSearch *strsrch,
                                               int32_t    position,
                                               UErrorCode    *status)
{
    if (strsrch && U_SUCCESS(*status)) {
        strsrch->search->isForwardSearching = FALSE;
        // position checked in usearch_setOffset
        usearch_setOffset(strsrch, position, status);
        if (U_SUCCESS(*status)) {
            return usearch_previous(strsrch, status);
        }
    }
    return USEARCH_DONE;
}

/**
* If a direction switch is required, we'll count the number of ces till the
* beginning of the collation element iterator and iterate forwards that
* number of times. This is so that we get to the correct point within the
* string to continue the search in. Imagine when we are in the middle of the
* normalization buffer when the change in direction is request. arrrgghh....
* After searching the offset within the collation element iterator will be
* shifted to the start of the match. If a match is not found, the offset would
* have been set to the end of the text string in the collation element
* iterator.
* Okay, here's my take on normalization buffer. The only time when there can
* be 2 matches within the same normalization is when the pattern is consists
* of all accents. But since the offset returned is from the text string, we
* should not confuse the caller by returning the second match within the
* same normalization buffer. If we do, the 2 results will have the same match
* offsets, and that'll be confusing. I'll return the next match that doesn't
* fall within the same normalization buffer. Note this does not affect the
* results of matches spanning the text and the normalization buffer.
* The position to start searching is taken from the collation element
* iterator. Callers of this API would have to set the offset in the collation
* element iterator before using this method.
*/
U_CAPI int32_t U_EXPORT2 usearch_next(UStringSearch *strsrch,
                                          UErrorCode    *status)
{
    if (U_SUCCESS(*status) && strsrch) {
        // note offset is either equivalent to the start of the previous match
        // or is set by the user
        int32_t      offset       = usearch_getOffset(strsrch);
        USearch     *search       = strsrch->search;
        search->reset             = FALSE;
        int32_t      textlength   = search->textLength;
        if (search->isForwardSearching) {
#if BOYER_MOORE
            if (offset == textlength
                || (!search->isOverlap &&
                    (offset + strsrch->pattern.defaultShiftSize > textlength ||
                    (search->matchedIndex != USEARCH_DONE &&
                     offset + search->matchedLength >= textlength)))) {
                // not enough characters to match
                setMatchNotFound(strsrch);
                return USEARCH_DONE;
            }
#else
            if (offset == textlength ||
                (! search->isOverlap &&
                (search->matchedIndex != USEARCH_DONE &&
                offset + search->matchedLength > textlength))) {
                    // not enough characters to match
                    setMatchNotFound(strsrch);
                    return USEARCH_DONE;
            }
#endif
        }
        else {
            // switching direction.
            // if matchedIndex == USEARCH_DONE, it means that either a
            // setOffset has been called or that previous ran off the text
            // string. the iterator would have been set to offset 0 if a
            // match is not found.
            search->isForwardSearching = TRUE;
            if (search->matchedIndex != USEARCH_DONE) {
                // there's no need to set the collation element iterator
                // the next call to next will set the offset.
                return search->matchedIndex;
            }
        }

        if (U_SUCCESS(*status)) {
            if (strsrch->pattern.cesLength == 0) {
                if (search->matchedIndex == USEARCH_DONE) {
                    search->matchedIndex = offset;
                }
                else { // moves by codepoints
                    U16_FWD_1(search->text, search->matchedIndex, textlength);
                }

                search->matchedLength = 0;
                setColEIterOffset(strsrch->textIter, search->matchedIndex);
                // status checked below
                if (search->matchedIndex == textlength) {
                    search->matchedIndex = USEARCH_DONE;
                }
            }
            else {
                if (search->matchedLength > 0) {
                    // if matchlength is 0 we are at the start of the iteration
                    if (search->isOverlap) {
                        ucol_setOffset(strsrch->textIter, offset + 1, status);
                    }
                    else {
                        ucol_setOffset(strsrch->textIter,
                                       offset + search->matchedLength, status);
                    }
                }
                else {
                    // for boundary check purposes. this will ensure that the
                    // next match will not preceed the current offset
                    // note search->matchedIndex will always be set to something
                    // in the code
                    search->matchedIndex = offset - 1;
                }

                if (search->isCanonicalMatch) {
                    // can't use exact here since extra accents are allowed.
                    usearch_handleNextCanonical(strsrch, status);
                }
                else {
                    usearch_handleNextExact(strsrch, status);
                }
            }

            if (U_FAILURE(*status)) {
                return USEARCH_DONE;
            }

#if !BOYER_MOORE
            if (search->matchedIndex == USEARCH_DONE) {
                ucol_setOffset(strsrch->textIter, search->textLength, status);
            } else {
                ucol_setOffset(strsrch->textIter, search->matchedIndex, status);
            }
#endif

            return search->matchedIndex;
        }
    }
    return USEARCH_DONE;
}

U_CAPI int32_t U_EXPORT2 usearch_previous(UStringSearch *strsrch,
                                              UErrorCode *status)
{
    if (U_SUCCESS(*status) && strsrch) {
        int32_t offset;
        USearch *search = strsrch->search;
        if (search->reset) {
            offset                     = search->textLength;
            search->isForwardSearching = FALSE;
            search->reset              = FALSE;
            setColEIterOffset(strsrch->textIter, offset);
        }
        else {
            offset = usearch_getOffset(strsrch);
        }

        int32_t matchedindex = search->matchedIndex;
        if (search->isForwardSearching == TRUE) {
            // switching direction.
            // if matchedIndex == USEARCH_DONE, it means that either a
            // setOffset has been called or that next ran off the text
            // string. the iterator would have been set to offset textLength if
            // a match is not found.
            search->isForwardSearching = FALSE;
            if (matchedindex != USEARCH_DONE) {
                return matchedindex;
            }
        }
        else {
#if BOYER_MOORE
            if (offset == 0 || matchedindex == 0 ||
                (!search->isOverlap &&
                    (offset < strsrch->pattern.defaultShiftSize ||
                    (matchedindex != USEARCH_DONE &&
                    matchedindex < strsrch->pattern.defaultShiftSize)))) {
                // not enough characters to match
                setMatchNotFound(strsrch);
                return USEARCH_DONE;
            }
#else
            // Could check pattern length, but the
            // linear search will do the right thing
            if (offset == 0 || matchedindex == 0) {
                setMatchNotFound(strsrch);
                return USEARCH_DONE;
            }
#endif
        }

        if (U_SUCCESS(*status)) {
            if (strsrch->pattern.cesLength == 0) {
                search->matchedIndex =
                      (matchedindex == USEARCH_DONE ? offset : matchedindex);
                if (search->matchedIndex == 0) {
                    setMatchNotFound(strsrch);
                    // status checked below
                }
                else { // move by codepoints
                    U16_BACK_1(search->text, 0, search->matchedIndex);
                    setColEIterOffset(strsrch->textIter, search->matchedIndex);
                    // status checked below
                    search->matchedLength = 0;
                }
            }
            else {
                if (strsrch->search->isCanonicalMatch) {
                    // can't use exact here since extra accents are allowed.
                    usearch_handlePreviousCanonical(strsrch, status);
                    // status checked below
                }
                else {
                    usearch_handlePreviousExact(strsrch, status);
                    // status checked below
                }
            }

            if (U_FAILURE(*status)) {
                return USEARCH_DONE;
            }

            return search->matchedIndex;
        }
    }
    return USEARCH_DONE;
}



U_CAPI void U_EXPORT2 usearch_reset(UStringSearch *strsrch)
{
    /*
    reset is setting the attributes that are already in
    string search, hence all attributes in the collator should
    be retrieved without any problems
    */
    if (strsrch) {
        UErrorCode status            = U_ZERO_ERROR;
        UBool      sameCollAttribute = TRUE;
        uint32_t   ceMask;
        UBool      shift;
        uint32_t   varTop;

        // **** hack to deal w/ how processed CEs encode quaternary ****
        UCollationStrength newStrength = ucol_getStrength(strsrch->collator);
        if ((strsrch->strength < UCOL_QUATERNARY && newStrength >= UCOL_QUATERNARY) ||
            (strsrch->strength >= UCOL_QUATERNARY && newStrength < UCOL_QUATERNARY)) {
                sameCollAttribute = FALSE;
        }

        strsrch->strength    = ucol_getStrength(strsrch->collator);
        ceMask = getMask(strsrch->strength);
        if (strsrch->ceMask != ceMask) {
            strsrch->ceMask = ceMask;
            sameCollAttribute = FALSE;
        }

        // if status is a failure, ucol_getAttribute returns UCOL_DEFAULT
        shift = ucol_getAttribute(strsrch->collator, UCOL_ALTERNATE_HANDLING,
                                  &status) == UCOL_SHIFTED;
        if (strsrch->toShift != shift) {
            strsrch->toShift  = shift;
            sameCollAttribute = FALSE;
        }

        // if status is a failure, ucol_getVariableTop returns 0
        varTop = ucol_getVariableTop(strsrch->collator, &status);
        if (strsrch->variableTop != varTop) {
            strsrch->variableTop = varTop;
            sameCollAttribute    = FALSE;
        }
        if (!sameCollAttribute) {
            initialize(strsrch, &status);
        }
        ucol_setText(strsrch->textIter, strsrch->search->text,
                              strsrch->search->textLength,
                              &status);
        strsrch->search->matchedLength      = 0;
        strsrch->search->matchedIndex       = USEARCH_DONE;
        strsrch->search->isOverlap          = FALSE;
        strsrch->search->isCanonicalMatch   = FALSE;
        strsrch->search->elementComparisonType = 0;
        strsrch->search->isForwardSearching = TRUE;
        strsrch->search->reset              = TRUE;
    }
}

//
//  CEI  Collation Element + source text index.
//       These structs are kept in the circular buffer.
//
struct  CEI {
    int64_t ce;
    int32_t lowIndex;
    int32_t highIndex;
};

U_NAMESPACE_BEGIN

namespace {
//
//  CEIBuffer   A circular buffer of CEs-with-index from the text being searched.
//
#define   DEFAULT_CEBUFFER_SIZE 96
#define   CEBUFFER_EXTRA 32
// Some typical max values to make buffer size more reasonable for asymmetric search.
// #8694 is for a better long-term solution to allocation of this buffer.
#define   MAX_TARGET_IGNORABLES_PER_PAT_JAMO_L 8
#define   MAX_TARGET_IGNORABLES_PER_PAT_OTHER 3
#define   MIGHT_BE_JAMO_L(c) ((c >= 0x1100 && c <= 0x115E) || (c >= 0x3131 && c <= 0x314E) || (c >= 0x3165 && c <= 0x3186))
struct CEIBuffer {
    CEI                  defBuf[DEFAULT_CEBUFFER_SIZE];
    CEI                 *buf;
    int32_t              bufSize;
    int32_t              firstIx;
    int32_t              limitIx;
    UCollationElements  *ceIter;
    UStringSearch       *strSearch;



               CEIBuffer(UStringSearch *ss, UErrorCode *status);
               ~CEIBuffer();
   const CEI   *get(int32_t index);
   const CEI   *getPrevious(int32_t index);
};


CEIBuffer::CEIBuffer(UStringSearch *ss, UErrorCode *status) {
    buf = defBuf;
    strSearch = ss;
    bufSize = ss->pattern.pcesLength + CEBUFFER_EXTRA;
    if (ss->search->elementComparisonType != 0) {
        const UChar * patText = ss->pattern.text;
        if (patText) {
            const UChar * patTextLimit = patText + ss->pattern.textLength;
            while ( patText < patTextLimit ) {
                UChar c = *patText++;
                if (MIGHT_BE_JAMO_L(c)) {
                    bufSize += MAX_TARGET_IGNORABLES_PER_PAT_JAMO_L;
                } else {
                    // No check for surrogates, we might allocate slightly more buffer than necessary.
                    bufSize += MAX_TARGET_IGNORABLES_PER_PAT_OTHER;
                }
            }
        }
    }
    ceIter    = ss->textIter;
    firstIx = 0;
    limitIx = 0;

    if (!initTextProcessedIter(ss, status)) { return; }

    if (bufSize>DEFAULT_CEBUFFER_SIZE) {
        buf = (CEI *)uprv_malloc(bufSize * sizeof(CEI));
        if (buf == NULL) {
            *status = U_MEMORY_ALLOCATION_ERROR;
        }
    }
}

// TODO: add a reset or init function so that allocated
//       buffers can be retained & reused.

CEIBuffer::~CEIBuffer() {
    if (buf != defBuf) {
        uprv_free(buf);
    }
}


// Get the CE with the specified index.
//   Index must be in the range
//          n-history_size < index < n+1
//   where n is the largest index to have been fetched by some previous call to this function.
//   The CE value will be UCOL__PROCESSED_NULLORDER at end of input.
//
const CEI *CEIBuffer::get(int32_t index) {
    int i = index % bufSize;

    if (index>=firstIx && index<limitIx) {
        // The request was for an entry already in our buffer.
        //  Just return it.
        return &buf[i];
    }

    // Caller is requesting a new, never accessed before, CE.
    //   Verify that it is the next one in sequence, which is all
    //   that is allowed.
    if (index != limitIx) {
        U_ASSERT(FALSE);
        // TODO: In ICU 64 the above assert was changed to use UPRV_UNREACHABLE instead
        // which unconditionally calls abort(). However, there were cases where this was
        // being hit. This change is reverted for now, restoring the existing behavior.
        // ICU-20792 tracks the follow-up work/further investigation on this.
        return NULL;
    }

    // Manage the circular CE buffer indexing
    limitIx++;

    if (limitIx - firstIx >= bufSize) {
        // The buffer is full, knock out the lowest-indexed entry.
        firstIx++;
    }

    UErrorCode status = U_ZERO_ERROR;

    buf[i].ce = strSearch->textProcessedIter->nextProcessed(&buf[i].lowIndex, &buf[i].highIndex, &status);

    return &buf[i];
}

// Get the CE with the specified index.
//   Index must be in the range
//          n-history_size < index < n+1
//   where n is the largest index to have been fetched by some previous call to this function.
//   The CE value will be UCOL__PROCESSED_NULLORDER at end of input.
//
const CEI *CEIBuffer::getPrevious(int32_t index) {
    int i = index % bufSize;

    if (index>=firstIx && index<limitIx) {
        // The request was for an entry already in our buffer.
        //  Just return it.
        return &buf[i];
    }

    // Caller is requesting a new, never accessed before, CE.
    //   Verify that it is the next one in sequence, which is all
    //   that is allowed.
    if (index != limitIx) {
        U_ASSERT(FALSE);
        // TODO: In ICU 64 the above assert was changed to use UPRV_UNREACHABLE instead
        // which unconditionally calls abort(). However, there were cases where this was
        // being hit. This change is reverted for now, restoring the existing behavior.
        // ICU-20792 tracks the follow-up work/further investigation on this.
        return NULL;
    }

    // Manage the circular CE buffer indexing
    limitIx++;

    if (limitIx - firstIx >= bufSize) {
        // The buffer is full, knock out the lowest-indexed entry.
        firstIx++;
    }

    UErrorCode status = U_ZERO_ERROR;

    buf[i].ce = strSearch->textProcessedIter->previousProcessed(&buf[i].lowIndex, &buf[i].highIndex, &status);

    return &buf[i];
}

}

U_NAMESPACE_END


// #define USEARCH_DEBUG

#ifdef USEARCH_DEBUG
#include <stdio.h>
#include <stdlib.h>
#endif

/*
 * Find the next break boundary after startIndex. If the UStringSearch object
 * has an external break iterator, use that. Otherwise use the internal character
 * break iterator.
 */
static int32_t nextBoundaryAfter(UStringSearch *strsrch, int32_t startIndex) {
#if 0
    const UChar *text = strsrch->search->text;
    int32_t textLen   = strsrch->search->textLength;

    U_ASSERT(startIndex>=0);
    U_ASSERT(startIndex<=textLen);

    if (startIndex >= textLen) {
        return startIndex;
    }

    UChar32  c;
    int32_t  i = startIndex;
    U16_NEXT(text, i, textLen, c);

    // If we are on a control character, stop without looking for combining marks.
    //    Control characters do not combine.
    int32_t gcProperty = u_getIntPropertyValue(c, UCHAR_GRAPHEME_CLUSTER_BREAK);
    if (gcProperty==U_GCB_CONTROL || gcProperty==U_GCB_LF || gcProperty==U_GCB_CR) {
        return i;
    }

    // The initial character was not a control, and can thus accept trailing
    //   combining characters.  Advance over however many of them there are.
    int32_t  indexOfLastCharChecked;
    for (;;) {
        indexOfLastCharChecked = i;
        if (i>=textLen) {
            break;
        }
        U16_NEXT(text, i, textLen, c);
        gcProperty = u_getIntPropertyValue(c, UCHAR_GRAPHEME_CLUSTER_BREAK);
        if (gcProperty != U_GCB_EXTEND && gcProperty != U_GCB_SPACING_MARK) {
            break;
        }
    }
    return indexOfLastCharChecked;
#elif !UCONFIG_NO_BREAK_ITERATION
    UBreakIterator *breakiterator = strsrch->search->breakIter;

    if (breakiterator == NULL) {
        breakiterator = strsrch->search->internalBreakIter;
    }

    if (breakiterator != NULL) {
        return ubrk_following(breakiterator, startIndex);
    }

    return startIndex;
#else
    // **** or should we use the original code? ****
    return startIndex;
#endif

}

/*
 * Returns TRUE if index is on a break boundary. If the UStringSearch
 * has an external break iterator, test using that, otherwise test
 * using the internal character break iterator.
 */
static UBool isBreakBoundary(UStringSearch *strsrch, int32_t index) {
#if 0
    const UChar *text = strsrch->search->text;
    int32_t textLen   = strsrch->search->textLength;

    U_ASSERT(index>=0);
    U_ASSERT(index<=textLen);

    if (index>=textLen || index<=0) {
        return TRUE;
    }

    // If the character at the current index is not a GRAPHEME_EXTEND
    //    then we can not be within a combining sequence.
    UChar32  c;
    U16_GET(text, 0, index, textLen, c);
    int32_t gcProperty = u_getIntPropertyValue(c, UCHAR_GRAPHEME_CLUSTER_BREAK);
    if (gcProperty != U_GCB_EXTEND && gcProperty != U_GCB_SPACING_MARK) {
        return TRUE;
    }

    // We are at a combining mark.  If the preceding character is anything
    //   except a CONTROL, CR or LF, we are in a combining sequence.
    U16_PREV(text, 0, index, c);
    gcProperty = u_getIntPropertyValue(c, UCHAR_GRAPHEME_CLUSTER_BREAK);
    UBool combining =  !(gcProperty==U_GCB_CONTROL || gcProperty==U_GCB_LF || gcProperty==U_GCB_CR);
    return !combining;
#elif !UCONFIG_NO_BREAK_ITERATION
    UBreakIterator *breakiterator = strsrch->search->breakIter;

    if (breakiterator == NULL) {
        breakiterator = strsrch->search->internalBreakIter;
    }

    return (breakiterator != NULL && ubrk_isBoundary(breakiterator, index));
#else
    // **** or use the original code? ****
    return TRUE;
#endif
}

#if 0
static UBool onBreakBoundaries(const UStringSearch *strsrch, int32_t start, int32_t end)
{
#if !UCONFIG_NO_BREAK_ITERATION
    UBreakIterator *breakiterator = strsrch->search->breakIter;

    if (breakiterator != NULL) {
        int32_t startindex = ubrk_first(breakiterator);
        int32_t endindex   = ubrk_last(breakiterator);

        // out-of-range indexes are never boundary positions
        if (start < startindex || start > endindex ||
            end < startindex || end > endindex) {
            return FALSE;
        }

        return ubrk_isBoundary(breakiterator, start) &&
               ubrk_isBoundary(breakiterator, end);
    }
#endif

    return TRUE;
}
#endif

typedef enum {
    U_CE_MATCH = -1,
    U_CE_NO_MATCH = 0,
    U_CE_SKIP_TARG,
    U_CE_SKIP_PATN
} UCompareCEsResult;
#define U_CE_LEVEL2_BASE 0x00000005
#define U_CE_LEVEL3_BASE 0x00050000

static UCompareCEsResult compareCE64s(int64_t targCE, int64_t patCE, int16_t compareType) {
    if (targCE == patCE) {
        return U_CE_MATCH;
    }
    if (compareType == 0) {
        return U_CE_NO_MATCH;
    }
    
    int64_t targCEshifted = targCE >> 32;
    int64_t patCEshifted = patCE >> 32;
    int64_t mask;

    mask = 0xFFFF0000;
    int32_t targLev1 = (int32_t)(targCEshifted & mask);
    int32_t patLev1 = (int32_t)(patCEshifted & mask);
    if ( targLev1 != patLev1 ) {
        if ( targLev1 == 0 ) {
            return U_CE_SKIP_TARG;
        }
        if ( patLev1 == 0 && compareType == USEARCH_ANY_BASE_WEIGHT_IS_WILDCARD ) {
            return U_CE_SKIP_PATN;
        }
        return U_CE_NO_MATCH;
    }

    mask = 0x0000FFFF;
    int32_t targLev2 = (int32_t)(targCEshifted & mask);
    int32_t patLev2 = (int32_t)(patCEshifted & mask);
    if ( targLev2 != patLev2 ) {
        if ( targLev2 == 0 ) {
            return U_CE_SKIP_TARG;
        }
        if ( patLev2 == 0 && compareType == USEARCH_ANY_BASE_WEIGHT_IS_WILDCARD ) {
            return U_CE_SKIP_PATN;
        }
        return (patLev2 == U_CE_LEVEL2_BASE || (compareType == USEARCH_ANY_BASE_WEIGHT_IS_WILDCARD && targLev2 == U_CE_LEVEL2_BASE) )?
            U_CE_MATCH: U_CE_NO_MATCH;
    }
    
    mask = 0xFFFF0000;
    int32_t targLev3 = (int32_t)(targCE & mask);
    int32_t patLev3 = (int32_t)(patCE & mask);
    if ( targLev3 != patLev3 ) {
        return (patLev3 == U_CE_LEVEL3_BASE || (compareType == USEARCH_ANY_BASE_WEIGHT_IS_WILDCARD && targLev3 == U_CE_LEVEL3_BASE) )?
            U_CE_MATCH: U_CE_NO_MATCH;
   }

    return U_CE_MATCH;
}

#if BOYER_MOORE
// TODO: #if BOYER_MOORE, need 32-bit version of compareCE64s
#endif

namespace {

UChar32 codePointAt(const USearch &search, int32_t index) {
    if (index < search.textLength) {
        UChar32 c;
        U16_NEXT(search.text, index, search.textLength, c);
        return c;
    }
    return U_SENTINEL;
}

UChar32 codePointBefore(const USearch &search, int32_t index) {
    if (0 < index) {
        UChar32 c;
        U16_PREV(search.text, 0, index, c);
        return c;
    }
    return U_SENTINEL;
}

}  // namespace

U_CAPI UBool U_EXPORT2 usearch_search(UStringSearch  *strsrch,
                                       int32_t        startIdx,
                                       int32_t        *matchStart,
                                       int32_t        *matchLimit,
                                       UErrorCode     *status)
{
    if (U_FAILURE(*status)) {
        return FALSE;
    }

    // TODO:  reject search patterns beginning with a combining char.

#ifdef USEARCH_DEBUG
    if (getenv("USEARCH_DEBUG") != NULL) {
        printf("Pattern CEs\n");
        for (int ii=0; ii<strsrch->pattern.cesLength; ii++) {
            printf(" %8x", strsrch->pattern.ces[ii]);
        }
        printf("\n");
    }

#endif
    // Input parameter sanity check.
    //  TODO:  should input indices clip to the text length
    //         in the same way that UText does.
    if(strsrch->pattern.cesLength == 0         ||
       startIdx < 0                           ||
       startIdx > strsrch->search->textLength ||
       strsrch->pattern.ces == NULL) {
           *status = U_ILLEGAL_ARGUMENT_ERROR;
           return FALSE;
    }

    if (strsrch->pattern.pces == NULL) {
        initializePatternPCETable(strsrch, status);
    }

    ucol_setOffset(strsrch->textIter, startIdx, status);
    CEIBuffer ceb(strsrch, status);


    int32_t    targetIx = 0;
    const CEI *targetCEI = NULL;
    int32_t    patIx;
    UBool      found;

    int32_t  mStart = -1;
    int32_t  mLimit = -1;
    int32_t  minLimit;
    int32_t  maxLimit;



    // Outer loop moves over match starting positions in the
    //      target CE space.
    // Here we see the target as a sequence of collation elements, resulting from the following:
    // 1. Target characters were decomposed, and (if appropriate) other compressions and expansions are applied
    //    (for example, digraphs such as IJ may be broken into two characters).
    // 2. An int64_t CE weight is determined for each resulting unit (high 16 bits are primary strength, next
    //    16 bits are secondary, next 16 (the high 16 bits of the low 32-bit half) are tertiary. Any of these
    //    fields that are for strengths below that of the collator are set to 0. If this makes the int64_t
    //    CE weight 0 (as for a combining diacritic with secondary weight when the collator strentgh is primary),
    //    then the CE is deleted, so the following code sees only CEs that are relevant.
    // For each CE, the lowIndex and highIndex correspond to where this CE begins and ends in the original text.
    // If lowIndex==highIndex, either the CE resulted from an expansion/decomposition of one of the original text
    // characters, or the CE marks the limit of the target text (in which case the CE weight is UCOL_PROCESSED_NULLORDER).
    //
    for(targetIx=0; ; targetIx++)
    {
        found = TRUE;
        //  Inner loop checks for a match beginning at each
        //  position from the outer loop.
        int32_t targetIxOffset = 0;
        int64_t patCE = 0;
        // For targetIx > 0, this ceb.get gets a CE that is as far back in the ring buffer
        // (compared to the last CE fetched for the previous targetIx value) as we need to go
        // for this targetIx value, so if it is non-NULL then other ceb.get calls should be OK.
        const CEI *firstCEI = ceb.get(targetIx);
        if (firstCEI == NULL) {
            *status = U_INTERNAL_PROGRAM_ERROR;
            found = FALSE;
            break;
        }
        
        for (patIx=0; patIx<strsrch->pattern.pcesLength; patIx++) {
            patCE = strsrch->pattern.pces[patIx];
            targetCEI = ceb.get(targetIx+patIx+targetIxOffset);
            //  Compare CE from target string with CE from the pattern.
            //    Note that the target CE will be UCOL_PROCESSED_NULLORDER if we reach the end of input,
            //    which will fail the compare, below.
            UCompareCEsResult ceMatch = compareCE64s(targetCEI->ce, patCE, strsrch->search->elementComparisonType);
            if ( ceMatch == U_CE_NO_MATCH ) {
                found = FALSE;
                break;
            } else if ( ceMatch > U_CE_NO_MATCH ) {
                if ( ceMatch == U_CE_SKIP_TARG ) {
                    // redo with same patCE, next targCE
                    patIx--;
                    targetIxOffset++;
                } else { // ceMatch == U_CE_SKIP_PATN
                    // redo with same targCE, next patCE
                    targetIxOffset--;
                }
            }
        }
        targetIxOffset += strsrch->pattern.pcesLength; // this is now the offset in target CE space to end of the match so far

        if (!found && ((targetCEI == NULL) || (targetCEI->ce != UCOL_PROCESSED_NULLORDER))) {
            // No match at this targetIx.  Try again at the next.
            continue;
        }

        if (!found) {
            // No match at all, we have run off the end of the target text.
            break;
        }


        // We have found a match in CE space.
        // Now determine the bounds in string index space.
        //  There still is a chance of match failure if the CE range not correspond to
        //     an acceptable character range.
        //
        const CEI *lastCEI  = ceb.get(targetIx + targetIxOffset - 1);

        mStart   = firstCEI->lowIndex;
        minLimit = lastCEI->lowIndex;

        // Look at the CE following the match.  If it is UCOL_NULLORDER the match
        //   extended to the end of input, and the match is good.

        // Look at the high and low indices of the CE following the match. If
        // they are the same it means one of two things:
        //    1. The match extended to the last CE from the target text, which is OK, or
        //    2. The last CE that was part of the match is in an expansion that extends
        //       to the first CE after the match. In this case, we reject the match.
        const CEI *nextCEI = 0;
        if (strsrch->search->elementComparisonType == 0) {
            nextCEI  = ceb.get(targetIx + targetIxOffset);
            maxLimit = nextCEI->lowIndex;
            if (nextCEI->lowIndex == nextCEI->highIndex && nextCEI->ce != UCOL_PROCESSED_NULLORDER) {
                found = FALSE;
            }
        } else {
            for ( ; ; ++targetIxOffset ) {
                nextCEI = ceb.get(targetIx + targetIxOffset);
                maxLimit = nextCEI->lowIndex;
                // If we are at the end of the target too, match succeeds
                if (  nextCEI->ce == UCOL_PROCESSED_NULLORDER ) {
                    break;
                }
                // As long as the next CE has primary weight of 0,
                // it is part of the last target element matched by the pattern;
                // make sure it can be part of a match with the last patCE
                if ( (((nextCEI->ce) >> 32) & 0xFFFF0000UL) == 0 ) {
                    UCompareCEsResult ceMatch = compareCE64s(nextCEI->ce, patCE, strsrch->search->elementComparisonType);
                    if ( ceMatch == U_CE_NO_MATCH || ceMatch == U_CE_SKIP_PATN ) {
                        found = FALSE;
                        break;
                    }
                // If lowIndex == highIndex, this target CE is part of an expansion of the last matched
                // target element, but it has non-zero primary weight => match fails
                } else if ( nextCEI->lowIndex == nextCEI->highIndex ) {
                    found = false;
                    break;
                // Else the target CE is not part of an expansion of the last matched element, match succeeds
                } else {
                    break;
                }
            }
        }


        // Check for the start of the match being within a combining sequence.
        //   This can happen if the pattern itself begins with a combining char, and
        //   the match found combining marks in the target text that were attached
        //    to something else.
        //   This type of match should be rejected for not completely consuming a
        //   combining sequence.
        if (!isBreakBoundary(strsrch, mStart)) {
            found = FALSE;
        }

        // Check for the start of the match being within an Collation Element Expansion,
        //   meaning that the first char of the match is only partially matched.
        //   With expansions, the first CE will report the index of the source
        //   character, and all subsequent (expansions) CEs will report the source index of the
        //    _following_ character.
        int32_t secondIx = firstCEI->highIndex;
        if (mStart == secondIx) {
            found = FALSE;
        }

        // Allow matches to end in the middle of a grapheme cluster if the following
        // conditions are met; this is needed to make prefix search work properly in
        // Indic, see #11750
        // * the default breakIter is being used
        // * the next collation element after this combining sequence
        //   - has non-zero primary weight
        //   - corresponds to a separate character following the one at end of the current match
        //   (the second of these conditions, and perhaps both, may be redundant given the
        //   subsequent check for normalization boundary; however they are likely much faster
        //   tests in any case)
        // * the match limit is a normalization boundary
        UBool allowMidclusterMatch = FALSE;
        if (strsrch->search->text != NULL && strsrch->search->textLength > maxLimit) {
            allowMidclusterMatch =
                    strsrch->search->breakIter == NULL &&
                    nextCEI != NULL && (((nextCEI->ce) >> 32) & 0xFFFF0000UL) != 0 &&
                    maxLimit >= lastCEI->highIndex && nextCEI->highIndex > maxLimit &&
                    (strsrch->nfd->hasBoundaryBefore(codePointAt(*strsrch->search, maxLimit)) ||
                        strsrch->nfd->hasBoundaryAfter(codePointBefore(*strsrch->search, maxLimit)));
        }
        // If those conditions are met, then:
        // * do NOT advance the candidate match limit (mLimit) to a break boundary; however
        //   the match limit may be backed off to a previous break boundary. This handles
        //   cases in which mLimit includes target characters that are ignorable with current
        //   settings (such as space) and which extend beyond the pattern match.
        // * do NOT require that end of the combining sequence not extend beyond the match in CE space
        // * do NOT require that match limit be on a breakIter boundary

        //  Advance the match end position to the first acceptable match boundary.
        //    This advances the index over any combining charcters.
        mLimit = maxLimit;
        if (minLimit < maxLimit) {
            // When the last CE's low index is same with its high index, the CE is likely
            // a part of expansion. In this case, the index is located just after the
            // character corresponding to the CEs compared above. If the index is right
            // at the break boundary, move the position to the next boundary will result
            // incorrect match length when there are ignorable characters exist between
            // the position and the next character produces CE(s). See ticket#8482.
            if (minLimit == lastCEI->highIndex && isBreakBoundary(strsrch, minLimit)) {
                mLimit = minLimit;
            } else {
                int32_t nba = nextBoundaryAfter(strsrch, minLimit);
                // Note that we can have nba < maxLimit && nba >= minLImit, in which
                // case we want to set mLimit to nba regardless of allowMidclusterMatch
                // (i.e. we back off mLimit to the previous breakIterator boundary).
                if (nba >= lastCEI->highIndex && (!allowMidclusterMatch || nba < maxLimit)) {
                    mLimit = nba;
                }
            }
        }

    #ifdef USEARCH_DEBUG
        if (getenv("USEARCH_DEBUG") != NULL) {
            printf("minLimit, maxLimit, mLimit = %d, %d, %d\n", minLimit, maxLimit, mLimit);
        }
    #endif

        if (!allowMidclusterMatch) {
            // If advancing to the end of a combining sequence in character indexing space
            //   advanced us beyond the end of the match in CE space, reject this match.
            if (mLimit > maxLimit) {
                found = FALSE;
            }

            if (!isBreakBoundary(strsrch, mLimit)) {
                found = FALSE;
            }
        }

        if (! checkIdentical(strsrch, mStart, mLimit)) {
            found = FALSE;
        }

        if (found) {
            break;
        }
    }

    #ifdef USEARCH_DEBUG
    if (getenv("USEARCH_DEBUG") != NULL) {
        printf("Target CEs [%d .. %d]\n", ceb.firstIx, ceb.limitIx);
        int32_t  lastToPrint = ceb.limitIx+2;
        for (int ii=ceb.firstIx; ii<lastToPrint; ii++) {
            printf("%8x@%d ", ceb.get(ii)->ce, ceb.get(ii)->srcIndex);
        }
        printf("\n%s\n", found? "match found" : "no match");
    }
    #endif

    // All Done.  Store back the match bounds to the caller.
    //
    if (found==FALSE) {
        mLimit = -1;
        mStart = -1;
    }

    if (matchStart != NULL) {
        *matchStart= mStart;
    }

    if (matchLimit != NULL) {
        *matchLimit = mLimit;
    }

    return found;
}

U_CAPI UBool U_EXPORT2 usearch_searchBackwards(UStringSearch  *strsrch,
                                                int32_t        startIdx,
                                                int32_t        *matchStart,
                                                int32_t        *matchLimit,
                                                UErrorCode     *status)
{
    if (U_FAILURE(*status)) {
        return FALSE;
    }

    // TODO:  reject search patterns beginning with a combining char.

#ifdef USEARCH_DEBUG
    if (getenv("USEARCH_DEBUG") != NULL) {
        printf("Pattern CEs\n");
        for (int ii=0; ii<strsrch->pattern.cesLength; ii++) {
            printf(" %8x", strsrch->pattern.ces[ii]);
        }
        printf("\n");
    }

#endif
    // Input parameter sanity check.
    //  TODO:  should input indicies clip to the text length
    //         in the same way that UText does.
    if(strsrch->pattern.cesLength == 0         ||
       startIdx < 0                           ||
       startIdx > strsrch->search->textLength ||
       strsrch->pattern.ces == NULL) {
           *status = U_ILLEGAL_ARGUMENT_ERROR;
           return FALSE;
    }

    if (strsrch->pattern.pces == NULL) {
        initializePatternPCETable(strsrch, status);
    }

    CEIBuffer ceb(strsrch, status);
    int32_t    targetIx = 0;

    /*
     * Pre-load the buffer with the CE's for the grapheme
     * after our starting position so that we're sure that
     * we can look at the CE following the match when we
     * check the match boundaries.
     *
     * This will also pre-fetch the first CE that we'll
     * consider for the match.
     */
    if (startIdx < strsrch->search->textLength) {
        UBreakIterator *bi = strsrch->search->internalBreakIter;
        int32_t next = ubrk_following(bi, startIdx);

        ucol_setOffset(strsrch->textIter, next, status);

        for (targetIx = 0; ; targetIx += 1) {
            if (ceb.getPrevious(targetIx)->lowIndex < startIdx) {
                break;
            }
        }
    } else {
        ucol_setOffset(strsrch->textIter, startIdx, status);
    }


    const CEI *targetCEI = NULL;
    int32_t    patIx;
    UBool      found;

    int32_t  limitIx = targetIx;
    int32_t  mStart = -1;
    int32_t  mLimit = -1;
    int32_t  minLimit;
    int32_t  maxLimit;



    // Outer loop moves over match starting positions in the
    //      target CE space.
    // Here, targetIx values increase toward the beginning of the base text (i.e. we get the text CEs in reverse order).
    // But  patIx is 0 at the beginning of the pattern and increases toward the end.
    // So this loop performs a comparison starting with the end of pattern, and prcessd toward the beginning of the pattern
    // and the beginning of the base text.
    for(targetIx = limitIx; ; targetIx += 1)
    {
        found = TRUE;
        // For targetIx > limitIx, this ceb.getPrevious gets a CE that is as far back in the ring buffer
        // (compared to the last CE fetched for the previous targetIx value) as we need to go
        // for this targetIx value, so if it is non-NULL then other ceb.getPrevious calls should be OK.
        const CEI *lastCEI  = ceb.getPrevious(targetIx);
        if (lastCEI == NULL) {
            *status = U_INTERNAL_PROGRAM_ERROR;
            found = FALSE;
             break;
        }
        //  Inner loop checks for a match beginning at each
        //  position from the outer loop.
        int32_t targetIxOffset = 0;
        for (patIx = strsrch->pattern.pcesLength - 1; patIx >= 0; patIx -= 1) {
            int64_t patCE = strsrch->pattern.pces[patIx];

            targetCEI = ceb.getPrevious(targetIx + strsrch->pattern.pcesLength - 1 - patIx + targetIxOffset);
            //  Compare CE from target string with CE from the pattern.
            //    Note that the target CE will be UCOL_NULLORDER if we reach the end of input,
            //    which will fail the compare, below.
            UCompareCEsResult ceMatch = compareCE64s(targetCEI->ce, patCE, strsrch->search->elementComparisonType);
            if ( ceMatch == U_CE_NO_MATCH ) {
                found = FALSE;
                break;
            } else if ( ceMatch > U_CE_NO_MATCH ) {
                if ( ceMatch == U_CE_SKIP_TARG ) {
                    // redo with same patCE, next targCE
                    patIx++;
                    targetIxOffset++;
                } else { // ceMatch == U_CE_SKIP_PATN
                    // redo with same targCE, next patCE
                    targetIxOffset--;
                }
            }
        }

        if (!found && ((targetCEI == NULL) || (targetCEI->ce != UCOL_PROCESSED_NULLORDER))) {
            // No match at this targetIx.  Try again at the next.
            continue;
        }

        if (!found) {
            // No match at all, we have run off the end of the target text.
            break;
        }


        // We have found a match in CE space.
        // Now determine the bounds in string index space.
        //  There still is a chance of match failure if the CE range not correspond to
        //     an acceptable character range.
        //
        const CEI *firstCEI = ceb.getPrevious(targetIx + strsrch->pattern.pcesLength - 1 + targetIxOffset);
        mStart   = firstCEI->lowIndex;

        // Check for the start of the match being within a combining sequence.
        //   This can happen if the pattern itself begins with a combining char, and
        //   the match found combining marks in the target text that were attached
        //    to something else.
        //   This type of match should be rejected for not completely consuming a
        //   combining sequence.
        if (!isBreakBoundary(strsrch, mStart)) {
            found = FALSE;
        }

        // Look at the high index of the first CE in the match. If it's the same as the
        // low index, the first CE in the match is in the middle of an expansion.
        if (mStart == firstCEI->highIndex) {
            found = FALSE;
        }


        minLimit = lastCEI->lowIndex;

        if (targetIx > 0) {
            // Look at the CE following the match.  If it is UCOL_NULLORDER the match
            //   extended to the end of input, and the match is good.

            // Look at the high and low indices of the CE following the match. If
            // they are the same it means one of two things:
            //    1. The match extended to the last CE from the target text, which is OK, or
            //    2. The last CE that was part of the match is in an expansion that extends
            //       to the first CE after the match. In this case, we reject the match.
            const CEI *nextCEI  = ceb.getPrevious(targetIx - 1);

            if (nextCEI->lowIndex == nextCEI->highIndex && nextCEI->ce != UCOL_PROCESSED_NULLORDER) {
                found = FALSE;
            }

            mLimit = maxLimit = nextCEI->lowIndex;

            // Allow matches to end in the middle of a grapheme cluster if the following
            // conditions are met; this is needed to make prefix search work properly in
            // Indic, see #11750
            // * the default breakIter is being used
            // * the next collation element after this combining sequence
            //   - has non-zero primary weight
            //   - corresponds to a separate character following the one at end of the current match
            //   (the second of these conditions, and perhaps both, may be redundant given the
            //   subsequent check for normalization boundary; however they are likely much faster
            //   tests in any case)
            // * the match limit is a normalization boundary
            UBool allowMidclusterMatch = FALSE;
            if (strsrch->search->text != NULL && strsrch->search->textLength > maxLimit) {
                allowMidclusterMatch =
                        strsrch->search->breakIter == NULL &&
                        nextCEI != NULL && (((nextCEI->ce) >> 32) & 0xFFFF0000UL) != 0 &&
                        maxLimit >= lastCEI->highIndex && nextCEI->highIndex > maxLimit &&
                        (strsrch->nfd->hasBoundaryBefore(codePointAt(*strsrch->search, maxLimit)) ||
                            strsrch->nfd->hasBoundaryAfter(codePointBefore(*strsrch->search, maxLimit)));
            }
            // If those conditions are met, then:
            // * do NOT advance the candidate match limit (mLimit) to a break boundary; however
            //   the match limit may be backed off to a previous break boundary. This handles
            //   cases in which mLimit includes target characters that are ignorable with current
            //   settings (such as space) and which extend beyond the pattern match.
            // * do NOT require that end of the combining sequence not extend beyond the match in CE space
            // * do NOT require that match limit be on a breakIter boundary

            //  Advance the match end position to the first acceptable match boundary.
            //    This advances the index over any combining characters.
            if (minLimit < maxLimit) {
                int32_t nba = nextBoundaryAfter(strsrch, minLimit);
                // Note that we can have nba < maxLimit && nba >= minLImit, in which
                // case we want to set mLimit to nba regardless of allowMidclusterMatch
                // (i.e. we back off mLimit to the previous breakIterator boundary).
                if (nba >= lastCEI->highIndex && (!allowMidclusterMatch || nba < maxLimit)) {
                    mLimit = nba;
                }
            }

            if (!allowMidclusterMatch) {
                // If advancing to the end of a combining sequence in character indexing space
                //   advanced us beyond the end of the match in CE space, reject this match.
                if (mLimit > maxLimit) {
                    found = FALSE;
                }

                // Make sure the end of the match is on a break boundary
                if (!isBreakBoundary(strsrch, mLimit)) {
                    found = FALSE;
                }
            }

        } else {
            // No non-ignorable CEs after this point.
            // The maximum position is detected by boundary after
            // the last non-ignorable CE. Combining sequence
            // across the start index will be truncated.
            int32_t nba = nextBoundaryAfter(strsrch, minLimit);
            mLimit = maxLimit = (nba > 0) && (startIdx > nba) ? nba : startIdx;
        }

    #ifdef USEARCH_DEBUG
        if (getenv("USEARCH_DEBUG") != NULL) {
            printf("minLimit, maxLimit, mLimit = %d, %d, %d\n", minLimit, maxLimit, mLimit);
        }
    #endif


        if (! checkIdentical(strsrch, mStart, mLimit)) {
            found = FALSE;
        }

        if (found) {
            break;
        }
    }

    #ifdef USEARCH_DEBUG
    if (getenv("USEARCH_DEBUG") != NULL) {
        printf("Target CEs [%d .. %d]\n", ceb.firstIx, ceb.limitIx);
        int32_t  lastToPrint = ceb.limitIx+2;
        for (int ii=ceb.firstIx; ii<lastToPrint; ii++) {
            printf("%8x@%d ", ceb.get(ii)->ce, ceb.get(ii)->srcIndex);
        }
        printf("\n%s\n", found? "match found" : "no match");
    }
    #endif

    // All Done.  Store back the match bounds to the caller.
    //
    if (found==FALSE) {
        mLimit = -1;
        mStart = -1;
    }

    if (matchStart != NULL) {
        *matchStart= mStart;
    }

    if (matchLimit != NULL) {
        *matchLimit = mLimit;
    }

    return found;
}

// internal use methods declared in usrchimp.h -----------------------------

UBool usearch_handleNextExact(UStringSearch *strsrch, UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        setMatchNotFound(strsrch);
        return FALSE;
    }

#if BOYER_MOORE
    UCollationElements *coleiter        = strsrch->textIter;
    int32_t             textlength      = strsrch->search->textLength;
    int32_t            *patternce       = strsrch->pattern.ces;
    int32_t             patterncelength = strsrch->pattern.cesLength;
    int32_t             textoffset      = ucol_getOffset(coleiter);

    // status used in setting coleiter offset, since offset is checked in
    // shiftForward before setting the coleiter offset, status never
    // a failure
    textoffset = shiftForward(strsrch, textoffset, UCOL_NULLORDER,
                              patterncelength);
    while (textoffset <= textlength)
    {
        uint32_t    patternceindex = patterncelength - 1;
        int32_t     targetce;
        UBool       found          = FALSE;
        int32_t    lastce          = UCOL_NULLORDER;

        setColEIterOffset(coleiter, textoffset);

        for (;;) {
            // finding the last pattern ce match, imagine composite characters
            // for example: search for pattern A in text \u00C0
            // we'll have to skip \u0300 the grave first before we get to A
            targetce = ucol_previous(coleiter, status);
            if (U_FAILURE(*status) || targetce == UCOL_NULLORDER) {
                found = FALSE;
                break;
            }
            targetce = getCE(strsrch, targetce);
            if (targetce == UCOL_IGNORABLE && inNormBuf(coleiter)) {
                // this is for the text \u0315\u0300 that requires
                // normalization and pattern \u0300, where \u0315 is ignorable
                continue;
            }
            if (lastce == UCOL_NULLORDER || lastce == UCOL_IGNORABLE) {
                lastce = targetce;
            }
            // TODO: #if BOYER_MOORE, replace with code using 32-bit version of compareCE64s
            if (targetce == patternce[patternceindex]) {
                // the first ce can be a contraction
                found = TRUE;
                break;
            }
            if (!hasExpansion(coleiter)) {
                found = FALSE;
                break;
            }
        }

        //targetce = lastce;

        while (found && patternceindex > 0) {
            lastce = targetce;
            targetce    = ucol_previous(coleiter, status);
            if (U_FAILURE(*status) || targetce == UCOL_NULLORDER) {
                found = FALSE;
                break;
            }
            targetce    = getCE(strsrch, targetce);
            if (targetce == UCOL_IGNORABLE) {
                continue;
            }

            patternceindex --;
            // TODO: #if BOYER_MOORE, replace with code using 32-bit version of compareCE64s
            found = found && targetce == patternce[patternceindex];
        }

        targetce = lastce;

        if (!found) {
            if (U_FAILURE(*status)) {
                break;
            }
            textoffset = shiftForward(strsrch, textoffset, lastce,
                                      patternceindex);
            // status checked at loop.
            patternceindex = patterncelength;
            continue;
        }

        if (checkNextExactMatch(strsrch, &textoffset, status)) {
            // status checked in ucol_setOffset
            setColEIterOffset(coleiter, strsrch->search->matchedIndex);
            return TRUE;
        }
    }
    setMatchNotFound(strsrch);
    return FALSE;
#else
    int32_t textOffset = ucol_getOffset(strsrch->textIter);
    int32_t start = -1;
    int32_t end = -1;

    if (usearch_search(strsrch, textOffset, &start, &end, status)) {
        strsrch->search->matchedIndex  = start;
        strsrch->search->matchedLength = end - start;
        return TRUE;
    } else {
        setMatchNotFound(strsrch);
        return FALSE;
    }
#endif
}

UBool usearch_handleNextCanonical(UStringSearch *strsrch, UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        setMatchNotFound(strsrch);
        return FALSE;
    }

#if BOYER_MOORE
    UCollationElements *coleiter        = strsrch->textIter;
    int32_t             textlength      = strsrch->search->textLength;
    int32_t            *patternce       = strsrch->pattern.ces;
    int32_t             patterncelength = strsrch->pattern.cesLength;
    int32_t             textoffset      = ucol_getOffset(coleiter);
    UBool               hasPatternAccents =
       strsrch->pattern.hasSuffixAccents || strsrch->pattern.hasPrefixAccents;

    textoffset = shiftForward(strsrch, textoffset, UCOL_NULLORDER,
                              patterncelength);
    strsrch->canonicalPrefixAccents[0] = 0;
    strsrch->canonicalSuffixAccents[0] = 0;

    while (textoffset <= textlength)
    {
        int32_t     patternceindex = patterncelength - 1;
        int32_t     targetce;
        UBool       found          = FALSE;
        int32_t     lastce         = UCOL_NULLORDER;

        setColEIterOffset(coleiter, textoffset);

        for (;;) {
            // finding the last pattern ce match, imagine composite characters
            // for example: search for pattern A in text \u00C0
            // we'll have to skip \u0300 the grave first before we get to A
            targetce = ucol_previous(coleiter, status);
            if (U_FAILURE(*status) || targetce == UCOL_NULLORDER) {
                found = FALSE;
                break;
            }
            targetce = getCE(strsrch, targetce);
            if (lastce == UCOL_NULLORDER || lastce == UCOL_IGNORABLE) {
                lastce = targetce;
            }
            // TODO: #if BOYER_MOORE, replace with code using 32-bit version of compareCE64s
            if (targetce == patternce[patternceindex]) {
                // the first ce can be a contraction
                found = TRUE;
                break;
            }
            if (!hasExpansion(coleiter)) {
                found = FALSE;
                break;
            }
        }

        while (found && patternceindex > 0) {
            targetce    = ucol_previous(coleiter, status);
            if (U_FAILURE(*status) || targetce == UCOL_NULLORDER) {
                found = FALSE;
                break;
            }
            targetce    = getCE(strsrch, targetce);
            if (targetce == UCOL_IGNORABLE) {
                continue;
            }

            patternceindex --;
            // TODO: #if BOYER_MOORE, replace with code using 32-bit version of compareCE64s
            found = found && targetce == patternce[patternceindex];
        }

        // initializing the rearranged accent array
        if (hasPatternAccents && !found) {
            strsrch->canonicalPrefixAccents[0] = 0;
            strsrch->canonicalSuffixAccents[0] = 0;
            if (U_FAILURE(*status)) {
                break;
            }
            found = doNextCanonicalMatch(strsrch, textoffset, status);
        }

        if (!found) {
            if (U_FAILURE(*status)) {
                break;
            }
            textoffset = shiftForward(strsrch, textoffset, lastce,
                                      patternceindex);
            // status checked at loop
            patternceindex = patterncelength;
            continue;
        }

        if (checkNextCanonicalMatch(strsrch, &textoffset, status)) {
            setColEIterOffset(coleiter, strsrch->search->matchedIndex);
            return TRUE;
        }
    }
    setMatchNotFound(strsrch);
    return FALSE;
#else
    int32_t textOffset = ucol_getOffset(strsrch->textIter);
    int32_t start = -1;
    int32_t end = -1;

    if (usearch_search(strsrch, textOffset, &start, &end, status)) {
        strsrch->search->matchedIndex  = start;
        strsrch->search->matchedLength = end - start;
        return TRUE;
    } else {
        setMatchNotFound(strsrch);
        return FALSE;
    }
#endif
}

UBool usearch_handlePreviousExact(UStringSearch *strsrch, UErrorCode *status)
{
    if (U_FAILURE(*status)) {
        setMatchNotFound(strsrch);
        return FALSE;
    }

#if BOYER_MOORE
    UCollationElements *coleiter        = strsrch->textIter;
    int32_t            *patternce       = strsrch->pattern.ces;
    int32_t             patterncelength = strsrch->pattern.cesLength;
    int32_t             textoffset      = ucol_getOffset(coleiter);

    // shifting it check for setting offset
    // if setOffset is called previously or there was no previous match, we
    // leave the offset as it is.
    if (strsrch->search->matchedIndex != USEARCH_DONE) {
        textoffset = strsrch->search->matchedIndex;
    }

    textoffset = reverseShift(strsrch, textoffset, UCOL_NULLORDER,
                              patterncelength);

    while (textoffset >= 0)
    {
        int32_t     patternceindex = 1;
        int32_t     targetce;
        UBool       found          = FALSE;
        int32_t     firstce        = UCOL_NULLORDER;

        // if status is a failure, ucol_setOffset does nothing
        setColEIterOffset(coleiter, textoffset);

        for (;;) {
            // finding the first pattern ce match, imagine composite
            // characters. for example: search for pattern \u0300 in text
            // \u00C0, we'll have to skip A first before we get to
            // \u0300 the grave accent
            targetce = ucol_next(coleiter, status);
            if (U_FAILURE(*status) || targetce == UCOL_NULLORDER) {
                found = FALSE;
                break;
            }
            targetce = getCE(strsrch, targetce);
            if (firstce == UCOL_NULLORDER || firstce == UCOL_IGNORABLE) {
                firstce = targetce;
            }
            if (targetce == UCOL_IGNORABLE && strsrch->strength != UCOL_PRIMARY) {
                continue;
            }
            // TODO: #if BOYER_MOORE, replace with code using 32-bit version of compareCE64s
            if (targetce == patternce[0]) {
                found = TRUE;
                break;
            }
            if (!hasExpansion(coleiter)) {
                // checking for accents in composite character
                found = FALSE;
                break;
            }
        }

        //targetce = firstce;

        while (found && (patternceindex < patterncelength)) {
            firstce = targetce;
            targetce    = ucol_next(coleiter, status);
            if (U_FAILURE(*status) || targetce == UCOL_NULLORDER) {
                found = FALSE;
                break;
            }
            targetce    = getCE(strsrch, targetce);
            if (targetce == UCOL_IGNORABLE) {
                continue;
            }

            // TODO: #if BOYER_MOORE, replace with code using 32-bit version of compareCE64s
            found = found && targetce == patternce[patternceindex];
            patternceindex ++;
        }

        targetce = firstce;

        if (!found) {
            if (U_FAILURE(*status)) {
                break;
            }

            textoffset = reverseShift(strsrch, textoffset, targetce,
                                      patternceindex);
            patternceindex = 0;
            continue;
        }

        if (checkPreviousExactMatch(strsrch, &textoffset, status)) {
            setColEIterOffset(coleiter, textoffset);
            return TRUE;
        }
    }
    setMatchNotFound(strsrch);
    return FALSE;
#else
    int32_t textOffset;

    if (strsrch->search->isOverlap) {
        if (strsrch->search->matchedIndex != USEARCH_DONE) {
            textOffset = strsrch->search->matchedIndex + strsrch->search->matchedLength - 1;
        } else {
            // move the start position at the end of possible match
            initializePatternPCETable(strsrch, status);
            if (!initTextProcessedIter(strsrch, status)) {
                setMatchNotFound(strsrch);
                return FALSE;
            }
            for (int32_t nPCEs = 0; nPCEs < strsrch->pattern.pcesLength - 1; nPCEs++) {
                int64_t pce = strsrch->textProcessedIter->nextProcessed(NULL, NULL, status);
                if (pce == UCOL_PROCESSED_NULLORDER) {
                    // at the end of the text
                    break;
                }
            }
            if (U_FAILURE(*status)) {
                setMatchNotFound(strsrch);
                return FALSE;
            }
            textOffset = ucol_getOffset(strsrch->textIter);
        }
    } else {
        textOffset = ucol_getOffset(strsrch->textIter);
    }

    int32_t start = -1;
    int32_t end = -1;

    if (usearch_searchBackwards(strsrch, textOffset, &start, &end, status)) {
        strsrch->search->matchedIndex = start;
        strsrch->search->matchedLength = end - start;
        return TRUE;
    } else {
        setMatchNotFound(strsrch);
        return FALSE;
    }
#endif
}

UBool usearch_handlePreviousCanonical(UStringSearch *strsrch,
                                      UErrorCode    *status)
{
    if (U_FAILURE(*status)) {
        setMatchNotFound(strsrch);
        return FALSE;
    }

#if BOYER_MOORE
    UCollationElements *coleiter        = strsrch->textIter;
    int32_t            *patternce       = strsrch->pattern.ces;
    int32_t             patterncelength = strsrch->pattern.cesLength;
    int32_t             textoffset      = ucol_getOffset(coleiter);
    UBool               hasPatternAccents =
       strsrch->pattern.hasSuffixAccents || strsrch->pattern.hasPrefixAccents;

    // shifting it check for setting offset
    // if setOffset is called previously or there was no previous match, we
    // leave the offset as it is.
    if (strsrch->search->matchedIndex != USEARCH_DONE) {
        textoffset = strsrch->search->matchedIndex;
    }

    textoffset = reverseShift(strsrch, textoffset, UCOL_NULLORDER,
                              patterncelength);
    strsrch->canonicalPrefixAccents[0] = 0;
    strsrch->canonicalSuffixAccents[0] = 0;

    while (textoffset >= 0)
    {
        int32_t     patternceindex = 1;
        int32_t     targetce;
        UBool       found          = FALSE;
        int32_t     firstce        = UCOL_NULLORDER;

        setColEIterOffset(coleiter, textoffset);
        for (;;) {
            // finding the first pattern ce match, imagine composite
            // characters. for example: search for pattern \u0300 in text
            // \u00C0, we'll have to skip A first before we get to
            // \u0300 the grave accent
            targetce = ucol_next(coleiter, status);
            if (U_FAILURE(*status) || targetce == UCOL_NULLORDER) {
                found = FALSE;
                break;
            }
            targetce = getCE(strsrch, targetce);
            if (firstce == UCOL_NULLORDER || firstce == UCOL_IGNORABLE) {
                firstce = targetce;
            }

            // TODO: #if BOYER_MOORE, replace with code using 32-bit version of compareCE64s
            if (targetce == patternce[0]) {
                // the first ce can be a contraction
                found = TRUE;
                break;
            }
            if (!hasExpansion(coleiter)) {
                // checking for accents in composite character
                found = FALSE;
                break;
            }
        }

        targetce = firstce;

        while (found && patternceindex < patterncelength) {
            targetce    = ucol_next(coleiter, status);
            if (U_FAILURE(*status) || targetce == UCOL_NULLORDER) {
                found = FALSE;
                break;
            }
            targetce = getCE(strsrch, targetce);
            if (targetce == UCOL_IGNORABLE) {
                continue;
            }

            // TODO: #if BOYER_MOORE, replace with code using 32-bit version of compareCE64s
            found = found && targetce == patternce[patternceindex];
            patternceindex ++;
        }

        // initializing the rearranged accent array
        if (hasPatternAccents && !found) {
            strsrch->canonicalPrefixAccents[0] = 0;
            strsrch->canonicalSuffixAccents[0] = 0;
            if (U_FAILURE(*status)) {
                break;
            }
            found = doPreviousCanonicalMatch(strsrch, textoffset, status);
        }

        if (!found) {
            if (U_FAILURE(*status)) {
                break;
            }
            textoffset = reverseShift(strsrch, textoffset, targetce,
                                      patternceindex);
            patternceindex = 0;
            continue;
        }

        if (checkPreviousCanonicalMatch(strsrch, &textoffset, status)) {
            setColEIterOffset(coleiter, textoffset);
            return TRUE;
        }
    }
    setMatchNotFound(strsrch);
    return FALSE;
#else
    int32_t textOffset;

    if (strsrch->search->isOverlap) {
        if (strsrch->search->matchedIndex != USEARCH_DONE) {
            textOffset = strsrch->search->matchedIndex + strsrch->search->matchedLength - 1;
        } else {
            // move the start position at the end of possible match
            initializePatternPCETable(strsrch, status);
            if (!initTextProcessedIter(strsrch, status)) {
                setMatchNotFound(strsrch);
                return FALSE;
            }
            for (int32_t nPCEs = 0; nPCEs < strsrch->pattern.pcesLength - 1; nPCEs++) {
                int64_t pce = strsrch->textProcessedIter->nextProcessed(NULL, NULL, status);
                if (pce == UCOL_PROCESSED_NULLORDER) {
                    // at the end of the text
                    break;
                }
            }
            if (U_FAILURE(*status)) {
                setMatchNotFound(strsrch);
                return FALSE;
            }
            textOffset = ucol_getOffset(strsrch->textIter);
        }
    } else {
        textOffset = ucol_getOffset(strsrch->textIter);
    }

    int32_t start = -1;
    int32_t end = -1;

    if (usearch_searchBackwards(strsrch, textOffset, &start, &end, status)) {
        strsrch->search->matchedIndex = start;
        strsrch->search->matchedLength = end - start;
        return TRUE;
    } else {
        setMatchNotFound(strsrch);
        return FALSE;
    }
#endif
}

#endif /* #if !UCONFIG_NO_COLLATION */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2010-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* utf16collationiterator.cpp
*
* created on: 2010oct27
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION











U_NAMESPACE_BEGIN

UTF16CollationIterator::UTF16CollationIterator(const UTF16CollationIterator &other,
                                               const UChar *newText)
        : CollationIterator(other),
          start(newText),
          pos(newText + (other.pos - other.start)),
          limit(other.limit == NULL ? NULL : newText + (other.limit - other.start)) {
}

UTF16CollationIterator::~UTF16CollationIterator() {}

UBool
UTF16CollationIterator::operator==(const CollationIterator &other) const {
    if(!CollationIterator::operator==(other)) { return FALSE; }
    const UTF16CollationIterator &o = static_cast<const UTF16CollationIterator &>(other);
    // Compare the iterator state but not the text: Assume that the caller does that.
    return (pos - start) == (o.pos - o.start);
}

void
UTF16CollationIterator::resetToOffset(int32_t newOffset) {
    reset();
    pos = start + newOffset;
}

int32_t
UTF16CollationIterator::getOffset() const {
    return (int32_t)(pos - start);
}

uint32_t
UTF16CollationIterator::handleNextCE32(UChar32 &c, UErrorCode & /*errorCode*/) {
    if(pos == limit) {
        c = U_SENTINEL;
        return Collation::FALLBACK_CE32;
    }
    c = *pos++;
    return UTRIE2_GET32_FROM_U16_SINGLE_LEAD(trie, c);
}

UChar
UTF16CollationIterator::handleGetTrailSurrogate() {
    if(pos == limit) { return 0; }
    UChar trail;
    if(U16_IS_TRAIL(trail = *pos)) { ++pos; }
    return trail;
}

UBool
UTF16CollationIterator::foundNULTerminator() {
    if(limit == NULL) {
        limit = --pos;
        return TRUE;
    } else {
        return FALSE;
    }
}

UChar32
UTF16CollationIterator::nextCodePoint(UErrorCode & /*errorCode*/) {
    if(pos == limit) {
        return U_SENTINEL;
    }
    UChar32 c = *pos;
    if(c == 0 && limit == NULL) {
        limit = pos;
        return U_SENTINEL;
    }
    ++pos;
    UChar trail;
    if(U16_IS_LEAD(c) && pos != limit && U16_IS_TRAIL(trail = *pos)) {
        ++pos;
        return U16_GET_SUPPLEMENTARY(c, trail);
    } else {
        return c;
    }
}

UChar32
UTF16CollationIterator::previousCodePoint(UErrorCode & /*errorCode*/) {
    if(pos == start) {
        return U_SENTINEL;
    }
    UChar32 c = *--pos;
    UChar lead;
    if(U16_IS_TRAIL(c) && pos != start && U16_IS_LEAD(lead = *(pos - 1))) {
        --pos;
        return U16_GET_SUPPLEMENTARY(lead, c);
    } else {
        return c;
    }
}

void
UTF16CollationIterator::forwardNumCodePoints(int32_t num, UErrorCode & /*errorCode*/) {
    while(num > 0 && pos != limit) {
        UChar32 c = *pos;
        if(c == 0 && limit == NULL) {
            limit = pos;
            break;
        }
        ++pos;
        --num;
        if(U16_IS_LEAD(c) && pos != limit && U16_IS_TRAIL(*pos)) {
            ++pos;
        }
    }
}

void
UTF16CollationIterator::backwardNumCodePoints(int32_t num, UErrorCode & /*errorCode*/) {
    while(num > 0 && pos != start) {
        UChar32 c = *--pos;
        --num;
        if(U16_IS_TRAIL(c) && pos != start && U16_IS_LEAD(*(pos-1))) {
            --pos;
        }
    }
}

// FCDUTF16CollationIterator ----------------------------------------------- ***

FCDUTF16CollationIterator::FCDUTF16CollationIterator(const FCDUTF16CollationIterator &other,
                                                     const UChar *newText)
        : UTF16CollationIterator(other),
          rawStart(newText),
          segmentStart(newText + (other.segmentStart - other.rawStart)),
          segmentLimit(other.segmentLimit == NULL ? NULL : newText + (other.segmentLimit - other.rawStart)),
          rawLimit(other.rawLimit == NULL ? NULL : newText + (other.rawLimit - other.rawStart)),
          nfcImpl(other.nfcImpl),
          normalized(other.normalized),
          checkDir(other.checkDir) {
    if(checkDir != 0 || other.start == other.segmentStart) {
        start = newText + (other.start - other.rawStart);
        pos = newText + (other.pos - other.rawStart);
        limit = other.limit == NULL ? NULL : newText + (other.limit - other.rawStart);
    } else {
        start = normalized.getBuffer();
        pos = start + (other.pos - other.start);
        limit = start + normalized.length();
    }
}

FCDUTF16CollationIterator::~FCDUTF16CollationIterator() {}

UBool
FCDUTF16CollationIterator::operator==(const CollationIterator &other) const {
    // Skip the UTF16CollationIterator and call its parent.
    if(!CollationIterator::operator==(other)) { return FALSE; }
    const FCDUTF16CollationIterator &o = static_cast<const FCDUTF16CollationIterator &>(other);
    // Compare the iterator state but not the text: Assume that the caller does that.
    if(checkDir != o.checkDir) { return FALSE; }
    if(checkDir == 0 && (start == segmentStart) != (o.start == o.segmentStart)) { return FALSE; }
    if(checkDir != 0 || start == segmentStart) {
        return (pos - rawStart) == (o.pos - o.rawStart);
    } else {
        return (segmentStart - rawStart) == (o.segmentStart - o.rawStart) &&
                (pos - start) == (o.pos - o.start);
    }
}

void
FCDUTF16CollationIterator::resetToOffset(int32_t newOffset) {
    reset();
    start = segmentStart = pos = rawStart + newOffset;
    limit = rawLimit;
    checkDir = 1;
}

int32_t
FCDUTF16CollationIterator::getOffset() const {
    if(checkDir != 0 || start == segmentStart) {
        return (int32_t)(pos - rawStart);
    } else if(pos == start) {
        return (int32_t)(segmentStart - rawStart);
    } else {
        return (int32_t)(segmentLimit - rawStart);
    }
}

uint32_t
FCDUTF16CollationIterator::handleNextCE32(UChar32 &c, UErrorCode &errorCode) {
    for(;;) {
        if(checkDir > 0) {
            if(pos == limit) {
                c = U_SENTINEL;
                return Collation::FALLBACK_CE32;
            }
            c = *pos++;
            if(CollationFCD::hasTccc(c)) {
                if(CollationFCD::maybeTibetanCompositeVowel(c) ||
                        (pos != limit && CollationFCD::hasLccc(*pos))) {
                    --pos;
                    if(!nextSegment(errorCode)) {
                        c = U_SENTINEL;
                        return Collation::FALLBACK_CE32;
                    }
                    c = *pos++;
                }
            }
            break;
        } else if(checkDir == 0 && pos != limit) {
            c = *pos++;
            break;
        } else {
            switchToForward();
        }
    }
    return UTRIE2_GET32_FROM_U16_SINGLE_LEAD(trie, c);
}

UBool
FCDUTF16CollationIterator::foundNULTerminator() {
    if(limit == NULL) {
        limit = rawLimit = --pos;
        return TRUE;
    } else {
        return FALSE;
    }
}

UChar32
FCDUTF16CollationIterator::nextCodePoint(UErrorCode &errorCode) {
    UChar32 c;
    for(;;) {
        if(checkDir > 0) {
            if(pos == limit) {
                return U_SENTINEL;
            }
            c = *pos++;
            if(CollationFCD::hasTccc(c)) {
                if(CollationFCD::maybeTibetanCompositeVowel(c) ||
                        (pos != limit && CollationFCD::hasLccc(*pos))) {
                    --pos;
                    if(!nextSegment(errorCode)) {
                        return U_SENTINEL;
                    }
                    c = *pos++;
                }
            } else if(c == 0 && limit == NULL) {
                limit = rawLimit = --pos;
                return U_SENTINEL;
            }
            break;
        } else if(checkDir == 0 && pos != limit) {
            c = *pos++;
            break;
        } else {
            switchToForward();
        }
    }
    UChar trail;
    if(U16_IS_LEAD(c) && pos != limit && U16_IS_TRAIL(trail = *pos)) {
        ++pos;
        return U16_GET_SUPPLEMENTARY(c, trail);
    } else {
        return c;
    }
}

UChar32
FCDUTF16CollationIterator::previousCodePoint(UErrorCode &errorCode) {
    UChar32 c;
    for(;;) {
        if(checkDir < 0) {
            if(pos == start) {
                return U_SENTINEL;
            }
            c = *--pos;
            if(CollationFCD::hasLccc(c)) {
                if(CollationFCD::maybeTibetanCompositeVowel(c) ||
                        (pos != start && CollationFCD::hasTccc(*(pos - 1)))) {
                    ++pos;
                    if(!previousSegment(errorCode)) {
                        return U_SENTINEL;
                    }
                    c = *--pos;
                }
            }
            break;
        } else if(checkDir == 0 && pos != start) {
            c = *--pos;
            break;
        } else {
            switchToBackward();
        }
    }
    UChar lead;
    if(U16_IS_TRAIL(c) && pos != start && U16_IS_LEAD(lead = *(pos - 1))) {
        --pos;
        return U16_GET_SUPPLEMENTARY(lead, c);
    } else {
        return c;
    }
}

void
FCDUTF16CollationIterator::forwardNumCodePoints(int32_t num, UErrorCode &errorCode) {
    // Specify the class to avoid a virtual-function indirection.
    // In Java, we would declare this class final.
    while(num > 0 && FCDUTF16CollationIterator::nextCodePoint(errorCode) >= 0) {
        --num;
    }
}

void
FCDUTF16CollationIterator::backwardNumCodePoints(int32_t num, UErrorCode &errorCode) {
    // Specify the class to avoid a virtual-function indirection.
    // In Java, we would declare this class final.
    while(num > 0 && FCDUTF16CollationIterator::previousCodePoint(errorCode) >= 0) {
        --num;
    }
}

void
FCDUTF16CollationIterator::switchToForward() {
    U_ASSERT(checkDir < 0 || (checkDir == 0 && pos == limit));
    if(checkDir < 0) {
        // Turn around from backward checking.
        start = segmentStart = pos;
        if(pos == segmentLimit) {
            limit = rawLimit;
            checkDir = 1;  // Check forward.
        } else {  // pos < segmentLimit
            checkDir = 0;  // Stay in FCD segment.
        }
    } else {
        // Reached the end of the FCD segment.
        if(start == segmentStart) {
            // The input text segment is FCD, extend it forward.
        } else {
            // The input text segment needed to be normalized.
            // Switch to checking forward from it.
            pos = start = segmentStart = segmentLimit;
            // Note: If this segment is at the end of the input text,
            // then it might help to return FALSE to indicate that, so that
            // we do not have to re-check and normalize when we turn around and go backwards.
            // However, that would complicate the call sites for an optimization of an unusual case.
        }
        limit = rawLimit;
        checkDir = 1;
    }
}

UBool
FCDUTF16CollationIterator::nextSegment(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    U_ASSERT(checkDir > 0 && pos != limit);
    // The input text [segmentStart..pos[ passes the FCD check.
    const UChar *p = pos;
    uint8_t prevCC = 0;
    for(;;) {
        // Fetch the next character's fcd16 value.
        const UChar *q = p;
        uint16_t fcd16 = nfcImpl.nextFCD16(p, rawLimit);
        uint8_t leadCC = (uint8_t)(fcd16 >> 8);
        if(leadCC == 0 && q != pos) {
            // FCD boundary before the [q, p[ character.
            limit = segmentLimit = q;
            break;
        }
        if(leadCC != 0 && (prevCC > leadCC || CollationFCD::isFCD16OfTibetanCompositeVowel(fcd16))) {
            // Fails FCD check. Find the next FCD boundary and normalize.
            do {
                q = p;
            } while(p != rawLimit && nfcImpl.nextFCD16(p, rawLimit) > 0xff);
            if(!normalize(pos, q, errorCode)) { return FALSE; }
            pos = start;
            break;
        }
        prevCC = (uint8_t)fcd16;
        if(p == rawLimit || prevCC == 0) {
            // FCD boundary after the last character.
            limit = segmentLimit = p;
            break;
        }
    }
    U_ASSERT(pos != limit);
    checkDir = 0;
    return TRUE;
}

void
FCDUTF16CollationIterator::switchToBackward() {
    U_ASSERT(checkDir > 0 || (checkDir == 0 && pos == start));
    if(checkDir > 0) {
        // Turn around from forward checking.
        limit = segmentLimit = pos;
        if(pos == segmentStart) {
            start = rawStart;
            checkDir = -1;  // Check backward.
        } else {  // pos > segmentStart
            checkDir = 0;  // Stay in FCD segment.
        }
    } else {
        // Reached the start of the FCD segment.
        if(start == segmentStart) {
            // The input text segment is FCD, extend it backward.
        } else {
            // The input text segment needed to be normalized.
            // Switch to checking backward from it.
            pos = limit = segmentLimit = segmentStart;
        }
        start = rawStart;
        checkDir = -1;
    }
}

UBool
FCDUTF16CollationIterator::previousSegment(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    U_ASSERT(checkDir < 0 && pos != start);
    // The input text [pos..segmentLimit[ passes the FCD check.
    const UChar *p = pos;
    uint8_t nextCC = 0;
    for(;;) {
        // Fetch the previous character's fcd16 value.
        const UChar *q = p;
        uint16_t fcd16 = nfcImpl.previousFCD16(rawStart, p);
        uint8_t trailCC = (uint8_t)fcd16;
        if(trailCC == 0 && q != pos) {
            // FCD boundary after the [p, q[ character.
            start = segmentStart = q;
            break;
        }
        if(trailCC != 0 && ((nextCC != 0 && trailCC > nextCC) ||
                            CollationFCD::isFCD16OfTibetanCompositeVowel(fcd16))) {
            // Fails FCD check. Find the previous FCD boundary and normalize.
            do {
                q = p;
            } while(fcd16 > 0xff && p != rawStart &&
                    (fcd16 = nfcImpl.previousFCD16(rawStart, p)) != 0);
            if(!normalize(q, pos, errorCode)) { return FALSE; }
            pos = limit;
            break;
        }
        nextCC = (uint8_t)(fcd16 >> 8);
        if(p == rawStart || nextCC == 0) {
            // FCD boundary before the following character.
            start = segmentStart = p;
            break;
        }
    }
    U_ASSERT(pos != start);
    checkDir = 0;
    return TRUE;
}

UBool
FCDUTF16CollationIterator::normalize(const UChar *from, const UChar *to, UErrorCode &errorCode) {
    // NFD without argument checking.
    U_ASSERT(U_SUCCESS(errorCode));
    nfcImpl.decompose(from, to, normalized, (int32_t)(to - from), errorCode);
    if(U_FAILURE(errorCode)) { return FALSE; }
    // Switch collation processing into the FCD buffer
    // with the result of normalizing [segmentStart, segmentLimit[.
    segmentStart = from;
    segmentLimit = to;
    start = normalized.getBuffer();
    limit = start + normalized.length();
    return TRUE;
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2012-2014, International Business Machines
* Corporation and others.  All Rights Reserved.
*******************************************************************************
* utf8collationiterator.cpp
*
* created on: 2012nov12 (from utf16collationiterator.cpp & uitercollationiterator.cpp)
* created by: Markus W. Scherer
*/



#if !UCONFIG_NO_COLLATION












U_NAMESPACE_BEGIN

UTF8CollationIterator::~UTF8CollationIterator() {}

void
UTF8CollationIterator::resetToOffset(int32_t newOffset) {
    reset();
    pos = newOffset;
}

int32_t
UTF8CollationIterator::getOffset() const {
    return pos;
}

uint32_t
UTF8CollationIterator::handleNextCE32(UChar32 &c, UErrorCode & /*errorCode*/) {
    if(pos == length) {
        c = U_SENTINEL;
        return Collation::FALLBACK_CE32;
    }
    // Optimized combination of U8_NEXT_OR_FFFD() and UTRIE2_U8_NEXT32().
    c = u8[pos++];
    if(U8_IS_SINGLE(c)) {
        // ASCII 00..7F
        return trie->data32[c];
    }
    uint8_t t1, t2;
    if(0xe0 <= c && c < 0xf0 &&
            ((pos + 1) < length || length < 0) &&
            U8_IS_VALID_LEAD3_AND_T1(c, t1 = u8[pos]) &&
            (t2 = (u8[pos + 1] - 0x80)) <= 0x3f) {
        // U+0800..U+FFFF except surrogates
        c = (((c & 0xf) << 12) | ((t1 & 0x3f) << 6) | t2);
        pos += 2;
        return UTRIE2_GET32_FROM_U16_SINGLE_LEAD(trie, c);
    } else if(c < 0xe0 && c >= 0xc2 && pos != length && (t1 = (u8[pos] - 0x80)) <= 0x3f) {
        // U+0080..U+07FF
        uint32_t ce32 = trie->data32[trie->index[(UTRIE2_UTF8_2B_INDEX_2_OFFSET - 0xc0) + c] + t1];
        c = ((c & 0x1f) << 6) | t1;
        ++pos;
        return ce32;
    } else {
        // Function call for supplementary code points and error cases.
        // Illegal byte sequences yield U+FFFD.
        c = utf8_nextCharSafeBody(u8, &pos, length, c, -3);
        return data->getCE32(c);
    }
}

UBool
UTF8CollationIterator::foundNULTerminator() {
    if(length < 0) {
        length = --pos;
        return TRUE;
    } else {
        return FALSE;
    }
}

UBool
UTF8CollationIterator::forbidSurrogateCodePoints() const {
    return TRUE;
}

UChar32
UTF8CollationIterator::nextCodePoint(UErrorCode & /*errorCode*/) {
    if(pos == length) {
        return U_SENTINEL;
    }
    if(u8[pos] == 0 && length < 0) {
        length = pos;
        return U_SENTINEL;
    }
    UChar32 c;
    U8_NEXT_OR_FFFD(u8, pos, length, c);
    return c;
}

UChar32
UTF8CollationIterator::previousCodePoint(UErrorCode & /*errorCode*/) {
    if(pos == 0) {
        return U_SENTINEL;
    }
    UChar32 c;
    U8_PREV_OR_FFFD(u8, 0, pos, c);
    return c;
}

void
UTF8CollationIterator::forwardNumCodePoints(int32_t num, UErrorCode & /*errorCode*/) {
    U8_FWD_N(u8, pos, length, num);
}

void
UTF8CollationIterator::backwardNumCodePoints(int32_t num, UErrorCode & /*errorCode*/) {
    U8_BACK_N(u8, 0, pos, num);
}

// FCDUTF8CollationIterator ------------------------------------------------ ***

FCDUTF8CollationIterator::~FCDUTF8CollationIterator() {}

void
FCDUTF8CollationIterator::resetToOffset(int32_t newOffset) {
    reset();
    start = pos = newOffset;
    state = CHECK_FWD;
}

int32_t
FCDUTF8CollationIterator::getOffset() const {
    if(state != IN_NORMALIZED) {
        return pos;
    } else if(pos == 0) {
        return start;
    } else {
        return limit;
    }
}

uint32_t
FCDUTF8CollationIterator::handleNextCE32(UChar32 &c, UErrorCode &errorCode) {
    for(;;) {
        if(state == CHECK_FWD) {
            // Combination of UTF8CollationIterator::handleNextCE32() with FCD check fastpath.
            if(pos == length) {
                c = U_SENTINEL;
                return Collation::FALLBACK_CE32;
            }
            c = u8[pos++];
            if(U8_IS_SINGLE(c)) {
                // ASCII 00..7F
                return trie->data32[c];
            }
            uint8_t t1, t2;
            if(0xe0 <= c && c < 0xf0 &&
                    ((pos + 1) < length || length < 0) &&
                    U8_IS_VALID_LEAD3_AND_T1(c, t1 = u8[pos]) &&
                    (t2 = (u8[pos + 1] - 0x80)) <= 0x3f) {
                // U+0800..U+FFFF except surrogates
                c = (((c & 0xf) << 12) | ((t1 & 0x3f) << 6) | t2);
                pos += 2;
                if(CollationFCD::hasTccc(c) &&
                        (CollationFCD::maybeTibetanCompositeVowel(c) ||
                            (pos != length && nextHasLccc()))) {
                    pos -= 3;
                } else {
                    break;  // return CE32(BMP)
                }
            } else if(c < 0xe0 && c >= 0xc2 && pos != length && (t1 = (u8[pos] - 0x80)) <= 0x3f) {
                // U+0080..U+07FF
                uint32_t ce32 = trie->data32[trie->index[(UTRIE2_UTF8_2B_INDEX_2_OFFSET - 0xc0) + c] + t1];
                c = ((c & 0x1f) << 6) | t1;
                ++pos;
                if(CollationFCD::hasTccc(c) && pos != length && nextHasLccc()) {
                    pos -= 2;
                } else {
                    return ce32;
                }
            } else {
                // Function call for supplementary code points and error cases.
                // Illegal byte sequences yield U+FFFD.
                c = utf8_nextCharSafeBody(u8, &pos, length, c, -3);
                if(c == 0xfffd) {
                    return Collation::FFFD_CE32;
                } else {
                    U_ASSERT(c > 0xffff);
                    if(CollationFCD::hasTccc(U16_LEAD(c)) && pos != length && nextHasLccc()) {
                        pos -= 4;
                    } else {
                        return data->getCE32FromSupplementary(c);
                    }
                }
            }
            if(!nextSegment(errorCode)) {
                c = U_SENTINEL;
                return Collation::FALLBACK_CE32;
            }
            continue;
        } else if(state == IN_FCD_SEGMENT && pos != limit) {
            return UTF8CollationIterator::handleNextCE32(c, errorCode);
        } else if(state == IN_NORMALIZED && pos != normalized.length()) {
            c = normalized[pos++];
            break;
        } else {
            switchToForward();
        }
    }
    return UTRIE2_GET32_FROM_U16_SINGLE_LEAD(trie, c);
}

UBool
FCDUTF8CollationIterator::nextHasLccc() const {
    U_ASSERT(state == CHECK_FWD && pos != length);
    // The lowest code point with ccc!=0 is U+0300 which is CC 80 in UTF-8.
    // CJK U+4000..U+DFFF except U+Axxx are also FCD-inert. (Lead bytes E4..ED except EA.)
    UChar32 c = u8[pos];
    if(c < 0xcc || (0xe4 <= c && c <= 0xed && c != 0xea)) { return FALSE; }
    int32_t i = pos;
    U8_NEXT_OR_FFFD(u8, i, length, c);
    if(c > 0xffff) { c = U16_LEAD(c); }
    return CollationFCD::hasLccc(c);
}

UBool
FCDUTF8CollationIterator::previousHasTccc() const {
    U_ASSERT(state == CHECK_BWD && pos != 0);
    UChar32 c = u8[pos - 1];
    if(U8_IS_SINGLE(c)) { return FALSE; }
    int32_t i = pos;
    U8_PREV_OR_FFFD(u8, 0, i, c);
    if(c > 0xffff) { c = U16_LEAD(c); }
    return CollationFCD::hasTccc(c);
}

UChar
FCDUTF8CollationIterator::handleGetTrailSurrogate() {
    if(state != IN_NORMALIZED) { return 0; }
    U_ASSERT(pos < normalized.length());
    UChar trail;
    if(U16_IS_TRAIL(trail = normalized[pos])) { ++pos; }
    return trail;
}

UBool
FCDUTF8CollationIterator::foundNULTerminator() {
    if(state == CHECK_FWD && length < 0) {
        length = --pos;
        return TRUE;
    } else {
        return FALSE;
    }
}

UChar32
FCDUTF8CollationIterator::nextCodePoint(UErrorCode &errorCode) {
    UChar32 c;
    for(;;) {
        if(state == CHECK_FWD) {
            if(pos == length || ((c = u8[pos]) == 0 && length < 0)) {
                return U_SENTINEL;
            }
            if(U8_IS_SINGLE(c)) {
                ++pos;
                return c;
            }
            U8_NEXT_OR_FFFD(u8, pos, length, c);
            if(CollationFCD::hasTccc(c <= 0xffff ? c : U16_LEAD(c)) &&
                    (CollationFCD::maybeTibetanCompositeVowel(c) ||
                        (pos != length && nextHasLccc()))) {
                // c is not FCD-inert, therefore it is not U+FFFD and it has a valid byte sequence
                // and we can use U8_LENGTH() rather than a previous-position variable.
                pos -= U8_LENGTH(c);
                if(!nextSegment(errorCode)) {
                    return U_SENTINEL;
                }
                continue;
            }
            return c;
        } else if(state == IN_FCD_SEGMENT && pos != limit) {
            U8_NEXT_OR_FFFD(u8, pos, length, c);
            return c;
        } else if(state == IN_NORMALIZED && pos != normalized.length()) {
            c = normalized.char32At(pos);
            pos += U16_LENGTH(c);
            return c;
        } else {
            switchToForward();
        }
    }
}

UChar32
FCDUTF8CollationIterator::previousCodePoint(UErrorCode &errorCode) {
    UChar32 c;
    for(;;) {
        if(state == CHECK_BWD) {
            if(pos == 0) {
                return U_SENTINEL;
            }
            if(U8_IS_SINGLE(c = u8[pos - 1])) {
                --pos;
                return c;
            }
            U8_PREV_OR_FFFD(u8, 0, pos, c);
            if(CollationFCD::hasLccc(c <= 0xffff ? c : U16_LEAD(c)) &&
                    (CollationFCD::maybeTibetanCompositeVowel(c) ||
                        (pos != 0 && previousHasTccc()))) {
                // c is not FCD-inert, therefore it is not U+FFFD and it has a valid byte sequence
                // and we can use U8_LENGTH() rather than a previous-position variable.
                pos += U8_LENGTH(c);
                if(!previousSegment(errorCode)) {
                    return U_SENTINEL;
                }
                continue;
            }
            return c;
        } else if(state == IN_FCD_SEGMENT && pos != start) {
            U8_PREV_OR_FFFD(u8, 0, pos, c);
            return c;
        } else if(state >= IN_NORMALIZED && pos != 0) {
            c = normalized.char32At(pos - 1);
            pos -= U16_LENGTH(c);
            return c;
        } else {
            switchToBackward();
        }
    }
}

void
FCDUTF8CollationIterator::forwardNumCodePoints(int32_t num, UErrorCode &errorCode) {
    // Specify the class to avoid a virtual-function indirection.
    // In Java, we would declare this class final.
    while(num > 0 && FCDUTF8CollationIterator::nextCodePoint(errorCode) >= 0) {
        --num;
    }
}

void
FCDUTF8CollationIterator::backwardNumCodePoints(int32_t num, UErrorCode &errorCode) {
    // Specify the class to avoid a virtual-function indirection.
    // In Java, we would declare this class final.
    while(num > 0 && FCDUTF8CollationIterator::previousCodePoint(errorCode) >= 0) {
        --num;
    }
}

void
FCDUTF8CollationIterator::switchToForward() {
    U_ASSERT(state == CHECK_BWD ||
             (state == IN_FCD_SEGMENT && pos == limit) ||
             (state == IN_NORMALIZED && pos == normalized.length()));
    if(state == CHECK_BWD) {
        // Turn around from backward checking.
        start = pos;
        if(pos == limit) {
            state = CHECK_FWD;  // Check forward.
        } else {  // pos < limit
            state = IN_FCD_SEGMENT;  // Stay in FCD segment.
        }
    } else {
        // Reached the end of the FCD segment.
        if(state == IN_FCD_SEGMENT) {
            // The input text segment is FCD, extend it forward.
        } else {
            // The input text segment needed to be normalized.
            // Switch to checking forward from it.
            start = pos = limit;
        }
        state = CHECK_FWD;
    }
}

UBool
FCDUTF8CollationIterator::nextSegment(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    U_ASSERT(state == CHECK_FWD && pos != length);
    // The input text [start..pos[ passes the FCD check.
    int32_t segmentStart = pos;
    // Collect the characters being checked, in case they need to be normalized.
    UnicodeString s;
    uint8_t prevCC = 0;
    for(;;) {
        // Fetch the next character and its fcd16 value.
        int32_t cpStart = pos;
        UChar32 c;
        U8_NEXT_OR_FFFD(u8, pos, length, c);
        uint16_t fcd16 = nfcImpl.getFCD16(c);
        uint8_t leadCC = (uint8_t)(fcd16 >> 8);
        if(leadCC == 0 && cpStart != segmentStart) {
            // FCD boundary before this character.
            pos = cpStart;
            break;
        }
        s.append(c);
        if(leadCC != 0 && (prevCC > leadCC || CollationFCD::isFCD16OfTibetanCompositeVowel(fcd16))) {
            // Fails FCD check. Find the next FCD boundary and normalize.
            while(pos != length) {
                cpStart = pos;
                U8_NEXT_OR_FFFD(u8, pos, length, c);
                if(nfcImpl.getFCD16(c) <= 0xff) {
                    pos = cpStart;
                    break;
                }
                s.append(c);
            }
            if(!normalize(s, errorCode)) { return FALSE; }
            start = segmentStart;
            limit = pos;
            state = IN_NORMALIZED;
            pos = 0;
            return TRUE;
        }
        prevCC = (uint8_t)fcd16;
        if(pos == length || prevCC == 0) {
            // FCD boundary after the last character.
            break;
        }
    }
    limit = pos;
    pos = segmentStart;
    U_ASSERT(pos != limit);
    state = IN_FCD_SEGMENT;
    return TRUE;
}

void
FCDUTF8CollationIterator::switchToBackward() {
    U_ASSERT(state == CHECK_FWD ||
             (state == IN_FCD_SEGMENT && pos == start) ||
             (state >= IN_NORMALIZED && pos == 0));
    if(state == CHECK_FWD) {
        // Turn around from forward checking.
        limit = pos;
        if(pos == start) {
            state = CHECK_BWD;  // Check backward.
        } else {  // pos > start
            state = IN_FCD_SEGMENT;  // Stay in FCD segment.
        }
    } else {
        // Reached the start of the FCD segment.
        if(state == IN_FCD_SEGMENT) {
            // The input text segment is FCD, extend it backward.
        } else {
            // The input text segment needed to be normalized.
            // Switch to checking backward from it.
            limit = pos = start;
        }
        state = CHECK_BWD;
    }
}

UBool
FCDUTF8CollationIterator::previousSegment(UErrorCode &errorCode) {
    if(U_FAILURE(errorCode)) { return FALSE; }
    U_ASSERT(state == CHECK_BWD && pos != 0);
    // The input text [pos..limit[ passes the FCD check.
    int32_t segmentLimit = pos;
    // Collect the characters being checked, in case they need to be normalized.
    UnicodeString s;
    uint8_t nextCC = 0;
    for(;;) {
        // Fetch the previous character and its fcd16 value.
        int32_t cpLimit = pos;
        UChar32 c;
        U8_PREV_OR_FFFD(u8, 0, pos, c);
        uint16_t fcd16 = nfcImpl.getFCD16(c);
        uint8_t trailCC = (uint8_t)fcd16;
        if(trailCC == 0 && cpLimit != segmentLimit) {
            // FCD boundary after this character.
            pos = cpLimit;
            break;
        }
        s.append(c);
        if(trailCC != 0 && ((nextCC != 0 && trailCC > nextCC) ||
                            CollationFCD::isFCD16OfTibetanCompositeVowel(fcd16))) {
            // Fails FCD check. Find the previous FCD boundary and normalize.
            while(fcd16 > 0xff && pos != 0) {
                cpLimit = pos;
                U8_PREV_OR_FFFD(u8, 0, pos, c);
                fcd16 = nfcImpl.getFCD16(c);
                if(fcd16 == 0) {
                    pos = cpLimit;
                    break;
                }
                s.append(c);
            }
            s.reverse();
            if(!normalize(s, errorCode)) { return FALSE; }
            limit = segmentLimit;
            start = pos;
            state = IN_NORMALIZED;
            pos = normalized.length();
            return TRUE;
        }
        nextCC = (uint8_t)(fcd16 >> 8);
        if(pos == 0 || nextCC == 0) {
            // FCD boundary before the following character.
            break;
        }
    }
    start = pos;
    pos = segmentLimit;
    U_ASSERT(pos != start);
    state = IN_FCD_SEGMENT;
    return TRUE;
}

UBool
FCDUTF8CollationIterator::normalize(const UnicodeString &s, UErrorCode &errorCode) {
    // NFD without argument checking.
    U_ASSERT(U_SUCCESS(errorCode));
    nfcImpl.decompose(s, normalized, errorCode);
    return U_SUCCESS(errorCode);
}

U_NAMESPACE_END

#endif  // !UCONFIG_NO_COLLATION
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2004-2012, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2004 - 2008, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/




#if !UCONFIG_NO_FORMATTING

/** 
 * \file
 * \brief C API: Universal Time Scale
 *
 * There are quite a few different conventions for binary datetime, depending on different
 * platforms and protocols. Some of these have severe drawbacks. For example, people using
 * Unix time (seconds since Jan 1, 1970) think that they are safe until near the year 2038.
 * But cases can and do arise where arithmetic manipulations causes serious problems. Consider
 * the computation of the average of two datetimes, for example: if one calculates them with
 * <code>averageTime = (time1 + time2)/2</code>, there will be overflow even with dates
 * around the present. Moreover, even if these problems don't occur, there is the issue of
 * conversion back and forth between different systems.
 *
 * <p>
 * Binary datetimes differ in a number of ways: the datatype, the unit,
 * and the epoch (origin). We'll refer to these as time scales. For example:
 *
 * <table border="1" cellspacing="0" cellpadding="4">
 *  <caption>Table 1: Binary Time Scales</caption>
 *  <tr>
 *    <th align="left">Source</th>
 *    <th align="left">Datatype</th>
 *    <th align="left">Unit</th>
 *    <th align="left">Epoch</th>
 *  </tr>
 *
 *  <tr>
 *    <td>UDTS_JAVA_TIME</td>
 *    <td>int64_t</td>
 *    <td>milliseconds</td>
 *    <td>Jan 1, 1970</td>
 *  </tr>
 *  <tr>
 *
 *    <td>UDTS_UNIX_TIME</td>
 *    <td>int32_t or int64_t</td>
 *    <td>seconds</td>
 *    <td>Jan 1, 1970</td>
 *  </tr>
 *  <tr>
 *    <td>UDTS_ICU4C_TIME</td>
 *
 *    <td>double</td>
 *    <td>milliseconds</td>
 *    <td>Jan 1, 1970</td>
 *  </tr>
 *  <tr>
 *    <td>UDTS_WINDOWS_FILE_TIME</td>
 *    <td>int64_t</td>
 *
 *    <td>ticks (100 nanoseconds)</td>
 *    <td>Jan 1, 1601</td>
 *  </tr>
 *  <tr>
 *    <td>UDTS_DOTNET_DATE_TIME</td>
 *    <td>int64_t</td>
 *    <td>ticks (100 nanoseconds)</td>
 *
 *    <td>Jan 1, 0001</td>
 *  </tr>
 *  <tr>
 *    <td>UDTS_MAC_OLD_TIME</td>
 *    <td>int32_t or int64_t</td>
 *    <td>seconds</td>
 *    <td>Jan 1, 1904</td>
 *
 *  </tr>
 *  <tr>
 *    <td>UDTS_MAC_TIME</td>
 *    <td>double</td>
 *    <td>seconds</td>
 *    <td>Jan 1, 2001</td>
 *  </tr>
 *
 *  <tr>
 *    <td>UDTS_EXCEL_TIME</td>
 *    <td>?</td>
 *    <td>days</td>
 *    <td>Dec 31, 1899</td>
 *  </tr>
 *  <tr>
 *
 *    <td>UDTS_DB2_TIME</td>
 *    <td>?</td>
 *    <td>days</td>
 *    <td>Dec 31, 1899</td>
 *  </tr>
 *
 *  <tr>
 *    <td>UDTS_UNIX_MICROSECONDS_TIME</td>
 *    <td>int64_t</td>
 *    <td>microseconds</td>
 *    <td>Jan 1, 1970</td>
 *  </tr>
 * </table>
 *
 * <p>
 * All of the epochs start at 00:00 am (the earliest possible time on the day in question),
 * and are assumed to be UTC.
 *
 * <p>
 * The ranges for different datatypes are given in the following table (all values in years).
 * The range of years includes the entire range expressible with positive and negative
 * values of the datatype. The range of years for double is the range that would be allowed
 * without losing precision to the corresponding unit.
 *
 * <table border="1" cellspacing="0" cellpadding="4">
 *  <tr>
 *    <th align="left">Units</th>
 *    <th align="left">int64_t</th>
 *    <th align="left">double</th>
 *    <th align="left">int32_t</th>
 *  </tr>
 *
 *  <tr>
 *    <td>1 sec</td>
 *    <td align="right">5.84542x10<sup>11</sup></td>
 *    <td align="right">285,420,920.94</td>
 *    <td align="right">136.10</td>
 *  </tr>
 *  <tr>
 *
 *    <td>1 millisecond</td>
 *    <td align="right">584,542,046.09</td>
 *    <td align="right">285,420.92</td>
 *    <td align="right">0.14</td>
 *  </tr>
 *  <tr>
 *    <td>1 microsecond</td>
 *
 *    <td align="right">584,542.05</td>
 *    <td align="right">285.42</td>
 *    <td align="right">0.00</td>
 *  </tr>
 *  <tr>
 *    <td>100 nanoseconds (tick)</td>
 *    <td align="right">58,454.20</td>
 *    <td align="right">28.54</td>
 *    <td align="right">0.00</td>
 *  </tr>
 *  <tr>
 *    <td>1 nanosecond</td>
 *    <td align="right">584.5420461</td>
 *    <td align="right">0.2854</td>
 *    <td align="right">0.00</td>
 *  </tr>
 * </table>
 *
 * <p>
 * These functions implement a universal time scale which can be used as a 'pivot',
 * and provide conversion functions to and from all other major time scales.
 * This datetimes to be converted to the pivot time, safely manipulated,
 * and converted back to any other datetime time scale.
 *
 *<p>
 * So what to use for this pivot? Java time has plenty of range, but cannot represent
 * .NET <code>System.DateTime</code> values without severe loss of precision. ICU4C time addresses this by using a
 * <code>double</code> that is otherwise equivalent to the Java time. However, there are disadvantages
 * with <code>doubles</code>. They provide for much more graceful degradation in arithmetic operations.
 * But they only have 53 bits of accuracy, which means that they will lose precision when
 * converting back and forth to ticks. What would really be nice would be a
 * <code>long double</code> (80 bits -- 64 bit mantissa), but that is not supported on most systems.
 *
 *<p>
 * The Unix extended time uses a structure with two components: time in seconds and a
 * fractional field (microseconds). However, this is clumsy, slow, and
 * prone to error (you always have to keep track of overflow and underflow in the
 * fractional field). <code>BigDecimal</code> would allow for arbitrary precision and arbitrary range,
 * but we do not want to use this as the normal type, because it is slow and does not
 * have a fixed size.
 *
 *<p>
 * Because of these issues, we ended up concluding that the .NET framework's
 * <code>System.DateTime</code> would be the best pivot. However, we use the full range
 * allowed by the datatype, allowing for datetimes back to 29,000 BC and up to 29,000 AD.
 * This time scale is very fine grained, does not lose precision, and covers a range that
 * will meet almost all requirements. It will not handle the range that Java times do,
 * but frankly, being able to handle dates before 29,000 BC or after 29,000 AD is of very limited interest.
 *
 */

/**
 * <code>UDateTimeScale</code> values are used to specify the time scale used for
 * conversion into or out if the universal time scale.
 *
 * @stable ICU 3.2
 */
typedef enum UDateTimeScale {
    /**
     * Used in the JDK. Data is a Java <code>long</code> (<code>int64_t</code>). Value
     * is milliseconds since January 1, 1970.
     *
     * @stable ICU 3.2
     */
    UDTS_JAVA_TIME = 0,

    /**
     * Used on Unix systems. Data is <code>int32_t</code> or <code>int64_t</code>. Value
     * is seconds since January 1, 1970.
     *
     * @stable ICU 3.2
     */
    UDTS_UNIX_TIME,
    
    /**
     * Used in IUC4C. Data is a <code>double</code>. Value
     * is milliseconds since January 1, 1970.
     *
     * @stable ICU 3.2
     */
    UDTS_ICU4C_TIME,
    
    /**
     * Used in Windows for file times. Data is an <code>int64_t</code>. Value
     * is ticks (1 tick == 100 nanoseconds) since January 1, 1601.
     *
     * @stable ICU 3.2
     */
    UDTS_WINDOWS_FILE_TIME,
    
    /**
     * Used in the .NET framework's <code>System.DateTime</code> structure. Data is an <code>int64_t</code>. Value
     * is ticks (1 tick == 100 nanoseconds) since January 1, 0001.
     *
     * @stable ICU 3.2
     */
    UDTS_DOTNET_DATE_TIME,
    
    /**
     * Used in older Macintosh systems. Data is <code>int32_t</code> or <code>int64_t</code>. Value
     * is seconds since January 1, 1904.
     *
     * @stable ICU 3.2
     */
    UDTS_MAC_OLD_TIME,
    
    /**
     * Used in newer Macintosh systems. Data is a <code>double</code>. Value
     * is seconds since January 1, 2001.
     *
     * @stable ICU 3.2
     */
    UDTS_MAC_TIME,
    
    /**
     * Used in Excel. Data is an <code>?unknown?</code>. Value
     * is days since December 31, 1899.
     *
     * @stable ICU 3.2
     */
    UDTS_EXCEL_TIME,
    
    /**
     * Used in DB2. Data is an <code>?unknown?</code>. Value
     * is days since December 31, 1899.
     *
     * @stable ICU 3.2
     */
    UDTS_DB2_TIME,

    /**
     * Data is a <code>long</code>. Value is microseconds since January 1, 1970.
     * Similar to Unix time (linear value from 1970) and struct timeval
     * (microseconds resolution).
     *
     * @stable ICU 3.8
     */
    UDTS_UNIX_MICROSECONDS_TIME,

#ifndef U_HIDE_DEPRECATED_API
    /**
     * The first unused time scale value. The limit of this enum
     * @deprecated ICU 59 The numeric value may change over time, see ICU ticket #12420.
     */
    UDTS_MAX_SCALE
#endif  /* U_HIDE_DEPRECATED_API */

} UDateTimeScale;

/**
 * <code>UTimeScaleValue</code> values are used to specify the time scale values
 * to <code>utmscale_getTimeScaleValue</code>.
 *
 * @see utmscale_getTimeScaleValue
 *
 * @stable ICU 3.2
 */
typedef enum UTimeScaleValue {
    /**
     * The constant used to select the units vale
     * for a time scale.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @stable ICU 3.2
     */
    UTSV_UNITS_VALUE = 0,

    /**
     * The constant used to select the epoch offset value
     * for a time scale.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @stable ICU 3.2
     */
    UTSV_EPOCH_OFFSET_VALUE=1,

    /**
     * The constant used to select the minimum from value
     * for a time scale.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @stable ICU 3.2
     */
    UTSV_FROM_MIN_VALUE=2,

    /**
     * The constant used to select the maximum from value
     * for a time scale.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @stable ICU 3.2
     */
    UTSV_FROM_MAX_VALUE=3,

    /**
     * The constant used to select the minimum to value
     * for a time scale.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @stable ICU 3.2
     */
    UTSV_TO_MIN_VALUE=4,

    /**
     * The constant used to select the maximum to value
     * for a time scale.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @stable ICU 3.2
     */
    UTSV_TO_MAX_VALUE=5,

#ifndef U_HIDE_INTERNAL_API
    /**
     * The constant used to select the epoch plus one value
     * for a time scale.
     * 
     * NOTE: This is an internal value. DO NOT USE IT. May not
     * actually be equal to the epoch offset value plus one.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @internal ICU 3.2
     */
    UTSV_EPOCH_OFFSET_PLUS_1_VALUE=6,

    /**
     * The constant used to select the epoch plus one value
     * for a time scale.
     * 
     * NOTE: This is an internal value. DO NOT USE IT. May not
     * actually be equal to the epoch offset value plus one.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @internal ICU 3.2
     */
    UTSV_EPOCH_OFFSET_MINUS_1_VALUE=7,

    /**
     * The constant used to select the units round value
     * for a time scale.
     * 
     * NOTE: This is an internal value. DO NOT USE IT.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @internal ICU 3.2
     */
    UTSV_UNITS_ROUND_VALUE=8,

    /**
     * The constant used to select the minimum safe rounding value
     * for a time scale.
     * 
     * NOTE: This is an internal value. DO NOT USE IT.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @internal ICU 3.2
     */
    UTSV_MIN_ROUND_VALUE=9,

    /**
     * The constant used to select the maximum safe rounding value
     * for a time scale.
     * 
     * NOTE: This is an internal value. DO NOT USE IT.
     * 
     * @see utmscale_getTimeScaleValue
     *
     * @internal ICU 3.2
     */
    UTSV_MAX_ROUND_VALUE=10,

#endif /* U_HIDE_INTERNAL_API */

#ifndef U_HIDE_DEPRECATED_API
    /**
     * The number of time scale values, in other words limit of this enum.
     * 
     * @see utmscale_getTimeScaleValue
     * @deprecated ICU 59 The numeric value may change over time, see ICU ticket #12420.
     */
    UTSV_MAX_SCALE_VALUE=11
#endif  /* U_HIDE_DEPRECATED_API */

} UTimeScaleValue;

/**
 * Get a value associated with a particular time scale.
 * 
 * @param timeScale The time scale
 * @param value A constant representing the value to get
 * @param status The status code. Set to <code>U_ILLEGAL_ARGUMENT_ERROR</code> if arguments are invalid.
 * @return - the value.
 * 
 * @stable ICU 3.2
 */
U_STABLE int64_t U_EXPORT2
    utmscale_getTimeScaleValue(UDateTimeScale timeScale, UTimeScaleValue value, UErrorCode *status);

/* Conversion to 'universal time scale' */

/**
 * Convert a <code>int64_t</code> datetime from the given time scale to the universal time scale.
 *
 * @param otherTime The <code>int64_t</code> datetime
 * @param timeScale The time scale to convert from
 * @param status The status code. Set to <code>U_ILLEGAL_ARGUMENT_ERROR</code> if the conversion is out of range.
 * 
 * @return The datetime converted to the universal time scale
 *
 * @stable ICU 3.2
 */
U_STABLE int64_t U_EXPORT2
    utmscale_fromInt64(int64_t otherTime, UDateTimeScale timeScale, UErrorCode *status);

/* Conversion from 'universal time scale' */

/**
 * Convert a datetime from the universal time scale to a <code>int64_t</code> in the given time scale.
 *
 * @param universalTime The datetime in the universal time scale
 * @param timeScale The time scale to convert to
 * @param status The status code. Set to <code>U_ILLEGAL_ARGUMENT_ERROR</code> if the conversion is out of range.
 * 
 * @return The datetime converted to the given time scale
 *
 * @stable ICU 3.2
 */
U_STABLE int64_t U_EXPORT2
    utmscale_toInt64(int64_t universalTime, UDateTimeScale timeScale, UErrorCode *status);

#endif /* #if !UCONFIG_NO_FORMATTING */



#define utmscale_ticks        INT64_C(1)
#define utmscale_microseconds (utmscale_ticks * 10)
#define utmscale_milliseconds (utmscale_microseconds * 1000)
#define utmscale_seconds      (utmscale_milliseconds * 1000)
#define utmscale_minutes      (utmscale_seconds * 60)
#define utmscale_hours        (utmscale_minutes * 60)
#define utmscale_days         (utmscale_hours * 24)

/* Constants generated by ICU4J com.ibm.icu.dev.tool.timescale.GenerateCTimeScaleData. */
static const int64_t timeScaleTable[UDTS_MAX_SCALE][UTSV_MAX_SCALE_VALUE] = {
  /*    units             epochOffset                     fromMin                        fromMax                            toMin                        toMax                    epochOffsetP1                epochOffsetM1              unitsRound                    minRound                     maxRound          */
    {utmscale_milliseconds, INT64_C(62135596800000),     INT64_C(-984472800485477),     INT64_C(860201606885477),     INT64_C(-9223372036854774999), INT64_C(9223372036854774999), INT64_C(62135596800001),     INT64_C(62135596799999),     INT64_C(5000),         INT64_C(-9223372036854770808), INT64_C(9223372036854770807)},
    {utmscale_seconds,      INT64_C(62135596800),        INT64_C(-984472800485),        INT64_C(860201606885),        U_INT64_MIN,                   U_INT64_MAX,                  INT64_C(62135596801),        INT64_C(62135596799),        INT64_C(5000000),      INT64_C(-9223372036849775808), INT64_C(9223372036849775807)},
    {utmscale_milliseconds, INT64_C(62135596800000),     INT64_C(-984472800485477),     INT64_C(860201606885477),     INT64_C(-9223372036854774999), INT64_C(9223372036854774999), INT64_C(62135596800001),     INT64_C(62135596799999),     INT64_C(5000),         INT64_C(-9223372036854770808), INT64_C(9223372036854770807)},
    {utmscale_ticks,        INT64_C(504911232000000000), U_INT64_MIN,                   INT64_C(8718460804854775807), INT64_C(-8718460804854775808), U_INT64_MAX,                  INT64_C(504911232000000000), INT64_C(504911232000000000), INT64_C(0),            U_INT64_MIN,                   U_INT64_MAX},
    {utmscale_ticks,        INT64_C(0),                  U_INT64_MIN,                   U_INT64_MAX,                  U_INT64_MIN,                   U_INT64_MAX,                  INT64_C(0),                  INT64_C(0),                  INT64_C(0),            U_INT64_MIN,                   U_INT64_MAX},
    {utmscale_seconds,      INT64_C(60052752000),        INT64_C(-982389955685),        INT64_C(862284451685),        U_INT64_MIN,                   U_INT64_MAX,                  INT64_C(60052752001),        INT64_C(60052751999),        INT64_C(5000000),      INT64_C(-9223372036849775808), INT64_C(9223372036849775807)},
    {utmscale_seconds,      INT64_C(63113904000),        INT64_C(-985451107685),        INT64_C(859223299685),        U_INT64_MIN,                   U_INT64_MAX,                  INT64_C(63113904001),        INT64_C(63113903999),        INT64_C(5000000),      INT64_C(-9223372036849775808), INT64_C(9223372036849775807)},
    {utmscale_days,         INT64_C(693594),             INT64_C(-11368793),            INT64_C(9981605),             U_INT64_MIN,                   U_INT64_MAX,                  INT64_C(693595),             INT64_C(693593),             INT64_C(432000000000), INT64_C(-9223371604854775808), INT64_C(9223371604854775807)},
    {utmscale_days,         INT64_C(693594),             INT64_C(-11368793),            INT64_C(9981605),             U_INT64_MIN,                   U_INT64_MAX,                  INT64_C(693595),             INT64_C(693593),             INT64_C(432000000000), INT64_C(-9223371604854775808), INT64_C(9223371604854775807)},
    {utmscale_microseconds, INT64_C(62135596800000000),  INT64_C(-984472800485477580),  INT64_C(860201606885477580),  INT64_C(-9223372036854775804), INT64_C(9223372036854775804), INT64_C(62135596800000001),  INT64_C(62135596799999999),  INT64_C(5),            INT64_C(-9223372036854775803), INT64_C(9223372036854775802)},
};

U_CAPI int64_t U_EXPORT2
utmscale_getTimeScaleValue(UDateTimeScale timeScale, UTimeScaleValue value, UErrorCode *status)
{
    if (status == NULL || U_FAILURE(*status)) {
        return 0;
    }

    if (timeScale < UDTS_JAVA_TIME || UDTS_MAX_SCALE <= timeScale
        || value < UTSV_UNITS_VALUE || UTSV_MAX_SCALE_VALUE <= value)
    {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    return timeScaleTable[timeScale][value];
}

U_CAPI int64_t U_EXPORT2
utmscale_fromInt64(int64_t otherTime, UDateTimeScale timeScale, UErrorCode *status)
{
    const int64_t *data;

    if (status == NULL || U_FAILURE(*status)) {
        return 0;
    }

    if ((int32_t)timeScale < 0 || timeScale >= UDTS_MAX_SCALE) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    data = (const int64_t *)(&timeScaleTable[timeScale]);

    if (otherTime < data[UTSV_FROM_MIN_VALUE] || otherTime > data[UTSV_FROM_MAX_VALUE]) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    return (otherTime + data[UTSV_EPOCH_OFFSET_VALUE]) * data[UTSV_UNITS_VALUE];
}

U_CAPI int64_t U_EXPORT2
utmscale_toInt64(int64_t universalTime, UDateTimeScale timeScale, UErrorCode *status)
{
    const int64_t *data;

    if (status == NULL || U_FAILURE(*status)) {
        return 0;
    }

    if ((int32_t)timeScale < 0 || timeScale >= UDTS_MAX_SCALE) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    data = (const int64_t *)(&timeScaleTable[timeScale]);

    if (universalTime < data[UTSV_TO_MIN_VALUE] || universalTime > data[UTSV_TO_MAX_VALUE]) {
        *status = U_ILLEGAL_ARGUMENT_ERROR;
        return 0;
    }

    if (universalTime < 0) {
        if (universalTime < data[UTSV_MIN_ROUND_VALUE]) {
            return (universalTime + data[UTSV_UNITS_ROUND_VALUE]) / data[UTSV_UNITS_VALUE] - data[UTSV_EPOCH_OFFSET_PLUS_1_VALUE];
        }

        return (universalTime - data[UTSV_UNITS_ROUND_VALUE]) / data[UTSV_UNITS_VALUE] - data[UTSV_EPOCH_OFFSET_VALUE];
    }

    if (universalTime > data[UTSV_MAX_ROUND_VALUE]) {
        return (universalTime - data[UTSV_UNITS_ROUND_VALUE]) / data[UTSV_UNITS_VALUE] - data[UTSV_EPOCH_OFFSET_MINUS_1_VALUE];
    }

    return (universalTime + data[UTSV_UNITS_ROUND_VALUE]) / data[UTSV_UNITS_VALUE] - data[UTSV_EPOCH_OFFSET_VALUE];
}

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

  // for 'typeid' to work



#if !UCONFIG_NO_FORMATTING










U_NAMESPACE_BEGIN

// This is the deleter that will be use to remove TimeZoneRule
U_CDECL_BEGIN
static void U_CALLCONV
deleteTimeZoneRule(void* obj) {
    delete (TimeZoneRule*) obj;
}
U_CDECL_END

// Smybol characters used by RFC2445 VTIMEZONE
static const UChar vtzone_COLON = 0x3A; /* : */
static const UChar vtzone_SEMICOLON = 0x3B; /* ; */
static const UChar vtzone_EQUALS_SIGN = 0x3D; /* = */
static const UChar vtzone_COMMA = 0x2C; /* , */
static const UChar vtzone_PLUS = 0x2B; /* + */
static const UChar vtzone_MINUS = 0x2D; /* - */

// RFC2445 VTIMEZONE tokens
static const UChar ICAL_BEGIN_VTIMEZONE[] = {0x42, 0x45, 0x47, 0x49, 0x4E, 0x3A, 0x56, 0x54, 0x49, 0x4D, 0x45, 0x5A, 0x4F, 0x4E, 0x45, 0}; /* "BEGIN:VTIMEZONE" */
static const UChar ICAL_END_VTIMEZONE[] = {0x45, 0x4E, 0x44, 0x3A, 0x56, 0x54, 0x49, 0x4D, 0x45, 0x5A, 0x4F, 0x4E, 0x45, 0}; /* "END:VTIMEZONE" */
static const UChar ICAL_BEGIN[] = {0x42, 0x45, 0x47, 0x49, 0x4E, 0}; /* "BEGIN" */
static const UChar ICAL_END[] = {0x45, 0x4E, 0x44, 0}; /* "END" */
static const UChar ICAL_VTIMEZONE[] = {0x56, 0x54, 0x49, 0x4D, 0x45, 0x5A, 0x4F, 0x4E, 0x45, 0}; /* "VTIMEZONE" */
static const UChar ICAL_TZID[] = {0x54, 0x5A, 0x49, 0x44, 0}; /* "TZID" */
static const UChar ICAL_STANDARD[] = {0x53, 0x54, 0x41, 0x4E, 0x44, 0x41, 0x52, 0x44, 0}; /* "STANDARD" */
static const UChar ICAL_DAYLIGHT[] = {0x44, 0x41, 0x59, 0x4C, 0x49, 0x47, 0x48, 0x54, 0}; /* "DAYLIGHT" */
static const UChar ICAL_DTSTART[] = {0x44, 0x54, 0x53, 0x54, 0x41, 0x52, 0x54, 0}; /* "DTSTART" */
static const UChar ICAL_TZOFFSETFROM[] = {0x54, 0x5A, 0x4F, 0x46, 0x46, 0x53, 0x45, 0x54, 0x46, 0x52, 0x4F, 0x4D, 0}; /* "TZOFFSETFROM" */
static const UChar ICAL_TZOFFSETTO[] = {0x54, 0x5A, 0x4F, 0x46, 0x46, 0x53, 0x45, 0x54, 0x54, 0x4F, 0}; /* "TZOFFSETTO" */
static const UChar ICAL_RDATE[] = {0x52, 0x44, 0x41, 0x54, 0x45, 0}; /* "RDATE" */
static const UChar ICAL_RRULE[] = {0x52, 0x52, 0x55, 0x4C, 0x45, 0}; /* "RRULE" */
static const UChar ICAL_TZNAME[] = {0x54, 0x5A, 0x4E, 0x41, 0x4D, 0x45, 0}; /* "TZNAME" */
static const UChar ICAL_TZURL[] = {0x54, 0x5A, 0x55, 0x52, 0x4C, 0}; /* "TZURL" */
static const UChar ICAL_LASTMOD[] = {0x4C, 0x41, 0x53, 0x54, 0x2D, 0x4D, 0x4F, 0x44, 0x49, 0x46, 0x49, 0x45, 0x44, 0}; /* "LAST-MODIFIED" */

static const UChar ICAL_FREQ[] = {0x46, 0x52, 0x45, 0x51, 0}; /* "FREQ" */
static const UChar ICAL_UNTIL[] = {0x55, 0x4E, 0x54, 0x49, 0x4C, 0}; /* "UNTIL" */
static const UChar ICAL_YEARLY[] = {0x59, 0x45, 0x41, 0x52, 0x4C, 0x59, 0}; /* "YEARLY" */
static const UChar ICAL_BYMONTH[] = {0x42, 0x59, 0x4D, 0x4F, 0x4E, 0x54, 0x48, 0}; /* "BYMONTH" */
static const UChar ICAL_BYDAY[] = {0x42, 0x59, 0x44, 0x41, 0x59, 0}; /* "BYDAY" */
static const UChar ICAL_BYMONTHDAY[] = {0x42, 0x59, 0x4D, 0x4F, 0x4E, 0x54, 0x48, 0x44, 0x41, 0x59, 0}; /* "BYMONTHDAY" */

static const UChar ICAL_NEWLINE[] = {0x0D, 0x0A, 0}; /* CRLF */

static const UChar ICAL_DOW_NAMES[7][3] = {
    {0x53, 0x55, 0}, /* "SU" */
    {0x4D, 0x4F, 0}, /* "MO" */
    {0x54, 0x55, 0}, /* "TU" */
    {0x57, 0x45, 0}, /* "WE" */
    {0x54, 0x48, 0}, /* "TH" */
    {0x46, 0x52, 0}, /* "FR" */
    {0x53, 0x41, 0}  /* "SA" */};

// Month length for non-leap year
static const int32_t MONTHLENGTH[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

// ICU custom property
static const UChar ICU_TZINFO_PROP[] = {0x58, 0x2D, 0x54, 0x5A, 0x49, 0x4E, 0x46, 0x4F, 0x3A, 0}; /* "X-TZINFO:" */
static const UChar ICU_TZINFO_PARTIAL[] = {0x2F, 0x50, 0x61, 0x72, 0x74, 0x69, 0x61, 0x6C, 0x40, 0}; /* "/Partial@" */
static const UChar ICU_TZINFO_SIMPLE[] = {0x2F, 0x53, 0x69, 0x6D, 0x70, 0x6C, 0x65, 0x40, 0}; /* "/Simple@" */


/*
 * Simple fixed digit ASCII number to integer converter
 */
static int32_t parseAsciiDigits(const UnicodeString& str, int32_t start, int32_t length, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return 0;
    }
    if (length <= 0 || str.length() < start || (start + length) > str.length()) {
        status = U_INVALID_FORMAT_ERROR;
        return 0;
    }
    int32_t sign = 1;
    if (str.charAt(start) == vtzone_PLUS) {
        start++;
        length--;
    } else if (str.charAt(start) == vtzone_MINUS) {
        sign = -1;
        start++;
        length--;
    }
    int32_t num = 0;
    for (int32_t i = 0; i < length; i++) {
        int32_t digit = str.charAt(start + i) - 0x0030;
        if (digit < 0 || digit > 9) {
            status = U_INVALID_FORMAT_ERROR;
            return 0;
        }
        num = 10 * num + digit;
    }
    return sign * num;
}

static UnicodeString& appendAsciiDigits(int32_t number, uint8_t length, UnicodeString& str) {
    UBool negative = FALSE;
    int32_t digits[10]; // max int32_t is 10 decimal digits
    int32_t i;

    if (number < 0) {
        negative = TRUE;
        number *= -1;
    }

    length = length > 10 ? 10 : length;
    if (length == 0) {
        // variable length
        i = 0;
        do {
            digits[i++] = number % 10;
            number /= 10;
        } while (number != 0);
        length = static_cast<uint8_t>(i);
    } else {
        // fixed digits
        for (i = 0; i < length; i++) {
           digits[i] = number % 10;
           number /= 10;
        }
    }
    if (negative) {
        str.append(vtzone_MINUS);
    }
    for (i = length - 1; i >= 0; i--) {
        str.append((UChar)(digits[i] + 0x0030));
    }
    return str;
}

static UnicodeString& appendMillis(UDate date, UnicodeString& str) {
    UBool negative = FALSE;
    int32_t digits[20]; // max int64_t is 20 decimal digits
    int32_t i;
    int64_t number;

    if (date < MIN_MILLIS) {
        number = (int64_t)MIN_MILLIS;
    } else if (date > MAX_MILLIS) {
        number = (int64_t)MAX_MILLIS;
    } else {
        number = (int64_t)date;
    }
    if (number < 0) {
        negative = TRUE;
        number *= -1;
    }
    i = 0;
    do {
        digits[i++] = (int32_t)(number % 10);
        number /= 10;
    } while (number != 0);

    if (negative) {
        str.append(vtzone_MINUS);
    }
    i--;
    while (i >= 0) {
        str.append((UChar)(digits[i--] + 0x0030));
    }
    return str;
}

/*
 * Convert date/time to RFC2445 Date-Time form #1 DATE WITH LOCAL TIME
 */
static UnicodeString& getDateTimeString(UDate time, UnicodeString& str) {
    int32_t year, month, dom, dow, doy, mid;
    Grego::timeToFields(time, year, month, dom, dow, doy, mid);

    str.remove();
    appendAsciiDigits(year, 4, str);
    appendAsciiDigits(month + 1, 2, str);
    appendAsciiDigits(dom, 2, str);
    str.append((UChar)0x0054 /*'T'*/);

    int32_t t = mid;
    int32_t hour = t / U_MILLIS_PER_HOUR;
    t %= U_MILLIS_PER_HOUR;
    int32_t min = t / U_MILLIS_PER_MINUTE;
    t %= U_MILLIS_PER_MINUTE;
    int32_t sec = t / U_MILLIS_PER_SECOND;

    appendAsciiDigits(hour, 2, str);
    appendAsciiDigits(min, 2, str);
    appendAsciiDigits(sec, 2, str);
    return str;
}

/*
 * Convert date/time to RFC2445 Date-Time form #2 DATE WITH UTC TIME
 */
static UnicodeString& getUTCDateTimeString(UDate time, UnicodeString& str) {
    getDateTimeString(time, str);
    str.append((UChar)0x005A /*'Z'*/);
    return str;
}

/*
 * Parse RFC2445 Date-Time form #1 DATE WITH LOCAL TIME and
 * #2 DATE WITH UTC TIME
 */
static UDate parseDateTimeString(const UnicodeString& str, int32_t offset, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return 0.0;
    }

    int32_t year = 0, month = 0, day = 0, hour = 0, min = 0, sec = 0;
    UBool isUTC = FALSE;
    UBool isValid = FALSE;
    do {
        int length = str.length();
        if (length != 15 && length != 16) {
            // FORM#1 15 characters, such as "20060317T142115"
            // FORM#2 16 characters, such as "20060317T142115Z"
            break;
        }
        if (str.charAt(8) != 0x0054) {
            // charcter "T" must be used for separating date and time
            break;
        }
        if (length == 16) {
            if (str.charAt(15) != 0x005A) {
                // invalid format
                break;
            }
            isUTC = TRUE;
        }

        year = parseAsciiDigits(str, 0, 4, status);
        month = parseAsciiDigits(str, 4, 2, status) - 1;  // 0-based
        day = parseAsciiDigits(str, 6, 2, status);
        hour = parseAsciiDigits(str, 9, 2, status);
        min = parseAsciiDigits(str, 11, 2, status);
        sec = parseAsciiDigits(str, 13, 2, status);

        if (U_FAILURE(status)) {
            break;
        }

        // check valid range
        int32_t maxDayOfMonth = Grego::monthLength(year, month);
        if (year < 0 || month < 0 || month > 11 || day < 1 || day > maxDayOfMonth ||
                hour < 0 || hour >= 24 || min < 0 || min >= 60 || sec < 0 || sec >= 60) {
            break;
        }

        isValid = TRUE;
    } while(false);

    if (!isValid) {
        status = U_INVALID_FORMAT_ERROR;
        return 0.0;
    }
    // Calculate the time
    UDate time = Grego::fieldsToDay(year, month, day) * U_MILLIS_PER_DAY;
    time += (hour * U_MILLIS_PER_HOUR + min * U_MILLIS_PER_MINUTE + sec * U_MILLIS_PER_SECOND);
    if (!isUTC) {
        time -= offset;
    }
    return time;
}

/*
 * Convert RFC2445 utc-offset string to milliseconds
 */
static int32_t offsetStrToMillis(const UnicodeString& str, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return 0;
    }

    UBool isValid = FALSE;
    int32_t sign = 0, hour = 0, min = 0, sec = 0;

    do {
        int length = str.length();
        if (length != 5 && length != 7) {
            // utf-offset must be 5 or 7 characters
            break;
        }
        // sign
        UChar s = str.charAt(0);
        if (s == vtzone_PLUS) {
            sign = 1;
        } else if (s == vtzone_MINUS) {
            sign = -1;
        } else {
            // utf-offset must start with "+" or "-"
            break;
        }
        hour = parseAsciiDigits(str, 1, 2, status);
        min = parseAsciiDigits(str, 3, 2, status);
        if (length == 7) {
            sec = parseAsciiDigits(str, 5, 2, status);
        }
        if (U_FAILURE(status)) {
            break;
        }
        isValid = true;
    } while(false);

    if (!isValid) {
        status = U_INVALID_FORMAT_ERROR;
        return 0;
    }
    int32_t millis = sign * ((hour * 60 + min) * 60 + sec) * 1000;
    return millis;
}

/*
 * Convert milliseconds to RFC2445 utc-offset string
 */
static void millisToOffset(int32_t millis, UnicodeString& str) {
    str.remove();
    if (millis >= 0) {
        str.append(vtzone_PLUS);
    } else {
        str.append(vtzone_MINUS);
        millis = -millis;
    }
    int32_t hour, min, sec;
    int32_t t = millis / 1000;

    sec = t % 60;
    t = (t - sec) / 60;
    min = t % 60;
    hour = t / 60;

    appendAsciiDigits(hour, 2, str);
    appendAsciiDigits(min, 2, str);
    appendAsciiDigits(sec, 2, str);
}

/*
 * Create a default TZNAME from TZID
 */
static void getDefaultTZName(const UnicodeString &tzid, UBool isDST, UnicodeString& zonename) {
    zonename = tzid;
    if (isDST) {
        zonename += UNICODE_STRING_SIMPLE("(DST)");
    } else {
        zonename += UNICODE_STRING_SIMPLE("(STD)");
    }
}

/*
 * Parse individual RRULE
 *
 * On return -
 *
 * month    calculated by BYMONTH-1, or -1 when not found
 * dow      day of week in BYDAY, or 0 when not found
 * wim      day of week ordinal number in BYDAY, or 0 when not found
 * dom      an array of day of month
 * domCount number of availble days in dom (domCount is specifying the size of dom on input)
 * until    time defined by UNTIL attribute or MIN_MILLIS if not available
 */
static void parseRRULE(const UnicodeString& rrule, int32_t& month, int32_t& dow, int32_t& wim,
                       int32_t* dom, int32_t& domCount, UDate& until, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    int32_t numDom = 0;

    month = -1;
    dow = 0;
    wim = 0;
    until = MIN_MILLIS;

    UBool yearly = FALSE;
    //UBool parseError = FALSE;

    int32_t prop_start = 0;
    int32_t prop_end;
    UnicodeString prop, attr, value;
    UBool nextProp = TRUE;

    while (nextProp) {
        prop_end = rrule.indexOf(vtzone_SEMICOLON, prop_start);
        if (prop_end == -1) {
            prop.setTo(rrule, prop_start);
            nextProp = FALSE;
        } else {
            prop.setTo(rrule, prop_start, prop_end - prop_start);
            prop_start = prop_end + 1;
        }
        int32_t eql = prop.indexOf(vtzone_EQUALS_SIGN);
        if (eql != -1) {
            attr.setTo(prop, 0, eql);
            value.setTo(prop, eql + 1);
        } else {
            goto rruleParseError;
        }

        if (attr.compare(ICAL_FREQ, -1) == 0) {
            // only support YEARLY frequency type
            if (value.compare(ICAL_YEARLY, -1) == 0) {
                yearly = TRUE;
            } else {
                goto rruleParseError;
            }
        } else if (attr.compare(ICAL_UNTIL, -1) == 0) {
            // ISO8601 UTC format, for example, "20060315T020000Z"
            until = parseDateTimeString(value, 0, status);
            if (U_FAILURE(status)) {
                goto rruleParseError;
            }
        } else if (attr.compare(ICAL_BYMONTH, -1) == 0) {
            // Note: BYMONTH may contain multiple months, but only single month make sense for
            // VTIMEZONE property.
            if (value.length() > 2) {
                goto rruleParseError;
            }
            month = parseAsciiDigits(value, 0, value.length(), status) - 1;
            if (U_FAILURE(status) || month < 0 || month >= 12) {
                goto rruleParseError;
            }
        } else if (attr.compare(ICAL_BYDAY, -1) == 0) {
            // Note: BYDAY may contain multiple day of week separated by comma.  It is unlikely used for
            // VTIMEZONE property.  We do not support the case.

            // 2-letter format is used just for representing a day of week, for example, "SU" for Sunday
            // 3 or 4-letter format is used for represeinging Nth day of week, for example, "-1SA" for last Saturday
            int32_t length = value.length();
            if (length < 2 || length > 4) {
                goto rruleParseError;
            }
            if (length > 2) {
                // Nth day of week
                int32_t sign = 1;
                if (value.charAt(0) == vtzone_PLUS) {
                    sign = 1;
                } else if (value.charAt(0) == vtzone_MINUS) {
                    sign = -1;
                } else if (length == 4) {
                    goto rruleParseError;
                }
                int32_t n = parseAsciiDigits(value, length - 3, 1, status);
                if (U_FAILURE(status) || n == 0 || n > 4) {
                    goto rruleParseError;
                }
                wim = n * sign;
                value.remove(0, length - 2);
            }
            int32_t wday;
            for (wday = 0; wday < 7; wday++) {
                if (value.compare(ICAL_DOW_NAMES[wday], 2) == 0) {
                    break;
                }
            }
            if (wday < 7) {
                // Sunday(1) - Saturday(7)
                dow = wday + 1;
            } else {
                goto rruleParseError;
            }
        } else if (attr.compare(ICAL_BYMONTHDAY, -1) == 0) {
            // Note: BYMONTHDAY may contain multiple days delimitted by comma
            //
            // A value of BYMONTHDAY could be negative, for example, -1 means
            // the last day in a month
            int32_t dom_idx = 0;
            int32_t dom_start = 0;
            int32_t dom_end;
            UBool nextDOM = TRUE;
            while (nextDOM) {
                dom_end = value.indexOf(vtzone_COMMA, dom_start);
                if (dom_end == -1) {
                    dom_end = value.length();
                    nextDOM = FALSE;
                }
                if (dom_idx < domCount) {
                    dom[dom_idx] = parseAsciiDigits(value, dom_start, dom_end - dom_start, status);
                    if (U_FAILURE(status)) {
                        goto rruleParseError;
                    }
                    dom_idx++;
                } else {
                    status = U_BUFFER_OVERFLOW_ERROR;
                    goto rruleParseError;
                }
                dom_start = dom_end + 1;
            }
            numDom = dom_idx;
        }
    }
    if (!yearly) {
        // FREQ=YEARLY must be set
        goto rruleParseError;
    }
    // Set actual number of parsed DOM (ICAL_BYMONTHDAY)
    domCount = numDom;
    return;

rruleParseError:
    if (U_SUCCESS(status)) {
        // Set error status
        status = U_INVALID_FORMAT_ERROR;
    }
}

static TimeZoneRule* createRuleByRRULE(const UnicodeString& zonename, int rawOffset, int dstSavings, UDate start,
                                       UVector* dates, int fromOffset, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    if (dates == NULL || dates->size() == 0) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    int32_t i, j;
    DateTimeRule *adtr = NULL;

    // Parse the first rule
    UnicodeString rrule = *((UnicodeString*)dates->elementAt(0));
    int32_t month, dayOfWeek, nthDayOfWeek, dayOfMonth = 0;
    int32_t days[7];
    int32_t daysCount = UPRV_LENGTHOF(days);
    UDate until;

    parseRRULE(rrule, month, dayOfWeek, nthDayOfWeek, days, daysCount, until, status);
    if (U_FAILURE(status)) {
        return NULL;
    }

    if (dates->size() == 1) {
        // No more rules
        if (daysCount > 1) {
            // Multiple BYMONTHDAY values
            if (daysCount != 7 || month == -1 || dayOfWeek == 0) {
                // Only support the rule using 7 continuous days
                // BYMONTH and BYDAY must be set at the same time
                goto unsupportedRRule;
            }
            int32_t firstDay = 31; // max possible number of dates in a month
            for (i = 0; i < 7; i++) {
                // Resolve negative day numbers.  A negative day number should
                // not be used in February, but if we see such case, we use 28
                // as the base.
                if (days[i] < 0) {
                    days[i] = MONTHLENGTH[month] + days[i] + 1;
                }
                if (days[i] < firstDay) {
                    firstDay = days[i];
                }
            }
            // Make sure days are continuous
            for (i = 1; i < 7; i++) {
                UBool found = FALSE;
                for (j = 0; j < 7; j++) {
                    if (days[j] == firstDay + i) {
                        found = TRUE;
                        break;
                    }
                }
                if (!found) {
                    // days are not continuous
                    goto unsupportedRRule;
                }
            }
            // Use DOW_GEQ_DOM rule with firstDay as the start date
            dayOfMonth = firstDay;
        }
    } else {
        // Check if BYMONTH + BYMONTHDAY + BYDAY rule with multiple RRULE lines.
        // Otherwise, not supported.
        if (month == -1 || dayOfWeek == 0 || daysCount == 0) {
            // This is not the case
            goto unsupportedRRule;
        }
        // Parse the rest of rules if number of rules is not exceeding 7.
        // We can only support 7 continuous days starting from a day of month.
        if (dates->size() > 7) {
            goto unsupportedRRule;
        }

        // Note: To check valid date range across multiple rule is a little
        // bit complicated.  For now, this code is not doing strict range
        // checking across month boundary

        int32_t earliestMonth = month;
        int32_t earliestDay = 31;
        for (i = 0; i < daysCount; i++) {
            int32_t dom = days[i];
            dom = dom > 0 ? dom : MONTHLENGTH[month] + dom + 1;
            earliestDay = dom < earliestDay ? dom : earliestDay;
        }

        int32_t anotherMonth = -1;
        for (i = 1; i < dates->size(); i++) {
            rrule = *((UnicodeString*)dates->elementAt(i));
            UDate tmp_until;
            int32_t tmp_month, tmp_dayOfWeek, tmp_nthDayOfWeek;
            int32_t tmp_days[7];
            int32_t tmp_daysCount = UPRV_LENGTHOF(tmp_days);
            parseRRULE(rrule, tmp_month, tmp_dayOfWeek, tmp_nthDayOfWeek, tmp_days, tmp_daysCount, tmp_until, status);
            if (U_FAILURE(status)) {
                return NULL;
            }
            // If UNTIL is newer than previous one, use the one
            if (tmp_until > until) {
                until = tmp_until;
            }

            // Check if BYMONTH + BYMONTHDAY + BYDAY rule
            if (tmp_month == -1 || tmp_dayOfWeek == 0 || tmp_daysCount == 0) {
                goto unsupportedRRule;
            }
            // Count number of BYMONTHDAY
            if (daysCount + tmp_daysCount > 7) {
                // We cannot support BYMONTHDAY more than 7
                goto unsupportedRRule;
            }
            // Check if the same BYDAY is used.  Otherwise, we cannot
            // support the rule
            if (tmp_dayOfWeek != dayOfWeek) {
                goto unsupportedRRule;
            }
            // Check if the month is same or right next to the primary month
            if (tmp_month != month) {
                if (anotherMonth == -1) {
                    int32_t diff = tmp_month - month;
                    if (diff == -11 || diff == -1) {
                        // Previous month
                        anotherMonth = tmp_month;
                        earliestMonth = anotherMonth;
                        // Reset earliest day
                        earliestDay = 31;
                    } else if (diff == 11 || diff == 1) {
                        // Next month
                        anotherMonth = tmp_month;
                    } else {
                        // The day range cannot exceed more than 2 months
                        goto unsupportedRRule;
                    }
                } else if (tmp_month != month && tmp_month != anotherMonth) {
                    // The day range cannot exceed more than 2 months
                    goto unsupportedRRule;
                }
            }
            // If ealier month, go through days to find the earliest day
            if (tmp_month == earliestMonth) {
                for (j = 0; j < tmp_daysCount; j++) {
                    tmp_days[j] = tmp_days[j] > 0 ? tmp_days[j] : MONTHLENGTH[tmp_month] + tmp_days[j] + 1;
                    earliestDay = tmp_days[j] < earliestDay ? tmp_days[j] : earliestDay;
                }
            }
            daysCount += tmp_daysCount;
        }
        if (daysCount != 7) {
            // Number of BYMONTHDAY entries must be 7
            goto unsupportedRRule;
        }
        month = earliestMonth;
        dayOfMonth = earliestDay;
    }

    // Calculate start/end year and missing fields
    int32_t startYear, startMonth, startDOM, startDOW, startDOY, startMID;
    Grego::timeToFields(start + fromOffset, startYear, startMonth, startDOM,
        startDOW, startDOY, startMID);
    if (month == -1) {
        // If BYMONTH is not set, use the month of DTSTART
        month = startMonth;
    }
    if (dayOfWeek == 0 && nthDayOfWeek == 0 && dayOfMonth == 0) {
        // If only YEARLY is set, use the day of DTSTART as BYMONTHDAY
        dayOfMonth = startDOM;
    }

    int32_t endYear;
    if (until != MIN_MILLIS) {
        int32_t endMonth, endDOM, endDOW, endDOY, endMID;
        Grego::timeToFields(until, endYear, endMonth, endDOM, endDOW, endDOY, endMID);
    } else {
        endYear = AnnualTimeZoneRule::MAX_YEAR;
    }

    // Create the AnnualDateTimeRule
    if (dayOfWeek == 0 && nthDayOfWeek == 0 && dayOfMonth != 0) {
        // Day in month rule, for example, 15th day in the month
        adtr = new DateTimeRule(month, dayOfMonth, startMID, DateTimeRule::WALL_TIME);
    } else if (dayOfWeek != 0 && nthDayOfWeek != 0 && dayOfMonth == 0) {
        // Nth day of week rule, for example, last Sunday
        adtr = new DateTimeRule(month, nthDayOfWeek, dayOfWeek, startMID, DateTimeRule::WALL_TIME);
    } else if (dayOfWeek != 0 && nthDayOfWeek == 0 && dayOfMonth != 0) {
        // First day of week after day of month rule, for example,
        // first Sunday after 15th day in the month
        adtr = new DateTimeRule(month, dayOfMonth, dayOfWeek, TRUE, startMID, DateTimeRule::WALL_TIME);
    }
    if (adtr == NULL) {
        goto unsupportedRRule;
    }
    return new AnnualTimeZoneRule(zonename, rawOffset, dstSavings, adtr, startYear, endYear);

unsupportedRRule:
    status = U_INVALID_STATE_ERROR;
    return NULL;
}

/*
 * Create a TimeZoneRule by the RDATE definition
 */
static TimeZoneRule* createRuleByRDATE(const UnicodeString& zonename, int32_t rawOffset, int32_t dstSavings,
                                       UDate start, UVector* dates, int32_t fromOffset, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    TimeArrayTimeZoneRule *retVal = NULL;
    if (dates == NULL || dates->size() == 0) {
        // When no RDATE line is provided, use start (DTSTART)
        // as the transition time
        retVal = new TimeArrayTimeZoneRule(zonename, rawOffset, dstSavings,
            &start, 1, DateTimeRule::UTC_TIME);
    } else {
        // Create an array of transition times
        int32_t size = dates->size();
        UDate* times = (UDate*)uprv_malloc(sizeof(UDate) * size);
        if (times == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            return NULL;
        }
        for (int32_t i = 0; i < size; i++) {
            UnicodeString *datestr = (UnicodeString*)dates->elementAt(i);
            times[i] = parseDateTimeString(*datestr, fromOffset, status);
            if (U_FAILURE(status)) {
                uprv_free(times);
                return NULL;
            }
        }
        retVal = new TimeArrayTimeZoneRule(zonename, rawOffset, dstSavings,
            times, size, DateTimeRule::UTC_TIME);
        uprv_free(times);
    }
    return retVal;
}

/*
 * Check if the DOW rule specified by month, weekInMonth and dayOfWeek is equivalent
 * to the DateTimerule.
 */
static UBool isEquivalentDateRule(int32_t month, int32_t weekInMonth, int32_t dayOfWeek, const DateTimeRule *dtrule) {
    if (month != dtrule->getRuleMonth() || dayOfWeek != dtrule->getRuleDayOfWeek()) {
        return FALSE;
    }
    if (dtrule->getTimeRuleType() != DateTimeRule::WALL_TIME) {
        // Do not try to do more intelligent comparison for now.
        return FALSE;
    }
    if (dtrule->getDateRuleType() == DateTimeRule::DOW
            && dtrule->getRuleWeekInMonth() == weekInMonth) {
        return TRUE;
    }
    int32_t ruleDOM = dtrule->getRuleDayOfMonth();
    if (dtrule->getDateRuleType() == DateTimeRule::DOW_GEQ_DOM) {
        if (ruleDOM%7 == 1 && (ruleDOM + 6)/7 == weekInMonth) {
            return TRUE;
        }
        if (month != UCAL_FEBRUARY && (MONTHLENGTH[month] - ruleDOM)%7 == 6
                && weekInMonth == -1*((MONTHLENGTH[month]-ruleDOM+1)/7)) {
            return TRUE;
        }
    }
    if (dtrule->getDateRuleType() == DateTimeRule::DOW_LEQ_DOM) {
        if (ruleDOM%7 == 0 && ruleDOM/7 == weekInMonth) {
            return TRUE;
        }
        if (month != UCAL_FEBRUARY && (MONTHLENGTH[month] - ruleDOM)%7 == 0
                && weekInMonth == -1*((MONTHLENGTH[month] - ruleDOM)/7 + 1)) {
            return TRUE;
        }
    }
    return FALSE;
}

/*
 * Convert the rule to its equivalent rule using WALL_TIME mode.
 * This function returns NULL when the specified DateTimeRule is already
 * using WALL_TIME mode.
 */
static DateTimeRule* toWallTimeRule(const DateTimeRule* rule, int32_t rawOffset, int32_t dstSavings) {
    if (rule->getTimeRuleType() == DateTimeRule::WALL_TIME) {
        return NULL;
    }
    int32_t wallt = rule->getRuleMillisInDay();
    if (rule->getTimeRuleType() == DateTimeRule::UTC_TIME) {
        wallt += (rawOffset + dstSavings);
    } else if (rule->getTimeRuleType() == DateTimeRule::STANDARD_TIME) {
        wallt += dstSavings;
    }

    int32_t month = -1, dom = 0, dow = 0;
    DateTimeRule::DateRuleType dtype;
    int32_t dshift = 0;
    if (wallt < 0) {
        dshift = -1;
        wallt += U_MILLIS_PER_DAY;
    } else if (wallt >= U_MILLIS_PER_DAY) {
        dshift = 1;
        wallt -= U_MILLIS_PER_DAY;
    }

    month = rule->getRuleMonth();
    dom = rule->getRuleDayOfMonth();
    dow = rule->getRuleDayOfWeek();
    dtype = rule->getDateRuleType();

    if (dshift != 0) {
        if (dtype == DateTimeRule::DOW) {
            // Convert to DOW_GEW_DOM or DOW_LEQ_DOM rule first
            int32_t wim = rule->getRuleWeekInMonth();
            if (wim > 0) {
                dtype = DateTimeRule::DOW_GEQ_DOM;
                dom = 7 * (wim - 1) + 1;
            } else {
                dtype = DateTimeRule::DOW_LEQ_DOM;
                dom = MONTHLENGTH[month] + 7 * (wim + 1);
            }
        }
        // Shift one day before or after
        dom += dshift;
        if (dom == 0) {
            month--;
            month = month < UCAL_JANUARY ? UCAL_DECEMBER : month;
            dom = MONTHLENGTH[month];
        } else if (dom > MONTHLENGTH[month]) {
            month++;
            month = month > UCAL_DECEMBER ? UCAL_JANUARY : month;
            dom = 1;
        }
        if (dtype != DateTimeRule::DOM) {
            // Adjust day of week
            dow += dshift;
            if (dow < UCAL_SUNDAY) {
                dow = UCAL_SATURDAY;
            } else if (dow > UCAL_SATURDAY) {
                dow = UCAL_SUNDAY;
            }
        }
    }
    // Create a new rule
    DateTimeRule *modifiedRule;
    if (dtype == DateTimeRule::DOM) {
        modifiedRule = new DateTimeRule(month, dom, wallt, DateTimeRule::WALL_TIME);
    } else {
        modifiedRule = new DateTimeRule(month, dom, dow,
            (dtype == DateTimeRule::DOW_GEQ_DOM), wallt, DateTimeRule::WALL_TIME);
    }
    return modifiedRule;
}

/*
 * Minumum implementations of stream writer/reader, writing/reading
 * UnicodeString.  For now, we do not want to introduce the dependency
 * on the ICU I/O stream in this module.  But we want to keep the code
 * equivalent to the ICU4J implementation, which utilizes java.io.Writer/
 * Reader.
 */
class VTZWriter {
public:
    VTZWriter(UnicodeString& out);
    ~VTZWriter();

    void write(const UnicodeString& str);
    void write(UChar ch);
    void write(const UChar* str);
    //void write(const UChar* str, int32_t length);
private:
    UnicodeString* out;
};

VTZWriter::VTZWriter(UnicodeString& output) {
    out = &output;
}

VTZWriter::~VTZWriter() {
}

void
VTZWriter::write(const UnicodeString& str) {
    out->append(str);
}

void
VTZWriter::write(UChar ch) {
    out->append(ch);
}

void
VTZWriter::write(const UChar* str) {
    out->append(str, -1);
}

/*
void
VTZWriter::write(const UChar* str, int32_t length) {
    out->append(str, length);
}
*/

class VTZReader {
public:
    VTZReader(const UnicodeString& input);
    ~VTZReader();

    UChar read(void);
private:
    const UnicodeString* in;
    int32_t index;
};

VTZReader::VTZReader(const UnicodeString& input) {
    in = &input;
    index = 0;
}

VTZReader::~VTZReader() {
}

UChar
VTZReader::read(void) {
    UChar ch = 0xFFFF;
    if (index < in->length()) {
        ch = in->charAt(index);
    }
    index++;
    return ch;
}


UOBJECT_DEFINE_RTTI_IMPLEMENTATION(VTimeZone)

VTimeZone::VTimeZone()
:   BasicTimeZone(), tz(NULL), vtzlines(NULL),
    lastmod(MAX_MILLIS) {
}

VTimeZone::VTimeZone(const VTimeZone& source)
:   BasicTimeZone(source), tz(NULL), vtzlines(NULL),
    tzurl(source.tzurl), lastmod(source.lastmod),
    olsonzid(source.olsonzid), icutzver(source.icutzver) {
    if (source.tz != NULL) {
        tz = source.tz->clone();
    }
    if (source.vtzlines != NULL) {
        UErrorCode status = U_ZERO_ERROR;
        int32_t size = source.vtzlines->size();
        vtzlines = new UVector(uprv_deleteUObject, uhash_compareUnicodeString, size, status);
        if (U_SUCCESS(status)) {
            for (int32_t i = 0; i < size; i++) {
                UnicodeString *line = (UnicodeString*)source.vtzlines->elementAt(i);
                vtzlines->addElement(line->clone(), status);
                if (U_FAILURE(status)) {
                    break;
                }
            }
        }
        if (U_FAILURE(status) && vtzlines != NULL) {
            delete vtzlines;
        }
    }
}

VTimeZone::~VTimeZone() {
    if (tz != NULL) {
        delete tz;
    }
    if (vtzlines != NULL) {
        delete vtzlines;
    }
}

VTimeZone&
VTimeZone::operator=(const VTimeZone& right) {
    if (this == &right) {
        return *this;
    }
    if (*this != right) {
        BasicTimeZone::operator=(right);
        if (tz != NULL) {
            delete tz;
            tz = NULL;
        }
        if (right.tz != NULL) {
            tz = right.tz->clone();
        }
        if (vtzlines != NULL) {
            delete vtzlines;
        }
        if (right.vtzlines != NULL) {
            UErrorCode status = U_ZERO_ERROR;
            int32_t size = right.vtzlines->size();
            vtzlines = new UVector(uprv_deleteUObject, uhash_compareUnicodeString, size, status);
            if (U_SUCCESS(status)) {
                for (int32_t i = 0; i < size; i++) {
                    UnicodeString *line = (UnicodeString*)right.vtzlines->elementAt(i);
                    vtzlines->addElement(line->clone(), status);
                    if (U_FAILURE(status)) {
                        break;
                    }
                }
            }
            if (U_FAILURE(status) && vtzlines != NULL) {
                delete vtzlines;
                vtzlines = NULL;
            }
        }
        tzurl = right.tzurl;
        lastmod = right.lastmod;
        olsonzid = right.olsonzid;
        icutzver = right.icutzver;
    }
    return *this;
}

UBool
VTimeZone::operator==(const TimeZone& that) const {
    if (this == &that) {
        return TRUE;
    }
    if (typeid(*this) != typeid(that) || !BasicTimeZone::operator==(that)) {
        return FALSE;
    }
    VTimeZone *vtz = (VTimeZone*)&that;
    if (*tz == *(vtz->tz)
        && tzurl == vtz->tzurl
        && lastmod == vtz->lastmod
        /* && olsonzid = that.olsonzid */
        /* && icutzver = that.icutzver */) {
        return TRUE;
    }
    return FALSE;
}

UBool
VTimeZone::operator!=(const TimeZone& that) const {
    return !operator==(that);
}

VTimeZone*
VTimeZone::createVTimeZoneByID(const UnicodeString& ID) {
    VTimeZone *vtz = new VTimeZone();
    vtz->tz = (BasicTimeZone*)TimeZone::createTimeZone(ID);
    vtz->tz->getID(vtz->olsonzid);

    // Set ICU tzdata version
    UErrorCode status = U_ZERO_ERROR;
    UResourceBundle *bundle = NULL;
    const UChar* versionStr = NULL;
    int32_t len = 0;
    bundle = ures_openDirect(NULL, "zoneinfo64", &status);
    versionStr = ures_getStringByKey(bundle, "TZVersion", &len, &status);
    if (U_SUCCESS(status)) {
        vtz->icutzver.setTo(versionStr, len);
    }
    ures_close(bundle);
    return vtz;
}

VTimeZone*
VTimeZone::createVTimeZoneFromBasicTimeZone(const BasicTimeZone& basic_time_zone, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    VTimeZone *vtz = new VTimeZone();
    if (vtz == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return NULL;
    }
    vtz->tz = basic_time_zone.clone();
    if (vtz->tz == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        delete vtz;
        return NULL;
    }
    vtz->tz->getID(vtz->olsonzid);

    // Set ICU tzdata version
    UResourceBundle *bundle = NULL;
    const UChar* versionStr = NULL;
    int32_t len = 0;
    bundle = ures_openDirect(NULL, "zoneinfo64", &status);
    versionStr = ures_getStringByKey(bundle, "TZVersion", &len, &status);
    if (U_SUCCESS(status)) {
        vtz->icutzver.setTo(versionStr, len);
    }
    ures_close(bundle);
    return vtz;
}

VTimeZone*
VTimeZone::createVTimeZone(const UnicodeString& vtzdata, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }
    VTZReader reader(vtzdata);
    VTimeZone *vtz = new VTimeZone();
    vtz->load(reader, status);
    if (U_FAILURE(status)) {
        delete vtz;
        return NULL;
    }
    return vtz;
}

UBool
VTimeZone::getTZURL(UnicodeString& url) const {
    if (tzurl.length() > 0) {
        url = tzurl;
        return TRUE;
    }
    return FALSE;
}

void
VTimeZone::setTZURL(const UnicodeString& url) {
    tzurl = url;
}

UBool
VTimeZone::getLastModified(UDate& lastModified) const {
    if (lastmod != MAX_MILLIS) {
        lastModified = lastmod;
        return TRUE;
    }
    return FALSE;
}

void
VTimeZone::setLastModified(UDate lastModified) {
    lastmod = lastModified;
}

void
VTimeZone::write(UnicodeString& result, UErrorCode& status) const {
    result.remove();
    VTZWriter writer(result);
    write(writer, status);
}

void
VTimeZone::write(UDate start, UnicodeString& result, UErrorCode& status) const {
    result.remove();
    VTZWriter writer(result);
    write(start, writer, status);
}

void
VTimeZone::writeSimple(UDate time, UnicodeString& result, UErrorCode& status) const {
    result.remove();
    VTZWriter writer(result);
    writeSimple(time, writer, status);
}

VTimeZone*
VTimeZone::clone() const {
    return new VTimeZone(*this);
}

int32_t
VTimeZone::getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                     uint8_t dayOfWeek, int32_t millis, UErrorCode& status) const {
    return tz->getOffset(era, year, month, day, dayOfWeek, millis, status);
}

int32_t
VTimeZone::getOffset(uint8_t era, int32_t year, int32_t month, int32_t day,
                     uint8_t dayOfWeek, int32_t millis,
                     int32_t monthLength, UErrorCode& status) const {
    return tz->getOffset(era, year, month, day, dayOfWeek, millis, monthLength, status);
}

void
VTimeZone::getOffset(UDate date, UBool local, int32_t& rawOffset,
                     int32_t& dstOffset, UErrorCode& status) const {
    return tz->getOffset(date, local, rawOffset, dstOffset, status);
}

void
VTimeZone::setRawOffset(int32_t offsetMillis) {
    tz->setRawOffset(offsetMillis);
}

int32_t
VTimeZone::getRawOffset(void) const {
    return tz->getRawOffset();
}

UBool
VTimeZone::useDaylightTime(void) const {
    return tz->useDaylightTime();
}

UBool
VTimeZone::inDaylightTime(UDate date, UErrorCode& status) const {
    return tz->inDaylightTime(date, status);
}

UBool
VTimeZone::hasSameRules(const TimeZone& other) const {
    return tz->hasSameRules(other);
}

UBool
VTimeZone::getNextTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const {
    return tz->getNextTransition(base, inclusive, result);
}

UBool
VTimeZone::getPreviousTransition(UDate base, UBool inclusive, TimeZoneTransition& result) const {
    return tz->getPreviousTransition(base, inclusive, result);
}

int32_t
VTimeZone::countTransitionRules(UErrorCode& status) const {
    return tz->countTransitionRules(status);
}

void
VTimeZone::getTimeZoneRules(const InitialTimeZoneRule*& initial,
                            const TimeZoneRule* trsrules[], int32_t& trscount,
                            UErrorCode& status) const {
    tz->getTimeZoneRules(initial, trsrules, trscount, status);
}

void
VTimeZone::load(VTZReader& reader, UErrorCode& status) {
    vtzlines = new UVector(uprv_deleteUObject, uhash_compareUnicodeString, DEFAULT_VTIMEZONE_LINES, status);
    if (U_FAILURE(status)) {
        return;
    }
    UBool eol = FALSE;
    UBool start = FALSE;
    UBool success = FALSE;
    UnicodeString line;

    while (TRUE) {
        UChar ch = reader.read();
        if (ch == 0xFFFF) {
            // end of file
            if (start && line.startsWith(ICAL_END_VTIMEZONE, -1)) {
                vtzlines->addElement(new UnicodeString(line), status);
                if (U_FAILURE(status)) {
                    goto cleanupVtzlines;
                }
                success = TRUE;
            }
            break;
        }
        if (ch == 0x000D) {
            // CR, must be followed by LF according to the definition in RFC2445
            continue;
        }
        if (eol) {
            if (ch != 0x0009 && ch != 0x0020) {
                // NOT followed by TAB/SP -> new line
                if (start) {
                    if (line.length() > 0) {
                        vtzlines->addElement(new UnicodeString(line), status);
                        if (U_FAILURE(status)) {
                            goto cleanupVtzlines;
                        }
                    }
                }
                line.remove();
                if (ch != 0x000A) {
                    line.append(ch);
                }
            }
            eol = FALSE;
        } else {
            if (ch == 0x000A) {
                // LF
                eol = TRUE;
                if (start) {
                    if (line.startsWith(ICAL_END_VTIMEZONE, -1)) {
                        vtzlines->addElement(new UnicodeString(line), status);
                        if (U_FAILURE(status)) {
                            goto cleanupVtzlines;
                        }
                        success = TRUE;
                        break;
                    }
                } else {
                    if (line.startsWith(ICAL_BEGIN_VTIMEZONE, -1)) {
                        vtzlines->addElement(new UnicodeString(line), status);
                        if (U_FAILURE(status)) {
                            goto cleanupVtzlines;
                        }
                        line.remove();
                        start = TRUE;
                        eol = FALSE;
                    }
                }
            } else {
                line.append(ch);
            }
        }
    }
    if (!success) {
        if (U_SUCCESS(status)) {
            status = U_INVALID_STATE_ERROR;
        }
        goto cleanupVtzlines;
    }
    parse(status);
    return;

cleanupVtzlines:
    delete vtzlines;
    vtzlines = NULL;
}

// parser state
#define vtzone_INI 0   // Initial state
#define vtzone_VTZ 1   // In VTIMEZONE
#define vtzone_TZI 2   // In STANDARD or DAYLIGHT

#define DEF_DSTSAVINGS (60*60*1000)
#define DEF_TZSTARTTIME (0.0)

void
VTimeZone::parse(UErrorCode& status) {
    if (U_FAILURE(status)) {
        return;
    }
    if (vtzlines == NULL || vtzlines->size() == 0) {
        status = U_INVALID_STATE_ERROR;
        return;
    }
    InitialTimeZoneRule *initialRule = NULL;
    RuleBasedTimeZone *rbtz = NULL;

    // timezone ID
    UnicodeString tzid;

    int32_t state = vtzone_INI;
    int32_t n = 0;
    UBool dst = FALSE;      // current zone type
    UnicodeString from;     // current zone from offset
    UnicodeString to;       // current zone offset
    UnicodeString zonename;   // current zone name
    UnicodeString dtstart;  // current zone starts
    UBool isRRULE = FALSE;  // true if the rule is described by RRULE
    int32_t initialRawOffset = 0;   // initial offset
    int32_t initialDSTSavings = 0;  // initial offset
    UDate firstStart = MAX_MILLIS;  // the earliest rule start time
    UnicodeString name;     // RFC2445 prop name
    UnicodeString value;    // RFC2445 prop value

    UVector *dates = NULL;  // list of RDATE or RRULE strings
    UVector *rules = NULL;  // list of TimeZoneRule instances

    int32_t finalRuleIdx = -1;
    int32_t finalRuleCount = 0;

    rules = new UVector(status);
    if (U_FAILURE(status)) {
        goto cleanupParse;
    }
     // Set the deleter to remove TimeZoneRule vectors to avoid memory leaks due to unowned TimeZoneRules.
    rules->setDeleter(deleteTimeZoneRule);

    dates = new UVector(uprv_deleteUObject, uhash_compareUnicodeString, status);
    if (U_FAILURE(status)) {
        goto cleanupParse;
    }
    if (rules == NULL || dates == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        goto cleanupParse;
    }

    for (n = 0; n < vtzlines->size(); n++) {
        UnicodeString *line = (UnicodeString*)vtzlines->elementAt(n);
        int32_t valueSep = line->indexOf(vtzone_COLON);
        if (valueSep < 0) {
            continue;
        }
        name.setTo(*line, 0, valueSep);
        value.setTo(*line, valueSep + 1);

        switch (state) {
        case vtzone_INI:
            if (name.compare(ICAL_BEGIN, -1) == 0
                && value.compare(ICAL_VTIMEZONE, -1) == 0) {
                state = vtzone_VTZ;
            }
            break;

        case vtzone_VTZ:
            if (name.compare(ICAL_TZID, -1) == 0) {
                tzid = value;
            } else if (name.compare(ICAL_TZURL, -1) == 0) {
                tzurl = value;
            } else if (name.compare(ICAL_LASTMOD, -1) == 0) {
                // Always in 'Z' format, so the offset argument for the parse method
                // can be any value.
                lastmod = parseDateTimeString(value, 0, status);
                if (U_FAILURE(status)) {
                    goto cleanupParse;
                }
            } else if (name.compare(ICAL_BEGIN, -1) == 0) {
                UBool isDST = (value.compare(ICAL_DAYLIGHT, -1) == 0);
                if (value.compare(ICAL_STANDARD, -1) == 0 || isDST) {
                    // tzid must be ready at this point
                    if (tzid.length() == 0) {
                        goto cleanupParse;
                    }
                    // initialize current zone properties
                    if (dates->size() != 0) {
                        dates->removeAllElements();
                    }
                    isRRULE = FALSE;
                    from.remove();
                    to.remove();
                    zonename.remove();
                    dst = isDST;
                    state = vtzone_TZI;
                } else {
                    // BEGIN property other than STANDARD/DAYLIGHT
                    // must not be there.
                    goto cleanupParse;
                }
            } else if (name.compare(ICAL_END, -1) == 0) {
                break;
            }
            break;
        case vtzone_TZI:
            if (name.compare(ICAL_DTSTART, -1) == 0) {
                dtstart = value;
            } else if (name.compare(ICAL_TZNAME, -1) == 0) {
                zonename = value;
            } else if (name.compare(ICAL_TZOFFSETFROM, -1) == 0) {
                from = value;
            } else if (name.compare(ICAL_TZOFFSETTO, -1) == 0) {
                to = value;
            } else if (name.compare(ICAL_RDATE, -1) == 0) {
                // RDATE mixed with RRULE is not supported
                if (isRRULE) {
                    goto cleanupParse;
                }
                // RDATE value may contain multiple date delimited
                // by comma
                UBool nextDate = TRUE;
                int32_t dstart = 0;
                UnicodeString *dstr;
                while (nextDate) {
                    int32_t dend = value.indexOf(vtzone_COMMA, dstart);
                    if (dend == -1) {
                        dstr = new UnicodeString(value, dstart);
                        nextDate = FALSE;
                    } else {
                        dstr = new UnicodeString(value, dstart, dend - dstart);
                    }
                    dates->addElement(dstr, status);
                    if (U_FAILURE(status)) {
                        goto cleanupParse;
                    }
                    dstart = dend + 1;
                }
            } else if (name.compare(ICAL_RRULE, -1) == 0) {
                // RRULE mixed with RDATE is not supported
                if (!isRRULE && dates->size() != 0) {
                    goto cleanupParse;
                }
                isRRULE = true;
                dates->addElement(new UnicodeString(value), status);
                if (U_FAILURE(status)) {
                    goto cleanupParse;
                }
            } else if (name.compare(ICAL_END, -1) == 0) {
                // Mandatory properties
                if (dtstart.length() == 0 || from.length() == 0 || to.length() == 0) {
                    goto cleanupParse;
                }
                // if zonename is not available, create one from tzid
                if (zonename.length() == 0) {
                    getDefaultTZName(tzid, dst, zonename);
                }

                // create a time zone rule
                TimeZoneRule *rule = NULL;
                int32_t fromOffset = 0;
                int32_t toOffset = 0;
                int32_t rawOffset = 0;
                int32_t dstSavings = 0;
                UDate start = 0;

                // Parse TZOFFSETFROM/TZOFFSETTO
                fromOffset = offsetStrToMillis(from, status);
                toOffset = offsetStrToMillis(to, status);
                if (U_FAILURE(status)) {
                    goto cleanupParse;
                }

                if (dst) {
                    // If daylight, use the previous offset as rawoffset if positive
                    if (toOffset - fromOffset > 0) {
                        rawOffset = fromOffset;
                        dstSavings = toOffset - fromOffset;
                    } else {
                        // This is rare case..  just use 1 hour DST savings
                        rawOffset = toOffset - DEF_DSTSAVINGS;
                        dstSavings = DEF_DSTSAVINGS;
                    }
                } else {
                    rawOffset = toOffset;
                    dstSavings = 0;
                }

                // start time
                start = parseDateTimeString(dtstart, fromOffset, status);
                if (U_FAILURE(status)) {
                    goto cleanupParse;
                }

                // Create the rule
                UDate actualStart = MAX_MILLIS;
                if (isRRULE) {
                    rule = createRuleByRRULE(zonename, rawOffset, dstSavings, start, dates, fromOffset, status);
                } else {
                    rule = createRuleByRDATE(zonename, rawOffset, dstSavings, start, dates, fromOffset, status);
                }
                if (U_FAILURE(status) || rule == NULL) {
                    goto cleanupParse;
                } else {
                    UBool startAvail = rule->getFirstStart(fromOffset, 0, actualStart);
                    if (startAvail && actualStart < firstStart) {
                        // save from offset information for the earliest rule
                        firstStart = actualStart;
                        // If this is STD, assume the time before this transtion
                        // is DST when the difference is 1 hour.  This might not be
                        // accurate, but VTIMEZONE data does not have such info.
                        if (dstSavings > 0) {
                            initialRawOffset = fromOffset;
                            initialDSTSavings = 0;
                        } else {
                            if (fromOffset - toOffset == DEF_DSTSAVINGS) {
                                initialRawOffset = fromOffset - DEF_DSTSAVINGS;
                                initialDSTSavings = DEF_DSTSAVINGS;
                            } else {
                                initialRawOffset = fromOffset;
                                initialDSTSavings = 0;
                            }
                        }
                    }
                }
                rules->addElement(rule, status);
                if (U_FAILURE(status)) {
                    goto cleanupParse;
                }
                state = vtzone_VTZ;
            }
            break;
        }
    }
    // Must have at least one rule
    if (rules->size() == 0) {
        goto cleanupParse;
    }

    // Create a initial rule
    getDefaultTZName(tzid, FALSE, zonename);
    initialRule = new InitialTimeZoneRule(zonename,
        initialRawOffset, initialDSTSavings);
    if (initialRule == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        goto cleanupParse;
    }

    // Finally, create the RuleBasedTimeZone
    rbtz = new RuleBasedTimeZone(tzid, initialRule);
    if (rbtz == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        goto cleanupParse;
    }
    initialRule = NULL; // already adopted by RBTZ, no need to delete

    for (n = 0; n < rules->size(); n++) {
        TimeZoneRule *r = (TimeZoneRule*)rules->elementAt(n);
        AnnualTimeZoneRule *atzrule = dynamic_cast<AnnualTimeZoneRule *>(r);
        if (atzrule != NULL) {
            if (atzrule->getEndYear() == AnnualTimeZoneRule::MAX_YEAR) {
                finalRuleCount++;
                finalRuleIdx = n;
            }
        }
    }
    if (finalRuleCount > 2) {
        // Too many final rules
        status = U_ILLEGAL_ARGUMENT_ERROR;
        goto cleanupParse;
    }

    if (finalRuleCount == 1) {
        if (rules->size() == 1) {
            // Only one final rule, only governs the initial rule,
            // which is already initialized, thus, we do not need to
            // add this transition rule
            rules->removeAllElements();
        } else {
            // Normalize the final rule
            AnnualTimeZoneRule *finalRule = (AnnualTimeZoneRule*)rules->elementAt(finalRuleIdx);
            int32_t tmpRaw = finalRule->getRawOffset();
            int32_t tmpDST = finalRule->getDSTSavings();

            // Find the last non-final rule
            UDate finalStart, start;
            finalRule->getFirstStart(initialRawOffset, initialDSTSavings, finalStart);
            start = finalStart;
            for (n = 0; n < rules->size(); n++) {
                if (finalRuleIdx == n) {
                    continue;
                }
                TimeZoneRule *r = (TimeZoneRule*)rules->elementAt(n);
                UDate lastStart;
                r->getFinalStart(tmpRaw, tmpDST, lastStart);
                if (lastStart > start) {
                    finalRule->getNextStart(lastStart,
                        r->getRawOffset(),
                        r->getDSTSavings(),
                        FALSE,
                        start);
                }
            }

            TimeZoneRule *newRule;
            UnicodeString tznam;
            if (start == finalStart) {
                // Transform this into a single transition
                newRule = new TimeArrayTimeZoneRule(
                        finalRule->getName(tznam),
                        finalRule->getRawOffset(),
                        finalRule->getDSTSavings(),
                        &finalStart,
                        1,
                        DateTimeRule::UTC_TIME);
            } else {
                // Update the end year
                int32_t y, m, d, dow, doy, mid;
                Grego::timeToFields(start, y, m, d, dow, doy, mid);
                newRule = new AnnualTimeZoneRule(
                        finalRule->getName(tznam),
                        finalRule->getRawOffset(),
                        finalRule->getDSTSavings(),
                        *(finalRule->getRule()),
                        finalRule->getStartYear(),
                        y);
            }
            if (newRule == NULL) {
                status = U_MEMORY_ALLOCATION_ERROR;
                goto cleanupParse;
            }
            rules->removeElementAt(finalRuleIdx);
            rules->addElement(newRule, status);
            if (U_FAILURE(status)) {
                delete newRule;
                goto cleanupParse;
            }
        }
    }

    while (!rules->isEmpty()) {
        TimeZoneRule *tzr = (TimeZoneRule*)rules->orphanElementAt(0);
        rbtz->addTransitionRule(tzr, status);
        if (U_FAILURE(status)) {
            goto cleanupParse;
        }
    }
    rbtz->complete(status);
    if (U_FAILURE(status)) {
        goto cleanupParse;
    }
    delete rules;
    delete dates;

    tz = rbtz;
    setID(tzid);
    return;

cleanupParse:
    if (rules != NULL) {
        while (!rules->isEmpty()) {
            TimeZoneRule *r = (TimeZoneRule*)rules->orphanElementAt(0);
            delete r;
        }
        delete rules;
    }
    if (dates != NULL) {
        delete dates;
    }
    if (initialRule != NULL) {
        delete initialRule;
    }
    if (rbtz != NULL) {
        delete rbtz;
    }
    return;
}

void
VTimeZone::write(VTZWriter& writer, UErrorCode& status) const {
    if (vtzlines != NULL) {
        for (int32_t i = 0; i < vtzlines->size(); i++) {
            UnicodeString *line = (UnicodeString*)vtzlines->elementAt(i);
            if (line->startsWith(ICAL_TZURL, -1)
                && line->charAt(u_strlen(ICAL_TZURL)) == vtzone_COLON) {
                writer.write(ICAL_TZURL);
                writer.write(vtzone_COLON);
                writer.write(tzurl);
                writer.write(ICAL_NEWLINE);
            } else if (line->startsWith(ICAL_LASTMOD, -1)
                && line->charAt(u_strlen(ICAL_LASTMOD)) == vtzone_COLON) {
                UnicodeString utcString;
                writer.write(ICAL_LASTMOD);
                writer.write(vtzone_COLON);
                writer.write(getUTCDateTimeString(lastmod, utcString));
                writer.write(ICAL_NEWLINE);
            } else {
                writer.write(*line);
                writer.write(ICAL_NEWLINE);
            }
        }
    } else {
        UnicodeString icutzprop;
        UVector customProps(nullptr, uhash_compareUnicodeString, status);
        if (olsonzid.length() > 0 && icutzver.length() > 0) {
            icutzprop.append(olsonzid);
            icutzprop.append(u'[');
            icutzprop.append(icutzver);
            icutzprop.append(u']');
            customProps.addElement(&icutzprop, status);
        }
        writeZone(writer, *tz, &customProps, status);
    }
}

void
VTimeZone::write(UDate start, VTZWriter& writer, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    InitialTimeZoneRule *initial = NULL;
    UVector *transitionRules = NULL;
    UVector customProps(uprv_deleteUObject, uhash_compareUnicodeString, status);
    UnicodeString tzid;

    // Extract rules applicable to dates after the start time
    getTimeZoneRulesAfter(start, initial, transitionRules, status);
    if (U_FAILURE(status)) {
        return;
    }

    // Create a RuleBasedTimeZone with the subset rule
    getID(tzid);
    RuleBasedTimeZone rbtz(tzid, initial);
    if (transitionRules != NULL) {
        while (!transitionRules->isEmpty()) {
            TimeZoneRule *tr = (TimeZoneRule*)transitionRules->orphanElementAt(0);
            rbtz.addTransitionRule(tr, status);
            if (U_FAILURE(status)) {
                goto cleanupWritePartial;
            }
        }
        delete transitionRules;
        transitionRules = NULL;
    }
    rbtz.complete(status);
    if (U_FAILURE(status)) {
        goto cleanupWritePartial;
    }

    if (olsonzid.length() > 0 && icutzver.length() > 0) {
        UnicodeString *icutzprop = new UnicodeString(ICU_TZINFO_PROP);
        icutzprop->append(olsonzid);
        icutzprop->append((UChar)0x005B/*'['*/);
        icutzprop->append(icutzver);
        icutzprop->append(ICU_TZINFO_PARTIAL, -1);
        appendMillis(start, *icutzprop);
        icutzprop->append((UChar)0x005D/*']'*/);
        customProps.addElement(icutzprop, status);
        if (U_FAILURE(status)) {
            delete icutzprop;
            goto cleanupWritePartial;
        }
    }
    writeZone(writer, rbtz, &customProps, status);
    return;

cleanupWritePartial:
    if (initial != NULL) {
        delete initial;
    }
    if (transitionRules != NULL) {
        while (!transitionRules->isEmpty()) {
            TimeZoneRule *tr = (TimeZoneRule*)transitionRules->orphanElementAt(0);
            delete tr;
        }
        delete transitionRules;
    }
}

void
VTimeZone::writeSimple(UDate time, VTZWriter& writer, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }

    UVector customProps(uprv_deleteUObject, uhash_compareUnicodeString, status);
    UnicodeString tzid;

    // Extract simple rules
    InitialTimeZoneRule *initial = NULL;
    AnnualTimeZoneRule *std = NULL, *dst = NULL;
    getSimpleRulesNear(time, initial, std, dst, status);
    if (U_SUCCESS(status)) {
        // Create a RuleBasedTimeZone with the subset rule
        getID(tzid);
        RuleBasedTimeZone rbtz(tzid, initial);
        if (std != NULL && dst != NULL) {
            rbtz.addTransitionRule(std, status);
            rbtz.addTransitionRule(dst, status);
        }
        if (U_FAILURE(status)) {
            goto cleanupWriteSimple;
        }

        if (olsonzid.length() > 0 && icutzver.length() > 0) {
            UnicodeString *icutzprop = new UnicodeString(ICU_TZINFO_PROP);
            icutzprop->append(olsonzid);
            icutzprop->append((UChar)0x005B/*'['*/);
            icutzprop->append(icutzver);
            icutzprop->append(ICU_TZINFO_SIMPLE, -1);
            appendMillis(time, *icutzprop);
            icutzprop->append((UChar)0x005D/*']'*/);
            customProps.addElement(icutzprop, status);
            if (U_FAILURE(status)) {
                delete icutzprop;
                goto cleanupWriteSimple;
            }
        }
        writeZone(writer, rbtz, &customProps, status);
    }
    return;

cleanupWriteSimple:
    if (initial != NULL) {
        delete initial;
    }
    if (std != NULL) {
        delete std;
    }
    if (dst != NULL) {
        delete dst;
    }
}

void
VTimeZone::writeZone(VTZWriter& w, BasicTimeZone& basictz,
                     UVector* customProps, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    writeHeaders(w, status);
    if (U_FAILURE(status)) {
        return;
    }

    if (customProps != NULL) {
        for (int32_t i = 0; i < customProps->size(); i++) {
            UnicodeString *custprop = (UnicodeString*)customProps->elementAt(i);
            w.write(*custprop);
            w.write(ICAL_NEWLINE);
        }
    }

    UDate t = MIN_MILLIS;
    UnicodeString dstName;
    int32_t dstFromOffset = 0;
    int32_t dstFromDSTSavings = 0;
    int32_t dstToOffset = 0;
    int32_t dstStartYear = 0;
    int32_t dstMonth = 0;
    int32_t dstDayOfWeek = 0;
    int32_t dstWeekInMonth = 0;
    int32_t dstMillisInDay = 0;
    UDate dstStartTime = 0.0;
    UDate dstUntilTime = 0.0;
    int32_t dstCount = 0;
    AnnualTimeZoneRule *finalDstRule = NULL;

    UnicodeString stdName;
    int32_t stdFromOffset = 0;
    int32_t stdFromDSTSavings = 0;
    int32_t stdToOffset = 0;
    int32_t stdStartYear = 0;
    int32_t stdMonth = 0;
    int32_t stdDayOfWeek = 0;
    int32_t stdWeekInMonth = 0;
    int32_t stdMillisInDay = 0;
    UDate stdStartTime = 0.0;
    UDate stdUntilTime = 0.0;
    int32_t stdCount = 0;
    AnnualTimeZoneRule *finalStdRule = NULL;

    int32_t year, month, dom, dow, doy, mid;
    UBool hasTransitions = FALSE;
    TimeZoneTransition tzt;
    UBool tztAvail;
    UnicodeString name;
    UBool isDst;

    // Going through all transitions
    while (TRUE) {
        tztAvail = basictz.getNextTransition(t, FALSE, tzt);
        if (!tztAvail) {
            break;
        }
        hasTransitions = TRUE;
        t = tzt.getTime();
        tzt.getTo()->getName(name);
        isDst = (tzt.getTo()->getDSTSavings() != 0);
        int32_t fromOffset = tzt.getFrom()->getRawOffset() + tzt.getFrom()->getDSTSavings();
        int32_t fromDSTSavings = tzt.getFrom()->getDSTSavings();
        int32_t toOffset = tzt.getTo()->getRawOffset() + tzt.getTo()->getDSTSavings();
        Grego::timeToFields(tzt.getTime() + fromOffset, year, month, dom, dow, doy, mid);
        int32_t weekInMonth = Grego::dayOfWeekInMonth(year, month, dom);
        UBool sameRule = FALSE;
        const AnnualTimeZoneRule *atzrule;
        if (isDst) {
            if (finalDstRule == NULL
                && (atzrule = dynamic_cast<const AnnualTimeZoneRule *>(tzt.getTo())) != NULL
                && atzrule->getEndYear() == AnnualTimeZoneRule::MAX_YEAR
            ) {
                finalDstRule = atzrule->clone();
            }
            if (dstCount > 0) {
                if (year == dstStartYear + dstCount
                        && name.compare(dstName) == 0
                        && dstFromOffset == fromOffset
                        && dstToOffset == toOffset
                        && dstMonth == month
                        && dstDayOfWeek == dow
                        && dstWeekInMonth == weekInMonth
                        && dstMillisInDay == mid) {
                    // Update until time
                    dstUntilTime = t;
                    dstCount++;
                    sameRule = TRUE;
                }
                if (!sameRule) {
                    if (dstCount == 1) {
                        writeZonePropsByTime(w, TRUE, dstName, dstFromOffset, dstToOffset, dstStartTime,
                                TRUE, status);
                    } else {
                        writeZonePropsByDOW(w, TRUE, dstName, dstFromOffset, dstToOffset,
                                dstMonth, dstWeekInMonth, dstDayOfWeek, dstStartTime, dstUntilTime, status);
                    }
                    if (U_FAILURE(status)) {
                        goto cleanupWriteZone;
                    }
                }
            }
            if (!sameRule) {
                // Reset this DST information
                dstName = name;
                dstFromOffset = fromOffset;
                dstFromDSTSavings = fromDSTSavings;
                dstToOffset = toOffset;
                dstStartYear = year;
                dstMonth = month;
                dstDayOfWeek = dow;
                dstWeekInMonth = weekInMonth;
                dstMillisInDay = mid;
                dstStartTime = dstUntilTime = t;
                dstCount = 1;
            }
            if (finalStdRule != NULL && finalDstRule != NULL) {
                break;
            }
        } else {
            if (finalStdRule == NULL
                && (atzrule = dynamic_cast<const AnnualTimeZoneRule *>(tzt.getTo())) != NULL
                && atzrule->getEndYear() == AnnualTimeZoneRule::MAX_YEAR
            ) {
                finalStdRule = atzrule->clone();
            }
            if (stdCount > 0) {
                if (year == stdStartYear + stdCount
                        && name.compare(stdName) == 0
                        && stdFromOffset == fromOffset
                        && stdToOffset == toOffset
                        && stdMonth == month
                        && stdDayOfWeek == dow
                        && stdWeekInMonth == weekInMonth
                        && stdMillisInDay == mid) {
                    // Update until time
                    stdUntilTime = t;
                    stdCount++;
                    sameRule = TRUE;
                }
                if (!sameRule) {
                    if (stdCount == 1) {
                        writeZonePropsByTime(w, FALSE, stdName, stdFromOffset, stdToOffset, stdStartTime,
                                TRUE, status);
                    } else {
                        writeZonePropsByDOW(w, FALSE, stdName, stdFromOffset, stdToOffset,
                                stdMonth, stdWeekInMonth, stdDayOfWeek, stdStartTime, stdUntilTime, status);
                    }
                    if (U_FAILURE(status)) {
                        goto cleanupWriteZone;
                    }
                }
            }
            if (!sameRule) {
                // Reset this STD information
                stdName = name;
                stdFromOffset = fromOffset;
                stdFromDSTSavings = fromDSTSavings;
                stdToOffset = toOffset;
                stdStartYear = year;
                stdMonth = month;
                stdDayOfWeek = dow;
                stdWeekInMonth = weekInMonth;
                stdMillisInDay = mid;
                stdStartTime = stdUntilTime = t;
                stdCount = 1;
            }
            if (finalStdRule != NULL && finalDstRule != NULL) {
                break;
            }
        }
    }
    if (!hasTransitions) {
        // No transition - put a single non transition RDATE
        int32_t raw, dst, offset;
        basictz.getOffset(0.0/*any time*/, FALSE, raw, dst, status);
        if (U_FAILURE(status)) {
            goto cleanupWriteZone;
        }
        offset = raw + dst;
        isDst = (dst != 0);
        UnicodeString tzid;
        basictz.getID(tzid);
        getDefaultTZName(tzid, isDst, name);
        writeZonePropsByTime(w, isDst, name,
                offset, offset, DEF_TZSTARTTIME - offset, FALSE, status);
        if (U_FAILURE(status)) {
            goto cleanupWriteZone;
        }
    } else {
        if (dstCount > 0) {
            if (finalDstRule == NULL) {
                if (dstCount == 1) {
                    writeZonePropsByTime(w, TRUE, dstName, dstFromOffset, dstToOffset, dstStartTime,
                            TRUE, status);
                } else {
                    writeZonePropsByDOW(w, TRUE, dstName, dstFromOffset, dstToOffset,
                            dstMonth, dstWeekInMonth, dstDayOfWeek, dstStartTime, dstUntilTime, status);
                }
                if (U_FAILURE(status)) {
                    goto cleanupWriteZone;
                }
            } else {
                if (dstCount == 1) {
                    writeFinalRule(w, TRUE, finalDstRule,
                            dstFromOffset - dstFromDSTSavings, dstFromDSTSavings, dstStartTime, status);
                } else {
                    // Use a single rule if possible
                    if (isEquivalentDateRule(dstMonth, dstWeekInMonth, dstDayOfWeek, finalDstRule->getRule())) {
                        writeZonePropsByDOW(w, TRUE, dstName, dstFromOffset, dstToOffset,
                                dstMonth, dstWeekInMonth, dstDayOfWeek, dstStartTime, MAX_MILLIS, status);
                    } else {
                        // Not equivalent rule - write out two different rules
                        writeZonePropsByDOW(w, TRUE, dstName, dstFromOffset, dstToOffset,
                                dstMonth, dstWeekInMonth, dstDayOfWeek, dstStartTime, dstUntilTime, status);
                        if (U_FAILURE(status)) {
                            goto cleanupWriteZone;
                        }
                        UDate nextStart;
                        UBool nextStartAvail = finalDstRule->getNextStart(dstUntilTime, dstFromOffset - dstFromDSTSavings, dstFromDSTSavings, false, nextStart);
                        U_ASSERT(nextStartAvail);
                        if (nextStartAvail) {
                            writeFinalRule(w, TRUE, finalDstRule,
                                    dstFromOffset - dstFromDSTSavings, dstFromDSTSavings, nextStart, status);
                        }
                    }
                }
                if (U_FAILURE(status)) {
                    goto cleanupWriteZone;
                }
            }
        }
        if (stdCount > 0) {
            if (finalStdRule == NULL) {
                if (stdCount == 1) {
                    writeZonePropsByTime(w, FALSE, stdName, stdFromOffset, stdToOffset, stdStartTime,
                            TRUE, status);
                } else {
                    writeZonePropsByDOW(w, FALSE, stdName, stdFromOffset, stdToOffset,
                            stdMonth, stdWeekInMonth, stdDayOfWeek, stdStartTime, stdUntilTime, status);
                }
                if (U_FAILURE(status)) {
                    goto cleanupWriteZone;
                }
            } else {
                if (stdCount == 1) {
                    writeFinalRule(w, FALSE, finalStdRule,
                            stdFromOffset - stdFromDSTSavings, stdFromDSTSavings, stdStartTime, status);
                } else {
                    // Use a single rule if possible
                    if (isEquivalentDateRule(stdMonth, stdWeekInMonth, stdDayOfWeek, finalStdRule->getRule())) {
                        writeZonePropsByDOW(w, FALSE, stdName, stdFromOffset, stdToOffset,
                                stdMonth, stdWeekInMonth, stdDayOfWeek, stdStartTime, MAX_MILLIS, status);
                    } else {
                        // Not equivalent rule - write out two different rules
                        writeZonePropsByDOW(w, FALSE, stdName, stdFromOffset, stdToOffset,
                                stdMonth, stdWeekInMonth, stdDayOfWeek, stdStartTime, stdUntilTime, status);
                        if (U_FAILURE(status)) {
                            goto cleanupWriteZone;
                        }
                        UDate nextStart;
                        UBool nextStartAvail = finalStdRule->getNextStart(stdUntilTime, stdFromOffset - stdFromDSTSavings, stdFromDSTSavings, false, nextStart);
                        U_ASSERT(nextStartAvail);
                        if (nextStartAvail) {
                            writeFinalRule(w, FALSE, finalStdRule,
                                    stdFromOffset - stdFromDSTSavings, stdFromDSTSavings, nextStart, status);
                        }
                    }
                }
                if (U_FAILURE(status)) {
                    goto cleanupWriteZone;
                }
            }
        }
    }
    writeFooter(w, status);

cleanupWriteZone:

    if (finalStdRule != NULL) {
        delete finalStdRule;
    }
    if (finalDstRule != NULL) {
        delete finalDstRule;
    }
}

void
VTimeZone::writeHeaders(VTZWriter& writer, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    UnicodeString tzid;
    tz->getID(tzid);

    writer.write(ICAL_BEGIN);
    writer.write(vtzone_COLON);
    writer.write(ICAL_VTIMEZONE);
    writer.write(ICAL_NEWLINE);
    writer.write(ICAL_TZID);
    writer.write(vtzone_COLON);
    writer.write(tzid);
    writer.write(ICAL_NEWLINE);
    if (tzurl.length() != 0) {
        writer.write(ICAL_TZURL);
        writer.write(vtzone_COLON);
        writer.write(tzurl);
        writer.write(ICAL_NEWLINE);
    }
    if (lastmod != MAX_MILLIS) {
        UnicodeString lastmodStr;
        writer.write(ICAL_LASTMOD);
        writer.write(vtzone_COLON);
        writer.write(getUTCDateTimeString(lastmod, lastmodStr));
        writer.write(ICAL_NEWLINE);
    }
}

/*
 * Write the closing section of the VTIMEZONE definition block
 */
void
VTimeZone::writeFooter(VTZWriter& writer, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    writer.write(ICAL_END);
    writer.write(vtzone_COLON);
    writer.write(ICAL_VTIMEZONE);
    writer.write(ICAL_NEWLINE);
}

/*
 * Write a single start time
 */
void
VTimeZone::writeZonePropsByTime(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                                int32_t fromOffset, int32_t toOffset, UDate time, UBool withRDATE,
                                UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    beginZoneProps(writer, isDst, zonename, fromOffset, toOffset, time, status);
    if (U_FAILURE(status)) {
        return;
    }
    if (withRDATE) {
        writer.write(ICAL_RDATE);
        writer.write(vtzone_COLON);
        UnicodeString timestr;
        writer.write(getDateTimeString(time + fromOffset, timestr));
        writer.write(ICAL_NEWLINE);
    }
    endZoneProps(writer, isDst, status);
    if (U_FAILURE(status)) {
        return;
    }
}

/*
 * Write start times defined by a DOM rule using VTIMEZONE RRULE
 */
void
VTimeZone::writeZonePropsByDOM(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                               int32_t fromOffset, int32_t toOffset,
                               int32_t month, int32_t dayOfMonth, UDate startTime, UDate untilTime,
                               UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    beginZoneProps(writer, isDst, zonename, fromOffset, toOffset, startTime, status);
    if (U_FAILURE(status)) {
        return;
    }
    beginRRULE(writer, month, status);
    if (U_FAILURE(status)) {
        return;
    }
    writer.write(ICAL_BYMONTHDAY);
    writer.write(vtzone_EQUALS_SIGN);
    UnicodeString dstr;
    appendAsciiDigits(dayOfMonth, 0, dstr);
    writer.write(dstr);
    if (untilTime != MAX_MILLIS) {
        appendUNTIL(writer, getDateTimeString(untilTime + fromOffset, dstr), status);
        if (U_FAILURE(status)) {
            return;
        }
    }
    writer.write(ICAL_NEWLINE);
    endZoneProps(writer, isDst, status);
}

/*
 * Write start times defined by a DOW rule using VTIMEZONE RRULE
 */
void
VTimeZone::writeZonePropsByDOW(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                               int32_t fromOffset, int32_t toOffset,
                               int32_t month, int32_t weekInMonth, int32_t dayOfWeek,
                               UDate startTime, UDate untilTime, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    beginZoneProps(writer, isDst, zonename, fromOffset, toOffset, startTime, status);
    if (U_FAILURE(status)) {
        return;
    }
    beginRRULE(writer, month, status);
    if (U_FAILURE(status)) {
        return;
    }
    writer.write(ICAL_BYDAY);
    writer.write(vtzone_EQUALS_SIGN);
    UnicodeString dstr;
    appendAsciiDigits(weekInMonth, 0, dstr);
    writer.write(dstr);    // -4, -3, -2, -1, 1, 2, 3, 4
    writer.write(ICAL_DOW_NAMES[dayOfWeek - 1]);    // SU, MO, TU...

    if (untilTime != MAX_MILLIS) {
        appendUNTIL(writer, getDateTimeString(untilTime + fromOffset, dstr), status);
        if (U_FAILURE(status)) {
            return;
        }
    }
    writer.write(ICAL_NEWLINE);
    endZoneProps(writer, isDst, status);
}

/*
 * Write start times defined by a DOW_GEQ_DOM rule using VTIMEZONE RRULE
 */
void
VTimeZone::writeZonePropsByDOW_GEQ_DOM(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                                       int32_t fromOffset, int32_t toOffset,
                                       int32_t month, int32_t dayOfMonth, int32_t dayOfWeek,
                                       UDate startTime, UDate untilTime, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    // Check if this rule can be converted to DOW rule
    if (dayOfMonth%7 == 1) {
        // Can be represented by DOW rule
        writeZonePropsByDOW(writer, isDst, zonename, fromOffset, toOffset,
                month, (dayOfMonth + 6)/7, dayOfWeek, startTime, untilTime, status);
        if (U_FAILURE(status)) {
            return;
        }
    } else if (month != UCAL_FEBRUARY && (MONTHLENGTH[month] - dayOfMonth)%7 == 6) {
        // Can be represented by DOW rule with negative week number
        writeZonePropsByDOW(writer, isDst, zonename, fromOffset, toOffset,
                month, -1*((MONTHLENGTH[month] - dayOfMonth + 1)/7), dayOfWeek, startTime, untilTime, status);
        if (U_FAILURE(status)) {
            return;
        }
    } else {
        // Otherwise, use BYMONTHDAY to include all possible dates
        beginZoneProps(writer, isDst, zonename, fromOffset, toOffset, startTime, status);
        if (U_FAILURE(status)) {
            return;
        }
        // Check if all days are in the same month
        int32_t startDay = dayOfMonth;
        int32_t currentMonthDays = 7;

        if (dayOfMonth <= 0) {
            // The start day is in previous month
            int32_t prevMonthDays = 1 - dayOfMonth;
            currentMonthDays -= prevMonthDays;

            int32_t prevMonth = (month - 1) < 0 ? 11 : month - 1;

            // Note: When a rule is separated into two, UNTIL attribute needs to be
            // calculated for each of them.  For now, we skip this, because we basically use this method
            // only for final rules, which does not have the UNTIL attribute
            writeZonePropsByDOW_GEQ_DOM_sub(writer, prevMonth, -prevMonthDays, dayOfWeek, prevMonthDays,
                MAX_MILLIS /* Do not use UNTIL */, fromOffset, status);
            if (U_FAILURE(status)) {
                return;
            }

            // Start from 1 for the rest
            startDay = 1;
        } else if (dayOfMonth + 6 > MONTHLENGTH[month]) {
            // Note: This code does not actually work well in February.  For now, days in month in
            // non-leap year.
            int32_t nextMonthDays = dayOfMonth + 6 - MONTHLENGTH[month];
            currentMonthDays -= nextMonthDays;

            int32_t nextMonth = (month + 1) > 11 ? 0 : month + 1;

            writeZonePropsByDOW_GEQ_DOM_sub(writer, nextMonth, 1, dayOfWeek, nextMonthDays,
                MAX_MILLIS /* Do not use UNTIL */, fromOffset, status);
            if (U_FAILURE(status)) {
                return;
            }
        }
        writeZonePropsByDOW_GEQ_DOM_sub(writer, month, startDay, dayOfWeek, currentMonthDays,
            untilTime, fromOffset, status);
        if (U_FAILURE(status)) {
            return;
        }
        endZoneProps(writer, isDst, status);
    }
}

/*
 * Called from writeZonePropsByDOW_GEQ_DOM
 */
void
VTimeZone::writeZonePropsByDOW_GEQ_DOM_sub(VTZWriter& writer, int32_t month, int32_t dayOfMonth,
                                           int32_t dayOfWeek, int32_t numDays,
                                           UDate untilTime, int32_t fromOffset, UErrorCode& status) const {

    if (U_FAILURE(status)) {
        return;
    }
    int32_t startDayNum = dayOfMonth;
    UBool isFeb = (month == UCAL_FEBRUARY);
    if (dayOfMonth < 0 && !isFeb) {
        // Use positive number if possible
        startDayNum = MONTHLENGTH[month] + dayOfMonth + 1;
    }
    beginRRULE(writer, month, status);
    if (U_FAILURE(status)) {
        return;
    }
    writer.write(ICAL_BYDAY);
    writer.write(vtzone_EQUALS_SIGN);
    writer.write(ICAL_DOW_NAMES[dayOfWeek - 1]);    // SU, MO, TU...
    writer.write(vtzone_SEMICOLON);
    writer.write(ICAL_BYMONTHDAY);
    writer.write(vtzone_EQUALS_SIGN);

    UnicodeString dstr;
    appendAsciiDigits(startDayNum, 0, dstr);
    writer.write(dstr);
    for (int32_t i = 1; i < numDays; i++) {
        writer.write(vtzone_COMMA);
        dstr.remove();
        appendAsciiDigits(startDayNum + i, 0, dstr);
        writer.write(dstr);
    }

    if (untilTime != MAX_MILLIS) {
        appendUNTIL(writer, getDateTimeString(untilTime + fromOffset, dstr), status);
        if (U_FAILURE(status)) {
            return;
        }
    }
    writer.write(ICAL_NEWLINE);
}

/*
 * Write start times defined by a DOW_LEQ_DOM rule using VTIMEZONE RRULE
 */
void
VTimeZone::writeZonePropsByDOW_LEQ_DOM(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                                       int32_t fromOffset, int32_t toOffset,
                                       int32_t month, int32_t dayOfMonth, int32_t dayOfWeek,
                                       UDate startTime, UDate untilTime, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    // Check if this rule can be converted to DOW rule
    if (dayOfMonth%7 == 0) {
        // Can be represented by DOW rule
        writeZonePropsByDOW(writer, isDst, zonename, fromOffset, toOffset,
                month, dayOfMonth/7, dayOfWeek, startTime, untilTime, status);
    } else if (month != UCAL_FEBRUARY && (MONTHLENGTH[month] - dayOfMonth)%7 == 0){
        // Can be represented by DOW rule with negative week number
        writeZonePropsByDOW(writer, isDst, zonename, fromOffset, toOffset,
                month, -1*((MONTHLENGTH[month] - dayOfMonth)/7 + 1), dayOfWeek, startTime, untilTime, status);
    } else if (month == UCAL_FEBRUARY && dayOfMonth == 29) {
        // Specical case for February
        writeZonePropsByDOW(writer, isDst, zonename, fromOffset, toOffset,
                UCAL_FEBRUARY, -1, dayOfWeek, startTime, untilTime, status);
    } else {
        // Otherwise, convert this to DOW_GEQ_DOM rule
        writeZonePropsByDOW_GEQ_DOM(writer, isDst, zonename, fromOffset, toOffset,
                month, dayOfMonth - 6, dayOfWeek, startTime, untilTime, status);
    }
}

/*
 * Write the final time zone rule using RRULE, with no UNTIL attribute
 */
void
VTimeZone::writeFinalRule(VTZWriter& writer, UBool isDst, const AnnualTimeZoneRule* rule,
                          int32_t fromRawOffset, int32_t fromDSTSavings,
                          UDate startTime, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    UBool modifiedRule = TRUE;
    const DateTimeRule *dtrule = toWallTimeRule(rule->getRule(), fromRawOffset, fromDSTSavings);
    if (dtrule == NULL) {
        modifiedRule = FALSE;
        dtrule = rule->getRule();
    }

    // If the rule's mills in a day is out of range, adjust start time.
    // Olson tzdata supports 24:00 of a day, but VTIMEZONE does not.
    // See ticket#7008/#7518

    int32_t timeInDay = dtrule->getRuleMillisInDay();
    if (timeInDay < 0) {
        startTime = startTime + (0 - timeInDay);
    } else if (timeInDay >= U_MILLIS_PER_DAY) {
        startTime = startTime - (timeInDay - (U_MILLIS_PER_DAY - 1));
    }

    int32_t toOffset = rule->getRawOffset() + rule->getDSTSavings();
    UnicodeString name;
    rule->getName(name);
    switch (dtrule->getDateRuleType()) {
    case DateTimeRule::DOM:
        writeZonePropsByDOM(writer, isDst, name, fromRawOffset + fromDSTSavings, toOffset,
                dtrule->getRuleMonth(), dtrule->getRuleDayOfMonth(), startTime, MAX_MILLIS, status);
        break;
    case DateTimeRule::DOW:
        writeZonePropsByDOW(writer, isDst, name, fromRawOffset + fromDSTSavings, toOffset,
                dtrule->getRuleMonth(), dtrule->getRuleWeekInMonth(), dtrule->getRuleDayOfWeek(), startTime, MAX_MILLIS, status);
        break;
    case DateTimeRule::DOW_GEQ_DOM:
        writeZonePropsByDOW_GEQ_DOM(writer, isDst, name, fromRawOffset + fromDSTSavings, toOffset,
                dtrule->getRuleMonth(), dtrule->getRuleDayOfMonth(), dtrule->getRuleDayOfWeek(), startTime, MAX_MILLIS, status);
        break;
    case DateTimeRule::DOW_LEQ_DOM:
        writeZonePropsByDOW_LEQ_DOM(writer, isDst, name, fromRawOffset + fromDSTSavings, toOffset,
                dtrule->getRuleMonth(), dtrule->getRuleDayOfMonth(), dtrule->getRuleDayOfWeek(), startTime, MAX_MILLIS, status);
        break;
    }
    if (modifiedRule) {
        delete dtrule;
    }
}

/*
 * Write the opening section of zone properties
 */
void
VTimeZone::beginZoneProps(VTZWriter& writer, UBool isDst, const UnicodeString& zonename,
                          int32_t fromOffset, int32_t toOffset, UDate startTime, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    writer.write(ICAL_BEGIN);
    writer.write(vtzone_COLON);
    if (isDst) {
        writer.write(ICAL_DAYLIGHT);
    } else {
        writer.write(ICAL_STANDARD);
    }
    writer.write(ICAL_NEWLINE);

    UnicodeString dstr;

    // TZOFFSETTO
    writer.write(ICAL_TZOFFSETTO);
    writer.write(vtzone_COLON);
    millisToOffset(toOffset, dstr);
    writer.write(dstr);
    writer.write(ICAL_NEWLINE);

    // TZOFFSETFROM
    writer.write(ICAL_TZOFFSETFROM);
    writer.write(vtzone_COLON);
    millisToOffset(fromOffset, dstr);
    writer.write(dstr);
    writer.write(ICAL_NEWLINE);

    // TZNAME
    writer.write(ICAL_TZNAME);
    writer.write(vtzone_COLON);
    writer.write(zonename);
    writer.write(ICAL_NEWLINE);

    // DTSTART
    writer.write(ICAL_DTSTART);
    writer.write(vtzone_COLON);
    writer.write(getDateTimeString(startTime + fromOffset, dstr));
    writer.write(ICAL_NEWLINE);
}

/*
 * Writes the closing section of zone properties
 */
void
VTimeZone::endZoneProps(VTZWriter& writer, UBool isDst, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    // END:STANDARD or END:DAYLIGHT
    writer.write(ICAL_END);
    writer.write(vtzone_COLON);
    if (isDst) {
        writer.write(ICAL_DAYLIGHT);
    } else {
        writer.write(ICAL_STANDARD);
    }
    writer.write(ICAL_NEWLINE);
}

/*
 * Write the beggining part of RRULE line
 */
void
VTimeZone::beginRRULE(VTZWriter& writer, int32_t month, UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    UnicodeString dstr;
    writer.write(ICAL_RRULE);
    writer.write(vtzone_COLON);
    writer.write(ICAL_FREQ);
    writer.write(vtzone_EQUALS_SIGN);
    writer.write(ICAL_YEARLY);
    writer.write(vtzone_SEMICOLON);
    writer.write(ICAL_BYMONTH);
    writer.write(vtzone_EQUALS_SIGN);
    appendAsciiDigits(month + 1, 0, dstr);
    writer.write(dstr);
    writer.write(vtzone_SEMICOLON);
}

/*
 * Append the UNTIL attribute after RRULE line
 */
void
VTimeZone::appendUNTIL(VTZWriter& writer, const UnicodeString& until,  UErrorCode& status) const {
    if (U_FAILURE(status)) {
        return;
    }
    if (until.length() > 0) {
        writer.write(vtzone_SEMICOLON);
        writer.write(ICAL_UNTIL);
        writer.write(vtzone_EQUALS_SIGN);
        writer.write(until);
    }
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

//eof
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2011, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

/**
 * \file 
 * \brief C API: VTimeZone classes
 */



#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

/**
* \file
* \brief C API: RFC2445 VTIMEZONE support
*
* <p>This is a C wrapper around the C++ VTimeZone class.</p>
*/




#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

/**
 * \file 
 * \brief C API: Time zone transition classes
 */



#if !UCONFIG_NO_FORMATTING



#ifndef UCNV_H

/**
 * A TimeZoneTransition.  Use the ztrans_* API to manipulate.  Create with
 * ztrans_open*, and destroy with ztrans_close.
 */
struct ZTrans;
typedef struct ZTrans ZTrans;

#endif

/**
 * Constructs a time zone transition with the time and the rules before/after
 * the transition.
 * 
 * @param time  The time of transition in milliseconds since the base time.
 * @param from  The time zone rule used before the transition.
 * @param to    The time zone rule used after the transition.
 */
U_CAPI ZTrans* U_EXPORT2
ztrans_open(UDate time, const void* from, const void* to);

/**
 * Constructs an empty <code>TimeZoneTransition</code>
 */
U_CAPI ZTrans* U_EXPORT2
ztrans_openEmpty();

/**
 * Disposes of the storage used by a ZTrans object.  This function should
 * be called exactly once for objects returned by ztrans_open*.
 * @param trans the object to dispose of
 */
U_CAPI void U_EXPORT2
ztrans_close(ZTrans *trans);

/**
 * Returns a copy of this object.
 * @param rule the original ZRule
 * @return the newly allocated copy of the ZRule
 */
U_CAPI ZTrans* U_EXPORT2
ztrans_clone(ZTrans *trans);

/**
 * Returns true if trans1 is identical to trans2
 * and vis versa.
 * @param trans1 to be checked for containment
 * @param trans2 to be checked for containment
 * @return true if the test condition is met
 */
U_CAPI UBool U_EXPORT2
ztrans_equals(const ZTrans* trans1, const ZTrans* trans2);

/**
 * Returns the time of transition in milliseconds.
 * param trans, the transition to use
 * @return The time of the transition in milliseconds since the 1970 Jan 1 epoch time.
 */
U_CAPI UDate U_EXPORT2
ztrans_getTime(ZTrans* trans);

/**
 * Sets the time of transition in milliseconds.
 * param trans, the transition to use
 * @param time The time of the transition in milliseconds since the 1970 Jan 1 epoch time.
 */
U_CAPI void U_EXPORT2
ztrans_setTime(ZTrans* trans, UDate time);

/**
 * Returns the rule used before the transition.
 * param trans, the transition to use
 * @return The time zone rule used after the transition.
 */
U_CAPI void* U_EXPORT2
ztrans_getFrom(ZTrans* & trans);

/**
 * Sets the rule used before the transition.  The caller remains
 * responsible for deleting the TimeZoneRule object.
 * param trans, the transition to use
 * param trans, the transition to use
 * @param from The time zone rule used before the transition.
 */
U_CAPI void U_EXPORT2
ztrans_setFrom(ZTrans* trans, const void* from);

/**
 * Adopts the rule used before the transition.  The caller must
 * not delete the TimeZoneRule object passed in.
 * param trans, the transition to use
 * @param from The time zone rule used before the transition.
 */
U_CAPI void U_EXPORT2
ztrans_adoptFrom(ZTrans* trans, void* from);

/**
 * Returns the rule used after the transition.
 * param trans, the transition to use
 * @return The time zone rule used after the transition.
 */
U_CAPI void* U_EXPORT2
ztrans_getTo(ZTrans* trans);

/**
 * Sets the rule used after the transition.  The caller remains
 * responsible for deleting the TimeZoneRule object.
 * param trans, the transition to use
 * @param to The time zone rule used after the transition.
 */
U_CAPI void U_EXPORT2
ztrans_setTo(ZTrans* trans, const void* to);

/**
 * Adopts the rule used after the transition.  The caller must
 * not delete the TimeZoneRule object passed in.
 * param trans, the transition to use
 * @param to The time zone rule used after the transition.
 */
U_CAPI void U_EXPORT2
ztrans_adoptTo(ZTrans* trans, void* to);

/**
 * Return the class ID for this class. This is useful only for comparing to
 * a return value from getDynamicClassID(). For example:
 * <pre>
 * .   Base* polymorphic_pointer = createPolymorphicObject();
 * .   if (polymorphic_pointer->getDynamicClassID() ==
 * .       erived::getStaticClassID()) ...
 * </pre>
 * param trans, the transition to use
 * @return          The class ID for all objects of this class.
 */
U_CAPI UClassID U_EXPORT2
ztrans_getStaticClassID(ZTrans* trans);

/**
 * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
 * method is to implement a simple version of RTTI, since not all C++
 * compilers support genuine RTTI. Polymorphic operator==() and clone()
 * methods call this method.
 *
 * param trans, the transition to use
 * @return          The class ID for this object. All objects of a
 *                  given class have the same class ID.  Objects of
 *                  other classes have different class IDs.
 */
U_CAPI UClassID U_EXPORT2
ztrans_getDynamicClassID(ZTrans* trans);

#endif



#ifndef UCNV_H
struct VZone;
/**
 * A UnicodeSet.  Use the vzone_* API to manipulate.  Create with
 * vzone_open*, and destroy with vzone_close.
 */
typedef struct VZone VZone;
#endif

/*********************************************************************
 * VZone API
 *********************************************************************/

/**
 * Creates a vzone from the given time zone ID.
 * @param ID The time zone ID, such as America/New_York
 * @param idLength, length of the ID parameter
 * @return A vzone object initialized by the time zone ID,
 * or NULL when the ID is unknown.
 */
U_CAPI VZone* U_EXPORT2
vzone_openID(const UChar* ID, int32_t idLength);

/**
 * Create a vzone instance by RFC2445 VTIMEZONE data
 * @param vtzdata The string including VTIMEZONE data block
 * @param vtzdataLength, length of the vtzdata
 * @param status Output param to filled in with a success or an error.
 * @return A vzone initialized by the VTIMEZONE data or
 * NULL if failed to load the rule from the VTIMEZONE data.
 */
U_CAPI VZone* U_EXPORT2
vzone_openData(const UChar* vtzdata, int32_t vtzdataLength, UErrorCode& status);

/**
 * Disposes of the storage used by a VZone object.  This function should
 * be called exactly once for objects returned by vzone_open*.
 * @param set the object to dispose of
 */
U_CAPI void U_EXPORT2
vzone_close(VZone* zone);

/**
 * Returns a copy of this object.
 * @param zone the original vzone
 * @return the newly allocated copy of the vzone
 */
U_CAPI VZone* U_EXPORT2
vzone_clone(const VZone *zone);

/**
 * Returns true if zone1 is identical to zone2
 * and vis versa.
 * @param zone1 to be checked for containment
 * @param zone2 to be checked for containment
 * @return true if the test condition is met
 */
U_CAPI UBool U_EXPORT2
vzone_equals(const VZone* zone1, const VZone* zone2);

/**
 * Gets the RFC2445 TZURL property value.  When a vzone instance was
 * created from VTIMEZONE data, the initial value is set by the TZURL 
 * property value in the data.  Otherwise, the initial value is not set.
 * @param zone, the vzone to use
 * @param url Receives the RFC2445 TZURL property value.
 * @param urlLength, length of the url
 * @return TRUE if TZURL attribute is available and value is set.
 */
U_CAPI UBool U_EXPORT2
vzone_getTZURL(VZone* zone, UChar* & url, int32_t & urlLength);

/**
 * Sets the RFC2445 TZURL property value.
 * @param zone, the vzone to use
 * @param url The TZURL property value.
 * @param urlLength, length of the url
 */
U_CAPI void U_EXPORT2
vzone_setTZURL(VZone* zone, UChar* url, int32_t urlLength);

/**
 * Gets the RFC2445 LAST-MODIFIED property value.  When a vzone instance
 * was created from VTIMEZONE data, the initial value is set by the 
 * LAST-MODIFIED property value in the data.  Otherwise, the initial value 
 * is not set.
 * @param zone, the vzone to use
 * @param lastModified Receives the last modified date.
 * @return TRUE if lastModified attribute is available and value is set.
 */
U_CAPI UBool U_EXPORT2
vzone_getLastModified(VZone* zone, UDate& lastModified);

/**
 * Sets the RFC2445 LAST-MODIFIED property value.
 * @param zone, the vzone to use
 * @param lastModified The LAST-MODIFIED date.
 */
U_CAPI void U_EXPORT2
vzone_setLastModified(VZone* zone, UDate lastModified);

/**
 * Writes RFC2445 VTIMEZONE data for this time zone
 * @param zone, the vzone to use
 * @param result Output param to filled in with the VTIMEZONE data.
 * @param resultLength, length of the result output
 * @param status Output param to filled in with a success or an error.
 */
U_CAPI void U_EXPORT2
vzone_write(VZone* zone, UChar* & result, int32_t & resultLength, UErrorCode& status);

/**
 * Writes RFC2445 VTIMEZONE data for this time zone applicalbe
 * for dates after the specified start time.
 * @param zone, the vzone to use
 * @param start The start date.
 * @param result Output param to filled in with the VTIMEZONE data.
 * @param resultLength, length of the result output
 * @param status Output param to filled in with a success or an error.
 */
U_CAPI void U_EXPORT2
vzone_writeFromStart(VZone* zone, UDate start, UChar* & result, int32_t & resultLength, UErrorCode& status);

/**
 * Writes RFC2445 VTIMEZONE data applicalbe for the specified date.
 * Some common iCalendar implementations can only handle a single time
 * zone property or a pair of standard and daylight time properties using
 * BYDAY rule with day of week (such as BYDAY=1SUN).  This method produce
 * the VTIMEZONE data which can be handled these implementations.  The rules
 * produced by this method can be used only for calculating time zone offset
 * around the specified date.
 * @param zone, the vzone to use
 * @param time The date used for rule extraction.
 * @param result Output param to filled in with the VTIMEZONE data.
 * @param status Output param to filled in with a success or an error.
 */
U_CAPI void U_EXPORT2
vzone_writeSimple(VZone* zone, UDate time, UChar* & result, int32_t & resultLength, UErrorCode& status);

/**
 * Returns the TimeZone's adjusted GMT offset (i.e., the number of milliseconds to add
 * to GMT to get local time in this time zone, taking daylight savings time into
 * account) as of a particular reference date.  The reference date is used to determine
 * whether daylight savings time is in effect and needs to be figured into the offset
 * that is returned (in other words, what is the adjusted GMT offset in this time zone
 * at this particular date and time?).  For the time zones produced by createTimeZone(),
 * the reference data is specified according to the Gregorian calendar, and the date
 * and time fields are local standard time.
 *
 * <p>Note: Don't call this method. Instead, call the getOffset(UDate...) overload,
 * which returns both the raw and the DST offset for a given time. This method
 * is retained only for backward compatibility.
 *
 * @param zone, the vzone to use
 * @param era        The reference date's era
 * @param year       The reference date's year
 * @param month      The reference date's month (0-based; 0 is January)
 * @param day        The reference date's day-in-month (1-based)
 * @param dayOfWeek  The reference date's day-of-week (1-based; 1 is Sunday)
 * @param millis     The reference date's milliseconds in day, local standard time
 * @param status     Output param to filled in with a success or an error.
 * @return           The offset in milliseconds to add to GMT to get local time.
 */
U_CAPI int32_t U_EXPORT2
vzone_getOffset(VZone* zone, uint8_t era, int32_t year, int32_t month, int32_t day,
                uint8_t dayOfWeek, int32_t millis, UErrorCode& status);

/**
 * Gets the time zone offset, for current date, modified in case of
 * daylight savings. This is the offset to add *to* UTC to get local time.
 *
 * <p>Note: Don't call this method. Instead, call the getOffset(UDate...) overload,
 * which returns both the raw and the DST offset for a given time. This method
 * is retained only for backward compatibility.
 *
 * @param zone, the vzone to use
 * @param era        The reference date's era
 * @param year       The reference date's year
 * @param month      The reference date's month (0-based; 0 is January)
 * @param day        The reference date's day-in-month (1-based)
 * @param dayOfWeek  The reference date's day-of-week (1-based; 1 is Sunday)
 * @param millis     The reference date's milliseconds in day, local standard time
 * @param monthLength The length of the given month in days.
 * @param status     Output param to filled in with a success or an error.
 * @return           The offset in milliseconds to add to GMT to get local time.
 */
U_CAPI int32_t U_EXPORT2
vzone_getOffset2(VZone* zone, uint8_t era, int32_t year, int32_t month, int32_t day,
                uint8_t dayOfWeek, int32_t millis,
                int32_t monthLength, UErrorCode& status);

/**
 * Returns the time zone raw and GMT offset for the given moment
 * in time.  Upon return, local-millis = GMT-millis + rawOffset +
 * dstOffset.  All computations are performed in the proleptic
 * Gregorian calendar.  The default implementation in the TimeZone
 * class delegates to the 8-argument getOffset().
 *
 * @param zone, the vzone to use
 * @param date moment in time for which to return offsets, in
 * units of milliseconds from January 1, 1970 0:00 GMT, either GMT
 * time or local wall time, depending on `local'.
 * @param local if true, `date' is local wall time; otherwise it
 * is in GMT time.
 * @param rawOffset output parameter to receive the raw offset, that
 * is, the offset not including DST adjustments
 * @param dstOffset output parameter to receive the DST offset,
 * that is, the offset to be added to `rawOffset' to obtain the
 * total offset between local and GMT time. If DST is not in
 * effect, this value is zero; otherwise it is a positive value,
 * typically one hour.
 * @param ec input-output error code
 */
U_CAPI void U_EXPORT2
vzone_getOffset3(VZone* zone, UDate date, UBool local, int32_t& rawOffset,
                int32_t& dstOffset, UErrorCode& ec);

/**
 * Sets the TimeZone's raw GMT offset (i.e., the number of milliseconds to add
 * to GMT to get local time, before taking daylight savings time into account).
 *
 * @param zone, the vzone to use
 * @param offsetMillis  The new raw GMT offset for this time zone.
 */
U_CAPI void U_EXPORT2
vzone_setRawOffset(VZone* zone, int32_t offsetMillis);

/**
 * Returns the TimeZone's raw GMT offset (i.e., the number of milliseconds to add
 * to GMT to get local time, before taking daylight savings time into account).
 *
 * @param zone, the vzone to use
 * @return   The TimeZone's raw GMT offset.
 */
U_CAPI int32_t U_EXPORT2
vzone_getRawOffset(VZone* zone);

/**
 * Queries if this time zone uses daylight savings time.
 * @param zone, the vzone to use
 * @return true if this time zone uses daylight savings time,
 * false, otherwise.
 */
U_CAPI UBool U_EXPORT2
vzone_useDaylightTime(VZone* zone);

/**
 * Queries if the given date is in daylight savings time in
 * this time zone.
 * This method is wasteful since it creates a new GregorianCalendar and
 * deletes it each time it is called. This is a deprecated method
 * and provided only for Java compatibility.
 *
 * @param zone, the vzone to use
 * @param date the given UDate.
 * @param status Output param filled in with success/error code.
 * @return true if the given date is in daylight savings time,
 * false, otherwise.
 */
U_INTERNAL UBool U_EXPORT2
vzone_inDaylightTime(VZone* zone, UDate date, UErrorCode& status);

/**
 * Returns true if this zone has the same rule and offset as another zone.
 * That is, if this zone differs only in ID, if at all.
 * @param zone, the vzone to use
 * @param other the <code>TimeZone</code> object to be compared with
 * @return true if the given zone is the same as this one,
 * with the possible exception of the ID
 */
U_CAPI UBool U_EXPORT2
vzone_hasSameRules(VZone* zone, const VZone* other);

/**
 * Gets the first time zone transition after the base time.
 * @param zone, the vzone to use
 * @param base      The base time.
 * @param inclusive Whether the base time is inclusive or not.
 * @param result    Receives the first transition after the base time.
 * @return  TRUE if the transition is found.
 */
U_CAPI UBool U_EXPORT2
vzone_getNextTransition(VZone* zone, UDate base, UBool inclusive, ZTrans* result);

/**
 * Gets the most recent time zone transition before the base time.
 * @param zone, the vzone to use
 * @param base      The base time.
 * @param inclusive Whether the base time is inclusive or not.
 * @param result    Receives the most recent transition before the base time.
 * @return  TRUE if the transition is found.
 */
U_CAPI UBool U_EXPORT2
vzone_getPreviousTransition(VZone* zone, UDate base, UBool inclusive, ZTrans* result);

/**
 * Returns the number of <code>TimeZoneRule</code>s which represents time transitions,
 * for this time zone, that is, all <code>TimeZoneRule</code>s for this time zone except
 * <code>InitialTimeZoneRule</code>.  The return value range is 0 or any positive value.
 * @param zone, the vzone to use     
 * @param status    Receives error status code.
 * @return The number of <code>TimeZoneRule</code>s representing time transitions.
 */
U_CAPI int32_t U_EXPORT2
vzone_countTransitionRules(VZone* zone, UErrorCode& status);

/**
 * Return the class ID for this class. This is useful only for comparing to
 * a return value from getDynamicClassID(). For example:
 * <pre>
 * .   Base* polymorphic_pointer = createPolymorphicObject();
 * .   if (polymorphic_pointer->getDynamicClassID() ==
 * .       erived::getStaticClassID()) ...
 * </pre>
 * @param zone, the vzone to use
 * @return          The class ID for all objects of this class.
 */
U_CAPI UClassID U_EXPORT2
vzone_getStaticClassID(VZone* zone);

/**
 * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
 * method is to implement a simple version of RTTI, since not all C++
 * compilers support genuine RTTI. Polymorphic operator==() and clone()
 * methods call this method.
 *
 * @param zone, the vzone to use
 * @return          The class ID for this object. All objects of a
 *                  given class have the same class ID.  Objects of
 *                  other classes have different class IDs.
 */
U_CAPI UClassID U_EXPORT2
vzone_getDynamicClassID(VZone* zone);

#endif // __VZONE_H







U_NAMESPACE_USE

U_CAPI VZone* U_EXPORT2
vzone_openID(const UChar* ID, int32_t idLength){
    UnicodeString s(idLength==-1, ID, idLength);
    return (VZone*) (VTimeZone::createVTimeZoneByID(s));
}
    
U_CAPI VZone* U_EXPORT2
vzone_openData(const UChar* vtzdata, int32_t vtzdataLength, UErrorCode& status) {
    UnicodeString s(vtzdataLength==-1, vtzdata, vtzdataLength);
    return (VZone*) (VTimeZone::createVTimeZone(s,status));
}

U_CAPI void U_EXPORT2
vzone_close(VZone* zone) {
    delete (VTimeZone*)zone;
}

U_CAPI VZone* U_EXPORT2
vzone_clone(const VZone *zone) {
    return (VZone*) (((VTimeZone*)zone)->VTimeZone::clone());
}

U_CAPI UBool U_EXPORT2
vzone_equals(const VZone* zone1, const VZone* zone2) {
    return *(const VTimeZone*)zone1 == *(const VTimeZone*)zone2;
}

U_CAPI UBool U_EXPORT2
vzone_getTZURL(VZone* zone, UChar* & url, int32_t & urlLength) {
    UnicodeString s;
    UBool b = ((VTimeZone*)zone)->VTimeZone::getTZURL(s);

    urlLength = s.length();
    memcpy(url,s.getBuffer(),urlLength);
    
    return b;
}

U_CAPI void U_EXPORT2
vzone_setTZURL(VZone* zone, UChar* url, int32_t urlLength) {
    UnicodeString s(urlLength==-1, url, urlLength);
    ((VTimeZone*)zone)->VTimeZone::setTZURL(s);
}

U_CAPI UBool U_EXPORT2
vzone_getLastModified(VZone* zone, UDate& lastModified) {
    return ((VTimeZone*)zone)->VTimeZone::getLastModified(lastModified);
}

U_CAPI void U_EXPORT2
vzone_setLastModified(VZone* zone, UDate lastModified) {
    return ((VTimeZone*)zone)->VTimeZone::setLastModified(lastModified);
}

U_CAPI void U_EXPORT2
vzone_write(VZone* zone, UChar* & result, int32_t & resultLength, UErrorCode& status) {
    UnicodeString s;
    ((VTimeZone*)zone)->VTimeZone::write(s, status);

    resultLength = s.length();
    result = (UChar*)uprv_malloc(resultLength);
    memcpy(result,s.getBuffer(),resultLength);

    return;
}

U_CAPI void U_EXPORT2
vzone_writeFromStart(VZone* zone, UDate start, UChar* & result, int32_t & resultLength, UErrorCode& status) {
    UnicodeString s;
    ((VTimeZone*)zone)->VTimeZone::write(start, s, status);

    resultLength = s.length();
    result = (UChar*)uprv_malloc(resultLength);
    memcpy(result,s.getBuffer(),resultLength);

    return;
}

U_CAPI void U_EXPORT2
vzone_writeSimple(VZone* zone, UDate time, UChar* & result, int32_t & resultLength, UErrorCode& status) {
    UnicodeString s;
    ((VTimeZone*)zone)->VTimeZone::writeSimple(time, s, status);

    resultLength = s.length();
    result = (UChar*)uprv_malloc(resultLength);
    memcpy(result,s.getBuffer(),resultLength);

    return;
}

U_CAPI int32_t U_EXPORT2
vzone_getOffset(VZone* zone, uint8_t era, int32_t year, int32_t month, int32_t day,
                uint8_t dayOfWeek, int32_t millis, UErrorCode& status) {
    return ((VTimeZone*)zone)->VTimeZone::getOffset(era, year, month, day, dayOfWeek, millis, status);
}

U_CAPI int32_t U_EXPORT2
vzone_getOffset2(VZone* zone, uint8_t era, int32_t year, int32_t month, int32_t day,
                uint8_t dayOfWeek, int32_t millis,
                int32_t monthLength, UErrorCode& status) {
    return ((VTimeZone*)zone)->VTimeZone::getOffset(era, year, month, day, dayOfWeek, millis, monthLength, status);
}

U_CAPI void U_EXPORT2
vzone_getOffset3(VZone* zone, UDate date, UBool local, int32_t& rawOffset,
                int32_t& dstOffset, UErrorCode& ec) {
    return ((VTimeZone*)zone)->VTimeZone::getOffset(date, local, rawOffset, dstOffset, ec);
}

U_CAPI void U_EXPORT2
vzone_setRawOffset(VZone* zone, int32_t offsetMillis) {
    return ((VTimeZone*)zone)->VTimeZone::setRawOffset(offsetMillis);
}

U_CAPI int32_t U_EXPORT2
vzone_getRawOffset(VZone* zone) {
    return ((VTimeZone*)zone)->VTimeZone::getRawOffset();
}

U_CAPI UBool U_EXPORT2
vzone_useDaylightTime(VZone* zone) {
    return ((VTimeZone*)zone)->VTimeZone::useDaylightTime();
}

U_CAPI UBool U_EXPORT2
vzone_inDaylightTime(VZone* zone, UDate date, UErrorCode& status) {
    return ((VTimeZone*)zone)->VTimeZone::inDaylightTime(date, status);
}

U_CAPI UBool U_EXPORT2
vzone_hasSameRules(VZone* zone, const VZone* other) {
    return ((VTimeZone*)zone)->VTimeZone::hasSameRules(*(VTimeZone*)other);
}

U_CAPI UBool U_EXPORT2
vzone_getNextTransition(VZone* zone, UDate base, UBool inclusive, ZTrans* result) {
    return ((VTimeZone*)zone)->VTimeZone::getNextTransition(base, inclusive, *(TimeZoneTransition*)result);
}

U_CAPI UBool U_EXPORT2
vzone_getPreviousTransition(VZone* zone, UDate base, UBool inclusive, ZTrans* result) {
    return ((VTimeZone*)zone)->VTimeZone::getPreviousTransition(base, inclusive, *(TimeZoneTransition*)result);
}

U_CAPI int32_t U_EXPORT2
vzone_countTransitionRules(VZone* zone, UErrorCode& status) {
    return ((VTimeZone*)zone)->VTimeZone::countTransitionRules(status);
}

U_CAPI UClassID U_EXPORT2
vzone_getStaticClassID(VZone* zone) {
    return ((VTimeZone*)zone)->VTimeZone::getStaticClassID();
}

U_CAPI UClassID U_EXPORT2
vzone_getDynamicClassID(VZone* zone) {
    return ((VTimeZone*)zone)->VTimeZone::getDynamicClassID();
}

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2007-2014, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/



#if !UCONFIG_NO_FORMATTING




















static icu::UMutex gZoneMetaLock;

// CLDR Canonical ID mapping table
static UHashtable *gCanonicalIDCache = NULL;
static icu::UInitOnce gCanonicalIDCacheInitOnce = U_INITONCE_INITIALIZER;

// Metazone mapping table
static UHashtable *gOlsonToMeta = NULL;
static icu::UInitOnce gOlsonToMetaInitOnce = U_INITONCE_INITIALIZER;

// Available metazone IDs vector and table
static icu::UVector *gMetaZoneIDs = NULL;
static UHashtable *gMetaZoneIDTable = NULL;
static icu::UInitOnce gMetaZoneIDsInitOnce = U_INITONCE_INITIALIZER;

// Country info vectors
static icu::UVector *gSingleZoneCountries = NULL;
static icu::UVector *gMultiZonesCountries = NULL;
static icu::UInitOnce gCountryInfoVectorsInitOnce = U_INITONCE_INITIALIZER;

U_CDECL_BEGIN

/**
 * Cleanup callback func
 */
static UBool U_CALLCONV zoneMeta_cleanup(void)
{
    if (gCanonicalIDCache != NULL) {
        uhash_close(gCanonicalIDCache);
        gCanonicalIDCache = NULL;
    }
    gCanonicalIDCacheInitOnce.reset();

    if (gOlsonToMeta != NULL) {
        uhash_close(gOlsonToMeta);
        gOlsonToMeta = NULL;
    }
    gOlsonToMetaInitOnce.reset();

    if (gMetaZoneIDTable != NULL) {
        uhash_close(gMetaZoneIDTable);
        gMetaZoneIDTable = NULL;
    }
    // delete after closing gMetaZoneIDTable, because it holds
    // value objects held by the hashtable
    delete gMetaZoneIDs;
    gMetaZoneIDs = NULL;
    gMetaZoneIDsInitOnce.reset();

    delete gSingleZoneCountries;
    gSingleZoneCountries = NULL;
    delete gMultiZonesCountries;
    gMultiZonesCountries = NULL;
    gCountryInfoVectorsInitOnce.reset();

    return TRUE;
}

/**
 * Deleter for UChar* string
 */
static void U_CALLCONV
deleteUCharString(void *obj) {
    UChar *entry = (UChar*)obj;
    uprv_free(entry);
}

/**
 * Deleter for UVector
 */
static void U_CALLCONV
deleteUVector(void *obj) {
   delete (icu::UVector*) obj;
}

/**
 * Deleter for OlsonToMetaMappingEntry
 */
static void U_CALLCONV
deleteOlsonToMetaMappingEntry(void *obj) {
    icu::OlsonToMetaMappingEntry *entry = (icu::OlsonToMetaMappingEntry*)obj;
    uprv_free(entry);
}

U_CDECL_END

U_NAMESPACE_BEGIN

#define ZID_KEY_MAX 128

static const char gMetaZones[]          = "metaZones";
static const char gMetazoneInfo[]       = "metazoneInfo";
static const char gMapTimezonesTag[]    = "mapTimezones";

static const char gKeyTypeData[]        = "keyTypeData";
static const char gTypeAliasTag[]       = "typeAlias";
static const char gTypeMapTag[]         = "typeMap";
static const char gTimezoneTag[]        = "timezone";

static const char gPrimaryZonesTag[]    = "primaryZones";

static const char gWorldTag[]           = "001";

static const UChar gWorld[] = {0x30, 0x30, 0x31, 0x00}; // "001"

static const UChar gDefaultFrom[] = {0x31, 0x39, 0x37, 0x30, 0x2D, 0x30, 0x31, 0x2D, 0x30, 0x31,
                                     0x20, 0x30, 0x30, 0x3A, 0x30, 0x30, 0x00}; // "1970-01-01 00:00"
static const UChar gDefaultTo[]   = {0x39, 0x39, 0x39, 0x39, 0x2D, 0x31, 0x32, 0x2D, 0x33, 0x31,
                                     0x20, 0x32, 0x33, 0x3A, 0x35, 0x39, 0x00}; // "9999-12-31 23:59"

static const UChar gCustomTzPrefix[]    = {0x47, 0x4D, 0x54, 0};    // "GMT"

#define ASCII_DIGIT(c) (((c)>=0x30 && (c)<=0x39) ? (c)-0x30 : -1)

/*
 * Convert a date string used by metazone mappings to UDate.
 * The format used by CLDR metazone mapping is "yyyy-MM-dd HH:mm".
 */
static UDate
parseDate (const UChar *text, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return 0;
    }
    int32_t len = u_strlen(text);
    if (len != 16 && len != 10) {
        // It must be yyyy-MM-dd HH:mm (length 16) or yyyy-MM-dd (length 10)
        status = U_INVALID_FORMAT_ERROR;
        return 0;
    }

    int32_t year = 0, month = 0, day = 0, hour = 0, min = 0, n;
    int32_t idx;

    // "yyyy" (0 - 3)
    for (idx = 0; idx <= 3 && U_SUCCESS(status); idx++) {
        n = ASCII_DIGIT((int32_t)text[idx]);
        if (n >= 0) {
            year = 10*year + n;
        } else {
            status = U_INVALID_FORMAT_ERROR;
        }
    }
    // "MM" (5 - 6)
    for (idx = 5; idx <= 6 && U_SUCCESS(status); idx++) {
        n = ASCII_DIGIT((int32_t)text[idx]);
        if (n >= 0) {
            month = 10*month + n;
        } else {
            status = U_INVALID_FORMAT_ERROR;
        }
    }
    // "dd" (8 - 9)
    for (idx = 8; idx <= 9 && U_SUCCESS(status); idx++) {
        n = ASCII_DIGIT((int32_t)text[idx]);
        if (n >= 0) {
            day = 10*day + n;
        } else {
            status = U_INVALID_FORMAT_ERROR;
        }
    }
    if (len == 16) {
        // "HH" (11 - 12)
        for (idx = 11; idx <= 12 && U_SUCCESS(status); idx++) {
            n = ASCII_DIGIT((int32_t)text[idx]);
            if (n >= 0) {
                hour = 10*hour + n;
            } else {
                status = U_INVALID_FORMAT_ERROR;
            }
        }
        // "mm" (14 - 15)
        for (idx = 14; idx <= 15 && U_SUCCESS(status); idx++) {
            n = ASCII_DIGIT((int32_t)text[idx]);
            if (n >= 0) {
                min = 10*min + n;
            } else {
                status = U_INVALID_FORMAT_ERROR;
            }
        }
    }

    if (U_SUCCESS(status)) {
        UDate date = Grego::fieldsToDay(year, month - 1, day) * U_MILLIS_PER_DAY
            + hour * U_MILLIS_PER_HOUR + min * U_MILLIS_PER_MINUTE;
        return date;
    }
    return 0;
}

static void U_CALLCONV initCanonicalIDCache(UErrorCode &status) {
    gCanonicalIDCache = uhash_open(uhash_hashUChars, uhash_compareUChars, NULL, &status);
    if (gCanonicalIDCache == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    if (U_FAILURE(status)) {
        gCanonicalIDCache = NULL;
    }
    // No key/value deleters - keys/values are from a resource bundle
    ucln_i18n_registerCleanup(UCLN_I18N_ZONEMETA, zoneMeta_cleanup);
}


const UChar* U_EXPORT2
ZoneMeta::getCanonicalCLDRID(const UnicodeString &tzid, UErrorCode& status) {
    if (U_FAILURE(status)) {
        return NULL;
    }

    if (tzid.isBogus() || tzid.length() > ZID_KEY_MAX) {
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    // Checking the cached results
    umtx_initOnce(gCanonicalIDCacheInitOnce, &initCanonicalIDCache, status);
    if (U_FAILURE(status)) {
        return NULL;
    }

    const UChar *canonicalID = NULL;

    UErrorCode tmpStatus = U_ZERO_ERROR;
    UChar utzid[ZID_KEY_MAX + 1];
    tzid.extract(utzid, ZID_KEY_MAX + 1, tmpStatus);
    U_ASSERT(tmpStatus == U_ZERO_ERROR);    // we checked the length of tzid already

    if (!uprv_isInvariantUString(utzid, -1)) {
        // All of known tz IDs are only containing ASCII invariant characters.
        status = U_ILLEGAL_ARGUMENT_ERROR;
        return NULL;
    }

    // Check if it was already cached
    umtx_lock(&gZoneMetaLock);
    {
        canonicalID = (const UChar *)uhash_get(gCanonicalIDCache, utzid);
    }
    umtx_unlock(&gZoneMetaLock);

    if (canonicalID != NULL) {
        return canonicalID;
    }

    // If not, resolve CLDR canonical ID with resource data
    UBool isInputCanonical = FALSE;
    char id[ZID_KEY_MAX + 1];
    tzid.extract(0, 0x7fffffff, id, UPRV_LENGTHOF(id), US_INV);

    // replace '/' with ':'
    char *p = id;
    while (*p++) {
        if (*p == '/') {
            *p = ':';
        }
    }

    UResourceBundle *top = ures_openDirect(NULL, gKeyTypeData, &tmpStatus);
    UResourceBundle *rb = ures_getByKey(top, gTypeMapTag, NULL, &tmpStatus);
    ures_getByKey(rb, gTimezoneTag, rb, &tmpStatus);
    ures_getByKey(rb, id, rb, &tmpStatus);
    if (U_SUCCESS(tmpStatus)) {
        // type entry (canonical) found
        // the input is the canonical ID. resolve to const UChar*
        canonicalID = TimeZone::findID(tzid);
        isInputCanonical = TRUE;
    }

    if (canonicalID == NULL) {
        // If a map element not found, then look for an alias
        tmpStatus = U_ZERO_ERROR;
        ures_getByKey(top, gTypeAliasTag, rb, &tmpStatus);
        ures_getByKey(rb, gTimezoneTag, rb, &tmpStatus);
        const UChar *canonical = ures_getStringByKey(rb,id,NULL,&tmpStatus);
        if (U_SUCCESS(tmpStatus)) {
            // canonical map found
            canonicalID = canonical;
        }

        if (canonicalID == NULL) {
            // Dereference the input ID using the tz data
            const UChar *derefer = TimeZone::dereferOlsonLink(tzid);
            if (derefer == NULL) {
                status = U_ILLEGAL_ARGUMENT_ERROR;
            } else {
                int32_t len = u_strlen(derefer);
                u_UCharsToChars(derefer,id,len);
                id[len] = (char) 0; // Make sure it is null terminated.

                // replace '/' with ':'
                char *q = id;
                while (*q++) {
                    if (*q == '/') {
                        *q = ':';
                    }
                }

                // If a dereference turned something up then look for an alias.
                // rb still points to the alias table, so we don't have to go looking
                // for it.
                tmpStatus = U_ZERO_ERROR;
                canonical = ures_getStringByKey(rb,id,NULL,&tmpStatus);
                if (U_SUCCESS(tmpStatus)) {
                    // canonical map for the dereferenced ID found
                    canonicalID = canonical;
                } else {
                    canonicalID = derefer;
                    isInputCanonical = TRUE;
                }
            }
        }
    }
    ures_close(rb);
    ures_close(top);

    if (U_SUCCESS(status)) {
        U_ASSERT(canonicalID != NULL);  // canocanilD must be non-NULL here

        // Put the resolved canonical ID to the cache
        umtx_lock(&gZoneMetaLock);
        {
            const UChar* idInCache = (const UChar *)uhash_get(gCanonicalIDCache, utzid);
            if (idInCache == NULL) {
                const UChar* key = ZoneMeta::findTimeZoneID(tzid);
                U_ASSERT(key != NULL);
                if (key != NULL) {
                    idInCache = (const UChar *)uhash_put(gCanonicalIDCache, (void *)key, (void *)canonicalID, &status);
                    U_ASSERT(idInCache == NULL);
                }
            }
            if (U_SUCCESS(status) && isInputCanonical) {
                // Also put canonical ID itself into the cache if not exist
                const UChar *canonicalInCache = (const UChar*)uhash_get(gCanonicalIDCache, canonicalID);
                if (canonicalInCache == NULL) {
                    canonicalInCache = (const UChar *)uhash_put(gCanonicalIDCache, (void *)canonicalID, (void *)canonicalID, &status);
                    U_ASSERT(canonicalInCache == NULL);
                }
            }
        }
        umtx_unlock(&gZoneMetaLock);
    }

    return canonicalID;
}

UnicodeString& U_EXPORT2
ZoneMeta::getCanonicalCLDRID(const UnicodeString &tzid, UnicodeString &systemID, UErrorCode& status) {
    const UChar *canonicalID = getCanonicalCLDRID(tzid, status);
    if (U_FAILURE(status) || canonicalID == NULL) {
        systemID.setToBogus();
        return systemID;
    }
    systemID.setTo(TRUE, canonicalID, -1);
    return systemID;
}

const UChar* U_EXPORT2
ZoneMeta::getCanonicalCLDRID(const TimeZone& tz) {
    if (dynamic_cast<const OlsonTimeZone *>(&tz) != NULL) {
        // short cut for OlsonTimeZone
        const OlsonTimeZone *otz = (const OlsonTimeZone*)&tz;
        return otz->getCanonicalID();
    }
    UErrorCode status = U_ZERO_ERROR;
    UnicodeString tzID;
    return getCanonicalCLDRID(tz.getID(tzID), status);
}

static void U_CALLCONV countryInfoVectorsInit(UErrorCode &status) {
    // Create empty vectors
    // No deleters for these UVectors, it's a reference to a resource bundle string.
    gSingleZoneCountries = new UVector(NULL, uhash_compareUChars, status);
    if (gSingleZoneCountries == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }
    gMultiZonesCountries = new UVector(NULL, uhash_compareUChars, status);
    if (gMultiZonesCountries == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
    }

    if (U_FAILURE(status)) {
        delete gSingleZoneCountries;
        delete gMultiZonesCountries;
        gSingleZoneCountries = NULL;
        gMultiZonesCountries  = NULL;
    }
    ucln_i18n_registerCleanup(UCLN_I18N_ZONEMETA, zoneMeta_cleanup);
}


UnicodeString& U_EXPORT2
ZoneMeta::getCanonicalCountry(const UnicodeString &tzid, UnicodeString &country, UBool *isPrimary /* = NULL */) {
    if (isPrimary != NULL) {
        *isPrimary = FALSE;
    }

    const UChar *region = TimeZone::getRegion(tzid);
    if (region != NULL && u_strcmp(gWorld, region) != 0) {
        country.setTo(region, -1);
    } else {
        country.setToBogus();
        return country;
    }

    if (isPrimary != NULL) {
        char regionBuf[] = {0, 0, 0};

        // Checking the cached results
        UErrorCode status = U_ZERO_ERROR;
        umtx_initOnce(gCountryInfoVectorsInitOnce, &countryInfoVectorsInit, status);
        if (U_FAILURE(status)) {
            return country;
        }

        // Check if it was already cached
        UBool cached = FALSE;
        UBool singleZone = FALSE;
        umtx_lock(&gZoneMetaLock);
        {
            singleZone = cached = gSingleZoneCountries->contains((void*)region);
            if (!cached) {
                cached = gMultiZonesCountries->contains((void*)region);
            }
        }
        umtx_unlock(&gZoneMetaLock);

        if (!cached) {
            // We need to go through all zones associated with the region.
            // This is relatively heavy operation.

            U_ASSERT(u_strlen(region) == 2);

            u_UCharsToChars(region, regionBuf, 2);

            StringEnumeration *ids = TimeZone::createTimeZoneIDEnumeration(UCAL_ZONE_TYPE_CANONICAL_LOCATION, regionBuf, NULL, status);
            int32_t idsLen = ids->count(status);
            if (U_SUCCESS(status) && idsLen == 1) {
                // only the single zone is available for the region
                singleZone = TRUE;
            }
            delete ids;

            // Cache the result
            umtx_lock(&gZoneMetaLock);
            {
                UErrorCode ec = U_ZERO_ERROR;
                if (singleZone) {
                    if (!gSingleZoneCountries->contains((void*)region)) {
                        gSingleZoneCountries->addElement((void*)region, ec);
                    }
                } else {
                    if (!gMultiZonesCountries->contains((void*)region)) {
                        gMultiZonesCountries->addElement((void*)region, ec);
                    }
                }
            }
            umtx_unlock(&gZoneMetaLock);
        }

        if (singleZone) {
            *isPrimary = TRUE;
        } else {
            // Note: We may cache the primary zone map in future.

            // Even a country has multiple zones, one of them might be
            // dominant and treated as a primary zone
            int32_t idLen = 0;
            if (regionBuf[0] == 0) {
                u_UCharsToChars(region, regionBuf, 2);
            }

            UResourceBundle *rb = ures_openDirect(NULL, gMetaZones, &status);
            ures_getByKey(rb, gPrimaryZonesTag, rb, &status);
            const UChar *primaryZone = ures_getStringByKey(rb, regionBuf, &idLen, &status);
            if (U_SUCCESS(status)) {
                if (tzid.compare(primaryZone, idLen) == 0) {
                    *isPrimary = TRUE;
                } else {
                    // The given ID might not be a canonical ID
                    UnicodeString canonicalID;
                    TimeZone::getCanonicalID(tzid, canonicalID, status);
                    if (U_SUCCESS(status) && canonicalID.compare(primaryZone, idLen) == 0) {
                        *isPrimary = TRUE;
                    }
                }
            }
            ures_close(rb);
        }
    }

    return country;
}

UnicodeString& U_EXPORT2
ZoneMeta::getMetazoneID(const UnicodeString &tzid, UDate date, UnicodeString &result) {
    UBool isSet = FALSE;
    const UVector *mappings = getMetazoneMappings(tzid);
    if (mappings != NULL) {
        for (int32_t i = 0; i < mappings->size(); i++) {
            OlsonToMetaMappingEntry *mzm = (OlsonToMetaMappingEntry*)mappings->elementAt(i);
            if (mzm->from <= date && mzm->to > date) {
                result.setTo(mzm->mzid, -1);
                isSet = TRUE;
                break;
            }
        }
    }
    if (!isSet) {
        result.setToBogus();
    }
    return result;
}

static void U_CALLCONV olsonToMetaInit(UErrorCode &status) {
    U_ASSERT(gOlsonToMeta == NULL);
    ucln_i18n_registerCleanup(UCLN_I18N_ZONEMETA, zoneMeta_cleanup);
    gOlsonToMeta = uhash_open(uhash_hashUChars, uhash_compareUChars, NULL, &status);
    if (U_FAILURE(status)) {
        gOlsonToMeta = NULL;
    } else {
        uhash_setKeyDeleter(gOlsonToMeta, deleteUCharString);
        uhash_setValueDeleter(gOlsonToMeta, deleteUVector);
    }
}


const UVector* U_EXPORT2
ZoneMeta::getMetazoneMappings(const UnicodeString &tzid) {
    UErrorCode status = U_ZERO_ERROR;
    UChar tzidUChars[ZID_KEY_MAX + 1];
    tzid.extract(tzidUChars, ZID_KEY_MAX + 1, status);
    if (U_FAILURE(status) || status == U_STRING_NOT_TERMINATED_WARNING) {
        return NULL;
    }

    umtx_initOnce(gOlsonToMetaInitOnce, &olsonToMetaInit, status);
    if (U_FAILURE(status)) {
        return NULL;
    }

    // get the mapping from cache
    const UVector *result = NULL;

    umtx_lock(&gZoneMetaLock);
    {
        result = (UVector*) uhash_get(gOlsonToMeta, tzidUChars);
    }
    umtx_unlock(&gZoneMetaLock);

    if (result != NULL) {
        return result;
    }

    // miss the cache - create new one
    UVector *tmpResult = createMetazoneMappings(tzid);
    if (tmpResult == NULL) {
        // not available
        return NULL;
    }

    // put the new one into the cache
    umtx_lock(&gZoneMetaLock);
    {
        // make sure it's already created
        result = (UVector*) uhash_get(gOlsonToMeta, tzidUChars);
        if (result == NULL) {
            // add the one just created
            int32_t tzidLen = tzid.length() + 1;
            UChar *key = (UChar*)uprv_malloc(tzidLen * sizeof(UChar));
            if (key == NULL) {
                // memory allocation error..  just return NULL
                result = NULL;
                delete tmpResult;
            } else {
                tzid.extract(key, tzidLen, status);
                uhash_put(gOlsonToMeta, key, tmpResult, &status);
                if (U_FAILURE(status)) {
                    // delete the mapping
                    result = NULL;
                    delete tmpResult;
                } else {
                    result = tmpResult;
                }
            }
        } else {
            // another thread already put the one
            delete tmpResult;
        }
    }
    umtx_unlock(&gZoneMetaLock);

    return result;
}

UVector*
ZoneMeta::createMetazoneMappings(const UnicodeString &tzid) {
    UVector *mzMappings = NULL;
    UErrorCode status = U_ZERO_ERROR;

    UnicodeString canonicalID;
    UResourceBundle *rb = ures_openDirect(NULL, gMetaZones, &status);
    ures_getByKey(rb, gMetazoneInfo, rb, &status);
    getCanonicalCLDRID(tzid, canonicalID, status);

    if (U_SUCCESS(status)) {
        char tzKey[ZID_KEY_MAX + 1];
        int32_t tzKeyLen = canonicalID.extract(0, canonicalID.length(), tzKey, sizeof(tzKey), US_INV);
        tzKey[tzKeyLen] = 0;

        // tzid keys are using ':' as separators
        char *p = tzKey;
        while (*p) {
            if (*p == '/') {
                *p = ':';
            }
            p++;
        }

        ures_getByKey(rb, tzKey, rb, &status);

        if (U_SUCCESS(status)) {
            UResourceBundle *mz = NULL;
            while (ures_hasNext(rb)) {
                mz = ures_getNextResource(rb, mz, &status);

                const UChar *mz_name = ures_getStringByIndex(mz, 0, NULL, &status);
                const UChar *mz_from = gDefaultFrom;
                const UChar *mz_to = gDefaultTo;

                if (ures_getSize(mz) == 3) {
                    mz_from = ures_getStringByIndex(mz, 1, NULL, &status);
                    mz_to   = ures_getStringByIndex(mz, 2, NULL, &status);
                }

                if(U_FAILURE(status)){
                    status = U_ZERO_ERROR;
                    continue;
                }
                // We do not want to use SimpleDateformat to parse boundary dates,
                // because this code could be triggered by the initialization code
                // used by SimpleDateFormat.
                UDate from = parseDate(mz_from, status);
                UDate to = parseDate(mz_to, status);
                if (U_FAILURE(status)) {
                    status = U_ZERO_ERROR;
                    continue;
                }

                OlsonToMetaMappingEntry *entry = (OlsonToMetaMappingEntry*)uprv_malloc(sizeof(OlsonToMetaMappingEntry));
                if (entry == NULL) {
                    status = U_MEMORY_ALLOCATION_ERROR;
                    break;
                }
                entry->mzid = mz_name;
                entry->from = from;
                entry->to = to;

                if (mzMappings == NULL) {
                    mzMappings = new UVector(deleteOlsonToMetaMappingEntry, NULL, status);
                    if (U_FAILURE(status)) {
                        delete mzMappings;
                        mzMappings = NULL;
                        uprv_free(entry);
                        break;
                    }
                }

                mzMappings->addElement(entry, status);
                if (U_FAILURE(status)) {
                    break;
                }
            }
            ures_close(mz);
            if (U_FAILURE(status)) {
                if (mzMappings != NULL) {
                    delete mzMappings;
                    mzMappings = NULL;
                }
            }
        }
    }
    ures_close(rb);
    return mzMappings;
}

UnicodeString& U_EXPORT2
ZoneMeta::getZoneIdByMetazone(const UnicodeString &mzid, const UnicodeString &region, UnicodeString &result) {
    UErrorCode status = U_ZERO_ERROR;
    const UChar *tzid = NULL;
    int32_t tzidLen = 0;
    char keyBuf[ZID_KEY_MAX + 1];
    int32_t keyLen = 0;

    if (mzid.isBogus() || mzid.length() > ZID_KEY_MAX) {
        result.setToBogus();
        return result;
    }

    keyLen = mzid.extract(0, mzid.length(), keyBuf, ZID_KEY_MAX + 1, US_INV);
    keyBuf[keyLen] = 0;

    UResourceBundle *rb = ures_openDirect(NULL, gMetaZones, &status);
    ures_getByKey(rb, gMapTimezonesTag, rb, &status);
    ures_getByKey(rb, keyBuf, rb, &status);

    if (U_SUCCESS(status)) {
        // check region mapping
        if (region.length() == 2 || region.length() == 3) {
            keyLen = region.extract(0, region.length(), keyBuf, ZID_KEY_MAX + 1, US_INV);
            keyBuf[keyLen] = 0;
            tzid = ures_getStringByKey(rb, keyBuf, &tzidLen, &status);
            if (status == U_MISSING_RESOURCE_ERROR) {
                status = U_ZERO_ERROR;
            }
        }
        if (U_SUCCESS(status) && tzid == NULL) {
            // try "001"
            tzid = ures_getStringByKey(rb, gWorldTag, &tzidLen, &status);
        }
    }
    ures_close(rb);

    if (tzid == NULL) {
        result.setToBogus();
    } else {
        result.setTo(tzid, tzidLen);
    }

    return result;
}

static void U_CALLCONV initAvailableMetaZoneIDs () {
    U_ASSERT(gMetaZoneIDs == NULL);
    U_ASSERT(gMetaZoneIDTable == NULL);
    ucln_i18n_registerCleanup(UCLN_I18N_ZONEMETA, zoneMeta_cleanup);

    UErrorCode status = U_ZERO_ERROR;
    gMetaZoneIDTable = uhash_open(uhash_hashUnicodeString, uhash_compareUnicodeString, NULL, &status);
    if (U_FAILURE(status) || gMetaZoneIDTable == NULL) {
        gMetaZoneIDTable = NULL;
        return;
    }
    uhash_setKeyDeleter(gMetaZoneIDTable, uprv_deleteUObject);
    // No valueDeleter, because the vector maintain the value objects
    gMetaZoneIDs = new UVector(NULL, uhash_compareUChars, status);
    if (U_FAILURE(status) || gMetaZoneIDs == NULL) {
        gMetaZoneIDs = NULL;
        uhash_close(gMetaZoneIDTable);
        gMetaZoneIDTable = NULL;
        return;
    }
    gMetaZoneIDs->setDeleter(uprv_free);

    UResourceBundle *rb = ures_openDirect(NULL, gMetaZones, &status);
    UResourceBundle *bundle = ures_getByKey(rb, gMapTimezonesTag, NULL, &status);
    StackUResourceBundle res;
    while (U_SUCCESS(status) && ures_hasNext(bundle)) {
        ures_getNextResource(bundle, res.getAlias(), &status);
        if (U_FAILURE(status)) {
            break;
        }
        const char *mzID = ures_getKey(res.getAlias());
        int32_t len = static_cast<int32_t>(uprv_strlen(mzID));
        UChar *uMzID = (UChar*)uprv_malloc(sizeof(UChar) * (len + 1));
        if (uMzID == NULL) {
            status = U_MEMORY_ALLOCATION_ERROR;
            break;
        }
        u_charsToUChars(mzID, uMzID, len);
        uMzID[len] = 0;
        UnicodeString *usMzID = new UnicodeString(uMzID);
        if (uhash_get(gMetaZoneIDTable, usMzID) == NULL) {
            gMetaZoneIDs->addElement((void *)uMzID, status);
            uhash_put(gMetaZoneIDTable, (void *)usMzID, (void *)uMzID, &status);
        } else {
            uprv_free(uMzID);
            delete usMzID;
        }
    }
    ures_close(bundle);
    ures_close(rb);

    if (U_FAILURE(status)) {
        uhash_close(gMetaZoneIDTable);
        delete gMetaZoneIDs;
        gMetaZoneIDTable = NULL;
        gMetaZoneIDs = NULL;
    }
}

const UVector*
ZoneMeta::getAvailableMetazoneIDs() {
    umtx_initOnce(gMetaZoneIDsInitOnce, &initAvailableMetaZoneIDs);
    return gMetaZoneIDs;
}

const UChar*
ZoneMeta::findMetaZoneID(const UnicodeString& mzid) {
    umtx_initOnce(gMetaZoneIDsInitOnce, &initAvailableMetaZoneIDs);
    if (gMetaZoneIDTable == NULL) {
        return NULL;
    }
    return (const UChar*)uhash_get(gMetaZoneIDTable, &mzid);
}

const UChar*
ZoneMeta::findTimeZoneID(const UnicodeString& tzid) {
    return TimeZone::findID(tzid);
}


TimeZone*
ZoneMeta::createCustomTimeZone(int32_t offset) {
    UBool negative = FALSE;
    int32_t tmp = offset;
    if (offset < 0) {
        negative = TRUE;
        tmp = -offset;
    }
    uint8_t hour, min, sec;

    tmp /= 1000;
    sec = static_cast<uint8_t>(tmp % 60);
    tmp /= 60;
    min = static_cast<uint8_t>(tmp % 60);
    hour = static_cast<uint8_t>(tmp / 60);

    UnicodeString zid;
    formatCustomID(hour, min, sec, negative, zid);
    return new SimpleTimeZone(offset, zid);
}

UnicodeString&
ZoneMeta::formatCustomID(uint8_t hour, uint8_t min, uint8_t sec, UBool negative, UnicodeString& id) {
    // Create normalized time zone ID - GMT[+|-]HH:mm[:ss]
    id.setTo(gCustomTzPrefix, -1);
    if (hour != 0 || min != 0) {
        if (negative) {
          id.append((UChar)0x2D);    // '-'
        } else {
          id.append((UChar)0x2B);    // '+'
        }
        // Always use US-ASCII digits
        id.append((UChar)(0x30 + (hour%100)/10));
        id.append((UChar)(0x30 + (hour%10)));
        id.append((UChar)0x3A);    // ':'
        id.append((UChar)(0x30 + (min%100)/10));
        id.append((UChar)(0x30 + (min%10)));
        if (sec != 0) {
          id.append((UChar)0x3A);    // ':'
          id.append((UChar)(0x30 + (sec%100)/10));
          id.append((UChar)(0x30 + (sec%10)));
        }
    }
    return id;
}

const UChar*
ZoneMeta::getShortID(const TimeZone& tz) {
    const UChar* canonicalID = NULL;
    if (dynamic_cast<const OlsonTimeZone *>(&tz) != NULL) {
        // short cut for OlsonTimeZone
        const OlsonTimeZone *otz = (const OlsonTimeZone*)&tz;
        canonicalID = otz->getCanonicalID();
    }
    if (canonicalID == NULL) {
        return NULL;
    }
    return getShortIDFromCanonical(canonicalID);
}

const UChar*
ZoneMeta::getShortID(const UnicodeString& id) {
    UErrorCode status = U_ZERO_ERROR;
    const UChar* canonicalID = ZoneMeta::getCanonicalCLDRID(id, status);
    if (U_FAILURE(status) || canonicalID == NULL) {
        return NULL;
    }
    return ZoneMeta::getShortIDFromCanonical(canonicalID);
}

const UChar*
ZoneMeta::getShortIDFromCanonical(const UChar* canonicalID) {
    const UChar* shortID = NULL;
    int32_t len = u_strlen(canonicalID);
    char tzidKey[ZID_KEY_MAX + 1];

    u_UCharsToChars(canonicalID, tzidKey, len);
    tzidKey[len] = (char) 0; // Make sure it is null terminated.

    // replace '/' with ':'
    char *p = tzidKey;
    while (*p++) {
        if (*p == '/') {
            *p = ':';
        }
    }

    UErrorCode status = U_ZERO_ERROR;
    UResourceBundle *rb = ures_openDirect(NULL, gKeyTypeData, &status);
    ures_getByKey(rb, gTypeMapTag, rb, &status);
    ures_getByKey(rb, gTimezoneTag, rb, &status);
    shortID = ures_getStringByKey(rb, tzidKey, NULL, &status);
    ures_close(rb);

    return shortID;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2011, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

/**
 * \file 
 * \brief C API: Time zone rule classes
 */



#if !UCONFIG_NO_FORMATTING


// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2016, International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*/

/**
 * \file 
 * \brief C API: Time zone rule classes
 */



#if !UCONFIG_NO_FORMATTING



#ifndef UCNV_H

/**
 * A TimeZoneRule.  Use the zrule_* API to manipulate.  Create with
 * zrule_open*, and destroy with zrule_close.
 */
struct ZRule;
typedef struct ZRule ZRule;

/**
 * An InitialTimeZoneRule.  Use the izrule_* API to manipulate.  Create with
 * izrule_open*, and destroy with izrule_close.
 */
struct IZRule;
typedef struct IZRule IZRule;

/**
 * An AnnualTimeZoneRule.  Use the azrule_* API to manipulate.  Create with
 * azrule_open*, and destroy with azrule_close.
 */
struct AZRule;
typedef struct AZRule AZRule;

#endif

/*********************************************************************
 * ZRule API
 *********************************************************************/

/**
 * Disposes of the storage used by a ZRule object.  This function should
 * be called exactly once for objects returned by zrule_open*.
 * @param set the object to dispose of
 */
U_CAPI void U_EXPORT2
zrule_close(ZRule* rule);

/**
 * Returns true if rule1 is identical to rule2
 * and vis versa.
 * @param rule1 to be checked for containment
 * @param rule2 to be checked for containment
 * @return true if the test condition is met
 */
U_CAPI UBool U_EXPORT2
zrule_equals(const ZRule* rule1, const ZRule* rule2);

/**
 * Fills in "name" with the name of this time zone.
 * @param rule, the Zrule to use
 * @param name  Receives the name of this time zone.
 * @param nameLength, length of the returned name
 */
U_CAPI void U_EXPORT2
zrule_getName(ZRule* rule, UChar* name, int32_t nameLength);

/**
 * Gets the standard time offset.
 * @param rule, the Zrule to use
 * @return  The standard time offset from UTC in milliseconds.
 */
U_CAPI int32_t U_EXPORT2
zrule_getRawOffset(ZRule* rule);

/**
 * Gets the amount of daylight saving delta time from the standard time.
 * @param rule, the Zrule to use
 * @return  The amount of daylight saving offset used by this rule
 *          in milliseconds.
 */
U_CAPI int32_t U_EXPORT2
zrule_getDSTSavings(ZRule* rule);

/**
 * Returns if this rule represents the same rule and offsets as another.
 * When two ZRule objects differ only its names, this method
 * returns true.
 * @param rule1 to be checked for containment
 * @param rule2 to be checked for containment
 * @return  true if the other <code>TimeZoneRule</code> is the same as this one.
 */
U_CAPI UBool U_EXPORT2
zrule_isEquivalentTo(ZRule* rule1,  ZRule* rule2);

/*********************************************************************
 * IZRule API
 *********************************************************************/

/**
 * Constructs an IZRule with the name, the GMT offset of its
 * standard time and the amount of daylight saving offset adjustment.
 * @param name          The time zone name.
 * @param nameLength    The length of the time zone name.
 * @param rawOffset     The UTC offset of its standard time in milliseconds.
 * @param dstSavings    The amount of daylight saving offset adjustment in milliseconds.
 *                      If this ia a rule for standard time, the value of this argument is 0.
 */
U_CAPI IZRule* U_EXPORT2
izrule_open(const UChar* name, int32_t nameLength, int32_t rawOffset, int32_t dstSavings);

/**
 * Disposes of the storage used by a IZRule object.  This function should
 * be called exactly once for objects returned by izrule_open*.
 * @param set the object to dispose of
 */
U_CAPI void U_EXPORT2
izrule_close(IZRule* rule);

/**
 * Returns a copy of this object.
 * @param rule the original IZRule
 * @return the newly allocated copy of the IZRule
 */
U_CAPI IZRule* U_EXPORT2
izrule_clone(IZRule *rule);

/**
 * Returns true if rule1 is identical to rule2
 * and vis versa.
 * @param rule1 to be checked for containment
 * @param rule2 to be checked for containment
 * @return true if the test condition is met
 */
U_CAPI UBool U_EXPORT2
izrule_equals(const IZRule* rule1, const IZRule* rule2);

/**
 * Fills in "name" with the name of this time zone.
 * @param rule, the IZrule to use
 * @param name  Receives the name of this time zone.
 * @param nameLength, length of the returned name
 */
U_CAPI void U_EXPORT2
izrule_getName(IZRule* rule, UChar* & name, int32_t & nameLength);

/**
 * Gets the standard time offset.
 * @param rule, the IZrule to use
 * @return  The standard time offset from UTC in milliseconds.
 */
U_CAPI int32_t U_EXPORT2
izrule_getRawOffset(IZRule* rule);

/**
 * Gets the amount of daylight saving delta time from the standard time.
 * @param rule, the IZrule to use
 * @return  The amount of daylight saving offset used by this rule
 *          in milliseconds.
 */
U_CAPI int32_t U_EXPORT2
izrule_getDSTSavings(IZRule* rule);

/**
 * Returns if this rule represents the same rule and offsets as another.
 * When two IZRule objects differ only its names, this method
 * returns true.
 * @param rule1 to be checked for containment
 * @param rule2 to be checked for containment
 * @return  true if the other <code>TimeZoneRule</code> is the same as this one.
 */
U_CAPI UBool U_EXPORT2
izrule_isEquivalentTo(IZRule* rule1,  IZRule* rule2);

/**
 * Gets the very first time when this rule takes effect.
 * @param rule              The IZrule to use
 * @param prevRawOffset     The standard time offset from UTC before this rule
 *                          takes effect in milliseconds.
 * @param prevDSTSavings    The amount of daylight saving offset from the
 *                          standard time.
 * @param result            Receives the very first time when this rule takes effect.
 * @return  true if the start time is available.  When false is returned, output parameter
 *          "result" is unchanged.
 */
U_CAPI UBool U_EXPORT2
izrule_getFirstStart(IZRule* rule, int32_t prevRawOffset, int32_t prevDSTSavings, 
                    UDate& result);

/**
 * Gets the final time when this rule takes effect.
 * @param rule              The IZrule to use     
 * @param prevRawOffset     The standard time offset from UTC before this rule
 *                          takes effect in milliseconds.
 * @param prevDSTSavings    The amount of daylight saving offset from the
 *                          standard time.
 * @param result            Receives the final time when this rule takes effect.
 * @return  true if the start time is available.  When false is returned, output parameter
 *          "result" is unchanged.
 */
U_CAPI UBool U_EXPORT2
izrule_getFinalStart(IZRule* rule, int32_t prevRawOffset, int32_t prevDSTSavings, 
                    UDate& result);

/**
 * Gets the first time when this rule takes effect after the specified time.
 * @param rule              The IZrule to use
 * @param base              The first start time after this base time will be returned.
 * @param prevRawOffset     The standard time offset from UTC before this rule
 *                          takes effect in milliseconds.
 * @param prevDSTSavings    The amount of daylight saving offset from the
 *                          standard time.
 * @param inclusive         Whether the base time is inclusive or not.
 * @param result            Receives The first time when this rule takes effect after
 *                          the specified base time.
 * @return  true if the start time is available.  When false is returned, output parameter
 *          "result" is unchanged.
 */
U_CAPI UBool U_EXPORT2
izrule_getNextStart(IZRule* rule, UDate base, int32_t prevRawOffset, 
                   int32_t prevDSTSavings, UBool inclusive, UDate& result);

/**
 * Gets the most recent time when this rule takes effect before the specified time.
 * @param rule              The IZrule to use
 * @param base              The most recent time before this base time will be returned.
 * @param prevRawOffset     The standard time offset from UTC before this rule
 *                          takes effect in milliseconds.
 * @param prevDSTSavings    The amount of daylight saving offset from the
 *                          standard time.
 * @param inclusive         Whether the base time is inclusive or not.
 * @param result            Receives The most recent time when this rule takes effect before
 *                          the specified base time.
 * @return  true if the start time is available.  When false is returned, output parameter
 *          "result" is unchanged.
 */
U_CAPI UBool U_EXPORT2
izrule_getPreviousStart(IZRule* rule, UDate base, int32_t prevRawOffset, 
                       int32_t prevDSTSavings, UBool inclusive, UDate& result);


/**
 * Return the class ID for this class. This is useful only for comparing to
 * a return value from getDynamicClassID(). For example:
 * <pre>
 * .   Base* polymorphic_pointer = createPolymorphicObject();
 * .   if (polymorphic_pointer->getDynamicClassID() ==
 * .       erived::getStaticClassID()) ...
 * </pre>
 * @param rule              The IZrule to use
 * @return          The class ID for all objects of this class.
 */
U_CAPI UClassID U_EXPORT2
izrule_getStaticClassID(IZRule* rule);

/**
 * Returns a unique class ID POLYMORPHICALLY. Pure virtual override. This
 * method is to implement a simple version of RTTI, since not all C++
 * compilers support genuine RTTI. Polymorphic operator==() and clone()
 * methods call this method.
 *
 * @param rule              The IZrule to use
 * @return          The class ID for this object. All objects of a
 *                  given class have the same class ID.  Objects of
 *                  other classes have different class IDs.
 */
U_CAPI UClassID U_EXPORT2
izrule_getDynamicClassID(IZRule* rule);

#endif







U_NAMESPACE_USE

/*********************************************************************
 * ZRule API
 *********************************************************************/

U_CAPI void U_EXPORT2
zrule_close(ZRule* rule) {
    delete (TimeZoneRule*)rule;
}

U_CAPI UBool U_EXPORT2
zrule_equals(const ZRule* rule1, const ZRule* rule2) {
    return *(const TimeZoneRule*)rule1 == *(const TimeZoneRule*)rule2;
}

U_CAPI void U_EXPORT2
zrule_getName(ZRule* rule, UChar* name, int32_t nameLength) {
    UnicodeString s(nameLength==-1, name, nameLength);
    s = ((TimeZoneRule*)rule)->TimeZoneRule::getName(s);
    nameLength = s.length();
    memcpy(name, s.getBuffer(), nameLength);
    return;
}

U_CAPI int32_t U_EXPORT2
zrule_getRawOffset(ZRule* rule) {
    return ((TimeZoneRule*)rule)->TimeZoneRule::getRawOffset();
}

U_CAPI int32_t U_EXPORT2
zrule_getDSTSavings(ZRule* rule) {
    return ((TimeZoneRule*)rule)->TimeZoneRule::getDSTSavings();
}

U_CAPI UBool U_EXPORT2
zrule_isEquivalentTo(ZRule* rule1,  ZRule* rule2) {
    return ((TimeZoneRule*)rule1)->TimeZoneRule::isEquivalentTo(*(TimeZoneRule*)rule2);
}

/*********************************************************************
 * IZRule API
 *********************************************************************/

U_CAPI IZRule* U_EXPORT2
izrule_open(const UChar* name, int32_t nameLength, int32_t rawOffset, int32_t dstSavings) {
    UnicodeString s(nameLength==-1, name, nameLength);
    return (IZRule*) new InitialTimeZoneRule(s, rawOffset, dstSavings);
}

U_CAPI void U_EXPORT2
izrule_close(IZRule* rule) {
    delete (InitialTimeZoneRule*)rule;
}

U_CAPI IZRule* U_EXPORT2
izrule_clone(IZRule *rule) {
    return (IZRule*) (((InitialTimeZoneRule*)rule)->InitialTimeZoneRule::clone());
}

U_CAPI UBool U_EXPORT2
izrule_equals(const IZRule* rule1, const IZRule* rule2) {
    return *(const InitialTimeZoneRule*)rule1 == *(const InitialTimeZoneRule*)rule2;
}

U_CAPI void U_EXPORT2
izrule_getName(IZRule* rule, UChar* & name, int32_t & nameLength) {
    // UnicodeString s(nameLength==-1, name, nameLength);
    UnicodeString s;
    ((InitialTimeZoneRule*)rule)->InitialTimeZoneRule::getName(s);
    nameLength = s.length();
    name = (UChar*)uprv_malloc(nameLength);
    memcpy(name, s.getBuffer(), nameLength);
    return;
}

U_CAPI int32_t U_EXPORT2
izrule_getRawOffset(IZRule* rule) {
    return ((InitialTimeZoneRule*)rule)->InitialTimeZoneRule::getRawOffset();
}

U_CAPI int32_t U_EXPORT2
izrule_getDSTSavings(IZRule* rule) {
    return ((InitialTimeZoneRule*)rule)->InitialTimeZoneRule::getDSTSavings();
}

U_CAPI UBool U_EXPORT2
izrule_isEquivalentTo(IZRule* rule1,  IZRule* rule2) {
    return ((InitialTimeZoneRule*)rule1)->InitialTimeZoneRule::isEquivalentTo(*(InitialTimeZoneRule*)rule2);
}

U_CAPI UBool U_EXPORT2
izrule_getFirstStart(IZRule* rule, int32_t prevRawOffset, int32_t prevDSTSavings, 
                    UDate& result) {
    return ((const InitialTimeZoneRule*)rule)->InitialTimeZoneRule::getFirstStart(prevRawOffset, prevDSTSavings, result);
}

U_CAPI UBool U_EXPORT2
izrule_getFinalStart(IZRule* rule, int32_t prevRawOffset, int32_t prevDSTSavings, 
                    UDate& result) {
    return ((InitialTimeZoneRule*)rule)->InitialTimeZoneRule::getFinalStart(prevRawOffset, prevDSTSavings, result);
}

U_CAPI UBool U_EXPORT2
izrule_getNextStart(IZRule* rule, UDate base, int32_t prevRawOffset, 
                   int32_t prevDSTSavings, UBool inclusive, UDate& result) {
    return ((InitialTimeZoneRule*)rule)->InitialTimeZoneRule::getNextStart(base, prevRawOffset, prevDSTSavings, inclusive, result);
}

U_CAPI UBool U_EXPORT2
izrule_getPreviousStart(IZRule* rule, UDate base, int32_t prevRawOffset, 
                       int32_t prevDSTSavings, UBool inclusive, UDate& result) {
    return ((InitialTimeZoneRule*)rule)->InitialTimeZoneRule::getPreviousStart(base, prevRawOffset, prevDSTSavings, inclusive, result);
}

U_CAPI UClassID U_EXPORT2
izrule_getStaticClassID(IZRule* rule) {
    return ((InitialTimeZoneRule*)rule)->InitialTimeZoneRule::getStaticClassID();
}

U_CAPI UClassID U_EXPORT2
izrule_getDynamicClassID(IZRule* rule) {
    return ((InitialTimeZoneRule*)rule)->InitialTimeZoneRule::getDynamicClassID();
}

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2009-2010, International Business Machines Corporation and         *
* others. All Rights Reserved.                                                *
*******************************************************************************
*/

/**
 * \file 
 * \brief C API: Time zone transition classes
 */



#if !UCONFIG_NO_FORMATTING








U_NAMESPACE_USE

U_CAPI ZTrans* U_EXPORT2
ztrans_open(UDate time, const void* from, const void* to){
    return (ZTrans*) new TimeZoneTransition(time,*(TimeZoneRule*)from,*(TimeZoneRule*)to);
}

U_CAPI ZTrans* U_EXPORT2
ztrans_openEmpty() {
    return (ZTrans*) new TimeZoneTransition();
}

U_CAPI void U_EXPORT2
ztrans_close(ZTrans *trans) {
    delete (TimeZoneTransition*)trans;
}

U_CAPI ZTrans* U_EXPORT2
ztrans_clone(ZTrans *trans) {
    return (ZTrans*) (((TimeZoneTransition*)trans)->TimeZoneTransition::clone());
}

U_CAPI UBool U_EXPORT2
ztrans_equals(const ZTrans* trans1, const ZTrans* trans2){
    return *(const TimeZoneTransition*)trans1 == *(const TimeZoneTransition*)trans2;
}

U_CAPI UDate U_EXPORT2
ztrans_getTime(ZTrans* trans) {
    return ((TimeZoneTransition*)trans)->TimeZoneTransition::getTime();
}

U_CAPI void U_EXPORT2
ztrans_setTime(ZTrans* trans, UDate time) {
    return ((TimeZoneTransition*)trans)->TimeZoneTransition::setTime(time);
}

U_CAPI void* U_EXPORT2
ztrans_getFrom(ZTrans* & trans) {
    return (void*) (((TimeZoneTransition*)trans)->TimeZoneTransition::getFrom());
}

U_CAPI void U_EXPORT2
ztrans_setFrom(ZTrans* trans, const void* from) {
    return ((TimeZoneTransition*)trans)->TimeZoneTransition::setFrom(*(TimeZoneRule*)from);
}

U_CAPI void U_EXPORT2
ztrans_adoptFrom(ZTrans* trans, void* from) {
    return ((TimeZoneTransition*)trans)->TimeZoneTransition::adoptFrom((TimeZoneRule*)from);
}

U_CAPI void* U_EXPORT2
ztrans_getTo(ZTrans* trans){
    return (void*) (((TimeZoneTransition*)trans)->TimeZoneTransition::getTo());
}

U_CAPI void U_EXPORT2
ztrans_setTo(ZTrans* trans, const void* to) {
    return ((TimeZoneTransition*)trans)->TimeZoneTransition::setTo(*(TimeZoneRule*)to);
}

U_CAPI void U_EXPORT2
ztrans_adoptTo(ZTrans* trans, void* to) {
    return ((TimeZoneTransition*)trans)->TimeZoneTransition::adoptTo((TimeZoneRule*)to);
}

U_CAPI UClassID U_EXPORT2
ztrans_getStaticClassID(ZTrans* trans) {
    return ((TimeZoneTransition*)trans)->TimeZoneTransition::getStaticClassID();
}

U_CAPI UClassID U_EXPORT2
ztrans_getDynamicClassID(ZTrans* trans){
    return ((TimeZoneTransition*)trans)->TimeZoneTransition::getDynamicClassID();
}

#endif

// This file is generated by scripts/inline-data.py, do not edit it manually //





extern "C" U_EXPORT const unsigned char U_ICUDATA_ENTRY_POINT [] = {
	144, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 67, 109, 110, 68, 1, 0, 0, 0, 3, 0, 0, 0, 32, 67, 111, 112, 121, 114, 105, 103, 104, 116, 32, 40, 67, 41, 32, 50, 48, 49, 54, 32, 97, 110, 100, 32, 108, 97, 116, 101, 114, 58, 32, 85, 110, 105, 99, 111, 100, 101, 44, 32, 73, 110, 99, 46, 32, 97, 110, 100, 32, 111, 116, 104, 101, 114, 115, 46, 32, 76, 105, 99, 101, 110, 115, 101, 32, 38, 32, 116, 101, 114, 109, 115, 32, 111, 102, 32, 117, 115, 101, 58, 32, 104, 116, 116, 112, 58, 47, 47, 119, 119, 119, 46, 117, 110, 105, 99, 111, 100, 101, 46, 111, 114, 103, 47, 99, 111, 112, 121, 114, 105, 103, 104, 116, 46, 104, 116, 109, 108, 32, 0, 180, 0, 0, 0, 164, 5, 0, 0, 176, 21, 0, 0, 185, 5, 0, 0, 0, 51, 0, 0, 206, 5, 0, 0, 32, 53, 0, 0, 227, 5, 0, 0, 16, 168, 0, 0, 249, 5, 0, 0, 112, 168, 0, 0, 14, 6, 0, 0, 128, 196, 0, 0, 35, 6, 0, 0, 0, 29, 1, 0, 56, 6, 0, 0, 240, 56, 1, 0, 77, 6, 0, 0, 16, 59, 1, 0, 98, 6, 0, 0, 128, 190, 1, 0, 119, 6, 0, 0, 224, 190, 1, 0, 140, 6, 0, 0, 48, 13, 2, 0, 166, 6, 0, 0, 192, 39, 2, 0, 187, 6, 0, 0, 176, 76, 2, 0, 209, 6, 0, 0, 240, 108, 2, 0, 231, 6, 0, 0, 16, 111, 2, 0, 252, 6, 0, 0, 128, 147, 2, 0, 17, 7, 0, 0, 32, 188, 2, 0, 38, 7, 0, 0, 128, 42, 3, 0, 59, 7, 0, 0, 0, 125, 3, 0, 81, 7, 0, 0, 80, 125, 3, 0, 105, 7, 0, 0, 48, 163, 3, 0, 137, 7, 0, 0, 176, 163, 3, 0, 159, 7, 0, 0, 96, 205, 3, 0, 180, 7, 0, 0, 192, 205, 3, 0, 201, 7, 0, 0, 32, 251, 3, 0, 222, 7, 0, 0, 96, 253, 3, 0, 243, 7, 0, 0, 192, 253, 3, 0, 11, 8, 0, 0, 32, 254, 3, 0, 41, 8, 0, 0, 176, 60, 4, 0, 62, 8, 0, 0, 0, 98, 4, 0, 83, 8, 0, 0, 80, 198, 4, 0, 105, 8, 0, 0, 160, 198, 4, 0, 139, 8, 0, 0, 32, 199, 4, 0, 160, 8, 0, 0, 128, 240, 4, 0, 181, 8, 0, 0, 160, 15, 5, 0, 205, 8, 0, 0, 208, 49, 5, 0, 226, 8, 0, 0, 0, 187, 5, 0, 248, 8, 0, 0, 64, 219, 5, 0, 13, 9, 0, 0, 176, 73, 6, 0, 34, 9, 0, 0, 16, 74, 6, 0, 58, 9, 0, 0, 0, 75, 6, 0, 79, 9, 0, 0, 96, 75, 6, 0, 100, 9, 0, 0, 32, 142, 6, 0, 121, 9, 0, 0, 16, 169, 6, 0, 142, 9, 0, 0, 240, 206, 6, 0, 164, 9, 0, 0, 32, 251, 6, 0, 185, 9, 0, 0, 16, 57, 7, 0, 209, 9, 0, 0, 96, 57, 7, 0, 230, 9, 0, 0, 64, 84, 7, 0, 251, 9, 0, 0, 224, 163, 7, 0, 17, 10, 0, 0, 144, 203, 7, 0, 38, 10, 0, 0, 96, 21, 8, 0, 59, 10, 0, 0, 16, 49, 8, 0, 80, 10, 0, 0, 112, 49, 8, 0, 104, 10, 0, 0, 192, 49, 8, 0, 125, 10, 0, 0, 0, 99, 8, 0, 146, 10, 0, 0, 96, 99, 8, 0, 170, 10, 0, 0, 192, 99, 8, 0, 191, 10, 0, 0, 176, 195, 8, 0, 212, 10, 0, 0, 16, 196, 8, 0, 233, 10, 0, 0, 112, 196, 8, 0, 1, 11, 0, 0, 208, 196, 8, 0, 22, 11, 0, 0, 224, 0, 11, 0, 43, 11, 0, 0, 0, 3, 11, 0, 64, 11, 0, 0, 80, 31, 11, 0, 85, 11, 0, 0, 32, 123, 11, 0, 106, 11, 0, 0, 240, 158, 11, 0, 127, 11, 0, 0, 0, 37, 12, 0, 148, 11, 0, 0, 176, 144, 16, 0, 170, 11, 0, 0, 48, 172, 16, 0, 191, 11, 0, 0, 32, 214, 16, 0, 212, 11, 0, 0, 208, 241, 16, 0, 233, 11, 0, 0, 48, 242, 16, 0, 255, 11, 0, 0, 160, 21, 17, 0, 20, 12, 0, 0, 208, 110, 17, 0, 41, 12, 0, 0, 240, 112, 17, 0, 62, 12, 0, 0, 16, 153, 17, 0, 83, 12, 0, 0, 0, 190, 17, 0, 104, 12, 0, 0, 64, 218, 17, 0, 125, 12, 0, 0, 32, 248, 17, 0, 146, 12, 0, 0, 80, 250, 17, 0, 167, 12, 0, 0, 176, 250, 17, 0, 188, 12, 0, 0, 96, 22, 18, 0, 209, 12, 0, 0, 192, 22, 18, 0, 230, 12, 0, 0, 16, 55, 18, 0, 251, 12, 0, 0, 144, 208, 18, 0, 16, 13, 0, 0, 48, 63, 19, 0, 40, 13, 0, 0, 128, 63, 19, 0, 61, 13, 0, 0, 160, 65, 19, 0, 82, 13, 0, 0, 0, 66, 19, 0, 103, 13, 0, 0, 144, 175, 19, 0, 124, 13, 0, 0, 240, 175, 19, 0, 148, 13, 0, 0, 80, 176, 19, 0, 169, 13, 0, 0, 176, 216, 19, 0, 190, 13, 0, 0, 32, 244, 19, 0, 211, 13, 0, 0, 144, 16, 20, 0, 237, 13, 0, 0, 224, 16, 20, 0, 10, 14, 0, 0, 48, 17, 20, 0, 34, 14, 0, 0, 160, 17, 20, 0, 55, 14, 0, 0, 192, 57, 20, 0, 76, 14, 0, 0, 112, 93, 20, 0, 97, 14, 0, 0, 208, 93, 20, 0, 125, 14, 0, 0, 96, 99, 20, 0, 146, 14, 0, 0, 160, 134, 20, 0, 169, 14, 0, 0, 64, 255, 24, 0, 190, 14, 0, 0, 96, 1, 25, 0, 211, 14, 0, 0, 64, 115, 25, 0, 232, 14, 0, 0, 160, 115, 25, 0, 0, 15, 0, 0, 16, 116, 25, 0, 24, 15, 0, 0, 128, 116, 25, 0, 48, 15, 0, 0, 240, 116, 25, 0, 69, 15, 0, 0, 0, 171, 25, 0, 90, 15, 0, 0, 48, 15, 26, 0, 111, 15, 0, 0, 32, 49, 26, 0, 133, 15, 0, 0, 240, 144, 26, 0, 154, 15, 0, 0, 112, 191, 26, 0, 175, 15, 0, 0, 64, 218, 26, 0, 199, 15, 0, 0, 176, 218, 26, 0, 225, 15, 0, 0, 0, 219, 26, 0, 254, 15, 0, 0, 80, 219, 26, 0, 27, 16, 0, 0, 160, 219, 26, 0, 56, 16, 0, 0, 240, 219, 26, 0, 82, 16, 0, 0, 64, 42, 27, 0, 111, 16, 0, 0, 144, 42, 27, 0, 140, 16, 0, 0, 224, 42, 27, 0, 164, 16, 0, 0, 80, 43, 27, 0, 188, 16, 0, 0, 192, 43, 27, 0, 209, 16, 0, 0, 16, 183, 27, 0, 230, 16, 0, 0, 112, 183, 27, 0, 251, 16, 0, 0, 16, 215, 27, 0, 16, 17, 0, 0, 240, 241, 27, 0, 37, 17, 0, 0, 80, 13, 28, 0, 58, 17, 0, 0, 176, 54, 28, 0, 79, 17, 0, 0, 64, 102, 28, 0, 100, 17, 0, 0, 192, 185, 28, 0, 126, 17, 0, 0, 224, 15, 37, 0, 147, 17, 0, 0, 32, 45, 37, 0, 168, 17, 0, 0, 160, 73, 37, 0, 189, 17, 0, 0, 176, 106, 37, 0, 210, 17, 0, 0, 144, 140, 37, 0, 231, 17, 0, 0, 128, 29, 38, 0, 253, 17, 0, 0, 224, 29, 38, 0, 18, 18, 0, 0, 64, 69, 38, 0, 39, 18, 0, 0, 160, 69, 38, 0, 60, 18, 0, 0, 0, 135, 38, 0, 81, 18, 0, 0, 208, 170, 38, 0, 103, 18, 0, 0, 48, 171, 38, 0, 128, 18, 0, 0, 144, 171, 38, 0, 155, 18, 0, 0, 240, 171, 38, 0, 176, 18, 0, 0, 80, 175, 52, 0, 200, 18, 0, 0, 192, 175, 52, 0, 224, 18, 0, 0, 48, 176, 52, 0, 250, 18, 0, 0, 128, 176, 52, 0, 23, 19, 0, 0, 208, 176, 52, 0, 52, 19, 0, 0, 32, 177, 52, 0, 78, 19, 0, 0, 176, 177, 52, 0, 107, 19, 0, 0, 0, 178, 52, 0, 136, 19, 0, 0, 80, 178, 52, 0, 165, 19, 0, 0, 160, 178, 52, 0, 189, 19, 0, 0, 16, 179, 52, 0, 213, 19, 0, 0, 128, 179, 52, 0, 237, 19, 0, 0, 240, 179, 52, 0, 2, 20, 0, 0, 80, 180, 52, 0, 36, 20, 0, 0, 96, 192, 52, 0, 60, 20, 0, 0, 80, 10, 53, 0, 84, 20, 0, 0, 112, 12, 53, 0, 104, 20, 0, 0, 208, 12, 53, 0, 124, 20, 0, 0, 80, 13, 53, 0, 149, 20, 0, 0, 64, 40, 53, 0, 171, 20, 0, 0, 128, 191, 53, 0, 198, 20, 0, 0, 240, 145, 54, 0, 221, 20, 0, 0, 0, 52, 55, 0, 243, 20, 0, 0, 176, 186, 55, 0, 17, 21, 0, 0, 96, 209, 55, 0, 43, 21, 0, 0, 128, 222, 55, 0, 64, 21, 0, 0, 112, 96, 56, 0, 94, 21, 0, 0, 160, 255, 59, 0, 121, 21, 0, 0, 0, 78, 60, 0, 147, 21, 0, 0, 0, 166, 60, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 97, 102, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 97, 109, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 97, 114, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 97, 114, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 97, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 97, 122, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 98, 101, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 98, 103, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 98, 110, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 98, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 98, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 98, 115, 95, 67, 121, 114, 108, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 99, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 99, 101, 98, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 99, 104, 114, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 99, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 99, 121, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 100, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 100, 101, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 100, 101, 95, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 100, 101, 95, 65, 84, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 100, 101, 95, 95, 80, 72, 79, 78, 69, 66, 79, 79, 75, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 100, 115, 98, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 100, 122, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 101, 101, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 101, 108, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 101, 110, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 101, 110, 95, 85, 83, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 101, 110, 95, 85, 83, 95, 80, 79, 83, 73, 88, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 101, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 101, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 101, 115, 95, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 101, 115, 95, 95, 84, 82, 65, 68, 73, 84, 73, 79, 78, 65, 76, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 101, 116, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 102, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 102, 97, 95, 65, 70, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 102, 105, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 102, 105, 108, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 102, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 102, 114, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 102, 114, 95, 67, 65, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 103, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 103, 108, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 103, 117, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 104, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 104, 97, 119, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 104, 101, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 104, 101, 95, 73, 76, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 104, 105, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 104, 114, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 104, 115, 98, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 104, 117, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 104, 121, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 105, 100, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 105, 100, 95, 73, 68, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 105, 103, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 105, 110, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 105, 110, 95, 73, 68, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 105, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 105, 116, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 105, 119, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 105, 119, 95, 73, 76, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 106, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 107, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 107, 107, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 107, 108, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 107, 109, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 107, 110, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 107, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 107, 111, 107, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 107, 117, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 107, 121, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 108, 98, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 108, 107, 116, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 108, 110, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 108, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 108, 116, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 108, 118, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 109, 107, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 109, 108, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 109, 110, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 109, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 109, 114, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 109, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 109, 116, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 109, 121, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 110, 98, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 110, 98, 95, 78, 79, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 110, 101, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 110, 108, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 110, 110, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 110, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 110, 111, 95, 78, 79, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 111, 109, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 111, 114, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 112, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 112, 97, 95, 71, 117, 114, 117, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 112, 97, 95, 71, 117, 114, 117, 95, 73, 78, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 112, 97, 95, 73, 78, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 112, 108, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 112, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 112, 116, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 114, 101, 115, 95, 105, 110, 100, 101, 120, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 114, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 114, 111, 111, 116, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 114, 117, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 101, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 104, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 104, 95, 66, 65, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 104, 95, 67, 83, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 104, 95, 89, 85, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 105, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 107, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 108, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 109, 110, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 113, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 95, 66, 65, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 95, 67, 121, 114, 108, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 95, 67, 121, 114, 108, 95, 66, 65, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 95, 67, 121, 114, 108, 95, 77, 69, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 95, 67, 121, 114, 108, 95, 82, 83, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 95, 76, 97, 116, 110, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 95, 76, 97, 116, 110, 95, 66, 65, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 95, 76, 97, 116, 110, 95, 82, 83, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 95, 77, 69, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 114, 95, 82, 83, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 118, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 115, 119, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 116, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 116, 101, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 116, 104, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 116, 107, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 116, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 116, 114, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 117, 99, 97, 100, 97, 116, 97, 46, 105, 99, 117, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 117, 103, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 117, 107, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 117, 114, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 117, 122, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 118, 105, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 119, 97, 101, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 119, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 120, 104, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 121, 105, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 121, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 121, 117, 101, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 121, 117, 101, 95, 67, 78, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 121, 117, 101, 95, 72, 97, 110, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 67, 78, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 72, 75, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 72, 97, 110, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 72, 97, 110, 115, 95, 67, 78, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 72, 97, 110, 115, 95, 83, 71, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 72, 97, 110, 116, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 72, 97, 110, 116, 95, 72, 75, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 72, 97, 110, 116, 95, 77, 79, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 72, 97, 110, 116, 95, 84, 87, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 77, 79, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 83, 71, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 104, 95, 84, 87, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 111, 108, 108, 47, 122, 117, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 99, 117, 114, 114, 101, 110, 99, 121, 78, 117, 109, 101, 114, 105, 99, 67, 111, 100, 101, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 100, 97, 121, 80, 101, 114, 105, 111, 100, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 103, 101, 110, 100, 101, 114, 76, 105, 115, 116, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 105, 99, 117, 115, 116, 100, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 105, 99, 117, 118, 101, 114, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 107, 101, 121, 84, 121, 112, 101, 68, 97, 116, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 108, 97, 110, 103, 73, 110, 102, 111, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 108, 105, 107, 101, 108, 121, 83, 117, 98, 116, 97, 103, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 109, 101, 116, 97, 90, 111, 110, 101, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 109, 101, 116, 97, 100, 97, 116, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 110, 117, 109, 98, 101, 114, 105, 110, 103, 83, 121, 115, 116, 101, 109, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 112, 108, 117, 114, 97, 108, 82, 97, 110, 103, 101, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 112, 108, 117, 114, 97, 108, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 115, 117, 112, 112, 108, 101, 109, 101, 110, 116, 97, 108, 68, 97, 116, 97, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 116, 105, 109, 101, 122, 111, 110, 101, 84, 121, 112, 101, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 119, 105, 110, 100, 111, 119, 115, 90, 111, 110, 101, 115, 46, 114, 101, 115, 0, 105, 99, 117, 100, 116, 54, 54, 108, 47, 122, 111, 110, 101, 105, 110, 102, 111, 54, 52, 46, 114, 101, 115, 0, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 70, 7, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 74, 7, 0, 0, 74, 7, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 28, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 38, 0, 78, 0, 60, 0, 60, 0, 60, 0, 73, 1, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 124, 28, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 100, 24, 0, 0, 100, 24, 0, 0, 100, 24, 0, 0, 100, 24, 0, 0, 168, 24, 0, 0, 168, 24, 0, 0, 168, 24, 0, 0, 168, 24, 0, 0, 100, 28, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 121, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 104, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 29, 5, 0, 68, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 22, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 185, 112, 185, 120, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 185, 152, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 168, 120, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 168, 152, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 153, 114, 185, 121, 153, 121, 249, 121, 153, 122, 89, 122, 136, 14, 217, 122, 185, 153, 153, 153, 249, 153, 89, 154, 153, 169, 185, 176, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 136, 121, 232, 121, 136, 122, 72, 122, 128, 14, 200, 122, 168, 153, 136, 153, 232, 153, 72, 154, 136, 169, 168, 176, 72, 170, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 153, 113, 136, 113, 249, 114, 232, 114, 25, 114, 8, 114, 170, 112, 185, 116, 168, 116, 57, 123, 40, 123, 217, 121, 200, 121, 121, 122, 104, 122, 10, 8, 12, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 250, 142, 232, 142, 26, 142, 8, 142, 249, 146, 232, 146, 25, 146, 8, 146, 185, 148, 168, 148, 153, 154, 136, 154, 57, 155, 40, 155, 217, 153, 200, 153, 57, 154, 40, 154, 121, 154, 104, 154, 25, 155, 8, 155, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 185, 174, 168, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 14, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 16, 8, 1, 0, 8, 14, 18, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 20, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 31, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 63, 7, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 68, 7, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 124, 0, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 128, 0, 0, 0, 128, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 69, 0, 116, 0, 104, 0, 105, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 68, 1, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 8, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 32, 0, 0, 0, 44, 0, 0, 0, 44, 1, 0, 0, 11, 0, 0, 0, 0, 0, 143, 103, 192, 255, 0, 104, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 0, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 3, 0, 69, 0, 60, 0, 32, 0, 35, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 117, 0, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 122, 0, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 174, 28, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 178, 28, 0, 0, 178, 28, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 49, 8, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 99, 111, 109, 112, 97, 116, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 56, 220, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 65, 0, 114, 0, 97, 0, 98, 0, 93, 0, 38, 0, 42, 6, 60, 0, 60, 0, 41, 6, 60, 0, 60, 0, 60, 0, 148, 254, 60, 0, 60, 0, 60, 0, 147, 254, 38, 0, 74, 6, 60, 0, 60, 0, 73, 6, 60, 0, 60, 0, 60, 0, 232, 251, 60, 0, 60, 0, 60, 0, 233, 251, 60, 0, 60, 0, 60, 0, 240, 254, 60, 0, 60, 0, 60, 0, 239, 254, 60, 0, 60, 0, 60, 0, 144, 252, 60, 0, 60, 0, 60, 0, 93, 252, 0, 0, 239, 223, 241, 15, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 65, 0, 114, 0, 97, 0, 98, 0, 93, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 42, 6, 60, 0, 60, 0, 14, 32, 41, 6, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 204, 6, 60, 0, 60, 0, 14, 32, 73, 6, 38, 0, 14, 32, 33, 6, 60, 0, 60, 0, 14, 32, 35, 6, 60, 0, 60, 0, 14, 32, 36, 6, 60, 0, 60, 0, 14, 32, 37, 6, 60, 0, 60, 0, 14, 32, 38, 6, 60, 0, 60, 0, 14, 32, 73, 6, 85, 6, 60, 0, 60, 0, 14, 32, 39, 6, 14, 32, 38, 0, 33, 6, 39, 6, 14, 32, 60, 0, 60, 0, 14, 32, 34, 6, 14, 32, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 115, 0, 101, 0, 99, 0, 111, 0, 110, 0, 100, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 60, 0, 60, 0, 60, 0, 75, 6, 60, 0, 60, 0, 60, 0, 240, 8, 60, 0, 60, 0, 60, 0, 231, 8, 60, 0, 60, 0, 60, 0, 76, 6, 60, 0, 60, 0, 60, 0, 241, 8, 60, 0, 60, 0, 60, 0, 232, 8, 60, 0, 60, 0, 60, 0, 77, 6, 60, 0, 60, 0, 60, 0, 242, 8, 60, 0, 60, 0, 60, 0, 233, 8, 60, 0, 60, 0, 60, 0, 78, 6, 60, 0, 60, 0, 60, 0, 228, 8, 60, 0, 60, 0, 60, 0, 244, 8, 60, 0, 60, 0, 60, 0, 245, 8, 60, 0, 60, 0, 60, 0, 79, 6, 60, 0, 60, 0, 60, 0, 229, 8, 60, 0, 60, 0, 60, 0, 254, 8, 60, 0, 60, 0, 60, 0, 80, 6, 60, 0, 60, 0, 60, 0, 230, 8, 60, 0, 60, 0, 60, 0, 246, 8, 60, 0, 60, 0, 60, 0, 81, 6, 60, 0, 60, 0, 60, 0, 82, 6, 60, 0, 60, 0, 60, 0, 83, 6, 60, 0, 60, 0, 60, 0, 223, 6, 60, 0, 60, 0, 60, 0, 225, 6, 60, 0, 60, 0, 60, 0, 95, 6, 60, 0, 60, 0, 60, 0, 86, 6, 60, 0, 60, 0, 60, 0, 87, 6, 60, 0, 60, 0, 60, 0, 88, 6, 60, 0, 60, 0, 60, 0, 255, 8, 60, 0, 60, 0, 60, 0, 89, 6, 60, 0, 60, 0, 60, 0, 90, 6, 60, 0, 60, 0, 60, 0, 91, 6, 60, 0, 60, 0, 60, 0, 92, 6, 60, 0, 60, 0, 60, 0, 93, 6, 60, 0, 60, 0, 60, 0, 94, 6, 60, 0, 60, 0, 60, 0, 247, 8, 60, 0, 60, 0, 60, 0, 248, 8, 60, 0, 60, 0, 60, 0, 253, 8, 60, 0, 60, 0, 60, 0, 251, 8, 60, 0, 60, 0, 60, 0, 252, 8, 60, 0, 60, 0, 60, 0, 249, 8, 60, 0, 60, 0, 60, 0, 250, 8, 60, 0, 60, 0, 60, 0, 112, 6, 38, 0, 14, 32, 39, 0, 40, 6, 51, 6, 69, 6, 32, 0, 39, 6, 68, 6, 68, 6, 71, 6, 32, 0, 39, 6, 68, 6, 49, 6, 45, 6, 69, 6, 70, 6, 32, 0, 39, 6, 68, 6, 49, 6, 45, 6, 74, 6, 69, 6, 39, 0, 60, 0, 60, 0, 60, 0, 14, 32, 253, 253, 14, 32, 38, 0, 33, 6, 14, 32, 61, 0, 128, 254, 14, 32, 38, 0, 39, 6, 14, 32, 61, 0, 141, 254, 14, 32, 61, 0, 142, 254, 14, 32, 38, 0, 39, 6, 67, 6, 40, 6, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 243, 253, 14, 32, 38, 0, 39, 6, 68, 6, 68, 6, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 242, 253, 14, 32, 38, 0, 39, 6, 75, 6, 14, 32, 61, 0, 60, 253, 14, 32, 61, 0, 61, 253, 14, 32, 38, 0, 34, 6, 14, 32, 61, 0, 129, 254, 14, 32, 61, 0, 130, 254, 14, 32, 38, 0, 35, 6, 14, 32, 61, 0, 131, 254, 14, 32, 61, 0, 132, 254, 14, 32, 38, 0, 37, 6, 14, 32, 61, 0, 135, 254, 14, 32, 61, 0, 136, 254, 14, 32, 38, 0, 40, 6, 14, 32, 61, 0, 143, 254, 14, 32, 61, 0, 144, 254, 14, 32, 61, 0, 145, 254, 14, 32, 61, 0, 146, 254, 14, 32, 38, 0, 40, 6, 44, 6, 14, 32, 61, 0, 5, 252, 14, 32, 61, 0, 156, 252, 14, 32, 38, 0, 40, 6, 45, 6, 14, 32, 61, 0, 6, 252, 14, 32, 61, 0, 157, 252, 14, 32, 38, 0, 40, 6, 45, 6, 74, 6, 14, 32, 61, 0, 194, 253, 14, 32, 38, 0, 40, 6, 46, 6, 14, 32, 61, 0, 7, 252, 14, 32, 61, 0, 158, 252, 14, 32, 38, 0, 40, 6, 46, 6, 74, 6, 14, 32, 61, 0, 158, 253, 14, 32, 38, 0, 40, 6, 49, 6, 14, 32, 61, 0, 106, 252, 14, 32, 38, 0, 40, 6, 50, 6, 14, 32, 61, 0, 107, 252, 14, 32, 38, 0, 40, 6, 69, 6, 14, 32, 61, 0, 8, 252, 14, 32, 61, 0, 108, 252, 14, 32, 61, 0, 159, 252, 14, 32, 61, 0, 225, 252, 14, 32, 38, 0, 40, 6, 70, 6, 14, 32, 61, 0, 109, 252, 14, 32, 38, 0, 40, 6, 71, 6, 14, 32, 61, 0, 160, 252, 14, 32, 61, 0, 226, 252, 14, 32, 38, 0, 40, 6, 73, 6, 14, 32, 61, 0, 9, 252, 14, 32, 61, 0, 110, 252, 14, 32, 38, 0, 40, 6, 74, 6, 14, 32, 61, 0, 10, 252, 14, 32, 61, 0, 111, 252, 14, 32, 38, 0, 41, 6, 14, 32, 61, 0, 147, 254, 14, 32, 61, 0, 148, 254, 14, 32, 38, 0, 42, 6, 14, 32, 61, 0, 149, 254, 14, 32, 61, 0, 150, 254, 14, 32, 61, 0, 151, 254, 14, 32, 61, 0, 152, 254, 14, 32, 38, 0, 42, 6, 44, 6, 14, 32, 61, 0, 11, 252, 14, 32, 61, 0, 161, 252, 14, 32, 38, 0, 42, 6, 44, 6, 69, 6, 14, 32, 61, 0, 80, 253, 14, 32, 38, 0, 42, 6, 44, 6, 73, 6, 14, 32, 61, 0, 160, 253, 14, 32, 38, 0, 42, 6, 44, 6, 74, 6, 14, 32, 61, 0, 159, 253, 14, 32, 38, 0, 42, 6, 45, 6, 14, 32, 61, 0, 12, 252, 14, 32, 61, 0, 162, 252, 14, 32, 38, 0, 42, 6, 45, 6, 44, 6, 14, 32, 61, 0, 81, 253, 14, 32, 61, 0, 82, 253, 14, 32, 38, 0, 42, 6, 45, 6, 69, 6, 14, 32, 61, 0, 83, 253, 14, 32, 38, 0, 42, 6, 46, 6, 14, 32, 61, 0, 13, 252, 14, 32, 61, 0, 163, 252, 14, 32, 38, 0, 42, 6, 46, 6, 69, 6, 14, 32, 61, 0, 84, 253, 14, 32, 38, 0, 42, 6, 46, 6, 73, 6, 14, 32, 61, 0, 162, 253, 14, 32, 38, 0, 42, 6, 46, 6, 74, 6, 14, 32, 61, 0, 161, 253, 14, 32, 38, 0, 42, 6, 49, 6, 14, 32, 61, 0, 112, 252, 14, 32, 38, 0, 42, 6, 50, 6, 14, 32, 61, 0, 113, 252, 14, 32, 38, 0, 42, 6, 69, 6, 14, 32, 61, 0, 14, 252, 14, 32, 61, 0, 114, 252, 14, 32, 61, 0, 164, 252, 14, 32, 61, 0, 227, 252, 14, 32, 38, 0, 42, 6, 69, 6, 44, 6, 14, 32, 61, 0, 85, 253, 14, 32, 38, 0, 42, 6, 69, 6, 45, 6, 14, 32, 61, 0, 86, 253, 14, 32, 38, 0, 42, 6, 69, 6, 46, 6, 14, 32, 61, 0, 87, 253, 14, 32, 38, 0, 42, 6, 69, 6, 73, 6, 14, 32, 61, 0, 164, 253, 14, 32, 38, 0, 42, 6, 69, 6, 74, 6, 14, 32, 61, 0, 163, 253, 14, 32, 38, 0, 42, 6, 70, 6, 14, 32, 61, 0, 115, 252, 14, 32, 38, 0, 42, 6, 71, 6, 14, 32, 61, 0, 165, 252, 14, 32, 61, 0, 228, 252, 14, 32, 38, 0, 42, 6, 73, 6, 14, 32, 61, 0, 15, 252, 14, 32, 61, 0, 116, 252, 14, 32, 38, 0, 42, 6, 74, 6, 14, 32, 61, 0, 16, 252, 14, 32, 61, 0, 117, 252, 14, 32, 38, 0, 43, 6, 14, 32, 61, 0, 153, 254, 14, 32, 61, 0, 154, 254, 14, 32, 61, 0, 155, 254, 14, 32, 61, 0, 156, 254, 14, 32, 38, 0, 43, 6, 44, 6, 14, 32, 61, 0, 17, 252, 14, 32, 38, 0, 43, 6, 49, 6, 14, 32, 61, 0, 118, 252, 14, 32, 38, 0, 43, 6, 50, 6, 14, 32, 61, 0, 119, 252, 14, 32, 38, 0, 43, 6, 69, 6, 14, 32, 61, 0, 18, 252, 14, 32, 61, 0, 120, 252, 14, 32, 61, 0, 166, 252, 14, 32, 61, 0, 229, 252, 14, 32, 38, 0, 43, 6, 70, 6, 14, 32, 61, 0, 121, 252, 14, 32, 38, 0, 43, 6, 71, 6, 14, 32, 61, 0, 230, 252, 14, 32, 38, 0, 43, 6, 73, 6, 14, 32, 61, 0, 19, 252, 14, 32, 61, 0, 122, 252, 14, 32, 38, 0, 43, 6, 74, 6, 14, 32, 61, 0, 20, 252, 14, 32, 61, 0, 123, 252, 14, 32, 38, 0, 44, 6, 14, 32, 61, 0, 157, 254, 14, 32, 61, 0, 158, 254, 14, 32, 61, 0, 159, 254, 14, 32, 61, 0, 160, 254, 14, 32, 38, 0, 44, 6, 45, 6, 14, 32, 61, 0, 21, 252, 14, 32, 61, 0, 167, 252, 14, 32, 38, 0, 44, 6, 45, 6, 73, 6, 14, 32, 61, 0, 166, 253, 14, 32, 38, 0, 44, 6, 45, 6, 74, 6, 14, 32, 61, 0, 190, 253, 14, 32, 38, 0, 39, 0, 44, 6, 68, 6, 32, 0, 44, 6, 68, 6, 39, 6, 68, 6, 71, 6, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 251, 253, 14, 32, 38, 0, 44, 6, 69, 6, 14, 32, 61, 0, 22, 252, 14, 32, 61, 0, 168, 252, 14, 32, 38, 0, 44, 6, 69, 6, 45, 6, 14, 32, 61, 0, 88, 253, 14, 32, 61, 0, 89, 253, 14, 32, 38, 0, 44, 6, 69, 6, 73, 6, 14, 32, 61, 0, 167, 253, 14, 32, 38, 0, 44, 6, 69, 6, 74, 6, 14, 32, 61, 0, 165, 253, 14, 32, 38, 0, 44, 6, 73, 6, 14, 32, 61, 0, 1, 253, 14, 32, 61, 0, 29, 253, 14, 32, 38, 0, 44, 6, 74, 6, 14, 32, 61, 0, 2, 253, 14, 32, 61, 0, 30, 253, 14, 32, 38, 0, 45, 6, 14, 32, 61, 0, 161, 254, 14, 32, 61, 0, 162, 254, 14, 32, 61, 0, 163, 254, 14, 32, 61, 0, 164, 254, 14, 32, 38, 0, 45, 6, 44, 6, 14, 32, 61, 0, 23, 252, 14, 32, 61, 0, 169, 252, 14, 32, 38, 0, 45, 6, 44, 6, 74, 6, 14, 32, 61, 0, 191, 253, 14, 32, 38, 0, 45, 6, 69, 6, 14, 32, 61, 0, 24, 252, 14, 32, 61, 0, 170, 252, 14, 32, 38, 0, 45, 6, 69, 6, 73, 6, 14, 32, 61, 0, 91, 253, 14, 32, 38, 0, 45, 6, 69, 6, 74, 6, 14, 32, 61, 0, 90, 253, 14, 32, 38, 0, 45, 6, 73, 6, 14, 32, 61, 0, 255, 252, 14, 32, 61, 0, 27, 253, 14, 32, 38, 0, 45, 6, 74, 6, 14, 32, 61, 0, 0, 253, 14, 32, 61, 0, 28, 253, 14, 32, 38, 0, 46, 6, 14, 32, 61, 0, 165, 254, 14, 32, 61, 0, 166, 254, 14, 32, 61, 0, 167, 254, 14, 32, 61, 0, 168, 254, 14, 32, 38, 0, 46, 6, 44, 6, 14, 32, 61, 0, 25, 252, 14, 32, 61, 0, 171, 252, 14, 32, 38, 0, 46, 6, 45, 6, 14, 32, 61, 0, 26, 252, 14, 32, 38, 0, 46, 6, 69, 6, 14, 32, 61, 0, 27, 252, 14, 32, 61, 0, 172, 252, 14, 32, 38, 0, 46, 6, 73, 6, 14, 32, 61, 0, 3, 253, 14, 32, 61, 0, 31, 253, 14, 32, 38, 0, 46, 6, 74, 6, 14, 32, 61, 0, 4, 253, 14, 32, 61, 0, 32, 253, 14, 32, 38, 0, 47, 6, 14, 32, 61, 0, 169, 254, 14, 32, 61, 0, 170, 254, 14, 32, 38, 0, 48, 6, 14, 32, 61, 0, 171, 254, 14, 32, 61, 0, 172, 254, 14, 32, 38, 0, 48, 6, 112, 6, 14, 32, 61, 0, 91, 252, 14, 32, 38, 0, 49, 6, 14, 32, 61, 0, 173, 254, 14, 32, 61, 0, 174, 254, 14, 32, 38, 0, 49, 6, 51, 6, 72, 6, 68, 6, 14, 32, 60, 0, 60, 0, 60, 0, 246, 253, 14, 32, 38, 0, 49, 6, 112, 6, 14, 32, 61, 0, 92, 252, 14, 32, 38, 0, 49, 6, 204, 6, 39, 6, 68, 6, 14, 32, 60, 0, 60, 0, 60, 0, 252, 253, 14, 32, 38, 0, 50, 6, 14, 32, 61, 0, 175, 254, 14, 32, 61, 0, 176, 254, 14, 32, 38, 0, 51, 6, 14, 32, 61, 0, 177, 254, 14, 32, 61, 0, 178, 254, 14, 32, 61, 0, 179, 254, 14, 32, 61, 0, 180, 254, 14, 32, 38, 0, 51, 6, 44, 6, 14, 32, 61, 0, 28, 252, 14, 32, 61, 0, 173, 252, 14, 32, 61, 0, 52, 253, 14, 32, 38, 0, 51, 6, 44, 6, 45, 6, 14, 32, 61, 0, 93, 253, 14, 32, 38, 0, 51, 6, 44, 6, 73, 6, 14, 32, 61, 0, 94, 253, 14, 32, 38, 0, 51, 6, 45, 6, 14, 32, 61, 0, 29, 252, 14, 32, 61, 0, 174, 252, 14, 32, 61, 0, 53, 253, 14, 32, 38, 0, 51, 6, 45, 6, 44, 6, 14, 32, 61, 0, 92, 253, 14, 32, 38, 0, 51, 6, 46, 6, 14, 32, 61, 0, 30, 252, 14, 32, 61, 0, 175, 252, 14, 32, 61, 0, 54, 253, 14, 32, 38, 0, 51, 6, 46, 6, 73, 6, 14, 32, 61, 0, 168, 253, 14, 32, 38, 0, 51, 6, 46, 6, 74, 6, 14, 32, 61, 0, 198, 253, 14, 32, 38, 0, 51, 6, 49, 6, 14, 32, 61, 0, 14, 253, 14, 32, 61, 0, 42, 253, 14, 32, 38, 0, 51, 6, 69, 6, 14, 32, 61, 0, 31, 252, 14, 32, 61, 0, 176, 252, 14, 32, 61, 0, 231, 252, 14, 32, 38, 0, 51, 6, 69, 6, 44, 6, 14, 32, 61, 0, 97, 253, 14, 32, 38, 0, 51, 6, 69, 6, 45, 6, 14, 32, 61, 0, 95, 253, 14, 32, 61, 0, 96, 253, 14, 32, 38, 0, 51, 6, 69, 6, 69, 6, 14, 32, 61, 0, 98, 253, 14, 32, 61, 0, 99, 253, 14, 32, 38, 0, 51, 6, 71, 6, 14, 32, 61, 0, 232, 252, 14, 32, 61, 0, 49, 253, 14, 32, 38, 0, 51, 6, 73, 6, 14, 32, 61, 0, 251, 252, 14, 32, 61, 0, 23, 253, 14, 32, 38, 0, 51, 6, 74, 6, 14, 32, 61, 0, 252, 252, 14, 32, 61, 0, 24, 253, 14, 32, 38, 0, 52, 6, 14, 32, 61, 0, 181, 254, 14, 32, 61, 0, 182, 254, 14, 32, 61, 0, 183, 254, 14, 32, 61, 0, 184, 254, 14, 32, 38, 0, 52, 6, 44, 6, 14, 32, 61, 0, 9, 253, 14, 32, 61, 0, 37, 253, 14, 32, 61, 0, 45, 253, 14, 32, 61, 0, 55, 253, 14, 32, 38, 0, 52, 6, 44, 6, 74, 6, 14, 32, 61, 0, 105, 253, 14, 32, 38, 0, 52, 6, 45, 6, 14, 32, 61, 0, 10, 253, 14, 32, 61, 0, 38, 253, 14, 32, 61, 0, 46, 253, 14, 32, 61, 0, 56, 253, 14, 32, 38, 0, 52, 6, 45, 6, 69, 6, 14, 32, 61, 0, 103, 253, 14, 32, 61, 0, 104, 253, 14, 32, 38, 0, 52, 6, 45, 6, 74, 6, 14, 32, 61, 0, 170, 253, 14, 32, 38, 0, 52, 6, 46, 6, 14, 32, 61, 0, 11, 253, 14, 32, 61, 0, 39, 253, 14, 32, 61, 0, 47, 253, 14, 32, 61, 0, 57, 253, 14, 32, 38, 0, 52, 6, 49, 6, 14, 32, 61, 0, 13, 253, 14, 32, 61, 0, 41, 253, 14, 32, 38, 0, 52, 6, 69, 6, 14, 32, 61, 0, 233, 252, 14, 32, 61, 0, 12, 253, 14, 32, 61, 0, 40, 253, 14, 32, 61, 0, 48, 253, 14, 32, 38, 0, 52, 6, 69, 6, 46, 6, 14, 32, 61, 0, 106, 253, 14, 32, 61, 0, 107, 253, 14, 32, 38, 0, 52, 6, 69, 6, 69, 6, 14, 32, 61, 0, 108, 253, 14, 32, 61, 0, 109, 253, 14, 32, 38, 0, 52, 6, 71, 6, 14, 32, 61, 0, 234, 252, 14, 32, 61, 0, 50, 253, 14, 32, 38, 0, 52, 6, 73, 6, 14, 32, 61, 0, 253, 252, 14, 32, 61, 0, 25, 253, 14, 32, 38, 0, 52, 6, 74, 6, 14, 32, 61, 0, 254, 252, 14, 32, 61, 0, 26, 253, 14, 32, 38, 0, 53, 6, 14, 32, 61, 0, 185, 254, 14, 32, 61, 0, 186, 254, 14, 32, 61, 0, 187, 254, 14, 32, 61, 0, 188, 254, 14, 32, 38, 0, 53, 6, 45, 6, 14, 32, 61, 0, 32, 252, 14, 32, 61, 0, 177, 252, 14, 32, 38, 0, 53, 6, 45, 6, 45, 6, 14, 32, 61, 0, 100, 253, 14, 32, 61, 0, 101, 253, 14, 32, 38, 0, 53, 6, 45, 6, 74, 6, 14, 32, 61, 0, 169, 253, 14, 32, 38, 0, 53, 6, 46, 6, 14, 32, 61, 0, 178, 252, 14, 32, 38, 0, 53, 6, 49, 6, 14, 32, 61, 0, 15, 253, 14, 32, 61, 0, 43, 253, 14, 32, 38, 0, 53, 6, 68, 6, 57, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 245, 253, 14, 32, 38, 0, 53, 6, 68, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 249, 253, 14, 32, 38, 0, 39, 0, 53, 6, 68, 6, 73, 6, 32, 0, 39, 6, 68, 6, 68, 6, 71, 6, 32, 0, 57, 6, 68, 6, 74, 6, 71, 6, 32, 0, 72, 6, 51, 6, 68, 6, 69, 6, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 250, 253, 14, 32, 38, 0, 53, 6, 68, 6, 210, 6, 14, 32, 60, 0, 60, 0, 60, 0, 240, 253, 14, 32, 38, 0, 53, 6, 69, 6, 14, 32, 61, 0, 33, 252, 14, 32, 61, 0, 179, 252, 14, 32, 38, 0, 53, 6, 69, 6, 69, 6, 14, 32, 61, 0, 102, 253, 14, 32, 61, 0, 197, 253, 14, 32, 38, 0, 53, 6, 73, 6, 14, 32, 61, 0, 5, 253, 14, 32, 61, 0, 33, 253, 14, 32, 38, 0, 53, 6, 74, 6, 14, 32, 61, 0, 6, 253, 14, 32, 61, 0, 34, 253, 14, 32, 38, 0, 54, 6, 14, 32, 61, 0, 189, 254, 14, 32, 61, 0, 190, 254, 14, 32, 61, 0, 191, 254, 14, 32, 61, 0, 192, 254, 14, 32, 38, 0, 54, 6, 44, 6, 14, 32, 61, 0, 34, 252, 14, 32, 61, 0, 180, 252, 14, 32, 38, 0, 54, 6, 45, 6, 14, 32, 61, 0, 35, 252, 14, 32, 61, 0, 181, 252, 14, 32, 38, 0, 54, 6, 45, 6, 73, 6, 14, 32, 61, 0, 110, 253, 14, 32, 38, 0, 54, 6, 45, 6, 74, 6, 14, 32, 61, 0, 171, 253, 14, 32, 38, 0, 54, 6, 46, 6, 14, 32, 61, 0, 36, 252, 14, 32, 61, 0, 182, 252, 14, 32, 38, 0, 54, 6, 46, 6, 69, 6, 14, 32, 61, 0, 111, 253, 14, 32, 61, 0, 112, 253, 14, 32, 38, 0, 54, 6, 49, 6, 14, 32, 61, 0, 16, 253, 14, 32, 61, 0, 44, 253, 14, 32, 38, 0, 54, 6, 69, 6, 14, 32, 61, 0, 37, 252, 14, 32, 61, 0, 183, 252, 14, 32, 38, 0, 54, 6, 73, 6, 14, 32, 61, 0, 7, 253, 14, 32, 61, 0, 35, 253, 14, 32, 38, 0, 54, 6, 74, 6, 14, 32, 61, 0, 8, 253, 14, 32, 61, 0, 36, 253, 14, 32, 38, 0, 55, 6, 14, 32, 61, 0, 193, 254, 14, 32, 61, 0, 194, 254, 14, 32, 61, 0, 195, 254, 14, 32, 61, 0, 196, 254, 14, 32, 38, 0, 55, 6, 45, 6, 14, 32, 61, 0, 38, 252, 14, 32, 61, 0, 184, 252, 14, 32, 38, 0, 55, 6, 69, 6, 14, 32, 61, 0, 39, 252, 14, 32, 61, 0, 51, 253, 14, 32, 61, 0, 58, 253, 14, 32, 38, 0, 55, 6, 69, 6, 45, 6, 14, 32, 61, 0, 113, 253, 14, 32, 61, 0, 114, 253, 14, 32, 38, 0, 55, 6, 69, 6, 69, 6, 14, 32, 61, 0, 115, 253, 14, 32, 38, 0, 55, 6, 69, 6, 74, 6, 14, 32, 61, 0, 116, 253, 14, 32, 38, 0, 55, 6, 73, 6, 14, 32, 61, 0, 245, 252, 14, 32, 61, 0, 17, 253, 14, 32, 38, 0, 55, 6, 74, 6, 14, 32, 61, 0, 246, 252, 14, 32, 61, 0, 18, 253, 14, 32, 38, 0, 56, 6, 14, 32, 61, 0, 197, 254, 14, 32, 61, 0, 198, 254, 14, 32, 61, 0, 199, 254, 14, 32, 61, 0, 200, 254, 14, 32, 38, 0, 56, 6, 69, 6, 14, 32, 61, 0, 40, 252, 14, 32, 61, 0, 185, 252, 14, 32, 61, 0, 59, 253, 14, 32, 38, 0, 57, 6, 14, 32, 61, 0, 201, 254, 14, 32, 61, 0, 202, 254, 14, 32, 61, 0, 203, 254, 14, 32, 61, 0, 204, 254, 14, 32, 38, 0, 57, 6, 44, 6, 14, 32, 61, 0, 41, 252, 14, 32, 61, 0, 186, 252, 14, 32, 38, 0, 57, 6, 44, 6, 69, 6, 14, 32, 61, 0, 117, 253, 14, 32, 61, 0, 196, 253, 14, 32, 38, 0, 57, 6, 68, 6, 74, 6, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 247, 253, 14, 32, 38, 0, 57, 6, 69, 6, 14, 32, 61, 0, 42, 252, 14, 32, 61, 0, 187, 252, 14, 32, 38, 0, 57, 6, 69, 6, 69, 6, 14, 32, 61, 0, 118, 253, 14, 32, 61, 0, 119, 253, 14, 32, 38, 0, 57, 6, 69, 6, 73, 6, 14, 32, 61, 0, 120, 253, 14, 32, 38, 0, 57, 6, 69, 6, 74, 6, 14, 32, 61, 0, 182, 253, 14, 32, 38, 0, 57, 6, 73, 6, 14, 32, 61, 0, 247, 252, 14, 32, 61, 0, 19, 253, 14, 32, 38, 0, 57, 6, 74, 6, 14, 32, 61, 0, 248, 252, 14, 32, 61, 0, 20, 253, 14, 32, 38, 0, 58, 6, 14, 32, 61, 0, 205, 254, 14, 32, 61, 0, 206, 254, 14, 32, 61, 0, 207, 254, 14, 32, 61, 0, 208, 254, 14, 32, 38, 0, 58, 6, 44, 6, 14, 32, 61, 0, 43, 252, 14, 32, 61, 0, 188, 252, 14, 32, 38, 0, 58, 6, 69, 6, 14, 32, 61, 0, 44, 252, 14, 32, 61, 0, 189, 252, 14, 32, 38, 0, 58, 6, 69, 6, 69, 6, 14, 32, 61, 0, 121, 253, 14, 32, 38, 0, 58, 6, 69, 6, 73, 6, 14, 32, 61, 0, 123, 253, 14, 32, 38, 0, 58, 6, 69, 6, 74, 6, 14, 32, 61, 0, 122, 253, 14, 32, 38, 0, 58, 6, 73, 6, 14, 32, 61, 0, 249, 252, 14, 32, 61, 0, 21, 253, 14, 32, 38, 0, 58, 6, 74, 6, 14, 32, 61, 0, 250, 252, 14, 32, 61, 0, 22, 253, 14, 32, 38, 0, 64, 6, 75, 6, 14, 32, 61, 0, 113, 254, 14, 32, 38, 0, 64, 6, 78, 6, 14, 32, 61, 0, 119, 254, 14, 32, 38, 0, 64, 6, 78, 6, 81, 6, 14, 32, 61, 0, 242, 252, 14, 32, 38, 0, 64, 6, 79, 6, 14, 32, 61, 0, 121, 254, 14, 32, 38, 0, 64, 6, 79, 6, 81, 6, 14, 32, 61, 0, 243, 252, 14, 32, 38, 0, 64, 6, 80, 6, 14, 32, 61, 0, 123, 254, 14, 32, 38, 0, 64, 6, 80, 6, 81, 6, 14, 32, 61, 0, 244, 252, 14, 32, 38, 0, 64, 6, 81, 6, 14, 32, 61, 0, 125, 254, 14, 32, 38, 0, 64, 6, 82, 6, 14, 32, 61, 0, 127, 254, 14, 32, 38, 0, 65, 6, 14, 32, 61, 0, 209, 254, 14, 32, 61, 0, 210, 254, 14, 32, 61, 0, 211, 254, 14, 32, 61, 0, 212, 254, 14, 32, 38, 0, 65, 6, 44, 6, 14, 32, 61, 0, 45, 252, 14, 32, 61, 0, 190, 252, 14, 32, 38, 0, 65, 6, 45, 6, 14, 32, 61, 0, 46, 252, 14, 32, 61, 0, 191, 252, 14, 32, 38, 0, 65, 6, 46, 6, 14, 32, 61, 0, 47, 252, 14, 32, 61, 0, 192, 252, 14, 32, 38, 0, 65, 6, 46, 6, 69, 6, 14, 32, 61, 0, 124, 253, 14, 32, 61, 0, 125, 253, 14, 32, 38, 0, 65, 6, 69, 6, 14, 32, 61, 0, 48, 252, 14, 32, 61, 0, 193, 252, 14, 32, 38, 0, 65, 6, 69, 6, 74, 6, 14, 32, 61, 0, 193, 253, 14, 32, 38, 0, 65, 6, 73, 6, 14, 32, 61, 0, 49, 252, 14, 32, 61, 0, 124, 252, 14, 32, 38, 0, 65, 6, 74, 6, 14, 32, 61, 0, 50, 252, 14, 32, 61, 0, 125, 252, 14, 32, 38, 0, 66, 6, 14, 32, 61, 0, 213, 254, 14, 32, 61, 0, 214, 254, 14, 32, 61, 0, 215, 254, 14, 32, 61, 0, 216, 254, 14, 32, 38, 0, 66, 6, 45, 6, 14, 32, 61, 0, 51, 252, 14, 32, 61, 0, 194, 252, 14, 32, 38, 0, 66, 6, 68, 6, 210, 6, 14, 32, 60, 0, 60, 0, 60, 0, 241, 253, 14, 32, 38, 0, 66, 6, 69, 6, 14, 32, 61, 0, 52, 252, 14, 32, 61, 0, 195, 252, 14, 32, 38, 0, 66, 6, 69, 6, 45, 6, 14, 32, 61, 0, 126, 253, 14, 32, 61, 0, 180, 253, 14, 32, 38, 0, 66, 6, 69, 6, 69, 6, 14, 32, 61, 0, 127, 253, 14, 32, 38, 0, 66, 6, 69, 6, 74, 6, 14, 32, 61, 0, 178, 253, 14, 32, 38, 0, 66, 6, 73, 6, 14, 32, 61, 0, 53, 252, 14, 32, 61, 0, 126, 252, 14, 32, 38, 0, 66, 6, 74, 6, 14, 32, 61, 0, 54, 252, 14, 32, 61, 0, 127, 252, 14, 32, 38, 0, 67, 6, 14, 32, 61, 0, 217, 254, 14, 32, 61, 0, 218, 254, 14, 32, 61, 0, 219, 254, 14, 32, 61, 0, 220, 254, 14, 32, 38, 0, 67, 6, 39, 6, 14, 32, 61, 0, 55, 252, 14, 32, 61, 0, 128, 252, 14, 32, 38, 0, 67, 6, 44, 6, 14, 32, 61, 0, 56, 252, 14, 32, 61, 0, 196, 252, 14, 32, 38, 0, 67, 6, 45, 6, 14, 32, 61, 0, 57, 252, 14, 32, 61, 0, 197, 252, 14, 32, 38, 0, 67, 6, 46, 6, 14, 32, 61, 0, 58, 252, 14, 32, 61, 0, 198, 252, 14, 32, 38, 0, 67, 6, 68, 6, 14, 32, 61, 0, 59, 252, 14, 32, 61, 0, 129, 252, 14, 32, 61, 0, 199, 252, 14, 32, 61, 0, 235, 252, 14, 32, 38, 0, 67, 6, 69, 6, 14, 32, 61, 0, 60, 252, 14, 32, 61, 0, 130, 252, 14, 32, 61, 0, 200, 252, 14, 32, 61, 0, 236, 252, 14, 32, 38, 0, 67, 6, 69, 6, 69, 6, 14, 32, 61, 0, 187, 253, 14, 32, 61, 0, 195, 253, 14, 32, 38, 0, 67, 6, 69, 6, 74, 6, 14, 32, 61, 0, 183, 253, 14, 32, 38, 0, 67, 6, 73, 6, 14, 32, 61, 0, 61, 252, 14, 32, 61, 0, 131, 252, 14, 32, 38, 0, 67, 6, 74, 6, 14, 32, 61, 0, 62, 252, 14, 32, 61, 0, 132, 252, 14, 32, 38, 0, 68, 6, 14, 32, 61, 0, 221, 254, 14, 32, 61, 0, 222, 254, 14, 32, 61, 0, 223, 254, 14, 32, 61, 0, 224, 254, 14, 32, 38, 0, 68, 6, 39, 6, 14, 32, 61, 0, 251, 254, 14, 32, 61, 0, 252, 254, 14, 32, 38, 0, 68, 6, 34, 6, 14, 32, 61, 0, 245, 254, 14, 32, 61, 0, 246, 254, 14, 32, 38, 0, 68, 6, 35, 6, 14, 32, 61, 0, 247, 254, 14, 32, 61, 0, 248, 254, 14, 32, 38, 0, 68, 6, 37, 6, 14, 32, 61, 0, 249, 254, 14, 32, 61, 0, 250, 254, 14, 32, 38, 0, 68, 6, 44, 6, 14, 32, 61, 0, 63, 252, 14, 32, 61, 0, 201, 252, 14, 32, 38, 0, 68, 6, 44, 6, 44, 6, 14, 32, 61, 0, 131, 253, 14, 32, 61, 0, 132, 253, 14, 32, 38, 0, 68, 6, 44, 6, 69, 6, 14, 32, 61, 0, 186, 253, 14, 32, 61, 0, 188, 253, 14, 32, 38, 0, 68, 6, 44, 6, 74, 6, 14, 32, 61, 0, 172, 253, 14, 32, 38, 0, 68, 6, 45, 6, 14, 32, 61, 0, 64, 252, 14, 32, 61, 0, 202, 252, 14, 32, 38, 0, 68, 6, 45, 6, 69, 6, 14, 32, 61, 0, 128, 253, 14, 32, 61, 0, 181, 253, 14, 32, 38, 0, 68, 6, 45, 6, 73, 6, 14, 32, 61, 0, 130, 253, 14, 32, 38, 0, 68, 6, 45, 6, 74, 6, 14, 32, 61, 0, 129, 253, 14, 32, 38, 0, 68, 6, 46, 6, 14, 32, 61, 0, 65, 252, 14, 32, 61, 0, 203, 252, 14, 32, 38, 0, 68, 6, 46, 6, 69, 6, 14, 32, 61, 0, 133, 253, 14, 32, 61, 0, 134, 253, 14, 32, 38, 0, 68, 6, 69, 6, 14, 32, 61, 0, 66, 252, 14, 32, 61, 0, 133, 252, 14, 32, 61, 0, 204, 252, 14, 32, 61, 0, 237, 252, 14, 32, 38, 0, 68, 6, 69, 6, 45, 6, 14, 32, 61, 0, 135, 253, 14, 32, 61, 0, 136, 253, 14, 32, 38, 0, 68, 6, 69, 6, 74, 6, 14, 32, 61, 0, 173, 253, 14, 32, 38, 0, 68, 6, 71, 6, 14, 32, 61, 0, 205, 252, 14, 32, 38, 0, 68, 6, 73, 6, 14, 32, 61, 0, 67, 252, 14, 32, 61, 0, 134, 252, 14, 32, 38, 0, 68, 6, 74, 6, 14, 32, 61, 0, 68, 252, 14, 32, 61, 0, 135, 252, 14, 32, 38, 0, 69, 6, 14, 32, 61, 0, 225, 254, 14, 32, 61, 0, 226, 254, 14, 32, 61, 0, 227, 254, 14, 32, 61, 0, 228, 254, 14, 32, 38, 0, 69, 6, 39, 6, 14, 32, 61, 0, 136, 252, 14, 32, 38, 0, 69, 6, 44, 6, 14, 32, 61, 0, 69, 252, 14, 32, 61, 0, 206, 252, 14, 32, 38, 0, 69, 6, 44, 6, 45, 6, 14, 32, 61, 0, 140, 253, 14, 32, 38, 0, 69, 6, 44, 6, 46, 6, 14, 32, 61, 0, 146, 253, 14, 32, 38, 0, 69, 6, 44, 6, 69, 6, 14, 32, 61, 0, 141, 253, 14, 32, 38, 0, 69, 6, 44, 6, 74, 6, 14, 32, 61, 0, 192, 253, 14, 32, 38, 0, 69, 6, 45, 6, 14, 32, 61, 0, 70, 252, 14, 32, 61, 0, 207, 252, 14, 32, 38, 0, 69, 6, 45, 6, 44, 6, 14, 32, 61, 0, 137, 253, 14, 32, 38, 0, 69, 6, 45, 6, 69, 6, 14, 32, 61, 0, 138, 253, 14, 32, 38, 0, 69, 6, 45, 6, 69, 6, 47, 6, 14, 32, 60, 0, 60, 0, 60, 0, 244, 253, 14, 32, 38, 0, 69, 6, 45, 6, 74, 6, 14, 32, 61, 0, 139, 253, 14, 32, 38, 0, 69, 6, 46, 6, 14, 32, 61, 0, 71, 252, 14, 32, 61, 0, 208, 252, 14, 32, 38, 0, 69, 6, 46, 6, 44, 6, 14, 32, 61, 0, 142, 253, 14, 32, 38, 0, 69, 6, 46, 6, 69, 6, 14, 32, 61, 0, 143, 253, 14, 32, 38, 0, 69, 6, 46, 6, 74, 6, 14, 32, 61, 0, 185, 253, 14, 32, 38, 0, 69, 6, 69, 6, 14, 32, 61, 0, 72, 252, 14, 32, 61, 0, 137, 252, 14, 32, 61, 0, 209, 252, 14, 32, 38, 0, 69, 6, 69, 6, 74, 6, 14, 32, 61, 0, 177, 253, 14, 32, 38, 0, 69, 6, 73, 6, 14, 32, 61, 0, 73, 252, 14, 32, 38, 0, 69, 6, 74, 6, 14, 32, 61, 0, 74, 252, 14, 32, 38, 0, 70, 6, 14, 32, 61, 0, 229, 254, 14, 32, 61, 0, 230, 254, 14, 32, 61, 0, 231, 254, 14, 32, 61, 0, 232, 254, 14, 32, 38, 0, 70, 6, 44, 6, 14, 32, 61, 0, 75, 252, 14, 32, 61, 0, 210, 252, 14, 32, 38, 0, 70, 6, 44, 6, 45, 6, 14, 32, 61, 0, 184, 253, 14, 32, 61, 0, 189, 253, 14, 32, 38, 0, 70, 6, 44, 6, 69, 6, 14, 32, 61, 0, 151, 253, 14, 32, 61, 0, 152, 253, 14, 32, 38, 0, 70, 6, 44, 6, 73, 6, 14, 32, 61, 0, 153, 253, 14, 32, 38, 0, 70, 6, 44, 6, 74, 6, 14, 32, 61, 0, 199, 253, 14, 32, 38, 0, 70, 6, 45, 6, 14, 32, 61, 0, 76, 252, 14, 32, 61, 0, 211, 252, 14, 32, 38, 0, 70, 6, 45, 6, 69, 6, 14, 32, 61, 0, 149, 253, 14, 32, 38, 0, 70, 6, 45, 6, 73, 6, 14, 32, 61, 0, 150, 253, 14, 32, 38, 0, 70, 6, 45, 6, 74, 6, 14, 32, 61, 0, 179, 253, 14, 32, 38, 0, 70, 6, 46, 6, 14, 32, 61, 0, 77, 252, 14, 32, 61, 0, 212, 252, 14, 32, 38, 0, 70, 6, 49, 6, 14, 32, 61, 0, 138, 252, 14, 32, 38, 0, 70, 6, 50, 6, 14, 32, 61, 0, 139, 252, 14, 32, 38, 0, 70, 6, 69, 6, 14, 32, 61, 0, 78, 252, 14, 32, 61, 0, 140, 252, 14, 32, 61, 0, 213, 252, 14, 32, 61, 0, 238, 252, 14, 32, 38, 0, 70, 6, 69, 6, 73, 6, 14, 32, 61, 0, 155, 253, 14, 32, 38, 0, 70, 6, 69, 6, 74, 6, 14, 32, 61, 0, 154, 253, 14, 32, 38, 0, 70, 6, 70, 6, 14, 32, 61, 0, 141, 252, 14, 32, 38, 0, 70, 6, 71, 6, 14, 32, 61, 0, 214, 252, 14, 32, 61, 0, 239, 252, 14, 32, 38, 0, 70, 6, 73, 6, 14, 32, 61, 0, 79, 252, 14, 32, 61, 0, 142, 252, 14, 32, 38, 0, 70, 6, 74, 6, 14, 32, 61, 0, 80, 252, 14, 32, 61, 0, 143, 252, 14, 32, 38, 0, 71, 6, 14, 32, 61, 0, 233, 254, 14, 32, 61, 0, 234, 254, 14, 32, 61, 0, 235, 254, 14, 32, 61, 0, 236, 254, 14, 32, 38, 0, 71, 6, 44, 6, 14, 32, 61, 0, 81, 252, 14, 32, 61, 0, 215, 252, 14, 32, 38, 0, 71, 6, 69, 6, 14, 32, 61, 0, 82, 252, 14, 32, 61, 0, 216, 252, 14, 32, 38, 0, 71, 6, 69, 6, 44, 6, 14, 32, 61, 0, 147, 253, 14, 32, 38, 0, 71, 6, 69, 6, 69, 6, 14, 32, 61, 0, 148, 253, 14, 32, 38, 0, 71, 6, 73, 6, 14, 32, 61, 0, 83, 252, 14, 32, 38, 0, 71, 6, 74, 6, 14, 32, 61, 0, 84, 252, 14, 32, 38, 0, 71, 6, 112, 6, 14, 32, 61, 0, 217, 252, 14, 32, 38, 0, 72, 6, 14, 32, 61, 0, 237, 254, 14, 32, 61, 0, 238, 254, 14, 32, 38, 0, 72, 6, 51, 6, 68, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 248, 253, 14, 32, 38, 0, 36, 6, 14, 32, 61, 0, 133, 254, 14, 32, 61, 0, 134, 254, 14, 32, 38, 0, 73, 6, 14, 32, 61, 0, 232, 251, 14, 32, 61, 0, 233, 251, 14, 32, 61, 0, 239, 254, 14, 32, 61, 0, 240, 254, 14, 32, 38, 0, 73, 6, 112, 6, 14, 32, 61, 0, 93, 252, 14, 32, 61, 0, 144, 252, 14, 32, 38, 0, 74, 6, 14, 32, 61, 0, 241, 254, 14, 32, 61, 0, 242, 254, 14, 32, 61, 0, 243, 254, 14, 32, 61, 0, 244, 254, 14, 32, 38, 0, 74, 6, 44, 6, 14, 32, 61, 0, 85, 252, 14, 32, 61, 0, 218, 252, 14, 32, 38, 0, 74, 6, 44, 6, 74, 6, 14, 32, 61, 0, 175, 253, 14, 32, 38, 0, 74, 6, 45, 6, 14, 32, 61, 0, 86, 252, 14, 32, 61, 0, 219, 252, 14, 32, 38, 0, 74, 6, 45, 6, 74, 6, 14, 32, 61, 0, 174, 253, 14, 32, 38, 0, 74, 6, 46, 6, 14, 32, 61, 0, 87, 252, 14, 32, 61, 0, 220, 252, 14, 32, 38, 0, 74, 6, 49, 6, 14, 32, 61, 0, 145, 252, 14, 32, 38, 0, 74, 6, 50, 6, 14, 32, 61, 0, 146, 252, 14, 32, 38, 0, 74, 6, 69, 6, 14, 32, 61, 0, 88, 252, 14, 32, 61, 0, 147, 252, 14, 32, 61, 0, 221, 252, 14, 32, 61, 0, 240, 252, 14, 32, 38, 0, 74, 6, 69, 6, 69, 6, 14, 32, 61, 0, 156, 253, 14, 32, 61, 0, 157, 253, 14, 32, 38, 0, 74, 6, 69, 6, 74, 6, 14, 32, 61, 0, 176, 253, 14, 32, 38, 0, 74, 6, 70, 6, 14, 32, 61, 0, 148, 252, 14, 32, 38, 0, 74, 6, 71, 6, 14, 32, 61, 0, 222, 252, 14, 32, 61, 0, 241, 252, 14, 32, 38, 0, 74, 6, 73, 6, 14, 32, 61, 0, 89, 252, 14, 32, 61, 0, 149, 252, 14, 32, 38, 0, 74, 6, 74, 6, 14, 32, 61, 0, 90, 252, 14, 32, 61, 0, 150, 252, 14, 32, 38, 0, 38, 6, 14, 32, 61, 0, 137, 254, 14, 32, 61, 0, 138, 254, 14, 32, 61, 0, 139, 254, 14, 32, 61, 0, 140, 254, 14, 32, 38, 0, 38, 6, 39, 6, 14, 32, 61, 0, 234, 251, 14, 32, 61, 0, 235, 251, 14, 32, 38, 0, 38, 6, 44, 6, 14, 32, 61, 0, 0, 252, 14, 32, 61, 0, 151, 252, 14, 32, 38, 0, 38, 6, 45, 6, 14, 32, 61, 0, 1, 252, 14, 32, 61, 0, 152, 252, 14, 32, 38, 0, 38, 6, 46, 6, 14, 32, 61, 0, 153, 252, 14, 32, 38, 0, 38, 6, 49, 6, 14, 32, 61, 0, 100, 252, 14, 32, 38, 0, 38, 6, 50, 6, 14, 32, 61, 0, 101, 252, 14, 32, 38, 0, 38, 6, 69, 6, 14, 32, 61, 0, 2, 252, 14, 32, 61, 0, 102, 252, 14, 32, 61, 0, 154, 252, 14, 32, 61, 0, 223, 252, 14, 32, 38, 0, 38, 6, 70, 6, 14, 32, 61, 0, 103, 252, 14, 32, 38, 0, 38, 6, 71, 6, 14, 32, 61, 0, 155, 252, 14, 32, 61, 0, 224, 252, 14, 32, 38, 0, 38, 6, 72, 6, 14, 32, 61, 0, 238, 251, 14, 32, 61, 0, 239, 251, 14, 32, 38, 0, 38, 6, 73, 6, 14, 32, 61, 0, 249, 251, 14, 32, 61, 0, 250, 251, 14, 32, 61, 0, 251, 251, 14, 32, 61, 0, 3, 252, 14, 32, 61, 0, 104, 252, 14, 32, 38, 0, 38, 6, 74, 6, 14, 32, 61, 0, 4, 252, 14, 32, 61, 0, 105, 252, 14, 32, 38, 0, 38, 6, 198, 6, 14, 32, 61, 0, 242, 251, 14, 32, 61, 0, 243, 251, 14, 32, 38, 0, 38, 6, 199, 6, 14, 32, 61, 0, 240, 251, 14, 32, 61, 0, 241, 251, 14, 32, 38, 0, 38, 6, 200, 6, 14, 32, 61, 0, 244, 251, 14, 32, 61, 0, 245, 251, 14, 32, 38, 0, 38, 6, 208, 6, 14, 32, 61, 0, 246, 251, 14, 32, 61, 0, 247, 251, 14, 32, 61, 0, 248, 251, 14, 32, 38, 0, 38, 6, 213, 6, 14, 32, 61, 0, 236, 251, 14, 32, 61, 0, 237, 251, 14, 32, 38, 0, 75, 6, 14, 32, 61, 0, 112, 254, 14, 32, 38, 0, 76, 6, 14, 32, 61, 0, 114, 254, 14, 32, 38, 0, 76, 6, 81, 6, 14, 32, 61, 0, 94, 252, 14, 32, 38, 0, 77, 6, 14, 32, 61, 0, 116, 254, 14, 32, 38, 0, 77, 6, 81, 6, 14, 32, 61, 0, 95, 252, 14, 32, 38, 0, 78, 6, 14, 32, 61, 0, 118, 254, 14, 32, 38, 0, 78, 6, 81, 6, 14, 32, 61, 0, 96, 252, 14, 32, 38, 0, 79, 6, 14, 32, 61, 0, 120, 254, 14, 32, 38, 0, 79, 6, 81, 6, 14, 32, 61, 0, 97, 252, 14, 32, 38, 0, 80, 6, 14, 32, 61, 0, 122, 254, 14, 32, 38, 0, 80, 6, 81, 6, 14, 32, 61, 0, 98, 252, 14, 32, 38, 0, 81, 6, 14, 32, 61, 0, 124, 254, 14, 32, 38, 0, 81, 6, 112, 6, 14, 32, 61, 0, 99, 252, 14, 32, 38, 0, 82, 6, 14, 32, 61, 0, 126, 254, 14, 32, 38, 0, 113, 6, 14, 32, 61, 0, 80, 251, 14, 32, 61, 0, 81, 251, 14, 32, 38, 0, 121, 6, 14, 32, 61, 0, 102, 251, 14, 32, 61, 0, 103, 251, 14, 32, 61, 0, 104, 251, 14, 32, 61, 0, 105, 251, 14, 32, 38, 0, 122, 6, 14, 32, 61, 0, 94, 251, 14, 32, 61, 0, 95, 251, 14, 32, 61, 0, 96, 251, 14, 32, 61, 0, 97, 251, 14, 32, 38, 0, 123, 6, 14, 32, 61, 0, 82, 251, 14, 32, 61, 0, 83, 251, 14, 32, 61, 0, 84, 251, 14, 32, 61, 0, 85, 251, 14, 32, 38, 0, 126, 6, 14, 32, 61, 0, 86, 251, 14, 32, 61, 0, 87, 251, 14, 32, 61, 0, 88, 251, 14, 32, 61, 0, 89, 251, 14, 32, 38, 0, 127, 6, 14, 32, 61, 0, 98, 251, 14, 32, 61, 0, 99, 251, 14, 32, 61, 0, 100, 251, 14, 32, 61, 0, 101, 251, 14, 32, 38, 0, 128, 6, 14, 32, 61, 0, 90, 251, 14, 32, 61, 0, 91, 251, 14, 32, 61, 0, 92, 251, 14, 32, 61, 0, 93, 251, 14, 32, 38, 0, 131, 6, 14, 32, 61, 0, 118, 251, 14, 32, 61, 0, 119, 251, 14, 32, 61, 0, 120, 251, 14, 32, 61, 0, 121, 251, 14, 32, 38, 0, 132, 6, 14, 32, 61, 0, 114, 251, 14, 32, 61, 0, 115, 251, 14, 32, 61, 0, 116, 251, 14, 32, 61, 0, 117, 251, 14, 32, 38, 0, 134, 6, 14, 32, 61, 0, 122, 251, 14, 32, 61, 0, 123, 251, 14, 32, 61, 0, 124, 251, 14, 32, 61, 0, 125, 251, 14, 32, 38, 0, 135, 6, 14, 32, 61, 0, 126, 251, 14, 32, 61, 0, 127, 251, 14, 32, 61, 0, 128, 251, 14, 32, 61, 0, 129, 251, 14, 32, 38, 0, 136, 6, 14, 32, 61, 0, 136, 251, 14, 32, 61, 0, 137, 251, 14, 32, 38, 0, 140, 6, 14, 32, 61, 0, 132, 251, 14, 32, 61, 0, 133, 251, 14, 32, 38, 0, 141, 6, 14, 32, 61, 0, 130, 251, 14, 32, 61, 0, 131, 251, 14, 32, 38, 0, 142, 6, 14, 32, 61, 0, 134, 251, 14, 32, 61, 0, 135, 251, 14, 32, 38, 0, 145, 6, 14, 32, 61, 0, 140, 251, 14, 32, 61, 0, 141, 251, 14, 32, 38, 0, 152, 6, 14, 32, 61, 0, 138, 251, 14, 32, 61, 0, 139, 251, 14, 32, 38, 0, 164, 6, 14, 32, 61, 0, 106, 251, 14, 32, 61, 0, 107, 251, 14, 32, 61, 0, 108, 251, 14, 32, 61, 0, 109, 251, 14, 32, 38, 0, 166, 6, 14, 32, 61, 0, 110, 251, 14, 32, 61, 0, 111, 251, 14, 32, 61, 0, 112, 251, 14, 32, 61, 0, 113, 251, 14, 32, 38, 0, 169, 6, 14, 32, 61, 0, 142, 251, 14, 32, 61, 0, 143, 251, 14, 32, 61, 0, 144, 251, 14, 32, 61, 0, 145, 251, 14, 32, 38, 0, 173, 6, 14, 32, 61, 0, 211, 251, 14, 32, 61, 0, 212, 251, 14, 32, 61, 0, 213, 251, 14, 32, 61, 0, 214, 251, 14, 32, 38, 0, 175, 6, 14, 32, 61, 0, 146, 251, 14, 32, 61, 0, 147, 251, 14, 32, 61, 0, 148, 251, 14, 32, 61, 0, 149, 251, 14, 32, 38, 0, 177, 6, 14, 32, 61, 0, 154, 251, 14, 32, 61, 0, 155, 251, 14, 32, 61, 0, 156, 251, 14, 32, 61, 0, 157, 251, 14, 32, 38, 0, 179, 6, 14, 32, 61, 0, 150, 251, 14, 32, 61, 0, 151, 251, 14, 32, 61, 0, 152, 251, 14, 32, 61, 0, 153, 251, 14, 32, 38, 0, 186, 6, 14, 32, 61, 0, 158, 251, 14, 32, 61, 0, 159, 251, 14, 32, 38, 0, 187, 6, 14, 32, 61, 0, 160, 251, 14, 32, 61, 0, 161, 251, 14, 32, 61, 0, 162, 251, 14, 32, 61, 0, 163, 251, 14, 32, 38, 0, 190, 6, 14, 32, 61, 0, 170, 251, 14, 32, 61, 0, 171, 251, 14, 32, 61, 0, 172, 251, 14, 32, 61, 0, 173, 251, 14, 32, 38, 0, 193, 6, 14, 32, 61, 0, 166, 251, 14, 32, 61, 0, 167, 251, 14, 32, 61, 0, 168, 251, 14, 32, 61, 0, 169, 251, 14, 32, 38, 0, 197, 6, 14, 32, 61, 0, 224, 251, 14, 32, 61, 0, 225, 251, 14, 32, 38, 0, 198, 6, 14, 32, 61, 0, 217, 251, 14, 32, 61, 0, 218, 251, 14, 32, 38, 0, 199, 6, 14, 32, 61, 0, 215, 251, 14, 32, 61, 0, 216, 251, 14, 32, 38, 0, 199, 6, 116, 6, 14, 32, 61, 0, 221, 251, 14, 32, 38, 0, 200, 6, 14, 32, 61, 0, 219, 251, 14, 32, 61, 0, 220, 251, 14, 32, 38, 0, 201, 6, 14, 32, 61, 0, 226, 251, 14, 32, 61, 0, 227, 251, 14, 32, 38, 0, 203, 6, 14, 32, 61, 0, 222, 251, 14, 32, 61, 0, 223, 251, 14, 32, 38, 0, 204, 6, 14, 32, 61, 0, 252, 251, 14, 32, 61, 0, 253, 251, 14, 32, 61, 0, 254, 251, 14, 32, 61, 0, 255, 251, 14, 32, 38, 0, 208, 6, 14, 32, 61, 0, 228, 251, 14, 32, 61, 0, 229, 251, 14, 32, 61, 0, 230, 251, 14, 32, 61, 0, 231, 251, 14, 32, 38, 0, 210, 6, 14, 32, 61, 0, 174, 251, 14, 32, 61, 0, 175, 251, 14, 32, 38, 0, 211, 6, 14, 32, 61, 0, 176, 251, 14, 32, 61, 0, 177, 251, 14, 32, 38, 0, 192, 6, 14, 32, 61, 0, 164, 251, 14, 32, 61, 0, 165, 251, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 84, 28, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 56, 0, 0, 0, 56, 1, 0, 0, 248, 27, 0, 0, 248, 27, 0, 0, 248, 27, 0, 0, 248, 27, 0, 0, 60, 28, 0, 0, 2, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 39, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 163, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 94, 0, 102, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 154, 0, 118, 0, 48, 0, 48, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 124, 0, 48, 0, 48, 0, 127, 0, 48, 0, 135, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 140, 0, 48, 0, 48, 0, 146, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 120, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 42, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 70, 238, 102, 7, 70, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 11, 70, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 10, 70, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 42, 102, 6, 70, 42, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 70, 238, 102, 8, 70, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 67, 0, 58, 0, 32, 0, 51, 8, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 84, 53, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 15, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 60, 0, 0, 0, 64, 0, 0, 0, 64, 1, 0, 0, 144, 40, 0, 0, 144, 40, 0, 0, 56, 42, 0, 0, 56, 42, 0, 0, 248, 52, 0, 0, 248, 52, 0, 0, 60, 53, 0, 0, 2, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 39, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 108, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 94, 0, 102, 0, 48, 0, 56, 0, 107, 0, 115, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 138, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 99, 1, 131, 0, 48, 0, 48, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 146, 0, 154, 0, 162, 0, 170, 0, 175, 0, 183, 0, 191, 0, 199, 0, 207, 0, 215, 0, 223, 0, 231, 0, 239, 0, 247, 0, 255, 0, 7, 1, 15, 1, 23, 1, 31, 1, 39, 1, 47, 1, 51, 1, 48, 0, 48, 0, 48, 0, 59, 1, 67, 1, 75, 1, 83, 1, 91, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 120, 1, 192, 0, 172, 1, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 192, 84, 0, 202, 225, 84, 0, 202, 2, 85, 0, 202, 35, 85, 0, 202, 68, 85, 0, 202, 101, 85, 0, 202, 134, 85, 0, 202, 167, 85, 0, 202, 200, 85, 0, 202, 233, 85, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 2, 84, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 66, 84, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 130, 84, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 5, 70, 6, 102, 5, 71, 6, 102, 5, 72, 6, 102, 5, 73, 6, 102, 201, 6, 0, 0, 192, 0, 0, 0, 5, 2, 42, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 1, 0, 201, 134, 2, 0, 201, 102, 3, 0, 198, 1, 0, 0, 198, 33, 0, 0, 198, 65, 0, 0, 198, 97, 0, 0, 198, 129, 0, 0, 198, 161, 0, 0, 198, 193, 0, 0, 198, 225, 0, 0, 198, 1, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 1, 0, 198, 65, 1, 0, 198, 97, 1, 0, 198, 129, 1, 0, 198, 161, 1, 0, 198, 193, 1, 0, 198, 225, 1, 0, 198, 1, 2, 0, 198, 33, 2, 0, 198, 65, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 191, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 2, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 97, 2, 0, 192, 0, 0, 0, 198, 129, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 2, 0, 198, 193, 2, 0, 198, 225, 2, 0, 198, 1, 3, 0, 198, 33, 3, 0, 198, 65, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 97, 3, 0, 198, 129, 3, 0, 198, 161, 3, 0, 192, 0, 0, 0, 198, 193, 3, 0, 198, 225, 3, 0, 198, 1, 4, 0, 198, 33, 4, 0, 198, 65, 4, 0, 198, 97, 4, 0, 190, 0, 0, 0, 198, 129, 4, 0, 198, 161, 4, 0, 198, 193, 4, 0, 198, 225, 4, 0, 198, 1, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 102, 5, 5, 14, 102, 5, 5, 32, 102, 5, 5, 32, 102, 5, 5, 32, 102, 5, 5, 32, 102, 5, 5, 34, 102, 5, 5, 34, 102, 5, 5, 34, 102, 5, 5, 34, 102, 5, 5, 36, 102, 5, 5, 36, 102, 5, 5, 36, 102, 5, 5, 36, 102, 5, 5, 48, 102, 5, 5, 48, 102, 5, 5, 48, 102, 5, 5, 48, 102, 5, 5, 54, 102, 5, 5, 54, 102, 5, 5, 54, 102, 5, 5, 54, 102, 5, 5, 46, 102, 5, 5, 46, 102, 5, 5, 46, 102, 5, 5, 46, 102, 193, 30, 176, 102, 193, 30, 176, 102, 193, 30, 176, 102, 193, 30, 176, 102, 193, 51, 176, 102, 193, 51, 176, 102, 193, 51, 176, 102, 193, 51, 176, 102, 5, 5, 62, 102, 5, 5, 62, 102, 5, 5, 62, 102, 5, 5, 62, 102, 5, 5, 60, 102, 5, 5, 60, 102, 5, 5, 60, 102, 5, 5, 60, 102, 5, 5, 64, 102, 5, 5, 64, 102, 5, 5, 64, 102, 5, 5, 64, 102, 5, 5, 68, 102, 5, 5, 68, 102, 5, 5, 68, 102, 5, 5, 68, 102, 5, 5, 98, 102, 5, 5, 98, 102, 5, 5, 96, 102, 5, 5, 96, 102, 5, 5, 102, 102, 5, 5, 102, 102, 5, 5, 88, 102, 5, 5, 88, 102, 5, 5, 128, 102, 5, 5, 128, 102, 5, 5, 114, 102, 5, 5, 114, 102, 5, 5, 184, 102, 5, 5, 184, 102, 5, 5, 184, 102, 5, 5, 184, 102, 5, 5, 192, 102, 5, 5, 192, 102, 5, 5, 192, 102, 5, 5, 192, 102, 193, 30, 194, 102, 193, 30, 194, 102, 193, 30, 194, 102, 193, 30, 194, 102, 193, 16, 194, 102, 193, 16, 194, 102, 193, 16, 194, 102, 193, 16, 194, 102, 5, 5, 206, 102, 5, 5, 206, 102, 193, 9, 208, 102, 193, 9, 208, 102, 193, 9, 208, 102, 193, 9, 208, 102, 197, 98, 0, 0, 197, 98, 0, 0, 5, 5, 218, 102, 5, 5, 218, 102, 5, 5, 218, 102, 5, 5, 218, 102, 5, 5, 216, 102, 5, 5, 216, 102, 5, 5, 216, 102, 5, 5, 216, 102, 5, 5, 250, 102, 5, 5, 250, 102, 197, 162, 0, 0, 197, 162, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 16, 190, 102, 193, 16, 190, 102, 193, 16, 190, 102, 193, 16, 190, 102, 5, 5, 228, 102, 5, 5, 228, 102, 193, 9, 226, 102, 193, 9, 226, 102, 5, 5, 230, 102, 5, 5, 230, 102, 197, 226, 0, 0, 193, 9, 234, 102, 193, 9, 234, 102, 193, 2, 226, 102, 193, 2, 226, 102, 5, 5, 232, 102, 5, 5, 232, 102, 5, 5, 246, 102, 5, 5, 246, 102, 5, 5, 246, 102, 5, 5, 246, 102, 5, 3, 238, 102, 5, 3, 238, 102, 197, 34, 1, 0, 197, 34, 1, 0, 197, 98, 1, 0, 197, 98, 1, 0, 197, 162, 1, 0, 197, 162, 1, 0, 197, 226, 1, 0, 197, 226, 1, 0, 197, 34, 2, 0, 197, 34, 2, 0, 197, 98, 2, 0, 197, 98, 2, 0, 197, 162, 2, 0, 197, 162, 2, 0, 197, 162, 2, 0, 197, 226, 2, 0, 197, 226, 2, 0, 197, 226, 2, 0, 5, 2, 238, 102, 5, 2, 238, 102, 5, 2, 238, 102, 5, 2, 238, 102, 197, 34, 3, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 226, 2, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 226, 4, 0, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 197, 226, 6, 0, 197, 34, 7, 0, 197, 98, 7, 0, 197, 162, 7, 0, 197, 226, 7, 0, 197, 34, 8, 0, 197, 98, 8, 0, 197, 66, 8, 0, 197, 162, 8, 0, 197, 226, 8, 0, 197, 34, 9, 0, 197, 98, 9, 0, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 197, 34, 14, 0, 197, 98, 14, 0, 197, 162, 14, 0, 197, 226, 14, 0, 197, 34, 15, 0, 197, 98, 15, 0, 197, 162, 15, 0, 197, 226, 15, 0, 197, 34, 16, 0, 197, 98, 16, 0, 197, 162, 16, 0, 197, 226, 16, 0, 197, 34, 17, 0, 197, 98, 17, 0, 197, 162, 17, 0, 197, 226, 17, 0, 197, 34, 18, 0, 197, 98, 18, 0, 197, 162, 18, 0, 197, 226, 18, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 197, 226, 19, 0, 197, 130, 8, 0, 197, 194, 8, 0, 197, 34, 20, 0, 197, 98, 20, 0, 197, 162, 20, 0, 197, 226, 20, 0, 197, 34, 21, 0, 197, 98, 21, 0, 197, 162, 21, 0, 197, 226, 21, 0, 197, 34, 22, 0, 197, 98, 22, 0, 197, 162, 22, 0, 197, 226, 22, 0, 197, 34, 23, 0, 197, 2, 8, 0, 197, 98, 23, 0, 197, 162, 23, 0, 197, 194, 19, 0, 197, 226, 23, 0, 197, 66, 23, 0, 197, 34, 24, 0, 197, 98, 24, 0, 197, 162, 24, 0, 198, 34, 5, 0, 198, 98, 5, 0, 198, 162, 5, 0, 198, 226, 5, 0, 198, 162, 0, 0, 198, 34, 6, 0, 197, 226, 24, 0, 197, 34, 25, 0, 197, 162, 3, 0, 197, 98, 25, 0, 197, 226, 2, 0, 197, 226, 3, 0, 197, 162, 25, 0, 197, 226, 25, 0, 197, 226, 4, 0, 197, 34, 26, 0, 197, 34, 5, 0, 197, 98, 5, 0, 197, 98, 26, 0, 197, 162, 26, 0, 197, 98, 6, 0, 197, 226, 26, 0, 197, 162, 6, 0, 197, 226, 6, 0, 197, 34, 27, 0, 197, 98, 27, 0, 197, 98, 7, 0, 197, 162, 27, 0, 197, 162, 7, 0, 197, 226, 7, 0, 197, 226, 14, 0, 197, 34, 15, 0, 197, 226, 15, 0, 197, 34, 16, 0, 197, 98, 16, 0, 197, 98, 17, 0, 197, 162, 17, 0, 197, 226, 17, 0, 197, 34, 18, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 197, 226, 27, 0, 197, 34, 20, 0, 197, 34, 28, 0, 197, 98, 28, 0, 197, 162, 21, 0, 197, 162, 28, 0, 197, 226, 21, 0, 197, 34, 22, 0, 197, 162, 24, 0, 197, 226, 28, 0, 197, 34, 29, 0, 197, 194, 19, 0, 197, 194, 20, 0, 197, 226, 23, 0, 197, 66, 23, 0, 197, 34, 3, 0, 197, 98, 3, 0, 197, 98, 29, 0, 197, 162, 3, 0, 197, 162, 29, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 226, 4, 0, 197, 226, 29, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 34, 30, 0, 197, 98, 7, 0, 197, 34, 8, 0, 197, 98, 8, 0, 197, 66, 8, 0, 197, 162, 8, 0, 197, 226, 8, 0, 197, 98, 9, 0, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 98, 30, 0, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 197, 34, 14, 0, 197, 98, 14, 0, 197, 162, 14, 0, 197, 98, 15, 0, 197, 162, 15, 0, 197, 162, 16, 0, 197, 226, 16, 0, 197, 34, 17, 0, 197, 98, 17, 0, 197, 162, 17, 0, 197, 98, 18, 0, 197, 162, 18, 0, 197, 226, 18, 0, 197, 34, 19, 0, 197, 162, 30, 0, 197, 226, 19, 0, 197, 130, 8, 0, 197, 194, 8, 0, 197, 34, 20, 0, 197, 226, 20, 0, 197, 34, 21, 0, 197, 98, 21, 0, 197, 162, 21, 0, 197, 226, 30, 0, 197, 98, 22, 0, 197, 162, 22, 0, 197, 34, 31, 0, 197, 2, 8, 0, 197, 98, 23, 0, 197, 162, 23, 0, 197, 194, 19, 0, 197, 66, 22, 0, 197, 162, 3, 0, 197, 162, 29, 0, 197, 226, 4, 0, 197, 226, 29, 0, 197, 98, 6, 0, 197, 34, 30, 0, 197, 98, 7, 0, 197, 98, 31, 0, 197, 98, 10, 0, 197, 162, 31, 0, 197, 226, 31, 0, 197, 34, 32, 0, 197, 98, 17, 0, 197, 162, 17, 0, 197, 34, 19, 0, 197, 162, 21, 0, 197, 226, 30, 0, 197, 194, 19, 0, 197, 66, 22, 0, 198, 162, 5, 0, 198, 226, 5, 0, 198, 162, 0, 0, 197, 98, 32, 0, 197, 162, 32, 0, 197, 226, 32, 0, 197, 34, 33, 0, 197, 98, 33, 0, 197, 162, 33, 0, 197, 226, 33, 0, 197, 34, 34, 0, 197, 98, 34, 0, 197, 162, 34, 0, 197, 226, 34, 0, 197, 130, 23, 0, 197, 34, 35, 0, 197, 98, 35, 0, 197, 162, 35, 0, 197, 194, 23, 0, 197, 226, 35, 0, 197, 34, 36, 0, 197, 98, 36, 0, 197, 162, 36, 0, 197, 226, 36, 0, 197, 34, 37, 0, 197, 98, 37, 0, 197, 226, 31, 0, 197, 162, 37, 0, 197, 226, 37, 0, 197, 34, 38, 0, 197, 98, 38, 0, 197, 98, 32, 0, 197, 162, 32, 0, 197, 226, 32, 0, 197, 34, 33, 0, 197, 98, 33, 0, 197, 162, 33, 0, 197, 226, 33, 0, 197, 34, 34, 0, 197, 98, 34, 0, 197, 162, 34, 0, 197, 226, 34, 0, 197, 130, 23, 0, 197, 34, 35, 0, 197, 98, 35, 0, 197, 162, 35, 0, 197, 194, 23, 0, 197, 226, 35, 0, 197, 34, 36, 0, 197, 98, 36, 0, 197, 162, 36, 0, 197, 226, 36, 0, 197, 34, 37, 0, 197, 98, 37, 0, 197, 226, 31, 0, 197, 162, 37, 0, 197, 226, 37, 0, 197, 34, 38, 0, 197, 98, 38, 0, 197, 226, 36, 0, 197, 34, 37, 0, 197, 98, 37, 0, 197, 226, 31, 0, 197, 162, 31, 0, 197, 34, 32, 0, 197, 98, 12, 0, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 226, 36, 0, 197, 34, 37, 0, 197, 98, 37, 0, 197, 98, 12, 0, 197, 162, 12, 0, 198, 98, 6, 0, 198, 98, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 38, 0, 197, 3, 39, 0, 197, 3, 39, 0, 197, 99, 39, 0, 197, 195, 39, 0, 197, 35, 40, 0, 197, 131, 40, 0, 197, 227, 40, 0, 197, 99, 8, 0, 197, 99, 8, 0, 197, 67, 41, 0, 197, 163, 41, 0, 197, 3, 42, 0, 197, 99, 42, 0, 197, 195, 42, 0, 197, 35, 43, 0, 197, 35, 43, 0, 197, 131, 43, 0, 197, 227, 43, 0, 197, 227, 43, 0, 197, 67, 44, 0, 197, 67, 44, 0, 197, 163, 44, 0, 197, 3, 45, 0, 197, 3, 45, 0, 197, 99, 45, 0, 197, 195, 45, 0, 197, 195, 45, 0, 197, 35, 46, 0, 197, 35, 46, 0, 197, 131, 46, 0, 197, 227, 46, 0, 197, 227, 46, 0, 197, 67, 47, 0, 197, 67, 47, 0, 197, 163, 47, 0, 197, 3, 48, 0, 197, 99, 48, 0, 197, 195, 48, 0, 197, 195, 48, 0, 197, 35, 49, 0, 197, 131, 49, 0, 197, 227, 49, 0, 197, 67, 50, 0, 197, 163, 50, 0, 197, 163, 50, 0, 197, 3, 51, 0, 197, 99, 51, 0, 197, 195, 51, 0, 197, 35, 52, 0, 197, 131, 52, 0, 197, 227, 52, 0, 197, 227, 52, 0, 197, 67, 53, 0, 197, 67, 53, 0, 197, 163, 53, 0, 197, 163, 53, 0, 197, 3, 54, 0, 197, 131, 8, 0, 197, 99, 54, 0, 197, 195, 54, 0, 197, 227, 19, 0, 197, 195, 8, 0, 197, 35, 55, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 55, 0, 197, 227, 55, 0, 197, 67, 56, 0, 197, 163, 56, 0, 197, 3, 57, 0, 197, 99, 57, 0, 197, 99, 57, 0, 197, 195, 57, 0, 197, 35, 58, 0, 197, 131, 58, 0, 197, 227, 58, 0, 197, 227, 58, 0, 197, 67, 59, 0, 197, 163, 59, 0, 197, 3, 60, 0, 197, 99, 60, 0, 197, 195, 60, 0, 197, 35, 61, 0, 197, 131, 61, 0, 197, 227, 61, 0, 197, 67, 62, 0, 197, 163, 62, 0, 197, 3, 63, 0, 197, 99, 63, 0, 197, 195, 63, 0, 197, 35, 64, 0, 197, 131, 64, 0, 197, 227, 64, 0, 197, 99, 23, 0, 197, 67, 65, 0, 197, 163, 65, 0, 197, 3, 66, 0, 197, 99, 66, 0, 197, 195, 66, 0, 197, 3, 51, 0, 197, 195, 51, 0, 197, 35, 67, 0, 197, 131, 67, 0, 197, 227, 67, 0, 197, 67, 68, 0, 197, 163, 68, 0, 197, 3, 69, 0, 197, 163, 68, 0, 197, 227, 67, 0, 197, 99, 69, 0, 197, 195, 69, 0, 197, 35, 70, 0, 197, 131, 70, 0, 197, 227, 70, 0, 197, 3, 69, 0, 197, 99, 48, 0, 197, 163, 44, 0, 197, 67, 71, 0, 197, 163, 71, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 72, 0, 197, 99, 72, 0, 197, 196, 72, 0, 197, 68, 73, 0, 197, 196, 73, 0, 197, 68, 74, 0, 197, 196, 74, 0, 197, 68, 75, 0, 197, 196, 75, 0, 197, 67, 76, 0, 197, 178, 76, 0, 197, 232, 78, 0, 197, 228, 79, 0, 197, 118, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 198, 1, 0, 0, 198, 33, 0, 0, 192, 0, 0, 0, 198, 65, 0, 0, 192, 0, 0, 0, 198, 97, 0, 0, 198, 97, 0, 0, 198, 129, 0, 0, 198, 129, 0, 0, 198, 161, 0, 0, 198, 161, 0, 0, 198, 193, 0, 0, 198, 193, 0, 0, 198, 225, 0, 0, 198, 225, 0, 0, 5, 5, 6, 102, 197, 34, 0, 0, 197, 34, 0, 0, 5, 70, 6, 102, 5, 70, 6, 102, 5, 71, 6, 102, 5, 71, 6, 102, 5, 72, 6, 102, 5, 72, 6, 102, 5, 73, 6, 102, 5, 73, 6, 102, 5, 73, 6, 102, 5, 73, 6, 102, 5, 75, 6, 102, 5, 75, 6, 102, 5, 5, 30, 102, 5, 5, 30, 102, 5, 5, 30, 102, 5, 5, 30, 102, 5, 2, 42, 102, 5, 2, 42, 102, 5, 5, 42, 102, 5, 5, 42, 102, 5, 5, 42, 102, 5, 5, 42, 102, 5, 5, 44, 102, 5, 5, 44, 102, 5, 5, 44, 102, 5, 5, 44, 102, 5, 5, 58, 102, 5, 5, 58, 102, 5, 5, 58, 102, 5, 5, 58, 102, 5, 5, 72, 102, 5, 5, 72, 102, 5, 5, 72, 102, 5, 5, 72, 102, 5, 5, 74, 102, 5, 5, 74, 102, 5, 5, 74, 102, 5, 5, 74, 102, 5, 5, 84, 102, 5, 5, 84, 102, 5, 5, 86, 102, 5, 5, 86, 102, 5, 5, 110, 102, 5, 5, 110, 102, 5, 5, 112, 102, 5, 5, 112, 102, 5, 5, 134, 102, 5, 5, 134, 102, 5, 5, 134, 102, 5, 5, 134, 102, 5, 5, 136, 102, 5, 5, 136, 102, 5, 5, 136, 102, 5, 5, 136, 102, 5, 5, 146, 102, 5, 5, 146, 102, 5, 5, 146, 102, 5, 5, 146, 102, 5, 5, 148, 102, 5, 5, 148, 102, 5, 5, 148, 102, 5, 5, 148, 102, 5, 5, 158, 102, 5, 5, 158, 102, 5, 5, 158, 102, 5, 5, 158, 102, 5, 5, 160, 102, 5, 5, 160, 102, 5, 5, 160, 102, 5, 5, 160, 102, 5, 5, 166, 102, 5, 5, 166, 102, 5, 5, 166, 102, 5, 5, 166, 102, 5, 5, 168, 102, 5, 5, 168, 102, 5, 5, 168, 102, 5, 5, 168, 102, 5, 5, 174, 102, 5, 5, 174, 102, 5, 5, 174, 102, 5, 5, 174, 102, 5, 5, 178, 102, 5, 5, 178, 102, 5, 5, 178, 102, 5, 5, 178, 102, 5, 5, 182, 102, 5, 5, 182, 102, 5, 5, 182, 102, 5, 5, 182, 102, 5, 5, 196, 102, 5, 5, 196, 102, 5, 5, 196, 102, 5, 5, 196, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 204, 102, 5, 5, 204, 102, 5, 5, 204, 102, 5, 5, 204, 102, 5, 5, 214, 102, 5, 5, 214, 102, 5, 5, 214, 102, 5, 5, 214, 102, 5, 5, 222, 102, 5, 5, 222, 102, 5, 3, 238, 102, 5, 3, 238, 102, 5, 5, 238, 102, 5, 5, 238, 102, 5, 5, 238, 102, 5, 5, 238, 102, 197, 35, 83, 0, 197, 35, 83, 0, 197, 130, 83, 0, 197, 130, 83, 0, 197, 194, 83, 0, 197, 194, 83, 0, 197, 98, 79, 0, 197, 98, 79, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 3, 189, 0, 0, 0, 0, 0, 0, 6, 189, 0, 0, 0, 0, 0, 0, 9, 189, 0, 0, 0, 0, 0, 0, 12, 189, 0, 0, 0, 0, 0, 0, 16, 189, 0, 0, 0, 0, 0, 0, 19, 189, 0, 0, 0, 0, 0, 0, 22, 189, 0, 0, 0, 0, 0, 0, 23, 189, 0, 0, 0, 0, 0, 0, 24, 189, 0, 0, 0, 0, 0, 0, 28, 189, 0, 0, 0, 0, 0, 0, 29, 189, 0, 0, 0, 0, 0, 0, 30, 189, 0, 0, 0, 0, 0, 0, 32, 189, 0, 0, 0, 0, 0, 0, 33, 189, 0, 0, 0, 0, 0, 0, 34, 189, 0, 0, 0, 0, 0, 0, 35, 189, 0, 0, 0, 0, 0, 0, 36, 189, 0, 0, 0, 0, 0, 0, 37, 189, 0, 0, 0, 0, 0, 0, 27, 189, 0, 0, 0, 0, 0, 0, 25, 189, 0, 0, 0, 0, 0, 0, 26, 189, 0, 0, 0, 0, 0, 0, 13, 189, 0, 0, 0, 0, 0, 0, 17, 189, 0, 0, 0, 0, 0, 0, 20, 189, 0, 0, 0, 0, 0, 0, 5, 189, 0, 0, 0, 0, 0, 0, 8, 189, 0, 0, 0, 0, 0, 0, 11, 189, 0, 0, 0, 0, 0, 0, 4, 189, 0, 0, 0, 0, 0, 0, 7, 189, 0, 0, 0, 0, 0, 0, 10, 189, 0, 0, 0, 0, 0, 0, 14, 189, 0, 0, 0, 0, 0, 0, 15, 189, 0, 0, 0, 0, 0, 0, 21, 189, 0, 0, 0, 0, 0, 0, 38, 189, 0, 0, 0, 0, 0, 0, 39, 189, 0, 0, 0, 0, 0, 0, 43, 189, 0, 0, 0, 0, 0, 0, 41, 189, 0, 0, 0, 0, 0, 0, 42, 189, 0, 0, 0, 0, 0, 0, 40, 189, 0, 0, 0, 0, 0, 0, 18, 189, 0, 0, 0, 0, 0, 0, 31, 189, 0, 0, 0, 0, 0, 0, 6, 189, 0, 0, 0, 0, 0, 0, 22, 189, 0, 0, 0, 0, 0, 0, 9, 189, 0, 0, 0, 0, 0, 0, 22, 189, 0, 0, 0, 0, 0, 0, 12, 189, 0, 0, 0, 0, 0, 0, 22, 189, 0, 0, 0, 0, 0, 0, 16, 189, 0, 0, 0, 0, 0, 0, 22, 189, 0, 0, 0, 0, 0, 0, 22, 189, 0, 0, 0, 0, 0, 0, 0, 191, 0, 0, 0, 0, 0, 0, 0, 5, 0, 75, 0, 0, 6, 102, 3, 189, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 6, 102, 5, 76, 6, 102, 193, 16, 220, 102, 194, 5, 128, 232, 5, 5, 250, 102, 194, 5, 128, 232, 5, 5, 228, 102, 32, 5, 6, 102, 5, 73, 6, 102, 5, 75, 6, 102, 5, 73, 6, 102, 193, 16, 220, 102, 5, 73, 6, 102, 5, 5, 222, 102, 5, 73, 6, 102, 5, 5, 228, 102, 5, 73, 6, 102, 193, 9, 226, 102, 5, 73, 6, 102, 5, 5, 230, 102, 5, 73, 6, 102, 5, 5, 246, 102, 5, 73, 6, 102, 5, 3, 238, 102, 5, 73, 6, 102, 5, 5, 58, 102, 5, 73, 6, 102, 5, 5, 72, 102, 5, 73, 6, 102, 5, 5, 200, 102, 5, 73, 6, 102, 5, 5, 238, 102, 5, 5, 30, 102, 5, 5, 58, 102, 5, 5, 30, 102, 5, 5, 72, 102, 5, 5, 30, 102, 5, 5, 74, 102, 5, 5, 30, 102, 5, 5, 200, 102, 5, 5, 30, 102, 5, 3, 238, 102, 5, 5, 30, 102, 5, 5, 238, 102, 5, 5, 42, 102, 5, 5, 58, 102, 5, 5, 42, 102, 5, 5, 72, 102, 5, 5, 42, 102, 5, 5, 74, 102, 5, 5, 42, 102, 5, 5, 200, 102, 5, 5, 42, 102, 5, 3, 238, 102, 5, 5, 42, 102, 5, 5, 238, 102, 5, 5, 44, 102, 5, 5, 58, 102, 5, 5, 44, 102, 5, 5, 200, 102, 5, 5, 44, 102, 5, 3, 238, 102, 5, 5, 44, 102, 5, 5, 238, 102, 5, 5, 58, 102, 5, 5, 72, 102, 5, 5, 58, 102, 5, 5, 200, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 5, 74, 102, 5, 5, 58, 102, 5, 5, 74, 102, 5, 5, 72, 102, 5, 5, 74, 102, 5, 5, 200, 102, 5, 5, 134, 102, 5, 5, 58, 102, 5, 5, 134, 102, 5, 5, 72, 102, 5, 5, 134, 102, 5, 5, 74, 102, 5, 5, 134, 102, 5, 5, 200, 102, 5, 5, 146, 102, 5, 5, 72, 102, 5, 5, 146, 102, 5, 5, 200, 102, 5, 5, 148, 102, 5, 5, 58, 102, 5, 5, 148, 102, 5, 5, 72, 102, 5, 5, 148, 102, 5, 5, 74, 102, 5, 5, 148, 102, 5, 5, 200, 102, 5, 5, 158, 102, 5, 5, 72, 102, 5, 5, 158, 102, 5, 5, 200, 102, 5, 5, 160, 102, 5, 5, 200, 102, 5, 5, 166, 102, 5, 5, 58, 102, 5, 5, 166, 102, 5, 5, 200, 102, 5, 5, 168, 102, 5, 5, 58, 102, 5, 5, 168, 102, 5, 5, 200, 102, 5, 5, 174, 102, 5, 5, 58, 102, 5, 5, 174, 102, 5, 5, 72, 102, 5, 5, 174, 102, 5, 5, 74, 102, 5, 5, 174, 102, 5, 5, 200, 102, 5, 5, 174, 102, 5, 3, 238, 102, 5, 5, 174, 102, 5, 5, 238, 102, 5, 5, 178, 102, 5, 5, 72, 102, 5, 5, 178, 102, 5, 5, 200, 102, 5, 5, 178, 102, 5, 3, 238, 102, 5, 5, 178, 102, 5, 5, 238, 102, 5, 5, 182, 102, 5, 75, 6, 102, 5, 5, 182, 102, 5, 5, 58, 102, 5, 5, 182, 102, 5, 5, 72, 102, 5, 5, 182, 102, 5, 5, 74, 102, 5, 5, 182, 102, 5, 5, 196, 102, 5, 5, 182, 102, 5, 5, 200, 102, 5, 5, 182, 102, 5, 3, 238, 102, 5, 5, 182, 102, 5, 5, 238, 102, 5, 5, 196, 102, 5, 5, 58, 102, 5, 5, 196, 102, 5, 5, 72, 102, 5, 5, 196, 102, 5, 5, 74, 102, 5, 5, 196, 102, 5, 5, 200, 102, 5, 5, 196, 102, 5, 3, 238, 102, 5, 5, 196, 102, 5, 5, 238, 102, 5, 5, 200, 102, 5, 5, 58, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 3, 238, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 204, 102, 5, 5, 58, 102, 5, 5, 204, 102, 5, 5, 72, 102, 5, 5, 204, 102, 5, 5, 74, 102, 5, 5, 204, 102, 5, 5, 200, 102, 5, 5, 204, 102, 5, 3, 238, 102, 5, 5, 204, 102, 5, 5, 238, 102, 5, 5, 214, 102, 5, 5, 58, 102, 5, 5, 214, 102, 5, 5, 200, 102, 5, 5, 214, 102, 5, 3, 238, 102, 5, 5, 214, 102, 5, 5, 238, 102, 5, 5, 238, 102, 5, 5, 72, 102, 5, 5, 238, 102, 5, 5, 74, 102, 5, 5, 238, 102, 5, 3, 238, 102, 5, 5, 86, 102, 191, 0, 0, 0, 5, 5, 110, 102, 191, 0, 0, 0, 5, 3, 238, 102, 191, 0, 0, 0, 5, 73, 6, 102, 5, 5, 110, 102, 5, 73, 6, 102, 5, 5, 112, 102, 5, 73, 6, 102, 5, 5, 204, 102, 5, 5, 30, 102, 5, 5, 110, 102, 5, 5, 30, 102, 5, 5, 112, 102, 5, 5, 30, 102, 5, 5, 204, 102, 5, 5, 42, 102, 5, 5, 110, 102, 5, 5, 42, 102, 5, 5, 112, 102, 5, 5, 42, 102, 5, 5, 204, 102, 5, 5, 44, 102, 5, 5, 110, 102, 5, 5, 44, 102, 5, 5, 112, 102, 5, 5, 44, 102, 5, 5, 204, 102, 5, 5, 200, 102, 5, 75, 6, 102, 5, 5, 204, 102, 5, 5, 110, 102, 5, 5, 204, 102, 5, 5, 112, 102, 5, 5, 204, 102, 5, 5, 204, 102, 5, 5, 238, 102, 5, 5, 110, 102, 5, 5, 238, 102, 5, 5, 112, 102, 5, 73, 6, 102, 5, 5, 74, 102, 5, 73, 6, 102, 5, 5, 214, 102, 5, 5, 30, 102, 5, 5, 214, 102, 5, 5, 42, 102, 5, 5, 214, 102, 5, 5, 146, 102, 5, 5, 74, 102, 5, 5, 196, 102, 5, 5, 214, 102, 5, 5, 204, 102, 5, 5, 214, 102, 5, 5, 214, 102, 191, 0, 0, 0, 5, 5, 44, 102, 5, 5, 214, 102, 5, 5, 134, 102, 5, 5, 214, 102, 5, 5, 136, 102, 5, 5, 200, 102, 5, 5, 136, 102, 5, 5, 214, 102, 5, 5, 158, 102, 5, 3, 238, 102, 5, 5, 158, 102, 5, 5, 238, 102, 5, 5, 166, 102, 5, 3, 238, 102, 5, 5, 166, 102, 5, 5, 238, 102, 5, 5, 168, 102, 5, 3, 238, 102, 5, 5, 168, 102, 5, 5, 238, 102, 5, 5, 134, 102, 5, 3, 238, 102, 5, 5, 134, 102, 5, 5, 238, 102, 5, 5, 136, 102, 5, 3, 238, 102, 5, 5, 136, 102, 5, 5, 238, 102, 5, 5, 72, 102, 5, 3, 238, 102, 5, 5, 58, 102, 5, 3, 238, 102, 5, 5, 58, 102, 5, 5, 238, 102, 5, 5, 74, 102, 5, 3, 238, 102, 5, 5, 146, 102, 5, 3, 238, 102, 5, 5, 146, 102, 5, 5, 238, 102, 5, 5, 148, 102, 5, 3, 238, 102, 5, 5, 148, 102, 5, 5, 238, 102, 5, 5, 136, 102, 5, 5, 58, 102, 5, 5, 136, 102, 5, 5, 72, 102, 5, 5, 136, 102, 5, 5, 74, 102, 5, 5, 136, 102, 5, 5, 110, 102, 5, 5, 134, 102, 5, 5, 110, 102, 5, 5, 146, 102, 5, 5, 110, 102, 5, 5, 148, 102, 5, 5, 110, 102, 5, 5, 42, 102, 5, 5, 58, 102, 5, 5, 200, 102, 5, 5, 42, 102, 5, 5, 72, 102, 5, 5, 58, 102, 5, 5, 42, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 5, 42, 102, 5, 5, 74, 102, 5, 5, 200, 102, 5, 5, 42, 102, 5, 5, 200, 102, 5, 5, 58, 102, 5, 5, 42, 102, 5, 5, 200, 102, 5, 5, 72, 102, 5, 5, 42, 102, 5, 5, 200, 102, 5, 5, 74, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 3, 238, 102, 5, 5, 134, 102, 5, 5, 72, 102, 5, 5, 58, 102, 5, 5, 134, 102, 5, 5, 58, 102, 5, 5, 72, 102, 5, 5, 134, 102, 5, 5, 58, 102, 5, 3, 238, 102, 5, 5, 134, 102, 5, 5, 200, 102, 5, 5, 72, 102, 5, 5, 134, 102, 5, 5, 200, 102, 5, 5, 58, 102, 5, 5, 134, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 146, 102, 5, 5, 72, 102, 5, 5, 72, 102, 5, 5, 146, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 136, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 5, 136, 102, 5, 5, 58, 102, 5, 5, 238, 102, 5, 5, 136, 102, 5, 5, 200, 102, 5, 5, 74, 102, 5, 5, 136, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 148, 102, 5, 5, 72, 102, 5, 3, 238, 102, 5, 5, 148, 102, 5, 5, 74, 102, 5, 5, 200, 102, 5, 5, 158, 102, 5, 5, 200, 102, 5, 5, 72, 102, 5, 5, 158, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 158, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 166, 102, 5, 5, 58, 102, 5, 5, 200, 102, 5, 5, 166, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 166, 102, 5, 5, 200, 102, 5, 3, 238, 102, 5, 5, 168, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 168, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 168, 102, 5, 5, 200, 102, 5, 3, 238, 102, 5, 5, 174, 102, 5, 5, 74, 102, 5, 5, 200, 102, 5, 5, 178, 102, 5, 5, 200, 102, 5, 5, 72, 102, 5, 5, 178, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 196, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 5, 196, 102, 5, 5, 72, 102, 5, 5, 238, 102, 5, 5, 196, 102, 5, 5, 72, 102, 5, 3, 238, 102, 5, 5, 196, 102, 5, 5, 58, 102, 5, 5, 58, 102, 5, 5, 196, 102, 5, 5, 74, 102, 5, 5, 200, 102, 5, 5, 196, 102, 5, 5, 200, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 5, 72, 102, 5, 5, 58, 102, 5, 5, 200, 102, 5, 5, 72, 102, 5, 5, 238, 102, 5, 5, 200, 102, 5, 5, 58, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 5, 74, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 58, 102, 5, 5, 74, 102, 5, 5, 214, 102, 5, 5, 200, 102, 5, 5, 58, 102, 5, 5, 214, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 204, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 5, 204, 102, 5, 5, 72, 102, 5, 3, 238, 102, 5, 5, 204, 102, 5, 5, 58, 102, 5, 5, 200, 102, 5, 5, 204, 102, 5, 5, 58, 102, 5, 3, 238, 102, 5, 5, 204, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 204, 102, 5, 5, 200, 102, 5, 3, 238, 102, 5, 5, 238, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 30, 102, 5, 5, 74, 102, 5, 5, 238, 102, 5, 5, 42, 102, 5, 5, 58, 102, 5, 5, 238, 102, 5, 5, 42, 102, 5, 5, 58, 102, 5, 3, 238, 102, 5, 5, 42, 102, 5, 5, 74, 102, 5, 5, 238, 102, 5, 5, 42, 102, 5, 5, 74, 102, 5, 3, 238, 102, 5, 5, 42, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 42, 102, 5, 5, 200, 102, 5, 3, 238, 102, 5, 5, 58, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 58, 102, 5, 5, 72, 102, 5, 3, 238, 102, 5, 5, 58, 102, 5, 5, 200, 102, 5, 3, 238, 102, 5, 5, 134, 102, 5, 5, 74, 102, 5, 3, 238, 102, 5, 5, 146, 102, 5, 5, 72, 102, 5, 5, 238, 102, 5, 5, 136, 102, 5, 5, 72, 102, 5, 5, 238, 102, 5, 5, 148, 102, 5, 5, 72, 102, 5, 5, 238, 102, 5, 5, 196, 102, 5, 5, 58, 102, 5, 5, 238, 102, 5, 5, 196, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 238, 102, 5, 5, 58, 102, 5, 5, 238, 102, 5, 5, 238, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 178, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 204, 102, 5, 5, 72, 102, 5, 5, 238, 102, 5, 5, 166, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 182, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 204, 102, 5, 5, 58, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 5, 74, 102, 5, 5, 238, 102, 5, 5, 196, 102, 5, 5, 58, 102, 5, 5, 200, 102, 5, 5, 182, 102, 5, 5, 200, 102, 5, 5, 200, 102, 5, 5, 58, 102, 5, 5, 72, 102, 5, 5, 238, 102, 5, 5, 72, 102, 5, 5, 58, 102, 5, 5, 238, 102, 5, 5, 200, 102, 5, 5, 58, 102, 5, 5, 238, 102, 5, 5, 174, 102, 5, 5, 200, 102, 5, 5, 238, 102, 5, 5, 30, 102, 5, 5, 72, 102, 5, 5, 238, 102, 5, 5, 134, 102, 5, 5, 74, 102, 5, 5, 238, 102, 5, 5, 204, 102, 5, 5, 58, 102, 5, 5, 238, 102, 5, 5, 146, 102, 5, 5, 196, 102, 7, 5, 250, 102, 5, 5, 178, 102, 5, 5, 196, 102, 6, 5, 250, 102, 5, 75, 6, 102, 5, 5, 196, 102, 5, 5, 196, 102, 8, 5, 214, 102, 5, 75, 6, 102, 5, 5, 182, 102, 5, 5, 30, 102, 6, 5, 110, 102, 5, 5, 200, 102, 5, 5, 72, 102, 5, 5, 200, 102, 6, 5, 84, 102, 5, 5, 146, 102, 5, 5, 196, 102, 5, 5, 166, 102, 8, 5, 200, 102, 5, 5, 110, 102, 5, 5, 134, 102, 5, 5, 222, 102, 7, 5, 196, 102, 5, 5, 166, 102, 5, 5, 196, 102, 5, 5, 238, 102, 6, 5, 214, 102, 5, 5, 222, 102, 5, 5, 134, 102, 5, 5, 196, 102, 6, 5, 200, 102, 5, 5, 146, 102, 5, 5, 196, 102, 6, 3, 238, 102, 5, 5, 146, 102, 5, 5, 196, 102, 5, 3, 238, 102, 5, 5, 0, 4, 5, 75, 6, 102, 5, 5, 196, 102, 5, 5, 196, 102, 5, 5, 214, 102, 5, 5, 0, 4, 5, 5, 166, 102, 5, 5, 196, 102, 5, 5, 238, 102, 5, 5, 214, 102, 5, 5, 0, 4, 5, 5, 222, 102, 5, 5, 134, 102, 5, 5, 196, 102, 7, 5, 200, 102, 5, 5, 58, 102, 5, 5, 196, 102, 5, 5, 0, 4, 5, 5, 58, 102, 5, 5, 196, 102, 5, 75, 6, 102, 5, 5, 196, 102, 7, 5, 214, 102, 5, 5, 110, 102, 5, 2, 238, 102, 5, 75, 6, 102, 6, 5, 196, 102, 5, 5, 30, 102, 5, 5, 134, 102, 5, 5, 200, 102, 5, 5, 0, 4, 5, 75, 6, 102, 5, 5, 196, 102, 5, 5, 196, 102, 5, 5, 214, 102, 5, 5, 0, 4, 5, 75, 6, 102, 5, 5, 196, 102, 5, 5, 110, 102, 5, 5, 72, 102, 5, 5, 200, 102, 5, 5, 204, 102, 5, 5, 0, 4, 5, 75, 6, 102, 5, 5, 196, 102, 5, 5, 110, 102, 5, 5, 72, 102, 5, 5, 238, 102, 9, 5, 200, 102, 5, 5, 196, 102, 5, 5, 6, 102, 5, 76, 6, 102, 5, 5, 196, 102, 5, 70, 6, 102, 5, 5, 196, 102, 5, 72, 6, 102, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 6, 102, 5, 75, 2, 0, 83, 6, 197, 162, 84, 6, 255, 255, 6, 102, 5, 70, 85, 6, 255, 255, 6, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 6, 102, 5, 71, 238, 102, 5, 3, 48, 0, 85, 6, 255, 255, 6, 102, 5, 74, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 6, 102, 5, 73, 3, 0, 67, 0, 58, 0, 32, 0, 79, 15, 0, 16, 64, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 73, 15, 0, 32, 165, 28, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 170, 28, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 14, 0, 0, 0, 14, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 14, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 97, 0, 114, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 245, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 249, 6, 0, 0, 249, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 80, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 109, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 148, 9, 60, 0, 130, 9, 60, 0, 129, 9, 60, 0, 131, 9, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 164, 9, 60, 0, 206, 9, 61, 0, 164, 9, 205, 9, 13, 32, 38, 0, 185, 9, 60, 0, 149, 9, 205, 9, 183, 9, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 102, 26, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 128, 0, 0, 0, 128, 1, 0, 0, 224, 25, 0, 0, 224, 25, 0, 0, 224, 25, 0, 0, 224, 25, 0, 0, 36, 26, 0, 0, 36, 26, 0, 0, 68, 26, 0, 0, 78, 26, 0, 0, 4, 0, 0, 0, 10, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 105, 190, 255, 0, 106, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 39, 41, 42, 43, 44, 45, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 125, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 103, 0, 108, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 3, 35, 105, 193, 2, 35, 105, 193, 4, 35, 105, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 2, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 2, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 105, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 36, 105, 5, 5, 49, 0, 205, 9, 183, 9, 255, 255, 103, 105, 5, 5, 66, 105, 5, 5, 49, 0, 205, 9, 13, 32, 255, 255, 65, 105, 5, 5, 4, 0, 183, 9, 184, 9, 13, 32, 14, 32, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 83, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 238, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 243, 6, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 20, 22, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 24, 22, 0, 0, 24, 22, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 123, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 65, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 97, 0, 122, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 111, 0, 116, 0, 104, 0, 101, 0, 114, 0, 115, 0, 93, 0, 0, 0, 125, 220, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 76, 0, 97, 0, 116, 0, 110, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 93, 0, 38, 0, 67, 0, 60, 0, 99, 0, 39, 3, 60, 0, 60, 0, 60, 0, 67, 0, 39, 3, 38, 0, 71, 0, 60, 0, 103, 0, 6, 3, 60, 0, 60, 0, 60, 0, 71, 0, 6, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 105, 0, 60, 0, 49, 1, 60, 0, 60, 0, 60, 0, 73, 0, 38, 0, 105, 0, 60, 0, 60, 0, 60, 0, 48, 1, 38, 0, 79, 0, 60, 0, 111, 0, 8, 3, 60, 0, 60, 0, 60, 0, 79, 0, 8, 3, 38, 0, 83, 0, 60, 0, 115, 0, 39, 3, 60, 0, 60, 0, 60, 0, 83, 0, 39, 3, 38, 0, 85, 0, 60, 0, 117, 0, 8, 3, 60, 0, 60, 0, 60, 0, 85, 0, 8, 3, 38, 0, 75, 0, 60, 0, 113, 0, 60, 0, 60, 0, 60, 0, 81, 0, 38, 0, 69, 0, 60, 0, 89, 2, 60, 0, 60, 0, 60, 0, 143, 1, 38, 0, 72, 0, 60, 0, 120, 0, 60, 0, 60, 0, 60, 0, 88, 0, 38, 0, 90, 0, 60, 0, 119, 0, 60, 0, 60, 0, 60, 0, 87, 0, 0, 0, 156, 45, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 214, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 212, 38, 0, 0, 212, 38, 0, 0, 212, 38, 0, 0, 212, 38, 0, 0, 96, 43, 0, 0, 96, 43, 0, 0, 132, 45, 0, 0, 132, 45, 0, 0, 132, 45, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 96, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 93, 0, 101, 0, 107, 0, 115, 0, 120, 0, 128, 0, 134, 0, 142, 0, 145, 0, 153, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 152, 0, 160, 0, 166, 0, 174, 0, 182, 0, 190, 0, 48, 0, 56, 0, 189, 0, 197, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 214, 0, 222, 0, 48, 0, 48, 0, 48, 0, 230, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 238, 0, 48, 0, 48, 0, 246, 0, 48, 0, 252, 0, 4, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 7, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 14, 1, 19, 1, 48, 0, 23, 1, 31, 1, 48, 0, 38, 1, 46, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 54, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 57, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 212, 0, 48, 0, 48, 0, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 87, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 65, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 73, 1, 48, 0, 48, 0, 79, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 116, 1, 172, 1, 224, 1, 24, 2, 68, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 96, 2, 152, 2, 216, 2, 192, 0, 244, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 35, 0, 202, 65, 35, 0, 202, 98, 35, 0, 202, 131, 35, 0, 202, 164, 35, 0, 202, 197, 35, 0, 202, 230, 35, 0, 202, 7, 36, 0, 202, 40, 36, 0, 202, 73, 36, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 198, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 134, 1, 0, 156, 5, 0, 56, 201, 70, 2, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 6, 3, 0, 156, 5, 0, 72, 134, 5, 2, 63, 156, 5, 0, 76, 201, 230, 3, 0, 156, 5, 0, 80, 201, 198, 4, 0, 156, 5, 0, 84, 134, 5, 2, 93, 134, 5, 2, 57, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 230, 5, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 166, 6, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 102, 7, 0, 5, 5, 0, 72, 5, 5, 2, 63, 5, 5, 0, 76, 201, 70, 8, 0, 5, 5, 0, 80, 201, 38, 9, 0, 5, 5, 0, 84, 5, 5, 2, 93, 5, 5, 2, 57, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 2, 26, 0, 134, 5, 2, 47, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 197, 162, 0, 0, 197, 226, 0, 0, 197, 34, 1, 0, 197, 98, 1, 0, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 134, 5, 2, 71, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 5, 2, 83, 196, 136, 156, 90, 160, 5, 38, 93, 197, 66, 26, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 130, 26, 0, 5, 5, 2, 47, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 5, 5, 2, 71, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 5, 2, 83, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 55, 5, 5, 2, 55, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 10, 0, 201, 38, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 197, 98, 2, 0, 192, 0, 0, 0, 197, 162, 2, 0, 192, 0, 0, 0, 201, 230, 11, 0, 192, 0, 0, 0, 134, 5, 0, 58, 5, 5, 27, 57, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 12, 0, 201, 102, 13, 0, 197, 162, 3, 0, 197, 226, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 14, 0, 201, 230, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 15, 0, 201, 102, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 17, 0, 201, 230, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 7, 0, 197, 34, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 166, 18, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 20, 0, 5, 71, 238, 102, 201, 102, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 8, 0, 192, 0, 0, 0, 197, 34, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 9, 0, 192, 0, 0, 0, 197, 226, 9, 0, 192, 0, 0, 0, 197, 34, 10, 0, 192, 0, 0, 0, 197, 98, 10, 0, 192, 0, 0, 0, 197, 162, 10, 0, 197, 227, 10, 0, 197, 66, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 11, 0, 197, 195, 11, 0, 197, 34, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 98, 8, 0, 197, 162, 12, 0, 5, 5, 10, 121, 197, 226, 12, 0, 197, 34, 13, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 197, 34, 14, 0, 197, 98, 14, 0, 197, 162, 14, 0, 197, 226, 14, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 2, 9, 0, 5, 5, 24, 121, 197, 34, 9, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 15, 0, 197, 98, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 22, 0, 192, 0, 0, 0, 197, 227, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 23, 0, 201, 198, 23, 0, 201, 134, 24, 0, 201, 70, 25, 0, 201, 6, 26, 0, 201, 198, 26, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 17, 0, 197, 2, 18, 0, 197, 66, 18, 0, 197, 130, 18, 0, 197, 194, 18, 0, 197, 2, 19, 0, 197, 66, 19, 0, 197, 130, 19, 0, 197, 194, 19, 0, 197, 2, 20, 0, 197, 66, 20, 0, 197, 130, 20, 0, 197, 194, 20, 0, 197, 2, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 21, 0, 192, 0, 0, 0, 201, 134, 27, 0, 192, 0, 0, 0, 201, 70, 28, 0, 201, 6, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 29, 0, 201, 134, 30, 0, 201, 70, 31, 0, 201, 6, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 32, 0, 201, 134, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 83, 5, 136, 0, 0, 5, 5, 2, 83, 5, 136, 0, 0, 134, 5, 27, 57, 5, 138, 0, 0, 134, 5, 27, 57, 5, 136, 0, 0, 134, 5, 27, 57, 5, 142, 0, 0, 134, 5, 27, 57, 5, 150, 0, 0, 134, 5, 2, 55, 5, 160, 0, 0, 5, 5, 2, 55, 5, 160, 0, 0, 134, 5, 27, 57, 5, 154, 0, 0, 134, 5, 27, 57, 5, 164, 0, 0, 134, 5, 27, 57, 5, 140, 0, 0, 134, 5, 27, 57, 5, 162, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 134, 5, 2, 93, 5, 142, 0, 0, 5, 5, 2, 93, 5, 142, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 134, 5, 27, 57, 5, 144, 0, 0, 134, 5, 2, 83, 5, 164, 0, 0, 5, 5, 2, 83, 5, 164, 0, 0, 134, 5, 2, 83, 5, 144, 0, 0, 5, 5, 2, 83, 5, 144, 0, 0, 134, 5, 2, 83, 5, 138, 0, 0, 5, 5, 2, 83, 5, 138, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 27, 57, 5, 184, 0, 0, 134, 5, 27, 57, 5, 188, 0, 0, 134, 5, 2, 71, 5, 164, 0, 0, 5, 5, 2, 71, 5, 164, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 5, 2, 47, 5, 136, 0, 0, 5, 5, 2, 47, 5, 136, 0, 0, 134, 5, 27, 57, 5, 208, 0, 0, 134, 5, 27, 57, 5, 150, 0, 0, 5, 136, 0, 0, 134, 5, 2, 83, 5, 198, 0, 0, 5, 5, 2, 83, 5, 198, 0, 0, 134, 5, 2, 83, 5, 208, 0, 0, 5, 5, 2, 83, 5, 208, 0, 0, 134, 5, 2, 83, 5, 204, 0, 0, 5, 5, 2, 83, 5, 204, 0, 0, 134, 5, 2, 93, 5, 138, 0, 0, 5, 5, 2, 93, 5, 138, 0, 0, 134, 5, 2, 93, 5, 136, 0, 0, 5, 5, 2, 93, 5, 136, 0, 0, 134, 5, 2, 93, 5, 150, 0, 0, 5, 5, 2, 93, 5, 150, 0, 0, 134, 5, 2, 93, 5, 156, 0, 0, 5, 5, 2, 93, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 57, 5, 156, 0, 0, 5, 5, 2, 57, 5, 156, 0, 0, 134, 5, 2, 57, 5, 150, 0, 0, 5, 5, 2, 57, 5, 150, 0, 0, 5, 5, 2, 93, 5, 146, 0, 0, 134, 5, 27, 57, 5, 182, 0, 0, 134, 5, 27, 57, 5, 196, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 196, 0, 0, 5, 5, 2, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 98, 8, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 162, 8, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 226, 8, 0, 5, 5, 24, 121, 197, 34, 9, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 98, 9, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 162, 9, 0, 5, 5, 107, 121, 197, 226, 9, 0, 5, 5, 111, 121, 197, 34, 10, 0, 5, 5, 115, 121, 197, 98, 10, 0, 5, 5, 119, 121, 197, 162, 10, 0, 197, 227, 10, 0, 197, 66, 11, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 130, 11, 0, 197, 195, 11, 0, 197, 34, 12, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 98, 12, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 98, 8, 0, 197, 162, 12, 0, 5, 5, 10, 121, 197, 226, 12, 0, 197, 34, 13, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 197, 34, 14, 0, 197, 98, 14, 0, 197, 162, 14, 0, 197, 226, 14, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 2, 9, 0, 5, 5, 24, 121, 197, 34, 9, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 46, 156, 5, 48, 0, 39, 3, 2, 239, 134, 5, 0, 54, 156, 5, 48, 0, 6, 3, 2, 247, 134, 5, 27, 57, 134, 5, 48, 0, 7, 3, 0, 250, 134, 5, 0, 70, 156, 5, 48, 0, 8, 3, 255, 255, 2, 71, 134, 5, 0, 78, 156, 5, 48, 0, 39, 3, 255, 255, 2, 79, 134, 5, 0, 82, 156, 5, 1, 0, 8, 3, 255, 255, 2, 83, 134, 5, 68, 3, 197, 162, 0, 46, 5, 5, 48, 0, 39, 3, 2, 239, 5, 5, 0, 54, 5, 5, 48, 0, 6, 3, 2, 247, 5, 5, 0, 70, 5, 5, 48, 0, 8, 3, 255, 255, 2, 71, 5, 5, 0, 78, 5, 5, 48, 0, 39, 3, 255, 255, 2, 79, 5, 5, 0, 82, 5, 5, 1, 0, 8, 3, 255, 255, 2, 83, 5, 5, 68, 3, 0, 192, 197, 98, 156, 54, 196, 160, 48, 0, 6, 3, 1, 192, 197, 162, 5, 54, 196, 160, 48, 0, 6, 3, 1, 192, 197, 226, 2, 0, 197, 226, 48, 0, 7, 3, 134, 250, 196, 162, 156, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 34, 5, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 98, 156, 70, 196, 190, 48, 0, 8, 3, 4, 192, 197, 34, 5, 70, 196, 190, 48, 0, 8, 3, 4, 192, 197, 98, 156, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 162, 5, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 226, 156, 70, 196, 162, 48, 0, 8, 3, 6, 192, 197, 226, 5, 70, 196, 162, 48, 0, 8, 3, 7, 192, 197, 34, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 15, 0, 197, 162, 48, 0, 7, 3, 134, 250, 196, 208, 156, 82, 196, 198, 48, 0, 8, 3, 16, 192, 197, 66, 5, 82, 196, 198, 48, 0, 8, 3, 16, 192, 197, 130, 156, 82, 196, 208, 48, 0, 8, 3, 16, 192, 197, 194, 5, 82, 196, 208, 48, 0, 8, 3, 17, 192, 197, 2, 156, 82, 196, 204, 48, 0, 8, 3, 17, 192, 197, 66, 5, 82, 196, 204, 48, 0, 8, 3, 17, 192, 197, 130, 21, 0, 197, 194, 48, 0, 7, 3, 134, 250, 196, 196, 156, 70, 196, 196, 48, 0, 8, 3, 22, 192, 197, 2, 5, 70, 196, 196, 48, 0, 8, 3, 22, 192, 197, 66, 22, 0, 197, 131, 48, 0, 8, 3, 22, 192, 197, 227, 23, 0, 197, 67, 48, 0, 8, 3, 23, 192, 197, 163, 156, 82, 196, 196, 48, 0, 8, 3, 24, 192, 197, 2, 5, 82, 196, 196, 48, 0, 8, 3, 24, 192, 197, 66, 24, 0, 197, 131, 48, 0, 8, 3, 24, 192, 197, 227, 25, 0, 197, 67, 48, 0, 8, 3, 25, 192, 197, 163, 3, 0, 67, 0, 58, 0, 32, 0, 123, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 122, 40, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 76, 0, 0, 0, 76, 1, 0, 0, 220, 31, 0, 0, 220, 31, 0, 0, 220, 31, 0, 0, 220, 31, 0, 0, 132, 34, 0, 0, 132, 34, 0, 0, 98, 36, 0, 0, 98, 36, 0, 0, 98, 40, 0, 0, 25, 0, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 94, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 224, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 93, 0, 101, 0, 107, 0, 115, 0, 120, 0, 128, 0, 134, 0, 142, 0, 145, 0, 153, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 175, 0, 180, 0, 48, 0, 184, 0, 192, 0, 48, 0, 199, 0, 207, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 215, 0, 169, 0, 48, 0, 48, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 116, 1, 172, 1, 224, 1, 24, 2, 68, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 20, 0, 202, 33, 20, 0, 202, 66, 20, 0, 202, 99, 20, 0, 202, 132, 20, 0, 202, 165, 20, 0, 202, 198, 20, 0, 202, 231, 20, 0, 202, 8, 21, 0, 202, 41, 21, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 198, 0, 0, 156, 5, 0, 56, 201, 134, 1, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 70, 2, 0, 156, 5, 0, 72, 134, 5, 2, 63, 156, 5, 0, 76, 201, 38, 3, 0, 156, 5, 0, 80, 201, 6, 4, 0, 156, 5, 0, 84, 134, 5, 2, 93, 134, 5, 2, 57, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 38, 5, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 230, 5, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 166, 6, 0, 5, 5, 0, 72, 5, 5, 2, 63, 5, 5, 0, 76, 201, 134, 7, 0, 5, 5, 0, 80, 201, 102, 8, 0, 5, 5, 0, 84, 5, 5, 2, 93, 5, 5, 2, 57, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 66, 19, 0, 134, 5, 2, 47, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 197, 162, 0, 0, 197, 226, 0, 0, 197, 34, 1, 0, 197, 98, 1, 0, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 134, 5, 2, 71, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 5, 2, 83, 196, 136, 156, 90, 160, 5, 38, 93, 197, 130, 19, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 194, 19, 0, 5, 5, 2, 47, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 5, 5, 2, 71, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 5, 2, 83, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 55, 5, 5, 2, 55, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 9, 0, 201, 102, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 197, 98, 2, 0, 192, 0, 0, 0, 197, 162, 2, 0, 192, 0, 0, 0, 201, 38, 11, 0, 192, 0, 0, 0, 134, 5, 0, 58, 5, 5, 27, 57, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 11, 0, 201, 166, 12, 0, 197, 162, 3, 0, 197, 226, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 13, 0, 201, 38, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 14, 0, 201, 166, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 16, 0, 201, 38, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 7, 0, 197, 34, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 8, 0, 197, 162, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 17, 0, 192, 0, 0, 0, 197, 35, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 18, 0, 201, 102, 19, 0, 201, 38, 20, 0, 201, 230, 20, 0, 201, 166, 21, 0, 201, 102, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 11, 0, 197, 66, 11, 0, 197, 130, 11, 0, 197, 194, 11, 0, 197, 2, 12, 0, 197, 66, 12, 0, 197, 130, 12, 0, 197, 194, 12, 0, 197, 2, 13, 0, 197, 66, 13, 0, 197, 130, 13, 0, 197, 194, 13, 0, 197, 2, 14, 0, 197, 66, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 14, 0, 192, 0, 0, 0, 201, 38, 23, 0, 192, 0, 0, 0, 201, 230, 23, 0, 201, 166, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 25, 0, 201, 38, 26, 0, 201, 230, 26, 0, 201, 166, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 28, 0, 201, 38, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 83, 5, 136, 0, 0, 5, 5, 2, 83, 5, 136, 0, 0, 134, 5, 27, 57, 5, 138, 0, 0, 134, 5, 27, 57, 5, 136, 0, 0, 134, 5, 27, 57, 5, 142, 0, 0, 134, 5, 27, 57, 5, 150, 0, 0, 134, 5, 2, 55, 5, 160, 0, 0, 5, 5, 2, 55, 5, 160, 0, 0, 134, 5, 27, 57, 5, 154, 0, 0, 134, 5, 27, 57, 5, 164, 0, 0, 134, 5, 27, 57, 5, 140, 0, 0, 134, 5, 27, 57, 5, 162, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 134, 5, 2, 93, 5, 142, 0, 0, 5, 5, 2, 93, 5, 142, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 134, 5, 27, 57, 5, 144, 0, 0, 134, 5, 2, 83, 5, 164, 0, 0, 5, 5, 2, 83, 5, 164, 0, 0, 134, 5, 2, 83, 5, 144, 0, 0, 5, 5, 2, 83, 5, 144, 0, 0, 134, 5, 2, 83, 5, 138, 0, 0, 5, 5, 2, 83, 5, 138, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 27, 57, 5, 184, 0, 0, 134, 5, 27, 57, 5, 188, 0, 0, 134, 5, 2, 71, 5, 164, 0, 0, 5, 5, 2, 71, 5, 164, 0, 0, 134, 5, 2, 47, 5, 136, 0, 0, 5, 5, 2, 47, 5, 136, 0, 0, 134, 5, 27, 57, 5, 208, 0, 0, 134, 5, 27, 57, 5, 150, 0, 0, 5, 136, 0, 0, 134, 5, 2, 83, 5, 198, 0, 0, 5, 5, 2, 83, 5, 198, 0, 0, 134, 5, 2, 83, 5, 208, 0, 0, 5, 5, 2, 83, 5, 208, 0, 0, 134, 5, 2, 83, 5, 204, 0, 0, 5, 5, 2, 83, 5, 204, 0, 0, 134, 5, 2, 93, 5, 138, 0, 0, 5, 5, 2, 93, 5, 138, 0, 0, 134, 5, 2, 93, 5, 136, 0, 0, 5, 5, 2, 93, 5, 136, 0, 0, 134, 5, 2, 93, 5, 150, 0, 0, 5, 5, 2, 93, 5, 150, 0, 0, 134, 5, 2, 93, 5, 156, 0, 0, 5, 5, 2, 93, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 57, 5, 156, 0, 0, 5, 5, 2, 57, 5, 156, 0, 0, 134, 5, 2, 57, 5, 150, 0, 0, 5, 5, 2, 57, 5, 150, 0, 0, 5, 5, 2, 93, 5, 146, 0, 0, 134, 5, 27, 57, 5, 182, 0, 0, 134, 5, 27, 57, 5, 196, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 196, 0, 0, 5, 5, 2, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 48, 0, 39, 3, 2, 239, 134, 5, 0, 54, 156, 5, 48, 0, 6, 3, 2, 247, 134, 5, 27, 57, 134, 5, 48, 0, 7, 3, 0, 250, 134, 5, 0, 70, 156, 5, 48, 0, 8, 3, 255, 255, 2, 71, 134, 5, 0, 78, 156, 5, 48, 0, 39, 3, 255, 255, 2, 79, 134, 5, 0, 82, 156, 5, 1, 0, 8, 3, 255, 255, 2, 83, 134, 5, 68, 3, 197, 162, 0, 46, 5, 5, 48, 0, 39, 3, 2, 239, 5, 5, 0, 54, 5, 5, 48, 0, 6, 3, 2, 247, 5, 5, 0, 70, 5, 5, 48, 0, 8, 3, 255, 255, 2, 71, 5, 5, 0, 78, 5, 5, 48, 0, 39, 3, 255, 255, 2, 79, 5, 5, 0, 82, 5, 5, 1, 0, 8, 3, 255, 255, 2, 83, 5, 5, 68, 3, 0, 192, 197, 98, 156, 54, 196, 160, 48, 0, 6, 3, 1, 192, 197, 162, 5, 54, 196, 160, 48, 0, 6, 3, 1, 192, 197, 226, 2, 0, 197, 226, 48, 0, 7, 3, 134, 250, 196, 162, 156, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 34, 5, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 98, 156, 70, 196, 190, 48, 0, 8, 3, 4, 192, 197, 34, 5, 70, 196, 190, 48, 0, 8, 3, 4, 192, 197, 98, 156, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 162, 5, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 226, 156, 70, 196, 162, 48, 0, 8, 3, 6, 192, 197, 226, 5, 70, 196, 162, 48, 0, 8, 3, 7, 192, 197, 34, 8, 0, 197, 226, 48, 0, 7, 3, 134, 250, 196, 208, 156, 82, 196, 198, 48, 0, 8, 3, 9, 192, 197, 130, 5, 82, 196, 198, 48, 0, 8, 3, 9, 192, 197, 194, 156, 82, 196, 208, 48, 0, 8, 3, 10, 192, 197, 2, 5, 82, 196, 208, 48, 0, 8, 3, 10, 192, 197, 66, 156, 82, 196, 204, 48, 0, 8, 3, 10, 192, 197, 130, 5, 82, 196, 204, 48, 0, 8, 3, 10, 192, 197, 194, 15, 0, 197, 2, 48, 0, 7, 3, 134, 250, 196, 196, 156, 70, 196, 196, 48, 0, 8, 3, 15, 192, 197, 66, 5, 70, 196, 196, 48, 0, 8, 3, 15, 192, 197, 130, 15, 0, 197, 195, 48, 0, 8, 3, 16, 192, 197, 35, 16, 0, 197, 131, 48, 0, 8, 3, 16, 192, 197, 227, 156, 82, 196, 196, 48, 0, 8, 3, 17, 192, 197, 66, 5, 82, 196, 196, 48, 0, 8, 3, 17, 192, 197, 130, 17, 0, 197, 195, 48, 0, 8, 3, 18, 192, 197, 35, 18, 0, 197, 131, 48, 0, 8, 3, 18, 192, 197, 227, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 72, 186, 76, 185, 80, 28, 4, 185, 92, 30, 4, 186, 108, 185, 112, 185, 120, 185, 124, 186, 128, 32, 4, 185, 144, 185, 116, 185, 152, 34, 4, 185, 164, 36, 4, 185, 180, 185, 192, 185, 96, 185, 184, 185, 188, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 38, 4, 168, 72, 168, 76, 168, 80, 40, 4, 168, 92, 168, 104, 168, 108, 168, 112, 168, 120, 168, 124, 168, 128, 42, 4, 168, 144, 168, 116, 168, 152, 44, 4, 168, 164, 46, 4, 168, 180, 168, 192, 168, 96, 168, 184, 168, 188, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 136, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 185, 68, 186, 77, 154, 77, 250, 77, 90, 78, 185, 101, 153, 101, 249, 101, 89, 102, 217, 72, 154, 130, 185, 137, 153, 137, 249, 137, 153, 138, 185, 140, 136, 14, 217, 138, 185, 173, 153, 173, 249, 173, 185, 176, 153, 185, 185, 196, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 168, 68, 168, 77, 136, 77, 232, 77, 72, 78, 168, 105, 136, 105, 232, 105, 72, 106, 200, 72, 136, 130, 168, 137, 136, 137, 232, 137, 136, 138, 168, 140, 128, 14, 200, 138, 168, 173, 136, 173, 232, 173, 168, 176, 136, 185, 168, 196, 72, 186, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 249, 85, 232, 85, 185, 88, 168, 88, 185, 86, 168, 86, 48, 4, 50, 4, 249, 93, 232, 93, 89, 95, 72, 95, 153, 102, 136, 106, 57, 103, 40, 107, 217, 101, 200, 105, 52, 4, 8, 107, 185, 104, 168, 100, 6, 8, 8, 8, 250, 109, 232, 109, 249, 114, 232, 114, 168, 148, 153, 121, 136, 121, 249, 122, 232, 122, 25, 122, 8, 122, 121, 123, 104, 123, 89, 123, 72, 123, 154, 129, 136, 129, 250, 130, 232, 130, 26, 130, 8, 130, 10, 8, 185, 132, 168, 132, 57, 139, 40, 139, 217, 137, 200, 137, 121, 138, 104, 138, 12, 8, 14, 8, 153, 153, 136, 153, 249, 154, 232, 154, 25, 154, 8, 154, 154, 157, 136, 157, 250, 157, 232, 157, 185, 160, 168, 160, 26, 158, 8, 158, 249, 166, 232, 166, 25, 166, 8, 166, 185, 168, 168, 168, 153, 174, 136, 174, 57, 175, 40, 175, 217, 173, 200, 173, 57, 174, 40, 174, 121, 174, 104, 174, 54, 4, 56, 4, 249, 193, 232, 193, 249, 185, 232, 185, 89, 186, 153, 189, 136, 189, 185, 190, 168, 190, 25, 190, 8, 190, 232, 156, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 156, 169, 156, 200, 56, 169, 76, 187, 104, 187, 108, 170, 104, 169, 108, 169, 200, 169, 128, 217, 136, 187, 76, 200, 136, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 185, 84, 255, 5, 185, 100, 255, 5, 185, 136, 255, 5, 186, 156, 255, 5, 185, 172, 255, 5, 168, 64, 255, 5, 168, 84, 255, 5, 168, 136, 255, 5, 168, 156, 255, 5, 168, 172, 255, 5, 249, 86, 255, 5, 232, 86, 255, 5, 25, 103, 255, 5, 25, 175, 255, 5, 8, 175, 255, 1, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 235, 11, 0, 16, 73, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 227, 11, 0, 32, 11, 22, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 16, 22, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 240, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 244, 6, 0, 0, 244, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 40, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 93, 0, 38, 0, 21, 4, 60, 0, 81, 4, 60, 0, 60, 0, 60, 0, 1, 4, 38, 0, 67, 4, 60, 0, 94, 4, 60, 0, 60, 0, 60, 0, 14, 4, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 244, 26, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 15, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 60, 0, 0, 0, 64, 0, 0, 0, 64, 1, 0, 0, 96, 26, 0, 0, 96, 26, 0, 0, 96, 26, 0, 0, 96, 26, 0, 0, 164, 26, 0, 0, 164, 26, 0, 0, 220, 26, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 39, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 137, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 96, 0, 104, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 120, 0, 48, 0, 48, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 128, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 134, 5, 41, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 173, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 40, 97, 174, 5, 48, 0, 8, 3, 255, 255, 41, 97, 134, 5, 172, 97, 174, 5, 48, 0, 6, 3, 255, 255, 173, 97, 134, 5, 40, 97, 5, 5, 48, 0, 8, 3, 255, 255, 41, 97, 5, 5, 172, 97, 5, 5, 48, 0, 6, 3, 255, 255, 173, 97, 5, 5, 3, 0, 69, 0, 60, 0, 32, 0, 43, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 233, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 238, 6, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 122, 0, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 126, 0, 0, 0, 126, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 60, 1, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 8, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 32, 0, 0, 0, 36, 0, 0, 0, 36, 1, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 39, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 3, 0, 69, 0, 60, 0, 32, 0, 35, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 115, 0, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 120, 0, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 208, 32, 0, 32, 7, 0, 0, 0, 24, 0, 0, 0, 212, 32, 0, 0, 212, 32, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 113, 9, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 116, 114, 97, 100, 105, 116, 105, 111, 110, 97, 108, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 85, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 148, 9, 60, 0, 130, 9, 60, 0, 131, 9, 60, 0, 129, 9, 0, 0, 239, 223, 81, 18, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 148, 9, 60, 0, 130, 9, 60, 0, 131, 9, 60, 0, 129, 9, 60, 0, 149, 9, 205, 9, 60, 0, 150, 9, 205, 9, 60, 0, 151, 9, 205, 9, 60, 0, 152, 9, 205, 9, 60, 0, 153, 9, 205, 9, 60, 0, 154, 9, 205, 9, 60, 0, 155, 9, 205, 9, 60, 0, 156, 9, 205, 9, 60, 0, 157, 9, 205, 9, 60, 0, 158, 9, 205, 9, 60, 0, 159, 9, 205, 9, 60, 0, 160, 9, 205, 9, 60, 0, 161, 9, 205, 9, 60, 0, 162, 9, 205, 9, 60, 0, 163, 9, 205, 9, 60, 0, 206, 9, 61, 0, 164, 9, 205, 9, 13, 32, 60, 0, 60, 0, 164, 9, 205, 9, 60, 0, 165, 9, 205, 9, 60, 0, 166, 9, 205, 9, 60, 0, 167, 9, 205, 9, 60, 0, 168, 9, 205, 9, 60, 0, 170, 9, 205, 9, 60, 0, 171, 9, 205, 9, 60, 0, 172, 9, 205, 9, 60, 0, 173, 9, 205, 9, 60, 0, 174, 9, 205, 9, 60, 0, 175, 9, 205, 9, 60, 0, 176, 9, 205, 9, 60, 0, 240, 9, 205, 9, 60, 0, 178, 9, 205, 9, 60, 0, 241, 9, 205, 9, 60, 0, 182, 9, 205, 9, 60, 0, 183, 9, 205, 9, 60, 0, 184, 9, 205, 9, 60, 0, 185, 9, 205, 9, 38, 0, 149, 9, 205, 9, 133, 9, 61, 0, 149, 9, 38, 0, 149, 9, 205, 9, 134, 9, 61, 0, 149, 9, 190, 9, 38, 0, 149, 9, 205, 9, 135, 9, 61, 0, 149, 9, 191, 9, 38, 0, 149, 9, 205, 9, 136, 9, 61, 0, 149, 9, 192, 9, 38, 0, 149, 9, 205, 9, 137, 9, 61, 0, 149, 9, 193, 9, 38, 0, 149, 9, 205, 9, 138, 9, 61, 0, 149, 9, 194, 9, 38, 0, 149, 9, 205, 9, 139, 9, 61, 0, 149, 9, 195, 9, 38, 0, 149, 9, 205, 9, 224, 9, 61, 0, 149, 9, 196, 9, 38, 0, 149, 9, 205, 9, 140, 9, 61, 0, 149, 9, 226, 9, 38, 0, 149, 9, 205, 9, 225, 9, 61, 0, 149, 9, 227, 9, 38, 0, 149, 9, 205, 9, 143, 9, 61, 0, 149, 9, 199, 9, 38, 0, 149, 9, 205, 9, 144, 9, 61, 0, 149, 9, 200, 9, 38, 0, 149, 9, 205, 9, 147, 9, 61, 0, 149, 9, 203, 9, 38, 0, 149, 9, 205, 9, 148, 9, 61, 0, 149, 9, 204, 9, 38, 0, 150, 9, 205, 9, 133, 9, 61, 0, 150, 9, 38, 0, 150, 9, 205, 9, 134, 9, 61, 0, 150, 9, 190, 9, 38, 0, 150, 9, 205, 9, 135, 9, 61, 0, 150, 9, 191, 9, 38, 0, 150, 9, 205, 9, 136, 9, 61, 0, 150, 9, 192, 9, 38, 0, 150, 9, 205, 9, 137, 9, 61, 0, 150, 9, 193, 9, 38, 0, 150, 9, 205, 9, 138, 9, 61, 0, 150, 9, 194, 9, 38, 0, 150, 9, 205, 9, 139, 9, 61, 0, 150, 9, 195, 9, 38, 0, 150, 9, 205, 9, 224, 9, 61, 0, 150, 9, 196, 9, 38, 0, 150, 9, 205, 9, 140, 9, 61, 0, 150, 9, 226, 9, 38, 0, 150, 9, 205, 9, 225, 9, 61, 0, 150, 9, 227, 9, 38, 0, 150, 9, 205, 9, 143, 9, 61, 0, 150, 9, 199, 9, 38, 0, 150, 9, 205, 9, 144, 9, 61, 0, 150, 9, 200, 9, 38, 0, 150, 9, 205, 9, 147, 9, 61, 0, 150, 9, 203, 9, 38, 0, 150, 9, 205, 9, 148, 9, 61, 0, 150, 9, 204, 9, 38, 0, 151, 9, 205, 9, 133, 9, 61, 0, 151, 9, 38, 0, 151, 9, 205, 9, 134, 9, 61, 0, 151, 9, 190, 9, 38, 0, 151, 9, 205, 9, 135, 9, 61, 0, 151, 9, 191, 9, 38, 0, 151, 9, 205, 9, 136, 9, 61, 0, 151, 9, 192, 9, 38, 0, 151, 9, 205, 9, 137, 9, 61, 0, 151, 9, 193, 9, 38, 0, 151, 9, 205, 9, 138, 9, 61, 0, 151, 9, 194, 9, 38, 0, 151, 9, 205, 9, 139, 9, 61, 0, 151, 9, 195, 9, 38, 0, 151, 9, 205, 9, 224, 9, 61, 0, 151, 9, 196, 9, 38, 0, 151, 9, 205, 9, 140, 9, 61, 0, 151, 9, 226, 9, 38, 0, 151, 9, 205, 9, 225, 9, 61, 0, 151, 9, 227, 9, 38, 0, 151, 9, 205, 9, 143, 9, 61, 0, 151, 9, 199, 9, 38, 0, 151, 9, 205, 9, 144, 9, 61, 0, 151, 9, 200, 9, 38, 0, 151, 9, 205, 9, 147, 9, 61, 0, 151, 9, 203, 9, 38, 0, 151, 9, 205, 9, 148, 9, 61, 0, 151, 9, 204, 9, 38, 0, 152, 9, 205, 9, 133, 9, 61, 0, 152, 9, 38, 0, 152, 9, 205, 9, 134, 9, 61, 0, 152, 9, 190, 9, 38, 0, 152, 9, 205, 9, 135, 9, 61, 0, 152, 9, 191, 9, 38, 0, 152, 9, 205, 9, 136, 9, 61, 0, 152, 9, 192, 9, 38, 0, 152, 9, 205, 9, 137, 9, 61, 0, 152, 9, 193, 9, 38, 0, 152, 9, 205, 9, 138, 9, 61, 0, 152, 9, 194, 9, 38, 0, 152, 9, 205, 9, 139, 9, 61, 0, 152, 9, 195, 9, 38, 0, 152, 9, 205, 9, 224, 9, 61, 0, 152, 9, 196, 9, 38, 0, 152, 9, 205, 9, 140, 9, 61, 0, 152, 9, 226, 9, 38, 0, 152, 9, 205, 9, 225, 9, 61, 0, 152, 9, 227, 9, 38, 0, 152, 9, 205, 9, 143, 9, 61, 0, 152, 9, 199, 9, 38, 0, 152, 9, 205, 9, 144, 9, 61, 0, 152, 9, 200, 9, 38, 0, 152, 9, 205, 9, 147, 9, 61, 0, 152, 9, 203, 9, 38, 0, 152, 9, 205, 9, 148, 9, 61, 0, 152, 9, 204, 9, 38, 0, 153, 9, 205, 9, 133, 9, 61, 0, 153, 9, 38, 0, 153, 9, 205, 9, 134, 9, 61, 0, 153, 9, 190, 9, 38, 0, 153, 9, 205, 9, 135, 9, 61, 0, 153, 9, 191, 9, 38, 0, 153, 9, 205, 9, 136, 9, 61, 0, 153, 9, 192, 9, 38, 0, 153, 9, 205, 9, 137, 9, 61, 0, 153, 9, 193, 9, 38, 0, 153, 9, 205, 9, 138, 9, 61, 0, 153, 9, 194, 9, 38, 0, 153, 9, 205, 9, 139, 9, 61, 0, 153, 9, 195, 9, 38, 0, 153, 9, 205, 9, 224, 9, 61, 0, 153, 9, 196, 9, 38, 0, 153, 9, 205, 9, 140, 9, 61, 0, 153, 9, 226, 9, 38, 0, 153, 9, 205, 9, 225, 9, 61, 0, 153, 9, 227, 9, 38, 0, 153, 9, 205, 9, 143, 9, 61, 0, 153, 9, 199, 9, 38, 0, 153, 9, 205, 9, 144, 9, 61, 0, 153, 9, 200, 9, 38, 0, 153, 9, 205, 9, 147, 9, 61, 0, 153, 9, 203, 9, 38, 0, 153, 9, 205, 9, 148, 9, 61, 0, 153, 9, 204, 9, 38, 0, 154, 9, 205, 9, 133, 9, 61, 0, 154, 9, 38, 0, 154, 9, 205, 9, 134, 9, 61, 0, 154, 9, 190, 9, 38, 0, 154, 9, 205, 9, 135, 9, 61, 0, 154, 9, 191, 9, 38, 0, 154, 9, 205, 9, 136, 9, 61, 0, 154, 9, 192, 9, 38, 0, 154, 9, 205, 9, 137, 9, 61, 0, 154, 9, 193, 9, 38, 0, 154, 9, 205, 9, 138, 9, 61, 0, 154, 9, 194, 9, 38, 0, 154, 9, 205, 9, 139, 9, 61, 0, 154, 9, 195, 9, 38, 0, 154, 9, 205, 9, 224, 9, 61, 0, 154, 9, 196, 9, 38, 0, 154, 9, 205, 9, 140, 9, 61, 0, 154, 9, 226, 9, 38, 0, 154, 9, 205, 9, 225, 9, 61, 0, 154, 9, 227, 9, 38, 0, 154, 9, 205, 9, 143, 9, 61, 0, 154, 9, 199, 9, 38, 0, 154, 9, 205, 9, 144, 9, 61, 0, 154, 9, 200, 9, 38, 0, 154, 9, 205, 9, 147, 9, 61, 0, 154, 9, 203, 9, 38, 0, 154, 9, 205, 9, 148, 9, 61, 0, 154, 9, 204, 9, 38, 0, 155, 9, 205, 9, 133, 9, 61, 0, 155, 9, 38, 0, 155, 9, 205, 9, 134, 9, 61, 0, 155, 9, 190, 9, 38, 0, 155, 9, 205, 9, 135, 9, 61, 0, 155, 9, 191, 9, 38, 0, 155, 9, 205, 9, 136, 9, 61, 0, 155, 9, 192, 9, 38, 0, 155, 9, 205, 9, 137, 9, 61, 0, 155, 9, 193, 9, 38, 0, 155, 9, 205, 9, 138, 9, 61, 0, 155, 9, 194, 9, 38, 0, 155, 9, 205, 9, 139, 9, 61, 0, 155, 9, 195, 9, 38, 0, 155, 9, 205, 9, 224, 9, 61, 0, 155, 9, 196, 9, 38, 0, 155, 9, 205, 9, 140, 9, 61, 0, 155, 9, 226, 9, 38, 0, 155, 9, 205, 9, 225, 9, 61, 0, 155, 9, 227, 9, 38, 0, 155, 9, 205, 9, 143, 9, 61, 0, 155, 9, 199, 9, 38, 0, 155, 9, 205, 9, 144, 9, 61, 0, 155, 9, 200, 9, 38, 0, 155, 9, 205, 9, 147, 9, 61, 0, 155, 9, 203, 9, 38, 0, 155, 9, 205, 9, 148, 9, 61, 0, 155, 9, 204, 9, 38, 0, 156, 9, 205, 9, 133, 9, 61, 0, 156, 9, 38, 0, 156, 9, 205, 9, 134, 9, 61, 0, 156, 9, 190, 9, 38, 0, 156, 9, 205, 9, 135, 9, 61, 0, 156, 9, 191, 9, 38, 0, 156, 9, 205, 9, 136, 9, 61, 0, 156, 9, 192, 9, 38, 0, 156, 9, 205, 9, 137, 9, 61, 0, 156, 9, 193, 9, 38, 0, 156, 9, 205, 9, 138, 9, 61, 0, 156, 9, 194, 9, 38, 0, 156, 9, 205, 9, 139, 9, 61, 0, 156, 9, 195, 9, 38, 0, 156, 9, 205, 9, 224, 9, 61, 0, 156, 9, 196, 9, 38, 0, 156, 9, 205, 9, 140, 9, 61, 0, 156, 9, 226, 9, 38, 0, 156, 9, 205, 9, 225, 9, 61, 0, 156, 9, 227, 9, 38, 0, 156, 9, 205, 9, 143, 9, 61, 0, 156, 9, 199, 9, 38, 0, 156, 9, 205, 9, 144, 9, 61, 0, 156, 9, 200, 9, 38, 0, 156, 9, 205, 9, 147, 9, 61, 0, 156, 9, 203, 9, 38, 0, 156, 9, 205, 9, 148, 9, 61, 0, 156, 9, 204, 9, 38, 0, 157, 9, 205, 9, 133, 9, 61, 0, 157, 9, 38, 0, 157, 9, 205, 9, 134, 9, 61, 0, 157, 9, 190, 9, 38, 0, 157, 9, 205, 9, 135, 9, 61, 0, 157, 9, 191, 9, 38, 0, 157, 9, 205, 9, 136, 9, 61, 0, 157, 9, 192, 9, 38, 0, 157, 9, 205, 9, 137, 9, 61, 0, 157, 9, 193, 9, 38, 0, 157, 9, 205, 9, 138, 9, 61, 0, 157, 9, 194, 9, 38, 0, 157, 9, 205, 9, 139, 9, 61, 0, 157, 9, 195, 9, 38, 0, 157, 9, 205, 9, 224, 9, 61, 0, 157, 9, 196, 9, 38, 0, 157, 9, 205, 9, 140, 9, 61, 0, 157, 9, 226, 9, 38, 0, 157, 9, 205, 9, 225, 9, 61, 0, 157, 9, 227, 9, 38, 0, 157, 9, 205, 9, 143, 9, 61, 0, 157, 9, 199, 9, 38, 0, 157, 9, 205, 9, 144, 9, 61, 0, 157, 9, 200, 9, 38, 0, 157, 9, 205, 9, 147, 9, 61, 0, 157, 9, 203, 9, 38, 0, 157, 9, 205, 9, 148, 9, 61, 0, 157, 9, 204, 9, 38, 0, 158, 9, 205, 9, 133, 9, 61, 0, 158, 9, 38, 0, 158, 9, 205, 9, 134, 9, 61, 0, 158, 9, 190, 9, 38, 0, 158, 9, 205, 9, 135, 9, 61, 0, 158, 9, 191, 9, 38, 0, 158, 9, 205, 9, 136, 9, 61, 0, 158, 9, 192, 9, 38, 0, 158, 9, 205, 9, 137, 9, 61, 0, 158, 9, 193, 9, 38, 0, 158, 9, 205, 9, 138, 9, 61, 0, 158, 9, 194, 9, 38, 0, 158, 9, 205, 9, 139, 9, 61, 0, 158, 9, 195, 9, 38, 0, 158, 9, 205, 9, 224, 9, 61, 0, 158, 9, 196, 9, 38, 0, 158, 9, 205, 9, 140, 9, 61, 0, 158, 9, 226, 9, 38, 0, 158, 9, 205, 9, 225, 9, 61, 0, 158, 9, 227, 9, 38, 0, 158, 9, 205, 9, 143, 9, 61, 0, 158, 9, 199, 9, 38, 0, 158, 9, 205, 9, 144, 9, 61, 0, 158, 9, 200, 9, 38, 0, 158, 9, 205, 9, 147, 9, 61, 0, 158, 9, 203, 9, 38, 0, 158, 9, 205, 9, 148, 9, 61, 0, 158, 9, 204, 9, 38, 0, 159, 9, 205, 9, 133, 9, 61, 0, 159, 9, 38, 0, 159, 9, 205, 9, 134, 9, 61, 0, 159, 9, 190, 9, 38, 0, 159, 9, 205, 9, 135, 9, 61, 0, 159, 9, 191, 9, 38, 0, 159, 9, 205, 9, 136, 9, 61, 0, 159, 9, 192, 9, 38, 0, 159, 9, 205, 9, 137, 9, 61, 0, 159, 9, 193, 9, 38, 0, 159, 9, 205, 9, 138, 9, 61, 0, 159, 9, 194, 9, 38, 0, 159, 9, 205, 9, 139, 9, 61, 0, 159, 9, 195, 9, 38, 0, 159, 9, 205, 9, 224, 9, 61, 0, 159, 9, 196, 9, 38, 0, 159, 9, 205, 9, 140, 9, 61, 0, 159, 9, 226, 9, 38, 0, 159, 9, 205, 9, 225, 9, 61, 0, 159, 9, 227, 9, 38, 0, 159, 9, 205, 9, 143, 9, 61, 0, 159, 9, 199, 9, 38, 0, 159, 9, 205, 9, 144, 9, 61, 0, 159, 9, 200, 9, 38, 0, 159, 9, 205, 9, 147, 9, 61, 0, 159, 9, 203, 9, 38, 0, 159, 9, 205, 9, 148, 9, 61, 0, 159, 9, 204, 9, 38, 0, 160, 9, 205, 9, 133, 9, 61, 0, 160, 9, 38, 0, 160, 9, 205, 9, 134, 9, 61, 0, 160, 9, 190, 9, 38, 0, 160, 9, 205, 9, 135, 9, 61, 0, 160, 9, 191, 9, 38, 0, 160, 9, 205, 9, 136, 9, 61, 0, 160, 9, 192, 9, 38, 0, 160, 9, 205, 9, 137, 9, 61, 0, 160, 9, 193, 9, 38, 0, 160, 9, 205, 9, 138, 9, 61, 0, 160, 9, 194, 9, 38, 0, 160, 9, 205, 9, 139, 9, 61, 0, 160, 9, 195, 9, 38, 0, 160, 9, 205, 9, 224, 9, 61, 0, 160, 9, 196, 9, 38, 0, 160, 9, 205, 9, 140, 9, 61, 0, 160, 9, 226, 9, 38, 0, 160, 9, 205, 9, 225, 9, 61, 0, 160, 9, 227, 9, 38, 0, 160, 9, 205, 9, 143, 9, 61, 0, 160, 9, 199, 9, 38, 0, 160, 9, 205, 9, 144, 9, 61, 0, 160, 9, 200, 9, 38, 0, 160, 9, 205, 9, 147, 9, 61, 0, 160, 9, 203, 9, 38, 0, 160, 9, 205, 9, 148, 9, 61, 0, 160, 9, 204, 9, 38, 0, 161, 9, 205, 9, 133, 9, 61, 0, 161, 9, 38, 0, 161, 9, 205, 9, 134, 9, 61, 0, 161, 9, 190, 9, 38, 0, 161, 9, 205, 9, 135, 9, 61, 0, 161, 9, 191, 9, 38, 0, 161, 9, 205, 9, 136, 9, 61, 0, 161, 9, 192, 9, 38, 0, 161, 9, 205, 9, 137, 9, 61, 0, 161, 9, 193, 9, 38, 0, 161, 9, 205, 9, 138, 9, 61, 0, 161, 9, 194, 9, 38, 0, 161, 9, 205, 9, 139, 9, 61, 0, 161, 9, 195, 9, 38, 0, 161, 9, 205, 9, 224, 9, 61, 0, 161, 9, 196, 9, 38, 0, 161, 9, 205, 9, 140, 9, 61, 0, 161, 9, 226, 9, 38, 0, 161, 9, 205, 9, 225, 9, 61, 0, 161, 9, 227, 9, 38, 0, 161, 9, 205, 9, 143, 9, 61, 0, 161, 9, 199, 9, 38, 0, 161, 9, 205, 9, 144, 9, 61, 0, 161, 9, 200, 9, 38, 0, 161, 9, 205, 9, 147, 9, 61, 0, 161, 9, 203, 9, 38, 0, 161, 9, 205, 9, 148, 9, 61, 0, 161, 9, 204, 9, 38, 0, 162, 9, 205, 9, 133, 9, 61, 0, 162, 9, 38, 0, 162, 9, 205, 9, 134, 9, 61, 0, 162, 9, 190, 9, 38, 0, 162, 9, 205, 9, 135, 9, 61, 0, 162, 9, 191, 9, 38, 0, 162, 9, 205, 9, 136, 9, 61, 0, 162, 9, 192, 9, 38, 0, 162, 9, 205, 9, 137, 9, 61, 0, 162, 9, 193, 9, 38, 0, 162, 9, 205, 9, 138, 9, 61, 0, 162, 9, 194, 9, 38, 0, 162, 9, 205, 9, 139, 9, 61, 0, 162, 9, 195, 9, 38, 0, 162, 9, 205, 9, 224, 9, 61, 0, 162, 9, 196, 9, 38, 0, 162, 9, 205, 9, 140, 9, 61, 0, 162, 9, 226, 9, 38, 0, 162, 9, 205, 9, 225, 9, 61, 0, 162, 9, 227, 9, 38, 0, 162, 9, 205, 9, 143, 9, 61, 0, 162, 9, 199, 9, 38, 0, 162, 9, 205, 9, 144, 9, 61, 0, 162, 9, 200, 9, 38, 0, 162, 9, 205, 9, 147, 9, 61, 0, 162, 9, 203, 9, 38, 0, 162, 9, 205, 9, 148, 9, 61, 0, 162, 9, 204, 9, 38, 0, 163, 9, 205, 9, 133, 9, 61, 0, 163, 9, 38, 0, 163, 9, 205, 9, 134, 9, 61, 0, 163, 9, 190, 9, 38, 0, 163, 9, 205, 9, 135, 9, 61, 0, 163, 9, 191, 9, 38, 0, 163, 9, 205, 9, 136, 9, 61, 0, 163, 9, 192, 9, 38, 0, 163, 9, 205, 9, 137, 9, 61, 0, 163, 9, 193, 9, 38, 0, 163, 9, 205, 9, 138, 9, 61, 0, 163, 9, 194, 9, 38, 0, 163, 9, 205, 9, 139, 9, 61, 0, 163, 9, 195, 9, 38, 0, 163, 9, 205, 9, 224, 9, 61, 0, 163, 9, 196, 9, 38, 0, 163, 9, 205, 9, 140, 9, 61, 0, 163, 9, 226, 9, 38, 0, 163, 9, 205, 9, 225, 9, 61, 0, 163, 9, 227, 9, 38, 0, 163, 9, 205, 9, 143, 9, 61, 0, 163, 9, 199, 9, 38, 0, 163, 9, 205, 9, 144, 9, 61, 0, 163, 9, 200, 9, 38, 0, 163, 9, 205, 9, 147, 9, 61, 0, 163, 9, 203, 9, 38, 0, 163, 9, 205, 9, 148, 9, 61, 0, 163, 9, 204, 9, 38, 0, 164, 9, 205, 9, 133, 9, 61, 0, 164, 9, 38, 0, 164, 9, 205, 9, 134, 9, 61, 0, 164, 9, 190, 9, 38, 0, 164, 9, 205, 9, 135, 9, 61, 0, 164, 9, 191, 9, 38, 0, 164, 9, 205, 9, 136, 9, 61, 0, 164, 9, 192, 9, 38, 0, 164, 9, 205, 9, 137, 9, 61, 0, 164, 9, 193, 9, 38, 0, 164, 9, 205, 9, 138, 9, 61, 0, 164, 9, 194, 9, 38, 0, 164, 9, 205, 9, 139, 9, 61, 0, 164, 9, 195, 9, 38, 0, 164, 9, 205, 9, 224, 9, 61, 0, 164, 9, 196, 9, 38, 0, 164, 9, 205, 9, 140, 9, 61, 0, 164, 9, 226, 9, 38, 0, 164, 9, 205, 9, 225, 9, 61, 0, 164, 9, 227, 9, 38, 0, 164, 9, 205, 9, 143, 9, 61, 0, 164, 9, 199, 9, 38, 0, 164, 9, 205, 9, 144, 9, 61, 0, 164, 9, 200, 9, 38, 0, 164, 9, 205, 9, 147, 9, 61, 0, 164, 9, 203, 9, 38, 0, 164, 9, 205, 9, 148, 9, 61, 0, 164, 9, 204, 9, 38, 0, 165, 9, 205, 9, 133, 9, 61, 0, 165, 9, 38, 0, 165, 9, 205, 9, 134, 9, 61, 0, 165, 9, 190, 9, 38, 0, 165, 9, 205, 9, 135, 9, 61, 0, 165, 9, 191, 9, 38, 0, 165, 9, 205, 9, 136, 9, 61, 0, 165, 9, 192, 9, 38, 0, 165, 9, 205, 9, 137, 9, 61, 0, 165, 9, 193, 9, 38, 0, 165, 9, 205, 9, 138, 9, 61, 0, 165, 9, 194, 9, 38, 0, 165, 9, 205, 9, 139, 9, 61, 0, 165, 9, 195, 9, 38, 0, 165, 9, 205, 9, 224, 9, 61, 0, 165, 9, 196, 9, 38, 0, 165, 9, 205, 9, 140, 9, 61, 0, 165, 9, 226, 9, 38, 0, 165, 9, 205, 9, 225, 9, 61, 0, 165, 9, 227, 9, 38, 0, 165, 9, 205, 9, 143, 9, 61, 0, 165, 9, 199, 9, 38, 0, 165, 9, 205, 9, 144, 9, 61, 0, 165, 9, 200, 9, 38, 0, 165, 9, 205, 9, 147, 9, 61, 0, 165, 9, 203, 9, 38, 0, 165, 9, 205, 9, 148, 9, 61, 0, 165, 9, 204, 9, 38, 0, 166, 9, 205, 9, 133, 9, 61, 0, 166, 9, 38, 0, 166, 9, 205, 9, 134, 9, 61, 0, 166, 9, 190, 9, 38, 0, 166, 9, 205, 9, 135, 9, 61, 0, 166, 9, 191, 9, 38, 0, 166, 9, 205, 9, 136, 9, 61, 0, 166, 9, 192, 9, 38, 0, 166, 9, 205, 9, 137, 9, 61, 0, 166, 9, 193, 9, 38, 0, 166, 9, 205, 9, 138, 9, 61, 0, 166, 9, 194, 9, 38, 0, 166, 9, 205, 9, 139, 9, 61, 0, 166, 9, 195, 9, 38, 0, 166, 9, 205, 9, 224, 9, 61, 0, 166, 9, 196, 9, 38, 0, 166, 9, 205, 9, 140, 9, 61, 0, 166, 9, 226, 9, 38, 0, 166, 9, 205, 9, 225, 9, 61, 0, 166, 9, 227, 9, 38, 0, 166, 9, 205, 9, 143, 9, 61, 0, 166, 9, 199, 9, 38, 0, 166, 9, 205, 9, 144, 9, 61, 0, 166, 9, 200, 9, 38, 0, 166, 9, 205, 9, 147, 9, 61, 0, 166, 9, 203, 9, 38, 0, 166, 9, 205, 9, 148, 9, 61, 0, 166, 9, 204, 9, 38, 0, 167, 9, 205, 9, 133, 9, 61, 0, 167, 9, 38, 0, 167, 9, 205, 9, 134, 9, 61, 0, 167, 9, 190, 9, 38, 0, 167, 9, 205, 9, 135, 9, 61, 0, 167, 9, 191, 9, 38, 0, 167, 9, 205, 9, 136, 9, 61, 0, 167, 9, 192, 9, 38, 0, 167, 9, 205, 9, 137, 9, 61, 0, 167, 9, 193, 9, 38, 0, 167, 9, 205, 9, 138, 9, 61, 0, 167, 9, 194, 9, 38, 0, 167, 9, 205, 9, 139, 9, 61, 0, 167, 9, 195, 9, 38, 0, 167, 9, 205, 9, 224, 9, 61, 0, 167, 9, 196, 9, 38, 0, 167, 9, 205, 9, 140, 9, 61, 0, 167, 9, 226, 9, 38, 0, 167, 9, 205, 9, 225, 9, 61, 0, 167, 9, 227, 9, 38, 0, 167, 9, 205, 9, 143, 9, 61, 0, 167, 9, 199, 9, 38, 0, 167, 9, 205, 9, 144, 9, 61, 0, 167, 9, 200, 9, 38, 0, 167, 9, 205, 9, 147, 9, 61, 0, 167, 9, 203, 9, 38, 0, 167, 9, 205, 9, 148, 9, 61, 0, 167, 9, 204, 9, 38, 0, 168, 9, 205, 9, 133, 9, 61, 0, 168, 9, 38, 0, 168, 9, 205, 9, 134, 9, 61, 0, 168, 9, 190, 9, 38, 0, 168, 9, 205, 9, 135, 9, 61, 0, 168, 9, 191, 9, 38, 0, 168, 9, 205, 9, 136, 9, 61, 0, 168, 9, 192, 9, 38, 0, 168, 9, 205, 9, 137, 9, 61, 0, 168, 9, 193, 9, 38, 0, 168, 9, 205, 9, 138, 9, 61, 0, 168, 9, 194, 9, 38, 0, 168, 9, 205, 9, 139, 9, 61, 0, 168, 9, 195, 9, 38, 0, 168, 9, 205, 9, 224, 9, 61, 0, 168, 9, 196, 9, 38, 0, 168, 9, 205, 9, 140, 9, 61, 0, 168, 9, 226, 9, 38, 0, 168, 9, 205, 9, 225, 9, 61, 0, 168, 9, 227, 9, 38, 0, 168, 9, 205, 9, 143, 9, 61, 0, 168, 9, 199, 9, 38, 0, 168, 9, 205, 9, 144, 9, 61, 0, 168, 9, 200, 9, 38, 0, 168, 9, 205, 9, 147, 9, 61, 0, 168, 9, 203, 9, 38, 0, 168, 9, 205, 9, 148, 9, 61, 0, 168, 9, 204, 9, 38, 0, 170, 9, 205, 9, 133, 9, 61, 0, 170, 9, 38, 0, 170, 9, 205, 9, 134, 9, 61, 0, 170, 9, 190, 9, 38, 0, 170, 9, 205, 9, 135, 9, 61, 0, 170, 9, 191, 9, 38, 0, 170, 9, 205, 9, 136, 9, 61, 0, 170, 9, 192, 9, 38, 0, 170, 9, 205, 9, 137, 9, 61, 0, 170, 9, 193, 9, 38, 0, 170, 9, 205, 9, 138, 9, 61, 0, 170, 9, 194, 9, 38, 0, 170, 9, 205, 9, 139, 9, 61, 0, 170, 9, 195, 9, 38, 0, 170, 9, 205, 9, 224, 9, 61, 0, 170, 9, 196, 9, 38, 0, 170, 9, 205, 9, 140, 9, 61, 0, 170, 9, 226, 9, 38, 0, 170, 9, 205, 9, 225, 9, 61, 0, 170, 9, 227, 9, 38, 0, 170, 9, 205, 9, 143, 9, 61, 0, 170, 9, 199, 9, 38, 0, 170, 9, 205, 9, 144, 9, 61, 0, 170, 9, 200, 9, 38, 0, 170, 9, 205, 9, 147, 9, 61, 0, 170, 9, 203, 9, 38, 0, 170, 9, 205, 9, 148, 9, 61, 0, 170, 9, 204, 9, 38, 0, 171, 9, 205, 9, 133, 9, 61, 0, 171, 9, 38, 0, 171, 9, 205, 9, 134, 9, 61, 0, 171, 9, 190, 9, 38, 0, 171, 9, 205, 9, 135, 9, 61, 0, 171, 9, 191, 9, 38, 0, 171, 9, 205, 9, 136, 9, 61, 0, 171, 9, 192, 9, 38, 0, 171, 9, 205, 9, 137, 9, 61, 0, 171, 9, 193, 9, 38, 0, 171, 9, 205, 9, 138, 9, 61, 0, 171, 9, 194, 9, 38, 0, 171, 9, 205, 9, 139, 9, 61, 0, 171, 9, 195, 9, 38, 0, 171, 9, 205, 9, 224, 9, 61, 0, 171, 9, 196, 9, 38, 0, 171, 9, 205, 9, 140, 9, 61, 0, 171, 9, 226, 9, 38, 0, 171, 9, 205, 9, 225, 9, 61, 0, 171, 9, 227, 9, 38, 0, 171, 9, 205, 9, 143, 9, 61, 0, 171, 9, 199, 9, 38, 0, 171, 9, 205, 9, 144, 9, 61, 0, 171, 9, 200, 9, 38, 0, 171, 9, 205, 9, 147, 9, 61, 0, 171, 9, 203, 9, 38, 0, 171, 9, 205, 9, 148, 9, 61, 0, 171, 9, 204, 9, 38, 0, 172, 9, 205, 9, 133, 9, 61, 0, 172, 9, 38, 0, 172, 9, 205, 9, 134, 9, 61, 0, 172, 9, 190, 9, 38, 0, 172, 9, 205, 9, 135, 9, 61, 0, 172, 9, 191, 9, 38, 0, 172, 9, 205, 9, 136, 9, 61, 0, 172, 9, 192, 9, 38, 0, 172, 9, 205, 9, 137, 9, 61, 0, 172, 9, 193, 9, 38, 0, 172, 9, 205, 9, 138, 9, 61, 0, 172, 9, 194, 9, 38, 0, 172, 9, 205, 9, 139, 9, 61, 0, 172, 9, 195, 9, 38, 0, 172, 9, 205, 9, 224, 9, 61, 0, 172, 9, 196, 9, 38, 0, 172, 9, 205, 9, 140, 9, 61, 0, 172, 9, 226, 9, 38, 0, 172, 9, 205, 9, 225, 9, 61, 0, 172, 9, 227, 9, 38, 0, 172, 9, 205, 9, 143, 9, 61, 0, 172, 9, 199, 9, 38, 0, 172, 9, 205, 9, 144, 9, 61, 0, 172, 9, 200, 9, 38, 0, 172, 9, 205, 9, 147, 9, 61, 0, 172, 9, 203, 9, 38, 0, 172, 9, 205, 9, 148, 9, 61, 0, 172, 9, 204, 9, 38, 0, 173, 9, 205, 9, 133, 9, 61, 0, 173, 9, 38, 0, 173, 9, 205, 9, 134, 9, 61, 0, 173, 9, 190, 9, 38, 0, 173, 9, 205, 9, 135, 9, 61, 0, 173, 9, 191, 9, 38, 0, 173, 9, 205, 9, 136, 9, 61, 0, 173, 9, 192, 9, 38, 0, 173, 9, 205, 9, 137, 9, 61, 0, 173, 9, 193, 9, 38, 0, 173, 9, 205, 9, 138, 9, 61, 0, 173, 9, 194, 9, 38, 0, 173, 9, 205, 9, 139, 9, 61, 0, 173, 9, 195, 9, 38, 0, 173, 9, 205, 9, 224, 9, 61, 0, 173, 9, 196, 9, 38, 0, 173, 9, 205, 9, 140, 9, 61, 0, 173, 9, 226, 9, 38, 0, 173, 9, 205, 9, 225, 9, 61, 0, 173, 9, 227, 9, 38, 0, 173, 9, 205, 9, 143, 9, 61, 0, 173, 9, 199, 9, 38, 0, 173, 9, 205, 9, 144, 9, 61, 0, 173, 9, 200, 9, 38, 0, 173, 9, 205, 9, 147, 9, 61, 0, 173, 9, 203, 9, 38, 0, 173, 9, 205, 9, 148, 9, 61, 0, 173, 9, 204, 9, 38, 0, 174, 9, 205, 9, 133, 9, 61, 0, 174, 9, 38, 0, 174, 9, 205, 9, 134, 9, 61, 0, 174, 9, 190, 9, 38, 0, 174, 9, 205, 9, 135, 9, 61, 0, 174, 9, 191, 9, 38, 0, 174, 9, 205, 9, 136, 9, 61, 0, 174, 9, 192, 9, 38, 0, 174, 9, 205, 9, 137, 9, 61, 0, 174, 9, 193, 9, 38, 0, 174, 9, 205, 9, 138, 9, 61, 0, 174, 9, 194, 9, 38, 0, 174, 9, 205, 9, 139, 9, 61, 0, 174, 9, 195, 9, 38, 0, 174, 9, 205, 9, 224, 9, 61, 0, 174, 9, 196, 9, 38, 0, 174, 9, 205, 9, 140, 9, 61, 0, 174, 9, 226, 9, 38, 0, 174, 9, 205, 9, 225, 9, 61, 0, 174, 9, 227, 9, 38, 0, 174, 9, 205, 9, 143, 9, 61, 0, 174, 9, 199, 9, 38, 0, 174, 9, 205, 9, 144, 9, 61, 0, 174, 9, 200, 9, 38, 0, 174, 9, 205, 9, 147, 9, 61, 0, 174, 9, 203, 9, 38, 0, 174, 9, 205, 9, 148, 9, 61, 0, 174, 9, 204, 9, 38, 0, 175, 9, 205, 9, 133, 9, 61, 0, 175, 9, 38, 0, 175, 9, 205, 9, 134, 9, 61, 0, 175, 9, 190, 9, 38, 0, 175, 9, 205, 9, 135, 9, 61, 0, 175, 9, 191, 9, 38, 0, 175, 9, 205, 9, 136, 9, 61, 0, 175, 9, 192, 9, 38, 0, 175, 9, 205, 9, 137, 9, 61, 0, 175, 9, 193, 9, 38, 0, 175, 9, 205, 9, 138, 9, 61, 0, 175, 9, 194, 9, 38, 0, 175, 9, 205, 9, 139, 9, 61, 0, 175, 9, 195, 9, 38, 0, 175, 9, 205, 9, 224, 9, 61, 0, 175, 9, 196, 9, 38, 0, 175, 9, 205, 9, 140, 9, 61, 0, 175, 9, 226, 9, 38, 0, 175, 9, 205, 9, 225, 9, 61, 0, 175, 9, 227, 9, 38, 0, 175, 9, 205, 9, 143, 9, 61, 0, 175, 9, 199, 9, 38, 0, 175, 9, 205, 9, 144, 9, 61, 0, 175, 9, 200, 9, 38, 0, 175, 9, 205, 9, 147, 9, 61, 0, 175, 9, 203, 9, 38, 0, 175, 9, 205, 9, 148, 9, 61, 0, 175, 9, 204, 9, 38, 0, 176, 9, 205, 9, 133, 9, 61, 0, 176, 9, 38, 0, 176, 9, 205, 9, 134, 9, 61, 0, 176, 9, 190, 9, 38, 0, 176, 9, 205, 9, 135, 9, 61, 0, 176, 9, 191, 9, 38, 0, 176, 9, 205, 9, 136, 9, 61, 0, 176, 9, 192, 9, 38, 0, 176, 9, 205, 9, 137, 9, 61, 0, 176, 9, 193, 9, 38, 0, 176, 9, 205, 9, 138, 9, 61, 0, 176, 9, 194, 9, 38, 0, 176, 9, 205, 9, 139, 9, 61, 0, 176, 9, 195, 9, 38, 0, 176, 9, 205, 9, 224, 9, 61, 0, 176, 9, 196, 9, 38, 0, 176, 9, 205, 9, 140, 9, 61, 0, 176, 9, 226, 9, 38, 0, 176, 9, 205, 9, 225, 9, 61, 0, 176, 9, 227, 9, 38, 0, 176, 9, 205, 9, 143, 9, 61, 0, 176, 9, 199, 9, 38, 0, 176, 9, 205, 9, 144, 9, 61, 0, 176, 9, 200, 9, 38, 0, 176, 9, 205, 9, 147, 9, 61, 0, 176, 9, 203, 9, 38, 0, 176, 9, 205, 9, 148, 9, 61, 0, 176, 9, 204, 9, 38, 0, 240, 9, 205, 9, 133, 9, 61, 0, 240, 9, 38, 0, 240, 9, 205, 9, 134, 9, 61, 0, 240, 9, 190, 9, 38, 0, 240, 9, 205, 9, 135, 9, 61, 0, 240, 9, 191, 9, 38, 0, 240, 9, 205, 9, 136, 9, 61, 0, 240, 9, 192, 9, 38, 0, 240, 9, 205, 9, 137, 9, 61, 0, 240, 9, 193, 9, 38, 0, 240, 9, 205, 9, 138, 9, 61, 0, 240, 9, 194, 9, 38, 0, 240, 9, 205, 9, 139, 9, 61, 0, 240, 9, 195, 9, 38, 0, 240, 9, 205, 9, 224, 9, 61, 0, 240, 9, 196, 9, 38, 0, 240, 9, 205, 9, 140, 9, 61, 0, 240, 9, 226, 9, 38, 0, 240, 9, 205, 9, 225, 9, 61, 0, 240, 9, 227, 9, 38, 0, 240, 9, 205, 9, 143, 9, 61, 0, 240, 9, 199, 9, 38, 0, 240, 9, 205, 9, 144, 9, 61, 0, 240, 9, 200, 9, 38, 0, 240, 9, 205, 9, 147, 9, 61, 0, 240, 9, 203, 9, 38, 0, 240, 9, 205, 9, 148, 9, 61, 0, 240, 9, 204, 9, 38, 0, 178, 9, 205, 9, 133, 9, 61, 0, 178, 9, 38, 0, 178, 9, 205, 9, 134, 9, 61, 0, 178, 9, 190, 9, 38, 0, 178, 9, 205, 9, 135, 9, 61, 0, 178, 9, 191, 9, 38, 0, 178, 9, 205, 9, 136, 9, 61, 0, 178, 9, 192, 9, 38, 0, 178, 9, 205, 9, 137, 9, 61, 0, 178, 9, 193, 9, 38, 0, 178, 9, 205, 9, 138, 9, 61, 0, 178, 9, 194, 9, 38, 0, 178, 9, 205, 9, 139, 9, 61, 0, 178, 9, 195, 9, 38, 0, 178, 9, 205, 9, 224, 9, 61, 0, 178, 9, 196, 9, 38, 0, 178, 9, 205, 9, 140, 9, 61, 0, 178, 9, 226, 9, 38, 0, 178, 9, 205, 9, 225, 9, 61, 0, 178, 9, 227, 9, 38, 0, 178, 9, 205, 9, 143, 9, 61, 0, 178, 9, 199, 9, 38, 0, 178, 9, 205, 9, 144, 9, 61, 0, 178, 9, 200, 9, 38, 0, 178, 9, 205, 9, 147, 9, 61, 0, 178, 9, 203, 9, 38, 0, 178, 9, 205, 9, 148, 9, 61, 0, 178, 9, 204, 9, 38, 0, 241, 9, 205, 9, 133, 9, 61, 0, 241, 9, 38, 0, 241, 9, 205, 9, 134, 9, 61, 0, 241, 9, 190, 9, 38, 0, 241, 9, 205, 9, 135, 9, 61, 0, 241, 9, 191, 9, 38, 0, 241, 9, 205, 9, 136, 9, 61, 0, 241, 9, 192, 9, 38, 0, 241, 9, 205, 9, 137, 9, 61, 0, 241, 9, 193, 9, 38, 0, 241, 9, 205, 9, 138, 9, 61, 0, 241, 9, 194, 9, 38, 0, 241, 9, 205, 9, 139, 9, 61, 0, 241, 9, 195, 9, 38, 0, 241, 9, 205, 9, 224, 9, 61, 0, 241, 9, 196, 9, 38, 0, 241, 9, 205, 9, 140, 9, 61, 0, 241, 9, 226, 9, 38, 0, 241, 9, 205, 9, 225, 9, 61, 0, 241, 9, 227, 9, 38, 0, 241, 9, 205, 9, 143, 9, 61, 0, 241, 9, 199, 9, 38, 0, 241, 9, 205, 9, 144, 9, 61, 0, 241, 9, 200, 9, 38, 0, 241, 9, 205, 9, 147, 9, 61, 0, 241, 9, 203, 9, 38, 0, 241, 9, 205, 9, 148, 9, 61, 0, 241, 9, 204, 9, 38, 0, 182, 9, 205, 9, 133, 9, 61, 0, 182, 9, 38, 0, 182, 9, 205, 9, 134, 9, 61, 0, 182, 9, 190, 9, 38, 0, 182, 9, 205, 9, 135, 9, 61, 0, 182, 9, 191, 9, 38, 0, 182, 9, 205, 9, 136, 9, 61, 0, 182, 9, 192, 9, 38, 0, 182, 9, 205, 9, 137, 9, 61, 0, 182, 9, 193, 9, 38, 0, 182, 9, 205, 9, 138, 9, 61, 0, 182, 9, 194, 9, 38, 0, 182, 9, 205, 9, 139, 9, 61, 0, 182, 9, 195, 9, 38, 0, 182, 9, 205, 9, 224, 9, 61, 0, 182, 9, 196, 9, 38, 0, 182, 9, 205, 9, 140, 9, 61, 0, 182, 9, 226, 9, 38, 0, 182, 9, 205, 9, 225, 9, 61, 0, 182, 9, 227, 9, 38, 0, 182, 9, 205, 9, 143, 9, 61, 0, 182, 9, 199, 9, 38, 0, 182, 9, 205, 9, 144, 9, 61, 0, 182, 9, 200, 9, 38, 0, 182, 9, 205, 9, 147, 9, 61, 0, 182, 9, 203, 9, 38, 0, 182, 9, 205, 9, 148, 9, 61, 0, 182, 9, 204, 9, 38, 0, 183, 9, 205, 9, 133, 9, 61, 0, 183, 9, 38, 0, 183, 9, 205, 9, 134, 9, 61, 0, 183, 9, 190, 9, 38, 0, 183, 9, 205, 9, 135, 9, 61, 0, 183, 9, 191, 9, 38, 0, 183, 9, 205, 9, 136, 9, 61, 0, 183, 9, 192, 9, 38, 0, 183, 9, 205, 9, 137, 9, 61, 0, 183, 9, 193, 9, 38, 0, 183, 9, 205, 9, 138, 9, 61, 0, 183, 9, 194, 9, 38, 0, 183, 9, 205, 9, 139, 9, 61, 0, 183, 9, 195, 9, 38, 0, 183, 9, 205, 9, 224, 9, 61, 0, 183, 9, 196, 9, 38, 0, 183, 9, 205, 9, 140, 9, 61, 0, 183, 9, 226, 9, 38, 0, 183, 9, 205, 9, 225, 9, 61, 0, 183, 9, 227, 9, 38, 0, 183, 9, 205, 9, 143, 9, 61, 0, 183, 9, 199, 9, 38, 0, 183, 9, 205, 9, 144, 9, 61, 0, 183, 9, 200, 9, 38, 0, 183, 9, 205, 9, 147, 9, 61, 0, 183, 9, 203, 9, 38, 0, 183, 9, 205, 9, 148, 9, 61, 0, 183, 9, 204, 9, 38, 0, 184, 9, 205, 9, 133, 9, 61, 0, 184, 9, 38, 0, 184, 9, 205, 9, 134, 9, 61, 0, 184, 9, 190, 9, 38, 0, 184, 9, 205, 9, 135, 9, 61, 0, 184, 9, 191, 9, 38, 0, 184, 9, 205, 9, 136, 9, 61, 0, 184, 9, 192, 9, 38, 0, 184, 9, 205, 9, 137, 9, 61, 0, 184, 9, 193, 9, 38, 0, 184, 9, 205, 9, 138, 9, 61, 0, 184, 9, 194, 9, 38, 0, 184, 9, 205, 9, 139, 9, 61, 0, 184, 9, 195, 9, 38, 0, 184, 9, 205, 9, 224, 9, 61, 0, 184, 9, 196, 9, 38, 0, 184, 9, 205, 9, 140, 9, 61, 0, 184, 9, 226, 9, 38, 0, 184, 9, 205, 9, 225, 9, 61, 0, 184, 9, 227, 9, 38, 0, 184, 9, 205, 9, 143, 9, 61, 0, 184, 9, 199, 9, 38, 0, 184, 9, 205, 9, 144, 9, 61, 0, 184, 9, 200, 9, 38, 0, 184, 9, 205, 9, 147, 9, 61, 0, 184, 9, 203, 9, 38, 0, 184, 9, 205, 9, 148, 9, 61, 0, 184, 9, 204, 9, 38, 0, 185, 9, 205, 9, 133, 9, 61, 0, 185, 9, 38, 0, 185, 9, 205, 9, 134, 9, 61, 0, 185, 9, 190, 9, 38, 0, 185, 9, 205, 9, 135, 9, 61, 0, 185, 9, 191, 9, 38, 0, 185, 9, 205, 9, 136, 9, 61, 0, 185, 9, 192, 9, 38, 0, 185, 9, 205, 9, 137, 9, 61, 0, 185, 9, 193, 9, 38, 0, 185, 9, 205, 9, 138, 9, 61, 0, 185, 9, 194, 9, 38, 0, 185, 9, 205, 9, 139, 9, 61, 0, 185, 9, 195, 9, 38, 0, 185, 9, 205, 9, 224, 9, 61, 0, 185, 9, 196, 9, 38, 0, 185, 9, 205, 9, 140, 9, 61, 0, 185, 9, 226, 9, 38, 0, 185, 9, 205, 9, 225, 9, 61, 0, 185, 9, 227, 9, 38, 0, 185, 9, 205, 9, 143, 9, 61, 0, 185, 9, 199, 9, 38, 0, 185, 9, 205, 9, 144, 9, 61, 0, 185, 9, 200, 9, 38, 0, 185, 9, 205, 9, 147, 9, 61, 0, 185, 9, 203, 9, 38, 0, 185, 9, 205, 9, 148, 9, 61, 0, 185, 9, 204, 9, 38, 0, 149, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 38, 0, 149, 9, 190, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 190, 9, 38, 0, 149, 9, 191, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 191, 9, 38, 0, 149, 9, 192, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 192, 9, 38, 0, 149, 9, 193, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 193, 9, 38, 0, 149, 9, 194, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 194, 9, 38, 0, 149, 9, 195, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 195, 9, 38, 0, 149, 9, 196, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 196, 9, 38, 0, 149, 9, 226, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 226, 9, 38, 0, 149, 9, 227, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 227, 9, 38, 0, 149, 9, 199, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 199, 9, 38, 0, 149, 9, 200, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 200, 9, 38, 0, 149, 9, 203, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 203, 9, 38, 0, 149, 9, 204, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 204, 9, 38, 0, 149, 9, 205, 9, 60, 0, 60, 0, 60, 0, 149, 9, 188, 9, 205, 9, 38, 0, 150, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 38, 0, 150, 9, 190, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 190, 9, 38, 0, 150, 9, 191, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 191, 9, 38, 0, 150, 9, 192, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 192, 9, 38, 0, 150, 9, 193, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 193, 9, 38, 0, 150, 9, 194, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 194, 9, 38, 0, 150, 9, 195, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 195, 9, 38, 0, 150, 9, 196, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 196, 9, 38, 0, 150, 9, 226, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 226, 9, 38, 0, 150, 9, 227, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 227, 9, 38, 0, 150, 9, 199, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 199, 9, 38, 0, 150, 9, 200, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 200, 9, 38, 0, 150, 9, 203, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 203, 9, 38, 0, 150, 9, 204, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 204, 9, 38, 0, 150, 9, 205, 9, 60, 0, 60, 0, 60, 0, 150, 9, 188, 9, 205, 9, 38, 0, 151, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 38, 0, 151, 9, 190, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 190, 9, 38, 0, 151, 9, 191, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 191, 9, 38, 0, 151, 9, 192, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 192, 9, 38, 0, 151, 9, 193, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 193, 9, 38, 0, 151, 9, 194, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 194, 9, 38, 0, 151, 9, 195, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 195, 9, 38, 0, 151, 9, 196, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 196, 9, 38, 0, 151, 9, 226, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 226, 9, 38, 0, 151, 9, 227, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 227, 9, 38, 0, 151, 9, 199, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 199, 9, 38, 0, 151, 9, 200, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 200, 9, 38, 0, 151, 9, 203, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 203, 9, 38, 0, 151, 9, 204, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 204, 9, 38, 0, 151, 9, 205, 9, 60, 0, 60, 0, 60, 0, 151, 9, 188, 9, 205, 9, 38, 0, 156, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 38, 0, 156, 9, 190, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 190, 9, 38, 0, 156, 9, 191, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 191, 9, 38, 0, 156, 9, 192, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 192, 9, 38, 0, 156, 9, 193, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 193, 9, 38, 0, 156, 9, 194, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 194, 9, 38, 0, 156, 9, 195, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 195, 9, 38, 0, 156, 9, 196, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 196, 9, 38, 0, 156, 9, 226, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 226, 9, 38, 0, 156, 9, 227, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 227, 9, 38, 0, 156, 9, 199, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 199, 9, 38, 0, 156, 9, 200, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 200, 9, 38, 0, 156, 9, 203, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 203, 9, 38, 0, 156, 9, 204, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 204, 9, 38, 0, 156, 9, 205, 9, 60, 0, 60, 0, 60, 0, 156, 9, 188, 9, 205, 9, 38, 0, 161, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 38, 0, 161, 9, 190, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 190, 9, 38, 0, 161, 9, 191, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 191, 9, 38, 0, 161, 9, 192, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 192, 9, 38, 0, 161, 9, 193, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 193, 9, 38, 0, 161, 9, 194, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 194, 9, 38, 0, 161, 9, 195, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 195, 9, 38, 0, 161, 9, 196, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 196, 9, 38, 0, 161, 9, 226, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 226, 9, 38, 0, 161, 9, 227, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 227, 9, 38, 0, 161, 9, 199, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 199, 9, 38, 0, 161, 9, 200, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 200, 9, 38, 0, 161, 9, 203, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 203, 9, 38, 0, 161, 9, 204, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 204, 9, 38, 0, 161, 9, 205, 9, 60, 0, 60, 0, 60, 0, 161, 9, 188, 9, 205, 9, 38, 0, 162, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 38, 0, 162, 9, 190, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 190, 9, 38, 0, 162, 9, 191, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 191, 9, 38, 0, 162, 9, 192, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 192, 9, 38, 0, 162, 9, 193, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 193, 9, 38, 0, 162, 9, 194, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 194, 9, 38, 0, 162, 9, 195, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 195, 9, 38, 0, 162, 9, 196, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 196, 9, 38, 0, 162, 9, 226, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 226, 9, 38, 0, 162, 9, 227, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 227, 9, 38, 0, 162, 9, 199, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 199, 9, 38, 0, 162, 9, 200, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 200, 9, 38, 0, 162, 9, 203, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 203, 9, 38, 0, 162, 9, 204, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 204, 9, 38, 0, 162, 9, 205, 9, 60, 0, 60, 0, 60, 0, 162, 9, 188, 9, 205, 9, 38, 0, 171, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 38, 0, 171, 9, 190, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 190, 9, 38, 0, 171, 9, 191, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 191, 9, 38, 0, 171, 9, 192, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 192, 9, 38, 0, 171, 9, 193, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 193, 9, 38, 0, 171, 9, 194, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 194, 9, 38, 0, 171, 9, 195, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 195, 9, 38, 0, 171, 9, 196, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 196, 9, 38, 0, 171, 9, 226, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 226, 9, 38, 0, 171, 9, 227, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 227, 9, 38, 0, 171, 9, 199, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 199, 9, 38, 0, 171, 9, 200, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 200, 9, 38, 0, 171, 9, 203, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 203, 9, 38, 0, 171, 9, 204, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 204, 9, 38, 0, 171, 9, 205, 9, 60, 0, 60, 0, 60, 0, 171, 9, 188, 9, 205, 9, 38, 0, 172, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 38, 0, 172, 9, 190, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 190, 9, 38, 0, 172, 9, 191, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 191, 9, 38, 0, 172, 9, 192, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 192, 9, 38, 0, 172, 9, 193, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 193, 9, 38, 0, 172, 9, 194, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 194, 9, 38, 0, 172, 9, 195, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 195, 9, 38, 0, 172, 9, 196, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 196, 9, 38, 0, 172, 9, 226, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 226, 9, 38, 0, 172, 9, 227, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 227, 9, 38, 0, 172, 9, 199, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 199, 9, 38, 0, 172, 9, 200, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 200, 9, 38, 0, 172, 9, 203, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 203, 9, 38, 0, 172, 9, 204, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 204, 9, 38, 0, 172, 9, 205, 9, 60, 0, 60, 0, 60, 0, 172, 9, 188, 9, 205, 9, 38, 0, 175, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 38, 0, 175, 9, 190, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 190, 9, 38, 0, 175, 9, 191, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 191, 9, 38, 0, 175, 9, 192, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 192, 9, 38, 0, 175, 9, 193, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 193, 9, 38, 0, 175, 9, 194, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 194, 9, 38, 0, 175, 9, 195, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 195, 9, 38, 0, 175, 9, 196, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 196, 9, 38, 0, 175, 9, 226, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 226, 9, 38, 0, 175, 9, 227, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 227, 9, 38, 0, 175, 9, 199, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 199, 9, 38, 0, 175, 9, 200, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 200, 9, 38, 0, 175, 9, 203, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 203, 9, 38, 0, 175, 9, 204, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 204, 9, 38, 0, 175, 9, 205, 9, 60, 0, 60, 0, 60, 0, 175, 9, 188, 9, 205, 9, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 112, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 116, 0, 0, 0, 116, 1, 0, 0, 20, 25, 0, 0, 20, 25, 0, 0, 20, 25, 0, 0, 20, 25, 0, 0, 88, 25, 0, 0, 4, 0, 0, 0, 10, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 105, 190, 255, 0, 106, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 39, 41, 42, 43, 44, 45, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 113, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 4, 35, 105, 193, 2, 35, 105, 193, 3, 35, 105, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 69, 0, 60, 0, 32, 0, 115, 9, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 188, 67, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 128, 0, 0, 0, 128, 1, 0, 0, 144, 26, 0, 0, 144, 26, 0, 0, 184, 27, 0, 0, 184, 27, 0, 0, 92, 46, 0, 0, 92, 46, 0, 0, 142, 67, 0, 0, 164, 67, 0, 0, 4, 0, 0, 0, 10, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 105, 190, 255, 0, 106, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 39, 41, 42, 43, 44, 45, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 136, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 104, 0, 111, 0, 119, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 147, 0, 202, 1, 148, 0, 202, 34, 148, 0, 202, 67, 148, 0, 202, 100, 148, 0, 202, 133, 148, 0, 202, 166, 148, 0, 202, 199, 148, 0, 202, 232, 148, 0, 202, 9, 149, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 147, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 147, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 147, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 4, 35, 105, 193, 2, 35, 105, 193, 3, 35, 105, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 4, 0, 0, 201, 164, 14, 0, 201, 68, 29, 0, 201, 4, 44, 0, 201, 100, 51, 0, 201, 196, 58, 0, 201, 36, 66, 0, 201, 132, 73, 0, 201, 68, 88, 0, 201, 164, 95, 0, 201, 4, 103, 0, 201, 100, 110, 0, 201, 196, 117, 0, 201, 132, 132, 0, 201, 68, 147, 0, 201, 164, 154, 0, 201, 132, 162, 0, 201, 228, 169, 0, 201, 68, 177, 0, 201, 164, 184, 0, 192, 0, 0, 0, 201, 4, 192, 0, 201, 100, 199, 0, 201, 36, 214, 0, 201, 228, 228, 0, 201, 68, 236, 0, 201, 164, 243, 0, 201, 100, 2, 1, 192, 0, 0, 0, 201, 196, 9, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 36, 17, 1, 201, 132, 24, 1, 201, 228, 31, 1, 201, 68, 39, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 20, 35, 105, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 164, 46, 1, 201, 228, 53, 1, 192, 0, 0, 0, 201, 36, 61, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 100, 68, 1, 201, 196, 75, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 6, 0, 5, 0, 5, 35, 105, 0, 6, 0, 5, 0, 6, 35, 105, 0, 6, 0, 5, 0, 7, 35, 105, 0, 6, 0, 5, 0, 12, 35, 105, 0, 6, 0, 5, 0, 17, 35, 105, 0, 6, 0, 5, 0, 18, 35, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 8, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 10, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 12, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 14, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 16, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 18, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 20, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 22, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 28, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 32, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 34, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 30, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 24, 105, 0, 5, 0, 70, 0, 20, 35, 105, 0, 5, 0, 5, 0, 0, 26, 105, 0, 6, 0, 5, 0, 26, 35, 105, 0, 6, 0, 5, 0, 27, 35, 105, 0, 6, 0, 5, 0, 30, 35, 105, 0, 0, 0, 0, 193, 5, 35, 105, 5, 5, 8, 105, 193, 5, 35, 105, 14, 5, 8, 105, 193, 5, 35, 105, 14, 5, 10, 105, 193, 5, 35, 105, 14, 5, 12, 105, 193, 5, 35, 105, 14, 5, 14, 105, 193, 5, 35, 105, 14, 5, 16, 105, 193, 5, 35, 105, 14, 5, 18, 105, 193, 5, 35, 105, 14, 5, 20, 105, 193, 5, 35, 105, 14, 5, 22, 105, 193, 5, 35, 105, 14, 5, 28, 105, 193, 5, 35, 105, 14, 5, 32, 105, 193, 5, 35, 105, 14, 5, 34, 105, 193, 5, 35, 105, 14, 5, 30, 105, 193, 5, 35, 105, 14, 5, 24, 105, 193, 5, 35, 105, 14, 5, 26, 105, 193, 5, 35, 105, 5, 5, 10, 105, 193, 5, 35, 105, 5, 5, 12, 105, 193, 5, 35, 105, 5, 5, 14, 105, 193, 5, 35, 105, 5, 5, 16, 105, 193, 5, 35, 105, 5, 5, 18, 105, 193, 5, 35, 105, 5, 5, 20, 105, 193, 5, 35, 105, 5, 5, 22, 105, 193, 5, 35, 105, 5, 5, 28, 105, 193, 5, 35, 105, 5, 5, 32, 105, 193, 5, 35, 105, 5, 5, 34, 105, 193, 5, 35, 105, 5, 5, 30, 105, 193, 5, 35, 105, 5, 5, 24, 105, 193, 5, 35, 105, 5, 5, 26, 105, 193, 6, 35, 105, 5, 5, 8, 105, 193, 6, 35, 105, 13, 5, 8, 105, 193, 6, 35, 105, 13, 5, 10, 105, 193, 6, 35, 105, 13, 5, 12, 105, 193, 6, 35, 105, 13, 5, 14, 105, 193, 6, 35, 105, 13, 5, 16, 105, 193, 6, 35, 105, 13, 5, 18, 105, 193, 6, 35, 105, 13, 5, 20, 105, 193, 6, 35, 105, 13, 5, 22, 105, 193, 6, 35, 105, 13, 5, 28, 105, 193, 6, 35, 105, 13, 5, 32, 105, 193, 6, 35, 105, 13, 5, 34, 105, 193, 6, 35, 105, 13, 5, 30, 105, 193, 6, 35, 105, 13, 5, 24, 105, 193, 6, 35, 105, 13, 5, 26, 105, 193, 6, 35, 105, 5, 5, 10, 105, 193, 6, 35, 105, 5, 5, 12, 105, 193, 6, 35, 105, 5, 5, 14, 105, 193, 6, 35, 105, 5, 5, 16, 105, 193, 6, 35, 105, 5, 5, 18, 105, 193, 6, 35, 105, 5, 5, 20, 105, 193, 6, 35, 105, 5, 5, 22, 105, 193, 6, 35, 105, 5, 5, 28, 105, 193, 6, 35, 105, 5, 5, 32, 105, 193, 6, 35, 105, 5, 5, 34, 105, 193, 6, 35, 105, 5, 5, 30, 105, 193, 6, 35, 105, 5, 5, 24, 105, 193, 6, 35, 105, 5, 5, 26, 105, 193, 7, 35, 105, 5, 5, 8, 105, 193, 7, 35, 105, 12, 5, 8, 105, 193, 7, 35, 105, 12, 5, 10, 105, 193, 7, 35, 105, 12, 5, 12, 105, 193, 7, 35, 105, 12, 5, 14, 105, 193, 7, 35, 105, 12, 5, 16, 105, 193, 7, 35, 105, 12, 5, 18, 105, 193, 7, 35, 105, 12, 5, 20, 105, 193, 7, 35, 105, 12, 5, 22, 105, 193, 7, 35, 105, 12, 5, 28, 105, 193, 7, 35, 105, 12, 5, 32, 105, 193, 7, 35, 105, 12, 5, 34, 105, 193, 7, 35, 105, 12, 5, 30, 105, 193, 7, 35, 105, 12, 5, 24, 105, 193, 7, 35, 105, 12, 5, 26, 105, 193, 7, 35, 105, 5, 5, 10, 105, 193, 7, 35, 105, 5, 5, 12, 105, 193, 7, 35, 105, 5, 5, 14, 105, 193, 7, 35, 105, 5, 5, 16, 105, 193, 7, 35, 105, 5, 5, 18, 105, 193, 7, 35, 105, 5, 5, 20, 105, 193, 7, 35, 105, 5, 5, 22, 105, 193, 7, 35, 105, 5, 5, 28, 105, 193, 7, 35, 105, 5, 5, 32, 105, 193, 7, 35, 105, 5, 5, 34, 105, 193, 7, 35, 105, 5, 5, 30, 105, 193, 7, 35, 105, 5, 5, 24, 105, 193, 7, 35, 105, 5, 5, 26, 105, 193, 8, 35, 105, 5, 5, 8, 105, 193, 8, 35, 105, 5, 5, 10, 105, 193, 8, 35, 105, 5, 5, 12, 105, 193, 8, 35, 105, 5, 5, 14, 105, 193, 8, 35, 105, 5, 5, 16, 105, 193, 8, 35, 105, 5, 5, 18, 105, 193, 8, 35, 105, 5, 5, 20, 105, 193, 8, 35, 105, 5, 5, 22, 105, 193, 8, 35, 105, 5, 5, 28, 105, 193, 8, 35, 105, 5, 5, 32, 105, 193, 8, 35, 105, 5, 5, 34, 105, 193, 8, 35, 105, 5, 5, 30, 105, 193, 8, 35, 105, 5, 5, 24, 105, 193, 8, 35, 105, 5, 5, 26, 105, 193, 9, 35, 105, 5, 5, 8, 105, 193, 9, 35, 105, 5, 5, 10, 105, 193, 9, 35, 105, 5, 5, 12, 105, 193, 9, 35, 105, 5, 5, 14, 105, 193, 9, 35, 105, 5, 5, 16, 105, 193, 9, 35, 105, 5, 5, 18, 105, 193, 9, 35, 105, 5, 5, 20, 105, 193, 9, 35, 105, 5, 5, 22, 105, 193, 9, 35, 105, 5, 5, 28, 105, 193, 9, 35, 105, 5, 5, 32, 105, 193, 9, 35, 105, 5, 5, 34, 105, 193, 9, 35, 105, 5, 5, 30, 105, 193, 9, 35, 105, 5, 5, 24, 105, 193, 9, 35, 105, 5, 5, 26, 105, 193, 10, 35, 105, 5, 5, 8, 105, 193, 10, 35, 105, 5, 5, 10, 105, 193, 10, 35, 105, 5, 5, 12, 105, 193, 10, 35, 105, 5, 5, 14, 105, 193, 10, 35, 105, 5, 5, 16, 105, 193, 10, 35, 105, 5, 5, 18, 105, 193, 10, 35, 105, 5, 5, 20, 105, 193, 10, 35, 105, 5, 5, 22, 105, 193, 10, 35, 105, 5, 5, 28, 105, 193, 10, 35, 105, 5, 5, 32, 105, 193, 10, 35, 105, 5, 5, 34, 105, 193, 10, 35, 105, 5, 5, 30, 105, 193, 10, 35, 105, 5, 5, 24, 105, 193, 10, 35, 105, 5, 5, 26, 105, 193, 11, 35, 105, 5, 5, 8, 105, 193, 11, 35, 105, 5, 5, 10, 105, 193, 11, 35, 105, 5, 5, 12, 105, 193, 11, 35, 105, 5, 5, 14, 105, 193, 11, 35, 105, 5, 5, 16, 105, 193, 11, 35, 105, 5, 5, 18, 105, 193, 11, 35, 105, 5, 5, 20, 105, 193, 11, 35, 105, 5, 5, 22, 105, 193, 11, 35, 105, 5, 5, 28, 105, 193, 11, 35, 105, 5, 5, 32, 105, 193, 11, 35, 105, 5, 5, 34, 105, 193, 11, 35, 105, 5, 5, 30, 105, 193, 11, 35, 105, 5, 5, 24, 105, 193, 11, 35, 105, 5, 5, 26, 105, 193, 12, 35, 105, 5, 5, 8, 105, 193, 12, 35, 105, 11, 5, 8, 105, 193, 12, 35, 105, 11, 5, 10, 105, 193, 12, 35, 105, 11, 5, 12, 105, 193, 12, 35, 105, 11, 5, 14, 105, 193, 12, 35, 105, 11, 5, 16, 105, 193, 12, 35, 105, 11, 5, 18, 105, 193, 12, 35, 105, 11, 5, 20, 105, 193, 12, 35, 105, 11, 5, 22, 105, 193, 12, 35, 105, 11, 5, 28, 105, 193, 12, 35, 105, 11, 5, 32, 105, 193, 12, 35, 105, 11, 5, 34, 105, 193, 12, 35, 105, 11, 5, 30, 105, 193, 12, 35, 105, 11, 5, 24, 105, 193, 12, 35, 105, 11, 5, 26, 105, 193, 12, 35, 105, 5, 5, 10, 105, 193, 12, 35, 105, 5, 5, 12, 105, 193, 12, 35, 105, 5, 5, 14, 105, 193, 12, 35, 105, 5, 5, 16, 105, 193, 12, 35, 105, 5, 5, 18, 105, 193, 12, 35, 105, 5, 5, 20, 105, 193, 12, 35, 105, 5, 5, 22, 105, 193, 12, 35, 105, 5, 5, 28, 105, 193, 12, 35, 105, 5, 5, 32, 105, 193, 12, 35, 105, 5, 5, 34, 105, 193, 12, 35, 105, 5, 5, 30, 105, 193, 12, 35, 105, 5, 5, 24, 105, 193, 12, 35, 105, 5, 5, 26, 105, 193, 13, 35, 105, 5, 5, 8, 105, 193, 13, 35, 105, 5, 5, 10, 105, 193, 13, 35, 105, 5, 5, 12, 105, 193, 13, 35, 105, 5, 5, 14, 105, 193, 13, 35, 105, 5, 5, 16, 105, 193, 13, 35, 105, 5, 5, 18, 105, 193, 13, 35, 105, 5, 5, 20, 105, 193, 13, 35, 105, 5, 5, 22, 105, 193, 13, 35, 105, 5, 5, 28, 105, 193, 13, 35, 105, 5, 5, 32, 105, 193, 13, 35, 105, 5, 5, 34, 105, 193, 13, 35, 105, 5, 5, 30, 105, 193, 13, 35, 105, 5, 5, 24, 105, 193, 13, 35, 105, 5, 5, 26, 105, 193, 14, 35, 105, 5, 5, 8, 105, 193, 14, 35, 105, 5, 5, 10, 105, 193, 14, 35, 105, 5, 5, 12, 105, 193, 14, 35, 105, 5, 5, 14, 105, 193, 14, 35, 105, 5, 5, 16, 105, 193, 14, 35, 105, 5, 5, 18, 105, 193, 14, 35, 105, 5, 5, 20, 105, 193, 14, 35, 105, 5, 5, 22, 105, 193, 14, 35, 105, 5, 5, 28, 105, 193, 14, 35, 105, 5, 5, 32, 105, 193, 14, 35, 105, 5, 5, 34, 105, 193, 14, 35, 105, 5, 5, 30, 105, 193, 14, 35, 105, 5, 5, 24, 105, 193, 14, 35, 105, 5, 5, 26, 105, 193, 15, 35, 105, 5, 5, 8, 105, 193, 15, 35, 105, 5, 5, 10, 105, 193, 15, 35, 105, 5, 5, 12, 105, 193, 15, 35, 105, 5, 5, 14, 105, 193, 15, 35, 105, 5, 5, 16, 105, 193, 15, 35, 105, 5, 5, 18, 105, 193, 15, 35, 105, 5, 5, 20, 105, 193, 15, 35, 105, 5, 5, 22, 105, 193, 15, 35, 105, 5, 5, 28, 105, 193, 15, 35, 105, 5, 5, 32, 105, 193, 15, 35, 105, 5, 5, 34, 105, 193, 15, 35, 105, 5, 5, 30, 105, 193, 15, 35, 105, 5, 5, 24, 105, 193, 15, 35, 105, 5, 5, 26, 105, 193, 16, 35, 105, 5, 5, 8, 105, 193, 16, 35, 105, 5, 5, 10, 105, 193, 16, 35, 105, 5, 5, 12, 105, 193, 16, 35, 105, 5, 5, 14, 105, 193, 16, 35, 105, 5, 5, 16, 105, 193, 16, 35, 105, 5, 5, 18, 105, 193, 16, 35, 105, 5, 5, 20, 105, 193, 16, 35, 105, 5, 5, 22, 105, 193, 16, 35, 105, 5, 5, 28, 105, 193, 16, 35, 105, 5, 5, 32, 105, 193, 16, 35, 105, 5, 5, 34, 105, 193, 16, 35, 105, 5, 5, 30, 105, 193, 16, 35, 105, 5, 5, 24, 105, 193, 16, 35, 105, 5, 5, 26, 105, 193, 17, 35, 105, 5, 5, 8, 105, 193, 17, 35, 105, 10, 5, 8, 105, 193, 17, 35, 105, 10, 5, 10, 105, 193, 17, 35, 105, 10, 5, 12, 105, 193, 17, 35, 105, 10, 5, 14, 105, 193, 17, 35, 105, 10, 5, 16, 105, 193, 17, 35, 105, 10, 5, 18, 105, 193, 17, 35, 105, 10, 5, 20, 105, 193, 17, 35, 105, 10, 5, 22, 105, 193, 17, 35, 105, 10, 5, 28, 105, 193, 17, 35, 105, 10, 5, 32, 105, 193, 17, 35, 105, 10, 5, 34, 105, 193, 17, 35, 105, 10, 5, 30, 105, 193, 17, 35, 105, 10, 5, 24, 105, 193, 17, 35, 105, 10, 5, 26, 105, 193, 17, 35, 105, 5, 5, 10, 105, 193, 17, 35, 105, 5, 5, 12, 105, 193, 17, 35, 105, 5, 5, 14, 105, 193, 17, 35, 105, 5, 5, 16, 105, 193, 17, 35, 105, 5, 5, 18, 105, 193, 17, 35, 105, 5, 5, 20, 105, 193, 17, 35, 105, 5, 5, 22, 105, 193, 17, 35, 105, 5, 5, 28, 105, 193, 17, 35, 105, 5, 5, 32, 105, 193, 17, 35, 105, 5, 5, 34, 105, 193, 17, 35, 105, 5, 5, 30, 105, 193, 17, 35, 105, 5, 5, 24, 105, 193, 17, 35, 105, 5, 5, 26, 105, 193, 18, 35, 105, 5, 5, 8, 105, 193, 18, 35, 105, 9, 5, 8, 105, 193, 18, 35, 105, 9, 5, 10, 105, 193, 18, 35, 105, 9, 5, 12, 105, 193, 18, 35, 105, 9, 5, 14, 105, 193, 18, 35, 105, 9, 5, 16, 105, 193, 18, 35, 105, 9, 5, 18, 105, 193, 18, 35, 105, 9, 5, 20, 105, 193, 18, 35, 105, 9, 5, 22, 105, 193, 18, 35, 105, 9, 5, 28, 105, 193, 18, 35, 105, 9, 5, 32, 105, 193, 18, 35, 105, 9, 5, 34, 105, 193, 18, 35, 105, 9, 5, 30, 105, 193, 18, 35, 105, 9, 5, 24, 105, 193, 18, 35, 105, 9, 5, 26, 105, 193, 18, 35, 105, 5, 5, 10, 105, 193, 18, 35, 105, 5, 5, 12, 105, 193, 18, 35, 105, 5, 5, 14, 105, 193, 18, 35, 105, 5, 5, 16, 105, 193, 18, 35, 105, 5, 5, 18, 105, 193, 18, 35, 105, 5, 5, 20, 105, 193, 18, 35, 105, 5, 5, 22, 105, 193, 18, 35, 105, 5, 5, 28, 105, 193, 18, 35, 105, 5, 5, 32, 105, 193, 18, 35, 105, 5, 5, 34, 105, 193, 18, 35, 105, 5, 5, 30, 105, 193, 18, 35, 105, 5, 5, 24, 105, 193, 18, 35, 105, 5, 5, 26, 105, 193, 19, 35, 105, 5, 5, 8, 105, 193, 19, 35, 105, 5, 5, 10, 105, 193, 19, 35, 105, 5, 5, 12, 105, 193, 19, 35, 105, 5, 5, 14, 105, 193, 19, 35, 105, 5, 5, 16, 105, 193, 19, 35, 105, 5, 5, 18, 105, 193, 19, 35, 105, 5, 5, 20, 105, 193, 19, 35, 105, 5, 5, 22, 105, 193, 19, 35, 105, 5, 5, 28, 105, 193, 19, 35, 105, 5, 5, 32, 105, 193, 19, 35, 105, 5, 5, 34, 105, 193, 19, 35, 105, 5, 5, 30, 105, 193, 19, 35, 105, 5, 5, 24, 105, 193, 19, 35, 105, 5, 5, 26, 105, 193, 21, 35, 105, 5, 5, 8, 105, 193, 21, 35, 105, 5, 5, 10, 105, 193, 21, 35, 105, 5, 5, 12, 105, 193, 21, 35, 105, 5, 5, 14, 105, 193, 21, 35, 105, 5, 5, 16, 105, 193, 21, 35, 105, 5, 5, 18, 105, 193, 21, 35, 105, 5, 5, 20, 105, 193, 21, 35, 105, 5, 5, 22, 105, 193, 21, 35, 105, 5, 5, 28, 105, 193, 21, 35, 105, 5, 5, 32, 105, 193, 21, 35, 105, 5, 5, 34, 105, 193, 21, 35, 105, 5, 5, 30, 105, 193, 21, 35, 105, 5, 5, 24, 105, 193, 21, 35, 105, 5, 5, 26, 105, 193, 22, 35, 105, 5, 5, 8, 105, 193, 22, 35, 105, 5, 5, 10, 105, 193, 22, 35, 105, 5, 5, 12, 105, 193, 22, 35, 105, 5, 5, 14, 105, 193, 22, 35, 105, 5, 5, 16, 105, 193, 22, 35, 105, 5, 5, 18, 105, 193, 22, 35, 105, 5, 5, 20, 105, 193, 22, 35, 105, 5, 5, 22, 105, 193, 22, 35, 105, 5, 5, 28, 105, 193, 22, 35, 105, 5, 5, 32, 105, 193, 22, 35, 105, 5, 5, 34, 105, 193, 22, 35, 105, 5, 5, 30, 105, 193, 22, 35, 105, 5, 5, 24, 105, 193, 22, 35, 105, 5, 5, 26, 105, 193, 23, 35, 105, 5, 5, 8, 105, 193, 23, 35, 105, 5, 5, 10, 105, 193, 23, 35, 105, 5, 5, 12, 105, 193, 23, 35, 105, 5, 5, 14, 105, 193, 23, 35, 105, 5, 5, 16, 105, 193, 23, 35, 105, 5, 5, 18, 105, 193, 23, 35, 105, 5, 5, 20, 105, 193, 23, 35, 105, 5, 5, 22, 105, 193, 23, 35, 105, 5, 5, 28, 105, 193, 23, 35, 105, 5, 5, 32, 105, 193, 23, 35, 105, 5, 5, 34, 105, 193, 23, 35, 105, 5, 5, 30, 105, 193, 23, 35, 105, 5, 5, 24, 105, 193, 23, 35, 105, 5, 5, 26, 105, 193, 24, 35, 105, 5, 5, 8, 105, 193, 24, 35, 105, 5, 5, 10, 105, 193, 24, 35, 105, 5, 5, 12, 105, 193, 24, 35, 105, 5, 5, 14, 105, 193, 24, 35, 105, 5, 5, 16, 105, 193, 24, 35, 105, 5, 5, 18, 105, 193, 24, 35, 105, 5, 5, 20, 105, 193, 24, 35, 105, 5, 5, 22, 105, 193, 24, 35, 105, 5, 5, 28, 105, 193, 24, 35, 105, 5, 5, 32, 105, 193, 24, 35, 105, 5, 5, 34, 105, 193, 24, 35, 105, 5, 5, 30, 105, 193, 24, 35, 105, 5, 5, 24, 105, 193, 24, 35, 105, 5, 5, 26, 105, 193, 25, 35, 105, 5, 5, 8, 105, 193, 25, 35, 105, 5, 5, 10, 105, 193, 25, 35, 105, 5, 5, 12, 105, 193, 25, 35, 105, 5, 5, 14, 105, 193, 25, 35, 105, 5, 5, 16, 105, 193, 25, 35, 105, 5, 5, 18, 105, 193, 25, 35, 105, 5, 5, 20, 105, 193, 25, 35, 105, 5, 5, 22, 105, 193, 25, 35, 105, 5, 5, 28, 105, 193, 25, 35, 105, 5, 5, 32, 105, 193, 25, 35, 105, 5, 5, 34, 105, 193, 25, 35, 105, 5, 5, 30, 105, 193, 25, 35, 105, 5, 5, 24, 105, 193, 25, 35, 105, 5, 5, 26, 105, 193, 26, 35, 105, 5, 5, 8, 105, 193, 26, 35, 105, 8, 5, 8, 105, 193, 26, 35, 105, 8, 5, 10, 105, 193, 26, 35, 105, 8, 5, 12, 105, 193, 26, 35, 105, 8, 5, 14, 105, 193, 26, 35, 105, 8, 5, 16, 105, 193, 26, 35, 105, 8, 5, 18, 105, 193, 26, 35, 105, 8, 5, 20, 105, 193, 26, 35, 105, 8, 5, 22, 105, 193, 26, 35, 105, 8, 5, 28, 105, 193, 26, 35, 105, 8, 5, 32, 105, 193, 26, 35, 105, 8, 5, 34, 105, 193, 26, 35, 105, 8, 5, 30, 105, 193, 26, 35, 105, 8, 5, 24, 105, 193, 26, 35, 105, 8, 5, 26, 105, 193, 26, 35, 105, 5, 5, 10, 105, 193, 26, 35, 105, 5, 5, 12, 105, 193, 26, 35, 105, 5, 5, 14, 105, 193, 26, 35, 105, 5, 5, 16, 105, 193, 26, 35, 105, 5, 5, 18, 105, 193, 26, 35, 105, 5, 5, 20, 105, 193, 26, 35, 105, 5, 5, 22, 105, 193, 26, 35, 105, 5, 5, 28, 105, 193, 26, 35, 105, 5, 5, 32, 105, 193, 26, 35, 105, 5, 5, 34, 105, 193, 26, 35, 105, 5, 5, 30, 105, 193, 26, 35, 105, 5, 5, 24, 105, 193, 26, 35, 105, 5, 5, 26, 105, 193, 27, 35, 105, 5, 5, 8, 105, 193, 27, 35, 105, 7, 5, 8, 105, 193, 27, 35, 105, 7, 5, 10, 105, 193, 27, 35, 105, 7, 5, 12, 105, 193, 27, 35, 105, 7, 5, 14, 105, 193, 27, 35, 105, 7, 5, 16, 105, 193, 27, 35, 105, 7, 5, 18, 105, 193, 27, 35, 105, 7, 5, 20, 105, 193, 27, 35, 105, 7, 5, 22, 105, 193, 27, 35, 105, 7, 5, 28, 105, 193, 27, 35, 105, 7, 5, 32, 105, 193, 27, 35, 105, 7, 5, 34, 105, 193, 27, 35, 105, 7, 5, 30, 105, 193, 27, 35, 105, 7, 5, 24, 105, 193, 27, 35, 105, 7, 5, 26, 105, 193, 27, 35, 105, 5, 5, 10, 105, 193, 27, 35, 105, 5, 5, 12, 105, 193, 27, 35, 105, 5, 5, 14, 105, 193, 27, 35, 105, 5, 5, 16, 105, 193, 27, 35, 105, 5, 5, 18, 105, 193, 27, 35, 105, 5, 5, 20, 105, 193, 27, 35, 105, 5, 5, 22, 105, 193, 27, 35, 105, 5, 5, 28, 105, 193, 27, 35, 105, 5, 5, 32, 105, 193, 27, 35, 105, 5, 5, 34, 105, 193, 27, 35, 105, 5, 5, 30, 105, 193, 27, 35, 105, 5, 5, 24, 105, 193, 27, 35, 105, 5, 5, 26, 105, 193, 28, 35, 105, 5, 5, 8, 105, 193, 28, 35, 105, 5, 5, 10, 105, 193, 28, 35, 105, 5, 5, 12, 105, 193, 28, 35, 105, 5, 5, 14, 105, 193, 28, 35, 105, 5, 5, 16, 105, 193, 28, 35, 105, 5, 5, 18, 105, 193, 28, 35, 105, 5, 5, 20, 105, 193, 28, 35, 105, 5, 5, 22, 105, 193, 28, 35, 105, 5, 5, 28, 105, 193, 28, 35, 105, 5, 5, 32, 105, 193, 28, 35, 105, 5, 5, 34, 105, 193, 28, 35, 105, 5, 5, 30, 105, 193, 28, 35, 105, 5, 5, 24, 105, 193, 28, 35, 105, 5, 5, 26, 105, 193, 29, 35, 105, 5, 5, 8, 105, 193, 29, 35, 105, 5, 5, 10, 105, 193, 29, 35, 105, 5, 5, 12, 105, 193, 29, 35, 105, 5, 5, 14, 105, 193, 29, 35, 105, 5, 5, 16, 105, 193, 29, 35, 105, 5, 5, 18, 105, 193, 29, 35, 105, 5, 5, 20, 105, 193, 29, 35, 105, 5, 5, 22, 105, 193, 29, 35, 105, 5, 5, 28, 105, 193, 29, 35, 105, 5, 5, 32, 105, 193, 29, 35, 105, 5, 5, 34, 105, 193, 29, 35, 105, 5, 5, 30, 105, 193, 29, 35, 105, 5, 5, 24, 105, 193, 29, 35, 105, 5, 5, 26, 105, 193, 30, 35, 105, 5, 5, 8, 105, 193, 30, 35, 105, 6, 5, 8, 105, 193, 30, 35, 105, 6, 5, 10, 105, 193, 30, 35, 105, 6, 5, 12, 105, 193, 30, 35, 105, 6, 5, 14, 105, 193, 30, 35, 105, 6, 5, 16, 105, 193, 30, 35, 105, 6, 5, 18, 105, 193, 30, 35, 105, 6, 5, 20, 105, 193, 30, 35, 105, 6, 5, 22, 105, 193, 30, 35, 105, 6, 5, 28, 105, 193, 30, 35, 105, 6, 5, 32, 105, 193, 30, 35, 105, 6, 5, 34, 105, 193, 30, 35, 105, 6, 5, 30, 105, 193, 30, 35, 105, 6, 5, 24, 105, 193, 30, 35, 105, 6, 5, 26, 105, 193, 30, 35, 105, 5, 5, 10, 105, 193, 30, 35, 105, 5, 5, 12, 105, 193, 30, 35, 105, 5, 5, 14, 105, 193, 30, 35, 105, 5, 5, 16, 105, 193, 30, 35, 105, 5, 5, 18, 105, 193, 30, 35, 105, 5, 5, 20, 105, 193, 30, 35, 105, 5, 5, 22, 105, 193, 30, 35, 105, 5, 5, 28, 105, 193, 30, 35, 105, 5, 5, 32, 105, 193, 30, 35, 105, 5, 5, 34, 105, 193, 30, 35, 105, 5, 5, 30, 105, 193, 30, 35, 105, 5, 5, 24, 105, 193, 30, 35, 105, 5, 5, 26, 105, 193, 31, 35, 105, 5, 5, 8, 105, 193, 31, 35, 105, 5, 5, 10, 105, 193, 31, 35, 105, 5, 5, 12, 105, 193, 31, 35, 105, 5, 5, 14, 105, 193, 31, 35, 105, 5, 5, 16, 105, 193, 31, 35, 105, 5, 5, 18, 105, 193, 31, 35, 105, 5, 5, 20, 105, 193, 31, 35, 105, 5, 5, 22, 105, 193, 31, 35, 105, 5, 5, 28, 105, 193, 31, 35, 105, 5, 5, 32, 105, 193, 31, 35, 105, 5, 5, 34, 105, 193, 31, 35, 105, 5, 5, 30, 105, 193, 31, 35, 105, 5, 5, 24, 105, 193, 31, 35, 105, 5, 5, 26, 105, 193, 33, 35, 105, 5, 5, 8, 105, 193, 33, 35, 105, 5, 5, 10, 105, 193, 33, 35, 105, 5, 5, 12, 105, 193, 33, 35, 105, 5, 5, 14, 105, 193, 33, 35, 105, 5, 5, 16, 105, 193, 33, 35, 105, 5, 5, 18, 105, 193, 33, 35, 105, 5, 5, 20, 105, 193, 33, 35, 105, 5, 5, 22, 105, 193, 33, 35, 105, 5, 5, 28, 105, 193, 33, 35, 105, 5, 5, 32, 105, 193, 33, 35, 105, 5, 5, 34, 105, 193, 33, 35, 105, 5, 5, 30, 105, 193, 33, 35, 105, 5, 5, 24, 105, 193, 33, 35, 105, 5, 5, 26, 105, 193, 35, 35, 105, 5, 5, 8, 105, 193, 35, 35, 105, 5, 5, 10, 105, 193, 35, 35, 105, 5, 5, 12, 105, 193, 35, 35, 105, 5, 5, 14, 105, 193, 35, 35, 105, 5, 5, 16, 105, 193, 35, 35, 105, 5, 5, 18, 105, 193, 35, 35, 105, 5, 5, 20, 105, 193, 35, 35, 105, 5, 5, 22, 105, 193, 35, 35, 105, 5, 5, 28, 105, 193, 35, 35, 105, 5, 5, 32, 105, 193, 35, 35, 105, 5, 5, 34, 105, 193, 35, 35, 105, 5, 5, 30, 105, 193, 35, 35, 105, 5, 5, 24, 105, 193, 35, 35, 105, 5, 5, 26, 105, 193, 36, 35, 105, 5, 5, 8, 105, 193, 36, 35, 105, 5, 5, 10, 105, 193, 36, 35, 105, 5, 5, 12, 105, 193, 36, 35, 105, 5, 5, 14, 105, 193, 36, 35, 105, 5, 5, 16, 105, 193, 36, 35, 105, 5, 5, 18, 105, 193, 36, 35, 105, 5, 5, 20, 105, 193, 36, 35, 105, 5, 5, 22, 105, 193, 36, 35, 105, 5, 5, 28, 105, 193, 36, 35, 105, 5, 5, 32, 105, 193, 36, 35, 105, 5, 5, 34, 105, 193, 36, 35, 105, 5, 5, 30, 105, 193, 36, 35, 105, 5, 5, 24, 105, 193, 36, 35, 105, 5, 5, 26, 105, 193, 37, 35, 105, 5, 5, 8, 105, 193, 37, 35, 105, 5, 5, 10, 105, 193, 37, 35, 105, 5, 5, 12, 105, 193, 37, 35, 105, 5, 5, 14, 105, 193, 37, 35, 105, 5, 5, 16, 105, 193, 37, 35, 105, 5, 5, 18, 105, 193, 37, 35, 105, 5, 5, 20, 105, 193, 37, 35, 105, 5, 5, 22, 105, 193, 37, 35, 105, 5, 5, 28, 105, 193, 37, 35, 105, 5, 5, 32, 105, 193, 37, 35, 105, 5, 5, 34, 105, 193, 37, 35, 105, 5, 5, 30, 105, 193, 37, 35, 105, 5, 5, 24, 105, 193, 37, 35, 105, 5, 5, 26, 105, 193, 38, 35, 105, 5, 5, 8, 105, 193, 38, 35, 105, 5, 5, 10, 105, 193, 38, 35, 105, 5, 5, 12, 105, 193, 38, 35, 105, 5, 5, 14, 105, 193, 38, 35, 105, 5, 5, 16, 105, 193, 38, 35, 105, 5, 5, 18, 105, 193, 38, 35, 105, 5, 5, 20, 105, 193, 38, 35, 105, 5, 5, 22, 105, 193, 38, 35, 105, 5, 5, 28, 105, 193, 38, 35, 105, 5, 5, 32, 105, 193, 38, 35, 105, 5, 5, 34, 105, 193, 38, 35, 105, 5, 5, 30, 105, 193, 38, 35, 105, 5, 5, 24, 105, 193, 38, 35, 105, 5, 5, 26, 105, 193, 32, 35, 105, 5, 5, 8, 105, 193, 32, 35, 105, 5, 5, 10, 105, 193, 32, 35, 105, 5, 5, 12, 105, 193, 32, 35, 105, 5, 5, 14, 105, 193, 32, 35, 105, 5, 5, 16, 105, 193, 32, 35, 105, 5, 5, 18, 105, 193, 32, 35, 105, 5, 5, 20, 105, 193, 32, 35, 105, 5, 5, 22, 105, 193, 32, 35, 105, 5, 5, 28, 105, 193, 32, 35, 105, 5, 5, 32, 105, 193, 32, 35, 105, 5, 5, 34, 105, 193, 32, 35, 105, 5, 5, 30, 105, 193, 32, 35, 105, 5, 5, 24, 105, 193, 32, 35, 105, 5, 5, 26, 105, 193, 34, 35, 105, 5, 5, 8, 105, 193, 34, 35, 105, 5, 5, 10, 105, 193, 34, 35, 105, 5, 5, 12, 105, 193, 34, 35, 105, 5, 5, 14, 105, 193, 34, 35, 105, 5, 5, 16, 105, 193, 34, 35, 105, 5, 5, 18, 105, 193, 34, 35, 105, 5, 5, 20, 105, 193, 34, 35, 105, 5, 5, 22, 105, 193, 34, 35, 105, 5, 5, 28, 105, 193, 34, 35, 105, 5, 5, 32, 105, 193, 34, 35, 105, 5, 5, 34, 105, 193, 34, 35, 105, 5, 5, 30, 105, 193, 34, 35, 105, 5, 5, 24, 105, 193, 34, 35, 105, 5, 5, 26, 105, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 0, 197, 34, 14, 0, 196, 9, 34, 0, 204, 9, 13, 0, 204, 9, 6, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 5, 226, 9, 6, 192, 197, 162, 227, 9, 6, 192, 197, 226, 196, 9, 5, 192, 197, 98, 199, 9, 6, 0, 200, 9, 6, 192, 197, 98, 203, 9, 5, 192, 197, 226, 129, 65, 197, 162, 190, 9, 5, 192, 197, 226, 215, 9, 6, 192, 197, 34, 192, 9, 12, 0, 192, 9, 4, 192, 197, 98, 193, 9, 4, 192, 197, 162, 194, 9, 4, 192, 197, 226, 195, 9, 5, 192, 197, 34, 188, 9, 6, 0, 190, 9, 3, 192, 197, 226, 191, 9, 4, 192, 197, 34, 77, 64, 197, 98, 199, 9, 29, 0, 204, 9, 11, 0, 204, 9, 2, 192, 197, 226, 205, 9, 198, 129, 226, 9, 3, 192, 197, 98, 227, 9, 3, 192, 197, 162, 199, 9, 6, 0, 200, 9, 3, 192, 197, 34, 203, 9, 2, 192, 197, 162, 193, 64, 197, 98, 190, 9, 2, 192, 197, 162, 215, 9, 2, 192, 197, 226, 193, 9, 12, 0, 193, 9, 1, 192, 197, 98, 194, 9, 1, 192, 197, 162, 195, 9, 1, 192, 197, 226, 196, 9, 2, 192, 197, 34, 190, 9, 0, 192, 197, 162, 191, 9, 0, 192, 197, 226, 192, 9, 1, 192, 197, 34, 7, 0, 197, 34, 14, 0, 196, 9, 34, 0, 204, 9, 13, 0, 204, 9, 13, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 6, 226, 9, 13, 192, 197, 162, 227, 9, 13, 192, 197, 226, 196, 9, 12, 192, 197, 98, 199, 9, 6, 0, 200, 9, 13, 192, 197, 98, 203, 9, 12, 192, 197, 226, 65, 67, 197, 162, 190, 9, 12, 192, 197, 226, 215, 9, 13, 192, 197, 34, 192, 9, 12, 0, 192, 9, 11, 192, 197, 98, 193, 9, 11, 192, 197, 162, 194, 9, 11, 192, 197, 226, 195, 9, 12, 192, 197, 34, 188, 9, 6, 0, 190, 9, 10, 192, 197, 226, 191, 9, 11, 192, 197, 34, 13, 66, 197, 98, 199, 9, 29, 0, 204, 9, 11, 0, 204, 9, 9, 192, 197, 226, 205, 9, 198, 161, 226, 9, 10, 192, 197, 98, 227, 9, 10, 192, 197, 162, 199, 9, 6, 0, 200, 9, 10, 192, 197, 34, 203, 9, 9, 192, 197, 162, 129, 66, 197, 98, 190, 9, 9, 192, 197, 162, 215, 9, 9, 192, 197, 226, 193, 9, 12, 0, 193, 9, 8, 192, 197, 98, 194, 9, 8, 192, 197, 162, 195, 9, 8, 192, 197, 226, 196, 9, 9, 192, 197, 34, 190, 9, 7, 192, 197, 162, 191, 9, 7, 192, 197, 226, 192, 9, 8, 192, 197, 34, 14, 0, 197, 34, 14, 0, 196, 9, 34, 0, 204, 9, 13, 0, 204, 9, 20, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 7, 226, 9, 20, 192, 197, 162, 227, 9, 20, 192, 197, 226, 196, 9, 19, 192, 197, 98, 199, 9, 6, 0, 200, 9, 20, 192, 197, 98, 203, 9, 19, 192, 197, 226, 1, 69, 197, 162, 190, 9, 19, 192, 197, 226, 215, 9, 20, 192, 197, 34, 192, 9, 12, 0, 192, 9, 18, 192, 197, 98, 193, 9, 18, 192, 197, 162, 194, 9, 18, 192, 197, 226, 195, 9, 19, 192, 197, 34, 188, 9, 6, 0, 190, 9, 17, 192, 197, 226, 191, 9, 18, 192, 197, 34, 205, 67, 197, 98, 199, 9, 30, 0, 204, 9, 12, 0, 204, 9, 16, 192, 197, 226, 205, 9, 0, 192, 198, 65, 226, 9, 17, 192, 197, 98, 227, 9, 17, 192, 197, 162, 199, 9, 6, 0, 200, 9, 17, 192, 197, 34, 203, 9, 16, 192, 197, 162, 65, 68, 197, 98, 190, 9, 16, 192, 197, 162, 215, 9, 16, 192, 197, 226, 193, 9, 12, 0, 193, 9, 15, 192, 197, 98, 194, 9, 15, 192, 197, 162, 195, 9, 15, 192, 197, 226, 196, 9, 16, 192, 197, 34, 190, 9, 14, 192, 197, 162, 191, 9, 14, 192, 197, 226, 192, 9, 15, 192, 197, 34, 21, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 23, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 8, 226, 9, 24, 192, 197, 34, 227, 9, 24, 192, 197, 98, 199, 9, 6, 0, 200, 9, 23, 192, 197, 226, 203, 9, 23, 192, 197, 98, 1, 70, 197, 34, 190, 9, 23, 192, 197, 98, 215, 9, 23, 192, 197, 162, 193, 9, 12, 0, 193, 9, 22, 192, 197, 34, 194, 9, 22, 192, 197, 98, 195, 9, 22, 192, 197, 162, 196, 9, 22, 192, 197, 226, 190, 9, 21, 192, 197, 98, 191, 9, 21, 192, 197, 162, 192, 9, 21, 192, 197, 226, 24, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 27, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 9, 226, 9, 27, 192, 197, 162, 227, 9, 27, 192, 197, 226, 199, 9, 6, 0, 200, 9, 27, 192, 197, 98, 203, 9, 26, 192, 197, 226, 193, 70, 197, 162, 190, 9, 26, 192, 197, 226, 215, 9, 27, 192, 197, 34, 193, 9, 12, 0, 193, 9, 25, 192, 197, 162, 194, 9, 25, 192, 197, 226, 195, 9, 26, 192, 197, 34, 196, 9, 26, 192, 197, 98, 190, 9, 24, 192, 197, 226, 191, 9, 25, 192, 197, 34, 192, 9, 25, 192, 197, 98, 28, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 30, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 10, 226, 9, 31, 192, 197, 34, 227, 9, 31, 192, 197, 98, 199, 9, 6, 0, 200, 9, 30, 192, 197, 226, 203, 9, 30, 192, 197, 98, 193, 71, 197, 34, 190, 9, 30, 192, 197, 98, 215, 9, 30, 192, 197, 162, 193, 9, 12, 0, 193, 9, 29, 192, 197, 34, 194, 9, 29, 192, 197, 98, 195, 9, 29, 192, 197, 162, 196, 9, 29, 192, 197, 226, 190, 9, 28, 192, 197, 98, 191, 9, 28, 192, 197, 162, 192, 9, 28, 192, 197, 226, 31, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 34, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 11, 226, 9, 34, 192, 197, 162, 227, 9, 34, 192, 197, 226, 199, 9, 6, 0, 200, 9, 34, 192, 197, 98, 203, 9, 33, 192, 197, 226, 129, 72, 197, 162, 190, 9, 33, 192, 197, 226, 215, 9, 34, 192, 197, 34, 193, 9, 12, 0, 193, 9, 32, 192, 197, 162, 194, 9, 32, 192, 197, 226, 195, 9, 33, 192, 197, 34, 196, 9, 33, 192, 197, 98, 190, 9, 31, 192, 197, 226, 191, 9, 32, 192, 197, 34, 192, 9, 32, 192, 197, 98, 35, 0, 197, 34, 14, 0, 196, 9, 34, 0, 204, 9, 13, 0, 204, 9, 41, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 12, 226, 9, 41, 192, 197, 162, 227, 9, 41, 192, 197, 226, 196, 9, 40, 192, 197, 98, 199, 9, 6, 0, 200, 9, 41, 192, 197, 98, 203, 9, 40, 192, 197, 226, 65, 74, 197, 162, 190, 9, 40, 192, 197, 226, 215, 9, 41, 192, 197, 34, 192, 9, 12, 0, 192, 9, 39, 192, 197, 98, 193, 9, 39, 192, 197, 162, 194, 9, 39, 192, 197, 226, 195, 9, 40, 192, 197, 34, 188, 9, 6, 0, 190, 9, 38, 192, 197, 226, 191, 9, 39, 192, 197, 34, 13, 73, 197, 98, 199, 9, 30, 0, 204, 9, 12, 0, 204, 9, 37, 192, 197, 226, 205, 9, 0, 192, 198, 97, 226, 9, 38, 192, 197, 98, 227, 9, 38, 192, 197, 162, 199, 9, 6, 0, 200, 9, 38, 192, 197, 34, 203, 9, 37, 192, 197, 162, 129, 73, 197, 98, 190, 9, 37, 192, 197, 162, 215, 9, 37, 192, 197, 226, 193, 9, 12, 0, 193, 9, 36, 192, 197, 98, 194, 9, 36, 192, 197, 162, 195, 9, 36, 192, 197, 226, 196, 9, 37, 192, 197, 34, 190, 9, 35, 192, 197, 162, 191, 9, 35, 192, 197, 226, 192, 9, 36, 192, 197, 34, 42, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 44, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 13, 226, 9, 45, 192, 197, 34, 227, 9, 45, 192, 197, 98, 199, 9, 6, 0, 200, 9, 44, 192, 197, 226, 203, 9, 44, 192, 197, 98, 65, 75, 197, 34, 190, 9, 44, 192, 197, 98, 215, 9, 44, 192, 197, 162, 193, 9, 12, 0, 193, 9, 43, 192, 197, 34, 194, 9, 43, 192, 197, 98, 195, 9, 43, 192, 197, 162, 196, 9, 43, 192, 197, 226, 190, 9, 42, 192, 197, 98, 191, 9, 42, 192, 197, 162, 192, 9, 42, 192, 197, 226, 45, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 48, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 14, 226, 9, 48, 192, 197, 162, 227, 9, 48, 192, 197, 226, 199, 9, 6, 0, 200, 9, 48, 192, 197, 98, 203, 9, 47, 192, 197, 226, 1, 76, 197, 162, 190, 9, 47, 192, 197, 226, 215, 9, 48, 192, 197, 34, 193, 9, 12, 0, 193, 9, 46, 192, 197, 162, 194, 9, 46, 192, 197, 226, 195, 9, 47, 192, 197, 34, 196, 9, 47, 192, 197, 98, 190, 9, 45, 192, 197, 226, 191, 9, 46, 192, 197, 34, 192, 9, 46, 192, 197, 98, 49, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 51, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 15, 226, 9, 52, 192, 197, 34, 227, 9, 52, 192, 197, 98, 199, 9, 6, 0, 200, 9, 51, 192, 197, 226, 203, 9, 51, 192, 197, 98, 1, 77, 197, 34, 190, 9, 51, 192, 197, 98, 215, 9, 51, 192, 197, 162, 193, 9, 12, 0, 193, 9, 50, 192, 197, 34, 194, 9, 50, 192, 197, 98, 195, 9, 50, 192, 197, 162, 196, 9, 50, 192, 197, 226, 190, 9, 49, 192, 197, 98, 191, 9, 49, 192, 197, 162, 192, 9, 49, 192, 197, 226, 52, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 55, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 16, 226, 9, 55, 192, 197, 162, 227, 9, 55, 192, 197, 226, 199, 9, 6, 0, 200, 9, 55, 192, 197, 98, 203, 9, 54, 192, 197, 226, 193, 77, 197, 162, 190, 9, 54, 192, 197, 226, 215, 9, 55, 192, 197, 34, 193, 9, 12, 0, 193, 9, 53, 192, 197, 162, 194, 9, 53, 192, 197, 226, 195, 9, 54, 192, 197, 34, 196, 9, 54, 192, 197, 98, 190, 9, 52, 192, 197, 226, 191, 9, 53, 192, 197, 34, 192, 9, 53, 192, 197, 98, 56, 0, 197, 34, 14, 0, 196, 9, 34, 0, 204, 9, 13, 0, 204, 9, 62, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 17, 226, 9, 62, 192, 197, 162, 227, 9, 62, 192, 197, 226, 196, 9, 61, 192, 197, 98, 199, 9, 6, 0, 200, 9, 62, 192, 197, 98, 203, 9, 61, 192, 197, 226, 129, 79, 197, 162, 190, 9, 61, 192, 197, 226, 215, 9, 62, 192, 197, 34, 192, 9, 12, 0, 192, 9, 60, 192, 197, 98, 193, 9, 60, 192, 197, 162, 194, 9, 60, 192, 197, 226, 195, 9, 61, 192, 197, 34, 188, 9, 6, 0, 190, 9, 59, 192, 197, 226, 191, 9, 60, 192, 197, 34, 77, 78, 197, 98, 199, 9, 30, 0, 204, 9, 12, 0, 204, 9, 58, 192, 197, 226, 205, 9, 0, 192, 198, 129, 226, 9, 59, 192, 197, 98, 227, 9, 59, 192, 197, 162, 199, 9, 6, 0, 200, 9, 59, 192, 197, 34, 203, 9, 58, 192, 197, 162, 193, 78, 197, 98, 190, 9, 58, 192, 197, 162, 215, 9, 58, 192, 197, 226, 193, 9, 12, 0, 193, 9, 57, 192, 197, 98, 194, 9, 57, 192, 197, 162, 195, 9, 57, 192, 197, 226, 196, 9, 58, 192, 197, 34, 190, 9, 56, 192, 197, 162, 191, 9, 56, 192, 197, 226, 192, 9, 57, 192, 197, 34, 63, 0, 197, 34, 14, 0, 196, 9, 34, 0, 204, 9, 13, 0, 204, 9, 69, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 18, 226, 9, 69, 192, 197, 162, 227, 9, 69, 192, 197, 226, 196, 9, 68, 192, 197, 98, 199, 9, 6, 0, 200, 9, 69, 192, 197, 98, 203, 9, 68, 192, 197, 226, 65, 81, 197, 162, 190, 9, 68, 192, 197, 226, 215, 9, 69, 192, 197, 34, 192, 9, 12, 0, 192, 9, 67, 192, 197, 98, 193, 9, 67, 192, 197, 162, 194, 9, 67, 192, 197, 226, 195, 9, 68, 192, 197, 34, 188, 9, 6, 0, 190, 9, 66, 192, 197, 226, 191, 9, 67, 192, 197, 34, 13, 80, 197, 98, 199, 9, 30, 0, 204, 9, 12, 0, 204, 9, 65, 192, 197, 226, 205, 9, 0, 192, 198, 161, 226, 9, 66, 192, 197, 98, 227, 9, 66, 192, 197, 162, 199, 9, 6, 0, 200, 9, 66, 192, 197, 34, 203, 9, 65, 192, 197, 162, 129, 80, 197, 98, 190, 9, 65, 192, 197, 162, 215, 9, 65, 192, 197, 226, 193, 9, 12, 0, 193, 9, 64, 192, 197, 98, 194, 9, 64, 192, 197, 162, 195, 9, 64, 192, 197, 226, 196, 9, 65, 192, 197, 34, 190, 9, 63, 192, 197, 162, 191, 9, 63, 192, 197, 226, 192, 9, 64, 192, 197, 34, 70, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 72, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 19, 226, 9, 73, 192, 197, 34, 227, 9, 73, 192, 197, 98, 199, 9, 6, 0, 200, 9, 72, 192, 197, 226, 203, 9, 72, 192, 197, 98, 65, 82, 197, 34, 190, 9, 72, 192, 197, 98, 215, 9, 72, 192, 197, 162, 193, 9, 12, 0, 193, 9, 71, 192, 197, 34, 194, 9, 71, 192, 197, 98, 195, 9, 71, 192, 197, 162, 196, 9, 71, 192, 197, 226, 190, 9, 70, 192, 197, 98, 191, 9, 70, 192, 197, 162, 192, 9, 70, 192, 197, 226, 0, 0, 198, 194, 13, 0, 199, 9, 35, 0, 204, 9, 17, 0, 204, 9, 3, 192, 198, 66, 205, 9, 6, 0, 226, 9, 3, 192, 198, 194, 227, 9, 4, 192, 198, 2, 112, 64, 198, 193, 13, 32, 255, 255, 35, 105, 193, 20, 199, 9, 6, 0, 200, 9, 3, 192, 198, 130, 203, 9, 3, 192, 198, 2, 193, 64, 198, 194, 190, 9, 3, 192, 198, 2, 215, 9, 3, 192, 198, 66, 193, 9, 12, 0, 193, 9, 1, 192, 198, 194, 194, 9, 2, 192, 198, 2, 195, 9, 2, 192, 198, 66, 196, 9, 2, 192, 198, 130, 190, 9, 1, 192, 198, 2, 191, 9, 1, 192, 198, 66, 192, 9, 1, 192, 198, 130, 73, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 76, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 21, 226, 9, 76, 192, 197, 162, 227, 9, 76, 192, 197, 226, 199, 9, 6, 0, 200, 9, 76, 192, 197, 98, 203, 9, 75, 192, 197, 226, 1, 83, 197, 162, 190, 9, 75, 192, 197, 226, 215, 9, 76, 192, 197, 34, 193, 9, 12, 0, 193, 9, 74, 192, 197, 162, 194, 9, 74, 192, 197, 226, 195, 9, 75, 192, 197, 34, 196, 9, 75, 192, 197, 98, 190, 9, 73, 192, 197, 226, 191, 9, 74, 192, 197, 34, 192, 9, 74, 192, 197, 98, 77, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 79, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 22, 226, 9, 80, 192, 197, 34, 227, 9, 80, 192, 197, 98, 199, 9, 6, 0, 200, 9, 79, 192, 197, 226, 203, 9, 79, 192, 197, 98, 1, 84, 197, 34, 190, 9, 79, 192, 197, 98, 215, 9, 79, 192, 197, 162, 193, 9, 12, 0, 193, 9, 78, 192, 197, 34, 194, 9, 78, 192, 197, 98, 195, 9, 78, 192, 197, 162, 196, 9, 78, 192, 197, 226, 190, 9, 77, 192, 197, 98, 191, 9, 77, 192, 197, 162, 192, 9, 77, 192, 197, 226, 80, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 83, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 23, 226, 9, 83, 192, 197, 162, 227, 9, 83, 192, 197, 226, 199, 9, 6, 0, 200, 9, 83, 192, 197, 98, 203, 9, 82, 192, 197, 226, 193, 84, 197, 162, 190, 9, 82, 192, 197, 226, 215, 9, 83, 192, 197, 34, 193, 9, 12, 0, 193, 9, 81, 192, 197, 162, 194, 9, 81, 192, 197, 226, 195, 9, 82, 192, 197, 34, 196, 9, 82, 192, 197, 98, 190, 9, 80, 192, 197, 226, 191, 9, 81, 192, 197, 34, 192, 9, 81, 192, 197, 98, 84, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 86, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 24, 226, 9, 87, 192, 197, 34, 227, 9, 87, 192, 197, 98, 199, 9, 6, 0, 200, 9, 86, 192, 197, 226, 203, 9, 86, 192, 197, 98, 193, 85, 197, 34, 190, 9, 86, 192, 197, 98, 215, 9, 86, 192, 197, 162, 193, 9, 12, 0, 193, 9, 85, 192, 197, 34, 194, 9, 85, 192, 197, 98, 195, 9, 85, 192, 197, 162, 196, 9, 85, 192, 197, 226, 190, 9, 84, 192, 197, 98, 191, 9, 84, 192, 197, 162, 192, 9, 84, 192, 197, 226, 87, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 90, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 25, 226, 9, 90, 192, 197, 162, 227, 9, 90, 192, 197, 226, 199, 9, 6, 0, 200, 9, 90, 192, 197, 98, 203, 9, 89, 192, 197, 226, 129, 86, 197, 162, 190, 9, 89, 192, 197, 226, 215, 9, 90, 192, 197, 34, 193, 9, 12, 0, 193, 9, 88, 192, 197, 162, 194, 9, 88, 192, 197, 226, 195, 9, 89, 192, 197, 34, 196, 9, 89, 192, 197, 98, 190, 9, 87, 192, 197, 226, 191, 9, 88, 192, 197, 34, 192, 9, 88, 192, 197, 98, 91, 0, 197, 34, 14, 0, 196, 9, 34, 0, 204, 9, 13, 0, 204, 9, 97, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 26, 226, 9, 97, 192, 197, 162, 227, 9, 97, 192, 197, 226, 196, 9, 96, 192, 197, 98, 199, 9, 6, 0, 200, 9, 97, 192, 197, 98, 203, 9, 96, 192, 197, 226, 65, 88, 197, 162, 190, 9, 96, 192, 197, 226, 215, 9, 97, 192, 197, 34, 192, 9, 12, 0, 192, 9, 95, 192, 197, 98, 193, 9, 95, 192, 197, 162, 194, 9, 95, 192, 197, 226, 195, 9, 96, 192, 197, 34, 188, 9, 6, 0, 190, 9, 94, 192, 197, 226, 191, 9, 95, 192, 197, 34, 13, 87, 197, 98, 199, 9, 30, 0, 204, 9, 12, 0, 204, 9, 93, 192, 197, 226, 205, 9, 4, 192, 198, 65, 226, 9, 94, 192, 197, 98, 227, 9, 94, 192, 197, 162, 199, 9, 6, 0, 200, 9, 94, 192, 197, 34, 203, 9, 93, 192, 197, 162, 129, 87, 197, 98, 190, 9, 93, 192, 197, 162, 215, 9, 93, 192, 197, 226, 193, 9, 12, 0, 193, 9, 92, 192, 197, 98, 194, 9, 92, 192, 197, 162, 195, 9, 92, 192, 197, 226, 196, 9, 93, 192, 197, 34, 190, 9, 91, 192, 197, 162, 191, 9, 91, 192, 197, 226, 192, 9, 92, 192, 197, 34, 98, 0, 197, 34, 14, 0, 196, 9, 34, 0, 204, 9, 13, 0, 204, 9, 104, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 27, 226, 9, 104, 192, 197, 162, 227, 9, 104, 192, 197, 226, 196, 9, 103, 192, 197, 98, 199, 9, 6, 0, 200, 9, 104, 192, 197, 98, 203, 9, 103, 192, 197, 226, 1, 90, 197, 162, 190, 9, 103, 192, 197, 226, 215, 9, 104, 192, 197, 34, 192, 9, 12, 0, 192, 9, 102, 192, 197, 98, 193, 9, 102, 192, 197, 162, 194, 9, 102, 192, 197, 226, 195, 9, 103, 192, 197, 34, 188, 9, 6, 0, 190, 9, 101, 192, 197, 226, 191, 9, 102, 192, 197, 34, 205, 88, 197, 98, 199, 9, 30, 0, 204, 9, 12, 0, 204, 9, 100, 192, 197, 226, 205, 9, 4, 192, 198, 97, 226, 9, 101, 192, 197, 98, 227, 9, 101, 192, 197, 162, 199, 9, 6, 0, 200, 9, 101, 192, 197, 34, 203, 9, 100, 192, 197, 162, 65, 89, 197, 98, 190, 9, 100, 192, 197, 162, 215, 9, 100, 192, 197, 226, 193, 9, 12, 0, 193, 9, 99, 192, 197, 98, 194, 9, 99, 192, 197, 162, 195, 9, 99, 192, 197, 226, 196, 9, 100, 192, 197, 34, 190, 9, 98, 192, 197, 162, 191, 9, 98, 192, 197, 226, 192, 9, 99, 192, 197, 34, 105, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 107, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 28, 226, 9, 108, 192, 197, 34, 227, 9, 108, 192, 197, 98, 199, 9, 6, 0, 200, 9, 107, 192, 197, 226, 203, 9, 107, 192, 197, 98, 1, 91, 197, 34, 190, 9, 107, 192, 197, 98, 215, 9, 107, 192, 197, 162, 193, 9, 12, 0, 193, 9, 106, 192, 197, 34, 194, 9, 106, 192, 197, 98, 195, 9, 106, 192, 197, 162, 196, 9, 106, 192, 197, 226, 190, 9, 105, 192, 197, 98, 191, 9, 105, 192, 197, 162, 192, 9, 105, 192, 197, 226, 108, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 111, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 29, 226, 9, 111, 192, 197, 162, 227, 9, 111, 192, 197, 226, 199, 9, 6, 0, 200, 9, 111, 192, 197, 98, 203, 9, 110, 192, 197, 226, 193, 91, 197, 162, 190, 9, 110, 192, 197, 226, 215, 9, 111, 192, 197, 34, 193, 9, 12, 0, 193, 9, 109, 192, 197, 162, 194, 9, 109, 192, 197, 226, 195, 9, 110, 192, 197, 34, 196, 9, 110, 192, 197, 98, 190, 9, 108, 192, 197, 226, 191, 9, 109, 192, 197, 34, 192, 9, 109, 192, 197, 98, 112, 0, 197, 34, 14, 0, 196, 9, 34, 0, 204, 9, 13, 0, 204, 9, 118, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 30, 226, 9, 118, 192, 197, 162, 227, 9, 118, 192, 197, 226, 196, 9, 117, 192, 197, 98, 199, 9, 6, 0, 200, 9, 118, 192, 197, 98, 203, 9, 117, 192, 197, 226, 129, 93, 197, 162, 190, 9, 117, 192, 197, 226, 215, 9, 118, 192, 197, 34, 192, 9, 12, 0, 192, 9, 116, 192, 197, 98, 193, 9, 116, 192, 197, 162, 194, 9, 116, 192, 197, 226, 195, 9, 117, 192, 197, 34, 188, 9, 6, 0, 190, 9, 115, 192, 197, 226, 191, 9, 116, 192, 197, 34, 77, 92, 197, 98, 199, 9, 30, 0, 204, 9, 12, 0, 204, 9, 114, 192, 197, 226, 205, 9, 4, 192, 198, 129, 226, 9, 115, 192, 197, 98, 227, 9, 115, 192, 197, 162, 199, 9, 6, 0, 200, 9, 115, 192, 197, 34, 203, 9, 114, 192, 197, 162, 193, 92, 197, 98, 190, 9, 114, 192, 197, 162, 215, 9, 114, 192, 197, 226, 193, 9, 12, 0, 193, 9, 113, 192, 197, 98, 194, 9, 113, 192, 197, 162, 195, 9, 113, 192, 197, 226, 196, 9, 114, 192, 197, 34, 190, 9, 112, 192, 197, 162, 191, 9, 112, 192, 197, 226, 192, 9, 113, 192, 197, 34, 119, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 121, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 31, 226, 9, 122, 192, 197, 34, 227, 9, 122, 192, 197, 98, 199, 9, 6, 0, 200, 9, 121, 192, 197, 226, 203, 9, 121, 192, 197, 98, 129, 94, 197, 34, 190, 9, 121, 192, 197, 98, 215, 9, 121, 192, 197, 162, 193, 9, 12, 0, 193, 9, 120, 192, 197, 34, 194, 9, 120, 192, 197, 98, 195, 9, 120, 192, 197, 162, 196, 9, 120, 192, 197, 226, 190, 9, 119, 192, 197, 98, 191, 9, 119, 192, 197, 162, 192, 9, 119, 192, 197, 226, 122, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 125, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 33, 226, 9, 125, 192, 197, 162, 227, 9, 125, 192, 197, 226, 199, 9, 6, 0, 200, 9, 125, 192, 197, 98, 203, 9, 124, 192, 197, 226, 65, 95, 197, 162, 190, 9, 124, 192, 197, 226, 215, 9, 125, 192, 197, 34, 193, 9, 12, 0, 193, 9, 123, 192, 197, 162, 194, 9, 123, 192, 197, 226, 195, 9, 124, 192, 197, 34, 196, 9, 124, 192, 197, 98, 190, 9, 122, 192, 197, 226, 191, 9, 123, 192, 197, 34, 192, 9, 123, 192, 197, 98, 126, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 128, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 35, 226, 9, 129, 192, 197, 34, 227, 9, 129, 192, 197, 98, 199, 9, 6, 0, 200, 9, 128, 192, 197, 226, 203, 9, 128, 192, 197, 98, 65, 96, 197, 34, 190, 9, 128, 192, 197, 98, 215, 9, 128, 192, 197, 162, 193, 9, 12, 0, 193, 9, 127, 192, 197, 34, 194, 9, 127, 192, 197, 98, 195, 9, 127, 192, 197, 162, 196, 9, 127, 192, 197, 226, 190, 9, 126, 192, 197, 98, 191, 9, 126, 192, 197, 162, 192, 9, 126, 192, 197, 226, 129, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 132, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 36, 226, 9, 132, 192, 197, 162, 227, 9, 132, 192, 197, 226, 199, 9, 6, 0, 200, 9, 132, 192, 197, 98, 203, 9, 131, 192, 197, 226, 1, 97, 197, 162, 190, 9, 131, 192, 197, 226, 215, 9, 132, 192, 197, 34, 193, 9, 12, 0, 193, 9, 130, 192, 197, 162, 194, 9, 130, 192, 197, 226, 195, 9, 131, 192, 197, 34, 196, 9, 131, 192, 197, 98, 190, 9, 129, 192, 197, 226, 191, 9, 130, 192, 197, 34, 192, 9, 130, 192, 197, 98, 133, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 135, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 37, 226, 9, 136, 192, 197, 34, 227, 9, 136, 192, 197, 98, 199, 9, 6, 0, 200, 9, 135, 192, 197, 226, 203, 9, 135, 192, 197, 98, 1, 98, 197, 34, 190, 9, 135, 192, 197, 98, 215, 9, 135, 192, 197, 162, 193, 9, 12, 0, 193, 9, 134, 192, 197, 34, 194, 9, 134, 192, 197, 98, 195, 9, 134, 192, 197, 162, 196, 9, 134, 192, 197, 226, 190, 9, 133, 192, 197, 98, 191, 9, 133, 192, 197, 162, 192, 9, 133, 192, 197, 226, 136, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 139, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 38, 226, 9, 139, 192, 197, 162, 227, 9, 139, 192, 197, 226, 199, 9, 6, 0, 200, 9, 139, 192, 197, 98, 203, 9, 138, 192, 197, 226, 193, 98, 197, 162, 190, 9, 138, 192, 197, 226, 215, 9, 139, 192, 197, 34, 193, 9, 12, 0, 193, 9, 137, 192, 197, 162, 194, 9, 137, 192, 197, 226, 195, 9, 138, 192, 197, 34, 196, 9, 138, 192, 197, 98, 190, 9, 136, 192, 197, 226, 191, 9, 137, 192, 197, 34, 192, 9, 137, 192, 197, 98, 56, 0, 197, 98, 13, 0, 199, 9, 30, 0, 204, 9, 12, 0, 204, 9, 58, 192, 197, 226, 205, 9, 0, 192, 198, 129, 226, 9, 59, 192, 197, 98, 227, 9, 59, 192, 197, 162, 199, 9, 6, 0, 200, 9, 59, 192, 197, 34, 203, 9, 58, 192, 197, 162, 193, 78, 197, 98, 190, 9, 58, 192, 197, 162, 215, 9, 58, 192, 197, 226, 193, 9, 12, 0, 193, 9, 57, 192, 197, 98, 194, 9, 57, 192, 197, 162, 195, 9, 57, 192, 197, 226, 196, 9, 58, 192, 197, 34, 190, 9, 56, 192, 197, 162, 191, 9, 56, 192, 197, 226, 192, 9, 57, 192, 197, 34, 63, 0, 197, 98, 13, 0, 199, 9, 30, 0, 204, 9, 12, 0, 204, 9, 65, 192, 197, 226, 205, 9, 0, 192, 198, 161, 226, 9, 66, 192, 197, 98, 227, 9, 66, 192, 197, 162, 199, 9, 6, 0, 200, 9, 66, 192, 197, 34, 203, 9, 65, 192, 197, 162, 129, 80, 197, 98, 190, 9, 65, 192, 197, 162, 215, 9, 65, 192, 197, 226, 193, 9, 12, 0, 193, 9, 64, 192, 197, 98, 194, 9, 64, 192, 197, 162, 195, 9, 64, 192, 197, 226, 196, 9, 65, 192, 197, 34, 190, 9, 63, 192, 197, 162, 191, 9, 63, 192, 197, 226, 192, 9, 64, 192, 197, 34, 112, 0, 197, 98, 13, 0, 199, 9, 30, 0, 204, 9, 12, 0, 204, 9, 114, 192, 197, 226, 205, 9, 4, 192, 198, 129, 226, 9, 115, 192, 197, 98, 227, 9, 115, 192, 197, 162, 199, 9, 6, 0, 200, 9, 115, 192, 197, 34, 203, 9, 114, 192, 197, 162, 193, 92, 197, 98, 190, 9, 114, 192, 197, 162, 215, 9, 114, 192, 197, 226, 193, 9, 12, 0, 193, 9, 113, 192, 197, 98, 194, 9, 113, 192, 197, 162, 195, 9, 113, 192, 197, 226, 196, 9, 114, 192, 197, 34, 190, 9, 112, 192, 197, 162, 191, 9, 112, 192, 197, 226, 192, 9, 113, 192, 197, 34, 140, 0, 197, 34, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 142, 192, 197, 162, 205, 9, 255, 255, 35, 105, 193, 32, 226, 9, 143, 192, 197, 34, 227, 9, 143, 192, 197, 98, 199, 9, 6, 0, 200, 9, 142, 192, 197, 226, 203, 9, 142, 192, 197, 98, 193, 99, 197, 34, 190, 9, 142, 192, 197, 98, 215, 9, 142, 192, 197, 162, 193, 9, 12, 0, 193, 9, 141, 192, 197, 34, 194, 9, 141, 192, 197, 98, 195, 9, 141, 192, 197, 162, 196, 9, 141, 192, 197, 226, 190, 9, 140, 192, 197, 98, 191, 9, 140, 192, 197, 162, 192, 9, 140, 192, 197, 226, 143, 0, 197, 162, 13, 0, 199, 9, 31, 0, 204, 9, 13, 0, 204, 9, 146, 192, 197, 34, 205, 9, 255, 255, 35, 105, 193, 34, 226, 9, 146, 192, 197, 162, 227, 9, 146, 192, 197, 226, 199, 9, 6, 0, 200, 9, 146, 192, 197, 98, 203, 9, 145, 192, 197, 226, 129, 100, 197, 162, 190, 9, 145, 192, 197, 226, 215, 9, 146, 192, 197, 34, 193, 9, 12, 0, 193, 9, 144, 192, 197, 162, 194, 9, 144, 192, 197, 226, 195, 9, 145, 192, 197, 34, 196, 9, 145, 192, 197, 98, 190, 9, 143, 192, 197, 226, 191, 9, 144, 192, 197, 34, 192, 9, 144, 192, 197, 98, 10, 0, 191, 9, 197, 9, 199, 9, 201, 9, 203, 9, 205, 9, 226, 9, 228, 9, 13, 32, 14, 32, 3, 0, 69, 0, 60, 0, 32, 0, 215, 15, 0, 16, 93, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 84, 0, 170, 170, 208, 15, 0, 32, 199, 32, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 204, 32, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 136, 19, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 140, 19, 0, 0, 140, 19, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 44, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 104, 0, 114, 0, 93, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 104, 0, 114, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 252, 40, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 113, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 180, 35, 0, 0, 180, 35, 0, 0, 180, 35, 0, 0, 180, 35, 0, 0, 172, 38, 0, 0, 172, 38, 0, 0, 198, 40, 0, 0, 228, 40, 0, 0, 228, 40, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 46, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 89, 0, 97, 0, 48, 0, 56, 0, 105, 0, 113, 0, 115, 0, 123, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 122, 0, 130, 0, 136, 0, 144, 0, 152, 0, 160, 0, 48, 0, 56, 0, 159, 0, 167, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 184, 0, 192, 0, 48, 0, 48, 0, 48, 0, 200, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 208, 0, 48, 0, 48, 0, 216, 0, 48, 0, 222, 0, 230, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 233, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 240, 0, 48, 0, 48, 0, 248, 0, 252, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 4, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 7, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 182, 0, 48, 0, 48, 0, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 15, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 23, 1, 48, 0, 48, 0, 29, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 100, 1, 192, 0, 164, 1, 204, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 232, 1, 32, 2, 96, 2, 192, 0, 124, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 128, 22, 0, 202, 161, 22, 0, 202, 194, 22, 0, 202, 227, 22, 0, 202, 4, 23, 0, 202, 37, 23, 0, 202, 70, 23, 0, 202, 103, 23, 0, 202, 136, 23, 0, 202, 169, 23, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 198, 0, 0, 201, 68, 2, 0, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 224, 4, 0, 156, 5, 0, 66, 201, 64, 7, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 198, 9, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 166, 10, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 134, 11, 0, 201, 4, 13, 0, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 128, 14, 0, 5, 5, 0, 66, 201, 32, 16, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 198, 17, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 166, 18, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 13, 0, 201, 134, 19, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 13, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 13, 0, 201, 6, 21, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 47, 5, 5, 3, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 49, 5, 5, 3, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 22, 0, 201, 70, 23, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 138, 5, 2, 49, 72, 5, 2, 49, 6, 5, 2, 49, 138, 5, 2, 65, 72, 5, 2, 65, 6, 5, 2, 65, 138, 5, 2, 69, 72, 5, 2, 69, 6, 5, 2, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 24, 0, 201, 198, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 134, 25, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 27, 0, 5, 71, 238, 102, 201, 70, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 5, 0, 192, 0, 0, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 5, 0, 192, 0, 0, 0, 197, 34, 6, 0, 192, 0, 0, 0, 197, 98, 6, 0, 192, 0, 0, 0, 197, 162, 6, 0, 192, 0, 0, 0, 197, 226, 6, 0, 197, 35, 7, 0, 197, 130, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 7, 0, 197, 3, 8, 0, 197, 98, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 162, 4, 0, 197, 226, 8, 0, 5, 5, 10, 121, 197, 34, 9, 0, 197, 98, 9, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 226, 10, 0, 197, 34, 11, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 5, 0, 5, 5, 24, 121, 197, 98, 5, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 2, 0, 197, 34, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 29, 0, 201, 230, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 11, 0, 197, 34, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 30, 0, 201, 102, 31, 0, 201, 38, 32, 0, 201, 230, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 137, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 144, 0, 0, 137, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 144, 0, 0, 5, 5, 2, 65, 5, 142, 0, 0, 5, 5, 2, 65, 5, 144, 0, 0, 5, 5, 2, 69, 5, 142, 0, 0, 5, 5, 2, 69, 5, 144, 0, 0, 134, 5, 3, 47, 5, 160, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 3, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 162, 4, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 226, 4, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 34, 5, 0, 5, 5, 24, 121, 197, 98, 5, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 162, 5, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 226, 5, 0, 5, 5, 107, 121, 197, 34, 6, 0, 5, 5, 111, 121, 197, 98, 6, 0, 5, 5, 115, 121, 197, 162, 6, 0, 5, 5, 119, 121, 197, 226, 6, 0, 197, 35, 7, 0, 197, 130, 7, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 194, 7, 0, 197, 3, 8, 0, 197, 98, 8, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 162, 8, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 162, 4, 0, 197, 226, 8, 0, 5, 5, 10, 121, 197, 34, 9, 0, 197, 98, 9, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 226, 10, 0, 197, 34, 11, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 5, 0, 5, 5, 24, 121, 197, 98, 5, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 46, 156, 5, 2, 0, 1, 3, 3, 239, 134, 5, 12, 3, 2, 239, 134, 5, 65, 3, 3, 239, 134, 5, 0, 48, 156, 5, 3, 0, 90, 0, 8, 0, 122, 0, 10, 0, 125, 1, 2, 241, 137, 5, 126, 1, 2, 241, 71, 5, 48, 0, 12, 3, 2, 241, 137, 5, 48, 0, 12, 3, 2, 241, 71, 5, 0, 64, 156, 5, 4, 0, 74, 0, 255, 255, 2, 65, 137, 5, 106, 0, 255, 255, 2, 65, 71, 5, 52, 1, 197, 162, 53, 1, 0, 192, 197, 98, 240, 1, 0, 192, 197, 162, 0, 68, 156, 5, 4, 0, 74, 0, 255, 255, 2, 69, 137, 5, 106, 0, 255, 255, 2, 69, 71, 5, 52, 1, 0, 192, 197, 226, 53, 1, 1, 192, 197, 34, 240, 1, 1, 192, 197, 98, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 46, 5, 5, 2, 0, 1, 3, 3, 239, 5, 5, 12, 3, 2, 239, 5, 5, 65, 3, 3, 239, 5, 5, 0, 48, 5, 5, 1, 0, 122, 0, 3, 0, 126, 1, 2, 241, 5, 5, 48, 0, 12, 3, 2, 241, 5, 5, 0, 64, 5, 5, 2, 0, 106, 0, 255, 255, 2, 65, 5, 5, 53, 1, 1, 192, 197, 162, 240, 1, 1, 192, 197, 226, 0, 68, 5, 5, 2, 0, 106, 0, 255, 255, 2, 69, 5, 5, 53, 1, 2, 192, 197, 34, 240, 1, 2, 192, 197, 98, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 2, 0, 1, 3, 2, 192, 197, 162, 12, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 5, 46, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 12, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 156, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 226, 156, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 98, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 78, 196, 196, 48, 0, 12, 3, 11, 192, 197, 98, 5, 78, 196, 196, 48, 0, 12, 3, 11, 192, 197, 162, 156, 92, 196, 196, 48, 0, 12, 3, 12, 192, 197, 98, 5, 92, 196, 196, 48, 0, 12, 3, 12, 192, 197, 162, 156, 92, 196, 210, 48, 0, 12, 3, 12, 192, 197, 226, 5, 92, 196, 210, 48, 0, 12, 3, 13, 192, 197, 34, 14, 0, 74, 0, 75, 0, 90, 0, 91, 0, 106, 0, 107, 0, 122, 0, 123, 0, 52, 1, 54, 1, 125, 1, 127, 1, 240, 1, 241, 1, 3, 0, 67, 0, 58, 0, 32, 0, 47, 0, 0, 16, 18, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 28, 36, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 76, 0, 0, 0, 76, 1, 0, 0, 188, 28, 0, 0, 188, 28, 0, 0, 188, 28, 0, 0, 188, 28, 0, 0, 208, 29, 0, 0, 208, 29, 0, 0, 164, 31, 0, 0, 194, 31, 0, 0, 4, 36, 0, 0, 25, 0, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 94, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 174, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 89, 0, 97, 0, 48, 0, 56, 0, 105, 0, 113, 0, 115, 0, 123, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 145, 0, 48, 0, 48, 0, 153, 0, 157, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 139, 0, 48, 0, 48, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 100, 1, 192, 0, 164, 1, 204, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 7, 0, 202, 129, 7, 0, 202, 162, 7, 0, 202, 195, 7, 0, 202, 228, 7, 0, 202, 5, 8, 0, 202, 38, 8, 0, 202, 71, 8, 0, 202, 104, 8, 0, 202, 137, 8, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 201, 132, 1, 0, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 32, 4, 0, 156, 5, 0, 66, 201, 128, 6, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 6, 9, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 230, 9, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 198, 10, 0, 201, 68, 12, 0, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 192, 13, 0, 5, 5, 0, 66, 201, 96, 15, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 6, 17, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 230, 17, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 6, 0, 201, 198, 18, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 6, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 7, 0, 201, 70, 20, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 47, 5, 5, 3, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 49, 5, 5, 3, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 21, 0, 201, 134, 22, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 138, 5, 2, 49, 72, 5, 2, 49, 6, 5, 2, 49, 138, 5, 2, 65, 72, 5, 2, 65, 6, 5, 2, 65, 138, 5, 2, 69, 72, 5, 2, 69, 6, 5, 2, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 23, 0, 201, 6, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 2, 0, 197, 34, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 24, 0, 201, 134, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 5, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 26, 0, 201, 6, 27, 0, 201, 198, 27, 0, 201, 134, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 137, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 144, 0, 0, 137, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 144, 0, 0, 5, 5, 2, 65, 5, 142, 0, 0, 5, 5, 2, 65, 5, 144, 0, 0, 5, 5, 2, 69, 5, 142, 0, 0, 5, 5, 2, 69, 5, 144, 0, 0, 134, 5, 3, 47, 5, 160, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 3, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 2, 0, 1, 3, 3, 239, 134, 5, 12, 3, 2, 239, 134, 5, 65, 3, 3, 239, 134, 5, 0, 48, 156, 5, 3, 0, 90, 0, 8, 0, 122, 0, 10, 0, 125, 1, 2, 241, 137, 5, 126, 1, 2, 241, 71, 5, 48, 0, 12, 3, 2, 241, 137, 5, 48, 0, 12, 3, 2, 241, 71, 5, 0, 64, 156, 5, 4, 0, 74, 0, 255, 255, 2, 65, 137, 5, 106, 0, 255, 255, 2, 65, 71, 5, 52, 1, 197, 162, 53, 1, 0, 192, 197, 98, 240, 1, 0, 192, 197, 162, 0, 68, 156, 5, 4, 0, 74, 0, 255, 255, 2, 69, 137, 5, 106, 0, 255, 255, 2, 69, 71, 5, 52, 1, 0, 192, 197, 226, 53, 1, 1, 192, 197, 34, 240, 1, 1, 192, 197, 98, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 46, 5, 5, 2, 0, 1, 3, 3, 239, 5, 5, 12, 3, 2, 239, 5, 5, 65, 3, 3, 239, 5, 5, 0, 48, 5, 5, 1, 0, 122, 0, 3, 0, 126, 1, 2, 241, 5, 5, 48, 0, 12, 3, 2, 241, 5, 5, 0, 64, 5, 5, 2, 0, 106, 0, 255, 255, 2, 65, 5, 5, 53, 1, 1, 192, 197, 162, 240, 1, 1, 192, 197, 226, 0, 68, 5, 5, 2, 0, 106, 0, 255, 255, 2, 69, 5, 5, 53, 1, 2, 192, 197, 34, 240, 1, 2, 192, 197, 98, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 2, 0, 1, 3, 2, 192, 197, 162, 12, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 5, 46, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 12, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 156, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 226, 156, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 98, 156, 78, 196, 196, 48, 0, 12, 3, 4, 192, 197, 162, 5, 78, 196, 196, 48, 0, 12, 3, 4, 192, 197, 226, 156, 92, 196, 196, 48, 0, 12, 3, 5, 192, 197, 162, 5, 92, 196, 196, 48, 0, 12, 3, 5, 192, 197, 226, 156, 92, 196, 210, 48, 0, 12, 3, 6, 192, 197, 34, 5, 92, 196, 210, 48, 0, 12, 3, 6, 192, 197, 98, 14, 0, 74, 0, 75, 0, 90, 0, 91, 0, 106, 0, 107, 0, 122, 0, 123, 0, 52, 1, 54, 1, 125, 1, 127, 1, 240, 1, 241, 1, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 28, 4, 186, 88, 185, 92, 185, 96, 185, 100, 186, 104, 186, 112, 185, 116, 36, 4, 185, 128, 46, 4, 185, 144, 185, 148, 185, 152, 185, 160, 56, 4, 185, 172, 185, 180, 185, 184, 185, 188, 185, 192, 185, 196, 58, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 60, 4, 62, 4, 168, 88, 168, 92, 168, 96, 168, 100, 168, 104, 168, 112, 168, 116, 67, 4, 168, 128, 73, 4, 168, 144, 168, 148, 168, 152, 168, 160, 79, 4, 168, 172, 168, 180, 168, 184, 168, 188, 168, 192, 168, 196, 81, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 144, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 83, 4, 186, 89, 154, 89, 250, 89, 90, 90, 186, 105, 154, 105, 250, 105, 90, 106, 217, 76, 154, 134, 185, 145, 153, 145, 249, 145, 153, 146, 89, 146, 136, 14, 217, 146, 185, 181, 153, 181, 249, 181, 89, 182, 153, 197, 185, 208, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 85, 4, 168, 89, 136, 89, 232, 89, 72, 90, 168, 105, 136, 105, 232, 105, 72, 106, 200, 76, 136, 134, 168, 145, 136, 145, 232, 145, 136, 146, 72, 146, 128, 14, 200, 146, 168, 181, 136, 181, 232, 181, 72, 182, 136, 197, 168, 208, 72, 198, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 185, 72, 168, 72, 249, 65, 232, 65, 185, 66, 168, 66, 185, 68, 168, 68, 25, 78, 8, 78, 185, 84, 168, 84, 58, 91, 40, 91, 218, 89, 200, 89, 186, 90, 168, 90, 26, 91, 8, 91, 26, 90, 8, 90, 249, 97, 232, 97, 217, 97, 200, 97, 185, 98, 168, 98, 249, 98, 232, 98, 249, 101, 232, 101, 89, 103, 72, 103, 154, 106, 136, 106, 58, 107, 40, 107, 218, 105, 200, 105, 26, 107, 8, 107, 186, 106, 168, 108, 6, 8, 8, 8, 250, 113, 232, 113, 249, 118, 232, 118, 168, 156, 153, 121, 136, 121, 249, 122, 232, 122, 25, 122, 8, 122, 121, 123, 104, 123, 89, 123, 72, 123, 154, 133, 136, 133, 250, 134, 232, 134, 26, 134, 8, 134, 10, 8, 185, 140, 168, 140, 57, 147, 40, 147, 217, 145, 200, 145, 121, 146, 104, 146, 12, 8, 14, 8, 153, 161, 136, 161, 249, 162, 232, 162, 25, 162, 8, 162, 154, 165, 136, 165, 250, 165, 232, 165, 87, 4, 89, 4, 185, 168, 168, 168, 249, 174, 232, 174, 25, 174, 8, 174, 185, 176, 168, 176, 153, 182, 136, 182, 57, 183, 40, 183, 217, 181, 200, 181, 57, 182, 40, 182, 121, 182, 104, 182, 25, 183, 8, 183, 249, 189, 232, 189, 249, 197, 232, 197, 89, 198, 153, 201, 136, 201, 185, 202, 168, 202, 185, 204, 168, 204, 232, 164, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 88, 200, 164, 169, 164, 200, 56, 169, 88, 187, 104, 187, 112, 169, 104, 169, 112, 169, 212, 169, 132, 217, 144, 187, 88, 200, 144, 169, 88, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 185, 76, 90, 2, 122, 2, 125, 5, 186, 80, 126, 5, 177, 80, 255, 5, 185, 120, 74, 4, 186, 124, 106, 4, 177, 124, 52, 5, 250, 125, 53, 5, 241, 125, 255, 5, 186, 132, 74, 4, 186, 136, 106, 4, 177, 136, 52, 5, 250, 137, 53, 5, 241, 137, 255, 5, 186, 164, 255, 5, 185, 200, 255, 5, 168, 64, 255, 5, 168, 76, 122, 2, 126, 5, 168, 80, 255, 5, 168, 120, 106, 4, 168, 124, 53, 5, 232, 125, 255, 5, 168, 132, 106, 4, 168, 136, 53, 5, 232, 137, 255, 5, 168, 164, 255, 5, 168, 200, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 250, 166, 255, 5, 232, 166, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 119, 10, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 111, 10, 0, 32, 127, 19, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 132, 19, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 152, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 156, 6, 0, 0, 156, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 30, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 115, 0, 114, 0, 93, 0, 0, 0, 170, 170, 170, 170, 196, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 56, 0, 0, 0, 56, 1, 0, 0, 88, 25, 0, 0, 88, 25, 0, 0, 88, 25, 0, 0, 88, 25, 0, 0, 172, 25, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 39, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 121, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 104, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 1, 0, 202, 129, 1, 0, 202, 162, 1, 0, 202, 195, 1, 0, 202, 228, 1, 0, 202, 5, 2, 0, 202, 38, 2, 0, 202, 71, 2, 0, 202, 104, 2, 0, 202, 137, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 174, 5, 70, 97, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 97, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 174, 5, 70, 97, 5, 140, 0, 0, 5, 5, 70, 97, 5, 140, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 69, 0, 60, 0, 32, 0, 31, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 145, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 150, 6, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 47, 9, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 51, 9, 0, 0, 51, 9, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 44, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 170, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 38, 0, 76, 0, 60, 0, 64, 1, 61, 0, 108, 0, 183, 0, 60, 0, 60, 0, 60, 0, 63, 1, 61, 0, 76, 0, 183, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 224, 35, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 61, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 36, 33, 0, 0, 36, 33, 0, 0, 36, 33, 0, 0, 36, 33, 0, 0, 76, 35, 0, 0, 76, 35, 0, 0, 190, 35, 0, 0, 200, 35, 0, 0, 200, 35, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 5, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 100, 0, 108, 0, 114, 0, 122, 0, 130, 0, 138, 0, 48, 0, 56, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 162, 0, 170, 0, 48, 0, 48, 0, 48, 0, 178, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 186, 0, 48, 0, 48, 0, 194, 0, 48, 0, 200, 0, 208, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 211, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 219, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 222, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 160, 0, 48, 0, 48, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 252, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 230, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 238, 0, 48, 0, 48, 0, 244, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 144, 1, 200, 1, 8, 2, 192, 0, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 16, 0, 202, 33, 16, 0, 202, 66, 16, 0, 202, 99, 16, 0, 202, 132, 16, 0, 202, 165, 16, 0, 202, 198, 16, 0, 202, 231, 16, 0, 202, 8, 17, 0, 202, 41, 17, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 192, 0, 0, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 32, 2, 0, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 6, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 7, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 7, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 65, 5, 5, 2, 65, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 134, 3, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 5, 0, 5, 71, 238, 102, 201, 70, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 0, 0, 192, 0, 0, 0, 197, 226, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 1, 0, 192, 0, 0, 0, 197, 162, 1, 0, 192, 0, 0, 0, 197, 226, 1, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 197, 98, 2, 0, 197, 163, 2, 0, 197, 2, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 3, 0, 197, 131, 3, 0, 197, 226, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 34, 0, 0, 197, 98, 4, 0, 5, 5, 10, 121, 197, 162, 4, 0, 197, 226, 4, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 34, 0, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 98, 0, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 162, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 34, 1, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 98, 1, 0, 5, 5, 107, 121, 197, 162, 1, 0, 5, 5, 111, 121, 197, 226, 1, 0, 5, 5, 115, 121, 197, 34, 2, 0, 5, 5, 119, 121, 197, 98, 2, 0, 197, 163, 2, 0, 197, 2, 3, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 66, 3, 0, 197, 131, 3, 0, 197, 226, 3, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 34, 4, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 34, 0, 0, 197, 98, 4, 0, 5, 5, 10, 121, 197, 162, 4, 0, 197, 226, 4, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 64, 156, 5, 1, 0, 183, 0, 255, 255, 2, 65, 134, 5, 135, 3, 255, 255, 2, 65, 134, 5, 0, 64, 5, 5, 1, 0, 183, 0, 255, 255, 2, 65, 5, 5, 135, 3, 255, 255, 2, 65, 5, 5, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 4, 0, 183, 0, 184, 0, 135, 3, 136, 3, 3, 0, 67, 0, 58, 0, 32, 0, 47, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 40, 9, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 45, 9, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 7, 8, 0, 0, 7, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 35, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 38, 0, 78, 0, 60, 0, 241, 0, 60, 0, 60, 0, 60, 0, 209, 0, 60, 0, 110, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 71, 0, 0, 0, 94, 31, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 0, 26, 0, 0, 0, 26, 0, 0, 32, 27, 0, 0, 58, 27, 0, 0, 70, 31, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 128, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 111, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 104, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 8, 0, 202, 65, 8, 0, 202, 98, 8, 0, 202, 131, 8, 0, 202, 164, 8, 0, 202, 197, 8, 0, 202, 230, 8, 0, 202, 7, 9, 0, 202, 40, 9, 0, 202, 73, 9, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 4, 0, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 196, 7, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 7, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 134, 5, 2, 69, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 7, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 7, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 5, 5, 2, 69, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 12, 0, 201, 198, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 13, 0, 201, 70, 14, 0, 201, 6, 15, 0, 201, 198, 15, 0, 201, 134, 16, 0, 201, 70, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 135, 5, 3, 69, 5, 142, 0, 0, 70, 5, 3, 69, 5, 142, 0, 0, 135, 5, 3, 69, 5, 140, 0, 0, 70, 5, 3, 69, 5, 140, 0, 0, 135, 5, 3, 69, 5, 156, 0, 0, 70, 5, 3, 69, 5, 156, 0, 0, 135, 5, 3, 69, 5, 160, 0, 0, 70, 5, 3, 69, 5, 160, 0, 0, 135, 5, 3, 69, 5, 144, 0, 0, 70, 5, 3, 69, 5, 144, 0, 0, 135, 5, 3, 69, 5, 136, 0, 0, 70, 5, 3, 69, 5, 136, 0, 0, 135, 5, 3, 69, 5, 164, 0, 0, 70, 5, 3, 69, 5, 164, 0, 0, 5, 5, 3, 69, 5, 142, 0, 0, 5, 5, 3, 69, 5, 140, 0, 0, 5, 5, 3, 69, 5, 156, 0, 0, 5, 5, 3, 69, 5, 160, 0, 0, 5, 5, 3, 69, 5, 144, 0, 0, 5, 5, 3, 69, 5, 136, 0, 0, 5, 5, 3, 69, 5, 164, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 68, 156, 5, 16, 0, 34, 1, 30, 0, 244, 1, 16, 0, 244, 1, 2, 192, 197, 162, 245, 1, 2, 192, 197, 226, 3, 3, 255, 255, 2, 69, 134, 5, 32, 30, 3, 192, 197, 34, 33, 30, 3, 192, 197, 98, 34, 1, 1, 192, 197, 162, 35, 1, 1, 192, 197, 226, 230, 1, 2, 192, 197, 34, 231, 1, 2, 192, 197, 98, 30, 1, 12, 0, 30, 1, 0, 192, 197, 162, 31, 1, 0, 192, 197, 226, 32, 1, 1, 192, 197, 34, 33, 1, 1, 192, 197, 98, 71, 0, 255, 255, 3, 69, 135, 5, 103, 0, 255, 255, 3, 69, 70, 5, 28, 1, 197, 162, 29, 1, 0, 192, 197, 98, 0, 68, 5, 5, 8, 0, 35, 1, 16, 0, 35, 1, 4, 192, 197, 98, 231, 1, 4, 192, 197, 162, 245, 1, 4, 192, 197, 226, 3, 3, 255, 255, 2, 69, 5, 5, 33, 30, 5, 192, 197, 34, 103, 0, 255, 255, 3, 69, 5, 5, 29, 1, 3, 192, 197, 162, 31, 1, 3, 192, 197, 226, 33, 1, 4, 192, 197, 34, 156, 68, 196, 160, 48, 0, 3, 3, 5, 192, 197, 98, 5, 68, 196, 160, 48, 0, 3, 3, 5, 192, 197, 162, 156, 68, 196, 196, 48, 0, 3, 3, 5, 192, 197, 226, 5, 68, 196, 196, 48, 0, 3, 3, 6, 192, 197, 34, 156, 68, 196, 210, 48, 0, 3, 3, 6, 192, 197, 98, 5, 68, 196, 210, 48, 0, 3, 3, 6, 192, 197, 162, 156, 68, 196, 204, 48, 0, 3, 3, 6, 192, 197, 226, 5, 68, 196, 204, 48, 0, 3, 3, 7, 192, 197, 34, 12, 0, 71, 0, 72, 0, 103, 0, 104, 0, 28, 1, 36, 1, 230, 1, 232, 1, 244, 1, 246, 1, 32, 30, 34, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 26, 4, 185, 128, 185, 132, 185, 136, 185, 144, 186, 148, 185, 152, 185, 160, 185, 164, 185, 168, 185, 172, 185, 176, 185, 180, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 48, 4, 168, 128, 168, 132, 168, 136, 168, 144, 168, 148, 168, 152, 168, 160, 168, 164, 168, 168, 168, 172, 168, 176, 168, 180, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 128, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 185, 116, 185, 129, 153, 129, 249, 129, 153, 130, 89, 130, 136, 14, 217, 130, 185, 161, 153, 161, 249, 161, 89, 162, 153, 177, 185, 184, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 168, 116, 168, 129, 136, 129, 232, 129, 136, 130, 72, 130, 128, 14, 200, 130, 168, 161, 136, 161, 232, 161, 72, 162, 136, 177, 168, 184, 72, 178, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 140, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 60, 4, 62, 4, 26, 114, 8, 114, 10, 8, 185, 124, 168, 124, 57, 131, 40, 131, 217, 129, 200, 129, 121, 130, 104, 130, 12, 8, 14, 8, 153, 145, 136, 145, 249, 146, 232, 146, 25, 146, 8, 146, 154, 149, 136, 149, 250, 149, 232, 149, 250, 150, 232, 150, 26, 150, 8, 150, 249, 154, 232, 154, 25, 154, 8, 154, 185, 156, 168, 156, 153, 162, 136, 162, 57, 163, 40, 163, 217, 161, 200, 161, 57, 162, 40, 162, 121, 162, 104, 162, 25, 163, 8, 163, 249, 169, 232, 169, 249, 177, 232, 177, 89, 178, 153, 181, 136, 181, 185, 182, 168, 182, 25, 182, 8, 182, 232, 148, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 148, 169, 148, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 188, 169, 112, 217, 128, 187, 72, 200, 128, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 186, 112, 71, 4, 186, 120, 103, 4, 177, 120, 28, 5, 250, 121, 29, 5, 241, 121, 30, 5, 218, 121, 31, 5, 209, 121, 32, 5, 186, 122, 33, 5, 177, 122, 34, 5, 250, 122, 35, 5, 241, 122, 255, 5, 168, 112, 103, 4, 168, 120, 29, 5, 232, 121, 31, 5, 200, 121, 33, 5, 168, 122, 35, 5, 232, 122, 255, 5, 250, 114, 255, 5, 232, 114, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 35, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 252, 7, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 1, 8, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 124, 0, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 128, 0, 0, 0, 128, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 67, 0, 104, 0, 101, 0, 114, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 68, 1, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 8, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 32, 0, 0, 0, 44, 0, 0, 0, 44, 1, 0, 0, 6, 0, 0, 0, 0, 0, 58, 118, 177, 255, 232, 118, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 0, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 3, 0, 69, 0, 60, 0, 32, 0, 35, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 117, 0, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 122, 0, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13, 9, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 17, 9, 0, 0, 17, 9, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 55, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 60, 220, 38, 0, 67, 0, 60, 0, 99, 0, 12, 3, 60, 0, 60, 0, 60, 0, 67, 0, 12, 3, 38, 0, 72, 0, 60, 0, 99, 0, 104, 0, 60, 0, 60, 0, 60, 0, 99, 0, 72, 0, 60, 0, 60, 0, 60, 0, 67, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 72, 0, 38, 0, 82, 0, 60, 0, 114, 0, 12, 3, 60, 0, 60, 0, 60, 0, 82, 0, 12, 3, 38, 0, 83, 0, 60, 0, 115, 0, 12, 3, 60, 0, 60, 0, 60, 0, 83, 0, 12, 3, 38, 0, 90, 0, 60, 0, 122, 0, 12, 3, 60, 0, 60, 0, 60, 0, 90, 0, 12, 3, 0, 0, 56, 35, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 244, 26, 0, 0, 244, 26, 0, 0, 244, 26, 0, 0, 244, 26, 0, 0, 200, 28, 0, 0, 200, 28, 0, 0, 238, 30, 0, 0, 8, 31, 0, 0, 32, 35, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 162, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 91, 0, 99, 0, 48, 0, 56, 0, 48, 0, 56, 0, 107, 0, 115, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 133, 0, 141, 0, 145, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 131, 0, 48, 0, 48, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 108, 1, 192, 0, 192, 0, 172, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 13, 0, 202, 129, 13, 0, 202, 162, 13, 0, 202, 195, 13, 0, 202, 228, 13, 0, 202, 5, 14, 0, 202, 38, 14, 0, 202, 71, 14, 0, 202, 104, 14, 0, 202, 137, 14, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 4, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 201, 198, 7, 0, 201, 166, 8, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 134, 9, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 100, 10, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 201, 70, 18, 0, 201, 38, 19, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 6, 20, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 12, 0, 201, 230, 20, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 12, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 13, 0, 201, 166, 21, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 22, 0, 201, 38, 23, 0, 134, 5, 2, 77, 5, 5, 2, 77, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 23, 0, 201, 166, 24, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 25, 0, 201, 38, 26, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 26, 0, 201, 166, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 28, 0, 201, 38, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 29, 0, 201, 166, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 197, 98, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 31, 0, 201, 38, 32, 0, 201, 230, 32, 0, 201, 166, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 136, 5, 2, 57, 5, 142, 0, 0, 71, 5, 2, 57, 5, 142, 0, 0, 136, 5, 2, 57, 5, 144, 0, 0, 71, 5, 2, 57, 5, 144, 0, 0, 136, 5, 2, 57, 5, 156, 0, 0, 71, 5, 2, 57, 5, 156, 0, 0, 136, 5, 2, 57, 5, 196, 0, 0, 71, 5, 2, 57, 5, 196, 0, 0, 136, 5, 2, 57, 5, 150, 0, 0, 71, 5, 2, 57, 5, 150, 0, 0, 136, 5, 2, 57, 5, 160, 0, 0, 71, 5, 2, 57, 5, 160, 0, 0, 136, 5, 2, 57, 5, 206, 0, 0, 71, 5, 2, 57, 5, 206, 0, 0, 71, 5, 2, 57, 5, 210, 0, 0, 70, 5, 2, 57, 5, 142, 0, 0, 5, 5, 2, 57, 5, 142, 0, 0, 70, 5, 2, 57, 5, 144, 0, 0, 5, 5, 2, 57, 5, 144, 0, 0, 70, 5, 2, 57, 5, 156, 0, 0, 5, 5, 2, 57, 5, 156, 0, 0, 70, 5, 2, 57, 5, 196, 0, 0, 5, 5, 2, 57, 5, 196, 0, 0, 70, 5, 2, 57, 5, 150, 0, 0, 5, 5, 2, 57, 5, 150, 0, 0, 70, 5, 2, 57, 5, 160, 0, 0, 5, 5, 2, 57, 5, 160, 0, 0, 70, 5, 2, 57, 5, 206, 0, 0, 5, 5, 2, 57, 5, 206, 0, 0, 5, 5, 2, 57, 5, 210, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 77, 5, 160, 0, 0, 5, 5, 2, 77, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 77, 5, 196, 0, 0, 5, 5, 2, 77, 5, 196, 0, 0, 134, 5, 2, 77, 5, 210, 0, 0, 5, 5, 2, 77, 5, 210, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 17, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 2, 192, 197, 162, 41, 30, 2, 192, 197, 226, 42, 30, 3, 192, 197, 34, 43, 30, 3, 192, 197, 98, 150, 30, 3, 192, 197, 162, 36, 30, 1, 192, 197, 162, 37, 30, 1, 192, 197, 226, 38, 30, 2, 192, 197, 34, 39, 30, 2, 192, 197, 98, 30, 2, 15, 0, 30, 2, 0, 192, 197, 162, 31, 2, 0, 192, 197, 226, 12, 3, 2, 239, 134, 5, 34, 30, 1, 192, 197, 34, 35, 30, 1, 192, 197, 98, 72, 0, 2, 249, 136, 5, 104, 0, 2, 249, 71, 5, 36, 1, 197, 162, 37, 1, 0, 192, 197, 98, 0, 76, 156, 5, 48, 0, 12, 3, 255, 255, 2, 77, 134, 5, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 46, 5, 5, 17, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 6, 192, 197, 98, 41, 30, 6, 192, 197, 162, 42, 30, 6, 192, 197, 226, 43, 30, 7, 192, 197, 34, 150, 30, 7, 192, 197, 98, 36, 30, 5, 192, 197, 98, 37, 30, 5, 192, 197, 162, 38, 30, 5, 192, 197, 226, 39, 30, 6, 192, 197, 34, 30, 2, 15, 0, 30, 2, 4, 192, 197, 98, 31, 2, 4, 192, 197, 162, 12, 3, 2, 239, 5, 5, 34, 30, 4, 192, 197, 226, 35, 30, 5, 192, 197, 34, 72, 0, 2, 249, 70, 5, 104, 0, 2, 249, 5, 5, 36, 1, 3, 192, 197, 226, 37, 1, 4, 192, 197, 34, 0, 76, 5, 5, 48, 0, 12, 3, 255, 255, 2, 77, 5, 5, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 48, 0, 12, 3, 7, 192, 197, 162, 5, 46, 196, 160, 48, 0, 12, 3, 7, 192, 197, 226, 156, 76, 196, 160, 48, 0, 12, 3, 8, 192, 197, 34, 5, 76, 196, 160, 48, 0, 12, 3, 8, 192, 197, 98, 156, 78, 196, 160, 48, 0, 12, 3, 8, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 8, 192, 197, 226, 156, 78, 196, 202, 48, 0, 12, 3, 9, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 9, 192, 197, 98, 156, 76, 196, 196, 48, 0, 12, 3, 9, 192, 197, 162, 5, 76, 196, 196, 48, 0, 12, 3, 9, 192, 197, 226, 156, 76, 196, 210, 48, 0, 12, 3, 10, 192, 197, 34, 5, 76, 196, 210, 48, 0, 12, 3, 10, 192, 197, 98, 156, 78, 196, 196, 48, 0, 12, 3, 10, 192, 197, 162, 5, 78, 196, 196, 48, 0, 12, 3, 10, 192, 197, 226, 156, 92, 196, 196, 48, 0, 12, 3, 11, 192, 197, 162, 5, 92, 196, 196, 48, 0, 12, 3, 11, 192, 197, 226, 156, 92, 196, 210, 48, 0, 12, 3, 12, 192, 197, 34, 5, 92, 196, 210, 48, 0, 12, 3, 12, 192, 197, 98, 12, 0, 72, 0, 73, 0, 104, 0, 105, 0, 36, 1, 38, 1, 30, 2, 32, 2, 34, 30, 44, 30, 150, 30, 151, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 72, 186, 76, 185, 80, 185, 84, 185, 88, 186, 96, 186, 104, 185, 108, 185, 112, 185, 116, 186, 120, 185, 128, 185, 132, 185, 136, 36, 4, 38, 4, 185, 160, 185, 168, 185, 172, 185, 176, 185, 180, 185, 184, 40, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 42, 4, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 104, 168, 108, 168, 112, 168, 116, 168, 120, 168, 128, 168, 132, 168, 136, 52, 4, 54, 4, 168, 160, 168, 168, 168, 172, 168, 176, 168, 180, 168, 184, 56, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 128, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 58, 4, 186, 77, 154, 77, 250, 77, 90, 78, 186, 97, 154, 97, 250, 97, 90, 98, 217, 72, 154, 122, 185, 129, 153, 129, 249, 129, 153, 130, 89, 130, 136, 14, 217, 130, 185, 169, 153, 169, 249, 169, 89, 170, 153, 185, 185, 196, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 60, 4, 168, 77, 136, 77, 232, 77, 72, 78, 168, 97, 136, 97, 232, 97, 72, 98, 200, 72, 136, 122, 168, 129, 136, 129, 232, 129, 136, 130, 72, 130, 128, 14, 200, 130, 168, 169, 136, 169, 232, 169, 72, 170, 136, 185, 168, 196, 72, 186, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 185, 68, 168, 68, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 249, 85, 232, 85, 217, 85, 200, 85, 185, 86, 168, 86, 249, 86, 232, 86, 249, 89, 232, 89, 89, 91, 72, 91, 154, 98, 136, 98, 58, 99, 40, 99, 218, 97, 200, 97, 26, 99, 8, 99, 186, 98, 168, 100, 6, 8, 8, 8, 250, 105, 232, 105, 249, 110, 232, 110, 168, 140, 153, 113, 136, 113, 249, 114, 232, 114, 25, 114, 8, 114, 121, 115, 104, 115, 89, 115, 72, 115, 154, 121, 136, 121, 250, 122, 232, 122, 26, 122, 8, 122, 10, 8, 185, 124, 168, 124, 57, 131, 40, 131, 217, 129, 200, 129, 121, 130, 104, 130, 12, 8, 14, 8, 153, 145, 136, 145, 62, 4, 64, 4, 185, 148, 168, 148, 154, 153, 136, 153, 250, 153, 232, 153, 66, 4, 68, 4, 185, 156, 168, 156, 249, 162, 232, 162, 25, 162, 8, 162, 185, 164, 168, 164, 153, 170, 136, 170, 57, 171, 40, 171, 217, 169, 200, 169, 57, 170, 40, 170, 121, 170, 104, 170, 25, 171, 8, 171, 249, 177, 232, 177, 249, 185, 232, 185, 89, 186, 153, 189, 136, 189, 185, 190, 168, 190, 185, 192, 168, 192, 232, 152, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 152, 169, 152, 200, 56, 169, 76, 187, 96, 187, 104, 169, 96, 169, 104, 169, 200, 169, 120, 217, 128, 187, 76, 200, 128, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 72, 4, 187, 92, 104, 4, 178, 92, 36, 5, 251, 93, 37, 5, 242, 93, 255, 5, 185, 144, 255, 5, 186, 152, 255, 5, 185, 188, 255, 5, 168, 64, 72, 4, 177, 92, 104, 4, 168, 92, 36, 5, 241, 93, 37, 5, 232, 93, 255, 5, 168, 144, 255, 5, 168, 152, 255, 5, 168, 188, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 249, 146, 255, 5, 232, 146, 255, 5, 250, 154, 255, 5, 232, 154, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 55, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 6, 9, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 11, 9, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26, 10, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 30, 10, 0, 0, 30, 10, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 85, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 120, 220, 38, 0, 67, 0, 60, 0, 99, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 72, 0, 38, 0, 68, 0, 60, 0, 100, 0, 100, 0, 60, 0, 60, 0, 60, 0, 68, 0, 100, 0, 60, 0, 60, 0, 60, 0, 68, 0, 68, 0, 38, 0, 70, 0, 60, 0, 102, 0, 102, 0, 60, 0, 60, 0, 60, 0, 70, 0, 102, 0, 60, 0, 60, 0, 60, 0, 70, 0, 70, 0, 38, 0, 71, 0, 60, 0, 110, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 71, 0, 38, 0, 76, 0, 60, 0, 108, 0, 108, 0, 60, 0, 60, 0, 60, 0, 76, 0, 108, 0, 60, 0, 60, 0, 60, 0, 76, 0, 76, 0, 38, 0, 80, 0, 60, 0, 112, 0, 104, 0, 60, 0, 60, 0, 60, 0, 80, 0, 104, 0, 60, 0, 60, 0, 60, 0, 80, 0, 72, 0, 38, 0, 82, 0, 60, 0, 114, 0, 104, 0, 60, 0, 60, 0, 60, 0, 82, 0, 104, 0, 60, 0, 60, 0, 60, 0, 82, 0, 72, 0, 38, 0, 84, 0, 60, 0, 116, 0, 104, 0, 60, 0, 60, 0, 60, 0, 84, 0, 104, 0, 60, 0, 60, 0, 60, 0, 84, 0, 72, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 234, 38, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 100, 23, 0, 0, 100, 23, 0, 0, 100, 23, 0, 0, 100, 23, 0, 0, 140, 28, 0, 0, 140, 28, 0, 0, 164, 33, 0, 0, 226, 33, 0, 0, 210, 38, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 105, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 40, 0, 202, 33, 40, 0, 202, 66, 40, 0, 202, 99, 40, 0, 202, 132, 40, 0, 202, 165, 40, 0, 202, 198, 40, 0, 202, 231, 40, 0, 202, 8, 41, 0, 202, 41, 41, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 0, 0, 0, 201, 96, 7, 0, 156, 5, 0, 50, 201, 192, 13, 0, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 160, 15, 0, 156, 5, 0, 66, 201, 0, 23, 0, 156, 5, 0, 70, 201, 32, 30, 0, 160, 5, 0, 74, 201, 224, 37, 0, 156, 5, 0, 78, 201, 160, 45, 0, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 96, 53, 0, 201, 96, 57, 0, 5, 5, 0, 50, 201, 160, 60, 0, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 192, 61, 0, 5, 5, 0, 66, 201, 128, 65, 0, 5, 5, 0, 70, 201, 32, 69, 0, 5, 5, 0, 74, 201, 64, 73, 0, 5, 5, 0, 78, 201, 96, 77, 0, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 66, 39, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 130, 39, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 194, 39, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 135, 5, 2, 47, 5, 142, 0, 0, 70, 5, 2, 47, 5, 142, 0, 0, 135, 5, 2, 47, 5, 144, 0, 0, 70, 5, 2, 47, 5, 144, 0, 0, 135, 5, 2, 47, 5, 156, 0, 0, 70, 5, 2, 47, 5, 156, 0, 0, 135, 5, 2, 47, 5, 196, 0, 0, 70, 5, 2, 47, 5, 196, 0, 0, 135, 5, 2, 47, 5, 150, 0, 0, 70, 5, 2, 47, 5, 150, 0, 0, 135, 5, 2, 47, 5, 160, 0, 0, 70, 5, 2, 47, 5, 160, 0, 0, 135, 5, 2, 47, 5, 206, 0, 0, 70, 5, 2, 47, 5, 206, 0, 0, 70, 5, 2, 47, 5, 210, 0, 0, 135, 5, 2, 49, 5, 144, 0, 0, 70, 5, 2, 49, 5, 144, 0, 0, 135, 5, 2, 49, 5, 156, 0, 0, 70, 5, 2, 49, 5, 156, 0, 0, 135, 5, 2, 49, 5, 196, 0, 0, 70, 5, 2, 49, 5, 196, 0, 0, 135, 5, 2, 49, 5, 210, 0, 0, 70, 5, 2, 49, 5, 210, 0, 0, 135, 5, 2, 49, 5, 160, 0, 0, 70, 5, 2, 49, 5, 160, 0, 0, 135, 5, 2, 49, 5, 204, 0, 0, 70, 5, 2, 49, 5, 204, 0, 0, 135, 5, 2, 53, 5, 156, 0, 0, 70, 5, 2, 53, 5, 156, 0, 0, 135, 5, 2, 65, 5, 136, 0, 0, 70, 5, 2, 65, 5, 136, 0, 0, 135, 5, 2, 65, 5, 160, 0, 0, 70, 5, 2, 65, 5, 160, 0, 0, 135, 5, 2, 65, 5, 144, 0, 0, 70, 5, 2, 65, 5, 144, 0, 0, 135, 5, 2, 65, 5, 196, 0, 0, 70, 5, 2, 65, 5, 196, 0, 0, 135, 5, 2, 65, 5, 196, 0, 0, 5, 164, 0, 0, 70, 5, 2, 65, 5, 196, 0, 0, 5, 164, 0, 0, 135, 5, 2, 65, 5, 210, 0, 0, 70, 5, 2, 65, 5, 210, 0, 0, 135, 5, 2, 65, 5, 204, 0, 0, 70, 5, 2, 65, 5, 204, 0, 0, 135, 5, 2, 55, 5, 142, 0, 0, 70, 5, 2, 55, 5, 142, 0, 0, 135, 5, 2, 55, 5, 140, 0, 0, 70, 5, 2, 55, 5, 140, 0, 0, 135, 5, 2, 55, 5, 156, 0, 0, 70, 5, 2, 55, 5, 156, 0, 0, 135, 5, 2, 55, 5, 160, 0, 0, 70, 5, 2, 55, 5, 160, 0, 0, 135, 5, 2, 55, 5, 144, 0, 0, 70, 5, 2, 55, 5, 144, 0, 0, 135, 5, 2, 55, 5, 136, 0, 0, 70, 5, 2, 55, 5, 136, 0, 0, 135, 5, 2, 55, 5, 164, 0, 0, 70, 5, 2, 55, 5, 164, 0, 0, 135, 5, 2, 73, 5, 142, 0, 0, 70, 5, 2, 73, 5, 142, 0, 0, 135, 5, 2, 73, 5, 144, 0, 0, 70, 5, 2, 73, 5, 144, 0, 0, 135, 5, 2, 73, 5, 156, 0, 0, 70, 5, 2, 73, 5, 156, 0, 0, 135, 5, 2, 73, 5, 196, 0, 0, 70, 5, 2, 73, 5, 196, 0, 0, 135, 5, 2, 73, 5, 150, 0, 0, 70, 5, 2, 73, 5, 150, 0, 0, 135, 5, 2, 73, 5, 160, 0, 0, 70, 5, 2, 73, 5, 160, 0, 0, 135, 5, 2, 73, 5, 206, 0, 0, 70, 5, 2, 73, 5, 206, 0, 0, 70, 5, 2, 73, 5, 210, 0, 0, 135, 5, 2, 77, 5, 142, 0, 0, 70, 5, 2, 77, 5, 142, 0, 0, 135, 5, 2, 77, 5, 144, 0, 0, 70, 5, 2, 77, 5, 144, 0, 0, 135, 5, 2, 77, 5, 156, 0, 0, 70, 5, 2, 77, 5, 156, 0, 0, 135, 5, 2, 77, 5, 196, 0, 0, 70, 5, 2, 77, 5, 196, 0, 0, 135, 5, 2, 77, 5, 150, 0, 0, 70, 5, 2, 77, 5, 150, 0, 0, 135, 5, 2, 77, 5, 160, 0, 0, 70, 5, 2, 77, 5, 160, 0, 0, 135, 5, 2, 77, 5, 206, 0, 0, 70, 5, 2, 77, 5, 206, 0, 0, 70, 5, 2, 77, 5, 210, 0, 0, 135, 5, 2, 81, 5, 142, 0, 0, 70, 5, 2, 81, 5, 142, 0, 0, 135, 5, 2, 81, 5, 144, 0, 0, 70, 5, 2, 81, 5, 144, 0, 0, 135, 5, 2, 81, 5, 156, 0, 0, 70, 5, 2, 81, 5, 156, 0, 0, 135, 5, 2, 81, 5, 196, 0, 0, 70, 5, 2, 81, 5, 196, 0, 0, 135, 5, 2, 81, 5, 150, 0, 0, 70, 5, 2, 81, 5, 150, 0, 0, 135, 5, 2, 81, 5, 160, 0, 0, 70, 5, 2, 81, 5, 160, 0, 0, 135, 5, 2, 81, 5, 206, 0, 0, 70, 5, 2, 81, 5, 206, 0, 0, 70, 5, 2, 81, 5, 210, 0, 0, 5, 5, 2, 47, 5, 142, 0, 0, 5, 5, 2, 47, 5, 144, 0, 0, 5, 5, 2, 47, 5, 156, 0, 0, 5, 5, 2, 47, 5, 196, 0, 0, 5, 5, 2, 47, 5, 150, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 206, 0, 0, 5, 5, 2, 47, 5, 210, 0, 0, 5, 5, 2, 49, 5, 144, 0, 0, 5, 5, 2, 49, 5, 156, 0, 0, 5, 5, 2, 49, 5, 196, 0, 0, 5, 5, 2, 49, 5, 210, 0, 0, 5, 5, 2, 49, 5, 160, 0, 0, 5, 5, 2, 49, 5, 204, 0, 0, 5, 5, 2, 53, 5, 156, 0, 0, 5, 5, 2, 65, 5, 136, 0, 0, 5, 5, 2, 65, 5, 160, 0, 0, 5, 5, 2, 65, 5, 144, 0, 0, 5, 5, 2, 65, 5, 196, 0, 0, 5, 5, 2, 65, 5, 196, 0, 0, 5, 164, 0, 0, 5, 5, 2, 65, 5, 210, 0, 0, 5, 5, 2, 65, 5, 204, 0, 0, 5, 5, 2, 55, 5, 142, 0, 0, 5, 5, 2, 55, 5, 140, 0, 0, 5, 5, 2, 55, 5, 156, 0, 0, 5, 5, 2, 55, 5, 160, 0, 0, 5, 5, 2, 55, 5, 144, 0, 0, 5, 5, 2, 55, 5, 136, 0, 0, 5, 5, 2, 55, 5, 164, 0, 0, 5, 5, 2, 73, 5, 142, 0, 0, 5, 5, 2, 73, 5, 144, 0, 0, 5, 5, 2, 73, 5, 156, 0, 0, 5, 5, 2, 73, 5, 196, 0, 0, 5, 5, 2, 73, 5, 150, 0, 0, 5, 5, 2, 73, 5, 160, 0, 0, 5, 5, 2, 73, 5, 206, 0, 0, 5, 5, 2, 73, 5, 210, 0, 0, 5, 5, 2, 77, 5, 142, 0, 0, 5, 5, 2, 77, 5, 144, 0, 0, 5, 5, 2, 77, 5, 156, 0, 0, 5, 5, 2, 77, 5, 196, 0, 0, 5, 5, 2, 77, 5, 150, 0, 0, 5, 5, 2, 77, 5, 160, 0, 0, 5, 5, 2, 77, 5, 206, 0, 0, 5, 5, 2, 77, 5, 210, 0, 0, 5, 5, 2, 81, 5, 142, 0, 0, 5, 5, 2, 81, 5, 144, 0, 0, 5, 5, 2, 81, 5, 156, 0, 0, 5, 5, 2, 81, 5, 196, 0, 0, 5, 5, 2, 81, 5, 150, 0, 0, 5, 5, 2, 81, 5, 160, 0, 0, 5, 5, 2, 81, 5, 206, 0, 0, 5, 5, 2, 81, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 2, 192, 197, 162, 41, 30, 2, 192, 197, 226, 42, 30, 3, 192, 197, 34, 43, 30, 3, 192, 197, 98, 150, 30, 3, 192, 197, 162, 36, 30, 1, 192, 197, 162, 37, 30, 1, 192, 197, 226, 38, 30, 2, 192, 197, 34, 39, 30, 2, 192, 197, 98, 30, 2, 12, 0, 30, 2, 0, 192, 197, 162, 31, 2, 0, 192, 197, 226, 34, 30, 1, 192, 197, 34, 35, 30, 1, 192, 197, 98, 72, 0, 2, 239, 135, 5, 104, 0, 2, 239, 70, 5, 36, 1, 197, 162, 37, 1, 0, 192, 197, 98, 0, 48, 156, 5, 13, 0, 13, 30, 23, 0, 16, 30, 12, 0, 16, 30, 5, 192, 197, 226, 17, 30, 6, 192, 197, 34, 18, 30, 6, 192, 197, 98, 19, 30, 6, 192, 197, 162, 13, 30, 5, 192, 197, 34, 14, 30, 5, 192, 197, 98, 15, 30, 5, 192, 197, 162, 15, 1, 12, 0, 15, 1, 4, 192, 197, 34, 10, 30, 4, 192, 197, 98, 11, 30, 4, 192, 197, 162, 12, 30, 4, 192, 197, 226, 68, 0, 2, 241, 135, 5, 100, 0, 2, 241, 70, 5, 14, 1, 3, 192, 197, 226, 0, 52, 156, 5, 3, 0, 70, 0, 2, 245, 135, 5, 102, 0, 2, 245, 70, 5, 30, 30, 6, 192, 197, 226, 31, 30, 7, 192, 197, 34, 0, 64, 156, 5, 15, 0, 54, 30, 26, 0, 58, 30, 12, 0, 58, 30, 10, 192, 197, 34, 59, 30, 10, 192, 197, 98, 60, 30, 10, 192, 197, 162, 61, 30, 10, 192, 197, 226, 54, 30, 8, 192, 197, 226, 55, 30, 9, 192, 197, 34, 56, 30, 9, 192, 197, 99, 57, 30, 9, 192, 197, 195, 59, 1, 12, 0, 59, 1, 7, 192, 197, 226, 60, 1, 8, 192, 197, 34, 61, 1, 8, 192, 197, 98, 62, 1, 8, 192, 197, 162, 76, 0, 255, 255, 2, 65, 135, 5, 108, 0, 255, 255, 2, 65, 70, 5, 57, 1, 7, 192, 197, 98, 58, 1, 7, 192, 197, 162, 0, 68, 156, 5, 15, 0, 34, 1, 26, 0, 244, 1, 12, 0, 244, 1, 13, 192, 197, 162, 245, 1, 13, 192, 197, 226, 32, 30, 14, 192, 197, 34, 33, 30, 14, 192, 197, 98, 34, 1, 12, 192, 197, 162, 35, 1, 12, 192, 197, 226, 230, 1, 13, 192, 197, 34, 231, 1, 13, 192, 197, 98, 30, 1, 12, 0, 30, 1, 11, 192, 197, 162, 31, 1, 11, 192, 197, 226, 32, 1, 12, 192, 197, 34, 33, 1, 12, 192, 197, 98, 71, 0, 2, 247, 135, 5, 103, 0, 2, 247, 70, 5, 28, 1, 11, 192, 197, 34, 29, 1, 11, 192, 197, 98, 0, 72, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 17, 192, 197, 34, 41, 30, 17, 192, 197, 98, 42, 30, 17, 192, 197, 162, 43, 30, 17, 192, 197, 226, 150, 30, 18, 192, 197, 34, 36, 30, 16, 192, 197, 34, 37, 30, 16, 192, 197, 98, 38, 30, 16, 192, 197, 162, 39, 30, 16, 192, 197, 226, 30, 2, 12, 0, 30, 2, 15, 192, 197, 34, 31, 2, 15, 192, 197, 98, 34, 30, 15, 192, 197, 162, 35, 30, 15, 192, 197, 226, 72, 0, 255, 255, 2, 73, 135, 5, 104, 0, 255, 255, 2, 73, 70, 5, 36, 1, 14, 192, 197, 162, 37, 1, 14, 192, 197, 226, 0, 76, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 20, 192, 197, 226, 41, 30, 21, 192, 197, 34, 42, 30, 21, 192, 197, 98, 43, 30, 21, 192, 197, 162, 150, 30, 21, 192, 197, 226, 36, 30, 19, 192, 197, 226, 37, 30, 20, 192, 197, 34, 38, 30, 20, 192, 197, 98, 39, 30, 20, 192, 197, 162, 30, 2, 12, 0, 30, 2, 18, 192, 197, 226, 31, 2, 19, 192, 197, 34, 34, 30, 19, 192, 197, 98, 35, 30, 19, 192, 197, 162, 72, 0, 255, 255, 2, 77, 135, 5, 104, 0, 255, 255, 2, 77, 70, 5, 36, 1, 18, 192, 197, 98, 37, 1, 18, 192, 197, 162, 0, 80, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 24, 192, 197, 162, 41, 30, 24, 192, 197, 226, 42, 30, 25, 192, 197, 34, 43, 30, 25, 192, 197, 98, 150, 30, 25, 192, 197, 162, 36, 30, 23, 192, 197, 162, 37, 30, 23, 192, 197, 226, 38, 30, 24, 192, 197, 34, 39, 30, 24, 192, 197, 98, 30, 2, 12, 0, 30, 2, 22, 192, 197, 162, 31, 2, 22, 192, 197, 226, 34, 30, 23, 192, 197, 34, 35, 30, 23, 192, 197, 98, 72, 0, 255, 255, 2, 81, 135, 5, 104, 0, 255, 255, 2, 81, 70, 5, 36, 1, 22, 192, 197, 34, 37, 1, 22, 192, 197, 98, 0, 46, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 26, 192, 197, 162, 39, 30, 26, 192, 197, 226, 41, 30, 27, 192, 197, 34, 43, 30, 27, 192, 197, 98, 150, 30, 27, 192, 197, 162, 104, 0, 2, 239, 5, 5, 37, 1, 25, 192, 197, 226, 31, 2, 26, 192, 197, 34, 35, 30, 26, 192, 197, 98, 0, 48, 5, 5, 6, 0, 13, 30, 12, 0, 13, 30, 28, 192, 197, 98, 15, 30, 28, 192, 197, 162, 17, 30, 28, 192, 197, 226, 19, 30, 29, 192, 197, 34, 100, 0, 2, 241, 5, 5, 15, 1, 27, 192, 197, 226, 11, 30, 28, 192, 197, 34, 0, 52, 5, 5, 1, 0, 102, 0, 2, 245, 5, 5, 31, 30, 29, 192, 197, 98, 0, 64, 5, 5, 7, 0, 55, 30, 12, 0, 55, 30, 30, 192, 197, 98, 57, 30, 30, 192, 197, 163, 59, 30, 31, 192, 197, 2, 61, 30, 31, 192, 197, 66, 108, 0, 255, 255, 2, 65, 5, 5, 58, 1, 29, 192, 197, 162, 60, 1, 29, 192, 197, 226, 62, 1, 30, 192, 197, 34, 0, 68, 5, 5, 7, 0, 35, 1, 12, 0, 35, 1, 32, 192, 197, 66, 231, 1, 32, 192, 197, 130, 245, 1, 32, 192, 197, 194, 33, 30, 33, 192, 197, 2, 103, 0, 2, 247, 5, 5, 29, 1, 31, 192, 197, 130, 31, 1, 31, 192, 197, 194, 33, 1, 32, 192, 197, 2, 0, 72, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 34, 192, 197, 2, 39, 30, 34, 192, 197, 66, 41, 30, 34, 192, 197, 130, 43, 30, 34, 192, 197, 194, 150, 30, 35, 192, 197, 2, 104, 0, 255, 255, 2, 73, 5, 5, 37, 1, 33, 192, 197, 66, 31, 2, 33, 192, 197, 130, 35, 30, 33, 192, 197, 194, 0, 76, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 36, 192, 197, 2, 39, 30, 36, 192, 197, 66, 41, 30, 36, 192, 197, 130, 43, 30, 36, 192, 197, 194, 150, 30, 37, 192, 197, 2, 104, 0, 255, 255, 2, 77, 5, 5, 37, 1, 35, 192, 197, 66, 31, 2, 35, 192, 197, 130, 35, 30, 35, 192, 197, 194, 0, 80, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 38, 192, 197, 2, 39, 30, 38, 192, 197, 66, 41, 30, 38, 192, 197, 130, 43, 30, 38, 192, 197, 194, 150, 30, 39, 192, 197, 2, 104, 0, 255, 255, 2, 81, 5, 5, 37, 1, 37, 192, 197, 66, 31, 2, 37, 192, 197, 130, 35, 30, 37, 192, 197, 194, 30, 0, 68, 0, 69, 0, 70, 0, 73, 0, 100, 0, 101, 0, 102, 0, 105, 0, 108, 0, 109, 0, 14, 1, 16, 1, 28, 1, 38, 1, 57, 1, 63, 1, 230, 1, 232, 1, 244, 1, 246, 1, 30, 2, 32, 2, 10, 30, 20, 30, 30, 30, 44, 30, 54, 30, 62, 30, 150, 30, 151, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 36, 4, 186, 80, 46, 4, 185, 92, 185, 100, 186, 104, 186, 112, 185, 116, 52, 4, 185, 128, 70, 4, 185, 140, 92, 4, 185, 152, 102, 4, 186, 168, 112, 4, 185, 184, 185, 188, 185, 192, 185, 196, 185, 200, 185, 204, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 122, 4, 128, 4, 168, 80, 134, 4, 168, 92, 168, 100, 168, 104, 168, 112, 168, 116, 138, 4, 168, 128, 148, 4, 168, 140, 160, 4, 168, 152, 166, 4, 168, 168, 172, 4, 168, 184, 168, 188, 168, 192, 168, 196, 168, 200, 168, 204, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 140, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 81, 154, 81, 250, 81, 90, 82, 186, 105, 154, 105, 250, 105, 90, 106, 217, 72, 154, 134, 185, 141, 153, 141, 249, 141, 153, 142, 89, 142, 136, 14, 217, 142, 185, 185, 153, 185, 249, 185, 89, 186, 153, 201, 185, 208, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 81, 136, 81, 232, 81, 72, 82, 168, 105, 136, 105, 232, 105, 72, 106, 200, 72, 136, 134, 168, 141, 136, 141, 232, 141, 136, 142, 72, 142, 128, 14, 200, 142, 168, 185, 136, 185, 232, 185, 72, 186, 136, 201, 168, 208, 72, 202, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 74, 8, 74, 89, 75, 72, 75, 58, 83, 40, 83, 218, 81, 200, 81, 186, 82, 168, 82, 26, 83, 8, 83, 26, 82, 8, 82, 249, 93, 232, 93, 217, 93, 200, 93, 185, 94, 168, 94, 249, 94, 232, 94, 249, 101, 232, 101, 89, 103, 72, 103, 154, 106, 136, 106, 58, 107, 40, 107, 218, 105, 200, 105, 26, 107, 8, 107, 186, 106, 168, 108, 6, 8, 8, 8, 250, 113, 232, 113, 249, 118, 232, 118, 168, 156, 153, 121, 136, 121, 249, 122, 232, 122, 25, 122, 8, 122, 121, 123, 104, 123, 89, 123, 72, 123, 154, 133, 136, 133, 250, 134, 232, 134, 26, 134, 8, 134, 10, 8, 185, 136, 168, 136, 57, 143, 40, 143, 217, 141, 200, 141, 121, 142, 104, 142, 12, 8, 14, 8, 153, 161, 136, 161, 249, 162, 232, 162, 25, 162, 8, 162, 154, 169, 136, 169, 250, 169, 232, 169, 250, 170, 232, 170, 26, 170, 8, 170, 249, 174, 232, 174, 25, 174, 8, 174, 185, 180, 168, 180, 153, 186, 136, 186, 57, 187, 40, 187, 217, 185, 200, 185, 57, 186, 40, 186, 121, 186, 104, 186, 25, 187, 8, 187, 249, 193, 232, 193, 249, 201, 232, 201, 89, 202, 153, 205, 136, 205, 185, 206, 168, 206, 25, 206, 8, 206, 232, 168, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 80, 200, 168, 169, 168, 200, 56, 169, 80, 187, 104, 187, 112, 169, 104, 169, 112, 169, 212, 169, 132, 217, 140, 187, 80, 200, 140, 169, 80, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 72, 4, 186, 68, 104, 4, 177, 68, 36, 5, 250, 69, 37, 5, 241, 69, 255, 5, 185, 72, 68, 4, 186, 76, 100, 4, 177, 76, 14, 5, 26, 78, 15, 5, 17, 78, 255, 5, 185, 84, 70, 4, 186, 88, 102, 4, 177, 88, 255, 5, 185, 120, 76, 4, 186, 124, 108, 4, 177, 124, 57, 5, 154, 125, 58, 5, 145, 125, 59, 5, 250, 126, 60, 5, 241, 126, 61, 5, 26, 126, 62, 5, 17, 126, 255, 5, 186, 132, 71, 4, 186, 96, 103, 4, 177, 96, 28, 5, 250, 97, 29, 5, 241, 97, 30, 5, 218, 97, 31, 5, 209, 97, 32, 5, 186, 98, 33, 5, 177, 98, 34, 5, 250, 98, 35, 5, 241, 98, 255, 5, 185, 144, 72, 4, 186, 148, 104, 4, 177, 148, 36, 5, 250, 149, 37, 5, 241, 149, 255, 5, 185, 160, 72, 4, 186, 164, 104, 4, 177, 164, 36, 5, 250, 165, 37, 5, 241, 165, 255, 5, 185, 172, 72, 4, 186, 176, 104, 4, 177, 176, 36, 5, 250, 177, 37, 5, 241, 177, 255, 5, 168, 64, 104, 4, 168, 68, 37, 5, 232, 69, 255, 5, 168, 72, 100, 4, 168, 76, 15, 5, 8, 78, 255, 5, 168, 84, 102, 4, 168, 88, 255, 5, 168, 120, 108, 4, 168, 124, 58, 5, 136, 125, 60, 5, 232, 126, 62, 5, 8, 126, 255, 5, 168, 132, 103, 4, 168, 96, 29, 5, 232, 97, 31, 5, 200, 97, 33, 5, 168, 98, 35, 5, 232, 98, 255, 5, 168, 144, 104, 4, 168, 148, 37, 5, 232, 149, 255, 5, 168, 160, 104, 4, 168, 164, 37, 5, 232, 165, 255, 5, 168, 172, 104, 4, 168, 176, 37, 5, 232, 177, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 87, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 19, 10, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 24, 10, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 138, 27, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 142, 27, 0, 0, 142, 27, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 132, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 64, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 100, 0, 97, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 91, 0, 99, 0, 97, 0, 115, 0, 101, 0, 70, 0, 105, 0, 114, 0, 115, 0, 116, 0, 32, 0, 111, 0, 102, 0, 102, 0, 93, 0, 0, 0, 143, 220, 91, 0, 99, 0, 97, 0, 115, 0, 101, 0, 70, 0, 105, 0, 114, 0, 115, 0, 116, 0, 32, 0, 117, 0, 112, 0, 112, 0, 101, 0, 114, 0, 93, 0, 38, 0, 68, 0, 60, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 60, 0, 60, 0, 240, 0, 60, 0, 60, 0, 60, 0, 208, 0, 38, 0, 116, 0, 104, 0, 60, 0, 60, 0, 60, 0, 254, 0, 38, 0, 84, 0, 72, 0, 60, 0, 60, 0, 60, 0, 222, 0, 38, 0, 89, 0, 60, 0, 60, 0, 117, 0, 8, 3, 60, 0, 60, 0, 60, 0, 85, 0, 8, 3, 60, 0, 60, 0, 117, 0, 11, 3, 60, 0, 60, 0, 60, 0, 85, 0, 11, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 192, 1, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 60, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 81, 1, 60, 0, 60, 0, 60, 0, 80, 1, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 60, 0, 60, 0, 60, 0, 97, 0, 97, 0, 60, 0, 60, 0, 60, 0, 65, 0, 97, 0, 60, 0, 60, 0, 60, 0, 65, 0, 65, 0, 38, 0, 111, 0, 101, 0, 60, 0, 60, 0, 83, 1, 60, 0, 60, 0, 60, 0, 82, 1, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 220, 56, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 167, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 100, 38, 0, 0, 100, 38, 0, 0, 108, 47, 0, 0, 108, 47, 0, 0, 60, 51, 0, 0, 60, 51, 0, 0, 138, 56, 0, 0, 192, 56, 0, 0, 192, 56, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 89, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 145, 0, 153, 0, 159, 0, 167, 0, 175, 0, 183, 0, 48, 0, 56, 0, 182, 0, 190, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 207, 0, 215, 0, 48, 0, 48, 0, 48, 0, 223, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 231, 0, 48, 0, 48, 0, 239, 0, 48, 0, 245, 0, 253, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 8, 1, 48, 0, 48, 0, 12, 1, 48, 0, 20, 1, 25, 1, 33, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 39, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 47, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 50, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 205, 0, 48, 0, 48, 0, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 58, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 66, 1, 48, 0, 48, 0, 72, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 68, 2, 124, 2, 188, 2, 192, 0, 216, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 64, 29, 0, 202, 97, 29, 0, 202, 130, 29, 0, 202, 163, 29, 0, 202, 196, 29, 0, 202, 229, 29, 0, 202, 6, 30, 0, 202, 39, 30, 0, 202, 72, 30, 0, 202, 105, 30, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 196, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 102, 29, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 134, 30, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 36, 32, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 230, 46, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 6, 48, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 129, 8, 0, 198, 161, 8, 0, 198, 33, 27, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 193, 17, 0, 5, 5, 127, 12, 198, 65, 27, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 162, 20, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 198, 65, 22, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 198, 225, 26, 0, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 49, 0, 201, 230, 50, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 225, 17, 0, 198, 1, 27, 0, 197, 98, 1, 0, 197, 162, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 6, 52, 0, 201, 38, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 54, 0, 201, 102, 55, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 56, 0, 201, 166, 57, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 226, 4, 0, 197, 34, 5, 0, 192, 0, 0, 0, 198, 34, 29, 0, 198, 98, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 29, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 58, 0, 201, 230, 59, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 30, 0, 197, 162, 5, 0, 198, 34, 31, 0, 197, 226, 5, 0, 198, 98, 31, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 31, 0, 198, 226, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 6, 61, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 62, 0, 5, 71, 238, 102, 201, 198, 63, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 192, 0, 0, 0, 197, 34, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 197, 34, 8, 0, 192, 0, 0, 0, 197, 98, 8, 0, 192, 0, 0, 0, 197, 162, 8, 0, 197, 227, 8, 0, 197, 66, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 9, 0, 197, 195, 9, 0, 197, 34, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 98, 6, 0, 197, 162, 10, 0, 5, 5, 10, 121, 197, 226, 10, 0, 197, 34, 11, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 2, 7, 0, 5, 5, 24, 121, 197, 34, 7, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 64, 0, 201, 198, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 66, 0, 201, 6, 68, 0, 201, 38, 69, 0, 201, 70, 70, 0, 201, 102, 71, 0, 201, 134, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 73, 0, 201, 198, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 75, 0, 201, 6, 77, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 78, 0, 201, 70, 79, 0, 201, 102, 80, 0, 201, 134, 81, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 82, 0, 201, 198, 83, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 72, 0, 5, 0, 4, 111, 93, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 156, 5, 0, 80, 29, 5, 0, 56, 5, 5, 0, 80, 6, 5, 0, 56, 193, 4, 111, 93, 5, 162, 0, 0, 133, 5, 0, 70, 134, 70, 0, 50, 5, 5, 0, 70, 5, 70, 0, 50, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 193, 4, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 6, 121, 197, 98, 6, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 162, 6, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 226, 6, 0, 5, 5, 24, 121, 197, 34, 7, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 98, 7, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 162, 7, 0, 5, 5, 107, 121, 197, 226, 7, 0, 5, 5, 111, 121, 197, 34, 8, 0, 5, 5, 115, 121, 197, 98, 8, 0, 5, 5, 119, 121, 197, 162, 8, 0, 197, 227, 8, 0, 197, 66, 9, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 130, 9, 0, 197, 195, 9, 0, 197, 34, 10, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 98, 10, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 98, 6, 0, 197, 162, 10, 0, 5, 5, 10, 121, 197, 226, 10, 0, 197, 34, 11, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 2, 7, 0, 5, 5, 24, 121, 197, 34, 7, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 0, 0, 62, 0, 3, 2, 118, 0, 169, 30, 54, 0, 177, 30, 26, 0, 181, 30, 12, 0, 181, 30, 16, 192, 198, 163, 182, 30, 17, 192, 198, 3, 183, 30, 17, 192, 198, 99, 43, 33, 1, 192, 198, 130, 177, 30, 15, 192, 198, 35, 178, 30, 15, 192, 198, 131, 179, 30, 15, 192, 198, 227, 180, 30, 16, 192, 198, 67, 173, 30, 12, 0, 173, 30, 13, 192, 198, 163, 174, 30, 14, 192, 198, 3, 175, 30, 14, 192, 198, 99, 176, 30, 14, 192, 198, 195, 169, 30, 12, 192, 198, 35, 170, 30, 12, 192, 198, 131, 171, 30, 12, 192, 198, 227, 172, 30, 13, 192, 198, 67, 161, 30, 26, 0, 165, 30, 12, 0, 165, 30, 10, 192, 198, 163, 166, 30, 11, 192, 198, 3, 167, 30, 11, 192, 198, 99, 168, 30, 11, 192, 198, 195, 161, 30, 9, 192, 198, 130, 162, 30, 9, 192, 198, 194, 163, 30, 10, 192, 198, 2, 164, 30, 10, 192, 198, 67, 10, 3, 12, 0, 10, 3, 8, 192, 198, 161, 0, 30, 8, 192, 198, 194, 1, 30, 9, 192, 198, 2, 160, 30, 9, 192, 198, 66, 3, 2, 8, 192, 198, 66, 38, 2, 5, 0, 39, 2, 8, 0, 8, 3, 8, 192, 198, 129, 240, 65, 198, 2, 4, 3, 6, 192, 198, 3, 240, 65, 198, 98, 4, 3, 6, 192, 198, 99, 1, 1, 54, 0, 223, 1, 26, 0, 251, 1, 12, 0, 251, 1, 7, 192, 198, 35, 0, 2, 7, 192, 198, 130, 1, 2, 7, 192, 198, 194, 2, 2, 8, 192, 198, 2, 223, 1, 5, 192, 198, 163, 224, 1, 6, 192, 198, 3, 225, 1, 6, 192, 198, 99, 250, 1, 6, 192, 198, 195, 5, 1, 12, 0, 5, 1, 4, 192, 198, 130, 205, 1, 4, 192, 198, 194, 206, 1, 5, 192, 198, 2, 222, 1, 5, 192, 198, 67, 1, 1, 3, 192, 198, 130, 2, 1, 3, 192, 198, 194, 3, 1, 4, 192, 198, 2, 4, 1, 4, 192, 198, 66, 197, 0, 26, 0, 227, 0, 12, 0, 227, 0, 2, 192, 198, 130, 228, 0, 2, 192, 198, 194, 229, 0, 3, 192, 198, 2, 0, 1, 3, 192, 198, 66, 197, 0, 1, 192, 198, 130, 224, 0, 1, 192, 198, 194, 225, 0, 2, 192, 198, 2, 226, 0, 2, 192, 198, 66, 193, 0, 12, 0, 193, 0, 0, 192, 198, 130, 194, 0, 0, 192, 198, 194, 195, 0, 1, 192, 198, 2, 196, 0, 1, 192, 198, 66, 65, 0, 198, 129, 97, 0, 198, 161, 192, 0, 0, 192, 198, 66, 0, 70, 156, 5, 1, 0, 8, 3, 17, 192, 198, 193, 11, 3, 17, 192, 198, 225, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 31, 0, 39, 2, 59, 0, 169, 30, 26, 0, 177, 30, 12, 0, 177, 30, 25, 192, 198, 99, 179, 30, 25, 192, 198, 195, 181, 30, 26, 192, 198, 35, 183, 30, 26, 192, 198, 131, 169, 30, 23, 192, 198, 227, 171, 30, 24, 192, 198, 67, 173, 30, 24, 192, 198, 163, 175, 30, 25, 192, 198, 3, 161, 30, 12, 0, 161, 30, 22, 192, 198, 162, 163, 30, 22, 192, 198, 226, 165, 30, 23, 192, 198, 35, 167, 30, 23, 192, 198, 131, 39, 2, 10, 0, 8, 3, 22, 192, 198, 65, 10, 3, 255, 255, 111, 93, 193, 4, 1, 30, 22, 192, 198, 98, 176, 69, 198, 2, 4, 3, 21, 192, 198, 3, 3, 1, 26, 0, 225, 1, 12, 0, 225, 1, 21, 192, 198, 3, 251, 1, 21, 192, 198, 99, 1, 2, 21, 192, 198, 194, 3, 2, 22, 192, 198, 2, 3, 1, 19, 192, 198, 226, 5, 1, 20, 192, 198, 34, 206, 1, 20, 192, 198, 98, 223, 1, 20, 192, 198, 163, 227, 0, 12, 0, 227, 0, 18, 192, 198, 226, 228, 0, 19, 192, 198, 34, 229, 0, 19, 192, 198, 98, 1, 1, 19, 192, 198, 162, 97, 0, 18, 192, 198, 1, 224, 0, 18, 192, 198, 34, 225, 0, 18, 192, 198, 98, 226, 0, 18, 192, 198, 162, 0, 70, 5, 5, 1, 0, 8, 3, 26, 192, 198, 225, 11, 3, 27, 192, 198, 1, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 27, 192, 198, 98, 10, 3, 27, 192, 198, 162, 5, 42, 196, 162, 1, 0, 8, 3, 27, 192, 198, 226, 10, 3, 1, 192, 197, 34, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 226, 11, 3, 2, 192, 197, 34, 5, 82, 196, 162, 1, 0, 8, 3, 2, 192, 197, 98, 11, 3, 2, 192, 197, 162, 156, 70, 196, 190, 1, 0, 8, 3, 28, 192, 198, 34, 11, 3, 28, 192, 198, 98, 5, 70, 196, 190, 1, 0, 8, 3, 28, 192, 198, 162, 11, 3, 28, 192, 198, 226, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 5, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 98, 11, 3, 3, 192, 197, 162, 156, 70, 196, 162, 1, 0, 8, 3, 29, 192, 198, 226, 11, 3, 30, 192, 198, 34, 5, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 98, 11, 3, 30, 192, 198, 162, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 1, 0, 8, 3, 32, 192, 198, 34, 10, 3, 32, 192, 198, 98, 5, 42, 196, 200, 1, 0, 8, 3, 32, 192, 198, 162, 10, 3, 13, 192, 197, 34, 156, 82, 196, 198, 1, 0, 8, 3, 13, 192, 197, 98, 11, 3, 13, 192, 197, 162, 5, 82, 196, 198, 1, 0, 8, 3, 13, 192, 197, 226, 11, 3, 14, 192, 197, 34, 156, 82, 196, 208, 1, 0, 8, 3, 14, 192, 197, 98, 11, 3, 14, 192, 197, 162, 5, 82, 196, 208, 1, 0, 8, 3, 14, 192, 197, 226, 11, 3, 15, 192, 197, 34, 156, 82, 196, 204, 1, 0, 8, 3, 15, 192, 197, 98, 11, 3, 15, 192, 197, 162, 5, 82, 196, 204, 1, 0, 8, 3, 15, 192, 197, 226, 11, 3, 16, 192, 197, 34, 156, 42, 196, 196, 1, 0, 8, 3, 32, 192, 198, 226, 10, 3, 33, 192, 198, 34, 5, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 98, 10, 3, 16, 192, 197, 98, 156, 70, 196, 196, 1, 0, 8, 3, 33, 192, 198, 162, 11, 3, 33, 192, 198, 226, 5, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 34, 11, 3, 34, 192, 198, 98, 16, 0, 197, 163, 1, 0, 8, 3, 34, 192, 198, 163, 11, 3, 35, 192, 198, 3, 17, 0, 197, 3, 1, 0, 8, 3, 35, 192, 198, 99, 11, 3, 35, 192, 198, 195, 156, 82, 196, 196, 1, 0, 8, 3, 17, 192, 197, 98, 11, 3, 17, 192, 197, 162, 5, 82, 196, 196, 1, 0, 8, 3, 17, 192, 197, 226, 11, 3, 18, 192, 197, 34, 18, 0, 197, 99, 1, 0, 8, 3, 18, 192, 197, 195, 11, 3, 19, 192, 197, 35, 19, 0, 197, 131, 1, 0, 8, 3, 19, 192, 197, 227, 11, 3, 20, 192, 197, 67, 26, 0, 65, 0, 66, 0, 97, 0, 98, 0, 192, 0, 198, 0, 224, 0, 230, 0, 0, 1, 6, 1, 205, 1, 207, 1, 222, 1, 226, 1, 250, 1, 252, 1, 0, 2, 4, 2, 38, 2, 40, 2, 0, 30, 2, 30, 160, 30, 184, 30, 43, 33, 44, 33, 3, 0, 67, 0, 58, 0, 32, 0, 135, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 226, 50, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 35, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 52, 30, 0, 0, 52, 30, 0, 0, 60, 39, 0, 0, 60, 39, 0, 0, 40, 41, 0, 0, 40, 41, 0, 0, 48, 46, 0, 0, 102, 46, 0, 0, 198, 50, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 214, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 48, 0, 48, 0, 170, 0, 48, 0, 178, 0, 183, 0, 191, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 197, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 158, 0, 48, 0, 48, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 14, 0, 202, 65, 14, 0, 202, 98, 14, 0, 202, 131, 14, 0, 202, 164, 14, 0, 202, 197, 14, 0, 202, 230, 14, 0, 202, 7, 15, 0, 202, 40, 15, 0, 202, 73, 15, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 4, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 166, 28, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 198, 29, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 100, 31, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 38, 46, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 70, 47, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 129, 8, 0, 198, 161, 8, 0, 198, 33, 27, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 193, 17, 0, 5, 5, 127, 12, 198, 65, 27, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 226, 13, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 198, 65, 22, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 198, 225, 26, 0, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 49, 0, 201, 38, 50, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 225, 17, 0, 198, 1, 27, 0, 197, 98, 1, 0, 197, 162, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 70, 51, 0, 201, 102, 52, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 53, 0, 201, 166, 54, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 55, 0, 201, 230, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 226, 4, 0, 197, 34, 5, 0, 192, 0, 0, 0, 198, 34, 29, 0, 198, 98, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 29, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 58, 0, 201, 38, 59, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 30, 0, 197, 162, 5, 0, 198, 34, 31, 0, 197, 226, 5, 0, 198, 98, 31, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 31, 0, 198, 226, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 60, 0, 201, 102, 61, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 62, 0, 201, 166, 63, 0, 201, 198, 64, 0, 201, 230, 65, 0, 201, 6, 67, 0, 201, 38, 68, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 69, 0, 201, 102, 70, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 71, 0, 201, 166, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 73, 0, 201, 230, 74, 0, 201, 6, 76, 0, 201, 38, 77, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 78, 0, 201, 102, 79, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 72, 0, 5, 0, 4, 111, 93, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 156, 5, 0, 80, 29, 5, 0, 56, 5, 5, 0, 80, 6, 5, 0, 56, 193, 4, 111, 93, 5, 162, 0, 0, 133, 5, 0, 70, 134, 70, 0, 50, 5, 5, 0, 70, 5, 70, 0, 50, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 0, 0, 62, 0, 3, 2, 118, 0, 169, 30, 54, 0, 177, 30, 26, 0, 181, 30, 12, 0, 181, 30, 16, 192, 198, 163, 182, 30, 17, 192, 198, 3, 183, 30, 17, 192, 198, 99, 43, 33, 1, 192, 198, 130, 177, 30, 15, 192, 198, 35, 178, 30, 15, 192, 198, 131, 179, 30, 15, 192, 198, 227, 180, 30, 16, 192, 198, 67, 173, 30, 12, 0, 173, 30, 13, 192, 198, 163, 174, 30, 14, 192, 198, 3, 175, 30, 14, 192, 198, 99, 176, 30, 14, 192, 198, 195, 169, 30, 12, 192, 198, 35, 170, 30, 12, 192, 198, 131, 171, 30, 12, 192, 198, 227, 172, 30, 13, 192, 198, 67, 161, 30, 26, 0, 165, 30, 12, 0, 165, 30, 10, 192, 198, 163, 166, 30, 11, 192, 198, 3, 167, 30, 11, 192, 198, 99, 168, 30, 11, 192, 198, 195, 161, 30, 9, 192, 198, 130, 162, 30, 9, 192, 198, 194, 163, 30, 10, 192, 198, 2, 164, 30, 10, 192, 198, 67, 10, 3, 12, 0, 10, 3, 8, 192, 198, 161, 0, 30, 8, 192, 198, 194, 1, 30, 9, 192, 198, 2, 160, 30, 9, 192, 198, 66, 3, 2, 8, 192, 198, 66, 38, 2, 5, 0, 39, 2, 8, 0, 8, 3, 8, 192, 198, 129, 240, 65, 198, 2, 4, 3, 6, 192, 198, 3, 240, 65, 198, 98, 4, 3, 6, 192, 198, 99, 1, 1, 54, 0, 223, 1, 26, 0, 251, 1, 12, 0, 251, 1, 7, 192, 198, 35, 0, 2, 7, 192, 198, 130, 1, 2, 7, 192, 198, 194, 2, 2, 8, 192, 198, 2, 223, 1, 5, 192, 198, 163, 224, 1, 6, 192, 198, 3, 225, 1, 6, 192, 198, 99, 250, 1, 6, 192, 198, 195, 5, 1, 12, 0, 5, 1, 4, 192, 198, 130, 205, 1, 4, 192, 198, 194, 206, 1, 5, 192, 198, 2, 222, 1, 5, 192, 198, 67, 1, 1, 3, 192, 198, 130, 2, 1, 3, 192, 198, 194, 3, 1, 4, 192, 198, 2, 4, 1, 4, 192, 198, 66, 197, 0, 26, 0, 227, 0, 12, 0, 227, 0, 2, 192, 198, 130, 228, 0, 2, 192, 198, 194, 229, 0, 3, 192, 198, 2, 0, 1, 3, 192, 198, 66, 197, 0, 1, 192, 198, 130, 224, 0, 1, 192, 198, 194, 225, 0, 2, 192, 198, 2, 226, 0, 2, 192, 198, 66, 193, 0, 12, 0, 193, 0, 0, 192, 198, 130, 194, 0, 0, 192, 198, 194, 195, 0, 1, 192, 198, 2, 196, 0, 1, 192, 198, 66, 65, 0, 198, 129, 97, 0, 198, 161, 192, 0, 0, 192, 198, 66, 0, 70, 156, 5, 1, 0, 8, 3, 17, 192, 198, 193, 11, 3, 17, 192, 198, 225, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 31, 0, 39, 2, 59, 0, 169, 30, 26, 0, 177, 30, 12, 0, 177, 30, 25, 192, 198, 99, 179, 30, 25, 192, 198, 195, 181, 30, 26, 192, 198, 35, 183, 30, 26, 192, 198, 131, 169, 30, 23, 192, 198, 227, 171, 30, 24, 192, 198, 67, 173, 30, 24, 192, 198, 163, 175, 30, 25, 192, 198, 3, 161, 30, 12, 0, 161, 30, 22, 192, 198, 162, 163, 30, 22, 192, 198, 226, 165, 30, 23, 192, 198, 35, 167, 30, 23, 192, 198, 131, 39, 2, 10, 0, 8, 3, 22, 192, 198, 65, 10, 3, 255, 255, 111, 93, 193, 4, 1, 30, 22, 192, 198, 98, 176, 69, 198, 2, 4, 3, 21, 192, 198, 3, 3, 1, 26, 0, 225, 1, 12, 0, 225, 1, 21, 192, 198, 3, 251, 1, 21, 192, 198, 99, 1, 2, 21, 192, 198, 194, 3, 2, 22, 192, 198, 2, 3, 1, 19, 192, 198, 226, 5, 1, 20, 192, 198, 34, 206, 1, 20, 192, 198, 98, 223, 1, 20, 192, 198, 163, 227, 0, 12, 0, 227, 0, 18, 192, 198, 226, 228, 0, 19, 192, 198, 34, 229, 0, 19, 192, 198, 98, 1, 1, 19, 192, 198, 162, 97, 0, 18, 192, 198, 1, 224, 0, 18, 192, 198, 34, 225, 0, 18, 192, 198, 98, 226, 0, 18, 192, 198, 162, 0, 70, 5, 5, 1, 0, 8, 3, 26, 192, 198, 225, 11, 3, 27, 192, 198, 1, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 27, 192, 198, 98, 10, 3, 27, 192, 198, 162, 5, 42, 196, 162, 1, 0, 8, 3, 27, 192, 198, 226, 10, 3, 1, 192, 197, 34, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 226, 11, 3, 2, 192, 197, 34, 5, 82, 196, 162, 1, 0, 8, 3, 2, 192, 197, 98, 11, 3, 2, 192, 197, 162, 156, 70, 196, 190, 1, 0, 8, 3, 28, 192, 198, 34, 11, 3, 28, 192, 198, 98, 5, 70, 196, 190, 1, 0, 8, 3, 28, 192, 198, 162, 11, 3, 28, 192, 198, 226, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 5, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 98, 11, 3, 3, 192, 197, 162, 156, 70, 196, 162, 1, 0, 8, 3, 29, 192, 198, 226, 11, 3, 30, 192, 198, 34, 5, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 98, 11, 3, 30, 192, 198, 162, 156, 42, 196, 200, 1, 0, 8, 3, 32, 192, 198, 34, 10, 3, 32, 192, 198, 98, 5, 42, 196, 200, 1, 0, 8, 3, 32, 192, 198, 162, 10, 3, 6, 192, 197, 98, 156, 82, 196, 198, 1, 0, 8, 3, 6, 192, 197, 162, 11, 3, 6, 192, 197, 226, 5, 82, 196, 198, 1, 0, 8, 3, 7, 192, 197, 34, 11, 3, 7, 192, 197, 98, 156, 82, 196, 208, 1, 0, 8, 3, 7, 192, 197, 162, 11, 3, 7, 192, 197, 226, 5, 82, 196, 208, 1, 0, 8, 3, 8, 192, 197, 34, 11, 3, 8, 192, 197, 98, 156, 82, 196, 204, 1, 0, 8, 3, 8, 192, 197, 162, 11, 3, 8, 192, 197, 226, 5, 82, 196, 204, 1, 0, 8, 3, 9, 192, 197, 34, 11, 3, 9, 192, 197, 98, 156, 42, 196, 196, 1, 0, 8, 3, 32, 192, 198, 226, 10, 3, 33, 192, 198, 34, 5, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 98, 10, 3, 9, 192, 197, 162, 156, 70, 196, 196, 1, 0, 8, 3, 33, 192, 198, 162, 11, 3, 33, 192, 198, 226, 5, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 34, 11, 3, 34, 192, 198, 98, 9, 0, 197, 227, 1, 0, 8, 3, 34, 192, 198, 163, 11, 3, 35, 192, 198, 3, 10, 0, 197, 67, 1, 0, 8, 3, 35, 192, 198, 99, 11, 3, 35, 192, 198, 195, 156, 82, 196, 196, 1, 0, 8, 3, 10, 192, 197, 162, 11, 3, 10, 192, 197, 226, 5, 82, 196, 196, 1, 0, 8, 3, 11, 192, 197, 34, 11, 3, 11, 192, 197, 98, 11, 0, 197, 163, 1, 0, 8, 3, 12, 192, 197, 3, 11, 3, 12, 192, 197, 99, 12, 0, 197, 195, 1, 0, 8, 3, 13, 192, 197, 35, 11, 3, 13, 192, 197, 131, 26, 0, 65, 0, 66, 0, 97, 0, 98, 0, 192, 0, 198, 0, 224, 0, 230, 0, 0, 1, 6, 1, 205, 1, 207, 1, 222, 1, 226, 1, 250, 1, 252, 1, 0, 2, 4, 2, 38, 2, 40, 2, 0, 30, 2, 30, 160, 30, 184, 30, 43, 33, 44, 33, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 185, 72, 185, 76, 185, 80, 186, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 68, 4, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 70, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 72, 4, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 94, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 96, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 121, 58, 217, 180, 185, 188, 185, 180, 185, 66, 185, 73, 153, 73, 249, 73, 89, 74, 186, 89, 154, 89, 250, 89, 90, 90, 249, 68, 122, 114, 185, 121, 153, 121, 249, 121, 121, 122, 217, 184, 136, 14, 185, 184, 185, 153, 153, 153, 249, 153, 217, 168, 153, 169, 0, 8, 2, 8, 168, 57, 136, 57, 232, 57, 104, 58, 200, 180, 168, 188, 168, 180, 168, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 232, 68, 104, 114, 168, 121, 136, 121, 232, 121, 104, 122, 200, 184, 128, 14, 168, 184, 168, 153, 136, 153, 232, 153, 200, 168, 136, 169, 4, 8, 72, 170, 249, 58, 232, 58, 217, 57, 200, 57, 98, 4, 100, 4, 153, 65, 136, 65, 249, 65, 232, 65, 153, 66, 136, 66, 25, 66, 8, 66, 25, 70, 8, 70, 217, 68, 200, 68, 249, 74, 232, 74, 217, 73, 200, 73, 153, 74, 136, 74, 217, 74, 200, 74, 25, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 153, 82, 136, 82, 185, 82, 168, 82, 250, 85, 232, 85, 26, 87, 8, 87, 122, 90, 104, 90, 250, 90, 232, 90, 218, 89, 200, 89, 218, 90, 200, 90, 154, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 185, 102, 168, 102, 168, 132, 153, 105, 136, 105, 185, 106, 168, 106, 25, 106, 8, 106, 57, 107, 40, 107, 25, 107, 8, 107, 154, 113, 136, 113, 186, 114, 168, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 249, 122, 232, 122, 217, 121, 200, 121, 249, 184, 232, 184, 12, 8, 14, 8, 153, 137, 136, 137, 185, 138, 168, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 186, 142, 168, 142, 26, 142, 8, 142, 185, 146, 168, 146, 25, 146, 8, 146, 185, 148, 168, 148, 121, 154, 104, 154, 249, 154, 232, 154, 217, 153, 200, 153, 57, 154, 40, 154, 249, 168, 232, 168, 102, 4, 104, 4, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 153, 174, 136, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 185, 144, 171, 84, 200, 140, 169, 140, 168, 144, 169, 84, 187, 88, 187, 96, 169, 88, 169, 96, 169, 176, 169, 112, 184, 120, 217, 72, 168, 120, 200, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 65, 4, 188, 188, 97, 4, 179, 188, 192, 4, 188, 189, 193, 4, 156, 189, 194, 4, 252, 189, 195, 4, 124, 190, 196, 4, 92, 190, 197, 4, 60, 190, 224, 4, 179, 189, 225, 4, 147, 189, 226, 4, 243, 189, 227, 4, 115, 190, 228, 4, 83, 190, 229, 4, 51, 190, 0, 5, 252, 190, 1, 5, 243, 190, 2, 5, 220, 189, 3, 5, 211, 189, 4, 5, 220, 190, 5, 5, 211, 190, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 97, 4, 170, 188, 224, 4, 170, 189, 225, 4, 138, 189, 226, 4, 234, 189, 227, 4, 106, 190, 228, 4, 74, 190, 229, 4, 42, 190, 1, 5, 234, 190, 3, 5, 202, 189, 5, 5, 202, 190, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 217, 58, 255, 5, 200, 58, 255, 5, 217, 154, 255, 5, 200, 154, 255, 1, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 199, 14, 0, 16, 72, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 191, 14, 0, 32, 129, 27, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 134, 27, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 148, 20, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 152, 20, 0, 0, 152, 20, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 67, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 112, 104, 111, 110, 101, 98, 111, 111, 107, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 101, 97, 114, 99, 104, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 38, 0, 65, 0, 69, 0, 60, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 38, 0, 79, 0, 69, 0, 60, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 38, 0, 85, 0, 69, 0, 60, 0, 60, 0, 252, 0, 60, 0, 60, 0, 60, 0, 220, 0, 0, 0, 48, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 100, 0, 101, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 104, 0, 111, 0, 110, 0, 101, 0, 98, 0, 107, 0, 93, 0, 0, 0, 170, 170, 82, 37, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 4, 29, 0, 0, 4, 29, 0, 0, 4, 29, 0, 0, 4, 29, 0, 0, 152, 31, 0, 0, 152, 31, 0, 0, 58, 33, 0, 0, 58, 33, 0, 0, 58, 37, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 195, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 84, 0, 92, 0, 97, 0, 105, 0, 110, 0, 118, 0, 121, 0, 129, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 153, 0, 48, 0, 48, 0, 157, 0, 48, 0, 165, 0, 170, 0, 178, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 186, 0, 145, 0, 48, 0, 48, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 80, 1, 132, 1, 184, 1, 228, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 19, 0, 202, 129, 19, 0, 202, 162, 19, 0, 202, 195, 19, 0, 202, 228, 19, 0, 202, 5, 20, 0, 202, 38, 20, 0, 202, 71, 20, 0, 202, 104, 20, 0, 202, 137, 20, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 166, 0, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 102, 1, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 134, 2, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 70, 3, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 6, 4, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 197, 34, 0, 0, 196, 146, 156, 42, 197, 162, 18, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 197, 98, 0, 0, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 197, 162, 0, 0, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 18, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 197, 66, 1, 0, 196, 146, 5, 42, 197, 34, 19, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 197, 130, 1, 0, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 197, 194, 1, 0, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 5, 0, 201, 230, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 6, 0, 201, 102, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 8, 0, 201, 230, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 9, 0, 201, 102, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 5, 0, 197, 195, 5, 0, 197, 227, 0, 0, 197, 3, 2, 0, 197, 35, 6, 0, 197, 131, 6, 0, 197, 227, 6, 0, 197, 67, 7, 0, 192, 0, 0, 0, 197, 163, 7, 0, 197, 3, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 11, 0, 201, 230, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 9, 0, 197, 131, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 12, 0, 201, 102, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 14, 0, 201, 230, 14, 0, 201, 166, 15, 0, 201, 102, 16, 0, 201, 38, 17, 0, 201, 230, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 18, 0, 201, 102, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 20, 0, 201, 230, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 21, 0, 201, 102, 22, 0, 201, 38, 23, 0, 201, 230, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 24, 0, 201, 102, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 156, 5, 0, 42, 6, 72, 0, 50, 156, 5, 0, 70, 6, 71, 0, 50, 156, 5, 0, 82, 6, 70, 0, 50, 156, 5, 0, 82, 6, 70, 0, 50, 5, 136, 0, 0, 28, 5, 0, 42, 5, 72, 0, 50, 28, 5, 0, 70, 5, 71, 0, 50, 28, 5, 0, 82, 5, 70, 0, 50, 28, 5, 0, 82, 5, 70, 0, 50, 5, 136, 0, 0, 156, 5, 0, 42, 6, 72, 0, 50, 5, 162, 0, 0, 28, 5, 0, 42, 5, 72, 0, 50, 5, 162, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 162, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 162, 0, 0, 156, 5, 0, 70, 6, 71, 0, 50, 5, 190, 0, 0, 28, 5, 0, 70, 5, 71, 0, 50, 5, 190, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 190, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 190, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 164, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 164, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 144, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 144, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 138, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 138, 0, 0, 156, 5, 0, 42, 6, 72, 0, 50, 5, 164, 0, 0, 28, 5, 0, 42, 5, 72, 0, 50, 5, 164, 0, 0, 156, 5, 0, 70, 6, 71, 0, 50, 5, 162, 0, 0, 28, 5, 0, 70, 5, 71, 0, 50, 5, 162, 0, 0, 156, 5, 0, 70, 6, 71, 0, 50, 5, 164, 0, 0, 28, 5, 0, 70, 5, 71, 0, 50, 5, 164, 0, 0, 156, 5, 0, 42, 6, 72, 0, 50, 5, 200, 0, 0, 28, 5, 0, 42, 5, 72, 0, 50, 5, 200, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 198, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 198, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 208, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 208, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 204, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 204, 0, 0, 156, 5, 0, 42, 6, 72, 0, 50, 5, 196, 0, 0, 28, 5, 0, 42, 5, 72, 0, 50, 5, 196, 0, 0, 156, 5, 0, 70, 6, 71, 0, 50, 5, 196, 0, 0, 28, 5, 0, 70, 5, 71, 0, 50, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 156, 5, 0, 70, 6, 71, 0, 50, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 28, 5, 0, 70, 5, 71, 0, 50, 5, 190, 0, 0, 5, 196, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 196, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 48, 0, 8, 3, 197, 162, 0, 70, 156, 5, 48, 0, 8, 3, 0, 192, 197, 98, 0, 82, 156, 5, 1, 0, 8, 3, 0, 192, 197, 162, 68, 3, 0, 192, 197, 227, 0, 42, 5, 5, 48, 0, 8, 3, 1, 192, 197, 66, 0, 70, 5, 5, 48, 0, 8, 3, 1, 192, 197, 130, 0, 82, 5, 5, 1, 0, 8, 3, 1, 192, 197, 194, 68, 3, 2, 192, 197, 3, 156, 42, 196, 162, 48, 0, 8, 3, 2, 192, 197, 99, 5, 42, 196, 162, 48, 0, 8, 3, 2, 192, 197, 195, 156, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 35, 5, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 131, 156, 70, 196, 190, 48, 0, 8, 3, 3, 192, 197, 227, 5, 70, 196, 190, 48, 0, 8, 3, 4, 192, 197, 67, 156, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 163, 5, 82, 196, 190, 48, 0, 8, 3, 5, 192, 197, 3, 156, 70, 196, 162, 48, 0, 8, 3, 8, 192, 197, 99, 5, 70, 196, 162, 48, 0, 8, 3, 8, 192, 197, 195, 156, 42, 196, 200, 48, 0, 8, 3, 9, 192, 197, 227, 5, 42, 196, 200, 48, 0, 8, 3, 10, 192, 197, 67, 156, 82, 196, 198, 48, 0, 8, 3, 10, 192, 197, 163, 5, 82, 196, 198, 48, 0, 8, 3, 11, 192, 197, 3, 156, 82, 196, 208, 48, 0, 8, 3, 11, 192, 197, 99, 5, 82, 196, 208, 48, 0, 8, 3, 11, 192, 197, 195, 156, 82, 196, 204, 48, 0, 8, 3, 12, 192, 197, 35, 5, 82, 196, 204, 48, 0, 8, 3, 12, 192, 197, 131, 156, 42, 196, 196, 48, 0, 8, 3, 12, 192, 197, 227, 5, 42, 196, 196, 48, 0, 8, 3, 13, 192, 197, 67, 156, 70, 196, 196, 48, 0, 8, 3, 13, 192, 197, 163, 5, 70, 196, 196, 48, 0, 8, 3, 14, 192, 197, 3, 14, 0, 197, 99, 48, 0, 8, 3, 14, 192, 197, 196, 15, 0, 197, 67, 48, 0, 8, 3, 15, 192, 197, 164, 156, 82, 196, 196, 48, 0, 8, 3, 16, 192, 197, 35, 5, 82, 196, 196, 48, 0, 8, 3, 16, 192, 197, 131, 16, 0, 197, 227, 48, 0, 8, 3, 17, 192, 197, 68, 17, 0, 197, 195, 48, 0, 8, 3, 18, 192, 197, 36, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 36, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 38, 4, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 40, 4, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 42, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 44, 4, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 46, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 48, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 0, 8, 57, 58, 2, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 154, 114, 185, 121, 153, 121, 249, 121, 153, 122, 4, 8, 136, 14, 217, 122, 185, 153, 153, 153, 249, 153, 6, 8, 153, 169, 185, 176, 8, 8, 168, 57, 136, 57, 232, 57, 136, 58, 10, 8, 40, 58, 12, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 136, 121, 232, 121, 136, 122, 14, 8, 128, 14, 200, 122, 168, 153, 136, 153, 232, 153, 16, 8, 136, 169, 168, 176, 72, 170, 57, 59, 40, 59, 217, 57, 200, 57, 50, 4, 52, 4, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 18, 8, 20, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 250, 114, 232, 114, 26, 114, 8, 114, 22, 8, 185, 116, 168, 116, 57, 123, 40, 123, 217, 121, 200, 121, 121, 122, 104, 122, 24, 8, 26, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 250, 142, 232, 142, 26, 142, 8, 142, 249, 146, 232, 146, 25, 146, 8, 146, 185, 148, 168, 148, 153, 154, 136, 154, 57, 155, 40, 155, 217, 153, 200, 153, 57, 154, 40, 154, 121, 154, 104, 154, 54, 4, 56, 4, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 185, 174, 168, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 28, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 30, 8, 1, 0, 8, 14, 32, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 34, 8, 200, 12, 72, 12, 32, 14, 185, 56, 9, 73, 217, 56, 187, 72, 185, 120, 233, 72, 185, 152, 201, 72, 200, 140, 169, 140, 169, 56, 8, 73, 200, 56, 169, 72, 169, 120, 232, 72, 169, 152, 200, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 25, 59, 255, 5, 8, 59, 255, 5, 25, 155, 255, 5, 8, 155, 255, 1, 170, 170, 3, 0, 70, 0, 61, 0, 32, 0, 67, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 172, 43, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 209, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 52, 37, 0, 0, 52, 37, 0, 0, 52, 37, 0, 0, 52, 37, 0, 0, 172, 41, 0, 0, 172, 41, 0, 0, 148, 43, 0, 0, 148, 43, 0, 0, 148, 43, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 70, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 84, 0, 92, 0, 97, 0, 105, 0, 110, 0, 118, 0, 121, 0, 129, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 132, 0, 140, 0, 146, 0, 154, 0, 162, 0, 170, 0, 48, 0, 56, 0, 169, 0, 177, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 194, 0, 202, 0, 48, 0, 48, 0, 48, 0, 210, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 218, 0, 48, 0, 48, 0, 226, 0, 48, 0, 232, 0, 240, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 243, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 251, 0, 48, 0, 48, 0, 255, 0, 48, 0, 7, 1, 12, 1, 20, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 28, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 31, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 185, 0, 192, 0, 48, 0, 48, 0, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 61, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 39, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 47, 1, 48, 0, 48, 0, 53, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 80, 1, 132, 1, 184, 1, 228, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 16, 2, 72, 2, 136, 2, 192, 0, 164, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 128, 34, 0, 202, 161, 34, 0, 202, 194, 34, 0, 202, 227, 34, 0, 202, 4, 35, 0, 202, 37, 35, 0, 202, 70, 35, 0, 202, 103, 35, 0, 202, 136, 35, 0, 202, 169, 35, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 198, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 102, 1, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 38, 2, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 70, 3, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 6, 4, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 198, 4, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 197, 34, 0, 0, 196, 146, 156, 42, 197, 98, 25, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 197, 98, 0, 0, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 197, 162, 0, 0, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 25, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 197, 66, 1, 0, 196, 146, 5, 42, 197, 226, 25, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 197, 130, 1, 0, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 197, 194, 1, 0, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 5, 0, 201, 166, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 7, 0, 201, 38, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 8, 0, 201, 166, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 10, 0, 201, 38, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 5, 0, 197, 195, 5, 0, 197, 227, 0, 0, 197, 3, 2, 0, 197, 35, 6, 0, 197, 131, 6, 0, 197, 227, 6, 0, 197, 67, 7, 0, 192, 0, 0, 0, 197, 163, 7, 0, 197, 3, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 11, 0, 201, 166, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 9, 0, 197, 131, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 102, 13, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 15, 0, 5, 71, 238, 102, 201, 38, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 10, 0, 192, 0, 0, 0, 197, 162, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 192, 0, 0, 0, 197, 98, 11, 0, 192, 0, 0, 0, 197, 162, 11, 0, 192, 0, 0, 0, 197, 226, 11, 0, 192, 0, 0, 0, 197, 34, 12, 0, 197, 99, 12, 0, 197, 194, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 13, 0, 197, 67, 13, 0, 197, 162, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 226, 9, 0, 197, 34, 14, 0, 5, 5, 10, 121, 197, 98, 14, 0, 197, 162, 14, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 14, 0, 197, 34, 15, 0, 197, 98, 15, 0, 197, 162, 15, 0, 197, 226, 15, 0, 197, 34, 16, 0, 197, 98, 16, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 10, 0, 5, 5, 24, 121, 197, 162, 10, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 17, 0, 201, 198, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 18, 0, 201, 70, 19, 0, 201, 6, 20, 0, 201, 198, 20, 0, 201, 134, 21, 0, 201, 70, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 23, 0, 201, 198, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 24, 0, 201, 70, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 26, 0, 201, 198, 26, 0, 201, 134, 27, 0, 201, 70, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 29, 0, 201, 198, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 156, 5, 0, 42, 6, 72, 0, 50, 156, 5, 0, 70, 6, 71, 0, 50, 156, 5, 0, 82, 6, 70, 0, 50, 156, 5, 0, 82, 6, 70, 0, 50, 5, 136, 0, 0, 28, 5, 0, 42, 5, 72, 0, 50, 28, 5, 0, 70, 5, 71, 0, 50, 28, 5, 0, 82, 5, 70, 0, 50, 28, 5, 0, 82, 5, 70, 0, 50, 5, 136, 0, 0, 156, 5, 0, 42, 6, 72, 0, 50, 5, 162, 0, 0, 28, 5, 0, 42, 5, 72, 0, 50, 5, 162, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 162, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 162, 0, 0, 156, 5, 0, 70, 6, 71, 0, 50, 5, 190, 0, 0, 28, 5, 0, 70, 5, 71, 0, 50, 5, 190, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 190, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 190, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 164, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 164, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 144, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 144, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 138, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 138, 0, 0, 156, 5, 0, 42, 6, 72, 0, 50, 5, 164, 0, 0, 28, 5, 0, 42, 5, 72, 0, 50, 5, 164, 0, 0, 156, 5, 0, 70, 6, 71, 0, 50, 5, 162, 0, 0, 28, 5, 0, 70, 5, 71, 0, 50, 5, 162, 0, 0, 156, 5, 0, 70, 6, 71, 0, 50, 5, 164, 0, 0, 28, 5, 0, 70, 5, 71, 0, 50, 5, 164, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 156, 5, 0, 42, 6, 72, 0, 50, 5, 200, 0, 0, 28, 5, 0, 42, 5, 72, 0, 50, 5, 200, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 198, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 198, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 208, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 208, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 204, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 204, 0, 0, 156, 5, 0, 42, 6, 72, 0, 50, 5, 196, 0, 0, 28, 5, 0, 42, 5, 72, 0, 50, 5, 196, 0, 0, 156, 5, 0, 70, 6, 71, 0, 50, 5, 196, 0, 0, 28, 5, 0, 70, 5, 71, 0, 50, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 156, 5, 0, 70, 6, 71, 0, 50, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 28, 5, 0, 70, 5, 71, 0, 50, 5, 190, 0, 0, 5, 196, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 196, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 156, 5, 0, 82, 6, 70, 0, 50, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 28, 5, 0, 82, 5, 70, 0, 50, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 226, 9, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 34, 10, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 98, 10, 0, 5, 5, 24, 121, 197, 162, 10, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 226, 10, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 34, 11, 0, 5, 5, 107, 121, 197, 98, 11, 0, 5, 5, 111, 121, 197, 162, 11, 0, 5, 5, 115, 121, 197, 226, 11, 0, 5, 5, 119, 121, 197, 34, 12, 0, 197, 99, 12, 0, 197, 194, 12, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 2, 13, 0, 197, 67, 13, 0, 197, 162, 13, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 226, 13, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 226, 9, 0, 197, 34, 14, 0, 5, 5, 10, 121, 197, 98, 14, 0, 197, 162, 14, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 14, 0, 197, 34, 15, 0, 197, 98, 15, 0, 197, 162, 15, 0, 197, 226, 15, 0, 197, 34, 16, 0, 197, 98, 16, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 10, 0, 5, 5, 24, 121, 197, 162, 10, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 48, 0, 8, 3, 197, 162, 0, 70, 156, 5, 48, 0, 8, 3, 0, 192, 197, 98, 0, 82, 156, 5, 1, 0, 8, 3, 0, 192, 197, 162, 68, 3, 0, 192, 197, 227, 0, 42, 5, 5, 48, 0, 8, 3, 1, 192, 197, 66, 0, 70, 5, 5, 48, 0, 8, 3, 1, 192, 197, 130, 0, 82, 5, 5, 1, 0, 8, 3, 1, 192, 197, 194, 68, 3, 2, 192, 197, 3, 156, 42, 196, 162, 48, 0, 8, 3, 2, 192, 197, 99, 5, 42, 196, 162, 48, 0, 8, 3, 2, 192, 197, 195, 156, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 35, 5, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 131, 156, 70, 196, 190, 48, 0, 8, 3, 3, 192, 197, 227, 5, 70, 196, 190, 48, 0, 8, 3, 4, 192, 197, 67, 156, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 163, 5, 82, 196, 190, 48, 0, 8, 3, 5, 192, 197, 3, 156, 70, 196, 162, 48, 0, 8, 3, 8, 192, 197, 99, 5, 70, 196, 162, 48, 0, 8, 3, 8, 192, 197, 195, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 48, 0, 8, 3, 16, 192, 197, 163, 5, 42, 196, 200, 48, 0, 8, 3, 17, 192, 197, 3, 156, 82, 196, 198, 48, 0, 8, 3, 17, 192, 197, 99, 5, 82, 196, 198, 48, 0, 8, 3, 17, 192, 197, 195, 156, 82, 196, 208, 48, 0, 8, 3, 18, 192, 197, 35, 5, 82, 196, 208, 48, 0, 8, 3, 18, 192, 197, 131, 156, 82, 196, 204, 48, 0, 8, 3, 18, 192, 197, 227, 5, 82, 196, 204, 48, 0, 8, 3, 19, 192, 197, 67, 156, 42, 196, 196, 48, 0, 8, 3, 19, 192, 197, 163, 5, 42, 196, 196, 48, 0, 8, 3, 20, 192, 197, 3, 156, 70, 196, 196, 48, 0, 8, 3, 20, 192, 197, 99, 5, 70, 196, 196, 48, 0, 8, 3, 20, 192, 197, 195, 21, 0, 197, 35, 48, 0, 8, 3, 21, 192, 197, 132, 22, 0, 197, 3, 48, 0, 8, 3, 22, 192, 197, 100, 156, 82, 196, 196, 48, 0, 8, 3, 22, 192, 197, 227, 5, 82, 196, 196, 48, 0, 8, 3, 23, 192, 197, 67, 23, 0, 197, 163, 48, 0, 8, 3, 24, 192, 197, 4, 24, 0, 197, 131, 48, 0, 8, 3, 24, 192, 197, 228, 3, 0, 70, 0, 61, 0, 32, 0, 159, 9, 0, 16, 37, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 85, 0, 170, 170, 153, 9, 0, 32, 139, 20, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 144, 20, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 106, 9, 0, 32, 7, 0, 0, 0, 22, 0, 0, 0, 110, 9, 0, 0, 110, 9, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 42, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 112, 104, 111, 110, 101, 98, 111, 111, 107, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 170, 170, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 38, 0, 97, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 38, 0, 111, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 38, 0, 117, 0, 60, 0, 252, 0, 60, 0, 60, 0, 60, 0, 220, 0, 38, 0, 115, 0, 115, 0, 60, 0, 223, 0, 60, 0, 60, 0, 60, 0, 158, 30, 0, 0, 170, 170, 170, 170, 218, 36, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 100, 29, 0, 0, 100, 29, 0, 0, 100, 29, 0, 0, 100, 29, 0, 0, 48, 31, 0, 0, 48, 31, 0, 0, 218, 32, 0, 0, 218, 32, 0, 0, 194, 36, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 201, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 84, 0, 92, 0, 97, 0, 105, 0, 110, 0, 118, 0, 121, 0, 129, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 153, 0, 48, 0, 48, 0, 157, 0, 163, 0, 171, 0, 176, 0, 184, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 145, 0, 48, 0, 48, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 80, 1, 132, 1, 184, 1, 228, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 13, 0, 202, 65, 13, 0, 202, 98, 13, 0, 202, 131, 13, 0, 202, 164, 13, 0, 202, 197, 13, 0, 202, 230, 13, 0, 202, 7, 14, 0, 202, 40, 14, 0, 202, 73, 14, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 198, 0, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 166, 1, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 198, 2, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 134, 3, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 102, 4, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 134, 5, 2, 43, 196, 146, 156, 42, 197, 162, 12, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 134, 5, 2, 71, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 5, 2, 83, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 5, 5, 2, 43, 196, 146, 5, 42, 197, 226, 12, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 5, 5, 2, 71, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 5, 2, 83, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 5, 0, 201, 102, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 7, 0, 201, 230, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 8, 0, 201, 102, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 10, 0, 201, 230, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 2, 0, 197, 34, 3, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 226, 3, 0, 197, 34, 4, 0, 192, 0, 0, 0, 197, 98, 4, 0, 197, 162, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 11, 0, 201, 102, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 5, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 13, 0, 201, 230, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 14, 0, 201, 102, 15, 0, 201, 38, 16, 0, 201, 230, 16, 0, 201, 166, 17, 0, 201, 102, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 201, 38, 19, 0, 201, 230, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 20, 0, 201, 102, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 22, 0, 201, 230, 22, 0, 201, 166, 23, 0, 201, 102, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 25, 0, 201, 230, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 83, 5, 136, 0, 0, 5, 5, 2, 83, 5, 136, 0, 0, 5, 5, 0, 78, 5, 5, 2, 79, 134, 5, 2, 43, 5, 162, 0, 0, 5, 5, 2, 43, 5, 162, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 134, 5, 2, 83, 5, 164, 0, 0, 5, 5, 2, 83, 5, 164, 0, 0, 134, 5, 2, 83, 5, 144, 0, 0, 5, 5, 2, 83, 5, 144, 0, 0, 134, 5, 2, 83, 5, 138, 0, 0, 5, 5, 2, 83, 5, 138, 0, 0, 134, 5, 2, 43, 5, 164, 0, 0, 5, 5, 2, 43, 5, 164, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 2, 71, 5, 164, 0, 0, 5, 5, 2, 71, 5, 164, 0, 0, 134, 5, 2, 43, 5, 200, 0, 0, 5, 5, 2, 43, 5, 200, 0, 0, 134, 5, 2, 83, 5, 198, 0, 0, 5, 5, 2, 83, 5, 198, 0, 0, 134, 5, 2, 83, 5, 208, 0, 0, 5, 5, 2, 83, 5, 208, 0, 0, 134, 5, 2, 83, 5, 204, 0, 0, 5, 5, 2, 83, 5, 204, 0, 0, 133, 5, 0, 78, 134, 5, 2, 79, 134, 5, 2, 43, 5, 196, 0, 0, 5, 5, 2, 43, 5, 196, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 196, 0, 0, 5, 5, 2, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 48, 0, 8, 3, 2, 235, 134, 5, 0, 70, 156, 5, 48, 0, 8, 3, 255, 255, 2, 71, 134, 5, 0, 82, 156, 5, 1, 0, 8, 3, 255, 255, 2, 83, 134, 5, 68, 3, 197, 162, 0, 42, 5, 5, 48, 0, 8, 3, 2, 235, 5, 5, 0, 70, 5, 5, 48, 0, 8, 3, 255, 255, 2, 71, 5, 5, 0, 82, 5, 5, 1, 0, 8, 3, 255, 255, 2, 83, 5, 5, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 48, 0, 8, 3, 0, 192, 197, 226, 5, 42, 196, 162, 48, 0, 8, 3, 1, 192, 197, 34, 156, 82, 196, 162, 48, 0, 8, 3, 1, 192, 197, 98, 5, 82, 196, 162, 48, 0, 8, 3, 1, 192, 197, 162, 156, 70, 196, 190, 48, 0, 8, 3, 1, 192, 197, 226, 5, 70, 196, 190, 48, 0, 8, 3, 2, 192, 197, 34, 156, 82, 196, 190, 48, 0, 8, 3, 2, 192, 197, 98, 5, 82, 196, 190, 48, 0, 8, 3, 2, 192, 197, 162, 156, 70, 196, 162, 48, 0, 8, 3, 4, 192, 197, 226, 5, 70, 196, 162, 48, 0, 8, 3, 5, 192, 197, 34, 156, 42, 196, 200, 48, 0, 8, 3, 5, 192, 197, 226, 5, 42, 196, 200, 48, 0, 8, 3, 6, 192, 197, 34, 156, 82, 196, 198, 48, 0, 8, 3, 6, 192, 197, 98, 5, 82, 196, 198, 48, 0, 8, 3, 6, 192, 197, 162, 156, 82, 196, 208, 48, 0, 8, 3, 6, 192, 197, 226, 5, 82, 196, 208, 48, 0, 8, 3, 7, 192, 197, 34, 156, 82, 196, 204, 48, 0, 8, 3, 7, 192, 197, 98, 5, 82, 196, 204, 48, 0, 8, 3, 7, 192, 197, 162, 156, 42, 196, 196, 48, 0, 8, 3, 8, 192, 197, 34, 5, 42, 196, 196, 48, 0, 8, 3, 8, 192, 197, 98, 156, 70, 196, 196, 48, 0, 8, 3, 8, 192, 197, 162, 5, 70, 196, 196, 48, 0, 8, 3, 8, 192, 197, 226, 9, 0, 197, 35, 48, 0, 8, 3, 9, 192, 197, 131, 9, 0, 197, 227, 48, 0, 8, 3, 10, 192, 197, 67, 156, 82, 196, 196, 48, 0, 8, 3, 10, 192, 197, 162, 5, 82, 196, 196, 48, 0, 8, 3, 10, 192, 197, 226, 11, 0, 197, 35, 48, 0, 8, 3, 11, 192, 197, 131, 11, 0, 197, 227, 48, 0, 8, 3, 12, 192, 197, 67, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 64, 185, 68, 185, 72, 186, 76, 185, 80, 185, 84, 185, 88, 186, 92, 186, 100, 185, 104, 185, 108, 185, 112, 186, 116, 28, 4, 185, 132, 185, 136, 185, 144, 185, 148, 185, 156, 30, 4, 185, 172, 185, 176, 185, 180, 185, 184, 185, 188, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 32, 4, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 92, 168, 100, 168, 104, 168, 108, 168, 112, 168, 116, 34, 4, 168, 132, 168, 136, 168, 144, 168, 148, 168, 156, 36, 4, 168, 172, 168, 176, 168, 180, 168, 184, 168, 188, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 124, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 185, 60, 57, 58, 0, 8, 249, 70, 186, 77, 154, 77, 250, 77, 90, 78, 186, 93, 154, 93, 250, 93, 90, 94, 217, 72, 154, 118, 185, 125, 153, 125, 249, 125, 153, 126, 185, 128, 136, 14, 217, 126, 185, 165, 153, 165, 249, 165, 185, 168, 153, 185, 185, 192, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 168, 60, 40, 58, 4, 8, 232, 70, 168, 77, 136, 77, 232, 77, 72, 78, 168, 93, 136, 93, 232, 93, 72, 94, 200, 72, 136, 118, 168, 125, 136, 125, 232, 125, 136, 126, 168, 128, 128, 14, 200, 126, 168, 165, 136, 165, 232, 165, 168, 168, 136, 185, 168, 192, 72, 186, 57, 59, 40, 59, 217, 57, 200, 57, 38, 4, 40, 4, 153, 69, 136, 69, 249, 69, 232, 69, 185, 70, 168, 70, 25, 70, 8, 70, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 249, 85, 232, 85, 217, 85, 200, 85, 185, 86, 168, 86, 249, 86, 232, 86, 249, 89, 232, 89, 89, 91, 72, 91, 154, 94, 136, 94, 58, 95, 40, 95, 218, 93, 200, 93, 26, 95, 8, 95, 186, 94, 168, 96, 6, 8, 8, 8, 250, 101, 232, 101, 249, 106, 232, 106, 168, 140, 153, 109, 136, 109, 249, 110, 232, 110, 25, 110, 8, 110, 121, 111, 104, 111, 89, 111, 72, 111, 154, 117, 136, 117, 250, 118, 232, 118, 26, 118, 8, 118, 10, 8, 185, 120, 168, 120, 57, 127, 40, 127, 217, 125, 200, 125, 121, 126, 104, 126, 12, 8, 14, 8, 153, 145, 136, 145, 249, 146, 232, 146, 25, 146, 8, 146, 153, 149, 136, 149, 249, 149, 232, 149, 249, 150, 232, 150, 25, 150, 8, 150, 249, 158, 232, 158, 25, 158, 8, 158, 185, 160, 168, 160, 153, 166, 136, 166, 57, 167, 40, 167, 217, 165, 200, 165, 57, 166, 40, 166, 121, 166, 104, 166, 42, 4, 44, 4, 249, 177, 232, 177, 249, 185, 232, 185, 89, 186, 153, 189, 136, 189, 185, 190, 168, 190, 25, 190, 8, 190, 200, 148, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 168, 148, 168, 152, 200, 56, 169, 76, 187, 92, 187, 100, 169, 92, 169, 100, 169, 196, 169, 116, 217, 124, 187, 76, 200, 124, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 124, 255, 5, 185, 164, 255, 5, 168, 56, 255, 5, 168, 124, 255, 5, 168, 164, 255, 5, 25, 59, 255, 5, 8, 59, 255, 5, 25, 167, 255, 5, 8, 167, 255, 1, 170, 170, 3, 0, 70, 0, 61, 0, 32, 0, 43, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 99, 9, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 104, 9, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 20, 0, 0, 32, 7, 0, 0, 0, 13, 0, 0, 0, 22, 0, 0, 0, 22, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 20, 0, 0, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 100, 101, 102, 97, 117, 108, 116, 0, 170, 0, 0, 112, 0, 104, 0, 111, 0, 110, 0, 101, 0, 98, 0, 111, 0, 111, 0, 107, 0, 0, 0, 1, 0, 43, 0, 1, 0, 1, 0, 32, 0, 11, 0, 0, 80, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 95, 10, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 99, 10, 0, 0, 99, 10, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 78, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 105, 220, 38, 0, 67, 0, 60, 0, 13, 1, 60, 0, 60, 0, 60, 0, 12, 1, 60, 0, 7, 1, 60, 0, 60, 0, 60, 0, 6, 1, 38, 0, 69, 0, 60, 0, 27, 1, 60, 0, 60, 0, 60, 0, 26, 1, 38, 0, 72, 0, 60, 0, 99, 0, 104, 0, 60, 0, 60, 0, 60, 0, 99, 0, 72, 0, 60, 0, 60, 0, 60, 0, 67, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 72, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 32, 0, 76, 0, 60, 0, 66, 1, 60, 0, 60, 0, 60, 0, 65, 1, 38, 0, 78, 0, 60, 0, 68, 1, 60, 0, 60, 0, 60, 0, 67, 1, 38, 0, 82, 0, 60, 0, 85, 1, 60, 0, 60, 0, 60, 0, 84, 1, 38, 0, 83, 0, 60, 0, 97, 1, 60, 0, 60, 0, 60, 0, 96, 1, 60, 0, 91, 1, 60, 0, 60, 0, 60, 0, 90, 1, 38, 0, 90, 0, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 60, 0, 122, 1, 60, 0, 60, 0, 60, 0, 121, 1, 0, 0, 170, 170, 170, 170, 170, 170, 32, 40, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 84, 28, 0, 0, 84, 28, 0, 0, 84, 28, 0, 0, 84, 28, 0, 0, 40, 31, 0, 0, 40, 31, 0, 0, 182, 35, 0, 0, 208, 35, 0, 0, 8, 40, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 184, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 48, 0, 56, 0, 48, 0, 56, 0, 112, 0, 120, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 142, 0, 48, 0, 149, 0, 157, 0, 161, 0, 167, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 136, 0, 48, 0, 48, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 0, 192, 0, 192, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 21, 0, 202, 129, 21, 0, 202, 162, 21, 0, 202, 195, 21, 0, 202, 228, 21, 0, 202, 5, 22, 0, 202, 38, 22, 0, 202, 71, 22, 0, 202, 104, 22, 0, 202, 137, 22, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 4, 0, 0, 156, 5, 0, 48, 201, 134, 8, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 70, 9, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 201, 166, 10, 0, 201, 6, 12, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 230, 13, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 196, 15, 0, 5, 5, 0, 48, 201, 102, 24, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 38, 25, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 201, 134, 26, 0, 201, 230, 27, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 198, 29, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 20, 0, 201, 166, 31, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 20, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 21, 0, 201, 38, 33, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 47, 5, 5, 3, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 34, 0, 201, 102, 35, 0, 134, 5, 2, 51, 5, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 21, 63, 5, 5, 21, 63, 134, 5, 2, 69, 5, 5, 2, 69, 201, 38, 36, 0, 201, 70, 37, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 77, 5, 5, 2, 77, 201, 102, 38, 0, 201, 134, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 79, 5, 5, 3, 79, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 40, 0, 201, 38, 42, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 93, 5, 5, 3, 93, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 43, 0, 201, 38, 45, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 46, 0, 201, 102, 47, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 7, 0, 197, 34, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 48, 0, 201, 230, 48, 0, 201, 166, 49, 0, 201, 102, 50, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 51, 0, 201, 70, 52, 0, 201, 102, 53, 0, 201, 134, 54, 0, 201, 166, 55, 0, 201, 198, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 57, 0, 201, 6, 59, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 60, 0, 201, 70, 61, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 62, 0, 201, 230, 63, 0, 197, 34, 17, 0, 197, 98, 17, 0, 197, 162, 17, 0, 197, 226, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 65, 0, 201, 230, 66, 0, 201, 102, 68, 0, 201, 230, 69, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 71, 0, 201, 38, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 136, 5, 2, 57, 5, 142, 0, 0, 71, 5, 2, 57, 5, 142, 0, 0, 136, 5, 2, 57, 5, 144, 0, 0, 71, 5, 2, 57, 5, 144, 0, 0, 136, 5, 2, 57, 5, 156, 0, 0, 71, 5, 2, 57, 5, 156, 0, 0, 136, 5, 2, 57, 5, 196, 0, 0, 71, 5, 2, 57, 5, 196, 0, 0, 136, 5, 2, 57, 5, 150, 0, 0, 71, 5, 2, 57, 5, 150, 0, 0, 136, 5, 2, 57, 5, 160, 0, 0, 71, 5, 2, 57, 5, 160, 0, 0, 136, 5, 2, 57, 5, 206, 0, 0, 71, 5, 2, 57, 5, 206, 0, 0, 71, 5, 2, 57, 5, 210, 0, 0, 70, 5, 2, 57, 5, 142, 0, 0, 5, 5, 2, 57, 5, 142, 0, 0, 70, 5, 2, 57, 5, 144, 0, 0, 5, 5, 2, 57, 5, 144, 0, 0, 70, 5, 2, 57, 5, 156, 0, 0, 5, 5, 2, 57, 5, 156, 0, 0, 70, 5, 2, 57, 5, 196, 0, 0, 5, 5, 2, 57, 5, 196, 0, 0, 70, 5, 2, 57, 5, 150, 0, 0, 5, 5, 2, 57, 5, 150, 0, 0, 70, 5, 2, 57, 5, 160, 0, 0, 5, 5, 2, 57, 5, 160, 0, 0, 70, 5, 2, 57, 5, 206, 0, 0, 5, 5, 2, 57, 5, 206, 0, 0, 5, 5, 2, 57, 5, 210, 0, 0, 134, 5, 3, 47, 5, 160, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 3, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 51, 5, 162, 0, 0, 5, 5, 2, 51, 5, 162, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 77, 5, 160, 0, 0, 5, 5, 2, 77, 5, 160, 0, 0, 134, 5, 3, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 3, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 3, 79, 5, 202, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 3, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 51, 5, 160, 0, 0, 5, 5, 2, 51, 5, 160, 0, 0, 134, 5, 2, 51, 5, 204, 0, 0, 5, 5, 2, 51, 5, 204, 0, 0, 134, 5, 2, 51, 5, 208, 0, 0, 5, 5, 2, 51, 5, 208, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 134, 5, 2, 77, 5, 196, 0, 0, 5, 5, 2, 77, 5, 196, 0, 0, 134, 5, 2, 77, 5, 210, 0, 0, 5, 5, 2, 77, 5, 210, 0, 0, 134, 5, 3, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 3, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 3, 79, 5, 156, 0, 0, 5, 5, 3, 79, 5, 156, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 3, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 3, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 3, 93, 5, 210, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 3, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 134, 5, 2, 51, 5, 196, 0, 0, 5, 5, 2, 51, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 19, 0, 35, 30, 32, 0, 40, 30, 15, 0, 40, 30, 2, 192, 197, 162, 41, 30, 2, 192, 197, 226, 42, 30, 3, 192, 197, 34, 43, 30, 3, 192, 197, 98, 150, 30, 3, 192, 197, 162, 35, 30, 1, 192, 197, 98, 36, 30, 1, 192, 197, 162, 37, 30, 1, 192, 197, 226, 38, 30, 2, 192, 197, 34, 39, 30, 2, 192, 197, 98, 31, 2, 15, 0, 31, 2, 0, 192, 197, 226, 1, 3, 3, 239, 134, 5, 12, 3, 2, 239, 134, 5, 65, 3, 3, 239, 134, 5, 34, 30, 1, 192, 197, 34, 72, 0, 2, 249, 136, 5, 104, 0, 2, 249, 71, 5, 36, 1, 197, 162, 37, 1, 0, 192, 197, 98, 30, 2, 0, 192, 197, 162, 0, 50, 156, 5, 48, 0, 12, 3, 2, 243, 134, 5, 0, 68, 156, 5, 1, 0, 1, 3, 255, 255, 2, 69, 134, 5, 65, 3, 255, 255, 2, 69, 134, 5, 0, 76, 156, 5, 1, 0, 1, 3, 255, 255, 2, 77, 134, 5, 65, 3, 255, 255, 2, 77, 134, 5, 0, 78, 156, 5, 2, 0, 1, 3, 255, 255, 3, 79, 134, 5, 12, 3, 255, 255, 2, 79, 134, 5, 65, 3, 255, 255, 3, 79, 134, 5, 0, 92, 156, 5, 2, 0, 1, 3, 255, 255, 3, 93, 134, 5, 12, 3, 255, 255, 2, 93, 134, 5, 65, 3, 255, 255, 3, 93, 134, 5, 0, 46, 5, 5, 19, 0, 35, 30, 32, 0, 40, 30, 15, 0, 40, 30, 6, 192, 197, 98, 41, 30, 6, 192, 197, 162, 42, 30, 6, 192, 197, 226, 43, 30, 7, 192, 197, 34, 150, 30, 7, 192, 197, 98, 35, 30, 5, 192, 197, 34, 36, 30, 5, 192, 197, 98, 37, 30, 5, 192, 197, 162, 38, 30, 5, 192, 197, 226, 39, 30, 6, 192, 197, 34, 31, 2, 15, 0, 31, 2, 4, 192, 197, 162, 1, 3, 3, 239, 5, 5, 12, 3, 2, 239, 5, 5, 65, 3, 3, 239, 5, 5, 34, 30, 4, 192, 197, 226, 72, 0, 2, 249, 70, 5, 104, 0, 2, 249, 5, 5, 36, 1, 3, 192, 197, 226, 37, 1, 4, 192, 197, 34, 30, 2, 4, 192, 197, 98, 0, 50, 5, 5, 48, 0, 12, 3, 2, 243, 5, 5, 0, 68, 5, 5, 1, 0, 1, 3, 255, 255, 2, 69, 5, 5, 65, 3, 255, 255, 2, 69, 5, 5, 0, 76, 5, 5, 1, 0, 1, 3, 255, 255, 2, 77, 5, 5, 65, 3, 255, 255, 2, 77, 5, 5, 0, 78, 5, 5, 2, 0, 1, 3, 255, 255, 3, 79, 5, 5, 12, 3, 255, 255, 2, 79, 5, 5, 65, 3, 255, 255, 3, 79, 5, 5, 0, 92, 5, 5, 2, 0, 1, 3, 255, 255, 3, 93, 5, 5, 12, 3, 255, 255, 2, 93, 5, 5, 65, 3, 255, 255, 3, 93, 5, 5, 156, 46, 196, 160, 2, 0, 1, 3, 7, 192, 197, 162, 12, 3, 7, 192, 197, 226, 65, 3, 7, 192, 197, 162, 5, 46, 196, 160, 2, 0, 1, 3, 8, 192, 197, 34, 12, 3, 8, 192, 197, 98, 65, 3, 8, 192, 197, 34, 156, 50, 196, 162, 48, 0, 12, 3, 8, 192, 197, 162, 5, 50, 196, 162, 48, 0, 12, 3, 8, 192, 197, 226, 156, 68, 196, 160, 1, 0, 1, 3, 9, 192, 197, 34, 65, 3, 9, 192, 197, 34, 5, 68, 196, 160, 1, 0, 1, 3, 9, 192, 197, 98, 65, 3, 9, 192, 197, 98, 156, 76, 196, 160, 1, 0, 1, 3, 9, 192, 197, 162, 65, 3, 9, 192, 197, 162, 5, 76, 196, 160, 1, 0, 1, 3, 9, 192, 197, 226, 65, 3, 9, 192, 197, 226, 156, 78, 196, 160, 2, 0, 1, 3, 10, 192, 197, 34, 12, 3, 10, 192, 197, 98, 65, 3, 10, 192, 197, 34, 5, 78, 196, 160, 2, 0, 1, 3, 10, 192, 197, 162, 12, 3, 10, 192, 197, 226, 65, 3, 10, 192, 197, 162, 156, 78, 196, 202, 2, 0, 1, 3, 11, 192, 197, 34, 12, 3, 11, 192, 197, 98, 65, 3, 11, 192, 197, 34, 5, 78, 196, 202, 2, 0, 1, 3, 11, 192, 197, 162, 12, 3, 11, 192, 197, 226, 65, 3, 11, 192, 197, 162, 156, 50, 196, 160, 48, 0, 12, 3, 12, 192, 197, 34, 5, 50, 196, 160, 48, 0, 12, 3, 12, 192, 197, 98, 156, 50, 196, 204, 48, 0, 12, 3, 12, 192, 197, 162, 5, 50, 196, 204, 48, 0, 12, 3, 12, 192, 197, 226, 156, 50, 196, 208, 48, 0, 12, 3, 13, 192, 197, 34, 5, 50, 196, 208, 48, 0, 12, 3, 13, 192, 197, 98, 156, 68, 196, 196, 1, 0, 1, 3, 13, 192, 197, 162, 65, 3, 13, 192, 197, 162, 5, 68, 196, 196, 1, 0, 1, 3, 13, 192, 197, 226, 65, 3, 13, 192, 197, 226, 156, 68, 196, 210, 1, 0, 1, 3, 14, 192, 197, 34, 65, 3, 14, 192, 197, 34, 5, 68, 196, 210, 1, 0, 1, 3, 14, 192, 197, 98, 65, 3, 14, 192, 197, 98, 156, 68, 196, 204, 1, 0, 1, 3, 14, 192, 197, 162, 65, 3, 14, 192, 197, 162, 5, 68, 196, 204, 1, 0, 1, 3, 14, 192, 197, 226, 65, 3, 14, 192, 197, 226, 156, 76, 196, 196, 1, 0, 1, 3, 15, 192, 197, 34, 65, 3, 15, 192, 197, 34, 5, 76, 196, 196, 1, 0, 1, 3, 15, 192, 197, 98, 65, 3, 15, 192, 197, 98, 156, 76, 196, 210, 1, 0, 1, 3, 15, 192, 197, 162, 65, 3, 15, 192, 197, 162, 5, 76, 196, 210, 1, 0, 1, 3, 15, 192, 197, 226, 65, 3, 15, 192, 197, 226, 156, 78, 196, 196, 2, 0, 1, 3, 16, 192, 197, 34, 12, 3, 16, 192, 197, 98, 65, 3, 16, 192, 197, 34, 5, 78, 196, 196, 2, 0, 1, 3, 16, 192, 197, 162, 12, 3, 16, 192, 197, 226, 65, 3, 16, 192, 197, 162, 156, 92, 196, 196, 2, 0, 1, 3, 18, 192, 197, 34, 12, 3, 18, 192, 197, 98, 65, 3, 18, 192, 197, 34, 5, 92, 196, 196, 2, 0, 1, 3, 18, 192, 197, 162, 12, 3, 18, 192, 197, 226, 65, 3, 18, 192, 197, 162, 156, 92, 196, 210, 2, 0, 1, 3, 19, 192, 197, 34, 12, 3, 19, 192, 197, 98, 65, 3, 19, 192, 197, 34, 5, 92, 196, 210, 2, 0, 1, 3, 19, 192, 197, 162, 12, 3, 19, 192, 197, 226, 65, 3, 19, 192, 197, 162, 156, 50, 196, 196, 48, 0, 12, 3, 20, 192, 197, 34, 5, 50, 196, 196, 48, 0, 12, 3, 20, 192, 197, 98, 12, 0, 72, 0, 73, 0, 104, 0, 105, 0, 36, 1, 38, 1, 30, 2, 32, 2, 34, 30, 44, 30, 150, 30, 151, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 76, 36, 4, 185, 88, 185, 92, 185, 96, 186, 104, 186, 112, 185, 116, 185, 124, 185, 128, 38, 4, 185, 144, 185, 148, 185, 152, 40, 4, 42, 4, 185, 180, 185, 188, 185, 192, 185, 196, 185, 200, 185, 204, 44, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 46, 4, 168, 76, 56, 4, 168, 88, 168, 92, 168, 96, 168, 104, 168, 112, 168, 116, 168, 124, 168, 128, 58, 4, 168, 144, 168, 148, 168, 152, 60, 4, 62, 4, 168, 180, 168, 188, 168, 192, 168, 196, 168, 200, 168, 204, 64, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 144, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 66, 4, 186, 81, 154, 81, 250, 81, 90, 82, 186, 105, 154, 105, 250, 105, 90, 106, 217, 76, 154, 134, 185, 145, 153, 145, 249, 145, 153, 146, 89, 146, 136, 14, 217, 146, 185, 189, 153, 189, 249, 189, 89, 190, 153, 205, 185, 220, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 68, 4, 168, 81, 136, 81, 232, 81, 72, 82, 168, 105, 136, 105, 232, 105, 72, 106, 200, 76, 136, 134, 168, 145, 136, 145, 232, 145, 136, 146, 72, 146, 128, 14, 200, 146, 168, 189, 136, 189, 232, 189, 72, 190, 136, 205, 168, 220, 72, 206, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 185, 72, 168, 72, 249, 65, 232, 65, 185, 66, 168, 66, 185, 68, 168, 68, 25, 78, 8, 78, 89, 79, 72, 79, 58, 83, 40, 83, 218, 81, 200, 81, 186, 82, 168, 82, 70, 4, 72, 4, 185, 84, 168, 84, 249, 93, 232, 93, 217, 93, 200, 93, 185, 94, 168, 94, 249, 94, 232, 94, 249, 97, 232, 97, 89, 99, 72, 99, 154, 106, 136, 106, 58, 107, 40, 107, 218, 105, 200, 105, 26, 107, 8, 107, 186, 106, 168, 108, 6, 8, 8, 8, 250, 113, 232, 113, 249, 118, 232, 118, 168, 156, 153, 125, 136, 125, 249, 126, 232, 126, 25, 126, 8, 126, 121, 127, 104, 127, 185, 120, 168, 120, 185, 136, 168, 136, 74, 4, 76, 4, 26, 134, 8, 134, 10, 8, 185, 140, 168, 140, 57, 147, 40, 147, 217, 145, 200, 145, 121, 146, 104, 146, 12, 8, 14, 8, 185, 164, 168, 164, 78, 4, 80, 4, 25, 162, 8, 162, 185, 176, 168, 176, 250, 169, 232, 169, 82, 4, 84, 4, 185, 172, 168, 172, 249, 182, 232, 182, 25, 182, 8, 182, 185, 184, 168, 184, 153, 190, 136, 190, 57, 191, 40, 191, 217, 189, 200, 189, 57, 190, 40, 190, 121, 190, 104, 190, 25, 191, 8, 191, 249, 197, 232, 197, 249, 205, 232, 205, 89, 206, 185, 216, 168, 216, 185, 210, 168, 210, 185, 212, 168, 212, 232, 168, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 80, 200, 168, 169, 168, 200, 56, 169, 80, 187, 104, 187, 112, 169, 104, 169, 112, 169, 224, 169, 132, 217, 144, 187, 80, 200, 144, 169, 80, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 72, 4, 187, 100, 104, 4, 178, 100, 36, 5, 251, 101, 37, 5, 242, 101, 255, 5, 186, 80, 255, 5, 186, 132, 255, 5, 185, 160, 255, 5, 186, 168, 255, 5, 185, 208, 255, 5, 168, 64, 72, 4, 177, 100, 104, 4, 168, 100, 36, 5, 241, 101, 37, 5, 232, 101, 255, 5, 168, 80, 255, 5, 168, 132, 255, 5, 168, 160, 255, 5, 168, 168, 255, 5, 168, 208, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 26, 83, 255, 5, 8, 83, 255, 5, 250, 134, 255, 5, 232, 134, 255, 5, 249, 162, 255, 5, 232, 162, 255, 5, 250, 170, 255, 5, 232, 170, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 79, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 88, 10, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 93, 10, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 76, 11, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 80, 11, 0, 0, 80, 11, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 94, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 138, 220, 38, 0, 12, 3, 60, 0, 60, 0, 2, 3, 38, 0, 68, 0, 60, 0, 100, 0, 122, 0, 60, 0, 60, 0, 60, 0, 68, 0, 122, 0, 60, 0, 60, 0, 60, 0, 68, 0, 90, 0, 60, 0, 86, 2, 60, 0, 60, 0, 60, 0, 137, 1, 38, 0, 69, 0, 60, 0, 91, 2, 60, 0, 60, 0, 60, 0, 144, 1, 38, 0, 70, 0, 60, 0, 146, 1, 60, 0, 60, 0, 60, 0, 145, 1, 38, 0, 71, 0, 60, 0, 103, 0, 98, 0, 60, 0, 60, 0, 60, 0, 71, 0, 98, 0, 60, 0, 60, 0, 60, 0, 71, 0, 66, 0, 60, 0, 99, 2, 60, 0, 60, 0, 60, 0, 148, 1, 38, 0, 72, 0, 60, 0, 120, 0, 60, 0, 60, 0, 60, 0, 88, 0, 38, 0, 75, 0, 60, 0, 107, 0, 112, 0, 60, 0, 60, 0, 60, 0, 75, 0, 112, 0, 60, 0, 60, 0, 60, 0, 75, 0, 80, 0, 38, 0, 78, 0, 60, 0, 110, 0, 121, 0, 60, 0, 60, 0, 60, 0, 78, 0, 121, 0, 60, 0, 60, 0, 60, 0, 78, 0, 89, 0, 60, 0, 75, 1, 60, 0, 60, 0, 60, 0, 74, 1, 38, 0, 79, 0, 60, 0, 84, 2, 60, 0, 60, 0, 60, 0, 134, 1, 38, 0, 84, 0, 60, 0, 116, 0, 115, 0, 60, 0, 60, 0, 60, 0, 84, 0, 115, 0, 60, 0, 60, 0, 60, 0, 84, 0, 83, 0, 38, 0, 86, 0, 60, 0, 139, 2, 60, 0, 60, 0, 60, 0, 178, 1, 0, 0, 170, 170, 170, 170, 146, 43, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 164, 30, 0, 0, 164, 30, 0, 0, 164, 30, 0, 0, 164, 30, 0, 0, 116, 35, 0, 0, 116, 35, 0, 0, 136, 38, 0, 0, 222, 38, 0, 0, 122, 43, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 221, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 94, 0, 102, 0, 109, 0, 117, 0, 48, 0, 56, 0, 48, 0, 56, 0, 122, 0, 130, 0, 136, 0, 144, 0, 48, 0, 56, 0, 152, 0, 160, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 182, 0, 189, 0, 197, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 204, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 176, 0, 48, 0, 48, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 120, 1, 180, 1, 192, 0, 192, 0, 232, 1, 32, 2, 192, 0, 96, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 64, 37, 0, 202, 97, 37, 0, 202, 130, 37, 0, 202, 163, 37, 0, 202, 196, 37, 0, 202, 229, 37, 0, 202, 6, 38, 0, 202, 39, 38, 0, 202, 72, 38, 0, 202, 105, 38, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 201, 0, 0, 0, 156, 5, 0, 50, 156, 5, 0, 52, 201, 64, 6, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 201, 224, 9, 0, 156, 5, 0, 64, 156, 5, 0, 66, 201, 192, 12, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 201, 64, 22, 0, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 134, 5, 2, 57, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 201, 96, 32, 0, 5, 5, 0, 50, 5, 5, 0, 52, 201, 160, 35, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 201, 128, 37, 0, 5, 5, 0, 64, 5, 5, 0, 66, 201, 0, 39, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 201, 32, 44, 0, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 2, 57, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 145, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 130, 36, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 145, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 145, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 145, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 145, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 194, 36, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 145, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 2, 37, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 145, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 145, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 145, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 145, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 145, 156, 46, 196, 145, 5, 46, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 145, 156, 54, 196, 145, 5, 54, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 145, 156, 56, 196, 145, 5, 56, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 145, 156, 60, 196, 145, 5, 60, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 69, 5, 5, 3, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 145, 156, 78, 196, 145, 5, 78, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 145, 160, 86, 196, 145, 5, 86, 196, 145, 156, 90, 196, 145, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 71, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 51, 134, 5, 2, 53, 5, 5, 2, 53, 192, 0, 0, 0, 134, 5, 3, 55, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 85, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 71, 192, 0, 0, 0, 5, 5, 3, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 55, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 85, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 145, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 24, 0, 197, 130, 24, 0, 197, 194, 24, 0, 197, 2, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 145, 156, 92, 196, 145, 5, 92, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 25, 0, 197, 163, 25, 0, 197, 3, 26, 0, 197, 99, 26, 0, 197, 195, 26, 0, 197, 35, 27, 0, 197, 131, 27, 0, 197, 227, 27, 0, 197, 67, 28, 0, 197, 163, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 29, 0, 197, 99, 29, 0, 197, 195, 29, 0, 197, 35, 30, 0, 197, 131, 30, 0, 197, 227, 30, 0, 197, 67, 31, 0, 197, 163, 31, 0, 197, 3, 32, 0, 197, 99, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 32, 0, 197, 35, 33, 0, 197, 131, 33, 0, 197, 227, 33, 0, 197, 67, 34, 0, 197, 163, 34, 0, 197, 3, 35, 0, 197, 99, 35, 0, 197, 195, 35, 0, 197, 35, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 224, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 135, 5, 2, 49, 5, 136, 0, 0, 70, 5, 2, 49, 5, 136, 0, 0, 135, 5, 2, 49, 5, 156, 0, 0, 70, 5, 2, 49, 5, 156, 0, 0, 135, 5, 2, 49, 5, 144, 0, 0, 70, 5, 2, 49, 5, 144, 0, 0, 135, 5, 2, 49, 5, 145, 0, 0, 70, 5, 2, 49, 5, 145, 0, 0, 135, 5, 2, 49, 5, 196, 0, 0, 70, 5, 2, 49, 5, 196, 0, 0, 135, 5, 2, 49, 5, 210, 0, 0, 70, 5, 2, 49, 5, 210, 0, 0, 135, 5, 2, 55, 5, 156, 0, 0, 70, 5, 2, 55, 5, 156, 0, 0, 135, 5, 2, 55, 5, 196, 0, 0, 70, 5, 2, 55, 5, 196, 0, 0, 135, 5, 2, 55, 5, 210, 0, 0, 70, 5, 2, 55, 5, 210, 0, 0, 135, 5, 2, 63, 5, 136, 0, 0, 70, 5, 2, 63, 5, 136, 0, 0, 135, 5, 2, 63, 5, 156, 0, 0, 70, 5, 2, 63, 5, 156, 0, 0, 135, 5, 2, 69, 5, 136, 0, 0, 70, 5, 2, 69, 5, 136, 0, 0, 70, 5, 2, 69, 5, 150, 0, 0, 135, 5, 2, 69, 5, 145, 0, 0, 70, 5, 2, 69, 5, 145, 0, 0, 135, 5, 2, 69, 5, 150, 0, 0, 135, 5, 2, 69, 5, 164, 0, 0, 70, 5, 2, 69, 5, 164, 0, 0, 135, 5, 2, 69, 5, 156, 0, 0, 70, 5, 2, 69, 5, 156, 0, 0, 70, 5, 2, 69, 5, 146, 0, 0, 135, 5, 2, 69, 5, 138, 0, 0, 70, 5, 2, 69, 5, 138, 0, 0, 135, 5, 2, 69, 5, 196, 0, 0, 70, 5, 2, 69, 5, 196, 0, 0, 135, 5, 2, 69, 5, 182, 0, 0, 70, 5, 2, 69, 5, 182, 0, 0, 135, 5, 2, 69, 5, 154, 0, 0, 70, 5, 2, 69, 5, 154, 0, 0, 135, 5, 2, 81, 5, 136, 0, 0, 70, 5, 2, 81, 5, 136, 0, 0, 135, 5, 2, 81, 5, 145, 0, 0, 70, 5, 2, 81, 5, 145, 0, 0, 135, 5, 2, 81, 5, 160, 0, 0, 70, 5, 2, 81, 5, 160, 0, 0, 135, 5, 2, 81, 5, 144, 0, 0, 70, 5, 2, 81, 5, 144, 0, 0, 135, 5, 2, 81, 5, 202, 0, 0, 70, 5, 2, 81, 5, 202, 0, 0, 135, 5, 2, 81, 5, 156, 0, 0, 70, 5, 2, 81, 5, 156, 0, 0, 135, 5, 2, 81, 5, 196, 0, 0, 70, 5, 2, 81, 5, 196, 0, 0, 135, 5, 2, 81, 5, 136, 0, 0, 5, 156, 0, 0, 70, 5, 2, 81, 5, 136, 0, 0, 5, 156, 0, 0, 135, 5, 2, 81, 5, 144, 0, 0, 5, 156, 0, 0, 70, 5, 2, 81, 5, 144, 0, 0, 5, 156, 0, 0, 135, 5, 2, 81, 5, 196, 0, 0, 5, 156, 0, 0, 70, 5, 2, 81, 5, 196, 0, 0, 5, 156, 0, 0, 5, 5, 2, 49, 5, 136, 0, 0, 5, 5, 2, 49, 5, 156, 0, 0, 5, 5, 2, 49, 5, 144, 0, 0, 5, 5, 2, 49, 5, 145, 0, 0, 5, 5, 2, 49, 5, 196, 0, 0, 5, 5, 2, 49, 5, 210, 0, 0, 5, 5, 2, 55, 5, 156, 0, 0, 5, 5, 2, 55, 5, 196, 0, 0, 5, 5, 2, 55, 5, 210, 0, 0, 5, 5, 2, 63, 5, 136, 0, 0, 5, 5, 2, 63, 5, 156, 0, 0, 5, 5, 2, 69, 5, 136, 0, 0, 5, 5, 2, 69, 5, 150, 0, 0, 5, 5, 2, 69, 5, 145, 0, 0, 5, 5, 2, 69, 5, 164, 0, 0, 5, 5, 2, 69, 5, 156, 0, 0, 5, 5, 2, 69, 5, 146, 0, 0, 5, 5, 2, 69, 5, 138, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 182, 0, 0, 5, 5, 2, 69, 5, 154, 0, 0, 5, 5, 2, 81, 5, 136, 0, 0, 5, 5, 2, 81, 5, 145, 0, 0, 5, 5, 2, 81, 5, 160, 0, 0, 5, 5, 2, 81, 5, 144, 0, 0, 5, 5, 2, 81, 5, 202, 0, 0, 5, 5, 2, 81, 5, 156, 0, 0, 5, 5, 2, 81, 5, 196, 0, 0, 5, 5, 2, 81, 5, 136, 0, 0, 5, 156, 0, 0, 5, 5, 2, 81, 5, 144, 0, 0, 5, 156, 0, 0, 5, 5, 2, 81, 5, 196, 0, 0, 5, 156, 0, 0, 134, 5, 2, 57, 5, 156, 0, 0, 5, 5, 2, 57, 5, 156, 0, 0, 134, 5, 2, 57, 5, 150, 0, 0, 5, 5, 2, 57, 5, 150, 0, 0, 156, 5, 0, 42, 5, 145, 0, 0, 5, 136, 0, 0, 5, 5, 0, 42, 5, 145, 0, 0, 5, 136, 0, 0, 156, 5, 0, 42, 5, 145, 0, 0, 5, 138, 0, 0, 5, 5, 0, 42, 5, 145, 0, 0, 5, 138, 0, 0, 156, 5, 0, 42, 5, 145, 0, 0, 5, 182, 0, 0, 5, 5, 0, 42, 5, 145, 0, 0, 5, 182, 0, 0, 156, 5, 0, 42, 5, 145, 0, 0, 5, 154, 0, 0, 5, 5, 0, 42, 5, 145, 0, 0, 5, 154, 0, 0, 156, 5, 0, 42, 5, 196, 0, 0, 5, 145, 0, 0, 5, 5, 0, 42, 5, 196, 0, 0, 5, 145, 0, 0, 156, 5, 0, 50, 5, 145, 0, 0, 5, 136, 0, 0, 5, 5, 0, 50, 5, 145, 0, 0, 5, 136, 0, 0, 156, 5, 0, 50, 5, 145, 0, 0, 5, 138, 0, 0, 5, 5, 0, 50, 5, 145, 0, 0, 5, 138, 0, 0, 156, 5, 0, 50, 5, 145, 0, 0, 5, 182, 0, 0, 5, 5, 0, 50, 5, 145, 0, 0, 5, 182, 0, 0, 156, 5, 0, 50, 5, 145, 0, 0, 5, 154, 0, 0, 5, 5, 0, 50, 5, 145, 0, 0, 5, 154, 0, 0, 156, 5, 0, 50, 5, 196, 0, 0, 5, 145, 0, 0, 5, 5, 0, 50, 5, 196, 0, 0, 5, 145, 0, 0, 156, 5, 0, 70, 5, 145, 0, 0, 5, 136, 0, 0, 5, 5, 0, 70, 5, 145, 0, 0, 5, 136, 0, 0, 156, 5, 0, 70, 5, 145, 0, 0, 5, 138, 0, 0, 5, 5, 0, 70, 5, 145, 0, 0, 5, 138, 0, 0, 156, 5, 0, 70, 5, 145, 0, 0, 5, 182, 0, 0, 5, 5, 0, 70, 5, 145, 0, 0, 5, 182, 0, 0, 156, 5, 0, 70, 5, 145, 0, 0, 5, 154, 0, 0, 5, 5, 0, 70, 5, 145, 0, 0, 5, 154, 0, 0, 156, 5, 0, 70, 5, 196, 0, 0, 5, 145, 0, 0, 5, 5, 0, 70, 5, 196, 0, 0, 5, 145, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 48, 156, 5, 13, 0, 126, 1, 23, 0, 146, 30, 12, 0, 146, 30, 2, 192, 197, 34, 147, 30, 2, 192, 197, 98, 148, 30, 2, 192, 197, 162, 149, 30, 2, 192, 197, 226, 126, 1, 1, 192, 197, 98, 144, 30, 1, 192, 197, 162, 145, 30, 1, 192, 197, 226, 122, 1, 12, 0, 122, 1, 0, 192, 197, 98, 123, 1, 0, 192, 197, 162, 124, 1, 0, 192, 197, 226, 125, 1, 1, 192, 197, 34, 90, 0, 2, 241, 135, 5, 122, 0, 2, 241, 70, 5, 121, 1, 197, 162, 0, 54, 156, 5, 7, 0, 4, 30, 12, 0, 4, 30, 3, 192, 197, 162, 5, 30, 3, 192, 197, 226, 6, 30, 4, 192, 197, 34, 7, 30, 4, 192, 197, 98, 66, 0, 2, 247, 135, 5, 98, 0, 2, 247, 70, 5, 2, 30, 3, 192, 197, 34, 3, 30, 3, 192, 197, 98, 0, 62, 156, 5, 5, 0, 85, 30, 9, 0, 85, 30, 4, 192, 197, 226, 86, 30, 5, 192, 197, 34, 87, 30, 5, 192, 197, 98, 80, 0, 2, 255, 135, 5, 112, 0, 2, 255, 70, 5, 84, 30, 4, 192, 197, 162, 0, 68, 156, 5, 20, 0, 142, 30, 37, 0, 244, 30, 20, 0, 247, 30, 9, 0, 247, 30, 9, 192, 197, 162, 248, 30, 9, 192, 197, 226, 249, 30, 10, 192, 197, 34, 244, 30, 8, 192, 197, 226, 245, 30, 9, 192, 197, 34, 246, 30, 9, 192, 197, 98, 142, 30, 7, 192, 197, 162, 143, 30, 7, 192, 197, 226, 153, 30, 8, 192, 197, 34, 242, 30, 8, 192, 197, 98, 243, 30, 8, 192, 197, 162, 118, 1, 15, 0, 118, 1, 6, 192, 197, 98, 119, 1, 6, 192, 197, 162, 120, 1, 6, 192, 197, 226, 50, 2, 7, 192, 197, 34, 51, 2, 7, 192, 197, 98, 89, 0, 255, 255, 2, 69, 135, 5, 121, 0, 255, 255, 2, 69, 70, 5, 221, 0, 5, 192, 197, 162, 253, 0, 5, 192, 197, 226, 255, 0, 6, 192, 197, 34, 0, 80, 156, 5, 21, 0, 25, 2, 37, 0, 100, 30, 20, 0, 103, 30, 9, 0, 103, 30, 15, 192, 197, 3, 104, 30, 15, 192, 197, 99, 105, 30, 15, 192, 197, 195, 100, 30, 13, 192, 197, 227, 101, 30, 14, 192, 197, 67, 102, 30, 14, 192, 197, 163, 25, 2, 12, 192, 197, 162, 96, 30, 12, 192, 197, 226, 97, 30, 13, 192, 197, 34, 98, 30, 13, 192, 197, 98, 99, 30, 13, 192, 197, 162, 93, 1, 20, 0, 96, 1, 9, 0, 96, 1, 11, 192, 197, 226, 97, 1, 12, 192, 197, 34, 24, 2, 12, 192, 197, 98, 93, 1, 11, 192, 197, 34, 94, 1, 11, 192, 197, 98, 95, 1, 11, 192, 197, 162, 83, 0, 255, 255, 2, 81, 135, 5, 115, 0, 255, 255, 2, 81, 70, 5, 90, 1, 10, 192, 197, 98, 91, 1, 10, 192, 197, 162, 92, 1, 10, 192, 197, 226, 0, 48, 5, 5, 6, 0, 126, 1, 12, 0, 126, 1, 16, 192, 197, 162, 145, 30, 16, 192, 197, 226, 147, 30, 17, 192, 197, 34, 149, 30, 17, 192, 197, 98, 122, 0, 2, 241, 5, 5, 122, 1, 16, 192, 197, 34, 124, 1, 16, 192, 197, 98, 0, 54, 5, 5, 3, 0, 98, 0, 2, 247, 5, 5, 3, 30, 17, 192, 197, 162, 5, 30, 17, 192, 197, 226, 7, 30, 18, 192, 197, 34, 0, 62, 5, 5, 2, 0, 112, 0, 2, 255, 5, 5, 85, 30, 18, 192, 197, 98, 87, 30, 18, 192, 197, 162, 0, 68, 5, 5, 10, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 20, 192, 197, 162, 247, 30, 20, 192, 197, 226, 249, 30, 21, 192, 197, 34, 143, 30, 19, 192, 197, 226, 153, 30, 20, 192, 197, 34, 243, 30, 20, 192, 197, 98, 121, 0, 255, 255, 2, 69, 5, 5, 253, 0, 18, 192, 197, 226, 255, 0, 19, 192, 197, 34, 119, 1, 19, 192, 197, 98, 51, 2, 19, 192, 197, 162, 0, 80, 5, 5, 10, 0, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 23, 192, 197, 35, 103, 30, 23, 192, 197, 131, 105, 30, 23, 192, 197, 227, 25, 2, 22, 192, 197, 98, 97, 30, 22, 192, 197, 162, 99, 30, 22, 192, 197, 226, 115, 0, 255, 255, 2, 81, 5, 5, 91, 1, 21, 192, 197, 98, 93, 1, 21, 192, 197, 162, 95, 1, 21, 192, 197, 226, 97, 1, 22, 192, 197, 34, 42, 0, 66, 0, 67, 0, 80, 0, 81, 0, 83, 0, 84, 0, 89, 0, 91, 0, 98, 0, 99, 0, 112, 0, 113, 0, 115, 0, 116, 0, 121, 0, 123, 0, 221, 0, 222, 0, 253, 0, 254, 0, 255, 0, 0, 1, 90, 1, 98, 1, 118, 1, 127, 1, 24, 2, 26, 2, 50, 2, 52, 2, 2, 30, 8, 30, 84, 30, 88, 30, 96, 30, 106, 30, 142, 30, 150, 30, 153, 30, 154, 30, 242, 30, 250, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 26, 4, 186, 76, 185, 80, 44, 4, 185, 92, 186, 100, 186, 108, 50, 4, 185, 120, 185, 124, 56, 4, 185, 140, 185, 144, 185, 148, 185, 156, 186, 160, 74, 4, 185, 176, 185, 180, 185, 184, 185, 96, 185, 188, 185, 192, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 96, 4, 168, 76, 168, 80, 106, 4, 168, 92, 168, 100, 168, 108, 110, 4, 168, 120, 168, 124, 114, 4, 168, 140, 168, 144, 168, 148, 168, 156, 168, 160, 124, 4, 168, 176, 168, 180, 168, 184, 168, 96, 168, 188, 168, 192, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 140, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 25, 58, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 77, 154, 77, 26, 78, 90, 78, 186, 101, 154, 101, 26, 102, 90, 102, 217, 68, 154, 130, 185, 141, 153, 141, 25, 142, 153, 142, 89, 142, 136, 14, 217, 142, 185, 177, 153, 177, 25, 178, 89, 178, 153, 189, 185, 196, 2, 8, 168, 57, 136, 57, 8, 58, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 77, 136, 77, 8, 78, 72, 78, 168, 101, 136, 101, 8, 102, 72, 102, 200, 68, 136, 130, 168, 141, 136, 141, 8, 142, 136, 142, 72, 142, 128, 14, 200, 142, 168, 177, 136, 177, 8, 178, 72, 178, 136, 189, 168, 196, 72, 190, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 25, 66, 8, 66, 185, 66, 168, 66, 249, 65, 232, 65, 249, 69, 232, 69, 89, 71, 72, 71, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 250, 77, 232, 77, 25, 86, 8, 86, 217, 85, 200, 85, 185, 86, 168, 86, 249, 86, 232, 86, 25, 94, 8, 94, 89, 95, 72, 95, 154, 102, 136, 102, 58, 103, 40, 103, 218, 101, 200, 101, 26, 103, 8, 103, 186, 102, 168, 104, 6, 8, 8, 8, 26, 110, 8, 110, 249, 114, 232, 114, 168, 152, 153, 121, 136, 121, 249, 122, 232, 122, 249, 121, 232, 121, 121, 123, 104, 123, 89, 123, 72, 123, 154, 129, 136, 129, 250, 130, 232, 130, 250, 129, 232, 129, 10, 8, 185, 136, 168, 136, 57, 143, 40, 143, 217, 141, 200, 141, 121, 142, 104, 142, 12, 8, 14, 8, 153, 157, 136, 157, 249, 158, 232, 158, 249, 157, 232, 157, 154, 161, 136, 161, 26, 162, 8, 162, 250, 162, 232, 162, 250, 161, 232, 161, 249, 166, 232, 166, 249, 165, 232, 165, 185, 172, 168, 172, 153, 178, 136, 178, 57, 179, 40, 179, 217, 177, 200, 177, 57, 178, 40, 178, 121, 178, 104, 178, 25, 179, 8, 179, 25, 186, 8, 186, 25, 190, 8, 190, 89, 190, 153, 193, 136, 193, 185, 194, 168, 194, 249, 193, 232, 193, 232, 160, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 160, 169, 160, 200, 56, 169, 76, 187, 100, 187, 108, 169, 100, 169, 108, 169, 200, 169, 128, 217, 140, 187, 76, 200, 140, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 68, 90, 4, 186, 72, 122, 4, 177, 72, 121, 5, 154, 73, 122, 5, 145, 73, 123, 5, 186, 74, 124, 5, 177, 74, 125, 5, 250, 73, 126, 5, 241, 73, 255, 5, 185, 84, 66, 4, 186, 88, 98, 4, 177, 88, 255, 5, 185, 112, 80, 4, 186, 116, 112, 4, 177, 116, 255, 5, 186, 128, 89, 4, 186, 132, 121, 4, 177, 132, 221, 4, 154, 133, 253, 4, 145, 133, 255, 4, 81, 134, 118, 5, 26, 134, 119, 5, 17, 134, 120, 5, 90, 134, 255, 5, 185, 164, 83, 4, 186, 168, 115, 4, 177, 168, 90, 5, 154, 169, 91, 5, 145, 169, 92, 5, 26, 170, 93, 5, 17, 170, 94, 5, 250, 170, 95, 5, 241, 170, 96, 5, 250, 169, 97, 5, 241, 169, 255, 5, 168, 68, 122, 4, 168, 72, 122, 5, 136, 73, 124, 5, 168, 74, 126, 5, 232, 73, 255, 5, 168, 84, 98, 4, 168, 88, 255, 5, 168, 112, 112, 4, 168, 116, 255, 5, 168, 128, 121, 4, 168, 132, 253, 4, 136, 133, 255, 4, 72, 134, 119, 5, 8, 134, 255, 5, 168, 164, 115, 4, 168, 168, 91, 5, 136, 169, 93, 5, 8, 170, 95, 5, 232, 170, 97, 5, 232, 169, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 95, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 69, 11, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 74, 11, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 132, 0, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 136, 0, 0, 0, 136, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 41, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 71, 0, 114, 0, 101, 0, 107, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 68, 1, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 8, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 32, 0, 0, 0, 44, 0, 0, 0, 44, 1, 0, 0, 14, 0, 0, 0, 0, 0, 0, 96, 199, 255, 96, 96, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 0, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 3, 0, 69, 0, 60, 0, 32, 0, 43, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 125, 0, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 130, 0, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 149, 15, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 153, 15, 0, 0, 153, 15, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 71, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 92, 220, 38, 0, 65, 0, 60, 0, 42, 0, 39, 0, 32, 0, 39, 0, 45, 0, 39, 0, 47, 0, 39, 0, 60, 0, 42, 0, 48, 0, 45, 0, 39, 0, 64, 0, 39, 0, 60, 0, 42, 0, 65, 0, 66, 0, 67, 0, 68, 0, 69, 0, 70, 0, 71, 0, 72, 0, 73, 0, 74, 0, 75, 0, 76, 0, 77, 0, 78, 0, 79, 0, 80, 0, 81, 0, 82, 0, 83, 0, 84, 0, 85, 0, 86, 0, 87, 0, 88, 0, 89, 0, 90, 0, 60, 0, 42, 0, 39, 0, 91, 0, 39, 0, 45, 0, 39, 0, 96, 0, 39, 0, 60, 0, 42, 0, 97, 0, 98, 0, 99, 0, 100, 0, 101, 0, 102, 0, 103, 0, 104, 0, 105, 0, 106, 0, 107, 0, 108, 0, 109, 0, 110, 0, 111, 0, 112, 0, 113, 0, 114, 0, 115, 0, 116, 0, 117, 0, 118, 0, 119, 0, 120, 0, 121, 0, 122, 0, 60, 0, 42, 0, 39, 0, 123, 0, 39, 0, 45, 0, 39, 0, 127, 0, 39, 0, 0, 0, 24, 61, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 84, 34, 0, 0, 84, 34, 0, 0, 252, 51, 0, 0, 252, 51, 0, 0, 228, 60, 0, 0, 228, 60, 0, 0, 252, 60, 0, 0, 252, 60, 0, 0, 252, 60, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 24, 1, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 125, 0, 133, 0, 141, 0, 149, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 154, 0, 162, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 186, 0, 194, 0, 202, 0, 210, 0, 218, 0, 226, 0, 234, 0, 242, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 249, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 255, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 7, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 178, 0, 48, 0, 48, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 1, 244, 1, 52, 2, 192, 0, 192, 0, 192, 0, 192, 0, 104, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 2, 43, 5, 5, 3, 43, 5, 5, 4, 43, 193, 2, 5, 43, 193, 3, 5, 43, 193, 4, 5, 43, 193, 5, 5, 43, 193, 6, 5, 43, 193, 7, 5, 43, 193, 8, 5, 43, 193, 9, 5, 43, 193, 10, 5, 43, 193, 11, 5, 43, 193, 12, 5, 43, 193, 13, 5, 43, 193, 14, 5, 43, 202, 0, 70, 0, 202, 33, 70, 0, 202, 66, 70, 0, 202, 99, 70, 0, 202, 132, 70, 0, 202, 165, 70, 0, 202, 198, 70, 0, 202, 231, 70, 0, 202, 8, 71, 0, 202, 41, 71, 0, 193, 25, 5, 43, 193, 26, 5, 43, 193, 27, 5, 43, 193, 28, 5, 43, 193, 29, 5, 43, 193, 30, 5, 43, 193, 31, 5, 43, 198, 1, 0, 0, 198, 33, 0, 0, 198, 65, 0, 0, 198, 97, 0, 0, 198, 129, 0, 0, 198, 161, 0, 0, 198, 193, 0, 0, 198, 225, 0, 0, 198, 1, 1, 0, 198, 33, 1, 0, 198, 65, 1, 0, 198, 97, 1, 0, 198, 129, 1, 0, 198, 161, 1, 0, 198, 193, 1, 0, 198, 225, 1, 0, 198, 1, 2, 0, 198, 33, 2, 0, 198, 65, 2, 0, 198, 97, 2, 0, 198, 129, 2, 0, 198, 161, 2, 0, 198, 193, 2, 0, 198, 225, 2, 0, 198, 1, 3, 0, 198, 33, 3, 0, 193, 58, 5, 43, 193, 59, 5, 43, 193, 60, 5, 43, 193, 61, 5, 43, 193, 62, 5, 43, 193, 63, 5, 43, 193, 64, 5, 43, 193, 65, 5, 43, 193, 66, 5, 43, 193, 67, 5, 43, 193, 68, 5, 43, 193, 69, 5, 43, 193, 70, 5, 43, 193, 71, 5, 43, 193, 72, 5, 43, 193, 73, 5, 43, 193, 74, 5, 43, 193, 75, 5, 43, 193, 76, 5, 43, 193, 77, 5, 43, 193, 78, 5, 43, 193, 79, 5, 43, 193, 80, 5, 43, 193, 81, 5, 43, 193, 82, 5, 43, 193, 83, 5, 43, 193, 84, 5, 43, 193, 85, 5, 43, 193, 86, 5, 43, 193, 87, 5, 43, 193, 88, 5, 43, 193, 89, 5, 43, 193, 90, 5, 43, 193, 91, 5, 43, 193, 92, 5, 43, 193, 93, 5, 43, 193, 94, 5, 43, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 3, 0, 198, 130, 3, 0, 198, 194, 3, 0, 198, 2, 4, 0, 198, 66, 4, 0, 198, 130, 4, 0, 197, 66, 69, 0, 198, 194, 4, 0, 198, 2, 5, 0, 198, 66, 5, 0, 198, 130, 5, 0, 198, 194, 5, 0, 198, 2, 6, 0, 198, 66, 6, 0, 198, 130, 6, 0, 198, 194, 6, 0, 160, 112, 0, 48, 198, 2, 7, 0, 198, 66, 7, 0, 198, 130, 7, 0, 198, 194, 7, 0, 198, 2, 8, 0, 198, 66, 8, 0, 5, 5, 127, 12, 196, 158, 156, 70, 198, 130, 8, 0, 198, 194, 8, 0, 198, 2, 9, 0, 198, 66, 9, 0, 198, 130, 9, 0, 160, 5, 38, 93, 197, 130, 69, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 162, 0, 0, 197, 226, 0, 0, 197, 34, 1, 0, 197, 98, 1, 0, 197, 194, 69, 0, 197, 162, 1, 0, 197, 226, 1, 0, 197, 34, 2, 0, 197, 98, 2, 0, 197, 162, 2, 0, 197, 226, 2, 0, 197, 34, 3, 0, 197, 98, 3, 0, 197, 162, 3, 0, 5, 112, 0, 48, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 226, 4, 0, 197, 34, 5, 0, 5, 5, 125, 12, 196, 158, 5, 70, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 5, 5, 38, 93, 197, 162, 6, 0, 198, 194, 9, 0, 197, 226, 6, 0, 198, 2, 10, 0, 197, 34, 7, 0, 198, 66, 10, 0, 197, 98, 7, 0, 198, 130, 10, 0, 197, 162, 7, 0, 198, 194, 10, 0, 197, 226, 7, 0, 198, 2, 11, 0, 197, 34, 8, 0, 198, 66, 11, 0, 197, 98, 8, 0, 198, 130, 11, 0, 197, 162, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 11, 0, 197, 226, 8, 0, 198, 2, 12, 0, 197, 34, 9, 0, 198, 66, 12, 0, 197, 98, 9, 0, 198, 130, 12, 0, 197, 162, 9, 0, 198, 194, 12, 0, 197, 226, 9, 0, 198, 2, 13, 0, 197, 34, 10, 0, 198, 66, 13, 0, 197, 98, 10, 0, 198, 130, 13, 0, 197, 162, 10, 0, 198, 194, 13, 0, 197, 226, 10, 0, 198, 2, 14, 0, 197, 34, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 14, 0, 197, 98, 11, 0, 198, 130, 14, 0, 197, 162, 11, 0, 198, 194, 14, 0, 197, 226, 11, 0, 198, 2, 15, 0, 197, 34, 12, 0, 198, 66, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 15, 0, 197, 98, 12, 0, 198, 194, 15, 0, 197, 162, 12, 0, 192, 0, 0, 0, 198, 2, 16, 0, 197, 226, 12, 0, 198, 66, 16, 0, 197, 34, 13, 0, 198, 130, 16, 0, 197, 98, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 16, 0, 197, 162, 13, 0, 198, 2, 17, 0, 197, 226, 13, 0, 198, 66, 17, 0, 197, 34, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 17, 0, 197, 98, 14, 0, 198, 194, 17, 0, 197, 162, 14, 0, 198, 2, 18, 0, 197, 226, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 18, 0, 197, 34, 15, 0, 198, 130, 18, 0, 197, 98, 15, 0, 198, 194, 18, 0, 197, 162, 15, 0, 198, 2, 19, 0, 197, 226, 15, 0, 198, 66, 19, 0, 197, 34, 16, 0, 198, 130, 19, 0, 197, 98, 16, 0, 198, 194, 19, 0, 197, 162, 16, 0, 198, 2, 20, 0, 197, 226, 16, 0, 198, 66, 20, 0, 197, 34, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 20, 0, 197, 98, 17, 0, 198, 194, 20, 0, 197, 162, 17, 0, 198, 2, 21, 0, 197, 226, 17, 0, 198, 66, 21, 0, 197, 34, 18, 0, 198, 130, 21, 0, 197, 98, 18, 0, 198, 194, 21, 0, 197, 162, 18, 0, 198, 2, 22, 0, 197, 226, 18, 0, 198, 66, 22, 0, 197, 34, 19, 0, 198, 130, 22, 0, 198, 194, 22, 0, 197, 98, 19, 0, 198, 2, 23, 0, 197, 162, 19, 0, 198, 66, 23, 0, 197, 226, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 23, 0, 197, 34, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 23, 0, 197, 98, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 24, 0, 197, 162, 20, 0, 198, 66, 24, 0, 197, 226, 20, 0, 198, 130, 24, 0, 197, 34, 21, 0, 198, 194, 24, 0, 197, 98, 21, 0, 198, 3, 25, 0, 197, 163, 21, 0, 198, 99, 25, 0, 197, 3, 22, 0, 198, 195, 25, 0, 197, 99, 22, 0, 198, 35, 26, 0, 197, 195, 22, 0, 192, 0, 0, 0, 198, 131, 26, 0, 197, 35, 23, 0, 198, 227, 26, 0, 197, 131, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 27, 0, 197, 226, 23, 0, 198, 130, 27, 0, 197, 34, 24, 0, 198, 194, 27, 0, 197, 98, 24, 0, 198, 3, 28, 0, 197, 163, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 28, 0, 197, 66, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 28, 0, 197, 130, 25, 0, 198, 227, 28, 0, 197, 195, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 29, 0, 197, 34, 26, 0, 198, 130, 29, 0, 197, 98, 26, 0, 198, 194, 29, 0, 197, 162, 26, 0, 198, 2, 30, 0, 197, 226, 26, 0, 198, 66, 30, 0, 197, 34, 27, 0, 198, 130, 30, 0, 197, 98, 27, 0, 198, 194, 30, 0, 197, 162, 27, 0, 198, 2, 31, 0, 197, 226, 27, 0, 198, 66, 31, 0, 197, 34, 28, 0, 198, 130, 31, 0, 197, 98, 28, 0, 198, 194, 31, 0, 197, 162, 28, 0, 198, 2, 32, 0, 197, 226, 28, 0, 198, 66, 32, 0, 197, 34, 29, 0, 198, 130, 32, 0, 197, 98, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 32, 0, 197, 162, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 0, 0, 201, 198, 0, 0, 198, 2, 33, 0, 197, 226, 29, 0, 198, 67, 33, 0, 197, 35, 30, 0, 198, 163, 33, 0, 197, 131, 30, 0, 198, 2, 34, 0, 197, 226, 30, 0, 198, 67, 34, 0, 197, 35, 31, 0, 198, 162, 34, 0, 197, 130, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 26, 5, 43, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 34, 0, 197, 194, 31, 0, 198, 34, 35, 0, 197, 2, 32, 0, 198, 98, 35, 0, 197, 66, 32, 0, 198, 162, 35, 0, 197, 130, 32, 0, 198, 227, 35, 0, 197, 195, 32, 0, 198, 66, 36, 0, 197, 34, 33, 0, 198, 130, 36, 0, 197, 98, 33, 0, 198, 194, 36, 0, 197, 162, 33, 0, 198, 2, 37, 0, 197, 226, 33, 0, 198, 66, 37, 0, 197, 34, 34, 0, 198, 131, 37, 0, 197, 99, 34, 0, 198, 227, 37, 0, 197, 195, 34, 0, 198, 66, 38, 0, 197, 34, 35, 0, 198, 130, 38, 0, 197, 98, 35, 0, 198, 195, 38, 0, 197, 163, 35, 0, 198, 34, 39, 0, 197, 2, 36, 0, 198, 98, 39, 0, 197, 66, 36, 0, 198, 162, 39, 0, 197, 130, 36, 0, 198, 226, 39, 0, 197, 194, 36, 0, 198, 34, 40, 0, 197, 2, 37, 0, 198, 98, 40, 0, 197, 66, 37, 0, 198, 162, 40, 0, 197, 130, 37, 0, 198, 226, 40, 0, 197, 194, 37, 0, 198, 35, 41, 0, 197, 3, 38, 0, 198, 130, 41, 0, 197, 98, 38, 0, 198, 194, 41, 0, 197, 162, 38, 0, 198, 2, 42, 0, 197, 226, 38, 0, 198, 66, 42, 0, 197, 34, 39, 0, 198, 131, 42, 0, 197, 99, 39, 0, 198, 226, 42, 0, 197, 194, 39, 0, 198, 34, 43, 0, 197, 2, 40, 0, 198, 98, 43, 0, 197, 66, 40, 0, 198, 162, 43, 0, 197, 130, 40, 0, 198, 226, 43, 0, 197, 194, 40, 0, 198, 34, 44, 0, 197, 2, 41, 0, 198, 98, 44, 0, 197, 66, 41, 0, 198, 162, 44, 0, 197, 130, 41, 0, 198, 226, 44, 0, 197, 194, 41, 0, 198, 35, 45, 0, 197, 3, 42, 0, 198, 131, 45, 0, 197, 99, 42, 0, 198, 227, 45, 0, 197, 195, 42, 0, 198, 67, 46, 0, 197, 35, 43, 0, 198, 162, 46, 0, 197, 130, 43, 0, 198, 226, 46, 0, 197, 194, 43, 0, 198, 34, 47, 0, 197, 2, 44, 0, 198, 98, 47, 0, 197, 66, 44, 0, 198, 163, 47, 0, 197, 131, 44, 0, 198, 2, 48, 0, 197, 226, 44, 0, 198, 66, 48, 0, 197, 34, 45, 0, 198, 130, 48, 0, 197, 98, 45, 0, 198, 195, 48, 0, 197, 163, 45, 0, 198, 35, 49, 0, 197, 3, 46, 0, 198, 131, 49, 0, 197, 99, 46, 0, 198, 226, 49, 0, 197, 194, 46, 0, 198, 34, 50, 0, 197, 2, 47, 0, 198, 98, 50, 0, 197, 66, 47, 0, 198, 162, 50, 0, 197, 130, 47, 0, 198, 226, 50, 0, 197, 194, 47, 0, 198, 34, 51, 0, 197, 2, 48, 0, 198, 98, 51, 0, 197, 66, 48, 0, 198, 163, 51, 0, 197, 131, 48, 0, 198, 3, 52, 0, 197, 227, 48, 0, 198, 98, 52, 0, 197, 66, 49, 0, 198, 162, 52, 0, 197, 130, 49, 0, 198, 226, 52, 0, 197, 194, 49, 0, 198, 34, 53, 0, 197, 2, 50, 0, 198, 98, 53, 0, 197, 66, 50, 0, 198, 162, 53, 0, 197, 130, 50, 0, 198, 226, 53, 0, 197, 194, 50, 0, 198, 34, 54, 0, 197, 2, 51, 0, 198, 98, 54, 0, 197, 66, 51, 0, 198, 162, 54, 0, 197, 130, 51, 0, 198, 226, 54, 0, 197, 194, 51, 0, 198, 34, 55, 0, 197, 2, 52, 0, 198, 98, 55, 0, 197, 66, 52, 0, 197, 130, 52, 0, 197, 194, 52, 0, 197, 2, 53, 0, 197, 66, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 55, 0, 197, 130, 53, 0, 198, 226, 55, 0, 197, 194, 53, 0, 198, 35, 56, 0, 197, 3, 54, 0, 198, 131, 56, 0, 197, 99, 54, 0, 198, 227, 56, 0, 197, 195, 54, 0, 198, 67, 57, 0, 197, 35, 55, 0, 198, 163, 57, 0, 197, 131, 55, 0, 198, 3, 58, 0, 197, 227, 55, 0, 198, 99, 58, 0, 197, 67, 56, 0, 198, 195, 58, 0, 197, 163, 56, 0, 198, 35, 59, 0, 197, 3, 57, 0, 198, 131, 59, 0, 197, 99, 57, 0, 198, 226, 59, 0, 197, 194, 57, 0, 198, 34, 60, 0, 197, 2, 58, 0, 198, 98, 60, 0, 197, 66, 58, 0, 198, 163, 60, 0, 197, 131, 58, 0, 198, 3, 61, 0, 197, 227, 58, 0, 198, 99, 61, 0, 197, 67, 59, 0, 198, 195, 61, 0, 197, 163, 59, 0, 198, 35, 62, 0, 197, 3, 60, 0, 198, 130, 62, 0, 197, 98, 60, 0, 198, 194, 62, 0, 197, 162, 60, 0, 198, 2, 63, 0, 197, 226, 60, 0, 198, 66, 63, 0, 197, 34, 61, 0, 198, 131, 63, 0, 197, 99, 61, 0, 198, 227, 63, 0, 197, 195, 61, 0, 198, 67, 64, 0, 197, 35, 62, 0, 198, 163, 64, 0, 197, 131, 62, 0, 198, 3, 65, 0, 197, 227, 62, 0, 198, 99, 65, 0, 197, 67, 63, 0, 198, 195, 65, 0, 197, 163, 63, 0, 198, 35, 66, 0, 197, 3, 64, 0, 198, 131, 66, 0, 197, 99, 64, 0, 198, 227, 66, 0, 197, 195, 64, 0, 198, 66, 67, 0, 197, 34, 65, 0, 198, 130, 67, 0, 197, 98, 65, 0, 198, 195, 67, 0, 197, 163, 65, 0, 198, 35, 68, 0, 197, 3, 66, 0, 198, 131, 68, 0, 197, 99, 66, 0, 198, 227, 68, 0, 197, 195, 66, 0, 198, 67, 69, 0, 197, 35, 67, 0, 198, 162, 69, 0, 197, 130, 67, 0, 198, 226, 69, 0, 197, 194, 67, 0, 198, 34, 70, 0, 197, 2, 68, 0, 198, 98, 70, 0, 197, 66, 68, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 63, 5, 43, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 65, 1, 0, 198, 130, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 68, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 68, 0, 197, 2, 69, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 133, 0, 5, 0, 33, 5, 43, 0, 133, 0, 5, 0, 34, 5, 43, 0, 133, 0, 5, 0, 35, 5, 43, 0, 133, 0, 5, 0, 36, 5, 43, 0, 133, 0, 5, 0, 37, 5, 43, 0, 133, 0, 5, 0, 38, 5, 43, 0, 133, 0, 5, 0, 39, 5, 43, 0, 133, 0, 5, 0, 40, 5, 43, 0, 133, 0, 5, 0, 41, 5, 43, 0, 133, 0, 5, 0, 42, 5, 43, 0, 133, 0, 5, 0, 43, 5, 43, 0, 133, 0, 5, 0, 44, 5, 43, 0, 133, 0, 5, 0, 45, 5, 43, 0, 133, 0, 5, 0, 46, 5, 43, 0, 133, 0, 5, 0, 47, 5, 43, 0, 133, 0, 5, 0, 48, 5, 43, 0, 133, 0, 5, 0, 49, 5, 43, 0, 133, 0, 5, 0, 50, 5, 43, 0, 133, 0, 5, 0, 51, 5, 43, 0, 133, 0, 5, 0, 52, 5, 43, 0, 133, 0, 5, 0, 53, 5, 43, 0, 133, 0, 5, 0, 54, 5, 43, 0, 133, 0, 5, 0, 55, 5, 43, 0, 133, 0, 5, 0, 56, 5, 43, 0, 133, 0, 5, 0, 57, 5, 43, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 146, 0, 0, 0, 0, 0, 133, 0, 5, 0, 34, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 45, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 56, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 140, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 162, 0, 0, 0, 0, 0, 133, 0, 5, 0, 34, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 34, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 34, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 34, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 35, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 140, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 162, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 38, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 38, 5, 43, 0, 5, 0, 140, 0, 0, 0, 0, 0, 133, 0, 5, 0, 38, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 38, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 39, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 140, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 162, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 41, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 42, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 43, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 43, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 43, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 45, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 45, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 45, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 140, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 152, 0, 0, 0, 0, 0, 133, 0, 5, 0, 49, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 49, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 49, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 50, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 50, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 50, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 50, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 51, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 51, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 140, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 146, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 152, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 162, 0, 0, 0, 0, 0, 133, 0, 5, 0, 54, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 56, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 56, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 57, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 57, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 57, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 38, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 42, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 162, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 38, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 45, 5, 43, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 184, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 188, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 184, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 188, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 184, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 188, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 184, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 188, 0, 0, 0, 0, 0, 133, 0, 5, 0, 49, 5, 43, 0, 5, 0, 184, 0, 0, 0, 0, 0, 133, 0, 5, 0, 49, 5, 43, 0, 5, 0, 188, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 184, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 188, 0, 0, 0, 0, 0, 133, 0, 5, 0, 50, 5, 43, 0, 5, 0, 202, 0, 0, 0, 0, 0, 133, 0, 5, 0, 51, 5, 43, 0, 5, 0, 202, 0, 0, 0, 0, 0, 133, 0, 5, 0, 39, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 56, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 200, 0, 0, 0, 0, 0, 133, 0, 5, 0, 33, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 33, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 33, 5, 43, 0, 5, 0, 210, 0, 0, 0, 0, 0, 133, 0, 5, 0, 34, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 35, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 35, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 35, 5, 43, 0, 5, 0, 210, 0, 0, 0, 0, 0, 133, 0, 5, 0, 35, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 35, 5, 43, 0, 5, 0, 204, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 204, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 208, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 133, 0, 5, 0, 37, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 38, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 39, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 39, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 39, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 39, 5, 43, 0, 5, 0, 160, 0, 0, 0, 0, 0, 133, 0, 5, 0, 39, 5, 43, 0, 5, 0, 206, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 208, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 42, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 42, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 42, 5, 43, 0, 5, 0, 210, 0, 0, 0, 0, 0, 133, 0, 5, 0, 43, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 43, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 43, 5, 43, 0, 5, 0, 210, 0, 0, 0, 0, 0, 133, 0, 5, 0, 43, 5, 43, 0, 5, 0, 204, 0, 0, 0, 0, 0, 133, 0, 5, 0, 44, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 44, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 44, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 45, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 45, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 45, 5, 43, 0, 5, 0, 210, 0, 0, 0, 0, 0, 133, 0, 5, 0, 45, 5, 43, 0, 5, 0, 204, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 47, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 47, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 49, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 49, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 49, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 133, 0, 5, 0, 49, 5, 43, 0, 5, 0, 210, 0, 0, 0, 0, 0, 133, 0, 5, 0, 50, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 50, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 50, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 50, 5, 43, 0, 5, 0, 144, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 50, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 51, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 51, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 51, 5, 43, 0, 5, 0, 210, 0, 0, 0, 0, 0, 133, 0, 5, 0, 51, 5, 43, 0, 5, 0, 204, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 198, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 208, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 204, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 53, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 53, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 54, 5, 43, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 54, 5, 43, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 54, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 54, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 54, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 55, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 55, 5, 43, 0, 5, 0, 150, 0, 0, 0, 0, 0, 133, 0, 5, 0, 56, 5, 43, 0, 5, 0, 156, 0, 0, 0, 0, 0, 133, 0, 5, 0, 57, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 57, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 57, 5, 43, 0, 5, 0, 210, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 32, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 36, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 40, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 46, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 133, 0, 5, 0, 52, 5, 43, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 56, 5, 43, 0, 5, 0, 138, 0, 0, 0, 0, 0, 133, 0, 5, 0, 56, 5, 43, 0, 5, 0, 196, 0, 0, 0, 0, 0, 133, 0, 5, 0, 56, 5, 43, 0, 5, 0, 182, 0, 0, 0, 0, 0, 133, 0, 5, 0, 56, 5, 43, 0, 5, 0, 154, 0, 0, 0, 0, 0, 0, 0, 0, 193, 64, 5, 43, 5, 138, 0, 0, 193, 64, 5, 43, 5, 136, 0, 0, 193, 64, 5, 43, 5, 142, 0, 0, 193, 64, 5, 43, 5, 154, 0, 0, 193, 64, 5, 43, 5, 150, 0, 0, 193, 64, 5, 43, 5, 146, 0, 0, 193, 66, 5, 43, 5, 160, 0, 0, 193, 68, 5, 43, 5, 138, 0, 0, 193, 68, 5, 43, 5, 136, 0, 0, 193, 68, 5, 43, 5, 142, 0, 0, 193, 68, 5, 43, 5, 150, 0, 0, 193, 72, 5, 43, 5, 138, 0, 0, 193, 72, 5, 43, 5, 136, 0, 0, 193, 72, 5, 43, 5, 142, 0, 0, 193, 72, 5, 43, 5, 150, 0, 0, 193, 77, 5, 43, 5, 154, 0, 0, 193, 78, 5, 43, 5, 138, 0, 0, 193, 78, 5, 43, 5, 136, 0, 0, 193, 78, 5, 43, 5, 142, 0, 0, 193, 78, 5, 43, 5, 154, 0, 0, 193, 78, 5, 43, 5, 150, 0, 0, 193, 84, 5, 43, 5, 138, 0, 0, 193, 84, 5, 43, 5, 136, 0, 0, 193, 84, 5, 43, 5, 142, 0, 0, 193, 84, 5, 43, 5, 150, 0, 0, 193, 88, 5, 43, 5, 136, 0, 0, 193, 88, 5, 43, 5, 150, 0, 0, 193, 64, 5, 43, 5, 164, 0, 0, 193, 64, 5, 43, 5, 140, 0, 0, 193, 64, 5, 43, 5, 162, 0, 0, 193, 66, 5, 43, 5, 136, 0, 0, 193, 66, 5, 43, 5, 142, 0, 0, 193, 66, 5, 43, 5, 156, 0, 0, 193, 66, 5, 43, 5, 144, 0, 0, 193, 67, 5, 43, 5, 144, 0, 0, 193, 68, 5, 43, 5, 164, 0, 0, 193, 68, 5, 43, 5, 140, 0, 0, 193, 68, 5, 43, 5, 156, 0, 0, 193, 68, 5, 43, 5, 162, 0, 0, 193, 68, 5, 43, 5, 144, 0, 0, 193, 70, 5, 43, 5, 142, 0, 0, 193, 70, 5, 43, 5, 140, 0, 0, 193, 70, 5, 43, 5, 156, 0, 0, 193, 70, 5, 43, 5, 160, 0, 0, 193, 71, 5, 43, 5, 142, 0, 0, 193, 72, 5, 43, 5, 154, 0, 0, 193, 72, 5, 43, 5, 164, 0, 0, 193, 72, 5, 43, 5, 140, 0, 0, 193, 72, 5, 43, 5, 162, 0, 0, 193, 73, 5, 43, 5, 142, 0, 0, 193, 74, 5, 43, 5, 160, 0, 0, 193, 75, 5, 43, 5, 136, 0, 0, 193, 75, 5, 43, 5, 160, 0, 0, 193, 75, 5, 43, 5, 144, 0, 0, 193, 77, 5, 43, 5, 136, 0, 0, 193, 77, 5, 43, 5, 160, 0, 0, 193, 77, 5, 43, 5, 144, 0, 0, 193, 78, 5, 43, 5, 164, 0, 0, 193, 78, 5, 43, 5, 140, 0, 0, 193, 78, 5, 43, 5, 152, 0, 0, 193, 81, 5, 43, 5, 136, 0, 0, 193, 81, 5, 43, 5, 160, 0, 0, 193, 81, 5, 43, 5, 144, 0, 0, 193, 82, 5, 43, 5, 136, 0, 0, 193, 82, 5, 43, 5, 142, 0, 0, 193, 82, 5, 43, 5, 160, 0, 0, 193, 82, 5, 43, 5, 144, 0, 0, 193, 83, 5, 43, 5, 160, 0, 0, 193, 83, 5, 43, 5, 144, 0, 0, 193, 84, 5, 43, 5, 154, 0, 0, 193, 84, 5, 43, 5, 164, 0, 0, 193, 84, 5, 43, 5, 140, 0, 0, 193, 84, 5, 43, 5, 146, 0, 0, 193, 84, 5, 43, 5, 152, 0, 0, 193, 84, 5, 43, 5, 162, 0, 0, 193, 86, 5, 43, 5, 142, 0, 0, 193, 88, 5, 43, 5, 142, 0, 0, 193, 89, 5, 43, 5, 136, 0, 0, 193, 89, 5, 43, 5, 156, 0, 0, 193, 89, 5, 43, 5, 144, 0, 0, 193, 78, 5, 43, 5, 190, 0, 0, 193, 84, 5, 43, 5, 190, 0, 0, 193, 64, 5, 43, 5, 144, 0, 0, 193, 72, 5, 43, 5, 144, 0, 0, 193, 78, 5, 43, 5, 144, 0, 0, 193, 84, 5, 43, 5, 144, 0, 0, 193, 84, 5, 43, 5, 150, 0, 0, 5, 164, 0, 0, 193, 84, 5, 43, 5, 150, 0, 0, 5, 136, 0, 0, 193, 84, 5, 43, 5, 150, 0, 0, 5, 144, 0, 0, 193, 84, 5, 43, 5, 150, 0, 0, 5, 138, 0, 0, 193, 64, 5, 43, 5, 150, 0, 0, 5, 164, 0, 0, 193, 64, 5, 43, 5, 156, 0, 0, 5, 164, 0, 0, 193, 70, 5, 43, 5, 144, 0, 0, 193, 74, 5, 43, 5, 144, 0, 0, 193, 78, 5, 43, 5, 162, 0, 0, 193, 78, 5, 43, 5, 162, 0, 0, 5, 164, 0, 0, 193, 73, 5, 43, 5, 144, 0, 0, 193, 70, 5, 43, 5, 136, 0, 0, 193, 77, 5, 43, 5, 138, 0, 0, 193, 64, 5, 43, 5, 146, 0, 0, 5, 136, 0, 0, 193, 64, 5, 43, 5, 184, 0, 0, 193, 64, 5, 43, 5, 188, 0, 0, 193, 68, 5, 43, 5, 184, 0, 0, 193, 68, 5, 43, 5, 188, 0, 0, 193, 72, 5, 43, 5, 184, 0, 0, 193, 72, 5, 43, 5, 188, 0, 0, 193, 78, 5, 43, 5, 184, 0, 0, 193, 78, 5, 43, 5, 188, 0, 0, 193, 81, 5, 43, 5, 184, 0, 0, 193, 81, 5, 43, 5, 188, 0, 0, 193, 84, 5, 43, 5, 184, 0, 0, 193, 84, 5, 43, 5, 188, 0, 0, 193, 82, 5, 43, 5, 202, 0, 0, 193, 83, 5, 43, 5, 202, 0, 0, 193, 71, 5, 43, 5, 144, 0, 0, 193, 68, 5, 43, 5, 160, 0, 0, 193, 78, 5, 43, 5, 150, 0, 0, 5, 164, 0, 0, 193, 78, 5, 43, 5, 154, 0, 0, 5, 164, 0, 0, 193, 78, 5, 43, 5, 156, 0, 0, 193, 78, 5, 43, 5, 156, 0, 0, 5, 164, 0, 0, 193, 88, 5, 43, 5, 164, 0, 0, 193, 64, 5, 43, 5, 200, 0, 0, 193, 65, 5, 43, 5, 156, 0, 0, 193, 65, 5, 43, 5, 196, 0, 0, 193, 65, 5, 43, 5, 210, 0, 0, 193, 66, 5, 43, 5, 160, 0, 0, 5, 136, 0, 0, 193, 67, 5, 43, 5, 156, 0, 0, 193, 67, 5, 43, 5, 196, 0, 0, 193, 67, 5, 43, 5, 210, 0, 0, 193, 67, 5, 43, 5, 160, 0, 0, 193, 67, 5, 43, 5, 204, 0, 0, 193, 68, 5, 43, 5, 164, 0, 0, 5, 138, 0, 0, 193, 68, 5, 43, 5, 164, 0, 0, 5, 136, 0, 0, 193, 68, 5, 43, 5, 204, 0, 0, 193, 68, 5, 43, 5, 208, 0, 0, 193, 68, 5, 43, 5, 160, 0, 0, 5, 140, 0, 0, 193, 69, 5, 43, 5, 156, 0, 0, 193, 70, 5, 43, 5, 164, 0, 0, 193, 71, 5, 43, 5, 156, 0, 0, 193, 71, 5, 43, 5, 196, 0, 0, 193, 71, 5, 43, 5, 150, 0, 0, 193, 71, 5, 43, 5, 160, 0, 0, 193, 71, 5, 43, 5, 206, 0, 0, 193, 72, 5, 43, 5, 208, 0, 0, 193, 72, 5, 43, 5, 150, 0, 0, 5, 136, 0, 0, 193, 74, 5, 43, 5, 136, 0, 0, 193, 74, 5, 43, 5, 196, 0, 0, 193, 74, 5, 43, 5, 210, 0, 0, 193, 75, 5, 43, 5, 196, 0, 0, 193, 75, 5, 43, 5, 196, 0, 0, 5, 164, 0, 0, 193, 75, 5, 43, 5, 210, 0, 0, 193, 75, 5, 43, 5, 204, 0, 0, 193, 76, 5, 43, 5, 136, 0, 0, 193, 76, 5, 43, 5, 156, 0, 0, 193, 76, 5, 43, 5, 196, 0, 0, 193, 77, 5, 43, 5, 156, 0, 0, 193, 77, 5, 43, 5, 196, 0, 0, 193, 77, 5, 43, 5, 210, 0, 0, 193, 77, 5, 43, 5, 204, 0, 0, 193, 78, 5, 43, 5, 154, 0, 0, 5, 136, 0, 0, 193, 78, 5, 43, 5, 154, 0, 0, 5, 150, 0, 0, 193, 78, 5, 43, 5, 164, 0, 0, 5, 138, 0, 0, 193, 78, 5, 43, 5, 164, 0, 0, 5, 136, 0, 0, 193, 79, 5, 43, 5, 136, 0, 0, 193, 79, 5, 43, 5, 156, 0, 0, 193, 81, 5, 43, 5, 156, 0, 0, 193, 81, 5, 43, 5, 196, 0, 0, 193, 81, 5, 43, 5, 196, 0, 0, 5, 164, 0, 0, 193, 81, 5, 43, 5, 210, 0, 0, 193, 82, 5, 43, 5, 156, 0, 0, 193, 82, 5, 43, 5, 196, 0, 0, 193, 82, 5, 43, 5, 136, 0, 0, 5, 156, 0, 0, 193, 82, 5, 43, 5, 144, 0, 0, 5, 156, 0, 0, 193, 82, 5, 43, 5, 196, 0, 0, 5, 156, 0, 0, 193, 83, 5, 43, 5, 156, 0, 0, 193, 83, 5, 43, 5, 196, 0, 0, 193, 83, 5, 43, 5, 210, 0, 0, 193, 83, 5, 43, 5, 204, 0, 0, 193, 84, 5, 43, 5, 198, 0, 0, 193, 84, 5, 43, 5, 208, 0, 0, 193, 84, 5, 43, 5, 204, 0, 0, 193, 84, 5, 43, 5, 154, 0, 0, 5, 136, 0, 0, 193, 84, 5, 43, 5, 164, 0, 0, 5, 150, 0, 0, 193, 85, 5, 43, 5, 154, 0, 0, 193, 85, 5, 43, 5, 196, 0, 0, 193, 86, 5, 43, 5, 138, 0, 0, 193, 86, 5, 43, 5, 136, 0, 0, 193, 86, 5, 43, 5, 150, 0, 0, 193, 86, 5, 43, 5, 156, 0, 0, 193, 86, 5, 43, 5, 196, 0, 0, 193, 87, 5, 43, 5, 156, 0, 0, 193, 87, 5, 43, 5, 150, 0, 0, 193, 88, 5, 43, 5, 156, 0, 0, 193, 89, 5, 43, 5, 142, 0, 0, 193, 89, 5, 43, 5, 196, 0, 0, 193, 89, 5, 43, 5, 210, 0, 0, 193, 71, 5, 43, 5, 210, 0, 0, 193, 83, 5, 43, 5, 150, 0, 0, 193, 86, 5, 43, 5, 146, 0, 0, 193, 88, 5, 43, 5, 146, 0, 0, 193, 64, 5, 43, 5, 196, 0, 0, 193, 64, 5, 43, 5, 182, 0, 0, 193, 64, 5, 43, 5, 142, 0, 0, 5, 136, 0, 0, 193, 64, 5, 43, 5, 142, 0, 0, 5, 138, 0, 0, 193, 64, 5, 43, 5, 142, 0, 0, 5, 182, 0, 0, 193, 64, 5, 43, 5, 142, 0, 0, 5, 154, 0, 0, 193, 64, 5, 43, 5, 196, 0, 0, 5, 142, 0, 0, 193, 64, 5, 43, 5, 140, 0, 0, 5, 136, 0, 0, 193, 64, 5, 43, 5, 140, 0, 0, 5, 138, 0, 0, 193, 64, 5, 43, 5, 140, 0, 0, 5, 182, 0, 0, 193, 64, 5, 43, 5, 140, 0, 0, 5, 154, 0, 0, 193, 64, 5, 43, 5, 196, 0, 0, 5, 140, 0, 0, 193, 68, 5, 43, 5, 196, 0, 0, 193, 68, 5, 43, 5, 182, 0, 0, 193, 68, 5, 43, 5, 154, 0, 0, 193, 68, 5, 43, 5, 142, 0, 0, 5, 136, 0, 0, 193, 68, 5, 43, 5, 142, 0, 0, 5, 138, 0, 0, 193, 68, 5, 43, 5, 142, 0, 0, 5, 182, 0, 0, 193, 68, 5, 43, 5, 142, 0, 0, 5, 154, 0, 0, 193, 68, 5, 43, 5, 196, 0, 0, 5, 142, 0, 0, 193, 72, 5, 43, 5, 182, 0, 0, 193, 72, 5, 43, 5, 196, 0, 0, 193, 78, 5, 43, 5, 196, 0, 0, 193, 78, 5, 43, 5, 182, 0, 0, 193, 78, 5, 43, 5, 142, 0, 0, 5, 136, 0, 0, 193, 78, 5, 43, 5, 142, 0, 0, 5, 138, 0, 0, 193, 78, 5, 43, 5, 142, 0, 0, 5, 182, 0, 0, 193, 78, 5, 43, 5, 142, 0, 0, 5, 154, 0, 0, 193, 78, 5, 43, 5, 196, 0, 0, 5, 142, 0, 0, 193, 78, 5, 43, 5, 190, 0, 0, 5, 136, 0, 0, 193, 78, 5, 43, 5, 190, 0, 0, 5, 138, 0, 0, 193, 78, 5, 43, 5, 190, 0, 0, 5, 182, 0, 0, 193, 78, 5, 43, 5, 190, 0, 0, 5, 154, 0, 0, 193, 78, 5, 43, 5, 190, 0, 0, 5, 196, 0, 0, 193, 84, 5, 43, 5, 196, 0, 0, 193, 84, 5, 43, 5, 182, 0, 0, 193, 84, 5, 43, 5, 190, 0, 0, 5, 136, 0, 0, 193, 84, 5, 43, 5, 190, 0, 0, 5, 138, 0, 0, 193, 84, 5, 43, 5, 190, 0, 0, 5, 182, 0, 0, 193, 84, 5, 43, 5, 190, 0, 0, 5, 154, 0, 0, 193, 84, 5, 43, 5, 190, 0, 0, 5, 196, 0, 0, 193, 88, 5, 43, 5, 138, 0, 0, 193, 88, 5, 43, 5, 196, 0, 0, 193, 88, 5, 43, 5, 182, 0, 0, 193, 88, 5, 43, 5, 154, 0, 0, 193, 28, 5, 43, 5, 158, 0, 0, 193, 27, 5, 43, 5, 158, 0, 0, 193, 29, 5, 43, 5, 158, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 193, 15, 5, 43, 193, 16, 5, 43, 193, 17, 5, 43, 193, 18, 5, 43, 193, 19, 5, 43, 193, 20, 5, 43, 193, 21, 5, 43, 193, 22, 5, 43, 193, 23, 5, 43, 193, 24, 5, 43, 26, 0, 198, 226, 48, 0, 4, 3, 26, 192, 198, 227, 23, 0, 197, 130, 48, 0, 4, 3, 23, 192, 197, 131, 3, 0, 69, 0, 60, 0, 32, 0, 71, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 142, 15, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 147, 15, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 72, 9, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 76, 9, 0, 0, 76, 9, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 49, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 48, 220, 38, 0, 67, 0, 60, 0, 9, 1, 60, 0, 60, 0, 60, 0, 8, 1, 38, 0, 71, 0, 60, 0, 29, 1, 60, 0, 60, 0, 60, 0, 28, 1, 38, 0, 72, 0, 60, 0, 37, 1, 60, 0, 60, 0, 60, 0, 36, 1, 38, 0, 74, 0, 60, 0, 53, 1, 60, 0, 60, 0, 60, 0, 52, 1, 38, 0, 83, 0, 60, 0, 93, 1, 60, 0, 60, 0, 60, 0, 92, 1, 38, 0, 85, 0, 60, 0, 109, 1, 60, 0, 60, 0, 60, 0, 108, 1, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 50, 36, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 164, 28, 0, 0, 164, 28, 0, 0, 164, 28, 0, 0, 164, 28, 0, 0, 0, 30, 0, 0, 0, 30, 0, 0, 10, 32, 0, 0, 10, 32, 0, 0, 26, 36, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 189, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 94, 0, 102, 0, 107, 0, 115, 0, 48, 0, 56, 0, 120, 0, 128, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 151, 0, 48, 0, 159, 0, 165, 0, 48, 0, 48, 0, 172, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 144, 0, 48, 0, 48, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 120, 1, 172, 1, 192, 0, 224, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 9, 0, 202, 193, 9, 0, 202, 226, 9, 0, 202, 3, 10, 0, 202, 36, 10, 0, 202, 69, 10, 0, 202, 102, 10, 0, 202, 135, 10, 0, 202, 168, 10, 0, 202, 201, 10, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 198, 0, 0, 201, 134, 1, 0, 156, 5, 0, 58, 201, 70, 2, 0, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 6, 3, 0, 156, 5, 0, 80, 201, 230, 3, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 198, 4, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 134, 5, 0, 201, 70, 6, 0, 5, 5, 0, 58, 201, 6, 7, 0, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 198, 7, 0, 5, 5, 0, 80, 201, 166, 8, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 8, 0, 201, 134, 9, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 9, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 9, 0, 201, 38, 10, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 55, 5, 5, 2, 55, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 10, 0, 201, 166, 11, 0, 134, 5, 2, 57, 5, 5, 2, 57, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 61, 5, 5, 2, 61, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 201, 102, 12, 0, 201, 38, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 83, 5, 5, 2, 83, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 13, 0, 201, 166, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 15, 0, 201, 38, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 16, 0, 201, 166, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 18, 0, 201, 38, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 19, 0, 201, 166, 20, 0, 201, 102, 21, 0, 201, 38, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 22, 0, 201, 166, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 24, 0, 201, 38, 25, 0, 201, 230, 25, 0, 201, 166, 26, 0, 201, 102, 27, 0, 201, 38, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 29, 0, 201, 102, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 31, 0, 201, 230, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 55, 5, 160, 0, 0, 5, 5, 2, 55, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 57, 5, 196, 0, 0, 5, 5, 2, 57, 5, 196, 0, 0, 134, 5, 2, 57, 5, 160, 0, 0, 5, 5, 2, 57, 5, 160, 0, 0, 134, 5, 2, 57, 5, 206, 0, 0, 5, 5, 2, 57, 5, 206, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 83, 5, 198, 0, 0, 5, 5, 2, 83, 5, 198, 0, 0, 134, 5, 2, 83, 5, 208, 0, 0, 5, 5, 2, 83, 5, 208, 0, 0, 134, 5, 2, 83, 5, 204, 0, 0, 5, 5, 2, 83, 5, 204, 0, 0, 5, 5, 2, 57, 5, 210, 0, 0, 134, 5, 2, 83, 5, 196, 0, 0, 5, 5, 2, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 48, 0, 2, 3, 2, 239, 134, 5, 0, 54, 156, 5, 48, 0, 2, 3, 2, 247, 134, 5, 0, 56, 156, 5, 48, 0, 2, 3, 2, 249, 134, 5, 0, 60, 156, 5, 48, 0, 2, 3, 2, 253, 134, 5, 0, 78, 156, 5, 48, 0, 2, 3, 255, 255, 2, 79, 134, 5, 0, 82, 156, 5, 48, 0, 6, 3, 255, 255, 2, 83, 134, 5, 0, 46, 5, 5, 48, 0, 2, 3, 2, 239, 5, 5, 0, 54, 5, 5, 48, 0, 2, 3, 2, 247, 5, 5, 0, 56, 5, 5, 48, 0, 2, 3, 2, 249, 5, 5, 0, 60, 5, 5, 48, 0, 2, 3, 2, 253, 5, 5, 0, 78, 5, 5, 48, 0, 2, 3, 255, 255, 2, 79, 5, 5, 0, 82, 5, 5, 48, 0, 6, 3, 255, 255, 2, 83, 5, 5, 156, 46, 196, 160, 48, 0, 2, 3, 197, 162, 5, 46, 196, 160, 48, 0, 2, 3, 0, 192, 197, 98, 156, 54, 196, 160, 48, 0, 2, 3, 0, 192, 197, 162, 5, 54, 196, 160, 48, 0, 2, 3, 0, 192, 197, 226, 156, 78, 196, 160, 48, 0, 2, 3, 1, 192, 197, 34, 5, 78, 196, 160, 48, 0, 2, 3, 1, 192, 197, 98, 156, 82, 196, 162, 48, 0, 6, 3, 1, 192, 197, 162, 5, 82, 196, 162, 48, 0, 6, 3, 1, 192, 197, 226, 156, 82, 196, 190, 48, 0, 6, 3, 2, 192, 197, 34, 5, 82, 196, 190, 48, 0, 6, 3, 2, 192, 197, 98, 156, 78, 196, 202, 48, 0, 2, 3, 2, 192, 197, 162, 5, 78, 196, 202, 48, 0, 2, 3, 2, 192, 197, 226, 156, 56, 196, 196, 48, 0, 2, 3, 3, 192, 197, 34, 5, 56, 196, 196, 48, 0, 2, 3, 3, 192, 197, 98, 156, 56, 196, 160, 48, 0, 2, 3, 3, 192, 197, 162, 5, 56, 196, 160, 48, 0, 2, 3, 3, 192, 197, 226, 156, 56, 196, 206, 48, 0, 2, 3, 4, 192, 197, 34, 5, 56, 196, 206, 48, 0, 2, 3, 4, 192, 197, 98, 156, 78, 196, 196, 48, 0, 2, 3, 4, 192, 197, 162, 5, 78, 196, 196, 48, 0, 2, 3, 4, 192, 197, 226, 156, 82, 196, 198, 48, 0, 6, 3, 5, 192, 197, 34, 5, 82, 196, 198, 48, 0, 6, 3, 5, 192, 197, 98, 156, 82, 196, 208, 48, 0, 6, 3, 5, 192, 197, 162, 5, 82, 196, 208, 48, 0, 6, 3, 5, 192, 197, 226, 156, 82, 196, 204, 48, 0, 6, 3, 6, 192, 197, 34, 5, 82, 196, 204, 48, 0, 6, 3, 6, 192, 197, 98, 5, 56, 196, 210, 48, 0, 2, 3, 6, 192, 197, 162, 156, 82, 196, 196, 48, 0, 6, 3, 6, 192, 197, 226, 5, 82, 196, 196, 48, 0, 6, 3, 7, 192, 197, 34, 7, 0, 197, 99, 48, 0, 6, 3, 7, 192, 197, 195, 8, 0, 197, 35, 48, 0, 6, 3, 8, 192, 197, 131, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 72, 186, 76, 185, 80, 28, 4, 30, 4, 186, 100, 32, 4, 185, 116, 185, 120, 185, 124, 186, 128, 185, 136, 185, 140, 185, 144, 185, 152, 34, 4, 185, 164, 36, 4, 185, 180, 185, 184, 185, 188, 185, 192, 185, 196, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 38, 4, 168, 72, 168, 76, 168, 80, 40, 4, 42, 4, 168, 100, 44, 4, 168, 116, 168, 120, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 168, 152, 46, 4, 168, 164, 48, 4, 168, 180, 168, 184, 168, 188, 168, 192, 168, 196, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 136, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 50, 4, 186, 77, 154, 77, 250, 77, 90, 78, 186, 101, 154, 101, 250, 101, 90, 102, 217, 72, 154, 130, 185, 137, 153, 137, 249, 137, 153, 138, 89, 138, 136, 14, 217, 138, 185, 173, 153, 173, 249, 173, 89, 174, 153, 193, 185, 200, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 52, 4, 168, 77, 136, 77, 232, 77, 72, 78, 168, 101, 136, 101, 232, 101, 72, 102, 200, 72, 136, 130, 168, 137, 136, 137, 232, 137, 136, 138, 72, 138, 128, 14, 200, 138, 168, 173, 136, 173, 232, 173, 72, 174, 136, 193, 168, 200, 72, 194, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 185, 68, 168, 68, 185, 66, 168, 66, 25, 66, 8, 66, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 185, 88, 168, 88, 217, 85, 200, 85, 185, 86, 168, 86, 54, 4, 56, 4, 185, 96, 168, 96, 89, 95, 72, 95, 154, 102, 136, 102, 58, 103, 40, 103, 218, 101, 200, 101, 26, 103, 8, 103, 186, 102, 168, 104, 6, 8, 8, 8, 185, 112, 168, 112, 249, 118, 232, 118, 168, 148, 153, 121, 136, 121, 249, 122, 232, 122, 25, 122, 8, 122, 121, 123, 104, 123, 89, 123, 72, 123, 154, 129, 136, 129, 250, 130, 232, 130, 26, 130, 8, 130, 10, 8, 185, 132, 168, 132, 57, 139, 40, 139, 217, 137, 200, 137, 121, 138, 104, 138, 12, 8, 14, 8, 153, 153, 136, 153, 249, 154, 232, 154, 25, 154, 8, 154, 154, 157, 136, 157, 185, 160, 168, 160, 58, 4, 60, 4, 26, 158, 8, 158, 249, 166, 232, 166, 25, 166, 8, 166, 185, 168, 168, 168, 153, 174, 136, 174, 57, 175, 40, 175, 185, 176, 168, 176, 57, 174, 40, 174, 121, 174, 104, 174, 62, 4, 64, 4, 249, 185, 232, 185, 249, 193, 232, 193, 89, 194, 153, 197, 136, 197, 185, 198, 168, 198, 25, 198, 8, 198, 232, 156, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 156, 169, 156, 200, 56, 169, 76, 187, 100, 187, 108, 169, 100, 169, 108, 169, 204, 169, 128, 217, 136, 187, 76, 200, 136, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 185, 84, 255, 5, 185, 92, 255, 5, 186, 108, 255, 5, 186, 156, 255, 5, 185, 172, 255, 5, 168, 64, 255, 5, 168, 84, 255, 5, 168, 92, 255, 5, 168, 108, 255, 5, 168, 156, 255, 5, 168, 172, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 249, 86, 255, 5, 232, 86, 255, 5, 250, 158, 255, 5, 232, 158, 255, 5, 25, 175, 255, 5, 8, 175, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 51, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 65, 9, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 70, 9, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 25, 0, 32, 7, 0, 0, 0, 26, 0, 0, 0, 10, 25, 0, 0, 10, 25, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 67, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 116, 114, 97, 100, 105, 116, 105, 111, 110, 97, 108, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 38, 0, 78, 0, 60, 0, 110, 0, 3, 3, 60, 0, 60, 0, 60, 0, 78, 0, 3, 3, 0, 0, 38, 0, 78, 0, 60, 0, 110, 0, 3, 3, 60, 0, 60, 0, 60, 0, 78, 0, 3, 3, 38, 0, 67, 0, 60, 0, 99, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 72, 0, 38, 0, 108, 0, 60, 0, 108, 0, 108, 0, 60, 0, 60, 0, 60, 0, 76, 0, 108, 0, 60, 0, 60, 0, 60, 0, 76, 0, 76, 0, 0, 0, 212, 35, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 77, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 148, 32, 0, 0, 148, 32, 0, 0, 148, 32, 0, 0, 148, 32, 0, 0, 252, 34, 0, 0, 252, 34, 0, 0, 188, 35, 0, 0, 188, 35, 0, 0, 188, 35, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 252, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 84, 0, 92, 0, 98, 0, 106, 0, 114, 0, 122, 0, 48, 0, 56, 0, 121, 0, 129, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 146, 0, 154, 0, 48, 0, 48, 0, 48, 0, 162, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 170, 0, 48, 0, 48, 0, 178, 0, 48, 0, 184, 0, 192, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 195, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 202, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 210, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 213, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 144, 0, 48, 0, 48, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 221, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 229, 0, 48, 0, 48, 0, 235, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 80, 1, 136, 1, 200, 1, 192, 0, 228, 1, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 18, 0, 202, 33, 18, 0, 202, 66, 18, 0, 202, 99, 18, 0, 202, 132, 18, 0, 202, 165, 18, 0, 202, 198, 18, 0, 202, 231, 18, 0, 202, 8, 19, 0, 202, 41, 19, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 198, 0, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 166, 1, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 8, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 134, 5, 2, 69, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 9, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 9, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 5, 5, 2, 69, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 2, 0, 201, 38, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 230, 3, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 5, 0, 5, 71, 238, 102, 201, 166, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 1, 0, 192, 0, 0, 0, 197, 98, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 1, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 197, 98, 2, 0, 192, 0, 0, 0, 197, 162, 2, 0, 192, 0, 0, 0, 197, 226, 2, 0, 197, 35, 3, 0, 197, 130, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 3, 0, 197, 3, 4, 0, 197, 98, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 162, 0, 0, 197, 226, 4, 0, 5, 5, 10, 121, 197, 34, 5, 0, 197, 98, 5, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 197, 226, 6, 0, 197, 34, 7, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 1, 0, 5, 5, 24, 121, 197, 98, 1, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 7, 0, 201, 70, 8, 0, 201, 6, 9, 0, 201, 198, 9, 0, 201, 134, 10, 0, 201, 70, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 162, 0, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 226, 0, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 34, 1, 0, 5, 5, 24, 121, 197, 98, 1, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 162, 1, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 226, 1, 0, 5, 5, 107, 121, 197, 34, 2, 0, 5, 5, 111, 121, 197, 98, 2, 0, 5, 5, 115, 121, 197, 162, 2, 0, 5, 5, 119, 121, 197, 226, 2, 0, 197, 35, 3, 0, 197, 130, 3, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 194, 3, 0, 197, 3, 4, 0, 197, 98, 4, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 162, 4, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 162, 0, 0, 197, 226, 4, 0, 5, 5, 10, 121, 197, 34, 5, 0, 197, 98, 5, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 197, 226, 6, 0, 197, 34, 7, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 1, 0, 5, 5, 24, 121, 197, 98, 1, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 68, 156, 5, 48, 0, 3, 3, 255, 255, 2, 69, 134, 5, 0, 68, 5, 5, 48, 0, 3, 3, 255, 255, 2, 69, 5, 5, 156, 68, 196, 160, 48, 0, 3, 3, 197, 162, 5, 68, 196, 160, 48, 0, 3, 3, 0, 192, 197, 98, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 68, 196, 196, 48, 0, 3, 3, 7, 192, 197, 98, 5, 68, 196, 196, 48, 0, 3, 3, 7, 192, 197, 162, 156, 68, 196, 210, 48, 0, 3, 3, 7, 192, 197, 226, 5, 68, 196, 210, 48, 0, 3, 3, 8, 192, 197, 34, 156, 68, 196, 204, 48, 0, 3, 3, 8, 192, 197, 98, 5, 68, 196, 204, 48, 0, 3, 3, 8, 192, 197, 162, 3, 0, 67, 0, 58, 0, 32, 0, 67, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 186, 29, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 88, 25, 0, 0, 88, 25, 0, 0, 210, 25, 0, 0, 210, 25, 0, 0, 162, 29, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 128, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 111, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 104, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 2, 0, 202, 1, 3, 0, 202, 34, 3, 0, 202, 67, 3, 0, 202, 100, 3, 0, 202, 133, 3, 0, 202, 166, 3, 0, 202, 199, 3, 0, 202, 232, 3, 0, 202, 9, 4, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 6, 0, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 230, 0, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 2, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 134, 5, 2, 69, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 2, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 2, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 5, 5, 2, 69, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 1, 0, 201, 102, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 3, 0, 201, 230, 3, 0, 201, 166, 4, 0, 201, 102, 5, 0, 201, 38, 6, 0, 201, 230, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 68, 156, 5, 48, 0, 3, 3, 255, 255, 2, 69, 134, 5, 0, 68, 5, 5, 48, 0, 3, 3, 255, 255, 2, 69, 5, 5, 156, 68, 196, 160, 48, 0, 3, 3, 197, 162, 5, 68, 196, 160, 48, 0, 3, 3, 0, 192, 197, 98, 156, 68, 196, 196, 48, 0, 3, 3, 0, 192, 197, 162, 5, 68, 196, 196, 48, 0, 3, 3, 0, 192, 197, 226, 156, 68, 196, 210, 48, 0, 3, 3, 1, 192, 197, 34, 5, 68, 196, 210, 48, 0, 3, 3, 1, 192, 197, 98, 156, 68, 196, 204, 48, 0, 3, 3, 1, 192, 197, 162, 5, 68, 196, 204, 48, 0, 3, 3, 1, 192, 197, 226, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 26, 4, 185, 124, 185, 128, 185, 132, 185, 140, 186, 144, 185, 148, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 185, 176, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 28, 4, 168, 124, 168, 128, 168, 132, 168, 140, 168, 144, 168, 148, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 168, 176, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 124, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 185, 116, 185, 125, 153, 125, 249, 125, 153, 126, 89, 126, 136, 14, 217, 126, 185, 157, 153, 157, 249, 157, 89, 158, 153, 173, 185, 180, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 168, 116, 168, 125, 136, 125, 232, 125, 136, 126, 72, 126, 128, 14, 200, 126, 168, 157, 136, 157, 232, 157, 72, 158, 136, 173, 168, 180, 72, 174, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 136, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 30, 4, 32, 4, 26, 114, 8, 114, 10, 8, 185, 120, 168, 120, 57, 127, 40, 127, 217, 125, 200, 125, 121, 126, 104, 126, 12, 8, 14, 8, 153, 141, 136, 141, 249, 142, 232, 142, 25, 142, 8, 142, 154, 145, 136, 145, 250, 145, 232, 145, 250, 146, 232, 146, 26, 146, 8, 146, 249, 150, 232, 150, 25, 150, 8, 150, 185, 152, 168, 152, 153, 158, 136, 158, 57, 159, 40, 159, 217, 157, 200, 157, 57, 158, 40, 158, 121, 158, 104, 158, 25, 159, 8, 159, 249, 165, 232, 165, 249, 173, 232, 173, 89, 174, 153, 177, 136, 177, 185, 178, 168, 178, 25, 178, 8, 178, 232, 144, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 144, 169, 144, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 184, 169, 112, 217, 124, 187, 72, 200, 124, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 186, 112, 255, 5, 168, 112, 255, 5, 250, 114, 255, 5, 232, 114, 255, 1, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 63, 9, 0, 16, 30, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 14, 33, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 196, 26, 0, 0, 196, 26, 0, 0, 168, 28, 0, 0, 206, 28, 0, 0, 246, 32, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 128, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 111, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 104, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 64, 14, 0, 202, 97, 14, 0, 202, 130, 14, 0, 202, 163, 14, 0, 202, 196, 14, 0, 202, 229, 14, 0, 202, 6, 15, 0, 202, 39, 15, 0, 202, 72, 15, 0, 202, 105, 15, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 0, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 96, 7, 0, 156, 5, 0, 66, 201, 198, 14, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 160, 15, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 160, 19, 0, 5, 5, 0, 66, 201, 102, 23, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 130, 13, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 134, 5, 2, 69, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 194, 13, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 2, 14, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 5, 5, 2, 69, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 24, 0, 201, 6, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 25, 0, 201, 134, 26, 0, 201, 70, 27, 0, 201, 6, 28, 0, 201, 198, 28, 0, 201, 134, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 135, 5, 2, 47, 5, 142, 0, 0, 70, 5, 2, 47, 5, 142, 0, 0, 135, 5, 2, 47, 5, 144, 0, 0, 70, 5, 2, 47, 5, 144, 0, 0, 135, 5, 2, 47, 5, 156, 0, 0, 70, 5, 2, 47, 5, 156, 0, 0, 135, 5, 2, 47, 5, 196, 0, 0, 70, 5, 2, 47, 5, 196, 0, 0, 135, 5, 2, 47, 5, 150, 0, 0, 70, 5, 2, 47, 5, 150, 0, 0, 135, 5, 2, 47, 5, 160, 0, 0, 70, 5, 2, 47, 5, 160, 0, 0, 135, 5, 2, 47, 5, 206, 0, 0, 70, 5, 2, 47, 5, 206, 0, 0, 70, 5, 2, 47, 5, 210, 0, 0, 135, 5, 2, 65, 5, 136, 0, 0, 70, 5, 2, 65, 5, 136, 0, 0, 135, 5, 2, 65, 5, 160, 0, 0, 70, 5, 2, 65, 5, 160, 0, 0, 135, 5, 2, 65, 5, 144, 0, 0, 70, 5, 2, 65, 5, 144, 0, 0, 135, 5, 2, 65, 5, 196, 0, 0, 70, 5, 2, 65, 5, 196, 0, 0, 135, 5, 2, 65, 5, 196, 0, 0, 5, 164, 0, 0, 70, 5, 2, 65, 5, 196, 0, 0, 5, 164, 0, 0, 135, 5, 2, 65, 5, 210, 0, 0, 70, 5, 2, 65, 5, 210, 0, 0, 135, 5, 2, 65, 5, 204, 0, 0, 70, 5, 2, 65, 5, 204, 0, 0, 5, 5, 2, 47, 5, 142, 0, 0, 5, 5, 2, 47, 5, 144, 0, 0, 5, 5, 2, 47, 5, 156, 0, 0, 5, 5, 2, 47, 5, 196, 0, 0, 5, 5, 2, 47, 5, 150, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 206, 0, 0, 5, 5, 2, 47, 5, 210, 0, 0, 5, 5, 2, 65, 5, 136, 0, 0, 5, 5, 2, 65, 5, 160, 0, 0, 5, 5, 2, 65, 5, 144, 0, 0, 5, 5, 2, 65, 5, 196, 0, 0, 5, 5, 2, 65, 5, 196, 0, 0, 5, 164, 0, 0, 5, 5, 2, 65, 5, 210, 0, 0, 5, 5, 2, 65, 5, 204, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 2, 192, 197, 162, 41, 30, 2, 192, 197, 226, 42, 30, 3, 192, 197, 34, 43, 30, 3, 192, 197, 98, 150, 30, 3, 192, 197, 162, 36, 30, 1, 192, 197, 162, 37, 30, 1, 192, 197, 226, 38, 30, 2, 192, 197, 34, 39, 30, 2, 192, 197, 98, 30, 2, 12, 0, 30, 2, 0, 192, 197, 162, 31, 2, 0, 192, 197, 226, 34, 30, 1, 192, 197, 34, 35, 30, 1, 192, 197, 98, 72, 0, 2, 239, 135, 5, 104, 0, 2, 239, 70, 5, 36, 1, 197, 162, 37, 1, 0, 192, 197, 98, 0, 64, 156, 5, 15, 0, 54, 30, 26, 0, 58, 30, 12, 0, 58, 30, 6, 192, 197, 162, 59, 30, 6, 192, 197, 226, 60, 30, 7, 192, 197, 34, 61, 30, 7, 192, 197, 98, 54, 30, 5, 192, 197, 98, 55, 30, 5, 192, 197, 162, 56, 30, 5, 192, 197, 227, 57, 30, 6, 192, 197, 67, 59, 1, 12, 0, 59, 1, 4, 192, 197, 98, 60, 1, 4, 192, 197, 162, 61, 1, 4, 192, 197, 226, 62, 1, 5, 192, 197, 34, 76, 0, 255, 255, 2, 65, 135, 5, 108, 0, 255, 255, 2, 65, 70, 5, 57, 1, 3, 192, 197, 226, 58, 1, 4, 192, 197, 34, 0, 68, 156, 5, 48, 0, 3, 3, 255, 255, 2, 69, 134, 5, 0, 46, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 8, 192, 197, 98, 39, 30, 8, 192, 197, 162, 41, 30, 8, 192, 197, 226, 43, 30, 9, 192, 197, 34, 150, 30, 9, 192, 197, 98, 104, 0, 2, 239, 5, 5, 37, 1, 7, 192, 197, 162, 31, 2, 7, 192, 197, 226, 35, 30, 8, 192, 197, 34, 0, 64, 5, 5, 7, 0, 55, 30, 12, 0, 55, 30, 10, 192, 197, 98, 57, 30, 10, 192, 197, 163, 59, 30, 11, 192, 197, 2, 61, 30, 11, 192, 197, 66, 108, 0, 255, 255, 2, 65, 5, 5, 58, 1, 9, 192, 197, 162, 60, 1, 9, 192, 197, 226, 62, 1, 10, 192, 197, 34, 0, 68, 5, 5, 48, 0, 3, 3, 255, 255, 2, 69, 5, 5, 156, 68, 196, 160, 48, 0, 3, 3, 11, 192, 197, 130, 5, 68, 196, 160, 48, 0, 3, 3, 11, 192, 197, 194, 156, 68, 196, 196, 48, 0, 3, 3, 12, 192, 197, 2, 5, 68, 196, 196, 48, 0, 3, 3, 12, 192, 197, 66, 156, 68, 196, 210, 48, 0, 3, 3, 12, 192, 197, 130, 5, 68, 196, 210, 48, 0, 3, 3, 12, 192, 197, 194, 156, 68, 196, 204, 48, 0, 3, 3, 13, 192, 197, 2, 5, 68, 196, 204, 48, 0, 3, 3, 13, 192, 197, 66, 18, 0, 72, 0, 73, 0, 104, 0, 105, 0, 108, 0, 109, 0, 36, 1, 38, 1, 57, 1, 63, 1, 30, 2, 32, 2, 34, 30, 44, 30, 54, 30, 62, 30, 150, 30, 151, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 72, 186, 76, 185, 80, 185, 84, 185, 88, 186, 92, 186, 100, 185, 104, 36, 4, 185, 116, 54, 4, 185, 132, 185, 136, 185, 140, 185, 148, 186, 152, 185, 156, 185, 164, 185, 168, 185, 172, 185, 176, 185, 180, 185, 184, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 56, 4, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 92, 168, 100, 168, 104, 62, 4, 168, 116, 72, 4, 168, 132, 168, 136, 168, 140, 168, 148, 168, 152, 168, 156, 168, 164, 168, 168, 168, 172, 168, 176, 168, 180, 168, 184, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 132, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 77, 154, 77, 250, 77, 90, 78, 186, 93, 154, 93, 250, 93, 90, 94, 217, 72, 185, 124, 185, 133, 153, 133, 249, 133, 153, 134, 89, 134, 136, 14, 217, 134, 185, 165, 153, 165, 249, 165, 89, 166, 153, 181, 185, 188, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 77, 136, 77, 232, 77, 72, 78, 168, 93, 136, 93, 232, 93, 72, 94, 200, 72, 168, 124, 168, 133, 136, 133, 232, 133, 136, 134, 72, 134, 128, 14, 200, 134, 168, 165, 136, 165, 232, 165, 72, 166, 136, 181, 168, 188, 72, 182, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 249, 85, 232, 85, 217, 85, 200, 85, 185, 86, 168, 86, 249, 86, 232, 86, 249, 89, 232, 89, 89, 91, 72, 91, 154, 94, 136, 94, 58, 95, 40, 95, 218, 93, 200, 93, 26, 95, 8, 95, 186, 94, 168, 96, 6, 8, 8, 8, 250, 101, 232, 101, 249, 106, 232, 106, 168, 144, 153, 109, 136, 109, 249, 110, 232, 110, 25, 110, 8, 110, 121, 111, 104, 111, 89, 111, 72, 111, 154, 121, 136, 121, 74, 4, 76, 4, 26, 122, 8, 122, 10, 8, 185, 128, 168, 128, 57, 135, 40, 135, 217, 133, 200, 133, 121, 134, 104, 134, 12, 8, 14, 8, 153, 149, 136, 149, 249, 150, 232, 150, 25, 150, 8, 150, 154, 153, 136, 153, 250, 153, 232, 153, 250, 154, 232, 154, 26, 154, 8, 154, 249, 158, 232, 158, 25, 158, 8, 158, 185, 160, 168, 160, 153, 166, 136, 166, 57, 167, 40, 167, 217, 165, 200, 165, 57, 166, 40, 166, 121, 166, 104, 166, 25, 167, 8, 167, 249, 173, 232, 173, 249, 181, 232, 181, 89, 182, 153, 185, 136, 185, 185, 186, 168, 186, 25, 186, 8, 186, 232, 152, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 152, 169, 152, 200, 56, 169, 76, 187, 92, 187, 100, 169, 92, 169, 100, 169, 192, 169, 120, 217, 132, 187, 76, 200, 132, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 72, 4, 186, 68, 104, 4, 177, 68, 36, 5, 250, 69, 37, 5, 241, 69, 255, 5, 185, 108, 76, 4, 186, 112, 108, 4, 177, 112, 57, 5, 154, 113, 58, 5, 145, 113, 59, 5, 250, 114, 60, 5, 241, 114, 61, 5, 26, 114, 62, 5, 17, 114, 255, 5, 186, 120, 255, 5, 168, 64, 104, 4, 168, 68, 37, 5, 232, 69, 255, 5, 168, 108, 108, 4, 168, 112, 58, 5, 136, 113, 60, 5, 232, 114, 62, 5, 8, 114, 255, 5, 168, 120, 255, 5, 250, 122, 255, 5, 232, 122, 255, 1, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 183, 16, 0, 16, 41, 0, 0, 96, 1, 0, 0, 96, 3, 0, 51, 0, 82, 0, 91, 0, 57, 9, 0, 32, 175, 16, 0, 32, 252, 24, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 1, 25, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 21, 0, 0, 32, 7, 0, 0, 0, 13, 0, 0, 0, 23, 0, 0, 0, 23, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 21, 0, 0, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 100, 101, 102, 97, 117, 108, 116, 0, 170, 0, 0, 116, 0, 114, 0, 97, 0, 100, 0, 105, 0, 116, 0, 105, 0, 111, 0, 110, 0, 97, 0, 108, 0, 0, 0, 1, 0, 43, 0, 1, 0, 1, 0, 32, 0, 13, 0, 0, 80, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 73, 10, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 77, 10, 0, 0, 77, 10, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 58, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 66, 220, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 84, 0, 60, 0, 97, 1, 60, 0, 60, 0, 60, 0, 96, 1, 60, 0, 122, 0, 60, 0, 60, 0, 60, 0, 90, 0, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 88, 0, 60, 0, 245, 0, 60, 0, 60, 0, 60, 0, 213, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 252, 0, 60, 0, 60, 0, 60, 0, 220, 0, 0, 0, 170, 170, 170, 170, 22, 40, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 116, 30, 0, 0, 116, 30, 0, 0, 116, 30, 0, 0, 116, 30, 0, 0, 96, 33, 0, 0, 96, 33, 0, 0, 254, 35, 0, 0, 254, 35, 0, 0, 254, 39, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 218, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 89, 0, 97, 0, 105, 0, 113, 0, 118, 0, 126, 0, 129, 0, 137, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 161, 0, 48, 0, 166, 0, 174, 0, 180, 0, 188, 0, 193, 0, 201, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 209, 0, 153, 0, 48, 0, 48, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 145, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 100, 1, 164, 1, 216, 1, 4, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 22, 0, 202, 65, 22, 0, 202, 98, 22, 0, 202, 131, 22, 0, 202, 164, 22, 0, 202, 197, 22, 0, 202, 230, 22, 0, 202, 7, 23, 0, 202, 40, 23, 0, 202, 73, 23, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 230, 0, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 70, 2, 0, 156, 5, 0, 80, 201, 38, 3, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 70, 4, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 38, 5, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 6, 6, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 102, 7, 0, 5, 5, 0, 80, 201, 70, 8, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 134, 9, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 134, 5, 14, 87, 196, 146, 156, 42, 197, 98, 21, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 134, 5, 13, 87, 134, 5, 15, 87, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 5, 16, 87, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 21, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 5, 5, 14, 87, 196, 146, 5, 42, 197, 226, 21, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 5, 5, 13, 87, 5, 5, 15, 87, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 5, 16, 87, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 10, 0, 201, 38, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 11, 0, 201, 166, 12, 0, 134, 5, 35, 79, 5, 5, 35, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 13, 0, 201, 38, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 2, 0, 197, 98, 2, 0, 197, 162, 2, 0, 197, 226, 2, 0, 134, 5, 37, 79, 5, 5, 37, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 14, 0, 201, 6, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 17, 0, 201, 230, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 197, 226, 4, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 192, 0, 0, 0, 197, 34, 6, 0, 197, 98, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 18, 0, 201, 198, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 20, 0, 201, 166, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 8, 0, 197, 98, 8, 0, 197, 162, 8, 0, 197, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 22, 0, 201, 38, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 98, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 23, 0, 201, 166, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 197, 98, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 25, 0, 201, 38, 26, 0, 201, 230, 26, 0, 201, 166, 27, 0, 201, 102, 28, 0, 201, 38, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 13, 0, 197, 98, 13, 0, 201, 230, 29, 0, 201, 166, 30, 0, 201, 102, 31, 0, 201, 38, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 32, 0, 201, 166, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 34, 0, 201, 134, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 3, 0, 197, 162, 3, 0, 201, 166, 36, 0, 201, 198, 37, 0, 201, 230, 38, 0, 201, 166, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 40, 0, 201, 38, 41, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 16, 87, 5, 136, 0, 0, 5, 5, 16, 87, 5, 136, 0, 0, 134, 5, 14, 87, 5, 162, 0, 0, 5, 5, 14, 87, 5, 162, 0, 0, 134, 5, 35, 79, 5, 160, 0, 0, 5, 5, 35, 79, 5, 160, 0, 0, 134, 5, 16, 87, 5, 162, 0, 0, 5, 5, 16, 87, 5, 162, 0, 0, 134, 5, 36, 79, 5, 136, 0, 0, 5, 5, 36, 79, 5, 136, 0, 0, 134, 5, 36, 79, 5, 156, 0, 0, 5, 5, 36, 79, 5, 156, 0, 0, 134, 5, 13, 87, 5, 190, 0, 0, 134, 5, 15, 87, 5, 190, 0, 0, 5, 5, 13, 87, 5, 190, 0, 0, 5, 5, 15, 87, 5, 190, 0, 0, 134, 5, 16, 87, 5, 190, 0, 0, 5, 5, 16, 87, 5, 190, 0, 0, 134, 5, 16, 87, 5, 164, 0, 0, 5, 5, 16, 87, 5, 164, 0, 0, 134, 5, 16, 87, 5, 144, 0, 0, 5, 5, 16, 87, 5, 144, 0, 0, 134, 5, 16, 87, 5, 138, 0, 0, 5, 5, 16, 87, 5, 138, 0, 0, 134, 5, 14, 87, 5, 164, 0, 0, 5, 5, 14, 87, 5, 164, 0, 0, 134, 5, 13, 87, 5, 162, 0, 0, 134, 5, 15, 87, 5, 162, 0, 0, 5, 5, 13, 87, 5, 162, 0, 0, 5, 5, 15, 87, 5, 162, 0, 0, 134, 5, 35, 79, 5, 202, 0, 0, 5, 5, 35, 79, 5, 202, 0, 0, 134, 5, 15, 87, 5, 164, 0, 0, 5, 5, 15, 87, 5, 164, 0, 0, 134, 5, 13, 87, 5, 164, 0, 0, 5, 5, 13, 87, 5, 164, 0, 0, 134, 5, 14, 87, 5, 200, 0, 0, 5, 5, 14, 87, 5, 200, 0, 0, 134, 5, 13, 87, 5, 136, 0, 0, 5, 5, 13, 87, 5, 136, 0, 0, 134, 5, 13, 87, 5, 150, 0, 0, 5, 5, 13, 87, 5, 150, 0, 0, 134, 5, 35, 79, 5, 196, 0, 0, 5, 5, 35, 79, 5, 196, 0, 0, 134, 5, 35, 79, 5, 156, 0, 0, 5, 5, 35, 79, 5, 156, 0, 0, 134, 5, 16, 87, 5, 198, 0, 0, 5, 5, 16, 87, 5, 198, 0, 0, 134, 5, 16, 87, 5, 208, 0, 0, 5, 5, 16, 87, 5, 208, 0, 0, 134, 5, 16, 87, 5, 204, 0, 0, 5, 5, 16, 87, 5, 204, 0, 0, 134, 5, 36, 79, 5, 142, 0, 0, 5, 5, 36, 79, 5, 142, 0, 0, 134, 5, 36, 79, 5, 196, 0, 0, 134, 5, 37, 79, 5, 196, 0, 0, 5, 5, 36, 79, 5, 196, 0, 0, 5, 5, 37, 79, 5, 196, 0, 0, 134, 5, 36, 79, 5, 210, 0, 0, 134, 5, 37, 79, 5, 210, 0, 0, 5, 5, 36, 79, 5, 210, 0, 0, 5, 5, 37, 79, 5, 210, 0, 0, 134, 5, 14, 87, 5, 196, 0, 0, 5, 5, 14, 87, 5, 196, 0, 0, 134, 5, 13, 87, 5, 196, 0, 0, 134, 5, 15, 87, 5, 196, 0, 0, 5, 5, 13, 87, 5, 196, 0, 0, 5, 5, 15, 87, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 13, 87, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 15, 87, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 13, 87, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 15, 87, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 16, 87, 5, 196, 0, 0, 5, 5, 16, 87, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 16, 87, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 16, 87, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 48, 0, 8, 3, 255, 255, 14, 87, 134, 5, 0, 70, 156, 5, 1, 0, 3, 3, 255, 255, 13, 87, 134, 5, 8, 3, 255, 255, 15, 87, 134, 5, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 35, 79, 134, 5, 0, 82, 156, 5, 1, 0, 8, 3, 255, 255, 16, 87, 134, 5, 68, 3, 197, 162, 36, 79, 134, 5, 48, 0, 12, 3, 255, 255, 37, 79, 134, 5, 0, 42, 5, 5, 48, 0, 8, 3, 255, 255, 14, 87, 5, 5, 0, 70, 5, 5, 1, 0, 3, 3, 255, 255, 13, 87, 5, 5, 8, 3, 255, 255, 15, 87, 5, 5, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 35, 79, 5, 5, 0, 82, 5, 5, 1, 0, 8, 3, 255, 255, 16, 87, 5, 5, 68, 3, 0, 192, 197, 98, 36, 79, 5, 5, 48, 0, 12, 3, 255, 255, 37, 79, 5, 5, 156, 42, 196, 162, 48, 0, 8, 3, 0, 192, 197, 162, 5, 42, 196, 162, 48, 0, 8, 3, 0, 192, 197, 226, 156, 78, 196, 160, 48, 0, 12, 3, 1, 192, 197, 34, 5, 78, 196, 160, 48, 0, 12, 3, 1, 192, 197, 98, 156, 82, 196, 162, 48, 0, 8, 3, 1, 192, 197, 162, 5, 82, 196, 162, 48, 0, 8, 3, 1, 192, 197, 226, 156, 70, 196, 190, 1, 0, 3, 3, 3, 192, 197, 34, 8, 3, 3, 192, 197, 98, 5, 70, 196, 190, 1, 0, 3, 3, 3, 192, 197, 162, 8, 3, 3, 192, 197, 226, 156, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 34, 5, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 98, 156, 70, 196, 162, 1, 0, 3, 3, 6, 192, 197, 162, 8, 3, 6, 192, 197, 226, 5, 70, 196, 162, 1, 0, 3, 3, 7, 192, 197, 34, 8, 3, 7, 192, 197, 98, 156, 78, 196, 202, 48, 0, 12, 3, 7, 192, 197, 162, 5, 78, 196, 202, 48, 0, 12, 3, 7, 192, 197, 226, 156, 42, 196, 200, 48, 0, 8, 3, 9, 192, 197, 34, 5, 42, 196, 200, 48, 0, 8, 3, 9, 192, 197, 98, 156, 78, 196, 196, 48, 0, 12, 3, 10, 192, 197, 162, 5, 78, 196, 196, 48, 0, 12, 3, 10, 192, 197, 226, 156, 82, 196, 198, 48, 0, 8, 3, 11, 192, 197, 162, 5, 82, 196, 198, 48, 0, 8, 3, 11, 192, 197, 226, 156, 82, 196, 208, 48, 0, 8, 3, 12, 192, 197, 34, 5, 82, 196, 208, 48, 0, 8, 3, 12, 192, 197, 98, 156, 82, 196, 204, 48, 0, 8, 3, 12, 192, 197, 162, 5, 82, 196, 204, 48, 0, 8, 3, 12, 192, 197, 226, 13, 0, 197, 162, 48, 0, 12, 3, 13, 192, 197, 226, 14, 0, 197, 34, 48, 0, 12, 3, 14, 192, 197, 98, 14, 0, 197, 162, 48, 0, 12, 3, 14, 192, 197, 226, 15, 0, 197, 34, 48, 0, 12, 3, 15, 192, 197, 98, 156, 42, 196, 196, 48, 0, 8, 3, 15, 192, 197, 162, 5, 42, 196, 196, 48, 0, 8, 3, 15, 192, 197, 226, 156, 70, 196, 196, 1, 0, 3, 3, 16, 192, 197, 34, 8, 3, 16, 192, 197, 98, 5, 70, 196, 196, 1, 0, 3, 3, 16, 192, 197, 162, 8, 3, 16, 192, 197, 226, 17, 0, 197, 35, 1, 0, 3, 3, 17, 192, 197, 131, 8, 3, 17, 192, 197, 227, 18, 0, 197, 67, 1, 0, 3, 3, 18, 192, 197, 163, 8, 3, 19, 192, 197, 3, 156, 82, 196, 196, 48, 0, 8, 3, 19, 192, 197, 98, 5, 82, 196, 196, 48, 0, 8, 3, 19, 192, 197, 162, 19, 0, 197, 227, 48, 0, 8, 3, 20, 192, 197, 67, 20, 0, 197, 163, 48, 0, 8, 3, 21, 192, 197, 3, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 28, 4, 185, 124, 185, 128, 185, 136, 30, 4, 185, 156, 32, 4, 185, 168, 185, 172, 185, 192, 185, 196, 34, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 36, 4, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 38, 4, 168, 124, 168, 128, 168, 136, 40, 4, 168, 156, 42, 4, 168, 168, 168, 172, 168, 192, 168, 196, 44, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 185, 180, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 154, 114, 185, 121, 153, 121, 249, 121, 185, 176, 185, 184, 136, 14, 217, 122, 185, 165, 153, 165, 249, 165, 185, 188, 153, 197, 185, 200, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 168, 180, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 136, 121, 232, 121, 168, 176, 168, 184, 128, 14, 200, 122, 168, 165, 136, 165, 232, 165, 168, 188, 136, 197, 168, 200, 72, 198, 57, 59, 40, 59, 217, 57, 200, 57, 46, 4, 48, 4, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 250, 114, 232, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 57, 123, 40, 123, 217, 121, 200, 121, 121, 122, 104, 122, 12, 8, 14, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 50, 4, 52, 4, 185, 144, 168, 144, 249, 158, 232, 158, 25, 158, 8, 158, 185, 160, 168, 160, 153, 166, 136, 166, 57, 167, 40, 167, 217, 165, 200, 165, 57, 166, 40, 166, 121, 166, 104, 166, 54, 4, 56, 4, 249, 173, 232, 173, 249, 197, 232, 197, 89, 198, 153, 149, 136, 149, 185, 150, 168, 150, 185, 152, 168, 152, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 204, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 120, 255, 5, 186, 140, 255, 5, 185, 164, 255, 5, 185, 148, 255, 5, 168, 56, 255, 5, 168, 120, 255, 5, 168, 140, 255, 5, 168, 164, 255, 5, 168, 148, 255, 5, 25, 59, 255, 5, 8, 59, 255, 5, 250, 142, 255, 5, 232, 142, 255, 5, 25, 167, 255, 5, 8, 167, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 59, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 66, 10, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 71, 10, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 185, 7, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 189, 7, 0, 0, 189, 7, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 97, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 143, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 65, 0, 114, 0, 97, 0, 98, 0, 93, 0, 38, 0, 78, 6, 60, 0, 60, 0, 80, 6, 60, 0, 60, 0, 79, 6, 60, 0, 60, 0, 75, 6, 60, 0, 60, 0, 77, 6, 60, 0, 60, 0, 76, 6, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 39, 6, 60, 0, 34, 6, 38, 0, 39, 6, 60, 0, 60, 0, 113, 6, 60, 0, 33, 6, 60, 0, 60, 0, 35, 6, 60, 0, 60, 0, 114, 6, 60, 0, 60, 0, 37, 6, 60, 0, 60, 0, 115, 6, 60, 0, 60, 0, 36, 6, 60, 0, 60, 0, 204, 6, 84, 6, 60, 0, 60, 0, 60, 0, 73, 6, 84, 6, 60, 0, 60, 0, 60, 0, 38, 6, 38, 0, 169, 6, 60, 0, 60, 0, 42, 0, 170, 6, 171, 6, 67, 6, 172, 6, 173, 6, 174, 6, 38, 0, 207, 6, 60, 0, 71, 6, 60, 0, 60, 0, 213, 6, 60, 0, 60, 0, 193, 6, 60, 0, 60, 0, 41, 6, 60, 0, 60, 0, 195, 6, 60, 0, 60, 0, 192, 6, 60, 0, 60, 0, 190, 6, 38, 0, 204, 6, 60, 0, 60, 0, 42, 0, 73, 6, 210, 6, 74, 6, 208, 6, 209, 6, 205, 6, 206, 6, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 54, 29, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 15, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 60, 0, 0, 0, 64, 0, 0, 0, 64, 1, 0, 0, 48, 28, 0, 0, 48, 28, 0, 0, 112, 28, 0, 0, 112, 28, 0, 0, 188, 28, 0, 0, 188, 28, 0, 0, 30, 29, 0, 0, 2, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 39, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 166, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 96, 0, 104, 0, 109, 0, 117, 0, 125, 0, 133, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 157, 0, 149, 0, 48, 0, 48, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 141, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 128, 1, 180, 1, 244, 1, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 1, 0, 202, 65, 1, 0, 202, 98, 1, 0, 202, 131, 1, 0, 202, 164, 1, 0, 202, 197, 1, 0, 202, 230, 1, 0, 202, 7, 2, 0, 202, 40, 2, 0, 202, 73, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 102, 5, 5, 25, 102, 5, 70, 27, 102, 5, 74, 27, 102, 5, 72, 27, 102, 7, 75, 27, 102, 201, 6, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 72, 184, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 24, 234, 102, 201, 230, 1, 0, 201, 198, 2, 0, 201, 166, 3, 0, 194, 5, 210, 230, 194, 5, 212, 230, 194, 5, 211, 230, 192, 0, 0, 0, 194, 5, 209, 230, 194, 5, 208, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 27, 102, 5, 73, 27, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 184, 102, 5, 71, 184, 102, 5, 73, 184, 102, 5, 74, 184, 102, 5, 75, 184, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 192, 0, 0, 0, 198, 65, 0, 0, 198, 97, 0, 0, 198, 130, 0, 0, 198, 193, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 4, 0, 5, 75, 240, 102, 5, 76, 240, 102, 192, 0, 0, 0, 5, 73, 240, 102, 5, 74, 240, 102, 5, 71, 240, 102, 197, 34, 0, 0, 192, 0, 0, 0, 201, 102, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 5, 0, 72, 0, 24, 234, 102, 0, 5, 0, 75, 0, 24, 234, 102, 0, 5, 0, 74, 0, 24, 234, 102, 0, 5, 0, 71, 0, 24, 234, 102, 0, 5, 0, 71, 0, 24, 234, 102, 0, 5, 128, 232, 0, 0, 0, 0, 0, 5, 0, 73, 0, 24, 234, 102, 0, 5, 0, 70, 0, 24, 234, 102, 0, 0, 0, 0, 5, 71, 240, 102, 194, 5, 128, 232, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 25, 102, 5, 5, 84, 6, 255, 255, 27, 102, 5, 70, 85, 6, 255, 255, 27, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 27, 102, 5, 74, 240, 102, 5, 70, 48, 0, 84, 6, 255, 255, 27, 102, 6, 75, 240, 102, 5, 72, 48, 0, 84, 6, 255, 255, 27, 102, 7, 75, 240, 102, 5, 5, 48, 0, 84, 6, 255, 255, 27, 102, 5, 75, 0, 0, 198, 225, 48, 0, 84, 6, 0, 192, 198, 65, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 99, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 178, 7, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 183, 7, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 128, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 132, 8, 0, 0, 132, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 30, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 112, 0, 115, 0, 93, 0, 0, 0, 170, 170, 170, 170, 100, 33, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 72, 0, 0, 0, 72, 1, 0, 0, 184, 28, 0, 0, 184, 28, 0, 0, 184, 28, 0, 0, 184, 28, 0, 0, 12, 29, 0, 0, 12, 29, 0, 0, 140, 29, 0, 0, 140, 29, 0, 0, 76, 33, 0, 0, 2, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 39, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 174, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 128, 0, 136, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 157, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 152, 0, 48, 0, 48, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 128, 1, 192, 1, 0, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 1, 0, 202, 129, 1, 0, 202, 162, 1, 0, 202, 195, 1, 0, 202, 228, 1, 0, 202, 5, 2, 0, 202, 38, 2, 0, 202, 71, 2, 0, 202, 104, 2, 0, 202, 137, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 102, 5, 5, 25, 102, 5, 70, 26, 102, 5, 70, 222, 102, 5, 73, 26, 102, 7, 75, 240, 102, 201, 6, 0, 0, 192, 0, 0, 0, 5, 73, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 71, 184, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 2, 0, 201, 166, 3, 0, 201, 134, 4, 0, 201, 102, 5, 0, 194, 5, 210, 230, 194, 5, 212, 230, 194, 5, 211, 230, 192, 0, 0, 0, 194, 5, 209, 230, 194, 5, 208, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 72, 26, 102, 5, 71, 26, 102, 5, 74, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 43, 102, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 102, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 85, 102, 5, 5, 85, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 111, 102, 192, 0, 0, 0, 5, 5, 111, 102, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 184, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 188, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 210, 102, 192, 0, 0, 0, 192, 0, 0, 0, 5, 72, 214, 102, 192, 0, 0, 0, 6, 5, 214, 102, 5, 71, 214, 102, 197, 34, 0, 0, 5, 74, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 71, 222, 102, 192, 0, 0, 0, 5, 72, 222, 102, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 6, 0, 5, 74, 240, 102, 192, 0, 0, 0, 192, 0, 0, 0, 5, 73, 240, 102, 192, 0, 0, 0, 5, 71, 240, 102, 197, 98, 0, 0, 192, 0, 0, 0, 201, 38, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 0, 4, 5, 71, 0, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 71, 214, 102, 194, 5, 128, 232, 5, 71, 240, 102, 194, 5, 128, 232, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 25, 102, 5, 5, 84, 6, 255, 255, 26, 102, 5, 70, 85, 6, 255, 255, 26, 102, 5, 73, 72, 102, 5, 5, 48, 0, 84, 6, 255, 255, 59, 102, 5, 70, 214, 102, 5, 5, 48, 0, 84, 6, 255, 255, 214, 102, 7, 5, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 240, 102, 5, 70, 48, 0, 84, 6, 255, 255, 240, 102, 6, 75, 240, 102, 5, 72, 48, 0, 84, 6, 255, 255, 240, 102, 7, 75, 240, 102, 5, 5, 48, 0, 84, 6, 255, 255, 240, 102, 5, 75, 214, 102, 5, 70, 48, 0, 84, 6, 255, 255, 214, 102, 6, 5, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 185, 120, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 185, 152, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 168, 120, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 168, 152, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 154, 114, 185, 121, 153, 121, 249, 121, 153, 122, 89, 122, 136, 14, 217, 122, 185, 153, 153, 153, 249, 153, 89, 154, 153, 169, 185, 176, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 136, 121, 232, 121, 136, 122, 72, 122, 128, 14, 200, 122, 168, 153, 136, 153, 232, 153, 72, 154, 136, 169, 168, 176, 72, 170, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 250, 114, 232, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 57, 123, 40, 123, 217, 121, 200, 121, 121, 122, 104, 122, 12, 8, 14, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 250, 142, 232, 142, 26, 142, 8, 142, 249, 146, 232, 146, 25, 146, 8, 146, 185, 148, 168, 148, 153, 154, 136, 154, 57, 155, 40, 155, 217, 153, 200, 153, 57, 154, 40, 154, 121, 154, 104, 154, 25, 155, 8, 155, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 185, 174, 168, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 31, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 121, 8, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 126, 8, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 63, 34, 0, 32, 7, 0, 0, 0, 26, 0, 0, 0, 67, 34, 0, 0, 67, 34, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 170, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 116, 114, 97, 100, 105, 116, 105, 111, 110, 97, 108, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 45, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 102, 0, 105, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 116, 0, 114, 0, 97, 0, 100, 0, 93, 0, 0, 0, 103, 220, 38, 0, 68, 0, 53, 3, 60, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 38, 0, 71, 0, 53, 3, 60, 0, 60, 0, 229, 1, 60, 0, 60, 0, 60, 0, 228, 1, 38, 0, 78, 0, 53, 3, 60, 0, 60, 0, 75, 1, 60, 0, 60, 0, 60, 0, 74, 1, 38, 0, 84, 0, 53, 3, 60, 0, 60, 0, 103, 1, 60, 0, 60, 0, 60, 0, 102, 1, 38, 0, 89, 0, 60, 0, 60, 0, 252, 0, 60, 0, 60, 0, 60, 0, 220, 0, 38, 0, 90, 0, 53, 3, 60, 0, 60, 0, 146, 2, 60, 0, 60, 0, 60, 0, 183, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 192, 1, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 0, 0, 128, 220, 38, 0, 68, 0, 60, 0, 60, 0, 240, 0, 60, 0, 60, 0, 60, 0, 208, 0, 60, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 38, 0, 78, 0, 60, 0, 60, 0, 75, 1, 60, 0, 60, 0, 60, 0, 74, 1, 38, 0, 84, 0, 60, 0, 60, 0, 254, 0, 47, 0, 104, 0, 60, 0, 60, 0, 60, 0, 222, 0, 47, 0, 104, 0, 38, 0, 86, 0, 60, 0, 60, 0, 119, 0, 60, 0, 60, 0, 60, 0, 87, 0, 38, 0, 89, 0, 60, 0, 60, 0, 252, 0, 60, 0, 60, 0, 60, 0, 220, 0, 60, 0, 60, 0, 113, 1, 60, 0, 60, 0, 60, 0, 112, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 192, 1, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 60, 0, 60, 0, 81, 1, 60, 0, 60, 0, 60, 0, 80, 1, 60, 0, 60, 0, 245, 0, 60, 0, 60, 0, 60, 0, 213, 0, 60, 0, 60, 0, 83, 1, 60, 0, 60, 0, 60, 0, 82, 1, 0, 0, 170, 170, 170, 170, 6, 50, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 202, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 116, 39, 0, 0, 116, 39, 0, 0, 148, 42, 0, 0, 148, 42, 0, 0, 240, 46, 0, 0, 240, 46, 0, 0, 234, 49, 0, 0, 234, 49, 0, 0, 234, 49, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 106, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 94, 0, 102, 0, 108, 0, 116, 0, 121, 0, 129, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 149, 0, 157, 0, 163, 0, 171, 0, 179, 0, 187, 0, 48, 0, 56, 0, 186, 0, 194, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 211, 0, 219, 0, 48, 0, 48, 0, 48, 0, 227, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 235, 0, 48, 0, 48, 0, 243, 0, 48, 0, 249, 0, 1, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 4, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 12, 1, 48, 0, 17, 1, 21, 1, 29, 1, 37, 1, 42, 1, 50, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 56, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 64, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 67, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 209, 0, 48, 0, 48, 0, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 97, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 75, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 83, 1, 48, 0, 48, 0, 89, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 120, 1, 176, 1, 228, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 84, 2, 140, 2, 204, 2, 192, 0, 232, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 33, 0, 202, 193, 33, 0, 202, 226, 33, 0, 202, 3, 34, 0, 202, 36, 34, 0, 202, 69, 34, 0, 202, 102, 34, 0, 202, 135, 34, 0, 202, 168, 34, 0, 202, 201, 34, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 198, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 166, 1, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 38, 3, 0, 156, 5, 0, 84, 134, 70, 0, 84, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 198, 4, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 38, 6, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 198, 7, 0, 5, 5, 0, 84, 5, 70, 0, 84, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 1, 0, 0, 198, 33, 0, 0, 198, 225, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 70, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 198, 65, 0, 0, 198, 97, 0, 0, 5, 5, 127, 12, 198, 1, 1, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 2, 25, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 193, 3, 111, 93, 193, 2, 111, 93, 198, 33, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 70, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 198, 161, 0, 0, 193, 4, 111, 93, 5, 5, 125, 12, 198, 65, 1, 0, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 9, 0, 201, 166, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 48, 5, 71, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 68, 5, 70, 0, 68, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 129, 0, 0, 198, 193, 0, 0, 198, 225, 1, 0, 198, 1, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 198, 11, 0, 201, 230, 12, 0, 197, 162, 2, 0, 197, 226, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 14, 0, 201, 134, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 17, 0, 201, 38, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 4, 0, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 192, 0, 0, 0, 198, 98, 3, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 3, 0, 198, 226, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 19, 0, 201, 198, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 5, 0, 197, 98, 6, 0, 198, 162, 5, 0, 198, 226, 5, 0, 198, 34, 6, 0, 198, 98, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 6, 0, 197, 162, 6, 0, 198, 226, 6, 0, 198, 34, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 70, 22, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 24, 0, 5, 71, 238, 102, 201, 6, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 8, 0, 192, 0, 0, 0, 197, 98, 8, 0, 192, 0, 0, 0, 197, 162, 8, 0, 192, 0, 0, 0, 197, 226, 8, 0, 192, 0, 0, 0, 197, 34, 9, 0, 197, 99, 9, 0, 197, 194, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 10, 0, 197, 67, 10, 0, 197, 162, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 226, 6, 0, 197, 34, 11, 0, 5, 5, 10, 121, 197, 98, 11, 0, 197, 162, 11, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 7, 0, 5, 5, 24, 121, 197, 162, 7, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 25, 0, 201, 6, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 7, 0, 198, 34, 8, 0, 198, 98, 8, 0, 198, 162, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 28, 0, 201, 70, 29, 0, 201, 102, 30, 0, 201, 134, 31, 0, 201, 166, 32, 0, 201, 198, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 17, 0, 197, 98, 17, 0, 197, 162, 17, 0, 197, 226, 17, 0, 197, 34, 18, 0, 197, 98, 18, 0, 197, 162, 18, 0, 197, 226, 18, 0, 197, 34, 19, 0, 197, 98, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 34, 0, 201, 6, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 37, 0, 201, 166, 38, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 2, 0, 198, 226, 2, 0, 201, 38, 40, 0, 201, 166, 41, 0, 201, 38, 43, 0, 201, 70, 44, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 45, 0, 201, 134, 46, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 72, 0, 4, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 71, 0, 4, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 73, 0, 4, 111, 93, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 134, 70, 0, 80, 5, 5, 0, 56, 5, 70, 0, 80, 5, 5, 0, 56, 193, 3, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 84, 5, 142, 0, 0, 5, 70, 0, 84, 5, 142, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 3, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 164, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 193, 3, 111, 93, 5, 200, 0, 0, 193, 2, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 134, 70, 0, 84, 5, 138, 0, 0, 5, 70, 0, 84, 5, 138, 0, 0, 134, 70, 0, 84, 5, 136, 0, 0, 5, 70, 0, 84, 5, 136, 0, 0, 134, 70, 0, 84, 5, 150, 0, 0, 5, 70, 0, 84, 5, 150, 0, 0, 134, 70, 0, 84, 5, 156, 0, 0, 5, 70, 0, 84, 5, 156, 0, 0, 134, 70, 0, 84, 5, 196, 0, 0, 5, 70, 0, 84, 5, 196, 0, 0, 5, 70, 0, 84, 5, 146, 0, 0, 193, 3, 111, 93, 5, 196, 0, 0, 193, 2, 111, 93, 5, 196, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 6, 121, 197, 226, 6, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 34, 7, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 98, 7, 0, 5, 5, 24, 121, 197, 162, 7, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 226, 7, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 34, 8, 0, 5, 5, 107, 121, 197, 98, 8, 0, 5, 5, 111, 121, 197, 162, 8, 0, 5, 5, 115, 121, 197, 226, 8, 0, 5, 5, 119, 121, 197, 34, 9, 0, 197, 99, 9, 0, 197, 194, 9, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 2, 10, 0, 197, 67, 10, 0, 197, 162, 10, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 226, 10, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 226, 6, 0, 197, 34, 11, 0, 5, 5, 10, 121, 197, 98, 11, 0, 197, 162, 11, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 7, 0, 5, 5, 24, 121, 197, 162, 7, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 1, 0, 8, 3, 198, 129, 10, 3, 198, 161, 0, 70, 156, 5, 2, 0, 3, 3, 0, 192, 198, 65, 8, 3, 0, 192, 198, 97, 11, 3, 0, 192, 198, 129, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 1, 0, 8, 3, 255, 255, 111, 93, 193, 3, 10, 3, 255, 255, 111, 93, 193, 2, 0, 70, 5, 5, 2, 0, 3, 3, 0, 192, 198, 161, 8, 3, 255, 255, 111, 93, 193, 4, 11, 3, 0, 192, 198, 193, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 1, 192, 198, 98, 10, 3, 1, 192, 198, 162, 5, 42, 196, 162, 1, 0, 8, 3, 1, 192, 197, 34, 10, 3, 1, 192, 197, 98, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 162, 11, 3, 1, 192, 197, 226, 5, 82, 196, 162, 1, 0, 8, 3, 2, 192, 197, 34, 11, 3, 2, 192, 197, 98, 156, 70, 196, 190, 2, 0, 3, 3, 2, 192, 198, 34, 8, 3, 2, 192, 198, 98, 11, 3, 2, 192, 198, 162, 5, 70, 196, 190, 2, 0, 3, 3, 2, 192, 198, 226, 8, 3, 3, 192, 197, 34, 11, 3, 3, 192, 198, 34, 156, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 98, 11, 3, 3, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 226, 11, 3, 4, 192, 197, 34, 156, 70, 196, 162, 2, 0, 3, 3, 4, 192, 198, 34, 8, 3, 4, 192, 198, 98, 11, 3, 4, 192, 198, 162, 5, 70, 196, 162, 2, 0, 3, 3, 4, 192, 198, 226, 8, 3, 6, 192, 197, 34, 11, 3, 5, 192, 198, 34, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 1, 0, 8, 3, 7, 192, 198, 98, 10, 3, 7, 192, 198, 162, 5, 42, 196, 200, 1, 0, 8, 3, 13, 192, 197, 162, 10, 3, 13, 192, 197, 226, 156, 82, 196, 198, 1, 0, 8, 3, 14, 192, 197, 34, 11, 3, 14, 192, 197, 98, 5, 82, 196, 198, 1, 0, 8, 3, 14, 192, 197, 162, 11, 3, 14, 192, 197, 226, 156, 82, 196, 208, 1, 0, 8, 3, 15, 192, 197, 34, 11, 3, 15, 192, 197, 98, 5, 82, 196, 208, 1, 0, 8, 3, 15, 192, 197, 162, 11, 3, 15, 192, 197, 226, 156, 82, 196, 204, 1, 0, 8, 3, 16, 192, 197, 34, 11, 3, 16, 192, 197, 98, 5, 82, 196, 204, 1, 0, 8, 3, 16, 192, 197, 162, 11, 3, 16, 192, 197, 226, 156, 42, 196, 196, 1, 0, 8, 3, 8, 192, 198, 226, 10, 3, 9, 192, 198, 34, 5, 42, 196, 196, 1, 0, 8, 3, 19, 192, 197, 226, 10, 3, 20, 192, 197, 34, 156, 70, 196, 196, 2, 0, 3, 3, 9, 192, 198, 98, 8, 3, 9, 192, 198, 162, 11, 3, 9, 192, 198, 226, 5, 70, 196, 196, 2, 0, 3, 3, 10, 192, 198, 34, 8, 3, 20, 192, 197, 98, 11, 3, 10, 192, 198, 98, 20, 0, 197, 163, 2, 0, 3, 3, 10, 192, 198, 163, 8, 3, 11, 192, 198, 3, 11, 3, 11, 192, 198, 99, 21, 0, 197, 3, 2, 0, 3, 3, 11, 192, 198, 195, 8, 3, 21, 192, 197, 99, 11, 3, 12, 192, 198, 35, 156, 82, 196, 196, 1, 0, 8, 3, 21, 192, 197, 194, 11, 3, 22, 192, 197, 2, 5, 82, 196, 196, 1, 0, 8, 3, 22, 192, 197, 66, 11, 3, 22, 192, 197, 130, 22, 0, 197, 195, 1, 0, 8, 3, 23, 192, 197, 35, 11, 3, 23, 192, 197, 131, 23, 0, 197, 227, 1, 0, 8, 3, 24, 192, 197, 67, 11, 3, 24, 192, 197, 163, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 171, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 92, 40, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 36, 31, 0, 0, 36, 31, 0, 0, 148, 32, 0, 0, 148, 32, 0, 0, 124, 34, 0, 0, 124, 34, 0, 0, 84, 36, 0, 0, 84, 36, 0, 0, 64, 40, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 229, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 94, 0, 102, 0, 107, 0, 115, 0, 121, 0, 129, 0, 137, 0, 145, 0, 48, 0, 56, 0, 149, 0, 157, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 181, 0, 48, 0, 48, 0, 185, 0, 48, 0, 193, 0, 198, 0, 206, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 212, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 220, 0, 173, 0, 48, 0, 48, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 120, 1, 172, 1, 228, 1, 36, 2, 192, 0, 84, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 14, 0, 202, 33, 14, 0, 202, 66, 14, 0, 202, 99, 14, 0, 202, 132, 14, 0, 202, 165, 14, 0, 202, 198, 14, 0, 202, 231, 14, 0, 202, 8, 15, 0, 202, 41, 15, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 230, 0, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 166, 1, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 198, 2, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 38, 4, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 6, 5, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 1, 0, 0, 198, 33, 0, 0, 198, 97, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 65, 0, 0, 5, 5, 127, 12, 198, 129, 0, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 160, 5, 38, 93, 197, 194, 13, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 193, 3, 111, 93, 193, 2, 111, 93, 198, 161, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 193, 4, 111, 93, 5, 5, 125, 12, 198, 193, 0, 0, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 6, 0, 201, 102, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 1, 0, 197, 98, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 1, 0, 197, 226, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 2, 0, 197, 98, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 8, 0, 201, 70, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 10, 0, 201, 198, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 11, 0, 201, 70, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 162, 4, 0, 197, 226, 4, 0, 197, 34, 5, 0, 197, 98, 5, 0, 192, 0, 0, 0, 198, 162, 1, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 1, 0, 198, 34, 2, 0, 197, 226, 5, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 13, 0, 201, 198, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 6, 0, 197, 3, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 2, 0, 197, 98, 7, 0, 198, 226, 2, 0, 198, 34, 3, 0, 198, 98, 3, 0, 198, 162, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 3, 0, 197, 162, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 14, 0, 201, 166, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 16, 0, 201, 134, 17, 0, 201, 70, 18, 0, 201, 6, 19, 0, 201, 198, 19, 0, 201, 134, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 21, 0, 201, 102, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 23, 0, 201, 70, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 25, 0, 201, 198, 25, 0, 201, 134, 26, 0, 201, 70, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 28, 0, 201, 198, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 193, 3, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 162, 0, 0, 156, 5, 0, 48, 194, 6, 6, 179, 28, 5, 0, 48, 194, 5, 6, 179, 156, 5, 0, 68, 194, 6, 4, 179, 28, 5, 0, 68, 194, 5, 4, 179, 156, 5, 0, 80, 194, 6, 3, 179, 28, 5, 0, 80, 194, 5, 3, 179, 134, 70, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 156, 5, 0, 92, 194, 6, 2, 179, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 3, 111, 93, 5, 164, 0, 0, 156, 5, 0, 54, 194, 6, 5, 179, 28, 5, 0, 54, 194, 5, 5, 179, 193, 4, 111, 93, 5, 162, 0, 0, 156, 5, 0, 92, 194, 6, 2, 179, 5, 144, 0, 0, 28, 5, 0, 92, 194, 5, 2, 179, 5, 144, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 164, 0, 0, 193, 3, 111, 93, 5, 200, 0, 0, 193, 2, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 193, 3, 111, 93, 5, 196, 0, 0, 193, 2, 111, 93, 5, 196, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 1, 0, 8, 3, 198, 129, 10, 3, 198, 161, 0, 70, 156, 5, 48, 0, 8, 3, 0, 192, 198, 65, 0, 82, 156, 5, 1, 0, 8, 3, 255, 255, 0, 90, 134, 70, 68, 3, 197, 162, 0, 42, 5, 5, 1, 0, 8, 3, 255, 255, 111, 93, 193, 3, 10, 3, 255, 255, 111, 93, 193, 2, 0, 70, 5, 5, 48, 0, 8, 3, 255, 255, 111, 93, 193, 4, 0, 82, 5, 5, 1, 0, 8, 3, 255, 255, 0, 90, 5, 70, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 0, 192, 198, 226, 10, 3, 1, 192, 198, 34, 5, 42, 196, 162, 1, 0, 8, 3, 0, 192, 197, 162, 10, 3, 0, 192, 197, 226, 156, 82, 196, 162, 48, 0, 8, 3, 2, 192, 197, 162, 5, 82, 196, 162, 48, 0, 8, 3, 2, 192, 197, 226, 156, 70, 196, 190, 48, 0, 8, 3, 1, 192, 198, 98, 5, 70, 196, 190, 48, 0, 8, 3, 3, 192, 197, 34, 156, 82, 196, 190, 48, 0, 8, 3, 3, 192, 197, 98, 5, 82, 196, 190, 48, 0, 8, 3, 3, 192, 197, 162, 156, 70, 196, 162, 48, 0, 8, 3, 2, 192, 198, 98, 5, 70, 196, 162, 48, 0, 8, 3, 6, 192, 197, 98, 156, 42, 196, 200, 1, 0, 8, 3, 4, 192, 198, 34, 10, 3, 4, 192, 198, 98, 5, 42, 196, 200, 1, 0, 8, 3, 7, 192, 197, 226, 10, 3, 8, 192, 197, 34, 156, 82, 196, 198, 48, 0, 8, 3, 8, 192, 197, 98, 5, 82, 196, 198, 48, 0, 8, 3, 8, 192, 197, 162, 156, 82, 196, 208, 48, 0, 8, 3, 8, 192, 197, 226, 5, 82, 196, 208, 48, 0, 8, 3, 9, 192, 197, 34, 156, 82, 196, 204, 48, 0, 8, 3, 9, 192, 197, 98, 5, 82, 196, 204, 48, 0, 8, 3, 9, 192, 197, 162, 156, 42, 196, 196, 1, 0, 8, 3, 4, 192, 198, 162, 10, 3, 4, 192, 198, 226, 5, 42, 196, 196, 1, 0, 8, 3, 9, 192, 197, 226, 10, 3, 10, 192, 197, 34, 156, 70, 196, 196, 48, 0, 8, 3, 5, 192, 198, 34, 5, 70, 196, 196, 48, 0, 8, 3, 10, 192, 197, 98, 10, 0, 197, 163, 48, 0, 8, 3, 5, 192, 198, 99, 11, 0, 197, 3, 48, 0, 8, 3, 11, 192, 197, 99, 156, 82, 196, 196, 48, 0, 8, 3, 11, 192, 197, 194, 5, 82, 196, 196, 48, 0, 8, 3, 12, 192, 197, 2, 12, 0, 197, 67, 48, 0, 8, 3, 12, 192, 197, 163, 13, 0, 197, 3, 48, 0, 8, 3, 13, 192, 197, 99, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 26, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 28, 4, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 30, 4, 185, 120, 185, 124, 185, 132, 186, 136, 185, 140, 32, 4, 185, 148, 185, 152, 185, 156, 185, 160, 185, 164, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 34, 4, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 36, 4, 168, 120, 168, 124, 168, 132, 168, 136, 168, 140, 38, 4, 168, 148, 168, 152, 168, 156, 168, 160, 168, 164, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 116, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 185, 180, 185, 176, 217, 180, 217, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 154, 114, 185, 117, 153, 117, 249, 117, 153, 118, 185, 184, 136, 14, 217, 184, 185, 145, 153, 145, 249, 145, 217, 160, 153, 161, 185, 168, 0, 8, 168, 57, 136, 57, 232, 57, 136, 58, 168, 180, 168, 176, 200, 180, 200, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 117, 136, 117, 232, 117, 136, 118, 168, 184, 128, 14, 200, 184, 168, 145, 136, 145, 232, 145, 200, 160, 136, 161, 168, 168, 72, 162, 25, 59, 8, 59, 217, 57, 200, 57, 40, 4, 42, 4, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 2, 8, 137, 71, 26, 75, 8, 75, 218, 73, 200, 73, 186, 74, 168, 74, 250, 74, 232, 74, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 217, 82, 200, 82, 249, 85, 232, 85, 57, 87, 40, 87, 154, 90, 136, 90, 26, 91, 8, 91, 218, 89, 200, 89, 250, 90, 232, 90, 186, 90, 168, 92, 4, 8, 6, 8, 250, 97, 232, 97, 217, 102, 200, 102, 168, 128, 153, 105, 136, 105, 217, 106, 200, 106, 25, 106, 8, 106, 185, 107, 168, 107, 57, 107, 40, 107, 154, 113, 136, 113, 218, 114, 200, 114, 26, 114, 8, 114, 8, 8, 10, 8, 106, 115, 25, 119, 8, 119, 217, 117, 200, 117, 121, 118, 104, 118, 12, 8, 14, 8, 153, 133, 136, 133, 217, 134, 200, 134, 25, 134, 8, 134, 154, 137, 136, 137, 250, 137, 232, 137, 218, 138, 200, 138, 26, 138, 8, 138, 217, 142, 200, 142, 25, 142, 8, 142, 16, 8, 73, 143, 153, 146, 136, 146, 25, 147, 8, 147, 217, 145, 200, 145, 57, 146, 40, 146, 121, 146, 104, 146, 44, 4, 46, 4, 249, 153, 232, 153, 249, 161, 232, 161, 89, 162, 153, 165, 136, 165, 185, 166, 168, 166, 25, 166, 8, 166, 232, 136, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 18, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 20, 8, 1, 0, 8, 14, 22, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 24, 8, 200, 12, 72, 12, 32, 14, 200, 136, 169, 136, 185, 68, 137, 3, 187, 88, 187, 96, 169, 88, 169, 96, 169, 172, 169, 112, 186, 112, 105, 3, 217, 116, 187, 72, 200, 116, 169, 72, 185, 140, 73, 3, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 116, 255, 5, 185, 144, 255, 5, 168, 56, 255, 5, 168, 116, 255, 5, 168, 144, 255, 5, 249, 58, 255, 5, 232, 58, 255, 5, 249, 146, 255, 5, 232, 146, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 51, 13, 0, 16, 53, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 132, 43, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 52, 31, 0, 0, 52, 31, 0, 0, 84, 34, 0, 0, 84, 34, 0, 0, 204, 36, 0, 0, 204, 36, 0, 0, 128, 39, 0, 0, 128, 39, 0, 0, 104, 43, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 230, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 94, 0, 102, 0, 108, 0, 116, 0, 121, 0, 129, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 169, 0, 48, 0, 174, 0, 178, 0, 186, 0, 194, 0, 199, 0, 207, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 213, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 221, 0, 161, 0, 48, 0, 48, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 120, 1, 176, 1, 228, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 128, 18, 0, 202, 161, 18, 0, 202, 194, 18, 0, 202, 227, 18, 0, 202, 4, 19, 0, 202, 37, 19, 0, 202, 70, 19, 0, 202, 103, 19, 0, 202, 136, 19, 0, 202, 169, 19, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 230, 0, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 102, 2, 0, 156, 5, 0, 84, 134, 70, 0, 84, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 6, 4, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 102, 5, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 6, 7, 0, 5, 5, 0, 84, 5, 70, 0, 84, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 1, 0, 0, 198, 33, 0, 0, 198, 225, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 70, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 198, 65, 0, 0, 198, 97, 0, 0, 5, 5, 127, 12, 198, 1, 1, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 66, 18, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 193, 3, 111, 93, 193, 2, 111, 93, 198, 33, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 70, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 198, 161, 0, 0, 193, 4, 111, 93, 5, 5, 125, 12, 198, 65, 1, 0, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 8, 0, 201, 230, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 48, 5, 71, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 68, 5, 70, 0, 68, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 129, 0, 0, 198, 193, 0, 0, 198, 225, 1, 0, 198, 1, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 6, 11, 0, 201, 38, 12, 0, 197, 162, 2, 0, 197, 226, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 13, 0, 201, 198, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 16, 0, 201, 102, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 4, 0, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 192, 0, 0, 0, 198, 98, 3, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 3, 0, 198, 226, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 18, 0, 201, 6, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 5, 0, 197, 98, 6, 0, 198, 162, 5, 0, 198, 226, 5, 0, 198, 34, 6, 0, 198, 98, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 6, 0, 197, 162, 6, 0, 198, 226, 6, 0, 198, 34, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 21, 0, 201, 166, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 7, 0, 198, 34, 8, 0, 198, 98, 8, 0, 198, 162, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 23, 0, 201, 230, 24, 0, 201, 6, 26, 0, 201, 38, 27, 0, 201, 70, 28, 0, 201, 102, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 197, 162, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 30, 0, 201, 166, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 32, 0, 201, 70, 34, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 2, 0, 198, 226, 2, 0, 201, 198, 35, 0, 201, 70, 37, 0, 201, 198, 38, 0, 201, 230, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 41, 0, 201, 38, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 72, 0, 4, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 71, 0, 4, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 73, 0, 4, 111, 93, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 72, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 134, 70, 0, 80, 5, 5, 0, 56, 5, 70, 0, 80, 5, 5, 0, 56, 193, 3, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 84, 5, 142, 0, 0, 5, 70, 0, 84, 5, 142, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 3, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 164, 0, 0, 193, 3, 111, 93, 5, 200, 0, 0, 193, 2, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 134, 70, 0, 84, 5, 138, 0, 0, 5, 70, 0, 84, 5, 138, 0, 0, 134, 70, 0, 84, 5, 136, 0, 0, 5, 70, 0, 84, 5, 136, 0, 0, 134, 70, 0, 84, 5, 150, 0, 0, 5, 70, 0, 84, 5, 150, 0, 0, 134, 70, 0, 84, 5, 156, 0, 0, 5, 70, 0, 84, 5, 156, 0, 0, 134, 70, 0, 84, 5, 196, 0, 0, 5, 70, 0, 84, 5, 196, 0, 0, 5, 70, 0, 84, 5, 146, 0, 0, 193, 3, 111, 93, 5, 196, 0, 0, 193, 2, 111, 93, 5, 196, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 1, 0, 8, 3, 198, 129, 10, 3, 198, 161, 0, 70, 156, 5, 2, 0, 3, 3, 0, 192, 198, 65, 8, 3, 0, 192, 198, 97, 11, 3, 0, 192, 198, 129, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 1, 0, 8, 3, 255, 255, 111, 93, 193, 3, 10, 3, 255, 255, 111, 93, 193, 2, 0, 70, 5, 5, 2, 0, 3, 3, 0, 192, 198, 161, 8, 3, 255, 255, 111, 93, 193, 4, 11, 3, 0, 192, 198, 193, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 1, 192, 198, 98, 10, 3, 1, 192, 198, 162, 5, 42, 196, 162, 1, 0, 8, 3, 1, 192, 197, 34, 10, 3, 1, 192, 197, 98, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 162, 11, 3, 1, 192, 197, 226, 5, 82, 196, 162, 1, 0, 8, 3, 2, 192, 197, 34, 11, 3, 2, 192, 197, 98, 156, 70, 196, 190, 2, 0, 3, 3, 2, 192, 198, 34, 8, 3, 2, 192, 198, 98, 11, 3, 2, 192, 198, 162, 5, 70, 196, 190, 2, 0, 3, 3, 2, 192, 198, 226, 8, 3, 3, 192, 197, 34, 11, 3, 3, 192, 198, 34, 156, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 98, 11, 3, 3, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 226, 11, 3, 4, 192, 197, 34, 156, 70, 196, 162, 2, 0, 3, 3, 4, 192, 198, 34, 8, 3, 4, 192, 198, 98, 11, 3, 4, 192, 198, 162, 5, 70, 196, 162, 2, 0, 3, 3, 4, 192, 198, 226, 8, 3, 6, 192, 197, 34, 11, 3, 5, 192, 198, 34, 156, 42, 196, 200, 1, 0, 8, 3, 7, 192, 198, 98, 10, 3, 7, 192, 198, 162, 5, 42, 196, 200, 1, 0, 8, 3, 6, 192, 197, 226, 10, 3, 7, 192, 197, 34, 156, 82, 196, 198, 1, 0, 8, 3, 7, 192, 197, 98, 11, 3, 7, 192, 197, 162, 5, 82, 196, 198, 1, 0, 8, 3, 7, 192, 197, 226, 11, 3, 8, 192, 197, 34, 156, 82, 196, 208, 1, 0, 8, 3, 8, 192, 197, 98, 11, 3, 8, 192, 197, 162, 5, 82, 196, 208, 1, 0, 8, 3, 8, 192, 197, 226, 11, 3, 9, 192, 197, 34, 156, 82, 196, 204, 1, 0, 8, 3, 9, 192, 197, 98, 11, 3, 9, 192, 197, 162, 5, 82, 196, 204, 1, 0, 8, 3, 9, 192, 197, 226, 11, 3, 10, 192, 197, 34, 156, 42, 196, 196, 1, 0, 8, 3, 8, 192, 198, 226, 10, 3, 9, 192, 198, 34, 5, 42, 196, 196, 1, 0, 8, 3, 13, 192, 197, 34, 10, 3, 13, 192, 197, 98, 156, 70, 196, 196, 2, 0, 3, 3, 9, 192, 198, 98, 8, 3, 9, 192, 198, 162, 11, 3, 9, 192, 198, 226, 5, 70, 196, 196, 2, 0, 3, 3, 10, 192, 198, 34, 8, 3, 13, 192, 197, 162, 11, 3, 10, 192, 198, 98, 13, 0, 197, 227, 2, 0, 3, 3, 10, 192, 198, 163, 8, 3, 11, 192, 198, 3, 11, 3, 11, 192, 198, 99, 14, 0, 197, 67, 2, 0, 3, 3, 11, 192, 198, 195, 8, 3, 14, 192, 197, 163, 11, 3, 12, 192, 198, 35, 156, 82, 196, 196, 1, 0, 8, 3, 15, 192, 197, 2, 11, 3, 15, 192, 197, 66, 5, 82, 196, 196, 1, 0, 8, 3, 15, 192, 197, 130, 11, 3, 15, 192, 197, 194, 16, 0, 197, 3, 1, 0, 8, 3, 16, 192, 197, 99, 11, 3, 16, 192, 197, 195, 17, 0, 197, 35, 1, 0, 8, 3, 17, 192, 197, 131, 11, 3, 17, 192, 197, 227, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 185, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 28, 4, 185, 120, 185, 124, 185, 132, 186, 136, 185, 140, 30, 4, 185, 152, 217, 152, 185, 156, 185, 160, 185, 164, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 32, 4, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 34, 4, 168, 120, 168, 124, 168, 132, 168, 136, 168, 140, 36, 4, 168, 152, 200, 152, 168, 156, 168, 160, 168, 164, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 116, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 121, 58, 185, 176, 185, 172, 217, 176, 185, 66, 185, 73, 153, 73, 249, 73, 89, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 122, 114, 185, 117, 153, 117, 249, 117, 25, 181, 185, 180, 136, 14, 217, 180, 185, 149, 153, 149, 249, 149, 217, 160, 153, 161, 0, 8, 2, 8, 168, 57, 136, 57, 232, 57, 104, 58, 168, 176, 168, 172, 200, 176, 168, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 104, 114, 168, 117, 136, 117, 232, 117, 8, 181, 168, 180, 128, 14, 200, 180, 168, 149, 136, 149, 232, 149, 200, 160, 136, 161, 4, 8, 72, 162, 249, 58, 232, 58, 217, 57, 200, 57, 38, 4, 40, 4, 153, 65, 136, 65, 249, 65, 232, 65, 153, 66, 136, 66, 25, 66, 8, 66, 25, 70, 8, 70, 249, 68, 232, 68, 249, 74, 232, 74, 217, 73, 200, 73, 153, 74, 136, 74, 217, 74, 200, 74, 25, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 153, 82, 136, 82, 185, 82, 168, 82, 249, 85, 232, 85, 25, 87, 8, 87, 122, 90, 104, 90, 250, 90, 232, 90, 218, 89, 200, 89, 218, 90, 200, 90, 154, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 185, 102, 168, 102, 168, 128, 153, 105, 136, 105, 185, 106, 168, 106, 25, 106, 8, 106, 57, 107, 40, 107, 25, 107, 8, 107, 154, 113, 136, 113, 186, 114, 168, 114, 26, 114, 8, 114, 10, 8, 217, 112, 200, 112, 249, 118, 232, 118, 217, 117, 200, 117, 249, 180, 232, 180, 57, 181, 40, 181, 153, 133, 136, 133, 185, 134, 168, 134, 25, 134, 8, 134, 154, 137, 136, 137, 250, 137, 232, 137, 186, 138, 168, 138, 26, 138, 8, 138, 185, 142, 168, 142, 25, 142, 8, 142, 185, 144, 168, 144, 121, 150, 104, 150, 249, 150, 232, 150, 217, 149, 200, 149, 57, 150, 40, 150, 249, 160, 232, 160, 42, 4, 44, 4, 12, 8, 14, 8, 249, 161, 232, 161, 89, 162, 153, 165, 136, 165, 153, 166, 136, 166, 25, 166, 8, 166, 232, 136, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 140, 168, 84, 200, 136, 169, 136, 200, 140, 168, 84, 187, 88, 187, 96, 169, 88, 169, 96, 169, 168, 169, 112, 217, 152, 232, 1, 200, 152, 232, 1, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 116, 255, 5, 185, 148, 255, 5, 168, 56, 255, 5, 168, 116, 255, 5, 168, 148, 255, 5, 217, 58, 255, 5, 200, 58, 255, 5, 217, 150, 255, 5, 200, 150, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 83, 23, 0, 16, 158, 0, 0, 96, 1, 0, 0, 96, 3, 0, 51, 0, 82, 0, 91, 0, 46, 13, 0, 32, 75, 23, 0, 32, 53, 34, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 58, 34, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 7, 8, 0, 0, 7, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 35, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 38, 0, 78, 0, 60, 0, 241, 0, 60, 0, 60, 0, 60, 0, 209, 0, 60, 0, 110, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 71, 0, 0, 0, 94, 31, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 0, 26, 0, 0, 0, 26, 0, 0, 32, 27, 0, 0, 58, 27, 0, 0, 70, 31, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 128, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 111, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 104, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 8, 0, 202, 65, 8, 0, 202, 98, 8, 0, 202, 131, 8, 0, 202, 164, 8, 0, 202, 197, 8, 0, 202, 230, 8, 0, 202, 7, 9, 0, 202, 40, 9, 0, 202, 73, 9, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 4, 0, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 196, 7, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 7, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 134, 5, 2, 69, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 7, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 7, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 5, 5, 2, 69, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 12, 0, 201, 198, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 13, 0, 201, 70, 14, 0, 201, 6, 15, 0, 201, 198, 15, 0, 201, 134, 16, 0, 201, 70, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 135, 5, 3, 69, 5, 142, 0, 0, 70, 5, 3, 69, 5, 142, 0, 0, 135, 5, 3, 69, 5, 140, 0, 0, 70, 5, 3, 69, 5, 140, 0, 0, 135, 5, 3, 69, 5, 156, 0, 0, 70, 5, 3, 69, 5, 156, 0, 0, 135, 5, 3, 69, 5, 160, 0, 0, 70, 5, 3, 69, 5, 160, 0, 0, 135, 5, 3, 69, 5, 144, 0, 0, 70, 5, 3, 69, 5, 144, 0, 0, 135, 5, 3, 69, 5, 136, 0, 0, 70, 5, 3, 69, 5, 136, 0, 0, 135, 5, 3, 69, 5, 164, 0, 0, 70, 5, 3, 69, 5, 164, 0, 0, 5, 5, 3, 69, 5, 142, 0, 0, 5, 5, 3, 69, 5, 140, 0, 0, 5, 5, 3, 69, 5, 156, 0, 0, 5, 5, 3, 69, 5, 160, 0, 0, 5, 5, 3, 69, 5, 144, 0, 0, 5, 5, 3, 69, 5, 136, 0, 0, 5, 5, 3, 69, 5, 164, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 68, 156, 5, 16, 0, 34, 1, 30, 0, 244, 1, 16, 0, 244, 1, 2, 192, 197, 162, 245, 1, 2, 192, 197, 226, 3, 3, 255, 255, 2, 69, 134, 5, 32, 30, 3, 192, 197, 34, 33, 30, 3, 192, 197, 98, 34, 1, 1, 192, 197, 162, 35, 1, 1, 192, 197, 226, 230, 1, 2, 192, 197, 34, 231, 1, 2, 192, 197, 98, 30, 1, 12, 0, 30, 1, 0, 192, 197, 162, 31, 1, 0, 192, 197, 226, 32, 1, 1, 192, 197, 34, 33, 1, 1, 192, 197, 98, 71, 0, 255, 255, 3, 69, 135, 5, 103, 0, 255, 255, 3, 69, 70, 5, 28, 1, 197, 162, 29, 1, 0, 192, 197, 98, 0, 68, 5, 5, 8, 0, 35, 1, 16, 0, 35, 1, 4, 192, 197, 98, 231, 1, 4, 192, 197, 162, 245, 1, 4, 192, 197, 226, 3, 3, 255, 255, 2, 69, 5, 5, 33, 30, 5, 192, 197, 34, 103, 0, 255, 255, 3, 69, 5, 5, 29, 1, 3, 192, 197, 162, 31, 1, 3, 192, 197, 226, 33, 1, 4, 192, 197, 34, 156, 68, 196, 160, 48, 0, 3, 3, 5, 192, 197, 98, 5, 68, 196, 160, 48, 0, 3, 3, 5, 192, 197, 162, 156, 68, 196, 196, 48, 0, 3, 3, 5, 192, 197, 226, 5, 68, 196, 196, 48, 0, 3, 3, 6, 192, 197, 34, 156, 68, 196, 210, 48, 0, 3, 3, 6, 192, 197, 98, 5, 68, 196, 210, 48, 0, 3, 3, 6, 192, 197, 162, 156, 68, 196, 204, 48, 0, 3, 3, 6, 192, 197, 226, 5, 68, 196, 204, 48, 0, 3, 3, 7, 192, 197, 34, 12, 0, 71, 0, 72, 0, 103, 0, 104, 0, 28, 1, 36, 1, 230, 1, 232, 1, 244, 1, 246, 1, 32, 30, 34, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 26, 4, 185, 128, 185, 132, 185, 136, 185, 144, 186, 148, 185, 152, 185, 160, 185, 164, 185, 168, 185, 172, 185, 176, 185, 180, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 48, 4, 168, 128, 168, 132, 168, 136, 168, 144, 168, 148, 168, 152, 168, 160, 168, 164, 168, 168, 168, 172, 168, 176, 168, 180, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 128, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 185, 116, 185, 129, 153, 129, 249, 129, 153, 130, 89, 130, 136, 14, 217, 130, 185, 161, 153, 161, 249, 161, 89, 162, 153, 177, 185, 184, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 168, 116, 168, 129, 136, 129, 232, 129, 136, 130, 72, 130, 128, 14, 200, 130, 168, 161, 136, 161, 232, 161, 72, 162, 136, 177, 168, 184, 72, 178, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 140, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 60, 4, 62, 4, 26, 114, 8, 114, 10, 8, 185, 124, 168, 124, 57, 131, 40, 131, 217, 129, 200, 129, 121, 130, 104, 130, 12, 8, 14, 8, 153, 145, 136, 145, 249, 146, 232, 146, 25, 146, 8, 146, 154, 149, 136, 149, 250, 149, 232, 149, 250, 150, 232, 150, 26, 150, 8, 150, 249, 154, 232, 154, 25, 154, 8, 154, 185, 156, 168, 156, 153, 162, 136, 162, 57, 163, 40, 163, 217, 161, 200, 161, 57, 162, 40, 162, 121, 162, 104, 162, 25, 163, 8, 163, 249, 169, 232, 169, 249, 177, 232, 177, 89, 178, 153, 181, 136, 181, 185, 182, 168, 182, 25, 182, 8, 182, 232, 148, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 148, 169, 148, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 188, 169, 112, 217, 128, 187, 72, 200, 128, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 186, 112, 71, 4, 186, 120, 103, 4, 177, 120, 28, 5, 250, 121, 29, 5, 241, 121, 30, 5, 218, 121, 31, 5, 209, 121, 32, 5, 186, 122, 33, 5, 177, 122, 34, 5, 250, 122, 35, 5, 241, 122, 255, 5, 168, 112, 103, 4, 168, 120, 29, 5, 232, 121, 31, 5, 200, 121, 33, 5, 168, 122, 35, 5, 232, 122, 255, 5, 250, 114, 255, 5, 232, 114, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 35, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 252, 7, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 1, 8, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 144, 27, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 148, 27, 0, 0, 148, 27, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 120, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 49, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 102, 0, 111, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 0, 0, 134, 220, 38, 0, 68, 0, 60, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 60, 0, 60, 0, 240, 0, 60, 0, 60, 0, 60, 0, 208, 0, 38, 0, 116, 0, 60, 0, 60, 0, 60, 0, 254, 0, 47, 0, 104, 0, 38, 0, 84, 0, 60, 0, 60, 0, 60, 0, 222, 0, 47, 0, 72, 0, 38, 0, 89, 0, 60, 0, 60, 0, 117, 0, 8, 3, 60, 0, 60, 0, 60, 0, 85, 0, 8, 3, 60, 0, 60, 0, 117, 0, 11, 3, 60, 0, 60, 0, 60, 0, 85, 0, 11, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 192, 1, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 60, 0, 101, 0, 40, 3, 60, 0, 60, 0, 60, 0, 69, 0, 40, 3, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 60, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 81, 1, 60, 0, 60, 0, 60, 0, 80, 1, 60, 0, 60, 0, 83, 1, 60, 0, 60, 0, 60, 0, 82, 1, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 60, 0, 60, 0, 60, 0, 97, 0, 97, 0, 60, 0, 60, 0, 60, 0, 65, 0, 97, 0, 60, 0, 60, 0, 60, 0, 65, 0, 65, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 4, 57, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 163, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 100, 38, 0, 0, 100, 38, 0, 0, 140, 47, 0, 0, 140, 47, 0, 0, 76, 51, 0, 0, 76, 51, 0, 0, 178, 56, 0, 0, 232, 56, 0, 0, 232, 56, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 89, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 145, 0, 153, 0, 159, 0, 167, 0, 175, 0, 183, 0, 48, 0, 56, 0, 182, 0, 190, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 207, 0, 215, 0, 48, 0, 48, 0, 48, 0, 223, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 231, 0, 48, 0, 48, 0, 239, 0, 48, 0, 245, 0, 253, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 8, 1, 48, 0, 48, 0, 12, 1, 48, 0, 20, 1, 25, 1, 33, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 39, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 47, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 50, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 205, 0, 48, 0, 48, 0, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 58, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 66, 1, 48, 0, 48, 0, 72, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 68, 2, 124, 2, 188, 2, 192, 0, 216, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 192, 28, 0, 202, 225, 28, 0, 202, 2, 29, 0, 202, 35, 29, 0, 202, 68, 29, 0, 202, 101, 29, 0, 202, 134, 29, 0, 202, 167, 29, 0, 202, 200, 29, 0, 202, 233, 29, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 196, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 102, 29, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 38, 30, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 70, 31, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 228, 32, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 166, 47, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 102, 48, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 134, 49, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 129, 8, 0, 198, 161, 8, 0, 198, 97, 27, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 225, 17, 0, 5, 5, 127, 12, 198, 129, 27, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 34, 20, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 198, 97, 22, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 198, 33, 27, 0, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 51, 0, 201, 102, 52, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 193, 17, 0, 198, 1, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 18, 0, 198, 65, 27, 0, 198, 97, 28, 0, 198, 129, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 134, 53, 0, 201, 166, 54, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 55, 0, 201, 230, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 58, 0, 201, 38, 59, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 192, 0, 0, 0, 198, 162, 29, 0, 198, 226, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 30, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 60, 0, 201, 102, 61, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 31, 0, 197, 34, 5, 0, 198, 162, 31, 0, 197, 98, 5, 0, 198, 226, 31, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 32, 0, 198, 98, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 134, 62, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 64, 0, 5, 71, 238, 102, 201, 70, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 6, 0, 192, 0, 0, 0, 197, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 7, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 197, 34, 8, 0, 197, 99, 8, 0, 197, 194, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 9, 0, 197, 67, 9, 0, 197, 162, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 226, 5, 0, 197, 34, 10, 0, 5, 5, 10, 121, 197, 98, 10, 0, 197, 162, 10, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 66, 0, 201, 70, 67, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 68, 0, 201, 134, 69, 0, 201, 166, 70, 0, 201, 198, 71, 0, 201, 230, 72, 0, 201, 6, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 75, 0, 201, 70, 76, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 77, 0, 201, 134, 78, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 79, 0, 201, 198, 80, 0, 201, 230, 81, 0, 201, 6, 83, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 84, 0, 201, 70, 85, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 72, 0, 5, 0, 4, 111, 93, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 134, 0, 71, 0, 2, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 71, 0, 2, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 3, 111, 93, 0, 5, 0, 72, 0, 3, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 4, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 193, 4, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 6, 121, 197, 226, 5, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 34, 6, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 98, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 226, 6, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 34, 7, 0, 5, 5, 107, 121, 197, 98, 7, 0, 5, 5, 111, 121, 197, 162, 7, 0, 5, 5, 115, 121, 197, 226, 7, 0, 5, 5, 119, 121, 197, 34, 8, 0, 197, 99, 8, 0, 197, 194, 8, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 2, 9, 0, 197, 67, 9, 0, 197, 162, 9, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 226, 9, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 226, 5, 0, 197, 34, 10, 0, 5, 5, 10, 121, 197, 98, 10, 0, 197, 162, 10, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 0, 0, 62, 0, 3, 2, 118, 0, 169, 30, 54, 0, 177, 30, 26, 0, 181, 30, 12, 0, 181, 30, 16, 192, 198, 163, 182, 30, 17, 192, 198, 3, 183, 30, 17, 192, 198, 99, 43, 33, 1, 192, 198, 130, 177, 30, 15, 192, 198, 35, 178, 30, 15, 192, 198, 131, 179, 30, 15, 192, 198, 227, 180, 30, 16, 192, 198, 67, 173, 30, 12, 0, 173, 30, 13, 192, 198, 163, 174, 30, 14, 192, 198, 3, 175, 30, 14, 192, 198, 99, 176, 30, 14, 192, 198, 195, 169, 30, 12, 192, 198, 35, 170, 30, 12, 192, 198, 131, 171, 30, 12, 192, 198, 227, 172, 30, 13, 192, 198, 67, 161, 30, 26, 0, 165, 30, 12, 0, 165, 30, 10, 192, 198, 163, 166, 30, 11, 192, 198, 3, 167, 30, 11, 192, 198, 99, 168, 30, 11, 192, 198, 195, 161, 30, 9, 192, 198, 130, 162, 30, 9, 192, 198, 194, 163, 30, 10, 192, 198, 2, 164, 30, 10, 192, 198, 67, 10, 3, 12, 0, 10, 3, 8, 192, 198, 161, 0, 30, 8, 192, 198, 194, 1, 30, 9, 192, 198, 2, 160, 30, 9, 192, 198, 66, 3, 2, 8, 192, 198, 66, 38, 2, 5, 0, 39, 2, 8, 0, 8, 3, 8, 192, 198, 129, 240, 65, 198, 2, 4, 3, 6, 192, 198, 3, 240, 65, 198, 98, 4, 3, 6, 192, 198, 99, 1, 1, 54, 0, 223, 1, 26, 0, 251, 1, 12, 0, 251, 1, 7, 192, 198, 35, 0, 2, 7, 192, 198, 130, 1, 2, 7, 192, 198, 194, 2, 2, 8, 192, 198, 2, 223, 1, 5, 192, 198, 163, 224, 1, 6, 192, 198, 3, 225, 1, 6, 192, 198, 99, 250, 1, 6, 192, 198, 195, 5, 1, 12, 0, 5, 1, 4, 192, 198, 130, 205, 1, 4, 192, 198, 194, 206, 1, 5, 192, 198, 2, 222, 1, 5, 192, 198, 67, 1, 1, 3, 192, 198, 130, 2, 1, 3, 192, 198, 194, 3, 1, 4, 192, 198, 2, 4, 1, 4, 192, 198, 66, 197, 0, 26, 0, 227, 0, 12, 0, 227, 0, 2, 192, 198, 130, 228, 0, 2, 192, 198, 194, 229, 0, 3, 192, 198, 2, 0, 1, 3, 192, 198, 66, 197, 0, 1, 192, 198, 130, 224, 0, 1, 192, 198, 194, 225, 0, 2, 192, 198, 2, 226, 0, 2, 192, 198, 66, 193, 0, 12, 0, 193, 0, 0, 192, 198, 130, 194, 0, 0, 192, 198, 194, 195, 0, 1, 192, 198, 2, 196, 0, 1, 192, 198, 66, 65, 0, 198, 129, 97, 0, 198, 161, 192, 0, 0, 192, 198, 66, 0, 50, 156, 5, 48, 0, 40, 3, 17, 192, 198, 193, 0, 70, 156, 5, 1, 0, 8, 3, 17, 192, 198, 225, 11, 3, 18, 192, 198, 1, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 31, 0, 39, 2, 59, 0, 169, 30, 26, 0, 177, 30, 12, 0, 177, 30, 25, 192, 198, 131, 179, 30, 25, 192, 198, 227, 181, 30, 26, 192, 198, 67, 183, 30, 26, 192, 198, 163, 169, 30, 24, 192, 198, 3, 171, 30, 24, 192, 198, 99, 173, 30, 24, 192, 198, 195, 175, 30, 25, 192, 198, 35, 161, 30, 12, 0, 161, 30, 22, 192, 198, 194, 163, 30, 23, 192, 198, 2, 165, 30, 23, 192, 198, 67, 167, 30, 23, 192, 198, 163, 39, 2, 10, 0, 8, 3, 22, 192, 198, 97, 10, 3, 255, 255, 111, 93, 193, 4, 1, 30, 22, 192, 198, 130, 176, 69, 198, 34, 4, 3, 21, 192, 198, 35, 3, 1, 26, 0, 225, 1, 12, 0, 225, 1, 21, 192, 198, 35, 251, 1, 21, 192, 198, 131, 1, 2, 21, 192, 198, 226, 3, 2, 22, 192, 198, 34, 3, 1, 20, 192, 198, 2, 5, 1, 20, 192, 198, 66, 206, 1, 20, 192, 198, 130, 223, 1, 20, 192, 198, 195, 227, 0, 12, 0, 227, 0, 19, 192, 198, 2, 228, 0, 19, 192, 198, 66, 229, 0, 19, 192, 198, 130, 1, 1, 19, 192, 198, 194, 97, 0, 18, 192, 198, 33, 224, 0, 18, 192, 198, 66, 225, 0, 18, 192, 198, 130, 226, 0, 18, 192, 198, 194, 0, 50, 5, 5, 48, 0, 40, 3, 27, 192, 198, 1, 0, 70, 5, 5, 1, 0, 8, 3, 27, 192, 198, 33, 11, 3, 27, 192, 198, 65, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 27, 192, 198, 162, 10, 3, 27, 192, 198, 226, 5, 42, 196, 162, 1, 0, 8, 3, 28, 192, 198, 34, 10, 3, 1, 192, 197, 34, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 98, 11, 3, 1, 192, 197, 162, 5, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 226, 11, 3, 2, 192, 197, 34, 156, 70, 196, 190, 1, 0, 8, 3, 28, 192, 198, 162, 11, 3, 28, 192, 198, 226, 5, 70, 196, 190, 1, 0, 8, 3, 29, 192, 198, 34, 11, 3, 29, 192, 198, 98, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 98, 11, 3, 2, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 156, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 98, 11, 3, 30, 192, 198, 162, 5, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 226, 11, 3, 31, 192, 198, 34, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 1, 0, 8, 3, 32, 192, 198, 162, 10, 3, 32, 192, 198, 226, 5, 42, 196, 200, 1, 0, 8, 3, 33, 192, 198, 34, 10, 3, 12, 192, 197, 162, 156, 82, 196, 198, 1, 0, 8, 3, 12, 192, 197, 226, 11, 3, 13, 192, 197, 34, 5, 82, 196, 198, 1, 0, 8, 3, 13, 192, 197, 98, 11, 3, 13, 192, 197, 162, 156, 82, 196, 208, 1, 0, 8, 3, 13, 192, 197, 226, 11, 3, 14, 192, 197, 34, 5, 82, 196, 208, 1, 0, 8, 3, 14, 192, 197, 98, 11, 3, 14, 192, 197, 162, 156, 82, 196, 204, 1, 0, 8, 3, 14, 192, 197, 226, 11, 3, 15, 192, 197, 34, 5, 82, 196, 204, 1, 0, 8, 3, 15, 192, 197, 98, 11, 3, 15, 192, 197, 162, 156, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 98, 10, 3, 33, 192, 198, 162, 5, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 226, 10, 3, 15, 192, 197, 226, 156, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 34, 11, 3, 34, 192, 198, 98, 5, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 162, 11, 3, 34, 192, 198, 226, 16, 0, 197, 35, 1, 0, 8, 3, 35, 192, 198, 35, 11, 3, 35, 192, 198, 131, 16, 0, 197, 131, 1, 0, 8, 3, 35, 192, 198, 227, 11, 3, 36, 192, 198, 67, 156, 82, 196, 196, 1, 0, 8, 3, 16, 192, 197, 226, 11, 3, 17, 192, 197, 34, 5, 82, 196, 196, 1, 0, 8, 3, 17, 192, 197, 98, 11, 3, 17, 192, 197, 162, 17, 0, 197, 227, 1, 0, 8, 3, 18, 192, 197, 67, 11, 3, 18, 192, 197, 163, 19, 0, 197, 3, 1, 0, 8, 3, 19, 192, 197, 99, 11, 3, 19, 192, 197, 195, 26, 0, 65, 0, 66, 0, 97, 0, 98, 0, 192, 0, 198, 0, 224, 0, 230, 0, 0, 1, 6, 1, 205, 1, 207, 1, 222, 1, 226, 1, 250, 1, 252, 1, 0, 2, 4, 2, 38, 2, 40, 2, 0, 30, 2, 30, 160, 30, 184, 30, 43, 33, 44, 33, 3, 0, 67, 0, 58, 0, 32, 0, 123, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 10, 51, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 52, 30, 0, 0, 52, 30, 0, 0, 92, 39, 0, 0, 92, 39, 0, 0, 56, 41, 0, 0, 56, 41, 0, 0, 88, 46, 0, 0, 142, 46, 0, 0, 238, 50, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 214, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 48, 0, 48, 0, 170, 0, 48, 0, 178, 0, 183, 0, 191, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 197, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 158, 0, 48, 0, 48, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 13, 0, 202, 193, 13, 0, 202, 226, 13, 0, 202, 3, 14, 0, 202, 36, 14, 0, 202, 69, 14, 0, 202, 102, 14, 0, 202, 135, 14, 0, 202, 168, 14, 0, 202, 201, 14, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 4, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 166, 28, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 102, 29, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 134, 30, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 36, 32, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 230, 46, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 166, 47, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 198, 48, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 129, 8, 0, 198, 161, 8, 0, 198, 97, 27, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 225, 17, 0, 5, 5, 127, 12, 198, 129, 27, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 98, 13, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 198, 97, 22, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 198, 33, 27, 0, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 50, 0, 201, 166, 51, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 193, 17, 0, 198, 1, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 18, 0, 198, 65, 27, 0, 198, 97, 28, 0, 198, 129, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 198, 52, 0, 201, 230, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 55, 0, 201, 38, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 57, 0, 201, 102, 58, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 192, 0, 0, 0, 198, 162, 29, 0, 198, 226, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 30, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 59, 0, 201, 166, 60, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 31, 0, 197, 34, 5, 0, 198, 162, 31, 0, 197, 98, 5, 0, 198, 226, 31, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 32, 0, 198, 98, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 61, 0, 201, 230, 62, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 64, 0, 201, 38, 65, 0, 201, 70, 66, 0, 201, 102, 67, 0, 201, 134, 68, 0, 201, 166, 69, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 70, 0, 201, 230, 71, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 73, 0, 201, 38, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 75, 0, 201, 102, 76, 0, 201, 134, 77, 0, 201, 166, 78, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 79, 0, 201, 230, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 72, 0, 5, 0, 4, 111, 93, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 137, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 72, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 134, 0, 71, 0, 2, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 7, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 71, 0, 2, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 3, 111, 93, 0, 5, 0, 72, 0, 3, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 4, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 0, 0, 62, 0, 3, 2, 118, 0, 169, 30, 54, 0, 177, 30, 26, 0, 181, 30, 12, 0, 181, 30, 16, 192, 198, 163, 182, 30, 17, 192, 198, 3, 183, 30, 17, 192, 198, 99, 43, 33, 1, 192, 198, 130, 177, 30, 15, 192, 198, 35, 178, 30, 15, 192, 198, 131, 179, 30, 15, 192, 198, 227, 180, 30, 16, 192, 198, 67, 173, 30, 12, 0, 173, 30, 13, 192, 198, 163, 174, 30, 14, 192, 198, 3, 175, 30, 14, 192, 198, 99, 176, 30, 14, 192, 198, 195, 169, 30, 12, 192, 198, 35, 170, 30, 12, 192, 198, 131, 171, 30, 12, 192, 198, 227, 172, 30, 13, 192, 198, 67, 161, 30, 26, 0, 165, 30, 12, 0, 165, 30, 10, 192, 198, 163, 166, 30, 11, 192, 198, 3, 167, 30, 11, 192, 198, 99, 168, 30, 11, 192, 198, 195, 161, 30, 9, 192, 198, 130, 162, 30, 9, 192, 198, 194, 163, 30, 10, 192, 198, 2, 164, 30, 10, 192, 198, 67, 10, 3, 12, 0, 10, 3, 8, 192, 198, 161, 0, 30, 8, 192, 198, 194, 1, 30, 9, 192, 198, 2, 160, 30, 9, 192, 198, 66, 3, 2, 8, 192, 198, 66, 38, 2, 5, 0, 39, 2, 8, 0, 8, 3, 8, 192, 198, 129, 240, 65, 198, 2, 4, 3, 6, 192, 198, 3, 240, 65, 198, 98, 4, 3, 6, 192, 198, 99, 1, 1, 54, 0, 223, 1, 26, 0, 251, 1, 12, 0, 251, 1, 7, 192, 198, 35, 0, 2, 7, 192, 198, 130, 1, 2, 7, 192, 198, 194, 2, 2, 8, 192, 198, 2, 223, 1, 5, 192, 198, 163, 224, 1, 6, 192, 198, 3, 225, 1, 6, 192, 198, 99, 250, 1, 6, 192, 198, 195, 5, 1, 12, 0, 5, 1, 4, 192, 198, 130, 205, 1, 4, 192, 198, 194, 206, 1, 5, 192, 198, 2, 222, 1, 5, 192, 198, 67, 1, 1, 3, 192, 198, 130, 2, 1, 3, 192, 198, 194, 3, 1, 4, 192, 198, 2, 4, 1, 4, 192, 198, 66, 197, 0, 26, 0, 227, 0, 12, 0, 227, 0, 2, 192, 198, 130, 228, 0, 2, 192, 198, 194, 229, 0, 3, 192, 198, 2, 0, 1, 3, 192, 198, 66, 197, 0, 1, 192, 198, 130, 224, 0, 1, 192, 198, 194, 225, 0, 2, 192, 198, 2, 226, 0, 2, 192, 198, 66, 193, 0, 12, 0, 193, 0, 0, 192, 198, 130, 194, 0, 0, 192, 198, 194, 195, 0, 1, 192, 198, 2, 196, 0, 1, 192, 198, 66, 65, 0, 198, 129, 97, 0, 198, 161, 192, 0, 0, 192, 198, 66, 0, 50, 156, 5, 48, 0, 40, 3, 17, 192, 198, 193, 0, 70, 156, 5, 1, 0, 8, 3, 17, 192, 198, 225, 11, 3, 18, 192, 198, 1, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 31, 0, 39, 2, 59, 0, 169, 30, 26, 0, 177, 30, 12, 0, 177, 30, 25, 192, 198, 131, 179, 30, 25, 192, 198, 227, 181, 30, 26, 192, 198, 67, 183, 30, 26, 192, 198, 163, 169, 30, 24, 192, 198, 3, 171, 30, 24, 192, 198, 99, 173, 30, 24, 192, 198, 195, 175, 30, 25, 192, 198, 35, 161, 30, 12, 0, 161, 30, 22, 192, 198, 194, 163, 30, 23, 192, 198, 2, 165, 30, 23, 192, 198, 67, 167, 30, 23, 192, 198, 163, 39, 2, 10, 0, 8, 3, 22, 192, 198, 97, 10, 3, 255, 255, 111, 93, 193, 4, 1, 30, 22, 192, 198, 130, 176, 69, 198, 34, 4, 3, 21, 192, 198, 35, 3, 1, 26, 0, 225, 1, 12, 0, 225, 1, 21, 192, 198, 35, 251, 1, 21, 192, 198, 131, 1, 2, 21, 192, 198, 226, 3, 2, 22, 192, 198, 34, 3, 1, 20, 192, 198, 2, 5, 1, 20, 192, 198, 66, 206, 1, 20, 192, 198, 130, 223, 1, 20, 192, 198, 195, 227, 0, 12, 0, 227, 0, 19, 192, 198, 2, 228, 0, 19, 192, 198, 66, 229, 0, 19, 192, 198, 130, 1, 1, 19, 192, 198, 194, 97, 0, 18, 192, 198, 33, 224, 0, 18, 192, 198, 66, 225, 0, 18, 192, 198, 130, 226, 0, 18, 192, 198, 194, 0, 50, 5, 5, 48, 0, 40, 3, 27, 192, 198, 1, 0, 70, 5, 5, 1, 0, 8, 3, 27, 192, 198, 33, 11, 3, 27, 192, 198, 65, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 27, 192, 198, 162, 10, 3, 27, 192, 198, 226, 5, 42, 196, 162, 1, 0, 8, 3, 28, 192, 198, 34, 10, 3, 1, 192, 197, 34, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 98, 11, 3, 1, 192, 197, 162, 5, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 226, 11, 3, 2, 192, 197, 34, 156, 70, 196, 190, 1, 0, 8, 3, 28, 192, 198, 162, 11, 3, 28, 192, 198, 226, 5, 70, 196, 190, 1, 0, 8, 3, 29, 192, 198, 34, 11, 3, 29, 192, 198, 98, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 98, 11, 3, 2, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 156, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 98, 11, 3, 30, 192, 198, 162, 5, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 226, 11, 3, 31, 192, 198, 34, 156, 42, 196, 200, 1, 0, 8, 3, 32, 192, 198, 162, 10, 3, 32, 192, 198, 226, 5, 42, 196, 200, 1, 0, 8, 3, 33, 192, 198, 34, 10, 3, 5, 192, 197, 226, 156, 82, 196, 198, 1, 0, 8, 3, 6, 192, 197, 34, 11, 3, 6, 192, 197, 98, 5, 82, 196, 198, 1, 0, 8, 3, 6, 192, 197, 162, 11, 3, 6, 192, 197, 226, 156, 82, 196, 208, 1, 0, 8, 3, 7, 192, 197, 34, 11, 3, 7, 192, 197, 98, 5, 82, 196, 208, 1, 0, 8, 3, 7, 192, 197, 162, 11, 3, 7, 192, 197, 226, 156, 82, 196, 204, 1, 0, 8, 3, 8, 192, 197, 34, 11, 3, 8, 192, 197, 98, 5, 82, 196, 204, 1, 0, 8, 3, 8, 192, 197, 162, 11, 3, 8, 192, 197, 226, 156, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 98, 10, 3, 33, 192, 198, 162, 5, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 226, 10, 3, 9, 192, 197, 34, 156, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 34, 11, 3, 34, 192, 198, 98, 5, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 162, 11, 3, 34, 192, 198, 226, 9, 0, 197, 99, 1, 0, 8, 3, 35, 192, 198, 35, 11, 3, 35, 192, 198, 131, 9, 0, 197, 195, 1, 0, 8, 3, 35, 192, 198, 227, 11, 3, 36, 192, 198, 67, 156, 82, 196, 196, 1, 0, 8, 3, 10, 192, 197, 34, 11, 3, 10, 192, 197, 98, 5, 82, 196, 196, 1, 0, 8, 3, 10, 192, 197, 162, 11, 3, 10, 192, 197, 226, 11, 0, 197, 35, 1, 0, 8, 3, 11, 192, 197, 131, 11, 3, 11, 192, 197, 227, 12, 0, 197, 67, 1, 0, 8, 3, 12, 192, 197, 163, 11, 3, 13, 192, 197, 3, 26, 0, 65, 0, 66, 0, 97, 0, 98, 0, 192, 0, 198, 0, 224, 0, 230, 0, 0, 1, 6, 1, 205, 1, 207, 1, 222, 1, 226, 1, 250, 1, 252, 1, 0, 2, 4, 2, 38, 2, 40, 2, 0, 30, 2, 30, 160, 30, 184, 30, 43, 33, 44, 33, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 20, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 22, 4, 185, 60, 185, 64, 185, 68, 64, 4, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 66, 4, 185, 124, 185, 128, 185, 136, 186, 140, 186, 144, 68, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 70, 4, 168, 60, 168, 64, 168, 68, 92, 4, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 94, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 96, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 121, 58, 217, 180, 185, 188, 185, 180, 185, 66, 185, 73, 153, 73, 249, 73, 89, 74, 186, 89, 154, 89, 250, 89, 90, 90, 249, 68, 122, 114, 185, 121, 153, 121, 249, 121, 121, 122, 217, 184, 136, 14, 185, 184, 185, 153, 153, 153, 249, 153, 217, 168, 153, 169, 0, 8, 2, 8, 168, 57, 136, 57, 232, 57, 104, 58, 200, 180, 168, 188, 168, 180, 168, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 232, 68, 104, 114, 168, 121, 136, 121, 232, 121, 104, 122, 200, 184, 128, 14, 168, 184, 168, 153, 136, 153, 232, 153, 200, 168, 136, 169, 4, 8, 72, 170, 249, 58, 232, 58, 217, 57, 200, 57, 98, 4, 100, 4, 153, 65, 136, 65, 249, 65, 232, 65, 153, 66, 136, 66, 25, 66, 8, 66, 25, 70, 8, 70, 217, 68, 200, 68, 249, 74, 232, 74, 217, 73, 200, 73, 153, 74, 136, 74, 249, 180, 232, 180, 25, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 153, 82, 136, 82, 185, 82, 168, 82, 249, 85, 232, 85, 25, 87, 8, 87, 122, 90, 104, 90, 250, 90, 232, 90, 218, 89, 200, 89, 218, 90, 200, 90, 154, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 185, 102, 168, 102, 168, 132, 153, 105, 136, 105, 185, 106, 168, 106, 25, 106, 8, 106, 57, 107, 40, 107, 25, 107, 8, 107, 154, 113, 136, 113, 186, 114, 168, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 249, 122, 232, 122, 217, 121, 200, 121, 249, 184, 232, 184, 25, 185, 8, 185, 153, 137, 136, 137, 185, 138, 168, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 186, 142, 168, 142, 26, 142, 8, 142, 186, 146, 168, 146, 26, 146, 8, 146, 185, 148, 168, 148, 121, 154, 104, 154, 249, 154, 232, 154, 217, 153, 200, 153, 57, 154, 40, 154, 249, 168, 232, 168, 102, 4, 104, 4, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 153, 174, 136, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 12, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 14, 8, 1, 0, 8, 14, 16, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 18, 8, 200, 12, 72, 12, 32, 14, 187, 144, 185, 84, 200, 140, 169, 140, 169, 144, 168, 84, 187, 88, 187, 96, 169, 88, 169, 96, 169, 176, 169, 112, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 65, 4, 188, 188, 97, 4, 179, 188, 192, 4, 188, 189, 193, 4, 156, 189, 194, 4, 252, 189, 195, 4, 124, 190, 196, 4, 92, 190, 197, 4, 60, 190, 224, 4, 179, 189, 225, 4, 147, 189, 226, 4, 243, 189, 227, 4, 115, 190, 228, 4, 83, 190, 229, 4, 51, 190, 0, 5, 252, 190, 1, 5, 243, 190, 2, 5, 220, 189, 3, 5, 211, 189, 4, 5, 220, 190, 5, 5, 211, 190, 255, 5, 185, 72, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 97, 4, 170, 188, 224, 4, 170, 189, 225, 4, 138, 189, 226, 4, 234, 189, 227, 4, 106, 190, 228, 4, 74, 190, 229, 4, 42, 190, 1, 5, 234, 190, 3, 5, 202, 189, 5, 5, 202, 190, 255, 5, 168, 72, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 217, 58, 255, 5, 200, 58, 255, 5, 217, 154, 255, 5, 200, 154, 255, 1, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 195, 14, 0, 16, 57, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 189, 14, 0, 32, 135, 27, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 140, 27, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 47, 0, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 51, 0, 0, 0, 51, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 31, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 98, 0, 97, 0, 99, 0, 107, 0, 119, 0, 97, 0, 114, 0, 100, 0, 115, 0, 32, 0, 50, 0, 93, 0, 0, 0, 32, 0, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 2, 0, 0, 0, 16, 40, 2, 15, 3, 0, 69, 0, 60, 0, 32, 0, 31, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 40, 0, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 45, 0, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 164, 16, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 168, 16, 0, 0, 168, 16, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 44, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 101, 0, 115, 0, 93, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 101, 0, 115, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 212, 35, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 77, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 148, 32, 0, 0, 148, 32, 0, 0, 148, 32, 0, 0, 148, 32, 0, 0, 252, 34, 0, 0, 252, 34, 0, 0, 188, 35, 0, 0, 188, 35, 0, 0, 188, 35, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 252, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 84, 0, 92, 0, 98, 0, 106, 0, 114, 0, 122, 0, 48, 0, 56, 0, 121, 0, 129, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 146, 0, 154, 0, 48, 0, 48, 0, 48, 0, 162, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 170, 0, 48, 0, 48, 0, 178, 0, 48, 0, 184, 0, 192, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 195, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 202, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 210, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 213, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 144, 0, 48, 0, 48, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 243, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 221, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 229, 0, 48, 0, 48, 0, 235, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 80, 1, 136, 1, 200, 1, 192, 0, 228, 1, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 18, 0, 202, 33, 18, 0, 202, 66, 18, 0, 202, 99, 18, 0, 202, 132, 18, 0, 202, 165, 18, 0, 202, 198, 18, 0, 202, 231, 18, 0, 202, 8, 19, 0, 202, 41, 19, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 198, 0, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 166, 1, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 8, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 134, 5, 2, 69, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 9, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 9, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 5, 5, 2, 69, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 2, 0, 201, 38, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 230, 3, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 5, 0, 5, 71, 238, 102, 201, 166, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 1, 0, 192, 0, 0, 0, 197, 98, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 1, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 197, 98, 2, 0, 192, 0, 0, 0, 197, 162, 2, 0, 192, 0, 0, 0, 197, 226, 2, 0, 197, 35, 3, 0, 197, 130, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 3, 0, 197, 3, 4, 0, 197, 98, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 162, 0, 0, 197, 226, 4, 0, 5, 5, 10, 121, 197, 34, 5, 0, 197, 98, 5, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 197, 226, 6, 0, 197, 34, 7, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 1, 0, 5, 5, 24, 121, 197, 98, 1, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 7, 0, 201, 70, 8, 0, 201, 6, 9, 0, 201, 198, 9, 0, 201, 134, 10, 0, 201, 70, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 162, 0, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 226, 0, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 34, 1, 0, 5, 5, 24, 121, 197, 98, 1, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 162, 1, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 226, 1, 0, 5, 5, 107, 121, 197, 34, 2, 0, 5, 5, 111, 121, 197, 98, 2, 0, 5, 5, 115, 121, 197, 162, 2, 0, 5, 5, 119, 121, 197, 226, 2, 0, 197, 35, 3, 0, 197, 130, 3, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 194, 3, 0, 197, 3, 4, 0, 197, 98, 4, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 162, 4, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 162, 0, 0, 197, 226, 4, 0, 5, 5, 10, 121, 197, 34, 5, 0, 197, 98, 5, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 197, 226, 6, 0, 197, 34, 7, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 1, 0, 5, 5, 24, 121, 197, 98, 1, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 68, 156, 5, 48, 0, 3, 3, 255, 255, 2, 69, 134, 5, 0, 68, 5, 5, 48, 0, 3, 3, 255, 255, 2, 69, 5, 5, 156, 68, 196, 160, 48, 0, 3, 3, 197, 162, 5, 68, 196, 160, 48, 0, 3, 3, 0, 192, 197, 98, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 68, 196, 196, 48, 0, 3, 3, 7, 192, 197, 98, 5, 68, 196, 196, 48, 0, 3, 3, 7, 192, 197, 162, 156, 68, 196, 210, 48, 0, 3, 3, 7, 192, 197, 226, 5, 68, 196, 210, 48, 0, 3, 3, 8, 192, 197, 34, 156, 68, 196, 204, 48, 0, 3, 3, 8, 192, 197, 98, 5, 68, 196, 204, 48, 0, 3, 3, 8, 192, 197, 162, 3, 0, 67, 0, 58, 0, 32, 0, 47, 0, 0, 16, 18, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 186, 29, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 212, 24, 0, 0, 88, 25, 0, 0, 88, 25, 0, 0, 210, 25, 0, 0, 210, 25, 0, 0, 162, 29, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 128, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 111, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 104, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 2, 0, 202, 1, 3, 0, 202, 34, 3, 0, 202, 67, 3, 0, 202, 100, 3, 0, 202, 133, 3, 0, 202, 166, 3, 0, 202, 199, 3, 0, 202, 232, 3, 0, 202, 9, 4, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 6, 0, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 230, 0, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 2, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 134, 5, 2, 69, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 2, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 2, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 5, 5, 2, 69, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 1, 0, 201, 102, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 3, 0, 201, 230, 3, 0, 201, 166, 4, 0, 201, 102, 5, 0, 201, 38, 6, 0, 201, 230, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 68, 156, 5, 48, 0, 3, 3, 255, 255, 2, 69, 134, 5, 0, 68, 5, 5, 48, 0, 3, 3, 255, 255, 2, 69, 5, 5, 156, 68, 196, 160, 48, 0, 3, 3, 197, 162, 5, 68, 196, 160, 48, 0, 3, 3, 0, 192, 197, 98, 156, 68, 196, 196, 48, 0, 3, 3, 0, 192, 197, 162, 5, 68, 196, 196, 48, 0, 3, 3, 0, 192, 197, 226, 156, 68, 196, 210, 48, 0, 3, 3, 1, 192, 197, 34, 5, 68, 196, 210, 48, 0, 3, 3, 1, 192, 197, 98, 156, 68, 196, 204, 48, 0, 3, 3, 1, 192, 197, 162, 5, 68, 196, 204, 48, 0, 3, 3, 1, 192, 197, 226, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 26, 4, 185, 124, 185, 128, 185, 132, 185, 140, 186, 144, 185, 148, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 185, 176, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 28, 4, 168, 124, 168, 128, 168, 132, 168, 140, 168, 144, 168, 148, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 168, 176, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 124, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 185, 116, 185, 125, 153, 125, 249, 125, 153, 126, 89, 126, 136, 14, 217, 126, 185, 157, 153, 157, 249, 157, 89, 158, 153, 173, 185, 180, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 168, 116, 168, 125, 136, 125, 232, 125, 136, 126, 72, 126, 128, 14, 200, 126, 168, 157, 136, 157, 232, 157, 72, 158, 136, 173, 168, 180, 72, 174, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 136, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 30, 4, 32, 4, 26, 114, 8, 114, 10, 8, 185, 120, 168, 120, 57, 127, 40, 127, 217, 125, 200, 125, 121, 126, 104, 126, 12, 8, 14, 8, 153, 141, 136, 141, 249, 142, 232, 142, 25, 142, 8, 142, 154, 145, 136, 145, 250, 145, 232, 145, 250, 146, 232, 146, 26, 146, 8, 146, 249, 150, 232, 150, 25, 150, 8, 150, 185, 152, 168, 152, 153, 158, 136, 158, 57, 159, 40, 159, 217, 157, 200, 157, 57, 158, 40, 158, 121, 158, 104, 158, 25, 159, 8, 159, 249, 165, 232, 165, 249, 173, 232, 173, 89, 174, 153, 177, 136, 177, 185, 178, 168, 178, 25, 178, 8, 178, 232, 144, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 144, 169, 144, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 184, 169, 112, 217, 124, 187, 72, 200, 124, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 186, 112, 255, 5, 168, 112, 255, 5, 250, 114, 255, 5, 232, 114, 255, 1, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 43, 9, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 37, 9, 0, 32, 155, 16, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 160, 16, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 174, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 178, 6, 0, 0, 178, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 68, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 86, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 208, 10, 60, 0, 130, 10, 60, 0, 60, 0, 129, 10, 60, 0, 131, 10, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 124, 25, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 13, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 116, 0, 0, 0, 116, 1, 0, 0, 20, 25, 0, 0, 20, 25, 0, 0, 28, 25, 0, 0, 28, 25, 0, 0, 96, 25, 0, 0, 15, 0, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 107, 188, 255, 0, 108, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 41, 42, 39, 43, 44, 45, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 113, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 193, 2, 7, 107, 193, 3, 7, 107, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 5, 0, 70, 0, 2, 7, 107, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 69, 0, 60, 0, 32, 0, 71, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 167, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 172, 6, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 106, 9, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 110, 9, 0, 0, 110, 9, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 67, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 84, 220, 38, 0, 66, 0, 60, 0, 83, 2, 60, 0, 60, 0, 60, 0, 129, 1, 38, 0, 68, 0, 60, 0, 87, 2, 60, 0, 60, 0, 60, 0, 138, 1, 38, 0, 75, 0, 60, 0, 153, 1, 60, 0, 60, 0, 60, 0, 152, 1, 38, 0, 83, 0, 60, 0, 115, 0, 104, 0, 60, 0, 60, 0, 60, 0, 83, 0, 104, 0, 60, 0, 60, 0, 60, 0, 83, 0, 72, 0, 38, 0, 84, 0, 60, 0, 116, 0, 115, 0, 60, 0, 60, 0, 60, 0, 84, 0, 115, 0, 60, 0, 60, 0, 60, 0, 84, 0, 83, 0, 38, 0, 89, 0, 60, 0, 180, 1, 60, 0, 60, 0, 60, 0, 188, 2, 121, 0, 60, 0, 60, 0, 60, 0, 39, 0, 39, 0, 121, 0, 60, 0, 60, 0, 60, 0, 179, 1, 60, 0, 60, 0, 60, 0, 188, 2, 89, 0, 60, 0, 60, 0, 60, 0, 39, 0, 39, 0, 89, 0, 0, 0, 122, 36, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 164, 25, 0, 0, 164, 25, 0, 0, 164, 25, 0, 0, 164, 25, 0, 0, 228, 28, 0, 0, 228, 28, 0, 0, 196, 31, 0, 0, 26, 32, 0, 0, 98, 36, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 141, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 94, 0, 102, 0, 100, 0, 108, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 124, 0, 48, 0, 48, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 64, 1, 192, 0, 192, 0, 120, 1, 144, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 201, 0, 0, 0, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 192, 24, 0, 202, 225, 24, 0, 202, 2, 25, 0, 202, 35, 25, 0, 202, 68, 25, 0, 202, 101, 25, 0, 202, 134, 25, 0, 202, 167, 25, 0, 202, 200, 25, 0, 202, 233, 25, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 96, 9, 0, 201, 32, 17, 0, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 64, 27, 0, 201, 96, 31, 0, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 2, 24, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 66, 24, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 130, 24, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 134, 5, 2, 45, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 63, 5, 5, 2, 63, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 136, 5, 2, 91, 5, 5, 2, 91, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 45, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 128, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 74, 5, 2, 91, 5, 136, 0, 0, 7, 5, 2, 91, 5, 136, 0, 0, 7, 5, 2, 91, 5, 150, 0, 0, 74, 5, 2, 91, 5, 142, 0, 0, 7, 5, 2, 91, 5, 142, 0, 0, 74, 5, 2, 91, 5, 150, 0, 0, 74, 5, 2, 91, 5, 164, 0, 0, 7, 5, 2, 91, 5, 164, 0, 0, 74, 5, 2, 91, 5, 156, 0, 0, 7, 5, 2, 91, 5, 156, 0, 0, 7, 5, 2, 91, 5, 146, 0, 0, 74, 5, 2, 91, 5, 138, 0, 0, 7, 5, 2, 91, 5, 138, 0, 0, 74, 5, 2, 91, 5, 196, 0, 0, 7, 5, 2, 91, 5, 196, 0, 0, 74, 5, 2, 91, 5, 182, 0, 0, 7, 5, 2, 91, 5, 182, 0, 0, 74, 5, 2, 91, 5, 154, 0, 0, 7, 5, 2, 91, 5, 154, 0, 0, 135, 5, 2, 79, 5, 142, 0, 0, 70, 5, 2, 79, 5, 142, 0, 0, 135, 5, 2, 79, 5, 144, 0, 0, 70, 5, 2, 79, 5, 144, 0, 0, 135, 5, 2, 79, 5, 156, 0, 0, 70, 5, 2, 79, 5, 156, 0, 0, 135, 5, 2, 79, 5, 196, 0, 0, 70, 5, 2, 79, 5, 196, 0, 0, 135, 5, 2, 79, 5, 150, 0, 0, 70, 5, 2, 79, 5, 150, 0, 0, 135, 5, 2, 79, 5, 160, 0, 0, 70, 5, 2, 79, 5, 160, 0, 0, 135, 5, 2, 79, 5, 206, 0, 0, 70, 5, 2, 79, 5, 206, 0, 0, 70, 5, 2, 79, 5, 210, 0, 0, 135, 5, 2, 81, 5, 136, 0, 0, 70, 5, 2, 81, 5, 136, 0, 0, 135, 5, 2, 81, 5, 142, 0, 0, 70, 5, 2, 81, 5, 142, 0, 0, 135, 5, 2, 81, 5, 160, 0, 0, 70, 5, 2, 81, 5, 160, 0, 0, 135, 5, 2, 81, 5, 144, 0, 0, 70, 5, 2, 81, 5, 144, 0, 0, 135, 5, 2, 81, 5, 202, 0, 0, 70, 5, 2, 81, 5, 202, 0, 0, 135, 5, 2, 81, 5, 156, 0, 0, 70, 5, 2, 81, 5, 156, 0, 0, 135, 5, 2, 81, 5, 196, 0, 0, 70, 5, 2, 81, 5, 196, 0, 0, 135, 5, 2, 81, 5, 136, 0, 0, 5, 156, 0, 0, 70, 5, 2, 81, 5, 136, 0, 0, 5, 156, 0, 0, 135, 5, 2, 81, 5, 144, 0, 0, 5, 156, 0, 0, 70, 5, 2, 81, 5, 144, 0, 0, 5, 156, 0, 0, 135, 5, 2, 81, 5, 196, 0, 0, 5, 156, 0, 0, 70, 5, 2, 81, 5, 196, 0, 0, 5, 156, 0, 0, 5, 5, 2, 79, 5, 142, 0, 0, 5, 5, 2, 79, 5, 144, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 150, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 206, 0, 0, 5, 5, 2, 79, 5, 210, 0, 0, 5, 5, 2, 81, 5, 136, 0, 0, 5, 5, 2, 81, 5, 142, 0, 0, 5, 5, 2, 81, 5, 160, 0, 0, 5, 5, 2, 81, 5, 144, 0, 0, 5, 5, 2, 81, 5, 202, 0, 0, 5, 5, 2, 81, 5, 156, 0, 0, 5, 5, 2, 81, 5, 196, 0, 0, 5, 5, 2, 81, 5, 136, 0, 0, 5, 156, 0, 0, 5, 5, 2, 81, 5, 144, 0, 0, 5, 156, 0, 0, 5, 5, 2, 81, 5, 196, 0, 0, 5, 156, 0, 0, 73, 5, 2, 91, 5, 136, 0, 0, 6, 5, 2, 91, 5, 136, 0, 0, 6, 5, 2, 91, 5, 150, 0, 0, 73, 5, 2, 91, 5, 142, 0, 0, 6, 5, 2, 91, 5, 142, 0, 0, 73, 5, 2, 91, 5, 150, 0, 0, 73, 5, 2, 91, 5, 164, 0, 0, 6, 5, 2, 91, 5, 164, 0, 0, 73, 5, 2, 91, 5, 156, 0, 0, 6, 5, 2, 91, 5, 156, 0, 0, 6, 5, 2, 91, 5, 146, 0, 0, 73, 5, 2, 91, 5, 138, 0, 0, 6, 5, 2, 91, 5, 138, 0, 0, 73, 5, 2, 91, 5, 196, 0, 0, 6, 5, 2, 91, 5, 196, 0, 0, 73, 5, 2, 91, 5, 182, 0, 0, 6, 5, 2, 91, 5, 182, 0, 0, 73, 5, 2, 91, 5, 154, 0, 0, 6, 5, 2, 91, 5, 154, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 106, 9, 5, 5, 20, 0, 142, 30, 37, 0, 244, 30, 20, 0, 247, 30, 9, 0, 247, 30, 4, 192, 197, 34, 248, 30, 4, 192, 197, 98, 249, 30, 4, 192, 197, 162, 244, 30, 3, 192, 197, 98, 245, 30, 3, 192, 197, 162, 246, 30, 3, 192, 197, 226, 142, 30, 2, 192, 197, 34, 143, 30, 2, 192, 197, 98, 153, 30, 2, 192, 197, 162, 242, 30, 2, 192, 197, 226, 243, 30, 3, 192, 197, 34, 118, 1, 15, 0, 118, 1, 0, 192, 197, 226, 119, 1, 1, 192, 197, 34, 120, 1, 1, 192, 197, 98, 50, 2, 1, 192, 197, 162, 51, 2, 1, 192, 197, 226, 89, 0, 255, 255, 2, 91, 74, 5, 121, 0, 255, 255, 2, 91, 7, 5, 221, 0, 197, 162, 253, 0, 0, 192, 197, 98, 255, 0, 0, 192, 197, 162, 0, 78, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 7, 192, 197, 98, 41, 30, 7, 192, 197, 162, 42, 30, 7, 192, 197, 226, 43, 30, 8, 192, 197, 34, 150, 30, 8, 192, 197, 98, 36, 30, 6, 192, 197, 98, 37, 30, 6, 192, 197, 162, 38, 30, 6, 192, 197, 226, 39, 30, 7, 192, 197, 34, 30, 2, 12, 0, 30, 2, 5, 192, 197, 98, 31, 2, 5, 192, 197, 162, 34, 30, 5, 192, 197, 226, 35, 30, 6, 192, 197, 34, 72, 0, 255, 255, 2, 79, 135, 5, 104, 0, 255, 255, 2, 79, 70, 5, 36, 1, 4, 192, 197, 226, 37, 1, 5, 192, 197, 34, 0, 80, 156, 5, 21, 0, 25, 2, 37, 0, 100, 30, 20, 0, 103, 30, 9, 0, 103, 30, 13, 192, 197, 67, 104, 30, 13, 192, 197, 163, 105, 30, 14, 192, 197, 3, 100, 30, 12, 192, 197, 35, 101, 30, 12, 192, 197, 131, 102, 30, 12, 192, 197, 227, 25, 2, 10, 192, 197, 226, 96, 30, 11, 192, 197, 34, 97, 30, 11, 192, 197, 98, 98, 30, 11, 192, 197, 162, 99, 30, 11, 192, 197, 226, 93, 1, 20, 0, 96, 1, 9, 0, 96, 1, 10, 192, 197, 34, 97, 1, 10, 192, 197, 98, 24, 2, 10, 192, 197, 162, 93, 1, 9, 192, 197, 98, 94, 1, 9, 192, 197, 162, 95, 1, 9, 192, 197, 226, 83, 0, 255, 255, 2, 81, 135, 5, 115, 0, 255, 255, 2, 81, 70, 5, 90, 1, 8, 192, 197, 162, 91, 1, 8, 192, 197, 226, 92, 1, 9, 192, 197, 34, 0, 78, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 15, 192, 197, 34, 39, 30, 15, 192, 197, 98, 41, 30, 15, 192, 197, 162, 43, 30, 15, 192, 197, 226, 150, 30, 16, 192, 197, 34, 104, 0, 255, 255, 2, 79, 5, 5, 37, 1, 14, 192, 197, 98, 31, 2, 14, 192, 197, 162, 35, 30, 14, 192, 197, 226, 0, 80, 5, 5, 10, 0, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 18, 192, 197, 35, 103, 30, 18, 192, 197, 131, 105, 30, 18, 192, 197, 227, 25, 2, 17, 192, 197, 98, 97, 30, 17, 192, 197, 162, 99, 30, 17, 192, 197, 226, 115, 0, 255, 255, 2, 81, 5, 5, 91, 1, 16, 192, 197, 98, 93, 1, 16, 192, 197, 162, 95, 1, 16, 192, 197, 226, 97, 1, 17, 192, 197, 34, 82, 93, 5, 5, 20, 0, 142, 30, 37, 0, 244, 30, 20, 0, 247, 30, 9, 0, 247, 30, 23, 192, 197, 66, 248, 30, 23, 192, 197, 130, 249, 30, 23, 192, 197, 194, 244, 30, 22, 192, 197, 130, 245, 30, 22, 192, 197, 194, 246, 30, 23, 192, 197, 2, 142, 30, 21, 192, 197, 66, 143, 30, 21, 192, 197, 130, 153, 30, 21, 192, 197, 194, 242, 30, 22, 192, 197, 2, 243, 30, 22, 192, 197, 66, 118, 1, 15, 0, 118, 1, 20, 192, 197, 2, 119, 1, 20, 192, 197, 66, 120, 1, 20, 192, 197, 130, 50, 2, 20, 192, 197, 194, 51, 2, 21, 192, 197, 2, 89, 0, 255, 255, 2, 91, 73, 5, 121, 0, 255, 255, 2, 91, 6, 5, 221, 0, 19, 192, 197, 66, 253, 0, 19, 192, 197, 130, 255, 0, 19, 192, 197, 194, 42, 0, 72, 0, 73, 0, 83, 0, 84, 0, 89, 0, 90, 0, 104, 0, 105, 0, 115, 0, 116, 0, 121, 0, 122, 0, 221, 0, 222, 0, 253, 0, 254, 0, 255, 0, 0, 1, 36, 1, 38, 1, 90, 1, 98, 1, 118, 1, 121, 1, 24, 2, 26, 2, 30, 2, 32, 2, 50, 2, 52, 2, 34, 30, 44, 30, 96, 30, 106, 30, 142, 30, 144, 30, 150, 30, 151, 30, 153, 30, 154, 30, 242, 30, 250, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 26, 4, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 185, 120, 185, 124, 185, 128, 185, 136, 44, 4, 54, 4, 185, 160, 185, 164, 185, 168, 185, 172, 185, 176, 185, 184, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 168, 120, 168, 124, 168, 128, 168, 136, 76, 4, 82, 4, 168, 160, 168, 164, 168, 168, 168, 172, 168, 176, 168, 184, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 154, 114, 185, 121, 153, 121, 249, 121, 153, 122, 89, 122, 136, 14, 217, 122, 185, 161, 153, 161, 249, 161, 89, 162, 153, 177, 185, 188, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 136, 121, 232, 121, 136, 122, 72, 122, 128, 14, 200, 122, 168, 161, 136, 161, 232, 161, 72, 162, 136, 177, 168, 188, 72, 178, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 250, 114, 232, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 57, 123, 40, 123, 217, 121, 200, 121, 121, 122, 104, 122, 12, 8, 14, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 250, 142, 232, 142, 26, 142, 8, 142, 249, 150, 232, 150, 25, 150, 8, 150, 185, 156, 168, 156, 153, 162, 136, 162, 57, 163, 40, 163, 217, 161, 200, 161, 57, 162, 40, 162, 121, 162, 104, 162, 25, 163, 8, 163, 249, 169, 232, 169, 249, 177, 232, 177, 89, 178, 153, 185, 136, 185, 185, 186, 168, 186, 25, 186, 8, 186, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 192, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 216, 12, 89, 4, 178, 180, 121, 4, 169, 180, 221, 4, 146, 181, 253, 4, 137, 181, 255, 4, 73, 182, 118, 5, 242, 181, 119, 5, 233, 181, 120, 5, 82, 182, 255, 5, 186, 140, 72, 4, 186, 144, 104, 4, 177, 144, 36, 5, 250, 145, 37, 5, 241, 145, 255, 5, 185, 148, 83, 4, 186, 152, 115, 4, 177, 152, 90, 5, 154, 153, 91, 5, 145, 153, 92, 5, 250, 153, 93, 5, 241, 153, 94, 5, 250, 154, 95, 5, 241, 154, 96, 5, 26, 154, 97, 5, 17, 154, 255, 5, 168, 140, 104, 4, 168, 144, 37, 5, 232, 145, 255, 5, 168, 148, 115, 4, 168, 152, 91, 5, 136, 153, 93, 5, 232, 153, 95, 5, 232, 154, 97, 5, 8, 154, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 67, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 99, 9, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 104, 9, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 253, 10, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 1, 11, 0, 0, 1, 11, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 40, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 38, 0, 97, 0, 60, 0, 101, 0, 60, 0, 60, 0, 60, 0, 69, 0, 60, 0, 105, 0, 60, 0, 60, 0, 60, 0, 73, 0, 60, 0, 111, 0, 60, 0, 60, 0, 60, 0, 79, 0, 60, 0, 117, 0, 60, 0, 60, 0, 60, 0, 85, 0, 38, 0, 119, 0, 60, 0, 187, 2, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 40, 43, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 84, 31, 0, 0, 84, 31, 0, 0, 84, 31, 0, 0, 84, 31, 0, 0, 80, 39, 0, 0, 80, 39, 0, 0, 80, 39, 0, 0, 80, 39, 0, 0, 16, 43, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 232, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 93, 0, 101, 0, 106, 0, 114, 0, 119, 0, 127, 0, 134, 0, 142, 0, 48, 0, 56, 0, 147, 0, 155, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 174, 0, 182, 0, 187, 0, 192, 0, 48, 0, 199, 0, 207, 0, 215, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 171, 0, 48, 0, 48, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 163, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 116, 1, 168, 1, 220, 1, 24, 2, 192, 0, 76, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 62, 0, 202, 193, 62, 0, 202, 226, 62, 0, 202, 3, 63, 0, 202, 36, 63, 0, 202, 69, 63, 0, 202, 102, 63, 0, 202, 135, 63, 0, 202, 168, 63, 0, 202, 201, 63, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 134, 5, 2, 43, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 134, 5, 3, 43, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 134, 5, 4, 43, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 134, 5, 5, 43, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 2, 43, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 3, 43, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 4, 43, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 5, 43, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 61, 0, 196, 160, 156, 46, 197, 34, 0, 0, 197, 98, 0, 0, 197, 162, 0, 0, 197, 226, 0, 0, 197, 34, 1, 0, 197, 98, 1, 0, 197, 162, 1, 0, 197, 226, 1, 0, 160, 112, 0, 48, 196, 154, 156, 68, 197, 34, 2, 0, 197, 98, 2, 0, 197, 162, 2, 0, 197, 226, 2, 0, 197, 34, 3, 0, 5, 5, 127, 12, 196, 158, 156, 70, 197, 98, 3, 0, 197, 162, 3, 0, 197, 226, 3, 0, 197, 34, 4, 0, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 62, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 62, 0, 196, 160, 5, 46, 197, 98, 4, 0, 197, 162, 4, 0, 197, 226, 4, 0, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 5, 112, 0, 48, 196, 154, 5, 68, 197, 98, 6, 0, 197, 162, 6, 0, 197, 226, 6, 0, 197, 34, 7, 0, 197, 98, 7, 0, 5, 5, 125, 12, 196, 158, 5, 70, 197, 162, 7, 0, 197, 226, 7, 0, 197, 34, 8, 0, 197, 98, 8, 0, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 8, 0, 197, 226, 8, 0, 197, 34, 9, 0, 197, 98, 9, 0, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 226, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 197, 34, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 197, 34, 14, 0, 197, 98, 14, 0, 197, 162, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 14, 0, 197, 34, 15, 0, 197, 98, 15, 0, 197, 162, 15, 0, 197, 226, 15, 0, 197, 34, 16, 0, 197, 98, 16, 0, 197, 162, 16, 0, 197, 226, 16, 0, 197, 34, 17, 0, 197, 98, 17, 0, 197, 162, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 17, 0, 197, 34, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 18, 0, 197, 162, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 18, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 197, 226, 19, 0, 197, 34, 20, 0, 197, 99, 20, 0, 197, 195, 20, 0, 197, 35, 21, 0, 197, 131, 21, 0, 197, 227, 21, 0, 197, 67, 22, 0, 197, 163, 22, 0, 197, 3, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 23, 0, 197, 162, 23, 0, 197, 227, 23, 0, 197, 67, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 24, 0, 197, 226, 24, 0, 197, 34, 25, 0, 197, 98, 25, 0, 197, 162, 25, 0, 197, 226, 25, 0, 197, 34, 26, 0, 197, 98, 26, 0, 197, 162, 26, 0, 197, 226, 26, 0, 197, 34, 27, 0, 197, 98, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 27, 0, 197, 226, 27, 0, 197, 34, 28, 0, 197, 98, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 28, 0, 197, 226, 28, 0, 197, 35, 29, 0, 197, 131, 29, 0, 197, 227, 29, 0, 197, 67, 30, 0, 197, 162, 30, 0, 197, 226, 30, 0, 197, 35, 31, 0, 197, 131, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 87, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 31, 0, 197, 67, 32, 0, 197, 163, 32, 0, 197, 3, 33, 0, 197, 98, 33, 0, 197, 162, 33, 0, 197, 226, 33, 0, 197, 34, 34, 0, 197, 99, 34, 0, 197, 195, 34, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 35, 0, 197, 98, 35, 0, 197, 163, 35, 0, 197, 3, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 36, 0, 197, 195, 36, 0, 197, 35, 37, 0, 197, 131, 37, 0, 197, 227, 37, 0, 197, 67, 38, 0, 197, 163, 38, 0, 197, 3, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 39, 0, 197, 162, 39, 0, 197, 226, 39, 0, 197, 34, 40, 0, 197, 98, 40, 0, 197, 162, 40, 0, 197, 227, 40, 0, 197, 67, 41, 0, 197, 163, 41, 0, 197, 3, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 42, 0, 197, 162, 42, 0, 197, 226, 42, 0, 197, 34, 43, 0, 197, 98, 43, 0, 197, 162, 43, 0, 197, 227, 43, 0, 197, 67, 44, 0, 197, 163, 44, 0, 197, 3, 45, 0, 197, 99, 45, 0, 197, 195, 45, 0, 197, 35, 46, 0, 197, 131, 46, 0, 197, 227, 46, 0, 197, 67, 47, 0, 197, 162, 47, 0, 197, 226, 47, 0, 197, 34, 48, 0, 197, 98, 48, 0, 197, 162, 48, 0, 197, 226, 48, 0, 197, 34, 49, 0, 197, 98, 49, 0, 197, 163, 49, 0, 197, 3, 50, 0, 197, 99, 50, 0, 197, 195, 50, 0, 197, 35, 51, 0, 197, 131, 51, 0, 197, 227, 51, 0, 197, 67, 52, 0, 197, 163, 52, 0, 197, 3, 53, 0, 197, 99, 53, 0, 197, 195, 53, 0, 197, 35, 54, 0, 197, 131, 54, 0, 197, 227, 54, 0, 197, 67, 55, 0, 197, 163, 55, 0, 197, 3, 56, 0, 197, 99, 56, 0, 197, 195, 56, 0, 197, 34, 57, 0, 197, 98, 57, 0, 197, 162, 57, 0, 197, 226, 57, 0, 197, 35, 58, 0, 197, 131, 58, 0, 197, 227, 58, 0, 197, 67, 59, 0, 197, 163, 59, 0, 197, 3, 60, 0, 197, 99, 60, 0, 197, 195, 60, 0, 197, 35, 61, 0, 197, 131, 61, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 43, 5, 138, 0, 0, 134, 5, 2, 43, 5, 136, 0, 0, 134, 5, 2, 43, 5, 142, 0, 0, 134, 5, 2, 43, 5, 150, 0, 0, 134, 5, 3, 43, 5, 138, 0, 0, 134, 5, 3, 43, 5, 136, 0, 0, 134, 5, 3, 43, 5, 142, 0, 0, 134, 5, 3, 43, 5, 150, 0, 0, 134, 5, 4, 43, 5, 138, 0, 0, 134, 5, 4, 43, 5, 136, 0, 0, 134, 5, 4, 43, 5, 142, 0, 0, 134, 5, 4, 43, 5, 154, 0, 0, 134, 5, 4, 43, 5, 150, 0, 0, 134, 5, 5, 43, 5, 138, 0, 0, 134, 5, 5, 43, 5, 136, 0, 0, 134, 5, 5, 43, 5, 142, 0, 0, 134, 5, 5, 43, 5, 150, 0, 0, 5, 5, 2, 43, 5, 138, 0, 0, 5, 5, 2, 43, 5, 136, 0, 0, 5, 5, 2, 43, 5, 142, 0, 0, 5, 5, 2, 43, 5, 150, 0, 0, 5, 5, 3, 43, 5, 138, 0, 0, 5, 5, 3, 43, 5, 136, 0, 0, 5, 5, 3, 43, 5, 142, 0, 0, 5, 5, 3, 43, 5, 150, 0, 0, 5, 5, 4, 43, 5, 138, 0, 0, 5, 5, 4, 43, 5, 136, 0, 0, 5, 5, 4, 43, 5, 142, 0, 0, 5, 5, 4, 43, 5, 154, 0, 0, 5, 5, 4, 43, 5, 150, 0, 0, 5, 5, 5, 43, 5, 138, 0, 0, 5, 5, 5, 43, 5, 136, 0, 0, 5, 5, 5, 43, 5, 142, 0, 0, 5, 5, 5, 43, 5, 150, 0, 0, 134, 5, 2, 43, 5, 164, 0, 0, 5, 5, 2, 43, 5, 164, 0, 0, 134, 5, 2, 43, 5, 140, 0, 0, 5, 5, 2, 43, 5, 140, 0, 0, 134, 5, 2, 43, 5, 156, 0, 0, 5, 5, 2, 43, 5, 156, 0, 0, 134, 5, 2, 43, 5, 162, 0, 0, 5, 5, 2, 43, 5, 162, 0, 0, 134, 5, 2, 43, 5, 144, 0, 0, 5, 5, 2, 43, 5, 144, 0, 0, 134, 5, 3, 43, 5, 154, 0, 0, 5, 5, 3, 43, 5, 154, 0, 0, 134, 5, 3, 43, 5, 164, 0, 0, 5, 5, 3, 43, 5, 164, 0, 0, 134, 5, 3, 43, 5, 140, 0, 0, 5, 5, 3, 43, 5, 140, 0, 0, 134, 5, 3, 43, 5, 162, 0, 0, 5, 5, 3, 43, 5, 162, 0, 0, 134, 5, 3, 43, 5, 156, 0, 0, 134, 5, 4, 43, 5, 164, 0, 0, 5, 5, 4, 43, 5, 164, 0, 0, 134, 5, 4, 43, 5, 140, 0, 0, 5, 5, 4, 43, 5, 140, 0, 0, 134, 5, 4, 43, 5, 152, 0, 0, 5, 5, 4, 43, 5, 152, 0, 0, 134, 5, 5, 43, 5, 154, 0, 0, 5, 5, 5, 43, 5, 154, 0, 0, 134, 5, 5, 43, 5, 164, 0, 0, 5, 5, 5, 43, 5, 164, 0, 0, 134, 5, 5, 43, 5, 140, 0, 0, 5, 5, 5, 43, 5, 140, 0, 0, 134, 5, 5, 43, 5, 146, 0, 0, 5, 5, 5, 43, 5, 146, 0, 0, 134, 5, 5, 43, 5, 152, 0, 0, 5, 5, 5, 43, 5, 152, 0, 0, 134, 5, 5, 43, 5, 162, 0, 0, 5, 5, 5, 43, 5, 162, 0, 0, 134, 5, 4, 43, 5, 190, 0, 0, 5, 5, 4, 43, 5, 190, 0, 0, 134, 5, 5, 43, 5, 190, 0, 0, 5, 5, 5, 43, 5, 190, 0, 0, 134, 5, 3, 43, 5, 144, 0, 0, 5, 5, 3, 43, 5, 144, 0, 0, 134, 5, 4, 43, 5, 144, 0, 0, 5, 5, 4, 43, 5, 144, 0, 0, 134, 5, 5, 43, 5, 144, 0, 0, 5, 5, 5, 43, 5, 144, 0, 0, 134, 5, 5, 43, 5, 150, 0, 0, 5, 164, 0, 0, 5, 5, 5, 43, 5, 150, 0, 0, 5, 164, 0, 0, 134, 5, 5, 43, 5, 150, 0, 0, 5, 136, 0, 0, 5, 5, 5, 43, 5, 150, 0, 0, 5, 136, 0, 0, 134, 5, 5, 43, 5, 150, 0, 0, 5, 144, 0, 0, 5, 5, 5, 43, 5, 150, 0, 0, 5, 144, 0, 0, 134, 5, 5, 43, 5, 150, 0, 0, 5, 138, 0, 0, 5, 5, 5, 43, 5, 150, 0, 0, 5, 138, 0, 0, 134, 5, 4, 43, 5, 162, 0, 0, 5, 5, 4, 43, 5, 162, 0, 0, 134, 5, 4, 43, 5, 162, 0, 0, 5, 164, 0, 0, 5, 5, 4, 43, 5, 162, 0, 0, 5, 164, 0, 0, 134, 5, 2, 43, 5, 184, 0, 0, 5, 5, 2, 43, 5, 184, 0, 0, 134, 5, 2, 43, 5, 188, 0, 0, 5, 5, 2, 43, 5, 188, 0, 0, 134, 5, 3, 43, 5, 184, 0, 0, 5, 5, 3, 43, 5, 184, 0, 0, 134, 5, 3, 43, 5, 188, 0, 0, 5, 5, 3, 43, 5, 188, 0, 0, 134, 5, 4, 43, 5, 184, 0, 0, 5, 5, 4, 43, 5, 184, 0, 0, 134, 5, 4, 43, 5, 188, 0, 0, 5, 5, 4, 43, 5, 188, 0, 0, 134, 5, 5, 43, 5, 184, 0, 0, 5, 5, 5, 43, 5, 184, 0, 0, 134, 5, 5, 43, 5, 188, 0, 0, 5, 5, 5, 43, 5, 188, 0, 0, 134, 5, 2, 43, 5, 160, 0, 0, 5, 5, 2, 43, 5, 160, 0, 0, 134, 5, 4, 43, 5, 150, 0, 0, 5, 164, 0, 0, 5, 5, 4, 43, 5, 150, 0, 0, 5, 164, 0, 0, 134, 5, 4, 43, 5, 154, 0, 0, 5, 164, 0, 0, 5, 5, 4, 43, 5, 154, 0, 0, 5, 164, 0, 0, 134, 5, 4, 43, 5, 156, 0, 0, 5, 5, 4, 43, 5, 156, 0, 0, 134, 5, 4, 43, 5, 156, 0, 0, 5, 164, 0, 0, 5, 5, 4, 43, 5, 156, 0, 0, 5, 164, 0, 0, 134, 5, 2, 43, 5, 164, 0, 0, 5, 138, 0, 0, 5, 5, 2, 43, 5, 164, 0, 0, 5, 138, 0, 0, 134, 5, 2, 43, 5, 164, 0, 0, 5, 136, 0, 0, 5, 5, 2, 43, 5, 164, 0, 0, 5, 136, 0, 0, 134, 5, 2, 43, 5, 204, 0, 0, 5, 5, 2, 43, 5, 204, 0, 0, 134, 5, 2, 43, 5, 208, 0, 0, 5, 5, 2, 43, 5, 208, 0, 0, 134, 5, 2, 43, 5, 160, 0, 0, 5, 140, 0, 0, 5, 5, 2, 43, 5, 160, 0, 0, 5, 140, 0, 0, 134, 5, 3, 43, 5, 208, 0, 0, 5, 5, 3, 43, 5, 208, 0, 0, 134, 5, 3, 43, 5, 150, 0, 0, 5, 136, 0, 0, 5, 5, 3, 43, 5, 150, 0, 0, 5, 136, 0, 0, 134, 5, 4, 43, 5, 154, 0, 0, 5, 136, 0, 0, 5, 5, 4, 43, 5, 154, 0, 0, 5, 136, 0, 0, 134, 5, 4, 43, 5, 154, 0, 0, 5, 150, 0, 0, 5, 5, 4, 43, 5, 154, 0, 0, 5, 150, 0, 0, 134, 5, 4, 43, 5, 164, 0, 0, 5, 138, 0, 0, 5, 5, 4, 43, 5, 164, 0, 0, 5, 138, 0, 0, 134, 5, 4, 43, 5, 164, 0, 0, 5, 136, 0, 0, 5, 5, 4, 43, 5, 164, 0, 0, 5, 136, 0, 0, 134, 5, 5, 43, 5, 198, 0, 0, 5, 5, 5, 43, 5, 198, 0, 0, 134, 5, 5, 43, 5, 208, 0, 0, 5, 5, 5, 43, 5, 208, 0, 0, 134, 5, 5, 43, 5, 204, 0, 0, 5, 5, 5, 43, 5, 204, 0, 0, 134, 5, 5, 43, 5, 154, 0, 0, 5, 136, 0, 0, 5, 5, 5, 43, 5, 154, 0, 0, 5, 136, 0, 0, 134, 5, 5, 43, 5, 164, 0, 0, 5, 150, 0, 0, 5, 5, 5, 43, 5, 164, 0, 0, 5, 150, 0, 0, 134, 5, 2, 43, 5, 196, 0, 0, 5, 5, 2, 43, 5, 196, 0, 0, 134, 5, 2, 43, 5, 182, 0, 0, 5, 5, 2, 43, 5, 182, 0, 0, 134, 5, 2, 43, 5, 154, 0, 0, 5, 5, 2, 43, 5, 154, 0, 0, 134, 5, 2, 43, 5, 142, 0, 0, 5, 136, 0, 0, 5, 5, 2, 43, 5, 142, 0, 0, 5, 136, 0, 0, 134, 5, 2, 43, 5, 142, 0, 0, 5, 138, 0, 0, 5, 5, 2, 43, 5, 142, 0, 0, 5, 138, 0, 0, 134, 5, 2, 43, 5, 142, 0, 0, 5, 182, 0, 0, 5, 5, 2, 43, 5, 142, 0, 0, 5, 182, 0, 0, 134, 5, 2, 43, 5, 142, 0, 0, 5, 154, 0, 0, 5, 5, 2, 43, 5, 142, 0, 0, 5, 154, 0, 0, 134, 5, 2, 43, 5, 196, 0, 0, 5, 142, 0, 0, 5, 5, 2, 43, 5, 196, 0, 0, 5, 142, 0, 0, 134, 5, 3, 43, 5, 182, 0, 0, 5, 5, 3, 43, 5, 182, 0, 0, 134, 5, 3, 43, 5, 196, 0, 0, 5, 5, 3, 43, 5, 196, 0, 0, 134, 5, 4, 43, 5, 196, 0, 0, 5, 5, 4, 43, 5, 196, 0, 0, 134, 5, 4, 43, 5, 182, 0, 0, 5, 5, 4, 43, 5, 182, 0, 0, 134, 5, 4, 43, 5, 142, 0, 0, 5, 136, 0, 0, 5, 5, 4, 43, 5, 142, 0, 0, 5, 136, 0, 0, 134, 5, 4, 43, 5, 142, 0, 0, 5, 138, 0, 0, 5, 5, 4, 43, 5, 142, 0, 0, 5, 138, 0, 0, 134, 5, 4, 43, 5, 142, 0, 0, 5, 182, 0, 0, 5, 5, 4, 43, 5, 142, 0, 0, 5, 182, 0, 0, 134, 5, 4, 43, 5, 142, 0, 0, 5, 154, 0, 0, 5, 5, 4, 43, 5, 142, 0, 0, 5, 154, 0, 0, 134, 5, 4, 43, 5, 196, 0, 0, 5, 142, 0, 0, 5, 5, 4, 43, 5, 196, 0, 0, 5, 142, 0, 0, 134, 5, 4, 43, 5, 190, 0, 0, 5, 136, 0, 0, 5, 5, 4, 43, 5, 190, 0, 0, 5, 136, 0, 0, 134, 5, 4, 43, 5, 190, 0, 0, 5, 138, 0, 0, 5, 5, 4, 43, 5, 190, 0, 0, 5, 138, 0, 0, 134, 5, 4, 43, 5, 190, 0, 0, 5, 182, 0, 0, 5, 5, 4, 43, 5, 190, 0, 0, 5, 182, 0, 0, 134, 5, 4, 43, 5, 190, 0, 0, 5, 154, 0, 0, 5, 5, 4, 43, 5, 190, 0, 0, 5, 154, 0, 0, 134, 5, 4, 43, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 4, 43, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 5, 43, 5, 196, 0, 0, 5, 5, 5, 43, 5, 196, 0, 0, 134, 5, 5, 43, 5, 182, 0, 0, 5, 5, 5, 43, 5, 182, 0, 0, 134, 5, 5, 43, 5, 190, 0, 0, 5, 136, 0, 0, 5, 5, 5, 43, 5, 190, 0, 0, 5, 136, 0, 0, 134, 5, 5, 43, 5, 190, 0, 0, 5, 138, 0, 0, 5, 5, 5, 43, 5, 190, 0, 0, 5, 138, 0, 0, 134, 5, 5, 43, 5, 190, 0, 0, 5, 182, 0, 0, 5, 5, 5, 43, 5, 190, 0, 0, 5, 182, 0, 0, 134, 5, 5, 43, 5, 190, 0, 0, 5, 154, 0, 0, 5, 5, 5, 43, 5, 190, 0, 0, 5, 154, 0, 0, 134, 5, 5, 43, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 5, 43, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 76, 185, 80, 185, 84, 185, 60, 185, 92, 185, 96, 185, 100, 185, 64, 186, 112, 185, 116, 185, 120, 185, 124, 186, 128, 185, 68, 185, 140, 185, 144, 185, 152, 186, 156, 185, 160, 185, 72, 185, 168, 185, 172, 185, 176, 185, 180, 185, 184, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 76, 168, 80, 168, 84, 168, 60, 168, 92, 168, 96, 168, 100, 168, 64, 168, 112, 168, 116, 168, 120, 168, 124, 168, 128, 168, 68, 168, 140, 168, 144, 168, 152, 168, 156, 168, 160, 168, 72, 168, 168, 168, 172, 168, 176, 168, 180, 168, 184, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 136, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 82, 185, 61, 153, 61, 249, 61, 89, 62, 185, 65, 153, 65, 249, 65, 89, 66, 217, 84, 154, 130, 185, 69, 153, 69, 249, 69, 153, 70, 89, 70, 136, 14, 217, 138, 185, 73, 153, 73, 249, 73, 89, 74, 153, 181, 185, 188, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 82, 168, 61, 136, 61, 232, 61, 72, 62, 168, 65, 136, 65, 232, 65, 72, 66, 200, 84, 136, 130, 168, 69, 136, 69, 232, 69, 136, 70, 72, 70, 128, 14, 200, 138, 168, 73, 136, 73, 232, 73, 72, 74, 136, 181, 168, 188, 72, 182, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 81, 136, 81, 249, 81, 232, 81, 185, 82, 168, 82, 25, 82, 8, 82, 25, 86, 8, 86, 89, 87, 72, 87, 57, 63, 40, 63, 217, 61, 200, 61, 185, 62, 168, 62, 25, 63, 8, 63, 25, 62, 8, 62, 249, 97, 232, 97, 217, 97, 200, 97, 185, 98, 168, 98, 249, 98, 232, 98, 249, 101, 232, 101, 89, 103, 72, 103, 153, 66, 136, 66, 57, 67, 40, 67, 217, 65, 200, 65, 25, 67, 8, 67, 185, 66, 168, 108, 6, 8, 8, 8, 250, 113, 232, 113, 249, 118, 232, 118, 168, 148, 153, 121, 136, 121, 249, 122, 232, 122, 25, 122, 8, 122, 121, 123, 104, 123, 89, 123, 72, 123, 154, 129, 136, 129, 250, 130, 232, 130, 26, 130, 8, 130, 10, 8, 185, 132, 168, 132, 57, 71, 40, 71, 217, 69, 200, 69, 121, 70, 104, 70, 12, 8, 14, 8, 153, 153, 136, 153, 249, 154, 232, 154, 25, 154, 8, 154, 154, 157, 136, 157, 250, 157, 232, 157, 250, 158, 232, 158, 26, 158, 8, 158, 249, 162, 232, 162, 25, 162, 8, 162, 185, 164, 168, 164, 153, 74, 136, 74, 57, 75, 40, 75, 217, 73, 200, 73, 57, 74, 40, 74, 121, 74, 104, 74, 25, 75, 8, 75, 249, 173, 232, 173, 249, 181, 232, 181, 89, 182, 153, 185, 136, 185, 185, 186, 168, 186, 25, 186, 8, 186, 232, 156, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 186, 88, 200, 156, 169, 156, 200, 56, 169, 88, 186, 104, 187, 112, 169, 104, 169, 112, 169, 192, 169, 128, 217, 136, 186, 88, 200, 136, 169, 88, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 43, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 246, 10, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 251, 10, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 112, 15, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 116, 15, 0, 0, 116, 15, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 80, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 14, 32, 38, 0, 243, 5, 60, 0, 60, 0, 39, 0, 39, 0, 14, 32, 38, 0, 244, 5, 60, 0, 60, 0, 39, 0, 34, 0, 39, 0, 0, 0, 65, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 72, 0, 101, 0, 98, 0, 114, 0, 93, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 39, 0, 39, 0, 60, 0, 60, 0, 243, 5, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 39, 0, 34, 0, 39, 0, 60, 0, 60, 0, 244, 5, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 106, 34, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 61, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 228, 31, 0, 0, 228, 31, 0, 0, 228, 31, 0, 0, 228, 31, 0, 0, 12, 34, 0, 0, 12, 34, 0, 0, 82, 34, 0, 0, 82, 34, 0, 0, 82, 34, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 241, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 94, 0, 102, 0, 110, 0, 118, 0, 48, 0, 56, 0, 117, 0, 125, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 142, 0, 150, 0, 48, 0, 48, 0, 48, 0, 158, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 48, 0, 48, 0, 174, 0, 48, 0, 180, 0, 188, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 191, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 199, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 202, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 140, 0, 48, 0, 48, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 210, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 218, 0, 48, 0, 48, 0, 224, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 120, 1, 184, 1, 192, 0, 212, 1, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 253, 0, 0, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 194, 5, 221, 251, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 16, 0, 202, 33, 16, 0, 202, 66, 16, 0, 202, 99, 16, 0, 202, 132, 16, 0, 202, 165, 16, 0, 202, 198, 16, 0, 202, 231, 16, 0, 202, 8, 17, 0, 202, 41, 17, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 6, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 7, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 7, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 220, 251, 5, 252, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 198, 0, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 2, 0, 5, 71, 238, 102, 201, 134, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 0, 0, 192, 0, 0, 0, 197, 226, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 1, 0, 192, 0, 0, 0, 197, 162, 1, 0, 192, 0, 0, 0, 197, 226, 1, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 197, 98, 2, 0, 197, 163, 2, 0, 197, 2, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 3, 0, 197, 131, 3, 0, 197, 226, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 34, 0, 0, 197, 98, 4, 0, 5, 5, 10, 121, 197, 162, 4, 0, 197, 226, 4, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 34, 0, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 98, 0, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 162, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 34, 1, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 98, 1, 0, 5, 5, 107, 121, 197, 162, 1, 0, 5, 5, 111, 121, 197, 226, 1, 0, 5, 5, 115, 121, 197, 34, 2, 0, 5, 5, 119, 121, 197, 98, 2, 0, 197, 163, 2, 0, 197, 2, 3, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 66, 3, 0, 197, 131, 3, 0, 197, 226, 3, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 34, 4, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 34, 0, 0, 197, 98, 4, 0, 5, 5, 10, 121, 197, 162, 4, 0, 197, 226, 4, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 83, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 188, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 64, 0, 0, 0, 64, 1, 0, 0, 96, 25, 0, 0, 96, 25, 0, 0, 96, 25, 0, 0, 96, 25, 0, 0, 164, 25, 0, 0, 19, 0, 0, 0, 0, 0, 0, 101, 194, 255, 50, 101, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 0, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 121, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 104, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 2, 106, 9, 5, 2, 120, 9, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 67, 0, 58, 0, 32, 0, 247, 8, 0, 16, 46, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 239, 8, 0, 32, 103, 15, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 108, 15, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 172, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 176, 6, 0, 0, 176, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 68, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 86, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 80, 9, 60, 0, 2, 9, 60, 0, 60, 0, 1, 9, 60, 0, 3, 9, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 116, 25, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 13, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 108, 0, 0, 0, 108, 1, 0, 0, 12, 25, 0, 0, 12, 25, 0, 0, 20, 25, 0, 0, 20, 25, 0, 0, 88, 25, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 39, 40, 41, 42, 43, 44, 45, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 113, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 193, 2, 7, 104, 193, 3, 7, 104, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 5, 0, 70, 0, 2, 7, 104, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 69, 0, 60, 0, 32, 0, 71, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 165, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 170, 6, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 220, 19, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 224, 19, 0, 0, 224, 19, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 129, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 65, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 104, 0, 114, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 111, 0, 116, 0, 104, 0, 101, 0, 114, 0, 115, 0, 93, 0, 0, 0, 136, 220, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 76, 0, 97, 0, 116, 0, 110, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 93, 0, 38, 0, 67, 0, 60, 0, 13, 1, 60, 0, 60, 0, 60, 0, 12, 1, 60, 0, 7, 1, 60, 0, 60, 0, 60, 0, 6, 1, 38, 0, 68, 0, 60, 0, 100, 0, 126, 1, 60, 0, 60, 0, 60, 0, 198, 1, 60, 0, 60, 0, 60, 0, 68, 0, 126, 1, 60, 0, 60, 0, 60, 0, 197, 1, 60, 0, 60, 0, 60, 0, 68, 0, 125, 1, 60, 0, 60, 0, 60, 0, 196, 1, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 38, 0, 76, 0, 60, 0, 108, 0, 106, 0, 60, 0, 60, 0, 60, 0, 201, 1, 60, 0, 60, 0, 60, 0, 76, 0, 106, 0, 60, 0, 60, 0, 60, 0, 200, 1, 60, 0, 60, 0, 60, 0, 76, 0, 74, 0, 60, 0, 60, 0, 60, 0, 199, 1, 38, 0, 78, 0, 60, 0, 110, 0, 106, 0, 60, 0, 60, 0, 60, 0, 204, 1, 60, 0, 60, 0, 60, 0, 78, 0, 106, 0, 60, 0, 60, 0, 60, 0, 203, 1, 60, 0, 60, 0, 60, 0, 78, 0, 74, 0, 60, 0, 60, 0, 60, 0, 202, 1, 38, 0, 83, 0, 60, 0, 97, 1, 60, 0, 60, 0, 60, 0, 96, 1, 38, 0, 90, 0, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 252, 40, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 113, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 180, 35, 0, 0, 180, 35, 0, 0, 180, 35, 0, 0, 180, 35, 0, 0, 172, 38, 0, 0, 172, 38, 0, 0, 198, 40, 0, 0, 228, 40, 0, 0, 228, 40, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 46, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 89, 0, 97, 0, 48, 0, 56, 0, 105, 0, 113, 0, 115, 0, 123, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 122, 0, 130, 0, 136, 0, 144, 0, 152, 0, 160, 0, 48, 0, 56, 0, 159, 0, 167, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 184, 0, 192, 0, 48, 0, 48, 0, 48, 0, 200, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 208, 0, 48, 0, 48, 0, 216, 0, 48, 0, 222, 0, 230, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 233, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 240, 0, 48, 0, 48, 0, 248, 0, 252, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 4, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 7, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 182, 0, 48, 0, 48, 0, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 15, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 23, 1, 48, 0, 48, 0, 29, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 100, 1, 192, 0, 164, 1, 204, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 232, 1, 32, 2, 96, 2, 192, 0, 124, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 128, 22, 0, 202, 161, 22, 0, 202, 194, 22, 0, 202, 227, 22, 0, 202, 4, 23, 0, 202, 37, 23, 0, 202, 70, 23, 0, 202, 103, 23, 0, 202, 136, 23, 0, 202, 169, 23, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 198, 0, 0, 201, 68, 2, 0, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 224, 4, 0, 156, 5, 0, 66, 201, 64, 7, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 198, 9, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 166, 10, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 134, 11, 0, 201, 4, 13, 0, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 128, 14, 0, 5, 5, 0, 66, 201, 32, 16, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 198, 17, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 166, 18, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 13, 0, 201, 134, 19, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 13, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 13, 0, 201, 6, 21, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 47, 5, 5, 3, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 49, 5, 5, 3, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 22, 0, 201, 70, 23, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 138, 5, 2, 49, 72, 5, 2, 49, 6, 5, 2, 49, 138, 5, 2, 65, 72, 5, 2, 65, 6, 5, 2, 65, 138, 5, 2, 69, 72, 5, 2, 69, 6, 5, 2, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 24, 0, 201, 198, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 134, 25, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 27, 0, 5, 71, 238, 102, 201, 70, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 5, 0, 192, 0, 0, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 5, 0, 192, 0, 0, 0, 197, 34, 6, 0, 192, 0, 0, 0, 197, 98, 6, 0, 192, 0, 0, 0, 197, 162, 6, 0, 192, 0, 0, 0, 197, 226, 6, 0, 197, 35, 7, 0, 197, 130, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 7, 0, 197, 3, 8, 0, 197, 98, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 162, 4, 0, 197, 226, 8, 0, 5, 5, 10, 121, 197, 34, 9, 0, 197, 98, 9, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 226, 10, 0, 197, 34, 11, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 5, 0, 5, 5, 24, 121, 197, 98, 5, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 2, 0, 197, 34, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 29, 0, 201, 230, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 11, 0, 197, 34, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 30, 0, 201, 102, 31, 0, 201, 38, 32, 0, 201, 230, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 137, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 144, 0, 0, 137, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 144, 0, 0, 5, 5, 2, 65, 5, 142, 0, 0, 5, 5, 2, 65, 5, 144, 0, 0, 5, 5, 2, 69, 5, 142, 0, 0, 5, 5, 2, 69, 5, 144, 0, 0, 134, 5, 3, 47, 5, 160, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 3, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 162, 4, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 226, 4, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 34, 5, 0, 5, 5, 24, 121, 197, 98, 5, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 162, 5, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 226, 5, 0, 5, 5, 107, 121, 197, 34, 6, 0, 5, 5, 111, 121, 197, 98, 6, 0, 5, 5, 115, 121, 197, 162, 6, 0, 5, 5, 119, 121, 197, 226, 6, 0, 197, 35, 7, 0, 197, 130, 7, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 194, 7, 0, 197, 3, 8, 0, 197, 98, 8, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 162, 8, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 162, 4, 0, 197, 226, 8, 0, 5, 5, 10, 121, 197, 34, 9, 0, 197, 98, 9, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 226, 10, 0, 197, 34, 11, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 5, 0, 5, 5, 24, 121, 197, 98, 5, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 46, 156, 5, 2, 0, 1, 3, 3, 239, 134, 5, 12, 3, 2, 239, 134, 5, 65, 3, 3, 239, 134, 5, 0, 48, 156, 5, 3, 0, 90, 0, 8, 0, 122, 0, 10, 0, 125, 1, 2, 241, 137, 5, 126, 1, 2, 241, 71, 5, 48, 0, 12, 3, 2, 241, 137, 5, 48, 0, 12, 3, 2, 241, 71, 5, 0, 64, 156, 5, 4, 0, 74, 0, 255, 255, 2, 65, 137, 5, 106, 0, 255, 255, 2, 65, 71, 5, 52, 1, 197, 162, 53, 1, 0, 192, 197, 98, 240, 1, 0, 192, 197, 162, 0, 68, 156, 5, 4, 0, 74, 0, 255, 255, 2, 69, 137, 5, 106, 0, 255, 255, 2, 69, 71, 5, 52, 1, 0, 192, 197, 226, 53, 1, 1, 192, 197, 34, 240, 1, 1, 192, 197, 98, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 46, 5, 5, 2, 0, 1, 3, 3, 239, 5, 5, 12, 3, 2, 239, 5, 5, 65, 3, 3, 239, 5, 5, 0, 48, 5, 5, 1, 0, 122, 0, 3, 0, 126, 1, 2, 241, 5, 5, 48, 0, 12, 3, 2, 241, 5, 5, 0, 64, 5, 5, 2, 0, 106, 0, 255, 255, 2, 65, 5, 5, 53, 1, 1, 192, 197, 162, 240, 1, 1, 192, 197, 226, 0, 68, 5, 5, 2, 0, 106, 0, 255, 255, 2, 69, 5, 5, 53, 1, 2, 192, 197, 34, 240, 1, 2, 192, 197, 98, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 2, 0, 1, 3, 2, 192, 197, 162, 12, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 5, 46, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 12, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 156, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 226, 156, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 98, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 78, 196, 196, 48, 0, 12, 3, 11, 192, 197, 98, 5, 78, 196, 196, 48, 0, 12, 3, 11, 192, 197, 162, 156, 92, 196, 196, 48, 0, 12, 3, 12, 192, 197, 98, 5, 92, 196, 196, 48, 0, 12, 3, 12, 192, 197, 162, 156, 92, 196, 210, 48, 0, 12, 3, 12, 192, 197, 226, 5, 92, 196, 210, 48, 0, 12, 3, 13, 192, 197, 34, 14, 0, 74, 0, 75, 0, 90, 0, 91, 0, 106, 0, 107, 0, 122, 0, 123, 0, 52, 1, 54, 1, 125, 1, 127, 1, 240, 1, 241, 1, 3, 0, 67, 0, 58, 0, 32, 0, 131, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 28, 36, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 76, 0, 0, 0, 76, 1, 0, 0, 188, 28, 0, 0, 188, 28, 0, 0, 188, 28, 0, 0, 188, 28, 0, 0, 208, 29, 0, 0, 208, 29, 0, 0, 164, 31, 0, 0, 194, 31, 0, 0, 4, 36, 0, 0, 25, 0, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 94, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 174, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 89, 0, 97, 0, 48, 0, 56, 0, 105, 0, 113, 0, 115, 0, 123, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 145, 0, 48, 0, 48, 0, 153, 0, 157, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 139, 0, 48, 0, 48, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 100, 1, 192, 0, 164, 1, 204, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 7, 0, 202, 129, 7, 0, 202, 162, 7, 0, 202, 195, 7, 0, 202, 228, 7, 0, 202, 5, 8, 0, 202, 38, 8, 0, 202, 71, 8, 0, 202, 104, 8, 0, 202, 137, 8, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 201, 132, 1, 0, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 32, 4, 0, 156, 5, 0, 66, 201, 128, 6, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 6, 9, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 230, 9, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 198, 10, 0, 201, 68, 12, 0, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 192, 13, 0, 5, 5, 0, 66, 201, 96, 15, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 6, 17, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 230, 17, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 6, 0, 201, 198, 18, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 6, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 7, 0, 201, 70, 20, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 47, 5, 5, 3, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 49, 5, 5, 3, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 21, 0, 201, 134, 22, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 138, 5, 2, 49, 72, 5, 2, 49, 6, 5, 2, 49, 138, 5, 2, 65, 72, 5, 2, 65, 6, 5, 2, 65, 138, 5, 2, 69, 72, 5, 2, 69, 6, 5, 2, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 23, 0, 201, 6, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 2, 0, 197, 34, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 24, 0, 201, 134, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 5, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 26, 0, 201, 6, 27, 0, 201, 198, 27, 0, 201, 134, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 137, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 144, 0, 0, 137, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 144, 0, 0, 5, 5, 2, 65, 5, 142, 0, 0, 5, 5, 2, 65, 5, 144, 0, 0, 5, 5, 2, 69, 5, 142, 0, 0, 5, 5, 2, 69, 5, 144, 0, 0, 134, 5, 3, 47, 5, 160, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 3, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 2, 0, 1, 3, 3, 239, 134, 5, 12, 3, 2, 239, 134, 5, 65, 3, 3, 239, 134, 5, 0, 48, 156, 5, 3, 0, 90, 0, 8, 0, 122, 0, 10, 0, 125, 1, 2, 241, 137, 5, 126, 1, 2, 241, 71, 5, 48, 0, 12, 3, 2, 241, 137, 5, 48, 0, 12, 3, 2, 241, 71, 5, 0, 64, 156, 5, 4, 0, 74, 0, 255, 255, 2, 65, 137, 5, 106, 0, 255, 255, 2, 65, 71, 5, 52, 1, 197, 162, 53, 1, 0, 192, 197, 98, 240, 1, 0, 192, 197, 162, 0, 68, 156, 5, 4, 0, 74, 0, 255, 255, 2, 69, 137, 5, 106, 0, 255, 255, 2, 69, 71, 5, 52, 1, 0, 192, 197, 226, 53, 1, 1, 192, 197, 34, 240, 1, 1, 192, 197, 98, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 46, 5, 5, 2, 0, 1, 3, 3, 239, 5, 5, 12, 3, 2, 239, 5, 5, 65, 3, 3, 239, 5, 5, 0, 48, 5, 5, 1, 0, 122, 0, 3, 0, 126, 1, 2, 241, 5, 5, 48, 0, 12, 3, 2, 241, 5, 5, 0, 64, 5, 5, 2, 0, 106, 0, 255, 255, 2, 65, 5, 5, 53, 1, 1, 192, 197, 162, 240, 1, 1, 192, 197, 226, 0, 68, 5, 5, 2, 0, 106, 0, 255, 255, 2, 69, 5, 5, 53, 1, 2, 192, 197, 34, 240, 1, 2, 192, 197, 98, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 2, 0, 1, 3, 2, 192, 197, 162, 12, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 5, 46, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 12, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 156, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 226, 156, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 98, 156, 78, 196, 196, 48, 0, 12, 3, 4, 192, 197, 162, 5, 78, 196, 196, 48, 0, 12, 3, 4, 192, 197, 226, 156, 92, 196, 196, 48, 0, 12, 3, 5, 192, 197, 162, 5, 92, 196, 196, 48, 0, 12, 3, 5, 192, 197, 226, 156, 92, 196, 210, 48, 0, 12, 3, 6, 192, 197, 34, 5, 92, 196, 210, 48, 0, 12, 3, 6, 192, 197, 98, 14, 0, 74, 0, 75, 0, 90, 0, 91, 0, 106, 0, 107, 0, 122, 0, 123, 0, 52, 1, 54, 1, 125, 1, 127, 1, 240, 1, 241, 1, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 28, 4, 186, 88, 185, 92, 185, 96, 185, 100, 186, 104, 186, 112, 185, 116, 36, 4, 185, 128, 46, 4, 185, 144, 185, 148, 185, 152, 185, 160, 56, 4, 185, 172, 185, 180, 185, 184, 185, 188, 185, 192, 185, 196, 58, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 60, 4, 62, 4, 168, 88, 168, 92, 168, 96, 168, 100, 168, 104, 168, 112, 168, 116, 67, 4, 168, 128, 73, 4, 168, 144, 168, 148, 168, 152, 168, 160, 79, 4, 168, 172, 168, 180, 168, 184, 168, 188, 168, 192, 168, 196, 81, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 144, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 83, 4, 186, 89, 154, 89, 250, 89, 90, 90, 186, 105, 154, 105, 250, 105, 90, 106, 217, 76, 154, 134, 185, 145, 153, 145, 249, 145, 153, 146, 89, 146, 136, 14, 217, 146, 185, 181, 153, 181, 249, 181, 89, 182, 153, 197, 185, 208, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 85, 4, 168, 89, 136, 89, 232, 89, 72, 90, 168, 105, 136, 105, 232, 105, 72, 106, 200, 76, 136, 134, 168, 145, 136, 145, 232, 145, 136, 146, 72, 146, 128, 14, 200, 146, 168, 181, 136, 181, 232, 181, 72, 182, 136, 197, 168, 208, 72, 198, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 185, 72, 168, 72, 249, 65, 232, 65, 185, 66, 168, 66, 185, 68, 168, 68, 25, 78, 8, 78, 185, 84, 168, 84, 58, 91, 40, 91, 218, 89, 200, 89, 186, 90, 168, 90, 26, 91, 8, 91, 26, 90, 8, 90, 249, 97, 232, 97, 217, 97, 200, 97, 185, 98, 168, 98, 249, 98, 232, 98, 249, 101, 232, 101, 89, 103, 72, 103, 154, 106, 136, 106, 58, 107, 40, 107, 218, 105, 200, 105, 26, 107, 8, 107, 186, 106, 168, 108, 6, 8, 8, 8, 250, 113, 232, 113, 249, 118, 232, 118, 168, 156, 153, 121, 136, 121, 249, 122, 232, 122, 25, 122, 8, 122, 121, 123, 104, 123, 89, 123, 72, 123, 154, 133, 136, 133, 250, 134, 232, 134, 26, 134, 8, 134, 10, 8, 185, 140, 168, 140, 57, 147, 40, 147, 217, 145, 200, 145, 121, 146, 104, 146, 12, 8, 14, 8, 153, 161, 136, 161, 249, 162, 232, 162, 25, 162, 8, 162, 154, 165, 136, 165, 250, 165, 232, 165, 87, 4, 89, 4, 185, 168, 168, 168, 249, 174, 232, 174, 25, 174, 8, 174, 185, 176, 168, 176, 153, 182, 136, 182, 57, 183, 40, 183, 217, 181, 200, 181, 57, 182, 40, 182, 121, 182, 104, 182, 25, 183, 8, 183, 249, 189, 232, 189, 249, 197, 232, 197, 89, 198, 153, 201, 136, 201, 185, 202, 168, 202, 185, 204, 168, 204, 232, 164, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 88, 200, 164, 169, 164, 200, 56, 169, 88, 187, 104, 187, 112, 169, 104, 169, 112, 169, 212, 169, 132, 217, 144, 187, 88, 200, 144, 169, 88, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 185, 76, 90, 2, 122, 2, 125, 5, 186, 80, 126, 5, 177, 80, 255, 5, 185, 120, 74, 4, 186, 124, 106, 4, 177, 124, 52, 5, 250, 125, 53, 5, 241, 125, 255, 5, 186, 132, 74, 4, 186, 136, 106, 4, 177, 136, 52, 5, 250, 137, 53, 5, 241, 137, 255, 5, 186, 164, 255, 5, 185, 200, 255, 5, 168, 64, 255, 5, 168, 76, 122, 2, 126, 5, 168, 80, 255, 5, 168, 120, 106, 4, 168, 124, 53, 5, 232, 125, 255, 5, 168, 132, 106, 4, 168, 136, 53, 5, 232, 137, 255, 5, 168, 164, 255, 5, 168, 200, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 250, 166, 255, 5, 232, 166, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 203, 10, 0, 16, 73, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 195, 10, 0, 32, 211, 19, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 216, 19, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 221, 9, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 225, 9, 0, 0, 225, 9, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 71, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 220, 38, 0, 67, 0, 60, 0, 13, 1, 60, 0, 60, 0, 60, 0, 12, 1, 60, 0, 7, 1, 60, 0, 60, 0, 60, 0, 6, 1, 38, 0, 69, 0, 60, 0, 27, 1, 60, 0, 60, 0, 60, 0, 26, 1, 38, 0, 72, 0, 60, 0, 99, 0, 104, 0, 60, 0, 60, 0, 60, 0, 99, 0, 72, 0, 60, 0, 60, 0, 60, 0, 67, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 72, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 32, 0, 76, 0, 60, 0, 66, 1, 60, 0, 60, 0, 60, 0, 65, 1, 38, 0, 82, 0, 60, 0, 89, 1, 60, 0, 60, 0, 60, 0, 88, 1, 38, 0, 83, 0, 60, 0, 97, 1, 60, 0, 60, 0, 60, 0, 96, 1, 38, 0, 90, 0, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 60, 0, 122, 1, 60, 0, 60, 0, 60, 0, 121, 1, 0, 0, 170, 170, 56, 38, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 84, 28, 0, 0, 84, 28, 0, 0, 84, 28, 0, 0, 84, 28, 0, 0, 168, 30, 0, 0, 168, 30, 0, 0, 222, 33, 0, 0, 248, 33, 0, 0, 32, 38, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 184, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 48, 0, 56, 0, 48, 0, 56, 0, 112, 0, 120, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 142, 0, 48, 0, 149, 0, 157, 0, 161, 0, 167, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 136, 0, 48, 0, 48, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 0, 192, 0, 192, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 17, 0, 202, 129, 17, 0, 202, 162, 17, 0, 202, 195, 17, 0, 202, 228, 17, 0, 202, 5, 18, 0, 202, 38, 18, 0, 202, 71, 18, 0, 202, 104, 18, 0, 202, 137, 18, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 4, 0, 0, 156, 5, 0, 48, 201, 134, 8, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 201, 70, 9, 0, 201, 38, 10, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 6, 11, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 228, 12, 0, 5, 5, 0, 48, 201, 134, 21, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 201, 70, 22, 0, 201, 38, 23, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 6, 24, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 16, 0, 201, 230, 25, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 16, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 17, 0, 201, 102, 27, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 47, 5, 5, 3, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 28, 0, 201, 166, 29, 0, 134, 5, 2, 51, 5, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 21, 63, 5, 5, 21, 63, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 30, 0, 201, 38, 31, 0, 134, 5, 2, 77, 5, 5, 2, 77, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 31, 0, 201, 166, 32, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 93, 5, 5, 3, 93, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 33, 0, 201, 38, 34, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 34, 0, 201, 166, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 7, 0, 197, 34, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 36, 0, 201, 38, 37, 0, 201, 230, 37, 0, 201, 166, 38, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 39, 0, 201, 38, 40, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 40, 0, 201, 166, 41, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 42, 0, 201, 38, 43, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 13, 0, 197, 226, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 43, 0, 201, 102, 45, 0, 201, 230, 46, 0, 201, 102, 48, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 49, 0, 201, 166, 50, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 136, 5, 2, 57, 5, 142, 0, 0, 71, 5, 2, 57, 5, 142, 0, 0, 136, 5, 2, 57, 5, 144, 0, 0, 71, 5, 2, 57, 5, 144, 0, 0, 136, 5, 2, 57, 5, 156, 0, 0, 71, 5, 2, 57, 5, 156, 0, 0, 136, 5, 2, 57, 5, 196, 0, 0, 71, 5, 2, 57, 5, 196, 0, 0, 136, 5, 2, 57, 5, 150, 0, 0, 71, 5, 2, 57, 5, 150, 0, 0, 136, 5, 2, 57, 5, 160, 0, 0, 71, 5, 2, 57, 5, 160, 0, 0, 136, 5, 2, 57, 5, 206, 0, 0, 71, 5, 2, 57, 5, 206, 0, 0, 71, 5, 2, 57, 5, 210, 0, 0, 70, 5, 2, 57, 5, 142, 0, 0, 5, 5, 2, 57, 5, 142, 0, 0, 70, 5, 2, 57, 5, 144, 0, 0, 5, 5, 2, 57, 5, 144, 0, 0, 70, 5, 2, 57, 5, 156, 0, 0, 5, 5, 2, 57, 5, 156, 0, 0, 70, 5, 2, 57, 5, 196, 0, 0, 5, 5, 2, 57, 5, 196, 0, 0, 70, 5, 2, 57, 5, 150, 0, 0, 5, 5, 2, 57, 5, 150, 0, 0, 70, 5, 2, 57, 5, 160, 0, 0, 5, 5, 2, 57, 5, 160, 0, 0, 70, 5, 2, 57, 5, 206, 0, 0, 5, 5, 2, 57, 5, 206, 0, 0, 5, 5, 2, 57, 5, 210, 0, 0, 134, 5, 3, 47, 5, 160, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 3, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 51, 5, 162, 0, 0, 5, 5, 2, 51, 5, 162, 0, 0, 134, 5, 2, 77, 5, 160, 0, 0, 5, 5, 2, 77, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 51, 5, 160, 0, 0, 5, 5, 2, 51, 5, 160, 0, 0, 134, 5, 2, 51, 5, 204, 0, 0, 5, 5, 2, 51, 5, 204, 0, 0, 134, 5, 2, 51, 5, 208, 0, 0, 5, 5, 2, 51, 5, 208, 0, 0, 134, 5, 2, 77, 5, 196, 0, 0, 5, 5, 2, 77, 5, 196, 0, 0, 134, 5, 2, 77, 5, 210, 0, 0, 5, 5, 2, 77, 5, 210, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 3, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 3, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 3, 93, 5, 210, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 3, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 134, 5, 2, 51, 5, 196, 0, 0, 5, 5, 2, 51, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 19, 0, 35, 30, 32, 0, 40, 30, 15, 0, 40, 30, 2, 192, 197, 162, 41, 30, 2, 192, 197, 226, 42, 30, 3, 192, 197, 34, 43, 30, 3, 192, 197, 98, 150, 30, 3, 192, 197, 162, 35, 30, 1, 192, 197, 98, 36, 30, 1, 192, 197, 162, 37, 30, 1, 192, 197, 226, 38, 30, 2, 192, 197, 34, 39, 30, 2, 192, 197, 98, 31, 2, 15, 0, 31, 2, 0, 192, 197, 226, 1, 3, 3, 239, 134, 5, 12, 3, 2, 239, 134, 5, 65, 3, 3, 239, 134, 5, 34, 30, 1, 192, 197, 34, 72, 0, 2, 249, 136, 5, 104, 0, 2, 249, 71, 5, 36, 1, 197, 162, 37, 1, 0, 192, 197, 98, 30, 2, 0, 192, 197, 162, 0, 50, 156, 5, 48, 0, 12, 3, 2, 243, 134, 5, 0, 76, 156, 5, 48, 0, 12, 3, 255, 255, 2, 77, 134, 5, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 2, 0, 1, 3, 255, 255, 3, 93, 134, 5, 12, 3, 255, 255, 2, 93, 134, 5, 65, 3, 255, 255, 3, 93, 134, 5, 0, 46, 5, 5, 19, 0, 35, 30, 32, 0, 40, 30, 15, 0, 40, 30, 6, 192, 197, 98, 41, 30, 6, 192, 197, 162, 42, 30, 6, 192, 197, 226, 43, 30, 7, 192, 197, 34, 150, 30, 7, 192, 197, 98, 35, 30, 5, 192, 197, 34, 36, 30, 5, 192, 197, 98, 37, 30, 5, 192, 197, 162, 38, 30, 5, 192, 197, 226, 39, 30, 6, 192, 197, 34, 31, 2, 15, 0, 31, 2, 4, 192, 197, 162, 1, 3, 3, 239, 5, 5, 12, 3, 2, 239, 5, 5, 65, 3, 3, 239, 5, 5, 34, 30, 4, 192, 197, 226, 72, 0, 2, 249, 70, 5, 104, 0, 2, 249, 5, 5, 36, 1, 3, 192, 197, 226, 37, 1, 4, 192, 197, 34, 30, 2, 4, 192, 197, 98, 0, 50, 5, 5, 48, 0, 12, 3, 2, 243, 5, 5, 0, 76, 5, 5, 48, 0, 12, 3, 255, 255, 2, 77, 5, 5, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 2, 0, 1, 3, 255, 255, 3, 93, 5, 5, 12, 3, 255, 255, 2, 93, 5, 5, 65, 3, 255, 255, 3, 93, 5, 5, 156, 46, 196, 160, 2, 0, 1, 3, 7, 192, 197, 162, 12, 3, 7, 192, 197, 226, 65, 3, 7, 192, 197, 162, 5, 46, 196, 160, 2, 0, 1, 3, 8, 192, 197, 34, 12, 3, 8, 192, 197, 98, 65, 3, 8, 192, 197, 34, 156, 50, 196, 162, 48, 0, 12, 3, 8, 192, 197, 162, 5, 50, 196, 162, 48, 0, 12, 3, 8, 192, 197, 226, 156, 76, 196, 160, 48, 0, 12, 3, 9, 192, 197, 34, 5, 76, 196, 160, 48, 0, 12, 3, 9, 192, 197, 98, 156, 78, 196, 160, 48, 0, 12, 3, 9, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 9, 192, 197, 226, 156, 78, 196, 202, 48, 0, 12, 3, 10, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 10, 192, 197, 98, 156, 50, 196, 160, 48, 0, 12, 3, 10, 192, 197, 162, 5, 50, 196, 160, 48, 0, 12, 3, 10, 192, 197, 226, 156, 50, 196, 204, 48, 0, 12, 3, 11, 192, 197, 34, 5, 50, 196, 204, 48, 0, 12, 3, 11, 192, 197, 98, 156, 50, 196, 208, 48, 0, 12, 3, 11, 192, 197, 162, 5, 50, 196, 208, 48, 0, 12, 3, 11, 192, 197, 226, 156, 76, 196, 196, 48, 0, 12, 3, 12, 192, 197, 34, 5, 76, 196, 196, 48, 0, 12, 3, 12, 192, 197, 98, 156, 76, 196, 210, 48, 0, 12, 3, 12, 192, 197, 162, 5, 76, 196, 210, 48, 0, 12, 3, 12, 192, 197, 226, 156, 78, 196, 196, 48, 0, 12, 3, 13, 192, 197, 34, 5, 78, 196, 196, 48, 0, 12, 3, 13, 192, 197, 98, 156, 92, 196, 196, 2, 0, 1, 3, 14, 192, 197, 34, 12, 3, 14, 192, 197, 98, 65, 3, 14, 192, 197, 34, 5, 92, 196, 196, 2, 0, 1, 3, 14, 192, 197, 162, 12, 3, 14, 192, 197, 226, 65, 3, 14, 192, 197, 162, 156, 92, 196, 210, 2, 0, 1, 3, 15, 192, 197, 34, 12, 3, 15, 192, 197, 98, 65, 3, 15, 192, 197, 34, 5, 92, 196, 210, 2, 0, 1, 3, 15, 192, 197, 162, 12, 3, 15, 192, 197, 226, 65, 3, 15, 192, 197, 162, 156, 50, 196, 196, 48, 0, 12, 3, 16, 192, 197, 34, 5, 50, 196, 196, 48, 0, 12, 3, 16, 192, 197, 98, 12, 0, 72, 0, 73, 0, 104, 0, 105, 0, 36, 1, 38, 1, 30, 2, 32, 2, 34, 30, 44, 30, 150, 30, 151, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 76, 36, 4, 185, 88, 185, 92, 185, 96, 186, 104, 186, 112, 185, 116, 185, 124, 185, 128, 186, 132, 185, 140, 185, 144, 185, 148, 38, 4, 40, 4, 185, 172, 185, 180, 185, 184, 185, 188, 185, 192, 185, 196, 42, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 44, 4, 168, 76, 54, 4, 168, 88, 168, 92, 168, 96, 168, 104, 168, 112, 168, 116, 168, 124, 168, 128, 168, 132, 168, 140, 168, 144, 168, 148, 56, 4, 58, 4, 168, 172, 168, 180, 168, 184, 168, 188, 168, 192, 168, 196, 60, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 140, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 62, 4, 186, 81, 154, 81, 250, 81, 90, 82, 186, 105, 154, 105, 250, 105, 90, 106, 217, 76, 154, 134, 185, 141, 153, 141, 249, 141, 153, 142, 89, 142, 136, 14, 217, 142, 185, 181, 153, 181, 249, 181, 89, 182, 153, 197, 185, 212, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 64, 4, 168, 81, 136, 81, 232, 81, 72, 82, 168, 105, 136, 105, 232, 105, 72, 106, 200, 76, 136, 134, 168, 141, 136, 141, 232, 141, 136, 142, 72, 142, 128, 14, 200, 142, 168, 181, 136, 181, 232, 181, 72, 182, 136, 197, 168, 212, 72, 198, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 185, 72, 168, 72, 249, 65, 232, 65, 185, 66, 168, 66, 185, 68, 168, 68, 25, 78, 8, 78, 89, 79, 72, 79, 58, 83, 40, 83, 218, 81, 200, 81, 186, 82, 168, 82, 66, 4, 68, 4, 185, 84, 168, 84, 249, 93, 232, 93, 217, 93, 200, 93, 185, 94, 168, 94, 249, 94, 232, 94, 249, 97, 232, 97, 89, 99, 72, 99, 154, 106, 136, 106, 58, 107, 40, 107, 218, 105, 200, 105, 26, 107, 8, 107, 186, 106, 168, 108, 6, 8, 8, 8, 250, 113, 232, 113, 249, 118, 232, 118, 168, 152, 153, 125, 136, 125, 249, 126, 232, 126, 25, 126, 8, 126, 121, 127, 104, 127, 185, 120, 168, 120, 154, 133, 136, 133, 250, 134, 232, 134, 26, 134, 8, 134, 10, 8, 185, 136, 168, 136, 57, 143, 40, 143, 217, 141, 200, 141, 121, 142, 104, 142, 12, 8, 14, 8, 153, 157, 136, 157, 70, 4, 72, 4, 185, 160, 168, 160, 154, 165, 136, 165, 250, 165, 232, 165, 74, 4, 76, 4, 185, 168, 168, 168, 249, 174, 232, 174, 25, 174, 8, 174, 185, 176, 168, 176, 153, 182, 136, 182, 57, 183, 40, 183, 217, 181, 200, 181, 57, 182, 40, 182, 121, 182, 104, 182, 25, 183, 8, 183, 249, 189, 232, 189, 249, 197, 232, 197, 89, 198, 185, 208, 168, 208, 185, 202, 168, 202, 185, 204, 168, 204, 232, 164, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 80, 200, 164, 169, 164, 200, 56, 169, 80, 187, 104, 187, 112, 169, 104, 169, 112, 169, 216, 169, 132, 217, 140, 187, 80, 200, 140, 169, 80, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 72, 4, 187, 100, 104, 4, 178, 100, 36, 5, 251, 101, 37, 5, 242, 101, 255, 5, 186, 80, 255, 5, 185, 156, 255, 5, 186, 164, 255, 5, 185, 200, 255, 5, 168, 64, 72, 4, 177, 100, 104, 4, 168, 100, 36, 5, 241, 101, 37, 5, 232, 101, 255, 5, 168, 80, 255, 5, 168, 156, 255, 5, 168, 164, 255, 5, 168, 200, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 26, 83, 255, 5, 8, 83, 255, 5, 249, 158, 255, 5, 232, 158, 255, 5, 250, 166, 255, 5, 232, 166, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 71, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 214, 9, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 219, 9, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 18, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 108, 18, 0, 0, 108, 18, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 20, 1, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 246, 221, 38, 0, 67, 0, 60, 0, 99, 0, 115, 0, 60, 0, 60, 0, 60, 0, 67, 0, 115, 0, 60, 0, 60, 0, 60, 0, 67, 0, 83, 0, 38, 0, 68, 0, 60, 0, 100, 0, 122, 0, 60, 0, 60, 0, 60, 0, 68, 0, 122, 0, 60, 0, 60, 0, 60, 0, 68, 0, 90, 0, 38, 0, 68, 0, 90, 0, 60, 0, 100, 0, 122, 0, 115, 0, 60, 0, 60, 0, 60, 0, 68, 0, 122, 0, 115, 0, 60, 0, 60, 0, 60, 0, 68, 0, 90, 0, 83, 0, 38, 0, 71, 0, 60, 0, 103, 0, 121, 0, 60, 0, 60, 0, 60, 0, 71, 0, 121, 0, 60, 0, 60, 0, 60, 0, 71, 0, 89, 0, 38, 0, 76, 0, 60, 0, 108, 0, 121, 0, 60, 0, 60, 0, 60, 0, 76, 0, 121, 0, 60, 0, 60, 0, 60, 0, 76, 0, 89, 0, 38, 0, 78, 0, 60, 0, 110, 0, 121, 0, 60, 0, 60, 0, 60, 0, 78, 0, 121, 0, 60, 0, 60, 0, 60, 0, 78, 0, 89, 0, 38, 0, 83, 0, 60, 0, 115, 0, 122, 0, 60, 0, 60, 0, 60, 0, 83, 0, 122, 0, 60, 0, 60, 0, 60, 0, 83, 0, 90, 0, 38, 0, 84, 0, 60, 0, 116, 0, 121, 0, 60, 0, 60, 0, 60, 0, 84, 0, 121, 0, 60, 0, 60, 0, 60, 0, 84, 0, 89, 0, 38, 0, 90, 0, 60, 0, 122, 0, 115, 0, 60, 0, 60, 0, 60, 0, 90, 0, 115, 0, 60, 0, 60, 0, 60, 0, 90, 0, 83, 0, 38, 0, 79, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 81, 1, 60, 0, 60, 0, 60, 0, 80, 1, 38, 0, 85, 0, 60, 0, 252, 0, 60, 0, 60, 0, 60, 0, 220, 0, 60, 0, 60, 0, 113, 1, 60, 0, 60, 0, 60, 0, 112, 1, 38, 0, 99, 0, 115, 0, 60, 0, 60, 0, 60, 0, 99, 0, 99, 0, 115, 0, 47, 0, 99, 0, 115, 0, 38, 0, 67, 0, 115, 0, 60, 0, 60, 0, 60, 0, 67, 0, 99, 0, 115, 0, 47, 0, 99, 0, 115, 0, 38, 0, 67, 0, 83, 0, 60, 0, 60, 0, 60, 0, 67, 0, 67, 0, 83, 0, 47, 0, 67, 0, 83, 0, 38, 0, 100, 0, 122, 0, 60, 0, 60, 0, 60, 0, 100, 0, 100, 0, 122, 0, 47, 0, 100, 0, 122, 0, 38, 0, 68, 0, 122, 0, 60, 0, 60, 0, 60, 0, 68, 0, 100, 0, 122, 0, 47, 0, 100, 0, 122, 0, 38, 0, 68, 0, 90, 0, 60, 0, 60, 0, 60, 0, 68, 0, 68, 0, 90, 0, 47, 0, 68, 0, 90, 0, 38, 0, 100, 0, 122, 0, 115, 0, 60, 0, 60, 0, 60, 0, 100, 0, 100, 0, 122, 0, 115, 0, 47, 0, 100, 0, 122, 0, 115, 0, 38, 0, 68, 0, 122, 0, 115, 0, 60, 0, 60, 0, 60, 0, 68, 0, 100, 0, 122, 0, 115, 0, 47, 0, 100, 0, 122, 0, 115, 0, 38, 0, 68, 0, 90, 0, 83, 0, 60, 0, 60, 0, 60, 0, 68, 0, 68, 0, 90, 0, 83, 0, 47, 0, 68, 0, 90, 0, 83, 0, 38, 0, 103, 0, 121, 0, 60, 0, 60, 0, 60, 0, 103, 0, 103, 0, 121, 0, 47, 0, 103, 0, 121, 0, 38, 0, 71, 0, 121, 0, 60, 0, 60, 0, 60, 0, 71, 0, 103, 0, 121, 0, 47, 0, 103, 0, 121, 0, 38, 0, 71, 0, 89, 0, 60, 0, 60, 0, 60, 0, 71, 0, 71, 0, 89, 0, 47, 0, 71, 0, 89, 0, 38, 0, 108, 0, 121, 0, 60, 0, 60, 0, 60, 0, 108, 0, 108, 0, 121, 0, 47, 0, 108, 0, 121, 0, 38, 0, 76, 0, 121, 0, 60, 0, 60, 0, 60, 0, 76, 0, 108, 0, 121, 0, 47, 0, 108, 0, 121, 0, 38, 0, 76, 0, 89, 0, 60, 0, 60, 0, 60, 0, 76, 0, 76, 0, 89, 0, 47, 0, 76, 0, 89, 0, 38, 0, 110, 0, 121, 0, 60, 0, 60, 0, 60, 0, 110, 0, 110, 0, 121, 0, 47, 0, 110, 0, 121, 0, 38, 0, 78, 0, 121, 0, 60, 0, 60, 0, 60, 0, 78, 0, 110, 0, 121, 0, 47, 0, 110, 0, 121, 0, 38, 0, 78, 0, 89, 0, 60, 0, 60, 0, 60, 0, 78, 0, 78, 0, 89, 0, 47, 0, 78, 0, 89, 0, 38, 0, 115, 0, 122, 0, 60, 0, 60, 0, 60, 0, 115, 0, 115, 0, 122, 0, 47, 0, 115, 0, 122, 0, 38, 0, 83, 0, 122, 0, 60, 0, 60, 0, 60, 0, 83, 0, 115, 0, 122, 0, 47, 0, 115, 0, 122, 0, 38, 0, 83, 0, 90, 0, 60, 0, 60, 0, 60, 0, 83, 0, 83, 0, 90, 0, 47, 0, 83, 0, 90, 0, 38, 0, 116, 0, 121, 0, 60, 0, 60, 0, 60, 0, 116, 0, 116, 0, 121, 0, 47, 0, 116, 0, 121, 0, 38, 0, 84, 0, 121, 0, 60, 0, 60, 0, 60, 0, 84, 0, 116, 0, 121, 0, 47, 0, 116, 0, 121, 0, 38, 0, 84, 0, 89, 0, 60, 0, 60, 0, 60, 0, 84, 0, 84, 0, 89, 0, 47, 0, 84, 0, 89, 0, 38, 0, 122, 0, 115, 0, 60, 0, 60, 0, 60, 0, 122, 0, 122, 0, 115, 0, 47, 0, 122, 0, 115, 0, 38, 0, 90, 0, 115, 0, 60, 0, 60, 0, 60, 0, 90, 0, 122, 0, 115, 0, 47, 0, 122, 0, 115, 0, 38, 0, 90, 0, 83, 0, 60, 0, 60, 0, 60, 0, 90, 0, 90, 0, 83, 0, 47, 0, 90, 0, 83, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 34, 69, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 212, 27, 0, 0, 212, 27, 0, 0, 212, 27, 0, 0, 212, 27, 0, 0, 104, 45, 0, 0, 104, 45, 0, 0, 214, 62, 0, 0, 48, 63, 0, 0, 10, 69, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 176, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 80, 0, 88, 0, 93, 0, 101, 0, 106, 0, 114, 0, 117, 0, 125, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 145, 0, 48, 0, 48, 0, 151, 0, 159, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 167, 0, 141, 0, 48, 0, 48, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 64, 1, 116, 1, 168, 1, 212, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 139, 0, 202, 129, 139, 0, 202, 162, 139, 0, 202, 195, 139, 0, 202, 228, 139, 0, 202, 5, 140, 0, 202, 38, 140, 0, 202, 71, 140, 0, 202, 104, 140, 0, 202, 137, 140, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 0, 0, 0, 201, 32, 20, 0, 156, 5, 0, 50, 156, 5, 0, 52, 201, 32, 52, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 64, 71, 0, 156, 5, 0, 66, 201, 160, 90, 0, 201, 6, 110, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 96, 111, 0, 201, 128, 124, 0, 201, 230, 143, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 160, 145, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 64, 166, 0, 201, 128, 176, 0, 5, 5, 0, 50, 5, 5, 0, 52, 201, 160, 192, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 224, 202, 0, 5, 5, 0, 66, 201, 64, 213, 0, 201, 166, 223, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 0, 225, 0, 201, 160, 231, 0, 201, 6, 242, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 192, 243, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 138, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 134, 5, 2, 71, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 5, 2, 83, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 138, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 139, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 5, 5, 2, 71, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 5, 2, 83, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 2, 71, 5, 70, 2, 71, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 2, 83, 5, 70, 2, 83, 201, 38, 254, 0, 201, 70, 255, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 0, 1, 201, 134, 1, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 2, 1, 201, 198, 3, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 126, 0, 197, 98, 126, 0, 197, 226, 76, 0, 197, 130, 118, 0, 197, 162, 126, 0, 197, 226, 126, 0, 197, 34, 127, 0, 197, 98, 127, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 4, 1, 201, 6, 6, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 128, 0, 197, 226, 128, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 7, 1, 201, 70, 8, 1, 201, 102, 9, 1, 201, 134, 10, 1, 201, 166, 11, 1, 201, 198, 12, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 13, 1, 201, 6, 15, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 16, 1, 201, 70, 17, 1, 201, 102, 18, 1, 201, 134, 19, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 20, 1, 201, 198, 21, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 138, 5, 2, 47, 137, 5, 2, 47, 138, 5, 2, 47, 137, 5, 2, 47, 5, 136, 0, 0, 138, 5, 2, 47, 137, 5, 2, 47, 5, 142, 0, 0, 138, 5, 2, 47, 137, 5, 2, 47, 5, 160, 0, 0, 138, 5, 2, 47, 137, 5, 2, 47, 5, 144, 0, 0, 138, 5, 2, 47, 137, 5, 2, 47, 5, 202, 0, 0, 138, 5, 2, 47, 137, 5, 2, 47, 5, 156, 0, 0, 138, 5, 2, 47, 137, 5, 2, 47, 5, 196, 0, 0, 138, 5, 2, 47, 137, 5, 2, 47, 5, 136, 0, 0, 5, 156, 0, 0, 138, 5, 2, 47, 137, 5, 2, 47, 5, 144, 0, 0, 5, 156, 0, 0, 138, 5, 2, 47, 137, 5, 2, 47, 5, 196, 0, 0, 5, 156, 0, 0, 72, 5, 2, 47, 5, 5, 2, 47, 72, 5, 2, 47, 5, 5, 2, 47, 5, 136, 0, 0, 72, 5, 2, 47, 5, 5, 2, 47, 5, 142, 0, 0, 72, 5, 2, 47, 5, 5, 2, 47, 5, 160, 0, 0, 72, 5, 2, 47, 5, 5, 2, 47, 5, 144, 0, 0, 72, 5, 2, 47, 5, 5, 2, 47, 5, 202, 0, 0, 72, 5, 2, 47, 5, 5, 2, 47, 5, 156, 0, 0, 72, 5, 2, 47, 5, 5, 2, 47, 5, 196, 0, 0, 72, 5, 2, 47, 5, 5, 2, 47, 5, 136, 0, 0, 5, 156, 0, 0, 72, 5, 2, 47, 5, 5, 2, 47, 5, 144, 0, 0, 5, 156, 0, 0, 72, 5, 2, 47, 5, 5, 2, 47, 5, 196, 0, 0, 5, 156, 0, 0, 71, 5, 2, 47, 5, 136, 0, 0, 71, 5, 2, 47, 5, 142, 0, 0, 71, 5, 2, 47, 5, 160, 0, 0, 71, 5, 2, 47, 5, 144, 0, 0, 71, 5, 2, 47, 5, 202, 0, 0, 71, 5, 2, 47, 5, 156, 0, 0, 71, 5, 2, 47, 5, 196, 0, 0, 71, 5, 2, 47, 5, 136, 0, 0, 5, 156, 0, 0, 71, 5, 2, 47, 5, 144, 0, 0, 5, 156, 0, 0, 71, 5, 2, 47, 5, 196, 0, 0, 5, 156, 0, 0, 138, 5, 2, 49, 137, 5, 2, 49, 138, 5, 3, 49, 137, 5, 3, 49, 138, 5, 3, 49, 137, 5, 3, 49, 5, 136, 0, 0, 138, 5, 3, 49, 137, 5, 3, 49, 5, 142, 0, 0, 138, 5, 3, 49, 137, 5, 3, 49, 5, 160, 0, 0, 138, 5, 3, 49, 137, 5, 3, 49, 5, 144, 0, 0, 138, 5, 3, 49, 137, 5, 3, 49, 5, 202, 0, 0, 138, 5, 3, 49, 137, 5, 3, 49, 5, 156, 0, 0, 138, 5, 3, 49, 137, 5, 3, 49, 5, 196, 0, 0, 138, 5, 3, 49, 137, 5, 3, 49, 5, 136, 0, 0, 5, 156, 0, 0, 138, 5, 3, 49, 137, 5, 3, 49, 5, 144, 0, 0, 5, 156, 0, 0, 138, 5, 3, 49, 137, 5, 3, 49, 5, 196, 0, 0, 5, 156, 0, 0, 138, 5, 2, 49, 137, 5, 2, 49, 5, 136, 0, 0, 138, 5, 2, 49, 137, 5, 2, 49, 5, 156, 0, 0, 138, 5, 2, 49, 137, 5, 2, 49, 5, 144, 0, 0, 138, 5, 2, 49, 137, 5, 2, 49, 5, 142, 0, 0, 138, 5, 2, 49, 137, 5, 2, 49, 5, 196, 0, 0, 138, 5, 2, 49, 137, 5, 2, 49, 5, 210, 0, 0, 72, 5, 2, 49, 5, 5, 2, 49, 72, 5, 3, 49, 5, 5, 3, 49, 72, 5, 3, 49, 5, 5, 3, 49, 5, 136, 0, 0, 72, 5, 3, 49, 5, 5, 3, 49, 5, 142, 0, 0, 72, 5, 3, 49, 5, 5, 3, 49, 5, 160, 0, 0, 72, 5, 3, 49, 5, 5, 3, 49, 5, 144, 0, 0, 72, 5, 3, 49, 5, 5, 3, 49, 5, 202, 0, 0, 72, 5, 3, 49, 5, 5, 3, 49, 5, 156, 0, 0, 72, 5, 3, 49, 5, 5, 3, 49, 5, 196, 0, 0, 72, 5, 3, 49, 5, 5, 3, 49, 5, 136, 0, 0, 5, 156, 0, 0, 72, 5, 3, 49, 5, 5, 3, 49, 5, 144, 0, 0, 5, 156, 0, 0, 72, 5, 3, 49, 5, 5, 3, 49, 5, 196, 0, 0, 5, 156, 0, 0, 72, 5, 2, 49, 5, 5, 2, 49, 5, 136, 0, 0, 72, 5, 2, 49, 5, 5, 2, 49, 5, 156, 0, 0, 72, 5, 2, 49, 5, 5, 2, 49, 5, 144, 0, 0, 72, 5, 2, 49, 5, 5, 2, 49, 5, 142, 0, 0, 72, 5, 2, 49, 5, 5, 2, 49, 5, 196, 0, 0, 72, 5, 2, 49, 5, 5, 2, 49, 5, 210, 0, 0, 71, 5, 3, 49, 5, 136, 0, 0, 71, 5, 3, 49, 5, 142, 0, 0, 71, 5, 3, 49, 5, 160, 0, 0, 71, 5, 3, 49, 5, 144, 0, 0, 71, 5, 3, 49, 5, 202, 0, 0, 71, 5, 3, 49, 5, 156, 0, 0, 71, 5, 3, 49, 5, 196, 0, 0, 71, 5, 3, 49, 5, 136, 0, 0, 5, 156, 0, 0, 71, 5, 3, 49, 5, 144, 0, 0, 5, 156, 0, 0, 71, 5, 3, 49, 5, 196, 0, 0, 5, 156, 0, 0, 71, 5, 2, 49, 5, 136, 0, 0, 71, 5, 2, 49, 5, 156, 0, 0, 71, 5, 2, 49, 5, 144, 0, 0, 71, 5, 2, 49, 5, 142, 0, 0, 71, 5, 2, 49, 5, 196, 0, 0, 71, 5, 2, 49, 5, 210, 0, 0, 138, 5, 2, 55, 137, 5, 2, 55, 138, 5, 2, 55, 137, 5, 2, 55, 5, 136, 0, 0, 138, 5, 2, 55, 137, 5, 2, 55, 5, 142, 0, 0, 138, 5, 2, 55, 137, 5, 2, 55, 5, 150, 0, 0, 138, 5, 2, 55, 137, 5, 2, 55, 5, 164, 0, 0, 138, 5, 2, 55, 137, 5, 2, 55, 5, 156, 0, 0, 138, 5, 2, 55, 137, 5, 2, 55, 5, 138, 0, 0, 138, 5, 2, 55, 137, 5, 2, 55, 5, 196, 0, 0, 138, 5, 2, 55, 137, 5, 2, 55, 5, 182, 0, 0, 138, 5, 2, 55, 137, 5, 2, 55, 5, 154, 0, 0, 72, 5, 2, 55, 5, 5, 2, 55, 72, 5, 2, 55, 5, 5, 2, 55, 5, 136, 0, 0, 72, 5, 2, 55, 5, 5, 2, 55, 5, 150, 0, 0, 72, 5, 2, 55, 5, 5, 2, 55, 5, 142, 0, 0, 72, 5, 2, 55, 5, 5, 2, 55, 5, 164, 0, 0, 72, 5, 2, 55, 5, 5, 2, 55, 5, 156, 0, 0, 72, 5, 2, 55, 5, 5, 2, 55, 5, 146, 0, 0, 72, 5, 2, 55, 5, 5, 2, 55, 5, 138, 0, 0, 72, 5, 2, 55, 5, 5, 2, 55, 5, 196, 0, 0, 72, 5, 2, 55, 5, 5, 2, 55, 5, 182, 0, 0, 72, 5, 2, 55, 5, 5, 2, 55, 5, 154, 0, 0, 71, 5, 2, 55, 5, 136, 0, 0, 71, 5, 2, 55, 5, 150, 0, 0, 71, 5, 2, 55, 5, 142, 0, 0, 71, 5, 2, 55, 5, 164, 0, 0, 71, 5, 2, 55, 5, 156, 0, 0, 71, 5, 2, 55, 5, 146, 0, 0, 71, 5, 2, 55, 5, 138, 0, 0, 71, 5, 2, 55, 5, 196, 0, 0, 71, 5, 2, 55, 5, 182, 0, 0, 71, 5, 2, 55, 5, 154, 0, 0, 138, 5, 2, 65, 137, 5, 2, 65, 138, 5, 2, 65, 137, 5, 2, 65, 5, 136, 0, 0, 138, 5, 2, 65, 137, 5, 2, 65, 5, 142, 0, 0, 138, 5, 2, 65, 137, 5, 2, 65, 5, 150, 0, 0, 138, 5, 2, 65, 137, 5, 2, 65, 5, 164, 0, 0, 138, 5, 2, 65, 137, 5, 2, 65, 5, 156, 0, 0, 138, 5, 2, 65, 137, 5, 2, 65, 5, 138, 0, 0, 138, 5, 2, 65, 137, 5, 2, 65, 5, 196, 0, 0, 138, 5, 2, 65, 137, 5, 2, 65, 5, 182, 0, 0, 138, 5, 2, 65, 137, 5, 2, 65, 5, 154, 0, 0, 72, 5, 2, 65, 5, 5, 2, 65, 72, 5, 2, 65, 5, 5, 2, 65, 5, 136, 0, 0, 72, 5, 2, 65, 5, 5, 2, 65, 5, 150, 0, 0, 72, 5, 2, 65, 5, 5, 2, 65, 5, 142, 0, 0, 72, 5, 2, 65, 5, 5, 2, 65, 5, 164, 0, 0, 72, 5, 2, 65, 5, 5, 2, 65, 5, 156, 0, 0, 72, 5, 2, 65, 5, 5, 2, 65, 5, 146, 0, 0, 72, 5, 2, 65, 5, 5, 2, 65, 5, 138, 0, 0, 72, 5, 2, 65, 5, 5, 2, 65, 5, 196, 0, 0, 72, 5, 2, 65, 5, 5, 2, 65, 5, 182, 0, 0, 72, 5, 2, 65, 5, 5, 2, 65, 5, 154, 0, 0, 71, 5, 2, 65, 5, 136, 0, 0, 71, 5, 2, 65, 5, 150, 0, 0, 71, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 164, 0, 0, 71, 5, 2, 65, 5, 156, 0, 0, 71, 5, 2, 65, 5, 146, 0, 0, 71, 5, 2, 65, 5, 138, 0, 0, 71, 5, 2, 65, 5, 196, 0, 0, 71, 5, 2, 65, 5, 182, 0, 0, 71, 5, 2, 65, 5, 154, 0, 0, 138, 5, 2, 69, 137, 5, 2, 69, 138, 5, 2, 69, 137, 5, 2, 69, 5, 136, 0, 0, 138, 5, 2, 69, 137, 5, 2, 69, 5, 142, 0, 0, 138, 5, 2, 69, 137, 5, 2, 69, 5, 150, 0, 0, 138, 5, 2, 69, 137, 5, 2, 69, 5, 164, 0, 0, 138, 5, 2, 69, 137, 5, 2, 69, 5, 156, 0, 0, 138, 5, 2, 69, 137, 5, 2, 69, 5, 138, 0, 0, 138, 5, 2, 69, 137, 5, 2, 69, 5, 196, 0, 0, 138, 5, 2, 69, 137, 5, 2, 69, 5, 182, 0, 0, 138, 5, 2, 69, 137, 5, 2, 69, 5, 154, 0, 0, 72, 5, 2, 69, 5, 5, 2, 69, 72, 5, 2, 69, 5, 5, 2, 69, 5, 136, 0, 0, 72, 5, 2, 69, 5, 5, 2, 69, 5, 150, 0, 0, 72, 5, 2, 69, 5, 5, 2, 69, 5, 142, 0, 0, 72, 5, 2, 69, 5, 5, 2, 69, 5, 164, 0, 0, 72, 5, 2, 69, 5, 5, 2, 69, 5, 156, 0, 0, 72, 5, 2, 69, 5, 5, 2, 69, 5, 146, 0, 0, 72, 5, 2, 69, 5, 5, 2, 69, 5, 138, 0, 0, 72, 5, 2, 69, 5, 5, 2, 69, 5, 196, 0, 0, 72, 5, 2, 69, 5, 5, 2, 69, 5, 182, 0, 0, 72, 5, 2, 69, 5, 5, 2, 69, 5, 154, 0, 0, 71, 5, 2, 69, 5, 136, 0, 0, 71, 5, 2, 69, 5, 150, 0, 0, 71, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 164, 0, 0, 71, 5, 2, 69, 5, 156, 0, 0, 71, 5, 2, 69, 5, 146, 0, 0, 71, 5, 2, 69, 5, 138, 0, 0, 71, 5, 2, 69, 5, 196, 0, 0, 71, 5, 2, 69, 5, 182, 0, 0, 71, 5, 2, 69, 5, 154, 0, 0, 138, 5, 2, 79, 137, 5, 2, 79, 138, 5, 2, 79, 137, 5, 2, 79, 5, 136, 0, 0, 138, 5, 2, 79, 137, 5, 2, 79, 5, 156, 0, 0, 138, 5, 2, 79, 137, 5, 2, 79, 5, 144, 0, 0, 138, 5, 2, 79, 137, 5, 2, 79, 5, 142, 0, 0, 138, 5, 2, 79, 137, 5, 2, 79, 5, 196, 0, 0, 138, 5, 2, 79, 137, 5, 2, 79, 5, 210, 0, 0, 72, 5, 2, 79, 5, 5, 2, 79, 72, 5, 2, 79, 5, 5, 2, 79, 5, 136, 0, 0, 72, 5, 2, 79, 5, 5, 2, 79, 5, 156, 0, 0, 72, 5, 2, 79, 5, 5, 2, 79, 5, 144, 0, 0, 72, 5, 2, 79, 5, 5, 2, 79, 5, 142, 0, 0, 72, 5, 2, 79, 5, 5, 2, 79, 5, 196, 0, 0, 72, 5, 2, 79, 5, 5, 2, 79, 5, 210, 0, 0, 71, 5, 2, 79, 5, 136, 0, 0, 71, 5, 2, 79, 5, 156, 0, 0, 71, 5, 2, 79, 5, 144, 0, 0, 71, 5, 2, 79, 5, 142, 0, 0, 71, 5, 2, 79, 5, 196, 0, 0, 71, 5, 2, 79, 5, 210, 0, 0, 138, 5, 2, 81, 137, 5, 2, 81, 138, 5, 2, 81, 137, 5, 2, 81, 5, 136, 0, 0, 138, 5, 2, 81, 137, 5, 2, 81, 5, 142, 0, 0, 138, 5, 2, 81, 137, 5, 2, 81, 5, 150, 0, 0, 138, 5, 2, 81, 137, 5, 2, 81, 5, 164, 0, 0, 138, 5, 2, 81, 137, 5, 2, 81, 5, 156, 0, 0, 138, 5, 2, 81, 137, 5, 2, 81, 5, 138, 0, 0, 138, 5, 2, 81, 137, 5, 2, 81, 5, 196, 0, 0, 138, 5, 2, 81, 137, 5, 2, 81, 5, 182, 0, 0, 138, 5, 2, 81, 137, 5, 2, 81, 5, 154, 0, 0, 72, 5, 2, 81, 5, 5, 2, 81, 72, 5, 2, 81, 5, 5, 2, 81, 5, 136, 0, 0, 72, 5, 2, 81, 5, 5, 2, 81, 5, 150, 0, 0, 72, 5, 2, 81, 5, 5, 2, 81, 5, 142, 0, 0, 72, 5, 2, 81, 5, 5, 2, 81, 5, 164, 0, 0, 72, 5, 2, 81, 5, 5, 2, 81, 5, 156, 0, 0, 72, 5, 2, 81, 5, 5, 2, 81, 5, 146, 0, 0, 72, 5, 2, 81, 5, 5, 2, 81, 5, 138, 0, 0, 72, 5, 2, 81, 5, 5, 2, 81, 5, 196, 0, 0, 72, 5, 2, 81, 5, 5, 2, 81, 5, 182, 0, 0, 72, 5, 2, 81, 5, 5, 2, 81, 5, 154, 0, 0, 71, 5, 2, 81, 5, 136, 0, 0, 71, 5, 2, 81, 5, 150, 0, 0, 71, 5, 2, 81, 5, 142, 0, 0, 71, 5, 2, 81, 5, 164, 0, 0, 71, 5, 2, 81, 5, 156, 0, 0, 71, 5, 2, 81, 5, 146, 0, 0, 71, 5, 2, 81, 5, 138, 0, 0, 71, 5, 2, 81, 5, 196, 0, 0, 71, 5, 2, 81, 5, 182, 0, 0, 71, 5, 2, 81, 5, 154, 0, 0, 134, 5, 2, 83, 5, 136, 0, 0, 138, 5, 2, 93, 137, 5, 2, 93, 138, 5, 2, 93, 137, 5, 2, 93, 5, 136, 0, 0, 138, 5, 2, 93, 137, 5, 2, 93, 5, 142, 0, 0, 138, 5, 2, 93, 137, 5, 2, 93, 5, 160, 0, 0, 138, 5, 2, 93, 137, 5, 2, 93, 5, 144, 0, 0, 138, 5, 2, 93, 137, 5, 2, 93, 5, 202, 0, 0, 138, 5, 2, 93, 137, 5, 2, 93, 5, 156, 0, 0, 138, 5, 2, 93, 137, 5, 2, 93, 5, 196, 0, 0, 138, 5, 2, 93, 137, 5, 2, 93, 5, 136, 0, 0, 5, 156, 0, 0, 138, 5, 2, 93, 137, 5, 2, 93, 5, 144, 0, 0, 5, 156, 0, 0, 138, 5, 2, 93, 137, 5, 2, 93, 5, 196, 0, 0, 5, 156, 0, 0, 72, 5, 2, 93, 5, 5, 2, 93, 72, 5, 2, 93, 5, 5, 2, 93, 5, 136, 0, 0, 72, 5, 2, 93, 5, 5, 2, 93, 5, 142, 0, 0, 72, 5, 2, 93, 5, 5, 2, 93, 5, 160, 0, 0, 72, 5, 2, 93, 5, 5, 2, 93, 5, 144, 0, 0, 72, 5, 2, 93, 5, 5, 2, 93, 5, 202, 0, 0, 72, 5, 2, 93, 5, 5, 2, 93, 5, 156, 0, 0, 72, 5, 2, 93, 5, 5, 2, 93, 5, 196, 0, 0, 72, 5, 2, 93, 5, 5, 2, 93, 5, 136, 0, 0, 5, 156, 0, 0, 72, 5, 2, 93, 5, 5, 2, 93, 5, 144, 0, 0, 5, 156, 0, 0, 72, 5, 2, 93, 5, 5, 2, 93, 5, 196, 0, 0, 5, 156, 0, 0, 71, 5, 2, 93, 5, 136, 0, 0, 71, 5, 2, 93, 5, 142, 0, 0, 71, 5, 2, 93, 5, 160, 0, 0, 71, 5, 2, 93, 5, 144, 0, 0, 71, 5, 2, 93, 5, 202, 0, 0, 71, 5, 2, 93, 5, 156, 0, 0, 71, 5, 2, 93, 5, 196, 0, 0, 71, 5, 2, 93, 5, 136, 0, 0, 5, 156, 0, 0, 71, 5, 2, 93, 5, 144, 0, 0, 5, 156, 0, 0, 71, 5, 2, 93, 5, 196, 0, 0, 5, 156, 0, 0, 6, 5, 2, 47, 5, 5, 2, 47, 6, 5, 2, 47, 5, 5, 2, 47, 5, 136, 0, 0, 6, 5, 2, 47, 5, 5, 2, 47, 5, 142, 0, 0, 6, 5, 2, 47, 5, 5, 2, 47, 5, 160, 0, 0, 6, 5, 2, 47, 5, 5, 2, 47, 5, 144, 0, 0, 6, 5, 2, 47, 5, 5, 2, 47, 5, 202, 0, 0, 6, 5, 2, 47, 5, 5, 2, 47, 5, 156, 0, 0, 6, 5, 2, 47, 5, 5, 2, 47, 5, 196, 0, 0, 6, 5, 2, 47, 5, 5, 2, 47, 5, 136, 0, 0, 5, 156, 0, 0, 6, 5, 2, 47, 5, 5, 2, 47, 5, 144, 0, 0, 5, 156, 0, 0, 6, 5, 2, 47, 5, 5, 2, 47, 5, 196, 0, 0, 5, 156, 0, 0, 6, 5, 2, 49, 5, 5, 2, 49, 6, 5, 3, 49, 5, 5, 3, 49, 6, 5, 3, 49, 5, 5, 3, 49, 5, 136, 0, 0, 6, 5, 3, 49, 5, 5, 3, 49, 5, 142, 0, 0, 6, 5, 3, 49, 5, 5, 3, 49, 5, 160, 0, 0, 6, 5, 3, 49, 5, 5, 3, 49, 5, 144, 0, 0, 6, 5, 3, 49, 5, 5, 3, 49, 5, 202, 0, 0, 6, 5, 3, 49, 5, 5, 3, 49, 5, 156, 0, 0, 6, 5, 3, 49, 5, 5, 3, 49, 5, 196, 0, 0, 6, 5, 3, 49, 5, 5, 3, 49, 5, 136, 0, 0, 5, 156, 0, 0, 6, 5, 3, 49, 5, 5, 3, 49, 5, 144, 0, 0, 5, 156, 0, 0, 6, 5, 3, 49, 5, 5, 3, 49, 5, 196, 0, 0, 5, 156, 0, 0, 6, 5, 2, 49, 5, 5, 2, 49, 5, 136, 0, 0, 6, 5, 2, 49, 5, 5, 2, 49, 5, 156, 0, 0, 6, 5, 2, 49, 5, 5, 2, 49, 5, 144, 0, 0, 6, 5, 2, 49, 5, 5, 2, 49, 5, 142, 0, 0, 6, 5, 2, 49, 5, 5, 2, 49, 5, 196, 0, 0, 6, 5, 2, 49, 5, 5, 2, 49, 5, 210, 0, 0, 6, 5, 2, 55, 5, 5, 2, 55, 6, 5, 2, 55, 5, 5, 2, 55, 5, 136, 0, 0, 6, 5, 2, 55, 5, 5, 2, 55, 5, 150, 0, 0, 6, 5, 2, 55, 5, 5, 2, 55, 5, 142, 0, 0, 6, 5, 2, 55, 5, 5, 2, 55, 5, 164, 0, 0, 6, 5, 2, 55, 5, 5, 2, 55, 5, 156, 0, 0, 6, 5, 2, 55, 5, 5, 2, 55, 5, 146, 0, 0, 6, 5, 2, 55, 5, 5, 2, 55, 5, 138, 0, 0, 6, 5, 2, 55, 5, 5, 2, 55, 5, 196, 0, 0, 6, 5, 2, 55, 5, 5, 2, 55, 5, 182, 0, 0, 6, 5, 2, 55, 5, 5, 2, 55, 5, 154, 0, 0, 6, 5, 2, 65, 5, 5, 2, 65, 6, 5, 2, 65, 5, 5, 2, 65, 5, 136, 0, 0, 6, 5, 2, 65, 5, 5, 2, 65, 5, 150, 0, 0, 6, 5, 2, 65, 5, 5, 2, 65, 5, 142, 0, 0, 6, 5, 2, 65, 5, 5, 2, 65, 5, 164, 0, 0, 6, 5, 2, 65, 5, 5, 2, 65, 5, 156, 0, 0, 6, 5, 2, 65, 5, 5, 2, 65, 5, 146, 0, 0, 6, 5, 2, 65, 5, 5, 2, 65, 5, 138, 0, 0, 6, 5, 2, 65, 5, 5, 2, 65, 5, 196, 0, 0, 6, 5, 2, 65, 5, 5, 2, 65, 5, 182, 0, 0, 6, 5, 2, 65, 5, 5, 2, 65, 5, 154, 0, 0, 6, 5, 2, 69, 5, 5, 2, 69, 6, 5, 2, 69, 5, 5, 2, 69, 5, 136, 0, 0, 6, 5, 2, 69, 5, 5, 2, 69, 5, 150, 0, 0, 6, 5, 2, 69, 5, 5, 2, 69, 5, 142, 0, 0, 6, 5, 2, 69, 5, 5, 2, 69, 5, 164, 0, 0, 6, 5, 2, 69, 5, 5, 2, 69, 5, 156, 0, 0, 6, 5, 2, 69, 5, 5, 2, 69, 5, 146, 0, 0, 6, 5, 2, 69, 5, 5, 2, 69, 5, 138, 0, 0, 6, 5, 2, 69, 5, 5, 2, 69, 5, 196, 0, 0, 6, 5, 2, 69, 5, 5, 2, 69, 5, 182, 0, 0, 6, 5, 2, 69, 5, 5, 2, 69, 5, 154, 0, 0, 6, 5, 2, 79, 5, 5, 2, 79, 6, 5, 2, 79, 5, 5, 2, 79, 5, 136, 0, 0, 6, 5, 2, 79, 5, 5, 2, 79, 5, 156, 0, 0, 6, 5, 2, 79, 5, 5, 2, 79, 5, 144, 0, 0, 6, 5, 2, 79, 5, 5, 2, 79, 5, 142, 0, 0, 6, 5, 2, 79, 5, 5, 2, 79, 5, 196, 0, 0, 6, 5, 2, 79, 5, 5, 2, 79, 5, 210, 0, 0, 6, 5, 2, 81, 5, 5, 2, 81, 6, 5, 2, 81, 5, 5, 2, 81, 5, 136, 0, 0, 6, 5, 2, 81, 5, 5, 2, 81, 5, 150, 0, 0, 6, 5, 2, 81, 5, 5, 2, 81, 5, 142, 0, 0, 6, 5, 2, 81, 5, 5, 2, 81, 5, 164, 0, 0, 6, 5, 2, 81, 5, 5, 2, 81, 5, 156, 0, 0, 6, 5, 2, 81, 5, 5, 2, 81, 5, 146, 0, 0, 6, 5, 2, 81, 5, 5, 2, 81, 5, 138, 0, 0, 6, 5, 2, 81, 5, 5, 2, 81, 5, 196, 0, 0, 6, 5, 2, 81, 5, 5, 2, 81, 5, 182, 0, 0, 6, 5, 2, 81, 5, 5, 2, 81, 5, 154, 0, 0, 5, 5, 2, 83, 5, 136, 0, 0, 6, 5, 2, 93, 5, 5, 2, 93, 6, 5, 2, 93, 5, 5, 2, 93, 5, 136, 0, 0, 6, 5, 2, 93, 5, 5, 2, 93, 5, 142, 0, 0, 6, 5, 2, 93, 5, 5, 2, 93, 5, 160, 0, 0, 6, 5, 2, 93, 5, 5, 2, 93, 5, 144, 0, 0, 6, 5, 2, 93, 5, 5, 2, 93, 5, 202, 0, 0, 6, 5, 2, 93, 5, 5, 2, 93, 5, 156, 0, 0, 6, 5, 2, 93, 5, 5, 2, 93, 5, 196, 0, 0, 6, 5, 2, 93, 5, 5, 2, 93, 5, 136, 0, 0, 5, 156, 0, 0, 6, 5, 2, 93, 5, 5, 2, 93, 5, 144, 0, 0, 5, 156, 0, 0, 6, 5, 2, 93, 5, 5, 2, 93, 5, 196, 0, 0, 5, 156, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 134, 70, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 5, 70, 2, 83, 5, 162, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 134, 70, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 70, 2, 71, 5, 190, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 134, 70, 2, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 70, 2, 83, 5, 190, 0, 0, 134, 5, 2, 83, 5, 164, 0, 0, 5, 5, 2, 83, 5, 164, 0, 0, 134, 5, 2, 83, 5, 144, 0, 0, 5, 5, 2, 83, 5, 144, 0, 0, 134, 5, 2, 83, 5, 138, 0, 0, 5, 5, 2, 83, 5, 138, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 134, 70, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 5, 70, 2, 71, 5, 162, 0, 0, 134, 5, 2, 71, 5, 164, 0, 0, 5, 5, 2, 71, 5, 164, 0, 0, 134, 5, 2, 83, 5, 198, 0, 0, 134, 70, 2, 83, 5, 198, 0, 0, 5, 5, 2, 83, 5, 198, 0, 0, 5, 70, 2, 83, 5, 198, 0, 0, 134, 5, 2, 83, 5, 208, 0, 0, 134, 70, 2, 83, 5, 208, 0, 0, 5, 5, 2, 83, 5, 208, 0, 0, 5, 70, 2, 83, 5, 208, 0, 0, 134, 5, 2, 83, 5, 204, 0, 0, 134, 70, 2, 83, 5, 204, 0, 0, 5, 5, 2, 83, 5, 204, 0, 0, 5, 70, 2, 83, 5, 204, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 134, 70, 2, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 5, 70, 2, 71, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 196, 0, 0, 134, 70, 2, 83, 5, 196, 0, 0, 5, 5, 2, 83, 5, 196, 0, 0, 5, 70, 2, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 23, 0, 24, 2, 42, 0, 100, 30, 20, 0, 103, 30, 9, 0, 103, 30, 11, 192, 197, 3, 104, 30, 4, 192, 197, 35, 105, 30, 11, 192, 197, 99, 100, 30, 3, 192, 197, 35, 101, 30, 10, 192, 197, 163, 102, 30, 3, 192, 197, 163, 97, 30, 9, 0, 97, 30, 10, 192, 197, 34, 98, 30, 2, 192, 197, 194, 99, 30, 10, 192, 197, 98, 24, 2, 2, 192, 197, 2, 25, 2, 9, 192, 197, 226, 96, 30, 2, 192, 197, 98, 92, 1, 20, 0, 95, 1, 9, 0, 95, 1, 9, 192, 197, 98, 96, 1, 1, 192, 197, 162, 97, 1, 9, 192, 197, 162, 92, 1, 0, 192, 197, 226, 93, 1, 9, 192, 197, 34, 94, 1, 1, 192, 197, 66, 115, 0, 9, 0, 115, 0, 2, 239, 71, 5, 90, 1, 0, 192, 197, 130, 91, 1, 8, 192, 197, 226, 67, 0, 42, 0, 83, 0, 2, 239, 137, 5, 99, 0, 10, 0, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 7, 192, 197, 100, 103, 30, 7, 192, 197, 228, 105, 30, 8, 192, 197, 100, 25, 2, 6, 192, 197, 67, 97, 30, 6, 192, 197, 163, 99, 30, 7, 192, 197, 3, 115, 0, 4, 192, 197, 130, 91, 1, 4, 192, 197, 195, 93, 1, 5, 192, 197, 35, 95, 1, 5, 192, 197, 131, 97, 1, 5, 192, 197, 227, 10, 0, 24, 2, 20, 0, 100, 30, 9, 0, 100, 30, 3, 192, 197, 4, 102, 30, 3, 192, 197, 132, 104, 30, 4, 192, 197, 4, 24, 2, 1, 192, 197, 227, 96, 30, 2, 192, 197, 67, 98, 30, 2, 192, 197, 163, 83, 0, 197, 162, 90, 1, 0, 192, 197, 99, 92, 1, 0, 192, 197, 195, 94, 1, 1, 192, 197, 35, 96, 1, 1, 192, 197, 131, 0, 48, 156, 5, 15, 0, 125, 1, 26, 0, 146, 30, 12, 0, 146, 30, 18, 192, 197, 2, 147, 30, 29, 192, 197, 98, 148, 30, 18, 192, 197, 98, 149, 30, 29, 192, 197, 162, 125, 1, 17, 192, 197, 66, 126, 1, 28, 192, 197, 226, 144, 30, 17, 192, 197, 162, 145, 30, 29, 192, 197, 34, 121, 1, 12, 0, 121, 1, 16, 192, 197, 130, 122, 1, 28, 192, 197, 98, 123, 1, 16, 192, 197, 226, 124, 1, 28, 192, 197, 162, 68, 0, 45, 0, 90, 0, 105, 0, 100, 0, 143, 0, 122, 0, 202, 127, 2, 49, 71, 5, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 27, 192, 197, 67, 103, 30, 27, 192, 197, 163, 105, 30, 28, 192, 197, 3, 25, 2, 26, 192, 197, 130, 97, 30, 26, 192, 197, 194, 99, 30, 27, 192, 197, 2, 115, 0, 3, 241, 71, 5, 91, 1, 25, 192, 197, 130, 93, 1, 25, 192, 197, 194, 95, 1, 26, 192, 197, 2, 97, 1, 26, 192, 197, 66, 6, 0, 125, 1, 12, 0, 125, 1, 17, 192, 197, 35, 144, 30, 17, 192, 197, 131, 146, 30, 17, 192, 197, 227, 148, 30, 18, 192, 197, 67, 90, 0, 6, 0, 121, 1, 16, 192, 197, 99, 123, 1, 16, 192, 197, 195, 10, 67, 197, 194, 24, 2, 20, 0, 100, 30, 9, 0, 100, 30, 14, 192, 197, 228, 102, 30, 15, 192, 197, 100, 104, 30, 15, 192, 197, 228, 24, 2, 13, 192, 197, 195, 96, 30, 14, 192, 197, 35, 98, 30, 14, 192, 197, 131, 83, 0, 12, 192, 197, 2, 90, 1, 12, 192, 197, 67, 92, 1, 12, 192, 197, 163, 94, 1, 13, 192, 197, 3, 96, 1, 13, 192, 197, 99, 202, 127, 2, 49, 137, 5, 24, 2, 20, 0, 100, 30, 9, 0, 100, 30, 15, 192, 197, 3, 102, 30, 15, 192, 197, 131, 104, 30, 16, 192, 197, 3, 24, 2, 13, 192, 197, 226, 96, 30, 14, 192, 197, 66, 98, 30, 14, 192, 197, 162, 83, 0, 3, 241, 137, 5, 90, 1, 12, 192, 197, 98, 92, 1, 12, 192, 197, 194, 94, 1, 13, 192, 197, 34, 96, 1, 13, 192, 197, 130, 6, 0, 126, 1, 12, 0, 126, 1, 24, 192, 197, 3, 145, 30, 24, 192, 197, 99, 147, 30, 24, 192, 197, 195, 149, 30, 25, 192, 197, 35, 122, 0, 6, 0, 122, 1, 23, 192, 197, 67, 124, 1, 23, 192, 197, 163, 202, 68, 197, 162, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 21, 192, 197, 196, 103, 30, 22, 192, 197, 68, 105, 30, 22, 192, 197, 196, 25, 2, 20, 192, 197, 163, 97, 30, 21, 192, 197, 3, 99, 30, 21, 192, 197, 99, 115, 0, 18, 192, 197, 226, 91, 1, 19, 192, 197, 35, 93, 1, 19, 192, 197, 131, 95, 1, 19, 192, 197, 227, 97, 1, 20, 192, 197, 67, 0, 54, 156, 5, 22, 0, 51, 2, 42, 0, 244, 30, 20, 0, 247, 30, 9, 0, 247, 30, 39, 192, 197, 130, 248, 30, 33, 192, 197, 66, 249, 30, 39, 192, 197, 194, 244, 30, 32, 192, 197, 130, 245, 30, 39, 192, 197, 66, 246, 30, 32, 192, 197, 226, 153, 30, 9, 0, 153, 30, 38, 192, 197, 194, 242, 30, 32, 192, 197, 34, 243, 30, 39, 192, 197, 2, 51, 2, 38, 192, 197, 66, 142, 30, 31, 192, 197, 194, 143, 30, 38, 192, 197, 130, 253, 0, 20, 0, 119, 1, 9, 0, 119, 1, 38, 192, 197, 2, 120, 1, 31, 192, 197, 2, 50, 2, 31, 192, 197, 98, 253, 0, 37, 192, 197, 130, 255, 0, 37, 192, 197, 194, 118, 1, 30, 192, 197, 162, 71, 0, 11, 0, 89, 0, 2, 247, 137, 5, 103, 0, 39, 0, 121, 0, 2, 247, 71, 5, 221, 0, 30, 192, 197, 66, 9, 0, 142, 30, 15, 0, 142, 30, 31, 192, 197, 163, 242, 30, 32, 192, 197, 3, 244, 30, 32, 192, 197, 99, 246, 30, 32, 192, 197, 195, 248, 30, 33, 192, 197, 35, 89, 0, 29, 192, 197, 226, 221, 0, 30, 192, 197, 35, 118, 1, 30, 192, 197, 131, 120, 1, 30, 192, 197, 227, 50, 2, 31, 192, 197, 67, 10, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 36, 192, 197, 99, 247, 30, 36, 192, 197, 195, 249, 30, 37, 192, 197, 35, 143, 30, 35, 192, 197, 67, 153, 30, 35, 192, 197, 163, 243, 30, 36, 192, 197, 3, 121, 0, 33, 192, 197, 130, 253, 0, 33, 192, 197, 195, 255, 0, 34, 192, 197, 35, 119, 1, 34, 192, 197, 131, 51, 2, 34, 192, 197, 227, 0, 64, 156, 5, 22, 0, 51, 2, 42, 0, 244, 30, 20, 0, 247, 30, 9, 0, 247, 30, 49, 192, 197, 162, 248, 30, 43, 192, 197, 98, 249, 30, 49, 192, 197, 226, 244, 30, 42, 192, 197, 162, 245, 30, 49, 192, 197, 98, 246, 30, 43, 192, 197, 2, 153, 30, 9, 0, 153, 30, 48, 192, 197, 226, 242, 30, 42, 192, 197, 66, 243, 30, 49, 192, 197, 34, 51, 2, 48, 192, 197, 98, 142, 30, 41, 192, 197, 226, 143, 30, 48, 192, 197, 162, 253, 0, 20, 0, 119, 1, 9, 0, 119, 1, 48, 192, 197, 34, 120, 1, 41, 192, 197, 34, 50, 2, 41, 192, 197, 130, 253, 0, 47, 192, 197, 162, 255, 0, 47, 192, 197, 226, 118, 1, 40, 192, 197, 194, 76, 0, 13, 0, 89, 0, 255, 255, 2, 65, 137, 5, 108, 0, 40, 0, 121, 0, 255, 255, 2, 65, 71, 5, 221, 0, 40, 192, 197, 98, 9, 0, 142, 30, 15, 0, 142, 30, 41, 192, 197, 195, 242, 30, 42, 192, 197, 35, 244, 30, 42, 192, 197, 131, 246, 30, 42, 192, 197, 227, 248, 30, 43, 192, 197, 67, 89, 0, 40, 192, 197, 2, 221, 0, 40, 192, 197, 67, 118, 1, 40, 192, 197, 163, 120, 1, 41, 192, 197, 3, 50, 2, 41, 192, 197, 99, 10, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 46, 192, 197, 131, 247, 30, 46, 192, 197, 227, 249, 30, 47, 192, 197, 67, 143, 30, 45, 192, 197, 99, 153, 30, 45, 192, 197, 195, 243, 30, 46, 192, 197, 35, 121, 0, 43, 192, 197, 162, 253, 0, 43, 192, 197, 227, 255, 0, 44, 192, 197, 67, 119, 1, 44, 192, 197, 163, 51, 2, 45, 192, 197, 3, 0, 68, 156, 5, 22, 0, 51, 2, 42, 0, 244, 30, 20, 0, 247, 30, 9, 0, 247, 30, 59, 192, 197, 194, 248, 30, 53, 192, 197, 130, 249, 30, 60, 192, 197, 2, 244, 30, 52, 192, 197, 194, 245, 30, 59, 192, 197, 130, 246, 30, 53, 192, 197, 34, 153, 30, 9, 0, 153, 30, 59, 192, 197, 2, 242, 30, 52, 192, 197, 98, 243, 30, 59, 192, 197, 66, 51, 2, 58, 192, 197, 130, 142, 30, 52, 192, 197, 2, 143, 30, 58, 192, 197, 194, 253, 0, 20, 0, 119, 1, 9, 0, 119, 1, 58, 192, 197, 66, 120, 1, 51, 192, 197, 66, 50, 2, 51, 192, 197, 162, 253, 0, 57, 192, 197, 194, 255, 0, 58, 192, 197, 2, 118, 1, 50, 192, 197, 226, 78, 0, 13, 0, 89, 0, 255, 255, 2, 69, 137, 5, 110, 0, 40, 0, 121, 0, 255, 255, 2, 69, 71, 5, 221, 0, 50, 192, 197, 130, 9, 0, 142, 30, 15, 0, 142, 30, 51, 192, 197, 227, 242, 30, 52, 192, 197, 67, 244, 30, 52, 192, 197, 163, 246, 30, 53, 192, 197, 3, 248, 30, 53, 192, 197, 99, 89, 0, 50, 192, 197, 34, 221, 0, 50, 192, 197, 99, 118, 1, 50, 192, 197, 195, 120, 1, 51, 192, 197, 35, 50, 2, 51, 192, 197, 131, 10, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 56, 192, 197, 163, 247, 30, 57, 192, 197, 3, 249, 30, 57, 192, 197, 99, 143, 30, 55, 192, 197, 131, 153, 30, 55, 192, 197, 227, 243, 30, 56, 192, 197, 67, 121, 0, 53, 192, 197, 194, 253, 0, 54, 192, 197, 3, 255, 0, 54, 192, 197, 99, 119, 1, 54, 192, 197, 195, 51, 2, 55, 192, 197, 35, 0, 70, 156, 5, 1, 0, 8, 3, 255, 255, 2, 71, 134, 5, 11, 3, 255, 255, 2, 71, 134, 70, 0, 78, 156, 5, 15, 0, 125, 1, 26, 0, 146, 30, 12, 0, 146, 30, 62, 192, 197, 34, 147, 30, 66, 192, 197, 66, 148, 30, 62, 192, 197, 130, 149, 30, 66, 192, 197, 130, 125, 1, 61, 192, 197, 98, 126, 1, 65, 192, 197, 194, 144, 30, 61, 192, 197, 194, 145, 30, 66, 192, 197, 2, 121, 1, 12, 0, 121, 1, 60, 192, 197, 162, 122, 1, 65, 192, 197, 66, 123, 1, 61, 192, 197, 2, 124, 1, 65, 192, 197, 130, 83, 0, 10, 0, 90, 0, 255, 255, 2, 79, 137, 5, 115, 0, 28, 0, 122, 0, 255, 255, 2, 79, 71, 5, 6, 0, 125, 1, 12, 0, 125, 1, 61, 192, 197, 67, 144, 30, 61, 192, 197, 163, 146, 30, 62, 192, 197, 3, 148, 30, 62, 192, 197, 99, 90, 0, 60, 192, 197, 66, 121, 1, 60, 192, 197, 131, 123, 1, 60, 192, 197, 227, 6, 0, 126, 1, 12, 0, 126, 1, 63, 192, 197, 195, 145, 30, 64, 192, 197, 35, 147, 30, 64, 192, 197, 131, 149, 30, 64, 192, 197, 227, 122, 0, 62, 192, 197, 194, 122, 1, 63, 192, 197, 3, 124, 1, 63, 192, 197, 99, 0, 80, 156, 5, 22, 0, 51, 2, 42, 0, 244, 30, 20, 0, 247, 30, 9, 0, 247, 30, 76, 192, 197, 98, 248, 30, 70, 192, 197, 34, 249, 30, 76, 192, 197, 162, 244, 30, 69, 192, 197, 98, 245, 30, 76, 192, 197, 34, 246, 30, 69, 192, 197, 194, 153, 30, 9, 0, 153, 30, 75, 192, 197, 162, 242, 30, 69, 192, 197, 2, 243, 30, 75, 192, 197, 226, 51, 2, 75, 192, 197, 34, 142, 30, 68, 192, 197, 162, 143, 30, 75, 192, 197, 98, 253, 0, 20, 0, 119, 1, 9, 0, 119, 1, 74, 192, 197, 226, 120, 1, 67, 192, 197, 226, 50, 2, 68, 192, 197, 66, 253, 0, 74, 192, 197, 98, 255, 0, 74, 192, 197, 162, 118, 1, 67, 192, 197, 130, 84, 0, 13, 0, 89, 0, 255, 255, 2, 81, 137, 5, 116, 0, 40, 0, 121, 0, 255, 255, 2, 81, 71, 5, 221, 0, 67, 192, 197, 34, 9, 0, 142, 30, 15, 0, 142, 30, 68, 192, 197, 131, 242, 30, 68, 192, 197, 227, 244, 30, 69, 192, 197, 67, 246, 30, 69, 192, 197, 163, 248, 30, 70, 192, 197, 3, 89, 0, 66, 192, 197, 194, 221, 0, 67, 192, 197, 3, 118, 1, 67, 192, 197, 99, 120, 1, 67, 192, 197, 195, 50, 2, 68, 192, 197, 35, 10, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 73, 192, 197, 67, 247, 30, 73, 192, 197, 163, 249, 30, 74, 192, 197, 3, 143, 30, 72, 192, 197, 35, 153, 30, 72, 192, 197, 131, 243, 30, 72, 192, 197, 227, 121, 0, 70, 192, 197, 98, 253, 0, 70, 192, 197, 163, 255, 0, 71, 192, 197, 3, 119, 1, 71, 192, 197, 99, 51, 2, 71, 192, 197, 195, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 2, 83, 134, 5, 11, 3, 255, 255, 2, 83, 134, 70, 68, 3, 76, 192, 197, 226, 0, 92, 156, 5, 23, 0, 24, 2, 42, 0, 100, 30, 20, 0, 103, 30, 9, 0, 103, 30, 88, 192, 197, 3, 104, 30, 81, 192, 197, 35, 105, 30, 88, 192, 197, 99, 100, 30, 80, 192, 197, 35, 101, 30, 87, 192, 197, 163, 102, 30, 80, 192, 197, 163, 97, 30, 9, 0, 97, 30, 87, 192, 197, 34, 98, 30, 79, 192, 197, 194, 99, 30, 87, 192, 197, 98, 24, 2, 79, 192, 197, 2, 25, 2, 86, 192, 197, 226, 96, 30, 79, 192, 197, 98, 92, 1, 20, 0, 95, 1, 9, 0, 95, 1, 86, 192, 197, 98, 96, 1, 78, 192, 197, 162, 97, 1, 86, 192, 197, 162, 92, 1, 77, 192, 197, 226, 93, 1, 86, 192, 197, 34, 94, 1, 78, 192, 197, 66, 122, 0, 46, 0, 122, 0, 6, 0, 90, 1, 77, 192, 197, 130, 91, 1, 85, 192, 197, 226, 10, 0, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 84, 192, 197, 100, 103, 30, 84, 192, 197, 228, 105, 30, 85, 192, 197, 100, 25, 2, 83, 192, 197, 67, 97, 30, 83, 192, 197, 163, 99, 30, 84, 192, 197, 3, 115, 0, 81, 192, 197, 130, 91, 1, 81, 192, 197, 195, 93, 1, 82, 192, 197, 35, 95, 1, 82, 192, 197, 131, 97, 1, 82, 192, 197, 227, 83, 0, 255, 255, 2, 93, 137, 5, 90, 0, 4, 0, 115, 0, 255, 255, 2, 93, 71, 5, 10, 0, 24, 2, 20, 0, 100, 30, 9, 0, 100, 30, 80, 192, 197, 4, 102, 30, 80, 192, 197, 132, 104, 30, 81, 192, 197, 4, 24, 2, 78, 192, 197, 227, 96, 30, 79, 192, 197, 67, 98, 30, 79, 192, 197, 163, 83, 0, 77, 192, 197, 34, 90, 1, 77, 192, 197, 99, 92, 1, 77, 192, 197, 195, 94, 1, 78, 192, 197, 35, 96, 1, 78, 192, 197, 131, 0, 46, 5, 5, 11, 0, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 7, 192, 197, 131, 103, 30, 8, 192, 197, 3, 105, 30, 8, 192, 197, 131, 25, 2, 6, 192, 197, 98, 97, 30, 6, 192, 197, 194, 99, 30, 7, 192, 197, 34, 93, 1, 9, 0, 93, 1, 5, 192, 197, 66, 95, 1, 5, 192, 197, 162, 97, 1, 6, 192, 197, 2, 99, 0, 6, 0, 115, 0, 2, 239, 5, 5, 91, 1, 4, 192, 197, 226, 10, 0, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 91, 192, 197, 164, 103, 30, 92, 192, 197, 36, 105, 30, 92, 192, 197, 164, 25, 2, 90, 192, 197, 131, 97, 30, 90, 192, 197, 227, 99, 30, 91, 192, 197, 67, 115, 0, 88, 192, 197, 194, 91, 1, 89, 192, 197, 3, 93, 1, 89, 192, 197, 99, 95, 1, 89, 192, 197, 195, 97, 1, 90, 192, 197, 35, 0, 48, 5, 5, 7, 0, 126, 1, 12, 0, 126, 1, 24, 192, 197, 34, 145, 30, 24, 192, 197, 130, 147, 30, 24, 192, 197, 226, 149, 30, 25, 192, 197, 66, 100, 0, 8, 0, 122, 0, 68, 0, 122, 1, 23, 192, 197, 98, 124, 1, 23, 192, 197, 194, 6, 0, 126, 1, 12, 0, 126, 1, 98, 192, 197, 131, 145, 30, 98, 192, 197, 227, 147, 30, 99, 192, 197, 67, 149, 30, 99, 192, 197, 163, 122, 0, 6, 0, 122, 1, 97, 192, 197, 195, 124, 1, 98, 192, 197, 35, 138, 87, 197, 34, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 96, 192, 197, 68, 103, 30, 96, 192, 197, 196, 105, 30, 97, 192, 197, 68, 25, 2, 95, 192, 197, 35, 97, 30, 95, 192, 197, 131, 99, 30, 95, 192, 197, 227, 115, 0, 93, 192, 197, 98, 91, 1, 93, 192, 197, 163, 93, 1, 94, 192, 197, 3, 95, 1, 94, 192, 197, 99, 97, 1, 94, 192, 197, 195, 202, 127, 2, 49, 5, 5, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 21, 192, 197, 227, 103, 30, 22, 192, 197, 99, 105, 30, 22, 192, 197, 227, 25, 2, 20, 192, 197, 194, 97, 30, 21, 192, 197, 34, 99, 30, 21, 192, 197, 130, 115, 0, 3, 241, 5, 5, 91, 1, 19, 192, 197, 66, 93, 1, 19, 192, 197, 162, 95, 1, 20, 192, 197, 2, 97, 1, 20, 192, 197, 98, 0, 54, 5, 5, 11, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 36, 192, 197, 130, 247, 30, 36, 192, 197, 226, 249, 30, 37, 192, 197, 66, 143, 30, 35, 192, 197, 98, 153, 30, 35, 192, 197, 194, 243, 30, 36, 192, 197, 34, 255, 0, 9, 0, 255, 0, 34, 192, 197, 66, 119, 1, 34, 192, 197, 162, 51, 2, 35, 192, 197, 2, 103, 0, 6, 0, 121, 0, 2, 247, 5, 5, 253, 0, 33, 192, 197, 226, 10, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 102, 192, 197, 227, 247, 30, 103, 192, 197, 67, 249, 30, 103, 192, 197, 163, 143, 30, 101, 192, 197, 195, 153, 30, 102, 192, 197, 35, 243, 30, 102, 192, 197, 131, 121, 0, 100, 192, 197, 2, 253, 0, 100, 192, 197, 67, 255, 0, 100, 192, 197, 163, 119, 1, 101, 192, 197, 3, 51, 2, 101, 192, 197, 99, 0, 64, 5, 5, 11, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 46, 192, 197, 162, 247, 30, 47, 192, 197, 2, 249, 30, 47, 192, 197, 98, 143, 30, 45, 192, 197, 130, 153, 30, 45, 192, 197, 226, 243, 30, 46, 192, 197, 66, 255, 0, 9, 0, 255, 0, 44, 192, 197, 98, 119, 1, 44, 192, 197, 194, 51, 2, 45, 192, 197, 34, 108, 0, 7, 0, 121, 0, 255, 255, 2, 65, 5, 5, 253, 0, 44, 192, 197, 2, 10, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 106, 192, 197, 227, 247, 30, 107, 192, 197, 67, 249, 30, 107, 192, 197, 163, 143, 30, 105, 192, 197, 195, 153, 30, 106, 192, 197, 35, 243, 30, 106, 192, 197, 131, 121, 0, 104, 192, 197, 2, 253, 0, 104, 192, 197, 67, 255, 0, 104, 192, 197, 163, 119, 1, 105, 192, 197, 3, 51, 2, 105, 192, 197, 99, 0, 68, 5, 5, 11, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 56, 192, 197, 194, 247, 30, 57, 192, 197, 34, 249, 30, 57, 192, 197, 130, 143, 30, 55, 192, 197, 162, 153, 30, 56, 192, 197, 2, 243, 30, 56, 192, 197, 98, 255, 0, 9, 0, 255, 0, 54, 192, 197, 130, 119, 1, 54, 192, 197, 226, 51, 2, 55, 192, 197, 66, 110, 0, 7, 0, 121, 0, 255, 255, 2, 69, 5, 5, 253, 0, 54, 192, 197, 34, 10, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 110, 192, 197, 227, 247, 30, 111, 192, 197, 67, 249, 30, 111, 192, 197, 163, 143, 30, 109, 192, 197, 195, 153, 30, 110, 192, 197, 35, 243, 30, 110, 192, 197, 131, 121, 0, 108, 192, 197, 2, 253, 0, 108, 192, 197, 67, 255, 0, 108, 192, 197, 163, 119, 1, 109, 192, 197, 3, 51, 2, 109, 192, 197, 99, 0, 70, 5, 5, 1, 0, 8, 3, 255, 255, 2, 71, 5, 5, 11, 3, 255, 255, 2, 71, 5, 70, 0, 78, 5, 5, 7, 0, 126, 1, 12, 0, 126, 1, 63, 192, 197, 226, 145, 30, 64, 192, 197, 66, 147, 30, 64, 192, 197, 162, 149, 30, 65, 192, 197, 2, 115, 0, 10, 0, 122, 0, 255, 255, 2, 79, 5, 5, 122, 1, 63, 192, 197, 34, 124, 1, 63, 192, 197, 130, 6, 0, 126, 1, 12, 0, 126, 1, 113, 192, 197, 3, 145, 30, 113, 192, 197, 99, 147, 30, 113, 192, 197, 195, 149, 30, 114, 192, 197, 35, 122, 0, 112, 192, 197, 2, 122, 1, 112, 192, 197, 67, 124, 1, 112, 192, 197, 163, 0, 80, 5, 5, 11, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 73, 192, 197, 98, 247, 30, 73, 192, 197, 194, 249, 30, 74, 192, 197, 34, 143, 30, 72, 192, 197, 66, 153, 30, 72, 192, 197, 162, 243, 30, 73, 192, 197, 2, 255, 0, 9, 0, 255, 0, 71, 192, 197, 34, 119, 1, 71, 192, 197, 130, 51, 2, 71, 192, 197, 226, 116, 0, 7, 0, 121, 0, 255, 255, 2, 81, 5, 5, 253, 0, 70, 192, 197, 194, 10, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 117, 192, 197, 99, 247, 30, 117, 192, 197, 195, 249, 30, 118, 192, 197, 35, 143, 30, 116, 192, 197, 67, 153, 30, 116, 192, 197, 163, 243, 30, 117, 192, 197, 3, 121, 0, 114, 192, 197, 130, 253, 0, 114, 192, 197, 195, 255, 0, 115, 192, 197, 35, 119, 1, 115, 192, 197, 131, 51, 2, 115, 192, 197, 227, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 2, 83, 5, 5, 11, 3, 255, 255, 2, 83, 5, 70, 68, 3, 118, 192, 197, 130, 0, 92, 5, 5, 11, 0, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 84, 192, 197, 131, 103, 30, 85, 192, 197, 3, 105, 30, 85, 192, 197, 131, 25, 2, 83, 192, 197, 98, 97, 30, 83, 192, 197, 194, 99, 30, 84, 192, 197, 34, 93, 1, 9, 0, 93, 1, 82, 192, 197, 66, 95, 1, 82, 192, 197, 162, 97, 1, 83, 192, 197, 2, 115, 0, 255, 255, 2, 93, 5, 5, 122, 0, 3, 0, 91, 1, 81, 192, 197, 226, 10, 0, 25, 2, 20, 0, 101, 30, 9, 0, 101, 30, 121, 192, 197, 164, 103, 30, 122, 192, 197, 36, 105, 30, 122, 192, 197, 164, 25, 2, 120, 192, 197, 131, 97, 30, 120, 192, 197, 227, 99, 30, 121, 192, 197, 67, 115, 0, 118, 192, 197, 194, 91, 1, 119, 192, 197, 3, 93, 1, 119, 192, 197, 99, 95, 1, 119, 192, 197, 195, 97, 1, 120, 192, 197, 35, 156, 82, 196, 162, 1, 0, 8, 3, 123, 192, 197, 34, 11, 3, 123, 192, 197, 98, 5, 82, 196, 162, 1, 0, 8, 3, 123, 192, 197, 162, 11, 3, 123, 192, 197, 226, 156, 70, 196, 190, 1, 0, 8, 3, 124, 192, 197, 34, 11, 3, 124, 192, 197, 98, 5, 70, 196, 190, 1, 0, 8, 3, 124, 192, 197, 162, 11, 3, 124, 192, 197, 226, 156, 82, 196, 190, 1, 0, 8, 3, 125, 192, 197, 34, 11, 3, 125, 192, 197, 98, 5, 82, 196, 190, 1, 0, 8, 3, 125, 192, 197, 162, 11, 3, 125, 192, 197, 226, 156, 70, 196, 162, 1, 0, 8, 3, 127, 192, 197, 162, 11, 3, 127, 192, 197, 226, 5, 70, 196, 162, 1, 0, 8, 3, 128, 192, 197, 34, 11, 3, 128, 192, 197, 98, 156, 82, 196, 198, 1, 0, 8, 3, 129, 192, 197, 34, 11, 3, 129, 192, 197, 98, 5, 82, 196, 198, 1, 0, 8, 3, 129, 192, 197, 162, 11, 3, 129, 192, 197, 226, 156, 82, 196, 208, 1, 0, 8, 3, 130, 192, 197, 34, 11, 3, 130, 192, 197, 98, 5, 82, 196, 208, 1, 0, 8, 3, 130, 192, 197, 162, 11, 3, 130, 192, 197, 226, 156, 82, 196, 204, 1, 0, 8, 3, 131, 192, 197, 34, 11, 3, 131, 192, 197, 98, 5, 82, 196, 204, 1, 0, 8, 3, 131, 192, 197, 162, 11, 3, 131, 192, 197, 226, 156, 70, 196, 196, 1, 0, 8, 3, 132, 192, 197, 34, 11, 3, 132, 192, 197, 98, 5, 70, 196, 196, 1, 0, 8, 3, 132, 192, 197, 162, 11, 3, 132, 192, 197, 226, 133, 0, 197, 35, 1, 0, 8, 3, 133, 192, 197, 131, 11, 3, 133, 192, 197, 227, 134, 0, 197, 67, 1, 0, 8, 3, 134, 192, 197, 163, 11, 3, 135, 192, 197, 3, 156, 82, 196, 196, 1, 0, 8, 3, 135, 192, 197, 98, 11, 3, 135, 192, 197, 162, 5, 82, 196, 196, 1, 0, 8, 3, 135, 192, 197, 226, 11, 3, 136, 192, 197, 34, 136, 0, 197, 99, 1, 0, 8, 3, 136, 192, 197, 195, 11, 3, 137, 192, 197, 35, 137, 0, 197, 131, 1, 0, 8, 3, 137, 192, 197, 227, 11, 3, 138, 192, 197, 67, 44, 0, 67, 0, 69, 0, 71, 0, 72, 0, 78, 0, 79, 0, 83, 0, 85, 0, 89, 0, 91, 0, 99, 0, 101, 0, 103, 0, 104, 0, 108, 0, 109, 0, 110, 0, 111, 0, 115, 0, 117, 0, 121, 0, 123, 0, 221, 0, 222, 0, 253, 0, 254, 0, 255, 0, 0, 1, 90, 1, 98, 1, 118, 1, 127, 1, 24, 2, 26, 2, 50, 2, 52, 2, 96, 30, 106, 30, 142, 30, 150, 30, 153, 30, 154, 30, 242, 30, 250, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 50, 4, 186, 80, 185, 84, 68, 4, 185, 96, 186, 100, 186, 108, 185, 112, 88, 4, 185, 124, 108, 4, 128, 4, 185, 148, 185, 152, 185, 160, 130, 4, 150, 4, 170, 4, 185, 192, 185, 196, 185, 200, 185, 204, 172, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 196, 4, 209, 4, 168, 80, 168, 84, 219, 4, 168, 96, 168, 100, 168, 108, 168, 112, 230, 4, 168, 124, 241, 4, 252, 4, 168, 148, 168, 152, 168, 160, 254, 4, 9, 5, 20, 5, 168, 192, 168, 196, 168, 200, 168, 204, 22, 5, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 140, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 121, 58, 89, 58, 57, 58, 0, 8, 217, 66, 186, 81, 154, 81, 250, 81, 90, 82, 186, 101, 154, 101, 250, 101, 90, 102, 217, 72, 122, 130, 185, 141, 153, 141, 249, 141, 121, 142, 185, 144, 136, 14, 185, 142, 185, 185, 153, 185, 249, 185, 185, 188, 153, 205, 185, 216, 2, 8, 168, 57, 136, 57, 232, 57, 104, 58, 72, 58, 40, 58, 4, 8, 200, 66, 168, 81, 136, 81, 232, 81, 72, 82, 168, 101, 136, 101, 232, 101, 72, 102, 200, 72, 104, 130, 168, 141, 136, 141, 232, 141, 104, 142, 168, 144, 128, 14, 168, 142, 168, 185, 136, 185, 232, 185, 168, 188, 136, 205, 168, 216, 72, 206, 25, 59, 8, 59, 217, 57, 200, 57, 249, 58, 232, 58, 153, 65, 136, 65, 249, 65, 232, 65, 153, 66, 136, 66, 25, 66, 8, 66, 25, 74, 8, 74, 57, 75, 40, 75, 26, 83, 8, 83, 218, 81, 200, 81, 154, 82, 136, 82, 250, 82, 232, 82, 26, 82, 8, 82, 249, 89, 232, 89, 217, 89, 200, 89, 153, 90, 136, 90, 217, 90, 200, 90, 249, 97, 232, 97, 57, 99, 40, 99, 122, 102, 104, 102, 26, 103, 8, 103, 218, 101, 200, 101, 250, 102, 232, 102, 154, 102, 168, 104, 6, 8, 8, 8, 250, 109, 232, 109, 217, 114, 200, 114, 168, 156, 153, 117, 136, 117, 217, 118, 200, 118, 25, 118, 8, 118, 89, 119, 72, 119, 57, 119, 40, 119, 154, 129, 136, 129, 218, 130, 200, 130, 26, 130, 8, 130, 10, 8, 185, 136, 168, 136, 25, 143, 8, 143, 217, 141, 200, 141, 217, 144, 200, 144, 12, 8, 14, 8, 153, 161, 136, 161, 217, 162, 200, 162, 25, 162, 8, 162, 154, 165, 136, 165, 250, 165, 232, 165, 218, 166, 200, 166, 26, 166, 8, 166, 217, 174, 200, 174, 25, 174, 8, 174, 185, 180, 168, 180, 121, 186, 104, 186, 25, 187, 8, 187, 217, 185, 200, 185, 57, 186, 40, 186, 217, 188, 200, 188, 35, 5, 37, 5, 249, 197, 232, 197, 249, 205, 232, 205, 89, 206, 153, 209, 136, 209, 153, 210, 136, 210, 25, 210, 8, 210, 232, 164, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 80, 200, 164, 169, 164, 200, 56, 169, 80, 187, 100, 187, 108, 169, 100, 169, 108, 169, 220, 169, 128, 217, 140, 187, 80, 200, 140, 169, 80, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 67, 2, 83, 4, 186, 68, 99, 2, 115, 4, 177, 68, 90, 5, 154, 69, 91, 5, 145, 69, 92, 5, 250, 69, 93, 5, 241, 69, 94, 5, 218, 70, 95, 5, 209, 70, 96, 5, 26, 70, 97, 5, 17, 70, 255, 5, 185, 72, 68, 2, 90, 2, 100, 2, 122, 2, 121, 5, 154, 77, 122, 5, 145, 77, 123, 5, 154, 78, 124, 5, 145, 78, 125, 5, 26, 78, 126, 5, 17, 78, 255, 5, 185, 88, 71, 2, 89, 4, 186, 92, 103, 2, 121, 4, 177, 92, 221, 4, 154, 93, 253, 4, 145, 93, 255, 4, 81, 94, 118, 5, 250, 93, 119, 5, 241, 93, 120, 5, 90, 94, 255, 5, 185, 116, 76, 2, 89, 4, 186, 120, 108, 2, 121, 4, 177, 120, 221, 4, 154, 121, 253, 4, 145, 121, 255, 4, 81, 122, 118, 5, 250, 121, 119, 5, 241, 121, 120, 5, 90, 122, 255, 5, 186, 128, 78, 2, 89, 4, 186, 132, 110, 2, 121, 4, 177, 132, 221, 4, 154, 133, 253, 4, 145, 133, 255, 4, 81, 134, 118, 5, 250, 133, 119, 5, 241, 133, 120, 5, 90, 134, 255, 5, 185, 140, 255, 5, 186, 164, 83, 2, 90, 4, 186, 168, 115, 2, 122, 4, 177, 168, 121, 5, 154, 169, 122, 5, 145, 169, 123, 5, 154, 170, 124, 5, 145, 170, 125, 5, 26, 170, 126, 5, 17, 170, 255, 5, 185, 172, 84, 2, 89, 4, 186, 176, 116, 2, 121, 4, 177, 176, 221, 4, 154, 177, 253, 4, 145, 177, 255, 4, 81, 178, 118, 5, 250, 177, 119, 5, 241, 177, 120, 5, 90, 178, 255, 5, 185, 184, 255, 5, 185, 208, 83, 4, 186, 212, 90, 2, 115, 4, 177, 212, 122, 2, 90, 5, 154, 213, 91, 5, 145, 213, 92, 5, 250, 213, 93, 5, 241, 213, 94, 5, 218, 214, 95, 5, 209, 214, 96, 5, 26, 214, 97, 5, 17, 214, 255, 5, 168, 64, 99, 2, 115, 4, 168, 68, 91, 5, 136, 69, 93, 5, 232, 69, 95, 5, 200, 70, 97, 5, 8, 70, 255, 5, 168, 72, 100, 2, 122, 2, 122, 5, 136, 77, 124, 5, 136, 78, 126, 5, 8, 78, 255, 5, 168, 88, 103, 2, 121, 4, 168, 92, 253, 4, 136, 93, 255, 4, 72, 94, 119, 5, 232, 93, 255, 5, 168, 116, 108, 2, 121, 4, 168, 120, 253, 4, 136, 121, 255, 4, 72, 122, 119, 5, 232, 121, 255, 5, 168, 128, 110, 2, 121, 4, 168, 132, 253, 4, 136, 133, 255, 4, 72, 134, 119, 5, 232, 133, 255, 5, 168, 140, 255, 5, 168, 164, 115, 2, 122, 4, 168, 168, 122, 5, 136, 169, 124, 5, 136, 170, 126, 5, 8, 170, 255, 5, 168, 172, 116, 2, 121, 4, 168, 176, 253, 4, 136, 177, 255, 4, 72, 178, 119, 5, 232, 177, 255, 5, 168, 184, 255, 5, 168, 208, 115, 4, 168, 212, 122, 2, 91, 5, 136, 213, 93, 5, 232, 213, 95, 5, 200, 214, 97, 5, 8, 214, 255, 5, 249, 186, 255, 5, 232, 186, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 23, 1, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 97, 18, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 102, 18, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 224, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 228, 6, 0, 0, 228, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 36, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 65, 0, 114, 0, 109, 0, 110, 0, 93, 0, 38, 0, 132, 5, 60, 0, 135, 5, 60, 0, 60, 0, 60, 0, 53, 5, 130, 5, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 196, 26, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 68, 0, 0, 0, 68, 1, 0, 0, 84, 26, 0, 0, 84, 26, 0, 0, 84, 26, 0, 0, 84, 26, 0, 0, 152, 26, 0, 0, 152, 26, 0, 0, 166, 26, 0, 0, 172, 26, 0, 0, 3, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 39, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 136, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 95, 0, 103, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 119, 0, 48, 0, 48, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 192, 0, 124, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 100, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 16, 100, 174, 5, 48, 0, 130, 5, 255, 255, 81, 100, 70, 5, 2, 0, 130, 5, 131, 5, 3, 0, 69, 0, 60, 0, 32, 0, 39, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 217, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 222, 6, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 67, 12, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 71, 12, 0, 0, 71, 12, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 113, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 175, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 38, 0, 66, 0, 60, 0, 99, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 72, 0, 38, 0, 71, 0, 60, 0, 103, 0, 98, 0, 60, 0, 60, 0, 60, 0, 71, 0, 98, 0, 60, 0, 60, 0, 60, 0, 71, 0, 66, 0, 60, 0, 103, 0, 104, 0, 60, 0, 60, 0, 60, 0, 71, 0, 104, 0, 60, 0, 60, 0, 60, 0, 71, 0, 72, 0, 60, 0, 103, 0, 119, 0, 60, 0, 60, 0, 60, 0, 71, 0, 119, 0, 60, 0, 60, 0, 60, 0, 71, 0, 87, 0, 38, 0, 73, 0, 60, 0, 203, 30, 60, 0, 60, 0, 60, 0, 202, 30, 38, 0, 75, 0, 60, 0, 107, 0, 112, 0, 60, 0, 60, 0, 60, 0, 75, 0, 112, 0, 60, 0, 60, 0, 60, 0, 75, 0, 80, 0, 60, 0, 107, 0, 119, 0, 60, 0, 60, 0, 60, 0, 75, 0, 119, 0, 60, 0, 60, 0, 60, 0, 75, 0, 87, 0, 38, 0, 78, 0, 60, 0, 69, 30, 60, 0, 60, 0, 60, 0, 68, 30, 60, 0, 110, 0, 119, 0, 60, 0, 60, 0, 60, 0, 78, 0, 119, 0, 60, 0, 60, 0, 60, 0, 78, 0, 87, 0, 60, 0, 110, 0, 121, 0, 60, 0, 60, 0, 60, 0, 78, 0, 121, 0, 60, 0, 60, 0, 60, 0, 78, 0, 89, 0, 38, 0, 79, 0, 60, 0, 205, 30, 60, 0, 60, 0, 60, 0, 204, 30, 38, 0, 83, 0, 60, 0, 115, 0, 104, 0, 60, 0, 60, 0, 60, 0, 83, 0, 104, 0, 60, 0, 60, 0, 60, 0, 83, 0, 72, 0, 38, 0, 85, 0, 60, 0, 229, 30, 60, 0, 60, 0, 60, 0, 228, 30, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 30, 47, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 164, 28, 0, 0, 164, 28, 0, 0, 164, 28, 0, 0, 164, 28, 0, 0, 240, 34, 0, 0, 240, 34, 0, 0, 176, 41, 0, 0, 22, 42, 0, 0, 6, 47, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 189, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 95, 0, 103, 0, 108, 0, 116, 0, 121, 0, 129, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 152, 0, 48, 0, 48, 0, 48, 0, 158, 0, 166, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 172, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 180, 0, 145, 0, 48, 0, 48, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 137, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 124, 1, 176, 1, 228, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 49, 0, 202, 65, 49, 0, 202, 98, 49, 0, 202, 131, 49, 0, 202, 164, 49, 0, 202, 197, 49, 0, 202, 230, 49, 0, 202, 7, 50, 0, 202, 40, 50, 0, 202, 73, 50, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 0, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 96, 7, 0, 156, 5, 0, 56, 201, 134, 24, 0, 156, 5, 0, 60, 201, 64, 25, 0, 156, 5, 0, 64, 156, 5, 0, 66, 201, 132, 34, 0, 201, 38, 51, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 0, 52, 0, 156, 5, 0, 80, 201, 198, 59, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 160, 60, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 160, 64, 0, 5, 5, 0, 56, 201, 230, 73, 0, 5, 5, 0, 60, 201, 160, 74, 0, 5, 5, 0, 64, 5, 5, 0, 66, 201, 164, 79, 0, 201, 166, 88, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 128, 89, 0, 5, 5, 0, 80, 201, 166, 93, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 48, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 48, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 48, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 94, 0, 201, 70, 95, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 96, 0, 201, 198, 96, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 97, 0, 201, 70, 98, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 99, 0, 201, 198, 99, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 100, 0, 201, 70, 101, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 102, 0, 201, 198, 102, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 69, 5, 5, 2, 69, 201, 134, 103, 0, 201, 70, 104, 0, 201, 6, 105, 0, 201, 198, 105, 0, 201, 134, 106, 0, 201, 70, 107, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 59, 5, 5, 2, 59, 134, 5, 2, 71, 5, 5, 2, 71, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 47, 0, 197, 34, 48, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 44, 0, 197, 34, 45, 0, 134, 5, 2, 83, 5, 5, 2, 83, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 45, 0, 197, 162, 45, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 64, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 135, 5, 2, 45, 5, 142, 0, 0, 70, 5, 2, 45, 5, 142, 0, 0, 135, 5, 2, 45, 5, 144, 0, 0, 70, 5, 2, 45, 5, 144, 0, 0, 135, 5, 2, 45, 5, 156, 0, 0, 70, 5, 2, 45, 5, 156, 0, 0, 135, 5, 2, 45, 5, 196, 0, 0, 70, 5, 2, 45, 5, 196, 0, 0, 135, 5, 2, 45, 5, 150, 0, 0, 70, 5, 2, 45, 5, 150, 0, 0, 135, 5, 2, 45, 5, 160, 0, 0, 70, 5, 2, 45, 5, 160, 0, 0, 135, 5, 2, 45, 5, 206, 0, 0, 70, 5, 2, 45, 5, 206, 0, 0, 70, 5, 2, 45, 5, 210, 0, 0, 135, 5, 3, 55, 5, 142, 0, 0, 70, 5, 3, 55, 5, 142, 0, 0, 135, 5, 4, 55, 5, 142, 0, 0, 70, 5, 4, 55, 5, 142, 0, 0, 135, 5, 3, 55, 5, 144, 0, 0, 70, 5, 3, 55, 5, 144, 0, 0, 135, 5, 2, 55, 5, 156, 0, 0, 70, 5, 2, 55, 5, 156, 0, 0, 135, 5, 2, 55, 5, 196, 0, 0, 70, 5, 2, 55, 5, 196, 0, 0, 135, 5, 2, 55, 5, 210, 0, 0, 70, 5, 2, 55, 5, 210, 0, 0, 135, 5, 3, 55, 5, 156, 0, 0, 70, 5, 3, 55, 5, 156, 0, 0, 135, 5, 3, 55, 5, 196, 0, 0, 70, 5, 3, 55, 5, 196, 0, 0, 135, 5, 3, 55, 5, 150, 0, 0, 70, 5, 3, 55, 5, 150, 0, 0, 135, 5, 3, 55, 5, 160, 0, 0, 70, 5, 3, 55, 5, 160, 0, 0, 135, 5, 3, 55, 5, 206, 0, 0, 70, 5, 3, 55, 5, 206, 0, 0, 135, 5, 4, 55, 5, 138, 0, 0, 70, 5, 4, 55, 5, 138, 0, 0, 135, 5, 4, 55, 5, 136, 0, 0, 70, 5, 4, 55, 5, 136, 0, 0, 135, 5, 4, 55, 5, 150, 0, 0, 70, 5, 4, 55, 5, 150, 0, 0, 135, 5, 4, 55, 5, 156, 0, 0, 70, 5, 4, 55, 5, 156, 0, 0, 135, 5, 4, 55, 5, 196, 0, 0, 70, 5, 4, 55, 5, 196, 0, 0, 70, 5, 3, 55, 5, 210, 0, 0, 70, 5, 4, 55, 5, 146, 0, 0, 135, 5, 3, 63, 5, 142, 0, 0, 70, 5, 3, 63, 5, 142, 0, 0, 135, 5, 2, 63, 5, 136, 0, 0, 70, 5, 2, 63, 5, 136, 0, 0, 135, 5, 2, 63, 5, 156, 0, 0, 70, 5, 2, 63, 5, 156, 0, 0, 135, 5, 3, 63, 5, 138, 0, 0, 70, 5, 3, 63, 5, 138, 0, 0, 135, 5, 3, 63, 5, 136, 0, 0, 70, 5, 3, 63, 5, 136, 0, 0, 135, 5, 3, 63, 5, 150, 0, 0, 70, 5, 3, 63, 5, 150, 0, 0, 135, 5, 3, 63, 5, 156, 0, 0, 70, 5, 3, 63, 5, 156, 0, 0, 135, 5, 3, 63, 5, 196, 0, 0, 70, 5, 3, 63, 5, 196, 0, 0, 70, 5, 3, 63, 5, 146, 0, 0, 135, 5, 4, 69, 5, 136, 0, 0, 70, 5, 4, 69, 5, 136, 0, 0, 70, 5, 4, 69, 5, 150, 0, 0, 135, 5, 3, 69, 5, 142, 0, 0, 70, 5, 3, 69, 5, 142, 0, 0, 135, 5, 4, 69, 5, 142, 0, 0, 70, 5, 4, 69, 5, 142, 0, 0, 135, 5, 4, 69, 5, 150, 0, 0, 135, 5, 4, 69, 5, 164, 0, 0, 70, 5, 4, 69, 5, 164, 0, 0, 135, 5, 3, 69, 5, 138, 0, 0, 70, 5, 3, 69, 5, 138, 0, 0, 135, 5, 3, 69, 5, 136, 0, 0, 70, 5, 3, 69, 5, 136, 0, 0, 135, 5, 3, 69, 5, 150, 0, 0, 70, 5, 3, 69, 5, 150, 0, 0, 135, 5, 3, 69, 5, 156, 0, 0, 70, 5, 3, 69, 5, 156, 0, 0, 135, 5, 3, 69, 5, 196, 0, 0, 70, 5, 3, 69, 5, 196, 0, 0, 135, 5, 4, 69, 5, 156, 0, 0, 70, 5, 4, 69, 5, 156, 0, 0, 70, 5, 3, 69, 5, 146, 0, 0, 70, 5, 4, 69, 5, 146, 0, 0, 135, 5, 4, 69, 5, 138, 0, 0, 70, 5, 4, 69, 5, 138, 0, 0, 135, 5, 4, 69, 5, 196, 0, 0, 70, 5, 4, 69, 5, 196, 0, 0, 135, 5, 4, 69, 5, 182, 0, 0, 70, 5, 4, 69, 5, 182, 0, 0, 135, 5, 4, 69, 5, 154, 0, 0, 70, 5, 4, 69, 5, 154, 0, 0, 135, 5, 2, 79, 5, 142, 0, 0, 70, 5, 2, 79, 5, 142, 0, 0, 135, 5, 2, 79, 5, 144, 0, 0, 70, 5, 2, 79, 5, 144, 0, 0, 135, 5, 2, 79, 5, 156, 0, 0, 70, 5, 2, 79, 5, 156, 0, 0, 135, 5, 2, 79, 5, 196, 0, 0, 70, 5, 2, 79, 5, 196, 0, 0, 135, 5, 2, 79, 5, 150, 0, 0, 70, 5, 2, 79, 5, 150, 0, 0, 135, 5, 2, 79, 5, 160, 0, 0, 70, 5, 2, 79, 5, 160, 0, 0, 135, 5, 2, 79, 5, 206, 0, 0, 70, 5, 2, 79, 5, 206, 0, 0, 70, 5, 2, 79, 5, 210, 0, 0, 5, 5, 2, 45, 5, 142, 0, 0, 5, 5, 2, 45, 5, 144, 0, 0, 5, 5, 2, 45, 5, 156, 0, 0, 5, 5, 2, 45, 5, 196, 0, 0, 5, 5, 2, 45, 5, 150, 0, 0, 5, 5, 2, 45, 5, 160, 0, 0, 5, 5, 2, 45, 5, 206, 0, 0, 5, 5, 2, 45, 5, 210, 0, 0, 5, 5, 3, 55, 5, 142, 0, 0, 5, 5, 4, 55, 5, 142, 0, 0, 5, 5, 3, 55, 5, 144, 0, 0, 5, 5, 2, 55, 5, 156, 0, 0, 5, 5, 2, 55, 5, 196, 0, 0, 5, 5, 2, 55, 5, 210, 0, 0, 5, 5, 3, 55, 5, 156, 0, 0, 5, 5, 3, 55, 5, 196, 0, 0, 5, 5, 3, 55, 5, 150, 0, 0, 5, 5, 3, 55, 5, 160, 0, 0, 5, 5, 3, 55, 5, 206, 0, 0, 5, 5, 4, 55, 5, 138, 0, 0, 5, 5, 4, 55, 5, 136, 0, 0, 5, 5, 4, 55, 5, 150, 0, 0, 5, 5, 4, 55, 5, 156, 0, 0, 5, 5, 4, 55, 5, 196, 0, 0, 5, 5, 3, 55, 5, 210, 0, 0, 5, 5, 4, 55, 5, 146, 0, 0, 5, 5, 3, 63, 5, 142, 0, 0, 5, 5, 2, 63, 5, 136, 0, 0, 5, 5, 2, 63, 5, 156, 0, 0, 5, 5, 3, 63, 5, 138, 0, 0, 5, 5, 3, 63, 5, 136, 0, 0, 5, 5, 3, 63, 5, 150, 0, 0, 5, 5, 3, 63, 5, 156, 0, 0, 5, 5, 3, 63, 5, 196, 0, 0, 5, 5, 3, 63, 5, 146, 0, 0, 5, 5, 4, 69, 5, 136, 0, 0, 5, 5, 4, 69, 5, 150, 0, 0, 5, 5, 3, 69, 5, 142, 0, 0, 5, 5, 4, 69, 5, 142, 0, 0, 5, 5, 4, 69, 5, 164, 0, 0, 5, 5, 3, 69, 5, 138, 0, 0, 5, 5, 3, 69, 5, 136, 0, 0, 5, 5, 3, 69, 5, 150, 0, 0, 5, 5, 3, 69, 5, 156, 0, 0, 5, 5, 3, 69, 5, 196, 0, 0, 5, 5, 4, 69, 5, 156, 0, 0, 5, 5, 3, 69, 5, 146, 0, 0, 5, 5, 4, 69, 5, 146, 0, 0, 5, 5, 4, 69, 5, 138, 0, 0, 5, 5, 4, 69, 5, 196, 0, 0, 5, 5, 4, 69, 5, 182, 0, 0, 5, 5, 4, 69, 5, 154, 0, 0, 5, 5, 2, 79, 5, 142, 0, 0, 5, 5, 2, 79, 5, 144, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 150, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 206, 0, 0, 5, 5, 2, 79, 5, 210, 0, 0, 134, 5, 2, 59, 5, 162, 0, 0, 5, 5, 2, 59, 5, 162, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 134, 5, 2, 71, 5, 142, 0, 0, 5, 5, 2, 71, 5, 142, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 2, 192, 197, 162, 41, 30, 2, 192, 197, 226, 42, 30, 3, 192, 197, 34, 43, 30, 3, 192, 197, 98, 150, 30, 3, 192, 197, 162, 36, 30, 1, 192, 197, 162, 37, 30, 1, 192, 197, 226, 38, 30, 2, 192, 197, 34, 39, 30, 2, 192, 197, 98, 30, 2, 12, 0, 30, 2, 0, 192, 197, 162, 31, 2, 0, 192, 197, 226, 34, 30, 1, 192, 197, 34, 35, 30, 1, 192, 197, 98, 72, 0, 2, 237, 135, 5, 104, 0, 2, 237, 70, 5, 36, 1, 197, 162, 37, 1, 0, 192, 197, 98, 0, 54, 156, 5, 39, 0, 36, 30, 66, 0, 130, 30, 32, 0, 135, 30, 15, 0, 135, 30, 11, 192, 197, 34, 136, 30, 11, 192, 197, 98, 137, 30, 11, 192, 197, 162, 150, 30, 11, 192, 197, 226, 152, 30, 12, 192, 197, 34, 130, 30, 9, 192, 197, 226, 131, 30, 10, 192, 197, 34, 132, 30, 10, 192, 197, 98, 133, 30, 10, 192, 197, 162, 134, 30, 10, 192, 197, 226, 41, 30, 15, 0, 41, 30, 8, 192, 197, 162, 42, 30, 8, 192, 197, 226, 43, 30, 9, 192, 197, 34, 128, 30, 9, 192, 197, 98, 129, 30, 9, 192, 197, 162, 36, 30, 7, 192, 197, 98, 37, 30, 7, 192, 197, 162, 38, 30, 7, 192, 197, 226, 39, 30, 8, 192, 197, 34, 40, 30, 8, 192, 197, 98, 30, 2, 32, 0, 5, 30, 15, 0, 5, 30, 6, 192, 197, 34, 6, 30, 6, 192, 197, 98, 7, 30, 6, 192, 197, 162, 34, 30, 6, 192, 197, 226, 35, 30, 7, 192, 197, 34, 30, 2, 4, 192, 197, 226, 31, 2, 5, 192, 197, 34, 2, 30, 5, 192, 197, 98, 3, 30, 5, 192, 197, 162, 4, 30, 5, 192, 197, 226, 119, 0, 15, 0, 119, 0, 4, 247, 70, 5, 36, 1, 3, 192, 197, 226, 37, 1, 4, 192, 197, 34, 116, 1, 4, 192, 197, 98, 117, 1, 4, 192, 197, 162, 66, 0, 2, 247, 135, 5, 72, 0, 3, 247, 135, 5, 87, 0, 4, 247, 135, 5, 98, 0, 2, 247, 70, 5, 104, 0, 3, 247, 70, 5, 0, 58, 156, 5, 48, 0, 35, 3, 2, 251, 134, 5, 0, 62, 156, 5, 20, 0, 128, 30, 37, 0, 133, 30, 20, 0, 136, 30, 9, 0, 136, 30, 15, 192, 197, 226, 137, 30, 16, 192, 197, 34, 152, 30, 16, 192, 197, 98, 133, 30, 15, 192, 197, 34, 134, 30, 15, 192, 197, 98, 135, 30, 15, 192, 197, 162, 128, 30, 13, 192, 197, 226, 129, 30, 14, 192, 197, 34, 130, 30, 14, 192, 197, 98, 131, 30, 14, 192, 197, 162, 132, 30, 14, 192, 197, 226, 117, 1, 15, 0, 117, 1, 12, 192, 197, 162, 84, 30, 12, 192, 197, 226, 85, 30, 13, 192, 197, 34, 86, 30, 13, 192, 197, 98, 87, 30, 13, 192, 197, 162, 80, 0, 2, 255, 135, 5, 87, 0, 3, 255, 135, 5, 112, 0, 2, 255, 70, 5, 119, 0, 3, 255, 70, 5, 116, 1, 12, 192, 197, 98, 0, 68, 156, 5, 36, 0, 131, 30, 63, 0, 152, 30, 32, 0, 245, 30, 15, 0, 245, 30, 23, 192, 197, 98, 246, 30, 23, 192, 197, 162, 247, 30, 23, 192, 197, 226, 248, 30, 24, 192, 197, 34, 249, 30, 24, 192, 197, 98, 152, 30, 22, 192, 197, 34, 153, 30, 22, 192, 197, 98, 242, 30, 22, 192, 197, 162, 243, 30, 22, 192, 197, 226, 244, 30, 23, 192, 197, 34, 135, 30, 15, 0, 135, 30, 20, 192, 197, 226, 136, 30, 21, 192, 197, 34, 137, 30, 21, 192, 197, 98, 142, 30, 21, 192, 197, 162, 143, 30, 21, 192, 197, 226, 131, 30, 19, 192, 197, 226, 132, 30, 20, 192, 197, 34, 133, 30, 20, 192, 197, 98, 134, 30, 20, 192, 197, 162, 118, 1, 30, 0, 51, 2, 16, 0, 51, 2, 18, 192, 197, 226, 7, 3, 255, 255, 2, 69, 134, 5, 128, 30, 19, 192, 197, 34, 129, 30, 19, 192, 197, 98, 130, 30, 19, 192, 197, 162, 118, 1, 17, 192, 197, 226, 119, 1, 18, 192, 197, 34, 120, 1, 18, 192, 197, 98, 50, 2, 18, 192, 197, 162, 221, 0, 15, 0, 221, 0, 16, 192, 197, 162, 253, 0, 16, 192, 197, 226, 255, 0, 17, 192, 197, 34, 116, 1, 17, 192, 197, 98, 117, 1, 17, 192, 197, 162, 87, 0, 255, 255, 3, 69, 135, 5, 89, 0, 255, 255, 4, 69, 135, 5, 119, 0, 255, 255, 3, 69, 70, 5, 121, 0, 255, 255, 4, 69, 70, 5, 0, 70, 156, 5, 48, 0, 35, 3, 255, 255, 2, 71, 134, 5, 0, 78, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 27, 192, 197, 34, 41, 30, 27, 192, 197, 98, 42, 30, 27, 192, 197, 162, 43, 30, 27, 192, 197, 226, 150, 30, 28, 192, 197, 34, 36, 30, 26, 192, 197, 34, 37, 30, 26, 192, 197, 98, 38, 30, 26, 192, 197, 162, 39, 30, 26, 192, 197, 226, 30, 2, 12, 0, 30, 2, 25, 192, 197, 34, 31, 2, 25, 192, 197, 98, 34, 30, 25, 192, 197, 162, 35, 30, 25, 192, 197, 226, 72, 0, 255, 255, 2, 79, 135, 5, 104, 0, 255, 255, 2, 79, 70, 5, 36, 1, 24, 192, 197, 162, 37, 1, 24, 192, 197, 226, 0, 82, 156, 5, 48, 0, 35, 3, 255, 255, 2, 83, 134, 5, 0, 46, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 29, 192, 197, 34, 39, 30, 29, 192, 197, 98, 41, 30, 29, 192, 197, 162, 43, 30, 29, 192, 197, 226, 150, 30, 30, 192, 197, 34, 104, 0, 2, 237, 5, 5, 37, 1, 28, 192, 197, 98, 31, 2, 28, 192, 197, 162, 35, 30, 28, 192, 197, 226, 0, 54, 5, 5, 20, 0, 37, 30, 37, 0, 131, 30, 20, 0, 137, 30, 9, 0, 137, 30, 34, 192, 197, 34, 150, 30, 34, 192, 197, 98, 152, 30, 34, 192, 197, 162, 131, 30, 33, 192, 197, 98, 133, 30, 33, 192, 197, 162, 135, 30, 33, 192, 197, 226, 37, 30, 32, 192, 197, 34, 39, 30, 32, 192, 197, 98, 41, 30, 32, 192, 197, 162, 43, 30, 32, 192, 197, 226, 129, 30, 33, 192, 197, 34, 31, 2, 15, 0, 31, 2, 30, 192, 197, 226, 3, 30, 31, 192, 197, 34, 5, 30, 31, 192, 197, 98, 7, 30, 31, 192, 197, 162, 35, 30, 31, 192, 197, 226, 98, 0, 2, 247, 5, 5, 104, 0, 3, 247, 5, 5, 119, 0, 4, 247, 5, 5, 37, 1, 30, 192, 197, 98, 117, 1, 30, 192, 197, 162, 0, 58, 5, 5, 48, 0, 35, 3, 2, 251, 5, 5, 0, 62, 5, 5, 10, 0, 129, 30, 20, 0, 135, 30, 9, 0, 135, 30, 36, 192, 197, 98, 137, 30, 36, 192, 197, 162, 152, 30, 36, 192, 197, 226, 129, 30, 35, 192, 197, 162, 131, 30, 35, 192, 197, 226, 133, 30, 36, 192, 197, 34, 112, 0, 2, 255, 5, 5, 119, 0, 3, 255, 5, 5, 117, 1, 34, 192, 197, 226, 85, 30, 35, 192, 197, 34, 87, 30, 35, 192, 197, 98, 0, 68, 5, 5, 19, 0, 133, 30, 32, 0, 153, 30, 15, 0, 153, 30, 40, 192, 197, 34, 243, 30, 40, 192, 197, 98, 245, 30, 40, 192, 197, 162, 247, 30, 40, 192, 197, 226, 249, 30, 41, 192, 197, 34, 133, 30, 38, 192, 197, 226, 135, 30, 39, 192, 197, 34, 137, 30, 39, 192, 197, 98, 143, 30, 39, 192, 197, 162, 152, 30, 39, 192, 197, 226, 119, 1, 16, 0, 119, 1, 37, 192, 197, 226, 51, 2, 38, 192, 197, 34, 7, 3, 255, 255, 2, 69, 5, 5, 129, 30, 38, 192, 197, 98, 131, 30, 38, 192, 197, 162, 119, 0, 255, 255, 3, 69, 5, 5, 121, 0, 255, 255, 4, 69, 5, 5, 253, 0, 37, 192, 197, 34, 255, 0, 37, 192, 197, 98, 117, 1, 37, 192, 197, 162, 0, 70, 5, 5, 48, 0, 35, 3, 255, 255, 2, 71, 5, 5, 0, 78, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 42, 192, 197, 34, 39, 30, 42, 192, 197, 98, 41, 30, 42, 192, 197, 162, 43, 30, 42, 192, 197, 226, 150, 30, 43, 192, 197, 34, 104, 0, 255, 255, 2, 79, 5, 5, 37, 1, 41, 192, 197, 98, 31, 2, 41, 192, 197, 162, 35, 30, 41, 192, 197, 226, 0, 82, 5, 5, 48, 0, 35, 3, 255, 255, 2, 83, 5, 5, 156, 58, 196, 162, 48, 0, 35, 3, 43, 192, 197, 98, 5, 58, 196, 162, 48, 0, 35, 3, 43, 192, 197, 162, 156, 68, 196, 160, 48, 0, 7, 3, 43, 192, 197, 226, 5, 68, 196, 160, 48, 0, 7, 3, 44, 192, 197, 34, 156, 82, 196, 162, 48, 0, 35, 3, 44, 192, 197, 98, 5, 82, 196, 162, 48, 0, 35, 3, 44, 192, 197, 162, 156, 70, 196, 190, 48, 0, 35, 3, 44, 192, 197, 226, 5, 70, 196, 190, 48, 0, 35, 3, 45, 192, 197, 34, 156, 82, 196, 190, 48, 0, 35, 3, 45, 192, 197, 98, 5, 82, 196, 190, 48, 0, 35, 3, 45, 192, 197, 162, 156, 70, 196, 162, 48, 0, 35, 3, 45, 192, 197, 226, 5, 70, 196, 162, 48, 0, 35, 3, 46, 192, 197, 34, 156, 68, 196, 196, 48, 0, 7, 3, 46, 192, 197, 98, 5, 68, 196, 196, 48, 0, 7, 3, 46, 192, 197, 162, 156, 68, 196, 210, 48, 0, 7, 3, 46, 192, 197, 226, 5, 68, 196, 210, 48, 0, 7, 3, 47, 192, 197, 34, 156, 68, 196, 204, 48, 0, 7, 3, 47, 192, 197, 98, 5, 68, 196, 204, 48, 0, 7, 3, 47, 192, 197, 162, 50, 0, 66, 0, 67, 0, 72, 0, 73, 0, 80, 0, 81, 0, 87, 0, 88, 0, 89, 0, 90, 0, 98, 0, 99, 0, 104, 0, 105, 0, 112, 0, 113, 0, 119, 0, 120, 0, 121, 0, 122, 0, 221, 0, 222, 0, 253, 0, 254, 0, 255, 0, 0, 1, 36, 1, 38, 1, 116, 1, 121, 1, 30, 2, 32, 2, 50, 2, 52, 2, 2, 30, 8, 30, 34, 30, 44, 30, 84, 30, 88, 30, 128, 30, 138, 30, 142, 30, 144, 30, 150, 30, 151, 30, 152, 30, 154, 30, 242, 30, 250, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 72, 186, 76, 185, 80, 36, 4, 185, 100, 58, 4, 186, 112, 60, 4, 185, 128, 185, 132, 74, 4, 100, 4, 185, 156, 185, 160, 185, 168, 102, 4, 185, 180, 112, 4, 185, 192, 185, 196, 185, 200, 185, 204, 185, 208, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 114, 4, 168, 72, 168, 76, 168, 80, 120, 4, 168, 100, 132, 4, 168, 112, 134, 4, 168, 128, 168, 132, 142, 4, 156, 4, 168, 156, 168, 160, 168, 168, 158, 4, 168, 180, 164, 4, 168, 192, 168, 196, 168, 200, 168, 204, 168, 208, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 152, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 70, 186, 77, 154, 77, 250, 77, 90, 78, 186, 105, 154, 105, 250, 105, 90, 106, 217, 72, 154, 138, 185, 153, 153, 153, 249, 153, 153, 154, 89, 154, 136, 14, 217, 154, 185, 189, 153, 189, 249, 189, 89, 190, 153, 205, 185, 212, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 70, 168, 77, 136, 77, 232, 77, 72, 78, 168, 105, 136, 105, 232, 105, 72, 106, 200, 72, 136, 138, 168, 153, 136, 153, 232, 153, 136, 154, 72, 154, 128, 14, 200, 154, 168, 189, 136, 189, 232, 189, 72, 190, 136, 205, 168, 212, 72, 206, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 69, 136, 69, 249, 69, 232, 69, 185, 70, 168, 70, 25, 70, 8, 70, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 249, 85, 232, 85, 217, 85, 200, 85, 185, 86, 168, 86, 249, 86, 232, 86, 249, 101, 232, 101, 89, 103, 72, 103, 154, 106, 136, 106, 58, 107, 40, 107, 218, 105, 200, 105, 166, 4, 168, 4, 186, 106, 168, 108, 6, 8, 8, 8, 250, 113, 232, 113, 249, 118, 232, 118, 168, 164, 153, 129, 136, 129, 249, 130, 232, 130, 25, 130, 8, 130, 121, 131, 104, 131, 89, 131, 72, 131, 154, 137, 136, 137, 170, 4, 172, 4, 26, 138, 8, 138, 10, 8, 185, 148, 168, 148, 57, 155, 40, 155, 217, 153, 200, 153, 121, 154, 104, 154, 12, 8, 14, 8, 153, 169, 136, 169, 249, 170, 232, 170, 25, 170, 8, 170, 154, 173, 136, 173, 250, 173, 232, 173, 250, 174, 232, 174, 26, 174, 8, 174, 249, 182, 232, 182, 25, 182, 8, 182, 185, 184, 168, 184, 153, 190, 136, 190, 57, 191, 40, 191, 217, 189, 200, 189, 57, 190, 40, 190, 121, 190, 104, 190, 174, 4, 176, 4, 249, 197, 232, 197, 249, 205, 232, 205, 89, 206, 153, 209, 136, 209, 185, 210, 168, 210, 25, 210, 8, 210, 232, 172, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 172, 169, 172, 200, 56, 169, 76, 187, 104, 187, 112, 169, 104, 169, 112, 169, 216, 169, 136, 217, 152, 187, 76, 200, 152, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 68, 72, 4, 186, 64, 104, 4, 177, 64, 36, 5, 250, 65, 37, 5, 241, 65, 255, 5, 185, 84, 66, 4, 186, 88, 72, 4, 186, 92, 87, 4, 186, 96, 98, 4, 177, 88, 104, 4, 177, 92, 119, 4, 177, 96, 36, 5, 250, 93, 37, 5, 241, 93, 116, 5, 250, 97, 117, 5, 241, 97, 255, 5, 186, 104, 255, 5, 185, 116, 80, 4, 186, 120, 87, 4, 186, 124, 112, 4, 177, 120, 119, 4, 177, 124, 116, 5, 250, 125, 117, 5, 241, 125, 255, 5, 186, 136, 87, 4, 186, 140, 89, 4, 186, 144, 119, 4, 177, 140, 121, 4, 177, 144, 221, 4, 154, 145, 253, 4, 145, 145, 255, 4, 81, 146, 116, 5, 250, 141, 117, 5, 241, 141, 118, 5, 250, 145, 119, 5, 241, 145, 120, 5, 90, 146, 255, 5, 185, 152, 255, 5, 186, 172, 72, 4, 186, 176, 104, 4, 177, 176, 36, 5, 250, 177, 37, 5, 241, 177, 255, 5, 185, 188, 255, 5, 168, 68, 104, 4, 168, 64, 37, 5, 232, 65, 255, 5, 168, 84, 98, 4, 168, 88, 104, 4, 168, 92, 119, 4, 168, 96, 37, 5, 232, 93, 117, 5, 232, 97, 255, 5, 168, 104, 255, 5, 168, 116, 112, 4, 168, 120, 119, 4, 168, 124, 117, 5, 232, 125, 255, 5, 168, 136, 119, 4, 168, 140, 121, 4, 168, 144, 253, 4, 136, 145, 255, 4, 72, 146, 117, 5, 232, 141, 119, 5, 232, 145, 255, 5, 168, 152, 255, 5, 168, 172, 104, 4, 168, 176, 37, 5, 232, 177, 255, 5, 168, 188, 255, 5, 26, 107, 255, 5, 8, 107, 255, 5, 250, 138, 255, 5, 232, 138, 255, 5, 25, 191, 255, 5, 8, 191, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 115, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 60, 12, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 65, 12, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 14, 0, 0, 0, 14, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 14, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 105, 0, 100, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 105, 0, 100, 0, 95, 0, 73, 0, 68, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 240, 23, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 244, 23, 0, 0, 244, 23, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 136, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 49, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 105, 0, 115, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 0, 0, 167, 220, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 98, 0, 60, 0, 225, 0, 60, 0, 60, 0, 60, 0, 193, 0, 38, 0, 100, 0, 60, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 60, 0, 240, 0, 60, 0, 60, 0, 60, 0, 208, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 102, 0, 60, 0, 233, 0, 60, 0, 60, 0, 60, 0, 201, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 106, 0, 60, 0, 237, 0, 60, 0, 60, 0, 60, 0, 205, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 112, 0, 60, 0, 243, 0, 60, 0, 60, 0, 60, 0, 211, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 118, 0, 60, 0, 250, 0, 60, 0, 60, 0, 60, 0, 218, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 122, 0, 60, 0, 253, 0, 60, 0, 60, 0, 60, 0, 221, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 192, 1, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 206, 49, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 188, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 244, 38, 0, 0, 244, 38, 0, 0, 132, 40, 0, 0, 132, 40, 0, 0, 168, 44, 0, 0, 168, 44, 0, 0, 178, 49, 0, 0, 178, 49, 0, 0, 178, 49, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 98, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 145, 0, 153, 0, 159, 0, 167, 0, 175, 0, 183, 0, 48, 0, 56, 0, 182, 0, 190, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 207, 0, 215, 0, 48, 0, 48, 0, 48, 0, 223, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 231, 0, 48, 0, 48, 0, 239, 0, 48, 0, 245, 0, 253, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 8, 1, 15, 1, 48, 0, 19, 1, 48, 0, 27, 1, 34, 1, 42, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 56, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 59, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 205, 0, 48, 0, 48, 0, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 67, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 75, 1, 48, 0, 48, 0, 81, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 68, 2, 124, 2, 188, 2, 192, 0, 216, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 31, 0, 202, 1, 32, 0, 202, 34, 32, 0, 202, 67, 32, 0, 202, 100, 32, 0, 202, 133, 32, 0, 202, 166, 32, 0, 202, 199, 32, 0, 202, 232, 32, 0, 202, 9, 33, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 198, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 102, 2, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 134, 3, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 166, 4, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 102, 6, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 201, 198, 7, 0, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 38, 9, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 38, 11, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 70, 12, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 102, 13, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 70, 15, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 201, 166, 16, 0, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 134, 5, 31, 43, 196, 142, 156, 42, 196, 154, 156, 42, 198, 1, 0, 0, 198, 33, 0, 0, 198, 129, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 134, 5, 49, 51, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 134, 5, 27, 59, 196, 142, 156, 58, 196, 150, 156, 58, 134, 5, 2, 49, 196, 154, 156, 68, 196, 138, 156, 70, 134, 5, 61, 71, 196, 142, 156, 70, 196, 154, 156, 70, 198, 65, 0, 0, 5, 5, 127, 12, 198, 161, 0, 0, 196, 138, 156, 82, 134, 5, 51, 83, 196, 142, 156, 82, 196, 150, 156, 82, 134, 5, 17, 91, 160, 5, 38, 93, 197, 66, 23, 0, 196, 138, 5, 42, 5, 5, 31, 43, 196, 142, 5, 42, 196, 154, 5, 42, 198, 97, 0, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 5, 5, 49, 51, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 5, 5, 27, 59, 196, 142, 5, 58, 196, 150, 5, 58, 5, 5, 2, 49, 196, 154, 5, 68, 196, 138, 5, 70, 5, 5, 61, 71, 196, 142, 5, 70, 196, 154, 5, 70, 193, 3, 111, 93, 5, 5, 125, 12, 198, 193, 0, 0, 196, 138, 5, 82, 5, 5, 51, 83, 196, 142, 5, 82, 196, 150, 5, 82, 5, 5, 17, 91, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 18, 0, 201, 166, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 21, 0, 201, 166, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 23, 0, 201, 230, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 26, 0, 201, 38, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 28, 0, 201, 198, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 31, 0, 201, 102, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 1, 0, 198, 34, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 2, 0, 197, 162, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 33, 0, 201, 6, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 2, 0, 197, 162, 4, 0, 198, 34, 3, 0, 197, 226, 4, 0, 198, 98, 3, 0, 198, 162, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 36, 0, 201, 166, 37, 0, 198, 226, 3, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 198, 38, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 40, 0, 5, 71, 238, 102, 201, 134, 41, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 6, 0, 192, 0, 0, 0, 197, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 7, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 197, 34, 8, 0, 197, 99, 8, 0, 197, 194, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 9, 0, 197, 67, 9, 0, 197, 162, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 226, 5, 0, 197, 34, 10, 0, 5, 5, 10, 121, 197, 98, 10, 0, 197, 162, 10, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 42, 0, 201, 70, 44, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 46, 0, 201, 70, 47, 0, 201, 102, 48, 0, 201, 134, 49, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 50, 0, 201, 198, 51, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 52, 0, 201, 6, 54, 0, 201, 38, 55, 0, 201, 70, 56, 0, 201, 102, 57, 0, 201, 134, 58, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 59, 0, 201, 134, 61, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 63, 0, 201, 134, 64, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 65, 0, 201, 198, 66, 0, 201, 230, 67, 0, 201, 102, 69, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 2, 0, 197, 162, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 70, 0, 201, 102, 72, 0, 201, 230, 73, 0, 201, 6, 75, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 3, 0, 197, 98, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 76, 0, 201, 70, 77, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 78, 0, 201, 134, 79, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 5, 31, 43, 5, 162, 0, 0, 5, 5, 31, 43, 5, 162, 0, 0, 193, 4, 111, 93, 5, 162, 0, 0, 134, 5, 49, 51, 5, 162, 0, 0, 5, 5, 49, 51, 5, 162, 0, 0, 134, 5, 27, 59, 5, 162, 0, 0, 5, 5, 27, 59, 5, 162, 0, 0, 134, 5, 51, 83, 5, 162, 0, 0, 5, 5, 51, 83, 5, 162, 0, 0, 134, 5, 61, 71, 5, 190, 0, 0, 5, 5, 61, 71, 5, 190, 0, 0, 193, 3, 111, 93, 5, 190, 0, 0, 134, 5, 51, 83, 5, 190, 0, 0, 5, 5, 51, 83, 5, 190, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 134, 5, 61, 71, 5, 162, 0, 0, 5, 5, 61, 71, 5, 162, 0, 0, 193, 3, 111, 93, 5, 162, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 134, 5, 49, 51, 5, 160, 0, 0, 5, 5, 49, 51, 5, 160, 0, 0, 193, 3, 111, 93, 5, 164, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 5, 31, 43, 5, 200, 0, 0, 5, 5, 31, 43, 5, 200, 0, 0, 193, 4, 111, 93, 5, 200, 0, 0, 134, 5, 49, 51, 5, 204, 0, 0, 5, 5, 49, 51, 5, 204, 0, 0, 134, 5, 49, 51, 5, 208, 0, 0, 5, 5, 49, 51, 5, 208, 0, 0, 134, 5, 27, 59, 5, 208, 0, 0, 5, 5, 27, 59, 5, 208, 0, 0, 134, 5, 51, 83, 5, 198, 0, 0, 5, 5, 51, 83, 5, 198, 0, 0, 134, 5, 51, 83, 5, 208, 0, 0, 5, 5, 51, 83, 5, 208, 0, 0, 134, 5, 51, 83, 5, 204, 0, 0, 5, 5, 51, 83, 5, 204, 0, 0, 134, 5, 31, 43, 5, 196, 0, 0, 5, 5, 31, 43, 5, 196, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 134, 5, 49, 51, 5, 196, 0, 0, 5, 5, 49, 51, 5, 196, 0, 0, 134, 5, 27, 59, 5, 196, 0, 0, 5, 5, 27, 59, 5, 196, 0, 0, 134, 5, 61, 71, 5, 196, 0, 0, 5, 5, 61, 71, 5, 196, 0, 0, 193, 3, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 61, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 61, 71, 5, 190, 0, 0, 5, 196, 0, 0, 193, 3, 111, 93, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 51, 83, 5, 196, 0, 0, 5, 5, 51, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 51, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 51, 83, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 17, 91, 5, 196, 0, 0, 5, 5, 17, 91, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 6, 121, 197, 226, 5, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 34, 6, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 98, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 226, 6, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 34, 7, 0, 5, 5, 107, 121, 197, 98, 7, 0, 5, 5, 111, 121, 197, 162, 7, 0, 5, 5, 115, 121, 197, 226, 7, 0, 5, 5, 119, 121, 197, 34, 8, 0, 197, 99, 8, 0, 197, 194, 8, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 2, 9, 0, 197, 67, 9, 0, 197, 162, 9, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 226, 9, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 226, 5, 0, 197, 34, 10, 0, 5, 5, 10, 121, 197, 98, 10, 0, 197, 162, 10, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 3, 0, 1, 3, 31, 235, 134, 5, 8, 3, 198, 129, 10, 3, 198, 161, 65, 3, 31, 235, 134, 5, 0, 50, 156, 5, 1, 0, 1, 3, 49, 243, 134, 5, 65, 3, 49, 243, 134, 5, 0, 58, 156, 5, 1, 0, 1, 3, 27, 251, 134, 5, 65, 3, 27, 251, 134, 5, 0, 70, 156, 5, 2, 0, 1, 3, 255, 255, 61, 71, 134, 5, 8, 3, 0, 192, 198, 65, 65, 3, 255, 255, 61, 71, 134, 5, 0, 82, 156, 5, 1, 0, 1, 3, 255, 255, 51, 83, 134, 5, 65, 3, 255, 255, 51, 83, 134, 5, 0, 90, 156, 5, 1, 0, 1, 3, 255, 255, 17, 91, 134, 5, 65, 3, 255, 255, 17, 91, 134, 5, 0, 42, 5, 5, 3, 0, 1, 3, 31, 235, 5, 5, 8, 3, 0, 192, 198, 97, 10, 3, 255, 255, 111, 93, 193, 4, 65, 3, 31, 235, 5, 5, 0, 50, 5, 5, 1, 0, 1, 3, 49, 243, 5, 5, 65, 3, 49, 243, 5, 5, 0, 58, 5, 5, 1, 0, 1, 3, 27, 251, 5, 5, 65, 3, 27, 251, 5, 5, 0, 70, 5, 5, 2, 0, 1, 3, 255, 255, 61, 71, 5, 5, 8, 3, 255, 255, 111, 93, 193, 3, 65, 3, 255, 255, 61, 71, 5, 5, 0, 82, 5, 5, 1, 0, 1, 3, 255, 255, 51, 83, 5, 5, 65, 3, 255, 255, 51, 83, 5, 5, 0, 90, 5, 5, 1, 0, 1, 3, 255, 255, 17, 91, 5, 5, 65, 3, 255, 255, 17, 91, 5, 5, 156, 42, 196, 162, 3, 0, 1, 3, 197, 162, 8, 3, 0, 192, 198, 226, 10, 3, 1, 192, 198, 34, 65, 3, 197, 162, 5, 42, 196, 162, 3, 0, 1, 3, 0, 192, 197, 98, 8, 3, 1, 192, 198, 98, 10, 3, 0, 192, 197, 162, 65, 3, 0, 192, 197, 98, 156, 50, 196, 162, 1, 0, 1, 3, 0, 192, 197, 226, 65, 3, 0, 192, 197, 226, 5, 50, 196, 162, 1, 0, 1, 3, 1, 192, 197, 34, 65, 3, 1, 192, 197, 34, 156, 58, 196, 162, 1, 0, 1, 3, 1, 192, 197, 98, 65, 3, 1, 192, 197, 98, 5, 58, 196, 162, 1, 0, 1, 3, 1, 192, 197, 162, 65, 3, 1, 192, 197, 162, 156, 82, 196, 162, 1, 0, 1, 3, 1, 192, 197, 226, 65, 3, 1, 192, 197, 226, 5, 82, 196, 162, 1, 0, 1, 3, 2, 192, 197, 34, 65, 3, 2, 192, 197, 34, 156, 70, 196, 190, 2, 0, 1, 3, 2, 192, 197, 98, 8, 3, 1, 192, 198, 162, 65, 3, 2, 192, 197, 98, 5, 70, 196, 190, 2, 0, 1, 3, 2, 192, 197, 162, 8, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 156, 82, 196, 190, 1, 0, 1, 3, 3, 192, 197, 34, 65, 3, 3, 192, 197, 34, 5, 82, 196, 190, 1, 0, 1, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 98, 156, 70, 196, 162, 2, 0, 1, 3, 3, 192, 197, 226, 8, 3, 2, 192, 198, 162, 65, 3, 3, 192, 197, 226, 5, 70, 196, 162, 2, 0, 1, 3, 4, 192, 197, 34, 8, 3, 4, 192, 197, 98, 65, 3, 4, 192, 197, 34, 156, 50, 196, 160, 1, 0, 1, 3, 5, 192, 197, 34, 65, 3, 5, 192, 197, 34, 5, 50, 196, 160, 1, 0, 1, 3, 5, 192, 197, 98, 65, 3, 5, 192, 197, 98, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 3, 0, 1, 3, 12, 192, 197, 162, 8, 3, 4, 192, 198, 34, 10, 3, 4, 192, 198, 98, 65, 3, 12, 192, 197, 162, 5, 42, 196, 200, 3, 0, 1, 3, 12, 192, 197, 226, 8, 3, 4, 192, 198, 162, 10, 3, 13, 192, 197, 34, 65, 3, 12, 192, 197, 226, 156, 50, 196, 204, 1, 0, 1, 3, 13, 192, 197, 98, 65, 3, 13, 192, 197, 98, 5, 50, 196, 204, 1, 0, 1, 3, 13, 192, 197, 162, 65, 3, 13, 192, 197, 162, 156, 50, 196, 208, 1, 0, 1, 3, 13, 192, 197, 226, 65, 3, 13, 192, 197, 226, 5, 50, 196, 208, 1, 0, 1, 3, 14, 192, 197, 34, 65, 3, 14, 192, 197, 34, 156, 58, 196, 208, 1, 0, 1, 3, 14, 192, 197, 98, 65, 3, 14, 192, 197, 98, 5, 58, 196, 208, 1, 0, 1, 3, 14, 192, 197, 162, 65, 3, 14, 192, 197, 162, 156, 82, 196, 198, 1, 0, 1, 3, 14, 192, 197, 226, 65, 3, 14, 192, 197, 226, 5, 82, 196, 198, 1, 0, 1, 3, 15, 192, 197, 34, 65, 3, 15, 192, 197, 34, 156, 82, 196, 208, 1, 0, 1, 3, 15, 192, 197, 98, 65, 3, 15, 192, 197, 98, 5, 82, 196, 208, 1, 0, 1, 3, 15, 192, 197, 162, 65, 3, 15, 192, 197, 162, 156, 82, 196, 204, 1, 0, 1, 3, 15, 192, 197, 226, 65, 3, 15, 192, 197, 226, 5, 82, 196, 204, 1, 0, 1, 3, 16, 192, 197, 34, 65, 3, 16, 192, 197, 34, 156, 42, 196, 196, 3, 0, 1, 3, 16, 192, 197, 98, 8, 3, 4, 192, 198, 226, 10, 3, 5, 192, 198, 34, 65, 3, 16, 192, 197, 98, 5, 42, 196, 196, 3, 0, 1, 3, 16, 192, 197, 162, 8, 3, 5, 192, 198, 98, 10, 3, 16, 192, 197, 226, 65, 3, 16, 192, 197, 162, 156, 50, 196, 196, 1, 0, 1, 3, 17, 192, 197, 34, 65, 3, 17, 192, 197, 34, 5, 50, 196, 196, 1, 0, 1, 3, 17, 192, 197, 98, 65, 3, 17, 192, 197, 98, 156, 58, 196, 196, 1, 0, 1, 3, 17, 192, 197, 162, 65, 3, 17, 192, 197, 162, 5, 58, 196, 196, 1, 0, 1, 3, 17, 192, 197, 226, 65, 3, 17, 192, 197, 226, 156, 70, 196, 196, 2, 0, 1, 3, 18, 192, 197, 34, 8, 3, 5, 192, 198, 162, 65, 3, 18, 192, 197, 34, 5, 70, 196, 196, 2, 0, 1, 3, 18, 192, 197, 98, 8, 3, 18, 192, 197, 162, 65, 3, 18, 192, 197, 98, 18, 0, 197, 227, 2, 0, 1, 3, 19, 192, 197, 67, 8, 3, 5, 192, 198, 227, 65, 3, 19, 192, 197, 67, 19, 0, 197, 163, 2, 0, 1, 3, 20, 192, 197, 3, 8, 3, 20, 192, 197, 99, 65, 3, 20, 192, 197, 3, 156, 82, 196, 196, 1, 0, 1, 3, 20, 192, 197, 194, 65, 3, 20, 192, 197, 194, 5, 82, 196, 196, 1, 0, 1, 3, 21, 192, 197, 2, 65, 3, 21, 192, 197, 2, 21, 0, 197, 67, 1, 0, 1, 3, 21, 192, 197, 163, 65, 3, 21, 192, 197, 163, 22, 0, 197, 3, 1, 0, 1, 3, 22, 192, 197, 99, 65, 3, 22, 192, 197, 99, 156, 90, 196, 196, 1, 0, 1, 3, 22, 192, 197, 194, 65, 3, 22, 192, 197, 194, 5, 90, 196, 196, 1, 0, 1, 3, 23, 192, 197, 2, 65, 3, 23, 192, 197, 2, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 139, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 124, 43, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 196, 30, 0, 0, 196, 30, 0, 0, 84, 32, 0, 0, 84, 32, 0, 0, 148, 34, 0, 0, 148, 34, 0, 0, 88, 39, 0, 0, 88, 39, 0, 0, 96, 43, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 223, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 173, 0, 48, 0, 177, 0, 48, 0, 185, 0, 192, 0, 200, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 206, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 158, 0, 48, 0, 48, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 192, 16, 0, 202, 225, 16, 0, 202, 2, 17, 0, 202, 35, 17, 0, 202, 68, 17, 0, 202, 101, 17, 0, 202, 134, 17, 0, 202, 167, 17, 0, 202, 200, 17, 0, 202, 233, 17, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 166, 1, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 198, 2, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 230, 3, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 166, 5, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 201, 6, 7, 0, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 102, 8, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 102, 10, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 134, 11, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 166, 12, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 134, 14, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 201, 230, 15, 0, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 134, 5, 31, 43, 196, 142, 156, 42, 196, 154, 156, 42, 198, 1, 0, 0, 198, 33, 0, 0, 198, 129, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 134, 5, 49, 51, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 134, 5, 27, 59, 196, 142, 156, 58, 196, 150, 156, 58, 134, 5, 2, 49, 196, 154, 156, 68, 196, 138, 156, 70, 134, 5, 61, 71, 196, 142, 156, 70, 196, 154, 156, 70, 198, 65, 0, 0, 5, 5, 127, 12, 198, 161, 0, 0, 196, 138, 156, 82, 134, 5, 51, 83, 196, 142, 156, 82, 196, 150, 156, 82, 134, 5, 17, 91, 160, 5, 38, 93, 197, 130, 16, 0, 196, 138, 5, 42, 5, 5, 31, 43, 196, 142, 5, 42, 196, 154, 5, 42, 198, 97, 0, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 5, 5, 49, 51, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 5, 5, 27, 59, 196, 142, 5, 58, 196, 150, 5, 58, 5, 5, 2, 49, 196, 154, 5, 68, 196, 138, 5, 70, 5, 5, 61, 71, 196, 142, 5, 70, 196, 154, 5, 70, 193, 3, 111, 93, 5, 5, 125, 12, 198, 193, 0, 0, 196, 138, 5, 82, 5, 5, 51, 83, 196, 142, 5, 82, 196, 150, 5, 82, 5, 5, 17, 91, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 17, 0, 201, 230, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 20, 0, 201, 230, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 23, 0, 201, 38, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 25, 0, 201, 102, 26, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 27, 0, 201, 6, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 30, 0, 201, 166, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 1, 0, 198, 34, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 2, 0, 197, 162, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 32, 0, 201, 70, 34, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 2, 0, 197, 162, 4, 0, 198, 34, 3, 0, 197, 226, 4, 0, 198, 98, 3, 0, 198, 162, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 35, 0, 201, 230, 36, 0, 198, 226, 3, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 38, 0, 201, 230, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 41, 0, 201, 230, 42, 0, 201, 6, 44, 0, 201, 38, 45, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 46, 0, 201, 102, 47, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 48, 0, 201, 166, 49, 0, 201, 198, 50, 0, 201, 230, 51, 0, 201, 6, 53, 0, 201, 38, 54, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 55, 0, 201, 38, 57, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 59, 0, 201, 38, 60, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 61, 0, 201, 102, 62, 0, 201, 134, 63, 0, 201, 6, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 2, 0, 197, 162, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 66, 0, 201, 6, 68, 0, 201, 134, 69, 0, 201, 166, 70, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 3, 0, 197, 98, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 71, 0, 201, 230, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 74, 0, 201, 38, 75, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 5, 31, 43, 5, 162, 0, 0, 5, 5, 31, 43, 5, 162, 0, 0, 193, 4, 111, 93, 5, 162, 0, 0, 134, 5, 49, 51, 5, 162, 0, 0, 5, 5, 49, 51, 5, 162, 0, 0, 134, 5, 27, 59, 5, 162, 0, 0, 5, 5, 27, 59, 5, 162, 0, 0, 134, 5, 51, 83, 5, 162, 0, 0, 5, 5, 51, 83, 5, 162, 0, 0, 134, 5, 61, 71, 5, 190, 0, 0, 5, 5, 61, 71, 5, 190, 0, 0, 193, 3, 111, 93, 5, 190, 0, 0, 134, 5, 51, 83, 5, 190, 0, 0, 5, 5, 51, 83, 5, 190, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 134, 5, 61, 71, 5, 162, 0, 0, 5, 5, 61, 71, 5, 162, 0, 0, 193, 3, 111, 93, 5, 162, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 134, 5, 49, 51, 5, 160, 0, 0, 5, 5, 49, 51, 5, 160, 0, 0, 193, 3, 111, 93, 5, 164, 0, 0, 134, 5, 31, 43, 5, 200, 0, 0, 5, 5, 31, 43, 5, 200, 0, 0, 193, 4, 111, 93, 5, 200, 0, 0, 134, 5, 49, 51, 5, 204, 0, 0, 5, 5, 49, 51, 5, 204, 0, 0, 134, 5, 49, 51, 5, 208, 0, 0, 5, 5, 49, 51, 5, 208, 0, 0, 134, 5, 27, 59, 5, 208, 0, 0, 5, 5, 27, 59, 5, 208, 0, 0, 134, 5, 51, 83, 5, 198, 0, 0, 5, 5, 51, 83, 5, 198, 0, 0, 134, 5, 51, 83, 5, 208, 0, 0, 5, 5, 51, 83, 5, 208, 0, 0, 134, 5, 51, 83, 5, 204, 0, 0, 5, 5, 51, 83, 5, 204, 0, 0, 134, 5, 31, 43, 5, 196, 0, 0, 5, 5, 31, 43, 5, 196, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 134, 5, 49, 51, 5, 196, 0, 0, 5, 5, 49, 51, 5, 196, 0, 0, 134, 5, 27, 59, 5, 196, 0, 0, 5, 5, 27, 59, 5, 196, 0, 0, 134, 5, 61, 71, 5, 196, 0, 0, 5, 5, 61, 71, 5, 196, 0, 0, 193, 3, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 61, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 61, 71, 5, 190, 0, 0, 5, 196, 0, 0, 193, 3, 111, 93, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 51, 83, 5, 196, 0, 0, 5, 5, 51, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 51, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 51, 83, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 17, 91, 5, 196, 0, 0, 5, 5, 17, 91, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 3, 0, 1, 3, 31, 235, 134, 5, 8, 3, 198, 129, 10, 3, 198, 161, 65, 3, 31, 235, 134, 5, 0, 50, 156, 5, 1, 0, 1, 3, 49, 243, 134, 5, 65, 3, 49, 243, 134, 5, 0, 58, 156, 5, 1, 0, 1, 3, 27, 251, 134, 5, 65, 3, 27, 251, 134, 5, 0, 70, 156, 5, 2, 0, 1, 3, 255, 255, 61, 71, 134, 5, 8, 3, 0, 192, 198, 65, 65, 3, 255, 255, 61, 71, 134, 5, 0, 82, 156, 5, 1, 0, 1, 3, 255, 255, 51, 83, 134, 5, 65, 3, 255, 255, 51, 83, 134, 5, 0, 90, 156, 5, 1, 0, 1, 3, 255, 255, 17, 91, 134, 5, 65, 3, 255, 255, 17, 91, 134, 5, 0, 42, 5, 5, 3, 0, 1, 3, 31, 235, 5, 5, 8, 3, 0, 192, 198, 97, 10, 3, 255, 255, 111, 93, 193, 4, 65, 3, 31, 235, 5, 5, 0, 50, 5, 5, 1, 0, 1, 3, 49, 243, 5, 5, 65, 3, 49, 243, 5, 5, 0, 58, 5, 5, 1, 0, 1, 3, 27, 251, 5, 5, 65, 3, 27, 251, 5, 5, 0, 70, 5, 5, 2, 0, 1, 3, 255, 255, 61, 71, 5, 5, 8, 3, 255, 255, 111, 93, 193, 3, 65, 3, 255, 255, 61, 71, 5, 5, 0, 82, 5, 5, 1, 0, 1, 3, 255, 255, 51, 83, 5, 5, 65, 3, 255, 255, 51, 83, 5, 5, 0, 90, 5, 5, 1, 0, 1, 3, 255, 255, 17, 91, 5, 5, 65, 3, 255, 255, 17, 91, 5, 5, 156, 42, 196, 162, 3, 0, 1, 3, 197, 162, 8, 3, 0, 192, 198, 226, 10, 3, 1, 192, 198, 34, 65, 3, 197, 162, 5, 42, 196, 162, 3, 0, 1, 3, 0, 192, 197, 98, 8, 3, 1, 192, 198, 98, 10, 3, 0, 192, 197, 162, 65, 3, 0, 192, 197, 98, 156, 50, 196, 162, 1, 0, 1, 3, 0, 192, 197, 226, 65, 3, 0, 192, 197, 226, 5, 50, 196, 162, 1, 0, 1, 3, 1, 192, 197, 34, 65, 3, 1, 192, 197, 34, 156, 58, 196, 162, 1, 0, 1, 3, 1, 192, 197, 98, 65, 3, 1, 192, 197, 98, 5, 58, 196, 162, 1, 0, 1, 3, 1, 192, 197, 162, 65, 3, 1, 192, 197, 162, 156, 82, 196, 162, 1, 0, 1, 3, 1, 192, 197, 226, 65, 3, 1, 192, 197, 226, 5, 82, 196, 162, 1, 0, 1, 3, 2, 192, 197, 34, 65, 3, 2, 192, 197, 34, 156, 70, 196, 190, 2, 0, 1, 3, 2, 192, 197, 98, 8, 3, 1, 192, 198, 162, 65, 3, 2, 192, 197, 98, 5, 70, 196, 190, 2, 0, 1, 3, 2, 192, 197, 162, 8, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 156, 82, 196, 190, 1, 0, 1, 3, 3, 192, 197, 34, 65, 3, 3, 192, 197, 34, 5, 82, 196, 190, 1, 0, 1, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 98, 156, 70, 196, 162, 2, 0, 1, 3, 3, 192, 197, 226, 8, 3, 2, 192, 198, 162, 65, 3, 3, 192, 197, 226, 5, 70, 196, 162, 2, 0, 1, 3, 4, 192, 197, 34, 8, 3, 4, 192, 197, 98, 65, 3, 4, 192, 197, 34, 156, 50, 196, 160, 1, 0, 1, 3, 5, 192, 197, 34, 65, 3, 5, 192, 197, 34, 5, 50, 196, 160, 1, 0, 1, 3, 5, 192, 197, 98, 65, 3, 5, 192, 197, 98, 156, 42, 196, 200, 3, 0, 1, 3, 5, 192, 197, 226, 8, 3, 4, 192, 198, 34, 10, 3, 4, 192, 198, 98, 65, 3, 5, 192, 197, 226, 5, 42, 196, 200, 3, 0, 1, 3, 6, 192, 197, 34, 8, 3, 4, 192, 198, 162, 10, 3, 6, 192, 197, 98, 65, 3, 6, 192, 197, 34, 156, 50, 196, 204, 1, 0, 1, 3, 6, 192, 197, 162, 65, 3, 6, 192, 197, 162, 5, 50, 196, 204, 1, 0, 1, 3, 6, 192, 197, 226, 65, 3, 6, 192, 197, 226, 156, 50, 196, 208, 1, 0, 1, 3, 7, 192, 197, 34, 65, 3, 7, 192, 197, 34, 5, 50, 196, 208, 1, 0, 1, 3, 7, 192, 197, 98, 65, 3, 7, 192, 197, 98, 156, 58, 196, 208, 1, 0, 1, 3, 7, 192, 197, 162, 65, 3, 7, 192, 197, 162, 5, 58, 196, 208, 1, 0, 1, 3, 7, 192, 197, 226, 65, 3, 7, 192, 197, 226, 156, 82, 196, 198, 1, 0, 1, 3, 8, 192, 197, 34, 65, 3, 8, 192, 197, 34, 5, 82, 196, 198, 1, 0, 1, 3, 8, 192, 197, 98, 65, 3, 8, 192, 197, 98, 156, 82, 196, 208, 1, 0, 1, 3, 8, 192, 197, 162, 65, 3, 8, 192, 197, 162, 5, 82, 196, 208, 1, 0, 1, 3, 8, 192, 197, 226, 65, 3, 8, 192, 197, 226, 156, 82, 196, 204, 1, 0, 1, 3, 9, 192, 197, 34, 65, 3, 9, 192, 197, 34, 5, 82, 196, 204, 1, 0, 1, 3, 9, 192, 197, 98, 65, 3, 9, 192, 197, 98, 156, 42, 196, 196, 3, 0, 1, 3, 9, 192, 197, 162, 8, 3, 4, 192, 198, 226, 10, 3, 5, 192, 198, 34, 65, 3, 9, 192, 197, 162, 5, 42, 196, 196, 3, 0, 1, 3, 9, 192, 197, 226, 8, 3, 5, 192, 198, 98, 10, 3, 10, 192, 197, 34, 65, 3, 9, 192, 197, 226, 156, 50, 196, 196, 1, 0, 1, 3, 10, 192, 197, 98, 65, 3, 10, 192, 197, 98, 5, 50, 196, 196, 1, 0, 1, 3, 10, 192, 197, 162, 65, 3, 10, 192, 197, 162, 156, 58, 196, 196, 1, 0, 1, 3, 10, 192, 197, 226, 65, 3, 10, 192, 197, 226, 5, 58, 196, 196, 1, 0, 1, 3, 11, 192, 197, 34, 65, 3, 11, 192, 197, 34, 156, 70, 196, 196, 2, 0, 1, 3, 11, 192, 197, 98, 8, 3, 5, 192, 198, 162, 65, 3, 11, 192, 197, 98, 5, 70, 196, 196, 2, 0, 1, 3, 11, 192, 197, 162, 8, 3, 11, 192, 197, 226, 65, 3, 11, 192, 197, 162, 12, 0, 197, 35, 2, 0, 1, 3, 12, 192, 197, 131, 8, 3, 5, 192, 198, 227, 65, 3, 12, 192, 197, 131, 12, 0, 197, 227, 2, 0, 1, 3, 13, 192, 197, 67, 8, 3, 13, 192, 197, 163, 65, 3, 13, 192, 197, 67, 156, 82, 196, 196, 1, 0, 1, 3, 14, 192, 197, 2, 65, 3, 14, 192, 197, 2, 5, 82, 196, 196, 1, 0, 1, 3, 14, 192, 197, 66, 65, 3, 14, 192, 197, 66, 14, 0, 197, 131, 1, 0, 1, 3, 14, 192, 197, 227, 65, 3, 14, 192, 197, 227, 15, 0, 197, 67, 1, 0, 1, 3, 15, 192, 197, 163, 65, 3, 15, 192, 197, 163, 156, 90, 196, 196, 1, 0, 1, 3, 16, 192, 197, 2, 65, 3, 16, 192, 197, 2, 5, 90, 196, 196, 1, 0, 1, 3, 16, 192, 197, 66, 65, 3, 16, 192, 197, 66, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 20, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 22, 4, 185, 64, 185, 68, 185, 72, 24, 4, 185, 88, 185, 92, 185, 96, 26, 4, 186, 112, 185, 116, 185, 120, 185, 124, 186, 128, 28, 4, 185, 144, 185, 148, 185, 156, 186, 160, 185, 164, 30, 4, 185, 180, 185, 184, 185, 188, 32, 4, 185, 200, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 34, 4, 168, 64, 168, 68, 168, 72, 36, 4, 168, 88, 168, 92, 168, 96, 38, 4, 168, 112, 168, 116, 168, 120, 168, 124, 168, 128, 40, 4, 168, 144, 168, 148, 168, 156, 168, 160, 168, 164, 42, 4, 168, 180, 168, 184, 168, 188, 44, 4, 168, 200, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 136, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 185, 60, 249, 57, 153, 58, 217, 212, 185, 220, 185, 212, 217, 70, 186, 81, 185, 84, 250, 81, 90, 82, 186, 101, 185, 108, 250, 101, 90, 102, 185, 76, 154, 130, 185, 137, 185, 140, 249, 137, 153, 138, 185, 216, 136, 14, 217, 216, 185, 173, 185, 176, 249, 173, 89, 174, 185, 196, 185, 204, 0, 8, 168, 57, 168, 60, 232, 57, 136, 58, 200, 212, 168, 220, 168, 212, 200, 70, 168, 81, 168, 84, 232, 81, 72, 82, 168, 101, 168, 108, 232, 101, 72, 102, 168, 76, 136, 130, 168, 137, 168, 140, 232, 137, 136, 138, 168, 216, 128, 14, 200, 216, 168, 173, 168, 176, 232, 173, 72, 174, 168, 196, 168, 204, 72, 194, 25, 59, 8, 59, 217, 57, 200, 57, 46, 4, 48, 4, 153, 69, 136, 69, 249, 69, 232, 69, 185, 70, 168, 70, 25, 70, 8, 70, 25, 74, 8, 74, 217, 72, 200, 72, 26, 83, 8, 83, 218, 81, 200, 81, 186, 82, 168, 82, 50, 4, 52, 4, 26, 82, 8, 82, 249, 93, 232, 93, 217, 93, 200, 93, 185, 94, 168, 94, 217, 94, 200, 94, 249, 97, 232, 97, 57, 99, 40, 99, 154, 102, 136, 102, 26, 103, 8, 103, 218, 101, 200, 101, 54, 4, 56, 4, 186, 102, 168, 104, 2, 8, 4, 8, 250, 113, 232, 113, 217, 118, 200, 118, 168, 152, 153, 121, 136, 121, 217, 122, 200, 122, 25, 122, 8, 122, 89, 123, 72, 123, 57, 123, 40, 123, 154, 129, 136, 129, 218, 130, 200, 130, 26, 130, 8, 130, 6, 8, 185, 132, 168, 132, 25, 139, 8, 139, 217, 137, 200, 137, 121, 138, 104, 138, 8, 8, 10, 8, 153, 157, 136, 157, 217, 158, 200, 158, 25, 158, 8, 158, 154, 161, 136, 161, 250, 161, 232, 161, 218, 162, 200, 162, 26, 162, 8, 162, 217, 166, 200, 166, 25, 166, 8, 166, 185, 168, 168, 168, 153, 174, 136, 174, 25, 175, 8, 175, 217, 173, 200, 173, 57, 174, 40, 174, 121, 174, 104, 174, 58, 4, 60, 4, 249, 185, 232, 185, 249, 193, 232, 193, 89, 194, 153, 201, 136, 201, 185, 202, 168, 202, 25, 202, 8, 202, 232, 160, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 12, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 14, 8, 1, 0, 8, 14, 16, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 18, 8, 200, 12, 72, 12, 32, 14, 200, 160, 169, 160, 187, 100, 187, 112, 169, 100, 169, 112, 169, 208, 169, 128, 217, 136, 187, 80, 200, 136, 169, 80, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 80, 255, 5, 186, 100, 255, 5, 185, 136, 255, 5, 185, 172, 255, 5, 185, 192, 255, 5, 168, 56, 255, 5, 168, 80, 255, 5, 168, 100, 255, 5, 168, 136, 255, 5, 168, 172, 255, 5, 168, 192, 255, 5, 249, 58, 255, 5, 232, 58, 255, 5, 250, 82, 255, 5, 232, 82, 255, 5, 250, 102, 255, 5, 232, 102, 255, 5, 249, 174, 255, 5, 232, 174, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 7, 13, 0, 16, 57, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 0, 13, 0, 32, 231, 23, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 236, 23, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 14, 0, 0, 0, 14, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 14, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 104, 0, 101, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 104, 0, 101, 0, 95, 0, 73, 0, 76, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 245, 142, 0, 32, 7, 0, 0, 0, 26, 0, 0, 0, 249, 142, 0, 0, 249, 142, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 219, 22, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 112, 114, 105, 118, 97, 116, 101, 45, 107, 97, 110, 97, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 117, 110, 105, 104, 97, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 61, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 117, 0, 110, 0, 105, 0, 104, 0, 97, 0, 110, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 106, 0, 97, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 107, 0, 97, 0, 110, 0, 97, 0, 93, 0, 0, 0, 239, 223, 39, 20, 91, 0, 115, 0, 116, 0, 114, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 32, 0, 51, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 76, 0, 97, 0, 116, 0, 110, 0, 32, 0, 75, 0, 97, 0, 110, 0, 97, 0, 32, 0, 72, 0, 97, 0, 110, 0, 105, 0, 93, 0, 38, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 65, 48, 60, 0, 60, 0, 60, 0, 65, 48, 124, 0, 252, 48, 61, 0, 66, 48, 124, 0, 252, 48, 61, 0, 75, 48, 124, 0, 252, 48, 61, 0, 149, 48, 124, 0, 252, 48, 61, 0, 76, 48, 124, 0, 252, 48, 61, 0, 85, 48, 124, 0, 252, 48, 61, 0, 86, 48, 124, 0, 252, 48, 61, 0, 95, 48, 124, 0, 252, 48, 61, 0, 96, 48, 124, 0, 252, 48, 61, 0, 106, 48, 124, 0, 252, 48, 61, 0, 111, 48, 124, 0, 252, 48, 61, 0, 112, 48, 124, 0, 252, 48, 61, 0, 113, 48, 124, 0, 252, 48, 61, 0, 126, 48, 124, 0, 252, 48, 61, 0, 131, 48, 124, 0, 252, 48, 61, 0, 132, 48, 124, 0, 252, 48, 61, 0, 137, 48, 124, 0, 252, 48, 61, 0, 142, 48, 124, 0, 252, 48, 61, 0, 143, 48, 124, 0, 252, 48, 60, 0, 60, 0, 60, 0, 60, 0, 161, 48, 124, 0, 252, 48, 61, 0, 103, 255, 124, 0, 252, 48, 61, 0, 162, 48, 124, 0, 252, 48, 61, 0, 113, 255, 124, 0, 252, 48, 61, 0, 171, 48, 124, 0, 252, 48, 61, 0, 118, 255, 124, 0, 252, 48, 61, 0, 172, 48, 124, 0, 252, 48, 61, 0, 181, 48, 124, 0, 252, 48, 61, 0, 123, 255, 124, 0, 252, 48, 61, 0, 182, 48, 124, 0, 252, 48, 61, 0, 191, 48, 124, 0, 252, 48, 61, 0, 128, 255, 124, 0, 252, 48, 61, 0, 192, 48, 124, 0, 252, 48, 61, 0, 202, 48, 124, 0, 252, 48, 61, 0, 133, 255, 124, 0, 252, 48, 61, 0, 207, 48, 124, 0, 252, 48, 61, 0, 138, 255, 124, 0, 252, 48, 61, 0, 245, 49, 124, 0, 252, 48, 61, 0, 208, 48, 124, 0, 252, 48, 61, 0, 209, 48, 124, 0, 252, 48, 61, 0, 222, 48, 124, 0, 252, 48, 61, 0, 143, 255, 124, 0, 252, 48, 61, 0, 227, 48, 124, 0, 252, 48, 61, 0, 108, 255, 124, 0, 252, 48, 61, 0, 228, 48, 124, 0, 252, 48, 61, 0, 148, 255, 124, 0, 252, 48, 61, 0, 233, 48, 124, 0, 252, 48, 61, 0, 151, 255, 124, 0, 252, 48, 61, 0, 251, 49, 124, 0, 252, 48, 61, 0, 238, 48, 124, 0, 252, 48, 61, 0, 239, 48, 124, 0, 252, 48, 61, 0, 156, 255, 124, 0, 252, 48, 61, 0, 245, 48, 124, 0, 252, 48, 61, 0, 247, 48, 124, 0, 252, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 67, 48, 60, 0, 60, 0, 60, 0, 67, 48, 124, 0, 252, 48, 61, 0, 68, 48, 124, 0, 252, 48, 61, 0, 77, 48, 124, 0, 252, 48, 61, 0, 78, 48, 124, 0, 252, 48, 61, 0, 87, 48, 124, 0, 252, 48, 61, 0, 88, 48, 124, 0, 252, 48, 61, 0, 97, 48, 124, 0, 252, 48, 61, 0, 98, 48, 124, 0, 252, 48, 61, 0, 107, 48, 124, 0, 252, 48, 61, 0, 114, 48, 124, 0, 252, 48, 61, 0, 115, 48, 124, 0, 252, 48, 61, 0, 116, 48, 124, 0, 252, 48, 61, 0, 127, 48, 124, 0, 252, 48, 61, 0, 138, 48, 124, 0, 252, 48, 61, 0, 144, 48, 124, 0, 252, 48, 60, 0, 60, 0, 60, 0, 60, 0, 163, 48, 124, 0, 252, 48, 61, 0, 104, 255, 124, 0, 252, 48, 61, 0, 164, 48, 124, 0, 252, 48, 61, 0, 114, 255, 124, 0, 252, 48, 61, 0, 173, 48, 124, 0, 252, 48, 61, 0, 119, 255, 124, 0, 252, 48, 61, 0, 174, 48, 124, 0, 252, 48, 61, 0, 183, 48, 124, 0, 252, 48, 61, 0, 124, 255, 124, 0, 252, 48, 61, 0, 241, 49, 124, 0, 252, 48, 61, 0, 184, 48, 124, 0, 252, 48, 61, 0, 193, 48, 124, 0, 252, 48, 61, 0, 129, 255, 124, 0, 252, 48, 61, 0, 194, 48, 124, 0, 252, 48, 61, 0, 203, 48, 124, 0, 252, 48, 61, 0, 134, 255, 124, 0, 252, 48, 61, 0, 210, 48, 124, 0, 252, 48, 61, 0, 139, 255, 124, 0, 252, 48, 61, 0, 246, 49, 124, 0, 252, 48, 61, 0, 211, 48, 124, 0, 252, 48, 61, 0, 212, 48, 124, 0, 252, 48, 61, 0, 223, 48, 124, 0, 252, 48, 61, 0, 144, 255, 124, 0, 252, 48, 61, 0, 234, 48, 124, 0, 252, 48, 61, 0, 152, 255, 124, 0, 252, 48, 61, 0, 252, 49, 124, 0, 252, 48, 61, 0, 240, 48, 124, 0, 252, 48, 61, 0, 248, 48, 124, 0, 252, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 69, 48, 60, 0, 60, 0, 60, 0, 69, 48, 124, 0, 252, 48, 61, 0, 70, 48, 124, 0, 252, 48, 61, 0, 79, 48, 124, 0, 252, 48, 61, 0, 80, 48, 124, 0, 252, 48, 61, 0, 89, 48, 124, 0, 252, 48, 61, 0, 90, 48, 124, 0, 252, 48, 61, 0, 99, 48, 124, 0, 252, 48, 61, 0, 100, 48, 124, 0, 252, 48, 61, 0, 101, 48, 124, 0, 252, 48, 61, 0, 108, 48, 124, 0, 252, 48, 61, 0, 117, 48, 124, 0, 252, 48, 61, 0, 118, 48, 124, 0, 252, 48, 61, 0, 119, 48, 124, 0, 252, 48, 61, 0, 128, 48, 124, 0, 252, 48, 61, 0, 133, 48, 124, 0, 252, 48, 61, 0, 134, 48, 124, 0, 252, 48, 61, 0, 139, 48, 124, 0, 252, 48, 61, 0, 148, 48, 124, 0, 252, 48, 60, 0, 60, 0, 60, 0, 60, 0, 165, 48, 124, 0, 252, 48, 61, 0, 105, 255, 124, 0, 252, 48, 61, 0, 166, 48, 124, 0, 252, 48, 61, 0, 115, 255, 124, 0, 252, 48, 61, 0, 175, 48, 124, 0, 252, 48, 61, 0, 120, 255, 124, 0, 252, 48, 61, 0, 240, 49, 124, 0, 252, 48, 61, 0, 176, 48, 124, 0, 252, 48, 61, 0, 185, 48, 124, 0, 252, 48, 61, 0, 125, 255, 124, 0, 252, 48, 61, 0, 242, 49, 124, 0, 252, 48, 61, 0, 186, 48, 124, 0, 252, 48, 61, 0, 195, 48, 124, 0, 252, 48, 61, 0, 111, 255, 124, 0, 252, 48, 61, 0, 196, 48, 124, 0, 252, 48, 61, 0, 130, 255, 124, 0, 252, 48, 61, 0, 197, 48, 124, 0, 252, 48, 61, 0, 204, 48, 124, 0, 252, 48, 61, 0, 135, 255, 124, 0, 252, 48, 61, 0, 244, 49, 124, 0, 252, 48, 61, 0, 213, 48, 124, 0, 252, 48, 61, 0, 140, 255, 124, 0, 252, 48, 61, 0, 247, 49, 124, 0, 252, 48, 61, 0, 214, 48, 124, 0, 252, 48, 61, 0, 215, 48, 124, 0, 252, 48, 61, 0, 224, 48, 124, 0, 252, 48, 61, 0, 145, 255, 124, 0, 252, 48, 61, 0, 250, 49, 124, 0, 252, 48, 61, 0, 229, 48, 124, 0, 252, 48, 61, 0, 109, 255, 124, 0, 252, 48, 61, 0, 230, 48, 124, 0, 252, 48, 61, 0, 149, 255, 124, 0, 252, 48, 61, 0, 235, 48, 124, 0, 252, 48, 61, 0, 153, 255, 124, 0, 252, 48, 61, 0, 253, 49, 124, 0, 252, 48, 61, 0, 244, 48, 124, 0, 252, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 71, 48, 60, 0, 60, 0, 60, 0, 71, 48, 124, 0, 252, 48, 61, 0, 72, 48, 124, 0, 252, 48, 61, 0, 81, 48, 124, 0, 252, 48, 61, 0, 150, 48, 124, 0, 252, 48, 61, 0, 82, 48, 124, 0, 252, 48, 61, 0, 91, 48, 124, 0, 252, 48, 61, 0, 92, 48, 124, 0, 252, 48, 61, 0, 102, 48, 124, 0, 252, 48, 61, 0, 103, 48, 124, 0, 252, 48, 61, 0, 109, 48, 124, 0, 252, 48, 61, 0, 120, 48, 124, 0, 252, 48, 61, 0, 121, 48, 124, 0, 252, 48, 61, 0, 122, 48, 124, 0, 252, 48, 61, 0, 129, 48, 124, 0, 252, 48, 61, 0, 140, 48, 124, 0, 252, 48, 61, 0, 145, 48, 124, 0, 252, 48, 60, 0, 60, 0, 60, 0, 60, 0, 167, 48, 124, 0, 252, 48, 61, 0, 106, 255, 124, 0, 252, 48, 61, 0, 168, 48, 124, 0, 252, 48, 61, 0, 116, 255, 124, 0, 252, 48, 61, 0, 177, 48, 124, 0, 252, 48, 61, 0, 121, 255, 124, 0, 252, 48, 61, 0, 178, 48, 124, 0, 252, 48, 61, 0, 187, 48, 124, 0, 252, 48, 61, 0, 126, 255, 124, 0, 252, 48, 61, 0, 188, 48, 124, 0, 252, 48, 61, 0, 198, 48, 124, 0, 252, 48, 61, 0, 131, 255, 124, 0, 252, 48, 61, 0, 199, 48, 124, 0, 252, 48, 61, 0, 205, 48, 124, 0, 252, 48, 61, 0, 136, 255, 124, 0, 252, 48, 61, 0, 216, 48, 124, 0, 252, 48, 61, 0, 141, 255, 124, 0, 252, 48, 61, 0, 248, 49, 124, 0, 252, 48, 61, 0, 217, 48, 124, 0, 252, 48, 61, 0, 218, 48, 124, 0, 252, 48, 61, 0, 225, 48, 124, 0, 252, 48, 61, 0, 146, 255, 124, 0, 252, 48, 61, 0, 236, 48, 124, 0, 252, 48, 61, 0, 154, 255, 124, 0, 252, 48, 61, 0, 254, 49, 124, 0, 252, 48, 61, 0, 241, 48, 124, 0, 252, 48, 61, 0, 246, 48, 124, 0, 252, 48, 61, 0, 249, 48, 124, 0, 252, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 73, 48, 60, 0, 60, 0, 60, 0, 73, 48, 124, 0, 252, 48, 61, 0, 74, 48, 124, 0, 252, 48, 61, 0, 83, 48, 124, 0, 252, 48, 61, 0, 84, 48, 124, 0, 252, 48, 61, 0, 93, 48, 124, 0, 252, 48, 61, 0, 94, 48, 124, 0, 252, 48, 61, 0, 104, 48, 124, 0, 252, 48, 61, 0, 105, 48, 124, 0, 252, 48, 61, 0, 110, 48, 124, 0, 252, 48, 61, 0, 123, 48, 124, 0, 252, 48, 61, 0, 124, 48, 124, 0, 252, 48, 61, 0, 125, 48, 124, 0, 252, 48, 61, 0, 130, 48, 124, 0, 252, 48, 61, 0, 135, 48, 124, 0, 252, 48, 61, 0, 136, 48, 124, 0, 252, 48, 61, 0, 141, 48, 124, 0, 252, 48, 61, 0, 146, 48, 124, 0, 252, 48, 60, 0, 60, 0, 60, 0, 60, 0, 169, 48, 124, 0, 252, 48, 61, 0, 107, 255, 124, 0, 252, 48, 61, 0, 170, 48, 124, 0, 252, 48, 61, 0, 117, 255, 124, 0, 252, 48, 61, 0, 179, 48, 124, 0, 252, 48, 61, 0, 122, 255, 124, 0, 252, 48, 61, 0, 180, 48, 124, 0, 252, 48, 61, 0, 189, 48, 124, 0, 252, 48, 61, 0, 127, 255, 124, 0, 252, 48, 61, 0, 190, 48, 124, 0, 252, 48, 61, 0, 200, 48, 124, 0, 252, 48, 61, 0, 132, 255, 124, 0, 252, 48, 61, 0, 243, 49, 124, 0, 252, 48, 61, 0, 201, 48, 124, 0, 252, 48, 61, 0, 206, 48, 124, 0, 252, 48, 61, 0, 137, 255, 124, 0, 252, 48, 61, 0, 219, 48, 124, 0, 252, 48, 61, 0, 142, 255, 124, 0, 252, 48, 61, 0, 249, 49, 124, 0, 252, 48, 61, 0, 220, 48, 124, 0, 252, 48, 61, 0, 221, 48, 124, 0, 252, 48, 61, 0, 226, 48, 124, 0, 252, 48, 61, 0, 147, 255, 124, 0, 252, 48, 61, 0, 231, 48, 124, 0, 252, 48, 61, 0, 110, 255, 124, 0, 252, 48, 61, 0, 232, 48, 124, 0, 252, 48, 61, 0, 150, 255, 124, 0, 252, 48, 61, 0, 237, 48, 124, 0, 252, 48, 61, 0, 155, 255, 124, 0, 252, 48, 61, 0, 255, 49, 124, 0, 252, 48, 61, 0, 242, 48, 124, 0, 252, 48, 61, 0, 102, 255, 124, 0, 252, 48, 61, 0, 250, 48, 124, 0, 252, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 66, 48, 60, 0, 60, 0, 60, 0, 66, 48, 124, 0, 157, 48, 61, 0, 65, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 162, 48, 124, 0, 253, 48, 61, 0, 113, 255, 124, 0, 253, 48, 61, 0, 161, 48, 124, 0, 253, 48, 61, 0, 103, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 68, 48, 60, 0, 60, 0, 60, 0, 68, 48, 124, 0, 157, 48, 61, 0, 67, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 164, 48, 124, 0, 253, 48, 61, 0, 114, 255, 124, 0, 253, 48, 61, 0, 163, 48, 124, 0, 253, 48, 61, 0, 104, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 70, 48, 60, 0, 60, 0, 60, 0, 70, 48, 124, 0, 157, 48, 61, 0, 69, 48, 124, 0, 157, 48, 61, 0, 148, 48, 124, 0, 157, 48, 61, 0, 70, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 69, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 148, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 166, 48, 124, 0, 253, 48, 61, 0, 115, 255, 124, 0, 253, 48, 61, 0, 165, 48, 124, 0, 253, 48, 61, 0, 105, 255, 124, 0, 253, 48, 61, 0, 244, 48, 124, 0, 253, 48, 61, 0, 166, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 115, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 165, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 105, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 244, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 72, 48, 60, 0, 60, 0, 60, 0, 72, 48, 124, 0, 157, 48, 61, 0, 71, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 168, 48, 124, 0, 253, 48, 61, 0, 116, 255, 124, 0, 253, 48, 61, 0, 167, 48, 124, 0, 253, 48, 61, 0, 106, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 74, 48, 60, 0, 60, 0, 60, 0, 74, 48, 124, 0, 157, 48, 61, 0, 73, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 170, 48, 124, 0, 253, 48, 61, 0, 117, 255, 124, 0, 253, 48, 61, 0, 169, 48, 124, 0, 253, 48, 61, 0, 107, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 75, 48, 60, 0, 60, 0, 60, 0, 75, 48, 124, 0, 157, 48, 61, 0, 149, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 171, 48, 124, 0, 253, 48, 61, 0, 118, 255, 124, 0, 253, 48, 61, 0, 245, 48, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 76, 48, 60, 0, 60, 0, 60, 0, 76, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 172, 48, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 77, 48, 60, 0, 60, 0, 60, 0, 77, 48, 124, 0, 157, 48, 61, 0, 78, 48, 124, 0, 157, 48, 61, 0, 77, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 78, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 173, 48, 124, 0, 253, 48, 61, 0, 119, 255, 124, 0, 253, 48, 61, 0, 174, 48, 124, 0, 253, 48, 61, 0, 173, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 119, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 174, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 79, 48, 60, 0, 60, 0, 60, 0, 79, 48, 124, 0, 157, 48, 61, 0, 80, 48, 124, 0, 157, 48, 61, 0, 79, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 80, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 175, 48, 124, 0, 253, 48, 61, 0, 120, 255, 124, 0, 253, 48, 61, 0, 240, 49, 124, 0, 253, 48, 61, 0, 176, 48, 124, 0, 253, 48, 61, 0, 175, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 120, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 240, 49, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 176, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 81, 48, 60, 0, 60, 0, 60, 0, 81, 48, 124, 0, 157, 48, 61, 0, 150, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 177, 48, 124, 0, 253, 48, 61, 0, 121, 255, 124, 0, 253, 48, 61, 0, 246, 48, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 82, 48, 60, 0, 60, 0, 60, 0, 82, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 178, 48, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 83, 48, 60, 0, 60, 0, 60, 0, 83, 48, 124, 0, 157, 48, 61, 0, 84, 48, 124, 0, 157, 48, 61, 0, 83, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 84, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 179, 48, 124, 0, 253, 48, 61, 0, 122, 255, 124, 0, 253, 48, 61, 0, 180, 48, 124, 0, 253, 48, 61, 0, 179, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 122, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 180, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 85, 48, 60, 0, 60, 0, 60, 0, 85, 48, 124, 0, 157, 48, 61, 0, 86, 48, 124, 0, 157, 48, 61, 0, 85, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 86, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 181, 48, 124, 0, 253, 48, 61, 0, 123, 255, 124, 0, 253, 48, 61, 0, 182, 48, 124, 0, 253, 48, 61, 0, 181, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 123, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 182, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 87, 48, 60, 0, 60, 0, 60, 0, 87, 48, 124, 0, 157, 48, 61, 0, 88, 48, 124, 0, 157, 48, 61, 0, 87, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 88, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 183, 48, 124, 0, 253, 48, 61, 0, 124, 255, 124, 0, 253, 48, 61, 0, 241, 49, 124, 0, 253, 48, 61, 0, 184, 48, 124, 0, 253, 48, 61, 0, 183, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 124, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 241, 49, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 184, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 89, 48, 60, 0, 60, 0, 60, 0, 89, 48, 124, 0, 157, 48, 61, 0, 90, 48, 124, 0, 157, 48, 61, 0, 89, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 90, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 185, 48, 124, 0, 253, 48, 61, 0, 125, 255, 124, 0, 253, 48, 61, 0, 242, 49, 124, 0, 253, 48, 61, 0, 186, 48, 124, 0, 253, 48, 61, 0, 185, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 125, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 242, 49, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 186, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 91, 48, 60, 0, 60, 0, 60, 0, 91, 48, 124, 0, 157, 48, 61, 0, 92, 48, 124, 0, 157, 48, 61, 0, 91, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 92, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 187, 48, 124, 0, 253, 48, 61, 0, 126, 255, 124, 0, 253, 48, 61, 0, 188, 48, 124, 0, 253, 48, 61, 0, 187, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 126, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 188, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 93, 48, 60, 0, 60, 0, 60, 0, 93, 48, 124, 0, 157, 48, 61, 0, 94, 48, 124, 0, 157, 48, 61, 0, 93, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 94, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 189, 48, 124, 0, 253, 48, 61, 0, 127, 255, 124, 0, 253, 48, 61, 0, 190, 48, 124, 0, 253, 48, 61, 0, 189, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 127, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 190, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 95, 48, 60, 0, 60, 0, 60, 0, 95, 48, 124, 0, 157, 48, 61, 0, 96, 48, 124, 0, 157, 48, 61, 0, 95, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 96, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 191, 48, 124, 0, 253, 48, 61, 0, 128, 255, 124, 0, 253, 48, 61, 0, 192, 48, 124, 0, 253, 48, 61, 0, 191, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 128, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 192, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 97, 48, 60, 0, 60, 0, 60, 0, 97, 48, 124, 0, 157, 48, 61, 0, 98, 48, 124, 0, 157, 48, 61, 0, 97, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 98, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 193, 48, 124, 0, 253, 48, 61, 0, 129, 255, 124, 0, 253, 48, 61, 0, 194, 48, 124, 0, 253, 48, 61, 0, 193, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 129, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 194, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 100, 48, 60, 0, 60, 0, 60, 0, 100, 48, 124, 0, 157, 48, 61, 0, 99, 48, 124, 0, 157, 48, 61, 0, 101, 48, 124, 0, 157, 48, 61, 0, 100, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 101, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 100, 48, 124, 0, 157, 48, 61, 0, 99, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 100, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 196, 48, 124, 0, 253, 48, 61, 0, 130, 255, 124, 0, 253, 48, 61, 0, 195, 48, 124, 0, 253, 48, 61, 0, 111, 255, 124, 0, 253, 48, 61, 0, 197, 48, 124, 0, 253, 48, 61, 0, 196, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 130, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 197, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 196, 48, 124, 0, 253, 48, 61, 0, 130, 255, 124, 0, 253, 48, 61, 0, 195, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 111, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 196, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 130, 255, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 102, 48, 60, 0, 60, 0, 60, 0, 102, 48, 124, 0, 157, 48, 61, 0, 103, 48, 124, 0, 157, 48, 61, 0, 102, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 103, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 198, 48, 124, 0, 253, 48, 61, 0, 131, 255, 124, 0, 253, 48, 61, 0, 199, 48, 124, 0, 253, 48, 61, 0, 198, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 131, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 199, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 104, 48, 60, 0, 60, 0, 60, 0, 104, 48, 124, 0, 157, 48, 61, 0, 105, 48, 124, 0, 157, 48, 61, 0, 104, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 105, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 200, 48, 124, 0, 253, 48, 61, 0, 132, 255, 124, 0, 253, 48, 61, 0, 243, 49, 124, 0, 253, 48, 61, 0, 201, 48, 124, 0, 253, 48, 61, 0, 200, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 132, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 243, 49, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 201, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 106, 48, 60, 0, 60, 0, 60, 0, 106, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 202, 48, 124, 0, 253, 48, 61, 0, 133, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 107, 48, 60, 0, 60, 0, 60, 0, 107, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 203, 48, 124, 0, 253, 48, 61, 0, 134, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 108, 48, 60, 0, 60, 0, 60, 0, 108, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 204, 48, 124, 0, 253, 48, 61, 0, 135, 255, 124, 0, 253, 48, 61, 0, 244, 49, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 109, 48, 60, 0, 60, 0, 60, 0, 109, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 205, 48, 124, 0, 253, 48, 61, 0, 136, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 110, 48, 60, 0, 60, 0, 60, 0, 110, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 206, 48, 124, 0, 253, 48, 61, 0, 137, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 111, 48, 60, 0, 60, 0, 60, 0, 111, 48, 124, 0, 157, 48, 61, 0, 112, 48, 124, 0, 157, 48, 61, 0, 111, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 112, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 113, 48, 124, 0, 157, 48, 61, 0, 113, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 207, 48, 124, 0, 253, 48, 61, 0, 138, 255, 124, 0, 253, 48, 61, 0, 245, 49, 124, 0, 253, 48, 61, 0, 208, 48, 124, 0, 253, 48, 61, 0, 207, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 138, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 245, 49, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 208, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 209, 48, 124, 0, 253, 48, 61, 0, 209, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 114, 48, 60, 0, 60, 0, 60, 0, 114, 48, 124, 0, 157, 48, 61, 0, 115, 48, 124, 0, 157, 48, 61, 0, 114, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 115, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 116, 48, 124, 0, 157, 48, 61, 0, 116, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 210, 48, 124, 0, 253, 48, 61, 0, 139, 255, 124, 0, 253, 48, 61, 0, 246, 49, 124, 0, 253, 48, 61, 0, 211, 48, 124, 0, 253, 48, 61, 0, 210, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 139, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 246, 49, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 211, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 212, 48, 124, 0, 253, 48, 61, 0, 212, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 117, 48, 60, 0, 60, 0, 60, 0, 117, 48, 124, 0, 157, 48, 61, 0, 118, 48, 124, 0, 157, 48, 61, 0, 117, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 118, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 119, 48, 124, 0, 157, 48, 61, 0, 119, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 213, 48, 124, 0, 253, 48, 61, 0, 140, 255, 124, 0, 253, 48, 61, 0, 247, 49, 124, 0, 253, 48, 61, 0, 214, 48, 124, 0, 253, 48, 61, 0, 213, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 140, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 247, 49, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 214, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 215, 48, 124, 0, 253, 48, 61, 0, 215, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 120, 48, 60, 0, 60, 0, 60, 0, 120, 48, 124, 0, 157, 48, 61, 0, 121, 48, 124, 0, 157, 48, 61, 0, 120, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 121, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 122, 48, 124, 0, 157, 48, 61, 0, 122, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 216, 48, 124, 0, 253, 48, 61, 0, 141, 255, 124, 0, 253, 48, 61, 0, 248, 49, 124, 0, 253, 48, 61, 0, 217, 48, 124, 0, 253, 48, 61, 0, 216, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 141, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 248, 49, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 217, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 218, 48, 124, 0, 253, 48, 61, 0, 218, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 123, 48, 60, 0, 60, 0, 60, 0, 123, 48, 124, 0, 157, 48, 61, 0, 124, 48, 124, 0, 157, 48, 61, 0, 123, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 124, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 125, 48, 124, 0, 157, 48, 61, 0, 125, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 219, 48, 124, 0, 253, 48, 61, 0, 142, 255, 124, 0, 253, 48, 61, 0, 249, 49, 124, 0, 253, 48, 61, 0, 220, 48, 124, 0, 253, 48, 61, 0, 219, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 142, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 249, 49, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 220, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 221, 48, 124, 0, 253, 48, 61, 0, 221, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 126, 48, 60, 0, 60, 0, 60, 0, 126, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 222, 48, 124, 0, 253, 48, 61, 0, 143, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 127, 48, 60, 0, 60, 0, 60, 0, 127, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 223, 48, 124, 0, 253, 48, 61, 0, 144, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 128, 48, 60, 0, 60, 0, 60, 0, 128, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 224, 48, 124, 0, 253, 48, 61, 0, 145, 255, 124, 0, 253, 48, 61, 0, 250, 49, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 129, 48, 60, 0, 60, 0, 60, 0, 129, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 225, 48, 124, 0, 253, 48, 61, 0, 146, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 130, 48, 60, 0, 60, 0, 60, 0, 130, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 226, 48, 124, 0, 253, 48, 61, 0, 147, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 132, 48, 60, 0, 60, 0, 60, 0, 132, 48, 124, 0, 157, 48, 61, 0, 131, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 228, 48, 124, 0, 253, 48, 61, 0, 148, 255, 124, 0, 253, 48, 61, 0, 227, 48, 124, 0, 253, 48, 61, 0, 108, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 134, 48, 60, 0, 60, 0, 60, 0, 134, 48, 124, 0, 157, 48, 61, 0, 133, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 230, 48, 124, 0, 253, 48, 61, 0, 149, 255, 124, 0, 253, 48, 61, 0, 229, 48, 124, 0, 253, 48, 61, 0, 109, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 136, 48, 60, 0, 60, 0, 60, 0, 136, 48, 124, 0, 157, 48, 61, 0, 135, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 232, 48, 124, 0, 253, 48, 61, 0, 150, 255, 124, 0, 253, 48, 61, 0, 231, 48, 124, 0, 253, 48, 61, 0, 110, 255, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 137, 48, 60, 0, 60, 0, 60, 0, 137, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 233, 48, 124, 0, 253, 48, 61, 0, 151, 255, 124, 0, 253, 48, 61, 0, 251, 49, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 138, 48, 60, 0, 60, 0, 60, 0, 138, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 234, 48, 124, 0, 253, 48, 61, 0, 152, 255, 124, 0, 253, 48, 61, 0, 252, 49, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 139, 48, 60, 0, 60, 0, 60, 0, 139, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 235, 48, 124, 0, 253, 48, 61, 0, 153, 255, 124, 0, 253, 48, 61, 0, 253, 49, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 140, 48, 60, 0, 60, 0, 60, 0, 140, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 236, 48, 124, 0, 253, 48, 61, 0, 154, 255, 124, 0, 253, 48, 61, 0, 254, 49, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 141, 48, 60, 0, 60, 0, 60, 0, 141, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 237, 48, 124, 0, 253, 48, 61, 0, 155, 255, 124, 0, 253, 48, 61, 0, 255, 49, 124, 0, 253, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 143, 48, 60, 0, 60, 0, 60, 0, 143, 48, 124, 0, 157, 48, 61, 0, 142, 48, 124, 0, 157, 48, 61, 0, 143, 48, 124, 0, 158, 48, 47, 0, 153, 48, 61, 0, 142, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 239, 48, 124, 0, 253, 48, 61, 0, 156, 255, 124, 0, 253, 48, 61, 0, 238, 48, 124, 0, 253, 48, 61, 0, 247, 48, 124, 0, 253, 48, 61, 0, 239, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 156, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 247, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 238, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 144, 48, 60, 0, 60, 0, 60, 0, 144, 48, 124, 0, 157, 48, 61, 0, 144, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 240, 48, 124, 0, 253, 48, 61, 0, 248, 48, 124, 0, 253, 48, 61, 0, 240, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 248, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 145, 48, 60, 0, 60, 0, 60, 0, 145, 48, 124, 0, 157, 48, 61, 0, 145, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 241, 48, 124, 0, 253, 48, 61, 0, 249, 48, 124, 0, 253, 48, 61, 0, 241, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 249, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 146, 48, 60, 0, 60, 0, 60, 0, 146, 48, 124, 0, 157, 48, 61, 0, 146, 48, 124, 0, 158, 48, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 60, 0, 242, 48, 124, 0, 253, 48, 61, 0, 102, 255, 124, 0, 253, 48, 61, 0, 250, 48, 124, 0, 253, 48, 61, 0, 242, 48, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 102, 255, 124, 0, 254, 48, 47, 0, 153, 48, 61, 0, 250, 48, 124, 0, 254, 48, 47, 0, 153, 48, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 51, 0, 93, 0, 147, 48, 60, 0, 60, 0, 60, 0, 147, 48, 124, 0, 157, 48, 60, 0, 60, 0, 60, 0, 60, 0, 243, 48, 124, 0, 253, 48, 61, 0, 157, 255, 124, 0, 253, 48, 38, 0, 65, 48, 60, 0, 60, 0, 60, 0, 60, 0, 161, 48, 61, 0, 103, 255, 38, 0, 66, 48, 60, 0, 60, 0, 60, 0, 60, 0, 162, 48, 61, 0, 113, 255, 38, 0, 67, 48, 60, 0, 60, 0, 60, 0, 60, 0, 163, 48, 61, 0, 104, 255, 38, 0, 68, 48, 60, 0, 60, 0, 60, 0, 60, 0, 164, 48, 61, 0, 114, 255, 38, 0, 69, 48, 60, 0, 60, 0, 60, 0, 60, 0, 165, 48, 61, 0, 105, 255, 38, 0, 70, 48, 60, 0, 60, 0, 60, 0, 60, 0, 166, 48, 61, 0, 115, 255, 38, 0, 71, 48, 60, 0, 60, 0, 60, 0, 60, 0, 167, 48, 61, 0, 106, 255, 38, 0, 72, 48, 60, 0, 60, 0, 60, 0, 60, 0, 168, 48, 61, 0, 116, 255, 38, 0, 73, 48, 60, 0, 60, 0, 60, 0, 60, 0, 169, 48, 61, 0, 107, 255, 38, 0, 74, 48, 60, 0, 60, 0, 60, 0, 60, 0, 170, 48, 61, 0, 117, 255, 38, 0, 75, 48, 60, 0, 60, 0, 60, 0, 60, 0, 171, 48, 61, 0, 118, 255, 38, 0, 77, 48, 60, 0, 60, 0, 60, 0, 60, 0, 173, 48, 61, 0, 119, 255, 38, 0, 79, 48, 60, 0, 60, 0, 60, 0, 60, 0, 175, 48, 61, 0, 120, 255, 38, 0, 81, 48, 60, 0, 60, 0, 60, 0, 60, 0, 177, 48, 61, 0, 121, 255, 38, 0, 83, 48, 60, 0, 60, 0, 60, 0, 60, 0, 179, 48, 61, 0, 122, 255, 38, 0, 85, 48, 60, 0, 60, 0, 60, 0, 60, 0, 181, 48, 61, 0, 123, 255, 38, 0, 87, 48, 60, 0, 60, 0, 60, 0, 60, 0, 183, 48, 61, 0, 124, 255, 38, 0, 89, 48, 60, 0, 60, 0, 60, 0, 60, 0, 185, 48, 61, 0, 125, 255, 38, 0, 91, 48, 60, 0, 60, 0, 60, 0, 60, 0, 187, 48, 61, 0, 126, 255, 38, 0, 93, 48, 60, 0, 60, 0, 60, 0, 60, 0, 189, 48, 61, 0, 127, 255, 38, 0, 95, 48, 60, 0, 60, 0, 60, 0, 60, 0, 191, 48, 61, 0, 128, 255, 38, 0, 97, 48, 60, 0, 60, 0, 60, 0, 60, 0, 193, 48, 61, 0, 129, 255, 38, 0, 99, 48, 60, 0, 60, 0, 60, 0, 60, 0, 195, 48, 61, 0, 111, 255, 38, 0, 100, 48, 60, 0, 60, 0, 60, 0, 60, 0, 196, 48, 61, 0, 130, 255, 38, 0, 102, 48, 60, 0, 60, 0, 60, 0, 60, 0, 198, 48, 61, 0, 131, 255, 38, 0, 104, 48, 60, 0, 60, 0, 60, 0, 60, 0, 200, 48, 61, 0, 132, 255, 38, 0, 106, 48, 60, 0, 60, 0, 60, 0, 60, 0, 202, 48, 61, 0, 133, 255, 38, 0, 107, 48, 60, 0, 60, 0, 60, 0, 60, 0, 203, 48, 61, 0, 134, 255, 38, 0, 108, 48, 60, 0, 60, 0, 60, 0, 60, 0, 204, 48, 61, 0, 135, 255, 38, 0, 109, 48, 60, 0, 60, 0, 60, 0, 60, 0, 205, 48, 61, 0, 136, 255, 38, 0, 110, 48, 60, 0, 60, 0, 60, 0, 60, 0, 206, 48, 61, 0, 137, 255, 38, 0, 111, 48, 60, 0, 60, 0, 60, 0, 60, 0, 207, 48, 61, 0, 138, 255, 38, 0, 114, 48, 60, 0, 60, 0, 60, 0, 60, 0, 210, 48, 61, 0, 139, 255, 38, 0, 117, 48, 60, 0, 60, 0, 60, 0, 60, 0, 213, 48, 61, 0, 140, 255, 38, 0, 120, 48, 60, 0, 60, 0, 60, 0, 60, 0, 216, 48, 61, 0, 141, 255, 38, 0, 123, 48, 60, 0, 60, 0, 60, 0, 60, 0, 219, 48, 61, 0, 142, 255, 38, 0, 126, 48, 60, 0, 60, 0, 60, 0, 60, 0, 222, 48, 61, 0, 143, 255, 38, 0, 127, 48, 60, 0, 60, 0, 60, 0, 60, 0, 223, 48, 61, 0, 144, 255, 38, 0, 128, 48, 60, 0, 60, 0, 60, 0, 60, 0, 224, 48, 61, 0, 145, 255, 38, 0, 129, 48, 60, 0, 60, 0, 60, 0, 60, 0, 225, 48, 61, 0, 146, 255, 38, 0, 130, 48, 60, 0, 60, 0, 60, 0, 60, 0, 226, 48, 61, 0, 147, 255, 38, 0, 131, 48, 60, 0, 60, 0, 60, 0, 60, 0, 227, 48, 61, 0, 108, 255, 38, 0, 132, 48, 60, 0, 60, 0, 60, 0, 60, 0, 228, 48, 61, 0, 148, 255, 38, 0, 133, 48, 60, 0, 60, 0, 60, 0, 60, 0, 229, 48, 61, 0, 109, 255, 38, 0, 134, 48, 60, 0, 60, 0, 60, 0, 60, 0, 230, 48, 61, 0, 149, 255, 38, 0, 135, 48, 60, 0, 60, 0, 60, 0, 60, 0, 231, 48, 61, 0, 110, 255, 38, 0, 136, 48, 60, 0, 60, 0, 60, 0, 60, 0, 232, 48, 61, 0, 150, 255, 38, 0, 137, 48, 60, 0, 60, 0, 60, 0, 60, 0, 233, 48, 61, 0, 151, 255, 38, 0, 138, 48, 60, 0, 60, 0, 60, 0, 60, 0, 234, 48, 61, 0, 152, 255, 38, 0, 139, 48, 60, 0, 60, 0, 60, 0, 60, 0, 235, 48, 61, 0, 153, 255, 38, 0, 140, 48, 60, 0, 60, 0, 60, 0, 60, 0, 236, 48, 61, 0, 154, 255, 38, 0, 141, 48, 60, 0, 60, 0, 60, 0, 60, 0, 237, 48, 61, 0, 155, 255, 38, 0, 142, 48, 60, 0, 60, 0, 60, 0, 60, 0, 238, 48, 38, 0, 143, 48, 60, 0, 60, 0, 60, 0, 60, 0, 239, 48, 61, 0, 156, 255, 38, 0, 144, 48, 60, 0, 60, 0, 60, 0, 60, 0, 240, 48, 38, 0, 145, 48, 60, 0, 60, 0, 60, 0, 60, 0, 241, 48, 38, 0, 146, 48, 60, 0, 60, 0, 60, 0, 60, 0, 242, 48, 61, 0, 102, 255, 38, 0, 147, 48, 60, 0, 60, 0, 60, 0, 60, 0, 243, 48, 61, 0, 157, 255, 38, 0, 149, 48, 60, 0, 60, 0, 60, 0, 60, 0, 245, 48, 38, 0, 150, 48, 60, 0, 60, 0, 60, 0, 60, 0, 246, 48, 38, 0, 136, 48, 138, 48, 60, 0, 60, 0, 159, 48, 38, 0, 179, 48, 200, 48, 60, 0, 60, 0, 255, 48, 38, 0, 39, 0, 32, 0, 39, 0, 61, 0, 42, 0, 39, 0, 0, 48, 39, 0, 227, 255, 38, 0, 39, 0, 33, 0, 39, 0, 61, 0, 1, 255, 38, 0, 39, 0, 34, 0, 39, 0, 61, 0, 2, 255, 38, 0, 39, 0, 35, 0, 39, 0, 61, 0, 3, 255, 38, 0, 39, 0, 36, 0, 39, 0, 61, 0, 4, 255, 38, 0, 39, 0, 37, 0, 39, 0, 61, 0, 5, 255, 38, 0, 39, 0, 38, 0, 39, 0, 61, 0, 6, 255, 38, 0, 39, 0, 39, 0, 61, 0, 7, 255, 38, 0, 39, 0, 40, 0, 39, 0, 61, 0, 8, 255, 38, 0, 39, 0, 41, 0, 39, 0, 61, 0, 9, 255, 38, 0, 39, 0, 42, 0, 39, 0, 61, 0, 10, 255, 38, 0, 39, 0, 43, 0, 39, 0, 61, 0, 11, 255, 38, 0, 39, 0, 44, 0, 39, 0, 61, 0, 12, 255, 38, 0, 39, 0, 45, 0, 39, 0, 61, 0, 13, 255, 38, 0, 39, 0, 46, 0, 39, 0, 61, 0, 14, 255, 38, 0, 39, 0, 47, 0, 39, 0, 61, 0, 15, 255, 38, 0, 48, 0, 61, 0, 16, 255, 38, 0, 49, 0, 61, 0, 17, 255, 38, 0, 50, 0, 61, 0, 18, 255, 38, 0, 51, 0, 61, 0, 19, 255, 38, 0, 52, 0, 61, 0, 20, 255, 38, 0, 53, 0, 61, 0, 21, 255, 38, 0, 54, 0, 61, 0, 22, 255, 38, 0, 55, 0, 61, 0, 23, 255, 38, 0, 56, 0, 61, 0, 24, 255, 38, 0, 57, 0, 61, 0, 25, 255, 38, 0, 39, 0, 58, 0, 39, 0, 61, 0, 26, 255, 38, 0, 39, 0, 59, 0, 39, 0, 61, 0, 27, 255, 38, 0, 39, 0, 60, 0, 39, 0, 61, 0, 28, 255, 38, 0, 39, 0, 61, 0, 39, 0, 61, 0, 29, 255, 38, 0, 39, 0, 62, 0, 39, 0, 61, 0, 30, 255, 38, 0, 39, 0, 63, 0, 39, 0, 61, 0, 31, 255, 38, 0, 39, 0, 64, 0, 39, 0, 61, 0, 32, 255, 38, 0, 65, 0, 61, 0, 33, 255, 38, 0, 66, 0, 61, 0, 34, 255, 38, 0, 67, 0, 61, 0, 35, 255, 38, 0, 68, 0, 61, 0, 36, 255, 38, 0, 69, 0, 61, 0, 37, 255, 38, 0, 70, 0, 61, 0, 38, 255, 38, 0, 71, 0, 61, 0, 39, 255, 38, 0, 72, 0, 61, 0, 40, 255, 38, 0, 73, 0, 61, 0, 41, 255, 38, 0, 74, 0, 61, 0, 42, 255, 38, 0, 75, 0, 61, 0, 43, 255, 38, 0, 76, 0, 61, 0, 44, 255, 38, 0, 77, 0, 61, 0, 45, 255, 38, 0, 78, 0, 61, 0, 46, 255, 38, 0, 79, 0, 61, 0, 47, 255, 38, 0, 80, 0, 61, 0, 48, 255, 38, 0, 81, 0, 61, 0, 49, 255, 38, 0, 82, 0, 61, 0, 50, 255, 38, 0, 83, 0, 61, 0, 51, 255, 38, 0, 84, 0, 61, 0, 52, 255, 38, 0, 85, 0, 61, 0, 53, 255, 38, 0, 86, 0, 61, 0, 54, 255, 38, 0, 87, 0, 61, 0, 55, 255, 38, 0, 88, 0, 61, 0, 56, 255, 38, 0, 89, 0, 61, 0, 57, 255, 38, 0, 90, 0, 61, 0, 58, 255, 38, 0, 39, 0, 91, 0, 39, 0, 61, 0, 59, 255, 38, 0, 39, 0, 92, 0, 39, 0, 61, 0, 60, 255, 38, 0, 39, 0, 93, 0, 39, 0, 61, 0, 61, 255, 38, 0, 39, 0, 94, 0, 39, 0, 61, 0, 62, 255, 38, 0, 39, 0, 95, 0, 39, 0, 61, 0, 63, 255, 38, 0, 39, 0, 96, 0, 39, 0, 61, 0, 64, 255, 38, 0, 97, 0, 61, 0, 65, 255, 38, 0, 98, 0, 61, 0, 66, 255, 38, 0, 99, 0, 61, 0, 67, 255, 38, 0, 100, 0, 61, 0, 68, 255, 38, 0, 101, 0, 61, 0, 69, 255, 38, 0, 102, 0, 61, 0, 70, 255, 38, 0, 103, 0, 61, 0, 71, 255, 38, 0, 104, 0, 61, 0, 72, 255, 38, 0, 105, 0, 61, 0, 73, 255, 38, 0, 106, 0, 61, 0, 74, 255, 38, 0, 107, 0, 61, 0, 75, 255, 38, 0, 108, 0, 61, 0, 76, 255, 38, 0, 109, 0, 61, 0, 77, 255, 38, 0, 110, 0, 61, 0, 78, 255, 38, 0, 111, 0, 61, 0, 79, 255, 38, 0, 112, 0, 61, 0, 80, 255, 38, 0, 113, 0, 61, 0, 81, 255, 38, 0, 114, 0, 61, 0, 82, 255, 38, 0, 115, 0, 61, 0, 83, 255, 38, 0, 116, 0, 61, 0, 84, 255, 38, 0, 117, 0, 61, 0, 85, 255, 38, 0, 118, 0, 61, 0, 86, 255, 38, 0, 119, 0, 61, 0, 87, 255, 38, 0, 120, 0, 61, 0, 88, 255, 38, 0, 121, 0, 61, 0, 89, 255, 38, 0, 122, 0, 61, 0, 90, 255, 38, 0, 39, 0, 123, 0, 39, 0, 61, 0, 91, 255, 38, 0, 39, 0, 124, 0, 39, 0, 61, 0, 92, 255, 38, 0, 39, 0, 125, 0, 39, 0, 61, 0, 93, 255, 38, 0, 39, 0, 126, 0, 39, 0, 61, 0, 94, 255, 38, 0, 162, 0, 61, 0, 224, 255, 38, 0, 163, 0, 61, 0, 225, 255, 38, 0, 165, 0, 61, 0, 229, 255, 38, 0, 166, 0, 61, 0, 228, 255, 38, 0, 172, 0, 61, 0, 226, 255, 38, 0, 0, 17, 61, 0, 161, 255, 61, 0, 49, 49, 38, 0, 1, 17, 61, 0, 162, 255, 61, 0, 50, 49, 38, 0, 2, 17, 61, 0, 164, 255, 61, 0, 52, 49, 38, 0, 3, 17, 61, 0, 167, 255, 61, 0, 55, 49, 38, 0, 4, 17, 61, 0, 168, 255, 61, 0, 56, 49, 38, 0, 5, 17, 61, 0, 169, 255, 61, 0, 57, 49, 38, 0, 6, 17, 61, 0, 177, 255, 61, 0, 65, 49, 38, 0, 7, 17, 61, 0, 178, 255, 61, 0, 66, 49, 38, 0, 8, 17, 61, 0, 179, 255, 61, 0, 67, 49, 38, 0, 9, 17, 61, 0, 181, 255, 61, 0, 69, 49, 38, 0, 10, 17, 61, 0, 182, 255, 61, 0, 70, 49, 38, 0, 11, 17, 61, 0, 183, 255, 61, 0, 71, 49, 38, 0, 12, 17, 61, 0, 184, 255, 61, 0, 72, 49, 38, 0, 13, 17, 61, 0, 185, 255, 61, 0, 73, 49, 38, 0, 14, 17, 61, 0, 186, 255, 61, 0, 74, 49, 38, 0, 15, 17, 61, 0, 187, 255, 61, 0, 75, 49, 38, 0, 16, 17, 61, 0, 188, 255, 61, 0, 76, 49, 38, 0, 17, 17, 61, 0, 189, 255, 61, 0, 77, 49, 38, 0, 18, 17, 61, 0, 190, 255, 61, 0, 78, 49, 38, 0, 26, 17, 61, 0, 176, 255, 61, 0, 64, 49, 38, 0, 33, 17, 61, 0, 180, 255, 61, 0, 68, 49, 38, 0, 96, 17, 61, 0, 160, 255, 61, 0, 100, 49, 38, 0, 97, 17, 61, 0, 194, 255, 61, 0, 79, 49, 38, 0, 98, 17, 61, 0, 195, 255, 61, 0, 80, 49, 38, 0, 99, 17, 61, 0, 196, 255, 61, 0, 81, 49, 38, 0, 100, 17, 61, 0, 197, 255, 61, 0, 82, 49, 38, 0, 101, 17, 61, 0, 198, 255, 61, 0, 83, 49, 38, 0, 102, 17, 61, 0, 199, 255, 61, 0, 84, 49, 38, 0, 103, 17, 61, 0, 202, 255, 61, 0, 85, 49, 38, 0, 104, 17, 61, 0, 203, 255, 61, 0, 86, 49, 38, 0, 105, 17, 61, 0, 204, 255, 61, 0, 87, 49, 38, 0, 106, 17, 61, 0, 205, 255, 61, 0, 88, 49, 38, 0, 107, 17, 61, 0, 206, 255, 61, 0, 89, 49, 38, 0, 108, 17, 61, 0, 207, 255, 61, 0, 90, 49, 38, 0, 109, 17, 61, 0, 210, 255, 61, 0, 91, 49, 38, 0, 110, 17, 61, 0, 211, 255, 61, 0, 92, 49, 38, 0, 111, 17, 61, 0, 212, 255, 61, 0, 93, 49, 38, 0, 112, 17, 61, 0, 213, 255, 61, 0, 94, 49, 38, 0, 113, 17, 61, 0, 214, 255, 61, 0, 95, 49, 38, 0, 114, 17, 61, 0, 215, 255, 61, 0, 96, 49, 38, 0, 115, 17, 61, 0, 218, 255, 61, 0, 97, 49, 38, 0, 116, 17, 61, 0, 219, 255, 61, 0, 98, 49, 38, 0, 117, 17, 61, 0, 220, 255, 61, 0, 99, 49, 38, 0, 170, 17, 61, 0, 163, 255, 61, 0, 51, 49, 38, 0, 172, 17, 61, 0, 165, 255, 61, 0, 53, 49, 38, 0, 173, 17, 61, 0, 166, 255, 61, 0, 54, 49, 38, 0, 176, 17, 61, 0, 170, 255, 61, 0, 58, 49, 38, 0, 177, 17, 61, 0, 171, 255, 61, 0, 59, 49, 38, 0, 178, 17, 61, 0, 172, 255, 61, 0, 60, 49, 38, 0, 179, 17, 61, 0, 173, 255, 61, 0, 61, 49, 38, 0, 180, 17, 61, 0, 174, 255, 61, 0, 62, 49, 38, 0, 181, 17, 61, 0, 175, 255, 61, 0, 63, 49, 38, 0, 169, 32, 61, 0, 230, 255, 38, 0, 144, 33, 61, 0, 233, 255, 38, 0, 145, 33, 61, 0, 234, 255, 38, 0, 146, 33, 61, 0, 235, 255, 38, 0, 147, 33, 61, 0, 236, 255, 38, 0, 2, 37, 61, 0, 232, 255, 38, 0, 160, 37, 61, 0, 237, 255, 38, 0, 203, 37, 61, 0, 238, 255, 38, 0, 1, 48, 61, 0, 100, 255, 38, 0, 2, 48, 61, 0, 97, 255, 38, 0, 12, 48, 61, 0, 98, 255, 38, 0, 13, 48, 61, 0, 99, 255, 0, 0, 239, 223, 10, 25, 38, 0, 253, 48, 60, 0, 42, 0, 126, 51, 125, 51, 124, 51, 123, 51, 255, 50, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 106, 0, 97, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 107, 0, 97, 0, 110, 0, 97, 0, 93, 0, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 114, 0, 101, 0, 103, 0, 117, 0, 108, 0, 97, 0, 114, 0, 93, 0, 60, 0, 42, 0, 156, 78, 22, 85, 3, 90, 63, 150, 192, 84, 27, 97, 40, 99, 246, 89, 34, 144, 117, 132, 28, 131, 80, 122, 170, 96, 225, 99, 37, 110, 237, 101, 102, 132, 166, 130, 245, 155, 147, 104, 39, 87, 161, 101, 113, 98, 155, 91, 208, 89, 123, 134, 244, 152, 98, 125, 190, 125, 142, 155, 22, 98, 159, 124, 183, 136, 137, 91, 181, 94, 9, 99, 151, 102, 72, 104, 199, 149, 141, 151, 79, 103, 229, 78, 10, 79, 77, 79, 157, 79, 73, 80, 242, 86, 55, 89, 212, 89, 1, 90, 9, 92, 223, 96, 15, 97, 112, 97, 19, 102, 5, 105, 186, 112, 79, 117, 112, 117, 251, 121, 173, 125, 239, 125, 195, 128, 14, 132, 99, 136, 2, 139, 85, 144, 122, 144, 59, 83, 149, 78, 165, 78, 223, 87, 178, 128, 193, 144, 239, 120, 0, 78, 241, 88, 162, 110, 56, 144, 50, 122, 40, 131, 139, 130, 47, 156, 65, 81, 112, 83, 189, 84, 225, 84, 224, 86, 251, 89, 21, 95, 242, 152, 235, 109, 228, 128, 45, 133, 98, 150, 112, 150, 160, 150, 251, 151, 11, 84, 243, 83, 135, 91, 207, 112, 189, 127, 194, 143, 232, 150, 111, 83, 92, 157, 186, 122, 17, 78, 147, 120, 252, 129, 38, 110, 24, 86, 4, 85, 29, 107, 26, 133, 59, 156, 229, 89, 169, 83, 102, 109, 220, 116, 143, 149, 66, 86, 145, 78, 75, 144, 242, 150, 79, 131, 12, 153, 225, 83, 182, 85, 48, 91, 113, 95, 32, 102, 243, 102, 4, 104, 56, 108, 243, 108, 41, 109, 91, 116, 200, 118, 78, 122, 52, 152, 241, 130, 91, 136, 96, 138, 237, 146, 178, 109, 171, 117, 202, 118, 197, 153, 166, 96, 1, 139, 138, 141, 178, 149, 142, 105, 173, 83, 134, 81, 18, 87, 48, 88, 68, 89, 180, 91, 246, 94, 40, 96, 169, 99, 244, 99, 191, 108, 20, 111, 142, 112, 20, 113, 89, 113, 213, 113, 63, 115, 1, 126, 118, 130, 209, 130, 151, 133, 96, 144, 91, 146, 27, 157, 105, 88, 188, 101, 90, 108, 37, 117, 249, 81, 46, 89, 101, 89, 128, 95, 220, 95, 188, 98, 250, 101, 42, 106, 39, 107, 180, 107, 139, 115, 193, 127, 86, 137, 44, 157, 14, 157, 196, 158, 161, 92, 150, 108, 123, 131, 4, 81, 75, 92, 182, 97, 198, 129, 118, 104, 97, 114, 89, 78, 250, 79, 120, 83, 105, 96, 41, 110, 79, 122, 243, 151, 11, 78, 22, 83, 238, 78, 85, 79, 61, 79, 161, 79, 115, 79, 160, 82, 239, 83, 9, 86, 15, 89, 193, 90, 182, 91, 225, 91, 209, 121, 135, 102, 156, 103, 182, 103, 76, 107, 179, 108, 107, 112, 194, 115, 141, 121, 190, 121, 60, 122, 135, 123, 177, 130, 219, 130, 4, 131, 119, 131, 239, 131, 211, 131, 102, 135, 178, 138, 41, 86, 168, 140, 230, 143, 78, 144, 30, 151, 138, 134, 196, 79, 232, 92, 17, 98, 89, 114, 59, 117, 229, 129, 189, 130, 254, 134, 192, 140, 197, 150, 19, 153, 213, 153, 203, 78, 26, 79, 227, 137, 222, 86, 74, 88, 202, 88, 251, 94, 235, 95, 42, 96, 148, 96, 98, 96, 208, 97, 18, 98, 208, 98, 57, 101, 65, 155, 102, 102, 176, 104, 119, 109, 112, 112, 76, 117, 134, 118, 117, 125, 165, 130, 249, 135, 139, 149, 142, 150, 157, 140, 241, 81, 190, 82, 22, 89, 179, 84, 179, 91, 22, 93, 104, 97, 130, 105, 175, 109, 141, 120, 203, 132, 87, 136, 114, 138, 167, 147, 184, 154, 108, 109, 168, 153, 217, 134, 163, 87, 255, 103, 206, 134, 14, 146, 131, 82, 135, 86, 4, 84, 211, 94, 225, 98, 185, 100, 60, 104, 56, 104, 187, 107, 114, 115, 186, 120, 107, 122, 154, 137, 210, 137, 107, 141, 3, 143, 237, 144, 163, 149, 148, 150, 105, 151, 102, 91, 179, 92, 125, 105, 77, 152, 78, 152, 155, 99, 32, 123, 43, 106, 127, 106, 182, 104, 13, 156, 95, 111, 114, 82, 157, 85, 112, 96, 236, 98, 59, 109, 7, 110, 209, 110, 91, 132, 16, 137, 68, 143, 20, 78, 57, 156, 246, 83, 27, 105, 58, 106, 132, 151, 42, 104, 92, 81, 195, 122, 178, 132, 220, 145, 140, 147, 91, 86, 40, 157, 34, 104, 5, 131, 49, 132, 165, 124, 8, 82, 197, 130, 230, 116, 126, 78, 131, 79, 160, 81, 210, 91, 10, 82, 216, 82, 231, 82, 251, 93, 154, 85, 42, 88, 230, 89, 140, 91, 152, 91, 219, 91, 114, 94, 121, 94, 163, 96, 31, 97, 99, 97, 190, 97, 219, 99, 98, 101, 209, 103, 83, 104, 250, 104, 62, 107, 83, 107, 87, 108, 34, 111, 151, 111, 69, 111, 176, 116, 24, 117, 227, 118, 11, 119, 255, 122, 161, 123, 33, 124, 233, 125, 54, 127, 240, 127, 157, 128, 102, 130, 158, 131, 179, 137, 204, 138, 171, 140, 132, 144, 81, 148, 147, 149, 145, 149, 162, 149, 101, 150, 211, 151, 40, 153, 24, 130, 56, 78, 43, 84, 184, 92, 204, 93, 169, 115, 76, 118, 60, 119, 169, 92, 235, 127, 11, 141, 193, 150, 17, 152, 84, 152, 88, 152, 1, 79, 14, 79, 113, 83, 156, 85, 104, 86, 250, 87, 71, 89, 9, 91, 196, 91, 144, 92, 12, 94, 126, 94, 204, 95, 238, 99, 58, 103, 215, 101, 226, 101, 31, 103, 203, 104, 196, 104, 95, 106, 48, 94, 197, 107, 23, 108, 125, 108, 127, 117, 72, 121, 99, 91, 0, 122, 0, 125, 189, 95, 143, 137, 24, 138, 180, 140, 119, 141, 204, 142, 29, 143, 226, 152, 14, 154, 60, 155, 128, 78, 125, 80, 0, 81, 147, 89, 156, 91, 47, 98, 128, 98, 236, 100, 58, 107, 160, 114, 145, 117, 71, 121, 169, 127, 251, 135, 188, 138, 112, 139, 172, 99, 202, 131, 160, 151, 9, 84, 3, 84, 171, 85, 84, 104, 88, 106, 112, 138, 39, 120, 117, 103, 205, 158, 116, 83, 162, 91, 26, 129, 80, 134, 6, 144, 24, 78, 69, 78, 199, 78, 17, 79, 202, 83, 56, 84, 174, 91, 19, 95, 37, 96, 81, 101, 61, 103, 66, 108, 114, 108, 227, 108, 120, 112, 3, 116, 118, 122, 174, 122, 8, 123, 26, 125, 254, 124, 102, 125, 231, 101, 91, 114, 187, 83, 69, 92, 232, 93, 210, 98, 224, 98, 25, 99, 32, 110, 90, 134, 49, 138, 221, 141, 248, 146, 1, 111, 166, 121, 90, 155, 168, 78, 171, 78, 172, 78, 155, 79, 160, 79, 209, 80, 71, 81, 246, 122, 113, 81, 246, 81, 84, 83, 33, 83, 127, 83, 235, 83, 172, 85, 131, 88, 225, 92, 55, 95, 74, 95, 47, 96, 80, 96, 109, 96, 31, 99, 89, 101, 75, 106, 193, 108, 194, 114, 237, 114, 239, 119, 248, 128, 5, 129, 8, 130, 78, 133, 247, 144, 225, 147, 255, 151, 87, 153, 90, 154, 240, 78, 221, 81, 45, 92, 129, 102, 109, 105, 64, 92, 242, 102, 117, 105, 137, 115, 80, 104, 129, 124, 197, 80, 228, 82, 71, 87, 254, 93, 38, 147, 164, 101, 35, 107, 61, 107, 52, 116, 129, 121, 189, 121, 75, 123, 202, 125, 185, 130, 204, 131, 127, 136, 95, 137, 57, 139, 209, 143, 209, 145, 31, 84, 128, 146, 93, 78, 54, 80, 229, 83, 58, 83, 215, 114, 150, 115, 233, 119, 230, 130, 175, 142, 198, 153, 200, 153, 210, 153, 119, 81, 26, 97, 94, 134, 176, 85, 122, 122, 118, 80, 211, 91, 71, 144, 133, 150, 50, 78, 219, 106, 231, 145, 81, 92, 72, 92, 152, 99, 159, 122, 147, 108, 116, 151, 97, 143, 170, 122, 138, 113, 136, 150, 130, 124, 23, 104, 112, 126, 81, 104, 108, 147, 242, 82, 27, 84, 171, 133, 19, 138, 164, 127, 205, 142, 225, 144, 102, 83, 136, 136, 65, 121, 194, 79, 190, 80, 17, 82, 68, 81, 83, 85, 45, 87, 234, 115, 139, 87, 81, 89, 98, 95, 132, 95, 117, 96, 118, 97, 103, 97, 169, 97, 178, 99, 58, 100, 108, 101, 111, 102, 66, 104, 19, 110, 102, 117, 61, 122, 251, 124, 76, 125, 153, 125, 75, 126, 107, 127, 14, 131, 74, 131, 205, 134, 8, 138, 99, 138, 102, 139, 253, 142, 26, 152, 143, 157, 184, 130, 206, 143, 232, 155, 135, 82, 31, 98, 131, 100, 192, 111, 153, 150, 65, 104, 145, 80, 32, 107, 122, 108, 84, 111, 116, 122, 80, 125, 64, 136, 35, 138, 8, 103, 246, 78, 57, 80, 38, 80, 101, 80, 124, 81, 56, 82, 99, 82, 167, 85, 15, 87, 5, 88, 204, 90, 250, 94, 178, 97, 248, 97, 243, 98, 114, 99, 28, 105, 41, 106, 125, 114, 172, 114, 46, 115, 20, 120, 111, 120, 121, 125, 12, 119, 169, 128, 139, 137, 25, 139, 226, 140, 210, 142, 99, 144, 117, 147, 122, 150, 85, 152, 19, 154, 120, 158, 67, 81, 159, 83, 179, 83, 123, 94, 38, 95, 27, 110, 144, 110, 132, 115, 254, 115, 67, 125, 55, 130, 0, 138, 250, 138, 80, 150, 78, 78, 11, 80, 228, 83, 124, 84, 250, 86, 209, 89, 100, 91, 241, 93, 171, 94, 39, 95, 56, 98, 69, 101, 175, 103, 86, 110, 208, 114, 202, 124, 180, 136, 161, 128, 225, 128, 240, 131, 78, 134, 135, 138, 232, 141, 55, 146, 199, 150, 103, 152, 19, 159, 148, 78, 146, 78, 13, 79, 72, 83, 73, 84, 62, 84, 47, 90, 140, 95, 161, 95, 159, 96, 167, 104, 142, 106, 90, 116, 129, 120, 158, 138, 164, 138, 119, 139, 144, 145, 94, 78, 201, 155, 164, 78, 124, 79, 175, 79, 25, 80, 22, 80, 73, 81, 108, 81, 159, 82, 185, 82, 254, 82, 154, 83, 227, 83, 17, 84, 14, 84, 137, 85, 81, 87, 162, 87, 125, 89, 84, 91, 93, 91, 143, 91, 229, 93, 231, 93, 247, 93, 120, 94, 131, 94, 154, 94, 183, 94, 24, 95, 82, 96, 76, 97, 151, 98, 216, 98, 167, 99, 59, 101, 2, 102, 67, 102, 244, 102, 109, 103, 33, 104, 151, 104, 203, 105, 95, 108, 42, 109, 105, 109, 47, 110, 157, 110, 50, 117, 135, 118, 108, 120, 63, 122, 224, 124, 5, 125, 24, 125, 94, 125, 177, 125, 21, 128, 3, 128, 175, 128, 177, 128, 84, 129, 143, 129, 42, 130, 82, 131, 76, 136, 97, 136, 27, 139, 162, 140, 252, 140, 202, 144, 117, 145, 113, 146, 63, 120, 252, 146, 164, 149, 77, 150, 5, 152, 153, 153, 216, 154, 59, 157, 91, 82, 171, 82, 247, 83, 8, 84, 213, 88, 247, 98, 224, 111, 106, 140, 95, 143, 185, 158, 75, 81, 59, 82, 74, 84, 253, 86, 64, 122, 119, 145, 96, 157, 210, 158, 68, 115, 9, 111, 112, 129, 17, 117, 253, 95, 218, 96, 168, 154, 219, 114, 188, 143, 100, 107, 3, 152, 202, 78, 240, 86, 100, 87, 190, 88, 90, 90, 104, 96, 199, 97, 15, 102, 6, 102, 57, 104, 177, 104, 247, 109, 213, 117, 58, 125, 110, 130, 66, 155, 155, 78, 80, 79, 201, 83, 6, 85, 111, 93, 230, 93, 238, 93, 251, 103, 153, 108, 115, 116, 2, 120, 80, 138, 150, 147, 223, 136, 80, 87, 167, 94, 43, 99, 181, 80, 172, 80, 141, 81, 0, 103, 201, 84, 94, 88, 187, 89, 176, 91, 105, 95, 77, 98, 161, 99, 61, 104, 115, 107, 8, 110, 125, 112, 199, 145, 128, 114, 21, 120, 38, 120, 109, 121, 142, 101, 48, 125, 220, 131, 193, 136, 9, 143, 155, 150, 100, 82, 40, 87, 80, 103, 106, 127, 161, 140, 180, 81, 66, 87, 42, 150, 58, 88, 138, 105, 180, 128, 178, 84, 14, 93, 252, 87, 149, 120, 250, 157, 92, 79, 74, 82, 139, 84, 62, 100, 40, 102, 20, 103, 245, 103, 132, 122, 86, 123, 34, 125, 47, 147, 92, 104, 173, 155, 57, 123, 25, 83, 138, 81, 55, 82, 223, 91, 246, 98, 174, 100, 230, 100, 45, 103, 186, 107, 169, 133, 209, 150, 144, 118, 214, 155, 76, 99, 6, 147, 171, 155, 191, 118, 82, 102, 9, 78, 152, 80, 194, 83, 113, 92, 232, 96, 146, 100, 99, 101, 95, 104, 230, 113, 202, 115, 35, 117, 151, 123, 130, 126, 149, 134, 131, 139, 219, 140, 120, 145, 16, 153, 172, 101, 171, 102, 139, 107, 213, 78, 212, 78, 58, 79, 127, 79, 58, 82, 248, 83, 242, 83, 227, 85, 219, 86, 235, 88, 203, 89, 201, 89, 255, 89, 80, 91, 77, 92, 2, 94, 43, 94, 215, 95, 29, 96, 7, 99, 47, 101, 92, 91, 175, 101, 189, 101, 232, 101, 157, 103, 98, 107, 123, 107, 15, 108, 69, 115, 73, 121, 193, 121, 248, 124, 25, 125, 43, 125, 162, 128, 2, 129, 243, 129, 150, 137, 94, 138, 105, 138, 102, 138, 140, 138, 238, 138, 199, 140, 220, 140, 204, 150, 252, 152, 111, 107, 139, 78, 60, 79, 141, 79, 80, 81, 87, 91, 250, 91, 72, 97, 1, 99, 66, 102, 33, 107, 203, 110, 187, 108, 62, 114, 189, 116, 212, 117, 193, 120, 58, 121, 12, 128, 51, 128, 234, 129, 148, 132, 158, 143, 80, 108, 127, 158, 15, 95, 88, 139, 43, 157, 250, 122, 248, 142, 141, 91, 235, 150, 3, 78, 241, 83, 247, 87, 49, 89, 201, 90, 164, 91, 137, 96, 127, 110, 6, 111, 190, 117, 234, 140, 159, 91, 0, 133, 224, 123, 114, 80, 244, 103, 157, 130, 97, 92, 74, 133, 30, 126, 14, 130, 153, 81, 4, 92, 104, 99, 102, 141, 156, 101, 110, 113, 62, 121, 23, 125, 5, 128, 29, 139, 202, 142, 110, 144, 199, 134, 170, 144, 31, 80, 250, 82, 58, 92, 83, 103, 124, 112, 53, 114, 76, 145, 200, 145, 43, 147, 229, 130, 194, 91, 49, 95, 249, 96, 59, 78, 214, 83, 136, 91, 75, 98, 49, 103, 138, 107, 233, 114, 224, 115, 46, 122, 107, 129, 163, 141, 82, 145, 150, 153, 18, 81, 215, 83, 106, 84, 255, 91, 136, 99, 57, 106, 172, 125, 0, 151, 218, 86, 206, 83, 104, 84, 151, 91, 49, 92, 222, 93, 238, 79, 1, 97, 254, 98, 50, 109, 192, 121, 203, 121, 66, 125, 77, 126, 210, 127, 237, 129, 31, 130, 144, 132, 70, 136, 114, 137, 144, 139, 116, 142, 47, 143, 49, 144, 75, 145, 108, 145, 198, 150, 156, 145, 192, 78, 79, 79, 69, 81, 65, 83, 147, 95, 14, 98, 212, 103, 65, 108, 11, 110, 99, 115, 38, 126, 205, 145, 131, 146, 212, 83, 25, 89, 191, 91, 209, 109, 93, 121, 46, 126, 155, 124, 126, 88, 159, 113, 250, 81, 83, 136, 240, 143, 202, 79, 251, 92, 37, 102, 172, 119, 227, 122, 28, 130, 255, 153, 198, 81, 170, 95, 236, 101, 111, 105, 137, 107, 243, 109, 150, 110, 100, 111, 254, 118, 20, 125, 225, 93, 117, 144, 135, 145, 6, 152, 230, 81, 29, 82, 64, 98, 145, 102, 217, 102, 26, 110, 182, 94, 210, 125, 114, 127, 248, 102, 175, 133, 247, 133, 248, 138, 169, 82, 217, 83, 115, 89, 143, 94, 144, 95, 85, 96, 228, 146, 100, 150, 183, 80, 31, 81, 221, 82, 32, 83, 71, 83, 236, 83, 232, 84, 70, 85, 49, 85, 23, 86, 104, 89, 190, 89, 60, 90, 181, 91, 6, 92, 15, 92, 17, 92, 26, 92, 132, 94, 138, 94, 224, 94, 112, 95, 127, 98, 132, 98, 219, 98, 140, 99, 119, 99, 7, 102, 12, 102, 45, 102, 118, 102, 126, 103, 162, 104, 31, 106, 53, 106, 188, 108, 136, 109, 9, 110, 88, 110, 60, 113, 38, 113, 103, 113, 199, 117, 1, 119, 93, 120, 1, 121, 101, 121, 240, 121, 224, 122, 17, 123, 167, 124, 57, 125, 150, 128, 214, 131, 139, 132, 73, 133, 93, 136, 243, 136, 31, 138, 60, 138, 84, 138, 115, 138, 97, 140, 222, 140, 164, 145, 102, 146, 126, 147, 24, 148, 156, 150, 152, 151, 10, 78, 8, 78, 30, 78, 87, 78, 151, 81, 112, 82, 206, 87, 52, 88, 204, 88, 34, 91, 56, 94, 197, 96, 254, 100, 97, 103, 86, 103, 68, 109, 182, 114, 115, 117, 99, 122, 184, 132, 114, 139, 184, 145, 32, 147, 49, 86, 244, 87, 254, 152, 237, 98, 13, 105, 150, 107, 237, 113, 84, 126, 119, 128, 114, 130, 230, 137, 223, 152, 85, 135, 177, 143, 59, 92, 56, 79, 225, 79, 181, 79, 7, 85, 32, 90, 221, 91, 233, 91, 195, 95, 78, 97, 47, 99, 176, 101, 75, 102, 238, 104, 155, 105, 120, 109, 241, 109, 51, 117, 185, 117, 31, 119, 94, 121, 230, 121, 51, 125, 227, 129, 175, 130, 170, 133, 170, 137, 58, 138, 171, 142, 155, 143, 50, 144, 221, 145, 7, 151, 186, 78, 193, 78, 3, 82, 117, 88, 236, 88, 11, 92, 26, 117, 61, 92, 78, 129, 10, 138, 197, 143, 99, 150, 109, 151, 37, 123, 207, 138, 8, 152, 98, 145, 243, 86, 168, 83, 23, 144, 57, 84, 130, 87, 37, 94, 168, 99, 52, 108, 138, 112, 97, 119, 139, 124, 224, 127, 112, 136, 66, 144, 84, 145, 16, 147, 24, 147, 143, 150, 94, 116, 196, 154, 7, 93, 105, 93, 112, 101, 162, 103, 168, 141, 219, 150, 110, 99, 73, 103, 25, 105, 197, 131, 23, 152, 192, 150, 254, 136, 132, 111, 122, 100, 248, 91, 22, 78, 44, 112, 93, 117, 47, 102, 196, 81, 54, 82, 226, 82, 211, 89, 129, 95, 39, 96, 16, 98, 63, 101, 116, 101, 31, 102, 116, 102, 242, 104, 22, 104, 99, 107, 5, 110, 114, 114, 31, 117, 219, 118, 190, 124, 86, 128, 240, 88, 253, 136, 127, 137, 160, 138, 147, 138, 203, 138, 29, 144, 146, 145, 82, 151, 89, 151, 137, 101, 14, 122, 6, 129, 187, 150, 45, 94, 220, 96, 26, 98, 165, 101, 20, 102, 144, 103, 243, 119, 77, 122, 77, 124, 62, 126, 10, 129, 172, 140, 100, 141, 225, 141, 95, 142, 169, 120, 7, 82, 217, 98, 165, 99, 66, 100, 152, 98, 45, 138, 131, 122, 192, 123, 172, 138, 234, 150, 118, 125, 12, 130, 73, 135, 217, 78, 72, 81, 67, 83, 96, 83, 163, 91, 2, 92, 22, 92, 221, 93, 38, 98, 71, 98, 176, 100, 19, 104, 52, 104, 201, 108, 69, 109, 23, 109, 211, 103, 92, 111, 78, 113, 125, 113, 203, 101, 127, 122, 173, 123, 218, 125, 74, 126, 168, 127, 122, 129, 27, 130, 57, 130, 166, 133, 110, 138, 206, 140, 245, 141, 120, 144, 119, 144, 173, 146, 145, 146, 131, 149, 174, 155, 77, 82, 132, 85, 56, 111, 54, 113, 104, 81, 133, 121, 85, 126, 179, 129, 206, 124, 76, 86, 81, 88, 168, 92, 170, 99, 254, 102, 253, 102, 90, 105, 217, 114, 143, 117, 142, 117, 14, 121, 86, 121, 223, 121, 151, 124, 32, 125, 68, 125, 7, 134, 52, 138, 59, 150, 97, 144, 32, 159, 231, 80, 117, 82, 204, 83, 226, 83, 9, 80, 170, 85, 238, 88, 79, 89, 61, 114, 139, 91, 100, 92, 29, 83, 227, 96, 243, 96, 92, 99, 131, 99, 63, 99, 187, 99, 205, 100, 233, 101, 249, 102, 227, 93, 205, 105, 253, 105, 21, 111, 229, 113, 137, 78, 233, 117, 248, 118, 147, 122, 223, 124, 207, 125, 156, 125, 97, 128, 73, 131, 88, 131, 108, 132, 188, 132, 251, 133, 197, 136, 112, 141, 1, 144, 109, 144, 151, 147, 28, 151, 18, 154, 207, 80, 151, 88, 142, 97, 211, 129, 53, 133, 8, 141, 32, 144, 195, 79, 116, 80, 71, 82, 115, 83, 111, 96, 73, 99, 95, 103, 44, 110, 179, 141, 31, 144, 215, 79, 94, 92, 202, 140, 207, 101, 154, 125, 82, 83, 150, 136, 118, 81, 195, 99, 88, 91, 107, 91, 10, 92, 13, 100, 81, 103, 92, 144, 214, 78, 26, 89, 42, 89, 112, 108, 81, 138, 62, 85, 21, 88, 165, 89, 240, 96, 83, 98, 193, 103, 53, 130, 85, 105, 64, 150, 196, 153, 40, 154, 83, 79, 6, 88, 254, 91, 16, 128, 177, 92, 47, 94, 133, 95, 32, 96, 75, 97, 52, 98, 255, 102, 240, 108, 222, 110, 206, 128, 127, 129, 212, 130, 139, 136, 184, 140, 0, 144, 46, 144, 138, 150, 219, 158, 219, 155, 227, 78, 240, 83, 39, 89, 44, 123, 141, 145, 76, 152, 249, 157, 221, 110, 39, 112, 83, 83, 68, 85, 133, 91, 88, 98, 158, 98, 211, 98, 162, 108, 239, 111, 34, 116, 23, 138, 56, 148, 193, 111, 254, 138, 56, 131, 231, 81, 248, 134, 234, 83, 233, 83, 70, 79, 84, 144, 176, 143, 106, 89, 49, 129, 253, 93, 234, 122, 191, 143, 218, 104, 55, 140, 248, 114, 72, 156, 61, 106, 176, 138, 57, 78, 88, 83, 6, 86, 102, 87, 197, 98, 162, 99, 230, 101, 78, 107, 225, 109, 91, 110, 173, 112, 237, 119, 239, 122, 170, 123, 187, 125, 61, 128, 198, 128, 203, 134, 149, 138, 91, 147, 227, 86, 199, 88, 62, 95, 173, 101, 150, 102, 128, 106, 181, 107, 55, 117, 199, 138, 36, 80, 229, 119, 48, 87, 27, 95, 101, 96, 122, 102, 96, 108, 244, 117, 26, 122, 110, 127, 244, 129, 24, 135, 69, 144, 179, 153, 201, 123, 92, 117, 249, 122, 81, 123, 196, 132, 16, 144, 233, 121, 146, 122, 54, 131, 225, 90, 64, 119, 45, 78, 242, 78, 153, 91, 224, 95, 189, 98, 60, 102, 241, 103, 232, 108, 107, 134, 119, 136, 59, 138, 78, 145, 243, 146, 208, 153, 23, 106, 38, 112, 42, 115, 231, 130, 87, 132, 175, 140, 1, 78, 70, 81, 203, 81, 139, 85, 245, 91, 22, 94, 51, 94, 129, 94, 20, 95, 53, 95, 107, 95, 180, 95, 242, 97, 17, 99, 162, 102, 29, 103, 110, 111, 82, 114, 58, 117, 58, 119, 116, 128, 57, 129, 120, 129, 118, 135, 191, 138, 220, 138, 133, 141, 243, 141, 154, 146, 119, 149, 2, 152, 229, 156, 197, 82, 87, 99, 244, 118, 21, 103, 136, 108, 205, 115, 195, 140, 174, 147, 115, 150, 37, 109, 156, 88, 14, 105, 204, 105, 253, 143, 154, 147, 219, 117, 26, 144, 90, 88, 2, 104, 180, 99, 251, 105, 67, 79, 44, 111, 216, 103, 187, 143, 38, 133, 180, 125, 84, 147, 63, 105, 112, 111, 106, 87, 247, 88, 44, 91, 44, 125, 42, 114, 10, 84, 227, 145, 180, 157, 173, 78, 78, 79, 92, 80, 117, 80, 67, 82, 158, 140, 72, 84, 36, 88, 154, 91, 29, 94, 149, 94, 173, 94, 247, 94, 31, 95, 140, 96, 181, 98, 58, 99, 208, 99, 175, 104, 64, 108, 135, 120, 142, 121, 11, 122, 224, 125, 71, 130, 2, 138, 230, 138, 68, 142, 19, 144, 184, 144, 45, 145, 216, 145, 14, 159, 229, 108, 88, 100, 226, 100, 117, 101, 244, 110, 132, 118, 27, 123, 105, 144, 209, 147, 186, 110, 242, 84, 185, 95, 164, 100, 77, 143, 237, 143, 68, 146, 120, 81, 107, 88, 41, 89, 85, 92, 151, 94, 251, 109, 143, 126, 28, 117, 188, 140, 226, 142, 91, 152, 185, 112, 29, 79, 191, 107, 177, 111, 48, 117, 251, 150, 78, 81, 16, 84, 53, 88, 87, 88, 172, 89, 96, 92, 146, 95, 151, 101, 92, 103, 33, 110, 123, 118, 223, 131, 237, 140, 20, 144, 253, 144, 77, 147, 37, 120, 58, 120, 170, 82, 166, 94, 31, 87, 116, 89, 18, 96, 18, 80, 90, 81, 172, 81, 205, 81, 0, 82, 16, 85, 84, 88, 88, 88, 87, 89, 149, 91, 246, 92, 139, 93, 188, 96, 149, 98, 45, 100, 113, 103, 67, 104, 188, 104, 223, 104, 215, 118, 216, 109, 111, 110, 155, 109, 111, 112, 200, 113, 83, 95, 216, 117, 119, 121, 73, 123, 84, 123, 82, 123, 214, 124, 113, 125, 48, 82, 99, 132, 105, 133, 228, 133, 14, 138, 4, 139, 70, 140, 15, 142, 3, 144, 15, 144, 25, 148, 118, 150, 45, 152, 48, 154, 216, 149, 205, 80, 213, 82, 12, 84, 2, 88, 14, 92, 167, 97, 158, 100, 30, 109, 179, 119, 229, 122, 244, 128, 4, 132, 83, 144, 133, 146, 224, 92, 7, 157, 63, 83, 151, 95, 179, 95, 156, 109, 121, 114, 99, 119, 191, 121, 228, 123, 210, 107, 236, 114, 173, 138, 3, 104, 97, 106, 248, 81, 129, 122, 52, 105, 74, 92, 246, 156, 235, 130, 197, 91, 73, 145, 30, 112, 120, 86, 111, 92, 199, 96, 102, 101, 140, 108, 90, 140, 65, 144, 19, 152, 81, 84, 199, 102, 13, 146, 72, 89, 163, 144, 133, 81, 77, 78, 234, 81, 153, 133, 14, 139, 88, 112, 122, 99, 75, 147, 98, 105, 180, 153, 4, 126, 119, 117, 87, 83, 96, 105, 223, 142, 227, 150, 93, 108, 140, 78, 60, 92, 16, 95, 233, 143, 2, 83, 209, 140, 137, 128, 121, 134, 255, 94, 229, 101, 115, 78, 101, 81, 130, 89, 63, 92, 238, 151, 251, 78, 138, 89, 205, 95, 141, 138, 225, 111, 176, 121, 98, 121, 231, 91, 113, 132, 43, 115, 177, 113, 116, 94, 245, 95, 123, 99, 154, 100, 195, 113, 152, 124, 67, 78, 252, 94, 75, 78, 220, 87, 162, 86, 169, 96, 195, 111, 13, 125, 253, 128, 51, 129, 191, 129, 178, 143, 151, 137, 164, 134, 244, 93, 138, 98, 173, 100, 135, 137, 119, 103, 226, 108, 62, 109, 54, 116, 52, 120, 70, 90, 117, 127, 173, 130, 172, 153, 243, 79, 195, 94, 221, 98, 146, 99, 87, 101, 111, 103, 195, 118, 76, 114, 204, 128, 186, 128, 41, 143, 77, 145, 13, 80, 249, 87, 146, 90, 133, 104, 115, 105, 100, 113, 253, 114, 183, 140, 242, 88, 224, 140, 106, 150, 25, 144, 127, 135, 228, 121, 231, 119, 41, 132, 47, 79, 101, 82, 90, 83, 205, 98, 207, 103, 202, 108, 125, 118, 148, 123, 149, 124, 54, 130, 132, 133, 235, 143, 221, 102, 32, 111, 6, 114, 27, 126, 171, 131, 193, 153, 166, 158, 253, 81, 177, 123, 114, 120, 184, 123, 135, 128, 72, 123, 232, 106, 97, 94, 140, 128, 81, 117, 96, 117, 107, 81, 98, 146, 140, 110, 122, 118, 151, 145, 234, 154, 16, 79, 112, 127, 156, 98, 79, 123, 165, 149, 233, 156, 122, 86, 89, 88, 228, 134, 188, 150, 52, 79, 36, 82, 74, 83, 205, 83, 219, 83, 6, 94, 44, 100, 145, 101, 127, 103, 62, 108, 78, 108, 72, 114, 175, 114, 237, 115, 84, 117, 65, 126, 44, 130, 233, 133, 169, 140, 196, 123, 198, 145, 105, 113, 18, 152, 239, 152, 61, 99, 105, 102, 106, 117, 228, 118, 208, 120, 67, 133, 238, 134, 42, 83, 81, 83, 38, 84, 131, 89, 135, 94, 124, 95, 178, 96, 73, 98, 121, 98, 171, 98, 144, 101, 212, 107, 204, 108, 178, 117, 174, 118, 145, 120, 216, 121, 203, 125, 119, 127, 165, 128, 171, 136, 185, 138, 187, 140, 127, 144, 94, 151, 219, 152, 11, 106, 56, 124, 153, 80, 62, 92, 174, 95, 135, 103, 216, 107, 53, 116, 9, 119, 142, 127, 59, 159, 202, 103, 23, 122, 57, 83, 139, 117, 237, 154, 102, 95, 157, 129, 241, 131, 152, 128, 60, 95, 197, 95, 98, 117, 70, 123, 60, 144, 103, 104, 235, 89, 155, 90, 16, 125, 126, 118, 44, 139, 245, 79, 106, 95, 25, 106, 55, 108, 2, 111, 226, 116, 104, 121, 104, 136, 85, 138, 121, 140, 223, 94, 207, 99, 197, 117, 210, 121, 215, 130, 40, 147, 242, 146, 156, 132, 237, 134, 45, 156, 193, 84, 108, 95, 140, 101, 92, 109, 21, 112, 167, 140, 211, 140, 59, 152, 79, 101, 246, 116, 13, 78, 216, 78, 224, 87, 43, 89, 102, 90, 204, 91, 168, 81, 3, 94, 156, 94, 22, 96, 118, 98, 119, 101, 167, 101, 110, 102, 110, 109, 54, 114, 38, 123, 80, 129, 154, 129, 153, 130, 92, 139, 160, 140, 230, 140, 116, 141, 28, 150, 68, 150, 174, 79, 171, 100, 102, 107, 30, 130, 97, 132, 106, 133, 232, 144, 1, 92, 83, 105, 168, 152, 122, 132, 87, 133, 15, 79, 111, 82, 169, 95, 69, 94, 13, 103, 143, 121, 121, 129, 7, 137, 134, 137, 245, 109, 23, 95, 85, 98, 184, 108, 207, 78, 105, 114, 146, 155, 6, 82, 59, 84, 116, 86, 179, 88, 164, 97, 110, 98, 26, 113, 110, 89, 137, 124, 222, 124, 27, 125, 240, 150, 135, 101, 94, 128, 25, 78, 117, 79, 117, 81, 64, 88, 99, 94, 115, 94, 10, 95, 196, 103, 38, 78, 61, 133, 137, 149, 91, 150, 115, 124, 1, 152, 251, 80, 193, 88, 86, 118, 167, 120, 37, 82, 165, 119, 17, 133, 134, 123, 79, 80, 9, 89, 71, 114, 199, 123, 232, 125, 186, 143, 212, 143, 77, 144, 191, 79, 201, 82, 41, 90, 1, 95, 173, 151, 221, 79, 23, 130, 234, 146, 3, 87, 85, 99, 105, 107, 43, 117, 220, 136, 20, 143, 66, 122, 223, 82, 147, 88, 85, 97, 10, 98, 174, 102, 205, 107, 63, 124, 233, 131, 35, 80, 248, 79, 5, 83, 70, 84, 49, 88, 73, 89, 157, 91, 240, 92, 239, 92, 41, 93, 150, 94, 177, 98, 103, 99, 62, 101, 185, 101, 11, 103, 213, 108, 225, 108, 249, 112, 50, 120, 43, 126, 222, 128, 179, 130, 12, 132, 236, 132, 2, 135, 18, 137, 42, 138, 74, 140, 166, 144, 210, 146, 253, 152, 243, 156, 108, 157, 79, 78, 161, 78, 141, 80, 86, 82, 74, 87, 168, 89, 61, 94, 216, 95, 217, 95, 63, 98, 180, 102, 27, 103, 208, 103, 210, 104, 146, 81, 33, 125, 170, 128, 168, 129, 0, 139, 140, 140, 191, 140, 126, 146, 50, 150, 32, 84, 44, 152, 23, 83, 213, 80, 92, 83, 168, 88, 178, 100, 52, 103, 103, 114, 102, 119, 70, 122, 230, 145, 195, 82, 161, 108, 134, 107, 0, 88, 76, 94, 84, 89, 44, 103, 251, 127, 225, 81, 198, 118, 105, 100, 232, 120, 84, 155, 187, 158, 203, 87, 185, 89, 39, 102, 154, 103, 206, 107, 233, 84, 217, 105, 85, 94, 156, 129, 149, 103, 170, 155, 254, 103, 82, 156, 93, 104, 166, 78, 227, 79, 200, 83, 185, 98, 43, 103, 171, 108, 196, 143, 173, 79, 109, 126, 191, 158, 7, 78, 98, 97, 128, 110, 43, 111, 19, 133, 115, 84, 42, 103, 69, 155, 243, 93, 149, 123, 172, 92, 198, 91, 28, 135, 74, 110, 209, 132, 20, 122, 8, 129, 153, 89, 141, 124, 17, 108, 32, 119, 217, 82, 34, 89, 33, 113, 95, 114, 219, 119, 39, 151, 97, 157, 11, 105, 127, 90, 24, 90, 165, 81, 13, 84, 125, 84, 14, 102, 223, 118, 247, 143, 152, 146, 244, 156, 234, 89, 93, 114, 197, 110, 77, 81, 201, 104, 191, 125, 236, 125, 98, 151, 186, 158, 120, 100, 33, 106, 2, 131, 132, 89, 95, 91, 219, 107, 27, 115, 242, 118, 178, 125, 23, 128, 153, 132, 50, 81, 40, 103, 217, 158, 238, 118, 98, 103, 255, 82, 5, 153, 36, 92, 59, 98, 126, 124, 176, 140, 79, 85, 182, 96, 11, 125, 128, 149, 1, 83, 95, 78, 182, 81, 28, 89, 58, 114, 54, 128, 206, 145, 37, 95, 226, 119, 132, 83, 121, 95, 4, 125, 172, 133, 51, 138, 141, 142, 86, 151, 243, 103, 174, 133, 83, 148, 9, 97, 8, 97, 185, 108, 82, 118, 237, 138, 56, 143, 47, 85, 81, 79, 42, 81, 199, 82, 203, 83, 165, 91, 125, 94, 160, 96, 130, 97, 214, 99, 9, 103, 218, 103, 103, 110, 140, 109, 54, 115, 55, 115, 49, 117, 80, 121, 213, 136, 152, 138, 74, 144, 145, 144, 245, 144, 196, 150, 141, 135, 21, 89, 136, 78, 89, 79, 14, 78, 137, 138, 63, 143, 16, 152, 173, 80, 124, 94, 150, 89, 185, 91, 184, 94, 218, 99, 250, 99, 193, 100, 220, 102, 74, 105, 216, 105, 11, 109, 182, 110, 148, 113, 40, 117, 175, 122, 138, 127, 0, 128, 73, 132, 201, 132, 129, 137, 33, 139, 10, 142, 101, 144, 125, 150, 10, 153, 126, 97, 145, 98, 50, 107, 131, 108, 116, 109, 204, 127, 252, 127, 192, 109, 133, 127, 186, 135, 248, 136, 101, 103, 177, 131, 60, 152, 247, 150, 27, 109, 97, 125, 61, 132, 106, 145, 113, 78, 117, 83, 80, 93, 4, 107, 235, 111, 205, 133, 45, 134, 167, 137, 41, 82, 15, 84, 101, 92, 78, 103, 168, 104, 6, 116, 131, 116, 226, 117, 207, 136, 225, 136, 204, 145, 226, 150, 120, 150, 139, 95, 135, 115, 203, 122, 78, 132, 160, 99, 101, 117, 137, 82, 65, 109, 156, 110, 9, 116, 89, 117, 107, 120, 146, 124, 134, 150, 220, 122, 141, 159, 182, 79, 110, 97, 197, 101, 92, 134, 134, 78, 174, 78, 218, 80, 33, 78, 204, 81, 238, 91, 153, 101, 129, 104, 188, 109, 31, 115, 66, 118, 173, 119, 28, 122, 231, 124, 111, 130, 210, 138, 124, 144, 207, 145, 117, 150, 24, 152, 155, 82, 209, 125, 43, 80, 152, 83, 151, 103, 203, 109, 208, 113, 51, 116, 232, 129, 42, 143, 163, 150, 87, 156, 159, 158, 96, 116, 65, 88, 153, 109, 47, 125, 94, 152, 228, 78, 54, 79, 139, 79, 183, 81, 177, 82, 186, 93, 28, 96, 178, 115, 60, 121, 211, 130, 52, 146, 183, 150, 246, 150, 10, 151, 151, 158, 98, 159, 166, 102, 116, 107, 23, 82, 163, 82, 200, 112, 194, 136, 201, 94, 75, 96, 144, 97, 35, 111, 73, 113, 62, 124, 244, 125, 111, 128, 238, 132, 35, 144, 44, 147, 66, 84, 111, 155, 211, 106, 137, 112, 194, 140, 239, 141, 50, 151, 180, 82, 65, 90, 202, 94, 4, 95, 23, 103, 124, 105, 148, 105, 106, 109, 15, 111, 98, 114, 252, 114, 237, 123, 1, 128, 126, 128, 75, 135, 206, 144, 109, 81, 147, 158, 132, 121, 139, 128, 50, 147, 214, 138, 45, 80, 140, 84, 113, 138, 106, 107, 196, 140, 7, 129, 209, 96, 160, 103, 242, 157, 153, 78, 152, 78, 16, 156, 107, 138, 193, 133, 104, 133, 0, 105, 126, 110, 151, 120, 85, 129, 12, 95, 16, 78, 21, 78, 42, 78, 49, 78, 54, 78, 60, 78, 63, 78, 66, 78, 86, 78, 88, 78, 130, 78, 133, 78, 107, 140, 138, 78, 18, 130, 13, 95, 142, 78, 158, 78, 159, 78, 160, 78, 162, 78, 176, 78, 179, 78, 182, 78, 206, 78, 205, 78, 196, 78, 198, 78, 194, 78, 215, 78, 222, 78, 237, 78, 223, 78, 247, 78, 9, 79, 90, 79, 48, 79, 91, 79, 93, 79, 87, 79, 71, 79, 118, 79, 136, 79, 143, 79, 152, 79, 123, 79, 105, 79, 112, 79, 145, 79, 111, 79, 134, 79, 150, 79, 24, 81, 212, 79, 223, 79, 206, 79, 216, 79, 219, 79, 209, 79, 218, 79, 208, 79, 228, 79, 229, 79, 26, 80, 40, 80, 20, 80, 42, 80, 37, 80, 5, 80, 28, 79, 246, 79, 33, 80, 41, 80, 44, 80, 254, 79, 239, 79, 17, 80, 6, 80, 67, 80, 71, 80, 3, 103, 85, 80, 80, 80, 72, 80, 90, 80, 86, 80, 108, 80, 120, 80, 128, 80, 154, 80, 133, 80, 180, 80, 178, 80, 201, 80, 202, 80, 179, 80, 194, 80, 214, 80, 222, 80, 229, 80, 237, 80, 227, 80, 238, 80, 249, 80, 245, 80, 9, 81, 1, 81, 2, 81, 22, 81, 21, 81, 20, 81, 26, 81, 33, 81, 58, 81, 55, 81, 60, 81, 59, 81, 63, 81, 64, 81, 82, 81, 76, 81, 84, 81, 98, 81, 248, 122, 105, 81, 106, 81, 110, 81, 128, 81, 130, 81, 216, 86, 140, 81, 137, 81, 143, 81, 145, 81, 147, 81, 149, 81, 150, 81, 164, 81, 166, 81, 162, 81, 169, 81, 170, 81, 171, 81, 179, 81, 177, 81, 178, 81, 176, 81, 181, 81, 189, 81, 197, 81, 201, 81, 219, 81, 224, 81, 85, 134, 233, 81, 237, 81, 240, 81, 245, 81, 254, 81, 4, 82, 11, 82, 20, 82, 14, 82, 39, 82, 42, 82, 46, 82, 51, 82, 57, 82, 79, 82, 68, 82, 75, 82, 76, 82, 94, 82, 84, 82, 106, 82, 116, 82, 105, 82, 115, 82, 127, 82, 125, 82, 141, 82, 148, 82, 146, 82, 113, 82, 136, 82, 145, 82, 168, 143, 167, 143, 172, 82, 173, 82, 188, 82, 181, 82, 193, 82, 205, 82, 215, 82, 222, 82, 227, 82, 230, 82, 237, 152, 224, 82, 243, 82, 245, 82, 248, 82, 249, 82, 6, 83, 8, 83, 56, 117, 13, 83, 16, 83, 15, 83, 21, 83, 26, 83, 35, 83, 47, 83, 49, 83, 51, 83, 56, 83, 64, 83, 70, 83, 69, 83, 23, 78, 73, 83, 77, 83, 214, 81, 94, 83, 105, 83, 110, 83, 24, 89, 123, 83, 119, 83, 130, 83, 150, 83, 160, 83, 166, 83, 165, 83, 174, 83, 176, 83, 182, 83, 195, 83, 18, 124, 217, 150, 223, 83, 252, 102, 238, 113, 238, 83, 232, 83, 237, 83, 250, 83, 1, 84, 61, 84, 64, 84, 44, 84, 45, 84, 60, 84, 46, 84, 54, 84, 41, 84, 29, 84, 78, 84, 143, 84, 117, 84, 142, 84, 95, 84, 113, 84, 119, 84, 112, 84, 146, 84, 123, 84, 128, 84, 118, 84, 132, 84, 144, 84, 134, 84, 199, 84, 162, 84, 184, 84, 165, 84, 172, 84, 196, 84, 200, 84, 168, 84, 171, 84, 194, 84, 164, 84, 190, 84, 188, 84, 216, 84, 229, 84, 230, 84, 15, 85, 20, 85, 253, 84, 238, 84, 237, 84, 250, 84, 226, 84, 57, 85, 64, 85, 99, 85, 76, 85, 46, 85, 92, 85, 69, 85, 86, 85, 87, 85, 56, 85, 51, 85, 93, 85, 153, 85, 128, 85, 175, 84, 138, 85, 159, 85, 123, 85, 126, 85, 152, 85, 158, 85, 174, 85, 124, 85, 131, 85, 169, 85, 135, 85, 168, 85, 218, 85, 197, 85, 223, 85, 196, 85, 220, 85, 228, 85, 212, 85, 20, 86, 247, 85, 22, 86, 254, 85, 253, 85, 27, 86, 249, 85, 78, 86, 80, 86, 223, 113, 52, 86, 54, 86, 50, 86, 56, 86, 107, 86, 100, 86, 47, 86, 108, 86, 106, 86, 134, 86, 128, 86, 138, 86, 160, 86, 148, 86, 143, 86, 165, 86, 174, 86, 182, 86, 180, 86, 194, 86, 188, 86, 193, 86, 195, 86, 192, 86, 200, 86, 206, 86, 209, 86, 211, 86, 215, 86, 238, 86, 249, 86, 0, 87, 255, 86, 4, 87, 9, 87, 8, 87, 11, 87, 13, 87, 19, 87, 24, 87, 22, 87, 199, 85, 28, 87, 38, 87, 55, 87, 56, 87, 78, 87, 59, 87, 64, 87, 79, 87, 105, 87, 192, 87, 136, 87, 97, 87, 127, 87, 137, 87, 147, 87, 160, 87, 179, 87, 164, 87, 170, 87, 176, 87, 195, 87, 198, 87, 212, 87, 210, 87, 211, 87, 10, 88, 214, 87, 227, 87, 11, 88, 25, 88, 29, 88, 114, 88, 33, 88, 98, 88, 75, 88, 112, 88, 192, 107, 82, 88, 61, 88, 121, 88, 133, 88, 185, 88, 159, 88, 171, 88, 186, 88, 222, 88, 187, 88, 184, 88, 174, 88, 197, 88, 211, 88, 209, 88, 215, 88, 217, 88, 216, 88, 229, 88, 220, 88, 228, 88, 223, 88, 239, 88, 250, 88, 249, 88, 251, 88, 252, 88, 253, 88, 2, 89, 10, 89, 16, 89, 27, 89, 166, 104, 37, 89, 44, 89, 45, 89, 50, 89, 56, 89, 62, 89, 210, 122, 85, 89, 80, 89, 78, 89, 90, 89, 88, 89, 98, 89, 96, 89, 103, 89, 108, 89, 105, 89, 120, 89, 129, 89, 157, 89, 94, 79, 171, 79, 163, 89, 178, 89, 198, 89, 232, 89, 220, 89, 141, 89, 217, 89, 218, 89, 37, 90, 31, 90, 17, 90, 28, 90, 9, 90, 26, 90, 64, 90, 108, 90, 73, 90, 53, 90, 54, 90, 98, 90, 106, 90, 154, 90, 188, 90, 190, 90, 203, 90, 194, 90, 189, 90, 227, 90, 215, 90, 230, 90, 233, 90, 214, 90, 250, 90, 251, 90, 12, 91, 11, 91, 22, 91, 50, 91, 208, 90, 42, 91, 54, 91, 62, 91, 67, 91, 69, 91, 64, 91, 81, 91, 85, 91, 90, 91, 91, 91, 101, 91, 105, 91, 112, 91, 115, 91, 117, 91, 120, 91, 136, 101, 122, 91, 128, 91, 131, 91, 166, 91, 184, 91, 195, 91, 199, 91, 201, 91, 212, 91, 208, 91, 228, 91, 230, 91, 226, 91, 222, 91, 229, 91, 235, 91, 240, 91, 246, 91, 243, 91, 5, 92, 7, 92, 8, 92, 13, 92, 19, 92, 32, 92, 34, 92, 40, 92, 56, 92, 57, 92, 65, 92, 70, 92, 78, 92, 83, 92, 80, 92, 79, 92, 113, 91, 108, 92, 110, 92, 98, 78, 118, 92, 121, 92, 140, 92, 145, 92, 148, 92, 155, 89, 171, 92, 187, 92, 182, 92, 188, 92, 183, 92, 197, 92, 190, 92, 199, 92, 217, 92, 233, 92, 253, 92, 250, 92, 237, 92, 140, 93, 234, 92, 11, 93, 21, 93, 23, 93, 92, 93, 31, 93, 27, 93, 17, 93, 20, 93, 34, 93, 26, 93, 25, 93, 24, 93, 76, 93, 82, 93, 78, 93, 75, 93, 108, 93, 115, 93, 118, 93, 135, 93, 132, 93, 130, 93, 162, 93, 157, 93, 172, 93, 174, 93, 189, 93, 144, 93, 183, 93, 188, 93, 201, 93, 205, 93, 211, 93, 210, 93, 214, 93, 219, 93, 235, 93, 242, 93, 245, 93, 11, 94, 26, 94, 25, 94, 17, 94, 27, 94, 54, 94, 55, 94, 68, 94, 67, 94, 64, 94, 78, 94, 87, 94, 84, 94, 95, 94, 98, 94, 100, 94, 71, 94, 117, 94, 118, 94, 122, 94, 188, 158, 127, 94, 160, 94, 193, 94, 194, 94, 200, 94, 208, 94, 207, 94, 214, 94, 227, 94, 221, 94, 218, 94, 219, 94, 226, 94, 225, 94, 232, 94, 233, 94, 236, 94, 241, 94, 243, 94, 240, 94, 244, 94, 248, 94, 254, 94, 3, 95, 9, 95, 93, 95, 92, 95, 11, 95, 17, 95, 22, 95, 41, 95, 45, 95, 56, 95, 65, 95, 72, 95, 76, 95, 78, 95, 47, 95, 81, 95, 86, 95, 87, 95, 89, 95, 97, 95, 109, 95, 115, 95, 119, 95, 131, 95, 130, 95, 127, 95, 138, 95, 136, 95, 145, 95, 135, 95, 158, 95, 153, 95, 152, 95, 160, 95, 168, 95, 173, 95, 188, 95, 214, 95, 251, 95, 228, 95, 248, 95, 241, 95, 221, 95, 179, 96, 255, 95, 33, 96, 96, 96, 25, 96, 16, 96, 41, 96, 14, 96, 49, 96, 27, 96, 21, 96, 43, 96, 38, 96, 15, 96, 58, 96, 90, 96, 65, 96, 106, 96, 119, 96, 95, 96, 74, 96, 70, 96, 77, 96, 99, 96, 67, 96, 100, 96, 66, 96, 108, 96, 107, 96, 89, 96, 129, 96, 141, 96, 231, 96, 131, 96, 154, 96, 132, 96, 155, 96, 150, 96, 151, 96, 146, 96, 167, 96, 139, 96, 225, 96, 184, 96, 224, 96, 211, 96, 180, 96, 240, 95, 189, 96, 198, 96, 181, 96, 216, 96, 77, 97, 21, 97, 6, 97, 246, 96, 247, 96, 0, 97, 244, 96, 250, 96, 3, 97, 33, 97, 251, 96, 241, 96, 13, 97, 14, 97, 71, 97, 62, 97, 40, 97, 39, 97, 74, 97, 63, 97, 60, 97, 44, 97, 52, 97, 61, 97, 66, 97, 68, 97, 115, 97, 119, 97, 88, 97, 89, 97, 90, 97, 107, 97, 116, 97, 111, 97, 101, 97, 113, 97, 95, 97, 93, 97, 83, 97, 117, 97, 153, 97, 150, 97, 135, 97, 172, 97, 148, 97, 154, 97, 138, 97, 145, 97, 171, 97, 174, 97, 204, 97, 202, 97, 201, 97, 247, 97, 200, 97, 195, 97, 198, 97, 186, 97, 203, 97, 121, 127, 205, 97, 230, 97, 227, 97, 246, 97, 250, 97, 244, 97, 255, 97, 253, 97, 252, 97, 254, 97, 0, 98, 8, 98, 9, 98, 13, 98, 12, 98, 20, 98, 27, 98, 30, 98, 33, 98, 42, 98, 46, 98, 48, 98, 50, 98, 51, 98, 65, 98, 78, 98, 94, 98, 99, 98, 91, 98, 96, 98, 104, 98, 124, 98, 130, 98, 137, 98, 126, 98, 146, 98, 147, 98, 150, 98, 212, 98, 131, 98, 148, 98, 215, 98, 209, 98, 187, 98, 207, 98, 255, 98, 198, 98, 212, 100, 200, 98, 220, 98, 204, 98, 202, 98, 194, 98, 199, 98, 155, 98, 201, 98, 12, 99, 238, 98, 241, 98, 39, 99, 2, 99, 8, 99, 239, 98, 245, 98, 80, 99, 62, 99, 77, 99, 28, 100, 79, 99, 150, 99, 142, 99, 128, 99, 171, 99, 118, 99, 163, 99, 143, 99, 137, 99, 159, 99, 181, 99, 107, 99, 105, 99, 190, 99, 233, 99, 192, 99, 198, 99, 227, 99, 201, 99, 210, 99, 246, 99, 196, 99, 22, 100, 52, 100, 6, 100, 19, 100, 38, 100, 54, 100, 29, 101, 23, 100, 40, 100, 15, 100, 103, 100, 111, 100, 118, 100, 78, 100, 42, 101, 149, 100, 147, 100, 165, 100, 169, 100, 136, 100, 188, 100, 218, 100, 210, 100, 197, 100, 199, 100, 187, 100, 216, 100, 194, 100, 241, 100, 231, 100, 9, 130, 224, 100, 225, 100, 172, 98, 227, 100, 239, 100, 44, 101, 246, 100, 244, 100, 242, 100, 250, 100, 0, 101, 253, 100, 24, 101, 28, 101, 5, 101, 36, 101, 35, 101, 43, 101, 52, 101, 53, 101, 55, 101, 54, 101, 56, 101, 75, 117, 72, 101, 86, 101, 85, 101, 77, 101, 88, 101, 94, 101, 93, 101, 114, 101, 120, 101, 130, 101, 131, 101, 138, 139, 155, 101, 159, 101, 171, 101, 183, 101, 195, 101, 198, 101, 193, 101, 196, 101, 204, 101, 210, 101, 219, 101, 217, 101, 224, 101, 225, 101, 241, 101, 114, 103, 10, 102, 3, 102, 251, 101, 115, 103, 53, 102, 54, 102, 52, 102, 28, 102, 79, 102, 68, 102, 73, 102, 65, 102, 94, 102, 93, 102, 100, 102, 103, 102, 104, 102, 95, 102, 98, 102, 112, 102, 131, 102, 136, 102, 142, 102, 137, 102, 132, 102, 152, 102, 157, 102, 193, 102, 185, 102, 201, 102, 190, 102, 188, 102, 196, 102, 184, 102, 214, 102, 218, 102, 224, 102, 63, 102, 230, 102, 233, 102, 240, 102, 245, 102, 247, 102, 15, 103, 22, 103, 30, 103, 38, 103, 39, 103, 56, 151, 46, 103, 63, 103, 54, 103, 65, 103, 56, 103, 55, 103, 70, 103, 94, 103, 96, 103, 89, 103, 99, 103, 100, 103, 137, 103, 112, 103, 169, 103, 124, 103, 106, 103, 140, 103, 139, 103, 166, 103, 161, 103, 133, 103, 183, 103, 239, 103, 180, 103, 236, 103, 179, 103, 233, 103, 184, 103, 228, 103, 222, 103, 221, 103, 226, 103, 238, 103, 185, 103, 206, 103, 198, 103, 231, 103, 156, 106, 30, 104, 70, 104, 41, 104, 64, 104, 77, 104, 50, 104, 78, 104, 179, 104, 43, 104, 89, 104, 99, 104, 119, 104, 127, 104, 159, 104, 143, 104, 173, 104, 148, 104, 157, 104, 155, 104, 131, 104, 174, 106, 185, 104, 116, 104, 181, 104, 160, 104, 186, 104, 15, 105, 141, 104, 126, 104, 1, 105, 202, 104, 8, 105, 216, 104, 34, 105, 38, 105, 225, 104, 12, 105, 205, 104, 212, 104, 231, 104, 213, 104, 54, 105, 18, 105, 4, 105, 215, 104, 227, 104, 37, 105, 249, 104, 224, 104, 239, 104, 40, 105, 42, 105, 26, 105, 35, 105, 33, 105, 198, 104, 121, 105, 119, 105, 92, 105, 120, 105, 107, 105, 84, 105, 126, 105, 110, 105, 57, 105, 116, 105, 61, 105, 89, 105, 48, 105, 97, 105, 94, 105, 93, 105, 129, 105, 106, 105, 178, 105, 174, 105, 208, 105, 191, 105, 193, 105, 211, 105, 190, 105, 206, 105, 232, 91, 202, 105, 221, 105, 187, 105, 195, 105, 167, 105, 46, 106, 145, 105, 160, 105, 156, 105, 149, 105, 180, 105, 222, 105, 232, 105, 2, 106, 27, 106, 255, 105, 10, 107, 249, 105, 242, 105, 231, 105, 5, 106, 177, 105, 30, 106, 237, 105, 20, 106, 235, 105, 10, 106, 18, 106, 193, 106, 35, 106, 19, 106, 68, 106, 12, 106, 114, 106, 54, 106, 120, 106, 71, 106, 98, 106, 89, 106, 102, 106, 72, 106, 56, 106, 34, 106, 144, 106, 141, 106, 160, 106, 132, 106, 162, 106, 163, 106, 151, 106, 23, 134, 187, 106, 195, 106, 194, 106, 184, 106, 179, 106, 172, 106, 222, 106, 209, 106, 223, 106, 170, 106, 218, 106, 234, 106, 251, 106, 5, 107, 22, 134, 250, 106, 18, 107, 22, 107, 49, 155, 31, 107, 56, 107, 55, 107, 220, 118, 57, 107, 238, 152, 71, 107, 67, 107, 73, 107, 80, 107, 89, 107, 84, 107, 91, 107, 95, 107, 97, 107, 120, 107, 121, 107, 127, 107, 128, 107, 132, 107, 131, 107, 141, 107, 152, 107, 149, 107, 158, 107, 164, 107, 170, 107, 171, 107, 175, 107, 178, 107, 177, 107, 179, 107, 183, 107, 188, 107, 198, 107, 203, 107, 211, 107, 223, 107, 236, 107, 235, 107, 243, 107, 239, 107, 190, 158, 8, 108, 19, 108, 20, 108, 27, 108, 36, 108, 35, 108, 94, 108, 85, 108, 98, 108, 106, 108, 130, 108, 141, 108, 154, 108, 129, 108, 155, 108, 126, 108, 104, 108, 115, 108, 146, 108, 144, 108, 196, 108, 241, 108, 211, 108, 189, 108, 215, 108, 197, 108, 221, 108, 174, 108, 177, 108, 190, 108, 186, 108, 219, 108, 239, 108, 217, 108, 234, 108, 31, 109, 77, 136, 54, 109, 43, 109, 61, 109, 56, 109, 25, 109, 53, 109, 51, 109, 18, 109, 12, 109, 99, 109, 147, 109, 100, 109, 90, 109, 121, 109, 89, 109, 142, 109, 149, 109, 228, 111, 133, 109, 249, 109, 21, 110, 10, 110, 181, 109, 199, 109, 230, 109, 184, 109, 198, 109, 236, 109, 222, 109, 204, 109, 232, 109, 210, 109, 197, 109, 250, 109, 217, 109, 228, 109, 213, 109, 234, 109, 238, 109, 45, 110, 110, 110, 46, 110, 25, 110, 114, 110, 95, 110, 62, 110, 35, 110, 107, 110, 43, 110, 118, 110, 77, 110, 31, 110, 67, 110, 58, 110, 78, 110, 36, 110, 255, 110, 29, 110, 56, 110, 130, 110, 170, 110, 152, 110, 201, 110, 183, 110, 211, 110, 189, 110, 175, 110, 196, 110, 178, 110, 212, 110, 213, 110, 143, 110, 165, 110, 194, 110, 159, 110, 65, 111, 17, 111, 76, 112, 236, 110, 248, 110, 254, 110, 63, 111, 242, 110, 49, 111, 239, 110, 50, 111, 204, 110, 62, 111, 19, 111, 247, 110, 134, 111, 122, 111, 120, 111, 129, 111, 128, 111, 111, 111, 91, 111, 243, 111, 109, 111, 130, 111, 124, 111, 88, 111, 142, 111, 145, 111, 194, 111, 102, 111, 179, 111, 163, 111, 161, 111, 164, 111, 185, 111, 198, 111, 170, 111, 223, 111, 213, 111, 236, 111, 212, 111, 216, 111, 241, 111, 238, 111, 219, 111, 9, 112, 11, 112, 250, 111, 17, 112, 1, 112, 15, 112, 254, 111, 27, 112, 26, 112, 116, 111, 29, 112, 24, 112, 31, 112, 48, 112, 62, 112, 50, 112, 81, 112, 99, 112, 153, 112, 146, 112, 175, 112, 241, 112, 172, 112, 184, 112, 179, 112, 174, 112, 223, 112, 203, 112, 221, 112, 217, 112, 9, 113, 253, 112, 28, 113, 25, 113, 101, 113, 85, 113, 136, 113, 102, 113, 98, 113, 76, 113, 86, 113, 108, 113, 143, 113, 251, 113, 132, 113, 149, 113, 168, 113, 172, 113, 215, 113, 185, 113, 190, 113, 210, 113, 201, 113, 212, 113, 206, 113, 224, 113, 236, 113, 231, 113, 245, 113, 252, 113, 249, 113, 255, 113, 13, 114, 16, 114, 27, 114, 40, 114, 45, 114, 44, 114, 48, 114, 50, 114, 59, 114, 60, 114, 63, 114, 64, 114, 70, 114, 75, 114, 88, 114, 116, 114, 126, 114, 130, 114, 129, 114, 135, 114, 146, 114, 150, 114, 162, 114, 167, 114, 185, 114, 178, 114, 195, 114, 198, 114, 196, 114, 206, 114, 210, 114, 226, 114, 224, 114, 225, 114, 249, 114, 247, 114, 15, 80, 23, 115, 10, 115, 28, 115, 22, 115, 29, 115, 52, 115, 47, 115, 41, 115, 37, 115, 62, 115, 78, 115, 79, 115, 216, 158, 87, 115, 106, 115, 104, 115, 112, 115, 120, 115, 117, 115, 123, 115, 122, 115, 200, 115, 179, 115, 206, 115, 187, 115, 192, 115, 229, 115, 238, 115, 222, 115, 162, 116, 5, 116, 111, 116, 37, 116, 248, 115, 50, 116, 58, 116, 85, 116, 63, 116, 95, 116, 89, 116, 65, 116, 92, 116, 105, 116, 112, 116, 99, 116, 106, 116, 118, 116, 126, 116, 139, 116, 158, 116, 167, 116, 202, 116, 207, 116, 212, 116, 241, 115, 224, 116, 227, 116, 231, 116, 233, 116, 238, 116, 242, 116, 240, 116, 241, 116, 248, 116, 247, 116, 4, 117, 3, 117, 5, 117, 12, 117, 14, 117, 13, 117, 21, 117, 19, 117, 30, 117, 38, 117, 44, 117, 60, 117, 68, 117, 77, 117, 74, 117, 73, 117, 91, 117, 70, 117, 90, 117, 105, 117, 100, 117, 103, 117, 107, 117, 109, 117, 120, 117, 118, 117, 134, 117, 135, 117, 116, 117, 138, 117, 137, 117, 130, 117, 148, 117, 154, 117, 157, 117, 165, 117, 163, 117, 194, 117, 179, 117, 195, 117, 181, 117, 189, 117, 184, 117, 188, 117, 177, 117, 205, 117, 202, 117, 210, 117, 217, 117, 227, 117, 222, 117, 254, 117, 255, 117, 252, 117, 1, 118, 240, 117, 250, 117, 242, 117, 243, 117, 11, 118, 13, 118, 9, 118, 31, 118, 39, 118, 32, 118, 33, 118, 34, 118, 36, 118, 52, 118, 48, 118, 59, 118, 71, 118, 72, 118, 70, 118, 92, 118, 88, 118, 97, 118, 98, 118, 104, 118, 105, 118, 106, 118, 103, 118, 108, 118, 112, 118, 114, 118, 118, 118, 120, 118, 124, 118, 128, 118, 131, 118, 136, 118, 139, 118, 142, 118, 150, 118, 147, 118, 153, 118, 154, 118, 176, 118, 180, 118, 184, 118, 185, 118, 186, 118, 194, 118, 205, 118, 214, 118, 210, 118, 222, 118, 225, 118, 229, 118, 231, 118, 234, 118, 47, 134, 251, 118, 8, 119, 7, 119, 4, 119, 41, 119, 36, 119, 30, 119, 37, 119, 38, 119, 27, 119, 55, 119, 56, 119, 71, 119, 90, 119, 104, 119, 107, 119, 91, 119, 101, 119, 127, 119, 126, 119, 121, 119, 142, 119, 139, 119, 145, 119, 160, 119, 158, 119, 176, 119, 182, 119, 185, 119, 191, 119, 188, 119, 189, 119, 187, 119, 199, 119, 205, 119, 215, 119, 218, 119, 220, 119, 227, 119, 238, 119, 252, 119, 12, 120, 18, 120, 38, 121, 32, 120, 42, 121, 69, 120, 142, 120, 116, 120, 134, 120, 124, 120, 154, 120, 140, 120, 163, 120, 181, 120, 170, 120, 175, 120, 209, 120, 198, 120, 203, 120, 212, 120, 190, 120, 188, 120, 197, 120, 202, 120, 236, 120, 231, 120, 218, 120, 253, 120, 244, 120, 7, 121, 18, 121, 17, 121, 25, 121, 44, 121, 43, 121, 64, 121, 96, 121, 87, 121, 95, 121, 90, 121, 85, 121, 83, 121, 122, 121, 127, 121, 138, 121, 157, 121, 167, 121, 75, 159, 170, 121, 174, 121, 179, 121, 185, 121, 186, 121, 201, 121, 213, 121, 231, 121, 236, 121, 225, 121, 227, 121, 8, 122, 13, 122, 24, 122, 25, 122, 32, 122, 31, 122, 128, 121, 49, 122, 59, 122, 62, 122, 55, 122, 67, 122, 87, 122, 73, 122, 97, 122, 98, 122, 105, 122, 157, 159, 112, 122, 121, 122, 125, 122, 136, 122, 151, 122, 149, 122, 152, 122, 150, 122, 169, 122, 200, 122, 176, 122, 182, 122, 197, 122, 196, 122, 191, 122, 131, 144, 199, 122, 202, 122, 205, 122, 207, 122, 213, 122, 211, 122, 217, 122, 218, 122, 221, 122, 225, 122, 226, 122, 230, 122, 237, 122, 240, 122, 2, 123, 15, 123, 10, 123, 6, 123, 51, 123, 24, 123, 25, 123, 30, 123, 53, 123, 40, 123, 54, 123, 80, 123, 122, 123, 4, 123, 77, 123, 11, 123, 76, 123, 69, 123, 117, 123, 101, 123, 116, 123, 103, 123, 112, 123, 113, 123, 108, 123, 110, 123, 157, 123, 152, 123, 159, 123, 141, 123, 156, 123, 154, 123, 139, 123, 146, 123, 143, 123, 93, 123, 153, 123, 203, 123, 193, 123, 204, 123, 207, 123, 180, 123, 198, 123, 221, 123, 233, 123, 17, 124, 20, 124, 230, 123, 229, 123, 96, 124, 0, 124, 7, 124, 19, 124, 243, 123, 247, 123, 23, 124, 13, 124, 246, 123, 35, 124, 39, 124, 42, 124, 31, 124, 55, 124, 43, 124, 61, 124, 76, 124, 67, 124, 84, 124, 79, 124, 64, 124, 80, 124, 88, 124, 95, 124, 100, 124, 86, 124, 101, 124, 108, 124, 117, 124, 131, 124, 144, 124, 164, 124, 173, 124, 162, 124, 171, 124, 161, 124, 168, 124, 179, 124, 178, 124, 177, 124, 174, 124, 185, 124, 189, 124, 192, 124, 197, 124, 194, 124, 216, 124, 210, 124, 220, 124, 226, 124, 59, 155, 239, 124, 242, 124, 244, 124, 246, 124, 250, 124, 6, 125, 2, 125, 28, 125, 21, 125, 10, 125, 69, 125, 75, 125, 46, 125, 50, 125, 63, 125, 53, 125, 70, 125, 115, 125, 86, 125, 78, 125, 114, 125, 104, 125, 110, 125, 79, 125, 99, 125, 147, 125, 137, 125, 91, 125, 143, 125, 125, 125, 155, 125, 186, 125, 174, 125, 163, 125, 181, 125, 199, 125, 189, 125, 171, 125, 61, 126, 162, 125, 175, 125, 220, 125, 184, 125, 159, 125, 176, 125, 216, 125, 221, 125, 228, 125, 222, 125, 251, 125, 242, 125, 225, 125, 5, 126, 10, 126, 35, 126, 33, 126, 18, 126, 49, 126, 31, 126, 9, 126, 11, 126, 34, 126, 70, 126, 102, 126, 59, 126, 53, 126, 57, 126, 67, 126, 55, 126, 50, 126, 58, 126, 103, 126, 93, 126, 86, 126, 94, 126, 89, 126, 90, 126, 121, 126, 106, 126, 105, 126, 124, 126, 123, 126, 131, 126, 213, 125, 125, 126, 174, 143, 127, 126, 136, 126, 137, 126, 140, 126, 146, 126, 144, 126, 147, 126, 148, 126, 150, 126, 142, 126, 155, 126, 156, 126, 56, 127, 58, 127, 69, 127, 76, 127, 77, 127, 78, 127, 80, 127, 81, 127, 85, 127, 84, 127, 88, 127, 95, 127, 96, 127, 104, 127, 105, 127, 103, 127, 120, 127, 130, 127, 134, 127, 131, 127, 136, 127, 135, 127, 140, 127, 148, 127, 158, 127, 157, 127, 154, 127, 163, 127, 175, 127, 178, 127, 185, 127, 174, 127, 182, 127, 184, 127, 113, 139, 197, 127, 198, 127, 202, 127, 213, 127, 212, 127, 225, 127, 230, 127, 233, 127, 243, 127, 249, 127, 220, 152, 6, 128, 4, 128, 11, 128, 18, 128, 24, 128, 25, 128, 28, 128, 33, 128, 40, 128, 63, 128, 59, 128, 74, 128, 70, 128, 82, 128, 88, 128, 90, 128, 95, 128, 98, 128, 104, 128, 115, 128, 114, 128, 112, 128, 118, 128, 121, 128, 125, 128, 127, 128, 132, 128, 134, 128, 133, 128, 155, 128, 147, 128, 154, 128, 173, 128, 144, 81, 172, 128, 219, 128, 229, 128, 217, 128, 221, 128, 196, 128, 218, 128, 214, 128, 9, 129, 239, 128, 241, 128, 27, 129, 41, 129, 35, 129, 47, 129, 75, 129, 139, 150, 70, 129, 62, 129, 83, 129, 81, 129, 252, 128, 113, 129, 110, 129, 101, 129, 102, 129, 116, 129, 131, 129, 136, 129, 138, 129, 128, 129, 130, 129, 160, 129, 149, 129, 164, 129, 163, 129, 95, 129, 147, 129, 169, 129, 176, 129, 181, 129, 190, 129, 184, 129, 189, 129, 192, 129, 194, 129, 186, 129, 201, 129, 205, 129, 209, 129, 217, 129, 216, 129, 200, 129, 218, 129, 223, 129, 224, 129, 231, 129, 250, 129, 251, 129, 254, 129, 1, 130, 2, 130, 5, 130, 7, 130, 10, 130, 13, 130, 16, 130, 22, 130, 41, 130, 43, 130, 56, 130, 51, 130, 64, 130, 89, 130, 88, 130, 93, 130, 90, 130, 95, 130, 100, 130, 98, 130, 104, 130, 106, 130, 107, 130, 46, 130, 113, 130, 119, 130, 120, 130, 126, 130, 141, 130, 146, 130, 171, 130, 159, 130, 187, 130, 172, 130, 225, 130, 227, 130, 223, 130, 210, 130, 244, 130, 243, 130, 250, 130, 147, 131, 3, 131, 251, 130, 249, 130, 222, 130, 6, 131, 220, 130, 9, 131, 217, 130, 53, 131, 52, 131, 22, 131, 50, 131, 49, 131, 64, 131, 57, 131, 80, 131, 69, 131, 47, 131, 43, 131, 23, 131, 24, 131, 133, 131, 154, 131, 170, 131, 159, 131, 162, 131, 150, 131, 35, 131, 142, 131, 135, 131, 138, 131, 124, 131, 181, 131, 115, 131, 117, 131, 160, 131, 137, 131, 168, 131, 244, 131, 19, 132, 235, 131, 206, 131, 253, 131, 3, 132, 216, 131, 11, 132, 193, 131, 247, 131, 7, 132, 224, 131, 242, 131, 13, 132, 34, 132, 32, 132, 189, 131, 56, 132, 6, 133, 251, 131, 109, 132, 42, 132, 60, 132, 90, 133, 132, 132, 119, 132, 107, 132, 173, 132, 110, 132, 130, 132, 105, 132, 70, 132, 44, 132, 111, 132, 121, 132, 53, 132, 202, 132, 98, 132, 185, 132, 191, 132, 159, 132, 217, 132, 205, 132, 187, 132, 218, 132, 208, 132, 193, 132, 198, 132, 214, 132, 161, 132, 33, 133, 255, 132, 244, 132, 23, 133, 24, 133, 44, 133, 31, 133, 21, 133, 20, 133, 252, 132, 64, 133, 99, 133, 88, 133, 72, 133, 65, 133, 2, 134, 75, 133, 85, 133, 128, 133, 164, 133, 136, 133, 145, 133, 138, 133, 168, 133, 109, 133, 148, 133, 155, 133, 234, 133, 135, 133, 156, 133, 119, 133, 126, 133, 144, 133, 201, 133, 186, 133, 207, 133, 185, 133, 208, 133, 213, 133, 221, 133, 229, 133, 220, 133, 249, 133, 10, 134, 19, 134, 11, 134, 254, 133, 250, 133, 6, 134, 34, 134, 26, 134, 48, 134, 63, 134, 77, 134, 85, 78, 84, 134, 95, 134, 103, 134, 113, 134, 147, 134, 163, 134, 169, 134, 170, 134, 139, 134, 140, 134, 182, 134, 175, 134, 196, 134, 198, 134, 176, 134, 201, 134, 35, 136, 171, 134, 212, 134, 222, 134, 233, 134, 236, 134, 223, 134, 219, 134, 239, 134, 18, 135, 6, 135, 8, 135, 0, 135, 3, 135, 251, 134, 17, 135, 9, 135, 13, 135, 249, 134, 10, 135, 52, 135, 63, 135, 55, 135, 59, 135, 37, 135, 41, 135, 26, 135, 96, 135, 95, 135, 120, 135, 76, 135, 78, 135, 116, 135, 87, 135, 104, 135, 110, 135, 89, 135, 83, 135, 99, 135, 106, 135, 5, 136, 162, 135, 159, 135, 130, 135, 175, 135, 203, 135, 189, 135, 192, 135, 208, 135, 214, 150, 171, 135, 196, 135, 179, 135, 199, 135, 198, 135, 187, 135, 239, 135, 242, 135, 224, 135, 15, 136, 13, 136, 254, 135, 246, 135, 247, 135, 14, 136, 210, 135, 17, 136, 22, 136, 21, 136, 34, 136, 33, 136, 49, 136, 54, 136, 57, 136, 39, 136, 59, 136, 68, 136, 66, 136, 82, 136, 89, 136, 94, 136, 98, 136, 107, 136, 129, 136, 126, 136, 158, 136, 117, 136, 125, 136, 181, 136, 114, 136, 130, 136, 151, 136, 146, 136, 174, 136, 153, 136, 162, 136, 141, 136, 164, 136, 176, 136, 191, 136, 177, 136, 195, 136, 196, 136, 212, 136, 216, 136, 217, 136, 221, 136, 249, 136, 2, 137, 252, 136, 244, 136, 232, 136, 242, 136, 4, 137, 12, 137, 10, 137, 19, 137, 67, 137, 30, 137, 37, 137, 42, 137, 43, 137, 65, 137, 68, 137, 59, 137, 54, 137, 56, 137, 76, 137, 29, 137, 96, 137, 94, 137, 102, 137, 100, 137, 109, 137, 106, 137, 111, 137, 116, 137, 119, 137, 126, 137, 131, 137, 136, 137, 138, 137, 147, 137, 152, 137, 161, 137, 169, 137, 166, 137, 172, 137, 175, 137, 178, 137, 186, 137, 189, 137, 191, 137, 192, 137, 218, 137, 220, 137, 221, 137, 231, 137, 244, 137, 248, 137, 3, 138, 22, 138, 16, 138, 12, 138, 27, 138, 29, 138, 37, 138, 54, 138, 65, 138, 91, 138, 82, 138, 70, 138, 72, 138, 124, 138, 109, 138, 108, 138, 98, 138, 133, 138, 130, 138, 132, 138, 168, 138, 161, 138, 145, 138, 165, 138, 166, 138, 154, 138, 163, 138, 196, 138, 205, 138, 194, 138, 218, 138, 235, 138, 243, 138, 231, 138, 228, 138, 241, 138, 20, 139, 224, 138, 226, 138, 247, 138, 222, 138, 219, 138, 12, 139, 7, 139, 26, 139, 225, 138, 22, 139, 16, 139, 23, 139, 32, 139, 51, 139, 171, 151, 38, 139, 43, 139, 62, 139, 40, 139, 65, 139, 76, 139, 79, 139, 78, 139, 73, 139, 86, 139, 91, 139, 90, 139, 107, 139, 95, 139, 108, 139, 111, 139, 116, 139, 125, 139, 128, 139, 140, 139, 142, 139, 146, 139, 147, 139, 150, 139, 153, 139, 154, 139, 58, 140, 65, 140, 63, 140, 72, 140, 76, 140, 78, 140, 80, 140, 85, 140, 98, 140, 108, 140, 120, 140, 122, 140, 130, 140, 137, 140, 133, 140, 138, 140, 141, 140, 142, 140, 148, 140, 124, 140, 152, 140, 29, 98, 173, 140, 170, 140, 189, 140, 178, 140, 179, 140, 174, 140, 182, 140, 200, 140, 193, 140, 228, 140, 227, 140, 218, 140, 253, 140, 250, 140, 251, 140, 4, 141, 5, 141, 10, 141, 7, 141, 15, 141, 13, 141, 16, 141, 78, 159, 19, 141, 205, 140, 20, 141, 22, 141, 103, 141, 109, 141, 113, 141, 115, 141, 129, 141, 153, 141, 194, 141, 190, 141, 186, 141, 207, 141, 218, 141, 214, 141, 204, 141, 219, 141, 203, 141, 234, 141, 235, 141, 223, 141, 227, 141, 252, 141, 8, 142, 9, 142, 255, 141, 29, 142, 30, 142, 16, 142, 31, 142, 66, 142, 53, 142, 48, 142, 52, 142, 74, 142, 71, 142, 73, 142, 76, 142, 80, 142, 72, 142, 89, 142, 100, 142, 96, 142, 42, 142, 99, 142, 85, 142, 118, 142, 114, 142, 124, 142, 129, 142, 135, 142, 133, 142, 132, 142, 139, 142, 138, 142, 147, 142, 145, 142, 148, 142, 153, 142, 170, 142, 161, 142, 172, 142, 176, 142, 198, 142, 177, 142, 190, 142, 197, 142, 200, 142, 203, 142, 219, 142, 227, 142, 252, 142, 251, 142, 235, 142, 254, 142, 10, 143, 5, 143, 21, 143, 18, 143, 25, 143, 19, 143, 28, 143, 31, 143, 27, 143, 12, 143, 38, 143, 51, 143, 59, 143, 57, 143, 69, 143, 66, 143, 62, 143, 76, 143, 73, 143, 70, 143, 78, 143, 87, 143, 92, 143, 98, 143, 99, 143, 100, 143, 156, 143, 159, 143, 163, 143, 173, 143, 175, 143, 183, 143, 218, 143, 229, 143, 226, 143, 234, 143, 239, 143, 135, 144, 244, 143, 5, 144, 249, 143, 250, 143, 17, 144, 21, 144, 33, 144, 13, 144, 30, 144, 22, 144, 11, 144, 39, 144, 54, 144, 53, 144, 57, 144, 248, 143, 79, 144, 80, 144, 81, 144, 82, 144, 14, 144, 73, 144, 62, 144, 86, 144, 88, 144, 94, 144, 104, 144, 111, 144, 118, 144, 168, 150, 114, 144, 130, 144, 125, 144, 129, 144, 128, 144, 138, 144, 137, 144, 143, 144, 168, 144, 175, 144, 177, 144, 181, 144, 226, 144, 228, 144, 72, 98, 219, 144, 2, 145, 18, 145, 25, 145, 50, 145, 48, 145, 74, 145, 86, 145, 88, 145, 99, 145, 101, 145, 105, 145, 115, 145, 114, 145, 139, 145, 137, 145, 130, 145, 162, 145, 171, 145, 175, 145, 170, 145, 181, 145, 180, 145, 186, 145, 192, 145, 193, 145, 201, 145, 203, 145, 208, 145, 214, 145, 223, 145, 225, 145, 219, 145, 252, 145, 245, 145, 246, 145, 30, 146, 255, 145, 20, 146, 44, 146, 21, 146, 17, 146, 94, 146, 87, 146, 69, 146, 73, 146, 100, 146, 72, 146, 149, 146, 63, 146, 75, 146, 80, 146, 156, 146, 150, 146, 147, 146, 155, 146, 90, 146, 207, 146, 185, 146, 183, 146, 233, 146, 15, 147, 250, 146, 68, 147, 46, 147, 25, 147, 34, 147, 26, 147, 35, 147, 58, 147, 53, 147, 59, 147, 92, 147, 96, 147, 124, 147, 110, 147, 86, 147, 176, 147, 172, 147, 173, 147, 148, 147, 185, 147, 214, 147, 215, 147, 232, 147, 229, 147, 216, 147, 195, 147, 221, 147, 208, 147, 200, 147, 228, 147, 26, 148, 20, 148, 19, 148, 3, 148, 7, 148, 16, 148, 54, 148, 43, 148, 53, 148, 33, 148, 58, 148, 65, 148, 82, 148, 68, 148, 91, 148, 96, 148, 98, 148, 94, 148, 106, 148, 41, 146, 112, 148, 117, 148, 119, 148, 125, 148, 90, 148, 124, 148, 126, 148, 129, 148, 127, 148, 130, 149, 135, 149, 138, 149, 148, 149, 150, 149, 152, 149, 153, 149, 160, 149, 168, 149, 167, 149, 173, 149, 188, 149, 187, 149, 185, 149, 190, 149, 202, 149, 246, 111, 195, 149, 205, 149, 204, 149, 213, 149, 212, 149, 214, 149, 220, 149, 225, 149, 229, 149, 226, 149, 33, 150, 40, 150, 46, 150, 47, 150, 66, 150, 76, 150, 79, 150, 75, 150, 119, 150, 92, 150, 94, 150, 93, 150, 95, 150, 102, 150, 114, 150, 108, 150, 141, 150, 152, 150, 149, 150, 151, 150, 170, 150, 167, 150, 177, 150, 178, 150, 176, 150, 180, 150, 182, 150, 184, 150, 185, 150, 206, 150, 203, 150, 201, 150, 205, 150, 77, 137, 220, 150, 13, 151, 213, 150, 249, 150, 4, 151, 6, 151, 8, 151, 19, 151, 14, 151, 17, 151, 15, 151, 22, 151, 25, 151, 36, 151, 42, 151, 48, 151, 57, 151, 61, 151, 62, 151, 68, 151, 70, 151, 72, 151, 66, 151, 73, 151, 92, 151, 96, 151, 100, 151, 102, 151, 104, 151, 210, 82, 107, 151, 113, 151, 121, 151, 133, 151, 124, 151, 129, 151, 122, 151, 134, 151, 139, 151, 143, 151, 144, 151, 156, 151, 168, 151, 166, 151, 163, 151, 179, 151, 180, 151, 195, 151, 198, 151, 200, 151, 203, 151, 220, 151, 237, 151, 79, 159, 242, 151, 223, 122, 246, 151, 245, 151, 15, 152, 12, 152, 56, 152, 36, 152, 33, 152, 55, 152, 61, 152, 70, 152, 79, 152, 75, 152, 107, 152, 111, 152, 112, 152, 113, 152, 116, 152, 115, 152, 170, 152, 175, 152, 177, 152, 182, 152, 196, 152, 195, 152, 198, 152, 233, 152, 235, 152, 3, 153, 9, 153, 18, 153, 20, 153, 24, 153, 33, 153, 29, 153, 30, 153, 36, 153, 32, 153, 44, 153, 46, 153, 61, 153, 62, 153, 66, 153, 73, 153, 69, 153, 80, 153, 75, 153, 81, 153, 82, 153, 76, 153, 85, 153, 151, 153, 152, 153, 165, 153, 173, 153, 174, 153, 188, 153, 223, 153, 219, 153, 221, 153, 216, 153, 209, 153, 237, 153, 238, 153, 241, 153, 242, 153, 251, 153, 248, 153, 1, 154, 15, 154, 5, 154, 226, 153, 25, 154, 43, 154, 55, 154, 69, 154, 66, 154, 64, 154, 67, 154, 62, 154, 85, 154, 77, 154, 91, 154, 87, 154, 95, 154, 98, 154, 101, 154, 100, 154, 105, 154, 107, 154, 106, 154, 173, 154, 176, 154, 188, 154, 192, 154, 207, 154, 209, 154, 211, 154, 212, 154, 222, 154, 223, 154, 226, 154, 227, 154, 230, 154, 239, 154, 235, 154, 238, 154, 244, 154, 241, 154, 247, 154, 251, 154, 6, 155, 24, 155, 26, 155, 31, 155, 34, 155, 35, 155, 37, 155, 39, 155, 40, 155, 41, 155, 42, 155, 46, 155, 47, 155, 50, 155, 68, 155, 67, 155, 79, 155, 77, 155, 78, 155, 81, 155, 88, 155, 116, 155, 147, 155, 131, 155, 145, 155, 150, 155, 151, 155, 159, 155, 160, 155, 168, 155, 180, 155, 192, 155, 202, 155, 185, 155, 198, 155, 207, 155, 209, 155, 210, 155, 227, 155, 226, 155, 228, 155, 212, 155, 225, 155, 58, 156, 242, 155, 241, 155, 240, 155, 21, 156, 20, 156, 9, 156, 19, 156, 12, 156, 6, 156, 8, 156, 18, 156, 10, 156, 4, 156, 46, 156, 27, 156, 37, 156, 36, 156, 33, 156, 48, 156, 71, 156, 50, 156, 70, 156, 62, 156, 90, 156, 96, 156, 103, 156, 118, 156, 120, 156, 231, 156, 236, 156, 240, 156, 9, 157, 8, 157, 235, 156, 3, 157, 6, 157, 42, 157, 38, 157, 175, 157, 35, 157, 31, 157, 68, 157, 21, 157, 18, 157, 65, 157, 63, 157, 62, 157, 70, 157, 72, 157, 93, 157, 94, 157, 100, 157, 81, 157, 80, 157, 89, 157, 114, 157, 137, 157, 135, 157, 171, 157, 111, 157, 122, 157, 154, 157, 164, 157, 169, 157, 178, 157, 196, 157, 193, 157, 187, 157, 184, 157, 186, 157, 198, 157, 207, 157, 194, 157, 217, 157, 211, 157, 248, 157, 230, 157, 237, 157, 239, 157, 253, 157, 26, 158, 27, 158, 30, 158, 117, 158, 121, 158, 125, 158, 129, 158, 136, 158, 139, 158, 140, 158, 146, 158, 149, 158, 145, 158, 157, 158, 165, 158, 169, 158, 184, 158, 170, 158, 173, 158, 97, 151, 204, 158, 206, 158, 207, 158, 208, 158, 212, 158, 220, 158, 222, 158, 221, 158, 224, 158, 229, 158, 232, 158, 239, 158, 244, 158, 246, 158, 247, 158, 249, 158, 251, 158, 252, 158, 253, 158, 7, 159, 8, 159, 183, 118, 21, 159, 33, 159, 44, 159, 62, 159, 74, 159, 82, 159, 84, 159, 99, 159, 95, 159, 96, 159, 97, 159, 102, 159, 103, 159, 108, 159, 106, 159, 119, 159, 114, 159, 118, 159, 149, 159, 156, 159, 160, 159, 47, 88, 199, 105, 89, 144, 100, 116, 220, 81, 153, 113, 0, 0, 2, 0, 64, 0, 32, 0, 69, 0, 1, 0, 170, 170, 38, 141, 1, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 15, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 60, 0, 0, 0, 72, 0, 0, 0, 72, 1, 0, 0, 136, 103, 1, 0, 136, 103, 1, 0, 112, 112, 1, 0, 112, 112, 1, 0, 108, 114, 1, 0, 108, 114, 1, 0, 14, 141, 1, 0, 25, 0, 0, 0, 22, 0, 0, 0, 17, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 94, 249, 250, 251, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 254, 255, 50, 105, 114, 84, 1, 0, 192, 8, 75, 21, 64, 0, 192, 0, 96, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 48, 0, 48, 0, 48, 0, 97, 0, 105, 0, 113, 0, 121, 0, 48, 0, 129, 0, 137, 0, 145, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 147, 0, 48, 0, 48, 0, 48, 0, 155, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 163, 0, 171, 0, 179, 0, 187, 0, 195, 0, 203, 0, 211, 0, 219, 0, 227, 0, 235, 0, 243, 0, 251, 0, 3, 1, 11, 1, 19, 1, 27, 1, 35, 1, 43, 1, 51, 1, 59, 1, 67, 1, 74, 1, 82, 1, 90, 1, 98, 1, 106, 1, 114, 1, 122, 1, 130, 1, 138, 1, 146, 1, 154, 1, 162, 1, 170, 1, 178, 1, 186, 1, 194, 1, 202, 1, 210, 1, 218, 1, 226, 1, 234, 1, 242, 1, 250, 1, 2, 2, 10, 2, 18, 2, 26, 2, 34, 2, 42, 2, 50, 2, 58, 2, 66, 2, 74, 2, 82, 2, 90, 2, 98, 2, 104, 2, 112, 2, 120, 2, 128, 2, 136, 2, 143, 2, 151, 2, 159, 2, 166, 2, 174, 2, 181, 2, 189, 2, 197, 2, 205, 2, 213, 2, 221, 2, 229, 2, 237, 2, 245, 2, 253, 2, 5, 3, 13, 3, 21, 3, 29, 3, 37, 3, 45, 3, 53, 3, 61, 3, 69, 3, 77, 3, 85, 3, 93, 3, 101, 3, 109, 3, 117, 3, 125, 3, 133, 3, 141, 3, 149, 3, 157, 3, 165, 3, 173, 3, 181, 3, 189, 3, 196, 3, 204, 3, 212, 3, 219, 3, 227, 3, 235, 3, 243, 3, 251, 3, 3, 4, 11, 4, 19, 4, 27, 4, 35, 4, 43, 4, 51, 4, 58, 4, 66, 4, 74, 4, 82, 4, 90, 4, 98, 4, 104, 4, 112, 4, 120, 4, 128, 4, 136, 4, 144, 4, 152, 4, 160, 4, 168, 4, 176, 4, 184, 4, 192, 4, 200, 4, 208, 4, 216, 4, 224, 4, 232, 4, 240, 4, 248, 4, 0, 5, 8, 5, 16, 5, 24, 5, 32, 5, 40, 5, 48, 5, 56, 5, 64, 5, 72, 5, 80, 5, 88, 5, 96, 5, 104, 5, 112, 5, 120, 5, 127, 5, 135, 5, 143, 5, 151, 5, 159, 5, 167, 5, 175, 5, 183, 5, 191, 5, 199, 5, 207, 5, 215, 5, 223, 5, 231, 5, 239, 5, 247, 5, 255, 5, 7, 6, 15, 6, 22, 6, 30, 6, 38, 6, 45, 6, 53, 6, 61, 6, 69, 6, 77, 6, 85, 6, 93, 6, 101, 6, 109, 6, 117, 6, 125, 6, 133, 6, 141, 6, 149, 6, 157, 6, 165, 6, 173, 6, 181, 6, 189, 6, 197, 6, 205, 6, 213, 6, 221, 6, 229, 6, 237, 6, 245, 6, 253, 6, 5, 7, 13, 7, 21, 7, 29, 7, 37, 7, 45, 7, 53, 7, 61, 7, 69, 7, 77, 7, 85, 7, 93, 7, 101, 7, 109, 7, 117, 7, 125, 7, 133, 7, 141, 7, 149, 7, 157, 7, 165, 7, 173, 7, 181, 7, 189, 7, 197, 7, 205, 7, 212, 7, 220, 7, 228, 7, 236, 7, 244, 7, 252, 7, 4, 8, 12, 8, 18, 8, 26, 8, 34, 8, 42, 8, 50, 8, 58, 8, 66, 8, 74, 8, 80, 8, 88, 8, 96, 8, 104, 8, 111, 8, 119, 8, 127, 8, 135, 8, 142, 8, 150, 8, 158, 8, 166, 8, 174, 8, 182, 8, 190, 8, 198, 8, 206, 8, 214, 8, 222, 8, 230, 8, 238, 8, 246, 8, 254, 8, 6, 9, 14, 9, 22, 9, 30, 9, 38, 9, 46, 9, 53, 9, 60, 9, 68, 9, 76, 9, 84, 9, 92, 9, 100, 9, 108, 9, 116, 9, 124, 9, 132, 9, 140, 9, 147, 9, 155, 9, 163, 9, 171, 9, 179, 9, 187, 9, 195, 9, 203, 9, 211, 9, 219, 9, 227, 9, 234, 9, 240, 9, 248, 9, 0, 10, 8, 10, 16, 10, 24, 10, 32, 10, 40, 10, 48, 10, 56, 10, 64, 10, 72, 10, 80, 10, 88, 10, 96, 10, 104, 10, 112, 10, 120, 10, 128, 10, 136, 10, 144, 10, 152, 10, 160, 10, 168, 10, 175, 10, 183, 10, 191, 10, 199, 10, 207, 10, 215, 10, 223, 10, 231, 10, 239, 10, 247, 10, 255, 10, 7, 11, 15, 11, 23, 11, 31, 11, 39, 11, 47, 11, 55, 11, 62, 11, 70, 11, 77, 11, 85, 11, 93, 11, 101, 11, 109, 11, 117, 11, 125, 11, 133, 11, 141, 11, 149, 11, 157, 11, 165, 11, 173, 11, 181, 11, 189, 11, 197, 11, 205, 11, 212, 11, 220, 11, 227, 11, 235, 11, 243, 11, 251, 11, 3, 12, 11, 12, 19, 12, 27, 12, 35, 12, 43, 12, 51, 12, 59, 12, 67, 12, 75, 12, 83, 12, 91, 12, 99, 12, 107, 12, 115, 12, 123, 12, 131, 12, 139, 12, 147, 12, 155, 12, 163, 12, 48, 0, 48, 0, 48, 0, 48, 0, 171, 12, 178, 12, 186, 12, 194, 12, 202, 12, 210, 12, 218, 12, 226, 12, 234, 12, 242, 12, 250, 12, 2, 13, 10, 13, 18, 13, 26, 13, 34, 13, 42, 13, 50, 13, 58, 13, 66, 13, 74, 13, 82, 13, 90, 13, 98, 13, 106, 13, 114, 13, 122, 13, 130, 13, 138, 13, 146, 13, 154, 13, 162, 13, 170, 13, 178, 13, 185, 13, 193, 13, 201, 13, 209, 13, 217, 13, 225, 13, 233, 13, 241, 13, 249, 13, 1, 14, 9, 14, 17, 14, 24, 14, 32, 14, 40, 14, 48, 14, 56, 14, 64, 14, 72, 14, 80, 14, 88, 14, 96, 14, 104, 14, 112, 14, 120, 14, 127, 14, 135, 14, 142, 14, 150, 14, 158, 14, 166, 14, 174, 14, 182, 14, 190, 14, 198, 14, 206, 14, 214, 14, 222, 14, 230, 14, 238, 14, 246, 14, 254, 14, 6, 15, 14, 15, 22, 15, 30, 15, 38, 15, 46, 15, 54, 15, 62, 15, 70, 15, 78, 15, 86, 15, 94, 15, 102, 15, 110, 15, 118, 15, 126, 15, 134, 15, 142, 15, 150, 15, 158, 15, 166, 15, 174, 15, 182, 15, 190, 15, 48, 0, 48, 0, 48, 0, 48, 0, 197, 15, 205, 15, 213, 15, 221, 15, 229, 15, 237, 15, 245, 15, 253, 15, 48, 0, 48, 0, 4, 16, 12, 16, 20, 16, 28, 16, 36, 16, 44, 16, 52, 16, 60, 16, 68, 16, 76, 16, 84, 16, 92, 16, 100, 16, 108, 16, 116, 16, 124, 16, 132, 16, 134, 16, 142, 16, 150, 16, 158, 16, 166, 16, 174, 16, 182, 16, 190, 16, 198, 16, 206, 16, 214, 16, 222, 16, 230, 16, 237, 16, 243, 16, 251, 16, 3, 17, 11, 17, 19, 17, 27, 17, 35, 17, 43, 17, 51, 17, 59, 17, 67, 17, 75, 17, 82, 17, 89, 17, 97, 17, 105, 17, 112, 17, 120, 17, 128, 17, 135, 17, 143, 17, 151, 17, 159, 17, 167, 17, 175, 17, 183, 17, 191, 17, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 194, 17, 202, 17, 210, 17, 218, 17, 226, 17, 228, 17, 236, 17, 244, 17, 252, 17, 4, 18, 12, 18, 20, 18, 28, 18, 36, 18, 44, 18, 52, 18, 60, 18, 68, 18, 76, 18, 84, 18, 92, 18, 100, 18, 108, 18, 116, 18, 124, 18, 48, 0, 130, 18, 138, 18, 146, 18, 154, 18, 162, 18, 170, 18, 48, 0, 176, 18, 183, 18, 191, 18, 199, 18, 207, 18, 214, 18, 222, 18, 230, 18, 48, 0, 236, 18, 244, 18, 252, 18, 3, 19, 11, 19, 19, 19, 26, 19, 34, 19, 42, 19, 50, 19, 58, 19, 65, 19, 73, 19, 81, 19, 89, 19, 48, 0, 48, 0, 48, 0, 96, 19, 104, 19, 112, 19, 120, 19, 128, 19, 135, 19, 142, 19, 150, 19, 157, 19, 165, 19, 48, 0, 48, 0, 173, 19, 181, 19, 189, 19, 197, 19, 205, 19, 213, 19, 221, 19, 229, 19, 237, 19, 243, 19, 251, 19, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 58, 21, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 66, 21, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 3, 20, 11, 20, 19, 20, 27, 20, 35, 20, 43, 20, 51, 20, 59, 20, 67, 20, 75, 20, 83, 20, 91, 20, 99, 20, 107, 20, 115, 20, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 123, 20, 16, 0, 131, 20, 139, 20, 147, 20, 155, 20, 163, 20, 171, 20, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 128, 8, 179, 20, 187, 20, 195, 20, 203, 20, 211, 20, 219, 20, 227, 20, 235, 20, 243, 20, 251, 20, 3, 21, 11, 21, 19, 21, 27, 21, 34, 21, 42, 21, 50, 21, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 14, 0, 202, 193, 14, 0, 202, 226, 14, 0, 202, 3, 15, 0, 202, 36, 15, 0, 202, 69, 15, 0, 202, 102, 15, 0, 202, 135, 15, 0, 202, 168, 15, 0, 202, 201, 15, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 13, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 14, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 14, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 200, 0, 0, 0, 200, 32, 63, 0, 197, 162, 13, 0, 192, 0, 0, 0, 198, 1, 0, 0, 198, 33, 0, 0, 198, 65, 0, 0, 198, 97, 0, 0, 198, 129, 0, 0, 198, 161, 0, 0, 198, 193, 0, 0, 198, 225, 0, 0, 198, 1, 1, 0, 198, 33, 1, 0, 198, 65, 1, 0, 198, 98, 1, 0, 198, 161, 1, 0, 198, 194, 1, 0, 198, 1, 2, 0, 198, 34, 2, 0, 198, 97, 2, 0, 198, 130, 2, 0, 198, 193, 2, 0, 198, 226, 2, 0, 198, 33, 3, 0, 198, 66, 3, 0, 198, 129, 3, 0, 198, 162, 3, 0, 198, 225, 3, 0, 198, 2, 4, 0, 198, 65, 4, 0, 198, 98, 4, 0, 198, 161, 4, 0, 198, 194, 4, 0, 198, 1, 5, 0, 198, 34, 5, 0, 198, 97, 5, 0, 198, 130, 5, 0, 198, 193, 5, 0, 198, 225, 5, 0, 198, 2, 6, 0, 198, 65, 6, 0, 198, 98, 6, 0, 198, 161, 6, 0, 198, 194, 6, 0, 198, 1, 7, 0, 198, 33, 7, 0, 198, 65, 7, 0, 198, 97, 7, 0, 198, 129, 7, 0, 198, 161, 7, 0, 198, 194, 7, 0, 198, 2, 8, 0, 198, 65, 8, 0, 198, 98, 8, 0, 198, 162, 8, 0, 198, 225, 8, 0, 198, 2, 9, 0, 198, 66, 9, 0, 198, 129, 9, 0, 198, 162, 9, 0, 198, 226, 9, 0, 198, 33, 10, 0, 198, 66, 10, 0, 198, 130, 10, 0, 198, 193, 10, 0, 198, 225, 10, 0, 198, 1, 11, 0, 198, 33, 11, 0, 198, 65, 11, 0, 198, 97, 11, 0, 198, 129, 11, 0, 198, 161, 11, 0, 198, 193, 11, 0, 198, 225, 11, 0, 198, 1, 12, 0, 198, 33, 12, 0, 198, 65, 12, 0, 198, 97, 12, 0, 198, 129, 12, 0, 198, 161, 12, 0, 198, 193, 12, 0, 198, 225, 12, 0, 198, 1, 13, 0, 198, 33, 13, 0, 198, 65, 13, 0, 198, 97, 13, 0, 198, 130, 13, 0, 198, 193, 13, 0, 198, 225, 13, 0, 198, 2, 14, 0, 198, 66, 14, 0, 198, 130, 14, 0, 198, 194, 14, 0, 192, 0, 0, 0, 200, 0, 113, 0, 200, 224, 3, 1, 200, 0, 87, 1, 198, 66, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 5, 5, 8, 121, 5, 5, 183, 121, 5, 5, 10, 121, 5, 5, 187, 121, 5, 5, 189, 121, 5, 5, 12, 121, 5, 5, 14, 121, 5, 5, 16, 121, 5, 5, 195, 121, 5, 5, 197, 121, 5, 5, 199, 121, 5, 5, 201, 121, 5, 5, 203, 121, 5, 5, 205, 121, 5, 5, 52, 121, 5, 5, 18, 121, 5, 5, 20, 121, 5, 5, 22, 121, 5, 5, 66, 121, 5, 5, 24, 121, 5, 5, 26, 121, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 32, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 5, 105, 121, 5, 5, 107, 121, 5, 5, 109, 121, 5, 5, 111, 121, 5, 5, 113, 121, 5, 5, 115, 121, 5, 5, 117, 121, 5, 5, 119, 121, 5, 5, 121, 121, 5, 5, 123, 121, 5, 5, 125, 121, 5, 5, 127, 121, 5, 5, 129, 121, 5, 5, 131, 121, 5, 5, 133, 121, 5, 5, 135, 121, 5, 5, 137, 121, 5, 5, 139, 121, 5, 5, 141, 121, 5, 5, 143, 121, 193, 247, 101, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 13, 5, 5, 123, 13, 5, 5, 122, 13, 5, 5, 121, 13, 192, 0, 0, 0, 5, 5, 0, 202, 5, 5, 254, 232, 192, 0, 0, 0, 5, 5, 37, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 235, 5, 5, 18, 231, 5, 5, 190, 229, 5, 5, 17, 231, 5, 5, 118, 226, 192, 0, 0, 0, 5, 5, 18, 235, 5, 5, 119, 236, 192, 0, 0, 0, 5, 5, 75, 237, 5, 5, 11, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 227, 5, 5, 76, 237, 5, 5, 140, 231, 5, 5, 41, 238, 5, 5, 172, 227, 5, 5, 86, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 237, 5, 5, 48, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 237, 192, 0, 0, 0, 5, 5, 85, 227, 5, 5, 180, 232, 192, 0, 0, 0, 5, 5, 85, 230, 5, 5, 80, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 237, 5, 5, 53, 234, 192, 0, 0, 0, 5, 5, 173, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 234, 192, 0, 0, 0, 5, 5, 5, 234, 5, 5, 181, 228, 5, 5, 173, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 246, 5, 5, 83, 237, 5, 5, 20, 231, 5, 5, 84, 237, 5, 5, 111, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 228, 5, 5, 226, 228, 5, 5, 67, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 236, 192, 0, 0, 0, 5, 5, 31, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 227, 192, 0, 0, 0, 5, 5, 139, 227, 192, 0, 0, 0, 5, 5, 85, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 237, 5, 5, 209, 236, 192, 0, 0, 0, 5, 5, 117, 236, 5, 5, 48, 232, 5, 5, 88, 237, 5, 5, 6, 230, 5, 5, 21, 234, 192, 0, 0, 0, 5, 5, 91, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 226, 5, 5, 209, 228, 192, 0, 0, 0, 5, 5, 208, 228, 5, 5, 0, 196, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 237, 5, 5, 64, 237, 192, 0, 0, 0, 5, 5, 99, 229, 5, 5, 0, 127, 192, 0, 0, 0, 5, 5, 92, 237, 5, 5, 93, 237, 5, 5, 94, 237, 5, 5, 174, 235, 5, 5, 95, 237, 192, 0, 0, 0, 5, 5, 228, 228, 5, 5, 0, 197, 5, 5, 236, 235, 192, 0, 0, 0, 5, 5, 210, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 227, 5, 5, 212, 227, 5, 5, 70, 233, 5, 5, 210, 236, 192, 0, 0, 0, 5, 5, 96, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 230, 5, 5, 88, 231, 5, 5, 103, 237, 192, 0, 0, 0, 5, 5, 101, 237, 192, 0, 0, 0, 5, 5, 102, 237, 5, 5, 174, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 229, 5, 5, 170, 226, 192, 0, 0, 0, 5, 5, 100, 237, 5, 5, 99, 237, 5, 5, 69, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 229, 5, 5, 211, 229, 5, 5, 100, 232, 5, 5, 104, 237, 5, 5, 19, 235, 5, 5, 207, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 237, 5, 5, 107, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 232, 5, 5, 247, 236, 5, 5, 0, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 237, 5, 5, 120, 226, 192, 0, 0, 0, 5, 5, 248, 227, 192, 0, 0, 0, 5, 5, 235, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 228, 5, 5, 108, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 237, 5, 5, 0, 169, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 228, 5, 5, 100, 227, 5, 5, 56, 235, 5, 5, 144, 234, 5, 5, 175, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 226, 192, 0, 0, 0, 5, 5, 144, 237, 5, 5, 131, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 234, 5, 5, 111, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 234, 192, 0, 0, 0, 5, 5, 248, 236, 192, 0, 0, 0, 5, 5, 55, 231, 192, 0, 0, 0, 5, 5, 213, 229, 192, 0, 0, 0, 5, 5, 7, 230, 5, 5, 122, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 232, 5, 5, 115, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 170, 5, 5, 71, 233, 5, 5, 135, 230, 5, 5, 100, 229, 5, 5, 92, 236, 192, 0, 0, 0, 5, 5, 116, 232, 192, 0, 0, 0, 5, 5, 121, 226, 192, 0, 0, 0, 5, 5, 114, 237, 192, 0, 0, 0, 5, 5, 118, 236, 5, 5, 110, 237, 5, 5, 112, 237, 5, 5, 158, 229, 5, 5, 113, 237, 5, 5, 39, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 237, 5, 5, 122, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 226, 192, 0, 0, 0, 5, 5, 87, 235, 5, 5, 116, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 237, 5, 5, 229, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 237, 192, 0, 0, 0, 5, 5, 117, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 236, 192, 0, 0, 0, 5, 5, 8, 230, 192, 0, 0, 0, 5, 5, 118, 237, 192, 0, 0, 0, 5, 5, 123, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 237, 192, 0, 0, 0, 5, 5, 119, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 227, 192, 0, 0, 0, 5, 5, 0, 171, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 227, 5, 5, 123, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 239, 192, 0, 0, 0, 5, 5, 243, 235, 5, 5, 44, 235, 5, 5, 230, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 231, 5, 5, 205, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 228, 5, 5, 75, 232, 5, 5, 158, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 237, 192, 0, 0, 0, 5, 5, 135, 237, 5, 5, 133, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 232, 5, 5, 131, 237, 192, 0, 0, 0, 5, 5, 134, 237, 5, 5, 132, 237, 192, 0, 0, 0, 5, 5, 121, 235, 192, 0, 0, 0, 5, 5, 129, 237, 192, 0, 0, 0, 5, 5, 56, 231, 192, 0, 0, 0, 5, 5, 237, 235, 5, 5, 136, 237, 5, 5, 137, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 230, 5, 5, 150, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 234, 192, 0, 0, 0, 5, 5, 242, 234, 5, 5, 145, 237, 192, 0, 0, 0, 5, 5, 140, 235, 192, 0, 0, 0, 5, 5, 112, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 237, 5, 5, 152, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 232, 192, 0, 0, 0, 5, 5, 182, 228, 192, 0, 0, 0, 5, 5, 92, 234, 192, 0, 0, 0, 5, 5, 92, 243, 192, 0, 0, 0, 5, 5, 151, 237, 5, 5, 159, 233, 192, 0, 0, 0, 5, 5, 140, 237, 192, 0, 0, 0, 5, 5, 232, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 228, 5, 5, 138, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 230, 192, 0, 0, 0, 5, 5, 146, 237, 192, 0, 0, 0, 5, 5, 139, 235, 5, 5, 209, 232, 5, 5, 142, 237, 5, 5, 133, 228, 192, 0, 0, 0, 5, 5, 139, 237, 5, 5, 147, 237, 5, 5, 141, 237, 5, 5, 231, 236, 5, 5, 148, 237, 5, 5, 55, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 237, 5, 5, 158, 237, 5, 5, 0, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 235, 5, 5, 157, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 237, 5, 5, 160, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 237, 192, 0, 0, 0, 5, 5, 72, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 230, 192, 0, 0, 0, 5, 5, 76, 232, 5, 5, 73, 233, 5, 5, 44, 228, 192, 0, 0, 0, 5, 5, 162, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 229, 5, 5, 213, 234, 5, 5, 164, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 229, 5, 5, 123, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 237, 5, 5, 170, 237, 5, 5, 166, 237, 5, 5, 116, 229, 192, 0, 0, 0, 5, 5, 201, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 237, 5, 5, 169, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 233, 192, 0, 0, 0, 5, 5, 68, 232, 192, 0, 0, 0, 5, 5, 215, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 235, 5, 5, 172, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 237, 192, 0, 0, 0, 5, 5, 174, 237, 192, 0, 0, 0, 5, 5, 22, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 237, 5, 5, 177, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 237, 192, 0, 0, 0, 5, 5, 100, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 227, 5, 5, 181, 237, 5, 5, 182, 237, 192, 0, 0, 0, 5, 5, 105, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 230, 192, 0, 0, 0, 5, 5, 185, 237, 5, 5, 184, 237, 5, 5, 183, 237, 192, 0, 0, 0, 5, 5, 127, 237, 192, 0, 0, 0, 5, 5, 186, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 230, 192, 0, 0, 0, 5, 5, 187, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 237, 5, 5, 191, 237, 5, 5, 190, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 237, 5, 5, 193, 237, 5, 5, 0, 210, 192, 0, 0, 0, 5, 5, 167, 228, 5, 5, 79, 228, 5, 5, 136, 230, 5, 5, 255, 232, 5, 5, 216, 227, 5, 5, 208, 231, 5, 5, 233, 228, 192, 0, 0, 0, 5, 5, 64, 229, 5, 5, 195, 237, 5, 5, 34, 236, 5, 5, 136, 233, 192, 0, 0, 0, 5, 5, 9, 230, 192, 0, 0, 0, 5, 5, 194, 237, 192, 0, 0, 0, 5, 5, 196, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 233, 192, 0, 0, 0, 5, 5, 15, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 231, 5, 5, 199, 237, 5, 5, 200, 237, 5, 5, 138, 234, 5, 5, 234, 228, 5, 5, 49, 237, 5, 5, 201, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 235, 5, 5, 92, 232, 5, 5, 39, 228, 5, 5, 119, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 237, 192, 0, 0, 0, 5, 5, 203, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 234, 5, 5, 59, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 237, 5, 5, 173, 229, 192, 0, 0, 0, 5, 5, 205, 237, 5, 5, 118, 229, 192, 0, 0, 0, 5, 5, 207, 237, 5, 5, 187, 245, 5, 5, 208, 237, 5, 5, 187, 235, 5, 5, 209, 237, 192, 0, 0, 0, 5, 5, 210, 237, 5, 5, 211, 237, 5, 5, 21, 231, 192, 0, 0, 0, 5, 5, 58, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 227, 192, 0, 0, 0, 5, 5, 214, 237, 192, 0, 0, 0, 5, 5, 212, 237, 5, 5, 23, 236, 5, 5, 213, 237, 192, 0, 0, 0, 5, 5, 24, 235, 5, 5, 215, 237, 5, 5, 216, 237, 5, 5, 217, 237, 5, 5, 161, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 237, 5, 5, 219, 237, 5, 5, 220, 237, 5, 5, 218, 237, 5, 5, 147, 229, 5, 5, 222, 237, 5, 5, 68, 236, 5, 5, 250, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 231, 5, 5, 224, 237, 5, 5, 166, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 237, 192, 0, 0, 0, 5, 5, 2, 233, 5, 5, 213, 236, 5, 5, 162, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 237, 5, 5, 160, 250, 5, 5, 249, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 237, 5, 5, 216, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 230, 5, 5, 162, 232, 192, 0, 0, 0, 5, 5, 229, 237, 5, 5, 6, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 237, 5, 5, 198, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 237, 5, 5, 219, 227, 192, 0, 0, 0, 5, 5, 236, 233, 5, 5, 86, 226, 5, 5, 156, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 234, 5, 5, 233, 237, 192, 0, 0, 0, 5, 5, 163, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 231, 5, 5, 234, 237, 192, 0, 0, 0, 5, 5, 72, 235, 5, 5, 194, 231, 5, 5, 26, 227, 192, 0, 0, 0, 5, 5, 33, 227, 5, 5, 235, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 234, 5, 5, 104, 235, 192, 0, 0, 0, 5, 5, 238, 237, 192, 0, 0, 0, 5, 5, 176, 236, 5, 5, 239, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 237, 192, 0, 0, 0, 5, 5, 192, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 231, 5, 5, 174, 229, 5, 5, 136, 228, 5, 5, 242, 237, 5, 5, 215, 229, 5, 5, 65, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 233, 5, 5, 244, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 229, 5, 5, 245, 237, 5, 5, 246, 237, 5, 5, 246, 231, 192, 0, 0, 0, 5, 5, 243, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 237, 192, 0, 0, 0, 5, 5, 176, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 228, 5, 5, 142, 229, 5, 5, 109, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 237, 5, 5, 249, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 235, 5, 5, 22, 231, 5, 5, 4, 238, 5, 5, 252, 226, 5, 5, 252, 237, 5, 5, 250, 237, 5, 5, 23, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 237, 192, 0, 0, 0, 5, 5, 253, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 228, 5, 5, 5, 238, 5, 5, 195, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 238, 5, 5, 3, 238, 192, 0, 0, 0, 5, 5, 2, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 228, 5, 5, 125, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 230, 5, 5, 154, 233, 5, 5, 55, 229, 5, 5, 9, 238, 5, 5, 10, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 237, 5, 5, 12, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 238, 192, 0, 0, 0, 5, 5, 199, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 238, 192, 0, 0, 0, 5, 5, 208, 235, 192, 0, 0, 0, 5, 5, 32, 233, 192, 0, 0, 0, 5, 5, 94, 236, 192, 0, 0, 0, 5, 5, 117, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 233, 192, 0, 0, 0, 5, 5, 15, 238, 5, 5, 34, 227, 5, 5, 13, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 230, 5, 5, 16, 238, 5, 5, 131, 235, 5, 5, 20, 238, 192, 0, 0, 0, 5, 5, 146, 231, 5, 5, 17, 238, 5, 5, 6, 228, 192, 0, 0, 0, 5, 5, 18, 238, 5, 5, 35, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 228, 5, 5, 21, 238, 192, 0, 0, 0, 5, 5, 22, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 238, 5, 5, 24, 238, 5, 5, 73, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 228, 5, 5, 56, 236, 192, 0, 0, 0, 5, 5, 66, 236, 5, 5, 25, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 235, 5, 5, 25, 238, 192, 0, 0, 0, 5, 5, 26, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 238, 192, 0, 0, 0, 5, 5, 30, 238, 5, 5, 29, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 238, 5, 5, 119, 226, 5, 5, 198, 235, 192, 0, 0, 0, 5, 5, 172, 229, 5, 5, 32, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 230, 5, 5, 221, 227, 192, 0, 0, 0, 5, 5, 33, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 238, 192, 0, 0, 0, 5, 5, 35, 238, 192, 0, 0, 0, 5, 5, 36, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 238, 5, 5, 224, 234, 5, 5, 30, 228, 5, 5, 0, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 233, 5, 5, 38, 238, 5, 5, 137, 230, 192, 0, 0, 0, 5, 5, 209, 231, 192, 0, 0, 0, 5, 5, 40, 238, 5, 5, 39, 238, 5, 5, 205, 230, 5, 5, 211, 228, 5, 5, 42, 238, 5, 5, 156, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 234, 5, 5, 90, 232, 5, 5, 148, 232, 5, 5, 220, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 234, 5, 5, 181, 232, 192, 0, 0, 0, 5, 5, 110, 234, 192, 0, 0, 0, 5, 5, 200, 235, 192, 0, 0, 0, 5, 5, 45, 238, 192, 0, 0, 0, 5, 5, 210, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 238, 5, 5, 8, 226, 5, 5, 0, 211, 5, 5, 101, 227, 192, 0, 0, 0, 5, 5, 78, 232, 5, 5, 167, 227, 5, 5, 169, 236, 192, 0, 0, 0, 5, 5, 50, 238, 5, 5, 113, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 238, 192, 0, 0, 0, 5, 5, 75, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 238, 192, 0, 0, 0, 5, 5, 232, 236, 192, 0, 0, 0, 5, 5, 238, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 228, 5, 5, 53, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 238, 5, 5, 54, 238, 192, 0, 0, 0, 5, 5, 105, 231, 5, 5, 21, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 226, 5, 5, 56, 238, 192, 0, 0, 0, 5, 5, 57, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 229, 5, 5, 59, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 235, 5, 5, 101, 229, 5, 5, 176, 227, 5, 5, 95, 236, 5, 5, 24, 232, 5, 5, 157, 234, 5, 5, 107, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 230, 192, 0, 0, 0, 5, 5, 86, 230, 5, 5, 99, 230, 192, 0, 0, 0, 5, 5, 194, 230, 192, 0, 0, 0, 5, 5, 158, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 238, 192, 0, 0, 0, 5, 5, 31, 226, 5, 5, 25, 232, 5, 5, 239, 228, 5, 5, 183, 228, 5, 5, 29, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 238, 5, 5, 165, 232, 5, 5, 164, 232, 5, 5, 223, 227, 5, 5, 206, 230, 5, 5, 67, 238, 5, 5, 65, 238, 5, 5, 126, 226, 5, 5, 140, 232, 5, 5, 38, 230, 5, 5, 217, 229, 5, 5, 2, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 227, 5, 5, 56, 229, 5, 5, 216, 229, 192, 0, 0, 0, 5, 5, 68, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 238, 192, 0, 0, 0, 5, 5, 159, 227, 5, 5, 222, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 229, 5, 5, 158, 227, 5, 5, 67, 233, 5, 5, 0, 225, 5, 5, 209, 233, 5, 5, 24, 236, 5, 5, 241, 228, 5, 5, 177, 236, 5, 5, 137, 233, 5, 5, 240, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 228, 192, 0, 0, 0, 5, 5, 78, 238, 192, 0, 0, 0, 5, 5, 25, 228, 5, 5, 196, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 238, 192, 0, 0, 0, 5, 5, 86, 227, 5, 5, 72, 238, 5, 5, 73, 238, 5, 5, 75, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 238, 192, 0, 0, 0, 5, 5, 177, 227, 5, 5, 107, 231, 192, 0, 0, 0, 5, 5, 73, 235, 5, 5, 74, 238, 5, 5, 70, 238, 5, 5, 213, 228, 192, 0, 0, 0, 5, 5, 71, 238, 192, 0, 0, 0, 5, 5, 26, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 235, 192, 0, 0, 0, 5, 5, 76, 233, 5, 5, 212, 228, 5, 5, 66, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 230, 192, 0, 0, 0, 5, 5, 100, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 238, 5, 5, 84, 238, 192, 0, 0, 0, 5, 5, 251, 235, 192, 0, 0, 0, 5, 5, 81, 238, 5, 5, 90, 238, 5, 5, 85, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 238, 5, 5, 184, 228, 5, 5, 25, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 238, 192, 0, 0, 0, 5, 5, 93, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 229, 5, 5, 56, 237, 192, 0, 0, 0, 5, 5, 82, 238, 5, 5, 80, 238, 5, 5, 92, 238, 192, 0, 0, 0, 5, 5, 87, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 238, 192, 0, 0, 0, 5, 5, 104, 238, 5, 5, 97, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 238, 5, 5, 98, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 229, 5, 5, 201, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 238, 5, 5, 0, 212, 5, 5, 105, 238, 192, 0, 0, 0, 5, 5, 0, 131, 5, 5, 8, 235, 5, 5, 103, 238, 192, 0, 0, 0, 5, 5, 99, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 238, 5, 5, 100, 238, 5, 5, 120, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 213, 5, 5, 116, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 238, 5, 5, 109, 238, 192, 0, 0, 0, 5, 5, 207, 230, 5, 5, 227, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 238, 5, 5, 113, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 226, 192, 0, 0, 0, 5, 5, 102, 229, 5, 5, 58, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 238, 5, 5, 164, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 238, 192, 0, 0, 0, 5, 5, 0, 128, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 238, 5, 5, 91, 236, 192, 0, 0, 0, 5, 5, 209, 230, 192, 0, 0, 0, 5, 5, 127, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 238, 5, 5, 117, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 232, 192, 0, 0, 0, 5, 5, 118, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 232, 5, 5, 123, 238, 5, 5, 208, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 238, 5, 5, 125, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 238, 5, 5, 128, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 238, 5, 5, 139, 238, 192, 0, 0, 0, 5, 5, 135, 238, 192, 0, 0, 0, 5, 5, 130, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 238, 5, 5, 247, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 238, 192, 0, 0, 0, 5, 5, 242, 228, 5, 5, 132, 238, 5, 5, 3, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 238, 5, 5, 129, 238, 5, 5, 37, 227, 192, 0, 0, 0, 5, 5, 102, 227, 5, 5, 253, 226, 5, 5, 137, 238, 5, 5, 133, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 228, 5, 5, 143, 238, 5, 5, 141, 238, 5, 5, 27, 232, 5, 5, 160, 227, 5, 5, 224, 227, 192, 0, 0, 0, 5, 5, 138, 238, 192, 0, 0, 0, 5, 5, 42, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 238, 5, 5, 145, 238, 192, 0, 0, 0, 5, 5, 202, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 238, 192, 0, 0, 0, 5, 5, 148, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 229, 5, 5, 149, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 238, 192, 0, 0, 0, 5, 5, 157, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 238, 5, 5, 154, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 238, 192, 0, 0, 0, 5, 5, 153, 238, 5, 5, 210, 230, 5, 5, 15, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 238, 192, 0, 0, 0, 5, 5, 40, 231, 5, 5, 163, 238, 192, 0, 0, 0, 5, 5, 161, 238, 192, 0, 0, 0, 5, 5, 162, 238, 192, 0, 0, 0, 5, 5, 164, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 231, 192, 0, 0, 0, 5, 5, 158, 238, 192, 0, 0, 0, 5, 5, 159, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 227, 192, 0, 0, 0, 5, 5, 169, 238, 5, 5, 165, 238, 5, 5, 168, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 233, 192, 0, 0, 0, 5, 5, 150, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 238, 5, 5, 221, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 238, 192, 0, 0, 0, 5, 5, 57, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 238, 192, 0, 0, 0, 5, 5, 178, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 238, 5, 5, 182, 238, 5, 5, 180, 238, 5, 5, 183, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 238, 192, 0, 0, 0, 5, 5, 188, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 238, 5, 5, 204, 237, 192, 0, 0, 0, 5, 5, 106, 230, 5, 5, 219, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 226, 192, 0, 0, 0, 5, 5, 0, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 238, 192, 0, 0, 0, 5, 5, 84, 229, 192, 0, 0, 0, 5, 5, 0, 173, 5, 5, 104, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 238, 5, 5, 185, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 229, 192, 0, 0, 0, 5, 5, 193, 238, 5, 5, 192, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 235, 5, 5, 194, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 238, 5, 5, 195, 238, 192, 0, 0, 0, 5, 5, 197, 238, 192, 0, 0, 0, 5, 5, 198, 238, 192, 0, 0, 0, 5, 5, 139, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 226, 5, 5, 199, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 238, 192, 0, 0, 0, 5, 5, 200, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 238, 5, 5, 0, 147, 5, 5, 143, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 238, 5, 5, 206, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 238, 192, 0, 0, 0, 5, 5, 148, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 238, 5, 5, 210, 238, 5, 5, 113, 229, 5, 5, 243, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 229, 192, 0, 0, 0, 5, 5, 183, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 238, 5, 5, 62, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 238, 5, 5, 216, 238, 192, 0, 0, 0, 5, 5, 83, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 238, 192, 0, 0, 0, 5, 5, 244, 228, 5, 5, 216, 226, 5, 5, 220, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 238, 5, 5, 227, 238, 5, 5, 225, 238, 192, 0, 0, 0, 5, 5, 229, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 198, 5, 5, 20, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 230, 192, 0, 0, 0, 5, 5, 93, 234, 5, 5, 104, 227, 192, 0, 0, 0, 5, 5, 155, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 235, 192, 0, 0, 0, 5, 5, 210, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 228, 5, 5, 117, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 238, 5, 5, 231, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 250, 5, 5, 61, 226, 5, 5, 143, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 231, 5, 5, 138, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 235, 5, 5, 243, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 226, 5, 5, 237, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 232, 5, 5, 240, 238, 192, 0, 0, 0, 5, 5, 165, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 233, 5, 5, 166, 233, 5, 5, 151, 234, 5, 5, 49, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 226, 192, 0, 0, 0, 5, 5, 120, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 238, 192, 0, 0, 0, 5, 5, 234, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 227, 192, 0, 0, 0, 5, 5, 243, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 238, 5, 5, 244, 238, 5, 5, 247, 238, 5, 5, 249, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 238, 192, 0, 0, 0, 5, 5, 201, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 226, 192, 0, 0, 0, 5, 5, 25, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 238, 192, 0, 0, 0, 5, 5, 253, 238, 192, 0, 0, 0, 5, 5, 58, 229, 192, 0, 0, 0, 5, 5, 255, 238, 5, 5, 3, 239, 5, 5, 2, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 239, 192, 0, 0, 0, 5, 5, 248, 238, 5, 5, 7, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 239, 5, 5, 4, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 229, 5, 5, 91, 231, 192, 0, 0, 0, 5, 5, 28, 232, 5, 5, 8, 239, 5, 5, 164, 231, 5, 5, 0, 203, 5, 5, 100, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 233, 192, 0, 0, 0, 5, 5, 10, 239, 5, 5, 9, 239, 5, 5, 11, 239, 5, 5, 12, 239, 5, 5, 13, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 235, 5, 5, 15, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 226, 5, 5, 16, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 236, 5, 5, 200, 226, 192, 0, 0, 0, 5, 5, 48, 238, 5, 5, 148, 230, 5, 5, 101, 232, 5, 5, 17, 239, 5, 5, 69, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 239, 192, 0, 0, 0, 5, 5, 141, 232, 192, 0, 0, 0, 5, 5, 121, 233, 5, 5, 102, 232, 5, 5, 21, 235, 5, 5, 20, 239, 5, 5, 21, 239, 5, 5, 87, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 230, 5, 5, 22, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 174, 5, 5, 23, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 227, 5, 5, 2, 234, 5, 5, 144, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 239, 5, 5, 29, 232, 5, 5, 27, 239, 5, 5, 84, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 235, 5, 5, 26, 239, 192, 0, 0, 0, 5, 5, 167, 233, 5, 5, 30, 239, 192, 0, 0, 0, 5, 5, 29, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 239, 192, 0, 0, 0, 5, 5, 31, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 226, 192, 0, 0, 0, 5, 5, 33, 239, 5, 5, 211, 230, 5, 5, 35, 239, 5, 5, 169, 232, 192, 0, 0, 0, 5, 5, 34, 239, 192, 0, 0, 0, 5, 5, 79, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 230, 5, 5, 157, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 239, 5, 5, 33, 234, 5, 5, 188, 234, 5, 5, 43, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 236, 192, 0, 0, 0, 5, 5, 141, 239, 192, 0, 0, 0, 5, 5, 38, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 239, 192, 0, 0, 0, 5, 5, 107, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 235, 192, 0, 0, 0, 5, 5, 122, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 229, 192, 0, 0, 0, 5, 5, 221, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 151, 5, 5, 186, 228, 192, 0, 0, 0, 5, 5, 147, 231, 5, 5, 0, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 239, 5, 5, 48, 239, 192, 0, 0, 0, 5, 5, 45, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 226, 5, 5, 39, 227, 192, 0, 0, 0, 5, 5, 44, 239, 192, 0, 0, 0, 5, 5, 31, 236, 5, 5, 237, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 134, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 229, 192, 0, 0, 0, 5, 5, 0, 176, 192, 0, 0, 0, 5, 5, 0, 129, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 236, 192, 0, 0, 0, 5, 5, 54, 239, 192, 0, 0, 0, 5, 5, 52, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 239, 5, 5, 59, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 239, 5, 5, 59, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 239, 5, 5, 34, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 239, 192, 0, 0, 0, 5, 5, 56, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 239, 5, 5, 238, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 239, 5, 5, 67, 239, 5, 5, 64, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 226, 5, 5, 66, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 230, 192, 0, 0, 0, 5, 5, 65, 239, 5, 5, 141, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 239, 5, 5, 69, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 232, 192, 0, 0, 0, 5, 5, 68, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 239, 5, 5, 74, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 227, 192, 0, 0, 0, 5, 5, 76, 239, 5, 5, 75, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 239, 192, 0, 0, 0, 5, 5, 64, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 226, 192, 0, 0, 0, 5, 5, 78, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 239, 192, 0, 0, 0, 5, 5, 85, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 239, 192, 0, 0, 0, 5, 5, 84, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 229, 5, 5, 86, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 228, 5, 5, 87, 239, 192, 0, 0, 0, 5, 5, 10, 230, 5, 5, 94, 232, 192, 0, 0, 0, 5, 5, 88, 239, 5, 5, 89, 239, 5, 5, 232, 229, 5, 5, 247, 228, 192, 0, 0, 0, 5, 5, 44, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 227, 5, 5, 187, 228, 5, 5, 90, 239, 5, 5, 240, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 239, 192, 0, 0, 0, 5, 5, 95, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 239, 5, 5, 132, 239, 192, 0, 0, 0, 5, 5, 93, 239, 192, 0, 0, 0, 5, 5, 94, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 239, 192, 0, 0, 0, 5, 5, 97, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 239, 192, 0, 0, 0, 5, 5, 150, 232, 192, 0, 0, 0, 5, 5, 3, 226, 5, 5, 87, 230, 5, 5, 0, 160, 192, 0, 0, 0, 5, 5, 31, 232, 5, 5, 40, 227, 5, 5, 35, 230, 192, 0, 0, 0, 5, 5, 248, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 233, 192, 0, 0, 0, 5, 5, 109, 230, 5, 5, 41, 227, 5, 5, 236, 232, 5, 5, 78, 233, 5, 5, 0, 150, 5, 5, 143, 227, 5, 5, 145, 235, 192, 0, 0, 0, 5, 5, 48, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 227, 5, 5, 211, 231, 5, 5, 42, 230, 5, 5, 96, 236, 5, 5, 100, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 227, 192, 0, 0, 0, 5, 5, 123, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 226, 5, 5, 63, 226, 5, 5, 214, 230, 5, 5, 130, 226, 192, 0, 0, 0, 5, 5, 101, 239, 5, 5, 126, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 230, 5, 5, 102, 239, 5, 5, 107, 227, 5, 5, 242, 233, 5, 5, 3, 236, 5, 5, 103, 239, 192, 0, 0, 0, 5, 5, 104, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 239, 192, 0, 0, 0, 5, 5, 32, 227, 5, 5, 45, 228, 5, 5, 105, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 227, 192, 0, 0, 0, 5, 5, 60, 231, 5, 5, 110, 239, 5, 5, 175, 229, 192, 0, 0, 0, 5, 5, 131, 226, 5, 5, 109, 239, 192, 0, 0, 0, 5, 5, 107, 239, 5, 5, 111, 239, 5, 5, 108, 239, 5, 5, 43, 234, 5, 5, 230, 241, 5, 5, 61, 231, 192, 0, 0, 0, 5, 5, 112, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 236, 192, 0, 0, 0, 5, 5, 113, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 239, 192, 0, 0, 0, 5, 5, 4, 233, 5, 5, 114, 239, 192, 0, 0, 0, 5, 5, 139, 231, 192, 0, 0, 0, 5, 5, 11, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 232, 5, 5, 101, 230, 192, 0, 0, 0, 5, 5, 51, 235, 5, 5, 212, 231, 192, 0, 0, 0, 5, 5, 59, 230, 5, 5, 116, 239, 5, 5, 215, 230, 5, 5, 117, 239, 5, 5, 118, 239, 5, 5, 0, 177, 5, 5, 96, 232, 5, 5, 92, 231, 192, 0, 0, 0, 5, 5, 119, 239, 5, 5, 211, 233, 5, 5, 216, 230, 192, 0, 0, 0, 5, 5, 217, 230, 192, 0, 0, 0, 5, 5, 120, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 239, 192, 0, 0, 0, 5, 5, 122, 239, 192, 0, 0, 0, 5, 5, 58, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 239, 5, 5, 125, 239, 5, 5, 74, 230, 5, 5, 54, 231, 5, 5, 22, 234, 5, 5, 94, 231, 5, 5, 214, 234, 5, 5, 34, 234, 5, 5, 253, 227, 5, 5, 126, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 227, 5, 5, 127, 239, 192, 0, 0, 0, 5, 5, 52, 228, 192, 0, 0, 0, 5, 5, 239, 233, 5, 5, 106, 226, 192, 0, 0, 0, 5, 5, 225, 229, 5, 5, 128, 239, 5, 5, 131, 239, 5, 5, 130, 239, 5, 5, 51, 228, 192, 0, 0, 0, 5, 5, 129, 239, 192, 0, 0, 0, 5, 5, 122, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 232, 192, 0, 0, 0, 5, 5, 141, 233, 5, 5, 54, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 232, 5, 5, 178, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 239, 192, 0, 0, 0, 5, 5, 134, 239, 5, 5, 246, 233, 192, 0, 0, 0, 5, 5, 193, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 227, 5, 5, 139, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 232, 5, 5, 92, 227, 192, 0, 0, 0, 5, 5, 142, 239, 5, 5, 2, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 232, 192, 0, 0, 0, 5, 5, 241, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 239, 5, 5, 146, 239, 5, 5, 87, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 239, 5, 5, 145, 239, 192, 0, 0, 0, 5, 5, 148, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 239, 192, 0, 0, 0, 5, 5, 149, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 233, 5, 5, 226, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 226, 5, 5, 151, 239, 5, 5, 156, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 239, 192, 0, 0, 0, 5, 5, 147, 235, 5, 5, 146, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 239, 5, 5, 160, 230, 192, 0, 0, 0, 5, 5, 152, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 239, 5, 5, 158, 239, 5, 5, 203, 226, 5, 5, 159, 239, 5, 5, 168, 239, 5, 5, 167, 239, 5, 5, 166, 239, 5, 5, 162, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 239, 5, 5, 169, 239, 192, 0, 0, 0, 5, 5, 171, 239, 192, 0, 0, 0, 5, 5, 170, 236, 192, 0, 0, 0, 5, 5, 170, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 239, 192, 0, 0, 0, 5, 5, 177, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 233, 5, 5, 155, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 239, 192, 0, 0, 0, 5, 5, 182, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 236, 192, 0, 0, 0, 5, 5, 186, 239, 5, 5, 183, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 227, 5, 5, 188, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 239, 5, 5, 189, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 239, 192, 0, 0, 0, 5, 5, 214, 231, 5, 5, 111, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 230, 192, 0, 0, 0, 5, 5, 43, 232, 192, 0, 0, 0, 5, 5, 249, 228, 5, 5, 104, 229, 5, 5, 250, 228, 5, 5, 198, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 228, 5, 5, 194, 239, 5, 5, 254, 235, 5, 5, 67, 234, 5, 5, 195, 239, 192, 0, 0, 0, 5, 5, 251, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 227, 192, 0, 0, 0, 5, 5, 171, 232, 5, 5, 8, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 229, 5, 5, 25, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 239, 5, 5, 109, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 239, 5, 5, 197, 239, 5, 5, 200, 239, 192, 0, 0, 0, 5, 5, 79, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 229, 192, 0, 0, 0, 5, 5, 178, 231, 192, 0, 0, 0, 5, 5, 121, 232, 5, 5, 120, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 239, 5, 5, 202, 239, 5, 5, 27, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 239, 5, 5, 203, 239, 5, 5, 59, 235, 192, 0, 0, 0, 5, 5, 212, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 235, 192, 0, 0, 0, 5, 5, 206, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 239, 5, 5, 229, 235, 192, 0, 0, 0, 5, 5, 207, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 239, 192, 0, 0, 0, 5, 5, 134, 234, 5, 5, 210, 239, 5, 5, 90, 235, 5, 5, 211, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 227, 5, 5, 91, 235, 5, 5, 47, 234, 5, 5, 213, 239, 5, 5, 214, 239, 192, 0, 0, 0, 5, 5, 252, 228, 5, 5, 44, 227, 5, 5, 215, 239, 5, 5, 170, 228, 5, 5, 124, 236, 5, 5, 97, 236, 5, 5, 110, 227, 5, 5, 217, 239, 192, 0, 0, 0, 5, 5, 7, 233, 192, 0, 0, 0, 5, 5, 253, 228, 5, 5, 219, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 233, 5, 5, 149, 235, 5, 5, 123, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 228, 192, 0, 0, 0, 5, 5, 26, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 233, 5, 5, 114, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 228, 192, 0, 0, 0, 5, 5, 81, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 161, 5, 5, 186, 230, 5, 5, 255, 228, 5, 5, 127, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 239, 5, 5, 220, 239, 5, 5, 81, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 239, 5, 5, 15, 237, 5, 5, 35, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 239, 5, 5, 222, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 239, 5, 5, 228, 239, 192, 0, 0, 0, 5, 5, 226, 239, 192, 0, 0, 0, 5, 5, 252, 234, 5, 5, 221, 230, 5, 5, 230, 239, 5, 5, 229, 239, 5, 5, 225, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 239, 5, 5, 232, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 239, 5, 5, 234, 239, 192, 0, 0, 0, 5, 5, 235, 239, 5, 5, 237, 239, 192, 0, 0, 0, 5, 5, 64, 226, 5, 5, 82, 233, 5, 5, 238, 239, 192, 0, 0, 0, 5, 5, 142, 228, 5, 5, 176, 226, 5, 5, 54, 234, 192, 0, 0, 0, 5, 5, 239, 239, 5, 5, 29, 234, 192, 0, 0, 0, 5, 5, 119, 235, 192, 0, 0, 0, 5, 5, 240, 239, 5, 5, 36, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 239, 5, 5, 92, 235, 5, 5, 244, 239, 5, 5, 74, 237, 5, 5, 90, 237, 192, 0, 0, 0, 5, 5, 30, 230, 5, 5, 23, 234, 5, 5, 245, 239, 192, 0, 0, 0, 5, 5, 179, 227, 5, 5, 8, 233, 5, 5, 0, 216, 5, 5, 246, 239, 5, 5, 66, 235, 5, 5, 2, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 236, 5, 5, 171, 228, 5, 5, 190, 228, 192, 0, 0, 0, 5, 5, 247, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 239, 192, 0, 0, 0, 5, 5, 254, 239, 192, 0, 0, 0, 5, 5, 83, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 233, 192, 0, 0, 0, 5, 5, 227, 227, 5, 5, 249, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 234, 192, 0, 0, 0, 5, 5, 202, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 239, 192, 0, 0, 0, 5, 5, 228, 227, 192, 0, 0, 0, 5, 5, 252, 239, 192, 0, 0, 0, 5, 5, 253, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 239, 192, 0, 0, 0, 5, 5, 184, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 240, 5, 5, 3, 240, 192, 0, 0, 0, 5, 5, 4, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 239, 5, 5, 242, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 240, 5, 5, 85, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 229, 5, 5, 243, 234, 5, 5, 10, 233, 5, 5, 9, 235, 5, 5, 6, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 230, 5, 5, 34, 226, 192, 0, 0, 0, 5, 5, 7, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 240, 192, 0, 0, 0, 5, 5, 76, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 240, 5, 5, 89, 226, 5, 5, 148, 231, 5, 5, 10, 240, 5, 5, 9, 240, 5, 5, 86, 228, 5, 5, 122, 232, 192, 0, 0, 0, 5, 5, 15, 240, 5, 5, 13, 240, 192, 0, 0, 0, 5, 5, 12, 240, 5, 5, 189, 236, 5, 5, 215, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 230, 5, 5, 14, 240, 5, 5, 142, 233, 5, 5, 138, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 233, 5, 5, 18, 240, 5, 5, 17, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 240, 192, 0, 0, 0, 5, 5, 19, 240, 5, 5, 216, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 240, 5, 5, 58, 235, 5, 5, 167, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 240, 5, 5, 215, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 233, 5, 5, 11, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 240, 5, 5, 129, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 231, 192, 0, 0, 0, 5, 5, 232, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 227, 5, 5, 38, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 240, 5, 5, 228, 229, 5, 5, 180, 235, 5, 5, 181, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 226, 5, 5, 28, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 240, 5, 5, 27, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 240, 192, 0, 0, 0, 5, 5, 76, 229, 192, 0, 0, 0, 5, 5, 30, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 240, 5, 5, 42, 240, 5, 5, 34, 240, 192, 0, 0, 0, 5, 5, 158, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 240, 5, 5, 27, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 240, 192, 0, 0, 0, 5, 5, 38, 240, 5, 5, 253, 236, 5, 5, 229, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 232, 5, 5, 31, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 227, 5, 5, 41, 240, 5, 5, 149, 231, 5, 5, 65, 226, 5, 5, 35, 240, 5, 5, 178, 226, 5, 5, 40, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 227, 192, 0, 0, 0, 5, 5, 37, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 240, 5, 5, 55, 240, 5, 5, 53, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 240, 5, 5, 16, 237, 192, 0, 0, 0, 5, 5, 51, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 227, 192, 0, 0, 0, 5, 5, 3, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 240, 5, 5, 44, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 240, 5, 5, 32, 240, 192, 0, 0, 0, 5, 5, 180, 226, 5, 5, 52, 240, 5, 5, 54, 240, 5, 5, 213, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 229, 5, 5, 114, 226, 5, 5, 46, 240, 5, 5, 57, 240, 5, 5, 56, 240, 5, 5, 231, 227, 192, 0, 0, 0, 5, 5, 79, 232, 5, 5, 254, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 228, 192, 0, 0, 0, 5, 5, 47, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 240, 192, 0, 0, 0, 5, 5, 62, 240, 5, 5, 64, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 230, 192, 0, 0, 0, 5, 5, 70, 240, 5, 5, 84, 233, 5, 5, 60, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 240, 192, 0, 0, 0, 5, 5, 179, 226, 192, 0, 0, 0, 5, 5, 66, 240, 5, 5, 67, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 240, 5, 5, 65, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 228, 5, 5, 98, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 226, 5, 5, 69, 240, 192, 0, 0, 0, 5, 5, 58, 234, 5, 5, 0, 139, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 234, 5, 5, 29, 240, 5, 5, 75, 240, 5, 5, 79, 240, 5, 5, 63, 236, 192, 0, 0, 0, 5, 5, 72, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 233, 5, 5, 77, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 231, 5, 5, 78, 240, 5, 5, 247, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 237, 192, 0, 0, 0, 5, 5, 74, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 240, 192, 0, 0, 0, 5, 5, 77, 229, 192, 0, 0, 0, 5, 5, 179, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 178, 5, 5, 73, 240, 5, 5, 71, 240, 192, 0, 0, 0, 5, 5, 34, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 240, 5, 5, 194, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 232, 5, 5, 92, 240, 192, 0, 0, 0, 5, 5, 35, 232, 5, 5, 87, 240, 192, 0, 0, 0, 5, 5, 84, 240, 5, 5, 85, 240, 192, 0, 0, 0, 5, 5, 84, 230, 5, 5, 88, 240, 5, 5, 91, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 240, 5, 5, 113, 230, 192, 0, 0, 0, 5, 5, 89, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 240, 192, 0, 0, 0, 5, 5, 86, 236, 5, 5, 85, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 240, 5, 5, 94, 240, 5, 5, 0, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 228, 5, 5, 0, 132, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 227, 192, 0, 0, 0, 5, 5, 90, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 240, 5, 5, 97, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 240, 5, 5, 104, 240, 5, 5, 96, 240, 5, 5, 100, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 240, 192, 0, 0, 0, 5, 5, 106, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 240, 5, 5, 12, 230, 192, 0, 0, 0, 5, 5, 99, 240, 5, 5, 124, 232, 5, 5, 4, 229, 5, 5, 81, 240, 5, 5, 63, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 240, 192, 0, 0, 0, 5, 5, 133, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 240, 5, 5, 110, 240, 5, 5, 111, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 240, 192, 0, 0, 0, 5, 5, 117, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 235, 5, 5, 47, 227, 192, 0, 0, 0, 5, 5, 115, 240, 192, 0, 0, 0, 5, 5, 89, 228, 5, 5, 204, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 236, 5, 5, 114, 240, 5, 5, 0, 180, 5, 5, 116, 240, 192, 0, 0, 0, 5, 5, 107, 240, 5, 5, 113, 240, 5, 5, 120, 240, 5, 5, 88, 228, 5, 5, 108, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 232, 192, 0, 0, 0, 5, 5, 17, 237, 5, 5, 128, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 240, 192, 0, 0, 0, 5, 5, 122, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 240, 5, 5, 126, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 233, 192, 0, 0, 0, 5, 5, 90, 228, 192, 0, 0, 0, 5, 5, 129, 240, 5, 5, 124, 240, 192, 0, 0, 0, 5, 5, 130, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 240, 5, 5, 89, 229, 5, 5, 135, 240, 5, 5, 133, 240, 5, 5, 132, 240, 5, 5, 139, 240, 5, 5, 131, 240, 5, 5, 141, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 233, 192, 0, 0, 0, 5, 5, 146, 240, 192, 0, 0, 0, 5, 5, 144, 240, 5, 5, 134, 240, 5, 5, 144, 228, 192, 0, 0, 0, 5, 5, 145, 240, 192, 0, 0, 0, 5, 5, 149, 240, 5, 5, 148, 240, 5, 5, 150, 240, 5, 5, 147, 240, 5, 5, 151, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 240, 5, 5, 153, 240, 5, 5, 134, 235, 192, 0, 0, 0, 5, 5, 155, 240, 5, 5, 154, 240, 5, 5, 139, 230, 192, 0, 0, 0, 5, 5, 150, 231, 5, 5, 160, 226, 5, 5, 182, 226, 192, 0, 0, 0, 5, 5, 156, 240, 192, 0, 0, 0, 5, 5, 0, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 231, 5, 5, 157, 240, 192, 0, 0, 0, 5, 5, 200, 247, 5, 5, 158, 240, 5, 5, 117, 228, 192, 0, 0, 0, 5, 5, 159, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 240, 5, 5, 144, 227, 5, 5, 162, 240, 192, 0, 0, 0, 5, 5, 163, 240, 5, 5, 164, 240, 5, 5, 125, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 235, 5, 5, 182, 230, 5, 5, 165, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 231, 5, 5, 128, 248, 5, 5, 192, 234, 192, 0, 0, 0, 5, 5, 88, 230, 192, 0, 0, 0, 5, 5, 125, 229, 5, 5, 166, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 232, 192, 0, 0, 0, 5, 5, 67, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 240, 192, 0, 0, 0, 5, 5, 170, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 149, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 240, 192, 0, 0, 0, 5, 5, 175, 240, 5, 5, 223, 230, 5, 5, 145, 227, 192, 0, 0, 0, 5, 5, 173, 240, 5, 5, 180, 240, 5, 5, 224, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 240, 5, 5, 68, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 236, 5, 5, 176, 240, 5, 5, 177, 240, 5, 5, 181, 240, 5, 5, 172, 233, 5, 5, 178, 240, 5, 5, 5, 229, 5, 5, 198, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 240, 5, 5, 146, 234, 192, 0, 0, 0, 5, 5, 152, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 234, 5, 5, 10, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 235, 192, 0, 0, 0, 5, 5, 184, 240, 5, 5, 91, 226, 5, 5, 238, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 232, 5, 5, 187, 240, 5, 5, 194, 240, 5, 5, 189, 240, 5, 5, 196, 240, 5, 5, 192, 240, 192, 0, 0, 0, 5, 5, 191, 240, 5, 5, 111, 234, 192, 0, 0, 0, 5, 5, 185, 240, 5, 5, 183, 226, 5, 5, 183, 240, 5, 5, 199, 227, 5, 5, 153, 232, 5, 5, 179, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 240, 5, 5, 6, 229, 5, 5, 195, 231, 192, 0, 0, 0, 5, 5, 225, 230, 5, 5, 190, 240, 5, 5, 82, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 227, 5, 5, 224, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 226, 5, 5, 43, 231, 5, 5, 198, 240, 5, 5, 203, 240, 192, 0, 0, 0, 5, 5, 199, 240, 192, 0, 0, 0, 5, 5, 145, 228, 192, 0, 0, 0, 5, 5, 204, 240, 5, 5, 176, 229, 5, 5, 59, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 230, 5, 5, 186, 240, 192, 0, 0, 0, 5, 5, 13, 230, 5, 5, 201, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 229, 5, 5, 202, 240, 5, 5, 0, 162, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 240, 5, 5, 0, 133, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 234, 5, 5, 206, 240, 5, 5, 38, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 229, 5, 5, 207, 240, 192, 0, 0, 0, 5, 5, 209, 240, 5, 5, 205, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 235, 192, 0, 0, 0, 5, 5, 33, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 235, 5, 5, 60, 230, 5, 5, 221, 240, 192, 0, 0, 0, 5, 5, 220, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 240, 5, 5, 227, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 234, 5, 5, 49, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 230, 5, 5, 217, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 230, 192, 0, 0, 0, 5, 5, 211, 240, 5, 5, 216, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 240, 192, 0, 0, 0, 5, 5, 53, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 240, 5, 5, 193, 236, 5, 5, 126, 229, 5, 5, 185, 232, 5, 5, 215, 240, 192, 0, 0, 0, 5, 5, 196, 231, 192, 0, 0, 0, 5, 5, 7, 229, 5, 5, 110, 231, 5, 5, 66, 226, 5, 5, 4, 232, 5, 5, 213, 240, 5, 5, 155, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 228, 192, 0, 0, 0, 5, 5, 51, 233, 5, 5, 219, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 240, 192, 0, 0, 0, 5, 5, 224, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 232, 5, 5, 230, 240, 192, 0, 0, 0, 5, 5, 225, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 234, 5, 5, 87, 233, 192, 0, 0, 0, 5, 5, 228, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 236, 5, 5, 49, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 140, 192, 0, 0, 0, 5, 5, 226, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 226, 192, 0, 0, 0, 5, 5, 229, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 232, 192, 0, 0, 0, 5, 5, 240, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 240, 5, 5, 238, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 240, 192, 0, 0, 0, 5, 5, 239, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 234, 5, 5, 173, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 240, 192, 0, 0, 0, 5, 5, 236, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 240, 192, 0, 0, 0, 5, 5, 218, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 240, 192, 0, 0, 0, 5, 5, 40, 236, 192, 0, 0, 0, 5, 5, 138, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 229, 5, 5, 247, 240, 192, 0, 0, 0, 5, 5, 246, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 233, 5, 5, 248, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 240, 192, 0, 0, 0, 5, 5, 45, 235, 192, 0, 0, 0, 5, 5, 69, 234, 5, 5, 177, 229, 192, 0, 0, 0, 5, 5, 217, 231, 192, 0, 0, 0, 5, 5, 202, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 226, 192, 0, 0, 0, 5, 5, 2, 241, 5, 5, 251, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 236, 5, 5, 4, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 240, 192, 0, 0, 0, 5, 5, 255, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 240, 192, 0, 0, 0, 5, 5, 188, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 241, 192, 0, 0, 0, 5, 5, 252, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 241, 5, 5, 9, 241, 5, 5, 105, 233, 5, 5, 11, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 229, 5, 5, 6, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 241, 192, 0, 0, 0, 5, 5, 5, 241, 5, 5, 16, 241, 192, 0, 0, 0, 5, 5, 15, 241, 192, 0, 0, 0, 5, 5, 14, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 241, 5, 5, 29, 231, 192, 0, 0, 0, 5, 5, 18, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 241, 5, 5, 237, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 241, 5, 5, 23, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 240, 5, 5, 25, 241, 5, 5, 13, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 241, 5, 5, 27, 241, 5, 5, 29, 241, 5, 5, 28, 241, 5, 5, 30, 241, 5, 5, 184, 226, 192, 0, 0, 0, 5, 5, 8, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 235, 5, 5, 151, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 241, 192, 0, 0, 0, 5, 5, 16, 235, 192, 0, 0, 0, 5, 5, 181, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 241, 5, 5, 33, 241, 5, 5, 84, 234, 5, 5, 36, 241, 5, 5, 233, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 241, 5, 5, 37, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 227, 5, 5, 196, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 231, 192, 0, 0, 0, 5, 5, 39, 241, 192, 0, 0, 0, 5, 5, 152, 231, 5, 5, 106, 233, 192, 0, 0, 0, 5, 5, 29, 235, 5, 5, 40, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 241, 5, 5, 42, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 235, 5, 5, 96, 239, 5, 5, 174, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 235, 192, 0, 0, 0, 5, 5, 136, 229, 192, 0, 0, 0, 5, 5, 195, 234, 5, 5, 161, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 233, 192, 0, 0, 0, 5, 5, 215, 236, 192, 0, 0, 0, 5, 5, 44, 241, 5, 5, 62, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 241, 192, 0, 0, 0, 5, 5, 0, 148, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 228, 5, 5, 181, 231, 192, 0, 0, 0, 5, 5, 30, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 241, 5, 5, 208, 229, 5, 5, 203, 232, 192, 0, 0, 0, 5, 5, 233, 229, 5, 5, 65, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 241, 192, 0, 0, 0, 5, 5, 153, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 226, 5, 5, 234, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 241, 192, 0, 0, 0, 5, 5, 48, 241, 5, 5, 51, 241, 5, 5, 207, 236, 5, 5, 49, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 231, 5, 5, 52, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 227, 192, 0, 0, 0, 5, 5, 55, 241, 192, 0, 0, 0, 5, 5, 54, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 241, 5, 5, 57, 241, 5, 5, 115, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 234, 5, 5, 186, 232, 5, 5, 194, 227, 5, 5, 235, 229, 5, 5, 41, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 230, 5, 5, 0, 142, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 226, 5, 5, 62, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 229, 5, 5, 61, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 229, 5, 5, 228, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 241, 192, 0, 0, 0, 5, 5, 229, 230, 192, 0, 0, 0, 5, 5, 26, 236, 5, 5, 90, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 181, 5, 5, 182, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 231, 5, 5, 35, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 230, 192, 0, 0, 0, 5, 5, 224, 235, 5, 5, 162, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 230, 192, 0, 0, 0, 5, 5, 143, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 241, 5, 5, 64, 241, 5, 5, 65, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 241, 192, 0, 0, 0, 5, 5, 71, 241, 5, 5, 14, 230, 5, 5, 10, 229, 5, 5, 69, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 241, 192, 0, 0, 0, 5, 5, 66, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 241, 5, 5, 72, 241, 5, 5, 77, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 241, 192, 0, 0, 0, 5, 5, 74, 241, 192, 0, 0, 0, 5, 5, 186, 226, 5, 5, 75, 241, 5, 5, 76, 241, 5, 5, 179, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 235, 5, 5, 94, 228, 5, 5, 79, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 231, 192, 0, 0, 0, 5, 5, 231, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 227, 192, 0, 0, 0, 5, 5, 80, 241, 5, 5, 84, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 226, 5, 5, 81, 241, 5, 5, 83, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 232, 5, 5, 0, 163, 5, 5, 85, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 241, 5, 5, 88, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 241, 192, 0, 0, 0, 5, 5, 90, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 233, 192, 0, 0, 0, 5, 5, 89, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 230, 5, 5, 95, 241, 192, 0, 0, 0, 5, 5, 131, 236, 5, 5, 120, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 241, 192, 0, 0, 0, 5, 5, 254, 227, 5, 5, 36, 226, 5, 5, 11, 229, 5, 5, 101, 241, 192, 0, 0, 0, 5, 5, 102, 241, 5, 5, 189, 230, 5, 5, 42, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 238, 5, 5, 6, 232, 5, 5, 5, 232, 5, 5, 126, 232, 5, 5, 119, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 228, 5, 5, 101, 236, 192, 0, 0, 0, 5, 5, 154, 235, 192, 0, 0, 0, 5, 5, 60, 235, 192, 0, 0, 0, 5, 5, 103, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 229, 5, 5, 35, 233, 5, 5, 104, 241, 5, 5, 37, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 235, 192, 0, 0, 0, 5, 5, 15, 233, 5, 5, 105, 241, 5, 5, 116, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 241, 5, 5, 107, 241, 5, 5, 52, 236, 192, 0, 0, 0, 5, 5, 252, 235, 5, 5, 240, 235, 5, 5, 214, 235, 5, 5, 179, 229, 5, 5, 109, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 235, 192, 0, 0, 0, 5, 5, 111, 241, 5, 5, 114, 241, 5, 5, 113, 241, 192, 0, 0, 0, 5, 5, 113, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 227, 192, 0, 0, 0, 5, 5, 110, 241, 192, 0, 0, 0, 5, 5, 112, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 236, 5, 5, 0, 167, 5, 5, 144, 229, 5, 5, 98, 232, 192, 0, 0, 0, 5, 5, 75, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 233, 192, 0, 0, 0, 5, 5, 116, 241, 5, 5, 81, 232, 5, 5, 117, 241, 5, 5, 30, 231, 5, 5, 55, 236, 5, 5, 119, 241, 5, 5, 120, 241, 5, 5, 160, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 229, 192, 0, 0, 0, 5, 5, 85, 234, 5, 5, 122, 241, 5, 5, 174, 233, 5, 5, 59, 241, 5, 5, 63, 241, 192, 0, 0, 0, 5, 5, 165, 227, 192, 0, 0, 0, 5, 5, 71, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 241, 192, 0, 0, 0, 5, 5, 232, 230, 5, 5, 162, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 241, 192, 0, 0, 0, 5, 5, 216, 234, 192, 0, 0, 0, 5, 5, 121, 241, 192, 0, 0, 0, 5, 5, 127, 241, 5, 5, 126, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 235, 192, 0, 0, 0, 5, 5, 233, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 235, 192, 0, 0, 0, 5, 5, 134, 226, 5, 5, 236, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 237, 5, 5, 129, 241, 5, 5, 127, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 241, 5, 5, 133, 241, 192, 0, 0, 0, 5, 5, 135, 226, 5, 5, 131, 241, 5, 5, 137, 241, 5, 5, 143, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 235, 192, 0, 0, 0, 5, 5, 145, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 241, 5, 5, 112, 234, 5, 5, 185, 235, 5, 5, 51, 227, 192, 0, 0, 0, 5, 5, 223, 231, 5, 5, 140, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 233, 192, 0, 0, 0, 5, 5, 102, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 241, 5, 5, 139, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 241, 192, 0, 0, 0, 5, 5, 138, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 241, 192, 0, 0, 0, 5, 5, 136, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 241, 192, 0, 0, 0, 5, 5, 142, 241, 5, 5, 132, 241, 192, 0, 0, 0, 5, 5, 240, 232, 192, 0, 0, 0, 5, 5, 82, 236, 5, 5, 52, 230, 5, 5, 164, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 235, 5, 5, 217, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 233, 5, 5, 234, 233, 5, 5, 37, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 231, 5, 5, 62, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 229, 5, 5, 22, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 241, 5, 5, 14, 227, 5, 5, 156, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 241, 192, 0, 0, 0, 5, 5, 219, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 226, 5, 5, 92, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 226, 5, 5, 127, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 241, 5, 5, 121, 228, 5, 5, 95, 228, 5, 5, 175, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 241, 192, 0, 0, 0, 5, 5, 0, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 241, 5, 5, 154, 241, 192, 0, 0, 0, 5, 5, 3, 228, 5, 5, 64, 228, 192, 0, 0, 0, 5, 5, 52, 227, 5, 5, 161, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 229, 5, 5, 235, 235, 192, 0, 0, 0, 5, 5, 197, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 241, 192, 0, 0, 0, 5, 5, 109, 226, 5, 5, 159, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 241, 5, 5, 160, 241, 192, 0, 0, 0, 5, 5, 216, 236, 192, 0, 0, 0, 5, 5, 167, 241, 192, 0, 0, 0, 5, 5, 95, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 241, 192, 0, 0, 0, 5, 5, 162, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 146, 5, 5, 164, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 241, 192, 0, 0, 0, 5, 5, 165, 241, 192, 0, 0, 0, 5, 5, 161, 241, 5, 5, 172, 241, 192, 0, 0, 0, 5, 5, 233, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 239, 5, 5, 218, 228, 5, 5, 180, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 241, 192, 0, 0, 0, 5, 5, 88, 233, 5, 5, 187, 226, 5, 5, 93, 229, 192, 0, 0, 0, 5, 5, 155, 241, 192, 0, 0, 0, 5, 5, 171, 241, 5, 5, 249, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 241, 5, 5, 173, 241, 192, 0, 0, 0, 5, 5, 176, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 227, 192, 0, 0, 0, 5, 5, 203, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 236, 5, 5, 178, 241, 5, 5, 117, 227, 192, 0, 0, 0, 5, 5, 185, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 235, 192, 0, 0, 0, 5, 5, 186, 241, 5, 5, 188, 241, 192, 0, 0, 0, 5, 5, 192, 241, 5, 5, 180, 241, 192, 0, 0, 0, 5, 5, 174, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 233, 5, 5, 196, 241, 5, 5, 183, 241, 192, 0, 0, 0, 5, 5, 193, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 231, 5, 5, 197, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 241, 5, 5, 53, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 237, 5, 5, 177, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 241, 5, 5, 0, 182, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 236, 5, 5, 184, 241, 5, 5, 44, 231, 5, 5, 43, 233, 5, 5, 174, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 231, 5, 5, 200, 241, 5, 5, 11, 227, 5, 5, 147, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 241, 5, 5, 181, 241, 5, 5, 201, 241, 192, 0, 0, 0, 5, 5, 194, 241, 5, 5, 182, 241, 192, 0, 0, 0, 5, 5, 198, 241, 192, 0, 0, 0, 5, 5, 199, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 233, 192, 0, 0, 0, 5, 5, 189, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 241, 192, 0, 0, 0, 5, 5, 60, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 235, 5, 5, 209, 241, 5, 5, 112, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 241, 5, 5, 7, 232, 192, 0, 0, 0, 5, 5, 206, 241, 5, 5, 219, 241, 5, 5, 218, 241, 192, 0, 0, 0, 5, 5, 17, 234, 5, 5, 217, 241, 5, 5, 12, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 241, 5, 5, 208, 241, 192, 0, 0, 0, 5, 5, 252, 227, 5, 5, 211, 241, 5, 5, 169, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 234, 5, 5, 213, 241, 5, 5, 255, 227, 192, 0, 0, 0, 5, 5, 205, 241, 5, 5, 207, 241, 5, 5, 204, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 237, 5, 5, 242, 226, 5, 5, 210, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 241, 5, 5, 205, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 237, 5, 5, 240, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 231, 5, 5, 239, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 241, 5, 5, 222, 241, 192, 0, 0, 0, 5, 5, 241, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 241, 5, 5, 225, 241, 192, 0, 0, 0, 5, 5, 226, 241, 192, 0, 0, 0, 5, 5, 234, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 241, 5, 5, 15, 229, 5, 5, 44, 233, 5, 5, 44, 232, 5, 5, 229, 241, 192, 0, 0, 0, 5, 5, 224, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 236, 5, 5, 228, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 241, 5, 5, 242, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 241, 5, 5, 243, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 242, 192, 0, 0, 0, 5, 5, 254, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 241, 192, 0, 0, 0, 5, 5, 52, 233, 192, 0, 0, 0, 5, 5, 45, 232, 192, 0, 0, 0, 5, 5, 246, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 242, 5, 5, 211, 234, 5, 5, 9, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 242, 5, 5, 7, 242, 5, 5, 255, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 232, 192, 0, 0, 0, 5, 5, 244, 234, 192, 0, 0, 0, 5, 5, 245, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 241, 5, 5, 234, 230, 192, 0, 0, 0, 5, 5, 41, 236, 5, 5, 19, 242, 5, 5, 6, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 228, 5, 5, 93, 226, 5, 5, 247, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 230, 5, 5, 11, 242, 192, 0, 0, 0, 5, 5, 18, 242, 5, 5, 103, 230, 5, 5, 12, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 242, 5, 5, 17, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 227, 5, 5, 15, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 227, 192, 0, 0, 0, 5, 5, 235, 233, 5, 5, 14, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 226, 5, 5, 205, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 242, 5, 5, 219, 228, 192, 0, 0, 0, 5, 5, 20, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 242, 192, 0, 0, 0, 5, 5, 24, 242, 5, 5, 25, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 242, 192, 0, 0, 0, 5, 5, 33, 242, 192, 0, 0, 0, 5, 5, 168, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 242, 5, 5, 30, 242, 5, 5, 29, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 242, 192, 0, 0, 0, 5, 5, 28, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 242, 5, 5, 49, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 242, 5, 5, 36, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 234, 192, 0, 0, 0, 5, 5, 39, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 242, 5, 5, 40, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 236, 5, 5, 41, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 226, 192, 0, 0, 0, 5, 5, 47, 242, 5, 5, 123, 228, 5, 5, 15, 230, 192, 0, 0, 0, 5, 5, 11, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 242, 5, 5, 48, 242, 5, 5, 51, 242, 5, 5, 147, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 228, 5, 5, 54, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 242, 192, 0, 0, 0, 5, 5, 55, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 226, 192, 0, 0, 0, 5, 5, 187, 232, 192, 0, 0, 0, 5, 5, 56, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 227, 5, 5, 58, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 242, 192, 0, 0, 0, 5, 5, 59, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 242, 192, 0, 0, 0, 5, 5, 61, 242, 5, 5, 237, 229, 5, 5, 157, 231, 5, 5, 81, 229, 192, 0, 0, 0, 5, 5, 46, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 235, 5, 5, 58, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 229, 5, 5, 10, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 242, 5, 5, 63, 242, 192, 0, 0, 0, 5, 5, 238, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 242, 5, 5, 65, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 242, 5, 5, 66, 242, 192, 0, 0, 0, 5, 5, 210, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 230, 5, 5, 90, 230, 5, 5, 210, 229, 192, 0, 0, 0, 5, 5, 68, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 242, 5, 5, 45, 231, 192, 0, 0, 0, 5, 5, 69, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 242, 5, 5, 74, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 242, 192, 0, 0, 0, 5, 5, 77, 242, 5, 5, 76, 242, 5, 5, 78, 242, 5, 5, 95, 226, 5, 5, 206, 232, 192, 0, 0, 0, 5, 5, 79, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 229, 5, 5, 228, 226, 5, 5, 80, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 233, 5, 5, 239, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 227, 5, 5, 81, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 242, 192, 0, 0, 0, 5, 5, 136, 235, 5, 5, 226, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 233, 5, 5, 83, 242, 5, 5, 196, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 242, 5, 5, 85, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 229, 192, 0, 0, 0, 5, 5, 11, 236, 192, 0, 0, 0, 5, 5, 91, 242, 5, 5, 92, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 242, 5, 5, 94, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 234, 5, 5, 38, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 234, 192, 0, 0, 0, 5, 5, 89, 233, 5, 5, 141, 230, 5, 5, 183, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 234, 192, 0, 0, 0, 5, 5, 28, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 242, 192, 0, 0, 0, 5, 5, 56, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 234, 5, 5, 96, 242, 5, 5, 16, 229, 5, 5, 215, 232, 192, 0, 0, 0, 5, 5, 98, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 242, 192, 0, 0, 0, 5, 5, 99, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 232, 192, 0, 0, 0, 5, 5, 184, 227, 5, 5, 107, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 227, 5, 5, 105, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 242, 5, 5, 100, 242, 5, 5, 152, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 233, 5, 5, 101, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 242, 192, 0, 0, 0, 5, 5, 108, 242, 5, 5, 55, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 229, 5, 5, 102, 242, 5, 5, 104, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 235, 5, 5, 154, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 242, 192, 0, 0, 0, 5, 5, 137, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 235, 5, 5, 87, 236, 5, 5, 120, 242, 5, 5, 17, 230, 5, 5, 236, 230, 5, 5, 113, 242, 5, 5, 119, 242, 5, 5, 68, 226, 192, 0, 0, 0, 5, 5, 235, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 242, 5, 5, 115, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 231, 5, 5, 113, 234, 192, 0, 0, 0, 5, 5, 197, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 242, 192, 0, 0, 0, 5, 5, 155, 235, 192, 0, 0, 0, 5, 5, 114, 242, 192, 0, 0, 0, 5, 5, 123, 242, 192, 0, 0, 0, 5, 5, 121, 242, 192, 0, 0, 0, 5, 5, 116, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 235, 5, 5, 72, 234, 5, 5, 185, 227, 192, 0, 0, 0, 5, 5, 103, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 232, 192, 0, 0, 0, 5, 5, 124, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 242, 5, 5, 127, 232, 5, 5, 111, 242, 192, 0, 0, 0, 5, 5, 39, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 236, 5, 5, 135, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 231, 192, 0, 0, 0, 5, 5, 131, 242, 192, 0, 0, 0, 5, 5, 164, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 233, 5, 5, 125, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 226, 5, 5, 17, 229, 5, 5, 128, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 230, 5, 5, 133, 242, 192, 0, 0, 0, 5, 5, 132, 242, 5, 5, 127, 242, 192, 0, 0, 0, 5, 5, 130, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 227, 192, 0, 0, 0, 5, 5, 129, 242, 5, 5, 73, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 231, 5, 5, 221, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 242, 5, 5, 139, 242, 192, 0, 0, 0, 5, 5, 11, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 242, 5, 5, 138, 242, 192, 0, 0, 0, 5, 5, 22, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 229, 5, 5, 40, 237, 192, 0, 0, 0, 5, 5, 213, 226, 192, 0, 0, 0, 5, 5, 32, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 226, 5, 5, 69, 231, 5, 5, 140, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 236, 192, 0, 0, 0, 5, 5, 142, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 242, 192, 0, 0, 0, 5, 5, 143, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 236, 192, 0, 0, 0, 5, 5, 181, 233, 5, 5, 226, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 242, 5, 5, 153, 242, 5, 5, 150, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 236, 5, 5, 156, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 230, 5, 5, 158, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 233, 5, 5, 161, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 242, 192, 0, 0, 0, 5, 5, 151, 242, 192, 0, 0, 0, 5, 5, 157, 242, 192, 0, 0, 0, 5, 5, 164, 242, 5, 5, 0, 218, 5, 5, 154, 242, 192, 0, 0, 0, 5, 5, 165, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 231, 192, 0, 0, 0, 5, 5, 171, 230, 192, 0, 0, 0, 5, 5, 65, 235, 192, 0, 0, 0, 5, 5, 94, 229, 192, 0, 0, 0, 5, 5, 146, 242, 5, 5, 160, 242, 5, 5, 124, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 231, 192, 0, 0, 0, 5, 5, 3, 227, 5, 5, 129, 229, 5, 5, 238, 230, 5, 5, 148, 242, 5, 5, 142, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 228, 192, 0, 0, 0, 5, 5, 147, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 242, 5, 5, 185, 230, 5, 5, 172, 228, 192, 0, 0, 0, 5, 5, 184, 242, 192, 0, 0, 0, 5, 5, 178, 242, 5, 5, 202, 227, 5, 5, 145, 233, 192, 0, 0, 0, 5, 5, 173, 242, 5, 5, 182, 242, 5, 5, 0, 141, 5, 5, 14, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 226, 192, 0, 0, 0, 5, 5, 175, 242, 5, 5, 82, 232, 5, 5, 166, 242, 5, 5, 168, 242, 5, 5, 19, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 242, 192, 0, 0, 0, 5, 5, 180, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 242, 5, 5, 181, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 228, 192, 0, 0, 0, 5, 5, 239, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 242, 5, 5, 180, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 237, 5, 5, 44, 230, 5, 5, 248, 235, 192, 0, 0, 0, 5, 5, 186, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 242, 5, 5, 173, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 230, 192, 0, 0, 0, 5, 5, 188, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 236, 5, 5, 20, 229, 192, 0, 0, 0, 5, 5, 201, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 236, 5, 5, 190, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 242, 192, 0, 0, 0, 5, 5, 194, 242, 5, 5, 33, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 242, 192, 0, 0, 0, 5, 5, 16, 230, 5, 5, 213, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 227, 192, 0, 0, 0, 5, 5, 191, 242, 5, 5, 196, 242, 5, 5, 197, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 232, 5, 5, 128, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 242, 192, 0, 0, 0, 5, 5, 107, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 242, 5, 5, 206, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 242, 5, 5, 183, 242, 192, 0, 0, 0, 5, 5, 207, 227, 5, 5, 246, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 237, 192, 0, 0, 0, 5, 5, 203, 242, 192, 0, 0, 0, 5, 5, 215, 242, 5, 5, 69, 226, 5, 5, 46, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 234, 192, 0, 0, 0, 5, 5, 57, 227, 5, 5, 18, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 235, 5, 5, 54, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 242, 5, 5, 212, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 242, 5, 5, 208, 242, 192, 0, 0, 0, 5, 5, 202, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 242, 5, 5, 224, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 230, 192, 0, 0, 0, 5, 5, 232, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 242, 5, 5, 16, 233, 5, 5, 222, 242, 5, 5, 61, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 242, 192, 0, 0, 0, 5, 5, 218, 242, 192, 0, 0, 0, 5, 5, 227, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 242, 5, 5, 220, 242, 5, 5, 226, 242, 192, 0, 0, 0, 5, 5, 137, 231, 192, 0, 0, 0, 5, 5, 217, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 242, 192, 0, 0, 0, 5, 5, 234, 242, 5, 5, 236, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 233, 192, 0, 0, 0, 5, 5, 233, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 228, 5, 5, 159, 232, 5, 5, 231, 242, 5, 5, 59, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 242, 5, 5, 241, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 242, 5, 5, 60, 229, 5, 5, 40, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 236, 5, 5, 242, 242, 192, 0, 0, 0, 5, 5, 246, 242, 5, 5, 155, 232, 192, 0, 0, 0, 5, 5, 245, 242, 192, 0, 0, 0, 5, 5, 224, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 242, 192, 0, 0, 0, 5, 5, 249, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 242, 192, 0, 0, 0, 5, 5, 251, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 243, 192, 0, 0, 0, 5, 5, 2, 243, 5, 5, 255, 242, 192, 0, 0, 0, 5, 5, 4, 243, 5, 5, 244, 233, 5, 5, 6, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 232, 5, 5, 147, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 243, 192, 0, 0, 0, 5, 5, 9, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 233, 5, 5, 189, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 230, 5, 5, 130, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 237, 5, 5, 112, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 243, 5, 5, 190, 232, 5, 5, 19, 243, 5, 5, 14, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 243, 5, 5, 130, 233, 5, 5, 0, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 243, 192, 0, 0, 0, 5, 5, 20, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 243, 5, 5, 78, 235, 192, 0, 0, 0, 5, 5, 26, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 243, 192, 0, 0, 0, 5, 5, 225, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 243, 5, 5, 34, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 243, 192, 0, 0, 0, 5, 5, 97, 234, 5, 5, 28, 243, 5, 5, 31, 243, 5, 5, 242, 230, 192, 0, 0, 0, 5, 5, 175, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 243, 192, 0, 0, 0, 5, 5, 63, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 243, 192, 0, 0, 0, 5, 5, 59, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 236, 5, 5, 39, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 233, 5, 5, 46, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 243, 192, 0, 0, 0, 5, 5, 235, 236, 192, 0, 0, 0, 5, 5, 45, 243, 192, 0, 0, 0, 5, 5, 47, 243, 5, 5, 73, 226, 192, 0, 0, 0, 5, 5, 42, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 238, 5, 5, 49, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 232, 5, 5, 198, 229, 5, 5, 51, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 243, 5, 5, 46, 231, 5, 5, 64, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 243, 192, 0, 0, 0, 5, 5, 37, 243, 5, 5, 53, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 243, 192, 0, 0, 0, 5, 5, 66, 233, 192, 0, 0, 0, 5, 5, 61, 243, 5, 5, 60, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 243, 192, 0, 0, 0, 5, 5, 63, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 230, 5, 5, 33, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 236, 5, 5, 64, 243, 5, 5, 65, 243, 5, 5, 30, 232, 5, 5, 18, 230, 5, 5, 66, 243, 5, 5, 67, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 243, 5, 5, 110, 235, 5, 5, 165, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 243, 5, 5, 87, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 243, 5, 5, 161, 226, 192, 0, 0, 0, 5, 5, 195, 227, 192, 0, 0, 0, 5, 5, 32, 236, 192, 0, 0, 0, 5, 5, 16, 236, 192, 0, 0, 0, 5, 5, 110, 226, 5, 5, 42, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 235, 192, 0, 0, 0, 5, 5, 70, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 231, 192, 0, 0, 0, 5, 5, 71, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 228, 5, 5, 72, 243, 192, 0, 0, 0, 5, 5, 132, 229, 5, 5, 74, 243, 5, 5, 73, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 227, 192, 0, 0, 0, 5, 5, 78, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 227, 5, 5, 82, 243, 5, 5, 84, 243, 192, 0, 0, 0, 5, 5, 83, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 243, 192, 0, 0, 0, 5, 5, 195, 228, 192, 0, 0, 0, 5, 5, 86, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 228, 192, 0, 0, 0, 5, 5, 8, 232, 192, 0, 0, 0, 5, 5, 79, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 243, 5, 5, 89, 243, 5, 5, 87, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 233, 5, 5, 237, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 243, 5, 5, 176, 232, 5, 5, 90, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 237, 5, 5, 98, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 243, 5, 5, 93, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 236, 5, 5, 95, 243, 5, 5, 97, 243, 192, 0, 0, 0, 5, 5, 218, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 243, 5, 5, 250, 232, 5, 5, 45, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 228, 5, 5, 99, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 243, 192, 0, 0, 0, 5, 5, 105, 236, 5, 5, 106, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 243, 5, 5, 74, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 229, 5, 5, 240, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 243, 5, 5, 104, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 243, 192, 0, 0, 0, 5, 5, 107, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 243, 192, 0, 0, 0, 5, 5, 229, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 243, 192, 0, 0, 0, 5, 5, 113, 243, 5, 5, 112, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 236, 192, 0, 0, 0, 5, 5, 2, 228, 192, 0, 0, 0, 5, 5, 96, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 236, 5, 5, 115, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 243, 192, 0, 0, 0, 5, 5, 139, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 243, 192, 0, 0, 0, 5, 5, 199, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 233, 5, 5, 116, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 243, 192, 0, 0, 0, 5, 5, 92, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 234, 5, 5, 120, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 227, 192, 0, 0, 0, 5, 5, 123, 243, 5, 5, 181, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 243, 5, 5, 236, 236, 5, 5, 13, 228, 5, 5, 218, 234, 5, 5, 74, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 243, 192, 0, 0, 0, 5, 5, 133, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 243, 5, 5, 220, 228, 5, 5, 41, 226, 5, 5, 134, 243, 192, 0, 0, 0, 5, 5, 122, 231, 5, 5, 131, 243, 5, 5, 242, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 243, 5, 5, 159, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 243, 5, 5, 138, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 243, 5, 5, 136, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 243, 192, 0, 0, 0, 5, 5, 247, 234, 5, 5, 149, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 227, 5, 5, 150, 243, 192, 0, 0, 0, 5, 5, 151, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 243, 192, 0, 0, 0, 5, 5, 154, 243, 5, 5, 155, 243, 5, 5, 153, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 235, 5, 5, 157, 243, 5, 5, 156, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 243, 5, 5, 158, 243, 5, 5, 160, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 243, 5, 5, 163, 243, 5, 5, 162, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 229, 192, 0, 0, 0, 5, 5, 165, 243, 192, 0, 0, 0, 5, 5, 164, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 227, 192, 0, 0, 0, 5, 5, 93, 231, 192, 0, 0, 0, 5, 5, 126, 233, 192, 0, 0, 0, 5, 5, 166, 243, 5, 5, 160, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 229, 192, 0, 0, 0, 5, 5, 85, 226, 5, 5, 167, 243, 192, 0, 0, 0, 5, 5, 137, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 235, 5, 5, 168, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 233, 5, 5, 107, 236, 5, 5, 21, 229, 5, 5, 71, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 232, 5, 5, 27, 238, 192, 0, 0, 0, 5, 5, 18, 233, 5, 5, 162, 226, 5, 5, 169, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 243, 192, 0, 0, 0, 5, 5, 175, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 243, 5, 5, 172, 243, 5, 5, 31, 241, 5, 5, 190, 226, 5, 5, 171, 243, 192, 0, 0, 0, 5, 5, 0, 184, 192, 0, 0, 0, 5, 5, 136, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 236, 5, 5, 176, 243, 5, 5, 174, 243, 5, 5, 224, 232, 5, 5, 142, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 234, 192, 0, 0, 0, 5, 5, 233, 234, 192, 0, 0, 0, 5, 5, 178, 243, 5, 5, 194, 236, 5, 5, 97, 228, 5, 5, 179, 243, 192, 0, 0, 0, 5, 5, 177, 243, 5, 5, 180, 234, 5, 5, 180, 243, 192, 0, 0, 0, 5, 5, 181, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 185, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 231, 5, 5, 186, 243, 192, 0, 0, 0, 5, 5, 183, 243, 5, 5, 15, 234, 5, 5, 182, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 243, 5, 5, 185, 243, 192, 0, 0, 0, 5, 5, 188, 243, 5, 5, 187, 243, 5, 5, 225, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 232, 5, 5, 9, 232, 192, 0, 0, 0, 5, 5, 149, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 243, 192, 0, 0, 0, 5, 5, 193, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 243, 5, 5, 198, 234, 5, 5, 196, 243, 192, 0, 0, 0, 5, 5, 198, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 243, 5, 5, 72, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 243, 5, 5, 199, 243, 5, 5, 46, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 243, 5, 5, 197, 243, 192, 0, 0, 0, 5, 5, 254, 234, 192, 0, 0, 0, 5, 5, 243, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 243, 192, 0, 0, 0, 5, 5, 20, 230, 5, 5, 95, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 233, 5, 5, 206, 243, 192, 0, 0, 0, 5, 5, 47, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 236, 5, 5, 207, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 243, 192, 0, 0, 0, 5, 5, 215, 243, 5, 5, 216, 243, 5, 5, 216, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 243, 192, 0, 0, 0, 5, 5, 211, 243, 192, 0, 0, 0, 5, 5, 209, 243, 5, 5, 210, 243, 192, 0, 0, 0, 5, 5, 212, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 243, 192, 0, 0, 0, 5, 5, 217, 243, 192, 0, 0, 0, 5, 5, 218, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 243, 5, 5, 222, 243, 5, 5, 223, 243, 5, 5, 224, 243, 192, 0, 0, 0, 5, 5, 225, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 243, 5, 5, 229, 243, 5, 5, 230, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 235, 192, 0, 0, 0, 5, 5, 233, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 243, 5, 5, 235, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 243, 5, 5, 236, 243, 5, 5, 237, 243, 5, 5, 238, 243, 192, 0, 0, 0, 5, 5, 240, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 243, 192, 0, 0, 0, 5, 5, 242, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 243, 192, 0, 0, 0, 5, 5, 244, 243, 192, 0, 0, 0, 5, 5, 141, 234, 5, 5, 146, 233, 5, 5, 245, 243, 5, 5, 114, 234, 5, 5, 240, 234, 192, 0, 0, 0, 5, 5, 246, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 243, 5, 5, 108, 233, 192, 0, 0, 0, 5, 5, 191, 226, 5, 5, 22, 229, 5, 5, 248, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 243, 192, 0, 0, 0, 5, 5, 183, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 243, 5, 5, 254, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 234, 192, 0, 0, 0, 5, 5, 255, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 250, 5, 5, 3, 244, 5, 5, 4, 244, 5, 5, 5, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 244, 5, 5, 86, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 235, 192, 0, 0, 0, 5, 5, 42, 226, 192, 0, 0, 0, 5, 5, 51, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 244, 5, 5, 178, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 231, 5, 5, 50, 242, 192, 0, 0, 0, 5, 5, 10, 244, 5, 5, 27, 236, 192, 0, 0, 0, 5, 5, 11, 244, 192, 0, 0, 0, 5, 5, 62, 227, 5, 5, 181, 234, 5, 5, 12, 244, 192, 0, 0, 0, 5, 5, 13, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 236, 192, 0, 0, 0, 5, 5, 34, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 244, 5, 5, 17, 244, 5, 5, 219, 234, 192, 0, 0, 0, 5, 5, 63, 227, 5, 5, 155, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 244, 5, 5, 73, 231, 5, 5, 12, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 244, 5, 5, 23, 244, 5, 5, 24, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 244, 5, 5, 27, 244, 192, 0, 0, 0, 5, 5, 19, 233, 192, 0, 0, 0, 5, 5, 91, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 244, 5, 5, 32, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 231, 192, 0, 0, 0, 5, 5, 228, 233, 192, 0, 0, 0, 5, 5, 33, 244, 5, 5, 205, 235, 192, 0, 0, 0, 5, 5, 30, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 244, 5, 5, 34, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 244, 192, 0, 0, 0, 5, 5, 40, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 230, 5, 5, 220, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 244, 192, 0, 0, 0, 5, 5, 48, 244, 5, 5, 46, 244, 5, 5, 47, 244, 192, 0, 0, 0, 5, 5, 45, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 244, 5, 5, 17, 236, 5, 5, 53, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 236, 5, 5, 54, 244, 192, 0, 0, 0, 5, 5, 210, 232, 192, 0, 0, 0, 5, 5, 106, 234, 192, 0, 0, 0, 5, 5, 33, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 232, 5, 5, 55, 244, 5, 5, 238, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 244, 192, 0, 0, 0, 5, 5, 152, 228, 5, 5, 133, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 233, 5, 5, 134, 229, 5, 5, 164, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 235, 192, 0, 0, 0, 5, 5, 75, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 236, 5, 5, 23, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 234, 192, 0, 0, 0, 5, 5, 64, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 244, 5, 5, 90, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 244, 5, 5, 207, 226, 5, 5, 63, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 234, 192, 0, 0, 0, 5, 5, 12, 226, 192, 0, 0, 0, 5, 5, 156, 229, 192, 0, 0, 0, 5, 5, 72, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 235, 192, 0, 0, 0, 5, 5, 193, 231, 5, 5, 71, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 226, 192, 0, 0, 0, 5, 5, 78, 244, 192, 0, 0, 0, 5, 5, 77, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 244, 5, 5, 74, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 244, 5, 5, 75, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 234, 5, 5, 73, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 244, 5, 5, 219, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 244, 5, 5, 87, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 244, 5, 5, 91, 244, 5, 5, 90, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 230, 192, 0, 0, 0, 5, 5, 255, 236, 192, 0, 0, 0, 5, 5, 64, 230, 192, 0, 0, 0, 5, 5, 92, 244, 5, 5, 75, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 227, 5, 5, 125, 227, 5, 5, 241, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 244, 192, 0, 0, 0, 5, 5, 97, 244, 5, 5, 12, 232, 5, 5, 94, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 230, 5, 5, 74, 231, 5, 5, 95, 244, 5, 5, 93, 244, 192, 0, 0, 0, 5, 5, 42, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 244, 5, 5, 122, 244, 5, 5, 14, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 237, 5, 5, 251, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 226, 5, 5, 91, 233, 5, 5, 61, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 227, 5, 5, 103, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 244, 192, 0, 0, 0, 5, 5, 41, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 244, 5, 5, 109, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 228, 5, 5, 141, 226, 5, 5, 229, 233, 5, 5, 116, 230, 5, 5, 242, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 244, 192, 0, 0, 0, 5, 5, 117, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 226, 5, 5, 255, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 232, 192, 0, 0, 0, 5, 5, 114, 244, 192, 0, 0, 0, 5, 5, 115, 244, 5, 5, 105, 234, 192, 0, 0, 0, 5, 5, 75, 231, 5, 5, 112, 244, 192, 0, 0, 0, 5, 5, 229, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 233, 192, 0, 0, 0, 5, 5, 117, 244, 5, 5, 175, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 234, 5, 5, 118, 244, 5, 5, 119, 244, 5, 5, 217, 232, 192, 0, 0, 0, 5, 5, 221, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 244, 5, 5, 120, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 244, 5, 5, 0, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 244, 5, 5, 142, 226, 5, 5, 98, 228, 5, 5, 125, 244, 5, 5, 24, 229, 5, 5, 68, 229, 192, 0, 0, 0, 5, 5, 130, 235, 5, 5, 127, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 231, 5, 5, 43, 226, 5, 5, 116, 226, 5, 5, 0, 138, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 244, 5, 5, 131, 244, 5, 5, 35, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 244, 192, 0, 0, 0, 5, 5, 231, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 228, 192, 0, 0, 0, 5, 5, 188, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 244, 5, 5, 43, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 244, 192, 0, 0, 0, 5, 5, 228, 231, 192, 0, 0, 0, 5, 5, 237, 233, 192, 0, 0, 0, 5, 5, 200, 231, 5, 5, 165, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 232, 5, 5, 51, 232, 192, 0, 0, 0, 5, 5, 139, 244, 5, 5, 141, 244, 5, 5, 138, 244, 5, 5, 140, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 244, 5, 5, 58, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 227, 5, 5, 138, 236, 5, 5, 144, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 227, 5, 5, 147, 244, 5, 5, 146, 244, 192, 0, 0, 0, 5, 5, 150, 244, 5, 5, 143, 244, 192, 0, 0, 0, 5, 5, 151, 244, 5, 5, 191, 236, 192, 0, 0, 0, 5, 5, 152, 244, 192, 0, 0, 0, 5, 5, 153, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 239, 5, 5, 155, 244, 192, 0, 0, 0, 5, 5, 154, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 244, 5, 5, 157, 244, 192, 0, 0, 0, 5, 5, 203, 236, 5, 5, 158, 244, 192, 0, 0, 0, 5, 5, 112, 249, 5, 5, 249, 230, 5, 5, 159, 244, 5, 5, 160, 244, 5, 5, 163, 230, 192, 0, 0, 0, 5, 5, 216, 233, 5, 5, 161, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 244, 192, 0, 0, 0, 5, 5, 192, 232, 5, 5, 163, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 227, 192, 0, 0, 0, 5, 5, 198, 237, 5, 5, 225, 232, 5, 5, 33, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 244, 192, 0, 0, 0, 5, 5, 177, 244, 192, 0, 0, 0, 5, 5, 167, 244, 192, 0, 0, 0, 5, 5, 190, 227, 192, 0, 0, 0, 5, 5, 166, 244, 5, 5, 179, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 244, 192, 0, 0, 0, 5, 5, 250, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 244, 5, 5, 170, 244, 192, 0, 0, 0, 5, 5, 109, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 244, 192, 0, 0, 0, 5, 5, 246, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 231, 5, 5, 34, 235, 192, 0, 0, 0, 5, 5, 173, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 244, 192, 0, 0, 0, 5, 5, 172, 244, 5, 5, 174, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 244, 5, 5, 234, 234, 192, 0, 0, 0, 5, 5, 132, 234, 5, 5, 187, 233, 192, 0, 0, 0, 5, 5, 16, 228, 5, 5, 180, 244, 5, 5, 178, 244, 192, 0, 0, 0, 5, 5, 147, 234, 5, 5, 175, 244, 5, 5, 226, 232, 5, 5, 189, 233, 192, 0, 0, 0, 5, 5, 188, 233, 192, 0, 0, 0, 5, 5, 166, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 244, 192, 0, 0, 0, 5, 5, 185, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 244, 192, 0, 0, 0, 5, 5, 189, 244, 192, 0, 0, 0, 5, 5, 186, 244, 5, 5, 187, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 244, 5, 5, 182, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 235, 5, 5, 143, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 244, 192, 0, 0, 0, 5, 5, 193, 244, 192, 0, 0, 0, 5, 5, 198, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 244, 192, 0, 0, 0, 5, 5, 115, 234, 5, 5, 255, 235, 192, 0, 0, 0, 5, 5, 201, 229, 5, 5, 191, 244, 5, 5, 200, 244, 5, 5, 195, 244, 192, 0, 0, 0, 5, 5, 194, 244, 5, 5, 190, 244, 192, 0, 0, 0, 5, 5, 192, 244, 192, 0, 0, 0, 5, 5, 65, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 231, 5, 5, 202, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 234, 192, 0, 0, 0, 5, 5, 206, 244, 5, 5, 111, 235, 192, 0, 0, 0, 5, 5, 223, 232, 192, 0, 0, 0, 5, 5, 201, 244, 5, 5, 203, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 233, 5, 5, 212, 244, 5, 5, 211, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 244, 5, 5, 218, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 244, 5, 5, 60, 238, 5, 5, 216, 244, 5, 5, 210, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 244, 192, 0, 0, 0, 5, 5, 66, 227, 192, 0, 0, 0, 5, 5, 222, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 244, 5, 5, 227, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 244, 5, 5, 212, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 244, 5, 5, 20, 237, 5, 5, 137, 235, 5, 5, 233, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 244, 5, 5, 186, 231, 192, 0, 0, 0, 5, 5, 232, 244, 5, 5, 234, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 244, 192, 0, 0, 0, 5, 5, 238, 244, 192, 0, 0, 0, 5, 5, 235, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 244, 5, 5, 213, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 244, 5, 5, 239, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 235, 192, 0, 0, 0, 5, 5, 241, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 228, 5, 5, 61, 228, 5, 5, 242, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 235, 192, 0, 0, 0, 5, 5, 114, 231, 192, 0, 0, 0, 5, 5, 10, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 244, 192, 0, 0, 0, 5, 5, 201, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 234, 192, 0, 0, 0, 5, 5, 14, 232, 5, 5, 52, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 158, 192, 0, 0, 0, 5, 5, 248, 244, 5, 5, 246, 244, 192, 0, 0, 0, 5, 5, 244, 244, 5, 5, 25, 227, 192, 0, 0, 0, 5, 5, 251, 230, 5, 5, 249, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 244, 192, 0, 0, 0, 5, 5, 245, 244, 5, 5, 253, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 244, 5, 5, 251, 244, 5, 5, 250, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 244, 5, 5, 162, 231, 192, 0, 0, 0, 5, 5, 2, 245, 192, 0, 0, 0, 5, 5, 4, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 233, 192, 0, 0, 0, 5, 5, 5, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 245, 192, 0, 0, 0, 5, 5, 81, 235, 5, 5, 52, 232, 5, 5, 25, 229, 192, 0, 0, 0, 5, 5, 8, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 245, 192, 0, 0, 0, 5, 5, 12, 245, 192, 0, 0, 0, 5, 5, 13, 245, 192, 0, 0, 0, 5, 5, 243, 229, 192, 0, 0, 0, 5, 5, 14, 245, 5, 5, 99, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 227, 192, 0, 0, 0, 5, 5, 128, 227, 192, 0, 0, 0, 5, 5, 16, 245, 192, 0, 0, 0, 5, 5, 77, 236, 5, 5, 26, 229, 5, 5, 15, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 245, 5, 5, 64, 236, 192, 0, 0, 0, 5, 5, 60, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 230, 5, 5, 18, 245, 192, 0, 0, 0, 5, 5, 65, 230, 5, 5, 27, 229, 5, 5, 244, 229, 5, 5, 191, 227, 5, 5, 82, 235, 5, 5, 17, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 232, 5, 5, 188, 235, 5, 5, 167, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 229, 5, 5, 65, 233, 192, 0, 0, 0, 5, 5, 22, 245, 5, 5, 245, 236, 5, 5, 137, 229, 192, 0, 0, 0, 5, 5, 23, 245, 5, 5, 76, 231, 192, 0, 0, 0, 5, 5, 25, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 230, 5, 5, 96, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 230, 5, 5, 176, 228, 5, 5, 16, 232, 5, 5, 20, 245, 5, 5, 26, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 245, 5, 5, 100, 228, 192, 0, 0, 0, 5, 5, 29, 245, 5, 5, 33, 245, 5, 5, 127, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 236, 5, 5, 0, 154, 5, 5, 34, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 227, 192, 0, 0, 0, 5, 5, 31, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 233, 5, 5, 30, 245, 5, 5, 27, 245, 192, 0, 0, 0, 5, 5, 192, 226, 5, 5, 204, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 228, 5, 5, 89, 232, 5, 5, 40, 245, 5, 5, 54, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 245, 5, 5, 43, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 245, 5, 5, 104, 230, 5, 5, 0, 187, 5, 5, 42, 245, 5, 5, 50, 245, 5, 5, 54, 245, 5, 5, 29, 229, 5, 5, 48, 236, 192, 0, 0, 0, 5, 5, 58, 233, 5, 5, 44, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 245, 192, 0, 0, 0, 5, 5, 41, 245, 5, 5, 194, 232, 192, 0, 0, 0, 5, 5, 46, 245, 5, 5, 0, 155, 5, 5, 36, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 228, 5, 5, 202, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 232, 192, 0, 0, 0, 5, 5, 230, 236, 5, 5, 187, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 245, 192, 0, 0, 0, 5, 5, 230, 231, 192, 0, 0, 0, 5, 5, 51, 245, 5, 5, 56, 245, 5, 5, 58, 245, 192, 0, 0, 0, 5, 5, 93, 233, 5, 5, 61, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 235, 5, 5, 67, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 188, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 245, 192, 0, 0, 0, 5, 5, 21, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 234, 5, 5, 62, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 245, 5, 5, 63, 245, 5, 5, 70, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 230, 5, 5, 68, 245, 192, 0, 0, 0, 5, 5, 65, 245, 5, 5, 71, 245, 5, 5, 64, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 245, 5, 5, 79, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 245, 192, 0, 0, 0, 5, 5, 78, 245, 192, 0, 0, 0, 5, 5, 76, 245, 5, 5, 80, 245, 5, 5, 74, 245, 192, 0, 0, 0, 5, 5, 48, 245, 5, 5, 187, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 234, 192, 0, 0, 0, 5, 5, 77, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 231, 5, 5, 102, 228, 192, 0, 0, 0, 5, 5, 119, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 231, 5, 5, 252, 231, 5, 5, 83, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 245, 5, 5, 86, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 245, 5, 5, 84, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 245, 5, 5, 81, 245, 192, 0, 0, 0, 5, 5, 89, 245, 5, 5, 88, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 245, 192, 0, 0, 0, 5, 5, 91, 245, 5, 5, 90, 245, 5, 5, 94, 245, 192, 0, 0, 0, 5, 5, 96, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 229, 5, 5, 92, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 245, 5, 5, 98, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 245, 192, 0, 0, 0, 5, 5, 105, 245, 5, 5, 125, 233, 5, 5, 101, 245, 192, 0, 0, 0, 5, 5, 100, 245, 5, 5, 102, 245, 5, 5, 103, 245, 192, 0, 0, 0, 5, 5, 104, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 245, 5, 5, 107, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 227, 192, 0, 0, 0, 5, 5, 108, 245, 192, 0, 0, 0, 5, 5, 109, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 245, 5, 5, 112, 245, 5, 5, 113, 245, 192, 0, 0, 0, 5, 5, 114, 245, 5, 5, 115, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 245, 5, 5, 116, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 245, 5, 5, 120, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 245, 5, 5, 121, 245, 5, 5, 122, 245, 5, 5, 145, 229, 5, 5, 103, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 232, 192, 0, 0, 0, 5, 5, 145, 234, 192, 0, 0, 0, 5, 5, 188, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 234, 192, 0, 0, 0, 5, 5, 203, 234, 5, 5, 124, 245, 5, 5, 140, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 245, 5, 5, 127, 245, 192, 0, 0, 0, 5, 5, 157, 236, 5, 5, 126, 245, 5, 5, 129, 245, 5, 5, 128, 245, 192, 0, 0, 0, 5, 5, 139, 236, 192, 0, 0, 0, 5, 5, 130, 245, 192, 0, 0, 0, 5, 5, 220, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 245, 5, 5, 132, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 245, 5, 5, 70, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 231, 5, 5, 151, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 245, 5, 5, 136, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 245, 192, 0, 0, 0, 5, 5, 141, 245, 5, 5, 138, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 245, 5, 5, 144, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 245, 192, 0, 0, 0, 5, 5, 154, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 230, 192, 0, 0, 0, 5, 5, 147, 245, 5, 5, 146, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 231, 5, 5, 148, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 245, 192, 0, 0, 0, 5, 5, 93, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 245, 192, 0, 0, 0, 5, 5, 215, 235, 5, 5, 155, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 236, 5, 5, 45, 237, 192, 0, 0, 0, 5, 5, 31, 229, 5, 5, 155, 245, 5, 5, 66, 230, 5, 5, 154, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 245, 5, 5, 23, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 232, 192, 0, 0, 0, 5, 5, 157, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 229, 192, 0, 0, 0, 5, 5, 49, 236, 5, 5, 158, 245, 5, 5, 159, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 245, 192, 0, 0, 0, 5, 5, 195, 232, 192, 0, 0, 0, 5, 5, 163, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 231, 192, 0, 0, 0, 5, 5, 168, 245, 192, 0, 0, 0, 5, 5, 169, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 235, 5, 5, 170, 245, 192, 0, 0, 0, 5, 5, 55, 232, 5, 5, 171, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 237, 5, 5, 175, 245, 192, 0, 0, 0, 5, 5, 174, 245, 5, 5, 173, 245, 5, 5, 20, 233, 192, 0, 0, 0, 5, 5, 176, 245, 5, 5, 48, 231, 192, 0, 0, 0, 5, 5, 177, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 245, 5, 5, 46, 237, 5, 5, 179, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 245, 5, 5, 182, 245, 5, 5, 181, 245, 5, 5, 131, 234, 192, 0, 0, 0, 5, 5, 27, 234, 192, 0, 0, 0, 5, 5, 52, 237, 5, 5, 135, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 230, 192, 0, 0, 0, 5, 5, 230, 234, 192, 0, 0, 0, 5, 5, 185, 245, 5, 5, 183, 245, 192, 0, 0, 0, 5, 5, 70, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 228, 5, 5, 246, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 228, 5, 5, 189, 235, 192, 0, 0, 0, 5, 5, 188, 245, 5, 5, 186, 245, 192, 0, 0, 0, 5, 5, 32, 229, 192, 0, 0, 0, 5, 5, 33, 229, 5, 5, 0, 199, 192, 0, 0, 0, 5, 5, 152, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 189, 5, 5, 193, 245, 192, 0, 0, 0, 5, 5, 196, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 234, 192, 0, 0, 0, 5, 5, 129, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 245, 5, 5, 194, 245, 5, 5, 189, 245, 192, 0, 0, 0, 5, 5, 192, 245, 5, 5, 160, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 219, 5, 5, 190, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 245, 192, 0, 0, 0, 5, 5, 198, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 245, 5, 5, 61, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 227, 5, 5, 176, 231, 5, 5, 60, 237, 5, 5, 8, 236, 5, 5, 196, 245, 5, 5, 188, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 227, 5, 5, 199, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 245, 192, 0, 0, 0, 5, 5, 170, 232, 192, 0, 0, 0, 5, 5, 62, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 231, 192, 0, 0, 0, 5, 5, 35, 235, 5, 5, 208, 245, 192, 0, 0, 0, 5, 5, 207, 245, 5, 5, 34, 229, 5, 5, 73, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 245, 5, 5, 213, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 245, 192, 0, 0, 0, 5, 5, 74, 229, 5, 5, 210, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 233, 5, 5, 62, 235, 5, 5, 233, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 232, 5, 5, 218, 245, 192, 0, 0, 0, 5, 5, 219, 245, 5, 5, 215, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 245, 192, 0, 0, 0, 5, 5, 217, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 245, 192, 0, 0, 0, 5, 5, 221, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 235, 192, 0, 0, 0, 5, 5, 230, 235, 5, 5, 228, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 245, 5, 5, 222, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 235, 5, 5, 226, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 231, 192, 0, 0, 0, 5, 5, 228, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 245, 192, 0, 0, 0, 5, 5, 234, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 245, 5, 5, 229, 245, 5, 5, 63, 234, 5, 5, 232, 245, 192, 0, 0, 0, 5, 5, 233, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 226, 192, 0, 0, 0, 5, 5, 240, 245, 5, 5, 235, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 245, 192, 0, 0, 0, 5, 5, 71, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 245, 5, 5, 238, 245, 5, 5, 241, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 245, 5, 5, 243, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 231, 192, 0, 0, 0, 5, 5, 163, 226, 192, 0, 0, 0, 5, 5, 244, 245, 5, 5, 237, 236, 192, 0, 0, 0, 5, 5, 25, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 229, 5, 5, 219, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 245, 5, 5, 246, 245, 5, 5, 13, 226, 192, 0, 0, 0, 5, 5, 247, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 245, 5, 5, 249, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 245, 192, 0, 0, 0, 5, 5, 251, 245, 5, 5, 241, 227, 5, 5, 7, 241, 5, 5, 252, 245, 192, 0, 0, 0, 5, 5, 205, 231, 5, 5, 253, 245, 5, 5, 57, 230, 192, 0, 0, 0, 5, 5, 254, 245, 192, 0, 0, 0, 5, 5, 89, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 245, 5, 5, 122, 235, 5, 5, 84, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 231, 5, 5, 164, 230, 192, 0, 0, 0, 5, 5, 47, 235, 5, 5, 122, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 246, 5, 5, 36, 229, 5, 5, 3, 246, 5, 5, 170, 234, 192, 0, 0, 0, 5, 5, 17, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 246, 192, 0, 0, 0, 5, 5, 111, 232, 5, 5, 117, 234, 5, 5, 177, 228, 5, 5, 4, 246, 5, 5, 235, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 246, 5, 5, 7, 246, 5, 5, 10, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 246, 192, 0, 0, 0, 5, 5, 11, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 246, 192, 0, 0, 0, 5, 5, 12, 246, 192, 0, 0, 0, 5, 5, 71, 227, 192, 0, 0, 0, 5, 5, 14, 246, 192, 0, 0, 0, 5, 5, 15, 246, 5, 5, 16, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 229, 5, 5, 223, 236, 192, 0, 0, 0, 5, 5, 18, 246, 5, 5, 49, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 226, 5, 5, 19, 246, 5, 5, 20, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 208, 192, 0, 0, 0, 5, 5, 22, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 230, 192, 0, 0, 0, 5, 5, 25, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 226, 5, 5, 0, 144, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 246, 5, 5, 27, 246, 5, 5, 78, 234, 192, 0, 0, 0, 5, 5, 78, 231, 192, 0, 0, 0, 5, 5, 144, 226, 192, 0, 0, 0, 5, 5, 161, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 228, 5, 5, 18, 228, 192, 0, 0, 0, 5, 5, 26, 246, 192, 0, 0, 0, 5, 5, 164, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 226, 5, 5, 31, 246, 5, 5, 2, 237, 5, 5, 131, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 235, 192, 0, 0, 0, 5, 5, 43, 246, 192, 0, 0, 0, 5, 5, 145, 226, 5, 5, 41, 246, 192, 0, 0, 0, 5, 5, 39, 246, 5, 5, 30, 246, 192, 0, 0, 0, 5, 5, 28, 246, 192, 0, 0, 0, 5, 5, 29, 246, 192, 0, 0, 0, 5, 5, 81, 230, 5, 5, 34, 228, 5, 5, 251, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 226, 192, 0, 0, 0, 5, 5, 33, 246, 5, 5, 32, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 246, 5, 5, 34, 246, 5, 5, 37, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 236, 5, 5, 36, 246, 5, 5, 146, 226, 5, 5, 23, 227, 5, 5, 40, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 246, 5, 5, 55, 246, 5, 5, 56, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 137, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 246, 192, 0, 0, 0, 5, 5, 48, 246, 5, 5, 47, 246, 192, 0, 0, 0, 5, 5, 45, 246, 5, 5, 44, 246, 5, 5, 231, 232, 192, 0, 0, 0, 5, 5, 161, 232, 5, 5, 50, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 232, 5, 5, 105, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 226, 5, 5, 51, 246, 192, 0, 0, 0, 5, 5, 37, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 246, 192, 0, 0, 0, 5, 5, 70, 246, 192, 0, 0, 0, 5, 5, 147, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 226, 5, 5, 67, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 246, 192, 0, 0, 0, 5, 5, 65, 246, 192, 0, 0, 0, 5, 5, 72, 246, 5, 5, 66, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 227, 5, 5, 60, 246, 5, 5, 71, 246, 192, 0, 0, 0, 5, 5, 61, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 246, 192, 0, 0, 0, 5, 5, 59, 246, 5, 5, 124, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 227, 192, 0, 0, 0, 5, 5, 19, 228, 192, 0, 0, 0, 5, 5, 77, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 230, 192, 0, 0, 0, 5, 5, 80, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 233, 5, 5, 85, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 235, 192, 0, 0, 0, 5, 5, 76, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 226, 5, 5, 200, 228, 5, 5, 229, 234, 5, 5, 86, 246, 192, 0, 0, 0, 5, 5, 74, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 246, 192, 0, 0, 0, 5, 5, 78, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 246, 5, 5, 218, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 246, 5, 5, 162, 235, 5, 5, 87, 246, 5, 5, 0, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 246, 192, 0, 0, 0, 5, 5, 88, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 234, 5, 5, 95, 246, 192, 0, 0, 0, 5, 5, 106, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 246, 5, 5, 166, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 235, 5, 5, 111, 246, 5, 5, 193, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 143, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 246, 192, 0, 0, 0, 5, 5, 100, 246, 5, 5, 58, 232, 5, 5, 94, 246, 5, 5, 102, 246, 5, 5, 107, 246, 192, 0, 0, 0, 5, 5, 44, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 136, 192, 0, 0, 0, 5, 5, 99, 246, 192, 0, 0, 0, 5, 5, 108, 246, 5, 5, 54, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 246, 192, 0, 0, 0, 5, 5, 98, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 246, 192, 0, 0, 0, 5, 5, 123, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 231, 5, 5, 112, 246, 192, 0, 0, 0, 5, 5, 117, 246, 5, 5, 59, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 246, 192, 0, 0, 0, 5, 5, 120, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 232, 192, 0, 0, 0, 5, 5, 121, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 236, 5, 5, 110, 246, 5, 5, 208, 226, 192, 0, 0, 0, 5, 5, 116, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 246, 5, 5, 6, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 246, 5, 5, 118, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 235, 192, 0, 0, 0, 5, 5, 23, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 246, 5, 5, 49, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 235, 192, 0, 0, 0, 5, 5, 250, 235, 5, 5, 132, 246, 5, 5, 131, 246, 192, 0, 0, 0, 5, 5, 127, 246, 5, 5, 128, 246, 192, 0, 0, 0, 5, 5, 18, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 246, 192, 0, 0, 0, 5, 5, 124, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 246, 5, 5, 0, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 246, 5, 5, 138, 246, 192, 0, 0, 0, 5, 5, 183, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 246, 5, 5, 2, 231, 5, 5, 55, 230, 5, 5, 140, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 246, 192, 0, 0, 0, 5, 5, 55, 235, 5, 5, 136, 246, 192, 0, 0, 0, 5, 5, 97, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 237, 5, 5, 194, 233, 5, 5, 49, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 246, 192, 0, 0, 0, 5, 5, 142, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 246, 5, 5, 144, 246, 192, 0, 0, 0, 5, 5, 146, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 246, 5, 5, 145, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 226, 192, 0, 0, 0, 5, 5, 7, 234, 192, 0, 0, 0, 5, 5, 150, 246, 5, 5, 153, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 246, 192, 0, 0, 0, 5, 5, 236, 231, 192, 0, 0, 0, 5, 5, 147, 246, 5, 5, 181, 229, 5, 5, 79, 231, 5, 5, 68, 228, 5, 5, 78, 236, 192, 0, 0, 0, 5, 5, 83, 236, 5, 5, 190, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 246, 5, 5, 158, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 236, 192, 0, 0, 0, 5, 5, 159, 246, 5, 5, 161, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 246, 5, 5, 163, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 233, 5, 5, 164, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 234, 5, 5, 151, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 230, 192, 0, 0, 0, 5, 5, 166, 246, 5, 5, 171, 246, 5, 5, 60, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 246, 5, 5, 17, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 246, 5, 5, 169, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 242, 5, 5, 27, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 236, 192, 0, 0, 0, 5, 5, 15, 244, 5, 5, 175, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 246, 5, 5, 201, 228, 192, 0, 0, 0, 5, 5, 170, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 246, 5, 5, 228, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 227, 192, 0, 0, 0, 5, 5, 208, 236, 192, 0, 0, 0, 5, 5, 41, 228, 5, 5, 180, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 234, 192, 0, 0, 0, 5, 5, 0, 152, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 226, 5, 5, 187, 246, 5, 5, 188, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 246, 192, 0, 0, 0, 5, 5, 203, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 246, 5, 5, 66, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 246, 5, 5, 186, 246, 5, 5, 196, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 246, 5, 5, 193, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 246, 192, 0, 0, 0, 5, 5, 192, 246, 5, 5, 70, 230, 192, 0, 0, 0, 5, 5, 194, 246, 192, 0, 0, 0, 5, 5, 197, 232, 192, 0, 0, 0, 5, 5, 106, 228, 5, 5, 218, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 226, 192, 0, 0, 0, 5, 5, 202, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 246, 5, 5, 201, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 246, 5, 5, 6, 235, 5, 5, 184, 234, 5, 5, 203, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 232, 5, 5, 213, 246, 192, 0, 0, 0, 5, 5, 209, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 226, 192, 0, 0, 0, 5, 5, 207, 246, 192, 0, 0, 0, 5, 5, 164, 235, 5, 5, 208, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 246, 192, 0, 0, 0, 5, 5, 206, 246, 5, 5, 211, 246, 5, 5, 214, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 246, 5, 5, 204, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 232, 192, 0, 0, 0, 5, 5, 221, 246, 192, 0, 0, 0, 5, 5, 4, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 231, 192, 0, 0, 0, 5, 5, 47, 237, 5, 5, 225, 246, 192, 0, 0, 0, 5, 5, 226, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 246, 192, 0, 0, 0, 5, 5, 52, 231, 192, 0, 0, 0, 5, 5, 228, 246, 192, 0, 0, 0, 5, 5, 231, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 246, 5, 5, 222, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 226, 192, 0, 0, 0, 5, 5, 229, 246, 192, 0, 0, 0, 5, 5, 234, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 246, 192, 0, 0, 0, 5, 5, 23, 233, 192, 0, 0, 0, 5, 5, 224, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 236, 5, 5, 250, 246, 192, 0, 0, 0, 5, 5, 241, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 246, 192, 0, 0, 0, 5, 5, 249, 246, 5, 5, 248, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 246, 192, 0, 0, 0, 5, 5, 6, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 247, 5, 5, 4, 247, 192, 0, 0, 0, 5, 5, 194, 226, 192, 0, 0, 0, 5, 5, 152, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 246, 5, 5, 5, 247, 5, 5, 254, 246, 192, 0, 0, 0, 5, 5, 7, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 247, 5, 5, 8, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 247, 5, 5, 10, 247, 5, 5, 195, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 247, 192, 0, 0, 0, 5, 5, 16, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 228, 192, 0, 0, 0, 5, 5, 18, 247, 192, 0, 0, 0, 5, 5, 17, 247, 192, 0, 0, 0, 5, 5, 124, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 229, 5, 5, 126, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 247, 5, 5, 157, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 226, 192, 0, 0, 0, 5, 5, 20, 247, 192, 0, 0, 0, 5, 5, 46, 226, 192, 0, 0, 0, 5, 5, 3, 231, 5, 5, 21, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 229, 5, 5, 22, 247, 5, 5, 0, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 231, 192, 0, 0, 0, 5, 5, 30, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 247, 192, 0, 0, 0, 5, 5, 243, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 247, 5, 5, 25, 247, 5, 5, 20, 228, 192, 0, 0, 0, 5, 5, 24, 247, 5, 5, 31, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 232, 192, 0, 0, 0, 5, 5, 37, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 232, 5, 5, 32, 247, 192, 0, 0, 0, 5, 5, 35, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 247, 192, 0, 0, 0, 5, 5, 38, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 247, 192, 0, 0, 0, 5, 5, 39, 247, 5, 5, 41, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 228, 5, 5, 29, 247, 192, 0, 0, 0, 5, 5, 0, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 247, 192, 0, 0, 0, 5, 5, 139, 229, 5, 5, 14, 237, 5, 5, 42, 247, 5, 5, 43, 247, 5, 5, 61, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 247, 5, 5, 109, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 247, 5, 5, 46, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 235, 5, 5, 47, 247, 192, 0, 0, 0, 5, 5, 112, 229, 192, 0, 0, 0, 5, 5, 185, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 247, 5, 5, 4, 231, 5, 5, 51, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 236, 5, 5, 48, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 247, 5, 5, 165, 231, 5, 5, 136, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 247, 192, 0, 0, 0, 5, 5, 54, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 247, 192, 0, 0, 0, 5, 5, 55, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 227, 192, 0, 0, 0, 5, 5, 165, 235, 5, 5, 57, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 247, 5, 5, 59, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 247, 5, 5, 62, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 247, 192, 0, 0, 0, 5, 5, 67, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 247, 192, 0, 0, 0, 5, 5, 58, 247, 5, 5, 64, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 247, 5, 5, 56, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 247, 5, 5, 21, 228, 5, 5, 70, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 247, 192, 0, 0, 0, 5, 5, 72, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 247, 192, 0, 0, 0, 5, 5, 76, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 230, 192, 0, 0, 0, 5, 5, 77, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 247, 5, 5, 166, 231, 192, 0, 0, 0, 5, 5, 143, 236, 192, 0, 0, 0, 5, 5, 80, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 235, 5, 5, 70, 234, 5, 5, 81, 247, 192, 0, 0, 0, 5, 5, 82, 247, 5, 5, 157, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 229, 5, 5, 65, 234, 5, 5, 84, 247, 192, 0, 0, 0, 5, 5, 232, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 247, 5, 5, 175, 236, 192, 0, 0, 0, 5, 5, 86, 247, 5, 5, 80, 231, 192, 0, 0, 0, 5, 5, 88, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 247, 5, 5, 73, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 247, 192, 0, 0, 0, 5, 5, 93, 247, 5, 5, 94, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 247, 192, 0, 0, 0, 5, 5, 96, 247, 5, 5, 97, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 231, 5, 5, 98, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 228, 192, 0, 0, 0, 5, 5, 95, 233, 5, 5, 101, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 228, 192, 0, 0, 0, 5, 5, 96, 231, 192, 0, 0, 0, 5, 5, 104, 247, 192, 0, 0, 0, 5, 5, 196, 233, 192, 0, 0, 0, 5, 5, 103, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 247, 5, 5, 157, 232, 5, 5, 131, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 247, 192, 0, 0, 0, 5, 5, 106, 247, 192, 0, 0, 0, 5, 5, 5, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 228, 192, 0, 0, 0, 5, 5, 107, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 227, 192, 0, 0, 0, 5, 5, 79, 236, 5, 5, 18, 232, 192, 0, 0, 0, 5, 5, 108, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 231, 5, 5, 244, 232, 5, 5, 6, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 247, 192, 0, 0, 0, 5, 5, 113, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 229, 5, 5, 104, 232, 5, 5, 111, 247, 192, 0, 0, 0, 5, 5, 7, 231, 5, 5, 250, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 229, 192, 0, 0, 0, 5, 5, 47, 226, 192, 0, 0, 0, 5, 5, 117, 247, 5, 5, 108, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 229, 192, 0, 0, 0, 5, 5, 67, 237, 5, 5, 116, 247, 5, 5, 115, 247, 5, 5, 237, 231, 192, 0, 0, 0, 5, 5, 163, 227, 5, 5, 57, 237, 5, 5, 210, 226, 5, 5, 8, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 247, 192, 0, 0, 0, 5, 5, 120, 247, 5, 5, 118, 247, 192, 0, 0, 0, 5, 5, 202, 228, 192, 0, 0, 0, 5, 5, 120, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 229, 5, 5, 39, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 247, 192, 0, 0, 0, 5, 5, 168, 231, 192, 0, 0, 0, 5, 5, 198, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 236, 192, 0, 0, 0, 5, 5, 126, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 228, 192, 0, 0, 0, 5, 5, 167, 231, 5, 5, 122, 247, 192, 0, 0, 0, 5, 5, 127, 247, 5, 5, 223, 228, 5, 5, 124, 247, 5, 5, 125, 247, 192, 0, 0, 0, 5, 5, 121, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 231, 5, 5, 233, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 232, 192, 0, 0, 0, 5, 5, 151, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 247, 192, 0, 0, 0, 5, 5, 128, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 231, 5, 5, 74, 227, 5, 5, 129, 247, 192, 0, 0, 0, 5, 5, 101, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 247, 5, 5, 142, 247, 5, 5, 25, 233, 192, 0, 0, 0, 5, 5, 141, 247, 192, 0, 0, 0, 5, 5, 138, 247, 5, 5, 146, 247, 5, 5, 139, 247, 192, 0, 0, 0, 5, 5, 135, 247, 192, 0, 0, 0, 5, 5, 96, 233, 5, 5, 134, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 247, 192, 0, 0, 0, 5, 5, 89, 236, 5, 5, 254, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 247, 192, 0, 0, 0, 5, 5, 133, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 247, 5, 5, 192, 230, 192, 0, 0, 0, 5, 5, 179, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 232, 192, 0, 0, 0, 5, 5, 191, 235, 5, 5, 54, 226, 5, 5, 0, 192, 192, 0, 0, 0, 5, 5, 197, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 247, 192, 0, 0, 0, 5, 5, 8, 234, 192, 0, 0, 0, 5, 5, 148, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 247, 192, 0, 0, 0, 5, 5, 147, 247, 5, 5, 149, 247, 192, 0, 0, 0, 5, 5, 158, 228, 5, 5, 145, 247, 5, 5, 40, 229, 192, 0, 0, 0, 5, 5, 67, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 247, 5, 5, 144, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 247, 192, 0, 0, 0, 5, 5, 156, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 247, 5, 5, 241, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 247, 192, 0, 0, 0, 5, 5, 160, 247, 5, 5, 159, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 247, 192, 0, 0, 0, 5, 5, 31, 230, 192, 0, 0, 0, 5, 5, 164, 247, 5, 5, 163, 247, 5, 5, 38, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 247, 5, 5, 167, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 247, 5, 5, 154, 227, 5, 5, 142, 245, 5, 5, 37, 231, 192, 0, 0, 0, 5, 5, 169, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 241, 192, 0, 0, 0, 5, 5, 172, 247, 192, 0, 0, 0, 5, 5, 173, 247, 192, 0, 0, 0, 5, 5, 126, 230, 192, 0, 0, 0, 5, 5, 174, 247, 5, 5, 175, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 247, 5, 5, 178, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 247, 192, 0, 0, 0, 5, 5, 180, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 233, 192, 0, 0, 0, 5, 5, 182, 247, 192, 0, 0, 0, 5, 5, 167, 235, 192, 0, 0, 0, 5, 5, 183, 247, 192, 0, 0, 0, 5, 5, 184, 247, 192, 0, 0, 0, 5, 5, 185, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 231, 5, 5, 187, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 229, 5, 5, 87, 237, 5, 5, 188, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 247, 5, 5, 251, 234, 5, 5, 190, 247, 192, 0, 0, 0, 5, 5, 198, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 247, 5, 5, 194, 247, 192, 0, 0, 0, 5, 5, 192, 235, 5, 5, 195, 247, 5, 5, 196, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 226, 5, 5, 75, 233, 192, 0, 0, 0, 5, 5, 39, 235, 5, 5, 146, 229, 5, 5, 41, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 235, 5, 5, 153, 226, 5, 5, 172, 234, 5, 5, 202, 247, 5, 5, 75, 227, 5, 5, 189, 231, 5, 5, 201, 247, 5, 5, 206, 247, 5, 5, 253, 232, 5, 5, 61, 236, 192, 0, 0, 0, 5, 5, 204, 247, 5, 5, 205, 247, 5, 5, 132, 227, 192, 0, 0, 0, 5, 5, 207, 247, 5, 5, 99, 234, 5, 5, 133, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 234, 5, 5, 127, 233, 5, 5, 203, 247, 192, 0, 0, 0, 5, 5, 193, 235, 5, 5, 166, 226, 5, 5, 209, 247, 5, 5, 30, 237, 5, 5, 38, 233, 5, 5, 59, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 229, 5, 5, 208, 247, 192, 0, 0, 0, 5, 5, 87, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 247, 5, 5, 238, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 234, 192, 0, 0, 0, 5, 5, 14, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 247, 5, 5, 205, 229, 5, 5, 2, 230, 192, 0, 0, 0, 5, 5, 10, 231, 192, 0, 0, 0, 5, 5, 101, 234, 192, 0, 0, 0, 5, 5, 159, 228, 5, 5, 211, 247, 5, 5, 210, 247, 192, 0, 0, 0, 5, 5, 40, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 247, 5, 5, 215, 247, 5, 5, 42, 229, 5, 5, 213, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 247, 5, 5, 217, 247, 192, 0, 0, 0, 5, 5, 219, 247, 5, 5, 73, 232, 192, 0, 0, 0, 5, 5, 218, 247, 5, 5, 94, 227, 192, 0, 0, 0, 5, 5, 221, 247, 192, 0, 0, 0, 5, 5, 220, 247, 5, 5, 222, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 247, 5, 5, 226, 247, 192, 0, 0, 0, 5, 5, 227, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 231, 192, 0, 0, 0, 5, 5, 61, 230, 5, 5, 228, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 226, 192, 0, 0, 0, 5, 5, 229, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 232, 5, 5, 230, 247, 192, 0, 0, 0, 5, 5, 231, 247, 5, 5, 41, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 247, 5, 5, 240, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 247, 5, 5, 241, 247, 192, 0, 0, 0, 5, 5, 205, 227, 192, 0, 0, 0, 5, 5, 245, 247, 192, 0, 0, 0, 5, 5, 191, 231, 192, 0, 0, 0, 5, 5, 246, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 228, 192, 0, 0, 0, 5, 5, 243, 247, 5, 5, 244, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 233, 192, 0, 0, 0, 5, 5, 239, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 247, 5, 5, 249, 247, 5, 5, 145, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 233, 5, 5, 253, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 247, 5, 5, 252, 247, 5, 5, 254, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 248, 5, 5, 2, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 247, 192, 0, 0, 0, 5, 5, 97, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 248, 5, 5, 10, 248, 5, 5, 7, 248, 5, 5, 5, 248, 192, 0, 0, 0, 5, 5, 8, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 231, 5, 5, 13, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 248, 5, 5, 12, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 248, 192, 0, 0, 0, 5, 5, 127, 230, 192, 0, 0, 0, 5, 5, 17, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 248, 5, 5, 22, 248, 192, 0, 0, 0, 5, 5, 21, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 248, 5, 5, 24, 248, 192, 0, 0, 0, 5, 5, 80, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 248, 192, 0, 0, 0, 5, 5, 26, 248, 5, 5, 28, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 248, 5, 5, 82, 231, 5, 5, 32, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 228, 5, 5, 33, 248, 5, 5, 35, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 248, 5, 5, 34, 248, 192, 0, 0, 0, 5, 5, 38, 248, 192, 0, 0, 0, 5, 5, 68, 230, 5, 5, 39, 248, 5, 5, 134, 227, 5, 5, 71, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 233, 5, 5, 41, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 248, 5, 5, 42, 248, 5, 5, 110, 228, 5, 5, 45, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 226, 192, 0, 0, 0, 5, 5, 47, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 229, 5, 5, 46, 248, 192, 0, 0, 0, 5, 5, 55, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 248, 5, 5, 51, 248, 5, 5, 129, 235, 5, 5, 48, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 248, 192, 0, 0, 0, 5, 5, 54, 248, 5, 5, 52, 248, 5, 5, 135, 227, 192, 0, 0, 0, 5, 5, 53, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 234, 5, 5, 238, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 236, 5, 5, 59, 248, 192, 0, 0, 0, 5, 5, 58, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 248, 5, 5, 121, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 248, 192, 0, 0, 0, 5, 5, 7, 227, 5, 5, 60, 248, 5, 5, 65, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 248, 5, 5, 116, 233, 5, 5, 66, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 229, 192, 0, 0, 0, 5, 5, 57, 228, 5, 5, 69, 248, 5, 5, 70, 248, 5, 5, 71, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 231, 5, 5, 72, 248, 192, 0, 0, 0, 5, 5, 27, 230, 5, 5, 73, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 238, 5, 5, 7, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 248, 5, 5, 95, 245, 5, 5, 76, 248, 5, 5, 168, 232, 5, 5, 53, 231, 5, 5, 64, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 235, 5, 5, 56, 233, 5, 5, 80, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 226, 192, 0, 0, 0, 5, 5, 242, 235, 5, 5, 97, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 248, 5, 5, 154, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 234, 5, 5, 81, 248, 5, 5, 119, 234, 192, 0, 0, 0, 5, 5, 117, 233, 192, 0, 0, 0, 5, 5, 82, 248, 5, 5, 158, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 236, 5, 5, 99, 248, 5, 5, 86, 248, 5, 5, 87, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 232, 5, 5, 63, 232, 192, 0, 0, 0, 5, 5, 200, 233, 192, 0, 0, 0, 5, 5, 85, 248, 5, 5, 171, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 248, 192, 0, 0, 0, 5, 5, 91, 248, 5, 5, 104, 248, 5, 5, 201, 233, 5, 5, 228, 232, 5, 5, 88, 248, 192, 0, 0, 0, 5, 5, 98, 233, 5, 5, 149, 233, 5, 5, 89, 248, 5, 5, 93, 248, 5, 5, 106, 231, 192, 0, 0, 0, 5, 5, 103, 234, 5, 5, 48, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 231, 5, 5, 92, 248, 5, 5, 84, 232, 5, 5, 74, 232, 5, 5, 90, 248, 5, 5, 0, 135, 5, 5, 24, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 230, 5, 5, 84, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 248, 5, 5, 96, 248, 192, 0, 0, 0, 5, 5, 0, 205, 5, 5, 98, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 234, 192, 0, 0, 0, 5, 5, 106, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 233, 5, 5, 117, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 232, 192, 0, 0, 0, 5, 5, 46, 228, 192, 0, 0, 0, 5, 5, 105, 248, 5, 5, 111, 236, 5, 5, 27, 226, 192, 0, 0, 0, 5, 5, 115, 235, 5, 5, 155, 226, 5, 5, 100, 248, 5, 5, 101, 248, 5, 5, 102, 248, 5, 5, 103, 248, 5, 5, 219, 233, 5, 5, 167, 232, 5, 5, 0, 193, 5, 5, 107, 248, 192, 0, 0, 0, 5, 5, 108, 248, 5, 5, 158, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 232, 192, 0, 0, 0, 5, 5, 109, 248, 192, 0, 0, 0, 5, 5, 79, 226, 5, 5, 20, 232, 192, 0, 0, 0, 5, 5, 161, 228, 192, 0, 0, 0, 5, 5, 146, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 248, 5, 5, 110, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 232, 5, 5, 69, 230, 5, 5, 111, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 230, 5, 5, 112, 248, 5, 5, 241, 231, 5, 5, 240, 231, 192, 0, 0, 0, 5, 5, 0, 194, 192, 0, 0, 0, 5, 5, 225, 236, 5, 5, 116, 248, 192, 0, 0, 0, 5, 5, 208, 234, 5, 5, 118, 248, 5, 5, 117, 248, 5, 5, 115, 248, 5, 5, 149, 244, 5, 5, 76, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 248, 192, 0, 0, 0, 5, 5, 120, 248, 5, 5, 119, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 248, 192, 0, 0, 0, 5, 5, 112, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 235, 192, 0, 0, 0, 5, 5, 122, 248, 192, 0, 0, 0, 5, 5, 71, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 248, 192, 0, 0, 0, 5, 5, 124, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 228, 5, 5, 126, 248, 192, 0, 0, 0, 5, 5, 127, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 236, 192, 0, 0, 0, 5, 5, 243, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 248, 192, 0, 0, 0, 5, 5, 133, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 233, 5, 5, 135, 248, 5, 5, 130, 230, 5, 5, 78, 230, 5, 5, 91, 234, 5, 5, 245, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 230, 192, 0, 0, 0, 5, 5, 118, 231, 192, 0, 0, 0, 5, 5, 136, 248, 192, 0, 0, 0, 5, 5, 137, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 231, 5, 5, 138, 248, 192, 0, 0, 0, 5, 5, 139, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 248, 5, 5, 167, 236, 192, 0, 0, 0, 5, 5, 131, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 248, 5, 5, 141, 248, 192, 0, 0, 0, 5, 5, 44, 229, 192, 0, 0, 0, 5, 5, 69, 229, 5, 5, 206, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 230, 192, 0, 0, 0, 5, 5, 144, 248, 192, 0, 0, 0, 5, 5, 143, 248, 192, 0, 0, 0, 5, 5, 143, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 228, 192, 0, 0, 0, 5, 5, 171, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 248, 192, 0, 0, 0, 5, 5, 11, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 248, 5, 5, 147, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 248, 5, 5, 150, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 231, 192, 0, 0, 0, 5, 5, 152, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 248, 5, 5, 154, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 234, 5, 5, 131, 229, 5, 5, 79, 230, 5, 5, 155, 248, 192, 0, 0, 0, 5, 5, 156, 248, 5, 5, 186, 236, 5, 5, 145, 230, 5, 5, 72, 236, 5, 5, 226, 236, 5, 5, 157, 248, 5, 5, 24, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 248, 192, 0, 0, 0, 5, 5, 101, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 248, 5, 5, 18, 227, 5, 5, 85, 231, 192, 0, 0, 0, 5, 5, 159, 248, 192, 0, 0, 0, 5, 5, 160, 248, 192, 0, 0, 0, 5, 5, 68, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 235, 5, 5, 50, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 248, 5, 5, 164, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 233, 5, 5, 219, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 248, 5, 5, 169, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 233, 5, 5, 173, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 248, 5, 5, 174, 248, 192, 0, 0, 0, 5, 5, 179, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 248, 5, 5, 80, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 234, 192, 0, 0, 0, 5, 5, 175, 248, 192, 0, 0, 0, 5, 5, 12, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 235, 192, 0, 0, 0, 5, 5, 26, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 230, 192, 0, 0, 0, 5, 5, 220, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 231, 192, 0, 0, 0, 5, 5, 183, 248, 192, 0, 0, 0, 5, 5, 177, 248, 5, 5, 182, 248, 192, 0, 0, 0, 5, 5, 29, 236, 192, 0, 0, 0, 5, 5, 28, 233, 5, 5, 184, 248, 5, 5, 181, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 248, 192, 0, 0, 0, 5, 5, 187, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 248, 5, 5, 123, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 235, 5, 5, 246, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 227, 192, 0, 0, 0, 5, 5, 191, 248, 192, 0, 0, 0, 5, 5, 47, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 248, 5, 5, 119, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 231, 5, 5, 194, 248, 5, 5, 196, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 231, 192, 0, 0, 0, 5, 5, 195, 248, 5, 5, 197, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 228, 192, 0, 0, 0, 5, 5, 3, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 230, 5, 5, 25, 237, 192, 0, 0, 0, 5, 5, 193, 248, 5, 5, 168, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 248, 5, 5, 200, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 234, 192, 0, 0, 0, 5, 5, 151, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 233, 192, 0, 0, 0, 5, 5, 205, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 232, 5, 5, 201, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 228, 192, 0, 0, 0, 5, 5, 204, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 248, 192, 0, 0, 0, 5, 5, 13, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 248, 192, 0, 0, 0, 5, 5, 111, 229, 5, 5, 65, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 248, 5, 5, 208, 248, 5, 5, 39, 233, 192, 0, 0, 0, 5, 5, 206, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 248, 5, 5, 111, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 248, 5, 5, 212, 248, 5, 5, 215, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 248, 5, 5, 214, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 248, 5, 5, 222, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 231, 5, 5, 202, 233, 5, 5, 221, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 248, 5, 5, 227, 248, 192, 0, 0, 0, 5, 5, 158, 232, 192, 0, 0, 0, 5, 5, 231, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 227, 5, 5, 233, 248, 5, 5, 84, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 248, 5, 5, 235, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 248, 192, 0, 0, 0, 5, 5, 236, 248, 192, 0, 0, 0, 5, 5, 237, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 248, 192, 0, 0, 0, 5, 5, 243, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 248, 5, 5, 244, 248, 5, 5, 247, 248, 5, 5, 249, 248, 192, 0, 0, 0, 5, 5, 248, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 236, 192, 0, 0, 0, 5, 5, 250, 248, 5, 5, 244, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 248, 192, 0, 0, 0, 5, 5, 96, 235, 5, 5, 252, 248, 5, 5, 195, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 226, 192, 0, 0, 0, 5, 5, 79, 227, 192, 0, 0, 0, 5, 5, 78, 227, 5, 5, 253, 248, 192, 0, 0, 0, 5, 5, 254, 248, 192, 0, 0, 0, 5, 5, 255, 248, 5, 5, 2, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 249, 192, 0, 0, 0, 5, 5, 80, 227, 5, 5, 237, 226, 5, 5, 48, 229, 5, 5, 148, 234, 192, 0, 0, 0, 5, 5, 5, 249, 5, 5, 4, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 249, 192, 0, 0, 0, 5, 5, 8, 249, 5, 5, 7, 249, 192, 0, 0, 0, 5, 5, 10, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 165, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 249, 192, 0, 0, 0, 5, 5, 15, 249, 5, 5, 14, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 249, 5, 5, 16, 249, 5, 5, 18, 249, 192, 0, 0, 0, 5, 5, 206, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 249, 5, 5, 22, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 249, 192, 0, 0, 0, 5, 5, 149, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 249, 5, 5, 26, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 130, 5, 5, 113, 232, 192, 0, 0, 0, 5, 5, 27, 249, 192, 0, 0, 0, 5, 5, 43, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 249, 5, 5, 28, 249, 5, 5, 49, 229, 192, 0, 0, 0, 5, 5, 29, 249, 5, 5, 180, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 235, 5, 5, 32, 249, 5, 5, 34, 249, 5, 5, 33, 249, 5, 5, 35, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 221, 5, 5, 98, 231, 5, 5, 200, 230, 5, 5, 81, 227, 5, 5, 36, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 234, 192, 0, 0, 0, 5, 5, 38, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 222, 192, 0, 0, 0, 5, 5, 37, 249, 5, 5, 40, 233, 192, 0, 0, 0, 5, 5, 227, 236, 5, 5, 203, 233, 5, 5, 31, 249, 5, 5, 188, 236, 192, 0, 0, 0, 5, 5, 163, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 228, 5, 5, 202, 236, 192, 0, 0, 0, 5, 5, 60, 228, 192, 0, 0, 0, 5, 5, 136, 232, 5, 5, 204, 245, 192, 0, 0, 0, 5, 5, 39, 249, 5, 5, 196, 226, 5, 5, 121, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 226, 5, 5, 41, 249, 192, 0, 0, 0, 5, 5, 42, 249, 5, 5, 40, 249, 5, 5, 120, 228, 192, 0, 0, 0, 5, 5, 141, 229, 5, 5, 15, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 249, 5, 5, 113, 248, 192, 0, 0, 0, 5, 5, 43, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 249, 5, 5, 45, 249, 5, 5, 46, 249, 192, 0, 0, 0, 5, 5, 48, 249, 192, 0, 0, 0, 5, 5, 49, 249, 5, 5, 4, 237, 5, 5, 50, 249, 5, 5, 51, 249, 192, 0, 0, 0, 5, 5, 177, 231, 5, 5, 153, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 231, 5, 5, 95, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 236, 5, 5, 167, 226, 5, 5, 132, 230, 5, 5, 205, 228, 192, 0, 0, 0, 5, 5, 54, 249, 192, 0, 0, 0, 5, 5, 53, 249, 5, 5, 3, 230, 5, 5, 55, 249, 5, 5, 52, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 249, 5, 5, 244, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 238, 192, 0, 0, 0, 5, 5, 129, 231, 5, 5, 57, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 236, 5, 5, 19, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 226, 192, 0, 0, 0, 5, 5, 203, 231, 5, 5, 36, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 235, 192, 0, 0, 0, 5, 5, 28, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 237, 5, 5, 163, 236, 192, 0, 0, 0, 5, 5, 60, 249, 192, 0, 0, 0, 5, 5, 135, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 249, 192, 0, 0, 0, 5, 5, 62, 249, 5, 5, 86, 231, 5, 5, 63, 249, 192, 0, 0, 0, 5, 5, 6, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 249, 5, 5, 65, 249, 5, 5, 67, 249, 192, 0, 0, 0, 5, 5, 66, 249, 192, 0, 0, 0, 5, 5, 64, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 232, 192, 0, 0, 0, 5, 5, 156, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 249, 192, 0, 0, 0, 5, 5, 32, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 241, 5, 5, 73, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 249, 5, 5, 75, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 249, 192, 0, 0, 0, 5, 5, 76, 249, 192, 0, 0, 0, 5, 5, 77, 249, 192, 0, 0, 0, 5, 5, 78, 249, 5, 5, 80, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 249, 192, 0, 0, 0, 5, 5, 209, 234, 192, 0, 0, 0, 5, 5, 82, 249, 5, 5, 112, 250, 5, 5, 38, 236, 192, 0, 0, 0, 5, 5, 83, 249, 192, 0, 0, 0, 5, 5, 84, 249, 192, 0, 0, 0, 5, 5, 85, 249, 5, 5, 239, 226, 192, 0, 0, 0, 5, 5, 87, 249, 192, 0, 0, 0, 5, 5, 99, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 249, 5, 5, 93, 249, 192, 0, 0, 0, 5, 5, 91, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 227, 5, 5, 90, 249, 5, 5, 94, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 249, 192, 0, 0, 0, 5, 5, 0, 166, 192, 0, 0, 0, 5, 5, 96, 249, 5, 5, 97, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 249, 192, 0, 0, 0, 5, 5, 99, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 247, 192, 0, 0, 0, 5, 5, 120, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 249, 5, 5, 103, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 249, 192, 0, 0, 0, 5, 5, 106, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 249, 5, 5, 35, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 249, 5, 5, 117, 226, 192, 0, 0, 0, 5, 5, 114, 249, 5, 5, 113, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 227, 192, 0, 0, 0, 5, 5, 99, 235, 5, 5, 30, 233, 5, 5, 82, 229, 192, 0, 0, 0, 5, 5, 50, 229, 5, 5, 179, 230, 192, 0, 0, 0, 5, 5, 102, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 249, 5, 5, 122, 236, 5, 5, 96, 227, 5, 5, 176, 234, 5, 5, 252, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 231, 5, 5, 228, 236, 192, 0, 0, 0, 5, 5, 111, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 235, 5, 5, 204, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 249, 5, 5, 117, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 235, 5, 5, 162, 236, 5, 5, 121, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 249, 5, 5, 144, 232, 5, 5, 243, 226, 5, 5, 244, 226, 5, 5, 123, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 227, 5, 5, 164, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 249, 5, 5, 127, 249, 5, 5, 128, 249, 192, 0, 0, 0, 5, 5, 130, 249, 5, 5, 129, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 235, 192, 0, 0, 0, 5, 5, 131, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 249, 192, 0, 0, 0, 5, 5, 133, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 249, 5, 5, 135, 249, 192, 0, 0, 0, 5, 5, 137, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 234, 5, 5, 153, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 249, 192, 0, 0, 0, 5, 5, 139, 249, 192, 0, 0, 0, 5, 5, 19, 238, 5, 5, 52, 242, 5, 5, 177, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 217, 192, 0, 0, 0, 5, 5, 0, 153, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 230, 5, 5, 170, 235, 5, 5, 42, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 249, 192, 0, 0, 0, 5, 5, 57, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 249, 5, 5, 148, 236, 192, 0, 0, 0, 5, 5, 30, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 229, 192, 0, 0, 0, 5, 5, 142, 249, 5, 5, 168, 226, 5, 5, 143, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 249, 5, 5, 147, 249, 192, 0, 0, 0, 5, 5, 149, 249, 5, 5, 145, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 249, 192, 0, 0, 0, 5, 5, 151, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 249, 5, 5, 153, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 249, 192, 0, 0, 0, 5, 5, 158, 249, 5, 5, 161, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 249, 5, 5, 159, 249, 5, 5, 160, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 249, 192, 0, 0, 0, 5, 5, 246, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 230, 5, 5, 163, 249, 5, 5, 164, 249, 5, 5, 51, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 234, 5, 5, 166, 249, 5, 5, 167, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 232, 5, 5, 13, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 232, 5, 5, 52, 226, 5, 5, 36, 228, 192, 0, 0, 0, 5, 5, 37, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 232, 5, 5, 173, 249, 5, 5, 38, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 249, 192, 0, 0, 0, 5, 5, 171, 249, 192, 0, 0, 0, 5, 5, 169, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 249, 5, 5, 175, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 249, 5, 5, 177, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 230, 192, 0, 0, 0, 5, 5, 180, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 227, 5, 5, 181, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 232, 5, 5, 165, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 249, 192, 0, 0, 0, 5, 5, 189, 249, 192, 0, 0, 0, 5, 5, 188, 249, 5, 5, 190, 249, 192, 0, 0, 0, 5, 5, 187, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 249, 192, 0, 0, 0, 5, 5, 195, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 227, 5, 5, 194, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 249, 192, 0, 0, 0, 5, 5, 199, 249, 5, 5, 198, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 249, 5, 5, 202, 249, 5, 5, 201, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 249, 192, 0, 0, 0, 5, 5, 208, 249, 192, 0, 0, 0, 5, 5, 209, 249, 5, 5, 210, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 249, 5, 5, 212, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 249, 5, 5, 214, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 234, 5, 5, 217, 249, 192, 0, 0, 0, 5, 5, 226, 234, 5, 5, 218, 249, 5, 5, 216, 249, 192, 0, 0, 0, 5, 5, 220, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 249, 192, 0, 0, 0, 5, 5, 225, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 249, 5, 5, 228, 249, 192, 0, 0, 0, 5, 5, 229, 249, 192, 0, 0, 0, 5, 5, 230, 249, 5, 5, 231, 249, 5, 5, 232, 249, 5, 5, 233, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 249, 5, 5, 235, 249, 192, 0, 0, 0, 5, 5, 46, 242, 5, 5, 236, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 245, 5, 5, 138, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 226, 5, 5, 98, 229, 5, 5, 238, 249, 5, 5, 237, 249, 5, 5, 253, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 249, 5, 5, 241, 249, 5, 5, 239, 249, 192, 0, 0, 0, 5, 5, 242, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 249, 192, 0, 0, 0, 5, 5, 209, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 156, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 249, 5, 5, 71, 235, 5, 5, 245, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 249, 5, 5, 249, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 249, 5, 5, 251, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 249, 192, 0, 0, 0, 5, 5, 232, 235, 5, 5, 187, 229, 192, 0, 0, 0, 5, 5, 170, 229, 5, 5, 245, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 228, 5, 5, 255, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 250, 192, 0, 0, 0, 5, 5, 5, 250, 5, 5, 6, 250, 192, 0, 0, 0, 5, 5, 10, 250, 192, 0, 0, 0, 5, 5, 184, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 250, 5, 5, 8, 250, 5, 5, 7, 250, 5, 5, 9, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 250, 5, 5, 14, 250, 5, 5, 13, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 145, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 250, 192, 0, 0, 0, 5, 5, 21, 250, 192, 0, 0, 0, 5, 5, 22, 250, 5, 5, 18, 250, 5, 5, 24, 250, 192, 0, 0, 0, 5, 5, 20, 250, 5, 5, 250, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 237, 192, 0, 0, 0, 5, 5, 23, 250, 5, 5, 19, 250, 5, 5, 17, 250, 5, 5, 16, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 250, 5, 5, 28, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 235, 5, 5, 26, 250, 5, 5, 0, 209, 5, 5, 31, 250, 192, 0, 0, 0, 5, 5, 33, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 227, 5, 5, 12, 250, 5, 5, 19, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 250, 5, 5, 32, 250, 5, 5, 177, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 250, 192, 0, 0, 0, 5, 5, 40, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 233, 192, 0, 0, 0, 5, 5, 41, 250, 192, 0, 0, 0, 5, 5, 149, 234, 192, 0, 0, 0, 5, 5, 46, 250, 5, 5, 42, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 235, 5, 5, 30, 236, 192, 0, 0, 0, 5, 5, 240, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 250, 5, 5, 222, 233, 5, 5, 45, 250, 5, 5, 44, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 250, 192, 0, 0, 0, 5, 5, 21, 227, 192, 0, 0, 0, 5, 5, 49, 250, 5, 5, 32, 230, 5, 5, 99, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 250, 5, 5, 58, 250, 192, 0, 0, 0, 5, 5, 57, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 250, 192, 0, 0, 0, 5, 5, 60, 250, 192, 0, 0, 0, 5, 5, 61, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 250, 5, 5, 65, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 226, 5, 5, 62, 250, 5, 5, 63, 250, 192, 0, 0, 0, 5, 5, 70, 229, 5, 5, 19, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 250, 192, 0, 0, 0, 5, 5, 69, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 250, 192, 0, 0, 0, 5, 5, 71, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 250, 192, 0, 0, 0, 5, 5, 69, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 250, 192, 0, 0, 0, 5, 5, 82, 250, 5, 5, 80, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 250, 5, 5, 85, 250, 192, 0, 0, 0, 5, 5, 78, 250, 192, 0, 0, 0, 5, 5, 83, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 250, 192, 0, 0, 0, 5, 5, 91, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 250, 5, 5, 145, 232, 5, 5, 157, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 250, 5, 5, 94, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 228, 5, 5, 97, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 250, 192, 0, 0, 0, 5, 5, 29, 230, 192, 0, 0, 0, 5, 5, 99, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 250, 5, 5, 102, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 250, 5, 5, 103, 250, 5, 5, 50, 237, 192, 0, 0, 0, 5, 5, 104, 250, 192, 0, 0, 0, 5, 5, 7, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 250, 192, 0, 0, 0, 5, 5, 241, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 250, 5, 5, 126, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 250, 5, 5, 110, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 250, 5, 5, 63, 229, 5, 5, 39, 236, 5, 5, 221, 235, 5, 5, 216, 239, 192, 0, 0, 0, 5, 5, 89, 242, 5, 5, 245, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 250, 5, 5, 166, 227, 5, 5, 114, 250, 5, 5, 115, 250, 5, 5, 116, 250, 192, 0, 0, 0, 5, 5, 71, 229, 192, 0, 0, 0, 5, 5, 117, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 243, 5, 5, 53, 236, 192, 0, 0, 0, 5, 5, 137, 232, 5, 5, 118, 250, 5, 5, 120, 250, 5, 5, 119, 250, 192, 0, 0, 0, 5, 5, 121, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 250, 192, 0, 0, 0, 5, 5, 126, 250, 5, 5, 127, 250, 192, 0, 0, 0, 5, 5, 128, 250, 192, 0, 0, 0, 5, 5, 129, 250, 5, 5, 130, 250, 5, 5, 131, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 250, 5, 5, 133, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 228, 192, 0, 0, 0, 5, 5, 135, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 232, 5, 5, 136, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 250, 5, 5, 104, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 247, 5, 5, 110, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 250, 192, 0, 0, 0, 5, 5, 141, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 250, 5, 5, 144, 250, 5, 5, 145, 250, 5, 5, 8, 237, 5, 5, 142, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 250, 5, 5, 147, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 250, 192, 0, 0, 0, 5, 5, 148, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 250, 5, 5, 150, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 250, 5, 5, 133, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 247, 5, 5, 11, 229, 5, 5, 68, 230, 5, 5, 208, 247, 5, 5, 4, 227, 5, 5, 48, 228, 5, 5, 29, 228, 5, 5, 154, 250, 5, 5, 154, 250, 5, 5, 84, 228, 5, 5, 24, 228, 5, 5, 142, 238, 5, 5, 2, 234, 5, 5, 144, 240, 5, 5, 237, 243, 5, 5, 157, 236, 5, 5, 176, 246, 5, 5, 158, 236, 5, 5, 159, 236, 5, 5, 121, 248, 5, 5, 244, 241, 5, 5, 164, 236, 5, 5, 23, 243, 5, 5, 121, 243, 5, 5, 166, 236, 5, 5, 167, 236, 5, 5, 176, 249, 5, 5, 85, 237, 5, 5, 169, 236, 5, 5, 171, 236, 5, 5, 58, 243, 5, 5, 174, 236, 5, 5, 95, 250, 5, 5, 170, 236, 5, 5, 172, 236, 5, 5, 173, 236, 5, 5, 73, 247, 5, 5, 196, 240, 5, 5, 239, 245, 192, 0, 0, 0, 5, 5, 35, 237, 5, 5, 37, 237, 5, 5, 40, 237, 5, 5, 43, 237, 5, 5, 48, 237, 5, 5, 125, 237, 5, 5, 250, 236, 5, 5, 16, 238, 192, 0, 0, 0, 5, 5, 28, 237, 5, 5, 57, 243, 5, 5, 13, 244, 5, 5, 45, 237, 5, 5, 172, 246, 5, 5, 208, 236, 5, 5, 31, 237, 5, 5, 32, 237, 5, 5, 27, 237, 5, 5, 157, 229, 5, 5, 68, 244, 5, 5, 100, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 230, 5, 5, 54, 237, 5, 5, 7, 239, 5, 5, 36, 237, 5, 5, 213, 244, 5, 5, 46, 237, 5, 5, 42, 237, 5, 5, 80, 244, 5, 5, 30, 237, 5, 5, 163, 236, 5, 5, 3, 239, 192, 0, 0, 0, 5, 5, 7, 242, 192, 0, 0, 0, 5, 5, 41, 237, 5, 5, 245, 236, 5, 5, 78, 245, 5, 5, 30, 249, 5, 5, 86, 249, 5, 5, 52, 237, 5, 5, 160, 250, 5, 5, 213, 236, 5, 5, 221, 236, 5, 5, 0, 155, 5, 5, 229, 234, 5, 5, 227, 236, 5, 5, 171, 247, 5, 5, 185, 240, 5, 5, 244, 241, 5, 5, 160, 232, 5, 5, 180, 232, 5, 5, 43, 234, 5, 5, 158, 233, 5, 5, 190, 236, 5, 5, 0, 185, 5, 5, 198, 235, 192, 0, 0, 0, 5, 5, 116, 235, 5, 5, 58, 235, 5, 5, 18, 235, 5, 5, 197, 234, 5, 5, 40, 241, 5, 5, 167, 229, 5, 5, 59, 238, 5, 5, 121, 229, 5, 5, 244, 230, 5, 5, 141, 236, 192, 0, 0, 0, 5, 5, 180, 229, 5, 5, 168, 232, 5, 5, 36, 233, 5, 5, 114, 230, 5, 5, 81, 230, 5, 5, 193, 236, 5, 5, 194, 236, 5, 5, 210, 236, 5, 5, 199, 237, 5, 5, 225, 237, 5, 5, 216, 236, 5, 5, 222, 236, 5, 5, 223, 236, 5, 5, 224, 236, 5, 5, 226, 236, 5, 5, 22, 238, 5, 5, 26, 237, 5, 5, 195, 230, 5, 5, 233, 239, 5, 5, 207, 236, 5, 5, 254, 242, 5, 5, 61, 244, 5, 5, 6, 249, 5, 5, 202, 249, 5, 5, 7, 237, 5, 5, 114, 250, 5, 5, 229, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 248, 5, 5, 47, 234, 5, 5, 17, 237, 5, 5, 151, 240, 5, 5, 50, 234, 5, 5, 18, 237, 5, 5, 19, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 237, 5, 5, 22, 237, 5, 5, 56, 248, 5, 5, 23, 237, 5, 5, 24, 237, 192, 0, 0, 0, 5, 5, 11, 237, 5, 5, 12, 237, 5, 5, 0, 212, 5, 5, 13, 237, 5, 5, 14, 237, 192, 0, 0, 0, 5, 5, 15, 237, 5, 5, 48, 234, 5, 5, 49, 234, 192, 0, 0, 0, 5, 5, 20, 237, 5, 5, 111, 243, 5, 5, 247, 236, 5, 5, 191, 238, 5, 5, 43, 234, 5, 5, 252, 236, 5, 5, 253, 236, 5, 5, 254, 236, 5, 5, 135, 243, 5, 5, 134, 245, 5, 5, 166, 245, 5, 5, 3, 237, 5, 5, 5, 237, 5, 5, 78, 249, 5, 5, 228, 236, 5, 5, 249, 236, 5, 5, 106, 244, 5, 5, 151, 248, 5, 5, 50, 249, 5, 5, 71, 240, 5, 5, 209, 236, 5, 5, 211, 236, 5, 5, 214, 236, 5, 5, 34, 234, 5, 5, 215, 236, 5, 5, 244, 241, 5, 5, 48, 243, 5, 5, 219, 236, 5, 5, 133, 246, 5, 5, 225, 236, 5, 5, 204, 236, 5, 5, 81, 241, 5, 5, 25, 249, 5, 5, 195, 236, 192, 0, 0, 0, 5, 5, 82, 236, 5, 5, 196, 236, 5, 5, 197, 236, 5, 5, 198, 236, 5, 5, 199, 236, 5, 5, 200, 236, 5, 5, 239, 234, 5, 5, 246, 236, 5, 5, 49, 237, 5, 5, 161, 240, 5, 5, 188, 236, 5, 5, 231, 236, 5, 5, 167, 239, 5, 5, 164, 242, 5, 5, 238, 236, 5, 5, 189, 236, 5, 5, 106, 240, 5, 5, 62, 228, 5, 5, 190, 236, 5, 5, 202, 236, 5, 5, 176, 236, 5, 5, 177, 236, 5, 5, 178, 236, 5, 5, 0, 181, 5, 5, 179, 236, 5, 5, 180, 236, 5, 5, 103, 233, 5, 5, 181, 236, 5, 5, 183, 236, 5, 5, 140, 240, 5, 5, 184, 236, 5, 5, 185, 236, 5, 5, 186, 236, 5, 5, 187, 236, 5, 5, 223, 233, 5, 5, 112, 233, 5, 5, 78, 238, 5, 5, 235, 236, 192, 0, 0, 0, 5, 5, 171, 246, 5, 5, 239, 236, 5, 5, 240, 236, 5, 5, 241, 236, 5, 5, 233, 236, 5, 5, 234, 236, 5, 5, 237, 236, 5, 5, 191, 236, 5, 5, 246, 226, 5, 5, 201, 236, 192, 0, 0, 0, 5, 5, 12, 243, 5, 5, 31, 230, 5, 5, 134, 230, 5, 5, 231, 232, 5, 5, 215, 229, 5, 5, 194, 231, 5, 5, 155, 233, 5, 5, 153, 232, 5, 5, 190, 233, 5, 5, 150, 232, 5, 5, 214, 233, 5, 5, 183, 235, 5, 5, 58, 248, 5, 5, 38, 229, 5, 5, 49, 229, 5, 5, 157, 228, 5, 5, 223, 226, 5, 5, 193, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 233, 192, 0, 0, 0, 5, 5, 154, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 232, 5, 5, 51, 226, 5, 5, 255, 236, 5, 5, 74, 231, 5, 5, 247, 230, 5, 5, 61, 235, 5, 5, 81, 236, 5, 5, 162, 231, 5, 5, 5, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 205, 5, 5, 150, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 234, 5, 5, 4, 230, 5, 5, 83, 227, 5, 5, 69, 233, 192, 0, 0, 0, 5, 5, 4, 237, 5, 5, 44, 235, 5, 5, 22, 232, 5, 5, 34, 236, 5, 5, 117, 235, 5, 5, 6, 228, 5, 5, 186, 234, 5, 5, 253, 226, 5, 5, 182, 232, 5, 5, 103, 227, 5, 5, 89, 235, 5, 5, 201, 235, 5, 5, 32, 232, 5, 5, 134, 239, 5, 5, 179, 226, 5, 5, 204, 226, 5, 5, 70, 232, 5, 5, 12, 233, 5, 5, 16, 235, 5, 5, 115, 227, 5, 5, 183, 230, 5, 5, 95, 234, 5, 5, 188, 226, 5, 5, 185, 230, 5, 5, 57, 227, 5, 5, 63, 230, 192, 0, 0, 0, 5, 5, 156, 232, 5, 5, 200, 234, 5, 5, 64, 230, 5, 5, 241, 229, 5, 5, 125, 227, 5, 5, 108, 236, 5, 5, 12, 232, 5, 5, 151, 230, 5, 5, 140, 226, 5, 5, 91, 233, 5, 5, 68, 229, 5, 5, 237, 233, 5, 5, 201, 231, 5, 5, 21, 237, 5, 5, 69, 245, 5, 5, 169, 234, 5, 5, 188, 230, 5, 5, 66, 230, 5, 5, 121, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 232, 5, 5, 6, 227, 5, 5, 249, 229, 5, 5, 54, 226, 5, 5, 22, 228, 5, 5, 14, 235, 5, 5, 73, 232, 192, 0, 0, 0, 5, 5, 0, 205, 5, 5, 19, 234, 5, 5, 245, 227, 5, 5, 15, 235, 5, 5, 87, 228, 192, 0, 0, 0, 5, 5, 84, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 235, 5, 5, 222, 237, 5, 5, 250, 231, 192, 0, 0, 0, 5, 5, 136, 230, 5, 5, 202, 237, 5, 5, 94, 236, 5, 5, 73, 230, 5, 5, 253, 226, 192, 0, 0, 0, 5, 5, 129, 238, 192, 0, 0, 0, 5, 5, 49, 233, 5, 5, 75, 235, 5, 5, 62, 226, 5, 5, 213, 235, 5, 5, 60, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 229, 5, 5, 21, 240, 5, 5, 80, 240, 5, 5, 63, 231, 5, 5, 86, 236, 5, 5, 70, 232, 192, 0, 0, 0, 5, 5, 12, 233, 5, 5, 125, 232, 5, 5, 230, 240, 5, 5, 208, 240, 192, 0, 0, 0, 5, 5, 33, 241, 5, 5, 154, 231, 5, 5, 37, 237, 5, 5, 184, 235, 5, 5, 31, 231, 5, 5, 63, 242, 5, 5, 180, 229, 5, 5, 196, 236, 192, 0, 0, 0, 5, 5, 16, 230, 5, 5, 57, 227, 5, 5, 244, 233, 5, 5, 63, 230, 192, 0, 0, 0, 5, 5, 77, 230, 5, 5, 166, 234, 5, 5, 250, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 226, 192, 0, 0, 0, 5, 5, 220, 243, 5, 5, 51, 226, 5, 5, 161, 231, 5, 5, 34, 233, 192, 0, 0, 0, 5, 5, 233, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 231, 192, 0, 0, 0, 5, 5, 37, 245, 5, 5, 21, 237, 192, 0, 0, 0, 5, 5, 66, 230, 5, 5, 37, 229, 5, 5, 148, 226, 192, 0, 0, 0, 5, 5, 63, 247, 5, 5, 64, 235, 5, 5, 249, 229, 5, 5, 24, 233, 5, 5, 192, 230, 5, 5, 169, 231, 5, 5, 54, 226, 5, 5, 160, 232, 5, 5, 89, 236, 5, 5, 22, 228, 5, 5, 43, 241, 5, 5, 73, 232, 5, 5, 90, 236, 5, 5, 114, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 234, 5, 5, 81, 236, 192, 0, 0, 0, 5, 5, 245, 227, 192, 0, 0, 0, 5, 5, 15, 235, 192, 0, 0, 0, 5, 5, 154, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 14, 0, 202, 193, 14, 0, 202, 226, 14, 0, 202, 3, 15, 0, 202, 36, 15, 0, 202, 69, 15, 0, 202, 102, 15, 0, 202, 135, 15, 0, 202, 168, 15, 0, 202, 201, 15, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 192, 0, 0, 0, 192, 0, 0, 0, 133, 5, 24, 9, 133, 5, 70, 10, 133, 5, 72, 10, 133, 5, 34, 7, 192, 0, 0, 0, 198, 65, 13, 0, 198, 1, 0, 0, 198, 65, 0, 0, 198, 129, 0, 0, 198, 193, 0, 0, 198, 1, 1, 0, 198, 97, 11, 0, 198, 161, 11, 0, 198, 225, 11, 0, 198, 193, 5, 0, 192, 0, 0, 0, 198, 33, 0, 0, 198, 97, 0, 0, 198, 161, 0, 0, 198, 225, 0, 0, 198, 33, 1, 0, 198, 65, 1, 0, 198, 161, 1, 0, 198, 1, 2, 0, 198, 97, 2, 0, 198, 193, 2, 0, 198, 33, 3, 0, 198, 129, 3, 0, 198, 225, 3, 0, 198, 65, 4, 0, 198, 161, 4, 0, 198, 1, 5, 0, 198, 97, 5, 0, 198, 225, 5, 0, 198, 65, 6, 0, 198, 161, 6, 0, 198, 1, 7, 0, 198, 33, 7, 0, 198, 65, 7, 0, 198, 97, 7, 0, 198, 129, 7, 0, 198, 161, 7, 0, 198, 65, 8, 0, 198, 225, 8, 0, 198, 129, 9, 0, 198, 33, 10, 0, 198, 193, 10, 0, 198, 225, 10, 0, 198, 1, 11, 0, 198, 33, 11, 0, 198, 65, 11, 0, 198, 129, 11, 0, 198, 193, 11, 0, 198, 1, 12, 0, 198, 33, 12, 0, 198, 65, 12, 0, 198, 97, 12, 0, 198, 129, 12, 0, 198, 161, 12, 0, 198, 225, 12, 0, 198, 97, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 129, 35, 0, 133, 5, 6, 121, 133, 5, 8, 121, 133, 5, 183, 121, 133, 5, 10, 121, 133, 5, 187, 121, 133, 5, 189, 121, 133, 5, 12, 121, 133, 5, 14, 121, 133, 5, 16, 121, 133, 5, 195, 121, 133, 5, 197, 121, 133, 5, 199, 121, 133, 5, 201, 121, 133, 5, 203, 121, 133, 5, 205, 121, 133, 5, 52, 121, 133, 5, 18, 121, 133, 5, 20, 121, 133, 5, 22, 121, 133, 5, 66, 121, 133, 5, 24, 121, 133, 5, 26, 121, 133, 5, 28, 121, 133, 5, 30, 121, 133, 5, 32, 121, 133, 5, 34, 121, 133, 5, 36, 121, 133, 5, 38, 121, 133, 5, 40, 121, 133, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 133, 5, 103, 121, 133, 5, 105, 121, 133, 5, 107, 121, 133, 5, 109, 121, 133, 5, 111, 121, 133, 5, 113, 121, 192, 0, 0, 0, 192, 0, 0, 0, 133, 5, 115, 121, 133, 5, 117, 121, 133, 5, 119, 121, 133, 5, 121, 121, 133, 5, 123, 121, 133, 5, 125, 121, 192, 0, 0, 0, 192, 0, 0, 0, 133, 5, 127, 121, 133, 5, 129, 121, 133, 5, 131, 121, 133, 5, 133, 121, 133, 5, 135, 121, 133, 5, 137, 121, 192, 0, 0, 0, 192, 0, 0, 0, 133, 5, 139, 121, 133, 5, 141, 121, 133, 5, 143, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 13, 5, 5, 140, 13, 5, 5, 135, 12, 5, 5, 0, 4, 5, 5, 139, 12, 5, 5, 142, 13, 5, 5, 192, 13, 192, 0, 0, 0, 133, 5, 172, 12, 133, 5, 99, 12, 133, 5, 103, 12, 133, 5, 101, 12, 133, 5, 105, 12, 133, 5, 180, 12, 133, 5, 185, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 234, 5, 5, 22, 232, 5, 5, 68, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 236, 5, 5, 196, 237, 192, 0, 0, 0, 5, 5, 39, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 229, 192, 0, 0, 0, 5, 5, 21, 231, 5, 5, 212, 237, 192, 0, 0, 0, 5, 5, 161, 233, 5, 5, 222, 237, 192, 0, 0, 0, 5, 5, 232, 237, 5, 5, 89, 231, 192, 0, 0, 0, 5, 5, 65, 229, 192, 0, 0, 0, 5, 5, 252, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 236, 5, 5, 117, 235, 5, 5, 6, 228, 5, 5, 73, 230, 5, 5, 141, 235, 5, 5, 25, 238, 5, 5, 198, 235, 5, 5, 42, 238, 5, 5, 186, 234, 5, 5, 110, 234, 5, 5, 78, 232, 192, 0, 0, 0, 5, 5, 222, 227, 5, 5, 222, 227, 5, 5, 222, 227, 192, 0, 0, 0, 5, 5, 189, 226, 5, 5, 176, 227, 5, 5, 62, 238, 192, 0, 0, 0, 5, 5, 223, 227, 5, 5, 38, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 227, 5, 5, 76, 233, 5, 5, 108, 230, 5, 5, 95, 238, 192, 0, 0, 0, 5, 5, 164, 233, 5, 5, 80, 228, 5, 5, 119, 238, 5, 5, 247, 231, 5, 5, 247, 231, 5, 5, 129, 238, 5, 5, 160, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 238, 5, 5, 182, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 235, 5, 5, 194, 231, 5, 5, 28, 232, 5, 5, 23, 231, 5, 5, 41, 231, 192, 0, 0, 0, 5, 5, 83, 228, 192, 0, 0, 0, 5, 5, 143, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 234, 5, 5, 63, 233, 192, 0, 0, 0, 5, 5, 101, 232, 5, 5, 14, 236, 5, 5, 31, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 239, 5, 5, 82, 239, 192, 0, 0, 0, 5, 5, 102, 239, 192, 0, 0, 0, 5, 5, 43, 234, 5, 5, 115, 239, 192, 0, 0, 0, 5, 5, 101, 230, 5, 5, 215, 230, 5, 5, 184, 233, 5, 5, 122, 239, 192, 0, 0, 0, 5, 5, 141, 233, 5, 5, 134, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 232, 192, 0, 0, 0, 5, 5, 179, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 230, 5, 5, 35, 237, 192, 0, 0, 0, 5, 5, 239, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 228, 5, 5, 10, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 234, 5, 5, 228, 229, 192, 0, 0, 0, 5, 5, 59, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 226, 192, 0, 0, 0, 5, 5, 247, 233, 5, 5, 12, 230, 5, 5, 4, 229, 5, 5, 63, 231, 5, 5, 4, 229, 192, 0, 0, 0, 5, 5, 70, 232, 5, 5, 143, 228, 5, 5, 76, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 233, 5, 5, 144, 240, 5, 5, 150, 231, 5, 5, 157, 240, 192, 0, 0, 0, 5, 5, 150, 235, 5, 5, 179, 240, 5, 5, 205, 240, 192, 0, 0, 0, 5, 5, 178, 234, 192, 0, 0, 0, 5, 5, 60, 230, 5, 5, 37, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 226, 192, 0, 0, 0, 5, 5, 218, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 235, 5, 5, 93, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 230, 5, 5, 70, 241, 192, 0, 0, 0, 5, 5, 183, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 235, 5, 5, 210, 237, 5, 5, 119, 229, 192, 0, 0, 0, 5, 5, 186, 245, 192, 0, 0, 0, 5, 5, 37, 237, 5, 5, 184, 235, 192, 0, 0, 0, 5, 5, 116, 241, 5, 5, 75, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 241, 192, 0, 0, 0, 5, 5, 95, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 228, 192, 0, 0, 0, 5, 5, 15, 230, 192, 0, 0, 0, 5, 5, 58, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 229, 5, 5, 228, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 234, 192, 0, 0, 0, 5, 5, 68, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 234, 5, 5, 188, 226, 5, 5, 196, 236, 5, 5, 18, 229, 5, 5, 69, 231, 5, 5, 145, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 229, 5, 5, 167, 242, 192, 0, 0, 0, 5, 5, 16, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 242, 5, 5, 16, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 242, 192, 0, 0, 0, 5, 5, 244, 233, 5, 5, 255, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 229, 192, 0, 0, 0, 5, 5, 190, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 243, 5, 5, 77, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 243, 5, 5, 96, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 244, 5, 5, 73, 231, 5, 5, 73, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 229, 192, 0, 0, 0, 5, 5, 116, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 244, 5, 5, 223, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 227, 192, 0, 0, 0, 5, 5, 34, 245, 192, 0, 0, 0, 5, 5, 45, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 245, 192, 0, 0, 0, 5, 5, 27, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 208, 5, 5, 53, 230, 192, 0, 0, 0, 5, 5, 144, 226, 5, 5, 161, 235, 5, 5, 164, 226, 5, 5, 34, 228, 192, 0, 0, 0, 5, 5, 81, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 246, 5, 5, 90, 246, 192, 0, 0, 0, 5, 5, 252, 232, 192, 0, 0, 0, 5, 5, 156, 227, 5, 5, 19, 228, 5, 5, 138, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 227, 5, 5, 208, 236, 5, 5, 181, 246, 192, 0, 0, 0, 5, 5, 185, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 231, 5, 5, 89, 236, 5, 5, 43, 241, 5, 5, 186, 247, 192, 0, 0, 0, 5, 5, 75, 227, 5, 5, 209, 247, 192, 0, 0, 0, 5, 5, 133, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 235, 192, 0, 0, 0, 5, 5, 128, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 64, 3, 0, 5, 0, 0, 6, 122, 64, 133, 0, 5, 0, 0, 6, 122, 64, 3, 0, 5, 0, 0, 8, 122, 64, 133, 0, 5, 0, 0, 8, 122, 64, 3, 0, 5, 0, 0, 10, 122, 64, 133, 0, 5, 0, 0, 10, 122, 64, 3, 0, 5, 0, 0, 14, 122, 64, 133, 0, 5, 0, 0, 14, 122, 64, 3, 0, 5, 0, 0, 16, 122, 64, 133, 0, 5, 0, 0, 16, 122, 64, 133, 0, 5, 0, 0, 18, 122, 64, 133, 0, 5, 0, 0, 18, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 20, 122, 64, 133, 0, 5, 0, 0, 20, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 22, 122, 64, 133, 0, 5, 0, 0, 22, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 24, 122, 64, 133, 0, 5, 0, 0, 24, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 26, 122, 64, 133, 0, 5, 0, 0, 26, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 28, 122, 64, 133, 0, 5, 0, 0, 28, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 30, 122, 64, 133, 0, 5, 0, 0, 30, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 32, 122, 64, 133, 0, 5, 0, 0, 32, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 34, 122, 64, 133, 0, 5, 0, 0, 34, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 36, 122, 64, 133, 0, 5, 0, 0, 36, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 38, 122, 64, 133, 0, 5, 0, 0, 38, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 40, 122, 64, 133, 0, 5, 0, 0, 40, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 3, 0, 5, 0, 0, 42, 122, 64, 133, 0, 5, 0, 0, 42, 122, 64, 133, 0, 5, 0, 0, 42, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 44, 122, 64, 133, 0, 5, 0, 0, 44, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 46, 122, 64, 133, 0, 5, 0, 0, 46, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 48, 122, 64, 133, 0, 5, 0, 0, 50, 122, 64, 133, 0, 5, 0, 0, 52, 122, 64, 133, 0, 5, 0, 0, 54, 122, 64, 133, 0, 5, 0, 0, 56, 122, 64, 133, 0, 5, 0, 0, 58, 122, 64, 133, 0, 5, 0, 0, 58, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 58, 122, 0, 5, 0, 176, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 60, 122, 64, 133, 0, 5, 0, 0, 60, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 60, 122, 0, 5, 0, 176, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 62, 122, 64, 133, 0, 5, 0, 0, 62, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 62, 122, 0, 5, 0, 176, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 64, 122, 64, 133, 0, 5, 0, 0, 64, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 64, 122, 0, 5, 0, 176, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 66, 122, 64, 133, 0, 5, 0, 0, 66, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 66, 122, 0, 5, 0, 176, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 68, 122, 64, 133, 0, 5, 0, 0, 70, 122, 64, 133, 0, 5, 0, 0, 72, 122, 64, 133, 0, 5, 0, 0, 74, 122, 64, 133, 0, 5, 0, 0, 76, 122, 64, 3, 0, 5, 0, 0, 78, 122, 64, 133, 0, 5, 0, 0, 78, 122, 64, 3, 0, 5, 0, 0, 80, 122, 64, 133, 0, 5, 0, 0, 80, 122, 64, 3, 0, 5, 0, 0, 82, 122, 64, 133, 0, 5, 0, 0, 82, 122, 64, 133, 0, 5, 0, 0, 84, 122, 64, 133, 0, 5, 0, 0, 86, 122, 64, 133, 0, 5, 0, 0, 88, 122, 64, 133, 0, 5, 0, 0, 90, 122, 64, 133, 0, 5, 0, 0, 92, 122, 64, 3, 0, 5, 0, 0, 94, 122, 64, 133, 0, 5, 0, 0, 94, 122, 64, 133, 0, 5, 0, 0, 96, 122, 64, 133, 0, 5, 0, 0, 98, 122, 64, 133, 0, 5, 0, 0, 100, 122, 64, 133, 0, 5, 0, 0, 102, 122, 64, 133, 0, 5, 0, 0, 10, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 3, 0, 5, 0, 0, 18, 122, 64, 3, 0, 5, 0, 0, 24, 122, 64, 133, 0, 5, 0, 0, 94, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 96, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 98, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 100, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 6, 122, 64, 2, 0, 5, 0, 0, 8, 122, 64, 2, 0, 5, 0, 0, 10, 122, 64, 2, 0, 5, 0, 0, 14, 122, 64, 2, 0, 5, 0, 0, 16, 122, 64, 5, 0, 5, 0, 0, 116, 13, 64, 4, 0, 5, 0, 0, 6, 122, 64, 4, 0, 5, 0, 0, 8, 122, 64, 4, 0, 5, 0, 0, 10, 122, 64, 4, 0, 5, 0, 0, 14, 122, 64, 4, 0, 5, 0, 0, 16, 122, 64, 4, 0, 5, 0, 0, 18, 122, 0, 5, 0, 5, 0, 0, 18, 122, 64, 2, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 20, 122, 64, 2, 0, 5, 0, 0, 22, 122, 64, 4, 0, 5, 0, 0, 24, 122, 0, 5, 0, 5, 0, 0, 24, 122, 64, 3, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 26, 122, 64, 2, 0, 5, 0, 0, 28, 122, 64, 4, 0, 5, 0, 0, 30, 122, 64, 4, 0, 5, 0, 0, 32, 122, 64, 2, 0, 5, 0, 0, 34, 122, 64, 2, 0, 5, 0, 0, 36, 122, 64, 2, 0, 5, 0, 0, 38, 122, 64, 2, 0, 5, 0, 0, 40, 122, 64, 4, 0, 5, 0, 0, 42, 122, 64, 2, 0, 5, 0, 0, 44, 122, 64, 2, 0, 5, 0, 0, 46, 122, 64, 2, 0, 5, 0, 0, 48, 122, 64, 2, 0, 5, 0, 0, 50, 122, 64, 2, 0, 5, 0, 0, 52, 122, 64, 2, 0, 5, 0, 0, 54, 122, 64, 2, 0, 5, 0, 0, 56, 122, 64, 2, 0, 5, 0, 0, 58, 122, 64, 2, 0, 5, 0, 0, 60, 122, 64, 2, 0, 5, 0, 0, 62, 122, 64, 2, 0, 5, 0, 0, 64, 122, 64, 2, 0, 5, 0, 0, 66, 122, 64, 4, 0, 5, 0, 0, 68, 122, 64, 2, 0, 5, 0, 0, 70, 122, 64, 2, 0, 5, 0, 0, 72, 122, 64, 4, 0, 5, 0, 0, 74, 122, 64, 2, 0, 5, 0, 0, 76, 122, 64, 4, 0, 5, 0, 0, 78, 122, 64, 4, 0, 5, 0, 0, 80, 122, 64, 4, 0, 5, 0, 0, 82, 122, 64, 2, 0, 5, 0, 0, 84, 122, 64, 2, 0, 5, 0, 0, 86, 122, 64, 2, 0, 5, 0, 0, 88, 122, 64, 2, 0, 5, 0, 0, 90, 122, 64, 2, 0, 5, 0, 0, 92, 122, 64, 4, 0, 5, 0, 0, 94, 122, 64, 4, 0, 5, 0, 0, 96, 122, 64, 4, 0, 5, 0, 0, 98, 122, 64, 4, 0, 5, 0, 0, 100, 122, 64, 2, 0, 5, 0, 0, 102, 122, 64, 5, 0, 5, 0, 0, 116, 13, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 6, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 8, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 10, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 14, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 16, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 18, 122, 0, 5, 0, 174, 0, 0, 0, 0, 0, 5, 0, 5, 0, 0, 18, 122, 64, 2, 0, 174, 0, 0, 0, 0, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 20, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 22, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 24, 122, 0, 5, 0, 174, 0, 0, 0, 0, 0, 5, 0, 5, 0, 0, 24, 122, 64, 3, 0, 174, 0, 0, 0, 0, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 26, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 28, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 30, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 32, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 34, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 36, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 38, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 40, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 42, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 44, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 46, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 48, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 50, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 52, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 54, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 56, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 58, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 60, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 62, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 64, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 66, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 68, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 70, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 72, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 74, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 76, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 78, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 80, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 82, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 84, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 86, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 88, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 90, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 92, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 94, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 96, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 98, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 100, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 102, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 5, 0, 5, 0, 0, 26, 122, 0, 5, 0, 70, 0, 0, 46, 122, 0, 133, 0, 5, 0, 247, 101, 121, 0, 0, 0, 0, 5, 5, 18, 122, 2, 174, 0, 0, 5, 5, 24, 122, 3, 174, 0, 0, 5, 5, 116, 13, 5, 174, 0, 0, 4, 5, 6, 122, 5, 174, 0, 0, 4, 5, 8, 122, 5, 174, 0, 0, 4, 5, 10, 122, 5, 174, 0, 0, 4, 5, 14, 122, 5, 174, 0, 0, 4, 5, 16, 122, 5, 174, 0, 0, 4, 5, 18, 122, 5, 174, 0, 0, 5, 5, 18, 122, 2, 174, 0, 0, 5, 174, 0, 0, 2, 5, 20, 122, 5, 174, 0, 0, 2, 5, 22, 122, 5, 174, 0, 0, 4, 5, 24, 122, 5, 174, 0, 0, 5, 5, 24, 122, 3, 174, 0, 0, 5, 174, 0, 0, 2, 5, 26, 122, 5, 174, 0, 0, 2, 5, 28, 122, 5, 174, 0, 0, 4, 5, 30, 122, 5, 174, 0, 0, 4, 5, 32, 122, 5, 174, 0, 0, 2, 5, 34, 122, 5, 174, 0, 0, 2, 5, 36, 122, 5, 174, 0, 0, 2, 5, 38, 122, 5, 174, 0, 0, 2, 5, 40, 122, 5, 174, 0, 0, 4, 5, 42, 122, 5, 174, 0, 0, 2, 5, 44, 122, 5, 174, 0, 0, 2, 5, 46, 122, 5, 174, 0, 0, 2, 5, 48, 122, 5, 174, 0, 0, 2, 5, 50, 122, 5, 174, 0, 0, 2, 5, 52, 122, 5, 174, 0, 0, 2, 5, 54, 122, 5, 174, 0, 0, 2, 5, 56, 122, 5, 174, 0, 0, 2, 5, 58, 122, 5, 174, 0, 0, 2, 5, 60, 122, 5, 174, 0, 0, 2, 5, 62, 122, 5, 174, 0, 0, 2, 5, 64, 122, 5, 174, 0, 0, 2, 5, 66, 122, 5, 174, 0, 0, 4, 5, 68, 122, 5, 174, 0, 0, 2, 5, 70, 122, 5, 174, 0, 0, 2, 5, 72, 122, 5, 174, 0, 0, 4, 5, 74, 122, 5, 174, 0, 0, 2, 5, 76, 122, 5, 174, 0, 0, 4, 5, 78, 122, 5, 174, 0, 0, 4, 5, 80, 122, 5, 174, 0, 0, 4, 5, 82, 122, 5, 174, 0, 0, 2, 5, 84, 122, 5, 174, 0, 0, 2, 5, 86, 122, 5, 174, 0, 0, 2, 5, 88, 122, 5, 174, 0, 0, 2, 5, 90, 122, 5, 174, 0, 0, 2, 5, 92, 122, 5, 174, 0, 0, 4, 5, 94, 122, 5, 174, 0, 0, 4, 5, 96, 122, 5, 174, 0, 0, 4, 5, 98, 122, 5, 174, 0, 0, 4, 5, 100, 122, 5, 174, 0, 0, 2, 5, 102, 122, 5, 174, 0, 0, 5, 5, 82, 122, 5, 70, 86, 122, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 116, 13, 5, 5, 0, 0, 87, 0, 109, 48, 48, 1, 131, 48, 204, 0, 142, 48, 154, 0, 147, 48, 132, 0, 150, 48, 118, 0, 150, 48, 255, 255, 24, 122, 4, 5, 153, 48, 22, 0, 154, 48, 4, 0, 111, 48, 255, 255, 58, 122, 2, 5, 114, 48, 255, 255, 60, 122, 2, 5, 117, 48, 255, 255, 62, 122, 2, 5, 120, 48, 255, 255, 64, 122, 2, 5, 123, 48, 255, 255, 66, 122, 2, 5, 20, 0, 93, 48, 48, 0, 104, 48, 26, 0, 117, 48, 12, 0, 117, 48, 255, 255, 62, 122, 2, 5, 120, 48, 255, 255, 64, 122, 2, 5, 123, 48, 255, 255, 66, 122, 2, 5, 104, 48, 255, 255, 46, 122, 2, 5, 111, 48, 255, 255, 58, 122, 2, 5, 114, 48, 255, 255, 60, 122, 2, 5, 93, 48, 255, 255, 36, 122, 2, 5, 95, 48, 255, 255, 38, 122, 2, 5, 97, 48, 255, 255, 40, 122, 2, 5, 100, 48, 255, 255, 42, 122, 4, 5, 102, 48, 255, 255, 44, 122, 2, 5, 83, 48, 20, 0, 83, 48, 255, 255, 26, 122, 2, 5, 85, 48, 255, 255, 28, 122, 2, 5, 87, 48, 255, 255, 30, 122, 4, 5, 89, 48, 255, 255, 32, 122, 4, 5, 91, 48, 255, 255, 34, 122, 2, 5, 70, 48, 255, 255, 10, 122, 4, 5, 75, 48, 197, 162, 77, 48, 255, 255, 20, 122, 2, 5, 79, 48, 255, 255, 22, 122, 2, 5, 81, 48, 0, 192, 197, 98, 147, 48, 255, 255, 102, 122, 2, 5, 148, 48, 255, 255, 10, 122, 4, 5, 149, 48, 255, 255, 18, 122, 4, 5, 142, 48, 255, 255, 94, 122, 4, 5, 143, 48, 255, 255, 94, 122, 4, 5, 144, 48, 255, 255, 96, 122, 4, 5, 145, 48, 255, 255, 98, 122, 4, 5, 146, 48, 255, 255, 100, 122, 4, 5, 136, 48, 26, 0, 139, 48, 12, 0, 139, 48, 255, 255, 88, 122, 2, 5, 140, 48, 255, 255, 90, 122, 2, 5, 141, 48, 255, 255, 92, 122, 2, 5, 136, 48, 255, 255, 82, 122, 4, 5, 137, 48, 255, 255, 84, 122, 2, 5, 138, 48, 255, 255, 86, 122, 2, 5, 131, 48, 255, 255, 78, 122, 4, 5, 132, 48, 255, 255, 78, 122, 4, 5, 133, 48, 255, 255, 80, 122, 4, 5, 134, 48, 255, 255, 80, 122, 4, 5, 135, 48, 255, 255, 82, 122, 4, 5, 120, 48, 48, 0, 125, 48, 26, 0, 128, 48, 12, 0, 128, 48, 255, 255, 72, 122, 2, 5, 129, 48, 255, 255, 74, 122, 4, 5, 130, 48, 255, 255, 76, 122, 2, 5, 125, 48, 255, 255, 66, 122, 2, 5, 126, 48, 255, 255, 68, 122, 4, 5, 127, 48, 255, 255, 70, 122, 2, 5, 120, 48, 255, 255, 64, 122, 2, 5, 121, 48, 255, 255, 64, 122, 2, 5, 122, 48, 255, 255, 64, 122, 2, 5, 123, 48, 255, 255, 66, 122, 2, 5, 124, 48, 255, 255, 66, 122, 2, 5, 114, 48, 26, 0, 117, 48, 12, 0, 117, 48, 255, 255, 62, 122, 2, 5, 118, 48, 255, 255, 62, 122, 2, 5, 119, 48, 255, 255, 62, 122, 2, 5, 114, 48, 255, 255, 60, 122, 2, 5, 115, 48, 255, 255, 60, 122, 2, 5, 116, 48, 255, 255, 60, 122, 2, 5, 109, 48, 255, 255, 54, 122, 2, 5, 110, 48, 255, 255, 56, 122, 2, 5, 111, 48, 255, 255, 58, 122, 2, 5, 112, 48, 255, 255, 58, 122, 2, 5, 113, 48, 255, 255, 58, 122, 2, 5, 87, 48, 98, 0, 98, 48, 48, 0, 103, 48, 26, 0, 106, 48, 12, 0, 106, 48, 255, 255, 48, 122, 2, 5, 107, 48, 255, 255, 50, 122, 2, 5, 108, 48, 255, 255, 52, 122, 2, 5, 103, 48, 255, 255, 44, 122, 2, 5, 104, 48, 255, 255, 46, 122, 2, 5, 105, 48, 255, 255, 46, 122, 2, 5, 98, 48, 255, 255, 40, 122, 2, 5, 99, 48, 255, 255, 42, 122, 4, 5, 100, 48, 255, 255, 42, 122, 4, 5, 101, 48, 255, 255, 42, 122, 4, 5, 102, 48, 255, 255, 44, 122, 2, 5, 92, 48, 26, 0, 95, 48, 12, 0, 95, 48, 255, 255, 38, 122, 2, 5, 96, 48, 255, 255, 38, 122, 2, 5, 97, 48, 255, 255, 40, 122, 2, 5, 92, 48, 255, 255, 34, 122, 2, 5, 93, 48, 255, 255, 36, 122, 2, 5, 94, 48, 255, 255, 36, 122, 2, 5, 87, 48, 255, 255, 30, 122, 4, 5, 88, 48, 255, 255, 30, 122, 4, 5, 89, 48, 255, 255, 32, 122, 4, 5, 90, 48, 255, 255, 32, 122, 4, 5, 91, 48, 255, 255, 34, 122, 2, 5, 76, 48, 45, 0, 81, 48, 25, 0, 84, 48, 12, 0, 84, 48, 255, 255, 26, 122, 2, 5, 85, 48, 255, 255, 28, 122, 2, 5, 86, 48, 255, 255, 28, 122, 2, 5, 81, 48, 255, 255, 24, 122, 4, 5, 82, 48, 0, 192, 197, 98, 83, 48, 255, 255, 26, 122, 2, 5, 76, 48, 197, 162, 77, 48, 255, 255, 20, 122, 2, 5, 78, 48, 255, 255, 20, 122, 2, 5, 79, 48, 255, 255, 22, 122, 2, 5, 80, 48, 255, 255, 22, 122, 2, 5, 70, 48, 26, 0, 73, 48, 12, 0, 73, 48, 255, 255, 16, 122, 4, 5, 74, 48, 255, 255, 16, 122, 4, 5, 75, 48, 255, 255, 18, 122, 4, 5, 70, 48, 255, 255, 10, 122, 4, 5, 71, 48, 255, 255, 14, 122, 4, 5, 72, 48, 255, 255, 14, 122, 4, 5, 65, 48, 255, 255, 6, 122, 4, 5, 66, 48, 255, 255, 6, 122, 4, 5, 67, 48, 255, 255, 8, 122, 4, 5, 68, 48, 255, 255, 8, 122, 4, 5, 69, 48, 255, 255, 10, 122, 4, 5, 0, 0, 197, 162, 0, 0, 87, 0, 109, 48, 239, 0, 131, 48, 161, 0, 142, 48, 122, 0, 147, 48, 105, 0, 150, 48, 94, 0, 150, 48, 3, 192, 197, 66, 153, 48, 17, 0, 154, 48, 4, 0, 111, 48, 7, 192, 197, 226, 114, 48, 8, 192, 197, 34, 117, 48, 8, 192, 197, 98, 120, 48, 8, 192, 197, 162, 123, 48, 8, 192, 197, 226, 20, 0, 93, 48, 37, 0, 104, 48, 20, 0, 117, 48, 9, 0, 117, 48, 8, 192, 197, 98, 120, 48, 8, 192, 197, 162, 123, 48, 8, 192, 197, 226, 104, 48, 6, 192, 197, 98, 111, 48, 7, 192, 197, 226, 114, 48, 8, 192, 197, 34, 93, 48, 5, 192, 197, 34, 95, 48, 5, 192, 197, 98, 97, 48, 5, 192, 197, 162, 100, 48, 5, 192, 197, 226, 102, 48, 6, 192, 197, 34, 83, 48, 15, 0, 83, 48, 3, 192, 197, 226, 85, 48, 4, 192, 197, 34, 87, 48, 4, 192, 197, 98, 89, 48, 4, 192, 197, 162, 91, 48, 4, 192, 197, 226, 70, 48, 1, 192, 197, 98, 75, 48, 2, 192, 197, 99, 77, 48, 2, 192, 197, 194, 79, 48, 3, 192, 197, 2, 81, 48, 3, 192, 197, 131, 147, 48, 13, 192, 197, 98, 148, 48, 1, 192, 197, 98, 149, 48, 2, 192, 197, 34, 142, 48, 12, 192, 197, 98, 143, 48, 12, 192, 197, 98, 144, 48, 12, 192, 197, 162, 145, 48, 12, 192, 197, 226, 146, 48, 13, 192, 197, 34, 136, 48, 20, 0, 139, 48, 9, 0, 139, 48, 11, 192, 197, 162, 140, 48, 11, 192, 197, 226, 141, 48, 12, 192, 197, 34, 136, 48, 10, 192, 197, 226, 137, 48, 11, 192, 197, 34, 138, 48, 11, 192, 197, 98, 131, 48, 10, 192, 197, 98, 132, 48, 10, 192, 197, 98, 133, 48, 10, 192, 197, 162, 134, 48, 10, 192, 197, 162, 135, 48, 10, 192, 197, 226, 120, 48, 37, 0, 125, 48, 20, 0, 128, 48, 9, 0, 128, 48, 9, 192, 197, 162, 129, 48, 9, 192, 197, 226, 130, 48, 10, 192, 197, 34, 125, 48, 8, 192, 197, 226, 126, 48, 9, 192, 197, 34, 127, 48, 9, 192, 197, 98, 120, 48, 8, 192, 197, 162, 121, 48, 8, 192, 197, 162, 122, 48, 8, 192, 197, 162, 123, 48, 8, 192, 197, 226, 124, 48, 8, 192, 197, 226, 114, 48, 20, 0, 117, 48, 9, 0, 117, 48, 8, 192, 197, 98, 118, 48, 8, 192, 197, 98, 119, 48, 8, 192, 197, 98, 114, 48, 8, 192, 197, 34, 115, 48, 8, 192, 197, 34, 116, 48, 8, 192, 197, 34, 109, 48, 7, 192, 197, 98, 110, 48, 7, 192, 197, 162, 111, 48, 7, 192, 197, 226, 112, 48, 7, 192, 197, 226, 113, 48, 7, 192, 197, 226, 87, 48, 76, 0, 98, 48, 37, 0, 103, 48, 20, 0, 106, 48, 9, 0, 106, 48, 6, 192, 197, 162, 107, 48, 6, 192, 197, 226, 108, 48, 7, 192, 197, 34, 103, 48, 6, 192, 197, 34, 104, 48, 6, 192, 197, 98, 105, 48, 6, 192, 197, 98, 98, 48, 5, 192, 197, 162, 99, 48, 5, 192, 197, 226, 100, 48, 5, 192, 197, 226, 101, 48, 5, 192, 197, 226, 102, 48, 6, 192, 197, 34, 92, 48, 20, 0, 95, 48, 9, 0, 95, 48, 5, 192, 197, 98, 96, 48, 5, 192, 197, 98, 97, 48, 5, 192, 197, 162, 92, 48, 4, 192, 197, 226, 93, 48, 5, 192, 197, 34, 94, 48, 5, 192, 197, 34, 87, 48, 4, 192, 197, 98, 88, 48, 4, 192, 197, 98, 89, 48, 4, 192, 197, 162, 90, 48, 4, 192, 197, 162, 91, 48, 4, 192, 197, 226, 76, 48, 37, 0, 81, 48, 20, 0, 84, 48, 9, 0, 84, 48, 3, 192, 197, 226, 85, 48, 4, 192, 197, 34, 86, 48, 4, 192, 197, 34, 81, 48, 3, 192, 197, 66, 82, 48, 3, 192, 197, 131, 83, 48, 3, 192, 197, 226, 76, 48, 2, 192, 197, 99, 77, 48, 2, 192, 197, 194, 78, 48, 2, 192, 197, 194, 79, 48, 3, 192, 197, 2, 80, 48, 3, 192, 197, 2, 70, 48, 20, 0, 73, 48, 9, 0, 73, 48, 1, 192, 197, 226, 74, 48, 1, 192, 197, 226, 75, 48, 2, 192, 197, 34, 70, 48, 1, 192, 197, 98, 71, 48, 1, 192, 197, 162, 72, 48, 1, 192, 197, 162, 65, 48, 0, 192, 197, 226, 66, 48, 0, 192, 197, 226, 67, 48, 1, 192, 197, 34, 68, 48, 1, 192, 197, 34, 69, 48, 1, 192, 197, 98, 118, 13, 5, 5, 0, 0, 245, 0, 197, 48, 175, 1, 248, 49, 216, 0, 126, 255, 107, 0, 141, 255, 54, 0, 149, 255, 26, 0, 153, 255, 12, 0, 153, 255, 15, 192, 198, 65, 154, 255, 15, 192, 198, 97, 155, 255, 15, 192, 198, 129, 156, 255, 15, 192, 198, 1, 149, 255, 15, 192, 198, 65, 150, 255, 15, 192, 198, 129, 151, 255, 15, 192, 198, 1, 152, 255, 15, 192, 198, 33, 145, 255, 12, 0, 145, 255, 15, 192, 198, 65, 146, 255, 15, 192, 198, 97, 147, 255, 15, 192, 198, 129, 148, 255, 15, 192, 198, 1, 141, 255, 15, 192, 198, 97, 142, 255, 15, 192, 198, 129, 143, 255, 15, 192, 198, 1, 144, 255, 15, 192, 198, 33, 133, 255, 26, 0, 137, 255, 12, 0, 137, 255, 15, 192, 198, 129, 138, 255, 15, 192, 198, 1, 139, 255, 15, 192, 198, 33, 140, 255, 15, 192, 198, 65, 133, 255, 15, 192, 198, 1, 134, 255, 15, 192, 198, 33, 135, 255, 15, 192, 198, 65, 136, 255, 15, 192, 198, 97, 129, 255, 12, 0, 129, 255, 15, 192, 198, 33, 130, 255, 15, 192, 198, 65, 131, 255, 15, 192, 198, 97, 132, 255, 15, 192, 198, 129, 126, 255, 15, 192, 198, 97, 127, 255, 15, 192, 198, 129, 128, 255, 15, 192, 198, 1, 109, 255, 54, 0, 118, 255, 26, 0, 122, 255, 12, 0, 122, 255, 15, 192, 198, 129, 123, 255, 15, 192, 198, 1, 124, 255, 15, 192, 198, 33, 125, 255, 15, 192, 198, 65, 118, 255, 15, 192, 198, 1, 119, 255, 15, 192, 198, 33, 120, 255, 15, 192, 198, 65, 121, 255, 15, 192, 198, 97, 114, 255, 12, 0, 114, 255, 15, 192, 198, 33, 115, 255, 15, 192, 198, 65, 116, 255, 15, 192, 198, 97, 117, 255, 15, 192, 198, 129, 109, 255, 15, 192, 198, 65, 110, 255, 15, 192, 198, 129, 111, 255, 15, 192, 198, 65, 113, 255, 15, 192, 198, 1, 255, 49, 26, 0, 105, 255, 12, 0, 105, 255, 15, 192, 198, 65, 106, 255, 15, 192, 198, 97, 107, 255, 15, 192, 198, 129, 108, 255, 15, 192, 198, 1, 255, 49, 15, 192, 198, 129, 102, 255, 15, 192, 198, 129, 103, 255, 15, 192, 198, 1, 104, 255, 15, 192, 198, 33, 251, 49, 12, 0, 251, 49, 15, 192, 198, 1, 252, 49, 15, 192, 198, 33, 253, 49, 15, 192, 198, 65, 254, 49, 15, 192, 198, 97, 248, 49, 15, 192, 198, 97, 249, 49, 15, 192, 198, 129, 250, 49, 15, 192, 198, 65, 227, 48, 107, 0, 242, 48, 54, 0, 240, 49, 26, 0, 244, 49, 12, 0, 244, 49, 15, 192, 198, 65, 245, 49, 15, 192, 198, 1, 246, 49, 15, 192, 198, 33, 247, 49, 15, 192, 198, 65, 240, 49, 15, 192, 198, 65, 241, 49, 15, 192, 198, 33, 242, 49, 15, 192, 198, 65, 243, 49, 15, 192, 198, 129, 247, 48, 12, 0, 247, 48, 15, 192, 198, 1, 248, 48, 15, 192, 198, 33, 249, 48, 15, 192, 198, 97, 250, 48, 15, 192, 198, 129, 242, 48, 15, 192, 198, 129, 244, 48, 15, 192, 198, 65, 245, 48, 15, 192, 198, 1, 246, 48, 15, 192, 198, 97, 234, 48, 26, 0, 238, 48, 12, 0, 238, 48, 15, 192, 198, 1, 239, 48, 15, 192, 198, 1, 240, 48, 15, 192, 198, 33, 241, 48, 15, 192, 198, 97, 234, 48, 15, 192, 198, 33, 235, 48, 15, 192, 198, 65, 236, 48, 15, 192, 198, 97, 237, 48, 15, 192, 198, 129, 230, 48, 12, 0, 230, 48, 15, 192, 198, 65, 231, 48, 15, 192, 198, 129, 232, 48, 15, 192, 198, 129, 233, 48, 15, 192, 198, 1, 227, 48, 15, 192, 198, 1, 228, 48, 15, 192, 198, 1, 229, 48, 15, 192, 198, 65, 212, 48, 51, 0, 219, 48, 26, 0, 223, 48, 12, 0, 223, 48, 15, 192, 198, 33, 224, 48, 15, 192, 198, 65, 225, 48, 15, 192, 198, 97, 226, 48, 15, 192, 198, 129, 219, 48, 15, 192, 198, 129, 220, 48, 15, 192, 198, 129, 221, 48, 15, 192, 198, 129, 222, 48, 15, 192, 198, 1, 215, 48, 12, 0, 215, 48, 15, 192, 198, 65, 216, 48, 15, 192, 198, 97, 217, 48, 15, 192, 198, 97, 218, 48, 15, 192, 198, 97, 212, 48, 15, 192, 198, 33, 213, 48, 15, 192, 198, 65, 214, 48, 15, 192, 198, 65, 204, 48, 26, 0, 208, 48, 12, 0, 208, 48, 15, 192, 198, 1, 209, 48, 15, 192, 198, 1, 210, 48, 15, 192, 198, 33, 211, 48, 15, 192, 198, 33, 204, 48, 15, 192, 198, 65, 205, 48, 15, 192, 198, 97, 206, 48, 15, 192, 198, 129, 207, 48, 15, 192, 198, 1, 200, 48, 12, 0, 200, 48, 15, 192, 198, 129, 201, 48, 15, 192, 198, 129, 202, 48, 15, 192, 198, 1, 203, 48, 15, 192, 198, 33, 197, 48, 15, 192, 198, 65, 198, 48, 15, 192, 198, 97, 199, 48, 15, 192, 198, 97, 126, 48, 206, 1, 166, 48, 107, 0, 181, 48, 54, 0, 189, 48, 26, 0, 193, 48, 12, 0, 193, 48, 15, 192, 198, 33, 194, 48, 15, 192, 198, 33, 195, 48, 15, 192, 198, 65, 196, 48, 15, 192, 198, 65, 189, 48, 15, 192, 198, 129, 190, 48, 15, 192, 198, 129, 191, 48, 15, 192, 198, 1, 192, 48, 15, 192, 198, 1, 185, 48, 12, 0, 185, 48, 15, 192, 198, 65, 186, 48, 15, 192, 198, 65, 187, 48, 15, 192, 198, 97, 188, 48, 15, 192, 198, 97, 181, 48, 15, 192, 198, 1, 182, 48, 15, 192, 198, 1, 183, 48, 15, 192, 198, 33, 184, 48, 15, 192, 198, 33, 173, 48, 26, 0, 177, 48, 12, 0, 177, 48, 15, 192, 198, 97, 178, 48, 15, 192, 198, 97, 179, 48, 15, 192, 198, 129, 180, 48, 15, 192, 198, 129, 173, 48, 15, 192, 198, 33, 174, 48, 15, 192, 198, 33, 175, 48, 15, 192, 198, 65, 176, 48, 15, 192, 198, 65, 169, 48, 12, 0, 169, 48, 15, 192, 198, 129, 170, 48, 15, 192, 198, 129, 171, 48, 15, 192, 198, 1, 172, 48, 15, 192, 198, 1, 166, 48, 15, 192, 198, 65, 167, 48, 15, 192, 198, 97, 168, 48, 15, 192, 198, 97, 141, 48, 29, 1, 150, 48, 249, 0, 162, 48, 12, 0, 162, 48, 15, 192, 198, 1, 163, 48, 15, 192, 198, 33, 164, 48, 15, 192, 198, 33, 165, 48, 15, 192, 198, 65, 150, 48, 255, 255, 14, 122, 2, 5, 153, 48, 5, 0, 154, 48, 189, 0, 161, 48, 15, 192, 198, 1, 45, 0, 173, 48, 81, 0, 196, 48, 42, 0, 216, 48, 20, 0, 240, 48, 9, 0, 240, 48, 15, 192, 198, 33, 241, 48, 15, 192, 198, 97, 242, 48, 15, 192, 198, 129, 216, 48, 15, 192, 198, 97, 219, 48, 15, 192, 198, 129, 239, 48, 15, 192, 198, 1, 207, 48, 9, 0, 207, 48, 15, 192, 198, 1, 210, 48, 15, 192, 198, 33, 213, 48, 15, 192, 198, 65, 196, 48, 15, 192, 198, 65, 198, 48, 15, 192, 198, 97, 200, 48, 15, 192, 198, 129, 183, 48, 20, 0, 189, 48, 9, 0, 189, 48, 15, 192, 198, 129, 191, 48, 15, 192, 198, 1, 193, 48, 15, 192, 198, 33, 183, 48, 15, 192, 198, 33, 185, 48, 15, 192, 198, 65, 187, 48, 15, 192, 198, 97, 173, 48, 15, 192, 198, 33, 175, 48, 15, 192, 198, 65, 177, 48, 15, 192, 198, 97, 179, 48, 15, 192, 198, 129, 181, 48, 15, 192, 198, 1, 95, 48, 52, 0, 114, 48, 24, 0, 123, 48, 10, 0, 123, 48, 255, 255, 16, 122, 2, 5, 166, 48, 15, 192, 198, 65, 171, 48, 15, 192, 198, 1, 114, 48, 255, 255, 8, 122, 2, 5, 117, 48, 255, 255, 10, 122, 2, 5, 120, 48, 255, 255, 14, 122, 2, 5, 102, 48, 12, 0, 102, 48, 255, 255, 14, 122, 2, 5, 104, 48, 255, 255, 16, 122, 2, 5, 111, 48, 255, 255, 6, 122, 2, 5, 95, 48, 255, 255, 6, 122, 2, 5, 97, 48, 255, 255, 8, 122, 2, 5, 100, 48, 255, 255, 10, 122, 2, 5, 83, 48, 26, 0, 89, 48, 12, 0, 89, 48, 255, 255, 10, 122, 2, 5, 91, 48, 255, 255, 14, 122, 2, 5, 93, 48, 255, 255, 16, 122, 2, 5, 83, 48, 255, 255, 16, 122, 2, 5, 85, 48, 255, 255, 6, 122, 2, 5, 87, 48, 255, 255, 8, 122, 2, 5, 70, 48, 255, 255, 10, 122, 2, 5, 75, 48, 255, 255, 6, 122, 2, 5, 77, 48, 255, 255, 8, 122, 2, 5, 79, 48, 255, 255, 10, 122, 2, 5, 81, 48, 255, 255, 14, 122, 2, 5, 9, 0, 207, 48, 15, 0, 207, 48, 15, 192, 198, 1, 210, 48, 15, 192, 198, 33, 213, 48, 15, 192, 198, 65, 216, 48, 15, 192, 198, 97, 219, 48, 15, 192, 198, 129, 111, 48, 255, 255, 6, 122, 2, 5, 114, 48, 255, 255, 8, 122, 2, 5, 117, 48, 255, 255, 10, 122, 2, 5, 120, 48, 255, 255, 14, 122, 2, 5, 123, 48, 255, 255, 16, 122, 2, 5, 145, 48, 16, 0, 145, 48, 255, 255, 14, 122, 2, 5, 146, 48, 255, 255, 16, 122, 2, 5, 148, 48, 255, 255, 10, 122, 2, 5, 149, 48, 255, 255, 6, 122, 2, 5, 141, 48, 255, 255, 16, 122, 2, 5, 142, 48, 255, 255, 6, 122, 2, 5, 143, 48, 255, 255, 6, 122, 2, 5, 144, 48, 255, 255, 8, 122, 2, 5, 133, 48, 34, 0, 137, 48, 16, 0, 137, 48, 255, 255, 6, 122, 2, 5, 138, 48, 255, 255, 8, 122, 2, 5, 139, 48, 255, 255, 10, 122, 2, 5, 140, 48, 255, 255, 14, 122, 2, 5, 133, 48, 255, 255, 10, 122, 2, 5, 134, 48, 255, 255, 10, 122, 2, 5, 135, 48, 255, 255, 16, 122, 2, 5, 136, 48, 255, 255, 16, 122, 2, 5, 129, 48, 16, 0, 129, 48, 255, 255, 14, 122, 2, 5, 130, 48, 255, 255, 16, 122, 2, 5, 131, 48, 255, 255, 6, 122, 2, 5, 132, 48, 255, 255, 6, 122, 2, 5, 126, 48, 255, 255, 6, 122, 2, 5, 127, 48, 255, 255, 8, 122, 2, 5, 128, 48, 255, 255, 10, 122, 2, 5, 95, 48, 138, 0, 110, 48, 70, 0, 118, 48, 34, 0, 122, 48, 16, 0, 122, 48, 255, 255, 14, 122, 2, 5, 123, 48, 255, 255, 16, 122, 2, 5, 124, 48, 255, 255, 16, 122, 2, 5, 125, 48, 255, 255, 16, 122, 2, 5, 118, 48, 255, 255, 10, 122, 2, 5, 119, 48, 255, 255, 10, 122, 2, 5, 120, 48, 255, 255, 14, 122, 2, 5, 121, 48, 255, 255, 14, 122, 2, 5, 114, 48, 16, 0, 114, 48, 255, 255, 8, 122, 2, 5, 115, 48, 255, 255, 8, 122, 2, 5, 116, 48, 255, 255, 8, 122, 2, 5, 117, 48, 255, 255, 10, 122, 2, 5, 110, 48, 255, 255, 16, 122, 2, 5, 111, 48, 255, 255, 6, 122, 2, 5, 112, 48, 255, 255, 6, 122, 2, 5, 113, 48, 255, 255, 6, 122, 2, 5, 102, 48, 34, 0, 106, 48, 16, 0, 106, 48, 255, 255, 6, 122, 2, 5, 107, 48, 255, 255, 8, 122, 2, 5, 108, 48, 255, 255, 10, 122, 2, 5, 109, 48, 255, 255, 14, 122, 2, 5, 102, 48, 255, 255, 14, 122, 2, 5, 103, 48, 255, 255, 14, 122, 2, 5, 104, 48, 255, 255, 16, 122, 2, 5, 105, 48, 255, 255, 16, 122, 2, 5, 98, 48, 16, 0, 98, 48, 255, 255, 8, 122, 2, 5, 99, 48, 255, 255, 10, 122, 2, 5, 100, 48, 255, 255, 10, 122, 2, 5, 101, 48, 255, 255, 10, 122, 2, 5, 95, 48, 255, 255, 6, 122, 2, 5, 96, 48, 255, 255, 6, 122, 2, 5, 97, 48, 255, 255, 8, 122, 2, 5, 80, 48, 66, 0, 87, 48, 34, 0, 91, 48, 16, 0, 91, 48, 255, 255, 14, 122, 2, 5, 92, 48, 255, 255, 14, 122, 2, 5, 93, 48, 255, 255, 16, 122, 2, 5, 94, 48, 255, 255, 16, 122, 2, 5, 87, 48, 255, 255, 8, 122, 2, 5, 88, 48, 255, 255, 8, 122, 2, 5, 89, 48, 255, 255, 10, 122, 2, 5, 90, 48, 255, 255, 10, 122, 2, 5, 83, 48, 16, 0, 83, 48, 255, 255, 16, 122, 2, 5, 84, 48, 255, 255, 16, 122, 2, 5, 85, 48, 255, 255, 6, 122, 2, 5, 86, 48, 255, 255, 6, 122, 2, 5, 80, 48, 255, 255, 10, 122, 2, 5, 81, 48, 255, 255, 14, 122, 2, 5, 82, 48, 255, 255, 14, 122, 2, 5, 72, 48, 34, 0, 76, 48, 16, 0, 76, 48, 255, 255, 6, 122, 2, 5, 77, 48, 255, 255, 8, 122, 2, 5, 78, 48, 255, 255, 8, 122, 2, 5, 79, 48, 255, 255, 10, 122, 2, 5, 72, 48, 255, 255, 14, 122, 2, 5, 73, 48, 255, 255, 16, 122, 2, 5, 74, 48, 255, 255, 16, 122, 2, 5, 75, 48, 255, 255, 6, 122, 2, 5, 68, 48, 16, 0, 68, 48, 255, 255, 8, 122, 2, 5, 69, 48, 255, 255, 10, 122, 2, 5, 70, 48, 255, 255, 10, 122, 2, 5, 71, 48, 255, 255, 14, 122, 2, 5, 65, 48, 255, 255, 6, 122, 2, 5, 66, 48, 255, 255, 6, 122, 2, 5, 67, 48, 255, 255, 8, 122, 2, 5, 15, 0, 198, 161, 0, 0, 162, 0, 240, 48, 24, 1, 117, 255, 139, 0, 137, 255, 71, 0, 147, 255, 37, 0, 152, 255, 20, 0, 155, 255, 9, 0, 155, 255, 21, 192, 198, 129, 156, 255, 21, 192, 198, 161, 157, 255, 22, 192, 198, 33, 152, 255, 21, 192, 198, 33, 153, 255, 21, 192, 198, 65, 154, 255, 21, 192, 198, 97, 147, 255, 20, 192, 198, 129, 148, 255, 20, 192, 198, 161, 149, 255, 20, 192, 198, 193, 150, 255, 20, 192, 198, 225, 151, 255, 21, 192, 198, 1, 142, 255, 15, 0, 142, 255, 19, 192, 198, 225, 143, 255, 20, 192, 198, 1, 144, 255, 20, 192, 198, 33, 145, 255, 20, 192, 198, 65, 146, 255, 20, 192, 198, 97, 137, 255, 19, 192, 198, 65, 138, 255, 19, 192, 198, 97, 139, 255, 19, 192, 198, 129, 140, 255, 19, 192, 198, 161, 141, 255, 19, 192, 198, 193, 127, 255, 32, 0, 132, 255, 15, 0, 132, 255, 18, 192, 198, 161, 133, 255, 18, 192, 198, 193, 134, 255, 18, 192, 198, 225, 135, 255, 19, 192, 198, 1, 136, 255, 19, 192, 198, 33, 127, 255, 18, 192, 198, 1, 128, 255, 18, 192, 198, 33, 129, 255, 18, 192, 198, 65, 130, 255, 18, 192, 198, 97, 131, 255, 18, 192, 198, 129, 122, 255, 15, 0, 122, 255, 17, 192, 198, 97, 123, 255, 17, 192, 198, 129, 124, 255, 17, 192, 198, 161, 125, 255, 17, 192, 198, 193, 126, 255, 17, 192, 198, 225, 117, 255, 16, 192, 198, 65, 118, 255, 16, 192, 198, 97, 119, 255, 16, 192, 198, 193, 120, 255, 16, 192, 198, 225, 121, 255, 17, 192, 198, 1, 249, 49, 71, 0, 105, 255, 37, 0, 110, 255, 20, 0, 114, 255, 9, 0, 114, 255, 15, 192, 198, 225, 115, 255, 16, 192, 198, 1, 116, 255, 16, 192, 198, 33, 110, 255, 20, 192, 198, 225, 111, 255, 18, 192, 198, 97, 113, 255, 15, 192, 198, 193, 105, 255, 16, 192, 198, 1, 106, 255, 16, 192, 198, 33, 107, 255, 16, 192, 198, 65, 108, 255, 20, 192, 198, 161, 109, 255, 20, 192, 198, 193, 254, 49, 15, 0, 254, 49, 21, 192, 198, 97, 255, 49, 21, 192, 198, 129, 102, 255, 22, 192, 198, 1, 103, 255, 15, 192, 198, 193, 104, 255, 15, 192, 198, 225, 249, 49, 19, 192, 198, 225, 250, 49, 20, 192, 198, 65, 251, 49, 21, 192, 198, 1, 252, 49, 21, 192, 198, 33, 253, 49, 21, 192, 198, 65, 250, 48, 32, 0, 244, 49, 15, 0, 244, 49, 19, 192, 198, 1, 245, 49, 19, 192, 198, 97, 246, 49, 19, 192, 198, 129, 247, 49, 19, 192, 198, 161, 248, 49, 19, 192, 198, 193, 250, 48, 22, 192, 198, 1, 240, 49, 16, 192, 198, 225, 241, 49, 17, 192, 198, 161, 242, 49, 17, 192, 198, 193, 243, 49, 18, 192, 198, 161, 245, 48, 15, 0, 245, 48, 16, 192, 198, 97, 246, 48, 17, 192, 198, 1, 247, 48, 21, 192, 198, 161, 248, 48, 21, 192, 198, 193, 249, 48, 21, 192, 198, 225, 240, 48, 21, 192, 198, 193, 241, 48, 21, 192, 198, 225, 242, 48, 22, 192, 198, 1, 243, 48, 22, 192, 198, 33, 244, 48, 16, 192, 198, 1, 199, 48, 139, 0, 219, 48, 71, 0, 229, 48, 37, 0, 234, 48, 20, 0, 237, 48, 9, 0, 237, 48, 21, 192, 198, 129, 238, 48, 21, 192, 198, 161, 239, 48, 21, 192, 198, 161, 234, 48, 21, 192, 198, 33, 235, 48, 21, 192, 198, 65, 236, 48, 21, 192, 198, 97, 229, 48, 20, 192, 198, 193, 230, 48, 20, 192, 198, 193, 231, 48, 20, 192, 198, 225, 232, 48, 20, 192, 198, 225, 233, 48, 21, 192, 198, 1, 224, 48, 15, 0, 224, 48, 20, 192, 198, 65, 225, 48, 20, 192, 198, 97, 226, 48, 20, 192, 198, 129, 227, 48, 20, 192, 198, 161, 228, 48, 20, 192, 198, 161, 219, 48, 19, 192, 198, 225, 220, 48, 19, 192, 198, 225, 221, 48, 19, 192, 198, 225, 222, 48, 20, 192, 198, 1, 223, 48, 20, 192, 198, 33, 209, 48, 32, 0, 214, 48, 15, 0, 214, 48, 19, 192, 198, 161, 215, 48, 19, 192, 198, 161, 216, 48, 19, 192, 198, 193, 217, 48, 19, 192, 198, 193, 218, 48, 19, 192, 198, 193, 209, 48, 19, 192, 198, 97, 210, 48, 19, 192, 198, 129, 211, 48, 19, 192, 198, 129, 212, 48, 19, 192, 198, 129, 213, 48, 19, 192, 198, 161, 204, 48, 15, 0, 204, 48, 19, 192, 198, 1, 205, 48, 19, 192, 198, 33, 206, 48, 19, 192, 198, 65, 207, 48, 19, 192, 198, 97, 208, 48, 19, 192, 198, 97, 199, 48, 18, 192, 198, 129, 200, 48, 18, 192, 198, 161, 201, 48, 18, 192, 198, 161, 202, 48, 18, 192, 198, 193, 203, 48, 18, 192, 198, 225, 179, 48, 66, 0, 189, 48, 32, 0, 194, 48, 15, 0, 194, 48, 18, 192, 198, 65, 195, 48, 18, 192, 198, 97, 196, 48, 18, 192, 198, 97, 197, 48, 18, 192, 198, 97, 198, 48, 18, 192, 198, 129, 189, 48, 18, 192, 198, 1, 190, 48, 18, 192, 198, 1, 191, 48, 18, 192, 198, 33, 192, 48, 18, 192, 198, 33, 193, 48, 18, 192, 198, 65, 184, 48, 15, 0, 184, 48, 17, 192, 198, 161, 185, 48, 17, 192, 198, 193, 186, 48, 17, 192, 198, 193, 187, 48, 17, 192, 198, 225, 188, 48, 17, 192, 198, 225, 179, 48, 17, 192, 198, 97, 180, 48, 17, 192, 198, 97, 181, 48, 17, 192, 198, 129, 182, 48, 17, 192, 198, 129, 183, 48, 17, 192, 198, 161, 169, 48, 32, 0, 174, 48, 15, 0, 174, 48, 16, 192, 198, 193, 175, 48, 16, 192, 198, 225, 176, 48, 16, 192, 198, 225, 177, 48, 17, 192, 198, 1, 178, 48, 17, 192, 198, 34, 169, 48, 16, 192, 198, 65, 170, 48, 16, 192, 198, 65, 171, 48, 16, 192, 198, 97, 172, 48, 16, 192, 198, 130, 173, 48, 16, 192, 198, 193, 164, 48, 15, 0, 164, 48, 15, 192, 198, 225, 165, 48, 16, 192, 198, 1, 166, 48, 16, 192, 198, 1, 167, 48, 16, 192, 198, 33, 168, 48, 16, 192, 198, 33, 153, 48, 11, 0, 154, 48, 99, 0, 161, 48, 15, 192, 198, 193, 162, 48, 15, 192, 198, 193, 163, 48, 15, 192, 198, 225, 24, 0, 193, 48, 45, 0, 213, 48, 23, 0, 239, 48, 12, 0, 239, 48, 21, 192, 198, 161, 240, 48, 21, 192, 198, 193, 241, 48, 21, 192, 198, 225, 242, 48, 22, 192, 198, 1, 213, 48, 19, 192, 198, 161, 216, 48, 19, 192, 198, 193, 219, 48, 19, 192, 198, 225, 200, 48, 9, 0, 200, 48, 18, 192, 198, 161, 207, 48, 19, 192, 198, 97, 210, 48, 19, 192, 198, 129, 193, 48, 18, 192, 198, 65, 196, 48, 18, 192, 198, 97, 198, 48, 18, 192, 198, 129, 181, 48, 20, 0, 187, 48, 9, 0, 187, 48, 17, 192, 198, 225, 189, 48, 18, 192, 198, 1, 191, 48, 18, 192, 198, 33, 181, 48, 17, 192, 198, 129, 183, 48, 17, 192, 198, 161, 185, 48, 17, 192, 198, 193, 175, 48, 9, 0, 175, 48, 16, 192, 198, 225, 177, 48, 17, 192, 198, 34, 179, 48, 17, 192, 198, 97, 166, 48, 16, 192, 198, 1, 171, 48, 16, 192, 198, 130, 173, 48, 16, 192, 198, 193, 4, 0, 207, 48, 19, 192, 198, 97, 210, 48, 19, 192, 198, 129, 213, 48, 19, 192, 198, 161, 216, 48, 19, 192, 198, 193, 219, 48, 19, 192, 198, 225, 22, 0, 198, 66, 0, 0, 162, 0, 240, 48, 24, 1, 117, 255, 139, 0, 137, 255, 71, 0, 147, 255, 37, 0, 152, 255, 20, 0, 155, 255, 9, 0, 155, 255, 33, 192, 198, 194, 156, 255, 34, 192, 198, 2, 157, 255, 35, 192, 198, 2, 152, 255, 33, 192, 198, 2, 153, 255, 33, 192, 198, 66, 154, 255, 33, 192, 198, 130, 147, 255, 31, 192, 198, 194, 148, 255, 32, 192, 198, 2, 149, 255, 32, 192, 198, 66, 150, 255, 32, 192, 198, 130, 151, 255, 32, 192, 198, 194, 142, 255, 15, 0, 142, 255, 30, 192, 198, 130, 143, 255, 30, 192, 198, 194, 144, 255, 31, 192, 198, 2, 145, 255, 31, 192, 198, 66, 146, 255, 31, 192, 198, 130, 137, 255, 29, 192, 198, 66, 138, 255, 29, 192, 198, 130, 139, 255, 29, 192, 198, 194, 140, 255, 30, 192, 198, 2, 141, 255, 30, 192, 198, 66, 127, 255, 32, 0, 132, 255, 15, 0, 132, 255, 28, 192, 198, 2, 133, 255, 28, 192, 198, 66, 134, 255, 28, 192, 198, 130, 135, 255, 28, 192, 198, 194, 136, 255, 29, 192, 198, 2, 127, 255, 26, 192, 198, 194, 128, 255, 27, 192, 198, 2, 129, 255, 27, 192, 198, 66, 130, 255, 27, 192, 198, 130, 131, 255, 27, 192, 198, 194, 122, 255, 15, 0, 122, 255, 25, 192, 198, 130, 123, 255, 25, 192, 198, 194, 124, 255, 26, 192, 198, 2, 125, 255, 26, 192, 198, 66, 126, 255, 26, 192, 198, 130, 117, 255, 23, 192, 198, 130, 118, 255, 23, 192, 198, 194, 119, 255, 24, 192, 198, 98, 120, 255, 24, 192, 198, 162, 121, 255, 24, 192, 198, 226, 249, 49, 71, 0, 105, 255, 37, 0, 110, 255, 20, 0, 114, 255, 9, 0, 114, 255, 22, 192, 198, 194, 115, 255, 23, 192, 198, 2, 116, 255, 23, 192, 198, 66, 110, 255, 32, 192, 198, 130, 111, 255, 27, 192, 198, 130, 113, 255, 22, 192, 198, 130, 105, 255, 23, 192, 198, 2, 106, 255, 23, 192, 198, 66, 107, 255, 23, 192, 198, 130, 108, 255, 32, 192, 198, 2, 109, 255, 32, 192, 198, 66, 254, 49, 15, 0, 254, 49, 33, 192, 198, 130, 255, 49, 33, 192, 198, 194, 102, 255, 34, 192, 198, 194, 103, 255, 22, 192, 198, 130, 104, 255, 22, 192, 198, 194, 249, 49, 30, 192, 198, 130, 250, 49, 31, 192, 198, 66, 251, 49, 32, 192, 198, 194, 252, 49, 33, 192, 198, 2, 253, 49, 33, 192, 198, 66, 250, 48, 32, 0, 244, 49, 15, 0, 244, 49, 28, 192, 198, 194, 245, 49, 29, 192, 198, 130, 246, 49, 29, 192, 198, 194, 247, 49, 30, 192, 198, 2, 248, 49, 30, 192, 198, 66, 250, 48, 34, 192, 198, 194, 240, 49, 24, 192, 198, 162, 241, 49, 26, 192, 198, 2, 242, 49, 26, 192, 198, 66, 243, 49, 28, 192, 198, 2, 245, 48, 15, 0, 245, 48, 23, 192, 198, 194, 246, 48, 24, 192, 198, 226, 247, 48, 34, 192, 198, 2, 248, 48, 34, 192, 198, 66, 249, 48, 34, 192, 198, 130, 240, 48, 34, 192, 198, 66, 241, 48, 34, 192, 198, 130, 242, 48, 34, 192, 198, 194, 243, 48, 35, 192, 198, 2, 244, 48, 23, 192, 198, 2, 199, 48, 139, 0, 219, 48, 71, 0, 229, 48, 37, 0, 234, 48, 20, 0, 237, 48, 9, 0, 237, 48, 33, 192, 198, 194, 238, 48, 34, 192, 198, 2, 239, 48, 34, 192, 198, 2, 234, 48, 33, 192, 198, 2, 235, 48, 33, 192, 198, 66, 236, 48, 33, 192, 198, 130, 229, 48, 32, 192, 198, 66, 230, 48, 32, 192, 198, 66, 231, 48, 32, 192, 198, 130, 232, 48, 32, 192, 198, 130, 233, 48, 32, 192, 198, 194, 224, 48, 15, 0, 224, 48, 31, 192, 198, 66, 225, 48, 31, 192, 198, 130, 226, 48, 31, 192, 198, 194, 227, 48, 32, 192, 198, 2, 228, 48, 32, 192, 198, 2, 219, 48, 30, 192, 198, 130, 220, 48, 30, 192, 198, 130, 221, 48, 30, 192, 198, 130, 222, 48, 30, 192, 198, 194, 223, 48, 31, 192, 198, 2, 209, 48, 32, 0, 214, 48, 15, 0, 214, 48, 30, 192, 198, 2, 215, 48, 30, 192, 198, 2, 216, 48, 30, 192, 198, 66, 217, 48, 30, 192, 198, 66, 218, 48, 30, 192, 198, 66, 209, 48, 29, 192, 198, 130, 210, 48, 29, 192, 198, 194, 211, 48, 29, 192, 198, 194, 212, 48, 29, 192, 198, 194, 213, 48, 30, 192, 198, 2, 204, 48, 15, 0, 204, 48, 28, 192, 198, 194, 205, 48, 29, 192, 198, 2, 206, 48, 29, 192, 198, 66, 207, 48, 29, 192, 198, 130, 208, 48, 29, 192, 198, 130, 199, 48, 27, 192, 198, 194, 200, 48, 28, 192, 198, 2, 201, 48, 28, 192, 198, 2, 202, 48, 28, 192, 198, 66, 203, 48, 28, 192, 198, 130, 179, 48, 66, 0, 189, 48, 32, 0, 194, 48, 15, 0, 194, 48, 27, 192, 198, 66, 195, 48, 27, 192, 198, 130, 196, 48, 27, 192, 198, 130, 197, 48, 27, 192, 198, 130, 198, 48, 27, 192, 198, 194, 189, 48, 26, 192, 198, 194, 190, 48, 26, 192, 198, 194, 191, 48, 27, 192, 198, 2, 192, 48, 27, 192, 198, 2, 193, 48, 27, 192, 198, 66, 184, 48, 15, 0, 184, 48, 26, 192, 198, 2, 185, 48, 26, 192, 198, 66, 186, 48, 26, 192, 198, 66, 187, 48, 26, 192, 198, 130, 188, 48, 26, 192, 198, 130, 179, 48, 25, 192, 198, 130, 180, 48, 25, 192, 198, 130, 181, 48, 25, 192, 198, 194, 182, 48, 25, 192, 198, 194, 183, 48, 26, 192, 198, 2, 169, 48, 32, 0, 174, 48, 15, 0, 174, 48, 24, 192, 198, 98, 175, 48, 24, 192, 198, 162, 176, 48, 24, 192, 198, 162, 177, 48, 24, 192, 198, 226, 178, 48, 25, 192, 198, 35, 169, 48, 23, 192, 198, 130, 170, 48, 23, 192, 198, 130, 171, 48, 23, 192, 198, 194, 172, 48, 24, 192, 198, 3, 173, 48, 24, 192, 198, 98, 164, 48, 15, 0, 164, 48, 22, 192, 198, 194, 165, 48, 23, 192, 198, 2, 166, 48, 23, 192, 198, 2, 167, 48, 23, 192, 198, 66, 168, 48, 23, 192, 198, 66, 153, 48, 11, 0, 154, 48, 99, 0, 161, 48, 22, 192, 198, 130, 162, 48, 22, 192, 198, 130, 163, 48, 22, 192, 198, 194, 24, 0, 193, 48, 45, 0, 213, 48, 23, 0, 239, 48, 12, 0, 239, 48, 34, 192, 198, 2, 240, 48, 34, 192, 198, 66, 241, 48, 34, 192, 198, 130, 242, 48, 34, 192, 198, 194, 213, 48, 30, 192, 198, 2, 216, 48, 30, 192, 198, 66, 219, 48, 30, 192, 198, 130, 200, 48, 9, 0, 200, 48, 28, 192, 198, 2, 207, 48, 29, 192, 198, 130, 210, 48, 29, 192, 198, 194, 193, 48, 27, 192, 198, 66, 196, 48, 27, 192, 198, 130, 198, 48, 27, 192, 198, 194, 181, 48, 20, 0, 187, 48, 9, 0, 187, 48, 26, 192, 198, 130, 189, 48, 26, 192, 198, 194, 191, 48, 27, 192, 198, 2, 181, 48, 25, 192, 198, 194, 183, 48, 26, 192, 198, 2, 185, 48, 26, 192, 198, 66, 175, 48, 9, 0, 175, 48, 24, 192, 198, 162, 177, 48, 25, 192, 198, 35, 179, 48, 25, 192, 198, 130, 166, 48, 23, 192, 198, 2, 171, 48, 24, 192, 198, 3, 173, 48, 24, 192, 198, 98, 4, 0, 207, 48, 29, 192, 198, 130, 210, 48, 29, 192, 198, 194, 213, 48, 30, 192, 198, 2, 216, 48, 30, 192, 198, 66, 219, 48, 30, 192, 198, 130, 170, 170, 3, 0, 82, 0, 64, 0, 32, 0, 219, 22, 0, 16, 111, 20, 0, 96, 1, 0, 0, 96, 252, 82, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 16, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 76, 0, 0, 0, 76, 1, 0, 0, 188, 31, 0, 0, 188, 31, 0, 0, 172, 40, 0, 0, 172, 40, 0, 0, 168, 42, 0, 0, 168, 42, 0, 0, 168, 78, 0, 0, 224, 82, 0, 0, 25, 0, 0, 0, 22, 0, 0, 0, 17, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 94, 249, 250, 251, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 222, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 48, 0, 48, 0, 48, 0, 97, 0, 105, 0, 113, 0, 121, 0, 48, 0, 129, 0, 137, 0, 145, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 149, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 157, 0, 16, 0, 165, 0, 173, 0, 181, 0, 189, 0, 197, 0, 205, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 14, 0, 202, 193, 14, 0, 202, 226, 14, 0, 202, 3, 15, 0, 202, 36, 15, 0, 202, 69, 15, 0, 202, 102, 15, 0, 202, 135, 15, 0, 202, 168, 15, 0, 202, 201, 15, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 13, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 14, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 14, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 200, 0, 0, 0, 200, 32, 63, 0, 197, 162, 13, 0, 192, 0, 0, 0, 198, 1, 0, 0, 198, 33, 0, 0, 198, 65, 0, 0, 198, 97, 0, 0, 198, 129, 0, 0, 198, 161, 0, 0, 198, 193, 0, 0, 198, 225, 0, 0, 198, 1, 1, 0, 198, 33, 1, 0, 198, 65, 1, 0, 198, 98, 1, 0, 198, 161, 1, 0, 198, 194, 1, 0, 198, 1, 2, 0, 198, 34, 2, 0, 198, 97, 2, 0, 198, 130, 2, 0, 198, 193, 2, 0, 198, 226, 2, 0, 198, 33, 3, 0, 198, 66, 3, 0, 198, 129, 3, 0, 198, 162, 3, 0, 198, 225, 3, 0, 198, 2, 4, 0, 198, 65, 4, 0, 198, 98, 4, 0, 198, 161, 4, 0, 198, 194, 4, 0, 198, 1, 5, 0, 198, 34, 5, 0, 198, 97, 5, 0, 198, 130, 5, 0, 198, 193, 5, 0, 198, 225, 5, 0, 198, 2, 6, 0, 198, 65, 6, 0, 198, 98, 6, 0, 198, 161, 6, 0, 198, 194, 6, 0, 198, 1, 7, 0, 198, 33, 7, 0, 198, 65, 7, 0, 198, 97, 7, 0, 198, 129, 7, 0, 198, 161, 7, 0, 198, 194, 7, 0, 198, 2, 8, 0, 198, 65, 8, 0, 198, 98, 8, 0, 198, 162, 8, 0, 198, 225, 8, 0, 198, 2, 9, 0, 198, 66, 9, 0, 198, 129, 9, 0, 198, 162, 9, 0, 198, 226, 9, 0, 198, 33, 10, 0, 198, 66, 10, 0, 198, 130, 10, 0, 198, 193, 10, 0, 198, 225, 10, 0, 198, 1, 11, 0, 198, 33, 11, 0, 198, 65, 11, 0, 198, 97, 11, 0, 198, 129, 11, 0, 198, 161, 11, 0, 198, 193, 11, 0, 198, 225, 11, 0, 198, 1, 12, 0, 198, 33, 12, 0, 198, 65, 12, 0, 198, 97, 12, 0, 198, 129, 12, 0, 198, 161, 12, 0, 198, 193, 12, 0, 198, 225, 12, 0, 198, 1, 13, 0, 198, 33, 13, 0, 198, 65, 13, 0, 198, 97, 13, 0, 198, 130, 13, 0, 198, 193, 13, 0, 198, 225, 13, 0, 198, 2, 14, 0, 198, 66, 14, 0, 198, 130, 14, 0, 198, 194, 14, 0, 192, 0, 0, 0, 200, 0, 113, 0, 200, 224, 3, 1, 200, 0, 87, 1, 198, 66, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 5, 5, 8, 121, 5, 5, 183, 121, 5, 5, 10, 121, 5, 5, 187, 121, 5, 5, 189, 121, 5, 5, 12, 121, 5, 5, 14, 121, 5, 5, 16, 121, 5, 5, 195, 121, 5, 5, 197, 121, 5, 5, 199, 121, 5, 5, 201, 121, 5, 5, 203, 121, 5, 5, 205, 121, 5, 5, 52, 121, 5, 5, 18, 121, 5, 5, 20, 121, 5, 5, 22, 121, 5, 5, 66, 121, 5, 5, 24, 121, 5, 5, 26, 121, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 32, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 5, 105, 121, 5, 5, 107, 121, 5, 5, 109, 121, 5, 5, 111, 121, 5, 5, 113, 121, 5, 5, 115, 121, 5, 5, 117, 121, 5, 5, 119, 121, 5, 5, 121, 121, 5, 5, 123, 121, 5, 5, 125, 121, 5, 5, 127, 121, 5, 5, 129, 121, 5, 5, 131, 121, 5, 5, 133, 121, 5, 5, 135, 121, 5, 5, 137, 121, 5, 5, 139, 121, 5, 5, 141, 121, 5, 5, 143, 121, 193, 247, 101, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 32, 170, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 14, 0, 202, 193, 14, 0, 202, 226, 14, 0, 202, 3, 15, 0, 202, 36, 15, 0, 202, 69, 15, 0, 202, 102, 15, 0, 202, 135, 15, 0, 202, 168, 15, 0, 202, 201, 15, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 192, 0, 0, 0, 192, 0, 0, 0, 133, 5, 24, 9, 133, 5, 70, 10, 133, 5, 72, 10, 133, 5, 34, 7, 192, 0, 0, 0, 198, 65, 13, 0, 198, 1, 0, 0, 198, 65, 0, 0, 198, 129, 0, 0, 198, 193, 0, 0, 198, 1, 1, 0, 198, 97, 11, 0, 198, 161, 11, 0, 198, 225, 11, 0, 198, 193, 5, 0, 192, 0, 0, 0, 198, 33, 0, 0, 198, 97, 0, 0, 198, 161, 0, 0, 198, 225, 0, 0, 198, 33, 1, 0, 198, 65, 1, 0, 198, 161, 1, 0, 198, 1, 2, 0, 198, 97, 2, 0, 198, 193, 2, 0, 198, 33, 3, 0, 198, 129, 3, 0, 198, 225, 3, 0, 198, 65, 4, 0, 198, 161, 4, 0, 198, 1, 5, 0, 198, 97, 5, 0, 198, 225, 5, 0, 198, 65, 6, 0, 198, 161, 6, 0, 198, 1, 7, 0, 198, 33, 7, 0, 198, 65, 7, 0, 198, 97, 7, 0, 198, 129, 7, 0, 198, 161, 7, 0, 198, 65, 8, 0, 198, 225, 8, 0, 198, 129, 9, 0, 198, 33, 10, 0, 198, 193, 10, 0, 198, 225, 10, 0, 198, 1, 11, 0, 198, 33, 11, 0, 198, 65, 11, 0, 198, 129, 11, 0, 198, 193, 11, 0, 198, 1, 12, 0, 198, 33, 12, 0, 198, 65, 12, 0, 198, 97, 12, 0, 198, 129, 12, 0, 198, 161, 12, 0, 198, 225, 12, 0, 198, 97, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 35, 0, 133, 5, 6, 121, 133, 5, 8, 121, 133, 5, 183, 121, 133, 5, 10, 121, 133, 5, 187, 121, 133, 5, 189, 121, 133, 5, 12, 121, 133, 5, 14, 121, 133, 5, 16, 121, 133, 5, 195, 121, 133, 5, 197, 121, 133, 5, 199, 121, 133, 5, 201, 121, 133, 5, 203, 121, 133, 5, 205, 121, 133, 5, 52, 121, 133, 5, 18, 121, 133, 5, 20, 121, 133, 5, 22, 121, 133, 5, 66, 121, 133, 5, 24, 121, 133, 5, 26, 121, 133, 5, 28, 121, 133, 5, 30, 121, 133, 5, 32, 121, 133, 5, 34, 121, 133, 5, 36, 121, 133, 5, 38, 121, 133, 5, 40, 121, 133, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 133, 5, 103, 121, 133, 5, 105, 121, 133, 5, 107, 121, 133, 5, 109, 121, 133, 5, 111, 121, 133, 5, 113, 121, 192, 0, 0, 0, 192, 0, 0, 0, 133, 5, 115, 121, 133, 5, 117, 121, 133, 5, 119, 121, 133, 5, 121, 121, 133, 5, 123, 121, 133, 5, 125, 121, 192, 0, 0, 0, 192, 0, 0, 0, 133, 5, 127, 121, 133, 5, 129, 121, 133, 5, 131, 121, 133, 5, 133, 121, 133, 5, 135, 121, 133, 5, 137, 121, 192, 0, 0, 0, 192, 0, 0, 0, 133, 5, 139, 121, 133, 5, 141, 121, 133, 5, 143, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 13, 5, 5, 140, 13, 5, 5, 135, 12, 5, 5, 0, 4, 5, 5, 139, 12, 5, 5, 142, 13, 5, 5, 192, 13, 192, 0, 0, 0, 133, 5, 172, 12, 133, 5, 99, 12, 133, 5, 103, 12, 133, 5, 101, 12, 133, 5, 105, 12, 133, 5, 180, 12, 133, 5, 185, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 64, 3, 0, 5, 0, 0, 6, 122, 64, 133, 0, 5, 0, 0, 6, 122, 64, 3, 0, 5, 0, 0, 8, 122, 64, 133, 0, 5, 0, 0, 8, 122, 64, 3, 0, 5, 0, 0, 10, 122, 64, 133, 0, 5, 0, 0, 10, 122, 64, 3, 0, 5, 0, 0, 14, 122, 64, 133, 0, 5, 0, 0, 14, 122, 64, 3, 0, 5, 0, 0, 16, 122, 64, 133, 0, 5, 0, 0, 16, 122, 64, 133, 0, 5, 0, 0, 18, 122, 64, 133, 0, 5, 0, 0, 18, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 20, 122, 64, 133, 0, 5, 0, 0, 20, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 22, 122, 64, 133, 0, 5, 0, 0, 22, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 24, 122, 64, 133, 0, 5, 0, 0, 24, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 26, 122, 64, 133, 0, 5, 0, 0, 26, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 28, 122, 64, 133, 0, 5, 0, 0, 28, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 30, 122, 64, 133, 0, 5, 0, 0, 30, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 32, 122, 64, 133, 0, 5, 0, 0, 32, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 34, 122, 64, 133, 0, 5, 0, 0, 34, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 36, 122, 64, 133, 0, 5, 0, 0, 36, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 38, 122, 64, 133, 0, 5, 0, 0, 38, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 40, 122, 64, 133, 0, 5, 0, 0, 40, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 3, 0, 5, 0, 0, 42, 122, 64, 133, 0, 5, 0, 0, 42, 122, 64, 133, 0, 5, 0, 0, 42, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 44, 122, 64, 133, 0, 5, 0, 0, 44, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 46, 122, 64, 133, 0, 5, 0, 0, 46, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 48, 122, 64, 133, 0, 5, 0, 0, 50, 122, 64, 133, 0, 5, 0, 0, 52, 122, 64, 133, 0, 5, 0, 0, 54, 122, 64, 133, 0, 5, 0, 0, 56, 122, 64, 133, 0, 5, 0, 0, 58, 122, 64, 133, 0, 5, 0, 0, 58, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 58, 122, 0, 5, 0, 176, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 60, 122, 64, 133, 0, 5, 0, 0, 60, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 60, 122, 0, 5, 0, 176, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 62, 122, 64, 133, 0, 5, 0, 0, 62, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 62, 122, 0, 5, 0, 176, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 64, 122, 64, 133, 0, 5, 0, 0, 64, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 64, 122, 0, 5, 0, 176, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 66, 122, 64, 133, 0, 5, 0, 0, 66, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 66, 122, 0, 5, 0, 176, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 68, 122, 64, 133, 0, 5, 0, 0, 70, 122, 64, 133, 0, 5, 0, 0, 72, 122, 64, 133, 0, 5, 0, 0, 74, 122, 64, 133, 0, 5, 0, 0, 76, 122, 64, 3, 0, 5, 0, 0, 78, 122, 64, 133, 0, 5, 0, 0, 78, 122, 64, 3, 0, 5, 0, 0, 80, 122, 64, 133, 0, 5, 0, 0, 80, 122, 64, 3, 0, 5, 0, 0, 82, 122, 64, 133, 0, 5, 0, 0, 82, 122, 64, 133, 0, 5, 0, 0, 84, 122, 64, 133, 0, 5, 0, 0, 86, 122, 64, 133, 0, 5, 0, 0, 88, 122, 64, 133, 0, 5, 0, 0, 90, 122, 64, 133, 0, 5, 0, 0, 92, 122, 64, 3, 0, 5, 0, 0, 94, 122, 64, 133, 0, 5, 0, 0, 94, 122, 64, 133, 0, 5, 0, 0, 96, 122, 64, 133, 0, 5, 0, 0, 98, 122, 64, 133, 0, 5, 0, 0, 100, 122, 64, 133, 0, 5, 0, 0, 102, 122, 64, 133, 0, 5, 0, 0, 10, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 3, 0, 5, 0, 0, 18, 122, 64, 3, 0, 5, 0, 0, 24, 122, 64, 133, 0, 5, 0, 0, 94, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 96, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 98, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 133, 0, 5, 0, 0, 100, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 6, 122, 64, 2, 0, 5, 0, 0, 8, 122, 64, 2, 0, 5, 0, 0, 10, 122, 64, 2, 0, 5, 0, 0, 14, 122, 64, 2, 0, 5, 0, 0, 16, 122, 64, 5, 0, 5, 0, 0, 116, 13, 64, 4, 0, 5, 0, 0, 6, 122, 64, 4, 0, 5, 0, 0, 8, 122, 64, 4, 0, 5, 0, 0, 10, 122, 64, 4, 0, 5, 0, 0, 14, 122, 64, 4, 0, 5, 0, 0, 16, 122, 64, 4, 0, 5, 0, 0, 18, 122, 0, 5, 0, 5, 0, 0, 18, 122, 64, 2, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 20, 122, 64, 2, 0, 5, 0, 0, 22, 122, 64, 4, 0, 5, 0, 0, 24, 122, 0, 5, 0, 5, 0, 0, 24, 122, 64, 3, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 26, 122, 64, 2, 0, 5, 0, 0, 28, 122, 64, 4, 0, 5, 0, 0, 30, 122, 64, 4, 0, 5, 0, 0, 32, 122, 64, 2, 0, 5, 0, 0, 34, 122, 64, 2, 0, 5, 0, 0, 36, 122, 64, 2, 0, 5, 0, 0, 38, 122, 64, 2, 0, 5, 0, 0, 40, 122, 64, 4, 0, 5, 0, 0, 42, 122, 64, 2, 0, 5, 0, 0, 44, 122, 64, 2, 0, 5, 0, 0, 46, 122, 64, 2, 0, 5, 0, 0, 48, 122, 64, 2, 0, 5, 0, 0, 50, 122, 64, 2, 0, 5, 0, 0, 52, 122, 64, 2, 0, 5, 0, 0, 54, 122, 64, 2, 0, 5, 0, 0, 56, 122, 64, 2, 0, 5, 0, 0, 58, 122, 64, 2, 0, 5, 0, 0, 60, 122, 64, 2, 0, 5, 0, 0, 62, 122, 64, 2, 0, 5, 0, 0, 64, 122, 64, 2, 0, 5, 0, 0, 66, 122, 64, 4, 0, 5, 0, 0, 68, 122, 64, 2, 0, 5, 0, 0, 70, 122, 64, 2, 0, 5, 0, 0, 72, 122, 64, 4, 0, 5, 0, 0, 74, 122, 64, 2, 0, 5, 0, 0, 76, 122, 64, 4, 0, 5, 0, 0, 78, 122, 64, 4, 0, 5, 0, 0, 80, 122, 64, 4, 0, 5, 0, 0, 82, 122, 64, 2, 0, 5, 0, 0, 84, 122, 64, 2, 0, 5, 0, 0, 86, 122, 64, 2, 0, 5, 0, 0, 88, 122, 64, 2, 0, 5, 0, 0, 90, 122, 64, 2, 0, 5, 0, 0, 92, 122, 64, 4, 0, 5, 0, 0, 94, 122, 64, 4, 0, 5, 0, 0, 96, 122, 64, 4, 0, 5, 0, 0, 98, 122, 64, 4, 0, 5, 0, 0, 100, 122, 64, 2, 0, 5, 0, 0, 102, 122, 64, 5, 0, 5, 0, 0, 116, 13, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 6, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 8, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 10, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 14, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 16, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 18, 122, 0, 5, 0, 174, 0, 0, 0, 0, 0, 5, 0, 5, 0, 0, 18, 122, 64, 2, 0, 174, 0, 0, 0, 0, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 20, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 22, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 24, 122, 0, 5, 0, 174, 0, 0, 0, 0, 0, 5, 0, 5, 0, 0, 24, 122, 64, 3, 0, 174, 0, 0, 0, 0, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 26, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 28, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 30, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 32, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 34, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 36, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 38, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 40, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 42, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 44, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 46, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 48, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 50, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 52, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 54, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 56, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 58, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 60, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 62, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 64, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 66, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 68, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 70, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 72, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 74, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 76, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 78, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 80, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 82, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 84, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 86, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 88, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 90, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 92, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 94, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 96, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 98, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 4, 0, 5, 0, 0, 100, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 2, 0, 5, 0, 0, 102, 122, 0, 5, 0, 174, 0, 0, 0, 0, 64, 5, 0, 5, 0, 0, 26, 122, 0, 5, 0, 70, 0, 0, 46, 122, 0, 5, 0, 5, 212, 55, 18, 254, 0, 133, 0, 5, 0, 247, 101, 121, 0, 0, 0, 0, 5, 5, 18, 122, 2, 174, 0, 0, 5, 5, 24, 122, 3, 174, 0, 0, 5, 5, 116, 13, 5, 174, 0, 0, 4, 5, 6, 122, 5, 174, 0, 0, 4, 5, 8, 122, 5, 174, 0, 0, 4, 5, 10, 122, 5, 174, 0, 0, 4, 5, 14, 122, 5, 174, 0, 0, 4, 5, 16, 122, 5, 174, 0, 0, 4, 5, 18, 122, 5, 174, 0, 0, 5, 5, 18, 122, 2, 174, 0, 0, 5, 174, 0, 0, 2, 5, 20, 122, 5, 174, 0, 0, 2, 5, 22, 122, 5, 174, 0, 0, 4, 5, 24, 122, 5, 174, 0, 0, 5, 5, 24, 122, 3, 174, 0, 0, 5, 174, 0, 0, 2, 5, 26, 122, 5, 174, 0, 0, 2, 5, 28, 122, 5, 174, 0, 0, 4, 5, 30, 122, 5, 174, 0, 0, 4, 5, 32, 122, 5, 174, 0, 0, 2, 5, 34, 122, 5, 174, 0, 0, 2, 5, 36, 122, 5, 174, 0, 0, 2, 5, 38, 122, 5, 174, 0, 0, 2, 5, 40, 122, 5, 174, 0, 0, 4, 5, 42, 122, 5, 174, 0, 0, 2, 5, 44, 122, 5, 174, 0, 0, 2, 5, 46, 122, 5, 174, 0, 0, 2, 5, 48, 122, 5, 174, 0, 0, 2, 5, 50, 122, 5, 174, 0, 0, 2, 5, 52, 122, 5, 174, 0, 0, 2, 5, 54, 122, 5, 174, 0, 0, 2, 5, 56, 122, 5, 174, 0, 0, 2, 5, 58, 122, 5, 174, 0, 0, 2, 5, 60, 122, 5, 174, 0, 0, 2, 5, 62, 122, 5, 174, 0, 0, 2, 5, 64, 122, 5, 174, 0, 0, 2, 5, 66, 122, 5, 174, 0, 0, 4, 5, 68, 122, 5, 174, 0, 0, 2, 5, 70, 122, 5, 174, 0, 0, 2, 5, 72, 122, 5, 174, 0, 0, 4, 5, 74, 122, 5, 174, 0, 0, 2, 5, 76, 122, 5, 174, 0, 0, 4, 5, 78, 122, 5, 174, 0, 0, 4, 5, 80, 122, 5, 174, 0, 0, 4, 5, 82, 122, 5, 174, 0, 0, 2, 5, 84, 122, 5, 174, 0, 0, 2, 5, 86, 122, 5, 174, 0, 0, 2, 5, 88, 122, 5, 174, 0, 0, 2, 5, 90, 122, 5, 174, 0, 0, 2, 5, 92, 122, 5, 174, 0, 0, 4, 5, 94, 122, 5, 174, 0, 0, 4, 5, 96, 122, 5, 174, 0, 0, 4, 5, 98, 122, 5, 174, 0, 0, 4, 5, 100, 122, 5, 174, 0, 0, 2, 5, 102, 122, 5, 174, 0, 0, 5, 5, 82, 122, 5, 70, 86, 122, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 116, 13, 5, 5, 0, 0, 87, 0, 109, 48, 48, 1, 131, 48, 204, 0, 142, 48, 154, 0, 147, 48, 132, 0, 150, 48, 118, 0, 150, 48, 255, 255, 24, 122, 4, 5, 153, 48, 22, 0, 154, 48, 4, 0, 111, 48, 255, 255, 58, 122, 2, 5, 114, 48, 255, 255, 60, 122, 2, 5, 117, 48, 255, 255, 62, 122, 2, 5, 120, 48, 255, 255, 64, 122, 2, 5, 123, 48, 255, 255, 66, 122, 2, 5, 20, 0, 93, 48, 48, 0, 104, 48, 26, 0, 117, 48, 12, 0, 117, 48, 255, 255, 62, 122, 2, 5, 120, 48, 255, 255, 64, 122, 2, 5, 123, 48, 255, 255, 66, 122, 2, 5, 104, 48, 255, 255, 46, 122, 2, 5, 111, 48, 255, 255, 58, 122, 2, 5, 114, 48, 255, 255, 60, 122, 2, 5, 93, 48, 255, 255, 36, 122, 2, 5, 95, 48, 255, 255, 38, 122, 2, 5, 97, 48, 255, 255, 40, 122, 2, 5, 100, 48, 255, 255, 42, 122, 4, 5, 102, 48, 255, 255, 44, 122, 2, 5, 83, 48, 20, 0, 83, 48, 255, 255, 26, 122, 2, 5, 85, 48, 255, 255, 28, 122, 2, 5, 87, 48, 255, 255, 30, 122, 4, 5, 89, 48, 255, 255, 32, 122, 4, 5, 91, 48, 255, 255, 34, 122, 2, 5, 70, 48, 255, 255, 10, 122, 4, 5, 75, 48, 197, 162, 77, 48, 255, 255, 20, 122, 2, 5, 79, 48, 255, 255, 22, 122, 2, 5, 81, 48, 0, 192, 197, 98, 147, 48, 255, 255, 102, 122, 2, 5, 148, 48, 255, 255, 10, 122, 4, 5, 149, 48, 255, 255, 18, 122, 4, 5, 142, 48, 255, 255, 94, 122, 4, 5, 143, 48, 255, 255, 94, 122, 4, 5, 144, 48, 255, 255, 96, 122, 4, 5, 145, 48, 255, 255, 98, 122, 4, 5, 146, 48, 255, 255, 100, 122, 4, 5, 136, 48, 26, 0, 139, 48, 12, 0, 139, 48, 255, 255, 88, 122, 2, 5, 140, 48, 255, 255, 90, 122, 2, 5, 141, 48, 255, 255, 92, 122, 2, 5, 136, 48, 255, 255, 82, 122, 4, 5, 137, 48, 255, 255, 84, 122, 2, 5, 138, 48, 255, 255, 86, 122, 2, 5, 131, 48, 255, 255, 78, 122, 4, 5, 132, 48, 255, 255, 78, 122, 4, 5, 133, 48, 255, 255, 80, 122, 4, 5, 134, 48, 255, 255, 80, 122, 4, 5, 135, 48, 255, 255, 82, 122, 4, 5, 120, 48, 48, 0, 125, 48, 26, 0, 128, 48, 12, 0, 128, 48, 255, 255, 72, 122, 2, 5, 129, 48, 255, 255, 74, 122, 4, 5, 130, 48, 255, 255, 76, 122, 2, 5, 125, 48, 255, 255, 66, 122, 2, 5, 126, 48, 255, 255, 68, 122, 4, 5, 127, 48, 255, 255, 70, 122, 2, 5, 120, 48, 255, 255, 64, 122, 2, 5, 121, 48, 255, 255, 64, 122, 2, 5, 122, 48, 255, 255, 64, 122, 2, 5, 123, 48, 255, 255, 66, 122, 2, 5, 124, 48, 255, 255, 66, 122, 2, 5, 114, 48, 26, 0, 117, 48, 12, 0, 117, 48, 255, 255, 62, 122, 2, 5, 118, 48, 255, 255, 62, 122, 2, 5, 119, 48, 255, 255, 62, 122, 2, 5, 114, 48, 255, 255, 60, 122, 2, 5, 115, 48, 255, 255, 60, 122, 2, 5, 116, 48, 255, 255, 60, 122, 2, 5, 109, 48, 255, 255, 54, 122, 2, 5, 110, 48, 255, 255, 56, 122, 2, 5, 111, 48, 255, 255, 58, 122, 2, 5, 112, 48, 255, 255, 58, 122, 2, 5, 113, 48, 255, 255, 58, 122, 2, 5, 87, 48, 98, 0, 98, 48, 48, 0, 103, 48, 26, 0, 106, 48, 12, 0, 106, 48, 255, 255, 48, 122, 2, 5, 107, 48, 255, 255, 50, 122, 2, 5, 108, 48, 255, 255, 52, 122, 2, 5, 103, 48, 255, 255, 44, 122, 2, 5, 104, 48, 255, 255, 46, 122, 2, 5, 105, 48, 255, 255, 46, 122, 2, 5, 98, 48, 255, 255, 40, 122, 2, 5, 99, 48, 255, 255, 42, 122, 4, 5, 100, 48, 255, 255, 42, 122, 4, 5, 101, 48, 255, 255, 42, 122, 4, 5, 102, 48, 255, 255, 44, 122, 2, 5, 92, 48, 26, 0, 95, 48, 12, 0, 95, 48, 255, 255, 38, 122, 2, 5, 96, 48, 255, 255, 38, 122, 2, 5, 97, 48, 255, 255, 40, 122, 2, 5, 92, 48, 255, 255, 34, 122, 2, 5, 93, 48, 255, 255, 36, 122, 2, 5, 94, 48, 255, 255, 36, 122, 2, 5, 87, 48, 255, 255, 30, 122, 4, 5, 88, 48, 255, 255, 30, 122, 4, 5, 89, 48, 255, 255, 32, 122, 4, 5, 90, 48, 255, 255, 32, 122, 4, 5, 91, 48, 255, 255, 34, 122, 2, 5, 76, 48, 45, 0, 81, 48, 25, 0, 84, 48, 12, 0, 84, 48, 255, 255, 26, 122, 2, 5, 85, 48, 255, 255, 28, 122, 2, 5, 86, 48, 255, 255, 28, 122, 2, 5, 81, 48, 255, 255, 24, 122, 4, 5, 82, 48, 0, 192, 197, 98, 83, 48, 255, 255, 26, 122, 2, 5, 76, 48, 197, 162, 77, 48, 255, 255, 20, 122, 2, 5, 78, 48, 255, 255, 20, 122, 2, 5, 79, 48, 255, 255, 22, 122, 2, 5, 80, 48, 255, 255, 22, 122, 2, 5, 70, 48, 26, 0, 73, 48, 12, 0, 73, 48, 255, 255, 16, 122, 4, 5, 74, 48, 255, 255, 16, 122, 4, 5, 75, 48, 255, 255, 18, 122, 4, 5, 70, 48, 255, 255, 10, 122, 4, 5, 71, 48, 255, 255, 14, 122, 4, 5, 72, 48, 255, 255, 14, 122, 4, 5, 65, 48, 255, 255, 6, 122, 4, 5, 66, 48, 255, 255, 6, 122, 4, 5, 67, 48, 255, 255, 8, 122, 4, 5, 68, 48, 255, 255, 8, 122, 4, 5, 69, 48, 255, 255, 10, 122, 4, 5, 0, 0, 197, 162, 0, 0, 87, 0, 109, 48, 239, 0, 131, 48, 161, 0, 142, 48, 122, 0, 147, 48, 105, 0, 150, 48, 94, 0, 150, 48, 3, 192, 197, 66, 153, 48, 17, 0, 154, 48, 4, 0, 111, 48, 7, 192, 197, 226, 114, 48, 8, 192, 197, 34, 117, 48, 8, 192, 197, 98, 120, 48, 8, 192, 197, 162, 123, 48, 8, 192, 197, 226, 20, 0, 93, 48, 37, 0, 104, 48, 20, 0, 117, 48, 9, 0, 117, 48, 8, 192, 197, 98, 120, 48, 8, 192, 197, 162, 123, 48, 8, 192, 197, 226, 104, 48, 6, 192, 197, 98, 111, 48, 7, 192, 197, 226, 114, 48, 8, 192, 197, 34, 93, 48, 5, 192, 197, 34, 95, 48, 5, 192, 197, 98, 97, 48, 5, 192, 197, 162, 100, 48, 5, 192, 197, 226, 102, 48, 6, 192, 197, 34, 83, 48, 15, 0, 83, 48, 3, 192, 197, 226, 85, 48, 4, 192, 197, 34, 87, 48, 4, 192, 197, 98, 89, 48, 4, 192, 197, 162, 91, 48, 4, 192, 197, 226, 70, 48, 1, 192, 197, 98, 75, 48, 2, 192, 197, 99, 77, 48, 2, 192, 197, 194, 79, 48, 3, 192, 197, 2, 81, 48, 3, 192, 197, 131, 147, 48, 13, 192, 197, 98, 148, 48, 1, 192, 197, 98, 149, 48, 2, 192, 197, 34, 142, 48, 12, 192, 197, 98, 143, 48, 12, 192, 197, 98, 144, 48, 12, 192, 197, 162, 145, 48, 12, 192, 197, 226, 146, 48, 13, 192, 197, 34, 136, 48, 20, 0, 139, 48, 9, 0, 139, 48, 11, 192, 197, 162, 140, 48, 11, 192, 197, 226, 141, 48, 12, 192, 197, 34, 136, 48, 10, 192, 197, 226, 137, 48, 11, 192, 197, 34, 138, 48, 11, 192, 197, 98, 131, 48, 10, 192, 197, 98, 132, 48, 10, 192, 197, 98, 133, 48, 10, 192, 197, 162, 134, 48, 10, 192, 197, 162, 135, 48, 10, 192, 197, 226, 120, 48, 37, 0, 125, 48, 20, 0, 128, 48, 9, 0, 128, 48, 9, 192, 197, 162, 129, 48, 9, 192, 197, 226, 130, 48, 10, 192, 197, 34, 125, 48, 8, 192, 197, 226, 126, 48, 9, 192, 197, 34, 127, 48, 9, 192, 197, 98, 120, 48, 8, 192, 197, 162, 121, 48, 8, 192, 197, 162, 122, 48, 8, 192, 197, 162, 123, 48, 8, 192, 197, 226, 124, 48, 8, 192, 197, 226, 114, 48, 20, 0, 117, 48, 9, 0, 117, 48, 8, 192, 197, 98, 118, 48, 8, 192, 197, 98, 119, 48, 8, 192, 197, 98, 114, 48, 8, 192, 197, 34, 115, 48, 8, 192, 197, 34, 116, 48, 8, 192, 197, 34, 109, 48, 7, 192, 197, 98, 110, 48, 7, 192, 197, 162, 111, 48, 7, 192, 197, 226, 112, 48, 7, 192, 197, 226, 113, 48, 7, 192, 197, 226, 87, 48, 76, 0, 98, 48, 37, 0, 103, 48, 20, 0, 106, 48, 9, 0, 106, 48, 6, 192, 197, 162, 107, 48, 6, 192, 197, 226, 108, 48, 7, 192, 197, 34, 103, 48, 6, 192, 197, 34, 104, 48, 6, 192, 197, 98, 105, 48, 6, 192, 197, 98, 98, 48, 5, 192, 197, 162, 99, 48, 5, 192, 197, 226, 100, 48, 5, 192, 197, 226, 101, 48, 5, 192, 197, 226, 102, 48, 6, 192, 197, 34, 92, 48, 20, 0, 95, 48, 9, 0, 95, 48, 5, 192, 197, 98, 96, 48, 5, 192, 197, 98, 97, 48, 5, 192, 197, 162, 92, 48, 4, 192, 197, 226, 93, 48, 5, 192, 197, 34, 94, 48, 5, 192, 197, 34, 87, 48, 4, 192, 197, 98, 88, 48, 4, 192, 197, 98, 89, 48, 4, 192, 197, 162, 90, 48, 4, 192, 197, 162, 91, 48, 4, 192, 197, 226, 76, 48, 37, 0, 81, 48, 20, 0, 84, 48, 9, 0, 84, 48, 3, 192, 197, 226, 85, 48, 4, 192, 197, 34, 86, 48, 4, 192, 197, 34, 81, 48, 3, 192, 197, 66, 82, 48, 3, 192, 197, 131, 83, 48, 3, 192, 197, 226, 76, 48, 2, 192, 197, 99, 77, 48, 2, 192, 197, 194, 78, 48, 2, 192, 197, 194, 79, 48, 3, 192, 197, 2, 80, 48, 3, 192, 197, 2, 70, 48, 20, 0, 73, 48, 9, 0, 73, 48, 1, 192, 197, 226, 74, 48, 1, 192, 197, 226, 75, 48, 2, 192, 197, 34, 70, 48, 1, 192, 197, 98, 71, 48, 1, 192, 197, 162, 72, 48, 1, 192, 197, 162, 65, 48, 0, 192, 197, 226, 66, 48, 0, 192, 197, 226, 67, 48, 1, 192, 197, 34, 68, 48, 1, 192, 197, 34, 69, 48, 1, 192, 197, 98, 118, 13, 5, 5, 0, 0, 245, 0, 197, 48, 175, 1, 248, 49, 216, 0, 126, 255, 107, 0, 141, 255, 54, 0, 149, 255, 26, 0, 153, 255, 12, 0, 153, 255, 15, 192, 198, 65, 154, 255, 15, 192, 198, 97, 155, 255, 15, 192, 198, 129, 156, 255, 15, 192, 198, 1, 149, 255, 15, 192, 198, 65, 150, 255, 15, 192, 198, 129, 151, 255, 15, 192, 198, 1, 152, 255, 15, 192, 198, 33, 145, 255, 12, 0, 145, 255, 15, 192, 198, 65, 146, 255, 15, 192, 198, 97, 147, 255, 15, 192, 198, 129, 148, 255, 15, 192, 198, 1, 141, 255, 15, 192, 198, 97, 142, 255, 15, 192, 198, 129, 143, 255, 15, 192, 198, 1, 144, 255, 15, 192, 198, 33, 133, 255, 26, 0, 137, 255, 12, 0, 137, 255, 15, 192, 198, 129, 138, 255, 15, 192, 198, 1, 139, 255, 15, 192, 198, 33, 140, 255, 15, 192, 198, 65, 133, 255, 15, 192, 198, 1, 134, 255, 15, 192, 198, 33, 135, 255, 15, 192, 198, 65, 136, 255, 15, 192, 198, 97, 129, 255, 12, 0, 129, 255, 15, 192, 198, 33, 130, 255, 15, 192, 198, 65, 131, 255, 15, 192, 198, 97, 132, 255, 15, 192, 198, 129, 126, 255, 15, 192, 198, 97, 127, 255, 15, 192, 198, 129, 128, 255, 15, 192, 198, 1, 109, 255, 54, 0, 118, 255, 26, 0, 122, 255, 12, 0, 122, 255, 15, 192, 198, 129, 123, 255, 15, 192, 198, 1, 124, 255, 15, 192, 198, 33, 125, 255, 15, 192, 198, 65, 118, 255, 15, 192, 198, 1, 119, 255, 15, 192, 198, 33, 120, 255, 15, 192, 198, 65, 121, 255, 15, 192, 198, 97, 114, 255, 12, 0, 114, 255, 15, 192, 198, 33, 115, 255, 15, 192, 198, 65, 116, 255, 15, 192, 198, 97, 117, 255, 15, 192, 198, 129, 109, 255, 15, 192, 198, 65, 110, 255, 15, 192, 198, 129, 111, 255, 15, 192, 198, 65, 113, 255, 15, 192, 198, 1, 255, 49, 26, 0, 105, 255, 12, 0, 105, 255, 15, 192, 198, 65, 106, 255, 15, 192, 198, 97, 107, 255, 15, 192, 198, 129, 108, 255, 15, 192, 198, 1, 255, 49, 15, 192, 198, 129, 102, 255, 15, 192, 198, 129, 103, 255, 15, 192, 198, 1, 104, 255, 15, 192, 198, 33, 251, 49, 12, 0, 251, 49, 15, 192, 198, 1, 252, 49, 15, 192, 198, 33, 253, 49, 15, 192, 198, 65, 254, 49, 15, 192, 198, 97, 248, 49, 15, 192, 198, 97, 249, 49, 15, 192, 198, 129, 250, 49, 15, 192, 198, 65, 227, 48, 107, 0, 242, 48, 54, 0, 240, 49, 26, 0, 244, 49, 12, 0, 244, 49, 15, 192, 198, 65, 245, 49, 15, 192, 198, 1, 246, 49, 15, 192, 198, 33, 247, 49, 15, 192, 198, 65, 240, 49, 15, 192, 198, 65, 241, 49, 15, 192, 198, 33, 242, 49, 15, 192, 198, 65, 243, 49, 15, 192, 198, 129, 247, 48, 12, 0, 247, 48, 15, 192, 198, 1, 248, 48, 15, 192, 198, 33, 249, 48, 15, 192, 198, 97, 250, 48, 15, 192, 198, 129, 242, 48, 15, 192, 198, 129, 244, 48, 15, 192, 198, 65, 245, 48, 15, 192, 198, 1, 246, 48, 15, 192, 198, 97, 234, 48, 26, 0, 238, 48, 12, 0, 238, 48, 15, 192, 198, 1, 239, 48, 15, 192, 198, 1, 240, 48, 15, 192, 198, 33, 241, 48, 15, 192, 198, 97, 234, 48, 15, 192, 198, 33, 235, 48, 15, 192, 198, 65, 236, 48, 15, 192, 198, 97, 237, 48, 15, 192, 198, 129, 230, 48, 12, 0, 230, 48, 15, 192, 198, 65, 231, 48, 15, 192, 198, 129, 232, 48, 15, 192, 198, 129, 233, 48, 15, 192, 198, 1, 227, 48, 15, 192, 198, 1, 228, 48, 15, 192, 198, 1, 229, 48, 15, 192, 198, 65, 212, 48, 51, 0, 219, 48, 26, 0, 223, 48, 12, 0, 223, 48, 15, 192, 198, 33, 224, 48, 15, 192, 198, 65, 225, 48, 15, 192, 198, 97, 226, 48, 15, 192, 198, 129, 219, 48, 15, 192, 198, 129, 220, 48, 15, 192, 198, 129, 221, 48, 15, 192, 198, 129, 222, 48, 15, 192, 198, 1, 215, 48, 12, 0, 215, 48, 15, 192, 198, 65, 216, 48, 15, 192, 198, 97, 217, 48, 15, 192, 198, 97, 218, 48, 15, 192, 198, 97, 212, 48, 15, 192, 198, 33, 213, 48, 15, 192, 198, 65, 214, 48, 15, 192, 198, 65, 204, 48, 26, 0, 208, 48, 12, 0, 208, 48, 15, 192, 198, 1, 209, 48, 15, 192, 198, 1, 210, 48, 15, 192, 198, 33, 211, 48, 15, 192, 198, 33, 204, 48, 15, 192, 198, 65, 205, 48, 15, 192, 198, 97, 206, 48, 15, 192, 198, 129, 207, 48, 15, 192, 198, 1, 200, 48, 12, 0, 200, 48, 15, 192, 198, 129, 201, 48, 15, 192, 198, 129, 202, 48, 15, 192, 198, 1, 203, 48, 15, 192, 198, 33, 197, 48, 15, 192, 198, 65, 198, 48, 15, 192, 198, 97, 199, 48, 15, 192, 198, 97, 126, 48, 206, 1, 166, 48, 107, 0, 181, 48, 54, 0, 189, 48, 26, 0, 193, 48, 12, 0, 193, 48, 15, 192, 198, 33, 194, 48, 15, 192, 198, 33, 195, 48, 15, 192, 198, 65, 196, 48, 15, 192, 198, 65, 189, 48, 15, 192, 198, 129, 190, 48, 15, 192, 198, 129, 191, 48, 15, 192, 198, 1, 192, 48, 15, 192, 198, 1, 185, 48, 12, 0, 185, 48, 15, 192, 198, 65, 186, 48, 15, 192, 198, 65, 187, 48, 15, 192, 198, 97, 188, 48, 15, 192, 198, 97, 181, 48, 15, 192, 198, 1, 182, 48, 15, 192, 198, 1, 183, 48, 15, 192, 198, 33, 184, 48, 15, 192, 198, 33, 173, 48, 26, 0, 177, 48, 12, 0, 177, 48, 15, 192, 198, 97, 178, 48, 15, 192, 198, 97, 179, 48, 15, 192, 198, 129, 180, 48, 15, 192, 198, 129, 173, 48, 15, 192, 198, 33, 174, 48, 15, 192, 198, 33, 175, 48, 15, 192, 198, 65, 176, 48, 15, 192, 198, 65, 169, 48, 12, 0, 169, 48, 15, 192, 198, 129, 170, 48, 15, 192, 198, 129, 171, 48, 15, 192, 198, 1, 172, 48, 15, 192, 198, 1, 166, 48, 15, 192, 198, 65, 167, 48, 15, 192, 198, 97, 168, 48, 15, 192, 198, 97, 141, 48, 29, 1, 150, 48, 249, 0, 162, 48, 12, 0, 162, 48, 15, 192, 198, 1, 163, 48, 15, 192, 198, 33, 164, 48, 15, 192, 198, 33, 165, 48, 15, 192, 198, 65, 150, 48, 255, 255, 14, 122, 2, 5, 153, 48, 5, 0, 154, 48, 189, 0, 161, 48, 15, 192, 198, 1, 45, 0, 173, 48, 81, 0, 196, 48, 42, 0, 216, 48, 20, 0, 240, 48, 9, 0, 240, 48, 15, 192, 198, 33, 241, 48, 15, 192, 198, 97, 242, 48, 15, 192, 198, 129, 216, 48, 15, 192, 198, 97, 219, 48, 15, 192, 198, 129, 239, 48, 15, 192, 198, 1, 207, 48, 9, 0, 207, 48, 15, 192, 198, 1, 210, 48, 15, 192, 198, 33, 213, 48, 15, 192, 198, 65, 196, 48, 15, 192, 198, 65, 198, 48, 15, 192, 198, 97, 200, 48, 15, 192, 198, 129, 183, 48, 20, 0, 189, 48, 9, 0, 189, 48, 15, 192, 198, 129, 191, 48, 15, 192, 198, 1, 193, 48, 15, 192, 198, 33, 183, 48, 15, 192, 198, 33, 185, 48, 15, 192, 198, 65, 187, 48, 15, 192, 198, 97, 173, 48, 15, 192, 198, 33, 175, 48, 15, 192, 198, 65, 177, 48, 15, 192, 198, 97, 179, 48, 15, 192, 198, 129, 181, 48, 15, 192, 198, 1, 95, 48, 52, 0, 114, 48, 24, 0, 123, 48, 10, 0, 123, 48, 255, 255, 16, 122, 2, 5, 166, 48, 15, 192, 198, 65, 171, 48, 15, 192, 198, 1, 114, 48, 255, 255, 8, 122, 2, 5, 117, 48, 255, 255, 10, 122, 2, 5, 120, 48, 255, 255, 14, 122, 2, 5, 102, 48, 12, 0, 102, 48, 255, 255, 14, 122, 2, 5, 104, 48, 255, 255, 16, 122, 2, 5, 111, 48, 255, 255, 6, 122, 2, 5, 95, 48, 255, 255, 6, 122, 2, 5, 97, 48, 255, 255, 8, 122, 2, 5, 100, 48, 255, 255, 10, 122, 2, 5, 83, 48, 26, 0, 89, 48, 12, 0, 89, 48, 255, 255, 10, 122, 2, 5, 91, 48, 255, 255, 14, 122, 2, 5, 93, 48, 255, 255, 16, 122, 2, 5, 83, 48, 255, 255, 16, 122, 2, 5, 85, 48, 255, 255, 6, 122, 2, 5, 87, 48, 255, 255, 8, 122, 2, 5, 70, 48, 255, 255, 10, 122, 2, 5, 75, 48, 255, 255, 6, 122, 2, 5, 77, 48, 255, 255, 8, 122, 2, 5, 79, 48, 255, 255, 10, 122, 2, 5, 81, 48, 255, 255, 14, 122, 2, 5, 9, 0, 207, 48, 15, 0, 207, 48, 15, 192, 198, 1, 210, 48, 15, 192, 198, 33, 213, 48, 15, 192, 198, 65, 216, 48, 15, 192, 198, 97, 219, 48, 15, 192, 198, 129, 111, 48, 255, 255, 6, 122, 2, 5, 114, 48, 255, 255, 8, 122, 2, 5, 117, 48, 255, 255, 10, 122, 2, 5, 120, 48, 255, 255, 14, 122, 2, 5, 123, 48, 255, 255, 16, 122, 2, 5, 145, 48, 16, 0, 145, 48, 255, 255, 14, 122, 2, 5, 146, 48, 255, 255, 16, 122, 2, 5, 148, 48, 255, 255, 10, 122, 2, 5, 149, 48, 255, 255, 6, 122, 2, 5, 141, 48, 255, 255, 16, 122, 2, 5, 142, 48, 255, 255, 6, 122, 2, 5, 143, 48, 255, 255, 6, 122, 2, 5, 144, 48, 255, 255, 8, 122, 2, 5, 133, 48, 34, 0, 137, 48, 16, 0, 137, 48, 255, 255, 6, 122, 2, 5, 138, 48, 255, 255, 8, 122, 2, 5, 139, 48, 255, 255, 10, 122, 2, 5, 140, 48, 255, 255, 14, 122, 2, 5, 133, 48, 255, 255, 10, 122, 2, 5, 134, 48, 255, 255, 10, 122, 2, 5, 135, 48, 255, 255, 16, 122, 2, 5, 136, 48, 255, 255, 16, 122, 2, 5, 129, 48, 16, 0, 129, 48, 255, 255, 14, 122, 2, 5, 130, 48, 255, 255, 16, 122, 2, 5, 131, 48, 255, 255, 6, 122, 2, 5, 132, 48, 255, 255, 6, 122, 2, 5, 126, 48, 255, 255, 6, 122, 2, 5, 127, 48, 255, 255, 8, 122, 2, 5, 128, 48, 255, 255, 10, 122, 2, 5, 95, 48, 138, 0, 110, 48, 70, 0, 118, 48, 34, 0, 122, 48, 16, 0, 122, 48, 255, 255, 14, 122, 2, 5, 123, 48, 255, 255, 16, 122, 2, 5, 124, 48, 255, 255, 16, 122, 2, 5, 125, 48, 255, 255, 16, 122, 2, 5, 118, 48, 255, 255, 10, 122, 2, 5, 119, 48, 255, 255, 10, 122, 2, 5, 120, 48, 255, 255, 14, 122, 2, 5, 121, 48, 255, 255, 14, 122, 2, 5, 114, 48, 16, 0, 114, 48, 255, 255, 8, 122, 2, 5, 115, 48, 255, 255, 8, 122, 2, 5, 116, 48, 255, 255, 8, 122, 2, 5, 117, 48, 255, 255, 10, 122, 2, 5, 110, 48, 255, 255, 16, 122, 2, 5, 111, 48, 255, 255, 6, 122, 2, 5, 112, 48, 255, 255, 6, 122, 2, 5, 113, 48, 255, 255, 6, 122, 2, 5, 102, 48, 34, 0, 106, 48, 16, 0, 106, 48, 255, 255, 6, 122, 2, 5, 107, 48, 255, 255, 8, 122, 2, 5, 108, 48, 255, 255, 10, 122, 2, 5, 109, 48, 255, 255, 14, 122, 2, 5, 102, 48, 255, 255, 14, 122, 2, 5, 103, 48, 255, 255, 14, 122, 2, 5, 104, 48, 255, 255, 16, 122, 2, 5, 105, 48, 255, 255, 16, 122, 2, 5, 98, 48, 16, 0, 98, 48, 255, 255, 8, 122, 2, 5, 99, 48, 255, 255, 10, 122, 2, 5, 100, 48, 255, 255, 10, 122, 2, 5, 101, 48, 255, 255, 10, 122, 2, 5, 95, 48, 255, 255, 6, 122, 2, 5, 96, 48, 255, 255, 6, 122, 2, 5, 97, 48, 255, 255, 8, 122, 2, 5, 80, 48, 66, 0, 87, 48, 34, 0, 91, 48, 16, 0, 91, 48, 255, 255, 14, 122, 2, 5, 92, 48, 255, 255, 14, 122, 2, 5, 93, 48, 255, 255, 16, 122, 2, 5, 94, 48, 255, 255, 16, 122, 2, 5, 87, 48, 255, 255, 8, 122, 2, 5, 88, 48, 255, 255, 8, 122, 2, 5, 89, 48, 255, 255, 10, 122, 2, 5, 90, 48, 255, 255, 10, 122, 2, 5, 83, 48, 16, 0, 83, 48, 255, 255, 16, 122, 2, 5, 84, 48, 255, 255, 16, 122, 2, 5, 85, 48, 255, 255, 6, 122, 2, 5, 86, 48, 255, 255, 6, 122, 2, 5, 80, 48, 255, 255, 10, 122, 2, 5, 81, 48, 255, 255, 14, 122, 2, 5, 82, 48, 255, 255, 14, 122, 2, 5, 72, 48, 34, 0, 76, 48, 16, 0, 76, 48, 255, 255, 6, 122, 2, 5, 77, 48, 255, 255, 8, 122, 2, 5, 78, 48, 255, 255, 8, 122, 2, 5, 79, 48, 255, 255, 10, 122, 2, 5, 72, 48, 255, 255, 14, 122, 2, 5, 73, 48, 255, 255, 16, 122, 2, 5, 74, 48, 255, 255, 16, 122, 2, 5, 75, 48, 255, 255, 6, 122, 2, 5, 68, 48, 16, 0, 68, 48, 255, 255, 8, 122, 2, 5, 69, 48, 255, 255, 10, 122, 2, 5, 70, 48, 255, 255, 10, 122, 2, 5, 71, 48, 255, 255, 14, 122, 2, 5, 65, 48, 255, 255, 6, 122, 2, 5, 66, 48, 255, 255, 6, 122, 2, 5, 67, 48, 255, 255, 8, 122, 2, 5, 15, 0, 198, 161, 0, 0, 162, 0, 240, 48, 24, 1, 117, 255, 139, 0, 137, 255, 71, 0, 147, 255, 37, 0, 152, 255, 20, 0, 155, 255, 9, 0, 155, 255, 21, 192, 198, 129, 156, 255, 21, 192, 198, 161, 157, 255, 22, 192, 198, 33, 152, 255, 21, 192, 198, 33, 153, 255, 21, 192, 198, 65, 154, 255, 21, 192, 198, 97, 147, 255, 20, 192, 198, 129, 148, 255, 20, 192, 198, 161, 149, 255, 20, 192, 198, 193, 150, 255, 20, 192, 198, 225, 151, 255, 21, 192, 198, 1, 142, 255, 15, 0, 142, 255, 19, 192, 198, 225, 143, 255, 20, 192, 198, 1, 144, 255, 20, 192, 198, 33, 145, 255, 20, 192, 198, 65, 146, 255, 20, 192, 198, 97, 137, 255, 19, 192, 198, 65, 138, 255, 19, 192, 198, 97, 139, 255, 19, 192, 198, 129, 140, 255, 19, 192, 198, 161, 141, 255, 19, 192, 198, 193, 127, 255, 32, 0, 132, 255, 15, 0, 132, 255, 18, 192, 198, 161, 133, 255, 18, 192, 198, 193, 134, 255, 18, 192, 198, 225, 135, 255, 19, 192, 198, 1, 136, 255, 19, 192, 198, 33, 127, 255, 18, 192, 198, 1, 128, 255, 18, 192, 198, 33, 129, 255, 18, 192, 198, 65, 130, 255, 18, 192, 198, 97, 131, 255, 18, 192, 198, 129, 122, 255, 15, 0, 122, 255, 17, 192, 198, 97, 123, 255, 17, 192, 198, 129, 124, 255, 17, 192, 198, 161, 125, 255, 17, 192, 198, 193, 126, 255, 17, 192, 198, 225, 117, 255, 16, 192, 198, 65, 118, 255, 16, 192, 198, 97, 119, 255, 16, 192, 198, 193, 120, 255, 16, 192, 198, 225, 121, 255, 17, 192, 198, 1, 249, 49, 71, 0, 105, 255, 37, 0, 110, 255, 20, 0, 114, 255, 9, 0, 114, 255, 15, 192, 198, 225, 115, 255, 16, 192, 198, 1, 116, 255, 16, 192, 198, 33, 110, 255, 20, 192, 198, 225, 111, 255, 18, 192, 198, 97, 113, 255, 15, 192, 198, 193, 105, 255, 16, 192, 198, 1, 106, 255, 16, 192, 198, 33, 107, 255, 16, 192, 198, 65, 108, 255, 20, 192, 198, 161, 109, 255, 20, 192, 198, 193, 254, 49, 15, 0, 254, 49, 21, 192, 198, 97, 255, 49, 21, 192, 198, 129, 102, 255, 22, 192, 198, 1, 103, 255, 15, 192, 198, 193, 104, 255, 15, 192, 198, 225, 249, 49, 19, 192, 198, 225, 250, 49, 20, 192, 198, 65, 251, 49, 21, 192, 198, 1, 252, 49, 21, 192, 198, 33, 253, 49, 21, 192, 198, 65, 250, 48, 32, 0, 244, 49, 15, 0, 244, 49, 19, 192, 198, 1, 245, 49, 19, 192, 198, 97, 246, 49, 19, 192, 198, 129, 247, 49, 19, 192, 198, 161, 248, 49, 19, 192, 198, 193, 250, 48, 22, 192, 198, 1, 240, 49, 16, 192, 198, 225, 241, 49, 17, 192, 198, 161, 242, 49, 17, 192, 198, 193, 243, 49, 18, 192, 198, 161, 245, 48, 15, 0, 245, 48, 16, 192, 198, 97, 246, 48, 17, 192, 198, 1, 247, 48, 21, 192, 198, 161, 248, 48, 21, 192, 198, 193, 249, 48, 21, 192, 198, 225, 240, 48, 21, 192, 198, 193, 241, 48, 21, 192, 198, 225, 242, 48, 22, 192, 198, 1, 243, 48, 22, 192, 198, 33, 244, 48, 16, 192, 198, 1, 199, 48, 139, 0, 219, 48, 71, 0, 229, 48, 37, 0, 234, 48, 20, 0, 237, 48, 9, 0, 237, 48, 21, 192, 198, 129, 238, 48, 21, 192, 198, 161, 239, 48, 21, 192, 198, 161, 234, 48, 21, 192, 198, 33, 235, 48, 21, 192, 198, 65, 236, 48, 21, 192, 198, 97, 229, 48, 20, 192, 198, 193, 230, 48, 20, 192, 198, 193, 231, 48, 20, 192, 198, 225, 232, 48, 20, 192, 198, 225, 233, 48, 21, 192, 198, 1, 224, 48, 15, 0, 224, 48, 20, 192, 198, 65, 225, 48, 20, 192, 198, 97, 226, 48, 20, 192, 198, 129, 227, 48, 20, 192, 198, 161, 228, 48, 20, 192, 198, 161, 219, 48, 19, 192, 198, 225, 220, 48, 19, 192, 198, 225, 221, 48, 19, 192, 198, 225, 222, 48, 20, 192, 198, 1, 223, 48, 20, 192, 198, 33, 209, 48, 32, 0, 214, 48, 15, 0, 214, 48, 19, 192, 198, 161, 215, 48, 19, 192, 198, 161, 216, 48, 19, 192, 198, 193, 217, 48, 19, 192, 198, 193, 218, 48, 19, 192, 198, 193, 209, 48, 19, 192, 198, 97, 210, 48, 19, 192, 198, 129, 211, 48, 19, 192, 198, 129, 212, 48, 19, 192, 198, 129, 213, 48, 19, 192, 198, 161, 204, 48, 15, 0, 204, 48, 19, 192, 198, 1, 205, 48, 19, 192, 198, 33, 206, 48, 19, 192, 198, 65, 207, 48, 19, 192, 198, 97, 208, 48, 19, 192, 198, 97, 199, 48, 18, 192, 198, 129, 200, 48, 18, 192, 198, 161, 201, 48, 18, 192, 198, 161, 202, 48, 18, 192, 198, 193, 203, 48, 18, 192, 198, 225, 179, 48, 66, 0, 189, 48, 32, 0, 194, 48, 15, 0, 194, 48, 18, 192, 198, 65, 195, 48, 18, 192, 198, 97, 196, 48, 18, 192, 198, 97, 197, 48, 18, 192, 198, 97, 198, 48, 18, 192, 198, 129, 189, 48, 18, 192, 198, 1, 190, 48, 18, 192, 198, 1, 191, 48, 18, 192, 198, 33, 192, 48, 18, 192, 198, 33, 193, 48, 18, 192, 198, 65, 184, 48, 15, 0, 184, 48, 17, 192, 198, 161, 185, 48, 17, 192, 198, 193, 186, 48, 17, 192, 198, 193, 187, 48, 17, 192, 198, 225, 188, 48, 17, 192, 198, 225, 179, 48, 17, 192, 198, 97, 180, 48, 17, 192, 198, 97, 181, 48, 17, 192, 198, 129, 182, 48, 17, 192, 198, 129, 183, 48, 17, 192, 198, 161, 169, 48, 32, 0, 174, 48, 15, 0, 174, 48, 16, 192, 198, 193, 175, 48, 16, 192, 198, 225, 176, 48, 16, 192, 198, 225, 177, 48, 17, 192, 198, 1, 178, 48, 17, 192, 198, 34, 169, 48, 16, 192, 198, 65, 170, 48, 16, 192, 198, 65, 171, 48, 16, 192, 198, 97, 172, 48, 16, 192, 198, 130, 173, 48, 16, 192, 198, 193, 164, 48, 15, 0, 164, 48, 15, 192, 198, 225, 165, 48, 16, 192, 198, 1, 166, 48, 16, 192, 198, 1, 167, 48, 16, 192, 198, 33, 168, 48, 16, 192, 198, 33, 153, 48, 11, 0, 154, 48, 99, 0, 161, 48, 15, 192, 198, 193, 162, 48, 15, 192, 198, 193, 163, 48, 15, 192, 198, 225, 24, 0, 193, 48, 45, 0, 213, 48, 23, 0, 239, 48, 12, 0, 239, 48, 21, 192, 198, 161, 240, 48, 21, 192, 198, 193, 241, 48, 21, 192, 198, 225, 242, 48, 22, 192, 198, 1, 213, 48, 19, 192, 198, 161, 216, 48, 19, 192, 198, 193, 219, 48, 19, 192, 198, 225, 200, 48, 9, 0, 200, 48, 18, 192, 198, 161, 207, 48, 19, 192, 198, 97, 210, 48, 19, 192, 198, 129, 193, 48, 18, 192, 198, 65, 196, 48, 18, 192, 198, 97, 198, 48, 18, 192, 198, 129, 181, 48, 20, 0, 187, 48, 9, 0, 187, 48, 17, 192, 198, 225, 189, 48, 18, 192, 198, 1, 191, 48, 18, 192, 198, 33, 181, 48, 17, 192, 198, 129, 183, 48, 17, 192, 198, 161, 185, 48, 17, 192, 198, 193, 175, 48, 9, 0, 175, 48, 16, 192, 198, 225, 177, 48, 17, 192, 198, 34, 179, 48, 17, 192, 198, 97, 166, 48, 16, 192, 198, 1, 171, 48, 16, 192, 198, 130, 173, 48, 16, 192, 198, 193, 4, 0, 207, 48, 19, 192, 198, 97, 210, 48, 19, 192, 198, 129, 213, 48, 19, 192, 198, 161, 216, 48, 19, 192, 198, 193, 219, 48, 19, 192, 198, 225, 22, 0, 198, 66, 0, 0, 162, 0, 240, 48, 24, 1, 117, 255, 139, 0, 137, 255, 71, 0, 147, 255, 37, 0, 152, 255, 20, 0, 155, 255, 9, 0, 155, 255, 33, 192, 198, 194, 156, 255, 34, 192, 198, 2, 157, 255, 35, 192, 198, 2, 152, 255, 33, 192, 198, 2, 153, 255, 33, 192, 198, 66, 154, 255, 33, 192, 198, 130, 147, 255, 31, 192, 198, 194, 148, 255, 32, 192, 198, 2, 149, 255, 32, 192, 198, 66, 150, 255, 32, 192, 198, 130, 151, 255, 32, 192, 198, 194, 142, 255, 15, 0, 142, 255, 30, 192, 198, 130, 143, 255, 30, 192, 198, 194, 144, 255, 31, 192, 198, 2, 145, 255, 31, 192, 198, 66, 146, 255, 31, 192, 198, 130, 137, 255, 29, 192, 198, 66, 138, 255, 29, 192, 198, 130, 139, 255, 29, 192, 198, 194, 140, 255, 30, 192, 198, 2, 141, 255, 30, 192, 198, 66, 127, 255, 32, 0, 132, 255, 15, 0, 132, 255, 28, 192, 198, 2, 133, 255, 28, 192, 198, 66, 134, 255, 28, 192, 198, 130, 135, 255, 28, 192, 198, 194, 136, 255, 29, 192, 198, 2, 127, 255, 26, 192, 198, 194, 128, 255, 27, 192, 198, 2, 129, 255, 27, 192, 198, 66, 130, 255, 27, 192, 198, 130, 131, 255, 27, 192, 198, 194, 122, 255, 15, 0, 122, 255, 25, 192, 198, 130, 123, 255, 25, 192, 198, 194, 124, 255, 26, 192, 198, 2, 125, 255, 26, 192, 198, 66, 126, 255, 26, 192, 198, 130, 117, 255, 23, 192, 198, 130, 118, 255, 23, 192, 198, 194, 119, 255, 24, 192, 198, 98, 120, 255, 24, 192, 198, 162, 121, 255, 24, 192, 198, 226, 249, 49, 71, 0, 105, 255, 37, 0, 110, 255, 20, 0, 114, 255, 9, 0, 114, 255, 22, 192, 198, 194, 115, 255, 23, 192, 198, 2, 116, 255, 23, 192, 198, 66, 110, 255, 32, 192, 198, 130, 111, 255, 27, 192, 198, 130, 113, 255, 22, 192, 198, 130, 105, 255, 23, 192, 198, 2, 106, 255, 23, 192, 198, 66, 107, 255, 23, 192, 198, 130, 108, 255, 32, 192, 198, 2, 109, 255, 32, 192, 198, 66, 254, 49, 15, 0, 254, 49, 33, 192, 198, 130, 255, 49, 33, 192, 198, 194, 102, 255, 34, 192, 198, 194, 103, 255, 22, 192, 198, 130, 104, 255, 22, 192, 198, 194, 249, 49, 30, 192, 198, 130, 250, 49, 31, 192, 198, 66, 251, 49, 32, 192, 198, 194, 252, 49, 33, 192, 198, 2, 253, 49, 33, 192, 198, 66, 250, 48, 32, 0, 244, 49, 15, 0, 244, 49, 28, 192, 198, 194, 245, 49, 29, 192, 198, 130, 246, 49, 29, 192, 198, 194, 247, 49, 30, 192, 198, 2, 248, 49, 30, 192, 198, 66, 250, 48, 34, 192, 198, 194, 240, 49, 24, 192, 198, 162, 241, 49, 26, 192, 198, 2, 242, 49, 26, 192, 198, 66, 243, 49, 28, 192, 198, 2, 245, 48, 15, 0, 245, 48, 23, 192, 198, 194, 246, 48, 24, 192, 198, 226, 247, 48, 34, 192, 198, 2, 248, 48, 34, 192, 198, 66, 249, 48, 34, 192, 198, 130, 240, 48, 34, 192, 198, 66, 241, 48, 34, 192, 198, 130, 242, 48, 34, 192, 198, 194, 243, 48, 35, 192, 198, 2, 244, 48, 23, 192, 198, 2, 199, 48, 139, 0, 219, 48, 71, 0, 229, 48, 37, 0, 234, 48, 20, 0, 237, 48, 9, 0, 237, 48, 33, 192, 198, 194, 238, 48, 34, 192, 198, 2, 239, 48, 34, 192, 198, 2, 234, 48, 33, 192, 198, 2, 235, 48, 33, 192, 198, 66, 236, 48, 33, 192, 198, 130, 229, 48, 32, 192, 198, 66, 230, 48, 32, 192, 198, 66, 231, 48, 32, 192, 198, 130, 232, 48, 32, 192, 198, 130, 233, 48, 32, 192, 198, 194, 224, 48, 15, 0, 224, 48, 31, 192, 198, 66, 225, 48, 31, 192, 198, 130, 226, 48, 31, 192, 198, 194, 227, 48, 32, 192, 198, 2, 228, 48, 32, 192, 198, 2, 219, 48, 30, 192, 198, 130, 220, 48, 30, 192, 198, 130, 221, 48, 30, 192, 198, 130, 222, 48, 30, 192, 198, 194, 223, 48, 31, 192, 198, 2, 209, 48, 32, 0, 214, 48, 15, 0, 214, 48, 30, 192, 198, 2, 215, 48, 30, 192, 198, 2, 216, 48, 30, 192, 198, 66, 217, 48, 30, 192, 198, 66, 218, 48, 30, 192, 198, 66, 209, 48, 29, 192, 198, 130, 210, 48, 29, 192, 198, 194, 211, 48, 29, 192, 198, 194, 212, 48, 29, 192, 198, 194, 213, 48, 30, 192, 198, 2, 204, 48, 15, 0, 204, 48, 28, 192, 198, 194, 205, 48, 29, 192, 198, 2, 206, 48, 29, 192, 198, 66, 207, 48, 29, 192, 198, 130, 208, 48, 29, 192, 198, 130, 199, 48, 27, 192, 198, 194, 200, 48, 28, 192, 198, 2, 201, 48, 28, 192, 198, 2, 202, 48, 28, 192, 198, 66, 203, 48, 28, 192, 198, 130, 179, 48, 66, 0, 189, 48, 32, 0, 194, 48, 15, 0, 194, 48, 27, 192, 198, 66, 195, 48, 27, 192, 198, 130, 196, 48, 27, 192, 198, 130, 197, 48, 27, 192, 198, 130, 198, 48, 27, 192, 198, 194, 189, 48, 26, 192, 198, 194, 190, 48, 26, 192, 198, 194, 191, 48, 27, 192, 198, 2, 192, 48, 27, 192, 198, 2, 193, 48, 27, 192, 198, 66, 184, 48, 15, 0, 184, 48, 26, 192, 198, 2, 185, 48, 26, 192, 198, 66, 186, 48, 26, 192, 198, 66, 187, 48, 26, 192, 198, 130, 188, 48, 26, 192, 198, 130, 179, 48, 25, 192, 198, 130, 180, 48, 25, 192, 198, 130, 181, 48, 25, 192, 198, 194, 182, 48, 25, 192, 198, 194, 183, 48, 26, 192, 198, 2, 169, 48, 32, 0, 174, 48, 15, 0, 174, 48, 24, 192, 198, 98, 175, 48, 24, 192, 198, 162, 176, 48, 24, 192, 198, 162, 177, 48, 24, 192, 198, 226, 178, 48, 25, 192, 198, 35, 169, 48, 23, 192, 198, 130, 170, 48, 23, 192, 198, 130, 171, 48, 23, 192, 198, 194, 172, 48, 24, 192, 198, 3, 173, 48, 24, 192, 198, 98, 164, 48, 15, 0, 164, 48, 22, 192, 198, 194, 165, 48, 23, 192, 198, 2, 166, 48, 23, 192, 198, 2, 167, 48, 23, 192, 198, 66, 168, 48, 23, 192, 198, 66, 153, 48, 11, 0, 154, 48, 99, 0, 161, 48, 22, 192, 198, 130, 162, 48, 22, 192, 198, 130, 163, 48, 22, 192, 198, 194, 24, 0, 193, 48, 45, 0, 213, 48, 23, 0, 239, 48, 12, 0, 239, 48, 34, 192, 198, 2, 240, 48, 34, 192, 198, 66, 241, 48, 34, 192, 198, 130, 242, 48, 34, 192, 198, 194, 213, 48, 30, 192, 198, 2, 216, 48, 30, 192, 198, 66, 219, 48, 30, 192, 198, 130, 200, 48, 9, 0, 200, 48, 28, 192, 198, 2, 207, 48, 29, 192, 198, 130, 210, 48, 29, 192, 198, 194, 193, 48, 27, 192, 198, 66, 196, 48, 27, 192, 198, 130, 198, 48, 27, 192, 198, 194, 181, 48, 20, 0, 187, 48, 9, 0, 187, 48, 26, 192, 198, 130, 189, 48, 26, 192, 198, 194, 191, 48, 27, 192, 198, 2, 181, 48, 25, 192, 198, 194, 183, 48, 26, 192, 198, 2, 185, 48, 26, 192, 198, 66, 175, 48, 9, 0, 175, 48, 24, 192, 198, 162, 177, 48, 25, 192, 198, 35, 179, 48, 25, 192, 198, 130, 166, 48, 23, 192, 198, 2, 171, 48, 24, 192, 198, 3, 173, 48, 24, 192, 198, 98, 4, 0, 207, 48, 29, 192, 198, 130, 210, 48, 29, 192, 198, 194, 213, 48, 30, 192, 198, 2, 216, 48, 30, 192, 198, 66, 219, 48, 30, 192, 198, 130, 35, 0, 198, 129, 0, 0, 0, 1, 12, 128, 107, 2, 117, 152, 75, 1, 254, 158, 187, 0, 7, 249, 74, 0, 196, 249, 38, 0, 10, 250, 20, 0, 10, 250, 255, 255, 25, 253, 193, 42, 30, 250, 255, 255, 195, 252, 193, 136, 60, 250, 255, 255, 135, 251, 193, 8, 149, 250, 255, 255, 11, 252, 193, 192, 206, 250, 255, 255, 224, 253, 193, 118, 196, 249, 255, 255, 224, 253, 193, 12, 233, 249, 255, 255, 98, 253, 193, 98, 247, 249, 255, 255, 152, 252, 193, 118, 8, 250, 255, 255, 15, 253, 193, 118, 64, 249, 16, 0, 64, 249, 255, 255, 209, 253, 193, 178, 113, 249, 255, 255, 78, 253, 193, 128, 129, 249, 255, 255, 106, 251, 193, 100, 138, 249, 255, 255, 42, 251, 193, 8, 7, 249, 255, 255, 224, 253, 193, 118, 8, 249, 255, 255, 224, 253, 193, 118, 10, 249, 255, 255, 98, 253, 193, 242, 52, 249, 255, 255, 198, 252, 193, 84, 127, 159, 75, 0, 159, 159, 57, 0, 159, 159, 255, 255, 224, 253, 193, 218, 160, 159, 255, 255, 224, 253, 193, 236, 126, 216, 4, 0, 2, 249, 255, 255, 70, 253, 193, 218, 9, 0, 210, 221, 20, 0, 210, 221, 255, 255, 45, 253, 193, 112, 21, 222, 255, 255, 213, 253, 193, 94, 23, 222, 255, 255, 217, 253, 193, 36, 24, 222, 255, 255, 217, 253, 193, 222, 28, 222, 255, 255, 219, 253, 193, 240, 29, 220, 255, 255, 34, 251, 193, 200, 117, 220, 255, 255, 130, 251, 193, 144, 120, 220, 255, 255, 135, 251, 193, 8, 144, 220, 255, 255, 160, 251, 193, 102, 196, 221, 255, 255, 16, 253, 193, 52, 127, 159, 255, 255, 223, 253, 193, 110, 141, 159, 255, 255, 224, 253, 193, 12, 153, 159, 255, 255, 224, 253, 193, 96, 156, 159, 255, 255, 224, 253, 193, 118, 59, 159, 16, 0, 59, 159, 255, 255, 219, 253, 193, 240, 74, 159, 255, 255, 220, 253, 193, 176, 80, 159, 255, 255, 220, 253, 193, 236, 82, 159, 255, 255, 220, 253, 193, 248, 254, 158, 255, 255, 217, 253, 193, 222, 14, 159, 255, 255, 218, 253, 193, 6, 19, 159, 255, 255, 218, 253, 193, 62, 32, 159, 255, 255, 218, 253, 193, 220, 50, 155, 70, 0, 165, 158, 34, 0, 205, 158, 16, 0, 205, 158, 255, 255, 214, 253, 193, 124, 209, 158, 255, 255, 214, 253, 193, 252, 249, 158, 255, 255, 217, 253, 193, 36, 253, 158, 255, 255, 217, 253, 193, 62, 165, 158, 255, 255, 211, 253, 193, 56, 166, 158, 255, 255, 212, 253, 193, 196, 187, 158, 255, 255, 213, 253, 193, 94, 195, 158, 255, 255, 213, 253, 193, 210, 229, 156, 16, 0, 229, 156, 255, 255, 195, 253, 193, 218, 31, 158, 255, 255, 206, 253, 193, 72, 117, 158, 255, 255, 208, 253, 193, 240, 127, 158, 255, 255, 209, 253, 193, 178, 50, 155, 255, 255, 180, 253, 193, 172, 60, 155, 255, 255, 181, 253, 193, 166, 90, 155, 255, 255, 184, 253, 193, 2, 124, 156, 255, 255, 193, 253, 193, 64, 153, 153, 34, 0, 216, 154, 16, 0, 216, 154, 255, 255, 175, 253, 193, 232, 223, 154, 255, 255, 176, 253, 193, 186, 37, 155, 255, 255, 180, 253, 193, 58, 47, 155, 255, 255, 180, 253, 193, 136, 153, 153, 255, 255, 164, 253, 193, 252, 172, 153, 255, 255, 165, 253, 193, 208, 108, 154, 255, 255, 171, 253, 193, 244, 168, 154, 255, 255, 173, 253, 193, 168, 222, 152, 16, 0, 222, 152, 255, 255, 157, 253, 193, 164, 223, 152, 255, 255, 157, 253, 193, 170, 99, 153, 255, 255, 163, 253, 193, 20, 150, 153, 255, 255, 164, 253, 193, 110, 117, 152, 255, 255, 153, 253, 193, 70, 168, 152, 255, 255, 154, 253, 193, 82, 206, 152, 255, 255, 156, 253, 193, 242, 219, 152, 255, 255, 157, 253, 193, 122, 100, 141, 142, 0, 127, 149, 70, 0, 94, 151, 34, 0, 230, 151, 16, 0, 230, 151, 255, 255, 147, 253, 193, 30, 237, 151, 255, 255, 147, 253, 193, 108, 243, 151, 255, 255, 147, 253, 193, 184, 1, 152, 255, 255, 148, 253, 193, 142, 94, 151, 255, 255, 140, 253, 193, 44, 98, 151, 255, 255, 140, 253, 193, 152, 105, 151, 255, 255, 141, 253, 193, 160, 203, 151, 255, 255, 145, 253, 193, 196, 182, 150, 16, 0, 182, 150, 255, 255, 131, 253, 193, 186, 185, 150, 255, 255, 131, 253, 193, 228, 232, 150, 255, 255, 134, 253, 193, 178, 81, 151, 255, 255, 139, 253, 193, 216, 127, 149, 255, 255, 120, 253, 193, 2, 128, 149, 255, 255, 120, 253, 193, 6, 232, 149, 255, 255, 124, 253, 193, 84, 28, 150, 255, 255, 125, 253, 193, 154, 181, 143, 34, 0, 204, 145, 16, 0, 204, 145, 255, 255, 98, 253, 193, 98, 209, 145, 255, 255, 98, 253, 193, 242, 133, 148, 255, 255, 114, 253, 193, 86, 119, 149, 255, 255, 118, 253, 193, 200, 181, 143, 255, 255, 78, 253, 193, 176, 145, 144, 255, 255, 87, 253, 193, 98, 73, 145, 255, 255, 93, 253, 193, 148, 198, 145, 255, 255, 98, 253, 193, 30, 202, 142, 16, 0, 202, 142, 255, 255, 70, 253, 193, 218, 102, 143, 255, 255, 76, 253, 193, 100, 155, 143, 255, 255, 77, 253, 193, 186, 176, 143, 255, 255, 78, 253, 193, 128, 100, 141, 255, 255, 54, 253, 193, 190, 112, 141, 255, 255, 55, 253, 193, 84, 179, 141, 255, 255, 59, 253, 193, 84, 171, 142, 255, 255, 68, 253, 193, 230, 107, 134, 70, 0, 0, 138, 34, 0, 85, 140, 16, 0, 85, 140, 255, 255, 45, 253, 193, 112, 120, 140, 255, 255, 47, 253, 193, 36, 157, 140, 255, 255, 48, 253, 193, 190, 29, 141, 255, 255, 53, 253, 193, 202, 0, 138, 255, 255, 30, 253, 193, 38, 160, 139, 255, 255, 40, 253, 193, 192, 55, 140, 255, 255, 43, 253, 193, 160, 70, 140, 255, 255, 44, 253, 193, 90, 126, 137, 16, 0, 126, 137, 255, 255, 24, 253, 193, 136, 139, 137, 255, 255, 25, 253, 193, 42, 193, 137, 255, 255, 27, 253, 193, 148, 210, 137, 255, 255, 27, 253, 193, 240, 107, 134, 255, 255, 254, 252, 193, 150, 64, 136, 255, 255, 14, 253, 193, 202, 76, 136, 255, 255, 15, 253, 193, 118, 99, 136, 255, 255, 16, 253, 193, 52, 252, 129, 34, 0, 110, 130, 16, 0, 110, 130, 255, 255, 222, 252, 193, 106, 114, 130, 255, 255, 222, 252, 193, 140, 120, 130, 255, 255, 223, 252, 193, 6, 77, 134, 255, 255, 252, 252, 193, 174, 252, 129, 255, 255, 217, 252, 193, 10, 12, 130, 255, 255, 218, 252, 193, 106, 27, 130, 255, 255, 219, 252, 193, 90, 31, 130, 255, 255, 219, 252, 193, 128, 137, 128, 16, 0, 137, 128, 255, 255, 203, 252, 193, 226, 227, 129, 255, 255, 215, 252, 193, 90, 234, 129, 255, 255, 215, 252, 193, 222, 243, 129, 255, 255, 216, 252, 193, 152, 12, 128, 255, 255, 198, 252, 193, 214, 18, 128, 255, 255, 199, 252, 193, 40, 51, 128, 255, 255, 200, 252, 193, 74, 127, 128, 255, 255, 203, 252, 193, 128, 54, 98, 30, 1, 132, 115, 142, 0, 226, 119, 70, 0, 115, 124, 34, 0, 81, 127, 16, 0, 81, 127, 255, 255, 189, 252, 193, 238, 138, 127, 255, 255, 193, 252, 193, 14, 189, 127, 255, 255, 195, 252, 193, 136, 1, 128, 255, 255, 198, 252, 193, 84, 115, 124, 255, 255, 168, 252, 193, 76, 248, 124, 255, 255, 174, 252, 193, 70, 159, 126, 255, 255, 185, 252, 193, 220, 54, 127, 255, 255, 188, 252, 193, 198, 190, 121, 16, 0, 190, 121, 255, 255, 140, 252, 193, 156, 116, 122, 255, 255, 148, 252, 193, 22, 203, 122, 255, 255, 152, 252, 193, 118, 249, 122, 255, 255, 154, 252, 193, 154, 226, 119, 255, 255, 124, 252, 193, 16, 243, 119, 255, 255, 125, 252, 193, 134, 58, 121, 255, 255, 134, 252, 193, 210, 184, 121, 255, 255, 140, 252, 193, 88, 139, 117, 34, 0, 174, 118, 16, 0, 174, 118, 255, 255, 109, 252, 193, 26, 191, 118, 255, 255, 110, 252, 193, 184, 238, 118, 255, 255, 113, 252, 193, 128, 219, 119, 255, 255, 123, 252, 193, 54, 139, 117, 255, 255, 98, 252, 193, 100, 146, 117, 255, 255, 98, 252, 193, 176, 118, 118, 255, 255, 106, 252, 193, 156, 125, 118, 255, 255, 106, 252, 193, 250, 24, 117, 16, 0, 24, 117, 255, 255, 92, 252, 193, 188, 31, 117, 255, 255, 93, 252, 193, 36, 40, 117, 255, 255, 93, 252, 193, 208, 48, 117, 255, 255, 94, 252, 193, 32, 132, 115, 255, 255, 79, 252, 193, 112, 137, 115, 255, 255, 79, 252, 193, 132, 220, 116, 255, 255, 89, 252, 193, 158, 230, 116, 255, 255, 90, 252, 193, 104, 179, 107, 70, 0, 42, 114, 34, 0, 71, 114, 16, 0, 71, 114, 255, 255, 66, 252, 193, 252, 89, 114, 255, 255, 68, 252, 193, 28, 91, 114, 255, 255, 68, 252, 193, 94, 172, 114, 255, 255, 72, 252, 193, 46, 42, 114, 255, 255, 64, 252, 193, 48, 54, 114, 255, 255, 65, 252, 193, 150, 59, 114, 255, 255, 65, 252, 193, 218, 63, 114, 255, 255, 66, 252, 193, 14, 15, 108, 16, 0, 15, 108, 255, 255, 20, 252, 193, 236, 20, 108, 255, 255, 21, 252, 193, 54, 52, 108, 255, 255, 21, 252, 193, 230, 107, 112, 255, 255, 48, 252, 193, 142, 179, 107, 255, 255, 14, 252, 193, 174, 203, 107, 255, 255, 16, 252, 193, 104, 212, 107, 255, 255, 16, 252, 193, 180, 219, 107, 255, 255, 17, 252, 193, 22, 224, 101, 34, 0, 40, 103, 16, 0, 40, 103, 255, 255, 235, 251, 193, 114, 32, 107, 255, 255, 6, 252, 193, 188, 98, 107, 255, 255, 9, 252, 193, 190, 121, 107, 255, 255, 11, 252, 193, 192, 224, 101, 255, 255, 222, 251, 193, 42, 229, 101, 255, 255, 222, 251, 193, 86, 240, 102, 255, 255, 232, 251, 193, 90, 8, 103, 255, 255, 233, 251, 193, 86, 135, 101, 16, 0, 135, 101, 255, 255, 217, 251, 193, 254, 151, 101, 255, 255, 218, 251, 193, 178, 164, 101, 255, 255, 219, 251, 193, 74, 185, 101, 255, 255, 220, 251, 193, 68, 54, 98, 255, 255, 190, 251, 193, 162, 75, 98, 255, 255, 191, 251, 193, 156, 47, 101, 255, 255, 212, 251, 193, 60, 52, 101, 255, 255, 212, 251, 193, 230, 215, 86, 142, 0, 219, 93, 70, 0, 254, 94, 34, 0, 97, 95, 16, 0, 97, 95, 255, 255, 165, 251, 193, 122, 115, 95, 255, 255, 166, 251, 193, 88, 195, 95, 255, 255, 169, 251, 193, 252, 8, 98, 255, 255, 188, 251, 193, 76, 254, 94, 255, 255, 160, 251, 193, 102, 11, 95, 255, 255, 161, 251, 193, 108, 19, 95, 255, 255, 161, 251, 193, 216, 80, 95, 255, 255, 164, 251, 193, 210, 114, 94, 16, 0, 114, 94, 255, 255, 153, 251, 193, 170, 122, 94, 255, 255, 154, 251, 193, 90, 127, 94, 255, 255, 154, 251, 193, 218, 244, 94, 255, 255, 159, 251, 193, 254, 219, 93, 255, 255, 147, 251, 193, 66, 229, 93, 255, 255, 147, 251, 193, 210, 241, 93, 255, 255, 148, 251, 193, 132, 254, 93, 255, 255, 149, 251, 193, 20, 80, 91, 34, 0, 34, 92, 16, 0, 34, 92, 255, 255, 130, 251, 193, 144, 56, 92, 255, 255, 131, 251, 193, 192, 110, 92, 255, 255, 135, 251, 193, 8, 113, 92, 255, 255, 135, 251, 193, 128, 80, 91, 255, 255, 119, 251, 193, 108, 128, 91, 255, 255, 121, 251, 193, 174, 248, 91, 255, 255, 128, 251, 193, 22, 15, 92, 255, 255, 129, 251, 193, 34, 10, 89, 16, 0, 10, 89, 255, 255, 100, 251, 193, 232, 21, 89, 255, 255, 101, 251, 193, 136, 39, 89, 255, 255, 102, 251, 193, 204, 115, 89, 255, 255, 106, 251, 193, 100, 215, 86, 255, 255, 83, 251, 193, 76, 31, 87, 255, 255, 85, 251, 193, 186, 235, 88, 255, 255, 99, 251, 193, 220, 2, 89, 255, 255, 100, 251, 193, 178, 224, 81, 70, 0, 65, 83, 34, 0, 130, 83, 16, 0, 130, 83, 255, 255, 50, 251, 193, 178, 182, 83, 255, 255, 53, 251, 193, 120, 200, 83, 255, 255, 54, 251, 193, 122, 227, 83, 255, 255, 56, 251, 193, 112, 65, 83, 255, 255, 47, 251, 193, 180, 92, 83, 255, 255, 49, 251, 193, 44, 100, 83, 255, 255, 209, 253, 193, 152, 105, 83, 255, 255, 49, 251, 193, 240, 249, 82, 16, 0, 249, 82, 255, 255, 44, 251, 193, 244, 21, 83, 255, 255, 45, 251, 193, 234, 26, 83, 255, 255, 46, 251, 193, 82, 56, 83, 255, 255, 47, 251, 193, 118, 224, 81, 255, 255, 33, 251, 193, 196, 245, 81, 255, 255, 34, 251, 193, 200, 0, 82, 255, 255, 35, 251, 193, 132, 155, 82, 255, 255, 42, 251, 193, 8, 160, 78, 34, 0, 107, 81, 16, 0, 107, 81, 255, 255, 28, 251, 193, 74, 130, 81, 255, 255, 29, 251, 193, 104, 150, 81, 255, 255, 30, 251, 193, 74, 171, 81, 255, 255, 31, 251, 193, 62, 160, 78, 255, 255, 8, 251, 193, 232, 186, 78, 255, 255, 10, 251, 193, 68, 63, 81, 255, 255, 26, 251, 193, 4, 101, 81, 255, 255, 27, 251, 193, 188, 63, 78, 16, 0, 63, 78, 255, 255, 4, 251, 193, 252, 89, 78, 255, 255, 6, 251, 193, 26, 133, 78, 255, 255, 7, 251, 193, 248, 140, 78, 255, 255, 8, 251, 193, 74, 0, 78, 255, 255, 2, 251, 193, 2, 40, 78, 255, 255, 4, 251, 193, 60, 44, 78, 255, 255, 66, 252, 193, 246, 54, 78, 255, 255, 4, 251, 193, 200, 26, 130, 246, 1, 0, 78, 1, 78, 40, 78, 41, 78, 44, 78, 45, 78, 54, 78, 55, 78, 63, 78, 64, 78, 89, 78, 90, 78, 133, 78, 134, 78, 140, 78, 141, 78, 160, 78, 161, 78, 186, 78, 187, 78, 63, 81, 64, 81, 101, 81, 102, 81, 107, 81, 108, 81, 130, 81, 131, 81, 150, 81, 151, 81, 171, 81, 172, 81, 224, 81, 225, 81, 245, 81, 246, 81, 0, 82, 1, 82, 155, 82, 156, 82, 249, 82, 250, 82, 21, 83, 22, 83, 26, 83, 27, 83, 56, 83, 57, 83, 65, 83, 66, 83, 92, 83, 93, 83, 100, 83, 101, 83, 105, 83, 106, 83, 130, 83, 131, 83, 182, 83, 183, 83, 200, 83, 201, 83, 227, 83, 228, 83, 215, 86, 216, 86, 31, 87, 32, 87, 235, 88, 236, 88, 2, 89, 3, 89, 10, 89, 11, 89, 21, 89, 22, 89, 39, 89, 40, 89, 115, 89, 116, 89, 80, 91, 81, 91, 128, 91, 129, 91, 248, 91, 249, 91, 15, 92, 16, 92, 34, 92, 35, 92, 56, 92, 57, 92, 110, 92, 111, 92, 113, 92, 114, 92, 219, 93, 220, 93, 229, 93, 230, 93, 241, 93, 242, 93, 254, 93, 255, 93, 114, 94, 115, 94, 122, 94, 123, 94, 127, 94, 128, 94, 244, 94, 245, 94, 254, 94, 255, 94, 11, 95, 12, 95, 19, 95, 20, 95, 80, 95, 81, 95, 97, 95, 98, 95, 115, 95, 116, 95, 195, 95, 196, 95, 8, 98, 9, 98, 54, 98, 55, 98, 75, 98, 76, 98, 47, 101, 48, 101, 52, 101, 53, 101, 135, 101, 136, 101, 151, 101, 152, 101, 164, 101, 165, 101, 185, 101, 186, 101, 224, 101, 225, 101, 229, 101, 230, 101, 240, 102, 241, 102, 8, 103, 9, 103, 40, 103, 41, 103, 32, 107, 33, 107, 98, 107, 99, 107, 121, 107, 122, 107, 179, 107, 180, 107, 203, 107, 204, 107, 212, 107, 213, 107, 219, 107, 220, 107, 15, 108, 16, 108, 20, 108, 21, 108, 52, 108, 53, 108, 107, 112, 108, 112, 42, 114, 43, 114, 54, 114, 55, 114, 59, 114, 60, 114, 63, 114, 64, 114, 71, 114, 72, 114, 89, 114, 90, 114, 91, 114, 92, 114, 172, 114, 173, 114, 132, 115, 133, 115, 137, 115, 138, 115, 220, 116, 221, 116, 230, 116, 231, 116, 24, 117, 25, 117, 31, 117, 32, 117, 40, 117, 41, 117, 48, 117, 49, 117, 139, 117, 140, 117, 146, 117, 147, 117, 118, 118, 119, 118, 125, 118, 126, 118, 174, 118, 175, 118, 191, 118, 192, 118, 238, 118, 239, 118, 219, 119, 220, 119, 226, 119, 227, 119, 243, 119, 244, 119, 58, 121, 59, 121, 184, 121, 185, 121, 190, 121, 191, 121, 116, 122, 117, 122, 203, 122, 204, 122, 249, 122, 250, 122, 115, 124, 116, 124, 248, 124, 249, 124, 159, 126, 160, 126, 54, 127, 55, 127, 81, 127, 82, 127, 138, 127, 139, 127, 189, 127, 190, 127, 1, 128, 2, 128, 12, 128, 13, 128, 18, 128, 19, 128, 51, 128, 52, 128, 127, 128, 128, 128, 137, 128, 138, 128, 227, 129, 228, 129, 234, 129, 235, 129, 243, 129, 244, 129, 252, 129, 253, 129, 12, 130, 13, 130, 27, 130, 28, 130, 31, 130, 32, 130, 110, 130, 111, 130, 114, 130, 115, 130, 120, 130, 121, 130, 77, 134, 78, 134, 107, 134, 108, 134, 64, 136, 65, 136, 76, 136, 77, 136, 99, 136, 100, 136, 126, 137, 127, 137, 139, 137, 140, 137, 193, 137, 194, 137, 210, 137, 211, 137, 0, 138, 1, 138, 160, 139, 161, 139, 55, 140, 56, 140, 70, 140, 71, 140, 85, 140, 86, 140, 120, 140, 121, 140, 157, 140, 158, 140, 29, 141, 30, 141, 100, 141, 101, 141, 112, 141, 113, 141, 179, 141, 180, 141, 171, 142, 172, 142, 202, 142, 203, 142, 102, 143, 103, 143, 155, 143, 156, 143, 176, 143, 177, 143, 181, 143, 182, 143, 145, 144, 146, 144, 73, 145, 74, 145, 198, 145, 199, 145, 204, 145, 205, 145, 209, 145, 210, 145, 133, 148, 134, 148, 119, 149, 120, 149, 127, 149, 129, 149, 232, 149, 233, 149, 28, 150, 29, 150, 182, 150, 183, 150, 185, 150, 186, 150, 232, 150, 233, 150, 81, 151, 82, 151, 94, 151, 95, 151, 98, 151, 99, 151, 105, 151, 106, 151, 203, 151, 204, 151, 230, 151, 231, 151, 237, 151, 238, 151, 243, 151, 244, 151, 1, 152, 2, 152, 117, 152, 118, 152, 168, 152, 169, 152, 206, 152, 207, 152, 219, 152, 220, 152, 222, 152, 224, 152, 99, 153, 100, 153, 150, 153, 151, 153, 153, 153, 154, 153, 172, 153, 173, 153, 108, 154, 109, 154, 168, 154, 169, 154, 216, 154, 217, 154, 223, 154, 224, 154, 37, 155, 38, 155, 47, 155, 48, 155, 50, 155, 51, 155, 60, 155, 61, 155, 90, 155, 91, 155, 124, 156, 125, 156, 229, 156, 230, 156, 31, 158, 32, 158, 117, 158, 118, 158, 127, 158, 128, 158, 165, 158, 167, 158, 187, 158, 188, 158, 195, 158, 196, 158, 205, 158, 206, 158, 209, 158, 210, 158, 249, 158, 250, 158, 253, 158, 255, 158, 14, 159, 15, 159, 19, 159, 20, 159, 32, 159, 33, 159, 59, 159, 60, 159, 74, 159, 75, 159, 80, 159, 81, 159, 82, 159, 83, 159, 127, 159, 128, 159, 141, 159, 142, 159, 153, 159, 154, 159, 156, 159, 157, 159, 159, 159, 161, 159, 126, 216, 127, 216, 2, 249, 3, 249, 7, 249, 9, 249, 10, 249, 11, 249, 52, 249, 53, 249, 64, 249, 65, 249, 113, 249, 114, 249, 129, 249, 130, 249, 138, 249, 139, 249, 196, 249, 197, 249, 233, 249, 234, 249, 247, 249, 248, 249, 8, 250, 9, 250, 10, 250, 11, 250, 30, 250, 31, 250, 60, 250, 61, 250, 149, 250, 150, 250, 206, 250, 207, 250, 2, 0, 29, 248, 2, 0, 30, 248, 2, 0, 117, 248, 2, 0, 118, 248, 2, 0, 120, 248, 2, 0, 121, 248, 2, 0, 144, 248, 2, 0, 145, 248, 2, 0, 196, 249, 2, 0, 197, 249, 2, 0, 210, 249, 2, 0, 211, 249, 2, 0, 21, 250, 2, 0, 22, 250, 2, 0, 23, 250, 2, 0, 25, 250, 2, 0, 28, 250, 2, 0, 29, 250, 3, 0, 82, 0, 64, 0, 32, 0, 43, 122, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 3, 0, 51, 0, 73, 0, 97, 0, 124, 45, 0, 80, 38, 122, 0, 32, 235, 142, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 240, 142, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 122, 0, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 126, 0, 0, 0, 126, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 71, 0, 101, 0, 111, 0, 114, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 60, 1, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 8, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 32, 0, 0, 0, 36, 0, 0, 0, 36, 1, 0, 0, 12, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 39, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 3, 0, 69, 0, 60, 0, 32, 0, 35, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 115, 0, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 120, 0, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 7, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 9, 7, 0, 0, 9, 7, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 49, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 48, 220, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 93, 0, 38, 0, 21, 4, 60, 0, 81, 4, 60, 0, 60, 0, 60, 0, 1, 4, 38, 0, 176, 4, 60, 0, 175, 4, 60, 0, 60, 0, 60, 0, 174, 4, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 76, 4, 60, 0, 86, 4, 60, 0, 60, 0, 60, 0, 6, 4, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 40, 27, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 15, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 60, 0, 0, 0, 64, 0, 0, 0, 64, 1, 0, 0, 160, 26, 0, 0, 160, 26, 0, 0, 160, 26, 0, 0, 160, 26, 0, 0, 244, 26, 0, 0, 244, 26, 0, 0, 16, 27, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 39, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 141, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 94, 0, 102, 0, 100, 0, 108, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 124, 0, 48, 0, 48, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 116, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 120, 1, 144, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 1, 0, 202, 129, 1, 0, 202, 162, 1, 0, 202, 195, 1, 0, 202, 228, 1, 0, 202, 5, 2, 0, 202, 38, 2, 0, 202, 71, 2, 0, 202, 104, 2, 0, 202, 137, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 134, 5, 41, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 231, 97, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 97, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 177, 97, 5, 5, 177, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 231, 97, 5, 150, 0, 0, 5, 5, 231, 97, 5, 150, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 40, 97, 174, 5, 48, 0, 8, 3, 255, 255, 41, 97, 134, 5, 40, 97, 5, 5, 48, 0, 8, 3, 255, 255, 41, 97, 5, 5, 3, 0, 69, 0, 60, 0, 32, 0, 51, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 254, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 3, 7, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 230, 22, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 234, 22, 0, 0, 234, 22, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 118, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 49, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 107, 0, 108, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 0, 0, 130, 220, 38, 0, 68, 0, 60, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 60, 0, 60, 0, 240, 0, 60, 0, 60, 0, 60, 0, 208, 0, 38, 0, 81, 0, 60, 0, 60, 0, 56, 1, 60, 0, 60, 0, 60, 0, 75, 0, 39, 0, 39, 0, 38, 0, 116, 0, 60, 0, 60, 0, 60, 0, 254, 0, 47, 0, 104, 0, 38, 0, 84, 0, 60, 0, 60, 0, 60, 0, 222, 0, 47, 0, 72, 0, 38, 0, 89, 0, 60, 0, 60, 0, 117, 0, 8, 3, 60, 0, 60, 0, 60, 0, 85, 0, 8, 3, 60, 0, 60, 0, 117, 0, 11, 3, 60, 0, 60, 0, 60, 0, 85, 0, 11, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 192, 1, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 60, 0, 101, 0, 40, 3, 60, 0, 60, 0, 60, 0, 69, 0, 40, 3, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 60, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 81, 1, 60, 0, 60, 0, 60, 0, 80, 1, 60, 0, 60, 0, 83, 1, 60, 0, 60, 0, 60, 0, 82, 1, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 0, 0, 170, 170, 170, 170, 170, 170, 238, 47, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 163, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 148, 38, 0, 0, 148, 38, 0, 0, 44, 41, 0, 0, 44, 41, 0, 0, 236, 44, 0, 0, 236, 44, 0, 0, 204, 47, 0, 0, 210, 47, 0, 0, 210, 47, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 92, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 95, 0, 103, 0, 108, 0, 116, 0, 121, 0, 129, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 148, 0, 156, 0, 162, 0, 170, 0, 178, 0, 186, 0, 48, 0, 56, 0, 185, 0, 193, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 210, 0, 218, 0, 48, 0, 48, 0, 48, 0, 226, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 234, 0, 48, 0, 48, 0, 242, 0, 48, 0, 248, 0, 0, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 3, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 11, 1, 48, 0, 48, 0, 15, 1, 48, 0, 23, 1, 28, 1, 36, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 42, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 50, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 53, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 208, 0, 48, 0, 48, 0, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 83, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 61, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 69, 1, 48, 0, 48, 0, 75, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 124, 1, 176, 1, 228, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 80, 2, 136, 2, 200, 2, 192, 0, 228, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 192, 28, 0, 202, 225, 28, 0, 202, 2, 29, 0, 202, 35, 29, 0, 202, 68, 29, 0, 202, 101, 29, 0, 202, 134, 29, 0, 202, 167, 29, 0, 202, 200, 29, 0, 202, 233, 29, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 198, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 166, 1, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 201, 96, 2, 0, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 70, 3, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 102, 4, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 6, 6, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 70, 7, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 6, 8, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 38, 9, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 1, 0, 0, 198, 33, 0, 0, 198, 33, 1, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 97, 0, 0, 5, 5, 127, 12, 198, 65, 1, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 34, 20, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 198, 161, 0, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 198, 225, 0, 0, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 10, 0, 201, 6, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 65, 0, 0, 198, 193, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 0, 74, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 129, 0, 0, 198, 1, 1, 0, 198, 33, 2, 0, 198, 65, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 38, 13, 0, 201, 70, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 15, 0, 201, 134, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 17, 0, 201, 198, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 192, 0, 0, 0, 198, 98, 3, 0, 198, 162, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 3, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 19, 0, 201, 6, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 5, 0, 197, 34, 5, 0, 198, 98, 5, 0, 197, 98, 5, 0, 198, 162, 5, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 5, 0, 198, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 38, 22, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 24, 0, 5, 71, 238, 102, 201, 230, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 6, 0, 192, 0, 0, 0, 197, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 7, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 197, 34, 8, 0, 197, 99, 8, 0, 197, 194, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 9, 0, 197, 67, 9, 0, 197, 162, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 226, 5, 0, 197, 34, 10, 0, 5, 5, 10, 121, 197, 98, 10, 0, 197, 162, 10, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 25, 0, 201, 230, 26, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 28, 0, 201, 38, 29, 0, 201, 70, 30, 0, 201, 102, 31, 0, 201, 134, 32, 0, 201, 166, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 34, 0, 201, 230, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 37, 0, 201, 38, 38, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 39, 0, 201, 102, 40, 0, 201, 134, 41, 0, 201, 166, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 43, 0, 201, 230, 44, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 96, 2, 0, 198, 33, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 71, 0, 2, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 71, 0, 2, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 3, 111, 93, 0, 5, 0, 72, 0, 3, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 4, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 193, 4, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 6, 121, 197, 226, 5, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 34, 6, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 98, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 226, 6, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 34, 7, 0, 5, 5, 107, 121, 197, 98, 7, 0, 5, 5, 111, 121, 197, 162, 7, 0, 5, 5, 115, 121, 197, 226, 7, 0, 5, 5, 119, 121, 197, 34, 8, 0, 197, 99, 8, 0, 197, 194, 8, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 2, 9, 0, 197, 67, 9, 0, 197, 162, 9, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 226, 9, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 226, 5, 0, 197, 34, 10, 0, 5, 5, 10, 121, 197, 98, 10, 0, 197, 162, 10, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 1, 0, 8, 3, 198, 129, 10, 3, 198, 161, 0, 50, 156, 5, 48, 0, 40, 3, 0, 192, 198, 65, 0, 62, 156, 5, 48, 0, 39, 0, 255, 255, 0, 74, 70, 70, 0, 70, 156, 5, 1, 0, 8, 3, 0, 192, 198, 97, 11, 3, 0, 192, 198, 129, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 1, 0, 8, 3, 0, 192, 198, 161, 10, 3, 255, 255, 111, 93, 193, 4, 0, 50, 5, 5, 48, 0, 40, 3, 0, 192, 198, 193, 0, 70, 5, 5, 1, 0, 8, 3, 0, 192, 198, 225, 11, 3, 1, 192, 198, 1, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 1, 192, 198, 98, 10, 3, 1, 192, 198, 162, 5, 42, 196, 162, 1, 0, 8, 3, 1, 192, 198, 226, 10, 3, 1, 192, 197, 34, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 98, 11, 3, 1, 192, 197, 162, 5, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 226, 11, 3, 2, 192, 197, 34, 156, 70, 196, 190, 1, 0, 8, 3, 2, 192, 198, 98, 11, 3, 2, 192, 198, 162, 5, 70, 196, 190, 1, 0, 8, 3, 2, 192, 198, 226, 11, 3, 3, 192, 198, 34, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 98, 11, 3, 2, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 156, 70, 196, 162, 1, 0, 8, 3, 4, 192, 198, 34, 11, 3, 4, 192, 198, 98, 5, 70, 196, 162, 1, 0, 8, 3, 4, 192, 198, 162, 11, 3, 4, 192, 198, 226, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 1, 0, 8, 3, 6, 192, 198, 98, 10, 3, 6, 192, 198, 162, 5, 42, 196, 200, 1, 0, 8, 3, 6, 192, 198, 226, 10, 3, 12, 192, 197, 162, 156, 82, 196, 198, 1, 0, 8, 3, 12, 192, 197, 226, 11, 3, 13, 192, 197, 34, 5, 82, 196, 198, 1, 0, 8, 3, 13, 192, 197, 98, 11, 3, 13, 192, 197, 162, 156, 82, 196, 208, 1, 0, 8, 3, 13, 192, 197, 226, 11, 3, 14, 192, 197, 34, 5, 82, 196, 208, 1, 0, 8, 3, 14, 192, 197, 98, 11, 3, 14, 192, 197, 162, 156, 82, 196, 204, 1, 0, 8, 3, 14, 192, 197, 226, 11, 3, 15, 192, 197, 34, 5, 82, 196, 204, 1, 0, 8, 3, 15, 192, 197, 98, 11, 3, 15, 192, 197, 162, 156, 42, 196, 196, 1, 0, 8, 3, 7, 192, 198, 34, 10, 3, 7, 192, 198, 98, 5, 42, 196, 196, 1, 0, 8, 3, 7, 192, 198, 162, 10, 3, 15, 192, 197, 226, 156, 70, 196, 196, 1, 0, 8, 3, 7, 192, 198, 226, 11, 3, 8, 192, 198, 34, 5, 70, 196, 196, 1, 0, 8, 3, 8, 192, 198, 98, 11, 3, 8, 192, 198, 162, 16, 0, 197, 35, 1, 0, 8, 3, 8, 192, 198, 227, 11, 3, 9, 192, 198, 67, 16, 0, 197, 131, 1, 0, 8, 3, 9, 192, 198, 163, 11, 3, 10, 192, 198, 3, 156, 82, 196, 196, 1, 0, 8, 3, 16, 192, 197, 226, 11, 3, 17, 192, 197, 34, 5, 82, 196, 196, 1, 0, 8, 3, 17, 192, 197, 98, 11, 3, 17, 192, 197, 162, 17, 0, 197, 227, 1, 0, 8, 3, 18, 192, 197, 67, 11, 3, 18, 192, 197, 163, 19, 0, 197, 3, 1, 0, 8, 3, 19, 192, 197, 99, 11, 3, 19, 192, 197, 195, 2, 0, 39, 0, 40, 0, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 119, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 132, 41, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 100, 30, 0, 0, 100, 30, 0, 0, 252, 32, 0, 0, 252, 32, 0, 0, 216, 34, 0, 0, 216, 34, 0, 0, 114, 37, 0, 0, 120, 37, 0, 0, 104, 41, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 217, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 95, 0, 103, 0, 108, 0, 116, 0, 121, 0, 129, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 169, 0, 48, 0, 48, 0, 173, 0, 48, 0, 181, 0, 186, 0, 194, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 200, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 208, 0, 161, 0, 48, 0, 48, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 124, 1, 176, 1, 228, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 13, 0, 202, 193, 13, 0, 202, 226, 13, 0, 202, 3, 14, 0, 202, 36, 14, 0, 202, 69, 14, 0, 202, 102, 14, 0, 202, 135, 14, 0, 202, 168, 14, 0, 202, 201, 14, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 230, 0, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 201, 160, 1, 0, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 134, 2, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 166, 3, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 70, 5, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 134, 6, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 70, 7, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 102, 8, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 1, 0, 0, 198, 33, 0, 0, 198, 33, 1, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 97, 0, 0, 5, 5, 127, 12, 198, 65, 1, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 98, 13, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 198, 161, 0, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 198, 225, 0, 0, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 10, 0, 201, 70, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 65, 0, 0, 198, 193, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 0, 74, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 129, 0, 0, 198, 1, 1, 0, 198, 33, 2, 0, 198, 65, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 102, 12, 0, 201, 134, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 14, 0, 201, 198, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 16, 0, 201, 6, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 192, 0, 0, 0, 198, 98, 3, 0, 198, 162, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 3, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 19, 0, 201, 70, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 5, 0, 197, 34, 5, 0, 198, 98, 5, 0, 197, 98, 5, 0, 198, 162, 5, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 5, 0, 198, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 21, 0, 201, 134, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 23, 0, 201, 198, 24, 0, 201, 230, 25, 0, 201, 6, 27, 0, 201, 38, 28, 0, 201, 70, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 30, 0, 201, 134, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 32, 0, 201, 198, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 34, 0, 201, 6, 36, 0, 201, 38, 37, 0, 201, 70, 38, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 39, 0, 201, 134, 40, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 160, 1, 0, 198, 33, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 71, 0, 2, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 71, 0, 2, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 3, 111, 93, 0, 5, 0, 72, 0, 3, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 4, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 1, 0, 8, 3, 198, 129, 10, 3, 198, 161, 0, 50, 156, 5, 48, 0, 40, 3, 0, 192, 198, 65, 0, 62, 156, 5, 48, 0, 39, 0, 255, 255, 0, 74, 70, 70, 0, 70, 156, 5, 1, 0, 8, 3, 0, 192, 198, 97, 11, 3, 0, 192, 198, 129, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 1, 0, 8, 3, 0, 192, 198, 161, 10, 3, 255, 255, 111, 93, 193, 4, 0, 50, 5, 5, 48, 0, 40, 3, 0, 192, 198, 193, 0, 70, 5, 5, 1, 0, 8, 3, 0, 192, 198, 225, 11, 3, 1, 192, 198, 1, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 1, 192, 198, 98, 10, 3, 1, 192, 198, 162, 5, 42, 196, 162, 1, 0, 8, 3, 1, 192, 198, 226, 10, 3, 1, 192, 197, 34, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 98, 11, 3, 1, 192, 197, 162, 5, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 226, 11, 3, 2, 192, 197, 34, 156, 70, 196, 190, 1, 0, 8, 3, 2, 192, 198, 98, 11, 3, 2, 192, 198, 162, 5, 70, 196, 190, 1, 0, 8, 3, 2, 192, 198, 226, 11, 3, 3, 192, 198, 34, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 98, 11, 3, 2, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 156, 70, 196, 162, 1, 0, 8, 3, 4, 192, 198, 34, 11, 3, 4, 192, 198, 98, 5, 70, 196, 162, 1, 0, 8, 3, 4, 192, 198, 162, 11, 3, 4, 192, 198, 226, 156, 42, 196, 200, 1, 0, 8, 3, 6, 192, 198, 98, 10, 3, 6, 192, 198, 162, 5, 42, 196, 200, 1, 0, 8, 3, 6, 192, 198, 226, 10, 3, 5, 192, 197, 226, 156, 82, 196, 198, 1, 0, 8, 3, 6, 192, 197, 34, 11, 3, 6, 192, 197, 98, 5, 82, 196, 198, 1, 0, 8, 3, 6, 192, 197, 162, 11, 3, 6, 192, 197, 226, 156, 82, 196, 208, 1, 0, 8, 3, 7, 192, 197, 34, 11, 3, 7, 192, 197, 98, 5, 82, 196, 208, 1, 0, 8, 3, 7, 192, 197, 162, 11, 3, 7, 192, 197, 226, 156, 82, 196, 204, 1, 0, 8, 3, 8, 192, 197, 34, 11, 3, 8, 192, 197, 98, 5, 82, 196, 204, 1, 0, 8, 3, 8, 192, 197, 162, 11, 3, 8, 192, 197, 226, 156, 42, 196, 196, 1, 0, 8, 3, 7, 192, 198, 34, 10, 3, 7, 192, 198, 98, 5, 42, 196, 196, 1, 0, 8, 3, 7, 192, 198, 162, 10, 3, 9, 192, 197, 34, 156, 70, 196, 196, 1, 0, 8, 3, 7, 192, 198, 226, 11, 3, 8, 192, 198, 34, 5, 70, 196, 196, 1, 0, 8, 3, 8, 192, 198, 98, 11, 3, 8, 192, 198, 162, 9, 0, 197, 99, 1, 0, 8, 3, 8, 192, 198, 227, 11, 3, 9, 192, 198, 67, 9, 0, 197, 195, 1, 0, 8, 3, 9, 192, 198, 163, 11, 3, 10, 192, 198, 3, 156, 82, 196, 196, 1, 0, 8, 3, 10, 192, 197, 34, 11, 3, 10, 192, 197, 98, 5, 82, 196, 196, 1, 0, 8, 3, 10, 192, 197, 162, 11, 3, 10, 192, 197, 226, 11, 0, 197, 35, 1, 0, 8, 3, 11, 192, 197, 131, 11, 3, 11, 192, 197, 227, 12, 0, 197, 67, 1, 0, 8, 3, 12, 192, 197, 163, 11, 3, 13, 192, 197, 3, 2, 0, 39, 0, 40, 0, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 20, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 22, 4, 185, 60, 185, 64, 185, 68, 24, 4, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 26, 4, 185, 104, 185, 108, 186, 112, 30, 4, 185, 124, 185, 128, 185, 132, 186, 136, 186, 140, 32, 4, 185, 152, 185, 156, 185, 160, 185, 164, 185, 168, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 34, 4, 168, 60, 168, 64, 168, 68, 36, 4, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 38, 4, 168, 124, 168, 128, 168, 132, 168, 136, 168, 140, 40, 4, 168, 152, 168, 156, 168, 160, 168, 164, 168, 168, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 121, 58, 217, 176, 185, 184, 185, 176, 185, 66, 185, 73, 153, 73, 249, 73, 89, 74, 186, 89, 154, 89, 250, 89, 90, 90, 249, 68, 122, 114, 185, 121, 153, 121, 249, 121, 121, 122, 217, 180, 136, 14, 185, 180, 185, 149, 153, 149, 249, 149, 217, 164, 153, 165, 0, 8, 2, 8, 168, 57, 136, 57, 232, 57, 104, 58, 200, 176, 168, 184, 168, 176, 168, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 232, 68, 104, 114, 168, 121, 136, 121, 232, 121, 104, 122, 200, 180, 128, 14, 168, 180, 168, 149, 136, 149, 232, 149, 200, 164, 136, 165, 4, 8, 72, 166, 249, 58, 232, 58, 217, 57, 200, 57, 42, 4, 44, 4, 153, 65, 136, 65, 249, 65, 232, 65, 153, 66, 136, 66, 25, 66, 8, 66, 25, 70, 8, 70, 217, 68, 200, 68, 249, 74, 232, 74, 217, 73, 200, 73, 153, 74, 136, 74, 249, 176, 232, 176, 25, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 153, 82, 136, 82, 185, 82, 168, 82, 249, 85, 232, 85, 25, 87, 8, 87, 122, 90, 104, 90, 250, 90, 232, 90, 218, 89, 200, 89, 218, 90, 200, 90, 154, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 185, 102, 168, 102, 200, 128, 153, 105, 136, 105, 185, 106, 168, 106, 25, 106, 8, 106, 57, 107, 40, 107, 25, 107, 8, 107, 154, 113, 136, 113, 186, 114, 168, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 249, 122, 232, 122, 217, 121, 200, 121, 249, 180, 232, 180, 25, 181, 8, 181, 153, 133, 136, 133, 185, 134, 168, 134, 25, 134, 8, 134, 154, 137, 136, 137, 250, 137, 232, 137, 186, 138, 168, 138, 26, 138, 8, 138, 186, 142, 168, 142, 26, 142, 8, 142, 185, 144, 168, 144, 121, 150, 104, 150, 249, 150, 232, 150, 217, 149, 200, 149, 57, 150, 40, 150, 249, 164, 232, 164, 46, 4, 48, 4, 249, 157, 232, 157, 249, 165, 232, 165, 89, 166, 153, 169, 136, 169, 153, 170, 136, 170, 25, 170, 8, 170, 232, 136, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 12, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 14, 8, 1, 0, 8, 14, 16, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 18, 8, 200, 12, 72, 12, 32, 14, 187, 140, 185, 84, 200, 136, 169, 136, 169, 140, 168, 84, 187, 88, 187, 96, 169, 88, 169, 96, 169, 172, 169, 112, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 72, 255, 5, 185, 100, 39, 4, 209, 128, 255, 5, 185, 120, 255, 5, 185, 148, 255, 5, 168, 56, 255, 5, 168, 72, 255, 5, 168, 120, 255, 5, 168, 148, 255, 5, 217, 58, 255, 5, 200, 58, 255, 5, 217, 150, 255, 5, 200, 150, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 123, 12, 0, 16, 57, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 116, 12, 0, 32, 221, 22, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 226, 22, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 230, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 234, 8, 0, 0, 234, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 34, 1, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 18, 222, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 75, 0, 104, 0, 109, 0, 114, 0, 93, 0, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 116, 0, 101, 0, 114, 0, 116, 0, 105, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 61, 0, 180, 23, 61, 0, 181, 23, 38, 0, 212, 23, 155, 23, 212, 23, 60, 0, 60, 0, 60, 0, 216, 23, 38, 0, 200, 23, 60, 0, 60, 0, 206, 23, 60, 0, 60, 0, 207, 23, 60, 0, 60, 0, 209, 23, 60, 0, 60, 0, 208, 23, 60, 0, 60, 0, 200, 23, 60, 0, 60, 0, 221, 23, 60, 0, 60, 0, 203, 23, 60, 0, 60, 0, 201, 23, 60, 0, 60, 0, 202, 23, 60, 0, 60, 0, 205, 23, 38, 0, 154, 23, 60, 0, 171, 23, 60, 0, 172, 23, 38, 0, 155, 23, 60, 0, 173, 23, 60, 0, 174, 23, 38, 0, 154, 23, 210, 23, 128, 23, 60, 0, 60, 0, 128, 23, 204, 23, 38, 0, 154, 23, 210, 23, 129, 23, 60, 0, 60, 0, 129, 23, 204, 23, 38, 0, 154, 23, 210, 23, 130, 23, 60, 0, 60, 0, 130, 23, 204, 23, 38, 0, 154, 23, 210, 23, 131, 23, 60, 0, 60, 0, 131, 23, 204, 23, 38, 0, 154, 23, 210, 23, 132, 23, 60, 0, 60, 0, 132, 23, 204, 23, 38, 0, 154, 23, 210, 23, 133, 23, 60, 0, 60, 0, 133, 23, 204, 23, 38, 0, 154, 23, 210, 23, 134, 23, 60, 0, 60, 0, 134, 23, 204, 23, 38, 0, 154, 23, 210, 23, 135, 23, 60, 0, 60, 0, 135, 23, 204, 23, 38, 0, 154, 23, 210, 23, 136, 23, 60, 0, 60, 0, 136, 23, 204, 23, 38, 0, 154, 23, 210, 23, 137, 23, 60, 0, 60, 0, 137, 23, 204, 23, 38, 0, 154, 23, 210, 23, 138, 23, 60, 0, 60, 0, 138, 23, 204, 23, 38, 0, 154, 23, 210, 23, 139, 23, 60, 0, 60, 0, 139, 23, 204, 23, 38, 0, 154, 23, 210, 23, 140, 23, 60, 0, 60, 0, 140, 23, 204, 23, 38, 0, 154, 23, 210, 23, 141, 23, 60, 0, 60, 0, 141, 23, 204, 23, 38, 0, 154, 23, 210, 23, 142, 23, 60, 0, 60, 0, 142, 23, 204, 23, 38, 0, 154, 23, 210, 23, 143, 23, 60, 0, 60, 0, 143, 23, 204, 23, 38, 0, 154, 23, 210, 23, 144, 23, 60, 0, 60, 0, 144, 23, 204, 23, 38, 0, 154, 23, 210, 23, 145, 23, 60, 0, 60, 0, 145, 23, 204, 23, 38, 0, 154, 23, 210, 23, 146, 23, 60, 0, 60, 0, 146, 23, 204, 23, 38, 0, 154, 23, 210, 23, 147, 23, 60, 0, 60, 0, 147, 23, 204, 23, 38, 0, 154, 23, 210, 23, 148, 23, 60, 0, 60, 0, 148, 23, 204, 23, 38, 0, 154, 23, 210, 23, 149, 23, 60, 0, 60, 0, 149, 23, 204, 23, 38, 0, 154, 23, 210, 23, 150, 23, 60, 0, 60, 0, 150, 23, 204, 23, 38, 0, 154, 23, 210, 23, 151, 23, 60, 0, 60, 0, 151, 23, 204, 23, 38, 0, 154, 23, 210, 23, 152, 23, 60, 0, 60, 0, 152, 23, 204, 23, 38, 0, 154, 23, 210, 23, 153, 23, 60, 0, 60, 0, 153, 23, 204, 23, 38, 0, 154, 23, 210, 23, 154, 23, 60, 0, 60, 0, 154, 23, 204, 23, 38, 0, 154, 23, 210, 23, 171, 23, 60, 0, 60, 0, 171, 23, 204, 23, 38, 0, 154, 23, 210, 23, 172, 23, 60, 0, 60, 0, 172, 23, 204, 23, 38, 0, 154, 23, 210, 23, 155, 23, 60, 0, 60, 0, 155, 23, 204, 23, 38, 0, 154, 23, 210, 23, 173, 23, 60, 0, 60, 0, 173, 23, 204, 23, 38, 0, 154, 23, 210, 23, 174, 23, 60, 0, 60, 0, 174, 23, 204, 23, 38, 0, 154, 23, 210, 23, 156, 23, 60, 0, 60, 0, 156, 23, 204, 23, 38, 0, 154, 23, 210, 23, 157, 23, 60, 0, 60, 0, 157, 23, 204, 23, 38, 0, 154, 23, 210, 23, 158, 23, 60, 0, 60, 0, 158, 23, 204, 23, 38, 0, 154, 23, 210, 23, 159, 23, 60, 0, 60, 0, 159, 23, 204, 23, 38, 0, 154, 23, 210, 23, 160, 23, 60, 0, 60, 0, 160, 23, 204, 23, 38, 0, 154, 23, 210, 23, 161, 23, 60, 0, 60, 0, 161, 23, 204, 23, 38, 0, 154, 23, 210, 23, 162, 23, 60, 0, 60, 0, 162, 23, 204, 23, 61, 0, 163, 23, 204, 23, 38, 0, 162, 23, 61, 0, 163, 23, 38, 0, 162, 23, 182, 23, 61, 0, 164, 23, 38, 0, 162, 23, 183, 23, 60, 0, 60, 0, 165, 23, 38, 0, 162, 23, 184, 23, 60, 0, 60, 0, 166, 23, 38, 0, 162, 23, 187, 23, 60, 0, 60, 0, 167, 23, 38, 0, 162, 23, 188, 23, 60, 0, 60, 0, 169, 23, 38, 0, 162, 23, 194, 23, 60, 0, 60, 0, 175, 23, 38, 0, 162, 23, 195, 23, 60, 0, 60, 0, 176, 23, 38, 0, 162, 23, 196, 23, 60, 0, 60, 0, 177, 23, 38, 0, 162, 23, 197, 23, 60, 0, 60, 0, 179, 23, 38, 0, 167, 23, 128, 23, 60, 0, 60, 0, 60, 0, 168, 23, 38, 0, 169, 23, 156, 23, 60, 0, 60, 0, 60, 0, 170, 23, 38, 0, 177, 23, 60, 0, 60, 0, 60, 0, 178, 23, 38, 0, 197, 23, 60, 0, 187, 23, 198, 23, 60, 0, 198, 23, 60, 0, 182, 23, 198, 23, 60, 0, 199, 23, 60, 0, 183, 23, 199, 23, 60, 0, 187, 23, 199, 23, 60, 0, 193, 23, 199, 23, 60, 0, 196, 23, 199, 23, 0, 0, 170, 170, 170, 170, 236, 30, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 16, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 76, 0, 0, 0, 76, 1, 0, 0, 236, 25, 0, 0, 236, 25, 0, 0, 76, 26, 0, 0, 76, 26, 0, 0, 168, 28, 0, 0, 168, 28, 0, 0, 198, 30, 0, 0, 208, 30, 0, 0, 23, 0, 0, 0, 0, 0, 98, 117, 178, 255, 0, 118, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 0, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 129, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 104, 0, 112, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 120, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 17, 0, 202, 193, 17, 0, 202, 226, 17, 0, 202, 3, 18, 0, 202, 36, 18, 0, 202, 69, 18, 0, 202, 102, 18, 0, 202, 135, 18, 0, 202, 168, 18, 0, 202, 201, 18, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 16, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 17, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 17, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 0, 0, 0, 201, 160, 0, 0, 201, 96, 1, 0, 201, 32, 2, 0, 201, 224, 2, 0, 201, 160, 3, 0, 201, 96, 4, 0, 201, 32, 5, 0, 201, 224, 5, 0, 201, 160, 6, 0, 201, 96, 7, 0, 201, 32, 8, 0, 201, 224, 8, 0, 201, 160, 9, 0, 201, 96, 10, 0, 201, 32, 11, 0, 201, 224, 11, 0, 201, 160, 12, 0, 201, 96, 13, 0, 201, 32, 14, 0, 201, 224, 14, 0, 201, 160, 15, 0, 201, 96, 16, 0, 201, 32, 17, 0, 201, 224, 17, 0, 201, 160, 18, 0, 201, 96, 19, 0, 201, 32, 20, 0, 201, 224, 20, 0, 201, 160, 21, 0, 201, 96, 22, 0, 201, 32, 23, 0, 201, 224, 23, 0, 201, 160, 24, 0, 201, 96, 25, 0, 201, 96, 25, 0, 197, 66, 13, 0, 197, 130, 13, 0, 197, 194, 13, 0, 197, 2, 14, 0, 197, 67, 14, 0, 197, 162, 14, 0, 197, 227, 14, 0, 201, 32, 26, 0, 201, 192, 26, 0, 201, 128, 27, 0, 201, 64, 28, 0, 197, 66, 15, 0, 197, 130, 15, 0, 197, 194, 15, 0, 197, 2, 16, 0, 197, 66, 16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 201, 0, 29, 0, 201, 224, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 192, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 32, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 0, 33, 0, 192, 0, 0, 0, 193, 3, 237, 117, 193, 5, 237, 117, 194, 5, 164, 245, 194, 5, 167, 245, 194, 5, 168, 245, 194, 5, 166, 245, 192, 0, 0, 0, 194, 5, 169, 245, 194, 5, 160, 245, 194, 5, 161, 245, 194, 5, 163, 245, 194, 5, 162, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 165, 245, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 5, 0, 5, 0, 0, 152, 117, 0, 5, 0, 5, 0, 0, 238, 117, 0, 5, 0, 70, 0, 2, 153, 117, 0, 5, 0, 5, 0, 0, 152, 117, 0, 5, 0, 5, 0, 0, 238, 117, 0, 5, 0, 70, 0, 3, 153, 117, 0, 5, 0, 5, 0, 0, 152, 117, 0, 5, 0, 5, 0, 0, 238, 117, 0, 5, 0, 70, 0, 2, 155, 117, 0, 5, 0, 5, 0, 0, 152, 117, 0, 5, 0, 5, 0, 0, 238, 117, 0, 5, 0, 70, 0, 3, 155, 117, 0, 0, 0, 0, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 100, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 102, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 104, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 106, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 108, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 110, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 112, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 114, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 116, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 118, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 120, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 122, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 124, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 126, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 128, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 130, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 132, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 134, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 136, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 138, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 140, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 142, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 144, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 146, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 148, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 150, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 152, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 154, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 156, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 158, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 160, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 162, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 164, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 166, 117, 5, 5, 152, 117, 5, 5, 238, 117, 5, 70, 168, 117, 5, 5, 168, 117, 5, 5, 206, 117, 5, 5, 168, 117, 5, 70, 208, 117, 5, 5, 168, 117, 5, 70, 210, 117, 5, 5, 168, 117, 5, 70, 216, 117, 5, 5, 168, 117, 5, 70, 216, 117, 6, 5, 100, 117, 5, 5, 168, 117, 5, 70, 218, 117, 5, 5, 168, 117, 5, 70, 218, 117, 6, 5, 156, 117, 5, 5, 168, 117, 5, 70, 230, 117, 5, 5, 168, 117, 5, 70, 232, 117, 5, 5, 168, 117, 5, 70, 234, 117, 5, 5, 168, 117, 6, 70, 234, 117, 5, 5, 168, 117, 5, 70, 236, 117, 5, 5, 48, 9, 5, 5, 154, 117, 6, 5, 48, 9, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 100, 117, 5, 5, 48, 0, 204, 23, 197, 163, 102, 117, 5, 5, 48, 0, 204, 23, 0, 192, 197, 131, 104, 117, 5, 5, 48, 0, 204, 23, 0, 192, 197, 227, 106, 117, 5, 5, 48, 0, 204, 23, 1, 192, 197, 67, 108, 117, 5, 5, 48, 0, 204, 23, 1, 192, 197, 163, 110, 117, 5, 5, 48, 0, 204, 23, 2, 192, 197, 3, 112, 117, 5, 5, 48, 0, 204, 23, 2, 192, 197, 99, 114, 117, 5, 5, 48, 0, 204, 23, 2, 192, 197, 195, 116, 117, 5, 5, 48, 0, 204, 23, 3, 192, 197, 35, 118, 117, 5, 5, 48, 0, 204, 23, 3, 192, 197, 131, 120, 117, 5, 5, 48, 0, 204, 23, 3, 192, 197, 227, 122, 117, 5, 5, 48, 0, 204, 23, 4, 192, 197, 67, 124, 117, 5, 5, 48, 0, 204, 23, 4, 192, 197, 163, 126, 117, 5, 5, 48, 0, 204, 23, 5, 192, 197, 3, 128, 117, 5, 5, 48, 0, 204, 23, 5, 192, 197, 99, 130, 117, 5, 5, 48, 0, 204, 23, 5, 192, 197, 195, 132, 117, 5, 5, 48, 0, 204, 23, 6, 192, 197, 35, 134, 117, 5, 5, 48, 0, 204, 23, 6, 192, 197, 131, 136, 117, 5, 5, 48, 0, 204, 23, 6, 192, 197, 227, 138, 117, 5, 5, 48, 0, 204, 23, 7, 192, 197, 67, 140, 117, 5, 5, 48, 0, 204, 23, 7, 192, 197, 163, 142, 117, 5, 5, 48, 0, 204, 23, 8, 192, 197, 3, 144, 117, 5, 5, 48, 0, 204, 23, 8, 192, 197, 99, 146, 117, 5, 5, 48, 0, 204, 23, 8, 192, 197, 195, 148, 117, 5, 5, 48, 0, 204, 23, 9, 192, 197, 35, 150, 117, 5, 5, 48, 0, 204, 23, 9, 192, 197, 131, 152, 117, 5, 5, 48, 0, 204, 23, 9, 192, 197, 227, 154, 117, 5, 5, 48, 0, 204, 23, 10, 192, 197, 67, 156, 117, 5, 5, 48, 0, 204, 23, 10, 192, 197, 163, 158, 117, 5, 5, 48, 0, 204, 23, 11, 192, 197, 3, 160, 117, 5, 5, 48, 0, 204, 23, 11, 192, 197, 99, 162, 117, 5, 5, 48, 0, 204, 23, 11, 192, 197, 195, 164, 117, 5, 5, 48, 0, 204, 23, 12, 192, 197, 35, 166, 117, 5, 5, 48, 0, 204, 23, 12, 192, 197, 131, 168, 117, 5, 5, 48, 0, 204, 23, 12, 192, 197, 227, 153, 117, 193, 2, 48, 0, 204, 23, 198, 131, 153, 117, 193, 3, 48, 0, 204, 23, 0, 192, 198, 99, 155, 117, 193, 2, 48, 0, 204, 23, 0, 192, 198, 195, 155, 117, 193, 3, 48, 0, 204, 23, 1, 192, 198, 35, 206, 117, 5, 5, 48, 0, 198, 23, 255, 255, 237, 117, 193, 4, 208, 117, 5, 5, 48, 0, 199, 23, 255, 255, 237, 117, 193, 6, 216, 117, 5, 5, 1, 0, 198, 23, 255, 255, 237, 117, 193, 2, 199, 23, 255, 255, 237, 117, 193, 7, 228, 117, 5, 5, 48, 0, 199, 23, 255, 255, 237, 117, 193, 8, 234, 117, 5, 5, 48, 0, 199, 23, 255, 255, 237, 117, 193, 9, 4, 0, 198, 23, 200, 23, 204, 23, 205, 23, 3, 0, 69, 0, 60, 0, 32, 0, 35, 1, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 223, 8, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 228, 8, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 117, 33, 0, 32, 7, 0, 0, 0, 24, 0, 0, 0, 121, 33, 0, 0, 121, 33, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 119, 9, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 116, 114, 97, 100, 105, 116, 105, 111, 110, 97, 108, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 87, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 148, 12, 60, 0, 130, 12, 60, 0, 131, 12, 60, 0, 241, 12, 60, 0, 242, 12, 0, 0, 239, 223, 92, 18, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 148, 12, 60, 0, 130, 12, 60, 0, 131, 12, 60, 0, 241, 12, 60, 0, 242, 12, 60, 0, 149, 12, 205, 12, 60, 0, 150, 12, 205, 12, 60, 0, 151, 12, 205, 12, 60, 0, 152, 12, 205, 12, 60, 0, 153, 12, 205, 12, 60, 0, 154, 12, 205, 12, 60, 0, 155, 12, 205, 12, 60, 0, 156, 12, 205, 12, 60, 0, 157, 12, 205, 12, 60, 0, 158, 12, 205, 12, 60, 0, 159, 12, 205, 12, 60, 0, 160, 12, 205, 12, 60, 0, 161, 12, 205, 12, 60, 0, 162, 12, 205, 12, 60, 0, 163, 12, 205, 12, 60, 0, 164, 12, 205, 12, 60, 0, 165, 12, 205, 12, 60, 0, 166, 12, 205, 12, 60, 0, 167, 12, 205, 12, 60, 0, 168, 12, 205, 12, 60, 0, 170, 12, 205, 12, 60, 0, 171, 12, 205, 12, 60, 0, 172, 12, 205, 12, 60, 0, 173, 12, 205, 12, 60, 0, 174, 12, 205, 12, 60, 0, 175, 12, 205, 12, 60, 0, 176, 12, 205, 12, 60, 0, 177, 12, 205, 12, 60, 0, 178, 12, 205, 12, 60, 0, 181, 12, 205, 12, 60, 0, 182, 12, 205, 12, 60, 0, 183, 12, 205, 12, 60, 0, 184, 12, 205, 12, 60, 0, 185, 12, 205, 12, 60, 0, 179, 12, 205, 12, 60, 0, 222, 12, 205, 12, 38, 0, 149, 12, 205, 12, 133, 12, 61, 0, 149, 12, 38, 0, 149, 12, 205, 12, 134, 12, 61, 0, 149, 12, 190, 12, 38, 0, 149, 12, 205, 12, 135, 12, 61, 0, 149, 12, 191, 12, 38, 0, 149, 12, 205, 12, 136, 12, 61, 0, 149, 12, 192, 12, 38, 0, 149, 12, 205, 12, 137, 12, 61, 0, 149, 12, 193, 12, 38, 0, 149, 12, 205, 12, 138, 12, 61, 0, 149, 12, 194, 12, 38, 0, 149, 12, 205, 12, 139, 12, 61, 0, 149, 12, 195, 12, 38, 0, 149, 12, 205, 12, 224, 12, 61, 0, 149, 12, 196, 12, 38, 0, 149, 12, 205, 12, 140, 12, 61, 0, 149, 12, 226, 12, 38, 0, 149, 12, 205, 12, 225, 12, 61, 0, 149, 12, 227, 12, 38, 0, 149, 12, 205, 12, 142, 12, 61, 0, 149, 12, 198, 12, 38, 0, 149, 12, 205, 12, 143, 12, 61, 0, 149, 12, 199, 12, 38, 0, 149, 12, 205, 12, 144, 12, 61, 0, 149, 12, 200, 12, 38, 0, 149, 12, 205, 12, 146, 12, 61, 0, 149, 12, 202, 12, 38, 0, 149, 12, 205, 12, 147, 12, 61, 0, 149, 12, 203, 12, 38, 0, 149, 12, 205, 12, 148, 12, 61, 0, 149, 12, 204, 12, 38, 0, 150, 12, 205, 12, 133, 12, 61, 0, 150, 12, 38, 0, 150, 12, 205, 12, 134, 12, 61, 0, 150, 12, 190, 12, 38, 0, 150, 12, 205, 12, 135, 12, 61, 0, 150, 12, 191, 12, 38, 0, 150, 12, 205, 12, 136, 12, 61, 0, 150, 12, 192, 12, 38, 0, 150, 12, 205, 12, 137, 12, 61, 0, 150, 12, 193, 12, 38, 0, 150, 12, 205, 12, 138, 12, 61, 0, 150, 12, 194, 12, 38, 0, 150, 12, 205, 12, 139, 12, 61, 0, 150, 12, 195, 12, 38, 0, 150, 12, 205, 12, 224, 12, 61, 0, 150, 12, 196, 12, 38, 0, 150, 12, 205, 12, 140, 12, 61, 0, 150, 12, 226, 12, 38, 0, 150, 12, 205, 12, 225, 12, 61, 0, 150, 12, 227, 12, 38, 0, 150, 12, 205, 12, 142, 12, 61, 0, 150, 12, 198, 12, 38, 0, 150, 12, 205, 12, 143, 12, 61, 0, 150, 12, 199, 12, 38, 0, 150, 12, 205, 12, 144, 12, 61, 0, 150, 12, 200, 12, 38, 0, 150, 12, 205, 12, 146, 12, 61, 0, 150, 12, 202, 12, 38, 0, 150, 12, 205, 12, 147, 12, 61, 0, 150, 12, 203, 12, 38, 0, 150, 12, 205, 12, 148, 12, 61, 0, 150, 12, 204, 12, 38, 0, 151, 12, 205, 12, 133, 12, 61, 0, 151, 12, 38, 0, 151, 12, 205, 12, 134, 12, 61, 0, 151, 12, 190, 12, 38, 0, 151, 12, 205, 12, 135, 12, 61, 0, 151, 12, 191, 12, 38, 0, 151, 12, 205, 12, 136, 12, 61, 0, 151, 12, 192, 12, 38, 0, 151, 12, 205, 12, 137, 12, 61, 0, 151, 12, 193, 12, 38, 0, 151, 12, 205, 12, 138, 12, 61, 0, 151, 12, 194, 12, 38, 0, 151, 12, 205, 12, 139, 12, 61, 0, 151, 12, 195, 12, 38, 0, 151, 12, 205, 12, 224, 12, 61, 0, 151, 12, 196, 12, 38, 0, 151, 12, 205, 12, 140, 12, 61, 0, 151, 12, 226, 12, 38, 0, 151, 12, 205, 12, 225, 12, 61, 0, 151, 12, 227, 12, 38, 0, 151, 12, 205, 12, 142, 12, 61, 0, 151, 12, 198, 12, 38, 0, 151, 12, 205, 12, 143, 12, 61, 0, 151, 12, 199, 12, 38, 0, 151, 12, 205, 12, 144, 12, 61, 0, 151, 12, 200, 12, 38, 0, 151, 12, 205, 12, 146, 12, 61, 0, 151, 12, 202, 12, 38, 0, 151, 12, 205, 12, 147, 12, 61, 0, 151, 12, 203, 12, 38, 0, 151, 12, 205, 12, 148, 12, 61, 0, 151, 12, 204, 12, 38, 0, 152, 12, 205, 12, 133, 12, 61, 0, 152, 12, 38, 0, 152, 12, 205, 12, 134, 12, 61, 0, 152, 12, 190, 12, 38, 0, 152, 12, 205, 12, 135, 12, 61, 0, 152, 12, 191, 12, 38, 0, 152, 12, 205, 12, 136, 12, 61, 0, 152, 12, 192, 12, 38, 0, 152, 12, 205, 12, 137, 12, 61, 0, 152, 12, 193, 12, 38, 0, 152, 12, 205, 12, 138, 12, 61, 0, 152, 12, 194, 12, 38, 0, 152, 12, 205, 12, 139, 12, 61, 0, 152, 12, 195, 12, 38, 0, 152, 12, 205, 12, 224, 12, 61, 0, 152, 12, 196, 12, 38, 0, 152, 12, 205, 12, 140, 12, 61, 0, 152, 12, 226, 12, 38, 0, 152, 12, 205, 12, 225, 12, 61, 0, 152, 12, 227, 12, 38, 0, 152, 12, 205, 12, 142, 12, 61, 0, 152, 12, 198, 12, 38, 0, 152, 12, 205, 12, 143, 12, 61, 0, 152, 12, 199, 12, 38, 0, 152, 12, 205, 12, 144, 12, 61, 0, 152, 12, 200, 12, 38, 0, 152, 12, 205, 12, 146, 12, 61, 0, 152, 12, 202, 12, 38, 0, 152, 12, 205, 12, 147, 12, 61, 0, 152, 12, 203, 12, 38, 0, 152, 12, 205, 12, 148, 12, 61, 0, 152, 12, 204, 12, 38, 0, 153, 12, 205, 12, 133, 12, 61, 0, 153, 12, 38, 0, 153, 12, 205, 12, 134, 12, 61, 0, 153, 12, 190, 12, 38, 0, 153, 12, 205, 12, 135, 12, 61, 0, 153, 12, 191, 12, 38, 0, 153, 12, 205, 12, 136, 12, 61, 0, 153, 12, 192, 12, 38, 0, 153, 12, 205, 12, 137, 12, 61, 0, 153, 12, 193, 12, 38, 0, 153, 12, 205, 12, 138, 12, 61, 0, 153, 12, 194, 12, 38, 0, 153, 12, 205, 12, 139, 12, 61, 0, 153, 12, 195, 12, 38, 0, 153, 12, 205, 12, 224, 12, 61, 0, 153, 12, 196, 12, 38, 0, 153, 12, 205, 12, 140, 12, 61, 0, 153, 12, 226, 12, 38, 0, 153, 12, 205, 12, 225, 12, 61, 0, 153, 12, 227, 12, 38, 0, 153, 12, 205, 12, 142, 12, 61, 0, 153, 12, 198, 12, 38, 0, 153, 12, 205, 12, 143, 12, 61, 0, 153, 12, 199, 12, 38, 0, 153, 12, 205, 12, 144, 12, 61, 0, 153, 12, 200, 12, 38, 0, 153, 12, 205, 12, 146, 12, 61, 0, 153, 12, 202, 12, 38, 0, 153, 12, 205, 12, 147, 12, 61, 0, 153, 12, 203, 12, 38, 0, 153, 12, 205, 12, 148, 12, 61, 0, 153, 12, 204, 12, 38, 0, 154, 12, 205, 12, 133, 12, 61, 0, 154, 12, 38, 0, 154, 12, 205, 12, 134, 12, 61, 0, 154, 12, 190, 12, 38, 0, 154, 12, 205, 12, 135, 12, 61, 0, 154, 12, 191, 12, 38, 0, 154, 12, 205, 12, 136, 12, 61, 0, 154, 12, 192, 12, 38, 0, 154, 12, 205, 12, 137, 12, 61, 0, 154, 12, 193, 12, 38, 0, 154, 12, 205, 12, 138, 12, 61, 0, 154, 12, 194, 12, 38, 0, 154, 12, 205, 12, 139, 12, 61, 0, 154, 12, 195, 12, 38, 0, 154, 12, 205, 12, 224, 12, 61, 0, 154, 12, 196, 12, 38, 0, 154, 12, 205, 12, 140, 12, 61, 0, 154, 12, 226, 12, 38, 0, 154, 12, 205, 12, 225, 12, 61, 0, 154, 12, 227, 12, 38, 0, 154, 12, 205, 12, 142, 12, 61, 0, 154, 12, 198, 12, 38, 0, 154, 12, 205, 12, 143, 12, 61, 0, 154, 12, 199, 12, 38, 0, 154, 12, 205, 12, 144, 12, 61, 0, 154, 12, 200, 12, 38, 0, 154, 12, 205, 12, 146, 12, 61, 0, 154, 12, 202, 12, 38, 0, 154, 12, 205, 12, 147, 12, 61, 0, 154, 12, 203, 12, 38, 0, 154, 12, 205, 12, 148, 12, 61, 0, 154, 12, 204, 12, 38, 0, 155, 12, 205, 12, 133, 12, 61, 0, 155, 12, 38, 0, 155, 12, 205, 12, 134, 12, 61, 0, 155, 12, 190, 12, 38, 0, 155, 12, 205, 12, 135, 12, 61, 0, 155, 12, 191, 12, 38, 0, 155, 12, 205, 12, 136, 12, 61, 0, 155, 12, 192, 12, 38, 0, 155, 12, 205, 12, 137, 12, 61, 0, 155, 12, 193, 12, 38, 0, 155, 12, 205, 12, 138, 12, 61, 0, 155, 12, 194, 12, 38, 0, 155, 12, 205, 12, 139, 12, 61, 0, 155, 12, 195, 12, 38, 0, 155, 12, 205, 12, 224, 12, 61, 0, 155, 12, 196, 12, 38, 0, 155, 12, 205, 12, 140, 12, 61, 0, 155, 12, 226, 12, 38, 0, 155, 12, 205, 12, 225, 12, 61, 0, 155, 12, 227, 12, 38, 0, 155, 12, 205, 12, 142, 12, 61, 0, 155, 12, 198, 12, 38, 0, 155, 12, 205, 12, 143, 12, 61, 0, 155, 12, 199, 12, 38, 0, 155, 12, 205, 12, 144, 12, 61, 0, 155, 12, 200, 12, 38, 0, 155, 12, 205, 12, 146, 12, 61, 0, 155, 12, 202, 12, 38, 0, 155, 12, 205, 12, 147, 12, 61, 0, 155, 12, 203, 12, 38, 0, 155, 12, 205, 12, 148, 12, 61, 0, 155, 12, 204, 12, 38, 0, 156, 12, 205, 12, 133, 12, 61, 0, 156, 12, 38, 0, 156, 12, 205, 12, 134, 12, 61, 0, 156, 12, 190, 12, 38, 0, 156, 12, 205, 12, 135, 12, 61, 0, 156, 12, 191, 12, 38, 0, 156, 12, 205, 12, 136, 12, 61, 0, 156, 12, 192, 12, 38, 0, 156, 12, 205, 12, 137, 12, 61, 0, 156, 12, 193, 12, 38, 0, 156, 12, 205, 12, 138, 12, 61, 0, 156, 12, 194, 12, 38, 0, 156, 12, 205, 12, 139, 12, 61, 0, 156, 12, 195, 12, 38, 0, 156, 12, 205, 12, 224, 12, 61, 0, 156, 12, 196, 12, 38, 0, 156, 12, 205, 12, 140, 12, 61, 0, 156, 12, 226, 12, 38, 0, 156, 12, 205, 12, 225, 12, 61, 0, 156, 12, 227, 12, 38, 0, 156, 12, 205, 12, 142, 12, 61, 0, 156, 12, 198, 12, 38, 0, 156, 12, 205, 12, 143, 12, 61, 0, 156, 12, 199, 12, 38, 0, 156, 12, 205, 12, 144, 12, 61, 0, 156, 12, 200, 12, 38, 0, 156, 12, 205, 12, 146, 12, 61, 0, 156, 12, 202, 12, 38, 0, 156, 12, 205, 12, 147, 12, 61, 0, 156, 12, 203, 12, 38, 0, 156, 12, 205, 12, 148, 12, 61, 0, 156, 12, 204, 12, 38, 0, 157, 12, 205, 12, 133, 12, 61, 0, 157, 12, 38, 0, 157, 12, 205, 12, 134, 12, 61, 0, 157, 12, 190, 12, 38, 0, 157, 12, 205, 12, 135, 12, 61, 0, 157, 12, 191, 12, 38, 0, 157, 12, 205, 12, 136, 12, 61, 0, 157, 12, 192, 12, 38, 0, 157, 12, 205, 12, 137, 12, 61, 0, 157, 12, 193, 12, 38, 0, 157, 12, 205, 12, 138, 12, 61, 0, 157, 12, 194, 12, 38, 0, 157, 12, 205, 12, 139, 12, 61, 0, 157, 12, 195, 12, 38, 0, 157, 12, 205, 12, 224, 12, 61, 0, 157, 12, 196, 12, 38, 0, 157, 12, 205, 12, 140, 12, 61, 0, 157, 12, 226, 12, 38, 0, 157, 12, 205, 12, 225, 12, 61, 0, 157, 12, 227, 12, 38, 0, 157, 12, 205, 12, 142, 12, 61, 0, 157, 12, 198, 12, 38, 0, 157, 12, 205, 12, 143, 12, 61, 0, 157, 12, 199, 12, 38, 0, 157, 12, 205, 12, 144, 12, 61, 0, 157, 12, 200, 12, 38, 0, 157, 12, 205, 12, 146, 12, 61, 0, 157, 12, 202, 12, 38, 0, 157, 12, 205, 12, 147, 12, 61, 0, 157, 12, 203, 12, 38, 0, 157, 12, 205, 12, 148, 12, 61, 0, 157, 12, 204, 12, 38, 0, 158, 12, 205, 12, 133, 12, 61, 0, 158, 12, 38, 0, 158, 12, 205, 12, 134, 12, 61, 0, 158, 12, 190, 12, 38, 0, 158, 12, 205, 12, 135, 12, 61, 0, 158, 12, 191, 12, 38, 0, 158, 12, 205, 12, 136, 12, 61, 0, 158, 12, 192, 12, 38, 0, 158, 12, 205, 12, 137, 12, 61, 0, 158, 12, 193, 12, 38, 0, 158, 12, 205, 12, 138, 12, 61, 0, 158, 12, 194, 12, 38, 0, 158, 12, 205, 12, 139, 12, 61, 0, 158, 12, 195, 12, 38, 0, 158, 12, 205, 12, 224, 12, 61, 0, 158, 12, 196, 12, 38, 0, 158, 12, 205, 12, 140, 12, 61, 0, 158, 12, 226, 12, 38, 0, 158, 12, 205, 12, 225, 12, 61, 0, 158, 12, 227, 12, 38, 0, 158, 12, 205, 12, 142, 12, 61, 0, 158, 12, 198, 12, 38, 0, 158, 12, 205, 12, 143, 12, 61, 0, 158, 12, 199, 12, 38, 0, 158, 12, 205, 12, 144, 12, 61, 0, 158, 12, 200, 12, 38, 0, 158, 12, 205, 12, 146, 12, 61, 0, 158, 12, 202, 12, 38, 0, 158, 12, 205, 12, 147, 12, 61, 0, 158, 12, 203, 12, 38, 0, 158, 12, 205, 12, 148, 12, 61, 0, 158, 12, 204, 12, 38, 0, 159, 12, 205, 12, 133, 12, 61, 0, 159, 12, 38, 0, 159, 12, 205, 12, 134, 12, 61, 0, 159, 12, 190, 12, 38, 0, 159, 12, 205, 12, 135, 12, 61, 0, 159, 12, 191, 12, 38, 0, 159, 12, 205, 12, 136, 12, 61, 0, 159, 12, 192, 12, 38, 0, 159, 12, 205, 12, 137, 12, 61, 0, 159, 12, 193, 12, 38, 0, 159, 12, 205, 12, 138, 12, 61, 0, 159, 12, 194, 12, 38, 0, 159, 12, 205, 12, 139, 12, 61, 0, 159, 12, 195, 12, 38, 0, 159, 12, 205, 12, 224, 12, 61, 0, 159, 12, 196, 12, 38, 0, 159, 12, 205, 12, 140, 12, 61, 0, 159, 12, 226, 12, 38, 0, 159, 12, 205, 12, 225, 12, 61, 0, 159, 12, 227, 12, 38, 0, 159, 12, 205, 12, 142, 12, 61, 0, 159, 12, 198, 12, 38, 0, 159, 12, 205, 12, 143, 12, 61, 0, 159, 12, 199, 12, 38, 0, 159, 12, 205, 12, 144, 12, 61, 0, 159, 12, 200, 12, 38, 0, 159, 12, 205, 12, 146, 12, 61, 0, 159, 12, 202, 12, 38, 0, 159, 12, 205, 12, 147, 12, 61, 0, 159, 12, 203, 12, 38, 0, 159, 12, 205, 12, 148, 12, 61, 0, 159, 12, 204, 12, 38, 0, 160, 12, 205, 12, 133, 12, 61, 0, 160, 12, 38, 0, 160, 12, 205, 12, 134, 12, 61, 0, 160, 12, 190, 12, 38, 0, 160, 12, 205, 12, 135, 12, 61, 0, 160, 12, 191, 12, 38, 0, 160, 12, 205, 12, 136, 12, 61, 0, 160, 12, 192, 12, 38, 0, 160, 12, 205, 12, 137, 12, 61, 0, 160, 12, 193, 12, 38, 0, 160, 12, 205, 12, 138, 12, 61, 0, 160, 12, 194, 12, 38, 0, 160, 12, 205, 12, 139, 12, 61, 0, 160, 12, 195, 12, 38, 0, 160, 12, 205, 12, 224, 12, 61, 0, 160, 12, 196, 12, 38, 0, 160, 12, 205, 12, 140, 12, 61, 0, 160, 12, 226, 12, 38, 0, 160, 12, 205, 12, 225, 12, 61, 0, 160, 12, 227, 12, 38, 0, 160, 12, 205, 12, 142, 12, 61, 0, 160, 12, 198, 12, 38, 0, 160, 12, 205, 12, 143, 12, 61, 0, 160, 12, 199, 12, 38, 0, 160, 12, 205, 12, 144, 12, 61, 0, 160, 12, 200, 12, 38, 0, 160, 12, 205, 12, 146, 12, 61, 0, 160, 12, 202, 12, 38, 0, 160, 12, 205, 12, 147, 12, 61, 0, 160, 12, 203, 12, 38, 0, 160, 12, 205, 12, 148, 12, 61, 0, 160, 12, 204, 12, 38, 0, 161, 12, 205, 12, 133, 12, 61, 0, 161, 12, 38, 0, 161, 12, 205, 12, 134, 12, 61, 0, 161, 12, 190, 12, 38, 0, 161, 12, 205, 12, 135, 12, 61, 0, 161, 12, 191, 12, 38, 0, 161, 12, 205, 12, 136, 12, 61, 0, 161, 12, 192, 12, 38, 0, 161, 12, 205, 12, 137, 12, 61, 0, 161, 12, 193, 12, 38, 0, 161, 12, 205, 12, 138, 12, 61, 0, 161, 12, 194, 12, 38, 0, 161, 12, 205, 12, 139, 12, 61, 0, 161, 12, 195, 12, 38, 0, 161, 12, 205, 12, 224, 12, 61, 0, 161, 12, 196, 12, 38, 0, 161, 12, 205, 12, 140, 12, 61, 0, 161, 12, 226, 12, 38, 0, 161, 12, 205, 12, 225, 12, 61, 0, 161, 12, 227, 12, 38, 0, 161, 12, 205, 12, 142, 12, 61, 0, 161, 12, 198, 12, 38, 0, 161, 12, 205, 12, 143, 12, 61, 0, 161, 12, 199, 12, 38, 0, 161, 12, 205, 12, 144, 12, 61, 0, 161, 12, 200, 12, 38, 0, 161, 12, 205, 12, 146, 12, 61, 0, 161, 12, 202, 12, 38, 0, 161, 12, 205, 12, 147, 12, 61, 0, 161, 12, 203, 12, 38, 0, 161, 12, 205, 12, 148, 12, 61, 0, 161, 12, 204, 12, 38, 0, 162, 12, 205, 12, 133, 12, 61, 0, 162, 12, 38, 0, 162, 12, 205, 12, 134, 12, 61, 0, 162, 12, 190, 12, 38, 0, 162, 12, 205, 12, 135, 12, 61, 0, 162, 12, 191, 12, 38, 0, 162, 12, 205, 12, 136, 12, 61, 0, 162, 12, 192, 12, 38, 0, 162, 12, 205, 12, 137, 12, 61, 0, 162, 12, 193, 12, 38, 0, 162, 12, 205, 12, 138, 12, 61, 0, 162, 12, 194, 12, 38, 0, 162, 12, 205, 12, 139, 12, 61, 0, 162, 12, 195, 12, 38, 0, 162, 12, 205, 12, 224, 12, 61, 0, 162, 12, 196, 12, 38, 0, 162, 12, 205, 12, 140, 12, 61, 0, 162, 12, 226, 12, 38, 0, 162, 12, 205, 12, 225, 12, 61, 0, 162, 12, 227, 12, 38, 0, 162, 12, 205, 12, 142, 12, 61, 0, 162, 12, 198, 12, 38, 0, 162, 12, 205, 12, 143, 12, 61, 0, 162, 12, 199, 12, 38, 0, 162, 12, 205, 12, 144, 12, 61, 0, 162, 12, 200, 12, 38, 0, 162, 12, 205, 12, 146, 12, 61, 0, 162, 12, 202, 12, 38, 0, 162, 12, 205, 12, 147, 12, 61, 0, 162, 12, 203, 12, 38, 0, 162, 12, 205, 12, 148, 12, 61, 0, 162, 12, 204, 12, 38, 0, 163, 12, 205, 12, 133, 12, 61, 0, 163, 12, 38, 0, 163, 12, 205, 12, 134, 12, 61, 0, 163, 12, 190, 12, 38, 0, 163, 12, 205, 12, 135, 12, 61, 0, 163, 12, 191, 12, 38, 0, 163, 12, 205, 12, 136, 12, 61, 0, 163, 12, 192, 12, 38, 0, 163, 12, 205, 12, 137, 12, 61, 0, 163, 12, 193, 12, 38, 0, 163, 12, 205, 12, 138, 12, 61, 0, 163, 12, 194, 12, 38, 0, 163, 12, 205, 12, 139, 12, 61, 0, 163, 12, 195, 12, 38, 0, 163, 12, 205, 12, 224, 12, 61, 0, 163, 12, 196, 12, 38, 0, 163, 12, 205, 12, 140, 12, 61, 0, 163, 12, 226, 12, 38, 0, 163, 12, 205, 12, 225, 12, 61, 0, 163, 12, 227, 12, 38, 0, 163, 12, 205, 12, 142, 12, 61, 0, 163, 12, 198, 12, 38, 0, 163, 12, 205, 12, 143, 12, 61, 0, 163, 12, 199, 12, 38, 0, 163, 12, 205, 12, 144, 12, 61, 0, 163, 12, 200, 12, 38, 0, 163, 12, 205, 12, 146, 12, 61, 0, 163, 12, 202, 12, 38, 0, 163, 12, 205, 12, 147, 12, 61, 0, 163, 12, 203, 12, 38, 0, 163, 12, 205, 12, 148, 12, 61, 0, 163, 12, 204, 12, 38, 0, 164, 12, 205, 12, 133, 12, 61, 0, 164, 12, 38, 0, 164, 12, 205, 12, 134, 12, 61, 0, 164, 12, 190, 12, 38, 0, 164, 12, 205, 12, 135, 12, 61, 0, 164, 12, 191, 12, 38, 0, 164, 12, 205, 12, 136, 12, 61, 0, 164, 12, 192, 12, 38, 0, 164, 12, 205, 12, 137, 12, 61, 0, 164, 12, 193, 12, 38, 0, 164, 12, 205, 12, 138, 12, 61, 0, 164, 12, 194, 12, 38, 0, 164, 12, 205, 12, 139, 12, 61, 0, 164, 12, 195, 12, 38, 0, 164, 12, 205, 12, 224, 12, 61, 0, 164, 12, 196, 12, 38, 0, 164, 12, 205, 12, 140, 12, 61, 0, 164, 12, 226, 12, 38, 0, 164, 12, 205, 12, 225, 12, 61, 0, 164, 12, 227, 12, 38, 0, 164, 12, 205, 12, 142, 12, 61, 0, 164, 12, 198, 12, 38, 0, 164, 12, 205, 12, 143, 12, 61, 0, 164, 12, 199, 12, 38, 0, 164, 12, 205, 12, 144, 12, 61, 0, 164, 12, 200, 12, 38, 0, 164, 12, 205, 12, 146, 12, 61, 0, 164, 12, 202, 12, 38, 0, 164, 12, 205, 12, 147, 12, 61, 0, 164, 12, 203, 12, 38, 0, 164, 12, 205, 12, 148, 12, 61, 0, 164, 12, 204, 12, 38, 0, 165, 12, 205, 12, 133, 12, 61, 0, 165, 12, 38, 0, 165, 12, 205, 12, 134, 12, 61, 0, 165, 12, 190, 12, 38, 0, 165, 12, 205, 12, 135, 12, 61, 0, 165, 12, 191, 12, 38, 0, 165, 12, 205, 12, 136, 12, 61, 0, 165, 12, 192, 12, 38, 0, 165, 12, 205, 12, 137, 12, 61, 0, 165, 12, 193, 12, 38, 0, 165, 12, 205, 12, 138, 12, 61, 0, 165, 12, 194, 12, 38, 0, 165, 12, 205, 12, 139, 12, 61, 0, 165, 12, 195, 12, 38, 0, 165, 12, 205, 12, 224, 12, 61, 0, 165, 12, 196, 12, 38, 0, 165, 12, 205, 12, 140, 12, 61, 0, 165, 12, 226, 12, 38, 0, 165, 12, 205, 12, 225, 12, 61, 0, 165, 12, 227, 12, 38, 0, 165, 12, 205, 12, 142, 12, 61, 0, 165, 12, 198, 12, 38, 0, 165, 12, 205, 12, 143, 12, 61, 0, 165, 12, 199, 12, 38, 0, 165, 12, 205, 12, 144, 12, 61, 0, 165, 12, 200, 12, 38, 0, 165, 12, 205, 12, 146, 12, 61, 0, 165, 12, 202, 12, 38, 0, 165, 12, 205, 12, 147, 12, 61, 0, 165, 12, 203, 12, 38, 0, 165, 12, 205, 12, 148, 12, 61, 0, 165, 12, 204, 12, 38, 0, 166, 12, 205, 12, 133, 12, 61, 0, 166, 12, 38, 0, 166, 12, 205, 12, 134, 12, 61, 0, 166, 12, 190, 12, 38, 0, 166, 12, 205, 12, 135, 12, 61, 0, 166, 12, 191, 12, 38, 0, 166, 12, 205, 12, 136, 12, 61, 0, 166, 12, 192, 12, 38, 0, 166, 12, 205, 12, 137, 12, 61, 0, 166, 12, 193, 12, 38, 0, 166, 12, 205, 12, 138, 12, 61, 0, 166, 12, 194, 12, 38, 0, 166, 12, 205, 12, 139, 12, 61, 0, 166, 12, 195, 12, 38, 0, 166, 12, 205, 12, 224, 12, 61, 0, 166, 12, 196, 12, 38, 0, 166, 12, 205, 12, 140, 12, 61, 0, 166, 12, 226, 12, 38, 0, 166, 12, 205, 12, 225, 12, 61, 0, 166, 12, 227, 12, 38, 0, 166, 12, 205, 12, 142, 12, 61, 0, 166, 12, 198, 12, 38, 0, 166, 12, 205, 12, 143, 12, 61, 0, 166, 12, 199, 12, 38, 0, 166, 12, 205, 12, 144, 12, 61, 0, 166, 12, 200, 12, 38, 0, 166, 12, 205, 12, 146, 12, 61, 0, 166, 12, 202, 12, 38, 0, 166, 12, 205, 12, 147, 12, 61, 0, 166, 12, 203, 12, 38, 0, 166, 12, 205, 12, 148, 12, 61, 0, 166, 12, 204, 12, 38, 0, 167, 12, 205, 12, 133, 12, 61, 0, 167, 12, 38, 0, 167, 12, 205, 12, 134, 12, 61, 0, 167, 12, 190, 12, 38, 0, 167, 12, 205, 12, 135, 12, 61, 0, 167, 12, 191, 12, 38, 0, 167, 12, 205, 12, 136, 12, 61, 0, 167, 12, 192, 12, 38, 0, 167, 12, 205, 12, 137, 12, 61, 0, 167, 12, 193, 12, 38, 0, 167, 12, 205, 12, 138, 12, 61, 0, 167, 12, 194, 12, 38, 0, 167, 12, 205, 12, 139, 12, 61, 0, 167, 12, 195, 12, 38, 0, 167, 12, 205, 12, 224, 12, 61, 0, 167, 12, 196, 12, 38, 0, 167, 12, 205, 12, 140, 12, 61, 0, 167, 12, 226, 12, 38, 0, 167, 12, 205, 12, 225, 12, 61, 0, 167, 12, 227, 12, 38, 0, 167, 12, 205, 12, 142, 12, 61, 0, 167, 12, 198, 12, 38, 0, 167, 12, 205, 12, 143, 12, 61, 0, 167, 12, 199, 12, 38, 0, 167, 12, 205, 12, 144, 12, 61, 0, 167, 12, 200, 12, 38, 0, 167, 12, 205, 12, 146, 12, 61, 0, 167, 12, 202, 12, 38, 0, 167, 12, 205, 12, 147, 12, 61, 0, 167, 12, 203, 12, 38, 0, 167, 12, 205, 12, 148, 12, 61, 0, 167, 12, 204, 12, 38, 0, 168, 12, 205, 12, 133, 12, 61, 0, 168, 12, 38, 0, 168, 12, 205, 12, 134, 12, 61, 0, 168, 12, 190, 12, 38, 0, 168, 12, 205, 12, 135, 12, 61, 0, 168, 12, 191, 12, 38, 0, 168, 12, 205, 12, 136, 12, 61, 0, 168, 12, 192, 12, 38, 0, 168, 12, 205, 12, 137, 12, 61, 0, 168, 12, 193, 12, 38, 0, 168, 12, 205, 12, 138, 12, 61, 0, 168, 12, 194, 12, 38, 0, 168, 12, 205, 12, 139, 12, 61, 0, 168, 12, 195, 12, 38, 0, 168, 12, 205, 12, 224, 12, 61, 0, 168, 12, 196, 12, 38, 0, 168, 12, 205, 12, 140, 12, 61, 0, 168, 12, 226, 12, 38, 0, 168, 12, 205, 12, 225, 12, 61, 0, 168, 12, 227, 12, 38, 0, 168, 12, 205, 12, 142, 12, 61, 0, 168, 12, 198, 12, 38, 0, 168, 12, 205, 12, 143, 12, 61, 0, 168, 12, 199, 12, 38, 0, 168, 12, 205, 12, 144, 12, 61, 0, 168, 12, 200, 12, 38, 0, 168, 12, 205, 12, 146, 12, 61, 0, 168, 12, 202, 12, 38, 0, 168, 12, 205, 12, 147, 12, 61, 0, 168, 12, 203, 12, 38, 0, 168, 12, 205, 12, 148, 12, 61, 0, 168, 12, 204, 12, 38, 0, 170, 12, 205, 12, 133, 12, 61, 0, 170, 12, 38, 0, 170, 12, 205, 12, 134, 12, 61, 0, 170, 12, 190, 12, 38, 0, 170, 12, 205, 12, 135, 12, 61, 0, 170, 12, 191, 12, 38, 0, 170, 12, 205, 12, 136, 12, 61, 0, 170, 12, 192, 12, 38, 0, 170, 12, 205, 12, 137, 12, 61, 0, 170, 12, 193, 12, 38, 0, 170, 12, 205, 12, 138, 12, 61, 0, 170, 12, 194, 12, 38, 0, 170, 12, 205, 12, 139, 12, 61, 0, 170, 12, 195, 12, 38, 0, 170, 12, 205, 12, 224, 12, 61, 0, 170, 12, 196, 12, 38, 0, 170, 12, 205, 12, 140, 12, 61, 0, 170, 12, 226, 12, 38, 0, 170, 12, 205, 12, 225, 12, 61, 0, 170, 12, 227, 12, 38, 0, 170, 12, 205, 12, 142, 12, 61, 0, 170, 12, 198, 12, 38, 0, 170, 12, 205, 12, 143, 12, 61, 0, 170, 12, 199, 12, 38, 0, 170, 12, 205, 12, 144, 12, 61, 0, 170, 12, 200, 12, 38, 0, 170, 12, 205, 12, 146, 12, 61, 0, 170, 12, 202, 12, 38, 0, 170, 12, 205, 12, 147, 12, 61, 0, 170, 12, 203, 12, 38, 0, 170, 12, 205, 12, 148, 12, 61, 0, 170, 12, 204, 12, 38, 0, 171, 12, 205, 12, 133, 12, 61, 0, 171, 12, 38, 0, 171, 12, 205, 12, 134, 12, 61, 0, 171, 12, 190, 12, 38, 0, 171, 12, 205, 12, 135, 12, 61, 0, 171, 12, 191, 12, 38, 0, 171, 12, 205, 12, 136, 12, 61, 0, 171, 12, 192, 12, 38, 0, 171, 12, 205, 12, 137, 12, 61, 0, 171, 12, 193, 12, 38, 0, 171, 12, 205, 12, 138, 12, 61, 0, 171, 12, 194, 12, 38, 0, 171, 12, 205, 12, 139, 12, 61, 0, 171, 12, 195, 12, 38, 0, 171, 12, 205, 12, 224, 12, 61, 0, 171, 12, 196, 12, 38, 0, 171, 12, 205, 12, 140, 12, 61, 0, 171, 12, 226, 12, 38, 0, 171, 12, 205, 12, 225, 12, 61, 0, 171, 12, 227, 12, 38, 0, 171, 12, 205, 12, 142, 12, 61, 0, 171, 12, 198, 12, 38, 0, 171, 12, 205, 12, 143, 12, 61, 0, 171, 12, 199, 12, 38, 0, 171, 12, 205, 12, 144, 12, 61, 0, 171, 12, 200, 12, 38, 0, 171, 12, 205, 12, 146, 12, 61, 0, 171, 12, 202, 12, 38, 0, 171, 12, 205, 12, 147, 12, 61, 0, 171, 12, 203, 12, 38, 0, 171, 12, 205, 12, 148, 12, 61, 0, 171, 12, 204, 12, 38, 0, 172, 12, 205, 12, 133, 12, 61, 0, 172, 12, 38, 0, 172, 12, 205, 12, 134, 12, 61, 0, 172, 12, 190, 12, 38, 0, 172, 12, 205, 12, 135, 12, 61, 0, 172, 12, 191, 12, 38, 0, 172, 12, 205, 12, 136, 12, 61, 0, 172, 12, 192, 12, 38, 0, 172, 12, 205, 12, 137, 12, 61, 0, 172, 12, 193, 12, 38, 0, 172, 12, 205, 12, 138, 12, 61, 0, 172, 12, 194, 12, 38, 0, 172, 12, 205, 12, 139, 12, 61, 0, 172, 12, 195, 12, 38, 0, 172, 12, 205, 12, 224, 12, 61, 0, 172, 12, 196, 12, 38, 0, 172, 12, 205, 12, 140, 12, 61, 0, 172, 12, 226, 12, 38, 0, 172, 12, 205, 12, 225, 12, 61, 0, 172, 12, 227, 12, 38, 0, 172, 12, 205, 12, 142, 12, 61, 0, 172, 12, 198, 12, 38, 0, 172, 12, 205, 12, 143, 12, 61, 0, 172, 12, 199, 12, 38, 0, 172, 12, 205, 12, 144, 12, 61, 0, 172, 12, 200, 12, 38, 0, 172, 12, 205, 12, 146, 12, 61, 0, 172, 12, 202, 12, 38, 0, 172, 12, 205, 12, 147, 12, 61, 0, 172, 12, 203, 12, 38, 0, 172, 12, 205, 12, 148, 12, 61, 0, 172, 12, 204, 12, 38, 0, 173, 12, 205, 12, 133, 12, 61, 0, 173, 12, 38, 0, 173, 12, 205, 12, 134, 12, 61, 0, 173, 12, 190, 12, 38, 0, 173, 12, 205, 12, 135, 12, 61, 0, 173, 12, 191, 12, 38, 0, 173, 12, 205, 12, 136, 12, 61, 0, 173, 12, 192, 12, 38, 0, 173, 12, 205, 12, 137, 12, 61, 0, 173, 12, 193, 12, 38, 0, 173, 12, 205, 12, 138, 12, 61, 0, 173, 12, 194, 12, 38, 0, 173, 12, 205, 12, 139, 12, 61, 0, 173, 12, 195, 12, 38, 0, 173, 12, 205, 12, 224, 12, 61, 0, 173, 12, 196, 12, 38, 0, 173, 12, 205, 12, 140, 12, 61, 0, 173, 12, 226, 12, 38, 0, 173, 12, 205, 12, 225, 12, 61, 0, 173, 12, 227, 12, 38, 0, 173, 12, 205, 12, 142, 12, 61, 0, 173, 12, 198, 12, 38, 0, 173, 12, 205, 12, 143, 12, 61, 0, 173, 12, 199, 12, 38, 0, 173, 12, 205, 12, 144, 12, 61, 0, 173, 12, 200, 12, 38, 0, 173, 12, 205, 12, 146, 12, 61, 0, 173, 12, 202, 12, 38, 0, 173, 12, 205, 12, 147, 12, 61, 0, 173, 12, 203, 12, 38, 0, 173, 12, 205, 12, 148, 12, 61, 0, 173, 12, 204, 12, 38, 0, 174, 12, 205, 12, 133, 12, 61, 0, 174, 12, 38, 0, 174, 12, 205, 12, 134, 12, 61, 0, 174, 12, 190, 12, 38, 0, 174, 12, 205, 12, 135, 12, 61, 0, 174, 12, 191, 12, 38, 0, 174, 12, 205, 12, 136, 12, 61, 0, 174, 12, 192, 12, 38, 0, 174, 12, 205, 12, 137, 12, 61, 0, 174, 12, 193, 12, 38, 0, 174, 12, 205, 12, 138, 12, 61, 0, 174, 12, 194, 12, 38, 0, 174, 12, 205, 12, 139, 12, 61, 0, 174, 12, 195, 12, 38, 0, 174, 12, 205, 12, 224, 12, 61, 0, 174, 12, 196, 12, 38, 0, 174, 12, 205, 12, 140, 12, 61, 0, 174, 12, 226, 12, 38, 0, 174, 12, 205, 12, 225, 12, 61, 0, 174, 12, 227, 12, 38, 0, 174, 12, 205, 12, 142, 12, 61, 0, 174, 12, 198, 12, 38, 0, 174, 12, 205, 12, 143, 12, 61, 0, 174, 12, 199, 12, 38, 0, 174, 12, 205, 12, 144, 12, 61, 0, 174, 12, 200, 12, 38, 0, 174, 12, 205, 12, 146, 12, 61, 0, 174, 12, 202, 12, 38, 0, 174, 12, 205, 12, 147, 12, 61, 0, 174, 12, 203, 12, 38, 0, 174, 12, 205, 12, 148, 12, 61, 0, 174, 12, 204, 12, 38, 0, 175, 12, 205, 12, 133, 12, 61, 0, 175, 12, 38, 0, 175, 12, 205, 12, 134, 12, 61, 0, 175, 12, 190, 12, 38, 0, 175, 12, 205, 12, 135, 12, 61, 0, 175, 12, 191, 12, 38, 0, 175, 12, 205, 12, 136, 12, 61, 0, 175, 12, 192, 12, 38, 0, 175, 12, 205, 12, 137, 12, 61, 0, 175, 12, 193, 12, 38, 0, 175, 12, 205, 12, 138, 12, 61, 0, 175, 12, 194, 12, 38, 0, 175, 12, 205, 12, 139, 12, 61, 0, 175, 12, 195, 12, 38, 0, 175, 12, 205, 12, 224, 12, 61, 0, 175, 12, 196, 12, 38, 0, 175, 12, 205, 12, 140, 12, 61, 0, 175, 12, 226, 12, 38, 0, 175, 12, 205, 12, 225, 12, 61, 0, 175, 12, 227, 12, 38, 0, 175, 12, 205, 12, 142, 12, 61, 0, 175, 12, 198, 12, 38, 0, 175, 12, 205, 12, 143, 12, 61, 0, 175, 12, 199, 12, 38, 0, 175, 12, 205, 12, 144, 12, 61, 0, 175, 12, 200, 12, 38, 0, 175, 12, 205, 12, 146, 12, 61, 0, 175, 12, 202, 12, 38, 0, 175, 12, 205, 12, 147, 12, 61, 0, 175, 12, 203, 12, 38, 0, 175, 12, 205, 12, 148, 12, 61, 0, 175, 12, 204, 12, 38, 0, 176, 12, 205, 12, 133, 12, 61, 0, 176, 12, 38, 0, 176, 12, 205, 12, 134, 12, 61, 0, 176, 12, 190, 12, 38, 0, 176, 12, 205, 12, 135, 12, 61, 0, 176, 12, 191, 12, 38, 0, 176, 12, 205, 12, 136, 12, 61, 0, 176, 12, 192, 12, 38, 0, 176, 12, 205, 12, 137, 12, 61, 0, 176, 12, 193, 12, 38, 0, 176, 12, 205, 12, 138, 12, 61, 0, 176, 12, 194, 12, 38, 0, 176, 12, 205, 12, 139, 12, 61, 0, 176, 12, 195, 12, 38, 0, 176, 12, 205, 12, 224, 12, 61, 0, 176, 12, 196, 12, 38, 0, 176, 12, 205, 12, 140, 12, 61, 0, 176, 12, 226, 12, 38, 0, 176, 12, 205, 12, 225, 12, 61, 0, 176, 12, 227, 12, 38, 0, 176, 12, 205, 12, 142, 12, 61, 0, 176, 12, 198, 12, 38, 0, 176, 12, 205, 12, 143, 12, 61, 0, 176, 12, 199, 12, 38, 0, 176, 12, 205, 12, 144, 12, 61, 0, 176, 12, 200, 12, 38, 0, 176, 12, 205, 12, 146, 12, 61, 0, 176, 12, 202, 12, 38, 0, 176, 12, 205, 12, 147, 12, 61, 0, 176, 12, 203, 12, 38, 0, 176, 12, 205, 12, 148, 12, 61, 0, 176, 12, 204, 12, 38, 0, 177, 12, 205, 12, 133, 12, 61, 0, 177, 12, 38, 0, 177, 12, 205, 12, 134, 12, 61, 0, 177, 12, 190, 12, 38, 0, 177, 12, 205, 12, 135, 12, 61, 0, 177, 12, 191, 12, 38, 0, 177, 12, 205, 12, 136, 12, 61, 0, 177, 12, 192, 12, 38, 0, 177, 12, 205, 12, 137, 12, 61, 0, 177, 12, 193, 12, 38, 0, 177, 12, 205, 12, 138, 12, 61, 0, 177, 12, 194, 12, 38, 0, 177, 12, 205, 12, 139, 12, 61, 0, 177, 12, 195, 12, 38, 0, 177, 12, 205, 12, 224, 12, 61, 0, 177, 12, 196, 12, 38, 0, 177, 12, 205, 12, 140, 12, 61, 0, 177, 12, 226, 12, 38, 0, 177, 12, 205, 12, 225, 12, 61, 0, 177, 12, 227, 12, 38, 0, 177, 12, 205, 12, 142, 12, 61, 0, 177, 12, 198, 12, 38, 0, 177, 12, 205, 12, 143, 12, 61, 0, 177, 12, 199, 12, 38, 0, 177, 12, 205, 12, 144, 12, 61, 0, 177, 12, 200, 12, 38, 0, 177, 12, 205, 12, 146, 12, 61, 0, 177, 12, 202, 12, 38, 0, 177, 12, 205, 12, 147, 12, 61, 0, 177, 12, 203, 12, 38, 0, 177, 12, 205, 12, 148, 12, 61, 0, 177, 12, 204, 12, 38, 0, 178, 12, 205, 12, 133, 12, 61, 0, 178, 12, 38, 0, 178, 12, 205, 12, 134, 12, 61, 0, 178, 12, 190, 12, 38, 0, 178, 12, 205, 12, 135, 12, 61, 0, 178, 12, 191, 12, 38, 0, 178, 12, 205, 12, 136, 12, 61, 0, 178, 12, 192, 12, 38, 0, 178, 12, 205, 12, 137, 12, 61, 0, 178, 12, 193, 12, 38, 0, 178, 12, 205, 12, 138, 12, 61, 0, 178, 12, 194, 12, 38, 0, 178, 12, 205, 12, 139, 12, 61, 0, 178, 12, 195, 12, 38, 0, 178, 12, 205, 12, 224, 12, 61, 0, 178, 12, 196, 12, 38, 0, 178, 12, 205, 12, 140, 12, 61, 0, 178, 12, 226, 12, 38, 0, 178, 12, 205, 12, 225, 12, 61, 0, 178, 12, 227, 12, 38, 0, 178, 12, 205, 12, 142, 12, 61, 0, 178, 12, 198, 12, 38, 0, 178, 12, 205, 12, 143, 12, 61, 0, 178, 12, 199, 12, 38, 0, 178, 12, 205, 12, 144, 12, 61, 0, 178, 12, 200, 12, 38, 0, 178, 12, 205, 12, 146, 12, 61, 0, 178, 12, 202, 12, 38, 0, 178, 12, 205, 12, 147, 12, 61, 0, 178, 12, 203, 12, 38, 0, 178, 12, 205, 12, 148, 12, 61, 0, 178, 12, 204, 12, 38, 0, 181, 12, 205, 12, 133, 12, 61, 0, 181, 12, 38, 0, 181, 12, 205, 12, 134, 12, 61, 0, 181, 12, 190, 12, 38, 0, 181, 12, 205, 12, 135, 12, 61, 0, 181, 12, 191, 12, 38, 0, 181, 12, 205, 12, 136, 12, 61, 0, 181, 12, 192, 12, 38, 0, 181, 12, 205, 12, 137, 12, 61, 0, 181, 12, 193, 12, 38, 0, 181, 12, 205, 12, 138, 12, 61, 0, 181, 12, 194, 12, 38, 0, 181, 12, 205, 12, 139, 12, 61, 0, 181, 12, 195, 12, 38, 0, 181, 12, 205, 12, 224, 12, 61, 0, 181, 12, 196, 12, 38, 0, 181, 12, 205, 12, 140, 12, 61, 0, 181, 12, 226, 12, 38, 0, 181, 12, 205, 12, 225, 12, 61, 0, 181, 12, 227, 12, 38, 0, 181, 12, 205, 12, 142, 12, 61, 0, 181, 12, 198, 12, 38, 0, 181, 12, 205, 12, 143, 12, 61, 0, 181, 12, 199, 12, 38, 0, 181, 12, 205, 12, 144, 12, 61, 0, 181, 12, 200, 12, 38, 0, 181, 12, 205, 12, 146, 12, 61, 0, 181, 12, 202, 12, 38, 0, 181, 12, 205, 12, 147, 12, 61, 0, 181, 12, 203, 12, 38, 0, 181, 12, 205, 12, 148, 12, 61, 0, 181, 12, 204, 12, 38, 0, 182, 12, 205, 12, 133, 12, 61, 0, 182, 12, 38, 0, 182, 12, 205, 12, 134, 12, 61, 0, 182, 12, 190, 12, 38, 0, 182, 12, 205, 12, 135, 12, 61, 0, 182, 12, 191, 12, 38, 0, 182, 12, 205, 12, 136, 12, 61, 0, 182, 12, 192, 12, 38, 0, 182, 12, 205, 12, 137, 12, 61, 0, 182, 12, 193, 12, 38, 0, 182, 12, 205, 12, 138, 12, 61, 0, 182, 12, 194, 12, 38, 0, 182, 12, 205, 12, 139, 12, 61, 0, 182, 12, 195, 12, 38, 0, 182, 12, 205, 12, 224, 12, 61, 0, 182, 12, 196, 12, 38, 0, 182, 12, 205, 12, 140, 12, 61, 0, 182, 12, 226, 12, 38, 0, 182, 12, 205, 12, 225, 12, 61, 0, 182, 12, 227, 12, 38, 0, 182, 12, 205, 12, 142, 12, 61, 0, 182, 12, 198, 12, 38, 0, 182, 12, 205, 12, 143, 12, 61, 0, 182, 12, 199, 12, 38, 0, 182, 12, 205, 12, 144, 12, 61, 0, 182, 12, 200, 12, 38, 0, 182, 12, 205, 12, 146, 12, 61, 0, 182, 12, 202, 12, 38, 0, 182, 12, 205, 12, 147, 12, 61, 0, 182, 12, 203, 12, 38, 0, 182, 12, 205, 12, 148, 12, 61, 0, 182, 12, 204, 12, 38, 0, 183, 12, 205, 12, 133, 12, 61, 0, 183, 12, 38, 0, 183, 12, 205, 12, 134, 12, 61, 0, 183, 12, 190, 12, 38, 0, 183, 12, 205, 12, 135, 12, 61, 0, 183, 12, 191, 12, 38, 0, 183, 12, 205, 12, 136, 12, 61, 0, 183, 12, 192, 12, 38, 0, 183, 12, 205, 12, 137, 12, 61, 0, 183, 12, 193, 12, 38, 0, 183, 12, 205, 12, 138, 12, 61, 0, 183, 12, 194, 12, 38, 0, 183, 12, 205, 12, 139, 12, 61, 0, 183, 12, 195, 12, 38, 0, 183, 12, 205, 12, 224, 12, 61, 0, 183, 12, 196, 12, 38, 0, 183, 12, 205, 12, 140, 12, 61, 0, 183, 12, 226, 12, 38, 0, 183, 12, 205, 12, 225, 12, 61, 0, 183, 12, 227, 12, 38, 0, 183, 12, 205, 12, 142, 12, 61, 0, 183, 12, 198, 12, 38, 0, 183, 12, 205, 12, 143, 12, 61, 0, 183, 12, 199, 12, 38, 0, 183, 12, 205, 12, 144, 12, 61, 0, 183, 12, 200, 12, 38, 0, 183, 12, 205, 12, 146, 12, 61, 0, 183, 12, 202, 12, 38, 0, 183, 12, 205, 12, 147, 12, 61, 0, 183, 12, 203, 12, 38, 0, 183, 12, 205, 12, 148, 12, 61, 0, 183, 12, 204, 12, 38, 0, 184, 12, 205, 12, 133, 12, 61, 0, 184, 12, 38, 0, 184, 12, 205, 12, 134, 12, 61, 0, 184, 12, 190, 12, 38, 0, 184, 12, 205, 12, 135, 12, 61, 0, 184, 12, 191, 12, 38, 0, 184, 12, 205, 12, 136, 12, 61, 0, 184, 12, 192, 12, 38, 0, 184, 12, 205, 12, 137, 12, 61, 0, 184, 12, 193, 12, 38, 0, 184, 12, 205, 12, 138, 12, 61, 0, 184, 12, 194, 12, 38, 0, 184, 12, 205, 12, 139, 12, 61, 0, 184, 12, 195, 12, 38, 0, 184, 12, 205, 12, 224, 12, 61, 0, 184, 12, 196, 12, 38, 0, 184, 12, 205, 12, 140, 12, 61, 0, 184, 12, 226, 12, 38, 0, 184, 12, 205, 12, 225, 12, 61, 0, 184, 12, 227, 12, 38, 0, 184, 12, 205, 12, 142, 12, 61, 0, 184, 12, 198, 12, 38, 0, 184, 12, 205, 12, 143, 12, 61, 0, 184, 12, 199, 12, 38, 0, 184, 12, 205, 12, 144, 12, 61, 0, 184, 12, 200, 12, 38, 0, 184, 12, 205, 12, 146, 12, 61, 0, 184, 12, 202, 12, 38, 0, 184, 12, 205, 12, 147, 12, 61, 0, 184, 12, 203, 12, 38, 0, 184, 12, 205, 12, 148, 12, 61, 0, 184, 12, 204, 12, 38, 0, 185, 12, 205, 12, 133, 12, 61, 0, 185, 12, 38, 0, 185, 12, 205, 12, 134, 12, 61, 0, 185, 12, 190, 12, 38, 0, 185, 12, 205, 12, 135, 12, 61, 0, 185, 12, 191, 12, 38, 0, 185, 12, 205, 12, 136, 12, 61, 0, 185, 12, 192, 12, 38, 0, 185, 12, 205, 12, 137, 12, 61, 0, 185, 12, 193, 12, 38, 0, 185, 12, 205, 12, 138, 12, 61, 0, 185, 12, 194, 12, 38, 0, 185, 12, 205, 12, 139, 12, 61, 0, 185, 12, 195, 12, 38, 0, 185, 12, 205, 12, 224, 12, 61, 0, 185, 12, 196, 12, 38, 0, 185, 12, 205, 12, 140, 12, 61, 0, 185, 12, 226, 12, 38, 0, 185, 12, 205, 12, 225, 12, 61, 0, 185, 12, 227, 12, 38, 0, 185, 12, 205, 12, 142, 12, 61, 0, 185, 12, 198, 12, 38, 0, 185, 12, 205, 12, 143, 12, 61, 0, 185, 12, 199, 12, 38, 0, 185, 12, 205, 12, 144, 12, 61, 0, 185, 12, 200, 12, 38, 0, 185, 12, 205, 12, 146, 12, 61, 0, 185, 12, 202, 12, 38, 0, 185, 12, 205, 12, 147, 12, 61, 0, 185, 12, 203, 12, 38, 0, 185, 12, 205, 12, 148, 12, 61, 0, 185, 12, 204, 12, 38, 0, 179, 12, 205, 12, 133, 12, 61, 0, 179, 12, 38, 0, 179, 12, 205, 12, 134, 12, 61, 0, 179, 12, 190, 12, 38, 0, 179, 12, 205, 12, 135, 12, 61, 0, 179, 12, 191, 12, 38, 0, 179, 12, 205, 12, 136, 12, 61, 0, 179, 12, 192, 12, 38, 0, 179, 12, 205, 12, 137, 12, 61, 0, 179, 12, 193, 12, 38, 0, 179, 12, 205, 12, 138, 12, 61, 0, 179, 12, 194, 12, 38, 0, 179, 12, 205, 12, 139, 12, 61, 0, 179, 12, 195, 12, 38, 0, 179, 12, 205, 12, 224, 12, 61, 0, 179, 12, 196, 12, 38, 0, 179, 12, 205, 12, 140, 12, 61, 0, 179, 12, 226, 12, 38, 0, 179, 12, 205, 12, 225, 12, 61, 0, 179, 12, 227, 12, 38, 0, 179, 12, 205, 12, 142, 12, 61, 0, 179, 12, 198, 12, 38, 0, 179, 12, 205, 12, 143, 12, 61, 0, 179, 12, 199, 12, 38, 0, 179, 12, 205, 12, 144, 12, 61, 0, 179, 12, 200, 12, 38, 0, 179, 12, 205, 12, 146, 12, 61, 0, 179, 12, 202, 12, 38, 0, 179, 12, 205, 12, 147, 12, 61, 0, 179, 12, 203, 12, 38, 0, 179, 12, 205, 12, 148, 12, 61, 0, 179, 12, 204, 12, 38, 0, 222, 12, 205, 12, 133, 12, 61, 0, 222, 12, 38, 0, 222, 12, 205, 12, 134, 12, 61, 0, 222, 12, 190, 12, 38, 0, 222, 12, 205, 12, 135, 12, 61, 0, 222, 12, 191, 12, 38, 0, 222, 12, 205, 12, 136, 12, 61, 0, 222, 12, 192, 12, 38, 0, 222, 12, 205, 12, 137, 12, 61, 0, 222, 12, 193, 12, 38, 0, 222, 12, 205, 12, 138, 12, 61, 0, 222, 12, 194, 12, 38, 0, 222, 12, 205, 12, 139, 12, 61, 0, 222, 12, 195, 12, 38, 0, 222, 12, 205, 12, 224, 12, 61, 0, 222, 12, 196, 12, 38, 0, 222, 12, 205, 12, 140, 12, 61, 0, 222, 12, 226, 12, 38, 0, 222, 12, 205, 12, 225, 12, 61, 0, 222, 12, 227, 12, 38, 0, 222, 12, 205, 12, 142, 12, 61, 0, 222, 12, 198, 12, 38, 0, 222, 12, 205, 12, 143, 12, 61, 0, 222, 12, 199, 12, 38, 0, 222, 12, 205, 12, 144, 12, 61, 0, 222, 12, 200, 12, 38, 0, 222, 12, 205, 12, 146, 12, 61, 0, 222, 12, 202, 12, 38, 0, 222, 12, 205, 12, 147, 12, 61, 0, 222, 12, 203, 12, 38, 0, 222, 12, 205, 12, 148, 12, 61, 0, 222, 12, 204, 12, 38, 0, 153, 12, 205, 12, 149, 12, 60, 0, 60, 0, 60, 0, 130, 12, 149, 12, 38, 0, 153, 12, 205, 12, 150, 12, 60, 0, 60, 0, 60, 0, 130, 12, 150, 12, 38, 0, 153, 12, 205, 12, 151, 12, 60, 0, 60, 0, 60, 0, 130, 12, 151, 12, 38, 0, 153, 12, 205, 12, 152, 12, 60, 0, 60, 0, 60, 0, 130, 12, 152, 12, 38, 0, 153, 12, 205, 12, 153, 12, 60, 0, 60, 0, 60, 0, 130, 12, 153, 12, 38, 0, 158, 12, 205, 12, 154, 12, 60, 0, 60, 0, 60, 0, 130, 12, 154, 12, 38, 0, 158, 12, 205, 12, 155, 12, 60, 0, 60, 0, 60, 0, 130, 12, 155, 12, 38, 0, 158, 12, 205, 12, 156, 12, 60, 0, 60, 0, 60, 0, 130, 12, 156, 12, 38, 0, 158, 12, 205, 12, 157, 12, 60, 0, 60, 0, 60, 0, 130, 12, 157, 12, 38, 0, 158, 12, 205, 12, 158, 12, 60, 0, 60, 0, 60, 0, 130, 12, 158, 12, 38, 0, 163, 12, 205, 12, 159, 12, 60, 0, 60, 0, 60, 0, 130, 12, 159, 12, 38, 0, 163, 12, 205, 12, 160, 12, 60, 0, 60, 0, 60, 0, 130, 12, 160, 12, 38, 0, 163, 12, 205, 12, 161, 12, 60, 0, 60, 0, 60, 0, 130, 12, 161, 12, 38, 0, 163, 12, 205, 12, 162, 12, 60, 0, 60, 0, 60, 0, 130, 12, 162, 12, 38, 0, 163, 12, 205, 12, 163, 12, 60, 0, 60, 0, 60, 0, 130, 12, 163, 12, 38, 0, 168, 12, 205, 12, 164, 12, 60, 0, 60, 0, 60, 0, 130, 12, 164, 12, 38, 0, 168, 12, 205, 12, 165, 12, 60, 0, 60, 0, 60, 0, 130, 12, 165, 12, 38, 0, 168, 12, 205, 12, 166, 12, 60, 0, 60, 0, 60, 0, 130, 12, 166, 12, 38, 0, 168, 12, 205, 12, 167, 12, 60, 0, 60, 0, 60, 0, 130, 12, 167, 12, 38, 0, 168, 12, 205, 12, 168, 12, 60, 0, 60, 0, 60, 0, 130, 12, 168, 12, 38, 0, 174, 12, 205, 12, 170, 12, 60, 0, 60, 0, 60, 0, 130, 12, 170, 12, 38, 0, 174, 12, 205, 12, 171, 12, 60, 0, 60, 0, 60, 0, 130, 12, 171, 12, 38, 0, 174, 12, 205, 12, 172, 12, 60, 0, 60, 0, 60, 0, 130, 12, 172, 12, 38, 0, 174, 12, 205, 12, 173, 12, 60, 0, 60, 0, 60, 0, 130, 12, 173, 12, 38, 0, 174, 12, 205, 12, 174, 12, 60, 0, 60, 0, 60, 0, 130, 12, 174, 12, 38, 0, 156, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 38, 0, 156, 12, 190, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 190, 12, 38, 0, 156, 12, 191, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 191, 12, 38, 0, 156, 12, 192, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 192, 12, 38, 0, 156, 12, 193, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 193, 12, 38, 0, 156, 12, 194, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 194, 12, 38, 0, 156, 12, 195, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 195, 12, 38, 0, 156, 12, 196, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 196, 12, 38, 0, 156, 12, 226, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 226, 12, 38, 0, 156, 12, 227, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 227, 12, 38, 0, 156, 12, 198, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 198, 12, 38, 0, 156, 12, 199, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 199, 12, 38, 0, 156, 12, 200, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 200, 12, 38, 0, 156, 12, 202, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 202, 12, 38, 0, 156, 12, 203, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 203, 12, 38, 0, 156, 12, 204, 12, 60, 0, 60, 0, 60, 0, 156, 12, 188, 12, 204, 12, 38, 0, 171, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 38, 0, 171, 12, 190, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 190, 12, 38, 0, 171, 12, 191, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 191, 12, 38, 0, 171, 12, 192, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 192, 12, 38, 0, 171, 12, 193, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 193, 12, 38, 0, 171, 12, 194, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 194, 12, 38, 0, 171, 12, 195, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 195, 12, 38, 0, 171, 12, 196, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 196, 12, 38, 0, 171, 12, 226, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 226, 12, 38, 0, 171, 12, 227, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 227, 12, 38, 0, 171, 12, 198, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 198, 12, 38, 0, 171, 12, 199, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 199, 12, 38, 0, 171, 12, 200, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 200, 12, 38, 0, 171, 12, 202, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 202, 12, 38, 0, 171, 12, 203, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 203, 12, 38, 0, 171, 12, 204, 12, 60, 0, 60, 0, 60, 0, 171, 12, 188, 12, 204, 12, 0, 0, 176, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 116, 0, 0, 0, 116, 1, 0, 0, 84, 25, 0, 0, 84, 25, 0, 0, 84, 25, 0, 0, 84, 25, 0, 0, 152, 25, 0, 0, 21, 0, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 111, 184, 255, 0, 112, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 41, 42, 43, 44, 45, 46, 39, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 117, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 48, 0, 48, 0, 100, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 2, 37, 111, 193, 3, 37, 111, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 4, 37, 111, 193, 5, 37, 111, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 69, 0, 60, 0, 32, 0, 119, 9, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 0, 70, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 128, 0, 0, 0, 128, 1, 0, 0, 144, 26, 0, 0, 144, 26, 0, 0, 144, 26, 0, 0, 144, 26, 0, 0, 0, 47, 0, 0, 0, 47, 0, 0, 202, 69, 0, 0, 232, 69, 0, 0, 21, 0, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 111, 184, 255, 0, 112, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 41, 42, 43, 44, 45, 46, 39, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 136, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 104, 0, 111, 0, 119, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 64, 162, 0, 202, 97, 162, 0, 202, 130, 162, 0, 202, 163, 162, 0, 202, 196, 162, 0, 202, 229, 162, 0, 202, 6, 163, 0, 202, 39, 163, 0, 202, 72, 163, 0, 202, 105, 163, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 130, 161, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 194, 161, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 2, 162, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 0, 0, 0, 193, 3, 37, 111, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 100, 11, 0, 201, 164, 20, 0, 201, 228, 29, 0, 201, 36, 39, 0, 201, 100, 48, 0, 201, 164, 57, 0, 201, 228, 66, 0, 201, 36, 76, 0, 201, 132, 95, 0, 201, 196, 104, 0, 201, 4, 114, 0, 201, 68, 123, 0, 201, 132, 132, 0, 201, 196, 141, 0, 201, 4, 151, 0, 201, 68, 160, 0, 201, 132, 169, 0, 201, 196, 178, 0, 201, 4, 188, 0, 201, 68, 197, 0, 192, 0, 0, 0, 201, 132, 206, 0, 201, 196, 215, 0, 201, 36, 235, 0, 201, 100, 244, 0, 201, 164, 253, 0, 201, 228, 6, 1, 201, 36, 16, 1, 201, 100, 25, 1, 201, 164, 34, 1, 201, 228, 43, 1, 192, 0, 0, 0, 201, 36, 53, 1, 201, 100, 62, 1, 201, 164, 71, 1, 201, 228, 80, 1, 201, 36, 90, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 100, 99, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 4, 37, 111, 193, 5, 37, 111, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 193, 10, 37, 111, 193, 6, 37, 111, 32, 5, 6, 111, 193, 10, 37, 111, 193, 7, 37, 111, 31, 5, 6, 111, 193, 10, 37, 111, 193, 8, 37, 111, 30, 5, 6, 111, 193, 10, 37, 111, 193, 9, 37, 111, 29, 5, 6, 111, 193, 10, 37, 111, 193, 10, 37, 111, 28, 5, 6, 111, 193, 15, 37, 111, 193, 11, 37, 111, 27, 5, 6, 111, 193, 15, 37, 111, 193, 12, 37, 111, 26, 5, 6, 111, 193, 15, 37, 111, 193, 13, 37, 111, 25, 5, 6, 111, 193, 15, 37, 111, 193, 14, 37, 111, 24, 5, 6, 111, 193, 15, 37, 111, 193, 15, 37, 111, 23, 5, 6, 111, 193, 20, 37, 111, 193, 16, 37, 111, 22, 5, 6, 111, 193, 20, 37, 111, 193, 17, 37, 111, 21, 5, 6, 111, 193, 20, 37, 111, 193, 18, 37, 111, 20, 5, 6, 111, 193, 20, 37, 111, 193, 19, 37, 111, 19, 5, 6, 111, 193, 20, 37, 111, 193, 20, 37, 111, 18, 5, 6, 111, 193, 25, 37, 111, 193, 21, 37, 111, 17, 5, 6, 111, 193, 25, 37, 111, 193, 22, 37, 111, 16, 5, 6, 111, 193, 25, 37, 111, 193, 23, 37, 111, 15, 5, 6, 111, 193, 25, 37, 111, 193, 24, 37, 111, 14, 5, 6, 111, 193, 25, 37, 111, 193, 25, 37, 111, 13, 5, 6, 111, 193, 30, 37, 111, 193, 26, 37, 111, 12, 5, 6, 111, 193, 30, 37, 111, 193, 27, 37, 111, 11, 5, 6, 111, 193, 30, 37, 111, 193, 28, 37, 111, 10, 5, 6, 111, 193, 30, 37, 111, 193, 29, 37, 111, 9, 5, 6, 111, 193, 30, 37, 111, 193, 30, 37, 111, 8, 5, 6, 111, 193, 6, 37, 111, 5, 5, 6, 111, 193, 6, 37, 111, 5, 5, 8, 111, 193, 6, 37, 111, 5, 5, 10, 111, 193, 6, 37, 111, 5, 5, 12, 111, 193, 6, 37, 111, 5, 5, 14, 111, 193, 6, 37, 111, 5, 5, 16, 111, 193, 6, 37, 111, 5, 5, 18, 111, 193, 6, 37, 111, 5, 5, 20, 111, 193, 6, 37, 111, 5, 5, 26, 111, 193, 6, 37, 111, 5, 5, 32, 111, 193, 6, 37, 111, 5, 5, 34, 111, 193, 6, 37, 111, 5, 5, 28, 111, 193, 6, 37, 111, 5, 5, 30, 111, 193, 6, 37, 111, 5, 5, 36, 111, 193, 6, 37, 111, 5, 5, 22, 111, 193, 6, 37, 111, 5, 5, 24, 111, 193, 7, 37, 111, 5, 5, 6, 111, 193, 7, 37, 111, 5, 5, 8, 111, 193, 7, 37, 111, 5, 5, 10, 111, 193, 7, 37, 111, 5, 5, 12, 111, 193, 7, 37, 111, 5, 5, 14, 111, 193, 7, 37, 111, 5, 5, 16, 111, 193, 7, 37, 111, 5, 5, 18, 111, 193, 7, 37, 111, 5, 5, 20, 111, 193, 7, 37, 111, 5, 5, 26, 111, 193, 7, 37, 111, 5, 5, 32, 111, 193, 7, 37, 111, 5, 5, 34, 111, 193, 7, 37, 111, 5, 5, 28, 111, 193, 7, 37, 111, 5, 5, 30, 111, 193, 7, 37, 111, 5, 5, 36, 111, 193, 7, 37, 111, 5, 5, 22, 111, 193, 7, 37, 111, 5, 5, 24, 111, 193, 8, 37, 111, 5, 5, 6, 111, 193, 8, 37, 111, 5, 5, 8, 111, 193, 8, 37, 111, 5, 5, 10, 111, 193, 8, 37, 111, 5, 5, 12, 111, 193, 8, 37, 111, 5, 5, 14, 111, 193, 8, 37, 111, 5, 5, 16, 111, 193, 8, 37, 111, 5, 5, 18, 111, 193, 8, 37, 111, 5, 5, 20, 111, 193, 8, 37, 111, 5, 5, 26, 111, 193, 8, 37, 111, 5, 5, 32, 111, 193, 8, 37, 111, 5, 5, 34, 111, 193, 8, 37, 111, 5, 5, 28, 111, 193, 8, 37, 111, 5, 5, 30, 111, 193, 8, 37, 111, 5, 5, 36, 111, 193, 8, 37, 111, 5, 5, 22, 111, 193, 8, 37, 111, 5, 5, 24, 111, 193, 9, 37, 111, 5, 5, 6, 111, 193, 9, 37, 111, 5, 5, 8, 111, 193, 9, 37, 111, 5, 5, 10, 111, 193, 9, 37, 111, 5, 5, 12, 111, 193, 9, 37, 111, 5, 5, 14, 111, 193, 9, 37, 111, 5, 5, 16, 111, 193, 9, 37, 111, 5, 5, 18, 111, 193, 9, 37, 111, 5, 5, 20, 111, 193, 9, 37, 111, 5, 5, 26, 111, 193, 9, 37, 111, 5, 5, 32, 111, 193, 9, 37, 111, 5, 5, 34, 111, 193, 9, 37, 111, 5, 5, 28, 111, 193, 9, 37, 111, 5, 5, 30, 111, 193, 9, 37, 111, 5, 5, 36, 111, 193, 9, 37, 111, 5, 5, 22, 111, 193, 9, 37, 111, 5, 5, 24, 111, 193, 10, 37, 111, 5, 5, 6, 111, 193, 10, 37, 111, 5, 5, 8, 111, 193, 10, 37, 111, 5, 5, 10, 111, 193, 10, 37, 111, 5, 5, 12, 111, 193, 10, 37, 111, 5, 5, 14, 111, 193, 10, 37, 111, 5, 5, 16, 111, 193, 10, 37, 111, 5, 5, 18, 111, 193, 10, 37, 111, 5, 5, 20, 111, 193, 10, 37, 111, 5, 5, 26, 111, 193, 10, 37, 111, 5, 5, 32, 111, 193, 10, 37, 111, 5, 5, 34, 111, 193, 10, 37, 111, 5, 5, 28, 111, 193, 10, 37, 111, 5, 5, 30, 111, 193, 10, 37, 111, 5, 5, 36, 111, 193, 10, 37, 111, 5, 5, 22, 111, 193, 10, 37, 111, 5, 5, 24, 111, 193, 11, 37, 111, 5, 5, 6, 111, 193, 11, 37, 111, 5, 5, 8, 111, 193, 11, 37, 111, 5, 5, 10, 111, 193, 11, 37, 111, 5, 5, 12, 111, 193, 11, 37, 111, 5, 5, 14, 111, 193, 11, 37, 111, 5, 5, 16, 111, 193, 11, 37, 111, 5, 5, 18, 111, 193, 11, 37, 111, 5, 5, 20, 111, 193, 11, 37, 111, 5, 5, 26, 111, 193, 11, 37, 111, 5, 5, 32, 111, 193, 11, 37, 111, 5, 5, 34, 111, 193, 11, 37, 111, 5, 5, 28, 111, 193, 11, 37, 111, 5, 5, 30, 111, 193, 11, 37, 111, 5, 5, 36, 111, 193, 11, 37, 111, 5, 5, 22, 111, 193, 11, 37, 111, 5, 5, 24, 111, 193, 12, 37, 111, 5, 5, 6, 111, 193, 12, 37, 111, 5, 5, 8, 111, 193, 12, 37, 111, 5, 5, 10, 111, 193, 12, 37, 111, 5, 5, 12, 111, 193, 12, 37, 111, 5, 5, 14, 111, 193, 12, 37, 111, 5, 5, 16, 111, 193, 12, 37, 111, 5, 5, 18, 111, 193, 12, 37, 111, 5, 5, 20, 111, 193, 12, 37, 111, 5, 5, 26, 111, 193, 12, 37, 111, 5, 5, 32, 111, 193, 12, 37, 111, 5, 5, 34, 111, 193, 12, 37, 111, 5, 5, 28, 111, 193, 12, 37, 111, 5, 5, 30, 111, 193, 12, 37, 111, 5, 5, 36, 111, 193, 12, 37, 111, 5, 5, 22, 111, 193, 12, 37, 111, 5, 5, 24, 111, 193, 13, 37, 111, 5, 5, 6, 111, 193, 13, 37, 111, 7, 5, 6, 111, 193, 13, 37, 111, 7, 5, 8, 111, 193, 13, 37, 111, 7, 5, 10, 111, 193, 13, 37, 111, 7, 5, 12, 111, 193, 13, 37, 111, 7, 5, 14, 111, 193, 13, 37, 111, 7, 5, 16, 111, 193, 13, 37, 111, 7, 5, 18, 111, 193, 13, 37, 111, 7, 5, 20, 111, 193, 13, 37, 111, 7, 5, 26, 111, 193, 13, 37, 111, 7, 5, 32, 111, 193, 13, 37, 111, 7, 5, 34, 111, 193, 13, 37, 111, 7, 5, 28, 111, 193, 13, 37, 111, 7, 5, 30, 111, 193, 13, 37, 111, 7, 5, 36, 111, 193, 13, 37, 111, 7, 5, 22, 111, 193, 13, 37, 111, 7, 5, 24, 111, 193, 13, 37, 111, 5, 5, 8, 111, 193, 13, 37, 111, 5, 5, 10, 111, 193, 13, 37, 111, 5, 5, 12, 111, 193, 13, 37, 111, 5, 5, 14, 111, 193, 13, 37, 111, 5, 5, 16, 111, 193, 13, 37, 111, 5, 5, 18, 111, 193, 13, 37, 111, 5, 5, 20, 111, 193, 13, 37, 111, 5, 5, 26, 111, 193, 13, 37, 111, 5, 5, 32, 111, 193, 13, 37, 111, 5, 5, 34, 111, 193, 13, 37, 111, 5, 5, 28, 111, 193, 13, 37, 111, 5, 5, 30, 111, 193, 13, 37, 111, 5, 5, 36, 111, 193, 13, 37, 111, 5, 5, 22, 111, 193, 13, 37, 111, 5, 5, 24, 111, 193, 14, 37, 111, 5, 5, 6, 111, 193, 14, 37, 111, 5, 5, 8, 111, 193, 14, 37, 111, 5, 5, 10, 111, 193, 14, 37, 111, 5, 5, 12, 111, 193, 14, 37, 111, 5, 5, 14, 111, 193, 14, 37, 111, 5, 5, 16, 111, 193, 14, 37, 111, 5, 5, 18, 111, 193, 14, 37, 111, 5, 5, 20, 111, 193, 14, 37, 111, 5, 5, 26, 111, 193, 14, 37, 111, 5, 5, 32, 111, 193, 14, 37, 111, 5, 5, 34, 111, 193, 14, 37, 111, 5, 5, 28, 111, 193, 14, 37, 111, 5, 5, 30, 111, 193, 14, 37, 111, 5, 5, 36, 111, 193, 14, 37, 111, 5, 5, 22, 111, 193, 14, 37, 111, 5, 5, 24, 111, 193, 15, 37, 111, 5, 5, 6, 111, 193, 15, 37, 111, 5, 5, 8, 111, 193, 15, 37, 111, 5, 5, 10, 111, 193, 15, 37, 111, 5, 5, 12, 111, 193, 15, 37, 111, 5, 5, 14, 111, 193, 15, 37, 111, 5, 5, 16, 111, 193, 15, 37, 111, 5, 5, 18, 111, 193, 15, 37, 111, 5, 5, 20, 111, 193, 15, 37, 111, 5, 5, 26, 111, 193, 15, 37, 111, 5, 5, 32, 111, 193, 15, 37, 111, 5, 5, 34, 111, 193, 15, 37, 111, 5, 5, 28, 111, 193, 15, 37, 111, 5, 5, 30, 111, 193, 15, 37, 111, 5, 5, 36, 111, 193, 15, 37, 111, 5, 5, 22, 111, 193, 15, 37, 111, 5, 5, 24, 111, 193, 16, 37, 111, 5, 5, 6, 111, 193, 16, 37, 111, 5, 5, 8, 111, 193, 16, 37, 111, 5, 5, 10, 111, 193, 16, 37, 111, 5, 5, 12, 111, 193, 16, 37, 111, 5, 5, 14, 111, 193, 16, 37, 111, 5, 5, 16, 111, 193, 16, 37, 111, 5, 5, 18, 111, 193, 16, 37, 111, 5, 5, 20, 111, 193, 16, 37, 111, 5, 5, 26, 111, 193, 16, 37, 111, 5, 5, 32, 111, 193, 16, 37, 111, 5, 5, 34, 111, 193, 16, 37, 111, 5, 5, 28, 111, 193, 16, 37, 111, 5, 5, 30, 111, 193, 16, 37, 111, 5, 5, 36, 111, 193, 16, 37, 111, 5, 5, 22, 111, 193, 16, 37, 111, 5, 5, 24, 111, 193, 17, 37, 111, 5, 5, 6, 111, 193, 17, 37, 111, 5, 5, 8, 111, 193, 17, 37, 111, 5, 5, 10, 111, 193, 17, 37, 111, 5, 5, 12, 111, 193, 17, 37, 111, 5, 5, 14, 111, 193, 17, 37, 111, 5, 5, 16, 111, 193, 17, 37, 111, 5, 5, 18, 111, 193, 17, 37, 111, 5, 5, 20, 111, 193, 17, 37, 111, 5, 5, 26, 111, 193, 17, 37, 111, 5, 5, 32, 111, 193, 17, 37, 111, 5, 5, 34, 111, 193, 17, 37, 111, 5, 5, 28, 111, 193, 17, 37, 111, 5, 5, 30, 111, 193, 17, 37, 111, 5, 5, 36, 111, 193, 17, 37, 111, 5, 5, 22, 111, 193, 17, 37, 111, 5, 5, 24, 111, 193, 18, 37, 111, 5, 5, 6, 111, 193, 18, 37, 111, 5, 5, 8, 111, 193, 18, 37, 111, 5, 5, 10, 111, 193, 18, 37, 111, 5, 5, 12, 111, 193, 18, 37, 111, 5, 5, 14, 111, 193, 18, 37, 111, 5, 5, 16, 111, 193, 18, 37, 111, 5, 5, 18, 111, 193, 18, 37, 111, 5, 5, 20, 111, 193, 18, 37, 111, 5, 5, 26, 111, 193, 18, 37, 111, 5, 5, 32, 111, 193, 18, 37, 111, 5, 5, 34, 111, 193, 18, 37, 111, 5, 5, 28, 111, 193, 18, 37, 111, 5, 5, 30, 111, 193, 18, 37, 111, 5, 5, 36, 111, 193, 18, 37, 111, 5, 5, 22, 111, 193, 18, 37, 111, 5, 5, 24, 111, 193, 19, 37, 111, 5, 5, 6, 111, 193, 19, 37, 111, 5, 5, 8, 111, 193, 19, 37, 111, 5, 5, 10, 111, 193, 19, 37, 111, 5, 5, 12, 111, 193, 19, 37, 111, 5, 5, 14, 111, 193, 19, 37, 111, 5, 5, 16, 111, 193, 19, 37, 111, 5, 5, 18, 111, 193, 19, 37, 111, 5, 5, 20, 111, 193, 19, 37, 111, 5, 5, 26, 111, 193, 19, 37, 111, 5, 5, 32, 111, 193, 19, 37, 111, 5, 5, 34, 111, 193, 19, 37, 111, 5, 5, 28, 111, 193, 19, 37, 111, 5, 5, 30, 111, 193, 19, 37, 111, 5, 5, 36, 111, 193, 19, 37, 111, 5, 5, 22, 111, 193, 19, 37, 111, 5, 5, 24, 111, 193, 20, 37, 111, 5, 5, 6, 111, 193, 20, 37, 111, 5, 5, 8, 111, 193, 20, 37, 111, 5, 5, 10, 111, 193, 20, 37, 111, 5, 5, 12, 111, 193, 20, 37, 111, 5, 5, 14, 111, 193, 20, 37, 111, 5, 5, 16, 111, 193, 20, 37, 111, 5, 5, 18, 111, 193, 20, 37, 111, 5, 5, 20, 111, 193, 20, 37, 111, 5, 5, 26, 111, 193, 20, 37, 111, 5, 5, 32, 111, 193, 20, 37, 111, 5, 5, 34, 111, 193, 20, 37, 111, 5, 5, 28, 111, 193, 20, 37, 111, 5, 5, 30, 111, 193, 20, 37, 111, 5, 5, 36, 111, 193, 20, 37, 111, 5, 5, 22, 111, 193, 20, 37, 111, 5, 5, 24, 111, 193, 21, 37, 111, 5, 5, 6, 111, 193, 21, 37, 111, 5, 5, 8, 111, 193, 21, 37, 111, 5, 5, 10, 111, 193, 21, 37, 111, 5, 5, 12, 111, 193, 21, 37, 111, 5, 5, 14, 111, 193, 21, 37, 111, 5, 5, 16, 111, 193, 21, 37, 111, 5, 5, 18, 111, 193, 21, 37, 111, 5, 5, 20, 111, 193, 21, 37, 111, 5, 5, 26, 111, 193, 21, 37, 111, 5, 5, 32, 111, 193, 21, 37, 111, 5, 5, 34, 111, 193, 21, 37, 111, 5, 5, 28, 111, 193, 21, 37, 111, 5, 5, 30, 111, 193, 21, 37, 111, 5, 5, 36, 111, 193, 21, 37, 111, 5, 5, 22, 111, 193, 21, 37, 111, 5, 5, 24, 111, 193, 22, 37, 111, 5, 5, 6, 111, 193, 22, 37, 111, 5, 5, 8, 111, 193, 22, 37, 111, 5, 5, 10, 111, 193, 22, 37, 111, 5, 5, 12, 111, 193, 22, 37, 111, 5, 5, 14, 111, 193, 22, 37, 111, 5, 5, 16, 111, 193, 22, 37, 111, 5, 5, 18, 111, 193, 22, 37, 111, 5, 5, 20, 111, 193, 22, 37, 111, 5, 5, 26, 111, 193, 22, 37, 111, 5, 5, 32, 111, 193, 22, 37, 111, 5, 5, 34, 111, 193, 22, 37, 111, 5, 5, 28, 111, 193, 22, 37, 111, 5, 5, 30, 111, 193, 22, 37, 111, 5, 5, 36, 111, 193, 22, 37, 111, 5, 5, 22, 111, 193, 22, 37, 111, 5, 5, 24, 111, 193, 23, 37, 111, 5, 5, 6, 111, 193, 23, 37, 111, 5, 5, 8, 111, 193, 23, 37, 111, 5, 5, 10, 111, 193, 23, 37, 111, 5, 5, 12, 111, 193, 23, 37, 111, 5, 5, 14, 111, 193, 23, 37, 111, 5, 5, 16, 111, 193, 23, 37, 111, 5, 5, 18, 111, 193, 23, 37, 111, 5, 5, 20, 111, 193, 23, 37, 111, 5, 5, 26, 111, 193, 23, 37, 111, 5, 5, 32, 111, 193, 23, 37, 111, 5, 5, 34, 111, 193, 23, 37, 111, 5, 5, 28, 111, 193, 23, 37, 111, 5, 5, 30, 111, 193, 23, 37, 111, 5, 5, 36, 111, 193, 23, 37, 111, 5, 5, 22, 111, 193, 23, 37, 111, 5, 5, 24, 111, 193, 24, 37, 111, 5, 5, 6, 111, 193, 24, 37, 111, 5, 5, 8, 111, 193, 24, 37, 111, 5, 5, 10, 111, 193, 24, 37, 111, 5, 5, 12, 111, 193, 24, 37, 111, 5, 5, 14, 111, 193, 24, 37, 111, 5, 5, 16, 111, 193, 24, 37, 111, 5, 5, 18, 111, 193, 24, 37, 111, 5, 5, 20, 111, 193, 24, 37, 111, 5, 5, 26, 111, 193, 24, 37, 111, 5, 5, 32, 111, 193, 24, 37, 111, 5, 5, 34, 111, 193, 24, 37, 111, 5, 5, 28, 111, 193, 24, 37, 111, 5, 5, 30, 111, 193, 24, 37, 111, 5, 5, 36, 111, 193, 24, 37, 111, 5, 5, 22, 111, 193, 24, 37, 111, 5, 5, 24, 111, 193, 25, 37, 111, 5, 5, 6, 111, 193, 25, 37, 111, 5, 5, 8, 111, 193, 25, 37, 111, 5, 5, 10, 111, 193, 25, 37, 111, 5, 5, 12, 111, 193, 25, 37, 111, 5, 5, 14, 111, 193, 25, 37, 111, 5, 5, 16, 111, 193, 25, 37, 111, 5, 5, 18, 111, 193, 25, 37, 111, 5, 5, 20, 111, 193, 25, 37, 111, 5, 5, 26, 111, 193, 25, 37, 111, 5, 5, 32, 111, 193, 25, 37, 111, 5, 5, 34, 111, 193, 25, 37, 111, 5, 5, 28, 111, 193, 25, 37, 111, 5, 5, 30, 111, 193, 25, 37, 111, 5, 5, 36, 111, 193, 25, 37, 111, 5, 5, 22, 111, 193, 25, 37, 111, 5, 5, 24, 111, 193, 26, 37, 111, 5, 5, 6, 111, 193, 26, 37, 111, 5, 5, 8, 111, 193, 26, 37, 111, 5, 5, 10, 111, 193, 26, 37, 111, 5, 5, 12, 111, 193, 26, 37, 111, 5, 5, 14, 111, 193, 26, 37, 111, 5, 5, 16, 111, 193, 26, 37, 111, 5, 5, 18, 111, 193, 26, 37, 111, 5, 5, 20, 111, 193, 26, 37, 111, 5, 5, 26, 111, 193, 26, 37, 111, 5, 5, 32, 111, 193, 26, 37, 111, 5, 5, 34, 111, 193, 26, 37, 111, 5, 5, 28, 111, 193, 26, 37, 111, 5, 5, 30, 111, 193, 26, 37, 111, 5, 5, 36, 111, 193, 26, 37, 111, 5, 5, 22, 111, 193, 26, 37, 111, 5, 5, 24, 111, 193, 27, 37, 111, 5, 5, 6, 111, 193, 27, 37, 111, 6, 5, 6, 111, 193, 27, 37, 111, 6, 5, 8, 111, 193, 27, 37, 111, 6, 5, 10, 111, 193, 27, 37, 111, 6, 5, 12, 111, 193, 27, 37, 111, 6, 5, 14, 111, 193, 27, 37, 111, 6, 5, 16, 111, 193, 27, 37, 111, 6, 5, 18, 111, 193, 27, 37, 111, 6, 5, 20, 111, 193, 27, 37, 111, 6, 5, 26, 111, 193, 27, 37, 111, 6, 5, 32, 111, 193, 27, 37, 111, 6, 5, 34, 111, 193, 27, 37, 111, 6, 5, 28, 111, 193, 27, 37, 111, 6, 5, 30, 111, 193, 27, 37, 111, 6, 5, 36, 111, 193, 27, 37, 111, 6, 5, 22, 111, 193, 27, 37, 111, 6, 5, 24, 111, 193, 27, 37, 111, 5, 5, 8, 111, 193, 27, 37, 111, 5, 5, 10, 111, 193, 27, 37, 111, 5, 5, 12, 111, 193, 27, 37, 111, 5, 5, 14, 111, 193, 27, 37, 111, 5, 5, 16, 111, 193, 27, 37, 111, 5, 5, 18, 111, 193, 27, 37, 111, 5, 5, 20, 111, 193, 27, 37, 111, 5, 5, 26, 111, 193, 27, 37, 111, 5, 5, 32, 111, 193, 27, 37, 111, 5, 5, 34, 111, 193, 27, 37, 111, 5, 5, 28, 111, 193, 27, 37, 111, 5, 5, 30, 111, 193, 27, 37, 111, 5, 5, 36, 111, 193, 27, 37, 111, 5, 5, 22, 111, 193, 27, 37, 111, 5, 5, 24, 111, 193, 28, 37, 111, 5, 5, 6, 111, 193, 28, 37, 111, 5, 5, 8, 111, 193, 28, 37, 111, 5, 5, 10, 111, 193, 28, 37, 111, 5, 5, 12, 111, 193, 28, 37, 111, 5, 5, 14, 111, 193, 28, 37, 111, 5, 5, 16, 111, 193, 28, 37, 111, 5, 5, 18, 111, 193, 28, 37, 111, 5, 5, 20, 111, 193, 28, 37, 111, 5, 5, 26, 111, 193, 28, 37, 111, 5, 5, 32, 111, 193, 28, 37, 111, 5, 5, 34, 111, 193, 28, 37, 111, 5, 5, 28, 111, 193, 28, 37, 111, 5, 5, 30, 111, 193, 28, 37, 111, 5, 5, 36, 111, 193, 28, 37, 111, 5, 5, 22, 111, 193, 28, 37, 111, 5, 5, 24, 111, 193, 29, 37, 111, 5, 5, 6, 111, 193, 29, 37, 111, 5, 5, 8, 111, 193, 29, 37, 111, 5, 5, 10, 111, 193, 29, 37, 111, 5, 5, 12, 111, 193, 29, 37, 111, 5, 5, 14, 111, 193, 29, 37, 111, 5, 5, 16, 111, 193, 29, 37, 111, 5, 5, 18, 111, 193, 29, 37, 111, 5, 5, 20, 111, 193, 29, 37, 111, 5, 5, 26, 111, 193, 29, 37, 111, 5, 5, 32, 111, 193, 29, 37, 111, 5, 5, 34, 111, 193, 29, 37, 111, 5, 5, 28, 111, 193, 29, 37, 111, 5, 5, 30, 111, 193, 29, 37, 111, 5, 5, 36, 111, 193, 29, 37, 111, 5, 5, 22, 111, 193, 29, 37, 111, 5, 5, 24, 111, 193, 30, 37, 111, 5, 5, 6, 111, 193, 30, 37, 111, 5, 5, 8, 111, 193, 30, 37, 111, 5, 5, 10, 111, 193, 30, 37, 111, 5, 5, 12, 111, 193, 30, 37, 111, 5, 5, 14, 111, 193, 30, 37, 111, 5, 5, 16, 111, 193, 30, 37, 111, 5, 5, 18, 111, 193, 30, 37, 111, 5, 5, 20, 111, 193, 30, 37, 111, 5, 5, 26, 111, 193, 30, 37, 111, 5, 5, 32, 111, 193, 30, 37, 111, 5, 5, 34, 111, 193, 30, 37, 111, 5, 5, 28, 111, 193, 30, 37, 111, 5, 5, 30, 111, 193, 30, 37, 111, 5, 5, 36, 111, 193, 30, 37, 111, 5, 5, 22, 111, 193, 30, 37, 111, 5, 5, 24, 111, 193, 31, 37, 111, 5, 5, 6, 111, 193, 31, 37, 111, 5, 5, 8, 111, 193, 31, 37, 111, 5, 5, 10, 111, 193, 31, 37, 111, 5, 5, 12, 111, 193, 31, 37, 111, 5, 5, 14, 111, 193, 31, 37, 111, 5, 5, 16, 111, 193, 31, 37, 111, 5, 5, 18, 111, 193, 31, 37, 111, 5, 5, 20, 111, 193, 31, 37, 111, 5, 5, 26, 111, 193, 31, 37, 111, 5, 5, 32, 111, 193, 31, 37, 111, 5, 5, 34, 111, 193, 31, 37, 111, 5, 5, 28, 111, 193, 31, 37, 111, 5, 5, 30, 111, 193, 31, 37, 111, 5, 5, 36, 111, 193, 31, 37, 111, 5, 5, 22, 111, 193, 31, 37, 111, 5, 5, 24, 111, 193, 32, 37, 111, 5, 5, 6, 111, 193, 32, 37, 111, 5, 5, 8, 111, 193, 32, 37, 111, 5, 5, 10, 111, 193, 32, 37, 111, 5, 5, 12, 111, 193, 32, 37, 111, 5, 5, 14, 111, 193, 32, 37, 111, 5, 5, 16, 111, 193, 32, 37, 111, 5, 5, 18, 111, 193, 32, 37, 111, 5, 5, 20, 111, 193, 32, 37, 111, 5, 5, 26, 111, 193, 32, 37, 111, 5, 5, 32, 111, 193, 32, 37, 111, 5, 5, 34, 111, 193, 32, 37, 111, 5, 5, 28, 111, 193, 32, 37, 111, 5, 5, 30, 111, 193, 32, 37, 111, 5, 5, 36, 111, 193, 32, 37, 111, 5, 5, 22, 111, 193, 32, 37, 111, 5, 5, 24, 111, 193, 33, 37, 111, 5, 5, 6, 111, 193, 33, 37, 111, 5, 5, 8, 111, 193, 33, 37, 111, 5, 5, 10, 111, 193, 33, 37, 111, 5, 5, 12, 111, 193, 33, 37, 111, 5, 5, 14, 111, 193, 33, 37, 111, 5, 5, 16, 111, 193, 33, 37, 111, 5, 5, 18, 111, 193, 33, 37, 111, 5, 5, 20, 111, 193, 33, 37, 111, 5, 5, 26, 111, 193, 33, 37, 111, 5, 5, 32, 111, 193, 33, 37, 111, 5, 5, 34, 111, 193, 33, 37, 111, 5, 5, 28, 111, 193, 33, 37, 111, 5, 5, 30, 111, 193, 33, 37, 111, 5, 5, 36, 111, 193, 33, 37, 111, 5, 5, 22, 111, 193, 33, 37, 111, 5, 5, 24, 111, 193, 34, 37, 111, 5, 5, 6, 111, 193, 34, 37, 111, 5, 5, 8, 111, 193, 34, 37, 111, 5, 5, 10, 111, 193, 34, 37, 111, 5, 5, 12, 111, 193, 34, 37, 111, 5, 5, 14, 111, 193, 34, 37, 111, 5, 5, 16, 111, 193, 34, 37, 111, 5, 5, 18, 111, 193, 34, 37, 111, 5, 5, 20, 111, 193, 34, 37, 111, 5, 5, 26, 111, 193, 34, 37, 111, 5, 5, 32, 111, 193, 34, 37, 111, 5, 5, 34, 111, 193, 34, 37, 111, 5, 5, 28, 111, 193, 34, 37, 111, 5, 5, 30, 111, 193, 34, 37, 111, 5, 5, 36, 111, 193, 34, 37, 111, 5, 5, 22, 111, 193, 34, 37, 111, 5, 5, 24, 111, 193, 40, 37, 111, 5, 5, 6, 111, 193, 40, 37, 111, 5, 5, 8, 111, 193, 40, 37, 111, 5, 5, 10, 111, 193, 40, 37, 111, 5, 5, 12, 111, 193, 40, 37, 111, 5, 5, 14, 111, 193, 40, 37, 111, 5, 5, 16, 111, 193, 40, 37, 111, 5, 5, 18, 111, 193, 40, 37, 111, 5, 5, 20, 111, 193, 40, 37, 111, 5, 5, 26, 111, 193, 40, 37, 111, 5, 5, 32, 111, 193, 40, 37, 111, 5, 5, 34, 111, 193, 40, 37, 111, 5, 5, 28, 111, 193, 40, 37, 111, 5, 5, 30, 111, 193, 40, 37, 111, 5, 5, 36, 111, 193, 40, 37, 111, 5, 5, 22, 111, 193, 40, 37, 111, 5, 5, 24, 111, 193, 35, 37, 111, 5, 5, 6, 111, 193, 35, 37, 111, 5, 5, 8, 111, 193, 35, 37, 111, 5, 5, 10, 111, 193, 35, 37, 111, 5, 5, 12, 111, 193, 35, 37, 111, 5, 5, 14, 111, 193, 35, 37, 111, 5, 5, 16, 111, 193, 35, 37, 111, 5, 5, 18, 111, 193, 35, 37, 111, 5, 5, 20, 111, 193, 35, 37, 111, 5, 5, 26, 111, 193, 35, 37, 111, 5, 5, 32, 111, 193, 35, 37, 111, 5, 5, 34, 111, 193, 35, 37, 111, 5, 5, 28, 111, 193, 35, 37, 111, 5, 5, 30, 111, 193, 35, 37, 111, 5, 5, 36, 111, 193, 35, 37, 111, 5, 5, 22, 111, 193, 35, 37, 111, 5, 5, 24, 111, 193, 36, 37, 111, 5, 5, 6, 111, 193, 36, 37, 111, 5, 5, 8, 111, 193, 36, 37, 111, 5, 5, 10, 111, 193, 36, 37, 111, 5, 5, 12, 111, 193, 36, 37, 111, 5, 5, 14, 111, 193, 36, 37, 111, 5, 5, 16, 111, 193, 36, 37, 111, 5, 5, 18, 111, 193, 36, 37, 111, 5, 5, 20, 111, 193, 36, 37, 111, 5, 5, 26, 111, 193, 36, 37, 111, 5, 5, 32, 111, 193, 36, 37, 111, 5, 5, 34, 111, 193, 36, 37, 111, 5, 5, 28, 111, 193, 36, 37, 111, 5, 5, 30, 111, 193, 36, 37, 111, 5, 5, 36, 111, 193, 36, 37, 111, 5, 5, 22, 111, 193, 36, 37, 111, 5, 5, 24, 111, 193, 37, 37, 111, 5, 5, 6, 111, 193, 37, 37, 111, 5, 5, 8, 111, 193, 37, 37, 111, 5, 5, 10, 111, 193, 37, 37, 111, 5, 5, 12, 111, 193, 37, 37, 111, 5, 5, 14, 111, 193, 37, 37, 111, 5, 5, 16, 111, 193, 37, 37, 111, 5, 5, 18, 111, 193, 37, 37, 111, 5, 5, 20, 111, 193, 37, 37, 111, 5, 5, 26, 111, 193, 37, 37, 111, 5, 5, 32, 111, 193, 37, 37, 111, 5, 5, 34, 111, 193, 37, 37, 111, 5, 5, 28, 111, 193, 37, 37, 111, 5, 5, 30, 111, 193, 37, 37, 111, 5, 5, 36, 111, 193, 37, 37, 111, 5, 5, 22, 111, 193, 37, 37, 111, 5, 5, 24, 111, 193, 38, 37, 111, 5, 5, 6, 111, 193, 38, 37, 111, 5, 5, 8, 111, 193, 38, 37, 111, 5, 5, 10, 111, 193, 38, 37, 111, 5, 5, 12, 111, 193, 38, 37, 111, 5, 5, 14, 111, 193, 38, 37, 111, 5, 5, 16, 111, 193, 38, 37, 111, 5, 5, 18, 111, 193, 38, 37, 111, 5, 5, 20, 111, 193, 38, 37, 111, 5, 5, 26, 111, 193, 38, 37, 111, 5, 5, 32, 111, 193, 38, 37, 111, 5, 5, 34, 111, 193, 38, 37, 111, 5, 5, 28, 111, 193, 38, 37, 111, 5, 5, 30, 111, 193, 38, 37, 111, 5, 5, 36, 111, 193, 38, 37, 111, 5, 5, 22, 111, 193, 38, 37, 111, 5, 5, 24, 111, 193, 39, 37, 111, 5, 5, 6, 111, 193, 39, 37, 111, 5, 5, 8, 111, 193, 39, 37, 111, 5, 5, 10, 111, 193, 39, 37, 111, 5, 5, 12, 111, 193, 39, 37, 111, 5, 5, 14, 111, 193, 39, 37, 111, 5, 5, 16, 111, 193, 39, 37, 111, 5, 5, 18, 111, 193, 39, 37, 111, 5, 5, 20, 111, 193, 39, 37, 111, 5, 5, 26, 111, 193, 39, 37, 111, 5, 5, 32, 111, 193, 39, 37, 111, 5, 5, 34, 111, 193, 39, 37, 111, 5, 5, 28, 111, 193, 39, 37, 111, 5, 5, 30, 111, 193, 39, 37, 111, 5, 5, 36, 111, 193, 39, 37, 111, 5, 5, 22, 111, 193, 39, 37, 111, 5, 5, 24, 111, 193, 41, 37, 111, 5, 5, 6, 111, 193, 41, 37, 111, 5, 5, 8, 111, 193, 41, 37, 111, 5, 5, 10, 111, 193, 41, 37, 111, 5, 5, 12, 111, 193, 41, 37, 111, 5, 5, 14, 111, 193, 41, 37, 111, 5, 5, 16, 111, 193, 41, 37, 111, 5, 5, 18, 111, 193, 41, 37, 111, 5, 5, 20, 111, 193, 41, 37, 111, 5, 5, 26, 111, 193, 41, 37, 111, 5, 5, 32, 111, 193, 41, 37, 111, 5, 5, 34, 111, 193, 41, 37, 111, 5, 5, 28, 111, 193, 41, 37, 111, 5, 5, 30, 111, 193, 41, 37, 111, 5, 5, 36, 111, 193, 41, 37, 111, 5, 5, 22, 111, 193, 41, 37, 111, 5, 5, 24, 111, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 37, 111, 193, 2, 24, 0, 161, 12, 45, 0, 167, 12, 23, 0, 171, 12, 12, 0, 171, 12, 8, 192, 197, 3, 172, 12, 8, 192, 197, 99, 173, 12, 8, 192, 197, 195, 174, 12, 9, 192, 197, 35, 167, 12, 6, 192, 197, 227, 168, 12, 7, 192, 197, 67, 170, 12, 7, 192, 197, 163, 164, 12, 9, 0, 164, 12, 5, 192, 197, 195, 165, 12, 6, 192, 197, 35, 166, 12, 6, 192, 197, 131, 161, 12, 4, 192, 197, 163, 162, 12, 5, 192, 197, 3, 163, 12, 5, 192, 197, 99, 155, 12, 20, 0, 158, 12, 9, 0, 158, 12, 3, 192, 197, 131, 159, 12, 3, 192, 197, 227, 160, 12, 4, 192, 197, 67, 155, 12, 2, 192, 197, 99, 156, 12, 2, 192, 197, 195, 157, 12, 3, 192, 197, 35, 152, 12, 9, 0, 152, 12, 1, 192, 197, 67, 153, 12, 1, 192, 197, 163, 154, 12, 2, 192, 197, 3, 149, 12, 197, 163, 150, 12, 0, 192, 197, 131, 151, 12, 0, 192, 197, 227, 9, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 12, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 6, 226, 12, 13, 192, 197, 2, 227, 12, 13, 192, 197, 66, 199, 12, 12, 192, 197, 66, 200, 12, 12, 192, 197, 130, 202, 12, 25, 0, 203, 12, 12, 192, 197, 2, 194, 12, 25, 0, 194, 12, 10, 192, 197, 194, 195, 12, 11, 192, 197, 2, 196, 12, 11, 192, 197, 66, 198, 12, 2, 67, 197, 130, 194, 12, 6, 0, 213, 12, 12, 192, 197, 66, 214, 12, 12, 192, 197, 130, 48, 67, 197, 194, 213, 12, 12, 192, 197, 2, 190, 12, 9, 192, 197, 194, 191, 12, 6, 0, 192, 12, 10, 192, 197, 66, 193, 12, 10, 192, 197, 130, 240, 66, 197, 2, 213, 12, 10, 192, 197, 66, 13, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 16, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 7, 226, 12, 17, 192, 197, 2, 227, 12, 17, 192, 197, 66, 199, 12, 16, 192, 197, 66, 200, 12, 16, 192, 197, 130, 202, 12, 25, 0, 203, 12, 16, 192, 197, 2, 194, 12, 25, 0, 194, 12, 14, 192, 197, 194, 195, 12, 15, 192, 197, 2, 196, 12, 15, 192, 197, 66, 198, 12, 2, 68, 197, 130, 194, 12, 6, 0, 213, 12, 16, 192, 197, 66, 214, 12, 16, 192, 197, 130, 48, 68, 197, 194, 213, 12, 16, 192, 197, 2, 190, 12, 13, 192, 197, 194, 191, 12, 6, 0, 192, 12, 14, 192, 197, 66, 193, 12, 14, 192, 197, 130, 240, 67, 197, 2, 213, 12, 14, 192, 197, 66, 17, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 20, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 8, 226, 12, 21, 192, 197, 2, 227, 12, 21, 192, 197, 66, 199, 12, 20, 192, 197, 66, 200, 12, 20, 192, 197, 130, 202, 12, 25, 0, 203, 12, 20, 192, 197, 2, 194, 12, 25, 0, 194, 12, 18, 192, 197, 194, 195, 12, 19, 192, 197, 2, 196, 12, 19, 192, 197, 66, 198, 12, 2, 69, 197, 130, 194, 12, 6, 0, 213, 12, 20, 192, 197, 66, 214, 12, 20, 192, 197, 130, 48, 69, 197, 194, 213, 12, 20, 192, 197, 2, 190, 12, 17, 192, 197, 194, 191, 12, 6, 0, 192, 12, 18, 192, 197, 66, 193, 12, 18, 192, 197, 130, 240, 68, 197, 2, 213, 12, 18, 192, 197, 66, 21, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 24, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 9, 226, 12, 25, 192, 197, 2, 227, 12, 25, 192, 197, 66, 199, 12, 24, 192, 197, 66, 200, 12, 24, 192, 197, 130, 202, 12, 25, 0, 203, 12, 24, 192, 197, 2, 194, 12, 25, 0, 194, 12, 22, 192, 197, 194, 195, 12, 23, 192, 197, 2, 196, 12, 23, 192, 197, 66, 198, 12, 2, 70, 197, 130, 194, 12, 6, 0, 213, 12, 24, 192, 197, 66, 214, 12, 24, 192, 197, 130, 48, 70, 197, 194, 213, 12, 24, 192, 197, 2, 190, 12, 21, 192, 197, 194, 191, 12, 6, 0, 192, 12, 22, 192, 197, 66, 193, 12, 22, 192, 197, 130, 240, 69, 197, 2, 213, 12, 22, 192, 197, 66, 25, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 28, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 10, 226, 12, 29, 192, 197, 2, 227, 12, 29, 192, 197, 66, 199, 12, 28, 192, 197, 66, 200, 12, 28, 192, 197, 130, 202, 12, 25, 0, 203, 12, 28, 192, 197, 2, 194, 12, 25, 0, 194, 12, 26, 192, 197, 194, 195, 12, 27, 192, 197, 2, 196, 12, 27, 192, 197, 66, 198, 12, 2, 71, 197, 130, 194, 12, 6, 0, 213, 12, 28, 192, 197, 66, 214, 12, 28, 192, 197, 130, 48, 71, 197, 194, 213, 12, 28, 192, 197, 2, 190, 12, 25, 192, 197, 194, 191, 12, 6, 0, 192, 12, 26, 192, 197, 66, 193, 12, 26, 192, 197, 130, 240, 70, 197, 2, 213, 12, 26, 192, 197, 66, 29, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 32, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 11, 226, 12, 33, 192, 197, 2, 227, 12, 33, 192, 197, 66, 199, 12, 32, 192, 197, 66, 200, 12, 32, 192, 197, 130, 202, 12, 25, 0, 203, 12, 32, 192, 197, 2, 194, 12, 25, 0, 194, 12, 30, 192, 197, 194, 195, 12, 31, 192, 197, 2, 196, 12, 31, 192, 197, 66, 198, 12, 2, 72, 197, 130, 194, 12, 6, 0, 213, 12, 32, 192, 197, 66, 214, 12, 32, 192, 197, 130, 48, 72, 197, 194, 213, 12, 32, 192, 197, 2, 190, 12, 29, 192, 197, 194, 191, 12, 6, 0, 192, 12, 30, 192, 197, 66, 193, 12, 30, 192, 197, 130, 240, 71, 197, 2, 213, 12, 30, 192, 197, 66, 33, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 36, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 12, 226, 12, 37, 192, 197, 2, 227, 12, 37, 192, 197, 66, 199, 12, 36, 192, 197, 66, 200, 12, 36, 192, 197, 130, 202, 12, 25, 0, 203, 12, 36, 192, 197, 2, 194, 12, 25, 0, 194, 12, 34, 192, 197, 194, 195, 12, 35, 192, 197, 2, 196, 12, 35, 192, 197, 66, 198, 12, 2, 73, 197, 130, 194, 12, 6, 0, 213, 12, 36, 192, 197, 66, 214, 12, 36, 192, 197, 130, 48, 73, 197, 194, 213, 12, 36, 192, 197, 2, 190, 12, 33, 192, 197, 194, 191, 12, 6, 0, 192, 12, 34, 192, 197, 66, 193, 12, 34, 192, 197, 130, 240, 72, 197, 2, 213, 12, 34, 192, 197, 66, 37, 0, 197, 130, 16, 0, 198, 12, 47, 0, 203, 12, 16, 0, 203, 12, 44, 192, 197, 2, 204, 12, 44, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 13, 226, 12, 45, 192, 197, 2, 227, 12, 45, 192, 197, 66, 198, 12, 12, 0, 199, 12, 44, 192, 197, 66, 200, 12, 44, 192, 197, 130, 202, 12, 48, 75, 197, 194, 213, 12, 44, 192, 197, 2, 2, 75, 197, 130, 194, 12, 6, 0, 213, 12, 44, 192, 197, 66, 214, 12, 44, 192, 197, 130, 48, 75, 197, 194, 213, 12, 44, 192, 197, 2, 193, 12, 12, 0, 193, 12, 42, 192, 197, 130, 194, 12, 42, 192, 197, 194, 195, 12, 43, 192, 197, 2, 196, 12, 43, 192, 197, 66, 188, 12, 8, 0, 190, 12, 41, 192, 197, 194, 191, 12, 77, 0, 192, 12, 42, 192, 197, 66, 142, 73, 197, 194, 198, 12, 43, 0, 203, 12, 12, 0, 203, 12, 40, 192, 197, 66, 204, 12, 41, 192, 197, 2, 226, 12, 41, 192, 197, 66, 227, 12, 41, 192, 197, 130, 198, 12, 12, 0, 199, 12, 40, 192, 197, 130, 200, 12, 40, 192, 197, 194, 202, 12, 112, 74, 197, 2, 213, 12, 40, 192, 197, 66, 2, 74, 197, 194, 194, 12, 6, 0, 213, 12, 40, 192, 197, 130, 214, 12, 40, 192, 197, 194, 112, 74, 197, 2, 213, 12, 40, 192, 197, 66, 193, 12, 12, 0, 193, 12, 38, 192, 197, 194, 194, 12, 39, 192, 197, 2, 195, 12, 39, 192, 197, 66, 196, 12, 39, 192, 197, 130, 190, 12, 38, 192, 197, 2, 191, 12, 3, 0, 192, 12, 38, 192, 197, 130, 240, 73, 197, 66, 213, 12, 38, 192, 197, 130, 240, 74, 197, 2, 213, 12, 42, 192, 197, 66, 45, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 48, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 14, 226, 12, 49, 192, 197, 2, 227, 12, 49, 192, 197, 66, 199, 12, 48, 192, 197, 66, 200, 12, 48, 192, 197, 130, 202, 12, 25, 0, 203, 12, 48, 192, 197, 2, 194, 12, 25, 0, 194, 12, 46, 192, 197, 194, 195, 12, 47, 192, 197, 2, 196, 12, 47, 192, 197, 66, 198, 12, 2, 76, 197, 130, 194, 12, 6, 0, 213, 12, 48, 192, 197, 66, 214, 12, 48, 192, 197, 130, 48, 76, 197, 194, 213, 12, 48, 192, 197, 2, 190, 12, 45, 192, 197, 194, 191, 12, 6, 0, 192, 12, 46, 192, 197, 66, 193, 12, 46, 192, 197, 130, 240, 75, 197, 2, 213, 12, 46, 192, 197, 66, 49, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 52, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 15, 226, 12, 53, 192, 197, 2, 227, 12, 53, 192, 197, 66, 199, 12, 52, 192, 197, 66, 200, 12, 52, 192, 197, 130, 202, 12, 25, 0, 203, 12, 52, 192, 197, 2, 194, 12, 25, 0, 194, 12, 50, 192, 197, 194, 195, 12, 51, 192, 197, 2, 196, 12, 51, 192, 197, 66, 198, 12, 2, 77, 197, 130, 194, 12, 6, 0, 213, 12, 52, 192, 197, 66, 214, 12, 52, 192, 197, 130, 48, 77, 197, 194, 213, 12, 52, 192, 197, 2, 190, 12, 49, 192, 197, 194, 191, 12, 6, 0, 192, 12, 50, 192, 197, 66, 193, 12, 50, 192, 197, 130, 240, 76, 197, 2, 213, 12, 50, 192, 197, 66, 53, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 56, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 16, 226, 12, 57, 192, 197, 2, 227, 12, 57, 192, 197, 66, 199, 12, 56, 192, 197, 66, 200, 12, 56, 192, 197, 130, 202, 12, 25, 0, 203, 12, 56, 192, 197, 2, 194, 12, 25, 0, 194, 12, 54, 192, 197, 194, 195, 12, 55, 192, 197, 2, 196, 12, 55, 192, 197, 66, 198, 12, 2, 78, 197, 130, 194, 12, 6, 0, 213, 12, 56, 192, 197, 66, 214, 12, 56, 192, 197, 130, 48, 78, 197, 194, 213, 12, 56, 192, 197, 2, 190, 12, 53, 192, 197, 194, 191, 12, 6, 0, 192, 12, 54, 192, 197, 66, 193, 12, 54, 192, 197, 130, 240, 77, 197, 2, 213, 12, 54, 192, 197, 66, 57, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 60, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 17, 226, 12, 61, 192, 197, 2, 227, 12, 61, 192, 197, 66, 199, 12, 60, 192, 197, 66, 200, 12, 60, 192, 197, 130, 202, 12, 25, 0, 203, 12, 60, 192, 197, 2, 194, 12, 25, 0, 194, 12, 58, 192, 197, 194, 195, 12, 59, 192, 197, 2, 196, 12, 59, 192, 197, 66, 198, 12, 2, 79, 197, 130, 194, 12, 6, 0, 213, 12, 60, 192, 197, 66, 214, 12, 60, 192, 197, 130, 48, 79, 197, 194, 213, 12, 60, 192, 197, 2, 190, 12, 57, 192, 197, 194, 191, 12, 6, 0, 192, 12, 58, 192, 197, 66, 193, 12, 58, 192, 197, 130, 240, 78, 197, 2, 213, 12, 58, 192, 197, 66, 61, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 64, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 18, 226, 12, 65, 192, 197, 2, 227, 12, 65, 192, 197, 66, 199, 12, 64, 192, 197, 66, 200, 12, 64, 192, 197, 130, 202, 12, 25, 0, 203, 12, 64, 192, 197, 2, 194, 12, 25, 0, 194, 12, 62, 192, 197, 194, 195, 12, 63, 192, 197, 2, 196, 12, 63, 192, 197, 66, 198, 12, 2, 80, 197, 130, 194, 12, 6, 0, 213, 12, 64, 192, 197, 66, 214, 12, 64, 192, 197, 130, 48, 80, 197, 194, 213, 12, 64, 192, 197, 2, 190, 12, 61, 192, 197, 194, 191, 12, 6, 0, 192, 12, 62, 192, 197, 66, 193, 12, 62, 192, 197, 130, 240, 79, 197, 2, 213, 12, 62, 192, 197, 66, 65, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 68, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 19, 226, 12, 69, 192, 197, 2, 227, 12, 69, 192, 197, 66, 199, 12, 68, 192, 197, 66, 200, 12, 68, 192, 197, 130, 202, 12, 25, 0, 203, 12, 68, 192, 197, 2, 194, 12, 25, 0, 194, 12, 66, 192, 197, 194, 195, 12, 67, 192, 197, 2, 196, 12, 67, 192, 197, 66, 198, 12, 2, 81, 197, 130, 194, 12, 6, 0, 213, 12, 68, 192, 197, 66, 214, 12, 68, 192, 197, 130, 48, 81, 197, 194, 213, 12, 68, 192, 197, 2, 190, 12, 65, 192, 197, 194, 191, 12, 6, 0, 192, 12, 66, 192, 197, 66, 193, 12, 66, 192, 197, 130, 240, 80, 197, 2, 213, 12, 66, 192, 197, 66, 69, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 72, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 20, 226, 12, 73, 192, 197, 2, 227, 12, 73, 192, 197, 66, 199, 12, 72, 192, 197, 66, 200, 12, 72, 192, 197, 130, 202, 12, 25, 0, 203, 12, 72, 192, 197, 2, 194, 12, 25, 0, 194, 12, 70, 192, 197, 194, 195, 12, 71, 192, 197, 2, 196, 12, 71, 192, 197, 66, 198, 12, 2, 82, 197, 130, 194, 12, 6, 0, 213, 12, 72, 192, 197, 66, 214, 12, 72, 192, 197, 130, 48, 82, 197, 194, 213, 12, 72, 192, 197, 2, 190, 12, 69, 192, 197, 194, 191, 12, 6, 0, 192, 12, 70, 192, 197, 66, 193, 12, 70, 192, 197, 130, 240, 81, 197, 2, 213, 12, 70, 192, 197, 66, 73, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 76, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 21, 226, 12, 77, 192, 197, 2, 227, 12, 77, 192, 197, 66, 199, 12, 76, 192, 197, 66, 200, 12, 76, 192, 197, 130, 202, 12, 25, 0, 203, 12, 76, 192, 197, 2, 194, 12, 25, 0, 194, 12, 74, 192, 197, 194, 195, 12, 75, 192, 197, 2, 196, 12, 75, 192, 197, 66, 198, 12, 2, 83, 197, 130, 194, 12, 6, 0, 213, 12, 76, 192, 197, 66, 214, 12, 76, 192, 197, 130, 48, 83, 197, 194, 213, 12, 76, 192, 197, 2, 190, 12, 73, 192, 197, 194, 191, 12, 6, 0, 192, 12, 74, 192, 197, 66, 193, 12, 74, 192, 197, 130, 240, 82, 197, 2, 213, 12, 74, 192, 197, 66, 77, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 80, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 22, 226, 12, 81, 192, 197, 2, 227, 12, 81, 192, 197, 66, 199, 12, 80, 192, 197, 66, 200, 12, 80, 192, 197, 130, 202, 12, 25, 0, 203, 12, 80, 192, 197, 2, 194, 12, 25, 0, 194, 12, 78, 192, 197, 194, 195, 12, 79, 192, 197, 2, 196, 12, 79, 192, 197, 66, 198, 12, 2, 84, 197, 130, 194, 12, 6, 0, 213, 12, 80, 192, 197, 66, 214, 12, 80, 192, 197, 130, 48, 84, 197, 194, 213, 12, 80, 192, 197, 2, 190, 12, 77, 192, 197, 194, 191, 12, 6, 0, 192, 12, 78, 192, 197, 66, 193, 12, 78, 192, 197, 130, 240, 83, 197, 2, 213, 12, 78, 192, 197, 66, 81, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 84, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 23, 226, 12, 85, 192, 197, 2, 227, 12, 85, 192, 197, 66, 199, 12, 84, 192, 197, 66, 200, 12, 84, 192, 197, 130, 202, 12, 25, 0, 203, 12, 84, 192, 197, 2, 194, 12, 25, 0, 194, 12, 82, 192, 197, 194, 195, 12, 83, 192, 197, 2, 196, 12, 83, 192, 197, 66, 198, 12, 2, 85, 197, 130, 194, 12, 6, 0, 213, 12, 84, 192, 197, 66, 214, 12, 84, 192, 197, 130, 48, 85, 197, 194, 213, 12, 84, 192, 197, 2, 190, 12, 81, 192, 197, 194, 191, 12, 6, 0, 192, 12, 82, 192, 197, 66, 193, 12, 82, 192, 197, 130, 240, 84, 197, 2, 213, 12, 82, 192, 197, 66, 85, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 88, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 24, 226, 12, 89, 192, 197, 2, 227, 12, 89, 192, 197, 66, 199, 12, 88, 192, 197, 66, 200, 12, 88, 192, 197, 130, 202, 12, 25, 0, 203, 12, 88, 192, 197, 2, 194, 12, 25, 0, 194, 12, 86, 192, 197, 194, 195, 12, 87, 192, 197, 2, 196, 12, 87, 192, 197, 66, 198, 12, 2, 86, 197, 130, 194, 12, 6, 0, 213, 12, 88, 192, 197, 66, 214, 12, 88, 192, 197, 130, 48, 86, 197, 194, 213, 12, 88, 192, 197, 2, 190, 12, 85, 192, 197, 194, 191, 12, 6, 0, 192, 12, 86, 192, 197, 66, 193, 12, 86, 192, 197, 130, 240, 85, 197, 2, 213, 12, 86, 192, 197, 66, 89, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 92, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 25, 226, 12, 93, 192, 197, 2, 227, 12, 93, 192, 197, 66, 199, 12, 92, 192, 197, 66, 200, 12, 92, 192, 197, 130, 202, 12, 25, 0, 203, 12, 92, 192, 197, 2, 194, 12, 25, 0, 194, 12, 90, 192, 197, 194, 195, 12, 91, 192, 197, 2, 196, 12, 91, 192, 197, 66, 198, 12, 2, 87, 197, 130, 194, 12, 6, 0, 213, 12, 92, 192, 197, 66, 214, 12, 92, 192, 197, 130, 48, 87, 197, 194, 213, 12, 92, 192, 197, 2, 190, 12, 89, 192, 197, 194, 191, 12, 6, 0, 192, 12, 90, 192, 197, 66, 193, 12, 90, 192, 197, 130, 240, 86, 197, 2, 213, 12, 90, 192, 197, 66, 93, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 96, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 26, 226, 12, 97, 192, 197, 2, 227, 12, 97, 192, 197, 66, 199, 12, 96, 192, 197, 66, 200, 12, 96, 192, 197, 130, 202, 12, 25, 0, 203, 12, 96, 192, 197, 2, 194, 12, 25, 0, 194, 12, 94, 192, 197, 194, 195, 12, 95, 192, 197, 2, 196, 12, 95, 192, 197, 66, 198, 12, 2, 88, 197, 130, 194, 12, 6, 0, 213, 12, 96, 192, 197, 66, 214, 12, 96, 192, 197, 130, 48, 88, 197, 194, 213, 12, 96, 192, 197, 2, 190, 12, 93, 192, 197, 194, 191, 12, 6, 0, 192, 12, 94, 192, 197, 66, 193, 12, 94, 192, 197, 130, 240, 87, 197, 2, 213, 12, 94, 192, 197, 66, 97, 0, 197, 130, 16, 0, 198, 12, 47, 0, 203, 12, 16, 0, 203, 12, 104, 192, 197, 2, 204, 12, 104, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 27, 226, 12, 105, 192, 197, 2, 227, 12, 105, 192, 197, 66, 198, 12, 12, 0, 199, 12, 104, 192, 197, 66, 200, 12, 104, 192, 197, 130, 202, 12, 48, 90, 197, 194, 213, 12, 104, 192, 197, 2, 2, 90, 197, 130, 194, 12, 6, 0, 213, 12, 104, 192, 197, 66, 214, 12, 104, 192, 197, 130, 48, 90, 197, 194, 213, 12, 104, 192, 197, 2, 193, 12, 12, 0, 193, 12, 102, 192, 197, 130, 194, 12, 102, 192, 197, 194, 195, 12, 103, 192, 197, 2, 196, 12, 103, 192, 197, 66, 188, 12, 8, 0, 190, 12, 101, 192, 197, 194, 191, 12, 77, 0, 192, 12, 102, 192, 197, 66, 142, 88, 197, 194, 198, 12, 43, 0, 203, 12, 12, 0, 203, 12, 100, 192, 197, 66, 204, 12, 101, 192, 197, 2, 226, 12, 101, 192, 197, 66, 227, 12, 101, 192, 197, 130, 198, 12, 12, 0, 199, 12, 100, 192, 197, 130, 200, 12, 100, 192, 197, 194, 202, 12, 112, 89, 197, 2, 213, 12, 100, 192, 197, 66, 2, 89, 197, 194, 194, 12, 6, 0, 213, 12, 100, 192, 197, 130, 214, 12, 100, 192, 197, 194, 112, 89, 197, 2, 213, 12, 100, 192, 197, 66, 193, 12, 12, 0, 193, 12, 98, 192, 197, 194, 194, 12, 99, 192, 197, 2, 195, 12, 99, 192, 197, 66, 196, 12, 99, 192, 197, 130, 190, 12, 98, 192, 197, 2, 191, 12, 3, 0, 192, 12, 98, 192, 197, 130, 240, 88, 197, 66, 213, 12, 98, 192, 197, 130, 240, 89, 197, 2, 213, 12, 102, 192, 197, 66, 105, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 108, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 28, 226, 12, 109, 192, 197, 2, 227, 12, 109, 192, 197, 66, 199, 12, 108, 192, 197, 66, 200, 12, 108, 192, 197, 130, 202, 12, 25, 0, 203, 12, 108, 192, 197, 2, 194, 12, 25, 0, 194, 12, 106, 192, 197, 194, 195, 12, 107, 192, 197, 2, 196, 12, 107, 192, 197, 66, 198, 12, 2, 91, 197, 130, 194, 12, 6, 0, 213, 12, 108, 192, 197, 66, 214, 12, 108, 192, 197, 130, 48, 91, 197, 194, 213, 12, 108, 192, 197, 2, 190, 12, 105, 192, 197, 194, 191, 12, 6, 0, 192, 12, 106, 192, 197, 66, 193, 12, 106, 192, 197, 130, 240, 90, 197, 2, 213, 12, 106, 192, 197, 66, 109, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 112, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 29, 226, 12, 113, 192, 197, 2, 227, 12, 113, 192, 197, 66, 199, 12, 112, 192, 197, 66, 200, 12, 112, 192, 197, 130, 202, 12, 25, 0, 203, 12, 112, 192, 197, 2, 194, 12, 25, 0, 194, 12, 110, 192, 197, 194, 195, 12, 111, 192, 197, 2, 196, 12, 111, 192, 197, 66, 198, 12, 2, 92, 197, 130, 194, 12, 6, 0, 213, 12, 112, 192, 197, 66, 214, 12, 112, 192, 197, 130, 48, 92, 197, 194, 213, 12, 112, 192, 197, 2, 190, 12, 109, 192, 197, 194, 191, 12, 6, 0, 192, 12, 110, 192, 197, 66, 193, 12, 110, 192, 197, 130, 240, 91, 197, 2, 213, 12, 110, 192, 197, 66, 113, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 116, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 30, 226, 12, 117, 192, 197, 2, 227, 12, 117, 192, 197, 66, 199, 12, 116, 192, 197, 66, 200, 12, 116, 192, 197, 130, 202, 12, 25, 0, 203, 12, 116, 192, 197, 2, 194, 12, 25, 0, 194, 12, 114, 192, 197, 194, 195, 12, 115, 192, 197, 2, 196, 12, 115, 192, 197, 66, 198, 12, 2, 93, 197, 130, 194, 12, 6, 0, 213, 12, 116, 192, 197, 66, 214, 12, 116, 192, 197, 130, 48, 93, 197, 194, 213, 12, 116, 192, 197, 2, 190, 12, 113, 192, 197, 194, 191, 12, 6, 0, 192, 12, 114, 192, 197, 66, 193, 12, 114, 192, 197, 130, 240, 92, 197, 2, 213, 12, 114, 192, 197, 66, 117, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 120, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 31, 226, 12, 121, 192, 197, 2, 227, 12, 121, 192, 197, 66, 199, 12, 120, 192, 197, 66, 200, 12, 120, 192, 197, 130, 202, 12, 25, 0, 203, 12, 120, 192, 197, 2, 194, 12, 25, 0, 194, 12, 118, 192, 197, 194, 195, 12, 119, 192, 197, 2, 196, 12, 119, 192, 197, 66, 198, 12, 2, 94, 197, 130, 194, 12, 6, 0, 213, 12, 120, 192, 197, 66, 214, 12, 120, 192, 197, 130, 48, 94, 197, 194, 213, 12, 120, 192, 197, 2, 190, 12, 117, 192, 197, 194, 191, 12, 6, 0, 192, 12, 118, 192, 197, 66, 193, 12, 118, 192, 197, 130, 240, 93, 197, 2, 213, 12, 118, 192, 197, 66, 121, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 124, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 32, 226, 12, 125, 192, 197, 2, 227, 12, 125, 192, 197, 66, 199, 12, 124, 192, 197, 66, 200, 12, 124, 192, 197, 130, 202, 12, 25, 0, 203, 12, 124, 192, 197, 2, 194, 12, 25, 0, 194, 12, 122, 192, 197, 194, 195, 12, 123, 192, 197, 2, 196, 12, 123, 192, 197, 66, 198, 12, 2, 95, 197, 130, 194, 12, 6, 0, 213, 12, 124, 192, 197, 66, 214, 12, 124, 192, 197, 130, 48, 95, 197, 194, 213, 12, 124, 192, 197, 2, 190, 12, 121, 192, 197, 194, 191, 12, 6, 0, 192, 12, 122, 192, 197, 66, 193, 12, 122, 192, 197, 130, 240, 94, 197, 2, 213, 12, 122, 192, 197, 66, 125, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 128, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 33, 226, 12, 129, 192, 197, 2, 227, 12, 129, 192, 197, 66, 199, 12, 128, 192, 197, 66, 200, 12, 128, 192, 197, 130, 202, 12, 25, 0, 203, 12, 128, 192, 197, 2, 194, 12, 25, 0, 194, 12, 126, 192, 197, 194, 195, 12, 127, 192, 197, 2, 196, 12, 127, 192, 197, 66, 198, 12, 2, 96, 197, 130, 194, 12, 6, 0, 213, 12, 128, 192, 197, 66, 214, 12, 128, 192, 197, 130, 48, 96, 197, 194, 213, 12, 128, 192, 197, 2, 190, 12, 125, 192, 197, 194, 191, 12, 6, 0, 192, 12, 126, 192, 197, 66, 193, 12, 126, 192, 197, 130, 240, 95, 197, 2, 213, 12, 126, 192, 197, 66, 129, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 132, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 34, 226, 12, 133, 192, 197, 2, 227, 12, 133, 192, 197, 66, 199, 12, 132, 192, 197, 66, 200, 12, 132, 192, 197, 130, 202, 12, 25, 0, 203, 12, 132, 192, 197, 2, 194, 12, 25, 0, 194, 12, 130, 192, 197, 194, 195, 12, 131, 192, 197, 2, 196, 12, 131, 192, 197, 66, 198, 12, 2, 97, 197, 130, 194, 12, 6, 0, 213, 12, 132, 192, 197, 66, 214, 12, 132, 192, 197, 130, 48, 97, 197, 194, 213, 12, 132, 192, 197, 2, 190, 12, 129, 192, 197, 194, 191, 12, 6, 0, 192, 12, 130, 192, 197, 66, 193, 12, 130, 192, 197, 130, 240, 96, 197, 2, 213, 12, 130, 192, 197, 66, 133, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 136, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 40, 226, 12, 137, 192, 197, 2, 227, 12, 137, 192, 197, 66, 199, 12, 136, 192, 197, 66, 200, 12, 136, 192, 197, 130, 202, 12, 25, 0, 203, 12, 136, 192, 197, 2, 194, 12, 25, 0, 194, 12, 134, 192, 197, 194, 195, 12, 135, 192, 197, 2, 196, 12, 135, 192, 197, 66, 198, 12, 2, 98, 197, 130, 194, 12, 6, 0, 213, 12, 136, 192, 197, 66, 214, 12, 136, 192, 197, 130, 48, 98, 197, 194, 213, 12, 136, 192, 197, 2, 190, 12, 133, 192, 197, 194, 191, 12, 6, 0, 192, 12, 134, 192, 197, 66, 193, 12, 134, 192, 197, 130, 240, 97, 197, 2, 213, 12, 134, 192, 197, 66, 137, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 140, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 35, 226, 12, 141, 192, 197, 2, 227, 12, 141, 192, 197, 66, 199, 12, 140, 192, 197, 66, 200, 12, 140, 192, 197, 130, 202, 12, 25, 0, 203, 12, 140, 192, 197, 2, 194, 12, 25, 0, 194, 12, 138, 192, 197, 194, 195, 12, 139, 192, 197, 2, 196, 12, 139, 192, 197, 66, 198, 12, 2, 99, 197, 130, 194, 12, 6, 0, 213, 12, 140, 192, 197, 66, 214, 12, 140, 192, 197, 130, 48, 99, 197, 194, 213, 12, 140, 192, 197, 2, 190, 12, 137, 192, 197, 194, 191, 12, 6, 0, 192, 12, 138, 192, 197, 66, 193, 12, 138, 192, 197, 130, 240, 98, 197, 2, 213, 12, 138, 192, 197, 66, 141, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 144, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 36, 226, 12, 145, 192, 197, 2, 227, 12, 145, 192, 197, 66, 199, 12, 144, 192, 197, 66, 200, 12, 144, 192, 197, 130, 202, 12, 25, 0, 203, 12, 144, 192, 197, 2, 194, 12, 25, 0, 194, 12, 142, 192, 197, 194, 195, 12, 143, 192, 197, 2, 196, 12, 143, 192, 197, 66, 198, 12, 2, 100, 197, 130, 194, 12, 6, 0, 213, 12, 144, 192, 197, 66, 214, 12, 144, 192, 197, 130, 48, 100, 197, 194, 213, 12, 144, 192, 197, 2, 190, 12, 141, 192, 197, 194, 191, 12, 6, 0, 192, 12, 142, 192, 197, 66, 193, 12, 142, 192, 197, 130, 240, 99, 197, 2, 213, 12, 142, 192, 197, 66, 145, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 148, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 37, 226, 12, 149, 192, 197, 2, 227, 12, 149, 192, 197, 66, 199, 12, 148, 192, 197, 66, 200, 12, 148, 192, 197, 130, 202, 12, 25, 0, 203, 12, 148, 192, 197, 2, 194, 12, 25, 0, 194, 12, 146, 192, 197, 194, 195, 12, 147, 192, 197, 2, 196, 12, 147, 192, 197, 66, 198, 12, 2, 101, 197, 130, 194, 12, 6, 0, 213, 12, 148, 192, 197, 66, 214, 12, 148, 192, 197, 130, 48, 101, 197, 194, 213, 12, 148, 192, 197, 2, 190, 12, 145, 192, 197, 194, 191, 12, 6, 0, 192, 12, 146, 192, 197, 66, 193, 12, 146, 192, 197, 130, 240, 100, 197, 2, 213, 12, 146, 192, 197, 66, 149, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 152, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 38, 226, 12, 153, 192, 197, 2, 227, 12, 153, 192, 197, 66, 199, 12, 152, 192, 197, 66, 200, 12, 152, 192, 197, 130, 202, 12, 25, 0, 203, 12, 152, 192, 197, 2, 194, 12, 25, 0, 194, 12, 150, 192, 197, 194, 195, 12, 151, 192, 197, 2, 196, 12, 151, 192, 197, 66, 198, 12, 2, 102, 197, 130, 194, 12, 6, 0, 213, 12, 152, 192, 197, 66, 214, 12, 152, 192, 197, 130, 48, 102, 197, 194, 213, 12, 152, 192, 197, 2, 190, 12, 149, 192, 197, 194, 191, 12, 6, 0, 192, 12, 150, 192, 197, 66, 193, 12, 150, 192, 197, 130, 240, 101, 197, 2, 213, 12, 150, 192, 197, 66, 153, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 156, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 39, 226, 12, 157, 192, 197, 2, 227, 12, 157, 192, 197, 66, 199, 12, 156, 192, 197, 66, 200, 12, 156, 192, 197, 130, 202, 12, 25, 0, 203, 12, 156, 192, 197, 2, 194, 12, 25, 0, 194, 12, 154, 192, 197, 194, 195, 12, 155, 192, 197, 2, 196, 12, 155, 192, 197, 66, 198, 12, 2, 103, 197, 130, 194, 12, 6, 0, 213, 12, 156, 192, 197, 66, 214, 12, 156, 192, 197, 130, 48, 103, 197, 194, 213, 12, 156, 192, 197, 2, 190, 12, 153, 192, 197, 194, 191, 12, 6, 0, 192, 12, 154, 192, 197, 66, 193, 12, 154, 192, 197, 130, 240, 102, 197, 2, 213, 12, 154, 192, 197, 66, 157, 0, 197, 130, 15, 0, 199, 12, 26, 0, 204, 12, 13, 0, 204, 12, 160, 192, 197, 194, 205, 12, 255, 255, 37, 111, 193, 41, 226, 12, 161, 192, 197, 2, 227, 12, 161, 192, 197, 66, 199, 12, 160, 192, 197, 66, 200, 12, 160, 192, 197, 130, 202, 12, 25, 0, 203, 12, 160, 192, 197, 2, 194, 12, 25, 0, 194, 12, 158, 192, 197, 194, 195, 12, 159, 192, 197, 2, 196, 12, 159, 192, 197, 66, 198, 12, 2, 104, 197, 130, 194, 12, 6, 0, 213, 12, 160, 192, 197, 66, 214, 12, 160, 192, 197, 130, 48, 104, 197, 194, 213, 12, 160, 192, 197, 2, 190, 12, 157, 192, 197, 194, 191, 12, 6, 0, 192, 12, 158, 192, 197, 66, 193, 12, 158, 192, 197, 130, 240, 103, 197, 2, 213, 12, 158, 192, 197, 66, 14, 0, 150, 12, 169, 12, 170, 12, 175, 12, 190, 12, 194, 12, 195, 12, 197, 12, 198, 12, 201, 12, 202, 12, 205, 12, 226, 12, 228, 12, 3, 0, 69, 0, 60, 0, 32, 0, 235, 15, 0, 16, 95, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 84, 0, 170, 170, 228, 15, 0, 32, 108, 33, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 113, 33, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 221, 26, 1, 32, 7, 0, 0, 0, 27, 0, 0, 0, 225, 26, 1, 0, 225, 26, 1, 0, 4, 0, 0, 0, 0, 0, 0, 0, 74, 27, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 101, 97, 114, 99, 104, 106, 108, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 117, 110, 105, 104, 97, 110, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 51, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 117, 0, 110, 0, 105, 0, 104, 0, 97, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 72, 0, 97, 0, 110, 0, 103, 0, 32, 0, 72, 0, 97, 0, 110, 0, 105, 0, 93, 0, 0, 0, 237, 221, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 115, 0, 117, 0, 112, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 67, 0, 111, 0, 110, 0, 116, 0, 114, 0, 97, 0, 99, 0, 116, 0, 105, 0, 111, 0, 110, 0, 115, 0, 32, 0, 91, 0, 64, 14, 45, 0, 68, 14, 32, 0, 192, 14, 45, 0, 196, 14, 32, 0, 181, 170, 32, 0, 182, 170, 32, 0, 185, 170, 32, 0, 187, 170, 32, 0, 188, 170, 181, 25, 45, 0, 183, 25, 186, 25, 93, 0, 93, 0, 38, 0, 39, 6, 60, 0, 60, 0, 60, 0, 142, 254, 60, 0, 60, 0, 60, 0, 141, 254, 60, 0, 60, 0, 34, 6, 60, 0, 60, 0, 60, 0, 130, 254, 60, 0, 60, 0, 60, 0, 129, 254, 60, 0, 60, 0, 35, 6, 60, 0, 60, 0, 60, 0, 132, 254, 60, 0, 60, 0, 60, 0, 131, 254, 60, 0, 60, 0, 37, 6, 60, 0, 60, 0, 60, 0, 136, 254, 60, 0, 60, 0, 60, 0, 135, 254, 38, 0, 72, 6, 60, 0, 60, 0, 60, 0, 229, 6, 60, 0, 60, 0, 60, 0, 238, 254, 60, 0, 60, 0, 60, 0, 237, 254, 60, 0, 60, 0, 36, 6, 60, 0, 60, 0, 60, 0, 134, 254, 60, 0, 60, 0, 60, 0, 133, 254, 38, 0, 74, 6, 60, 0, 60, 0, 60, 0, 230, 6, 60, 0, 60, 0, 60, 0, 243, 254, 60, 0, 60, 0, 60, 0, 244, 254, 60, 0, 60, 0, 60, 0, 242, 254, 60, 0, 60, 0, 60, 0, 241, 254, 60, 0, 60, 0, 38, 6, 60, 0, 60, 0, 60, 0, 139, 254, 60, 0, 60, 0, 60, 0, 140, 254, 60, 0, 60, 0, 60, 0, 138, 254, 60, 0, 60, 0, 60, 0, 137, 254, 60, 0, 60, 0, 73, 6, 60, 0, 60, 0, 60, 0, 232, 251, 60, 0, 60, 0, 60, 0, 233, 251, 60, 0, 60, 0, 60, 0, 240, 254, 60, 0, 60, 0, 60, 0, 239, 254, 38, 0, 71, 6, 60, 0, 60, 0, 60, 0, 235, 254, 60, 0, 60, 0, 60, 0, 236, 254, 60, 0, 60, 0, 60, 0, 234, 254, 60, 0, 60, 0, 60, 0, 233, 254, 60, 0, 60, 0, 41, 6, 60, 0, 60, 0, 60, 0, 148, 254, 60, 0, 60, 0, 60, 0, 147, 254, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 112, 0, 114, 0, 105, 0, 109, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 60, 0, 60, 0, 243, 5, 60, 0, 60, 0, 244, 5, 60, 0, 60, 0, 64, 6, 60, 0, 60, 0, 58, 14, 60, 0, 60, 0, 97, 17, 60, 0, 60, 0, 98, 17, 60, 0, 60, 0, 99, 17, 60, 0, 60, 0, 100, 17, 60, 0, 60, 0, 101, 17, 60, 0, 60, 0, 102, 17, 60, 0, 60, 0, 103, 17, 60, 0, 60, 0, 104, 17, 60, 0, 60, 0, 105, 17, 60, 0, 60, 0, 106, 17, 60, 0, 60, 0, 107, 17, 60, 0, 60, 0, 108, 17, 60, 0, 60, 0, 109, 17, 60, 0, 60, 0, 110, 17, 60, 0, 60, 0, 111, 17, 60, 0, 60, 0, 112, 17, 60, 0, 60, 0, 113, 17, 60, 0, 60, 0, 114, 17, 60, 0, 60, 0, 115, 17, 60, 0, 60, 0, 116, 17, 60, 0, 60, 0, 117, 17, 60, 0, 60, 0, 168, 17, 60, 0, 60, 0, 169, 17, 60, 0, 60, 0, 170, 17, 60, 0, 60, 0, 171, 17, 60, 0, 60, 0, 172, 17, 60, 0, 60, 0, 173, 17, 60, 0, 60, 0, 174, 17, 60, 0, 60, 0, 175, 17, 60, 0, 60, 0, 176, 17, 60, 0, 60, 0, 177, 17, 60, 0, 60, 0, 178, 17, 60, 0, 60, 0, 179, 17, 60, 0, 60, 0, 180, 17, 60, 0, 60, 0, 181, 17, 60, 0, 60, 0, 182, 17, 60, 0, 60, 0, 183, 17, 60, 0, 60, 0, 184, 17, 60, 0, 60, 0, 185, 17, 60, 0, 60, 0, 186, 17, 60, 0, 60, 0, 187, 17, 60, 0, 60, 0, 188, 17, 60, 0, 60, 0, 189, 17, 60, 0, 60, 0, 190, 17, 60, 0, 60, 0, 191, 17, 60, 0, 60, 0, 192, 17, 60, 0, 60, 0, 193, 17, 60, 0, 60, 0, 194, 17, 38, 0, 91, 0, 102, 0, 105, 0, 114, 0, 115, 0, 116, 0, 32, 0, 112, 0, 114, 0, 105, 0, 109, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 61, 0, 0, 17, 124, 0, 0, 17, 61, 0, 3, 17, 124, 0, 3, 17, 61, 0, 7, 17, 124, 0, 7, 17, 61, 0, 9, 17, 124, 0, 9, 17, 61, 0, 12, 17, 124, 0, 12, 17, 38, 0, 0, 17, 0, 17, 61, 0, 1, 17, 60, 0, 60, 0, 60, 0, 50, 49, 38, 0, 3, 17, 3, 17, 61, 0, 4, 17, 60, 0, 60, 0, 60, 0, 56, 49, 38, 0, 7, 17, 7, 17, 61, 0, 8, 17, 60, 0, 60, 0, 60, 0, 67, 49, 38, 0, 9, 17, 9, 17, 61, 0, 10, 17, 60, 0, 60, 0, 60, 0, 70, 49, 38, 0, 12, 17, 12, 17, 61, 0, 13, 17, 60, 0, 60, 0, 60, 0, 73, 49, 0, 0, 14, 223, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 38, 0, 0, 17, 2, 17, 61, 0, 250, 17, 38, 0, 0, 17, 3, 17, 61, 0, 90, 17, 38, 0, 0, 17, 5, 17, 61, 0, 195, 17, 38, 0, 0, 17, 7, 17, 61, 0, 251, 17, 38, 0, 0, 17, 9, 17, 0, 17, 61, 0, 196, 17, 38, 0, 0, 17, 14, 17, 61, 0, 252, 17, 38, 0, 0, 17, 15, 17, 61, 0, 253, 17, 38, 0, 0, 17, 18, 17, 61, 0, 254, 17, 38, 0, 2, 17, 0, 17, 61, 0, 19, 17, 61, 0, 197, 17, 38, 0, 2, 17, 2, 17, 61, 0, 20, 17, 61, 0, 255, 17, 38, 0, 2, 17, 3, 17, 61, 0, 21, 17, 61, 0, 198, 17, 38, 0, 2, 17, 7, 17, 61, 0, 22, 17, 38, 0, 2, 17, 9, 17, 61, 0, 91, 17, 61, 0, 199, 17, 38, 0, 2, 17, 12, 17, 61, 0, 92, 17, 38, 0, 2, 17, 16, 17, 61, 0, 201, 17, 38, 0, 2, 17, 18, 17, 61, 0, 93, 17, 38, 0, 3, 17, 0, 17, 61, 0, 23, 17, 61, 0, 202, 17, 38, 0, 3, 17, 5, 17, 61, 0, 94, 17, 61, 0, 203, 17, 38, 0, 5, 17, 0, 17, 9, 17, 61, 0, 204, 17, 38, 0, 5, 17, 2, 17, 61, 0, 24, 17, 61, 0, 205, 17, 38, 0, 5, 17, 3, 17, 61, 0, 206, 17, 38, 0, 5, 17, 3, 17, 18, 17, 61, 0, 207, 17, 38, 0, 5, 17, 5, 17, 61, 0, 25, 17, 61, 0, 208, 17, 38, 0, 5, 17, 6, 17, 0, 17, 61, 0, 209, 17, 38, 0, 5, 17, 6, 17, 9, 17, 61, 0, 210, 17, 38, 0, 5, 17, 7, 17, 9, 17, 61, 0, 211, 17, 38, 0, 5, 17, 7, 17, 11, 17, 61, 0, 213, 17, 38, 0, 5, 17, 7, 17, 18, 17, 61, 0, 212, 17, 38, 0, 5, 17, 9, 17, 9, 17, 61, 0, 214, 17, 38, 0, 5, 17, 11, 17, 61, 0, 27, 17, 38, 0, 5, 17, 15, 17, 61, 0, 216, 17, 38, 0, 5, 17, 18, 17, 61, 0, 26, 17, 38, 0, 6, 17, 0, 17, 61, 0, 218, 17, 38, 0, 6, 17, 5, 17, 61, 0, 219, 17, 38, 0, 6, 17, 7, 17, 61, 0, 28, 17, 61, 0, 220, 17, 38, 0, 6, 17, 9, 17, 61, 0, 221, 17, 38, 0, 6, 17, 9, 17, 9, 17, 61, 0, 222, 17, 38, 0, 6, 17, 11, 17, 61, 0, 29, 17, 61, 0, 226, 17, 38, 0, 6, 17, 14, 17, 61, 0, 224, 17, 38, 0, 6, 17, 18, 17, 61, 0, 225, 17, 38, 0, 7, 17, 0, 17, 61, 0, 30, 17, 38, 0, 7, 17, 2, 17, 61, 0, 31, 17, 38, 0, 7, 17, 3, 17, 61, 0, 32, 17, 38, 0, 7, 17, 5, 17, 61, 0, 227, 17, 38, 0, 7, 17, 7, 17, 11, 17, 61, 0, 44, 17, 38, 0, 7, 17, 9, 17, 61, 0, 33, 17, 38, 0, 7, 17, 9, 17, 0, 17, 61, 0, 34, 17, 38, 0, 7, 17, 9, 17, 3, 17, 61, 0, 35, 17, 38, 0, 7, 17, 9, 17, 7, 17, 61, 0, 36, 17, 38, 0, 7, 17, 9, 17, 9, 17, 61, 0, 37, 17, 38, 0, 7, 17, 9, 17, 12, 17, 61, 0, 38, 17, 38, 0, 7, 17, 11, 17, 61, 0, 43, 17, 61, 0, 230, 17, 38, 0, 7, 17, 12, 17, 61, 0, 39, 17, 38, 0, 7, 17, 14, 17, 61, 0, 40, 17, 38, 0, 7, 17, 16, 17, 61, 0, 41, 17, 38, 0, 7, 17, 17, 17, 61, 0, 42, 17, 61, 0, 228, 17, 38, 0, 7, 17, 18, 17, 61, 0, 229, 17, 38, 0, 9, 17, 0, 17, 61, 0, 45, 17, 61, 0, 231, 17, 38, 0, 9, 17, 2, 17, 61, 0, 46, 17, 38, 0, 9, 17, 3, 17, 61, 0, 47, 17, 61, 0, 232, 17, 38, 0, 9, 17, 5, 17, 61, 0, 48, 17, 61, 0, 233, 17, 38, 0, 9, 17, 6, 17, 61, 0, 49, 17, 38, 0, 9, 17, 7, 17, 61, 0, 50, 17, 61, 0, 234, 17, 38, 0, 9, 17, 7, 17, 0, 17, 61, 0, 51, 17, 38, 0, 9, 17, 9, 17, 9, 17, 61, 0, 52, 17, 38, 0, 9, 17, 11, 17, 61, 0, 53, 17, 38, 0, 9, 17, 12, 17, 61, 0, 54, 17, 38, 0, 9, 17, 14, 17, 61, 0, 55, 17, 38, 0, 9, 17, 15, 17, 61, 0, 56, 17, 38, 0, 9, 17, 16, 17, 61, 0, 57, 17, 38, 0, 9, 17, 17, 17, 61, 0, 58, 17, 38, 0, 9, 17, 18, 17, 61, 0, 59, 17, 38, 0, 11, 17, 0, 17, 61, 0, 65, 17, 61, 0, 236, 17, 38, 0, 11, 17, 0, 17, 0, 17, 61, 0, 237, 17, 38, 0, 11, 17, 3, 17, 61, 0, 66, 17, 38, 0, 11, 17, 6, 17, 61, 0, 67, 17, 38, 0, 11, 17, 7, 17, 61, 0, 68, 17, 38, 0, 11, 17, 9, 17, 61, 0, 69, 17, 38, 0, 11, 17, 11, 17, 61, 0, 71, 17, 61, 0, 238, 17, 38, 0, 11, 17, 12, 17, 61, 0, 72, 17, 38, 0, 11, 17, 14, 17, 61, 0, 73, 17, 38, 0, 11, 17, 15, 17, 61, 0, 239, 17, 38, 0, 11, 17, 16, 17, 61, 0, 74, 17, 38, 0, 11, 17, 17, 17, 61, 0, 75, 17, 38, 0, 12, 17, 11, 17, 61, 0, 77, 17, 38, 0, 14, 17, 15, 17, 61, 0, 82, 17, 38, 0, 14, 17, 18, 17, 61, 0, 83, 17, 38, 0, 17, 17, 7, 17, 61, 0, 86, 17, 61, 0, 243, 17, 38, 0, 17, 17, 11, 17, 61, 0, 87, 17, 61, 0, 244, 17, 38, 0, 18, 17, 2, 17, 61, 0, 245, 17, 38, 0, 18, 17, 5, 17, 61, 0, 246, 17, 38, 0, 18, 17, 6, 17, 61, 0, 247, 17, 38, 0, 18, 17, 7, 17, 61, 0, 248, 17, 38, 0, 18, 17, 18, 17, 61, 0, 88, 17, 38, 0, 97, 17, 105, 17, 61, 0, 118, 17, 38, 0, 97, 17, 110, 17, 61, 0, 119, 17, 38, 0, 99, 17, 105, 17, 61, 0, 120, 17, 38, 0, 99, 17, 109, 17, 61, 0, 121, 17, 38, 0, 101, 17, 105, 17, 61, 0, 122, 17, 38, 0, 101, 17, 110, 17, 61, 0, 123, 17, 38, 0, 101, 17, 115, 17, 61, 0, 124, 17, 38, 0, 103, 17, 105, 17, 61, 0, 125, 17, 38, 0, 103, 17, 110, 17, 61, 0, 126, 17, 38, 0, 105, 17, 101, 17, 61, 0, 127, 17, 38, 0, 105, 17, 101, 17, 117, 17, 61, 0, 128, 17, 38, 0, 105, 17, 103, 17, 117, 17, 61, 0, 129, 17, 38, 0, 105, 17, 105, 17, 61, 0, 130, 17, 38, 0, 105, 17, 110, 17, 61, 0, 131, 17, 38, 0, 109, 17, 99, 17, 61, 0, 132, 17, 38, 0, 109, 17, 99, 17, 117, 17, 61, 0, 133, 17, 38, 0, 109, 17, 103, 17, 61, 0, 134, 17, 38, 0, 109, 17, 105, 17, 61, 0, 135, 17, 38, 0, 109, 17, 117, 17, 61, 0, 136, 17, 38, 0, 110, 17, 97, 17, 61, 0, 137, 17, 38, 0, 110, 17, 97, 17, 117, 17, 61, 0, 138, 17, 38, 0, 110, 17, 101, 17, 115, 17, 61, 0, 139, 17, 38, 0, 110, 17, 103, 17, 117, 17, 61, 0, 140, 17, 38, 0, 110, 17, 110, 17, 61, 0, 141, 17, 38, 0, 114, 17, 97, 17, 61, 0, 142, 17, 38, 0, 114, 17, 101, 17, 61, 0, 143, 17, 38, 0, 114, 17, 101, 17, 117, 17, 61, 0, 144, 17, 38, 0, 114, 17, 103, 17, 61, 0, 145, 17, 38, 0, 114, 17, 103, 17, 117, 17, 61, 0, 146, 17, 38, 0, 114, 17, 110, 17, 61, 0, 147, 17, 38, 0, 114, 17, 117, 17, 61, 0, 148, 17, 38, 0, 115, 17, 110, 17, 61, 0, 149, 17, 38, 0, 115, 17, 115, 17, 61, 0, 150, 17, 38, 0, 116, 17, 110, 17, 61, 0, 151, 17, 38, 0, 117, 17, 97, 17, 61, 0, 152, 17, 38, 0, 117, 17, 99, 17, 61, 0, 153, 17, 38, 0, 117, 17, 105, 17, 61, 0, 154, 17, 38, 0, 117, 17, 110, 17, 61, 0, 155, 17, 38, 0, 117, 17, 115, 17, 61, 0, 156, 17, 38, 0, 97, 17, 115, 17, 61, 0, 163, 17, 38, 0, 99, 17, 110, 17, 61, 0, 164, 17, 38, 0, 103, 17, 99, 17, 61, 0, 165, 17, 38, 0, 105, 17, 99, 17, 61, 0, 166, 17, 38, 0, 105, 17, 99, 17, 117, 17, 61, 0, 167, 17, 0, 0, 239, 223, 33, 49, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 72, 0, 97, 0, 110, 0, 103, 0, 32, 0, 72, 0, 97, 0, 110, 0, 105, 0, 93, 0, 91, 0, 111, 0, 112, 0, 116, 0, 105, 0, 109, 0, 105, 0, 122, 0, 101, 0, 32, 0, 91, 0, 0, 172, 45, 0, 1, 172, 4, 172, 7, 172, 45, 0, 10, 172, 16, 172, 45, 0, 23, 172, 25, 172, 45, 0, 29, 172, 32, 172, 36, 172, 44, 172, 45, 0, 45, 172, 47, 172, 45, 0, 49, 172, 56, 172, 45, 0, 57, 172, 60, 172, 64, 172, 75, 172, 77, 172, 84, 172, 88, 172, 92, 172, 112, 172, 45, 0, 113, 172, 116, 172, 119, 172, 45, 0, 120, 172, 122, 172, 128, 172, 45, 0, 129, 172, 131, 172, 45, 0, 134, 172, 137, 172, 45, 0, 140, 172, 144, 172, 148, 172, 156, 172, 45, 0, 157, 172, 159, 172, 45, 0, 161, 172, 168, 172, 45, 0, 170, 172, 172, 172, 175, 172, 45, 0, 176, 172, 184, 172, 45, 0, 185, 172, 187, 172, 45, 0, 189, 172, 193, 172, 196, 172, 200, 172, 204, 172, 213, 172, 215, 172, 224, 172, 45, 0, 225, 172, 228, 172, 231, 172, 45, 0, 232, 172, 234, 172, 236, 172, 239, 172, 45, 0, 241, 172, 243, 172, 245, 172, 45, 0, 246, 172, 252, 172, 45, 0, 253, 172, 0, 173, 4, 173, 6, 173, 12, 173, 45, 0, 13, 173, 15, 173, 17, 173, 24, 173, 28, 173, 32, 173, 41, 173, 44, 173, 45, 0, 45, 173, 52, 173, 45, 0, 53, 173, 56, 173, 60, 173, 68, 173, 45, 0, 69, 173, 71, 173, 73, 173, 80, 173, 84, 173, 88, 173, 97, 173, 99, 173, 108, 173, 45, 0, 109, 173, 112, 173, 115, 173, 45, 0, 118, 173, 123, 173, 45, 0, 125, 173, 127, 173, 129, 173, 45, 0, 130, 173, 136, 173, 45, 0, 137, 173, 140, 173, 144, 173, 156, 173, 45, 0, 157, 173, 164, 173, 183, 173, 192, 173, 45, 0, 193, 173, 196, 173, 200, 173, 208, 173, 45, 0, 209, 173, 211, 173, 220, 173, 224, 173, 228, 173, 248, 173, 45, 0, 249, 173, 252, 173, 255, 173, 45, 0, 1, 174, 8, 174, 45, 0, 9, 174, 11, 174, 13, 174, 20, 174, 48, 174, 45, 0, 49, 174, 52, 174, 55, 174, 45, 0, 56, 174, 58, 174, 64, 174, 45, 0, 65, 174, 67, 174, 69, 174, 45, 0, 70, 174, 74, 174, 76, 174, 45, 0, 78, 174, 80, 174, 84, 174, 86, 174, 92, 174, 45, 0, 93, 174, 95, 174, 45, 0, 97, 174, 101, 174, 104, 174, 45, 0, 105, 174, 108, 174, 112, 174, 120, 174, 45, 0, 121, 174, 123, 174, 45, 0, 125, 174, 132, 174, 45, 0, 133, 174, 140, 174, 188, 174, 45, 0, 190, 174, 192, 174, 196, 174, 204, 174, 45, 0, 205, 174, 207, 174, 45, 0, 209, 174, 216, 174, 45, 0, 217, 174, 220, 174, 232, 174, 235, 174, 237, 174, 244, 174, 248, 174, 252, 174, 7, 175, 45, 0, 8, 175, 13, 175, 16, 175, 44, 175, 45, 0, 45, 175, 48, 175, 50, 175, 52, 175, 60, 175, 45, 0, 61, 175, 63, 175, 65, 175, 45, 0, 67, 175, 72, 175, 45, 0, 73, 175, 80, 175, 92, 175, 45, 0, 93, 175, 100, 175, 45, 0, 101, 175, 121, 175, 128, 175, 132, 175, 136, 175, 144, 175, 45, 0, 145, 175, 149, 175, 156, 175, 184, 175, 45, 0, 185, 175, 188, 175, 192, 175, 199, 175, 45, 0, 201, 175, 203, 175, 205, 175, 45, 0, 206, 175, 212, 175, 220, 175, 232, 175, 45, 0, 233, 175, 240, 175, 45, 0, 241, 175, 244, 175, 248, 175, 0, 176, 45, 0, 1, 176, 4, 176, 12, 176, 16, 176, 20, 176, 28, 176, 45, 0, 29, 176, 40, 176, 68, 176, 45, 0, 69, 176, 72, 176, 74, 176, 76, 176, 78, 176, 83, 176, 45, 0, 85, 176, 87, 176, 89, 176, 93, 176, 124, 176, 45, 0, 125, 176, 128, 176, 132, 176, 140, 176, 45, 0, 141, 176, 143, 176, 145, 176, 152, 176, 45, 0, 154, 176, 156, 176, 159, 176, 45, 0, 162, 176, 168, 176, 45, 0, 169, 176, 171, 176, 45, 0, 175, 176, 177, 176, 179, 176, 45, 0, 181, 176, 184, 176, 188, 176, 196, 176, 45, 0, 197, 176, 199, 176, 45, 0, 201, 176, 208, 176, 45, 0, 209, 176, 212, 176, 216, 176, 224, 176, 229, 176, 8, 177, 45, 0, 9, 177, 11, 177, 45, 0, 12, 177, 16, 177, 18, 177, 45, 0, 19, 177, 24, 177, 45, 0, 25, 177, 27, 177, 45, 0, 29, 177, 35, 177, 45, 0, 37, 177, 40, 177, 44, 177, 52, 177, 45, 0, 53, 177, 55, 177, 45, 0, 57, 177, 64, 177, 45, 0, 65, 177, 68, 177, 72, 177, 80, 177, 45, 0, 81, 177, 84, 177, 45, 0, 85, 177, 88, 177, 92, 177, 96, 177, 120, 177, 45, 0, 121, 177, 124, 177, 128, 177, 130, 177, 136, 177, 45, 0, 137, 177, 139, 177, 141, 177, 146, 177, 45, 0, 148, 177, 152, 177, 156, 177, 168, 177, 204, 177, 208, 177, 212, 177, 220, 177, 45, 0, 221, 177, 223, 177, 232, 177, 45, 0, 233, 177, 236, 177, 240, 177, 249, 177, 251, 177, 253, 177, 4, 178, 45, 0, 5, 178, 8, 178, 11, 178, 45, 0, 12, 178, 20, 178, 45, 0, 21, 178, 23, 178, 25, 178, 32, 178, 52, 178, 60, 178, 88, 178, 92, 178, 96, 178, 104, 178, 45, 0, 105, 178, 116, 178, 45, 0, 117, 178, 124, 178, 132, 178, 45, 0, 133, 178, 137, 178, 144, 178, 45, 0, 145, 178, 148, 178, 152, 178, 45, 0, 154, 178, 160, 178, 45, 0, 161, 178, 163, 178, 165, 178, 45, 0, 166, 178, 170, 178, 172, 178, 176, 178, 180, 178, 200, 178, 45, 0, 201, 178, 204, 178, 208, 178, 214, 178, 216, 178, 45, 0, 217, 178, 219, 178, 221, 178, 226, 178, 228, 178, 45, 0, 230, 178, 232, 178, 235, 178, 45, 0, 239, 178, 243, 178, 45, 0, 245, 178, 247, 178, 45, 0, 251, 178, 255, 178, 45, 0, 1, 179, 4, 179, 8, 179, 16, 179, 45, 0, 17, 179, 19, 179, 45, 0, 21, 179, 28, 179, 84, 179, 45, 0, 86, 179, 88, 179, 91, 179, 45, 0, 92, 179, 94, 179, 45, 0, 95, 179, 100, 179, 45, 0, 101, 179, 103, 179, 105, 179, 107, 179, 110, 179, 112, 179, 45, 0, 113, 179, 116, 179, 120, 179, 128, 179, 45, 0, 129, 179, 131, 179, 45, 0, 133, 179, 140, 179, 144, 179, 148, 179, 160, 179, 45, 0, 161, 179, 168, 179, 172, 179, 196, 179, 45, 0, 197, 179, 200, 179, 203, 179, 45, 0, 204, 179, 206, 179, 208, 179, 212, 179, 45, 0, 213, 179, 215, 179, 217, 179, 219, 179, 221, 179, 224, 179, 228, 179, 232, 179, 252, 179, 16, 180, 24, 180, 28, 180, 32, 180, 40, 180, 45, 0, 41, 180, 43, 180, 52, 180, 80, 180, 45, 0, 81, 180, 84, 180, 88, 180, 96, 180, 45, 0, 97, 180, 99, 180, 101, 180, 108, 180, 128, 180, 136, 180, 157, 180, 164, 180, 168, 180, 172, 180, 181, 180, 183, 180, 185, 180, 192, 180, 196, 180, 200, 180, 208, 180, 213, 180, 220, 180, 45, 0, 221, 180, 224, 180, 227, 180, 45, 0, 228, 180, 230, 180, 236, 180, 45, 0, 237, 180, 239, 180, 241, 180, 248, 180, 20, 181, 45, 0, 21, 181, 24, 181, 27, 181, 45, 0, 28, 181, 36, 181, 45, 0, 37, 181, 39, 181, 45, 0, 42, 181, 48, 181, 45, 0, 49, 181, 52, 181, 56, 181, 64, 181, 45, 0, 65, 181, 67, 181, 45, 0, 69, 181, 75, 181, 45, 0, 77, 181, 80, 181, 84, 181, 92, 181, 45, 0, 93, 181, 95, 181, 45, 0, 97, 181, 160, 181, 45, 0, 161, 181, 164, 181, 168, 181, 170, 181, 45, 0, 171, 181, 176, 181, 45, 0, 177, 181, 179, 181, 45, 0, 181, 181, 187, 181, 45, 0, 189, 181, 192, 181, 196, 181, 204, 181, 45, 0, 205, 181, 207, 181, 45, 0, 209, 181, 216, 181, 236, 181, 16, 182, 45, 0, 17, 182, 20, 182, 24, 182, 37, 182, 44, 182, 52, 182, 72, 182, 100, 182, 104, 182, 156, 182, 45, 0, 157, 182, 160, 182, 164, 182, 171, 182, 45, 0, 172, 182, 177, 182, 212, 182, 240, 182, 244, 182, 248, 182, 0, 183, 45, 0, 1, 183, 5, 183, 40, 183, 45, 0, 41, 183, 44, 183, 47, 183, 45, 0, 48, 183, 56, 183, 45, 0, 57, 183, 59, 183, 68, 183, 72, 183, 76, 183, 84, 183, 45, 0, 85, 183, 96, 183, 100, 183, 104, 183, 112, 183, 45, 0, 113, 183, 115, 183, 117, 183, 124, 183, 45, 0, 125, 183, 128, 183, 132, 183, 140, 183, 45, 0, 141, 183, 143, 183, 45, 0, 146, 183, 150, 183, 45, 0, 153, 183, 156, 183, 160, 183, 168, 183, 45, 0, 169, 183, 171, 183, 45, 0, 173, 183, 180, 183, 45, 0, 181, 183, 184, 183, 199, 183, 201, 183, 236, 183, 45, 0, 237, 183, 240, 183, 244, 183, 252, 183, 45, 0, 253, 183, 255, 183, 45, 0, 1, 184, 7, 184, 45, 0, 9, 184, 12, 184, 16, 184, 24, 184, 45, 0, 25, 184, 27, 184, 29, 184, 36, 184, 45, 0, 37, 184, 40, 184, 44, 184, 52, 184, 45, 0, 53, 184, 55, 184, 45, 0, 57, 184, 64, 184, 68, 184, 81, 184, 83, 184, 92, 184, 45, 0, 93, 184, 96, 184, 100, 184, 108, 184, 45, 0, 109, 184, 111, 184, 113, 184, 120, 184, 124, 184, 141, 184, 168, 184, 176, 184, 180, 184, 184, 184, 192, 184, 45, 0, 193, 184, 195, 184, 197, 184, 204, 184, 208, 184, 212, 184, 221, 184, 223, 184, 225, 184, 232, 184, 45, 0, 233, 184, 236, 184, 240, 184, 248, 184, 45, 0, 249, 184, 251, 184, 253, 184, 4, 185, 24, 185, 32, 185, 60, 185, 45, 0, 61, 185, 64, 185, 68, 185, 76, 185, 79, 185, 81, 185, 88, 185, 45, 0, 89, 185, 92, 185, 96, 185, 104, 185, 45, 0, 105, 185, 107, 185, 109, 185, 116, 185, 45, 0, 117, 185, 120, 185, 124, 185, 132, 185, 45, 0, 133, 185, 135, 185, 137, 185, 45, 0, 138, 185, 141, 185, 45, 0, 142, 185, 172, 185, 45, 0, 173, 185, 176, 185, 180, 185, 188, 185, 45, 0, 189, 185, 191, 185, 193, 185, 200, 185, 45, 0, 201, 185, 204, 185, 206, 185, 45, 0, 210, 185, 216, 185, 45, 0, 217, 185, 219, 185, 221, 185, 45, 0, 222, 185, 225, 185, 227, 185, 45, 0, 229, 185, 232, 185, 236, 185, 244, 185, 45, 0, 245, 185, 247, 185, 45, 0, 250, 185, 0, 186, 45, 0, 1, 186, 8, 186, 21, 186, 56, 186, 45, 0, 57, 186, 60, 186, 64, 186, 66, 186, 72, 186, 45, 0, 73, 186, 75, 186, 77, 186, 45, 0, 78, 186, 83, 186, 45, 0, 85, 186, 88, 186, 92, 186, 100, 186, 45, 0, 101, 186, 103, 186, 45, 0, 105, 186, 112, 186, 45, 0, 113, 186, 116, 186, 120, 186, 131, 186, 45, 0, 133, 186, 135, 186, 140, 186, 168, 186, 45, 0, 169, 186, 171, 186, 45, 0, 172, 186, 176, 186, 178, 186, 184, 186, 45, 0, 185, 186, 187, 186, 189, 186, 196, 186, 200, 186, 216, 186, 45, 0, 217, 186, 252, 186, 0, 187, 4, 187, 13, 187, 15, 187, 17, 187, 24, 187, 28, 187, 32, 187, 41, 187, 43, 187, 52, 187, 45, 0, 54, 187, 56, 187, 59, 187, 45, 0, 62, 187, 68, 187, 45, 0, 69, 187, 71, 187, 73, 187, 77, 187, 79, 187, 45, 0, 80, 187, 84, 187, 88, 187, 97, 187, 99, 187, 108, 187, 136, 187, 140, 187, 144, 187, 164, 187, 168, 187, 172, 187, 180, 187, 183, 187, 192, 187, 196, 187, 200, 187, 208, 187, 211, 187, 248, 187, 45, 0, 249, 187, 252, 187, 255, 187, 45, 0, 0, 188, 2, 188, 8, 188, 45, 0, 9, 188, 11, 188, 45, 0, 13, 188, 15, 188, 17, 188, 20, 188, 45, 0, 24, 188, 27, 188, 45, 0, 31, 188, 36, 188, 45, 0, 37, 188, 39, 188, 41, 188, 45, 188, 48, 188, 45, 0, 49, 188, 52, 188, 56, 188, 64, 188, 45, 0, 65, 188, 67, 188, 45, 0, 69, 188, 73, 188, 76, 188, 45, 0, 77, 188, 80, 188, 93, 188, 132, 188, 45, 0, 133, 188, 136, 188, 139, 188, 45, 0, 140, 188, 142, 188, 148, 188, 45, 0, 149, 188, 151, 188, 153, 188, 45, 0, 154, 188, 160, 188, 45, 0, 161, 188, 164, 188, 167, 188, 45, 0, 168, 188, 176, 188, 45, 0, 177, 188, 179, 188, 45, 0, 181, 188, 188, 188, 45, 0, 189, 188, 192, 188, 196, 188, 205, 188, 207, 188, 45, 0, 209, 188, 213, 188, 216, 188, 220, 188, 244, 188, 45, 0, 246, 188, 248, 188, 252, 188, 4, 189, 45, 0, 5, 189, 7, 189, 9, 189, 16, 189, 20, 189, 36, 189, 44, 189, 64, 189, 72, 189, 45, 0, 73, 189, 76, 189, 80, 189, 88, 189, 45, 0, 89, 189, 100, 189, 104, 189, 128, 189, 45, 0, 129, 189, 132, 189, 135, 189, 45, 0, 138, 189, 144, 189, 45, 0, 145, 189, 147, 189, 149, 189, 153, 189, 45, 0, 154, 189, 156, 189, 164, 189, 176, 189, 184, 189, 212, 189, 45, 0, 213, 189, 216, 189, 220, 189, 233, 189, 240, 189, 244, 189, 248, 189, 0, 190, 3, 190, 5, 190, 12, 190, 45, 0, 13, 190, 16, 190, 20, 190, 28, 190, 45, 0, 29, 190, 31, 190, 68, 190, 45, 0, 69, 190, 72, 190, 76, 190, 78, 190, 84, 190, 45, 0, 85, 190, 87, 190, 89, 190, 45, 0, 91, 190, 96, 190, 45, 0, 97, 190, 100, 190, 104, 190, 106, 190, 112, 190, 45, 0, 113, 190, 115, 190, 45, 0, 117, 190, 123, 190, 45, 0, 125, 190, 128, 190, 132, 190, 140, 190, 45, 0, 141, 190, 143, 190, 45, 0, 145, 190, 152, 190, 45, 0, 153, 190, 168, 190, 208, 190, 45, 0, 209, 190, 212, 190, 215, 190, 45, 0, 216, 190, 224, 190, 227, 190, 45, 0, 229, 190, 236, 190, 1, 191, 8, 191, 45, 0, 9, 191, 24, 191, 45, 0, 25, 191, 27, 191, 45, 0, 29, 191, 64, 191, 45, 0, 65, 191, 68, 191, 72, 191, 80, 191, 45, 0, 81, 191, 85, 191, 148, 191, 176, 191, 197, 191, 204, 191, 45, 0, 205, 191, 208, 191, 212, 191, 220, 191, 223, 191, 225, 191, 60, 192, 81, 192, 88, 192, 92, 192, 96, 192, 104, 192, 45, 0, 105, 192, 144, 192, 45, 0, 145, 192, 148, 192, 152, 192, 160, 192, 45, 0, 161, 192, 163, 192, 165, 192, 172, 192, 45, 0, 173, 192, 175, 192, 45, 0, 176, 192, 179, 192, 45, 0, 182, 192, 188, 192, 45, 0, 189, 192, 191, 192, 45, 0, 193, 192, 197, 192, 200, 192, 45, 0, 201, 192, 204, 192, 208, 192, 216, 192, 45, 0, 217, 192, 219, 192, 45, 0, 221, 192, 228, 192, 45, 0, 229, 192, 232, 192, 236, 192, 244, 192, 45, 0, 245, 192, 247, 192, 249, 192, 0, 193, 4, 193, 8, 193, 16, 193, 21, 193, 28, 193, 45, 0, 32, 193, 35, 193, 45, 0, 36, 193, 38, 193, 45, 0, 39, 193, 44, 193, 45, 0, 45, 193, 47, 193, 45, 0, 49, 193, 54, 193, 56, 193, 45, 0, 57, 193, 60, 193, 64, 193, 72, 193, 45, 0, 73, 193, 75, 193, 45, 0, 77, 193, 84, 193, 45, 0, 85, 193, 88, 193, 92, 193, 100, 193, 45, 0, 101, 193, 103, 193, 45, 0, 105, 193, 112, 193, 116, 193, 120, 193, 133, 193, 140, 193, 45, 0, 142, 193, 144, 193, 148, 193, 150, 193, 156, 193, 45, 0, 157, 193, 159, 193, 161, 193, 165, 193, 168, 193, 45, 0, 169, 193, 172, 193, 176, 193, 189, 193, 196, 193, 200, 193, 204, 193, 212, 193, 215, 193, 45, 0, 216, 193, 224, 193, 228, 193, 232, 193, 240, 193, 45, 0, 241, 193, 243, 193, 252, 193, 45, 0, 253, 193, 0, 194, 4, 194, 12, 194, 45, 0, 13, 194, 15, 194, 17, 194, 24, 194, 45, 0, 25, 194, 28, 194, 31, 194, 45, 0, 32, 194, 40, 194, 45, 0, 41, 194, 43, 194, 45, 194, 47, 194, 49, 194, 45, 0, 50, 194, 52, 194, 72, 194, 80, 194, 45, 0, 81, 194, 84, 194, 88, 194, 96, 194, 101, 194, 108, 194, 45, 0, 109, 194, 112, 194, 116, 194, 124, 194, 45, 0, 125, 194, 127, 194, 129, 194, 136, 194, 45, 0, 137, 194, 144, 194, 152, 194, 155, 194, 157, 194, 164, 194, 45, 0, 165, 194, 168, 194, 172, 194, 45, 0, 173, 194, 180, 194, 45, 0, 181, 194, 183, 194, 185, 194, 220, 194, 45, 0, 221, 194, 224, 194, 227, 194, 45, 0, 228, 194, 235, 194, 45, 0, 237, 194, 239, 194, 241, 194, 246, 194, 248, 194, 45, 0, 249, 194, 251, 194, 45, 0, 252, 194, 0, 195, 8, 195, 45, 0, 9, 195, 12, 195, 45, 0, 13, 195, 19, 195, 45, 0, 21, 195, 24, 195, 28, 195, 36, 195, 45, 0, 37, 195, 40, 195, 45, 0, 41, 195, 69, 195, 104, 195, 45, 0, 105, 195, 108, 195, 112, 195, 114, 195, 120, 195, 45, 0, 121, 195, 124, 195, 45, 0, 125, 195, 132, 195, 136, 195, 140, 195, 192, 195, 216, 195, 45, 0, 217, 195, 220, 195, 223, 195, 45, 0, 224, 195, 226, 195, 232, 195, 45, 0, 233, 195, 237, 195, 244, 195, 45, 0, 245, 195, 248, 195, 8, 196, 16, 196, 36, 196, 44, 196, 48, 196, 52, 196, 60, 196, 45, 0, 61, 196, 72, 196, 100, 196, 45, 0, 101, 196, 104, 196, 108, 196, 116, 196, 45, 0, 117, 196, 121, 196, 128, 196, 148, 196, 156, 196, 184, 196, 188, 196, 233, 196, 240, 196, 45, 0, 241, 196, 244, 196, 248, 196, 250, 196, 255, 196, 45, 0, 1, 197, 12, 197, 16, 197, 20, 197, 28, 197, 40, 197, 45, 0, 41, 197, 44, 197, 48, 197, 56, 197, 45, 0, 57, 197, 59, 197, 61, 197, 68, 197, 45, 0, 69, 197, 72, 197, 45, 0, 74, 197, 76, 197, 45, 0, 78, 197, 83, 197, 45, 0, 85, 197, 87, 197, 45, 0, 89, 197, 93, 197, 45, 0, 94, 197, 96, 197, 45, 0, 97, 197, 100, 197, 104, 197, 112, 197, 45, 0, 113, 197, 115, 197, 45, 0, 117, 197, 124, 197, 45, 0, 125, 197, 128, 197, 132, 197, 135, 197, 140, 197, 45, 0, 141, 197, 143, 197, 145, 197, 149, 197, 151, 197, 45, 0, 152, 197, 156, 197, 160, 197, 169, 197, 180, 197, 45, 0, 181, 197, 184, 197, 45, 0, 185, 197, 187, 197, 45, 0, 190, 197, 196, 197, 45, 0, 202, 197, 204, 197, 206, 197, 208, 197, 45, 0, 209, 197, 212, 197, 216, 197, 224, 197, 45, 0, 225, 197, 227, 197, 229, 197, 236, 197, 45, 0, 238, 197, 240, 197, 244, 197, 246, 197, 45, 0, 247, 197, 252, 197, 45, 0, 1, 198, 5, 198, 45, 0, 8, 198, 12, 198, 16, 198, 24, 198, 45, 0, 25, 198, 27, 198, 45, 0, 28, 198, 36, 198, 45, 0, 37, 198, 40, 198, 44, 198, 45, 0, 46, 198, 48, 198, 51, 198, 45, 0, 53, 198, 55, 198, 57, 198, 59, 198, 64, 198, 45, 0, 65, 198, 68, 198, 72, 198, 80, 198, 45, 0, 81, 198, 83, 198, 45, 0, 85, 198, 92, 198, 45, 0, 93, 198, 96, 198, 108, 198, 111, 198, 113, 198, 120, 198, 45, 0, 121, 198, 124, 198, 128, 198, 136, 198, 45, 0, 137, 198, 139, 198, 141, 198, 148, 198, 45, 0, 149, 198, 152, 198, 156, 198, 164, 198, 45, 0, 165, 198, 167, 198, 169, 198, 176, 198, 45, 0, 177, 198, 180, 198, 184, 198, 45, 0, 186, 198, 192, 198, 45, 0, 193, 198, 195, 198, 197, 198, 204, 198, 45, 0, 205, 198, 208, 198, 212, 198, 220, 198, 45, 0, 221, 198, 224, 198, 45, 0, 225, 198, 232, 198, 45, 0, 233, 198, 236, 198, 240, 198, 248, 198, 45, 0, 249, 198, 253, 198, 4, 199, 45, 0, 5, 199, 8, 199, 12, 199, 20, 199, 45, 0, 21, 199, 23, 199, 25, 199, 32, 199, 45, 0, 33, 199, 36, 199, 40, 199, 48, 199, 45, 0, 49, 199, 51, 199, 53, 199, 55, 199, 60, 199, 45, 0, 61, 199, 64, 199, 68, 199, 74, 199, 76, 199, 45, 0, 77, 199, 79, 199, 81, 199, 45, 0, 88, 199, 92, 199, 96, 199, 104, 199, 107, 199, 116, 199, 45, 0, 117, 199, 120, 199, 124, 199, 45, 0, 126, 199, 131, 199, 45, 0, 133, 199, 135, 199, 45, 0, 138, 199, 142, 199, 144, 199, 45, 0, 145, 199, 148, 199, 150, 199, 45, 0, 152, 199, 154, 199, 160, 199, 45, 0, 161, 199, 163, 199, 45, 0, 166, 199, 172, 199, 45, 0, 173, 199, 176, 199, 180, 199, 188, 199, 45, 0, 189, 199, 191, 199, 45, 0, 193, 199, 200, 199, 45, 0, 201, 199, 204, 199, 206, 199, 208, 199, 216, 199, 221, 199, 228, 199, 232, 199, 236, 199, 0, 200, 45, 0, 1, 200, 4, 200, 8, 200, 10, 200, 16, 200, 45, 0, 17, 200, 19, 200, 21, 200, 45, 0, 22, 200, 28, 200, 45, 0, 29, 200, 32, 200, 36, 200, 44, 200, 45, 0, 45, 200, 47, 200, 49, 200, 56, 200, 60, 200, 64, 200, 72, 200, 45, 0, 73, 200, 76, 200, 45, 0, 77, 200, 84, 200, 112, 200, 45, 0, 113, 200, 116, 200, 120, 200, 122, 200, 128, 200, 45, 0, 129, 200, 131, 200, 133, 200, 45, 0, 135, 200, 139, 200, 45, 0, 141, 200, 148, 200, 157, 200, 159, 200, 161, 200, 168, 200, 188, 200, 45, 0, 189, 200, 196, 200, 200, 200, 204, 200, 212, 200, 45, 0, 213, 200, 215, 200, 217, 200, 224, 200, 45, 0, 225, 200, 228, 200, 245, 200, 252, 200, 45, 0, 253, 200, 0, 201, 4, 201, 45, 0, 6, 201, 12, 201, 45, 0, 13, 201, 15, 201, 17, 201, 24, 201, 44, 201, 52, 201, 80, 201, 45, 0, 81, 201, 84, 201, 88, 201, 96, 201, 45, 0, 97, 201, 99, 201, 108, 201, 112, 201, 116, 201, 124, 201, 136, 201, 45, 0, 137, 201, 140, 201, 144, 201, 152, 201, 45, 0, 153, 201, 155, 201, 157, 201, 192, 201, 45, 0, 193, 201, 196, 201, 199, 201, 45, 0, 200, 201, 202, 201, 208, 201, 45, 0, 209, 201, 211, 201, 213, 201, 45, 0, 214, 201, 217, 201, 45, 0, 218, 201, 220, 201, 45, 0, 221, 201, 224, 201, 226, 201, 228, 201, 231, 201, 236, 201, 45, 0, 237, 201, 239, 201, 45, 0, 241, 201, 248, 201, 45, 0, 249, 201, 252, 201, 0, 202, 8, 202, 45, 0, 9, 202, 11, 202, 45, 0, 13, 202, 20, 202, 24, 202, 41, 202, 76, 202, 45, 0, 77, 202, 80, 202, 84, 202, 92, 202, 45, 0, 93, 202, 95, 202, 45, 0, 97, 202, 104, 202, 125, 202, 132, 202, 152, 202, 188, 202, 45, 0, 189, 202, 192, 202, 196, 202, 204, 202, 45, 0, 205, 202, 207, 202, 209, 202, 211, 202, 216, 202, 45, 0, 217, 202, 224, 202, 236, 202, 244, 202, 8, 203, 16, 203, 20, 203, 24, 203, 32, 203, 45, 0, 33, 203, 65, 203, 72, 203, 45, 0, 73, 203, 76, 203, 80, 203, 88, 203, 45, 0, 89, 203, 93, 203, 100, 203, 120, 203, 45, 0, 121, 203, 156, 203, 184, 203, 212, 203, 228, 203, 231, 203, 233, 203, 12, 204, 45, 0, 13, 204, 16, 204, 20, 204, 28, 204, 45, 0, 29, 204, 33, 204, 45, 0, 34, 204, 39, 204, 45, 0, 41, 204, 44, 204, 46, 204, 48, 204, 56, 204, 45, 0, 57, 204, 59, 204, 45, 0, 62, 204, 68, 204, 45, 0, 69, 204, 72, 204, 76, 204, 84, 204, 45, 0, 85, 204, 87, 204, 45, 0, 89, 204, 96, 204, 100, 204, 102, 204, 104, 204, 112, 204, 117, 204, 152, 204, 45, 0, 153, 204, 156, 204, 160, 204, 168, 204, 45, 0, 169, 204, 171, 204, 45, 0, 173, 204, 180, 204, 45, 0, 181, 204, 184, 204, 188, 204, 196, 204, 45, 0, 197, 204, 199, 204, 201, 204, 208, 204, 212, 204, 228, 204, 236, 204, 240, 204, 1, 205, 8, 205, 45, 0, 9, 205, 12, 205, 16, 205, 24, 205, 45, 0, 25, 205, 27, 205, 29, 205, 36, 205, 40, 205, 44, 205, 57, 205, 92, 205, 96, 205, 100, 205, 108, 205, 45, 0, 109, 205, 111, 205, 113, 205, 120, 205, 136, 205, 148, 205, 45, 0, 149, 205, 152, 205, 156, 205, 164, 205, 45, 0, 165, 205, 167, 205, 169, 205, 176, 205, 196, 205, 204, 205, 208, 205, 232, 205, 236, 205, 240, 205, 248, 205, 45, 0, 249, 205, 251, 205, 253, 205, 4, 206, 8, 206, 12, 206, 20, 206, 25, 206, 32, 206, 45, 0, 33, 206, 36, 206, 40, 206, 48, 206, 45, 0, 49, 206, 51, 206, 53, 206, 88, 206, 45, 0, 89, 206, 92, 206, 95, 206, 45, 0, 97, 206, 104, 206, 45, 0, 105, 206, 107, 206, 109, 206, 116, 206, 45, 0, 117, 206, 120, 206, 124, 206, 132, 206, 45, 0, 133, 206, 135, 206, 137, 206, 144, 206, 45, 0, 145, 206, 148, 206, 152, 206, 160, 206, 45, 0, 161, 206, 163, 206, 45, 0, 165, 206, 172, 206, 45, 0, 173, 206, 193, 206, 228, 206, 45, 0, 229, 206, 232, 206, 235, 206, 45, 0, 236, 206, 244, 206, 45, 0, 245, 206, 247, 206, 45, 0, 249, 206, 0, 207, 45, 0, 1, 207, 4, 207, 8, 207, 16, 207, 45, 0, 17, 207, 19, 207, 21, 207, 28, 207, 32, 207, 36, 207, 44, 207, 45, 0, 45, 207, 47, 207, 45, 0, 49, 207, 56, 207, 84, 207, 45, 0, 85, 207, 88, 207, 92, 207, 100, 207, 45, 0, 101, 207, 103, 207, 105, 207, 112, 207, 45, 0, 113, 207, 116, 207, 120, 207, 128, 207, 133, 207, 140, 207, 161, 207, 168, 207, 176, 207, 196, 207, 224, 207, 45, 0, 225, 207, 228, 207, 232, 207, 240, 207, 45, 0, 241, 207, 243, 207, 245, 207, 252, 207, 0, 208, 4, 208, 17, 208, 24, 208, 45, 208, 52, 208, 45, 0, 53, 208, 56, 208, 60, 208, 68, 208, 45, 0, 69, 208, 71, 208, 73, 208, 80, 208, 84, 208, 88, 208, 96, 208, 108, 208, 45, 0, 109, 208, 112, 208, 116, 208, 124, 208, 45, 0, 125, 208, 129, 208, 164, 208, 45, 0, 165, 208, 168, 208, 172, 208, 180, 208, 45, 0, 181, 208, 183, 208, 185, 208, 192, 208, 45, 0, 193, 208, 196, 208, 200, 208, 45, 0, 201, 208, 208, 208, 45, 0, 209, 208, 211, 208, 45, 0, 213, 208, 220, 208, 45, 0, 221, 208, 224, 208, 228, 208, 236, 208, 45, 0, 237, 208, 239, 208, 45, 0, 241, 208, 248, 208, 13, 209, 48, 209, 45, 0, 49, 209, 52, 209, 56, 209, 58, 209, 64, 209, 45, 0, 65, 209, 67, 209, 45, 0, 69, 209, 76, 209, 45, 0, 77, 209, 80, 209, 84, 209, 92, 209, 45, 0, 93, 209, 95, 209, 97, 209, 104, 209, 108, 209, 124, 209, 132, 209, 136, 209, 160, 209, 45, 0, 161, 209, 164, 209, 168, 209, 176, 209, 45, 0, 177, 209, 179, 209, 181, 209, 186, 209, 188, 209, 192, 209, 216, 209, 244, 209, 248, 209, 7, 210, 9, 210, 16, 210, 44, 210, 45, 0, 45, 210, 48, 210, 52, 210, 60, 210, 45, 0, 61, 210, 63, 210, 65, 210, 72, 210, 92, 210, 100, 210, 128, 210, 45, 0, 129, 210, 132, 210, 136, 210, 144, 210, 45, 0, 145, 210, 149, 210, 156, 210, 160, 210, 164, 210, 172, 210, 177, 210, 184, 210, 45, 0, 185, 210, 188, 210, 191, 210, 45, 0, 192, 210, 194, 210, 200, 210, 45, 0, 201, 210, 203, 210, 212, 210, 216, 210, 220, 210, 228, 210, 45, 0, 229, 210, 240, 210, 45, 0, 241, 210, 244, 210, 248, 210, 0, 211, 45, 0, 1, 211, 3, 211, 5, 211, 12, 211, 45, 0, 14, 211, 16, 211, 20, 211, 22, 211, 28, 211, 45, 0, 29, 211, 31, 211, 45, 0, 33, 211, 37, 211, 40, 211, 45, 0, 41, 211, 44, 211, 48, 211, 56, 211, 45, 0, 57, 211, 59, 211, 45, 0, 61, 211, 68, 211, 45, 0, 69, 211, 124, 211, 45, 0, 125, 211, 128, 211, 132, 211, 140, 211, 45, 0, 141, 211, 143, 211, 45, 0, 145, 211, 152, 211, 45, 0, 153, 211, 156, 211, 160, 211, 168, 211, 45, 0, 169, 211, 171, 211, 173, 211, 180, 211, 184, 211, 188, 211, 196, 211, 45, 0, 197, 211, 200, 211, 45, 0, 201, 211, 208, 211, 216, 211, 225, 211, 227, 211, 236, 211, 45, 0, 237, 211, 240, 211, 244, 211, 252, 211, 45, 0, 253, 211, 255, 211, 1, 212, 8, 212, 29, 212, 64, 212, 68, 212, 92, 212, 96, 212, 100, 212, 109, 212, 111, 212, 120, 212, 45, 0, 121, 212, 124, 212, 127, 212, 45, 0, 128, 212, 130, 212, 136, 212, 45, 0, 137, 212, 139, 212, 141, 212, 148, 212, 169, 212, 204, 212, 208, 212, 212, 212, 220, 212, 223, 212, 232, 212, 236, 212, 240, 212, 248, 212, 251, 212, 253, 212, 4, 213, 8, 213, 12, 213, 20, 213, 45, 0, 21, 213, 23, 213, 60, 213, 45, 0, 61, 213, 64, 213, 68, 213, 76, 213, 45, 0, 77, 213, 79, 213, 81, 213, 88, 213, 45, 0, 89, 213, 92, 213, 96, 213, 101, 213, 104, 213, 45, 0, 105, 213, 107, 213, 109, 213, 116, 213, 45, 0, 117, 213, 120, 213, 124, 213, 132, 213, 45, 0, 133, 213, 135, 213, 45, 0, 137, 213, 144, 213, 165, 213, 200, 213, 45, 0, 201, 213, 204, 213, 208, 213, 210, 213, 216, 213, 45, 0, 217, 213, 219, 213, 221, 213, 228, 213, 45, 0, 229, 213, 232, 213, 236, 213, 244, 213, 45, 0, 245, 213, 247, 213, 249, 213, 0, 214, 45, 0, 1, 214, 4, 214, 8, 214, 16, 214, 45, 0, 17, 214, 19, 214, 45, 0, 21, 214, 28, 214, 32, 214, 36, 214, 45, 214, 56, 214, 45, 0, 57, 214, 60, 214, 64, 214, 69, 214, 72, 214, 45, 0, 73, 214, 75, 214, 77, 214, 81, 214, 84, 214, 45, 0, 85, 214, 88, 214, 92, 214, 103, 214, 105, 214, 112, 214, 45, 0, 113, 214, 116, 214, 131, 214, 133, 214, 140, 214, 45, 0, 141, 214, 144, 214, 148, 214, 157, 214, 159, 214, 161, 214, 168, 214, 172, 214, 176, 214, 185, 214, 187, 214, 196, 214, 45, 0, 197, 214, 200, 214, 204, 214, 209, 214, 212, 214, 215, 214, 217, 214, 224, 214, 228, 214, 232, 214, 240, 214, 245, 214, 252, 214, 45, 0, 253, 214, 0, 215, 4, 215, 17, 215, 24, 215, 45, 0, 25, 215, 28, 215, 32, 215, 40, 215, 45, 0, 41, 215, 43, 215, 45, 215, 52, 215, 45, 0, 53, 215, 56, 215, 60, 215, 68, 215, 71, 215, 73, 215, 80, 215, 45, 0, 81, 215, 84, 215, 86, 215, 45, 0, 89, 215, 96, 215, 45, 0, 97, 215, 99, 215, 101, 215, 105, 215, 108, 215, 112, 215, 116, 215, 124, 215, 45, 0, 125, 215, 129, 215, 136, 215, 45, 0, 137, 215, 140, 215, 144, 215, 152, 215, 45, 0, 153, 215, 155, 215, 157, 215, 93, 0, 93, 0, 38, 0, 0, 172, 60, 0, 60, 0, 42, 0, 61, 79, 115, 79, 71, 80, 249, 80, 160, 82, 239, 83, 117, 84, 229, 84, 9, 86, 193, 90, 182, 91, 135, 102, 182, 103, 183, 103, 239, 103, 76, 107, 194, 115, 194, 117, 60, 122, 219, 130, 4, 131, 87, 136, 136, 136, 54, 138, 200, 140, 207, 141, 251, 142, 230, 143, 213, 153, 238, 78, 162, 80, 150, 84, 255, 84, 119, 87, 138, 91, 157, 101, 142, 105, 159, 106, 200, 115, 51, 123, 30, 128, 56, 130, 109, 132, 12, 139, 38, 0, 1, 172, 60, 0, 60, 0, 42, 0, 59, 82, 116, 83, 4, 84, 106, 96, 100, 97, 188, 107, 207, 115, 26, 129, 186, 137, 210, 137, 163, 149, 123, 83, 175, 84, 198, 87, 9, 100, 241, 100, 119, 104, 38, 0, 4, 172, 60, 0, 60, 0, 42, 0, 131, 79, 10, 82, 190, 88, 120, 89, 230, 89, 114, 94, 121, 94, 199, 97, 192, 99, 70, 103, 236, 103, 127, 104, 151, 111, 78, 118, 11, 119, 245, 120, 8, 122, 255, 122, 33, 124, 157, 128, 110, 130, 113, 130, 235, 138, 147, 149, 88, 80, 115, 97, 30, 104, 166, 105, 149, 115, 198, 121, 27, 131, 78, 136, 118, 141, 192, 143, 102, 159, 38, 0, 8, 172, 60, 0, 60, 0, 42, 0, 107, 78, 157, 85, 247, 102, 52, 110, 163, 120, 237, 122, 91, 132, 16, 137, 78, 135, 168, 151, 118, 86, 108, 105, 248, 121, 175, 127, 13, 136, 161, 157, 38, 0, 16, 172, 60, 0, 60, 0, 42, 0, 216, 82, 78, 87, 42, 88, 76, 93, 31, 97, 190, 97, 33, 98, 98, 101, 209, 103, 68, 106, 27, 110, 24, 117, 179, 117, 227, 118, 176, 119, 58, 125, 175, 144, 81, 148, 82, 148, 149, 159, 105, 87, 243, 87, 65, 93, 7, 95, 168, 97, 188, 100, 63, 107, 91, 107, 212, 108, 230, 109, 137, 111, 217, 119, 87, 143, 99, 145, 123, 158, 38, 0, 17, 172, 60, 0, 60, 0, 42, 0, 35, 83, 172, 92, 50, 117, 219, 128, 64, 146, 152, 149, 208, 151, 38, 0, 21, 172, 60, 0, 60, 0, 42, 0, 91, 82, 8, 88, 220, 89, 161, 92, 23, 93, 183, 94, 58, 95, 74, 95, 119, 97, 95, 108, 122, 117, 134, 117, 224, 124, 115, 125, 177, 125, 140, 127, 84, 129, 33, 130, 145, 133, 65, 137, 27, 139, 252, 146, 77, 150, 71, 156, 139, 80, 245, 80, 195, 88, 252, 95, 91, 98, 96, 103, 127, 106, 173, 107, 252, 119, 69, 122, 72, 126, 97, 127, 151, 127, 171, 127, 51, 131, 71, 140, 193, 151, 38, 0, 28, 172, 60, 0, 60, 0, 42, 0, 203, 78, 247, 78, 11, 80, 241, 81, 79, 88, 55, 97, 62, 97, 104, 97, 57, 101, 234, 105, 17, 111, 165, 117, 134, 118, 214, 118, 135, 123, 165, 130, 203, 132, 167, 147, 139, 149, 116, 82, 3, 83, 233, 99, 233, 105, 160, 115, 213, 120, 211, 149, 38, 0, 29, 172, 60, 0, 60, 0, 42, 0, 128, 85, 162, 91, 38, 0, 49, 172, 60, 0, 60, 0, 42, 0, 81, 87, 179, 124, 185, 127, 92, 120, 225, 140, 215, 147, 38, 0, 57, 172, 60, 0, 60, 0, 181, 145, 38, 0, 112, 172, 60, 0, 60, 0, 42, 0, 40, 80, 187, 83, 69, 92, 232, 93, 210, 98, 110, 99, 218, 100, 231, 100, 32, 110, 172, 112, 91, 121, 221, 141, 30, 142, 125, 144, 69, 146, 248, 146, 127, 84, 27, 102, 236, 121, 101, 123, 103, 124, 224, 128, 82, 129, 227, 130, 146, 131, 86, 133, 39, 134, 170, 136, 254, 136, 207, 153, 38, 0, 116, 172, 60, 0, 60, 0, 42, 0, 126, 78, 246, 78, 101, 80, 254, 93, 250, 94, 6, 97, 87, 105, 113, 129, 84, 134, 71, 142, 117, 147, 43, 154, 245, 99, 141, 114, 119, 119, 48, 137, 7, 139, 172, 151, 38, 0, 120, 172, 60, 0, 60, 0, 42, 0, 94, 78, 145, 80, 112, 103, 64, 104, 108, 78, 5, 103, 164, 105, 38, 0, 128, 172, 60, 0, 60, 0, 42, 0, 9, 81, 141, 82, 146, 82, 162, 106, 188, 119, 16, 146, 212, 158, 191, 100, 161, 130, 38, 0, 129, 172, 60, 0, 60, 0, 42, 0, 171, 82, 47, 96, 242, 143, 38, 82, 39, 82, 38, 0, 140, 172, 60, 0, 60, 0, 42, 0, 72, 80, 169, 97, 237, 99, 38, 0, 169, 172, 60, 0, 60, 0, 42, 0, 202, 100, 60, 104, 132, 106, 192, 111, 136, 129, 161, 137, 148, 150, 12, 99, 196, 107, 195, 149, 188, 154, 50, 155, 3, 157, 38, 0, 172, 172, 60, 0, 60, 0, 42, 0, 5, 88, 125, 114, 172, 114, 4, 117, 121, 125, 109, 126, 169, 128, 139, 137, 116, 139, 99, 144, 81, 157, 43, 106, 247, 114, 78, 117, 103, 123, 51, 126, 126, 126, 130, 127, 50, 136, 57, 156, 38, 0, 176, 172, 60, 0, 60, 0, 42, 0, 137, 98, 122, 108, 84, 111, 80, 125, 58, 127, 35, 138, 166, 115, 214, 137, 203, 149, 38, 0, 184, 172, 60, 0, 60, 0, 42, 0, 124, 81, 74, 97, 157, 123, 25, 139, 87, 146, 140, 147, 148, 80, 219, 85, 146, 92, 209, 98, 73, 107, 17, 126, 185, 132, 218, 158, 56, 159, 38, 0, 189, 172, 60, 0, 60, 0, 42, 0, 172, 78, 211, 79, 30, 80, 190, 80, 6, 81, 193, 82, 205, 82, 127, 83, 112, 87, 131, 88, 154, 94, 145, 95, 118, 97, 172, 97, 206, 100, 108, 101, 111, 102, 187, 102, 244, 102, 151, 104, 135, 109, 133, 112, 241, 112, 159, 116, 165, 116, 202, 116, 217, 117, 108, 120, 236, 120, 223, 122, 246, 122, 69, 125, 147, 125, 21, 128, 63, 128, 27, 129, 150, 131, 102, 139, 21, 143, 21, 144, 225, 147, 3, 152, 56, 152, 90, 154, 232, 155, 143, 81, 68, 82, 253, 84, 248, 96, 188, 97, 67, 98, 160, 106, 98, 113, 45, 113, 178, 113, 74, 117, 248, 122, 134, 125, 72, 152, 68, 127, 39, 137, 38, 139, 201, 153, 193, 155, 229, 158, 38, 0, 196, 172, 60, 0, 60, 0, 42, 0, 194, 79, 83, 85, 58, 88, 81, 89, 99, 91, 70, 92, 184, 96, 18, 98, 66, 104, 176, 104, 232, 104, 170, 110, 76, 117, 120, 118, 206, 120, 61, 122, 251, 124, 107, 126, 124, 126, 8, 138, 161, 138, 63, 140, 142, 150, 196, 157, 38, 88, 211, 112, 8, 118, 138, 121, 83, 123, 174, 125, 24, 126, 125, 127, 106, 132, 138, 133, 222, 150, 251, 154, 38, 0, 224, 172, 60, 0, 60, 0, 42, 0, 228, 83, 233, 83, 74, 84, 113, 84, 250, 86, 209, 89, 100, 91, 59, 92, 171, 94, 247, 98, 55, 101, 69, 101, 114, 101, 160, 102, 175, 103, 193, 105, 189, 108, 252, 117, 144, 118, 126, 119, 63, 122, 148, 127, 3, 128, 161, 128, 143, 129, 230, 130, 253, 130, 240, 131, 193, 133, 49, 136, 180, 136, 165, 138, 156, 143, 46, 147, 199, 150, 103, 152, 216, 154, 19, 159, 48, 79, 197, 81, 51, 82, 89, 88, 114, 103, 50, 104, 192, 105, 249, 105, 220, 106, 111, 114, 139, 118, 236, 118, 189, 119, 1, 122, 141, 123, 217, 123, 213, 124, 95, 127, 150, 127, 250, 127, 239, 128, 218, 137, 65, 138, 220, 144, 100, 145, 55, 146, 96, 151, 35, 157, 241, 157, 38, 0, 225, 172, 60, 0, 60, 0, 42, 0, 237, 84, 155, 101, 242, 102, 143, 104, 64, 122, 55, 140, 96, 157, 179, 86, 242, 105, 32, 126, 243, 137, 66, 143, 38, 0, 228, 172, 60, 0, 60, 0, 42, 0, 240, 86, 100, 87, 17, 93, 6, 102, 177, 104, 205, 104, 254, 110, 40, 116, 158, 136, 228, 155, 3, 88, 16, 93, 131, 96, 70, 99, 196, 125, 110, 136, 205, 136, 12, 137, 21, 147, 171, 149, 225, 154, 126, 157, 164, 157, 107, 159, 38, 0, 232, 172, 60, 0, 60, 0, 42, 0, 104, 108, 168, 154, 48, 100, 190, 105, 251, 119, 187, 157, 38, 0, 245, 172, 60, 0, 60, 0, 42, 0, 155, 79, 108, 81, 113, 81, 159, 82, 84, 91, 229, 93, 80, 96, 109, 96, 241, 98, 167, 99, 59, 101, 217, 115, 122, 122, 163, 134, 162, 140, 143, 151, 37, 80, 6, 93, 190, 96, 49, 104, 211, 105, 156, 123, 233, 134, 236, 134, 27, 141, 235, 141, 237, 145, 148, 159, 38, 0, 246, 172, 60, 0, 60, 0, 50, 78, 38, 0, 252, 172, 60, 0, 60, 0, 42, 0, 225, 91, 8, 98, 156, 103, 220, 116, 209, 121, 211, 131, 135, 138, 178, 138, 232, 141, 78, 144, 75, 147, 70, 152, 137, 79, 29, 88, 37, 89, 56, 89, 190, 100, 19, 115, 30, 122, 160, 122, 76, 135, 249, 136, 29, 142, 153, 146, 13, 154, 38, 0, 253, 172, 60, 0, 60, 0, 42, 0, 211, 94, 232, 105, 255, 133, 237, 144, 1, 105, 104, 118, 169, 142, 13, 151, 185, 151, 38, 0, 0, 173, 60, 0, 60, 0, 42, 0, 160, 81, 152, 91, 236, 91, 99, 97, 250, 104, 62, 107, 76, 112, 47, 116, 216, 116, 161, 123, 80, 127, 197, 131, 192, 137, 171, 140, 220, 149, 40, 153, 49, 78, 171, 109, 31, 114, 229, 118, 124, 121, 190, 122, 102, 123, 176, 125, 40, 143, 39, 147, 117, 148, 218, 150, 116, 152, 214, 154, 27, 158, 38, 0, 4, 173, 60, 0, 60, 0, 42, 0, 46, 82, 93, 96, 236, 98, 2, 144, 120, 79, 29, 104, 72, 123, 82, 128, 250, 154, 48, 157, 38, 0, 17, 173, 60, 0, 60, 0, 42, 0, 138, 79, 73, 81, 33, 83, 217, 88, 227, 94, 224, 102, 56, 109, 154, 112, 194, 114, 214, 115, 80, 123, 241, 128, 91, 148, 71, 96, 68, 104, 70, 104, 12, 114, 119, 115, 250, 120, 86, 125, 138, 126, 42, 131, 134, 138, 145, 138, 38, 0, 24, 173, 60, 0, 60, 0, 42, 0, 102, 83, 155, 99, 107, 127, 188, 84, 2, 99, 99, 127, 127, 138, 38, 0, 52, 173, 60, 0, 60, 0, 42, 0, 86, 78, 128, 80, 74, 88, 222, 88, 42, 96, 39, 97, 208, 98, 208, 105, 65, 155, 191, 90, 229, 94, 96, 96, 112, 116, 157, 116, 175, 132, 88, 137, 38, 0, 53, 173, 60, 0, 60, 0, 152, 153, 38, 0, 73, 173, 60, 0, 60, 0, 42, 0, 143, 91, 24, 125, 177, 128, 95, 143, 100, 109, 229, 137, 7, 138, 142, 149, 38, 0, 80, 173, 60, 0, 60, 0, 42, 0, 164, 78, 209, 80, 172, 84, 172, 85, 12, 91, 160, 93, 231, 93, 42, 101, 78, 101, 33, 104, 75, 106, 225, 114, 142, 118, 239, 119, 94, 125, 249, 127, 160, 129, 78, 133, 223, 134, 3, 143, 78, 143, 202, 144, 3, 153, 85, 154, 171, 155, 124, 79, 4, 86, 16, 86, 109, 86, 153, 86, 227, 89, 141, 97, 159, 100, 72, 102, 158, 102, 183, 105, 253, 120, 150, 122, 171, 141, 123, 142, 120, 146, 185, 154, 65, 157, 105, 159, 38, 0, 108, 173, 60, 0, 60, 0, 42, 0, 24, 78, 69, 78, 93, 78, 199, 78, 241, 79, 119, 81, 254, 82, 64, 83, 227, 83, 229, 83, 142, 84, 20, 86, 117, 87, 162, 87, 199, 91, 135, 93, 208, 94, 252, 97, 216, 98, 81, 101, 184, 103, 233, 103, 203, 105, 80, 107, 198, 107, 236, 107, 66, 108, 157, 110, 120, 112, 215, 114, 150, 115, 3, 116, 191, 119, 233, 119, 118, 122, 127, 125, 9, 128, 252, 129, 5, 130, 10, 130, 223, 130, 98, 136, 51, 139, 252, 140, 192, 142, 17, 144, 177, 144, 100, 146, 182, 146, 210, 153, 69, 154, 233, 156, 215, 157, 156, 159, 73, 79, 93, 79, 197, 79, 180, 80, 147, 81, 172, 82, 54, 83, 185, 83, 244, 83, 120, 87, 228, 89, 190, 90, 215, 90, 104, 92, 163, 92, 64, 95, 53, 98, 99, 98, 68, 99, 6, 100, 115, 100, 43, 102, 152, 105, 26, 111, 134, 116, 12, 117, 154, 117, 192, 117, 111, 118, 155, 122, 182, 122, 221, 123, 215, 124, 202, 128, 159, 132, 175, 134, 216, 136, 175, 137, 108, 138, 88, 144, 230, 145, 221, 151, 237, 151, 238, 151, 182, 152, 200, 153, 46, 155, 199, 157, 28, 158, 38, 0, 109, 173, 60, 0, 60, 0, 42, 0, 11, 87, 64, 92, 202, 131, 160, 151, 171, 151, 180, 158, 10, 83, 172, 99, 252, 141, 175, 158, 38, 0, 112, 173, 60, 0, 60, 0, 42, 0, 27, 84, 152, 122, 164, 127, 217, 136, 205, 142, 225, 144, 67, 99, 126, 104, 184, 118, 38, 0, 116, 173, 60, 0, 60, 0, 42, 0, 0, 88, 72, 92, 152, 99, 159, 122, 20, 80, 27, 93, 200, 109, 88, 138, 38, 0, 129, 173, 60, 0, 60, 0, 42, 0, 174, 91, 19, 95, 121, 122, 174, 122, 142, 130, 172, 142, 179, 142, 38, 0, 140, 173, 60, 0, 60, 0, 42, 0, 38, 80, 56, 82, 248, 82, 119, 83, 8, 87, 243, 98, 114, 99, 10, 107, 195, 109, 55, 119, 204, 82, 211, 96, 236, 104, 96, 119, 163, 125, 55, 135, 38, 0, 144, 173, 60, 0, 60, 0, 42, 0, 165, 83, 87, 115, 104, 133, 118, 142, 213, 149, 38, 0, 164, 173, 60, 0, 60, 0, 42, 0, 58, 103, 195, 106, 112, 111, 109, 138, 204, 142, 75, 153, 121, 79, 224, 81, 130, 82, 49, 83, 146, 97, 133, 100, 59, 106, 63, 108, 11, 124, 98, 126, 234, 141, 224, 149, 61, 153, 130, 158, 38, 0, 192, 173, 60, 0, 60, 0, 42, 0, 119, 102, 120, 107, 180, 140, 60, 155, 38, 0, 220, 173, 60, 0, 60, 0, 42, 0, 235, 83, 45, 87, 78, 89, 198, 99, 251, 105, 234, 115, 69, 120, 186, 122, 197, 122, 254, 124, 117, 132, 143, 137, 115, 141, 53, 144, 168, 149, 50, 82, 226, 90, 0, 91, 203, 93, 140, 102, 79, 105, 27, 106, 89, 111, 125, 119, 250, 124, 108, 134, 111, 134, 236, 141, 189, 144, 218, 149, 13, 152, 151, 153, 38, 0, 224, 173, 60, 0, 60, 0, 42, 0, 251, 82, 71, 87, 71, 117, 96, 123, 204, 131, 30, 146, 247, 86, 143, 158, 38, 0, 228, 173, 60, 0, 60, 0, 88, 106, 38, 0, 249, 173, 60, 0, 60, 0, 42, 0, 75, 81, 75, 82, 135, 82, 31, 98, 216, 104, 117, 105, 153, 150, 159, 78, 5, 92, 80, 92, 196, 144, 38, 0, 252, 173, 60, 0, 60, 0, 42, 0, 197, 80, 164, 82, 228, 82, 195, 97, 164, 101, 57, 104, 255, 105, 126, 116, 75, 123, 185, 130, 235, 131, 178, 137, 57, 139, 209, 143, 73, 153, 122, 83, 170, 83, 144, 88, 249, 93, 209, 94, 12, 111, 212, 137, 223, 141, 255, 145, 115, 151, 38, 0, 8, 174, 60, 0, 60, 0, 42, 0, 202, 78, 151, 89, 210, 100, 17, 102, 142, 106, 52, 116, 129, 121, 189, 121, 169, 130, 126, 136, 127, 136, 95, 137, 38, 147, 43, 85, 100, 86, 148, 93, 18, 123, 197, 158, 38, 0, 9, 174, 60, 0, 60, 0, 42, 0, 11, 79, 202, 83, 37, 96, 113, 98, 114, 108, 26, 125, 102, 125, 62, 87, 140, 92, 128, 118, 15, 121, 8, 123, 168, 130, 38, 0, 13, 174, 60, 0, 60, 0, 42, 0, 152, 78, 98, 81, 220, 119, 175, 128, 153, 78, 145, 107, 38, 0, 48, 174, 60, 0, 60, 0, 42, 0, 1, 79, 14, 79, 118, 81, 128, 81, 220, 85, 104, 86, 59, 87, 250, 87, 252, 87, 20, 89, 71, 89, 147, 89, 196, 91, 144, 92, 14, 93, 241, 93, 126, 94, 204, 95, 128, 98, 215, 101, 227, 101, 30, 103, 31, 103, 94, 103, 203, 104, 196, 104, 95, 106, 58, 107, 35, 108, 125, 108, 130, 108, 199, 109, 152, 115, 38, 116, 42, 116, 130, 116, 163, 116, 120, 117, 127, 117, 129, 120, 239, 120, 65, 121, 71, 121, 72, 121, 122, 121, 149, 123, 0, 125, 186, 125, 136, 127, 6, 128, 45, 128, 140, 128, 24, 138, 79, 139, 72, 140, 119, 141, 33, 147, 36, 147, 226, 152, 81, 153, 14, 154, 15, 154, 101, 154, 146, 158, 219, 80, 94, 82, 141, 88, 122, 92, 139, 94, 3, 95, 238, 95, 45, 97, 142, 99, 50, 101, 194, 101, 163, 102, 193, 102, 202, 104, 103, 107, 129, 112, 9, 115, 168, 121, 165, 125, 166, 125, 135, 127, 181, 128, 170, 130, 176, 130, 4, 134, 65, 134, 29, 135, 227, 135, 137, 137, 172, 137, 194, 141, 145, 150, 14, 152, 16, 155, 45, 156, 214, 158, 38, 0, 52, 174, 60, 0, 60, 0, 202, 125, 38, 0, 56, 174, 60, 0, 60, 0, 42, 0, 118, 79, 9, 84, 238, 98, 84, 104, 222, 89, 227, 134, 38, 0, 64, 174, 60, 0, 60, 0, 209, 145, 38, 0, 125, 176, 60, 0, 60, 0, 171, 85, 38, 0, 152, 176, 60, 0, 60, 0, 42, 0, 58, 81, 28, 90, 230, 97, 207, 98, 255, 98, 163, 144, 16, 99, 42, 99, 155, 104, 229, 124, 239, 124, 38, 0, 153, 176, 60, 0, 60, 0, 254, 138, 38, 0, 156, 176, 60, 0, 60, 0, 42, 0, 150, 102, 86, 113, 227, 150, 68, 80, 87, 113, 103, 141, 42, 153, 38, 0, 160, 176, 60, 0, 60, 0, 42, 0, 79, 99, 122, 99, 38, 0, 168, 176, 60, 0, 60, 0, 42, 0, 87, 83, 143, 103, 96, 105, 115, 110, 55, 117, 131, 85, 223, 103, 38, 0, 169, 176, 60, 0, 60, 0, 42, 0, 13, 125, 114, 136, 38, 0, 173, 176, 60, 0, 60, 0, 42, 0, 202, 86, 24, 90, 233, 102, 38, 0, 180, 176, 60, 0, 60, 0, 42, 0, 67, 78, 103, 81, 72, 89, 240, 103, 16, 128, 2, 83, 118, 89, 45, 91, 250, 143, 16, 159, 38, 0, 64, 177, 60, 0, 60, 0, 115, 89, 38, 0, 65, 177, 60, 0, 60, 0, 196, 96, 38, 0, 68, 177, 60, 0, 60, 0, 42, 0, 116, 94, 154, 100, 202, 121, 190, 120, 38, 0, 80, 177, 60, 0, 60, 0, 42, 0, 245, 95, 108, 96, 200, 98, 123, 99, 38, 0, 85, 177, 60, 0, 60, 0, 42, 0, 231, 91, 215, 91, 94, 79, 28, 81, 128, 86, 216, 111, 38, 0, 120, 177, 60, 0, 60, 0, 42, 0, 170, 82, 116, 89, 41, 95, 18, 96, 89, 116, 209, 153, 118, 84, 101, 91, 241, 92, 49, 115, 47, 123, 209, 129, 38, 0, 141, 177, 60, 0, 60, 0, 42, 0, 195, 111, 191, 129, 178, 143, 2, 81, 101, 86, 96, 122, 178, 145, 38, 0, 204, 177, 60, 0, 60, 0, 42, 0, 241, 96, 102, 129, 18, 153, 38, 0, 232, 177, 60, 0, 60, 0, 42, 0, 63, 92, 203, 90, 50, 91, 214, 109, 224, 120, 202, 136, 3, 148, 38, 0, 4, 178, 60, 0, 60, 0, 42, 0, 66, 85, 40, 128, 38, 0, 8, 178, 60, 0, 60, 0, 233, 90, 38, 0, 12, 178, 60, 0, 60, 0, 42, 0, 37, 138, 54, 84, 173, 128, 38, 0, 116, 178, 60, 0, 60, 0, 42, 0, 123, 103, 16, 125, 248, 95, 117, 151, 38, 0, 117, 178, 60, 0, 60, 0, 68, 136, 38, 0, 165, 178, 60, 0, 60, 0, 253, 128, 38, 0, 200, 178, 60, 0, 60, 0, 42, 0, 60, 92, 229, 108, 98, 84, 41, 96, 197, 103, 98, 121, 176, 121, 169, 129, 38, 0, 201, 178, 60, 0, 60, 0, 42, 0, 63, 83, 186, 110, 53, 102, 177, 102, 38, 0, 228, 178, 60, 0, 60, 0, 42, 0, 26, 89, 54, 131, 57, 114, 38, 0, 232, 178, 60, 0, 60, 0, 42, 0, 57, 78, 182, 78, 70, 79, 174, 85, 24, 87, 199, 88, 86, 95, 183, 101, 230, 101, 128, 106, 181, 107, 77, 110, 237, 119, 239, 122, 30, 124, 222, 125, 203, 134, 146, 136, 50, 145, 91, 147, 113, 97, 197, 98, 52, 105, 25, 111, 73, 118, 17, 128, 198, 128, 118, 129, 17, 135, 38, 0, 236, 178, 60, 0, 60, 0, 42, 0, 187, 100, 190, 111, 122, 115, 184, 117, 84, 144, 178, 89, 27, 96, 229, 149, 124, 151, 195, 151, 38, 0, 244, 178, 60, 0, 60, 0, 42, 0, 86, 85, 77, 87, 186, 97, 212, 100, 199, 102, 225, 109, 91, 110, 109, 111, 185, 111, 240, 117, 67, 128, 189, 129, 65, 133, 131, 137, 199, 138, 90, 139, 31, 147, 11, 81, 87, 85, 73, 86, 176, 88, 220, 88, 239, 107, 171, 121, 78, 127, 157, 133, 239, 144, 238, 158, 245, 158, 38, 0, 245, 178, 60, 0, 60, 0, 42, 0, 147, 108, 83, 117, 84, 123, 15, 142, 93, 144, 38, 0, 249, 178, 60, 0, 60, 0, 42, 0, 16, 85, 2, 88, 88, 88, 98, 94, 7, 98, 158, 100, 224, 104, 118, 117, 214, 124, 179, 135, 232, 158, 24, 80, 59, 81, 90, 81, 42, 100, 148, 106, 143, 110, 109, 116, 171, 116, 160, 119, 17, 121, 247, 135, 96, 137, 156, 139, 220, 147, 58, 148, 51, 153, 57, 153, 38, 0, 0, 179, 60, 0, 60, 0, 42, 0, 227, 78, 136, 87, 110, 87, 39, 89, 13, 92, 177, 92, 54, 94, 133, 95, 52, 98, 225, 100, 179, 115, 250, 129, 139, 136, 184, 140, 138, 150, 219, 158, 19, 81, 223, 97, 242, 101, 79, 108, 147, 120, 19, 148, 38, 0, 1, 179, 60, 0, 60, 0, 133, 91, 38, 0, 85, 179, 60, 0, 60, 0, 42, 0, 183, 95, 179, 96, 38, 0, 196, 179, 60, 0, 60, 0, 42, 0, 18, 80, 0, 82, 48, 82, 22, 87, 53, 88, 87, 88, 14, 92, 96, 92, 246, 92, 139, 93, 166, 94, 146, 95, 188, 96, 17, 99, 137, 99, 23, 100, 67, 104, 249, 104, 194, 106, 216, 109, 33, 110, 212, 110, 228, 111, 254, 113, 220, 118, 121, 119, 177, 121, 59, 122, 4, 132, 169, 137, 237, 140, 243, 141, 72, 142, 3, 144, 20, 144, 83, 144, 253, 144, 77, 147, 118, 150, 220, 151, 232, 83, 212, 88, 34, 95, 201, 95, 70, 97, 143, 99, 47, 100, 227, 100, 174, 106, 46, 109, 130, 109, 12, 122, 223, 131, 116, 145, 205, 149, 128, 151, 177, 151, 85, 153, 23, 159, 38, 0, 197, 179, 60, 0, 60, 0, 42, 0, 210, 107, 6, 112, 88, 114, 162, 114, 104, 115, 99, 119, 191, 121, 228, 123, 155, 126, 128, 139, 221, 106, 247, 158, 38, 0, 200, 179, 60, 0, 60, 0, 42, 0, 169, 88, 199, 96, 102, 101, 253, 101, 190, 102, 140, 108, 30, 113, 201, 113, 90, 140, 19, 152, 52, 95, 97, 111, 137, 142, 38, 0, 204, 179, 60, 0, 60, 0, 42, 0, 109, 78, 129, 122, 132, 84, 23, 88, 38, 0, 217, 179, 60, 0, 60, 0, 42, 0, 221, 78, 172, 81, 205, 81, 213, 82, 12, 84, 167, 97, 113, 103, 80, 104, 223, 104, 30, 109, 124, 111, 188, 117, 179, 119, 229, 122, 244, 128, 99, 132, 133, 146, 151, 79, 238, 80, 195, 84, 140, 87, 210, 92, 100, 95, 35, 103, 102, 106, 183, 109, 95, 130, 243, 130, 60, 131, 107, 133, 64, 135, 44, 147, 151, 155, 38, 0, 80, 180, 60, 0, 60, 0, 42, 0, 92, 81, 151, 101, 92, 103, 147, 103, 216, 117, 199, 122, 115, 131, 70, 140, 23, 144, 45, 152, 150, 98, 129, 101, 154, 128, 48, 129, 170, 134, 57, 136, 97, 150, 38, 0, 84, 180, 60, 0, 60, 0, 42, 0, 111, 92, 192, 129, 154, 130, 65, 144, 111, 144, 13, 146, 128, 122, 205, 143, 38, 0, 88, 180, 60, 0, 60, 0, 103, 78, 38, 0, 221, 180, 60, 0, 60, 0, 151, 95, 38, 0, 241, 180, 60, 0, 60, 0, 42, 0, 157, 93, 89, 106, 200, 113, 123, 118, 73, 123, 228, 133, 4, 139, 39, 145, 48, 154, 243, 81, 177, 88, 213, 110, 244, 120, 80, 124, 34, 126, 163, 135, 25, 148, 38, 0, 124, 183, 60, 0, 60, 0, 42, 0, 135, 85, 246, 97, 105, 118, 133, 127, 63, 134, 186, 135, 248, 136, 143, 144, 46, 80, 201, 86, 234, 102, 48, 118, 34, 120, 221, 129, 124, 148, 62, 154, 88, 154, 38, 0, 125, 183, 60, 0, 60, 0, 42, 0, 27, 109, 217, 112, 222, 115, 97, 125, 61, 132, 106, 145, 241, 153, 224, 85, 150, 114, 38, 0, 128, 183, 60, 0, 60, 0, 42, 0, 130, 78, 117, 83, 4, 107, 18, 107, 62, 112, 27, 114, 45, 134, 30, 158, 62, 91, 113, 94, 20, 101, 83, 112, 116, 137, 126, 148, 204, 149, 38, 0, 132, 183, 60, 0, 60, 0, 42, 0, 76, 82, 163, 143, 210, 87, 162, 143, 38, 0, 140, 183, 60, 0, 60, 0, 42, 0, 80, 93, 229, 100, 44, 101, 22, 107, 235, 111, 67, 124, 156, 126, 205, 133, 100, 137, 189, 137, 106, 90, 207, 96, 38, 0, 141, 183, 60, 0, 60, 0, 42, 0, 201, 98, 216, 129, 31, 136, 94, 148, 38, 0, 145, 183, 60, 0, 60, 0, 42, 0, 202, 94, 23, 103, 106, 109, 252, 114, 5, 116, 111, 116, 130, 135, 222, 144, 148, 105, 96, 120, 2, 122, 168, 131, 11, 135, 172, 149, 38, 0, 152, 183, 60, 0, 60, 0, 42, 0, 134, 79, 13, 93, 160, 95, 10, 132, 246, 109, 11, 154, 38, 0, 173, 183, 60, 0, 60, 0, 183, 81, 38, 0, 181, 183, 60, 0, 60, 0, 42, 0, 160, 99, 101, 117, 103, 117, 38, 0, 201, 183, 60, 0, 60, 0, 42, 0, 174, 78, 6, 80, 105, 81, 201, 81, 129, 104, 17, 106, 174, 124, 177, 124, 231, 124, 111, 130, 210, 138, 27, 143, 207, 145, 168, 85, 162, 96, 11, 105, 188, 109, 9, 142, 78, 155, 38, 0, 36, 184, 60, 0, 60, 0, 42, 0, 182, 79, 55, 81, 245, 82, 66, 84, 236, 94, 110, 97, 62, 98, 197, 101, 218, 106, 254, 111, 42, 121, 220, 133, 35, 136, 173, 149, 98, 154, 106, 154, 151, 158, 206, 158, 34, 81, 178, 83, 51, 85, 160, 104, 88, 118, 242, 124, 130, 129, 218, 129, 33, 136, 140, 144, 98, 148, 38, 0, 37, 184, 60, 0, 60, 0, 42, 0, 155, 82, 198, 102, 119, 107, 29, 112, 43, 121, 98, 143, 66, 151, 10, 101, 223, 106, 234, 106, 103, 118, 99, 143, 72, 145, 38, 0, 40, 184, 60, 0, 60, 0, 42, 0, 144, 97, 0, 98, 35, 101, 35, 111, 73, 113, 137, 116, 244, 125, 111, 128, 238, 132, 38, 143, 35, 144, 74, 147, 76, 91, 93, 105, 69, 110, 224, 129, 200, 147, 10, 156, 49, 156, 38, 0, 44, 184, 60, 0, 60, 0, 42, 0, 189, 81, 23, 82, 163, 82, 12, 109, 200, 112, 194, 136, 18, 99, 105, 99, 178, 152, 38, 0, 52, 184, 60, 0, 60, 0, 42, 0, 201, 94, 130, 101, 174, 107, 194, 111, 62, 124, 105, 89, 50, 112, 207, 120, 38, 0, 53, 184, 60, 0, 60, 0, 42, 0, 117, 115, 144, 142, 35, 155, 38, 0, 57, 184, 60, 0, 60, 0, 42, 0, 228, 78, 54, 79, 249, 86, 186, 92, 186, 93, 28, 96, 178, 115, 45, 123, 154, 127, 206, 127, 70, 128, 30, 144, 52, 146, 246, 150, 72, 151, 24, 152, 97, 159, 230, 83, 100, 84, 200, 89, 173, 92, 36, 102, 30, 107, 224, 108, 226, 121, 211, 130, 201, 134, 232, 142, 18, 157, 119, 158, 38, 0, 64, 184, 60, 0, 60, 0, 42, 0, 139, 79, 167, 111, 174, 121, 180, 145, 183, 150, 184, 150, 103, 156, 38, 0, 92, 184, 60, 0, 60, 0, 42, 0, 222, 82, 136, 100, 196, 100, 211, 106, 94, 111, 24, 112, 16, 114, 231, 118, 1, 128, 6, 134, 92, 134, 239, 141, 5, 143, 50, 151, 111, 155, 250, 157, 117, 158, 218, 88, 247, 110, 136, 115, 70, 118, 130, 122, 106, 130, 107, 130, 100, 143, 42, 148, 106, 148, 113, 152, 215, 154, 120, 156, 21, 158, 38, 0, 93, 184, 60, 0, 60, 0, 42, 0, 140, 120, 127, 121, 160, 125, 201, 131, 4, 147, 127, 158, 147, 158, 37, 87, 84, 95, 229, 109, 9, 111, 15, 124, 70, 143, 4, 154, 38, 0, 96, 184, 60, 0, 60, 0, 214, 138, 38, 0, 113, 184, 60, 0, 60, 0, 42, 0, 223, 88, 4, 95, 39, 103, 39, 112, 207, 116, 96, 124, 126, 128, 49, 81, 15, 101, 232, 102, 49, 121, 34, 134, 180, 150, 142, 159, 38, 0, 176, 184, 60, 0, 60, 0, 42, 0, 33, 81, 40, 112, 98, 114, 202, 120, 194, 140, 218, 140, 244, 140, 247, 150, 2, 101, 12, 121, 39, 121, 95, 124, 135, 126, 77, 127, 18, 128, 126, 133, 132, 138, 121, 145, 66, 152, 38, 0, 204, 184, 60, 0, 60, 0, 42, 0, 134, 78, 218, 80, 238, 91, 214, 94, 153, 101, 206, 113, 66, 118, 173, 119, 74, 128, 252, 132, 124, 144, 39, 155, 57, 86, 253, 90, 169, 100, 184, 102, 102, 111, 96, 115, 90, 126, 139, 129, 170, 145, 16, 148, 194, 152, 201, 152, 38, 0, 225, 184, 60, 0, 60, 0, 42, 0, 141, 159, 146, 159, 38, 0, 232, 184, 60, 0, 60, 0, 42, 0, 216, 88, 65, 90, 98, 92, 19, 106, 218, 109, 15, 111, 59, 118, 47, 125, 55, 126, 30, 133, 56, 137, 228, 147, 75, 150, 194, 80, 13, 86, 129, 93, 122, 97, 44, 128, 187, 135, 207, 154, 38, 0, 88, 185, 60, 0, 60, 0, 42, 0, 137, 82, 210, 101, 243, 103, 180, 105, 65, 109, 156, 110, 15, 112, 9, 116, 96, 116, 89, 117, 36, 118, 107, 120, 44, 139, 94, 152, 74, 106, 50, 126, 141, 126, 91, 144, 185, 157, 38, 0, 89, 185, 60, 0, 60, 0, 42, 0, 109, 81, 46, 98, 120, 150, 224, 82, 38, 0, 92, 185, 60, 0, 60, 0, 42, 0, 150, 79, 43, 80, 25, 93, 234, 109, 184, 125, 42, 143, 132, 99, 38, 0, 96, 185, 60, 0, 60, 0, 42, 0, 139, 95, 68, 97, 23, 104, 66, 93, 167, 110, 38, 0, 109, 185, 60, 0, 60, 0, 42, 0, 134, 150, 67, 118, 191, 122, 38, 0, 117, 185, 60, 0, 60, 0, 42, 0, 210, 82, 139, 128, 208, 108, 38, 0, 132, 185, 60, 0, 60, 0, 42, 0, 220, 81, 219, 81, 233, 94, 159, 111, 38, 0, 137, 185, 60, 0, 60, 0, 42, 0, 204, 81, 94, 105, 28, 122, 190, 125, 241, 131, 117, 150, 48, 80, 6, 133, 38, 0, 172, 185, 60, 0, 60, 0, 42, 0, 218, 79, 41, 82, 152, 83, 15, 84, 14, 85, 101, 92, 167, 96, 78, 103, 168, 104, 108, 109, 129, 114, 248, 114, 6, 116, 131, 116, 226, 117, 108, 124, 121, 127, 184, 127, 137, 131, 207, 136, 225, 136, 204, 145, 208, 145, 226, 150, 201, 155, 208, 79, 122, 82, 233, 84, 224, 90, 150, 109, 19, 111, 187, 121, 133, 131, 10, 135, 173, 135, 141, 140, 144, 144, 81, 155, 208, 158, 38, 0, 176, 185, 60, 0, 60, 0, 42, 0, 29, 84, 126, 111, 208, 113, 152, 116, 250, 133, 170, 142, 163, 150, 87, 156, 159, 158, 153, 93, 139, 96, 92, 115, 247, 120, 166, 124, 188, 124, 87, 126, 153, 142, 84, 143, 48, 145, 251, 147, 78, 154, 38, 0, 188, 185, 60, 0, 60, 0, 42, 0, 151, 103, 203, 109, 51, 116, 232, 129, 22, 151, 243, 117, 38, 0, 189, 185, 60, 0, 60, 0, 42, 0, 44, 120, 203, 122, 32, 123, 146, 124, 166, 92, 38, 0, 200, 185, 60, 0, 60, 0, 42, 0, 105, 100, 106, 116, 242, 117, 188, 120, 232, 120, 172, 153, 84, 155, 187, 158, 152, 82, 189, 90, 158, 135, 199, 135, 189, 158, 191, 158, 38, 0, 201, 185, 60, 0, 60, 0, 42, 0, 222, 91, 85, 94, 32, 111, 156, 129, 171, 131, 136, 144, 153, 119, 204, 147, 38, 0, 204, 185, 60, 0, 60, 0, 42, 0, 7, 78, 77, 83, 41, 90, 210, 93, 78, 95, 98, 97, 61, 99, 105, 102, 252, 102, 255, 110, 43, 111, 99, 112, 158, 119, 44, 132, 19, 133, 59, 136, 19, 143, 69, 153, 59, 156, 129, 88, 218, 90, 84, 94, 53, 126, 62, 139, 99, 142, 203, 147, 221, 147, 24, 155, 38, 0, 208, 185, 60, 0, 60, 0, 42, 0, 28, 85, 185, 98, 43, 103, 171, 108, 9, 131, 106, 137, 122, 151, 21, 94, 227, 121, 38, 0, 221, 185, 60, 0, 60, 0, 42, 0, 161, 78, 132, 89, 216, 95, 217, 95, 27, 103, 178, 125, 84, 127, 146, 130, 43, 131, 189, 131, 30, 143, 153, 144, 216, 96, 82, 108, 45, 111, 190, 131, 210, 135, 77, 155, 38, 0, 228, 185, 60, 0, 60, 0, 42, 0, 203, 87, 185, 89, 146, 90, 208, 91, 39, 102, 154, 103, 133, 104, 207, 107, 100, 113, 117, 127, 183, 140, 227, 140, 129, 144, 69, 155, 70, 84, 115, 105, 172, 108, 171, 115, 27, 119, 250, 130, 147, 131, 118, 145, 9, 151, 38, 0, 229, 185, 60, 0, 60, 0, 42, 0, 8, 129, 138, 140, 76, 150, 64, 154, 165, 158, 9, 129, 131, 140, 152, 140, 38, 0, 249, 185, 60, 0, 60, 0, 42, 0, 95, 91, 19, 108, 27, 115, 242, 118, 223, 118, 12, 132, 26, 81, 13, 117, 63, 117, 123, 134, 38, 0, 113, 186, 60, 0, 60, 0, 42, 0, 170, 81, 147, 137, 78, 94, 248, 124, 38, 0, 116, 186, 60, 0, 60, 0, 42, 0, 77, 81, 149, 81, 201, 82, 201, 104, 148, 108, 4, 119, 32, 119, 191, 125, 236, 125, 98, 151, 181, 158, 219, 79, 78, 110, 198, 124, 220, 125, 170, 158, 38, 0, 120, 186, 60, 0, 60, 0, 42, 0, 197, 110, 17, 133, 254, 123, 74, 136, 38, 0, 133, 186, 60, 0, 60, 0, 42, 0, 165, 81, 13, 84, 125, 84, 14, 102, 157, 102, 39, 105, 159, 110, 191, 118, 145, 119, 23, 131, 194, 132, 159, 135, 105, 145, 152, 146, 244, 156, 58, 109, 38, 0, 140, 186, 60, 0, 60, 0, 130, 136, 38, 0, 168, 186, 60, 0, 60, 0, 42, 0, 174, 79, 146, 81, 223, 82, 198, 89, 61, 94, 85, 97, 120, 100, 121, 100, 174, 102, 208, 103, 33, 106, 205, 107, 219, 107, 95, 114, 97, 114, 65, 116, 56, 119, 219, 119, 23, 128, 188, 130, 5, 131, 0, 139, 40, 139, 140, 140, 148, 79, 229, 89, 162, 90, 235, 90, 72, 96, 196, 101, 131, 118, 10, 119, 141, 124, 226, 124, 4, 128, 101, 135, 202, 135, 126, 146, 230, 154, 38, 0, 169, 186, 60, 0, 60, 0, 42, 0, 40, 103, 144, 108, 103, 114, 238, 118, 102, 119, 70, 122, 169, 157, 233, 81, 220, 130, 38, 0, 176, 186, 60, 0, 60, 0, 42, 0, 127, 107, 146, 108, 38, 0, 189, 186, 60, 0, 60, 0, 42, 0, 34, 89, 38, 103, 153, 132, 106, 94, 222, 97, 218, 102, 149, 110, 219, 111, 162, 119, 199, 119, 104, 130, 250, 150, 15, 158, 38, 0, 24, 187, 60, 0, 60, 0, 42, 0, 111, 83, 147, 88, 153, 89, 223, 94, 207, 99, 52, 102, 115, 103, 58, 110, 43, 115, 215, 122, 215, 130, 40, 147, 252, 109, 7, 119, 208, 133, 147, 140, 38, 0, 52, 187, 60, 0, 60, 0, 42, 0, 217, 82, 235, 93, 174, 97, 203, 97, 10, 98, 199, 98, 171, 100, 224, 101, 89, 105, 102, 107, 203, 107, 33, 113, 247, 115, 93, 117, 70, 126, 30, 130, 2, 131, 106, 133, 163, 138, 191, 140, 39, 151, 97, 157, 27, 81, 56, 86, 225, 94, 180, 129, 22, 154, 38, 0, 53, 187, 60, 0, 60, 0, 42, 0, 168, 88, 216, 158, 63, 86, 38, 0, 56, 187, 60, 0, 60, 0, 42, 0, 17, 80, 14, 82, 59, 84, 79, 85, 135, 101, 118, 108, 10, 125, 11, 125, 94, 128, 138, 134, 128, 149, 239, 150, 1, 83, 151, 96, 227, 97, 134, 98, 107, 99, 134, 112, 138, 116, 38, 0, 60, 187, 60, 0, 60, 0, 42, 0, 255, 82, 149, 108, 105, 114, 38, 0, 248, 187, 60, 0, 60, 0, 42, 0, 115, 84, 154, 90, 62, 92, 75, 93, 76, 95, 174, 95, 42, 103, 182, 104, 99, 105, 60, 110, 68, 110, 9, 119, 115, 124, 142, 127, 135, 133, 14, 139, 247, 143, 97, 151, 244, 158, 185, 78, 19, 90, 132, 90, 186, 90, 37, 95, 45, 95, 73, 101, 48, 112, 124, 115, 220, 124, 59, 126, 255, 130, 60, 134, 139, 158, 38, 0, 252, 187, 60, 0, 60, 0, 42, 0, 183, 92, 182, 96, 13, 97, 171, 97, 79, 101, 251, 101, 252, 101, 17, 108, 239, 108, 159, 115, 201, 115, 225, 125, 148, 149, 222, 95, 223, 95, 139, 102, 99, 110, 205, 125, 96, 127, 224, 130, 169, 149, 53, 156, 253, 158, 38, 0, 0, 188, 60, 0, 60, 0, 42, 0, 198, 91, 28, 135, 16, 139, 18, 106, 245, 110, 38, 0, 21, 188, 60, 0, 60, 0, 42, 0, 93, 82, 90, 83, 205, 98, 15, 100, 178, 100, 52, 103, 56, 106, 202, 108, 192, 115, 158, 116, 148, 123, 149, 124, 27, 126, 138, 129, 54, 130, 132, 133, 235, 143, 249, 150, 193, 153, 179, 78, 2, 107, 84, 114, 155, 147, 238, 153, 198, 154, 38, 0, 24, 188, 60, 0, 60, 0, 42, 0, 52, 79, 74, 83, 205, 83, 219, 83, 204, 98, 44, 100, 0, 101, 145, 101, 195, 105, 238, 108, 88, 111, 237, 115, 84, 117, 34, 118, 228, 118, 252, 118, 208, 120, 251, 120, 44, 121, 70, 125, 44, 130, 224, 135, 212, 143, 18, 152, 239, 152, 187, 90, 115, 98, 43, 100, 61, 101, 12, 103, 214, 128, 140, 135, 22, 152, 38, 0, 28, 188, 60, 0, 60, 0, 42, 0, 195, 82, 212, 98, 165, 100, 36, 110, 81, 111, 124, 118, 203, 141, 177, 145, 98, 146, 238, 154, 67, 155, 241, 84, 97, 109, 22, 129, 56, 146, 83, 157, 38, 0, 41, 188, 60, 0, 60, 0, 42, 0, 35, 80, 141, 80, 74, 87, 168, 89, 40, 92, 71, 94, 119, 95, 63, 98, 62, 101, 185, 101, 193, 101, 9, 102, 139, 103, 156, 105, 194, 110, 197, 120, 33, 125, 170, 128, 128, 129, 43, 130, 179, 130, 161, 132, 140, 134, 42, 138, 23, 139, 166, 144, 50, 150, 144, 159, 255, 78, 150, 83, 107, 94, 172, 95, 18, 100, 202, 101, 134, 104, 83, 114, 61, 130, 131, 135, 138, 147, 227, 154, 116, 155, 38, 0, 48, 188, 60, 0, 60, 0, 42, 0, 13, 80, 243, 79, 249, 87, 152, 95, 220, 98, 146, 99, 111, 103, 67, 110, 25, 113, 195, 118, 204, 128, 218, 128, 244, 136, 245, 136, 25, 137, 224, 140, 41, 143, 77, 145, 106, 150, 79, 87, 111, 87, 82, 98, 50, 116, 211, 132, 38, 0, 49, 188, 60, 0, 60, 0, 42, 0, 47, 79, 112, 79, 27, 94, 207, 103, 34, 104, 125, 118, 126, 118, 68, 155, 168, 124, 38, 0, 136, 188, 60, 0, 60, 0, 42, 0, 97, 94, 10, 106, 105, 113, 212, 113, 106, 117, 65, 126, 67, 133, 233, 133, 220, 152, 89, 126, 251, 127, 176, 129, 41, 134, 162, 136, 38, 0, 140, 188, 60, 0, 60, 0, 42, 0, 16, 79, 79, 123, 112, 127, 165, 149, 67, 106, 120, 127, 38, 0, 148, 188, 60, 0, 60, 0, 42, 0, 225, 81, 6, 94, 181, 104, 62, 108, 78, 108, 219, 108, 175, 114, 196, 123, 3, 131, 53, 123, 9, 138, 191, 152, 38, 0, 149, 188, 60, 0, 60, 0, 42, 0, 213, 108, 58, 116, 38, 0, 189, 188, 60, 0, 60, 0, 42, 0, 251, 80, 136, 82, 193, 88, 216, 100, 151, 106, 167, 116, 86, 118, 167, 120, 23, 134, 226, 149, 57, 151, 215, 100, 19, 117, 136, 117, 94, 137, 255, 157, 10, 159, 38, 0, 192, 188, 60, 0, 60, 0, 42, 0, 94, 83, 1, 95, 138, 139, 168, 143, 175, 143, 138, 144, 237, 95, 131, 98, 105, 124, 174, 143, 65, 129, 198, 140, 226, 153, 191, 154, 24, 157, 38, 0, 196, 188, 60, 0, 60, 0, 42, 0, 37, 82, 165, 119, 73, 156, 8, 159, 70, 95, 233, 157, 38, 0, 209, 188, 60, 0, 60, 0, 42, 0, 25, 78, 2, 80, 117, 81, 91, 92, 119, 94, 30, 102, 58, 102, 196, 103, 197, 104, 179, 112, 1, 117, 197, 117, 201, 121, 221, 122, 39, 143, 32, 153, 8, 154, 38, 78, 64, 88, 99, 125, 62, 127, 248, 143, 53, 146, 242, 146, 124, 146, 38, 0, 244, 188, 60, 0, 60, 0, 42, 0, 221, 79, 33, 88, 49, 88, 246, 91, 110, 102, 101, 107, 17, 109, 122, 110, 125, 111, 228, 115, 43, 117, 233, 131, 220, 136, 19, 137, 92, 139, 20, 143, 204, 79, 217, 118, 32, 124, 70, 132, 76, 151, 7, 157, 252, 158, 38, 0, 245, 188, 60, 0, 60, 0, 42, 0, 15, 79, 213, 80, 16, 83, 92, 83, 147, 91, 169, 95, 13, 103, 143, 121, 121, 129, 47, 131, 20, 133, 7, 137, 134, 137, 57, 143, 59, 143, 165, 153, 18, 156, 163, 88, 94, 94, 81, 98, 238, 111, 153, 123, 212, 131, 96, 135, 110, 135, 105, 157, 38, 0, 248, 188, 60, 0, 60, 0, 44, 103, 38, 0, 252, 188, 60, 0, 60, 0, 118, 78, 38, 0, 9, 189, 60, 0, 60, 0, 42, 0, 248, 79, 73, 89, 1, 92, 239, 92, 240, 92, 103, 99, 210, 104, 253, 112, 162, 113, 43, 116, 43, 126, 236, 132, 2, 135, 34, 144, 210, 146, 243, 156, 48, 78, 6, 89, 247, 123, 152, 125, 246, 131, 12, 157, 38, 0, 128, 189, 60, 0, 60, 0, 42, 0, 13, 78, 216, 78, 239, 79, 133, 80, 86, 82, 111, 82, 38, 84, 144, 84, 224, 87, 43, 89, 102, 90, 90, 91, 117, 91, 204, 91, 156, 94, 118, 98, 119, 101, 167, 101, 110, 109, 165, 110, 54, 114, 38, 123, 63, 124, 54, 127, 80, 129, 81, 129, 154, 129, 64, 130, 153, 130, 169, 131, 3, 138, 160, 140, 230, 140, 251, 140, 116, 141, 186, 141, 232, 144, 220, 145, 28, 150, 68, 150, 217, 153, 231, 156, 198, 78, 216, 79, 141, 90, 148, 98, 202, 98, 138, 99, 116, 104, 145, 105, 170, 109, 158, 115, 84, 121, 95, 123, 88, 127, 102, 127, 213, 128, 163, 130, 251, 130, 0, 133, 168, 134, 9, 135, 157, 136, 210, 136, 215, 141, 7, 146, 43, 152, 146, 155, 169, 158, 38, 0, 129, 189, 60, 0, 60, 0, 23, 83, 38, 0, 132, 189, 60, 0, 60, 0, 42, 0, 6, 82, 41, 84, 116, 86, 179, 88, 84, 89, 110, 89, 255, 95, 164, 97, 110, 98, 16, 102, 126, 108, 26, 113, 198, 118, 137, 124, 222, 124, 27, 125, 172, 130, 193, 140, 240, 150, 83, 79, 76, 87, 9, 94, 140, 103, 251, 104, 252, 104, 27, 108, 83, 110, 198, 111, 135, 114, 90, 117, 15, 120, 40, 123, 166, 128, 185, 129, 97, 133, 82, 143, 250, 158, 34, 159, 38, 0, 136, 189, 60, 0, 60, 0, 42, 0, 91, 79, 23, 95, 127, 95, 194, 98, 170, 92, 83, 121, 49, 125, 116, 130, 0, 131, 205, 151, 244, 154, 251, 158, 38, 0, 149, 189, 60, 0, 60, 0, 42, 0, 41, 93, 11, 103, 218, 104, 124, 120, 67, 126, 108, 157, 11, 88, 48, 111, 5, 155, 38, 0, 68, 190, 60, 0, 60, 0, 42, 0, 21, 78, 153, 80, 21, 83, 42, 83, 81, 83, 131, 89, 98, 90, 135, 94, 178, 96, 138, 97, 73, 98, 121, 98, 144, 101, 135, 103, 167, 105, 212, 107, 214, 107, 215, 107, 216, 107, 184, 108, 53, 116, 250, 117, 18, 120, 145, 120, 213, 121, 216, 121, 131, 124, 203, 125, 225, 127, 165, 128, 62, 129, 194, 129, 242, 131, 26, 135, 232, 136, 185, 138, 108, 139, 187, 140, 25, 145, 94, 151, 219, 152, 59, 159, 243, 78, 254, 79, 85, 82, 46, 87, 228, 87, 163, 89, 65, 92, 179, 94, 177, 96, 208, 104, 17, 105, 152, 108, 221, 109, 224, 109, 222, 111, 201, 114, 210, 114, 222, 117, 249, 117, 101, 119, 85, 121, 230, 123, 21, 125, 134, 127, 83, 129, 152, 130, 190, 130, 6, 132, 214, 132, 141, 134, 148, 140, 20, 141, 97, 143, 179, 144, 235, 144, 159, 149, 116, 150, 15, 151, 180, 151, 17, 154, 27, 154, 192, 154, 25, 159, 38, 0, 72, 190, 60, 0, 60, 0, 42, 0, 172, 86, 42, 91, 108, 95, 140, 101, 179, 106, 175, 107, 92, 109, 241, 111, 21, 112, 93, 114, 173, 115, 167, 140, 211, 140, 59, 152, 16, 81, 239, 100, 201, 119, 125, 126, 207, 129, 11, 134, 115, 140, 160, 144, 76, 148, 38, 151, 112, 152, 2, 155, 34, 155, 38, 0, 89, 190, 60, 0, 60, 0, 42, 0, 145, 97, 55, 108, 88, 128, 1, 154, 176, 81, 237, 81, 244, 81, 9, 90, 38, 0, 172, 192, 60, 0, 60, 0, 42, 0, 77, 78, 139, 78, 155, 78, 213, 78, 58, 79, 60, 79, 127, 79, 223, 79, 255, 80, 242, 83, 248, 83, 6, 85, 227, 85, 219, 86, 235, 88, 98, 89, 17, 90, 235, 91, 250, 91, 4, 92, 243, 93, 43, 94, 153, 95, 29, 96, 104, 99, 156, 101, 175, 101, 246, 103, 251, 103, 173, 104, 123, 107, 153, 108, 215, 108, 35, 110, 9, 112, 69, 115, 2, 120, 62, 121, 64, 121, 96, 121, 193, 121, 233, 123, 23, 125, 114, 125, 134, 128, 13, 130, 142, 131, 209, 132, 199, 134, 223, 136, 80, 138, 94, 138, 29, 139, 220, 140, 102, 141, 173, 143, 170, 144, 252, 152, 223, 153, 157, 158, 158, 80, 90, 82, 120, 83, 139, 84, 210, 89, 66, 105, 173, 105, 92, 108, 231, 117, 182, 118, 226, 122, 37, 123, 55, 127, 33, 135, 151, 137, 219, 153, 102, 155, 202, 155, 36, 156, 38, 0, 173, 192, 60, 0, 60, 0, 42, 0, 74, 82, 20, 103, 202, 105, 13, 114, 180, 132, 96, 148, 38, 0, 176, 192, 60, 0, 60, 0, 42, 0, 152, 80, 42, 82, 113, 92, 99, 101, 85, 108, 202, 115, 35, 117, 157, 117, 151, 123, 156, 132, 120, 145, 48, 151, 119, 82, 205, 89, 127, 91, 117, 106, 120, 111, 152, 111, 251, 114, 86, 126, 21, 138, 223, 147, 138, 149, 255, 107, 38, 0, 180, 192, 60, 0, 60, 0, 42, 0, 119, 78, 146, 100, 186, 107, 94, 113, 169, 133, 38, 0, 188, 192, 60, 0, 60, 0, 42, 0, 9, 78, 73, 103, 238, 104, 23, 110, 159, 130, 24, 133, 107, 136, 221, 124, 228, 145, 22, 155, 38, 0, 189, 192, 60, 0, 60, 0, 42, 0, 247, 99, 129, 111, 18, 146, 175, 152, 69, 83, 60, 85, 67, 107, 227, 127, 100, 147, 5, 151, 14, 151, 38, 0, 193, 192, 60, 0, 60, 0, 42, 0, 10, 78, 183, 80, 207, 80, 31, 81, 70, 85, 170, 85, 23, 86, 64, 91, 25, 92, 224, 92, 56, 94, 138, 94, 160, 94, 194, 94, 243, 96, 81, 104, 97, 106, 88, 110, 61, 114, 64, 114, 192, 114, 248, 118, 101, 121, 177, 123, 212, 127, 243, 136, 244, 137, 115, 138, 97, 140, 222, 140, 28, 151, 125, 88, 156, 95, 76, 102, 164, 107, 30, 117, 215, 125, 159, 147, 89, 152, 58, 155, 38, 0, 200, 192, 60, 0, 60, 0, 42, 0, 94, 88, 189, 116, 253, 140, 19, 156, 38, 0, 201, 192, 60, 0, 60, 0, 42, 0, 199, 85, 97, 122, 34, 125, 114, 130, 237, 105, 199, 111, 18, 112, 38, 0, 221, 192, 60, 0, 60, 0, 42, 0, 114, 114, 31, 117, 37, 117, 25, 123, 26, 119, 78, 146, 38, 0, 28, 193, 60, 0, 60, 0, 42, 0, 133, 88, 251, 88, 188, 93, 143, 94, 182, 94, 144, 95, 85, 96, 146, 98, 127, 99, 77, 101, 145, 102, 217, 102, 248, 102, 22, 104, 242, 104, 128, 114, 94, 116, 110, 123, 110, 125, 214, 125, 114, 127, 229, 128, 18, 130, 175, 133, 127, 137, 147, 138, 29, 144, 228, 146, 205, 158, 32, 159, 108, 86, 127, 90, 223, 99, 149, 100, 81, 110, 168, 111, 19, 125, 33, 128, 167, 130, 79, 146, 38, 0, 29, 193, 60, 0, 60, 0, 42, 0, 21, 89, 109, 89, 45, 94, 220, 96, 20, 102, 115, 102, 144, 103, 80, 108, 197, 109, 95, 111, 243, 119, 169, 120, 198, 132, 203, 145, 43, 147, 112, 102, 253, 119, 74, 129, 3, 130, 37, 135, 80, 146, 43, 159, 38, 0, 32, 193, 60, 0, 60, 0, 42, 0, 217, 78, 202, 80, 72, 81, 132, 85, 11, 91, 163, 91, 71, 98, 126, 101, 203, 101, 50, 110, 125, 113, 1, 116, 68, 116, 135, 116, 191, 116, 108, 118, 170, 121, 218, 125, 85, 126, 168, 127, 122, 129, 179, 129, 57, 130, 26, 134, 236, 135, 117, 138, 227, 141, 120, 144, 145, 146, 37, 148, 77, 153, 174, 155, 161, 88, 217, 90, 31, 92, 32, 92, 115, 92, 3, 97, 90, 107, 175, 113, 69, 123, 171, 125, 84, 139, 113, 139, 199, 147, 56, 154, 83, 156, 123, 156, 38, 0, 36, 193, 60, 0, 60, 0, 42, 0, 104, 83, 81, 92, 84, 105, 196, 108, 41, 109, 43, 110, 12, 130, 155, 133, 59, 137, 45, 138, 170, 138, 234, 150, 103, 159, 112, 80, 159, 90, 242, 99, 172, 102, 7, 114, 159, 120, 39, 122, 50, 125, 38, 0, 44, 193, 60, 0, 60, 0, 42, 0, 97, 82, 185, 102, 178, 107, 150, 126, 254, 135, 13, 141, 131, 149, 93, 150, 69, 91, 184, 97, 123, 100, 82, 119, 107, 139, 155, 146, 241, 151, 38, 0, 45, 193, 60, 0, 60, 0, 42, 0, 29, 101, 137, 109, 238, 113, 193, 86, 254, 97, 68, 112, 118, 128, 161, 142, 119, 148, 115, 152, 38, 0, 49, 193, 60, 0, 60, 0, 42, 0, 206, 87, 211, 89, 172, 91, 39, 96, 250, 96, 16, 98, 31, 102, 95, 102, 41, 115, 249, 115, 219, 118, 1, 119, 108, 123, 86, 128, 114, 128, 101, 129, 160, 138, 146, 145, 70, 116, 2, 154, 38, 0, 56, 193, 60, 0, 60, 0, 42, 0, 22, 78, 226, 82, 114, 107, 23, 109, 5, 122, 57, 123, 48, 125, 176, 140, 40, 94, 18, 109, 80, 126, 251, 134, 38, 0, 140, 193, 60, 0, 60, 0, 42, 0, 236, 83, 47, 86, 81, 88, 181, 91, 15, 92, 17, 92, 226, 93, 64, 98, 131, 99, 20, 100, 45, 102, 179, 104, 188, 108, 136, 109, 175, 110, 31, 112, 164, 112, 210, 113, 38, 117, 143, 117, 142, 117, 25, 118, 17, 123, 224, 123, 43, 124, 32, 125, 57, 125, 44, 133, 109, 133, 7, 134, 52, 138, 13, 144, 97, 144, 181, 144, 183, 146, 246, 151, 55, 154, 75, 79, 207, 79, 114, 83, 201, 85, 253, 87, 80, 88, 44, 97, 78, 99, 20, 106, 221, 108, 113, 123, 190, 123, 69, 126, 219, 127, 134, 129, 88, 130, 248, 134, 8, 142, 101, 145, 4, 151, 72, 155, 185, 155, 58, 156, 38, 0, 141, 193, 60, 0, 60, 0, 42, 0, 215, 79, 108, 92, 95, 103, 145, 109, 159, 124, 140, 126, 22, 139, 22, 141, 31, 144, 44, 109, 108, 144, 38, 0, 144, 193, 60, 0, 60, 0, 42, 0, 107, 91, 253, 93, 13, 100, 192, 132, 92, 144, 225, 152, 231, 152, 241, 152, 38, 0, 148, 193, 60, 0, 60, 0, 42, 0, 135, 115, 163, 122, 192, 135, 38, 0, 161, 193, 60, 0, 60, 0, 42, 0, 139, 91, 154, 96, 126, 103, 222, 109, 31, 138, 166, 138, 1, 144, 12, 152, 215, 103, 230, 122, 6, 155, 38, 0, 196, 193, 60, 0, 60, 0, 42, 0, 55, 82, 81, 112, 142, 120, 150, 147, 226, 96, 236, 102, 99, 116, 38, 0, 224, 193, 60, 0, 60, 0, 42, 0, 112, 136, 215, 145, 38, 0, 24, 194, 60, 0, 60, 0, 42, 0, 238, 79, 215, 83, 253, 85, 218, 86, 130, 87, 253, 88, 194, 90, 136, 91, 171, 92, 192, 92, 37, 94, 1, 97, 13, 98, 75, 98, 136, 99, 28, 100, 54, 101, 120, 101, 57, 106, 138, 107, 52, 108, 25, 109, 49, 111, 231, 113, 233, 114, 120, 115, 7, 116, 178, 116, 38, 118, 97, 119, 192, 121, 87, 122, 234, 122, 185, 124, 143, 125, 172, 125, 97, 126, 158, 127, 41, 129, 49, 131, 144, 132, 218, 132, 234, 133, 150, 136, 176, 138, 144, 139, 56, 143, 66, 144, 131, 144, 108, 145, 150, 146, 185, 146, 139, 150, 167, 150, 168, 150, 214, 150, 0, 151, 8, 152, 150, 153, 211, 154, 26, 155, 223, 83, 46, 85, 203, 94, 108, 102, 179, 107, 197, 108, 178, 110, 201, 111, 95, 119, 98, 119, 141, 119, 95, 121, 84, 124, 58, 129, 132, 129, 184, 129, 142, 139, 78, 140, 114, 150, 188, 152, 72, 153, 38, 0, 25, 194, 60, 0, 60, 0, 42, 0, 212, 83, 126, 88, 25, 89, 112, 91, 191, 91, 209, 109, 90, 111, 159, 113, 33, 116, 185, 116, 133, 128, 253, 131, 246, 79, 15, 80, 53, 81, 76, 90, 90, 106, 76, 154, 235, 157, 38, 0, 28, 194, 60, 0, 60, 0, 42, 0, 225, 93, 135, 95, 170, 95, 66, 96, 236, 101, 18, 104, 111, 105, 83, 106, 137, 107, 53, 109, 243, 109, 227, 115, 254, 118, 172, 119, 77, 123, 20, 125, 35, 129, 28, 130, 64, 131, 244, 132, 99, 133, 98, 138, 196, 138, 135, 145, 30, 147, 6, 152, 180, 153, 154, 79, 229, 114, 249, 118, 52, 119, 3, 125, 171, 128, 232, 153, 10, 155, 137, 157, 38, 0, 32, 194, 60, 0, 60, 0, 42, 0, 12, 98, 83, 136, 240, 143, 101, 146, 73, 125, 38, 0, 45, 194, 60, 0, 60, 0, 42, 0, 7, 93, 39, 93, 105, 93, 216, 131, 38, 0, 108, 194, 60, 0, 60, 0, 42, 0, 5, 80, 236, 109, 32, 113, 38, 0, 172, 194, 60, 0, 60, 0, 42, 0, 95, 116, 157, 129, 104, 135, 113, 134, 38, 0, 181, 194, 60, 0, 60, 0, 42, 0, 213, 111, 254, 98, 210, 127, 54, 137, 114, 137, 116, 97, 160, 113, 176, 150, 38, 0, 185, 194, 60, 0, 60, 0, 42, 0, 30, 78, 88, 78, 231, 80, 221, 82, 71, 83, 127, 98, 7, 102, 105, 126, 5, 136, 94, 150, 77, 88, 25, 155, 38, 0, 220, 194, 60, 0, 60, 0, 42, 0, 141, 79, 25, 83, 54, 86, 203, 89, 164, 90, 56, 92, 78, 92, 77, 92, 2, 94, 17, 95, 67, 96, 189, 101, 47, 102, 66, 102, 190, 103, 244, 103, 28, 115, 226, 119, 58, 121, 197, 127, 148, 132, 205, 132, 150, 137, 102, 138, 105, 138, 225, 138, 85, 140, 122, 140, 114, 80, 85, 81, 174, 83, 123, 85, 82, 88, 221, 94, 178, 103, 249, 103, 140, 111, 230, 125, 228, 127, 223, 138, 240, 138, 73, 140, 195, 145, 73, 147, 75, 152, 38, 0, 221, 194, 60, 0, 60, 0, 42, 0, 244, 87, 212, 91, 15, 95, 111, 96, 237, 98, 13, 105, 150, 107, 92, 110, 132, 113, 210, 123, 85, 135, 88, 139, 254, 142, 223, 152, 254, 152, 176, 85, 179, 90, 59, 104, 38, 0, 224, 194, 60, 0, 60, 0, 42, 0, 56, 79, 129, 79, 225, 79, 123, 84, 32, 90, 184, 91, 60, 97, 176, 101, 104, 102, 252, 113, 51, 117, 94, 121, 51, 125, 78, 129, 227, 129, 152, 131, 170, 133, 206, 133, 3, 135, 10, 138, 171, 142, 155, 143, 197, 143, 194, 84, 122, 86, 223, 86, 250, 89, 91, 108, 231, 119, 36, 129, 16, 141, 35, 152, 234, 153, 38, 0, 228, 194, 60, 0, 60, 0, 42, 0, 49, 89, 164, 91, 230, 91, 137, 96, 203, 135, 203, 152, 38, 0, 236, 194, 60, 0, 60, 0, 42, 0, 233, 91, 11, 92, 195, 95, 129, 108, 241, 109, 11, 112, 26, 117, 175, 130, 246, 138, 163, 104, 111, 111, 214, 113, 90, 132, 20, 148, 79, 156, 38, 0, 237, 194, 60, 0, 60, 0, 42, 0, 192, 78, 65, 83, 187, 143, 38, 0, 13, 195, 60, 0, 60, 0, 217, 150, 38, 0, 40, 197, 60, 0, 60, 0, 15, 108, 38, 0, 68, 197, 60, 0, 60, 0, 42, 0, 158, 78, 196, 79, 82, 81, 94, 85, 37, 90, 232, 92, 17, 98, 89, 114, 189, 130, 170, 131, 254, 134, 89, 136, 29, 138, 63, 150, 197, 150, 19, 153, 9, 157, 93, 157, 43, 78, 230, 84, 63, 90, 64, 90, 233, 92, 180, 117, 17, 120, 12, 123, 211, 143, 15, 147, 94, 157, 38, 0, 69, 197, 60, 0, 60, 0, 42, 0, 10, 88, 179, 92, 189, 93, 68, 94, 225, 96, 21, 97, 225, 99, 2, 106, 37, 110, 2, 145, 84, 147, 78, 152, 16, 156, 119, 159, 83, 80, 126, 83, 162, 84, 148, 85, 105, 86, 109, 129, 60, 132, 168, 137, 228, 138, 154, 157, 118, 159, 38, 0, 72, 197, 60, 0, 60, 0, 42, 0, 137, 91, 184, 92, 9, 99, 79, 102, 72, 104, 60, 119, 193, 150, 141, 151, 84, 152, 159, 155, 73, 104, 180, 114, 11, 141, 8, 157, 38, 0, 76, 197, 60, 0, 60, 0, 42, 0, 161, 101, 1, 139, 203, 142, 188, 149, 14, 86, 30, 98, 224, 99, 117, 122, 16, 138, 79, 144, 30, 152, 54, 157, 38, 0, 84, 197, 60, 0, 60, 0, 42, 0, 53, 85, 169, 92, 214, 93, 181, 94, 151, 102, 76, 118, 244, 131, 199, 149, 125, 85, 149, 90, 83, 93, 123, 102, 100, 129, 74, 132, 237, 132, 243, 138, 55, 152, 163, 153, 239, 158, 38, 0, 85, 197, 60, 0, 60, 0, 42, 0, 211, 88, 188, 98, 206, 114, 40, 157, 38, 0, 89, 197, 60, 0, 60, 0, 42, 0, 240, 78, 46, 89, 15, 96, 59, 102, 131, 107, 231, 121, 38, 157, 108, 83, 113, 87, 241, 108, 206, 118, 133, 151, 38, 0, 96, 197, 60, 0, 60, 0, 42, 0, 147, 83, 192, 84, 195, 87, 22, 93, 27, 97, 214, 102, 175, 109, 141, 120, 126, 130, 152, 150, 68, 151, 254, 80, 9, 85, 64, 85, 111, 86, 45, 90, 21, 93, 40, 99, 113, 99, 56, 107, 4, 111, 67, 115, 154, 118, 90, 119, 185, 119, 209, 120, 25, 121, 134, 133, 249, 133, 73, 151, 3, 154, 38, 0, 97, 197, 60, 0, 60, 0, 42, 0, 132, 83, 124, 98, 150, 99, 178, 109, 10, 126, 75, 129, 77, 152, 93, 84, 57, 98, 36, 100, 40, 150, 38, 0, 117, 197, 60, 0, 60, 0, 42, 0, 251, 106, 76, 127, 175, 157, 26, 158, 182, 86, 200, 90, 67, 127, 234, 157, 38, 0, 124, 197, 60, 0, 60, 0, 42, 0, 95, 78, 59, 80, 182, 81, 28, 89, 249, 96, 246, 99, 48, 105, 58, 114, 54, 128, 206, 145, 220, 87, 38, 0, 125, 197, 60, 0, 60, 0, 42, 0, 49, 95, 4, 125, 229, 130, 111, 132, 187, 132, 229, 133, 141, 142, 26, 114, 180, 121, 219, 123, 101, 124, 112, 148, 47, 156, 184, 157, 160, 159, 38, 0, 145, 197, 60, 0, 60, 0, 42, 0, 111, 79, 228, 88, 67, 91, 89, 96, 218, 99, 24, 101, 109, 101, 152, 102, 74, 105, 35, 106, 11, 109, 1, 112, 108, 113, 210, 117, 13, 118, 179, 121, 112, 122, 138, 127, 68, 137, 147, 139, 192, 145, 125, 150, 10, 153, 137, 95, 62, 111, 60, 112, 202, 112, 98, 118, 59, 119, 24, 134, 48, 143, 114, 148, 186, 152, 100, 154, 38, 0, 180, 197, 60, 0, 60, 0, 42, 0, 4, 87, 161, 95, 188, 101, 1, 111, 0, 118, 166, 121, 158, 138, 173, 153, 90, 155, 108, 159, 9, 87, 84, 101, 228, 109, 235, 152, 38, 0, 181, 197, 60, 0, 60, 0, 42, 0, 4, 81, 182, 97, 145, 98, 141, 106, 198, 129, 118, 126, 38, 0, 184, 197, 60, 0, 60, 0, 42, 0, 67, 80, 48, 88, 102, 95, 9, 113, 0, 138, 250, 138, 191, 80, 61, 83, 227, 90, 158, 139, 34, 145, 52, 159, 57, 159, 38, 0, 188, 197, 60, 0, 60, 0, 42, 0, 124, 91, 22, 134, 236, 129, 38, 0, 196, 197, 60, 0, 60, 0, 42, 0, 250, 79, 60, 81, 180, 86, 68, 89, 169, 99, 249, 109, 38, 93, 127, 94, 238, 102, 104, 127, 131, 145, 185, 149, 38, 0, 197, 197, 60, 0, 60, 0, 42, 0, 170, 93, 109, 105, 171, 93, 52, 145, 38, 0, 208, 197, 60, 0, 60, 0, 42, 0, 90, 96, 192, 102, 38, 0, 212, 197, 60, 0, 60, 0, 134, 81, 38, 0, 236, 197, 60, 0, 60, 0, 42, 0, 136, 78, 89, 79, 130, 89, 95, 107, 93, 108, 181, 116, 22, 121, 7, 130, 69, 130, 57, 131, 63, 143, 93, 143, 24, 153, 1, 130, 38, 0, 237, 197, 60, 0, 60, 0, 42, 0, 166, 78, 223, 87, 121, 95, 19, 102, 171, 117, 121, 126, 111, 139, 6, 144, 91, 154, 167, 93, 204, 97, 226, 109, 190, 149, 38, 0, 240, 197, 60, 0, 60, 0, 42, 0, 165, 86, 39, 88, 248, 89, 31, 90, 180, 91, 246, 94, 80, 99, 59, 99, 61, 105, 135, 108, 191, 108, 142, 109, 147, 109, 245, 109, 20, 111, 223, 112, 54, 113, 89, 113, 195, 113, 213, 113, 79, 120, 111, 120, 117, 123, 227, 125, 47, 126, 77, 136, 223, 142, 91, 146, 246, 156, 87, 81, 230, 86, 207, 87, 63, 91, 129, 96, 190, 99, 227, 102, 222, 106, 55, 110, 217, 129, 154, 131, 53, 135, 21, 136, 140, 139, 240, 157, 38, 0, 244, 197, 60, 0, 60, 0, 42, 0, 133, 96, 133, 109, 177, 113, 177, 149, 78, 86, 38, 0, 252, 197, 60, 0, 60, 0, 42, 0, 173, 83, 211, 103, 142, 112, 48, 113, 48, 116, 118, 130, 210, 130, 187, 149, 229, 154, 125, 158, 137, 81, 105, 88, 213, 97, 74, 98, 191, 106, 182, 106, 78, 112, 105, 112, 197, 145, 92, 153, 88, 155, 246, 158, 38, 0, 253, 197, 60, 0, 60, 0, 42, 0, 196, 102, 193, 113, 73, 132, 197, 102, 128, 113, 23, 114, 104, 151, 38, 0, 1, 198, 60, 0, 60, 0, 42, 0, 75, 88, 184, 93, 113, 95, 32, 102, 142, 102, 121, 105, 174, 105, 56, 108, 243, 108, 54, 110, 65, 111, 218, 111, 27, 112, 47, 112, 80, 113, 223, 113, 112, 115, 91, 116, 212, 116, 200, 118, 78, 122, 147, 126, 241, 130, 96, 138, 206, 143, 72, 147, 25, 151, 143, 84, 52, 91, 48, 91, 103, 109, 244, 111, 109, 118, 164, 120, 8, 126, 17, 136, 15, 141, 226, 144, 250, 151, 38, 0, 8, 198, 60, 0, 60, 0, 42, 0, 66, 78, 42, 80, 8, 82, 225, 83, 243, 102, 109, 108, 202, 111, 10, 115, 127, 119, 98, 122, 174, 130, 221, 133, 2, 134, 212, 136, 99, 138, 125, 139, 107, 140, 179, 146, 19, 151, 16, 152, 200, 86, 219, 90, 253, 98, 156, 99, 152, 103, 105, 115, 104, 119, 150, 119, 68, 126, 243, 127, 197, 130, 74, 133, 75, 133, 137, 133, 139, 134, 58, 135, 226, 155, 214, 157, 145, 158, 38, 0, 36, 198, 60, 0, 60, 0, 42, 0, 148, 78, 13, 79, 201, 79, 178, 80, 72, 83, 62, 84, 51, 84, 218, 85, 98, 88, 186, 88, 103, 89, 27, 90, 228, 91, 159, 96, 202, 97, 86, 101, 255, 101, 100, 102, 167, 104, 90, 108, 179, 111, 207, 112, 172, 113, 82, 115, 125, 123, 8, 135, 164, 138, 50, 156, 7, 159, 245, 78, 227, 79, 20, 85, 247, 85, 65, 86, 44, 87, 170, 90, 239, 90, 228, 95, 96, 97, 66, 99, 89, 108, 185, 122, 113, 128, 35, 131, 86, 137, 55, 139, 195, 143, 213, 143, 104, 144, 202, 147, 214, 147, 169, 150, 65, 154, 47, 159, 38, 0, 37, 198, 60, 0, 60, 0, 42, 0, 75, 92, 131, 108, 68, 115, 137, 115, 58, 146, 38, 0, 40, 198, 60, 0, 60, 0, 42, 0, 171, 110, 101, 116, 31, 118, 105, 122, 21, 126, 10, 134, 188, 90, 77, 97, 55, 102, 51, 108, 133, 113, 128, 133, 64, 143, 158, 145, 222, 151, 66, 153, 46, 156, 38, 0, 44, 198, 60, 0, 60, 0, 42, 0, 64, 81, 226, 85, 131, 129, 38, 0, 57, 198, 60, 0, 60, 0, 42, 0, 197, 88, 193, 100, 238, 116, 21, 117, 112, 118, 193, 127, 149, 144, 205, 150, 84, 153, 129, 85, 241, 94, 195, 110, 85, 118, 186, 121, 75, 127, 202, 132, 221, 150, 82, 152, 38, 0, 64, 198, 60, 0, 60, 0, 42, 0, 38, 110, 230, 116, 169, 122, 170, 122, 229, 129, 217, 134, 120, 135, 27, 138, 199, 84, 238, 86, 80, 90, 153, 103, 60, 109, 39, 115, 138, 122, 53, 132, 76, 139, 38, 0, 68, 198, 60, 0, 60, 0, 42, 0, 73, 90, 140, 91, 155, 91, 161, 104, 0, 105, 99, 109, 169, 115, 19, 116, 44, 116, 151, 120, 233, 125, 235, 127, 24, 129, 85, 129, 158, 131, 76, 140, 46, 150, 17, 152, 19, 82, 184, 87, 167, 89, 143, 92, 232, 95, 203, 96, 180, 109, 204, 118, 38, 0, 72, 198, 60, 0, 60, 0, 240, 102, 38, 0, 85, 198, 60, 0, 60, 0, 42, 0, 128, 95, 250, 101, 137, 103, 106, 108, 139, 115, 43, 92, 7, 112, 236, 143, 38, 0, 92, 198, 60, 0, 60, 0, 42, 0, 45, 80, 3, 90, 106, 107, 238, 119, 167, 90, 38, 0, 120, 198, 60, 0, 60, 0, 42, 0, 22, 89, 108, 93, 205, 93, 37, 115, 79, 117, 78, 80, 52, 93, 84, 93, 40, 110, 104, 113, 168, 120, 200, 120, 117, 128, 151, 150, 38, 0, 148, 198, 60, 0, 60, 0, 42, 0, 229, 80, 249, 81, 47, 88, 45, 89, 150, 89, 218, 89, 229, 91, 162, 93, 215, 98, 22, 100, 147, 100, 254, 100, 220, 102, 72, 106, 255, 113, 100, 116, 136, 122, 175, 122, 71, 126, 94, 126, 0, 128, 112, 129, 239, 135, 129, 137, 32, 139, 89, 144, 128, 144, 82, 153, 96, 80, 147, 85, 115, 87, 157, 88, 8, 91, 122, 94, 173, 95, 188, 95, 128, 107, 134, 111, 69, 121, 126, 122, 133, 122, 88, 133, 118, 144, 194, 157, 38, 0, 149, 198, 60, 0, 60, 0, 42, 0, 126, 97, 50, 107, 116, 109, 31, 126, 37, 137, 177, 143, 189, 110, 208, 132, 38, 0, 169, 198, 60, 0, 60, 0, 42, 0, 209, 79, 173, 80, 151, 81, 199, 82, 199, 87, 137, 88, 185, 91, 184, 94, 66, 97, 149, 105, 140, 109, 103, 110, 182, 110, 148, 113, 98, 116, 40, 117, 44, 117, 115, 128, 56, 131, 201, 132, 10, 142, 148, 147, 222, 147, 155, 80, 130, 91, 113, 93, 117, 97, 131, 97, 230, 105, 103, 120, 2, 130, 249, 134, 52, 142, 38, 0, 176, 198, 60, 0, 60, 0, 42, 0, 142, 78, 81, 79, 118, 80, 42, 81, 200, 83, 203, 83, 243, 83, 135, 91, 211, 91, 36, 92, 26, 97, 130, 97, 244, 101, 91, 114, 151, 115, 64, 116, 194, 118, 80, 121, 145, 121, 185, 121, 6, 125, 189, 127, 139, 130, 213, 133, 94, 134, 194, 143, 71, 144, 245, 144, 234, 145, 133, 150, 232, 150, 233, 150, 143, 78, 180, 78, 193, 79, 74, 80, 1, 84, 35, 88, 78, 93, 189, 94, 69, 103, 163, 117, 241, 118, 253, 122, 38, 128, 48, 128, 35, 139, 61, 142, 93, 147, 128, 158, 140, 158, 114, 159, 38, 0, 177, 198, 60, 0, 60, 0, 42, 0, 214, 82, 103, 95, 237, 101, 49, 102, 47, 104, 92, 113, 54, 122, 193, 144, 10, 152, 224, 113, 38, 0, 180, 198, 60, 0, 60, 0, 42, 0, 145, 78, 82, 106, 158, 107, 144, 111, 137, 113, 24, 128, 184, 130, 83, 133, 75, 144, 149, 150, 242, 150, 251, 151, 242, 96, 132, 108, 212, 123, 28, 125, 35, 151, 245, 151, 38, 0, 184, 198, 60, 0, 60, 0, 42, 0, 26, 133, 49, 155, 144, 78, 38, 0, 197, 198, 60, 0, 60, 0, 42, 0, 138, 113, 196, 150, 38, 0, 208, 198, 60, 0, 60, 0, 42, 0, 67, 81, 159, 83, 225, 84, 19, 87, 18, 87, 163, 87, 155, 90, 196, 90, 195, 91, 40, 96, 63, 97, 244, 99, 133, 108, 57, 109, 114, 110, 144, 110, 48, 114, 63, 115, 87, 116, 209, 130, 129, 136, 69, 143, 96, 144, 98, 150, 88, 152, 27, 157, 164, 81, 28, 87, 108, 103, 101, 105, 40, 115, 169, 125, 171, 130, 151, 133, 63, 135, 28, 139, 250, 146, 53, 154, 119, 157, 255, 158, 38, 0, 212, 198, 60, 0, 60, 0, 42, 0, 8, 103, 138, 141, 94, 146, 22, 82, 164, 124, 38, 0, 4, 199, 60, 0, 60, 0, 42, 0, 77, 79, 73, 80, 222, 80, 113, 83, 13, 87, 212, 89, 1, 90, 9, 92, 112, 97, 144, 102, 45, 110, 50, 114, 75, 116, 239, 125, 195, 128, 14, 132, 102, 132, 63, 133, 95, 135, 91, 136, 24, 137, 2, 139, 85, 144, 203, 151, 79, 155, 159, 85, 67, 94, 82, 113, 168, 113, 255, 117, 115, 132, 94, 136, 201, 138, 54, 144, 200, 149, 217, 151, 225, 151, 39, 153, 170, 154, 38, 0, 32, 199, 60, 0, 60, 0, 42, 0, 115, 78, 145, 79, 18, 81, 106, 81, 47, 85, 169, 85, 122, 91, 165, 91, 124, 94, 125, 94, 190, 94, 160, 96, 223, 96, 8, 97, 9, 97, 196, 99, 56, 101, 9, 103, 212, 103, 218, 103, 97, 105, 98, 105, 185, 108, 39, 109, 56, 110, 225, 111, 54, 115, 55, 115, 92, 116, 49, 117, 82, 118, 173, 125, 254, 129, 56, 132, 213, 136, 152, 138, 219, 138, 237, 138, 48, 142, 66, 142, 74, 144, 62, 144, 122, 144, 73, 145, 201, 145, 110, 147, 152, 81, 102, 84, 255, 86, 221, 88, 55, 94, 201, 99, 191, 101, 209, 108, 86, 114, 9, 118, 16, 118, 172, 122, 179, 122, 114, 124, 197, 124, 204, 125, 116, 129, 160, 131, 85, 133, 100, 133, 176, 134, 180, 134, 100, 135, 21, 137, 137, 139, 12, 144, 163, 151, 170, 155, 221, 158, 44, 159, 165, 159, 38, 0, 33, 199, 60, 0, 60, 0, 42, 0, 9, 88, 211, 107, 137, 128, 178, 128, 37, 81, 38, 0, 36, 199, 60, 0, 60, 0, 42, 0, 65, 81, 107, 89, 57, 92, 100, 111, 167, 115, 228, 128, 7, 141, 23, 146, 143, 149, 0, 102, 198, 146, 38, 0, 40, 199, 60, 0, 60, 0, 42, 0, 127, 128, 79, 111, 222, 119, 38, 0, 53, 199, 60, 0, 60, 0, 42, 0, 14, 98, 28, 112, 104, 125, 141, 135, 232, 114, 38, 0, 64, 199, 60, 0, 60, 0, 42, 0, 160, 87, 105, 96, 71, 97, 183, 107, 190, 138, 128, 146, 177, 150, 44, 84, 154, 86, 1, 87, 189, 87, 150, 97, 188, 106, 181, 110, 250, 114, 226, 115, 110, 118, 20, 138, 30, 145, 87, 159, 38, 0, 68, 199, 60, 0, 60, 0, 42, 0, 89, 78, 230, 156, 38, 0, 76, 199, 60, 0, 60, 0, 42, 0, 31, 84, 235, 109, 45, 133, 112, 150, 243, 151, 238, 152, 145, 85, 31, 93, 213, 94, 20, 97, 42, 151, 38, 0, 77, 199, 60, 0, 60, 0, 42, 0, 214, 99, 227, 108, 145, 144, 146, 96, 57, 99, 101, 109, 38, 0, 81, 199, 60, 0, 60, 0, 42, 0, 221, 81, 201, 97, 186, 129, 249, 157, 38, 0, 88, 199, 60, 0, 60, 0, 42, 0, 157, 79, 26, 80, 0, 81, 156, 91, 15, 97, 255, 97, 236, 100, 5, 105, 197, 107, 145, 117, 227, 119, 169, 127, 100, 130, 143, 133, 251, 135, 99, 136, 188, 138, 112, 139, 171, 145, 23, 81, 210, 81, 147, 82, 183, 93, 57, 107, 42, 111, 23, 115, 18, 121, 152, 135, 80, 153, 38, 0, 116, 199, 60, 0, 60, 0, 42, 0, 140, 78, 229, 78, 10, 79, 55, 89, 232, 89, 242, 93, 27, 95, 91, 95, 33, 96, 62, 114, 229, 115, 112, 117, 205, 117, 251, 121, 12, 128, 51, 128, 132, 128, 225, 130, 81, 131, 189, 140, 179, 140, 135, 144, 244, 152, 12, 153, 191, 84, 104, 87, 20, 92, 93, 95, 46, 104, 31, 109, 198, 115, 17, 138, 81, 138, 228, 143, 182, 150, 38, 0, 117, 199, 60, 0, 60, 0, 42, 0, 55, 112, 202, 118, 202, 127, 204, 127, 252, 127, 26, 139, 11, 95, 164, 113, 193, 157, 38, 0, 120, 199, 60, 0, 60, 0, 42, 0, 186, 78, 193, 78, 3, 82, 112, 83, 189, 84, 224, 86, 251, 89, 197, 91, 21, 95, 205, 95, 110, 110, 106, 125, 53, 131, 147, 134, 141, 138, 109, 151, 119, 151, 222, 78, 25, 88, 36, 89, 99, 90, 104, 98, 36, 108, 7, 109, 139, 121, 126, 124, 162, 130, 192, 136, 38, 0, 124, 199, 60, 0, 60, 0, 42, 0, 0, 78, 90, 79, 126, 79, 249, 88, 229, 101, 162, 110, 56, 144, 176, 147, 185, 153, 198, 108, 252, 142, 38, 0, 132, 199, 60, 0, 60, 0, 42, 0, 251, 78, 236, 88, 138, 89, 217, 89, 65, 96, 20, 122, 79, 131, 195, 140, 77, 125, 125, 136, 139, 146, 234, 152, 38, 0, 133, 199, 60, 0, 60, 0, 42, 0, 101, 81, 68, 83, 255, 94, 38, 0, 137, 199, 60, 0, 60, 0, 42, 0, 205, 78, 105, 82, 85, 91, 191, 130, 181, 90, 38, 0, 144, 199, 60, 0, 60, 0, 42, 0, 212, 78, 58, 82, 168, 84, 201, 89, 255, 89, 80, 91, 87, 91, 92, 91, 99, 96, 72, 97, 203, 110, 153, 112, 110, 113, 134, 115, 247, 116, 181, 117, 193, 120, 43, 125, 5, 128, 234, 129, 40, 131, 23, 133, 201, 133, 238, 138, 199, 140, 204, 150, 112, 84, 40, 91, 86, 91, 118, 91, 216, 103, 218, 108, 120, 114, 37, 119, 38, 119, 162, 124, 20, 128, 254, 128, 8, 131, 50, 131, 191, 131, 120, 134, 220, 137, 62, 138, 178, 140, 109, 141, 161, 147, 63, 152, 237, 154, 147, 155, 191, 157, 211, 157, 38, 0, 145, 199, 60, 0, 60, 0, 42, 0, 92, 79, 250, 82, 188, 86, 171, 101, 40, 102, 124, 112, 184, 112, 53, 114, 189, 125, 141, 130, 76, 145, 192, 150, 114, 157, 157, 92, 13, 96, 177, 101, 222, 103, 75, 108, 47, 113, 179, 114, 143, 120, 38, 0, 148, 199, 60, 0, 60, 0, 42, 0, 113, 91, 231, 104, 152, 107, 122, 111, 222, 118, 87, 82, 20, 98, 79, 154, 38, 0, 160, 199, 60, 0, 60, 0, 42, 0, 145, 92, 171, 102, 91, 111, 180, 123, 42, 124, 54, 136, 148, 109, 92, 111, 243, 111, 38, 0, 161, 199, 60, 0, 60, 0, 42, 0, 220, 150, 97, 83, 195, 86, 40, 119, 252, 120, 77, 137, 38, 0, 165, 199, 60, 0, 60, 0, 42, 0, 8, 78, 215, 78, 32, 83, 52, 88, 187, 88, 239, 88, 108, 89, 7, 92, 51, 94, 132, 94, 53, 95, 140, 99, 178, 102, 86, 103, 31, 106, 163, 106, 12, 107, 63, 111, 70, 114, 80, 115, 139, 116, 224, 122, 167, 124, 120, 129, 223, 129, 231, 129, 138, 131, 108, 132, 35, 133, 148, 133, 207, 133, 221, 136, 19, 141, 172, 145, 119, 149, 156, 150, 189, 80, 88, 89, 157, 89, 25, 91, 130, 93, 231, 94, 21, 98, 51, 111, 66, 114, 52, 118, 218, 124, 152, 127, 7, 132, 197, 136, 236, 140, 35, 145, 216, 147, 38, 153, 158, 158, 38, 0, 172, 199, 60, 0, 60, 0, 42, 0, 141, 81, 201, 84, 40, 87, 176, 91, 77, 98, 80, 103, 61, 104, 147, 104, 61, 110, 211, 110, 125, 112, 33, 126, 193, 136, 161, 140, 9, 143, 75, 159, 78, 159, 8, 89, 61, 93, 87, 98, 159, 105, 126, 112, 148, 126, 38, 0, 193, 199, 60, 0, 60, 0, 42, 0, 45, 114, 143, 123, 205, 138, 26, 147, 34, 93, 25, 115, 36, 116, 151, 147, 38, 0, 0, 200, 60, 0, 60, 0, 42, 0, 71, 79, 78, 79, 50, 81, 128, 84, 208, 89, 149, 94, 181, 98, 117, 103, 110, 105, 23, 106, 174, 108, 26, 110, 217, 114, 42, 115, 189, 117, 184, 123, 53, 125, 231, 130, 249, 131, 87, 132, 247, 133, 91, 138, 175, 140, 135, 142, 25, 144, 184, 144, 206, 150, 95, 159, 129, 91, 168, 92, 124, 103, 226, 103, 16, 108, 116, 111, 38, 112, 116, 114, 93, 127, 157, 127, 244, 130, 198, 134, 155, 136, 26, 137, 221, 137, 70, 138, 108, 140, 124, 150, 38, 0, 1, 200, 60, 0, 60, 0, 42, 0, 227, 82, 10, 84, 225, 90, 194, 91, 88, 100, 117, 101, 244, 110, 196, 114, 132, 118, 77, 122, 27, 123, 77, 124, 62, 126, 223, 127, 123, 131, 43, 139, 202, 140, 100, 141, 225, 141, 95, 142, 234, 143, 249, 143, 105, 144, 209, 147, 0, 106, 231, 120, 244, 124, 194, 131, 191, 137, 22, 144, 176, 153, 38, 0, 4, 200, 60, 0, 60, 0, 42, 0, 67, 79, 122, 79, 179, 80, 104, 81, 120, 81, 77, 82, 106, 82, 97, 88, 124, 88, 96, 89, 8, 92, 85, 92, 219, 94, 155, 96, 48, 98, 19, 104, 191, 107, 8, 108, 177, 111, 78, 113, 32, 116, 48, 117, 56, 117, 81, 117, 114, 118, 76, 123, 139, 123, 173, 123, 198, 123, 143, 126, 110, 138, 62, 143, 73, 143, 63, 146, 147, 146, 34, 147, 43, 148, 251, 150, 90, 152, 107, 152, 30, 153, 46, 84, 192, 86, 229, 90, 71, 92, 211, 93, 41, 98, 195, 99, 195, 101, 52, 104, 199, 105, 84, 110, 182, 111, 75, 114, 113, 116, 14, 117, 75, 117, 96, 117, 202, 117, 92, 118, 218, 120, 91, 124, 182, 127, 230, 127, 70, 129, 158, 129, 67, 131, 148, 142, 7, 143, 133, 144, 61, 145, 209, 146, 42, 147, 91, 151, 102, 151, 83, 152, 230, 152, 48, 153, 11, 155, 99, 156, 7, 158, 38, 0, 8, 200, 60, 0, 60, 0, 42, 0, 7, 82, 42, 98, 152, 98, 89, 109, 100, 118, 202, 122, 192, 123, 118, 125, 138, 92, 98, 102, 131, 122, 38, 0, 16, 200, 60, 0, 60, 0, 42, 0, 96, 83, 190, 92, 151, 94, 56, 111, 185, 112, 152, 124, 17, 151, 142, 155, 222, 158, 84, 79, 138, 88, 183, 115, 24, 123, 31, 124, 235, 130, 42, 133, 197, 134, 152, 137, 173, 152, 207, 158, 38, 0, 17, 200, 60, 0, 60, 0, 42, 0, 165, 99, 122, 100, 118, 135, 4, 105, 106, 105, 40, 135, 213, 141, 64, 142, 8, 156, 38, 0, 21, 200, 60, 0, 60, 0, 42, 0, 1, 78, 149, 78, 173, 78, 92, 80, 117, 80, 72, 84, 195, 89, 154, 91, 64, 94, 173, 94, 247, 94, 129, 95, 197, 96, 58, 99, 63, 101, 116, 101, 204, 101, 118, 102, 120, 102, 254, 103, 104, 105, 137, 106, 99, 107, 64, 108, 192, 109, 232, 109, 31, 110, 94, 110, 30, 112, 161, 112, 142, 115, 253, 115, 58, 117, 91, 119, 135, 120, 142, 121, 11, 122, 125, 122, 190, 124, 142, 125, 71, 130, 2, 138, 234, 138, 158, 140, 45, 145, 74, 145, 216, 145, 102, 146, 204, 146, 32, 147, 6, 151, 86, 151, 92, 151, 2, 152, 14, 159, 66, 79, 238, 83, 103, 90, 119, 90, 20, 96, 159, 99, 111, 104, 131, 104, 214, 104, 111, 112, 245, 115, 148, 117, 115, 123, 155, 131, 60, 138, 73, 144, 114, 145, 229, 146, 90, 151, 38, 0, 28, 200, 60, 0, 60, 0, 42, 0, 54, 82, 145, 82, 124, 85, 36, 88, 29, 94, 31, 95, 140, 96, 208, 99, 175, 104, 223, 111, 109, 121, 44, 123, 205, 129, 186, 133, 253, 136, 248, 138, 68, 142, 141, 145, 100, 150, 155, 150, 61, 151, 76, 152, 74, 159, 21, 81, 35, 90, 224, 100, 24, 115, 69, 116, 71, 119, 148, 121, 10, 122, 249, 125, 54, 142, 79, 142, 139, 142, 87, 147, 132, 150, 242, 151, 167, 155, 247, 155, 38, 0, 112, 200, 60, 0, 60, 0, 42, 0, 206, 79, 70, 81, 203, 81, 169, 82, 50, 86, 20, 95, 107, 95, 170, 99, 205, 100, 233, 101, 65, 102, 250, 102, 249, 102, 29, 103, 157, 104, 215, 104, 253, 105, 21, 111, 110, 111, 103, 113, 229, 113, 42, 114, 170, 116, 58, 119, 86, 121, 90, 121, 223, 121, 32, 122, 149, 122, 151, 124, 223, 124, 68, 125, 112, 126, 135, 128, 251, 133, 164, 134, 84, 138, 191, 138, 153, 141, 129, 142, 32, 144, 109, 144, 227, 145, 59, 150, 213, 150, 229, 156, 123, 79, 174, 80, 1, 82, 157, 83, 8, 86, 106, 86, 37, 91, 130, 95, 198, 97, 126, 98, 130, 107, 161, 111, 49, 116, 129, 118, 103, 121, 200, 122, 10, 123, 217, 124, 246, 124, 105, 125, 91, 125, 217, 128, 202, 129, 90, 130, 38, 133, 41, 135, 130, 138, 95, 139, 31, 146, 154, 146, 253, 146, 219, 155, 112, 157, 2, 159, 38, 0, 113, 200, 60, 0, 60, 0, 42, 0, 207, 101, 7, 124, 179, 141, 195, 147, 47, 118, 38, 0, 116, 200, 60, 0, 60, 0, 42, 0, 88, 91, 10, 92, 245, 98, 38, 0, 120, 200, 60, 0, 60, 0, 42, 0, 82, 83, 217, 98, 29, 115, 38, 0, 133, 200, 60, 0, 60, 0, 42, 0, 39, 80, 151, 91, 158, 95, 176, 96, 107, 97, 213, 104, 217, 109, 46, 116, 46, 122, 66, 125, 156, 125, 49, 126, 107, 129, 42, 142, 53, 142, 126, 147, 24, 148, 0, 79, 82, 97, 202, 103, 54, 105, 5, 106, 125, 116, 7, 118, 189, 124, 189, 135, 100, 142, 38, 0, 140, 200, 60, 0, 60, 0, 42, 0, 80, 79, 80, 87, 230, 93, 167, 94, 43, 99, 73, 82, 228, 117, 157, 131, 253, 154, 38, 0, 196, 200, 60, 0, 60, 0, 106, 127, 38, 0, 252, 200, 60, 0, 60, 0, 42, 0, 59, 78, 79, 79, 143, 79, 90, 80, 221, 89, 196, 128, 106, 84, 104, 84, 254, 85, 79, 89, 153, 91, 222, 93, 218, 94, 93, 102, 49, 103, 241, 103, 42, 104, 232, 108, 50, 109, 74, 110, 141, 111, 183, 112, 224, 115, 135, 117, 76, 124, 2, 125, 44, 125, 162, 125, 31, 130, 219, 134, 59, 138, 133, 138, 112, 141, 138, 142, 51, 143, 49, 144, 78, 145, 82, 145, 68, 148, 208, 153, 31, 78, 156, 79, 20, 81, 12, 92, 108, 94, 196, 98, 67, 120, 82, 124, 152, 128, 96, 129, 31, 133, 192, 134, 239, 136, 75, 138, 217, 140, 142, 141, 8, 143, 82, 144, 82, 146, 12, 151, 20, 151, 4, 159, 38, 0, 253, 200, 60, 0, 60, 0, 42, 0, 249, 122, 165, 124, 38, 0, 0, 201, 60, 0, 60, 0, 42, 0, 202, 79, 1, 81, 198, 81, 200, 87, 239, 91, 251, 92, 89, 102, 61, 106, 90, 109, 150, 110, 236, 111, 12, 113, 111, 117, 227, 122, 34, 136, 33, 144, 117, 144, 203, 150, 255, 153, 66, 86, 251, 87, 171, 88, 247, 96, 153, 100, 180, 118, 167, 125, 71, 127, 6, 142, 114, 142, 15, 148, 188, 150, 21, 153, 82, 156, 84, 157, 38, 0, 4, 201, 60, 0, 60, 0, 42, 0, 1, 131, 124, 78, 38, 0, 17, 201, 60, 0, 60, 0, 42, 0, 45, 78, 242, 78, 70, 136, 205, 145, 62, 119, 38, 0, 137, 201, 60, 0, 60, 0, 42, 0, 125, 83, 115, 83, 158, 85, 38, 0, 144, 201, 60, 0, 60, 0, 42, 0, 219, 106, 45, 154, 38, 0, 153, 201, 60, 0, 60, 0, 42, 0, 107, 105, 65, 108, 122, 132, 157, 106, 122, 133, 38, 0, 157, 201, 60, 0, 60, 0, 42, 0, 158, 88, 142, 97, 254, 102, 239, 98, 221, 112, 17, 117, 199, 117, 82, 126, 184, 132, 73, 139, 8, 141, 146, 93, 240, 119, 126, 127, 38, 0, 192, 201, 60, 0, 60, 0, 42, 0, 75, 78, 234, 83, 171, 84, 48, 87, 64, 87, 215, 95, 1, 99, 7, 99, 111, 100, 47, 101, 232, 101, 122, 102, 157, 103, 179, 103, 98, 107, 96, 108, 154, 108, 44, 111, 229, 119, 37, 120, 73, 121, 87, 121, 25, 125, 162, 128, 2, 129, 243, 129, 157, 130, 183, 130, 24, 135, 140, 138, 4, 141, 190, 141, 114, 144, 123, 87, 128, 88, 122, 98, 176, 105, 220, 108, 227, 117, 234, 121, 234, 123, 16, 130, 31, 142, 147, 142, 249, 142, 47, 150, 168, 155, 217, 157, 38, 0, 193, 201, 60, 0, 60, 0, 42, 0, 244, 118, 25, 122, 55, 122, 84, 126, 119, 128, 157, 121, 38, 0, 196, 201, 60, 0, 60, 0, 42, 0, 7, 85, 212, 85, 117, 88, 47, 99, 34, 100, 73, 102, 75, 102, 109, 104, 155, 105, 132, 107, 37, 109, 177, 110, 205, 115, 104, 116, 161, 116, 91, 117, 185, 117, 225, 118, 30, 119, 139, 119, 230, 121, 9, 126, 29, 126, 251, 129, 47, 133, 151, 136, 58, 138, 209, 140, 235, 142, 176, 143, 50, 144, 173, 147, 99, 150, 115, 150, 7, 151, 178, 79, 24, 81, 210, 115, 57, 122, 193, 132, 180, 135, 129, 141, 65, 146, 18, 155, 38, 0, 200, 201, 60, 0, 60, 0, 42, 0, 132, 79, 241, 83, 234, 89, 201, 90, 25, 94, 78, 104, 198, 116, 190, 117, 233, 121, 146, 122, 163, 129, 237, 134, 234, 140, 204, 141, 237, 143, 164, 87, 112, 125, 186, 132, 197, 144, 85, 148, 38, 0, 208, 201, 60, 0, 60, 0, 42, 0, 159, 101, 21, 103, 6, 157, 38, 0, 209, 201, 60, 0, 60, 0, 42, 0, 247, 87, 87, 111, 221, 125, 47, 143, 246, 147, 198, 150, 160, 84, 34, 98, 38, 0, 213, 201, 60, 0, 60, 0, 42, 0, 181, 95, 242, 97, 132, 111, 130, 111, 19, 112, 101, 118, 170, 119, 38, 0, 40, 204, 60, 0, 60, 0, 42, 0, 20, 78, 152, 79, 31, 80, 201, 83, 223, 85, 111, 93, 238, 93, 33, 107, 100, 107, 203, 120, 154, 123, 73, 142, 202, 142, 110, 144, 125, 79, 86, 80, 83, 89, 148, 92, 163, 95, 206, 105, 115, 116, 104, 120, 38, 0, 41, 204, 60, 0, 60, 0, 42, 0, 73, 99, 62, 100, 64, 119, 132, 122, 47, 147, 127, 148, 106, 159, 51, 98, 201, 100, 178, 101, 38, 0, 44, 204, 60, 0, 60, 0, 42, 0, 176, 100, 175, 111, 230, 113, 168, 116, 218, 116, 196, 122, 18, 124, 130, 126, 178, 124, 152, 126, 154, 139, 10, 141, 125, 148, 16, 153, 76, 153, 39, 81, 57, 81, 151, 82, 209, 93, 34, 101, 17, 107, 40, 114, 178, 141, 38, 0, 48, 204, 60, 0, 60, 0, 42, 0, 57, 82, 223, 91, 230, 100, 45, 103, 46, 125, 78, 98, 246, 98, 38, 0, 56, 204, 60, 0, 60, 0, 42, 0, 237, 80, 195, 83, 121, 88, 88, 97, 89, 97, 250, 97, 172, 101, 217, 122, 146, 139, 150, 139, 51, 81, 197, 83, 132, 93, 201, 93, 90, 97, 175, 97, 25, 101, 231, 105, 3, 107, 218, 107, 86, 139, 232, 147, 113, 148, 94, 153, 66, 154, 242, 158, 38, 0, 61, 204, 60, 0, 60, 0, 42, 0, 9, 80, 33, 80, 117, 82, 49, 85, 60, 90, 224, 94, 112, 95, 52, 97, 94, 101, 12, 102, 54, 102, 162, 102, 205, 105, 196, 110, 50, 111, 22, 115, 33, 118, 147, 122, 57, 129, 89, 130, 214, 131, 188, 132, 0, 80, 150, 80, 212, 81, 49, 82, 181, 96, 221, 96, 39, 98, 54, 100, 25, 105, 5, 108, 114, 116, 151, 122, 187, 122, 76, 142, 249, 146, 6, 147, 182, 149, 47, 155, 172, 157, 38, 0, 68, 204, 60, 0, 60, 0, 42, 0, 181, 80, 240, 87, 192, 91, 232, 91, 105, 95, 161, 99, 38, 120, 181, 125, 220, 131, 33, 133, 199, 145, 245, 145, 204, 104, 29, 131, 38, 0, 69, 204, 60, 0, 60, 0, 42, 0, 138, 81, 245, 103, 86, 123, 172, 140, 22, 86, 88, 94, 212, 120, 39, 123, 0, 124, 177, 134, 38, 0, 152, 204, 60, 0, 60, 0, 42, 0, 196, 81, 187, 89, 189, 96, 85, 134, 210, 109, 11, 132, 4, 137, 183, 137, 234, 144, 38, 0, 153, 204, 60, 0, 60, 0, 42, 0, 28, 80, 84, 82, 58, 92, 125, 97, 26, 98, 211, 98, 242, 100, 165, 101, 204, 110, 32, 118, 10, 129, 96, 142, 95, 150, 187, 150, 78, 84, 103, 87, 73, 88, 213, 96, 87, 99, 109, 100, 52, 135, 214, 141, 145, 142, 38, 0, 156, 204, 60, 0, 60, 0, 42, 0, 223, 78, 67, 83, 152, 85, 41, 89, 221, 93, 197, 100, 201, 108, 250, 109, 148, 115, 127, 122, 27, 130, 166, 133, 228, 140, 16, 142, 119, 144, 231, 145, 225, 149, 33, 150, 198, 151, 244, 79, 41, 80, 226, 80, 3, 81, 10, 109, 250, 111, 233, 116, 70, 121, 129, 124, 246, 129, 138, 130, 28, 131, 80, 131, 168, 132, 70, 133, 149, 134, 191, 143, 93, 151, 38, 0, 160, 204, 60, 0, 60, 0, 42, 0, 248, 81, 242, 84, 134, 85, 185, 95, 164, 100, 136, 111, 180, 125, 31, 143, 77, 143, 53, 148, 95, 82, 92, 85, 209, 87, 217, 96, 135, 99, 96, 107, 149, 146, 35, 147, 251, 152, 46, 153, 38, 0, 168, 204, 60, 0, 60, 0, 42, 0, 201, 80, 22, 92, 190, 108, 251, 109, 27, 117, 187, 119, 61, 124, 100, 124, 121, 138, 194, 138, 104, 94, 221, 95, 201, 96, 144, 106, 252, 106, 56, 112, 55, 124, 92, 137, 38, 0, 169, 204, 60, 0, 60, 0, 42, 0, 30, 88, 190, 89, 22, 94, 119, 99, 82, 114, 138, 117, 107, 119, 220, 138, 188, 140, 18, 143, 34, 80, 107, 84, 139, 85, 23, 96, 58, 137, 38, 0, 173, 204, 60, 0, 60, 0, 42, 0, 243, 94, 116, 102, 248, 109, 125, 128, 193, 131, 203, 138, 81, 151, 214, 155, 202, 81, 10, 87, 59, 135, 132, 157, 38, 0, 180, 204, 60, 0, 60, 0, 42, 0, 67, 82, 255, 102, 149, 109, 239, 110, 224, 125, 230, 138, 46, 144, 94, 144, 212, 154, 143, 86, 88, 95, 227, 104, 162, 107, 12, 120, 130, 132, 21, 133, 94, 133, 198, 142, 70, 151, 240, 154, 38, 0, 8, 205, 60, 0, 60, 0, 42, 0, 29, 82, 127, 82, 232, 84, 148, 97, 132, 98, 219, 98, 162, 104, 18, 105, 90, 105, 53, 106, 146, 112, 38, 113, 93, 120, 1, 121, 14, 121, 210, 121, 13, 122, 150, 128, 120, 130, 213, 130, 73, 131, 73, 133, 130, 140, 133, 141, 98, 145, 139, 145, 174, 145, 98, 80, 236, 80, 173, 82, 230, 82, 77, 86, 246, 90, 167, 92, 237, 92, 149, 93, 10, 96, 132, 96, 0, 97, 106, 103, 203, 113, 131, 125, 22, 128, 154, 138, 89, 139, 160, 141, 250, 142, 226, 143, 20, 146, 107, 147, 108, 147, 152, 151, 102, 152, 235, 154, 230, 157, 96, 159, 38, 0, 9, 205, 60, 0, 60, 0, 42, 0, 195, 79, 209, 86, 237, 113, 215, 119, 0, 135, 248, 137, 239, 102, 37, 114, 218, 119, 165, 133, 133, 142, 209, 154, 38, 0, 12, 205, 60, 0, 60, 0, 42, 0, 248, 91, 214, 95, 81, 103, 168, 144, 11, 84, 38, 0, 29, 205, 60, 0, 60, 0, 42, 0, 226, 83, 90, 88, 245, 91, 164, 96, 129, 97, 96, 100, 61, 126, 112, 128, 37, 133, 131, 146, 113, 132, 239, 132, 230, 147, 24, 154, 68, 154, 38, 0, 44, 205, 60, 0, 60, 0, 174, 100, 38, 0, 92, 205, 60, 0, 60, 0, 42, 0, 172, 80, 20, 93, 0, 103, 44, 86, 103, 100, 177, 105, 60, 111, 128, 116, 234, 120, 23, 126, 39, 129, 38, 0, 148, 205, 60, 0, 60, 0, 42, 0, 156, 88, 189, 98, 168, 99, 14, 105, 120, 105, 30, 106, 107, 110, 186, 118, 203, 121, 187, 130, 41, 132, 207, 138, 168, 141, 253, 143, 18, 145, 75, 145, 156, 145, 16, 147, 24, 147, 154, 147, 219, 150, 54, 154, 13, 156, 230, 80, 126, 85, 53, 90, 26, 94, 198, 96, 118, 99, 235, 99, 37, 100, 3, 117, 51, 118, 157, 159, 146, 123, 160, 123, 9, 124, 11, 126, 16, 126, 173, 132, 108, 150, 185, 150, 166, 151, 5, 154, 75, 155, 12, 156, 123, 157, 150, 157, 181, 157, 132, 158, 164, 158, 38, 0, 149, 205, 60, 0, 60, 0, 42, 0, 17, 78, 92, 117, 93, 121, 250, 122, 81, 123, 201, 123, 46, 126, 196, 132, 89, 142, 116, 142, 248, 142, 16, 144, 175, 89, 51, 130, 86, 140, 92, 142, 0, 159, 38, 0, 152, 205, 60, 0, 60, 0, 42, 0, 37, 102, 63, 105, 67, 116, 38, 0, 156, 205, 60, 0, 60, 0, 42, 0, 250, 81, 46, 103, 220, 158, 235, 121, 38, 0, 169, 205, 60, 0, 60, 0, 42, 0, 69, 81, 224, 95, 150, 108, 242, 135, 93, 136, 119, 136, 178, 81, 225, 95, 235, 115, 38, 0, 204, 205, 60, 0, 60, 0, 42, 0, 180, 96, 181, 129, 3, 132, 5, 141, 244, 96, 227, 99, 169, 117, 1, 118, 71, 152, 38, 0, 232, 205, 60, 0, 60, 0, 42, 0, 214, 83, 57, 84, 52, 86, 54, 90, 49, 92, 138, 112, 224, 127, 90, 128, 6, 129, 237, 129, 163, 141, 137, 145, 95, 154, 242, 157, 163, 81, 71, 106, 243, 107, 38, 0, 33, 206, 60, 0, 60, 0, 42, 0, 116, 80, 196, 78, 160, 83, 251, 96, 44, 110, 193, 94, 3, 102, 38, 0, 53, 206, 60, 0, 60, 0, 100, 92, 38, 0, 88, 206, 60, 0, 60, 0, 42, 0, 136, 79, 36, 80, 228, 85, 217, 92, 95, 94, 101, 96, 148, 104, 187, 108, 196, 109, 190, 113, 212, 117, 244, 117, 97, 118, 26, 122, 73, 122, 199, 125, 251, 125, 110, 127, 244, 129, 169, 134, 28, 143, 201, 150, 179, 153, 82, 159, 110, 83, 198, 84, 216, 91, 245, 93, 100, 117, 211, 117, 122, 125, 209, 131, 153, 133, 43, 137, 120, 140, 241, 141, 25, 147, 36, 150, 212, 155, 25, 157, 31, 157, 68, 157, 38, 0, 89, 206, 60, 0, 60, 0, 42, 0, 71, 82, 197, 82, 237, 152, 85, 101, 38, 0, 92, 206, 60, 0, 60, 0, 42, 0, 170, 137, 236, 106, 253, 133, 111, 137, 84, 159, 38, 0, 96, 206, 60, 0, 60, 0, 42, 0, 3, 78, 210, 103, 6, 111, 38, 0, 104, 206, 60, 0, 60, 0, 42, 0, 181, 79, 226, 91, 149, 103, 136, 108, 120, 109, 27, 116, 39, 120, 221, 145, 124, 147, 214, 91, 241, 95, 57, 105, 137, 108, 244, 144, 223, 146, 248, 153, 38, 0, 105, 206, 60, 0, 60, 0, 196, 135, 38, 0, 109, 206, 60, 0, 60, 0, 42, 0, 228, 121, 49, 122, 38, 0, 140, 207, 60, 0, 60, 0, 42, 0, 235, 95, 114, 86, 44, 89, 38, 0, 192, 208, 60, 0, 60, 0, 42, 0, 214, 78, 164, 84, 62, 85, 174, 88, 165, 89, 240, 96, 83, 98, 214, 98, 54, 103, 85, 105, 53, 130, 64, 150, 177, 153, 221, 153, 87, 79, 158, 87, 213, 98, 193, 103, 98, 106, 177, 108, 107, 138, 206, 141, 177, 142, 222, 153, 128, 155, 21, 157, 9, 159, 38, 0, 193, 208, 60, 0, 60, 0, 42, 0, 44, 80, 83, 83, 68, 85, 124, 87, 88, 98, 226, 100, 107, 102, 221, 103, 193, 111, 239, 111, 34, 116, 56, 116, 23, 138, 56, 148, 198, 98, 80, 106, 176, 108, 191, 109, 250, 119, 92, 124, 0, 134, 20, 142, 52, 144, 38, 0, 196, 208, 60, 0, 60, 0, 42, 0, 81, 84, 6, 86, 102, 87, 72, 95, 154, 97, 78, 107, 88, 112, 173, 112, 187, 125, 149, 138, 187, 97, 36, 101, 171, 107, 113, 118, 82, 154, 38, 0, 200, 208, 60, 0, 60, 0, 42, 0, 106, 89, 43, 129, 187, 79, 38, 0, 208, 208, 60, 0, 60, 0, 42, 0, 162, 99, 8, 119, 61, 128, 170, 140, 255, 85, 208, 95, 86, 145, 38, 0, 209, 208, 60, 0, 60, 0, 42, 0, 84, 88, 45, 100, 187, 105, 157, 80, 76, 88, 40, 100, 38, 0, 213, 208, 60, 0, 60, 0, 42, 0, 149, 91, 17, 94, 111, 110, 105, 133, 217, 113, 234, 118, 173, 120, 47, 134, 38, 0, 220, 208, 60, 0, 60, 0, 42, 0, 76, 81, 240, 83, 42, 89, 32, 96, 75, 97, 134, 107, 112, 108, 240, 108, 30, 123, 206, 128, 212, 130, 198, 141, 176, 144, 177, 152, 237, 87, 39, 90, 97, 91, 172, 98, 232, 143, 196, 153, 216, 153, 38, 0, 221, 208, 60, 0, 60, 0, 42, 0, 199, 100, 164, 111, 38, 0, 241, 208, 60, 0, 60, 0, 42, 0, 145, 100, 144, 100, 90, 114, 38, 0, 48, 209, 60, 0, 60, 0, 4, 101, 38, 0, 160, 209, 60, 0, 60, 0, 42, 0, 78, 81, 16, 84, 31, 87, 14, 138, 38, 0, 164, 209, 60, 0, 60, 0, 42, 0, 75, 86, 120, 86, 242, 116, 38, 0, 181, 209, 60, 0, 60, 0, 42, 0, 95, 97, 118, 104, 219, 117, 82, 123, 113, 125, 26, 144, 107, 96, 11, 106, 105, 123, 38, 0, 244, 209, 60, 0, 60, 0, 42, 0, 6, 88, 204, 105, 127, 129, 42, 137, 0, 144, 57, 152, 164, 150, 38, 0, 44, 210, 60, 0, 60, 0, 42, 0, 120, 80, 87, 89, 172, 89, 149, 98, 15, 144, 42, 155, 146, 89, 29, 110, 176, 154, 38, 0, 65, 210, 60, 0, 60, 0, 95, 79, 38, 0, 185, 210, 60, 0, 60, 0, 42, 0, 93, 97, 121, 114, 210, 95, 38, 0, 200, 210, 60, 0, 60, 0, 214, 149, 38, 0, 12, 211, 60, 0, 60, 0, 42, 0, 97, 87, 70, 90, 244, 93, 138, 98, 173, 100, 250, 100, 119, 103, 226, 108, 62, 109, 44, 114, 54, 116, 52, 120, 119, 127, 173, 130, 219, 141, 23, 152, 245, 83, 145, 89, 165, 92, 21, 96, 94, 112, 56, 114, 187, 115, 164, 118, 6, 123, 56, 124, 25, 128, 224, 131, 105, 132, 49, 145, 38, 0, 16, 211, 60, 0, 60, 0, 42, 0, 36, 82, 66, 87, 127, 103, 72, 114, 227, 116, 169, 140, 166, 143, 17, 146, 42, 150, 116, 108, 38, 0, 20, 211, 60, 0, 60, 0, 42, 0, 107, 81, 237, 83, 76, 99, 51, 103, 67, 108, 38, 0, 40, 211, 60, 0, 60, 0, 42, 0, 105, 79, 4, 85, 150, 96, 87, 101, 155, 108, 127, 109, 76, 114, 253, 114, 23, 122, 135, 137, 157, 140, 91, 91, 198, 101, 238, 115, 8, 151, 56, 151, 38, 0, 61, 211, 60, 0, 60, 0, 42, 0, 109, 95, 142, 111, 249, 112, 168, 129, 48, 120, 74, 121, 218, 135, 219, 135, 38, 0, 69, 211, 60, 0, 60, 0, 14, 97, 38, 0, 184, 211, 60, 0, 60, 0, 42, 0, 191, 79, 79, 80, 65, 98, 71, 114, 199, 123, 232, 125, 233, 127, 77, 144, 173, 151, 25, 154, 62, 83, 167, 95, 252, 96, 246, 125, 81, 130, 57, 132, 89, 135, 10, 137, 222, 138, 38, 0, 196, 211, 60, 0, 60, 0, 42, 0, 182, 140, 45, 120, 134, 122, 38, 0, 201, 211, 60, 0, 60, 0, 42, 0, 106, 87, 115, 94, 176, 103, 13, 132, 85, 138, 38, 96, 168, 98, 217, 108, 249, 130, 241, 132, 131, 155, 38, 0, 208, 211, 60, 0, 60, 0, 42, 0, 32, 84, 22, 91, 99, 94, 226, 94, 10, 95, 131, 101, 186, 128, 61, 133, 137, 149, 91, 150, 93, 101, 244, 114, 88, 115, 72, 118, 38, 0, 236, 211, 60, 0, 60, 0, 42, 0, 72, 79, 5, 83, 13, 83, 15, 83, 134, 84, 250, 84, 3, 87, 3, 94, 22, 96, 155, 98, 177, 98, 85, 99, 225, 108, 102, 109, 177, 117, 50, 120, 222, 128, 47, 129, 222, 130, 97, 132, 178, 132, 141, 136, 18, 137, 11, 144, 234, 146, 253, 152, 145, 155, 36, 81, 150, 94, 97, 102, 211, 102, 174, 112, 176, 112, 22, 130, 167, 138, 75, 146, 132, 151, 20, 153, 198, 155, 38, 0, 237, 211, 60, 0, 60, 0, 42, 0, 69, 94, 180, 102, 221, 102, 17, 112, 6, 114, 38, 0, 92, 212, 60, 0, 60, 0, 42, 0, 245, 79, 125, 82, 106, 95, 83, 97, 83, 103, 25, 106, 2, 111, 226, 116, 104, 121, 104, 136, 121, 140, 199, 152, 196, 152, 67, 154, 196, 80, 225, 82, 12, 86, 214, 90, 125, 100, 141, 107, 155, 113, 57, 126, 241, 136, 226, 147, 99, 148, 223, 154, 62, 156, 38, 0, 136, 212, 60, 0, 60, 0, 42, 0, 193, 84, 31, 122, 128, 121, 38, 0, 141, 212, 60, 0, 60, 0, 42, 0, 83, 105, 247, 138, 74, 140, 168, 152, 174, 153, 11, 118, 81, 132, 38, 0, 60, 213, 60, 0, 60, 0, 42, 0, 124, 95, 171, 98, 178, 117, 174, 118, 171, 136, 127, 144, 66, 150, 86, 138, 159, 143, 129, 151, 242, 154, 38, 0, 61, 213, 60, 0, 60, 0, 119, 129, 38, 0, 68, 213, 60, 0, 60, 0, 42, 0, 57, 83, 60, 95, 197, 95, 204, 108, 204, 115, 98, 117, 139, 117, 70, 123, 254, 130, 157, 153, 86, 79, 135, 84, 237, 110, 243, 123, 124, 127, 253, 132, 241, 137, 85, 142, 184, 151, 224, 151, 220, 153, 111, 157, 38, 0, 77, 213, 60, 0, 60, 0, 42, 0, 79, 78, 60, 144, 106, 80, 38, 0, 88, 213, 60, 0, 60, 0, 42, 0, 11, 78, 85, 79, 166, 83, 15, 89, 200, 94, 48, 102, 179, 108, 85, 116, 119, 131, 102, 135, 192, 140, 80, 144, 30, 151, 21, 156, 64, 84, 135, 86, 136, 92, 215, 97, 70, 113, 21, 118, 69, 127, 92, 147, 38, 0, 89, 213, 60, 0, 60, 0, 42, 0, 209, 88, 120, 91, 80, 134, 20, 139, 180, 157, 226, 114, 39, 118, 172, 118, 110, 120, 221, 144, 253, 157, 38, 0, 92, 213, 60, 0, 60, 0, 42, 0, 210, 91, 104, 96, 141, 96, 241, 101, 87, 108, 34, 111, 163, 111, 26, 112, 85, 127, 240, 127, 145, 149, 146, 149, 80, 150, 211, 151, 233, 80, 250, 90, 251, 90, 77, 99, 181, 102, 136, 149, 251, 153, 243, 157, 62, 159, 38, 0, 96, 213, 60, 0, 60, 0, 42, 0, 114, 82, 68, 143, 142, 119, 38, 0, 104, 213, 60, 0, 60, 0, 42, 0, 253, 81, 43, 84, 184, 84, 99, 85, 138, 85, 187, 106, 181, 109, 216, 125, 102, 130, 156, 146, 119, 150, 121, 158, 159, 131, 225, 131, 244, 138, 94, 143, 222, 149, 38, 0, 105, 213, 60, 0, 60, 0, 42, 0, 8, 84, 200, 84, 210, 118, 228, 134, 164, 149, 212, 149, 92, 150, 12, 83, 209, 85, 217, 103, 188, 105, 152, 110, 205, 118, 195, 144, 38, 0, 109, 213, 60, 0, 60, 0, 42, 0, 162, 78, 9, 79, 238, 89, 230, 90, 247, 93, 82, 96, 151, 98, 109, 103, 65, 104, 134, 108, 47, 110, 56, 127, 155, 128, 42, 130, 5, 152, 47, 89, 70, 96, 149, 112, 63, 127, 15, 152, 38, 0, 116, 213, 60, 0, 60, 0, 42, 0, 165, 78, 85, 80, 179, 84, 147, 87, 90, 89, 105, 91, 179, 91, 200, 97, 119, 105, 119, 109, 35, 112, 249, 135, 227, 137, 114, 138, 231, 138, 130, 144, 237, 153, 184, 154, 141, 84, 176, 93, 232, 94, 44, 107, 108, 115, 78, 116, 206, 117, 164, 133, 162, 145, 38, 152, 173, 155, 38, 0, 117, 213, 60, 0, 60, 0, 42, 0, 190, 82, 56, 104, 238, 127, 136, 137, 38, 0, 137, 213, 60, 0, 60, 0, 42, 0, 22, 80, 120, 94, 79, 103, 71, 131, 76, 136, 187, 96, 38, 0, 165, 213, 60, 0, 60, 0, 42, 0, 171, 78, 17, 84, 174, 86, 230, 115, 21, 145, 255, 151, 9, 153, 87, 153, 153, 153, 140, 133, 38, 0, 200, 213, 60, 0, 60, 0, 42, 0, 83, 86, 159, 88, 91, 134, 49, 138, 84, 107, 38, 0, 204, 213, 60, 0, 60, 0, 42, 0, 178, 97, 246, 106, 123, 115, 210, 142, 218, 93, 112, 94, 7, 101, 38, 0, 208, 213, 60, 0, 60, 0, 71, 107, 38, 0, 216, 213, 60, 0, 60, 0, 42, 0, 170, 150, 87, 154, 174, 93, 107, 115, 129, 115, 38, 0, 1, 214, 60, 0, 60, 0, 42, 0, 85, 89, 0, 114, 107, 141, 105, 151, 8, 95, 43, 109, 49, 113, 41, 155, 38, 0, 4, 214, 60, 0, 60, 0, 42, 0, 212, 79, 244, 92, 38, 95, 248, 97, 91, 102, 235, 108, 171, 112, 132, 115, 185, 115, 254, 115, 41, 119, 77, 119, 67, 125, 98, 125, 35, 126, 55, 130, 82, 136, 226, 140, 73, 146, 111, 152, 7, 81, 27, 91, 33, 102, 4, 116, 195, 117, 12, 119, 111, 126, 254, 127, 6, 135, 162, 138, 215, 146, 253, 153, 38, 0, 8, 214, 60, 0, 60, 0, 42, 0, 81, 91, 116, 122, 64, 136, 1, 152, 92, 125, 144, 141, 38, 0, 16, 214, 60, 0, 60, 0, 204, 90, 38, 0, 17, 214, 60, 0, 60, 0, 42, 0, 224, 79, 84, 83, 62, 89, 253, 92, 62, 99, 121, 109, 249, 114, 5, 129, 7, 129, 162, 131, 207, 146, 48, 152, 39, 83, 246, 83, 201, 87, 74, 96, 143, 96, 28, 97, 203, 123, 38, 0, 21, 214, 60, 0, 60, 0, 42, 0, 168, 78, 68, 81, 17, 82, 139, 87, 98, 95, 194, 108, 206, 110, 5, 112, 80, 112, 175, 112, 146, 113, 233, 115, 105, 116, 74, 131, 162, 135, 97, 136, 8, 144, 162, 144, 163, 147, 168, 153, 16, 89, 25, 90, 87, 138, 229, 143, 88, 150, 38, 0, 28, 214, 60, 0, 60, 0, 42, 0, 110, 81, 87, 95, 224, 96, 103, 97, 179, 102, 89, 133, 74, 142, 175, 145, 139, 151, 146, 80, 18, 86, 237, 91, 175, 95, 229, 105, 251, 118, 17, 139, 127, 139, 38, 0, 56, 214, 60, 0, 60, 0, 42, 0, 78, 78, 146, 78, 124, 84, 213, 88, 250, 88, 125, 89, 181, 92, 39, 95, 54, 98, 72, 98, 10, 102, 103, 102, 235, 107, 105, 109, 207, 109, 86, 110, 248, 110, 148, 111, 224, 111, 233, 111, 93, 112, 208, 114, 37, 116, 90, 116, 224, 116, 147, 118, 92, 121, 202, 124, 30, 126, 225, 128, 166, 130, 107, 132, 191, 132, 78, 134, 95, 134, 116, 135, 119, 139, 106, 140, 172, 147, 0, 152, 101, 152, 43, 81, 177, 81, 55, 86, 237, 90, 238, 90, 25, 96, 141, 108, 200, 110, 236, 110, 146, 114, 34, 115, 156, 118, 158, 118, 142, 123, 85, 128, 144, 145, 44, 153, 13, 155, 38, 0, 57, 214, 60, 0, 60, 0, 42, 0, 209, 96, 22, 98, 119, 145, 38, 0, 60, 214, 60, 0, 60, 0, 42, 0, 90, 90, 15, 102, 247, 109, 62, 110, 63, 116, 66, 155, 2, 87, 219, 96, 183, 110, 28, 113, 189, 149, 38, 0, 64, 214, 60, 0, 60, 0, 42, 0, 253, 95, 218, 96, 15, 123, 235, 86, 38, 0, 77, 214, 60, 0, 60, 0, 42, 0, 196, 84, 24, 95, 94, 108, 211, 108, 42, 109, 216, 112, 5, 125, 121, 134, 12, 138, 59, 157, 78, 102, 146, 111, 202, 123, 119, 146, 40, 155, 38, 0, 84, 214, 60, 0, 60, 0, 42, 0, 22, 83, 140, 84, 5, 91, 58, 106, 107, 112, 117, 117, 141, 121, 190, 121, 177, 130, 239, 131, 113, 138, 65, 139, 168, 140, 116, 151, 240, 79, 41, 86, 114, 90, 237, 100, 107, 117, 74, 154, 162, 159, 38, 0, 85, 214, 60, 0, 60, 0, 42, 0, 244, 100, 43, 101, 186, 120, 187, 120, 107, 122, 205, 119, 225, 119, 45, 121, 74, 148, 38, 0, 88, 214, 60, 0, 60, 0, 42, 0, 56, 78, 154, 85, 80, 89, 166, 91, 123, 94, 163, 96, 219, 99, 97, 107, 101, 102, 83, 104, 25, 110, 101, 113, 176, 116, 8, 125, 132, 144, 105, 154, 37, 156, 240, 91, 253, 97, 208, 100, 219, 116, 150, 118, 70, 119, 89, 125, 98, 140, 88, 143, 112, 147, 54, 148, 31, 155, 38, 0, 92, 214, 60, 0, 60, 0, 42, 0, 59, 109, 209, 110, 62, 115, 65, 140, 202, 149, 222, 134, 38, 0, 105, 214, 60, 0, 60, 0, 42, 0, 240, 81, 76, 94, 168, 95, 77, 96, 246, 96, 48, 97, 76, 97, 67, 102, 68, 102, 165, 105, 193, 108, 95, 110, 201, 110, 98, 111, 76, 113, 156, 116, 135, 118, 193, 123, 39, 124, 82, 131, 87, 135, 81, 144, 141, 150, 195, 158, 164, 85, 147, 90, 51, 96, 93, 116, 147, 128, 186, 140, 164, 147, 38, 0, 140, 214, 60, 0, 60, 0, 42, 0, 47, 83, 222, 86, 251, 94, 138, 95, 98, 96, 148, 96, 247, 97, 102, 102, 3, 103, 156, 106, 238, 109, 174, 111, 112, 112, 106, 115, 106, 126, 190, 129, 52, 131, 212, 134, 168, 138, 196, 140, 106, 79, 3, 104, 4, 109, 217, 110, 212, 118, 124, 138, 244, 143, 46, 152, 96, 156, 38, 0, 141, 214, 60, 0, 60, 0, 42, 0, 131, 82, 114, 115, 132, 86, 38, 0, 161, 214, 60, 0, 60, 0, 42, 0, 150, 91, 107, 106, 4, 148, 139, 111, 28, 146, 204, 158, 38, 0, 168, 214, 60, 0, 60, 0, 42, 0, 238, 84, 134, 86, 93, 91, 72, 101, 133, 101, 201, 102, 159, 104, 141, 109, 198, 109, 59, 114, 180, 128, 117, 145, 77, 154, 154, 80, 194, 86, 36, 93, 189, 107, 135, 113, 155, 118, 83, 134, 26, 153, 38, 0, 196, 214, 60, 0, 60, 0, 42, 0, 175, 79, 25, 80, 154, 83, 14, 84, 60, 84, 137, 85, 197, 85, 63, 94, 140, 95, 61, 103, 102, 113, 221, 115, 5, 144, 61, 84, 163, 85, 149, 87, 32, 88, 184, 109, 52, 115, 204, 123, 97, 138, 67, 139, 87, 145, 49, 153, 38, 0, 200, 214, 60, 0, 60, 0, 42, 0, 219, 82, 243, 82, 100, 88, 206, 88, 4, 113, 143, 113, 251, 113, 176, 133, 19, 138, 136, 102, 219, 102, 11, 114, 111, 115, 129, 126, 119, 132, 66, 148, 38, 0, 204, 214, 60, 0, 60, 0, 59, 107, 38, 0, 217, 214, 60, 0, 60, 0, 168, 133, 38, 0, 228, 214, 60, 0, 60, 0, 42, 0, 167, 85, 132, 102, 74, 113, 49, 132, 220, 112, 224, 138, 252, 138, 38, 0, 252, 214, 60, 0, 60, 0, 42, 0, 73, 83, 153, 85, 193, 107, 236, 113, 148, 130, 107, 134, 122, 134, 38, 0, 24, 215, 60, 0, 60, 0, 42, 0, 89, 95, 189, 95, 238, 99, 137, 102, 71, 113, 241, 138, 29, 143, 190, 158, 157, 100, 236, 127, 38, 0, 52, 215, 60, 0, 60, 0, 42, 0, 17, 79, 58, 100, 203, 112, 102, 117, 103, 134, 187, 84, 213, 100, 179, 150, 249, 154, 66, 157, 38, 0, 60, 215, 60, 0, 60, 0, 42, 0, 100, 96, 78, 139, 248, 157, 121, 83, 38, 0, 73, 215, 60, 0, 60, 0, 42, 0, 71, 81, 246, 81, 8, 83, 54, 109, 248, 128, 95, 96, 247, 128, 38, 0, 81, 215, 60, 0, 60, 0, 209, 158, 38, 0, 84, 215, 60, 0, 60, 0, 42, 0, 21, 102, 35, 107, 152, 112, 213, 117, 136, 95, 251, 95, 128, 99, 46, 113, 34, 138, 193, 145, 38, 0, 88, 215, 60, 0, 60, 0, 42, 0, 3, 84, 121, 92, 7, 125, 22, 138, 225, 78, 84, 108, 153, 117, 196, 143, 85, 159, 38, 0, 96, 215, 60, 0, 60, 0, 42, 0, 32, 107, 61, 107, 70, 107, 222, 94, 38, 0, 97, 215, 60, 0, 60, 0, 42, 0, 56, 84, 112, 96, 61, 109, 213, 127, 79, 86, 89, 107, 93, 111, 214, 127, 38, 0, 101, 215, 60, 0, 60, 0, 8, 130, 38, 0, 108, 215, 60, 0, 60, 0, 42, 0, 214, 80, 222, 81, 156, 85, 107, 86, 205, 86, 236, 89, 9, 91, 12, 94, 153, 97, 152, 97, 49, 98, 94, 102, 230, 102, 153, 113, 185, 113, 186, 113, 167, 114, 167, 121, 0, 122, 178, 127, 165, 84, 15, 85, 59, 86, 149, 96, 50, 98, 191, 102, 55, 107, 249, 113, 20, 114, 104, 140, 60, 153, 38, 0, 144, 215, 60, 0, 60, 0, 42, 0, 112, 138, 181, 114, 136, 126, 109, 137, 33, 152, 224, 158, 0, 0, 170, 170, 170, 170, 98, 40, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 31, 1, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 84, 34, 0, 0, 84, 34, 0, 0, 84, 34, 0, 0, 84, 34, 0, 0, 4, 40, 0, 0, 4, 40, 0, 0, 74, 40, 0, 0, 74, 40, 0, 0, 74, 40, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 24, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 94, 0, 102, 0, 110, 0, 118, 0, 48, 0, 56, 0, 117, 0, 125, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 142, 0, 150, 0, 48, 0, 48, 0, 48, 0, 158, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 174, 0, 182, 0, 190, 0, 198, 0, 206, 0, 214, 0, 222, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 230, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 238, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 241, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 140, 0, 48, 0, 48, 0, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 249, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 1, 1, 48, 0, 48, 0, 7, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 120, 1, 184, 1, 192, 0, 212, 1, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 64, 44, 0, 202, 97, 44, 0, 202, 130, 44, 0, 202, 163, 44, 0, 202, 196, 44, 0, 202, 229, 44, 0, 202, 6, 45, 0, 202, 39, 45, 0, 202, 72, 45, 0, 202, 105, 45, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 35, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 35, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 35, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 198, 0, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 2, 0, 5, 71, 238, 102, 201, 134, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 0, 0, 192, 0, 0, 0, 197, 226, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 1, 0, 197, 162, 1, 0, 197, 226, 1, 0, 197, 34, 2, 0, 197, 98, 2, 0, 197, 162, 2, 0, 197, 226, 2, 0, 197, 34, 3, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 66, 2, 0, 197, 194, 0, 0, 197, 163, 4, 0, 197, 3, 5, 0, 197, 99, 5, 0, 197, 195, 0, 0, 197, 195, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 197, 226, 6, 0, 197, 34, 7, 0, 197, 99, 7, 0, 197, 194, 4, 0, 197, 194, 7, 0, 197, 34, 5, 0, 197, 2, 8, 0, 197, 66, 8, 0, 197, 130, 5, 0, 197, 131, 8, 0, 197, 227, 8, 0, 197, 66, 9, 0, 197, 2, 1, 0, 197, 130, 9, 0, 197, 194, 9, 0, 197, 2, 10, 0, 197, 66, 10, 0, 197, 130, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 10, 0, 197, 2, 11, 0, 197, 130, 3, 0, 197, 2, 4, 0, 197, 162, 7, 0, 192, 0, 0, 0, 197, 66, 11, 0, 197, 130, 11, 0, 197, 194, 11, 0, 197, 2, 12, 0, 197, 66, 12, 0, 192, 0, 0, 0, 197, 162, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 12, 0, 197, 194, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 7, 0, 197, 2, 13, 0, 197, 66, 13, 0, 192, 0, 0, 0, 197, 66, 0, 0, 197, 226, 7, 0, 197, 130, 13, 0, 197, 194, 13, 0, 197, 2, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 14, 0, 192, 0, 0, 0, 197, 130, 14, 0, 192, 0, 0, 0, 197, 194, 14, 0, 192, 0, 0, 0, 197, 2, 15, 0, 192, 0, 0, 0, 197, 66, 15, 0, 197, 131, 15, 0, 197, 226, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 16, 0, 197, 99, 16, 0, 197, 194, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 17, 0, 192, 0, 0, 0, 197, 98, 15, 0, 197, 66, 17, 0, 197, 130, 17, 0, 197, 194, 17, 0, 197, 2, 18, 0, 197, 66, 16, 0, 197, 66, 18, 0, 197, 130, 18, 0, 197, 194, 18, 0, 197, 34, 18, 0, 197, 3, 19, 0, 197, 99, 19, 0, 197, 194, 19, 0, 197, 2, 20, 0, 197, 66, 20, 0, 197, 131, 20, 0, 197, 226, 20, 0, 197, 34, 21, 0, 197, 98, 21, 0, 197, 162, 21, 0, 197, 227, 21, 0, 197, 67, 22, 0, 197, 163, 22, 0, 197, 2, 23, 0, 197, 66, 23, 0, 197, 130, 23, 0, 197, 195, 23, 0, 197, 34, 24, 0, 197, 99, 24, 0, 197, 194, 24, 0, 197, 2, 25, 0, 197, 130, 22, 0, 197, 66, 25, 0, 197, 131, 25, 0, 197, 34, 17, 0, 197, 98, 14, 0, 197, 34, 15, 0, 197, 2, 16, 0, 197, 226, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 25, 0, 197, 34, 26, 0, 197, 98, 26, 0, 197, 162, 17, 0, 197, 163, 26, 0, 5, 5, 6, 121, 197, 34, 0, 0, 197, 194, 8, 0, 5, 5, 10, 121, 197, 130, 13, 0, 197, 194, 13, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 2, 27, 0, 197, 66, 27, 0, 197, 130, 27, 0, 197, 34, 8, 0, 197, 194, 27, 0, 197, 2, 28, 0, 197, 34, 3, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 197, 130, 2, 0, 197, 67, 28, 0, 197, 98, 1, 0, 197, 226, 1, 0, 197, 226, 7, 0, 192, 0, 0, 0, 197, 162, 28, 0, 197, 98, 2, 0, 197, 2, 14, 0, 197, 227, 28, 0, 197, 162, 2, 0, 197, 66, 29, 0, 197, 131, 29, 0, 197, 226, 2, 0, 197, 227, 29, 0, 197, 67, 30, 0, 197, 163, 30, 0, 197, 3, 31, 0, 197, 99, 31, 0, 197, 195, 31, 0, 192, 0, 0, 0, 197, 34, 32, 0, 192, 0, 0, 0, 197, 2, 30, 0, 197, 98, 27, 0, 197, 162, 3, 0, 197, 98, 8, 0, 197, 99, 32, 0, 192, 0, 0, 0, 197, 194, 32, 0, 197, 2, 33, 0, 197, 226, 3, 0, 197, 162, 27, 0, 197, 226, 6, 0, 197, 34, 31, 0, 197, 34, 7, 0, 197, 194, 4, 0, 197, 34, 5, 0, 197, 2, 8, 0, 197, 130, 5, 0, 192, 0, 0, 0, 197, 194, 10, 0, 197, 67, 33, 0, 197, 66, 11, 0, 197, 162, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 7, 0, 197, 2, 13, 0, 197, 98, 13, 0, 197, 66, 3, 0, 197, 226, 33, 0, 197, 34, 34, 0, 192, 0, 0, 0, 197, 130, 1, 0, 197, 66, 4, 0, 197, 98, 34, 0, 197, 162, 34, 0, 197, 226, 34, 0, 197, 162, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 6, 121, 5, 5, 10, 121, 5, 5, 10, 121, 5, 5, 10, 121, 5, 5, 12, 121, 5, 5, 10, 121, 5, 5, 20, 121, 5, 5, 12, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 10, 121, 5, 5, 16, 121, 5, 5, 16, 121, 5, 5, 16, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 28, 121, 5, 5, 18, 121, 5, 5, 20, 121, 5, 5, 18, 121, 5, 5, 28, 121, 5, 5, 20, 121, 5, 5, 6, 121, 5, 5, 20, 121, 5, 5, 10, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 6, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 20, 121, 5, 5, 30, 121, 5, 5, 20, 121, 5, 5, 34, 121, 5, 5, 20, 121, 5, 5, 38, 121, 5, 5, 20, 121, 5, 5, 40, 121, 5, 5, 20, 121, 5, 5, 28, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 28, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 18, 121, 5, 5, 24, 121, 5, 5, 20, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 28, 121, 5, 5, 24, 121, 5, 5, 34, 121, 5, 5, 24, 121, 5, 5, 36, 121, 5, 5, 24, 121, 5, 5, 38, 121, 5, 5, 24, 121, 5, 5, 40, 121, 5, 5, 24, 121, 5, 5, 42, 121, 5, 5, 28, 121, 5, 5, 6, 121, 5, 5, 28, 121, 5, 5, 12, 121, 5, 5, 28, 121, 5, 5, 28, 121, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 28, 121, 5, 5, 34, 121, 5, 5, 28, 121, 5, 5, 38, 121, 5, 5, 28, 121, 5, 5, 40, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 34, 121, 5, 5, 42, 121, 5, 5, 40, 121, 5, 5, 28, 121, 5, 5, 42, 121, 5, 5, 42, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 12, 121, 5, 5, 16, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 103, 121, 5, 5, 129, 121, 5, 5, 107, 121, 5, 5, 119, 121, 5, 5, 107, 121, 5, 5, 127, 121, 5, 5, 111, 121, 5, 5, 119, 121, 5, 5, 111, 121, 5, 5, 139, 121, 5, 5, 115, 121, 5, 5, 119, 121, 5, 5, 115, 121, 5, 5, 129, 121, 5, 5, 119, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 119, 121, 5, 5, 119, 121, 5, 5, 129, 121, 5, 5, 127, 121, 5, 5, 107, 121, 5, 5, 127, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 127, 121, 5, 5, 115, 121, 5, 5, 127, 121, 5, 5, 119, 121, 5, 5, 127, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 103, 121, 5, 5, 129, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 139, 121, 5, 5, 129, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 129, 121, 5, 5, 137, 121, 5, 5, 103, 121, 5, 5, 137, 121, 5, 5, 111, 121, 5, 5, 137, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 137, 121, 5, 5, 115, 121, 5, 5, 137, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 137, 121, 5, 5, 129, 121, 5, 5, 137, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 139, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 103, 121, 5, 5, 139, 121, 5, 5, 107, 121, 5, 5, 129, 121, 5, 5, 115, 121, 5, 5, 107, 121, 5, 5, 119, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 6, 121, 5, 5, 10, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 12, 121, 5, 5, 16, 121, 5, 5, 12, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 28, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 36, 121, 5, 5, 18, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 18, 121, 5, 5, 34, 121, 5, 5, 18, 121, 5, 5, 42, 121, 5, 5, 28, 121, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 28, 121, 5, 5, 36, 121, 5, 5, 42, 121, 5, 5, 18, 121, 5, 5, 42, 121, 5, 5, 20, 121, 5, 5, 6, 121, 5, 5, 34, 121, 5, 5, 6, 121, 5, 5, 36, 121, 5, 5, 6, 121, 5, 5, 42, 121, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 34, 0, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 98, 0, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 162, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 34, 1, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 66, 14, 0, 5, 5, 107, 121, 197, 130, 14, 0, 5, 5, 111, 121, 197, 194, 14, 0, 5, 5, 115, 121, 197, 2, 15, 0, 5, 5, 119, 121, 197, 66, 15, 0, 197, 131, 15, 0, 197, 226, 15, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 34, 16, 0, 197, 99, 16, 0, 197, 194, 16, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 2, 17, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 34, 0, 0, 197, 194, 8, 0, 5, 5, 10, 121, 197, 130, 13, 0, 197, 194, 13, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 2, 27, 0, 197, 66, 27, 0, 197, 130, 27, 0, 197, 34, 8, 0, 197, 194, 27, 0, 197, 2, 28, 0, 197, 34, 3, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 75, 27, 0, 16, 42, 2, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 130, 34, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 27, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 84, 32, 0, 0, 84, 32, 0, 0, 84, 32, 0, 0, 84, 32, 0, 0, 244, 33, 0, 0, 244, 33, 0, 0, 106, 34, 0, 0, 106, 34, 0, 0, 106, 34, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 248, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 94, 0, 102, 0, 110, 0, 118, 0, 48, 0, 56, 0, 117, 0, 125, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 142, 0, 150, 0, 48, 0, 48, 0, 48, 0, 158, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 48, 0, 48, 0, 174, 0, 48, 0, 180, 0, 188, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 191, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 198, 0, 206, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 209, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 140, 0, 48, 0, 48, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 239, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 217, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 225, 0, 48, 0, 48, 0, 231, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 120, 1, 184, 1, 192, 0, 212, 1, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 192, 11, 0, 202, 225, 11, 0, 202, 2, 12, 0, 202, 35, 12, 0, 202, 68, 12, 0, 202, 101, 12, 0, 202, 134, 12, 0, 202, 167, 12, 0, 202, 200, 12, 0, 202, 233, 12, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 2, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 2, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 3, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 6, 0, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 1, 0, 5, 71, 238, 102, 201, 198, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 2, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 200, 160, 3, 0, 197, 34, 0, 0, 192, 0, 0, 0, 200, 96, 4, 0, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 200, 32, 5, 0, 197, 162, 0, 0, 200, 224, 5, 0, 197, 226, 0, 0, 192, 0, 0, 0, 200, 160, 6, 0, 197, 34, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 3, 255, 194, 5, 4, 255, 194, 5, 5, 255, 194, 5, 6, 255, 194, 5, 7, 255, 194, 5, 8, 255, 194, 5, 9, 255, 194, 5, 10, 255, 194, 5, 11, 255, 194, 5, 12, 255, 194, 5, 13, 255, 194, 5, 14, 255, 194, 5, 15, 255, 194, 5, 16, 255, 194, 5, 17, 255, 194, 5, 18, 255, 194, 5, 19, 255, 194, 5, 20, 255, 194, 5, 21, 255, 194, 5, 22, 255, 194, 5, 23, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 24, 255, 194, 5, 25, 255, 194, 5, 26, 255, 194, 5, 27, 255, 194, 5, 28, 255, 194, 5, 29, 255, 194, 5, 30, 255, 194, 5, 31, 255, 194, 5, 32, 255, 194, 5, 33, 255, 194, 5, 34, 255, 194, 5, 35, 255, 194, 5, 36, 255, 194, 5, 37, 255, 194, 5, 38, 255, 194, 5, 39, 255, 194, 5, 40, 255, 194, 5, 41, 255, 194, 5, 42, 255, 194, 5, 43, 255, 194, 5, 44, 255, 194, 5, 45, 255, 194, 5, 46, 255, 194, 5, 47, 255, 194, 5, 48, 255, 194, 5, 49, 255, 194, 5, 50, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 6, 121, 5, 130, 0, 0, 5, 5, 12, 121, 5, 130, 0, 0, 5, 5, 20, 121, 5, 130, 0, 0, 5, 5, 24, 121, 5, 130, 0, 0, 5, 5, 30, 121, 5, 130, 0, 0, 5, 5, 6, 121, 10, 130, 0, 0, 5, 5, 12, 121, 9, 130, 0, 0, 5, 5, 20, 121, 8, 130, 0, 0, 5, 5, 24, 121, 7, 130, 0, 0, 5, 5, 30, 121, 6, 130, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 200, 160, 3, 0, 197, 34, 0, 0, 5, 5, 10, 121, 200, 96, 4, 0, 197, 98, 0, 0, 5, 5, 16, 121, 5, 5, 18, 121, 200, 32, 5, 0, 197, 162, 0, 0, 200, 224, 5, 0, 197, 226, 0, 0, 5, 5, 28, 121, 200, 160, 6, 0, 197, 34, 1, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 194, 5, 3, 255, 194, 5, 4, 255, 194, 5, 5, 255, 194, 5, 6, 255, 194, 5, 7, 255, 194, 5, 8, 255, 194, 5, 9, 255, 194, 5, 10, 255, 194, 5, 11, 255, 194, 5, 12, 255, 194, 5, 13, 255, 194, 5, 14, 255, 194, 5, 15, 255, 194, 5, 16, 255, 194, 5, 17, 255, 194, 5, 18, 255, 194, 5, 19, 255, 194, 5, 20, 255, 194, 5, 21, 255, 194, 5, 22, 255, 194, 5, 23, 255, 194, 5, 24, 255, 194, 5, 25, 255, 194, 5, 26, 255, 194, 5, 27, 255, 194, 5, 28, 255, 194, 5, 29, 255, 194, 5, 30, 255, 194, 5, 31, 255, 194, 5, 32, 255, 194, 5, 33, 255, 194, 5, 34, 255, 194, 5, 35, 255, 194, 5, 36, 255, 194, 5, 37, 255, 194, 5, 38, 255, 194, 5, 39, 255, 194, 5, 40, 255, 194, 5, 41, 255, 194, 5, 42, 255, 194, 5, 43, 255, 194, 5, 44, 255, 194, 5, 45, 255, 194, 5, 46, 255, 194, 5, 47, 255, 194, 5, 48, 255, 194, 5, 49, 255, 194, 5, 50, 255, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 6, 121, 5, 5, 48, 0, 0, 17, 0, 192, 5, 130, 12, 121, 5, 5, 48, 0, 3, 17, 0, 192, 5, 130, 20, 121, 5, 5, 48, 0, 7, 17, 0, 192, 5, 130, 24, 121, 5, 5, 48, 0, 9, 17, 0, 192, 5, 130, 30, 121, 5, 5, 48, 0, 12, 17, 0, 192, 5, 130, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 107, 37, 0, 16, 59, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 52, 140, 3, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 13, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 60, 0, 0, 0, 60, 1, 0, 0, 156, 95, 1, 0, 156, 95, 1, 0, 12, 96, 3, 0, 12, 96, 3, 0, 24, 140, 3, 0, 18, 0, 0, 0, 17, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 39, 246, 247, 248, 249, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 254, 255, 50, 105, 114, 84, 1, 0, 192, 8, 205, 20, 64, 0, 192, 0, 96, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 104, 0, 112, 0, 120, 0, 128, 0, 136, 0, 144, 0, 152, 0, 160, 0, 165, 0, 173, 0, 181, 0, 189, 0, 197, 0, 205, 0, 213, 0, 221, 0, 229, 0, 237, 0, 245, 0, 253, 0, 5, 1, 13, 1, 21, 1, 29, 1, 37, 1, 45, 1, 53, 1, 61, 1, 69, 1, 77, 1, 85, 1, 93, 1, 101, 1, 109, 1, 117, 1, 125, 1, 133, 1, 141, 1, 149, 1, 157, 1, 165, 1, 173, 1, 181, 1, 189, 1, 197, 1, 205, 1, 213, 1, 221, 1, 229, 1, 237, 1, 245, 1, 253, 1, 5, 2, 13, 2, 21, 2, 29, 2, 37, 2, 45, 2, 53, 2, 61, 2, 69, 2, 76, 2, 84, 2, 92, 2, 98, 2, 106, 2, 113, 2, 121, 2, 128, 2, 136, 2, 144, 2, 152, 2, 160, 2, 168, 2, 176, 2, 184, 2, 192, 2, 200, 2, 208, 2, 216, 2, 224, 2, 232, 2, 240, 2, 248, 2, 0, 3, 8, 3, 16, 3, 24, 3, 32, 3, 40, 3, 48, 3, 56, 3, 64, 3, 72, 3, 80, 3, 88, 3, 96, 3, 104, 3, 112, 3, 120, 3, 128, 3, 136, 3, 144, 3, 152, 3, 159, 3, 167, 3, 175, 3, 183, 3, 191, 3, 199, 3, 207, 3, 215, 3, 223, 3, 231, 3, 239, 3, 246, 3, 254, 3, 6, 4, 14, 4, 22, 4, 30, 4, 38, 4, 44, 4, 52, 4, 60, 4, 68, 4, 76, 4, 84, 4, 92, 4, 100, 4, 108, 4, 116, 4, 124, 4, 132, 4, 140, 4, 148, 4, 156, 4, 164, 4, 172, 4, 180, 4, 188, 4, 196, 4, 204, 4, 212, 4, 220, 4, 228, 4, 236, 4, 244, 4, 252, 4, 4, 5, 12, 5, 20, 5, 28, 5, 36, 5, 44, 5, 52, 5, 59, 5, 67, 5, 75, 5, 83, 5, 91, 5, 99, 5, 107, 5, 115, 5, 122, 5, 130, 5, 138, 5, 146, 5, 152, 5, 160, 5, 167, 5, 175, 5, 183, 5, 191, 5, 199, 5, 206, 5, 214, 5, 222, 5, 230, 5, 238, 5, 246, 5, 254, 5, 6, 6, 14, 6, 22, 6, 30, 6, 38, 6, 46, 6, 54, 6, 62, 6, 70, 6, 78, 6, 86, 6, 94, 6, 102, 6, 109, 6, 117, 6, 125, 6, 133, 6, 141, 6, 149, 6, 157, 6, 165, 6, 173, 6, 181, 6, 189, 6, 197, 6, 205, 6, 213, 6, 221, 6, 227, 6, 235, 6, 243, 6, 250, 6, 2, 7, 10, 7, 17, 7, 25, 7, 33, 7, 40, 7, 48, 7, 56, 7, 63, 7, 71, 7, 79, 7, 87, 7, 95, 7, 103, 7, 111, 7, 119, 7, 127, 7, 135, 7, 143, 7, 151, 7, 159, 7, 167, 7, 175, 7, 183, 7, 190, 7, 198, 7, 206, 7, 214, 7, 222, 7, 230, 7, 237, 7, 245, 7, 253, 7, 4, 8, 12, 8, 20, 8, 28, 8, 36, 8, 42, 8, 50, 8, 58, 8, 66, 8, 74, 8, 82, 8, 90, 8, 97, 8, 105, 8, 113, 8, 120, 8, 128, 8, 136, 8, 144, 8, 152, 8, 160, 8, 168, 8, 176, 8, 184, 8, 192, 8, 200, 8, 208, 8, 216, 8, 224, 8, 232, 8, 239, 8, 247, 8, 255, 8, 7, 9, 14, 9, 22, 9, 30, 9, 38, 9, 46, 9, 54, 9, 62, 9, 70, 9, 78, 9, 86, 9, 94, 9, 102, 9, 110, 9, 118, 9, 126, 9, 134, 9, 142, 9, 150, 9, 158, 9, 166, 9, 174, 9, 182, 9, 190, 9, 198, 9, 206, 9, 214, 9, 222, 9, 230, 9, 238, 9, 246, 9, 254, 9, 6, 10, 14, 10, 22, 10, 30, 10, 38, 10, 46, 10, 54, 10, 62, 10, 70, 10, 78, 10, 86, 10, 94, 10, 102, 10, 110, 10, 118, 10, 126, 10, 134, 10, 142, 10, 150, 10, 158, 10, 166, 10, 174, 10, 182, 10, 190, 10, 198, 10, 206, 10, 214, 10, 222, 10, 230, 10, 238, 10, 246, 10, 254, 10, 5, 11, 13, 11, 21, 11, 29, 11, 37, 11, 45, 11, 53, 11, 61, 11, 69, 11, 77, 11, 85, 11, 93, 11, 101, 11, 109, 11, 117, 11, 125, 11, 133, 11, 140, 11, 148, 11, 156, 11, 164, 11, 172, 11, 180, 11, 188, 11, 196, 11, 204, 11, 212, 11, 220, 11, 228, 11, 236, 11, 244, 11, 252, 11, 4, 12, 12, 12, 20, 12, 28, 12, 36, 12, 44, 12, 52, 12, 59, 12, 67, 12, 75, 12, 83, 12, 91, 12, 48, 0, 48, 0, 48, 0, 48, 0, 99, 12, 107, 12, 115, 12, 123, 12, 131, 12, 139, 12, 147, 12, 155, 12, 163, 12, 171, 12, 179, 12, 187, 12, 195, 12, 203, 12, 211, 12, 219, 12, 227, 12, 235, 12, 243, 12, 251, 12, 3, 13, 11, 13, 19, 13, 27, 13, 35, 13, 43, 13, 51, 13, 59, 13, 67, 13, 75, 13, 83, 13, 91, 13, 99, 13, 107, 13, 112, 13, 119, 13, 127, 13, 135, 13, 143, 13, 151, 13, 157, 13, 165, 13, 173, 13, 181, 13, 189, 13, 197, 13, 204, 13, 212, 13, 220, 13, 228, 13, 236, 13, 244, 13, 252, 13, 4, 14, 12, 14, 20, 14, 28, 14, 36, 14, 44, 14, 51, 14, 59, 14, 67, 14, 75, 14, 83, 14, 91, 14, 99, 14, 107, 14, 115, 14, 123, 14, 131, 14, 139, 14, 147, 14, 155, 14, 163, 14, 171, 14, 179, 14, 187, 14, 195, 14, 203, 14, 211, 14, 218, 14, 226, 14, 234, 14, 242, 14, 250, 14, 2, 15, 10, 15, 18, 15, 26, 15, 34, 15, 42, 15, 50, 15, 58, 15, 66, 15, 74, 15, 82, 15, 90, 15, 98, 15, 106, 15, 114, 15, 48, 0, 48, 0, 48, 0, 48, 0, 122, 15, 130, 15, 138, 15, 146, 15, 154, 15, 162, 15, 170, 15, 178, 15, 48, 0, 48, 0, 185, 15, 193, 15, 201, 15, 209, 15, 217, 15, 225, 15, 233, 15, 241, 15, 249, 15, 1, 16, 9, 16, 17, 16, 25, 16, 33, 16, 41, 16, 49, 16, 57, 16, 59, 16, 67, 16, 75, 16, 83, 16, 91, 16, 99, 16, 107, 16, 115, 16, 123, 16, 131, 16, 139, 16, 147, 16, 155, 16, 163, 16, 171, 16, 179, 16, 187, 16, 195, 16, 203, 16, 211, 16, 219, 16, 227, 16, 235, 16, 243, 16, 251, 16, 3, 17, 10, 17, 18, 17, 26, 17, 34, 17, 40, 17, 48, 17, 56, 17, 64, 17, 72, 17, 80, 17, 88, 17, 95, 17, 103, 17, 111, 17, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 119, 17, 127, 17, 135, 17, 143, 17, 151, 17, 153, 17, 161, 17, 169, 17, 177, 17, 185, 17, 193, 17, 201, 17, 209, 17, 217, 17, 225, 17, 233, 17, 241, 17, 249, 17, 1, 18, 9, 18, 17, 18, 25, 18, 33, 18, 41, 18, 49, 18, 48, 0, 55, 18, 63, 18, 71, 18, 79, 18, 87, 18, 95, 18, 48, 0, 103, 18, 111, 18, 119, 18, 127, 18, 135, 18, 142, 18, 150, 18, 158, 18, 48, 0, 164, 18, 172, 18, 180, 18, 188, 18, 196, 18, 204, 18, 211, 18, 219, 18, 227, 18, 235, 18, 243, 18, 249, 18, 0, 19, 8, 19, 16, 19, 48, 0, 48, 0, 48, 0, 23, 19, 31, 19, 39, 19, 47, 19, 55, 19, 63, 19, 71, 19, 79, 19, 86, 19, 94, 19, 48, 0, 48, 0, 102, 19, 110, 19, 118, 19, 126, 19, 134, 19, 142, 19, 150, 19, 158, 19, 166, 19, 172, 19, 180, 19, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 188, 20, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 196, 20, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 188, 19, 196, 19, 204, 19, 212, 19, 220, 19, 228, 19, 236, 19, 244, 19, 252, 19, 4, 20, 12, 20, 20, 20, 28, 20, 36, 20, 44, 20, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 128, 8, 52, 20, 60, 20, 68, 20, 76, 20, 84, 20, 92, 20, 100, 20, 108, 20, 116, 20, 124, 20, 132, 20, 140, 20, 148, 20, 156, 20, 164, 20, 172, 20, 180, 20, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 95, 1, 202, 65, 95, 1, 202, 98, 95, 1, 202, 131, 95, 1, 202, 164, 95, 1, 202, 197, 95, 1, 202, 230, 95, 1, 202, 7, 96, 1, 202, 40, 96, 1, 202, 73, 96, 1, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 94, 1, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 94, 1, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 94, 1, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 0, 0, 198, 99, 0, 0, 192, 0, 0, 0, 198, 195, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 1, 0, 198, 131, 1, 0, 198, 227, 1, 0, 198, 67, 2, 0, 197, 34, 0, 0, 192, 0, 0, 0, 198, 162, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 3, 0, 198, 130, 3, 0, 198, 194, 3, 0, 192, 0, 0, 0, 198, 2, 4, 0, 198, 67, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 4, 0, 198, 2, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 5, 0, 192, 0, 0, 0, 198, 227, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 6, 0, 198, 163, 6, 0, 197, 99, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 0, 0, 198, 3, 7, 0, 192, 0, 0, 0, 198, 98, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 1, 0, 198, 162, 7, 0, 192, 0, 0, 0, 198, 226, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 1, 0, 192, 0, 0, 0, 198, 34, 8, 0, 197, 162, 1, 0, 197, 227, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 8, 0, 192, 0, 0, 0, 198, 163, 8, 0, 198, 3, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 9, 0, 198, 163, 9, 0, 197, 66, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 10, 0, 198, 195, 10, 0, 198, 35, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 11, 0, 198, 227, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 12, 0, 192, 0, 0, 0, 198, 163, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 13, 0, 192, 0, 0, 0, 197, 194, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 13, 0, 198, 226, 13, 0, 192, 0, 0, 0, 198, 34, 14, 0, 198, 98, 14, 0, 198, 163, 14, 0, 198, 3, 15, 0, 197, 2, 3, 0, 192, 0, 0, 0, 198, 98, 15, 0, 198, 163, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 16, 0, 198, 99, 16, 0, 192, 0, 0, 0, 198, 194, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 17, 0, 198, 67, 17, 0, 192, 0, 0, 0, 198, 163, 17, 0, 198, 3, 18, 0, 192, 0, 0, 0, 198, 98, 18, 0, 197, 66, 3, 0, 198, 163, 18, 0, 192, 0, 0, 0, 198, 3, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 19, 0, 198, 195, 19, 0, 198, 35, 20, 0, 198, 131, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 20, 0, 198, 66, 21, 0, 192, 0, 0, 0, 198, 131, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 21, 0, 198, 35, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 22, 0, 198, 227, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 23, 0, 192, 0, 0, 0, 198, 162, 23, 0, 198, 226, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 24, 0, 198, 130, 24, 0, 192, 0, 0, 0, 198, 195, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 25, 0, 198, 98, 25, 0, 197, 130, 3, 0, 198, 163, 25, 0, 198, 2, 26, 0, 198, 67, 26, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 26, 0, 198, 3, 27, 0, 198, 99, 27, 0, 192, 0, 0, 0, 197, 195, 3, 0, 192, 0, 0, 0, 198, 194, 27, 0, 198, 3, 28, 0, 198, 98, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 28, 0, 192, 0, 0, 0, 198, 227, 28, 0, 192, 0, 0, 0, 198, 67, 29, 0, 198, 162, 29, 0, 192, 0, 0, 0, 198, 226, 29, 0, 198, 35, 30, 0, 198, 130, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 31, 0, 198, 131, 31, 0, 198, 226, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 32, 0, 198, 130, 32, 0, 198, 195, 32, 0, 192, 0, 0, 0, 198, 34, 33, 0, 198, 98, 33, 0, 198, 163, 33, 0, 198, 3, 34, 0, 197, 34, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 34, 0, 198, 194, 34, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 35, 0, 192, 0, 0, 0, 198, 99, 35, 0, 192, 0, 0, 0, 198, 195, 35, 0, 192, 0, 0, 0, 198, 34, 36, 0, 192, 0, 0, 0, 198, 98, 36, 0, 198, 162, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 36, 0, 198, 67, 37, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 37, 0, 197, 98, 4, 0, 198, 2, 38, 0, 198, 66, 38, 0, 192, 0, 0, 0, 198, 130, 38, 0, 192, 0, 0, 0, 197, 162, 4, 0, 197, 226, 4, 0, 198, 194, 38, 0, 197, 34, 5, 0, 198, 2, 39, 0, 192, 0, 0, 0, 198, 67, 39, 0, 198, 163, 39, 0, 197, 98, 5, 0, 197, 163, 5, 0, 198, 2, 40, 0, 192, 0, 0, 0, 197, 2, 6, 0, 198, 67, 40, 0, 198, 163, 40, 0, 198, 3, 41, 0, 198, 98, 41, 0, 198, 163, 41, 0, 198, 3, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 6, 0, 197, 130, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 42, 0, 198, 195, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 43, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 43, 0, 192, 0, 0, 0, 198, 195, 43, 0, 197, 194, 6, 0, 198, 35, 44, 0, 198, 130, 44, 0, 198, 194, 44, 0, 198, 2, 45, 0, 198, 67, 45, 0, 198, 162, 45, 0, 192, 0, 0, 0, 198, 227, 45, 0, 192, 0, 0, 0, 198, 67, 46, 0, 198, 163, 46, 0, 192, 0, 0, 0, 198, 2, 47, 0, 192, 0, 0, 0, 197, 2, 7, 0, 198, 66, 47, 0, 198, 131, 47, 0, 198, 226, 47, 0, 192, 0, 0, 0, 198, 34, 48, 0, 192, 0, 0, 0, 198, 98, 48, 0, 192, 0, 0, 0, 197, 66, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 48, 0, 192, 0, 0, 0, 198, 227, 48, 0, 198, 67, 49, 0, 198, 162, 49, 0, 192, 0, 0, 0, 198, 227, 49, 0, 198, 67, 50, 0, 198, 162, 50, 0, 197, 130, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 50, 0, 197, 194, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 51, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 8, 0, 197, 98, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 51, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 51, 0, 192, 0, 0, 0, 198, 66, 52, 0, 198, 130, 52, 0, 198, 195, 52, 0, 198, 34, 53, 0, 198, 98, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 53, 0, 198, 227, 53, 0, 192, 0, 0, 0, 198, 66, 54, 0, 192, 0, 0, 0, 198, 130, 54, 0, 198, 194, 54, 0, 198, 2, 55, 0, 198, 67, 55, 0, 192, 0, 0, 0, 198, 163, 55, 0, 198, 3, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 56, 0, 198, 194, 56, 0, 192, 0, 0, 0, 198, 2, 57, 0, 198, 67, 57, 0, 192, 0, 0, 0, 198, 162, 57, 0, 192, 0, 0, 0, 198, 226, 57, 0, 197, 163, 8, 0, 198, 35, 58, 0, 192, 0, 0, 0, 198, 130, 58, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 58, 0, 198, 2, 59, 0, 197, 2, 9, 0, 198, 66, 59, 0, 192, 0, 0, 0, 198, 130, 59, 0, 198, 195, 59, 0, 197, 66, 9, 0, 198, 35, 60, 0, 192, 0, 0, 0, 198, 131, 60, 0, 192, 0, 0, 0, 198, 227, 60, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 61, 0, 192, 0, 0, 0, 198, 131, 61, 0, 192, 0, 0, 0, 198, 227, 61, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 62, 0, 198, 131, 62, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 62, 0, 192, 0, 0, 0, 198, 66, 63, 0, 192, 0, 0, 0, 198, 130, 63, 0, 192, 0, 0, 0, 198, 195, 63, 0, 192, 0, 0, 0, 198, 35, 64, 0, 198, 130, 64, 0, 192, 0, 0, 0, 198, 195, 64, 0, 192, 0, 0, 0, 198, 35, 65, 0, 192, 0, 0, 0, 198, 131, 65, 0, 197, 130, 9, 0, 197, 194, 9, 0, 192, 0, 0, 0, 198, 227, 65, 0, 192, 0, 0, 0, 198, 67, 66, 0, 198, 162, 66, 0, 192, 0, 0, 0, 198, 227, 66, 0, 198, 67, 67, 0, 198, 163, 67, 0, 197, 2, 10, 0, 198, 3, 68, 0, 198, 99, 68, 0, 198, 195, 68, 0, 198, 34, 69, 0, 198, 99, 69, 0, 197, 66, 10, 0, 198, 195, 69, 0, 197, 131, 10, 0, 197, 226, 10, 0, 198, 34, 70, 0, 192, 0, 0, 0, 198, 99, 70, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 70, 0, 198, 35, 71, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 71, 0, 198, 194, 71, 0, 197, 98, 11, 0, 198, 2, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 11, 0, 198, 67, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 72, 0, 192, 0, 0, 0, 197, 226, 11, 0, 198, 2, 73, 0, 192, 0, 0, 0, 198, 67, 73, 0, 192, 0, 0, 0, 198, 162, 73, 0, 192, 0, 0, 0, 198, 227, 73, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 74, 0, 192, 0, 0, 0, 198, 130, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 75, 0, 192, 0, 0, 0, 198, 130, 75, 0, 192, 0, 0, 0, 198, 195, 75, 0, 198, 35, 76, 0, 198, 130, 76, 0, 192, 0, 0, 0, 197, 130, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 76, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 77, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 77, 0, 192, 0, 0, 0, 198, 163, 77, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 78, 0, 197, 194, 12, 0, 192, 0, 0, 0, 198, 99, 78, 0, 192, 0, 0, 0, 198, 195, 78, 0, 192, 0, 0, 0, 198, 35, 79, 0, 198, 130, 79, 0, 197, 2, 13, 0, 198, 195, 79, 0, 192, 0, 0, 0, 197, 67, 13, 0, 198, 34, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 13, 0, 198, 163, 80, 0, 198, 2, 81, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 81, 0, 198, 131, 81, 0, 198, 226, 81, 0, 198, 34, 82, 0, 192, 0, 0, 0, 198, 99, 82, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 82, 0, 198, 35, 83, 0, 198, 131, 83, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 83, 0, 192, 0, 0, 0, 197, 226, 13, 0, 198, 35, 84, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 84, 0, 198, 227, 84, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 85, 0, 192, 0, 0, 0, 198, 162, 85, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 85, 0, 197, 34, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 86, 0, 198, 130, 86, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 86, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 14, 0, 198, 34, 87, 0, 198, 99, 87, 0, 192, 0, 0, 0, 198, 195, 87, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 88, 0, 198, 99, 88, 0, 198, 195, 88, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 89, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 89, 0, 192, 0, 0, 0, 198, 195, 89, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 90, 0, 198, 98, 90, 0, 197, 226, 14, 0, 198, 163, 90, 0, 198, 3, 91, 0, 198, 99, 91, 0, 198, 195, 91, 0, 192, 0, 0, 0, 198, 35, 92, 0, 198, 131, 92, 0, 192, 0, 0, 0, 198, 227, 92, 0, 192, 0, 0, 0, 198, 67, 93, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 93, 0, 192, 0, 0, 0, 197, 34, 15, 0, 198, 2, 94, 0, 198, 66, 94, 0, 197, 98, 15, 0, 192, 0, 0, 0, 197, 162, 15, 0, 198, 131, 94, 0, 192, 0, 0, 0, 198, 227, 94, 0, 198, 66, 95, 0, 198, 131, 95, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 95, 0, 192, 0, 0, 0, 198, 66, 96, 0, 197, 226, 15, 0, 192, 0, 0, 0, 198, 130, 96, 0, 198, 195, 96, 0, 192, 0, 0, 0, 198, 35, 97, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 97, 0, 197, 34, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 97, 0, 197, 98, 16, 0, 198, 35, 98, 0, 192, 0, 0, 0, 198, 131, 98, 0, 192, 0, 0, 0, 197, 162, 16, 0, 192, 0, 0, 0, 198, 227, 98, 0, 198, 66, 99, 0, 198, 131, 99, 0, 198, 227, 99, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 100, 0, 198, 163, 100, 0, 192, 0, 0, 0, 198, 3, 101, 0, 198, 99, 101, 0, 198, 195, 101, 0, 198, 34, 102, 0, 197, 227, 16, 0, 198, 99, 102, 0, 198, 195, 102, 0, 192, 0, 0, 0, 198, 35, 103, 0, 197, 66, 17, 0, 198, 131, 103, 0, 198, 226, 103, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 104, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 104, 0, 192, 0, 0, 0, 198, 163, 104, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 105, 0, 192, 0, 0, 0, 198, 98, 105, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 105, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 106, 0, 192, 0, 0, 0, 198, 98, 106, 0, 198, 163, 106, 0, 198, 3, 107, 0, 197, 130, 17, 0, 198, 99, 107, 0, 198, 195, 107, 0, 198, 35, 108, 0, 197, 194, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 108, 0, 198, 66, 109, 0, 198, 130, 109, 0, 198, 195, 109, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 110, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 110, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 110, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 111, 0, 198, 131, 111, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 111, 0, 192, 0, 0, 0, 198, 35, 112, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 112, 0, 198, 194, 112, 0, 192, 0, 0, 0, 198, 3, 113, 0, 192, 0, 0, 0, 198, 99, 113, 0, 197, 2, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 113, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 18, 0, 198, 35, 114, 0, 198, 131, 114, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 114, 0, 192, 0, 0, 0, 198, 67, 115, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 115, 0, 197, 130, 18, 0, 198, 3, 116, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 18, 0, 198, 99, 116, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 116, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 19, 0, 198, 34, 117, 0, 198, 99, 117, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 117, 0, 198, 35, 118, 0, 198, 130, 118, 0, 198, 195, 118, 0, 198, 35, 119, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 19, 0, 192, 0, 0, 0, 198, 131, 119, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 119, 0, 198, 67, 120, 0, 198, 163, 120, 0, 197, 130, 19, 0, 192, 0, 0, 0, 197, 194, 19, 0, 198, 3, 121, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 121, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 121, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 20, 0, 198, 34, 122, 0, 192, 0, 0, 0, 198, 99, 122, 0, 198, 195, 122, 0, 192, 0, 0, 0, 197, 99, 20, 0, 192, 0, 0, 0, 198, 35, 123, 0, 197, 194, 20, 0, 198, 131, 123, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 123, 0, 198, 34, 124, 0, 192, 0, 0, 0, 198, 99, 124, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 124, 0, 198, 35, 125, 0, 197, 98, 21, 0, 192, 0, 0, 0, 198, 131, 125, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 125, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 126, 0, 192, 0, 0, 0, 198, 163, 126, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 127, 0, 198, 99, 127, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 127, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 128, 0, 198, 99, 128, 0, 198, 195, 128, 0, 198, 35, 129, 0, 192, 0, 0, 0, 198, 130, 129, 0, 198, 195, 129, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 130, 0, 192, 0, 0, 0, 198, 130, 130, 0, 198, 195, 130, 0, 198, 34, 131, 0, 198, 98, 131, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 21, 0, 197, 226, 21, 0, 198, 163, 131, 0, 198, 3, 132, 0, 198, 98, 132, 0, 198, 163, 132, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 22, 0, 198, 3, 133, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 133, 0, 192, 0, 0, 0, 197, 98, 22, 0, 198, 195, 133, 0, 198, 35, 134, 0, 198, 131, 134, 0, 198, 227, 134, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 135, 0, 198, 162, 135, 0, 198, 226, 135, 0, 198, 35, 136, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 136, 0, 198, 195, 136, 0, 192, 0, 0, 0, 198, 35, 137, 0, 198, 130, 137, 0, 198, 195, 137, 0, 192, 0, 0, 0, 198, 35, 138, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 138, 0, 198, 227, 138, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 139, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 22, 0, 192, 0, 0, 0, 198, 130, 139, 0, 198, 195, 139, 0, 192, 0, 0, 0, 198, 35, 140, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 140, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 23, 0, 192, 0, 0, 0, 198, 194, 140, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 23, 0, 197, 131, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 141, 0, 198, 99, 141, 0, 198, 194, 141, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 142, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 23, 0, 198, 99, 142, 0, 197, 34, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 142, 0, 198, 34, 143, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 143, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 143, 0, 198, 34, 144, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 144, 0, 198, 195, 144, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 145, 0, 198, 98, 145, 0, 198, 163, 145, 0, 198, 2, 146, 0, 198, 66, 146, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 146, 0, 192, 0, 0, 0, 198, 227, 146, 0, 192, 0, 0, 0, 198, 67, 147, 0, 192, 0, 0, 0, 198, 163, 147, 0, 192, 0, 0, 0, 198, 3, 148, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 148, 0, 198, 195, 148, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 149, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 149, 0, 198, 227, 149, 0, 192, 0, 0, 0, 198, 67, 150, 0, 198, 162, 150, 0, 192, 0, 0, 0, 197, 195, 24, 0, 192, 0, 0, 0, 198, 227, 150, 0, 198, 66, 151, 0, 198, 130, 151, 0, 198, 195, 151, 0, 197, 34, 25, 0, 198, 34, 152, 0, 198, 99, 152, 0, 198, 195, 152, 0, 192, 0, 0, 0, 198, 34, 153, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 25, 0, 192, 0, 0, 0, 197, 194, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 153, 0, 192, 0, 0, 0, 198, 195, 153, 0, 198, 35, 154, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 154, 0, 198, 195, 154, 0, 192, 0, 0, 0, 198, 35, 155, 0, 198, 130, 155, 0, 198, 194, 155, 0, 192, 0, 0, 0, 198, 2, 156, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 26, 0, 192, 0, 0, 0, 198, 67, 156, 0, 192, 0, 0, 0, 197, 99, 26, 0, 198, 162, 156, 0, 192, 0, 0, 0, 198, 226, 156, 0, 198, 35, 157, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 157, 0, 197, 194, 26, 0, 198, 195, 157, 0, 192, 0, 0, 0, 198, 34, 158, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 158, 0, 198, 195, 158, 0, 192, 0, 0, 0, 198, 35, 159, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 159, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 27, 0, 192, 0, 0, 0, 197, 162, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 159, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 160, 0, 198, 99, 160, 0, 198, 195, 160, 0, 198, 34, 161, 0, 198, 99, 161, 0, 192, 0, 0, 0, 198, 195, 161, 0, 198, 35, 162, 0, 198, 131, 162, 0, 192, 0, 0, 0, 198, 227, 162, 0, 198, 66, 163, 0, 197, 66, 28, 0, 198, 131, 163, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 163, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 164, 0, 198, 131, 164, 0, 197, 131, 28, 0, 197, 227, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 164, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 165, 0, 192, 0, 0, 0, 198, 163, 165, 0, 192, 0, 0, 0, 198, 3, 166, 0, 192, 0, 0, 0, 198, 99, 166, 0, 198, 195, 166, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 29, 0, 192, 0, 0, 0, 198, 35, 167, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 167, 0, 192, 0, 0, 0, 197, 130, 29, 0, 198, 226, 167, 0, 198, 35, 168, 0, 197, 194, 29, 0, 198, 130, 168, 0, 198, 195, 168, 0, 198, 35, 169, 0, 198, 131, 169, 0, 192, 0, 0, 0, 198, 227, 169, 0, 198, 66, 170, 0, 197, 3, 30, 0, 198, 131, 170, 0, 198, 227, 170, 0, 192, 0, 0, 0, 198, 67, 171, 0, 198, 163, 171, 0, 198, 3, 172, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 172, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 172, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 173, 0, 192, 0, 0, 0, 198, 98, 173, 0, 192, 0, 0, 0, 197, 98, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 173, 0, 192, 0, 0, 0, 198, 227, 173, 0, 198, 67, 174, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 174, 0, 197, 162, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 175, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 175, 0, 198, 194, 175, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 176, 0, 192, 0, 0, 0, 198, 66, 176, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 176, 0, 192, 0, 0, 0, 198, 227, 176, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 177, 0, 198, 130, 177, 0, 198, 195, 177, 0, 198, 34, 178, 0, 192, 0, 0, 0, 198, 99, 178, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 178, 0, 192, 0, 0, 0, 197, 34, 31, 0, 198, 34, 179, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 179, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 179, 0, 192, 0, 0, 0, 197, 98, 31, 0, 198, 3, 180, 0, 198, 98, 180, 0, 198, 162, 180, 0, 198, 226, 180, 0, 198, 35, 181, 0, 192, 0, 0, 0, 198, 130, 181, 0, 198, 194, 181, 0, 197, 162, 31, 0, 198, 2, 182, 0, 198, 66, 182, 0, 198, 131, 182, 0, 198, 227, 182, 0, 198, 66, 183, 0, 197, 226, 31, 0, 198, 131, 183, 0, 198, 226, 183, 0, 198, 34, 184, 0, 198, 98, 184, 0, 197, 34, 32, 0, 197, 99, 32, 0, 192, 0, 0, 0, 198, 162, 184, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 184, 0, 192, 0, 0, 0, 197, 195, 32, 0, 198, 35, 185, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 33, 0, 198, 131, 185, 0, 198, 227, 185, 0, 198, 67, 186, 0, 198, 163, 186, 0, 198, 3, 187, 0, 197, 130, 33, 0, 198, 98, 187, 0, 198, 162, 187, 0, 198, 227, 187, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 188, 0, 192, 0, 0, 0, 198, 163, 188, 0, 192, 0, 0, 0, 198, 3, 189, 0, 197, 194, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 189, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 189, 0, 192, 0, 0, 0, 197, 3, 34, 0, 198, 3, 190, 0, 192, 0, 0, 0, 198, 99, 190, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 190, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 191, 0, 192, 0, 0, 0, 197, 99, 34, 0, 197, 194, 34, 0, 192, 0, 0, 0, 198, 99, 191, 0, 197, 2, 35, 0, 197, 66, 35, 0, 198, 194, 191, 0, 192, 0, 0, 0, 197, 130, 35, 0, 192, 0, 0, 0, 197, 194, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 192, 0, 192, 0, 0, 0, 198, 67, 192, 0, 192, 0, 0, 0, 198, 162, 192, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 192, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 193, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 193, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 194, 0, 192, 0, 0, 0, 198, 67, 194, 0, 192, 0, 0, 0, 198, 162, 194, 0, 192, 0, 0, 0, 198, 226, 194, 0, 192, 0, 0, 0, 198, 34, 195, 0, 198, 99, 195, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 195, 0, 198, 2, 196, 0, 192, 0, 0, 0, 198, 66, 196, 0, 192, 0, 0, 0, 198, 130, 196, 0, 198, 194, 196, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 197, 0, 197, 2, 36, 0, 198, 99, 197, 0, 192, 0, 0, 0, 198, 194, 197, 0, 197, 66, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 198, 0, 192, 0, 0, 0, 198, 98, 198, 0, 197, 131, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 198, 0, 197, 226, 36, 0, 197, 34, 37, 0, 198, 226, 198, 0, 198, 35, 199, 0, 198, 130, 199, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 199, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 200, 0, 192, 0, 0, 0, 198, 99, 200, 0, 192, 0, 0, 0, 197, 98, 37, 0, 197, 162, 37, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 200, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 37, 0, 198, 2, 201, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 201, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 38, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 38, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 38, 0, 197, 2, 39, 0, 198, 163, 201, 0, 192, 0, 0, 0, 198, 2, 202, 0, 198, 66, 202, 0, 197, 67, 39, 0, 198, 131, 202, 0, 198, 227, 202, 0, 197, 163, 39, 0, 192, 0, 0, 0, 197, 2, 40, 0, 197, 66, 40, 0, 197, 131, 40, 0, 198, 66, 203, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 203, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 203, 0, 198, 34, 204, 0, 192, 0, 0, 0, 198, 98, 204, 0, 198, 162, 204, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 204, 0, 197, 226, 40, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 205, 0, 198, 163, 205, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 206, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 206, 0, 192, 0, 0, 0, 198, 162, 206, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 41, 0, 192, 0, 0, 0, 198, 226, 206, 0, 198, 35, 207, 0, 192, 0, 0, 0, 198, 130, 207, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 207, 0, 197, 98, 41, 0, 198, 3, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 209, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 209, 0, 197, 162, 41, 0, 192, 0, 0, 0, 198, 163, 209, 0, 192, 0, 0, 0, 197, 226, 41, 0, 192, 0, 0, 0, 198, 3, 210, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 210, 0, 192, 0, 0, 0, 197, 34, 42, 0, 192, 0, 0, 0, 198, 194, 210, 0, 192, 0, 0, 0, 198, 2, 211, 0, 192, 0, 0, 0, 197, 99, 42, 0, 192, 0, 0, 0, 198, 67, 211, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 211, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 212, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 212, 0, 198, 163, 212, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 213, 0, 192, 0, 0, 0, 198, 98, 213, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 213, 0, 197, 34, 43, 0, 198, 227, 213, 0, 198, 66, 214, 0, 192, 0, 0, 0, 198, 131, 214, 0, 198, 227, 214, 0, 192, 0, 0, 0, 198, 67, 215, 0, 198, 163, 215, 0, 192, 0, 0, 0, 198, 3, 216, 0, 198, 98, 216, 0, 192, 0, 0, 0, 197, 98, 43, 0, 197, 163, 43, 0, 198, 163, 216, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 217, 0, 192, 0, 0, 0, 198, 98, 217, 0, 198, 163, 217, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 218, 0, 197, 2, 44, 0, 197, 67, 44, 0, 192, 0, 0, 0, 197, 162, 44, 0, 198, 99, 218, 0, 198, 195, 218, 0, 197, 226, 44, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 45, 0, 197, 99, 45, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 45, 0, 198, 35, 219, 0, 197, 34, 46, 0, 198, 131, 219, 0, 198, 227, 219, 0, 198, 66, 220, 0, 192, 0, 0, 0, 198, 131, 220, 0, 192, 0, 0, 0, 198, 227, 220, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 46, 0, 192, 0, 0, 0, 198, 67, 221, 0, 192, 0, 0, 0, 198, 162, 221, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 46, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 221, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 222, 0, 198, 131, 222, 0, 198, 226, 222, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 223, 0, 198, 99, 223, 0, 192, 0, 0, 0, 198, 195, 223, 0, 198, 34, 224, 0, 197, 2, 47, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 224, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 224, 0, 198, 226, 224, 0, 197, 67, 47, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 225, 0, 192, 0, 0, 0, 198, 99, 225, 0, 192, 0, 0, 0, 198, 194, 225, 0, 198, 2, 226, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 47, 0, 198, 66, 226, 0, 198, 131, 226, 0, 192, 0, 0, 0, 198, 226, 226, 0, 192, 0, 0, 0, 197, 226, 47, 0, 192, 0, 0, 0, 198, 34, 227, 0, 192, 0, 0, 0, 198, 99, 227, 0, 198, 195, 227, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 48, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 48, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 228, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 228, 0, 192, 0, 0, 0, 197, 162, 48, 0, 192, 0, 0, 0, 198, 162, 228, 0, 197, 226, 48, 0, 198, 226, 228, 0, 198, 34, 229, 0, 192, 0, 0, 0, 197, 34, 49, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 229, 0, 192, 0, 0, 0, 198, 194, 229, 0, 198, 3, 230, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 230, 0, 192, 0, 0, 0, 198, 195, 230, 0, 192, 0, 0, 0, 198, 34, 231, 0, 198, 99, 231, 0, 197, 99, 49, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 49, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 231, 0, 198, 35, 232, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 232, 0, 198, 195, 232, 0, 198, 34, 233, 0, 197, 2, 50, 0, 198, 98, 233, 0, 198, 162, 233, 0, 192, 0, 0, 0, 198, 226, 233, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 50, 0, 192, 0, 0, 0, 198, 35, 234, 0, 192, 0, 0, 0, 198, 131, 234, 0, 192, 0, 0, 0, 198, 227, 234, 0, 192, 0, 0, 0, 198, 67, 235, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 235, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 50, 0, 192, 0, 0, 0, 197, 226, 50, 0, 197, 34, 51, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 51, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 236, 0, 198, 67, 236, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 236, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 237, 0, 192, 0, 0, 0, 198, 99, 237, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 237, 0, 198, 35, 238, 0, 192, 0, 0, 0, 198, 131, 238, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 238, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 239, 0, 198, 163, 239, 0, 197, 162, 51, 0, 198, 3, 240, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 51, 0, 198, 98, 240, 0, 198, 163, 240, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 52, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 241, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 241, 0, 198, 162, 241, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 52, 0, 198, 227, 241, 0, 198, 67, 242, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 242, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 52, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 53, 0, 192, 0, 0, 0, 198, 3, 243, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 243, 0, 192, 0, 0, 0, 198, 195, 243, 0, 198, 34, 244, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 244, 0, 192, 0, 0, 0, 198, 163, 244, 0, 198, 3, 245, 0, 198, 98, 245, 0, 198, 162, 245, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 245, 0, 198, 66, 246, 0, 198, 131, 246, 0, 198, 227, 246, 0, 192, 0, 0, 0, 197, 66, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 247, 0, 198, 163, 247, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 248, 0, 192, 0, 0, 0, 198, 67, 248, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 248, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 249, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 249, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 249, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 249, 0, 198, 34, 250, 0, 198, 98, 250, 0, 192, 0, 0, 0, 198, 162, 250, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 250, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 251, 0, 192, 0, 0, 0, 198, 130, 251, 0, 192, 0, 0, 0, 198, 195, 251, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 252, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 252, 0, 192, 0, 0, 0, 198, 227, 252, 0, 198, 67, 253, 0, 198, 163, 253, 0, 198, 2, 254, 0, 197, 130, 53, 0, 198, 67, 254, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 254, 0, 192, 0, 0, 0, 198, 3, 255, 0, 198, 99, 255, 0, 198, 194, 255, 0, 198, 3, 0, 1, 198, 99, 0, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 0, 1, 198, 2, 1, 1, 198, 67, 1, 1, 198, 163, 1, 1, 192, 0, 0, 0, 198, 2, 2, 1, 192, 0, 0, 0, 198, 66, 2, 1, 192, 0, 0, 0, 198, 130, 2, 1, 198, 194, 2, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 3, 1, 197, 3, 54, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 3, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 3, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 3, 1, 198, 35, 4, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 54, 0, 192, 0, 0, 0, 197, 162, 54, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 4, 1, 192, 0, 0, 0, 198, 195, 4, 1, 192, 0, 0, 0, 198, 34, 5, 1, 198, 99, 5, 1, 198, 195, 5, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 6, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 6, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 6, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 7, 1, 198, 131, 7, 1, 198, 227, 7, 1, 197, 227, 54, 0, 192, 0, 0, 0, 198, 66, 8, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 8, 1, 198, 227, 8, 1, 192, 0, 0, 0, 198, 67, 9, 1, 198, 163, 9, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 55, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 10, 1, 198, 98, 10, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 10, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 10, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 11, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 11, 1, 198, 195, 11, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 12, 1, 192, 0, 0, 0, 198, 130, 12, 1, 198, 194, 12, 1, 198, 3, 13, 1, 198, 98, 13, 1, 198, 162, 13, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 13, 1, 192, 0, 0, 0, 198, 67, 14, 1, 198, 163, 14, 1, 192, 0, 0, 0, 198, 3, 15, 1, 197, 130, 55, 0, 192, 0, 0, 0, 198, 99, 15, 1, 198, 195, 15, 1, 198, 35, 16, 1, 198, 131, 16, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 16, 1, 192, 0, 0, 0, 198, 67, 17, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 17, 1, 197, 195, 55, 0, 192, 0, 0, 0, 197, 34, 56, 0, 198, 226, 17, 1, 192, 0, 0, 0, 198, 34, 18, 1, 197, 98, 56, 0, 192, 0, 0, 0, 198, 98, 18, 1, 198, 163, 18, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 19, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 56, 0, 198, 99, 19, 1, 198, 194, 19, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 20, 1, 198, 98, 20, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 20, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 20, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 21, 1, 198, 162, 21, 1, 198, 227, 21, 1, 197, 227, 56, 0, 198, 67, 22, 1, 192, 0, 0, 0, 198, 162, 22, 1, 198, 226, 22, 1, 198, 34, 23, 1, 198, 98, 23, 1, 192, 0, 0, 0, 197, 67, 57, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 23, 1, 198, 227, 23, 1, 198, 66, 24, 1, 198, 131, 24, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 24, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 25, 1, 198, 130, 25, 1, 192, 0, 0, 0, 197, 163, 57, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 25, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 26, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 26, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 26, 1, 198, 67, 27, 1, 198, 163, 27, 1, 192, 0, 0, 0, 198, 2, 28, 1, 198, 67, 28, 1, 192, 0, 0, 0, 198, 163, 28, 1, 192, 0, 0, 0, 198, 2, 29, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 29, 1, 198, 163, 29, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 30, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 30, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 30, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 58, 0, 198, 226, 30, 1, 198, 35, 31, 1, 197, 66, 58, 0, 192, 0, 0, 0, 198, 131, 31, 1, 192, 0, 0, 0, 198, 227, 31, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 32, 1, 198, 163, 32, 1, 192, 0, 0, 0, 198, 3, 33, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 58, 0, 192, 0, 0, 0, 198, 99, 33, 1, 198, 195, 33, 1, 192, 0, 0, 0, 198, 35, 34, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 34, 1, 198, 195, 34, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 35, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 35, 1, 192, 0, 0, 0, 198, 227, 35, 1, 192, 0, 0, 0, 198, 67, 36, 1, 192, 0, 0, 0, 198, 163, 36, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 58, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 59, 0, 192, 0, 0, 0, 198, 3, 37, 1, 198, 98, 37, 1, 197, 130, 59, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 37, 1, 198, 227, 37, 1, 198, 66, 38, 1, 192, 0, 0, 0, 198, 131, 38, 1, 198, 226, 38, 1, 198, 34, 39, 1, 198, 99, 39, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 39, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 40, 1, 198, 99, 40, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 40, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 41, 1, 197, 194, 59, 0, 198, 130, 41, 1, 192, 0, 0, 0, 198, 194, 41, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 42, 1, 192, 0, 0, 0, 198, 98, 42, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 60, 0, 198, 163, 42, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 43, 1, 198, 67, 43, 1, 197, 66, 60, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 43, 1, 198, 2, 44, 1, 192, 0, 0, 0, 197, 130, 60, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 44, 1, 192, 0, 0, 0, 198, 163, 44, 1, 198, 2, 45, 1, 192, 0, 0, 0, 198, 66, 45, 1, 192, 0, 0, 0, 198, 131, 45, 1, 197, 194, 60, 0, 198, 226, 45, 1, 197, 2, 61, 0, 197, 66, 61, 0, 198, 35, 46, 1, 198, 131, 46, 1, 192, 0, 0, 0, 198, 227, 46, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 47, 1, 198, 130, 47, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 61, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 47, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 48, 1, 198, 98, 48, 1, 198, 163, 48, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 49, 1, 198, 66, 49, 1, 197, 227, 61, 0, 198, 130, 49, 1, 192, 0, 0, 0, 198, 194, 49, 1, 198, 3, 50, 1, 197, 67, 62, 0, 192, 0, 0, 0, 197, 162, 62, 0, 198, 99, 50, 1, 192, 0, 0, 0, 197, 226, 62, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 50, 1, 192, 0, 0, 0, 198, 34, 51, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 51, 1, 192, 0, 0, 0, 198, 163, 51, 1, 198, 3, 52, 1, 198, 99, 52, 1, 198, 195, 52, 1, 198, 35, 53, 1, 198, 131, 53, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 63, 0, 198, 226, 53, 1, 192, 0, 0, 0, 198, 34, 54, 1, 192, 0, 0, 0, 198, 99, 54, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 63, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 63, 0, 198, 194, 54, 1, 198, 3, 55, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 55, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 63, 0, 197, 34, 64, 0, 198, 194, 55, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 64, 0, 198, 3, 56, 1, 192, 0, 0, 0, 198, 98, 56, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 56, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 57, 1, 192, 0, 0, 0, 198, 66, 57, 1, 192, 0, 0, 0, 198, 131, 57, 1, 198, 227, 57, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 64, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 58, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 58, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 59, 1, 192, 0, 0, 0, 197, 226, 64, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 59, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 59, 1, 192, 0, 0, 0, 198, 227, 59, 1, 198, 66, 60, 1, 192, 0, 0, 0, 198, 130, 60, 1, 192, 0, 0, 0, 198, 195, 60, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 65, 0, 198, 34, 61, 1, 198, 98, 61, 1, 198, 163, 61, 1, 197, 194, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 62, 1, 197, 2, 66, 0, 192, 0, 0, 0, 198, 99, 62, 1, 198, 194, 62, 1, 198, 3, 63, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 63, 1, 198, 162, 63, 1, 198, 226, 63, 1, 198, 35, 64, 1, 192, 0, 0, 0, 198, 130, 64, 1, 192, 0, 0, 0, 198, 195, 64, 1, 192, 0, 0, 0, 197, 66, 66, 0, 192, 0, 0, 0, 198, 35, 65, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 65, 1, 192, 0, 0, 0, 198, 227, 65, 1, 198, 67, 66, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 66, 1, 192, 0, 0, 0, 197, 130, 66, 0, 192, 0, 0, 0, 197, 194, 66, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 66, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 67, 1, 192, 0, 0, 0, 198, 130, 67, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 67, 1, 198, 35, 68, 1, 192, 0, 0, 0, 198, 130, 68, 1, 198, 194, 68, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 69, 1, 198, 99, 69, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 67, 0, 192, 0, 0, 0, 198, 194, 69, 1, 192, 0, 0, 0, 198, 2, 70, 1, 192, 0, 0, 0, 198, 67, 70, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 70, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 70, 1, 197, 66, 67, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 71, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 71, 1, 198, 194, 71, 1, 198, 2, 72, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 67, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 72, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 72, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 67, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 73, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 73, 1, 198, 163, 73, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 74, 1, 198, 67, 74, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 74, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 68, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 75, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 75, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 75, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 75, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 76, 1, 197, 67, 68, 0, 192, 0, 0, 0, 198, 99, 76, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 76, 1, 198, 3, 77, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 77, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 77, 1, 192, 0, 0, 0, 198, 2, 78, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 68, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 78, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 78, 1, 192, 0, 0, 0, 198, 227, 78, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 79, 1, 198, 131, 79, 1, 198, 227, 79, 1, 198, 66, 80, 1, 198, 130, 80, 1, 198, 194, 80, 1, 192, 0, 0, 0, 198, 2, 81, 1, 198, 66, 81, 1, 192, 0, 0, 0, 198, 131, 81, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 81, 1, 198, 67, 82, 1, 192, 0, 0, 0, 198, 162, 82, 1, 197, 227, 68, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 69, 0, 198, 226, 82, 1, 192, 0, 0, 0, 198, 35, 83, 1, 198, 131, 83, 1, 197, 130, 69, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 83, 1, 198, 35, 84, 1, 198, 130, 84, 1, 198, 195, 84, 1, 192, 0, 0, 0, 198, 35, 85, 1, 198, 131, 85, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 85, 1, 192, 0, 0, 0, 198, 66, 86, 1, 192, 0, 0, 0, 197, 194, 69, 0, 197, 2, 70, 0, 198, 130, 86, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 86, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 87, 1, 198, 99, 87, 1, 192, 0, 0, 0, 198, 194, 87, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 88, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 70, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 88, 1, 197, 130, 70, 0, 192, 0, 0, 0, 198, 131, 88, 1, 198, 226, 88, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 70, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 89, 1, 192, 0, 0, 0, 198, 131, 89, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 89, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 90, 1, 192, 0, 0, 0, 198, 99, 90, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 90, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 91, 1, 192, 0, 0, 0, 198, 98, 91, 1, 192, 0, 0, 0, 198, 163, 91, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 92, 1, 198, 99, 92, 1, 198, 195, 92, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 93, 1, 192, 0, 0, 0, 198, 131, 93, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 93, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 94, 1, 197, 3, 71, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 94, 1, 198, 227, 94, 1, 198, 66, 95, 1, 198, 130, 95, 1, 198, 195, 95, 1, 192, 0, 0, 0, 198, 34, 96, 1, 197, 98, 71, 0, 198, 98, 96, 1, 197, 162, 71, 0, 192, 0, 0, 0, 198, 163, 96, 1, 192, 0, 0, 0, 198, 2, 97, 1, 192, 0, 0, 0, 198, 66, 97, 1, 198, 130, 97, 1, 198, 194, 97, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 71, 0, 192, 0, 0, 0, 198, 3, 98, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 98, 1, 198, 195, 98, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 99, 1, 198, 98, 99, 1, 192, 0, 0, 0, 197, 35, 72, 0, 192, 0, 0, 0, 197, 130, 72, 0, 192, 0, 0, 0, 198, 163, 99, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 100, 1, 192, 0, 0, 0, 197, 194, 72, 0, 198, 99, 100, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 100, 1, 192, 0, 0, 0, 198, 34, 101, 1, 198, 98, 101, 1, 198, 163, 101, 1, 198, 2, 102, 1, 198, 67, 102, 1, 198, 163, 102, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 103, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 103, 1, 192, 0, 0, 0, 198, 195, 103, 1, 197, 3, 73, 0, 198, 35, 104, 1, 198, 131, 104, 1, 198, 226, 104, 1, 198, 35, 105, 1, 198, 131, 105, 1, 197, 98, 73, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 105, 1, 198, 67, 106, 1, 198, 163, 106, 1, 197, 162, 73, 0, 197, 227, 73, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 107, 1, 192, 0, 0, 0, 198, 99, 107, 1, 192, 0, 0, 0, 198, 194, 107, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 74, 0, 198, 3, 108, 1, 198, 98, 108, 1, 198, 162, 108, 1, 192, 0, 0, 0, 198, 227, 108, 1, 198, 67, 109, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 109, 1, 198, 2, 110, 1, 192, 0, 0, 0, 198, 67, 110, 1, 198, 163, 110, 1, 198, 2, 111, 1, 198, 67, 111, 1, 198, 163, 111, 1, 198, 2, 112, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 112, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 112, 1, 192, 0, 0, 0, 198, 195, 112, 1, 198, 34, 113, 1, 198, 99, 113, 1, 192, 0, 0, 0, 197, 131, 74, 0, 198, 195, 113, 1, 197, 226, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 114, 1, 198, 131, 114, 1, 192, 0, 0, 0, 197, 34, 75, 0, 197, 99, 75, 0, 192, 0, 0, 0, 198, 226, 114, 1, 197, 194, 75, 0, 198, 35, 115, 1, 198, 131, 115, 1, 198, 226, 115, 1, 198, 35, 116, 1, 192, 0, 0, 0, 198, 130, 116, 1, 198, 195, 116, 1, 197, 2, 76, 0, 198, 34, 117, 1, 198, 99, 117, 1, 197, 67, 76, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 117, 1, 198, 34, 118, 1, 192, 0, 0, 0, 198, 99, 118, 1, 192, 0, 0, 0, 198, 194, 118, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 119, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 119, 1, 198, 163, 119, 1, 192, 0, 0, 0, 198, 3, 120, 1, 198, 99, 120, 1, 197, 162, 76, 0, 198, 195, 120, 1, 198, 35, 121, 1, 198, 130, 121, 1, 198, 194, 121, 1, 198, 2, 122, 1, 198, 66, 122, 1, 192, 0, 0, 0, 198, 130, 122, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 122, 1, 192, 0, 0, 0, 198, 3, 123, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 123, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 123, 1, 198, 35, 124, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 124, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 124, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 125, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 76, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 125, 1, 198, 195, 125, 1, 198, 35, 126, 1, 198, 130, 126, 1, 198, 194, 126, 1, 192, 0, 0, 0, 198, 2, 127, 1, 198, 66, 127, 1, 198, 131, 127, 1, 198, 226, 127, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 128, 1, 198, 98, 128, 1, 198, 163, 128, 1, 198, 3, 129, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 129, 1, 192, 0, 0, 0, 198, 194, 129, 1, 198, 2, 130, 1, 192, 0, 0, 0, 198, 67, 130, 1, 198, 163, 130, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 131, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 131, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 131, 1, 192, 0, 0, 0, 198, 2, 132, 1, 192, 0, 0, 0, 198, 67, 132, 1, 198, 162, 132, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 132, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 133, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 133, 1, 192, 0, 0, 0, 198, 227, 133, 1, 192, 0, 0, 0, 198, 67, 134, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 77, 0, 198, 162, 134, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 77, 0, 192, 0, 0, 0, 198, 227, 134, 1, 192, 0, 0, 0, 198, 67, 135, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 135, 1, 198, 2, 136, 1, 198, 67, 136, 1, 198, 163, 136, 1, 192, 0, 0, 0, 198, 2, 137, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 137, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 137, 1, 192, 0, 0, 0, 198, 2, 138, 1, 192, 0, 0, 0, 197, 194, 77, 0, 198, 67, 138, 1, 198, 162, 138, 1, 197, 2, 78, 0, 198, 227, 138, 1, 198, 67, 139, 1, 198, 162, 139, 1, 198, 227, 139, 1, 198, 67, 140, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 140, 1, 192, 0, 0, 0, 198, 227, 140, 1, 192, 0, 0, 0, 197, 66, 78, 0, 192, 0, 0, 0, 198, 67, 141, 1, 198, 163, 141, 1, 192, 0, 0, 0, 198, 3, 142, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 142, 1, 192, 0, 0, 0, 198, 194, 142, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 143, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 78, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 143, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 143, 1, 198, 2, 144, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 144, 1, 192, 0, 0, 0, 198, 131, 144, 1, 198, 227, 144, 1, 198, 66, 145, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 145, 1, 192, 0, 0, 0, 198, 226, 145, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 146, 1, 192, 0, 0, 0, 197, 195, 78, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 146, 1, 198, 227, 146, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 147, 1, 198, 130, 147, 1, 192, 0, 0, 0, 198, 195, 147, 1, 198, 35, 148, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 79, 0, 198, 130, 148, 1, 198, 194, 148, 1, 198, 3, 149, 1, 192, 0, 0, 0, 198, 99, 149, 1, 192, 0, 0, 0, 198, 195, 149, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 150, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 150, 1, 192, 0, 0, 0, 197, 98, 79, 0, 192, 0, 0, 0, 198, 227, 150, 1, 198, 67, 151, 1, 192, 0, 0, 0, 198, 163, 151, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 79, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 152, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 152, 1, 198, 163, 152, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 153, 1, 198, 67, 153, 1, 192, 0, 0, 0, 198, 162, 153, 1, 192, 0, 0, 0, 198, 227, 153, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 154, 1, 197, 226, 79, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 154, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 155, 1, 192, 0, 0, 0, 198, 67, 155, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 155, 1, 198, 227, 155, 1, 192, 0, 0, 0, 198, 67, 156, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 156, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 156, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 157, 1, 192, 0, 0, 0, 198, 131, 157, 1, 198, 226, 157, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 158, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 158, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 158, 1, 192, 0, 0, 0, 197, 98, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 159, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 159, 1, 198, 227, 159, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 80, 0, 192, 0, 0, 0, 197, 2, 81, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 81, 0, 198, 67, 160, 1, 192, 0, 0, 0, 198, 163, 160, 1, 192, 0, 0, 0, 198, 2, 161, 1, 198, 67, 161, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 161, 1, 192, 0, 0, 0, 198, 2, 162, 1, 192, 0, 0, 0, 197, 130, 81, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 162, 1, 198, 163, 162, 1, 198, 3, 163, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 163, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 163, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 164, 1, 198, 130, 164, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 164, 1, 198, 34, 165, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 165, 1, 197, 194, 81, 0, 198, 194, 165, 1, 198, 2, 166, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 166, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 166, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 166, 1, 198, 2, 167, 1, 198, 66, 167, 1, 197, 2, 82, 0, 197, 66, 82, 0, 192, 0, 0, 0, 198, 131, 167, 1, 192, 0, 0, 0, 198, 227, 167, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 168, 1, 198, 163, 168, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 169, 1, 198, 66, 169, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 169, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 169, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 82, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 170, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 170, 1, 197, 195, 82, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 171, 1, 198, 98, 171, 1, 198, 163, 171, 1, 192, 0, 0, 0, 197, 34, 83, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 172, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 172, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 172, 1, 192, 0, 0, 0, 198, 195, 172, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 173, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 173, 1, 198, 194, 173, 1, 198, 3, 174, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 174, 1, 192, 0, 0, 0, 197, 98, 83, 0, 198, 163, 174, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 83, 0, 198, 3, 175, 1, 197, 227, 83, 0, 192, 0, 0, 0, 198, 99, 175, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 84, 0, 192, 0, 0, 0, 198, 195, 175, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 176, 1, 198, 131, 176, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 176, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 177, 1, 197, 162, 84, 0, 192, 0, 0, 0, 198, 163, 177, 1, 198, 3, 178, 1, 197, 226, 84, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 178, 1, 192, 0, 0, 0, 198, 195, 178, 1, 198, 35, 179, 1, 198, 130, 179, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 179, 1, 198, 35, 180, 1, 198, 131, 180, 1, 198, 227, 180, 1, 198, 67, 181, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 181, 1, 198, 3, 182, 1, 198, 99, 182, 1, 198, 194, 182, 1, 197, 35, 85, 0, 197, 130, 85, 0, 197, 194, 85, 0, 198, 2, 183, 1, 198, 67, 183, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 183, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 184, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 184, 1, 198, 162, 184, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 184, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 86, 0, 198, 34, 185, 1, 198, 99, 185, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 185, 1, 192, 0, 0, 0, 198, 34, 186, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 186, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 186, 1, 197, 66, 86, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 187, 1, 192, 0, 0, 0, 198, 67, 187, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 187, 1, 192, 0, 0, 0, 198, 227, 187, 1, 198, 66, 188, 1, 198, 131, 188, 1, 198, 227, 188, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 189, 1, 197, 130, 86, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 189, 1, 198, 227, 189, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 86, 0, 198, 67, 190, 1, 198, 163, 190, 1, 198, 2, 191, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 191, 1, 198, 131, 191, 1, 192, 0, 0, 0, 198, 227, 191, 1, 192, 0, 0, 0, 198, 67, 192, 1, 198, 162, 192, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 192, 1, 198, 35, 193, 1, 192, 0, 0, 0, 198, 130, 193, 1, 197, 3, 87, 0, 198, 194, 193, 1, 198, 3, 194, 1, 198, 98, 194, 1, 197, 98, 87, 0, 198, 163, 194, 1, 192, 0, 0, 0, 198, 2, 195, 1, 192, 0, 0, 0, 198, 67, 195, 1, 197, 162, 87, 0, 198, 163, 195, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 87, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 196, 1, 192, 0, 0, 0, 198, 99, 196, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 196, 1, 198, 35, 197, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 197, 1, 197, 34, 88, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 197, 1, 192, 0, 0, 0, 198, 67, 198, 1, 192, 0, 0, 0, 198, 162, 198, 1, 198, 227, 198, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 199, 1, 197, 99, 88, 0, 192, 0, 0, 0, 197, 194, 88, 0, 198, 163, 199, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 200, 1, 192, 0, 0, 0, 198, 98, 200, 1, 192, 0, 0, 0, 198, 163, 200, 1, 198, 2, 201, 1, 198, 67, 201, 1, 192, 0, 0, 0, 198, 163, 201, 1, 197, 3, 89, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 202, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 89, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 202, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 202, 1, 198, 195, 202, 1, 197, 162, 89, 0, 192, 0, 0, 0, 198, 34, 203, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 203, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 203, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 204, 1, 198, 99, 204, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 204, 1, 192, 0, 0, 0, 197, 227, 89, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 205, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 205, 1, 192, 0, 0, 0, 198, 195, 205, 1, 192, 0, 0, 0, 198, 35, 206, 1, 192, 0, 0, 0, 198, 130, 206, 1, 192, 0, 0, 0, 198, 195, 206, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 207, 1, 192, 0, 0, 0, 198, 131, 207, 1, 197, 66, 90, 0, 197, 130, 90, 0, 197, 194, 90, 0, 192, 0, 0, 0, 198, 226, 207, 1, 192, 0, 0, 0, 198, 34, 208, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 208, 1, 192, 0, 0, 0, 198, 195, 208, 1, 192, 0, 0, 0, 198, 35, 209, 1, 198, 131, 209, 1, 192, 0, 0, 0, 198, 226, 209, 1, 197, 2, 91, 0, 198, 34, 210, 1, 198, 99, 210, 1, 198, 195, 210, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 211, 1, 198, 131, 211, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 211, 1, 192, 0, 0, 0, 198, 67, 212, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 91, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 212, 1, 198, 3, 213, 1, 198, 99, 213, 1, 198, 194, 213, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 214, 1, 192, 0, 0, 0, 198, 67, 214, 1, 197, 131, 91, 0, 198, 163, 214, 1, 197, 226, 91, 0, 198, 3, 215, 1, 197, 34, 92, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 215, 1, 198, 163, 215, 1, 198, 2, 216, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 216, 1, 198, 162, 216, 1, 198, 226, 216, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 217, 1, 192, 0, 0, 0, 198, 131, 217, 1, 192, 0, 0, 0, 198, 226, 217, 1, 198, 34, 218, 1, 192, 0, 0, 0, 198, 98, 218, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 218, 1, 197, 99, 92, 0, 198, 3, 219, 1, 198, 99, 219, 1, 192, 0, 0, 0, 198, 195, 219, 1, 198, 34, 220, 1, 198, 98, 220, 1, 197, 194, 92, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 220, 1, 192, 0, 0, 0, 198, 3, 221, 1, 192, 0, 0, 0, 198, 99, 221, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 221, 1, 197, 2, 93, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 222, 1, 192, 0, 0, 0, 197, 67, 93, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 222, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 222, 1, 198, 227, 222, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 93, 0, 192, 0, 0, 0, 197, 3, 94, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 223, 1, 198, 162, 223, 1, 198, 227, 223, 1, 198, 67, 224, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 224, 1, 198, 3, 225, 1, 198, 99, 225, 1, 198, 195, 225, 1, 198, 35, 226, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 226, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 226, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 94, 0, 192, 0, 0, 0, 198, 35, 227, 1, 198, 130, 227, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 94, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 227, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 228, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 95, 0, 198, 99, 228, 1, 197, 99, 95, 0, 192, 0, 0, 0, 198, 195, 228, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 229, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 229, 1, 192, 0, 0, 0, 198, 195, 229, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 230, 1, 192, 0, 0, 0, 198, 99, 230, 1, 197, 194, 95, 0, 198, 194, 230, 1, 198, 3, 231, 1, 192, 0, 0, 0, 197, 2, 96, 0, 192, 0, 0, 0, 198, 99, 231, 1, 198, 195, 231, 1, 198, 34, 232, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 96, 0, 198, 98, 232, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 232, 1, 198, 3, 233, 1, 198, 99, 233, 1, 198, 195, 233, 1, 198, 34, 234, 1, 198, 98, 234, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 234, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 96, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 235, 1, 198, 99, 235, 1, 198, 194, 235, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 236, 1, 198, 99, 236, 1, 198, 194, 236, 1, 192, 0, 0, 0, 197, 227, 96, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 97, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 237, 1, 192, 0, 0, 0, 198, 99, 237, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 237, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 238, 1, 198, 98, 238, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 238, 1, 192, 0, 0, 0, 197, 163, 97, 0, 198, 2, 239, 1, 192, 0, 0, 0, 197, 2, 98, 0, 198, 66, 239, 1, 197, 67, 98, 0, 197, 162, 98, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 239, 1, 198, 227, 239, 1, 198, 67, 240, 1, 198, 163, 240, 1, 198, 3, 241, 1, 198, 99, 241, 1, 192, 0, 0, 0, 198, 195, 241, 1, 197, 227, 98, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 242, 1, 192, 0, 0, 0, 198, 131, 242, 1, 198, 226, 242, 1, 198, 35, 243, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 243, 1, 192, 0, 0, 0, 198, 227, 243, 1, 197, 66, 99, 0, 198, 67, 244, 1, 192, 0, 0, 0, 197, 131, 99, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 244, 1, 192, 0, 0, 0, 197, 226, 99, 0, 197, 34, 100, 0, 197, 98, 100, 0, 198, 3, 245, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 245, 1, 198, 195, 245, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 246, 1, 197, 162, 100, 0, 192, 0, 0, 0, 198, 99, 246, 1, 197, 227, 100, 0, 198, 195, 246, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 247, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 247, 1, 198, 194, 247, 1, 198, 2, 248, 1, 198, 67, 248, 1, 197, 66, 101, 0, 198, 163, 248, 1, 198, 3, 249, 1, 192, 0, 0, 0, 198, 98, 249, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 249, 1, 198, 2, 250, 1, 197, 130, 101, 0, 198, 67, 250, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 250, 1, 198, 3, 251, 1, 198, 98, 251, 1, 198, 163, 251, 1, 192, 0, 0, 0, 198, 3, 252, 1, 192, 0, 0, 0, 198, 99, 252, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 252, 1, 192, 0, 0, 0, 197, 195, 101, 0, 192, 0, 0, 0, 198, 35, 253, 1, 192, 0, 0, 0, 198, 131, 253, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 253, 1, 198, 67, 254, 1, 197, 35, 102, 0, 198, 163, 254, 1, 198, 3, 255, 1, 198, 99, 255, 1, 192, 0, 0, 0, 197, 130, 102, 0, 197, 194, 102, 0, 198, 195, 255, 1, 197, 2, 103, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 0, 2, 198, 98, 0, 2, 198, 163, 0, 2, 198, 3, 1, 2, 197, 66, 103, 0, 192, 0, 0, 0, 197, 131, 103, 0, 198, 99, 1, 2, 198, 195, 1, 2, 197, 226, 103, 0, 198, 35, 2, 2, 198, 131, 2, 2, 198, 227, 2, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 3, 2, 198, 130, 3, 2, 192, 0, 0, 0, 198, 195, 3, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 4, 2, 192, 0, 0, 0, 197, 34, 104, 0, 197, 98, 104, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 4, 2, 197, 163, 104, 0, 197, 2, 105, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 4, 2, 198, 67, 5, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 5, 2, 198, 226, 5, 2, 197, 67, 105, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 6, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 6, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 6, 2, 198, 2, 7, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 105, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 7, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 7, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 7, 2, 192, 0, 0, 0, 198, 66, 8, 2, 198, 131, 8, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 8, 2, 192, 0, 0, 0, 198, 67, 9, 2, 192, 0, 0, 0, 198, 162, 9, 2, 198, 227, 9, 2, 198, 66, 10, 2, 192, 0, 0, 0, 198, 131, 10, 2, 197, 2, 106, 0, 197, 67, 106, 0, 198, 227, 10, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 11, 2, 197, 162, 106, 0, 192, 0, 0, 0, 198, 162, 11, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 11, 2, 198, 67, 12, 2, 198, 163, 12, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 106, 0, 192, 0, 0, 0, 198, 3, 13, 2, 198, 98, 13, 2, 192, 0, 0, 0, 198, 163, 13, 2, 198, 3, 14, 2, 198, 99, 14, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 107, 0, 198, 194, 14, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 15, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 107, 0, 192, 0, 0, 0, 197, 194, 107, 0, 198, 99, 15, 2, 192, 0, 0, 0, 198, 195, 15, 2, 198, 35, 16, 2, 198, 131, 16, 2, 198, 227, 16, 2, 198, 67, 17, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 17, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 17, 2, 198, 67, 18, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 18, 2, 198, 2, 19, 2, 198, 67, 19, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 19, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 19, 2, 198, 35, 20, 2, 192, 0, 0, 0, 198, 131, 20, 2, 198, 227, 20, 2, 197, 2, 108, 0, 192, 0, 0, 0, 198, 66, 21, 2, 192, 0, 0, 0, 198, 131, 21, 2, 192, 0, 0, 0, 197, 66, 108, 0, 197, 130, 108, 0, 198, 227, 21, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 22, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 22, 2, 198, 3, 23, 2, 198, 98, 23, 2, 192, 0, 0, 0, 198, 163, 23, 2, 198, 3, 24, 2, 192, 0, 0, 0, 198, 98, 24, 2, 198, 163, 24, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 25, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 25, 2, 192, 0, 0, 0, 198, 195, 25, 2, 192, 0, 0, 0, 198, 35, 26, 2, 198, 131, 26, 2, 198, 227, 26, 2, 192, 0, 0, 0, 198, 67, 27, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 27, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 28, 2, 198, 67, 28, 2, 197, 194, 108, 0, 198, 163, 28, 2, 198, 2, 29, 2, 198, 66, 29, 2, 198, 131, 29, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 29, 2, 192, 0, 0, 0, 197, 2, 109, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 30, 2, 198, 162, 30, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 30, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 31, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 31, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 31, 2, 197, 66, 109, 0, 198, 35, 32, 2, 192, 0, 0, 0, 198, 131, 32, 2, 192, 0, 0, 0, 198, 226, 32, 2, 197, 131, 109, 0, 198, 35, 33, 2, 197, 226, 109, 0, 198, 131, 33, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 33, 2, 197, 34, 110, 0, 192, 0, 0, 0, 198, 66, 34, 2, 192, 0, 0, 0, 198, 131, 34, 2, 198, 226, 34, 2, 198, 35, 35, 2, 192, 0, 0, 0, 198, 131, 35, 2, 198, 226, 35, 2, 198, 34, 36, 2, 192, 0, 0, 0, 198, 99, 36, 2, 198, 195, 36, 2, 197, 99, 110, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 37, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 37, 2, 198, 227, 37, 2, 192, 0, 0, 0, 198, 67, 38, 2, 198, 163, 38, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 39, 2, 192, 0, 0, 0, 198, 99, 39, 2, 198, 195, 39, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 40, 2, 192, 0, 0, 0, 198, 131, 40, 2, 197, 194, 110, 0, 197, 2, 111, 0, 198, 227, 40, 2, 198, 66, 41, 2, 198, 130, 41, 2, 197, 66, 111, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 41, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 111, 0, 198, 35, 42, 2, 198, 130, 42, 2, 198, 195, 42, 2, 192, 0, 0, 0, 198, 35, 43, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 43, 2, 197, 194, 111, 0, 198, 226, 43, 2, 198, 35, 44, 2, 198, 130, 44, 2, 192, 0, 0, 0, 198, 194, 44, 2, 198, 3, 45, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 45, 2, 198, 194, 45, 2, 198, 2, 46, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 46, 2, 198, 131, 46, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 46, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 47, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 47, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 47, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 48, 2, 192, 0, 0, 0, 198, 163, 48, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 49, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 49, 2, 198, 130, 49, 2, 192, 0, 0, 0, 198, 195, 49, 2, 192, 0, 0, 0, 198, 34, 50, 2, 198, 99, 50, 2, 198, 194, 50, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 51, 2, 198, 98, 51, 2, 192, 0, 0, 0, 198, 163, 51, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 52, 2, 197, 2, 112, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 52, 2, 198, 194, 52, 2, 192, 0, 0, 0, 198, 2, 53, 2, 197, 66, 112, 0, 198, 66, 53, 2, 198, 131, 53, 2, 198, 227, 53, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 54, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 54, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 112, 0, 198, 226, 54, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 55, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 112, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 55, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 55, 2, 197, 2, 113, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 56, 2, 192, 0, 0, 0, 198, 130, 56, 2, 198, 195, 56, 2, 198, 35, 57, 2, 198, 130, 57, 2, 198, 195, 57, 2, 198, 35, 58, 2, 198, 130, 58, 2, 192, 0, 0, 0, 198, 195, 58, 2, 198, 35, 59, 2, 192, 0, 0, 0, 198, 130, 59, 2, 198, 194, 59, 2, 198, 3, 60, 2, 198, 98, 60, 2, 198, 163, 60, 2, 198, 3, 61, 2, 198, 98, 61, 2, 198, 162, 61, 2, 197, 66, 113, 0, 198, 226, 61, 2, 198, 35, 62, 2, 192, 0, 0, 0, 198, 130, 62, 2, 198, 194, 62, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 63, 2, 198, 99, 63, 2, 198, 195, 63, 2, 198, 35, 64, 2, 192, 0, 0, 0, 198, 131, 64, 2, 192, 0, 0, 0, 198, 227, 64, 2, 192, 0, 0, 0, 198, 67, 65, 2, 198, 163, 65, 2, 198, 2, 66, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 113, 0, 198, 67, 66, 2, 198, 162, 66, 2, 192, 0, 0, 0, 198, 226, 66, 2, 197, 194, 113, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 67, 2, 192, 0, 0, 0, 198, 99, 67, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 67, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 68, 2, 198, 98, 68, 2, 198, 163, 68, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 69, 2, 192, 0, 0, 0, 198, 66, 69, 2, 197, 2, 114, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 69, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 69, 2, 198, 67, 70, 2, 198, 163, 70, 2, 192, 0, 0, 0, 198, 3, 71, 2, 197, 67, 114, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 71, 2, 198, 163, 71, 2, 198, 2, 72, 2, 192, 0, 0, 0, 198, 67, 72, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 72, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 73, 2, 198, 99, 73, 2, 198, 194, 73, 2, 198, 3, 74, 2, 198, 99, 74, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 74, 2, 192, 0, 0, 0, 198, 3, 75, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 75, 2, 198, 194, 75, 2, 198, 3, 76, 2, 192, 0, 0, 0, 198, 99, 76, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 76, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 77, 2, 198, 67, 77, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 77, 2, 197, 163, 114, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 77, 2, 198, 67, 78, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 78, 2, 197, 3, 115, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 78, 2, 198, 35, 79, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 79, 2, 198, 226, 79, 2, 198, 34, 80, 2, 198, 98, 80, 2, 192, 0, 0, 0, 198, 163, 80, 2, 192, 0, 0, 0, 198, 2, 81, 2, 198, 66, 81, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 81, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 81, 2, 192, 0, 0, 0, 198, 35, 82, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 115, 0, 197, 162, 115, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 82, 2, 198, 227, 82, 2, 198, 66, 83, 2, 197, 227, 115, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 83, 2, 192, 0, 0, 0, 198, 227, 83, 2, 197, 66, 116, 0, 198, 67, 84, 2, 198, 162, 84, 2, 192, 0, 0, 0, 198, 227, 84, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 85, 2, 192, 0, 0, 0, 198, 163, 85, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 86, 2, 197, 130, 116, 0, 192, 0, 0, 0, 198, 98, 86, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 86, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 86, 2, 197, 194, 116, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 87, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 87, 2, 197, 3, 117, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 87, 2, 198, 35, 88, 2, 198, 131, 88, 2, 192, 0, 0, 0, 197, 98, 117, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 88, 2, 192, 0, 0, 0, 198, 34, 89, 2, 192, 0, 0, 0, 198, 98, 89, 2, 197, 162, 117, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 89, 2, 192, 0, 0, 0, 198, 3, 90, 2, 198, 99, 90, 2, 197, 226, 117, 0, 198, 195, 90, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 91, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 91, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 91, 2, 192, 0, 0, 0, 198, 35, 92, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 92, 2, 192, 0, 0, 0, 198, 226, 92, 2, 192, 0, 0, 0, 198, 34, 93, 2, 198, 98, 93, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 93, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 93, 2, 192, 0, 0, 0, 198, 67, 94, 2, 198, 162, 94, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 118, 0, 192, 0, 0, 0, 198, 227, 94, 2, 198, 67, 95, 2, 198, 163, 95, 2, 197, 131, 118, 0, 192, 0, 0, 0, 198, 2, 96, 2, 198, 67, 96, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 96, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 118, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 97, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 97, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 97, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 119, 0, 192, 0, 0, 0, 198, 34, 98, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 98, 2, 192, 0, 0, 0, 198, 194, 98, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 99, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 99, 2, 198, 130, 99, 2, 198, 195, 99, 2, 198, 35, 100, 2, 192, 0, 0, 0, 197, 98, 119, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 119, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 100, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 100, 2, 198, 35, 101, 2, 198, 131, 101, 2, 198, 226, 101, 2, 192, 0, 0, 0, 198, 34, 102, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 102, 2, 198, 195, 102, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 119, 0, 198, 35, 103, 2, 198, 130, 103, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 103, 2, 198, 35, 104, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 104, 2, 192, 0, 0, 0, 198, 194, 104, 2, 192, 0, 0, 0, 197, 34, 120, 0, 198, 3, 105, 2, 192, 0, 0, 0, 198, 99, 105, 2, 192, 0, 0, 0, 198, 195, 105, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 106, 2, 198, 131, 106, 2, 192, 0, 0, 0, 198, 226, 106, 2, 198, 35, 107, 2, 192, 0, 0, 0, 198, 131, 107, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 107, 2, 198, 35, 108, 2, 192, 0, 0, 0, 197, 99, 120, 0, 192, 0, 0, 0, 198, 131, 108, 2, 198, 227, 108, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 109, 2, 198, 131, 109, 2, 192, 0, 0, 0, 198, 227, 109, 2, 192, 0, 0, 0, 198, 67, 110, 2, 192, 0, 0, 0, 198, 163, 110, 2, 197, 194, 120, 0, 192, 0, 0, 0, 198, 3, 111, 2, 198, 99, 111, 2, 192, 0, 0, 0, 198, 194, 111, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 121, 0, 198, 2, 112, 2, 198, 67, 112, 2, 198, 162, 112, 2, 192, 0, 0, 0, 198, 227, 112, 2, 198, 67, 113, 2, 198, 162, 113, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 121, 0, 197, 130, 121, 0, 192, 0, 0, 0, 198, 227, 113, 2, 192, 0, 0, 0, 198, 67, 114, 2, 198, 163, 114, 2, 192, 0, 0, 0, 197, 195, 121, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 122, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 115, 2, 192, 0, 0, 0, 198, 67, 115, 2, 192, 0, 0, 0, 198, 162, 115, 2, 192, 0, 0, 0, 197, 98, 122, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 115, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 116, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 116, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 117, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 117, 2, 198, 195, 117, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 118, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 118, 2, 198, 227, 118, 2, 198, 67, 119, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 119, 2, 197, 163, 122, 0, 198, 227, 119, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 120, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 120, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 120, 2, 198, 2, 121, 2, 197, 2, 123, 0, 198, 66, 121, 2, 192, 0, 0, 0, 198, 131, 121, 2, 192, 0, 0, 0, 198, 227, 121, 2, 198, 67, 122, 2, 198, 163, 122, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 123, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 123, 0, 198, 66, 123, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 123, 2, 198, 194, 123, 2, 198, 3, 124, 2, 192, 0, 0, 0, 198, 98, 124, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 124, 2, 198, 227, 124, 2, 198, 66, 125, 2, 197, 130, 123, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 123, 0, 198, 131, 125, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 125, 2, 198, 67, 126, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 126, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 127, 2, 198, 99, 127, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 127, 2, 192, 0, 0, 0, 198, 3, 128, 2, 198, 99, 128, 2, 192, 0, 0, 0, 198, 194, 128, 2, 198, 2, 129, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 129, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 129, 2, 198, 227, 129, 2, 197, 2, 124, 0, 192, 0, 0, 0, 197, 66, 124, 0, 192, 0, 0, 0, 198, 67, 130, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 130, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 131, 2, 198, 67, 131, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 131, 2, 192, 0, 0, 0, 198, 226, 131, 2, 192, 0, 0, 0, 198, 35, 132, 2, 198, 130, 132, 2, 198, 194, 132, 2, 192, 0, 0, 0, 198, 3, 133, 2, 192, 0, 0, 0, 198, 99, 133, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 133, 2, 198, 35, 134, 2, 192, 0, 0, 0, 198, 130, 134, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 134, 2, 198, 35, 135, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 135, 2, 198, 195, 135, 2, 198, 35, 136, 2, 198, 131, 136, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 136, 2, 192, 0, 0, 0, 198, 67, 137, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 137, 2, 198, 226, 137, 2, 192, 0, 0, 0, 198, 34, 138, 2, 192, 0, 0, 0, 198, 99, 138, 2, 198, 194, 138, 2, 198, 3, 139, 2, 198, 98, 139, 2, 197, 130, 124, 0, 197, 194, 124, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 139, 2, 198, 3, 140, 2, 198, 99, 140, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 140, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 141, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 141, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 141, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 141, 2, 192, 0, 0, 0, 198, 35, 142, 2, 198, 131, 142, 2, 192, 0, 0, 0, 198, 226, 142, 2, 198, 34, 143, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 143, 2, 198, 195, 143, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 144, 2, 198, 130, 144, 2, 192, 0, 0, 0, 198, 194, 144, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 145, 2, 198, 99, 145, 2, 198, 195, 145, 2, 198, 35, 146, 2, 192, 0, 0, 0, 198, 130, 146, 2, 198, 195, 146, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 147, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 147, 2, 198, 227, 147, 2, 192, 0, 0, 0, 198, 67, 148, 2, 197, 2, 125, 0, 192, 0, 0, 0, 198, 163, 148, 2, 192, 0, 0, 0, 198, 3, 149, 2, 198, 99, 149, 2, 198, 195, 149, 2, 198, 35, 150, 2, 192, 0, 0, 0, 198, 131, 150, 2, 198, 227, 150, 2, 197, 67, 125, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 151, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 151, 2, 198, 227, 151, 2, 198, 67, 152, 2, 198, 163, 152, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 153, 2, 198, 99, 153, 2, 192, 0, 0, 0, 198, 194, 153, 2, 198, 3, 154, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 154, 2, 192, 0, 0, 0, 198, 162, 154, 2, 192, 0, 0, 0, 198, 226, 154, 2, 197, 162, 125, 0, 198, 35, 155, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 155, 2, 198, 195, 155, 2, 198, 35, 156, 2, 198, 131, 156, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 156, 2, 198, 67, 157, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 157, 2, 198, 226, 157, 2, 197, 227, 125, 0, 197, 67, 126, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 158, 2, 198, 99, 158, 2, 192, 0, 0, 0, 198, 195, 158, 2, 198, 35, 159, 2, 192, 0, 0, 0, 198, 131, 159, 2, 198, 227, 159, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 160, 2, 192, 0, 0, 0, 198, 163, 160, 2, 192, 0, 0, 0, 198, 2, 161, 2, 197, 162, 126, 0, 198, 67, 161, 2, 192, 0, 0, 0, 198, 162, 161, 2, 198, 227, 161, 2, 192, 0, 0, 0, 198, 66, 162, 2, 197, 227, 126, 0, 197, 66, 127, 0, 197, 130, 127, 0, 198, 131, 162, 2, 198, 227, 162, 2, 192, 0, 0, 0, 198, 67, 163, 2, 198, 162, 163, 2, 192, 0, 0, 0, 198, 226, 163, 2, 198, 35, 164, 2, 198, 131, 164, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 164, 2, 198, 67, 165, 2, 192, 0, 0, 0, 198, 163, 165, 2, 198, 2, 166, 2, 198, 67, 166, 2, 192, 0, 0, 0, 198, 162, 166, 2, 198, 227, 166, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 127, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 167, 2, 197, 35, 128, 0, 197, 130, 128, 0, 192, 0, 0, 0, 198, 131, 167, 2, 198, 226, 167, 2, 192, 0, 0, 0, 198, 35, 168, 2, 192, 0, 0, 0, 197, 194, 128, 0, 198, 130, 168, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 168, 2, 198, 35, 169, 2, 198, 131, 169, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 169, 2, 198, 66, 170, 2, 192, 0, 0, 0, 198, 130, 170, 2, 192, 0, 0, 0, 198, 195, 170, 2, 198, 34, 171, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 171, 2, 198, 195, 171, 2, 192, 0, 0, 0, 198, 34, 172, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 172, 2, 198, 195, 172, 2, 197, 2, 129, 0, 197, 67, 129, 0, 198, 35, 173, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 173, 2, 198, 195, 173, 2, 192, 0, 0, 0, 198, 35, 174, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 174, 2, 197, 162, 129, 0, 198, 226, 174, 2, 198, 34, 175, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 175, 2, 198, 195, 175, 2, 198, 35, 176, 2, 198, 131, 176, 2, 192, 0, 0, 0, 197, 226, 129, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 176, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 177, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 177, 2, 198, 163, 177, 2, 197, 35, 130, 0, 198, 2, 178, 2, 197, 131, 130, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 178, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 178, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 130, 0, 192, 0, 0, 0, 198, 3, 179, 2, 198, 99, 179, 2, 198, 194, 179, 2, 192, 0, 0, 0, 197, 34, 131, 0, 192, 0, 0, 0, 198, 3, 180, 2, 198, 99, 180, 2, 198, 195, 180, 2, 192, 0, 0, 0, 198, 35, 181, 2, 197, 98, 131, 0, 198, 131, 181, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 181, 2, 198, 66, 182, 2, 198, 130, 182, 2, 198, 194, 182, 2, 192, 0, 0, 0, 198, 3, 183, 2, 192, 0, 0, 0, 198, 99, 183, 2, 197, 162, 131, 0, 197, 226, 131, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 132, 0, 192, 0, 0, 0, 198, 195, 183, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 184, 2, 197, 98, 132, 0, 192, 0, 0, 0, 198, 131, 184, 2, 198, 227, 184, 2, 198, 67, 185, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 185, 2, 198, 3, 186, 2, 192, 0, 0, 0, 198, 99, 186, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 186, 2, 192, 0, 0, 0, 198, 34, 187, 2, 198, 98, 187, 2, 198, 162, 187, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 187, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 188, 2, 198, 163, 188, 2, 198, 3, 189, 2, 192, 0, 0, 0, 198, 98, 189, 2, 198, 163, 189, 2, 198, 3, 190, 2, 198, 99, 190, 2, 198, 195, 190, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 191, 2, 192, 0, 0, 0, 198, 99, 191, 2, 198, 195, 191, 2, 192, 0, 0, 0, 198, 34, 192, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 132, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 132, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 192, 2, 198, 163, 192, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 193, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 193, 2, 198, 163, 193, 2, 198, 2, 194, 2, 198, 67, 194, 2, 192, 0, 0, 0, 197, 34, 133, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 194, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 195, 2, 192, 0, 0, 0, 198, 66, 195, 2, 198, 131, 195, 2, 198, 227, 195, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 196, 2, 192, 0, 0, 0, 198, 131, 196, 2, 198, 227, 196, 2, 192, 0, 0, 0, 198, 66, 197, 2, 198, 131, 197, 2, 198, 227, 197, 2, 198, 66, 198, 2, 198, 130, 198, 2, 192, 0, 0, 0, 197, 98, 133, 0, 192, 0, 0, 0, 197, 162, 133, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 198, 2, 197, 226, 133, 0, 192, 0, 0, 0, 198, 3, 199, 2, 198, 99, 199, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 199, 2, 192, 0, 0, 0, 198, 3, 200, 2, 192, 0, 0, 0, 198, 99, 200, 2, 198, 195, 200, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 201, 2, 198, 131, 201, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 201, 2, 192, 0, 0, 0, 197, 35, 134, 0, 192, 0, 0, 0, 198, 35, 202, 2, 197, 130, 134, 0, 197, 194, 134, 0, 198, 130, 202, 2, 192, 0, 0, 0, 197, 2, 135, 0, 198, 194, 202, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 203, 2, 198, 67, 203, 2, 192, 0, 0, 0, 198, 162, 203, 2, 198, 226, 203, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 204, 2, 198, 98, 204, 2, 198, 162, 204, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 204, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 205, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 205, 2, 198, 194, 205, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 206, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 206, 2, 198, 194, 206, 2, 198, 3, 207, 2, 198, 99, 207, 2, 198, 195, 207, 2, 197, 66, 135, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 208, 2, 198, 130, 208, 2, 197, 131, 135, 0, 197, 226, 135, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 208, 2, 198, 35, 209, 2, 198, 131, 209, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 136, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 209, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 210, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 210, 2, 198, 194, 210, 2, 198, 2, 211, 2, 192, 0, 0, 0, 198, 66, 211, 2, 198, 130, 211, 2, 198, 195, 211, 2, 198, 34, 212, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 212, 2, 192, 0, 0, 0, 198, 162, 212, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 212, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 136, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 213, 2, 198, 163, 213, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 214, 2, 198, 67, 214, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 214, 2, 198, 3, 215, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 215, 2, 198, 163, 215, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 216, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 216, 2, 198, 131, 216, 2, 192, 0, 0, 0, 198, 227, 216, 2, 198, 66, 217, 2, 192, 0, 0, 0, 198, 131, 217, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 136, 0, 198, 227, 217, 2, 192, 0, 0, 0, 198, 67, 218, 2, 198, 163, 218, 2, 198, 2, 219, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 219, 2, 198, 163, 219, 2, 198, 2, 220, 2, 198, 66, 220, 2, 198, 131, 220, 2, 192, 0, 0, 0, 198, 227, 220, 2, 192, 0, 0, 0, 198, 67, 221, 2, 198, 163, 221, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 222, 2, 192, 0, 0, 0, 198, 99, 222, 2, 198, 195, 222, 2, 192, 0, 0, 0, 197, 3, 137, 0, 198, 35, 223, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 223, 2, 192, 0, 0, 0, 198, 227, 223, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 224, 2, 192, 0, 0, 0, 198, 131, 224, 2, 198, 227, 224, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 225, 2, 198, 163, 225, 2, 192, 0, 0, 0, 198, 3, 226, 2, 192, 0, 0, 0, 198, 99, 226, 2, 192, 0, 0, 0, 198, 194, 226, 2, 198, 3, 227, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 227, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 227, 2, 197, 98, 137, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 228, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 228, 2, 192, 0, 0, 0, 198, 162, 228, 2, 192, 0, 0, 0, 197, 162, 137, 0, 197, 226, 137, 0, 198, 227, 228, 2, 198, 66, 229, 2, 198, 131, 229, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 229, 2, 198, 34, 230, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 230, 2, 192, 0, 0, 0, 197, 34, 138, 0, 198, 162, 230, 2, 198, 227, 230, 2, 192, 0, 0, 0, 198, 66, 231, 2, 192, 0, 0, 0, 198, 131, 231, 2, 198, 226, 231, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 232, 2, 198, 99, 232, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 232, 2, 198, 34, 233, 2, 198, 98, 233, 2, 198, 162, 233, 2, 198, 227, 233, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 234, 2, 192, 0, 0, 0, 198, 131, 234, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 234, 2, 198, 67, 235, 2, 198, 162, 235, 2, 198, 227, 235, 2, 192, 0, 0, 0, 198, 67, 236, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 236, 2, 198, 3, 237, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 138, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 237, 2, 198, 194, 237, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 238, 2, 192, 0, 0, 0, 198, 99, 238, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 238, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 239, 2, 198, 67, 239, 2, 198, 163, 239, 2, 198, 3, 240, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 240, 2, 198, 195, 240, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 241, 2, 197, 162, 138, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 241, 2, 192, 0, 0, 0, 198, 227, 241, 2, 197, 226, 138, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 242, 2, 198, 130, 242, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 242, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 243, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 139, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 243, 2, 192, 0, 0, 0, 198, 227, 243, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 139, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 244, 2, 192, 0, 0, 0, 198, 163, 244, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 245, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 245, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 245, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 245, 2, 198, 67, 246, 2, 198, 162, 246, 2, 192, 0, 0, 0, 198, 227, 246, 2, 192, 0, 0, 0, 198, 66, 247, 2, 198, 130, 247, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 247, 2, 198, 35, 248, 2, 192, 0, 0, 0, 198, 131, 248, 2, 197, 194, 139, 0, 197, 2, 140, 0, 198, 226, 248, 2, 192, 0, 0, 0, 198, 35, 249, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 249, 2, 192, 0, 0, 0, 198, 227, 249, 2, 198, 67, 250, 2, 198, 163, 250, 2, 198, 3, 251, 2, 197, 66, 140, 0, 198, 99, 251, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 251, 2, 192, 0, 0, 0, 198, 3, 252, 2, 192, 0, 0, 0, 197, 130, 140, 0, 197, 194, 140, 0, 198, 99, 252, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 252, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 253, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 253, 2, 198, 163, 253, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 254, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 254, 2, 198, 163, 254, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 255, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 255, 2, 197, 3, 141, 0, 198, 163, 255, 2, 198, 2, 0, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 0, 3, 198, 131, 0, 3, 192, 0, 0, 0, 198, 226, 0, 3, 197, 98, 141, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 1, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 1, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 141, 0, 197, 3, 142, 0, 192, 0, 0, 0, 198, 163, 1, 3, 192, 0, 0, 0, 198, 2, 2, 3, 198, 66, 2, 3, 192, 0, 0, 0, 198, 131, 2, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 2, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 3, 3, 198, 162, 3, 3, 197, 98, 142, 0, 198, 227, 3, 3, 198, 66, 4, 3, 192, 0, 0, 0, 198, 130, 4, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 4, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 142, 0, 198, 35, 5, 3, 198, 131, 5, 3, 198, 227, 5, 3, 198, 66, 6, 3, 198, 130, 6, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 6, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 7, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 7, 3, 192, 0, 0, 0, 198, 194, 7, 3, 192, 0, 0, 0, 198, 2, 8, 3, 192, 0, 0, 0, 198, 67, 8, 3, 198, 163, 8, 3, 192, 0, 0, 0, 198, 3, 9, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 9, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 9, 3, 198, 35, 10, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 10, 3, 198, 227, 10, 3, 198, 66, 11, 3, 198, 130, 11, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 142, 0, 192, 0, 0, 0, 197, 34, 143, 0, 192, 0, 0, 0, 198, 194, 11, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 143, 0, 198, 3, 12, 3, 192, 0, 0, 0, 198, 98, 12, 3, 192, 0, 0, 0, 198, 163, 12, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 13, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 13, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 13, 3, 198, 2, 14, 3, 197, 162, 143, 0, 197, 226, 143, 0, 192, 0, 0, 0, 198, 67, 14, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 14, 3, 198, 3, 15, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 144, 0, 198, 98, 15, 3, 192, 0, 0, 0, 198, 162, 15, 3, 198, 226, 15, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 16, 3, 192, 0, 0, 0, 198, 131, 16, 3, 198, 227, 16, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 17, 3, 198, 163, 17, 3, 198, 3, 18, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 18, 3, 192, 0, 0, 0, 198, 163, 18, 3, 198, 2, 19, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 19, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 144, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 19, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 20, 3, 198, 99, 20, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 20, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 21, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 21, 3, 198, 227, 21, 3, 192, 0, 0, 0, 198, 67, 22, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 22, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 23, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 144, 0, 198, 99, 23, 3, 192, 0, 0, 0, 198, 194, 23, 3, 198, 3, 24, 3, 192, 0, 0, 0, 198, 99, 24, 3, 198, 195, 24, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 25, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 25, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 25, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 145, 0, 198, 35, 26, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 26, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 26, 3, 197, 98, 145, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 27, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 145, 0, 198, 99, 27, 3, 198, 194, 27, 3, 198, 3, 28, 3, 198, 99, 28, 3, 198, 195, 28, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 29, 3, 192, 0, 0, 0, 198, 131, 29, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 29, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 30, 3, 198, 163, 30, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 31, 3, 198, 99, 31, 3, 198, 195, 31, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 32, 3, 192, 0, 0, 0, 198, 131, 32, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 32, 3, 198, 67, 33, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 33, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 34, 3, 192, 0, 0, 0, 197, 227, 145, 0, 198, 98, 34, 3, 192, 0, 0, 0, 197, 67, 146, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 146, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 34, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 146, 0, 198, 2, 35, 3, 197, 34, 147, 0, 198, 66, 35, 3, 197, 99, 147, 0, 192, 0, 0, 0, 197, 195, 147, 0, 198, 131, 35, 3, 198, 227, 35, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 36, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 148, 0, 197, 131, 148, 0, 192, 0, 0, 0, 198, 163, 36, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 37, 3, 192, 0, 0, 0, 198, 67, 37, 3, 192, 0, 0, 0, 198, 162, 37, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 148, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 149, 0, 198, 227, 37, 3, 198, 67, 38, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 149, 0, 198, 163, 38, 3, 197, 195, 149, 0, 198, 2, 39, 3, 198, 67, 39, 3, 198, 162, 39, 3, 198, 226, 39, 3, 198, 34, 40, 3, 198, 98, 40, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 150, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 40, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 40, 3, 198, 66, 41, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 41, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 41, 3, 198, 34, 42, 3, 192, 0, 0, 0, 198, 98, 42, 3, 198, 163, 42, 3, 198, 3, 43, 3, 192, 0, 0, 0, 197, 98, 150, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 43, 3, 198, 194, 43, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 150, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 44, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 44, 3, 192, 0, 0, 0, 198, 195, 44, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 45, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 150, 0, 192, 0, 0, 0, 198, 131, 45, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 45, 3, 192, 0, 0, 0, 198, 67, 46, 3, 198, 163, 46, 3, 198, 3, 47, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 47, 3, 198, 194, 47, 3, 192, 0, 0, 0, 198, 3, 48, 3, 192, 0, 0, 0, 198, 99, 48, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 48, 3, 192, 0, 0, 0, 198, 35, 49, 3, 197, 34, 151, 0, 192, 0, 0, 0, 198, 131, 49, 3, 192, 0, 0, 0, 197, 98, 151, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 49, 3, 197, 162, 151, 0, 198, 66, 50, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 50, 3, 192, 0, 0, 0, 198, 194, 50, 3, 192, 0, 0, 0, 198, 2, 51, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 51, 3, 198, 163, 51, 3, 198, 2, 52, 3, 192, 0, 0, 0, 198, 66, 52, 3, 198, 130, 52, 3, 198, 194, 52, 3, 192, 0, 0, 0, 198, 3, 53, 3, 198, 98, 53, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 151, 0, 198, 162, 53, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 53, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 152, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 54, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 54, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 55, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 55, 3, 197, 98, 152, 0, 198, 163, 55, 3, 192, 0, 0, 0, 198, 3, 56, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 56, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 56, 3, 198, 3, 57, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 57, 3, 198, 194, 57, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 58, 3, 198, 99, 58, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 58, 3, 197, 162, 152, 0, 198, 35, 59, 3, 198, 131, 59, 3, 198, 226, 59, 3, 198, 35, 60, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 60, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 60, 3, 198, 66, 61, 3, 198, 131, 61, 3, 192, 0, 0, 0, 198, 227, 61, 3, 192, 0, 0, 0, 197, 227, 152, 0, 198, 67, 62, 3, 192, 0, 0, 0, 198, 163, 62, 3, 192, 0, 0, 0, 198, 2, 63, 3, 198, 66, 63, 3, 198, 131, 63, 3, 198, 226, 63, 3, 197, 66, 153, 0, 197, 131, 153, 0, 198, 35, 64, 3, 198, 130, 64, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 64, 3, 192, 0, 0, 0, 198, 35, 65, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 153, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 154, 0, 192, 0, 0, 0, 198, 131, 65, 3, 192, 0, 0, 0, 198, 227, 65, 3, 192, 0, 0, 0, 198, 67, 66, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 66, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 67, 3, 198, 67, 67, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 67, 3, 198, 2, 68, 3, 198, 67, 68, 3, 198, 163, 68, 3, 198, 3, 69, 3, 198, 99, 69, 3, 198, 195, 69, 3, 197, 99, 154, 0, 197, 195, 154, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 70, 3, 197, 34, 155, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 70, 3, 192, 0, 0, 0, 198, 227, 70, 3, 198, 67, 71, 3, 198, 163, 71, 3, 198, 3, 72, 3, 198, 99, 72, 3, 192, 0, 0, 0, 198, 194, 72, 3, 198, 2, 73, 3, 198, 66, 73, 3, 197, 98, 155, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 73, 3, 198, 226, 73, 3, 192, 0, 0, 0, 197, 162, 155, 0, 192, 0, 0, 0, 198, 35, 74, 3, 198, 130, 74, 3, 192, 0, 0, 0, 197, 226, 155, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 74, 3, 197, 34, 156, 0, 192, 0, 0, 0, 197, 98, 156, 0, 198, 2, 75, 3, 198, 66, 75, 3, 198, 131, 75, 3, 198, 227, 75, 3, 192, 0, 0, 0, 197, 163, 156, 0, 198, 67, 76, 3, 192, 0, 0, 0, 198, 163, 76, 3, 198, 2, 77, 3, 198, 67, 77, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 77, 3, 198, 3, 78, 3, 192, 0, 0, 0, 197, 3, 157, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 78, 3, 198, 194, 78, 3, 192, 0, 0, 0, 197, 99, 157, 0, 198, 3, 79, 3, 198, 99, 79, 3, 192, 0, 0, 0, 198, 194, 79, 3, 192, 0, 0, 0, 198, 3, 80, 3, 198, 98, 80, 3, 198, 163, 80, 3, 198, 2, 81, 3, 198, 66, 81, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 81, 3, 198, 226, 81, 3, 197, 194, 157, 0, 198, 35, 82, 3, 192, 0, 0, 0, 198, 130, 82, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 82, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 83, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 83, 3, 198, 195, 83, 3, 197, 2, 158, 0, 198, 35, 84, 3, 192, 0, 0, 0, 198, 131, 84, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 158, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 84, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 85, 3, 198, 163, 85, 3, 198, 3, 86, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 86, 3, 198, 162, 86, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 86, 3, 192, 0, 0, 0, 198, 34, 87, 3, 192, 0, 0, 0, 198, 99, 87, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 87, 3, 198, 34, 88, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 88, 3, 192, 0, 0, 0, 197, 130, 158, 0, 192, 0, 0, 0, 198, 195, 88, 3, 197, 195, 158, 0, 197, 35, 159, 0, 198, 35, 89, 3, 192, 0, 0, 0, 198, 130, 89, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 89, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 90, 3, 197, 131, 159, 0, 192, 0, 0, 0, 198, 99, 90, 3, 198, 195, 90, 3, 198, 35, 91, 3, 197, 227, 159, 0, 197, 66, 160, 0, 197, 131, 160, 0, 197, 226, 160, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 91, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 91, 3, 198, 35, 92, 3, 192, 0, 0, 0, 198, 131, 92, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 92, 3, 192, 0, 0, 0, 198, 67, 93, 3, 198, 163, 93, 3, 198, 3, 94, 3, 198, 98, 94, 3, 198, 163, 94, 3, 192, 0, 0, 0, 197, 34, 161, 0, 198, 3, 95, 3, 192, 0, 0, 0, 198, 98, 95, 3, 192, 0, 0, 0, 198, 162, 95, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 95, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 161, 0, 197, 163, 161, 0, 197, 3, 162, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 162, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 96, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 96, 3, 192, 0, 0, 0, 198, 227, 96, 3, 198, 66, 97, 3, 198, 131, 97, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 97, 3, 197, 162, 162, 0, 198, 67, 98, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 98, 3, 192, 0, 0, 0, 198, 3, 99, 3, 198, 99, 99, 3, 197, 226, 162, 0, 198, 194, 99, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 100, 3, 198, 67, 100, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 100, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 100, 3, 192, 0, 0, 0, 198, 67, 101, 3, 197, 35, 163, 0, 192, 0, 0, 0, 198, 163, 101, 3, 197, 130, 163, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 102, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 102, 3, 198, 195, 102, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 103, 3, 197, 194, 163, 0, 198, 131, 103, 3, 197, 2, 164, 0, 198, 226, 103, 3, 198, 35, 104, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 104, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 164, 0, 192, 0, 0, 0, 198, 227, 104, 3, 197, 130, 164, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 105, 3, 192, 0, 0, 0, 198, 130, 105, 3, 198, 195, 105, 3, 198, 34, 106, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 106, 3, 198, 163, 106, 3, 192, 0, 0, 0, 198, 2, 107, 3, 198, 67, 107, 3, 198, 163, 107, 3, 192, 0, 0, 0, 198, 2, 108, 3, 198, 67, 108, 3, 198, 163, 108, 3, 192, 0, 0, 0, 198, 3, 109, 3, 192, 0, 0, 0, 198, 99, 109, 3, 198, 195, 109, 3, 198, 34, 110, 3, 192, 0, 0, 0, 197, 194, 164, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 110, 3, 192, 0, 0, 0, 198, 195, 110, 3, 192, 0, 0, 0, 198, 35, 111, 3, 198, 130, 111, 3, 198, 195, 111, 3, 198, 35, 112, 3, 198, 131, 112, 3, 198, 227, 112, 3, 198, 67, 113, 3, 198, 162, 113, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 113, 3, 192, 0, 0, 0, 197, 3, 165, 0, 197, 98, 165, 0, 198, 67, 114, 3, 192, 0, 0, 0, 197, 162, 165, 0, 192, 0, 0, 0, 198, 163, 114, 3, 198, 2, 115, 3, 198, 66, 115, 3, 192, 0, 0, 0, 198, 130, 115, 3, 198, 195, 115, 3, 198, 35, 116, 3, 197, 226, 165, 0, 192, 0, 0, 0, 197, 34, 166, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 116, 3, 198, 227, 116, 3, 197, 98, 166, 0, 192, 0, 0, 0, 198, 67, 117, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 117, 3, 192, 0, 0, 0, 198, 3, 118, 3, 192, 0, 0, 0, 198, 99, 118, 3, 198, 195, 118, 3, 197, 162, 166, 0, 192, 0, 0, 0, 198, 34, 119, 3, 192, 0, 0, 0, 198, 98, 119, 3, 198, 162, 119, 3, 198, 227, 119, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 120, 3, 198, 130, 120, 3, 198, 195, 120, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 121, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 121, 3, 198, 195, 121, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 122, 3, 192, 0, 0, 0, 198, 99, 122, 3, 198, 195, 122, 3, 192, 0, 0, 0, 197, 226, 166, 0, 192, 0, 0, 0, 198, 35, 123, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 123, 3, 198, 227, 123, 3, 192, 0, 0, 0, 198, 67, 124, 3, 197, 35, 167, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 124, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 125, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 167, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 125, 3, 198, 195, 125, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 126, 3, 198, 131, 126, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 126, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 127, 3, 198, 131, 127, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 127, 3, 198, 67, 128, 3, 192, 0, 0, 0, 197, 195, 167, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 128, 3, 198, 226, 128, 3, 192, 0, 0, 0, 198, 35, 129, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 129, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 129, 3, 192, 0, 0, 0, 198, 35, 130, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 130, 3, 198, 195, 130, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 131, 3, 192, 0, 0, 0, 198, 99, 131, 3, 198, 194, 131, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 132, 3, 198, 99, 132, 3, 198, 195, 132, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 133, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 133, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 133, 3, 198, 67, 134, 3, 198, 163, 134, 3, 198, 3, 135, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 168, 0, 197, 99, 168, 0, 192, 0, 0, 0, 198, 98, 135, 3, 198, 163, 135, 3, 192, 0, 0, 0, 198, 3, 136, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 168, 0, 192, 0, 0, 0, 198, 98, 136, 3, 198, 162, 136, 3, 198, 227, 136, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 169, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 169, 0, 197, 163, 169, 0, 192, 0, 0, 0, 197, 2, 170, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 137, 3, 198, 163, 137, 3, 192, 0, 0, 0, 198, 2, 138, 3, 197, 66, 170, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 138, 3, 198, 163, 138, 3, 192, 0, 0, 0, 198, 2, 139, 3, 197, 130, 170, 0, 192, 0, 0, 0, 198, 66, 139, 3, 192, 0, 0, 0, 198, 131, 139, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 139, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 140, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 140, 3, 192, 0, 0, 0, 198, 226, 140, 3, 192, 0, 0, 0, 198, 34, 141, 3, 198, 99, 141, 3, 198, 194, 141, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 142, 3, 198, 98, 142, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 142, 3, 192, 0, 0, 0, 198, 3, 143, 3, 198, 99, 143, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 170, 0, 198, 195, 143, 3, 198, 34, 144, 3, 198, 99, 144, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 144, 3, 198, 34, 145, 3, 198, 99, 145, 3, 198, 195, 145, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 146, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 171, 0, 192, 0, 0, 0, 198, 131, 146, 3, 198, 227, 146, 3, 192, 0, 0, 0, 198, 67, 147, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 147, 3, 192, 0, 0, 0, 198, 3, 148, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 148, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 148, 3, 198, 35, 149, 3, 198, 130, 149, 3, 198, 195, 149, 3, 198, 35, 150, 3, 198, 131, 150, 3, 197, 67, 171, 0, 198, 226, 150, 3, 198, 35, 151, 3, 192, 0, 0, 0, 198, 131, 151, 3, 197, 163, 171, 0, 192, 0, 0, 0, 198, 227, 151, 3, 192, 0, 0, 0, 198, 66, 152, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 152, 3, 198, 226, 152, 3, 198, 35, 153, 3, 197, 2, 172, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 172, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 153, 3, 192, 0, 0, 0, 198, 227, 153, 3, 192, 0, 0, 0, 198, 67, 154, 3, 198, 162, 154, 3, 198, 227, 154, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 155, 3, 198, 163, 155, 3, 192, 0, 0, 0, 198, 3, 156, 3, 192, 0, 0, 0, 198, 98, 156, 3, 192, 0, 0, 0, 198, 163, 156, 3, 198, 3, 157, 3, 192, 0, 0, 0, 197, 131, 172, 0, 198, 98, 157, 3, 198, 162, 157, 3, 198, 227, 157, 3, 192, 0, 0, 0, 198, 67, 158, 3, 192, 0, 0, 0, 198, 163, 158, 3, 192, 0, 0, 0, 197, 226, 172, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 159, 3, 198, 99, 159, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 159, 3, 192, 0, 0, 0, 198, 34, 160, 3, 192, 0, 0, 0, 198, 99, 160, 3, 197, 35, 173, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 160, 3, 198, 2, 161, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 173, 0, 198, 67, 161, 3, 192, 0, 0, 0, 198, 163, 161, 3, 192, 0, 0, 0, 198, 2, 162, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 162, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 162, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 163, 3, 192, 0, 0, 0, 198, 99, 163, 3, 198, 195, 163, 3, 198, 35, 164, 3, 198, 131, 164, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 164, 3, 198, 67, 165, 3, 192, 0, 0, 0, 198, 162, 165, 3, 197, 194, 173, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 165, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 166, 3, 192, 0, 0, 0, 198, 163, 166, 3, 198, 3, 167, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 167, 3, 197, 2, 174, 0, 197, 66, 174, 0, 197, 130, 174, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 167, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 174, 0, 192, 0, 0, 0, 198, 227, 167, 3, 198, 67, 168, 3, 192, 0, 0, 0, 198, 163, 168, 3, 198, 3, 169, 3, 192, 0, 0, 0, 198, 99, 169, 3, 192, 0, 0, 0, 198, 195, 169, 3, 198, 35, 170, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 170, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 175, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 170, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 171, 3, 198, 163, 171, 3, 198, 3, 172, 3, 192, 0, 0, 0, 198, 98, 172, 3, 192, 0, 0, 0, 198, 163, 172, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 173, 3, 192, 0, 0, 0, 197, 67, 175, 0, 198, 67, 173, 3, 198, 163, 173, 3, 192, 0, 0, 0, 198, 3, 174, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 174, 3, 192, 0, 0, 0, 198, 163, 174, 3, 198, 3, 175, 3, 198, 99, 175, 3, 198, 195, 175, 3, 198, 35, 176, 3, 198, 130, 176, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 175, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 175, 0, 198, 195, 176, 3, 198, 34, 177, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 177, 3, 198, 194, 177, 3, 192, 0, 0, 0, 198, 3, 178, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 178, 3, 198, 195, 178, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 179, 3, 192, 0, 0, 0, 198, 131, 179, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 179, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 180, 3, 192, 0, 0, 0, 198, 163, 180, 3, 192, 0, 0, 0, 198, 3, 181, 3, 197, 34, 176, 0, 192, 0, 0, 0, 198, 99, 181, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 181, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 176, 0, 197, 162, 176, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 182, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 182, 3, 192, 0, 0, 0, 197, 226, 176, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 182, 3, 197, 34, 177, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 183, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 183, 3, 198, 226, 183, 3, 198, 34, 184, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 184, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 184, 3, 198, 3, 185, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 177, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 185, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 185, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 186, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 177, 0, 198, 98, 186, 3, 198, 163, 186, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 187, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 187, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 187, 3, 198, 2, 188, 3, 198, 67, 188, 3, 198, 162, 188, 3, 198, 227, 188, 3, 198, 66, 189, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 189, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 189, 3, 198, 35, 190, 3, 198, 131, 190, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 190, 3, 192, 0, 0, 0, 198, 67, 191, 3, 197, 2, 178, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 191, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 191, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 178, 0, 198, 35, 192, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 178, 0, 198, 131, 192, 3, 192, 0, 0, 0, 198, 227, 192, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 193, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 193, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 194, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 179, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 194, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 194, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 195, 3, 198, 99, 195, 3, 192, 0, 0, 0, 198, 195, 195, 3, 192, 0, 0, 0, 198, 35, 196, 3, 192, 0, 0, 0, 198, 130, 196, 3, 198, 194, 196, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 197, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 197, 3, 197, 99, 179, 0, 198, 163, 197, 3, 198, 3, 198, 3, 197, 195, 179, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 198, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 180, 0, 197, 98, 180, 0, 192, 0, 0, 0, 198, 195, 198, 3, 197, 163, 180, 0, 192, 0, 0, 0, 197, 3, 181, 0, 192, 0, 0, 0, 198, 35, 199, 3, 192, 0, 0, 0, 198, 131, 199, 3, 192, 0, 0, 0, 197, 98, 181, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 199, 3, 198, 67, 200, 3, 192, 0, 0, 0, 198, 163, 200, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 201, 3, 192, 0, 0, 0, 198, 99, 201, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 201, 3, 192, 0, 0, 0, 198, 34, 202, 3, 197, 163, 181, 0, 197, 2, 182, 0, 198, 98, 202, 3, 198, 162, 202, 3, 198, 227, 202, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 203, 3, 192, 0, 0, 0, 198, 162, 203, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 203, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 204, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 204, 3, 198, 3, 205, 3, 192, 0, 0, 0, 197, 66, 182, 0, 192, 0, 0, 0, 198, 99, 205, 3, 198, 195, 205, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 182, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 206, 3, 192, 0, 0, 0, 198, 131, 206, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 182, 0, 192, 0, 0, 0, 197, 34, 183, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 206, 3, 198, 67, 207, 3, 192, 0, 0, 0, 198, 163, 207, 3, 192, 0, 0, 0, 198, 3, 208, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 183, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 208, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 208, 3, 192, 0, 0, 0, 198, 3, 209, 3, 198, 99, 209, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 183, 0, 197, 226, 183, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 184, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 209, 3, 192, 0, 0, 0, 198, 35, 210, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 210, 3, 198, 227, 210, 3, 192, 0, 0, 0, 198, 66, 211, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 211, 3, 192, 0, 0, 0, 198, 195, 211, 3, 192, 0, 0, 0, 198, 34, 212, 3, 192, 0, 0, 0, 198, 99, 212, 3, 198, 195, 212, 3, 198, 35, 213, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 213, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 213, 3, 198, 67, 214, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 214, 3, 198, 227, 214, 3, 198, 66, 215, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 184, 0, 198, 131, 215, 3, 198, 227, 215, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 184, 0, 192, 0, 0, 0, 197, 227, 184, 0, 198, 67, 216, 3, 192, 0, 0, 0, 198, 162, 216, 3, 198, 226, 216, 3, 197, 67, 185, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 185, 0, 198, 35, 217, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 186, 0, 198, 131, 217, 3, 198, 227, 217, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 218, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 186, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 218, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 218, 3, 198, 67, 219, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 219, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 220, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 220, 3, 192, 0, 0, 0, 198, 194, 220, 3, 192, 0, 0, 0, 197, 162, 186, 0, 198, 3, 221, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 221, 3, 192, 0, 0, 0, 197, 226, 186, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 221, 3, 198, 34, 222, 3, 192, 0, 0, 0, 197, 34, 187, 0, 198, 98, 222, 3, 197, 98, 187, 0, 197, 162, 187, 0, 192, 0, 0, 0, 198, 163, 222, 3, 198, 2, 223, 3, 192, 0, 0, 0, 198, 67, 223, 3, 192, 0, 0, 0, 198, 163, 223, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 224, 3, 198, 99, 224, 3, 198, 195, 224, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 225, 3, 197, 226, 187, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 188, 0, 198, 131, 225, 3, 198, 227, 225, 3, 192, 0, 0, 0, 198, 66, 226, 3, 192, 0, 0, 0, 198, 131, 226, 3, 198, 226, 226, 3, 198, 35, 227, 3, 198, 130, 227, 3, 192, 0, 0, 0, 198, 195, 227, 3, 192, 0, 0, 0, 198, 34, 228, 3, 192, 0, 0, 0, 198, 98, 228, 3, 198, 162, 228, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 228, 3, 192, 0, 0, 0, 198, 67, 229, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 229, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 229, 3, 192, 0, 0, 0, 197, 130, 188, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 230, 3, 197, 195, 188, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 230, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 230, 3, 198, 34, 231, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 231, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 231, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 189, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 232, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 232, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 189, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 232, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 233, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 233, 3, 198, 3, 234, 3, 197, 163, 189, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 234, 3, 192, 0, 0, 0, 198, 195, 234, 3, 192, 0, 0, 0, 198, 35, 235, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 235, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 235, 3, 192, 0, 0, 0, 197, 2, 190, 0, 192, 0, 0, 0, 198, 34, 236, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 236, 3, 192, 0, 0, 0, 197, 66, 190, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 236, 3, 197, 131, 190, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 236, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 237, 3, 198, 162, 237, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 190, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 237, 3, 198, 66, 238, 3, 197, 35, 191, 0, 198, 131, 238, 3, 198, 227, 238, 3, 198, 67, 239, 3, 197, 130, 191, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 239, 3, 197, 194, 191, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 239, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 240, 3, 197, 2, 192, 0, 197, 66, 192, 0, 198, 131, 240, 3, 192, 0, 0, 0, 198, 227, 240, 3, 198, 66, 241, 3, 198, 131, 241, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 192, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 192, 0, 192, 0, 0, 0, 197, 2, 193, 0, 198, 227, 241, 3, 198, 67, 242, 3, 197, 66, 193, 0, 198, 162, 242, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 242, 3, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 193, 0, 192, 0, 0, 0, 197, 194, 193, 0, 197, 2, 194, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 194, 0, 198, 35, 243, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 243, 3, 198, 195, 243, 3, 198, 34, 244, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 244, 3, 192, 0, 0, 0, 198, 194, 244, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 245, 3, 197, 162, 194, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 245, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 245, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 246, 3, 197, 226, 194, 0, 197, 34, 195, 0, 197, 99, 195, 0, 197, 194, 195, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 196, 0, 198, 99, 246, 3, 192, 0, 0, 0, 197, 99, 196, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 246, 3, 192, 0, 0, 0, 198, 35, 247, 3, 198, 130, 247, 3, 192, 0, 0, 0, 198, 195, 247, 3, 198, 35, 248, 3, 198, 130, 248, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 196, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 248, 3, 192, 0, 0, 0, 198, 34, 249, 3, 198, 99, 249, 3, 198, 194, 249, 3, 198, 3, 250, 3, 192, 0, 0, 0, 198, 99, 250, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 250, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 251, 3, 198, 131, 251, 3, 198, 226, 251, 3, 198, 34, 252, 3, 198, 98, 252, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 252, 3, 198, 227, 252, 3, 198, 66, 253, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 253, 3, 198, 227, 253, 3, 192, 0, 0, 0, 198, 67, 254, 3, 192, 0, 0, 0, 198, 162, 254, 3, 192, 0, 0, 0, 198, 227, 254, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 255, 3, 198, 162, 255, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 255, 3, 192, 0, 0, 0, 198, 67, 0, 4, 192, 0, 0, 0, 197, 34, 197, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 0, 4, 192, 0, 0, 0, 198, 2, 1, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 1, 4, 192, 0, 0, 0, 198, 130, 1, 4, 198, 195, 1, 4, 198, 35, 2, 4, 192, 0, 0, 0, 197, 99, 197, 0, 198, 131, 2, 4, 192, 0, 0, 0, 198, 227, 2, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 3, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 3, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 4, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 197, 0, 198, 99, 4, 4, 192, 0, 0, 0, 198, 194, 4, 4, 192, 0, 0, 0, 198, 3, 5, 4, 198, 99, 5, 4, 198, 194, 5, 4, 198, 2, 6, 4, 198, 67, 6, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 6, 4, 198, 2, 7, 4, 198, 67, 7, 4, 192, 0, 0, 0, 198, 163, 7, 4, 197, 2, 198, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 8, 4, 192, 0, 0, 0, 198, 98, 8, 4, 192, 0, 0, 0, 198, 163, 8, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 9, 4, 198, 99, 9, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 9, 4, 192, 0, 0, 0, 198, 34, 10, 4, 198, 99, 10, 4, 198, 195, 10, 4, 198, 34, 11, 4, 198, 98, 11, 4, 192, 0, 0, 0, 198, 162, 11, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 11, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 198, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 12, 4, 198, 163, 12, 4, 198, 3, 13, 4, 192, 0, 0, 0, 198, 99, 13, 4, 197, 162, 198, 0, 198, 194, 13, 4, 192, 0, 0, 0, 198, 3, 14, 4, 192, 0, 0, 0, 198, 98, 14, 4, 198, 163, 14, 4, 198, 3, 15, 4, 192, 0, 0, 0, 198, 99, 15, 4, 198, 195, 15, 4, 198, 35, 16, 4, 198, 130, 16, 4, 198, 194, 16, 4, 198, 3, 17, 4, 198, 99, 17, 4, 198, 194, 17, 4, 197, 226, 198, 0, 192, 0, 0, 0, 198, 3, 18, 4, 192, 0, 0, 0, 198, 99, 18, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 18, 4, 198, 34, 19, 4, 198, 98, 19, 4, 192, 0, 0, 0, 198, 163, 19, 4, 198, 3, 20, 4, 197, 34, 199, 0, 198, 99, 20, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 199, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 199, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 199, 0, 192, 0, 0, 0, 198, 195, 20, 4, 192, 0, 0, 0, 198, 34, 21, 4, 197, 34, 200, 0, 198, 99, 21, 4, 197, 98, 200, 0, 197, 163, 200, 0, 198, 194, 21, 4, 198, 3, 22, 4, 198, 98, 22, 4, 192, 0, 0, 0, 198, 163, 22, 4, 197, 2, 201, 0, 198, 2, 23, 4, 198, 67, 23, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 23, 4, 198, 3, 24, 4, 198, 98, 24, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 24, 4, 192, 0, 0, 0, 198, 3, 25, 4, 198, 98, 25, 4, 198, 163, 25, 4, 198, 3, 26, 4, 198, 98, 26, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 26, 4, 198, 3, 27, 4, 192, 0, 0, 0, 197, 66, 201, 0, 192, 0, 0, 0, 198, 98, 27, 4, 198, 162, 27, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 27, 4, 198, 35, 28, 4, 192, 0, 0, 0, 198, 131, 28, 4, 198, 227, 28, 4, 198, 67, 29, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 29, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 201, 0, 198, 2, 30, 4, 198, 67, 30, 4, 198, 163, 30, 4, 192, 0, 0, 0, 198, 3, 31, 4, 192, 0, 0, 0, 198, 98, 31, 4, 192, 0, 0, 0, 198, 163, 31, 4, 192, 0, 0, 0, 198, 3, 32, 4, 198, 99, 32, 4, 192, 0, 0, 0, 198, 194, 32, 4, 198, 3, 33, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 201, 0, 192, 0, 0, 0, 198, 98, 33, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 202, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 33, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 34, 4, 192, 0, 0, 0, 198, 67, 34, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 34, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 35, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 35, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 35, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 36, 4, 192, 0, 0, 0, 198, 131, 36, 4, 198, 227, 36, 4, 197, 130, 202, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 202, 0, 192, 0, 0, 0, 197, 2, 203, 0, 198, 67, 37, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 203, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 37, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 38, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 38, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 38, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 39, 4, 192, 0, 0, 0, 198, 98, 39, 4, 198, 163, 39, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 40, 4, 198, 67, 40, 4, 198, 163, 40, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 41, 4, 192, 0, 0, 0, 198, 99, 41, 4, 192, 0, 0, 0, 198, 195, 41, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 42, 4, 192, 0, 0, 0, 198, 131, 42, 4, 198, 227, 42, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 43, 4, 198, 163, 43, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 44, 4, 192, 0, 0, 0, 198, 99, 44, 4, 198, 194, 44, 4, 192, 0, 0, 0, 198, 3, 45, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 45, 4, 198, 163, 45, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 46, 4, 197, 130, 203, 0, 198, 99, 46, 4, 198, 195, 46, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 47, 4, 198, 131, 47, 4, 192, 0, 0, 0, 198, 227, 47, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 48, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 48, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 49, 4, 198, 99, 49, 4, 198, 194, 49, 4, 192, 0, 0, 0, 198, 3, 50, 4, 197, 194, 203, 0, 192, 0, 0, 0, 198, 99, 50, 4, 192, 0, 0, 0, 198, 195, 50, 4, 198, 35, 51, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 51, 4, 198, 195, 51, 4, 198, 35, 52, 4, 198, 131, 52, 4, 198, 226, 52, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 53, 4, 192, 0, 0, 0, 197, 3, 204, 0, 192, 0, 0, 0, 197, 98, 204, 0, 198, 131, 53, 4, 197, 162, 204, 0, 198, 227, 53, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 54, 4, 197, 226, 204, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 54, 4, 198, 2, 55, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 205, 0, 198, 67, 55, 4, 192, 0, 0, 0, 198, 162, 55, 4, 192, 0, 0, 0, 198, 227, 55, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 56, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 56, 4, 198, 226, 56, 4, 198, 35, 57, 4, 192, 0, 0, 0, 197, 99, 205, 0, 197, 195, 205, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 57, 4, 198, 194, 57, 4, 192, 0, 0, 0, 197, 34, 206, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 58, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 206, 0, 198, 98, 58, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 58, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 59, 4, 192, 0, 0, 0, 198, 66, 59, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 206, 0, 192, 0, 0, 0, 198, 131, 59, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 59, 4, 197, 2, 207, 0, 198, 35, 60, 4, 198, 130, 60, 4, 198, 194, 60, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 61, 4, 198, 99, 61, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 61, 4, 197, 66, 207, 0, 198, 35, 62, 4, 192, 0, 0, 0, 198, 130, 62, 4, 192, 0, 0, 0, 198, 194, 62, 4, 198, 3, 63, 4, 192, 0, 0, 0, 198, 99, 63, 4, 192, 0, 0, 0, 198, 195, 63, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 64, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 64, 4, 197, 130, 207, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 64, 4, 197, 194, 207, 0, 197, 2, 208, 0, 197, 67, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 65, 4, 198, 99, 65, 4, 192, 0, 0, 0, 198, 195, 65, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 66, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 66, 4, 198, 162, 66, 4, 197, 162, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 66, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 67, 4, 192, 0, 0, 0, 198, 130, 67, 4, 198, 195, 67, 4, 197, 226, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 68, 4, 198, 98, 68, 4, 192, 0, 0, 0, 198, 162, 68, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 209, 0, 198, 226, 68, 4, 197, 98, 209, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 69, 4, 198, 130, 69, 4, 198, 194, 69, 4, 192, 0, 0, 0, 198, 3, 70, 4, 192, 0, 0, 0, 198, 99, 70, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 70, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 209, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 71, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 71, 4, 198, 163, 71, 4, 198, 3, 72, 4, 198, 99, 72, 4, 192, 0, 0, 0, 198, 194, 72, 4, 192, 0, 0, 0, 198, 2, 73, 4, 197, 227, 209, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 73, 4, 198, 162, 73, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 73, 4, 198, 35, 74, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 74, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 74, 4, 198, 3, 75, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 75, 4, 192, 0, 0, 0, 197, 66, 210, 0, 198, 163, 75, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 76, 4, 192, 0, 0, 0, 198, 99, 76, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 210, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 76, 4, 198, 35, 77, 4, 192, 0, 0, 0, 197, 194, 210, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 77, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 211, 0, 198, 227, 77, 4, 192, 0, 0, 0, 198, 67, 78, 4, 198, 163, 78, 4, 192, 0, 0, 0, 198, 3, 79, 4, 198, 99, 79, 4, 198, 194, 79, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 80, 4, 198, 99, 80, 4, 198, 195, 80, 4, 198, 34, 81, 4, 198, 99, 81, 4, 198, 195, 81, 4, 198, 35, 82, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 82, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 82, 4, 198, 35, 83, 4, 198, 131, 83, 4, 198, 227, 83, 4, 192, 0, 0, 0, 198, 67, 84, 4, 198, 162, 84, 4, 192, 0, 0, 0, 198, 226, 84, 4, 198, 35, 85, 4, 192, 0, 0, 0, 198, 130, 85, 4, 197, 67, 211, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 85, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 86, 4, 192, 0, 0, 0, 198, 66, 86, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 211, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 86, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 86, 4, 197, 226, 211, 0, 197, 34, 212, 0, 192, 0, 0, 0, 197, 98, 212, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 212, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 212, 0, 192, 0, 0, 0, 197, 66, 213, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 87, 4, 192, 0, 0, 0, 198, 130, 87, 4, 192, 0, 0, 0, 198, 195, 87, 4, 192, 0, 0, 0, 197, 130, 213, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 88, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 88, 4, 198, 194, 88, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 89, 4, 192, 0, 0, 0, 198, 99, 89, 4, 192, 0, 0, 0, 198, 195, 89, 4, 192, 0, 0, 0, 198, 35, 90, 4, 197, 195, 213, 0, 198, 131, 90, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 214, 0, 192, 0, 0, 0, 197, 130, 214, 0, 192, 0, 0, 0, 198, 226, 90, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 91, 4, 192, 0, 0, 0, 198, 99, 91, 4, 198, 194, 91, 4, 192, 0, 0, 0, 198, 3, 92, 4, 198, 99, 92, 4, 192, 0, 0, 0, 198, 195, 92, 4, 192, 0, 0, 0, 198, 35, 93, 4, 198, 131, 93, 4, 198, 227, 93, 4, 192, 0, 0, 0, 198, 66, 94, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 94, 4, 192, 0, 0, 0, 198, 226, 94, 4, 192, 0, 0, 0, 198, 35, 95, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 95, 4, 198, 195, 95, 4, 192, 0, 0, 0, 198, 35, 96, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 96, 4, 198, 227, 96, 4, 192, 0, 0, 0, 197, 194, 214, 0, 198, 67, 97, 4, 192, 0, 0, 0, 198, 163, 97, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 98, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 98, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 98, 4, 197, 3, 215, 0, 198, 2, 99, 4, 198, 66, 99, 4, 198, 131, 99, 4, 198, 227, 99, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 100, 4, 198, 162, 100, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 100, 4, 192, 0, 0, 0, 198, 67, 101, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 101, 4, 198, 226, 101, 4, 192, 0, 0, 0, 198, 35, 102, 4, 198, 131, 102, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 102, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 103, 4, 198, 162, 103, 4, 192, 0, 0, 0, 198, 226, 103, 4, 198, 35, 104, 4, 198, 131, 104, 4, 192, 0, 0, 0, 198, 227, 104, 4, 192, 0, 0, 0, 198, 67, 105, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 105, 4, 197, 98, 215, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 106, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 106, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 106, 4, 198, 3, 107, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 107, 4, 198, 163, 107, 4, 198, 2, 108, 4, 198, 66, 108, 4, 198, 131, 108, 4, 192, 0, 0, 0, 198, 226, 108, 4, 198, 35, 109, 4, 192, 0, 0, 0, 198, 130, 109, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 109, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 110, 4, 192, 0, 0, 0, 198, 130, 110, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 110, 4, 198, 2, 111, 4, 197, 162, 215, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 111, 4, 192, 0, 0, 0, 198, 162, 111, 4, 197, 227, 215, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 111, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 112, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 112, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 112, 4, 192, 0, 0, 0, 198, 67, 113, 4, 192, 0, 0, 0, 198, 163, 113, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 114, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 114, 4, 192, 0, 0, 0, 198, 195, 114, 4, 198, 34, 115, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 115, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 115, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 116, 4, 192, 0, 0, 0, 198, 99, 116, 4, 198, 195, 116, 4, 198, 34, 117, 4, 192, 0, 0, 0, 198, 98, 117, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 117, 4, 192, 0, 0, 0, 198, 3, 118, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 216, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 118, 4, 192, 0, 0, 0, 198, 195, 118, 4, 198, 35, 119, 4, 198, 130, 119, 4, 198, 195, 119, 4, 192, 0, 0, 0, 198, 35, 120, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 216, 0, 198, 130, 120, 4, 197, 2, 217, 0, 192, 0, 0, 0, 198, 194, 120, 4, 192, 0, 0, 0, 198, 3, 121, 4, 192, 0, 0, 0, 198, 98, 121, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 121, 4, 197, 66, 217, 0, 198, 2, 122, 4, 198, 67, 122, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 122, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 123, 4, 198, 99, 123, 4, 198, 195, 123, 4, 198, 35, 124, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 124, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 217, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 124, 4, 192, 0, 0, 0, 198, 34, 125, 4, 198, 98, 125, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 125, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 125, 4, 198, 34, 126, 4, 197, 195, 217, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 126, 4, 197, 35, 218, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 126, 4, 192, 0, 0, 0, 198, 34, 127, 4, 192, 0, 0, 0, 197, 130, 218, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 127, 4, 192, 0, 0, 0, 198, 162, 127, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 127, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 128, 4, 198, 162, 128, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 128, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 129, 4, 198, 162, 129, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 129, 4, 198, 35, 130, 4, 192, 0, 0, 0, 197, 195, 218, 0, 198, 131, 130, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 130, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 131, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 131, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 131, 4, 198, 66, 132, 4, 197, 34, 219, 0, 198, 131, 132, 4, 192, 0, 0, 0, 198, 226, 132, 4, 192, 0, 0, 0, 198, 34, 133, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 133, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 133, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 134, 4, 198, 131, 134, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 134, 4, 198, 66, 135, 4, 198, 131, 135, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 135, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 219, 0, 197, 195, 219, 0, 198, 66, 136, 4, 192, 0, 0, 0, 198, 131, 136, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 136, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 137, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 137, 4, 192, 0, 0, 0, 198, 194, 137, 4, 198, 2, 138, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 138, 4, 198, 131, 138, 4, 198, 227, 138, 4, 198, 66, 139, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 139, 4, 198, 226, 139, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 140, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 140, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 140, 4, 198, 34, 141, 4, 192, 0, 0, 0, 197, 34, 220, 0, 192, 0, 0, 0, 198, 99, 141, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 141, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 142, 4, 198, 99, 142, 4, 192, 0, 0, 0, 198, 195, 142, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 143, 4, 198, 131, 143, 4, 198, 227, 143, 4, 198, 66, 144, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 144, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 144, 4, 192, 0, 0, 0, 198, 2, 145, 4, 198, 67, 145, 4, 192, 0, 0, 0, 198, 163, 145, 4, 198, 2, 146, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 220, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 146, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 146, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 220, 0, 198, 195, 146, 4, 198, 35, 147, 4, 197, 227, 220, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 147, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 147, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 148, 4, 192, 0, 0, 0, 198, 98, 148, 4, 198, 162, 148, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 148, 4, 198, 35, 149, 4, 198, 130, 149, 4, 198, 194, 149, 4, 198, 2, 150, 4, 198, 67, 150, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 150, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 150, 4, 198, 66, 151, 4, 198, 130, 151, 4, 198, 194, 151, 4, 198, 2, 152, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 152, 4, 198, 130, 152, 4, 197, 66, 221, 0, 198, 195, 152, 4, 198, 35, 153, 4, 198, 130, 153, 4, 198, 194, 153, 4, 192, 0, 0, 0, 198, 2, 154, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 154, 4, 192, 0, 0, 0, 198, 162, 154, 4, 197, 130, 221, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 221, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 154, 4, 192, 0, 0, 0, 198, 35, 155, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 155, 4, 197, 3, 222, 0, 198, 227, 155, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 156, 4, 198, 131, 156, 4, 192, 0, 0, 0, 197, 98, 222, 0, 198, 227, 156, 4, 198, 67, 157, 4, 192, 0, 0, 0, 198, 162, 157, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 222, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 157, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 158, 4, 197, 226, 222, 0, 198, 130, 158, 4, 192, 0, 0, 0, 198, 195, 158, 4, 198, 35, 159, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 159, 4, 192, 0, 0, 0, 198, 194, 159, 4, 198, 2, 160, 4, 192, 0, 0, 0, 198, 67, 160, 4, 198, 163, 160, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 161, 4, 198, 67, 161, 4, 198, 162, 161, 4, 192, 0, 0, 0, 198, 227, 161, 4, 197, 34, 223, 0, 198, 67, 162, 4, 198, 162, 162, 4, 198, 226, 162, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 163, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 163, 4, 198, 227, 163, 4, 197, 98, 223, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 223, 0, 198, 66, 164, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 164, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 164, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 165, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 165, 4, 198, 163, 165, 4, 198, 3, 166, 4, 192, 0, 0, 0, 198, 99, 166, 4, 198, 195, 166, 4, 192, 0, 0, 0, 198, 35, 167, 4, 198, 130, 167, 4, 198, 195, 167, 4, 198, 34, 168, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 168, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 168, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 223, 0, 198, 2, 169, 4, 198, 67, 169, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 169, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 169, 4, 192, 0, 0, 0, 197, 34, 224, 0, 198, 67, 170, 4, 198, 162, 170, 4, 198, 226, 170, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 171, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 224, 0, 192, 0, 0, 0, 198, 131, 171, 4, 197, 162, 224, 0, 192, 0, 0, 0, 198, 227, 171, 4, 192, 0, 0, 0, 198, 66, 172, 4, 197, 227, 224, 0, 198, 130, 172, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 172, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 173, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 173, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 173, 4, 198, 67, 174, 4, 192, 0, 0, 0, 198, 163, 174, 4, 192, 0, 0, 0, 198, 2, 175, 4, 198, 66, 175, 4, 198, 130, 175, 4, 192, 0, 0, 0, 198, 194, 175, 4, 198, 3, 176, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 176, 4, 198, 195, 176, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 225, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 177, 4, 198, 131, 177, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 177, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 178, 4, 198, 131, 178, 4, 197, 130, 225, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 178, 4, 192, 0, 0, 0, 197, 195, 225, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 179, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 226, 0, 198, 163, 179, 4, 198, 2, 180, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 180, 4, 198, 163, 180, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 181, 4, 198, 98, 181, 4, 198, 163, 181, 4, 198, 3, 182, 4, 198, 99, 182, 4, 198, 194, 182, 4, 198, 3, 183, 4, 198, 99, 183, 4, 198, 194, 183, 4, 197, 131, 226, 0, 192, 0, 0, 0, 198, 2, 184, 4, 192, 0, 0, 0, 197, 226, 226, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 184, 4, 198, 163, 184, 4, 192, 0, 0, 0, 198, 2, 185, 4, 198, 66, 185, 4, 198, 131, 185, 4, 198, 227, 185, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 186, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 186, 4, 198, 226, 186, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 187, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 227, 0, 197, 98, 227, 0, 192, 0, 0, 0, 198, 130, 187, 4, 192, 0, 0, 0, 198, 195, 187, 4, 198, 34, 188, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 188, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 188, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 188, 4, 198, 34, 189, 4, 198, 99, 189, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 189, 4, 198, 35, 190, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 190, 4, 198, 226, 190, 4, 192, 0, 0, 0, 198, 34, 191, 4, 198, 98, 191, 4, 192, 0, 0, 0, 198, 163, 191, 4, 198, 3, 192, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 192, 4, 192, 0, 0, 0, 198, 163, 192, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 193, 4, 192, 0, 0, 0, 198, 99, 193, 4, 198, 195, 193, 4, 192, 0, 0, 0, 198, 34, 194, 4, 198, 99, 194, 4, 192, 0, 0, 0, 198, 195, 194, 4, 198, 35, 195, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 195, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 195, 4, 192, 0, 0, 0, 198, 35, 196, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 196, 4, 192, 0, 0, 0, 198, 227, 196, 4, 198, 67, 197, 4, 198, 163, 197, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 198, 4, 192, 0, 0, 0, 198, 67, 198, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 227, 0, 192, 0, 0, 0, 198, 163, 198, 4, 192, 0, 0, 0, 198, 2, 199, 4, 192, 0, 0, 0, 198, 66, 199, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 227, 0, 192, 0, 0, 0, 198, 130, 199, 4, 198, 195, 199, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 200, 4, 198, 131, 200, 4, 192, 0, 0, 0, 198, 227, 200, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 228, 0, 192, 0, 0, 0, 198, 67, 201, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 201, 4, 198, 226, 201, 4, 198, 35, 202, 4, 198, 131, 202, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 228, 0, 198, 227, 202, 4, 192, 0, 0, 0, 198, 66, 203, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 203, 4, 192, 0, 0, 0, 198, 195, 203, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 204, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 204, 4, 197, 195, 228, 0, 192, 0, 0, 0, 198, 195, 204, 4, 198, 35, 205, 4, 192, 0, 0, 0, 198, 131, 205, 4, 198, 227, 205, 4, 198, 67, 206, 4, 192, 0, 0, 0, 198, 163, 206, 4, 198, 3, 207, 4, 198, 99, 207, 4, 198, 195, 207, 4, 198, 34, 208, 4, 198, 99, 208, 4, 192, 0, 0, 0, 198, 195, 208, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 209, 4, 198, 99, 209, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 209, 4, 198, 3, 210, 4, 198, 99, 210, 4, 192, 0, 0, 0, 198, 195, 210, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 211, 4, 192, 0, 0, 0, 198, 130, 211, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 211, 4, 192, 0, 0, 0, 198, 3, 212, 4, 192, 0, 0, 0, 198, 99, 212, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 212, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 213, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 213, 4, 192, 0, 0, 0, 198, 162, 213, 4, 197, 34, 229, 0, 198, 227, 213, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 214, 4, 192, 0, 0, 0, 198, 131, 214, 4, 198, 226, 214, 4, 192, 0, 0, 0, 198, 35, 215, 4, 192, 0, 0, 0, 198, 131, 215, 4, 198, 226, 215, 4, 192, 0, 0, 0, 198, 35, 216, 4, 198, 131, 216, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 216, 4, 198, 35, 217, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 217, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 217, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 218, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 229, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 218, 4, 192, 0, 0, 0, 198, 227, 218, 4, 197, 163, 229, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 219, 4, 192, 0, 0, 0, 198, 163, 219, 4, 198, 3, 220, 4, 192, 0, 0, 0, 198, 99, 220, 4, 198, 195, 220, 4, 197, 3, 230, 0, 197, 98, 230, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 221, 4, 192, 0, 0, 0, 198, 131, 221, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 221, 4, 192, 0, 0, 0, 198, 35, 222, 4, 192, 0, 0, 0, 198, 130, 222, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 222, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 223, 4, 192, 0, 0, 0, 198, 98, 223, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 223, 4, 198, 226, 223, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 230, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 224, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 224, 4, 192, 0, 0, 0, 198, 227, 224, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 225, 4, 192, 0, 0, 0, 198, 162, 225, 4, 192, 0, 0, 0, 197, 2, 231, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 225, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 226, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 226, 4, 198, 3, 227, 4, 198, 98, 227, 4, 198, 163, 227, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 231, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 228, 4, 198, 98, 228, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 228, 4, 197, 162, 231, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 229, 4, 198, 99, 229, 4, 198, 194, 229, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 230, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 230, 4, 198, 163, 230, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 231, 4, 192, 0, 0, 0, 198, 98, 231, 4, 192, 0, 0, 0, 198, 162, 231, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 231, 4, 198, 67, 232, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 232, 4, 198, 227, 232, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 233, 4, 198, 163, 233, 4, 192, 0, 0, 0, 198, 2, 234, 4, 192, 0, 0, 0, 198, 67, 234, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 234, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 234, 4, 198, 34, 235, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 235, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 235, 4, 192, 0, 0, 0, 198, 34, 236, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 236, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 236, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 237, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 237, 4, 192, 0, 0, 0, 198, 194, 237, 4, 198, 3, 238, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 238, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 238, 4, 192, 0, 0, 0, 198, 3, 239, 4, 198, 99, 239, 4, 198, 195, 239, 4, 198, 35, 240, 4, 198, 131, 240, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 240, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 241, 4, 198, 163, 241, 4, 198, 3, 242, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 242, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 242, 4, 198, 3, 243, 4, 198, 99, 243, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 243, 4, 198, 3, 244, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 231, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 244, 4, 198, 163, 244, 4, 198, 2, 245, 4, 192, 0, 0, 0, 198, 66, 245, 4, 198, 131, 245, 4, 192, 0, 0, 0, 198, 226, 245, 4, 198, 35, 246, 4, 198, 131, 246, 4, 198, 226, 246, 4, 198, 35, 247, 4, 192, 0, 0, 0, 198, 130, 247, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 247, 4, 192, 0, 0, 0, 198, 3, 248, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 248, 4, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 232, 0, 192, 0, 0, 0, 198, 163, 248, 4, 192, 0, 0, 0, 198, 2, 249, 4, 192, 0, 0, 0, 198, 67, 249, 4, 192, 0, 0, 0, 198, 162, 249, 4, 198, 226, 249, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 250, 4, 192, 0, 0, 0, 198, 98, 250, 4, 192, 0, 0, 0, 198, 162, 250, 4, 198, 227, 250, 4, 198, 67, 251, 4, 197, 99, 232, 0, 198, 162, 251, 4, 197, 195, 232, 0, 197, 35, 233, 0, 192, 0, 0, 0, 198, 227, 251, 4, 198, 67, 252, 4, 192, 0, 0, 0, 198, 163, 252, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 253, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 253, 4, 198, 131, 253, 4, 198, 226, 253, 4, 192, 0, 0, 0, 198, 34, 254, 4, 198, 99, 254, 4, 198, 194, 254, 4, 198, 3, 255, 4, 198, 99, 255, 4, 198, 195, 255, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 0, 5, 198, 99, 0, 5, 198, 195, 0, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 1, 5, 198, 98, 1, 5, 192, 0, 0, 0, 198, 163, 1, 5, 198, 2, 2, 5, 198, 66, 2, 5, 198, 131, 2, 5, 198, 227, 2, 5, 198, 67, 3, 5, 192, 0, 0, 0, 197, 130, 233, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 3, 5, 198, 227, 3, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 4, 5, 198, 163, 4, 5, 198, 2, 5, 5, 198, 67, 5, 5, 198, 163, 5, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 6, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 6, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 6, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 7, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 7, 5, 192, 0, 0, 0, 198, 226, 7, 5, 197, 195, 233, 0, 192, 0, 0, 0, 198, 34, 8, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 8, 5, 198, 195, 8, 5, 198, 35, 9, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 9, 5, 192, 0, 0, 0, 198, 227, 9, 5, 198, 66, 10, 5, 198, 131, 10, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 10, 5, 192, 0, 0, 0, 198, 35, 11, 5, 198, 131, 11, 5, 198, 226, 11, 5, 198, 35, 12, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 12, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 12, 5, 197, 34, 234, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 13, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 13, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 13, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 14, 5, 198, 130, 14, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 14, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 15, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 15, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 15, 5, 192, 0, 0, 0, 198, 66, 16, 5, 198, 131, 16, 5, 192, 0, 0, 0, 198, 226, 16, 5, 198, 34, 17, 5, 198, 99, 17, 5, 192, 0, 0, 0, 198, 195, 17, 5, 192, 0, 0, 0, 198, 35, 18, 5, 198, 130, 18, 5, 197, 98, 234, 0, 198, 194, 18, 5, 192, 0, 0, 0, 198, 3, 19, 5, 198, 99, 19, 5, 198, 195, 19, 5, 192, 0, 0, 0, 198, 35, 20, 5, 198, 130, 20, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 20, 5, 192, 0, 0, 0, 198, 34, 21, 5, 198, 99, 21, 5, 192, 0, 0, 0, 198, 195, 21, 5, 198, 35, 22, 5, 198, 131, 22, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 22, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 234, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 23, 5, 198, 162, 23, 5, 198, 226, 23, 5, 198, 35, 24, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 24, 5, 198, 195, 24, 5, 197, 227, 234, 0, 192, 0, 0, 0, 198, 35, 25, 5, 192, 0, 0, 0, 198, 131, 25, 5, 198, 227, 25, 5, 198, 67, 26, 5, 192, 0, 0, 0, 197, 66, 235, 0, 198, 163, 26, 5, 192, 0, 0, 0, 198, 3, 27, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 27, 5, 192, 0, 0, 0, 198, 163, 27, 5, 198, 3, 28, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 28, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 235, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 28, 5, 192, 0, 0, 0, 198, 35, 29, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 235, 0, 192, 0, 0, 0, 197, 2, 236, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 29, 5, 198, 227, 29, 5, 198, 67, 30, 5, 198, 162, 30, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 30, 5, 198, 35, 31, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 31, 5, 192, 0, 0, 0, 197, 66, 236, 0, 198, 226, 31, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 32, 5, 192, 0, 0, 0, 198, 131, 32, 5, 197, 130, 236, 0, 198, 227, 32, 5, 198, 67, 33, 5, 198, 162, 33, 5, 198, 227, 33, 5, 198, 67, 34, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 34, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 35, 5, 198, 99, 35, 5, 192, 0, 0, 0, 198, 195, 35, 5, 198, 34, 36, 5, 198, 99, 36, 5, 192, 0, 0, 0, 198, 195, 36, 5, 192, 0, 0, 0, 198, 34, 37, 5, 192, 0, 0, 0, 197, 194, 236, 0, 192, 0, 0, 0, 198, 98, 37, 5, 192, 0, 0, 0, 198, 163, 37, 5, 198, 3, 38, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 38, 5, 192, 0, 0, 0, 198, 195, 38, 5, 198, 34, 39, 5, 198, 98, 39, 5, 198, 162, 39, 5, 198, 226, 39, 5, 198, 35, 40, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 40, 5, 192, 0, 0, 0, 198, 195, 40, 5, 192, 0, 0, 0, 198, 35, 41, 5, 198, 131, 41, 5, 198, 227, 41, 5, 198, 67, 42, 5, 192, 0, 0, 0, 198, 163, 42, 5, 198, 2, 43, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 43, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 43, 5, 197, 2, 237, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 43, 5, 197, 66, 237, 0, 198, 34, 44, 5, 192, 0, 0, 0, 198, 99, 44, 5, 192, 0, 0, 0, 198, 195, 44, 5, 198, 34, 45, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 45, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 45, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 46, 5, 198, 99, 46, 5, 198, 195, 46, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 237, 0, 198, 35, 47, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 47, 5, 197, 227, 237, 0, 192, 0, 0, 0, 198, 195, 47, 5, 192, 0, 0, 0, 198, 35, 48, 5, 198, 130, 48, 5, 192, 0, 0, 0, 198, 195, 48, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 49, 5, 198, 130, 49, 5, 192, 0, 0, 0, 198, 195, 49, 5, 192, 0, 0, 0, 198, 35, 50, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 50, 5, 198, 227, 50, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 51, 5, 198, 130, 51, 5, 198, 195, 51, 5, 192, 0, 0, 0, 198, 35, 52, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 52, 5, 198, 227, 52, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 53, 5, 198, 163, 53, 5, 197, 66, 238, 0, 192, 0, 0, 0, 198, 3, 54, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 54, 5, 198, 195, 54, 5, 198, 34, 55, 5, 198, 99, 55, 5, 192, 0, 0, 0, 198, 195, 55, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 56, 5, 198, 131, 56, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 56, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 238, 0, 192, 0, 0, 0, 198, 34, 57, 5, 198, 99, 57, 5, 198, 195, 57, 5, 192, 0, 0, 0, 197, 194, 238, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 58, 5, 192, 0, 0, 0, 198, 99, 58, 5, 192, 0, 0, 0, 197, 2, 239, 0, 197, 66, 239, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 239, 0, 192, 0, 0, 0, 198, 195, 58, 5, 192, 0, 0, 0, 198, 34, 59, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 59, 5, 192, 0, 0, 0, 197, 194, 239, 0, 198, 163, 59, 5, 198, 2, 60, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 240, 0, 198, 66, 60, 5, 198, 131, 60, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 60, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 61, 5, 198, 130, 61, 5, 198, 194, 61, 5, 198, 2, 62, 5, 192, 0, 0, 0, 198, 67, 62, 5, 192, 0, 0, 0, 198, 163, 62, 5, 192, 0, 0, 0, 198, 2, 63, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 63, 5, 192, 0, 0, 0, 198, 130, 63, 5, 198, 195, 63, 5, 198, 35, 64, 5, 192, 0, 0, 0, 198, 131, 64, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 240, 0, 198, 226, 64, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 65, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 65, 5, 197, 162, 240, 0, 192, 0, 0, 0, 198, 227, 65, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 66, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 240, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 66, 5, 192, 0, 0, 0, 198, 2, 67, 5, 198, 67, 67, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 67, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 67, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 68, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 68, 5, 192, 0, 0, 0, 198, 227, 68, 5, 192, 0, 0, 0, 198, 67, 69, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 69, 5, 192, 0, 0, 0, 198, 3, 70, 5, 197, 35, 241, 0, 197, 131, 241, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 70, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 70, 5, 197, 226, 241, 0, 198, 2, 71, 5, 192, 0, 0, 0, 198, 67, 71, 5, 198, 162, 71, 5, 192, 0, 0, 0, 198, 227, 71, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 72, 5, 192, 0, 0, 0, 198, 163, 72, 5, 197, 34, 242, 0, 192, 0, 0, 0, 197, 99, 242, 0, 192, 0, 0, 0, 198, 3, 73, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 73, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 73, 5, 198, 226, 73, 5, 198, 35, 74, 5, 198, 131, 74, 5, 192, 0, 0, 0, 198, 227, 74, 5, 192, 0, 0, 0, 198, 66, 75, 5, 198, 130, 75, 5, 192, 0, 0, 0, 198, 194, 75, 5, 198, 2, 76, 5, 198, 66, 76, 5, 198, 130, 76, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 76, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 77, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 77, 5, 198, 131, 77, 5, 198, 226, 77, 5, 192, 0, 0, 0, 198, 34, 78, 5, 198, 99, 78, 5, 198, 194, 78, 5, 198, 2, 79, 5, 198, 67, 79, 5, 197, 194, 242, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 79, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 79, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 80, 5, 192, 0, 0, 0, 198, 98, 80, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 80, 5, 198, 226, 80, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 81, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 81, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 243, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 243, 0, 192, 0, 0, 0, 198, 163, 81, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 82, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 82, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 82, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 83, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 83, 5, 198, 163, 83, 5, 192, 0, 0, 0, 198, 3, 84, 5, 192, 0, 0, 0, 197, 162, 243, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 84, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 84, 5, 198, 35, 85, 5, 198, 130, 85, 5, 198, 195, 85, 5, 198, 35, 86, 5, 192, 0, 0, 0, 198, 130, 86, 5, 198, 195, 86, 5, 198, 35, 87, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 87, 5, 198, 227, 87, 5, 198, 67, 88, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 88, 5, 198, 227, 88, 5, 198, 66, 89, 5, 198, 130, 89, 5, 192, 0, 0, 0, 198, 195, 89, 5, 192, 0, 0, 0, 198, 35, 90, 5, 198, 130, 90, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 243, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 90, 5, 192, 0, 0, 0, 198, 2, 91, 5, 192, 0, 0, 0, 198, 66, 91, 5, 198, 131, 91, 5, 192, 0, 0, 0, 198, 227, 91, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 92, 5, 198, 130, 92, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 92, 5, 198, 3, 93, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 93, 5, 198, 195, 93, 5, 198, 35, 94, 5, 192, 0, 0, 0, 198, 131, 94, 5, 192, 0, 0, 0, 198, 227, 94, 5, 192, 0, 0, 0, 198, 67, 95, 5, 198, 163, 95, 5, 192, 0, 0, 0, 197, 66, 244, 0, 198, 2, 96, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 244, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 244, 0, 198, 66, 96, 5, 192, 0, 0, 0, 198, 131, 96, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 96, 5, 192, 0, 0, 0, 198, 34, 97, 5, 192, 0, 0, 0, 197, 2, 245, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 97, 5, 198, 163, 97, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 98, 5, 198, 66, 98, 5, 198, 131, 98, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 98, 5, 192, 0, 0, 0, 198, 34, 99, 5, 197, 66, 245, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 99, 5, 198, 194, 99, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 100, 5, 192, 0, 0, 0, 198, 67, 100, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 100, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 245, 0, 197, 227, 245, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 101, 5, 198, 98, 101, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 101, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 246, 0, 197, 162, 246, 0, 197, 227, 246, 0, 198, 227, 101, 5, 198, 67, 102, 5, 198, 163, 102, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 103, 5, 192, 0, 0, 0, 198, 99, 103, 5, 192, 0, 0, 0, 198, 195, 103, 5, 198, 35, 104, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 104, 5, 198, 227, 104, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 247, 0, 192, 0, 0, 0, 198, 66, 105, 5, 192, 0, 0, 0, 198, 131, 105, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 105, 5, 198, 34, 106, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 106, 5, 198, 194, 106, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 107, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 107, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 107, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 108, 5, 198, 99, 108, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 108, 5, 192, 0, 0, 0, 198, 34, 109, 5, 198, 98, 109, 5, 198, 162, 109, 5, 198, 226, 109, 5, 198, 35, 110, 5, 198, 131, 110, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 110, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 111, 5, 198, 131, 111, 5, 198, 226, 111, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 112, 5, 192, 0, 0, 0, 198, 131, 112, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 112, 5, 198, 35, 113, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 113, 5, 192, 0, 0, 0, 198, 195, 113, 5, 197, 131, 247, 0, 198, 35, 114, 5, 198, 131, 114, 5, 198, 227, 114, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 247, 0, 198, 67, 115, 5, 192, 0, 0, 0, 197, 34, 248, 0, 198, 163, 115, 5, 198, 2, 116, 5, 192, 0, 0, 0, 198, 66, 116, 5, 192, 0, 0, 0, 198, 131, 116, 5, 192, 0, 0, 0, 198, 227, 116, 5, 198, 66, 117, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 117, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 117, 5, 192, 0, 0, 0, 198, 3, 118, 5, 198, 99, 118, 5, 198, 195, 118, 5, 192, 0, 0, 0, 198, 34, 119, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 119, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 119, 5, 198, 34, 120, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 120, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 120, 5, 198, 34, 121, 5, 197, 98, 248, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 121, 5, 198, 163, 121, 5, 198, 3, 122, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 122, 5, 197, 162, 248, 0, 198, 195, 122, 5, 198, 35, 123, 5, 192, 0, 0, 0, 198, 130, 123, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 123, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 124, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 248, 0, 192, 0, 0, 0, 198, 99, 124, 5, 192, 0, 0, 0, 198, 194, 124, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 125, 5, 198, 99, 125, 5, 197, 34, 249, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 125, 5, 192, 0, 0, 0, 198, 3, 126, 5, 198, 99, 126, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 126, 5, 192, 0, 0, 0, 197, 98, 249, 0, 192, 0, 0, 0, 198, 35, 127, 5, 198, 131, 127, 5, 192, 0, 0, 0, 198, 226, 127, 5, 192, 0, 0, 0, 198, 35, 128, 5, 197, 162, 249, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 128, 5, 197, 226, 249, 0, 192, 0, 0, 0, 198, 195, 128, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 129, 5, 198, 99, 129, 5, 198, 194, 129, 5, 192, 0, 0, 0, 197, 34, 250, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 250, 0, 198, 3, 130, 5, 198, 99, 130, 5, 192, 0, 0, 0, 198, 194, 130, 5, 192, 0, 0, 0, 197, 162, 250, 0, 197, 227, 250, 0, 192, 0, 0, 0, 198, 2, 131, 5, 192, 0, 0, 0, 198, 67, 131, 5, 198, 162, 131, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 131, 5, 192, 0, 0, 0, 198, 34, 132, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 132, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 132, 5, 198, 3, 133, 5, 192, 0, 0, 0, 198, 98, 133, 5, 198, 162, 133, 5, 192, 0, 0, 0, 198, 227, 133, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 134, 5, 198, 163, 134, 5, 198, 3, 135, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 135, 5, 192, 0, 0, 0, 198, 194, 135, 5, 198, 3, 136, 5, 198, 99, 136, 5, 198, 194, 136, 5, 198, 3, 137, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 137, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 137, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 251, 0, 192, 0, 0, 0, 198, 3, 138, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 138, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 138, 5, 198, 3, 139, 5, 198, 98, 139, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 139, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 251, 0, 192, 0, 0, 0, 197, 195, 251, 0, 192, 0, 0, 0, 198, 3, 140, 5, 192, 0, 0, 0, 198, 98, 140, 5, 198, 162, 140, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 140, 5, 198, 67, 141, 5, 192, 0, 0, 0, 198, 163, 141, 5, 198, 3, 142, 5, 192, 0, 0, 0, 198, 99, 142, 5, 192, 0, 0, 0, 198, 195, 142, 5, 192, 0, 0, 0, 198, 34, 143, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 143, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 143, 5, 198, 2, 144, 5, 192, 0, 0, 0, 198, 67, 144, 5, 198, 163, 144, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 145, 5, 192, 0, 0, 0, 197, 34, 252, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 145, 5, 198, 162, 145, 5, 198, 227, 145, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 146, 5, 192, 0, 0, 0, 198, 131, 146, 5, 192, 0, 0, 0, 198, 227, 146, 5, 198, 67, 147, 5, 198, 162, 147, 5, 192, 0, 0, 0, 198, 226, 147, 5, 197, 98, 252, 0, 198, 34, 148, 5, 198, 99, 148, 5, 198, 195, 148, 5, 198, 34, 149, 5, 198, 99, 149, 5, 198, 195, 149, 5, 197, 162, 252, 0, 197, 226, 252, 0, 198, 35, 150, 5, 198, 130, 150, 5, 197, 34, 253, 0, 192, 0, 0, 0, 198, 195, 150, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 151, 5, 198, 99, 151, 5, 198, 195, 151, 5, 192, 0, 0, 0, 197, 98, 253, 0, 198, 34, 152, 5, 198, 98, 152, 5, 198, 162, 152, 5, 198, 227, 152, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 153, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 153, 5, 198, 227, 153, 5, 198, 67, 154, 5, 197, 162, 253, 0, 198, 162, 154, 5, 192, 0, 0, 0, 198, 226, 154, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 155, 5, 198, 99, 155, 5, 192, 0, 0, 0, 198, 194, 155, 5, 198, 2, 156, 5, 198, 67, 156, 5, 198, 162, 156, 5, 192, 0, 0, 0, 198, 226, 156, 5, 192, 0, 0, 0, 198, 35, 157, 5, 198, 130, 157, 5, 197, 226, 253, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 157, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 158, 5, 192, 0, 0, 0, 198, 131, 158, 5, 197, 34, 254, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 158, 5, 198, 66, 159, 5, 198, 130, 159, 5, 192, 0, 0, 0, 198, 194, 159, 5, 197, 99, 254, 0, 198, 2, 160, 5, 198, 67, 160, 5, 198, 163, 160, 5, 198, 2, 161, 5, 198, 67, 161, 5, 198, 162, 161, 5, 198, 226, 161, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 162, 5, 198, 99, 162, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 162, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 163, 5, 198, 131, 163, 5, 198, 226, 163, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 164, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 164, 5, 192, 0, 0, 0, 198, 163, 164, 5, 198, 3, 165, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 165, 5, 192, 0, 0, 0, 198, 194, 165, 5, 198, 2, 166, 5, 198, 67, 166, 5, 197, 194, 254, 0, 198, 163, 166, 5, 198, 2, 167, 5, 192, 0, 0, 0, 198, 67, 167, 5, 197, 2, 255, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 167, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 168, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 168, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 168, 5, 192, 0, 0, 0, 198, 34, 169, 5, 198, 99, 169, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 169, 5, 198, 35, 170, 5, 198, 130, 170, 5, 197, 67, 255, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 170, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 255, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 171, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 171, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 171, 5, 198, 227, 171, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 172, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 172, 5, 198, 194, 172, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 173, 5, 192, 0, 0, 0, 198, 99, 173, 5, 192, 0, 0, 0, 198, 195, 173, 5, 198, 35, 174, 5, 192, 0, 0, 0, 197, 226, 255, 0, 198, 131, 174, 5, 197, 35, 0, 1, 198, 226, 174, 5, 192, 0, 0, 0, 197, 131, 0, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 175, 5, 198, 130, 175, 5, 198, 194, 175, 5, 198, 3, 176, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 176, 5, 198, 195, 176, 5, 198, 34, 177, 5, 192, 0, 0, 0, 198, 99, 177, 5, 198, 195, 177, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 178, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 178, 5, 198, 227, 178, 5, 192, 0, 0, 0, 198, 67, 179, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 0, 1, 192, 0, 0, 0, 197, 34, 1, 1, 198, 163, 179, 5, 192, 0, 0, 0, 198, 3, 180, 5, 192, 0, 0, 0, 198, 99, 180, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 180, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 181, 5, 197, 98, 1, 1, 197, 163, 1, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 181, 5, 192, 0, 0, 0, 198, 131, 181, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 2, 1, 198, 226, 181, 5, 198, 35, 182, 5, 198, 130, 182, 5, 192, 0, 0, 0, 198, 195, 182, 5, 192, 0, 0, 0, 198, 35, 183, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 2, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 183, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 2, 1, 198, 226, 183, 5, 192, 0, 0, 0, 198, 34, 184, 5, 198, 99, 184, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 184, 5, 197, 194, 2, 1, 198, 3, 185, 5, 198, 99, 185, 5, 197, 2, 3, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 3, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 185, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 3, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 3, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 4, 1, 198, 35, 186, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 186, 5, 198, 227, 186, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 187, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 187, 5, 198, 227, 187, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 188, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 4, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 188, 5, 198, 3, 189, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 189, 5, 192, 0, 0, 0, 198, 163, 189, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 4, 1, 198, 2, 190, 5, 197, 34, 5, 1, 198, 67, 190, 5, 198, 162, 190, 5, 192, 0, 0, 0, 198, 227, 190, 5, 192, 0, 0, 0, 198, 67, 191, 5, 192, 0, 0, 0, 198, 162, 191, 5, 192, 0, 0, 0, 198, 226, 191, 5, 192, 0, 0, 0, 197, 99, 5, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 192, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 5, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 192, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 192, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 193, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 193, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 193, 5, 192, 0, 0, 0, 198, 3, 194, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 194, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 194, 5, 192, 0, 0, 0, 198, 2, 195, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 195, 5, 192, 0, 0, 0, 198, 131, 195, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 195, 5, 198, 67, 196, 5, 198, 163, 196, 5, 198, 3, 197, 5, 198, 99, 197, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 6, 1, 198, 195, 197, 5, 198, 35, 198, 5, 198, 131, 198, 5, 192, 0, 0, 0, 198, 227, 198, 5, 192, 0, 0, 0, 198, 67, 199, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 199, 5, 198, 3, 200, 5, 198, 98, 200, 5, 192, 0, 0, 0, 198, 162, 200, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 200, 5, 198, 66, 201, 5, 192, 0, 0, 0, 198, 130, 201, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 201, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 202, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 202, 5, 198, 163, 202, 5, 198, 3, 203, 5, 198, 99, 203, 5, 192, 0, 0, 0, 198, 195, 203, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 204, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 204, 5, 197, 67, 6, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 204, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 205, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 205, 5, 192, 0, 0, 0, 198, 195, 205, 5, 198, 35, 206, 5, 197, 162, 6, 1, 192, 0, 0, 0, 198, 130, 206, 5, 198, 195, 206, 5, 192, 0, 0, 0, 198, 35, 207, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 207, 5, 198, 194, 207, 5, 192, 0, 0, 0, 198, 2, 208, 5, 192, 0, 0, 0, 198, 67, 208, 5, 192, 0, 0, 0, 198, 163, 208, 5, 198, 2, 209, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 209, 5, 192, 0, 0, 0, 198, 162, 209, 5, 198, 227, 209, 5, 192, 0, 0, 0, 198, 67, 210, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 1, 192, 0, 0, 0, 197, 34, 7, 1, 192, 0, 0, 0, 198, 163, 210, 5, 192, 0, 0, 0, 198, 3, 211, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 211, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 211, 5, 198, 2, 212, 5, 198, 66, 212, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 212, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 212, 5, 192, 0, 0, 0, 198, 34, 213, 5, 198, 98, 213, 5, 192, 0, 0, 0, 198, 162, 213, 5, 197, 98, 7, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 7, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 213, 5, 192, 0, 0, 0, 198, 67, 214, 5, 198, 162, 214, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 214, 5, 198, 67, 215, 5, 198, 162, 215, 5, 198, 227, 215, 5, 192, 0, 0, 0, 198, 66, 216, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 216, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 216, 5, 192, 0, 0, 0, 198, 35, 217, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 217, 5, 192, 0, 0, 0, 198, 226, 217, 5, 198, 34, 218, 5, 192, 0, 0, 0, 198, 98, 218, 5, 198, 162, 218, 5, 192, 0, 0, 0, 198, 227, 218, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 7, 1, 192, 0, 0, 0, 198, 67, 219, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 219, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 220, 5, 192, 0, 0, 0, 197, 34, 8, 1, 192, 0, 0, 0, 198, 99, 220, 5, 192, 0, 0, 0, 198, 195, 220, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 8, 1, 198, 35, 221, 5, 198, 131, 221, 5, 192, 0, 0, 0, 197, 163, 8, 1, 198, 227, 221, 5, 198, 67, 222, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 222, 5, 197, 3, 9, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 9, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 222, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 223, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 223, 5, 198, 195, 223, 5, 198, 35, 224, 5, 198, 130, 224, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 224, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 9, 1, 197, 226, 9, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 225, 5, 198, 99, 225, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 225, 5, 198, 34, 226, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 10, 1, 192, 0, 0, 0, 198, 98, 226, 5, 198, 163, 226, 5, 198, 2, 227, 5, 192, 0, 0, 0, 198, 67, 227, 5, 192, 0, 0, 0, 198, 163, 227, 5, 198, 3, 228, 5, 192, 0, 0, 0, 197, 98, 10, 1, 192, 0, 0, 0, 198, 98, 228, 5, 192, 0, 0, 0, 198, 162, 228, 5, 198, 226, 228, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 229, 5, 198, 131, 229, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 229, 5, 198, 67, 230, 5, 198, 163, 230, 5, 192, 0, 0, 0, 198, 3, 231, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 231, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 231, 5, 192, 0, 0, 0, 198, 3, 232, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 232, 5, 192, 0, 0, 0, 198, 195, 232, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 233, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 233, 5, 192, 0, 0, 0, 198, 162, 233, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 10, 1, 192, 0, 0, 0, 197, 227, 10, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 11, 1, 198, 227, 233, 5, 197, 130, 11, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 11, 1, 198, 66, 234, 5, 192, 0, 0, 0, 198, 130, 234, 5, 197, 2, 12, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 12, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 12, 1, 198, 194, 234, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 235, 5, 192, 0, 0, 0, 198, 67, 235, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 235, 5, 198, 227, 235, 5, 197, 194, 12, 1, 198, 67, 236, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 236, 5, 198, 2, 237, 5, 198, 67, 237, 5, 198, 162, 237, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 237, 5, 192, 0, 0, 0, 198, 67, 238, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 238, 5, 198, 2, 239, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 239, 5, 197, 2, 13, 1, 198, 130, 239, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 239, 5, 198, 2, 240, 5, 198, 67, 240, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 240, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 240, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 241, 5, 197, 66, 13, 1, 198, 130, 241, 5, 192, 0, 0, 0, 198, 195, 241, 5, 192, 0, 0, 0, 198, 35, 242, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 13, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 13, 1, 192, 0, 0, 0, 198, 130, 242, 5, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 14, 1, 198, 194, 242, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 243, 5, 197, 99, 14, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 243, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 243, 5, 198, 35, 244, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 244, 5, 198, 195, 244, 5, 192, 0, 0, 0, 198, 34, 245, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 245, 5, 192, 0, 0, 0, 198, 163, 245, 5, 192, 0, 0, 0, 198, 3, 246, 5, 198, 99, 246, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 246, 5, 192, 0, 0, 0, 198, 34, 247, 5, 198, 98, 247, 5, 198, 162, 247, 5, 198, 227, 247, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 248, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 248, 5, 192, 0, 0, 0, 198, 226, 248, 5, 192, 0, 0, 0, 198, 35, 249, 5, 198, 130, 249, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 249, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 250, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 250, 5, 198, 194, 250, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 251, 5, 198, 99, 251, 5, 192, 0, 0, 0, 198, 195, 251, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 252, 5, 198, 99, 252, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 252, 5, 198, 35, 253, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 253, 5, 192, 0, 0, 0, 198, 195, 253, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 254, 5, 192, 0, 0, 0, 197, 195, 14, 1, 192, 0, 0, 0, 198, 131, 254, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 15, 1, 198, 227, 254, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 255, 5, 198, 130, 255, 5, 197, 98, 15, 1, 192, 0, 0, 0, 198, 195, 255, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 0, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 15, 1, 192, 0, 0, 0, 198, 131, 0, 6, 192, 0, 0, 0, 197, 226, 15, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 0, 6, 198, 67, 1, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 1, 6, 198, 3, 2, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 2, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 2, 6, 198, 227, 2, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 3, 6, 198, 163, 3, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 4, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 4, 6, 198, 163, 4, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 5, 6, 198, 66, 5, 6, 192, 0, 0, 0, 198, 131, 5, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 5, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 16, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 6, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 6, 6, 198, 195, 6, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 7, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 7, 6, 198, 227, 7, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 8, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 8, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 8, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 9, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 9, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 9, 6, 192, 0, 0, 0, 197, 130, 16, 1, 192, 0, 0, 0, 197, 194, 16, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 10, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 10, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 11, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 11, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 11, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 12, 6, 192, 0, 0, 0, 197, 2, 17, 1, 198, 131, 12, 6, 197, 66, 17, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 12, 6, 198, 35, 13, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 13, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 13, 6, 192, 0, 0, 0, 198, 35, 14, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 17, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 14, 6, 192, 0, 0, 0, 198, 227, 14, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 15, 6, 192, 0, 0, 0, 198, 163, 15, 6, 198, 3, 16, 6, 198, 99, 16, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 16, 6, 198, 35, 17, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 17, 6, 192, 0, 0, 0, 198, 194, 17, 6, 192, 0, 0, 0, 197, 226, 17, 1, 192, 0, 0, 0, 198, 3, 18, 6, 198, 98, 18, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 18, 6, 198, 2, 19, 6, 197, 34, 18, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 18, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 19, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 19, 6, 192, 0, 0, 0, 198, 2, 20, 6, 192, 0, 0, 0, 198, 67, 20, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 20, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 21, 6, 198, 99, 21, 6, 198, 195, 21, 6, 192, 0, 0, 0, 198, 35, 22, 6, 198, 130, 22, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 22, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 23, 6, 192, 0, 0, 0, 198, 66, 23, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 23, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 23, 6, 198, 35, 24, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 18, 1, 192, 0, 0, 0, 198, 130, 24, 6, 198, 195, 24, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 25, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 25, 6, 197, 226, 18, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 25, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 26, 6, 198, 98, 26, 6, 198, 163, 26, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 27, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 27, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 27, 6, 192, 0, 0, 0, 198, 2, 28, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 28, 6, 192, 0, 0, 0, 197, 34, 19, 1, 197, 98, 19, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 28, 6, 198, 195, 28, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 29, 6, 198, 99, 29, 6, 192, 0, 0, 0, 198, 194, 29, 6, 192, 0, 0, 0, 198, 2, 30, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 30, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 30, 6, 192, 0, 0, 0, 197, 162, 19, 1, 192, 0, 0, 0, 198, 195, 30, 6, 198, 34, 31, 6, 198, 98, 31, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 31, 6, 198, 226, 31, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 19, 1, 198, 34, 32, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 20, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 32, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 32, 6, 192, 0, 0, 0, 198, 35, 33, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 33, 6, 192, 0, 0, 0, 198, 226, 33, 6, 198, 34, 34, 6, 192, 0, 0, 0, 198, 98, 34, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 20, 1, 198, 162, 34, 6, 197, 162, 20, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 34, 6, 192, 0, 0, 0, 198, 67, 35, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 20, 1, 197, 34, 21, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 35, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 36, 6, 192, 0, 0, 0, 198, 98, 36, 6, 192, 0, 0, 0, 198, 163, 36, 6, 198, 3, 37, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 37, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 37, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 38, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 38, 6, 192, 0, 0, 0, 198, 194, 38, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 39, 6, 192, 0, 0, 0, 198, 99, 39, 6, 198, 195, 39, 6, 198, 35, 40, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 40, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 40, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 21, 1, 192, 0, 0, 0, 198, 67, 41, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 41, 6, 192, 0, 0, 0, 198, 3, 42, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 42, 6, 192, 0, 0, 0, 198, 162, 42, 6, 192, 0, 0, 0, 198, 227, 42, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 43, 6, 197, 194, 21, 1, 197, 3, 22, 1, 198, 162, 43, 6, 192, 0, 0, 0, 198, 227, 43, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 44, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 44, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 44, 6, 198, 34, 45, 6, 198, 99, 45, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 45, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 46, 6, 198, 130, 46, 6, 198, 195, 46, 6, 192, 0, 0, 0, 198, 34, 47, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 47, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 47, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 22, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 48, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 48, 6, 192, 0, 0, 0, 198, 195, 48, 6, 198, 35, 49, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 49, 6, 192, 0, 0, 0, 198, 227, 49, 6, 192, 0, 0, 0, 198, 67, 50, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 50, 6, 192, 0, 0, 0, 198, 226, 50, 6, 198, 34, 51, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 22, 1, 192, 0, 0, 0, 198, 99, 51, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 51, 6, 192, 0, 0, 0, 198, 35, 52, 6, 198, 131, 52, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 52, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 53, 6, 192, 0, 0, 0, 198, 163, 53, 6, 198, 2, 54, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 54, 6, 198, 162, 54, 6, 198, 227, 54, 6, 198, 67, 55, 6, 192, 0, 0, 0, 198, 163, 55, 6, 192, 0, 0, 0, 198, 2, 56, 6, 192, 0, 0, 0, 198, 67, 56, 6, 198, 162, 56, 6, 192, 0, 0, 0, 197, 227, 22, 1, 198, 227, 56, 6, 198, 67, 57, 6, 197, 67, 23, 1, 198, 163, 57, 6, 198, 2, 58, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 23, 1, 192, 0, 0, 0, 198, 66, 58, 6, 192, 0, 0, 0, 198, 131, 58, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 23, 1, 198, 227, 58, 6, 192, 0, 0, 0, 198, 67, 59, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 59, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 60, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 24, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 60, 6, 192, 0, 0, 0, 198, 162, 60, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 60, 6, 198, 66, 61, 6, 198, 131, 61, 6, 192, 0, 0, 0, 198, 226, 61, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 62, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 62, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 62, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 62, 6, 198, 34, 63, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 63, 6, 198, 163, 63, 6, 197, 130, 24, 1, 198, 3, 64, 6, 198, 99, 64, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 24, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 64, 6, 192, 0, 0, 0, 198, 3, 65, 6, 197, 2, 25, 1, 198, 99, 65, 6, 197, 66, 25, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 65, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 66, 6, 192, 0, 0, 0, 198, 66, 66, 6, 198, 130, 66, 6, 197, 130, 25, 1, 198, 195, 66, 6, 192, 0, 0, 0, 197, 195, 25, 1, 197, 34, 26, 1, 197, 98, 26, 1, 198, 35, 67, 6, 192, 0, 0, 0, 198, 131, 67, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 67, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 26, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 26, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 68, 6, 192, 0, 0, 0, 198, 130, 68, 6, 198, 194, 68, 6, 198, 3, 69, 6, 197, 34, 27, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 69, 6, 198, 163, 69, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 70, 6, 192, 0, 0, 0, 198, 98, 70, 6, 192, 0, 0, 0, 198, 163, 70, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 27, 1, 192, 0, 0, 0, 198, 2, 71, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 71, 6, 192, 0, 0, 0, 198, 131, 71, 6, 198, 226, 71, 6, 198, 35, 72, 6, 198, 130, 72, 6, 198, 194, 72, 6, 198, 2, 73, 6, 198, 67, 73, 6, 198, 162, 73, 6, 197, 162, 27, 1, 192, 0, 0, 0, 198, 227, 73, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 74, 6, 192, 0, 0, 0, 197, 226, 27, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 74, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 28, 1, 192, 0, 0, 0, 198, 195, 74, 6, 198, 34, 75, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 75, 6, 192, 0, 0, 0, 198, 195, 75, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 76, 6, 192, 0, 0, 0, 198, 130, 76, 6, 192, 0, 0, 0, 198, 195, 76, 6, 192, 0, 0, 0, 198, 35, 77, 6, 192, 0, 0, 0, 197, 98, 28, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 77, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 77, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 28, 1, 198, 67, 78, 6, 192, 0, 0, 0, 198, 163, 78, 6, 198, 2, 79, 6, 197, 227, 28, 1, 198, 66, 79, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 79, 6, 192, 0, 0, 0, 197, 66, 29, 1, 197, 130, 29, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 79, 6, 198, 67, 80, 6, 192, 0, 0, 0, 197, 194, 29, 1, 198, 162, 80, 6, 192, 0, 0, 0, 198, 226, 80, 6, 197, 2, 30, 1, 192, 0, 0, 0, 198, 35, 81, 6, 197, 67, 30, 1, 192, 0, 0, 0, 198, 131, 81, 6, 198, 227, 81, 6, 197, 162, 30, 1, 192, 0, 0, 0, 198, 67, 82, 6, 192, 0, 0, 0, 197, 227, 30, 1, 192, 0, 0, 0, 198, 163, 82, 6, 192, 0, 0, 0, 198, 2, 83, 6, 198, 67, 83, 6, 197, 66, 31, 1, 192, 0, 0, 0, 198, 163, 83, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 84, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 84, 6, 192, 0, 0, 0, 198, 162, 84, 6, 192, 0, 0, 0, 198, 227, 84, 6, 197, 130, 31, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 31, 1, 192, 0, 0, 0, 198, 67, 85, 6, 198, 163, 85, 6, 192, 0, 0, 0, 198, 3, 86, 6, 198, 99, 86, 6, 198, 195, 86, 6, 198, 35, 87, 6, 198, 130, 87, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 87, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 88, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 88, 6, 192, 0, 0, 0, 198, 162, 88, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 88, 6, 198, 66, 89, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 89, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 89, 6, 192, 0, 0, 0, 198, 3, 90, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 90, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 32, 1, 192, 0, 0, 0, 197, 98, 32, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 90, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 32, 1, 192, 0, 0, 0, 197, 227, 32, 1, 198, 34, 91, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 91, 6, 192, 0, 0, 0, 198, 195, 91, 6, 192, 0, 0, 0, 198, 35, 92, 6, 198, 130, 92, 6, 198, 195, 92, 6, 192, 0, 0, 0, 198, 34, 93, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 93, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 93, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 94, 6, 198, 98, 94, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 94, 6, 197, 66, 33, 1, 198, 3, 95, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 95, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 33, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 33, 1, 192, 0, 0, 0, 197, 2, 34, 1, 198, 195, 95, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 96, 6, 198, 99, 96, 6, 192, 0, 0, 0, 198, 195, 96, 6, 192, 0, 0, 0, 198, 34, 97, 6, 192, 0, 0, 0, 198, 98, 97, 6, 192, 0, 0, 0, 198, 163, 97, 6, 198, 3, 98, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 98, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 98, 6, 192, 0, 0, 0, 198, 35, 99, 6, 192, 0, 0, 0, 198, 131, 99, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 99, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 100, 6, 192, 0, 0, 0, 197, 67, 34, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 100, 6, 198, 195, 100, 6, 198, 34, 101, 6, 198, 98, 101, 6, 198, 163, 101, 6, 192, 0, 0, 0, 198, 3, 102, 6, 192, 0, 0, 0, 198, 98, 102, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 102, 6, 198, 227, 102, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 103, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 103, 6, 198, 3, 104, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 34, 1, 192, 0, 0, 0, 197, 226, 34, 1, 197, 34, 35, 1, 198, 98, 104, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 104, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 35, 1, 197, 162, 35, 1, 198, 2, 105, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 105, 6, 198, 131, 105, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 105, 6, 192, 0, 0, 0, 198, 66, 106, 6, 198, 130, 106, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 106, 6, 198, 2, 107, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 107, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 35, 1, 198, 163, 107, 6, 192, 0, 0, 0, 198, 2, 108, 6, 198, 66, 108, 6, 198, 131, 108, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 108, 6, 197, 34, 36, 1, 198, 67, 109, 6, 197, 99, 36, 1, 198, 163, 109, 6, 198, 3, 110, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 36, 1, 198, 98, 110, 6, 192, 0, 0, 0, 198, 162, 110, 6, 198, 226, 110, 6, 198, 34, 111, 6, 192, 0, 0, 0, 197, 3, 37, 1, 198, 98, 111, 6, 198, 162, 111, 6, 192, 0, 0, 0, 197, 99, 37, 1, 198, 226, 111, 6, 198, 35, 112, 6, 198, 130, 112, 6, 192, 0, 0, 0, 198, 194, 112, 6, 197, 194, 37, 1, 198, 3, 113, 6, 198, 98, 113, 6, 198, 163, 113, 6, 197, 2, 38, 1, 192, 0, 0, 0, 198, 3, 114, 6, 198, 98, 114, 6, 198, 162, 114, 6, 192, 0, 0, 0, 198, 227, 114, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 115, 6, 192, 0, 0, 0, 198, 163, 115, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 116, 6, 198, 66, 116, 6, 192, 0, 0, 0, 198, 130, 116, 6, 192, 0, 0, 0, 198, 195, 116, 6, 192, 0, 0, 0, 198, 34, 117, 6, 198, 99, 117, 6, 198, 195, 117, 6, 198, 34, 118, 6, 198, 99, 118, 6, 192, 0, 0, 0, 198, 194, 118, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 119, 6, 192, 0, 0, 0, 198, 99, 119, 6, 198, 194, 119, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 120, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 120, 6, 198, 130, 120, 6, 198, 194, 120, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 121, 6, 197, 66, 38, 1, 192, 0, 0, 0, 198, 67, 121, 6, 198, 163, 121, 6, 192, 0, 0, 0, 198, 3, 122, 6, 198, 99, 122, 6, 192, 0, 0, 0, 198, 195, 122, 6, 192, 0, 0, 0, 198, 35, 123, 6, 198, 131, 123, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 123, 6, 198, 66, 124, 6, 192, 0, 0, 0, 198, 131, 124, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 124, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 125, 6, 192, 0, 0, 0, 198, 162, 125, 6, 198, 227, 125, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 38, 1, 192, 0, 0, 0, 198, 66, 126, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 126, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 126, 6, 198, 2, 127, 6, 192, 0, 0, 0, 198, 67, 127, 6, 198, 162, 127, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 127, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 128, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 128, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 128, 6, 192, 0, 0, 0, 197, 227, 38, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 129, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 129, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 39, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 39, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 129, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 129, 6, 198, 67, 130, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 130, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 130, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 131, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 39, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 131, 6, 198, 195, 131, 6, 192, 0, 0, 0, 198, 34, 132, 6, 198, 98, 132, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 132, 6, 198, 3, 133, 6, 198, 98, 133, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 40, 1, 192, 0, 0, 0, 198, 162, 133, 6, 192, 0, 0, 0, 198, 227, 133, 6, 192, 0, 0, 0, 198, 67, 134, 6, 192, 0, 0, 0, 198, 163, 134, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 40, 1, 192, 0, 0, 0, 197, 130, 40, 1, 198, 3, 135, 6, 198, 98, 135, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 135, 6, 192, 0, 0, 0, 197, 194, 40, 1, 192, 0, 0, 0, 198, 226, 135, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 136, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 136, 6, 192, 0, 0, 0, 198, 226, 136, 6, 198, 35, 137, 6, 198, 131, 137, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 137, 6, 198, 67, 138, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 138, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 139, 6, 198, 66, 139, 6, 198, 130, 139, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 139, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 41, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 140, 6, 198, 131, 140, 6, 197, 66, 41, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 140, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 141, 6, 192, 0, 0, 0, 197, 130, 41, 1, 192, 0, 0, 0, 197, 194, 41, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 141, 6, 198, 226, 141, 6, 198, 34, 142, 6, 197, 2, 42, 1, 192, 0, 0, 0, 198, 99, 142, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 42, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 142, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 143, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 143, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 143, 6, 198, 67, 144, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 144, 6, 198, 3, 145, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 145, 6, 192, 0, 0, 0, 198, 195, 145, 6, 192, 0, 0, 0, 198, 35, 146, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 146, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 146, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 147, 6, 192, 0, 0, 0, 197, 130, 42, 1, 192, 0, 0, 0, 198, 99, 147, 6, 198, 194, 147, 6, 197, 194, 42, 1, 192, 0, 0, 0, 198, 3, 148, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 148, 6, 198, 195, 148, 6, 192, 0, 0, 0, 198, 34, 149, 6, 198, 99, 149, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 149, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 150, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 150, 6, 198, 227, 150, 6, 198, 67, 151, 6, 198, 163, 151, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 152, 6, 192, 0, 0, 0, 198, 67, 152, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 152, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 43, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 152, 6, 198, 35, 153, 6, 197, 66, 43, 1, 198, 131, 153, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 153, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 154, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 154, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 155, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 155, 6, 198, 194, 155, 6, 198, 2, 156, 6, 198, 66, 156, 6, 198, 131, 156, 6, 192, 0, 0, 0, 198, 227, 156, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 157, 6, 192, 0, 0, 0, 198, 130, 157, 6, 192, 0, 0, 0, 198, 195, 157, 6, 198, 34, 158, 6, 198, 98, 158, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 158, 6, 198, 3, 159, 6, 198, 98, 159, 6, 198, 163, 159, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 160, 6, 197, 130, 43, 1, 197, 194, 43, 1, 198, 99, 160, 6, 198, 195, 160, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 161, 6, 198, 131, 161, 6, 198, 227, 161, 6, 198, 66, 162, 6, 198, 131, 162, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 162, 6, 198, 67, 163, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 163, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 163, 6, 198, 35, 164, 6, 192, 0, 0, 0, 198, 131, 164, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 164, 6, 197, 2, 44, 1, 198, 67, 165, 6, 192, 0, 0, 0, 198, 163, 165, 6, 192, 0, 0, 0, 197, 67, 44, 1, 198, 3, 166, 6, 198, 99, 166, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 166, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 167, 6, 198, 130, 167, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 167, 6, 192, 0, 0, 0, 198, 35, 168, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 168, 6, 198, 227, 168, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 44, 1, 197, 227, 44, 1, 198, 67, 169, 6, 192, 0, 0, 0, 198, 162, 169, 6, 198, 227, 169, 6, 198, 67, 170, 6, 198, 162, 170, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 170, 6, 198, 66, 171, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 45, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 171, 6, 198, 227, 171, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 172, 6, 192, 0, 0, 0, 198, 163, 172, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 173, 6, 198, 67, 173, 6, 198, 163, 173, 6, 198, 3, 174, 6, 198, 99, 174, 6, 198, 195, 174, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 175, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 175, 6, 198, 227, 175, 6, 192, 0, 0, 0, 198, 66, 176, 6, 198, 130, 176, 6, 197, 131, 45, 1, 198, 195, 176, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 177, 6, 198, 131, 177, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 177, 6, 192, 0, 0, 0, 198, 66, 178, 6, 198, 131, 178, 6, 198, 226, 178, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 179, 6, 192, 0, 0, 0, 198, 98, 179, 6, 198, 163, 179, 6, 198, 2, 180, 6, 192, 0, 0, 0, 198, 66, 180, 6, 192, 0, 0, 0, 198, 131, 180, 6, 198, 227, 180, 6, 198, 67, 181, 6, 198, 163, 181, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 182, 6, 192, 0, 0, 0, 198, 99, 182, 6, 192, 0, 0, 0, 197, 226, 45, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 182, 6, 198, 3, 183, 6, 198, 99, 183, 6, 198, 194, 183, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 46, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 46, 1, 198, 3, 184, 6, 198, 99, 184, 6, 198, 194, 184, 6, 192, 0, 0, 0, 197, 162, 46, 1, 198, 3, 185, 6, 192, 0, 0, 0, 198, 99, 185, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 185, 6, 198, 2, 186, 6, 198, 66, 186, 6, 192, 0, 0, 0, 197, 226, 46, 1, 198, 131, 186, 6, 198, 226, 186, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 187, 6, 198, 99, 187, 6, 198, 195, 187, 6, 198, 34, 188, 6, 192, 0, 0, 0, 197, 34, 47, 1, 198, 99, 188, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 188, 6, 198, 3, 189, 6, 198, 99, 189, 6, 198, 194, 189, 6, 198, 3, 190, 6, 198, 99, 190, 6, 198, 195, 190, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 47, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 191, 6, 198, 99, 191, 6, 192, 0, 0, 0, 198, 195, 191, 6, 198, 34, 192, 6, 198, 98, 192, 6, 192, 0, 0, 0, 198, 163, 192, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 47, 1, 192, 0, 0, 0, 197, 2, 48, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 193, 6, 198, 98, 193, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 193, 6, 192, 0, 0, 0, 198, 227, 193, 6, 197, 66, 48, 1, 198, 67, 194, 6, 192, 0, 0, 0, 198, 163, 194, 6, 197, 130, 48, 1, 198, 3, 195, 6, 197, 194, 48, 1, 197, 3, 49, 1, 198, 98, 195, 6, 198, 162, 195, 6, 198, 227, 195, 6, 197, 98, 49, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 196, 6, 198, 130, 196, 6, 192, 0, 0, 0, 198, 194, 196, 6, 198, 3, 197, 6, 198, 99, 197, 6, 197, 162, 49, 1, 192, 0, 0, 0, 198, 195, 197, 6, 198, 34, 198, 6, 192, 0, 0, 0, 198, 99, 198, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 198, 6, 198, 3, 199, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 199, 6, 198, 194, 199, 6, 198, 2, 200, 6, 198, 67, 200, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 200, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 200, 6, 198, 66, 201, 6, 198, 131, 201, 6, 198, 227, 201, 6, 192, 0, 0, 0, 197, 226, 49, 1, 192, 0, 0, 0, 198, 66, 202, 6, 198, 130, 202, 6, 192, 0, 0, 0, 197, 34, 50, 1, 198, 194, 202, 6, 198, 2, 203, 6, 197, 98, 50, 1, 198, 66, 203, 6, 197, 163, 50, 1, 198, 131, 203, 6, 192, 0, 0, 0, 198, 226, 203, 6, 198, 35, 204, 6, 192, 0, 0, 0, 198, 131, 204, 6, 192, 0, 0, 0, 197, 2, 51, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 204, 6, 198, 34, 205, 6, 198, 99, 205, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 205, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 206, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 206, 6, 192, 0, 0, 0, 198, 227, 206, 6, 198, 66, 207, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 207, 6, 192, 0, 0, 0, 198, 227, 207, 6, 192, 0, 0, 0, 198, 66, 208, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 208, 6, 198, 226, 208, 6, 198, 34, 209, 6, 192, 0, 0, 0, 198, 98, 209, 6, 192, 0, 0, 0, 198, 162, 209, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 51, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 209, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 210, 6, 192, 0, 0, 0, 197, 131, 51, 1, 198, 131, 210, 6, 198, 227, 210, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 211, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 211, 6, 197, 227, 51, 1, 198, 195, 211, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 212, 6, 198, 131, 212, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 212, 6, 192, 0, 0, 0, 197, 66, 52, 1, 198, 34, 213, 6, 192, 0, 0, 0, 198, 99, 213, 6, 192, 0, 0, 0, 198, 195, 213, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 52, 1, 198, 34, 214, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 214, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 214, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 52, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 215, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 215, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 215, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 216, 6, 198, 99, 216, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 216, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 217, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 217, 6, 197, 34, 53, 1, 198, 227, 217, 6, 192, 0, 0, 0, 198, 67, 218, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 218, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 219, 6, 197, 98, 53, 1, 198, 99, 219, 6, 197, 162, 53, 1, 198, 195, 219, 6, 198, 34, 220, 6, 198, 98, 220, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 220, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 53, 1, 197, 66, 54, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 220, 6, 198, 35, 221, 6, 198, 130, 221, 6, 198, 194, 221, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 222, 6, 198, 99, 222, 6, 192, 0, 0, 0, 198, 194, 222, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 223, 6, 192, 0, 0, 0, 198, 98, 223, 6, 197, 130, 54, 1, 198, 162, 223, 6, 197, 195, 54, 1, 198, 227, 223, 6, 198, 66, 224, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 224, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 224, 6, 192, 0, 0, 0, 197, 34, 55, 1, 192, 0, 0, 0, 198, 66, 225, 6, 192, 0, 0, 0, 197, 98, 55, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 225, 6, 192, 0, 0, 0, 198, 195, 225, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 55, 1, 192, 0, 0, 0, 198, 35, 226, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 55, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 226, 6, 197, 34, 56, 1, 198, 195, 226, 6, 192, 0, 0, 0, 198, 34, 227, 6, 197, 98, 56, 1, 192, 0, 0, 0, 198, 99, 227, 6, 198, 195, 227, 6, 192, 0, 0, 0, 198, 34, 228, 6, 198, 99, 228, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 228, 6, 197, 163, 56, 1, 192, 0, 0, 0, 198, 34, 229, 6, 192, 0, 0, 0, 198, 99, 229, 6, 192, 0, 0, 0, 198, 194, 229, 6, 192, 0, 0, 0, 197, 2, 57, 1, 192, 0, 0, 0, 198, 3, 230, 6, 198, 98, 230, 6, 198, 163, 230, 6, 197, 66, 57, 1, 198, 3, 231, 6, 198, 98, 231, 6, 198, 163, 231, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 232, 6, 198, 67, 232, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 232, 6, 197, 131, 57, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 232, 6, 198, 35, 233, 6, 192, 0, 0, 0, 198, 130, 233, 6, 198, 195, 233, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 234, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 234, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 234, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 235, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 235, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 235, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 236, 6, 198, 99, 236, 6, 198, 195, 236, 6, 192, 0, 0, 0, 198, 34, 237, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 237, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 57, 1, 192, 0, 0, 0, 198, 195, 237, 6, 198, 34, 238, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 238, 6, 198, 195, 238, 6, 192, 0, 0, 0, 198, 34, 239, 6, 198, 99, 239, 6, 192, 0, 0, 0, 198, 195, 239, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 240, 6, 198, 131, 240, 6, 198, 227, 240, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 241, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 241, 6, 192, 0, 0, 0, 198, 226, 241, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 242, 6, 198, 130, 242, 6, 198, 195, 242, 6, 192, 0, 0, 0, 198, 34, 243, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 243, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 243, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 244, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 244, 6, 192, 0, 0, 0, 198, 226, 244, 6, 198, 35, 245, 6, 198, 131, 245, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 245, 6, 198, 66, 246, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 246, 6, 192, 0, 0, 0, 198, 226, 246, 6, 192, 0, 0, 0, 198, 35, 247, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 247, 6, 192, 0, 0, 0, 198, 227, 247, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 248, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 248, 6, 192, 0, 0, 0, 198, 3, 249, 6, 192, 0, 0, 0, 198, 99, 249, 6, 198, 194, 249, 6, 192, 0, 0, 0, 198, 3, 250, 6, 198, 98, 250, 6, 198, 162, 250, 6, 198, 227, 250, 6, 197, 67, 58, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 58, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 251, 6, 198, 130, 251, 6, 192, 0, 0, 0, 198, 194, 251, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 252, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 252, 6, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 58, 1, 192, 0, 0, 0, 198, 195, 252, 6, 198, 35, 253, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 253, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 59, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 253, 6, 198, 35, 254, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 254, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 254, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 255, 6, 198, 99, 255, 6, 198, 195, 255, 6, 192, 0, 0, 0, 198, 35, 0, 7, 198, 130, 0, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 0, 7, 192, 0, 0, 0, 198, 35, 1, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 1, 7, 197, 162, 59, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 1, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 59, 1, 197, 34, 60, 1, 198, 35, 2, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 2, 7, 198, 227, 2, 7, 198, 67, 3, 7, 198, 162, 3, 7, 198, 227, 3, 7, 198, 67, 4, 7, 198, 162, 4, 7, 192, 0, 0, 0, 198, 227, 4, 7, 198, 67, 5, 7, 198, 162, 5, 7, 192, 0, 0, 0, 198, 227, 5, 7, 198, 67, 6, 7, 198, 163, 6, 7, 192, 0, 0, 0, 198, 2, 7, 7, 198, 67, 7, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 7, 7, 198, 2, 8, 7, 198, 67, 8, 7, 198, 162, 8, 7, 192, 0, 0, 0, 198, 226, 8, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 9, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 60, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 9, 7, 197, 162, 60, 1, 198, 226, 9, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 10, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 10, 7, 198, 194, 10, 7, 192, 0, 0, 0, 197, 226, 60, 1, 192, 0, 0, 0, 198, 3, 11, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 11, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 61, 1, 192, 0, 0, 0, 198, 195, 11, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 12, 7, 192, 0, 0, 0, 198, 131, 12, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 12, 7, 192, 0, 0, 0, 197, 130, 61, 1, 198, 67, 13, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 61, 1, 198, 163, 13, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 14, 7, 192, 0, 0, 0, 198, 98, 14, 7, 192, 0, 0, 0, 198, 163, 14, 7, 197, 3, 62, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 15, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 62, 1, 198, 67, 15, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 15, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 16, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 16, 7, 198, 195, 16, 7, 192, 0, 0, 0, 198, 34, 17, 7, 198, 99, 17, 7, 198, 195, 17, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 18, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 18, 7, 198, 195, 18, 7, 198, 35, 19, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 19, 7, 198, 227, 19, 7, 198, 67, 20, 7, 198, 163, 20, 7, 192, 0, 0, 0, 198, 3, 21, 7, 197, 162, 62, 1, 192, 0, 0, 0, 198, 98, 21, 7, 192, 0, 0, 0, 198, 163, 21, 7, 192, 0, 0, 0, 198, 3, 22, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 22, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 22, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 23, 7, 197, 227, 62, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 23, 7, 198, 194, 23, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 24, 7, 198, 67, 24, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 24, 7, 198, 3, 25, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 25, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 25, 7, 198, 3, 26, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 26, 7, 198, 195, 26, 7, 192, 0, 0, 0, 198, 35, 27, 7, 192, 0, 0, 0, 198, 131, 27, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 63, 1, 192, 0, 0, 0, 198, 226, 27, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 63, 1, 192, 0, 0, 0, 198, 35, 28, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 28, 7, 198, 227, 28, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 29, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 29, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 30, 7, 192, 0, 0, 0, 198, 99, 30, 7, 192, 0, 0, 0, 197, 2, 64, 1, 197, 66, 64, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 30, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 31, 7, 198, 99, 31, 7, 198, 195, 31, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 32, 7, 192, 0, 0, 0, 198, 131, 32, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 32, 7, 198, 35, 33, 7, 198, 131, 33, 7, 198, 227, 33, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 34, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 34, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 35, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 35, 7, 197, 130, 64, 1, 198, 195, 35, 7, 198, 34, 36, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 36, 7, 198, 195, 36, 7, 192, 0, 0, 0, 198, 35, 37, 7, 198, 131, 37, 7, 198, 227, 37, 7, 198, 67, 38, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 38, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 39, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 39, 7, 197, 195, 64, 1, 198, 163, 39, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 40, 7, 198, 67, 40, 7, 192, 0, 0, 0, 198, 163, 40, 7, 198, 3, 41, 7, 197, 34, 65, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 41, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 41, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 42, 7, 192, 0, 0, 0, 198, 131, 42, 7, 198, 227, 42, 7, 198, 67, 43, 7, 198, 163, 43, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 44, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 44, 7, 198, 194, 44, 7, 192, 0, 0, 0, 197, 99, 65, 1, 198, 3, 45, 7, 198, 99, 45, 7, 198, 194, 45, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 46, 7, 197, 195, 65, 1, 198, 67, 46, 7, 197, 35, 66, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 46, 7, 192, 0, 0, 0, 198, 227, 46, 7, 192, 0, 0, 0, 197, 131, 66, 1, 192, 0, 0, 0, 197, 226, 66, 1, 198, 67, 47, 7, 198, 163, 47, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 48, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 48, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 48, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 67, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 49, 7, 192, 0, 0, 0, 198, 99, 49, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 49, 7, 198, 34, 50, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 50, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 50, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 51, 7, 198, 66, 51, 7, 192, 0, 0, 0, 197, 98, 67, 1, 192, 0, 0, 0, 198, 130, 51, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 51, 7, 198, 3, 52, 7, 198, 99, 52, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 52, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 53, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 67, 1, 197, 227, 67, 1, 198, 131, 53, 7, 198, 227, 53, 7, 198, 67, 54, 7, 192, 0, 0, 0, 198, 162, 54, 7, 198, 227, 54, 7, 198, 67, 55, 7, 197, 66, 68, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 55, 7, 192, 0, 0, 0, 198, 226, 55, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 56, 7, 192, 0, 0, 0, 198, 130, 56, 7, 198, 195, 56, 7, 198, 34, 57, 7, 198, 99, 57, 7, 198, 194, 57, 7, 197, 131, 68, 1, 198, 3, 58, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 58, 7, 198, 163, 58, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 68, 1, 198, 2, 59, 7, 198, 67, 59, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 59, 7, 198, 226, 59, 7, 192, 0, 0, 0, 197, 35, 69, 1, 198, 34, 60, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 60, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 60, 7, 198, 3, 61, 7, 192, 0, 0, 0, 198, 98, 61, 7, 198, 162, 61, 7, 198, 227, 61, 7, 192, 0, 0, 0, 197, 130, 69, 1, 198, 67, 62, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 62, 7, 197, 194, 69, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 63, 7, 198, 66, 63, 7, 198, 130, 63, 7, 197, 3, 70, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 63, 7, 198, 35, 64, 7, 192, 0, 0, 0, 197, 98, 70, 1, 198, 131, 64, 7, 192, 0, 0, 0, 198, 226, 64, 7, 198, 34, 65, 7, 198, 98, 65, 7, 198, 162, 65, 7, 192, 0, 0, 0, 198, 227, 65, 7, 198, 67, 66, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 66, 7, 198, 3, 67, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 67, 7, 198, 194, 67, 7, 198, 3, 68, 7, 198, 98, 68, 7, 192, 0, 0, 0, 197, 162, 70, 1, 192, 0, 0, 0, 198, 163, 68, 7, 198, 2, 69, 7, 198, 67, 69, 7, 197, 226, 70, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 69, 7, 198, 226, 69, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 70, 7, 198, 131, 70, 7, 197, 34, 71, 1, 198, 227, 70, 7, 198, 67, 71, 7, 198, 162, 71, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 71, 7, 198, 35, 72, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 72, 7, 198, 194, 72, 7, 198, 3, 73, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 73, 7, 198, 195, 73, 7, 192, 0, 0, 0, 198, 35, 74, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 74, 7, 198, 226, 74, 7, 192, 0, 0, 0, 198, 35, 75, 7, 198, 131, 75, 7, 198, 227, 75, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 76, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 76, 7, 198, 195, 76, 7, 198, 35, 77, 7, 198, 131, 77, 7, 197, 98, 71, 1, 198, 226, 77, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 78, 7, 198, 99, 78, 7, 198, 195, 78, 7, 198, 34, 79, 7, 192, 0, 0, 0, 198, 99, 79, 7, 192, 0, 0, 0, 197, 162, 71, 1, 198, 194, 79, 7, 192, 0, 0, 0, 198, 3, 80, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 80, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 80, 7, 192, 0, 0, 0, 197, 226, 71, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 81, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 81, 7, 198, 226, 81, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 82, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 82, 7, 192, 0, 0, 0, 198, 226, 82, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 72, 1, 198, 35, 83, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 72, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 83, 7, 192, 0, 0, 0, 198, 226, 83, 7, 192, 0, 0, 0, 197, 162, 72, 1, 192, 0, 0, 0, 198, 35, 84, 7, 198, 130, 84, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 84, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 85, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 85, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 85, 7, 198, 35, 86, 7, 198, 131, 86, 7, 198, 227, 86, 7, 198, 66, 87, 7, 192, 0, 0, 0, 198, 130, 87, 7, 198, 194, 87, 7, 198, 3, 88, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 88, 7, 192, 0, 0, 0, 198, 195, 88, 7, 197, 227, 72, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 89, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 89, 7, 197, 66, 73, 1, 198, 226, 89, 7, 192, 0, 0, 0, 198, 35, 90, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 90, 7, 192, 0, 0, 0, 198, 227, 90, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 91, 7, 197, 130, 73, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 91, 7, 198, 195, 91, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 73, 1, 192, 0, 0, 0, 198, 35, 92, 7, 192, 0, 0, 0, 198, 131, 92, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 92, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 93, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 93, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 93, 7, 192, 0, 0, 0, 198, 3, 94, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 94, 7, 198, 195, 94, 7, 198, 35, 95, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 95, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 95, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 96, 7, 198, 99, 96, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 96, 7, 192, 0, 0, 0, 198, 35, 97, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 97, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 74, 1, 192, 0, 0, 0, 198, 227, 97, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 98, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 98, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 99, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 99, 7, 198, 130, 99, 7, 198, 195, 99, 7, 192, 0, 0, 0, 198, 35, 100, 7, 198, 130, 100, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 100, 7, 198, 2, 101, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 101, 7, 198, 162, 101, 7, 198, 227, 101, 7, 192, 0, 0, 0, 198, 67, 102, 7, 198, 162, 102, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 102, 7, 198, 67, 103, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 103, 7, 197, 66, 74, 1, 197, 131, 74, 1, 198, 3, 104, 7, 198, 99, 104, 7, 192, 0, 0, 0, 198, 195, 104, 7, 198, 35, 105, 7, 192, 0, 0, 0, 198, 130, 105, 7, 192, 0, 0, 0, 198, 195, 105, 7, 192, 0, 0, 0, 198, 35, 106, 7, 198, 130, 106, 7, 198, 194, 106, 7, 198, 3, 107, 7, 197, 226, 74, 1, 198, 99, 107, 7, 198, 195, 107, 7, 198, 35, 108, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 108, 7, 197, 34, 75, 1, 198, 227, 108, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 109, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 75, 1, 192, 0, 0, 0, 198, 163, 109, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 75, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 110, 7, 192, 0, 0, 0, 198, 66, 110, 7, 197, 2, 76, 1, 192, 0, 0, 0, 197, 67, 76, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 110, 7, 198, 227, 110, 7, 197, 162, 76, 1, 192, 0, 0, 0, 198, 67, 111, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 111, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 111, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 112, 7, 197, 226, 76, 1, 198, 99, 112, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 112, 7, 197, 34, 77, 1, 198, 3, 113, 7, 198, 99, 113, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 113, 7, 198, 35, 114, 7, 198, 131, 114, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 114, 7, 198, 67, 115, 7, 198, 163, 115, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 116, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 77, 1, 198, 66, 116, 7, 198, 130, 116, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 116, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 117, 7, 198, 131, 117, 7, 198, 226, 117, 7, 192, 0, 0, 0, 198, 35, 118, 7, 198, 131, 118, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 118, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 119, 7, 192, 0, 0, 0, 198, 163, 119, 7, 192, 0, 0, 0, 198, 2, 120, 7, 198, 67, 120, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 120, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 120, 7, 192, 0, 0, 0, 198, 66, 121, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 121, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 121, 7, 192, 0, 0, 0, 197, 162, 77, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 77, 1, 192, 0, 0, 0, 198, 34, 122, 7, 192, 0, 0, 0, 198, 99, 122, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 122, 7, 198, 3, 123, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 123, 7, 192, 0, 0, 0, 198, 195, 123, 7, 198, 34, 124, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 124, 7, 198, 195, 124, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 125, 7, 198, 130, 125, 7, 192, 0, 0, 0, 198, 195, 125, 7, 198, 35, 126, 7, 198, 131, 126, 7, 192, 0, 0, 0, 198, 227, 126, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 127, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 127, 7, 198, 226, 127, 7, 198, 34, 128, 7, 198, 99, 128, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 128, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 129, 7, 198, 99, 129, 7, 192, 0, 0, 0, 198, 194, 129, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 130, 7, 192, 0, 0, 0, 198, 98, 130, 7, 198, 162, 130, 7, 198, 226, 130, 7, 198, 35, 131, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 78, 1, 192, 0, 0, 0, 197, 98, 78, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 131, 7, 192, 0, 0, 0, 198, 227, 131, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 132, 7, 198, 162, 132, 7, 198, 227, 132, 7, 192, 0, 0, 0, 198, 67, 133, 7, 192, 0, 0, 0, 198, 162, 133, 7, 192, 0, 0, 0, 198, 227, 133, 7, 192, 0, 0, 0, 198, 67, 134, 7, 197, 162, 78, 1, 192, 0, 0, 0, 198, 163, 134, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 135, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 78, 1, 197, 34, 79, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 135, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 135, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 136, 7, 198, 99, 136, 7, 192, 0, 0, 0, 197, 98, 79, 1, 198, 195, 136, 7, 198, 35, 137, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 137, 7, 198, 194, 137, 7, 198, 2, 138, 7, 192, 0, 0, 0, 198, 67, 138, 7, 198, 162, 138, 7, 192, 0, 0, 0, 198, 227, 138, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 139, 7, 192, 0, 0, 0, 198, 163, 139, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 140, 7, 198, 66, 140, 7, 198, 131, 140, 7, 198, 227, 140, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 79, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 141, 7, 192, 0, 0, 0, 198, 163, 141, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 142, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 142, 7, 198, 162, 142, 7, 198, 227, 142, 7, 192, 0, 0, 0, 198, 67, 143, 7, 198, 162, 143, 7, 192, 0, 0, 0, 197, 2, 80, 1, 198, 227, 143, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 144, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 144, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 145, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 145, 7, 198, 131, 145, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 145, 7, 198, 34, 146, 7, 198, 98, 146, 7, 198, 162, 146, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 146, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 147, 7, 198, 98, 147, 7, 192, 0, 0, 0, 198, 162, 147, 7, 198, 227, 147, 7, 198, 66, 148, 7, 198, 130, 148, 7, 198, 194, 148, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 149, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 149, 7, 192, 0, 0, 0, 198, 195, 149, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 80, 1, 198, 35, 150, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 150, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 150, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 151, 7, 192, 0, 0, 0, 198, 163, 151, 7, 192, 0, 0, 0, 198, 3, 152, 7, 192, 0, 0, 0, 198, 99, 152, 7, 198, 195, 152, 7, 198, 34, 153, 7, 198, 99, 153, 7, 198, 194, 153, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 154, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 154, 7, 192, 0, 0, 0, 198, 162, 154, 7, 198, 226, 154, 7, 198, 34, 155, 7, 192, 0, 0, 0, 198, 98, 155, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 155, 7, 192, 0, 0, 0, 198, 227, 155, 7, 198, 67, 156, 7, 192, 0, 0, 0, 198, 162, 156, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 156, 7, 192, 0, 0, 0, 198, 67, 157, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 157, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 158, 7, 197, 130, 80, 1, 198, 98, 158, 7, 198, 163, 158, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 159, 7, 192, 0, 0, 0, 198, 67, 159, 7, 198, 162, 159, 7, 198, 227, 159, 7, 197, 194, 80, 1, 198, 67, 160, 7, 198, 162, 160, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 81, 1, 192, 0, 0, 0, 198, 227, 160, 7, 197, 66, 81, 1, 198, 67, 161, 7, 198, 163, 161, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 162, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 162, 7, 192, 0, 0, 0, 197, 131, 81, 1, 198, 162, 162, 7, 192, 0, 0, 0, 198, 227, 162, 7, 198, 67, 163, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 81, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 82, 1, 192, 0, 0, 0, 198, 163, 163, 7, 198, 2, 164, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 82, 1, 197, 194, 82, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 164, 7, 192, 0, 0, 0, 198, 162, 164, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 83, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 83, 1, 198, 226, 164, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 165, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 165, 7, 198, 226, 165, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 166, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 166, 7, 192, 0, 0, 0, 198, 195, 166, 7, 192, 0, 0, 0, 198, 34, 167, 7, 197, 130, 83, 1, 192, 0, 0, 0, 198, 99, 167, 7, 198, 194, 167, 7, 198, 2, 168, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 83, 1, 192, 0, 0, 0, 198, 67, 168, 7, 192, 0, 0, 0, 198, 163, 168, 7, 192, 0, 0, 0, 198, 3, 169, 7, 198, 98, 169, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 169, 7, 192, 0, 0, 0, 198, 226, 169, 7, 198, 35, 170, 7, 192, 0, 0, 0, 197, 2, 84, 1, 192, 0, 0, 0, 197, 66, 84, 1, 192, 0, 0, 0, 198, 131, 170, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 84, 1, 198, 227, 170, 7, 198, 66, 171, 7, 192, 0, 0, 0, 197, 194, 84, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 171, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 171, 7, 198, 67, 172, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 172, 7, 198, 3, 173, 7, 192, 0, 0, 0, 198, 98, 173, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 173, 7, 192, 0, 0, 0, 198, 227, 173, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 174, 7, 192, 0, 0, 0, 198, 163, 174, 7, 198, 3, 175, 7, 198, 98, 175, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 175, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 176, 7, 198, 99, 176, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 176, 7, 198, 3, 177, 7, 197, 3, 85, 1, 197, 98, 85, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 177, 7, 198, 163, 177, 7, 192, 0, 0, 0, 198, 3, 178, 7, 198, 99, 178, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 178, 7, 192, 0, 0, 0, 198, 34, 179, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 179, 7, 198, 163, 179, 7, 198, 3, 180, 7, 198, 99, 180, 7, 198, 194, 180, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 181, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 181, 7, 192, 0, 0, 0, 198, 131, 181, 7, 198, 227, 181, 7, 197, 162, 85, 1, 192, 0, 0, 0, 198, 66, 182, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 182, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 182, 7, 192, 0, 0, 0, 198, 34, 183, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 183, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 183, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 183, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 184, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 184, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 184, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 185, 7, 198, 130, 185, 7, 198, 194, 185, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 186, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 186, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 186, 7, 198, 2, 187, 7, 192, 0, 0, 0, 198, 66, 187, 7, 198, 130, 187, 7, 192, 0, 0, 0, 197, 226, 85, 1, 198, 195, 187, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 188, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 188, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 188, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 189, 7, 198, 66, 189, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 86, 1, 192, 0, 0, 0, 198, 131, 189, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 189, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 190, 7, 192, 0, 0, 0, 198, 99, 190, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 190, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 191, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 191, 7, 192, 0, 0, 0, 198, 195, 191, 7, 192, 0, 0, 0, 198, 34, 192, 7, 197, 98, 86, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 192, 7, 192, 0, 0, 0, 198, 195, 192, 7, 198, 34, 193, 7, 192, 0, 0, 0, 197, 162, 86, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 193, 7, 197, 227, 86, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 193, 7, 198, 227, 193, 7, 198, 67, 194, 7, 192, 0, 0, 0, 198, 163, 194, 7, 198, 2, 195, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 195, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 195, 7, 198, 2, 196, 7, 198, 67, 196, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 87, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 196, 7, 192, 0, 0, 0, 198, 3, 197, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 197, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 197, 7, 198, 35, 198, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 198, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 87, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 198, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 199, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 199, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 199, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 200, 7, 198, 99, 200, 7, 198, 194, 200, 7, 192, 0, 0, 0, 198, 2, 201, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 201, 7, 198, 163, 201, 7, 192, 0, 0, 0, 198, 3, 202, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 202, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 202, 7, 198, 34, 203, 7, 198, 99, 203, 7, 198, 194, 203, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 204, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 204, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 204, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 205, 7, 197, 194, 87, 1, 192, 0, 0, 0, 198, 99, 205, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 88, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 205, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 206, 7, 192, 0, 0, 0, 198, 99, 206, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 206, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 207, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 207, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 207, 7, 197, 66, 88, 1, 192, 0, 0, 0, 197, 130, 88, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 208, 7, 192, 0, 0, 0, 198, 131, 208, 7, 198, 227, 208, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 209, 7, 198, 130, 209, 7, 192, 0, 0, 0, 198, 195, 209, 7, 198, 34, 210, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 210, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 210, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 211, 7, 198, 130, 211, 7, 192, 0, 0, 0, 198, 195, 211, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 212, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 212, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 212, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 213, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 213, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 213, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 214, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 214, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 214, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 215, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 215, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 216, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 88, 1, 198, 98, 216, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 216, 7, 198, 2, 217, 7, 192, 0, 0, 0, 198, 67, 217, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 217, 7, 192, 0, 0, 0, 198, 227, 217, 7, 198, 66, 218, 7, 192, 0, 0, 0, 198, 130, 218, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 218, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 219, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 219, 7, 198, 130, 219, 7, 192, 0, 0, 0, 198, 194, 219, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 220, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 220, 7, 198, 163, 220, 7, 198, 3, 221, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 221, 7, 198, 194, 221, 7, 197, 34, 89, 1, 198, 3, 222, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 89, 1, 198, 99, 222, 7, 198, 194, 222, 7, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 89, 1, 192, 0, 0, 0, 198, 3, 223, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 223, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 223, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 224, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 224, 7, 198, 195, 224, 7, 198, 34, 225, 7, 192, 0, 0, 0, 198, 99, 225, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 225, 7, 192, 0, 0, 0, 198, 3, 226, 7, 192, 0, 0, 0, 197, 35, 90, 1, 192, 0, 0, 0, 198, 99, 226, 7, 192, 0, 0, 0, 198, 195, 226, 7, 192, 0, 0, 0, 198, 35, 227, 7, 198, 130, 227, 7, 192, 0, 0, 0, 197, 130, 90, 1, 192, 0, 0, 0, 198, 194, 227, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 228, 7, 198, 66, 228, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 228, 7, 192, 0, 0, 0, 198, 226, 228, 7, 198, 34, 229, 7, 198, 99, 229, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 90, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 229, 7, 198, 3, 230, 7, 198, 99, 230, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 230, 7, 198, 3, 231, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 231, 7, 198, 163, 231, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 232, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 232, 7, 198, 195, 232, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 233, 7, 192, 0, 0, 0, 198, 98, 233, 7, 197, 2, 91, 1, 198, 162, 233, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 91, 1, 192, 0, 0, 0, 198, 227, 233, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 91, 1, 198, 66, 234, 7, 197, 2, 92, 1, 198, 131, 234, 7, 198, 226, 234, 7, 197, 67, 92, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 235, 7, 192, 0, 0, 0, 198, 130, 235, 7, 192, 0, 0, 0, 198, 195, 235, 7, 192, 0, 0, 0, 198, 35, 236, 7, 198, 130, 236, 7, 198, 195, 236, 7, 198, 34, 237, 7, 198, 99, 237, 7, 192, 0, 0, 0, 197, 163, 92, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 237, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 238, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 238, 7, 198, 227, 238, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 239, 7, 192, 0, 0, 0, 198, 162, 239, 7, 198, 227, 239, 7, 198, 67, 240, 7, 198, 163, 240, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 241, 7, 198, 99, 241, 7, 198, 194, 241, 7, 198, 3, 242, 7, 192, 0, 0, 0, 198, 99, 242, 7, 198, 195, 242, 7, 192, 0, 0, 0, 198, 34, 243, 7, 192, 0, 0, 0, 198, 98, 243, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 243, 7, 198, 227, 243, 7, 198, 66, 244, 7, 198, 131, 244, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 244, 7, 192, 0, 0, 0, 198, 66, 245, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 245, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 245, 7, 192, 0, 0, 0, 198, 2, 246, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 246, 7, 192, 0, 0, 0, 198, 131, 246, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 246, 7, 198, 66, 247, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 247, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 247, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 248, 7, 198, 131, 248, 7, 192, 0, 0, 0, 198, 226, 248, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 249, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 93, 1, 198, 130, 249, 7, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 249, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 93, 1, 192, 0, 0, 0, 198, 3, 250, 7, 197, 131, 93, 1, 192, 0, 0, 0, 198, 99, 250, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 93, 1, 198, 194, 250, 7, 198, 3, 251, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 251, 7, 198, 195, 251, 7, 192, 0, 0, 0, 198, 34, 252, 7, 198, 99, 252, 7, 198, 195, 252, 7, 198, 34, 253, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 253, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 253, 7, 198, 3, 254, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 254, 7, 198, 195, 254, 7, 192, 0, 0, 0, 198, 35, 255, 7, 192, 0, 0, 0, 198, 131, 255, 7, 192, 0, 0, 0, 198, 227, 255, 7, 198, 67, 0, 8, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 0, 8, 198, 226, 0, 8, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 1, 8, 192, 0, 0, 0, 197, 34, 94, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 1, 8, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 101, 6, 198, 131, 181, 2, 198, 226, 152, 6, 198, 162, 114, 6, 197, 195, 168, 0, 197, 99, 0, 0, 198, 226, 180, 0, 198, 162, 0, 8, 198, 162, 0, 8, 198, 130, 49, 1, 198, 163, 231, 6, 198, 98, 216, 0, 198, 98, 48, 1, 198, 226, 31, 2, 198, 194, 79, 4, 198, 66, 61, 5, 198, 98, 233, 5, 198, 2, 5, 6, 198, 162, 31, 6, 198, 226, 204, 6, 198, 3, 9, 3, 198, 99, 87, 3, 198, 35, 192, 3, 198, 99, 4, 4, 198, 227, 186, 5, 198, 99, 222, 6, 198, 131, 150, 7, 198, 3, 13, 0, 198, 131, 169, 0, 198, 195, 31, 3, 198, 67, 219, 3, 198, 99, 232, 5, 198, 99, 225, 7, 198, 227, 153, 1, 198, 227, 167, 3, 198, 99, 223, 5, 198, 131, 40, 6, 198, 35, 58, 2, 198, 3, 125, 5, 198, 35, 12, 6, 198, 163, 190, 1, 198, 99, 186, 2, 198, 3, 95, 3, 198, 67, 239, 3, 192, 0, 0, 0, 198, 2, 47, 0, 198, 99, 116, 0, 198, 66, 151, 0, 198, 226, 107, 2, 198, 34, 27, 3, 198, 66, 218, 3, 198, 66, 94, 4, 198, 130, 75, 5, 198, 162, 228, 5, 198, 66, 234, 5, 198, 162, 135, 6, 198, 226, 82, 7, 198, 162, 183, 7, 198, 194, 222, 7, 198, 227, 131, 4, 198, 131, 155, 4, 198, 227, 15, 5, 198, 131, 178, 5, 198, 195, 0, 7, 198, 35, 227, 7, 198, 227, 77, 6, 198, 99, 39, 1, 198, 227, 198, 1, 198, 227, 232, 4, 198, 227, 87, 5, 198, 162, 228, 3, 198, 194, 137, 4, 198, 98, 113, 6, 198, 226, 74, 7, 198, 162, 37, 1, 198, 2, 132, 1, 198, 66, 11, 3, 198, 34, 106, 3, 198, 162, 140, 3, 198, 2, 2, 5, 198, 34, 37, 5, 198, 194, 51, 7, 198, 35, 149, 0, 198, 35, 90, 5, 198, 67, 120, 0, 198, 195, 118, 0, 198, 227, 171, 4, 198, 35, 22, 5, 198, 35, 182, 5, 198, 99, 57, 7, 198, 195, 95, 6, 198, 130, 59, 2, 198, 3, 9, 3, 198, 163, 82, 6, 198, 3, 7, 0, 198, 131, 115, 1, 198, 34, 230, 1, 198, 99, 249, 3, 198, 2, 55, 4, 198, 195, 157, 0, 198, 131, 143, 4, 198, 227, 51, 0, 198, 3, 219, 1, 198, 162, 2, 0, 197, 3, 157, 0, 198, 2, 129, 2, 198, 195, 0, 5, 198, 131, 176, 0, 198, 226, 24, 1, 198, 3, 98, 4, 198, 131, 187, 5, 198, 227, 73, 6, 198, 195, 48, 3, 198, 3, 174, 6, 197, 99, 154, 0, 198, 67, 66, 2, 198, 35, 157, 5, 198, 227, 82, 2, 198, 131, 53, 4, 198, 131, 20, 0, 198, 3, 107, 0, 198, 195, 117, 0, 198, 3, 226, 2, 198, 3, 248, 4, 198, 67, 141, 5, 198, 131, 77, 6, 198, 3, 231, 6, 197, 194, 25, 0, 197, 194, 35, 0, 197, 34, 63, 0, 197, 226, 87, 0, 197, 130, 124, 0, 197, 2, 175, 0, 197, 162, 220, 0, 197, 34, 65, 1, 197, 194, 82, 1, 197, 194, 90, 1, 197, 2, 92, 1, 198, 99, 143, 0, 198, 35, 176, 2, 198, 227, 40, 3, 198, 227, 169, 6, 198, 67, 181, 1, 198, 131, 20, 2, 198, 131, 33, 2, 198, 195, 102, 2, 198, 99, 143, 3, 198, 195, 198, 3, 198, 131, 28, 4, 198, 227, 163, 4, 198, 195, 28, 5, 198, 195, 84, 5, 198, 35, 161, 6, 198, 3, 203, 5, 198, 195, 190, 6, 198, 67, 8, 7, 198, 99, 127, 0, 198, 99, 144, 0, 198, 163, 201, 0, 198, 227, 190, 3, 198, 163, 26, 6, 198, 227, 73, 6, 198, 67, 190, 1, 198, 195, 227, 1, 198, 67, 78, 2, 198, 163, 46, 3, 198, 99, 229, 4, 198, 195, 246, 3, 198, 3, 28, 0, 198, 195, 243, 0, 198, 131, 115, 1, 198, 163, 160, 1, 198, 195, 231, 1, 198, 67, 255, 3, 198, 3, 24, 4, 198, 131, 64, 5, 198, 99, 82, 5, 198, 99, 238, 6, 198, 131, 74, 7, 198, 35, 84, 7, 198, 227, 108, 7, 198, 226, 47, 0, 198, 130, 159, 4, 198, 34, 228, 6, 198, 98, 65, 7, 198, 195, 255, 1, 198, 98, 13, 0, 198, 66, 86, 0, 198, 34, 117, 1, 198, 66, 127, 1, 198, 226, 131, 2, 198, 3, 9, 3, 198, 130, 211, 3, 198, 194, 74, 4, 198, 130, 204, 5, 198, 66, 202, 6, 198, 99, 254, 7, 197, 35, 128, 0, 198, 195, 49, 7, 198, 194, 141, 0, 198, 194, 198, 2, 198, 66, 211, 2, 198, 130, 91, 3, 198, 162, 128, 3, 198, 162, 11, 4, 198, 130, 51, 4, 198, 226, 129, 4, 198, 2, 253, 4, 198, 2, 116, 7, 198, 35, 108, 0, 198, 35, 40, 2, 198, 3, 58, 7, 198, 195, 69, 0, 198, 99, 149, 1, 198, 99, 109, 3, 198, 131, 162, 6, 198, 3, 215, 1, 198, 67, 9, 2, 198, 67, 214, 2, 198, 99, 249, 3, 198, 67, 59, 7, 198, 130, 129, 0, 198, 98, 187, 0, 198, 162, 132, 1, 198, 131, 150, 2, 198, 98, 193, 2, 198, 34, 230, 2, 198, 130, 82, 3, 198, 34, 11, 4, 198, 98, 66, 4, 198, 2, 60, 5, 198, 2, 28, 6, 198, 2, 30, 6, 198, 98, 230, 6, 198, 226, 71, 7, 198, 195, 160, 0, 198, 35, 133, 3, 198, 163, 188, 0, 198, 195, 211, 3, 198, 163, 29, 4, 198, 163, 226, 5, 198, 163, 62, 7, 198, 131, 198, 7, 198, 99, 230, 7, 198, 35, 202, 2, 198, 131, 104, 3, 198, 131, 127, 5, 198, 3, 192, 4, 198, 67, 201, 4, 198, 3, 237, 4, 198, 99, 234, 3, 198, 98, 186, 3, 198, 35, 92, 6, 198, 131, 22, 0, 198, 2, 167, 5, 198, 98, 132, 0, 198, 35, 125, 0, 198, 194, 186, 1, 198, 163, 60, 2, 198, 163, 244, 4, 198, 195, 100, 1, 198, 195, 87, 3, 197, 67, 129, 0, 198, 131, 164, 6, 198, 163, 15, 6, 198, 99, 52, 7, 198, 227, 43, 6, 198, 227, 191, 1, 198, 67, 100, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 24, 1, 192, 0, 0, 0, 198, 67, 165, 2, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 19, 0, 197, 66, 193, 0, 198, 195, 89, 4, 192, 0, 0, 0, 198, 35, 153, 4, 198, 67, 154, 4, 198, 67, 157, 4, 198, 99, 85, 7, 198, 99, 243, 4, 198, 162, 67, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 30, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 192, 6, 198, 98, 214, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 126, 7, 198, 226, 127, 7, 198, 227, 132, 7, 197, 195, 88, 1, 198, 195, 211, 6, 198, 34, 65, 7, 198, 226, 50, 0, 198, 99, 87, 0, 198, 131, 103, 0, 198, 3, 148, 0, 198, 195, 152, 0, 198, 66, 164, 0, 198, 99, 218, 0, 198, 99, 225, 0, 198, 130, 232, 0, 198, 227, 20, 1, 198, 67, 32, 1, 198, 67, 132, 1, 192, 0, 0, 0, 197, 226, 99, 0, 198, 194, 14, 2, 198, 35, 20, 2, 198, 131, 31, 2, 198, 3, 124, 2, 192, 0, 0, 0, 198, 130, 168, 2, 198, 194, 226, 2, 197, 98, 161, 0, 197, 98, 165, 0, 198, 3, 143, 3, 198, 162, 203, 3, 192, 0, 0, 0, 198, 3, 13, 4, 198, 226, 132, 4, 198, 34, 148, 4, 198, 2, 150, 4, 198, 194, 149, 4, 198, 162, 150, 4, 198, 194, 151, 4, 198, 195, 152, 4, 197, 98, 222, 0, 198, 227, 156, 4, 198, 3, 176, 4, 198, 99, 182, 4, 198, 227, 218, 4, 198, 195, 28, 5, 198, 227, 29, 5, 198, 99, 38, 5, 198, 34, 59, 5, 198, 66, 76, 5, 197, 162, 249, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 4, 1, 198, 131, 33, 6, 198, 226, 44, 6, 198, 67, 83, 6, 198, 3, 90, 6, 198, 163, 115, 6, 198, 163, 121, 6, 192, 0, 0, 0, 198, 163, 192, 6, 198, 35, 72, 7, 198, 163, 103, 7, 198, 67, 111, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 5, 0, 192, 0, 0, 0, 198, 163, 106, 0, 192, 0, 0, 0, 198, 195, 101, 0, 198, 130, 110, 0, 198, 163, 147, 0, 198, 195, 153, 0, 198, 99, 218, 0, 192, 0, 0, 0, 197, 2, 44, 0, 198, 195, 223, 0, 198, 131, 24, 1, 198, 35, 34, 1, 198, 195, 47, 1, 198, 3, 50, 1, 198, 98, 73, 1, 198, 34, 90, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 209, 1, 198, 34, 220, 1, 198, 227, 253, 1, 192, 0, 0, 0, 197, 34, 104, 0, 198, 35, 20, 2, 198, 98, 13, 2, 198, 131, 31, 2, 198, 66, 41, 2, 197, 130, 116, 0, 198, 162, 93, 2, 192, 0, 0, 0, 198, 66, 125, 2, 198, 67, 165, 2, 198, 99, 186, 2, 198, 195, 186, 2, 198, 163, 194, 2, 192, 0, 0, 0, 198, 195, 48, 3, 198, 130, 91, 3, 192, 0, 0, 0, 198, 98, 135, 3, 198, 3, 143, 3, 198, 35, 176, 3, 198, 162, 203, 3, 192, 0, 0, 0, 198, 195, 221, 3, 198, 67, 233, 3, 197, 66, 193, 0, 198, 163, 25, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 71, 4, 198, 195, 89, 4, 198, 99, 91, 4, 198, 35, 96, 4, 192, 0, 0, 0, 198, 67, 105, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 218, 4, 192, 0, 0, 0, 198, 226, 7, 5, 198, 195, 28, 5, 198, 131, 52, 5, 198, 66, 76, 5, 197, 67, 255, 0, 197, 2, 2, 1, 192, 0, 0, 0, 198, 99, 37, 6, 198, 67, 43, 6, 198, 226, 44, 6, 198, 34, 75, 6, 197, 162, 30, 1, 198, 195, 76, 6, 198, 67, 83, 6, 198, 163, 82, 6, 197, 194, 29, 1, 198, 3, 90, 6, 198, 99, 96, 6, 198, 163, 121, 6, 198, 226, 163, 6, 198, 162, 200, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 58, 7, 198, 35, 72, 7, 198, 99, 85, 7, 192, 0, 0, 0, 198, 163, 103, 7, 192, 0, 0, 0, 198, 67, 111, 7, 198, 227, 173, 7, 198, 162, 0, 8, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 254, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 50, 0, 198, 131, 51, 0, 198, 227, 61, 0, 192, 0, 0, 0, 198, 130, 79, 0, 198, 99, 87, 0, 198, 67, 85, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 103, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 109, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 106, 0, 198, 226, 111, 0, 192, 0, 0, 0, 198, 99, 113, 0, 198, 35, 114, 0, 192, 0, 0, 0, 198, 67, 115, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 124, 0, 192, 0, 0, 0, 198, 163, 132, 0, 192, 0, 0, 0, 197, 163, 22, 0, 198, 35, 140, 0, 192, 0, 0, 0, 198, 163, 147, 0, 198, 3, 148, 0, 198, 195, 152, 0, 198, 195, 153, 0, 198, 2, 156, 0, 192, 0, 0, 0, 198, 195, 157, 0, 197, 66, 28, 0, 198, 66, 164, 0, 198, 67, 165, 0, 198, 195, 168, 0, 198, 67, 171, 0, 198, 3, 172, 0, 198, 3, 172, 0, 198, 3, 172, 0, 192, 0, 0, 0, 197, 34, 177, 0, 198, 195, 177, 0, 198, 194, 179, 0, 192, 0, 0, 0, 198, 2, 182, 0, 198, 131, 183, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 34, 0, 198, 67, 192, 0, 198, 226, 194, 0, 198, 99, 200, 0, 192, 0, 0, 0, 198, 3, 208, 0, 198, 2, 212, 0, 197, 195, 42, 0, 198, 163, 215, 0, 198, 163, 215, 0, 197, 2, 44, 0, 198, 227, 219, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 248, 0, 198, 99, 225, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 234, 0, 198, 35, 125, 0, 192, 0, 0, 0, 198, 131, 8, 1, 198, 195, 11, 1, 192, 0, 0, 0, 198, 195, 3, 1, 192, 0, 0, 0, 198, 194, 19, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 42, 1, 198, 2, 44, 1, 198, 67, 44, 1, 198, 34, 51, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 66, 0, 198, 130, 68, 1, 198, 2, 70, 1, 192, 0, 0, 0, 198, 2, 74, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 88, 1, 198, 3, 92, 1, 198, 3, 92, 1, 192, 0, 0, 0, 198, 163, 110, 1, 197, 226, 74, 0, 198, 131, 115, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 131, 1, 192, 0, 0, 0, 198, 194, 142, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 164, 1, 198, 34, 165, 1, 192, 0, 0, 0, 198, 67, 168, 1, 198, 66, 172, 1, 198, 98, 174, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 187, 1, 198, 66, 188, 1, 198, 163, 190, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 250, 0, 198, 66, 202, 1, 198, 66, 202, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 208, 1, 198, 34, 210, 1, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 222, 1, 198, 162, 223, 1, 192, 0, 0, 0, 198, 35, 242, 1, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 99, 0, 192, 0, 0, 0, 198, 163, 251, 1, 198, 66, 10, 2, 197, 67, 106, 0, 192, 0, 0, 0, 197, 67, 106, 0, 198, 162, 17, 2, 198, 35, 20, 2, 198, 3, 25, 2, 198, 67, 22, 2, 198, 163, 24, 2, 198, 67, 30, 2, 198, 131, 31, 2, 198, 226, 31, 2, 198, 131, 35, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 55, 2, 198, 3, 61, 2, 198, 99, 74, 2, 192, 0, 0, 0, 198, 3, 71, 2, 192, 0, 0, 0, 198, 194, 75, 2, 198, 226, 78, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 93, 2, 192, 0, 0, 0, 198, 67, 84, 2, 192, 0, 0, 0, 198, 34, 98, 2, 192, 0, 0, 0, 197, 226, 119, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 3, 124, 2, 198, 3, 127, 2, 192, 0, 0, 0, 198, 226, 141, 2, 198, 66, 182, 2, 198, 99, 158, 2, 192, 0, 0, 0, 198, 130, 168, 2, 192, 0, 0, 0, 192, 0, 0, 0, 198, 130, 112, 0, 198, 3, 113, 0, 197, 226, 131, 0, 192, 0, 0, 0, 198, 131, 94, 5, 192, 0, 0, 0, 198, 99, 186, 2, 198, 195, 186, 2, 192, 0, 0, 0, 198, 66, 195, 2, 197, 34, 133, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 226, 2, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 245, 2, 192, 0, 0, 0, 198, 130, 6, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 27, 3, 192, 0, 0, 0, 198, 98, 34, 3, 192, 0, 0, 0, 197, 226, 148, 0, 192, 0, 0, 0, 198, 162, 40, 3, 192, 0, 0, 0, 198, 195, 48, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 60, 3, 192, 0, 0, 0, 198, 227, 75, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 160, 0, 197, 98, 161, 0, 198, 130, 91, 3, 197, 34, 161, 0, 197, 163, 161, 0, 198, 131, 96, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 117, 3, 198, 99, 125, 3, 192, 0, 0, 0, 198, 98, 135, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 112, 3, 198, 226, 152, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 164, 3, 192, 0, 0, 0, 198, 35, 176, 3, 198, 3, 175, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 183, 3, 192, 0, 0, 0, 198, 67, 188, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 220, 3, 198, 195, 221, 3, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 230, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 247, 3, 198, 99, 250, 3, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 200, 0, 198, 163, 25, 4, 192, 0, 0, 0, 198, 3, 35, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 55, 4, 192, 0, 0, 0, 198, 194, 70, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 96, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 100, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 111, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 131, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 151, 4, 192, 0, 0, 0, 192, 0, 0, 0, 198, 67, 157, 4, 198, 195, 167, 4, 192, 0, 0, 0, 198, 3, 176, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 219, 4, 198, 99, 220, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 250, 4, 192, 0, 0, 0, 198, 35, 9, 5, 192, 0, 0, 0, 197, 162, 234, 0, 192, 0, 0, 0, 198, 98, 39, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 73, 5, 198, 66, 76, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 85, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 163, 95, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 107, 5, 198, 227, 78, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 250, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 144, 5, 198, 66, 146, 5, 192, 0, 0, 0, 197, 34, 253, 0, 198, 195, 150, 5, 198, 98, 152, 5, 198, 130, 157, 5, 192, 0, 0, 0, 198, 35, 157, 5, 198, 226, 163, 5, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 164, 5, 198, 99, 176, 5, 192, 0, 0, 0, 197, 98, 4, 1, 192, 0, 0, 0, 198, 227, 178, 5, 198, 67, 179, 5, 198, 194, 180, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 195, 203, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 214, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 10, 1, 198, 66, 234, 5, 197, 66, 12, 1, 192, 0, 0, 0, 197, 2, 13, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 250, 5, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 18, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 131, 71, 6, 197, 194, 29, 1, 198, 99, 96, 6, 198, 162, 102, 6, 192, 0, 0, 0, 198, 163, 109, 6, 198, 3, 113, 6, 198, 227, 124, 6, 198, 162, 127, 6, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 131, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 163, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 99, 239, 6, 198, 131, 253, 6, 192, 0, 0, 0, 198, 131, 246, 6, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 36, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 81, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 100, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 124, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 227, 173, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 233, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 248, 7, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 89, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 251, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 16, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 180, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 79, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 185, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 102, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 226, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 32, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 25, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 128, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 7, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 107, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 181, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 49, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 70, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 198, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 117, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 219, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 175, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 154, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 38, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 30, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 158, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 108, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 128, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 31, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 50, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 100, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 109, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 74, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 9, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 81, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 78, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 10, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 193, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 176, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 74, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 213, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 2, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 53, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 129, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 47, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 92, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 124, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 112, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 220, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 185, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 252, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 236, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 240, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 130, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 99, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 41, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 32, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 51, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 84, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 209, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 10, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 35, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 55, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 253, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 107, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 180, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 125, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 39, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 232, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 121, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 63, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 122, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 26, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 14, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 110, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 24, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 168, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 134, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 47, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 131, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 80, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 227, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 176, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 158, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 138, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 145, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 136, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 77, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 48, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 27, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 36, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 220, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 170, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 214, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 214, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 169, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 79, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 234, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 241, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 39, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 52, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 49, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 116, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 186, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 40, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 109, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 200, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 152, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 247, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 89, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 78, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 99, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 132, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 133, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 252, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 52, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 32, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 40, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 6, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 161, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 22, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 31, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 93, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 40, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 59, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 12, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 90, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 181, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 151, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 162, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 229, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 97, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 248, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 153, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 253, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 29, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 53, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 33, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 151, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 109, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 39, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 91, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 134, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 100, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 231, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 51, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 130, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 4, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 3, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 16, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 83, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 32, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 92, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 229, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 175, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 26, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 159, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 27, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 71, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 68, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 240, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 15, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 85, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 126, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 23, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 49, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 100, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 163, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 187, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 249, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 61, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 105, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 23, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 17, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 150, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 56, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 23, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 62, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 15, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 246, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 175, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 45, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 135, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 101, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 215, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 144, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 228, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 111, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 84, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 164, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 55, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 159, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 124, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 171, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 179, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 182, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 11, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 108, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 157, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 170, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 83, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 56, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 145, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 138, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 45, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 45, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 202, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 61, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 57, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 27, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 158, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 6, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 206, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 43, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 113, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 224, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 60, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 165, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 230, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 229, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 56, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 72, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 75, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 254, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 19, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 127, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 86, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 14, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 40, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 255, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 33, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 254, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 21, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 76, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 215, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 168, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 111, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 25, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 255, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 94, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 32, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 229, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 150, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 207, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 75, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 48, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 180, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 224, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 129, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 173, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 188, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 28, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 151, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 152, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 188, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 34, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 164, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 11, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 103, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 115, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 58, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 78, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 81, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 80, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 3, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 177, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 104, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 85, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 110, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 54, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 103, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 166, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 2, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 51, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 76, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 77, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 21, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 176, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 151, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 255, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 135, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 59, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 136, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 250, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 223, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 167, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 222, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 59, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 43, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 57, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 254, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 8, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 152, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 72, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 241, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 28, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 63, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 230, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 105, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 9, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 33, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 185, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 197, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 95, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 70, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 31, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 57, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 198, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 241, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 203, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 125, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 114, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 169, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 180, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 11, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 121, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 106, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 178, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 4, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 34, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 164, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 45, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 101, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 112, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 56, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 204, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 88, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 237, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 39, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 159, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 35, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 109, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 8, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 28, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 221, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 29, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 183, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 41, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 112, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 36, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 42, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 42, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 113, 121, 0, 5, 71, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 112, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 84, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 25, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 100, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 69, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 165, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 165, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 152, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 138, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 206, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 6, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 198, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 159, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 146, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 127, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 126, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 250, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 248, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 251, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 110, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 153, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 107, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 50, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 160, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 181, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 214, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 213, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 139, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 195, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 196, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 147, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 171, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 150, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 148, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 24, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 19, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 139, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 153, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 123, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 21, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 78, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 232, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 146, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 12, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 2, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 108, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 251, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 49, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 104, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 194, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 132, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 133, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 247, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 225, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 49, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 140, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 182, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 70, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 249, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 114, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 44, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 48, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 93, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 101, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 21, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 102, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 35, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 5, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 37, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 62, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 172, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 230, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 118, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 189, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 126, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 161, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 104, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 34, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 170, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 135, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 38, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 231, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 187, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 159, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 236, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 18, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 50, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 36, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 136, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 32, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 58, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 59, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 199, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 215, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 239, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 30, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 2, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 252, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 81, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 108, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 197, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 129, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 166, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 78, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 60, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 206, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 22, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 153, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 166, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 123, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 61, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 218, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 161, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 186, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 66, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 41, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 52, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 107, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 70, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 224, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 8, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 120, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 82, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 79, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 115, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 152, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 105, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 113, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 227, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 157, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 163, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 188, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 7, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 51, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 8, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 9, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 111, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 130, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 108, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 84, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 81, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 5, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 134, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 131, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 167, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 79, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 95, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 26, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 114, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 64, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 146, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 40, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 88, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 53, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 189, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 90, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 181, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 132, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 75, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 16, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 162, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 112, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 60, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 50, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 34, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 155, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 43, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 37, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 124, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 24, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 111, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 94, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 3, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 116, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 190, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 95, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 104, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 110, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 15, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 62, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 243, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 48, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 235, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 248, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 154, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 181, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 27, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 36, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 96, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 102, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 113, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 168, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 61, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 22, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 32, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 96, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 28, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 117, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 97, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 91, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 43, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 145, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 92, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 205, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 130, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 115, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 209, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 116, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 94, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 178, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 210, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 55, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 239, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 9, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 53, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 3, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 52, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 137, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 98, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 169, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 138, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 128, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 95, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 30, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 121, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 144, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 162, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 7, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 249, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 46, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 36, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 129, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 208, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 91, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 232, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 22, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 156, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 73, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 191, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 25, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 147, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 190, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 121, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 2, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 211, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 75, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 117, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 250, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 33, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 95, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 2, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 37, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 36, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 7, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 73, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 212, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 213, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 4, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 203, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 102, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 8, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 76, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 12, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 10, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 191, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 117, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 236, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 233, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 29, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 17, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 16, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 49, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 86, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 105, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 125, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 71, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 49, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 122, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 177, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 26, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 182, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 5, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 118, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 51, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 19, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 65, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 217, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 25, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 11, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 102, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 30, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 139, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 205, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 60, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 250, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 191, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 216, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 236, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 37, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 206, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 160, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 136, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 89, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 61, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 105, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 106, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 148, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 24, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 237, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 255, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 35, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 102, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 114, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 144, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 3, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 91, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 200, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 21, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 179, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 26, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 221, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 8, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 97, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 22, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 17, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 147, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 82, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 112, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 120, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 107, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 8, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 33, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 87, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 41, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 40, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 95, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 25, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 206, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 192, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 49, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 43, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 29, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 8, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 115, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 140, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 217, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 146, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 38, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 110, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 118, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 155, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 6, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 82, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 129, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 111, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 118, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 88, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 108, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 240, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 109, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 85, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 64, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 65, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 188, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 218, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 14, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 2, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 114, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 80, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 121, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 38, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 224, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 44, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 18, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 156, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 50, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 112, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 62, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 23, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 91, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 115, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 123, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 231, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 113, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 157, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 27, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 248, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 37, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 72, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 126, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 32, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 153, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 74, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 77, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 35, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 230, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 233, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 50, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 147, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 141, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 124, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 126, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 57, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 169, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 111, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 79, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 213, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 3, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 158, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 18, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 12, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 6, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 117, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 16, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 154, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 45, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 184, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 183, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 141, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 42, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 207, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 47, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 232, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 27, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 219, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 56, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 173, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 2, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 45, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 123, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 3, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 105, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 106, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 208, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 41, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 238, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 67, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 102, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 120, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 170, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 119, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 76, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 72, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 86, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 78, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 138, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 103, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 63, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 44, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 23, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 119, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 31, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 170, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 32, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 148, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 137, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 13, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 178, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 13, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 149, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 120, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 185, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 66, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 3, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 159, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 50, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 106, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 154, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 252, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 107, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 76, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 58, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 36, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 7, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 210, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 234, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 174, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 4, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 12, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 87, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 25, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 11, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 85, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 46, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 15, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 47, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 26, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 41, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 192, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 179, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 193, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 127, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 181, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 254, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 156, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 13, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 139, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 5, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 46, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 129, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 47, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 28, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 68, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 73, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 47, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 82, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 142, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 25, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 199, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 77, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 33, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 209, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 59, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 172, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 27, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 241, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 115, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 143, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 193, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 250, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 131, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 79, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 39, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 193, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 113, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 207, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 23, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 40, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 133, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 44, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 33, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 199, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 64, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 20, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 155, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 60, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 105, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 97, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 38, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 24, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 205, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 208, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 126, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 186, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 25, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 16, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 3, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 151, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 24, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 194, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 83, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 233, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 137, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 152, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 191, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 43, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 59, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 179, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 68, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 6, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 124, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 4, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 4, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 34, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 63, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 249, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 142, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 80, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 84, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 92, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 21, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 149, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 176, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 42, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 30, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 48, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 73, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 45, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 238, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 247, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 140, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 6, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 139, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 148, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 235, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 37, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 66, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 137, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 50, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 7, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 127, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 49, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 160, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 160, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 57, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 39, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 26, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 41, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 25, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 116, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 41, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 143, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 195, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 222, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 13, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 170, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 85, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 74, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 26, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 198, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 164, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 234, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 133, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 33, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 81, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 50, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 25, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 39, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 117, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 175, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 2, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 4, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 209, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 44, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 19, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 108, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 3, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 71, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 96, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 50, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 86, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 237, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 214, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 192, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 77, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 82, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 128, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 34, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 33, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 114, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 171, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 93, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 235, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 51, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 53, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 53, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 29, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 125, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 127, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 51, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 149, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 144, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 233, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 234, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 31, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 196, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 242, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 30, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 103, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 103, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 5, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 182, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 63, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 4, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 161, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 119, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 120, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 69, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 238, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 58, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 12, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 134, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 141, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 236, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 54, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 235, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 55, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 51, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 234, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 16, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 109, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 145, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 214, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 186, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 169, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 94, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 87, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 220, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 41, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 138, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 235, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 114, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 16, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 216, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 172, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 40, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 7, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 150, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 187, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 33, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 54, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 78, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 173, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 209, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 200, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 20, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 122, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 71, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 80, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 75, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 54, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 67, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 95, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 221, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 81, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 77, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 78, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 66, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 72, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 25, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 180, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 88, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 118, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 44, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 157, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 74, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 249, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 165, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 238, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 79, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 237, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 10, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 59, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 109, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 250, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 178, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 249, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 52, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 31, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 136, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 75, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 53, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 29, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 237, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 54, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 22, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 6, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 27, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 215, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 204, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 168, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 46, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 113, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 238, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 76, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 122, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 238, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 174, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 25, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 99, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 151, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 85, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 32, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 65, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 239, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 125, 121, 0, 5, 249, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 113, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 110, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 225, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 139, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 40, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 4, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 240, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 92, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 181, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 138, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 78, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 242, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 28, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 31, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 242, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 8, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 104, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 156, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 47, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 13, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 123, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 188, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 51, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 128, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 222, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 121, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 239, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 110, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 62, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 100, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 26, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 228, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 210, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 45, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 197, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 139, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 227, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 14, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 155, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 145, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 140, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 55, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 253, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 132, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 48, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 140, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 146, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 161, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 147, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 42, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 86, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 42, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 30, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 156, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 73, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 140, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 144, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 243, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 244, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 73, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 4, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 110, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 210, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 211, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 101, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 210, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 188, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 88, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 171, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 63, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 216, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 31, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 215, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 77, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 46, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 176, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 62, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 28, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 76, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 42, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 204, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 90, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 89, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 43, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 156, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 43, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 184, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 145, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 70, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 120, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 251, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 174, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 63, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 17, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 226, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 212, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 106, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 86, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 124, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 5, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 52, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 70, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 50, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 42, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 164, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 129, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 175, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 101, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 82, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 137, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 185, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 152, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 135, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 97, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 141, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 255, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 122, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 86, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 80, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 51, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 153, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 179, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 111, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 104, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 121, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 83, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 162, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 163, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 205, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 24, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 146, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 254, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 44, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 61, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 131, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 53, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 180, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 171, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 64, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 51, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 130, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 231, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 46, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 98, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 75, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 130, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 62, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 150, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 43, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 88, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 248, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 216, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 69, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 130, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 225, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 146, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 63, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 30, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 175, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 83, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 119, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 33, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 122, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 147, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 216, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 39, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 84, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 217, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 141, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 89, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 218, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 55, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 57, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 210, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 81, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 22, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 2, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 34, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 73, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 200, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 183, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 77, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 138, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 59, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 149, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 41, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 132, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 245, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 32, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 90, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 63, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 42, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 31, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 54, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 52, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 148, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 55, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 98, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 169, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 216, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 91, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 13, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 196, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 42, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 111, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 243, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 55, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 5, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 154, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 154, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 15, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 149, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 75, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 87, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 142, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 5, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 112, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 72, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 5, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 3, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 211, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 200, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 201, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 211, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 89, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 127, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 75, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 194, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 13, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 250, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 236, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 74, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 60, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 11, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 236, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 79, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 68, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 185, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 46, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 237, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 26, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 55, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 131, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 88, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 135, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 113, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 107, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 23, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 34, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 52, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 65, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 240, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 114, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 148, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 217, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 6, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 177, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 139, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 24, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 131, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 157, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 131, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 30, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 115, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 217, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 127, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 207, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 123, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 99, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 56, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 99, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 56, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 42, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 44, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 116, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 40, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 162, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 65, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 7, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 41, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 120, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 215, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 201, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 145, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 32, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 7, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 217, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 41, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 61, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 108, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 64, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 89, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 178, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 27, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 92, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 243, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 110, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 129, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 182, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 53, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 180, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 236, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 25, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 199, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 237, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 228, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 35, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 89, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 132, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 176, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 198, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 120, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 133, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 217, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 184, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 233, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 44, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 54, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 74, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 14, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 180, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 74, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 187, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 15, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 111, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 250, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 89, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 163, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 211, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 212, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 211, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 183, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 183, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 8, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 158, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 143, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 133, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 17, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 245, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 25, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 66, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 149, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 10, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 86, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 150, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 134, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 226, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 132, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 7, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 43, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 96, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 114, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 134, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 237, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 27, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 218, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 198, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 23, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 27, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 28, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 157, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 181, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 56, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 130, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 141, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 4, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 34, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 35, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 14, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 135, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 136, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 122, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 128, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 222, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 6, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 66, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 109, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 12, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 53, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 153, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 27, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 85, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 165, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 38, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 200, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 56, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 14, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 9, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 20, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 82, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 151, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 55, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 118, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 83, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 79, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 189, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 34, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 35, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 130, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 83, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 135, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 93, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 67, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 16, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 96, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 33, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 56, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 26, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 251, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 5, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 50, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 42, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 55, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 63, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 150, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 96, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 104, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 13, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 34, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 28, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 60, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 181, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 86, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 117, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 140, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 218, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 64, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 45, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 158, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 66, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 45, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 198, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 121, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 133, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 252, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 227, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 178, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 136, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 24, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 183, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 123, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 29, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 50, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 150, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 45, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 31, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 9, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 6, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 26, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 15, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 7, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 236, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 124, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 78, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 44, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 37, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 19, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 76, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 184, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 4, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 3, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 28, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 232, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 116, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 17, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 103, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 80, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 28, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 97, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 87, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 155, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 213, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 218, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 124, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 100, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 5, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 102, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 52, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 70, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 36, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 28, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 110, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 164, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 173, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 105, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 174, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 190, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 158, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 246, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 182, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 151, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 56, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 43, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 17, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 242, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 73, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 24, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 25, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 35, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 98, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 223, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 185, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 141, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 97, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 175, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 228, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 58, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 2, 127, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 71, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 176, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 67, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 134, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 84, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 64, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 196, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 134, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 103, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 177, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 137, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 115, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 206, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 237, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 142, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 52, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 160, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 121, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 28, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 163, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 90, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 20, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 125, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 68, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 43, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 36, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 55, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 223, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 179, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 239, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 127, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 104, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 71, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 83, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 159, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 238, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 140, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 199, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 44, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 219, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 114, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 251, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 153, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 159, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 243, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 199, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 222, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 26, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 26, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 124, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 78, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 238, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 195, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 45, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 147, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 156, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 143, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 81, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 105, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 30, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 119, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 121, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 65, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 130, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 37, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 72, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 18, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 185, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 78, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 79, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 222, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 46, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 101, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 144, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 177, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 251, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 213, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 246, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 87, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 105, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 182, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 138, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 115, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 157, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 29, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 49, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 128, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 104, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 4, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 44, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 2, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 154, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 178, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 142, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 29, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 119, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 115, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 241, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 139, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 34, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 244, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 160, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 54, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 57, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 67, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 53, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 160, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 65, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 224, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 27, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 17, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 98, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 57, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 80, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 15, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 84, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 7, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 16, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 237, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 60, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 184, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 75, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 59, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 12, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 46, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 14, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 105, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 18, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 93, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 163, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 244, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 37, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 51, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 64, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 66, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 207, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 14, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 5, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 125, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 76, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 54, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 87, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 51, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 29, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 31, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 208, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 35, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 118, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 31, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 50, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 218, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 67, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 245, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 5, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 6, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 87, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 84, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 247, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 151, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 2, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 65, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 248, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 208, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 46, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 10, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 34, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 186, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 224, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 135, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 179, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 182, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 94, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 10, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 79, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 29, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 27, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 33, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 17, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 79, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 164, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 152, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 3, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 161, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 65, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 66, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 119, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 7, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 28, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 15, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 28, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 247, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 235, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 38, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 158, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 152, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 19, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 90, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 111, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 183, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 88, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 245, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 43, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 57, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 252, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 16, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 24, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 36, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 128, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 112, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 52, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 61, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 239, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 79, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 58, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 6, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 41, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 42, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 252, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 52, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 131, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 6, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 189, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 8, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 21, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 170, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 191, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 225, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 84, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 180, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 107, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 170, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 164, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 238, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 248, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 2, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 153, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 159, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 209, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 252, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 9, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 205, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 116, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 94, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 184, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 61, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 186, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 95, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 124, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 135, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 207, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 80, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 7, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 181, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 74, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 75, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 25, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 180, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 108, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 108, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 107, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 177, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 3, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 53, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 18, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 69, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 196, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 116, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 28, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 208, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 67, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 47, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 23, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 182, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 165, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 14, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 91, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 8, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 64, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 192, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 109, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 150, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 102, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 77, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 113, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 69, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 2, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 26, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 240, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 146, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 138, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 66, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 145, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 3, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 185, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 226, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 46, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 67, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 9, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 108, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 67, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 4, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 95, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 17, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 61, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 246, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 71, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 12, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 253, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 37, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 74, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 201, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 239, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 128, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 91, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 227, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 8, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 112, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 160, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 219, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 176, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 37, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 117, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 214, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 11, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 220, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 238, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 29, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 92, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 126, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 126, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 17, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 24, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 200, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 165, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 73, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 227, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 27, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 70, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 254, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 221, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 12, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 125, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 242, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 161, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 183, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 99, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 218, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 149, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 251, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 140, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 96, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 89, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 57, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 66, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 153, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 33, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 47, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 125, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 84, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 68, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 243, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 154, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 27, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 154, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 106, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 166, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 103, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 45, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 215, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 8, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 94, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 7, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 216, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 113, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 80, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 97, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 239, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 184, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 13, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 115, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 33, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 58, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 48, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 59, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 93, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 46, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 9, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 114, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 136, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 188, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 65, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 154, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 104, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 184, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 127, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 128, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 28, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 201, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 178, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 31, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 172, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 55, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 217, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 166, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 9, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 13, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 30, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 27, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 212, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 77, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 77, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 186, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 82, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 212, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 45, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 98, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 137, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 111, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 155, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 182, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 249, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 95, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 189, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 42, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 28, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 167, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 18, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 187, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 39, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 115, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 96, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 117, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 210, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 211, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 135, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 5, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 88, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 255, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 27, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 43, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 83, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 35, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 87, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 202, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 107, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 187, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 4, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 208, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 49, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 71, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 88, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 78, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 146, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 14, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 14, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 15, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 10, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 147, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 9, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 137, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 74, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 213, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 30, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 72, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 140, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 255, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 10, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 118, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 222, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 168, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 114, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 71, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 171, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 76, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 29, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 91, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 80, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 68, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 171, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 72, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 29, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 6, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 115, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 48, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 51, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 240, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 241, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 36, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 53, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 31, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 252, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 5, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 8, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 62, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 72, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 235, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 231, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 132, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 55, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 113, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 67, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 73, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 83, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 32, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 117, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 116, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 188, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 207, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 79, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 105, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 151, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 118, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 115, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 103, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 48, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 128, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 36, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 118, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 60, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 61, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 181, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 248, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 65, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 189, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 219, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 238, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 59, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 47, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 20, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 44, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 11, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 166, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 6, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 132, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 240, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 186, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 253, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 47, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 30, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 142, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 92, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 90, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 62, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 64, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 155, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 88, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 208, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 58, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 89, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 76, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 164, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 251, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 43, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 51, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 67, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 106, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 141, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 185, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 218, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 43, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 17, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 79, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 91, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 207, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 112, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 11, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 223, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 80, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 117, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 12, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 10, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 11, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 127, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 97, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 15, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 188, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 35, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 243, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 155, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 77, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 74, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 17, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 103, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 73, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 77, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 5, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 167, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 22, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 128, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 241, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 12, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 161, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 45, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 34, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 98, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 85, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 155, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 118, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 156, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 16, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 137, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 47, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 33, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 247, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 54, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 3, 127, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 93, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 156, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 46, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 75, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 190, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 46, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 252, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 81, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 85, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 42, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 78, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 31, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 77, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 59, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 31, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 154, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 17, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 18, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 89, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 95, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 56, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 165, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 117, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 32, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 212, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 58, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 20, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 99, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 109, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 136, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 242, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 111, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 59, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 113, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 240, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 68, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 32, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 74, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 249, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 119, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 241, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 55, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 164, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 203, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 254, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 157, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 32, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 173, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 250, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 197, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 220, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 64, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 189, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 167, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 116, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 63, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 23, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 174, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 179, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 165, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 229, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 60, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 120, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 38, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 152, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 166, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 197, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 114, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 33, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 125, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 187, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 34, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 155, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 142, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 44, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 45, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 242, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 60, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 27, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 122, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 30, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 180, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 56, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 187, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 11, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 60, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 244, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 35, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 138, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 193, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 2, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 75, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 71, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 157, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 7, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 219, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 49, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 57, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 250, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 51, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 18, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 221, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 209, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 17, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 28, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 19, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 92, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 56, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 138, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 122, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 39, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 214, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 34, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 21, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 135, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 159, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 183, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 213, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 219, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 45, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 80, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 4, 127, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 142, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 194, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 215, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 35, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 35, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 119, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 62, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 253, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 224, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 97, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 82, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 104, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 129, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 169, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 44, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 38, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 47, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 178, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 38, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 57, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 190, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 177, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 99, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 73, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 254, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 59, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 121, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 86, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 144, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 7, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 245, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 76, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 105, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 112, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 225, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 71, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 30, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 76, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 93, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 78, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 3, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 204, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 58, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 167, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 162, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 2, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 204, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 75, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 6, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 46, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 94, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 13, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 221, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 139, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 35, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 142, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 59, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 65, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 118, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 16, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 182, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 239, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 156, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 47, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 79, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 18, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 225, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 28, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 16, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 206, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 139, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 22, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 35, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 106, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 60, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 94, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 186, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 252, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 117, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 161, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 116, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 156, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 110, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 71, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 255, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 21, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 63, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 247, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 248, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 133, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 56, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 28, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 17, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 181, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 39, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 4, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 120, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 95, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 202, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 5, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 228, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 99, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 32, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 197, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 67, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 66, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 143, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 92, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 90, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 52, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 13, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 122, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 130, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 214, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 93, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 39, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 13, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 19, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 33, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 50, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 50, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 117, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 9, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 241, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 15, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 158, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 230, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 29, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 161, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 40, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 214, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 140, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 163, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 241, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 31, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 115, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 222, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 136, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 129, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 157, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 219, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 253, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 172, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 209, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 48, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 152, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 178, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 98, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 48, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 85, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 131, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 51, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 79, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 114, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 11, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 182, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 143, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 144, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 145, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 146, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 40, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 80, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 132, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 5, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 247, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 188, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 49, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 26, 121, 0, 5, 82, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 129, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 23, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 46, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 67, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 143, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 19, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 164, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 143, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 216, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 198, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 20, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 20, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 133, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 12, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 168, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 199, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 155, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 80, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 45, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 228, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 66, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 225, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 204, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 126, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 195, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 135, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 14, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 59, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 186, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 120, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 113, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 150, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 66, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 68, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 31, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 158, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 69, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 205, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 233, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 192, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 60, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 36, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 30, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 190, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 231, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 109, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 168, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 228, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 123, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 181, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 159, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 15, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 239, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 123, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 14, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 17, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 147, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 251, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 225, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 5, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 37, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 15, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 158, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 210, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 98, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 151, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 168, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 220, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 8, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 94, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 95, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 160, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 85, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 78, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 200, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 36, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 31, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 100, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 18, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 43, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 32, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 47, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 28, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 98, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 130, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 45, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 217, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 144, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 168, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 4, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 253, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 51, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 16, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 10, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 165, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 255, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 167, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 76, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 215, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 159, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 71, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 187, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 48, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 141, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 9, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 193, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 21, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 36, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 81, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 3, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 254, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 218, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 243, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 125, 121, 0, 5, 253, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 47, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 19, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 239, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 129, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 255, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 244, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 198, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 188, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 118, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 75, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 252, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 75, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 160, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 38, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 65, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 39, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 76, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 21, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 22, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 155, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 54, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 246, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 8, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 183, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 123, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 15, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 21, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 121, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 81, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 70, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 47, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 208, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 48, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 80, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 157, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 230, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 72, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 182, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 68, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 183, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 242, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 220, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 18, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 234, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 95, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 22, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 232, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 92, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 12, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 159, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 142, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 40, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 134, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 14, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 147, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 129, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 152, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 6, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 49, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 188, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 76, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 23, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 68, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 158, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 161, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 209, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 8, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 160, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 29, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 61, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 185, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 84, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 218, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 64, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 44, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 222, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 33, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 15, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 90, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 223, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 251, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 49, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 49, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 176, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 54, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 47, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 83, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 119, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 243, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 32, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 24, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 138, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 161, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 131, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 112, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 194, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 19, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 52, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 207, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 195, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 253, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 4, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 37, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 134, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 12, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 94, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 245, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 249, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 34, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 10, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 253, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 216, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 211, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 162, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 162, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 20, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 27, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 202, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 220, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 202, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 170, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 232, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 57, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 69, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 16, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 34, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 159, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 152, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 126, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 226, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 125, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 175, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 189, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 9, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 68, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 18, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 235, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 100, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 73, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 178, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 22, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 41, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 122, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 61, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 184, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 212, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 67, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 240, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 190, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 16, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 46, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 157, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 166, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 171, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 122, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 53, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 18, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 96, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 219, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 103, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 210, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 67, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 12, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 99, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 77, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 49, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 72, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 77, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 111, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 82, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 36, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 172, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 69, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 244, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 123, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 165, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 240, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 116, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 168, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 53, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 209, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 85, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 201, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 200, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 30, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 29, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 96, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 119, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 50, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 90, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 223, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 6, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 243, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 241, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 216, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 162, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 68, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 17, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 55, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 183, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 50, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 205, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 84, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 245, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 15, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 80, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 6, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 220, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 220, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 48, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 100, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 212, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 66, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 232, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 220, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 254, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 184, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 160, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 229, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 5, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 129, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 7, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 255, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 78, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 240, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 169, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 5, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 17, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 12, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 189, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 162, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 160, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 38, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 101, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 202, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 2, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 112, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 69, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 221, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 144, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 242, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 25, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 254, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 66, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 10, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 222, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 239, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 223, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 172, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 18, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 19, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 6, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 79, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 144, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 118, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 18, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 13, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 123, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 191, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 119, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 61, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 135, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 156, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 35, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 46, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 118, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 76, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 162, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 104, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 59, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 68, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 58, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 10, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 26, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 255, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 29, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 69, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 124, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 37, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 19, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 38, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 190, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 51, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 157, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 54, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 254, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 20, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 206, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 48, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 2, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 209, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 42, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 77, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 91, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 166, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 75, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 6, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 91, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 32, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 21, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 31, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 13, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 52, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 60, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 108, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 169, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 105, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 43, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 241, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 51, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 223, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 73, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 76, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 44, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 191, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 179, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 107, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 115, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 124, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 2, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 77, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 6, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 59, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 186, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 52, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 33, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 20, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 224, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 148, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 20, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 163, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 50, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 72, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 167, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 179, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 207, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 215, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 119, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 109, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 49, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 45, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 143, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 32, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 89, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 58, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 210, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 2, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 78, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 46, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 166, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 114, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 224, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 195, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 125, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 184, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 167, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 51, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 76, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 108, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 161, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 7, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 172, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 19, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 105, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 113, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 53, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 235, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 7, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 156, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 56, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 203, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 126, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 71, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 196, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 158, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 246, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 240, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 120, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 56, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 121, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 123, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 97, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 88, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 107, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 85, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 241, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 182, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 9, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 13, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 106, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 109, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 105, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 3, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 81, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 82, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 11, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 191, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 229, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 2, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 96, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 79, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 194, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 35, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 2, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 49, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 226, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 9, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 14, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 195, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 201, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 170, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 96, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 122, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 11, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 127, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 185, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 61, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 186, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 136, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 95, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 160, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 147, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 168, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 205, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 3, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 163, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 242, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 130, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 119, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 9, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 172, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 76, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 24, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 99, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 77, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 210, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 84, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 45, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 206, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 197, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 69, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 206, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 163, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 98, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 208, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 37, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 219, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 70, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 15, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 225, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 102, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 20, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 169, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 16, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 57, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 240, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 30, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 60, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 152, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 126, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 207, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 187, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 27, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 153, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 5, 127, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 137, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 106, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 71, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 23, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 131, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 191, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 71, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 173, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 244, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 124, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 4, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 83, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 146, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 61, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 31, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 169, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 14, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 77, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 36, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 132, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 229, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 217, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 99, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 242, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 12, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 38, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 3, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 52, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 164, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 143, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 54, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 57, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 38, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 21, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 60, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 129, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 100, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 62, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 53, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 85, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 28, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 32, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 187, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 138, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 46, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 131, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 4, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 170, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 39, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 245, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 52, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 51, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 8, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 77, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 72, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 176, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 73, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 168, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 246, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 231, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 145, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 106, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 163, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 105, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 209, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 29, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 148, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 9, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 96, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 16, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 107, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 83, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 139, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 188, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 94, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 198, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 201, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 226, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 36, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 143, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 40, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 164, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 8, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 11, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 218, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 22, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 208, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 208, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 132, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 43, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 86, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 189, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 45, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 246, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 87, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 74, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 5, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 185, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 189, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 242, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 163, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 99, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 211, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 44, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 170, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 78, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 219, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 75, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 79, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 140, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 185, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 127, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 209, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 171, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 52, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 80, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 190, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 57, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 80, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 67, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 227, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 146, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 186, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 3, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 108, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 170, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 193, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 116, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 61, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 191, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 10, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 196, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 186, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 29, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 87, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 210, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 147, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 194, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 77, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 161, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 7, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 137, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 97, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 230, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 98, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 3, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 165, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 55, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 166, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 53, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 136, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 109, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 20, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 54, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 29, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 30, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 107, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 110, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 88, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 35, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 206, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 55, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 110, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 101, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 41, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 50, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 220, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 33, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 54, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 89, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 253, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 254, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 79, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 7, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 58, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 198, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 76, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 136, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 77, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 137, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 53, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 148, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 5, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 4, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 63, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 187, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 231, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 133, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 179, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 213, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 20, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 78, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 122, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 62, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 17, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 221, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 169, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 58, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 188, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 15, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 47, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 192, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 211, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 90, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 59, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 7, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 92, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 81, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 100, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 187, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 6, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 37, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 196, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 246, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 209, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 212, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 188, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 149, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 226, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 10, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 247, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 49, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 10, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 76, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 8, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 11, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 6, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 12, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 70, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 173, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 102, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 42, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 172, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 119, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 232, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 11, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 124, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 74, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 59, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 59, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 127, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 62, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 141, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 244, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 203, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 141, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 91, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 21, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 82, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 203, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 249, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 61, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 223, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 191, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 241, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 165, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 189, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 195, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 130, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 56, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 36, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 144, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 211, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 157, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 158, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 125, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 164, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 98, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 128, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 180, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 3, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 96, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 227, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 4, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 70, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 19, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 197, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 137, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 13, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 108, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 92, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 33, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 228, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 173, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 9, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 46, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 181, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 62, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 86, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 162, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 117, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 26, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 222, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 79, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 103, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 157, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 114, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 115, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 4, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 192, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 134, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 25, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 104, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 123, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 161, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 104, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 144, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 87, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 169, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 50, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 224, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 99, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 245, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 48, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 100, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 125, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 223, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 170, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 80, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 81, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 36, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 33, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 162, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 122, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 84, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 128, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 22, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 90, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 223, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 96, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 5, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 34, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 181, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 71, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 30, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 213, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 126, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 173, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 214, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 16, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 189, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 193, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 2, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 2, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 227, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 163, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 78, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 228, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 224, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 215, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 14, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 3, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 216, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 192, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 210, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 255, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 195, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 204, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 145, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 60, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 80, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 170, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 72, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 8, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 61, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 5, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 139, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 2, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 10, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 52, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 97, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 57, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 85, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 238, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 104, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 100, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 17, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 127, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 140, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 50, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 97, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 217, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 83, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 16, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 63, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 158, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 89, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 164, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 51, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 123, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 150, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 233, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 18, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 17, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 97, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 93, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 21, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 76, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 61, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 47, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 123, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 61, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 136, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 179, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 46, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 43, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 82, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 48, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 153, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 78, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 31, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 225, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 55, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 171, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 151, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 156, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 247, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 173, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 85, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 242, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 142, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 3, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 84, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 116, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 210, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 71, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 63, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 221, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 4, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 12, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 34, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 37, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 224, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 105, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 73, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 90, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 192, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 92, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 81, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 145, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 211, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 83, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 79, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 153, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 246, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 220, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 21, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 106, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 37, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 120, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 62, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 18, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 63, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 126, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 211, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 2, 127, 0, 0, 141, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 74, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 19, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 243, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 107, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 73, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 101, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 165, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 166, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 80, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 31, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 171, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 81, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 82, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 19, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 94, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 109, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 186, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 24, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 190, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 129, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 20, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 130, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 70, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 34, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 110, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 216, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 167, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 36, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 124, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 165, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 83, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 158, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 226, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 30, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 50, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 145, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 32, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 116, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 110, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 109, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 11, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 120, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 192, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 6, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 18, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 37, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 164, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 9, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 236, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 111, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 204, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 23, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 31, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 17, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 174, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 168, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 6, 127, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 193, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 70, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 64, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 152, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 153, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 131, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 101, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 244, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 118, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 225, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 41, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 59, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 38, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 22, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 78, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 93, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 60, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 6, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 137, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 11, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 247, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 33, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 189, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 65, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 84, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 105, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 52, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 10, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 132, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 174, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 232, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 119, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 192, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 106, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 243, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 165, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 16, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 38, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 229, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 69, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 159, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 194, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 129, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 229, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 175, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 225, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 96, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 13, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 10, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 132, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 141, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 244, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 39, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 34, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 32, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 154, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 23, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 11, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 128, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 88, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 97, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 33, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 9, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 60, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 174, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 133, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 121, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 190, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 130, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 190, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 44, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 11, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 153, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 12, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 245, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 10, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 13, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 191, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 226, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 62, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 191, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 159, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 60, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 187, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 63, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 93, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 166, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 83, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 98, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 46, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 28, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 194, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 84, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 100, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 198, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 8, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 171, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 74, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 176, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 86, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 51, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 85, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 111, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 113, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 35, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 135, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 248, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 127, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 167, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 124, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 7, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 40, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 62, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 99, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 130, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 99, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 199, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 247, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 91, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 52, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 84, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 207, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 37, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 204, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 12, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 216, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 55, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 145, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 88, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 57, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 146, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 201, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 13, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 36, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 101, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 51, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 111, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 86, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 25, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 108, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 79, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 160, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 207, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 106, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 88, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 37, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 32, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 64, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 49, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 209, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 182, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 58, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 7, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 72, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 13, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 171, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 84, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 232, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 130, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 35, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 48, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 44, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 14, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 147, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 59, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 125, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 73, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 33, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 84, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 202, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 60, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 89, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 37, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 7, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 34, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 234, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 8, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 237, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 192, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 8, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 18, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 191, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 169, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 39, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 177, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 234, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 92, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 116, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 15, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 236, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 188, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 52, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 63, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 63, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 249, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 74, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 9, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 19, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 9, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 221, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 248, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 230, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 54, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 131, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 154, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 77, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 217, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 93, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 17, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 20, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 68, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 10, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 238, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 80, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 68, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 7, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 14, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 225, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 146, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 172, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 154, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 34, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 100, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 100, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 172, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 134, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 55, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 66, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 82, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 109, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 114, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 221, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 101, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 160, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 113, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 114, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 190, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 121, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 177, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 222, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 248, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 35, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 15, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 177, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 9, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 199, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 193, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 168, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 125, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 241, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 192, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 35, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 135, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 138, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 101, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 250, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 115, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 251, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 146, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 169, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 201, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 79, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 39, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 64, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 85, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 55, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 163, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 229, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 112, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 151, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 152, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 114, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 52, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 56, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 147, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 192, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 170, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 125, 121, 0, 5, 250, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 21, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 121, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 36, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 235, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 10, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 222, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 97, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 64, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 56, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 48, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 75, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 13, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 187, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 175, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 111, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 11, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 81, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 233, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 35, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 136, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 86, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 108, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 151, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 87, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 68, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 22, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 6, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 171, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 98, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 4, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 36, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 200, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 122, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 145, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 170, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 132, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 38, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 67, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 102, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 171, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 139, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 85, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 197, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 142, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 90, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 112, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 36, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 178, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 87, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 178, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 234, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 227, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 57, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 127, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 121, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 122, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 20, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 211, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 217, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 179, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 52, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 161, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 140, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 37, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 30, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 18, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 233, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 86, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 125, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 159, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 53, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 171, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 183, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 79, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 155, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 16, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 139, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 39, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 138, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 193, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 178, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 13, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 53, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 133, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 50, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 120, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 90, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 248, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 172, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 243, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 98, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 243, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 226, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 247, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 16, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 17, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 53, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 14, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 53, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 173, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 227, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 199, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 74, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 36, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 155, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 37, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 169, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 38, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 51, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 235, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 47, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 208, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 202, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 76, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 242, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 136, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 132, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 80, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 154, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 4, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 34, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 55, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 11, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 160, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 17, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 212, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 107, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 194, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 243, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 223, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 84, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 71, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 12, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 180, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 56, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 40, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 198, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 49, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 137, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 227, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 214, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 41, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 11, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 209, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 189, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 21, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 23, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 67, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 48, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 61, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 230, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 47, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 173, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 191, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 19, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 231, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 249, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 200, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 62, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 69, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 201, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 69, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 34, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 114, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 21, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 205, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 29, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 22, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 7, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 148, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 38, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 70, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 26, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 10, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 140, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 161, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 27, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 133, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 202, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 40, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 116, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 205, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 8, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 54, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 218, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 210, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 224, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 193, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 123, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 87, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 11, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 141, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 226, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 230, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 11, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 162, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 126, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 85, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 181, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 131, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 195, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 163, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 94, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 94, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 9, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 122, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 113, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 196, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 102, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 174, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 175, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 86, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 68, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 126, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 35, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 133, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 156, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 91, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 121, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 95, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 91, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 249, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 71, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 5, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 99, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 12, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 85, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 176, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 50, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 13, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 115, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 231, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 102, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 65, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 88, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 143, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 61, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 162, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 63, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 23, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 83, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 88, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 91, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 86, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 225, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 38, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 57, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 136, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 106, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 85, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 89, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 137, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 62, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 89, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 247, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 87, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 61, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 56, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 2, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 89, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 144, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 153, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 244, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 137, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 227, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 90, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 167, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 5, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 80, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 108, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 148, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 69, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 102, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 63, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 53, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 172, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 138, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 183, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 219, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 90, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 66, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 78, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 100, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 63, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 250, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 232, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 22, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 157, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 53, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 199, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 223, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 164, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 26, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 239, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 125, 121, 0, 5, 251, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 184, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 124, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 35, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 64, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 194, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 93, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 28, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 119, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 172, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 94, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 137, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 22, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 81, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 174, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 41, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 13, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 14, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 172, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 219, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 84, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 23, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 202, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 207, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 71, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 219, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 250, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 100, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 90, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 148, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 128, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 182, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 14, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 23, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 101, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 175, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 218, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 158, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 70, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 96, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 90, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 248, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 112, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 250, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 251, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 82, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 194, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 142, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 251, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 79, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 148, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 91, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 29, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 236, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 19, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 14, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 220, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 17, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 65, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 102, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 117, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 196, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 224, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 116, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 219, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 36, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 81, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 174, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 72, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 233, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 252, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 213, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 144, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 97, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 100, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 61, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 92, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 38, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 86, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 197, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 66, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 220, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 54, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 141, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 248, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 221, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 18, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 159, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 226, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 173, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 12, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 208, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 106, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 65, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 234, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 103, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 252, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 113, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 104, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 210, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 66, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 124, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 24, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 173, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 147, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 228, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 144, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 180, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 91, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 145, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 146, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 161, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 173, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 40, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 42, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 39, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 155, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 149, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 56, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 58, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 143, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 64, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 225, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 109, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 45, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 8, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 175, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 38, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 39, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 198, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 242, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 253, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 37, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 99, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 37, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 176, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 174, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 184, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 3, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 129, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 130, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 251, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 116, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 252, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 98, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 188, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 107, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 67, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 56, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 174, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 114, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 160, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 246, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 39, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 195, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 254, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 19, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 25, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 189, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 65, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 147, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 29, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 183, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 58, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 128, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 32, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 125, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 212, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 37, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 162, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 126, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 226, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 26, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 226, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 89, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 87, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 107, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 196, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 137, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 6, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 199, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 108, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 102, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 68, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 47, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 17, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 197, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 24, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 147, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 141, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 64, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 83, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 230, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 234, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 64, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 231, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 185, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 86, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 126, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 30, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 235, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 243, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 189, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 73, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 57, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 203, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 18, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 88, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 85, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 240, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 15, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 7, 127, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 175, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 24, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 228, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 67, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 24, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 40, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 118, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 183, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 86, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 164, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 133, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 239, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 168, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 186, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 150, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 98, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 48, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 69, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 23, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 85, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 176, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 65, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 94, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 134, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 24, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 10, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 105, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 174, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 84, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 127, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 91, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 114, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 66, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 65, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 252, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 18, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 138, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 255, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 108, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 214, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 41, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 47, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 87, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 149, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 147, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 149, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 228, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 47, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 109, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 131, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 177, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 64, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 19, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 190, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 56, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 90, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 236, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 54, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 160, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 142, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 179, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 54, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 166, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 197, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 127, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 133, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 27, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 81, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 193, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 100, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 20, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 5, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 64, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 50, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 149, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 77, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 167, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 87, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 25, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 173, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 55, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 106, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 40, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 231, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 140, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 2, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 65, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 21, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 184, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 103, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 249, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 233, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 195, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 227, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 177, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 18, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 45, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 26, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 106, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 213, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 54, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 68, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 232, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 178, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 171, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 244, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 16, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 39, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 85, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 169, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 206, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 184, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 104, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 203, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 175, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 106, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 198, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 185, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 170, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 150, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 245, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 141, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 87, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 151, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 62, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 31, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 57, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 233, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 194, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 119, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 68, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 190, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 110, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 77, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 80, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 20, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 107, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 138, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 175, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 65, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 12, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 93, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 229, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 249, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 81, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 178, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 70, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 7, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 8, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 101, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 227, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 19, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 85, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 34, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 55, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 152, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 20, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 38, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 113, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 58, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 187, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 13, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 23, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 24, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 21, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 176, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 75, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 227, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 9, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 56, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 44, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 144, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 108, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 213, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 162, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 8, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 58, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 156, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 180, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 115, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 43, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 237, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 69, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 232, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 116, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 109, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 51, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 115, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 146, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 233, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 82, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 186, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 76, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 212, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 139, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 134, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 58, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 18, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 127, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 116, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 14, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 15, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 13, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 123, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 143, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 23, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 199, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 74, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 86, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 87, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 8, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 117, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 244, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 225, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 128, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 221, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 117, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 116, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 80, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 81, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 146, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 88, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 2, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 31, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 154, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 10, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 228, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 199, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 134, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 179, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 40, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 140, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 32, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 102, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 196, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 21, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 93, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 22, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 81, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 66, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 176, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 103, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 54, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 59, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 102, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 34, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 49, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 50, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 62, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 38, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 181, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 18, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 109, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 53, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 171, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 117, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 51, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 210, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 33, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 25, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 139, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 66, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 27, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 161, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 223, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 129, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 201, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 92, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 29, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 176, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 81, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 67, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 128, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 11, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 155, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 214, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 188, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 12, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 89, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 35, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 101, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 132, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 23, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 103, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 145, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 133, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 26, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 133, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 243, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 120, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 146, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 3, 127, 0, 0, 141, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 217, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 17, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 24, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 156, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 26, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 200, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 245, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 226, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 86, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 249, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 154, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 218, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 104, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 141, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 48, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 95, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 96, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 110, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 130, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 200, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 15, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 124, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 212, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 174, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 57, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 95, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 244, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 68, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 107, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 9, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 238, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 211, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 75, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 103, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 195, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 30, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 221, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 235, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 49, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 52, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 8, 127, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 39, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 148, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 19, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 110, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 125, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 3, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 25, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 148, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 191, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 184, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 7, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 115, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 126, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 88, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 33, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 34, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 200, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 15, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 20, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 148, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 27, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 55, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 187, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 230, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 34, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 177, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 88, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 87, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 239, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 148, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 37, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 255, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 128, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 82, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 185, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 11, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 35, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 197, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 132, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 57, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 110, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 177, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 12, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 162, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 82, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 127, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 95, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 96, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 227, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 13, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 88, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 120, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 102, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 186, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 88, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 28, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 58, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 24, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 97, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 200, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 115, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 228, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 69, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 78, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 10, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 163, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 145, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 12, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 38, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 229, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 46, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 20, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 149, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 35, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 213, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 233, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 241, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 135, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 70, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 14, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 44, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 16, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 250, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 121, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 131, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 200, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 66, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 199, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 105, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 202, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 67, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 82, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 21, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 167, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 132, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 132, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 198, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 91, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 101, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 234, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 148, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 150, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 102, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 101, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 72, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 14, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 89, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 2, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 125, 121, 0, 5, 254, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 2, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 75, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 128, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 54, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 76, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 38, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 2, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 125, 121, 0, 5, 255, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 42, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 205, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 118, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 3, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 48, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 133, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 78, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 166, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 244, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 10, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 91, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 111, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 69, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 70, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 24, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 229, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 164, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 15, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 21, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 231, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 60, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 66, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 90, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 15, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 3, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 83, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 178, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 79, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 141, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 82, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 32, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 48, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 179, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 232, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 153, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 106, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 107, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 18, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 199, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 62, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 129, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 201, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 11, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 177, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 25, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 61, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 25, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 28, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 135, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 26, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 45, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 126, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 92, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 12, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 92, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 43, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 52, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 59, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 211, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 240, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 70, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 40, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 165, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 188, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 163, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 153, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 142, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 12, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 154, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 251, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 4, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 150, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 73, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 117, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 236, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 3, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 103, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 122, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 108, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 22, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 39, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 253, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 70, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 123, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 151, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 163, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 72, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 77, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 32, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 89, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 126, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 147, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 229, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 41, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 228, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 224, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 68, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 230, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 45, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 177, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 146, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 215, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 180, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 17, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 138, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 116, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 90, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 135, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 44, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 149, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 230, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 92, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 203, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 214, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 92, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 218, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 41, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 36, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 93, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 252, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 59, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 178, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 92, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 182, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 164, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 215, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 219, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 201, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 20, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 25, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 62, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 69, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 221, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 189, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 29, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 13, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 30, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 192, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 214, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 81, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 70, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 44, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 95, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 165, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 93, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 124, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 73, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 253, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 222, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 109, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 46, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 33, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 130, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 7, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 4, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 14, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 110, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 71, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 208, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 203, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 91, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 27, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 52, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 200, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 209, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 40, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 3, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 38, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 41, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 5, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 109, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 151, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 87, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 14, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 91, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 67, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 16, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 117, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 149, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 165, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 4, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 66, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 60, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 38, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 13, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 86, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 71, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 71, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 22, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 136, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 232, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 6, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 84, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 14, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 181, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 131, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 98, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 124, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 40, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 118, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 157, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 4, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 63, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 129, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 168, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 15, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 5, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 150, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 81, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 29, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 175, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 215, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 193, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 97, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 175, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 245, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 129, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 111, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 201, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 69, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 51, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 137, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 182, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 77, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 92, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 62, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 67, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 6, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 9, 127, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 94, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 246, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 72, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 106, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 85, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 143, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 94, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 122, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 204, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 117, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 7, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 30, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 133, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 212, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 64, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 19, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 23, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 30, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 82, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 36, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 71, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 102, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 138, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 202, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 52, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 82, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 118, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 169, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 53, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 22, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 39, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 124, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 170, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 112, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 158, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 19, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 14, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 72, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 41, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 237, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 182, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 238, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 139, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 31, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 222, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 170, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 36, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 192, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 3, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 40, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 47, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 139, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 72, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 204, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 250, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 23, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 41, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 39, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 123, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 144, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 59, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 37, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 63, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 234, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 103, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 83, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 42, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 99, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 224, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 81, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 62, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 213, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 119, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 247, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 20, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 248, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 151, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 37, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 127, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 133, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 58, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 53, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 104, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 233, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 42, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 43, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 95, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 24, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 99, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 37, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 103, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 119, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 100, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 86, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 39, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 5, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 196, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 60, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 93, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 118, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 42, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 31, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 246, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 195, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 162, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 99, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 234, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 14, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 24, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 64, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 190, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 122, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 123, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 63, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 22, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 76, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 201, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 237, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 64, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 32, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 104, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 56, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 225, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 98, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 8, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 125, 121, 0, 5, 252, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 202, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 244, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 250, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 131, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 120, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 241, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 8, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 6, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 110, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 63, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 183, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 69, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 64, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 234, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 201, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 222, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 65, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 180, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 10, 127, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 117, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 231, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 121, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 24, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 233, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 93, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 125, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 111, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 232, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 96, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 80, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 33, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 25, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 8, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 140, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 177, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 190, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 65, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 149, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 131, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 229, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 202, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 141, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 166, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 71, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 52, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 171, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 216, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 26, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 29, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 26, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 16, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 43, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 152, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 172, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 94, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 149, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 96, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 103, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 46, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 88, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 70, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 145, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 10, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 172, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 247, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 65, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 235, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 20, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 68, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 5, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 95, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 20, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 191, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 118, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 67, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 228, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 92, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 134, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 87, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 173, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 18, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 200, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 18, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 49, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 172, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 113, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 202, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 17, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 202, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 233, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 142, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 72, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 146, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 38, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 21, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 31, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 11, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 35, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 159, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 203, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 63, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 73, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 19, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 119, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 192, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 101, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 68, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 215, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 65, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 228, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 28, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 44, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 197, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 43, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 15, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 27, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 246, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 231, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 143, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 184, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 82, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 2, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 153, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 203, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 18, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 83, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 193, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 44, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 69, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 105, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 16, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 134, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 134, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 246, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 8, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 204, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 59, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 9, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 119, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 174, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 9, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 39, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 166, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 167, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 82, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 130, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 148, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 41, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 226, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 56, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 58, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 102, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 251, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 42, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 161, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 100, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 66, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 67, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 73, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 99, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 17, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 40, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 18, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 6, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 193, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 68, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 191, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 128, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 234, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 26, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 134, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 155, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 74, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 3, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 93, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 23, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 73, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 112, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 75, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 213, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 226, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 19, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 90, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 14, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 125, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 120, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 21, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 86, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 13, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 222, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 119, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 14, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 23, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 8, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 43, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 9, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 147, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 193, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 201, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 160, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 120, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 53, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 22, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 103, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 37, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 22, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 62, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 111, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 86, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 97, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 176, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 110, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 179, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 89, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 203, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 18, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 220, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 208, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 65, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 138, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 247, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 74, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 41, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 41, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 101, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 59, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 93, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 88, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 46, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 54, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 107, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 214, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 154, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 150, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 44, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 8, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 124, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 205, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 105, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 154, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 139, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 174, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 89, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 204, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 66, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 40, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 194, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 19, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 15, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 180, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 183, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 155, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 19, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 195, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 179, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 45, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 103, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 173, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 53, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 47, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 69, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 30, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 204, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 75, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 193, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 216, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 147, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 181, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 229, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 194, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 126, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 187, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 74, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 156, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 253, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 46, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 95, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 67, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 41, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 96, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 36, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 163, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 48, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 104, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 87, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 189, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 242, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 98, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 234, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 126, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 248, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 2, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 175, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 10, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 57, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 140, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 92, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 98, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 99, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 9, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 223, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 239, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 244, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 47, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 171, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 78, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 183, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 139, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 93, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 28, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 185, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 236, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 96, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 118, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 220, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 253, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 234, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 205, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 217, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 143, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 235, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 125, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 131, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 215, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 206, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 172, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 245, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 95, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 79, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 142, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 42, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 15, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 227, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 83, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 24, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 74, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 157, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 168, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 240, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 157, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 205, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 132, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 68, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 33, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 10, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 11, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 216, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 68, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 16, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 83, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 84, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 70, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 15, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 108, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 107, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 133, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 239, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 43, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 25, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 164, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 82, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 235, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 194, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 15, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 99, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 113, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 155, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 176, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 81, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 246, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 186, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 125, 121, 0, 5, 2, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 177, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 72, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 73, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 254, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 134, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 139, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 205, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 103, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 203, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 69, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 142, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 188, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 140, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 120, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 195, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 135, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 109, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 128, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 22, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 221, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 86, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 143, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 6, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 73, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 16, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 192, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 193, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 40, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 84, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 152, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 161, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 144, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 21, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 10, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 162, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 253, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 112, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 254, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 44, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 232, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 148, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 112, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 9, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 11, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 80, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 204, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 148, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 74, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 34, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 177, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 58, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 73, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 104, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 203, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 237, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 101, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 19, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 94, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 223, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 238, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 206, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 55, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 176, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 223, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 5, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 15, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 121, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 58, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 150, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 196, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 45, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 134, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 83, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 229, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 6, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 57, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 40, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 114, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 214, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 197, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 198, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 236, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 73, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 155, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 75, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 81, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 20, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 21, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 20, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 74, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 25, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 162, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 91, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 149, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 113, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 16, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 243, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 12, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 72, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 198, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 149, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 199, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 26, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 48, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 165, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 33, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 207, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 57, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 15, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 74, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 122, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 123, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 166, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 156, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 39, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 230, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 90, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 16, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 178, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 4, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 69, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 200, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 41, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 93, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 47, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 239, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 129, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 89, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 63, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 94, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 207, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 56, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 90, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 211, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 235, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 132, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 236, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 17, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 61, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 101, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 235, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 120, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 145, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 245, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 58, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 91, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 77, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 65, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 151, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 48, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 70, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 105, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 49, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 181, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 167, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 19, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 230, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 173, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 106, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 54, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 136, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 104, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 54, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 137, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 242, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 22, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 193, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 42, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 47, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 92, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 42, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 100, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 48, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 27, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 21, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 90, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 49, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 230, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 93, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 194, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 208, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 4, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 64, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 166, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 23, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 163, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 190, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 140, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 75, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 7, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 177, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 202, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 24, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 5, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 8, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 203, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 204, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 184, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 2, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 95, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 237, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 20, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 247, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 231, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 40, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 50, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 96, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 65, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 157, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 201, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 194, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 72, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 85, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 135, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 70, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 255, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 112, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 34, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 121, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 92, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 135, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 130, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 38, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 97, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 152, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 97, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 7, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 94, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 27, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 3, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 113, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 127, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 249, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 11, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 208, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 107, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 93, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 204, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 140, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 70, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 45, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 101, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 244, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 71, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 42, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 18, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 134, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 83, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 166, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 74, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 68, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 87, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 3, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 4, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 205, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 26, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 224, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 206, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 75, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 76, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 142, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 168, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 83, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 38, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 173, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 97, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 108, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 174, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 90, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 19, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 23, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 96, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 113, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 175, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 69, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 42, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 185, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 31, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 39, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 106, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 41, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 16, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 70, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 195, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 30, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 114, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 43, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 26, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 72, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 196, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 132, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 9, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 153, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 78, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 4, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 43, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 109, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 217, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 218, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 25, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 197, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 225, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 94, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 196, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 206, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 4, 127, 0, 0, 141, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 98, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 20, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 199, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 195, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 42, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 121, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 44, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 202, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 86, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 125, 121, 0, 5, 55, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 43, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 153, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 124, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 29, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 212, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 13, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 75, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 150, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 10, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 158, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 97, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 179, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 9, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 206, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 117, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 89, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 69, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 202, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 156, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 26, 127, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 10, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 12, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 203, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 6, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 11, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 21, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 186, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 72, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 165, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 131, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 184, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 7, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 2, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 82, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 54, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 13, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 145, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 95, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 78, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 238, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 22, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 197, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 135, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 16, 127, 0, 0, 121, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 99, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 100, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 209, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 67, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 143, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 94, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 210, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 224, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 72, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 149, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 217, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 21, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 221, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 236, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 205, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 237, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 45, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 144, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 157, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 60, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 228, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 120, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 131, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 107, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 237, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 98, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 217, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 27, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 101, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 60, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 10, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 125, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 87, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 73, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 211, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 185, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 87, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 60, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 203, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 167, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 135, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 245, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 189, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 245, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 110, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 106, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 102, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 95, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 215, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 191, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 57, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 58, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 19, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 158, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 75, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 166, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 110, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 5, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 136, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 248, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 194, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 207, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 60, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 204, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 11, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 20, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 104, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 64, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 9, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 207, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 196, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 111, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 88, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 141, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 17, 127, 0, 0, 135, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 39, 127, 0, 0, 125, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 17, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 216, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 159, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 120, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 42, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 23, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 49, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 125, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 29, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 213, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 122, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 14, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 204, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 121, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 246, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 22, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 88, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 57, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 32, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 39, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 96, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 190, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 94, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 3, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 45, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 28, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 108, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 207, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 43, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 118, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 44, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 16, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 205, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 152, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 182, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 12, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 178, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 99, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 4, 127, 0, 0, 113, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 32, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 254, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 24, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 12, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 125, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 82, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 206, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 137, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 26, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 240, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 255, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 212, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 143, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 212, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 44, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 198, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 149, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 196, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 169, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 27, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 222, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 215, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 111, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 22, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 136, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 223, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 150, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 101, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 13, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 158, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 174, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 20, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 55, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 89, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 49, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 66, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 98, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 115, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 180, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 105, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 23, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 74, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 205, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 20, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 42, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 62, 127, 0, 0, 211, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 58, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 136, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 4, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 126, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 203, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 221, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 28, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 116, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 127, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 71, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 62, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 134, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 155, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 40, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 7, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 183, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 163, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 218, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 25, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 190, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 156, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 167, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 26, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 23, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 96, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 191, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 195, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 197, 123, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 30, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 27, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 241, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 50, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 64, 127, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 97, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 180, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 36, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 46, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 208, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 98, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 13, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 159, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 46, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 103, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 199, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 35, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 61, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 70, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 21, 127, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 83, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 142, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 122, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 150, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 112, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 5, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 137, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 31, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 121, 121, 0, 5, 168, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 209, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 9, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 123, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 106, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 100, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 111, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 230, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 141, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 28, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 245, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 142, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 137, 121, 0, 5, 112, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 14, 127, 0, 0, 117, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 102, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 231, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 187, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 133, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 216, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 29, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 105, 121, 0, 5, 206, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 40, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 179, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 54, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 50, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 174, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 214, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 219, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 220, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 41, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 48, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 69, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 30, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 63, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 138, 127, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 187, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 55, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 151, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 21, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 29, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 58, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 119, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 201, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 10, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 154, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 46, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 231, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 11, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 123, 126, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 22, 127, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 57, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 192, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 67, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 18, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 144, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 131, 121, 0, 5, 219, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 48, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 184, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 91, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 213, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 197, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 38, 121, 0, 5, 24, 126, 0, 0, 103, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 151, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 51, 124, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 10, 121, 0, 5, 167, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 246, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 12, 121, 0, 5, 196, 126, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 212, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 49, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 129, 121, 0, 5, 58, 125, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 94, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 30, 127, 0, 0, 137, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 238, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 83, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 56, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 84, 124, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 169, 126, 0, 0, 143, 121, 0, 5, 0, 5, 0, 0, 42, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 84, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 40, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 30, 121, 0, 5, 41, 127, 0, 0, 105, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 143, 121, 0, 5, 136, 122, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 139, 121, 0, 5, 168, 123, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 104, 125, 0, 0, 119, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 115, 121, 0, 5, 93, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 11, 127, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 24, 121, 0, 5, 0, 5, 0, 0, 111, 121, 0, 5, 167, 126, 0, 0, 193, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 127, 127, 0, 0, 127, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 100, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 192, 126, 0, 0, 185, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 15, 127, 0, 0, 111, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 143, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 24, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 13, 126, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 61, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 76, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 20, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 233, 125, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 16, 121, 0, 5, 0, 5, 0, 0, 127, 121, 0, 5, 62, 126, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 119, 121, 0, 5, 54, 127, 0, 0, 219, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 0, 5, 0, 0, 103, 121, 0, 5, 85, 127, 0, 0, 209, 121, 0, 5, 0, 5, 0, 0, 6, 121, 0, 5, 160, 127, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 34, 121, 0, 5, 12, 126, 0, 0, 129, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 0, 5, 0, 0, 107, 121, 0, 5, 242, 125, 0, 0, 179, 121, 0, 5, 0, 5, 0, 0, 28, 121, 0, 5, 31, 127, 0, 0, 137, 121, 0, 0, 0, 0, 5, 5, 42, 121, 5, 70, 103, 121, 5, 5, 6, 121, 5, 5, 119, 121, 5, 70, 221, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 103, 185, 121, 5, 5, 28, 121, 5, 101, 117, 121, 5, 5, 30, 121, 5, 123, 143, 121, 5, 5, 42, 121, 5, 70, 119, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 111, 211, 121, 5, 5, 28, 121, 5, 70, 107, 121, 5, 5, 28, 121, 5, 80, 137, 121, 5, 5, 28, 121, 5, 70, 115, 121, 5, 5, 42, 121, 5, 71, 119, 121, 5, 5, 42, 121, 5, 70, 105, 121, 5, 5, 38, 121, 5, 122, 103, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 80, 193, 121, 5, 5, 42, 121, 5, 70, 137, 121, 5, 5, 30, 121, 5, 85, 111, 121, 5, 5, 28, 121, 5, 97, 135, 121, 5, 5, 30, 121, 5, 86, 111, 121, 5, 5, 30, 121, 5, 91, 121, 121, 5, 5, 42, 121, 5, 71, 103, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 120, 193, 121, 5, 5, 28, 121, 5, 71, 115, 121, 5, 5, 40, 121, 5, 99, 105, 121, 5, 5, 42, 121, 5, 90, 125, 121, 5, 5, 6, 121, 5, 92, 133, 121, 5, 5, 34, 121, 5, 81, 143, 121, 5, 5, 28, 121, 5, 81, 137, 121, 5, 5, 28, 121, 5, 101, 141, 121, 5, 5, 42, 121, 5, 70, 129, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 111, 209, 121, 5, 5, 16, 121, 5, 84, 115, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 78, 211, 121, 5, 5, 42, 121, 5, 84, 121, 121, 5, 5, 40, 121, 5, 91, 127, 121, 5, 5, 42, 121, 5, 71, 129, 121, 5, 5, 28, 121, 5, 102, 141, 121, 5, 5, 34, 121, 5, 82, 143, 121, 5, 5, 28, 121, 5, 102, 117, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 121, 179, 121, 5, 5, 28, 121, 5, 73, 123, 121, 5, 5, 28, 121, 5, 71, 107, 121, 5, 5, 28, 121, 5, 98, 135, 121, 5, 5, 28, 121, 5, 123, 125, 121, 5, 5, 42, 121, 5, 71, 105, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 113, 211, 121, 5, 5, 38, 121, 5, 94, 129, 121, 5, 5, 42, 121, 5, 79, 117, 121, 5, 5, 42, 121, 5, 83, 127, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 117, 211, 121, 5, 5, 34, 121, 5, 106, 125, 121, 5, 5, 40, 121, 5, 105, 127, 121, 5, 5, 42, 121, 5, 70, 141, 121, 5, 5, 28, 121, 5, 99, 135, 121, 5, 5, 28, 121, 5, 118, 127, 121, 5, 5, 28, 121, 5, 103, 141, 121, 5, 5, 28, 121, 5, 82, 137, 121, 5, 5, 30, 121, 5, 113, 113, 121, 5, 5, 28, 121, 5, 120, 141, 121, 5, 5, 16, 121, 5, 102, 115, 121, 5, 5, 42, 121, 5, 111, 119, 121, 5, 5, 30, 121, 5, 87, 111, 121, 5, 5, 16, 121, 5, 85, 115, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 71, 219, 121, 5, 5, 38, 121, 5, 115, 105, 121, 5, 5, 28, 121, 5, 83, 137, 121, 5, 5, 42, 121, 5, 70, 117, 121, 5, 5, 28, 121, 5, 126, 137, 121, 5, 5, 34, 121, 5, 94, 135, 121, 5, 5, 42, 121, 5, 112, 119, 121, 5, 5, 28, 121, 5, 72, 107, 121, 5, 5, 34, 121, 5, 76, 111, 121, 5, 5, 28, 121, 5, 121, 141, 121, 5, 5, 42, 121, 5, 71, 141, 121, 5, 5, 6, 121, 5, 93, 133, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 85, 219, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 72, 219, 121, 5, 5, 28, 121, 5, 119, 127, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 80, 209, 121, 5, 5, 28, 121, 5, 103, 117, 121, 5, 5, 30, 121, 5, 90, 113, 121, 5, 5, 24, 121, 5, 78, 123, 121, 5, 5, 34, 121, 5, 70, 113, 121, 5, 5, 30, 121, 5, 96, 121, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 107, 193, 121, 5, 5, 40, 121, 5, 92, 127, 121, 5, 5, 6, 121, 5, 94, 133, 121, 5, 5, 42, 121, 5, 5, 125, 121, 5, 71, 179, 121, 5, 5, 30, 121, 5, 91, 113, 121, 5, 5, 28, 121, 5, 122, 141, 121, 5, 5, 42, 121, 5, 5, 105, 121, 5, 94, 179, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 87, 185, 121, 5, 5, 40, 121, 5, 106, 127, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 88, 185, 121, 5, 5, 16, 121, 5, 86, 115, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 73, 219, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 104, 211, 121, 5, 5, 42, 121, 5, 70, 121, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 90, 211, 121, 5, 5, 42, 121, 5, 70, 125, 121, 5, 5, 6, 121, 5, 95, 133, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 110, 193, 121, 5, 5, 42, 121, 5, 70, 133, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 122, 179, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 79, 211, 121, 5, 5, 6, 121, 5, 85, 123, 121, 5, 5, 34, 121, 5, 105, 143, 121, 5, 5, 28, 121, 5, 100, 135, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 88, 193, 121, 5, 5, 42, 121, 5, 72, 129, 121, 5, 5, 42, 121, 5, 72, 103, 121, 5, 5, 16, 121, 5, 103, 115, 121, 5, 5, 34, 121, 5, 80, 135, 121, 5, 5, 28, 121, 5, 104, 117, 121, 5, 5, 30, 121, 5, 124, 143, 121, 5, 5, 38, 121, 5, 116, 105, 121, 5, 5, 40, 121, 5, 108, 103, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 91, 211, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 76, 193, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 97, 211, 121, 5, 5, 42, 121, 5, 73, 129, 121, 5, 5, 40, 121, 5, 87, 117, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 81, 209, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 70, 211, 121, 5, 5, 34, 121, 5, 81, 135, 121, 5, 5, 42, 121, 5, 74, 129, 121, 5, 5, 42, 121, 5, 83, 129, 121, 5, 5, 42, 121, 5, 84, 103, 121, 5, 5, 16, 121, 5, 87, 115, 121, 5, 5, 42, 121, 5, 72, 119, 121, 5, 5, 30, 121, 5, 88, 111, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 121, 193, 121, 5, 5, 42, 121, 5, 71, 121, 121, 5, 5, 42, 121, 5, 88, 105, 121, 5, 5, 38, 121, 5, 123, 103, 121, 5, 5, 42, 121, 5, 90, 141, 121, 5, 5, 30, 121, 5, 125, 143, 121, 5, 5, 42, 121, 5, 72, 105, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 82, 209, 121, 5, 5, 42, 121, 5, 75, 137, 121, 5, 5, 6, 121, 5, 88, 123, 121, 5, 5, 40, 121, 5, 5, 129, 121, 5, 97, 209, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 116, 219, 121, 5, 5, 34, 121, 5, 106, 143, 121, 5, 5, 28, 121, 5, 108, 121, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 98, 211, 121, 5, 5, 42, 121, 5, 70, 127, 121, 5, 5, 40, 121, 5, 100, 105, 121, 5, 5, 42, 121, 5, 91, 141, 121, 5, 5, 28, 121, 5, 84, 137, 121, 5, 5, 16, 121, 5, 104, 115, 121, 5, 5, 38, 121, 5, 124, 103, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 123, 179, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 83, 209, 121, 5, 5, 30, 121, 5, 92, 113, 121, 5, 5, 42, 121, 5, 75, 129, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 84, 209, 121, 5, 5, 42, 121, 5, 71, 133, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 104, 185, 121, 5, 5, 42, 121, 5, 72, 141, 121, 5, 5, 28, 121, 5, 122, 135, 121, 5, 5, 42, 121, 5, 84, 129, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 109, 219, 121, 5, 5, 42, 121, 5, 5, 131, 121, 5, 80, 185, 121, 5, 5, 28, 121, 5, 85, 137, 121, 5, 5, 42, 121, 5, 76, 129, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 105, 211, 121, 5, 5, 34, 121, 5, 83, 143, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 108, 209, 121, 5, 5, 40, 121, 5, 107, 127, 121, 5, 5, 42, 121, 5, 80, 117, 121, 5, 5, 42, 121, 5, 85, 121, 121, 5, 5, 34, 121, 5, 109, 125, 121, 5, 5, 34, 121, 5, 82, 135, 121, 5, 5, 42, 121, 5, 113, 119, 121, 5, 5, 42, 121, 5, 92, 141, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 74, 211, 121, 5, 5, 42, 121, 5, 70, 111, 121, 5, 5, 42, 121, 5, 73, 141, 121, 5, 5, 36, 121, 5, 71, 123, 121, 5, 5, 42, 121, 5, 5, 125, 121, 5, 73, 179, 121, 5, 5, 42, 121, 5, 71, 127, 121, 5, 5, 42, 121, 5, 85, 103, 121, 5, 5, 34, 121, 5, 79, 113, 121, 5, 5, 42, 121, 5, 84, 127, 121, 5, 5, 28, 121, 5, 121, 117, 121, 5, 5, 42, 121, 5, 74, 141, 121, 5, 5, 42, 121, 5, 71, 125, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 99, 193, 121, 5, 5, 28, 121, 5, 109, 121, 121, 5, 5, 28, 121, 5, 101, 135, 121, 5, 5, 30, 121, 5, 92, 121, 121, 5, 5, 40, 121, 5, 92, 103, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 124, 179, 121, 5, 5, 42, 121, 5, 73, 105, 121, 5, 5, 42, 121, 5, 85, 129, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 92, 211, 121, 5, 5, 28, 121, 5, 80, 107, 121, 5, 5, 38, 121, 5, 99, 125, 121, 5, 5, 34, 121, 5, 5, 111, 121, 5, 121, 211, 121, 5, 5, 42, 121, 5, 86, 129, 121, 5, 5, 30, 121, 5, 93, 113, 121, 5, 5, 28, 121, 5, 120, 127, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 118, 211, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 114, 211, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 89, 185, 121, 5, 5, 34, 121, 5, 103, 129, 121, 5, 5, 42, 121, 5, 71, 111, 121, 5, 5, 38, 121, 5, 125, 103, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 90, 185, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 98, 179, 121, 5, 5, 42, 121, 5, 73, 119, 121, 5, 5, 42, 121, 5, 74, 119, 121, 5, 5, 42, 121, 5, 73, 103, 121, 5, 5, 28, 121, 5, 118, 125, 121, 5, 5, 28, 121, 5, 73, 107, 121, 5, 5, 38, 121, 5, 117, 105, 121, 5, 5, 36, 121, 5, 72, 123, 121, 5, 5, 28, 121, 5, 121, 127, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 80, 211, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 105, 185, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 86, 179, 121, 5, 5, 38, 121, 5, 95, 129, 121, 5, 5, 42, 121, 5, 74, 105, 121, 5, 5, 10, 121, 5, 113, 115, 121, 5, 5, 42, 121, 5, 75, 119, 121, 5, 5, 28, 121, 5, 72, 115, 121, 5, 5, 40, 121, 5, 109, 103, 121, 5, 5, 38, 121, 5, 100, 129, 121, 5, 5, 28, 121, 5, 122, 127, 121, 5, 5, 38, 121, 5, 96, 129, 121, 5, 5, 34, 121, 5, 77, 111, 121, 5, 5, 34, 121, 5, 5, 111, 121, 5, 122, 211, 121, 5, 5, 30, 121, 5, 89, 111, 121, 5, 5, 28, 121, 5, 102, 135, 121, 5, 5, 28, 121, 5, 123, 127, 121, 5, 5, 42, 121, 5, 75, 141, 121, 5, 5, 28, 121, 5, 103, 135, 121, 5, 5, 28, 121, 5, 74, 123, 121, 5, 5, 30, 121, 5, 114, 113, 121, 5, 5, 34, 121, 5, 83, 135, 121, 5, 5, 40, 121, 5, 93, 103, 121, 5, 5, 28, 121, 5, 110, 121, 121, 5, 5, 42, 121, 5, 86, 121, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 110, 219, 121, 5, 5, 28, 121, 5, 77, 123, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 74, 209, 121, 5, 5, 40, 121, 5, 108, 127, 121, 5, 5, 28, 121, 5, 122, 117, 121, 5, 5, 42, 121, 5, 114, 119, 121, 5, 5, 42, 121, 5, 115, 119, 121, 5, 5, 42, 121, 5, 72, 121, 121, 5, 5, 42, 121, 5, 76, 141, 121, 5, 5, 40, 121, 5, 88, 117, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 100, 193, 121, 5, 5, 40, 121, 5, 110, 105, 121, 5, 5, 42, 121, 5, 72, 127, 121, 5, 5, 42, 121, 5, 75, 105, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 99, 179, 121, 5, 5, 28, 121, 5, 86, 137, 121, 5, 5, 30, 121, 5, 113, 111, 121, 5, 5, 42, 121, 5, 5, 125, 121, 5, 79, 219, 121, 5, 5, 28, 121, 5, 104, 141, 121, 5, 5, 28, 121, 5, 87, 137, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 106, 185, 121, 5, 5, 42, 121, 5, 76, 105, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 120, 209, 121, 5, 5, 34, 121, 5, 107, 143, 121, 5, 5, 6, 121, 5, 117, 121, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 112, 209, 121, 5, 5, 28, 121, 5, 124, 127, 121, 5, 5, 42, 121, 5, 81, 117, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 120, 185, 121, 5, 5, 28, 121, 5, 104, 135, 121, 5, 5, 34, 121, 5, 84, 135, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 77, 193, 121, 5, 5, 42, 121, 5, 86, 103, 121, 5, 5, 40, 121, 5, 110, 103, 121, 5, 5, 30, 121, 5, 114, 111, 121, 5, 5, 42, 121, 5, 76, 119, 121, 5, 5, 34, 121, 5, 84, 143, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 81, 211, 121, 5, 5, 34, 121, 5, 107, 125, 121, 5, 5, 42, 121, 5, 85, 127, 121, 5, 5, 28, 121, 5, 124, 125, 121, 5, 5, 28, 121, 5, 125, 125, 121, 5, 5, 28, 121, 5, 119, 125, 121, 5, 5, 28, 121, 5, 125, 127, 121, 5, 5, 42, 121, 5, 5, 111, 121, 5, 77, 209, 121, 5, 5, 42, 121, 5, 89, 105, 121, 5, 5, 28, 121, 5, 123, 141, 121, 5, 5, 28, 121, 5, 120, 125, 121, 5, 5, 30, 121, 5, 93, 121, 121, 5, 5, 40, 121, 5, 94, 103, 121, 5, 5, 34, 121, 5, 108, 143, 121, 5, 5, 42, 121, 5, 77, 141, 121, 5, 5, 34, 121, 5, 5, 111, 121, 5, 123, 211, 121, 5, 5, 30, 121, 5, 94, 113, 121, 5, 5, 42, 121, 5, 77, 129, 121, 5, 5, 28, 121, 5, 123, 135, 121, 5, 5, 40, 121, 5, 5, 119, 121, 5, 110, 179, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 86, 219, 121, 5, 5, 34, 121, 5, 85, 143, 121, 5, 5, 40, 121, 5, 89, 117, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 107, 185, 121, 5, 5, 28, 121, 5, 88, 137, 121, 5, 5, 28, 121, 5, 89, 137, 121, 5, 5, 30, 121, 5, 90, 111, 121, 5, 5, 30, 121, 5, 94, 121, 121, 5, 5, 28, 121, 5, 90, 137, 121, 5, 5, 42, 121, 5, 74, 103, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 73, 209, 121, 5, 5, 40, 121, 5, 90, 117, 121, 5, 5, 42, 121, 5, 90, 105, 121, 5, 5, 16, 121, 5, 88, 115, 121, 5, 5, 42, 121, 5, 72, 125, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 90, 179, 121, 5, 5, 40, 121, 5, 91, 117, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 117, 219, 121, 5, 5, 30, 121, 5, 95, 113, 121, 5, 5, 42, 121, 5, 77, 119, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 111, 193, 121, 5, 5, 42, 121, 5, 71, 117, 121, 5, 5, 34, 121, 5, 80, 113, 121, 5, 5, 42, 121, 5, 70, 135, 121, 5, 5, 40, 121, 5, 93, 127, 121, 5, 5, 40, 121, 5, 70, 143, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 74, 185, 121, 5, 5, 42, 121, 5, 73, 125, 121, 5, 5, 42, 121, 5, 78, 129, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 87, 219, 121, 5, 5, 42, 121, 5, 82, 117, 121, 5, 5, 42, 121, 5, 71, 135, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 112, 193, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 109, 209, 121, 5, 5, 38, 121, 5, 5, 139, 121, 5, 118, 179, 121, 5, 5, 36, 121, 5, 70, 123, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 121, 209, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 75, 185, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 96, 193, 121, 5, 5, 40, 121, 5, 111, 103, 121, 5, 5, 42, 121, 5, 116, 119, 121, 5, 5, 38, 121, 5, 118, 105, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 111, 219, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 93, 211, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 88, 219, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 76, 219, 121, 5, 5, 42, 121, 5, 74, 125, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 85, 193, 121, 5, 5, 34, 121, 5, 86, 143, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 71, 211, 121, 5, 5, 30, 121, 5, 96, 113, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 94, 211, 121, 5, 5, 42, 121, 5, 75, 125, 121, 5, 5, 42, 121, 5, 93, 141, 121, 5, 5, 40, 121, 5, 101, 105, 121, 5, 5, 28, 121, 5, 91, 137, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 108, 185, 121, 5, 5, 34, 121, 5, 77, 133, 121, 5, 5, 34, 121, 5, 78, 111, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 83, 179, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 97, 193, 121, 5, 5, 28, 121, 5, 92, 137, 121, 5, 5, 42, 121, 5, 72, 117, 121, 5, 5, 24, 121, 5, 82, 123, 121, 5, 5, 34, 121, 5, 81, 133, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 89, 219, 121, 5, 5, 28, 121, 5, 74, 107, 121, 5, 5, 28, 121, 5, 93, 137, 121, 5, 5, 28, 121, 5, 94, 137, 121, 5, 5, 40, 121, 5, 5, 107, 121, 5, 115, 179, 121, 5, 5, 28, 121, 5, 105, 141, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 95, 211, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 90, 219, 121, 5, 5, 38, 121, 5, 119, 105, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 91, 219, 121, 5, 5, 40, 121, 5, 94, 127, 121, 5, 5, 38, 121, 5, 5, 139, 121, 5, 116, 179, 121, 5, 5, 42, 121, 5, 73, 117, 121, 5, 5, 16, 121, 5, 89, 115, 121, 5, 5, 28, 121, 5, 105, 135, 121, 5, 5, 6, 121, 5, 96, 133, 121, 5, 5, 42, 121, 5, 79, 141, 121, 5, 5, 42, 121, 5, 78, 141, 121, 5, 5, 42, 121, 5, 77, 105, 121, 5, 5, 42, 121, 5, 87, 103, 121, 5, 5, 42, 121, 5, 76, 125, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 121, 185, 121, 5, 5, 28, 121, 5, 106, 141, 121, 5, 5, 6, 121, 5, 118, 121, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 84, 179, 121, 5, 5, 42, 121, 5, 80, 141, 121, 5, 5, 42, 121, 5, 94, 141, 121, 5, 5, 42, 121, 5, 78, 119, 121, 5, 5, 16, 121, 5, 90, 115, 121, 5, 5, 42, 121, 5, 79, 119, 121, 5, 5, 40, 121, 5, 95, 103, 121, 5, 5, 38, 121, 5, 97, 129, 121, 5, 5, 40, 121, 5, 71, 143, 121, 5, 5, 30, 121, 5, 91, 111, 121, 5, 5, 34, 121, 5, 104, 129, 121, 5, 5, 28, 121, 5, 126, 127, 121, 5, 5, 28, 121, 5, 123, 117, 121, 5, 5, 6, 121, 5, 89, 123, 121, 5, 5, 30, 121, 5, 95, 121, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 82, 211, 121, 5, 5, 34, 121, 5, 5, 111, 121, 5, 124, 211, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 76, 185, 121, 5, 5, 6, 121, 5, 86, 123, 121, 5, 5, 28, 121, 5, 124, 117, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 104, 209, 121, 5, 5, 34, 121, 5, 105, 129, 121, 5, 5, 28, 121, 5, 95, 137, 121, 5, 5, 30, 121, 5, 97, 113, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 109, 185, 121, 5, 5, 34, 121, 5, 82, 133, 121, 5, 5, 42, 121, 5, 72, 135, 121, 5, 5, 28, 121, 5, 75, 107, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 119, 211, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 115, 211, 121, 5, 5, 42, 121, 5, 71, 137, 121, 5, 5, 34, 121, 5, 110, 125, 121, 5, 5, 40, 121, 5, 109, 127, 121, 5, 5, 6, 121, 5, 97, 133, 121, 5, 5, 42, 121, 5, 78, 135, 121, 5, 5, 40, 121, 5, 96, 103, 121, 5, 5, 38, 121, 5, 5, 105, 121, 5, 119, 179, 121, 5, 5, 42, 121, 5, 76, 137, 121, 5, 5, 30, 121, 5, 115, 113, 121, 5, 5, 28, 121, 5, 107, 141, 121, 5, 5, 42, 121, 5, 87, 121, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 74, 179, 121, 5, 5, 40, 121, 5, 97, 103, 121, 5, 5, 38, 121, 5, 75, 111, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 75, 179, 121, 5, 5, 28, 121, 5, 96, 137, 121, 5, 5, 42, 121, 5, 73, 127, 121, 5, 5, 40, 121, 5, 102, 105, 121, 5, 5, 40, 121, 5, 97, 117, 121, 5, 5, 40, 121, 5, 92, 117, 121, 5, 5, 42, 121, 5, 74, 127, 121, 5, 5, 16, 121, 5, 91, 115, 121, 5, 5, 40, 121, 5, 111, 105, 121, 5, 5, 42, 121, 5, 80, 119, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 70, 185, 121, 5, 5, 42, 121, 5, 75, 103, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 92, 219, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 93, 219, 121, 5, 5, 42, 121, 5, 81, 141, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 111, 185, 121, 5, 5, 42, 121, 5, 77, 125, 121, 5, 5, 42, 121, 5, 81, 119, 121, 5, 5, 42, 121, 5, 5, 131, 121, 5, 81, 185, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 96, 185, 121, 5, 5, 42, 121, 5, 73, 135, 121, 5, 5, 28, 121, 5, 106, 135, 121, 5, 5, 42, 121, 5, 74, 117, 121, 5, 5, 40, 121, 5, 5, 119, 121, 5, 111, 179, 121, 5, 5, 42, 121, 5, 95, 141, 121, 5, 5, 42, 121, 5, 75, 127, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 97, 185, 121, 5, 5, 40, 121, 5, 5, 119, 121, 5, 112, 179, 121, 5, 5, 42, 121, 5, 82, 141, 121, 5, 5, 24, 121, 5, 83, 123, 121, 5, 5, 28, 121, 5, 105, 117, 121, 5, 5, 34, 121, 5, 71, 113, 121, 5, 5, 34, 121, 5, 108, 125, 121, 5, 5, 42, 121, 5, 78, 125, 121, 5, 5, 28, 121, 5, 97, 137, 121, 5, 5, 6, 121, 5, 86, 133, 121, 5, 5, 42, 121, 5, 79, 129, 121, 5, 5, 40, 121, 5, 95, 127, 121, 5, 5, 30, 121, 5, 92, 111, 121, 5, 5, 40, 121, 5, 98, 103, 121, 5, 5, 30, 121, 5, 115, 111, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 113, 209, 121, 5, 5, 28, 121, 5, 125, 117, 121, 5, 5, 28, 121, 5, 111, 121, 121, 5, 5, 6, 121, 5, 119, 121, 121, 5, 5, 28, 121, 5, 98, 137, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 106, 211, 121, 5, 5, 28, 121, 5, 99, 137, 121, 5, 5, 30, 121, 5, 116, 111, 121, 5, 5, 42, 121, 5, 91, 125, 121, 5, 5, 42, 121, 5, 5, 105, 121, 5, 95, 179, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 112, 185, 121, 5, 5, 34, 121, 5, 87, 143, 121, 5, 5, 42, 121, 5, 76, 127, 121, 5, 5, 16, 121, 5, 105, 115, 121, 5, 5, 30, 121, 5, 98, 113, 121, 5, 5, 34, 121, 5, 81, 113, 121, 5, 5, 28, 121, 5, 108, 141, 121, 5, 5, 34, 121, 5, 106, 129, 121, 5, 5, 28, 121, 5, 76, 107, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 122, 209, 121, 5, 5, 28, 121, 5, 100, 137, 121, 5, 5, 28, 121, 5, 101, 137, 121, 5, 5, 30, 121, 5, 93, 111, 121, 5, 5, 42, 121, 5, 78, 105, 121, 5, 5, 34, 121, 5, 107, 129, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 94, 219, 121, 5, 5, 34, 121, 5, 111, 125, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 116, 211, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 107, 211, 121, 5, 5, 38, 121, 5, 100, 125, 121, 5, 5, 42, 121, 5, 83, 117, 121, 5, 5, 30, 121, 5, 94, 111, 121, 5, 5, 40, 121, 5, 96, 127, 121, 5, 5, 34, 121, 5, 108, 129, 121, 5, 5, 42, 121, 5, 73, 121, 121, 5, 5, 6, 121, 5, 98, 133, 121, 5, 5, 34, 121, 5, 95, 135, 121, 5, 5, 42, 121, 5, 5, 125, 121, 5, 80, 219, 121, 5, 5, 42, 121, 5, 79, 125, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 85, 209, 121, 5, 5, 6, 121, 5, 87, 133, 121, 5, 5, 16, 121, 5, 92, 115, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 70, 209, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 71, 185, 121, 5, 5, 42, 121, 5, 91, 105, 121, 5, 5, 42, 121, 5, 96, 141, 121, 5, 5, 28, 121, 5, 124, 141, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 89, 193, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 71, 209, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 72, 209, 121, 5, 5, 42, 121, 5, 5, 111, 121, 5, 106, 193, 121, 5, 5, 42, 121, 5, 74, 111, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 75, 211, 121, 5, 5, 28, 121, 5, 73, 115, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 110, 185, 121, 5, 5, 28, 121, 5, 75, 123, 121, 5, 5, 38, 121, 5, 120, 105, 121, 5, 5, 40, 121, 5, 110, 127, 121, 5, 5, 34, 121, 5, 82, 113, 121, 5, 5, 42, 121, 5, 86, 127, 121, 5, 5, 42, 121, 5, 72, 133, 121, 5, 5, 28, 121, 5, 109, 141, 121, 5, 5, 42, 121, 5, 82, 119, 121, 5, 5, 34, 121, 5, 96, 135, 121, 5, 5, 30, 121, 5, 117, 111, 121, 5, 5, 6, 121, 5, 99, 133, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 81, 193, 121, 5, 5, 28, 121, 5, 74, 115, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 118, 219, 121, 5, 5, 28, 121, 5, 106, 117, 121, 5, 5, 38, 121, 5, 121, 105, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 114, 209, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 123, 209, 121, 5, 5, 42, 121, 5, 117, 119, 121, 5, 5, 40, 121, 5, 103, 105, 121, 5, 5, 30, 121, 5, 95, 111, 121, 5, 5, 42, 121, 5, 76, 103, 121, 5, 5, 28, 121, 5, 102, 137, 121, 5, 5, 34, 121, 5, 88, 143, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 95, 219, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 113, 193, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 119, 219, 121, 5, 5, 40, 121, 5, 99, 103, 121, 5, 5, 38, 121, 5, 122, 105, 121, 5, 5, 42, 121, 5, 92, 125, 121, 5, 5, 28, 121, 5, 103, 137, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 120, 219, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 91, 179, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 74, 219, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 90, 193, 121, 5, 5, 28, 121, 5, 112, 121, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 72, 211, 121, 5, 5, 40, 121, 5, 100, 103, 121, 5, 5, 42, 121, 5, 83, 119, 121, 5, 5, 42, 121, 5, 79, 105, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 115, 209, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 83, 211, 121, 5, 5, 40, 121, 5, 104, 105, 121, 5, 5, 42, 121, 5, 77, 127, 121, 5, 5, 34, 121, 5, 72, 113, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 86, 209, 121, 5, 5, 42, 121, 5, 87, 129, 121, 5, 5, 34, 121, 5, 89, 143, 121, 5, 5, 42, 121, 5, 78, 127, 121, 5, 5, 42, 121, 5, 84, 119, 121, 5, 5, 34, 121, 5, 80, 111, 121, 5, 5, 42, 121, 5, 80, 125, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 113, 185, 121, 5, 5, 30, 121, 5, 96, 111, 121, 5, 5, 38, 121, 5, 101, 129, 121, 5, 5, 28, 121, 5, 100, 121, 121, 5, 5, 28, 121, 5, 126, 125, 121, 5, 5, 28, 121, 5, 107, 135, 121, 5, 5, 28, 121, 5, 104, 137, 121, 5, 5, 42, 121, 5, 85, 119, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 96, 219, 121, 5, 5, 34, 121, 5, 109, 129, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 108, 211, 121, 5, 5, 42, 121, 5, 118, 119, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 97, 219, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 91, 193, 121, 5, 5, 42, 121, 5, 93, 125, 121, 5, 5, 42, 121, 5, 119, 119, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 122, 193, 121, 5, 5, 34, 121, 5, 73, 113, 121, 5, 5, 42, 121, 5, 86, 119, 121, 5, 5, 40, 121, 5, 97, 127, 121, 5, 5, 28, 121, 5, 125, 141, 121, 5, 5, 34, 121, 5, 112, 125, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 76, 211, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 98, 219, 121, 5, 5, 6, 121, 5, 88, 133, 121, 5, 5, 30, 121, 5, 118, 111, 121, 5, 5, 42, 121, 5, 5, 125, 121, 5, 82, 219, 121, 5, 5, 42, 121, 5, 87, 119, 121, 5, 5, 38, 121, 5, 5, 105, 121, 5, 120, 179, 121, 5, 5, 42, 121, 5, 81, 125, 121, 5, 5, 28, 121, 5, 107, 117, 121, 5, 5, 30, 121, 5, 99, 113, 121, 5, 5, 42, 121, 5, 88, 119, 121, 5, 5, 28, 121, 5, 105, 137, 121, 5, 5, 42, 121, 5, 89, 119, 121, 5, 5, 16, 121, 5, 93, 115, 121, 5, 5, 40, 121, 5, 5, 119, 121, 5, 113, 179, 121, 5, 5, 42, 121, 5, 80, 105, 121, 5, 5, 30, 121, 5, 119, 111, 121, 5, 5, 24, 121, 5, 79, 123, 121, 5, 5, 42, 121, 5, 90, 119, 121, 5, 5, 40, 121, 5, 112, 103, 121, 5, 5, 42, 121, 5, 74, 121, 121, 5, 5, 42, 121, 5, 82, 125, 121, 5, 5, 34, 121, 5, 85, 135, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 72, 185, 121, 5, 5, 42, 121, 5, 72, 137, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 121, 219, 121, 5, 5, 42, 121, 5, 5, 131, 121, 5, 84, 185, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 91, 185, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 77, 185, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 92, 179, 121, 5, 5, 42, 121, 5, 88, 103, 121, 5, 5, 42, 121, 5, 74, 135, 121, 5, 5, 42, 121, 5, 5, 131, 121, 5, 82, 185, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 99, 219, 121, 5, 5, 28, 121, 5, 124, 135, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 114, 185, 121, 5, 5, 42, 121, 5, 80, 129, 121, 5, 5, 42, 121, 5, 87, 127, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 92, 185, 121, 5, 5, 42, 121, 5, 83, 141, 121, 5, 5, 40, 121, 5, 111, 127, 121, 5, 5, 28, 121, 5, 125, 135, 121, 5, 5, 42, 121, 5, 84, 141, 121, 5, 5, 42, 121, 5, 85, 141, 121, 5, 5, 34, 121, 5, 90, 143, 121, 5, 5, 42, 121, 5, 73, 133, 121, 5, 5, 42, 121, 5, 97, 141, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 93, 185, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 87, 179, 121, 5, 5, 40, 121, 5, 5, 119, 121, 5, 114, 179, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 98, 185, 121, 5, 5, 42, 121, 5, 98, 141, 121, 5, 5, 40, 121, 5, 101, 103, 121, 5, 5, 28, 121, 5, 108, 135, 121, 5, 5, 40, 121, 5, 113, 103, 121, 5, 5, 28, 121, 5, 77, 107, 121, 5, 5, 42, 121, 5, 79, 127, 121, 5, 5, 40, 121, 5, 105, 105, 121, 5, 5, 34, 121, 5, 5, 111, 121, 5, 125, 211, 121, 5, 5, 30, 121, 5, 120, 111, 121, 5, 5, 38, 121, 5, 5, 139, 121, 5, 117, 179, 121, 5, 5, 42, 121, 5, 120, 119, 121, 5, 5, 42, 121, 5, 86, 141, 121, 5, 5, 42, 121, 5, 5, 143, 121, 5, 71, 193, 121, 5, 5, 42, 121, 5, 91, 119, 121, 5, 5, 30, 121, 5, 97, 111, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 103, 179, 121, 5, 5, 40, 121, 5, 98, 117, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 84, 211, 121, 5, 5, 40, 121, 5, 106, 105, 121, 5, 5, 28, 121, 5, 108, 117, 121, 5, 5, 28, 121, 5, 126, 141, 121, 5, 5, 30, 121, 5, 116, 113, 121, 5, 5, 42, 121, 5, 121, 119, 121, 5, 5, 28, 121, 5, 121, 125, 121, 5, 5, 28, 121, 5, 113, 121, 121, 5, 5, 30, 121, 5, 98, 111, 121, 5, 5, 42, 121, 5, 88, 129, 121, 5, 5, 28, 121, 5, 106, 137, 121, 5, 5, 28, 121, 5, 107, 137, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 92, 193, 121, 5, 5, 40, 121, 5, 99, 117, 121, 5, 5, 28, 121, 5, 126, 117, 121, 5, 5, 42, 121, 5, 83, 125, 121, 5, 5, 42, 121, 5, 5, 111, 121, 5, 78, 209, 121, 5, 5, 42, 121, 5, 92, 105, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 99, 185, 121, 5, 5, 42, 121, 5, 5, 125, 121, 5, 72, 179, 121, 5, 5, 42, 121, 5, 5, 111, 121, 5, 79, 209, 121, 5, 5, 40, 121, 5, 114, 103, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 114, 193, 121, 5, 5, 42, 121, 5, 81, 129, 121, 5, 5, 40, 121, 5, 112, 105, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 116, 209, 121, 5, 5, 42, 121, 5, 92, 119, 121, 5, 5, 40, 121, 5, 102, 103, 121, 5, 5, 30, 121, 5, 117, 113, 121, 5, 5, 28, 121, 5, 109, 135, 121, 5, 5, 42, 121, 5, 93, 105, 121, 5, 5, 42, 121, 5, 77, 103, 121, 5, 5, 42, 121, 5, 93, 119, 121, 5, 5, 28, 121, 5, 108, 137, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 112, 219, 121, 5, 5, 24, 121, 5, 84, 123, 121, 5, 5, 34, 121, 5, 113, 125, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 100, 219, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 115, 185, 121, 5, 5, 28, 121, 5, 75, 115, 121, 5, 5, 6, 121, 5, 120, 121, 121, 5, 5, 42, 121, 5, 94, 119, 121, 5, 5, 40, 121, 5, 98, 127, 121, 5, 5, 28, 121, 5, 101, 121, 121, 5, 5, 28, 121, 5, 109, 137, 121, 5, 5, 28, 121, 5, 122, 125, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 115, 193, 121, 5, 5, 34, 121, 5, 109, 143, 121, 5, 5, 42, 121, 5, 73, 137, 121, 5, 5, 42, 121, 5, 88, 121, 121, 5, 5, 42, 121, 5, 75, 121, 121, 5, 5, 34, 121, 5, 5, 111, 121, 5, 126, 211, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 116, 193, 121, 5, 5, 28, 121, 5, 110, 141, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 82, 193, 121, 5, 5, 34, 121, 5, 83, 133, 121, 5, 5, 40, 121, 5, 72, 143, 121, 5, 5, 30, 121, 5, 99, 111, 121, 5, 5, 42, 121, 5, 94, 105, 121, 5, 5, 34, 121, 5, 110, 143, 121, 5, 5, 34, 121, 5, 91, 143, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 73, 185, 121, 5, 5, 30, 121, 5, 97, 121, 121, 5, 5, 34, 121, 5, 92, 143, 121, 5, 5, 28, 121, 5, 126, 135, 121, 5, 5, 34, 121, 5, 84, 133, 121, 5, 5, 42, 121, 5, 89, 103, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 104, 179, 121, 5, 5, 40, 121, 5, 100, 117, 121, 5, 5, 28, 121, 5, 110, 137, 121, 5, 5, 16, 121, 5, 106, 115, 121, 5, 5, 34, 121, 5, 93, 143, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 101, 219, 121, 5, 5, 42, 121, 5, 95, 119, 121, 5, 5, 42, 121, 5, 88, 127, 121, 5, 5, 42, 121, 5, 122, 119, 121, 5, 5, 42, 121, 5, 123, 119, 121, 5, 5, 40, 121, 5, 115, 103, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 105, 179, 121, 5, 5, 40, 121, 5, 73, 143, 121, 5, 5, 34, 121, 5, 110, 129, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 109, 211, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 99, 211, 121, 5, 5, 42, 121, 5, 94, 125, 121, 5, 5, 42, 121, 5, 84, 117, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 105, 209, 121, 5, 5, 30, 121, 5, 118, 113, 121, 5, 5, 28, 121, 5, 109, 117, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 109, 193, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 79, 179, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 80, 179, 121, 5, 5, 28, 121, 5, 111, 141, 121, 5, 5, 28, 121, 5, 76, 123, 121, 5, 5, 34, 121, 5, 83, 113, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 117, 209, 121, 5, 5, 40, 121, 5, 5, 115, 121, 5, 101, 209, 121, 5, 5, 40, 121, 5, 103, 103, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 106, 179, 121, 5, 5, 24, 121, 5, 80, 123, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 76, 179, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 77, 179, 121, 5, 5, 34, 121, 5, 114, 125, 121, 5, 5, 28, 121, 5, 76, 115, 121, 5, 5, 16, 121, 5, 94, 115, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 81, 179, 121, 5, 5, 42, 121, 5, 96, 119, 121, 5, 5, 40, 121, 5, 99, 127, 121, 5, 5, 30, 121, 5, 100, 113, 121, 5, 5, 40, 121, 5, 5, 129, 121, 5, 99, 209, 121, 5, 5, 42, 121, 5, 76, 121, 121, 5, 5, 30, 121, 5, 119, 113, 121, 5, 5, 42, 121, 5, 87, 141, 121, 5, 5, 42, 121, 5, 77, 121, 121, 5, 5, 34, 121, 5, 111, 129, 121, 5, 5, 6, 121, 5, 121, 121, 121, 5, 5, 42, 121, 5, 88, 141, 121, 5, 5, 30, 121, 5, 120, 113, 121, 5, 5, 40, 121, 5, 107, 105, 121, 5, 5, 34, 121, 5, 94, 143, 121, 5, 5, 40, 121, 5, 5, 129, 121, 5, 98, 209, 121, 5, 5, 34, 121, 5, 95, 143, 121, 5, 5, 28, 121, 5, 110, 117, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 78, 179, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 101, 193, 121, 5, 5, 40, 121, 5, 5, 115, 121, 5, 102, 209, 121, 5, 5, 28, 121, 5, 114, 121, 121, 5, 5, 28, 121, 5, 102, 121, 121, 5, 5, 28, 121, 5, 103, 121, 121, 5, 5, 40, 121, 5, 116, 103, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 98, 193, 121, 5, 5, 38, 121, 5, 123, 105, 121, 5, 5, 30, 121, 5, 101, 113, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 117, 193, 121, 5, 5, 42, 121, 5, 124, 119, 121, 5, 5, 30, 121, 5, 100, 111, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 102, 219, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 96, 211, 121, 5, 5, 42, 121, 5, 89, 129, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 123, 193, 121, 5, 5, 6, 121, 5, 100, 133, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 103, 219, 121, 5, 5, 40, 121, 5, 117, 103, 121, 5, 5, 42, 121, 5, 97, 119, 121, 5, 5, 16, 121, 5, 107, 115, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 122, 219, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 78, 193, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 116, 185, 121, 5, 5, 30, 121, 5, 101, 111, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 104, 193, 121, 5, 5, 34, 121, 5, 111, 143, 121, 5, 5, 28, 121, 5, 111, 137, 121, 5, 5, 34, 121, 5, 96, 143, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 87, 209, 121, 5, 5, 34, 121, 5, 74, 113, 121, 5, 5, 28, 121, 5, 110, 135, 121, 5, 5, 30, 121, 5, 121, 113, 121, 5, 5, 34, 121, 5, 97, 143, 121, 5, 5, 34, 121, 5, 115, 125, 121, 5, 5, 42, 121, 5, 98, 119, 121, 5, 5, 40, 121, 5, 112, 127, 121, 5, 5, 6, 121, 5, 101, 133, 121, 5, 5, 42, 121, 5, 84, 125, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 100, 185, 121, 5, 5, 42, 121, 5, 5, 143, 121, 5, 72, 193, 121, 5, 5, 42, 121, 5, 90, 103, 121, 5, 5, 30, 121, 5, 121, 111, 121, 5, 5, 6, 121, 5, 90, 123, 121, 5, 5, 30, 121, 5, 117, 125, 121, 5, 5, 6, 121, 5, 87, 123, 121, 5, 5, 34, 121, 5, 98, 143, 121, 5, 5, 40, 121, 5, 104, 103, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 124, 193, 121, 5, 5, 30, 121, 5, 122, 111, 121, 5, 5, 28, 121, 5, 112, 141, 121, 5, 5, 42, 121, 5, 89, 141, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 73, 211, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 77, 211, 121, 5, 5, 34, 121, 5, 86, 135, 121, 5, 5, 42, 121, 5, 79, 135, 121, 5, 5, 42, 121, 5, 5, 105, 121, 5, 96, 179, 121, 5, 5, 40, 121, 5, 118, 103, 121, 5, 5, 28, 121, 5, 78, 107, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 106, 209, 121, 5, 5, 34, 121, 5, 87, 135, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 113, 219, 121, 5, 5, 42, 121, 5, 80, 127, 121, 5, 5, 40, 121, 5, 93, 117, 121, 5, 5, 28, 121, 5, 111, 135, 121, 5, 5, 38, 121, 5, 124, 105, 121, 5, 5, 42, 121, 5, 99, 119, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 77, 219, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 75, 219, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 85, 211, 121, 5, 5, 34, 121, 5, 88, 135, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 86, 211, 121, 5, 5, 34, 121, 5, 116, 125, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 109, 179, 121, 5, 5, 38, 121, 5, 101, 125, 121, 5, 5, 16, 121, 5, 108, 115, 121, 5, 5, 34, 121, 5, 78, 133, 121, 5, 5, 42, 121, 5, 85, 125, 121, 5, 5, 30, 121, 5, 102, 113, 121, 5, 5, 16, 121, 5, 109, 115, 121, 5, 5, 28, 121, 5, 104, 121, 121, 5, 5, 34, 121, 5, 89, 135, 121, 5, 5, 34, 121, 5, 99, 143, 121, 5, 5, 28, 121, 5, 112, 137, 121, 5, 5, 28, 121, 5, 83, 115, 121, 5, 5, 28, 121, 5, 77, 115, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 70, 219, 121, 5, 5, 28, 121, 5, 78, 115, 121, 5, 5, 28, 121, 5, 113, 141, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 88, 209, 121, 5, 5, 42, 121, 5, 74, 133, 121, 5, 5, 42, 121, 5, 100, 119, 121, 5, 5, 40, 121, 5, 105, 103, 121, 5, 5, 28, 121, 5, 111, 117, 121, 5, 5, 42, 121, 5, 78, 121, 121, 5, 5, 34, 121, 5, 112, 129, 121, 5, 5, 38, 121, 5, 125, 105, 121, 5, 5, 30, 121, 5, 102, 111, 121, 5, 5, 30, 121, 5, 123, 111, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 118, 193, 121, 5, 5, 42, 121, 5, 86, 125, 121, 5, 5, 28, 121, 5, 79, 115, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 104, 219, 121, 5, 5, 42, 121, 5, 78, 103, 121, 5, 5, 30, 121, 5, 98, 121, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 92, 209, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 87, 211, 121, 5, 5, 34, 121, 5, 112, 143, 121, 5, 5, 6, 121, 5, 122, 121, 121, 5, 5, 40, 121, 5, 119, 103, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 93, 209, 121, 5, 5, 42, 121, 5, 79, 121, 121, 5, 5, 30, 121, 5, 103, 111, 121, 5, 5, 34, 121, 5, 79, 133, 121, 5, 5, 34, 121, 5, 81, 111, 121, 5, 5, 28, 121, 5, 112, 135, 121, 5, 5, 34, 121, 5, 113, 129, 121, 5, 5, 42, 121, 5, 5, 131, 121, 5, 83, 185, 121, 5, 5, 28, 121, 5, 115, 121, 121, 5, 5, 28, 121, 5, 113, 137, 121, 5, 5, 30, 121, 5, 104, 111, 121, 5, 5, 28, 121, 5, 113, 135, 121, 5, 5, 40, 121, 5, 120, 103, 121, 5, 5, 42, 121, 5, 101, 119, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 101, 185, 121, 5, 5, 34, 121, 5, 84, 113, 121, 5, 5, 42, 121, 5, 102, 119, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 125, 193, 121, 5, 5, 34, 121, 5, 85, 113, 121, 5, 5, 40, 121, 5, 94, 117, 121, 5, 5, 28, 121, 5, 114, 135, 121, 5, 5, 42, 121, 5, 75, 117, 121, 5, 5, 34, 121, 5, 86, 113, 121, 5, 5, 28, 121, 5, 114, 141, 121, 5, 5, 34, 121, 5, 113, 143, 121, 5, 5, 42, 121, 5, 95, 105, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 78, 219, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 94, 185, 121, 5, 5, 30, 121, 5, 103, 113, 121, 5, 5, 16, 121, 5, 95, 115, 121, 5, 5, 28, 121, 5, 112, 117, 121, 5, 5, 30, 121, 5, 105, 111, 121, 5, 5, 28, 121, 5, 113, 117, 121, 5, 5, 42, 121, 5, 103, 119, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 100, 179, 121, 5, 5, 42, 121, 5, 89, 127, 121, 5, 5, 34, 121, 5, 79, 111, 121, 5, 5, 42, 121, 5, 72, 111, 121, 5, 5, 42, 121, 5, 104, 119, 121, 5, 5, 42, 121, 5, 74, 137, 121, 5, 5, 42, 121, 5, 75, 133, 121, 5, 5, 42, 121, 5, 76, 133, 121, 5, 5, 34, 121, 5, 100, 143, 121, 5, 5, 30, 121, 5, 124, 111, 121, 5, 5, 42, 121, 5, 87, 125, 121, 5, 5, 28, 121, 5, 105, 121, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 95, 193, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 100, 211, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 105, 219, 121, 5, 5, 28, 121, 5, 115, 135, 121, 5, 5, 42, 121, 5, 79, 103, 121, 5, 5, 42, 121, 5, 105, 119, 121, 5, 5, 28, 121, 5, 106, 121, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 120, 211, 121, 5, 5, 42, 121, 5, 81, 105, 121, 5, 5, 28, 121, 5, 115, 141, 121, 5, 5, 16, 121, 5, 110, 115, 121, 5, 5, 16, 121, 5, 96, 115, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 102, 193, 121, 5, 5, 28, 121, 5, 116, 135, 121, 5, 5, 28, 121, 5, 116, 141, 121, 5, 5, 40, 121, 5, 100, 127, 121, 5, 5, 30, 121, 5, 125, 111, 121, 5, 5, 40, 121, 5, 74, 143, 121, 5, 5, 28, 121, 5, 114, 117, 121, 5, 5, 28, 121, 5, 114, 137, 121, 5, 5, 40, 121, 5, 113, 127, 121, 5, 5, 30, 121, 5, 104, 113, 121, 5, 5, 34, 121, 5, 82, 111, 121, 5, 5, 28, 121, 5, 117, 135, 121, 5, 5, 30, 121, 5, 126, 111, 121, 5, 5, 38, 121, 5, 102, 125, 121, 5, 5, 34, 121, 5, 114, 143, 121, 5, 5, 42, 121, 5, 5, 143, 121, 5, 73, 193, 121, 5, 5, 40, 121, 5, 108, 105, 121, 5, 5, 42, 121, 5, 5, 105, 121, 5, 97, 179, 121, 5, 5, 34, 121, 5, 83, 111, 121, 5, 5, 42, 121, 5, 82, 105, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 95, 185, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 79, 193, 121, 5, 5, 28, 121, 5, 107, 121, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 78, 185, 121, 5, 5, 42, 121, 5, 73, 111, 121, 5, 5, 40, 121, 5, 77, 143, 121, 5, 5, 30, 121, 5, 106, 111, 121, 5, 5, 42, 121, 5, 90, 129, 121, 5, 5, 28, 121, 5, 115, 117, 121, 5, 5, 6, 121, 5, 89, 133, 121, 5, 5, 42, 121, 5, 5, 143, 121, 5, 70, 193, 121, 5, 5, 42, 121, 5, 80, 121, 121, 5, 5, 42, 121, 5, 83, 105, 121, 5, 5, 42, 121, 5, 95, 125, 121, 5, 5, 6, 121, 5, 91, 123, 121, 5, 5, 6, 121, 5, 123, 121, 121, 5, 5, 28, 121, 5, 115, 137, 121, 5, 5, 42, 121, 5, 88, 125, 121, 5, 5, 6, 121, 5, 124, 121, 121, 5, 5, 28, 121, 5, 117, 141, 121, 5, 5, 34, 121, 5, 114, 129, 121, 5, 5, 28, 121, 5, 116, 137, 121, 5, 5, 42, 121, 5, 5, 131, 121, 5, 85, 185, 121, 5, 5, 34, 121, 5, 75, 113, 121, 5, 5, 42, 121, 5, 84, 105, 121, 5, 5, 28, 121, 5, 117, 137, 121, 5, 5, 42, 121, 5, 75, 135, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 94, 209, 121, 5, 5, 30, 121, 5, 105, 113, 121, 5, 5, 42, 121, 5, 5, 131, 121, 5, 86, 185, 121, 5, 5, 28, 121, 5, 118, 135, 121, 5, 5, 42, 121, 5, 85, 117, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 101, 179, 121, 5, 5, 42, 121, 5, 81, 121, 121, 5, 5, 42, 121, 5, 91, 129, 121, 5, 5, 28, 121, 5, 116, 121, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 86, 193, 121, 5, 5, 28, 121, 5, 118, 141, 121, 5, 5, 42, 121, 5, 106, 119, 121, 5, 5, 28, 121, 5, 116, 117, 121, 5, 5, 42, 121, 5, 86, 117, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 93, 193, 121, 5, 5, 42, 121, 5, 99, 141, 121, 5, 5, 42, 121, 5, 107, 119, 121, 5, 5, 28, 121, 5, 117, 117, 121, 5, 5, 34, 121, 5, 115, 143, 121, 5, 5, 40, 121, 5, 101, 127, 121, 5, 5, 40, 121, 5, 109, 105, 121, 5, 5, 42, 121, 5, 82, 121, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 107, 209, 121, 5, 5, 30, 121, 5, 107, 111, 121, 5, 5, 40, 121, 5, 5, 115, 121, 5, 100, 209, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 114, 219, 121, 5, 5, 42, 121, 5, 80, 103, 121, 5, 5, 42, 121, 5, 89, 125, 121, 5, 5, 34, 121, 5, 80, 133, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 88, 179, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 105, 193, 121, 5, 5, 34, 121, 5, 90, 135, 121, 5, 5, 34, 121, 5, 115, 129, 121, 5, 5, 38, 121, 5, 126, 105, 121, 5, 5, 40, 121, 5, 106, 103, 121, 5, 5, 6, 121, 5, 125, 121, 121, 5, 5, 6, 121, 5, 102, 133, 121, 5, 5, 34, 121, 5, 116, 143, 121, 5, 5, 28, 121, 5, 118, 137, 121, 5, 5, 30, 121, 5, 122, 113, 121, 5, 5, 28, 121, 5, 119, 137, 121, 5, 5, 30, 121, 5, 106, 113, 121, 5, 5, 42, 121, 5, 76, 117, 121, 5, 5, 30, 121, 5, 123, 113, 121, 5, 5, 30, 121, 5, 108, 111, 121, 5, 5, 30, 121, 5, 124, 113, 121, 5, 5, 34, 121, 5, 87, 113, 121, 5, 5, 6, 121, 5, 90, 133, 121, 5, 5, 34, 121, 5, 101, 143, 121, 5, 5, 42, 121, 5, 76, 135, 121, 5, 5, 28, 121, 5, 80, 115, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 108, 193, 121, 5, 5, 28, 121, 5, 81, 115, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 95, 209, 121, 5, 5, 40, 121, 5, 78, 143, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 83, 193, 121, 5, 5, 42, 121, 5, 96, 125, 121, 5, 5, 34, 121, 5, 116, 129, 121, 5, 5, 38, 121, 5, 103, 125, 121, 5, 5, 42, 121, 5, 82, 129, 121, 5, 5, 38, 121, 5, 98, 129, 121, 5, 5, 30, 121, 5, 109, 111, 121, 5, 5, 34, 121, 5, 76, 113, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 112, 211, 121, 5, 5, 28, 121, 5, 121, 137, 121, 5, 5, 28, 121, 5, 120, 137, 121, 5, 5, 6, 121, 5, 126, 121, 121, 5, 5, 42, 121, 5, 81, 103, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 106, 219, 121, 5, 5, 28, 121, 5, 119, 135, 121, 5, 5, 34, 121, 5, 77, 113, 121, 5, 5, 28, 121, 5, 122, 137, 121, 5, 5, 40, 121, 5, 75, 143, 121, 5, 5, 42, 121, 5, 85, 105, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 117, 185, 121, 5, 5, 16, 121, 5, 111, 115, 121, 5, 5, 30, 121, 5, 110, 111, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 110, 211, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 107, 179, 121, 5, 5, 34, 121, 5, 84, 111, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 124, 209, 121, 5, 5, 34, 121, 5, 117, 129, 121, 5, 5, 40, 121, 5, 121, 103, 121, 5, 5, 28, 121, 5, 123, 137, 121, 5, 5, 34, 121, 5, 118, 129, 121, 5, 5, 38, 121, 5, 5, 103, 121, 5, 110, 209, 121, 5, 5, 42, 121, 5, 92, 129, 121, 5, 5, 42, 121, 5, 81, 127, 121, 5, 5, 42, 121, 5, 5, 119, 121, 5, 85, 179, 121, 5, 5, 34, 121, 5, 91, 135, 121, 5, 5, 30, 121, 5, 107, 113, 121, 5, 5, 34, 121, 5, 119, 129, 121, 5, 5, 42, 121, 5, 96, 105, 121, 5, 5, 28, 121, 5, 119, 141, 121, 5, 5, 42, 121, 5, 77, 117, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 79, 185, 121, 5, 5, 28, 121, 5, 124, 137, 121, 5, 5, 28, 121, 5, 79, 107, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 118, 209, 121, 5, 5, 42, 121, 5, 5, 125, 121, 5, 83, 219, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 89, 209, 121, 5, 5, 28, 121, 5, 118, 117, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 88, 211, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 125, 209, 121, 5, 5, 34, 121, 5, 120, 129, 121, 5, 5, 34, 121, 5, 121, 129, 121, 5, 5, 34, 121, 5, 117, 143, 121, 5, 5, 30, 121, 5, 125, 113, 121, 5, 5, 42, 121, 5, 91, 103, 121, 5, 5, 28, 121, 5, 125, 137, 121, 5, 5, 34, 121, 5, 5, 143, 121, 5, 119, 209, 121, 5, 5, 24, 121, 5, 81, 123, 121, 5, 5, 34, 121, 5, 122, 129, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 115, 219, 121, 5, 5, 42, 121, 5, 108, 119, 121, 5, 5, 40, 121, 5, 114, 127, 121, 5, 5, 42, 121, 5, 5, 125, 121, 5, 81, 219, 121, 5, 5, 42, 121, 5, 5, 129, 121, 5, 102, 185, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 82, 179, 121, 5, 5, 16, 121, 5, 112, 115, 121, 5, 5, 40, 121, 5, 115, 127, 121, 5, 5, 40, 121, 5, 95, 117, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 101, 211, 121, 5, 5, 16, 121, 5, 97, 115, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 94, 193, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 102, 211, 121, 5, 5, 38, 121, 5, 5, 139, 121, 5, 103, 209, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 96, 209, 121, 5, 5, 6, 121, 5, 103, 133, 121, 5, 5, 34, 121, 5, 118, 143, 121, 5, 5, 40, 121, 5, 76, 143, 121, 5, 5, 40, 121, 5, 96, 117, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 103, 211, 121, 5, 5, 30, 121, 5, 108, 113, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 90, 209, 121, 5, 5, 30, 121, 5, 126, 113, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 107, 219, 121, 5, 5, 30, 121, 5, 109, 113, 121, 5, 5, 38, 121, 5, 105, 125, 121, 5, 5, 42, 121, 5, 5, 111, 121, 5, 75, 209, 121, 5, 5, 42, 121, 5, 77, 137, 121, 5, 5, 34, 121, 5, 102, 143, 121, 5, 5, 30, 121, 5, 111, 111, 121, 5, 5, 34, 121, 5, 123, 129, 121, 5, 5, 40, 121, 5, 113, 105, 121, 5, 5, 28, 121, 5, 119, 117, 121, 5, 5, 42, 121, 5, 82, 103, 121, 5, 5, 40, 121, 5, 114, 105, 121, 5, 5, 30, 121, 5, 110, 113, 121, 5, 5, 34, 121, 5, 88, 113, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 89, 179, 121, 5, 5, 42, 121, 5, 83, 121, 121, 5, 5, 40, 121, 5, 79, 143, 121, 5, 5, 42, 121, 5, 78, 117, 121, 5, 5, 28, 121, 5, 120, 135, 121, 5, 5, 42, 121, 5, 109, 119, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 103, 193, 121, 5, 5, 28, 121, 5, 120, 117, 121, 5, 5, 40, 121, 5, 107, 103, 121, 5, 5, 42, 121, 5, 5, 143, 121, 5, 74, 193, 121, 5, 5, 42, 121, 5, 97, 105, 121, 5, 5, 42, 121, 5, 97, 125, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 89, 211, 121, 5, 5, 38, 121, 5, 104, 125, 121, 5, 5, 34, 121, 5, 85, 133, 121, 5, 5, 30, 121, 5, 111, 113, 121, 5, 5, 42, 121, 5, 110, 119, 121, 5, 5, 40, 121, 5, 103, 127, 121, 5, 5, 40, 121, 5, 102, 127, 121, 5, 5, 28, 121, 5, 82, 115, 121, 5, 5, 42, 121, 5, 90, 127, 121, 5, 5, 42, 121, 5, 93, 129, 121, 5, 5, 42, 121, 5, 100, 141, 121, 5, 5, 6, 121, 5, 104, 133, 121, 5, 5, 6, 121, 5, 91, 133, 121, 5, 5, 40, 121, 5, 5, 143, 121, 5, 119, 193, 121, 5, 5, 34, 121, 5, 103, 143, 121, 5, 5, 42, 121, 5, 86, 105, 121, 5, 5, 34, 121, 5, 124, 129, 121, 5, 5, 40, 121, 5, 104, 127, 121, 5, 5, 42, 121, 5, 89, 121, 121, 5, 5, 42, 121, 5, 82, 127, 121, 5, 5, 42, 121, 5, 5, 111, 121, 5, 76, 209, 121, 5, 5, 34, 121, 5, 92, 135, 121, 5, 5, 16, 121, 5, 98, 115, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 118, 185, 121, 5, 5, 16, 121, 5, 99, 115, 121, 5, 5, 38, 121, 5, 102, 129, 121, 5, 5, 42, 121, 5, 87, 105, 121, 5, 5, 34, 121, 5, 78, 113, 121, 5, 5, 40, 121, 5, 116, 127, 121, 5, 5, 34, 121, 5, 89, 113, 121, 5, 5, 40, 121, 5, 80, 143, 121, 5, 5, 42, 121, 5, 78, 137, 121, 5, 5, 30, 121, 5, 99, 121, 121, 5, 5, 42, 121, 5, 5, 115, 121, 5, 93, 179, 121, 5, 5, 38, 121, 5, 99, 129, 121, 5, 5, 28, 121, 5, 121, 135, 121, 5, 5, 42, 121, 5, 98, 105, 121, 5, 5, 34, 121, 5, 119, 143, 121, 5, 5, 34, 121, 5, 125, 129, 121, 5, 5, 42, 121, 5, 83, 103, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 119, 185, 121, 5, 5, 40, 121, 5, 117, 127, 121, 5, 5, 42, 121, 5, 98, 125, 121, 5, 5, 34, 121, 5, 120, 143, 121, 5, 5, 34, 121, 5, 121, 143, 121, 5, 5, 42, 121, 5, 79, 137, 121, 5, 5, 34, 121, 5, 122, 143, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 102, 179, 121, 5, 5, 34, 121, 5, 93, 135, 121, 5, 5, 42, 121, 5, 5, 137, 121, 5, 87, 193, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 108, 179, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 91, 209, 121, 5, 5, 6, 121, 5, 105, 133, 121, 5, 5, 16, 121, 5, 100, 115, 121, 5, 5, 42, 121, 5, 77, 135, 121, 5, 5, 42, 121, 5, 5, 121, 121, 5, 108, 219, 121, 5, 5, 42, 121, 5, 5, 125, 121, 5, 84, 219, 121, 5, 5, 16, 121, 5, 101, 115, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 70, 179, 121, 5, 5, 42, 121, 5, 5, 143, 121, 5, 75, 193, 121, 5, 5, 30, 121, 5, 112, 113, 121, 5, 5, 34, 121, 5, 104, 143, 121, 5, 5, 42, 121, 5, 5, 139, 121, 5, 84, 193, 121, 5, 5, 30, 121, 5, 112, 111, 121, 5, 5, 42, 121, 5, 90, 121, 121, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 67, 0, 58, 0, 32, 0, 19, 46, 0, 16, 58, 5, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 162, 38, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 72, 0, 0, 0, 72, 1, 0, 0, 168, 24, 0, 0, 168, 24, 0, 0, 176, 24, 0, 0, 176, 24, 0, 0, 244, 24, 0, 0, 244, 24, 0, 0, 82, 34, 0, 0, 138, 38, 0, 0, 18, 0, 0, 0, 17, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 39, 246, 247, 248, 249, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 109, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 100, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 92, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 5, 0, 5, 212, 55, 18, 254, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 0, 198, 1, 0, 0, 0, 1, 12, 128, 107, 2, 117, 152, 75, 1, 254, 158, 187, 0, 7, 249, 74, 0, 196, 249, 38, 0, 10, 250, 20, 0, 10, 250, 255, 255, 25, 253, 193, 42, 30, 250, 255, 255, 195, 252, 193, 136, 60, 250, 255, 255, 135, 251, 193, 8, 149, 250, 255, 255, 11, 252, 193, 192, 206, 250, 255, 255, 224, 253, 193, 118, 196, 249, 255, 255, 224, 253, 193, 12, 233, 249, 255, 255, 98, 253, 193, 98, 247, 249, 255, 255, 152, 252, 193, 118, 8, 250, 255, 255, 15, 253, 193, 118, 64, 249, 16, 0, 64, 249, 255, 255, 209, 253, 193, 178, 113, 249, 255, 255, 78, 253, 193, 128, 129, 249, 255, 255, 106, 251, 193, 100, 138, 249, 255, 255, 42, 251, 193, 8, 7, 249, 255, 255, 224, 253, 193, 118, 8, 249, 255, 255, 224, 253, 193, 118, 10, 249, 255, 255, 98, 253, 193, 242, 52, 249, 255, 255, 198, 252, 193, 84, 127, 159, 75, 0, 159, 159, 57, 0, 159, 159, 255, 255, 224, 253, 193, 218, 160, 159, 255, 255, 224, 253, 193, 236, 126, 216, 4, 0, 2, 249, 255, 255, 70, 253, 193, 218, 9, 0, 210, 221, 20, 0, 210, 221, 255, 255, 45, 253, 193, 112, 21, 222, 255, 255, 213, 253, 193, 94, 23, 222, 255, 255, 217, 253, 193, 36, 24, 222, 255, 255, 217, 253, 193, 222, 28, 222, 255, 255, 219, 253, 193, 240, 29, 220, 255, 255, 34, 251, 193, 200, 117, 220, 255, 255, 130, 251, 193, 144, 120, 220, 255, 255, 135, 251, 193, 8, 144, 220, 255, 255, 160, 251, 193, 102, 196, 221, 255, 255, 16, 253, 193, 52, 127, 159, 255, 255, 223, 253, 193, 110, 141, 159, 255, 255, 224, 253, 193, 12, 153, 159, 255, 255, 224, 253, 193, 96, 156, 159, 255, 255, 224, 253, 193, 118, 59, 159, 16, 0, 59, 159, 255, 255, 219, 253, 193, 240, 74, 159, 255, 255, 220, 253, 193, 176, 80, 159, 255, 255, 220, 253, 193, 236, 82, 159, 255, 255, 220, 253, 193, 248, 254, 158, 255, 255, 217, 253, 193, 222, 14, 159, 255, 255, 218, 253, 193, 6, 19, 159, 255, 255, 218, 253, 193, 62, 32, 159, 255, 255, 218, 253, 193, 220, 50, 155, 70, 0, 165, 158, 34, 0, 205, 158, 16, 0, 205, 158, 255, 255, 214, 253, 193, 124, 209, 158, 255, 255, 214, 253, 193, 252, 249, 158, 255, 255, 217, 253, 193, 36, 253, 158, 255, 255, 217, 253, 193, 62, 165, 158, 255, 255, 211, 253, 193, 56, 166, 158, 255, 255, 212, 253, 193, 196, 187, 158, 255, 255, 213, 253, 193, 94, 195, 158, 255, 255, 213, 253, 193, 210, 229, 156, 16, 0, 229, 156, 255, 255, 195, 253, 193, 218, 31, 158, 255, 255, 206, 253, 193, 72, 117, 158, 255, 255, 208, 253, 193, 240, 127, 158, 255, 255, 209, 253, 193, 178, 50, 155, 255, 255, 180, 253, 193, 172, 60, 155, 255, 255, 181, 253, 193, 166, 90, 155, 255, 255, 184, 253, 193, 2, 124, 156, 255, 255, 193, 253, 193, 64, 153, 153, 34, 0, 216, 154, 16, 0, 216, 154, 255, 255, 175, 253, 193, 232, 223, 154, 255, 255, 176, 253, 193, 186, 37, 155, 255, 255, 180, 253, 193, 58, 47, 155, 255, 255, 180, 253, 193, 136, 153, 153, 255, 255, 164, 253, 193, 252, 172, 153, 255, 255, 165, 253, 193, 208, 108, 154, 255, 255, 171, 253, 193, 244, 168, 154, 255, 255, 173, 253, 193, 168, 222, 152, 16, 0, 222, 152, 255, 255, 157, 253, 193, 164, 223, 152, 255, 255, 157, 253, 193, 170, 99, 153, 255, 255, 163, 253, 193, 20, 150, 153, 255, 255, 164, 253, 193, 110, 117, 152, 255, 255, 153, 253, 193, 70, 168, 152, 255, 255, 154, 253, 193, 82, 206, 152, 255, 255, 156, 253, 193, 242, 219, 152, 255, 255, 157, 253, 193, 122, 100, 141, 142, 0, 127, 149, 70, 0, 94, 151, 34, 0, 230, 151, 16, 0, 230, 151, 255, 255, 147, 253, 193, 30, 237, 151, 255, 255, 147, 253, 193, 108, 243, 151, 255, 255, 147, 253, 193, 184, 1, 152, 255, 255, 148, 253, 193, 142, 94, 151, 255, 255, 140, 253, 193, 44, 98, 151, 255, 255, 140, 253, 193, 152, 105, 151, 255, 255, 141, 253, 193, 160, 203, 151, 255, 255, 145, 253, 193, 196, 182, 150, 16, 0, 182, 150, 255, 255, 131, 253, 193, 186, 185, 150, 255, 255, 131, 253, 193, 228, 232, 150, 255, 255, 134, 253, 193, 178, 81, 151, 255, 255, 139, 253, 193, 216, 127, 149, 255, 255, 120, 253, 193, 2, 128, 149, 255, 255, 120, 253, 193, 6, 232, 149, 255, 255, 124, 253, 193, 84, 28, 150, 255, 255, 125, 253, 193, 154, 181, 143, 34, 0, 204, 145, 16, 0, 204, 145, 255, 255, 98, 253, 193, 98, 209, 145, 255, 255, 98, 253, 193, 242, 133, 148, 255, 255, 114, 253, 193, 86, 119, 149, 255, 255, 118, 253, 193, 200, 181, 143, 255, 255, 78, 253, 193, 176, 145, 144, 255, 255, 87, 253, 193, 98, 73, 145, 255, 255, 93, 253, 193, 148, 198, 145, 255, 255, 98, 253, 193, 30, 202, 142, 16, 0, 202, 142, 255, 255, 70, 253, 193, 218, 102, 143, 255, 255, 76, 253, 193, 100, 155, 143, 255, 255, 77, 253, 193, 186, 176, 143, 255, 255, 78, 253, 193, 128, 100, 141, 255, 255, 54, 253, 193, 190, 112, 141, 255, 255, 55, 253, 193, 84, 179, 141, 255, 255, 59, 253, 193, 84, 171, 142, 255, 255, 68, 253, 193, 230, 107, 134, 70, 0, 0, 138, 34, 0, 85, 140, 16, 0, 85, 140, 255, 255, 45, 253, 193, 112, 120, 140, 255, 255, 47, 253, 193, 36, 157, 140, 255, 255, 48, 253, 193, 190, 29, 141, 255, 255, 53, 253, 193, 202, 0, 138, 255, 255, 30, 253, 193, 38, 160, 139, 255, 255, 40, 253, 193, 192, 55, 140, 255, 255, 43, 253, 193, 160, 70, 140, 255, 255, 44, 253, 193, 90, 126, 137, 16, 0, 126, 137, 255, 255, 24, 253, 193, 136, 139, 137, 255, 255, 25, 253, 193, 42, 193, 137, 255, 255, 27, 253, 193, 148, 210, 137, 255, 255, 27, 253, 193, 240, 107, 134, 255, 255, 254, 252, 193, 150, 64, 136, 255, 255, 14, 253, 193, 202, 76, 136, 255, 255, 15, 253, 193, 118, 99, 136, 255, 255, 16, 253, 193, 52, 252, 129, 34, 0, 110, 130, 16, 0, 110, 130, 255, 255, 222, 252, 193, 106, 114, 130, 255, 255, 222, 252, 193, 140, 120, 130, 255, 255, 223, 252, 193, 6, 77, 134, 255, 255, 252, 252, 193, 174, 252, 129, 255, 255, 217, 252, 193, 10, 12, 130, 255, 255, 218, 252, 193, 106, 27, 130, 255, 255, 219, 252, 193, 90, 31, 130, 255, 255, 219, 252, 193, 128, 137, 128, 16, 0, 137, 128, 255, 255, 203, 252, 193, 226, 227, 129, 255, 255, 215, 252, 193, 90, 234, 129, 255, 255, 215, 252, 193, 222, 243, 129, 255, 255, 216, 252, 193, 152, 12, 128, 255, 255, 198, 252, 193, 214, 18, 128, 255, 255, 199, 252, 193, 40, 51, 128, 255, 255, 200, 252, 193, 74, 127, 128, 255, 255, 203, 252, 193, 128, 54, 98, 30, 1, 132, 115, 142, 0, 226, 119, 70, 0, 115, 124, 34, 0, 81, 127, 16, 0, 81, 127, 255, 255, 189, 252, 193, 238, 138, 127, 255, 255, 193, 252, 193, 14, 189, 127, 255, 255, 195, 252, 193, 136, 1, 128, 255, 255, 198, 252, 193, 84, 115, 124, 255, 255, 168, 252, 193, 76, 248, 124, 255, 255, 174, 252, 193, 70, 159, 126, 255, 255, 185, 252, 193, 220, 54, 127, 255, 255, 188, 252, 193, 198, 190, 121, 16, 0, 190, 121, 255, 255, 140, 252, 193, 156, 116, 122, 255, 255, 148, 252, 193, 22, 203, 122, 255, 255, 152, 252, 193, 118, 249, 122, 255, 255, 154, 252, 193, 154, 226, 119, 255, 255, 124, 252, 193, 16, 243, 119, 255, 255, 125, 252, 193, 134, 58, 121, 255, 255, 134, 252, 193, 210, 184, 121, 255, 255, 140, 252, 193, 88, 139, 117, 34, 0, 174, 118, 16, 0, 174, 118, 255, 255, 109, 252, 193, 26, 191, 118, 255, 255, 110, 252, 193, 184, 238, 118, 255, 255, 113, 252, 193, 128, 219, 119, 255, 255, 123, 252, 193, 54, 139, 117, 255, 255, 98, 252, 193, 100, 146, 117, 255, 255, 98, 252, 193, 176, 118, 118, 255, 255, 106, 252, 193, 156, 125, 118, 255, 255, 106, 252, 193, 250, 24, 117, 16, 0, 24, 117, 255, 255, 92, 252, 193, 188, 31, 117, 255, 255, 93, 252, 193, 36, 40, 117, 255, 255, 93, 252, 193, 208, 48, 117, 255, 255, 94, 252, 193, 32, 132, 115, 255, 255, 79, 252, 193, 112, 137, 115, 255, 255, 79, 252, 193, 132, 220, 116, 255, 255, 89, 252, 193, 158, 230, 116, 255, 255, 90, 252, 193, 104, 179, 107, 70, 0, 42, 114, 34, 0, 71, 114, 16, 0, 71, 114, 255, 255, 66, 252, 193, 252, 89, 114, 255, 255, 68, 252, 193, 28, 91, 114, 255, 255, 68, 252, 193, 94, 172, 114, 255, 255, 72, 252, 193, 46, 42, 114, 255, 255, 64, 252, 193, 48, 54, 114, 255, 255, 65, 252, 193, 150, 59, 114, 255, 255, 65, 252, 193, 218, 63, 114, 255, 255, 66, 252, 193, 14, 15, 108, 16, 0, 15, 108, 255, 255, 20, 252, 193, 236, 20, 108, 255, 255, 21, 252, 193, 54, 52, 108, 255, 255, 21, 252, 193, 230, 107, 112, 255, 255, 48, 252, 193, 142, 179, 107, 255, 255, 14, 252, 193, 174, 203, 107, 255, 255, 16, 252, 193, 104, 212, 107, 255, 255, 16, 252, 193, 180, 219, 107, 255, 255, 17, 252, 193, 22, 224, 101, 34, 0, 40, 103, 16, 0, 40, 103, 255, 255, 235, 251, 193, 114, 32, 107, 255, 255, 6, 252, 193, 188, 98, 107, 255, 255, 9, 252, 193, 190, 121, 107, 255, 255, 11, 252, 193, 192, 224, 101, 255, 255, 222, 251, 193, 42, 229, 101, 255, 255, 222, 251, 193, 86, 240, 102, 255, 255, 232, 251, 193, 90, 8, 103, 255, 255, 233, 251, 193, 86, 135, 101, 16, 0, 135, 101, 255, 255, 217, 251, 193, 254, 151, 101, 255, 255, 218, 251, 193, 178, 164, 101, 255, 255, 219, 251, 193, 74, 185, 101, 255, 255, 220, 251, 193, 68, 54, 98, 255, 255, 190, 251, 193, 162, 75, 98, 255, 255, 191, 251, 193, 156, 47, 101, 255, 255, 212, 251, 193, 60, 52, 101, 255, 255, 212, 251, 193, 230, 215, 86, 142, 0, 219, 93, 70, 0, 254, 94, 34, 0, 97, 95, 16, 0, 97, 95, 255, 255, 165, 251, 193, 122, 115, 95, 255, 255, 166, 251, 193, 88, 195, 95, 255, 255, 169, 251, 193, 252, 8, 98, 255, 255, 188, 251, 193, 76, 254, 94, 255, 255, 160, 251, 193, 102, 11, 95, 255, 255, 161, 251, 193, 108, 19, 95, 255, 255, 161, 251, 193, 216, 80, 95, 255, 255, 164, 251, 193, 210, 114, 94, 16, 0, 114, 94, 255, 255, 153, 251, 193, 170, 122, 94, 255, 255, 154, 251, 193, 90, 127, 94, 255, 255, 154, 251, 193, 218, 244, 94, 255, 255, 159, 251, 193, 254, 219, 93, 255, 255, 147, 251, 193, 66, 229, 93, 255, 255, 147, 251, 193, 210, 241, 93, 255, 255, 148, 251, 193, 132, 254, 93, 255, 255, 149, 251, 193, 20, 80, 91, 34, 0, 34, 92, 16, 0, 34, 92, 255, 255, 130, 251, 193, 144, 56, 92, 255, 255, 131, 251, 193, 192, 110, 92, 255, 255, 135, 251, 193, 8, 113, 92, 255, 255, 135, 251, 193, 128, 80, 91, 255, 255, 119, 251, 193, 108, 128, 91, 255, 255, 121, 251, 193, 174, 248, 91, 255, 255, 128, 251, 193, 22, 15, 92, 255, 255, 129, 251, 193, 34, 10, 89, 16, 0, 10, 89, 255, 255, 100, 251, 193, 232, 21, 89, 255, 255, 101, 251, 193, 136, 39, 89, 255, 255, 102, 251, 193, 204, 115, 89, 255, 255, 106, 251, 193, 100, 215, 86, 255, 255, 83, 251, 193, 76, 31, 87, 255, 255, 85, 251, 193, 186, 235, 88, 255, 255, 99, 251, 193, 220, 2, 89, 255, 255, 100, 251, 193, 178, 224, 81, 70, 0, 65, 83, 34, 0, 130, 83, 16, 0, 130, 83, 255, 255, 50, 251, 193, 178, 182, 83, 255, 255, 53, 251, 193, 120, 200, 83, 255, 255, 54, 251, 193, 122, 227, 83, 255, 255, 56, 251, 193, 112, 65, 83, 255, 255, 47, 251, 193, 180, 92, 83, 255, 255, 49, 251, 193, 44, 100, 83, 255, 255, 209, 253, 193, 152, 105, 83, 255, 255, 49, 251, 193, 240, 249, 82, 16, 0, 249, 82, 255, 255, 44, 251, 193, 244, 21, 83, 255, 255, 45, 251, 193, 234, 26, 83, 255, 255, 46, 251, 193, 82, 56, 83, 255, 255, 47, 251, 193, 118, 224, 81, 255, 255, 33, 251, 193, 196, 245, 81, 255, 255, 34, 251, 193, 200, 0, 82, 255, 255, 35, 251, 193, 132, 155, 82, 255, 255, 42, 251, 193, 8, 160, 78, 34, 0, 107, 81, 16, 0, 107, 81, 255, 255, 28, 251, 193, 74, 130, 81, 255, 255, 29, 251, 193, 104, 150, 81, 255, 255, 30, 251, 193, 74, 171, 81, 255, 255, 31, 251, 193, 62, 160, 78, 255, 255, 8, 251, 193, 232, 186, 78, 255, 255, 10, 251, 193, 68, 63, 81, 255, 255, 26, 251, 193, 4, 101, 81, 255, 255, 27, 251, 193, 188, 63, 78, 16, 0, 63, 78, 255, 255, 4, 251, 193, 252, 89, 78, 255, 255, 6, 251, 193, 26, 133, 78, 255, 255, 7, 251, 193, 248, 140, 78, 255, 255, 8, 251, 193, 74, 0, 78, 255, 255, 2, 251, 193, 2, 40, 78, 255, 255, 4, 251, 193, 60, 44, 78, 255, 255, 66, 252, 193, 246, 54, 78, 255, 255, 4, 251, 193, 200, 26, 130, 246, 1, 0, 78, 1, 78, 40, 78, 41, 78, 44, 78, 45, 78, 54, 78, 55, 78, 63, 78, 64, 78, 89, 78, 90, 78, 133, 78, 134, 78, 140, 78, 141, 78, 160, 78, 161, 78, 186, 78, 187, 78, 63, 81, 64, 81, 101, 81, 102, 81, 107, 81, 108, 81, 130, 81, 131, 81, 150, 81, 151, 81, 171, 81, 172, 81, 224, 81, 225, 81, 245, 81, 246, 81, 0, 82, 1, 82, 155, 82, 156, 82, 249, 82, 250, 82, 21, 83, 22, 83, 26, 83, 27, 83, 56, 83, 57, 83, 65, 83, 66, 83, 92, 83, 93, 83, 100, 83, 101, 83, 105, 83, 106, 83, 130, 83, 131, 83, 182, 83, 183, 83, 200, 83, 201, 83, 227, 83, 228, 83, 215, 86, 216, 86, 31, 87, 32, 87, 235, 88, 236, 88, 2, 89, 3, 89, 10, 89, 11, 89, 21, 89, 22, 89, 39, 89, 40, 89, 115, 89, 116, 89, 80, 91, 81, 91, 128, 91, 129, 91, 248, 91, 249, 91, 15, 92, 16, 92, 34, 92, 35, 92, 56, 92, 57, 92, 110, 92, 111, 92, 113, 92, 114, 92, 219, 93, 220, 93, 229, 93, 230, 93, 241, 93, 242, 93, 254, 93, 255, 93, 114, 94, 115, 94, 122, 94, 123, 94, 127, 94, 128, 94, 244, 94, 245, 94, 254, 94, 255, 94, 11, 95, 12, 95, 19, 95, 20, 95, 80, 95, 81, 95, 97, 95, 98, 95, 115, 95, 116, 95, 195, 95, 196, 95, 8, 98, 9, 98, 54, 98, 55, 98, 75, 98, 76, 98, 47, 101, 48, 101, 52, 101, 53, 101, 135, 101, 136, 101, 151, 101, 152, 101, 164, 101, 165, 101, 185, 101, 186, 101, 224, 101, 225, 101, 229, 101, 230, 101, 240, 102, 241, 102, 8, 103, 9, 103, 40, 103, 41, 103, 32, 107, 33, 107, 98, 107, 99, 107, 121, 107, 122, 107, 179, 107, 180, 107, 203, 107, 204, 107, 212, 107, 213, 107, 219, 107, 220, 107, 15, 108, 16, 108, 20, 108, 21, 108, 52, 108, 53, 108, 107, 112, 108, 112, 42, 114, 43, 114, 54, 114, 55, 114, 59, 114, 60, 114, 63, 114, 64, 114, 71, 114, 72, 114, 89, 114, 90, 114, 91, 114, 92, 114, 172, 114, 173, 114, 132, 115, 133, 115, 137, 115, 138, 115, 220, 116, 221, 116, 230, 116, 231, 116, 24, 117, 25, 117, 31, 117, 32, 117, 40, 117, 41, 117, 48, 117, 49, 117, 139, 117, 140, 117, 146, 117, 147, 117, 118, 118, 119, 118, 125, 118, 126, 118, 174, 118, 175, 118, 191, 118, 192, 118, 238, 118, 239, 118, 219, 119, 220, 119, 226, 119, 227, 119, 243, 119, 244, 119, 58, 121, 59, 121, 184, 121, 185, 121, 190, 121, 191, 121, 116, 122, 117, 122, 203, 122, 204, 122, 249, 122, 250, 122, 115, 124, 116, 124, 248, 124, 249, 124, 159, 126, 160, 126, 54, 127, 55, 127, 81, 127, 82, 127, 138, 127, 139, 127, 189, 127, 190, 127, 1, 128, 2, 128, 12, 128, 13, 128, 18, 128, 19, 128, 51, 128, 52, 128, 127, 128, 128, 128, 137, 128, 138, 128, 227, 129, 228, 129, 234, 129, 235, 129, 243, 129, 244, 129, 252, 129, 253, 129, 12, 130, 13, 130, 27, 130, 28, 130, 31, 130, 32, 130, 110, 130, 111, 130, 114, 130, 115, 130, 120, 130, 121, 130, 77, 134, 78, 134, 107, 134, 108, 134, 64, 136, 65, 136, 76, 136, 77, 136, 99, 136, 100, 136, 126, 137, 127, 137, 139, 137, 140, 137, 193, 137, 194, 137, 210, 137, 211, 137, 0, 138, 1, 138, 160, 139, 161, 139, 55, 140, 56, 140, 70, 140, 71, 140, 85, 140, 86, 140, 120, 140, 121, 140, 157, 140, 158, 140, 29, 141, 30, 141, 100, 141, 101, 141, 112, 141, 113, 141, 179, 141, 180, 141, 171, 142, 172, 142, 202, 142, 203, 142, 102, 143, 103, 143, 155, 143, 156, 143, 176, 143, 177, 143, 181, 143, 182, 143, 145, 144, 146, 144, 73, 145, 74, 145, 198, 145, 199, 145, 204, 145, 205, 145, 209, 145, 210, 145, 133, 148, 134, 148, 119, 149, 120, 149, 127, 149, 129, 149, 232, 149, 233, 149, 28, 150, 29, 150, 182, 150, 183, 150, 185, 150, 186, 150, 232, 150, 233, 150, 81, 151, 82, 151, 94, 151, 95, 151, 98, 151, 99, 151, 105, 151, 106, 151, 203, 151, 204, 151, 230, 151, 231, 151, 237, 151, 238, 151, 243, 151, 244, 151, 1, 152, 2, 152, 117, 152, 118, 152, 168, 152, 169, 152, 206, 152, 207, 152, 219, 152, 220, 152, 222, 152, 224, 152, 99, 153, 100, 153, 150, 153, 151, 153, 153, 153, 154, 153, 172, 153, 173, 153, 108, 154, 109, 154, 168, 154, 169, 154, 216, 154, 217, 154, 223, 154, 224, 154, 37, 155, 38, 155, 47, 155, 48, 155, 50, 155, 51, 155, 60, 155, 61, 155, 90, 155, 91, 155, 124, 156, 125, 156, 229, 156, 230, 156, 31, 158, 32, 158, 117, 158, 118, 158, 127, 158, 128, 158, 165, 158, 167, 158, 187, 158, 188, 158, 195, 158, 196, 158, 205, 158, 206, 158, 209, 158, 210, 158, 249, 158, 250, 158, 253, 158, 255, 158, 14, 159, 15, 159, 19, 159, 20, 159, 32, 159, 33, 159, 59, 159, 60, 159, 74, 159, 75, 159, 80, 159, 81, 159, 82, 159, 83, 159, 127, 159, 128, 159, 141, 159, 142, 159, 153, 159, 154, 159, 156, 159, 157, 159, 159, 159, 161, 159, 126, 216, 127, 216, 2, 249, 3, 249, 7, 249, 9, 249, 10, 249, 11, 249, 52, 249, 53, 249, 64, 249, 65, 249, 113, 249, 114, 249, 129, 249, 130, 249, 138, 249, 139, 249, 196, 249, 197, 249, 233, 249, 234, 249, 247, 249, 248, 249, 8, 250, 9, 250, 10, 250, 11, 250, 30, 250, 31, 250, 60, 250, 61, 250, 149, 250, 150, 250, 206, 250, 207, 250, 2, 0, 29, 248, 2, 0, 30, 248, 2, 0, 117, 248, 2, 0, 118, 248, 2, 0, 120, 248, 2, 0, 121, 248, 2, 0, 144, 248, 2, 0, 145, 248, 2, 0, 196, 249, 2, 0, 197, 249, 2, 0, 210, 249, 2, 0, 211, 249, 2, 0, 21, 250, 2, 0, 22, 250, 2, 0, 23, 250, 2, 0, 25, 250, 2, 0, 28, 250, 2, 0, 29, 250, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 39, 17, 1, 16, 6, 0, 0, 96, 1, 0, 0, 96, 4, 0, 51, 0, 82, 0, 91, 0, 100, 0, 170, 170, 101, 37, 0, 32, 13, 46, 0, 32, 33, 17, 1, 32, 209, 26, 1, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 214, 26, 1, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 210, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 214, 6, 0, 0, 214, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 72, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 94, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 80, 9, 60, 0, 2, 9, 60, 0, 60, 0, 1, 9, 60, 0, 3, 9, 38, 0, 57, 9, 60, 0, 51, 9, 60, 0, 21, 9, 77, 9, 55, 9, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 250, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 120, 0, 0, 0, 120, 1, 0, 0, 120, 25, 0, 0, 120, 25, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 204, 25, 0, 0, 204, 25, 0, 0, 220, 25, 0, 0, 226, 25, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 39, 40, 41, 42, 43, 44, 45, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 119, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 102, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 1, 0, 202, 65, 1, 0, 202, 98, 1, 0, 202, 131, 1, 0, 202, 164, 1, 0, 202, 197, 1, 0, 202, 230, 1, 0, 202, 7, 2, 0, 202, 40, 2, 0, 202, 73, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 193, 2, 7, 104, 193, 3, 7, 104, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 2, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 2, 143, 104, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 5, 0, 70, 0, 2, 7, 104, 0, 0, 0, 0, 193, 2, 143, 104, 194, 5, 175, 240, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 62, 104, 5, 5, 49, 0, 77, 9, 55, 9, 255, 255, 143, 104, 193, 3, 2, 0, 55, 9, 56, 9, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 75, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 203, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 208, 6, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 110, 10, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 114, 10, 0, 0, 114, 10, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 68, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 86, 220, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 76, 0, 97, 0, 116, 0, 110, 0, 32, 0, 65, 0, 114, 0, 97, 0, 98, 0, 93, 0, 38, 0, 99, 0, 60, 0, 231, 0, 60, 0, 60, 0, 60, 0, 199, 0, 38, 0, 101, 0, 60, 0, 234, 0, 60, 0, 60, 0, 60, 0, 202, 0, 38, 0, 105, 0, 60, 0, 238, 0, 60, 0, 60, 0, 60, 0, 206, 0, 38, 0, 115, 0, 60, 0, 95, 1, 60, 0, 60, 0, 60, 0, 94, 1, 38, 0, 117, 0, 60, 0, 251, 0, 60, 0, 60, 0, 60, 0, 219, 0, 38, 0, 234, 0, 60, 0, 60, 0, 233, 0, 60, 0, 60, 0, 60, 0, 201, 0, 38, 0, 238, 0, 60, 0, 60, 0, 237, 0, 60, 0, 60, 0, 60, 0, 205, 0, 38, 0, 251, 0, 60, 0, 60, 0, 250, 0, 60, 0, 60, 0, 60, 0, 218, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 124, 40, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 76, 0, 0, 0, 76, 1, 0, 0, 44, 30, 0, 0, 44, 30, 0, 0, 44, 30, 0, 0, 44, 30, 0, 0, 200, 32, 0, 0, 200, 32, 0, 0, 100, 36, 0, 0, 100, 36, 0, 0, 100, 40, 0, 0, 25, 0, 0, 0, 2, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 94, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 197, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 48, 0, 56, 0, 118, 0, 126, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 148, 0, 155, 0, 48, 0, 159, 0, 48, 0, 165, 0, 173, 0, 180, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 188, 0, 142, 0, 48, 0, 48, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 134, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 192, 0, 216, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 19, 0, 202, 193, 19, 0, 202, 226, 19, 0, 202, 3, 20, 0, 202, 36, 20, 0, 202, 69, 20, 0, 202, 102, 20, 0, 202, 135, 20, 0, 202, 168, 20, 0, 202, 201, 20, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 156, 5, 0, 48, 201, 198, 0, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 70, 2, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 198, 3, 0, 156, 5, 0, 80, 201, 166, 4, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 134, 6, 0, 5, 5, 0, 48, 201, 70, 7, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 198, 8, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 70, 10, 0, 5, 5, 0, 80, 201, 38, 11, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 18, 0, 134, 5, 2, 47, 196, 138, 156, 50, 134, 70, 2, 51, 134, 5, 2, 51, 196, 150, 156, 50, 196, 138, 156, 58, 134, 70, 2, 59, 134, 5, 2, 59, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 134, 70, 2, 83, 134, 5, 2, 83, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 19, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 19, 0, 5, 5, 2, 47, 196, 138, 5, 50, 5, 70, 2, 51, 5, 5, 2, 51, 196, 150, 5, 50, 196, 138, 5, 58, 5, 70, 2, 59, 5, 5, 2, 59, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 5, 70, 2, 83, 5, 5, 2, 83, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 13, 0, 201, 70, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 15, 0, 201, 70, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 18, 0, 201, 70, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 21, 0, 201, 70, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 24, 0, 201, 70, 26, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 5, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 27, 0, 201, 70, 29, 0, 201, 198, 30, 0, 201, 70, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 33, 0, 201, 70, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 36, 0, 201, 70, 38, 0, 201, 198, 39, 0, 201, 70, 41, 0, 201, 198, 42, 0, 201, 70, 44, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 45, 0, 201, 70, 47, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 12, 0, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 197, 34, 14, 0, 197, 98, 14, 0, 197, 226, 11, 0, 197, 98, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 48, 0, 201, 70, 50, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 51, 0, 201, 70, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 3, 0, 197, 162, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 54, 0, 201, 70, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 70, 2, 51, 5, 162, 0, 0, 134, 5, 2, 51, 5, 162, 0, 0, 5, 70, 2, 51, 5, 162, 0, 0, 5, 5, 2, 51, 5, 162, 0, 0, 134, 70, 2, 59, 5, 162, 0, 0, 134, 5, 2, 59, 5, 162, 0, 0, 5, 70, 2, 59, 5, 162, 0, 0, 5, 5, 2, 59, 5, 162, 0, 0, 134, 70, 2, 83, 5, 162, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 5, 70, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 134, 70, 2, 83, 5, 190, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 70, 2, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 134, 70, 2, 51, 5, 160, 0, 0, 134, 5, 2, 51, 5, 160, 0, 0, 5, 70, 2, 51, 5, 160, 0, 0, 5, 5, 2, 51, 5, 160, 0, 0, 134, 5, 2, 47, 5, 136, 0, 0, 5, 5, 2, 47, 5, 136, 0, 0, 134, 70, 2, 51, 5, 204, 0, 0, 134, 5, 2, 51, 5, 204, 0, 0, 5, 70, 2, 51, 5, 204, 0, 0, 5, 5, 2, 51, 5, 204, 0, 0, 134, 70, 2, 51, 5, 208, 0, 0, 134, 5, 2, 51, 5, 208, 0, 0, 5, 70, 2, 51, 5, 208, 0, 0, 5, 5, 2, 51, 5, 208, 0, 0, 134, 70, 2, 59, 5, 208, 0, 0, 134, 5, 2, 59, 5, 208, 0, 0, 5, 70, 2, 59, 5, 208, 0, 0, 5, 5, 2, 59, 5, 208, 0, 0, 134, 70, 2, 83, 5, 198, 0, 0, 134, 5, 2, 83, 5, 198, 0, 0, 5, 70, 2, 83, 5, 198, 0, 0, 5, 5, 2, 83, 5, 198, 0, 0, 134, 70, 2, 83, 5, 208, 0, 0, 134, 5, 2, 83, 5, 208, 0, 0, 5, 70, 2, 83, 5, 208, 0, 0, 5, 5, 2, 83, 5, 208, 0, 0, 134, 70, 2, 83, 5, 204, 0, 0, 134, 5, 2, 83, 5, 204, 0, 0, 5, 70, 2, 83, 5, 204, 0, 0, 5, 5, 2, 83, 5, 204, 0, 0, 134, 70, 2, 51, 5, 196, 0, 0, 134, 5, 2, 51, 5, 196, 0, 0, 5, 70, 2, 51, 5, 196, 0, 0, 5, 5, 2, 51, 5, 196, 0, 0, 134, 5, 2, 51, 5, 136, 0, 0, 5, 5, 2, 51, 5, 136, 0, 0, 134, 5, 2, 51, 5, 138, 0, 0, 5, 5, 2, 51, 5, 138, 0, 0, 134, 5, 2, 51, 5, 182, 0, 0, 5, 5, 2, 51, 5, 182, 0, 0, 134, 5, 2, 51, 5, 154, 0, 0, 5, 5, 2, 51, 5, 154, 0, 0, 134, 70, 2, 59, 5, 196, 0, 0, 134, 5, 2, 59, 5, 196, 0, 0, 5, 70, 2, 59, 5, 196, 0, 0, 5, 5, 2, 59, 5, 196, 0, 0, 134, 70, 2, 83, 5, 196, 0, 0, 134, 5, 2, 83, 5, 196, 0, 0, 5, 70, 2, 83, 5, 196, 0, 0, 5, 5, 2, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 48, 0, 39, 3, 2, 239, 134, 5, 0, 50, 156, 5, 2, 0, 1, 3, 2, 243, 134, 70, 2, 3, 2, 243, 134, 5, 65, 3, 2, 243, 134, 70, 0, 58, 156, 5, 2, 0, 1, 3, 2, 251, 134, 70, 2, 3, 2, 251, 134, 5, 65, 3, 2, 251, 134, 70, 0, 78, 156, 5, 48, 0, 39, 3, 255, 255, 2, 79, 134, 5, 0, 82, 156, 5, 2, 0, 1, 3, 255, 255, 2, 83, 134, 70, 2, 3, 255, 255, 2, 83, 134, 5, 65, 3, 255, 255, 2, 83, 134, 70, 0, 46, 5, 5, 48, 0, 39, 3, 2, 239, 5, 5, 0, 50, 5, 5, 2, 0, 1, 3, 2, 243, 5, 70, 2, 3, 2, 243, 5, 5, 65, 3, 2, 243, 5, 70, 0, 58, 5, 5, 2, 0, 1, 3, 2, 251, 5, 70, 2, 3, 2, 251, 5, 5, 65, 3, 2, 251, 5, 70, 0, 78, 5, 5, 48, 0, 39, 3, 255, 255, 2, 79, 5, 5, 0, 82, 5, 5, 2, 0, 1, 3, 255, 255, 2, 83, 5, 70, 2, 3, 255, 255, 2, 83, 5, 5, 65, 3, 255, 255, 2, 83, 5, 70, 156, 50, 196, 162, 2, 0, 1, 3, 197, 162, 2, 3, 0, 192, 197, 98, 65, 3, 197, 162, 5, 50, 196, 162, 2, 0, 1, 3, 0, 192, 197, 162, 2, 3, 0, 192, 197, 226, 65, 3, 0, 192, 197, 162, 156, 58, 196, 162, 2, 0, 1, 3, 1, 192, 197, 34, 2, 3, 1, 192, 197, 98, 65, 3, 1, 192, 197, 34, 5, 58, 196, 162, 2, 0, 1, 3, 1, 192, 197, 162, 2, 3, 1, 192, 197, 226, 65, 3, 1, 192, 197, 162, 156, 82, 196, 162, 2, 0, 1, 3, 2, 192, 197, 34, 2, 3, 2, 192, 197, 98, 65, 3, 2, 192, 197, 34, 5, 82, 196, 162, 2, 0, 1, 3, 2, 192, 197, 162, 2, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 156, 82, 196, 190, 2, 0, 1, 3, 3, 192, 197, 34, 2, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 5, 82, 196, 190, 2, 0, 1, 3, 3, 192, 197, 162, 2, 3, 3, 192, 197, 226, 65, 3, 3, 192, 197, 162, 156, 50, 196, 160, 2, 0, 1, 3, 4, 192, 197, 34, 2, 3, 4, 192, 197, 98, 65, 3, 4, 192, 197, 34, 5, 50, 196, 160, 2, 0, 1, 3, 4, 192, 197, 162, 2, 3, 4, 192, 197, 226, 65, 3, 4, 192, 197, 162, 156, 50, 196, 204, 2, 0, 1, 3, 5, 192, 197, 162, 2, 3, 5, 192, 197, 226, 65, 3, 5, 192, 197, 162, 5, 50, 196, 204, 2, 0, 1, 3, 6, 192, 197, 34, 2, 3, 6, 192, 197, 98, 65, 3, 6, 192, 197, 34, 156, 50, 196, 208, 2, 0, 1, 3, 6, 192, 197, 162, 2, 3, 6, 192, 197, 226, 65, 3, 6, 192, 197, 162, 5, 50, 196, 208, 2, 0, 1, 3, 7, 192, 197, 34, 2, 3, 7, 192, 197, 98, 65, 3, 7, 192, 197, 34, 156, 58, 196, 208, 2, 0, 1, 3, 7, 192, 197, 162, 2, 3, 7, 192, 197, 226, 65, 3, 7, 192, 197, 162, 5, 58, 196, 208, 2, 0, 1, 3, 8, 192, 197, 34, 2, 3, 8, 192, 197, 98, 65, 3, 8, 192, 197, 34, 156, 82, 196, 198, 2, 0, 1, 3, 8, 192, 197, 162, 2, 3, 8, 192, 197, 226, 65, 3, 8, 192, 197, 162, 5, 82, 196, 198, 2, 0, 1, 3, 9, 192, 197, 34, 2, 3, 9, 192, 197, 98, 65, 3, 9, 192, 197, 34, 156, 82, 196, 208, 2, 0, 1, 3, 9, 192, 197, 162, 2, 3, 9, 192, 197, 226, 65, 3, 9, 192, 197, 162, 5, 82, 196, 208, 2, 0, 1, 3, 10, 192, 197, 34, 2, 3, 10, 192, 197, 98, 65, 3, 10, 192, 197, 34, 156, 82, 196, 204, 2, 0, 1, 3, 10, 192, 197, 162, 2, 3, 10, 192, 197, 226, 65, 3, 10, 192, 197, 162, 5, 82, 196, 204, 2, 0, 1, 3, 11, 192, 197, 34, 2, 3, 11, 192, 197, 98, 65, 3, 11, 192, 197, 34, 156, 50, 196, 196, 2, 0, 1, 3, 11, 192, 197, 162, 2, 3, 11, 192, 197, 226, 65, 3, 11, 192, 197, 162, 5, 50, 196, 196, 2, 0, 1, 3, 12, 192, 197, 34, 2, 3, 12, 192, 197, 98, 65, 3, 12, 192, 197, 34, 156, 58, 196, 196, 2, 0, 1, 3, 14, 192, 197, 162, 2, 3, 14, 192, 197, 226, 65, 3, 14, 192, 197, 162, 5, 58, 196, 196, 2, 0, 1, 3, 15, 192, 197, 34, 2, 3, 15, 192, 197, 98, 65, 3, 15, 192, 197, 34, 156, 82, 196, 196, 2, 0, 1, 3, 15, 192, 197, 162, 2, 3, 15, 192, 197, 226, 65, 3, 15, 192, 197, 162, 5, 82, 196, 196, 2, 0, 1, 3, 16, 192, 197, 34, 2, 3, 16, 192, 197, 98, 65, 3, 16, 192, 197, 34, 16, 0, 197, 163, 2, 0, 1, 3, 17, 192, 197, 3, 2, 3, 17, 192, 197, 99, 65, 3, 17, 192, 197, 3, 17, 0, 197, 195, 2, 0, 1, 3, 18, 192, 197, 35, 2, 3, 18, 192, 197, 131, 65, 3, 18, 192, 197, 35, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 72, 28, 4, 185, 84, 185, 88, 185, 92, 30, 4, 186, 108, 185, 112, 185, 116, 185, 120, 186, 124, 185, 132, 185, 136, 185, 140, 185, 148, 32, 4, 185, 160, 34, 4, 185, 176, 185, 180, 185, 184, 185, 188, 185, 192, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 36, 4, 168, 72, 38, 4, 168, 84, 168, 88, 168, 92, 40, 4, 168, 108, 168, 112, 168, 116, 168, 120, 168, 124, 168, 132, 168, 136, 168, 140, 168, 148, 42, 4, 168, 160, 44, 4, 168, 176, 168, 180, 168, 184, 168, 188, 168, 192, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 132, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 185, 68, 186, 77, 217, 80, 185, 80, 90, 78, 186, 97, 217, 100, 185, 100, 90, 98, 217, 72, 154, 126, 185, 133, 153, 133, 249, 133, 153, 134, 89, 134, 136, 14, 217, 134, 185, 169, 217, 172, 185, 172, 89, 170, 153, 189, 185, 196, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 168, 68, 168, 77, 200, 80, 168, 80, 72, 78, 168, 97, 200, 100, 168, 100, 72, 98, 200, 72, 136, 126, 168, 133, 136, 133, 232, 133, 136, 134, 72, 134, 128, 14, 200, 134, 168, 169, 200, 172, 168, 172, 72, 170, 136, 189, 168, 196, 72, 190, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 46, 4, 48, 4, 26, 78, 8, 78, 249, 89, 232, 89, 217, 89, 200, 89, 185, 90, 168, 90, 249, 90, 232, 90, 249, 93, 232, 93, 89, 95, 72, 95, 154, 98, 136, 98, 58, 99, 40, 99, 218, 97, 200, 97, 50, 4, 52, 4, 186, 98, 168, 104, 6, 8, 8, 8, 250, 109, 232, 109, 249, 114, 232, 114, 168, 144, 153, 117, 136, 117, 249, 118, 232, 118, 25, 118, 8, 118, 121, 119, 104, 119, 89, 119, 72, 119, 154, 125, 136, 125, 250, 126, 232, 126, 26, 126, 8, 126, 10, 8, 185, 128, 168, 128, 57, 135, 40, 135, 217, 133, 200, 133, 121, 134, 104, 134, 12, 8, 14, 8, 153, 149, 136, 149, 249, 150, 232, 150, 25, 150, 8, 150, 154, 153, 136, 153, 250, 153, 232, 153, 185, 156, 168, 156, 26, 154, 8, 154, 249, 162, 232, 162, 25, 162, 8, 162, 185, 164, 168, 164, 153, 170, 136, 170, 57, 171, 40, 171, 217, 169, 200, 169, 57, 170, 40, 170, 121, 170, 104, 170, 54, 4, 56, 4, 249, 181, 232, 181, 249, 189, 232, 189, 89, 190, 153, 193, 136, 193, 185, 194, 168, 194, 25, 194, 8, 194, 232, 152, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 152, 169, 152, 200, 56, 169, 76, 187, 96, 187, 108, 169, 96, 169, 108, 169, 200, 169, 124, 217, 132, 187, 76, 200, 132, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 186, 76, 255, 5, 186, 96, 255, 5, 186, 152, 255, 5, 185, 168, 255, 5, 168, 64, 255, 5, 168, 76, 255, 5, 168, 96, 255, 5, 168, 152, 255, 5, 168, 168, 255, 5, 26, 79, 255, 5, 8, 79, 255, 5, 26, 99, 255, 5, 8, 99, 255, 5, 25, 171, 255, 5, 8, 171, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 71, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 103, 10, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 108, 10, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 221, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 225, 6, 0, 0, 225, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 36, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 93, 0, 38, 0, 53, 4, 60, 0, 81, 4, 60, 0, 60, 0, 60, 0, 1, 4, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 184, 26, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 15, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 60, 0, 0, 0, 64, 0, 0, 0, 64, 1, 0, 0, 64, 26, 0, 0, 64, 26, 0, 0, 64, 26, 0, 0, 64, 26, 0, 0, 132, 26, 0, 0, 132, 26, 0, 0, 160, 26, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 39, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 135, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 94, 0, 102, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 118, 0, 48, 0, 48, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 120, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 134, 5, 41, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 40, 97, 174, 5, 48, 0, 8, 3, 255, 255, 41, 97, 134, 5, 40, 97, 5, 5, 48, 0, 8, 3, 255, 255, 41, 97, 5, 5, 3, 0, 69, 0, 60, 0, 32, 0, 39, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 214, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 219, 6, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 208, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 212, 8, 0, 0, 212, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 48, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 46, 220, 38, 0, 67, 0, 60, 0, 99, 0, 12, 3, 60, 0, 60, 0, 60, 0, 67, 0, 12, 3, 38, 0, 71, 0, 60, 0, 231, 1, 60, 0, 60, 0, 60, 0, 230, 1, 38, 0, 72, 0, 60, 0, 31, 2, 60, 0, 60, 0, 60, 0, 30, 2, 38, 0, 83, 0, 60, 0, 115, 0, 12, 3, 60, 0, 60, 0, 60, 0, 83, 0, 12, 3, 38, 0, 90, 0, 60, 0, 122, 0, 12, 3, 60, 0, 60, 0, 60, 0, 90, 0, 12, 3, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 82, 34, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 196, 27, 0, 0, 196, 27, 0, 0, 196, 27, 0, 0, 196, 27, 0, 0, 192, 28, 0, 0, 192, 28, 0, 0, 58, 30, 0, 0, 58, 30, 0, 0, 58, 34, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 175, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 89, 0, 97, 0, 48, 0, 56, 0, 105, 0, 113, 0, 115, 0, 123, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 146, 0, 48, 0, 154, 0, 158, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 166, 0, 139, 0, 48, 0, 48, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 100, 1, 192, 0, 164, 1, 204, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 6, 0, 202, 193, 6, 0, 202, 226, 6, 0, 202, 3, 7, 0, 202, 36, 7, 0, 202, 69, 7, 0, 202, 102, 7, 0, 202, 135, 7, 0, 202, 168, 7, 0, 202, 201, 7, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 198, 0, 0, 201, 134, 1, 0, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 70, 2, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 38, 3, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 6, 4, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 198, 4, 0, 201, 134, 5, 0, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 70, 6, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 38, 7, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 5, 0, 201, 6, 8, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 6, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 6, 0, 201, 166, 8, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 9, 0, 201, 38, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 10, 0, 201, 166, 11, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 55, 5, 5, 2, 55, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 12, 0, 201, 38, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 57, 5, 5, 2, 57, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 13, 0, 201, 166, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 15, 0, 201, 38, 16, 0, 201, 230, 16, 0, 201, 166, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 18, 0, 201, 38, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 4, 0, 197, 98, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 19, 0, 201, 166, 20, 0, 201, 102, 21, 0, 201, 38, 22, 0, 201, 230, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 55, 5, 160, 0, 0, 5, 5, 2, 55, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 57, 5, 196, 0, 0, 5, 5, 2, 57, 5, 196, 0, 0, 134, 5, 2, 57, 5, 160, 0, 0, 5, 5, 2, 57, 5, 160, 0, 0, 134, 5, 2, 57, 5, 206, 0, 0, 5, 5, 2, 57, 5, 206, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 57, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 48, 0, 12, 3, 2, 239, 134, 5, 0, 54, 156, 5, 48, 0, 12, 3, 2, 247, 134, 5, 0, 56, 156, 5, 48, 0, 12, 3, 2, 249, 134, 5, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 46, 5, 5, 48, 0, 12, 3, 2, 239, 5, 5, 0, 54, 5, 5, 48, 0, 12, 3, 2, 247, 5, 5, 0, 56, 5, 5, 48, 0, 12, 3, 2, 249, 5, 5, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 48, 0, 12, 3, 197, 162, 5, 46, 196, 160, 48, 0, 12, 3, 0, 192, 197, 98, 156, 54, 196, 160, 48, 0, 12, 3, 0, 192, 197, 162, 5, 54, 196, 160, 48, 0, 12, 3, 0, 192, 197, 226, 156, 78, 196, 160, 48, 0, 12, 3, 1, 192, 197, 34, 5, 78, 196, 160, 48, 0, 12, 3, 1, 192, 197, 98, 156, 78, 196, 202, 48, 0, 12, 3, 1, 192, 197, 162, 5, 78, 196, 202, 48, 0, 12, 3, 1, 192, 197, 226, 156, 56, 196, 196, 48, 0, 12, 3, 2, 192, 197, 34, 5, 56, 196, 196, 48, 0, 12, 3, 2, 192, 197, 98, 156, 56, 196, 160, 48, 0, 12, 3, 2, 192, 197, 162, 5, 56, 196, 160, 48, 0, 12, 3, 2, 192, 197, 226, 156, 56, 196, 206, 48, 0, 12, 3, 3, 192, 197, 34, 5, 56, 196, 206, 48, 0, 12, 3, 3, 192, 197, 98, 156, 78, 196, 196, 48, 0, 12, 3, 3, 192, 197, 162, 5, 78, 196, 196, 48, 0, 12, 3, 3, 192, 197, 226, 156, 92, 196, 196, 48, 0, 12, 3, 4, 192, 197, 162, 5, 92, 196, 196, 48, 0, 12, 3, 4, 192, 197, 226, 156, 92, 196, 210, 48, 0, 12, 3, 5, 192, 197, 34, 5, 92, 196, 210, 48, 0, 12, 3, 5, 192, 197, 98, 5, 56, 196, 210, 48, 0, 12, 3, 5, 192, 197, 162, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 72, 186, 76, 185, 80, 28, 4, 30, 4, 186, 92, 186, 100, 185, 104, 185, 108, 185, 112, 186, 116, 185, 124, 185, 128, 185, 132, 185, 140, 32, 4, 185, 152, 185, 160, 185, 164, 185, 168, 185, 172, 185, 176, 34, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 36, 4, 168, 72, 168, 76, 168, 80, 38, 4, 40, 4, 168, 92, 168, 100, 168, 104, 168, 108, 168, 112, 168, 116, 168, 124, 168, 128, 168, 132, 168, 140, 42, 4, 168, 152, 168, 160, 168, 164, 168, 168, 168, 172, 168, 176, 44, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 124, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 46, 4, 186, 77, 154, 77, 250, 77, 90, 78, 186, 93, 154, 93, 250, 93, 90, 94, 217, 72, 154, 118, 185, 125, 153, 125, 249, 125, 153, 126, 89, 126, 136, 14, 217, 126, 185, 161, 153, 161, 249, 161, 89, 162, 153, 177, 185, 188, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 48, 4, 168, 77, 136, 77, 232, 77, 72, 78, 168, 93, 136, 93, 232, 93, 72, 94, 200, 72, 136, 118, 168, 125, 136, 125, 232, 125, 136, 126, 72, 126, 128, 14, 200, 126, 168, 161, 136, 161, 232, 161, 72, 162, 136, 177, 168, 188, 72, 178, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 185, 68, 168, 68, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 249, 85, 232, 85, 217, 85, 200, 85, 185, 86, 168, 86, 50, 4, 52, 4, 249, 89, 232, 89, 89, 91, 72, 91, 154, 94, 136, 94, 58, 95, 40, 95, 218, 93, 200, 93, 26, 95, 8, 95, 186, 94, 168, 96, 6, 8, 8, 8, 250, 101, 232, 101, 249, 106, 232, 106, 168, 136, 153, 109, 136, 109, 249, 110, 232, 110, 25, 110, 8, 110, 121, 111, 104, 111, 89, 111, 72, 111, 154, 117, 136, 117, 250, 118, 232, 118, 26, 118, 8, 118, 10, 8, 185, 120, 168, 120, 57, 127, 40, 127, 217, 125, 200, 125, 121, 126, 104, 126, 12, 8, 14, 8, 153, 141, 136, 141, 249, 142, 232, 142, 25, 142, 8, 142, 154, 145, 136, 145, 250, 145, 232, 145, 54, 4, 56, 4, 185, 148, 168, 148, 249, 154, 232, 154, 25, 154, 8, 154, 185, 156, 168, 156, 153, 162, 136, 162, 57, 163, 40, 163, 217, 161, 200, 161, 57, 162, 40, 162, 121, 162, 104, 162, 25, 163, 8, 163, 249, 169, 232, 169, 249, 177, 232, 177, 89, 178, 153, 181, 136, 181, 185, 182, 168, 182, 185, 184, 168, 184, 232, 144, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 144, 169, 144, 200, 56, 169, 76, 187, 92, 187, 100, 169, 92, 169, 100, 169, 192, 169, 116, 217, 124, 187, 76, 200, 124, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 185, 84, 255, 5, 185, 88, 255, 5, 186, 144, 255, 5, 185, 180, 255, 5, 168, 64, 255, 5, 168, 84, 255, 5, 168, 88, 255, 5, 168, 144, 255, 5, 168, 180, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 249, 86, 255, 5, 232, 86, 255, 5, 250, 146, 255, 5, 232, 146, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 51, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 201, 8, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 206, 8, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 61, 22, 0, 32, 7, 0, 0, 0, 24, 0, 0, 0, 65, 22, 0, 0, 65, 22, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 207, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 112, 104, 111, 110, 101, 116, 105, 99, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 170, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 38, 0, 69, 0, 60, 0, 91, 2, 60, 0, 60, 0, 60, 0, 144, 1, 38, 0, 79, 0, 60, 0, 60, 0, 84, 2, 60, 0, 60, 0, 60, 0, 134, 1, 0, 0, 83, 221, 38, 0, 69, 0, 60, 0, 91, 2, 60, 0, 60, 0, 60, 0, 144, 1, 38, 0, 79, 0, 60, 0, 60, 0, 84, 2, 60, 0, 60, 0, 60, 0, 134, 1, 38, 0, 71, 0, 60, 0, 103, 0, 98, 0, 60, 0, 60, 0, 60, 0, 103, 0, 66, 0, 60, 0, 60, 0, 60, 0, 71, 0, 98, 0, 60, 0, 60, 0, 60, 0, 71, 0, 66, 0, 38, 0, 75, 0, 60, 0, 107, 0, 112, 0, 60, 0, 60, 0, 60, 0, 107, 0, 80, 0, 60, 0, 60, 0, 60, 0, 75, 0, 112, 0, 60, 0, 60, 0, 60, 0, 75, 0, 80, 0, 38, 0, 77, 0, 60, 0, 109, 0, 98, 0, 60, 0, 60, 0, 60, 0, 109, 0, 66, 0, 60, 0, 60, 0, 60, 0, 77, 0, 98, 0, 60, 0, 60, 0, 60, 0, 77, 0, 66, 0, 60, 0, 109, 0, 102, 0, 60, 0, 60, 0, 60, 0, 109, 0, 70, 0, 60, 0, 60, 0, 60, 0, 77, 0, 102, 0, 60, 0, 60, 0, 60, 0, 77, 0, 70, 0, 60, 0, 109, 0, 112, 0, 60, 0, 60, 0, 60, 0, 109, 0, 80, 0, 60, 0, 60, 0, 60, 0, 77, 0, 112, 0, 60, 0, 60, 0, 60, 0, 77, 0, 80, 0, 60, 0, 109, 0, 118, 0, 60, 0, 60, 0, 60, 0, 109, 0, 86, 0, 60, 0, 60, 0, 60, 0, 77, 0, 118, 0, 60, 0, 60, 0, 60, 0, 77, 0, 86, 0, 38, 0, 78, 0, 60, 0, 110, 0, 100, 0, 60, 0, 60, 0, 60, 0, 110, 0, 68, 0, 60, 0, 60, 0, 60, 0, 78, 0, 100, 0, 60, 0, 60, 0, 60, 0, 78, 0, 68, 0, 60, 0, 110, 0, 103, 0, 60, 0, 60, 0, 60, 0, 110, 0, 71, 0, 60, 0, 60, 0, 60, 0, 78, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 71, 0, 60, 0, 110, 0, 103, 0, 98, 0, 60, 0, 60, 0, 60, 0, 110, 0, 103, 0, 66, 0, 60, 0, 60, 0, 60, 0, 110, 0, 71, 0, 98, 0, 60, 0, 60, 0, 60, 0, 110, 0, 71, 0, 66, 0, 60, 0, 60, 0, 60, 0, 78, 0, 103, 0, 98, 0, 60, 0, 60, 0, 60, 0, 78, 0, 103, 0, 66, 0, 60, 0, 60, 0, 60, 0, 78, 0, 71, 0, 66, 0, 60, 0, 110, 0, 107, 0, 60, 0, 60, 0, 60, 0, 110, 0, 75, 0, 60, 0, 60, 0, 60, 0, 78, 0, 107, 0, 60, 0, 60, 0, 60, 0, 78, 0, 75, 0, 60, 0, 110, 0, 115, 0, 60, 0, 60, 0, 60, 0, 110, 0, 83, 0, 60, 0, 60, 0, 60, 0, 78, 0, 115, 0, 60, 0, 60, 0, 60, 0, 78, 0, 83, 0, 60, 0, 110, 0, 116, 0, 60, 0, 60, 0, 60, 0, 110, 0, 84, 0, 60, 0, 60, 0, 60, 0, 78, 0, 116, 0, 60, 0, 60, 0, 60, 0, 78, 0, 84, 0, 60, 0, 110, 0, 121, 0, 60, 0, 60, 0, 60, 0, 110, 0, 89, 0, 60, 0, 60, 0, 60, 0, 78, 0, 121, 0, 60, 0, 60, 0, 60, 0, 78, 0, 89, 0, 60, 0, 110, 0, 122, 0, 60, 0, 60, 0, 60, 0, 110, 0, 90, 0, 60, 0, 60, 0, 60, 0, 78, 0, 122, 0, 60, 0, 60, 0, 60, 0, 78, 0, 90, 0, 38, 0, 83, 0, 60, 0, 115, 0, 104, 0, 60, 0, 60, 0, 60, 0, 115, 0, 72, 0, 60, 0, 60, 0, 60, 0, 83, 0, 104, 0, 60, 0, 60, 0, 60, 0, 83, 0, 72, 0, 38, 0, 84, 0, 60, 0, 116, 0, 115, 0, 60, 0, 60, 0, 60, 0, 116, 0, 83, 0, 60, 0, 60, 0, 60, 0, 84, 0, 115, 0, 60, 0, 60, 0, 60, 0, 84, 0, 83, 0, 0, 0, 170, 170, 10, 60, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 116, 25, 0, 0, 116, 25, 0, 0, 68, 33, 0, 0, 68, 33, 0, 0, 20, 41, 0, 0, 20, 41, 0, 0, 4, 53, 0, 0, 166, 53, 0, 0, 238, 59, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 138, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 91, 0, 99, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 121, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 115, 0, 48, 0, 48, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 64, 1, 192, 0, 192, 0, 108, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 64, 61, 0, 202, 97, 61, 0, 202, 130, 61, 0, 202, 163, 61, 0, 202, 196, 61, 0, 202, 229, 61, 0, 202, 6, 62, 0, 202, 39, 62, 0, 202, 72, 62, 0, 202, 105, 62, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 0, 0, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 201, 128, 3, 0, 156, 5, 0, 64, 201, 96, 6, 0, 201, 128, 18, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 32, 76, 0, 201, 224, 83, 0, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 0, 94, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 201, 160, 97, 0, 5, 5, 0, 64, 201, 128, 100, 0, 201, 160, 112, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 32, 173, 0, 201, 224, 180, 0, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 130, 60, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 194, 60, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 2, 61, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 128, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 136, 0, 5, 0, 2, 5, 69, 0, 136, 0, 5, 0, 3, 5, 69, 0, 136, 0, 5, 0, 4, 5, 69, 0, 136, 0, 5, 0, 5, 5, 69, 0, 136, 0, 5, 0, 6, 5, 69, 0, 71, 0, 5, 0, 2, 5, 69, 0, 71, 0, 5, 0, 3, 5, 69, 0, 71, 0, 5, 0, 4, 5, 69, 0, 71, 0, 5, 0, 5, 5, 69, 0, 71, 0, 5, 0, 6, 5, 69, 0, 136, 0, 5, 0, 5, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 71, 0, 5, 0, 5, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 71, 0, 5, 0, 5, 5, 69, 0, 5, 0, 150, 0, 0, 0, 0, 0, 136, 0, 5, 0, 2, 5, 69, 0, 5, 0, 160, 0, 0, 0, 0, 0, 71, 0, 5, 0, 2, 5, 69, 0, 5, 0, 160, 0, 0, 0, 0, 0, 136, 0, 5, 0, 3, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 71, 0, 5, 0, 3, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 136, 0, 5, 0, 3, 5, 69, 0, 5, 0, 142, 0, 0, 0, 0, 0, 71, 0, 5, 0, 3, 5, 69, 0, 5, 0, 142, 0, 0, 0, 0, 0, 136, 0, 5, 0, 3, 5, 69, 0, 5, 0, 160, 0, 0, 0, 0, 0, 71, 0, 5, 0, 3, 5, 69, 0, 5, 0, 160, 0, 0, 0, 0, 0, 136, 0, 5, 0, 3, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 71, 0, 5, 0, 3, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 136, 0, 5, 0, 4, 5, 69, 0, 5, 0, 160, 0, 0, 0, 0, 0, 71, 0, 5, 0, 4, 5, 69, 0, 5, 0, 160, 0, 0, 0, 0, 0, 136, 0, 5, 0, 4, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 71, 0, 5, 0, 4, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 136, 0, 5, 0, 5, 5, 69, 0, 5, 0, 142, 0, 0, 0, 0, 0, 71, 0, 5, 0, 5, 5, 69, 0, 5, 0, 142, 0, 0, 0, 0, 0, 136, 0, 5, 0, 5, 5, 69, 0, 5, 0, 150, 0, 0, 0, 0, 0, 136, 0, 5, 0, 6, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 71, 0, 5, 0, 6, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 136, 0, 5, 0, 6, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 71, 0, 5, 0, 6, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 136, 0, 5, 0, 6, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 71, 0, 5, 0, 6, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 136, 0, 5, 0, 2, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 71, 0, 5, 0, 2, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 136, 0, 5, 0, 3, 5, 69, 0, 5, 0, 202, 0, 0, 0, 0, 0, 71, 0, 5, 0, 3, 5, 69, 0, 5, 0, 202, 0, 0, 0, 0, 0, 136, 0, 5, 0, 4, 5, 69, 0, 5, 0, 202, 0, 0, 0, 0, 0, 71, 0, 5, 0, 4, 5, 69, 0, 5, 0, 202, 0, 0, 0, 0, 0, 136, 0, 5, 0, 5, 5, 69, 0, 5, 0, 164, 0, 0, 0, 0, 0, 71, 0, 5, 0, 5, 5, 69, 0, 5, 0, 164, 0, 0, 0, 0, 0, 136, 0, 5, 0, 2, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 71, 0, 5, 0, 2, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 136, 0, 5, 0, 2, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 71, 0, 5, 0, 2, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 136, 0, 5, 0, 2, 5, 69, 0, 5, 0, 210, 0, 0, 0, 0, 0, 71, 0, 5, 0, 2, 5, 69, 0, 5, 0, 210, 0, 0, 0, 0, 0, 136, 0, 5, 0, 3, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 71, 0, 5, 0, 3, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 136, 0, 5, 0, 3, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 71, 0, 5, 0, 3, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 136, 0, 5, 0, 3, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 71, 0, 5, 0, 3, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 136, 0, 5, 0, 3, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 71, 0, 5, 0, 3, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 136, 0, 5, 0, 3, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 71, 0, 5, 0, 3, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 136, 0, 5, 0, 4, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 71, 0, 5, 0, 4, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 136, 0, 5, 0, 4, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 71, 0, 5, 0, 4, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 136, 0, 5, 0, 4, 5, 69, 0, 5, 0, 210, 0, 0, 0, 0, 0, 71, 0, 5, 0, 4, 5, 69, 0, 5, 0, 210, 0, 0, 0, 0, 0, 136, 0, 5, 0, 4, 5, 69, 0, 5, 0, 204, 0, 0, 0, 0, 0, 71, 0, 5, 0, 4, 5, 69, 0, 5, 0, 204, 0, 0, 0, 0, 0, 136, 0, 5, 0, 5, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 71, 0, 5, 0, 5, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 136, 0, 5, 0, 6, 5, 69, 0, 5, 0, 142, 0, 0, 0, 0, 0, 71, 0, 5, 0, 6, 5, 69, 0, 5, 0, 142, 0, 0, 0, 0, 0, 136, 0, 5, 0, 6, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 71, 0, 5, 0, 6, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 136, 0, 5, 0, 6, 5, 69, 0, 5, 0, 210, 0, 0, 0, 0, 0, 71, 0, 5, 0, 6, 5, 69, 0, 5, 0, 210, 0, 0, 0, 0, 0, 71, 0, 5, 0, 4, 5, 69, 0, 5, 0, 150, 0, 0, 0, 0, 0, 71, 0, 5, 0, 5, 5, 69, 0, 5, 0, 146, 0, 0, 0, 0, 0, 136, 0, 5, 0, 5, 5, 69, 0, 5, 0, 138, 0, 0, 0, 0, 0, 71, 0, 5, 0, 5, 5, 69, 0, 5, 0, 138, 0, 0, 0, 0, 0, 136, 0, 5, 0, 5, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 71, 0, 5, 0, 5, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 136, 0, 5, 0, 5, 5, 69, 0, 5, 0, 182, 0, 0, 0, 0, 0, 71, 0, 5, 0, 5, 5, 69, 0, 5, 0, 182, 0, 0, 0, 0, 0, 136, 0, 5, 0, 5, 5, 69, 0, 5, 0, 154, 0, 0, 0, 0, 0, 71, 0, 5, 0, 5, 5, 69, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 69, 0, 70, 0, 5, 0, 3, 5, 69, 0, 70, 0, 5, 0, 4, 5, 69, 0, 70, 0, 5, 0, 5, 5, 69, 0, 70, 0, 5, 0, 6, 5, 69, 0, 70, 0, 5, 0, 5, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 69, 0, 5, 0, 160, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 69, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 69, 0, 5, 0, 160, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 69, 0, 5, 0, 160, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 5, 69, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 5, 69, 0, 5, 0, 150, 0, 0, 0, 0, 0, 70, 0, 5, 0, 6, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 5, 0, 6, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 6, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 69, 0, 5, 0, 202, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 69, 0, 5, 0, 202, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 5, 69, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 69, 0, 5, 0, 210, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 69, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 69, 0, 5, 0, 144, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 69, 0, 5, 0, 210, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 69, 0, 5, 0, 204, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 5, 69, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 6, 5, 69, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 5, 0, 6, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 5, 0, 6, 5, 69, 0, 5, 0, 210, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 5, 69, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 5, 69, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 5, 69, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 5, 69, 0, 5, 0, 154, 0, 0, 0, 0, 0, 0, 0, 0, 136, 5, 2, 55, 5, 156, 0, 0, 71, 5, 2, 55, 5, 156, 0, 0, 136, 5, 2, 55, 5, 196, 0, 0, 71, 5, 2, 55, 5, 196, 0, 0, 136, 5, 2, 55, 5, 210, 0, 0, 71, 5, 2, 55, 5, 210, 0, 0, 136, 5, 2, 63, 5, 136, 0, 0, 71, 5, 2, 63, 5, 136, 0, 0, 136, 5, 2, 63, 5, 156, 0, 0, 71, 5, 2, 63, 5, 156, 0, 0, 136, 5, 2, 67, 5, 156, 0, 0, 71, 5, 2, 67, 5, 156, 0, 0, 136, 5, 2, 67, 5, 196, 0, 0, 71, 5, 2, 67, 5, 196, 0, 0, 136, 5, 2, 67, 5, 210, 0, 0, 71, 5, 2, 67, 5, 210, 0, 0, 136, 5, 3, 67, 5, 156, 0, 0, 71, 5, 3, 67, 5, 156, 0, 0, 136, 5, 4, 67, 5, 136, 0, 0, 71, 5, 4, 67, 5, 136, 0, 0, 136, 5, 4, 67, 5, 156, 0, 0, 71, 5, 4, 67, 5, 156, 0, 0, 136, 5, 5, 67, 5, 154, 0, 0, 71, 5, 5, 67, 5, 154, 0, 0, 136, 5, 5, 67, 5, 196, 0, 0, 71, 5, 5, 67, 5, 196, 0, 0, 139, 5, 4, 69, 5, 156, 0, 0, 139, 5, 4, 69, 5, 196, 0, 0, 139, 5, 4, 69, 5, 210, 0, 0, 74, 5, 4, 69, 5, 156, 0, 0, 73, 5, 4, 69, 5, 156, 0, 0, 74, 5, 4, 69, 5, 196, 0, 0, 73, 5, 4, 69, 5, 196, 0, 0, 74, 5, 4, 69, 5, 210, 0, 0, 73, 5, 4, 69, 5, 210, 0, 0, 136, 5, 2, 69, 5, 144, 0, 0, 71, 5, 2, 69, 5, 144, 0, 0, 136, 5, 3, 69, 5, 142, 0, 0, 71, 5, 3, 69, 5, 142, 0, 0, 136, 5, 3, 69, 5, 140, 0, 0, 71, 5, 3, 69, 5, 140, 0, 0, 136, 5, 3, 69, 5, 156, 0, 0, 71, 5, 3, 69, 5, 156, 0, 0, 136, 5, 3, 69, 5, 160, 0, 0, 71, 5, 3, 69, 5, 160, 0, 0, 136, 5, 3, 69, 5, 144, 0, 0, 71, 5, 3, 69, 5, 144, 0, 0, 136, 5, 3, 69, 5, 136, 0, 0, 71, 5, 3, 69, 5, 136, 0, 0, 136, 5, 2, 69, 5, 156, 0, 0, 71, 5, 2, 69, 5, 156, 0, 0, 136, 5, 2, 69, 5, 196, 0, 0, 71, 5, 2, 69, 5, 196, 0, 0, 136, 5, 2, 69, 5, 210, 0, 0, 71, 5, 2, 69, 5, 210, 0, 0, 136, 5, 2, 69, 5, 160, 0, 0, 71, 5, 2, 69, 5, 160, 0, 0, 136, 5, 2, 69, 5, 204, 0, 0, 71, 5, 2, 69, 5, 204, 0, 0, 136, 5, 3, 69, 5, 164, 0, 0, 71, 5, 3, 69, 5, 164, 0, 0, 136, 5, 2, 79, 5, 142, 0, 0, 71, 5, 2, 79, 5, 142, 0, 0, 136, 5, 2, 79, 5, 144, 0, 0, 71, 5, 2, 79, 5, 144, 0, 0, 136, 5, 2, 79, 5, 156, 0, 0, 71, 5, 2, 79, 5, 156, 0, 0, 136, 5, 2, 79, 5, 196, 0, 0, 71, 5, 2, 79, 5, 196, 0, 0, 136, 5, 2, 79, 5, 150, 0, 0, 71, 5, 2, 79, 5, 150, 0, 0, 136, 5, 2, 79, 5, 160, 0, 0, 71, 5, 2, 79, 5, 160, 0, 0, 136, 5, 2, 79, 5, 206, 0, 0, 71, 5, 2, 79, 5, 206, 0, 0, 71, 5, 2, 79, 5, 210, 0, 0, 136, 5, 2, 81, 5, 136, 0, 0, 71, 5, 2, 81, 5, 136, 0, 0, 136, 5, 2, 81, 5, 142, 0, 0, 71, 5, 2, 81, 5, 142, 0, 0, 136, 5, 2, 81, 5, 160, 0, 0, 71, 5, 2, 81, 5, 160, 0, 0, 136, 5, 2, 81, 5, 144, 0, 0, 71, 5, 2, 81, 5, 144, 0, 0, 136, 5, 2, 81, 5, 202, 0, 0, 71, 5, 2, 81, 5, 202, 0, 0, 136, 5, 2, 81, 5, 156, 0, 0, 71, 5, 2, 81, 5, 156, 0, 0, 136, 5, 2, 81, 5, 196, 0, 0, 71, 5, 2, 81, 5, 196, 0, 0, 136, 5, 2, 81, 5, 136, 0, 0, 5, 156, 0, 0, 71, 5, 2, 81, 5, 136, 0, 0, 5, 156, 0, 0, 136, 5, 2, 81, 5, 144, 0, 0, 5, 156, 0, 0, 71, 5, 2, 81, 5, 144, 0, 0, 5, 156, 0, 0, 136, 5, 2, 81, 5, 196, 0, 0, 5, 156, 0, 0, 71, 5, 2, 81, 5, 196, 0, 0, 5, 156, 0, 0, 70, 5, 2, 55, 5, 156, 0, 0, 5, 5, 2, 55, 5, 156, 0, 0, 70, 5, 2, 55, 5, 196, 0, 0, 5, 5, 2, 55, 5, 196, 0, 0, 70, 5, 2, 55, 5, 210, 0, 0, 5, 5, 2, 55, 5, 210, 0, 0, 70, 5, 2, 63, 5, 136, 0, 0, 5, 5, 2, 63, 5, 136, 0, 0, 70, 5, 2, 63, 5, 156, 0, 0, 5, 5, 2, 63, 5, 156, 0, 0, 70, 5, 2, 67, 5, 156, 0, 0, 5, 5, 2, 67, 5, 156, 0, 0, 70, 5, 2, 67, 5, 196, 0, 0, 5, 5, 2, 67, 5, 196, 0, 0, 70, 5, 2, 67, 5, 210, 0, 0, 5, 5, 2, 67, 5, 210, 0, 0, 70, 5, 3, 67, 5, 156, 0, 0, 5, 5, 3, 67, 5, 156, 0, 0, 70, 5, 4, 67, 5, 136, 0, 0, 5, 5, 4, 67, 5, 136, 0, 0, 70, 5, 4, 67, 5, 156, 0, 0, 5, 5, 4, 67, 5, 156, 0, 0, 70, 5, 5, 67, 5, 154, 0, 0, 5, 5, 5, 67, 5, 154, 0, 0, 70, 5, 5, 67, 5, 196, 0, 0, 5, 5, 5, 67, 5, 196, 0, 0, 72, 5, 4, 69, 5, 156, 0, 0, 71, 5, 4, 69, 5, 156, 0, 0, 72, 5, 4, 69, 5, 196, 0, 0, 71, 5, 4, 69, 5, 196, 0, 0, 72, 5, 4, 69, 5, 210, 0, 0, 71, 5, 4, 69, 5, 210, 0, 0, 70, 5, 4, 69, 5, 156, 0, 0, 5, 5, 4, 69, 5, 156, 0, 0, 70, 5, 4, 69, 5, 196, 0, 0, 5, 5, 4, 69, 5, 196, 0, 0, 70, 5, 4, 69, 5, 210, 0, 0, 5, 5, 4, 69, 5, 210, 0, 0, 193, 5, 5, 69, 5, 136, 0, 0, 193, 5, 5, 69, 5, 150, 0, 0, 70, 5, 2, 69, 5, 144, 0, 0, 5, 5, 2, 69, 5, 144, 0, 0, 70, 5, 3, 69, 5, 142, 0, 0, 5, 5, 3, 69, 5, 142, 0, 0, 70, 5, 3, 69, 5, 140, 0, 0, 5, 5, 3, 69, 5, 140, 0, 0, 70, 5, 3, 69, 5, 156, 0, 0, 5, 5, 3, 69, 5, 156, 0, 0, 70, 5, 3, 69, 5, 160, 0, 0, 5, 5, 3, 69, 5, 160, 0, 0, 193, 2, 5, 69, 5, 160, 0, 0, 193, 3, 5, 69, 5, 136, 0, 0, 193, 3, 5, 69, 5, 142, 0, 0, 193, 3, 5, 69, 5, 160, 0, 0, 193, 3, 5, 69, 5, 144, 0, 0, 193, 4, 5, 69, 5, 160, 0, 0, 193, 4, 5, 69, 5, 144, 0, 0, 193, 5, 5, 69, 5, 142, 0, 0, 193, 6, 5, 69, 5, 136, 0, 0, 193, 6, 5, 69, 5, 156, 0, 0, 193, 6, 5, 69, 5, 144, 0, 0, 70, 5, 3, 69, 5, 144, 0, 0, 5, 5, 3, 69, 5, 144, 0, 0, 193, 2, 5, 69, 5, 144, 0, 0, 70, 5, 3, 69, 5, 136, 0, 0, 5, 5, 3, 69, 5, 136, 0, 0, 193, 3, 5, 69, 5, 202, 0, 0, 193, 4, 5, 69, 5, 202, 0, 0, 193, 5, 5, 69, 5, 164, 0, 0, 70, 5, 2, 69, 5, 156, 0, 0, 5, 5, 2, 69, 5, 156, 0, 0, 70, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 70, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 70, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 70, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 70, 5, 3, 69, 5, 164, 0, 0, 5, 5, 3, 69, 5, 164, 0, 0, 193, 2, 5, 69, 5, 136, 0, 0, 193, 2, 5, 69, 5, 196, 0, 0, 193, 2, 5, 69, 5, 210, 0, 0, 193, 3, 5, 69, 5, 156, 0, 0, 193, 3, 5, 69, 5, 196, 0, 0, 193, 3, 5, 69, 5, 136, 0, 0, 5, 156, 0, 0, 193, 3, 5, 69, 5, 144, 0, 0, 5, 156, 0, 0, 193, 3, 5, 69, 5, 196, 0, 0, 5, 156, 0, 0, 193, 4, 5, 69, 5, 156, 0, 0, 193, 4, 5, 69, 5, 196, 0, 0, 193, 4, 5, 69, 5, 210, 0, 0, 193, 4, 5, 69, 5, 204, 0, 0, 193, 5, 5, 69, 5, 156, 0, 0, 193, 6, 5, 69, 5, 142, 0, 0, 193, 6, 5, 69, 5, 196, 0, 0, 193, 6, 5, 69, 5, 210, 0, 0, 193, 4, 5, 69, 5, 150, 0, 0, 193, 5, 5, 69, 5, 146, 0, 0, 193, 5, 5, 69, 5, 138, 0, 0, 193, 5, 5, 69, 5, 196, 0, 0, 193, 5, 5, 69, 5, 182, 0, 0, 193, 5, 5, 69, 5, 154, 0, 0, 70, 5, 2, 79, 5, 142, 0, 0, 5, 5, 2, 79, 5, 142, 0, 0, 70, 5, 2, 79, 5, 144, 0, 0, 5, 5, 2, 79, 5, 144, 0, 0, 70, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 70, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 70, 5, 2, 79, 5, 150, 0, 0, 5, 5, 2, 79, 5, 150, 0, 0, 70, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 70, 5, 2, 79, 5, 206, 0, 0, 5, 5, 2, 79, 5, 206, 0, 0, 5, 5, 2, 79, 5, 210, 0, 0, 70, 5, 2, 81, 5, 136, 0, 0, 5, 5, 2, 81, 5, 136, 0, 0, 70, 5, 2, 81, 5, 142, 0, 0, 5, 5, 2, 81, 5, 142, 0, 0, 70, 5, 2, 81, 5, 160, 0, 0, 5, 5, 2, 81, 5, 160, 0, 0, 70, 5, 2, 81, 5, 144, 0, 0, 5, 5, 2, 81, 5, 144, 0, 0, 70, 5, 2, 81, 5, 202, 0, 0, 5, 5, 2, 81, 5, 202, 0, 0, 70, 5, 2, 81, 5, 156, 0, 0, 5, 5, 2, 81, 5, 156, 0, 0, 70, 5, 2, 81, 5, 196, 0, 0, 5, 5, 2, 81, 5, 196, 0, 0, 70, 5, 2, 81, 5, 136, 0, 0, 5, 156, 0, 0, 5, 5, 2, 81, 5, 136, 0, 0, 5, 156, 0, 0, 70, 5, 2, 81, 5, 144, 0, 0, 5, 156, 0, 0, 5, 5, 2, 81, 5, 144, 0, 0, 5, 156, 0, 0, 70, 5, 2, 81, 5, 196, 0, 0, 5, 156, 0, 0, 5, 5, 2, 81, 5, 196, 0, 0, 5, 156, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 54, 156, 5, 7, 0, 4, 30, 12, 0, 4, 30, 0, 192, 197, 162, 5, 30, 0, 192, 197, 226, 6, 30, 1, 192, 197, 34, 7, 30, 1, 192, 197, 98, 66, 0, 2, 247, 136, 5, 98, 0, 2, 247, 71, 5, 2, 30, 197, 162, 3, 30, 0, 192, 197, 98, 0, 62, 156, 5, 5, 0, 85, 30, 9, 0, 85, 30, 1, 192, 197, 226, 86, 30, 2, 192, 197, 34, 87, 30, 2, 192, 197, 98, 80, 0, 2, 255, 136, 5, 112, 0, 2, 255, 71, 5, 84, 30, 1, 192, 197, 162, 0, 66, 156, 5, 23, 0, 6, 30, 42, 0, 86, 30, 20, 0, 125, 30, 9, 0, 125, 30, 5, 192, 197, 226, 126, 30, 6, 192, 197, 34, 127, 30, 6, 192, 197, 98, 86, 30, 5, 192, 197, 34, 87, 30, 5, 192, 197, 98, 124, 30, 5, 192, 197, 162, 31, 30, 9, 0, 31, 30, 4, 192, 197, 98, 84, 30, 4, 192, 197, 162, 85, 30, 4, 192, 197, 226, 6, 30, 3, 192, 197, 162, 7, 30, 3, 192, 197, 226, 30, 30, 4, 192, 197, 34, 112, 0, 22, 0, 3, 30, 9, 0, 3, 30, 2, 192, 197, 226, 4, 30, 3, 192, 197, 34, 5, 30, 3, 192, 197, 98, 112, 0, 255, 255, 4, 67, 71, 5, 118, 0, 255, 255, 5, 67, 71, 5, 2, 30, 2, 192, 197, 162, 86, 0, 12, 0, 86, 0, 255, 255, 5, 67, 136, 5, 98, 0, 255, 255, 2, 67, 71, 5, 102, 0, 255, 255, 3, 67, 71, 5, 66, 0, 255, 255, 2, 67, 136, 5, 70, 0, 255, 255, 3, 67, 136, 5, 80, 0, 255, 255, 4, 67, 136, 5, 0, 68, 156, 5, 0, 0, 116, 0, 26, 2, 206, 0, 103, 30, 103, 0, 146, 30, 50, 0, 243, 30, 25, 0, 247, 30, 11, 0, 247, 30, 20, 192, 198, 66, 248, 30, 20, 192, 198, 130, 249, 30, 20, 192, 198, 194, 42, 33, 198, 129, 243, 30, 19, 192, 198, 66, 244, 30, 19, 192, 198, 130, 245, 30, 19, 192, 198, 194, 246, 30, 20, 192, 198, 2, 149, 30, 12, 0, 149, 30, 18, 192, 198, 66, 151, 30, 18, 192, 198, 130, 153, 30, 18, 192, 198, 194, 242, 30, 19, 192, 198, 2, 146, 30, 17, 192, 198, 130, 147, 30, 17, 192, 198, 194, 148, 30, 18, 192, 198, 2, 110, 30, 26, 0, 142, 30, 12, 0, 142, 30, 16, 192, 198, 130, 143, 30, 16, 192, 198, 194, 144, 30, 17, 192, 198, 2, 145, 30, 17, 192, 198, 66, 110, 30, 15, 192, 198, 130, 111, 30, 15, 192, 198, 194, 112, 30, 16, 192, 198, 2, 113, 30, 16, 192, 198, 66, 106, 30, 12, 0, 106, 30, 14, 192, 198, 130, 107, 30, 14, 192, 198, 194, 108, 30, 15, 192, 198, 2, 109, 30, 15, 192, 198, 66, 103, 30, 13, 192, 198, 99, 104, 30, 13, 192, 198, 195, 105, 30, 14, 192, 198, 35, 32, 30, 51, 0, 53, 30, 26, 0, 99, 30, 12, 0, 99, 30, 12, 192, 198, 2, 100, 30, 12, 192, 198, 67, 101, 30, 12, 192, 198, 163, 102, 30, 13, 192, 198, 3, 53, 30, 11, 192, 198, 2, 96, 30, 11, 192, 198, 66, 97, 30, 11, 192, 198, 130, 98, 30, 11, 192, 198, 194, 49, 30, 12, 0, 49, 30, 10, 192, 198, 2, 50, 30, 10, 192, 198, 66, 51, 30, 10, 192, 198, 130, 52, 30, 10, 192, 198, 194, 32, 30, 14, 192, 197, 226, 33, 30, 15, 192, 197, 34, 48, 30, 9, 192, 198, 194, 13, 30, 23, 0, 16, 30, 12, 0, 16, 30, 13, 192, 197, 226, 17, 30, 14, 192, 197, 34, 18, 30, 14, 192, 197, 98, 19, 30, 14, 192, 197, 162, 13, 30, 13, 192, 197, 34, 14, 30, 13, 192, 197, 98, 15, 30, 13, 192, 197, 162, 51, 2, 12, 0, 51, 2, 9, 192, 198, 130, 10, 30, 12, 192, 197, 98, 11, 30, 12, 192, 197, 162, 12, 30, 12, 192, 197, 226, 26, 2, 8, 192, 198, 194, 27, 2, 9, 192, 198, 2, 50, 2, 9, 192, 198, 66, 90, 1, 101, 0, 120, 1, 51, 0, 230, 1, 26, 0, 244, 1, 12, 0, 244, 1, 11, 192, 197, 226, 245, 1, 12, 192, 197, 34, 24, 2, 8, 192, 198, 66, 25, 2, 8, 192, 198, 130, 230, 1, 11, 192, 197, 98, 231, 1, 11, 192, 197, 162, 232, 1, 7, 192, 198, 194, 233, 1, 8, 192, 198, 2, 123, 1, 12, 0, 123, 1, 6, 192, 198, 194, 124, 1, 7, 192, 198, 2, 125, 1, 7, 192, 198, 66, 126, 1, 7, 192, 198, 130, 120, 1, 6, 192, 198, 2, 121, 1, 6, 192, 198, 66, 122, 1, 6, 192, 198, 130, 97, 1, 23, 0, 100, 1, 12, 0, 100, 1, 5, 192, 198, 2, 101, 1, 5, 192, 198, 66, 118, 1, 5, 192, 198, 130, 119, 1, 5, 192, 198, 194, 97, 1, 4, 192, 198, 66, 98, 1, 4, 192, 198, 130, 99, 1, 4, 192, 198, 194, 93, 1, 12, 0, 93, 1, 3, 192, 198, 66, 94, 1, 3, 192, 198, 130, 95, 1, 3, 192, 198, 194, 96, 1, 4, 192, 198, 2, 90, 1, 2, 192, 198, 130, 91, 1, 2, 192, 198, 194, 92, 1, 3, 192, 198, 2, 221, 0, 51, 0, 30, 1, 26, 0, 34, 1, 12, 0, 34, 1, 10, 192, 197, 226, 35, 1, 11, 192, 197, 34, 54, 1, 2, 192, 198, 2, 55, 1, 2, 192, 198, 66, 30, 1, 9, 192, 197, 226, 31, 1, 10, 192, 197, 34, 32, 1, 10, 192, 197, 98, 33, 1, 10, 192, 197, 162, 14, 1, 12, 0, 14, 1, 8, 192, 197, 226, 15, 1, 9, 192, 197, 34, 28, 1, 9, 192, 197, 98, 29, 1, 9, 192, 197, 162, 221, 0, 1, 192, 198, 66, 253, 0, 1, 192, 198, 130, 255, 0, 1, 192, 198, 194, 100, 0, 54, 0, 115, 0, 12, 0, 115, 0, 0, 192, 198, 193, 116, 0, 0, 192, 198, 225, 121, 0, 1, 192, 198, 1, 122, 0, 1, 192, 198, 33, 100, 0, 255, 255, 2, 69, 71, 5, 103, 0, 3, 0, 107, 0, 0, 192, 198, 161, 199, 127, 3, 69, 71, 5, 4, 30, 12, 0, 4, 30, 7, 192, 197, 226, 5, 30, 8, 192, 197, 34, 6, 30, 8, 192, 197, 98, 7, 30, 8, 192, 197, 162, 66, 0, 255, 255, 4, 69, 74, 5, 98, 0, 255, 255, 4, 69, 73, 5, 2, 30, 7, 192, 197, 98, 3, 30, 7, 192, 197, 162, 83, 0, 11, 0, 83, 0, 198, 161, 84, 0, 0, 192, 198, 65, 89, 0, 0, 192, 198, 97, 90, 0, 0, 192, 198, 129, 68, 0, 255, 255, 2, 69, 136, 5, 71, 0, 2, 0, 75, 0, 198, 129, 195, 127, 3, 69, 136, 5, 66, 0, 255, 255, 4, 69, 139, 5, 2, 30, 6, 192, 197, 162, 4, 30, 6, 192, 197, 226, 6, 30, 7, 192, 197, 34, 0, 78, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 17, 192, 197, 226, 41, 30, 18, 192, 197, 34, 42, 30, 18, 192, 197, 98, 43, 30, 18, 192, 197, 162, 150, 30, 18, 192, 197, 226, 36, 30, 16, 192, 197, 226, 37, 30, 17, 192, 197, 34, 38, 30, 17, 192, 197, 98, 39, 30, 17, 192, 197, 162, 30, 2, 12, 0, 30, 2, 15, 192, 197, 226, 31, 2, 16, 192, 197, 34, 34, 30, 16, 192, 197, 98, 35, 30, 16, 192, 197, 162, 72, 0, 255, 255, 2, 79, 136, 5, 104, 0, 255, 255, 2, 79, 71, 5, 36, 1, 15, 192, 197, 98, 37, 1, 15, 192, 197, 162, 0, 80, 156, 5, 21, 0, 25, 2, 37, 0, 100, 30, 20, 0, 103, 30, 9, 0, 103, 30, 23, 192, 197, 195, 104, 30, 24, 192, 197, 35, 105, 30, 24, 192, 197, 131, 100, 30, 22, 192, 197, 163, 101, 30, 23, 192, 197, 3, 102, 30, 23, 192, 197, 99, 25, 2, 21, 192, 197, 98, 96, 30, 21, 192, 197, 162, 97, 30, 21, 192, 197, 226, 98, 30, 22, 192, 197, 34, 99, 30, 22, 192, 197, 98, 93, 1, 20, 0, 96, 1, 9, 0, 96, 1, 20, 192, 197, 162, 97, 1, 20, 192, 197, 226, 24, 2, 21, 192, 197, 34, 93, 1, 19, 192, 197, 226, 94, 1, 20, 192, 197, 34, 95, 1, 20, 192, 197, 98, 83, 0, 255, 255, 2, 81, 136, 5, 115, 0, 255, 255, 2, 81, 71, 5, 90, 1, 19, 192, 197, 34, 91, 1, 19, 192, 197, 98, 92, 1, 19, 192, 197, 162, 0, 54, 5, 5, 7, 0, 4, 30, 12, 0, 4, 30, 25, 192, 197, 98, 5, 30, 25, 192, 197, 162, 6, 30, 25, 192, 197, 226, 7, 30, 26, 192, 197, 34, 66, 0, 2, 247, 70, 5, 98, 0, 2, 247, 5, 5, 2, 30, 24, 192, 197, 226, 3, 30, 25, 192, 197, 34, 0, 62, 5, 5, 5, 0, 85, 30, 9, 0, 85, 30, 26, 192, 197, 162, 86, 30, 26, 192, 197, 226, 87, 30, 27, 192, 197, 34, 80, 0, 2, 255, 70, 5, 112, 0, 2, 255, 5, 5, 84, 30, 26, 192, 197, 98, 0, 66, 5, 5, 23, 0, 6, 30, 42, 0, 86, 30, 20, 0, 125, 30, 9, 0, 125, 30, 30, 192, 197, 162, 126, 30, 30, 192, 197, 226, 127, 30, 31, 192, 197, 34, 86, 30, 29, 192, 197, 226, 87, 30, 30, 192, 197, 34, 124, 30, 30, 192, 197, 98, 31, 30, 9, 0, 31, 30, 29, 192, 197, 34, 84, 30, 29, 192, 197, 98, 85, 30, 29, 192, 197, 162, 6, 30, 28, 192, 197, 98, 7, 30, 28, 192, 197, 162, 30, 30, 28, 192, 197, 226, 112, 0, 22, 0, 3, 30, 9, 0, 3, 30, 27, 192, 197, 162, 4, 30, 27, 192, 197, 226, 5, 30, 28, 192, 197, 34, 112, 0, 255, 255, 4, 67, 5, 5, 118, 0, 255, 255, 5, 67, 5, 5, 2, 30, 27, 192, 197, 98, 86, 0, 12, 0, 86, 0, 255, 255, 5, 67, 70, 5, 98, 0, 255, 255, 2, 67, 5, 5, 102, 0, 255, 255, 3, 67, 5, 5, 66, 0, 255, 255, 2, 67, 70, 5, 70, 0, 255, 255, 3, 67, 70, 5, 80, 0, 255, 255, 4, 67, 70, 5, 0, 68, 5, 5, 0, 0, 116, 0, 26, 2, 207, 0, 103, 30, 104, 0, 146, 30, 51, 0, 243, 30, 26, 0, 247, 30, 12, 0, 247, 30, 50, 192, 197, 130, 248, 30, 31, 192, 198, 2, 249, 30, 50, 192, 197, 194, 42, 33, 21, 192, 198, 1, 243, 30, 50, 192, 197, 2, 244, 30, 30, 192, 198, 130, 245, 30, 50, 192, 197, 66, 246, 30, 30, 192, 198, 194, 149, 30, 12, 0, 149, 30, 49, 192, 197, 66, 151, 30, 49, 192, 197, 130, 153, 30, 49, 192, 197, 194, 242, 30, 30, 192, 198, 66, 146, 30, 29, 192, 198, 194, 147, 30, 49, 192, 197, 2, 148, 30, 30, 192, 198, 2, 110, 30, 26, 0, 142, 30, 12, 0, 142, 30, 29, 192, 198, 66, 143, 30, 48, 192, 197, 130, 144, 30, 29, 192, 198, 130, 145, 30, 48, 192, 197, 194, 110, 30, 28, 192, 198, 194, 111, 30, 48, 192, 197, 2, 112, 30, 29, 192, 198, 2, 113, 30, 48, 192, 197, 66, 106, 30, 12, 0, 106, 30, 28, 192, 198, 66, 107, 30, 47, 192, 197, 130, 108, 30, 28, 192, 198, 130, 109, 30, 47, 192, 197, 194, 103, 30, 46, 192, 197, 195, 104, 30, 27, 192, 198, 227, 105, 30, 47, 192, 197, 35, 32, 30, 51, 0, 53, 30, 26, 0, 99, 30, 12, 0, 99, 30, 46, 192, 197, 34, 100, 30, 27, 192, 198, 35, 101, 30, 46, 192, 197, 99, 102, 30, 27, 192, 198, 131, 53, 30, 45, 192, 197, 162, 96, 30, 26, 192, 198, 162, 97, 30, 45, 192, 197, 226, 98, 30, 26, 192, 198, 226, 49, 30, 12, 0, 49, 30, 45, 192, 197, 34, 50, 30, 26, 192, 198, 34, 51, 30, 45, 192, 197, 98, 52, 30, 26, 192, 198, 98, 32, 30, 44, 192, 197, 162, 33, 30, 44, 192, 197, 226, 48, 30, 25, 192, 198, 226, 13, 30, 23, 0, 16, 30, 12, 0, 16, 30, 43, 192, 197, 162, 17, 30, 43, 192, 197, 226, 18, 30, 44, 192, 197, 34, 19, 30, 44, 192, 197, 98, 13, 30, 42, 192, 197, 226, 14, 30, 43, 192, 197, 34, 15, 30, 43, 192, 197, 98, 51, 2, 12, 0, 51, 2, 41, 192, 197, 226, 10, 30, 42, 192, 197, 34, 11, 30, 42, 192, 197, 98, 12, 30, 42, 192, 197, 162, 26, 2, 25, 192, 198, 98, 27, 2, 41, 192, 197, 162, 50, 2, 25, 192, 198, 162, 90, 1, 101, 0, 120, 1, 51, 0, 230, 1, 26, 0, 244, 1, 12, 0, 244, 1, 40, 192, 197, 226, 245, 1, 41, 192, 197, 34, 24, 2, 25, 192, 198, 34, 25, 2, 41, 192, 197, 98, 230, 1, 40, 192, 197, 34, 231, 1, 40, 192, 197, 98, 232, 1, 24, 192, 198, 226, 233, 1, 40, 192, 197, 162, 123, 1, 12, 0, 123, 1, 24, 192, 198, 98, 124, 1, 39, 192, 197, 162, 125, 1, 24, 192, 198, 162, 126, 1, 39, 192, 197, 226, 120, 1, 23, 192, 198, 226, 121, 1, 24, 192, 198, 34, 122, 1, 39, 192, 197, 98, 97, 1, 23, 0, 100, 1, 12, 0, 100, 1, 23, 192, 198, 98, 101, 1, 38, 192, 197, 226, 118, 1, 23, 192, 198, 162, 119, 1, 39, 192, 197, 34, 97, 1, 38, 192, 197, 98, 98, 1, 23, 192, 198, 34, 99, 1, 38, 192, 197, 162, 93, 1, 12, 0, 93, 1, 37, 192, 197, 226, 94, 1, 22, 192, 198, 162, 95, 1, 38, 192, 197, 34, 96, 1, 22, 192, 198, 226, 90, 1, 22, 192, 198, 34, 91, 1, 37, 192, 197, 162, 92, 1, 22, 192, 198, 98, 221, 0, 51, 0, 30, 1, 26, 0, 34, 1, 12, 0, 34, 1, 36, 192, 197, 226, 35, 1, 37, 192, 197, 34, 54, 1, 21, 192, 198, 226, 55, 1, 37, 192, 197, 98, 30, 1, 35, 192, 197, 226, 31, 1, 36, 192, 197, 34, 32, 1, 36, 192, 197, 98, 33, 1, 36, 192, 197, 162, 14, 1, 12, 0, 14, 1, 34, 192, 197, 226, 15, 1, 35, 192, 197, 34, 28, 1, 35, 192, 197, 98, 29, 1, 35, 192, 197, 162, 221, 0, 21, 192, 198, 162, 253, 0, 34, 192, 197, 98, 255, 0, 34, 192, 197, 162, 100, 0, 59, 0, 115, 0, 16, 0, 115, 0, 255, 255, 5, 69, 193, 3, 116, 0, 255, 255, 5, 69, 193, 4, 121, 0, 255, 255, 5, 69, 193, 5, 122, 0, 255, 255, 5, 69, 193, 6, 100, 0, 255, 255, 2, 69, 5, 5, 103, 0, 4, 0, 107, 0, 255, 255, 5, 69, 193, 2, 199, 127, 3, 69, 5, 5, 4, 30, 12, 0, 4, 30, 33, 192, 197, 98, 5, 30, 33, 192, 197, 162, 6, 30, 33, 192, 197, 226, 7, 30, 34, 192, 197, 34, 66, 0, 255, 255, 4, 69, 70, 5, 98, 0, 255, 255, 4, 69, 5, 5, 2, 30, 32, 192, 197, 226, 3, 30, 33, 192, 197, 34, 83, 0, 12, 0, 83, 0, 21, 192, 198, 33, 84, 0, 21, 192, 198, 65, 89, 0, 21, 192, 198, 97, 90, 0, 21, 192, 198, 129, 68, 0, 255, 255, 2, 69, 70, 5, 71, 0, 3, 0, 75, 0, 21, 192, 198, 1, 199, 127, 3, 69, 70, 5, 4, 30, 12, 0, 4, 30, 31, 192, 197, 226, 5, 30, 32, 192, 197, 34, 6, 30, 32, 192, 197, 98, 7, 30, 32, 192, 197, 162, 66, 0, 255, 255, 4, 69, 72, 5, 98, 0, 255, 255, 4, 69, 71, 5, 2, 30, 31, 192, 197, 98, 3, 30, 31, 192, 197, 162, 0, 78, 5, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 53, 192, 197, 130, 41, 30, 53, 192, 197, 194, 42, 30, 54, 192, 197, 2, 43, 30, 54, 192, 197, 66, 150, 30, 54, 192, 197, 130, 36, 30, 52, 192, 197, 130, 37, 30, 52, 192, 197, 194, 38, 30, 53, 192, 197, 2, 39, 30, 53, 192, 197, 66, 30, 2, 12, 0, 30, 2, 51, 192, 197, 130, 31, 2, 51, 192, 197, 194, 34, 30, 52, 192, 197, 2, 35, 30, 52, 192, 197, 66, 72, 0, 255, 255, 2, 79, 70, 5, 104, 0, 255, 255, 2, 79, 5, 5, 36, 1, 51, 192, 197, 2, 37, 1, 51, 192, 197, 66, 0, 80, 5, 5, 21, 0, 25, 2, 37, 0, 100, 30, 20, 0, 103, 30, 9, 0, 103, 30, 59, 192, 197, 99, 104, 30, 59, 192, 197, 195, 105, 30, 60, 192, 197, 35, 100, 30, 58, 192, 197, 67, 101, 30, 58, 192, 197, 163, 102, 30, 59, 192, 197, 3, 25, 2, 57, 192, 197, 2, 96, 30, 57, 192, 197, 66, 97, 30, 57, 192, 197, 130, 98, 30, 57, 192, 197, 194, 99, 30, 58, 192, 197, 2, 93, 1, 20, 0, 96, 1, 9, 0, 96, 1, 56, 192, 197, 66, 97, 1, 56, 192, 197, 130, 24, 2, 56, 192, 197, 194, 93, 1, 55, 192, 197, 130, 94, 1, 55, 192, 197, 194, 95, 1, 56, 192, 197, 2, 83, 0, 255, 255, 2, 81, 70, 5, 115, 0, 255, 255, 2, 81, 5, 5, 90, 1, 54, 192, 197, 194, 91, 1, 55, 192, 197, 2, 92, 1, 55, 192, 197, 66, 80, 0, 66, 0, 67, 0, 68, 0, 69, 0, 70, 0, 73, 0, 75, 0, 76, 0, 80, 0, 81, 0, 83, 0, 85, 0, 86, 0, 87, 0, 89, 0, 91, 0, 98, 0, 99, 0, 100, 0, 101, 0, 102, 0, 105, 0, 107, 0, 108, 0, 112, 0, 113, 0, 115, 0, 117, 0, 118, 0, 119, 0, 121, 0, 123, 0, 221, 0, 222, 0, 253, 0, 254, 0, 255, 0, 0, 1, 14, 1, 16, 1, 28, 1, 38, 1, 54, 1, 56, 1, 90, 1, 102, 1, 118, 1, 127, 1, 230, 1, 234, 1, 244, 1, 246, 1, 24, 2, 28, 2, 30, 2, 32, 2, 50, 2, 52, 2, 2, 30, 8, 30, 10, 30, 20, 30, 30, 30, 44, 30, 48, 30, 54, 30, 84, 30, 88, 30, 96, 30, 114, 30, 124, 30, 128, 30, 142, 30, 152, 30, 153, 30, 154, 30, 242, 30, 250, 30, 42, 33, 43, 33, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 26, 4, 185, 88, 186, 92, 186, 100, 32, 4, 185, 112, 38, 4, 56, 4, 185, 172, 185, 176, 185, 180, 185, 188, 156, 4, 166, 4, 185, 212, 185, 216, 185, 220, 185, 224, 185, 228, 185, 232, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 188, 4, 168, 88, 168, 92, 168, 100, 194, 4, 168, 112, 200, 4, 218, 4, 168, 172, 168, 176, 168, 180, 168, 188, 62, 5, 72, 5, 168, 212, 168, 216, 168, 220, 168, 224, 168, 228, 168, 232, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 172, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 93, 154, 93, 250, 93, 90, 94, 217, 68, 154, 138, 185, 173, 153, 173, 249, 173, 153, 174, 89, 174, 136, 14, 217, 174, 185, 213, 153, 213, 249, 213, 89, 214, 153, 229, 185, 236, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 93, 136, 93, 232, 93, 72, 94, 200, 68, 136, 138, 168, 173, 136, 173, 232, 173, 136, 174, 72, 174, 128, 14, 200, 174, 168, 213, 136, 213, 232, 213, 72, 214, 136, 229, 168, 236, 72, 230, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 89, 232, 89, 89, 91, 72, 91, 154, 94, 136, 94, 58, 95, 40, 95, 218, 93, 200, 93, 26, 95, 8, 95, 186, 94, 168, 96, 6, 8, 8, 8, 250, 101, 232, 101, 249, 106, 232, 106, 168, 184, 153, 113, 136, 113, 249, 114, 232, 114, 25, 114, 8, 114, 121, 115, 104, 115, 89, 115, 72, 115, 154, 137, 136, 137, 250, 138, 232, 138, 26, 138, 8, 138, 10, 8, 185, 168, 168, 168, 57, 175, 40, 175, 217, 173, 200, 173, 121, 174, 104, 174, 12, 8, 14, 8, 153, 189, 136, 189, 249, 190, 232, 190, 25, 190, 8, 190, 154, 193, 136, 193, 250, 193, 232, 193, 250, 194, 232, 194, 26, 194, 8, 194, 249, 202, 232, 202, 25, 202, 8, 202, 185, 208, 168, 208, 153, 214, 136, 214, 57, 215, 40, 215, 217, 213, 200, 213, 57, 214, 40, 214, 121, 214, 104, 214, 25, 215, 8, 215, 249, 221, 232, 221, 249, 229, 232, 229, 89, 230, 153, 233, 136, 233, 185, 234, 168, 234, 25, 234, 8, 234, 232, 192, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 192, 169, 192, 200, 56, 169, 72, 187, 92, 187, 100, 169, 92, 169, 100, 169, 240, 169, 136, 217, 172, 187, 72, 200, 172, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 80, 66, 4, 187, 84, 98, 4, 178, 84, 255, 5, 185, 104, 80, 4, 187, 108, 112, 4, 178, 108, 255, 5, 185, 116, 66, 4, 187, 120, 70, 4, 187, 124, 80, 4, 187, 128, 86, 4, 187, 132, 98, 4, 178, 120, 102, 4, 178, 124, 112, 4, 178, 128, 118, 4, 178, 132, 255, 5, 186, 136, 68, 4, 187, 140, 71, 2, 75, 4, 187, 148, 83, 4, 187, 152, 84, 4, 187, 156, 89, 4, 187, 160, 90, 4, 187, 164, 100, 4, 178, 140, 103, 2, 107, 4, 178, 148, 115, 4, 178, 152, 116, 4, 178, 156, 121, 4, 178, 160, 122, 4, 178, 164, 221, 4, 155, 161, 253, 4, 146, 161, 255, 4, 82, 162, 14, 5, 27, 142, 15, 5, 18, 142, 28, 5, 251, 145, 29, 5, 242, 145, 30, 5, 219, 145, 31, 5, 210, 145, 32, 5, 187, 146, 33, 5, 178, 146, 34, 5, 251, 146, 35, 5, 242, 146, 54, 5, 251, 150, 55, 5, 242, 150, 90, 5, 155, 153, 91, 5, 146, 153, 92, 5, 251, 153, 93, 5, 242, 153, 94, 5, 251, 154, 95, 5, 242, 154, 96, 5, 27, 154, 97, 5, 18, 154, 98, 5, 251, 158, 99, 5, 242, 158, 100, 5, 27, 158, 101, 5, 18, 158, 118, 5, 251, 161, 119, 5, 242, 161, 120, 5, 91, 162, 121, 5, 155, 165, 122, 5, 146, 165, 123, 5, 187, 166, 124, 5, 178, 166, 125, 5, 27, 166, 126, 5, 18, 166, 255, 5, 186, 192, 72, 4, 187, 196, 104, 4, 178, 196, 36, 5, 251, 197, 37, 5, 242, 197, 255, 5, 185, 200, 83, 4, 187, 204, 115, 4, 178, 204, 90, 5, 155, 205, 91, 5, 146, 205, 92, 5, 251, 205, 93, 5, 242, 205, 94, 5, 251, 206, 95, 5, 242, 206, 96, 5, 27, 206, 97, 5, 18, 206, 255, 5, 168, 80, 66, 4, 177, 84, 98, 4, 168, 84, 255, 5, 168, 104, 80, 4, 177, 108, 112, 4, 168, 108, 255, 5, 168, 116, 66, 4, 177, 120, 70, 4, 177, 124, 80, 4, 177, 128, 86, 4, 177, 132, 98, 4, 168, 120, 102, 4, 168, 124, 112, 4, 168, 128, 118, 4, 168, 132, 255, 5, 168, 136, 68, 4, 177, 140, 71, 2, 75, 4, 177, 148, 83, 4, 177, 152, 84, 4, 177, 156, 89, 4, 177, 160, 90, 4, 177, 164, 100, 4, 168, 140, 103, 2, 107, 4, 168, 148, 115, 4, 168, 152, 116, 4, 168, 156, 121, 4, 168, 160, 122, 4, 168, 164, 221, 4, 145, 161, 253, 4, 136, 161, 255, 4, 72, 162, 14, 5, 17, 142, 15, 5, 8, 142, 28, 5, 241, 145, 29, 5, 232, 145, 30, 5, 209, 145, 31, 5, 200, 145, 32, 5, 177, 146, 33, 5, 168, 146, 34, 5, 241, 146, 35, 5, 232, 146, 54, 5, 241, 150, 55, 5, 232, 150, 90, 5, 145, 153, 91, 5, 136, 153, 92, 5, 241, 153, 93, 5, 232, 153, 94, 5, 241, 154, 95, 5, 232, 154, 96, 5, 17, 154, 97, 5, 8, 154, 98, 5, 241, 158, 99, 5, 232, 158, 100, 5, 17, 158, 101, 5, 8, 158, 118, 5, 241, 161, 119, 5, 232, 161, 120, 5, 81, 162, 121, 5, 145, 165, 122, 5, 136, 165, 123, 5, 177, 166, 124, 5, 168, 166, 125, 5, 17, 166, 126, 5, 8, 166, 255, 5, 168, 192, 72, 4, 177, 196, 104, 4, 168, 196, 36, 5, 241, 197, 37, 5, 232, 197, 255, 5, 168, 200, 83, 4, 177, 204, 115, 4, 168, 204, 90, 5, 145, 205, 91, 5, 136, 205, 92, 5, 241, 205, 93, 5, 232, 205, 94, 5, 241, 206, 95, 5, 232, 206, 96, 5, 17, 206, 97, 5, 8, 206, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 207, 0, 0, 16, 24, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 96, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 52, 0, 0, 0, 52, 0, 0, 0, 4, 25, 0, 0, 4, 25, 0, 0, 4, 25, 0, 0, 4, 25, 0, 0, 72, 25, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 132, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 91, 0, 99, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 123, 0, 115, 0, 48, 0, 48, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 107, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 64, 1, 192, 0, 192, 0, 108, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 69, 0, 60, 0, 32, 0, 219, 15, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 84, 0, 170, 170, 211, 15, 0, 32, 52, 22, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 57, 22, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 124, 0, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 128, 0, 0, 0, 128, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 76, 0, 97, 0, 111, 0, 111, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 68, 1, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 8, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 32, 0, 0, 0, 44, 0, 0, 0, 44, 1, 0, 0, 24, 0, 0, 0, 0, 0, 0, 115, 180, 255, 132, 115, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 0, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 3, 0, 69, 0, 60, 0, 32, 0, 35, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 117, 0, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 122, 0, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 250, 9, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 254, 9, 0, 0, 254, 9, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 73, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 96, 220, 38, 0, 0, 3, 61, 0, 7, 3, 0, 3, 38, 0, 1, 3, 61, 0, 7, 3, 1, 3, 38, 0, 3, 3, 61, 0, 7, 3, 3, 3, 38, 0, 65, 0, 60, 0, 60, 0, 5, 1, 60, 0, 60, 0, 60, 0, 4, 1, 38, 0, 67, 0, 60, 0, 13, 1, 60, 0, 60, 0, 60, 0, 12, 1, 38, 0, 69, 0, 60, 0, 60, 0, 25, 1, 60, 0, 60, 0, 60, 0, 24, 1, 60, 0, 60, 0, 23, 1, 60, 0, 60, 0, 60, 0, 22, 1, 38, 0, 73, 0, 60, 0, 60, 0, 47, 1, 60, 0, 60, 0, 60, 0, 46, 1, 60, 0, 60, 0, 121, 0, 60, 0, 60, 0, 60, 0, 89, 0, 38, 0, 83, 0, 60, 0, 97, 1, 60, 0, 60, 0, 60, 0, 96, 1, 38, 0, 85, 0, 60, 0, 60, 0, 115, 1, 60, 0, 60, 0, 60, 0, 114, 1, 60, 0, 60, 0, 107, 1, 60, 0, 60, 0, 60, 0, 106, 1, 38, 0, 90, 0, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 154, 38, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 212, 29, 0, 0, 212, 29, 0, 0, 212, 29, 0, 0, 212, 29, 0, 0, 240, 31, 0, 0, 240, 31, 0, 0, 98, 34, 0, 0, 98, 34, 0, 0, 130, 38, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 208, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 108, 0, 116, 0, 48, 0, 56, 0, 121, 0, 129, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 136, 0, 144, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 162, 0, 48, 0, 48, 0, 170, 0, 177, 0, 184, 0, 48, 0, 191, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 160, 0, 48, 0, 48, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 152, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 176, 1, 192, 0, 228, 1, 192, 0, 192, 0, 192, 0, 32, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 15, 0, 202, 193, 15, 0, 202, 226, 15, 0, 202, 3, 16, 0, 202, 36, 16, 0, 202, 69, 16, 0, 202, 102, 16, 0, 202, 135, 16, 0, 202, 168, 16, 0, 202, 201, 16, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 201, 198, 0, 0, 156, 5, 0, 48, 201, 134, 1, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 166, 2, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 102, 3, 0, 156, 5, 0, 80, 201, 70, 4, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 134, 71, 0, 58, 201, 166, 5, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 134, 6, 0, 5, 5, 0, 44, 201, 70, 7, 0, 5, 5, 0, 48, 201, 6, 8, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 38, 9, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 230, 9, 0, 5, 5, 0, 80, 201, 198, 10, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 71, 0, 58, 201, 38, 12, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 14, 0, 201, 6, 13, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 197, 98, 0, 0, 160, 5, 38, 93, 197, 34, 15, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 15, 0, 201, 166, 13, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 197, 226, 0, 0, 5, 5, 38, 93, 197, 34, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 42, 5, 70, 0, 42, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 50, 5, 71, 0, 50, 134, 70, 0, 50, 5, 70, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 58, 5, 70, 0, 58, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 14, 0, 201, 38, 15, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 82, 5, 71, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 82, 5, 70, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 1, 0, 197, 34, 2, 0, 197, 98, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 15, 0, 201, 166, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 17, 0, 201, 38, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 18, 0, 201, 166, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 4, 0, 197, 98, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 22, 0, 201, 102, 23, 0, 201, 38, 24, 0, 201, 230, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 25, 0, 201, 102, 26, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 6, 0, 197, 98, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 27, 0, 201, 230, 27, 0, 201, 166, 28, 0, 201, 102, 29, 0, 201, 38, 30, 0, 201, 230, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 8, 0, 197, 98, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 8, 0, 197, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 31, 0, 201, 102, 32, 0, 201, 38, 33, 0, 201, 230, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 34, 0, 201, 102, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 36, 0, 201, 230, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 37, 0, 201, 102, 38, 0, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 197, 34, 14, 0, 197, 98, 14, 0, 197, 162, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 134, 71, 0, 58, 5, 136, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 5, 71, 0, 58, 5, 136, 0, 0, 5, 71, 0, 58, 5, 150, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 71, 0, 58, 5, 142, 0, 0, 5, 71, 0, 58, 5, 142, 0, 0, 134, 71, 0, 58, 5, 150, 0, 0, 134, 71, 0, 82, 5, 190, 0, 0, 5, 71, 0, 82, 5, 190, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 71, 0, 50, 5, 160, 0, 0, 5, 71, 0, 50, 5, 160, 0, 0, 134, 71, 0, 58, 5, 164, 0, 0, 5, 71, 0, 58, 5, 164, 0, 0, 134, 71, 0, 50, 5, 204, 0, 0, 5, 71, 0, 50, 5, 204, 0, 0, 134, 71, 0, 50, 5, 208, 0, 0, 5, 71, 0, 50, 5, 208, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 71, 0, 82, 5, 198, 0, 0, 5, 71, 0, 82, 5, 198, 0, 0, 134, 71, 0, 82, 5, 208, 0, 0, 5, 71, 0, 82, 5, 208, 0, 0, 134, 71, 0, 82, 5, 204, 0, 0, 5, 71, 0, 82, 5, 204, 0, 0, 134, 71, 0, 82, 5, 150, 0, 0, 5, 71, 0, 82, 5, 150, 0, 0, 134, 71, 0, 58, 5, 156, 0, 0, 5, 71, 0, 58, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 5, 71, 0, 58, 5, 146, 0, 0, 134, 71, 0, 50, 5, 196, 0, 0, 5, 71, 0, 50, 5, 196, 0, 0, 134, 71, 0, 82, 5, 196, 0, 0, 5, 71, 0, 82, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 58, 5, 138, 0, 0, 5, 71, 0, 58, 5, 138, 0, 0, 134, 71, 0, 58, 5, 196, 0, 0, 5, 71, 0, 58, 5, 196, 0, 0, 134, 71, 0, 58, 5, 182, 0, 0, 5, 71, 0, 58, 5, 182, 0, 0, 134, 71, 0, 58, 5, 154, 0, 0, 5, 71, 0, 58, 5, 154, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 48, 0, 40, 3, 0, 234, 134, 70, 0, 46, 156, 5, 48, 0, 12, 3, 2, 239, 134, 5, 0, 50, 156, 5, 1, 0, 7, 3, 0, 242, 134, 71, 40, 3, 0, 242, 134, 70, 0, 58, 156, 5, 48, 0, 40, 3, 0, 250, 134, 70, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 82, 156, 5, 1, 0, 4, 3, 255, 255, 0, 82, 134, 71, 40, 3, 255, 255, 0, 82, 134, 70, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 42, 5, 5, 48, 0, 40, 3, 0, 234, 5, 70, 0, 46, 5, 5, 48, 0, 12, 3, 2, 239, 5, 5, 0, 50, 5, 5, 1, 0, 7, 3, 0, 242, 5, 71, 40, 3, 0, 242, 5, 70, 0, 58, 5, 5, 48, 0, 40, 3, 0, 250, 5, 70, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 82, 5, 5, 1, 0, 4, 3, 255, 255, 0, 82, 5, 71, 40, 3, 255, 255, 0, 82, 5, 70, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 48, 0, 12, 3, 197, 162, 5, 46, 196, 160, 48, 0, 12, 3, 0, 192, 197, 162, 156, 78, 196, 160, 48, 0, 12, 3, 1, 192, 197, 98, 5, 78, 196, 160, 48, 0, 12, 3, 1, 192, 197, 162, 156, 82, 196, 190, 48, 0, 4, 3, 2, 192, 197, 162, 5, 82, 196, 190, 48, 0, 4, 3, 2, 192, 197, 226, 156, 78, 196, 202, 48, 0, 12, 3, 3, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 3, 192, 197, 98, 156, 50, 196, 160, 48, 0, 7, 3, 3, 192, 197, 162, 5, 50, 196, 160, 48, 0, 7, 3, 3, 192, 197, 226, 0, 0, 5, 156, 4, 0, 0, 3, 0, 192, 5, 138, 1, 3, 0, 192, 5, 136, 3, 3, 0, 192, 5, 154, 64, 3, 0, 192, 5, 138, 65, 3, 0, 192, 5, 136, 156, 50, 196, 204, 48, 0, 7, 3, 4, 192, 197, 162, 5, 50, 196, 204, 48, 0, 7, 3, 4, 192, 197, 226, 156, 50, 196, 208, 48, 0, 7, 3, 5, 192, 197, 34, 5, 50, 196, 208, 48, 0, 7, 3, 5, 192, 197, 98, 156, 78, 196, 196, 48, 0, 12, 3, 5, 192, 197, 162, 5, 78, 196, 196, 48, 0, 12, 3, 5, 192, 197, 226, 156, 82, 196, 198, 48, 0, 4, 3, 6, 192, 197, 162, 5, 82, 196, 198, 48, 0, 4, 3, 6, 192, 197, 226, 156, 82, 196, 208, 48, 0, 4, 3, 7, 192, 197, 34, 5, 82, 196, 208, 48, 0, 4, 3, 7, 192, 197, 98, 156, 82, 196, 204, 48, 0, 4, 3, 7, 192, 197, 162, 5, 82, 196, 204, 48, 0, 4, 3, 7, 192, 197, 226, 156, 92, 196, 196, 48, 0, 12, 3, 9, 192, 197, 34, 5, 92, 196, 196, 48, 0, 12, 3, 9, 192, 197, 98, 156, 92, 196, 210, 48, 0, 12, 3, 9, 192, 197, 162, 5, 92, 196, 210, 48, 0, 12, 3, 9, 192, 197, 226, 156, 50, 196, 196, 48, 0, 7, 3, 10, 192, 197, 98, 5, 50, 196, 196, 48, 0, 7, 3, 10, 192, 197, 162, 156, 82, 196, 196, 48, 0, 4, 3, 10, 192, 197, 226, 5, 82, 196, 196, 48, 0, 4, 3, 11, 192, 197, 34, 11, 0, 197, 99, 48, 0, 4, 3, 11, 192, 197, 195, 12, 0, 197, 35, 48, 0, 4, 3, 12, 192, 197, 131, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 36, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 38, 4, 185, 60, 40, 4, 185, 72, 42, 4, 185, 80, 185, 84, 185, 88, 44, 4, 186, 100, 185, 104, 185, 108, 185, 112, 186, 116, 185, 124, 185, 128, 185, 132, 185, 140, 46, 4, 185, 152, 48, 4, 185, 164, 185, 168, 185, 172, 249, 92, 50, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 52, 4, 168, 60, 54, 4, 168, 72, 56, 4, 168, 80, 168, 84, 168, 88, 58, 4, 168, 100, 168, 104, 168, 108, 168, 112, 168, 116, 168, 124, 168, 128, 168, 132, 168, 140, 60, 4, 168, 152, 62, 4, 168, 164, 168, 168, 168, 172, 232, 92, 64, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 124, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 66, 4, 186, 77, 154, 77, 250, 77, 90, 78, 186, 93, 154, 93, 250, 93, 90, 94, 217, 72, 154, 118, 185, 125, 153, 125, 249, 125, 153, 126, 89, 126, 136, 14, 217, 126, 185, 161, 153, 161, 249, 161, 89, 162, 2, 8, 185, 184, 4, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 6, 8, 68, 4, 168, 77, 136, 77, 232, 77, 72, 78, 168, 93, 136, 93, 232, 93, 72, 94, 200, 72, 136, 118, 168, 125, 136, 125, 232, 125, 136, 126, 72, 126, 128, 14, 200, 126, 168, 161, 136, 161, 232, 161, 72, 162, 8, 8, 168, 184, 10, 8, 25, 59, 8, 59, 217, 57, 200, 57, 217, 56, 200, 56, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 185, 68, 168, 68, 25, 74, 8, 74, 57, 75, 40, 75, 26, 79, 8, 79, 218, 77, 200, 77, 249, 76, 232, 76, 217, 76, 200, 76, 26, 78, 8, 78, 249, 85, 232, 85, 217, 85, 200, 85, 185, 86, 168, 86, 249, 86, 232, 86, 249, 89, 232, 89, 57, 91, 40, 91, 154, 94, 136, 94, 26, 95, 8, 95, 218, 93, 200, 93, 217, 92, 200, 92, 186, 94, 168, 96, 12, 8, 14, 8, 250, 101, 232, 101, 249, 106, 232, 106, 168, 136, 153, 109, 136, 109, 249, 110, 232, 110, 25, 110, 8, 110, 89, 111, 72, 111, 57, 111, 40, 111, 154, 117, 136, 117, 250, 118, 232, 118, 26, 118, 8, 118, 16, 8, 185, 120, 168, 120, 25, 127, 8, 127, 217, 125, 200, 125, 121, 126, 104, 126, 18, 8, 20, 8, 153, 141, 136, 141, 249, 142, 232, 142, 25, 142, 8, 142, 154, 145, 136, 145, 250, 145, 232, 145, 70, 4, 72, 4, 185, 148, 168, 148, 249, 154, 232, 154, 25, 154, 8, 154, 185, 156, 168, 156, 153, 162, 136, 162, 249, 160, 232, 160, 217, 161, 200, 161, 57, 162, 40, 162, 121, 162, 104, 162, 217, 160, 200, 160, 249, 169, 232, 169, 22, 8, 24, 8, 26, 8, 153, 177, 136, 177, 185, 178, 168, 178, 185, 180, 168, 180, 232, 144, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 28, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 30, 8, 1, 0, 8, 14, 32, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 34, 8, 200, 12, 72, 12, 32, 14, 249, 56, 187, 76, 249, 92, 136, 1, 200, 144, 169, 144, 232, 56, 169, 76, 232, 92, 136, 1, 232, 92, 72, 2, 187, 92, 187, 100, 169, 92, 169, 100, 169, 188, 169, 116, 217, 124, 187, 76, 200, 124, 169, 76, 249, 92, 232, 1, 232, 92, 232, 1, 249, 92, 72, 2, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 64, 255, 5, 186, 76, 255, 5, 186, 92, 255, 5, 186, 144, 255, 5, 185, 160, 255, 5, 185, 176, 255, 5, 168, 56, 255, 5, 168, 64, 255, 5, 168, 76, 255, 5, 168, 92, 255, 5, 168, 144, 255, 5, 168, 160, 255, 5, 168, 176, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 250, 146, 255, 5, 232, 146, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 75, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 243, 9, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 248, 9, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 48, 9, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 52, 9, 0, 0, 52, 9, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 102, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 153, 220, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 68, 0, 60, 0, 13, 1, 60, 0, 60, 0, 60, 0, 12, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 72, 0, 60, 0, 35, 1, 60, 0, 60, 0, 60, 0, 34, 1, 38, 0, 73, 0, 60, 0, 60, 0, 121, 0, 60, 0, 60, 0, 60, 0, 89, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 76, 0, 60, 0, 55, 1, 60, 0, 60, 0, 60, 0, 54, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 77, 0, 60, 0, 60, 1, 60, 0, 60, 0, 60, 0, 59, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 79, 0, 60, 0, 70, 1, 60, 0, 60, 0, 60, 0, 69, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 83, 0, 60, 0, 87, 1, 60, 0, 60, 0, 60, 0, 86, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 84, 0, 60, 0, 97, 1, 60, 0, 60, 0, 60, 0, 96, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 183, 1, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 0, 0, 170, 170, 170, 170, 170, 170, 2, 35, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 20, 28, 0, 0, 20, 28, 0, 0, 20, 28, 0, 0, 20, 28, 0, 0, 96, 29, 0, 0, 96, 29, 0, 0, 194, 30, 0, 0, 194, 30, 0, 0, 234, 34, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 180, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 48, 0, 56, 0, 48, 0, 56, 0, 112, 0, 120, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 144, 0, 149, 0, 48, 0, 48, 0, 156, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 163, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 171, 0, 136, 0, 48, 0, 48, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 128, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 0, 192, 0, 192, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 9, 0, 202, 65, 9, 0, 202, 98, 9, 0, 202, 131, 9, 0, 202, 164, 9, 0, 202, 197, 9, 0, 202, 230, 9, 0, 202, 7, 10, 0, 202, 40, 10, 0, 202, 73, 10, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 198, 0, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 201, 134, 1, 0, 201, 70, 2, 0, 156, 5, 0, 66, 201, 38, 3, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 201, 6, 4, 0, 201, 230, 4, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 134, 70, 0, 58, 201, 198, 5, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 166, 6, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 102, 7, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 201, 38, 8, 0, 201, 230, 8, 0, 5, 5, 0, 66, 201, 198, 9, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 201, 166, 10, 0, 201, 134, 11, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 70, 0, 58, 201, 102, 12, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 8, 0, 201, 70, 13, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 197, 98, 0, 0, 160, 5, 38, 93, 197, 162, 8, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 8, 0, 201, 230, 13, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 197, 226, 0, 0, 5, 5, 38, 93, 197, 34, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 21, 47, 5, 5, 21, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 27, 55, 5, 5, 27, 55, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 21, 63, 5, 5, 21, 63, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 45, 65, 5, 5, 45, 65, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 33, 69, 5, 5, 33, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 63, 77, 5, 5, 63, 77, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 14, 0, 201, 102, 15, 0, 134, 5, 35, 79, 5, 5, 35, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 1, 0, 197, 34, 2, 0, 197, 98, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 25, 93, 5, 5, 25, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 16, 0, 201, 230, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 3, 0, 197, 98, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 17, 0, 201, 102, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 4, 0, 197, 98, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 19, 0, 201, 230, 19, 0, 201, 166, 20, 0, 201, 102, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 6, 0, 197, 162, 6, 0, 197, 226, 6, 0, 197, 34, 7, 0, 197, 98, 7, 0, 197, 162, 7, 0, 197, 226, 7, 0, 197, 34, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 21, 47, 5, 160, 0, 0, 134, 70, 0, 58, 5, 136, 0, 0, 5, 5, 21, 47, 5, 160, 0, 0, 5, 70, 0, 58, 5, 136, 0, 0, 5, 70, 0, 58, 5, 150, 0, 0, 134, 5, 35, 79, 5, 160, 0, 0, 5, 5, 35, 79, 5, 160, 0, 0, 134, 70, 0, 58, 5, 142, 0, 0, 5, 70, 0, 58, 5, 142, 0, 0, 134, 70, 0, 58, 5, 150, 0, 0, 134, 5, 35, 79, 5, 202, 0, 0, 5, 5, 35, 79, 5, 202, 0, 0, 134, 70, 0, 58, 5, 164, 0, 0, 5, 70, 0, 58, 5, 164, 0, 0, 134, 5, 35, 79, 5, 196, 0, 0, 5, 5, 35, 79, 5, 196, 0, 0, 134, 5, 35, 79, 5, 156, 0, 0, 5, 5, 35, 79, 5, 156, 0, 0, 134, 70, 0, 58, 5, 156, 0, 0, 5, 70, 0, 58, 5, 156, 0, 0, 134, 5, 25, 93, 5, 196, 0, 0, 5, 5, 25, 93, 5, 196, 0, 0, 134, 5, 25, 93, 5, 210, 0, 0, 5, 5, 25, 93, 5, 210, 0, 0, 5, 70, 0, 58, 5, 146, 0, 0, 134, 70, 0, 58, 5, 138, 0, 0, 5, 70, 0, 58, 5, 138, 0, 0, 134, 70, 0, 58, 5, 196, 0, 0, 5, 70, 0, 58, 5, 196, 0, 0, 134, 70, 0, 58, 5, 182, 0, 0, 5, 70, 0, 58, 5, 182, 0, 0, 134, 70, 0, 58, 5, 154, 0, 0, 5, 70, 0, 58, 5, 154, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 48, 0, 12, 3, 21, 239, 134, 5, 0, 54, 156, 5, 48, 0, 39, 3, 27, 247, 134, 5, 0, 62, 156, 5, 48, 0, 39, 3, 21, 255, 134, 5, 0, 64, 156, 5, 48, 0, 39, 3, 255, 255, 45, 65, 134, 5, 0, 68, 156, 5, 48, 0, 39, 3, 255, 255, 33, 69, 134, 5, 0, 76, 156, 5, 48, 0, 39, 3, 255, 255, 63, 77, 134, 5, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 35, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 25, 93, 134, 5, 0, 46, 5, 5, 48, 0, 12, 3, 21, 239, 5, 5, 0, 54, 5, 5, 48, 0, 39, 3, 27, 247, 5, 5, 0, 62, 5, 5, 48, 0, 39, 3, 21, 255, 5, 5, 0, 64, 5, 5, 48, 0, 39, 3, 255, 255, 45, 65, 5, 5, 0, 68, 5, 5, 48, 0, 39, 3, 255, 255, 33, 69, 5, 5, 0, 76, 5, 5, 48, 0, 39, 3, 255, 255, 63, 77, 5, 5, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 35, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 25, 93, 5, 5, 156, 46, 196, 160, 48, 0, 12, 3, 197, 162, 5, 46, 196, 160, 48, 0, 12, 3, 0, 192, 197, 162, 156, 78, 196, 160, 48, 0, 12, 3, 1, 192, 197, 98, 5, 78, 196, 160, 48, 0, 12, 3, 1, 192, 197, 162, 156, 78, 196, 202, 48, 0, 12, 3, 2, 192, 197, 162, 5, 78, 196, 202, 48, 0, 12, 3, 2, 192, 197, 226, 156, 78, 196, 196, 48, 0, 12, 3, 3, 192, 197, 162, 5, 78, 196, 196, 48, 0, 12, 3, 3, 192, 197, 226, 156, 92, 196, 196, 48, 0, 12, 3, 5, 192, 197, 34, 5, 92, 196, 196, 48, 0, 12, 3, 5, 192, 197, 98, 156, 92, 196, 210, 48, 0, 12, 3, 5, 192, 197, 162, 5, 92, 196, 210, 48, 0, 12, 3, 5, 192, 197, 226, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 36, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 38, 4, 185, 72, 186, 76, 185, 80, 40, 4, 185, 92, 186, 96, 186, 104, 42, 4, 44, 4, 185, 124, 46, 4, 185, 140, 185, 144, 185, 148, 48, 4, 50, 4, 185, 172, 185, 180, 185, 184, 185, 188, 185, 192, 217, 96, 52, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 54, 4, 168, 72, 168, 76, 168, 80, 56, 4, 168, 92, 168, 96, 168, 104, 58, 4, 60, 4, 168, 124, 62, 4, 168, 140, 168, 144, 168, 148, 64, 4, 66, 4, 168, 172, 168, 180, 168, 184, 168, 188, 168, 192, 200, 96, 68, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 140, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 70, 4, 186, 77, 154, 77, 250, 77, 90, 78, 186, 97, 154, 97, 250, 97, 90, 98, 217, 72, 154, 130, 185, 141, 153, 141, 249, 141, 153, 142, 89, 142, 136, 14, 217, 142, 185, 181, 153, 181, 249, 181, 89, 182, 2, 8, 185, 204, 4, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 6, 8, 72, 4, 168, 77, 136, 77, 232, 77, 72, 78, 168, 97, 136, 97, 232, 97, 72, 98, 200, 72, 136, 130, 168, 141, 136, 141, 232, 141, 136, 142, 72, 142, 128, 14, 200, 142, 168, 181, 136, 181, 232, 181, 72, 182, 8, 8, 168, 204, 10, 8, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 185, 68, 168, 68, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 249, 85, 232, 85, 217, 85, 200, 85, 185, 86, 168, 86, 185, 88, 168, 88, 249, 93, 232, 93, 89, 95, 72, 95, 154, 98, 136, 98, 58, 99, 40, 99, 218, 97, 200, 97, 26, 99, 8, 99, 186, 98, 168, 100, 12, 8, 14, 8, 250, 105, 232, 105, 185, 112, 168, 112, 168, 152, 153, 117, 136, 117, 185, 120, 168, 120, 25, 118, 8, 118, 121, 119, 104, 119, 89, 119, 72, 119, 154, 129, 136, 129, 185, 136, 168, 136, 26, 130, 8, 130, 16, 8, 185, 132, 168, 132, 57, 143, 40, 143, 217, 141, 200, 141, 121, 142, 104, 142, 18, 8, 20, 8, 153, 157, 136, 157, 185, 160, 168, 160, 25, 158, 8, 158, 154, 165, 136, 165, 250, 165, 232, 165, 74, 4, 76, 4, 185, 168, 168, 168, 249, 174, 232, 174, 25, 174, 8, 174, 185, 176, 168, 176, 153, 182, 136, 182, 57, 183, 40, 183, 217, 181, 200, 181, 57, 182, 40, 182, 121, 182, 104, 182, 25, 183, 8, 183, 249, 189, 232, 189, 22, 8, 24, 8, 26, 8, 153, 197, 136, 197, 185, 198, 168, 198, 185, 200, 168, 200, 232, 164, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 28, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 30, 8, 1, 0, 8, 14, 32, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 34, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 217, 96, 136, 1, 200, 164, 169, 164, 200, 56, 169, 76, 200, 96, 136, 1, 200, 96, 72, 2, 187, 96, 187, 104, 169, 96, 169, 104, 169, 208, 169, 128, 217, 140, 187, 76, 200, 140, 169, 76, 217, 96, 232, 1, 200, 96, 232, 1, 217, 96, 72, 2, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 185, 84, 255, 5, 185, 108, 255, 5, 185, 116, 255, 5, 186, 128, 255, 5, 185, 156, 255, 5, 186, 164, 255, 5, 185, 196, 255, 5, 168, 64, 255, 5, 168, 84, 255, 5, 168, 108, 255, 5, 168, 116, 255, 5, 168, 128, 255, 5, 168, 156, 255, 5, 168, 164, 255, 5, 168, 196, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 250, 166, 255, 5, 232, 166, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 103, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 41, 9, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 46, 9, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 7, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 8, 7, 0, 0, 8, 7, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 54, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 57, 220, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 93, 0, 91, 0, 115, 0, 117, 0, 112, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 67, 0, 111, 0, 110, 0, 116, 0, 114, 0, 97, 0, 99, 0, 116, 0, 105, 0, 111, 0, 110, 0, 115, 0, 32, 0, 91, 0, 24, 4, 56, 4, 93, 0, 93, 0, 38, 0, 3, 5, 60, 0, 83, 4, 60, 0, 60, 0, 60, 0, 3, 4, 38, 0, 91, 4, 60, 0, 92, 4, 60, 0, 60, 0, 60, 0, 12, 4, 0, 0, 170, 170, 170, 170, 170, 170, 20, 27, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 15, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 60, 0, 0, 0, 64, 0, 0, 0, 64, 1, 0, 0, 80, 26, 0, 0, 80, 26, 0, 0, 80, 26, 0, 0, 80, 26, 0, 0, 164, 26, 0, 0, 164, 26, 0, 0, 252, 26, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 39, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 136, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 95, 0, 103, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 127, 0, 119, 0, 48, 0, 48, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 124, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 1, 0, 202, 129, 1, 0, 202, 162, 1, 0, 202, 195, 1, 0, 202, 228, 1, 0, 202, 5, 2, 0, 202, 38, 2, 0, 202, 71, 2, 0, 202, 104, 2, 0, 202, 137, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 37, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 171, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 174, 5, 70, 97, 197, 34, 0, 0, 201, 102, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 97, 197, 98, 0, 0, 201, 38, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 174, 5, 70, 97, 5, 140, 0, 0, 5, 5, 70, 97, 5, 140, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 16, 97, 174, 5, 1, 0, 1, 3, 255, 255, 37, 97, 134, 5, 65, 3, 255, 255, 37, 97, 134, 5, 84, 97, 174, 5, 1, 0, 1, 3, 255, 255, 171, 97, 134, 5, 65, 3, 255, 255, 171, 97, 134, 5, 16, 97, 5, 5, 1, 0, 1, 3, 255, 255, 37, 97, 5, 5, 65, 3, 255, 255, 37, 97, 5, 5, 84, 97, 5, 5, 1, 0, 1, 3, 255, 255, 171, 97, 5, 5, 65, 3, 255, 255, 171, 97, 5, 5, 3, 0, 69, 0, 60, 0, 32, 0, 55, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 253, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 2, 7, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 108, 7, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 112, 7, 0, 0, 112, 7, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 108, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 165, 220, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 76, 0, 97, 0, 116, 0, 110, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 65, 0, 114, 0, 97, 0, 98, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 3, 13, 60, 0, 60, 0, 61, 13, 38, 0, 76, 13, 60, 0, 60, 0, 87, 13, 38, 0, 21, 13, 77, 13, 60, 0, 60, 0, 21, 13, 77, 13, 13, 32, 60, 0, 60, 0, 60, 0, 127, 13, 38, 0, 35, 13, 77, 13, 60, 0, 60, 0, 35, 13, 77, 13, 13, 32, 60, 0, 60, 0, 60, 0, 122, 13, 38, 0, 40, 13, 77, 13, 60, 0, 60, 0, 40, 13, 77, 13, 13, 32, 60, 0, 60, 0, 60, 0, 123, 13, 38, 0, 48, 13, 77, 13, 60, 0, 60, 0, 48, 13, 77, 13, 13, 32, 60, 0, 60, 0, 60, 0, 124, 13, 38, 0, 50, 13, 77, 13, 60, 0, 60, 0, 50, 13, 77, 13, 13, 32, 60, 0, 60, 0, 60, 0, 125, 13, 38, 0, 51, 13, 77, 13, 60, 0, 60, 0, 51, 13, 77, 13, 13, 32, 60, 0, 60, 0, 60, 0, 126, 13, 38, 0, 46, 13, 77, 13, 60, 0, 60, 0, 2, 13, 38, 0, 40, 13, 77, 13, 60, 0, 60, 0, 60, 0, 123, 13, 77, 13, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 210, 27, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 160, 0, 0, 0, 160, 1, 0, 0, 160, 26, 0, 0, 160, 26, 0, 0, 160, 26, 0, 0, 160, 26, 0, 0, 84, 27, 0, 0, 84, 27, 0, 0, 180, 27, 0, 0, 186, 27, 0, 0, 26, 0, 0, 0, 25, 0, 0, 0, 10, 0, 0, 0, 2, 0, 0, 0, 35, 0, 0, 0, 21, 0, 0, 0, 36, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 255, 255, 0, 96, 7, 0, 0, 102, 248, 255, 0, 103, 6, 0, 0, 104, 245, 255, 0, 105, 249, 255, 0, 109, 242, 255, 0, 110, 243, 255, 0, 111, 241, 255, 0, 112, 183, 255, 0, 113, 245, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 103, 104, 105, 106, 107, 108, 94, 109, 93, 98, 99, 100, 101, 95, 97, 96, 39, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 135, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 104, 0, 112, 0, 118, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 4, 0, 202, 129, 4, 0, 202, 162, 4, 0, 202, 195, 4, 0, 202, 228, 4, 0, 202, 5, 5, 0, 202, 38, 5, 0, 202, 71, 5, 0, 202, 104, 5, 0, 202, 137, 5, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 3, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 3, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 4, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 2, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 226, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 194, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 162, 2, 0, 192, 0, 0, 0, 201, 130, 3, 0, 201, 98, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 21, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 148, 112, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 1, 0, 201, 70, 5, 0, 197, 162, 2, 0, 197, 226, 2, 0, 197, 34, 3, 0, 197, 98, 3, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 90, 112, 5, 70, 152, 112, 5, 5, 40, 112, 5, 76, 152, 112, 5, 5, 68, 112, 5, 75, 152, 112, 5, 5, 78, 112, 5, 74, 152, 112, 5, 5, 94, 112, 5, 73, 152, 112, 5, 5, 96, 112, 5, 72, 152, 112, 5, 5, 108, 112, 5, 71, 152, 112, 5, 5, 68, 112, 6, 75, 152, 112, 5, 5, 78, 112, 6, 74, 152, 112, 5, 5, 78, 112, 6, 5, 152, 112, 5, 5, 94, 112, 6, 73, 152, 112, 5, 5, 96, 112, 6, 72, 152, 112, 5, 5, 108, 112, 6, 71, 152, 112, 5, 5, 40, 112, 6, 76, 152, 112, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 40, 112, 5, 5, 49, 0, 77, 13, 13, 32, 0, 192, 197, 98, 68, 112, 5, 5, 49, 0, 77, 13, 13, 32, 0, 192, 197, 162, 78, 112, 5, 5, 49, 0, 77, 13, 13, 32, 0, 192, 197, 226, 94, 112, 5, 5, 49, 0, 77, 13, 13, 32, 1, 192, 197, 34, 96, 112, 5, 5, 49, 0, 77, 13, 13, 32, 1, 192, 197, 98, 108, 112, 5, 5, 49, 0, 77, 13, 13, 32, 1, 192, 197, 162, 2, 0, 197, 34, 48, 0, 77, 13, 2, 192, 197, 98, 2, 0, 13, 32, 14, 32, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 111, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 101, 7, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 106, 7, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 127, 0, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 131, 0, 0, 0, 131, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 34, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 32, 0, 77, 0, 111, 0, 110, 0, 103, 0, 93, 0, 0, 0, 170, 170, 170, 170, 80, 1, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 8, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 32, 0, 0, 0, 56, 0, 0, 0, 56, 1, 0, 0, 8, 0, 0, 0, 27, 0, 0, 0, 0, 0, 0, 97, 198, 255, 0, 98, 0, 0, 52, 118, 178, 255, 56, 118, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 39, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 0, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 3, 0, 69, 0, 60, 0, 32, 0, 35, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 120, 0, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 125, 0, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 14, 0, 0, 0, 14, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 14, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 114, 0, 111, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 223, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 227, 6, 0, 0, 227, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 74, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 98, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 80, 9, 60, 0, 2, 9, 60, 0, 60, 0, 1, 9, 60, 0, 3, 9, 38, 0, 57, 9, 60, 0, 51, 9, 60, 0, 21, 9, 77, 9, 55, 9, 60, 0, 28, 9, 77, 9, 30, 9, 0, 0, 170, 170, 170, 170, 46, 26, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 120, 0, 0, 0, 120, 1, 0, 0, 152, 25, 0, 0, 152, 25, 0, 0, 160, 25, 0, 0, 160, 25, 0, 0, 236, 25, 0, 0, 236, 25, 0, 0, 12, 26, 0, 0, 22, 26, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 39, 40, 41, 42, 43, 44, 45, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 121, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 104, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 1, 0, 202, 65, 1, 0, 202, 98, 1, 0, 202, 131, 1, 0, 202, 164, 1, 0, 202, 197, 1, 0, 202, 230, 1, 0, 202, 7, 2, 0, 202, 40, 2, 0, 202, 73, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 193, 2, 7, 104, 193, 3, 7, 104, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 2, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 2, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 2, 143, 104, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 5, 0, 70, 0, 2, 7, 104, 0, 0, 0, 0, 193, 2, 143, 104, 194, 5, 175, 240, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 62, 104, 5, 5, 49, 0, 77, 9, 55, 9, 255, 255, 143, 104, 193, 3, 78, 104, 5, 5, 49, 0, 77, 9, 30, 9, 255, 255, 143, 104, 193, 4, 4, 0, 30, 9, 31, 9, 55, 9, 56, 9, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 75, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 216, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 221, 6, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 11, 8, 0, 0, 11, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 87, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 123, 220, 91, 0, 99, 0, 97, 0, 115, 0, 101, 0, 70, 0, 105, 0, 114, 0, 115, 0, 116, 0, 32, 0, 117, 0, 112, 0, 112, 0, 101, 0, 114, 0, 93, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 99, 0, 60, 0, 11, 1, 60, 0, 60, 0, 60, 0, 10, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 103, 0, 60, 0, 103, 0, 7, 3, 60, 0, 60, 0, 60, 0, 71, 0, 7, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 104, 0, 60, 0, 103, 0, 39, 1, 60, 0, 60, 0, 60, 0, 103, 0, 38, 1, 60, 0, 60, 0, 60, 0, 71, 0, 39, 1, 60, 0, 60, 0, 60, 0, 71, 0, 38, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 105, 0, 60, 0, 39, 1, 60, 0, 60, 0, 60, 0, 38, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 122, 0, 60, 0, 122, 0, 7, 3, 60, 0, 60, 0, 60, 0, 90, 0, 7, 3, 0, 0, 170, 170, 160, 30, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 35, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 25, 0, 0, 68, 25, 0, 0, 68, 25, 0, 0, 68, 25, 0, 0, 200, 25, 0, 0, 200, 25, 0, 0, 138, 26, 0, 0, 144, 26, 0, 0, 136, 30, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 135, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 90, 0, 98, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 118, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 126, 0, 114, 0, 48, 0, 48, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 106, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 104, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 2, 0, 202, 1, 3, 0, 202, 34, 3, 0, 202, 67, 3, 0, 202, 100, 3, 0, 202, 133, 3, 0, 202, 166, 3, 0, 202, 199, 3, 0, 202, 232, 3, 0, 202, 9, 4, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 196, 0, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 70, 2, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 38, 3, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 228, 3, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 102, 5, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 2, 0, 201, 70, 6, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 2, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 2, 0, 201, 230, 6, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 25, 45, 5, 5, 25, 45, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 21, 53, 5, 5, 21, 53, 201, 166, 7, 0, 201, 102, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 27, 57, 5, 5, 27, 57, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 17, 91, 5, 5, 17, 91, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 9, 0, 201, 230, 9, 0, 201, 166, 10, 0, 201, 102, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 25, 45, 5, 160, 0, 0, 5, 5, 25, 45, 5, 160, 0, 0, 134, 5, 21, 53, 5, 160, 0, 0, 5, 5, 21, 53, 5, 160, 0, 0, 134, 5, 17, 91, 5, 196, 0, 0, 5, 5, 17, 91, 5, 196, 0, 0, 134, 5, 17, 91, 5, 210, 0, 0, 5, 5, 17, 91, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 48, 0, 7, 3, 25, 237, 134, 5, 0, 54, 156, 5, 2, 0, 38, 1, 27, 247, 136, 5, 39, 1, 27, 247, 71, 5, 7, 3, 21, 245, 134, 5, 0, 92, 156, 5, 48, 0, 7, 3, 255, 255, 17, 91, 134, 5, 0, 46, 5, 5, 48, 0, 7, 3, 25, 237, 5, 5, 0, 54, 5, 5, 2, 0, 38, 1, 27, 247, 70, 5, 39, 1, 27, 247, 5, 5, 7, 3, 21, 245, 5, 5, 0, 92, 5, 5, 48, 0, 7, 3, 255, 255, 17, 91, 5, 5, 156, 46, 196, 160, 48, 0, 7, 3, 197, 162, 5, 46, 196, 160, 48, 0, 7, 3, 0, 192, 197, 98, 156, 54, 196, 160, 48, 0, 7, 3, 0, 192, 197, 162, 5, 54, 196, 160, 48, 0, 7, 3, 0, 192, 197, 226, 156, 92, 196, 196, 48, 0, 7, 3, 1, 192, 197, 34, 5, 92, 196, 196, 48, 0, 7, 3, 1, 192, 197, 98, 156, 92, 196, 210, 48, 0, 7, 3, 1, 192, 197, 162, 5, 92, 196, 210, 48, 0, 7, 3, 1, 192, 197, 226, 2, 0, 38, 1, 40, 1, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 72, 186, 76, 185, 80, 28, 4, 185, 96, 186, 104, 186, 112, 185, 116, 185, 120, 185, 124, 186, 128, 185, 136, 185, 140, 185, 144, 185, 152, 186, 156, 185, 160, 185, 168, 185, 172, 185, 176, 185, 180, 185, 184, 34, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 36, 4, 168, 72, 168, 76, 168, 80, 38, 4, 168, 96, 168, 104, 168, 112, 168, 116, 168, 120, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 168, 152, 168, 156, 168, 160, 168, 168, 168, 172, 168, 176, 168, 180, 168, 184, 44, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 136, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 46, 4, 186, 77, 154, 77, 250, 77, 90, 78, 186, 105, 154, 105, 250, 105, 90, 106, 217, 72, 154, 130, 185, 137, 153, 137, 249, 137, 153, 138, 89, 138, 136, 14, 217, 138, 185, 169, 153, 169, 249, 169, 89, 170, 153, 185, 185, 196, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 48, 4, 168, 77, 136, 77, 232, 77, 72, 78, 168, 105, 136, 105, 232, 105, 72, 106, 200, 72, 136, 130, 168, 137, 136, 137, 232, 137, 136, 138, 72, 138, 128, 14, 200, 138, 168, 169, 136, 169, 232, 169, 72, 170, 136, 185, 168, 196, 72, 186, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 69, 136, 69, 249, 69, 232, 69, 185, 64, 168, 64, 25, 70, 8, 70, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 249, 89, 232, 89, 217, 89, 200, 89, 185, 84, 168, 84, 50, 4, 52, 4, 249, 97, 232, 97, 185, 100, 168, 100, 154, 106, 136, 106, 58, 107, 40, 107, 218, 105, 200, 105, 26, 107, 8, 107, 186, 106, 168, 108, 6, 8, 8, 8, 250, 113, 232, 113, 249, 118, 232, 118, 168, 148, 153, 121, 136, 121, 249, 122, 232, 122, 25, 122, 8, 122, 121, 123, 104, 123, 89, 123, 72, 123, 154, 129, 136, 129, 250, 130, 232, 130, 26, 130, 8, 130, 10, 8, 185, 132, 168, 132, 57, 139, 40, 139, 217, 137, 200, 137, 121, 138, 104, 138, 12, 8, 14, 8, 153, 153, 136, 153, 249, 154, 232, 154, 25, 154, 8, 154, 154, 157, 136, 157, 250, 157, 232, 157, 250, 158, 232, 158, 26, 158, 8, 158, 249, 162, 232, 162, 25, 162, 8, 162, 185, 164, 168, 164, 153, 170, 136, 170, 57, 171, 40, 171, 217, 169, 200, 169, 57, 170, 40, 170, 121, 170, 104, 170, 25, 171, 8, 171, 249, 177, 232, 177, 249, 185, 232, 185, 89, 186, 153, 193, 136, 193, 185, 188, 168, 188, 25, 194, 8, 194, 232, 156, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 156, 169, 156, 200, 56, 169, 76, 187, 104, 187, 112, 169, 104, 169, 112, 169, 200, 169, 128, 217, 136, 187, 76, 200, 136, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 68, 255, 5, 185, 88, 38, 5, 187, 92, 39, 5, 178, 92, 255, 5, 185, 192, 255, 5, 168, 68, 255, 5, 168, 88, 38, 5, 177, 92, 39, 5, 168, 92, 255, 5, 168, 192, 255, 5, 249, 70, 255, 5, 232, 70, 255, 5, 249, 90, 255, 5, 232, 90, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 87, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 0, 8, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 5, 8, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 84, 38, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 88, 38, 0, 0, 88, 38, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 53, 13, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 239, 223, 55, 26, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 77, 0, 121, 0, 109, 0, 114, 0, 93, 0, 38, 0, 140, 16, 60, 0, 55, 16, 60, 0, 56, 16, 60, 0, 55, 16, 56, 16, 38, 0, 44, 16, 60, 0, 60, 0, 43, 16, 38, 0, 50, 16, 60, 0, 49, 16, 44, 16, 60, 0, 60, 0, 49, 16, 43, 16, 60, 0, 49, 16, 44, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 58, 16, 38, 0, 52, 16, 60, 0, 54, 16, 60, 0, 45, 16, 47, 16, 60, 0, 0, 16, 57, 16, 60, 0, 60, 0, 0, 16, 58, 16, 60, 0, 44, 16, 0, 16, 57, 16, 60, 0, 60, 0, 44, 16, 0, 16, 58, 16, 60, 0, 60, 0, 43, 16, 0, 16, 57, 16, 60, 0, 60, 0, 43, 16, 0, 16, 58, 16, 60, 0, 45, 16, 0, 16, 57, 16, 60, 0, 60, 0, 45, 16, 0, 16, 58, 16, 60, 0, 47, 16, 0, 16, 57, 16, 60, 0, 60, 0, 47, 16, 0, 16, 58, 16, 60, 0, 49, 16, 0, 16, 57, 16, 60, 0, 60, 0, 49, 16, 0, 16, 58, 16, 60, 0, 49, 16, 44, 16, 0, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 0, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 0, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 0, 16, 58, 16, 60, 0, 45, 16, 47, 16, 0, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 0, 16, 58, 16, 60, 0, 1, 16, 57, 16, 60, 0, 60, 0, 1, 16, 58, 16, 60, 0, 44, 16, 1, 16, 57, 16, 60, 0, 60, 0, 44, 16, 1, 16, 58, 16, 60, 0, 60, 0, 43, 16, 1, 16, 57, 16, 60, 0, 60, 0, 43, 16, 1, 16, 58, 16, 60, 0, 45, 16, 1, 16, 57, 16, 60, 0, 60, 0, 45, 16, 1, 16, 58, 16, 60, 0, 47, 16, 1, 16, 57, 16, 60, 0, 60, 0, 47, 16, 1, 16, 58, 16, 60, 0, 49, 16, 1, 16, 57, 16, 60, 0, 60, 0, 49, 16, 1, 16, 58, 16, 60, 0, 49, 16, 44, 16, 1, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 1, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 1, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 1, 16, 58, 16, 60, 0, 45, 16, 47, 16, 1, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 1, 16, 58, 16, 60, 0, 2, 16, 57, 16, 60, 0, 60, 0, 2, 16, 58, 16, 60, 0, 44, 16, 2, 16, 57, 16, 60, 0, 60, 0, 44, 16, 2, 16, 58, 16, 60, 0, 60, 0, 43, 16, 2, 16, 57, 16, 60, 0, 60, 0, 43, 16, 2, 16, 58, 16, 60, 0, 45, 16, 2, 16, 57, 16, 60, 0, 60, 0, 45, 16, 2, 16, 58, 16, 60, 0, 47, 16, 2, 16, 57, 16, 60, 0, 60, 0, 47, 16, 2, 16, 58, 16, 60, 0, 49, 16, 2, 16, 57, 16, 60, 0, 60, 0, 49, 16, 2, 16, 58, 16, 60, 0, 49, 16, 44, 16, 2, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 2, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 2, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 2, 16, 58, 16, 60, 0, 45, 16, 47, 16, 2, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 2, 16, 58, 16, 60, 0, 3, 16, 57, 16, 60, 0, 60, 0, 3, 16, 58, 16, 60, 0, 44, 16, 3, 16, 57, 16, 60, 0, 60, 0, 44, 16, 3, 16, 58, 16, 60, 0, 60, 0, 43, 16, 3, 16, 57, 16, 60, 0, 60, 0, 43, 16, 3, 16, 58, 16, 60, 0, 45, 16, 3, 16, 57, 16, 60, 0, 60, 0, 45, 16, 3, 16, 58, 16, 60, 0, 47, 16, 3, 16, 57, 16, 60, 0, 60, 0, 47, 16, 3, 16, 58, 16, 60, 0, 49, 16, 3, 16, 57, 16, 60, 0, 60, 0, 49, 16, 3, 16, 58, 16, 60, 0, 49, 16, 44, 16, 3, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 3, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 3, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 3, 16, 58, 16, 60, 0, 45, 16, 47, 16, 3, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 3, 16, 58, 16, 60, 0, 4, 16, 58, 16, 57, 16, 60, 0, 60, 0, 4, 16, 58, 16, 60, 0, 44, 16, 4, 16, 58, 16, 57, 16, 60, 0, 60, 0, 44, 16, 4, 16, 58, 16, 60, 0, 60, 0, 43, 16, 4, 16, 58, 16, 57, 16, 60, 0, 60, 0, 43, 16, 4, 16, 58, 16, 60, 0, 45, 16, 4, 16, 58, 16, 57, 16, 60, 0, 60, 0, 45, 16, 4, 16, 58, 16, 60, 0, 47, 16, 4, 16, 58, 16, 57, 16, 60, 0, 60, 0, 47, 16, 4, 16, 58, 16, 60, 0, 49, 16, 4, 16, 58, 16, 57, 16, 60, 0, 60, 0, 49, 16, 4, 16, 58, 16, 60, 0, 49, 16, 44, 16, 4, 16, 58, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 4, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 4, 16, 58, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 4, 16, 58, 16, 60, 0, 45, 16, 47, 16, 4, 16, 58, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 4, 16, 58, 16, 60, 0, 5, 16, 57, 16, 60, 0, 60, 0, 5, 16, 58, 16, 60, 0, 44, 16, 5, 16, 57, 16, 60, 0, 60, 0, 44, 16, 5, 16, 58, 16, 60, 0, 60, 0, 43, 16, 5, 16, 57, 16, 60, 0, 60, 0, 43, 16, 5, 16, 58, 16, 60, 0, 45, 16, 5, 16, 57, 16, 60, 0, 60, 0, 45, 16, 5, 16, 58, 16, 60, 0, 47, 16, 5, 16, 57, 16, 60, 0, 60, 0, 47, 16, 5, 16, 58, 16, 60, 0, 49, 16, 5, 16, 57, 16, 60, 0, 60, 0, 49, 16, 5, 16, 58, 16, 60, 0, 49, 16, 44, 16, 5, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 5, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 5, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 5, 16, 58, 16, 60, 0, 45, 16, 47, 16, 5, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 5, 16, 58, 16, 60, 0, 6, 16, 57, 16, 60, 0, 60, 0, 6, 16, 58, 16, 60, 0, 44, 16, 6, 16, 57, 16, 60, 0, 60, 0, 44, 16, 6, 16, 58, 16, 60, 0, 60, 0, 43, 16, 6, 16, 57, 16, 60, 0, 60, 0, 43, 16, 6, 16, 58, 16, 60, 0, 45, 16, 6, 16, 57, 16, 60, 0, 60, 0, 45, 16, 6, 16, 58, 16, 60, 0, 47, 16, 6, 16, 57, 16, 60, 0, 60, 0, 47, 16, 6, 16, 58, 16, 60, 0, 49, 16, 6, 16, 57, 16, 60, 0, 60, 0, 49, 16, 6, 16, 58, 16, 60, 0, 49, 16, 44, 16, 6, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 6, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 6, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 6, 16, 58, 16, 60, 0, 45, 16, 47, 16, 6, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 6, 16, 58, 16, 60, 0, 7, 16, 57, 16, 60, 0, 60, 0, 7, 16, 58, 16, 60, 0, 44, 16, 7, 16, 57, 16, 60, 0, 60, 0, 44, 16, 7, 16, 58, 16, 60, 0, 60, 0, 43, 16, 7, 16, 57, 16, 60, 0, 60, 0, 43, 16, 7, 16, 58, 16, 60, 0, 45, 16, 7, 16, 57, 16, 60, 0, 60, 0, 45, 16, 7, 16, 58, 16, 60, 0, 47, 16, 7, 16, 57, 16, 60, 0, 60, 0, 47, 16, 7, 16, 58, 16, 60, 0, 49, 16, 7, 16, 57, 16, 60, 0, 60, 0, 49, 16, 7, 16, 58, 16, 60, 0, 49, 16, 44, 16, 7, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 7, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 7, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 7, 16, 58, 16, 60, 0, 45, 16, 47, 16, 7, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 7, 16, 58, 16, 60, 0, 8, 16, 58, 16, 60, 0, 44, 16, 8, 16, 58, 16, 60, 0, 60, 0, 43, 16, 8, 16, 58, 16, 60, 0, 45, 16, 8, 16, 58, 16, 60, 0, 47, 16, 8, 16, 58, 16, 60, 0, 49, 16, 8, 16, 58, 16, 60, 0, 49, 16, 44, 16, 8, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 8, 16, 58, 16, 60, 0, 45, 16, 47, 16, 8, 16, 58, 16, 60, 0, 9, 16, 57, 16, 60, 0, 60, 0, 9, 16, 58, 16, 60, 0, 44, 16, 9, 16, 57, 16, 60, 0, 60, 0, 44, 16, 9, 16, 58, 16, 60, 0, 60, 0, 43, 16, 9, 16, 57, 16, 60, 0, 60, 0, 43, 16, 9, 16, 58, 16, 60, 0, 45, 16, 9, 16, 57, 16, 60, 0, 60, 0, 45, 16, 9, 16, 58, 16, 60, 0, 47, 16, 9, 16, 57, 16, 60, 0, 60, 0, 47, 16, 9, 16, 58, 16, 60, 0, 49, 16, 9, 16, 57, 16, 60, 0, 60, 0, 49, 16, 9, 16, 58, 16, 60, 0, 49, 16, 44, 16, 9, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 9, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 9, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 9, 16, 58, 16, 60, 0, 45, 16, 47, 16, 9, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 9, 16, 58, 16, 60, 0, 10, 16, 57, 16, 60, 0, 60, 0, 10, 16, 58, 16, 60, 0, 44, 16, 10, 16, 57, 16, 60, 0, 60, 0, 44, 16, 10, 16, 58, 16, 60, 0, 60, 0, 43, 16, 10, 16, 57, 16, 60, 0, 60, 0, 43, 16, 10, 16, 58, 16, 60, 0, 45, 16, 10, 16, 57, 16, 60, 0, 60, 0, 45, 16, 10, 16, 58, 16, 60, 0, 47, 16, 10, 16, 57, 16, 60, 0, 60, 0, 47, 16, 10, 16, 58, 16, 60, 0, 49, 16, 10, 16, 57, 16, 60, 0, 60, 0, 49, 16, 10, 16, 58, 16, 60, 0, 49, 16, 44, 16, 10, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 10, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 10, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 10, 16, 58, 16, 60, 0, 45, 16, 47, 16, 10, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 10, 16, 58, 16, 60, 0, 11, 16, 57, 16, 60, 0, 60, 0, 11, 16, 58, 16, 60, 0, 44, 16, 11, 16, 57, 16, 60, 0, 60, 0, 44, 16, 11, 16, 58, 16, 60, 0, 60, 0, 43, 16, 11, 16, 57, 16, 60, 0, 60, 0, 43, 16, 11, 16, 58, 16, 60, 0, 45, 16, 11, 16, 57, 16, 60, 0, 60, 0, 45, 16, 11, 16, 58, 16, 60, 0, 47, 16, 11, 16, 57, 16, 60, 0, 60, 0, 47, 16, 11, 16, 58, 16, 60, 0, 49, 16, 11, 16, 57, 16, 60, 0, 60, 0, 49, 16, 11, 16, 58, 16, 60, 0, 49, 16, 44, 16, 11, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 11, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 11, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 11, 16, 58, 16, 60, 0, 45, 16, 47, 16, 11, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 11, 16, 58, 16, 60, 0, 12, 16, 57, 16, 60, 0, 60, 0, 12, 16, 58, 16, 60, 0, 44, 16, 12, 16, 57, 16, 60, 0, 60, 0, 44, 16, 12, 16, 58, 16, 60, 0, 60, 0, 43, 16, 12, 16, 57, 16, 60, 0, 60, 0, 43, 16, 12, 16, 58, 16, 60, 0, 45, 16, 12, 16, 57, 16, 60, 0, 60, 0, 45, 16, 12, 16, 58, 16, 60, 0, 47, 16, 12, 16, 57, 16, 60, 0, 60, 0, 47, 16, 12, 16, 58, 16, 60, 0, 49, 16, 12, 16, 57, 16, 60, 0, 60, 0, 49, 16, 12, 16, 58, 16, 60, 0, 49, 16, 44, 16, 12, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 12, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 12, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 12, 16, 58, 16, 60, 0, 45, 16, 47, 16, 12, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 12, 16, 58, 16, 60, 0, 13, 16, 57, 16, 60, 0, 60, 0, 13, 16, 58, 16, 60, 0, 44, 16, 13, 16, 57, 16, 60, 0, 60, 0, 44, 16, 13, 16, 58, 16, 60, 0, 60, 0, 43, 16, 13, 16, 57, 16, 60, 0, 60, 0, 43, 16, 13, 16, 58, 16, 60, 0, 45, 16, 13, 16, 57, 16, 60, 0, 60, 0, 45, 16, 13, 16, 58, 16, 60, 0, 47, 16, 13, 16, 57, 16, 60, 0, 60, 0, 47, 16, 13, 16, 58, 16, 60, 0, 49, 16, 13, 16, 57, 16, 60, 0, 60, 0, 49, 16, 13, 16, 58, 16, 60, 0, 49, 16, 44, 16, 13, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 13, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 13, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 13, 16, 58, 16, 60, 0, 45, 16, 47, 16, 13, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 13, 16, 58, 16, 60, 0, 14, 16, 57, 16, 60, 0, 60, 0, 14, 16, 58, 16, 60, 0, 44, 16, 14, 16, 57, 16, 60, 0, 60, 0, 44, 16, 14, 16, 58, 16, 60, 0, 60, 0, 43, 16, 14, 16, 57, 16, 60, 0, 60, 0, 43, 16, 14, 16, 58, 16, 60, 0, 45, 16, 14, 16, 57, 16, 60, 0, 60, 0, 45, 16, 14, 16, 58, 16, 60, 0, 47, 16, 14, 16, 57, 16, 60, 0, 60, 0, 47, 16, 14, 16, 58, 16, 60, 0, 49, 16, 14, 16, 57, 16, 60, 0, 60, 0, 49, 16, 14, 16, 58, 16, 60, 0, 49, 16, 44, 16, 14, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 14, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 14, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 14, 16, 58, 16, 60, 0, 45, 16, 47, 16, 14, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 14, 16, 58, 16, 60, 0, 15, 16, 57, 16, 60, 0, 60, 0, 15, 16, 58, 16, 60, 0, 44, 16, 15, 16, 57, 16, 60, 0, 60, 0, 44, 16, 15, 16, 58, 16, 60, 0, 60, 0, 43, 16, 15, 16, 57, 16, 60, 0, 60, 0, 43, 16, 15, 16, 58, 16, 60, 0, 45, 16, 15, 16, 57, 16, 60, 0, 60, 0, 45, 16, 15, 16, 58, 16, 60, 0, 47, 16, 15, 16, 57, 16, 60, 0, 60, 0, 47, 16, 15, 16, 58, 16, 60, 0, 49, 16, 15, 16, 57, 16, 60, 0, 60, 0, 49, 16, 15, 16, 58, 16, 60, 0, 49, 16, 44, 16, 15, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 15, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 15, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 15, 16, 58, 16, 60, 0, 45, 16, 47, 16, 15, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 15, 16, 58, 16, 60, 0, 16, 16, 57, 16, 60, 0, 60, 0, 16, 16, 58, 16, 60, 0, 44, 16, 16, 16, 57, 16, 60, 0, 60, 0, 44, 16, 16, 16, 58, 16, 60, 0, 60, 0, 43, 16, 16, 16, 57, 16, 60, 0, 60, 0, 43, 16, 16, 16, 58, 16, 60, 0, 45, 16, 16, 16, 57, 16, 60, 0, 60, 0, 45, 16, 16, 16, 58, 16, 60, 0, 47, 16, 16, 16, 57, 16, 60, 0, 60, 0, 47, 16, 16, 16, 58, 16, 60, 0, 49, 16, 16, 16, 57, 16, 60, 0, 60, 0, 49, 16, 16, 16, 58, 16, 60, 0, 49, 16, 44, 16, 16, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 16, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 16, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 16, 16, 58, 16, 60, 0, 45, 16, 47, 16, 16, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 16, 16, 58, 16, 60, 0, 17, 16, 57, 16, 60, 0, 60, 0, 17, 16, 58, 16, 60, 0, 44, 16, 17, 16, 57, 16, 60, 0, 60, 0, 44, 16, 17, 16, 58, 16, 60, 0, 60, 0, 43, 16, 17, 16, 57, 16, 60, 0, 60, 0, 43, 16, 17, 16, 58, 16, 60, 0, 45, 16, 17, 16, 57, 16, 60, 0, 60, 0, 45, 16, 17, 16, 58, 16, 60, 0, 47, 16, 17, 16, 57, 16, 60, 0, 60, 0, 47, 16, 17, 16, 58, 16, 60, 0, 49, 16, 17, 16, 57, 16, 60, 0, 60, 0, 49, 16, 17, 16, 58, 16, 60, 0, 49, 16, 44, 16, 17, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 17, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 17, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 17, 16, 58, 16, 60, 0, 45, 16, 47, 16, 17, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 17, 16, 58, 16, 60, 0, 18, 16, 57, 16, 60, 0, 60, 0, 18, 16, 58, 16, 60, 0, 44, 16, 18, 16, 57, 16, 60, 0, 60, 0, 44, 16, 18, 16, 58, 16, 60, 0, 60, 0, 43, 16, 18, 16, 57, 16, 60, 0, 60, 0, 43, 16, 18, 16, 58, 16, 60, 0, 45, 16, 18, 16, 57, 16, 60, 0, 60, 0, 45, 16, 18, 16, 58, 16, 60, 0, 47, 16, 18, 16, 57, 16, 60, 0, 60, 0, 47, 16, 18, 16, 58, 16, 60, 0, 49, 16, 18, 16, 57, 16, 60, 0, 60, 0, 49, 16, 18, 16, 58, 16, 60, 0, 49, 16, 44, 16, 18, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 18, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 18, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 18, 16, 58, 16, 60, 0, 45, 16, 47, 16, 18, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 18, 16, 58, 16, 60, 0, 19, 16, 57, 16, 60, 0, 60, 0, 19, 16, 58, 16, 60, 0, 44, 16, 19, 16, 57, 16, 60, 0, 60, 0, 44, 16, 19, 16, 58, 16, 60, 0, 60, 0, 43, 16, 19, 16, 57, 16, 60, 0, 60, 0, 43, 16, 19, 16, 58, 16, 60, 0, 45, 16, 19, 16, 57, 16, 60, 0, 60, 0, 45, 16, 19, 16, 58, 16, 60, 0, 47, 16, 19, 16, 57, 16, 60, 0, 60, 0, 47, 16, 19, 16, 58, 16, 60, 0, 49, 16, 19, 16, 57, 16, 60, 0, 60, 0, 49, 16, 19, 16, 58, 16, 60, 0, 49, 16, 44, 16, 19, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 19, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 19, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 19, 16, 58, 16, 60, 0, 45, 16, 47, 16, 19, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 19, 16, 58, 16, 60, 0, 20, 16, 57, 16, 60, 0, 60, 0, 20, 16, 58, 16, 60, 0, 44, 16, 20, 16, 57, 16, 60, 0, 60, 0, 44, 16, 20, 16, 58, 16, 60, 0, 60, 0, 43, 16, 20, 16, 57, 16, 60, 0, 60, 0, 43, 16, 20, 16, 58, 16, 60, 0, 45, 16, 20, 16, 57, 16, 60, 0, 60, 0, 45, 16, 20, 16, 58, 16, 60, 0, 47, 16, 20, 16, 57, 16, 60, 0, 60, 0, 47, 16, 20, 16, 58, 16, 60, 0, 49, 16, 20, 16, 57, 16, 60, 0, 60, 0, 49, 16, 20, 16, 58, 16, 60, 0, 49, 16, 44, 16, 20, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 20, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 20, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 20, 16, 58, 16, 60, 0, 45, 16, 47, 16, 20, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 20, 16, 58, 16, 60, 0, 21, 16, 57, 16, 60, 0, 60, 0, 21, 16, 58, 16, 60, 0, 44, 16, 21, 16, 57, 16, 60, 0, 60, 0, 44, 16, 21, 16, 58, 16, 60, 0, 60, 0, 43, 16, 21, 16, 57, 16, 60, 0, 60, 0, 43, 16, 21, 16, 58, 16, 60, 0, 45, 16, 21, 16, 57, 16, 60, 0, 60, 0, 45, 16, 21, 16, 58, 16, 60, 0, 47, 16, 21, 16, 57, 16, 60, 0, 60, 0, 47, 16, 21, 16, 58, 16, 60, 0, 49, 16, 21, 16, 57, 16, 60, 0, 60, 0, 49, 16, 21, 16, 58, 16, 60, 0, 49, 16, 44, 16, 21, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 21, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 21, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 21, 16, 58, 16, 60, 0, 45, 16, 47, 16, 21, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 21, 16, 58, 16, 60, 0, 22, 16, 57, 16, 60, 0, 60, 0, 22, 16, 58, 16, 60, 0, 44, 16, 22, 16, 57, 16, 60, 0, 60, 0, 44, 16, 22, 16, 58, 16, 60, 0, 60, 0, 43, 16, 22, 16, 57, 16, 60, 0, 60, 0, 43, 16, 22, 16, 58, 16, 60, 0, 45, 16, 22, 16, 57, 16, 60, 0, 60, 0, 45, 16, 22, 16, 58, 16, 60, 0, 47, 16, 22, 16, 57, 16, 60, 0, 60, 0, 47, 16, 22, 16, 58, 16, 60, 0, 49, 16, 22, 16, 57, 16, 60, 0, 60, 0, 49, 16, 22, 16, 58, 16, 60, 0, 49, 16, 44, 16, 22, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 22, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 22, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 22, 16, 58, 16, 60, 0, 45, 16, 47, 16, 22, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 22, 16, 58, 16, 60, 0, 23, 16, 57, 16, 60, 0, 60, 0, 23, 16, 58, 16, 60, 0, 44, 16, 23, 16, 57, 16, 60, 0, 60, 0, 44, 16, 23, 16, 58, 16, 60, 0, 60, 0, 43, 16, 23, 16, 57, 16, 60, 0, 60, 0, 43, 16, 23, 16, 58, 16, 60, 0, 45, 16, 23, 16, 57, 16, 60, 0, 60, 0, 45, 16, 23, 16, 58, 16, 60, 0, 47, 16, 23, 16, 57, 16, 60, 0, 60, 0, 47, 16, 23, 16, 58, 16, 60, 0, 49, 16, 23, 16, 57, 16, 60, 0, 60, 0, 49, 16, 23, 16, 58, 16, 60, 0, 49, 16, 44, 16, 23, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 23, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 23, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 23, 16, 58, 16, 60, 0, 45, 16, 47, 16, 23, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 23, 16, 58, 16, 60, 0, 24, 16, 57, 16, 60, 0, 60, 0, 24, 16, 58, 16, 60, 0, 44, 16, 24, 16, 57, 16, 60, 0, 60, 0, 44, 16, 24, 16, 58, 16, 60, 0, 60, 0, 43, 16, 24, 16, 57, 16, 60, 0, 60, 0, 43, 16, 24, 16, 58, 16, 60, 0, 45, 16, 24, 16, 57, 16, 60, 0, 60, 0, 45, 16, 24, 16, 58, 16, 60, 0, 47, 16, 24, 16, 57, 16, 60, 0, 60, 0, 47, 16, 24, 16, 58, 16, 60, 0, 49, 16, 24, 16, 57, 16, 60, 0, 60, 0, 49, 16, 24, 16, 58, 16, 60, 0, 49, 16, 44, 16, 24, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 24, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 24, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 24, 16, 58, 16, 60, 0, 45, 16, 47, 16, 24, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 24, 16, 58, 16, 60, 0, 25, 16, 57, 16, 60, 0, 60, 0, 25, 16, 58, 16, 60, 0, 44, 16, 25, 16, 57, 16, 60, 0, 60, 0, 44, 16, 25, 16, 58, 16, 60, 0, 60, 0, 43, 16, 25, 16, 57, 16, 60, 0, 60, 0, 43, 16, 25, 16, 58, 16, 60, 0, 45, 16, 25, 16, 57, 16, 60, 0, 60, 0, 45, 16, 25, 16, 58, 16, 60, 0, 47, 16, 25, 16, 57, 16, 60, 0, 60, 0, 47, 16, 25, 16, 58, 16, 60, 0, 49, 16, 25, 16, 57, 16, 60, 0, 60, 0, 49, 16, 25, 16, 58, 16, 60, 0, 49, 16, 44, 16, 25, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 25, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 25, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 25, 16, 58, 16, 60, 0, 45, 16, 47, 16, 25, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 25, 16, 58, 16, 60, 0, 26, 16, 57, 16, 60, 0, 60, 0, 26, 16, 58, 16, 60, 0, 44, 16, 26, 16, 57, 16, 60, 0, 60, 0, 44, 16, 26, 16, 58, 16, 60, 0, 60, 0, 43, 16, 26, 16, 57, 16, 60, 0, 60, 0, 43, 16, 26, 16, 58, 16, 60, 0, 45, 16, 26, 16, 57, 16, 60, 0, 60, 0, 45, 16, 26, 16, 58, 16, 60, 0, 47, 16, 26, 16, 57, 16, 60, 0, 60, 0, 47, 16, 26, 16, 58, 16, 60, 0, 49, 16, 26, 16, 57, 16, 60, 0, 60, 0, 49, 16, 26, 16, 58, 16, 60, 0, 49, 16, 44, 16, 26, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 26, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 26, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 26, 16, 58, 16, 60, 0, 45, 16, 47, 16, 26, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 26, 16, 58, 16, 60, 0, 27, 16, 57, 16, 60, 0, 60, 0, 27, 16, 58, 16, 60, 0, 44, 16, 27, 16, 57, 16, 60, 0, 60, 0, 44, 16, 27, 16, 58, 16, 60, 0, 60, 0, 43, 16, 27, 16, 57, 16, 60, 0, 60, 0, 43, 16, 27, 16, 58, 16, 60, 0, 45, 16, 27, 16, 57, 16, 60, 0, 60, 0, 45, 16, 27, 16, 58, 16, 60, 0, 47, 16, 27, 16, 57, 16, 60, 0, 60, 0, 47, 16, 27, 16, 58, 16, 60, 0, 49, 16, 27, 16, 57, 16, 60, 0, 60, 0, 49, 16, 27, 16, 58, 16, 60, 0, 49, 16, 44, 16, 27, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 27, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 27, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 27, 16, 58, 16, 60, 0, 45, 16, 47, 16, 27, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 27, 16, 58, 16, 60, 0, 28, 16, 57, 16, 60, 0, 60, 0, 28, 16, 58, 16, 60, 0, 44, 16, 28, 16, 57, 16, 60, 0, 60, 0, 44, 16, 28, 16, 58, 16, 60, 0, 60, 0, 43, 16, 28, 16, 57, 16, 60, 0, 60, 0, 43, 16, 28, 16, 58, 16, 60, 0, 45, 16, 28, 16, 57, 16, 60, 0, 60, 0, 45, 16, 28, 16, 58, 16, 60, 0, 47, 16, 28, 16, 57, 16, 60, 0, 60, 0, 47, 16, 28, 16, 58, 16, 60, 0, 49, 16, 28, 16, 57, 16, 60, 0, 60, 0, 49, 16, 28, 16, 58, 16, 60, 0, 49, 16, 44, 16, 28, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 28, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 28, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 28, 16, 58, 16, 60, 0, 45, 16, 47, 16, 28, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 28, 16, 58, 16, 60, 0, 29, 16, 58, 16, 60, 0, 44, 16, 29, 16, 58, 16, 60, 0, 60, 0, 43, 16, 29, 16, 58, 16, 60, 0, 45, 16, 29, 16, 58, 16, 60, 0, 47, 16, 29, 16, 58, 16, 60, 0, 49, 16, 29, 16, 58, 16, 60, 0, 49, 16, 44, 16, 29, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 29, 16, 58, 16, 60, 0, 45, 16, 47, 16, 29, 16, 58, 16, 60, 0, 30, 16, 57, 16, 60, 0, 60, 0, 30, 16, 58, 16, 60, 0, 44, 16, 30, 16, 57, 16, 60, 0, 60, 0, 44, 16, 30, 16, 58, 16, 60, 0, 60, 0, 43, 16, 30, 16, 57, 16, 60, 0, 60, 0, 43, 16, 30, 16, 58, 16, 60, 0, 45, 16, 30, 16, 57, 16, 60, 0, 60, 0, 45, 16, 30, 16, 58, 16, 60, 0, 47, 16, 30, 16, 57, 16, 60, 0, 60, 0, 47, 16, 30, 16, 58, 16, 60, 0, 49, 16, 30, 16, 57, 16, 60, 0, 60, 0, 49, 16, 30, 16, 58, 16, 60, 0, 49, 16, 44, 16, 30, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 30, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 30, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 30, 16, 58, 16, 60, 0, 45, 16, 47, 16, 30, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 30, 16, 58, 16, 60, 0, 31, 16, 57, 16, 60, 0, 60, 0, 31, 16, 58, 16, 60, 0, 44, 16, 31, 16, 57, 16, 60, 0, 60, 0, 44, 16, 31, 16, 58, 16, 60, 0, 60, 0, 43, 16, 31, 16, 57, 16, 60, 0, 60, 0, 43, 16, 31, 16, 58, 16, 60, 0, 45, 16, 31, 16, 57, 16, 60, 0, 60, 0, 45, 16, 31, 16, 58, 16, 60, 0, 47, 16, 31, 16, 57, 16, 60, 0, 60, 0, 47, 16, 31, 16, 58, 16, 60, 0, 49, 16, 31, 16, 57, 16, 60, 0, 60, 0, 49, 16, 31, 16, 58, 16, 60, 0, 49, 16, 44, 16, 31, 16, 57, 16, 60, 0, 60, 0, 49, 16, 44, 16, 31, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 31, 16, 57, 16, 60, 0, 60, 0, 49, 16, 43, 16, 31, 16, 58, 16, 60, 0, 45, 16, 47, 16, 31, 16, 57, 16, 60, 0, 60, 0, 45, 16, 47, 16, 31, 16, 58, 16, 60, 0, 32, 16, 58, 16, 60, 0, 44, 16, 32, 16, 58, 16, 60, 0, 60, 0, 43, 16, 32, 16, 58, 16, 60, 0, 45, 16, 32, 16, 58, 16, 60, 0, 47, 16, 32, 16, 58, 16, 60, 0, 49, 16, 32, 16, 58, 16, 60, 0, 49, 16, 44, 16, 32, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 32, 16, 58, 16, 60, 0, 45, 16, 47, 16, 32, 16, 58, 16, 60, 0, 33, 16, 58, 16, 60, 0, 44, 16, 33, 16, 58, 16, 60, 0, 60, 0, 43, 16, 33, 16, 58, 16, 60, 0, 45, 16, 33, 16, 58, 16, 60, 0, 47, 16, 33, 16, 58, 16, 60, 0, 49, 16, 33, 16, 58, 16, 60, 0, 49, 16, 44, 16, 33, 16, 58, 16, 60, 0, 60, 0, 49, 16, 43, 16, 33, 16, 58, 16, 60, 0, 45, 16, 47, 16, 33, 16, 58, 16, 60, 0, 94, 16, 60, 0, 95, 16, 60, 0, 59, 16, 60, 0, 60, 16, 60, 0, 96, 16, 60, 0, 61, 16, 60, 0, 130, 16, 60, 0, 62, 16, 60, 0, 59, 16, 61, 16, 60, 0, 60, 16, 61, 16, 60, 0, 59, 16, 62, 16, 60, 0, 60, 16, 62, 16, 60, 0, 61, 16, 62, 16, 60, 0, 59, 16, 61, 16, 62, 16, 60, 0, 60, 16, 61, 16, 62, 16, 38, 0, 33, 16, 45, 16, 60, 0, 60, 0, 60, 0, 35, 16, 38, 0, 33, 16, 46, 16, 60, 0, 60, 0, 60, 0, 36, 16, 38, 0, 33, 16, 47, 16, 60, 0, 60, 0, 60, 0, 37, 16, 38, 0, 33, 16, 48, 16, 60, 0, 60, 0, 60, 0, 38, 16, 38, 0, 33, 16, 49, 16, 60, 0, 60, 0, 60, 0, 39, 16, 60, 0, 60, 0, 60, 0, 40, 16, 38, 0, 33, 16, 49, 16, 44, 16, 60, 0, 60, 0, 60, 0, 41, 16, 38, 0, 33, 16, 49, 16, 44, 16, 58, 16, 60, 0, 60, 0, 60, 0, 42, 16, 38, 0, 45, 16, 25, 16, 58, 16, 60, 0, 45, 16, 54, 16, 60, 0, 45, 16, 25, 16, 58, 16, 55, 16, 60, 0, 45, 16, 54, 16, 55, 16, 60, 0, 45, 16, 25, 16, 58, 16, 56, 16, 60, 0, 45, 16, 54, 16, 56, 16, 38, 0, 47, 16, 25, 16, 58, 16, 60, 0, 47, 16, 54, 16, 60, 0, 47, 16, 25, 16, 58, 16, 55, 16, 60, 0, 47, 16, 54, 16, 55, 16, 60, 0, 47, 16, 25, 16, 58, 16, 56, 16, 60, 0, 47, 16, 54, 16, 56, 16, 38, 0, 33, 16, 47, 16, 54, 16, 60, 0, 60, 0, 60, 0, 37, 16, 47, 16, 54, 16, 38, 0, 33, 16, 45, 16, 0, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 0, 16, 57, 16, 38, 0, 33, 16, 45, 16, 0, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 0, 16, 58, 16, 38, 0, 33, 16, 47, 16, 0, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 0, 16, 57, 16, 38, 0, 33, 16, 47, 16, 0, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 0, 16, 58, 16, 38, 0, 33, 16, 49, 16, 0, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 0, 16, 57, 16, 38, 0, 33, 16, 49, 16, 0, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 0, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 0, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 0, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 0, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 0, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 0, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 0, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 0, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 0, 16, 58, 16, 38, 0, 33, 16, 45, 16, 1, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 1, 16, 57, 16, 38, 0, 33, 16, 45, 16, 1, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 1, 16, 58, 16, 38, 0, 33, 16, 47, 16, 1, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 1, 16, 57, 16, 38, 0, 33, 16, 47, 16, 1, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 1, 16, 58, 16, 38, 0, 33, 16, 49, 16, 1, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 1, 16, 57, 16, 38, 0, 33, 16, 49, 16, 1, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 1, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 1, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 1, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 1, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 1, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 1, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 1, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 1, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 1, 16, 58, 16, 38, 0, 33, 16, 45, 16, 2, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 2, 16, 57, 16, 38, 0, 33, 16, 45, 16, 2, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 2, 16, 58, 16, 38, 0, 33, 16, 47, 16, 2, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 2, 16, 57, 16, 38, 0, 33, 16, 47, 16, 2, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 2, 16, 58, 16, 38, 0, 33, 16, 49, 16, 2, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 2, 16, 57, 16, 38, 0, 33, 16, 49, 16, 2, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 2, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 2, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 2, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 2, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 2, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 2, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 2, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 2, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 2, 16, 58, 16, 38, 0, 33, 16, 45, 16, 3, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 3, 16, 57, 16, 38, 0, 33, 16, 45, 16, 3, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 3, 16, 58, 16, 38, 0, 33, 16, 47, 16, 3, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 3, 16, 57, 16, 38, 0, 33, 16, 47, 16, 3, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 3, 16, 58, 16, 38, 0, 33, 16, 49, 16, 3, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 3, 16, 57, 16, 38, 0, 33, 16, 49, 16, 3, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 3, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 3, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 3, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 3, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 3, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 3, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 3, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 3, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 3, 16, 58, 16, 38, 0, 33, 16, 45, 16, 4, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 4, 16, 57, 16, 38, 0, 33, 16, 45, 16, 4, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 4, 16, 58, 16, 38, 0, 33, 16, 47, 16, 4, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 4, 16, 57, 16, 38, 0, 33, 16, 47, 16, 4, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 4, 16, 58, 16, 38, 0, 33, 16, 49, 16, 4, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 4, 16, 57, 16, 38, 0, 33, 16, 49, 16, 4, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 4, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 4, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 4, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 4, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 4, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 4, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 4, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 4, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 4, 16, 58, 16, 38, 0, 33, 16, 45, 16, 5, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 5, 16, 57, 16, 38, 0, 33, 16, 45, 16, 5, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 5, 16, 58, 16, 38, 0, 33, 16, 47, 16, 5, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 5, 16, 57, 16, 38, 0, 33, 16, 47, 16, 5, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 5, 16, 58, 16, 38, 0, 33, 16, 49, 16, 5, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 5, 16, 57, 16, 38, 0, 33, 16, 49, 16, 5, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 5, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 5, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 5, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 5, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 5, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 5, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 5, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 5, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 5, 16, 58, 16, 38, 0, 33, 16, 45, 16, 6, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 6, 16, 57, 16, 38, 0, 33, 16, 45, 16, 6, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 6, 16, 58, 16, 38, 0, 33, 16, 47, 16, 6, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 6, 16, 57, 16, 38, 0, 33, 16, 47, 16, 6, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 6, 16, 58, 16, 38, 0, 33, 16, 49, 16, 6, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 6, 16, 57, 16, 38, 0, 33, 16, 49, 16, 6, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 6, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 6, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 6, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 6, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 6, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 6, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 6, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 6, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 6, 16, 58, 16, 38, 0, 33, 16, 45, 16, 7, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 7, 16, 57, 16, 38, 0, 33, 16, 45, 16, 7, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 7, 16, 58, 16, 38, 0, 33, 16, 47, 16, 7, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 7, 16, 57, 16, 38, 0, 33, 16, 47, 16, 7, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 7, 16, 58, 16, 38, 0, 33, 16, 49, 16, 7, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 7, 16, 57, 16, 38, 0, 33, 16, 49, 16, 7, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 7, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 7, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 7, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 7, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 7, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 7, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 7, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 7, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 7, 16, 58, 16, 38, 0, 33, 16, 45, 16, 9, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 9, 16, 57, 16, 38, 0, 33, 16, 45, 16, 9, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 9, 16, 58, 16, 38, 0, 33, 16, 47, 16, 9, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 9, 16, 57, 16, 38, 0, 33, 16, 47, 16, 9, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 9, 16, 58, 16, 38, 0, 33, 16, 49, 16, 9, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 9, 16, 57, 16, 38, 0, 33, 16, 49, 16, 9, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 9, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 9, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 9, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 9, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 9, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 9, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 9, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 9, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 9, 16, 58, 16, 38, 0, 33, 16, 45, 16, 10, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 10, 16, 57, 16, 38, 0, 33, 16, 45, 16, 10, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 10, 16, 58, 16, 38, 0, 33, 16, 47, 16, 10, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 10, 16, 57, 16, 38, 0, 33, 16, 47, 16, 10, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 10, 16, 58, 16, 38, 0, 33, 16, 49, 16, 10, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 10, 16, 57, 16, 38, 0, 33, 16, 49, 16, 10, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 10, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 10, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 10, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 10, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 10, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 10, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 10, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 10, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 10, 16, 58, 16, 38, 0, 33, 16, 45, 16, 11, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 11, 16, 57, 16, 38, 0, 33, 16, 45, 16, 11, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 11, 16, 58, 16, 38, 0, 33, 16, 47, 16, 11, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 11, 16, 57, 16, 38, 0, 33, 16, 47, 16, 11, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 11, 16, 58, 16, 38, 0, 33, 16, 49, 16, 11, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 11, 16, 57, 16, 38, 0, 33, 16, 49, 16, 11, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 11, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 11, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 11, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 11, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 11, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 11, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 11, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 11, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 11, 16, 58, 16, 38, 0, 33, 16, 45, 16, 12, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 12, 16, 57, 16, 38, 0, 33, 16, 45, 16, 12, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 12, 16, 58, 16, 38, 0, 33, 16, 47, 16, 12, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 12, 16, 57, 16, 38, 0, 33, 16, 47, 16, 12, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 12, 16, 58, 16, 38, 0, 33, 16, 49, 16, 12, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 12, 16, 57, 16, 38, 0, 33, 16, 49, 16, 12, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 12, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 12, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 12, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 12, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 12, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 12, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 12, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 12, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 12, 16, 58, 16, 38, 0, 33, 16, 45, 16, 13, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 13, 16, 57, 16, 38, 0, 33, 16, 45, 16, 13, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 13, 16, 58, 16, 38, 0, 33, 16, 47, 16, 13, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 13, 16, 57, 16, 38, 0, 33, 16, 47, 16, 13, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 13, 16, 58, 16, 38, 0, 33, 16, 49, 16, 13, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 13, 16, 57, 16, 38, 0, 33, 16, 49, 16, 13, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 13, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 13, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 13, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 13, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 13, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 13, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 13, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 13, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 13, 16, 58, 16, 38, 0, 33, 16, 45, 16, 14, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 14, 16, 57, 16, 38, 0, 33, 16, 45, 16, 14, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 14, 16, 58, 16, 38, 0, 33, 16, 47, 16, 14, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 14, 16, 57, 16, 38, 0, 33, 16, 47, 16, 14, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 14, 16, 58, 16, 38, 0, 33, 16, 49, 16, 14, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 14, 16, 57, 16, 38, 0, 33, 16, 49, 16, 14, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 14, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 14, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 14, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 14, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 14, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 14, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 14, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 14, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 14, 16, 58, 16, 38, 0, 33, 16, 45, 16, 15, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 15, 16, 57, 16, 38, 0, 33, 16, 45, 16, 15, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 15, 16, 58, 16, 38, 0, 33, 16, 47, 16, 15, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 15, 16, 57, 16, 38, 0, 33, 16, 47, 16, 15, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 15, 16, 58, 16, 38, 0, 33, 16, 49, 16, 15, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 15, 16, 57, 16, 38, 0, 33, 16, 49, 16, 15, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 15, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 15, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 15, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 15, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 15, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 15, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 15, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 15, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 15, 16, 58, 16, 38, 0, 33, 16, 45, 16, 16, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 16, 16, 57, 16, 38, 0, 33, 16, 45, 16, 16, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 16, 16, 58, 16, 38, 0, 33, 16, 47, 16, 16, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 16, 16, 57, 16, 38, 0, 33, 16, 47, 16, 16, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 16, 16, 58, 16, 38, 0, 33, 16, 49, 16, 16, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 16, 16, 57, 16, 38, 0, 33, 16, 49, 16, 16, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 16, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 16, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 16, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 16, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 16, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 16, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 16, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 16, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 16, 16, 58, 16, 38, 0, 33, 16, 45, 16, 17, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 17, 16, 57, 16, 38, 0, 33, 16, 45, 16, 17, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 17, 16, 58, 16, 38, 0, 33, 16, 47, 16, 17, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 17, 16, 57, 16, 38, 0, 33, 16, 47, 16, 17, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 17, 16, 58, 16, 38, 0, 33, 16, 49, 16, 17, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 17, 16, 57, 16, 38, 0, 33, 16, 49, 16, 17, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 17, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 17, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 17, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 17, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 17, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 17, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 17, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 17, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 17, 16, 58, 16, 38, 0, 33, 16, 45, 16, 18, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 18, 16, 57, 16, 38, 0, 33, 16, 45, 16, 18, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 18, 16, 58, 16, 38, 0, 33, 16, 47, 16, 18, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 18, 16, 57, 16, 38, 0, 33, 16, 47, 16, 18, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 18, 16, 58, 16, 38, 0, 33, 16, 49, 16, 18, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 18, 16, 57, 16, 38, 0, 33, 16, 49, 16, 18, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 18, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 18, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 18, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 18, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 18, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 18, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 18, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 18, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 18, 16, 58, 16, 38, 0, 33, 16, 45, 16, 19, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 19, 16, 57, 16, 38, 0, 33, 16, 45, 16, 19, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 19, 16, 58, 16, 38, 0, 33, 16, 47, 16, 19, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 19, 16, 57, 16, 38, 0, 33, 16, 47, 16, 19, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 19, 16, 58, 16, 38, 0, 33, 16, 49, 16, 19, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 19, 16, 57, 16, 38, 0, 33, 16, 49, 16, 19, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 19, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 19, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 19, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 19, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 19, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 19, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 19, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 19, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 19, 16, 58, 16, 38, 0, 33, 16, 45, 16, 20, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 20, 16, 57, 16, 38, 0, 33, 16, 45, 16, 20, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 20, 16, 58, 16, 38, 0, 33, 16, 47, 16, 20, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 20, 16, 57, 16, 38, 0, 33, 16, 47, 16, 20, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 20, 16, 58, 16, 38, 0, 33, 16, 49, 16, 20, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 20, 16, 57, 16, 38, 0, 33, 16, 49, 16, 20, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 20, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 20, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 20, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 20, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 20, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 20, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 20, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 20, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 20, 16, 58, 16, 38, 0, 33, 16, 45, 16, 21, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 21, 16, 57, 16, 38, 0, 33, 16, 45, 16, 21, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 21, 16, 58, 16, 38, 0, 33, 16, 47, 16, 21, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 21, 16, 57, 16, 38, 0, 33, 16, 47, 16, 21, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 21, 16, 58, 16, 38, 0, 33, 16, 49, 16, 21, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 21, 16, 57, 16, 38, 0, 33, 16, 49, 16, 21, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 21, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 21, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 21, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 21, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 21, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 21, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 21, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 21, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 21, 16, 58, 16, 38, 0, 33, 16, 45, 16, 22, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 22, 16, 57, 16, 38, 0, 33, 16, 45, 16, 22, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 22, 16, 58, 16, 38, 0, 33, 16, 47, 16, 22, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 22, 16, 57, 16, 38, 0, 33, 16, 47, 16, 22, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 22, 16, 58, 16, 38, 0, 33, 16, 49, 16, 22, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 22, 16, 57, 16, 38, 0, 33, 16, 49, 16, 22, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 22, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 22, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 22, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 22, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 22, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 22, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 22, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 22, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 22, 16, 58, 16, 38, 0, 33, 16, 45, 16, 23, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 23, 16, 57, 16, 38, 0, 33, 16, 45, 16, 23, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 23, 16, 58, 16, 38, 0, 33, 16, 47, 16, 23, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 23, 16, 57, 16, 38, 0, 33, 16, 47, 16, 23, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 23, 16, 58, 16, 38, 0, 33, 16, 49, 16, 23, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 23, 16, 57, 16, 38, 0, 33, 16, 49, 16, 23, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 23, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 23, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 23, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 23, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 23, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 23, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 23, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 23, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 23, 16, 58, 16, 38, 0, 33, 16, 45, 16, 24, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 24, 16, 57, 16, 38, 0, 33, 16, 45, 16, 24, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 24, 16, 58, 16, 38, 0, 33, 16, 47, 16, 24, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 24, 16, 57, 16, 38, 0, 33, 16, 47, 16, 24, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 24, 16, 58, 16, 38, 0, 33, 16, 49, 16, 24, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 24, 16, 57, 16, 38, 0, 33, 16, 49, 16, 24, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 24, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 24, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 24, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 24, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 24, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 24, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 24, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 24, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 24, 16, 58, 16, 38, 0, 33, 16, 45, 16, 25, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 25, 16, 57, 16, 38, 0, 33, 16, 45, 16, 25, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 25, 16, 58, 16, 38, 0, 33, 16, 47, 16, 25, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 25, 16, 57, 16, 38, 0, 33, 16, 47, 16, 25, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 25, 16, 58, 16, 38, 0, 33, 16, 49, 16, 25, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 25, 16, 57, 16, 38, 0, 33, 16, 49, 16, 25, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 25, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 25, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 25, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 25, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 25, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 25, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 25, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 25, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 25, 16, 58, 16, 38, 0, 33, 16, 45, 16, 26, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 26, 16, 57, 16, 38, 0, 33, 16, 45, 16, 26, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 26, 16, 58, 16, 38, 0, 33, 16, 47, 16, 26, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 26, 16, 57, 16, 38, 0, 33, 16, 47, 16, 26, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 26, 16, 58, 16, 38, 0, 33, 16, 49, 16, 26, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 26, 16, 57, 16, 38, 0, 33, 16, 49, 16, 26, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 26, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 26, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 26, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 26, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 26, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 26, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 26, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 26, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 26, 16, 58, 16, 38, 0, 33, 16, 45, 16, 27, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 27, 16, 57, 16, 38, 0, 33, 16, 45, 16, 27, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 27, 16, 58, 16, 38, 0, 33, 16, 47, 16, 27, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 27, 16, 57, 16, 38, 0, 33, 16, 47, 16, 27, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 27, 16, 58, 16, 38, 0, 33, 16, 49, 16, 27, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 27, 16, 57, 16, 38, 0, 33, 16, 49, 16, 27, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 27, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 27, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 27, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 27, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 27, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 27, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 27, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 27, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 27, 16, 58, 16, 38, 0, 33, 16, 45, 16, 28, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 28, 16, 57, 16, 38, 0, 33, 16, 45, 16, 28, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 28, 16, 58, 16, 38, 0, 33, 16, 47, 16, 28, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 28, 16, 57, 16, 38, 0, 33, 16, 47, 16, 28, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 28, 16, 58, 16, 38, 0, 33, 16, 49, 16, 28, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 28, 16, 57, 16, 38, 0, 33, 16, 49, 16, 28, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 28, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 28, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 28, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 28, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 28, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 28, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 28, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 28, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 28, 16, 58, 16, 38, 0, 33, 16, 45, 16, 30, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 30, 16, 57, 16, 38, 0, 33, 16, 45, 16, 30, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 30, 16, 58, 16, 38, 0, 33, 16, 47, 16, 30, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 30, 16, 57, 16, 38, 0, 33, 16, 47, 16, 30, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 30, 16, 58, 16, 38, 0, 33, 16, 49, 16, 30, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 30, 16, 57, 16, 38, 0, 33, 16, 49, 16, 30, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 30, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 30, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 30, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 30, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 30, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 30, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 30, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 30, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 30, 16, 58, 16, 38, 0, 33, 16, 45, 16, 31, 16, 57, 16, 60, 0, 60, 0, 60, 0, 35, 16, 31, 16, 57, 16, 38, 0, 33, 16, 45, 16, 31, 16, 58, 16, 60, 0, 60, 0, 60, 0, 35, 16, 31, 16, 58, 16, 38, 0, 33, 16, 47, 16, 31, 16, 57, 16, 60, 0, 60, 0, 60, 0, 37, 16, 31, 16, 57, 16, 38, 0, 33, 16, 47, 16, 31, 16, 58, 16, 60, 0, 60, 0, 60, 0, 37, 16, 31, 16, 58, 16, 38, 0, 33, 16, 49, 16, 31, 16, 57, 16, 60, 0, 60, 0, 60, 0, 39, 16, 31, 16, 57, 16, 38, 0, 33, 16, 49, 16, 31, 16, 58, 16, 60, 0, 60, 0, 60, 0, 39, 16, 31, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 31, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 31, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 31, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 31, 16, 58, 16, 38, 0, 33, 16, 49, 16, 44, 16, 31, 16, 57, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 31, 16, 57, 16, 38, 0, 33, 16, 49, 16, 44, 16, 31, 16, 58, 16, 60, 0, 60, 0, 60, 0, 41, 16, 49, 16, 44, 16, 31, 16, 58, 16, 38, 0, 49, 16, 44, 16, 0, 16, 58, 16, 0, 16, 59, 16, 61, 0, 49, 16, 44, 16, 0, 16, 58, 16, 59, 16, 38, 0, 20, 16, 58, 16, 20, 16, 47, 16, 21, 16, 58, 16, 61, 0, 20, 16, 58, 16, 47, 16, 21, 16, 58, 16, 38, 0, 49, 16, 30, 16, 57, 16, 30, 16, 61, 0, 49, 16, 63, 16, 38, 0, 45, 16, 30, 16, 57, 16, 30, 16, 61, 0, 45, 16, 63, 16, 38, 0, 47, 16, 30, 16, 57, 16, 30, 16, 61, 0, 47, 16, 63, 16, 38, 0, 33, 16, 45, 16, 30, 16, 57, 16, 30, 16, 60, 0, 60, 0, 60, 0, 35, 16, 63, 16, 38, 0, 33, 16, 47, 16, 30, 16, 57, 16, 30, 16, 60, 0, 60, 0, 60, 0, 37, 16, 63, 16, 38, 0, 30, 16, 57, 16, 30, 16, 61, 0, 63, 16, 38, 0, 20, 16, 62, 16, 45, 16, 47, 16, 0, 16, 58, 16, 60, 0, 60, 0, 76, 16, 38, 0, 27, 16, 61, 16, 49, 16, 55, 16, 60, 0, 60, 0, 77, 16, 38, 0, 28, 16, 10, 16, 58, 16, 56, 16, 0, 16, 49, 16, 44, 16, 4, 16, 58, 16, 56, 16, 60, 0, 60, 0, 78, 16, 4, 16, 58, 16, 56, 16, 38, 0, 33, 16, 45, 16, 60, 0, 60, 0, 79, 16, 38, 0, 28, 16, 0, 16, 58, 16, 26, 16, 44, 16, 61, 0, 28, 16, 0, 16, 58, 16, 59, 16, 44, 16, 38, 0, 30, 16, 25, 16, 46, 16, 61, 0, 30, 16, 57, 16, 25, 16, 46, 16, 38, 0, 17, 16, 25, 16, 4, 16, 58, 16, 56, 16, 61, 0, 17, 16, 57, 16, 25, 16, 4, 16, 58, 16, 56, 16, 38, 0, 28, 16, 0, 16, 58, 16, 24, 16, 0, 16, 58, 16, 61, 0, 28, 16, 57, 16, 24, 16, 0, 16, 58, 16, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 82, 100, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 16, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 76, 0, 0, 0, 76, 1, 0, 0, 236, 26, 0, 0, 236, 26, 0, 0, 12, 66, 0, 0, 12, 66, 0, 0, 172, 66, 0, 0, 172, 66, 0, 0, 20, 100, 0, 0, 54, 100, 0, 0, 28, 0, 0, 0, 0, 0, 79, 117, 178, 255, 94, 117, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 0, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 145, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 104, 0, 112, 0, 120, 0, 128, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 136, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 192, 3, 0, 202, 225, 3, 0, 202, 2, 4, 0, 202, 35, 4, 0, 202, 68, 4, 0, 202, 101, 4, 0, 202, 134, 4, 0, 202, 167, 4, 0, 202, 200, 4, 0, 202, 233, 4, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 2, 3, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 66, 3, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 130, 3, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 0, 0, 201, 38, 1, 0, 201, 38, 2, 0, 201, 70, 3, 0, 201, 102, 4, 0, 201, 134, 5, 0, 201, 166, 6, 0, 201, 198, 7, 0, 201, 230, 8, 0, 201, 166, 9, 0, 201, 198, 10, 0, 201, 230, 11, 0, 201, 6, 13, 0, 201, 38, 14, 0, 201, 70, 15, 0, 201, 102, 16, 0, 201, 134, 17, 0, 201, 166, 18, 0, 201, 166, 20, 0, 201, 198, 21, 0, 201, 230, 22, 0, 201, 166, 24, 0, 201, 198, 25, 0, 201, 230, 26, 0, 201, 6, 28, 0, 201, 38, 29, 0, 201, 70, 30, 0, 201, 102, 31, 0, 201, 132, 32, 0, 201, 102, 35, 0, 201, 38, 36, 0, 201, 198, 37, 0, 201, 230, 38, 0, 201, 166, 39, 0, 192, 0, 0, 0, 201, 100, 40, 0, 198, 34, 25, 0, 201, 196, 75, 0, 198, 98, 40, 0, 201, 68, 112, 0, 198, 66, 56, 0, 201, 36, 147, 0, 198, 66, 86, 0, 201, 36, 217, 0, 201, 4, 255, 0, 201, 4, 37, 1, 192, 0, 0, 0, 201, 228, 115, 1, 192, 0, 0, 0, 201, 164, 156, 1, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 150, 89, 117, 201, 160, 16, 2, 193, 81, 93, 117, 192, 0, 0, 0, 192, 0, 0, 0, 201, 128, 17, 2, 201, 32, 19, 2, 201, 192, 20, 2, 198, 193, 153, 0, 198, 226, 153, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 35, 154, 0, 198, 132, 154, 0, 201, 128, 21, 2, 198, 194, 155, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 156, 0, 198, 33, 156, 0, 198, 65, 156, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 97, 156, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 5, 0, 70, 0, 152, 89, 117, 0, 5, 0, 5, 7, 154, 89, 117, 0, 5, 0, 70, 7, 154, 89, 117, 0, 5, 0, 5, 14, 154, 89, 117, 0, 5, 0, 70, 14, 154, 89, 117, 0, 5, 0, 5, 21, 154, 89, 117, 0, 5, 0, 70, 21, 154, 89, 117, 0, 5, 0, 70, 28, 154, 89, 117, 0, 5, 0, 5, 28, 154, 89, 117, 0, 5, 0, 5, 35, 154, 89, 117, 0, 5, 0, 70, 35, 154, 89, 117, 0, 5, 0, 5, 42, 154, 89, 117, 0, 5, 0, 70, 42, 154, 89, 117, 0, 5, 0, 5, 49, 154, 89, 117, 0, 5, 0, 70, 49, 154, 89, 117, 0, 5, 0, 5, 56, 154, 89, 117, 0, 5, 0, 5, 63, 154, 89, 117, 0, 5, 0, 70, 63, 154, 89, 117, 0, 5, 0, 5, 70, 154, 89, 117, 0, 5, 0, 70, 70, 154, 89, 117, 0, 5, 0, 5, 77, 154, 89, 117, 0, 5, 0, 70, 77, 154, 89, 117, 0, 5, 0, 5, 84, 154, 89, 117, 0, 5, 0, 70, 84, 154, 89, 117, 0, 5, 0, 5, 91, 154, 89, 117, 0, 5, 0, 70, 91, 154, 89, 117, 0, 5, 0, 5, 98, 154, 89, 117, 0, 5, 0, 70, 98, 154, 89, 117, 0, 5, 0, 5, 105, 154, 89, 117, 0, 5, 0, 70, 105, 154, 89, 117, 0, 5, 0, 5, 112, 154, 89, 117, 0, 5, 0, 70, 112, 154, 89, 117, 0, 5, 0, 5, 119, 154, 89, 117, 0, 5, 0, 5, 0, 119, 80, 117, 0, 5, 0, 5, 0, 33, 81, 117, 0, 5, 0, 70, 28, 154, 89, 117, 0, 5, 0, 5, 0, 81, 93, 117, 0, 5, 0, 70, 119, 154, 89, 117, 0, 5, 0, 5, 126, 154, 89, 117, 0, 5, 0, 70, 126, 154, 89, 117, 0, 5, 0, 5, 133, 154, 89, 117, 0, 5, 0, 70, 133, 154, 89, 117, 0, 5, 0, 5, 140, 154, 89, 117, 0, 5, 0, 70, 140, 154, 89, 117, 0, 5, 0, 70, 140, 154, 89, 117, 0, 5, 0, 5, 0, 168, 80, 117, 0, 5, 0, 70, 150, 154, 89, 117, 0, 5, 0, 5, 147, 154, 89, 117, 0, 5, 0, 70, 147, 154, 89, 117, 0, 5, 0, 5, 154, 154, 89, 117, 0, 5, 0, 70, 154, 154, 89, 117, 0, 5, 0, 5, 161, 154, 89, 117, 0, 5, 0, 70, 161, 154, 89, 117, 0, 5, 0, 5, 168, 154, 89, 117, 0, 5, 0, 70, 168, 154, 89, 117, 0, 5, 0, 5, 175, 154, 89, 117, 0, 5, 0, 70, 175, 154, 89, 117, 0, 5, 0, 5, 192, 154, 89, 117, 0, 5, 0, 70, 192, 154, 89, 117, 0, 5, 0, 5, 199, 154, 89, 117, 0, 5, 0, 70, 199, 154, 89, 117, 0, 5, 0, 5, 0, 89, 81, 117, 0, 5, 0, 70, 0, 152, 89, 117, 0, 5, 0, 5, 0, 47, 81, 117, 0, 5, 0, 5, 0, 0, 87, 117, 0, 5, 0, 5, 206, 154, 89, 117, 0, 5, 0, 5, 0, 89, 81, 117, 0, 5, 0, 70, 0, 152, 89, 117, 0, 5, 0, 5, 0, 12, 81, 117, 0, 5, 0, 70, 0, 152, 89, 117, 0, 5, 0, 70, 206, 154, 89, 117, 0, 5, 0, 5, 213, 154, 89, 117, 0, 5, 0, 5, 220, 154, 89, 117, 0, 5, 0, 70, 220, 154, 89, 117, 0, 5, 0, 5, 227, 154, 89, 117, 0, 5, 0, 70, 227, 154, 89, 117, 0, 5, 0, 5, 234, 154, 89, 117, 0, 5, 0, 5, 241, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 0, 23, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 2, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 2, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 9, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 9, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 16, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 16, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 23, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 23, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 30, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 37, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 37, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 44, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 44, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 51, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 51, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 65, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 65, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 72, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 72, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 79, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 79, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 86, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 86, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 93, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 93, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 100, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 100, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 107, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 107, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 114, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 114, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 121, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 121, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 128, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 128, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 135, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 135, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 142, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 142, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 149, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 149, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 156, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 156, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 163, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 163, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 170, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 170, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 177, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 177, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 194, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 194, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 201, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 201, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 208, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 208, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 222, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 222, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 229, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 229, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 5, 0, 5, 222, 154, 89, 117, 0, 7, 0, 5, 0, 0, 83, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 0, 37, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 0, 51, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 3, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 3, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 10, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 10, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 17, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 17, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 24, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 24, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 31, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 38, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 38, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 45, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 45, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 52, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 52, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 66, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 66, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 73, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 73, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 80, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 80, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 87, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 87, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 94, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 94, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 101, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 101, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 108, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 108, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 115, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 115, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 122, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 122, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 129, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 129, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 136, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 136, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 143, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 143, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 150, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 150, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 157, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 157, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 164, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 164, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 171, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 171, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 183, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 183, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 195, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 195, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 202, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 202, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 209, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 209, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 223, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 223, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 230, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 230, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 0, 72, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 184, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 5, 0, 5, 223, 154, 89, 117, 0, 6, 0, 5, 0, 0, 83, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 0, 107, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 4, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 4, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 11, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 11, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 18, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 18, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 25, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 25, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 32, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 39, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 39, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 46, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 46, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 53, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 53, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 67, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 67, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 74, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 74, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 81, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 81, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 88, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 88, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 95, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 95, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 102, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 102, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 109, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 109, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 116, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 116, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 123, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 123, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 130, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 130, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 137, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 137, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 144, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 144, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 151, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 151, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 158, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 158, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 165, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 165, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 172, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 172, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 189, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 189, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 196, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 196, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 203, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 203, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 210, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 210, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 224, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 224, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 231, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 231, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 0, 107, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 0, 136, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 5, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 5, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 12, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 12, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 19, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 19, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 26, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 26, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 33, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 40, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 40, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 47, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 47, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 54, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 54, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 68, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 68, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 75, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 75, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 82, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 82, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 89, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 89, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 96, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 96, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 103, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 103, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 110, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 110, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 117, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 117, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 124, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 124, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 131, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 131, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 138, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 138, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 145, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 145, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 152, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 152, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 159, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 159, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 166, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 166, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 173, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 173, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 190, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 190, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 197, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 197, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 204, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 204, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 211, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 211, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 225, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 225, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 5, 232, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 7, 0, 70, 232, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 5, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 5, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 12, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 12, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 19, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 19, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 26, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 26, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 33, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 40, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 40, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 47, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 47, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 54, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 54, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 68, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 68, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 75, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 75, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 82, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 82, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 89, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 89, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 96, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 96, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 103, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 103, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 110, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 110, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 117, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 117, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 124, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 124, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 131, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 131, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 138, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 138, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 145, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 145, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 152, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 152, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 159, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 159, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 166, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 166, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 173, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 173, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 190, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 190, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 197, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 197, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 204, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 204, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 211, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 211, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 225, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 225, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 232, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 70, 232, 154, 89, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 6, 0, 5, 0, 137, 89, 117, 0, 5, 0, 71, 0, 153, 89, 117, 0, 5, 0, 72, 0, 153, 89, 117, 0, 5, 0, 71, 8, 154, 89, 117, 0, 5, 0, 72, 8, 154, 89, 117, 0, 5, 0, 71, 15, 154, 89, 117, 0, 5, 0, 72, 15, 154, 89, 117, 0, 5, 0, 71, 22, 154, 89, 117, 0, 5, 0, 72, 22, 154, 89, 117, 0, 5, 0, 72, 29, 154, 89, 117, 0, 5, 0, 71, 29, 154, 89, 117, 0, 5, 0, 71, 36, 154, 89, 117, 0, 5, 0, 72, 36, 154, 89, 117, 0, 5, 0, 71, 43, 154, 89, 117, 0, 5, 0, 72, 43, 154, 89, 117, 0, 5, 0, 71, 50, 154, 89, 117, 0, 5, 0, 72, 50, 154, 89, 117, 0, 5, 0, 70, 57, 154, 89, 117, 0, 5, 0, 71, 64, 154, 89, 117, 0, 5, 0, 72, 64, 154, 89, 117, 0, 5, 0, 71, 71, 154, 89, 117, 0, 5, 0, 72, 71, 154, 89, 117, 0, 5, 0, 71, 78, 154, 89, 117, 0, 5, 0, 72, 78, 154, 89, 117, 0, 5, 0, 71, 85, 154, 89, 117, 0, 5, 0, 72, 85, 154, 89, 117, 0, 5, 0, 71, 92, 154, 89, 117, 0, 5, 0, 72, 92, 154, 89, 117, 0, 5, 0, 71, 99, 154, 89, 117, 0, 5, 0, 72, 99, 154, 89, 117, 0, 5, 0, 71, 106, 154, 89, 117, 0, 5, 0, 72, 106, 154, 89, 117, 0, 5, 0, 71, 113, 154, 89, 117, 0, 5, 0, 72, 113, 154, 89, 117, 0, 5, 0, 71, 120, 154, 89, 117, 0, 5, 0, 72, 120, 154, 89, 117, 0, 5, 0, 71, 127, 154, 89, 117, 0, 5, 0, 72, 127, 154, 89, 117, 0, 5, 0, 71, 134, 154, 89, 117, 0, 5, 0, 72, 134, 154, 89, 117, 0, 5, 0, 71, 141, 154, 89, 117, 0, 5, 0, 72, 141, 154, 89, 117, 0, 5, 0, 71, 148, 154, 89, 117, 0, 5, 0, 72, 148, 154, 89, 117, 0, 5, 0, 71, 155, 154, 89, 117, 0, 5, 0, 72, 155, 154, 89, 117, 0, 5, 0, 71, 162, 154, 89, 117, 0, 5, 0, 72, 162, 154, 89, 117, 0, 5, 0, 71, 169, 154, 89, 117, 0, 5, 0, 72, 169, 154, 89, 117, 0, 5, 0, 71, 176, 154, 89, 117, 0, 5, 0, 72, 176, 154, 89, 117, 0, 5, 0, 71, 193, 154, 89, 117, 0, 5, 0, 72, 193, 154, 89, 117, 0, 5, 0, 71, 200, 154, 89, 117, 0, 5, 0, 72, 200, 154, 89, 117, 0, 5, 0, 71, 207, 154, 89, 117, 0, 5, 0, 72, 207, 154, 89, 117, 0, 5, 0, 70, 214, 154, 89, 117, 0, 5, 0, 71, 221, 154, 89, 117, 0, 5, 0, 72, 221, 154, 89, 117, 0, 5, 0, 71, 228, 154, 89, 117, 0, 5, 0, 72, 228, 154, 89, 117, 0, 5, 0, 70, 235, 154, 89, 117, 0, 5, 0, 70, 242, 154, 89, 117, 0, 5, 0, 70, 0, 153, 89, 117, 0, 5, 0, 5, 8, 154, 89, 117, 0, 5, 0, 70, 8, 154, 89, 117, 0, 5, 0, 5, 15, 154, 89, 117, 0, 5, 0, 70, 15, 154, 89, 117, 0, 5, 0, 5, 22, 154, 89, 117, 0, 5, 0, 70, 22, 154, 89, 117, 0, 5, 0, 70, 29, 154, 89, 117, 0, 5, 0, 5, 29, 154, 89, 117, 0, 5, 0, 5, 36, 154, 89, 117, 0, 5, 0, 70, 36, 154, 89, 117, 0, 5, 0, 5, 43, 154, 89, 117, 0, 5, 0, 70, 43, 154, 89, 117, 0, 5, 0, 5, 50, 154, 89, 117, 0, 5, 0, 70, 50, 154, 89, 117, 0, 5, 0, 5, 57, 154, 89, 117, 0, 5, 0, 5, 64, 154, 89, 117, 0, 5, 0, 70, 64, 154, 89, 117, 0, 5, 0, 5, 71, 154, 89, 117, 0, 5, 0, 70, 71, 154, 89, 117, 0, 5, 0, 5, 78, 154, 89, 117, 0, 5, 0, 70, 78, 154, 89, 117, 0, 5, 0, 5, 85, 154, 89, 117, 0, 5, 0, 70, 85, 154, 89, 117, 0, 5, 0, 5, 92, 154, 89, 117, 0, 5, 0, 70, 92, 154, 89, 117, 0, 5, 0, 5, 99, 154, 89, 117, 0, 5, 0, 70, 99, 154, 89, 117, 0, 5, 0, 5, 106, 154, 89, 117, 0, 5, 0, 70, 106, 154, 89, 117, 0, 5, 0, 5, 113, 154, 89, 117, 0, 5, 0, 70, 113, 154, 89, 117, 0, 5, 0, 5, 120, 154, 89, 117, 0, 5, 0, 70, 120, 154, 89, 117, 0, 5, 0, 5, 127, 154, 89, 117, 0, 5, 0, 70, 127, 154, 89, 117, 0, 5, 0, 5, 134, 154, 89, 117, 0, 5, 0, 70, 134, 154, 89, 117, 0, 5, 0, 5, 141, 154, 89, 117, 0, 5, 0, 70, 141, 154, 89, 117, 0, 5, 0, 5, 148, 154, 89, 117, 0, 5, 0, 70, 148, 154, 89, 117, 0, 5, 0, 5, 155, 154, 89, 117, 0, 5, 0, 70, 155, 154, 89, 117, 0, 5, 0, 5, 162, 154, 89, 117, 0, 5, 0, 70, 162, 154, 89, 117, 0, 5, 0, 5, 169, 154, 89, 117, 0, 5, 0, 70, 169, 154, 89, 117, 0, 5, 0, 5, 176, 154, 89, 117, 0, 5, 0, 70, 176, 154, 89, 117, 0, 5, 0, 5, 193, 154, 89, 117, 0, 5, 0, 70, 193, 154, 89, 117, 0, 5, 0, 5, 200, 154, 89, 117, 0, 5, 0, 70, 200, 154, 89, 117, 0, 5, 0, 5, 207, 154, 89, 117, 0, 5, 0, 70, 207, 154, 89, 117, 0, 5, 0, 5, 214, 154, 89, 117, 0, 5, 0, 5, 221, 154, 89, 117, 0, 5, 0, 70, 221, 154, 89, 117, 0, 5, 0, 5, 228, 154, 89, 117, 0, 5, 0, 70, 228, 154, 89, 117, 0, 5, 0, 5, 235, 154, 89, 117, 0, 5, 0, 5, 242, 154, 89, 117, 0, 5, 0, 5, 2, 154, 89, 117, 0, 5, 0, 70, 2, 154, 89, 117, 0, 5, 0, 5, 9, 154, 89, 117, 0, 5, 0, 70, 9, 154, 89, 117, 0, 5, 0, 5, 16, 154, 89, 117, 0, 5, 0, 70, 16, 154, 89, 117, 0, 5, 0, 5, 23, 154, 89, 117, 0, 5, 0, 70, 23, 154, 89, 117, 0, 5, 0, 70, 30, 154, 89, 117, 0, 5, 0, 5, 30, 154, 89, 117, 0, 5, 0, 5, 37, 154, 89, 117, 0, 5, 0, 70, 37, 154, 89, 117, 0, 5, 0, 5, 44, 154, 89, 117, 0, 5, 0, 70, 44, 154, 89, 117, 0, 5, 0, 5, 51, 154, 89, 117, 0, 5, 0, 70, 51, 154, 89, 117, 0, 5, 0, 5, 58, 154, 89, 117, 0, 5, 0, 5, 65, 154, 89, 117, 0, 5, 0, 70, 65, 154, 89, 117, 0, 5, 0, 5, 72, 154, 89, 117, 0, 5, 0, 70, 72, 154, 89, 117, 0, 5, 0, 5, 79, 154, 89, 117, 0, 5, 0, 70, 79, 154, 89, 117, 0, 5, 0, 5, 86, 154, 89, 117, 0, 5, 0, 70, 86, 154, 89, 117, 0, 5, 0, 5, 93, 154, 89, 117, 0, 5, 0, 70, 93, 154, 89, 117, 0, 5, 0, 5, 100, 154, 89, 117, 0, 5, 0, 70, 100, 154, 89, 117, 0, 5, 0, 5, 107, 154, 89, 117, 0, 5, 0, 70, 107, 154, 89, 117, 0, 5, 0, 5, 114, 154, 89, 117, 0, 5, 0, 70, 114, 154, 89, 117, 0, 5, 0, 5, 121, 154, 89, 117, 0, 5, 0, 70, 121, 154, 89, 117, 0, 5, 0, 5, 128, 154, 89, 117, 0, 5, 0, 70, 128, 154, 89, 117, 0, 5, 0, 5, 135, 154, 89, 117, 0, 5, 0, 70, 135, 154, 89, 117, 0, 5, 0, 5, 142, 154, 89, 117, 0, 5, 0, 70, 142, 154, 89, 117, 0, 5, 0, 5, 149, 154, 89, 117, 0, 5, 0, 70, 149, 154, 89, 117, 0, 5, 0, 5, 156, 154, 89, 117, 0, 5, 0, 70, 156, 154, 89, 117, 0, 5, 0, 5, 163, 154, 89, 117, 0, 5, 0, 70, 163, 154, 89, 117, 0, 5, 0, 5, 170, 154, 89, 117, 0, 5, 0, 70, 170, 154, 89, 117, 0, 5, 0, 5, 179, 154, 89, 117, 0, 5, 0, 5, 177, 154, 89, 117, 0, 5, 0, 70, 177, 154, 89, 117, 0, 5, 0, 5, 181, 154, 89, 117, 0, 5, 0, 5, 194, 154, 89, 117, 0, 5, 0, 70, 194, 154, 89, 117, 0, 5, 0, 5, 201, 154, 89, 117, 0, 5, 0, 70, 201, 154, 89, 117, 0, 5, 0, 5, 208, 154, 89, 117, 0, 5, 0, 70, 208, 154, 89, 117, 0, 5, 0, 5, 215, 154, 89, 117, 0, 5, 0, 70, 222, 154, 89, 117, 0, 5, 0, 5, 229, 154, 89, 117, 0, 5, 0, 70, 229, 154, 89, 117, 0, 5, 0, 5, 236, 154, 89, 117, 0, 5, 0, 5, 243, 154, 89, 117, 0, 5, 0, 5, 6, 154, 89, 117, 0, 5, 0, 70, 6, 154, 89, 117, 0, 5, 0, 5, 13, 154, 89, 117, 0, 5, 0, 70, 13, 154, 89, 117, 0, 5, 0, 5, 20, 154, 89, 117, 0, 5, 0, 70, 20, 154, 89, 117, 0, 5, 0, 5, 27, 154, 89, 117, 0, 5, 0, 70, 27, 154, 89, 117, 0, 5, 0, 70, 34, 154, 89, 117, 0, 5, 0, 5, 34, 154, 89, 117, 0, 5, 0, 5, 41, 154, 89, 117, 0, 5, 0, 70, 41, 154, 89, 117, 0, 5, 0, 5, 48, 154, 89, 117, 0, 5, 0, 70, 48, 154, 89, 117, 0, 5, 0, 5, 55, 154, 89, 117, 0, 5, 0, 70, 55, 154, 89, 117, 0, 5, 0, 5, 62, 154, 89, 117, 0, 5, 0, 5, 69, 154, 89, 117, 0, 5, 0, 70, 69, 154, 89, 117, 0, 5, 0, 5, 76, 154, 89, 117, 0, 5, 0, 70, 76, 154, 89, 117, 0, 5, 0, 5, 83, 154, 89, 117, 0, 5, 0, 70, 83, 154, 89, 117, 0, 5, 0, 5, 90, 154, 89, 117, 0, 5, 0, 70, 90, 154, 89, 117, 0, 5, 0, 5, 97, 154, 89, 117, 0, 5, 0, 70, 97, 154, 89, 117, 0, 5, 0, 5, 104, 154, 89, 117, 0, 5, 0, 70, 104, 154, 89, 117, 0, 5, 0, 5, 111, 154, 89, 117, 0, 5, 0, 70, 111, 154, 89, 117, 0, 5, 0, 5, 118, 154, 89, 117, 0, 5, 0, 70, 118, 154, 89, 117, 0, 5, 0, 5, 125, 154, 89, 117, 0, 5, 0, 70, 125, 154, 89, 117, 0, 5, 0, 5, 132, 154, 89, 117, 0, 5, 0, 70, 132, 154, 89, 117, 0, 5, 0, 5, 139, 154, 89, 117, 0, 5, 0, 70, 139, 154, 89, 117, 0, 5, 0, 5, 146, 154, 89, 117, 0, 5, 0, 70, 146, 154, 89, 117, 0, 5, 0, 5, 153, 154, 89, 117, 0, 5, 0, 70, 153, 154, 89, 117, 0, 5, 0, 5, 160, 154, 89, 117, 0, 5, 0, 70, 160, 154, 89, 117, 0, 5, 0, 5, 167, 154, 89, 117, 0, 5, 0, 70, 167, 154, 89, 117, 0, 5, 0, 5, 174, 154, 89, 117, 0, 5, 0, 70, 174, 154, 89, 117, 0, 5, 0, 5, 191, 154, 89, 117, 0, 5, 0, 70, 191, 154, 89, 117, 0, 5, 0, 5, 198, 154, 89, 117, 0, 5, 0, 70, 198, 154, 89, 117, 0, 5, 0, 5, 205, 154, 89, 117, 0, 5, 0, 70, 205, 154, 89, 117, 0, 5, 0, 5, 212, 154, 89, 117, 0, 5, 0, 70, 212, 154, 89, 117, 0, 5, 0, 5, 219, 154, 89, 117, 0, 5, 0, 5, 226, 154, 89, 117, 0, 5, 0, 70, 226, 154, 89, 117, 0, 5, 0, 5, 233, 154, 89, 117, 0, 5, 0, 70, 233, 154, 89, 117, 0, 5, 0, 5, 240, 154, 89, 117, 0, 5, 0, 5, 247, 154, 89, 117, 0, 5, 0, 5, 178, 154, 89, 117, 0, 5, 0, 5, 180, 154, 89, 117, 0, 5, 0, 5, 182, 154, 89, 117, 0, 5, 0, 5, 222, 154, 89, 117, 0, 5, 0, 5, 0, 0, 83, 117, 0, 5, 0, 5, 3, 154, 89, 117, 0, 5, 0, 70, 3, 154, 89, 117, 0, 5, 0, 5, 10, 154, 89, 117, 0, 5, 0, 70, 10, 154, 89, 117, 0, 5, 0, 5, 17, 154, 89, 117, 0, 5, 0, 70, 17, 154, 89, 117, 0, 5, 0, 5, 24, 154, 89, 117, 0, 5, 0, 70, 24, 154, 89, 117, 0, 5, 0, 70, 31, 154, 89, 117, 0, 5, 0, 5, 31, 154, 89, 117, 0, 5, 0, 5, 38, 154, 89, 117, 0, 5, 0, 70, 38, 154, 89, 117, 0, 5, 0, 5, 45, 154, 89, 117, 0, 5, 0, 70, 45, 154, 89, 117, 0, 5, 0, 5, 52, 154, 89, 117, 0, 5, 0, 70, 52, 154, 89, 117, 0, 5, 0, 5, 59, 154, 89, 117, 0, 5, 0, 5, 66, 154, 89, 117, 0, 5, 0, 70, 66, 154, 89, 117, 0, 5, 0, 5, 73, 154, 89, 117, 0, 5, 0, 70, 73, 154, 89, 117, 0, 5, 0, 5, 80, 154, 89, 117, 0, 5, 0, 70, 80, 154, 89, 117, 0, 5, 0, 5, 87, 154, 89, 117, 0, 5, 0, 70, 87, 154, 89, 117, 0, 5, 0, 5, 94, 154, 89, 117, 0, 5, 0, 70, 94, 154, 89, 117, 0, 5, 0, 5, 101, 154, 89, 117, 0, 5, 0, 70, 101, 154, 89, 117, 0, 5, 0, 5, 108, 154, 89, 117, 0, 5, 0, 70, 108, 154, 89, 117, 0, 5, 0, 5, 115, 154, 89, 117, 0, 5, 0, 70, 115, 154, 89, 117, 0, 5, 0, 5, 122, 154, 89, 117, 0, 5, 0, 70, 122, 154, 89, 117, 0, 5, 0, 5, 129, 154, 89, 117, 0, 5, 0, 70, 129, 154, 89, 117, 0, 5, 0, 5, 136, 154, 89, 117, 0, 5, 0, 70, 136, 154, 89, 117, 0, 5, 0, 5, 143, 154, 89, 117, 0, 5, 0, 70, 143, 154, 89, 117, 0, 5, 0, 5, 150, 154, 89, 117, 0, 5, 0, 5, 157, 154, 89, 117, 0, 5, 0, 70, 157, 154, 89, 117, 0, 5, 0, 5, 164, 154, 89, 117, 0, 5, 0, 70, 164, 154, 89, 117, 0, 5, 0, 5, 171, 154, 89, 117, 0, 5, 0, 70, 171, 154, 89, 117, 0, 5, 0, 5, 185, 154, 89, 117, 0, 5, 0, 5, 183, 154, 89, 117, 0, 5, 0, 70, 183, 154, 89, 117, 0, 5, 0, 5, 187, 154, 89, 117, 0, 5, 0, 5, 195, 154, 89, 117, 0, 5, 0, 70, 195, 154, 89, 117, 0, 5, 0, 5, 202, 154, 89, 117, 0, 5, 0, 70, 202, 154, 89, 117, 0, 5, 0, 5, 209, 154, 89, 117, 0, 5, 0, 70, 209, 154, 89, 117, 0, 5, 0, 5, 216, 154, 89, 117, 0, 5, 0, 70, 223, 154, 89, 117, 0, 5, 0, 5, 230, 154, 89, 117, 0, 5, 0, 70, 230, 154, 89, 117, 0, 5, 0, 5, 237, 154, 89, 117, 0, 5, 0, 5, 244, 154, 89, 117, 0, 5, 0, 5, 184, 154, 89, 117, 0, 5, 0, 5, 186, 154, 89, 117, 0, 5, 0, 5, 188, 154, 89, 117, 0, 5, 0, 5, 223, 154, 89, 117, 0, 5, 0, 5, 0, 0, 83, 117, 0, 5, 0, 5, 4, 154, 89, 117, 0, 5, 0, 70, 4, 154, 89, 117, 0, 5, 0, 5, 11, 154, 89, 117, 0, 5, 0, 70, 11, 154, 89, 117, 0, 5, 0, 5, 18, 154, 89, 117, 0, 5, 0, 70, 18, 154, 89, 117, 0, 5, 0, 5, 25, 154, 89, 117, 0, 5, 0, 70, 25, 154, 89, 117, 0, 5, 0, 70, 32, 154, 89, 117, 0, 5, 0, 5, 32, 154, 89, 117, 0, 5, 0, 5, 39, 154, 89, 117, 0, 5, 0, 70, 39, 154, 89, 117, 0, 5, 0, 5, 46, 154, 89, 117, 0, 5, 0, 70, 46, 154, 89, 117, 0, 5, 0, 5, 53, 154, 89, 117, 0, 5, 0, 70, 53, 154, 89, 117, 0, 5, 0, 5, 60, 154, 89, 117, 0, 5, 0, 5, 67, 154, 89, 117, 0, 5, 0, 70, 67, 154, 89, 117, 0, 5, 0, 5, 74, 154, 89, 117, 0, 5, 0, 70, 74, 154, 89, 117, 0, 5, 0, 5, 81, 154, 89, 117, 0, 5, 0, 70, 81, 154, 89, 117, 0, 5, 0, 5, 88, 154, 89, 117, 0, 5, 0, 70, 88, 154, 89, 117, 0, 5, 0, 5, 95, 154, 89, 117, 0, 5, 0, 70, 95, 154, 89, 117, 0, 5, 0, 5, 102, 154, 89, 117, 0, 5, 0, 70, 102, 154, 89, 117, 0, 5, 0, 5, 109, 154, 89, 117, 0, 5, 0, 70, 109, 154, 89, 117, 0, 5, 0, 5, 116, 154, 89, 117, 0, 5, 0, 70, 116, 154, 89, 117, 0, 5, 0, 5, 123, 154, 89, 117, 0, 5, 0, 70, 123, 154, 89, 117, 0, 5, 0, 5, 130, 154, 89, 117, 0, 5, 0, 70, 130, 154, 89, 117, 0, 5, 0, 5, 137, 154, 89, 117, 0, 5, 0, 70, 137, 154, 89, 117, 0, 5, 0, 5, 144, 154, 89, 117, 0, 5, 0, 70, 144, 154, 89, 117, 0, 5, 0, 5, 151, 154, 89, 117, 0, 5, 0, 70, 151, 154, 89, 117, 0, 5, 0, 5, 158, 154, 89, 117, 0, 5, 0, 70, 158, 154, 89, 117, 0, 5, 0, 5, 165, 154, 89, 117, 0, 5, 0, 70, 165, 154, 89, 117, 0, 5, 0, 5, 172, 154, 89, 117, 0, 5, 0, 70, 172, 154, 89, 117, 0, 5, 0, 5, 189, 154, 89, 117, 0, 5, 0, 70, 189, 154, 89, 117, 0, 5, 0, 5, 196, 154, 89, 117, 0, 5, 0, 70, 196, 154, 89, 117, 0, 5, 0, 5, 203, 154, 89, 117, 0, 5, 0, 70, 203, 154, 89, 117, 0, 5, 0, 5, 210, 154, 89, 117, 0, 5, 0, 70, 210, 154, 89, 117, 0, 5, 0, 5, 217, 154, 89, 117, 0, 5, 0, 5, 224, 154, 89, 117, 0, 5, 0, 70, 224, 154, 89, 117, 0, 5, 0, 5, 231, 154, 89, 117, 0, 5, 0, 70, 231, 154, 89, 117, 0, 5, 0, 5, 238, 154, 89, 117, 0, 5, 0, 5, 245, 154, 89, 117, 0, 5, 0, 70, 0, 136, 89, 117, 0, 5, 0, 71, 5, 154, 89, 117, 0, 5, 0, 72, 5, 154, 89, 117, 0, 5, 0, 71, 12, 154, 89, 117, 0, 5, 0, 72, 12, 154, 89, 117, 0, 5, 0, 71, 19, 154, 89, 117, 0, 5, 0, 72, 19, 154, 89, 117, 0, 5, 0, 71, 26, 154, 89, 117, 0, 5, 0, 72, 26, 154, 89, 117, 0, 5, 0, 72, 33, 154, 89, 117, 0, 5, 0, 71, 33, 154, 89, 117, 0, 5, 0, 71, 40, 154, 89, 117, 0, 5, 0, 72, 40, 154, 89, 117, 0, 5, 0, 71, 47, 154, 89, 117, 0, 5, 0, 72, 47, 154, 89, 117, 0, 5, 0, 71, 54, 154, 89, 117, 0, 5, 0, 72, 54, 154, 89, 117, 0, 5, 0, 70, 61, 154, 89, 117, 0, 5, 0, 71, 68, 154, 89, 117, 0, 5, 0, 72, 68, 154, 89, 117, 0, 5, 0, 71, 75, 154, 89, 117, 0, 5, 0, 72, 75, 154, 89, 117, 0, 5, 0, 71, 82, 154, 89, 117, 0, 5, 0, 72, 82, 154, 89, 117, 0, 5, 0, 71, 89, 154, 89, 117, 0, 5, 0, 72, 89, 154, 89, 117, 0, 5, 0, 71, 96, 154, 89, 117, 0, 5, 0, 72, 96, 154, 89, 117, 0, 5, 0, 71, 103, 154, 89, 117, 0, 5, 0, 72, 103, 154, 89, 117, 0, 5, 0, 71, 110, 154, 89, 117, 0, 5, 0, 72, 110, 154, 89, 117, 0, 5, 0, 71, 117, 154, 89, 117, 0, 5, 0, 72, 117, 154, 89, 117, 0, 5, 0, 71, 124, 154, 89, 117, 0, 5, 0, 72, 124, 154, 89, 117, 0, 5, 0, 71, 131, 154, 89, 117, 0, 5, 0, 72, 131, 154, 89, 117, 0, 5, 0, 71, 138, 154, 89, 117, 0, 5, 0, 72, 138, 154, 89, 117, 0, 5, 0, 71, 145, 154, 89, 117, 0, 5, 0, 72, 145, 154, 89, 117, 0, 5, 0, 71, 152, 154, 89, 117, 0, 5, 0, 72, 152, 154, 89, 117, 0, 5, 0, 71, 159, 154, 89, 117, 0, 5, 0, 72, 159, 154, 89, 117, 0, 5, 0, 71, 166, 154, 89, 117, 0, 5, 0, 72, 166, 154, 89, 117, 0, 5, 0, 71, 173, 154, 89, 117, 0, 5, 0, 72, 173, 154, 89, 117, 0, 5, 0, 71, 190, 154, 89, 117, 0, 5, 0, 72, 190, 154, 89, 117, 0, 5, 0, 71, 197, 154, 89, 117, 0, 5, 0, 72, 197, 154, 89, 117, 0, 5, 0, 71, 204, 154, 89, 117, 0, 5, 0, 72, 204, 154, 89, 117, 0, 5, 0, 71, 211, 154, 89, 117, 0, 5, 0, 72, 211, 154, 89, 117, 0, 5, 0, 70, 218, 154, 89, 117, 0, 5, 0, 71, 225, 154, 89, 117, 0, 5, 0, 72, 225, 154, 89, 117, 0, 5, 0, 71, 232, 154, 89, 117, 0, 5, 0, 72, 232, 154, 89, 117, 0, 5, 0, 70, 239, 154, 89, 117, 0, 5, 0, 70, 246, 154, 89, 117, 0, 5, 0, 70, 0, 137, 89, 117, 0, 5, 0, 5, 5, 154, 89, 117, 0, 5, 0, 70, 5, 154, 89, 117, 0, 5, 0, 70, 5, 154, 89, 117, 0, 5, 0, 5, 0, 9, 79, 117, 0, 5, 0, 5, 250, 154, 89, 117, 0, 5, 0, 5, 12, 154, 89, 117, 0, 5, 0, 70, 12, 154, 89, 117, 0, 5, 0, 5, 19, 154, 89, 117, 0, 5, 0, 70, 19, 154, 89, 117, 0, 5, 0, 5, 26, 154, 89, 117, 0, 5, 0, 70, 26, 154, 89, 117, 0, 5, 0, 70, 33, 154, 89, 117, 0, 5, 0, 5, 33, 154, 89, 117, 0, 5, 0, 5, 40, 154, 89, 117, 0, 5, 0, 70, 40, 154, 89, 117, 0, 5, 0, 5, 47, 154, 89, 117, 0, 5, 0, 70, 47, 154, 89, 117, 0, 5, 0, 5, 54, 154, 89, 117, 0, 5, 0, 70, 54, 154, 89, 117, 0, 5, 0, 5, 61, 154, 89, 117, 0, 5, 0, 5, 68, 154, 89, 117, 0, 5, 0, 70, 68, 154, 89, 117, 0, 5, 0, 5, 75, 154, 89, 117, 0, 5, 0, 70, 75, 154, 89, 117, 0, 5, 0, 5, 82, 154, 89, 117, 0, 5, 0, 70, 82, 154, 89, 117, 0, 5, 0, 5, 89, 154, 89, 117, 0, 5, 0, 70, 89, 154, 89, 117, 0, 5, 0, 5, 96, 154, 89, 117, 0, 5, 0, 70, 96, 154, 89, 117, 0, 5, 0, 5, 103, 154, 89, 117, 0, 5, 0, 70, 103, 154, 89, 117, 0, 5, 0, 5, 110, 154, 89, 117, 0, 5, 0, 70, 110, 154, 89, 117, 0, 5, 0, 5, 117, 154, 89, 117, 0, 5, 0, 70, 117, 154, 89, 117, 0, 5, 0, 5, 124, 154, 89, 117, 0, 5, 0, 70, 124, 154, 89, 117, 0, 5, 0, 5, 131, 154, 89, 117, 0, 5, 0, 70, 131, 154, 89, 117, 0, 5, 0, 5, 138, 154, 89, 117, 0, 5, 0, 70, 138, 154, 89, 117, 0, 5, 0, 5, 145, 154, 89, 117, 0, 5, 0, 70, 145, 154, 89, 117, 0, 5, 0, 5, 152, 154, 89, 117, 0, 5, 0, 70, 152, 154, 89, 117, 0, 5, 0, 5, 159, 154, 89, 117, 0, 5, 0, 70, 159, 154, 89, 117, 0, 5, 0, 5, 166, 154, 89, 117, 0, 5, 0, 70, 166, 154, 89, 117, 0, 5, 0, 5, 173, 154, 89, 117, 0, 5, 0, 70, 173, 154, 89, 117, 0, 5, 0, 5, 190, 154, 89, 117, 0, 5, 0, 70, 190, 154, 89, 117, 0, 5, 0, 5, 197, 154, 89, 117, 0, 5, 0, 70, 197, 154, 89, 117, 0, 5, 0, 5, 204, 154, 89, 117, 0, 5, 0, 70, 204, 154, 89, 117, 0, 5, 0, 5, 211, 154, 89, 117, 0, 5, 0, 70, 211, 154, 89, 117, 0, 5, 0, 5, 218, 154, 89, 117, 0, 5, 0, 5, 225, 154, 89, 117, 0, 5, 0, 70, 225, 154, 89, 117, 0, 5, 0, 5, 232, 154, 89, 117, 0, 5, 0, 70, 232, 154, 89, 117, 0, 5, 0, 5, 239, 154, 89, 117, 0, 5, 0, 5, 246, 154, 89, 117, 0, 5, 0, 5, 224, 154, 89, 117, 0, 5, 0, 5, 0, 0, 83, 117, 0, 5, 0, 5, 2, 155, 89, 117, 0, 5, 0, 5, 7, 155, 89, 117, 0, 5, 0, 5, 4, 155, 89, 117, 0, 5, 0, 5, 251, 154, 89, 117, 0, 5, 0, 5, 3, 155, 89, 117, 0, 5, 0, 5, 8, 155, 89, 117, 0, 5, 0, 5, 5, 155, 89, 117, 0, 5, 0, 5, 253, 154, 89, 117, 0, 5, 0, 5, 6, 155, 89, 117, 0, 5, 0, 5, 255, 154, 89, 117, 0, 5, 0, 5, 220, 154, 89, 117, 0, 5, 0, 5, 0, 0, 83, 117, 0, 5, 0, 5, 0, 168, 80, 117, 0, 5, 0, 5, 255, 154, 89, 117, 0, 5, 0, 71, 6, 154, 89, 117, 0, 5, 0, 5, 0, 61, 81, 117, 0, 5, 0, 5, 253, 154, 89, 117, 0, 5, 0, 5, 0, 107, 89, 117, 0, 5, 0, 70, 0, 80, 93, 117, 0, 5, 0, 5, 0, 89, 81, 117, 0, 5, 0, 70, 70, 154, 89, 117, 0, 5, 0, 5, 0, 81, 93, 117, 0, 5, 0, 5, 0, 9, 79, 117, 0, 5, 0, 70, 33, 154, 89, 117, 0, 5, 0, 70, 0, 81, 93, 117, 0, 5, 0, 5, 0, 100, 85, 117, 0, 5, 0, 70, 0, 23, 89, 117, 0, 5, 0, 5, 248, 154, 89, 117, 0, 5, 0, 5, 249, 154, 89, 117, 0, 5, 0, 5, 252, 154, 89, 117, 0, 5, 0, 5, 254, 154, 89, 117, 0, 0, 0, 0, 5, 5, 83, 117, 193, 33, 81, 117, 193, 37, 89, 117, 193, 100, 85, 117, 193, 23, 89, 117, 193, 86, 79, 117, 10, 5, 91, 117, 193, 100, 85, 117, 193, 51, 89, 117, 193, 86, 79, 117, 9, 5, 91, 117, 193, 100, 85, 117, 193, 107, 89, 117, 193, 86, 79, 117, 8, 5, 91, 117, 193, 100, 85, 117, 193, 136, 89, 117, 193, 86, 79, 117, 7, 5, 91, 117, 193, 100, 85, 117, 193, 136, 89, 117, 193, 86, 79, 117, 6, 5, 91, 117, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 79, 117, 193, 9, 1, 0, 57, 16, 255, 255, 89, 117, 193, 152, 58, 16, 198, 129, 79, 117, 193, 23, 1, 0, 57, 16, 198, 161, 58, 16, 0, 192, 198, 65, 79, 117, 193, 37, 1, 0, 57, 16, 0, 192, 198, 97, 58, 16, 0, 192, 198, 129, 79, 117, 193, 65, 1, 0, 57, 16, 0, 192, 198, 161, 58, 16, 0, 192, 198, 193, 79, 117, 193, 86, 48, 0, 58, 16, 112, 64, 198, 225, 57, 16, 1, 192, 198, 1, 79, 117, 193, 100, 1, 0, 57, 16, 1, 192, 198, 33, 58, 16, 1, 192, 198, 65, 79, 117, 193, 121, 1, 0, 57, 16, 1, 192, 198, 97, 58, 16, 1, 192, 198, 129, 79, 117, 193, 149, 1, 0, 57, 16, 1, 192, 198, 161, 58, 16, 1, 192, 198, 193, 79, 117, 193, 184, 48, 0, 58, 16, 1, 192, 198, 225, 79, 117, 193, 233, 1, 0, 57, 16, 2, 192, 198, 1, 58, 16, 2, 192, 198, 33, 80, 117, 193, 7, 1, 0, 57, 16, 2, 192, 198, 65, 58, 16, 2, 192, 198, 97, 80, 117, 193, 14, 1, 0, 57, 16, 2, 192, 198, 129, 58, 16, 2, 192, 198, 161, 80, 117, 193, 28, 1, 0, 57, 16, 2, 192, 198, 193, 58, 16, 2, 192, 198, 225, 80, 117, 193, 42, 1, 0, 57, 16, 3, 192, 198, 1, 58, 16, 3, 192, 198, 33, 80, 117, 193, 63, 1, 0, 57, 16, 3, 192, 198, 65, 58, 16, 3, 192, 198, 97, 80, 117, 193, 84, 1, 0, 57, 16, 3, 192, 198, 129, 58, 16, 3, 192, 198, 161, 80, 117, 193, 112, 1, 0, 57, 16, 3, 192, 198, 193, 58, 16, 3, 192, 198, 225, 80, 117, 193, 119, 1, 0, 57, 16, 3, 0, 58, 16, 4, 192, 198, 161, 115, 65, 198, 1, 25, 16, 4, 16, 58, 16, 56, 16, 4, 192, 198, 36, 80, 117, 193, 126, 1, 0, 57, 16, 4, 192, 198, 193, 58, 16, 4, 192, 198, 225, 80, 117, 193, 147, 1, 0, 57, 16, 5, 192, 198, 1, 58, 16, 5, 192, 198, 33, 80, 117, 193, 168, 1, 0, 57, 16, 5, 192, 198, 65, 58, 16, 178, 65, 198, 97, 47, 16, 21, 16, 58, 16, 5, 192, 198, 131, 80, 117, 193, 196, 1, 0, 57, 16, 5, 192, 198, 225, 58, 16, 6, 192, 198, 1, 80, 117, 193, 203, 1, 0, 57, 16, 6, 192, 198, 33, 58, 16, 6, 192, 198, 65, 80, 117, 193, 245, 1, 0, 57, 16, 6, 192, 198, 97, 58, 16, 6, 192, 198, 129, 81, 117, 193, 12, 1, 0, 57, 16, 6, 192, 198, 161, 58, 16, 6, 192, 198, 193, 81, 117, 193, 33, 1, 0, 57, 16, 6, 192, 198, 225, 58, 16, 7, 192, 198, 1, 81, 117, 193, 47, 1, 0, 57, 16, 7, 192, 198, 33, 58, 16, 7, 192, 198, 65, 81, 117, 193, 61, 1, 0, 57, 16, 7, 192, 198, 97, 58, 16, 7, 192, 198, 129, 81, 117, 193, 89, 2, 0, 0, 16, 5, 0, 57, 16, 9, 0, 58, 16, 8, 192, 198, 193, 50, 0, 58, 16, 59, 16, 44, 16, 7, 192, 198, 164, 114, 66, 198, 33, 24, 16, 0, 16, 58, 16, 8, 192, 198, 68, 81, 117, 193, 103, 48, 0, 58, 16, 8, 192, 198, 225, 83, 117, 5, 5, 1, 0, 57, 16, 3, 0, 58, 16, 9, 192, 198, 33, 177, 66, 198, 1, 25, 16, 46, 16, 197, 163, 85, 117, 193, 9, 1, 0, 57, 16, 9, 192, 198, 65, 58, 16, 9, 192, 198, 97, 85, 117, 193, 51, 48, 0, 58, 16, 9, 192, 198, 129, 85, 117, 193, 100, 48, 0, 58, 16, 9, 192, 198, 161, 9, 0, 198, 194, 30, 0, 16, 16, 141, 0, 24, 16, 67, 0, 28, 16, 30, 0, 28, 16, 7, 0, 30, 16, 12, 0, 31, 16, 17, 0, 63, 16, 24, 192, 198, 195, 1, 0, 57, 16, 23, 192, 198, 66, 58, 16, 23, 192, 198, 130, 1, 0, 57, 16, 23, 192, 198, 194, 58, 16, 24, 192, 198, 2, 1, 0, 57, 16, 24, 192, 198, 66, 58, 16, 24, 192, 198, 130, 24, 16, 12, 0, 25, 16, 17, 0, 26, 16, 22, 0, 27, 16, 1, 0, 57, 16, 22, 192, 198, 194, 58, 16, 23, 192, 198, 2, 1, 0, 57, 16, 21, 192, 198, 66, 58, 16, 21, 192, 198, 130, 1, 0, 57, 16, 21, 192, 198, 194, 58, 16, 22, 192, 198, 2, 1, 0, 57, 16, 22, 192, 198, 66, 58, 16, 22, 192, 198, 130, 20, 16, 35, 0, 20, 16, 12, 0, 21, 16, 17, 0, 22, 16, 22, 0, 23, 16, 1, 0, 57, 16, 20, 192, 198, 194, 58, 16, 21, 192, 198, 2, 1, 0, 57, 16, 19, 192, 198, 66, 58, 16, 19, 192, 198, 130, 1, 0, 57, 16, 19, 192, 198, 194, 58, 16, 20, 192, 198, 2, 1, 0, 57, 16, 20, 192, 198, 66, 58, 16, 20, 192, 198, 130, 16, 16, 12, 0, 17, 16, 17, 0, 18, 16, 22, 0, 19, 16, 1, 0, 57, 16, 18, 192, 198, 194, 58, 16, 19, 192, 198, 2, 1, 0, 57, 16, 17, 192, 198, 66, 58, 16, 17, 192, 198, 130, 1, 0, 57, 16, 17, 192, 198, 194, 58, 16, 18, 192, 198, 2, 1, 0, 57, 16, 18, 192, 198, 66, 58, 16, 18, 192, 198, 130, 7, 16, 72, 0, 12, 16, 35, 0, 12, 16, 12, 0, 13, 16, 17, 0, 14, 16, 22, 0, 15, 16, 1, 0, 57, 16, 16, 192, 198, 194, 58, 16, 17, 192, 198, 2, 1, 0, 57, 16, 15, 192, 198, 66, 58, 16, 15, 192, 198, 130, 1, 0, 57, 16, 15, 192, 198, 194, 58, 16, 16, 192, 198, 2, 1, 0, 57, 16, 16, 192, 198, 66, 58, 16, 16, 192, 198, 130, 7, 16, 12, 0, 9, 16, 17, 0, 10, 16, 22, 0, 11, 16, 1, 0, 57, 16, 14, 192, 198, 194, 58, 16, 15, 192, 198, 2, 1, 0, 57, 16, 13, 192, 198, 66, 58, 16, 13, 192, 198, 130, 1, 0, 57, 16, 13, 192, 198, 194, 58, 16, 14, 192, 198, 2, 1, 0, 57, 16, 14, 192, 198, 66, 58, 16, 14, 192, 198, 130, 3, 16, 35, 0, 3, 16, 12, 0, 4, 16, 17, 0, 5, 16, 22, 0, 6, 16, 1, 0, 57, 16, 12, 192, 198, 194, 58, 16, 13, 192, 198, 2, 1, 0, 57, 16, 11, 192, 198, 130, 58, 16, 11, 192, 198, 194, 1, 0, 57, 16, 0, 192, 197, 132, 58, 16, 12, 192, 198, 2, 1, 0, 57, 16, 12, 192, 198, 66, 58, 16, 12, 192, 198, 130, 0, 16, 10, 0, 1, 16, 15, 0, 2, 16, 1, 0, 57, 16, 11, 192, 198, 2, 58, 16, 11, 192, 198, 66, 1, 0, 57, 16, 10, 192, 198, 2, 58, 16, 10, 192, 198, 66, 1, 0, 57, 16, 10, 192, 198, 130, 58, 16, 10, 192, 198, 194, 25, 0, 198, 98, 32, 0, 17, 16, 141, 0, 25, 16, 67, 0, 30, 16, 30, 0, 30, 16, 10, 0, 31, 16, 15, 0, 46, 16, 40, 192, 198, 98, 47, 16, 17, 0, 63, 16, 40, 192, 198, 227, 1, 0, 57, 16, 39, 192, 198, 98, 58, 16, 39, 192, 198, 162, 1, 0, 57, 16, 39, 192, 198, 226, 58, 16, 40, 192, 198, 34, 48, 0, 54, 16, 40, 192, 198, 162, 25, 16, 12, 0, 26, 16, 17, 0, 27, 16, 22, 0, 28, 16, 1, 0, 57, 16, 38, 192, 198, 226, 58, 16, 39, 192, 198, 34, 1, 0, 57, 16, 37, 192, 198, 98, 58, 16, 37, 192, 198, 162, 1, 0, 57, 16, 37, 192, 198, 226, 58, 16, 38, 192, 198, 34, 1, 0, 57, 16, 38, 192, 198, 98, 58, 16, 38, 192, 198, 162, 21, 16, 35, 0, 21, 16, 12, 0, 22, 16, 17, 0, 23, 16, 22, 0, 24, 16, 1, 0, 57, 16, 36, 192, 198, 226, 58, 16, 37, 192, 198, 34, 1, 0, 57, 16, 35, 192, 198, 98, 58, 16, 35, 192, 198, 162, 1, 0, 57, 16, 35, 192, 198, 226, 58, 16, 36, 192, 198, 34, 1, 0, 57, 16, 36, 192, 198, 98, 58, 16, 36, 192, 198, 162, 17, 16, 12, 0, 18, 16, 17, 0, 19, 16, 22, 0, 20, 16, 1, 0, 57, 16, 34, 192, 198, 226, 58, 16, 35, 192, 198, 34, 1, 0, 57, 16, 33, 192, 198, 98, 58, 16, 33, 192, 198, 162, 1, 0, 57, 16, 33, 192, 198, 226, 58, 16, 34, 192, 198, 34, 1, 0, 57, 16, 34, 192, 198, 98, 58, 16, 34, 192, 198, 162, 9, 16, 72, 0, 13, 16, 35, 0, 13, 16, 12, 0, 14, 16, 17, 0, 15, 16, 22, 0, 16, 16, 1, 0, 57, 16, 32, 192, 198, 226, 58, 16, 33, 192, 198, 34, 1, 0, 57, 16, 31, 192, 198, 98, 58, 16, 31, 192, 198, 162, 1, 0, 57, 16, 31, 192, 198, 226, 58, 16, 32, 192, 198, 34, 1, 0, 57, 16, 32, 192, 198, 98, 58, 16, 32, 192, 198, 162, 9, 16, 12, 0, 10, 16, 17, 0, 11, 16, 22, 0, 12, 16, 1, 0, 57, 16, 30, 192, 198, 226, 58, 16, 31, 192, 198, 34, 1, 0, 57, 16, 29, 192, 198, 98, 58, 16, 29, 192, 198, 162, 1, 0, 57, 16, 29, 192, 198, 226, 58, 16, 30, 192, 198, 34, 1, 0, 57, 16, 30, 192, 198, 98, 58, 16, 30, 192, 198, 162, 4, 16, 35, 0, 4, 16, 12, 0, 5, 16, 17, 0, 6, 16, 22, 0, 7, 16, 1, 0, 57, 16, 28, 192, 198, 226, 58, 16, 29, 192, 198, 34, 1, 0, 57, 16, 1, 192, 197, 4, 58, 16, 27, 192, 198, 162, 1, 0, 57, 16, 27, 192, 198, 226, 58, 16, 28, 192, 198, 34, 1, 0, 57, 16, 28, 192, 198, 98, 58, 16, 28, 192, 198, 162, 0, 16, 12, 0, 1, 16, 17, 0, 2, 16, 22, 0, 3, 16, 1, 0, 57, 16, 27, 192, 198, 34, 58, 16, 27, 192, 198, 98, 1, 0, 57, 16, 25, 192, 198, 162, 58, 16, 25, 192, 198, 226, 1, 0, 57, 16, 26, 192, 198, 34, 58, 16, 26, 192, 198, 98, 1, 0, 57, 16, 26, 192, 198, 162, 58, 16, 26, 192, 198, 226, 41, 0, 198, 66, 29, 0, 16, 16, 137, 0, 23, 16, 72, 0, 27, 16, 35, 0, 27, 16, 12, 0, 28, 16, 17, 0, 30, 16, 22, 0, 31, 16, 1, 0, 57, 16, 55, 192, 198, 194, 58, 16, 56, 192, 198, 2, 1, 0, 57, 16, 54, 192, 198, 66, 58, 16, 54, 192, 198, 130, 1, 0, 57, 16, 54, 192, 198, 194, 58, 16, 55, 192, 198, 2, 1, 0, 57, 16, 55, 192, 198, 66, 58, 16, 55, 192, 198, 130, 23, 16, 12, 0, 24, 16, 17, 0, 25, 16, 22, 0, 26, 16, 1, 0, 57, 16, 53, 192, 198, 194, 58, 16, 54, 192, 198, 2, 1, 0, 57, 16, 52, 192, 198, 66, 58, 16, 52, 192, 198, 130, 1, 0, 57, 16, 52, 192, 198, 194, 58, 16, 53, 192, 198, 2, 1, 0, 57, 16, 53, 192, 198, 66, 58, 16, 53, 192, 198, 130, 19, 16, 35, 0, 19, 16, 12, 0, 20, 16, 17, 0, 21, 16, 22, 0, 22, 16, 1, 0, 57, 16, 51, 192, 198, 194, 58, 16, 52, 192, 198, 2, 1, 0, 57, 16, 50, 192, 198, 66, 58, 16, 50, 192, 198, 130, 1, 0, 57, 16, 50, 192, 198, 194, 58, 16, 51, 192, 198, 2, 1, 0, 57, 16, 51, 192, 198, 66, 58, 16, 51, 192, 198, 130, 16, 16, 10, 0, 17, 16, 15, 0, 18, 16, 1, 0, 57, 16, 49, 192, 198, 194, 58, 16, 50, 192, 198, 2, 1, 0, 57, 16, 48, 192, 198, 194, 58, 16, 49, 192, 198, 2, 1, 0, 57, 16, 49, 192, 198, 66, 58, 16, 49, 192, 198, 130, 7, 16, 72, 0, 12, 16, 35, 0, 12, 16, 12, 0, 13, 16, 17, 0, 14, 16, 22, 0, 15, 16, 1, 0, 57, 16, 48, 192, 198, 66, 58, 16, 48, 192, 198, 130, 1, 0, 57, 16, 46, 192, 198, 194, 58, 16, 47, 192, 198, 2, 1, 0, 57, 16, 47, 192, 198, 66, 58, 16, 47, 192, 198, 130, 1, 0, 57, 16, 47, 192, 198, 194, 58, 16, 48, 192, 198, 2, 7, 16, 12, 0, 9, 16, 17, 0, 10, 16, 22, 0, 11, 16, 1, 0, 57, 16, 46, 192, 198, 66, 58, 16, 46, 192, 198, 130, 1, 0, 57, 16, 44, 192, 198, 194, 58, 16, 45, 192, 198, 2, 1, 0, 57, 16, 45, 192, 198, 66, 58, 16, 45, 192, 198, 130, 1, 0, 57, 16, 45, 192, 198, 194, 58, 16, 46, 192, 198, 2, 3, 16, 35, 0, 3, 16, 12, 0, 4, 16, 17, 0, 5, 16, 22, 0, 6, 16, 1, 0, 57, 16, 44, 192, 198, 66, 58, 16, 44, 192, 198, 130, 1, 0, 57, 16, 43, 192, 198, 2, 58, 16, 43, 192, 198, 66, 1, 0, 57, 16, 1, 192, 197, 132, 58, 16, 43, 192, 198, 130, 1, 0, 57, 16, 43, 192, 198, 194, 58, 16, 44, 192, 198, 2, 0, 16, 10, 0, 1, 16, 15, 0, 2, 16, 1, 0, 57, 16, 42, 192, 198, 130, 58, 16, 42, 192, 198, 194, 1, 0, 57, 16, 41, 192, 198, 130, 58, 16, 41, 192, 198, 194, 1, 0, 57, 16, 42, 192, 198, 2, 58, 16, 42, 192, 198, 66, 56, 0, 198, 130, 30, 0, 16, 16, 162, 1, 24, 16, 88, 1, 28, 16, 51, 1, 28, 16, 28, 1, 30, 16, 33, 1, 31, 16, 38, 1, 49, 16, 48, 0, 44, 16, 29, 0, 16, 16, 137, 0, 23, 16, 72, 0, 27, 16, 35, 0, 27, 16, 12, 0, 28, 16, 17, 0, 30, 16, 22, 0, 31, 16, 1, 0, 57, 16, 85, 192, 198, 194, 58, 16, 86, 192, 198, 2, 1, 0, 57, 16, 84, 192, 198, 66, 58, 16, 84, 192, 198, 130, 1, 0, 57, 16, 84, 192, 198, 194, 58, 16, 85, 192, 198, 2, 1, 0, 57, 16, 85, 192, 198, 66, 58, 16, 85, 192, 198, 130, 23, 16, 12, 0, 24, 16, 17, 0, 25, 16, 22, 0, 26, 16, 1, 0, 57, 16, 83, 192, 198, 194, 58, 16, 84, 192, 198, 2, 1, 0, 57, 16, 82, 192, 198, 66, 58, 16, 82, 192, 198, 130, 1, 0, 57, 16, 82, 192, 198, 194, 58, 16, 83, 192, 198, 2, 1, 0, 57, 16, 83, 192, 198, 66, 58, 16, 83, 192, 198, 130, 19, 16, 35, 0, 19, 16, 12, 0, 20, 16, 17, 0, 21, 16, 22, 0, 22, 16, 1, 0, 57, 16, 81, 192, 198, 194, 58, 16, 82, 192, 198, 2, 1, 0, 57, 16, 80, 192, 198, 66, 58, 16, 80, 192, 198, 130, 1, 0, 57, 16, 80, 192, 198, 194, 58, 16, 81, 192, 198, 2, 1, 0, 57, 16, 81, 192, 198, 66, 58, 16, 81, 192, 198, 130, 16, 16, 10, 0, 17, 16, 15, 0, 18, 16, 1, 0, 57, 16, 79, 192, 198, 194, 58, 16, 80, 192, 198, 2, 1, 0, 57, 16, 78, 192, 198, 194, 58, 16, 79, 192, 198, 2, 1, 0, 57, 16, 79, 192, 198, 66, 58, 16, 79, 192, 198, 130, 7, 16, 72, 0, 12, 16, 35, 0, 12, 16, 12, 0, 13, 16, 17, 0, 14, 16, 22, 0, 15, 16, 1, 0, 57, 16, 78, 192, 198, 66, 58, 16, 78, 192, 198, 130, 1, 0, 57, 16, 76, 192, 198, 194, 58, 16, 77, 192, 198, 2, 1, 0, 57, 16, 77, 192, 198, 66, 58, 16, 77, 192, 198, 130, 1, 0, 57, 16, 77, 192, 198, 194, 58, 16, 78, 192, 198, 2, 7, 16, 12, 0, 9, 16, 17, 0, 10, 16, 22, 0, 11, 16, 1, 0, 57, 16, 76, 192, 198, 66, 58, 16, 76, 192, 198, 130, 1, 0, 57, 16, 74, 192, 198, 194, 58, 16, 75, 192, 198, 2, 1, 0, 57, 16, 75, 192, 198, 66, 58, 16, 75, 192, 198, 130, 1, 0, 57, 16, 75, 192, 198, 194, 58, 16, 76, 192, 198, 2, 3, 16, 35, 0, 3, 16, 12, 0, 4, 16, 17, 0, 5, 16, 22, 0, 6, 16, 1, 0, 57, 16, 74, 192, 198, 66, 58, 16, 74, 192, 198, 130, 1, 0, 57, 16, 73, 192, 198, 2, 58, 16, 73, 192, 198, 66, 1, 0, 57, 16, 2, 192, 197, 132, 58, 16, 73, 192, 198, 130, 1, 0, 57, 16, 73, 192, 198, 194, 58, 16, 74, 192, 198, 2, 0, 16, 10, 0, 1, 16, 15, 0, 2, 16, 1, 0, 57, 16, 72, 192, 198, 130, 58, 16, 72, 192, 198, 194, 1, 0, 57, 16, 71, 192, 198, 130, 58, 16, 71, 192, 198, 194, 1, 0, 57, 16, 72, 192, 198, 2, 58, 16, 72, 192, 198, 66, 1, 0, 57, 16, 70, 192, 198, 2, 58, 16, 70, 192, 198, 66, 1, 0, 57, 16, 70, 192, 198, 130, 58, 16, 70, 192, 198, 194, 1, 0, 57, 16, 71, 192, 198, 2, 58, 16, 71, 192, 198, 66, 24, 16, 12, 0, 25, 16, 17, 0, 26, 16, 22, 0, 27, 16, 1, 0, 57, 16, 69, 192, 198, 130, 58, 16, 69, 192, 198, 194, 1, 0, 57, 16, 68, 192, 198, 2, 58, 16, 68, 192, 198, 66, 1, 0, 57, 16, 68, 192, 198, 130, 58, 16, 68, 192, 198, 194, 1, 0, 57, 16, 69, 192, 198, 2, 58, 16, 69, 192, 198, 66, 20, 16, 35, 0, 20, 16, 12, 0, 21, 16, 17, 0, 22, 16, 22, 0, 23, 16, 1, 0, 57, 16, 67, 192, 198, 130, 58, 16, 67, 192, 198, 194, 1, 0, 57, 16, 66, 192, 198, 2, 58, 16, 66, 192, 198, 66, 1, 0, 57, 16, 66, 192, 198, 130, 58, 16, 66, 192, 198, 194, 1, 0, 57, 16, 67, 192, 198, 2, 58, 16, 67, 192, 198, 66, 16, 16, 12, 0, 17, 16, 17, 0, 18, 16, 22, 0, 19, 16, 1, 0, 57, 16, 65, 192, 198, 130, 58, 16, 65, 192, 198, 194, 1, 0, 57, 16, 64, 192, 198, 2, 58, 16, 64, 192, 198, 66, 1, 0, 57, 16, 64, 192, 198, 130, 58, 16, 64, 192, 198, 194, 1, 0, 57, 16, 65, 192, 198, 2, 58, 16, 65, 192, 198, 66, 7, 16, 72, 0, 12, 16, 35, 0, 12, 16, 12, 0, 13, 16, 17, 0, 14, 16, 22, 0, 15, 16, 1, 0, 57, 16, 63, 192, 198, 130, 58, 16, 63, 192, 198, 194, 1, 0, 57, 16, 62, 192, 198, 2, 58, 16, 62, 192, 198, 66, 1, 0, 57, 16, 62, 192, 198, 130, 58, 16, 62, 192, 198, 194, 1, 0, 57, 16, 63, 192, 198, 2, 58, 16, 63, 192, 198, 66, 7, 16, 12, 0, 9, 16, 17, 0, 10, 16, 22, 0, 11, 16, 1, 0, 57, 16, 61, 192, 198, 130, 58, 16, 61, 192, 198, 194, 1, 0, 57, 16, 60, 192, 198, 2, 58, 16, 60, 192, 198, 66, 1, 0, 57, 16, 60, 192, 198, 130, 58, 16, 60, 192, 198, 194, 1, 0, 57, 16, 61, 192, 198, 2, 58, 16, 61, 192, 198, 66, 3, 16, 35, 0, 3, 16, 12, 0, 4, 16, 17, 0, 5, 16, 22, 0, 6, 16, 1, 0, 57, 16, 59, 192, 198, 130, 58, 16, 59, 192, 198, 194, 1, 0, 57, 16, 58, 192, 198, 66, 58, 16, 58, 192, 198, 130, 1, 0, 57, 16, 2, 192, 197, 4, 58, 16, 58, 192, 198, 194, 1, 0, 57, 16, 59, 192, 198, 2, 58, 16, 59, 192, 198, 66, 0, 16, 10, 0, 1, 16, 15, 0, 2, 16, 1, 0, 57, 16, 57, 192, 198, 194, 58, 16, 58, 192, 198, 2, 1, 0, 57, 16, 56, 192, 198, 194, 58, 16, 57, 192, 198, 2, 1, 0, 57, 16, 57, 192, 198, 66, 58, 16, 57, 192, 198, 130, 87, 117, 5, 70, 33, 0, 17, 16, 146, 0, 25, 16, 72, 0, 29, 16, 35, 0, 29, 16, 11, 0, 30, 16, 13, 0, 31, 16, 18, 0, 32, 16, 23, 0, 33, 16, 48, 0, 58, 16, 94, 192, 198, 97, 48, 0, 58, 16, 93, 192, 198, 161, 1, 0, 57, 16, 93, 192, 198, 193, 58, 16, 93, 192, 198, 225, 1, 0, 57, 16, 94, 192, 198, 1, 58, 16, 94, 192, 198, 33, 48, 0, 58, 16, 94, 192, 198, 65, 25, 16, 12, 0, 26, 16, 17, 0, 27, 16, 22, 0, 28, 16, 1, 0, 57, 16, 93, 192, 198, 97, 58, 16, 93, 192, 198, 129, 1, 0, 57, 16, 92, 192, 198, 161, 58, 16, 92, 192, 198, 193, 1, 0, 57, 16, 92, 192, 198, 225, 58, 16, 93, 192, 198, 1, 1, 0, 57, 16, 93, 192, 198, 33, 58, 16, 93, 192, 198, 65, 21, 16, 35, 0, 21, 16, 12, 0, 22, 16, 17, 0, 23, 16, 22, 0, 24, 16, 1, 0, 57, 16, 92, 192, 198, 97, 58, 16, 92, 192, 198, 129, 1, 0, 57, 16, 91, 192, 198, 161, 58, 16, 91, 192, 198, 193, 1, 0, 57, 16, 91, 192, 198, 225, 58, 16, 92, 192, 198, 1, 1, 0, 57, 16, 92, 192, 198, 33, 58, 16, 92, 192, 198, 65, 17, 16, 12, 0, 18, 16, 17, 0, 19, 16, 22, 0, 20, 16, 1, 0, 57, 16, 91, 192, 198, 97, 58, 16, 91, 192, 198, 129, 1, 0, 57, 16, 90, 192, 198, 161, 58, 16, 90, 192, 198, 193, 1, 0, 57, 16, 90, 192, 198, 225, 58, 16, 91, 192, 198, 1, 1, 0, 57, 16, 91, 192, 198, 33, 58, 16, 91, 192, 198, 65, 8, 16, 78, 0, 12, 16, 44, 0, 12, 16, 14, 0, 13, 16, 19, 0, 14, 16, 24, 0, 15, 16, 29, 0, 16, 16, 1, 0, 57, 16, 90, 192, 198, 97, 58, 16, 90, 192, 198, 129, 1, 0, 57, 16, 89, 192, 198, 97, 58, 16, 89, 192, 198, 129, 1, 0, 57, 16, 89, 192, 198, 161, 58, 16, 89, 192, 198, 193, 1, 0, 57, 16, 89, 192, 198, 225, 58, 16, 90, 192, 198, 1, 1, 0, 57, 16, 90, 192, 198, 33, 58, 16, 90, 192, 198, 65, 8, 16, 12, 0, 9, 16, 14, 0, 10, 16, 19, 0, 11, 16, 1, 0, 57, 16, 89, 192, 198, 33, 58, 16, 89, 192, 198, 65, 48, 0, 58, 16, 88, 192, 198, 129, 1, 0, 57, 16, 88, 192, 198, 161, 58, 16, 88, 192, 198, 193, 1, 0, 57, 16, 88, 192, 198, 225, 58, 16, 89, 192, 198, 1, 4, 16, 35, 0, 4, 16, 12, 0, 5, 16, 17, 0, 6, 16, 22, 0, 7, 16, 1, 0, 57, 16, 88, 192, 198, 65, 58, 16, 88, 192, 198, 97, 48, 0, 58, 16, 48, 86, 198, 129, 57, 16, 87, 192, 198, 161, 1, 0, 57, 16, 87, 192, 198, 193, 58, 16, 87, 192, 198, 225, 1, 0, 57, 16, 88, 192, 198, 1, 58, 16, 88, 192, 198, 33, 0, 16, 12, 0, 1, 16, 17, 0, 2, 16, 22, 0, 3, 16, 1, 0, 57, 16, 87, 192, 198, 65, 58, 16, 87, 192, 198, 97, 1, 0, 57, 16, 86, 192, 198, 129, 58, 16, 86, 192, 198, 161, 1, 0, 57, 16, 86, 192, 198, 193, 58, 16, 86, 192, 198, 225, 1, 0, 57, 16, 87, 192, 198, 1, 58, 16, 87, 192, 198, 33, 87, 117, 5, 5, 33, 0, 17, 16, 146, 0, 25, 16, 72, 0, 29, 16, 35, 0, 29, 16, 11, 0, 30, 16, 13, 0, 31, 16, 18, 0, 32, 16, 23, 0, 33, 16, 48, 0, 58, 16, 102, 192, 198, 65, 48, 0, 58, 16, 101, 192, 198, 129, 1, 0, 57, 16, 101, 192, 198, 161, 58, 16, 101, 192, 198, 193, 1, 0, 57, 16, 101, 192, 198, 225, 58, 16, 102, 192, 198, 1, 48, 0, 58, 16, 102, 192, 198, 33, 25, 16, 12, 0, 26, 16, 17, 0, 27, 16, 22, 0, 28, 16, 1, 0, 57, 16, 101, 192, 198, 65, 58, 16, 101, 192, 198, 97, 1, 0, 57, 16, 100, 192, 198, 129, 58, 16, 100, 192, 198, 161, 1, 0, 57, 16, 100, 192, 198, 193, 58, 16, 100, 192, 198, 225, 1, 0, 57, 16, 101, 192, 198, 1, 58, 16, 101, 192, 198, 33, 21, 16, 35, 0, 21, 16, 12, 0, 22, 16, 17, 0, 23, 16, 22, 0, 24, 16, 1, 0, 57, 16, 100, 192, 198, 65, 58, 16, 100, 192, 198, 97, 1, 0, 57, 16, 99, 192, 198, 129, 58, 16, 99, 192, 198, 161, 1, 0, 57, 16, 99, 192, 198, 193, 58, 16, 99, 192, 198, 225, 1, 0, 57, 16, 100, 192, 198, 1, 58, 16, 100, 192, 198, 33, 17, 16, 12, 0, 18, 16, 17, 0, 19, 16, 22, 0, 20, 16, 1, 0, 57, 16, 99, 192, 198, 65, 58, 16, 99, 192, 198, 97, 1, 0, 57, 16, 98, 192, 198, 129, 58, 16, 98, 192, 198, 161, 1, 0, 57, 16, 98, 192, 198, 193, 58, 16, 98, 192, 198, 225, 1, 0, 57, 16, 99, 192, 198, 1, 58, 16, 99, 192, 198, 33, 8, 16, 78, 0, 12, 16, 44, 0, 12, 16, 14, 0, 13, 16, 19, 0, 14, 16, 24, 0, 15, 16, 29, 0, 16, 16, 1, 0, 57, 16, 98, 192, 198, 65, 58, 16, 98, 192, 198, 97, 1, 0, 57, 16, 97, 192, 198, 65, 58, 16, 97, 192, 198, 97, 1, 0, 57, 16, 97, 192, 198, 129, 58, 16, 97, 192, 198, 161, 1, 0, 57, 16, 97, 192, 198, 193, 58, 16, 97, 192, 198, 225, 1, 0, 57, 16, 98, 192, 198, 1, 58, 16, 98, 192, 198, 33, 8, 16, 12, 0, 9, 16, 14, 0, 10, 16, 19, 0, 11, 16, 1, 0, 57, 16, 97, 192, 198, 1, 58, 16, 97, 192, 198, 33, 48, 0, 58, 16, 96, 192, 198, 97, 1, 0, 57, 16, 96, 192, 198, 129, 58, 16, 96, 192, 198, 161, 1, 0, 57, 16, 96, 192, 198, 193, 58, 16, 96, 192, 198, 225, 4, 16, 35, 0, 4, 16, 12, 0, 5, 16, 17, 0, 6, 16, 22, 0, 7, 16, 1, 0, 57, 16, 96, 192, 198, 33, 58, 16, 96, 192, 198, 65, 48, 0, 58, 16, 48, 88, 198, 97, 57, 16, 95, 192, 198, 129, 1, 0, 57, 16, 95, 192, 198, 161, 58, 16, 95, 192, 198, 193, 1, 0, 57, 16, 95, 192, 198, 225, 58, 16, 96, 192, 198, 1, 0, 16, 12, 0, 1, 16, 18, 0, 2, 16, 23, 0, 3, 16, 1, 0, 57, 16, 95, 192, 198, 33, 58, 16, 95, 192, 198, 65, 1, 0, 57, 16, 255, 255, 89, 117, 193, 153, 58, 16, 94, 192, 198, 129, 1, 0, 57, 16, 94, 192, 198, 161, 58, 16, 94, 192, 198, 193, 1, 0, 57, 16, 94, 192, 198, 225, 58, 16, 95, 192, 198, 1, 89, 117, 193, 23, 36, 0, 18, 16, 209, 1, 27, 16, 117, 1, 32, 16, 74, 1, 32, 16, 9, 0, 33, 16, 11, 0, 47, 16, 13, 0, 54, 16, 58, 1, 63, 16, 118, 192, 198, 226, 48, 0, 58, 16, 110, 192, 198, 65, 48, 0, 58, 16, 110, 192, 198, 97, 225, 127, 89, 117, 193, 151, 17, 16, 146, 0, 25, 16, 72, 0, 29, 16, 35, 0, 29, 16, 11, 0, 30, 16, 13, 0, 31, 16, 18, 0, 32, 16, 23, 0, 33, 16, 48, 0, 58, 16, 118, 192, 198, 97, 48, 0, 58, 16, 117, 192, 198, 161, 1, 0, 57, 16, 117, 192, 198, 193, 58, 16, 117, 192, 198, 225, 1, 0, 57, 16, 118, 192, 198, 1, 58, 16, 118, 192, 198, 33, 48, 0, 58, 16, 118, 192, 198, 65, 25, 16, 12, 0, 26, 16, 17, 0, 27, 16, 22, 0, 28, 16, 1, 0, 57, 16, 117, 192, 198, 97, 58, 16, 117, 192, 198, 129, 1, 0, 57, 16, 116, 192, 198, 161, 58, 16, 116, 192, 198, 193, 1, 0, 57, 16, 116, 192, 198, 225, 58, 16, 117, 192, 198, 1, 1, 0, 57, 16, 117, 192, 198, 33, 58, 16, 117, 192, 198, 65, 21, 16, 35, 0, 21, 16, 12, 0, 22, 16, 17, 0, 23, 16, 22, 0, 24, 16, 1, 0, 57, 16, 116, 192, 198, 97, 58, 16, 116, 192, 198, 129, 1, 0, 57, 16, 115, 192, 198, 161, 58, 16, 115, 192, 198, 193, 1, 0, 57, 16, 115, 192, 198, 225, 58, 16, 116, 192, 198, 1, 1, 0, 57, 16, 116, 192, 198, 33, 58, 16, 116, 192, 198, 65, 17, 16, 12, 0, 18, 16, 17, 0, 19, 16, 22, 0, 20, 16, 1, 0, 57, 16, 115, 192, 198, 97, 58, 16, 115, 192, 198, 129, 1, 0, 57, 16, 114, 192, 198, 161, 58, 16, 114, 192, 198, 193, 1, 0, 57, 16, 114, 192, 198, 225, 58, 16, 115, 192, 198, 1, 1, 0, 57, 16, 115, 192, 198, 33, 58, 16, 115, 192, 198, 65, 8, 16, 78, 0, 12, 16, 44, 0, 12, 16, 14, 0, 13, 16, 19, 0, 14, 16, 24, 0, 15, 16, 29, 0, 16, 16, 1, 0, 57, 16, 114, 192, 198, 97, 58, 16, 114, 192, 198, 129, 1, 0, 57, 16, 113, 192, 198, 97, 58, 16, 113, 192, 198, 129, 1, 0, 57, 16, 113, 192, 198, 161, 58, 16, 113, 192, 198, 193, 1, 0, 57, 16, 113, 192, 198, 225, 58, 16, 114, 192, 198, 1, 1, 0, 57, 16, 114, 192, 198, 33, 58, 16, 114, 192, 198, 65, 8, 16, 12, 0, 9, 16, 14, 0, 10, 16, 19, 0, 11, 16, 1, 0, 57, 16, 113, 192, 198, 33, 58, 16, 113, 192, 198, 65, 48, 0, 58, 16, 112, 192, 198, 129, 1, 0, 57, 16, 112, 192, 198, 161, 58, 16, 112, 192, 198, 193, 1, 0, 57, 16, 112, 192, 198, 225, 58, 16, 113, 192, 198, 1, 4, 16, 35, 0, 4, 16, 12, 0, 5, 16, 17, 0, 6, 16, 22, 0, 7, 16, 1, 0, 57, 16, 112, 192, 198, 65, 58, 16, 112, 192, 198, 97, 48, 0, 58, 16, 48, 92, 198, 129, 57, 16, 111, 192, 198, 161, 1, 0, 57, 16, 111, 192, 198, 193, 58, 16, 111, 192, 198, 225, 1, 0, 57, 16, 112, 192, 198, 1, 58, 16, 112, 192, 198, 33, 0, 16, 12, 0, 1, 16, 17, 0, 2, 16, 22, 0, 3, 16, 1, 0, 57, 16, 111, 192, 198, 65, 58, 16, 111, 192, 198, 97, 1, 0, 57, 16, 110, 192, 198, 129, 58, 16, 110, 192, 198, 161, 1, 0, 57, 16, 110, 192, 198, 193, 58, 16, 110, 192, 198, 225, 1, 0, 57, 16, 111, 192, 198, 1, 58, 16, 111, 192, 198, 33, 193, 93, 198, 129, 55, 16, 118, 192, 198, 161, 56, 16, 118, 192, 198, 193, 27, 16, 14, 0, 28, 16, 19, 0, 29, 16, 24, 0, 30, 16, 26, 0, 31, 16, 1, 0, 57, 16, 110, 192, 198, 1, 58, 16, 110, 192, 198, 33, 1, 0, 57, 16, 109, 192, 198, 65, 58, 16, 109, 192, 198, 97, 1, 0, 57, 16, 109, 192, 198, 129, 58, 16, 109, 192, 198, 161, 48, 0, 58, 16, 109, 192, 198, 193, 1, 0, 57, 16, 24, 192, 198, 225, 58, 16, 109, 192, 198, 225, 22, 16, 53, 0, 22, 16, 14, 0, 23, 16, 19, 0, 24, 16, 24, 0, 25, 16, 29, 0, 26, 16, 1, 0, 57, 16, 109, 192, 198, 1, 58, 16, 109, 192, 198, 33, 1, 0, 57, 16, 107, 192, 198, 193, 58, 16, 107, 192, 198, 225, 1, 0, 57, 16, 108, 192, 198, 1, 58, 16, 108, 192, 198, 33, 1, 0, 57, 16, 108, 192, 198, 65, 58, 16, 108, 192, 198, 97, 2, 0, 55, 16, 9, 0, 57, 16, 108, 192, 198, 161, 58, 16, 112, 91, 198, 193, 56, 16, 108, 192, 198, 225, 48, 0, 58, 16, 108, 192, 198, 129, 18, 16, 12, 0, 19, 16, 17, 0, 20, 16, 22, 0, 21, 16, 1, 0, 57, 16, 107, 192, 198, 129, 58, 16, 107, 192, 198, 161, 1, 0, 57, 16, 106, 192, 198, 193, 58, 16, 106, 192, 198, 225, 1, 0, 57, 16, 107, 192, 198, 1, 58, 16, 107, 192, 198, 33, 1, 0, 57, 16, 107, 192, 198, 65, 58, 16, 107, 192, 198, 97, 9, 16, 81, 0, 13, 16, 44, 0, 13, 16, 14, 0, 14, 16, 19, 0, 15, 16, 24, 0, 16, 16, 29, 0, 17, 16, 1, 0, 57, 16, 106, 192, 198, 129, 58, 16, 106, 192, 198, 161, 1, 0, 57, 16, 105, 192, 198, 129, 58, 16, 105, 192, 198, 161, 1, 0, 57, 16, 105, 192, 198, 193, 58, 16, 105, 192, 198, 225, 1, 0, 57, 16, 106, 192, 198, 1, 58, 16, 106, 192, 198, 33, 1, 0, 57, 16, 106, 192, 198, 65, 58, 16, 106, 192, 198, 97, 9, 16, 12, 0, 10, 16, 17, 0, 11, 16, 22, 0, 12, 16, 1, 0, 57, 16, 105, 192, 198, 65, 58, 16, 105, 192, 198, 97, 1, 0, 57, 16, 104, 192, 198, 129, 58, 16, 104, 192, 198, 161, 1, 0, 57, 16, 104, 192, 198, 193, 58, 16, 104, 192, 198, 225, 1, 0, 57, 16, 105, 192, 198, 1, 58, 16, 105, 192, 198, 33, 4, 16, 41, 0, 4, 16, 11, 0, 5, 16, 16, 0, 6, 16, 21, 0, 7, 16, 26, 0, 8, 16, 48, 0, 58, 16, 104, 192, 198, 97, 48, 0, 58, 16, 48, 90, 198, 97, 57, 16, 103, 192, 198, 129, 1, 0, 57, 16, 103, 192, 198, 161, 58, 16, 103, 192, 198, 193, 1, 0, 57, 16, 103, 192, 198, 225, 58, 16, 104, 192, 198, 1, 1, 0, 57, 16, 104, 192, 198, 33, 58, 16, 104, 192, 198, 65, 0, 16, 12, 0, 1, 16, 17, 0, 2, 16, 22, 0, 3, 16, 1, 0, 57, 16, 103, 192, 198, 33, 58, 16, 103, 192, 198, 65, 1, 0, 57, 16, 102, 192, 198, 97, 58, 16, 102, 192, 198, 129, 1, 0, 57, 16, 102, 192, 198, 161, 58, 16, 102, 192, 198, 193, 1, 0, 57, 16, 102, 192, 198, 225, 58, 16, 103, 192, 198, 1, 89, 117, 193, 51, 35, 0, 18, 16, 160, 0, 27, 16, 68, 0, 31, 16, 34, 0, 31, 16, 9, 0, 32, 16, 14, 0, 33, 16, 16, 0, 54, 16, 18, 0, 63, 16, 127, 192, 198, 130, 1, 0, 57, 16, 126, 192, 198, 161, 58, 16, 126, 192, 198, 193, 48, 0, 58, 16, 126, 192, 198, 225, 48, 0, 58, 16, 127, 192, 198, 1, 1, 96, 198, 33, 55, 16, 127, 192, 198, 65, 56, 16, 127, 192, 198, 97, 27, 16, 12, 0, 28, 16, 17, 0, 29, 16, 22, 0, 30, 16, 1, 0, 57, 16, 41, 192, 198, 1, 58, 16, 126, 192, 198, 129, 1, 0, 57, 16, 125, 192, 198, 225, 58, 16, 126, 192, 198, 1, 1, 0, 57, 16, 126, 192, 198, 33, 58, 16, 126, 192, 198, 65, 48, 0, 58, 16, 126, 192, 198, 97, 22, 16, 53, 0, 22, 16, 14, 0, 23, 16, 19, 0, 24, 16, 24, 0, 25, 16, 29, 0, 26, 16, 1, 0, 57, 16, 125, 192, 198, 161, 58, 16, 125, 192, 198, 193, 1, 0, 57, 16, 124, 192, 198, 97, 58, 16, 124, 192, 198, 129, 1, 0, 57, 16, 124, 192, 198, 161, 58, 16, 124, 192, 198, 193, 1, 0, 57, 16, 124, 192, 198, 225, 58, 16, 125, 192, 198, 1, 2, 0, 55, 16, 9, 0, 57, 16, 125, 192, 198, 65, 58, 16, 176, 95, 198, 97, 56, 16, 125, 192, 198, 129, 48, 0, 58, 16, 125, 192, 198, 33, 18, 16, 12, 0, 19, 16, 17, 0, 20, 16, 22, 0, 21, 16, 1, 0, 57, 16, 124, 192, 198, 65, 58, 16, 5, 192, 198, 193, 1, 0, 57, 16, 123, 192, 198, 129, 58, 16, 123, 192, 198, 161, 1, 0, 57, 16, 123, 192, 198, 193, 58, 16, 123, 192, 198, 225, 1, 0, 57, 16, 124, 192, 198, 1, 58, 16, 124, 192, 198, 33, 9, 16, 81, 0, 13, 16, 44, 0, 13, 16, 14, 0, 14, 16, 19, 0, 15, 16, 24, 0, 16, 16, 29, 0, 17, 16, 1, 0, 57, 16, 123, 192, 198, 65, 58, 16, 123, 192, 198, 97, 1, 0, 57, 16, 122, 192, 198, 65, 58, 16, 122, 192, 198, 97, 1, 0, 57, 16, 122, 192, 198, 129, 58, 16, 122, 192, 198, 161, 1, 0, 57, 16, 122, 192, 198, 193, 58, 16, 122, 192, 198, 225, 1, 0, 57, 16, 123, 192, 198, 1, 58, 16, 123, 192, 198, 33, 9, 16, 12, 0, 10, 16, 17, 0, 11, 16, 22, 0, 12, 16, 1, 0, 57, 16, 122, 192, 198, 1, 58, 16, 122, 192, 198, 33, 1, 0, 57, 16, 121, 192, 198, 65, 58, 16, 121, 192, 198, 97, 1, 0, 57, 16, 121, 192, 198, 129, 58, 16, 121, 192, 198, 161, 1, 0, 57, 16, 121, 192, 198, 193, 58, 16, 121, 192, 198, 225, 4, 16, 41, 0, 4, 16, 11, 0, 5, 16, 16, 0, 6, 16, 21, 0, 7, 16, 26, 0, 8, 16, 48, 0, 58, 16, 121, 192, 198, 33, 48, 0, 58, 16, 112, 94, 198, 33, 57, 16, 120, 192, 198, 65, 1, 0, 57, 16, 120, 192, 198, 97, 58, 16, 120, 192, 198, 129, 1, 0, 57, 16, 120, 192, 198, 161, 58, 16, 120, 192, 198, 193, 1, 0, 57, 16, 120, 192, 198, 225, 58, 16, 121, 192, 198, 1, 0, 16, 12, 0, 1, 16, 17, 0, 2, 16, 22, 0, 3, 16, 1, 0, 57, 16, 119, 192, 198, 225, 58, 16, 120, 192, 198, 1, 1, 0, 57, 16, 119, 192, 198, 33, 58, 16, 119, 192, 198, 65, 1, 0, 57, 16, 119, 192, 198, 97, 58, 16, 119, 192, 198, 129, 1, 0, 57, 16, 119, 192, 198, 161, 58, 16, 119, 192, 198, 193, 89, 117, 193, 107, 36, 0, 18, 16, 250, 2, 27, 16, 167, 2, 32, 16, 124, 2, 32, 16, 9, 0, 33, 16, 11, 0, 43, 16, 13, 0, 44, 16, 61, 1, 63, 16, 152, 192, 198, 98, 48, 0, 58, 16, 135, 192, 198, 129, 48, 0, 58, 16, 135, 192, 198, 161, 34, 98, 198, 193, 17, 16, 150, 0, 26, 16, 67, 0, 30, 16, 33, 0, 30, 16, 9, 0, 31, 16, 14, 0, 32, 16, 19, 0, 33, 16, 21, 0, 58, 16, 143, 192, 198, 225, 1, 0, 57, 16, 143, 192, 198, 33, 58, 16, 143, 192, 198, 65, 1, 0, 57, 16, 143, 192, 198, 97, 58, 16, 143, 192, 198, 129, 48, 0, 58, 16, 143, 192, 198, 161, 48, 0, 58, 16, 143, 192, 198, 193, 26, 16, 9, 0, 27, 16, 14, 0, 28, 16, 19, 0, 29, 16, 48, 0, 58, 16, 143, 192, 198, 1, 1, 0, 57, 16, 142, 192, 198, 65, 58, 16, 142, 192, 198, 97, 1, 0, 57, 16, 142, 192, 198, 129, 58, 16, 142, 192, 198, 161, 1, 0, 57, 16, 142, 192, 198, 193, 58, 16, 142, 192, 198, 225, 21, 16, 44, 0, 21, 16, 14, 0, 22, 16, 19, 0, 23, 16, 24, 0, 24, 16, 29, 0, 25, 16, 1, 0, 57, 16, 142, 192, 198, 1, 58, 16, 142, 192, 198, 33, 1, 0, 57, 16, 141, 192, 198, 1, 58, 16, 141, 192, 198, 33, 1, 0, 57, 16, 141, 192, 198, 65, 58, 16, 141, 192, 198, 97, 1, 0, 57, 16, 141, 192, 198, 129, 58, 16, 141, 192, 198, 161, 1, 0, 57, 16, 141, 192, 198, 193, 58, 16, 141, 192, 198, 225, 17, 16, 12, 0, 18, 16, 17, 0, 19, 16, 22, 0, 20, 16, 1, 0, 57, 16, 140, 192, 198, 193, 58, 16, 140, 192, 198, 225, 1, 0, 57, 16, 140, 192, 198, 1, 58, 16, 140, 192, 198, 33, 1, 0, 57, 16, 140, 192, 198, 65, 58, 16, 140, 192, 198, 97, 1, 0, 57, 16, 140, 192, 198, 129, 58, 16, 140, 192, 198, 161, 8, 16, 78, 0, 12, 16, 44, 0, 12, 16, 14, 0, 13, 16, 19, 0, 14, 16, 24, 0, 15, 16, 29, 0, 16, 16, 1, 0, 57, 16, 139, 192, 198, 193, 58, 16, 139, 192, 198, 225, 1, 0, 57, 16, 138, 192, 198, 193, 58, 16, 138, 192, 198, 225, 1, 0, 57, 16, 139, 192, 198, 1, 58, 16, 139, 192, 198, 33, 1, 0, 57, 16, 139, 192, 198, 65, 58, 16, 139, 192, 198, 97, 1, 0, 57, 16, 139, 192, 198, 129, 58, 16, 139, 192, 198, 161, 8, 16, 12, 0, 9, 16, 14, 0, 10, 16, 19, 0, 11, 16, 1, 0, 57, 16, 138, 192, 198, 129, 58, 16, 138, 192, 198, 161, 48, 0, 58, 16, 137, 192, 198, 225, 1, 0, 57, 16, 138, 192, 198, 1, 58, 16, 138, 192, 198, 33, 1, 0, 57, 16, 138, 192, 198, 65, 58, 16, 138, 192, 198, 97, 4, 16, 35, 0, 4, 16, 12, 0, 5, 16, 17, 0, 6, 16, 22, 0, 7, 16, 1, 0, 57, 16, 137, 192, 198, 161, 58, 16, 137, 192, 198, 193, 48, 0, 58, 16, 112, 98, 198, 225, 57, 16, 137, 192, 198, 1, 1, 0, 57, 16, 137, 192, 198, 33, 58, 16, 137, 192, 198, 65, 1, 0, 57, 16, 137, 192, 198, 97, 58, 16, 137, 192, 198, 129, 0, 16, 12, 0, 1, 16, 17, 0, 2, 16, 22, 0, 3, 16, 1, 0, 57, 16, 136, 192, 198, 161, 58, 16, 136, 192, 198, 193, 1, 0, 57, 16, 135, 192, 198, 225, 58, 16, 136, 192, 198, 1, 1, 0, 57, 16, 136, 192, 198, 33, 58, 16, 136, 192, 198, 65, 1, 0, 57, 16, 136, 192, 198, 97, 58, 16, 136, 192, 198, 129, 226, 127, 89, 117, 193, 136, 17, 16, 151, 0, 26, 16, 68, 0, 30, 16, 34, 0, 30, 16, 10, 0, 31, 16, 15, 0, 32, 16, 20, 0, 33, 16, 22, 0, 58, 16, 255, 255, 89, 117, 193, 137, 1, 0, 57, 16, 151, 192, 198, 161, 58, 16, 151, 192, 198, 193, 1, 0, 57, 16, 151, 192, 198, 225, 58, 16, 152, 192, 198, 1, 48, 0, 58, 16, 152, 192, 198, 33, 48, 0, 58, 16, 152, 192, 198, 65, 26, 16, 9, 0, 27, 16, 14, 0, 28, 16, 19, 0, 29, 16, 48, 0, 58, 16, 151, 192, 198, 129, 1, 0, 57, 16, 150, 192, 198, 193, 58, 16, 150, 192, 198, 225, 1, 0, 57, 16, 151, 192, 198, 1, 58, 16, 151, 192, 198, 33, 1, 0, 57, 16, 151, 192, 198, 65, 58, 16, 151, 192, 198, 97, 21, 16, 44, 0, 21, 16, 14, 0, 22, 16, 19, 0, 23, 16, 24, 0, 24, 16, 29, 0, 25, 16, 1, 0, 57, 16, 150, 192, 198, 129, 58, 16, 150, 192, 198, 161, 1, 0, 57, 16, 149, 192, 198, 129, 58, 16, 149, 192, 198, 161, 1, 0, 57, 16, 149, 192, 198, 193, 58, 16, 149, 192, 198, 225, 1, 0, 57, 16, 150, 192, 198, 1, 58, 16, 150, 192, 198, 33, 1, 0, 57, 16, 150, 192, 198, 65, 58, 16, 150, 192, 198, 97, 17, 16, 12, 0, 18, 16, 17, 0, 19, 16, 22, 0, 20, 16, 1, 0, 57, 16, 149, 192, 198, 65, 58, 16, 149, 192, 198, 97, 1, 0, 57, 16, 148, 192, 198, 129, 58, 16, 148, 192, 198, 161, 1, 0, 57, 16, 148, 192, 198, 193, 58, 16, 148, 192, 198, 225, 1, 0, 57, 16, 149, 192, 198, 1, 58, 16, 149, 192, 198, 33, 8, 16, 78, 0, 12, 16, 44, 0, 12, 16, 14, 0, 13, 16, 19, 0, 14, 16, 24, 0, 15, 16, 29, 0, 16, 16, 1, 0, 57, 16, 148, 192, 198, 65, 58, 16, 148, 192, 198, 97, 1, 0, 57, 16, 147, 192, 198, 65, 58, 16, 147, 192, 198, 97, 1, 0, 57, 16, 147, 192, 198, 129, 58, 16, 147, 192, 198, 161, 1, 0, 57, 16, 147, 192, 198, 193, 58, 16, 147, 192, 198, 225, 1, 0, 57, 16, 148, 192, 198, 1, 58, 16, 148, 192, 198, 33, 8, 16, 12, 0, 9, 16, 14, 0, 10, 16, 19, 0, 11, 16, 1, 0, 57, 16, 147, 192, 198, 1, 58, 16, 147, 192, 198, 33, 48, 0, 58, 16, 146, 192, 198, 97, 1, 0, 57, 16, 146, 192, 198, 129, 58, 16, 146, 192, 198, 161, 1, 0, 57, 16, 146, 192, 198, 193, 58, 16, 146, 192, 198, 225, 4, 16, 35, 0, 4, 16, 12, 0, 5, 16, 17, 0, 6, 16, 22, 0, 7, 16, 1, 0, 57, 16, 146, 192, 198, 33, 58, 16, 146, 192, 198, 65, 48, 0, 58, 16, 176, 100, 198, 97, 57, 16, 145, 192, 198, 129, 1, 0, 57, 16, 145, 192, 198, 161, 58, 16, 145, 192, 198, 193, 1, 0, 57, 16, 145, 192, 198, 225, 58, 16, 146, 192, 198, 1, 0, 16, 12, 0, 1, 16, 20, 0, 2, 16, 25, 0, 3, 16, 1, 0, 57, 16, 145, 192, 198, 33, 58, 16, 145, 192, 198, 65, 1, 0, 57, 16, 144, 192, 198, 1, 58, 16, 112, 100, 198, 33, 59, 16, 144, 192, 198, 67, 1, 0, 57, 16, 144, 192, 198, 161, 58, 16, 144, 192, 198, 193, 1, 0, 57, 16, 144, 192, 198, 225, 58, 16, 145, 192, 198, 1, 27, 16, 14, 0, 28, 16, 19, 0, 29, 16, 24, 0, 30, 16, 26, 0, 31, 16, 1, 0, 57, 16, 135, 192, 198, 65, 58, 16, 135, 192, 198, 97, 1, 0, 57, 16, 134, 192, 198, 97, 58, 16, 134, 192, 198, 129, 1, 0, 57, 16, 134, 192, 198, 161, 58, 16, 134, 192, 198, 193, 48, 0, 58, 16, 134, 192, 198, 225, 1, 0, 57, 16, 135, 192, 198, 1, 58, 16, 135, 192, 198, 33, 22, 16, 44, 0, 22, 16, 14, 0, 23, 16, 19, 0, 24, 16, 24, 0, 25, 16, 29, 0, 26, 16, 1, 0, 57, 16, 134, 192, 198, 33, 58, 16, 134, 192, 198, 65, 1, 0, 57, 16, 133, 192, 198, 33, 58, 16, 133, 192, 198, 65, 1, 0, 57, 16, 133, 192, 198, 97, 58, 16, 133, 192, 198, 129, 1, 0, 57, 16, 133, 192, 198, 161, 58, 16, 133, 192, 198, 193, 1, 0, 57, 16, 133, 192, 198, 225, 58, 16, 134, 192, 198, 1, 18, 16, 12, 0, 19, 16, 17, 0, 20, 16, 22, 0, 21, 16, 1, 0, 57, 16, 132, 192, 198, 225, 58, 16, 133, 192, 198, 1, 1, 0, 57, 16, 132, 192, 198, 33, 58, 16, 132, 192, 198, 65, 1, 0, 57, 16, 132, 192, 198, 97, 58, 16, 132, 192, 198, 129, 1, 0, 57, 16, 132, 192, 198, 161, 58, 16, 132, 192, 198, 193, 9, 16, 81, 0, 13, 16, 44, 0, 13, 16, 14, 0, 14, 16, 19, 0, 15, 16, 24, 0, 16, 16, 29, 0, 17, 16, 1, 0, 57, 16, 131, 192, 198, 225, 58, 16, 132, 192, 198, 1, 1, 0, 57, 16, 130, 192, 198, 225, 58, 16, 131, 192, 198, 1, 1, 0, 57, 16, 131, 192, 198, 33, 58, 16, 131, 192, 198, 65, 1, 0, 57, 16, 131, 192, 198, 97, 58, 16, 131, 192, 198, 129, 1, 0, 57, 16, 131, 192, 198, 161, 58, 16, 131, 192, 198, 193, 9, 16, 12, 0, 10, 16, 17, 0, 11, 16, 22, 0, 12, 16, 1, 0, 57, 16, 130, 192, 198, 161, 58, 16, 130, 192, 198, 193, 1, 0, 57, 16, 129, 192, 198, 225, 58, 16, 130, 192, 198, 1, 1, 0, 57, 16, 130, 192, 198, 33, 58, 16, 130, 192, 198, 65, 1, 0, 57, 16, 130, 192, 198, 97, 58, 16, 130, 192, 198, 129, 4, 16, 41, 0, 4, 16, 11, 0, 5, 16, 16, 0, 6, 16, 21, 0, 7, 16, 26, 0, 8, 16, 48, 0, 58, 16, 129, 192, 198, 193, 48, 0, 58, 16, 112, 96, 198, 193, 57, 16, 128, 192, 198, 225, 1, 0, 57, 16, 129, 192, 198, 1, 58, 16, 129, 192, 198, 33, 1, 0, 57, 16, 129, 192, 198, 65, 58, 16, 129, 192, 198, 97, 1, 0, 57, 16, 129, 192, 198, 129, 58, 16, 129, 192, 198, 161, 0, 16, 12, 0, 1, 16, 17, 0, 2, 16, 22, 0, 3, 16, 1, 0, 57, 16, 128, 192, 198, 129, 58, 16, 128, 192, 198, 161, 1, 0, 57, 16, 127, 192, 198, 193, 58, 16, 127, 192, 198, 225, 1, 0, 57, 16, 128, 192, 198, 1, 58, 16, 128, 192, 198, 33, 1, 0, 57, 16, 128, 192, 198, 65, 58, 16, 128, 192, 198, 97, 93, 117, 193, 80, 48, 0, 56, 16, 255, 255, 93, 117, 193, 82, 144, 0, 198, 129, 1, 0, 61, 16, 3, 0, 62, 16, 152, 192, 198, 225, 112, 102, 198, 161, 62, 16, 152, 192, 198, 193, 153, 0, 198, 1, 1, 0, 61, 16, 3, 0, 62, 16, 153, 192, 198, 97, 176, 102, 198, 33, 62, 16, 153, 192, 198, 65, 153, 0, 198, 129, 48, 0, 62, 16, 153, 192, 198, 161, 108, 11, 5, 5, 50, 0, 4, 16, 58, 16, 56, 16, 155, 192, 198, 6, 16, 0, 1, 16, 34, 16, 43, 16, 45, 16, 47, 16, 48, 16, 49, 16, 50, 16, 54, 16, 55, 16, 56, 16, 57, 16, 59, 16, 60, 16, 61, 16, 64, 16, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 55, 13, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 77, 38, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 82, 38, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 153, 27, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 157, 27, 0, 0, 157, 27, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 119, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 49, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 110, 0, 98, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 0, 0, 133, 220, 38, 0, 68, 0, 60, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 60, 0, 60, 0, 240, 0, 60, 0, 60, 0, 60, 0, 208, 0, 38, 0, 116, 0, 60, 0, 60, 0, 60, 0, 254, 0, 47, 0, 104, 0, 38, 0, 84, 0, 60, 0, 60, 0, 60, 0, 222, 0, 47, 0, 72, 0, 38, 0, 89, 0, 60, 0, 60, 0, 117, 0, 8, 3, 60, 0, 60, 0, 60, 0, 85, 0, 8, 3, 60, 0, 60, 0, 117, 0, 11, 3, 60, 0, 60, 0, 60, 0, 85, 0, 11, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 192, 1, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 60, 0, 101, 0, 40, 3, 60, 0, 60, 0, 60, 0, 69, 0, 40, 3, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 60, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 81, 1, 60, 0, 60, 0, 60, 0, 80, 1, 60, 0, 60, 0, 83, 1, 60, 0, 60, 0, 60, 0, 82, 1, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 60, 0, 60, 0, 97, 0, 97, 0, 60, 0, 60, 0, 60, 0, 65, 0, 97, 0, 60, 0, 60, 0, 60, 0, 65, 0, 65, 0, 0, 0, 4, 57, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 163, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 100, 38, 0, 0, 100, 38, 0, 0, 140, 47, 0, 0, 140, 47, 0, 0, 76, 51, 0, 0, 76, 51, 0, 0, 178, 56, 0, 0, 232, 56, 0, 0, 232, 56, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 89, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 145, 0, 153, 0, 159, 0, 167, 0, 175, 0, 183, 0, 48, 0, 56, 0, 182, 0, 190, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 207, 0, 215, 0, 48, 0, 48, 0, 48, 0, 223, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 231, 0, 48, 0, 48, 0, 239, 0, 48, 0, 245, 0, 253, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 8, 1, 48, 0, 48, 0, 12, 1, 48, 0, 20, 1, 25, 1, 33, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 39, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 47, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 50, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 205, 0, 48, 0, 48, 0, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 58, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 66, 1, 48, 0, 48, 0, 72, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 68, 2, 124, 2, 188, 2, 192, 0, 216, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 192, 28, 0, 202, 225, 28, 0, 202, 2, 29, 0, 202, 35, 29, 0, 202, 68, 29, 0, 202, 101, 29, 0, 202, 134, 29, 0, 202, 167, 29, 0, 202, 200, 29, 0, 202, 233, 29, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 196, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 102, 29, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 38, 30, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 70, 31, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 228, 32, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 166, 47, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 102, 48, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 134, 49, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 129, 8, 0, 198, 161, 8, 0, 198, 97, 27, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 225, 17, 0, 5, 5, 127, 12, 198, 129, 27, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 34, 20, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 198, 97, 22, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 198, 33, 27, 0, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 51, 0, 201, 102, 52, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 193, 17, 0, 198, 1, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 18, 0, 198, 65, 27, 0, 198, 97, 28, 0, 198, 129, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 134, 53, 0, 201, 166, 54, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 55, 0, 201, 230, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 58, 0, 201, 38, 59, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 192, 0, 0, 0, 198, 162, 29, 0, 198, 226, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 30, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 60, 0, 201, 102, 61, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 31, 0, 197, 34, 5, 0, 198, 162, 31, 0, 197, 98, 5, 0, 198, 226, 31, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 32, 0, 198, 98, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 134, 62, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 64, 0, 5, 71, 238, 102, 201, 70, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 6, 0, 192, 0, 0, 0, 197, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 7, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 197, 34, 8, 0, 197, 99, 8, 0, 197, 194, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 9, 0, 197, 67, 9, 0, 197, 162, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 226, 5, 0, 197, 34, 10, 0, 5, 5, 10, 121, 197, 98, 10, 0, 197, 162, 10, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 66, 0, 201, 70, 67, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 68, 0, 201, 134, 69, 0, 201, 166, 70, 0, 201, 198, 71, 0, 201, 230, 72, 0, 201, 6, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 75, 0, 201, 70, 76, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 77, 0, 201, 134, 78, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 79, 0, 201, 198, 80, 0, 201, 230, 81, 0, 201, 6, 83, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 84, 0, 201, 70, 85, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 70, 0, 70, 0, 4, 111, 93, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 134, 0, 71, 0, 2, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 71, 0, 2, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 3, 111, 93, 0, 5, 0, 72, 0, 3, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 4, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 193, 4, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 6, 121, 197, 226, 5, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 34, 6, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 98, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 226, 6, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 34, 7, 0, 5, 5, 107, 121, 197, 98, 7, 0, 5, 5, 111, 121, 197, 162, 7, 0, 5, 5, 115, 121, 197, 226, 7, 0, 5, 5, 119, 121, 197, 34, 8, 0, 197, 99, 8, 0, 197, 194, 8, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 2, 9, 0, 197, 67, 9, 0, 197, 162, 9, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 226, 9, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 226, 5, 0, 197, 34, 10, 0, 5, 5, 10, 121, 197, 98, 10, 0, 197, 162, 10, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 0, 0, 62, 0, 3, 2, 118, 0, 169, 30, 54, 0, 177, 30, 26, 0, 181, 30, 12, 0, 181, 30, 16, 192, 198, 163, 182, 30, 17, 192, 198, 3, 183, 30, 17, 192, 198, 99, 43, 33, 1, 192, 198, 130, 177, 30, 15, 192, 198, 35, 178, 30, 15, 192, 198, 131, 179, 30, 15, 192, 198, 227, 180, 30, 16, 192, 198, 67, 173, 30, 12, 0, 173, 30, 13, 192, 198, 163, 174, 30, 14, 192, 198, 3, 175, 30, 14, 192, 198, 99, 176, 30, 14, 192, 198, 195, 169, 30, 12, 192, 198, 35, 170, 30, 12, 192, 198, 131, 171, 30, 12, 192, 198, 227, 172, 30, 13, 192, 198, 67, 161, 30, 26, 0, 165, 30, 12, 0, 165, 30, 10, 192, 198, 163, 166, 30, 11, 192, 198, 3, 167, 30, 11, 192, 198, 99, 168, 30, 11, 192, 198, 195, 161, 30, 9, 192, 198, 130, 162, 30, 9, 192, 198, 194, 163, 30, 10, 192, 198, 2, 164, 30, 10, 192, 198, 67, 10, 3, 12, 0, 10, 3, 8, 192, 198, 161, 0, 30, 8, 192, 198, 194, 1, 30, 9, 192, 198, 2, 160, 30, 9, 192, 198, 66, 3, 2, 8, 192, 198, 66, 38, 2, 5, 0, 39, 2, 8, 0, 8, 3, 8, 192, 198, 129, 240, 65, 198, 2, 4, 3, 6, 192, 198, 3, 240, 65, 198, 98, 4, 3, 6, 192, 198, 99, 1, 1, 54, 0, 223, 1, 26, 0, 251, 1, 12, 0, 251, 1, 7, 192, 198, 35, 0, 2, 7, 192, 198, 130, 1, 2, 7, 192, 198, 194, 2, 2, 8, 192, 198, 2, 223, 1, 5, 192, 198, 163, 224, 1, 6, 192, 198, 3, 225, 1, 6, 192, 198, 99, 250, 1, 6, 192, 198, 195, 5, 1, 12, 0, 5, 1, 4, 192, 198, 130, 205, 1, 4, 192, 198, 194, 206, 1, 5, 192, 198, 2, 222, 1, 5, 192, 198, 67, 1, 1, 3, 192, 198, 130, 2, 1, 3, 192, 198, 194, 3, 1, 4, 192, 198, 2, 4, 1, 4, 192, 198, 66, 197, 0, 26, 0, 227, 0, 12, 0, 227, 0, 2, 192, 198, 130, 228, 0, 2, 192, 198, 194, 229, 0, 3, 192, 198, 2, 0, 1, 3, 192, 198, 66, 197, 0, 1, 192, 198, 130, 224, 0, 1, 192, 198, 194, 225, 0, 2, 192, 198, 2, 226, 0, 2, 192, 198, 66, 193, 0, 12, 0, 193, 0, 0, 192, 198, 130, 194, 0, 0, 192, 198, 194, 195, 0, 1, 192, 198, 2, 196, 0, 1, 192, 198, 66, 65, 0, 198, 129, 97, 0, 198, 161, 192, 0, 0, 192, 198, 66, 0, 50, 156, 5, 48, 0, 40, 3, 17, 192, 198, 193, 0, 70, 156, 5, 1, 0, 8, 3, 17, 192, 198, 225, 11, 3, 18, 192, 198, 1, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 31, 0, 39, 2, 59, 0, 169, 30, 26, 0, 177, 30, 12, 0, 177, 30, 25, 192, 198, 131, 179, 30, 25, 192, 198, 227, 181, 30, 26, 192, 198, 67, 183, 30, 26, 192, 198, 163, 169, 30, 24, 192, 198, 3, 171, 30, 24, 192, 198, 99, 173, 30, 24, 192, 198, 195, 175, 30, 25, 192, 198, 35, 161, 30, 12, 0, 161, 30, 22, 192, 198, 194, 163, 30, 23, 192, 198, 2, 165, 30, 23, 192, 198, 67, 167, 30, 23, 192, 198, 163, 39, 2, 10, 0, 8, 3, 22, 192, 198, 97, 10, 3, 255, 255, 111, 93, 193, 4, 1, 30, 22, 192, 198, 130, 176, 69, 198, 34, 4, 3, 21, 192, 198, 35, 3, 1, 26, 0, 225, 1, 12, 0, 225, 1, 21, 192, 198, 35, 251, 1, 21, 192, 198, 131, 1, 2, 21, 192, 198, 226, 3, 2, 22, 192, 198, 34, 3, 1, 20, 192, 198, 2, 5, 1, 20, 192, 198, 66, 206, 1, 20, 192, 198, 130, 223, 1, 20, 192, 198, 195, 227, 0, 12, 0, 227, 0, 19, 192, 198, 2, 228, 0, 19, 192, 198, 66, 229, 0, 19, 192, 198, 130, 1, 1, 19, 192, 198, 194, 97, 0, 18, 192, 198, 33, 224, 0, 18, 192, 198, 66, 225, 0, 18, 192, 198, 130, 226, 0, 18, 192, 198, 194, 0, 50, 5, 5, 48, 0, 40, 3, 27, 192, 198, 1, 0, 70, 5, 5, 1, 0, 8, 3, 27, 192, 198, 33, 11, 3, 27, 192, 198, 65, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 27, 192, 198, 162, 10, 3, 27, 192, 198, 226, 5, 42, 196, 162, 1, 0, 8, 3, 28, 192, 198, 34, 10, 3, 1, 192, 197, 34, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 98, 11, 3, 1, 192, 197, 162, 5, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 226, 11, 3, 2, 192, 197, 34, 156, 70, 196, 190, 1, 0, 8, 3, 28, 192, 198, 162, 11, 3, 28, 192, 198, 226, 5, 70, 196, 190, 1, 0, 8, 3, 29, 192, 198, 34, 11, 3, 29, 192, 198, 98, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 98, 11, 3, 2, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 156, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 98, 11, 3, 30, 192, 198, 162, 5, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 226, 11, 3, 31, 192, 198, 34, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 1, 0, 8, 3, 32, 192, 198, 162, 10, 3, 32, 192, 198, 226, 5, 42, 196, 200, 1, 0, 8, 3, 33, 192, 198, 34, 10, 3, 12, 192, 197, 162, 156, 82, 196, 198, 1, 0, 8, 3, 12, 192, 197, 226, 11, 3, 13, 192, 197, 34, 5, 82, 196, 198, 1, 0, 8, 3, 13, 192, 197, 98, 11, 3, 13, 192, 197, 162, 156, 82, 196, 208, 1, 0, 8, 3, 13, 192, 197, 226, 11, 3, 14, 192, 197, 34, 5, 82, 196, 208, 1, 0, 8, 3, 14, 192, 197, 98, 11, 3, 14, 192, 197, 162, 156, 82, 196, 204, 1, 0, 8, 3, 14, 192, 197, 226, 11, 3, 15, 192, 197, 34, 5, 82, 196, 204, 1, 0, 8, 3, 15, 192, 197, 98, 11, 3, 15, 192, 197, 162, 156, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 98, 10, 3, 33, 192, 198, 162, 5, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 226, 10, 3, 15, 192, 197, 226, 156, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 34, 11, 3, 34, 192, 198, 98, 5, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 162, 11, 3, 34, 192, 198, 226, 16, 0, 197, 35, 1, 0, 8, 3, 35, 192, 198, 35, 11, 3, 35, 192, 198, 131, 16, 0, 197, 131, 1, 0, 8, 3, 35, 192, 198, 227, 11, 3, 36, 192, 198, 67, 156, 82, 196, 196, 1, 0, 8, 3, 16, 192, 197, 226, 11, 3, 17, 192, 197, 34, 5, 82, 196, 196, 1, 0, 8, 3, 17, 192, 197, 98, 11, 3, 17, 192, 197, 162, 17, 0, 197, 227, 1, 0, 8, 3, 18, 192, 197, 67, 11, 3, 18, 192, 197, 163, 19, 0, 197, 3, 1, 0, 8, 3, 19, 192, 197, 99, 11, 3, 19, 192, 197, 195, 26, 0, 65, 0, 66, 0, 97, 0, 98, 0, 192, 0, 198, 0, 224, 0, 230, 0, 0, 1, 6, 1, 205, 1, 207, 1, 222, 1, 226, 1, 250, 1, 252, 1, 0, 2, 4, 2, 38, 2, 40, 2, 0, 30, 2, 30, 160, 30, 184, 30, 43, 33, 44, 33, 3, 0, 67, 0, 58, 0, 32, 0, 119, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 64, 51, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 52, 30, 0, 0, 52, 30, 0, 0, 92, 39, 0, 0, 92, 39, 0, 0, 56, 41, 0, 0, 56, 41, 0, 0, 88, 46, 0, 0, 142, 46, 0, 0, 36, 51, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 214, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 48, 0, 48, 0, 170, 0, 48, 0, 178, 0, 183, 0, 191, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 197, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 158, 0, 48, 0, 48, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 13, 0, 202, 193, 13, 0, 202, 226, 13, 0, 202, 3, 14, 0, 202, 36, 14, 0, 202, 69, 14, 0, 202, 102, 14, 0, 202, 135, 14, 0, 202, 168, 14, 0, 202, 201, 14, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 4, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 166, 28, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 102, 29, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 134, 30, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 36, 32, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 230, 46, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 166, 47, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 198, 48, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 129, 8, 0, 198, 161, 8, 0, 198, 97, 27, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 225, 17, 0, 5, 5, 127, 12, 198, 129, 27, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 98, 13, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 198, 97, 22, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 198, 33, 27, 0, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 50, 0, 201, 166, 51, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 193, 17, 0, 198, 1, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 18, 0, 198, 65, 27, 0, 198, 97, 28, 0, 198, 129, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 198, 52, 0, 201, 230, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 55, 0, 201, 38, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 57, 0, 201, 102, 58, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 192, 0, 0, 0, 198, 162, 29, 0, 198, 226, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 30, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 59, 0, 201, 166, 60, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 31, 0, 197, 34, 5, 0, 198, 162, 31, 0, 197, 98, 5, 0, 198, 226, 31, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 32, 0, 198, 98, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 61, 0, 201, 230, 62, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 64, 0, 201, 38, 65, 0, 201, 70, 66, 0, 201, 102, 67, 0, 201, 134, 68, 0, 201, 166, 69, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 70, 0, 201, 230, 71, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 73, 0, 201, 38, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 75, 0, 201, 102, 76, 0, 201, 134, 77, 0, 201, 166, 78, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 79, 0, 201, 230, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 70, 0, 70, 0, 4, 111, 93, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 134, 0, 71, 0, 2, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 71, 0, 2, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 3, 111, 93, 0, 5, 0, 72, 0, 3, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 4, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 0, 0, 62, 0, 3, 2, 118, 0, 169, 30, 54, 0, 177, 30, 26, 0, 181, 30, 12, 0, 181, 30, 16, 192, 198, 163, 182, 30, 17, 192, 198, 3, 183, 30, 17, 192, 198, 99, 43, 33, 1, 192, 198, 130, 177, 30, 15, 192, 198, 35, 178, 30, 15, 192, 198, 131, 179, 30, 15, 192, 198, 227, 180, 30, 16, 192, 198, 67, 173, 30, 12, 0, 173, 30, 13, 192, 198, 163, 174, 30, 14, 192, 198, 3, 175, 30, 14, 192, 198, 99, 176, 30, 14, 192, 198, 195, 169, 30, 12, 192, 198, 35, 170, 30, 12, 192, 198, 131, 171, 30, 12, 192, 198, 227, 172, 30, 13, 192, 198, 67, 161, 30, 26, 0, 165, 30, 12, 0, 165, 30, 10, 192, 198, 163, 166, 30, 11, 192, 198, 3, 167, 30, 11, 192, 198, 99, 168, 30, 11, 192, 198, 195, 161, 30, 9, 192, 198, 130, 162, 30, 9, 192, 198, 194, 163, 30, 10, 192, 198, 2, 164, 30, 10, 192, 198, 67, 10, 3, 12, 0, 10, 3, 8, 192, 198, 161, 0, 30, 8, 192, 198, 194, 1, 30, 9, 192, 198, 2, 160, 30, 9, 192, 198, 66, 3, 2, 8, 192, 198, 66, 38, 2, 5, 0, 39, 2, 8, 0, 8, 3, 8, 192, 198, 129, 240, 65, 198, 2, 4, 3, 6, 192, 198, 3, 240, 65, 198, 98, 4, 3, 6, 192, 198, 99, 1, 1, 54, 0, 223, 1, 26, 0, 251, 1, 12, 0, 251, 1, 7, 192, 198, 35, 0, 2, 7, 192, 198, 130, 1, 2, 7, 192, 198, 194, 2, 2, 8, 192, 198, 2, 223, 1, 5, 192, 198, 163, 224, 1, 6, 192, 198, 3, 225, 1, 6, 192, 198, 99, 250, 1, 6, 192, 198, 195, 5, 1, 12, 0, 5, 1, 4, 192, 198, 130, 205, 1, 4, 192, 198, 194, 206, 1, 5, 192, 198, 2, 222, 1, 5, 192, 198, 67, 1, 1, 3, 192, 198, 130, 2, 1, 3, 192, 198, 194, 3, 1, 4, 192, 198, 2, 4, 1, 4, 192, 198, 66, 197, 0, 26, 0, 227, 0, 12, 0, 227, 0, 2, 192, 198, 130, 228, 0, 2, 192, 198, 194, 229, 0, 3, 192, 198, 2, 0, 1, 3, 192, 198, 66, 197, 0, 1, 192, 198, 130, 224, 0, 1, 192, 198, 194, 225, 0, 2, 192, 198, 2, 226, 0, 2, 192, 198, 66, 193, 0, 12, 0, 193, 0, 0, 192, 198, 130, 194, 0, 0, 192, 198, 194, 195, 0, 1, 192, 198, 2, 196, 0, 1, 192, 198, 66, 65, 0, 198, 129, 97, 0, 198, 161, 192, 0, 0, 192, 198, 66, 0, 50, 156, 5, 48, 0, 40, 3, 17, 192, 198, 193, 0, 70, 156, 5, 1, 0, 8, 3, 17, 192, 198, 225, 11, 3, 18, 192, 198, 1, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 31, 0, 39, 2, 59, 0, 169, 30, 26, 0, 177, 30, 12, 0, 177, 30, 25, 192, 198, 131, 179, 30, 25, 192, 198, 227, 181, 30, 26, 192, 198, 67, 183, 30, 26, 192, 198, 163, 169, 30, 24, 192, 198, 3, 171, 30, 24, 192, 198, 99, 173, 30, 24, 192, 198, 195, 175, 30, 25, 192, 198, 35, 161, 30, 12, 0, 161, 30, 22, 192, 198, 194, 163, 30, 23, 192, 198, 2, 165, 30, 23, 192, 198, 67, 167, 30, 23, 192, 198, 163, 39, 2, 10, 0, 8, 3, 22, 192, 198, 97, 10, 3, 255, 255, 111, 93, 193, 4, 1, 30, 22, 192, 198, 130, 176, 69, 198, 34, 4, 3, 21, 192, 198, 35, 3, 1, 26, 0, 225, 1, 12, 0, 225, 1, 21, 192, 198, 35, 251, 1, 21, 192, 198, 131, 1, 2, 21, 192, 198, 226, 3, 2, 22, 192, 198, 34, 3, 1, 20, 192, 198, 2, 5, 1, 20, 192, 198, 66, 206, 1, 20, 192, 198, 130, 223, 1, 20, 192, 198, 195, 227, 0, 12, 0, 227, 0, 19, 192, 198, 2, 228, 0, 19, 192, 198, 66, 229, 0, 19, 192, 198, 130, 1, 1, 19, 192, 198, 194, 97, 0, 18, 192, 198, 33, 224, 0, 18, 192, 198, 66, 225, 0, 18, 192, 198, 130, 226, 0, 18, 192, 198, 194, 0, 50, 5, 5, 48, 0, 40, 3, 27, 192, 198, 1, 0, 70, 5, 5, 1, 0, 8, 3, 27, 192, 198, 33, 11, 3, 27, 192, 198, 65, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 27, 192, 198, 162, 10, 3, 27, 192, 198, 226, 5, 42, 196, 162, 1, 0, 8, 3, 28, 192, 198, 34, 10, 3, 1, 192, 197, 34, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 98, 11, 3, 1, 192, 197, 162, 5, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 226, 11, 3, 2, 192, 197, 34, 156, 70, 196, 190, 1, 0, 8, 3, 28, 192, 198, 162, 11, 3, 28, 192, 198, 226, 5, 70, 196, 190, 1, 0, 8, 3, 29, 192, 198, 34, 11, 3, 29, 192, 198, 98, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 98, 11, 3, 2, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 156, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 98, 11, 3, 30, 192, 198, 162, 5, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 226, 11, 3, 31, 192, 198, 34, 156, 42, 196, 200, 1, 0, 8, 3, 32, 192, 198, 162, 10, 3, 32, 192, 198, 226, 5, 42, 196, 200, 1, 0, 8, 3, 33, 192, 198, 34, 10, 3, 5, 192, 197, 226, 156, 82, 196, 198, 1, 0, 8, 3, 6, 192, 197, 34, 11, 3, 6, 192, 197, 98, 5, 82, 196, 198, 1, 0, 8, 3, 6, 192, 197, 162, 11, 3, 6, 192, 197, 226, 156, 82, 196, 208, 1, 0, 8, 3, 7, 192, 197, 34, 11, 3, 7, 192, 197, 98, 5, 82, 196, 208, 1, 0, 8, 3, 7, 192, 197, 162, 11, 3, 7, 192, 197, 226, 156, 82, 196, 204, 1, 0, 8, 3, 8, 192, 197, 34, 11, 3, 8, 192, 197, 98, 5, 82, 196, 204, 1, 0, 8, 3, 8, 192, 197, 162, 11, 3, 8, 192, 197, 226, 156, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 98, 10, 3, 33, 192, 198, 162, 5, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 226, 10, 3, 9, 192, 197, 34, 156, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 34, 11, 3, 34, 192, 198, 98, 5, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 162, 11, 3, 34, 192, 198, 226, 9, 0, 197, 99, 1, 0, 8, 3, 35, 192, 198, 35, 11, 3, 35, 192, 198, 131, 9, 0, 197, 195, 1, 0, 8, 3, 35, 192, 198, 227, 11, 3, 36, 192, 198, 67, 156, 82, 196, 196, 1, 0, 8, 3, 10, 192, 197, 34, 11, 3, 10, 192, 197, 98, 5, 82, 196, 196, 1, 0, 8, 3, 10, 192, 197, 162, 11, 3, 10, 192, 197, 226, 11, 0, 197, 35, 1, 0, 8, 3, 11, 192, 197, 131, 11, 3, 11, 192, 197, 227, 12, 0, 197, 67, 1, 0, 8, 3, 12, 192, 197, 163, 11, 3, 13, 192, 197, 3, 26, 0, 65, 0, 66, 0, 97, 0, 98, 0, 192, 0, 198, 0, 224, 0, 230, 0, 0, 1, 6, 1, 205, 1, 207, 1, 222, 1, 226, 1, 250, 1, 252, 1, 0, 2, 4, 2, 38, 2, 40, 2, 0, 30, 2, 30, 160, 30, 184, 30, 43, 33, 44, 33, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 20, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 22, 4, 185, 60, 185, 64, 185, 68, 82, 4, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 84, 4, 185, 124, 185, 128, 185, 136, 186, 140, 186, 144, 86, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 88, 4, 168, 60, 168, 64, 168, 68, 119, 4, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 121, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 123, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 121, 58, 217, 180, 185, 188, 185, 180, 185, 66, 185, 73, 153, 73, 249, 73, 89, 74, 186, 89, 154, 89, 250, 89, 90, 90, 249, 68, 122, 114, 185, 121, 153, 121, 249, 121, 121, 122, 217, 184, 136, 14, 185, 184, 185, 153, 153, 153, 249, 153, 217, 168, 153, 169, 0, 8, 2, 8, 168, 57, 136, 57, 232, 57, 104, 58, 200, 180, 168, 188, 168, 180, 168, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 232, 68, 104, 114, 168, 121, 136, 121, 232, 121, 104, 122, 200, 184, 128, 14, 168, 184, 168, 153, 136, 153, 232, 153, 200, 168, 136, 169, 4, 8, 72, 170, 249, 58, 232, 58, 217, 57, 200, 57, 125, 4, 127, 4, 153, 65, 136, 65, 249, 65, 232, 65, 153, 66, 136, 66, 25, 66, 8, 66, 25, 70, 8, 70, 217, 68, 200, 68, 249, 74, 232, 74, 217, 73, 200, 73, 153, 74, 136, 74, 249, 180, 232, 180, 25, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 153, 82, 136, 82, 185, 82, 168, 82, 249, 85, 232, 85, 25, 87, 8, 87, 122, 90, 104, 90, 250, 90, 232, 90, 218, 89, 200, 89, 218, 90, 200, 90, 154, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 185, 102, 168, 102, 168, 132, 153, 105, 136, 105, 185, 106, 168, 106, 25, 106, 8, 106, 57, 107, 40, 107, 25, 107, 8, 107, 154, 113, 136, 113, 186, 114, 168, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 249, 122, 232, 122, 217, 121, 200, 121, 249, 184, 232, 184, 25, 185, 8, 185, 153, 137, 136, 137, 185, 138, 168, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 186, 142, 168, 142, 26, 142, 8, 142, 186, 146, 168, 146, 26, 146, 8, 146, 185, 148, 168, 148, 121, 154, 104, 154, 249, 154, 232, 154, 217, 153, 200, 153, 57, 154, 40, 154, 249, 168, 232, 168, 129, 4, 131, 4, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 153, 174, 136, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 12, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 14, 8, 1, 0, 8, 14, 16, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 18, 8, 200, 12, 72, 12, 32, 14, 187, 144, 185, 84, 200, 140, 169, 140, 169, 144, 168, 84, 187, 88, 187, 96, 169, 88, 169, 96, 169, 176, 169, 112, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 65, 4, 218, 188, 97, 4, 209, 188, 192, 6, 218, 188, 168, 1, 193, 6, 218, 188, 136, 1, 194, 6, 218, 188, 232, 1, 195, 6, 218, 188, 104, 2, 196, 6, 218, 188, 72, 2, 197, 6, 218, 188, 40, 2, 224, 6, 209, 188, 168, 1, 225, 6, 209, 188, 136, 1, 226, 6, 209, 188, 232, 1, 227, 6, 209, 188, 104, 2, 228, 6, 209, 188, 72, 2, 229, 6, 209, 188, 40, 2, 0, 7, 218, 188, 232, 2, 1, 7, 209, 188, 232, 2, 2, 7, 218, 188, 200, 1, 3, 7, 209, 188, 200, 1, 4, 7, 218, 188, 200, 2, 5, 7, 209, 188, 200, 2, 255, 5, 185, 72, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 97, 4, 200, 188, 224, 6, 200, 188, 168, 1, 225, 6, 200, 188, 136, 1, 226, 6, 200, 188, 232, 1, 227, 6, 200, 188, 104, 2, 228, 6, 200, 188, 72, 2, 229, 6, 200, 188, 40, 2, 1, 7, 200, 188, 232, 2, 3, 7, 200, 188, 200, 1, 5, 7, 200, 188, 200, 2, 255, 5, 168, 72, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 217, 58, 255, 5, 200, 58, 255, 5, 217, 154, 255, 5, 200, 154, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 191, 14, 0, 16, 57, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 185, 14, 0, 32, 144, 27, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 149, 27, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 122, 0, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 126, 0, 0, 0, 126, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 60, 1, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 8, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 32, 0, 0, 0, 36, 0, 0, 0, 36, 1, 0, 0, 10, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 39, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 3, 0, 69, 0, 60, 0, 32, 0, 35, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 115, 0, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 120, 0, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 85, 27, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 89, 27, 0, 0, 89, 27, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 51, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 110, 0, 98, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 110, 0, 98, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 0, 0, 4, 57, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 163, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 100, 38, 0, 0, 100, 38, 0, 0, 140, 47, 0, 0, 140, 47, 0, 0, 76, 51, 0, 0, 76, 51, 0, 0, 178, 56, 0, 0, 232, 56, 0, 0, 232, 56, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 89, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 145, 0, 153, 0, 159, 0, 167, 0, 175, 0, 183, 0, 48, 0, 56, 0, 182, 0, 190, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 207, 0, 215, 0, 48, 0, 48, 0, 48, 0, 223, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 231, 0, 48, 0, 48, 0, 239, 0, 48, 0, 245, 0, 253, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 8, 1, 48, 0, 48, 0, 12, 1, 48, 0, 20, 1, 25, 1, 33, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 39, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 47, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 50, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 198, 0, 205, 0, 48, 0, 48, 0, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 80, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 58, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 66, 1, 48, 0, 48, 0, 72, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 68, 2, 124, 2, 188, 2, 192, 0, 216, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 192, 28, 0, 202, 225, 28, 0, 202, 2, 29, 0, 202, 35, 29, 0, 202, 68, 29, 0, 202, 101, 29, 0, 202, 134, 29, 0, 202, 167, 29, 0, 202, 200, 29, 0, 202, 233, 29, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 196, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 102, 29, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 38, 30, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 70, 31, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 228, 32, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 166, 47, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 102, 48, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 134, 49, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 129, 8, 0, 198, 161, 8, 0, 198, 97, 27, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 225, 17, 0, 5, 5, 127, 12, 198, 129, 27, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 34, 20, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 198, 97, 22, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 198, 33, 27, 0, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 51, 0, 201, 102, 52, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 193, 17, 0, 198, 1, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 18, 0, 198, 65, 27, 0, 198, 97, 28, 0, 198, 129, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 134, 53, 0, 201, 166, 54, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 55, 0, 201, 230, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 58, 0, 201, 38, 59, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 192, 0, 0, 0, 198, 162, 29, 0, 198, 226, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 30, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 60, 0, 201, 102, 61, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 31, 0, 197, 34, 5, 0, 198, 162, 31, 0, 197, 98, 5, 0, 198, 226, 31, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 32, 0, 198, 98, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 134, 62, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 64, 0, 5, 71, 238, 102, 201, 70, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 6, 0, 192, 0, 0, 0, 197, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 7, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 197, 34, 8, 0, 197, 99, 8, 0, 197, 194, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 9, 0, 197, 67, 9, 0, 197, 162, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 226, 5, 0, 197, 34, 10, 0, 5, 5, 10, 121, 197, 98, 10, 0, 197, 162, 10, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 66, 0, 201, 70, 67, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 68, 0, 201, 134, 69, 0, 201, 166, 70, 0, 201, 198, 71, 0, 201, 230, 72, 0, 201, 6, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 75, 0, 201, 70, 76, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 77, 0, 201, 134, 78, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 79, 0, 201, 198, 80, 0, 201, 230, 81, 0, 201, 6, 83, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 84, 0, 201, 70, 85, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 70, 0, 70, 0, 4, 111, 93, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 134, 0, 71, 0, 2, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 71, 0, 2, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 3, 111, 93, 0, 5, 0, 72, 0, 3, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 4, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 193, 4, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 6, 121, 197, 226, 5, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 34, 6, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 98, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 226, 6, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 34, 7, 0, 5, 5, 107, 121, 197, 98, 7, 0, 5, 5, 111, 121, 197, 162, 7, 0, 5, 5, 115, 121, 197, 226, 7, 0, 5, 5, 119, 121, 197, 34, 8, 0, 197, 99, 8, 0, 197, 194, 8, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 2, 9, 0, 197, 67, 9, 0, 197, 162, 9, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 226, 9, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 226, 5, 0, 197, 34, 10, 0, 5, 5, 10, 121, 197, 98, 10, 0, 197, 162, 10, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 6, 0, 5, 5, 24, 121, 197, 162, 6, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 0, 0, 62, 0, 3, 2, 118, 0, 169, 30, 54, 0, 177, 30, 26, 0, 181, 30, 12, 0, 181, 30, 16, 192, 198, 163, 182, 30, 17, 192, 198, 3, 183, 30, 17, 192, 198, 99, 43, 33, 1, 192, 198, 130, 177, 30, 15, 192, 198, 35, 178, 30, 15, 192, 198, 131, 179, 30, 15, 192, 198, 227, 180, 30, 16, 192, 198, 67, 173, 30, 12, 0, 173, 30, 13, 192, 198, 163, 174, 30, 14, 192, 198, 3, 175, 30, 14, 192, 198, 99, 176, 30, 14, 192, 198, 195, 169, 30, 12, 192, 198, 35, 170, 30, 12, 192, 198, 131, 171, 30, 12, 192, 198, 227, 172, 30, 13, 192, 198, 67, 161, 30, 26, 0, 165, 30, 12, 0, 165, 30, 10, 192, 198, 163, 166, 30, 11, 192, 198, 3, 167, 30, 11, 192, 198, 99, 168, 30, 11, 192, 198, 195, 161, 30, 9, 192, 198, 130, 162, 30, 9, 192, 198, 194, 163, 30, 10, 192, 198, 2, 164, 30, 10, 192, 198, 67, 10, 3, 12, 0, 10, 3, 8, 192, 198, 161, 0, 30, 8, 192, 198, 194, 1, 30, 9, 192, 198, 2, 160, 30, 9, 192, 198, 66, 3, 2, 8, 192, 198, 66, 38, 2, 5, 0, 39, 2, 8, 0, 8, 3, 8, 192, 198, 129, 240, 65, 198, 2, 4, 3, 6, 192, 198, 3, 240, 65, 198, 98, 4, 3, 6, 192, 198, 99, 1, 1, 54, 0, 223, 1, 26, 0, 251, 1, 12, 0, 251, 1, 7, 192, 198, 35, 0, 2, 7, 192, 198, 130, 1, 2, 7, 192, 198, 194, 2, 2, 8, 192, 198, 2, 223, 1, 5, 192, 198, 163, 224, 1, 6, 192, 198, 3, 225, 1, 6, 192, 198, 99, 250, 1, 6, 192, 198, 195, 5, 1, 12, 0, 5, 1, 4, 192, 198, 130, 205, 1, 4, 192, 198, 194, 206, 1, 5, 192, 198, 2, 222, 1, 5, 192, 198, 67, 1, 1, 3, 192, 198, 130, 2, 1, 3, 192, 198, 194, 3, 1, 4, 192, 198, 2, 4, 1, 4, 192, 198, 66, 197, 0, 26, 0, 227, 0, 12, 0, 227, 0, 2, 192, 198, 130, 228, 0, 2, 192, 198, 194, 229, 0, 3, 192, 198, 2, 0, 1, 3, 192, 198, 66, 197, 0, 1, 192, 198, 130, 224, 0, 1, 192, 198, 194, 225, 0, 2, 192, 198, 2, 226, 0, 2, 192, 198, 66, 193, 0, 12, 0, 193, 0, 0, 192, 198, 130, 194, 0, 0, 192, 198, 194, 195, 0, 1, 192, 198, 2, 196, 0, 1, 192, 198, 66, 65, 0, 198, 129, 97, 0, 198, 161, 192, 0, 0, 192, 198, 66, 0, 50, 156, 5, 48, 0, 40, 3, 17, 192, 198, 193, 0, 70, 156, 5, 1, 0, 8, 3, 17, 192, 198, 225, 11, 3, 18, 192, 198, 1, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 31, 0, 39, 2, 59, 0, 169, 30, 26, 0, 177, 30, 12, 0, 177, 30, 25, 192, 198, 131, 179, 30, 25, 192, 198, 227, 181, 30, 26, 192, 198, 67, 183, 30, 26, 192, 198, 163, 169, 30, 24, 192, 198, 3, 171, 30, 24, 192, 198, 99, 173, 30, 24, 192, 198, 195, 175, 30, 25, 192, 198, 35, 161, 30, 12, 0, 161, 30, 22, 192, 198, 194, 163, 30, 23, 192, 198, 2, 165, 30, 23, 192, 198, 67, 167, 30, 23, 192, 198, 163, 39, 2, 10, 0, 8, 3, 22, 192, 198, 97, 10, 3, 255, 255, 111, 93, 193, 4, 1, 30, 22, 192, 198, 130, 176, 69, 198, 34, 4, 3, 21, 192, 198, 35, 3, 1, 26, 0, 225, 1, 12, 0, 225, 1, 21, 192, 198, 35, 251, 1, 21, 192, 198, 131, 1, 2, 21, 192, 198, 226, 3, 2, 22, 192, 198, 34, 3, 1, 20, 192, 198, 2, 5, 1, 20, 192, 198, 66, 206, 1, 20, 192, 198, 130, 223, 1, 20, 192, 198, 195, 227, 0, 12, 0, 227, 0, 19, 192, 198, 2, 228, 0, 19, 192, 198, 66, 229, 0, 19, 192, 198, 130, 1, 1, 19, 192, 198, 194, 97, 0, 18, 192, 198, 33, 224, 0, 18, 192, 198, 66, 225, 0, 18, 192, 198, 130, 226, 0, 18, 192, 198, 194, 0, 50, 5, 5, 48, 0, 40, 3, 27, 192, 198, 1, 0, 70, 5, 5, 1, 0, 8, 3, 27, 192, 198, 33, 11, 3, 27, 192, 198, 65, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 27, 192, 198, 162, 10, 3, 27, 192, 198, 226, 5, 42, 196, 162, 1, 0, 8, 3, 28, 192, 198, 34, 10, 3, 1, 192, 197, 34, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 98, 11, 3, 1, 192, 197, 162, 5, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 226, 11, 3, 2, 192, 197, 34, 156, 70, 196, 190, 1, 0, 8, 3, 28, 192, 198, 162, 11, 3, 28, 192, 198, 226, 5, 70, 196, 190, 1, 0, 8, 3, 29, 192, 198, 34, 11, 3, 29, 192, 198, 98, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 98, 11, 3, 2, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 156, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 98, 11, 3, 30, 192, 198, 162, 5, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 226, 11, 3, 31, 192, 198, 34, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 1, 0, 8, 3, 32, 192, 198, 162, 10, 3, 32, 192, 198, 226, 5, 42, 196, 200, 1, 0, 8, 3, 33, 192, 198, 34, 10, 3, 12, 192, 197, 162, 156, 82, 196, 198, 1, 0, 8, 3, 12, 192, 197, 226, 11, 3, 13, 192, 197, 34, 5, 82, 196, 198, 1, 0, 8, 3, 13, 192, 197, 98, 11, 3, 13, 192, 197, 162, 156, 82, 196, 208, 1, 0, 8, 3, 13, 192, 197, 226, 11, 3, 14, 192, 197, 34, 5, 82, 196, 208, 1, 0, 8, 3, 14, 192, 197, 98, 11, 3, 14, 192, 197, 162, 156, 82, 196, 204, 1, 0, 8, 3, 14, 192, 197, 226, 11, 3, 15, 192, 197, 34, 5, 82, 196, 204, 1, 0, 8, 3, 15, 192, 197, 98, 11, 3, 15, 192, 197, 162, 156, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 98, 10, 3, 33, 192, 198, 162, 5, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 226, 10, 3, 15, 192, 197, 226, 156, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 34, 11, 3, 34, 192, 198, 98, 5, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 162, 11, 3, 34, 192, 198, 226, 16, 0, 197, 35, 1, 0, 8, 3, 35, 192, 198, 35, 11, 3, 35, 192, 198, 131, 16, 0, 197, 131, 1, 0, 8, 3, 35, 192, 198, 227, 11, 3, 36, 192, 198, 67, 156, 82, 196, 196, 1, 0, 8, 3, 16, 192, 197, 226, 11, 3, 17, 192, 197, 34, 5, 82, 196, 196, 1, 0, 8, 3, 17, 192, 197, 98, 11, 3, 17, 192, 197, 162, 17, 0, 197, 227, 1, 0, 8, 3, 18, 192, 197, 67, 11, 3, 18, 192, 197, 163, 19, 0, 197, 3, 1, 0, 8, 3, 19, 192, 197, 99, 11, 3, 19, 192, 197, 195, 26, 0, 65, 0, 66, 0, 97, 0, 98, 0, 192, 0, 198, 0, 224, 0, 230, 0, 0, 1, 6, 1, 205, 1, 207, 1, 222, 1, 226, 1, 250, 1, 252, 1, 0, 2, 4, 2, 38, 2, 40, 2, 0, 30, 2, 30, 160, 30, 184, 30, 43, 33, 44, 33, 3, 0, 67, 0, 58, 0, 32, 0, 51, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 64, 51, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 52, 30, 0, 0, 52, 30, 0, 0, 92, 39, 0, 0, 92, 39, 0, 0, 56, 41, 0, 0, 56, 41, 0, 0, 88, 46, 0, 0, 142, 46, 0, 0, 36, 51, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 214, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 48, 0, 48, 0, 170, 0, 48, 0, 178, 0, 183, 0, 191, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 197, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 158, 0, 48, 0, 48, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 13, 0, 202, 193, 13, 0, 202, 226, 13, 0, 202, 3, 14, 0, 202, 36, 14, 0, 202, 69, 14, 0, 202, 102, 14, 0, 202, 135, 14, 0, 202, 168, 14, 0, 202, 201, 14, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 4, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 166, 28, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 102, 29, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 134, 30, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 36, 32, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 230, 46, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 166, 47, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 198, 48, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 129, 8, 0, 198, 161, 8, 0, 198, 97, 27, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 225, 17, 0, 5, 5, 127, 12, 198, 129, 27, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 98, 13, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 198, 97, 22, 0, 193, 4, 111, 93, 193, 2, 111, 93, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 198, 33, 27, 0, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 50, 0, 201, 166, 51, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 193, 17, 0, 198, 1, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 18, 0, 198, 65, 27, 0, 198, 97, 28, 0, 198, 129, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 198, 52, 0, 201, 230, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 55, 0, 201, 38, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 57, 0, 201, 102, 58, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 3, 0, 197, 162, 3, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 192, 0, 0, 0, 198, 162, 29, 0, 198, 226, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 30, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 59, 0, 201, 166, 60, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 31, 0, 197, 34, 5, 0, 198, 162, 31, 0, 197, 98, 5, 0, 198, 226, 31, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 34, 32, 0, 198, 98, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 61, 0, 201, 230, 62, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 64, 0, 201, 38, 65, 0, 201, 70, 66, 0, 201, 102, 67, 0, 201, 134, 68, 0, 201, 166, 69, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 70, 0, 201, 230, 71, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 73, 0, 201, 38, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 75, 0, 201, 102, 76, 0, 201, 134, 77, 0, 201, 166, 78, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 79, 0, 201, 230, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 70, 0, 70, 0, 4, 111, 93, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 70, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 134, 0, 71, 0, 2, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 184, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 188, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 142, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 140, 0, 0, 0, 0, 0, 5, 0, 71, 0, 2, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 3, 111, 93, 0, 5, 0, 72, 0, 3, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 4, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 2, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 0, 0, 62, 0, 3, 2, 118, 0, 169, 30, 54, 0, 177, 30, 26, 0, 181, 30, 12, 0, 181, 30, 16, 192, 198, 163, 182, 30, 17, 192, 198, 3, 183, 30, 17, 192, 198, 99, 43, 33, 1, 192, 198, 130, 177, 30, 15, 192, 198, 35, 178, 30, 15, 192, 198, 131, 179, 30, 15, 192, 198, 227, 180, 30, 16, 192, 198, 67, 173, 30, 12, 0, 173, 30, 13, 192, 198, 163, 174, 30, 14, 192, 198, 3, 175, 30, 14, 192, 198, 99, 176, 30, 14, 192, 198, 195, 169, 30, 12, 192, 198, 35, 170, 30, 12, 192, 198, 131, 171, 30, 12, 192, 198, 227, 172, 30, 13, 192, 198, 67, 161, 30, 26, 0, 165, 30, 12, 0, 165, 30, 10, 192, 198, 163, 166, 30, 11, 192, 198, 3, 167, 30, 11, 192, 198, 99, 168, 30, 11, 192, 198, 195, 161, 30, 9, 192, 198, 130, 162, 30, 9, 192, 198, 194, 163, 30, 10, 192, 198, 2, 164, 30, 10, 192, 198, 67, 10, 3, 12, 0, 10, 3, 8, 192, 198, 161, 0, 30, 8, 192, 198, 194, 1, 30, 9, 192, 198, 2, 160, 30, 9, 192, 198, 66, 3, 2, 8, 192, 198, 66, 38, 2, 5, 0, 39, 2, 8, 0, 8, 3, 8, 192, 198, 129, 240, 65, 198, 2, 4, 3, 6, 192, 198, 3, 240, 65, 198, 98, 4, 3, 6, 192, 198, 99, 1, 1, 54, 0, 223, 1, 26, 0, 251, 1, 12, 0, 251, 1, 7, 192, 198, 35, 0, 2, 7, 192, 198, 130, 1, 2, 7, 192, 198, 194, 2, 2, 8, 192, 198, 2, 223, 1, 5, 192, 198, 163, 224, 1, 6, 192, 198, 3, 225, 1, 6, 192, 198, 99, 250, 1, 6, 192, 198, 195, 5, 1, 12, 0, 5, 1, 4, 192, 198, 130, 205, 1, 4, 192, 198, 194, 206, 1, 5, 192, 198, 2, 222, 1, 5, 192, 198, 67, 1, 1, 3, 192, 198, 130, 2, 1, 3, 192, 198, 194, 3, 1, 4, 192, 198, 2, 4, 1, 4, 192, 198, 66, 197, 0, 26, 0, 227, 0, 12, 0, 227, 0, 2, 192, 198, 130, 228, 0, 2, 192, 198, 194, 229, 0, 3, 192, 198, 2, 0, 1, 3, 192, 198, 66, 197, 0, 1, 192, 198, 130, 224, 0, 1, 192, 198, 194, 225, 0, 2, 192, 198, 2, 226, 0, 2, 192, 198, 66, 193, 0, 12, 0, 193, 0, 0, 192, 198, 130, 194, 0, 0, 192, 198, 194, 195, 0, 1, 192, 198, 2, 196, 0, 1, 192, 198, 66, 65, 0, 198, 129, 97, 0, 198, 161, 192, 0, 0, 192, 198, 66, 0, 50, 156, 5, 48, 0, 40, 3, 17, 192, 198, 193, 0, 70, 156, 5, 1, 0, 8, 3, 17, 192, 198, 225, 11, 3, 18, 192, 198, 1, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 31, 0, 39, 2, 59, 0, 169, 30, 26, 0, 177, 30, 12, 0, 177, 30, 25, 192, 198, 131, 179, 30, 25, 192, 198, 227, 181, 30, 26, 192, 198, 67, 183, 30, 26, 192, 198, 163, 169, 30, 24, 192, 198, 3, 171, 30, 24, 192, 198, 99, 173, 30, 24, 192, 198, 195, 175, 30, 25, 192, 198, 35, 161, 30, 12, 0, 161, 30, 22, 192, 198, 194, 163, 30, 23, 192, 198, 2, 165, 30, 23, 192, 198, 67, 167, 30, 23, 192, 198, 163, 39, 2, 10, 0, 8, 3, 22, 192, 198, 97, 10, 3, 255, 255, 111, 93, 193, 4, 1, 30, 22, 192, 198, 130, 176, 69, 198, 34, 4, 3, 21, 192, 198, 35, 3, 1, 26, 0, 225, 1, 12, 0, 225, 1, 21, 192, 198, 35, 251, 1, 21, 192, 198, 131, 1, 2, 21, 192, 198, 226, 3, 2, 22, 192, 198, 34, 3, 1, 20, 192, 198, 2, 5, 1, 20, 192, 198, 66, 206, 1, 20, 192, 198, 130, 223, 1, 20, 192, 198, 195, 227, 0, 12, 0, 227, 0, 19, 192, 198, 2, 228, 0, 19, 192, 198, 66, 229, 0, 19, 192, 198, 130, 1, 1, 19, 192, 198, 194, 97, 0, 18, 192, 198, 33, 224, 0, 18, 192, 198, 66, 225, 0, 18, 192, 198, 130, 226, 0, 18, 192, 198, 194, 0, 50, 5, 5, 48, 0, 40, 3, 27, 192, 198, 1, 0, 70, 5, 5, 1, 0, 8, 3, 27, 192, 198, 33, 11, 3, 27, 192, 198, 65, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 27, 192, 198, 162, 10, 3, 27, 192, 198, 226, 5, 42, 196, 162, 1, 0, 8, 3, 28, 192, 198, 34, 10, 3, 1, 192, 197, 34, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 98, 11, 3, 1, 192, 197, 162, 5, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 226, 11, 3, 2, 192, 197, 34, 156, 70, 196, 190, 1, 0, 8, 3, 28, 192, 198, 162, 11, 3, 28, 192, 198, 226, 5, 70, 196, 190, 1, 0, 8, 3, 29, 192, 198, 34, 11, 3, 29, 192, 198, 98, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 98, 11, 3, 2, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 156, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 98, 11, 3, 30, 192, 198, 162, 5, 70, 196, 162, 1, 0, 8, 3, 30, 192, 198, 226, 11, 3, 31, 192, 198, 34, 156, 42, 196, 200, 1, 0, 8, 3, 32, 192, 198, 162, 10, 3, 32, 192, 198, 226, 5, 42, 196, 200, 1, 0, 8, 3, 33, 192, 198, 34, 10, 3, 5, 192, 197, 226, 156, 82, 196, 198, 1, 0, 8, 3, 6, 192, 197, 34, 11, 3, 6, 192, 197, 98, 5, 82, 196, 198, 1, 0, 8, 3, 6, 192, 197, 162, 11, 3, 6, 192, 197, 226, 156, 82, 196, 208, 1, 0, 8, 3, 7, 192, 197, 34, 11, 3, 7, 192, 197, 98, 5, 82, 196, 208, 1, 0, 8, 3, 7, 192, 197, 162, 11, 3, 7, 192, 197, 226, 156, 82, 196, 204, 1, 0, 8, 3, 8, 192, 197, 34, 11, 3, 8, 192, 197, 98, 5, 82, 196, 204, 1, 0, 8, 3, 8, 192, 197, 162, 11, 3, 8, 192, 197, 226, 156, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 98, 10, 3, 33, 192, 198, 162, 5, 42, 196, 196, 1, 0, 8, 3, 33, 192, 198, 226, 10, 3, 9, 192, 197, 34, 156, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 34, 11, 3, 34, 192, 198, 98, 5, 70, 196, 196, 1, 0, 8, 3, 34, 192, 198, 162, 11, 3, 34, 192, 198, 226, 9, 0, 197, 99, 1, 0, 8, 3, 35, 192, 198, 35, 11, 3, 35, 192, 198, 131, 9, 0, 197, 195, 1, 0, 8, 3, 35, 192, 198, 227, 11, 3, 36, 192, 198, 67, 156, 82, 196, 196, 1, 0, 8, 3, 10, 192, 197, 34, 11, 3, 10, 192, 197, 98, 5, 82, 196, 196, 1, 0, 8, 3, 10, 192, 197, 162, 11, 3, 10, 192, 197, 226, 11, 0, 197, 35, 1, 0, 8, 3, 11, 192, 197, 131, 11, 3, 11, 192, 197, 227, 12, 0, 197, 67, 1, 0, 8, 3, 12, 192, 197, 163, 11, 3, 13, 192, 197, 3, 26, 0, 65, 0, 66, 0, 97, 0, 98, 0, 192, 0, 198, 0, 224, 0, 230, 0, 0, 1, 6, 1, 205, 1, 207, 1, 222, 1, 226, 1, 250, 1, 252, 1, 0, 2, 4, 2, 38, 2, 40, 2, 0, 30, 2, 30, 160, 30, 184, 30, 43, 33, 44, 33, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 20, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 22, 4, 185, 60, 185, 64, 185, 68, 82, 4, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 84, 4, 185, 124, 185, 128, 185, 136, 186, 140, 186, 144, 86, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 88, 4, 168, 60, 168, 64, 168, 68, 119, 4, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 121, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 123, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 121, 58, 217, 180, 185, 188, 185, 180, 185, 66, 185, 73, 153, 73, 249, 73, 89, 74, 186, 89, 154, 89, 250, 89, 90, 90, 249, 68, 122, 114, 185, 121, 153, 121, 249, 121, 121, 122, 217, 184, 136, 14, 185, 184, 185, 153, 153, 153, 249, 153, 217, 168, 153, 169, 0, 8, 2, 8, 168, 57, 136, 57, 232, 57, 104, 58, 200, 180, 168, 188, 168, 180, 168, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 232, 68, 104, 114, 168, 121, 136, 121, 232, 121, 104, 122, 200, 184, 128, 14, 168, 184, 168, 153, 136, 153, 232, 153, 200, 168, 136, 169, 4, 8, 72, 170, 249, 58, 232, 58, 217, 57, 200, 57, 125, 4, 127, 4, 153, 65, 136, 65, 249, 65, 232, 65, 153, 66, 136, 66, 25, 66, 8, 66, 25, 70, 8, 70, 217, 68, 200, 68, 249, 74, 232, 74, 217, 73, 200, 73, 153, 74, 136, 74, 249, 180, 232, 180, 25, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 153, 82, 136, 82, 185, 82, 168, 82, 249, 85, 232, 85, 25, 87, 8, 87, 122, 90, 104, 90, 250, 90, 232, 90, 218, 89, 200, 89, 218, 90, 200, 90, 154, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 185, 102, 168, 102, 168, 132, 153, 105, 136, 105, 185, 106, 168, 106, 25, 106, 8, 106, 57, 107, 40, 107, 25, 107, 8, 107, 154, 113, 136, 113, 186, 114, 168, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 249, 122, 232, 122, 217, 121, 200, 121, 249, 184, 232, 184, 25, 185, 8, 185, 153, 137, 136, 137, 185, 138, 168, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 186, 142, 168, 142, 26, 142, 8, 142, 186, 146, 168, 146, 26, 146, 8, 146, 185, 148, 168, 148, 121, 154, 104, 154, 249, 154, 232, 154, 217, 153, 200, 153, 57, 154, 40, 154, 249, 168, 232, 168, 129, 4, 131, 4, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 153, 174, 136, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 12, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 14, 8, 1, 0, 8, 14, 16, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 18, 8, 200, 12, 72, 12, 32, 14, 187, 144, 185, 84, 200, 140, 169, 140, 169, 144, 168, 84, 187, 88, 187, 96, 169, 88, 169, 96, 169, 176, 169, 112, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 65, 4, 218, 188, 97, 4, 209, 188, 192, 6, 218, 188, 168, 1, 193, 6, 218, 188, 136, 1, 194, 6, 218, 188, 232, 1, 195, 6, 218, 188, 104, 2, 196, 6, 218, 188, 72, 2, 197, 6, 218, 188, 40, 2, 224, 6, 209, 188, 168, 1, 225, 6, 209, 188, 136, 1, 226, 6, 209, 188, 232, 1, 227, 6, 209, 188, 104, 2, 228, 6, 209, 188, 72, 2, 229, 6, 209, 188, 40, 2, 0, 7, 218, 188, 232, 2, 1, 7, 209, 188, 232, 2, 2, 7, 218, 188, 200, 1, 3, 7, 209, 188, 200, 1, 4, 7, 218, 188, 200, 2, 5, 7, 209, 188, 200, 2, 255, 5, 185, 72, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 97, 4, 200, 188, 224, 6, 200, 188, 168, 1, 225, 6, 200, 188, 136, 1, 226, 6, 200, 188, 232, 1, 227, 6, 200, 188, 104, 2, 228, 6, 200, 188, 72, 2, 229, 6, 200, 188, 40, 2, 1, 7, 200, 188, 232, 2, 3, 7, 200, 188, 200, 1, 5, 7, 200, 188, 200, 2, 255, 5, 168, 72, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 217, 58, 255, 5, 200, 58, 255, 5, 217, 154, 255, 5, 200, 154, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 123, 14, 0, 16, 30, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 117, 14, 0, 32, 76, 27, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 81, 27, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 14, 0, 0, 0, 14, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 14, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 110, 0, 98, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 110, 0, 98, 0, 95, 0, 78, 0, 79, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 10, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 14, 10, 0, 0, 14, 10, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 63, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 75, 220, 38, 0, 90, 0, 60, 0, 99, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 72, 0, 60, 0, 100, 0, 104, 0, 60, 0, 60, 0, 60, 0, 68, 0, 104, 0, 60, 0, 60, 0, 60, 0, 68, 0, 72, 0, 60, 0, 107, 0, 104, 0, 60, 0, 60, 0, 60, 0, 75, 0, 104, 0, 60, 0, 60, 0, 60, 0, 75, 0, 72, 0, 60, 0, 110, 0, 121, 0, 60, 0, 60, 0, 60, 0, 78, 0, 121, 0, 60, 0, 60, 0, 60, 0, 78, 0, 89, 0, 60, 0, 112, 0, 104, 0, 60, 0, 60, 0, 60, 0, 80, 0, 104, 0, 60, 0, 60, 0, 60, 0, 80, 0, 72, 0, 60, 0, 115, 0, 104, 0, 60, 0, 60, 0, 60, 0, 83, 0, 104, 0, 0, 0, 170, 170, 12, 39, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 196, 23, 0, 0, 196, 23, 0, 0, 140, 26, 0, 0, 140, 26, 0, 0, 200, 29, 0, 0, 200, 29, 0, 0, 30, 34, 0, 0, 96, 34, 0, 0, 240, 38, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 111, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 94, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 102, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 24, 0, 202, 193, 24, 0, 202, 226, 24, 0, 202, 3, 25, 0, 202, 36, 25, 0, 202, 69, 25, 0, 202, 102, 25, 0, 202, 135, 25, 0, 202, 168, 25, 0, 202, 201, 25, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 0, 0, 0, 201, 160, 7, 0, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 201, 96, 15, 0, 156, 5, 0, 64, 156, 5, 0, 66, 201, 32, 23, 0, 156, 5, 0, 70, 201, 32, 32, 0, 160, 5, 0, 74, 156, 5, 0, 76, 201, 160, 39, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 160, 43, 0, 201, 192, 47, 0, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 201, 224, 51, 0, 5, 5, 0, 64, 5, 5, 0, 66, 201, 0, 56, 0, 5, 5, 0, 70, 201, 32, 61, 0, 5, 5, 0, 74, 5, 5, 0, 76, 201, 64, 65, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 23, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 24, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 24, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 96, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 135, 0, 5, 0, 2, 5, 93, 0, 70, 0, 5, 0, 2, 5, 93, 0, 135, 0, 5, 0, 2, 5, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 135, 0, 5, 0, 2, 5, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 5, 0, 2, 5, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 135, 0, 5, 0, 2, 5, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 135, 0, 5, 0, 2, 5, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 93, 0, 5, 0, 146, 0, 0, 0, 0, 0, 135, 0, 5, 0, 2, 5, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 135, 0, 5, 0, 2, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 135, 0, 5, 0, 2, 5, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 135, 0, 5, 0, 2, 5, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 70, 0, 5, 0, 2, 5, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 135, 0, 5, 0, 3, 5, 93, 0, 70, 0, 5, 0, 3, 5, 93, 0, 135, 0, 5, 0, 3, 5, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 5, 0, 3, 5, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 135, 0, 5, 0, 3, 5, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 135, 0, 5, 0, 3, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 135, 0, 5, 0, 3, 5, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 135, 0, 5, 0, 3, 5, 93, 0, 5, 0, 160, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 93, 0, 5, 0, 160, 0, 0, 0, 0, 0, 135, 0, 5, 0, 3, 5, 93, 0, 5, 0, 206, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 93, 0, 5, 0, 206, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 5, 93, 0, 5, 0, 210, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 93, 0, 70, 0, 5, 0, 4, 5, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 93, 0, 5, 0, 160, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 93, 0, 5, 0, 206, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 5, 93, 0, 5, 0, 210, 0, 0, 0, 0, 0, 0, 0, 0, 135, 5, 2, 93, 5, 142, 0, 0, 70, 5, 2, 93, 5, 142, 0, 0, 135, 5, 2, 93, 5, 144, 0, 0, 70, 5, 2, 93, 5, 144, 0, 0, 135, 5, 2, 93, 5, 156, 0, 0, 70, 5, 2, 93, 5, 156, 0, 0, 135, 5, 2, 93, 5, 196, 0, 0, 70, 5, 2, 93, 5, 196, 0, 0, 135, 5, 2, 93, 5, 150, 0, 0, 70, 5, 2, 93, 5, 150, 0, 0, 135, 5, 2, 93, 5, 160, 0, 0, 70, 5, 2, 93, 5, 160, 0, 0, 135, 5, 2, 93, 5, 206, 0, 0, 70, 5, 2, 93, 5, 206, 0, 0, 70, 5, 2, 93, 5, 210, 0, 0, 135, 5, 3, 93, 5, 142, 0, 0, 70, 5, 3, 93, 5, 142, 0, 0, 135, 5, 3, 93, 5, 144, 0, 0, 70, 5, 3, 93, 5, 144, 0, 0, 135, 5, 3, 93, 5, 156, 0, 0, 70, 5, 3, 93, 5, 156, 0, 0, 135, 5, 3, 93, 5, 196, 0, 0, 70, 5, 3, 93, 5, 196, 0, 0, 135, 5, 3, 93, 5, 150, 0, 0, 70, 5, 3, 93, 5, 150, 0, 0, 135, 5, 3, 93, 5, 160, 0, 0, 70, 5, 3, 93, 5, 160, 0, 0, 135, 5, 3, 93, 5, 206, 0, 0, 70, 5, 3, 93, 5, 206, 0, 0, 70, 5, 3, 93, 5, 210, 0, 0, 135, 5, 4, 93, 5, 142, 0, 0, 70, 5, 4, 93, 5, 142, 0, 0, 135, 5, 4, 93, 5, 144, 0, 0, 70, 5, 4, 93, 5, 144, 0, 0, 135, 5, 4, 93, 5, 156, 0, 0, 70, 5, 4, 93, 5, 156, 0, 0, 135, 5, 4, 93, 5, 196, 0, 0, 70, 5, 4, 93, 5, 196, 0, 0, 135, 5, 4, 93, 5, 150, 0, 0, 70, 5, 4, 93, 5, 150, 0, 0, 135, 5, 4, 93, 5, 160, 0, 0, 70, 5, 4, 93, 5, 160, 0, 0, 135, 5, 4, 93, 5, 206, 0, 0, 70, 5, 4, 93, 5, 206, 0, 0, 70, 5, 4, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 142, 0, 0, 5, 5, 2, 93, 5, 144, 0, 0, 5, 5, 2, 93, 5, 156, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 150, 0, 0, 5, 5, 2, 93, 5, 160, 0, 0, 5, 5, 2, 93, 5, 206, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 5, 5, 3, 93, 5, 142, 0, 0, 5, 5, 3, 93, 5, 144, 0, 0, 5, 5, 3, 93, 5, 156, 0, 0, 5, 5, 3, 93, 5, 196, 0, 0, 5, 5, 3, 93, 5, 150, 0, 0, 5, 5, 3, 93, 5, 160, 0, 0, 5, 5, 3, 93, 5, 206, 0, 0, 5, 5, 3, 93, 5, 210, 0, 0, 5, 5, 4, 93, 5, 142, 0, 0, 5, 5, 4, 93, 5, 144, 0, 0, 5, 5, 4, 93, 5, 156, 0, 0, 5, 5, 4, 93, 5, 196, 0, 0, 5, 5, 4, 93, 5, 150, 0, 0, 5, 5, 4, 93, 5, 160, 0, 0, 5, 5, 4, 93, 5, 206, 0, 0, 5, 5, 4, 93, 5, 210, 0, 0, 193, 2, 5, 93, 5, 136, 0, 0, 193, 2, 5, 93, 5, 150, 0, 0, 193, 2, 5, 93, 5, 142, 0, 0, 193, 2, 5, 93, 5, 164, 0, 0, 193, 2, 5, 93, 5, 156, 0, 0, 193, 2, 5, 93, 5, 146, 0, 0, 193, 2, 5, 93, 5, 138, 0, 0, 193, 2, 5, 93, 5, 196, 0, 0, 193, 2, 5, 93, 5, 182, 0, 0, 193, 2, 5, 93, 5, 154, 0, 0, 193, 3, 5, 93, 5, 142, 0, 0, 193, 3, 5, 93, 5, 144, 0, 0, 193, 3, 5, 93, 5, 156, 0, 0, 193, 3, 5, 93, 5, 196, 0, 0, 193, 3, 5, 93, 5, 150, 0, 0, 193, 3, 5, 93, 5, 160, 0, 0, 193, 3, 5, 93, 5, 206, 0, 0, 193, 3, 5, 93, 5, 210, 0, 0, 193, 4, 5, 93, 5, 142, 0, 0, 193, 4, 5, 93, 5, 144, 0, 0, 193, 4, 5, 93, 5, 156, 0, 0, 193, 4, 5, 93, 5, 196, 0, 0, 193, 4, 5, 93, 5, 150, 0, 0, 193, 4, 5, 93, 5, 160, 0, 0, 193, 4, 5, 93, 5, 206, 0, 0, 193, 4, 5, 93, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 2, 192, 197, 162, 41, 30, 2, 192, 197, 226, 42, 30, 3, 192, 197, 34, 43, 30, 3, 192, 197, 98, 150, 30, 3, 192, 197, 162, 36, 30, 1, 192, 197, 162, 37, 30, 1, 192, 197, 226, 38, 30, 2, 192, 197, 34, 39, 30, 2, 192, 197, 98, 30, 2, 12, 0, 30, 2, 0, 192, 197, 162, 31, 2, 0, 192, 197, 226, 34, 30, 1, 192, 197, 34, 35, 30, 1, 192, 197, 98, 72, 0, 255, 255, 2, 93, 135, 5, 104, 0, 255, 255, 2, 93, 70, 5, 36, 1, 197, 162, 37, 1, 0, 192, 197, 98, 0, 48, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 6, 192, 197, 98, 41, 30, 6, 192, 197, 162, 42, 30, 6, 192, 197, 226, 43, 30, 7, 192, 197, 34, 150, 30, 7, 192, 197, 98, 36, 30, 5, 192, 197, 98, 37, 30, 5, 192, 197, 162, 38, 30, 5, 192, 197, 226, 39, 30, 6, 192, 197, 34, 30, 2, 12, 0, 30, 2, 4, 192, 197, 98, 31, 2, 4, 192, 197, 162, 34, 30, 4, 192, 197, 226, 35, 30, 5, 192, 197, 34, 72, 0, 255, 255, 3, 93, 135, 5, 104, 0, 255, 255, 3, 93, 70, 5, 36, 1, 3, 192, 197, 226, 37, 1, 4, 192, 197, 34, 0, 62, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 10, 192, 197, 34, 41, 30, 10, 192, 197, 98, 42, 30, 10, 192, 197, 162, 43, 30, 10, 192, 197, 226, 150, 30, 11, 192, 197, 34, 36, 30, 9, 192, 197, 34, 37, 30, 9, 192, 197, 98, 38, 30, 9, 192, 197, 162, 39, 30, 9, 192, 197, 226, 30, 2, 12, 0, 30, 2, 8, 192, 197, 34, 31, 2, 8, 192, 197, 98, 34, 30, 8, 192, 197, 162, 35, 30, 8, 192, 197, 226, 72, 0, 255, 255, 4, 93, 135, 5, 104, 0, 255, 255, 4, 93, 70, 5, 36, 1, 7, 192, 197, 162, 37, 1, 7, 192, 197, 226, 0, 68, 156, 5, 20, 0, 142, 30, 37, 0, 244, 30, 20, 0, 247, 30, 9, 0, 247, 30, 4, 192, 198, 66, 248, 30, 4, 192, 198, 130, 249, 30, 4, 192, 198, 194, 244, 30, 3, 192, 198, 130, 245, 30, 3, 192, 198, 194, 246, 30, 4, 192, 198, 2, 142, 30, 2, 192, 198, 66, 143, 30, 2, 192, 198, 130, 153, 30, 2, 192, 198, 194, 242, 30, 3, 192, 198, 2, 243, 30, 3, 192, 198, 66, 118, 1, 15, 0, 118, 1, 1, 192, 198, 2, 119, 1, 1, 192, 198, 66, 120, 1, 1, 192, 198, 130, 50, 2, 1, 192, 198, 194, 51, 2, 2, 192, 198, 2, 89, 0, 198, 129, 121, 0, 198, 161, 221, 0, 0, 192, 198, 66, 253, 0, 0, 192, 198, 130, 255, 0, 0, 192, 198, 194, 0, 72, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 7, 192, 198, 194, 41, 30, 8, 192, 198, 2, 42, 30, 8, 192, 198, 66, 43, 30, 8, 192, 198, 130, 150, 30, 8, 192, 198, 194, 36, 30, 6, 192, 198, 194, 37, 30, 7, 192, 198, 2, 38, 30, 7, 192, 198, 66, 39, 30, 7, 192, 198, 130, 30, 2, 12, 0, 30, 2, 5, 192, 198, 194, 31, 2, 6, 192, 198, 2, 34, 30, 6, 192, 198, 66, 35, 30, 6, 192, 198, 130, 72, 0, 5, 192, 198, 1, 104, 0, 5, 192, 198, 33, 36, 1, 5, 192, 198, 66, 37, 1, 5, 192, 198, 130, 0, 78, 156, 5, 8, 0, 37, 30, 15, 0, 37, 30, 9, 192, 198, 226, 39, 30, 10, 192, 198, 34, 41, 30, 10, 192, 198, 98, 43, 30, 10, 192, 198, 162, 150, 30, 10, 192, 198, 226, 104, 0, 9, 192, 198, 1, 37, 1, 9, 192, 198, 34, 31, 2, 9, 192, 198, 98, 35, 30, 9, 192, 198, 162, 0, 46, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 12, 192, 197, 34, 39, 30, 12, 192, 197, 98, 41, 30, 12, 192, 197, 162, 43, 30, 12, 192, 197, 226, 150, 30, 13, 192, 197, 34, 104, 0, 255, 255, 2, 93, 5, 5, 37, 1, 11, 192, 197, 98, 31, 2, 11, 192, 197, 162, 35, 30, 11, 192, 197, 226, 0, 48, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 14, 192, 197, 34, 39, 30, 14, 192, 197, 98, 41, 30, 14, 192, 197, 162, 43, 30, 14, 192, 197, 226, 150, 30, 15, 192, 197, 34, 104, 0, 255, 255, 3, 93, 5, 5, 37, 1, 13, 192, 197, 98, 31, 2, 13, 192, 197, 162, 35, 30, 13, 192, 197, 226, 0, 62, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 16, 192, 197, 34, 39, 30, 16, 192, 197, 98, 41, 30, 16, 192, 197, 162, 43, 30, 16, 192, 197, 226, 150, 30, 17, 192, 197, 34, 104, 0, 255, 255, 4, 93, 5, 5, 37, 1, 15, 192, 197, 98, 31, 2, 15, 192, 197, 162, 35, 30, 15, 192, 197, 226, 0, 68, 5, 5, 10, 0, 143, 30, 20, 0, 245, 30, 9, 0, 245, 30, 19, 192, 197, 34, 247, 30, 19, 192, 197, 98, 249, 30, 19, 192, 197, 162, 143, 30, 18, 192, 197, 98, 153, 30, 18, 192, 197, 162, 243, 30, 18, 192, 197, 226, 121, 0, 255, 255, 5, 93, 193, 2, 253, 0, 17, 192, 197, 98, 255, 0, 17, 192, 197, 162, 119, 1, 17, 192, 197, 226, 51, 2, 18, 192, 197, 34, 0, 72, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 20, 192, 197, 162, 39, 30, 20, 192, 197, 226, 41, 30, 21, 192, 197, 34, 43, 30, 21, 192, 197, 98, 150, 30, 21, 192, 197, 162, 104, 0, 255, 255, 5, 93, 193, 3, 37, 1, 19, 192, 197, 226, 31, 2, 20, 192, 197, 34, 35, 30, 20, 192, 197, 98, 0, 78, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 22, 192, 197, 162, 39, 30, 22, 192, 197, 226, 41, 30, 23, 192, 197, 34, 43, 30, 23, 192, 197, 98, 150, 30, 23, 192, 197, 162, 104, 0, 255, 255, 5, 93, 193, 4, 37, 1, 21, 192, 197, 226, 31, 2, 22, 192, 197, 34, 35, 30, 22, 192, 197, 98, 32, 0, 72, 0, 73, 0, 89, 0, 90, 0, 104, 0, 105, 0, 121, 0, 122, 0, 221, 0, 222, 0, 253, 0, 254, 0, 255, 0, 0, 1, 36, 1, 38, 1, 118, 1, 121, 1, 30, 2, 32, 2, 50, 2, 52, 2, 34, 30, 44, 30, 142, 30, 144, 30, 150, 30, 151, 30, 153, 30, 154, 30, 242, 30, 250, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 36, 4, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 46, 4, 185, 104, 185, 108, 56, 4, 185, 120, 74, 4, 185, 128, 185, 136, 84, 4, 185, 144, 185, 152, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 90, 4, 96, 4, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 102, 4, 168, 104, 168, 108, 108, 4, 168, 120, 118, 4, 168, 128, 168, 136, 124, 4, 168, 144, 168, 152, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 154, 114, 185, 121, 153, 121, 249, 121, 153, 122, 89, 122, 136, 14, 217, 122, 185, 153, 153, 153, 249, 153, 89, 154, 153, 169, 185, 200, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 136, 121, 232, 121, 136, 122, 72, 122, 128, 14, 200, 122, 168, 153, 136, 153, 232, 153, 72, 154, 136, 169, 168, 200, 72, 170, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 250, 114, 232, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 57, 123, 40, 123, 217, 121, 200, 121, 121, 122, 104, 122, 12, 8, 14, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 250, 142, 232, 142, 26, 142, 8, 142, 249, 146, 232, 146, 25, 146, 8, 146, 185, 148, 168, 148, 153, 154, 136, 154, 57, 155, 40, 155, 217, 153, 200, 153, 57, 154, 40, 154, 121, 154, 104, 154, 25, 155, 8, 155, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 185, 174, 168, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 204, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 72, 4, 186, 176, 104, 4, 177, 176, 36, 5, 250, 177, 37, 5, 241, 177, 255, 5, 185, 68, 72, 4, 186, 180, 104, 4, 177, 180, 36, 5, 250, 181, 37, 5, 241, 181, 255, 5, 185, 100, 72, 4, 186, 184, 104, 4, 177, 184, 36, 5, 250, 185, 37, 5, 241, 185, 255, 5, 186, 112, 89, 4, 186, 188, 121, 4, 177, 188, 221, 4, 154, 189, 253, 4, 145, 189, 255, 4, 81, 190, 118, 5, 250, 189, 119, 5, 241, 189, 120, 5, 90, 190, 255, 5, 185, 124, 72, 4, 186, 192, 104, 4, 177, 192, 36, 5, 250, 193, 37, 5, 241, 193, 255, 5, 186, 140, 104, 4, 177, 196, 37, 5, 241, 197, 255, 5, 168, 64, 104, 4, 168, 176, 37, 5, 232, 177, 255, 5, 168, 68, 104, 4, 168, 180, 37, 5, 232, 181, 255, 5, 168, 100, 104, 4, 168, 184, 37, 5, 232, 185, 255, 5, 168, 112, 121, 4, 168, 188, 253, 4, 136, 189, 255, 4, 72, 190, 119, 5, 232, 189, 255, 5, 168, 124, 104, 4, 168, 192, 37, 5, 232, 193, 255, 5, 168, 140, 104, 4, 168, 196, 37, 5, 232, 197, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 63, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 3, 10, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 8, 10, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 208, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 212, 6, 0, 0, 212, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 73, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 96, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 20, 11, 60, 0, 1, 11, 60, 0, 2, 11, 60, 0, 3, 11, 38, 0, 57, 11, 60, 0, 21, 11, 77, 11, 55, 11, 38, 0, 47, 11, 60, 0, 60, 0, 95, 11, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 242, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 128, 0, 0, 0, 128, 1, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 196, 25, 0, 0, 196, 25, 0, 0, 212, 25, 0, 0, 218, 25, 0, 0, 31, 0, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 108, 187, 255, 0, 109, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 41, 42, 43, 39, 44, 45, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 119, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 48, 0, 102, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 110, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 2, 33, 108, 193, 3, 33, 108, 193, 4, 33, 108, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 2, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 84, 108, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 34, 108, 5, 5, 49, 0, 77, 11, 55, 11, 255, 255, 105, 108, 5, 5, 2, 0, 55, 11, 56, 11, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 75, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 201, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 206, 6, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14, 7, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 18, 7, 0, 0, 18, 7, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 75, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 100, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 32, 0, 65, 0, 114, 0, 97, 0, 98, 0, 93, 0, 38, 0, 113, 10, 60, 0, 60, 0, 112, 10, 60, 0, 60, 0, 2, 10, 60, 0, 60, 0, 1, 10, 60, 0, 60, 0, 60, 10, 38, 0, 92, 10, 60, 0, 77, 10, 0, 0, 236, 26, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 128, 0, 0, 0, 128, 1, 0, 0, 96, 26, 0, 0, 96, 26, 0, 0, 96, 26, 0, 0, 96, 26, 0, 0, 212, 26, 0, 0, 16, 0, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 41, 9, 0, 0, 96, 7, 0, 0, 102, 203, 255, 0, 103, 6, 0, 0, 104, 192, 255, 0, 106, 189, 255, 0, 107, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 103, 104, 105, 106, 107, 108, 49, 109, 40, 41, 39, 42, 43, 44, 45, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 133, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 100, 0, 108, 0, 116, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 124, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 2, 0, 202, 129, 2, 0, 202, 162, 2, 0, 202, 195, 2, 0, 202, 228, 2, 0, 202, 5, 3, 0, 202, 38, 3, 0, 202, 71, 3, 0, 202, 104, 3, 0, 202, 137, 3, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 1, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 1, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 2, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 122, 241, 194, 5, 121, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 123, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 106, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 0, 0, 197, 226, 0, 0, 197, 34, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 120, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 94, 106, 194, 5, 123, 241, 5, 5, 32, 106, 194, 5, 123, 241, 5, 5, 40, 106, 194, 5, 123, 241, 5, 5, 42, 106, 194, 5, 123, 241, 5, 5, 52, 106, 194, 5, 123, 241, 5, 5, 80, 106, 194, 5, 123, 241, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 69, 0, 60, 0, 32, 0, 75, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 7, 7, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 12, 7, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 112, 0, 97, 0, 95, 0, 71, 0, 117, 0, 114, 0, 117, 0, 95, 0, 73, 0, 78, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 251, 9, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 255, 9, 0, 0, 255, 9, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 60, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 70, 220, 38, 0, 65, 0, 60, 0, 5, 1, 60, 0, 60, 0, 60, 0, 4, 1, 38, 0, 67, 0, 60, 0, 7, 1, 60, 0, 60, 0, 60, 0, 6, 1, 38, 0, 69, 0, 60, 0, 25, 1, 60, 0, 60, 0, 60, 0, 24, 1, 38, 0, 76, 0, 60, 0, 66, 1, 60, 0, 60, 0, 60, 0, 65, 1, 38, 0, 78, 0, 60, 0, 68, 1, 60, 0, 60, 0, 60, 0, 67, 1, 38, 0, 79, 0, 60, 0, 243, 0, 60, 0, 60, 0, 60, 0, 211, 0, 38, 0, 83, 0, 60, 0, 91, 1, 60, 0, 60, 0, 60, 0, 90, 1, 38, 0, 90, 0, 60, 0, 122, 1, 60, 0, 60, 0, 60, 0, 121, 1, 60, 0, 124, 1, 60, 0, 60, 0, 60, 0, 123, 1, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 208, 38, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 20, 30, 0, 0, 20, 30, 0, 0, 20, 30, 0, 0, 20, 30, 0, 0, 136, 31, 0, 0, 136, 31, 0, 0, 168, 34, 0, 0, 168, 34, 0, 0, 184, 38, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 212, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 121, 0, 129, 0, 132, 0, 140, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 162, 0, 48, 0, 169, 0, 177, 0, 181, 0, 48, 0, 187, 0, 195, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 203, 0, 156, 0, 48, 0, 48, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 1, 228, 1, 16, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 10, 0, 202, 129, 10, 0, 202, 162, 10, 0, 202, 195, 10, 0, 202, 228, 10, 0, 202, 5, 11, 0, 202, 38, 11, 0, 202, 71, 11, 0, 202, 104, 11, 0, 202, 137, 11, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 201, 198, 0, 0, 156, 5, 0, 48, 201, 230, 1, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 166, 2, 0, 201, 6, 4, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 102, 5, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 198, 6, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 166, 8, 0, 5, 5, 0, 44, 201, 102, 9, 0, 5, 5, 0, 48, 201, 134, 10, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 70, 11, 0, 201, 166, 12, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 6, 14, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 102, 15, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 9, 0, 201, 70, 17, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 134, 5, 2, 71, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 9, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 10, 0, 201, 38, 18, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 5, 5, 2, 71, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 43, 5, 5, 2, 43, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 51, 5, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 65, 5, 5, 2, 65, 134, 5, 2, 69, 5, 5, 2, 69, 201, 70, 19, 0, 201, 102, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 21, 0, 201, 166, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 134, 5, 3, 93, 5, 5, 3, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 23, 0, 201, 230, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 26, 0, 201, 38, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 28, 0, 201, 102, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 30, 0, 201, 166, 31, 0, 201, 198, 32, 0, 201, 230, 33, 0, 201, 6, 35, 0, 201, 38, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 37, 0, 201, 102, 38, 0, 197, 34, 5, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 39, 0, 201, 6, 41, 0, 201, 134, 42, 0, 201, 6, 44, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 45, 0, 201, 166, 46, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 1, 0, 197, 226, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 47, 0, 201, 230, 48, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 134, 5, 3, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 5, 5, 3, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 134, 5, 3, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 5, 5, 3, 93, 5, 210, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 48, 0, 40, 3, 2, 235, 134, 5, 0, 46, 156, 5, 1, 0, 1, 3, 2, 239, 134, 5, 65, 3, 2, 239, 134, 5, 0, 50, 156, 5, 48, 0, 40, 3, 2, 243, 134, 5, 0, 68, 156, 5, 1, 0, 1, 3, 255, 255, 2, 69, 134, 5, 65, 3, 255, 255, 2, 69, 134, 5, 0, 70, 156, 5, 1, 0, 1, 3, 255, 255, 2, 71, 134, 5, 65, 3, 255, 255, 2, 71, 134, 5, 0, 78, 156, 5, 1, 0, 1, 3, 255, 255, 2, 79, 134, 5, 65, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 2, 0, 1, 3, 255, 255, 2, 93, 134, 5, 7, 3, 255, 255, 3, 93, 134, 5, 65, 3, 255, 255, 2, 93, 134, 5, 0, 42, 5, 5, 48, 0, 40, 3, 2, 235, 5, 5, 0, 46, 5, 5, 1, 0, 1, 3, 2, 239, 5, 5, 65, 3, 2, 239, 5, 5, 0, 50, 5, 5, 48, 0, 40, 3, 2, 243, 5, 5, 0, 68, 5, 5, 1, 0, 1, 3, 255, 255, 2, 69, 5, 5, 65, 3, 255, 255, 2, 69, 5, 5, 0, 70, 5, 5, 1, 0, 1, 3, 255, 255, 2, 71, 5, 5, 65, 3, 255, 255, 2, 71, 5, 5, 0, 78, 5, 5, 1, 0, 1, 3, 255, 255, 2, 79, 5, 5, 65, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 2, 0, 1, 3, 255, 255, 2, 93, 5, 5, 7, 3, 255, 255, 3, 93, 5, 5, 65, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 1, 0, 1, 3, 197, 162, 65, 3, 197, 162, 5, 46, 196, 160, 1, 0, 1, 3, 0, 192, 197, 98, 65, 3, 0, 192, 197, 98, 156, 68, 196, 160, 1, 0, 1, 3, 0, 192, 197, 162, 65, 3, 0, 192, 197, 162, 5, 68, 196, 160, 1, 0, 1, 3, 0, 192, 197, 226, 65, 3, 0, 192, 197, 226, 156, 78, 196, 160, 1, 0, 1, 3, 1, 192, 197, 34, 65, 3, 1, 192, 197, 34, 5, 78, 196, 160, 1, 0, 1, 3, 1, 192, 197, 98, 65, 3, 1, 192, 197, 98, 156, 70, 196, 190, 1, 0, 1, 3, 1, 192, 197, 162, 65, 3, 1, 192, 197, 162, 5, 70, 196, 190, 1, 0, 1, 3, 1, 192, 197, 226, 65, 3, 1, 192, 197, 226, 156, 70, 196, 162, 1, 0, 1, 3, 2, 192, 197, 34, 65, 3, 2, 192, 197, 34, 5, 70, 196, 162, 1, 0, 1, 3, 2, 192, 197, 98, 65, 3, 2, 192, 197, 98, 156, 78, 196, 202, 1, 0, 1, 3, 2, 192, 197, 162, 65, 3, 2, 192, 197, 162, 5, 78, 196, 202, 1, 0, 1, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 226, 156, 68, 196, 196, 1, 0, 1, 3, 3, 192, 197, 34, 65, 3, 3, 192, 197, 34, 5, 68, 196, 196, 1, 0, 1, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 98, 156, 68, 196, 210, 1, 0, 1, 3, 3, 192, 197, 162, 65, 3, 3, 192, 197, 162, 5, 68, 196, 210, 1, 0, 1, 3, 3, 192, 197, 226, 65, 3, 3, 192, 197, 226, 156, 68, 196, 204, 1, 0, 1, 3, 4, 192, 197, 34, 65, 3, 4, 192, 197, 34, 5, 68, 196, 204, 1, 0, 1, 3, 4, 192, 197, 98, 65, 3, 4, 192, 197, 98, 156, 78, 196, 196, 1, 0, 1, 3, 4, 192, 197, 162, 65, 3, 4, 192, 197, 162, 5, 78, 196, 196, 1, 0, 1, 3, 4, 192, 197, 226, 65, 3, 4, 192, 197, 226, 156, 92, 196, 196, 2, 0, 1, 3, 5, 192, 197, 162, 7, 3, 5, 192, 197, 226, 65, 3, 5, 192, 197, 162, 5, 92, 196, 196, 2, 0, 1, 3, 6, 192, 197, 34, 7, 3, 6, 192, 197, 98, 65, 3, 6, 192, 197, 34, 156, 92, 196, 210, 2, 0, 1, 3, 6, 192, 197, 162, 7, 3, 6, 192, 197, 226, 65, 3, 6, 192, 197, 162, 5, 92, 196, 210, 2, 0, 1, 3, 7, 192, 197, 34, 7, 3, 7, 192, 197, 98, 65, 3, 7, 192, 197, 34, 156, 70, 196, 196, 1, 0, 1, 3, 7, 192, 197, 162, 65, 3, 7, 192, 197, 162, 5, 70, 196, 196, 1, 0, 1, 3, 7, 192, 197, 226, 65, 3, 7, 192, 197, 226, 8, 0, 197, 35, 1, 0, 1, 3, 8, 192, 197, 131, 65, 3, 8, 192, 197, 131, 8, 0, 197, 227, 1, 0, 1, 3, 9, 192, 197, 67, 65, 3, 9, 192, 197, 67, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 64, 28, 4, 185, 76, 30, 4, 185, 88, 185, 92, 185, 96, 186, 100, 186, 108, 185, 112, 185, 116, 185, 124, 32, 4, 34, 4, 185, 148, 185, 152, 185, 160, 36, 4, 185, 172, 185, 180, 185, 184, 185, 188, 185, 192, 185, 196, 38, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 40, 4, 168, 64, 42, 4, 168, 76, 44, 4, 168, 88, 168, 92, 168, 96, 168, 100, 168, 108, 168, 112, 168, 116, 168, 124, 46, 4, 48, 4, 168, 148, 168, 152, 168, 160, 50, 4, 168, 172, 168, 180, 168, 184, 168, 188, 168, 192, 168, 196, 52, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 140, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 54, 4, 186, 81, 154, 81, 250, 81, 90, 82, 186, 101, 154, 101, 250, 101, 90, 102, 217, 76, 154, 130, 185, 141, 185, 144, 249, 141, 153, 142, 89, 142, 136, 14, 217, 142, 185, 181, 153, 181, 249, 181, 89, 182, 153, 197, 185, 212, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 56, 4, 168, 81, 136, 81, 232, 81, 72, 82, 168, 101, 136, 101, 232, 101, 72, 102, 200, 76, 136, 130, 168, 141, 168, 144, 232, 141, 136, 142, 72, 142, 128, 14, 200, 142, 168, 181, 136, 181, 232, 181, 72, 182, 136, 197, 168, 212, 72, 198, 57, 59, 40, 59, 217, 57, 200, 57, 185, 60, 168, 60, 185, 72, 168, 72, 249, 69, 232, 69, 185, 70, 168, 70, 25, 70, 8, 70, 25, 78, 8, 78, 89, 79, 72, 79, 58, 83, 40, 83, 218, 81, 200, 81, 186, 82, 168, 82, 185, 84, 168, 84, 26, 82, 8, 82, 249, 93, 232, 93, 217, 93, 200, 93, 185, 94, 168, 94, 249, 94, 232, 94, 249, 97, 232, 97, 89, 99, 72, 99, 154, 102, 136, 102, 58, 103, 40, 103, 218, 101, 200, 101, 26, 103, 8, 103, 186, 102, 168, 104, 6, 8, 8, 8, 250, 109, 232, 109, 249, 114, 232, 114, 168, 156, 153, 117, 136, 117, 249, 118, 232, 118, 25, 118, 8, 118, 121, 119, 104, 119, 185, 120, 168, 120, 185, 132, 168, 132, 58, 4, 60, 4, 26, 130, 8, 130, 10, 8, 185, 136, 168, 136, 57, 143, 40, 143, 217, 141, 200, 141, 121, 142, 104, 142, 12, 8, 14, 8, 153, 161, 136, 161, 249, 162, 232, 162, 25, 162, 8, 162, 185, 168, 168, 168, 250, 165, 232, 165, 62, 4, 64, 4, 26, 166, 8, 166, 249, 174, 232, 174, 25, 174, 8, 174, 185, 176, 168, 176, 153, 182, 136, 182, 57, 183, 40, 183, 217, 181, 200, 181, 57, 182, 40, 182, 121, 182, 104, 182, 25, 183, 8, 183, 249, 189, 232, 189, 249, 197, 232, 197, 89, 198, 185, 204, 168, 204, 185, 208, 168, 208, 25, 202, 8, 202, 232, 164, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 80, 200, 164, 169, 164, 200, 56, 169, 80, 187, 100, 187, 108, 169, 100, 169, 108, 169, 216, 169, 128, 217, 140, 187, 80, 200, 140, 169, 80, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 68, 255, 5, 186, 80, 255, 5, 186, 128, 255, 5, 185, 140, 255, 5, 186, 164, 255, 5, 185, 200, 255, 5, 168, 56, 255, 5, 168, 68, 255, 5, 168, 80, 255, 5, 168, 128, 255, 5, 168, 140, 255, 5, 168, 164, 255, 5, 168, 200, 255, 5, 249, 70, 255, 5, 232, 70, 255, 5, 250, 130, 255, 5, 232, 130, 255, 5, 250, 166, 255, 5, 232, 166, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 63, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 244, 9, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 249, 9, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 224, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 228, 8, 0, 0, 228, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 124, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 197, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 65, 0, 114, 0, 97, 0, 98, 0, 93, 0, 38, 0, 78, 6, 60, 0, 60, 0, 80, 6, 60, 0, 60, 0, 79, 6, 60, 0, 60, 0, 75, 6, 60, 0, 60, 0, 77, 6, 60, 0, 60, 0, 76, 6, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 39, 6, 60, 0, 34, 6, 38, 0, 39, 6, 60, 0, 60, 0, 35, 6, 60, 0, 60, 0, 114, 6, 60, 0, 60, 0, 113, 6, 60, 0, 60, 0, 37, 6, 60, 0, 60, 0, 115, 6, 60, 0, 33, 6, 38, 0, 42, 6, 60, 0, 124, 6, 60, 0, 60, 0, 121, 6, 38, 0, 44, 6, 60, 0, 129, 6, 60, 0, 60, 0, 45, 6, 84, 6, 38, 0, 134, 6, 60, 0, 133, 6, 38, 0, 47, 6, 60, 0, 137, 6, 60, 0, 60, 0, 136, 6, 38, 0, 49, 6, 60, 0, 147, 6, 60, 0, 60, 0, 145, 6, 38, 0, 152, 6, 60, 0, 150, 6, 38, 0, 169, 6, 60, 0, 60, 0, 42, 0, 170, 6, 67, 6, 38, 0, 171, 6, 60, 0, 60, 0, 175, 6, 38, 0, 188, 6, 60, 0, 60, 0, 187, 6, 38, 0, 72, 6, 60, 0, 60, 0, 36, 6, 60, 0, 60, 0, 199, 6, 60, 0, 60, 0, 201, 6, 38, 0, 71, 6, 60, 0, 60, 0, 60, 0, 192, 6, 60, 0, 60, 0, 60, 0, 71, 6, 84, 6, 60, 0, 60, 0, 42, 0, 213, 6, 193, 6, 190, 6, 41, 6, 195, 6, 38, 0, 204, 6, 60, 0, 60, 0, 42, 0, 73, 6, 210, 6, 74, 6, 208, 6, 205, 6, 60, 0, 60, 0, 204, 6, 84, 6, 60, 0, 60, 0, 60, 0, 73, 6, 84, 6, 60, 0, 60, 0, 60, 0, 38, 6, 38, 0, 160, 0, 60, 0, 60, 0, 12, 32, 60, 0, 60, 0, 13, 32, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 100, 33, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 72, 0, 0, 0, 72, 1, 0, 0, 184, 28, 0, 0, 184, 28, 0, 0, 184, 28, 0, 0, 184, 28, 0, 0, 12, 29, 0, 0, 12, 29, 0, 0, 140, 29, 0, 0, 140, 29, 0, 0, 76, 33, 0, 0, 2, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 39, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 174, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 128, 0, 136, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 157, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 152, 0, 48, 0, 48, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 128, 1, 192, 1, 0, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 1, 0, 202, 129, 1, 0, 202, 162, 1, 0, 202, 195, 1, 0, 202, 228, 1, 0, 202, 5, 2, 0, 202, 38, 2, 0, 202, 71, 2, 0, 202, 104, 2, 0, 202, 137, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 102, 5, 5, 25, 102, 5, 70, 26, 102, 5, 70, 222, 102, 5, 73, 26, 102, 7, 75, 240, 102, 201, 6, 0, 0, 192, 0, 0, 0, 5, 73, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 71, 184, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 2, 0, 201, 166, 3, 0, 201, 134, 4, 0, 201, 102, 5, 0, 194, 5, 210, 230, 194, 5, 212, 230, 194, 5, 211, 230, 192, 0, 0, 0, 194, 5, 209, 230, 194, 5, 208, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 72, 26, 102, 5, 71, 26, 102, 5, 74, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 43, 102, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 102, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 85, 102, 5, 5, 85, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 111, 102, 192, 0, 0, 0, 5, 5, 111, 102, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 184, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 188, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 210, 102, 192, 0, 0, 0, 192, 0, 0, 0, 5, 72, 214, 102, 192, 0, 0, 0, 6, 5, 214, 102, 5, 71, 214, 102, 197, 34, 0, 0, 5, 74, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 71, 222, 102, 192, 0, 0, 0, 5, 72, 222, 102, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 6, 0, 5, 74, 240, 102, 192, 0, 0, 0, 192, 0, 0, 0, 5, 73, 240, 102, 192, 0, 0, 0, 5, 71, 240, 102, 197, 98, 0, 0, 192, 0, 0, 0, 201, 38, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 0, 4, 5, 71, 0, 4, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 71, 214, 102, 194, 5, 128, 232, 5, 71, 240, 102, 194, 5, 128, 232, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 25, 102, 5, 5, 84, 6, 255, 255, 26, 102, 5, 70, 85, 6, 255, 255, 26, 102, 5, 73, 72, 102, 5, 5, 48, 0, 84, 6, 255, 255, 59, 102, 5, 70, 214, 102, 5, 5, 48, 0, 84, 6, 255, 255, 214, 102, 7, 5, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 240, 102, 5, 70, 48, 0, 84, 6, 255, 255, 240, 102, 6, 75, 240, 102, 5, 72, 48, 0, 84, 6, 255, 255, 240, 102, 7, 75, 240, 102, 5, 5, 48, 0, 84, 6, 255, 255, 240, 102, 5, 75, 214, 102, 5, 70, 48, 0, 84, 6, 255, 255, 214, 102, 6, 5, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 185, 120, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 185, 152, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 168, 120, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 168, 152, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 154, 114, 185, 121, 153, 121, 249, 121, 153, 122, 89, 122, 136, 14, 217, 122, 185, 153, 153, 153, 249, 153, 89, 154, 153, 169, 185, 176, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 136, 121, 232, 121, 136, 122, 72, 122, 128, 14, 200, 122, 168, 153, 136, 153, 232, 153, 72, 154, 136, 169, 168, 176, 72, 170, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 250, 114, 232, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 57, 123, 40, 123, 217, 121, 200, 121, 121, 122, 104, 122, 12, 8, 14, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 250, 142, 232, 142, 26, 142, 8, 142, 249, 146, 232, 146, 25, 146, 8, 146, 185, 148, 168, 148, 153, 154, 136, 154, 57, 155, 40, 155, 217, 153, 200, 153, 57, 154, 40, 154, 121, 154, 104, 154, 25, 155, 8, 155, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 185, 174, 168, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 127, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 217, 8, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 222, 8, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 86, 1, 0, 32, 7, 0, 0, 0, 184, 0, 0, 0, 90, 1, 0, 0, 90, 1, 0, 0, 132, 0, 0, 0, 1, 0, 0, 0, 86, 1, 0, 0, 73, 110, 115, 116, 97, 108, 108, 101, 100, 76, 111, 99, 97, 108, 101, 115, 0, 97, 102, 0, 97, 109, 0, 97, 114, 0, 97, 115, 0, 97, 122, 0, 98, 101, 0, 98, 103, 0, 98, 110, 0, 98, 111, 0, 98, 115, 0, 98, 115, 95, 67, 121, 114, 108, 0, 99, 97, 0, 99, 101, 98, 0, 99, 104, 114, 0, 99, 115, 0, 99, 121, 0, 100, 97, 0, 100, 101, 0, 100, 101, 95, 65, 84, 0, 100, 115, 98, 0, 100, 122, 0, 101, 101, 0, 101, 108, 0, 101, 110, 0, 101, 110, 95, 85, 83, 0, 101, 110, 95, 85, 83, 95, 80, 79, 83, 73, 88, 0, 101, 111, 0, 101, 116, 0, 102, 97, 0, 102, 97, 95, 65, 70, 0, 102, 105, 0, 102, 105, 108, 0, 102, 111, 0, 102, 114, 0, 102, 114, 95, 67, 65, 0, 103, 97, 0, 103, 108, 0, 103, 117, 0, 104, 97, 0, 104, 97, 119, 0, 104, 101, 0, 104, 101, 95, 73, 76, 0, 104, 105, 0, 104, 115, 98, 0, 104, 117, 0, 104, 121, 0, 105, 100, 0, 105, 100, 95, 73, 68, 0, 105, 103, 0, 105, 115, 0, 105, 116, 0, 106, 97, 0, 107, 97, 0, 107, 107, 0, 107, 108, 0, 107, 109, 0, 107, 110, 0, 107, 111, 0, 107, 111, 107, 0, 107, 117, 0, 107, 121, 0, 108, 98, 0, 108, 107, 116, 0, 108, 110, 0, 108, 111, 0, 108, 116, 0, 108, 118, 0, 109, 107, 0, 109, 108, 0, 109, 114, 0, 109, 115, 0, 109, 116, 0, 109, 121, 0, 110, 98, 0, 110, 98, 95, 78, 79, 0, 110, 101, 0, 110, 108, 0, 110, 110, 0, 111, 109, 0, 111, 114, 0, 112, 97, 0, 112, 97, 95, 71, 117, 114, 117, 0, 112, 97, 95, 71, 117, 114, 117, 95, 73, 78, 0, 112, 108, 0, 112, 115, 0, 112, 116, 0, 114, 111, 0, 115, 101, 0, 115, 105, 0, 115, 107, 0, 115, 108, 0, 115, 109, 110, 0, 115, 113, 0, 115, 114, 0, 115, 114, 95, 67, 121, 114, 108, 0, 115, 114, 95, 67, 121, 114, 108, 95, 66, 65, 0, 115, 114, 95, 67, 121, 114, 108, 95, 77, 69, 0, 115, 114, 95, 67, 121, 114, 108, 95, 82, 83, 0, 115, 114, 95, 76, 97, 116, 110, 0, 115, 114, 95, 76, 97, 116, 110, 95, 66, 65, 0, 115, 114, 95, 76, 97, 116, 110, 95, 82, 83, 0, 115, 118, 0, 115, 119, 0, 116, 97, 0, 116, 101, 0, 116, 104, 0, 116, 107, 0, 116, 111, 0, 116, 114, 0, 117, 103, 0, 117, 107, 0, 117, 114, 0, 117, 122, 0, 118, 105, 0, 119, 97, 101, 0, 119, 111, 0, 120, 104, 0, 121, 105, 0, 121, 111, 0, 122, 104, 0, 122, 104, 95, 72, 97, 110, 115, 0, 122, 104, 95, 72, 97, 110, 115, 95, 67, 78, 0, 122, 104, 95, 72, 97, 110, 115, 95, 83, 71, 0, 122, 104, 95, 72, 97, 110, 116, 0, 122, 104, 95, 72, 97, 110, 116, 95, 72, 75, 0, 122, 104, 95, 72, 97, 110, 116, 95, 77, 79, 0, 122, 104, 95, 72, 97, 110, 116, 95, 84, 87, 0, 122, 117, 0, 65, 108, 105, 97, 115, 76, 111, 99, 97, 108, 101, 115, 0, 97, 114, 115, 0, 105, 110, 0, 105, 110, 95, 73, 68, 0, 105, 119, 0, 105, 119, 95, 73, 76, 0, 109, 111, 0, 110, 111, 0, 110, 111, 95, 78, 79, 0, 112, 97, 95, 73, 78, 0, 115, 104, 0, 115, 104, 95, 66, 65, 0, 115, 104, 95, 67, 83, 0, 115, 104, 95, 89, 85, 0, 115, 114, 95, 66, 65, 0, 115, 114, 95, 77, 69, 0, 115, 114, 95, 82, 83, 0, 121, 117, 101, 0, 121, 117, 101, 95, 67, 78, 0, 121, 117, 101, 95, 72, 97, 110, 115, 0, 122, 104, 95, 67, 78, 0, 122, 104, 95, 72, 75, 0, 122, 104, 95, 77, 79, 0, 122, 104, 95, 83, 71, 0, 122, 104, 95, 84, 87, 0, 170, 170, 170, 0, 0, 24, 0, 92, 2, 96, 2, 99, 2, 105, 2, 108, 2, 114, 2, 117, 2, 120, 2, 126, 2, 132, 2, 135, 2, 141, 2, 147, 2, 153, 2, 159, 2, 165, 2, 171, 2, 175, 2, 182, 2, 191, 2, 197, 2, 203, 2, 209, 2, 215, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 132, 0, 49, 0, 52, 0, 55, 0, 58, 0, 61, 0, 64, 0, 67, 0, 70, 0, 73, 0, 76, 0, 79, 0, 87, 0, 90, 0, 94, 0, 98, 0, 101, 0, 104, 0, 107, 0, 110, 0, 116, 0, 120, 0, 123, 0, 126, 0, 129, 0, 132, 0, 138, 0, 150, 0, 89, 2, 153, 0, 156, 0, 159, 0, 165, 0, 168, 0, 172, 0, 175, 0, 178, 0, 184, 0, 187, 0, 190, 0, 193, 0, 196, 0, 200, 0, 203, 0, 209, 0, 95, 0, 212, 0, 216, 0, 219, 0, 222, 0, 225, 0, 231, 0, 234, 0, 237, 0, 240, 0, 243, 0, 246, 0, 249, 0, 252, 0, 255, 0, 2, 1, 5, 1, 9, 1, 12, 1, 15, 1, 18, 1, 22, 1, 25, 1, 28, 1, 31, 1, 34, 1, 37, 1, 123, 1, 40, 1, 43, 1, 46, 1, 49, 1, 52, 1, 55, 1, 61, 1, 64, 1, 67, 1, 70, 1, 73, 1, 76, 1, 79, 1, 87, 1, 98, 1, 101, 1, 104, 1, 107, 1, 84, 1, 110, 1, 113, 1, 116, 1, 119, 1, 122, 1, 126, 1, 129, 1, 132, 1, 140, 1, 151, 1, 162, 1, 173, 1, 181, 1, 192, 1, 203, 1, 206, 1, 209, 1, 212, 1, 215, 1, 218, 1, 221, 1, 224, 1, 227, 1, 230, 1, 233, 1, 236, 1, 239, 1, 242, 1, 246, 1, 249, 1, 252, 1, 255, 1, 2, 2, 5, 2, 13, 2, 24, 2, 35, 2, 43, 2, 54, 2, 65, 2, 76, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 170, 170, 2, 0, 79, 2, 32, 0, 170, 170, 1, 0, 0, 80, 50, 0, 0, 80, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 197, 8, 0, 0, 197, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 48, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 46, 220, 38, 0, 65, 0, 60, 0, 3, 1, 60, 0, 60, 0, 60, 0, 2, 1, 60, 0, 226, 0, 60, 0, 60, 0, 60, 0, 194, 0, 38, 0, 73, 0, 60, 0, 238, 0, 60, 0, 60, 0, 60, 0, 206, 0, 38, 0, 83, 0, 60, 0, 95, 1, 61, 0, 25, 2, 60, 0, 60, 0, 60, 0, 94, 1, 61, 0, 24, 2, 38, 0, 84, 0, 60, 0, 99, 1, 61, 0, 27, 2, 60, 0, 60, 0, 60, 0, 98, 1, 61, 0, 26, 2, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 22, 34, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 116, 27, 0, 0, 116, 27, 0, 0, 116, 27, 0, 0, 116, 27, 0, 0, 200, 28, 0, 0, 200, 28, 0, 0, 14, 30, 0, 0, 14, 30, 0, 0, 254, 33, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 170, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 48, 0, 56, 0, 48, 0, 56, 0, 102, 0, 110, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 134, 0, 139, 0, 48, 0, 48, 0, 48, 0, 147, 0, 153, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 161, 0, 126, 0, 48, 0, 48, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 118, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 192, 0, 192, 0, 152, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 9, 0, 202, 129, 9, 0, 202, 162, 9, 0, 202, 195, 9, 0, 202, 228, 9, 0, 202, 5, 10, 0, 202, 38, 10, 0, 202, 71, 10, 0, 202, 104, 10, 0, 202, 137, 10, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 38, 1, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 230, 1, 0, 201, 70, 3, 0, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 166, 4, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 198, 5, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 134, 6, 0, 201, 230, 7, 0, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 134, 5, 3, 43, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 8, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 134, 5, 2, 59, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 8, 0, 196, 138, 5, 42, 196, 136, 5, 42, 5, 5, 3, 43, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 9, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 5, 5, 2, 59, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 43, 5, 5, 2, 43, 201, 70, 9, 0, 201, 70, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 11, 0, 201, 38, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 81, 5, 5, 2, 81, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 134, 5, 2, 81, 5, 5, 2, 81, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 12, 0, 201, 6, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 15, 0, 201, 230, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 16, 0, 201, 198, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 4, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 226, 4, 0, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 3, 0, 197, 162, 3, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 197, 226, 6, 0, 197, 34, 7, 0, 197, 98, 7, 0, 197, 162, 7, 0, 197, 226, 7, 0, 197, 98, 3, 0, 197, 226, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 18, 0, 201, 166, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 3, 43, 5, 162, 0, 0, 134, 5, 2, 43, 5, 162, 0, 0, 5, 5, 3, 43, 5, 162, 0, 0, 5, 5, 2, 43, 5, 162, 0, 0, 134, 5, 2, 59, 5, 162, 0, 0, 5, 5, 2, 59, 5, 162, 0, 0, 134, 5, 3, 43, 5, 200, 0, 0, 134, 5, 2, 43, 5, 200, 0, 0, 5, 5, 3, 43, 5, 200, 0, 0, 5, 5, 2, 43, 5, 200, 0, 0, 134, 5, 2, 59, 5, 208, 0, 0, 5, 5, 2, 59, 5, 208, 0, 0, 134, 5, 3, 43, 5, 196, 0, 0, 134, 5, 2, 43, 5, 196, 0, 0, 5, 5, 3, 43, 5, 196, 0, 0, 5, 5, 2, 43, 5, 196, 0, 0, 134, 5, 3, 43, 5, 136, 0, 0, 5, 5, 3, 43, 5, 136, 0, 0, 134, 5, 3, 43, 5, 138, 0, 0, 5, 5, 3, 43, 5, 138, 0, 0, 134, 5, 3, 43, 5, 182, 0, 0, 5, 5, 3, 43, 5, 182, 0, 0, 134, 5, 3, 43, 5, 154, 0, 0, 5, 5, 3, 43, 5, 154, 0, 0, 134, 5, 2, 43, 5, 136, 0, 0, 5, 5, 2, 43, 5, 136, 0, 0, 134, 5, 2, 43, 5, 138, 0, 0, 5, 5, 2, 43, 5, 138, 0, 0, 134, 5, 2, 43, 5, 182, 0, 0, 5, 5, 2, 43, 5, 182, 0, 0, 134, 5, 2, 43, 5, 154, 0, 0, 5, 5, 2, 43, 5, 154, 0, 0, 134, 5, 2, 59, 5, 196, 0, 0, 5, 5, 2, 59, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 1, 0, 2, 3, 3, 235, 134, 5, 6, 3, 2, 235, 134, 5, 0, 58, 156, 5, 48, 0, 2, 3, 2, 251, 134, 5, 0, 78, 156, 5, 1, 0, 38, 3, 255, 255, 2, 79, 134, 5, 39, 3, 255, 255, 2, 79, 134, 5, 0, 80, 156, 5, 1, 0, 38, 3, 255, 255, 2, 81, 134, 5, 39, 3, 255, 255, 2, 81, 134, 5, 0, 42, 5, 5, 1, 0, 2, 3, 3, 235, 5, 5, 6, 3, 2, 235, 5, 5, 0, 58, 5, 5, 48, 0, 2, 3, 2, 251, 5, 5, 0, 78, 5, 5, 1, 0, 38, 3, 255, 255, 2, 79, 5, 5, 39, 3, 255, 255, 2, 79, 5, 5, 0, 80, 5, 5, 1, 0, 38, 3, 255, 255, 2, 81, 5, 5, 39, 3, 255, 255, 2, 81, 5, 5, 156, 42, 196, 162, 1, 0, 2, 3, 197, 162, 6, 3, 0, 192, 197, 98, 5, 42, 196, 162, 1, 0, 2, 3, 0, 192, 197, 162, 6, 3, 0, 192, 197, 226, 156, 58, 196, 162, 48, 0, 2, 3, 1, 192, 197, 34, 5, 58, 196, 162, 48, 0, 2, 3, 1, 192, 197, 98, 156, 42, 196, 200, 1, 0, 2, 3, 1, 192, 197, 162, 6, 3, 1, 192, 197, 226, 5, 42, 196, 200, 1, 0, 2, 3, 2, 192, 197, 34, 6, 3, 2, 192, 197, 98, 156, 58, 196, 208, 48, 0, 2, 3, 2, 192, 197, 162, 5, 58, 196, 208, 48, 0, 2, 3, 2, 192, 197, 226, 156, 42, 196, 196, 1, 0, 2, 3, 3, 192, 197, 34, 6, 3, 3, 192, 197, 98, 5, 42, 196, 196, 1, 0, 2, 3, 3, 192, 197, 162, 6, 3, 3, 192, 197, 226, 156, 58, 196, 196, 48, 0, 2, 3, 8, 192, 197, 34, 5, 58, 196, 196, 48, 0, 2, 3, 8, 192, 197, 98, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 68, 185, 72, 185, 76, 186, 80, 185, 84, 185, 88, 185, 92, 28, 4, 186, 108, 185, 112, 185, 116, 185, 120, 186, 124, 185, 132, 185, 136, 185, 140, 185, 148, 30, 4, 32, 4, 185, 172, 185, 176, 185, 180, 185, 184, 185, 188, 185, 192, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 34, 4, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 92, 36, 4, 168, 108, 168, 112, 168, 116, 168, 120, 168, 124, 168, 132, 168, 136, 168, 140, 168, 148, 38, 4, 40, 4, 168, 172, 168, 176, 168, 180, 168, 184, 168, 188, 168, 192, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 132, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 185, 64, 153, 58, 89, 58, 57, 58, 0, 8, 249, 74, 186, 81, 154, 81, 250, 81, 90, 82, 186, 97, 154, 97, 185, 100, 90, 98, 217, 76, 154, 126, 185, 133, 153, 133, 249, 133, 153, 134, 89, 134, 136, 14, 217, 134, 185, 173, 153, 173, 249, 173, 89, 174, 153, 189, 185, 196, 2, 8, 168, 57, 136, 57, 168, 64, 136, 58, 72, 58, 40, 58, 4, 8, 232, 74, 168, 81, 136, 81, 232, 81, 72, 82, 168, 97, 136, 97, 168, 100, 72, 98, 200, 76, 136, 126, 168, 133, 136, 133, 232, 133, 136, 134, 72, 134, 128, 14, 200, 134, 168, 173, 136, 173, 232, 173, 72, 174, 136, 189, 168, 196, 72, 190, 57, 59, 40, 59, 185, 60, 168, 60, 42, 4, 44, 4, 153, 73, 136, 73, 249, 73, 232, 73, 185, 74, 168, 74, 25, 74, 8, 74, 25, 78, 8, 78, 89, 79, 72, 79, 58, 83, 40, 83, 218, 81, 200, 81, 186, 82, 168, 82, 26, 83, 8, 83, 26, 82, 8, 82, 249, 89, 232, 89, 217, 89, 200, 89, 185, 90, 168, 90, 249, 90, 232, 90, 249, 93, 232, 93, 89, 95, 72, 95, 154, 98, 136, 98, 58, 99, 40, 99, 218, 97, 200, 97, 46, 4, 48, 4, 186, 98, 168, 104, 6, 8, 8, 8, 250, 109, 232, 109, 249, 114, 232, 114, 168, 144, 153, 117, 136, 117, 249, 118, 232, 118, 25, 118, 8, 118, 121, 119, 104, 119, 89, 119, 72, 119, 154, 125, 136, 125, 250, 126, 232, 126, 26, 126, 8, 126, 10, 8, 185, 128, 168, 128, 57, 135, 40, 135, 217, 133, 200, 133, 121, 134, 104, 134, 12, 8, 14, 8, 153, 149, 136, 149, 249, 150, 232, 150, 25, 150, 8, 150, 154, 153, 136, 153, 250, 153, 232, 153, 185, 156, 168, 156, 26, 154, 8, 154, 185, 164, 168, 164, 25, 162, 8, 162, 185, 168, 168, 168, 153, 174, 136, 174, 57, 175, 40, 175, 217, 173, 200, 173, 57, 174, 40, 174, 121, 174, 104, 174, 25, 175, 8, 175, 249, 181, 232, 181, 249, 189, 232, 189, 89, 190, 153, 193, 136, 193, 185, 194, 168, 194, 25, 194, 8, 194, 232, 152, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 80, 200, 152, 169, 152, 200, 56, 169, 80, 187, 96, 187, 108, 169, 96, 169, 108, 169, 200, 169, 124, 217, 132, 187, 80, 200, 132, 169, 80, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 96, 255, 5, 186, 152, 255, 5, 185, 160, 255, 5, 168, 56, 255, 5, 168, 96, 255, 5, 168, 152, 255, 5, 168, 160, 255, 5, 25, 59, 255, 5, 8, 59, 255, 5, 26, 99, 255, 5, 8, 99, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 51, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 186, 8, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 191, 8, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23, 30, 1, 32, 7, 0, 0, 0, 39, 0, 0, 0, 31, 30, 1, 0, 31, 30, 1, 0, 6, 0, 0, 0, 0, 0, 0, 0, 55, 230, 0, 0, 85, 67, 65, 82, 117, 108, 101, 115, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 100, 101, 102, 97, 117, 108, 116, 0, 101, 109, 111, 106, 105, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 101, 111, 114, 0, 112, 114, 105, 118, 97, 116, 101, 45, 117, 110, 105, 104, 97, 110, 0, 115, 101, 97, 114, 99, 104, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 37, 37, 68, 69, 80, 69, 78, 68, 69, 78, 67, 89, 0, 100, 101, 112, 101, 110, 100, 115, 0, 170, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 0, 0, 117, 0, 99, 0, 97, 0, 100, 0, 97, 0, 116, 0, 97, 0, 46, 0, 105, 0, 99, 0, 117, 0, 0, 0, 207, 221, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 115, 0, 117, 0, 112, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 67, 0, 111, 0, 110, 0, 116, 0, 114, 0, 97, 0, 99, 0, 116, 0, 105, 0, 111, 0, 110, 0, 115, 0, 32, 0, 91, 0, 64, 14, 45, 0, 68, 14, 32, 0, 192, 14, 45, 0, 196, 14, 32, 0, 181, 170, 32, 0, 182, 170, 32, 0, 185, 170, 32, 0, 187, 170, 32, 0, 188, 170, 181, 25, 45, 0, 183, 25, 186, 25, 93, 0, 93, 0, 38, 0, 39, 0, 61, 0, 39, 0, 60, 0, 39, 0, 96, 34, 39, 0, 38, 0, 39, 6, 60, 0, 60, 0, 60, 0, 142, 254, 60, 0, 60, 0, 60, 0, 141, 254, 60, 0, 60, 0, 34, 6, 60, 0, 60, 0, 60, 0, 130, 254, 60, 0, 60, 0, 60, 0, 129, 254, 60, 0, 60, 0, 35, 6, 60, 0, 60, 0, 60, 0, 132, 254, 60, 0, 60, 0, 60, 0, 131, 254, 60, 0, 60, 0, 37, 6, 60, 0, 60, 0, 60, 0, 136, 254, 60, 0, 60, 0, 60, 0, 135, 254, 38, 0, 72, 6, 60, 0, 60, 0, 60, 0, 229, 6, 60, 0, 60, 0, 60, 0, 238, 254, 60, 0, 60, 0, 60, 0, 237, 254, 60, 0, 60, 0, 36, 6, 60, 0, 60, 0, 60, 0, 134, 254, 60, 0, 60, 0, 60, 0, 133, 254, 38, 0, 74, 6, 60, 0, 60, 0, 60, 0, 230, 6, 60, 0, 60, 0, 60, 0, 243, 254, 60, 0, 60, 0, 60, 0, 244, 254, 60, 0, 60, 0, 60, 0, 242, 254, 60, 0, 60, 0, 60, 0, 241, 254, 60, 0, 60, 0, 38, 6, 60, 0, 60, 0, 60, 0, 139, 254, 60, 0, 60, 0, 60, 0, 140, 254, 60, 0, 60, 0, 60, 0, 138, 254, 60, 0, 60, 0, 60, 0, 137, 254, 60, 0, 60, 0, 73, 6, 60, 0, 60, 0, 60, 0, 232, 251, 60, 0, 60, 0, 60, 0, 233, 251, 60, 0, 60, 0, 60, 0, 240, 254, 60, 0, 60, 0, 60, 0, 239, 254, 38, 0, 71, 6, 60, 0, 60, 0, 60, 0, 235, 254, 60, 0, 60, 0, 60, 0, 236, 254, 60, 0, 60, 0, 60, 0, 234, 254, 60, 0, 60, 0, 60, 0, 233, 254, 60, 0, 60, 0, 41, 6, 60, 0, 60, 0, 60, 0, 148, 254, 60, 0, 60, 0, 60, 0, 147, 254, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 112, 0, 114, 0, 105, 0, 109, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 60, 0, 60, 0, 243, 5, 60, 0, 60, 0, 244, 5, 60, 0, 60, 0, 64, 6, 60, 0, 60, 0, 58, 14, 38, 0, 0, 17, 61, 0, 168, 17, 38, 0, 0, 17, 0, 17, 61, 0, 1, 17, 61, 0, 169, 17, 38, 0, 0, 17, 9, 17, 61, 0, 170, 17, 38, 0, 2, 17, 61, 0, 171, 17, 38, 0, 2, 17, 12, 17, 61, 0, 172, 17, 38, 0, 2, 17, 18, 17, 61, 0, 173, 17, 38, 0, 3, 17, 61, 0, 174, 17, 38, 0, 3, 17, 3, 17, 61, 0, 4, 17, 38, 0, 5, 17, 61, 0, 175, 17, 38, 0, 5, 17, 0, 17, 61, 0, 176, 17, 38, 0, 5, 17, 6, 17, 61, 0, 177, 17, 38, 0, 5, 17, 7, 17, 61, 0, 178, 17, 38, 0, 5, 17, 9, 17, 61, 0, 179, 17, 38, 0, 5, 17, 16, 17, 61, 0, 180, 17, 38, 0, 5, 17, 17, 17, 61, 0, 181, 17, 38, 0, 5, 17, 18, 17, 61, 0, 182, 17, 38, 0, 6, 17, 61, 0, 183, 17, 38, 0, 7, 17, 61, 0, 184, 17, 38, 0, 7, 17, 7, 17, 61, 0, 8, 17, 38, 0, 7, 17, 9, 17, 61, 0, 185, 17, 38, 0, 9, 17, 61, 0, 186, 17, 38, 0, 9, 17, 9, 17, 61, 0, 10, 17, 61, 0, 187, 17, 38, 0, 11, 17, 61, 0, 188, 17, 38, 0, 12, 17, 61, 0, 189, 17, 38, 0, 12, 17, 12, 17, 61, 0, 13, 17, 38, 0, 14, 17, 61, 0, 190, 17, 38, 0, 15, 17, 61, 0, 191, 17, 38, 0, 16, 17, 61, 0, 192, 17, 38, 0, 17, 17, 61, 0, 193, 17, 38, 0, 18, 17, 61, 0, 194, 17, 38, 0, 97, 17, 117, 17, 61, 0, 98, 17, 38, 0, 99, 17, 117, 17, 61, 0, 100, 17, 38, 0, 101, 17, 117, 17, 61, 0, 102, 17, 38, 0, 103, 17, 117, 17, 61, 0, 104, 17, 38, 0, 105, 17, 97, 17, 61, 0, 106, 17, 38, 0, 105, 17, 97, 17, 117, 17, 61, 0, 107, 17, 38, 0, 105, 17, 117, 17, 61, 0, 108, 17, 38, 0, 110, 17, 101, 17, 61, 0, 111, 17, 38, 0, 110, 17, 101, 17, 117, 17, 61, 0, 112, 17, 38, 0, 110, 17, 117, 17, 61, 0, 113, 17, 38, 0, 115, 17, 117, 17, 61, 0, 116, 17, 0, 0, 239, 223, 168, 4, 38, 0, 0, 78, 61, 0, 208, 253, 0, 78, 38, 0, 40, 78, 61, 0, 208, 253, 40, 78, 38, 0, 54, 78, 61, 0, 208, 253, 54, 78, 38, 0, 63, 78, 61, 0, 208, 253, 63, 78, 38, 0, 89, 78, 61, 0, 208, 253, 89, 78, 38, 0, 133, 78, 61, 0, 208, 253, 133, 78, 38, 0, 140, 78, 61, 0, 208, 253, 140, 78, 38, 0, 160, 78, 61, 0, 208, 253, 160, 78, 38, 0, 186, 78, 61, 0, 208, 253, 186, 78, 38, 0, 63, 81, 61, 0, 208, 253, 63, 81, 38, 0, 101, 81, 61, 0, 208, 253, 101, 81, 38, 0, 107, 81, 61, 0, 208, 253, 107, 81, 38, 0, 130, 81, 61, 0, 208, 253, 130, 81, 38, 0, 150, 81, 61, 0, 208, 253, 150, 81, 38, 0, 171, 81, 61, 0, 208, 253, 171, 81, 38, 0, 224, 81, 61, 0, 208, 253, 224, 81, 38, 0, 245, 81, 61, 0, 208, 253, 245, 81, 38, 0, 0, 82, 61, 0, 208, 253, 0, 82, 38, 0, 155, 82, 61, 0, 208, 253, 155, 82, 38, 0, 249, 82, 61, 0, 208, 253, 249, 82, 38, 0, 21, 83, 61, 0, 208, 253, 21, 83, 38, 0, 26, 83, 61, 0, 208, 253, 26, 83, 38, 0, 56, 83, 61, 0, 208, 253, 56, 83, 38, 0, 65, 83, 61, 0, 208, 253, 65, 83, 38, 0, 92, 83, 61, 0, 208, 253, 92, 83, 38, 0, 105, 83, 61, 0, 208, 253, 105, 83, 38, 0, 130, 83, 61, 0, 208, 253, 130, 83, 38, 0, 182, 83, 61, 0, 208, 253, 182, 83, 38, 0, 200, 83, 61, 0, 208, 253, 200, 83, 38, 0, 227, 83, 61, 0, 208, 253, 227, 83, 38, 0, 215, 86, 61, 0, 208, 253, 215, 86, 38, 0, 31, 87, 61, 0, 208, 253, 31, 87, 38, 0, 235, 88, 61, 0, 208, 253, 235, 88, 38, 0, 2, 89, 61, 0, 208, 253, 2, 89, 38, 0, 10, 89, 61, 0, 208, 253, 10, 89, 38, 0, 21, 89, 61, 0, 208, 253, 21, 89, 38, 0, 39, 89, 61, 0, 208, 253, 39, 89, 38, 0, 115, 89, 61, 0, 208, 253, 115, 89, 38, 0, 80, 91, 61, 0, 208, 253, 80, 91, 38, 0, 128, 91, 61, 0, 208, 253, 128, 91, 38, 0, 248, 91, 61, 0, 208, 253, 248, 91, 38, 0, 15, 92, 61, 0, 208, 253, 15, 92, 38, 0, 34, 92, 61, 0, 208, 253, 34, 92, 38, 0, 56, 92, 61, 0, 208, 253, 56, 92, 38, 0, 110, 92, 61, 0, 208, 253, 110, 92, 38, 0, 113, 92, 61, 0, 208, 253, 113, 92, 38, 0, 219, 93, 61, 0, 208, 253, 219, 93, 38, 0, 229, 93, 61, 0, 208, 253, 229, 93, 38, 0, 241, 93, 61, 0, 208, 253, 241, 93, 38, 0, 254, 93, 61, 0, 208, 253, 254, 93, 38, 0, 114, 94, 61, 0, 208, 253, 114, 94, 38, 0, 97, 78, 61, 0, 208, 253, 122, 94, 38, 0, 127, 94, 61, 0, 208, 253, 127, 94, 38, 0, 244, 94, 61, 0, 208, 253, 244, 94, 38, 0, 254, 94, 61, 0, 208, 253, 254, 94, 38, 0, 11, 95, 61, 0, 208, 253, 11, 95, 38, 0, 19, 95, 61, 0, 208, 253, 19, 95, 38, 0, 80, 95, 61, 0, 208, 253, 80, 95, 38, 0, 97, 95, 61, 0, 208, 253, 97, 95, 38, 0, 115, 95, 61, 0, 208, 253, 115, 95, 38, 0, 195, 95, 61, 0, 208, 253, 195, 95, 38, 0, 8, 98, 61, 0, 208, 253, 8, 98, 38, 0, 54, 98, 61, 0, 208, 253, 54, 98, 38, 0, 75, 98, 61, 0, 208, 253, 75, 98, 38, 0, 47, 101, 61, 0, 208, 253, 47, 101, 38, 0, 52, 101, 61, 0, 208, 253, 52, 101, 38, 0, 135, 101, 61, 0, 208, 253, 135, 101, 38, 0, 151, 101, 61, 0, 208, 253, 151, 101, 38, 0, 164, 101, 61, 0, 208, 253, 164, 101, 38, 0, 185, 101, 61, 0, 208, 253, 185, 101, 38, 0, 224, 101, 61, 0, 208, 253, 224, 101, 38, 0, 229, 101, 61, 0, 208, 253, 229, 101, 38, 0, 240, 102, 61, 0, 208, 253, 240, 102, 38, 0, 8, 103, 61, 0, 208, 253, 8, 103, 38, 0, 40, 103, 61, 0, 208, 253, 40, 103, 38, 0, 32, 107, 61, 0, 208, 253, 32, 107, 38, 0, 98, 107, 61, 0, 208, 253, 98, 107, 38, 0, 121, 107, 61, 0, 208, 253, 121, 107, 38, 0, 179, 107, 61, 0, 208, 253, 179, 107, 38, 0, 203, 107, 61, 0, 208, 253, 203, 107, 38, 0, 212, 107, 61, 0, 208, 253, 212, 107, 38, 0, 219, 107, 61, 0, 208, 253, 219, 107, 38, 0, 15, 108, 61, 0, 208, 253, 15, 108, 38, 0, 20, 108, 61, 0, 208, 253, 20, 108, 38, 0, 52, 108, 61, 0, 208, 253, 52, 108, 38, 0, 107, 112, 61, 0, 208, 253, 107, 112, 38, 0, 42, 114, 61, 0, 208, 253, 42, 114, 38, 0, 54, 114, 61, 0, 208, 253, 54, 114, 38, 0, 59, 114, 61, 0, 208, 253, 59, 114, 38, 0, 44, 78, 61, 0, 208, 253, 63, 114, 38, 0, 130, 216, 23, 220, 61, 0, 208, 253, 44, 78, 38, 0, 71, 114, 61, 0, 208, 253, 71, 114, 38, 0, 89, 114, 61, 0, 208, 253, 89, 114, 38, 0, 91, 114, 61, 0, 208, 253, 91, 114, 38, 0, 172, 114, 61, 0, 208, 253, 172, 114, 38, 0, 132, 115, 61, 0, 208, 253, 132, 115, 38, 0, 137, 115, 61, 0, 208, 253, 137, 115, 38, 0, 220, 116, 61, 0, 208, 253, 220, 116, 38, 0, 230, 116, 61, 0, 208, 253, 230, 116, 38, 0, 24, 117, 61, 0, 208, 253, 24, 117, 38, 0, 31, 117, 61, 0, 208, 253, 31, 117, 38, 0, 40, 117, 61, 0, 208, 253, 40, 117, 38, 0, 48, 117, 61, 0, 208, 253, 48, 117, 38, 0, 139, 117, 61, 0, 208, 253, 139, 117, 38, 0, 146, 117, 61, 0, 208, 253, 146, 117, 38, 0, 118, 118, 61, 0, 208, 253, 118, 118, 38, 0, 125, 118, 61, 0, 208, 253, 125, 118, 38, 0, 174, 118, 61, 0, 208, 253, 174, 118, 38, 0, 191, 118, 61, 0, 208, 253, 191, 118, 38, 0, 238, 118, 61, 0, 208, 253, 238, 118, 38, 0, 219, 119, 61, 0, 208, 253, 219, 119, 38, 0, 226, 119, 61, 0, 208, 253, 226, 119, 38, 0, 243, 119, 61, 0, 208, 253, 243, 119, 38, 0, 58, 121, 61, 0, 208, 253, 58, 121, 38, 0, 184, 121, 61, 0, 208, 253, 184, 121, 38, 0, 190, 121, 61, 0, 208, 253, 190, 121, 38, 0, 116, 122, 61, 0, 208, 253, 116, 122, 38, 0, 203, 122, 61, 0, 208, 253, 203, 122, 38, 0, 249, 122, 61, 0, 208, 253, 249, 122, 38, 0, 115, 124, 61, 0, 208, 253, 115, 124, 38, 0, 248, 124, 61, 0, 208, 253, 248, 124, 38, 0, 159, 126, 61, 0, 208, 253, 159, 126, 38, 0, 54, 127, 61, 0, 208, 253, 54, 127, 38, 0, 81, 127, 61, 0, 208, 253, 81, 127, 38, 0, 138, 127, 61, 0, 208, 253, 138, 127, 38, 0, 189, 127, 61, 0, 208, 253, 189, 127, 38, 0, 1, 128, 61, 0, 208, 253, 1, 128, 38, 0, 12, 128, 61, 0, 208, 253, 12, 128, 38, 0, 18, 128, 61, 0, 208, 253, 18, 128, 38, 0, 51, 128, 61, 0, 208, 253, 51, 128, 38, 0, 127, 128, 61, 0, 208, 253, 127, 128, 38, 0, 137, 128, 61, 0, 208, 253, 137, 128, 38, 0, 227, 129, 61, 0, 208, 253, 227, 129, 38, 0, 234, 129, 61, 0, 208, 253, 234, 129, 38, 0, 243, 129, 61, 0, 208, 253, 243, 129, 38, 0, 252, 129, 61, 0, 208, 253, 252, 129, 38, 0, 12, 130, 61, 0, 208, 253, 12, 130, 38, 0, 27, 130, 61, 0, 208, 253, 27, 130, 38, 0, 31, 130, 61, 0, 208, 253, 31, 130, 38, 0, 110, 130, 61, 0, 208, 253, 110, 130, 38, 0, 114, 130, 61, 0, 208, 253, 114, 130, 38, 0, 120, 130, 61, 0, 208, 253, 120, 130, 38, 0, 77, 134, 61, 0, 208, 253, 77, 134, 38, 0, 107, 134, 61, 0, 208, 253, 107, 134, 38, 0, 64, 136, 61, 0, 208, 253, 64, 136, 38, 0, 76, 136, 61, 0, 208, 253, 76, 136, 38, 0, 99, 136, 61, 0, 208, 253, 99, 136, 38, 0, 126, 137, 61, 0, 208, 253, 126, 137, 38, 0, 139, 137, 61, 0, 208, 253, 139, 137, 38, 0, 193, 137, 61, 0, 208, 253, 193, 137, 38, 0, 210, 137, 61, 0, 208, 253, 210, 137, 38, 0, 0, 138, 61, 0, 208, 253, 0, 138, 38, 0, 160, 139, 61, 0, 208, 253, 160, 139, 38, 0, 55, 140, 61, 0, 208, 253, 55, 140, 38, 0, 70, 140, 61, 0, 208, 253, 70, 140, 38, 0, 85, 140, 61, 0, 208, 253, 85, 140, 38, 0, 120, 140, 61, 0, 208, 253, 120, 140, 38, 0, 157, 140, 61, 0, 208, 253, 157, 140, 38, 0, 29, 141, 61, 0, 208, 253, 29, 141, 38, 0, 100, 141, 61, 0, 208, 253, 100, 141, 38, 0, 112, 141, 61, 0, 208, 253, 112, 141, 38, 0, 179, 141, 61, 0, 208, 253, 179, 141, 38, 0, 171, 142, 61, 0, 208, 253, 171, 142, 38, 0, 202, 142, 61, 0, 208, 253, 202, 142, 38, 0, 102, 143, 61, 0, 208, 253, 102, 143, 38, 0, 155, 143, 61, 0, 208, 253, 155, 143, 38, 0, 176, 143, 61, 0, 208, 253, 176, 143, 38, 0, 181, 143, 61, 0, 208, 253, 181, 143, 38, 0, 145, 144, 61, 0, 208, 253, 145, 144, 38, 0, 73, 145, 61, 0, 208, 253, 73, 145, 38, 0, 198, 145, 61, 0, 208, 253, 198, 145, 38, 0, 204, 145, 61, 0, 208, 253, 204, 145, 38, 0, 209, 145, 61, 0, 208, 253, 209, 145, 38, 0, 133, 148, 61, 0, 208, 253, 133, 148, 38, 0, 119, 149, 61, 0, 208, 253, 119, 149, 38, 0, 127, 149, 61, 0, 208, 253, 127, 149, 38, 0, 128, 149, 61, 0, 208, 253, 128, 149, 38, 0, 232, 149, 61, 0, 208, 253, 232, 149, 38, 0, 28, 150, 61, 0, 208, 253, 28, 150, 38, 0, 182, 150, 61, 0, 208, 253, 182, 150, 38, 0, 185, 150, 61, 0, 208, 253, 185, 150, 38, 0, 232, 150, 61, 0, 208, 253, 232, 150, 38, 0, 81, 151, 61, 0, 208, 253, 81, 151, 38, 0, 94, 151, 61, 0, 208, 253, 94, 151, 38, 0, 98, 151, 61, 0, 208, 253, 98, 151, 38, 0, 105, 151, 61, 0, 208, 253, 105, 151, 38, 0, 203, 151, 61, 0, 208, 253, 203, 151, 38, 0, 230, 151, 61, 0, 208, 253, 230, 151, 38, 0, 237, 151, 61, 0, 208, 253, 237, 151, 38, 0, 243, 151, 61, 0, 208, 253, 243, 151, 38, 0, 1, 152, 61, 0, 208, 253, 1, 152, 38, 0, 117, 152, 61, 0, 208, 253, 117, 152, 38, 0, 168, 152, 61, 0, 208, 253, 168, 152, 38, 0, 206, 152, 61, 0, 208, 253, 206, 152, 38, 0, 219, 152, 61, 0, 208, 253, 219, 152, 38, 0, 222, 152, 61, 0, 208, 253, 222, 152, 38, 0, 223, 152, 61, 0, 208, 253, 223, 152, 38, 0, 99, 153, 61, 0, 208, 253, 99, 153, 38, 0, 150, 153, 61, 0, 208, 253, 150, 153, 38, 0, 153, 153, 61, 0, 208, 253, 153, 153, 38, 0, 172, 153, 61, 0, 208, 253, 172, 153, 38, 0, 108, 154, 61, 0, 208, 253, 108, 154, 38, 0, 168, 154, 61, 0, 208, 253, 168, 154, 38, 0, 216, 154, 61, 0, 208, 253, 216, 154, 38, 0, 223, 154, 61, 0, 208, 253, 223, 154, 38, 0, 37, 155, 61, 0, 208, 253, 37, 155, 38, 0, 47, 155, 61, 0, 208, 253, 47, 155, 38, 0, 50, 155, 61, 0, 208, 253, 50, 155, 38, 0, 60, 155, 61, 0, 208, 253, 60, 155, 38, 0, 90, 155, 61, 0, 208, 253, 90, 155, 38, 0, 124, 156, 61, 0, 208, 253, 124, 156, 38, 0, 229, 156, 61, 0, 208, 253, 229, 156, 38, 0, 31, 158, 61, 0, 208, 253, 31, 158, 38, 0, 117, 158, 61, 0, 208, 253, 117, 158, 38, 0, 132, 216, 186, 222, 61, 0, 208, 253, 100, 83, 38, 0, 127, 158, 61, 0, 208, 253, 127, 158, 38, 0, 165, 158, 61, 0, 208, 253, 165, 158, 38, 0, 166, 158, 61, 0, 208, 253, 166, 158, 38, 0, 187, 158, 61, 0, 208, 253, 187, 158, 38, 0, 195, 158, 61, 0, 208, 253, 195, 158, 38, 0, 205, 158, 61, 0, 208, 253, 205, 158, 38, 0, 209, 158, 61, 0, 208, 253, 209, 158, 38, 0, 249, 158, 61, 0, 208, 253, 249, 158, 38, 0, 253, 158, 61, 0, 208, 253, 253, 158, 38, 0, 254, 158, 61, 0, 208, 253, 254, 158, 38, 0, 14, 159, 61, 0, 208, 253, 14, 159, 38, 0, 19, 159, 61, 0, 208, 253, 19, 159, 38, 0, 32, 159, 61, 0, 208, 253, 32, 159, 38, 0, 59, 159, 61, 0, 208, 253, 59, 159, 38, 0, 74, 159, 61, 0, 208, 253, 74, 159, 38, 0, 80, 159, 61, 0, 208, 253, 80, 159, 38, 0, 82, 159, 61, 0, 208, 253, 82, 159, 38, 0, 127, 159, 61, 0, 208, 253, 127, 159, 38, 0, 141, 159, 61, 0, 208, 253, 141, 159, 38, 0, 153, 159, 61, 0, 208, 253, 153, 159, 38, 0, 156, 159, 61, 0, 208, 253, 156, 159, 38, 0, 159, 159, 61, 0, 208, 253, 159, 159, 38, 0, 160, 159, 61, 0, 208, 253, 160, 159, 0, 0, 239, 223, 197, 6, 38, 0, 97, 0, 60, 0, 60, 0, 0, 29, 60, 0, 60, 0, 101, 44, 60, 0, 60, 0, 60, 0, 58, 2, 60, 0, 60, 0, 143, 29, 60, 0, 60, 0, 80, 2, 60, 0, 60, 0, 60, 0, 111, 44, 60, 0, 60, 0, 60, 0, 68, 29, 60, 0, 60, 0, 81, 2, 60, 0, 60, 0, 60, 0, 109, 44, 60, 0, 60, 0, 60, 0, 69, 29, 60, 0, 60, 0, 144, 29, 60, 0, 60, 0, 82, 2, 60, 0, 60, 0, 60, 0, 112, 44, 60, 0, 60, 0, 60, 0, 155, 29, 38, 0, 98, 0, 60, 0, 60, 0, 153, 2, 60, 0, 60, 0, 128, 1, 60, 0, 60, 0, 60, 0, 67, 2, 60, 0, 60, 0, 47, 29, 60, 0, 60, 0, 3, 29, 60, 0, 60, 0, 108, 29, 60, 0, 60, 0, 128, 29, 60, 0, 60, 0, 83, 2, 60, 0, 60, 0, 60, 0, 129, 1, 60, 0, 60, 0, 131, 1, 60, 0, 60, 0, 60, 0, 130, 1, 38, 0, 99, 0, 60, 0, 60, 0, 4, 29, 60, 0, 60, 0, 60, 2, 60, 0, 60, 0, 60, 0, 59, 2, 60, 0, 60, 0, 147, 167, 60, 0, 60, 0, 60, 0, 146, 167, 60, 0, 60, 0, 136, 1, 60, 0, 60, 0, 60, 0, 135, 1, 60, 0, 60, 0, 85, 2, 60, 0, 60, 0, 60, 0, 157, 29, 60, 0, 60, 0, 132, 33, 60, 0, 60, 0, 60, 0, 131, 33, 60, 0, 60, 0, 63, 167, 60, 0, 60, 0, 60, 0, 62, 167, 38, 0, 100, 0, 60, 0, 60, 0, 5, 29, 60, 0, 60, 0, 6, 29, 60, 0, 60, 0, 109, 29, 60, 0, 60, 0, 129, 29, 60, 0, 60, 0, 86, 2, 60, 0, 60, 0, 60, 0, 137, 1, 60, 0, 60, 0, 87, 2, 60, 0, 60, 0, 60, 0, 138, 1, 60, 0, 60, 0, 145, 29, 60, 0, 60, 0, 140, 1, 60, 0, 60, 0, 60, 0, 139, 1, 60, 0, 60, 0, 33, 2, 60, 0, 60, 0, 113, 167, 60, 0, 60, 0, 159, 30, 60, 0, 60, 0, 141, 1, 38, 0, 101, 0, 60, 0, 60, 0, 7, 29, 60, 0, 60, 0, 71, 2, 60, 0, 60, 0, 60, 0, 70, 2, 60, 0, 60, 0, 146, 29, 60, 0, 60, 0, 120, 44, 60, 0, 60, 0, 221, 1, 60, 0, 60, 0, 60, 0, 142, 1, 60, 0, 60, 0, 60, 0, 50, 29, 60, 0, 60, 0, 123, 44, 60, 0, 60, 0, 89, 2, 60, 0, 60, 0, 60, 0, 143, 1, 60, 0, 60, 0, 60, 0, 74, 29, 60, 0, 60, 0, 60, 0, 148, 32, 60, 0, 60, 0, 149, 29, 60, 0, 60, 0, 91, 2, 60, 0, 60, 0, 60, 0, 144, 1, 60, 0, 60, 0, 60, 0, 7, 33, 60, 0, 60, 0, 60, 0, 75, 29, 60, 0, 60, 0, 147, 29, 60, 0, 60, 0, 88, 2, 60, 0, 60, 0, 90, 2, 60, 0, 60, 0, 92, 2, 60, 0, 60, 0, 60, 0, 159, 29, 60, 0, 60, 0, 148, 29, 60, 0, 60, 0, 8, 29, 60, 0, 60, 0, 60, 0, 76, 29, 60, 0, 60, 0, 93, 2, 60, 0, 60, 0, 94, 2, 60, 0, 60, 0, 154, 2, 38, 0, 102, 0, 60, 0, 60, 0, 48, 167, 60, 0, 60, 0, 110, 29, 60, 0, 60, 0, 130, 29, 60, 0, 60, 0, 146, 1, 60, 0, 60, 0, 60, 0, 145, 1, 60, 0, 60, 0, 78, 33, 60, 0, 60, 0, 60, 0, 50, 33, 60, 0, 60, 0, 251, 167, 38, 0, 103, 0, 60, 0, 60, 0, 97, 2, 60, 0, 60, 0, 60, 0, 162, 29, 60, 0, 60, 0, 98, 2, 60, 0, 60, 0, 60, 0, 219, 29, 60, 0, 60, 0, 229, 1, 60, 0, 60, 0, 60, 0, 228, 1, 60, 0, 60, 0, 131, 29, 60, 0, 60, 0, 96, 2, 60, 0, 60, 0, 60, 0, 147, 1, 60, 0, 60, 0, 155, 2, 60, 0, 60, 0, 119, 29, 60, 0, 60, 0, 127, 167, 60, 0, 60, 0, 60, 0, 126, 167, 60, 0, 60, 0, 99, 2, 60, 0, 60, 0, 60, 0, 148, 1, 60, 0, 60, 0, 60, 0, 224, 2, 60, 0, 60, 0, 100, 2, 60, 0, 60, 0, 163, 1, 60, 0, 60, 0, 60, 0, 162, 1, 38, 0, 104, 0, 60, 0, 60, 0, 156, 2, 60, 0, 60, 0, 102, 2, 60, 0, 60, 0, 60, 0, 170, 167, 60, 0, 60, 0, 60, 0, 177, 2, 60, 0, 60, 0, 104, 44, 60, 0, 60, 0, 60, 0, 103, 44, 60, 0, 60, 0, 118, 44, 60, 0, 60, 0, 60, 0, 117, 44, 60, 0, 60, 0, 39, 167, 60, 0, 60, 0, 60, 0, 38, 167, 60, 0, 60, 0, 103, 2, 60, 0, 60, 0, 101, 2, 60, 0, 60, 0, 60, 0, 141, 167, 60, 0, 60, 0, 60, 0, 163, 29, 60, 0, 60, 0, 174, 2, 60, 0, 60, 0, 175, 2, 38, 0, 105, 0, 60, 0, 60, 0, 49, 1, 60, 0, 60, 0, 60, 0, 53, 216, 164, 222, 60, 0, 60, 0, 106, 2, 60, 0, 60, 0, 60, 0, 166, 29, 60, 0, 60, 0, 254, 167, 60, 0, 60, 0, 9, 29, 60, 0, 60, 0, 60, 0, 78, 29, 60, 0, 60, 0, 104, 2, 60, 0, 60, 0, 60, 0, 151, 1, 60, 0, 60, 0, 60, 0, 164, 29, 60, 0, 60, 0, 123, 29, 60, 0, 60, 0, 60, 0, 167, 29, 60, 0, 60, 0, 150, 29, 60, 0, 60, 0, 105, 2, 60, 0, 60, 0, 60, 0, 150, 1, 60, 0, 60, 0, 60, 0, 165, 29, 60, 0, 60, 0, 124, 29, 38, 0, 106, 0, 60, 0, 60, 0, 55, 2, 60, 0, 60, 0, 60, 0, 53, 216, 165, 222, 60, 0, 60, 0, 10, 29, 60, 0, 60, 0, 73, 2, 60, 0, 60, 0, 60, 0, 72, 2, 60, 0, 60, 0, 157, 2, 60, 0, 60, 0, 60, 0, 168, 29, 60, 0, 60, 0, 95, 2, 60, 0, 60, 0, 60, 0, 161, 29, 60, 0, 60, 0, 132, 2, 38, 0, 107, 0, 60, 0, 60, 0, 11, 29, 60, 0, 60, 0, 132, 29, 60, 0, 60, 0, 153, 1, 60, 0, 60, 0, 60, 0, 152, 1, 60, 0, 60, 0, 106, 44, 60, 0, 60, 0, 60, 0, 105, 44, 60, 0, 60, 0, 65, 167, 60, 0, 60, 0, 60, 0, 64, 167, 60, 0, 60, 0, 67, 167, 60, 0, 60, 0, 60, 0, 66, 167, 60, 0, 60, 0, 69, 167, 60, 0, 60, 0, 60, 0, 68, 167, 60, 0, 60, 0, 56, 1, 60, 0, 60, 0, 158, 2, 38, 0, 108, 0, 60, 0, 60, 0, 159, 2, 60, 0, 60, 0, 60, 0, 222, 29, 60, 0, 60, 0, 60, 0, 171, 29, 60, 0, 60, 0, 71, 167, 60, 0, 60, 0, 60, 0, 70, 167, 60, 0, 60, 0, 12, 29, 60, 0, 60, 0, 73, 167, 60, 0, 60, 0, 60, 0, 72, 167, 60, 0, 60, 0, 154, 1, 60, 0, 60, 0, 60, 0, 61, 2, 60, 0, 60, 0, 97, 44, 60, 0, 60, 0, 60, 0, 96, 44, 60, 0, 60, 0, 107, 2, 60, 0, 60, 0, 60, 0, 98, 44, 60, 0, 60, 0, 108, 2, 60, 0, 60, 0, 133, 29, 60, 0, 60, 0, 60, 0, 170, 29, 60, 0, 60, 0, 109, 2, 60, 0, 60, 0, 60, 0, 169, 29, 60, 0, 60, 0, 142, 167, 60, 0, 60, 0, 52, 2, 60, 0, 60, 0, 114, 167, 60, 0, 60, 0, 129, 167, 60, 0, 60, 0, 60, 0, 128, 167, 60, 0, 60, 0, 155, 1, 38, 0, 109, 0, 60, 0, 60, 0, 13, 29, 60, 0, 60, 0, 60, 0, 223, 29, 60, 0, 60, 0, 111, 29, 60, 0, 60, 0, 134, 29, 60, 0, 60, 0, 113, 2, 60, 0, 60, 0, 60, 0, 110, 44, 60, 0, 60, 0, 60, 0, 172, 29, 60, 0, 60, 0, 253, 167, 60, 0, 60, 0, 255, 167, 60, 0, 60, 0, 115, 167, 60, 0, 60, 0, 111, 2, 60, 0, 60, 0, 60, 0, 156, 1, 60, 0, 60, 0, 60, 0, 90, 29, 60, 0, 60, 0, 250, 167, 60, 0, 60, 0, 31, 29, 60, 0, 60, 0, 112, 2, 60, 0, 60, 0, 60, 0, 173, 29, 38, 0, 110, 0, 60, 0, 60, 0, 73, 1, 60, 0, 60, 0, 116, 2, 60, 0, 60, 0, 60, 0, 225, 29, 60, 0, 60, 0, 60, 0, 176, 29, 60, 0, 60, 0, 59, 29, 60, 0, 60, 0, 14, 29, 60, 0, 60, 0, 112, 29, 60, 0, 60, 0, 114, 2, 60, 0, 60, 0, 60, 0, 157, 1, 60, 0, 60, 0, 60, 0, 174, 29, 60, 0, 60, 0, 158, 1, 60, 0, 60, 0, 60, 0, 32, 2, 60, 0, 60, 0, 145, 167, 60, 0, 60, 0, 60, 0, 144, 167, 60, 0, 60, 0, 135, 29, 60, 0, 60, 0, 115, 2, 60, 0, 60, 0, 60, 0, 175, 29, 60, 0, 60, 0, 53, 2, 60, 0, 60, 0, 116, 167, 60, 0, 60, 0, 75, 1, 60, 0, 60, 0, 60, 0, 74, 1, 60, 0, 60, 0, 60, 0, 81, 29, 38, 0, 111, 0, 60, 0, 60, 0, 15, 29, 60, 0, 60, 0, 17, 29, 60, 0, 60, 0, 19, 29, 60, 0, 60, 0, 84, 2, 60, 0, 60, 0, 60, 0, 134, 1, 60, 0, 60, 0, 60, 0, 83, 29, 60, 0, 60, 0, 16, 29, 60, 0, 60, 0, 18, 29, 60, 0, 60, 0, 151, 29, 60, 0, 60, 0, 77, 167, 60, 0, 60, 0, 60, 0, 76, 167, 60, 0, 60, 0, 22, 29, 60, 0, 60, 0, 60, 0, 84, 29, 60, 0, 60, 0, 23, 29, 60, 0, 60, 0, 60, 0, 85, 29, 60, 0, 60, 0, 122, 44, 60, 0, 60, 0, 117, 2, 60, 0, 60, 0, 60, 0, 159, 1, 60, 0, 60, 0, 60, 0, 177, 29, 60, 0, 60, 0, 75, 167, 60, 0, 60, 0, 60, 0, 74, 167, 60, 0, 60, 0, 119, 2, 60, 0, 60, 0, 35, 2, 60, 0, 60, 0, 60, 0, 34, 2, 60, 0, 60, 0, 60, 0, 61, 29, 60, 0, 60, 0, 21, 29, 38, 0, 112, 0, 60, 0, 60, 0, 24, 29, 60, 0, 60, 0, 125, 29, 60, 0, 60, 0, 60, 0, 99, 44, 60, 0, 60, 0, 81, 167, 60, 0, 60, 0, 60, 0, 80, 167, 60, 0, 60, 0, 113, 29, 60, 0, 60, 0, 136, 29, 60, 0, 60, 0, 165, 1, 60, 0, 60, 0, 60, 0, 164, 1, 60, 0, 60, 0, 83, 167, 60, 0, 60, 0, 60, 0, 82, 167, 60, 0, 60, 0, 85, 167, 60, 0, 60, 0, 60, 0, 84, 167, 60, 0, 60, 0, 252, 167, 60, 0, 60, 0, 120, 2, 60, 0, 60, 0, 60, 0, 178, 29, 60, 0, 60, 0, 119, 44, 38, 0, 113, 0, 60, 0, 60, 0, 87, 167, 60, 0, 60, 0, 60, 0, 86, 167, 60, 0, 60, 0, 89, 167, 60, 0, 60, 0, 60, 0, 88, 167, 60, 0, 60, 0, 160, 2, 60, 0, 60, 0, 75, 2, 60, 0, 60, 0, 60, 0, 74, 2, 38, 0, 114, 0, 60, 0, 60, 0, 128, 2, 60, 0, 60, 0, 60, 0, 226, 29, 60, 0, 60, 0, 60, 0, 166, 1, 60, 0, 60, 0, 91, 167, 60, 0, 60, 0, 60, 0, 227, 29, 60, 0, 60, 0, 60, 0, 90, 167, 60, 0, 60, 0, 25, 29, 60, 0, 60, 0, 77, 2, 60, 0, 60, 0, 60, 0, 76, 2, 60, 0, 60, 0, 114, 29, 60, 0, 60, 0, 121, 2, 60, 0, 60, 0, 60, 0, 180, 2, 60, 0, 60, 0, 26, 29, 60, 0, 60, 0, 122, 2, 60, 0, 60, 0, 137, 29, 60, 0, 60, 0, 123, 2, 60, 0, 60, 0, 60, 0, 181, 2, 60, 0, 60, 0, 121, 44, 60, 0, 60, 0, 124, 2, 60, 0, 60, 0, 125, 2, 60, 0, 60, 0, 60, 0, 100, 44, 60, 0, 60, 0, 126, 2, 60, 0, 60, 0, 115, 29, 60, 0, 60, 0, 127, 2, 60, 0, 60, 0, 129, 2, 60, 0, 60, 0, 60, 0, 182, 2, 60, 0, 60, 0, 117, 167, 60, 0, 60, 0, 118, 167, 60, 0, 60, 0, 93, 167, 60, 0, 60, 0, 60, 0, 92, 167, 38, 0, 115, 0, 60, 0, 60, 0, 49, 167, 60, 0, 60, 0, 116, 29, 60, 0, 60, 0, 138, 29, 60, 0, 60, 0, 130, 2, 60, 0, 60, 0, 60, 0, 179, 29, 60, 0, 60, 0, 63, 2, 60, 0, 60, 0, 60, 0, 126, 44, 60, 0, 60, 0, 156, 30, 60, 0, 60, 0, 157, 30, 60, 0, 60, 0, 131, 2, 60, 0, 60, 0, 60, 0, 169, 1, 60, 0, 60, 0, 60, 0, 180, 29, 60, 0, 60, 0, 139, 29, 60, 0, 60, 0, 170, 1, 60, 0, 60, 0, 133, 2, 60, 0, 60, 0, 152, 29, 60, 0, 60, 0, 134, 2, 38, 0, 116, 0, 60, 0, 60, 0, 27, 29, 60, 0, 60, 0, 103, 1, 60, 0, 60, 0, 60, 0, 102, 1, 60, 0, 60, 0, 102, 44, 60, 0, 60, 0, 60, 0, 62, 2, 60, 0, 60, 0, 117, 29, 60, 0, 60, 0, 171, 1, 60, 0, 60, 0, 60, 0, 181, 29, 60, 0, 60, 0, 173, 1, 60, 0, 60, 0, 60, 0, 172, 1, 60, 0, 60, 0, 136, 2, 60, 0, 60, 0, 60, 0, 174, 1, 60, 0, 60, 0, 54, 2, 60, 0, 60, 0, 119, 167, 60, 0, 60, 0, 135, 2, 38, 0, 117, 0, 60, 0, 60, 0, 28, 29, 60, 0, 60, 0, 60, 0, 184, 29, 60, 0, 60, 0, 29, 29, 60, 0, 60, 0, 60, 0, 89, 29, 60, 0, 60, 0, 30, 29, 60, 0, 60, 0, 137, 2, 60, 0, 60, 0, 60, 0, 68, 2, 60, 0, 60, 0, 60, 0, 182, 29, 60, 0, 60, 0, 126, 29, 60, 0, 60, 0, 153, 29, 60, 0, 60, 0, 138, 2, 60, 0, 60, 0, 60, 0, 177, 1, 60, 0, 60, 0, 60, 0, 183, 29, 60, 0, 60, 0, 127, 29, 38, 0, 118, 0, 60, 0, 60, 0, 32, 29, 60, 0, 60, 0, 95, 167, 60, 0, 60, 0, 60, 0, 94, 167, 60, 0, 60, 0, 140, 29, 60, 0, 60, 0, 139, 2, 60, 0, 60, 0, 60, 0, 178, 1, 60, 0, 60, 0, 60, 0, 185, 29, 60, 0, 60, 0, 113, 44, 60, 0, 60, 0, 116, 44, 60, 0, 60, 0, 253, 30, 60, 0, 60, 0, 60, 0, 252, 30, 60, 0, 60, 0, 140, 2, 60, 0, 60, 0, 60, 0, 69, 2, 60, 0, 60, 0, 60, 0, 186, 29, 38, 0, 119, 0, 60, 0, 60, 0, 33, 29, 60, 0, 60, 0, 115, 44, 60, 0, 60, 0, 60, 0, 114, 44, 60, 0, 60, 0, 141, 2, 60, 0, 60, 0, 191, 1, 60, 0, 60, 0, 60, 0, 247, 1, 38, 0, 120, 0, 60, 0, 60, 0, 141, 29, 38, 0, 121, 0, 60, 0, 60, 0, 143, 2, 60, 0, 60, 0, 79, 2, 60, 0, 60, 0, 60, 0, 78, 2, 60, 0, 60, 0, 180, 1, 60, 0, 60, 0, 60, 0, 179, 1, 60, 0, 60, 0, 255, 30, 60, 0, 60, 0, 60, 0, 254, 30, 60, 0, 60, 0, 142, 2, 60, 0, 60, 0, 29, 2, 60, 0, 60, 0, 60, 0, 28, 2, 38, 0, 122, 0, 60, 0, 60, 0, 34, 29, 60, 0, 60, 0, 182, 1, 60, 0, 60, 0, 60, 0, 181, 1, 60, 0, 60, 0, 118, 29, 60, 0, 60, 0, 142, 29, 60, 0, 60, 0, 37, 2, 60, 0, 60, 0, 60, 0, 36, 2, 60, 0, 60, 0, 144, 2, 60, 0, 60, 0, 60, 0, 188, 29, 60, 0, 60, 0, 145, 2, 60, 0, 60, 0, 60, 0, 189, 29, 60, 0, 60, 0, 64, 2, 60, 0, 60, 0, 60, 0, 127, 44, 60, 0, 60, 0, 108, 44, 60, 0, 60, 0, 60, 0, 107, 44, 60, 0, 60, 0, 99, 167, 60, 0, 60, 0, 60, 0, 98, 167, 60, 0, 60, 0, 146, 2, 60, 0, 60, 0, 60, 0, 183, 1, 60, 0, 60, 0, 60, 0, 190, 29, 60, 0, 60, 0, 239, 1, 60, 0, 60, 0, 60, 0, 238, 1, 60, 0, 60, 0, 35, 29, 60, 0, 60, 0, 185, 1, 60, 0, 60, 0, 60, 0, 184, 1, 60, 0, 60, 0, 154, 29, 60, 0, 60, 0, 186, 1, 60, 0, 60, 0, 147, 2, 38, 0, 254, 0, 60, 0, 60, 0, 101, 167, 60, 0, 60, 0, 60, 0, 100, 167, 60, 0, 60, 0, 103, 167, 60, 0, 60, 0, 60, 0, 102, 167, 38, 0, 97, 0, 101, 0, 60, 0, 60, 0, 1, 29, 60, 0, 60, 0, 2, 29, 60, 0, 60, 0, 60, 0, 70, 29, 38, 0, 100, 0, 145, 2, 60, 0, 60, 0, 165, 2, 38, 0, 100, 0, 146, 2, 60, 0, 60, 0, 164, 2, 38, 0, 104, 0, 118, 0, 60, 0, 60, 0, 149, 1, 60, 0, 60, 0, 60, 0, 246, 1, 38, 0, 108, 0, 146, 2, 60, 0, 60, 0, 110, 2, 38, 0, 111, 0, 101, 0, 60, 0, 60, 0, 118, 2, 60, 0, 60, 0, 20, 29, 38, 0, 116, 0, 85, 2, 60, 0, 60, 0, 168, 2, 38, 0, 117, 0, 101, 0, 60, 0, 60, 0, 107, 29, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 116, 0, 101, 0, 114, 0, 116, 0, 105, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 61, 0, 187, 2, 61, 0, 189, 2, 61, 0, 105, 167, 61, 0, 104, 167, 61, 0, 107, 167, 61, 0, 106, 167, 61, 0, 109, 167, 61, 0, 108, 167, 61, 0, 111, 167, 61, 0, 210, 29, 61, 0, 110, 167, 61, 0, 112, 167, 61, 0, 120, 167, 61, 0, 187, 1, 61, 0, 43, 167, 61, 0, 42, 167, 61, 0, 45, 167, 61, 0, 44, 167, 61, 0, 47, 167, 61, 0, 46, 167, 61, 0, 168, 1, 61, 0, 167, 1, 61, 0, 189, 1, 61, 0, 188, 1, 61, 0, 133, 1, 61, 0, 132, 1, 61, 0, 148, 2, 61, 0, 66, 2, 61, 0, 65, 2, 61, 0, 192, 2, 61, 0, 188, 2, 61, 0, 238, 2, 61, 0, 190, 2, 61, 0, 35, 167, 61, 0, 34, 167, 61, 0, 140, 167, 61, 0, 139, 167, 61, 0, 149, 2, 61, 0, 228, 2, 61, 0, 191, 2, 61, 0, 193, 2, 61, 0, 36, 29, 61, 0, 37, 29, 61, 0, 92, 29, 61, 0, 37, 167, 61, 0, 36, 167, 61, 0, 161, 2, 61, 0, 162, 2, 61, 0, 150, 2, 61, 0, 192, 1, 61, 0, 193, 1, 61, 0, 194, 1, 61, 0, 195, 1, 61, 0, 151, 2, 61, 0, 152, 2, 61, 0, 172, 2, 61, 0, 173, 2, 38, 0, 179, 3, 60, 0, 60, 0, 38, 29, 38, 0, 187, 3, 60, 0, 60, 0, 39, 29, 38, 0, 192, 3, 60, 0, 60, 0, 40, 29, 38, 0, 193, 3, 60, 0, 60, 0, 41, 29, 38, 0, 200, 3, 60, 0, 60, 0, 42, 29, 38, 0, 82, 4, 60, 0, 60, 0, 83, 4, 60, 0, 60, 0, 60, 0, 3, 4, 38, 0, 57, 4, 60, 0, 139, 4, 60, 0, 60, 0, 60, 0, 138, 4, 38, 0, 59, 4, 60, 0, 60, 0, 43, 29, 38, 0, 91, 4, 60, 0, 60, 0, 92, 4, 60, 0, 60, 0, 60, 0, 12, 4, 38, 0, 132, 5, 60, 0, 135, 5, 0, 0, 239, 223, 180, 51, 38, 0, 32, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 112, 0, 114, 0, 105, 0, 109, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 60, 0, 60, 0, 42, 0, 62, 216, 176, 221, 62, 216, 177, 221, 62, 216, 179, 221, 62, 216, 178, 221, 60, 216, 251, 223, 60, 216, 252, 223, 60, 216, 253, 223, 60, 216, 254, 223, 60, 216, 255, 223, 38, 0, 32, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 209, 253, 172, 32, 60, 0, 42, 0, 61, 216, 0, 222, 61, 216, 3, 222, 61, 216, 4, 222, 61, 216, 1, 222, 61, 216, 6, 222, 61, 216, 5, 222, 62, 216, 35, 221, 61, 216, 2, 222, 61, 216, 66, 222, 61, 216, 67, 222, 61, 216, 9, 222, 61, 216, 10, 222, 61, 216, 7, 222, 60, 0, 42, 0, 62, 216, 112, 221, 61, 216, 13, 222, 62, 216, 41, 221, 61, 216, 24, 222, 61, 216, 23, 222, 58, 38, 61, 216, 26, 222, 61, 216, 25, 222, 62, 216, 114, 221, 60, 0, 42, 0, 61, 216, 11, 222, 61, 216, 27, 222, 61, 216, 28, 222, 62, 216, 42, 221, 61, 216, 29, 222, 62, 216, 17, 221, 60, 0, 42, 0, 62, 216, 23, 221, 62, 216, 45, 221, 62, 216, 43, 221, 62, 216, 20, 221, 60, 0, 42, 0, 62, 216, 16, 221, 62, 216, 40, 221, 61, 216, 16, 222, 61, 216, 17, 222, 61, 216, 54, 222, 61, 216, 15, 222, 61, 216, 18, 222, 61, 216, 68, 222, 61, 216, 44, 222, 62, 216, 37, 221, 60, 0, 42, 0, 61, 216, 12, 222, 61, 216, 20, 222, 61, 216, 42, 222, 62, 216, 36, 221, 61, 216, 52, 222, 60, 0, 42, 0, 61, 216, 55, 222, 62, 216, 18, 221, 62, 216, 21, 221, 62, 216, 34, 221, 62, 216, 46, 221, 62, 216, 39, 221, 62, 216, 117, 221, 62, 216, 118, 221, 62, 216, 116, 221, 61, 216, 53, 222, 62, 216, 47, 221, 60, 0, 42, 0, 62, 216, 32, 221, 62, 216, 115, 221, 62, 216, 120, 221, 60, 0, 42, 0, 61, 216, 14, 222, 62, 216, 19, 221, 62, 216, 208, 221, 60, 0, 42, 0, 61, 216, 21, 222, 61, 216, 31, 222, 61, 216, 65, 222, 57, 38, 61, 216, 46, 222, 61, 216, 47, 222, 61, 216, 50, 222, 61, 216, 51, 222, 62, 216, 122, 221, 61, 216, 38, 222, 61, 216, 39, 222, 61, 216, 40, 222, 61, 216, 48, 222, 61, 216, 37, 222, 61, 216, 34, 222, 61, 216, 45, 222, 61, 216, 49, 222, 61, 216, 22, 222, 61, 216, 35, 222, 61, 216, 30, 222, 61, 216, 19, 222, 61, 216, 41, 222, 61, 216, 43, 222, 62, 216, 113, 221, 60, 0, 42, 0, 61, 216, 36, 222, 61, 216, 33, 222, 61, 216, 32, 222, 62, 216, 44, 221, 61, 216, 8, 222, 61, 216, 127, 220, 61, 216, 128, 220, 32, 38, 60, 0, 42, 0, 61, 216, 169, 220, 62, 216, 33, 221, 61, 216, 121, 220, 61, 216, 122, 220, 61, 216, 123, 220, 61, 216, 125, 220, 61, 216, 126, 220, 62, 216, 22, 221, 60, 0, 42, 0, 61, 216, 58, 222, 61, 216, 56, 222, 61, 216, 57, 222, 61, 216, 59, 222, 61, 216, 60, 222, 61, 216, 61, 222, 61, 216, 64, 222, 61, 216, 63, 222, 61, 216, 62, 222, 60, 0, 42, 0, 61, 216, 72, 222, 61, 216, 73, 222, 61, 216, 74, 222, 60, 0, 42, 0, 61, 216, 139, 220, 61, 216, 140, 220, 61, 216, 152, 220, 61, 216, 157, 220, 61, 216, 150, 220, 61, 216, 151, 220, 61, 216, 147, 220, 61, 216, 158, 220, 61, 216, 149, 220, 61, 216, 159, 220, 99, 39, 61, 216, 148, 220, 100, 39, 62, 216, 225, 221, 61, 216, 155, 220, 61, 216, 154, 220, 61, 216, 153, 220, 61, 216, 156, 220, 62, 216, 14, 221, 61, 216, 164, 221, 62, 216, 13, 221, 61, 216, 175, 220, 61, 216, 162, 220, 61, 216, 165, 220, 61, 216, 171, 220, 61, 216, 166, 220, 61, 216, 168, 220, 61, 216, 115, 221, 61, 216, 163, 220, 61, 216, 172, 220, 60, 0, 32, 0, 61, 216, 65, 220, 13, 32, 61, 216, 232, 221, 32, 0, 61, 0, 32, 0, 61, 216, 65, 220, 15, 254, 13, 32, 61, 216, 232, 221, 60, 0, 42, 0, 61, 216, 232, 221, 61, 216, 239, 221, 61, 216, 173, 220, 61, 216, 164, 220, 60, 0, 42, 0, 61, 216, 75, 220, 62, 216, 26, 221, 61, 216, 144, 221, 11, 39, 61, 216, 150, 221, 60, 0, 42, 0, 61, 216, 76, 220, 62, 216, 12, 221, 62, 216, 15, 221, 12, 39, 62, 216, 30, 221, 62, 216, 31, 221, 62, 216, 24, 221, 62, 216, 25, 221, 60, 0, 42, 0, 61, 216, 72, 220, 61, 216, 73, 220, 61, 216, 70, 220, 61, 216, 149, 221, 61, 216, 71, 220, 29, 38, 60, 0, 42, 0, 61, 216, 77, 220, 61, 216, 78, 220, 10, 39, 61, 216, 74, 220, 62, 216, 27, 221, 62, 216, 28, 221, 60, 0, 42, 0, 61, 216, 79, 220, 61, 216, 76, 222, 61, 216, 80, 220, 62, 216, 50, 221, 62, 216, 29, 221, 61, 216, 79, 222, 60, 0, 42, 0, 13, 39, 61, 216, 133, 220, 62, 216, 51, 221, 60, 0, 42, 0, 61, 216, 170, 220, 62, 216, 190, 221, 62, 216, 191, 221, 62, 216, 181, 221, 62, 216, 182, 221, 61, 216, 66, 220, 62, 216, 187, 221, 61, 216, 67, 220, 62, 216, 224, 221, 62, 216, 192, 222, 62, 216, 193, 222, 62, 216, 183, 221, 62, 216, 180, 221, 61, 216, 64, 220, 61, 216, 65, 220, 61, 216, 69, 220, 61, 216, 68, 220, 60, 0, 42, 0, 61, 216, 118, 220, 62, 216, 210, 221, 61, 216, 102, 220, 61, 216, 103, 220, 62, 216, 209, 221, 61, 216, 113, 220, 61, 216, 104, 220, 62, 216, 212, 221, 61, 216, 105, 220, 60, 0, 32, 0, 61, 216, 113, 220, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 32, 0, 61, 216, 113, 220, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 42, 0, 62, 216, 211, 221, 61, 216, 116, 220, 61, 216, 117, 220, 60, 0, 42, 0, 61, 216, 77, 222, 60, 0, 32, 0, 61, 216, 77, 222, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 77, 222, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 77, 222, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 77, 222, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 77, 222, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 77, 222, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 77, 222, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 77, 222, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 77, 222, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 77, 222, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 77, 222, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 77, 222, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 78, 222, 60, 0, 32, 0, 61, 216, 78, 222, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 78, 222, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 78, 222, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 78, 222, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 78, 222, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 78, 222, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 78, 222, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 78, 222, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 78, 222, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 78, 222, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 78, 222, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 78, 222, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 69, 222, 60, 0, 32, 0, 61, 216, 69, 222, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 69, 222, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 69, 222, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 69, 222, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 69, 222, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 69, 222, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 69, 222, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 69, 222, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 69, 222, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 69, 222, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 69, 222, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 69, 222, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 70, 222, 60, 0, 32, 0, 61, 216, 70, 222, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 70, 222, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 70, 222, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 70, 222, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 70, 222, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 70, 222, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 70, 222, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 70, 222, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 70, 222, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 70, 222, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 70, 222, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 70, 222, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 129, 220, 60, 0, 32, 0, 61, 216, 129, 220, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 129, 220, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 129, 220, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 129, 220, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 129, 220, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 129, 220, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 129, 220, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 129, 220, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 129, 220, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 129, 220, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 129, 220, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 129, 220, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 75, 222, 60, 0, 32, 0, 61, 216, 75, 222, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 75, 222, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 75, 222, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 75, 222, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 75, 222, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 75, 222, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 75, 222, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 75, 222, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 75, 222, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 75, 222, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 75, 222, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 75, 222, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 207, 221, 60, 0, 32, 0, 62, 216, 207, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 207, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 207, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 207, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 207, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 207, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 207, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 207, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 207, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 207, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 207, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 207, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 71, 222, 60, 0, 32, 0, 61, 216, 71, 222, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 71, 222, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 71, 222, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 71, 222, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 71, 222, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 71, 222, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 71, 222, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 71, 222, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 71, 222, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 71, 222, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 71, 222, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 71, 222, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 38, 221, 60, 0, 32, 0, 62, 216, 38, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 38, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 38, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 38, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 38, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 38, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 38, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 38, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 38, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 38, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 38, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 38, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 55, 221, 60, 0, 32, 0, 62, 216, 55, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 55, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 55, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 55, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 55, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 55, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 55, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 55, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 55, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 55, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 55, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 55, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 149, 38, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 149, 38, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 149, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 149, 38, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 60, 216, 147, 223, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 60, 216, 147, 223, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 60, 216, 147, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 60, 216, 147, 223, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 60, 216, 235, 223, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 60, 216, 235, 223, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 60, 216, 235, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 60, 216, 235, 223, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 150, 38, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 150, 38, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 150, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 150, 38, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 60, 216, 62, 223, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 60, 216, 62, 223, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 60, 216, 62, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 60, 216, 62, 223, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 60, 216, 115, 223, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 60, 216, 115, 223, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 60, 216, 115, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 60, 216, 115, 223, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 61, 216, 39, 221, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 61, 216, 39, 221, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 61, 216, 39, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 61, 216, 39, 221, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 60, 216, 237, 223, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 60, 216, 237, 223, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 60, 216, 237, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 60, 216, 237, 223, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 61, 216, 188, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 61, 216, 188, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 61, 216, 188, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 61, 216, 188, 220, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 61, 216, 44, 221, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 61, 216, 44, 221, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 61, 216, 44, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 61, 216, 44, 221, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 61, 216, 187, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 61, 216, 187, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 61, 216, 187, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 61, 216, 187, 220, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 60, 216, 164, 223, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 60, 216, 164, 223, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 60, 216, 164, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 60, 216, 164, 223, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 60, 216, 168, 223, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 60, 216, 168, 223, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 60, 216, 168, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 60, 216, 168, 223, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 8, 39, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 8, 39, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 8, 39, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 8, 39, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 61, 216, 128, 222, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 61, 216, 128, 222, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 61, 216, 128, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 61, 216, 128, 222, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 61, 216, 146, 222, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 61, 216, 146, 222, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 61, 216, 146, 222, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 61, 216, 146, 222, 60, 0, 42, 0, 61, 216, 110, 220, 60, 0, 32, 0, 61, 216, 110, 220, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 110, 220, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 110, 220, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 110, 220, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 110, 220, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 110, 220, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 110, 220, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 110, 220, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 110, 220, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 110, 220, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 110, 220, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 110, 220, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 117, 221, 60, 0, 32, 0, 61, 216, 117, 221, 13, 32, 66, 38, 32, 0, 61, 0, 32, 0, 61, 216, 117, 221, 15, 254, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 117, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 117, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 117, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 117, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 117, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 117, 221, 13, 32, 64, 38, 32, 0, 61, 0, 32, 0, 61, 216, 117, 221, 15, 254, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 117, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 117, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 117, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 117, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 117, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 130, 220, 60, 0, 32, 0, 61, 216, 130, 220, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 130, 220, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 130, 220, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 130, 220, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 130, 220, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 130, 220, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 130, 220, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 130, 220, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 130, 220, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 130, 220, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 130, 220, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 130, 220, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 119, 221, 61, 216, 119, 220, 60, 0, 32, 0, 61, 216, 119, 220, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 119, 220, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 119, 220, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 119, 220, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 119, 220, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 119, 220, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 119, 220, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 119, 220, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 119, 220, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 119, 220, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 119, 220, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 119, 220, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 52, 221, 61, 216, 120, 220, 61, 216, 115, 220, 60, 0, 32, 0, 61, 216, 115, 220, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 115, 220, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 115, 220, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 115, 220, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 115, 220, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 115, 220, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 115, 220, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 115, 220, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 115, 220, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 115, 220, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 115, 220, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 115, 220, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 114, 220, 62, 216, 213, 221, 62, 216, 53, 221, 60, 0, 32, 0, 62, 216, 53, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 53, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 53, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 53, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 53, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 53, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 53, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 53, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 53, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 53, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 53, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 53, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 112, 220, 60, 0, 32, 0, 61, 216, 112, 220, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 112, 220, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 112, 220, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 112, 220, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 112, 220, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 112, 220, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 112, 220, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 112, 220, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 112, 220, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 112, 220, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 112, 220, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 112, 220, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 48, 221, 62, 216, 49, 221, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 60, 216, 124, 223, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 60, 216, 124, 223, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 60, 216, 124, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 60, 216, 124, 223, 60, 0, 42, 0, 61, 216, 124, 220, 60, 216, 133, 223, 62, 216, 54, 221, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 60, 216, 132, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 60, 216, 132, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 60, 216, 132, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 60, 216, 132, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 60, 216, 132, 223, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 60, 216, 132, 223, 60, 0, 42, 0, 62, 216, 184, 221, 60, 0, 32, 0, 62, 216, 184, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 184, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 184, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 184, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 184, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 184, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 184, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 184, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 184, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 184, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 184, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 184, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 185, 221, 60, 0, 32, 0, 62, 216, 185, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 185, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 185, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 185, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 185, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 185, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 185, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 185, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 185, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 185, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 185, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 185, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 217, 221, 60, 0, 32, 0, 62, 216, 217, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 217, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 217, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 217, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 217, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 217, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 217, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 217, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 217, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 217, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 217, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 217, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 218, 221, 60, 0, 32, 0, 62, 216, 218, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 218, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 218, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 218, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 218, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 218, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 218, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 218, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 218, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 218, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 218, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 218, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 219, 221, 60, 0, 32, 0, 62, 216, 219, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 219, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 219, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 219, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 219, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 219, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 219, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 219, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 219, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 219, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 219, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 219, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 220, 221, 60, 0, 32, 0, 62, 216, 220, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 220, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 220, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 220, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 220, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 220, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 220, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 220, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 220, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 220, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 220, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 220, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 221, 221, 60, 0, 32, 0, 62, 216, 221, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 221, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 221, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 221, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 221, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 221, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 221, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 221, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 221, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 221, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 221, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 221, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 222, 221, 60, 0, 32, 0, 62, 216, 222, 221, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 222, 221, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 223, 221, 60, 0, 32, 0, 62, 216, 223, 221, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 223, 221, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 134, 220, 60, 0, 32, 0, 61, 216, 134, 220, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 134, 220, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 134, 220, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 134, 220, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 134, 220, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 134, 220, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 134, 220, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 134, 220, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 134, 220, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 134, 220, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 134, 220, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 134, 220, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 135, 220, 60, 0, 32, 0, 61, 216, 135, 220, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 135, 220, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 135, 220, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 135, 220, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 135, 220, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 135, 220, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 135, 220, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 135, 220, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 135, 220, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 135, 220, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 135, 220, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 135, 220, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 182, 222, 60, 0, 32, 0, 61, 216, 182, 222, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 182, 222, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 182, 222, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 182, 222, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 182, 222, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 182, 222, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 182, 222, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 182, 222, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 182, 222, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 182, 222, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 182, 222, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 182, 222, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 205, 221, 60, 0, 32, 0, 62, 216, 205, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 205, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 205, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 205, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 205, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 205, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 205, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 205, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 205, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 205, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 205, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 205, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 206, 221, 60, 0, 32, 0, 62, 216, 206, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 206, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 206, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 206, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 206, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 206, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 206, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 206, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 206, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 206, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 206, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 206, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 62, 216, 175, 221, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 62, 216, 175, 221, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 62, 216, 175, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 62, 216, 175, 221, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 62, 216, 188, 221, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 62, 216, 188, 221, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 62, 216, 188, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 62, 216, 188, 221, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 62, 216, 189, 221, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 62, 216, 189, 221, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 62, 216, 189, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 62, 216, 189, 221, 60, 0, 42, 0, 60, 216, 195, 223, 60, 0, 32, 0, 60, 216, 195, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 195, 223, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 195, 223, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 195, 223, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 195, 223, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 195, 223, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 60, 216, 195, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 195, 223, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 195, 223, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 195, 223, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 195, 223, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 195, 223, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 131, 220, 61, 216, 122, 221, 61, 216, 116, 221, 61, 216, 111, 220, 60, 0, 32, 0, 61, 216, 111, 220, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 111, 220, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 214, 221, 60, 0, 32, 0, 62, 216, 214, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 214, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 214, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 214, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 214, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 214, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 214, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 214, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 214, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 214, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 214, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 214, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 215, 221, 60, 0, 32, 0, 62, 216, 215, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 215, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 215, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 215, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 215, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 215, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 215, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 215, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 215, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 215, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 215, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 215, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 58, 221, 60, 216, 199, 223, 247, 38, 60, 216, 194, 223, 60, 216, 204, 223, 60, 0, 32, 0, 60, 216, 204, 223, 13, 32, 66, 38, 32, 0, 61, 0, 32, 0, 60, 216, 204, 223, 15, 254, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 204, 223, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 204, 223, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 204, 223, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 204, 223, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 204, 223, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 60, 216, 204, 223, 13, 32, 64, 38, 32, 0, 61, 0, 32, 0, 60, 216, 204, 223, 15, 254, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 204, 223, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 204, 223, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 204, 223, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 204, 223, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 204, 223, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 60, 216, 196, 223, 60, 0, 32, 0, 60, 216, 196, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 196, 223, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 196, 223, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 196, 223, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 196, 223, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 196, 223, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 60, 216, 196, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 196, 223, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 196, 223, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 196, 223, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 196, 223, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 196, 223, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 163, 222, 60, 0, 32, 0, 61, 216, 163, 222, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 163, 222, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 163, 222, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 163, 222, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 163, 222, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 163, 222, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 163, 222, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 163, 222, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 163, 222, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 163, 222, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 163, 222, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 163, 222, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 60, 216, 202, 223, 60, 0, 32, 0, 60, 216, 202, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 202, 223, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 202, 223, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 202, 223, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 202, 223, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 202, 223, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 60, 216, 202, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 202, 223, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 202, 223, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 202, 223, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 202, 223, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 202, 223, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 249, 38, 60, 0, 32, 0, 249, 38, 13, 32, 66, 38, 32, 0, 61, 0, 32, 0, 249, 38, 15, 254, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 249, 38, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 249, 38, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 249, 38, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 249, 38, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 249, 38, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 249, 38, 13, 32, 64, 38, 32, 0, 61, 0, 32, 0, 249, 38, 15, 254, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 249, 38, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 249, 38, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 249, 38, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 249, 38, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 249, 38, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 60, 216, 203, 223, 60, 0, 32, 0, 60, 216, 203, 223, 13, 32, 66, 38, 32, 0, 61, 0, 32, 0, 60, 216, 203, 223, 15, 254, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 203, 223, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 203, 223, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 203, 223, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 203, 223, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 203, 223, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 60, 216, 203, 223, 13, 32, 64, 38, 32, 0, 61, 0, 32, 0, 60, 216, 203, 223, 15, 254, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 203, 223, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 203, 223, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 203, 223, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 203, 223, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 60, 216, 203, 223, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 180, 222, 60, 0, 32, 0, 61, 216, 180, 222, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 180, 222, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 180, 222, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 180, 222, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 180, 222, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 180, 222, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 180, 222, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 180, 222, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 180, 222, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 180, 222, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 180, 222, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 180, 222, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 181, 222, 60, 0, 32, 0, 61, 216, 181, 222, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 181, 222, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 181, 222, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 181, 222, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 181, 222, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 181, 222, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 61, 216, 181, 222, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 181, 222, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 181, 222, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 181, 222, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 181, 222, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 181, 222, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 56, 221, 60, 0, 32, 0, 62, 216, 56, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 56, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 56, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 56, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 56, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 56, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 56, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 56, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 56, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 56, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 56, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 56, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 60, 221, 60, 0, 32, 0, 62, 216, 60, 221, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 60, 221, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 61, 221, 60, 0, 32, 0, 62, 216, 61, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 61, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 61, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 61, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 61, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 61, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 61, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 61, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 61, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 61, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 61, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 61, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 62, 221, 60, 0, 32, 0, 62, 216, 62, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 62, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 62, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 62, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 62, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 62, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 62, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 62, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 62, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 62, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 62, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 62, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 57, 221, 60, 0, 32, 0, 62, 216, 57, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 57, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 57, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 57, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 57, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 57, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 57, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 57, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 57, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 57, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 57, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 57, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 62, 216, 216, 221, 60, 0, 32, 0, 62, 216, 216, 221, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 216, 221, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 216, 221, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 216, 221, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 216, 221, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 216, 221, 60, 216, 255, 223, 13, 32, 66, 38, 60, 0, 32, 0, 62, 216, 216, 221, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 216, 221, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 216, 221, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 216, 221, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 216, 221, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 216, 221, 60, 216, 255, 223, 13, 32, 64, 38, 60, 0, 42, 0, 61, 216, 192, 222, 61, 216, 204, 222, 60, 0, 32, 0, 62, 216, 209, 221, 13, 32, 62, 216, 29, 221, 13, 32, 62, 216, 209, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 251, 223, 13, 32, 62, 216, 29, 221, 13, 32, 62, 216, 209, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 252, 223, 13, 32, 62, 216, 29, 221, 13, 32, 62, 216, 209, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 253, 223, 13, 32, 62, 216, 29, 221, 13, 32, 62, 216, 209, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 254, 223, 13, 32, 62, 216, 29, 221, 13, 32, 62, 216, 209, 221, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 209, 221, 60, 216, 255, 223, 13, 32, 62, 216, 29, 221, 13, 32, 62, 216, 209, 221, 60, 0, 32, 0, 61, 216, 109, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 105, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 105, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 105, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 105, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 105, 220, 60, 0, 32, 0, 61, 216, 107, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 0, 32, 0, 61, 216, 108, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 38, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 105, 220, 60, 216, 251, 223, 61, 0, 61, 216, 109, 220, 60, 216, 251, 223, 38, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 105, 220, 60, 216, 252, 223, 61, 0, 61, 216, 109, 220, 60, 216, 252, 223, 38, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 105, 220, 60, 216, 253, 223, 61, 0, 61, 216, 109, 220, 60, 216, 253, 223, 38, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 105, 220, 60, 216, 254, 223, 61, 0, 61, 216, 109, 220, 60, 216, 254, 223, 38, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 105, 220, 60, 216, 255, 223, 61, 0, 61, 216, 109, 220, 60, 216, 255, 223, 38, 0, 32, 0, 61, 216, 105, 220, 60, 216, 251, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 216, 251, 223, 61, 0, 61, 216, 107, 220, 60, 216, 251, 223, 38, 0, 32, 0, 61, 216, 105, 220, 60, 216, 252, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 216, 252, 223, 61, 0, 61, 216, 107, 220, 60, 216, 252, 223, 38, 0, 32, 0, 61, 216, 105, 220, 60, 216, 253, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 216, 253, 223, 61, 0, 61, 216, 107, 220, 60, 216, 253, 223, 38, 0, 32, 0, 61, 216, 105, 220, 60, 216, 254, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 216, 254, 223, 61, 0, 61, 216, 107, 220, 60, 216, 254, 223, 38, 0, 32, 0, 61, 216, 105, 220, 60, 216, 255, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 216, 255, 223, 61, 0, 61, 216, 107, 220, 60, 216, 255, 223, 38, 0, 32, 0, 61, 216, 104, 220, 60, 216, 251, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 216, 251, 223, 61, 0, 61, 216, 108, 220, 60, 216, 251, 223, 38, 0, 32, 0, 61, 216, 104, 220, 60, 216, 252, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 216, 252, 223, 61, 0, 61, 216, 108, 220, 60, 216, 252, 223, 38, 0, 32, 0, 61, 216, 104, 220, 60, 216, 253, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 216, 253, 223, 61, 0, 61, 216, 108, 220, 60, 216, 253, 223, 38, 0, 32, 0, 61, 216, 104, 220, 60, 216, 254, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 216, 254, 223, 61, 0, 61, 216, 108, 220, 60, 216, 254, 223, 38, 0, 32, 0, 61, 216, 104, 220, 60, 216, 255, 223, 13, 32, 62, 216, 29, 221, 13, 32, 61, 216, 104, 220, 60, 216, 255, 223, 61, 0, 61, 216, 108, 220, 60, 216, 255, 223, 60, 0, 42, 0, 61, 216, 143, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 100, 39, 13, 32, 61, 216, 139, 220, 13, 32, 61, 216, 104, 220, 32, 0, 61, 0, 32, 0, 61, 216, 105, 220, 13, 32, 100, 39, 15, 254, 13, 32, 61, 216, 139, 220, 13, 32, 61, 216, 104, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 100, 39, 13, 32, 61, 216, 139, 220, 13, 32, 61, 216, 104, 220, 32, 0, 61, 0, 32, 0, 61, 216, 104, 220, 13, 32, 100, 39, 15, 254, 13, 32, 61, 216, 139, 220, 13, 32, 61, 216, 104, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 100, 39, 13, 32, 61, 216, 139, 220, 13, 32, 61, 216, 105, 220, 32, 0, 61, 0, 32, 0, 61, 216, 105, 220, 13, 32, 100, 39, 15, 254, 13, 32, 61, 216, 139, 220, 13, 32, 61, 216, 105, 220, 60, 0, 42, 0, 61, 216, 145, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 100, 39, 13, 32, 61, 216, 104, 220, 32, 0, 61, 0, 32, 0, 61, 216, 105, 220, 13, 32, 100, 39, 15, 254, 13, 32, 61, 216, 104, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 100, 39, 13, 32, 61, 216, 104, 220, 32, 0, 61, 0, 32, 0, 61, 216, 104, 220, 13, 32, 100, 39, 15, 254, 13, 32, 61, 216, 104, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 100, 39, 13, 32, 61, 216, 105, 220, 32, 0, 61, 0, 32, 0, 61, 216, 105, 220, 13, 32, 100, 39, 15, 254, 13, 32, 61, 216, 105, 220, 60, 0, 42, 0, 61, 216, 106, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 105, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 105, 220, 13, 32, 61, 216, 103, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 105, 220, 13, 32, 61, 216, 103, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 105, 220, 13, 32, 61, 216, 102, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 105, 220, 13, 32, 61, 216, 103, 220, 13, 32, 61, 216, 103, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 104, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 104, 220, 13, 32, 61, 216, 103, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 104, 220, 13, 32, 61, 216, 103, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 104, 220, 13, 32, 61, 216, 102, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 104, 220, 13, 32, 61, 216, 103, 220, 13, 32, 61, 216, 103, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 105, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 105, 220, 13, 32, 61, 216, 103, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 105, 220, 13, 32, 61, 216, 103, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 105, 220, 13, 32, 61, 216, 102, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 105, 220, 13, 32, 61, 216, 103, 220, 13, 32, 61, 216, 103, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 102, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 103, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 103, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 104, 220, 13, 32, 61, 216, 103, 220, 13, 32, 61, 216, 103, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 102, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 103, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 103, 220, 13, 32, 61, 216, 102, 220, 60, 0, 32, 0, 61, 216, 105, 220, 13, 32, 61, 216, 103, 220, 13, 32, 61, 216, 103, 220, 60, 0, 42, 0, 61, 216, 227, 221, 61, 216, 100, 220, 61, 216, 101, 220, 62, 216, 194, 222, 61, 216, 99, 220, 60, 0, 42, 0, 62, 216, 176, 221, 62, 216, 177, 221, 62, 216, 179, 221, 62, 216, 178, 221, 60, 0, 42, 0, 61, 216, 53, 220, 61, 216, 18, 220, 62, 216, 141, 221, 62, 216, 167, 221, 61, 216, 54, 220, 61, 216, 21, 220, 62, 216, 174, 221, 60, 0, 32, 0, 61, 216, 21, 220, 13, 32, 62, 216, 186, 221, 60, 0, 42, 0, 61, 216, 41, 220, 61, 216, 58, 220, 62, 216, 138, 221, 62, 216, 157, 221, 61, 216, 49, 220, 61, 216, 8, 220, 60, 0, 32, 0, 61, 216, 8, 220, 13, 32, 27, 43, 60, 0, 42, 0, 62, 216, 129, 221, 61, 216, 47, 220, 61, 216, 5, 220, 61, 216, 6, 220, 61, 216, 52, 220, 61, 216, 14, 220, 62, 216, 132, 221, 62, 216, 147, 221, 62, 216, 140, 221, 62, 216, 172, 221, 61, 216, 46, 220, 61, 216, 2, 220, 61, 216, 3, 220, 61, 216, 4, 220, 61, 216, 55, 220, 61, 216, 22, 220, 61, 216, 23, 220, 61, 216, 61, 220, 61, 216, 15, 220, 61, 216, 17, 220, 61, 216, 16, 220, 61, 216, 42, 220, 61, 216, 43, 220, 62, 216, 153, 221, 62, 216, 146, 221, 61, 216, 24, 220, 62, 216, 163, 221, 62, 216, 143, 221, 62, 216, 155, 221, 61, 216, 45, 220, 61, 216, 1, 220, 61, 216, 0, 220, 61, 216, 57, 220, 61, 216, 48, 220, 61, 216, 7, 220, 61, 216, 63, 220, 62, 216, 171, 221, 62, 216, 148, 221, 62, 216, 135, 221, 61, 216, 59, 220, 60, 0, 32, 0, 61, 216, 59, 220, 13, 32, 68, 39, 60, 0, 42, 0, 61, 216, 40, 220, 61, 216, 60, 220, 62, 216, 165, 221, 62, 216, 166, 221, 62, 216, 168, 221, 62, 216, 152, 221, 62, 216, 161, 221, 61, 216, 62, 220, 60, 0, 42, 0, 62, 216, 131, 221, 61, 216, 20, 220, 61, 216, 19, 220, 61, 216, 35, 220, 61, 216, 36, 220, 61, 216, 37, 220, 61, 216, 38, 220, 61, 216, 39, 220, 61, 216, 74, 221, 62, 216, 133, 221, 62, 216, 134, 221, 62, 216, 162, 221, 62, 216, 137, 221, 62, 216, 164, 221, 62, 216, 182, 222, 62, 216, 169, 221, 62, 216, 154, 221, 62, 216, 156, 221, 60, 0, 42, 0, 61, 216, 56, 220, 60, 0, 42, 0, 61, 216, 10, 220, 61, 216, 34, 220, 62, 216, 142, 221, 61, 216, 13, 220, 61, 216, 50, 220, 61, 216, 9, 220, 62, 216, 149, 221, 62, 216, 150, 221, 60, 0, 42, 0, 61, 216, 51, 220, 61, 216, 11, 220, 61, 216, 44, 220, 62, 216, 173, 221, 61, 216, 31, 220, 61, 216, 32, 220, 61, 216, 33, 220, 62, 216, 136, 221, 61, 216, 25, 220, 61, 216, 26, 220, 60, 0, 42, 0, 61, 216, 12, 220, 62, 216, 139, 221, 61, 216, 27, 220, 61, 216, 28, 220, 61, 216, 29, 220, 62, 216, 178, 222, 61, 216, 30, 220, 62, 216, 151, 221, 62, 216, 179, 222, 61, 216, 119, 221, 61, 216, 120, 221, 62, 216, 130, 221, 62, 216, 159, 221, 62, 216, 176, 222, 62, 216, 177, 222, 62, 216, 160, 221, 60, 0, 42, 0, 61, 216, 144, 220, 60, 216, 56, 223, 61, 216, 174, 220, 60, 216, 245, 223, 60, 216, 57, 223, 62, 216, 64, 221, 60, 216, 58, 223, 60, 216, 59, 223, 60, 216, 60, 223, 60, 216, 55, 223, 60, 0, 42, 0, 60, 216, 49, 223, 62, 216, 180, 222, 60, 216, 50, 223, 60, 216, 51, 223, 60, 216, 52, 223, 60, 216, 53, 223, 60, 216, 62, 223, 60, 216, 63, 223, 24, 38, 60, 216, 64, 223, 60, 216, 65, 223, 60, 216, 66, 223, 60, 216, 67, 223, 60, 0, 42, 0, 60, 216, 71, 223, 60, 216, 72, 223, 60, 216, 73, 223, 60, 216, 74, 223, 60, 216, 75, 223, 60, 216, 76, 223, 60, 216, 77, 223, 62, 216, 109, 221, 60, 216, 78, 223, 60, 216, 79, 223, 60, 216, 80, 223, 60, 216, 81, 223, 60, 216, 82, 223, 60, 216, 83, 223, 62, 216, 208, 222, 62, 216, 93, 221, 60, 216, 69, 223, 62, 216, 210, 222, 62, 216, 101, 221, 60, 0, 42, 0, 62, 216, 81, 221, 60, 216, 70, 223, 62, 216, 84, 221, 62, 216, 85, 221, 60, 216, 61, 223, 60, 216, 54, 223, 62, 216, 209, 222, 62, 216, 82, 221, 62, 216, 108, 221, 62, 216, 102, 221, 62, 216, 196, 221, 62, 216, 197, 221, 60, 216, 68, 223, 62, 216, 92, 221, 60, 216, 48, 223, 60, 0, 42, 0, 60, 216, 94, 223, 62, 216, 80, 221, 62, 216, 86, 221, 62, 216, 211, 222, 62, 216, 104, 221, 62, 216, 111, 221, 62, 216, 94, 221, 62, 216, 199, 221, 62, 216, 192, 221, 60, 216, 86, 223, 60, 216, 87, 223, 62, 216, 105, 221, 62, 216, 83, 221, 60, 216, 84, 223, 60, 216, 95, 223, 60, 216, 85, 223, 60, 216, 45, 223, 62, 216, 106, 221, 60, 216, 46, 223, 60, 216, 47, 223, 62, 216, 212, 222, 62, 216, 89, 221, 62, 216, 198, 221, 62, 216, 90, 221, 60, 216, 115, 223, 62, 216, 88, 221, 60, 216, 114, 223, 62, 216, 213, 222, 62, 216, 99, 221, 62, 216, 87, 221, 60, 216, 127, 223, 62, 216, 200, 221, 62, 216, 194, 221, 62, 216, 107, 221, 60, 0, 42, 0, 60, 216, 113, 223, 60, 216, 88, 223, 60, 216, 89, 223, 60, 216, 90, 223, 60, 216, 91, 223, 60, 216, 92, 223, 60, 216, 93, 223, 60, 216, 96, 223, 60, 216, 98, 223, 60, 216, 99, 223, 60, 216, 100, 223, 60, 216, 101, 223, 62, 216, 110, 221, 60, 216, 97, 223, 62, 216, 95, 221, 62, 216, 96, 221, 62, 216, 97, 221, 60, 0, 42, 0, 62, 216, 128, 221, 62, 216, 158, 221, 62, 216, 144, 221, 62, 216, 145, 221, 62, 216, 170, 221, 60, 0, 42, 0, 60, 216, 102, 223, 60, 216, 103, 223, 60, 216, 104, 223, 60, 216, 105, 223, 60, 216, 106, 223, 60, 216, 130, 223, 60, 216, 112, 223, 62, 216, 193, 221, 62, 216, 103, 221, 60, 216, 107, 223, 60, 216, 108, 223, 60, 216, 109, 223, 60, 216, 110, 223, 60, 216, 111, 223, 60, 0, 42, 0, 60, 216, 124, 223, 62, 216, 91, 221, 21, 38, 62, 216, 214, 222, 60, 216, 117, 223, 60, 216, 118, 223, 60, 216, 126, 223, 60, 216, 119, 223, 60, 216, 120, 223, 60, 216, 121, 223, 60, 216, 122, 223, 60, 216, 123, 223, 62, 216, 66, 221, 62, 216, 67, 221, 62, 216, 100, 221, 62, 216, 203, 221, 62, 216, 195, 221, 62, 216, 201, 221, 62, 216, 202, 221, 60, 0, 42, 0, 62, 216, 98, 221, 60, 216, 125, 223, 60, 216, 116, 223, 62, 216, 68, 221, 61, 216, 42, 221, 60, 216, 250, 223, 60, 0, 42, 0, 60, 216, 13, 223, 60, 216, 14, 223, 60, 216, 15, 223, 60, 216, 16, 223, 61, 216, 250, 221, 61, 216, 254, 221, 62, 216, 237, 221, 60, 0, 42, 0, 60, 216, 212, 223, 240, 38, 60, 216, 11, 223, 61, 216, 251, 221, 60, 216, 213, 223, 60, 216, 214, 223, 60, 216, 220, 223, 60, 216, 221, 223, 60, 216, 222, 223, 60, 0, 42, 0, 60, 216, 223, 223, 60, 216, 219, 223, 60, 216, 215, 223, 62, 216, 241, 221, 62, 216, 168, 222, 62, 216, 181, 222, 61, 216, 214, 222, 60, 216, 216, 223, 60, 216, 218, 223, 60, 216, 224, 223, 60, 216, 225, 223, 60, 216, 226, 223, 60, 216, 227, 223, 60, 216, 228, 223, 60, 216, 229, 223, 60, 216, 230, 223, 60, 216, 232, 223, 60, 216, 233, 223, 60, 216, 234, 223, 60, 216, 235, 223, 60, 216, 236, 223, 60, 216, 237, 223, 60, 216, 239, 223, 60, 216, 240, 223, 61, 216, 146, 220, 61, 216, 252, 221, 61, 216, 253, 221, 60, 0, 42, 0, 234, 38, 61, 216, 76, 221, 61, 216, 213, 222, 61, 216, 77, 221, 233, 38, 61, 216, 75, 221, 60, 0, 42, 0, 242, 38, 250, 38, 60, 216, 1, 223, 60, 216, 3, 223, 60, 216, 217, 223, 60, 216, 4, 223, 60, 216, 5, 223, 60, 216, 6, 223, 60, 216, 7, 223, 60, 216, 9, 223, 104, 38, 60, 216, 160, 223, 60, 216, 161, 223, 60, 216, 162, 223, 61, 216, 136, 220, 60, 216, 170, 223, 60, 0, 42, 0, 61, 216, 130, 222, 61, 216, 131, 222, 61, 216, 132, 222, 61, 216, 133, 222, 61, 216, 134, 222, 61, 216, 135, 222, 61, 216, 136, 222, 61, 216, 137, 222, 61, 216, 138, 222, 61, 216, 157, 222, 61, 216, 158, 222, 61, 216, 139, 222, 61, 216, 140, 222, 61, 216, 141, 222, 61, 216, 142, 222, 61, 216, 144, 222, 61, 216, 145, 222, 61, 216, 146, 222, 61, 216, 147, 222, 61, 216, 148, 222, 61, 216, 149, 222, 61, 216, 150, 222, 61, 216, 151, 222, 61, 216, 152, 222, 61, 216, 153, 222, 61, 216, 251, 222, 61, 216, 154, 222, 61, 216, 155, 222, 61, 216, 156, 222, 60, 216, 206, 223, 60, 216, 205, 223, 61, 216, 245, 222, 62, 216, 189, 221, 62, 216, 188, 221, 61, 216, 250, 222, 61, 216, 178, 222, 61, 216, 244, 222, 61, 216, 249, 222, 61, 216, 252, 222, 61, 216, 143, 222, 61, 216, 227, 222, 61, 216, 228, 222, 61, 216, 226, 222, 253, 38, 61, 216, 168, 222, 61, 216, 165, 222, 61, 216, 166, 222, 61, 216, 209, 222, 61, 216, 167, 222, 60, 0, 42, 0, 147, 38, 245, 38, 61, 216, 246, 222, 61, 216, 164, 222, 61, 216, 243, 222, 244, 38, 61, 216, 229, 222, 61, 216, 162, 222, 60, 0, 42, 0, 8, 39, 61, 216, 233, 222, 61, 216, 235, 222, 61, 216, 236, 222, 62, 216, 130, 222, 61, 216, 186, 220, 61, 216, 129, 222, 61, 216, 159, 222, 61, 216, 160, 222, 61, 216, 161, 222, 61, 216, 240, 222, 61, 216, 128, 222, 61, 216, 248, 222, 60, 0, 42, 0, 61, 216, 206, 222, 62, 216, 243, 221, 60, 0, 42, 0, 27, 35, 243, 35, 26, 35, 240, 35, 241, 35, 242, 35, 61, 216, 112, 221, 61, 216, 91, 221, 61, 216, 103, 221, 61, 216, 80, 221, 61, 216, 92, 221, 61, 216, 81, 221, 61, 216, 93, 221, 61, 216, 82, 221, 61, 216, 94, 221, 61, 216, 83, 221, 61, 216, 95, 221, 61, 216, 84, 221, 61, 216, 96, 221, 61, 216, 85, 221, 61, 216, 97, 221, 61, 216, 86, 221, 61, 216, 98, 221, 61, 216, 87, 221, 61, 216, 99, 221, 61, 216, 88, 221, 61, 216, 100, 221, 61, 216, 89, 221, 61, 216, 101, 221, 61, 216, 90, 221, 61, 216, 102, 221, 60, 0, 42, 0, 60, 216, 17, 223, 60, 216, 18, 223, 60, 216, 19, 223, 60, 216, 20, 223, 60, 216, 21, 223, 60, 216, 22, 223, 60, 216, 23, 223, 60, 216, 24, 223, 60, 216, 25, 223, 60, 216, 26, 223, 60, 216, 27, 223, 60, 216, 28, 223, 60, 216, 33, 223, 0, 38, 60, 216, 29, 223, 60, 216, 30, 223, 62, 216, 144, 222, 80, 43, 60, 216, 31, 223, 60, 216, 32, 223, 60, 216, 12, 223, 1, 38, 197, 38, 200, 38, 60, 216, 36, 223, 60, 216, 37, 223, 60, 216, 38, 223, 60, 216, 39, 223, 60, 216, 40, 223, 60, 216, 41, 223, 60, 216, 42, 223, 60, 216, 43, 223, 60, 216, 44, 223, 60, 216, 0, 223, 60, 216, 8, 223, 60, 216, 2, 223, 2, 38, 20, 38, 241, 38, 161, 38, 68, 39, 3, 38, 196, 38, 4, 38, 61, 216, 37, 221, 61, 216, 167, 220, 60, 216, 10, 223, 60, 0, 42, 0, 60, 216, 131, 223, 60, 216, 132, 223, 60, 216, 134, 223, 60, 216, 135, 223, 62, 216, 232, 221, 40, 39, 60, 216, 136, 223, 60, 216, 137, 223, 60, 216, 138, 223, 60, 216, 139, 223, 60, 216, 141, 223, 60, 216, 142, 223, 60, 216, 143, 223, 60, 216, 144, 223, 60, 216, 145, 223, 62, 216, 231, 221, 60, 216, 128, 223, 60, 216, 129, 223, 60, 216, 151, 223, 60, 216, 159, 223, 60, 216, 171, 223, 60, 0, 42, 0, 60, 216, 150, 223, 60, 216, 198, 223, 60, 216, 197, 223, 62, 216, 71, 221, 62, 216, 72, 221, 62, 216, 73, 221, 60, 0, 42, 0, 189, 38, 190, 38, 62, 216, 78, 221, 60, 216, 192, 223, 60, 216, 208, 223, 60, 216, 200, 223, 60, 216, 201, 223, 60, 216, 190, 223, 62, 216, 79, 221, 60, 216, 179, 223, 60, 216, 207, 223, 60, 216, 209, 223, 60, 216, 210, 223, 62, 216, 77, 221, 60, 216, 211, 223, 60, 216, 248, 223, 62, 216, 74, 221, 62, 216, 75, 221, 62, 216, 69, 221, 243, 38, 248, 38, 60, 216, 163, 223, 62, 216, 63, 221, 60, 216, 189, 223, 60, 216, 191, 223, 61, 216, 247, 222, 62, 216, 76, 221, 60, 0, 42, 0, 60, 216, 175, 223, 62, 216, 128, 222, 62, 216, 129, 222, 60, 216, 177, 223, 61, 216, 46, 221, 62, 216, 132, 222, 62, 216, 255, 221, 60, 216, 174, 223, 61, 216, 121, 221, 60, 216, 176, 223, 60, 216, 178, 223, 62, 216, 233, 221, 62, 216, 248, 221, 62, 216, 133, 222, 62, 216, 134, 222, 96, 38, 101, 38, 102, 38, 99, 38, 95, 38, 60, 216, 207, 220, 60, 216, 4, 220, 60, 216, 180, 223, 60, 0, 42, 0, 60, 216, 173, 223, 61, 216, 188, 221, 60, 216, 168, 223, 62, 216, 245, 221, 62, 216, 161, 222, 62, 216, 246, 221, 62, 216, 162, 222, 60, 0, 42, 0, 61, 216, 83, 220, 61, 216, 118, 221, 62, 216, 125, 221, 62, 216, 124, 221, 62, 216, 186, 221, 61, 216, 84, 220, 61, 216, 85, 220, 61, 216, 86, 220, 62, 216, 227, 221, 62, 216, 228, 221, 62, 216, 229, 221, 62, 216, 230, 221, 61, 216, 87, 220, 61, 216, 88, 220, 62, 216, 123, 221, 62, 216, 113, 222, 62, 216, 114, 222, 62, 216, 115, 222, 61, 216, 89, 220, 61, 216, 90, 220, 61, 216, 91, 220, 61, 216, 92, 220, 61, 216, 93, 220, 61, 216, 205, 222, 60, 216, 146, 223, 62, 216, 116, 222, 61, 216, 94, 220, 61, 216, 95, 220, 62, 216, 126, 221, 62, 216, 127, 221, 61, 216, 96, 220, 61, 216, 97, 220, 62, 216, 112, 222, 61, 216, 98, 220, 61, 216, 81, 220, 61, 216, 82, 220, 60, 216, 169, 223, 60, 216, 147, 223, 62, 216, 226, 221, 62, 216, 150, 222, 209, 38, 61, 216, 255, 220, 61, 216, 132, 220, 61, 216, 141, 220, 61, 216, 142, 220, 60, 0, 42, 0, 61, 216, 7, 221, 61, 216, 8, 221, 61, 216, 9, 221, 61, 216, 10, 221, 61, 216, 226, 220, 61, 216, 227, 220, 61, 216, 239, 220, 61, 216, 20, 221, 61, 216, 21, 221, 60, 0, 42, 0, 60, 216, 188, 223, 60, 216, 181, 223, 60, 216, 182, 223, 60, 216, 153, 223, 60, 216, 154, 223, 60, 216, 155, 223, 60, 216, 164, 223, 60, 216, 167, 223, 61, 216, 251, 220, 60, 0, 42, 0, 60, 216, 183, 223, 62, 216, 151, 222, 60, 216, 184, 223, 60, 216, 185, 223, 60, 216, 186, 223, 60, 216, 187, 223, 62, 216, 149, 222, 62, 216, 65, 221, 62, 216, 152, 222, 60, 0, 42, 0, 61, 216, 241, 220, 61, 216, 242, 220, 14, 38, 61, 216, 222, 220, 61, 216, 223, 220, 61, 216, 224, 220, 60, 0, 42, 0, 61, 216, 11, 221, 61, 216, 12, 221, 61, 216, 187, 220, 61, 216, 165, 221, 61, 216, 168, 221, 40, 35, 61, 216, 177, 221, 61, 216, 178, 221, 61, 216, 189, 220, 61, 216, 190, 220, 61, 216, 191, 220, 61, 216, 192, 220, 62, 216, 238, 221, 60, 0, 42, 0, 60, 216, 165, 223, 60, 216, 158, 223, 61, 216, 253, 220, 60, 216, 172, 223, 61, 216, 250, 220, 61, 216, 247, 220, 61, 216, 248, 220, 61, 216, 249, 220, 61, 216, 252, 220, 61, 216, 13, 221, 61, 216, 14, 221, 61, 216, 111, 221, 61, 216, 161, 220, 61, 216, 38, 221, 60, 216, 238, 223, 62, 216, 148, 222, 60, 0, 42, 0, 61, 216, 212, 220, 61, 216, 213, 220, 61, 216, 214, 220, 61, 216, 215, 220, 61, 216, 216, 220, 61, 216, 217, 220, 61, 216, 218, 220, 61, 216, 211, 220, 61, 216, 210, 220, 61, 216, 195, 220, 61, 216, 220, 220, 61, 216, 196, 220, 61, 216, 240, 220, 61, 216, 222, 221, 61, 216, 209, 220, 61, 216, 22, 221, 60, 216, 247, 223, 60, 0, 42, 0, 61, 216, 176, 220, 62, 216, 153, 222, 61, 216, 180, 220, 61, 216, 181, 220, 61, 216, 182, 220, 61, 216, 183, 220, 61, 216, 184, 220, 61, 216, 179, 220, 62, 216, 254, 221, 61, 216, 185, 220, 60, 0, 42, 0, 9, 39, 61, 216, 231, 220, 61, 216, 232, 220, 61, 216, 233, 220, 61, 216, 228, 220, 61, 216, 229, 220, 61, 216, 230, 220, 61, 216, 235, 220, 61, 216, 234, 220, 61, 216, 236, 220, 61, 216, 237, 220, 61, 216, 238, 220, 61, 216, 243, 221, 60, 0, 42, 0, 15, 39, 18, 39, 61, 216, 139, 221, 61, 216, 138, 221, 61, 216, 140, 221, 61, 216, 141, 221, 61, 216, 221, 220, 60, 0, 42, 0, 61, 216, 188, 220, 61, 216, 193, 220, 61, 216, 194, 220, 61, 216, 194, 221, 61, 216, 197, 220, 61, 216, 198, 220, 61, 216, 210, 221, 61, 216, 211, 221, 61, 216, 199, 220, 61, 216, 200, 220, 61, 216, 201, 220, 61, 216, 202, 220, 61, 216, 203, 220, 61, 216, 204, 220, 61, 216, 205, 220, 61, 216, 206, 220, 61, 216, 135, 221, 61, 216, 207, 220, 61, 216, 208, 220, 2, 39, 61, 216, 195, 221, 61, 216, 196, 221, 61, 216, 209, 221, 60, 0, 42, 0, 61, 216, 18, 221, 61, 216, 19, 221, 61, 216, 15, 221, 61, 216, 16, 221, 61, 216, 17, 221, 61, 216, 221, 221, 60, 0, 42, 0, 61, 216, 40, 221, 62, 216, 147, 222, 207, 38, 146, 38, 61, 216, 224, 222, 61, 216, 225, 221, 148, 38, 61, 216, 43, 221, 62, 216, 131, 222, 60, 216, 249, 223, 61, 216, 225, 222, 62, 216, 154, 222, 61, 216, 39, 221, 62, 216, 155, 222, 61, 216, 41, 221, 153, 38, 61, 216, 220, 221, 150, 38, 62, 216, 175, 221, 61, 216, 23, 221, 211, 38, 62, 216, 157, 222, 62, 216, 240, 221, 62, 216, 242, 221, 62, 216, 156, 222, 60, 0, 42, 0, 151, 38, 62, 216, 234, 221, 62, 216, 235, 221, 62, 216, 236, 221, 61, 216, 44, 221, 61, 216, 45, 221, 61, 216, 225, 220, 60, 0, 42, 0, 61, 216, 137, 220, 62, 216, 120, 222, 61, 216, 138, 220, 62, 216, 121, 222, 62, 216, 122, 222, 60, 0, 42, 0, 61, 216, 170, 222, 61, 216, 215, 222, 62, 216, 158, 222, 62, 216, 159, 222, 61, 216, 207, 222, 61, 216, 203, 222, 62, 216, 145, 222, 61, 216, 189, 222, 62, 216, 160, 222, 61, 216, 191, 222, 61, 216, 193, 222, 62, 216, 164, 222, 62, 216, 146, 222, 62, 216, 244, 221, 62, 216, 247, 221, 62, 216, 249, 221, 62, 216, 250, 221, 62, 216, 251, 221, 62, 216, 163, 222, 62, 216, 252, 221, 62, 216, 165, 222, 62, 216, 253, 221, 62, 216, 239, 221, 61, 216, 210, 222, 60, 0, 42, 0, 61, 216, 172, 222, 176, 38, 62, 216, 166, 222, 177, 38, 61, 216, 255, 221, 62, 216, 167, 222, 60, 0, 42, 0, 60, 216, 231, 223, 61, 216, 174, 222, 61, 216, 176, 222, 127, 38, 61, 216, 185, 222, 61, 216, 186, 222, 61, 216, 187, 222, 61, 216, 188, 222, 61, 216, 190, 222, 61, 216, 194, 222, 61, 216, 195, 222, 61, 216, 196, 222, 61, 216, 197, 222, 60, 0, 42, 0, 160, 38, 61, 216, 184, 222, 212, 38, 61, 216, 171, 222, 61, 216, 179, 222, 61, 216, 173, 222, 61, 216, 175, 222, 61, 216, 177, 222, 61, 216, 183, 222, 61, 216, 245, 220, 61, 216, 30, 221, 34, 38, 35, 38, 60, 0, 42, 0, 6, 43, 151, 33, 161, 39, 152, 33, 7, 43, 153, 33, 5, 43, 150, 33, 149, 33, 148, 33, 169, 33, 170, 33, 52, 41, 53, 41, 61, 216, 3, 221, 61, 216, 4, 221, 61, 216, 25, 221, 61, 216, 26, 221, 61, 216, 27, 221, 61, 216, 28, 221, 61, 216, 29, 221, 60, 0, 42, 0, 61, 216, 208, 222, 155, 38, 61, 216, 73, 221, 33, 39, 56, 38, 47, 38, 29, 39, 38, 38, 42, 38, 46, 38, 61, 216, 78, 221, 61, 216, 47, 221, 60, 0, 42, 0, 72, 38, 73, 38, 74, 38, 75, 38, 76, 38, 77, 38, 78, 38, 79, 38, 80, 38, 81, 38, 82, 38, 83, 38, 206, 38, 60, 0, 42, 0, 61, 216, 0, 221, 61, 216, 1, 221, 61, 216, 2, 221, 182, 37, 233, 35, 237, 35, 239, 35, 192, 37, 234, 35, 238, 35, 61, 216, 60, 221, 235, 35, 61, 216, 61, 221, 236, 35, 248, 35, 249, 35, 250, 35, 207, 35, 60, 216, 166, 223, 61, 216, 5, 221, 61, 216, 6, 221, 61, 216, 246, 220, 61, 216, 243, 220, 61, 216, 244, 220, 60, 0, 42, 0, 64, 38, 66, 38, 167, 38, 60, 0, 42, 0, 22, 39, 149, 39, 150, 39, 151, 39, 126, 38, 60, 0, 42, 0, 60, 32, 73, 32, 83, 39, 84, 39, 85, 39, 87, 39, 48, 48, 60, 0, 42, 0, 61, 216, 177, 220, 61, 216, 178, 220, 60, 0, 42, 0, 149, 38, 123, 38, 156, 38, 61, 216, 49, 221, 61, 216, 219, 220, 61, 216, 48, 221, 85, 43, 5, 39, 17, 38, 20, 39, 76, 39, 78, 39, 176, 39, 191, 39, 61, 48, 51, 39, 52, 39, 71, 39, 169, 0, 174, 0, 34, 33, 60, 0, 32, 0, 39, 0, 35, 0, 227, 32, 39, 0, 32, 0, 61, 0, 32, 0, 39, 0, 35, 0, 15, 254, 227, 32, 39, 0, 60, 0, 32, 0, 39, 0, 42, 0, 227, 32, 39, 0, 32, 0, 61, 0, 32, 0, 39, 0, 42, 0, 15, 254, 227, 32, 39, 0, 60, 0, 32, 0, 48, 0, 227, 32, 32, 0, 61, 0, 32, 0, 48, 0, 15, 254, 227, 32, 60, 0, 32, 0, 49, 0, 227, 32, 32, 0, 61, 0, 32, 0, 49, 0, 15, 254, 227, 32, 60, 0, 32, 0, 50, 0, 227, 32, 32, 0, 61, 0, 32, 0, 50, 0, 15, 254, 227, 32, 60, 0, 32, 0, 51, 0, 227, 32, 32, 0, 61, 0, 32, 0, 51, 0, 15, 254, 227, 32, 60, 0, 32, 0, 52, 0, 227, 32, 32, 0, 61, 0, 32, 0, 52, 0, 15, 254, 227, 32, 60, 0, 32, 0, 53, 0, 227, 32, 32, 0, 61, 0, 32, 0, 53, 0, 15, 254, 227, 32, 60, 0, 32, 0, 54, 0, 227, 32, 32, 0, 61, 0, 32, 0, 54, 0, 15, 254, 227, 32, 60, 0, 32, 0, 55, 0, 227, 32, 32, 0, 61, 0, 32, 0, 55, 0, 15, 254, 227, 32, 60, 0, 32, 0, 56, 0, 227, 32, 32, 0, 61, 0, 32, 0, 56, 0, 15, 254, 227, 32, 60, 0, 32, 0, 57, 0, 227, 32, 32, 0, 61, 0, 32, 0, 57, 0, 15, 254, 227, 32, 60, 0, 42, 0, 61, 216, 31, 221, 60, 0, 42, 0, 61, 216, 32, 221, 61, 216, 33, 221, 61, 216, 34, 221, 61, 216, 35, 221, 61, 216, 36, 221, 60, 216, 112, 221, 60, 216, 142, 221, 60, 216, 113, 221, 60, 216, 145, 221, 60, 216, 146, 221, 60, 216, 147, 221, 57, 33, 60, 216, 148, 221, 194, 36, 60, 216, 149, 221, 60, 216, 150, 221, 60, 216, 126, 221, 60, 216, 151, 221, 60, 216, 127, 221, 60, 216, 152, 221, 60, 216, 153, 221, 60, 216, 154, 221, 60, 216, 1, 222, 60, 216, 2, 222, 60, 216, 55, 222, 60, 216, 54, 222, 60, 216, 47, 222, 60, 216, 80, 222, 60, 216, 57, 222, 60, 216, 26, 222, 60, 216, 50, 222, 60, 216, 81, 222, 60, 216, 56, 222, 60, 216, 52, 222, 60, 216, 51, 222, 151, 50, 153, 50, 60, 216, 58, 222, 60, 216, 53, 222, 60, 0, 42, 0, 61, 216, 52, 221, 61, 216, 224, 223, 61, 216, 225, 223, 61, 216, 226, 223, 61, 216, 53, 221, 61, 216, 227, 223, 61, 216, 228, 223, 171, 38, 170, 38, 61, 216, 229, 223, 61, 216, 231, 223, 61, 216, 232, 223, 61, 216, 233, 223, 61, 216, 230, 223, 61, 216, 234, 223, 61, 216, 235, 223, 27, 43, 28, 43, 252, 37, 251, 37, 254, 37, 253, 37, 170, 37, 171, 37, 61, 216, 54, 221, 61, 216, 55, 221, 61, 216, 56, 221, 61, 216, 57, 221, 61, 216, 58, 221, 61, 216, 59, 221, 61, 216, 160, 220, 61, 216, 24, 221, 61, 216, 51, 221, 61, 216, 50, 221, 60, 0, 42, 0, 60, 216, 193, 223, 61, 216, 169, 222, 60, 216, 140, 223, 60, 216, 244, 223, 60, 216, 243, 223, 60, 0, 32, 0, 60, 216, 243, 223, 13, 32, 60, 216, 8, 223, 32, 0, 61, 0, 32, 0, 60, 216, 243, 223, 15, 254, 13, 32, 60, 216, 8, 223, 60, 0, 32, 0, 60, 216, 243, 223, 13, 32, 167, 38, 32, 0, 61, 0, 32, 0, 60, 216, 243, 223, 15, 254, 13, 32, 167, 38, 60, 0, 32, 0, 60, 216, 244, 223, 13, 32, 32, 38, 60, 0, 42, 0, 60, 216, 230, 221, 60, 216, 231, 221, 60, 216, 232, 221, 60, 216, 233, 221, 60, 216, 234, 221, 60, 216, 235, 221, 60, 216, 236, 221, 60, 216, 237, 221, 60, 216, 238, 221, 60, 216, 239, 221, 60, 216, 240, 221, 60, 216, 241, 221, 60, 216, 242, 221, 60, 216, 243, 221, 60, 216, 244, 221, 60, 216, 245, 221, 60, 216, 246, 221, 60, 216, 247, 221, 60, 216, 248, 221, 60, 216, 249, 221, 60, 216, 250, 221, 60, 216, 251, 221, 60, 216, 252, 221, 60, 216, 253, 221, 60, 216, 254, 221, 60, 216, 255, 221, 60, 0, 32, 0, 60, 216, 244, 223, 64, 219, 103, 220, 64, 219, 98, 220, 64, 219, 101, 220, 64, 219, 110, 220, 64, 219, 103, 220, 64, 219, 127, 220, 60, 0, 32, 0, 60, 216, 244, 223, 64, 219, 103, 220, 64, 219, 98, 220, 64, 219, 115, 220, 64, 219, 99, 220, 64, 219, 116, 220, 64, 219, 127, 220, 60, 0, 32, 0, 60, 216, 244, 223, 64, 219, 103, 220, 64, 219, 98, 220, 64, 219, 119, 220, 64, 219, 108, 220, 64, 219, 115, 220, 64, 219, 127, 220, 38, 0, 32, 0, 62, 216, 209, 221, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 38, 0, 32, 0, 61, 216, 104, 220, 32, 0, 60, 0, 60, 0, 32, 0, 62, 216, 212, 221, 32, 0, 60, 0, 60, 0, 61, 216, 113, 220, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 61, 216, 113, 220, 60, 216, 251, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 61, 216, 113, 220, 60, 216, 252, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 61, 216, 113, 220, 60, 216, 253, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 61, 216, 113, 220, 60, 216, 254, 223, 13, 32, 66, 38, 32, 0, 60, 0, 60, 0, 61, 216, 113, 220, 60, 216, 255, 223, 13, 32, 66, 38, 38, 0, 32, 0, 61, 216, 105, 220, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 251, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 252, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 253, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 254, 223, 13, 32, 64, 38, 32, 0, 60, 0, 60, 0, 32, 0, 61, 216, 113, 220, 60, 216, 255, 223, 13, 32, 64, 38, 0, 0, 240, 223, 0, 139, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 116, 0, 101, 0, 114, 0, 116, 0, 105, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 61, 0, 39, 0, 1, 0, 39, 0, 61, 0, 39, 0, 2, 0, 39, 0, 61, 0, 39, 0, 3, 0, 39, 0, 61, 0, 39, 0, 4, 0, 39, 0, 61, 0, 39, 0, 5, 0, 39, 0, 61, 0, 39, 0, 6, 0, 39, 0, 61, 0, 39, 0, 7, 0, 39, 0, 61, 0, 39, 0, 8, 0, 39, 0, 61, 0, 39, 0, 14, 0, 39, 0, 61, 0, 39, 0, 15, 0, 39, 0, 61, 0, 39, 0, 16, 0, 39, 0, 61, 0, 39, 0, 17, 0, 39, 0, 61, 0, 39, 0, 18, 0, 39, 0, 61, 0, 39, 0, 19, 0, 39, 0, 61, 0, 39, 0, 20, 0, 39, 0, 61, 0, 39, 0, 21, 0, 39, 0, 61, 0, 39, 0, 22, 0, 39, 0, 61, 0, 39, 0, 23, 0, 39, 0, 61, 0, 39, 0, 24, 0, 39, 0, 61, 0, 39, 0, 25, 0, 39, 0, 61, 0, 39, 0, 26, 0, 39, 0, 61, 0, 39, 0, 27, 0, 39, 0, 61, 0, 39, 0, 28, 0, 39, 0, 61, 0, 39, 0, 29, 0, 39, 0, 61, 0, 39, 0, 30, 0, 39, 0, 61, 0, 39, 0, 31, 0, 39, 0, 61, 0, 39, 0, 127, 0, 39, 0, 61, 0, 39, 0, 128, 0, 39, 0, 61, 0, 39, 0, 129, 0, 39, 0, 61, 0, 39, 0, 130, 0, 39, 0, 61, 0, 39, 0, 131, 0, 39, 0, 61, 0, 39, 0, 132, 0, 39, 0, 61, 0, 39, 0, 134, 0, 39, 0, 61, 0, 39, 0, 135, 0, 39, 0, 61, 0, 39, 0, 136, 0, 39, 0, 61, 0, 39, 0, 137, 0, 39, 0, 61, 0, 39, 0, 138, 0, 39, 0, 61, 0, 39, 0, 139, 0, 39, 0, 61, 0, 39, 0, 140, 0, 39, 0, 61, 0, 39, 0, 141, 0, 39, 0, 61, 0, 39, 0, 142, 0, 39, 0, 61, 0, 39, 0, 143, 0, 39, 0, 61, 0, 39, 0, 144, 0, 39, 0, 61, 0, 39, 0, 145, 0, 39, 0, 61, 0, 39, 0, 146, 0, 39, 0, 61, 0, 39, 0, 147, 0, 39, 0, 61, 0, 39, 0, 148, 0, 39, 0, 61, 0, 39, 0, 149, 0, 39, 0, 61, 0, 39, 0, 150, 0, 39, 0, 61, 0, 39, 0, 151, 0, 39, 0, 61, 0, 39, 0, 152, 0, 39, 0, 61, 0, 39, 0, 153, 0, 39, 0, 61, 0, 39, 0, 154, 0, 39, 0, 61, 0, 39, 0, 155, 0, 39, 0, 61, 0, 39, 0, 156, 0, 39, 0, 61, 0, 39, 0, 157, 0, 39, 0, 61, 0, 39, 0, 158, 0, 39, 0, 61, 0, 39, 0, 159, 0, 39, 0, 61, 0, 39, 0, 173, 0, 39, 0, 61, 0, 79, 3, 61, 0, 136, 4, 61, 0, 137, 4, 61, 0, 145, 5, 61, 0, 146, 5, 61, 0, 147, 5, 61, 0, 148, 5, 61, 0, 149, 5, 61, 0, 150, 5, 61, 0, 151, 5, 61, 0, 152, 5, 61, 0, 153, 5, 61, 0, 154, 5, 61, 0, 155, 5, 61, 0, 156, 5, 61, 0, 157, 5, 61, 0, 158, 5, 61, 0, 159, 5, 61, 0, 160, 5, 61, 0, 161, 5, 61, 0, 162, 5, 61, 0, 163, 5, 61, 0, 164, 5, 61, 0, 165, 5, 61, 0, 166, 5, 61, 0, 167, 5, 61, 0, 168, 5, 61, 0, 169, 5, 61, 0, 170, 5, 61, 0, 171, 5, 61, 0, 172, 5, 61, 0, 173, 5, 61, 0, 174, 5, 61, 0, 175, 5, 61, 0, 189, 5, 61, 0, 196, 5, 61, 0, 197, 5, 61, 0, 39, 0, 0, 6, 39, 0, 61, 0, 39, 0, 1, 6, 39, 0, 61, 0, 39, 0, 2, 6, 39, 0, 61, 0, 39, 0, 3, 6, 39, 0, 61, 0, 39, 0, 4, 6, 39, 0, 61, 0, 39, 0, 5, 6, 39, 0, 61, 0, 16, 6, 61, 0, 17, 6, 61, 0, 18, 6, 61, 0, 19, 6, 61, 0, 20, 6, 61, 0, 21, 6, 61, 0, 22, 6, 61, 0, 23, 6, 61, 0, 24, 6, 61, 0, 25, 6, 61, 0, 26, 6, 61, 0, 39, 0, 28, 6, 39, 0, 61, 0, 14, 32, 64, 6, 14, 32, 61, 0, 214, 6, 61, 0, 215, 6, 61, 0, 216, 6, 61, 0, 217, 6, 61, 0, 218, 6, 61, 0, 219, 6, 61, 0, 220, 6, 61, 0, 39, 0, 221, 6, 39, 0, 61, 0, 223, 6, 61, 0, 224, 6, 61, 0, 225, 6, 61, 0, 226, 6, 61, 0, 227, 6, 61, 0, 228, 6, 61, 0, 231, 6, 61, 0, 232, 6, 61, 0, 234, 6, 61, 0, 235, 6, 61, 0, 236, 6, 61, 0, 237, 6, 61, 0, 39, 0, 15, 7, 39, 0, 61, 0, 64, 7, 61, 0, 67, 7, 61, 0, 68, 7, 61, 0, 71, 7, 61, 0, 72, 7, 61, 0, 73, 7, 61, 0, 74, 7, 61, 0, 14, 32, 250, 7, 14, 32, 61, 0, 211, 8, 61, 0, 212, 8, 61, 0, 213, 8, 61, 0, 214, 8, 61, 0, 215, 8, 61, 0, 216, 8, 61, 0, 217, 8, 61, 0, 218, 8, 61, 0, 219, 8, 61, 0, 220, 8, 61, 0, 221, 8, 61, 0, 222, 8, 61, 0, 223, 8, 61, 0, 224, 8, 61, 0, 225, 8, 61, 0, 39, 0, 226, 8, 39, 0, 61, 0, 234, 8, 61, 0, 235, 8, 61, 0, 236, 8, 61, 0, 237, 8, 61, 0, 238, 8, 61, 0, 239, 8, 61, 0, 243, 8, 61, 0, 81, 9, 61, 0, 82, 9, 61, 0, 24, 15, 61, 0, 25, 15, 61, 0, 53, 15, 61, 0, 55, 15, 61, 0, 62, 15, 61, 0, 63, 15, 61, 0, 130, 15, 61, 0, 131, 15, 61, 0, 134, 15, 61, 0, 135, 15, 61, 0, 198, 15, 61, 0, 180, 23, 61, 0, 181, 23, 61, 0, 211, 23, 61, 0, 10, 24, 61, 0, 11, 24, 61, 0, 12, 24, 61, 0, 13, 24, 61, 0, 39, 0, 14, 24, 39, 0, 61, 0, 127, 26, 61, 0, 107, 27, 61, 0, 108, 27, 61, 0, 109, 27, 61, 0, 110, 27, 61, 0, 111, 27, 61, 0, 112, 27, 61, 0, 113, 27, 61, 0, 114, 27, 61, 0, 115, 27, 61, 0, 208, 28, 61, 0, 209, 28, 61, 0, 210, 28, 61, 0, 211, 28, 61, 0, 212, 28, 61, 0, 213, 28, 61, 0, 214, 28, 61, 0, 215, 28, 61, 0, 216, 28, 61, 0, 217, 28, 61, 0, 218, 28, 61, 0, 219, 28, 61, 0, 220, 28, 61, 0, 221, 28, 61, 0, 222, 28, 61, 0, 223, 28, 61, 0, 224, 28, 61, 0, 225, 28, 61, 0, 226, 28, 61, 0, 227, 28, 61, 0, 228, 28, 61, 0, 229, 28, 61, 0, 230, 28, 61, 0, 231, 28, 61, 0, 232, 28, 61, 0, 244, 28, 61, 0, 247, 28, 61, 0, 248, 28, 61, 0, 249, 28, 61, 0, 39, 0, 11, 32, 39, 0, 61, 0, 39, 0, 12, 32, 39, 0, 61, 0, 39, 0, 13, 32, 39, 0, 61, 0, 39, 0, 14, 32, 39, 0, 61, 0, 39, 0, 15, 32, 39, 0, 61, 0, 39, 0, 42, 32, 39, 0, 61, 0, 39, 0, 43, 32, 39, 0, 61, 0, 39, 0, 44, 32, 39, 0, 61, 0, 39, 0, 45, 32, 39, 0, 61, 0, 39, 0, 46, 32, 39, 0, 61, 0, 39, 0, 96, 32, 39, 0, 61, 0, 39, 0, 97, 32, 39, 0, 61, 0, 39, 0, 98, 32, 39, 0, 61, 0, 39, 0, 99, 32, 39, 0, 61, 0, 39, 0, 100, 32, 39, 0, 61, 0, 39, 0, 102, 32, 39, 0, 61, 0, 39, 0, 103, 32, 39, 0, 61, 0, 39, 0, 104, 32, 39, 0, 61, 0, 39, 0, 105, 32, 39, 0, 61, 0, 39, 0, 106, 32, 39, 0, 61, 0, 39, 0, 107, 32, 39, 0, 61, 0, 39, 0, 108, 32, 39, 0, 61, 0, 39, 0, 109, 32, 39, 0, 61, 0, 39, 0, 110, 32, 39, 0, 61, 0, 39, 0, 111, 32, 39, 0, 61, 0, 127, 45, 61, 0, 112, 166, 61, 0, 113, 166, 61, 0, 114, 166, 61, 0, 224, 168, 61, 0, 225, 168, 61, 0, 226, 168, 61, 0, 227, 168, 61, 0, 228, 168, 61, 0, 229, 168, 61, 0, 230, 168, 61, 0, 231, 168, 61, 0, 232, 168, 61, 0, 233, 168, 61, 0, 234, 168, 61, 0, 235, 168, 61, 0, 236, 168, 61, 0, 237, 168, 61, 0, 238, 168, 61, 0, 239, 168, 61, 0, 240, 168, 61, 0, 241, 168, 61, 0, 0, 254, 61, 0, 1, 254, 61, 0, 2, 254, 61, 0, 3, 254, 61, 0, 4, 254, 61, 0, 5, 254, 61, 0, 6, 254, 61, 0, 7, 254, 61, 0, 8, 254, 61, 0, 9, 254, 61, 0, 10, 254, 61, 0, 11, 254, 61, 0, 12, 254, 61, 0, 13, 254, 61, 0, 14, 254, 61, 0, 15, 254, 61, 0, 33, 254, 61, 0, 35, 254, 61, 0, 36, 254, 61, 0, 37, 254, 61, 0, 38, 254, 61, 0, 40, 254, 61, 0, 42, 254, 61, 0, 43, 254, 61, 0, 44, 254, 61, 0, 45, 254, 61, 0, 47, 254, 61, 0, 14, 32, 115, 254, 14, 32, 61, 0, 39, 0, 255, 254, 39, 0, 61, 0, 39, 0, 249, 255, 39, 0, 61, 0, 39, 0, 250, 255, 39, 0, 61, 0, 39, 0, 251, 255, 39, 0, 61, 0, 0, 216, 224, 222, 61, 0, 39, 0, 4, 216, 189, 220, 39, 0, 61, 0, 39, 0, 4, 216, 205, 220, 39, 0, 61, 0, 4, 216, 102, 223, 61, 0, 4, 216, 103, 223, 61, 0, 4, 216, 104, 223, 61, 0, 4, 216, 105, 223, 61, 0, 4, 216, 106, 223, 61, 0, 4, 216, 107, 223, 61, 0, 4, 216, 108, 223, 61, 0, 4, 216, 112, 223, 61, 0, 4, 216, 113, 223, 61, 0, 4, 216, 114, 223, 61, 0, 4, 216, 115, 223, 61, 0, 4, 216, 116, 223, 61, 0, 39, 0, 13, 216, 48, 220, 39, 0, 61, 0, 39, 0, 13, 216, 49, 220, 39, 0, 61, 0, 39, 0, 13, 216, 50, 220, 39, 0, 61, 0, 39, 0, 13, 216, 51, 220, 39, 0, 61, 0, 39, 0, 13, 216, 52, 220, 39, 0, 61, 0, 39, 0, 13, 216, 53, 220, 39, 0, 61, 0, 39, 0, 13, 216, 54, 220, 39, 0, 61, 0, 39, 0, 13, 216, 55, 220, 39, 0, 61, 0, 39, 0, 13, 216, 56, 220, 39, 0, 61, 0, 27, 216, 228, 223, 61, 0, 39, 0, 47, 216, 160, 220, 39, 0, 61, 0, 39, 0, 47, 216, 161, 220, 39, 0, 61, 0, 39, 0, 47, 216, 162, 220, 39, 0, 61, 0, 39, 0, 47, 216, 163, 220, 39, 0, 61, 0, 52, 216, 101, 221, 61, 0, 52, 216, 102, 221, 61, 0, 52, 216, 103, 221, 61, 0, 52, 216, 104, 221, 61, 0, 52, 216, 105, 221, 61, 0, 52, 216, 109, 221, 61, 0, 52, 216, 110, 221, 61, 0, 52, 216, 111, 221, 61, 0, 52, 216, 112, 221, 61, 0, 52, 216, 113, 221, 61, 0, 52, 216, 114, 221, 61, 0, 39, 0, 52, 216, 115, 221, 39, 0, 61, 0, 39, 0, 52, 216, 116, 221, 39, 0, 61, 0, 39, 0, 52, 216, 117, 221, 39, 0, 61, 0, 39, 0, 52, 216, 118, 221, 39, 0, 61, 0, 39, 0, 52, 216, 119, 221, 39, 0, 61, 0, 39, 0, 52, 216, 120, 221, 39, 0, 61, 0, 39, 0, 52, 216, 121, 221, 39, 0, 61, 0, 39, 0, 52, 216, 122, 221, 39, 0, 61, 0, 52, 216, 123, 221, 61, 0, 52, 216, 124, 221, 61, 0, 52, 216, 125, 221, 61, 0, 52, 216, 126, 221, 61, 0, 52, 216, 127, 221, 61, 0, 52, 216, 128, 221, 61, 0, 52, 216, 129, 221, 61, 0, 52, 216, 130, 221, 61, 0, 52, 216, 133, 221, 61, 0, 52, 216, 134, 221, 61, 0, 52, 216, 135, 221, 61, 0, 52, 216, 136, 221, 61, 0, 52, 216, 137, 221, 61, 0, 52, 216, 138, 221, 61, 0, 52, 216, 139, 221, 61, 0, 52, 216, 170, 221, 61, 0, 52, 216, 171, 221, 61, 0, 52, 216, 172, 221, 61, 0, 52, 216, 173, 221, 61, 0, 52, 216, 66, 222, 61, 0, 52, 216, 67, 222, 61, 0, 52, 216, 68, 222, 61, 0, 54, 216, 0, 222, 61, 0, 54, 216, 1, 222, 61, 0, 54, 216, 2, 222, 61, 0, 54, 216, 3, 222, 61, 0, 54, 216, 4, 222, 61, 0, 54, 216, 5, 222, 61, 0, 54, 216, 6, 222, 61, 0, 54, 216, 7, 222, 61, 0, 54, 216, 8, 222, 61, 0, 54, 216, 9, 222, 61, 0, 54, 216, 10, 222, 61, 0, 54, 216, 11, 222, 61, 0, 54, 216, 12, 222, 61, 0, 54, 216, 13, 222, 61, 0, 54, 216, 14, 222, 61, 0, 54, 216, 15, 222, 61, 0, 54, 216, 16, 222, 61, 0, 54, 216, 17, 222, 61, 0, 54, 216, 18, 222, 61, 0, 54, 216, 19, 222, 61, 0, 54, 216, 20, 222, 61, 0, 54, 216, 21, 222, 61, 0, 54, 216, 22, 222, 61, 0, 54, 216, 23, 222, 61, 0, 54, 216, 24, 222, 61, 0, 54, 216, 25, 222, 61, 0, 54, 216, 26, 222, 61, 0, 54, 216, 27, 222, 61, 0, 54, 216, 28, 222, 61, 0, 54, 216, 29, 222, 61, 0, 54, 216, 30, 222, 61, 0, 54, 216, 31, 222, 61, 0, 54, 216, 32, 222, 61, 0, 54, 216, 33, 222, 61, 0, 54, 216, 34, 222, 61, 0, 54, 216, 35, 222, 61, 0, 54, 216, 36, 222, 61, 0, 54, 216, 37, 222, 61, 0, 54, 216, 38, 222, 61, 0, 54, 216, 39, 222, 61, 0, 54, 216, 40, 222, 61, 0, 54, 216, 41, 222, 61, 0, 54, 216, 42, 222, 61, 0, 54, 216, 43, 222, 61, 0, 54, 216, 44, 222, 61, 0, 54, 216, 45, 222, 61, 0, 54, 216, 46, 222, 61, 0, 54, 216, 47, 222, 61, 0, 54, 216, 48, 222, 61, 0, 54, 216, 49, 222, 61, 0, 54, 216, 50, 222, 61, 0, 54, 216, 51, 222, 61, 0, 54, 216, 52, 222, 61, 0, 54, 216, 53, 222, 61, 0, 54, 216, 54, 222, 61, 0, 54, 216, 59, 222, 61, 0, 54, 216, 60, 222, 61, 0, 54, 216, 61, 222, 61, 0, 54, 216, 62, 222, 61, 0, 54, 216, 63, 222, 61, 0, 54, 216, 64, 222, 61, 0, 54, 216, 65, 222, 61, 0, 54, 216, 66, 222, 61, 0, 54, 216, 67, 222, 61, 0, 54, 216, 68, 222, 61, 0, 54, 216, 69, 222, 61, 0, 54, 216, 70, 222, 61, 0, 54, 216, 71, 222, 61, 0, 54, 216, 72, 222, 61, 0, 54, 216, 73, 222, 61, 0, 54, 216, 74, 222, 61, 0, 54, 216, 75, 222, 61, 0, 54, 216, 76, 222, 61, 0, 54, 216, 77, 222, 61, 0, 54, 216, 78, 222, 61, 0, 54, 216, 79, 222, 61, 0, 54, 216, 80, 222, 61, 0, 54, 216, 81, 222, 61, 0, 54, 216, 82, 222, 61, 0, 54, 216, 83, 222, 61, 0, 54, 216, 84, 222, 61, 0, 54, 216, 85, 222, 61, 0, 54, 216, 86, 222, 61, 0, 54, 216, 87, 222, 61, 0, 54, 216, 88, 222, 61, 0, 54, 216, 89, 222, 61, 0, 54, 216, 90, 222, 61, 0, 54, 216, 91, 222, 61, 0, 54, 216, 92, 222, 61, 0, 54, 216, 93, 222, 61, 0, 54, 216, 94, 222, 61, 0, 54, 216, 95, 222, 61, 0, 54, 216, 96, 222, 61, 0, 54, 216, 97, 222, 61, 0, 54, 216, 98, 222, 61, 0, 54, 216, 99, 222, 61, 0, 54, 216, 100, 222, 61, 0, 54, 216, 101, 222, 61, 0, 54, 216, 102, 222, 61, 0, 54, 216, 103, 222, 61, 0, 54, 216, 104, 222, 61, 0, 54, 216, 105, 222, 61, 0, 54, 216, 106, 222, 61, 0, 54, 216, 107, 222, 61, 0, 54, 216, 108, 222, 61, 0, 54, 216, 117, 222, 61, 0, 54, 216, 132, 222, 61, 0, 54, 216, 155, 222, 61, 0, 54, 216, 156, 222, 61, 0, 54, 216, 157, 222, 61, 0, 54, 216, 158, 222, 61, 0, 54, 216, 159, 222, 61, 0, 54, 216, 161, 222, 61, 0, 54, 216, 162, 222, 61, 0, 54, 216, 163, 222, 61, 0, 54, 216, 164, 222, 61, 0, 54, 216, 165, 222, 61, 0, 54, 216, 166, 222, 61, 0, 54, 216, 167, 222, 61, 0, 54, 216, 168, 222, 61, 0, 54, 216, 169, 222, 61, 0, 54, 216, 170, 222, 61, 0, 54, 216, 171, 222, 61, 0, 54, 216, 172, 222, 61, 0, 54, 216, 173, 222, 61, 0, 54, 216, 174, 222, 61, 0, 54, 216, 175, 222, 61, 0, 58, 216, 208, 220, 61, 0, 58, 216, 209, 220, 61, 0, 58, 216, 210, 220, 61, 0, 58, 216, 211, 220, 61, 0, 58, 216, 212, 220, 61, 0, 58, 216, 213, 220, 61, 0, 58, 216, 214, 220, 61, 0, 39, 0, 64, 219, 1, 220, 39, 0, 61, 0, 39, 0, 64, 219, 32, 220, 39, 0, 61, 0, 39, 0, 64, 219, 33, 220, 39, 0, 61, 0, 39, 0, 64, 219, 34, 220, 39, 0, 61, 0, 39, 0, 64, 219, 35, 220, 39, 0, 61, 0, 39, 0, 64, 219, 36, 220, 39, 0, 61, 0, 39, 0, 64, 219, 37, 220, 39, 0, 61, 0, 39, 0, 64, 219, 38, 220, 39, 0, 61, 0, 39, 0, 64, 219, 39, 220, 39, 0, 61, 0, 39, 0, 64, 219, 40, 220, 39, 0, 61, 0, 39, 0, 64, 219, 41, 220, 39, 0, 61, 0, 39, 0, 64, 219, 42, 220, 39, 0, 61, 0, 39, 0, 64, 219, 43, 220, 39, 0, 61, 0, 39, 0, 64, 219, 44, 220, 39, 0, 61, 0, 39, 0, 64, 219, 45, 220, 39, 0, 61, 0, 39, 0, 64, 219, 46, 220, 39, 0, 61, 0, 39, 0, 64, 219, 47, 220, 39, 0, 61, 0, 39, 0, 64, 219, 48, 220, 39, 0, 61, 0, 39, 0, 64, 219, 49, 220, 39, 0, 61, 0, 39, 0, 64, 219, 50, 220, 39, 0, 61, 0, 39, 0, 64, 219, 51, 220, 39, 0, 61, 0, 39, 0, 64, 219, 52, 220, 39, 0, 61, 0, 39, 0, 64, 219, 53, 220, 39, 0, 61, 0, 39, 0, 64, 219, 54, 220, 39, 0, 61, 0, 39, 0, 64, 219, 55, 220, 39, 0, 61, 0, 39, 0, 64, 219, 56, 220, 39, 0, 61, 0, 39, 0, 64, 219, 57, 220, 39, 0, 61, 0, 39, 0, 64, 219, 58, 220, 39, 0, 61, 0, 39, 0, 64, 219, 59, 220, 39, 0, 61, 0, 39, 0, 64, 219, 60, 220, 39, 0, 61, 0, 39, 0, 64, 219, 61, 220, 39, 0, 61, 0, 39, 0, 64, 219, 62, 220, 39, 0, 61, 0, 39, 0, 64, 219, 63, 220, 39, 0, 61, 0, 39, 0, 64, 219, 64, 220, 39, 0, 61, 0, 39, 0, 64, 219, 65, 220, 39, 0, 61, 0, 39, 0, 64, 219, 66, 220, 39, 0, 61, 0, 39, 0, 64, 219, 67, 220, 39, 0, 61, 0, 39, 0, 64, 219, 68, 220, 39, 0, 61, 0, 39, 0, 64, 219, 69, 220, 39, 0, 61, 0, 39, 0, 64, 219, 70, 220, 39, 0, 61, 0, 39, 0, 64, 219, 71, 220, 39, 0, 61, 0, 39, 0, 64, 219, 72, 220, 39, 0, 61, 0, 39, 0, 64, 219, 73, 220, 39, 0, 61, 0, 39, 0, 64, 219, 74, 220, 39, 0, 61, 0, 39, 0, 64, 219, 75, 220, 39, 0, 61, 0, 39, 0, 64, 219, 76, 220, 39, 0, 61, 0, 39, 0, 64, 219, 77, 220, 39, 0, 61, 0, 39, 0, 64, 219, 78, 220, 39, 0, 61, 0, 39, 0, 64, 219, 79, 220, 39, 0, 61, 0, 39, 0, 64, 219, 80, 220, 39, 0, 61, 0, 39, 0, 64, 219, 81, 220, 39, 0, 61, 0, 39, 0, 64, 219, 82, 220, 39, 0, 61, 0, 39, 0, 64, 219, 83, 220, 39, 0, 61, 0, 39, 0, 64, 219, 84, 220, 39, 0, 61, 0, 39, 0, 64, 219, 85, 220, 39, 0, 61, 0, 39, 0, 64, 219, 86, 220, 39, 0, 61, 0, 39, 0, 64, 219, 87, 220, 39, 0, 61, 0, 39, 0, 64, 219, 88, 220, 39, 0, 61, 0, 39, 0, 64, 219, 89, 220, 39, 0, 61, 0, 39, 0, 64, 219, 90, 220, 39, 0, 61, 0, 39, 0, 64, 219, 91, 220, 39, 0, 61, 0, 39, 0, 64, 219, 92, 220, 39, 0, 61, 0, 39, 0, 64, 219, 93, 220, 39, 0, 61, 0, 39, 0, 64, 219, 94, 220, 39, 0, 61, 0, 39, 0, 64, 219, 95, 220, 39, 0, 61, 0, 39, 0, 64, 219, 96, 220, 39, 0, 61, 0, 39, 0, 64, 219, 97, 220, 39, 0, 61, 0, 39, 0, 64, 219, 98, 220, 39, 0, 61, 0, 39, 0, 64, 219, 99, 220, 39, 0, 61, 0, 39, 0, 64, 219, 100, 220, 39, 0, 61, 0, 39, 0, 64, 219, 101, 220, 39, 0, 61, 0, 39, 0, 64, 219, 102, 220, 39, 0, 61, 0, 39, 0, 64, 219, 103, 220, 39, 0, 61, 0, 39, 0, 64, 219, 104, 220, 39, 0, 61, 0, 39, 0, 64, 219, 105, 220, 39, 0, 61, 0, 39, 0, 64, 219, 106, 220, 39, 0, 61, 0, 39, 0, 64, 219, 107, 220, 39, 0, 61, 0, 39, 0, 64, 219, 108, 220, 39, 0, 61, 0, 39, 0, 64, 219, 109, 220, 39, 0, 61, 0, 39, 0, 64, 219, 110, 220, 39, 0, 61, 0, 39, 0, 64, 219, 111, 220, 39, 0, 61, 0, 39, 0, 64, 219, 112, 220, 39, 0, 61, 0, 39, 0, 64, 219, 113, 220, 39, 0, 61, 0, 39, 0, 64, 219, 114, 220, 39, 0, 61, 0, 39, 0, 64, 219, 115, 220, 39, 0, 61, 0, 39, 0, 64, 219, 116, 220, 39, 0, 61, 0, 39, 0, 64, 219, 117, 220, 39, 0, 61, 0, 39, 0, 64, 219, 118, 220, 39, 0, 61, 0, 39, 0, 64, 219, 119, 220, 39, 0, 61, 0, 39, 0, 64, 219, 120, 220, 39, 0, 61, 0, 39, 0, 64, 219, 121, 220, 39, 0, 61, 0, 39, 0, 64, 219, 122, 220, 39, 0, 61, 0, 39, 0, 64, 219, 123, 220, 39, 0, 61, 0, 39, 0, 64, 219, 124, 220, 39, 0, 61, 0, 39, 0, 64, 219, 125, 220, 39, 0, 61, 0, 39, 0, 64, 219, 126, 220, 39, 0, 61, 0, 39, 0, 64, 219, 127, 220, 39, 0, 61, 0, 64, 219, 0, 221, 61, 0, 64, 219, 1, 221, 61, 0, 64, 219, 2, 221, 61, 0, 64, 219, 3, 221, 61, 0, 64, 219, 4, 221, 61, 0, 64, 219, 5, 221, 61, 0, 64, 219, 6, 221, 61, 0, 64, 219, 7, 221, 61, 0, 64, 219, 8, 221, 61, 0, 64, 219, 9, 221, 61, 0, 64, 219, 10, 221, 61, 0, 64, 219, 11, 221, 61, 0, 64, 219, 12, 221, 61, 0, 64, 219, 13, 221, 61, 0, 64, 219, 14, 221, 61, 0, 64, 219, 15, 221, 61, 0, 64, 219, 16, 221, 61, 0, 64, 219, 17, 221, 61, 0, 64, 219, 18, 221, 61, 0, 64, 219, 19, 221, 61, 0, 64, 219, 20, 221, 61, 0, 64, 219, 21, 221, 61, 0, 64, 219, 22, 221, 61, 0, 64, 219, 23, 221, 61, 0, 64, 219, 24, 221, 61, 0, 64, 219, 25, 221, 61, 0, 64, 219, 26, 221, 61, 0, 64, 219, 27, 221, 61, 0, 64, 219, 28, 221, 61, 0, 64, 219, 29, 221, 61, 0, 64, 219, 30, 221, 61, 0, 64, 219, 31, 221, 61, 0, 64, 219, 32, 221, 61, 0, 64, 219, 33, 221, 61, 0, 64, 219, 34, 221, 61, 0, 64, 219, 35, 221, 61, 0, 64, 219, 36, 221, 61, 0, 64, 219, 37, 221, 61, 0, 64, 219, 38, 221, 61, 0, 64, 219, 39, 221, 61, 0, 64, 219, 40, 221, 61, 0, 64, 219, 41, 221, 61, 0, 64, 219, 42, 221, 61, 0, 64, 219, 43, 221, 61, 0, 64, 219, 44, 221, 61, 0, 64, 219, 45, 221, 61, 0, 64, 219, 46, 221, 61, 0, 64, 219, 47, 221, 61, 0, 64, 219, 48, 221, 61, 0, 64, 219, 49, 221, 61, 0, 64, 219, 50, 221, 61, 0, 64, 219, 51, 221, 61, 0, 64, 219, 52, 221, 61, 0, 64, 219, 53, 221, 61, 0, 64, 219, 54, 221, 61, 0, 64, 219, 55, 221, 61, 0, 64, 219, 56, 221, 61, 0, 64, 219, 57, 221, 61, 0, 64, 219, 58, 221, 61, 0, 64, 219, 59, 221, 61, 0, 64, 219, 60, 221, 61, 0, 64, 219, 61, 221, 61, 0, 64, 219, 62, 221, 61, 0, 64, 219, 63, 221, 61, 0, 64, 219, 64, 221, 61, 0, 64, 219, 65, 221, 61, 0, 64, 219, 66, 221, 61, 0, 64, 219, 67, 221, 61, 0, 64, 219, 68, 221, 61, 0, 64, 219, 69, 221, 61, 0, 64, 219, 70, 221, 61, 0, 64, 219, 71, 221, 61, 0, 64, 219, 72, 221, 61, 0, 64, 219, 73, 221, 61, 0, 64, 219, 74, 221, 61, 0, 64, 219, 75, 221, 61, 0, 64, 219, 76, 221, 61, 0, 64, 219, 77, 221, 61, 0, 64, 219, 78, 221, 61, 0, 64, 219, 79, 221, 61, 0, 64, 219, 80, 221, 61, 0, 64, 219, 81, 221, 61, 0, 64, 219, 82, 221, 61, 0, 64, 219, 83, 221, 61, 0, 64, 219, 84, 221, 61, 0, 64, 219, 85, 221, 61, 0, 64, 219, 86, 221, 61, 0, 64, 219, 87, 221, 61, 0, 64, 219, 88, 221, 61, 0, 64, 219, 89, 221, 61, 0, 64, 219, 90, 221, 61, 0, 64, 219, 91, 221, 61, 0, 64, 219, 92, 221, 61, 0, 64, 219, 93, 221, 61, 0, 64, 219, 94, 221, 61, 0, 64, 219, 95, 221, 61, 0, 64, 219, 96, 221, 61, 0, 64, 219, 97, 221, 61, 0, 64, 219, 98, 221, 61, 0, 64, 219, 99, 221, 61, 0, 64, 219, 100, 221, 61, 0, 64, 219, 101, 221, 61, 0, 64, 219, 102, 221, 61, 0, 64, 219, 103, 221, 61, 0, 64, 219, 104, 221, 61, 0, 64, 219, 105, 221, 61, 0, 64, 219, 106, 221, 61, 0, 64, 219, 107, 221, 61, 0, 64, 219, 108, 221, 61, 0, 64, 219, 109, 221, 61, 0, 64, 219, 110, 221, 61, 0, 64, 219, 111, 221, 61, 0, 64, 219, 112, 221, 61, 0, 64, 219, 113, 221, 61, 0, 64, 219, 114, 221, 61, 0, 64, 219, 115, 221, 61, 0, 64, 219, 116, 221, 61, 0, 64, 219, 117, 221, 61, 0, 64, 219, 118, 221, 61, 0, 64, 219, 119, 221, 61, 0, 64, 219, 120, 221, 61, 0, 64, 219, 121, 221, 61, 0, 64, 219, 122, 221, 61, 0, 64, 219, 123, 221, 61, 0, 64, 219, 124, 221, 61, 0, 64, 219, 125, 221, 61, 0, 64, 219, 126, 221, 61, 0, 64, 219, 127, 221, 61, 0, 64, 219, 128, 221, 61, 0, 64, 219, 129, 221, 61, 0, 64, 219, 130, 221, 61, 0, 64, 219, 131, 221, 61, 0, 64, 219, 132, 221, 61, 0, 64, 219, 133, 221, 61, 0, 64, 219, 134, 221, 61, 0, 64, 219, 135, 221, 61, 0, 64, 219, 136, 221, 61, 0, 64, 219, 137, 221, 61, 0, 64, 219, 138, 221, 61, 0, 64, 219, 139, 221, 61, 0, 64, 219, 140, 221, 61, 0, 64, 219, 141, 221, 61, 0, 64, 219, 142, 221, 61, 0, 64, 219, 143, 221, 61, 0, 64, 219, 144, 221, 61, 0, 64, 219, 145, 221, 61, 0, 64, 219, 146, 221, 61, 0, 64, 219, 147, 221, 61, 0, 64, 219, 148, 221, 61, 0, 64, 219, 149, 221, 61, 0, 64, 219, 150, 221, 61, 0, 64, 219, 151, 221, 61, 0, 64, 219, 152, 221, 61, 0, 64, 219, 153, 221, 61, 0, 64, 219, 154, 221, 61, 0, 64, 219, 155, 221, 61, 0, 64, 219, 156, 221, 61, 0, 64, 219, 157, 221, 61, 0, 64, 219, 158, 221, 61, 0, 64, 219, 159, 221, 61, 0, 64, 219, 160, 221, 61, 0, 64, 219, 161, 221, 61, 0, 64, 219, 162, 221, 61, 0, 64, 219, 163, 221, 61, 0, 64, 219, 164, 221, 61, 0, 64, 219, 165, 221, 61, 0, 64, 219, 166, 221, 61, 0, 64, 219, 167, 221, 61, 0, 64, 219, 168, 221, 61, 0, 64, 219, 169, 221, 61, 0, 64, 219, 170, 221, 61, 0, 64, 219, 171, 221, 61, 0, 64, 219, 172, 221, 61, 0, 64, 219, 173, 221, 61, 0, 64, 219, 174, 221, 61, 0, 64, 219, 175, 221, 61, 0, 64, 219, 176, 221, 61, 0, 64, 219, 177, 221, 61, 0, 64, 219, 178, 221, 61, 0, 64, 219, 179, 221, 61, 0, 64, 219, 180, 221, 61, 0, 64, 219, 181, 221, 61, 0, 64, 219, 182, 221, 61, 0, 64, 219, 183, 221, 61, 0, 64, 219, 184, 221, 61, 0, 64, 219, 185, 221, 61, 0, 64, 219, 186, 221, 61, 0, 64, 219, 187, 221, 61, 0, 64, 219, 188, 221, 61, 0, 64, 219, 189, 221, 61, 0, 64, 219, 190, 221, 61, 0, 64, 219, 191, 221, 61, 0, 64, 219, 192, 221, 61, 0, 64, 219, 193, 221, 61, 0, 64, 219, 194, 221, 61, 0, 64, 219, 195, 221, 61, 0, 64, 219, 196, 221, 61, 0, 64, 219, 197, 221, 61, 0, 64, 219, 198, 221, 61, 0, 64, 219, 199, 221, 61, 0, 64, 219, 200, 221, 61, 0, 64, 219, 201, 221, 61, 0, 64, 219, 202, 221, 61, 0, 64, 219, 203, 221, 61, 0, 64, 219, 204, 221, 61, 0, 64, 219, 205, 221, 61, 0, 64, 219, 206, 221, 61, 0, 64, 219, 207, 221, 61, 0, 64, 219, 208, 221, 61, 0, 64, 219, 209, 221, 61, 0, 64, 219, 210, 221, 61, 0, 64, 219, 211, 221, 61, 0, 64, 219, 212, 221, 61, 0, 64, 219, 213, 221, 61, 0, 64, 219, 214, 221, 61, 0, 64, 219, 215, 221, 61, 0, 64, 219, 216, 221, 61, 0, 64, 219, 217, 221, 61, 0, 64, 219, 218, 221, 61, 0, 64, 219, 219, 221, 61, 0, 64, 219, 220, 221, 61, 0, 64, 219, 221, 221, 61, 0, 64, 219, 222, 221, 61, 0, 64, 219, 223, 221, 61, 0, 64, 219, 224, 221, 61, 0, 64, 219, 225, 221, 61, 0, 64, 219, 226, 221, 61, 0, 64, 219, 227, 221, 61, 0, 64, 219, 228, 221, 61, 0, 64, 219, 229, 221, 61, 0, 64, 219, 230, 221, 61, 0, 64, 219, 231, 221, 61, 0, 64, 219, 232, 221, 61, 0, 64, 219, 233, 221, 61, 0, 64, 219, 234, 221, 61, 0, 64, 219, 235, 221, 61, 0, 64, 219, 236, 221, 61, 0, 64, 219, 237, 221, 61, 0, 64, 219, 238, 221, 61, 0, 64, 219, 239, 221, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 112, 0, 114, 0, 105, 0, 109, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 60, 0, 60, 0, 50, 3, 60, 0, 60, 0, 19, 3, 61, 0, 134, 4, 61, 0, 241, 44, 60, 0, 60, 0, 20, 3, 61, 0, 133, 4, 61, 0, 240, 44, 60, 0, 60, 0, 1, 3, 61, 0, 84, 9, 60, 0, 60, 0, 0, 3, 61, 0, 83, 9, 60, 0, 60, 0, 6, 3, 60, 0, 60, 0, 2, 3, 60, 0, 60, 0, 12, 3, 60, 0, 60, 0, 10, 3, 60, 0, 60, 0, 66, 3, 60, 0, 60, 0, 8, 3, 60, 0, 60, 0, 11, 3, 60, 0, 60, 0, 3, 3, 60, 0, 60, 0, 7, 3, 60, 0, 60, 0, 56, 3, 60, 0, 60, 0, 39, 3, 60, 0, 60, 0, 40, 3, 60, 0, 60, 0, 4, 3, 60, 0, 60, 0, 13, 3, 61, 0, 14, 3, 61, 0, 18, 3, 61, 0, 21, 3, 61, 0, 26, 3, 61, 0, 61, 3, 61, 0, 62, 3, 61, 0, 63, 3, 61, 0, 70, 3, 61, 0, 74, 3, 61, 0, 75, 3, 61, 0, 76, 3, 61, 0, 80, 3, 61, 0, 81, 3, 61, 0, 82, 3, 61, 0, 87, 3, 61, 0, 91, 3, 61, 0, 93, 3, 61, 0, 94, 3, 61, 0, 132, 4, 61, 0, 135, 4, 61, 0, 65, 7, 61, 0, 69, 7, 61, 0, 85, 11, 61, 0, 203, 23, 61, 0, 204, 23, 61, 0, 205, 23, 61, 0, 206, 23, 61, 0, 207, 23, 61, 0, 208, 23, 61, 0, 209, 23, 61, 0, 221, 23, 61, 0, 176, 26, 61, 0, 177, 26, 61, 0, 178, 26, 61, 0, 179, 26, 61, 0, 180, 26, 61, 0, 187, 26, 61, 0, 188, 26, 61, 0, 192, 29, 61, 0, 193, 29, 61, 0, 195, 29, 61, 0, 196, 29, 61, 0, 197, 29, 61, 0, 198, 29, 61, 0, 199, 29, 61, 0, 200, 29, 61, 0, 201, 29, 61, 0, 203, 29, 61, 0, 204, 29, 61, 0, 205, 29, 61, 0, 206, 29, 61, 0, 209, 29, 61, 0, 245, 29, 61, 0, 246, 29, 61, 0, 247, 29, 61, 0, 248, 29, 61, 0, 251, 29, 61, 0, 254, 29, 61, 0, 240, 32, 61, 0, 239, 44, 61, 0, 124, 166, 61, 0, 125, 166, 61, 0, 2, 216, 229, 222, 61, 0, 3, 216, 36, 221, 61, 0, 3, 216, 37, 221, 61, 0, 3, 216, 38, 221, 61, 0, 3, 216, 39, 221, 61, 0, 3, 216, 72, 223, 61, 0, 3, 216, 73, 223, 61, 0, 3, 216, 74, 223, 61, 0, 3, 216, 76, 223, 61, 0, 47, 216, 157, 220, 60, 0, 60, 0, 22, 3, 61, 0, 23, 3, 61, 0, 24, 3, 61, 0, 25, 3, 61, 0, 28, 3, 61, 0, 29, 3, 61, 0, 30, 3, 61, 0, 31, 3, 61, 0, 32, 3, 61, 0, 41, 3, 61, 0, 42, 3, 61, 0, 43, 3, 61, 0, 44, 3, 61, 0, 47, 3, 61, 0, 51, 3, 61, 0, 58, 3, 61, 0, 59, 3, 61, 0, 60, 3, 61, 0, 71, 3, 61, 0, 72, 3, 61, 0, 73, 3, 61, 0, 77, 3, 61, 0, 78, 3, 61, 0, 83, 3, 61, 0, 84, 3, 61, 0, 85, 3, 61, 0, 86, 3, 61, 0, 89, 3, 61, 0, 90, 3, 61, 0, 92, 3, 61, 0, 95, 3, 61, 0, 98, 3, 61, 0, 66, 7, 61, 0, 70, 7, 61, 0, 253, 7, 61, 0, 89, 8, 61, 0, 90, 8, 61, 0, 91, 8, 61, 0, 181, 26, 61, 0, 182, 26, 61, 0, 183, 26, 61, 0, 184, 26, 61, 0, 185, 26, 61, 0, 186, 26, 61, 0, 189, 26, 61, 0, 194, 29, 61, 0, 207, 29, 61, 0, 208, 29, 61, 0, 249, 29, 61, 0, 252, 29, 61, 0, 253, 29, 61, 0, 255, 29, 61, 0, 236, 32, 61, 0, 237, 32, 61, 0, 238, 32, 61, 0, 239, 32, 61, 0, 39, 254, 61, 0, 2, 216, 13, 222, 61, 0, 2, 216, 230, 222, 61, 0, 3, 216, 70, 223, 61, 0, 3, 216, 71, 223, 61, 0, 3, 216, 75, 223, 61, 0, 3, 216, 77, 223, 61, 0, 3, 216, 78, 223, 61, 0, 3, 216, 79, 223, 61, 0, 3, 216, 80, 223, 60, 0, 60, 0, 54, 3, 61, 0, 55, 3, 61, 0, 216, 32, 61, 0, 217, 32, 61, 0, 218, 32, 61, 0, 229, 32, 61, 0, 234, 32, 61, 0, 235, 32, 61, 0, 47, 216, 158, 220, 60, 0, 60, 0, 190, 26, 61, 0, 221, 32, 61, 0, 222, 32, 61, 0, 223, 32, 61, 0, 224, 32, 61, 0, 226, 32, 61, 0, 227, 32, 61, 0, 228, 32, 60, 0, 60, 0, 153, 48, 60, 0, 60, 0, 60, 0, 158, 255, 60, 0, 60, 0, 154, 48, 60, 0, 60, 0, 60, 0, 159, 255, 60, 0, 60, 0, 53, 3, 60, 0, 60, 0, 5, 3, 60, 0, 60, 0, 9, 3, 60, 0, 60, 0, 15, 3, 60, 0, 60, 0, 16, 3, 60, 0, 60, 0, 17, 3, 60, 0, 60, 0, 27, 3, 60, 0, 60, 0, 33, 3, 60, 0, 60, 0, 34, 3, 60, 0, 60, 0, 35, 3, 60, 0, 60, 0, 36, 3, 60, 0, 60, 0, 37, 3, 60, 0, 60, 0, 38, 3, 60, 0, 60, 0, 45, 3, 60, 0, 60, 0, 46, 3, 60, 0, 60, 0, 48, 3, 60, 0, 60, 0, 49, 3, 60, 0, 60, 0, 52, 3, 60, 0, 60, 0, 57, 3, 60, 0, 60, 0, 69, 3, 60, 0, 60, 0, 88, 3, 60, 0, 60, 0, 96, 3, 61, 0, 34, 254, 61, 0, 41, 254, 60, 0, 60, 0, 97, 3, 61, 0, 32, 254, 60, 0, 60, 0, 131, 4, 61, 0, 46, 254, 60, 0, 60, 0, 111, 166, 60, 0, 60, 0, 176, 5, 60, 0, 60, 0, 177, 5, 60, 0, 60, 0, 178, 5, 60, 0, 60, 0, 179, 5, 60, 0, 60, 0, 180, 5, 60, 0, 60, 0, 181, 5, 60, 0, 60, 0, 182, 5, 60, 0, 60, 0, 183, 5, 60, 0, 60, 0, 184, 5, 61, 0, 199, 5, 60, 0, 60, 0, 185, 5, 61, 0, 186, 5, 60, 0, 60, 0, 187, 5, 60, 0, 60, 0, 194, 5, 60, 0, 60, 0, 193, 5, 60, 0, 60, 0, 188, 5, 60, 0, 60, 0, 191, 5, 60, 0, 60, 0, 30, 251, 60, 0, 60, 0, 28, 8, 61, 0, 29, 8, 60, 0, 60, 0, 30, 8, 61, 0, 31, 8, 61, 0, 32, 8, 60, 0, 60, 0, 33, 8, 61, 0, 34, 8, 61, 0, 35, 8, 60, 0, 60, 0, 14, 32, 36, 8, 14, 32, 61, 0, 37, 8, 60, 0, 60, 0, 38, 8, 61, 0, 39, 8, 60, 0, 60, 0, 14, 32, 40, 8, 14, 32, 61, 0, 41, 8, 61, 0, 42, 8, 60, 0, 60, 0, 43, 8, 60, 0, 60, 0, 44, 8, 60, 0, 60, 0, 24, 8, 60, 0, 60, 0, 25, 8, 60, 0, 60, 0, 45, 8, 60, 0, 60, 0, 75, 6, 60, 0, 60, 0, 60, 0, 14, 32, 113, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 112, 254, 14, 32, 60, 0, 60, 0, 240, 8, 60, 0, 60, 0, 231, 8, 60, 0, 60, 0, 76, 6, 60, 0, 60, 0, 60, 0, 14, 32, 114, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 94, 252, 14, 32, 47, 0, 4, 216, 55, 222, 60, 0, 60, 0, 241, 8, 60, 0, 60, 0, 232, 8, 60, 0, 60, 0, 77, 6, 60, 0, 60, 0, 60, 0, 14, 32, 116, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 95, 252, 14, 32, 47, 0, 4, 216, 55, 222, 60, 0, 60, 0, 242, 8, 60, 0, 60, 0, 233, 8, 60, 0, 60, 0, 78, 6, 60, 0, 60, 0, 60, 0, 14, 32, 119, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 242, 252, 14, 32, 47, 0, 4, 216, 55, 222, 60, 0, 60, 0, 60, 0, 14, 32, 118, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 96, 252, 14, 32, 47, 0, 4, 216, 55, 222, 60, 0, 60, 0, 228, 8, 60, 0, 60, 0, 244, 8, 60, 0, 60, 0, 245, 8, 60, 0, 60, 0, 79, 6, 60, 0, 60, 0, 60, 0, 14, 32, 121, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 243, 252, 14, 32, 47, 0, 4, 216, 55, 222, 60, 0, 60, 0, 60, 0, 14, 32, 120, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 97, 252, 14, 32, 47, 0, 4, 216, 55, 222, 60, 0, 60, 0, 229, 8, 60, 0, 60, 0, 254, 8, 60, 0, 60, 0, 80, 6, 60, 0, 60, 0, 60, 0, 14, 32, 123, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 244, 252, 14, 32, 47, 0, 4, 216, 55, 222, 60, 0, 60, 0, 60, 0, 14, 32, 122, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 98, 252, 14, 32, 47, 0, 4, 216, 55, 222, 60, 0, 60, 0, 230, 8, 60, 0, 60, 0, 246, 8, 60, 0, 60, 0, 81, 6, 61, 0, 251, 10, 61, 0, 4, 216, 55, 222, 60, 0, 60, 0, 60, 0, 14, 32, 125, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 124, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 99, 252, 14, 32, 47, 0, 112, 6, 60, 0, 60, 0, 82, 6, 61, 0, 250, 10, 61, 0, 4, 216, 62, 222, 60, 0, 60, 0, 60, 0, 14, 32, 127, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 126, 254, 14, 32, 60, 0, 60, 0, 83, 6, 61, 0, 252, 10, 61, 0, 3, 216, 172, 222, 60, 0, 60, 0, 84, 6, 61, 0, 3, 216, 171, 222, 60, 0, 60, 0, 85, 6, 60, 0, 60, 0, 95, 6, 60, 0, 60, 0, 86, 6, 60, 0, 60, 0, 87, 6, 60, 0, 60, 0, 88, 6, 60, 0, 60, 0, 255, 8, 60, 0, 60, 0, 89, 6, 60, 0, 60, 0, 90, 6, 60, 0, 60, 0, 91, 6, 60, 0, 60, 0, 92, 6, 60, 0, 60, 0, 93, 6, 60, 0, 60, 0, 94, 6, 60, 0, 60, 0, 227, 8, 60, 0, 60, 0, 247, 8, 60, 0, 60, 0, 248, 8, 60, 0, 60, 0, 253, 8, 60, 0, 60, 0, 251, 8, 60, 0, 60, 0, 252, 8, 60, 0, 60, 0, 249, 8, 60, 0, 60, 0, 250, 8, 60, 0, 60, 0, 112, 6, 60, 0, 60, 0, 17, 7, 60, 0, 60, 0, 48, 7, 60, 0, 60, 0, 49, 7, 60, 0, 60, 0, 50, 7, 60, 0, 60, 0, 51, 7, 60, 0, 60, 0, 52, 7, 60, 0, 60, 0, 53, 7, 60, 0, 60, 0, 54, 7, 60, 0, 60, 0, 55, 7, 60, 0, 60, 0, 56, 7, 60, 0, 60, 0, 57, 7, 60, 0, 60, 0, 58, 7, 60, 0, 60, 0, 59, 7, 60, 0, 60, 0, 60, 7, 60, 0, 60, 0, 61, 7, 60, 0, 60, 0, 62, 7, 60, 0, 60, 0, 63, 7, 60, 0, 60, 0, 235, 7, 60, 0, 60, 0, 236, 7, 60, 0, 60, 0, 237, 7, 60, 0, 60, 0, 238, 7, 60, 0, 60, 0, 239, 7, 60, 0, 60, 0, 240, 7, 60, 0, 60, 0, 241, 7, 60, 0, 60, 0, 242, 7, 60, 0, 60, 0, 243, 7, 60, 0, 60, 0, 95, 19, 60, 0, 60, 0, 94, 19, 60, 0, 60, 0, 93, 19, 60, 0, 60, 0, 240, 166, 60, 0, 60, 0, 241, 166, 60, 0, 60, 0, 26, 216, 240, 222, 60, 0, 60, 0, 26, 216, 241, 222, 60, 0, 60, 0, 26, 216, 242, 222, 60, 0, 60, 0, 26, 216, 243, 222, 60, 0, 60, 0, 26, 216, 244, 222, 60, 0, 60, 0, 58, 216, 68, 221, 61, 0, 58, 216, 69, 221, 61, 0, 58, 216, 70, 221, 60, 0, 60, 0, 58, 216, 74, 221, 60, 0, 60, 0, 58, 216, 71, 221, 60, 0, 60, 0, 58, 216, 72, 221, 60, 0, 60, 0, 58, 216, 73, 221, 60, 0, 60, 0, 60, 9, 61, 0, 188, 9, 61, 0, 60, 10, 61, 0, 188, 10, 61, 0, 253, 10, 61, 0, 254, 10, 61, 0, 255, 10, 61, 0, 60, 11, 61, 0, 188, 12, 61, 0, 52, 27, 61, 0, 230, 27, 61, 0, 55, 28, 61, 0, 179, 169, 61, 0, 4, 216, 186, 220, 61, 0, 4, 216, 115, 221, 61, 0, 4, 216, 202, 221, 61, 0, 4, 216, 54, 222, 61, 0, 4, 216, 233, 222, 61, 0, 4, 216, 59, 223, 61, 0, 4, 216, 60, 223, 61, 0, 5, 216, 70, 220, 61, 0, 5, 216, 195, 220, 61, 0, 5, 216, 192, 221, 61, 0, 5, 216, 183, 222, 61, 0, 6, 216, 58, 220, 61, 0, 6, 216, 67, 221, 61, 0, 6, 216, 51, 222, 61, 0, 7, 216, 66, 221, 60, 0, 60, 0, 0, 9, 61, 0, 1, 9, 61, 0, 129, 9, 61, 0, 1, 10, 61, 0, 129, 10, 61, 0, 1, 11, 61, 0, 0, 12, 61, 0, 1, 12, 61, 0, 129, 12, 61, 0, 1, 13, 61, 0, 129, 13, 61, 0, 0, 27, 61, 0, 1, 27, 61, 0, 197, 168, 61, 0, 128, 169, 61, 0, 4, 216, 0, 220, 61, 0, 4, 216, 128, 220, 61, 0, 4, 216, 0, 221, 61, 0, 4, 216, 128, 221, 61, 0, 4, 216, 207, 221, 61, 0, 4, 216, 1, 223, 61, 0, 5, 216, 67, 220, 61, 0, 5, 216, 191, 220, 61, 0, 5, 216, 188, 221, 61, 0, 5, 216, 64, 222, 61, 0, 6, 216, 60, 221, 61, 0, 6, 216, 53, 222, 61, 0, 6, 216, 54, 222, 61, 0, 6, 216, 55, 222, 61, 0, 7, 216, 60, 220, 61, 0, 7, 216, 182, 220, 61, 0, 7, 216, 67, 221, 60, 0, 60, 0, 2, 9, 61, 0, 130, 9, 61, 0, 2, 10, 61, 0, 130, 10, 61, 0, 2, 11, 61, 0, 130, 11, 61, 0, 2, 12, 61, 0, 4, 12, 61, 0, 130, 12, 61, 0, 0, 13, 61, 0, 2, 13, 61, 0, 130, 13, 61, 0, 126, 15, 61, 0, 54, 16, 61, 0, 198, 23, 61, 0, 116, 26, 61, 0, 2, 27, 61, 0, 128, 27, 61, 0, 237, 28, 61, 0, 11, 168, 61, 0, 128, 168, 61, 0, 129, 169, 61, 0, 2, 216, 14, 222, 61, 0, 4, 216, 1, 220, 61, 0, 4, 216, 129, 220, 61, 0, 4, 216, 1, 221, 61, 0, 4, 216, 129, 221, 61, 0, 4, 216, 52, 222, 61, 0, 4, 216, 223, 222, 61, 0, 4, 216, 0, 223, 61, 0, 4, 216, 2, 223, 61, 0, 5, 216, 68, 220, 61, 0, 5, 216, 192, 220, 61, 0, 5, 216, 189, 221, 61, 0, 5, 216, 61, 222, 61, 0, 5, 216, 171, 222, 61, 0, 6, 216, 55, 220, 61, 0, 6, 216, 59, 221, 61, 0, 6, 216, 222, 221, 61, 0, 6, 216, 56, 222, 61, 0, 6, 216, 150, 222, 61, 0, 7, 216, 61, 220, 61, 0, 7, 216, 181, 220, 61, 0, 7, 216, 64, 221, 61, 0, 7, 216, 149, 221, 60, 0, 60, 0, 3, 9, 61, 0, 131, 9, 61, 0, 3, 10, 61, 0, 131, 10, 61, 0, 3, 11, 61, 0, 3, 12, 61, 0, 131, 12, 61, 0, 3, 13, 61, 0, 131, 13, 61, 0, 127, 15, 61, 0, 56, 16, 61, 0, 199, 23, 61, 0, 4, 27, 61, 0, 130, 27, 61, 0, 242, 28, 61, 0, 243, 28, 61, 0, 129, 168, 61, 0, 131, 169, 61, 0, 2, 216, 15, 222, 61, 0, 4, 216, 2, 220, 61, 0, 4, 216, 130, 220, 61, 0, 4, 216, 2, 221, 61, 0, 4, 216, 130, 221, 61, 0, 4, 216, 3, 223, 61, 0, 5, 216, 69, 220, 61, 0, 5, 216, 193, 220, 61, 0, 5, 216, 190, 221, 61, 0, 5, 216, 62, 222, 61, 0, 5, 216, 172, 222, 61, 0, 6, 216, 56, 220, 61, 0, 6, 216, 223, 221, 61, 0, 6, 216, 57, 222, 61, 0, 6, 216, 151, 222, 61, 0, 7, 216, 62, 220, 61, 0, 7, 216, 65, 221, 61, 0, 7, 216, 150, 221, 60, 0, 60, 0, 254, 9, 61, 0, 4, 216, 201, 221, 61, 0, 5, 216, 94, 220, 60, 0, 60, 0, 112, 10, 60, 0, 60, 0, 113, 10, 60, 0, 60, 0, 3, 27, 60, 0, 60, 0, 130, 169, 60, 0, 60, 0, 129, 27, 60, 0, 60, 0, 236, 171, 60, 0, 60, 0, 2, 216, 56, 222, 60, 0, 60, 0, 2, 216, 57, 222, 60, 0, 60, 0, 2, 216, 58, 222, 60, 0, 60, 0, 4, 216, 203, 221, 60, 0, 60, 0, 4, 216, 204, 221, 60, 0, 60, 0, 6, 216, 152, 222, 60, 0, 60, 0, 78, 14, 60, 0, 60, 0, 71, 14, 60, 0, 60, 0, 72, 14, 60, 0, 60, 0, 73, 14, 60, 0, 60, 0, 74, 14, 60, 0, 60, 0, 75, 14, 60, 0, 60, 0, 76, 14, 60, 0, 60, 0, 77, 14, 60, 0, 60, 0, 200, 14, 60, 0, 60, 0, 201, 14, 60, 0, 60, 0, 202, 14, 60, 0, 60, 0, 203, 14, 60, 0, 60, 0, 204, 14, 60, 0, 60, 0, 205, 14, 60, 0, 60, 0, 191, 170, 60, 0, 60, 0, 193, 170, 60, 0, 60, 0, 57, 15, 60, 0, 60, 0, 43, 169, 60, 0, 60, 0, 44, 169, 60, 0, 60, 0, 45, 169, 60, 0, 60, 0, 55, 16, 60, 0, 60, 0, 200, 23, 60, 0, 60, 0, 201, 23, 60, 0, 60, 0, 202, 23, 60, 0, 60, 0, 117, 26, 60, 0, 60, 0, 118, 26, 60, 0, 60, 0, 119, 26, 60, 0, 60, 0, 120, 26, 60, 0, 60, 0, 121, 26, 60, 0, 60, 0, 122, 26, 60, 0, 60, 0, 123, 26, 60, 0, 60, 0, 124, 26, 60, 0, 60, 0, 57, 25, 60, 0, 60, 0, 58, 25, 60, 0, 60, 0, 59, 25, 60, 0, 60, 0, 26, 216, 48, 223, 61, 0, 56, 216, 49, 221, 60, 0, 60, 0, 26, 216, 49, 223, 61, 0, 56, 216, 54, 221, 60, 0, 60, 0, 26, 216, 50, 223, 61, 0, 56, 216, 50, 221, 60, 0, 60, 0, 26, 216, 51, 223, 61, 0, 56, 216, 51, 221, 60, 0, 60, 0, 26, 216, 52, 223, 61, 0, 56, 216, 48, 221, 60, 0, 60, 0, 26, 216, 53, 223, 61, 0, 56, 216, 52, 221, 60, 0, 60, 0, 26, 216, 54, 223, 61, 0, 56, 216, 53, 221, 60, 0, 60, 0, 56, 216, 236, 222, 60, 0, 60, 0, 56, 216, 237, 222, 60, 0, 60, 0, 56, 216, 238, 222, 60, 0, 60, 0, 56, 216, 239, 222, 60, 0, 60, 0, 42, 48, 60, 0, 60, 0, 43, 48, 60, 0, 60, 0, 44, 48, 60, 0, 60, 0, 45, 48, 60, 0, 60, 0, 46, 48, 60, 0, 60, 0, 47, 48, 60, 0, 60, 0, 27, 216, 240, 223, 60, 0, 60, 0, 27, 216, 241, 223, 60, 0, 60, 0, 208, 32, 60, 0, 60, 0, 209, 32, 60, 0, 60, 0, 210, 32, 61, 0, 211, 32, 60, 0, 60, 0, 212, 32, 60, 0, 60, 0, 213, 32, 60, 0, 60, 0, 214, 32, 60, 0, 60, 0, 215, 32, 60, 0, 60, 0, 219, 32, 60, 0, 60, 0, 220, 32, 60, 0, 60, 0, 225, 32, 60, 0, 60, 0, 230, 32, 60, 0, 60, 0, 231, 32, 60, 0, 60, 0, 232, 32, 60, 0, 60, 0, 233, 32, 60, 0, 60, 0, 0, 216, 253, 221, 38, 0, 39, 0, 47, 32, 39, 0, 60, 0, 39, 0, 9, 0, 39, 0, 60, 0, 39, 0, 10, 0, 39, 0, 60, 0, 39, 0, 11, 0, 39, 0, 60, 0, 39, 0, 12, 0, 39, 0, 60, 0, 39, 0, 13, 0, 39, 0, 60, 0, 39, 0, 133, 0, 39, 0, 60, 0, 39, 0, 40, 32, 39, 0, 60, 0, 39, 0, 41, 32, 39, 0, 60, 0, 39, 0, 32, 0, 39, 0, 60, 0, 60, 0, 60, 0, 39, 0, 0, 48, 39, 0, 60, 0, 60, 0, 60, 0, 39, 0, 128, 22, 39, 0, 61, 0, 39, 0, 2, 32, 39, 0, 61, 0, 39, 0, 3, 32, 39, 0, 61, 0, 39, 0, 4, 32, 39, 0, 61, 0, 39, 0, 5, 32, 39, 0, 61, 0, 39, 0, 6, 32, 39, 0, 61, 0, 39, 0, 8, 32, 39, 0, 61, 0, 39, 0, 9, 32, 39, 0, 61, 0, 39, 0, 10, 32, 39, 0, 61, 0, 39, 0, 95, 32, 39, 0, 60, 0, 60, 0, 60, 0, 39, 0, 160, 0, 39, 0, 61, 0, 39, 0, 7, 32, 39, 0, 61, 0, 39, 0, 47, 32, 39, 0, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 118, 0, 97, 0, 114, 0, 105, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 60, 0, 62, 32, 60, 0, 60, 0, 60, 0, 73, 254, 61, 0, 74, 254, 61, 0, 75, 254, 61, 0, 76, 254, 60, 0, 39, 0, 95, 0, 39, 0, 60, 0, 60, 0, 60, 0, 63, 255, 60, 0, 60, 0, 60, 0, 77, 254, 61, 0, 78, 254, 61, 0, 79, 254, 60, 0, 60, 0, 60, 0, 51, 254, 61, 0, 52, 254, 60, 0, 23, 32, 60, 0, 39, 0, 45, 0, 39, 0, 60, 0, 60, 0, 60, 0, 13, 255, 60, 0, 60, 0, 60, 0, 99, 254, 60, 0, 138, 5, 60, 0, 0, 20, 60, 0, 96, 27, 60, 0, 6, 24, 60, 0, 7, 24, 60, 0, 16, 32, 60, 0, 60, 0, 60, 0, 17, 32, 60, 0, 18, 32, 60, 0, 19, 32, 60, 0, 60, 0, 60, 0, 50, 254, 60, 0, 20, 32, 60, 0, 60, 0, 60, 0, 88, 254, 60, 0, 60, 0, 60, 0, 49, 254, 60, 0, 21, 32, 60, 0, 58, 46, 60, 0, 59, 46, 60, 0, 83, 32, 60, 0, 67, 46, 60, 0, 23, 46, 60, 0, 64, 46, 60, 0, 28, 48, 60, 0, 48, 48, 60, 0, 160, 48, 60, 0, 251, 48, 60, 0, 60, 0, 60, 0, 101, 255, 60, 0, 14, 32, 3, 216, 173, 222, 14, 32, 60, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 12, 255, 60, 0, 60, 0, 60, 0, 80, 254, 60, 0, 60, 0, 60, 0, 16, 254, 60, 0, 52, 46, 60, 0, 50, 46, 60, 0, 65, 46, 60, 0, 76, 46, 60, 0, 78, 46, 60, 0, 79, 46, 60, 0, 93, 5, 60, 0, 12, 6, 60, 0, 14, 32, 13, 6, 14, 32, 60, 0, 107, 6, 60, 0, 108, 6, 60, 0, 248, 7, 60, 0, 2, 24, 60, 0, 8, 24, 60, 0, 254, 164, 60, 0, 13, 166, 60, 0, 245, 166, 60, 0, 27, 216, 151, 222, 60, 0, 1, 48, 60, 0, 60, 0, 60, 0, 81, 254, 60, 0, 60, 0, 60, 0, 100, 255, 60, 0, 60, 0, 60, 0, 17, 254, 60, 0, 69, 254, 60, 0, 70, 254, 60, 0, 27, 216, 226, 223, 60, 0, 39, 0, 59, 0, 39, 0, 60, 0, 60, 0, 60, 0, 27, 255, 60, 0, 60, 0, 60, 0, 84, 254, 60, 0, 60, 0, 60, 0, 20, 254, 60, 0, 14, 32, 27, 6, 14, 32, 60, 0, 79, 32, 60, 0, 53, 46, 60, 0, 246, 166, 60, 0, 73, 46, 60, 0, 39, 0, 58, 0, 39, 0, 60, 0, 60, 0, 60, 0, 26, 255, 60, 0, 60, 0, 60, 0, 116, 42, 47, 0, 39, 0, 58, 0, 61, 0, 39, 0, 60, 0, 60, 0, 60, 0, 85, 254, 60, 0, 60, 0, 60, 0, 19, 254, 60, 0, 137, 5, 60, 0, 14, 32, 30, 6, 14, 32, 60, 0, 14, 32, 3, 7, 14, 32, 60, 0, 14, 32, 4, 7, 14, 32, 60, 0, 14, 32, 5, 7, 14, 32, 60, 0, 14, 32, 6, 7, 14, 32, 60, 0, 14, 32, 7, 7, 14, 32, 60, 0, 14, 32, 8, 7, 14, 32, 60, 0, 14, 32, 48, 8, 14, 32, 60, 0, 14, 32, 49, 8, 14, 32, 60, 0, 14, 32, 50, 8, 14, 32, 60, 0, 14, 32, 51, 8, 14, 32, 60, 0, 14, 32, 52, 8, 14, 32, 60, 0, 14, 32, 53, 8, 14, 32, 60, 0, 14, 32, 54, 8, 14, 32, 60, 0, 14, 32, 55, 8, 14, 32, 60, 0, 14, 32, 56, 8, 14, 32, 60, 0, 14, 32, 57, 8, 14, 32, 60, 0, 14, 32, 58, 8, 14, 32, 60, 0, 14, 32, 59, 8, 14, 32, 60, 0, 14, 32, 60, 8, 14, 32, 60, 0, 14, 32, 61, 8, 14, 32, 60, 0, 14, 32, 62, 8, 14, 32, 60, 0, 97, 19, 60, 0, 99, 19, 60, 0, 100, 19, 60, 0, 101, 19, 60, 0, 102, 19, 60, 0, 4, 24, 60, 0, 5, 24, 60, 0, 20, 15, 60, 0, 214, 23, 60, 0, 93, 27, 60, 0, 199, 169, 60, 0, 235, 22, 60, 0, 236, 22, 60, 0, 237, 22, 60, 0, 244, 166, 60, 0, 39, 0, 33, 0, 39, 0, 60, 0, 60, 0, 60, 0, 1, 255, 60, 0, 60, 0, 60, 0, 60, 32, 47, 0, 39, 0, 33, 0, 39, 0, 60, 0, 60, 0, 60, 0, 73, 32, 47, 0, 39, 0, 63, 0, 39, 0, 60, 0, 60, 0, 60, 0, 87, 254, 60, 0, 60, 0, 60, 0, 21, 254, 60, 0, 161, 0, 60, 0, 92, 5, 60, 0, 249, 7, 60, 0, 68, 25, 60, 0, 14, 32, 58, 216, 94, 221, 14, 32, 60, 0, 39, 0, 63, 0, 39, 0, 60, 0, 60, 0, 60, 0, 31, 255, 60, 0, 60, 0, 60, 0, 72, 32, 47, 0, 39, 0, 33, 0, 39, 0, 60, 0, 60, 0, 60, 0, 71, 32, 47, 0, 39, 0, 63, 0, 39, 0, 60, 0, 60, 0, 60, 0, 86, 254, 60, 0, 60, 0, 60, 0, 22, 254, 60, 0, 191, 0, 60, 0, 46, 46, 60, 0, 94, 5, 60, 0, 14, 32, 31, 6, 14, 32, 60, 0, 14, 32, 9, 7, 14, 32, 60, 0, 103, 19, 60, 0, 69, 25, 60, 0, 250, 44, 60, 0, 251, 44, 60, 0, 15, 166, 60, 0, 247, 166, 60, 0, 241, 170, 60, 0, 4, 216, 67, 221, 60, 0, 14, 32, 58, 216, 95, 221, 14, 32, 60, 0, 61, 32, 60, 0, 24, 46, 60, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 14, 255, 60, 0, 60, 0, 60, 0, 36, 32, 60, 0, 60, 0, 60, 0, 37, 32, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 38, 32, 47, 0, 39, 0, 46, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 82, 254, 60, 0, 60, 0, 60, 0, 48, 254, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 25, 254, 47, 0, 39, 0, 46, 0, 46, 0, 39, 0, 60, 0, 1, 24, 60, 0, 14, 32, 212, 6, 14, 32, 60, 0, 14, 32, 1, 7, 14, 32, 60, 0, 14, 32, 2, 7, 14, 32, 60, 0, 98, 19, 60, 0, 3, 24, 60, 0, 9, 24, 60, 0, 110, 22, 60, 0, 92, 27, 60, 0, 249, 44, 60, 0, 254, 44, 60, 0, 48, 46, 60, 0, 60, 46, 60, 0, 255, 164, 60, 0, 14, 166, 60, 0, 243, 166, 60, 0, 26, 216, 245, 222, 60, 0, 27, 216, 152, 222, 60, 0, 47, 216, 159, 220, 60, 0, 2, 48, 60, 0, 60, 0, 60, 0, 97, 255, 60, 0, 60, 0, 60, 0, 18, 254, 60, 0, 183, 0, 60, 0, 49, 46, 60, 0, 51, 46, 60, 0, 100, 9, 60, 0, 101, 9, 60, 0, 206, 168, 60, 0, 207, 168, 60, 0, 59, 28, 60, 0, 60, 28, 60, 0, 118, 168, 60, 0, 119, 168, 60, 0, 53, 23, 60, 0, 54, 23, 60, 0, 47, 169, 60, 0, 74, 16, 60, 0, 75, 16, 60, 0, 212, 23, 60, 0, 213, 23, 60, 0, 168, 26, 60, 0, 169, 26, 60, 0, 170, 26, 60, 0, 171, 26, 60, 0, 94, 27, 60, 0, 95, 27, 60, 0, 200, 169, 60, 0, 201, 169, 60, 0, 93, 170, 60, 0, 94, 170, 60, 0, 95, 170, 60, 0, 240, 170, 60, 0, 235, 171, 60, 0, 14, 32, 2, 216, 86, 222, 14, 32, 60, 0, 14, 32, 2, 216, 87, 222, 14, 32, 60, 0, 4, 216, 71, 220, 60, 0, 4, 216, 72, 220, 60, 0, 4, 216, 192, 220, 60, 0, 4, 216, 193, 220, 60, 0, 4, 216, 65, 221, 60, 0, 4, 216, 66, 221, 60, 0, 4, 216, 197, 221, 60, 0, 4, 216, 198, 221, 60, 0, 4, 216, 56, 222, 60, 0, 4, 216, 57, 222, 60, 0, 5, 216, 75, 220, 60, 0, 5, 216, 76, 220, 60, 0, 5, 216, 194, 221, 60, 0, 5, 216, 195, 221, 60, 0, 5, 216, 65, 222, 60, 0, 5, 216, 66, 222, 60, 0, 5, 216, 60, 223, 60, 0, 5, 216, 61, 223, 60, 0, 6, 216, 68, 221, 60, 0, 7, 216, 65, 220, 60, 0, 7, 216, 66, 220, 60, 0, 26, 216, 110, 222, 60, 0, 26, 216, 111, 222, 60, 0, 126, 28, 60, 0, 127, 28, 60, 0, 14, 32, 0, 7, 14, 32, 60, 0, 247, 7, 60, 0, 251, 16, 60, 0, 96, 19, 60, 0, 104, 19, 60, 0, 30, 26, 60, 0, 31, 26, 60, 0, 90, 27, 60, 0, 91, 27, 60, 0, 193, 169, 60, 0, 194, 169, 60, 0, 195, 169, 60, 0, 196, 169, 60, 0, 197, 169, 60, 0, 198, 169, 60, 0, 202, 169, 60, 0, 203, 169, 60, 0, 204, 169, 60, 0, 205, 169, 60, 0, 242, 166, 60, 0, 95, 169, 60, 0, 14, 32, 2, 216, 87, 220, 14, 32, 60, 0, 2, 216, 58, 223, 60, 0, 2, 216, 59, 223, 60, 0, 2, 216, 60, 223, 60, 0, 2, 216, 61, 223, 60, 0, 2, 216, 62, 223, 60, 0, 2, 216, 63, 223, 60, 0, 14, 32, 3, 216, 85, 223, 14, 32, 60, 0, 14, 32, 3, 216, 86, 223, 14, 32, 60, 0, 14, 32, 3, 216, 87, 223, 14, 32, 60, 0, 14, 32, 3, 216, 88, 223, 14, 32, 60, 0, 14, 32, 3, 216, 89, 223, 14, 32, 60, 0, 4, 216, 190, 220, 60, 0, 4, 216, 191, 220, 60, 0, 4, 216, 64, 221, 60, 0, 4, 216, 200, 221, 60, 0, 4, 216, 222, 221, 60, 0, 4, 216, 223, 221, 60, 0, 4, 216, 169, 222, 60, 0, 5, 216, 62, 223, 60, 0, 6, 216, 70, 221, 60, 0, 7, 216, 247, 222, 60, 0, 7, 216, 248, 222, 60, 0, 85, 32, 60, 0, 86, 32, 60, 0, 88, 32, 60, 0, 89, 32, 60, 0, 90, 32, 60, 0, 91, 32, 60, 0, 92, 32, 60, 0, 93, 32, 60, 0, 94, 32, 60, 0, 42, 46, 60, 0, 43, 46, 60, 0, 44, 46, 60, 0, 45, 46, 60, 0, 61, 46, 60, 0, 252, 44, 60, 0, 255, 44, 60, 0, 25, 46, 60, 0, 14, 32, 2, 216, 63, 221, 14, 32, 60, 0, 0, 216, 0, 221, 60, 0, 0, 216, 1, 221, 60, 0, 0, 216, 2, 221, 60, 0, 0, 216, 159, 223, 60, 0, 0, 216, 208, 223, 60, 0, 2, 216, 31, 221, 60, 0, 9, 216, 112, 220, 60, 0, 9, 216, 113, 220, 60, 0, 9, 216, 114, 220, 60, 0, 9, 216, 115, 220, 60, 0, 9, 216, 116, 220, 60, 0, 39, 0, 39, 0, 60, 0, 60, 0, 60, 0, 7, 255, 60, 0, 24, 32, 60, 0, 25, 32, 60, 0, 26, 32, 60, 0, 27, 32, 60, 0, 57, 32, 60, 0, 58, 32, 60, 0, 39, 0, 34, 0, 39, 0, 60, 0, 60, 0, 60, 0, 2, 255, 60, 0, 28, 32, 60, 0, 29, 32, 60, 0, 30, 32, 60, 0, 31, 32, 60, 0, 66, 46, 60, 0, 29, 48, 60, 0, 30, 48, 60, 0, 31, 48, 60, 0, 171, 0, 60, 0, 187, 0, 60, 0, 39, 0, 40, 0, 39, 0, 60, 0, 60, 0, 60, 0, 8, 255, 60, 0, 60, 0, 60, 0, 116, 36, 47, 0, 14, 32, 59, 216, 1, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 125, 36, 47, 0, 14, 32, 59, 216, 1, 221, 58, 216, 80, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 126, 36, 47, 0, 14, 32, 59, 216, 1, 221, 59, 216, 1, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 127, 36, 47, 0, 14, 32, 59, 216, 1, 221, 59, 216, 2, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 128, 36, 47, 0, 14, 32, 59, 216, 1, 221, 59, 216, 3, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 129, 36, 47, 0, 14, 32, 59, 216, 1, 221, 59, 216, 4, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 130, 36, 47, 0, 14, 32, 59, 216, 1, 221, 59, 216, 5, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 131, 36, 47, 0, 14, 32, 59, 216, 1, 221, 59, 216, 6, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 132, 36, 47, 0, 14, 32, 59, 216, 1, 221, 59, 216, 7, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 133, 36, 47, 0, 14, 32, 59, 216, 1, 221, 59, 216, 8, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 134, 36, 47, 0, 14, 32, 59, 216, 1, 221, 59, 216, 9, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 117, 36, 47, 0, 14, 32, 59, 216, 2, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 135, 36, 47, 0, 14, 32, 59, 216, 2, 221, 58, 216, 80, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 118, 36, 47, 0, 14, 32, 59, 216, 3, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 119, 36, 47, 0, 14, 32, 59, 216, 4, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 120, 36, 47, 0, 14, 32, 59, 216, 5, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 121, 36, 47, 0, 14, 32, 59, 216, 6, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 122, 36, 47, 0, 14, 32, 59, 216, 7, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 123, 36, 47, 0, 14, 32, 59, 216, 8, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 124, 36, 47, 0, 14, 32, 59, 216, 9, 221, 39, 0, 41, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 156, 36, 47, 0, 97, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 16, 221, 47, 0, 65, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 157, 36, 47, 0, 98, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 17, 221, 47, 0, 66, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 158, 36, 47, 0, 99, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 18, 221, 47, 0, 67, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 159, 36, 47, 0, 100, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 19, 221, 47, 0, 68, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 160, 36, 47, 0, 101, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 20, 221, 47, 0, 69, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 161, 36, 47, 0, 102, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 21, 221, 47, 0, 70, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 162, 36, 47, 0, 103, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 22, 221, 47, 0, 71, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 163, 36, 47, 0, 104, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 23, 221, 47, 0, 72, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 164, 36, 47, 0, 105, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 24, 221, 47, 0, 73, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 165, 36, 47, 0, 106, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 25, 221, 47, 0, 74, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 166, 36, 47, 0, 107, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 26, 221, 47, 0, 75, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 167, 36, 47, 0, 108, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 27, 221, 47, 0, 76, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 168, 36, 47, 0, 109, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 28, 221, 47, 0, 77, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 169, 36, 47, 0, 110, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 29, 221, 47, 0, 78, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 170, 36, 47, 0, 111, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 30, 221, 47, 0, 79, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 171, 36, 47, 0, 112, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 31, 221, 47, 0, 80, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 172, 36, 47, 0, 113, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 32, 221, 47, 0, 81, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 173, 36, 47, 0, 114, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 33, 221, 47, 0, 82, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 174, 36, 47, 0, 115, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 34, 221, 47, 0, 83, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 175, 36, 47, 0, 116, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 35, 221, 47, 0, 84, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 176, 36, 47, 0, 117, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 36, 221, 47, 0, 85, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 177, 36, 47, 0, 118, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 37, 221, 47, 0, 86, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 178, 36, 47, 0, 119, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 38, 221, 47, 0, 87, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 179, 36, 47, 0, 120, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 39, 221, 47, 0, 88, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 180, 36, 47, 0, 121, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 40, 221, 47, 0, 89, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 181, 36, 47, 0, 122, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 41, 221, 47, 0, 90, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 0, 50, 47, 0, 0, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 14, 50, 47, 0, 0, 172, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 1, 50, 47, 0, 2, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 15, 50, 47, 0, 152, 176, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 2, 50, 47, 0, 3, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 16, 50, 47, 0, 228, 178, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 3, 50, 47, 0, 5, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 17, 50, 47, 0, 124, 183, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 4, 50, 47, 0, 6, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 18, 50, 47, 0, 200, 185, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 5, 50, 47, 0, 7, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 19, 50, 47, 0, 20, 188, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 6, 50, 47, 0, 9, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 20, 50, 47, 0, 172, 192, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 7, 50, 47, 0, 11, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 21, 50, 47, 0, 68, 197, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 29, 50, 47, 0, 36, 198, 4, 200, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 30, 50, 47, 0, 36, 198, 196, 214, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 8, 50, 47, 0, 12, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 22, 50, 47, 0, 144, 199, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 28, 50, 47, 0, 252, 200, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 9, 50, 47, 0, 14, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 23, 50, 47, 0, 40, 204, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 10, 50, 47, 0, 15, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 24, 50, 47, 0, 116, 206, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 11, 50, 47, 0, 16, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 25, 50, 47, 0, 192, 208, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 12, 50, 47, 0, 17, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 26, 50, 47, 0, 12, 211, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 13, 50, 47, 0, 18, 17, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 27, 50, 47, 0, 88, 213, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 32, 50, 47, 0, 0, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 38, 50, 47, 0, 3, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 34, 50, 47, 0, 9, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 40, 50, 47, 0, 93, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 33, 50, 47, 0, 140, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 36, 50, 47, 0, 148, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 57, 50, 47, 0, 227, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 61, 50, 47, 0, 1, 79, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 65, 50, 47, 0, 17, 79, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 39, 50, 47, 0, 107, 81, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 37, 50, 47, 0, 109, 81, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 56, 50, 47, 0, 180, 82, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 41, 50, 47, 0, 65, 83, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 63, 50, 47, 0, 84, 83, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 52, 50, 47, 0, 13, 84, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 58, 50, 47, 0, 124, 84, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 35, 50, 47, 0, 219, 86, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 47, 50, 47, 0, 31, 87, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 59, 50, 47, 0, 102, 91, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 48, 50, 47, 0, 229, 101, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 42, 50, 47, 0, 8, 103, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 50, 50, 47, 0, 9, 103, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 45, 50, 47, 0, 40, 103, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 49, 50, 47, 0, 42, 104, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 44, 50, 47, 0, 52, 108, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 43, 50, 47, 0, 107, 112, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 53, 50, 47, 0, 121, 114, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 50, 47, 0, 227, 118, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 51, 50, 47, 0, 62, 121, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 55, 50, 47, 0, 93, 121, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 64, 50, 47, 0, 109, 121, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 66, 50, 47, 0, 234, 129, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 67, 50, 47, 0, 243, 129, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 54, 50, 47, 0, 161, 140, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 62, 50, 47, 0, 199, 140, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 46, 50, 47, 0, 209, 145, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 89, 254, 60, 0, 60, 0, 60, 0, 125, 32, 60, 0, 60, 0, 60, 0, 141, 32, 60, 0, 60, 0, 60, 0, 53, 254, 60, 0, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 9, 255, 60, 0, 60, 0, 60, 0, 90, 254, 60, 0, 60, 0, 60, 0, 126, 32, 60, 0, 60, 0, 60, 0, 142, 32, 60, 0, 60, 0, 60, 0, 54, 254, 60, 0, 39, 0, 91, 0, 39, 0, 60, 0, 60, 0, 60, 0, 59, 255, 60, 0, 60, 0, 60, 0, 71, 254, 60, 0, 39, 0, 93, 0, 39, 0, 60, 0, 60, 0, 60, 0, 61, 255, 60, 0, 60, 0, 60, 0, 72, 254, 60, 0, 39, 0, 123, 0, 39, 0, 60, 0, 60, 0, 60, 0, 91, 255, 60, 0, 60, 0, 60, 0, 91, 254, 60, 0, 60, 0, 60, 0, 55, 254, 60, 0, 39, 0, 125, 0, 39, 0, 60, 0, 60, 0, 60, 0, 93, 255, 60, 0, 60, 0, 60, 0, 92, 254, 60, 0, 60, 0, 60, 0, 56, 254, 60, 0, 58, 15, 60, 0, 59, 15, 60, 0, 60, 15, 60, 0, 61, 15, 60, 0, 155, 22, 60, 0, 156, 22, 60, 0, 69, 32, 60, 0, 70, 32, 60, 0, 8, 35, 60, 0, 9, 35, 60, 0, 10, 35, 60, 0, 11, 35, 60, 0, 252, 41, 60, 0, 253, 41, 60, 0, 131, 41, 60, 0, 132, 41, 60, 0, 133, 41, 60, 0, 60, 0, 60, 0, 95, 255, 60, 0, 134, 41, 60, 0, 60, 0, 60, 0, 96, 255, 60, 0, 135, 41, 60, 0, 136, 41, 60, 0, 137, 41, 60, 0, 138, 41, 60, 0, 139, 41, 60, 0, 140, 41, 60, 0, 141, 41, 60, 0, 142, 41, 60, 0, 143, 41, 60, 0, 144, 41, 60, 0, 145, 41, 60, 0, 146, 41, 60, 0, 147, 41, 60, 0, 148, 41, 60, 0, 149, 41, 60, 0, 150, 41, 60, 0, 151, 41, 60, 0, 152, 41, 60, 0, 197, 39, 60, 0, 198, 39, 60, 0, 230, 39, 60, 0, 231, 39, 60, 0, 232, 39, 60, 0, 233, 39, 60, 0, 234, 39, 60, 0, 235, 39, 60, 0, 236, 39, 60, 0, 237, 39, 60, 0, 238, 39, 60, 0, 239, 39, 60, 0, 104, 39, 60, 0, 105, 39, 60, 0, 106, 39, 60, 0, 107, 39, 60, 0, 108, 39, 60, 0, 109, 39, 60, 0, 110, 39, 60, 0, 111, 39, 60, 0, 112, 39, 60, 0, 113, 39, 60, 0, 114, 39, 60, 0, 115, 39, 60, 0, 116, 39, 60, 0, 117, 39, 60, 0, 2, 46, 60, 0, 3, 46, 60, 0, 4, 46, 60, 0, 5, 46, 60, 0, 9, 46, 60, 0, 10, 46, 60, 0, 12, 46, 60, 0, 13, 46, 60, 0, 28, 46, 60, 0, 29, 46, 60, 0, 32, 46, 60, 0, 33, 46, 60, 0, 34, 46, 60, 0, 35, 46, 60, 0, 36, 46, 60, 0, 37, 46, 60, 0, 38, 46, 60, 0, 39, 46, 60, 0, 40, 46, 60, 0, 41, 46, 60, 0, 8, 48, 60, 0, 60, 0, 60, 0, 63, 254, 60, 0, 9, 48, 60, 0, 60, 0, 60, 0, 64, 254, 60, 0, 10, 48, 60, 0, 60, 0, 60, 0, 61, 254, 60, 0, 11, 48, 60, 0, 60, 0, 60, 0, 62, 254, 60, 0, 12, 48, 60, 0, 60, 0, 60, 0, 98, 255, 60, 0, 60, 0, 60, 0, 65, 254, 60, 0, 13, 48, 60, 0, 60, 0, 60, 0, 99, 255, 60, 0, 60, 0, 60, 0, 66, 254, 60, 0, 14, 48, 60, 0, 60, 0, 60, 0, 67, 254, 60, 0, 15, 48, 60, 0, 60, 0, 60, 0, 68, 254, 60, 0, 16, 48, 60, 0, 60, 0, 60, 0, 59, 254, 60, 0, 17, 48, 60, 0, 60, 0, 60, 0, 60, 254, 60, 0, 20, 48, 60, 0, 60, 0, 60, 0, 60, 216, 42, 221, 47, 0, 83, 0, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 65, 222, 47, 0, 9, 78, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 66, 222, 47, 0, 140, 78, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 71, 222, 47, 0, 221, 82, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 67, 222, 47, 0, 137, 91, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 69, 222, 47, 0, 83, 98, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 72, 222, 47, 0, 87, 101, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 64, 222, 47, 0, 44, 103, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 68, 222, 47, 0, 185, 112, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 70, 222, 47, 0, 215, 118, 21, 48, 60, 0, 60, 0, 60, 0, 93, 254, 60, 0, 60, 0, 60, 0, 57, 254, 60, 0, 21, 48, 60, 0, 60, 0, 60, 0, 94, 254, 60, 0, 60, 0, 60, 0, 58, 254, 60, 0, 22, 48, 60, 0, 60, 0, 60, 0, 23, 254, 60, 0, 23, 48, 60, 0, 60, 0, 60, 0, 24, 254, 60, 0, 24, 48, 60, 0, 25, 48, 60, 0, 26, 48, 60, 0, 27, 48, 60, 0, 62, 253, 60, 0, 63, 253, 60, 0, 22, 32, 60, 0, 62, 46, 60, 0, 216, 41, 60, 0, 217, 41, 60, 0, 218, 41, 60, 0, 219, 41, 60, 0, 167, 0, 60, 0, 57, 46, 60, 0, 182, 0, 60, 0, 75, 32, 60, 0, 77, 46, 60, 0, 63, 46, 60, 0, 39, 0, 64, 0, 39, 0, 60, 0, 60, 0, 60, 0, 32, 255, 60, 0, 60, 0, 60, 0, 107, 254, 60, 0, 39, 0, 42, 0, 39, 0, 60, 0, 60, 0, 60, 0, 10, 255, 60, 0, 60, 0, 60, 0, 97, 254, 60, 0, 78, 32, 60, 0, 81, 32, 60, 0, 14, 32, 109, 6, 14, 32, 60, 0, 115, 166, 60, 0, 39, 0, 47, 0, 39, 0, 60, 0, 60, 0, 60, 0, 15, 255, 60, 0, 39, 0, 92, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 255, 60, 0, 60, 0, 60, 0, 104, 254, 60, 0, 74, 46, 60, 0, 39, 0, 38, 0, 39, 0, 60, 0, 60, 0, 60, 0, 6, 255, 60, 0, 60, 0, 60, 0, 96, 254, 60, 0, 74, 32, 60, 0, 82, 46, 60, 0, 39, 0, 35, 0, 39, 0, 60, 0, 60, 0, 60, 0, 3, 255, 60, 0, 60, 0, 60, 0, 95, 254, 60, 0, 39, 0, 37, 0, 39, 0, 60, 0, 60, 0, 60, 0, 5, 255, 60, 0, 60, 0, 60, 0, 106, 254, 60, 0, 106, 6, 60, 0, 48, 32, 60, 0, 9, 6, 60, 0, 49, 32, 60, 0, 10, 6, 60, 0, 32, 32, 60, 0, 33, 32, 60, 0, 54, 46, 60, 0, 55, 46, 60, 0, 56, 46, 60, 0, 75, 46, 60, 0, 34, 32, 60, 0, 35, 32, 60, 0, 39, 32, 60, 0, 67, 32, 60, 0, 76, 32, 60, 0, 77, 32, 60, 0, 50, 32, 60, 0, 60, 0, 60, 0, 51, 32, 47, 0, 50, 32, 60, 0, 60, 0, 60, 0, 52, 32, 47, 0, 50, 32, 50, 32, 60, 0, 60, 0, 60, 0, 87, 32, 47, 0, 50, 32, 50, 32, 50, 32, 60, 0, 53, 32, 60, 0, 60, 0, 60, 0, 54, 32, 47, 0, 53, 32, 60, 0, 60, 0, 60, 0, 55, 32, 47, 0, 53, 32, 53, 32, 60, 0, 3, 48, 60, 0, 61, 48, 60, 0, 56, 32, 60, 0, 59, 32, 60, 0, 63, 32, 60, 0, 84, 32, 60, 0, 64, 32, 60, 0, 80, 32, 60, 0, 65, 32, 60, 0, 66, 32, 60, 0, 0, 46, 60, 0, 1, 46, 60, 0, 6, 46, 60, 0, 7, 46, 60, 0, 8, 46, 60, 0, 11, 46, 60, 0, 14, 46, 60, 0, 15, 46, 60, 0, 16, 46, 60, 0, 17, 46, 60, 0, 18, 46, 60, 0, 19, 46, 60, 0, 20, 46, 60, 0, 21, 46, 60, 0, 22, 46, 60, 0, 26, 46, 60, 0, 27, 46, 60, 0, 30, 46, 60, 0, 31, 46, 60, 0, 68, 46, 60, 0, 69, 46, 60, 0, 70, 46, 60, 0, 71, 46, 60, 0, 72, 46, 60, 0, 126, 166, 60, 0, 90, 5, 60, 0, 91, 5, 60, 0, 95, 5, 60, 0, 14, 32, 190, 5, 14, 32, 60, 0, 14, 32, 192, 5, 14, 32, 60, 0, 14, 32, 195, 5, 14, 32, 60, 0, 14, 32, 198, 5, 14, 32, 60, 0, 14, 32, 243, 5, 14, 32, 60, 0, 14, 32, 244, 5, 14, 32, 60, 0, 14, 32, 10, 7, 14, 32, 60, 0, 14, 32, 11, 7, 14, 32, 60, 0, 14, 32, 12, 7, 14, 32, 60, 0, 14, 32, 13, 7, 14, 32, 60, 0, 14, 32, 94, 8, 14, 32, 60, 0, 0, 24, 60, 0, 5, 216, 96, 222, 60, 0, 5, 216, 97, 222, 60, 0, 5, 216, 98, 222, 60, 0, 5, 216, 99, 222, 60, 0, 5, 216, 100, 222, 60, 0, 5, 216, 101, 222, 60, 0, 5, 216, 102, 222, 60, 0, 5, 216, 103, 222, 60, 0, 5, 216, 104, 222, 60, 0, 5, 216, 105, 222, 60, 0, 5, 216, 106, 222, 60, 0, 5, 216, 107, 222, 60, 0, 5, 216, 108, 222, 60, 0, 112, 9, 60, 0, 248, 168, 60, 0, 249, 168, 60, 0, 250, 168, 60, 0, 252, 168, 60, 0, 253, 9, 60, 0, 118, 10, 60, 0, 240, 10, 60, 0, 119, 12, 60, 0, 132, 12, 60, 0, 244, 13, 60, 0, 79, 14, 60, 0, 90, 14, 60, 0, 91, 14, 60, 0, 222, 170, 60, 0, 223, 170, 60, 0, 4, 15, 60, 0, 5, 15, 60, 0, 6, 15, 60, 0, 7, 15, 60, 0, 8, 15, 60, 0, 9, 15, 60, 0, 10, 15, 60, 0, 208, 15, 60, 0, 209, 15, 60, 0, 11, 15, 60, 0, 60, 0, 60, 0, 12, 15, 60, 0, 13, 15, 60, 0, 14, 15, 60, 0, 15, 15, 60, 0, 16, 15, 60, 0, 17, 15, 60, 0, 18, 15, 60, 0, 133, 15, 60, 0, 210, 15, 60, 0, 211, 15, 60, 0, 212, 15, 60, 0, 217, 15, 60, 0, 218, 15, 60, 0, 6, 216, 63, 222, 60, 0, 6, 216, 64, 222, 60, 0, 6, 216, 65, 222, 60, 0, 6, 216, 66, 222, 60, 0, 6, 216, 67, 222, 60, 0, 6, 216, 68, 222, 60, 0, 6, 216, 69, 222, 60, 0, 6, 216, 70, 222, 60, 0, 6, 216, 154, 222, 60, 0, 6, 216, 155, 222, 60, 0, 6, 216, 156, 222, 60, 0, 6, 216, 158, 222, 60, 0, 6, 216, 159, 222, 60, 0, 6, 216, 160, 222, 60, 0, 6, 216, 161, 222, 60, 0, 6, 216, 162, 222, 60, 0, 7, 216, 112, 220, 60, 0, 7, 216, 113, 220, 60, 0, 61, 28, 60, 0, 62, 28, 60, 0, 63, 28, 60, 0, 76, 16, 60, 0, 77, 16, 60, 0, 78, 16, 60, 0, 79, 16, 60, 0, 216, 23, 60, 0, 217, 23, 60, 0, 218, 23, 60, 0, 160, 26, 60, 0, 161, 26, 60, 0, 162, 26, 60, 0, 163, 26, 60, 0, 164, 26, 60, 0, 165, 26, 60, 0, 166, 26, 60, 0, 172, 26, 60, 0, 173, 26, 60, 0, 192, 28, 60, 0, 193, 28, 60, 0, 194, 28, 60, 0, 195, 28, 60, 0, 196, 28, 60, 0, 197, 28, 60, 0, 198, 28, 60, 0, 199, 28, 60, 0, 112, 45, 60, 0, 116, 168, 60, 0, 117, 168, 60, 0, 252, 27, 60, 0, 253, 27, 60, 0, 254, 27, 60, 0, 255, 27, 60, 0, 46, 169, 60, 0, 222, 169, 60, 0, 223, 169, 60, 0, 92, 170, 60, 0, 1, 216, 111, 221, 60, 0, 4, 216, 73, 220, 60, 0, 4, 216, 74, 220, 60, 0, 4, 216, 75, 220, 60, 0, 4, 216, 76, 220, 60, 0, 4, 216, 77, 220, 60, 0, 14, 32, 2, 216, 80, 222, 14, 32, 60, 0, 14, 32, 2, 216, 81, 222, 14, 32, 60, 0, 14, 32, 2, 216, 82, 222, 14, 32, 60, 0, 14, 32, 2, 216, 83, 222, 14, 32, 60, 0, 14, 32, 2, 216, 84, 222, 14, 32, 60, 0, 14, 32, 2, 216, 85, 222, 14, 32, 60, 0, 14, 32, 2, 216, 88, 222, 14, 32, 60, 0, 7, 216, 67, 220, 60, 0, 7, 216, 68, 220, 60, 0, 7, 216, 69, 220, 60, 0, 2, 216, 57, 223, 60, 0, 14, 32, 2, 216, 240, 222, 14, 32, 60, 0, 14, 32, 2, 216, 241, 222, 14, 32, 60, 0, 14, 32, 2, 216, 242, 222, 14, 32, 60, 0, 14, 32, 2, 216, 243, 222, 14, 32, 60, 0, 14, 32, 2, 216, 244, 222, 14, 32, 60, 0, 14, 32, 2, 216, 245, 222, 14, 32, 60, 0, 14, 32, 2, 216, 246, 222, 14, 32, 60, 0, 14, 32, 2, 216, 153, 223, 14, 32, 60, 0, 14, 32, 2, 216, 154, 223, 14, 32, 60, 0, 14, 32, 2, 216, 155, 223, 14, 32, 60, 0, 14, 32, 2, 216, 156, 223, 14, 32, 60, 0, 4, 216, 187, 220, 60, 0, 4, 216, 188, 220, 60, 0, 4, 216, 116, 221, 60, 0, 4, 216, 117, 221, 60, 0, 4, 216, 205, 221, 60, 0, 4, 216, 199, 221, 60, 0, 4, 216, 219, 221, 60, 0, 4, 216, 221, 221, 60, 0, 4, 216, 58, 222, 60, 0, 4, 216, 59, 222, 60, 0, 4, 216, 60, 222, 60, 0, 4, 216, 61, 222, 60, 0, 5, 216, 77, 220, 60, 0, 5, 216, 90, 220, 60, 0, 5, 216, 78, 220, 60, 0, 5, 216, 79, 220, 60, 0, 5, 216, 91, 220, 60, 0, 5, 216, 93, 220, 60, 0, 5, 216, 198, 220, 60, 0, 5, 216, 193, 221, 60, 0, 5, 216, 196, 221, 60, 0, 5, 216, 197, 221, 60, 0, 5, 216, 198, 221, 60, 0, 5, 216, 199, 221, 60, 0, 5, 216, 200, 221, 60, 0, 5, 216, 201, 221, 60, 0, 5, 216, 202, 221, 60, 0, 5, 216, 203, 221, 60, 0, 5, 216, 204, 221, 60, 0, 5, 216, 205, 221, 60, 0, 5, 216, 206, 221, 60, 0, 5, 216, 207, 221, 60, 0, 5, 216, 208, 221, 60, 0, 5, 216, 209, 221, 60, 0, 5, 216, 210, 221, 60, 0, 5, 216, 211, 221, 60, 0, 5, 216, 212, 221, 60, 0, 5, 216, 213, 221, 60, 0, 5, 216, 214, 221, 60, 0, 5, 216, 215, 221, 60, 0, 5, 216, 67, 222, 60, 0, 6, 216, 59, 220, 60, 0, 6, 216, 69, 221, 60, 0, 6, 216, 226, 221, 60, 0, 7, 216, 255, 223, 60, 0, 26, 216, 55, 223, 60, 0, 26, 216, 56, 223, 60, 0, 26, 216, 57, 223, 60, 0, 26, 216, 58, 223, 60, 0, 26, 216, 59, 223, 60, 0, 26, 216, 68, 223, 60, 0, 27, 216, 153, 222, 60, 0, 27, 216, 154, 222, 60, 0, 54, 216, 135, 222, 60, 0, 54, 216, 136, 222, 60, 0, 54, 216, 137, 222, 60, 0, 54, 216, 138, 222, 60, 0, 54, 216, 139, 222, 60, 0, 14, 32, 2, 216, 127, 222, 14, 32, 38, 0, 253, 48, 60, 0, 39, 0, 96, 0, 39, 0, 60, 0, 60, 0, 60, 0, 64, 255, 60, 0, 180, 0, 61, 0, 132, 3, 60, 0, 220, 2, 60, 0, 39, 0, 94, 0, 39, 0, 60, 0, 60, 0, 60, 0, 62, 255, 60, 0, 175, 0, 60, 0, 60, 0, 60, 0, 227, 255, 60, 0, 216, 2, 60, 0, 217, 2, 60, 0, 168, 0, 60, 0, 218, 2, 60, 0, 221, 2, 60, 0, 184, 0, 60, 0, 219, 2, 60, 0, 189, 31, 61, 0, 191, 31, 60, 0, 254, 31, 60, 0, 192, 31, 60, 0, 155, 48, 60, 0, 156, 48, 60, 0, 185, 2, 60, 0, 117, 3, 60, 0, 186, 2, 60, 0, 194, 2, 60, 0, 195, 2, 60, 0, 196, 2, 60, 0, 197, 2, 60, 0, 198, 2, 60, 0, 199, 2, 60, 0, 200, 2, 60, 0, 201, 2, 60, 0, 202, 2, 60, 0, 203, 2, 60, 0, 204, 2, 60, 0, 205, 2, 60, 0, 206, 2, 60, 0, 207, 2, 60, 0, 210, 2, 60, 0, 211, 2, 60, 0, 212, 2, 60, 0, 213, 2, 60, 0, 106, 171, 60, 0, 107, 171, 60, 0, 214, 2, 60, 0, 215, 2, 60, 0, 222, 2, 60, 0, 223, 2, 60, 0, 229, 2, 60, 0, 230, 2, 60, 0, 231, 2, 60, 0, 232, 2, 60, 0, 233, 2, 60, 0, 234, 2, 60, 0, 235, 2, 60, 0, 236, 2, 60, 0, 237, 2, 60, 0, 239, 2, 60, 0, 240, 2, 60, 0, 241, 2, 60, 0, 242, 2, 60, 0, 243, 2, 60, 0, 244, 2, 60, 0, 245, 2, 60, 0, 246, 2, 60, 0, 247, 2, 60, 0, 248, 2, 60, 0, 249, 2, 60, 0, 250, 2, 60, 0, 251, 2, 60, 0, 252, 2, 60, 0, 253, 2, 60, 0, 254, 2, 60, 0, 255, 2, 60, 0, 144, 19, 60, 0, 145, 19, 60, 0, 146, 19, 60, 0, 147, 19, 60, 0, 148, 19, 60, 0, 149, 19, 60, 0, 150, 19, 60, 0, 151, 19, 60, 0, 152, 19, 60, 0, 153, 19, 60, 0, 0, 167, 60, 0, 1, 167, 60, 0, 2, 167, 60, 0, 3, 167, 60, 0, 4, 167, 60, 0, 5, 167, 60, 0, 6, 167, 60, 0, 7, 167, 60, 0, 8, 167, 60, 0, 9, 167, 60, 0, 10, 167, 60, 0, 11, 167, 60, 0, 12, 167, 60, 0, 13, 167, 60, 0, 14, 167, 60, 0, 15, 167, 60, 0, 16, 167, 60, 0, 17, 167, 60, 0, 18, 167, 60, 0, 19, 167, 60, 0, 20, 167, 60, 0, 21, 167, 60, 0, 22, 167, 60, 0, 23, 167, 60, 0, 24, 167, 60, 0, 25, 167, 60, 0, 26, 167, 60, 0, 27, 167, 60, 0, 28, 167, 60, 0, 29, 167, 60, 0, 30, 167, 60, 0, 31, 167, 60, 0, 32, 167, 60, 0, 33, 167, 60, 0, 136, 167, 60, 0, 137, 167, 60, 0, 138, 167, 60, 0, 91, 171, 60, 0, 176, 0, 60, 0, 60, 0, 60, 0, 3, 33, 47, 0, 67, 0, 60, 0, 60, 0, 60, 0, 9, 33, 47, 0, 70, 0, 60, 0, 130, 4, 60, 0, 141, 5, 60, 0, 142, 5, 60, 0, 14, 32, 8, 6, 14, 32, 60, 0, 59, 216, 240, 222, 60, 0, 59, 216, 241, 222, 60, 0, 14, 6, 60, 0, 15, 6, 60, 0, 222, 6, 60, 0, 233, 6, 60, 0, 253, 253, 60, 0, 14, 32, 178, 251, 14, 32, 60, 0, 14, 32, 179, 251, 14, 32, 60, 0, 14, 32, 180, 251, 14, 32, 60, 0, 14, 32, 181, 251, 14, 32, 60, 0, 14, 32, 182, 251, 14, 32, 60, 0, 14, 32, 183, 251, 14, 32, 60, 0, 14, 32, 184, 251, 14, 32, 60, 0, 14, 32, 185, 251, 14, 32, 60, 0, 14, 32, 186, 251, 14, 32, 60, 0, 14, 32, 187, 251, 14, 32, 60, 0, 14, 32, 188, 251, 14, 32, 60, 0, 14, 32, 189, 251, 14, 32, 60, 0, 14, 32, 190, 251, 14, 32, 60, 0, 14, 32, 191, 251, 14, 32, 60, 0, 14, 32, 192, 251, 14, 32, 60, 0, 14, 32, 193, 251, 14, 32, 60, 0, 246, 7, 60, 0, 250, 9, 60, 0, 112, 11, 60, 0, 243, 11, 60, 0, 244, 11, 60, 0, 245, 11, 60, 0, 246, 11, 60, 0, 247, 11, 60, 0, 248, 11, 60, 0, 250, 11, 60, 0, 7, 216, 213, 223, 60, 0, 7, 216, 214, 223, 60, 0, 7, 216, 215, 223, 60, 0, 7, 216, 216, 223, 60, 0, 7, 216, 217, 223, 60, 0, 7, 216, 218, 223, 60, 0, 7, 216, 219, 223, 60, 0, 7, 216, 220, 223, 60, 0, 7, 216, 225, 223, 60, 0, 7, 216, 226, 223, 60, 0, 7, 216, 227, 223, 60, 0, 7, 216, 228, 223, 60, 0, 7, 216, 229, 223, 60, 0, 7, 216, 230, 223, 60, 0, 7, 216, 231, 223, 60, 0, 7, 216, 232, 223, 60, 0, 7, 216, 233, 223, 60, 0, 7, 216, 234, 223, 60, 0, 7, 216, 235, 223, 60, 0, 7, 216, 236, 223, 60, 0, 7, 216, 237, 223, 60, 0, 7, 216, 238, 223, 60, 0, 7, 216, 239, 223, 60, 0, 7, 216, 240, 223, 60, 0, 7, 216, 241, 223, 60, 0, 127, 12, 60, 0, 79, 13, 60, 0, 121, 13, 60, 0, 40, 168, 60, 0, 41, 168, 60, 0, 42, 168, 60, 0, 43, 168, 60, 0, 54, 168, 60, 0, 55, 168, 60, 0, 57, 168, 60, 0, 1, 15, 60, 0, 2, 15, 60, 0, 3, 15, 60, 0, 19, 15, 60, 0, 21, 15, 60, 0, 22, 15, 60, 0, 23, 15, 60, 0, 26, 15, 60, 0, 27, 15, 60, 0, 28, 15, 60, 0, 29, 15, 60, 0, 30, 15, 60, 0, 31, 15, 60, 0, 52, 15, 60, 0, 54, 15, 60, 0, 56, 15, 60, 0, 190, 15, 60, 0, 191, 15, 60, 0, 192, 15, 60, 0, 193, 15, 60, 0, 194, 15, 60, 0, 195, 15, 60, 0, 196, 15, 60, 0, 197, 15, 60, 0, 199, 15, 60, 0, 200, 15, 60, 0, 201, 15, 60, 0, 202, 15, 60, 0, 203, 15, 60, 0, 204, 15, 60, 0, 206, 15, 60, 0, 207, 15, 60, 0, 213, 15, 60, 0, 214, 15, 60, 0, 215, 15, 60, 0, 216, 15, 60, 0, 109, 22, 60, 0, 64, 25, 60, 0, 158, 16, 60, 0, 159, 16, 60, 0, 119, 170, 60, 0, 120, 170, 60, 0, 121, 170, 60, 0, 5, 216, 63, 223, 60, 0, 224, 25, 60, 0, 225, 25, 60, 0, 226, 25, 60, 0, 227, 25, 60, 0, 228, 25, 60, 0, 229, 25, 60, 0, 230, 25, 60, 0, 231, 25, 60, 0, 232, 25, 60, 0, 233, 25, 60, 0, 234, 25, 60, 0, 235, 25, 60, 0, 236, 25, 60, 0, 237, 25, 60, 0, 238, 25, 60, 0, 239, 25, 60, 0, 240, 25, 60, 0, 241, 25, 60, 0, 242, 25, 60, 0, 243, 25, 60, 0, 244, 25, 60, 0, 245, 25, 60, 0, 246, 25, 60, 0, 247, 25, 60, 0, 248, 25, 60, 0, 249, 25, 60, 0, 250, 25, 60, 0, 251, 25, 60, 0, 252, 25, 60, 0, 253, 25, 60, 0, 254, 25, 60, 0, 255, 25, 60, 0, 97, 27, 60, 0, 98, 27, 60, 0, 99, 27, 60, 0, 100, 27, 60, 0, 101, 27, 60, 0, 102, 27, 60, 0, 103, 27, 60, 0, 104, 27, 60, 0, 105, 27, 60, 0, 106, 27, 60, 0, 116, 27, 60, 0, 117, 27, 60, 0, 118, 27, 60, 0, 119, 27, 60, 0, 120, 27, 60, 0, 121, 27, 60, 0, 122, 27, 60, 0, 123, 27, 60, 0, 124, 27, 60, 0, 169, 0, 60, 0, 60, 216, 47, 221, 60, 0, 174, 0, 60, 0, 60, 216, 173, 221, 60, 0, 4, 33, 60, 0, 8, 33, 60, 0, 20, 33, 60, 0, 23, 33, 60, 0, 24, 33, 60, 0, 30, 33, 60, 0, 31, 33, 60, 0, 35, 33, 60, 0, 37, 33, 60, 0, 39, 33, 60, 0, 41, 33, 60, 0, 46, 33, 60, 0, 58, 33, 60, 0, 65, 33, 60, 0, 66, 33, 60, 0, 67, 33, 60, 0, 68, 33, 60, 0, 74, 33, 60, 0, 76, 33, 60, 0, 79, 33, 60, 0, 138, 33, 60, 0, 139, 33, 60, 0, 144, 33, 60, 0, 60, 0, 60, 0, 233, 255, 60, 0, 146, 33, 60, 0, 60, 0, 60, 0, 235, 255, 60, 0, 145, 33, 60, 0, 60, 0, 60, 0, 234, 255, 60, 0, 147, 33, 60, 0, 60, 0, 60, 0, 236, 255, 60, 0, 148, 33, 60, 0, 149, 33, 60, 0, 150, 33, 60, 0, 151, 33, 60, 0, 152, 33, 60, 0, 153, 33, 60, 0, 156, 33, 60, 0, 157, 33, 60, 0, 158, 33, 60, 0, 159, 33, 60, 0, 160, 33, 60, 0, 161, 33, 60, 0, 162, 33, 60, 0, 163, 33, 60, 0, 164, 33, 60, 0, 165, 33, 60, 0, 166, 33, 60, 0, 167, 33, 60, 0, 168, 33, 60, 0, 169, 33, 60, 0, 170, 33, 60, 0, 171, 33, 60, 0, 172, 33, 60, 0, 173, 33, 60, 0, 175, 33, 60, 0, 176, 33, 60, 0, 177, 33, 60, 0, 178, 33, 60, 0, 179, 33, 60, 0, 180, 33, 60, 0, 181, 33, 60, 0, 182, 33, 60, 0, 183, 33, 60, 0, 184, 33, 60, 0, 185, 33, 60, 0, 186, 33, 60, 0, 187, 33, 60, 0, 188, 33, 60, 0, 189, 33, 60, 0, 190, 33, 60, 0, 191, 33, 60, 0, 192, 33, 60, 0, 193, 33, 60, 0, 194, 33, 60, 0, 195, 33, 60, 0, 196, 33, 60, 0, 197, 33, 60, 0, 198, 33, 60, 0, 199, 33, 60, 0, 200, 33, 60, 0, 201, 33, 60, 0, 202, 33, 60, 0, 203, 33, 60, 0, 204, 33, 60, 0, 208, 33, 60, 0, 209, 33, 60, 0, 210, 33, 60, 0, 211, 33, 60, 0, 212, 33, 60, 0, 213, 33, 60, 0, 214, 33, 60, 0, 215, 33, 60, 0, 216, 33, 60, 0, 217, 33, 60, 0, 218, 33, 60, 0, 219, 33, 60, 0, 220, 33, 60, 0, 221, 33, 60, 0, 222, 33, 60, 0, 223, 33, 60, 0, 224, 33, 60, 0, 225, 33, 60, 0, 226, 33, 60, 0, 227, 33, 60, 0, 228, 33, 60, 0, 229, 33, 60, 0, 230, 33, 60, 0, 231, 33, 60, 0, 232, 33, 60, 0, 233, 33, 60, 0, 234, 33, 60, 0, 235, 33, 60, 0, 236, 33, 60, 0, 237, 33, 60, 0, 238, 33, 60, 0, 239, 33, 60, 0, 240, 33, 60, 0, 241, 33, 60, 0, 242, 33, 60, 0, 243, 33, 60, 0, 244, 33, 60, 0, 245, 33, 60, 0, 246, 33, 60, 0, 247, 33, 60, 0, 248, 33, 60, 0, 249, 33, 60, 0, 250, 33, 60, 0, 251, 33, 60, 0, 252, 33, 60, 0, 253, 33, 60, 0, 254, 33, 60, 0, 255, 33, 60, 0, 0, 34, 60, 0, 1, 34, 60, 0, 2, 34, 60, 0, 60, 0, 60, 0, 53, 216, 219, 222, 61, 0, 53, 216, 21, 223, 61, 0, 53, 216, 79, 223, 61, 0, 53, 216, 137, 223, 61, 0, 53, 216, 195, 223, 60, 0, 3, 34, 60, 0, 5, 34, 60, 0, 6, 34, 60, 0, 7, 34, 60, 0, 60, 0, 60, 0, 53, 216, 193, 222, 61, 0, 53, 216, 251, 222, 61, 0, 53, 216, 53, 223, 61, 0, 53, 216, 111, 223, 61, 0, 53, 216, 169, 223, 60, 0, 8, 34, 60, 0, 10, 34, 60, 0, 11, 34, 60, 0, 13, 34, 60, 0, 246, 3, 60, 0, 14, 34, 60, 0, 15, 34, 60, 0, 16, 34, 60, 0, 17, 34, 60, 0, 60, 0, 60, 0, 64, 33, 60, 0, 39, 0, 43, 0, 39, 0, 60, 0, 60, 0, 60, 0, 11, 255, 60, 0, 60, 0, 60, 0, 41, 251, 60, 0, 60, 0, 60, 0, 98, 254, 60, 0, 60, 0, 60, 0, 122, 32, 60, 0, 60, 0, 60, 0, 138, 32, 60, 0, 177, 0, 60, 0, 247, 0, 60, 0, 215, 0, 60, 0, 39, 0, 60, 0, 39, 0, 60, 0, 60, 0, 60, 0, 28, 255, 60, 0, 60, 0, 60, 0, 100, 254, 60, 0, 39, 0, 61, 0, 39, 0, 60, 0, 60, 0, 60, 0, 29, 255, 60, 0, 60, 0, 60, 0, 117, 42, 47, 0, 39, 0, 61, 0, 39, 0, 60, 0, 60, 0, 60, 0, 118, 42, 47, 0, 39, 0, 61, 0, 61, 0, 39, 0, 60, 0, 60, 0, 60, 0, 102, 254, 60, 0, 60, 0, 60, 0, 124, 32, 60, 0, 60, 0, 60, 0, 140, 32, 60, 0, 39, 0, 62, 0, 39, 0, 60, 0, 60, 0, 60, 0, 30, 255, 60, 0, 60, 0, 60, 0, 101, 254, 60, 0, 172, 0, 60, 0, 60, 0, 60, 0, 226, 255, 60, 0, 39, 0, 124, 0, 39, 0, 60, 0, 60, 0, 60, 0, 92, 255, 60, 0, 166, 0, 60, 0, 60, 0, 60, 0, 228, 255, 60, 0, 39, 0, 126, 0, 39, 0, 60, 0, 60, 0, 60, 0, 94, 255, 60, 0, 18, 34, 60, 0, 60, 0, 60, 0, 123, 32, 60, 0, 60, 0, 60, 0, 139, 32, 60, 0, 82, 32, 60, 0, 19, 34, 60, 0, 20, 34, 60, 0, 21, 34, 60, 0, 68, 32, 60, 0, 22, 34, 60, 0, 23, 34, 60, 0, 24, 34, 60, 0, 25, 34, 60, 0, 26, 34, 60, 0, 27, 34, 60, 0, 6, 6, 60, 0, 28, 34, 60, 0, 7, 6, 60, 0, 29, 34, 60, 0, 30, 34, 60, 0, 31, 34, 60, 0, 32, 34, 60, 0, 33, 34, 60, 0, 34, 34, 60, 0, 35, 34, 60, 0, 37, 34, 60, 0, 39, 34, 60, 0, 40, 34, 60, 0, 41, 34, 60, 0, 42, 34, 60, 0, 43, 34, 60, 0, 60, 0, 60, 0, 44, 34, 47, 0, 43, 34, 60, 0, 60, 0, 60, 0, 45, 34, 47, 0, 43, 34, 43, 34, 60, 0, 60, 0, 60, 0, 12, 42, 47, 0, 43, 34, 43, 34, 43, 34, 60, 0, 46, 34, 60, 0, 60, 0, 60, 0, 47, 34, 47, 0, 46, 34, 60, 0, 60, 0, 60, 0, 48, 34, 47, 0, 46, 34, 46, 34, 60, 0, 49, 34, 60, 0, 50, 34, 60, 0, 51, 34, 60, 0, 52, 34, 60, 0, 53, 34, 60, 0, 54, 34, 60, 0, 55, 34, 60, 0, 56, 34, 60, 0, 57, 34, 60, 0, 58, 34, 60, 0, 59, 34, 60, 0, 60, 34, 60, 0, 61, 34, 60, 0, 62, 34, 60, 0, 63, 34, 60, 0, 64, 34, 60, 0, 66, 34, 60, 0, 67, 34, 60, 0, 69, 34, 60, 0, 70, 34, 60, 0, 72, 34, 60, 0, 74, 34, 60, 0, 75, 34, 60, 0, 76, 34, 60, 0, 77, 34, 60, 0, 78, 34, 60, 0, 79, 34, 60, 0, 80, 34, 60, 0, 81, 34, 60, 0, 82, 34, 60, 0, 83, 34, 60, 0, 84, 34, 60, 0, 85, 34, 60, 0, 86, 34, 60, 0, 87, 34, 60, 0, 88, 34, 60, 0, 89, 34, 60, 0, 90, 34, 60, 0, 91, 34, 60, 0, 92, 34, 60, 0, 93, 34, 60, 0, 94, 34, 60, 0, 95, 34, 60, 0, 97, 34, 60, 0, 99, 34, 60, 0, 100, 34, 60, 0, 101, 34, 60, 0, 102, 34, 60, 0, 103, 34, 60, 0, 104, 34, 60, 0, 105, 34, 60, 0, 106, 34, 60, 0, 107, 34, 60, 0, 108, 34, 60, 0, 114, 34, 60, 0, 115, 34, 60, 0, 118, 34, 60, 0, 119, 34, 60, 0, 122, 34, 60, 0, 123, 34, 60, 0, 124, 34, 60, 0, 125, 34, 60, 0, 126, 34, 60, 0, 127, 34, 60, 0, 130, 34, 60, 0, 131, 34, 60, 0, 134, 34, 60, 0, 135, 34, 60, 0, 138, 34, 60, 0, 139, 34, 60, 0, 140, 34, 60, 0, 141, 34, 60, 0, 142, 34, 60, 0, 143, 34, 60, 0, 144, 34, 60, 0, 145, 34, 60, 0, 146, 34, 60, 0, 147, 34, 60, 0, 148, 34, 60, 0, 149, 34, 60, 0, 150, 34, 60, 0, 151, 34, 60, 0, 152, 34, 60, 0, 153, 34, 60, 0, 154, 34, 60, 0, 155, 34, 60, 0, 156, 34, 60, 0, 157, 34, 60, 0, 158, 34, 60, 0, 159, 34, 60, 0, 160, 34, 60, 0, 161, 34, 60, 0, 162, 34, 60, 0, 163, 34, 60, 0, 164, 34, 60, 0, 165, 34, 60, 0, 166, 34, 60, 0, 167, 34, 60, 0, 168, 34, 60, 0, 169, 34, 60, 0, 170, 34, 60, 0, 171, 34, 60, 0, 176, 34, 60, 0, 177, 34, 60, 0, 178, 34, 60, 0, 179, 34, 60, 0, 180, 34, 60, 0, 181, 34, 60, 0, 182, 34, 60, 0, 183, 34, 60, 0, 184, 34, 60, 0, 185, 34, 60, 0, 186, 34, 60, 0, 187, 34, 60, 0, 188, 34, 60, 0, 75, 33, 60, 0, 189, 34, 60, 0, 190, 34, 60, 0, 191, 34, 60, 0, 192, 34, 60, 0, 193, 34, 60, 0, 194, 34, 60, 0, 195, 34, 60, 0, 196, 34, 60, 0, 197, 34, 60, 0, 198, 34, 60, 0, 199, 34, 60, 0, 200, 34, 60, 0, 201, 34, 60, 0, 202, 34, 60, 0, 203, 34, 60, 0, 204, 34, 60, 0, 205, 34, 60, 0, 206, 34, 60, 0, 207, 34, 60, 0, 208, 34, 60, 0, 209, 34, 60, 0, 210, 34, 60, 0, 211, 34, 60, 0, 212, 34, 60, 0, 213, 34, 60, 0, 214, 34, 60, 0, 215, 34, 60, 0, 216, 34, 60, 0, 217, 34, 60, 0, 218, 34, 60, 0, 219, 34, 60, 0, 220, 34, 60, 0, 221, 34, 60, 0, 222, 34, 60, 0, 223, 34, 60, 0, 228, 34, 60, 0, 229, 34, 60, 0, 230, 34, 60, 0, 231, 34, 60, 0, 232, 34, 60, 0, 233, 34, 60, 0, 238, 34, 60, 0, 239, 34, 60, 0, 240, 34, 60, 0, 241, 34, 60, 0, 242, 34, 60, 0, 243, 34, 60, 0, 244, 34, 60, 0, 245, 34, 60, 0, 246, 34, 60, 0, 247, 34, 60, 0, 248, 34, 60, 0, 249, 34, 60, 0, 250, 34, 60, 0, 251, 34, 60, 0, 252, 34, 60, 0, 253, 34, 60, 0, 254, 34, 60, 0, 255, 34, 60, 0, 0, 35, 60, 0, 1, 35, 60, 0, 2, 35, 60, 0, 3, 35, 60, 0, 4, 35, 60, 0, 5, 35, 60, 0, 6, 35, 60, 0, 7, 35, 60, 0, 12, 35, 60, 0, 13, 35, 60, 0, 14, 35, 60, 0, 15, 35, 60, 0, 16, 35, 60, 0, 17, 35, 60, 0, 18, 35, 60, 0, 19, 35, 60, 0, 20, 35, 60, 0, 21, 35, 60, 0, 22, 35, 60, 0, 23, 35, 60, 0, 24, 35, 60, 0, 25, 35, 60, 0, 26, 35, 60, 0, 27, 35, 60, 0, 28, 35, 60, 0, 29, 35, 60, 0, 30, 35, 60, 0, 31, 35, 60, 0, 32, 35, 60, 0, 33, 35, 60, 0, 34, 35, 60, 0, 35, 35, 60, 0, 36, 35, 60, 0, 37, 35, 60, 0, 38, 35, 60, 0, 39, 35, 60, 0, 40, 35, 60, 0, 43, 35, 60, 0, 44, 35, 60, 0, 45, 35, 60, 0, 46, 35, 60, 0, 47, 35, 60, 0, 48, 35, 60, 0, 49, 35, 60, 0, 50, 35, 60, 0, 51, 35, 60, 0, 52, 35, 60, 0, 53, 35, 60, 0, 54, 35, 60, 0, 55, 35, 60, 0, 56, 35, 60, 0, 57, 35, 60, 0, 58, 35, 60, 0, 59, 35, 60, 0, 60, 35, 60, 0, 61, 35, 60, 0, 62, 35, 60, 0, 63, 35, 60, 0, 64, 35, 60, 0, 65, 35, 60, 0, 66, 35, 60, 0, 67, 35, 60, 0, 68, 35, 60, 0, 69, 35, 60, 0, 70, 35, 60, 0, 71, 35, 60, 0, 72, 35, 60, 0, 73, 35, 60, 0, 74, 35, 60, 0, 75, 35, 60, 0, 76, 35, 60, 0, 77, 35, 60, 0, 78, 35, 60, 0, 79, 35, 60, 0, 80, 35, 60, 0, 81, 35, 60, 0, 82, 35, 60, 0, 83, 35, 60, 0, 84, 35, 60, 0, 85, 35, 60, 0, 86, 35, 60, 0, 87, 35, 60, 0, 88, 35, 60, 0, 89, 35, 60, 0, 90, 35, 60, 0, 91, 35, 60, 0, 92, 35, 60, 0, 93, 35, 60, 0, 94, 35, 60, 0, 95, 35, 60, 0, 96, 35, 60, 0, 97, 35, 60, 0, 98, 35, 60, 0, 99, 35, 60, 0, 100, 35, 60, 0, 101, 35, 60, 0, 102, 35, 60, 0, 103, 35, 60, 0, 104, 35, 60, 0, 105, 35, 60, 0, 106, 35, 60, 0, 107, 35, 60, 0, 108, 35, 60, 0, 109, 35, 60, 0, 110, 35, 60, 0, 111, 35, 60, 0, 112, 35, 60, 0, 113, 35, 60, 0, 114, 35, 60, 0, 115, 35, 60, 0, 116, 35, 60, 0, 117, 35, 60, 0, 118, 35, 60, 0, 119, 35, 60, 0, 120, 35, 60, 0, 121, 35, 60, 0, 122, 35, 60, 0, 123, 35, 60, 0, 124, 35, 60, 0, 125, 35, 60, 0, 126, 35, 60, 0, 127, 35, 60, 0, 128, 35, 60, 0, 129, 35, 60, 0, 130, 35, 60, 0, 131, 35, 60, 0, 132, 35, 60, 0, 133, 35, 60, 0, 134, 35, 60, 0, 135, 35, 60, 0, 136, 35, 60, 0, 137, 35, 60, 0, 138, 35, 60, 0, 139, 35, 60, 0, 140, 35, 60, 0, 141, 35, 60, 0, 142, 35, 60, 0, 143, 35, 60, 0, 144, 35, 60, 0, 145, 35, 60, 0, 146, 35, 60, 0, 147, 35, 60, 0, 148, 35, 60, 0, 149, 35, 60, 0, 150, 35, 60, 0, 151, 35, 60, 0, 152, 35, 60, 0, 153, 35, 60, 0, 154, 35, 60, 0, 155, 35, 60, 0, 156, 35, 60, 0, 157, 35, 60, 0, 158, 35, 60, 0, 159, 35, 60, 0, 160, 35, 60, 0, 161, 35, 60, 0, 162, 35, 60, 0, 163, 35, 60, 0, 164, 35, 60, 0, 165, 35, 60, 0, 166, 35, 60, 0, 167, 35, 60, 0, 168, 35, 60, 0, 169, 35, 60, 0, 170, 35, 60, 0, 171, 35, 60, 0, 172, 35, 60, 0, 173, 35, 60, 0, 174, 35, 60, 0, 175, 35, 60, 0, 176, 35, 60, 0, 177, 35, 60, 0, 178, 35, 60, 0, 179, 35, 60, 0, 180, 35, 60, 0, 181, 35, 60, 0, 182, 35, 60, 0, 183, 35, 60, 0, 184, 35, 60, 0, 185, 35, 60, 0, 186, 35, 60, 0, 187, 35, 60, 0, 188, 35, 60, 0, 189, 35, 60, 0, 190, 35, 60, 0, 191, 35, 60, 0, 192, 35, 60, 0, 193, 35, 60, 0, 194, 35, 60, 0, 195, 35, 60, 0, 196, 35, 60, 0, 197, 35, 60, 0, 198, 35, 60, 0, 199, 35, 60, 0, 200, 35, 60, 0, 201, 35, 60, 0, 202, 35, 60, 0, 203, 35, 60, 0, 204, 35, 60, 0, 205, 35, 60, 0, 206, 35, 60, 0, 207, 35, 60, 0, 208, 35, 60, 0, 209, 35, 60, 0, 210, 35, 60, 0, 211, 35, 60, 0, 212, 35, 60, 0, 213, 35, 60, 0, 214, 35, 60, 0, 215, 35, 60, 0, 216, 35, 60, 0, 217, 35, 60, 0, 218, 35, 60, 0, 219, 35, 60, 0, 220, 35, 60, 0, 221, 35, 60, 0, 222, 35, 60, 0, 223, 35, 60, 0, 224, 35, 60, 0, 225, 35, 60, 0, 226, 35, 60, 0, 227, 35, 60, 0, 228, 35, 60, 0, 229, 35, 60, 0, 230, 35, 60, 0, 231, 35, 60, 0, 232, 35, 60, 0, 233, 35, 60, 0, 234, 35, 60, 0, 235, 35, 60, 0, 236, 35, 60, 0, 237, 35, 60, 0, 238, 35, 60, 0, 239, 35, 60, 0, 240, 35, 60, 0, 241, 35, 60, 0, 242, 35, 60, 0, 243, 35, 60, 0, 244, 35, 60, 0, 245, 35, 60, 0, 246, 35, 60, 0, 247, 35, 60, 0, 248, 35, 60, 0, 249, 35, 60, 0, 250, 35, 60, 0, 251, 35, 60, 0, 252, 35, 60, 0, 253, 35, 60, 0, 254, 35, 60, 0, 255, 35, 60, 0, 0, 36, 60, 0, 1, 36, 60, 0, 2, 36, 60, 0, 3, 36, 60, 0, 4, 36, 60, 0, 5, 36, 60, 0, 6, 36, 60, 0, 7, 36, 60, 0, 8, 36, 60, 0, 9, 36, 60, 0, 10, 36, 60, 0, 11, 36, 60, 0, 12, 36, 60, 0, 13, 36, 60, 0, 14, 36, 60, 0, 15, 36, 60, 0, 16, 36, 60, 0, 17, 36, 60, 0, 18, 36, 60, 0, 19, 36, 60, 0, 20, 36, 60, 0, 21, 36, 60, 0, 22, 36, 60, 0, 23, 36, 60, 0, 24, 36, 60, 0, 25, 36, 60, 0, 26, 36, 60, 0, 27, 36, 60, 0, 28, 36, 60, 0, 29, 36, 60, 0, 30, 36, 60, 0, 31, 36, 60, 0, 32, 36, 60, 0, 33, 36, 60, 0, 34, 36, 60, 0, 35, 36, 60, 0, 36, 36, 60, 0, 37, 36, 60, 0, 38, 36, 60, 0, 64, 36, 60, 0, 65, 36, 60, 0, 66, 36, 60, 0, 67, 36, 60, 0, 68, 36, 60, 0, 69, 36, 60, 0, 70, 36, 60, 0, 71, 36, 60, 0, 72, 36, 60, 0, 73, 36, 60, 0, 74, 36, 60, 0, 0, 37, 60, 0, 1, 37, 60, 0, 2, 37, 60, 0, 60, 0, 60, 0, 232, 255, 60, 0, 3, 37, 60, 0, 4, 37, 60, 0, 5, 37, 60, 0, 6, 37, 60, 0, 7, 37, 60, 0, 8, 37, 60, 0, 9, 37, 60, 0, 10, 37, 60, 0, 11, 37, 60, 0, 12, 37, 60, 0, 13, 37, 60, 0, 14, 37, 60, 0, 15, 37, 60, 0, 16, 37, 60, 0, 17, 37, 60, 0, 18, 37, 60, 0, 19, 37, 60, 0, 20, 37, 60, 0, 21, 37, 60, 0, 22, 37, 60, 0, 23, 37, 60, 0, 24, 37, 60, 0, 25, 37, 60, 0, 26, 37, 60, 0, 27, 37, 60, 0, 28, 37, 60, 0, 29, 37, 60, 0, 30, 37, 60, 0, 31, 37, 60, 0, 32, 37, 60, 0, 33, 37, 60, 0, 34, 37, 60, 0, 35, 37, 60, 0, 36, 37, 60, 0, 37, 37, 60, 0, 38, 37, 60, 0, 39, 37, 60, 0, 40, 37, 60, 0, 41, 37, 60, 0, 42, 37, 60, 0, 43, 37, 60, 0, 44, 37, 60, 0, 45, 37, 60, 0, 46, 37, 60, 0, 47, 37, 60, 0, 48, 37, 60, 0, 49, 37, 60, 0, 50, 37, 60, 0, 51, 37, 60, 0, 52, 37, 60, 0, 53, 37, 60, 0, 54, 37, 60, 0, 55, 37, 60, 0, 56, 37, 60, 0, 57, 37, 60, 0, 58, 37, 60, 0, 59, 37, 60, 0, 60, 37, 60, 0, 61, 37, 60, 0, 62, 37, 60, 0, 63, 37, 60, 0, 64, 37, 60, 0, 65, 37, 60, 0, 66, 37, 60, 0, 67, 37, 60, 0, 68, 37, 60, 0, 69, 37, 60, 0, 70, 37, 60, 0, 71, 37, 60, 0, 72, 37, 60, 0, 73, 37, 60, 0, 74, 37, 60, 0, 75, 37, 60, 0, 76, 37, 60, 0, 77, 37, 60, 0, 78, 37, 60, 0, 79, 37, 60, 0, 80, 37, 60, 0, 81, 37, 60, 0, 82, 37, 60, 0, 83, 37, 60, 0, 84, 37, 60, 0, 85, 37, 60, 0, 86, 37, 60, 0, 87, 37, 60, 0, 88, 37, 60, 0, 89, 37, 60, 0, 90, 37, 60, 0, 91, 37, 60, 0, 92, 37, 60, 0, 93, 37, 60, 0, 94, 37, 60, 0, 95, 37, 60, 0, 96, 37, 60, 0, 97, 37, 60, 0, 98, 37, 60, 0, 99, 37, 60, 0, 100, 37, 60, 0, 101, 37, 60, 0, 102, 37, 60, 0, 103, 37, 60, 0, 104, 37, 60, 0, 105, 37, 60, 0, 106, 37, 60, 0, 107, 37, 60, 0, 108, 37, 60, 0, 109, 37, 60, 0, 110, 37, 60, 0, 111, 37, 60, 0, 112, 37, 60, 0, 113, 37, 60, 0, 114, 37, 60, 0, 115, 37, 60, 0, 116, 37, 60, 0, 117, 37, 60, 0, 118, 37, 60, 0, 119, 37, 60, 0, 120, 37, 60, 0, 121, 37, 60, 0, 122, 37, 60, 0, 123, 37, 60, 0, 124, 37, 60, 0, 125, 37, 60, 0, 126, 37, 60, 0, 127, 37, 60, 0, 128, 37, 60, 0, 129, 37, 60, 0, 130, 37, 60, 0, 131, 37, 60, 0, 132, 37, 60, 0, 133, 37, 60, 0, 134, 37, 60, 0, 135, 37, 60, 0, 136, 37, 60, 0, 137, 37, 60, 0, 138, 37, 60, 0, 139, 37, 60, 0, 140, 37, 60, 0, 141, 37, 60, 0, 142, 37, 60, 0, 143, 37, 60, 0, 144, 37, 60, 0, 145, 37, 60, 0, 146, 37, 60, 0, 147, 37, 60, 0, 148, 37, 60, 0, 149, 37, 60, 0, 150, 37, 60, 0, 151, 37, 60, 0, 152, 37, 60, 0, 153, 37, 60, 0, 154, 37, 60, 0, 155, 37, 60, 0, 156, 37, 60, 0, 157, 37, 60, 0, 158, 37, 60, 0, 159, 37, 60, 0, 160, 37, 60, 0, 60, 0, 60, 0, 237, 255, 60, 0, 161, 37, 60, 0, 162, 37, 60, 0, 163, 37, 60, 0, 164, 37, 60, 0, 165, 37, 60, 0, 166, 37, 60, 0, 167, 37, 60, 0, 168, 37, 60, 0, 169, 37, 60, 0, 170, 37, 60, 0, 171, 37, 60, 0, 172, 37, 60, 0, 173, 37, 60, 0, 174, 37, 60, 0, 175, 37, 60, 0, 176, 37, 60, 0, 177, 37, 60, 0, 178, 37, 60, 0, 179, 37, 60, 0, 180, 37, 60, 0, 181, 37, 60, 0, 182, 37, 60, 0, 183, 37, 60, 0, 184, 37, 60, 0, 185, 37, 60, 0, 186, 37, 60, 0, 187, 37, 60, 0, 188, 37, 60, 0, 189, 37, 60, 0, 190, 37, 60, 0, 191, 37, 60, 0, 192, 37, 60, 0, 193, 37, 60, 0, 194, 37, 60, 0, 195, 37, 60, 0, 196, 37, 60, 0, 197, 37, 60, 0, 198, 37, 60, 0, 199, 37, 60, 0, 200, 37, 60, 0, 201, 37, 60, 0, 202, 37, 60, 0, 203, 37, 60, 0, 60, 0, 60, 0, 238, 255, 60, 0, 204, 37, 60, 0, 205, 37, 60, 0, 206, 37, 60, 0, 207, 37, 60, 0, 208, 37, 60, 0, 209, 37, 60, 0, 210, 37, 60, 0, 211, 37, 60, 0, 212, 37, 60, 0, 213, 37, 60, 0, 214, 37, 60, 0, 215, 37, 60, 0, 216, 37, 60, 0, 217, 37, 60, 0, 218, 37, 60, 0, 219, 37, 60, 0, 220, 37, 60, 0, 221, 37, 60, 0, 222, 37, 60, 0, 223, 37, 60, 0, 224, 37, 60, 0, 225, 37, 60, 0, 226, 37, 60, 0, 227, 37, 60, 0, 228, 37, 60, 0, 229, 37, 60, 0, 230, 37, 60, 0, 231, 37, 60, 0, 232, 37, 60, 0, 233, 37, 60, 0, 234, 37, 60, 0, 235, 37, 60, 0, 236, 37, 60, 0, 237, 37, 60, 0, 238, 37, 60, 0, 239, 37, 60, 0, 240, 37, 60, 0, 241, 37, 60, 0, 242, 37, 60, 0, 243, 37, 60, 0, 244, 37, 60, 0, 245, 37, 60, 0, 246, 37, 60, 0, 247, 37, 60, 0, 248, 37, 60, 0, 249, 37, 60, 0, 250, 37, 60, 0, 251, 37, 60, 0, 252, 37, 60, 0, 253, 37, 60, 0, 254, 37, 60, 0, 255, 37, 60, 0, 62, 216, 0, 223, 60, 0, 62, 216, 1, 223, 60, 0, 62, 216, 2, 223, 60, 0, 62, 216, 3, 223, 60, 0, 62, 216, 4, 223, 60, 0, 62, 216, 5, 223, 60, 0, 62, 216, 6, 223, 60, 0, 62, 216, 7, 223, 60, 0, 62, 216, 8, 223, 60, 0, 62, 216, 9, 223, 60, 0, 62, 216, 10, 223, 60, 0, 62, 216, 11, 223, 60, 0, 62, 216, 12, 223, 60, 0, 62, 216, 13, 223, 60, 0, 62, 216, 14, 223, 60, 0, 62, 216, 15, 223, 60, 0, 62, 216, 16, 223, 60, 0, 62, 216, 17, 223, 60, 0, 62, 216, 18, 223, 60, 0, 62, 216, 19, 223, 60, 0, 62, 216, 20, 223, 60, 0, 62, 216, 21, 223, 60, 0, 62, 216, 22, 223, 60, 0, 62, 216, 23, 223, 60, 0, 62, 216, 24, 223, 60, 0, 62, 216, 25, 223, 60, 0, 62, 216, 26, 223, 60, 0, 62, 216, 27, 223, 60, 0, 62, 216, 28, 223, 60, 0, 62, 216, 29, 223, 60, 0, 62, 216, 30, 223, 60, 0, 62, 216, 31, 223, 60, 0, 62, 216, 32, 223, 60, 0, 62, 216, 33, 223, 60, 0, 62, 216, 34, 223, 60, 0, 62, 216, 35, 223, 60, 0, 62, 216, 36, 223, 60, 0, 62, 216, 37, 223, 60, 0, 62, 216, 38, 223, 60, 0, 62, 216, 39, 223, 60, 0, 62, 216, 40, 223, 60, 0, 62, 216, 41, 223, 60, 0, 62, 216, 42, 223, 60, 0, 62, 216, 43, 223, 60, 0, 62, 216, 44, 223, 60, 0, 62, 216, 45, 223, 60, 0, 62, 216, 46, 223, 60, 0, 62, 216, 47, 223, 60, 0, 62, 216, 48, 223, 60, 0, 62, 216, 49, 223, 60, 0, 62, 216, 50, 223, 60, 0, 62, 216, 51, 223, 60, 0, 62, 216, 52, 223, 60, 0, 62, 216, 53, 223, 60, 0, 62, 216, 54, 223, 60, 0, 62, 216, 55, 223, 60, 0, 62, 216, 56, 223, 60, 0, 62, 216, 57, 223, 60, 0, 62, 216, 58, 223, 60, 0, 62, 216, 59, 223, 60, 0, 62, 216, 60, 223, 60, 0, 62, 216, 61, 223, 60, 0, 62, 216, 62, 223, 60, 0, 62, 216, 63, 223, 60, 0, 62, 216, 64, 223, 60, 0, 62, 216, 65, 223, 60, 0, 62, 216, 66, 223, 60, 0, 62, 216, 67, 223, 60, 0, 62, 216, 68, 223, 60, 0, 62, 216, 69, 223, 60, 0, 62, 216, 70, 223, 60, 0, 62, 216, 71, 223, 60, 0, 62, 216, 72, 223, 60, 0, 62, 216, 73, 223, 60, 0, 62, 216, 74, 223, 60, 0, 62, 216, 75, 223, 60, 0, 62, 216, 76, 223, 60, 0, 62, 216, 77, 223, 60, 0, 62, 216, 78, 223, 60, 0, 62, 216, 79, 223, 60, 0, 62, 216, 80, 223, 60, 0, 62, 216, 81, 223, 60, 0, 62, 216, 82, 223, 60, 0, 62, 216, 83, 223, 60, 0, 62, 216, 84, 223, 60, 0, 62, 216, 85, 223, 60, 0, 62, 216, 86, 223, 60, 0, 62, 216, 87, 223, 60, 0, 62, 216, 88, 223, 60, 0, 62, 216, 89, 223, 60, 0, 62, 216, 90, 223, 60, 0, 62, 216, 91, 223, 60, 0, 62, 216, 92, 223, 60, 0, 62, 216, 93, 223, 60, 0, 62, 216, 94, 223, 60, 0, 62, 216, 95, 223, 60, 0, 62, 216, 96, 223, 60, 0, 62, 216, 97, 223, 60, 0, 62, 216, 98, 223, 60, 0, 62, 216, 99, 223, 60, 0, 62, 216, 100, 223, 60, 0, 62, 216, 101, 223, 60, 0, 62, 216, 102, 223, 60, 0, 62, 216, 103, 223, 60, 0, 62, 216, 104, 223, 60, 0, 62, 216, 105, 223, 60, 0, 62, 216, 106, 223, 60, 0, 62, 216, 107, 223, 60, 0, 62, 216, 108, 223, 60, 0, 62, 216, 109, 223, 60, 0, 62, 216, 110, 223, 60, 0, 62, 216, 111, 223, 60, 0, 62, 216, 112, 223, 60, 0, 62, 216, 113, 223, 60, 0, 62, 216, 114, 223, 60, 0, 62, 216, 115, 223, 60, 0, 62, 216, 116, 223, 60, 0, 62, 216, 117, 223, 60, 0, 62, 216, 118, 223, 60, 0, 62, 216, 119, 223, 60, 0, 62, 216, 120, 223, 60, 0, 62, 216, 121, 223, 60, 0, 62, 216, 122, 223, 60, 0, 62, 216, 123, 223, 60, 0, 62, 216, 124, 223, 60, 0, 62, 216, 125, 223, 60, 0, 62, 216, 126, 223, 60, 0, 62, 216, 127, 223, 60, 0, 62, 216, 128, 223, 60, 0, 62, 216, 129, 223, 60, 0, 62, 216, 130, 223, 60, 0, 62, 216, 131, 223, 60, 0, 62, 216, 132, 223, 60, 0, 62, 216, 133, 223, 60, 0, 62, 216, 134, 223, 60, 0, 62, 216, 135, 223, 60, 0, 62, 216, 136, 223, 60, 0, 62, 216, 137, 223, 60, 0, 62, 216, 138, 223, 60, 0, 62, 216, 139, 223, 60, 0, 62, 216, 140, 223, 60, 0, 62, 216, 141, 223, 60, 0, 62, 216, 142, 223, 60, 0, 62, 216, 143, 223, 60, 0, 62, 216, 144, 223, 60, 0, 62, 216, 145, 223, 60, 0, 62, 216, 146, 223, 60, 0, 62, 216, 148, 223, 60, 0, 62, 216, 149, 223, 60, 0, 62, 216, 150, 223, 60, 0, 62, 216, 151, 223, 60, 0, 62, 216, 152, 223, 60, 0, 62, 216, 153, 223, 60, 0, 62, 216, 154, 223, 60, 0, 62, 216, 155, 223, 60, 0, 62, 216, 156, 223, 60, 0, 62, 216, 157, 223, 60, 0, 62, 216, 158, 223, 60, 0, 62, 216, 159, 223, 60, 0, 62, 216, 160, 223, 60, 0, 62, 216, 161, 223, 60, 0, 62, 216, 162, 223, 60, 0, 62, 216, 163, 223, 60, 0, 62, 216, 164, 223, 60, 0, 62, 216, 165, 223, 60, 0, 62, 216, 166, 223, 60, 0, 62, 216, 167, 223, 60, 0, 62, 216, 168, 223, 60, 0, 62, 216, 169, 223, 60, 0, 62, 216, 170, 223, 60, 0, 62, 216, 171, 223, 60, 0, 62, 216, 172, 223, 60, 0, 62, 216, 173, 223, 60, 0, 62, 216, 174, 223, 60, 0, 62, 216, 175, 223, 60, 0, 62, 216, 176, 223, 60, 0, 62, 216, 177, 223, 60, 0, 62, 216, 178, 223, 60, 0, 62, 216, 179, 223, 60, 0, 62, 216, 180, 223, 60, 0, 62, 216, 181, 223, 60, 0, 62, 216, 182, 223, 60, 0, 62, 216, 183, 223, 60, 0, 62, 216, 184, 223, 60, 0, 62, 216, 185, 223, 60, 0, 62, 216, 186, 223, 60, 0, 62, 216, 187, 223, 60, 0, 62, 216, 188, 223, 60, 0, 62, 216, 189, 223, 60, 0, 62, 216, 190, 223, 60, 0, 62, 216, 191, 223, 60, 0, 62, 216, 192, 223, 60, 0, 62, 216, 193, 223, 60, 0, 62, 216, 194, 223, 60, 0, 62, 216, 195, 223, 60, 0, 62, 216, 196, 223, 60, 0, 62, 216, 197, 223, 60, 0, 62, 216, 198, 223, 60, 0, 62, 216, 199, 223, 60, 0, 62, 216, 200, 223, 60, 0, 62, 216, 201, 223, 60, 0, 62, 216, 202, 223, 60, 0, 0, 38, 60, 0, 1, 38, 60, 0, 2, 38, 60, 0, 3, 38, 60, 0, 4, 38, 60, 0, 5, 38, 60, 0, 6, 38, 60, 0, 7, 38, 60, 0, 8, 38, 60, 0, 9, 38, 60, 0, 10, 38, 60, 0, 11, 38, 60, 0, 12, 38, 60, 0, 13, 38, 60, 0, 14, 38, 60, 0, 15, 38, 60, 0, 16, 38, 60, 0, 17, 38, 60, 0, 18, 38, 60, 0, 19, 38, 60, 0, 20, 38, 60, 0, 21, 38, 60, 0, 22, 38, 60, 0, 23, 38, 60, 0, 24, 38, 60, 0, 25, 38, 60, 0, 26, 38, 60, 0, 27, 38, 60, 0, 28, 38, 60, 0, 29, 38, 60, 0, 30, 38, 60, 0, 31, 38, 60, 0, 32, 38, 60, 0, 33, 38, 60, 0, 34, 38, 60, 0, 35, 38, 60, 0, 36, 38, 60, 0, 37, 38, 60, 0, 38, 38, 60, 0, 39, 38, 60, 0, 40, 38, 60, 0, 41, 38, 60, 0, 42, 38, 60, 0, 43, 38, 60, 0, 44, 38, 60, 0, 45, 38, 60, 0, 46, 38, 60, 0, 47, 38, 60, 0, 56, 38, 60, 0, 57, 38, 60, 0, 58, 38, 60, 0, 59, 38, 60, 0, 60, 38, 60, 0, 61, 38, 60, 0, 62, 38, 60, 0, 63, 38, 60, 0, 64, 38, 60, 0, 65, 38, 60, 0, 66, 38, 60, 0, 67, 38, 60, 0, 68, 38, 60, 0, 69, 38, 60, 0, 70, 38, 60, 0, 71, 38, 60, 0, 72, 38, 60, 0, 73, 38, 60, 0, 74, 38, 60, 0, 75, 38, 60, 0, 76, 38, 60, 0, 77, 38, 60, 0, 78, 38, 60, 0, 79, 38, 60, 0, 80, 38, 60, 0, 81, 38, 60, 0, 82, 38, 60, 0, 83, 38, 60, 0, 84, 38, 60, 0, 85, 38, 60, 0, 86, 38, 60, 0, 87, 38, 60, 0, 88, 38, 60, 0, 89, 38, 60, 0, 90, 38, 60, 0, 91, 38, 60, 0, 92, 38, 60, 0, 93, 38, 60, 0, 94, 38, 60, 0, 95, 38, 60, 0, 96, 38, 60, 0, 97, 38, 60, 0, 98, 38, 60, 0, 99, 38, 60, 0, 100, 38, 60, 0, 101, 38, 60, 0, 102, 38, 60, 0, 103, 38, 60, 0, 104, 38, 60, 0, 105, 38, 60, 0, 106, 38, 60, 0, 107, 38, 60, 0, 108, 38, 60, 0, 112, 38, 60, 0, 113, 38, 60, 0, 114, 38, 60, 0, 115, 38, 60, 0, 116, 38, 60, 0, 117, 38, 60, 0, 118, 38, 60, 0, 119, 38, 60, 0, 120, 38, 60, 0, 121, 38, 60, 0, 122, 38, 60, 0, 123, 38, 60, 0, 124, 38, 60, 0, 125, 38, 60, 0, 126, 38, 60, 0, 127, 38, 60, 0, 128, 38, 60, 0, 129, 38, 60, 0, 130, 38, 60, 0, 131, 38, 60, 0, 132, 38, 60, 0, 133, 38, 60, 0, 134, 38, 60, 0, 135, 38, 60, 0, 136, 38, 60, 0, 137, 38, 60, 0, 144, 38, 60, 0, 145, 38, 60, 0, 146, 38, 60, 0, 147, 38, 60, 0, 148, 38, 60, 0, 149, 38, 60, 0, 150, 38, 60, 0, 151, 38, 60, 0, 152, 38, 60, 0, 153, 38, 60, 0, 154, 38, 60, 0, 155, 38, 60, 0, 156, 38, 60, 0, 157, 38, 60, 0, 158, 38, 60, 0, 159, 38, 60, 0, 160, 38, 60, 0, 161, 38, 60, 0, 162, 38, 60, 0, 163, 38, 60, 0, 164, 38, 60, 0, 165, 38, 60, 0, 166, 38, 60, 0, 167, 38, 60, 0, 168, 38, 60, 0, 169, 38, 60, 0, 170, 38, 60, 0, 171, 38, 60, 0, 172, 38, 60, 0, 173, 38, 60, 0, 174, 38, 60, 0, 175, 38, 60, 0, 176, 38, 60, 0, 177, 38, 60, 0, 178, 38, 60, 0, 179, 38, 60, 0, 180, 38, 60, 0, 181, 38, 60, 0, 182, 38, 60, 0, 183, 38, 60, 0, 184, 38, 60, 0, 185, 38, 60, 0, 186, 38, 60, 0, 187, 38, 60, 0, 188, 38, 60, 0, 189, 38, 60, 0, 190, 38, 60, 0, 191, 38, 60, 0, 192, 38, 60, 0, 193, 38, 60, 0, 194, 38, 60, 0, 195, 38, 60, 0, 196, 38, 60, 0, 197, 38, 60, 0, 198, 38, 60, 0, 199, 38, 60, 0, 200, 38, 60, 0, 201, 38, 60, 0, 202, 38, 60, 0, 203, 38, 60, 0, 204, 38, 60, 0, 205, 38, 60, 0, 206, 38, 60, 0, 207, 38, 60, 0, 208, 38, 60, 0, 209, 38, 60, 0, 210, 38, 60, 0, 211, 38, 60, 0, 212, 38, 60, 0, 213, 38, 60, 0, 214, 38, 60, 0, 215, 38, 60, 0, 216, 38, 60, 0, 217, 38, 60, 0, 218, 38, 60, 0, 219, 38, 60, 0, 220, 38, 60, 0, 221, 38, 60, 0, 222, 38, 60, 0, 223, 38, 60, 0, 224, 38, 60, 0, 225, 38, 60, 0, 226, 38, 60, 0, 227, 38, 60, 0, 228, 38, 60, 0, 229, 38, 60, 0, 230, 38, 60, 0, 231, 38, 60, 0, 232, 38, 60, 0, 233, 38, 60, 0, 234, 38, 60, 0, 235, 38, 60, 0, 236, 38, 60, 0, 237, 38, 60, 0, 238, 38, 60, 0, 239, 38, 60, 0, 240, 38, 60, 0, 241, 38, 60, 0, 242, 38, 60, 0, 243, 38, 60, 0, 244, 38, 60, 0, 245, 38, 60, 0, 246, 38, 60, 0, 247, 38, 60, 0, 248, 38, 60, 0, 249, 38, 60, 0, 250, 38, 60, 0, 251, 38, 60, 0, 252, 38, 60, 0, 253, 38, 60, 0, 254, 38, 60, 0, 255, 38, 60, 0, 60, 216, 230, 221, 60, 0, 60, 216, 231, 221, 60, 0, 60, 216, 232, 221, 60, 0, 60, 216, 233, 221, 60, 0, 60, 216, 234, 221, 60, 0, 60, 216, 235, 221, 60, 0, 60, 216, 236, 221, 60, 0, 60, 216, 237, 221, 60, 0, 60, 216, 238, 221, 60, 0, 60, 216, 239, 221, 60, 0, 60, 216, 240, 221, 60, 0, 60, 216, 241, 221, 60, 0, 60, 216, 242, 221, 60, 0, 60, 216, 243, 221, 60, 0, 60, 216, 244, 221, 60, 0, 60, 216, 245, 221, 60, 0, 60, 216, 246, 221, 60, 0, 60, 216, 247, 221, 60, 0, 60, 216, 248, 221, 60, 0, 60, 216, 249, 221, 60, 0, 60, 216, 250, 221, 60, 0, 60, 216, 251, 221, 60, 0, 60, 216, 252, 221, 60, 0, 60, 216, 253, 221, 60, 0, 60, 216, 254, 221, 60, 0, 60, 216, 255, 221, 60, 0, 0, 39, 60, 0, 1, 39, 60, 0, 2, 39, 60, 0, 3, 39, 60, 0, 4, 39, 60, 0, 5, 39, 60, 0, 6, 39, 60, 0, 7, 39, 60, 0, 8, 39, 60, 0, 9, 39, 60, 0, 10, 39, 60, 0, 11, 39, 60, 0, 12, 39, 60, 0, 13, 39, 60, 0, 14, 39, 60, 0, 15, 39, 60, 0, 16, 39, 60, 0, 17, 39, 60, 0, 18, 39, 60, 0, 19, 39, 60, 0, 20, 39, 60, 0, 21, 39, 60, 0, 22, 39, 60, 0, 23, 39, 60, 0, 24, 39, 60, 0, 25, 39, 60, 0, 26, 39, 60, 0, 27, 39, 60, 0, 28, 39, 60, 0, 29, 39, 60, 0, 30, 39, 60, 0, 31, 39, 60, 0, 32, 39, 60, 0, 80, 46, 60, 0, 81, 46, 60, 0, 33, 39, 60, 0, 34, 39, 60, 0, 35, 39, 60, 0, 36, 39, 60, 0, 37, 39, 60, 0, 38, 39, 60, 0, 39, 39, 60, 0, 40, 39, 60, 0, 41, 39, 60, 0, 42, 39, 60, 0, 43, 39, 60, 0, 44, 39, 60, 0, 45, 39, 60, 0, 46, 39, 60, 0, 47, 39, 60, 0, 48, 39, 60, 0, 49, 39, 60, 0, 50, 39, 60, 0, 51, 39, 60, 0, 52, 39, 60, 0, 53, 39, 60, 0, 54, 39, 60, 0, 55, 39, 60, 0, 56, 39, 60, 0, 57, 39, 60, 0, 58, 39, 60, 0, 59, 39, 60, 0, 60, 39, 60, 0, 61, 39, 60, 0, 62, 39, 60, 0, 63, 39, 60, 0, 64, 39, 60, 0, 65, 39, 60, 0, 66, 39, 60, 0, 67, 39, 60, 0, 68, 39, 60, 0, 69, 39, 60, 0, 70, 39, 60, 0, 71, 39, 60, 0, 72, 39, 60, 0, 73, 39, 60, 0, 74, 39, 60, 0, 75, 39, 60, 0, 76, 39, 60, 0, 77, 39, 60, 0, 78, 39, 60, 0, 79, 39, 60, 0, 80, 39, 60, 0, 81, 39, 60, 0, 82, 39, 60, 0, 83, 39, 60, 0, 84, 39, 60, 0, 85, 39, 60, 0, 86, 39, 60, 0, 87, 39, 60, 0, 88, 39, 60, 0, 89, 39, 60, 0, 90, 39, 60, 0, 91, 39, 60, 0, 92, 39, 60, 0, 93, 39, 60, 0, 94, 39, 60, 0, 95, 39, 60, 0, 96, 39, 60, 0, 97, 39, 60, 0, 98, 39, 60, 0, 99, 39, 60, 0, 100, 39, 60, 0, 101, 39, 60, 0, 102, 39, 60, 0, 103, 39, 60, 0, 148, 39, 60, 0, 149, 39, 60, 0, 150, 39, 60, 0, 151, 39, 60, 0, 152, 39, 60, 0, 153, 39, 60, 0, 154, 39, 60, 0, 155, 39, 60, 0, 156, 39, 60, 0, 157, 39, 60, 0, 158, 39, 60, 0, 159, 39, 60, 0, 160, 39, 60, 0, 161, 39, 60, 0, 162, 39, 60, 0, 163, 39, 60, 0, 164, 39, 60, 0, 165, 39, 60, 0, 166, 39, 60, 0, 167, 39, 60, 0, 168, 39, 60, 0, 169, 39, 60, 0, 170, 39, 60, 0, 171, 39, 60, 0, 172, 39, 60, 0, 173, 39, 60, 0, 174, 39, 60, 0, 175, 39, 60, 0, 176, 39, 60, 0, 177, 39, 60, 0, 178, 39, 60, 0, 179, 39, 60, 0, 180, 39, 60, 0, 181, 39, 60, 0, 182, 39, 60, 0, 183, 39, 60, 0, 184, 39, 60, 0, 185, 39, 60, 0, 186, 39, 60, 0, 187, 39, 60, 0, 188, 39, 60, 0, 189, 39, 60, 0, 190, 39, 60, 0, 191, 39, 60, 0, 192, 39, 60, 0, 193, 39, 60, 0, 194, 39, 60, 0, 195, 39, 60, 0, 196, 39, 60, 0, 199, 39, 60, 0, 200, 39, 60, 0, 201, 39, 60, 0, 202, 39, 60, 0, 203, 39, 60, 0, 204, 39, 60, 0, 205, 39, 60, 0, 206, 39, 60, 0, 207, 39, 60, 0, 208, 39, 60, 0, 209, 39, 60, 0, 210, 39, 60, 0, 211, 39, 60, 0, 212, 39, 60, 0, 213, 39, 60, 0, 214, 39, 60, 0, 215, 39, 60, 0, 216, 39, 60, 0, 217, 39, 60, 0, 218, 39, 60, 0, 219, 39, 60, 0, 220, 39, 60, 0, 221, 39, 60, 0, 222, 39, 60, 0, 223, 39, 60, 0, 224, 39, 60, 0, 225, 39, 60, 0, 226, 39, 60, 0, 227, 39, 60, 0, 228, 39, 60, 0, 229, 39, 60, 0, 240, 39, 60, 0, 241, 39, 60, 0, 242, 39, 60, 0, 243, 39, 60, 0, 244, 39, 60, 0, 245, 39, 60, 0, 246, 39, 60, 0, 247, 39, 60, 0, 248, 39, 60, 0, 249, 39, 60, 0, 250, 39, 60, 0, 251, 39, 60, 0, 252, 39, 60, 0, 253, 39, 60, 0, 254, 39, 60, 0, 255, 39, 60, 0, 0, 41, 60, 0, 1, 41, 60, 0, 2, 41, 60, 0, 3, 41, 60, 0, 4, 41, 60, 0, 5, 41, 60, 0, 6, 41, 60, 0, 7, 41, 60, 0, 8, 41, 60, 0, 9, 41, 60, 0, 10, 41, 60, 0, 11, 41, 60, 0, 12, 41, 60, 0, 13, 41, 60, 0, 14, 41, 60, 0, 15, 41, 60, 0, 16, 41, 60, 0, 17, 41, 60, 0, 18, 41, 60, 0, 19, 41, 60, 0, 20, 41, 60, 0, 21, 41, 60, 0, 22, 41, 60, 0, 23, 41, 60, 0, 24, 41, 60, 0, 25, 41, 60, 0, 26, 41, 60, 0, 27, 41, 60, 0, 28, 41, 60, 0, 29, 41, 60, 0, 30, 41, 60, 0, 31, 41, 60, 0, 32, 41, 60, 0, 33, 41, 60, 0, 34, 41, 60, 0, 35, 41, 60, 0, 36, 41, 60, 0, 37, 41, 60, 0, 38, 41, 60, 0, 39, 41, 60, 0, 40, 41, 60, 0, 41, 41, 60, 0, 42, 41, 60, 0, 43, 41, 60, 0, 44, 41, 60, 0, 45, 41, 60, 0, 46, 41, 60, 0, 47, 41, 60, 0, 48, 41, 60, 0, 49, 41, 60, 0, 50, 41, 60, 0, 51, 41, 60, 0, 52, 41, 60, 0, 53, 41, 60, 0, 54, 41, 60, 0, 55, 41, 60, 0, 56, 41, 60, 0, 57, 41, 60, 0, 58, 41, 60, 0, 59, 41, 60, 0, 60, 41, 60, 0, 61, 41, 60, 0, 62, 41, 60, 0, 63, 41, 60, 0, 64, 41, 60, 0, 65, 41, 60, 0, 66, 41, 60, 0, 67, 41, 60, 0, 68, 41, 60, 0, 69, 41, 60, 0, 70, 41, 60, 0, 71, 41, 60, 0, 72, 41, 60, 0, 73, 41, 60, 0, 74, 41, 60, 0, 75, 41, 60, 0, 76, 41, 60, 0, 77, 41, 60, 0, 78, 41, 60, 0, 79, 41, 60, 0, 80, 41, 60, 0, 81, 41, 60, 0, 82, 41, 60, 0, 83, 41, 60, 0, 84, 41, 60, 0, 85, 41, 60, 0, 86, 41, 60, 0, 87, 41, 60, 0, 88, 41, 60, 0, 89, 41, 60, 0, 90, 41, 60, 0, 91, 41, 60, 0, 92, 41, 60, 0, 93, 41, 60, 0, 94, 41, 60, 0, 95, 41, 60, 0, 96, 41, 60, 0, 97, 41, 60, 0, 98, 41, 60, 0, 99, 41, 60, 0, 100, 41, 60, 0, 101, 41, 60, 0, 102, 41, 60, 0, 103, 41, 60, 0, 104, 41, 60, 0, 105, 41, 60, 0, 106, 41, 60, 0, 107, 41, 60, 0, 108, 41, 60, 0, 109, 41, 60, 0, 110, 41, 60, 0, 111, 41, 60, 0, 112, 41, 60, 0, 113, 41, 60, 0, 114, 41, 60, 0, 115, 41, 60, 0, 116, 41, 60, 0, 117, 41, 60, 0, 118, 41, 60, 0, 119, 41, 60, 0, 120, 41, 60, 0, 121, 41, 60, 0, 122, 41, 60, 0, 123, 41, 60, 0, 124, 41, 60, 0, 125, 41, 60, 0, 126, 41, 60, 0, 127, 41, 60, 0, 128, 41, 60, 0, 129, 41, 60, 0, 130, 41, 60, 0, 153, 41, 60, 0, 154, 41, 60, 0, 155, 41, 60, 0, 156, 41, 60, 0, 157, 41, 60, 0, 158, 41, 60, 0, 159, 41, 60, 0, 160, 41, 60, 0, 161, 41, 60, 0, 162, 41, 60, 0, 163, 41, 60, 0, 164, 41, 60, 0, 165, 41, 60, 0, 166, 41, 60, 0, 167, 41, 60, 0, 168, 41, 60, 0, 169, 41, 60, 0, 170, 41, 60, 0, 171, 41, 60, 0, 172, 41, 60, 0, 173, 41, 60, 0, 174, 41, 60, 0, 175, 41, 60, 0, 176, 41, 60, 0, 177, 41, 60, 0, 178, 41, 60, 0, 179, 41, 60, 0, 180, 41, 60, 0, 181, 41, 60, 0, 182, 41, 60, 0, 183, 41, 60, 0, 184, 41, 60, 0, 185, 41, 60, 0, 186, 41, 60, 0, 187, 41, 60, 0, 188, 41, 60, 0, 189, 41, 60, 0, 190, 41, 60, 0, 191, 41, 60, 0, 192, 41, 60, 0, 193, 41, 60, 0, 194, 41, 60, 0, 195, 41, 60, 0, 196, 41, 60, 0, 197, 41, 60, 0, 198, 41, 60, 0, 199, 41, 60, 0, 200, 41, 60, 0, 201, 41, 60, 0, 202, 41, 60, 0, 203, 41, 60, 0, 204, 41, 60, 0, 205, 41, 60, 0, 206, 41, 60, 0, 207, 41, 60, 0, 208, 41, 60, 0, 209, 41, 60, 0, 210, 41, 60, 0, 211, 41, 60, 0, 212, 41, 60, 0, 213, 41, 60, 0, 214, 41, 60, 0, 215, 41, 60, 0, 220, 41, 60, 0, 221, 41, 60, 0, 222, 41, 60, 0, 223, 41, 60, 0, 224, 41, 60, 0, 225, 41, 60, 0, 226, 41, 60, 0, 227, 41, 60, 0, 228, 41, 60, 0, 229, 41, 60, 0, 230, 41, 60, 0, 231, 41, 60, 0, 232, 41, 60, 0, 233, 41, 60, 0, 234, 41, 60, 0, 235, 41, 60, 0, 236, 41, 60, 0, 237, 41, 60, 0, 238, 41, 60, 0, 239, 41, 60, 0, 240, 41, 60, 0, 241, 41, 60, 0, 242, 41, 60, 0, 243, 41, 60, 0, 244, 41, 60, 0, 245, 41, 60, 0, 246, 41, 60, 0, 247, 41, 60, 0, 248, 41, 60, 0, 249, 41, 60, 0, 250, 41, 60, 0, 251, 41, 60, 0, 254, 41, 60, 0, 255, 41, 60, 0, 0, 42, 60, 0, 1, 42, 60, 0, 2, 42, 60, 0, 3, 42, 60, 0, 4, 42, 60, 0, 5, 42, 60, 0, 6, 42, 60, 0, 7, 42, 60, 0, 8, 42, 60, 0, 9, 42, 60, 0, 10, 42, 60, 0, 11, 42, 60, 0, 13, 42, 60, 0, 14, 42, 60, 0, 15, 42, 60, 0, 16, 42, 60, 0, 17, 42, 60, 0, 18, 42, 60, 0, 19, 42, 60, 0, 20, 42, 60, 0, 21, 42, 60, 0, 22, 42, 60, 0, 23, 42, 60, 0, 24, 42, 60, 0, 25, 42, 60, 0, 26, 42, 60, 0, 27, 42, 60, 0, 28, 42, 60, 0, 29, 42, 60, 0, 30, 42, 60, 0, 31, 42, 60, 0, 32, 42, 60, 0, 33, 42, 60, 0, 34, 42, 60, 0, 35, 42, 60, 0, 36, 42, 60, 0, 37, 42, 60, 0, 38, 42, 60, 0, 39, 42, 60, 0, 40, 42, 60, 0, 41, 42, 60, 0, 42, 42, 60, 0, 43, 42, 60, 0, 44, 42, 60, 0, 45, 42, 60, 0, 46, 42, 60, 0, 47, 42, 60, 0, 48, 42, 60, 0, 49, 42, 60, 0, 50, 42, 60, 0, 51, 42, 60, 0, 52, 42, 60, 0, 53, 42, 60, 0, 54, 42, 60, 0, 55, 42, 60, 0, 56, 42, 60, 0, 57, 42, 60, 0, 58, 42, 60, 0, 59, 42, 60, 0, 60, 42, 60, 0, 61, 42, 60, 0, 62, 42, 60, 0, 63, 42, 60, 0, 64, 42, 60, 0, 65, 42, 60, 0, 66, 42, 60, 0, 67, 42, 60, 0, 68, 42, 60, 0, 69, 42, 60, 0, 70, 42, 60, 0, 71, 42, 60, 0, 72, 42, 60, 0, 73, 42, 60, 0, 74, 42, 60, 0, 75, 42, 60, 0, 76, 42, 60, 0, 77, 42, 60, 0, 78, 42, 60, 0, 79, 42, 60, 0, 80, 42, 60, 0, 81, 42, 60, 0, 82, 42, 60, 0, 83, 42, 60, 0, 84, 42, 60, 0, 85, 42, 60, 0, 86, 42, 60, 0, 87, 42, 60, 0, 88, 42, 60, 0, 89, 42, 60, 0, 90, 42, 60, 0, 91, 42, 60, 0, 92, 42, 60, 0, 93, 42, 60, 0, 94, 42, 60, 0, 95, 42, 60, 0, 96, 42, 60, 0, 97, 42, 60, 0, 98, 42, 60, 0, 99, 42, 60, 0, 100, 42, 60, 0, 101, 42, 60, 0, 102, 42, 60, 0, 103, 42, 60, 0, 104, 42, 60, 0, 105, 42, 60, 0, 106, 42, 60, 0, 107, 42, 60, 0, 108, 42, 60, 0, 109, 42, 60, 0, 110, 42, 60, 0, 111, 42, 60, 0, 112, 42, 60, 0, 113, 42, 60, 0, 114, 42, 60, 0, 115, 42, 60, 0, 119, 42, 60, 0, 120, 42, 60, 0, 121, 42, 60, 0, 122, 42, 60, 0, 123, 42, 60, 0, 124, 42, 60, 0, 125, 42, 60, 0, 126, 42, 60, 0, 127, 42, 60, 0, 128, 42, 60, 0, 129, 42, 60, 0, 130, 42, 60, 0, 131, 42, 60, 0, 132, 42, 60, 0, 133, 42, 60, 0, 134, 42, 60, 0, 135, 42, 60, 0, 136, 42, 60, 0, 137, 42, 60, 0, 138, 42, 60, 0, 139, 42, 60, 0, 140, 42, 60, 0, 141, 42, 60, 0, 142, 42, 60, 0, 143, 42, 60, 0, 144, 42, 60, 0, 145, 42, 60, 0, 146, 42, 60, 0, 147, 42, 60, 0, 148, 42, 60, 0, 149, 42, 60, 0, 150, 42, 60, 0, 151, 42, 60, 0, 152, 42, 60, 0, 153, 42, 60, 0, 154, 42, 60, 0, 155, 42, 60, 0, 156, 42, 60, 0, 157, 42, 60, 0, 158, 42, 60, 0, 159, 42, 60, 0, 160, 42, 60, 0, 161, 42, 60, 0, 162, 42, 60, 0, 163, 42, 60, 0, 164, 42, 60, 0, 165, 42, 60, 0, 166, 42, 60, 0, 167, 42, 60, 0, 168, 42, 60, 0, 169, 42, 60, 0, 170, 42, 60, 0, 171, 42, 60, 0, 172, 42, 60, 0, 173, 42, 60, 0, 174, 42, 60, 0, 175, 42, 60, 0, 176, 42, 60, 0, 177, 42, 60, 0, 178, 42, 60, 0, 179, 42, 60, 0, 180, 42, 60, 0, 181, 42, 60, 0, 182, 42, 60, 0, 183, 42, 60, 0, 184, 42, 60, 0, 185, 42, 60, 0, 186, 42, 60, 0, 187, 42, 60, 0, 188, 42, 60, 0, 189, 42, 60, 0, 190, 42, 60, 0, 191, 42, 60, 0, 192, 42, 60, 0, 193, 42, 60, 0, 194, 42, 60, 0, 195, 42, 60, 0, 196, 42, 60, 0, 197, 42, 60, 0, 198, 42, 60, 0, 199, 42, 60, 0, 200, 42, 60, 0, 201, 42, 60, 0, 202, 42, 60, 0, 203, 42, 60, 0, 204, 42, 60, 0, 205, 42, 60, 0, 206, 42, 60, 0, 207, 42, 60, 0, 208, 42, 60, 0, 209, 42, 60, 0, 210, 42, 60, 0, 211, 42, 60, 0, 212, 42, 60, 0, 213, 42, 60, 0, 214, 42, 60, 0, 215, 42, 60, 0, 216, 42, 60, 0, 217, 42, 60, 0, 218, 42, 60, 0, 219, 42, 60, 0, 221, 42, 60, 0, 222, 42, 60, 0, 223, 42, 60, 0, 224, 42, 60, 0, 225, 42, 60, 0, 226, 42, 60, 0, 227, 42, 60, 0, 228, 42, 60, 0, 229, 42, 60, 0, 230, 42, 60, 0, 231, 42, 60, 0, 232, 42, 60, 0, 233, 42, 60, 0, 234, 42, 60, 0, 235, 42, 60, 0, 236, 42, 60, 0, 237, 42, 60, 0, 238, 42, 60, 0, 239, 42, 60, 0, 240, 42, 60, 0, 241, 42, 60, 0, 242, 42, 60, 0, 243, 42, 60, 0, 244, 42, 60, 0, 245, 42, 60, 0, 246, 42, 60, 0, 247, 42, 60, 0, 248, 42, 60, 0, 249, 42, 60, 0, 250, 42, 60, 0, 251, 42, 60, 0, 252, 42, 60, 0, 253, 42, 60, 0, 254, 42, 60, 0, 255, 42, 60, 0, 0, 43, 60, 0, 1, 43, 60, 0, 2, 43, 60, 0, 3, 43, 60, 0, 4, 43, 60, 0, 5, 43, 60, 0, 6, 43, 60, 0, 7, 43, 60, 0, 8, 43, 60, 0, 9, 43, 60, 0, 10, 43, 60, 0, 11, 43, 60, 0, 12, 43, 60, 0, 13, 43, 60, 0, 14, 43, 60, 0, 15, 43, 60, 0, 16, 43, 60, 0, 17, 43, 60, 0, 18, 43, 60, 0, 19, 43, 60, 0, 20, 43, 60, 0, 21, 43, 60, 0, 22, 43, 60, 0, 23, 43, 60, 0, 24, 43, 60, 0, 25, 43, 60, 0, 26, 43, 60, 0, 27, 43, 60, 0, 28, 43, 60, 0, 29, 43, 60, 0, 30, 43, 60, 0, 31, 43, 60, 0, 32, 43, 60, 0, 33, 43, 60, 0, 34, 43, 60, 0, 35, 43, 60, 0, 36, 43, 60, 0, 37, 43, 60, 0, 38, 43, 60, 0, 39, 43, 60, 0, 40, 43, 60, 0, 41, 43, 60, 0, 42, 43, 60, 0, 43, 43, 60, 0, 44, 43, 60, 0, 45, 43, 60, 0, 46, 43, 60, 0, 47, 43, 60, 0, 48, 43, 60, 0, 49, 43, 60, 0, 50, 43, 60, 0, 51, 43, 60, 0, 52, 43, 60, 0, 53, 43, 60, 0, 54, 43, 60, 0, 55, 43, 60, 0, 56, 43, 60, 0, 57, 43, 60, 0, 58, 43, 60, 0, 59, 43, 60, 0, 60, 43, 60, 0, 61, 43, 60, 0, 62, 43, 60, 0, 63, 43, 60, 0, 64, 43, 60, 0, 65, 43, 60, 0, 66, 43, 60, 0, 67, 43, 60, 0, 68, 43, 60, 0, 69, 43, 60, 0, 70, 43, 60, 0, 71, 43, 60, 0, 72, 43, 60, 0, 73, 43, 60, 0, 74, 43, 60, 0, 75, 43, 60, 0, 76, 43, 60, 0, 77, 43, 60, 0, 78, 43, 60, 0, 79, 43, 60, 0, 80, 43, 60, 0, 81, 43, 60, 0, 82, 43, 60, 0, 83, 43, 60, 0, 84, 43, 60, 0, 85, 43, 60, 0, 86, 43, 60, 0, 87, 43, 60, 0, 88, 43, 60, 0, 89, 43, 60, 0, 90, 43, 60, 0, 91, 43, 60, 0, 92, 43, 60, 0, 93, 43, 60, 0, 94, 43, 60, 0, 95, 43, 60, 0, 96, 43, 60, 0, 97, 43, 60, 0, 98, 43, 60, 0, 99, 43, 60, 0, 100, 43, 60, 0, 101, 43, 60, 0, 102, 43, 60, 0, 103, 43, 60, 0, 104, 43, 60, 0, 105, 43, 60, 0, 106, 43, 60, 0, 107, 43, 60, 0, 108, 43, 60, 0, 109, 43, 60, 0, 110, 43, 60, 0, 111, 43, 60, 0, 112, 43, 60, 0, 113, 43, 60, 0, 114, 43, 60, 0, 115, 43, 60, 0, 118, 43, 60, 0, 119, 43, 60, 0, 120, 43, 60, 0, 121, 43, 60, 0, 122, 43, 60, 0, 123, 43, 60, 0, 124, 43, 60, 0, 125, 43, 60, 0, 126, 43, 60, 0, 127, 43, 60, 0, 128, 43, 60, 0, 129, 43, 60, 0, 130, 43, 60, 0, 131, 43, 60, 0, 132, 43, 60, 0, 133, 43, 60, 0, 134, 43, 60, 0, 135, 43, 60, 0, 136, 43, 60, 0, 137, 43, 60, 0, 138, 43, 60, 0, 139, 43, 60, 0, 140, 43, 60, 0, 141, 43, 60, 0, 142, 43, 60, 0, 143, 43, 60, 0, 144, 43, 60, 0, 145, 43, 60, 0, 146, 43, 60, 0, 147, 43, 60, 0, 148, 43, 60, 0, 149, 43, 60, 0, 151, 43, 60, 0, 152, 43, 60, 0, 153, 43, 60, 0, 154, 43, 60, 0, 155, 43, 60, 0, 156, 43, 60, 0, 157, 43, 60, 0, 158, 43, 60, 0, 159, 43, 60, 0, 160, 43, 60, 0, 161, 43, 60, 0, 162, 43, 60, 0, 163, 43, 60, 0, 164, 43, 60, 0, 165, 43, 60, 0, 166, 43, 60, 0, 167, 43, 60, 0, 168, 43, 60, 0, 169, 43, 60, 0, 170, 43, 60, 0, 171, 43, 60, 0, 172, 43, 60, 0, 173, 43, 60, 0, 174, 43, 60, 0, 175, 43, 60, 0, 176, 43, 60, 0, 177, 43, 60, 0, 178, 43, 60, 0, 179, 43, 60, 0, 180, 43, 60, 0, 181, 43, 60, 0, 182, 43, 60, 0, 183, 43, 60, 0, 184, 43, 60, 0, 185, 43, 60, 0, 186, 43, 60, 0, 187, 43, 60, 0, 188, 43, 60, 0, 189, 43, 60, 0, 190, 43, 60, 0, 191, 43, 60, 0, 192, 43, 60, 0, 193, 43, 60, 0, 194, 43, 60, 0, 195, 43, 60, 0, 196, 43, 60, 0, 197, 43, 60, 0, 198, 43, 60, 0, 199, 43, 60, 0, 200, 43, 60, 0, 201, 43, 60, 0, 202, 43, 60, 0, 203, 43, 60, 0, 204, 43, 60, 0, 205, 43, 60, 0, 206, 43, 60, 0, 207, 43, 60, 0, 208, 43, 60, 0, 209, 43, 60, 0, 210, 43, 60, 0, 211, 43, 60, 0, 212, 43, 60, 0, 213, 43, 60, 0, 214, 43, 60, 0, 215, 43, 60, 0, 216, 43, 60, 0, 217, 43, 60, 0, 218, 43, 60, 0, 219, 43, 60, 0, 220, 43, 60, 0, 221, 43, 60, 0, 222, 43, 60, 0, 223, 43, 60, 0, 224, 43, 60, 0, 225, 43, 60, 0, 226, 43, 60, 0, 227, 43, 60, 0, 228, 43, 60, 0, 229, 43, 60, 0, 230, 43, 60, 0, 231, 43, 60, 0, 232, 43, 60, 0, 233, 43, 60, 0, 234, 43, 60, 0, 235, 43, 60, 0, 236, 43, 60, 0, 237, 43, 60, 0, 238, 43, 60, 0, 239, 43, 60, 0, 240, 43, 60, 0, 241, 43, 60, 0, 242, 43, 60, 0, 243, 43, 60, 0, 244, 43, 60, 0, 245, 43, 60, 0, 246, 43, 60, 0, 247, 43, 60, 0, 248, 43, 60, 0, 249, 43, 60, 0, 250, 43, 60, 0, 251, 43, 60, 0, 252, 43, 60, 0, 253, 43, 60, 0, 254, 43, 60, 0, 255, 43, 60, 0, 229, 44, 60, 0, 230, 44, 60, 0, 231, 44, 60, 0, 232, 44, 60, 0, 233, 44, 60, 0, 234, 44, 60, 0, 0, 40, 60, 0, 1, 40, 60, 0, 2, 40, 60, 0, 3, 40, 60, 0, 4, 40, 60, 0, 5, 40, 60, 0, 6, 40, 60, 0, 7, 40, 60, 0, 8, 40, 60, 0, 9, 40, 60, 0, 10, 40, 60, 0, 11, 40, 60, 0, 12, 40, 60, 0, 13, 40, 60, 0, 14, 40, 60, 0, 15, 40, 60, 0, 16, 40, 60, 0, 17, 40, 60, 0, 18, 40, 60, 0, 19, 40, 60, 0, 20, 40, 60, 0, 21, 40, 60, 0, 22, 40, 60, 0, 23, 40, 60, 0, 24, 40, 60, 0, 25, 40, 60, 0, 26, 40, 60, 0, 27, 40, 60, 0, 28, 40, 60, 0, 29, 40, 60, 0, 30, 40, 60, 0, 31, 40, 60, 0, 32, 40, 60, 0, 33, 40, 60, 0, 34, 40, 60, 0, 35, 40, 60, 0, 36, 40, 60, 0, 37, 40, 60, 0, 38, 40, 60, 0, 39, 40, 60, 0, 40, 40, 60, 0, 41, 40, 60, 0, 42, 40, 60, 0, 43, 40, 60, 0, 44, 40, 60, 0, 45, 40, 60, 0, 46, 40, 60, 0, 47, 40, 60, 0, 48, 40, 60, 0, 49, 40, 60, 0, 50, 40, 60, 0, 51, 40, 60, 0, 52, 40, 60, 0, 53, 40, 60, 0, 54, 40, 60, 0, 55, 40, 60, 0, 56, 40, 60, 0, 57, 40, 60, 0, 58, 40, 60, 0, 59, 40, 60, 0, 60, 40, 60, 0, 61, 40, 60, 0, 62, 40, 60, 0, 63, 40, 60, 0, 64, 40, 60, 0, 65, 40, 60, 0, 66, 40, 60, 0, 67, 40, 60, 0, 68, 40, 60, 0, 69, 40, 60, 0, 70, 40, 60, 0, 71, 40, 60, 0, 72, 40, 60, 0, 73, 40, 60, 0, 74, 40, 60, 0, 75, 40, 60, 0, 76, 40, 60, 0, 77, 40, 60, 0, 78, 40, 60, 0, 79, 40, 60, 0, 80, 40, 60, 0, 81, 40, 60, 0, 82, 40, 60, 0, 83, 40, 60, 0, 84, 40, 60, 0, 85, 40, 60, 0, 86, 40, 60, 0, 87, 40, 60, 0, 88, 40, 60, 0, 89, 40, 60, 0, 90, 40, 60, 0, 91, 40, 60, 0, 92, 40, 60, 0, 93, 40, 60, 0, 94, 40, 60, 0, 95, 40, 60, 0, 96, 40, 60, 0, 97, 40, 60, 0, 98, 40, 60, 0, 99, 40, 60, 0, 100, 40, 60, 0, 101, 40, 60, 0, 102, 40, 60, 0, 103, 40, 60, 0, 104, 40, 60, 0, 105, 40, 60, 0, 106, 40, 60, 0, 107, 40, 60, 0, 108, 40, 60, 0, 109, 40, 60, 0, 110, 40, 60, 0, 111, 40, 60, 0, 112, 40, 60, 0, 113, 40, 60, 0, 114, 40, 60, 0, 115, 40, 60, 0, 116, 40, 60, 0, 117, 40, 60, 0, 118, 40, 60, 0, 119, 40, 60, 0, 120, 40, 60, 0, 121, 40, 60, 0, 122, 40, 60, 0, 123, 40, 60, 0, 124, 40, 60, 0, 125, 40, 60, 0, 126, 40, 60, 0, 127, 40, 60, 0, 128, 40, 60, 0, 129, 40, 60, 0, 130, 40, 60, 0, 131, 40, 60, 0, 132, 40, 60, 0, 133, 40, 60, 0, 134, 40, 60, 0, 135, 40, 60, 0, 136, 40, 60, 0, 137, 40, 60, 0, 138, 40, 60, 0, 139, 40, 60, 0, 140, 40, 60, 0, 141, 40, 60, 0, 142, 40, 60, 0, 143, 40, 60, 0, 144, 40, 60, 0, 145, 40, 60, 0, 146, 40, 60, 0, 147, 40, 60, 0, 148, 40, 60, 0, 149, 40, 60, 0, 150, 40, 60, 0, 151, 40, 60, 0, 152, 40, 60, 0, 153, 40, 60, 0, 154, 40, 60, 0, 155, 40, 60, 0, 156, 40, 60, 0, 157, 40, 60, 0, 158, 40, 60, 0, 159, 40, 60, 0, 160, 40, 60, 0, 161, 40, 60, 0, 162, 40, 60, 0, 163, 40, 60, 0, 164, 40, 60, 0, 165, 40, 60, 0, 166, 40, 60, 0, 167, 40, 60, 0, 168, 40, 60, 0, 169, 40, 60, 0, 170, 40, 60, 0, 171, 40, 60, 0, 172, 40, 60, 0, 173, 40, 60, 0, 174, 40, 60, 0, 175, 40, 60, 0, 176, 40, 60, 0, 177, 40, 60, 0, 178, 40, 60, 0, 179, 40, 60, 0, 180, 40, 60, 0, 181, 40, 60, 0, 182, 40, 60, 0, 183, 40, 60, 0, 184, 40, 60, 0, 185, 40, 60, 0, 186, 40, 60, 0, 187, 40, 60, 0, 188, 40, 60, 0, 189, 40, 60, 0, 190, 40, 60, 0, 191, 40, 60, 0, 192, 40, 60, 0, 193, 40, 60, 0, 194, 40, 60, 0, 195, 40, 60, 0, 196, 40, 60, 0, 197, 40, 60, 0, 198, 40, 60, 0, 199, 40, 60, 0, 200, 40, 60, 0, 201, 40, 60, 0, 202, 40, 60, 0, 203, 40, 60, 0, 204, 40, 60, 0, 205, 40, 60, 0, 206, 40, 60, 0, 207, 40, 60, 0, 208, 40, 60, 0, 209, 40, 60, 0, 210, 40, 60, 0, 211, 40, 60, 0, 212, 40, 60, 0, 213, 40, 60, 0, 214, 40, 60, 0, 215, 40, 60, 0, 216, 40, 60, 0, 217, 40, 60, 0, 218, 40, 60, 0, 219, 40, 60, 0, 220, 40, 60, 0, 221, 40, 60, 0, 222, 40, 60, 0, 223, 40, 60, 0, 224, 40, 60, 0, 225, 40, 60, 0, 226, 40, 60, 0, 227, 40, 60, 0, 228, 40, 60, 0, 229, 40, 60, 0, 230, 40, 60, 0, 231, 40, 60, 0, 232, 40, 60, 0, 233, 40, 60, 0, 234, 40, 60, 0, 235, 40, 60, 0, 236, 40, 60, 0, 237, 40, 60, 0, 238, 40, 60, 0, 239, 40, 60, 0, 240, 40, 60, 0, 241, 40, 60, 0, 242, 40, 60, 0, 243, 40, 60, 0, 244, 40, 60, 0, 245, 40, 60, 0, 246, 40, 60, 0, 247, 40, 60, 0, 248, 40, 60, 0, 249, 40, 60, 0, 250, 40, 60, 0, 251, 40, 60, 0, 252, 40, 60, 0, 253, 40, 60, 0, 254, 40, 60, 0, 255, 40, 60, 0, 138, 38, 60, 0, 139, 38, 60, 0, 140, 38, 60, 0, 141, 38, 60, 0, 142, 38, 60, 0, 143, 38, 60, 0, 48, 38, 60, 0, 49, 38, 60, 0, 50, 38, 60, 0, 51, 38, 60, 0, 52, 38, 60, 0, 53, 38, 60, 0, 54, 38, 60, 0, 55, 38, 60, 0, 192, 77, 60, 0, 193, 77, 60, 0, 194, 77, 60, 0, 195, 77, 60, 0, 196, 77, 60, 0, 197, 77, 60, 0, 198, 77, 60, 0, 199, 77, 60, 0, 200, 77, 60, 0, 201, 77, 60, 0, 202, 77, 60, 0, 203, 77, 60, 0, 204, 77, 60, 0, 205, 77, 60, 0, 206, 77, 60, 0, 207, 77, 60, 0, 208, 77, 60, 0, 209, 77, 60, 0, 210, 77, 60, 0, 211, 77, 60, 0, 212, 77, 60, 0, 213, 77, 60, 0, 214, 77, 60, 0, 215, 77, 60, 0, 216, 77, 60, 0, 217, 77, 60, 0, 218, 77, 60, 0, 219, 77, 60, 0, 220, 77, 60, 0, 221, 77, 60, 0, 222, 77, 60, 0, 223, 77, 60, 0, 224, 77, 60, 0, 225, 77, 60, 0, 226, 77, 60, 0, 227, 77, 60, 0, 228, 77, 60, 0, 229, 77, 60, 0, 230, 77, 60, 0, 231, 77, 60, 0, 232, 77, 60, 0, 233, 77, 60, 0, 234, 77, 60, 0, 235, 77, 60, 0, 236, 77, 60, 0, 237, 77, 60, 0, 238, 77, 60, 0, 239, 77, 60, 0, 240, 77, 60, 0, 241, 77, 60, 0, 242, 77, 60, 0, 243, 77, 60, 0, 244, 77, 60, 0, 245, 77, 60, 0, 246, 77, 60, 0, 247, 77, 60, 0, 248, 77, 60, 0, 249, 77, 60, 0, 250, 77, 60, 0, 251, 77, 60, 0, 252, 77, 60, 0, 253, 77, 60, 0, 254, 77, 60, 0, 255, 77, 60, 0, 52, 216, 0, 223, 60, 0, 52, 216, 1, 223, 60, 0, 52, 216, 2, 223, 60, 0, 52, 216, 3, 223, 60, 0, 52, 216, 4, 223, 60, 0, 52, 216, 5, 223, 60, 0, 52, 216, 6, 223, 60, 0, 52, 216, 7, 223, 60, 0, 52, 216, 8, 223, 60, 0, 52, 216, 9, 223, 60, 0, 52, 216, 10, 223, 60, 0, 52, 216, 11, 223, 60, 0, 52, 216, 12, 223, 60, 0, 52, 216, 13, 223, 60, 0, 52, 216, 14, 223, 60, 0, 52, 216, 15, 223, 60, 0, 52, 216, 16, 223, 60, 0, 52, 216, 17, 223, 60, 0, 52, 216, 18, 223, 60, 0, 52, 216, 19, 223, 60, 0, 52, 216, 20, 223, 60, 0, 52, 216, 21, 223, 60, 0, 52, 216, 22, 223, 60, 0, 52, 216, 23, 223, 60, 0, 52, 216, 24, 223, 60, 0, 52, 216, 25, 223, 60, 0, 52, 216, 26, 223, 60, 0, 52, 216, 27, 223, 60, 0, 52, 216, 28, 223, 60, 0, 52, 216, 29, 223, 60, 0, 52, 216, 30, 223, 60, 0, 52, 216, 31, 223, 60, 0, 52, 216, 32, 223, 60, 0, 52, 216, 33, 223, 60, 0, 52, 216, 34, 223, 60, 0, 52, 216, 35, 223, 60, 0, 52, 216, 36, 223, 60, 0, 52, 216, 37, 223, 60, 0, 52, 216, 38, 223, 60, 0, 52, 216, 39, 223, 60, 0, 52, 216, 40, 223, 60, 0, 52, 216, 41, 223, 60, 0, 52, 216, 42, 223, 60, 0, 52, 216, 43, 223, 60, 0, 52, 216, 44, 223, 60, 0, 52, 216, 45, 223, 60, 0, 52, 216, 46, 223, 60, 0, 52, 216, 47, 223, 60, 0, 52, 216, 48, 223, 60, 0, 52, 216, 49, 223, 60, 0, 52, 216, 50, 223, 60, 0, 52, 216, 51, 223, 60, 0, 52, 216, 52, 223, 60, 0, 52, 216, 53, 223, 60, 0, 52, 216, 54, 223, 60, 0, 52, 216, 55, 223, 60, 0, 52, 216, 56, 223, 60, 0, 52, 216, 57, 223, 60, 0, 52, 216, 58, 223, 60, 0, 52, 216, 59, 223, 60, 0, 52, 216, 60, 223, 60, 0, 52, 216, 61, 223, 60, 0, 52, 216, 62, 223, 60, 0, 52, 216, 63, 223, 60, 0, 52, 216, 64, 223, 60, 0, 52, 216, 65, 223, 60, 0, 52, 216, 66, 223, 60, 0, 52, 216, 67, 223, 60, 0, 52, 216, 68, 223, 60, 0, 52, 216, 69, 223, 60, 0, 52, 216, 70, 223, 60, 0, 52, 216, 71, 223, 60, 0, 52, 216, 72, 223, 60, 0, 52, 216, 73, 223, 60, 0, 52, 216, 74, 223, 60, 0, 52, 216, 75, 223, 60, 0, 52, 216, 76, 223, 60, 0, 52, 216, 77, 223, 60, 0, 52, 216, 78, 223, 60, 0, 52, 216, 79, 223, 60, 0, 52, 216, 80, 223, 60, 0, 52, 216, 81, 223, 60, 0, 52, 216, 82, 223, 60, 0, 52, 216, 83, 223, 60, 0, 52, 216, 84, 223, 60, 0, 52, 216, 85, 223, 60, 0, 52, 216, 86, 223, 60, 0, 144, 164, 60, 0, 145, 164, 60, 0, 146, 164, 60, 0, 147, 164, 60, 0, 148, 164, 60, 0, 149, 164, 60, 0, 150, 164, 60, 0, 151, 164, 60, 0, 152, 164, 60, 0, 153, 164, 60, 0, 154, 164, 60, 0, 155, 164, 60, 0, 156, 164, 60, 0, 157, 164, 60, 0, 158, 164, 60, 0, 159, 164, 60, 0, 160, 164, 60, 0, 161, 164, 60, 0, 162, 164, 60, 0, 163, 164, 60, 0, 164, 164, 60, 0, 165, 164, 60, 0, 166, 164, 60, 0, 167, 164, 60, 0, 168, 164, 60, 0, 169, 164, 60, 0, 170, 164, 60, 0, 171, 164, 60, 0, 172, 164, 60, 0, 173, 164, 60, 0, 174, 164, 60, 0, 175, 164, 60, 0, 176, 164, 60, 0, 177, 164, 60, 0, 178, 164, 60, 0, 179, 164, 60, 0, 180, 164, 60, 0, 181, 164, 60, 0, 182, 164, 60, 0, 183, 164, 60, 0, 184, 164, 60, 0, 185, 164, 60, 0, 186, 164, 60, 0, 187, 164, 60, 0, 188, 164, 60, 0, 189, 164, 60, 0, 190, 164, 60, 0, 191, 164, 60, 0, 192, 164, 60, 0, 193, 164, 60, 0, 194, 164, 60, 0, 195, 164, 60, 0, 196, 164, 60, 0, 197, 164, 60, 0, 198, 164, 60, 0, 0, 216, 55, 221, 60, 0, 0, 216, 56, 221, 60, 0, 0, 216, 57, 221, 60, 0, 0, 216, 58, 221, 60, 0, 0, 216, 59, 221, 60, 0, 0, 216, 60, 221, 60, 0, 0, 216, 61, 221, 60, 0, 0, 216, 62, 221, 60, 0, 0, 216, 63, 221, 60, 0, 0, 216, 121, 221, 60, 0, 0, 216, 122, 221, 60, 0, 0, 216, 123, 221, 60, 0, 0, 216, 124, 221, 60, 0, 0, 216, 125, 221, 60, 0, 0, 216, 126, 221, 60, 0, 0, 216, 127, 221, 60, 0, 0, 216, 128, 221, 60, 0, 0, 216, 129, 221, 60, 0, 0, 216, 130, 221, 60, 0, 0, 216, 131, 221, 60, 0, 0, 216, 132, 221, 60, 0, 0, 216, 133, 221, 60, 0, 0, 216, 134, 221, 60, 0, 0, 216, 135, 221, 60, 0, 0, 216, 136, 221, 60, 0, 0, 216, 137, 221, 60, 0, 0, 216, 140, 221, 60, 0, 0, 216, 141, 221, 60, 0, 0, 216, 142, 221, 60, 0, 0, 216, 144, 221, 60, 0, 0, 216, 145, 221, 60, 0, 0, 216, 146, 221, 60, 0, 0, 216, 147, 221, 60, 0, 0, 216, 148, 221, 60, 0, 0, 216, 149, 221, 60, 0, 0, 216, 150, 221, 60, 0, 0, 216, 151, 221, 60, 0, 0, 216, 152, 221, 60, 0, 0, 216, 153, 221, 60, 0, 0, 216, 154, 221, 60, 0, 0, 216, 155, 221, 60, 0, 0, 216, 156, 221, 60, 0, 0, 216, 160, 221, 60, 0, 0, 216, 208, 221, 60, 0, 0, 216, 209, 221, 60, 0, 0, 216, 210, 221, 60, 0, 0, 216, 211, 221, 60, 0, 0, 216, 212, 221, 60, 0, 0, 216, 213, 221, 60, 0, 0, 216, 214, 221, 60, 0, 0, 216, 215, 221, 60, 0, 0, 216, 216, 221, 60, 0, 0, 216, 217, 221, 60, 0, 0, 216, 218, 221, 60, 0, 0, 216, 219, 221, 60, 0, 0, 216, 220, 221, 60, 0, 0, 216, 221, 221, 60, 0, 0, 216, 222, 221, 60, 0, 0, 216, 223, 221, 60, 0, 0, 216, 224, 221, 60, 0, 0, 216, 225, 221, 60, 0, 0, 216, 226, 221, 60, 0, 0, 216, 227, 221, 60, 0, 0, 216, 228, 221, 60, 0, 0, 216, 229, 221, 60, 0, 0, 216, 230, 221, 60, 0, 0, 216, 231, 221, 60, 0, 0, 216, 232, 221, 60, 0, 0, 216, 233, 221, 60, 0, 0, 216, 234, 221, 60, 0, 0, 216, 235, 221, 60, 0, 0, 216, 236, 221, 60, 0, 0, 216, 237, 221, 60, 0, 0, 216, 238, 221, 60, 0, 0, 216, 239, 221, 60, 0, 0, 216, 240, 221, 60, 0, 0, 216, 241, 221, 60, 0, 0, 216, 242, 221, 60, 0, 0, 216, 243, 221, 60, 0, 0, 216, 244, 221, 60, 0, 0, 216, 245, 221, 60, 0, 0, 216, 246, 221, 60, 0, 0, 216, 247, 221, 60, 0, 0, 216, 248, 221, 60, 0, 0, 216, 249, 221, 60, 0, 0, 216, 250, 221, 60, 0, 0, 216, 251, 221, 60, 0, 0, 216, 252, 221, 60, 0, 14, 32, 2, 216, 119, 220, 14, 32, 60, 0, 14, 32, 2, 216, 120, 220, 14, 32, 60, 0, 26, 216, 60, 223, 60, 0, 26, 216, 61, 223, 60, 0, 26, 216, 62, 223, 60, 0, 26, 216, 63, 223, 60, 0, 26, 216, 69, 223, 60, 0, 56, 216, 79, 221, 60, 0, 52, 216, 0, 220, 60, 0, 52, 216, 1, 220, 60, 0, 52, 216, 2, 220, 60, 0, 52, 216, 3, 220, 60, 0, 52, 216, 4, 220, 60, 0, 52, 216, 5, 220, 60, 0, 52, 216, 6, 220, 60, 0, 52, 216, 7, 220, 60, 0, 52, 216, 8, 220, 60, 0, 52, 216, 9, 220, 60, 0, 52, 216, 10, 220, 60, 0, 52, 216, 11, 220, 60, 0, 52, 216, 12, 220, 60, 0, 52, 216, 13, 220, 60, 0, 52, 216, 14, 220, 60, 0, 52, 216, 15, 220, 60, 0, 52, 216, 16, 220, 60, 0, 52, 216, 17, 220, 60, 0, 52, 216, 18, 220, 60, 0, 52, 216, 19, 220, 60, 0, 52, 216, 20, 220, 60, 0, 52, 216, 21, 220, 60, 0, 52, 216, 22, 220, 60, 0, 52, 216, 23, 220, 60, 0, 52, 216, 24, 220, 60, 0, 52, 216, 25, 220, 60, 0, 52, 216, 26, 220, 60, 0, 52, 216, 27, 220, 60, 0, 52, 216, 28, 220, 60, 0, 52, 216, 29, 220, 60, 0, 52, 216, 30, 220, 60, 0, 52, 216, 31, 220, 60, 0, 52, 216, 32, 220, 60, 0, 52, 216, 33, 220, 60, 0, 52, 216, 34, 220, 60, 0, 52, 216, 35, 220, 60, 0, 52, 216, 36, 220, 60, 0, 52, 216, 37, 220, 60, 0, 52, 216, 38, 220, 60, 0, 52, 216, 39, 220, 60, 0, 52, 216, 40, 220, 60, 0, 52, 216, 41, 220, 60, 0, 52, 216, 42, 220, 60, 0, 52, 216, 43, 220, 60, 0, 52, 216, 44, 220, 60, 0, 52, 216, 45, 220, 60, 0, 52, 216, 46, 220, 60, 0, 52, 216, 47, 220, 60, 0, 52, 216, 48, 220, 60, 0, 52, 216, 49, 220, 60, 0, 52, 216, 50, 220, 60, 0, 52, 216, 51, 220, 60, 0, 52, 216, 52, 220, 60, 0, 52, 216, 53, 220, 60, 0, 52, 216, 54, 220, 60, 0, 52, 216, 55, 220, 60, 0, 52, 216, 56, 220, 60, 0, 52, 216, 57, 220, 60, 0, 52, 216, 58, 220, 60, 0, 52, 216, 59, 220, 60, 0, 52, 216, 60, 220, 60, 0, 52, 216, 61, 220, 60, 0, 52, 216, 62, 220, 60, 0, 52, 216, 63, 220, 60, 0, 52, 216, 64, 220, 60, 0, 52, 216, 65, 220, 60, 0, 52, 216, 66, 220, 60, 0, 52, 216, 67, 220, 60, 0, 52, 216, 68, 220, 60, 0, 52, 216, 69, 220, 60, 0, 52, 216, 70, 220, 60, 0, 52, 216, 71, 220, 60, 0, 52, 216, 72, 220, 60, 0, 52, 216, 73, 220, 60, 0, 52, 216, 74, 220, 60, 0, 52, 216, 75, 220, 60, 0, 52, 216, 76, 220, 60, 0, 52, 216, 77, 220, 60, 0, 52, 216, 78, 220, 60, 0, 52, 216, 79, 220, 60, 0, 52, 216, 80, 220, 60, 0, 52, 216, 81, 220, 60, 0, 52, 216, 82, 220, 60, 0, 52, 216, 83, 220, 60, 0, 52, 216, 84, 220, 60, 0, 52, 216, 85, 220, 60, 0, 52, 216, 86, 220, 60, 0, 52, 216, 87, 220, 60, 0, 52, 216, 88, 220, 60, 0, 52, 216, 89, 220, 60, 0, 52, 216, 90, 220, 60, 0, 52, 216, 91, 220, 60, 0, 52, 216, 92, 220, 60, 0, 52, 216, 93, 220, 60, 0, 52, 216, 94, 220, 60, 0, 52, 216, 95, 220, 60, 0, 52, 216, 96, 220, 60, 0, 52, 216, 97, 220, 60, 0, 52, 216, 98, 220, 60, 0, 52, 216, 99, 220, 60, 0, 52, 216, 100, 220, 60, 0, 52, 216, 101, 220, 60, 0, 52, 216, 102, 220, 60, 0, 52, 216, 103, 220, 60, 0, 52, 216, 104, 220, 60, 0, 52, 216, 105, 220, 60, 0, 52, 216, 106, 220, 60, 0, 52, 216, 107, 220, 60, 0, 52, 216, 108, 220, 60, 0, 52, 216, 109, 220, 60, 0, 52, 216, 110, 220, 60, 0, 52, 216, 111, 220, 60, 0, 52, 216, 112, 220, 60, 0, 52, 216, 113, 220, 60, 0, 52, 216, 114, 220, 60, 0, 52, 216, 115, 220, 60, 0, 52, 216, 116, 220, 60, 0, 52, 216, 117, 220, 60, 0, 52, 216, 118, 220, 60, 0, 52, 216, 119, 220, 60, 0, 52, 216, 120, 220, 60, 0, 52, 216, 121, 220, 60, 0, 52, 216, 122, 220, 60, 0, 52, 216, 123, 220, 60, 0, 52, 216, 124, 220, 60, 0, 52, 216, 125, 220, 60, 0, 52, 216, 126, 220, 60, 0, 52, 216, 127, 220, 60, 0, 52, 216, 128, 220, 60, 0, 52, 216, 129, 220, 60, 0, 52, 216, 130, 220, 60, 0, 52, 216, 131, 220, 60, 0, 52, 216, 132, 220, 60, 0, 52, 216, 133, 220, 60, 0, 52, 216, 134, 220, 60, 0, 52, 216, 135, 220, 60, 0, 52, 216, 136, 220, 60, 0, 52, 216, 137, 220, 60, 0, 52, 216, 138, 220, 60, 0, 52, 216, 139, 220, 60, 0, 52, 216, 140, 220, 60, 0, 52, 216, 141, 220, 60, 0, 52, 216, 142, 220, 60, 0, 52, 216, 143, 220, 60, 0, 52, 216, 144, 220, 60, 0, 52, 216, 145, 220, 60, 0, 52, 216, 146, 220, 60, 0, 52, 216, 147, 220, 60, 0, 52, 216, 148, 220, 60, 0, 52, 216, 149, 220, 60, 0, 52, 216, 150, 220, 60, 0, 52, 216, 151, 220, 60, 0, 52, 216, 152, 220, 60, 0, 52, 216, 153, 220, 60, 0, 52, 216, 154, 220, 60, 0, 52, 216, 155, 220, 60, 0, 52, 216, 156, 220, 60, 0, 52, 216, 157, 220, 60, 0, 52, 216, 158, 220, 60, 0, 52, 216, 159, 220, 60, 0, 52, 216, 160, 220, 60, 0, 52, 216, 161, 220, 60, 0, 52, 216, 162, 220, 60, 0, 52, 216, 163, 220, 60, 0, 52, 216, 164, 220, 60, 0, 52, 216, 165, 220, 60, 0, 52, 216, 166, 220, 60, 0, 52, 216, 167, 220, 60, 0, 52, 216, 168, 220, 60, 0, 52, 216, 169, 220, 60, 0, 52, 216, 170, 220, 60, 0, 52, 216, 171, 220, 60, 0, 52, 216, 172, 220, 60, 0, 52, 216, 173, 220, 60, 0, 52, 216, 174, 220, 60, 0, 52, 216, 175, 220, 60, 0, 52, 216, 176, 220, 60, 0, 52, 216, 177, 220, 60, 0, 52, 216, 178, 220, 60, 0, 52, 216, 179, 220, 60, 0, 52, 216, 180, 220, 60, 0, 52, 216, 181, 220, 60, 0, 52, 216, 182, 220, 60, 0, 52, 216, 183, 220, 60, 0, 52, 216, 184, 220, 60, 0, 52, 216, 185, 220, 60, 0, 52, 216, 186, 220, 60, 0, 52, 216, 187, 220, 60, 0, 52, 216, 188, 220, 60, 0, 52, 216, 189, 220, 60, 0, 52, 216, 190, 220, 60, 0, 52, 216, 191, 220, 60, 0, 52, 216, 192, 220, 60, 0, 52, 216, 193, 220, 60, 0, 52, 216, 194, 220, 60, 0, 52, 216, 195, 220, 60, 0, 52, 216, 196, 220, 60, 0, 52, 216, 197, 220, 60, 0, 52, 216, 198, 220, 60, 0, 52, 216, 199, 220, 60, 0, 52, 216, 200, 220, 60, 0, 52, 216, 201, 220, 60, 0, 52, 216, 202, 220, 60, 0, 52, 216, 203, 220, 60, 0, 52, 216, 204, 220, 60, 0, 52, 216, 205, 220, 60, 0, 52, 216, 206, 220, 60, 0, 52, 216, 207, 220, 60, 0, 52, 216, 208, 220, 60, 0, 52, 216, 209, 220, 60, 0, 52, 216, 210, 220, 60, 0, 52, 216, 211, 220, 60, 0, 52, 216, 212, 220, 60, 0, 52, 216, 213, 220, 60, 0, 52, 216, 214, 220, 60, 0, 52, 216, 215, 220, 60, 0, 52, 216, 216, 220, 60, 0, 52, 216, 217, 220, 60, 0, 52, 216, 218, 220, 60, 0, 52, 216, 219, 220, 60, 0, 52, 216, 220, 220, 60, 0, 52, 216, 221, 220, 60, 0, 52, 216, 222, 220, 60, 0, 52, 216, 223, 220, 60, 0, 52, 216, 224, 220, 60, 0, 52, 216, 225, 220, 60, 0, 52, 216, 226, 220, 60, 0, 52, 216, 227, 220, 60, 0, 52, 216, 228, 220, 60, 0, 52, 216, 229, 220, 60, 0, 52, 216, 230, 220, 60, 0, 52, 216, 231, 220, 60, 0, 52, 216, 232, 220, 60, 0, 52, 216, 233, 220, 60, 0, 52, 216, 234, 220, 60, 0, 52, 216, 235, 220, 60, 0, 52, 216, 236, 220, 60, 0, 52, 216, 237, 220, 60, 0, 52, 216, 238, 220, 60, 0, 52, 216, 239, 220, 60, 0, 52, 216, 240, 220, 60, 0, 52, 216, 241, 220, 60, 0, 52, 216, 242, 220, 60, 0, 52, 216, 243, 220, 60, 0, 52, 216, 244, 220, 60, 0, 52, 216, 245, 220, 60, 0, 52, 216, 0, 221, 60, 0, 52, 216, 1, 221, 60, 0, 52, 216, 2, 221, 60, 0, 52, 216, 3, 221, 60, 0, 52, 216, 4, 221, 60, 0, 52, 216, 5, 221, 60, 0, 52, 216, 6, 221, 60, 0, 52, 216, 7, 221, 60, 0, 52, 216, 8, 221, 60, 0, 52, 216, 9, 221, 60, 0, 52, 216, 10, 221, 60, 0, 52, 216, 11, 221, 60, 0, 52, 216, 12, 221, 60, 0, 52, 216, 13, 221, 60, 0, 52, 216, 14, 221, 60, 0, 52, 216, 15, 221, 60, 0, 52, 216, 16, 221, 60, 0, 52, 216, 17, 221, 60, 0, 52, 216, 18, 221, 60, 0, 52, 216, 19, 221, 60, 0, 52, 216, 20, 221, 60, 0, 52, 216, 21, 221, 60, 0, 52, 216, 22, 221, 60, 0, 52, 216, 23, 221, 60, 0, 52, 216, 24, 221, 60, 0, 52, 216, 25, 221, 60, 0, 52, 216, 26, 221, 60, 0, 52, 216, 27, 221, 60, 0, 52, 216, 28, 221, 60, 0, 52, 216, 29, 221, 60, 0, 52, 216, 30, 221, 60, 0, 52, 216, 31, 221, 60, 0, 52, 216, 32, 221, 60, 0, 52, 216, 33, 221, 60, 0, 52, 216, 34, 221, 60, 0, 52, 216, 35, 221, 60, 0, 52, 216, 36, 221, 60, 0, 52, 216, 37, 221, 60, 0, 52, 216, 38, 221, 60, 0, 109, 38, 60, 0, 110, 38, 60, 0, 111, 38, 60, 0, 52, 216, 42, 221, 60, 0, 52, 216, 43, 221, 60, 0, 52, 216, 44, 221, 60, 0, 52, 216, 45, 221, 60, 0, 52, 216, 46, 221, 60, 0, 52, 216, 47, 221, 60, 0, 52, 216, 48, 221, 60, 0, 52, 216, 49, 221, 60, 0, 52, 216, 50, 221, 60, 0, 52, 216, 51, 221, 60, 0, 52, 216, 52, 221, 60, 0, 52, 216, 53, 221, 60, 0, 52, 216, 54, 221, 60, 0, 52, 216, 55, 221, 60, 0, 52, 216, 56, 221, 60, 0, 52, 216, 57, 221, 60, 0, 52, 216, 41, 221, 60, 0, 52, 216, 58, 221, 60, 0, 52, 216, 59, 221, 60, 0, 52, 216, 60, 221, 60, 0, 52, 216, 61, 221, 60, 0, 52, 216, 62, 221, 60, 0, 52, 216, 63, 221, 60, 0, 52, 216, 64, 221, 60, 0, 52, 216, 65, 221, 60, 0, 52, 216, 66, 221, 60, 0, 52, 216, 67, 221, 60, 0, 52, 216, 68, 221, 60, 0, 52, 216, 69, 221, 60, 0, 52, 216, 70, 221, 60, 0, 52, 216, 71, 221, 60, 0, 52, 216, 72, 221, 60, 0, 52, 216, 73, 221, 60, 0, 52, 216, 74, 221, 60, 0, 52, 216, 75, 221, 60, 0, 52, 216, 76, 221, 60, 0, 52, 216, 77, 221, 60, 0, 52, 216, 78, 221, 60, 0, 52, 216, 79, 221, 60, 0, 52, 216, 80, 221, 60, 0, 52, 216, 81, 221, 60, 0, 52, 216, 82, 221, 60, 0, 52, 216, 83, 221, 60, 0, 52, 216, 84, 221, 60, 0, 52, 216, 85, 221, 60, 0, 52, 216, 86, 221, 60, 0, 52, 216, 87, 221, 60, 0, 52, 216, 88, 221, 60, 0, 52, 216, 89, 221, 60, 0, 52, 216, 90, 221, 60, 0, 52, 216, 91, 221, 60, 0, 52, 216, 92, 221, 60, 0, 52, 216, 93, 221, 60, 0, 52, 216, 106, 221, 60, 0, 52, 216, 107, 221, 60, 0, 52, 216, 108, 221, 60, 0, 52, 216, 131, 221, 60, 0, 52, 216, 132, 221, 60, 0, 52, 216, 140, 221, 60, 0, 52, 216, 141, 221, 60, 0, 52, 216, 142, 221, 60, 0, 52, 216, 143, 221, 60, 0, 52, 216, 144, 221, 60, 0, 52, 216, 145, 221, 60, 0, 52, 216, 146, 221, 60, 0, 52, 216, 147, 221, 60, 0, 52, 216, 148, 221, 60, 0, 52, 216, 149, 221, 60, 0, 52, 216, 150, 221, 60, 0, 52, 216, 151, 221, 60, 0, 52, 216, 152, 221, 60, 0, 52, 216, 153, 221, 60, 0, 52, 216, 154, 221, 60, 0, 52, 216, 155, 221, 60, 0, 52, 216, 156, 221, 60, 0, 52, 216, 157, 221, 60, 0, 52, 216, 158, 221, 60, 0, 52, 216, 159, 221, 60, 0, 52, 216, 160, 221, 60, 0, 52, 216, 161, 221, 60, 0, 52, 216, 162, 221, 60, 0, 52, 216, 163, 221, 60, 0, 52, 216, 164, 221, 60, 0, 52, 216, 165, 221, 60, 0, 52, 216, 166, 221, 60, 0, 52, 216, 167, 221, 60, 0, 52, 216, 168, 221, 60, 0, 52, 216, 169, 221, 60, 0, 52, 216, 174, 221, 60, 0, 52, 216, 175, 221, 60, 0, 52, 216, 176, 221, 60, 0, 52, 216, 177, 221, 60, 0, 52, 216, 178, 221, 60, 0, 52, 216, 179, 221, 60, 0, 52, 216, 180, 221, 60, 0, 52, 216, 181, 221, 60, 0, 52, 216, 182, 221, 60, 0, 52, 216, 183, 221, 60, 0, 52, 216, 184, 221, 60, 0, 52, 216, 185, 221, 60, 0, 52, 216, 186, 221, 60, 0, 52, 216, 193, 221, 60, 0, 52, 216, 194, 221, 60, 0, 52, 216, 195, 221, 60, 0, 52, 216, 196, 221, 60, 0, 52, 216, 197, 221, 60, 0, 52, 216, 198, 221, 60, 0, 52, 216, 199, 221, 60, 0, 52, 216, 200, 221, 60, 0, 52, 216, 201, 221, 60, 0, 52, 216, 202, 221, 60, 0, 52, 216, 203, 221, 60, 0, 52, 216, 204, 221, 60, 0, 52, 216, 205, 221, 60, 0, 52, 216, 206, 221, 60, 0, 52, 216, 207, 221, 60, 0, 52, 216, 208, 221, 60, 0, 52, 216, 209, 221, 60, 0, 52, 216, 210, 221, 60, 0, 52, 216, 211, 221, 60, 0, 52, 216, 212, 221, 60, 0, 52, 216, 213, 221, 60, 0, 52, 216, 214, 221, 60, 0, 52, 216, 215, 221, 60, 0, 52, 216, 216, 221, 60, 0, 52, 216, 217, 221, 60, 0, 52, 216, 218, 221, 60, 0, 52, 216, 219, 221, 60, 0, 52, 216, 220, 221, 60, 0, 52, 216, 221, 221, 60, 0, 52, 216, 222, 221, 60, 0, 52, 216, 223, 221, 60, 0, 52, 216, 224, 221, 60, 0, 52, 216, 225, 221, 60, 0, 52, 216, 226, 221, 60, 0, 52, 216, 227, 221, 60, 0, 52, 216, 228, 221, 60, 0, 52, 216, 229, 221, 60, 0, 52, 216, 230, 221, 60, 0, 52, 216, 231, 221, 60, 0, 52, 216, 232, 221, 60, 0, 52, 216, 0, 222, 60, 0, 52, 216, 1, 222, 60, 0, 52, 216, 2, 222, 60, 0, 52, 216, 3, 222, 60, 0, 52, 216, 4, 222, 60, 0, 52, 216, 5, 222, 60, 0, 52, 216, 6, 222, 60, 0, 52, 216, 7, 222, 60, 0, 52, 216, 8, 222, 60, 0, 52, 216, 9, 222, 60, 0, 52, 216, 10, 222, 60, 0, 52, 216, 11, 222, 60, 0, 52, 216, 12, 222, 60, 0, 52, 216, 13, 222, 60, 0, 52, 216, 14, 222, 60, 0, 52, 216, 15, 222, 60, 0, 52, 216, 16, 222, 60, 0, 52, 216, 17, 222, 60, 0, 52, 216, 18, 222, 60, 0, 52, 216, 19, 222, 60, 0, 52, 216, 20, 222, 60, 0, 52, 216, 21, 222, 60, 0, 52, 216, 22, 222, 60, 0, 52, 216, 23, 222, 60, 0, 52, 216, 24, 222, 60, 0, 52, 216, 25, 222, 60, 0, 52, 216, 26, 222, 60, 0, 52, 216, 27, 222, 60, 0, 52, 216, 28, 222, 60, 0, 52, 216, 29, 222, 60, 0, 52, 216, 30, 222, 60, 0, 52, 216, 31, 222, 60, 0, 52, 216, 32, 222, 60, 0, 52, 216, 33, 222, 60, 0, 52, 216, 34, 222, 60, 0, 52, 216, 35, 222, 60, 0, 52, 216, 36, 222, 60, 0, 52, 216, 37, 222, 60, 0, 52, 216, 38, 222, 60, 0, 52, 216, 39, 222, 60, 0, 52, 216, 40, 222, 60, 0, 52, 216, 41, 222, 60, 0, 52, 216, 42, 222, 60, 0, 52, 216, 43, 222, 60, 0, 52, 216, 44, 222, 60, 0, 52, 216, 45, 222, 60, 0, 52, 216, 46, 222, 60, 0, 52, 216, 47, 222, 60, 0, 52, 216, 48, 222, 60, 0, 52, 216, 49, 222, 60, 0, 52, 216, 50, 222, 60, 0, 52, 216, 51, 222, 60, 0, 52, 216, 52, 222, 60, 0, 52, 216, 53, 222, 60, 0, 52, 216, 54, 222, 60, 0, 52, 216, 55, 222, 60, 0, 52, 216, 56, 222, 60, 0, 52, 216, 57, 222, 60, 0, 52, 216, 58, 222, 60, 0, 52, 216, 59, 222, 60, 0, 52, 216, 60, 222, 60, 0, 52, 216, 61, 222, 60, 0, 52, 216, 62, 222, 60, 0, 52, 216, 63, 222, 60, 0, 52, 216, 64, 222, 60, 0, 52, 216, 65, 222, 60, 0, 52, 216, 69, 222, 60, 0, 47, 216, 156, 220, 60, 0, 14, 32, 59, 216, 172, 220, 14, 32, 60, 0, 14, 32, 59, 216, 46, 221, 14, 32, 60, 0, 60, 216, 0, 220, 60, 0, 60, 216, 1, 220, 60, 0, 60, 216, 2, 220, 60, 0, 60, 216, 3, 220, 60, 0, 60, 216, 4, 220, 60, 0, 60, 216, 5, 220, 60, 0, 60, 216, 6, 220, 60, 0, 60, 216, 7, 220, 60, 0, 60, 216, 8, 220, 60, 0, 60, 216, 9, 220, 60, 0, 60, 216, 10, 220, 60, 0, 60, 216, 11, 220, 60, 0, 60, 216, 12, 220, 60, 0, 60, 216, 13, 220, 60, 0, 60, 216, 14, 220, 60, 0, 60, 216, 15, 220, 60, 0, 60, 216, 16, 220, 60, 0, 60, 216, 17, 220, 60, 0, 60, 216, 18, 220, 60, 0, 60, 216, 19, 220, 60, 0, 60, 216, 20, 220, 60, 0, 60, 216, 21, 220, 60, 0, 60, 216, 22, 220, 60, 0, 60, 216, 23, 220, 60, 0, 60, 216, 24, 220, 60, 0, 60, 216, 25, 220, 60, 0, 60, 216, 26, 220, 60, 0, 60, 216, 27, 220, 60, 0, 60, 216, 28, 220, 60, 0, 60, 216, 29, 220, 60, 0, 60, 216, 30, 220, 60, 0, 60, 216, 31, 220, 60, 0, 60, 216, 32, 220, 60, 0, 60, 216, 33, 220, 60, 0, 60, 216, 34, 220, 60, 0, 60, 216, 35, 220, 60, 0, 60, 216, 36, 220, 60, 0, 60, 216, 37, 220, 60, 0, 60, 216, 38, 220, 60, 0, 60, 216, 39, 220, 60, 0, 60, 216, 40, 220, 60, 0, 60, 216, 41, 220, 60, 0, 60, 216, 42, 220, 60, 0, 60, 216, 43, 220, 60, 0, 60, 216, 48, 220, 60, 0, 60, 216, 49, 220, 60, 0, 60, 216, 50, 220, 60, 0, 60, 216, 51, 220, 60, 0, 60, 216, 52, 220, 60, 0, 60, 216, 53, 220, 60, 0, 60, 216, 54, 220, 60, 0, 60, 216, 55, 220, 60, 0, 60, 216, 56, 220, 60, 0, 60, 216, 57, 220, 60, 0, 60, 216, 58, 220, 60, 0, 60, 216, 59, 220, 60, 0, 60, 216, 60, 220, 60, 0, 60, 216, 61, 220, 60, 0, 60, 216, 62, 220, 60, 0, 60, 216, 63, 220, 60, 0, 60, 216, 64, 220, 60, 0, 60, 216, 65, 220, 60, 0, 60, 216, 66, 220, 60, 0, 60, 216, 67, 220, 60, 0, 60, 216, 68, 220, 60, 0, 60, 216, 69, 220, 60, 0, 60, 216, 70, 220, 60, 0, 60, 216, 71, 220, 60, 0, 60, 216, 72, 220, 60, 0, 60, 216, 73, 220, 60, 0, 60, 216, 74, 220, 60, 0, 60, 216, 75, 220, 60, 0, 60, 216, 76, 220, 60, 0, 60, 216, 77, 220, 60, 0, 60, 216, 78, 220, 60, 0, 60, 216, 79, 220, 60, 0, 60, 216, 80, 220, 60, 0, 60, 216, 81, 220, 60, 0, 60, 216, 82, 220, 60, 0, 60, 216, 83, 220, 60, 0, 60, 216, 84, 220, 60, 0, 60, 216, 85, 220, 60, 0, 60, 216, 86, 220, 60, 0, 60, 216, 87, 220, 60, 0, 60, 216, 88, 220, 60, 0, 60, 216, 89, 220, 60, 0, 60, 216, 90, 220, 60, 0, 60, 216, 91, 220, 60, 0, 60, 216, 92, 220, 60, 0, 60, 216, 93, 220, 60, 0, 60, 216, 94, 220, 60, 0, 60, 216, 95, 220, 60, 0, 60, 216, 96, 220, 60, 0, 60, 216, 97, 220, 60, 0, 60, 216, 98, 220, 60, 0, 60, 216, 99, 220, 60, 0, 60, 216, 100, 220, 60, 0, 60, 216, 101, 220, 60, 0, 60, 216, 102, 220, 60, 0, 60, 216, 103, 220, 60, 0, 60, 216, 104, 220, 60, 0, 60, 216, 105, 220, 60, 0, 60, 216, 106, 220, 60, 0, 60, 216, 107, 220, 60, 0, 60, 216, 108, 220, 60, 0, 60, 216, 109, 220, 60, 0, 60, 216, 110, 220, 60, 0, 60, 216, 111, 220, 60, 0, 60, 216, 112, 220, 60, 0, 60, 216, 113, 220, 60, 0, 60, 216, 114, 220, 60, 0, 60, 216, 115, 220, 60, 0, 60, 216, 116, 220, 60, 0, 60, 216, 117, 220, 60, 0, 60, 216, 118, 220, 60, 0, 60, 216, 119, 220, 60, 0, 60, 216, 120, 220, 60, 0, 60, 216, 121, 220, 60, 0, 60, 216, 122, 220, 60, 0, 60, 216, 123, 220, 60, 0, 60, 216, 124, 220, 60, 0, 60, 216, 125, 220, 60, 0, 60, 216, 126, 220, 60, 0, 60, 216, 127, 220, 60, 0, 60, 216, 128, 220, 60, 0, 60, 216, 129, 220, 60, 0, 60, 216, 130, 220, 60, 0, 60, 216, 131, 220, 60, 0, 60, 216, 132, 220, 60, 0, 60, 216, 133, 220, 60, 0, 60, 216, 134, 220, 60, 0, 60, 216, 135, 220, 60, 0, 60, 216, 136, 220, 60, 0, 60, 216, 137, 220, 60, 0, 60, 216, 138, 220, 60, 0, 60, 216, 139, 220, 60, 0, 60, 216, 140, 220, 60, 0, 60, 216, 141, 220, 60, 0, 60, 216, 142, 220, 60, 0, 60, 216, 143, 220, 60, 0, 60, 216, 144, 220, 60, 0, 60, 216, 145, 220, 60, 0, 60, 216, 146, 220, 60, 0, 60, 216, 147, 220, 60, 0, 60, 216, 160, 220, 60, 0, 60, 216, 161, 220, 60, 0, 60, 216, 162, 220, 60, 0, 60, 216, 163, 220, 60, 0, 60, 216, 164, 220, 60, 0, 60, 216, 165, 220, 60, 0, 60, 216, 166, 220, 60, 0, 60, 216, 167, 220, 60, 0, 60, 216, 168, 220, 60, 0, 60, 216, 169, 220, 60, 0, 60, 216, 170, 220, 60, 0, 60, 216, 171, 220, 60, 0, 60, 216, 172, 220, 60, 0, 60, 216, 173, 220, 60, 0, 60, 216, 174, 220, 60, 0, 60, 216, 177, 220, 60, 0, 60, 216, 178, 220, 60, 0, 60, 216, 179, 220, 60, 0, 60, 216, 180, 220, 60, 0, 60, 216, 181, 220, 60, 0, 60, 216, 182, 220, 60, 0, 60, 216, 183, 220, 60, 0, 60, 216, 184, 220, 60, 0, 60, 216, 185, 220, 60, 0, 60, 216, 186, 220, 60, 0, 60, 216, 187, 220, 60, 0, 60, 216, 188, 220, 60, 0, 60, 216, 189, 220, 60, 0, 60, 216, 190, 220, 60, 0, 60, 216, 191, 220, 60, 0, 60, 216, 193, 220, 60, 0, 60, 216, 194, 220, 60, 0, 60, 216, 195, 220, 60, 0, 60, 216, 196, 220, 60, 0, 60, 216, 197, 220, 60, 0, 60, 216, 198, 220, 60, 0, 60, 216, 199, 220, 60, 0, 60, 216, 200, 220, 60, 0, 60, 216, 201, 220, 60, 0, 60, 216, 202, 220, 60, 0, 60, 216, 203, 220, 60, 0, 60, 216, 204, 220, 60, 0, 60, 216, 205, 220, 60, 0, 60, 216, 206, 220, 60, 0, 60, 216, 207, 220, 60, 0, 60, 216, 209, 220, 60, 0, 60, 216, 210, 220, 60, 0, 60, 216, 211, 220, 60, 0, 60, 216, 212, 220, 60, 0, 60, 216, 213, 220, 60, 0, 60, 216, 214, 220, 60, 0, 60, 216, 215, 220, 60, 0, 60, 216, 216, 220, 60, 0, 60, 216, 217, 220, 60, 0, 60, 216, 218, 220, 60, 0, 60, 216, 219, 220, 60, 0, 60, 216, 220, 220, 60, 0, 60, 216, 221, 220, 60, 0, 60, 216, 222, 220, 60, 0, 60, 216, 223, 220, 60, 0, 60, 216, 224, 220, 60, 0, 60, 216, 225, 220, 60, 0, 60, 216, 226, 220, 60, 0, 60, 216, 227, 220, 60, 0, 60, 216, 228, 220, 60, 0, 60, 216, 229, 220, 60, 0, 60, 216, 230, 220, 60, 0, 60, 216, 231, 220, 60, 0, 60, 216, 232, 220, 60, 0, 60, 216, 233, 220, 60, 0, 60, 216, 234, 220, 60, 0, 60, 216, 235, 220, 60, 0, 60, 216, 236, 220, 60, 0, 60, 216, 237, 220, 60, 0, 60, 216, 238, 220, 60, 0, 60, 216, 239, 220, 60, 0, 60, 216, 240, 220, 60, 0, 60, 216, 241, 220, 60, 0, 60, 216, 242, 220, 60, 0, 60, 216, 243, 220, 60, 0, 60, 216, 244, 220, 60, 0, 60, 216, 245, 220, 60, 0, 62, 216, 0, 222, 60, 0, 62, 216, 1, 222, 60, 0, 62, 216, 2, 222, 60, 0, 62, 216, 3, 222, 60, 0, 62, 216, 4, 222, 60, 0, 62, 216, 5, 222, 60, 0, 62, 216, 6, 222, 60, 0, 62, 216, 7, 222, 60, 0, 62, 216, 8, 222, 60, 0, 62, 216, 9, 222, 60, 0, 62, 216, 10, 222, 60, 0, 62, 216, 11, 222, 60, 0, 62, 216, 12, 222, 60, 0, 62, 216, 13, 222, 60, 0, 62, 216, 14, 222, 60, 0, 62, 216, 15, 222, 60, 0, 62, 216, 16, 222, 60, 0, 62, 216, 17, 222, 60, 0, 62, 216, 18, 222, 60, 0, 62, 216, 19, 222, 60, 0, 62, 216, 20, 222, 60, 0, 62, 216, 21, 222, 60, 0, 62, 216, 22, 222, 60, 0, 62, 216, 23, 222, 60, 0, 62, 216, 24, 222, 60, 0, 62, 216, 25, 222, 60, 0, 62, 216, 26, 222, 60, 0, 62, 216, 27, 222, 60, 0, 62, 216, 28, 222, 60, 0, 62, 216, 29, 222, 60, 0, 62, 216, 30, 222, 60, 0, 62, 216, 31, 222, 60, 0, 62, 216, 32, 222, 60, 0, 62, 216, 33, 222, 60, 0, 62, 216, 34, 222, 60, 0, 62, 216, 35, 222, 60, 0, 62, 216, 36, 222, 60, 0, 62, 216, 37, 222, 60, 0, 62, 216, 38, 222, 60, 0, 62, 216, 39, 222, 60, 0, 62, 216, 40, 222, 60, 0, 62, 216, 41, 222, 60, 0, 62, 216, 42, 222, 60, 0, 62, 216, 43, 222, 60, 0, 62, 216, 44, 222, 60, 0, 62, 216, 45, 222, 60, 0, 62, 216, 46, 222, 60, 0, 62, 216, 47, 222, 60, 0, 62, 216, 48, 222, 60, 0, 62, 216, 49, 222, 60, 0, 62, 216, 50, 222, 60, 0, 62, 216, 51, 222, 60, 0, 62, 216, 52, 222, 60, 0, 62, 216, 53, 222, 60, 0, 62, 216, 54, 222, 60, 0, 62, 216, 55, 222, 60, 0, 62, 216, 56, 222, 60, 0, 62, 216, 57, 222, 60, 0, 62, 216, 58, 222, 60, 0, 62, 216, 59, 222, 60, 0, 62, 216, 60, 222, 60, 0, 62, 216, 61, 222, 60, 0, 62, 216, 62, 222, 60, 0, 62, 216, 63, 222, 60, 0, 62, 216, 64, 222, 60, 0, 62, 216, 65, 222, 60, 0, 62, 216, 66, 222, 60, 0, 62, 216, 67, 222, 60, 0, 62, 216, 68, 222, 60, 0, 62, 216, 69, 222, 60, 0, 62, 216, 70, 222, 60, 0, 62, 216, 71, 222, 60, 0, 62, 216, 72, 222, 60, 0, 62, 216, 73, 222, 60, 0, 62, 216, 74, 222, 60, 0, 62, 216, 75, 222, 60, 0, 62, 216, 76, 222, 60, 0, 62, 216, 77, 222, 60, 0, 62, 216, 78, 222, 60, 0, 62, 216, 79, 222, 60, 0, 62, 216, 80, 222, 60, 0, 62, 216, 81, 222, 60, 0, 62, 216, 82, 222, 60, 0, 62, 216, 83, 222, 60, 0, 62, 216, 96, 222, 60, 0, 62, 216, 97, 222, 60, 0, 62, 216, 98, 222, 60, 0, 62, 216, 99, 222, 60, 0, 62, 216, 100, 222, 60, 0, 62, 216, 101, 222, 60, 0, 62, 216, 102, 222, 60, 0, 62, 216, 103, 222, 60, 0, 62, 216, 104, 222, 60, 0, 62, 216, 105, 222, 60, 0, 62, 216, 106, 222, 60, 0, 62, 216, 107, 222, 60, 0, 62, 216, 108, 222, 60, 0, 62, 216, 109, 222, 60, 0, 60, 216, 13, 221, 60, 0, 60, 216, 14, 221, 60, 0, 60, 216, 15, 221, 60, 0, 60, 216, 109, 221, 60, 0, 60, 216, 110, 221, 60, 0, 60, 216, 111, 221, 60, 0, 60, 216, 96, 222, 60, 0, 60, 216, 97, 222, 60, 0, 60, 216, 98, 222, 60, 0, 60, 216, 99, 222, 60, 0, 60, 216, 100, 222, 60, 0, 60, 216, 101, 222, 60, 0, 60, 216, 0, 223, 60, 0, 60, 216, 1, 223, 60, 0, 60, 216, 2, 223, 60, 0, 60, 216, 3, 223, 60, 0, 60, 216, 4, 223, 60, 0, 60, 216, 5, 223, 60, 0, 60, 216, 6, 223, 60, 0, 60, 216, 7, 223, 60, 0, 60, 216, 8, 223, 60, 0, 60, 216, 9, 223, 60, 0, 60, 216, 10, 223, 60, 0, 60, 216, 11, 223, 60, 0, 60, 216, 12, 223, 60, 0, 60, 216, 13, 223, 60, 0, 60, 216, 14, 223, 60, 0, 60, 216, 15, 223, 60, 0, 60, 216, 16, 223, 60, 0, 60, 216, 17, 223, 60, 0, 60, 216, 18, 223, 60, 0, 60, 216, 19, 223, 60, 0, 60, 216, 20, 223, 60, 0, 60, 216, 21, 223, 60, 0, 60, 216, 22, 223, 60, 0, 60, 216, 23, 223, 60, 0, 60, 216, 24, 223, 60, 0, 60, 216, 25, 223, 60, 0, 60, 216, 26, 223, 60, 0, 60, 216, 27, 223, 60, 0, 60, 216, 28, 223, 60, 0, 60, 216, 29, 223, 60, 0, 60, 216, 30, 223, 60, 0, 60, 216, 31, 223, 60, 0, 60, 216, 32, 223, 60, 0, 60, 216, 33, 223, 60, 0, 60, 216, 34, 223, 60, 0, 60, 216, 35, 223, 60, 0, 60, 216, 36, 223, 60, 0, 60, 216, 37, 223, 60, 0, 60, 216, 38, 223, 60, 0, 60, 216, 39, 223, 60, 0, 60, 216, 40, 223, 60, 0, 60, 216, 41, 223, 60, 0, 60, 216, 42, 223, 60, 0, 60, 216, 43, 223, 60, 0, 60, 216, 44, 223, 60, 0, 60, 216, 45, 223, 60, 0, 60, 216, 46, 223, 60, 0, 60, 216, 47, 223, 60, 0, 60, 216, 48, 223, 60, 0, 60, 216, 49, 223, 60, 0, 60, 216, 50, 223, 60, 0, 60, 216, 51, 223, 60, 0, 60, 216, 52, 223, 60, 0, 60, 216, 53, 223, 60, 0, 60, 216, 54, 223, 60, 0, 60, 216, 55, 223, 60, 0, 60, 216, 56, 223, 60, 0, 60, 216, 57, 223, 60, 0, 60, 216, 58, 223, 60, 0, 60, 216, 59, 223, 60, 0, 60, 216, 60, 223, 60, 0, 60, 216, 61, 223, 60, 0, 60, 216, 62, 223, 60, 0, 60, 216, 63, 223, 60, 0, 60, 216, 64, 223, 60, 0, 60, 216, 65, 223, 60, 0, 60, 216, 66, 223, 60, 0, 60, 216, 67, 223, 60, 0, 60, 216, 68, 223, 60, 0, 60, 216, 69, 223, 60, 0, 60, 216, 70, 223, 60, 0, 60, 216, 71, 223, 60, 0, 60, 216, 72, 223, 60, 0, 60, 216, 73, 223, 60, 0, 60, 216, 74, 223, 60, 0, 60, 216, 75, 223, 60, 0, 60, 216, 76, 223, 60, 0, 60, 216, 77, 223, 60, 0, 60, 216, 78, 223, 60, 0, 60, 216, 79, 223, 60, 0, 60, 216, 80, 223, 60, 0, 60, 216, 81, 223, 60, 0, 60, 216, 82, 223, 60, 0, 60, 216, 83, 223, 60, 0, 60, 216, 84, 223, 60, 0, 60, 216, 85, 223, 60, 0, 60, 216, 86, 223, 60, 0, 60, 216, 87, 223, 60, 0, 60, 216, 88, 223, 60, 0, 60, 216, 89, 223, 60, 0, 60, 216, 90, 223, 60, 0, 60, 216, 91, 223, 60, 0, 60, 216, 92, 223, 60, 0, 60, 216, 93, 223, 60, 0, 60, 216, 94, 223, 60, 0, 60, 216, 95, 223, 60, 0, 60, 216, 96, 223, 60, 0, 60, 216, 97, 223, 60, 0, 60, 216, 98, 223, 60, 0, 60, 216, 99, 223, 60, 0, 60, 216, 100, 223, 60, 0, 60, 216, 101, 223, 60, 0, 60, 216, 102, 223, 60, 0, 60, 216, 103, 223, 60, 0, 60, 216, 104, 223, 60, 0, 60, 216, 105, 223, 60, 0, 60, 216, 106, 223, 60, 0, 60, 216, 107, 223, 60, 0, 60, 216, 108, 223, 60, 0, 60, 216, 109, 223, 60, 0, 60, 216, 110, 223, 60, 0, 60, 216, 111, 223, 60, 0, 60, 216, 112, 223, 60, 0, 60, 216, 113, 223, 60, 0, 60, 216, 114, 223, 60, 0, 60, 216, 115, 223, 60, 0, 60, 216, 116, 223, 60, 0, 60, 216, 117, 223, 60, 0, 60, 216, 118, 223, 60, 0, 60, 216, 119, 223, 60, 0, 60, 216, 120, 223, 60, 0, 60, 216, 121, 223, 60, 0, 60, 216, 122, 223, 60, 0, 60, 216, 123, 223, 60, 0, 60, 216, 124, 223, 60, 0, 60, 216, 125, 223, 60, 0, 60, 216, 126, 223, 60, 0, 60, 216, 127, 223, 60, 0, 60, 216, 128, 223, 60, 0, 60, 216, 129, 223, 60, 0, 60, 216, 130, 223, 60, 0, 60, 216, 131, 223, 60, 0, 60, 216, 132, 223, 60, 0, 60, 216, 133, 223, 60, 0, 60, 216, 134, 223, 60, 0, 60, 216, 135, 223, 60, 0, 60, 216, 136, 223, 60, 0, 60, 216, 137, 223, 60, 0, 60, 216, 138, 223, 60, 0, 60, 216, 139, 223, 60, 0, 60, 216, 140, 223, 60, 0, 60, 216, 141, 223, 60, 0, 60, 216, 142, 223, 60, 0, 60, 216, 143, 223, 60, 0, 60, 216, 144, 223, 60, 0, 60, 216, 145, 223, 60, 0, 60, 216, 146, 223, 60, 0, 60, 216, 147, 223, 60, 0, 60, 216, 148, 223, 60, 0, 60, 216, 149, 223, 60, 0, 60, 216, 150, 223, 60, 0, 60, 216, 151, 223, 60, 0, 60, 216, 152, 223, 60, 0, 60, 216, 153, 223, 60, 0, 60, 216, 154, 223, 60, 0, 60, 216, 155, 223, 60, 0, 60, 216, 156, 223, 60, 0, 60, 216, 157, 223, 60, 0, 60, 216, 158, 223, 60, 0, 60, 216, 159, 223, 60, 0, 60, 216, 160, 223, 60, 0, 60, 216, 161, 223, 60, 0, 60, 216, 162, 223, 60, 0, 60, 216, 163, 223, 60, 0, 60, 216, 164, 223, 60, 0, 60, 216, 165, 223, 60, 0, 60, 216, 166, 223, 60, 0, 60, 216, 167, 223, 60, 0, 60, 216, 168, 223, 60, 0, 60, 216, 169, 223, 60, 0, 60, 216, 170, 223, 60, 0, 60, 216, 171, 223, 60, 0, 60, 216, 172, 223, 60, 0, 60, 216, 173, 223, 60, 0, 60, 216, 174, 223, 60, 0, 60, 216, 175, 223, 60, 0, 60, 216, 176, 223, 60, 0, 60, 216, 177, 223, 60, 0, 60, 216, 178, 223, 60, 0, 60, 216, 179, 223, 60, 0, 60, 216, 180, 223, 60, 0, 60, 216, 181, 223, 60, 0, 60, 216, 182, 223, 60, 0, 60, 216, 183, 223, 60, 0, 60, 216, 184, 223, 60, 0, 60, 216, 185, 223, 60, 0, 60, 216, 186, 223, 60, 0, 60, 216, 187, 223, 60, 0, 60, 216, 188, 223, 60, 0, 60, 216, 189, 223, 60, 0, 60, 216, 190, 223, 60, 0, 60, 216, 191, 223, 60, 0, 60, 216, 192, 223, 60, 0, 60, 216, 193, 223, 60, 0, 60, 216, 194, 223, 60, 0, 60, 216, 195, 223, 60, 0, 60, 216, 196, 223, 60, 0, 60, 216, 197, 223, 60, 0, 60, 216, 198, 223, 60, 0, 60, 216, 199, 223, 60, 0, 60, 216, 200, 223, 60, 0, 60, 216, 201, 223, 60, 0, 60, 216, 202, 223, 60, 0, 60, 216, 203, 223, 60, 0, 60, 216, 204, 223, 60, 0, 60, 216, 205, 223, 60, 0, 60, 216, 206, 223, 60, 0, 60, 216, 207, 223, 60, 0, 60, 216, 208, 223, 60, 0, 60, 216, 209, 223, 60, 0, 60, 216, 210, 223, 60, 0, 60, 216, 211, 223, 60, 0, 60, 216, 212, 223, 60, 0, 60, 216, 213, 223, 60, 0, 60, 216, 214, 223, 60, 0, 60, 216, 215, 223, 60, 0, 60, 216, 216, 223, 60, 0, 60, 216, 217, 223, 60, 0, 60, 216, 218, 223, 60, 0, 60, 216, 219, 223, 60, 0, 60, 216, 220, 223, 60, 0, 60, 216, 221, 223, 60, 0, 60, 216, 222, 223, 60, 0, 60, 216, 223, 223, 60, 0, 60, 216, 224, 223, 60, 0, 60, 216, 225, 223, 60, 0, 60, 216, 226, 223, 60, 0, 60, 216, 227, 223, 60, 0, 60, 216, 228, 223, 60, 0, 60, 216, 229, 223, 60, 0, 60, 216, 230, 223, 60, 0, 60, 216, 231, 223, 60, 0, 60, 216, 232, 223, 60, 0, 60, 216, 233, 223, 60, 0, 60, 216, 234, 223, 60, 0, 60, 216, 235, 223, 60, 0, 60, 216, 236, 223, 60, 0, 60, 216, 237, 223, 60, 0, 60, 216, 238, 223, 60, 0, 60, 216, 239, 223, 60, 0, 60, 216, 240, 223, 60, 0, 60, 216, 241, 223, 60, 0, 60, 216, 242, 223, 60, 0, 60, 216, 243, 223, 60, 0, 60, 216, 244, 223, 60, 0, 60, 216, 245, 223, 60, 0, 60, 216, 246, 223, 60, 0, 60, 216, 247, 223, 60, 0, 60, 216, 248, 223, 60, 0, 60, 216, 249, 223, 60, 0, 60, 216, 250, 223, 60, 0, 60, 216, 251, 223, 60, 0, 60, 216, 252, 223, 60, 0, 60, 216, 253, 223, 60, 0, 60, 216, 254, 223, 60, 0, 60, 216, 255, 223, 60, 0, 61, 216, 0, 220, 60, 0, 61, 216, 1, 220, 60, 0, 61, 216, 2, 220, 60, 0, 61, 216, 3, 220, 60, 0, 61, 216, 4, 220, 60, 0, 61, 216, 5, 220, 60, 0, 61, 216, 6, 220, 60, 0, 61, 216, 7, 220, 60, 0, 61, 216, 8, 220, 60, 0, 61, 216, 9, 220, 60, 0, 61, 216, 10, 220, 60, 0, 61, 216, 11, 220, 60, 0, 61, 216, 12, 220, 60, 0, 61, 216, 13, 220, 60, 0, 61, 216, 14, 220, 60, 0, 61, 216, 15, 220, 60, 0, 61, 216, 16, 220, 60, 0, 61, 216, 17, 220, 60, 0, 61, 216, 18, 220, 60, 0, 61, 216, 19, 220, 60, 0, 61, 216, 20, 220, 60, 0, 61, 216, 21, 220, 60, 0, 61, 216, 22, 220, 60, 0, 61, 216, 23, 220, 60, 0, 61, 216, 24, 220, 60, 0, 61, 216, 25, 220, 60, 0, 61, 216, 26, 220, 60, 0, 61, 216, 27, 220, 60, 0, 61, 216, 28, 220, 60, 0, 61, 216, 29, 220, 60, 0, 61, 216, 30, 220, 60, 0, 61, 216, 31, 220, 60, 0, 61, 216, 32, 220, 60, 0, 61, 216, 33, 220, 60, 0, 61, 216, 34, 220, 60, 0, 61, 216, 35, 220, 60, 0, 61, 216, 36, 220, 60, 0, 61, 216, 37, 220, 60, 0, 61, 216, 38, 220, 60, 0, 61, 216, 39, 220, 60, 0, 61, 216, 40, 220, 60, 0, 61, 216, 41, 220, 60, 0, 61, 216, 42, 220, 60, 0, 61, 216, 43, 220, 60, 0, 61, 216, 44, 220, 60, 0, 61, 216, 45, 220, 60, 0, 61, 216, 46, 220, 60, 0, 61, 216, 47, 220, 60, 0, 61, 216, 48, 220, 60, 0, 61, 216, 49, 220, 60, 0, 61, 216, 50, 220, 60, 0, 61, 216, 51, 220, 60, 0, 61, 216, 52, 220, 60, 0, 61, 216, 53, 220, 60, 0, 61, 216, 54, 220, 60, 0, 61, 216, 55, 220, 60, 0, 61, 216, 56, 220, 60, 0, 61, 216, 57, 220, 60, 0, 61, 216, 58, 220, 60, 0, 61, 216, 59, 220, 60, 0, 61, 216, 60, 220, 60, 0, 61, 216, 61, 220, 60, 0, 61, 216, 62, 220, 60, 0, 61, 216, 63, 220, 60, 0, 61, 216, 64, 220, 60, 0, 61, 216, 65, 220, 60, 0, 61, 216, 66, 220, 60, 0, 61, 216, 67, 220, 60, 0, 61, 216, 68, 220, 60, 0, 61, 216, 69, 220, 60, 0, 61, 216, 70, 220, 60, 0, 61, 216, 71, 220, 60, 0, 61, 216, 72, 220, 60, 0, 61, 216, 73, 220, 60, 0, 61, 216, 74, 220, 60, 0, 61, 216, 75, 220, 60, 0, 61, 216, 76, 220, 60, 0, 61, 216, 77, 220, 60, 0, 61, 216, 78, 220, 60, 0, 61, 216, 79, 220, 60, 0, 61, 216, 80, 220, 60, 0, 61, 216, 81, 220, 60, 0, 61, 216, 82, 220, 60, 0, 61, 216, 83, 220, 60, 0, 61, 216, 84, 220, 60, 0, 61, 216, 85, 220, 60, 0, 61, 216, 86, 220, 60, 0, 61, 216, 87, 220, 60, 0, 61, 216, 88, 220, 60, 0, 61, 216, 89, 220, 60, 0, 61, 216, 90, 220, 60, 0, 61, 216, 91, 220, 60, 0, 61, 216, 92, 220, 60, 0, 61, 216, 93, 220, 60, 0, 61, 216, 94, 220, 60, 0, 61, 216, 95, 220, 60, 0, 61, 216, 96, 220, 60, 0, 61, 216, 97, 220, 60, 0, 61, 216, 98, 220, 60, 0, 61, 216, 99, 220, 60, 0, 61, 216, 100, 220, 60, 0, 61, 216, 101, 220, 60, 0, 61, 216, 102, 220, 60, 0, 61, 216, 103, 220, 60, 0, 61, 216, 104, 220, 60, 0, 61, 216, 105, 220, 60, 0, 61, 216, 106, 220, 60, 0, 61, 216, 107, 220, 60, 0, 61, 216, 108, 220, 60, 0, 61, 216, 109, 220, 60, 0, 61, 216, 110, 220, 60, 0, 61, 216, 111, 220, 60, 0, 61, 216, 112, 220, 60, 0, 61, 216, 113, 220, 60, 0, 61, 216, 114, 220, 60, 0, 61, 216, 115, 220, 60, 0, 61, 216, 116, 220, 60, 0, 61, 216, 117, 220, 60, 0, 61, 216, 118, 220, 60, 0, 61, 216, 119, 220, 60, 0, 61, 216, 120, 220, 60, 0, 61, 216, 121, 220, 60, 0, 61, 216, 122, 220, 60, 0, 61, 216, 123, 220, 60, 0, 61, 216, 124, 220, 60, 0, 61, 216, 125, 220, 60, 0, 61, 216, 126, 220, 60, 0, 61, 216, 127, 220, 60, 0, 61, 216, 128, 220, 60, 0, 61, 216, 129, 220, 60, 0, 61, 216, 130, 220, 60, 0, 61, 216, 131, 220, 60, 0, 61, 216, 132, 220, 60, 0, 61, 216, 133, 220, 60, 0, 61, 216, 134, 220, 60, 0, 61, 216, 135, 220, 60, 0, 61, 216, 136, 220, 60, 0, 61, 216, 137, 220, 60, 0, 61, 216, 138, 220, 60, 0, 61, 216, 139, 220, 60, 0, 61, 216, 140, 220, 60, 0, 61, 216, 141, 220, 60, 0, 61, 216, 142, 220, 60, 0, 61, 216, 143, 220, 60, 0, 61, 216, 144, 220, 60, 0, 61, 216, 145, 220, 60, 0, 61, 216, 146, 220, 60, 0, 61, 216, 147, 220, 60, 0, 61, 216, 148, 220, 60, 0, 61, 216, 149, 220, 60, 0, 61, 216, 150, 220, 60, 0, 61, 216, 151, 220, 60, 0, 61, 216, 152, 220, 60, 0, 61, 216, 153, 220, 60, 0, 61, 216, 154, 220, 60, 0, 61, 216, 155, 220, 60, 0, 61, 216, 156, 220, 60, 0, 61, 216, 157, 220, 60, 0, 61, 216, 158, 220, 60, 0, 61, 216, 159, 220, 60, 0, 61, 216, 160, 220, 60, 0, 61, 216, 161, 220, 60, 0, 61, 216, 162, 220, 60, 0, 61, 216, 163, 220, 60, 0, 61, 216, 164, 220, 60, 0, 61, 216, 165, 220, 60, 0, 61, 216, 166, 220, 60, 0, 61, 216, 167, 220, 60, 0, 61, 216, 168, 220, 60, 0, 61, 216, 169, 220, 60, 0, 61, 216, 170, 220, 60, 0, 61, 216, 171, 220, 60, 0, 61, 216, 172, 220, 60, 0, 61, 216, 173, 220, 60, 0, 61, 216, 174, 220, 60, 0, 61, 216, 175, 220, 60, 0, 61, 216, 176, 220, 60, 0, 61, 216, 177, 220, 60, 0, 61, 216, 178, 220, 60, 0, 61, 216, 179, 220, 60, 0, 61, 216, 180, 220, 60, 0, 61, 216, 181, 220, 60, 0, 61, 216, 182, 220, 60, 0, 61, 216, 183, 220, 60, 0, 61, 216, 184, 220, 60, 0, 61, 216, 185, 220, 60, 0, 61, 216, 186, 220, 60, 0, 61, 216, 187, 220, 60, 0, 61, 216, 188, 220, 60, 0, 61, 216, 189, 220, 60, 0, 61, 216, 190, 220, 60, 0, 61, 216, 191, 220, 60, 0, 61, 216, 192, 220, 60, 0, 61, 216, 193, 220, 60, 0, 61, 216, 194, 220, 60, 0, 61, 216, 195, 220, 60, 0, 61, 216, 196, 220, 60, 0, 61, 216, 197, 220, 60, 0, 61, 216, 198, 220, 60, 0, 61, 216, 199, 220, 60, 0, 61, 216, 200, 220, 60, 0, 61, 216, 201, 220, 60, 0, 61, 216, 202, 220, 60, 0, 61, 216, 203, 220, 60, 0, 61, 216, 204, 220, 60, 0, 61, 216, 205, 220, 60, 0, 61, 216, 206, 220, 60, 0, 61, 216, 207, 220, 60, 0, 61, 216, 208, 220, 60, 0, 61, 216, 209, 220, 60, 0, 61, 216, 210, 220, 60, 0, 61, 216, 211, 220, 60, 0, 61, 216, 212, 220, 60, 0, 61, 216, 213, 220, 60, 0, 61, 216, 214, 220, 60, 0, 61, 216, 215, 220, 60, 0, 61, 216, 216, 220, 60, 0, 61, 216, 217, 220, 60, 0, 61, 216, 218, 220, 60, 0, 61, 216, 219, 220, 60, 0, 61, 216, 220, 220, 60, 0, 61, 216, 221, 220, 60, 0, 61, 216, 222, 220, 60, 0, 61, 216, 223, 220, 60, 0, 61, 216, 224, 220, 60, 0, 61, 216, 225, 220, 60, 0, 61, 216, 226, 220, 60, 0, 61, 216, 227, 220, 60, 0, 61, 216, 228, 220, 60, 0, 61, 216, 229, 220, 60, 0, 61, 216, 230, 220, 60, 0, 61, 216, 231, 220, 60, 0, 61, 216, 232, 220, 60, 0, 61, 216, 233, 220, 60, 0, 61, 216, 234, 220, 60, 0, 61, 216, 235, 220, 60, 0, 61, 216, 236, 220, 60, 0, 61, 216, 237, 220, 60, 0, 61, 216, 238, 220, 60, 0, 61, 216, 239, 220, 60, 0, 61, 216, 240, 220, 60, 0, 61, 216, 241, 220, 60, 0, 61, 216, 242, 220, 60, 0, 61, 216, 243, 220, 60, 0, 61, 216, 244, 220, 60, 0, 61, 216, 245, 220, 60, 0, 61, 216, 246, 220, 60, 0, 61, 216, 247, 220, 60, 0, 61, 216, 248, 220, 60, 0, 61, 216, 249, 220, 60, 0, 61, 216, 250, 220, 60, 0, 61, 216, 251, 220, 60, 0, 61, 216, 252, 220, 60, 0, 61, 216, 253, 220, 60, 0, 61, 216, 254, 220, 60, 0, 61, 216, 255, 220, 60, 0, 61, 216, 0, 221, 60, 0, 61, 216, 1, 221, 60, 0, 61, 216, 2, 221, 60, 0, 61, 216, 3, 221, 60, 0, 61, 216, 4, 221, 60, 0, 61, 216, 5, 221, 60, 0, 61, 216, 6, 221, 60, 0, 61, 216, 7, 221, 60, 0, 61, 216, 8, 221, 60, 0, 61, 216, 9, 221, 60, 0, 61, 216, 10, 221, 60, 0, 61, 216, 11, 221, 60, 0, 61, 216, 12, 221, 60, 0, 61, 216, 13, 221, 60, 0, 61, 216, 14, 221, 60, 0, 61, 216, 15, 221, 60, 0, 61, 216, 16, 221, 60, 0, 61, 216, 17, 221, 60, 0, 61, 216, 18, 221, 60, 0, 61, 216, 19, 221, 60, 0, 61, 216, 20, 221, 60, 0, 61, 216, 21, 221, 60, 0, 61, 216, 22, 221, 60, 0, 61, 216, 23, 221, 60, 0, 61, 216, 24, 221, 60, 0, 61, 216, 25, 221, 60, 0, 61, 216, 26, 221, 60, 0, 61, 216, 27, 221, 60, 0, 61, 216, 28, 221, 60, 0, 61, 216, 29, 221, 60, 0, 61, 216, 30, 221, 60, 0, 61, 216, 31, 221, 60, 0, 61, 216, 32, 221, 60, 0, 61, 216, 33, 221, 60, 0, 61, 216, 34, 221, 60, 0, 61, 216, 35, 221, 60, 0, 61, 216, 36, 221, 60, 0, 61, 216, 37, 221, 60, 0, 61, 216, 38, 221, 60, 0, 61, 216, 39, 221, 60, 0, 61, 216, 40, 221, 60, 0, 61, 216, 41, 221, 60, 0, 61, 216, 42, 221, 60, 0, 61, 216, 43, 221, 60, 0, 61, 216, 44, 221, 60, 0, 61, 216, 45, 221, 60, 0, 61, 216, 46, 221, 60, 0, 61, 216, 47, 221, 60, 0, 61, 216, 48, 221, 60, 0, 61, 216, 49, 221, 60, 0, 61, 216, 50, 221, 60, 0, 61, 216, 51, 221, 60, 0, 61, 216, 52, 221, 60, 0, 61, 216, 53, 221, 60, 0, 61, 216, 54, 221, 60, 0, 61, 216, 55, 221, 60, 0, 61, 216, 56, 221, 60, 0, 61, 216, 57, 221, 60, 0, 61, 216, 58, 221, 60, 0, 61, 216, 59, 221, 60, 0, 61, 216, 60, 221, 60, 0, 61, 216, 61, 221, 60, 0, 61, 216, 62, 221, 60, 0, 61, 216, 63, 221, 60, 0, 61, 216, 64, 221, 60, 0, 61, 216, 65, 221, 60, 0, 61, 216, 66, 221, 60, 0, 61, 216, 67, 221, 60, 0, 61, 216, 68, 221, 60, 0, 61, 216, 69, 221, 60, 0, 61, 216, 70, 221, 60, 0, 61, 216, 71, 221, 60, 0, 61, 216, 72, 221, 60, 0, 61, 216, 73, 221, 60, 0, 61, 216, 74, 221, 60, 0, 61, 216, 75, 221, 60, 0, 61, 216, 76, 221, 60, 0, 61, 216, 77, 221, 60, 0, 61, 216, 78, 221, 60, 0, 61, 216, 79, 221, 60, 0, 61, 216, 80, 221, 60, 0, 61, 216, 81, 221, 60, 0, 61, 216, 82, 221, 60, 0, 61, 216, 83, 221, 60, 0, 61, 216, 84, 221, 60, 0, 61, 216, 85, 221, 60, 0, 61, 216, 86, 221, 60, 0, 61, 216, 87, 221, 60, 0, 61, 216, 88, 221, 60, 0, 61, 216, 89, 221, 60, 0, 61, 216, 90, 221, 60, 0, 61, 216, 91, 221, 60, 0, 61, 216, 92, 221, 60, 0, 61, 216, 93, 221, 60, 0, 61, 216, 94, 221, 60, 0, 61, 216, 95, 221, 60, 0, 61, 216, 96, 221, 60, 0, 61, 216, 97, 221, 60, 0, 61, 216, 98, 221, 60, 0, 61, 216, 99, 221, 60, 0, 61, 216, 100, 221, 60, 0, 61, 216, 101, 221, 60, 0, 61, 216, 102, 221, 60, 0, 61, 216, 103, 221, 60, 0, 61, 216, 104, 221, 60, 0, 61, 216, 105, 221, 60, 0, 61, 216, 106, 221, 60, 0, 61, 216, 107, 221, 60, 0, 61, 216, 108, 221, 60, 0, 61, 216, 109, 221, 60, 0, 61, 216, 110, 221, 60, 0, 61, 216, 111, 221, 60, 0, 61, 216, 112, 221, 60, 0, 61, 216, 113, 221, 60, 0, 61, 216, 114, 221, 60, 0, 61, 216, 115, 221, 60, 0, 61, 216, 116, 221, 60, 0, 61, 216, 117, 221, 60, 0, 61, 216, 118, 221, 60, 0, 61, 216, 119, 221, 60, 0, 61, 216, 120, 221, 60, 0, 61, 216, 121, 221, 60, 0, 61, 216, 122, 221, 60, 0, 61, 216, 123, 221, 60, 0, 61, 216, 124, 221, 60, 0, 61, 216, 125, 221, 60, 0, 61, 216, 126, 221, 60, 0, 61, 216, 127, 221, 60, 0, 61, 216, 128, 221, 60, 0, 61, 216, 129, 221, 60, 0, 61, 216, 130, 221, 60, 0, 61, 216, 131, 221, 60, 0, 61, 216, 132, 221, 60, 0, 61, 216, 133, 221, 60, 0, 61, 216, 134, 221, 60, 0, 61, 216, 135, 221, 60, 0, 61, 216, 136, 221, 60, 0, 61, 216, 137, 221, 60, 0, 61, 216, 138, 221, 60, 0, 61, 216, 139, 221, 60, 0, 61, 216, 140, 221, 60, 0, 61, 216, 141, 221, 60, 0, 61, 216, 142, 221, 60, 0, 61, 216, 143, 221, 60, 0, 61, 216, 144, 221, 60, 0, 61, 216, 145, 221, 60, 0, 61, 216, 146, 221, 60, 0, 61, 216, 147, 221, 60, 0, 61, 216, 148, 221, 60, 0, 61, 216, 149, 221, 60, 0, 61, 216, 150, 221, 60, 0, 61, 216, 151, 221, 60, 0, 61, 216, 152, 221, 60, 0, 61, 216, 153, 221, 60, 0, 61, 216, 154, 221, 60, 0, 61, 216, 155, 221, 60, 0, 61, 216, 156, 221, 60, 0, 61, 216, 157, 221, 60, 0, 61, 216, 158, 221, 60, 0, 61, 216, 159, 221, 60, 0, 61, 216, 160, 221, 60, 0, 61, 216, 161, 221, 60, 0, 61, 216, 162, 221, 60, 0, 61, 216, 163, 221, 60, 0, 61, 216, 164, 221, 60, 0, 61, 216, 165, 221, 60, 0, 61, 216, 166, 221, 60, 0, 61, 216, 167, 221, 60, 0, 61, 216, 168, 221, 60, 0, 61, 216, 169, 221, 60, 0, 61, 216, 170, 221, 60, 0, 61, 216, 171, 221, 60, 0, 61, 216, 172, 221, 60, 0, 61, 216, 173, 221, 60, 0, 61, 216, 174, 221, 60, 0, 61, 216, 175, 221, 60, 0, 61, 216, 176, 221, 60, 0, 61, 216, 177, 221, 60, 0, 61, 216, 178, 221, 60, 0, 61, 216, 179, 221, 60, 0, 61, 216, 180, 221, 60, 0, 61, 216, 181, 221, 60, 0, 61, 216, 182, 221, 60, 0, 61, 216, 183, 221, 60, 0, 61, 216, 184, 221, 60, 0, 61, 216, 185, 221, 60, 0, 61, 216, 186, 221, 60, 0, 61, 216, 187, 221, 60, 0, 61, 216, 188, 221, 60, 0, 61, 216, 189, 221, 60, 0, 61, 216, 190, 221, 60, 0, 61, 216, 191, 221, 60, 0, 61, 216, 192, 221, 60, 0, 61, 216, 193, 221, 60, 0, 61, 216, 194, 221, 60, 0, 61, 216, 195, 221, 60, 0, 61, 216, 196, 221, 60, 0, 61, 216, 197, 221, 60, 0, 61, 216, 198, 221, 60, 0, 61, 216, 199, 221, 60, 0, 61, 216, 200, 221, 60, 0, 61, 216, 201, 221, 60, 0, 61, 216, 202, 221, 60, 0, 61, 216, 203, 221, 60, 0, 61, 216, 204, 221, 60, 0, 61, 216, 205, 221, 60, 0, 61, 216, 206, 221, 60, 0, 61, 216, 207, 221, 60, 0, 61, 216, 208, 221, 60, 0, 61, 216, 209, 221, 60, 0, 61, 216, 210, 221, 60, 0, 61, 216, 211, 221, 60, 0, 61, 216, 212, 221, 60, 0, 61, 216, 213, 221, 60, 0, 61, 216, 214, 221, 60, 0, 61, 216, 215, 221, 60, 0, 61, 216, 216, 221, 60, 0, 61, 216, 217, 221, 60, 0, 61, 216, 218, 221, 60, 0, 61, 216, 219, 221, 60, 0, 61, 216, 220, 221, 60, 0, 61, 216, 221, 221, 60, 0, 61, 216, 222, 221, 60, 0, 61, 216, 223, 221, 60, 0, 61, 216, 224, 221, 60, 0, 61, 216, 225, 221, 60, 0, 61, 216, 226, 221, 60, 0, 61, 216, 227, 221, 60, 0, 61, 216, 228, 221, 60, 0, 61, 216, 229, 221, 60, 0, 61, 216, 230, 221, 60, 0, 61, 216, 231, 221, 60, 0, 61, 216, 232, 221, 60, 0, 61, 216, 233, 221, 60, 0, 61, 216, 234, 221, 60, 0, 61, 216, 235, 221, 60, 0, 61, 216, 236, 221, 60, 0, 61, 216, 237, 221, 60, 0, 61, 216, 238, 221, 60, 0, 61, 216, 239, 221, 60, 0, 61, 216, 240, 221, 60, 0, 61, 216, 241, 221, 60, 0, 61, 216, 242, 221, 60, 0, 61, 216, 243, 221, 60, 0, 61, 216, 244, 221, 60, 0, 61, 216, 245, 221, 60, 0, 61, 216, 246, 221, 60, 0, 61, 216, 247, 221, 60, 0, 61, 216, 248, 221, 60, 0, 61, 216, 249, 221, 60, 0, 61, 216, 250, 221, 60, 0, 61, 216, 251, 221, 60, 0, 61, 216, 252, 221, 60, 0, 61, 216, 253, 221, 60, 0, 61, 216, 254, 221, 60, 0, 61, 216, 255, 221, 60, 0, 62, 216, 0, 221, 60, 0, 62, 216, 1, 221, 60, 0, 62, 216, 2, 221, 60, 0, 62, 216, 3, 221, 60, 0, 62, 216, 4, 221, 60, 0, 62, 216, 5, 221, 60, 0, 62, 216, 6, 221, 60, 0, 62, 216, 7, 221, 60, 0, 62, 216, 8, 221, 60, 0, 62, 216, 9, 221, 60, 0, 62, 216, 10, 221, 60, 0, 62, 216, 11, 221, 60, 0, 62, 216, 12, 221, 60, 0, 62, 216, 13, 221, 60, 0, 62, 216, 14, 221, 60, 0, 62, 216, 15, 221, 60, 0, 62, 216, 16, 221, 60, 0, 62, 216, 17, 221, 60, 0, 62, 216, 18, 221, 60, 0, 62, 216, 19, 221, 60, 0, 62, 216, 20, 221, 60, 0, 62, 216, 21, 221, 60, 0, 62, 216, 22, 221, 60, 0, 62, 216, 23, 221, 60, 0, 62, 216, 24, 221, 60, 0, 62, 216, 25, 221, 60, 0, 62, 216, 26, 221, 60, 0, 62, 216, 27, 221, 60, 0, 62, 216, 28, 221, 60, 0, 62, 216, 29, 221, 60, 0, 62, 216, 30, 221, 60, 0, 62, 216, 31, 221, 60, 0, 62, 216, 32, 221, 60, 0, 62, 216, 33, 221, 60, 0, 62, 216, 34, 221, 60, 0, 62, 216, 35, 221, 60, 0, 62, 216, 36, 221, 60, 0, 62, 216, 37, 221, 60, 0, 62, 216, 38, 221, 60, 0, 62, 216, 39, 221, 60, 0, 62, 216, 40, 221, 60, 0, 62, 216, 41, 221, 60, 0, 62, 216, 42, 221, 60, 0, 62, 216, 43, 221, 60, 0, 62, 216, 44, 221, 60, 0, 62, 216, 45, 221, 60, 0, 62, 216, 46, 221, 60, 0, 62, 216, 47, 221, 60, 0, 62, 216, 48, 221, 60, 0, 62, 216, 49, 221, 60, 0, 62, 216, 50, 221, 60, 0, 62, 216, 51, 221, 60, 0, 62, 216, 52, 221, 60, 0, 62, 216, 53, 221, 60, 0, 62, 216, 54, 221, 60, 0, 62, 216, 55, 221, 60, 0, 62, 216, 56, 221, 60, 0, 62, 216, 57, 221, 60, 0, 62, 216, 58, 221, 60, 0, 62, 216, 59, 221, 60, 0, 62, 216, 60, 221, 60, 0, 62, 216, 61, 221, 60, 0, 62, 216, 62, 221, 60, 0, 62, 216, 63, 221, 60, 0, 62, 216, 64, 221, 60, 0, 62, 216, 65, 221, 60, 0, 62, 216, 66, 221, 60, 0, 62, 216, 67, 221, 60, 0, 62, 216, 68, 221, 60, 0, 62, 216, 69, 221, 60, 0, 62, 216, 70, 221, 60, 0, 62, 216, 71, 221, 60, 0, 62, 216, 72, 221, 60, 0, 62, 216, 73, 221, 60, 0, 62, 216, 74, 221, 60, 0, 62, 216, 75, 221, 60, 0, 62, 216, 76, 221, 60, 0, 62, 216, 77, 221, 60, 0, 62, 216, 78, 221, 60, 0, 62, 216, 79, 221, 60, 0, 62, 216, 80, 221, 60, 0, 62, 216, 81, 221, 60, 0, 62, 216, 82, 221, 60, 0, 62, 216, 83, 221, 60, 0, 62, 216, 84, 221, 60, 0, 62, 216, 85, 221, 60, 0, 62, 216, 86, 221, 60, 0, 62, 216, 87, 221, 60, 0, 62, 216, 88, 221, 60, 0, 62, 216, 89, 221, 60, 0, 62, 216, 90, 221, 60, 0, 62, 216, 91, 221, 60, 0, 62, 216, 92, 221, 60, 0, 62, 216, 93, 221, 60, 0, 62, 216, 94, 221, 60, 0, 62, 216, 95, 221, 60, 0, 62, 216, 96, 221, 60, 0, 62, 216, 97, 221, 60, 0, 62, 216, 98, 221, 60, 0, 62, 216, 99, 221, 60, 0, 62, 216, 100, 221, 60, 0, 62, 216, 101, 221, 60, 0, 62, 216, 102, 221, 60, 0, 62, 216, 103, 221, 60, 0, 62, 216, 104, 221, 60, 0, 62, 216, 105, 221, 60, 0, 62, 216, 106, 221, 60, 0, 62, 216, 107, 221, 60, 0, 62, 216, 108, 221, 60, 0, 62, 216, 109, 221, 60, 0, 62, 216, 110, 221, 60, 0, 62, 216, 111, 221, 60, 0, 62, 216, 112, 221, 60, 0, 62, 216, 113, 221, 60, 0, 62, 216, 114, 221, 60, 0, 62, 216, 115, 221, 60, 0, 62, 216, 116, 221, 60, 0, 62, 216, 117, 221, 60, 0, 62, 216, 118, 221, 60, 0, 62, 216, 119, 221, 60, 0, 62, 216, 120, 221, 60, 0, 62, 216, 122, 221, 60, 0, 62, 216, 123, 221, 60, 0, 62, 216, 124, 221, 60, 0, 62, 216, 125, 221, 60, 0, 62, 216, 126, 221, 60, 0, 62, 216, 127, 221, 60, 0, 62, 216, 128, 221, 60, 0, 62, 216, 129, 221, 60, 0, 62, 216, 130, 221, 60, 0, 62, 216, 131, 221, 60, 0, 62, 216, 132, 221, 60, 0, 62, 216, 133, 221, 60, 0, 62, 216, 134, 221, 60, 0, 62, 216, 135, 221, 60, 0, 62, 216, 136, 221, 60, 0, 62, 216, 137, 221, 60, 0, 62, 216, 138, 221, 60, 0, 62, 216, 139, 221, 60, 0, 62, 216, 140, 221, 60, 0, 62, 216, 141, 221, 60, 0, 62, 216, 142, 221, 60, 0, 62, 216, 143, 221, 60, 0, 62, 216, 144, 221, 60, 0, 62, 216, 145, 221, 60, 0, 62, 216, 146, 221, 60, 0, 62, 216, 147, 221, 60, 0, 62, 216, 148, 221, 60, 0, 62, 216, 149, 221, 60, 0, 62, 216, 150, 221, 60, 0, 62, 216, 151, 221, 60, 0, 62, 216, 152, 221, 60, 0, 62, 216, 153, 221, 60, 0, 62, 216, 154, 221, 60, 0, 62, 216, 155, 221, 60, 0, 62, 216, 156, 221, 60, 0, 62, 216, 157, 221, 60, 0, 62, 216, 158, 221, 60, 0, 62, 216, 159, 221, 60, 0, 62, 216, 160, 221, 60, 0, 62, 216, 161, 221, 60, 0, 62, 216, 162, 221, 60, 0, 62, 216, 163, 221, 60, 0, 62, 216, 164, 221, 60, 0, 62, 216, 165, 221, 60, 0, 62, 216, 166, 221, 60, 0, 62, 216, 167, 221, 60, 0, 62, 216, 168, 221, 60, 0, 62, 216, 169, 221, 60, 0, 62, 216, 170, 221, 60, 0, 62, 216, 171, 221, 60, 0, 62, 216, 172, 221, 60, 0, 62, 216, 173, 221, 60, 0, 62, 216, 174, 221, 60, 0, 62, 216, 175, 221, 60, 0, 62, 216, 176, 221, 60, 0, 62, 216, 177, 221, 60, 0, 62, 216, 178, 221, 60, 0, 62, 216, 179, 221, 60, 0, 62, 216, 180, 221, 60, 0, 62, 216, 181, 221, 60, 0, 62, 216, 182, 221, 60, 0, 62, 216, 183, 221, 60, 0, 62, 216, 184, 221, 60, 0, 62, 216, 185, 221, 60, 0, 62, 216, 186, 221, 60, 0, 62, 216, 187, 221, 60, 0, 62, 216, 188, 221, 60, 0, 62, 216, 189, 221, 60, 0, 62, 216, 190, 221, 60, 0, 62, 216, 191, 221, 60, 0, 62, 216, 192, 221, 60, 0, 62, 216, 193, 221, 60, 0, 62, 216, 194, 221, 60, 0, 62, 216, 195, 221, 60, 0, 62, 216, 196, 221, 60, 0, 62, 216, 197, 221, 60, 0, 62, 216, 198, 221, 60, 0, 62, 216, 199, 221, 60, 0, 62, 216, 200, 221, 60, 0, 62, 216, 201, 221, 60, 0, 62, 216, 202, 221, 60, 0, 62, 216, 203, 221, 60, 0, 62, 216, 205, 221, 60, 0, 62, 216, 206, 221, 60, 0, 62, 216, 207, 221, 60, 0, 62, 216, 208, 221, 60, 0, 62, 216, 209, 221, 60, 0, 62, 216, 210, 221, 60, 0, 62, 216, 211, 221, 60, 0, 62, 216, 212, 221, 60, 0, 62, 216, 213, 221, 60, 0, 62, 216, 214, 221, 60, 0, 62, 216, 215, 221, 60, 0, 62, 216, 216, 221, 60, 0, 62, 216, 217, 221, 60, 0, 62, 216, 218, 221, 60, 0, 62, 216, 219, 221, 60, 0, 62, 216, 220, 221, 60, 0, 62, 216, 221, 221, 60, 0, 62, 216, 222, 221, 60, 0, 62, 216, 223, 221, 60, 0, 62, 216, 224, 221, 60, 0, 62, 216, 225, 221, 60, 0, 62, 216, 226, 221, 60, 0, 62, 216, 227, 221, 60, 0, 62, 216, 228, 221, 60, 0, 62, 216, 229, 221, 60, 0, 62, 216, 230, 221, 60, 0, 62, 216, 231, 221, 60, 0, 62, 216, 232, 221, 60, 0, 62, 216, 233, 221, 60, 0, 62, 216, 234, 221, 60, 0, 62, 216, 235, 221, 60, 0, 62, 216, 236, 221, 60, 0, 62, 216, 237, 221, 60, 0, 62, 216, 238, 221, 60, 0, 62, 216, 239, 221, 60, 0, 62, 216, 240, 221, 60, 0, 62, 216, 241, 221, 60, 0, 62, 216, 242, 221, 60, 0, 62, 216, 243, 221, 60, 0, 62, 216, 244, 221, 60, 0, 62, 216, 245, 221, 60, 0, 62, 216, 246, 221, 60, 0, 62, 216, 247, 221, 60, 0, 62, 216, 248, 221, 60, 0, 62, 216, 249, 221, 60, 0, 62, 216, 250, 221, 60, 0, 62, 216, 251, 221, 60, 0, 62, 216, 252, 221, 60, 0, 62, 216, 253, 221, 60, 0, 62, 216, 254, 221, 60, 0, 62, 216, 255, 221, 60, 0, 62, 216, 112, 222, 60, 0, 62, 216, 113, 222, 60, 0, 62, 216, 114, 222, 60, 0, 62, 216, 115, 222, 60, 0, 62, 216, 116, 222, 60, 0, 62, 216, 120, 222, 60, 0, 62, 216, 121, 222, 60, 0, 62, 216, 122, 222, 60, 0, 62, 216, 128, 222, 60, 0, 62, 216, 129, 222, 60, 0, 62, 216, 130, 222, 60, 0, 62, 216, 131, 222, 60, 0, 62, 216, 132, 222, 60, 0, 62, 216, 133, 222, 60, 0, 62, 216, 134, 222, 60, 0, 62, 216, 144, 222, 60, 0, 62, 216, 145, 222, 60, 0, 62, 216, 146, 222, 60, 0, 62, 216, 147, 222, 60, 0, 62, 216, 148, 222, 60, 0, 62, 216, 149, 222, 60, 0, 62, 216, 150, 222, 60, 0, 62, 216, 151, 222, 60, 0, 62, 216, 152, 222, 60, 0, 62, 216, 153, 222, 60, 0, 62, 216, 154, 222, 60, 0, 62, 216, 155, 222, 60, 0, 62, 216, 156, 222, 60, 0, 62, 216, 157, 222, 60, 0, 62, 216, 158, 222, 60, 0, 62, 216, 159, 222, 60, 0, 62, 216, 160, 222, 60, 0, 62, 216, 161, 222, 60, 0, 62, 216, 162, 222, 60, 0, 62, 216, 163, 222, 60, 0, 62, 216, 164, 222, 60, 0, 62, 216, 165, 222, 60, 0, 62, 216, 166, 222, 60, 0, 62, 216, 167, 222, 60, 0, 62, 216, 168, 222, 60, 0, 62, 216, 176, 222, 60, 0, 62, 216, 177, 222, 60, 0, 62, 216, 178, 222, 60, 0, 62, 216, 179, 222, 60, 0, 62, 216, 180, 222, 60, 0, 62, 216, 181, 222, 60, 0, 62, 216, 182, 222, 60, 0, 62, 216, 192, 222, 60, 0, 62, 216, 193, 222, 60, 0, 62, 216, 194, 222, 60, 0, 62, 216, 208, 222, 60, 0, 62, 216, 209, 222, 60, 0, 62, 216, 210, 222, 60, 0, 62, 216, 211, 222, 60, 0, 62, 216, 212, 222, 60, 0, 62, 216, 213, 222, 60, 0, 62, 216, 214, 222, 60, 0, 61, 216, 0, 222, 60, 0, 61, 216, 1, 222, 60, 0, 61, 216, 2, 222, 60, 0, 61, 216, 3, 222, 60, 0, 61, 216, 4, 222, 60, 0, 61, 216, 5, 222, 60, 0, 61, 216, 6, 222, 60, 0, 61, 216, 7, 222, 60, 0, 61, 216, 8, 222, 60, 0, 61, 216, 9, 222, 60, 0, 61, 216, 10, 222, 60, 0, 61, 216, 11, 222, 60, 0, 61, 216, 12, 222, 60, 0, 61, 216, 13, 222, 60, 0, 61, 216, 14, 222, 60, 0, 61, 216, 15, 222, 60, 0, 61, 216, 16, 222, 60, 0, 61, 216, 17, 222, 60, 0, 61, 216, 18, 222, 60, 0, 61, 216, 19, 222, 60, 0, 61, 216, 20, 222, 60, 0, 61, 216, 21, 222, 60, 0, 61, 216, 22, 222, 60, 0, 61, 216, 23, 222, 60, 0, 61, 216, 24, 222, 60, 0, 61, 216, 25, 222, 60, 0, 61, 216, 26, 222, 60, 0, 61, 216, 27, 222, 60, 0, 61, 216, 28, 222, 60, 0, 61, 216, 29, 222, 60, 0, 61, 216, 30, 222, 60, 0, 61, 216, 31, 222, 60, 0, 61, 216, 32, 222, 60, 0, 61, 216, 33, 222, 60, 0, 61, 216, 34, 222, 60, 0, 61, 216, 35, 222, 60, 0, 61, 216, 36, 222, 60, 0, 61, 216, 37, 222, 60, 0, 61, 216, 38, 222, 60, 0, 61, 216, 39, 222, 60, 0, 61, 216, 40, 222, 60, 0, 61, 216, 41, 222, 60, 0, 61, 216, 42, 222, 60, 0, 61, 216, 43, 222, 60, 0, 61, 216, 44, 222, 60, 0, 61, 216, 45, 222, 60, 0, 61, 216, 46, 222, 60, 0, 61, 216, 47, 222, 60, 0, 61, 216, 48, 222, 60, 0, 61, 216, 49, 222, 60, 0, 61, 216, 50, 222, 60, 0, 61, 216, 51, 222, 60, 0, 61, 216, 52, 222, 60, 0, 61, 216, 53, 222, 60, 0, 61, 216, 54, 222, 60, 0, 61, 216, 55, 222, 60, 0, 61, 216, 56, 222, 60, 0, 61, 216, 57, 222, 60, 0, 61, 216, 58, 222, 60, 0, 61, 216, 59, 222, 60, 0, 61, 216, 60, 222, 60, 0, 61, 216, 61, 222, 60, 0, 61, 216, 62, 222, 60, 0, 61, 216, 63, 222, 60, 0, 61, 216, 64, 222, 60, 0, 61, 216, 65, 222, 60, 0, 61, 216, 66, 222, 60, 0, 61, 216, 67, 222, 60, 0, 61, 216, 68, 222, 60, 0, 61, 216, 69, 222, 60, 0, 61, 216, 70, 222, 60, 0, 61, 216, 71, 222, 60, 0, 61, 216, 72, 222, 60, 0, 61, 216, 73, 222, 60, 0, 61, 216, 74, 222, 60, 0, 61, 216, 75, 222, 60, 0, 61, 216, 76, 222, 60, 0, 61, 216, 77, 222, 60, 0, 61, 216, 78, 222, 60, 0, 61, 216, 79, 222, 60, 0, 61, 216, 80, 222, 60, 0, 61, 216, 81, 222, 60, 0, 61, 216, 82, 222, 60, 0, 61, 216, 83, 222, 60, 0, 61, 216, 84, 222, 60, 0, 61, 216, 85, 222, 60, 0, 61, 216, 86, 222, 60, 0, 61, 216, 87, 222, 60, 0, 61, 216, 88, 222, 60, 0, 61, 216, 89, 222, 60, 0, 61, 216, 90, 222, 60, 0, 61, 216, 91, 222, 60, 0, 61, 216, 92, 222, 60, 0, 61, 216, 93, 222, 60, 0, 61, 216, 94, 222, 60, 0, 61, 216, 95, 222, 60, 0, 61, 216, 96, 222, 60, 0, 61, 216, 97, 222, 60, 0, 61, 216, 98, 222, 60, 0, 61, 216, 99, 222, 60, 0, 61, 216, 100, 222, 60, 0, 61, 216, 101, 222, 60, 0, 61, 216, 102, 222, 60, 0, 61, 216, 103, 222, 60, 0, 61, 216, 104, 222, 60, 0, 61, 216, 105, 222, 60, 0, 61, 216, 106, 222, 60, 0, 61, 216, 107, 222, 60, 0, 61, 216, 108, 222, 60, 0, 61, 216, 109, 222, 60, 0, 61, 216, 110, 222, 60, 0, 61, 216, 111, 222, 60, 0, 61, 216, 112, 222, 60, 0, 61, 216, 113, 222, 60, 0, 61, 216, 114, 222, 60, 0, 61, 216, 115, 222, 60, 0, 61, 216, 116, 222, 60, 0, 61, 216, 117, 222, 60, 0, 61, 216, 118, 222, 60, 0, 61, 216, 119, 222, 60, 0, 61, 216, 120, 222, 60, 0, 61, 216, 121, 222, 60, 0, 61, 216, 122, 222, 60, 0, 61, 216, 123, 222, 60, 0, 61, 216, 124, 222, 60, 0, 61, 216, 125, 222, 60, 0, 61, 216, 126, 222, 60, 0, 61, 216, 127, 222, 60, 0, 61, 216, 128, 222, 60, 0, 61, 216, 129, 222, 60, 0, 61, 216, 130, 222, 60, 0, 61, 216, 131, 222, 60, 0, 61, 216, 132, 222, 60, 0, 61, 216, 133, 222, 60, 0, 61, 216, 134, 222, 60, 0, 61, 216, 135, 222, 60, 0, 61, 216, 136, 222, 60, 0, 61, 216, 137, 222, 60, 0, 61, 216, 138, 222, 60, 0, 61, 216, 139, 222, 60, 0, 61, 216, 140, 222, 60, 0, 61, 216, 141, 222, 60, 0, 61, 216, 142, 222, 60, 0, 61, 216, 143, 222, 60, 0, 61, 216, 144, 222, 60, 0, 61, 216, 145, 222, 60, 0, 61, 216, 146, 222, 60, 0, 61, 216, 147, 222, 60, 0, 61, 216, 148, 222, 60, 0, 61, 216, 149, 222, 60, 0, 61, 216, 150, 222, 60, 0, 61, 216, 151, 222, 60, 0, 61, 216, 152, 222, 60, 0, 61, 216, 153, 222, 60, 0, 61, 216, 154, 222, 60, 0, 61, 216, 155, 222, 60, 0, 61, 216, 156, 222, 60, 0, 61, 216, 157, 222, 60, 0, 61, 216, 158, 222, 60, 0, 61, 216, 159, 222, 60, 0, 61, 216, 160, 222, 60, 0, 61, 216, 161, 222, 60, 0, 61, 216, 162, 222, 60, 0, 61, 216, 163, 222, 60, 0, 61, 216, 164, 222, 60, 0, 61, 216, 165, 222, 60, 0, 61, 216, 166, 222, 60, 0, 61, 216, 167, 222, 60, 0, 61, 216, 168, 222, 60, 0, 61, 216, 169, 222, 60, 0, 61, 216, 170, 222, 60, 0, 61, 216, 171, 222, 60, 0, 61, 216, 172, 222, 60, 0, 61, 216, 173, 222, 60, 0, 61, 216, 174, 222, 60, 0, 61, 216, 175, 222, 60, 0, 61, 216, 176, 222, 60, 0, 61, 216, 177, 222, 60, 0, 61, 216, 178, 222, 60, 0, 61, 216, 179, 222, 60, 0, 61, 216, 180, 222, 60, 0, 61, 216, 181, 222, 60, 0, 61, 216, 182, 222, 60, 0, 61, 216, 183, 222, 60, 0, 61, 216, 184, 222, 60, 0, 61, 216, 185, 222, 60, 0, 61, 216, 186, 222, 60, 0, 61, 216, 187, 222, 60, 0, 61, 216, 188, 222, 60, 0, 61, 216, 189, 222, 60, 0, 61, 216, 190, 222, 60, 0, 61, 216, 191, 222, 60, 0, 61, 216, 192, 222, 60, 0, 61, 216, 193, 222, 60, 0, 61, 216, 194, 222, 60, 0, 61, 216, 195, 222, 60, 0, 61, 216, 196, 222, 60, 0, 61, 216, 197, 222, 60, 0, 61, 216, 198, 222, 60, 0, 61, 216, 199, 222, 60, 0, 61, 216, 200, 222, 60, 0, 61, 216, 201, 222, 60, 0, 61, 216, 202, 222, 60, 0, 61, 216, 203, 222, 60, 0, 61, 216, 204, 222, 60, 0, 61, 216, 205, 222, 60, 0, 61, 216, 206, 222, 60, 0, 61, 216, 207, 222, 60, 0, 61, 216, 208, 222, 60, 0, 61, 216, 209, 222, 60, 0, 61, 216, 210, 222, 60, 0, 61, 216, 211, 222, 60, 0, 61, 216, 212, 222, 60, 0, 61, 216, 213, 222, 60, 0, 61, 216, 214, 222, 60, 0, 61, 216, 215, 222, 60, 0, 61, 216, 224, 222, 60, 0, 61, 216, 225, 222, 60, 0, 61, 216, 226, 222, 60, 0, 61, 216, 227, 222, 60, 0, 61, 216, 228, 222, 60, 0, 61, 216, 229, 222, 60, 0, 61, 216, 230, 222, 60, 0, 61, 216, 231, 222, 60, 0, 61, 216, 232, 222, 60, 0, 61, 216, 233, 222, 60, 0, 61, 216, 234, 222, 60, 0, 61, 216, 235, 222, 60, 0, 61, 216, 236, 222, 60, 0, 61, 216, 240, 222, 60, 0, 61, 216, 241, 222, 60, 0, 61, 216, 242, 222, 60, 0, 61, 216, 243, 222, 60, 0, 61, 216, 244, 222, 60, 0, 61, 216, 245, 222, 60, 0, 61, 216, 246, 222, 60, 0, 61, 216, 247, 222, 60, 0, 61, 216, 248, 222, 60, 0, 61, 216, 249, 222, 60, 0, 61, 216, 250, 222, 60, 0, 61, 216, 251, 222, 60, 0, 61, 216, 252, 222, 60, 0, 61, 216, 0, 223, 60, 0, 61, 216, 1, 223, 60, 0, 61, 216, 2, 223, 60, 0, 61, 216, 3, 223, 60, 0, 61, 216, 4, 223, 60, 0, 61, 216, 5, 223, 60, 0, 61, 216, 6, 223, 60, 0, 61, 216, 7, 223, 60, 0, 61, 216, 8, 223, 60, 0, 61, 216, 9, 223, 60, 0, 61, 216, 10, 223, 60, 0, 61, 216, 11, 223, 60, 0, 61, 216, 12, 223, 60, 0, 61, 216, 13, 223, 60, 0, 61, 216, 14, 223, 60, 0, 61, 216, 15, 223, 60, 0, 61, 216, 16, 223, 60, 0, 61, 216, 17, 223, 60, 0, 61, 216, 18, 223, 60, 0, 61, 216, 19, 223, 60, 0, 61, 216, 20, 223, 60, 0, 61, 216, 21, 223, 60, 0, 61, 216, 22, 223, 60, 0, 61, 216, 23, 223, 60, 0, 61, 216, 24, 223, 60, 0, 61, 216, 25, 223, 60, 0, 61, 216, 26, 223, 60, 0, 61, 216, 27, 223, 60, 0, 61, 216, 28, 223, 60, 0, 61, 216, 29, 223, 60, 0, 61, 216, 30, 223, 60, 0, 61, 216, 31, 223, 60, 0, 61, 216, 32, 223, 60, 0, 61, 216, 33, 223, 60, 0, 61, 216, 34, 223, 60, 0, 61, 216, 35, 223, 60, 0, 61, 216, 36, 223, 60, 0, 61, 216, 37, 223, 60, 0, 61, 216, 38, 223, 60, 0, 61, 216, 39, 223, 60, 0, 61, 216, 40, 223, 60, 0, 61, 216, 41, 223, 60, 0, 61, 216, 42, 223, 60, 0, 61, 216, 43, 223, 60, 0, 61, 216, 44, 223, 60, 0, 61, 216, 45, 223, 60, 0, 61, 216, 46, 223, 60, 0, 61, 216, 47, 223, 60, 0, 61, 216, 48, 223, 60, 0, 61, 216, 49, 223, 60, 0, 61, 216, 50, 223, 60, 0, 61, 216, 51, 223, 60, 0, 61, 216, 52, 223, 60, 0, 61, 216, 53, 223, 60, 0, 61, 216, 54, 223, 60, 0, 61, 216, 55, 223, 60, 0, 61, 216, 56, 223, 60, 0, 61, 216, 57, 223, 60, 0, 61, 216, 58, 223, 60, 0, 61, 216, 59, 223, 60, 0, 61, 216, 60, 223, 60, 0, 61, 216, 61, 223, 60, 0, 61, 216, 62, 223, 60, 0, 61, 216, 63, 223, 60, 0, 61, 216, 64, 223, 60, 0, 61, 216, 65, 223, 60, 0, 61, 216, 66, 223, 60, 0, 61, 216, 67, 223, 60, 0, 61, 216, 68, 223, 60, 0, 61, 216, 69, 223, 60, 0, 61, 216, 70, 223, 60, 0, 61, 216, 71, 223, 60, 0, 61, 216, 72, 223, 60, 0, 61, 216, 73, 223, 60, 0, 61, 216, 74, 223, 60, 0, 61, 216, 75, 223, 60, 0, 61, 216, 76, 223, 60, 0, 61, 216, 77, 223, 60, 0, 61, 216, 78, 223, 60, 0, 61, 216, 79, 223, 60, 0, 61, 216, 80, 223, 60, 0, 61, 216, 81, 223, 60, 0, 61, 216, 82, 223, 60, 0, 61, 216, 83, 223, 60, 0, 61, 216, 84, 223, 60, 0, 61, 216, 85, 223, 60, 0, 61, 216, 86, 223, 60, 0, 61, 216, 87, 223, 60, 0, 61, 216, 88, 223, 60, 0, 61, 216, 89, 223, 60, 0, 61, 216, 90, 223, 60, 0, 61, 216, 91, 223, 60, 0, 61, 216, 92, 223, 60, 0, 61, 216, 93, 223, 60, 0, 61, 216, 94, 223, 60, 0, 61, 216, 95, 223, 60, 0, 61, 216, 96, 223, 60, 0, 61, 216, 97, 223, 60, 0, 61, 216, 98, 223, 60, 0, 61, 216, 99, 223, 60, 0, 61, 216, 100, 223, 60, 0, 61, 216, 101, 223, 60, 0, 61, 216, 102, 223, 60, 0, 61, 216, 103, 223, 60, 0, 61, 216, 104, 223, 60, 0, 61, 216, 105, 223, 60, 0, 61, 216, 106, 223, 60, 0, 61, 216, 107, 223, 60, 0, 61, 216, 108, 223, 60, 0, 61, 216, 109, 223, 60, 0, 61, 216, 110, 223, 60, 0, 61, 216, 111, 223, 60, 0, 61, 216, 112, 223, 60, 0, 61, 216, 113, 223, 60, 0, 61, 216, 114, 223, 60, 0, 61, 216, 115, 223, 60, 0, 61, 216, 128, 223, 60, 0, 61, 216, 129, 223, 60, 0, 61, 216, 130, 223, 60, 0, 61, 216, 131, 223, 60, 0, 61, 216, 132, 223, 60, 0, 61, 216, 133, 223, 60, 0, 61, 216, 134, 223, 60, 0, 61, 216, 135, 223, 60, 0, 61, 216, 136, 223, 60, 0, 61, 216, 137, 223, 60, 0, 61, 216, 138, 223, 60, 0, 61, 216, 139, 223, 60, 0, 61, 216, 140, 223, 60, 0, 61, 216, 141, 223, 60, 0, 61, 216, 142, 223, 60, 0, 61, 216, 143, 223, 60, 0, 61, 216, 144, 223, 60, 0, 61, 216, 145, 223, 60, 0, 61, 216, 146, 223, 60, 0, 61, 216, 147, 223, 60, 0, 61, 216, 148, 223, 60, 0, 61, 216, 149, 223, 60, 0, 61, 216, 150, 223, 60, 0, 61, 216, 151, 223, 60, 0, 61, 216, 152, 223, 60, 0, 61, 216, 153, 223, 60, 0, 61, 216, 154, 223, 60, 0, 61, 216, 155, 223, 60, 0, 61, 216, 156, 223, 60, 0, 61, 216, 157, 223, 60, 0, 61, 216, 158, 223, 60, 0, 61, 216, 159, 223, 60, 0, 61, 216, 160, 223, 60, 0, 61, 216, 161, 223, 60, 0, 61, 216, 162, 223, 60, 0, 61, 216, 163, 223, 60, 0, 61, 216, 164, 223, 60, 0, 61, 216, 165, 223, 60, 0, 61, 216, 166, 223, 60, 0, 61, 216, 167, 223, 60, 0, 61, 216, 168, 223, 60, 0, 61, 216, 169, 223, 60, 0, 61, 216, 170, 223, 60, 0, 61, 216, 171, 223, 60, 0, 61, 216, 172, 223, 60, 0, 61, 216, 173, 223, 60, 0, 61, 216, 174, 223, 60, 0, 61, 216, 175, 223, 60, 0, 61, 216, 176, 223, 60, 0, 61, 216, 177, 223, 60, 0, 61, 216, 178, 223, 60, 0, 61, 216, 179, 223, 60, 0, 61, 216, 180, 223, 60, 0, 61, 216, 181, 223, 60, 0, 61, 216, 182, 223, 60, 0, 61, 216, 183, 223, 60, 0, 61, 216, 184, 223, 60, 0, 61, 216, 185, 223, 60, 0, 61, 216, 186, 223, 60, 0, 61, 216, 187, 223, 60, 0, 61, 216, 188, 223, 60, 0, 61, 216, 189, 223, 60, 0, 61, 216, 190, 223, 60, 0, 61, 216, 191, 223, 60, 0, 61, 216, 192, 223, 60, 0, 61, 216, 193, 223, 60, 0, 61, 216, 194, 223, 60, 0, 61, 216, 195, 223, 60, 0, 61, 216, 196, 223, 60, 0, 61, 216, 197, 223, 60, 0, 61, 216, 198, 223, 60, 0, 61, 216, 199, 223, 60, 0, 61, 216, 200, 223, 60, 0, 61, 216, 201, 223, 60, 0, 61, 216, 202, 223, 60, 0, 61, 216, 203, 223, 60, 0, 61, 216, 204, 223, 60, 0, 61, 216, 205, 223, 60, 0, 61, 216, 206, 223, 60, 0, 61, 216, 207, 223, 60, 0, 61, 216, 208, 223, 60, 0, 61, 216, 209, 223, 60, 0, 61, 216, 210, 223, 60, 0, 61, 216, 211, 223, 60, 0, 61, 216, 212, 223, 60, 0, 61, 216, 213, 223, 60, 0, 61, 216, 214, 223, 60, 0, 61, 216, 215, 223, 60, 0, 61, 216, 216, 223, 60, 0, 61, 216, 224, 223, 60, 0, 61, 216, 225, 223, 60, 0, 61, 216, 226, 223, 60, 0, 61, 216, 227, 223, 60, 0, 61, 216, 228, 223, 60, 0, 61, 216, 229, 223, 60, 0, 61, 216, 230, 223, 60, 0, 61, 216, 231, 223, 60, 0, 61, 216, 232, 223, 60, 0, 61, 216, 233, 223, 60, 0, 61, 216, 234, 223, 60, 0, 61, 216, 235, 223, 60, 0, 62, 216, 0, 220, 60, 0, 62, 216, 1, 220, 60, 0, 62, 216, 2, 220, 60, 0, 62, 216, 3, 220, 60, 0, 62, 216, 4, 220, 60, 0, 62, 216, 5, 220, 60, 0, 62, 216, 6, 220, 60, 0, 62, 216, 7, 220, 60, 0, 62, 216, 8, 220, 60, 0, 62, 216, 9, 220, 60, 0, 62, 216, 10, 220, 60, 0, 62, 216, 11, 220, 60, 0, 62, 216, 16, 220, 60, 0, 62, 216, 17, 220, 60, 0, 62, 216, 18, 220, 60, 0, 62, 216, 19, 220, 60, 0, 62, 216, 20, 220, 60, 0, 62, 216, 21, 220, 60, 0, 62, 216, 22, 220, 60, 0, 62, 216, 23, 220, 60, 0, 62, 216, 24, 220, 60, 0, 62, 216, 25, 220, 60, 0, 62, 216, 26, 220, 60, 0, 62, 216, 27, 220, 60, 0, 62, 216, 28, 220, 60, 0, 62, 216, 29, 220, 60, 0, 62, 216, 30, 220, 60, 0, 62, 216, 31, 220, 60, 0, 62, 216, 32, 220, 60, 0, 62, 216, 33, 220, 60, 0, 62, 216, 34, 220, 60, 0, 62, 216, 35, 220, 60, 0, 62, 216, 36, 220, 60, 0, 62, 216, 37, 220, 60, 0, 62, 216, 38, 220, 60, 0, 62, 216, 39, 220, 60, 0, 62, 216, 40, 220, 60, 0, 62, 216, 41, 220, 60, 0, 62, 216, 42, 220, 60, 0, 62, 216, 43, 220, 60, 0, 62, 216, 44, 220, 60, 0, 62, 216, 45, 220, 60, 0, 62, 216, 46, 220, 60, 0, 62, 216, 47, 220, 60, 0, 62, 216, 48, 220, 60, 0, 62, 216, 49, 220, 60, 0, 62, 216, 50, 220, 60, 0, 62, 216, 51, 220, 60, 0, 62, 216, 52, 220, 60, 0, 62, 216, 53, 220, 60, 0, 62, 216, 54, 220, 60, 0, 62, 216, 55, 220, 60, 0, 62, 216, 56, 220, 60, 0, 62, 216, 57, 220, 60, 0, 62, 216, 58, 220, 60, 0, 62, 216, 59, 220, 60, 0, 62, 216, 60, 220, 60, 0, 62, 216, 61, 220, 60, 0, 62, 216, 62, 220, 60, 0, 62, 216, 63, 220, 60, 0, 62, 216, 64, 220, 60, 0, 62, 216, 65, 220, 60, 0, 62, 216, 66, 220, 60, 0, 62, 216, 67, 220, 60, 0, 62, 216, 68, 220, 60, 0, 62, 216, 69, 220, 60, 0, 62, 216, 70, 220, 60, 0, 62, 216, 71, 220, 60, 0, 62, 216, 80, 220, 60, 0, 62, 216, 81, 220, 60, 0, 62, 216, 82, 220, 60, 0, 62, 216, 83, 220, 60, 0, 62, 216, 84, 220, 60, 0, 62, 216, 85, 220, 60, 0, 62, 216, 86, 220, 60, 0, 62, 216, 87, 220, 60, 0, 62, 216, 88, 220, 60, 0, 62, 216, 89, 220, 60, 0, 62, 216, 96, 220, 60, 0, 62, 216, 97, 220, 60, 0, 62, 216, 98, 220, 60, 0, 62, 216, 99, 220, 60, 0, 62, 216, 100, 220, 60, 0, 62, 216, 101, 220, 60, 0, 62, 216, 102, 220, 60, 0, 62, 216, 103, 220, 60, 0, 62, 216, 104, 220, 60, 0, 62, 216, 105, 220, 60, 0, 62, 216, 106, 220, 60, 0, 62, 216, 107, 220, 60, 0, 62, 216, 108, 220, 60, 0, 62, 216, 109, 220, 60, 0, 62, 216, 110, 220, 60, 0, 62, 216, 111, 220, 60, 0, 62, 216, 112, 220, 60, 0, 62, 216, 113, 220, 60, 0, 62, 216, 114, 220, 60, 0, 62, 216, 115, 220, 60, 0, 62, 216, 116, 220, 60, 0, 62, 216, 117, 220, 60, 0, 62, 216, 118, 220, 60, 0, 62, 216, 119, 220, 60, 0, 62, 216, 120, 220, 60, 0, 62, 216, 121, 220, 60, 0, 62, 216, 122, 220, 60, 0, 62, 216, 123, 220, 60, 0, 62, 216, 124, 220, 60, 0, 62, 216, 125, 220, 60, 0, 62, 216, 126, 220, 60, 0, 62, 216, 127, 220, 60, 0, 62, 216, 128, 220, 60, 0, 62, 216, 129, 220, 60, 0, 62, 216, 130, 220, 60, 0, 62, 216, 131, 220, 60, 0, 62, 216, 132, 220, 60, 0, 62, 216, 133, 220, 60, 0, 62, 216, 134, 220, 60, 0, 62, 216, 135, 220, 60, 0, 62, 216, 144, 220, 60, 0, 62, 216, 145, 220, 60, 0, 62, 216, 146, 220, 60, 0, 62, 216, 147, 220, 60, 0, 62, 216, 148, 220, 60, 0, 62, 216, 149, 220, 60, 0, 62, 216, 150, 220, 60, 0, 62, 216, 151, 220, 60, 0, 62, 216, 152, 220, 60, 0, 62, 216, 153, 220, 60, 0, 62, 216, 154, 220, 60, 0, 62, 216, 155, 220, 60, 0, 62, 216, 156, 220, 60, 0, 62, 216, 157, 220, 60, 0, 62, 216, 158, 220, 60, 0, 62, 216, 159, 220, 60, 0, 62, 216, 160, 220, 60, 0, 62, 216, 161, 220, 60, 0, 62, 216, 162, 220, 60, 0, 62, 216, 163, 220, 60, 0, 62, 216, 164, 220, 60, 0, 62, 216, 165, 220, 60, 0, 62, 216, 166, 220, 60, 0, 62, 216, 167, 220, 60, 0, 62, 216, 168, 220, 60, 0, 62, 216, 169, 220, 60, 0, 62, 216, 170, 220, 60, 0, 62, 216, 171, 220, 60, 0, 62, 216, 172, 220, 60, 0, 62, 216, 173, 220, 60, 0, 62, 216, 176, 220, 60, 0, 62, 216, 177, 220, 60, 0, 54, 216, 0, 220, 60, 0, 54, 216, 1, 220, 60, 0, 54, 216, 2, 220, 60, 0, 54, 216, 3, 220, 60, 0, 54, 216, 4, 220, 60, 0, 54, 216, 5, 220, 60, 0, 54, 216, 6, 220, 60, 0, 54, 216, 7, 220, 60, 0, 54, 216, 8, 220, 60, 0, 54, 216, 9, 220, 60, 0, 54, 216, 10, 220, 60, 0, 54, 216, 11, 220, 60, 0, 54, 216, 12, 220, 60, 0, 54, 216, 13, 220, 60, 0, 54, 216, 14, 220, 60, 0, 54, 216, 15, 220, 60, 0, 54, 216, 16, 220, 60, 0, 54, 216, 17, 220, 60, 0, 54, 216, 18, 220, 60, 0, 54, 216, 19, 220, 60, 0, 54, 216, 20, 220, 60, 0, 54, 216, 21, 220, 60, 0, 54, 216, 22, 220, 60, 0, 54, 216, 23, 220, 60, 0, 54, 216, 24, 220, 60, 0, 54, 216, 25, 220, 60, 0, 54, 216, 26, 220, 60, 0, 54, 216, 27, 220, 60, 0, 54, 216, 28, 220, 60, 0, 54, 216, 29, 220, 60, 0, 54, 216, 30, 220, 60, 0, 54, 216, 31, 220, 60, 0, 54, 216, 32, 220, 60, 0, 54, 216, 33, 220, 60, 0, 54, 216, 34, 220, 60, 0, 54, 216, 35, 220, 60, 0, 54, 216, 36, 220, 60, 0, 54, 216, 37, 220, 60, 0, 54, 216, 38, 220, 60, 0, 54, 216, 39, 220, 60, 0, 54, 216, 40, 220, 60, 0, 54, 216, 41, 220, 60, 0, 54, 216, 42, 220, 60, 0, 54, 216, 43, 220, 60, 0, 54, 216, 44, 220, 60, 0, 54, 216, 45, 220, 60, 0, 54, 216, 46, 220, 60, 0, 54, 216, 47, 220, 60, 0, 54, 216, 48, 220, 60, 0, 54, 216, 49, 220, 60, 0, 54, 216, 50, 220, 60, 0, 54, 216, 51, 220, 60, 0, 54, 216, 52, 220, 60, 0, 54, 216, 53, 220, 60, 0, 54, 216, 54, 220, 60, 0, 54, 216, 55, 220, 60, 0, 54, 216, 56, 220, 60, 0, 54, 216, 57, 220, 60, 0, 54, 216, 58, 220, 60, 0, 54, 216, 59, 220, 60, 0, 54, 216, 60, 220, 60, 0, 54, 216, 61, 220, 60, 0, 54, 216, 62, 220, 60, 0, 54, 216, 63, 220, 60, 0, 54, 216, 64, 220, 60, 0, 54, 216, 65, 220, 60, 0, 54, 216, 66, 220, 60, 0, 54, 216, 67, 220, 60, 0, 54, 216, 68, 220, 60, 0, 54, 216, 69, 220, 60, 0, 54, 216, 70, 220, 60, 0, 54, 216, 71, 220, 60, 0, 54, 216, 72, 220, 60, 0, 54, 216, 73, 220, 60, 0, 54, 216, 74, 220, 60, 0, 54, 216, 75, 220, 60, 0, 54, 216, 76, 220, 60, 0, 54, 216, 77, 220, 60, 0, 54, 216, 78, 220, 60, 0, 54, 216, 79, 220, 60, 0, 54, 216, 80, 220, 60, 0, 54, 216, 81, 220, 60, 0, 54, 216, 82, 220, 60, 0, 54, 216, 83, 220, 60, 0, 54, 216, 84, 220, 60, 0, 54, 216, 85, 220, 60, 0, 54, 216, 86, 220, 60, 0, 54, 216, 87, 220, 60, 0, 54, 216, 88, 220, 60, 0, 54, 216, 89, 220, 60, 0, 54, 216, 90, 220, 60, 0, 54, 216, 91, 220, 60, 0, 54, 216, 92, 220, 60, 0, 54, 216, 93, 220, 60, 0, 54, 216, 94, 220, 60, 0, 54, 216, 95, 220, 60, 0, 54, 216, 96, 220, 60, 0, 54, 216, 97, 220, 60, 0, 54, 216, 98, 220, 60, 0, 54, 216, 99, 220, 60, 0, 54, 216, 100, 220, 60, 0, 54, 216, 101, 220, 60, 0, 54, 216, 102, 220, 60, 0, 54, 216, 103, 220, 60, 0, 54, 216, 104, 220, 60, 0, 54, 216, 105, 220, 60, 0, 54, 216, 106, 220, 60, 0, 54, 216, 107, 220, 60, 0, 54, 216, 108, 220, 60, 0, 54, 216, 109, 220, 60, 0, 54, 216, 110, 220, 60, 0, 54, 216, 111, 220, 60, 0, 54, 216, 112, 220, 60, 0, 54, 216, 113, 220, 60, 0, 54, 216, 114, 220, 60, 0, 54, 216, 115, 220, 60, 0, 54, 216, 116, 220, 60, 0, 54, 216, 117, 220, 60, 0, 54, 216, 118, 220, 60, 0, 54, 216, 119, 220, 60, 0, 54, 216, 120, 220, 60, 0, 54, 216, 121, 220, 60, 0, 54, 216, 122, 220, 60, 0, 54, 216, 123, 220, 60, 0, 54, 216, 124, 220, 60, 0, 54, 216, 125, 220, 60, 0, 54, 216, 126, 220, 60, 0, 54, 216, 127, 220, 60, 0, 54, 216, 128, 220, 60, 0, 54, 216, 129, 220, 60, 0, 54, 216, 130, 220, 60, 0, 54, 216, 131, 220, 60, 0, 54, 216, 132, 220, 60, 0, 54, 216, 133, 220, 60, 0, 54, 216, 134, 220, 60, 0, 54, 216, 135, 220, 60, 0, 54, 216, 136, 220, 60, 0, 54, 216, 137, 220, 60, 0, 54, 216, 138, 220, 60, 0, 54, 216, 139, 220, 60, 0, 54, 216, 140, 220, 60, 0, 54, 216, 141, 220, 60, 0, 54, 216, 142, 220, 60, 0, 54, 216, 143, 220, 60, 0, 54, 216, 144, 220, 60, 0, 54, 216, 145, 220, 60, 0, 54, 216, 146, 220, 60, 0, 54, 216, 147, 220, 60, 0, 54, 216, 148, 220, 60, 0, 54, 216, 149, 220, 60, 0, 54, 216, 150, 220, 60, 0, 54, 216, 151, 220, 60, 0, 54, 216, 152, 220, 60, 0, 54, 216, 153, 220, 60, 0, 54, 216, 154, 220, 60, 0, 54, 216, 155, 220, 60, 0, 54, 216, 156, 220, 60, 0, 54, 216, 157, 220, 60, 0, 54, 216, 158, 220, 60, 0, 54, 216, 159, 220, 60, 0, 54, 216, 160, 220, 60, 0, 54, 216, 161, 220, 60, 0, 54, 216, 162, 220, 60, 0, 54, 216, 163, 220, 60, 0, 54, 216, 164, 220, 60, 0, 54, 216, 165, 220, 60, 0, 54, 216, 166, 220, 60, 0, 54, 216, 167, 220, 60, 0, 54, 216, 168, 220, 60, 0, 54, 216, 169, 220, 60, 0, 54, 216, 170, 220, 60, 0, 54, 216, 171, 220, 60, 0, 54, 216, 172, 220, 60, 0, 54, 216, 173, 220, 60, 0, 54, 216, 174, 220, 60, 0, 54, 216, 175, 220, 60, 0, 54, 216, 176, 220, 60, 0, 54, 216, 177, 220, 60, 0, 54, 216, 178, 220, 60, 0, 54, 216, 179, 220, 60, 0, 54, 216, 180, 220, 60, 0, 54, 216, 181, 220, 60, 0, 54, 216, 182, 220, 60, 0, 54, 216, 183, 220, 60, 0, 54, 216, 184, 220, 60, 0, 54, 216, 185, 220, 60, 0, 54, 216, 186, 220, 60, 0, 54, 216, 187, 220, 60, 0, 54, 216, 188, 220, 60, 0, 54, 216, 189, 220, 60, 0, 54, 216, 190, 220, 60, 0, 54, 216, 191, 220, 60, 0, 54, 216, 192, 220, 60, 0, 54, 216, 193, 220, 60, 0, 54, 216, 194, 220, 60, 0, 54, 216, 195, 220, 60, 0, 54, 216, 196, 220, 60, 0, 54, 216, 197, 220, 60, 0, 54, 216, 198, 220, 60, 0, 54, 216, 199, 220, 60, 0, 54, 216, 200, 220, 60, 0, 54, 216, 201, 220, 60, 0, 54, 216, 202, 220, 60, 0, 54, 216, 203, 220, 60, 0, 54, 216, 204, 220, 60, 0, 54, 216, 205, 220, 60, 0, 54, 216, 206, 220, 60, 0, 54, 216, 207, 220, 60, 0, 54, 216, 208, 220, 60, 0, 54, 216, 209, 220, 60, 0, 54, 216, 210, 220, 60, 0, 54, 216, 211, 220, 60, 0, 54, 216, 212, 220, 60, 0, 54, 216, 213, 220, 60, 0, 54, 216, 214, 220, 60, 0, 54, 216, 215, 220, 60, 0, 54, 216, 216, 220, 60, 0, 54, 216, 217, 220, 60, 0, 54, 216, 218, 220, 60, 0, 54, 216, 219, 220, 60, 0, 54, 216, 220, 220, 60, 0, 54, 216, 221, 220, 60, 0, 54, 216, 222, 220, 60, 0, 54, 216, 223, 220, 60, 0, 54, 216, 224, 220, 60, 0, 54, 216, 225, 220, 60, 0, 54, 216, 226, 220, 60, 0, 54, 216, 227, 220, 60, 0, 54, 216, 228, 220, 60, 0, 54, 216, 229, 220, 60, 0, 54, 216, 230, 220, 60, 0, 54, 216, 231, 220, 60, 0, 54, 216, 232, 220, 60, 0, 54, 216, 233, 220, 60, 0, 54, 216, 234, 220, 60, 0, 54, 216, 235, 220, 60, 0, 54, 216, 236, 220, 60, 0, 54, 216, 237, 220, 60, 0, 54, 216, 238, 220, 60, 0, 54, 216, 239, 220, 60, 0, 54, 216, 240, 220, 60, 0, 54, 216, 241, 220, 60, 0, 54, 216, 242, 220, 60, 0, 54, 216, 243, 220, 60, 0, 54, 216, 244, 220, 60, 0, 54, 216, 245, 220, 60, 0, 54, 216, 246, 220, 60, 0, 54, 216, 247, 220, 60, 0, 54, 216, 248, 220, 60, 0, 54, 216, 249, 220, 60, 0, 54, 216, 250, 220, 60, 0, 54, 216, 251, 220, 60, 0, 54, 216, 252, 220, 60, 0, 54, 216, 253, 220, 60, 0, 54, 216, 254, 220, 60, 0, 54, 216, 255, 220, 60, 0, 54, 216, 0, 221, 60, 0, 54, 216, 1, 221, 60, 0, 54, 216, 2, 221, 60, 0, 54, 216, 3, 221, 60, 0, 54, 216, 4, 221, 60, 0, 54, 216, 5, 221, 60, 0, 54, 216, 6, 221, 60, 0, 54, 216, 7, 221, 60, 0, 54, 216, 8, 221, 60, 0, 54, 216, 9, 221, 60, 0, 54, 216, 10, 221, 60, 0, 54, 216, 11, 221, 60, 0, 54, 216, 12, 221, 60, 0, 54, 216, 13, 221, 60, 0, 54, 216, 14, 221, 60, 0, 54, 216, 15, 221, 60, 0, 54, 216, 16, 221, 60, 0, 54, 216, 17, 221, 60, 0, 54, 216, 18, 221, 60, 0, 54, 216, 19, 221, 60, 0, 54, 216, 20, 221, 60, 0, 54, 216, 21, 221, 60, 0, 54, 216, 22, 221, 60, 0, 54, 216, 23, 221, 60, 0, 54, 216, 24, 221, 60, 0, 54, 216, 25, 221, 60, 0, 54, 216, 26, 221, 60, 0, 54, 216, 27, 221, 60, 0, 54, 216, 28, 221, 60, 0, 54, 216, 29, 221, 60, 0, 54, 216, 30, 221, 60, 0, 54, 216, 31, 221, 60, 0, 54, 216, 32, 221, 60, 0, 54, 216, 33, 221, 60, 0, 54, 216, 34, 221, 60, 0, 54, 216, 35, 221, 60, 0, 54, 216, 36, 221, 60, 0, 54, 216, 37, 221, 60, 0, 54, 216, 38, 221, 60, 0, 54, 216, 39, 221, 60, 0, 54, 216, 40, 221, 60, 0, 54, 216, 41, 221, 60, 0, 54, 216, 42, 221, 60, 0, 54, 216, 43, 221, 60, 0, 54, 216, 44, 221, 60, 0, 54, 216, 45, 221, 60, 0, 54, 216, 46, 221, 60, 0, 54, 216, 47, 221, 60, 0, 54, 216, 48, 221, 60, 0, 54, 216, 49, 221, 60, 0, 54, 216, 50, 221, 60, 0, 54, 216, 51, 221, 60, 0, 54, 216, 52, 221, 60, 0, 54, 216, 53, 221, 60, 0, 54, 216, 54, 221, 60, 0, 54, 216, 55, 221, 60, 0, 54, 216, 56, 221, 60, 0, 54, 216, 57, 221, 60, 0, 54, 216, 58, 221, 60, 0, 54, 216, 59, 221, 60, 0, 54, 216, 60, 221, 60, 0, 54, 216, 61, 221, 60, 0, 54, 216, 62, 221, 60, 0, 54, 216, 63, 221, 60, 0, 54, 216, 64, 221, 60, 0, 54, 216, 65, 221, 60, 0, 54, 216, 66, 221, 60, 0, 54, 216, 67, 221, 60, 0, 54, 216, 68, 221, 60, 0, 54, 216, 69, 221, 60, 0, 54, 216, 70, 221, 60, 0, 54, 216, 71, 221, 60, 0, 54, 216, 72, 221, 60, 0, 54, 216, 73, 221, 60, 0, 54, 216, 74, 221, 60, 0, 54, 216, 75, 221, 60, 0, 54, 216, 76, 221, 60, 0, 54, 216, 77, 221, 60, 0, 54, 216, 78, 221, 60, 0, 54, 216, 79, 221, 60, 0, 54, 216, 80, 221, 60, 0, 54, 216, 81, 221, 60, 0, 54, 216, 82, 221, 60, 0, 54, 216, 83, 221, 60, 0, 54, 216, 84, 221, 60, 0, 54, 216, 85, 221, 60, 0, 54, 216, 86, 221, 60, 0, 54, 216, 87, 221, 60, 0, 54, 216, 88, 221, 60, 0, 54, 216, 89, 221, 60, 0, 54, 216, 90, 221, 60, 0, 54, 216, 91, 221, 60, 0, 54, 216, 92, 221, 60, 0, 54, 216, 93, 221, 60, 0, 54, 216, 94, 221, 60, 0, 54, 216, 95, 221, 60, 0, 54, 216, 96, 221, 60, 0, 54, 216, 97, 221, 60, 0, 54, 216, 98, 221, 60, 0, 54, 216, 99, 221, 60, 0, 54, 216, 100, 221, 60, 0, 54, 216, 101, 221, 60, 0, 54, 216, 102, 221, 60, 0, 54, 216, 103, 221, 60, 0, 54, 216, 104, 221, 60, 0, 54, 216, 105, 221, 60, 0, 54, 216, 106, 221, 60, 0, 54, 216, 107, 221, 60, 0, 54, 216, 108, 221, 60, 0, 54, 216, 109, 221, 60, 0, 54, 216, 110, 221, 60, 0, 54, 216, 111, 221, 60, 0, 54, 216, 112, 221, 60, 0, 54, 216, 113, 221, 60, 0, 54, 216, 114, 221, 60, 0, 54, 216, 115, 221, 60, 0, 54, 216, 116, 221, 60, 0, 54, 216, 117, 221, 60, 0, 54, 216, 118, 221, 60, 0, 54, 216, 119, 221, 60, 0, 54, 216, 120, 221, 60, 0, 54, 216, 121, 221, 60, 0, 54, 216, 122, 221, 60, 0, 54, 216, 123, 221, 60, 0, 54, 216, 124, 221, 60, 0, 54, 216, 125, 221, 60, 0, 54, 216, 126, 221, 60, 0, 54, 216, 127, 221, 60, 0, 54, 216, 128, 221, 60, 0, 54, 216, 129, 221, 60, 0, 54, 216, 130, 221, 60, 0, 54, 216, 131, 221, 60, 0, 54, 216, 132, 221, 60, 0, 54, 216, 133, 221, 60, 0, 54, 216, 134, 221, 60, 0, 54, 216, 135, 221, 60, 0, 54, 216, 136, 221, 60, 0, 54, 216, 137, 221, 60, 0, 54, 216, 138, 221, 60, 0, 54, 216, 139, 221, 60, 0, 54, 216, 140, 221, 60, 0, 54, 216, 141, 221, 60, 0, 54, 216, 142, 221, 60, 0, 54, 216, 143, 221, 60, 0, 54, 216, 144, 221, 60, 0, 54, 216, 145, 221, 60, 0, 54, 216, 146, 221, 60, 0, 54, 216, 147, 221, 60, 0, 54, 216, 148, 221, 60, 0, 54, 216, 149, 221, 60, 0, 54, 216, 150, 221, 60, 0, 54, 216, 151, 221, 60, 0, 54, 216, 152, 221, 60, 0, 54, 216, 153, 221, 60, 0, 54, 216, 154, 221, 60, 0, 54, 216, 155, 221, 60, 0, 54, 216, 156, 221, 60, 0, 54, 216, 157, 221, 60, 0, 54, 216, 158, 221, 60, 0, 54, 216, 159, 221, 60, 0, 54, 216, 160, 221, 60, 0, 54, 216, 161, 221, 60, 0, 54, 216, 162, 221, 60, 0, 54, 216, 163, 221, 60, 0, 54, 216, 164, 221, 60, 0, 54, 216, 165, 221, 60, 0, 54, 216, 166, 221, 60, 0, 54, 216, 167, 221, 60, 0, 54, 216, 168, 221, 60, 0, 54, 216, 169, 221, 60, 0, 54, 216, 170, 221, 60, 0, 54, 216, 171, 221, 60, 0, 54, 216, 172, 221, 60, 0, 54, 216, 173, 221, 60, 0, 54, 216, 174, 221, 60, 0, 54, 216, 175, 221, 60, 0, 54, 216, 176, 221, 60, 0, 54, 216, 177, 221, 60, 0, 54, 216, 178, 221, 60, 0, 54, 216, 179, 221, 60, 0, 54, 216, 180, 221, 60, 0, 54, 216, 181, 221, 60, 0, 54, 216, 182, 221, 60, 0, 54, 216, 183, 221, 60, 0, 54, 216, 184, 221, 60, 0, 54, 216, 185, 221, 60, 0, 54, 216, 186, 221, 60, 0, 54, 216, 187, 221, 60, 0, 54, 216, 188, 221, 60, 0, 54, 216, 189, 221, 60, 0, 54, 216, 190, 221, 60, 0, 54, 216, 191, 221, 60, 0, 54, 216, 192, 221, 60, 0, 54, 216, 193, 221, 60, 0, 54, 216, 194, 221, 60, 0, 54, 216, 195, 221, 60, 0, 54, 216, 196, 221, 60, 0, 54, 216, 197, 221, 60, 0, 54, 216, 198, 221, 60, 0, 54, 216, 199, 221, 60, 0, 54, 216, 200, 221, 60, 0, 54, 216, 201, 221, 60, 0, 54, 216, 202, 221, 60, 0, 54, 216, 203, 221, 60, 0, 54, 216, 204, 221, 60, 0, 54, 216, 205, 221, 60, 0, 54, 216, 206, 221, 60, 0, 54, 216, 207, 221, 60, 0, 54, 216, 208, 221, 60, 0, 54, 216, 209, 221, 60, 0, 54, 216, 210, 221, 60, 0, 54, 216, 211, 221, 60, 0, 54, 216, 212, 221, 60, 0, 54, 216, 213, 221, 60, 0, 54, 216, 214, 221, 60, 0, 54, 216, 215, 221, 60, 0, 54, 216, 216, 221, 60, 0, 54, 216, 217, 221, 60, 0, 54, 216, 218, 221, 60, 0, 54, 216, 219, 221, 60, 0, 54, 216, 220, 221, 60, 0, 54, 216, 221, 221, 60, 0, 54, 216, 222, 221, 60, 0, 54, 216, 223, 221, 60, 0, 54, 216, 224, 221, 60, 0, 54, 216, 225, 221, 60, 0, 54, 216, 226, 221, 60, 0, 54, 216, 227, 221, 60, 0, 54, 216, 228, 221, 60, 0, 54, 216, 229, 221, 60, 0, 54, 216, 230, 221, 60, 0, 54, 216, 231, 221, 60, 0, 54, 216, 232, 221, 60, 0, 54, 216, 233, 221, 60, 0, 54, 216, 234, 221, 60, 0, 54, 216, 235, 221, 60, 0, 54, 216, 236, 221, 60, 0, 54, 216, 237, 221, 60, 0, 54, 216, 238, 221, 60, 0, 54, 216, 239, 221, 60, 0, 54, 216, 240, 221, 60, 0, 54, 216, 241, 221, 60, 0, 54, 216, 242, 221, 60, 0, 54, 216, 243, 221, 60, 0, 54, 216, 244, 221, 60, 0, 54, 216, 245, 221, 60, 0, 54, 216, 246, 221, 60, 0, 54, 216, 247, 221, 60, 0, 54, 216, 248, 221, 60, 0, 54, 216, 249, 221, 60, 0, 54, 216, 250, 221, 60, 0, 54, 216, 251, 221, 60, 0, 54, 216, 252, 221, 60, 0, 54, 216, 253, 221, 60, 0, 54, 216, 254, 221, 60, 0, 54, 216, 255, 221, 60, 0, 54, 216, 55, 222, 60, 0, 54, 216, 56, 222, 60, 0, 54, 216, 57, 222, 60, 0, 54, 216, 58, 222, 60, 0, 54, 216, 109, 222, 60, 0, 54, 216, 110, 222, 60, 0, 54, 216, 111, 222, 60, 0, 54, 216, 112, 222, 60, 0, 54, 216, 113, 222, 60, 0, 54, 216, 114, 222, 60, 0, 54, 216, 115, 222, 60, 0, 54, 216, 116, 222, 60, 0, 54, 216, 118, 222, 60, 0, 54, 216, 119, 222, 60, 0, 54, 216, 120, 222, 60, 0, 54, 216, 121, 222, 60, 0, 54, 216, 122, 222, 60, 0, 54, 216, 123, 222, 60, 0, 54, 216, 124, 222, 60, 0, 54, 216, 125, 222, 60, 0, 54, 216, 126, 222, 60, 0, 54, 216, 127, 222, 60, 0, 54, 216, 128, 222, 60, 0, 54, 216, 129, 222, 60, 0, 54, 216, 130, 222, 60, 0, 54, 216, 131, 222, 60, 0, 54, 216, 133, 222, 60, 0, 54, 216, 134, 222, 60, 0, 240, 47, 60, 0, 241, 47, 60, 0, 242, 47, 60, 0, 243, 47, 60, 0, 244, 47, 60, 0, 245, 47, 60, 0, 246, 47, 60, 0, 247, 47, 60, 0, 248, 47, 60, 0, 249, 47, 60, 0, 250, 47, 60, 0, 251, 47, 60, 0, 192, 49, 60, 0, 193, 49, 60, 0, 194, 49, 60, 0, 195, 49, 60, 0, 196, 49, 60, 0, 197, 49, 60, 0, 198, 49, 60, 0, 199, 49, 60, 0, 200, 49, 60, 0, 201, 49, 60, 0, 202, 49, 60, 0, 203, 49, 60, 0, 204, 49, 60, 0, 205, 49, 60, 0, 206, 49, 60, 0, 207, 49, 60, 0, 208, 49, 60, 0, 209, 49, 60, 0, 210, 49, 60, 0, 211, 49, 60, 0, 212, 49, 60, 0, 213, 49, 60, 0, 214, 49, 60, 0, 215, 49, 60, 0, 216, 49, 60, 0, 217, 49, 60, 0, 218, 49, 60, 0, 219, 49, 60, 0, 220, 49, 60, 0, 221, 49, 60, 0, 222, 49, 60, 0, 223, 49, 60, 0, 224, 49, 60, 0, 225, 49, 60, 0, 226, 49, 60, 0, 227, 49, 60, 0, 4, 48, 60, 0, 18, 48, 60, 0, 60, 0, 60, 0, 54, 48, 60, 0, 19, 48, 60, 0, 32, 48, 60, 0, 55, 48, 60, 0, 62, 48, 60, 0, 63, 48, 60, 0, 144, 49, 60, 0, 145, 49, 60, 0, 127, 50, 60, 0, 252, 255, 60, 0, 208, 2, 60, 0, 209, 2, 60, 0, 113, 9, 60, 0, 70, 14, 60, 0, 198, 14, 60, 0, 215, 23, 60, 0, 167, 26, 60, 0, 207, 169, 60, 0, 230, 169, 60, 0, 112, 170, 60, 0, 221, 170, 60, 0, 243, 170, 60, 0, 244, 170, 60, 0, 26, 216, 66, 223, 60, 0, 26, 216, 67, 223, 60, 0, 56, 216, 60, 221, 60, 0, 56, 216, 61, 221, 60, 0, 5, 48, 60, 0, 59, 48, 60, 0, 27, 216, 224, 223, 60, 0, 27, 216, 225, 223, 60, 0, 27, 216, 227, 223, 60, 0, 49, 48, 60, 0, 60, 0, 60, 0, 50, 48, 47, 0, 153, 48, 60, 0, 51, 48, 60, 0, 60, 0, 60, 0, 52, 48, 47, 0, 153, 48, 60, 0, 53, 48, 60, 0, 157, 48, 60, 0, 252, 48, 60, 0, 60, 0, 60, 0, 112, 255, 60, 0, 253, 48, 38, 0, 252, 253, 60, 0, 164, 0, 60, 0, 162, 0, 60, 0, 60, 0, 60, 0, 224, 255, 60, 0, 39, 0, 36, 0, 39, 0, 60, 0, 60, 0, 60, 0, 4, 255, 60, 0, 60, 0, 60, 0, 105, 254, 60, 0, 163, 0, 60, 0, 60, 0, 60, 0, 225, 255, 60, 0, 165, 0, 60, 0, 60, 0, 60, 0, 229, 255, 60, 0, 143, 5, 60, 0, 14, 32, 11, 6, 14, 32, 60, 0, 14, 32, 254, 7, 14, 32, 60, 0, 14, 32, 255, 7, 14, 32, 60, 0, 242, 9, 60, 0, 243, 9, 60, 0, 251, 9, 60, 0, 241, 10, 60, 0, 56, 168, 60, 0, 249, 11, 60, 0, 7, 216, 221, 223, 60, 0, 7, 216, 222, 223, 60, 0, 7, 216, 223, 223, 60, 0, 7, 216, 224, 223, 60, 0, 63, 14, 60, 0, 219, 23, 60, 0, 56, 216, 255, 222, 60, 0, 160, 32, 60, 0, 161, 32, 60, 0, 162, 32, 60, 0, 163, 32, 60, 0, 164, 32, 60, 0, 165, 32, 60, 0, 166, 32, 60, 0, 167, 32, 60, 0, 169, 32, 60, 0, 60, 0, 60, 0, 230, 255, 60, 0, 170, 32, 60, 0, 171, 32, 60, 0, 172, 32, 60, 0, 173, 32, 60, 0, 174, 32, 60, 0, 175, 32, 60, 0, 176, 32, 60, 0, 177, 32, 60, 0, 178, 32, 60, 0, 179, 32, 60, 0, 180, 32, 60, 0, 181, 32, 60, 0, 182, 32, 60, 0, 183, 32, 60, 0, 184, 32, 60, 0, 185, 32, 60, 0, 14, 32, 59, 216, 176, 220, 14, 32, 60, 0, 186, 32, 60, 0, 187, 32, 60, 0, 188, 32, 60, 0, 189, 32, 60, 0, 190, 32, 60, 0, 191, 32, 60, 0, 168, 32, 60, 0, 14, 32, 252, 253, 14, 32, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 114, 0, 101, 0, 103, 0, 117, 0, 108, 0, 97, 0, 114, 0, 93, 0, 60, 0, 244, 9, 60, 0, 245, 9, 60, 0, 246, 9, 60, 0, 247, 9, 60, 0, 248, 9, 60, 0, 249, 9, 60, 0, 114, 11, 60, 0, 115, 11, 60, 0, 116, 11, 60, 0, 117, 11, 60, 0, 118, 11, 60, 0, 119, 11, 60, 0, 48, 168, 60, 0, 49, 168, 60, 0, 50, 168, 60, 0, 51, 168, 60, 0, 52, 168, 60, 0, 53, 168, 60, 0, 240, 11, 60, 0, 241, 11, 60, 0, 242, 11, 60, 0, 7, 216, 192, 223, 60, 0, 7, 216, 193, 223, 60, 0, 7, 216, 194, 223, 60, 0, 7, 216, 195, 223, 60, 0, 7, 216, 196, 223, 60, 0, 7, 216, 197, 223, 60, 0, 7, 216, 198, 223, 60, 0, 7, 216, 199, 223, 60, 0, 7, 216, 200, 223, 60, 0, 7, 216, 201, 223, 60, 0, 7, 216, 202, 223, 60, 0, 7, 216, 203, 223, 60, 0, 7, 216, 204, 223, 60, 0, 7, 216, 205, 223, 60, 0, 7, 216, 206, 223, 60, 0, 7, 216, 207, 223, 60, 0, 7, 216, 208, 223, 60, 0, 7, 216, 209, 223, 60, 0, 7, 216, 210, 223, 60, 0, 7, 216, 211, 223, 60, 0, 7, 216, 212, 223, 60, 0, 88, 13, 60, 0, 89, 13, 60, 0, 90, 13, 60, 0, 91, 13, 60, 0, 92, 13, 60, 0, 93, 13, 60, 0, 94, 13, 60, 0, 112, 13, 60, 0, 113, 13, 60, 0, 114, 13, 60, 0, 115, 13, 60, 0, 116, 13, 60, 0, 117, 13, 60, 0, 118, 13, 60, 0, 119, 13, 60, 0, 120, 13, 60, 0, 114, 19, 60, 0, 115, 19, 60, 0, 116, 19, 60, 0, 117, 19, 60, 0, 118, 19, 60, 0, 119, 19, 60, 0, 120, 19, 60, 0, 121, 19, 60, 0, 122, 19, 60, 0, 123, 19, 60, 0, 124, 19, 60, 0, 128, 33, 60, 0, 129, 33, 60, 0, 130, 33, 60, 0, 134, 33, 60, 0, 135, 33, 60, 0, 136, 33, 60, 0, 3, 216, 105, 222, 60, 0, 3, 216, 106, 222, 60, 0, 3, 216, 107, 222, 60, 0, 3, 216, 108, 222, 60, 0, 3, 216, 109, 222, 60, 0, 3, 216, 110, 222, 60, 0, 3, 216, 111, 222, 60, 0, 3, 216, 112, 222, 60, 0, 3, 216, 113, 222, 60, 0, 3, 216, 114, 222, 60, 0, 3, 216, 115, 222, 60, 0, 3, 216, 116, 222, 60, 0, 3, 216, 117, 222, 60, 0, 3, 216, 118, 222, 60, 0, 3, 216, 119, 222, 60, 0, 3, 216, 120, 222, 60, 0, 3, 216, 121, 222, 60, 0, 3, 216, 122, 222, 60, 0, 3, 216, 123, 222, 60, 0, 3, 216, 124, 222, 60, 0, 3, 216, 125, 222, 60, 0, 3, 216, 126, 222, 60, 0, 253, 44, 60, 0, 0, 216, 34, 223, 60, 0, 0, 216, 35, 223, 60, 0, 14, 32, 3, 216, 252, 220, 14, 32, 60, 0, 14, 32, 3, 216, 253, 220, 14, 32, 60, 0, 14, 32, 3, 216, 254, 220, 14, 32, 60, 0, 14, 32, 3, 216, 255, 220, 14, 32, 60, 0, 0, 216, 16, 221, 60, 0, 0, 216, 17, 221, 60, 0, 0, 216, 18, 221, 60, 0, 0, 216, 19, 221, 60, 0, 0, 216, 20, 221, 60, 0, 0, 216, 21, 221, 60, 0, 0, 216, 22, 221, 60, 0, 0, 216, 23, 221, 60, 0, 0, 216, 24, 221, 60, 0, 0, 216, 25, 221, 60, 0, 0, 216, 26, 221, 60, 0, 0, 216, 27, 221, 60, 0, 0, 216, 28, 221, 60, 0, 0, 216, 29, 221, 60, 0, 0, 216, 30, 221, 60, 0, 0, 216, 31, 221, 60, 0, 0, 216, 32, 221, 60, 0, 0, 216, 33, 221, 60, 0, 0, 216, 34, 221, 60, 0, 0, 216, 35, 221, 60, 0, 0, 216, 36, 221, 60, 0, 0, 216, 37, 221, 60, 0, 0, 216, 38, 221, 60, 0, 0, 216, 39, 221, 60, 0, 0, 216, 40, 221, 60, 0, 0, 216, 41, 221, 60, 0, 0, 216, 42, 221, 60, 0, 0, 216, 43, 221, 60, 0, 0, 216, 44, 221, 60, 0, 0, 216, 45, 221, 60, 0, 0, 216, 46, 221, 60, 0, 0, 216, 47, 221, 60, 0, 0, 216, 48, 221, 60, 0, 0, 216, 49, 221, 60, 0, 0, 216, 50, 221, 60, 0, 0, 216, 51, 221, 60, 0, 0, 216, 64, 221, 60, 0, 0, 216, 65, 221, 60, 0, 0, 216, 68, 221, 60, 0, 0, 216, 69, 221, 60, 0, 0, 216, 70, 221, 60, 0, 0, 216, 71, 221, 60, 0, 0, 216, 73, 221, 60, 0, 0, 216, 74, 221, 60, 0, 0, 216, 75, 221, 60, 0, 0, 216, 76, 221, 60, 0, 0, 216, 77, 221, 60, 0, 0, 216, 78, 221, 60, 0, 0, 216, 80, 221, 60, 0, 0, 216, 81, 221, 60, 0, 0, 216, 82, 221, 60, 0, 0, 216, 83, 221, 60, 0, 0, 216, 84, 221, 60, 0, 0, 216, 85, 221, 60, 0, 0, 216, 86, 221, 60, 0, 0, 216, 87, 221, 60, 0, 0, 216, 96, 221, 60, 0, 0, 216, 97, 221, 60, 0, 0, 216, 98, 221, 60, 0, 0, 216, 99, 221, 60, 0, 0, 216, 100, 221, 60, 0, 0, 216, 101, 221, 60, 0, 0, 216, 102, 221, 60, 0, 0, 216, 103, 221, 60, 0, 0, 216, 104, 221, 60, 0, 0, 216, 105, 221, 60, 0, 0, 216, 106, 221, 60, 0, 0, 216, 107, 221, 60, 0, 0, 216, 108, 221, 60, 0, 0, 216, 109, 221, 60, 0, 0, 216, 110, 221, 60, 0, 0, 216, 111, 221, 60, 0, 0, 216, 112, 221, 60, 0, 0, 216, 113, 221, 60, 0, 0, 216, 114, 221, 60, 0, 0, 216, 116, 221, 60, 0, 0, 216, 117, 221, 60, 0, 0, 216, 118, 221, 60, 0, 0, 216, 119, 221, 60, 0, 0, 216, 120, 221, 60, 0, 0, 216, 139, 221, 60, 0, 0, 216, 234, 222, 60, 0, 0, 216, 235, 222, 60, 0, 0, 216, 236, 222, 60, 0, 0, 216, 237, 222, 60, 0, 0, 216, 238, 222, 60, 0, 0, 216, 239, 222, 60, 0, 0, 216, 240, 222, 60, 0, 0, 216, 241, 222, 60, 0, 0, 216, 242, 222, 60, 0, 0, 216, 243, 222, 60, 0, 0, 216, 244, 222, 60, 0, 0, 216, 245, 222, 60, 0, 0, 216, 246, 222, 60, 0, 0, 216, 247, 222, 60, 0, 0, 216, 248, 222, 60, 0, 0, 216, 249, 222, 60, 0, 0, 216, 250, 222, 60, 0, 0, 216, 251, 222, 60, 0, 0, 216, 211, 223, 60, 0, 0, 216, 212, 223, 60, 0, 0, 216, 213, 223, 60, 0, 14, 32, 2, 216, 126, 220, 14, 32, 60, 0, 14, 32, 2, 216, 127, 220, 14, 32, 60, 0, 14, 32, 2, 216, 173, 220, 14, 32, 60, 0, 14, 32, 2, 216, 174, 220, 14, 32, 60, 0, 14, 32, 2, 216, 175, 220, 14, 32, 60, 0, 14, 32, 2, 216, 253, 220, 14, 32, 60, 0, 14, 32, 2, 216, 254, 220, 14, 32, 60, 0, 14, 32, 2, 216, 255, 220, 14, 32, 60, 0, 14, 32, 2, 216, 126, 222, 14, 32, 60, 0, 14, 32, 2, 216, 158, 222, 14, 32, 60, 0, 14, 32, 2, 216, 159, 222, 14, 32, 60, 0, 14, 32, 2, 216, 23, 221, 14, 32, 60, 0, 14, 32, 2, 216, 24, 221, 14, 32, 60, 0, 14, 32, 2, 216, 25, 221, 14, 32, 60, 0, 14, 32, 2, 216, 91, 220, 14, 32, 60, 0, 14, 32, 2, 216, 92, 220, 14, 32, 60, 0, 14, 32, 2, 216, 93, 220, 14, 32, 60, 0, 14, 32, 2, 216, 94, 220, 14, 32, 60, 0, 14, 32, 2, 216, 95, 220, 14, 32, 60, 0, 14, 32, 2, 216, 237, 222, 14, 32, 60, 0, 14, 32, 2, 216, 238, 222, 14, 32, 60, 0, 14, 32, 2, 216, 239, 222, 14, 32, 60, 0, 14, 32, 2, 216, 92, 223, 14, 32, 60, 0, 14, 32, 2, 216, 93, 223, 14, 32, 60, 0, 14, 32, 2, 216, 94, 223, 14, 32, 60, 0, 14, 32, 2, 216, 95, 223, 14, 32, 60, 0, 14, 32, 2, 216, 124, 223, 14, 32, 60, 0, 14, 32, 2, 216, 125, 223, 14, 32, 60, 0, 14, 32, 2, 216, 126, 223, 14, 32, 60, 0, 14, 32, 2, 216, 127, 223, 14, 32, 60, 0, 14, 32, 2, 216, 173, 223, 14, 32, 60, 0, 14, 32, 2, 216, 174, 223, 14, 32, 60, 0, 14, 32, 2, 216, 175, 223, 14, 32, 60, 0, 14, 32, 3, 216, 34, 223, 14, 32, 60, 0, 14, 32, 3, 216, 35, 223, 14, 32, 60, 0, 14, 32, 3, 216, 36, 223, 14, 32, 60, 0, 14, 32, 3, 216, 37, 223, 14, 32, 60, 0, 14, 32, 3, 216, 38, 223, 14, 32, 60, 0, 14, 32, 3, 216, 82, 223, 14, 32, 60, 0, 14, 32, 3, 216, 83, 223, 14, 32, 60, 0, 14, 32, 3, 216, 84, 223, 14, 32, 60, 0, 14, 32, 3, 216, 201, 223, 14, 32, 60, 0, 14, 32, 3, 216, 202, 223, 14, 32, 60, 0, 14, 32, 3, 216, 203, 223, 14, 32, 60, 0, 5, 216, 58, 223, 60, 0, 5, 216, 59, 223, 60, 0, 4, 216, 91, 220, 60, 0, 4, 216, 92, 220, 60, 0, 4, 216, 93, 220, 60, 0, 4, 216, 94, 220, 60, 0, 4, 216, 95, 220, 60, 0, 4, 216, 96, 220, 60, 0, 4, 216, 97, 220, 60, 0, 4, 216, 98, 220, 60, 0, 4, 216, 99, 220, 60, 0, 4, 216, 100, 220, 60, 0, 4, 216, 101, 220, 60, 0, 14, 32, 2, 216, 68, 222, 14, 32, 60, 0, 14, 32, 2, 216, 69, 222, 14, 32, 60, 0, 14, 32, 2, 216, 70, 222, 14, 32, 60, 0, 14, 32, 2, 216, 71, 222, 14, 32, 60, 0, 14, 32, 2, 216, 72, 222, 14, 32, 60, 0, 7, 216, 99, 220, 60, 0, 7, 216, 100, 220, 60, 0, 7, 216, 101, 220, 60, 0, 7, 216, 102, 220, 60, 0, 7, 216, 103, 220, 60, 0, 7, 216, 104, 220, 60, 0, 7, 216, 105, 220, 60, 0, 7, 216, 106, 220, 60, 0, 7, 216, 107, 220, 60, 0, 7, 216, 108, 220, 60, 0, 4, 216, 234, 221, 60, 0, 4, 216, 235, 221, 60, 0, 4, 216, 236, 221, 60, 0, 4, 216, 237, 221, 60, 0, 4, 216, 238, 221, 60, 0, 4, 216, 239, 221, 60, 0, 4, 216, 240, 221, 60, 0, 4, 216, 241, 221, 60, 0, 4, 216, 242, 221, 60, 0, 4, 216, 243, 221, 60, 0, 4, 216, 244, 221, 60, 0, 6, 216, 234, 220, 60, 0, 6, 216, 235, 220, 60, 0, 6, 216, 236, 220, 60, 0, 6, 216, 237, 220, 60, 0, 6, 216, 238, 220, 60, 0, 6, 216, 239, 220, 60, 0, 6, 216, 240, 220, 60, 0, 6, 216, 241, 220, 60, 0, 6, 216, 242, 220, 60, 0, 26, 216, 91, 223, 60, 0, 26, 216, 92, 223, 60, 0, 26, 216, 93, 223, 60, 0, 26, 216, 94, 223, 60, 0, 26, 216, 95, 223, 60, 0, 26, 216, 96, 223, 60, 0, 26, 216, 97, 223, 60, 0, 27, 216, 138, 222, 60, 0, 27, 216, 139, 222, 60, 0, 27, 216, 140, 222, 60, 0, 27, 216, 141, 222, 60, 0, 27, 216, 142, 222, 60, 0, 27, 216, 143, 222, 60, 0, 27, 216, 144, 222, 60, 0, 27, 216, 145, 222, 60, 0, 27, 216, 146, 222, 60, 0, 27, 216, 147, 222, 60, 0, 52, 216, 234, 222, 60, 0, 52, 216, 235, 222, 60, 0, 52, 216, 236, 222, 60, 0, 52, 216, 237, 222, 60, 0, 52, 216, 238, 222, 60, 0, 52, 216, 239, 222, 60, 0, 52, 216, 240, 222, 60, 0, 52, 216, 241, 222, 60, 0, 52, 216, 242, 222, 60, 0, 52, 216, 243, 222, 60, 0, 14, 32, 59, 216, 122, 220, 14, 32, 60, 0, 14, 32, 59, 216, 123, 220, 14, 32, 60, 0, 14, 32, 59, 216, 124, 220, 14, 32, 60, 0, 14, 32, 59, 216, 125, 220, 14, 32, 60, 0, 14, 32, 59, 216, 126, 220, 14, 32, 60, 0, 14, 32, 59, 216, 127, 220, 14, 32, 60, 0, 14, 32, 59, 216, 128, 220, 14, 32, 60, 0, 14, 32, 59, 216, 129, 220, 14, 32, 60, 0, 14, 32, 59, 216, 130, 220, 14, 32, 60, 0, 14, 32, 59, 216, 131, 220, 14, 32, 60, 0, 14, 32, 59, 216, 132, 220, 14, 32, 60, 0, 14, 32, 59, 216, 133, 220, 14, 32, 60, 0, 14, 32, 59, 216, 134, 220, 14, 32, 60, 0, 14, 32, 59, 216, 135, 220, 14, 32, 60, 0, 14, 32, 59, 216, 136, 220, 14, 32, 60, 0, 14, 32, 59, 216, 137, 220, 14, 32, 60, 0, 14, 32, 59, 216, 138, 220, 14, 32, 60, 0, 14, 32, 59, 216, 139, 220, 14, 32, 60, 0, 14, 32, 59, 216, 140, 220, 14, 32, 60, 0, 14, 32, 59, 216, 141, 220, 14, 32, 60, 0, 14, 32, 59, 216, 142, 220, 14, 32, 60, 0, 14, 32, 59, 216, 143, 220, 14, 32, 60, 0, 14, 32, 59, 216, 144, 220, 14, 32, 60, 0, 14, 32, 59, 216, 145, 220, 14, 32, 60, 0, 14, 32, 59, 216, 146, 220, 14, 32, 60, 0, 14, 32, 59, 216, 147, 220, 14, 32, 60, 0, 14, 32, 59, 216, 148, 220, 14, 32, 60, 0, 14, 32, 59, 216, 149, 220, 14, 32, 60, 0, 14, 32, 59, 216, 150, 220, 14, 32, 60, 0, 14, 32, 59, 216, 151, 220, 14, 32, 60, 0, 14, 32, 59, 216, 152, 220, 14, 32, 60, 0, 14, 32, 59, 216, 153, 220, 14, 32, 60, 0, 14, 32, 59, 216, 154, 220, 14, 32, 60, 0, 14, 32, 59, 216, 155, 220, 14, 32, 60, 0, 14, 32, 59, 216, 156, 220, 14, 32, 60, 0, 14, 32, 59, 216, 157, 220, 14, 32, 60, 0, 14, 32, 59, 216, 158, 220, 14, 32, 60, 0, 14, 32, 59, 216, 159, 220, 14, 32, 60, 0, 14, 32, 59, 216, 160, 220, 14, 32, 60, 0, 14, 32, 59, 216, 161, 220, 14, 32, 60, 0, 14, 32, 59, 216, 162, 220, 14, 32, 60, 0, 14, 32, 59, 216, 173, 220, 14, 32, 60, 0, 14, 32, 59, 216, 174, 220, 14, 32, 60, 0, 14, 32, 59, 216, 175, 220, 14, 32, 60, 0, 14, 32, 59, 216, 179, 220, 14, 32, 60, 0, 14, 32, 59, 216, 180, 220, 14, 32, 60, 0, 14, 32, 59, 216, 10, 221, 14, 32, 60, 0, 14, 32, 59, 216, 11, 221, 14, 32, 60, 0, 14, 32, 59, 216, 12, 221, 14, 32, 60, 0, 14, 32, 59, 216, 13, 221, 14, 32, 60, 0, 14, 32, 59, 216, 14, 221, 14, 32, 60, 0, 14, 32, 59, 216, 15, 221, 14, 32, 60, 0, 14, 32, 59, 216, 16, 221, 14, 32, 60, 0, 14, 32, 59, 216, 17, 221, 14, 32, 60, 0, 14, 32, 59, 216, 18, 221, 14, 32, 60, 0, 14, 32, 59, 216, 19, 221, 14, 32, 60, 0, 14, 32, 59, 216, 20, 221, 14, 32, 60, 0, 14, 32, 59, 216, 21, 221, 14, 32, 60, 0, 14, 32, 59, 216, 22, 221, 14, 32, 60, 0, 14, 32, 59, 216, 23, 221, 14, 32, 60, 0, 14, 32, 59, 216, 24, 221, 14, 32, 60, 0, 14, 32, 59, 216, 25, 221, 14, 32, 60, 0, 14, 32, 59, 216, 26, 221, 14, 32, 60, 0, 14, 32, 59, 216, 27, 221, 14, 32, 60, 0, 14, 32, 59, 216, 28, 221, 14, 32, 60, 0, 14, 32, 59, 216, 29, 221, 14, 32, 60, 0, 14, 32, 59, 216, 30, 221, 14, 32, 60, 0, 14, 32, 59, 216, 31, 221, 14, 32, 60, 0, 14, 32, 59, 216, 32, 221, 14, 32, 60, 0, 14, 32, 59, 216, 33, 221, 14, 32, 60, 0, 14, 32, 59, 216, 34, 221, 14, 32, 60, 0, 14, 32, 59, 216, 35, 221, 14, 32, 60, 0, 14, 32, 59, 216, 36, 221, 14, 32, 60, 0, 14, 32, 59, 216, 37, 221, 14, 32, 60, 0, 14, 32, 59, 216, 38, 221, 14, 32, 60, 0, 14, 32, 59, 216, 39, 221, 14, 32, 60, 0, 14, 32, 59, 216, 40, 221, 14, 32, 60, 0, 14, 32, 59, 216, 41, 221, 14, 32, 60, 0, 14, 32, 59, 216, 42, 221, 14, 32, 60, 0, 14, 32, 59, 216, 43, 221, 14, 32, 60, 0, 14, 32, 59, 216, 44, 221, 14, 32, 60, 0, 14, 32, 59, 216, 45, 221, 14, 32, 60, 0, 14, 32, 59, 216, 55, 221, 14, 32, 60, 0, 14, 32, 59, 216, 56, 221, 14, 32, 60, 0, 14, 32, 59, 216, 57, 221, 14, 32, 60, 0, 14, 32, 59, 216, 58, 221, 14, 32, 60, 0, 14, 32, 59, 216, 59, 221, 14, 32, 60, 0, 14, 32, 59, 216, 60, 221, 14, 32, 60, 0, 14, 32, 59, 216, 61, 221, 14, 32, 60, 0, 14, 32, 2, 216, 201, 221, 14, 32, 60, 0, 14, 32, 2, 216, 202, 221, 14, 32, 60, 0, 14, 32, 2, 216, 203, 221, 14, 32, 60, 0, 14, 32, 2, 216, 204, 221, 14, 32, 60, 0, 14, 32, 2, 216, 205, 221, 14, 32, 60, 0, 14, 32, 2, 216, 206, 221, 14, 32, 60, 0, 14, 32, 2, 216, 207, 221, 14, 32, 60, 0, 14, 32, 2, 216, 210, 221, 14, 32, 60, 0, 14, 32, 2, 216, 211, 221, 14, 32, 60, 0, 14, 32, 2, 216, 212, 221, 14, 32, 60, 0, 14, 32, 2, 216, 213, 221, 14, 32, 60, 0, 14, 32, 2, 216, 214, 221, 14, 32, 60, 0, 14, 32, 2, 216, 215, 221, 14, 32, 60, 0, 14, 32, 2, 216, 216, 221, 14, 32, 60, 0, 14, 32, 2, 216, 217, 221, 14, 32, 60, 0, 14, 32, 2, 216, 218, 221, 14, 32, 60, 0, 14, 32, 2, 216, 219, 221, 14, 32, 60, 0, 14, 32, 2, 216, 220, 221, 14, 32, 60, 0, 14, 32, 2, 216, 221, 221, 14, 32, 60, 0, 14, 32, 2, 216, 222, 221, 14, 32, 60, 0, 14, 32, 2, 216, 223, 221, 14, 32, 60, 0, 14, 32, 2, 216, 224, 221, 14, 32, 60, 0, 14, 32, 2, 216, 225, 221, 14, 32, 60, 0, 14, 32, 2, 216, 226, 221, 14, 32, 60, 0, 14, 32, 2, 216, 227, 221, 14, 32, 60, 0, 14, 32, 2, 216, 228, 221, 14, 32, 60, 0, 14, 32, 2, 216, 229, 221, 14, 32, 60, 0, 14, 32, 2, 216, 230, 221, 14, 32, 60, 0, 14, 32, 2, 216, 231, 221, 14, 32, 60, 0, 14, 32, 2, 216, 232, 221, 14, 32, 60, 0, 14, 32, 2, 216, 233, 221, 14, 32, 60, 0, 14, 32, 2, 216, 234, 221, 14, 32, 60, 0, 14, 32, 2, 216, 235, 221, 14, 32, 60, 0, 14, 32, 2, 216, 236, 221, 14, 32, 60, 0, 14, 32, 2, 216, 237, 221, 14, 32, 60, 0, 14, 32, 2, 216, 238, 221, 14, 32, 60, 0, 14, 32, 2, 216, 239, 221, 14, 32, 60, 0, 14, 32, 2, 216, 240, 221, 14, 32, 60, 0, 14, 32, 2, 216, 241, 221, 14, 32, 60, 0, 14, 32, 2, 216, 242, 221, 14, 32, 60, 0, 14, 32, 2, 216, 243, 221, 14, 32, 60, 0, 14, 32, 2, 216, 244, 221, 14, 32, 60, 0, 14, 32, 2, 216, 245, 221, 14, 32, 60, 0, 14, 32, 2, 216, 189, 221, 14, 32, 60, 0, 14, 32, 2, 216, 246, 221, 14, 32, 60, 0, 14, 32, 2, 216, 247, 221, 14, 32, 60, 0, 14, 32, 2, 216, 248, 221, 14, 32, 60, 0, 14, 32, 2, 216, 249, 221, 14, 32, 60, 0, 14, 32, 2, 216, 250, 221, 14, 32, 60, 0, 14, 32, 2, 216, 251, 221, 14, 32, 60, 0, 14, 32, 2, 216, 252, 221, 14, 32, 60, 0, 14, 32, 2, 216, 253, 221, 14, 32, 60, 0, 14, 32, 2, 216, 254, 221, 14, 32, 60, 0, 14, 32, 2, 216, 255, 221, 14, 32, 60, 0, 14, 32, 2, 216, 188, 221, 14, 32, 60, 0, 9, 216, 50, 220, 60, 0, 9, 216, 51, 220, 60, 0, 9, 216, 90, 220, 60, 0, 9, 216, 91, 220, 60, 0, 9, 216, 92, 220, 60, 0, 9, 216, 93, 220, 60, 0, 9, 216, 94, 220, 60, 0, 9, 216, 95, 220, 60, 0, 9, 216, 96, 220, 60, 0, 9, 216, 97, 220, 60, 0, 9, 216, 98, 220, 60, 0, 9, 216, 99, 220, 60, 0, 9, 216, 100, 220, 60, 0, 9, 216, 101, 220, 60, 0, 9, 216, 102, 220, 60, 0, 9, 216, 103, 220, 60, 0, 9, 216, 104, 220, 60, 0, 52, 216, 105, 223, 60, 0, 52, 216, 106, 223, 60, 0, 52, 216, 107, 223, 60, 0, 52, 216, 108, 223, 60, 0, 52, 216, 109, 223, 60, 0, 52, 216, 110, 223, 60, 0, 52, 216, 111, 223, 60, 0, 52, 216, 112, 223, 60, 0, 52, 216, 113, 223, 60, 0, 48, 0, 61, 0, 96, 6, 61, 0, 240, 6, 61, 0, 14, 32, 192, 7, 14, 32, 61, 0, 102, 9, 61, 0, 230, 9, 61, 0, 102, 10, 61, 0, 230, 10, 61, 0, 102, 11, 61, 0, 230, 11, 61, 0, 102, 12, 61, 0, 120, 12, 61, 0, 230, 12, 61, 0, 102, 13, 61, 0, 230, 13, 61, 0, 80, 14, 61, 0, 208, 14, 61, 0, 32, 15, 61, 0, 64, 16, 61, 0, 144, 16, 61, 0, 224, 23, 61, 0, 240, 23, 61, 0, 16, 24, 61, 0, 70, 25, 61, 0, 208, 25, 61, 0, 128, 26, 61, 0, 144, 26, 61, 0, 80, 27, 61, 0, 176, 27, 61, 0, 64, 28, 61, 0, 80, 28, 61, 0, 7, 48, 61, 0, 32, 166, 61, 0, 208, 168, 61, 0, 0, 169, 61, 0, 208, 169, 61, 0, 240, 169, 61, 0, 80, 170, 61, 0, 240, 171, 61, 0, 0, 216, 138, 221, 61, 0, 1, 216, 160, 220, 61, 0, 3, 216, 48, 221, 61, 0, 4, 216, 102, 220, 61, 0, 4, 216, 240, 220, 61, 0, 4, 216, 54, 221, 61, 0, 4, 216, 208, 221, 61, 0, 4, 216, 240, 222, 61, 0, 5, 216, 80, 220, 61, 0, 5, 216, 208, 220, 61, 0, 5, 216, 80, 222, 61, 0, 5, 216, 192, 222, 61, 0, 5, 216, 48, 223, 61, 0, 6, 216, 224, 220, 61, 0, 6, 216, 80, 221, 61, 0, 7, 216, 80, 220, 61, 0, 7, 216, 80, 221, 61, 0, 7, 216, 160, 221, 61, 0, 26, 216, 96, 222, 61, 0, 26, 216, 80, 223, 61, 0, 27, 216, 128, 222, 61, 0, 52, 216, 224, 222, 61, 0, 56, 216, 64, 221, 61, 0, 56, 216, 240, 222, 61, 0, 14, 32, 58, 216, 80, 221, 14, 32, 60, 0, 60, 0, 60, 0, 16, 255, 60, 0, 60, 0, 60, 0, 51, 15, 60, 0, 60, 0, 60, 0, 60, 216, 1, 221, 47, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 0, 221, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 88, 51, 47, 0, 185, 112, 60, 0, 60, 0, 60, 0, 53, 216, 206, 223, 61, 0, 53, 216, 216, 223, 61, 0, 53, 216, 226, 223, 61, 0, 53, 216, 236, 223, 61, 0, 53, 216, 246, 223, 61, 0, 62, 216, 240, 223, 60, 0, 60, 0, 60, 0, 234, 36, 61, 0, 255, 36, 61, 0, 60, 216, 11, 221, 61, 0, 60, 216, 12, 221, 60, 0, 60, 0, 60, 0, 112, 32, 60, 0, 60, 0, 60, 0, 128, 32, 60, 0, 60, 0, 60, 0, 137, 33, 47, 0, 14, 32, 68, 32, 59, 216, 3, 221, 14, 32, 60, 0, 49, 0, 61, 0, 97, 6, 61, 0, 241, 6, 61, 0, 14, 32, 193, 7, 14, 32, 61, 0, 103, 9, 61, 0, 231, 9, 61, 0, 103, 10, 61, 0, 231, 10, 61, 0, 103, 11, 61, 0, 231, 11, 61, 0, 103, 12, 61, 0, 121, 12, 61, 0, 124, 12, 61, 0, 231, 12, 61, 0, 103, 13, 61, 0, 231, 13, 61, 0, 81, 14, 61, 0, 209, 14, 61, 0, 33, 15, 61, 0, 65, 16, 61, 0, 145, 16, 61, 0, 105, 19, 61, 0, 225, 23, 61, 0, 241, 23, 61, 0, 17, 24, 61, 0, 71, 25, 61, 0, 209, 25, 61, 0, 218, 25, 61, 0, 129, 26, 61, 0, 145, 26, 61, 0, 81, 27, 61, 0, 177, 27, 61, 0, 65, 28, 61, 0, 81, 28, 61, 0, 33, 48, 61, 0, 33, 166, 61, 0, 209, 168, 61, 0, 1, 169, 61, 0, 209, 169, 61, 0, 241, 169, 61, 0, 81, 170, 61, 0, 241, 171, 61, 0, 0, 216, 7, 221, 61, 0, 0, 216, 66, 221, 61, 0, 0, 216, 88, 221, 61, 0, 0, 216, 89, 221, 61, 0, 0, 216, 90, 221, 61, 0, 0, 216, 225, 222, 61, 0, 0, 216, 32, 223, 61, 0, 0, 216, 209, 223, 61, 0, 1, 216, 161, 220, 61, 0, 14, 32, 2, 216, 88, 220, 14, 32, 61, 0, 14, 32, 2, 216, 121, 220, 14, 32, 61, 0, 14, 32, 2, 216, 167, 220, 14, 32, 61, 0, 14, 32, 2, 216, 251, 220, 14, 32, 61, 0, 14, 32, 2, 216, 22, 221, 14, 32, 61, 0, 14, 32, 2, 216, 192, 221, 14, 32, 61, 0, 14, 32, 2, 216, 64, 222, 14, 32, 61, 0, 14, 32, 2, 216, 125, 222, 14, 32, 61, 0, 14, 32, 2, 216, 157, 222, 14, 32, 61, 0, 14, 32, 2, 216, 235, 222, 14, 32, 61, 0, 14, 32, 2, 216, 88, 223, 14, 32, 61, 0, 14, 32, 2, 216, 120, 223, 14, 32, 61, 0, 14, 32, 2, 216, 169, 223, 14, 32, 61, 0, 14, 32, 3, 216, 250, 220, 14, 32, 61, 0, 3, 216, 49, 221, 61, 0, 3, 216, 96, 222, 61, 0, 14, 32, 3, 216, 29, 223, 14, 32, 61, 0, 14, 32, 3, 216, 81, 223, 14, 32, 61, 0, 14, 32, 3, 216, 197, 223, 14, 32, 61, 0, 4, 216, 82, 220, 61, 0, 4, 216, 103, 220, 61, 0, 4, 216, 241, 220, 61, 0, 4, 216, 55, 221, 61, 0, 4, 216, 209, 221, 61, 0, 4, 216, 225, 221, 61, 0, 4, 216, 241, 222, 61, 0, 5, 216, 81, 220, 61, 0, 5, 216, 209, 220, 61, 0, 5, 216, 81, 222, 61, 0, 5, 216, 193, 222, 61, 0, 5, 216, 49, 223, 61, 0, 6, 216, 225, 220, 61, 0, 6, 216, 81, 221, 61, 0, 7, 216, 81, 220, 61, 0, 7, 216, 90, 220, 61, 0, 7, 216, 81, 221, 61, 0, 7, 216, 161, 221, 61, 0, 9, 216, 21, 220, 61, 0, 9, 216, 30, 220, 61, 0, 9, 216, 44, 220, 61, 0, 9, 216, 52, 220, 61, 0, 9, 216, 79, 220, 61, 0, 9, 216, 88, 220, 61, 0, 26, 216, 97, 222, 61, 0, 26, 216, 81, 223, 61, 0, 27, 216, 129, 222, 61, 0, 52, 216, 225, 222, 61, 0, 52, 216, 96, 223, 61, 0, 52, 216, 114, 223, 61, 0, 52, 216, 119, 223, 61, 0, 56, 216, 65, 221, 61, 0, 56, 216, 241, 222, 61, 0, 14, 32, 58, 216, 199, 220, 14, 32, 61, 0, 14, 32, 58, 216, 81, 221, 14, 32, 61, 0, 14, 32, 59, 216, 113, 220, 14, 32, 61, 0, 14, 32, 59, 216, 163, 220, 14, 32, 61, 0, 14, 32, 59, 216, 1, 221, 14, 32, 60, 0, 60, 0, 60, 0, 17, 255, 60, 0, 60, 0, 60, 0, 42, 15, 60, 0, 60, 0, 60, 0, 60, 216, 2, 221, 47, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 136, 36, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 145, 36, 47, 0, 14, 32, 58, 216, 80, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 233, 51, 47, 0, 14, 32, 58, 216, 80, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 201, 50, 47, 0, 14, 32, 58, 216, 80, 221, 8, 103, 14, 32, 60, 0, 60, 0, 60, 0, 98, 51, 47, 0, 14, 32, 58, 216, 80, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 146, 36, 47, 0, 14, 32, 59, 216, 1, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 234, 51, 47, 0, 14, 32, 59, 216, 1, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 202, 50, 47, 0, 14, 32, 59, 216, 1, 221, 8, 103, 14, 32, 60, 0, 60, 0, 60, 0, 99, 51, 47, 0, 14, 32, 59, 216, 1, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 147, 36, 47, 0, 14, 32, 59, 216, 2, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 235, 51, 47, 0, 14, 32, 59, 216, 2, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 203, 50, 47, 0, 14, 32, 59, 216, 2, 221, 8, 103, 14, 32, 60, 0, 60, 0, 60, 0, 100, 51, 47, 0, 14, 32, 59, 216, 2, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 148, 36, 47, 0, 14, 32, 59, 216, 3, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 236, 51, 47, 0, 14, 32, 59, 216, 3, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 101, 51, 47, 0, 14, 32, 59, 216, 3, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 149, 36, 47, 0, 14, 32, 59, 216, 4, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 237, 51, 47, 0, 14, 32, 59, 216, 4, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 102, 51, 47, 0, 14, 32, 59, 216, 4, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 150, 36, 47, 0, 14, 32, 59, 216, 5, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 238, 51, 47, 0, 14, 32, 59, 216, 5, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 103, 51, 47, 0, 14, 32, 59, 216, 5, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 151, 36, 47, 0, 14, 32, 59, 216, 6, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 239, 51, 47, 0, 14, 32, 59, 216, 6, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 104, 51, 47, 0, 14, 32, 59, 216, 6, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 152, 36, 47, 0, 14, 32, 59, 216, 7, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 240, 51, 47, 0, 14, 32, 59, 216, 7, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 105, 51, 47, 0, 14, 32, 59, 216, 7, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 153, 36, 47, 0, 14, 32, 59, 216, 8, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 241, 51, 47, 0, 14, 32, 59, 216, 8, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 106, 51, 47, 0, 14, 32, 59, 216, 8, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 154, 36, 47, 0, 14, 32, 59, 216, 9, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 242, 51, 47, 0, 14, 32, 59, 216, 9, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 107, 51, 47, 0, 14, 32, 59, 216, 9, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 224, 51, 47, 0, 229, 101, 60, 0, 60, 0, 60, 0, 192, 50, 47, 0, 8, 103, 60, 0, 60, 0, 60, 0, 89, 51, 47, 0, 185, 112, 60, 0, 60, 0, 60, 0, 53, 216, 207, 223, 61, 0, 53, 216, 217, 223, 61, 0, 53, 216, 227, 223, 61, 0, 53, 216, 237, 223, 61, 0, 53, 216, 247, 223, 61, 0, 62, 216, 241, 223, 60, 0, 60, 0, 60, 0, 96, 36, 61, 0, 245, 36, 61, 0, 118, 39, 61, 0, 128, 39, 61, 0, 138, 39, 60, 0, 60, 0, 60, 0, 105, 36, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 61, 0, 254, 36, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 61, 0, 127, 39, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 61, 0, 137, 39, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 61, 0, 147, 39, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 61, 0, 72, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 60, 0, 60, 0, 60, 0, 106, 36, 47, 0, 14, 32, 59, 216, 1, 221, 14, 32, 61, 0, 235, 36, 47, 0, 14, 32, 59, 216, 1, 221, 14, 32, 60, 0, 60, 0, 60, 0, 107, 36, 47, 0, 14, 32, 59, 216, 2, 221, 14, 32, 61, 0, 236, 36, 47, 0, 14, 32, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 108, 36, 47, 0, 14, 32, 59, 216, 3, 221, 14, 32, 61, 0, 237, 36, 47, 0, 14, 32, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 109, 36, 47, 0, 14, 32, 59, 216, 4, 221, 14, 32, 61, 0, 238, 36, 47, 0, 14, 32, 59, 216, 4, 221, 14, 32, 60, 0, 60, 0, 60, 0, 110, 36, 47, 0, 14, 32, 59, 216, 5, 221, 14, 32, 61, 0, 239, 36, 47, 0, 14, 32, 59, 216, 5, 221, 14, 32, 60, 0, 60, 0, 60, 0, 111, 36, 47, 0, 14, 32, 59, 216, 6, 221, 14, 32, 61, 0, 240, 36, 47, 0, 14, 32, 59, 216, 6, 221, 14, 32, 60, 0, 60, 0, 60, 0, 112, 36, 47, 0, 14, 32, 59, 216, 7, 221, 14, 32, 61, 0, 241, 36, 47, 0, 14, 32, 59, 216, 7, 221, 14, 32, 60, 0, 60, 0, 60, 0, 113, 36, 47, 0, 14, 32, 59, 216, 8, 221, 14, 32, 61, 0, 242, 36, 47, 0, 14, 32, 59, 216, 8, 221, 14, 32, 60, 0, 60, 0, 60, 0, 114, 36, 47, 0, 14, 32, 59, 216, 9, 221, 14, 32, 61, 0, 243, 36, 47, 0, 14, 32, 59, 216, 9, 221, 14, 32, 60, 0, 60, 0, 60, 0, 185, 0, 60, 0, 60, 0, 60, 0, 129, 32, 60, 0, 60, 0, 60, 0, 60, 216, 164, 221, 47, 0, 14, 32, 59, 216, 2, 221, 58, 216, 80, 221, 80, 0, 14, 32, 60, 0, 60, 0, 60, 0, 95, 33, 47, 0, 68, 32, 60, 0, 60, 0, 60, 0, 82, 33, 47, 0, 14, 32, 68, 32, 59, 216, 1, 221, 58, 216, 80, 221, 14, 32, 60, 0, 60, 0, 60, 0, 189, 0, 47, 0, 14, 32, 68, 32, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 83, 33, 47, 0, 14, 32, 68, 32, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 188, 0, 47, 0, 14, 32, 68, 32, 59, 216, 4, 221, 14, 32, 60, 0, 60, 0, 60, 0, 85, 33, 47, 0, 14, 32, 68, 32, 59, 216, 5, 221, 14, 32, 60, 0, 60, 0, 60, 0, 89, 33, 47, 0, 14, 32, 68, 32, 59, 216, 6, 221, 14, 32, 60, 0, 60, 0, 60, 0, 80, 33, 47, 0, 14, 32, 68, 32, 59, 216, 7, 221, 14, 32, 60, 0, 60, 0, 60, 0, 91, 33, 47, 0, 14, 32, 68, 32, 59, 216, 8, 221, 14, 32, 60, 0, 60, 0, 60, 0, 81, 33, 47, 0, 14, 32, 68, 32, 59, 216, 9, 221, 14, 32, 60, 0, 60, 0, 27, 216, 148, 222, 61, 0, 14, 32, 59, 216, 177, 220, 14, 32, 60, 0, 50, 0, 61, 0, 98, 6, 61, 0, 242, 6, 61, 0, 14, 32, 194, 7, 14, 32, 61, 0, 104, 9, 61, 0, 232, 9, 61, 0, 104, 10, 61, 0, 232, 10, 61, 0, 104, 11, 61, 0, 232, 11, 61, 0, 104, 12, 61, 0, 122, 12, 61, 0, 125, 12, 61, 0, 232, 12, 61, 0, 104, 13, 61, 0, 232, 13, 61, 0, 82, 14, 61, 0, 210, 14, 61, 0, 34, 15, 61, 0, 66, 16, 61, 0, 146, 16, 61, 0, 106, 19, 61, 0, 226, 23, 61, 0, 242, 23, 61, 0, 18, 24, 61, 0, 72, 25, 61, 0, 210, 25, 61, 0, 130, 26, 61, 0, 146, 26, 61, 0, 82, 27, 61, 0, 178, 27, 61, 0, 66, 28, 61, 0, 82, 28, 61, 0, 34, 48, 61, 0, 34, 166, 61, 0, 210, 168, 61, 0, 2, 169, 61, 0, 210, 169, 61, 0, 242, 169, 61, 0, 82, 170, 61, 0, 242, 171, 61, 0, 0, 216, 8, 221, 61, 0, 0, 216, 91, 221, 61, 0, 0, 216, 92, 221, 61, 0, 0, 216, 93, 221, 61, 0, 0, 216, 94, 221, 61, 0, 0, 216, 226, 222, 61, 0, 0, 216, 210, 223, 61, 0, 1, 216, 162, 220, 61, 0, 14, 32, 2, 216, 89, 220, 14, 32, 61, 0, 14, 32, 2, 216, 122, 220, 14, 32, 61, 0, 14, 32, 2, 216, 168, 220, 14, 32, 61, 0, 14, 32, 2, 216, 26, 221, 14, 32, 61, 0, 14, 32, 2, 216, 193, 221, 14, 32, 61, 0, 14, 32, 2, 216, 65, 222, 14, 32, 61, 0, 14, 32, 2, 216, 89, 223, 14, 32, 61, 0, 14, 32, 2, 216, 121, 223, 14, 32, 61, 0, 14, 32, 2, 216, 170, 223, 14, 32, 61, 0, 3, 216, 50, 221, 61, 0, 3, 216, 97, 222, 61, 0, 14, 32, 3, 216, 30, 223, 14, 32, 61, 0, 14, 32, 3, 216, 198, 223, 14, 32, 61, 0, 4, 216, 83, 220, 61, 0, 4, 216, 104, 220, 61, 0, 4, 216, 242, 220, 61, 0, 4, 216, 56, 221, 61, 0, 4, 216, 210, 221, 61, 0, 4, 216, 226, 221, 61, 0, 4, 216, 242, 222, 61, 0, 5, 216, 82, 220, 61, 0, 5, 216, 210, 220, 61, 0, 5, 216, 82, 222, 61, 0, 5, 216, 194, 222, 61, 0, 5, 216, 50, 223, 61, 0, 6, 216, 226, 220, 61, 0, 6, 216, 82, 221, 61, 0, 7, 216, 82, 220, 61, 0, 7, 216, 91, 220, 61, 0, 7, 216, 82, 221, 61, 0, 7, 216, 162, 221, 61, 0, 9, 216, 0, 220, 61, 0, 9, 216, 22, 220, 61, 0, 9, 216, 31, 220, 61, 0, 9, 216, 35, 220, 61, 0, 9, 216, 45, 220, 61, 0, 9, 216, 53, 220, 61, 0, 9, 216, 74, 220, 61, 0, 9, 216, 80, 220, 61, 0, 9, 216, 86, 220, 61, 0, 9, 216, 89, 220, 61, 0, 26, 216, 98, 222, 61, 0, 26, 216, 82, 223, 61, 0, 27, 216, 130, 222, 61, 0, 52, 216, 226, 222, 61, 0, 52, 216, 97, 223, 61, 0, 52, 216, 115, 223, 61, 0, 56, 216, 66, 221, 61, 0, 56, 216, 242, 222, 61, 0, 14, 32, 58, 216, 200, 220, 14, 32, 61, 0, 14, 32, 58, 216, 82, 221, 14, 32, 61, 0, 14, 32, 59, 216, 114, 220, 14, 32, 61, 0, 14, 32, 59, 216, 164, 220, 14, 32, 61, 0, 14, 32, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 18, 255, 60, 0, 60, 0, 60, 0, 43, 15, 60, 0, 60, 0, 60, 0, 60, 216, 3, 221, 47, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 137, 36, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 155, 36, 47, 0, 14, 32, 58, 216, 80, 221, 39, 0, 46, 0, 39, 0, 14, 32, 60, 0, 60, 0, 60, 0, 243, 51, 47, 0, 14, 32, 58, 216, 80, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 108, 51, 47, 0, 14, 32, 58, 216, 80, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 244, 51, 47, 0, 14, 32, 59, 216, 1, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 109, 51, 47, 0, 14, 32, 59, 216, 1, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 245, 51, 47, 0, 14, 32, 59, 216, 2, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 110, 51, 47, 0, 14, 32, 59, 216, 2, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 246, 51, 47, 0, 14, 32, 59, 216, 3, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 111, 51, 47, 0, 14, 32, 59, 216, 3, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 247, 51, 47, 0, 14, 32, 59, 216, 4, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 112, 51, 47, 0, 14, 32, 59, 216, 4, 221, 185, 112, 14, 32, 60, 0, 60, 0, 60, 0, 248, 51, 47, 0, 14, 32, 59, 216, 5, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 249, 51, 47, 0, 14, 32, 59, 216, 6, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 250, 51, 47, 0, 14, 32, 59, 216, 7, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 251, 51, 47, 0, 14, 32, 59, 216, 8, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 252, 51, 47, 0, 14, 32, 59, 216, 9, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 225, 51, 47, 0, 229, 101, 60, 0, 60, 0, 60, 0, 193, 50, 47, 0, 8, 103, 60, 0, 60, 0, 60, 0, 90, 51, 47, 0, 185, 112, 60, 0, 60, 0, 60, 0, 53, 216, 208, 223, 61, 0, 53, 216, 218, 223, 61, 0, 53, 216, 228, 223, 61, 0, 53, 216, 238, 223, 61, 0, 53, 216, 248, 223, 61, 0, 62, 216, 242, 223, 60, 0, 60, 0, 60, 0, 97, 36, 61, 0, 246, 36, 61, 0, 119, 39, 61, 0, 129, 39, 61, 0, 139, 39, 60, 0, 60, 0, 60, 0, 115, 36, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 61, 0, 244, 36, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 61, 0, 73, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 60, 0, 60, 0, 60, 0, 81, 50, 47, 0, 14, 32, 59, 216, 1, 221, 14, 32, 60, 0, 60, 0, 60, 0, 82, 50, 47, 0, 14, 32, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 83, 50, 47, 0, 14, 32, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 84, 50, 47, 0, 14, 32, 59, 216, 4, 221, 14, 32, 60, 0, 60, 0, 60, 0, 85, 50, 47, 0, 14, 32, 59, 216, 5, 221, 14, 32, 60, 0, 60, 0, 60, 0, 86, 50, 47, 0, 14, 32, 59, 216, 6, 221, 14, 32, 60, 0, 60, 0, 60, 0, 87, 50, 47, 0, 14, 32, 59, 216, 7, 221, 14, 32, 60, 0, 60, 0, 60, 0, 88, 50, 47, 0, 14, 32, 59, 216, 8, 221, 14, 32, 60, 0, 60, 0, 60, 0, 89, 50, 47, 0, 14, 32, 59, 216, 9, 221, 14, 32, 60, 0, 60, 0, 60, 0, 178, 0, 60, 0, 60, 0, 60, 0, 130, 32, 60, 0, 60, 0, 60, 0, 60, 216, 162, 221, 47, 0, 14, 32, 59, 216, 2, 221, 39, 0, 46, 0, 39, 0, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 60, 216, 157, 221, 47, 0, 75, 0, 60, 0, 60, 0, 60, 0, 60, 216, 156, 221, 47, 0, 110, 0, 100, 0, 39, 0, 32, 0, 39, 0, 83, 0, 99, 0, 114, 0, 60, 0, 60, 0, 60, 0, 84, 33, 47, 0, 14, 32, 68, 32, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 86, 33, 47, 0, 14, 32, 68, 32, 59, 216, 5, 221, 14, 32, 60, 0, 60, 0, 27, 216, 149, 222, 61, 0, 14, 32, 59, 216, 178, 220, 14, 32, 61, 0, 14, 32, 59, 216, 47, 221, 14, 32, 60, 0, 51, 0, 61, 0, 99, 6, 61, 0, 243, 6, 61, 0, 14, 32, 195, 7, 14, 32, 61, 0, 105, 9, 61, 0, 233, 9, 61, 0, 105, 10, 61, 0, 233, 10, 61, 0, 105, 11, 61, 0, 233, 11, 61, 0, 105, 12, 61, 0, 123, 12, 61, 0, 126, 12, 61, 0, 233, 12, 61, 0, 105, 13, 61, 0, 233, 13, 61, 0, 83, 14, 61, 0, 211, 14, 61, 0, 35, 15, 61, 0, 67, 16, 61, 0, 147, 16, 61, 0, 107, 19, 61, 0, 227, 23, 61, 0, 243, 23, 61, 0, 19, 24, 61, 0, 73, 25, 61, 0, 211, 25, 61, 0, 131, 26, 61, 0, 147, 26, 61, 0, 83, 27, 61, 0, 179, 27, 61, 0, 67, 28, 61, 0, 83, 28, 61, 0, 35, 48, 61, 0, 35, 166, 61, 0, 211, 168, 61, 0, 3, 169, 61, 0, 211, 169, 61, 0, 243, 169, 61, 0, 83, 170, 61, 0, 243, 171, 61, 0, 0, 216, 9, 221, 61, 0, 0, 216, 227, 222, 61, 0, 1, 216, 163, 220, 61, 0, 14, 32, 2, 216, 90, 220, 14, 32, 61, 0, 14, 32, 2, 216, 123, 220, 14, 32, 61, 0, 14, 32, 2, 216, 169, 220, 14, 32, 61, 0, 14, 32, 2, 216, 27, 221, 14, 32, 61, 0, 14, 32, 2, 216, 194, 221, 14, 32, 61, 0, 14, 32, 2, 216, 66, 222, 14, 32, 61, 0, 14, 32, 2, 216, 90, 223, 14, 32, 61, 0, 14, 32, 2, 216, 122, 223, 14, 32, 61, 0, 14, 32, 2, 216, 171, 223, 14, 32, 61, 0, 3, 216, 51, 221, 61, 0, 3, 216, 98, 222, 61, 0, 14, 32, 3, 216, 31, 223, 14, 32, 61, 0, 14, 32, 3, 216, 199, 223, 14, 32, 61, 0, 4, 216, 84, 220, 61, 0, 4, 216, 105, 220, 61, 0, 4, 216, 243, 220, 61, 0, 4, 216, 57, 221, 61, 0, 4, 216, 211, 221, 61, 0, 4, 216, 227, 221, 61, 0, 4, 216, 243, 222, 61, 0, 5, 216, 83, 220, 61, 0, 5, 216, 211, 220, 61, 0, 5, 216, 83, 222, 61, 0, 5, 216, 195, 222, 61, 0, 5, 216, 51, 223, 61, 0, 6, 216, 227, 220, 61, 0, 6, 216, 83, 221, 61, 0, 7, 216, 83, 220, 61, 0, 7, 216, 92, 220, 61, 0, 7, 216, 83, 221, 61, 0, 7, 216, 163, 221, 61, 0, 9, 216, 1, 220, 61, 0, 9, 216, 8, 220, 61, 0, 9, 216, 23, 220, 61, 0, 9, 216, 32, 220, 61, 0, 9, 216, 36, 220, 61, 0, 9, 216, 37, 220, 61, 0, 9, 216, 46, 220, 61, 0, 9, 216, 47, 220, 61, 0, 9, 216, 54, 220, 61, 0, 9, 216, 55, 220, 61, 0, 9, 216, 58, 220, 61, 0, 9, 216, 59, 220, 61, 0, 9, 216, 75, 220, 61, 0, 9, 216, 81, 220, 61, 0, 9, 216, 87, 220, 61, 0, 26, 216, 99, 222, 61, 0, 26, 216, 83, 223, 61, 0, 27, 216, 131, 222, 61, 0, 52, 216, 227, 222, 61, 0, 52, 216, 98, 223, 61, 0, 52, 216, 116, 223, 61, 0, 56, 216, 67, 221, 61, 0, 56, 216, 243, 222, 61, 0, 14, 32, 58, 216, 201, 220, 14, 32, 61, 0, 14, 32, 58, 216, 83, 221, 14, 32, 61, 0, 14, 32, 59, 216, 115, 220, 14, 32, 61, 0, 14, 32, 59, 216, 165, 220, 14, 32, 61, 0, 14, 32, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 19, 255, 60, 0, 60, 0, 60, 0, 44, 15, 60, 0, 60, 0, 60, 0, 60, 216, 4, 221, 47, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 138, 36, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 253, 51, 47, 0, 14, 32, 58, 216, 80, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 254, 51, 47, 0, 14, 32, 59, 216, 1, 221, 229, 101, 14, 32, 60, 0, 60, 0, 60, 0, 226, 51, 47, 0, 229, 101, 60, 0, 60, 0, 60, 0, 194, 50, 47, 0, 8, 103, 60, 0, 60, 0, 60, 0, 91, 51, 47, 0, 185, 112, 60, 0, 60, 0, 60, 0, 53, 216, 209, 223, 61, 0, 53, 216, 219, 223, 61, 0, 53, 216, 229, 223, 61, 0, 53, 216, 239, 223, 61, 0, 53, 216, 249, 223, 61, 0, 62, 216, 243, 223, 60, 0, 60, 0, 60, 0, 98, 36, 61, 0, 247, 36, 61, 0, 120, 39, 61, 0, 130, 39, 61, 0, 140, 39, 60, 0, 60, 0, 60, 0, 74, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 61, 0, 90, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 60, 0, 60, 0, 60, 0, 91, 50, 47, 0, 14, 32, 59, 216, 1, 221, 14, 32, 60, 0, 60, 0, 60, 0, 92, 50, 47, 0, 14, 32, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 93, 50, 47, 0, 14, 32, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 94, 50, 47, 0, 14, 32, 59, 216, 4, 221, 14, 32, 60, 0, 60, 0, 60, 0, 95, 50, 47, 0, 14, 32, 59, 216, 5, 221, 14, 32, 60, 0, 60, 0, 60, 0, 177, 50, 47, 0, 14, 32, 59, 216, 6, 221, 14, 32, 60, 0, 60, 0, 60, 0, 178, 50, 47, 0, 14, 32, 59, 216, 7, 221, 14, 32, 60, 0, 60, 0, 60, 0, 179, 50, 47, 0, 14, 32, 59, 216, 8, 221, 14, 32, 60, 0, 60, 0, 60, 0, 180, 50, 47, 0, 14, 32, 59, 216, 9, 221, 14, 32, 60, 0, 60, 0, 60, 0, 179, 0, 60, 0, 60, 0, 60, 0, 131, 32, 60, 0, 60, 0, 60, 0, 60, 216, 155, 221, 47, 0, 68, 0, 60, 0, 60, 0, 60, 0, 190, 0, 47, 0, 14, 32, 68, 32, 59, 216, 4, 221, 14, 32, 60, 0, 60, 0, 60, 0, 87, 33, 47, 0, 14, 32, 68, 32, 59, 216, 5, 221, 14, 32, 60, 0, 60, 0, 60, 0, 92, 33, 47, 0, 14, 32, 68, 32, 59, 216, 8, 221, 14, 32, 60, 0, 60, 0, 27, 216, 150, 222, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 48, 221, 14, 32, 60, 0, 52, 0, 61, 0, 100, 6, 61, 0, 244, 6, 61, 0, 14, 32, 196, 7, 14, 32, 61, 0, 106, 9, 61, 0, 234, 9, 61, 0, 106, 10, 61, 0, 234, 10, 61, 0, 106, 11, 61, 0, 234, 11, 61, 0, 106, 12, 61, 0, 234, 12, 61, 0, 106, 13, 61, 0, 234, 13, 61, 0, 84, 14, 61, 0, 212, 14, 61, 0, 36, 15, 61, 0, 68, 16, 61, 0, 148, 16, 61, 0, 108, 19, 61, 0, 228, 23, 61, 0, 244, 23, 61, 0, 20, 24, 61, 0, 74, 25, 61, 0, 212, 25, 61, 0, 132, 26, 61, 0, 148, 26, 61, 0, 84, 27, 61, 0, 180, 27, 61, 0, 68, 28, 61, 0, 84, 28, 61, 0, 36, 48, 61, 0, 36, 166, 61, 0, 212, 168, 61, 0, 4, 169, 61, 0, 212, 169, 61, 0, 244, 169, 61, 0, 84, 170, 61, 0, 244, 171, 61, 0, 0, 216, 10, 221, 61, 0, 0, 216, 228, 222, 61, 0, 1, 216, 164, 220, 61, 0, 14, 32, 2, 216, 124, 220, 14, 32, 61, 0, 14, 32, 2, 216, 170, 220, 14, 32, 61, 0, 14, 32, 2, 216, 171, 220, 14, 32, 61, 0, 14, 32, 2, 216, 195, 221, 14, 32, 61, 0, 14, 32, 2, 216, 67, 222, 14, 32, 61, 0, 14, 32, 2, 216, 91, 223, 14, 32, 61, 0, 14, 32, 2, 216, 123, 223, 14, 32, 61, 0, 14, 32, 2, 216, 172, 223, 14, 32, 61, 0, 3, 216, 52, 221, 61, 0, 3, 216, 99, 222, 61, 0, 14, 32, 3, 216, 32, 223, 14, 32, 61, 0, 14, 32, 3, 216, 200, 223, 14, 32, 61, 0, 4, 216, 85, 220, 61, 0, 4, 216, 106, 220, 61, 0, 4, 216, 244, 220, 61, 0, 4, 216, 58, 221, 61, 0, 4, 216, 212, 221, 61, 0, 4, 216, 228, 221, 61, 0, 4, 216, 244, 222, 61, 0, 5, 216, 84, 220, 61, 0, 5, 216, 212, 220, 61, 0, 5, 216, 84, 222, 61, 0, 5, 216, 196, 222, 61, 0, 5, 216, 52, 223, 61, 0, 6, 216, 228, 220, 61, 0, 6, 216, 84, 221, 61, 0, 7, 216, 84, 220, 61, 0, 7, 216, 93, 220, 61, 0, 7, 216, 84, 221, 61, 0, 7, 216, 164, 221, 61, 0, 9, 216, 2, 220, 61, 0, 9, 216, 9, 220, 61, 0, 9, 216, 15, 220, 61, 0, 9, 216, 24, 220, 61, 0, 9, 216, 33, 220, 61, 0, 9, 216, 38, 220, 61, 0, 9, 216, 48, 220, 61, 0, 9, 216, 56, 220, 61, 0, 9, 216, 60, 220, 61, 0, 9, 216, 61, 220, 61, 0, 9, 216, 62, 220, 61, 0, 9, 216, 63, 220, 61, 0, 9, 216, 76, 220, 61, 0, 9, 216, 82, 220, 61, 0, 9, 216, 83, 220, 61, 0, 9, 216, 105, 220, 61, 0, 26, 216, 100, 222, 61, 0, 26, 216, 84, 223, 61, 0, 27, 216, 132, 222, 61, 0, 52, 216, 228, 222, 61, 0, 52, 216, 99, 223, 61, 0, 52, 216, 117, 223, 61, 0, 56, 216, 68, 221, 61, 0, 56, 216, 244, 222, 61, 0, 14, 32, 58, 216, 202, 220, 14, 32, 61, 0, 14, 32, 58, 216, 84, 221, 14, 32, 61, 0, 14, 32, 59, 216, 116, 220, 14, 32, 61, 0, 14, 32, 59, 216, 166, 220, 14, 32, 61, 0, 14, 32, 59, 216, 4, 221, 14, 32, 60, 0, 60, 0, 60, 0, 20, 255, 60, 0, 60, 0, 60, 0, 45, 15, 60, 0, 60, 0, 60, 0, 60, 216, 5, 221, 47, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 139, 36, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 227, 51, 47, 0, 229, 101, 60, 0, 60, 0, 60, 0, 195, 50, 47, 0, 8, 103, 60, 0, 60, 0, 60, 0, 92, 51, 47, 0, 185, 112, 60, 0, 60, 0, 60, 0, 53, 216, 210, 223, 61, 0, 53, 216, 220, 223, 61, 0, 53, 216, 230, 223, 61, 0, 53, 216, 240, 223, 61, 0, 53, 216, 250, 223, 61, 0, 62, 216, 244, 223, 60, 0, 60, 0, 60, 0, 99, 36, 61, 0, 248, 36, 61, 0, 121, 39, 61, 0, 131, 39, 61, 0, 141, 39, 60, 0, 60, 0, 60, 0, 75, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 61, 0, 181, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 60, 0, 60, 0, 60, 0, 182, 50, 47, 0, 14, 32, 59, 216, 1, 221, 14, 32, 60, 0, 60, 0, 60, 0, 183, 50, 47, 0, 14, 32, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 184, 50, 47, 0, 14, 32, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 185, 50, 47, 0, 14, 32, 59, 216, 4, 221, 14, 32, 60, 0, 60, 0, 60, 0, 186, 50, 47, 0, 14, 32, 59, 216, 5, 221, 14, 32, 60, 0, 60, 0, 60, 0, 187, 50, 47, 0, 14, 32, 59, 216, 6, 221, 14, 32, 60, 0, 60, 0, 60, 0, 188, 50, 47, 0, 14, 32, 59, 216, 7, 221, 14, 32, 60, 0, 60, 0, 60, 0, 189, 50, 47, 0, 14, 32, 59, 216, 8, 221, 14, 32, 60, 0, 60, 0, 60, 0, 190, 50, 47, 0, 14, 32, 59, 216, 9, 221, 14, 32, 60, 0, 60, 0, 60, 0, 116, 32, 60, 0, 60, 0, 60, 0, 132, 32, 60, 0, 60, 0, 60, 0, 60, 216, 158, 221, 47, 0, 75, 0, 60, 0, 60, 0, 60, 0, 88, 33, 47, 0, 14, 32, 68, 32, 59, 216, 5, 221, 14, 32, 60, 0, 60, 0, 14, 32, 59, 216, 49, 221, 14, 32, 60, 0, 53, 0, 61, 0, 101, 6, 61, 0, 245, 6, 61, 0, 14, 32, 197, 7, 14, 32, 61, 0, 107, 9, 61, 0, 235, 9, 61, 0, 107, 10, 61, 0, 235, 10, 61, 0, 107, 11, 61, 0, 235, 11, 61, 0, 107, 12, 61, 0, 235, 12, 61, 0, 107, 13, 61, 0, 235, 13, 61, 0, 85, 14, 61, 0, 213, 14, 61, 0, 37, 15, 61, 0, 69, 16, 61, 0, 149, 16, 61, 0, 109, 19, 61, 0, 229, 23, 61, 0, 245, 23, 61, 0, 21, 24, 61, 0, 75, 25, 61, 0, 213, 25, 61, 0, 133, 26, 61, 0, 149, 26, 61, 0, 85, 27, 61, 0, 181, 27, 61, 0, 69, 28, 61, 0, 85, 28, 61, 0, 37, 48, 61, 0, 37, 166, 61, 0, 213, 168, 61, 0, 5, 169, 61, 0, 213, 169, 61, 0, 245, 169, 61, 0, 85, 170, 61, 0, 245, 171, 61, 0, 0, 216, 11, 221, 61, 0, 0, 216, 67, 221, 61, 0, 0, 216, 72, 221, 61, 0, 0, 216, 79, 221, 61, 0, 0, 216, 95, 221, 61, 0, 0, 216, 115, 221, 61, 0, 0, 216, 229, 222, 61, 0, 0, 216, 33, 223, 61, 0, 1, 216, 165, 220, 61, 0, 14, 32, 2, 216, 125, 220, 14, 32, 61, 0, 14, 32, 2, 216, 172, 220, 14, 32, 61, 0, 14, 32, 2, 216, 252, 220, 14, 32, 61, 0, 14, 32, 2, 216, 196, 221, 14, 32, 61, 0, 14, 32, 2, 216, 236, 222, 14, 32, 61, 0, 14, 32, 3, 216, 251, 220, 14, 32, 61, 0, 3, 216, 53, 221, 61, 0, 3, 216, 100, 222, 61, 0, 14, 32, 3, 216, 33, 223, 14, 32, 61, 0, 4, 216, 86, 220, 61, 0, 4, 216, 107, 220, 61, 0, 4, 216, 245, 220, 61, 0, 4, 216, 59, 221, 61, 0, 4, 216, 213, 221, 61, 0, 4, 216, 229, 221, 61, 0, 4, 216, 245, 222, 61, 0, 5, 216, 85, 220, 61, 0, 5, 216, 213, 220, 61, 0, 5, 216, 85, 222, 61, 0, 5, 216, 197, 222, 61, 0, 5, 216, 53, 223, 61, 0, 6, 216, 229, 220, 61, 0, 6, 216, 85, 221, 61, 0, 7, 216, 85, 220, 61, 0, 7, 216, 94, 220, 61, 0, 7, 216, 85, 221, 61, 0, 7, 216, 165, 221, 61, 0, 9, 216, 3, 220, 61, 0, 9, 216, 10, 220, 61, 0, 9, 216, 16, 220, 61, 0, 9, 216, 25, 220, 61, 0, 9, 216, 34, 220, 61, 0, 9, 216, 39, 220, 61, 0, 9, 216, 49, 220, 61, 0, 9, 216, 57, 220, 61, 0, 9, 216, 77, 220, 61, 0, 9, 216, 84, 220, 61, 0, 9, 216, 85, 220, 61, 0, 9, 216, 106, 220, 61, 0, 26, 216, 101, 222, 61, 0, 26, 216, 85, 223, 61, 0, 27, 216, 133, 222, 61, 0, 52, 216, 229, 222, 61, 0, 52, 216, 100, 223, 61, 0, 52, 216, 118, 223, 61, 0, 52, 216, 120, 223, 61, 0, 56, 216, 69, 221, 61, 0, 56, 216, 245, 222, 61, 0, 14, 32, 58, 216, 203, 220, 14, 32, 61, 0, 14, 32, 58, 216, 85, 221, 14, 32, 61, 0, 14, 32, 59, 216, 117, 220, 14, 32, 61, 0, 14, 32, 59, 216, 167, 220, 14, 32, 61, 0, 14, 32, 59, 216, 5, 221, 14, 32, 60, 0, 60, 0, 60, 0, 21, 255, 60, 0, 60, 0, 60, 0, 46, 15, 60, 0, 60, 0, 60, 0, 60, 216, 6, 221, 47, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 140, 36, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 228, 51, 47, 0, 229, 101, 60, 0, 60, 0, 60, 0, 196, 50, 47, 0, 8, 103, 60, 0, 60, 0, 60, 0, 93, 51, 47, 0, 185, 112, 60, 0, 60, 0, 60, 0, 53, 216, 211, 223, 61, 0, 53, 216, 221, 223, 61, 0, 53, 216, 231, 223, 61, 0, 53, 216, 241, 223, 61, 0, 53, 216, 251, 223, 61, 0, 62, 216, 245, 223, 60, 0, 60, 0, 60, 0, 100, 36, 61, 0, 249, 36, 61, 0, 122, 39, 61, 0, 132, 39, 61, 0, 142, 39, 60, 0, 60, 0, 60, 0, 76, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 61, 0, 191, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 60, 0, 60, 0, 60, 0, 117, 32, 60, 0, 60, 0, 60, 0, 133, 32, 60, 0, 60, 0, 60, 0, 60, 216, 160, 221, 47, 0, 14, 32, 39, 0, 46, 0, 39, 0, 59, 216, 1, 221, 14, 32, 60, 0, 60, 0, 60, 0, 90, 33, 47, 0, 14, 32, 68, 32, 59, 216, 6, 221, 14, 32, 60, 0, 60, 0, 60, 0, 93, 33, 47, 0, 14, 32, 68, 32, 59, 216, 8, 221, 14, 32, 60, 0, 60, 0, 14, 32, 59, 216, 50, 221, 14, 32, 60, 0, 54, 0, 61, 0, 102, 6, 61, 0, 246, 6, 61, 0, 14, 32, 198, 7, 14, 32, 61, 0, 108, 9, 61, 0, 236, 9, 61, 0, 108, 10, 61, 0, 236, 10, 61, 0, 108, 11, 61, 0, 236, 11, 61, 0, 108, 12, 61, 0, 236, 12, 61, 0, 108, 13, 61, 0, 236, 13, 61, 0, 86, 14, 61, 0, 214, 14, 61, 0, 38, 15, 61, 0, 70, 16, 61, 0, 150, 16, 61, 0, 110, 19, 61, 0, 230, 23, 61, 0, 246, 23, 61, 0, 22, 24, 61, 0, 76, 25, 61, 0, 214, 25, 61, 0, 134, 26, 61, 0, 150, 26, 61, 0, 86, 27, 61, 0, 182, 27, 61, 0, 70, 28, 61, 0, 86, 28, 61, 0, 133, 33, 61, 0, 38, 48, 61, 0, 38, 166, 61, 0, 214, 168, 61, 0, 6, 169, 61, 0, 214, 169, 61, 0, 246, 169, 61, 0, 86, 170, 61, 0, 246, 171, 61, 0, 0, 216, 12, 221, 61, 0, 0, 216, 230, 222, 61, 0, 1, 216, 166, 220, 61, 0, 14, 32, 2, 216, 197, 221, 14, 32, 61, 0, 3, 216, 54, 221, 61, 0, 3, 216, 101, 222, 61, 0, 4, 216, 87, 220, 61, 0, 4, 216, 108, 220, 61, 0, 4, 216, 246, 220, 61, 0, 4, 216, 60, 221, 61, 0, 4, 216, 214, 221, 61, 0, 4, 216, 230, 221, 61, 0, 4, 216, 246, 222, 61, 0, 5, 216, 86, 220, 61, 0, 5, 216, 214, 220, 61, 0, 5, 216, 86, 222, 61, 0, 5, 216, 198, 222, 61, 0, 5, 216, 54, 223, 61, 0, 6, 216, 230, 220, 61, 0, 6, 216, 86, 221, 61, 0, 7, 216, 86, 220, 61, 0, 7, 216, 95, 220, 61, 0, 7, 216, 86, 221, 61, 0, 7, 216, 166, 221, 61, 0, 9, 216, 4, 220, 61, 0, 9, 216, 11, 220, 61, 0, 9, 216, 17, 220, 61, 0, 9, 216, 26, 220, 61, 0, 9, 216, 40, 220, 61, 0, 9, 216, 64, 220, 61, 0, 9, 216, 78, 220, 61, 0, 9, 216, 107, 220, 61, 0, 26, 216, 102, 222, 61, 0, 26, 216, 86, 223, 61, 0, 27, 216, 134, 222, 61, 0, 52, 216, 230, 222, 61, 0, 52, 216, 101, 223, 61, 0, 56, 216, 70, 221, 61, 0, 56, 216, 246, 222, 61, 0, 14, 32, 58, 216, 204, 220, 14, 32, 61, 0, 14, 32, 58, 216, 86, 221, 14, 32, 61, 0, 14, 32, 59, 216, 118, 220, 14, 32, 61, 0, 14, 32, 59, 216, 168, 220, 14, 32, 61, 0, 14, 32, 59, 216, 6, 221, 14, 32, 60, 0, 60, 0, 60, 0, 22, 255, 60, 0, 60, 0, 60, 0, 47, 15, 60, 0, 60, 0, 60, 0, 60, 216, 7, 221, 47, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 141, 36, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 229, 51, 47, 0, 229, 101, 60, 0, 60, 0, 60, 0, 197, 50, 47, 0, 8, 103, 60, 0, 60, 0, 60, 0, 94, 51, 47, 0, 185, 112, 60, 0, 60, 0, 60, 0, 53, 216, 212, 223, 61, 0, 53, 216, 222, 223, 61, 0, 53, 216, 232, 223, 61, 0, 53, 216, 242, 223, 61, 0, 53, 216, 252, 223, 61, 0, 62, 216, 246, 223, 60, 0, 60, 0, 60, 0, 101, 36, 61, 0, 250, 36, 61, 0, 123, 39, 61, 0, 133, 39, 61, 0, 143, 39, 60, 0, 60, 0, 60, 0, 77, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 60, 0, 60, 0, 60, 0, 118, 32, 60, 0, 60, 0, 60, 0, 134, 32, 60, 0, 60, 0, 60, 0, 60, 216, 163, 221, 47, 0, 14, 32, 58, 216, 80, 221, 80, 0, 14, 32, 60, 0, 60, 0, 14, 32, 59, 216, 51, 221, 14, 32, 60, 0, 55, 0, 61, 0, 103, 6, 61, 0, 247, 6, 61, 0, 14, 32, 199, 7, 14, 32, 61, 0, 109, 9, 61, 0, 237, 9, 61, 0, 109, 10, 61, 0, 237, 10, 61, 0, 109, 11, 61, 0, 237, 11, 61, 0, 109, 12, 61, 0, 237, 12, 61, 0, 109, 13, 61, 0, 237, 13, 61, 0, 87, 14, 61, 0, 215, 14, 61, 0, 39, 15, 61, 0, 71, 16, 61, 0, 151, 16, 61, 0, 111, 19, 61, 0, 231, 23, 61, 0, 247, 23, 61, 0, 23, 24, 61, 0, 77, 25, 61, 0, 215, 25, 61, 0, 135, 26, 61, 0, 151, 26, 61, 0, 87, 27, 61, 0, 183, 27, 61, 0, 71, 28, 61, 0, 87, 28, 61, 0, 39, 48, 61, 0, 39, 166, 61, 0, 215, 168, 61, 0, 7, 169, 61, 0, 215, 169, 61, 0, 247, 169, 61, 0, 87, 170, 61, 0, 247, 171, 61, 0, 0, 216, 13, 221, 61, 0, 0, 216, 231, 222, 61, 0, 1, 216, 167, 220, 61, 0, 14, 32, 2, 216, 198, 221, 14, 32, 61, 0, 3, 216, 55, 221, 61, 0, 3, 216, 102, 222, 61, 0, 4, 216, 88, 220, 61, 0, 4, 216, 109, 220, 61, 0, 4, 216, 247, 220, 61, 0, 4, 216, 61, 221, 61, 0, 4, 216, 215, 221, 61, 0, 4, 216, 231, 221, 61, 0, 4, 216, 247, 222, 61, 0, 5, 216, 87, 220, 61, 0, 5, 216, 215, 220, 61, 0, 5, 216, 87, 222, 61, 0, 5, 216, 199, 222, 61, 0, 5, 216, 55, 223, 61, 0, 6, 216, 231, 220, 61, 0, 6, 216, 87, 221, 61, 0, 7, 216, 87, 220, 61, 0, 7, 216, 96, 220, 61, 0, 7, 216, 87, 221, 61, 0, 7, 216, 167, 221, 61, 0, 9, 216, 5, 220, 61, 0, 9, 216, 12, 220, 61, 0, 9, 216, 18, 220, 61, 0, 9, 216, 27, 220, 61, 0, 9, 216, 41, 220, 61, 0, 9, 216, 65, 220, 61, 0, 9, 216, 66, 220, 61, 0, 9, 216, 67, 220, 61, 0, 9, 216, 108, 220, 61, 0, 26, 216, 103, 222, 61, 0, 26, 216, 87, 223, 61, 0, 27, 216, 135, 222, 61, 0, 52, 216, 231, 222, 61, 0, 52, 216, 102, 223, 61, 0, 56, 216, 71, 221, 61, 0, 56, 216, 247, 222, 61, 0, 14, 32, 58, 216, 205, 220, 14, 32, 61, 0, 14, 32, 58, 216, 87, 221, 14, 32, 61, 0, 14, 32, 59, 216, 119, 220, 14, 32, 61, 0, 14, 32, 59, 216, 169, 220, 14, 32, 61, 0, 14, 32, 59, 216, 7, 221, 14, 32, 60, 0, 60, 0, 60, 0, 23, 255, 60, 0, 60, 0, 60, 0, 48, 15, 60, 0, 60, 0, 60, 0, 60, 216, 8, 221, 47, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 142, 36, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 230, 51, 47, 0, 229, 101, 60, 0, 60, 0, 60, 0, 198, 50, 47, 0, 8, 103, 60, 0, 60, 0, 60, 0, 95, 51, 47, 0, 185, 112, 60, 0, 60, 0, 60, 0, 53, 216, 213, 223, 61, 0, 53, 216, 223, 223, 61, 0, 53, 216, 233, 223, 61, 0, 53, 216, 243, 223, 61, 0, 53, 216, 253, 223, 61, 0, 62, 216, 247, 223, 60, 0, 60, 0, 60, 0, 102, 36, 61, 0, 251, 36, 61, 0, 124, 39, 61, 0, 134, 39, 61, 0, 144, 39, 60, 0, 60, 0, 60, 0, 78, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 60, 0, 60, 0, 60, 0, 119, 32, 60, 0, 60, 0, 60, 0, 135, 32, 60, 0, 60, 0, 60, 0, 60, 216, 161, 221, 47, 0, 14, 32, 39, 0, 46, 0, 39, 0, 59, 216, 1, 221, 14, 32, 60, 0, 60, 0, 60, 0, 94, 33, 47, 0, 14, 32, 68, 32, 59, 216, 8, 221, 14, 32, 60, 0, 60, 0, 14, 32, 59, 216, 52, 221, 14, 32, 60, 0, 56, 0, 61, 0, 104, 6, 61, 0, 248, 6, 61, 0, 14, 32, 200, 7, 14, 32, 61, 0, 110, 9, 61, 0, 238, 9, 61, 0, 110, 10, 61, 0, 238, 10, 61, 0, 110, 11, 61, 0, 238, 11, 61, 0, 110, 12, 61, 0, 238, 12, 61, 0, 110, 13, 61, 0, 238, 13, 61, 0, 88, 14, 61, 0, 216, 14, 61, 0, 40, 15, 61, 0, 72, 16, 61, 0, 152, 16, 61, 0, 112, 19, 61, 0, 232, 23, 61, 0, 248, 23, 61, 0, 24, 24, 61, 0, 78, 25, 61, 0, 216, 25, 61, 0, 136, 26, 61, 0, 152, 26, 61, 0, 88, 27, 61, 0, 184, 27, 61, 0, 72, 28, 61, 0, 88, 28, 61, 0, 40, 48, 61, 0, 40, 166, 61, 0, 216, 168, 61, 0, 8, 169, 61, 0, 216, 169, 61, 0, 248, 169, 61, 0, 88, 170, 61, 0, 248, 171, 61, 0, 0, 216, 14, 221, 61, 0, 0, 216, 232, 222, 61, 0, 1, 216, 168, 220, 61, 0, 14, 32, 2, 216, 199, 221, 14, 32, 61, 0, 3, 216, 56, 221, 61, 0, 3, 216, 103, 222, 61, 0, 4, 216, 89, 220, 61, 0, 4, 216, 110, 220, 61, 0, 4, 216, 248, 220, 61, 0, 4, 216, 62, 221, 61, 0, 4, 216, 216, 221, 61, 0, 4, 216, 232, 221, 61, 0, 4, 216, 248, 222, 61, 0, 5, 216, 88, 220, 61, 0, 5, 216, 216, 220, 61, 0, 5, 216, 88, 222, 61, 0, 5, 216, 200, 222, 61, 0, 5, 216, 56, 223, 61, 0, 6, 216, 232, 220, 61, 0, 6, 216, 88, 221, 61, 0, 7, 216, 88, 220, 61, 0, 7, 216, 97, 220, 61, 0, 7, 216, 88, 221, 61, 0, 7, 216, 168, 221, 61, 0, 9, 216, 6, 220, 61, 0, 9, 216, 13, 220, 61, 0, 9, 216, 19, 220, 61, 0, 9, 216, 28, 220, 61, 0, 9, 216, 42, 220, 61, 0, 9, 216, 68, 220, 61, 0, 9, 216, 69, 220, 61, 0, 9, 216, 109, 220, 61, 0, 26, 216, 104, 222, 61, 0, 26, 216, 88, 223, 61, 0, 27, 216, 136, 222, 61, 0, 52, 216, 232, 222, 61, 0, 52, 216, 103, 223, 61, 0, 56, 216, 72, 221, 61, 0, 56, 216, 248, 222, 61, 0, 14, 32, 58, 216, 206, 220, 14, 32, 61, 0, 14, 32, 58, 216, 88, 221, 14, 32, 61, 0, 14, 32, 59, 216, 120, 220, 14, 32, 61, 0, 14, 32, 59, 216, 170, 220, 14, 32, 61, 0, 14, 32, 59, 216, 8, 221, 14, 32, 60, 0, 60, 0, 60, 0, 24, 255, 60, 0, 60, 0, 60, 0, 49, 15, 60, 0, 60, 0, 60, 0, 60, 216, 9, 221, 47, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 143, 36, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 231, 51, 47, 0, 229, 101, 60, 0, 60, 0, 60, 0, 199, 50, 47, 0, 8, 103, 60, 0, 60, 0, 60, 0, 96, 51, 47, 0, 185, 112, 60, 0, 60, 0, 60, 0, 53, 216, 214, 223, 61, 0, 53, 216, 224, 223, 61, 0, 53, 216, 234, 223, 61, 0, 53, 216, 244, 223, 61, 0, 53, 216, 254, 223, 61, 0, 62, 216, 248, 223, 60, 0, 60, 0, 60, 0, 103, 36, 61, 0, 252, 36, 61, 0, 125, 39, 61, 0, 135, 39, 61, 0, 145, 39, 60, 0, 60, 0, 60, 0, 79, 50, 47, 0, 14, 32, 58, 216, 80, 221, 14, 32, 60, 0, 60, 0, 60, 0, 120, 32, 60, 0, 60, 0, 60, 0, 136, 32, 60, 0, 60, 0, 60, 0, 60, 216, 159, 221, 47, 0, 75, 0, 60, 0, 60, 0, 14, 32, 59, 216, 53, 221, 14, 32, 60, 0, 57, 0, 61, 0, 105, 6, 61, 0, 249, 6, 61, 0, 14, 32, 201, 7, 14, 32, 61, 0, 111, 9, 61, 0, 239, 9, 61, 0, 111, 10, 61, 0, 239, 10, 61, 0, 111, 11, 61, 0, 239, 11, 61, 0, 111, 12, 61, 0, 239, 12, 61, 0, 111, 13, 61, 0, 239, 13, 61, 0, 89, 14, 61, 0, 217, 14, 61, 0, 41, 15, 61, 0, 73, 16, 61, 0, 153, 16, 61, 0, 113, 19, 61, 0, 233, 23, 61, 0, 249, 23, 61, 0, 25, 24, 61, 0, 79, 25, 61, 0, 217, 25, 61, 0, 137, 26, 61, 0, 153, 26, 61, 0, 89, 27, 61, 0, 185, 27, 61, 0, 73, 28, 61, 0, 89, 28, 61, 0, 41, 48, 61, 0, 41, 166, 61, 0, 217, 168, 61, 0, 9, 169, 61, 0, 217, 169, 61, 0, 249, 169, 61, 0, 89, 170, 61, 0, 249, 171, 61, 0, 0, 216, 15, 221, 61, 0, 0, 216, 233, 222, 61, 0, 1, 216, 169, 220, 61, 0, 14, 32, 2, 216, 200, 221, 14, 32, 61, 0, 3, 216, 57, 221, 61, 0, 3, 216, 104, 222, 61, 0, 4, 216, 90, 220, 61, 0, 4, 216, 111, 220, 61, 0, 4, 216, 249, 220, 61, 0, 4, 216, 63, 221, 61, 0, 4, 216, 217, 221, 61, 0, 4, 216, 233, 221, 61, 0, 4, 216, 249, 222, 61, 0, 5, 216, 89, 220, 61, 0, 5, 216, 217, 220, 61, 0, 5, 216, 89, 222, 61, 0, 5, 216, 201, 222, 61, 0, 5, 216, 57, 223, 61, 0, 6, 216, 233, 220, 61, 0, 6, 216, 89, 221, 61, 0, 7, 216, 89, 220, 61, 0, 7, 216, 98, 220, 61, 0, 7, 216, 89, 221, 61, 0, 7, 216, 169, 221, 61, 0, 9, 216, 7, 220, 61, 0, 9, 216, 14, 220, 61, 0, 9, 216, 20, 220, 61, 0, 9, 216, 29, 220, 61, 0, 9, 216, 43, 220, 61, 0, 9, 216, 70, 220, 61, 0, 9, 216, 71, 220, 61, 0, 9, 216, 72, 220, 61, 0, 9, 216, 73, 220, 61, 0, 9, 216, 110, 220, 61, 0, 26, 216, 105, 222, 61, 0, 26, 216, 89, 223, 61, 0, 27, 216, 137, 222, 61, 0, 52, 216, 233, 222, 61, 0, 52, 216, 104, 223, 61, 0, 56, 216, 73, 221, 61, 0, 56, 216, 249, 222, 61, 0, 14, 32, 58, 216, 207, 220, 14, 32, 61, 0, 14, 32, 58, 216, 89, 221, 14, 32, 61, 0, 14, 32, 59, 216, 121, 220, 14, 32, 61, 0, 14, 32, 59, 216, 171, 220, 14, 32, 61, 0, 14, 32, 59, 216, 9, 221, 14, 32, 60, 0, 60, 0, 60, 0, 25, 255, 60, 0, 60, 0, 60, 0, 50, 15, 60, 0, 60, 0, 60, 0, 60, 216, 10, 221, 47, 0, 39, 0, 44, 0, 39, 0, 60, 0, 60, 0, 60, 0, 144, 36, 47, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 232, 51, 47, 0, 229, 101, 60, 0, 60, 0, 60, 0, 200, 50, 47, 0, 8, 103, 60, 0, 60, 0, 60, 0, 97, 51, 47, 0, 185, 112, 60, 0, 60, 0, 60, 0, 53, 216, 215, 223, 61, 0, 53, 216, 225, 223, 61, 0, 53, 216, 235, 223, 61, 0, 53, 216, 245, 223, 61, 0, 53, 216, 255, 223, 61, 0, 62, 216, 249, 223, 60, 0, 60, 0, 60, 0, 104, 36, 61, 0, 253, 36, 61, 0, 126, 39, 61, 0, 136, 39, 61, 0, 146, 39, 60, 0, 60, 0, 60, 0, 121, 32, 60, 0, 60, 0, 60, 0, 137, 32, 60, 0, 60, 0, 14, 32, 59, 216, 54, 221, 14, 32, 60, 0, 97, 0, 60, 0, 60, 0, 60, 0, 65, 255, 60, 0, 60, 0, 60, 0, 99, 3, 60, 0, 60, 0, 60, 0, 242, 29, 47, 0, 8, 3, 61, 0, 155, 167, 47, 0, 8, 3, 60, 0, 60, 0, 60, 0, 0, 33, 47, 0, 39, 0, 47, 0, 39, 0, 99, 0, 60, 0, 60, 0, 60, 0, 1, 33, 47, 0, 39, 0, 47, 0, 39, 0, 115, 0, 60, 0, 60, 0, 60, 0, 51, 167, 47, 0, 97, 0, 60, 0, 60, 0, 60, 0, 213, 29, 47, 0, 111, 0, 61, 0, 53, 167, 47, 0, 111, 0, 60, 0, 60, 0, 60, 0, 55, 167, 47, 0, 117, 0, 60, 0, 60, 0, 60, 0, 214, 29, 47, 0, 118, 0, 61, 0, 57, 167, 47, 0, 118, 0, 60, 0, 60, 0, 60, 0, 61, 167, 47, 0, 121, 0, 60, 0, 60, 0, 60, 0, 154, 30, 47, 0, 190, 2, 60, 0, 60, 0, 60, 0, 53, 216, 26, 220, 61, 0, 53, 216, 78, 220, 61, 0, 53, 216, 130, 220, 61, 0, 53, 216, 182, 220, 61, 0, 53, 216, 234, 220, 61, 0, 53, 216, 30, 221, 61, 0, 53, 216, 82, 221, 61, 0, 53, 216, 134, 221, 61, 0, 53, 216, 186, 221, 61, 0, 53, 216, 238, 221, 61, 0, 53, 216, 34, 222, 61, 0, 53, 216, 86, 222, 61, 0, 53, 216, 138, 222, 60, 0, 60, 0, 60, 0, 208, 36, 60, 0, 60, 0, 60, 0, 65, 0, 60, 0, 60, 0, 60, 0, 33, 255, 60, 0, 60, 0, 60, 0, 154, 167, 47, 0, 8, 3, 60, 0, 60, 0, 60, 0, 77, 33, 47, 0, 39, 0, 47, 0, 39, 0, 83, 0, 60, 0, 60, 0, 60, 0, 50, 167, 47, 0, 65, 0, 60, 0, 60, 0, 60, 0, 52, 167, 47, 0, 79, 0, 60, 0, 60, 0, 60, 0, 54, 167, 47, 0, 85, 0, 60, 0, 60, 0, 60, 0, 56, 167, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 60, 167, 47, 0, 89, 0, 60, 0, 60, 0, 60, 0, 53, 216, 0, 220, 61, 0, 53, 216, 52, 220, 61, 0, 53, 216, 104, 220, 61, 0, 53, 216, 156, 220, 61, 0, 53, 216, 208, 220, 61, 0, 53, 216, 4, 221, 61, 0, 53, 216, 56, 221, 61, 0, 53, 216, 108, 221, 61, 0, 53, 216, 160, 221, 61, 0, 53, 216, 212, 221, 61, 0, 53, 216, 8, 222, 61, 0, 53, 216, 60, 222, 61, 0, 53, 216, 112, 222, 60, 0, 60, 0, 60, 0, 182, 36, 61, 0, 60, 216, 80, 221, 60, 0, 60, 0, 60, 0, 170, 0, 61, 0, 67, 29, 60, 0, 60, 0, 60, 0, 144, 32, 60, 0, 60, 0, 60, 0, 194, 51, 47, 0, 39, 0, 46, 0, 39, 0, 109, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 44, 29, 61, 0, 60, 216, 48, 221, 61, 0, 60, 216, 112, 221, 60, 0, 60, 0, 60, 0, 223, 51, 47, 0, 21, 34, 109, 0, 60, 0, 60, 0, 60, 0, 60, 216, 142, 221, 47, 0, 66, 0, 60, 0, 60, 0, 60, 0, 115, 51, 47, 0, 85, 0, 60, 0, 60, 0, 211, 29, 60, 0, 60, 0, 60, 0, 230, 0, 47, 0, 101, 0, 61, 0, 212, 29, 47, 0, 101, 0, 60, 0, 60, 0, 60, 0, 59, 167, 47, 0, 118, 0, 60, 0, 60, 0, 60, 0, 198, 0, 47, 0, 69, 0, 60, 0, 60, 0, 60, 0, 58, 167, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 45, 29, 47, 0, 101, 0, 60, 0, 0, 29, 60, 0, 101, 44, 60, 0, 60, 0, 60, 0, 58, 2, 60, 0, 143, 29, 60, 0, 187, 167, 60, 0, 60, 0, 60, 0, 186, 167, 60, 0, 1, 29, 60, 0, 2, 29, 60, 0, 60, 0, 60, 0, 70, 29, 60, 0, 49, 171, 60, 0, 80, 2, 60, 0, 60, 0, 60, 0, 111, 44, 60, 0, 60, 0, 60, 0, 68, 29, 60, 0, 81, 2, 60, 0, 60, 0, 60, 0, 231, 29, 60, 0, 60, 0, 60, 0, 109, 44, 60, 0, 60, 0, 60, 0, 69, 29, 60, 0, 48, 171, 60, 0, 144, 29, 60, 0, 82, 2, 60, 0, 60, 0, 60, 0, 112, 44, 60, 0, 60, 0, 60, 0, 155, 29, 60, 0, 100, 171, 60, 0, 98, 0, 60, 0, 60, 0, 60, 0, 66, 255, 60, 0, 60, 0, 60, 0, 232, 29, 60, 0, 60, 0, 60, 0, 53, 216, 27, 220, 61, 0, 53, 216, 79, 220, 61, 0, 53, 216, 131, 220, 61, 0, 53, 216, 183, 220, 61, 0, 53, 216, 235, 220, 61, 0, 53, 216, 31, 221, 61, 0, 53, 216, 83, 221, 61, 0, 53, 216, 135, 221, 61, 0, 53, 216, 187, 221, 61, 0, 53, 216, 239, 221, 61, 0, 53, 216, 35, 222, 61, 0, 53, 216, 87, 222, 61, 0, 53, 216, 139, 222, 60, 0, 60, 0, 60, 0, 209, 36, 60, 0, 60, 0, 60, 0, 66, 0, 60, 0, 60, 0, 60, 0, 34, 255, 60, 0, 60, 0, 60, 0, 44, 33, 61, 0, 53, 216, 1, 220, 61, 0, 53, 216, 53, 220, 61, 0, 53, 216, 105, 220, 61, 0, 53, 216, 209, 220, 61, 0, 53, 216, 5, 221, 61, 0, 53, 216, 57, 221, 61, 0, 53, 216, 109, 221, 61, 0, 53, 216, 161, 221, 61, 0, 53, 216, 213, 221, 61, 0, 53, 216, 9, 222, 61, 0, 53, 216, 61, 222, 61, 0, 53, 216, 113, 222, 60, 0, 60, 0, 60, 0, 183, 36, 61, 0, 60, 216, 81, 221, 60, 0, 60, 0, 60, 0, 71, 29, 60, 0, 60, 0, 60, 0, 116, 51, 47, 0, 97, 0, 114, 0, 60, 0, 60, 0, 60, 0, 46, 29, 61, 0, 60, 216, 49, 221, 61, 0, 60, 216, 113, 221, 60, 0, 60, 0, 60, 0, 195, 51, 47, 0, 113, 0, 60, 0, 153, 2, 60, 0, 128, 1, 60, 0, 60, 0, 60, 0, 67, 2, 60, 0, 47, 29, 60, 0, 3, 29, 60, 0, 108, 29, 60, 0, 151, 167, 60, 0, 60, 0, 60, 0, 150, 167, 60, 0, 128, 29, 60, 0, 83, 2, 60, 0, 60, 0, 60, 0, 129, 1, 60, 0, 131, 1, 60, 0, 60, 0, 60, 0, 130, 1, 60, 0, 181, 167, 60, 0, 60, 0, 60, 0, 233, 29, 60, 0, 60, 0, 60, 0, 180, 167, 60, 0, 99, 0, 60, 0, 60, 0, 60, 0, 67, 255, 60, 0, 60, 0, 60, 0, 104, 3, 61, 0, 125, 33, 60, 0, 60, 0, 60, 0, 215, 29, 47, 0, 39, 3, 60, 0, 60, 0, 60, 0, 5, 33, 47, 0, 39, 0, 47, 0, 39, 0, 111, 0, 60, 0, 60, 0, 60, 0, 6, 33, 47, 0, 39, 0, 47, 0, 39, 0, 117, 0, 60, 0, 60, 0, 60, 0, 53, 216, 28, 220, 61, 0, 53, 216, 80, 220, 61, 0, 53, 216, 132, 220, 61, 0, 53, 216, 184, 220, 61, 0, 53, 216, 236, 220, 61, 0, 53, 216, 32, 221, 61, 0, 53, 216, 84, 221, 61, 0, 53, 216, 136, 221, 61, 0, 53, 216, 188, 221, 61, 0, 53, 216, 240, 221, 61, 0, 53, 216, 36, 222, 61, 0, 53, 216, 88, 222, 61, 0, 53, 216, 140, 222, 60, 0, 60, 0, 60, 0, 210, 36, 60, 0, 60, 0, 60, 0, 67, 0, 60, 0, 60, 0, 60, 0, 35, 255, 60, 0, 60, 0, 60, 0, 109, 33, 60, 0, 60, 0, 60, 0, 2, 33, 61, 0, 45, 33, 61, 0, 53, 216, 2, 220, 61, 0, 53, 216, 54, 220, 61, 0, 53, 216, 106, 220, 61, 0, 53, 216, 158, 220, 61, 0, 53, 216, 210, 220, 61, 0, 53, 216, 110, 221, 61, 0, 53, 216, 162, 221, 61, 0, 53, 216, 214, 221, 61, 0, 53, 216, 10, 222, 61, 0, 53, 216, 62, 222, 61, 0, 53, 216, 114, 222, 60, 0, 60, 0, 60, 0, 184, 36, 61, 0, 60, 216, 43, 221, 61, 0, 60, 216, 82, 221, 60, 0, 60, 0, 60, 0, 60, 216, 45, 221, 47, 0, 68, 0, 60, 0, 60, 0, 60, 0, 156, 29, 60, 0, 60, 0, 60, 0, 136, 51, 47, 0, 97, 0, 108, 0, 60, 0, 60, 0, 60, 0, 196, 51, 47, 0, 99, 0, 60, 0, 60, 0, 60, 0, 197, 51, 47, 0, 100, 0, 60, 0, 60, 0, 60, 0, 157, 51, 47, 0, 109, 0, 60, 0, 60, 0, 60, 0, 160, 51, 47, 0, 14, 32, 109, 0, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 164, 51, 47, 0, 14, 32, 109, 0, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 60, 216, 50, 221, 61, 0, 60, 216, 114, 221, 60, 0, 60, 0, 60, 0, 198, 51, 47, 0, 21, 34, 107, 0, 103, 0, 60, 0, 60, 0, 60, 0, 60, 216, 145, 221, 47, 0, 76, 0, 60, 0, 60, 0, 60, 0, 199, 51, 47, 0, 111, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 146, 221, 47, 0, 79, 0, 79, 0, 76, 0, 60, 0, 4, 29, 60, 0, 60, 2, 60, 0, 60, 0, 60, 0, 59, 2, 60, 0, 147, 167, 60, 0, 60, 0, 60, 0, 146, 167, 60, 0, 148, 167, 60, 0, 60, 0, 60, 0, 196, 167, 60, 0, 136, 1, 60, 0, 60, 0, 60, 0, 135, 1, 60, 0, 85, 2, 60, 0, 60, 0, 60, 0, 157, 29, 60, 0, 132, 33, 60, 0, 60, 0, 60, 0, 131, 33, 60, 0, 63, 167, 60, 0, 60, 0, 60, 0, 62, 167, 60, 0, 100, 0, 60, 0, 60, 0, 60, 0, 17, 1, 47, 0, 53, 3, 60, 0, 60, 0, 60, 0, 68, 255, 60, 0, 60, 0, 60, 0, 105, 3, 61, 0, 126, 33, 60, 0, 60, 0, 60, 0, 56, 2, 47, 0, 98, 0, 60, 0, 60, 0, 60, 0, 243, 1, 47, 0, 122, 0, 61, 0, 163, 2, 47, 0, 122, 0, 60, 0, 60, 0, 60, 0, 198, 1, 47, 0, 126, 1, 60, 0, 60, 0, 60, 0, 102, 171, 47, 0, 144, 2, 60, 0, 60, 0, 60, 0, 165, 2, 47, 0, 145, 2, 60, 0, 60, 0, 60, 0, 164, 2, 47, 0, 146, 2, 60, 0, 60, 0, 60, 0, 70, 33, 61, 0, 53, 216, 29, 220, 61, 0, 53, 216, 81, 220, 61, 0, 53, 216, 133, 220, 61, 0, 53, 216, 185, 220, 61, 0, 53, 216, 237, 220, 61, 0, 53, 216, 33, 221, 61, 0, 53, 216, 85, 221, 61, 0, 53, 216, 137, 221, 61, 0, 53, 216, 189, 221, 61, 0, 53, 216, 241, 221, 61, 0, 53, 216, 37, 222, 61, 0, 53, 216, 89, 222, 61, 0, 53, 216, 141, 222, 60, 0, 60, 0, 60, 0, 211, 36, 60, 0, 60, 0, 60, 0, 68, 0, 60, 0, 60, 0, 60, 0, 16, 1, 47, 0, 53, 3, 60, 0, 60, 0, 60, 0, 36, 255, 60, 0, 60, 0, 60, 0, 110, 33, 60, 0, 60, 0, 60, 0, 242, 1, 47, 0, 122, 0, 60, 0, 60, 0, 60, 0, 241, 1, 47, 0, 90, 0, 60, 0, 60, 0, 60, 0, 197, 1, 47, 0, 126, 1, 60, 0, 60, 0, 60, 0, 196, 1, 47, 0, 125, 1, 60, 0, 60, 0, 60, 0, 69, 33, 61, 0, 53, 216, 3, 220, 61, 0, 53, 216, 55, 220, 61, 0, 53, 216, 107, 220, 61, 0, 53, 216, 159, 220, 61, 0, 53, 216, 211, 220, 61, 0, 53, 216, 7, 221, 61, 0, 53, 216, 59, 221, 61, 0, 53, 216, 111, 221, 61, 0, 53, 216, 163, 221, 61, 0, 53, 216, 215, 221, 61, 0, 53, 216, 11, 222, 61, 0, 53, 216, 63, 222, 61, 0, 53, 216, 115, 222, 60, 0, 60, 0, 60, 0, 185, 36, 61, 0, 60, 216, 83, 221, 60, 0, 60, 0, 60, 0, 72, 29, 60, 0, 60, 0, 60, 0, 60, 216, 165, 221, 60, 0, 60, 0, 60, 0, 114, 51, 47, 0, 97, 0, 60, 0, 60, 0, 60, 0, 200, 51, 47, 0, 66, 0, 60, 0, 60, 0, 60, 0, 151, 51, 47, 0, 108, 0, 60, 0, 60, 0, 60, 0, 119, 51, 47, 0, 109, 0, 60, 0, 60, 0, 60, 0, 120, 51, 47, 0, 14, 32, 109, 0, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 121, 51, 47, 0, 14, 32, 109, 0, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 48, 29, 61, 0, 60, 216, 51, 221, 61, 0, 60, 216, 115, 221, 60, 0, 60, 0, 60, 0, 60, 216, 144, 221, 47, 0, 74, 0, 60, 0, 60, 0, 240, 0, 61, 0, 217, 29, 60, 0, 60, 0, 60, 0, 208, 0, 60, 0, 60, 0, 60, 0, 158, 29, 60, 0, 60, 0, 216, 29, 61, 0, 122, 167, 60, 0, 60, 0, 60, 0, 121, 167, 60, 0, 5, 29, 60, 0, 6, 29, 60, 0, 200, 167, 60, 0, 60, 0, 60, 0, 199, 167, 60, 0, 109, 29, 60, 0, 129, 29, 60, 0, 86, 2, 60, 0, 60, 0, 60, 0, 137, 1, 60, 0, 87, 2, 60, 0, 60, 0, 60, 0, 138, 1, 60, 0, 145, 29, 60, 0, 140, 1, 60, 0, 60, 0, 60, 0, 139, 1, 60, 0, 33, 2, 60, 0, 113, 167, 60, 0, 159, 30, 60, 0, 101, 0, 60, 0, 60, 0, 60, 0, 69, 255, 60, 0, 60, 0, 60, 0, 100, 3, 60, 0, 60, 0, 60, 0, 47, 33, 61, 0, 71, 33, 61, 0, 53, 216, 30, 220, 61, 0, 53, 216, 82, 220, 61, 0, 53, 216, 134, 220, 61, 0, 53, 216, 238, 220, 61, 0, 53, 216, 34, 221, 61, 0, 53, 216, 86, 221, 61, 0, 53, 216, 138, 221, 61, 0, 53, 216, 190, 221, 61, 0, 53, 216, 242, 221, 61, 0, 53, 216, 38, 222, 61, 0, 53, 216, 90, 222, 61, 0, 53, 216, 142, 222, 60, 0, 60, 0, 60, 0, 212, 36, 60, 0, 60, 0, 60, 0, 69, 0, 60, 0, 60, 0, 60, 0, 37, 255, 60, 0, 60, 0, 60, 0, 48, 33, 61, 0, 53, 216, 4, 220, 61, 0, 53, 216, 56, 220, 61, 0, 53, 216, 108, 220, 61, 0, 53, 216, 212, 220, 61, 0, 53, 216, 8, 221, 61, 0, 53, 216, 60, 221, 61, 0, 53, 216, 112, 221, 61, 0, 53, 216, 164, 221, 61, 0, 53, 216, 216, 221, 61, 0, 53, 216, 12, 222, 61, 0, 53, 216, 64, 222, 61, 0, 53, 216, 116, 222, 60, 0, 60, 0, 60, 0, 186, 36, 61, 0, 60, 216, 84, 221, 60, 0, 60, 0, 60, 0, 73, 29, 60, 0, 60, 0, 60, 0, 145, 32, 60, 0, 60, 0, 60, 0, 205, 50, 47, 0, 114, 0, 103, 0, 60, 0, 60, 0, 60, 0, 206, 50, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 49, 29, 61, 0, 60, 216, 52, 221, 61, 0, 60, 216, 116, 221, 60, 0, 7, 29, 60, 0, 50, 171, 60, 0, 51, 171, 60, 0, 71, 2, 60, 0, 60, 0, 60, 0, 70, 2, 60, 0, 146, 29, 60, 0, 52, 171, 60, 0, 120, 44, 60, 0, 221, 1, 60, 0, 60, 0, 60, 0, 142, 1, 60, 0, 60, 0, 60, 0, 50, 29, 60, 0, 123, 44, 60, 0, 89, 2, 60, 0, 60, 0, 60, 0, 234, 29, 60, 0, 60, 0, 60, 0, 143, 1, 60, 0, 60, 0, 60, 0, 74, 29, 60, 0, 60, 0, 60, 0, 148, 32, 60, 0, 149, 29, 60, 0, 91, 2, 60, 0, 60, 0, 60, 0, 144, 1, 60, 0, 60, 0, 60, 0, 7, 33, 60, 0, 60, 0, 60, 0, 75, 29, 60, 0, 147, 29, 60, 0, 88, 2, 60, 0, 90, 2, 60, 0, 92, 2, 60, 0, 60, 0, 60, 0, 171, 167, 60, 0, 60, 0, 60, 0, 159, 29, 60, 0, 148, 29, 60, 0, 8, 29, 60, 0, 60, 0, 60, 0, 76, 29, 60, 0, 93, 2, 60, 0, 94, 2, 60, 0, 154, 2, 60, 0, 100, 2, 60, 0, 102, 0, 60, 0, 60, 0, 60, 0, 70, 255, 60, 0, 60, 0, 60, 0, 235, 29, 60, 0, 60, 0, 60, 0, 0, 251, 47, 0, 102, 0, 60, 0, 60, 0, 60, 0, 3, 251, 47, 0, 102, 0, 105, 0, 60, 0, 60, 0, 60, 0, 4, 251, 47, 0, 102, 0, 108, 0, 60, 0, 60, 0, 60, 0, 1, 251, 47, 0, 105, 0, 60, 0, 60, 0, 60, 0, 2, 251, 47, 0, 108, 0, 60, 0, 60, 0, 60, 0, 169, 2, 47, 0, 75, 1, 60, 0, 60, 0, 60, 0, 53, 216, 31, 220, 61, 0, 53, 216, 83, 220, 61, 0, 53, 216, 135, 220, 61, 0, 53, 216, 187, 220, 61, 0, 53, 216, 239, 220, 61, 0, 53, 216, 35, 221, 61, 0, 53, 216, 87, 221, 61, 0, 53, 216, 139, 221, 61, 0, 53, 216, 191, 221, 61, 0, 53, 216, 243, 221, 61, 0, 53, 216, 39, 222, 61, 0, 53, 216, 91, 222, 61, 0, 53, 216, 143, 222, 60, 0, 60, 0, 60, 0, 213, 36, 60, 0, 60, 0, 60, 0, 70, 0, 60, 0, 60, 0, 60, 0, 38, 255, 60, 0, 60, 0, 60, 0, 59, 33, 47, 0, 65, 0, 88, 0, 60, 0, 60, 0, 60, 0, 49, 33, 61, 0, 53, 216, 5, 220, 61, 0, 53, 216, 57, 220, 61, 0, 53, 216, 109, 220, 61, 0, 53, 216, 213, 220, 61, 0, 53, 216, 9, 221, 61, 0, 53, 216, 61, 221, 61, 0, 53, 216, 113, 221, 61, 0, 53, 216, 165, 221, 61, 0, 53, 216, 217, 221, 61, 0, 53, 216, 13, 222, 61, 0, 53, 216, 65, 222, 61, 0, 53, 216, 117, 222, 60, 0, 60, 0, 60, 0, 187, 36, 61, 0, 60, 216, 85, 221, 60, 0, 60, 0, 60, 0, 160, 29, 60, 0, 60, 0, 60, 0, 153, 51, 47, 0, 109, 0, 60, 0, 60, 0, 60, 0, 60, 216, 53, 221, 61, 0, 60, 216, 117, 221, 60, 0, 60, 0, 60, 0, 60, 216, 147, 221, 47, 0, 82, 0, 69, 0, 69, 0, 60, 0, 60, 0, 124, 167, 60, 0, 60, 0, 60, 0, 123, 167, 60, 0, 48, 167, 60, 0, 53, 171, 60, 0, 153, 167, 60, 0, 60, 0, 60, 0, 152, 167, 60, 0, 110, 29, 60, 0, 130, 29, 60, 0, 146, 1, 60, 0, 60, 0, 60, 0, 145, 1, 60, 0, 78, 33, 60, 0, 60, 0, 60, 0, 50, 33, 60, 0, 251, 167, 60, 0, 103, 0, 60, 0, 60, 0, 60, 0, 71, 255, 60, 0, 60, 0, 60, 0, 218, 29, 60, 0, 60, 0, 60, 0, 161, 167, 47, 0, 47, 216, 158, 220, 60, 0, 60, 0, 60, 0, 10, 33, 61, 0, 53, 216, 32, 220, 61, 0, 53, 216, 84, 220, 61, 0, 53, 216, 136, 220, 61, 0, 53, 216, 240, 220, 61, 0, 53, 216, 36, 221, 61, 0, 53, 216, 88, 221, 61, 0, 53, 216, 140, 221, 61, 0, 53, 216, 192, 221, 61, 0, 53, 216, 244, 221, 61, 0, 53, 216, 40, 222, 61, 0, 53, 216, 92, 222, 61, 0, 53, 216, 144, 222, 60, 0, 60, 0, 60, 0, 214, 36, 60, 0, 60, 0, 60, 0, 71, 0, 60, 0, 60, 0, 60, 0, 39, 255, 60, 0, 60, 0, 60, 0, 160, 167, 47, 0, 47, 216, 158, 220, 60, 0, 60, 0, 60, 0, 53, 216, 6, 220, 61, 0, 53, 216, 58, 220, 61, 0, 53, 216, 110, 220, 61, 0, 53, 216, 162, 220, 61, 0, 53, 216, 214, 220, 61, 0, 53, 216, 10, 221, 61, 0, 53, 216, 62, 221, 61, 0, 53, 216, 114, 221, 61, 0, 53, 216, 166, 221, 61, 0, 53, 216, 218, 221, 61, 0, 53, 216, 14, 222, 61, 0, 53, 216, 66, 222, 61, 0, 53, 216, 118, 222, 60, 0, 60, 0, 60, 0, 188, 36, 61, 0, 60, 216, 86, 221, 60, 0, 60, 0, 60, 0, 77, 29, 60, 0, 60, 0, 60, 0, 255, 51, 47, 0, 97, 0, 108, 0, 60, 0, 60, 0, 60, 0, 51, 29, 61, 0, 60, 216, 54, 221, 61, 0, 60, 216, 118, 221, 60, 0, 60, 0, 60, 0, 135, 51, 47, 0, 66, 0, 60, 0, 60, 0, 60, 0, 147, 51, 47, 0, 72, 0, 122, 0, 60, 0, 60, 0, 60, 0, 172, 51, 47, 0, 80, 0, 97, 0, 60, 0, 60, 0, 60, 0, 201, 51, 47, 0, 121, 0, 60, 0, 60, 0, 121, 29, 60, 0, 60, 0, 60, 0, 125, 167, 60, 0, 97, 2, 60, 0, 60, 0, 60, 0, 172, 167, 60, 0, 60, 0, 60, 0, 162, 29, 60, 0, 54, 171, 60, 0, 98, 2, 60, 0, 60, 0, 60, 0, 219, 29, 60, 0, 229, 1, 60, 0, 60, 0, 60, 0, 228, 1, 60, 0, 131, 29, 60, 0, 96, 2, 60, 0, 60, 0, 60, 0, 147, 1, 60, 0, 155, 2, 60, 0, 119, 29, 60, 0, 127, 167, 60, 0, 60, 0, 60, 0, 126, 167, 60, 0, 99, 2, 60, 0, 60, 0, 60, 0, 148, 1, 60, 0, 60, 0, 60, 0, 224, 2, 60, 0, 163, 1, 60, 0, 60, 0, 60, 0, 162, 1, 60, 0, 104, 0, 60, 0, 60, 0, 60, 0, 39, 1, 47, 0, 53, 3, 61, 0, 15, 33, 47, 0, 53, 3, 60, 0, 60, 0, 60, 0, 72, 255, 60, 0, 60, 0, 60, 0, 106, 3, 60, 0, 60, 0, 60, 0, 14, 33, 61, 0, 53, 216, 33, 220, 61, 0, 53, 216, 137, 220, 61, 0, 53, 216, 189, 220, 61, 0, 53, 216, 241, 220, 61, 0, 53, 216, 37, 221, 61, 0, 53, 216, 89, 221, 61, 0, 53, 216, 141, 221, 61, 0, 53, 216, 193, 221, 61, 0, 53, 216, 245, 221, 61, 0, 53, 216, 41, 222, 61, 0, 53, 216, 93, 222, 61, 0, 53, 216, 145, 222, 60, 0, 60, 0, 60, 0, 215, 36, 60, 0, 60, 0, 60, 0, 72, 0, 60, 0, 60, 0, 60, 0, 38, 1, 47, 0, 53, 3, 60, 0, 60, 0, 60, 0, 40, 255, 60, 0, 60, 0, 60, 0, 11, 33, 61, 0, 12, 33, 61, 0, 13, 33, 61, 0, 53, 216, 7, 220, 61, 0, 53, 216, 59, 220, 61, 0, 53, 216, 111, 220, 61, 0, 53, 216, 215, 220, 61, 0, 53, 216, 115, 221, 61, 0, 53, 216, 167, 221, 61, 0, 53, 216, 219, 221, 61, 0, 53, 216, 15, 222, 61, 0, 53, 216, 67, 222, 61, 0, 53, 216, 119, 222, 60, 0, 60, 0, 60, 0, 189, 36, 61, 0, 60, 216, 87, 221, 60, 0, 60, 0, 60, 0, 176, 2, 60, 0, 60, 0, 60, 0, 248, 167, 47, 0, 53, 3, 60, 0, 60, 0, 60, 0, 149, 32, 60, 0, 60, 0, 60, 0, 202, 51, 47, 0, 97, 0, 60, 0, 60, 0, 60, 0, 113, 51, 47, 0, 80, 0, 97, 0, 60, 0, 60, 0, 60, 0, 52, 29, 61, 0, 60, 216, 55, 221, 61, 0, 60, 216, 119, 221, 60, 0, 60, 0, 60, 0, 60, 216, 166, 221, 47, 0, 99, 0, 60, 0, 60, 0, 60, 0, 60, 216, 167, 221, 47, 0, 100, 0, 82, 0, 60, 0, 60, 0, 60, 0, 204, 50, 47, 0, 103, 0, 60, 0, 60, 0, 60, 0, 60, 216, 168, 221, 47, 0, 105, 0, 39, 0, 32, 0, 39, 0, 82, 0, 101, 0, 115, 0, 60, 0, 60, 0, 60, 0, 203, 51, 47, 0, 80, 0, 60, 0, 60, 0, 60, 0, 60, 216, 74, 221, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 144, 51, 47, 0, 122, 0, 60, 0, 156, 2, 60, 0, 149, 1, 60, 0, 60, 0, 60, 0, 246, 1, 60, 0, 149, 167, 60, 0, 102, 2, 60, 0, 60, 0, 60, 0, 170, 167, 60, 0, 60, 0, 60, 0, 177, 2, 60, 0, 104, 44, 60, 0, 60, 0, 60, 0, 103, 44, 60, 0, 118, 44, 60, 0, 60, 0, 60, 0, 117, 44, 60, 0, 246, 167, 60, 0, 60, 0, 60, 0, 245, 167, 60, 0, 39, 167, 60, 0, 60, 0, 60, 0, 38, 167, 60, 0, 60, 0, 60, 0, 92, 171, 60, 0, 103, 2, 60, 0, 187, 2, 60, 0, 189, 2, 60, 0, 105, 0, 60, 0, 60, 0, 60, 0, 73, 255, 60, 0, 60, 0, 60, 0, 101, 3, 61, 0, 112, 33, 60, 0, 60, 0, 60, 0, 113, 33, 47, 0, 105, 0, 60, 0, 60, 0, 60, 0, 114, 33, 47, 0, 105, 0, 105, 0, 60, 0, 60, 0, 60, 0, 51, 1, 47, 0, 106, 0, 60, 0, 60, 0, 60, 0, 115, 33, 47, 0, 118, 0, 60, 0, 60, 0, 60, 0, 120, 33, 47, 0, 120, 0, 60, 0, 60, 0, 60, 0, 57, 33, 61, 0, 72, 33, 61, 0, 53, 216, 34, 220, 61, 0, 53, 216, 86, 220, 61, 0, 53, 216, 138, 220, 61, 0, 53, 216, 190, 220, 61, 0, 53, 216, 242, 220, 61, 0, 53, 216, 38, 221, 61, 0, 53, 216, 90, 221, 61, 0, 53, 216, 142, 221, 61, 0, 53, 216, 194, 221, 61, 0, 53, 216, 246, 221, 61, 0, 53, 216, 42, 222, 61, 0, 53, 216, 94, 222, 61, 0, 53, 216, 146, 222, 60, 0, 60, 0, 60, 0, 216, 36, 60, 0, 60, 0, 60, 0, 73, 0, 60, 0, 60, 0, 60, 0, 41, 255, 60, 0, 60, 0, 60, 0, 96, 33, 60, 0, 60, 0, 60, 0, 97, 33, 47, 0, 73, 0, 60, 0, 60, 0, 60, 0, 98, 33, 47, 0, 73, 0, 73, 0, 60, 0, 60, 0, 60, 0, 50, 1, 47, 0, 74, 0, 60, 0, 60, 0, 60, 0, 99, 33, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 104, 33, 47, 0, 88, 0, 60, 0, 60, 0, 60, 0, 16, 33, 61, 0, 17, 33, 61, 0, 53, 216, 8, 220, 61, 0, 53, 216, 60, 220, 61, 0, 53, 216, 112, 220, 61, 0, 53, 216, 216, 220, 61, 0, 53, 216, 64, 221, 61, 0, 53, 216, 116, 221, 61, 0, 53, 216, 168, 221, 61, 0, 53, 216, 220, 221, 61, 0, 53, 216, 16, 222, 61, 0, 53, 216, 68, 222, 61, 0, 53, 216, 120, 222, 60, 0, 60, 0, 60, 0, 190, 36, 61, 0, 60, 216, 88, 221, 60, 0, 60, 0, 60, 0, 113, 32, 60, 0, 60, 0, 60, 0, 98, 29, 60, 0, 60, 0, 60, 0, 204, 51, 47, 0, 110, 0, 60, 0, 60, 0, 60, 0, 53, 29, 61, 0, 60, 216, 56, 221, 61, 0, 60, 216, 120, 221, 60, 0, 60, 0, 60, 0, 60, 216, 139, 221, 47, 0, 67, 0, 60, 0, 60, 0, 60, 0, 60, 216, 148, 221, 47, 0, 68, 0, 60, 0, 60, 0, 60, 0, 122, 51, 47, 0, 85, 0, 60, 0, 49, 1, 60, 0, 60, 0, 60, 0, 53, 216, 164, 222, 60, 0, 106, 2, 60, 0, 60, 0, 60, 0, 174, 167, 60, 0, 60, 0, 60, 0, 166, 29, 60, 0, 254, 167, 60, 0, 247, 167, 60, 0, 9, 29, 60, 0, 60, 0, 60, 0, 78, 29, 60, 0, 104, 2, 60, 0, 60, 0, 60, 0, 151, 1, 60, 0, 60, 0, 60, 0, 164, 29, 60, 0, 123, 29, 60, 0, 60, 0, 60, 0, 167, 29, 60, 0, 150, 29, 60, 0, 189, 167, 60, 0, 60, 0, 60, 0, 188, 167, 60, 0, 105, 2, 60, 0, 60, 0, 60, 0, 150, 1, 60, 0, 60, 0, 60, 0, 165, 29, 60, 0, 124, 29, 60, 0, 106, 0, 60, 0, 60, 0, 60, 0, 74, 255, 60, 0, 60, 0, 60, 0, 73, 33, 61, 0, 53, 216, 35, 220, 61, 0, 53, 216, 87, 220, 61, 0, 53, 216, 139, 220, 61, 0, 53, 216, 191, 220, 61, 0, 53, 216, 243, 220, 61, 0, 53, 216, 39, 221, 61, 0, 53, 216, 91, 221, 61, 0, 53, 216, 143, 221, 61, 0, 53, 216, 195, 221, 61, 0, 53, 216, 247, 221, 61, 0, 53, 216, 43, 222, 61, 0, 53, 216, 95, 222, 61, 0, 53, 216, 147, 222, 60, 0, 60, 0, 60, 0, 217, 36, 60, 0, 60, 0, 60, 0, 74, 0, 60, 0, 60, 0, 60, 0, 42, 255, 60, 0, 60, 0, 60, 0, 53, 216, 9, 220, 61, 0, 53, 216, 61, 220, 61, 0, 53, 216, 113, 220, 61, 0, 53, 216, 165, 220, 61, 0, 53, 216, 217, 220, 61, 0, 53, 216, 13, 221, 61, 0, 53, 216, 65, 221, 61, 0, 53, 216, 117, 221, 61, 0, 53, 216, 169, 221, 61, 0, 53, 216, 221, 221, 61, 0, 53, 216, 17, 222, 61, 0, 53, 216, 69, 222, 61, 0, 53, 216, 121, 222, 60, 0, 60, 0, 60, 0, 191, 36, 61, 0, 60, 216, 89, 221, 60, 0, 60, 0, 60, 0, 178, 2, 60, 0, 60, 0, 60, 0, 124, 44, 60, 0, 60, 0, 60, 0, 54, 29, 61, 0, 60, 216, 57, 221, 61, 0, 60, 216, 121, 221, 60, 0, 55, 2, 60, 0, 60, 0, 60, 0, 53, 216, 165, 222, 60, 0, 10, 29, 60, 0, 73, 2, 60, 0, 60, 0, 60, 0, 72, 2, 60, 0, 157, 2, 60, 0, 60, 0, 60, 0, 178, 167, 60, 0, 60, 0, 60, 0, 168, 29, 60, 0, 95, 2, 60, 0, 60, 0, 60, 0, 161, 29, 60, 0, 132, 2, 60, 0, 107, 0, 60, 0, 60, 0, 60, 0, 75, 255, 60, 0, 60, 0, 60, 0, 220, 29, 60, 0, 60, 0, 60, 0, 163, 167, 47, 0, 47, 216, 158, 220, 60, 0, 60, 0, 60, 0, 53, 216, 36, 220, 61, 0, 53, 216, 88, 220, 61, 0, 53, 216, 140, 220, 61, 0, 53, 216, 192, 220, 61, 0, 53, 216, 244, 220, 61, 0, 53, 216, 40, 221, 61, 0, 53, 216, 92, 221, 61, 0, 53, 216, 144, 221, 61, 0, 53, 216, 196, 221, 61, 0, 53, 216, 248, 221, 61, 0, 53, 216, 44, 222, 61, 0, 53, 216, 96, 222, 61, 0, 53, 216, 148, 222, 60, 0, 60, 0, 60, 0, 218, 36, 60, 0, 60, 0, 60, 0, 75, 0, 60, 0, 60, 0, 60, 0, 43, 255, 60, 0, 60, 0, 60, 0, 162, 167, 47, 0, 47, 216, 158, 220, 60, 0, 60, 0, 60, 0, 53, 216, 10, 220, 61, 0, 53, 216, 62, 220, 61, 0, 53, 216, 114, 220, 61, 0, 53, 216, 166, 220, 61, 0, 53, 216, 218, 220, 61, 0, 53, 216, 14, 221, 61, 0, 53, 216, 66, 221, 61, 0, 53, 216, 118, 221, 61, 0, 53, 216, 170, 221, 61, 0, 53, 216, 222, 221, 61, 0, 53, 216, 18, 222, 61, 0, 53, 216, 70, 222, 61, 0, 53, 216, 122, 222, 60, 0, 60, 0, 60, 0, 192, 36, 61, 0, 60, 216, 90, 221, 60, 0, 60, 0, 60, 0, 79, 29, 60, 0, 60, 0, 60, 0, 150, 32, 60, 0, 60, 0, 60, 0, 132, 51, 47, 0, 65, 0, 60, 0, 60, 0, 60, 0, 137, 51, 47, 0, 99, 0, 97, 0, 108, 0, 60, 0, 60, 0, 60, 0, 143, 51, 47, 0, 103, 0, 60, 0, 60, 0, 60, 0, 145, 51, 47, 0, 72, 0, 122, 0, 60, 0, 60, 0, 60, 0, 152, 51, 47, 0, 108, 0, 60, 0, 60, 0, 60, 0, 158, 51, 47, 0, 109, 0, 60, 0, 60, 0, 60, 0, 162, 51, 47, 0, 14, 32, 109, 0, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 166, 51, 47, 0, 14, 32, 109, 0, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 170, 51, 47, 0, 80, 0, 97, 0, 60, 0, 60, 0, 60, 0, 207, 51, 47, 0, 116, 0, 60, 0, 60, 0, 60, 0, 184, 51, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 190, 51, 47, 0, 87, 0, 60, 0, 60, 0, 60, 0, 192, 51, 47, 0, 169, 3, 60, 0, 60, 0, 60, 0, 55, 29, 61, 0, 60, 216, 58, 221, 61, 0, 60, 216, 122, 221, 60, 0, 60, 0, 60, 0, 133, 51, 47, 0, 66, 0, 60, 0, 60, 0, 60, 0, 205, 51, 47, 0, 75, 0, 60, 0, 60, 0, 60, 0, 206, 51, 47, 0, 77, 0, 60, 0, 11, 29, 60, 0, 132, 29, 60, 0, 153, 1, 60, 0, 60, 0, 60, 0, 152, 1, 60, 0, 106, 44, 60, 0, 60, 0, 60, 0, 105, 44, 60, 0, 65, 167, 60, 0, 60, 0, 60, 0, 64, 167, 60, 0, 67, 167, 60, 0, 60, 0, 60, 0, 66, 167, 60, 0, 69, 167, 60, 0, 60, 0, 60, 0, 68, 167, 60, 0, 158, 2, 60, 0, 60, 0, 60, 0, 176, 167, 60, 0, 108, 0, 60, 0, 60, 0, 60, 0, 66, 1, 47, 0, 53, 3, 60, 0, 60, 0, 60, 0, 76, 255, 60, 0, 60, 0, 60, 0, 221, 29, 61, 0, 124, 33, 60, 0, 60, 0, 60, 0, 201, 1, 47, 0, 106, 0, 60, 0, 60, 0, 60, 0, 251, 30, 47, 0, 108, 0, 60, 0, 60, 0, 60, 0, 170, 2, 47, 0, 115, 0, 60, 0, 60, 0, 60, 0, 171, 2, 47, 0, 122, 0, 60, 0, 60, 0, 60, 0, 19, 33, 61, 0, 53, 216, 37, 220, 61, 0, 53, 216, 89, 220, 61, 0, 53, 216, 141, 220, 61, 0, 53, 216, 193, 220, 61, 0, 53, 216, 245, 220, 61, 0, 53, 216, 41, 221, 61, 0, 53, 216, 93, 221, 61, 0, 53, 216, 145, 221, 61, 0, 53, 216, 197, 221, 61, 0, 53, 216, 249, 221, 61, 0, 53, 216, 45, 222, 61, 0, 53, 216, 97, 222, 61, 0, 53, 216, 149, 222, 60, 0, 60, 0, 60, 0, 219, 36, 60, 0, 60, 0, 60, 0, 76, 0, 60, 0, 60, 0, 60, 0, 65, 1, 47, 0, 53, 3, 60, 0, 60, 0, 60, 0, 44, 255, 60, 0, 60, 0, 60, 0, 108, 33, 60, 0, 60, 0, 60, 0, 200, 1, 47, 0, 106, 0, 60, 0, 60, 0, 60, 0, 199, 1, 47, 0, 74, 0, 60, 0, 60, 0, 60, 0, 250, 30, 47, 0, 76, 0, 60, 0, 60, 0, 60, 0, 18, 33, 61, 0, 53, 216, 11, 220, 61, 0, 53, 216, 63, 220, 61, 0, 53, 216, 115, 220, 61, 0, 53, 216, 219, 220, 61, 0, 53, 216, 15, 221, 61, 0, 53, 216, 67, 221, 61, 0, 53, 216, 119, 221, 61, 0, 53, 216, 171, 221, 61, 0, 53, 216, 223, 221, 61, 0, 53, 216, 19, 222, 61, 0, 53, 216, 71, 222, 61, 0, 53, 216, 123, 222, 60, 0, 60, 0, 60, 0, 193, 36, 61, 0, 60, 216, 91, 221, 60, 0, 60, 0, 60, 0, 225, 2, 60, 0, 60, 0, 60, 0, 151, 32, 60, 0, 60, 0, 60, 0, 208, 51, 47, 0, 109, 0, 60, 0, 60, 0, 60, 0, 209, 51, 47, 0, 110, 0, 60, 0, 60, 0, 60, 0, 210, 51, 47, 0, 111, 0, 103, 0, 60, 0, 60, 0, 60, 0, 211, 51, 47, 0, 120, 0, 60, 0, 60, 0, 60, 0, 56, 29, 61, 0, 60, 216, 59, 221, 61, 0, 60, 216, 123, 221, 60, 0, 60, 0, 60, 0, 60, 216, 169, 221, 47, 0, 111, 0, 115, 0, 115, 0, 108, 0, 101, 0, 115, 0, 115, 0, 60, 0, 60, 0, 60, 0, 207, 50, 47, 0, 84, 0, 68, 0, 60, 0, 60, 0, 108, 0, 183, 0, 61, 0, 108, 0, 183, 0, 61, 0, 64, 1, 60, 0, 60, 0, 60, 0, 76, 0, 183, 0, 61, 0, 76, 0, 183, 0, 61, 0, 63, 1, 60, 0, 159, 2, 60, 0, 60, 0, 60, 0, 222, 29, 60, 0, 60, 0, 60, 0, 171, 29, 60, 0, 71, 167, 60, 0, 60, 0, 60, 0, 70, 167, 60, 0, 12, 29, 60, 0, 73, 167, 60, 0, 60, 0, 60, 0, 72, 167, 60, 0, 154, 1, 60, 0, 60, 0, 60, 0, 61, 2, 60, 0, 97, 44, 60, 0, 60, 0, 60, 0, 96, 44, 60, 0, 107, 2, 60, 0, 60, 0, 60, 0, 98, 44, 60, 0, 60, 0, 60, 0, 94, 171, 60, 0, 56, 171, 60, 0, 60, 0, 60, 0, 236, 29, 60, 0, 57, 171, 60, 0, 108, 2, 60, 0, 60, 0, 60, 0, 173, 167, 60, 0, 55, 171, 60, 0, 60, 0, 60, 0, 93, 171, 60, 0, 133, 29, 60, 0, 60, 0, 60, 0, 170, 29, 60, 0, 109, 2, 60, 0, 60, 0, 60, 0, 169, 29, 60, 0, 142, 167, 60, 0, 52, 2, 60, 0, 114, 167, 60, 0, 110, 2, 60, 0, 129, 167, 60, 0, 60, 0, 60, 0, 128, 167, 60, 0, 155, 1, 60, 0, 142, 2, 60, 0, 109, 0, 60, 0, 60, 0, 60, 0, 77, 255, 60, 0, 60, 0, 60, 0, 107, 3, 61, 0, 127, 33, 60, 0, 60, 0, 60, 0, 53, 216, 38, 220, 61, 0, 53, 216, 90, 220, 61, 0, 53, 216, 142, 220, 61, 0, 53, 216, 194, 220, 61, 0, 53, 216, 246, 220, 61, 0, 53, 216, 42, 221, 61, 0, 53, 216, 94, 221, 61, 0, 53, 216, 146, 221, 61, 0, 53, 216, 198, 221, 61, 0, 53, 216, 250, 221, 61, 0, 53, 216, 46, 222, 61, 0, 53, 216, 98, 222, 61, 0, 53, 216, 150, 222, 60, 0, 60, 0, 60, 0, 220, 36, 60, 0, 60, 0, 60, 0, 77, 0, 60, 0, 60, 0, 60, 0, 45, 255, 60, 0, 60, 0, 60, 0, 111, 33, 60, 0, 60, 0, 60, 0, 51, 33, 61, 0, 53, 216, 12, 220, 61, 0, 53, 216, 64, 220, 61, 0, 53, 216, 116, 220, 61, 0, 53, 216, 220, 220, 61, 0, 53, 216, 16, 221, 61, 0, 53, 216, 68, 221, 61, 0, 53, 216, 120, 221, 61, 0, 53, 216, 172, 221, 61, 0, 53, 216, 224, 221, 61, 0, 53, 216, 20, 222, 61, 0, 53, 216, 72, 222, 61, 0, 53, 216, 124, 222, 60, 0, 60, 0, 60, 0, 194, 36, 61, 0, 60, 216, 92, 221, 60, 0, 60, 0, 60, 0, 80, 29, 60, 0, 60, 0, 60, 0, 60, 216, 106, 221, 47, 0, 99, 0, 60, 0, 60, 0, 60, 0, 60, 216, 107, 221, 47, 0, 100, 0, 60, 0, 60, 0, 60, 0, 60, 216, 108, 221, 47, 0, 114, 0, 60, 0, 60, 0, 60, 0, 152, 32, 60, 0, 60, 0, 60, 0, 167, 51, 47, 0, 21, 34, 115, 0, 60, 0, 60, 0, 60, 0, 168, 51, 47, 0, 14, 32, 21, 34, 115, 0, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 161, 51, 47, 0, 14, 32, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 165, 51, 47, 0, 14, 32, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 131, 51, 47, 0, 65, 0, 60, 0, 60, 0, 60, 0, 212, 51, 47, 0, 98, 0, 60, 0, 60, 0, 60, 0, 142, 51, 47, 0, 103, 0, 60, 0, 60, 0, 60, 0, 213, 51, 47, 0, 105, 0, 108, 0, 60, 0, 60, 0, 60, 0, 150, 51, 47, 0, 108, 0, 60, 0, 60, 0, 60, 0, 156, 51, 47, 0, 109, 0, 60, 0, 60, 0, 60, 0, 159, 51, 47, 0, 14, 32, 109, 0, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 163, 51, 47, 0, 14, 32, 109, 0, 59, 216, 3, 221, 14, 32, 60, 0, 60, 0, 60, 0, 214, 51, 47, 0, 111, 0, 108, 0, 60, 0, 60, 0, 60, 0, 179, 51, 47, 0, 115, 0, 60, 0, 60, 0, 60, 0, 183, 51, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 189, 51, 47, 0, 87, 0, 60, 0, 60, 0, 60, 0, 57, 29, 61, 0, 60, 216, 60, 221, 61, 0, 60, 216, 124, 221, 60, 0, 60, 0, 60, 0, 134, 51, 47, 0, 66, 0, 60, 0, 60, 0, 60, 0, 146, 51, 47, 0, 72, 0, 122, 0, 60, 0, 60, 0, 60, 0, 171, 51, 47, 0, 80, 0, 97, 0, 60, 0, 60, 0, 60, 0, 185, 51, 47, 0, 86, 0, 61, 0, 60, 216, 75, 221, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 191, 51, 47, 0, 87, 0, 60, 0, 60, 0, 60, 0, 193, 51, 47, 0, 169, 3, 60, 0, 13, 29, 60, 0, 60, 0, 60, 0, 223, 29, 60, 0, 111, 29, 60, 0, 134, 29, 60, 0, 113, 2, 60, 0, 60, 0, 60, 0, 110, 44, 60, 0, 60, 0, 60, 0, 172, 29, 60, 0, 58, 171, 60, 0, 253, 167, 60, 0, 255, 167, 60, 0, 115, 167, 60, 0, 110, 0, 60, 0, 60, 0, 60, 0, 78, 255, 60, 0, 60, 0, 60, 0, 224, 29, 60, 0, 60, 0, 60, 0, 165, 167, 47, 0, 47, 216, 158, 220, 60, 0, 60, 0, 60, 0, 204, 1, 47, 0, 106, 0, 60, 0, 60, 0, 60, 0, 53, 216, 39, 220, 61, 0, 53, 216, 91, 220, 61, 0, 53, 216, 143, 220, 61, 0, 53, 216, 195, 220, 61, 0, 53, 216, 247, 220, 61, 0, 53, 216, 43, 221, 61, 0, 53, 216, 95, 221, 61, 0, 53, 216, 147, 221, 61, 0, 53, 216, 199, 221, 61, 0, 53, 216, 251, 221, 61, 0, 53, 216, 47, 222, 61, 0, 53, 216, 99, 222, 61, 0, 53, 216, 151, 222, 60, 0, 60, 0, 60, 0, 221, 36, 60, 0, 60, 0, 60, 0, 78, 0, 60, 0, 60, 0, 60, 0, 46, 255, 60, 0, 60, 0, 60, 0, 164, 167, 47, 0, 47, 216, 158, 220, 60, 0, 60, 0, 60, 0, 203, 1, 47, 0, 106, 0, 60, 0, 60, 0, 60, 0, 202, 1, 47, 0, 74, 0, 60, 0, 60, 0, 60, 0, 22, 33, 47, 0, 111, 0, 60, 0, 60, 0, 60, 0, 21, 33, 61, 0, 53, 216, 13, 220, 61, 0, 53, 216, 65, 220, 61, 0, 53, 216, 117, 220, 61, 0, 53, 216, 169, 220, 61, 0, 53, 216, 221, 220, 61, 0, 53, 216, 17, 221, 61, 0, 53, 216, 121, 221, 61, 0, 53, 216, 173, 221, 61, 0, 53, 216, 225, 221, 61, 0, 53, 216, 21, 222, 61, 0, 53, 216, 73, 222, 61, 0, 53, 216, 125, 222, 60, 0, 60, 0, 60, 0, 195, 36, 61, 0, 60, 216, 93, 221, 60, 0, 60, 0, 60, 0, 127, 32, 60, 0, 60, 0, 60, 0, 153, 32, 60, 0, 60, 0, 60, 0, 129, 51, 47, 0, 65, 0, 60, 0, 60, 0, 60, 0, 139, 51, 47, 0, 70, 0, 60, 0, 60, 0, 60, 0, 154, 51, 47, 0, 109, 0, 60, 0, 60, 0, 60, 0, 177, 51, 47, 0, 115, 0, 60, 0, 60, 0, 60, 0, 181, 51, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 187, 51, 47, 0, 87, 0, 60, 0, 60, 0, 60, 0, 58, 29, 61, 0, 60, 216, 61, 221, 61, 0, 60, 216, 125, 221, 60, 0, 60, 0, 60, 0, 60, 216, 149, 221, 47, 0, 69, 0, 87, 0, 60, 0, 60, 0, 60, 0, 60, 216, 150, 221, 47, 0, 71, 0, 60, 0, 116, 2, 60, 0, 60, 0, 60, 0, 225, 29, 60, 0, 60, 0, 60, 0, 176, 29, 60, 0, 59, 29, 60, 0, 14, 29, 60, 0, 112, 29, 60, 0, 114, 2, 60, 0, 60, 0, 60, 0, 157, 1, 60, 0, 60, 0, 60, 0, 174, 29, 60, 0, 158, 1, 60, 0, 60, 0, 60, 0, 32, 2, 60, 0, 145, 167, 60, 0, 60, 0, 60, 0, 144, 167, 60, 0, 135, 29, 60, 0, 115, 2, 60, 0, 60, 0, 60, 0, 175, 29, 60, 0, 53, 2, 60, 0, 59, 171, 60, 0, 116, 167, 60, 0, 75, 1, 60, 0, 60, 0, 60, 0, 74, 1, 60, 0, 60, 0, 60, 0, 81, 29, 60, 0, 60, 171, 60, 0, 111, 0, 60, 0, 60, 0, 60, 0, 248, 0, 47, 0, 56, 3, 60, 0, 60, 0, 60, 0, 79, 255, 60, 0, 60, 0, 60, 0, 102, 3, 60, 0, 60, 0, 60, 0, 243, 29, 47, 0, 8, 3, 61, 0, 157, 167, 47, 0, 8, 3, 60, 0, 60, 0, 60, 0, 237, 29, 47, 0, 3, 216, 80, 223, 60, 0, 60, 0, 60, 0, 79, 167, 47, 0, 111, 0, 60, 0, 60, 0, 60, 0, 52, 33, 61, 0, 53, 216, 40, 220, 61, 0, 53, 216, 92, 220, 61, 0, 53, 216, 144, 220, 61, 0, 53, 216, 248, 220, 61, 0, 53, 216, 44, 221, 61, 0, 53, 216, 96, 221, 61, 0, 53, 216, 148, 221, 61, 0, 53, 216, 200, 221, 61, 0, 53, 216, 252, 221, 61, 0, 53, 216, 48, 222, 61, 0, 53, 216, 100, 222, 61, 0, 53, 216, 152, 222, 60, 0, 60, 0, 60, 0, 222, 36, 60, 0, 60, 0, 60, 0, 79, 0, 60, 0, 60, 0, 60, 0, 216, 0, 47, 0, 56, 3, 60, 0, 60, 0, 60, 0, 47, 255, 60, 0, 60, 0, 60, 0, 156, 167, 47, 0, 8, 3, 60, 0, 60, 0, 60, 0, 78, 167, 47, 0, 79, 0, 60, 0, 60, 0, 60, 0, 53, 216, 14, 220, 61, 0, 53, 216, 66, 220, 61, 0, 53, 216, 118, 220, 61, 0, 53, 216, 170, 220, 61, 0, 53, 216, 222, 220, 61, 0, 53, 216, 18, 221, 61, 0, 53, 216, 70, 221, 61, 0, 53, 216, 122, 221, 61, 0, 53, 216, 174, 221, 61, 0, 53, 216, 226, 221, 61, 0, 53, 216, 22, 222, 61, 0, 53, 216, 74, 222, 61, 0, 53, 216, 126, 222, 60, 0, 60, 0, 60, 0, 196, 36, 61, 0, 60, 216, 94, 221, 60, 0, 60, 0, 60, 0, 186, 0, 61, 0, 82, 29, 60, 0, 60, 0, 60, 0, 146, 32, 60, 0, 60, 0, 60, 0, 117, 51, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 60, 29, 61, 0, 60, 216, 62, 221, 61, 0, 60, 216, 126, 221, 60, 0, 60, 0, 60, 0, 60, 216, 151, 221, 47, 0, 75, 0, 60, 0, 60, 0, 83, 1, 47, 0, 101, 0, 60, 0, 60, 0, 60, 0, 82, 1, 47, 0, 69, 0, 60, 0, 60, 0, 60, 0, 249, 167, 47, 0, 101, 0, 60, 0, 15, 29, 60, 0, 17, 29, 60, 0, 61, 171, 60, 0, 118, 2, 60, 0, 20, 29, 60, 0, 65, 171, 60, 0, 66, 171, 60, 0, 64, 171, 60, 0, 67, 171, 60, 0, 68, 171, 60, 0, 19, 29, 60, 0, 62, 171, 60, 0, 84, 2, 60, 0, 60, 0, 60, 0, 134, 1, 60, 0, 60, 0, 60, 0, 83, 29, 60, 0, 16, 29, 60, 0, 18, 29, 60, 0, 63, 171, 60, 0, 151, 29, 60, 0, 98, 171, 60, 0, 77, 167, 60, 0, 60, 0, 60, 0, 76, 167, 60, 0, 22, 29, 60, 0, 60, 0, 60, 0, 84, 29, 60, 0, 23, 29, 60, 0, 60, 0, 60, 0, 85, 29, 60, 0, 122, 44, 60, 0, 117, 2, 60, 0, 60, 0, 60, 0, 159, 1, 60, 0, 60, 0, 60, 0, 177, 29, 60, 0, 75, 167, 60, 0, 60, 0, 60, 0, 74, 167, 60, 0, 119, 2, 60, 0, 183, 167, 60, 0, 60, 0, 60, 0, 182, 167, 60, 0, 35, 2, 60, 0, 60, 0, 60, 0, 34, 2, 60, 0, 60, 0, 60, 0, 61, 29, 60, 0, 21, 29, 60, 0, 112, 0, 60, 0, 60, 0, 60, 0, 80, 255, 60, 0, 60, 0, 60, 0, 238, 29, 60, 0, 60, 0, 60, 0, 53, 216, 41, 220, 61, 0, 53, 216, 93, 220, 61, 0, 53, 216, 145, 220, 61, 0, 53, 216, 197, 220, 61, 0, 53, 216, 249, 220, 61, 0, 53, 216, 45, 221, 61, 0, 53, 216, 97, 221, 61, 0, 53, 216, 149, 221, 61, 0, 53, 216, 201, 221, 61, 0, 53, 216, 253, 221, 61, 0, 53, 216, 49, 222, 61, 0, 53, 216, 101, 222, 61, 0, 53, 216, 153, 222, 60, 0, 60, 0, 60, 0, 223, 36, 60, 0, 60, 0, 60, 0, 80, 0, 60, 0, 60, 0, 60, 0, 48, 255, 60, 0, 60, 0, 60, 0, 25, 33, 61, 0, 53, 216, 15, 220, 61, 0, 53, 216, 67, 220, 61, 0, 53, 216, 119, 220, 61, 0, 53, 216, 171, 220, 61, 0, 53, 216, 223, 220, 61, 0, 53, 216, 19, 221, 61, 0, 53, 216, 123, 221, 61, 0, 53, 216, 175, 221, 61, 0, 53, 216, 227, 221, 61, 0, 53, 216, 23, 222, 61, 0, 53, 216, 75, 222, 61, 0, 53, 216, 127, 222, 60, 0, 60, 0, 60, 0, 197, 36, 61, 0, 60, 216, 95, 221, 60, 0, 60, 0, 60, 0, 86, 29, 60, 0, 60, 0, 60, 0, 154, 32, 60, 0, 60, 0, 60, 0, 216, 51, 47, 0, 39, 0, 46, 0, 39, 0, 109, 0, 39, 0, 46, 0, 39, 0, 60, 0, 60, 0, 60, 0, 128, 51, 47, 0, 65, 0, 60, 0, 60, 0, 60, 0, 118, 51, 47, 0, 99, 0, 60, 0, 60, 0, 60, 0, 138, 51, 47, 0, 70, 0, 60, 0, 60, 0, 60, 0, 176, 51, 47, 0, 115, 0, 60, 0, 60, 0, 60, 0, 180, 51, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 186, 51, 47, 0, 87, 0, 60, 0, 60, 0, 60, 0, 62, 29, 61, 0, 60, 216, 63, 221, 61, 0, 60, 216, 127, 221, 61, 0, 60, 216, 138, 221, 60, 0, 60, 0, 60, 0, 169, 51, 47, 0, 97, 0, 60, 0, 60, 0, 60, 0, 60, 216, 140, 221, 47, 0, 65, 0, 60, 0, 60, 0, 60, 0, 215, 51, 47, 0, 72, 0, 60, 0, 60, 0, 60, 0, 217, 51, 47, 0, 80, 0, 77, 0, 60, 0, 60, 0, 60, 0, 60, 216, 78, 221, 47, 0, 80, 0, 86, 0, 60, 0, 60, 0, 60, 0, 218, 51, 47, 0, 82, 0, 60, 0, 60, 0, 60, 0, 80, 50, 47, 0, 84, 0, 69, 0, 60, 0, 24, 29, 60, 0, 125, 29, 60, 0, 60, 0, 60, 0, 99, 44, 60, 0, 81, 167, 60, 0, 60, 0, 60, 0, 80, 167, 60, 0, 113, 29, 60, 0, 136, 29, 60, 0, 165, 1, 60, 0, 60, 0, 60, 0, 164, 1, 60, 0, 83, 167, 60, 0, 60, 0, 60, 0, 82, 167, 60, 0, 85, 167, 60, 0, 60, 0, 60, 0, 84, 167, 60, 0, 252, 167, 60, 0, 120, 2, 60, 0, 60, 0, 60, 0, 178, 29, 60, 0, 119, 44, 60, 0, 113, 0, 60, 0, 60, 0, 60, 0, 81, 255, 60, 0, 60, 0, 60, 0, 57, 2, 47, 0, 112, 0, 60, 0, 60, 0, 60, 0, 53, 216, 42, 220, 61, 0, 53, 216, 94, 220, 61, 0, 53, 216, 146, 220, 61, 0, 53, 216, 198, 220, 61, 0, 53, 216, 250, 220, 61, 0, 53, 216, 46, 221, 61, 0, 53, 216, 98, 221, 61, 0, 53, 216, 150, 221, 61, 0, 53, 216, 202, 221, 61, 0, 53, 216, 254, 221, 61, 0, 53, 216, 50, 222, 61, 0, 53, 216, 102, 222, 61, 0, 53, 216, 154, 222, 60, 0, 60, 0, 60, 0, 224, 36, 60, 0, 60, 0, 60, 0, 81, 0, 60, 0, 60, 0, 60, 0, 49, 255, 60, 0, 60, 0, 60, 0, 26, 33, 61, 0, 53, 216, 16, 220, 61, 0, 53, 216, 68, 220, 61, 0, 53, 216, 120, 220, 61, 0, 53, 216, 172, 220, 61, 0, 53, 216, 224, 220, 61, 0, 53, 216, 20, 221, 61, 0, 53, 216, 124, 221, 61, 0, 53, 216, 176, 221, 61, 0, 53, 216, 228, 221, 61, 0, 53, 216, 24, 222, 61, 0, 53, 216, 76, 222, 61, 0, 53, 216, 128, 222, 60, 0, 60, 0, 60, 0, 198, 36, 61, 0, 60, 216, 96, 221, 60, 0, 60, 0, 60, 0, 60, 216, 64, 221, 61, 0, 60, 216, 128, 221, 60, 0, 175, 167, 60, 0, 87, 167, 60, 0, 60, 0, 60, 0, 86, 167, 60, 0, 89, 167, 60, 0, 60, 0, 60, 0, 88, 167, 60, 0, 160, 2, 60, 0, 75, 2, 60, 0, 60, 0, 60, 0, 74, 2, 60, 0, 56, 1, 60, 0, 114, 0, 60, 0, 60, 0, 60, 0, 82, 255, 60, 0, 60, 0, 60, 0, 108, 3, 61, 0, 202, 29, 60, 0, 60, 0, 60, 0, 167, 167, 47, 0, 47, 216, 158, 220, 60, 0, 60, 0, 60, 0, 53, 216, 43, 220, 61, 0, 53, 216, 95, 220, 61, 0, 53, 216, 147, 220, 61, 0, 53, 216, 199, 220, 61, 0, 53, 216, 251, 220, 61, 0, 53, 216, 47, 221, 61, 0, 53, 216, 99, 221, 61, 0, 53, 216, 151, 221, 61, 0, 53, 216, 203, 221, 61, 0, 53, 216, 255, 221, 61, 0, 53, 216, 51, 222, 61, 0, 53, 216, 103, 222, 61, 0, 53, 216, 155, 222, 60, 0, 60, 0, 60, 0, 225, 36, 60, 0, 60, 0, 60, 0, 82, 0, 60, 0, 60, 0, 60, 0, 50, 255, 60, 0, 60, 0, 60, 0, 166, 167, 47, 0, 47, 216, 158, 220, 60, 0, 60, 0, 60, 0, 27, 33, 61, 0, 28, 33, 61, 0, 29, 33, 61, 0, 53, 216, 17, 220, 61, 0, 53, 216, 69, 220, 61, 0, 53, 216, 121, 220, 61, 0, 53, 216, 225, 220, 61, 0, 53, 216, 125, 221, 61, 0, 53, 216, 177, 221, 61, 0, 53, 216, 229, 221, 61, 0, 53, 216, 25, 222, 61, 0, 53, 216, 77, 222, 61, 0, 53, 216, 129, 222, 60, 0, 60, 0, 60, 0, 199, 36, 61, 0, 60, 216, 44, 221, 61, 0, 60, 216, 97, 221, 60, 0, 60, 0, 60, 0, 179, 2, 60, 0, 60, 0, 60, 0, 99, 29, 60, 0, 60, 0, 60, 0, 173, 51, 47, 0, 97, 0, 100, 0, 60, 0, 60, 0, 60, 0, 174, 51, 47, 0, 97, 0, 100, 0, 21, 34, 115, 0, 60, 0, 60, 0, 60, 0, 175, 51, 47, 0, 14, 32, 97, 0, 100, 0, 21, 34, 115, 0, 59, 216, 2, 221, 14, 32, 60, 0, 60, 0, 60, 0, 63, 29, 61, 0, 60, 216, 65, 221, 61, 0, 60, 216, 129, 221, 60, 0, 60, 0, 131, 167, 60, 0, 60, 0, 60, 0, 130, 167, 60, 0, 69, 171, 60, 0, 128, 2, 60, 0, 60, 0, 60, 0, 226, 29, 60, 0, 60, 0, 60, 0, 166, 1, 60, 0, 70, 171, 60, 0, 91, 167, 60, 0, 60, 0, 60, 0, 227, 29, 60, 0, 60, 0, 60, 0, 90, 167, 60, 0, 25, 29, 60, 0, 77, 2, 60, 0, 60, 0, 60, 0, 76, 2, 60, 0, 114, 29, 60, 0, 121, 2, 60, 0, 60, 0, 60, 0, 180, 2, 60, 0, 26, 29, 60, 0, 122, 2, 60, 0, 137, 29, 60, 0, 123, 2, 60, 0, 60, 0, 60, 0, 181, 2, 60, 0, 121, 44, 60, 0, 124, 2, 60, 0, 125, 2, 60, 0, 60, 0, 60, 0, 100, 44, 60, 0, 73, 171, 60, 0, 126, 2, 60, 0, 115, 29, 60, 0, 127, 2, 60, 0, 71, 171, 60, 0, 72, 171, 60, 0, 74, 171, 60, 0, 75, 171, 60, 0, 76, 171, 60, 0, 129, 2, 60, 0, 60, 0, 60, 0, 182, 2, 60, 0, 104, 171, 60, 0, 117, 167, 60, 0, 118, 167, 60, 0, 93, 167, 60, 0, 60, 0, 60, 0, 92, 167, 60, 0, 115, 0, 60, 0, 60, 0, 60, 0, 83, 255, 60, 0, 60, 0, 60, 0, 228, 29, 60, 0, 60, 0, 60, 0, 169, 167, 47, 0, 47, 216, 158, 220, 60, 0, 60, 0, 60, 0, 6, 251, 47, 0, 116, 0, 60, 0, 60, 0, 60, 0, 53, 216, 44, 220, 61, 0, 53, 216, 96, 220, 61, 0, 53, 216, 148, 220, 61, 0, 53, 216, 200, 220, 61, 0, 53, 216, 252, 220, 61, 0, 53, 216, 48, 221, 61, 0, 53, 216, 100, 221, 61, 0, 53, 216, 152, 221, 61, 0, 53, 216, 204, 221, 61, 0, 53, 216, 0, 222, 61, 0, 53, 216, 52, 222, 61, 0, 53, 216, 104, 222, 61, 0, 53, 216, 156, 222, 60, 0, 60, 0, 60, 0, 226, 36, 60, 0, 60, 0, 60, 0, 83, 0, 60, 0, 60, 0, 60, 0, 51, 255, 60, 0, 60, 0, 60, 0, 168, 167, 47, 0, 47, 216, 158, 220, 60, 0, 60, 0, 60, 0, 53, 216, 18, 220, 61, 0, 53, 216, 70, 220, 61, 0, 53, 216, 122, 220, 61, 0, 53, 216, 174, 220, 61, 0, 53, 216, 226, 220, 61, 0, 53, 216, 22, 221, 61, 0, 53, 216, 74, 221, 61, 0, 53, 216, 126, 221, 61, 0, 53, 216, 178, 221, 61, 0, 53, 216, 230, 221, 61, 0, 53, 216, 26, 222, 61, 0, 53, 216, 78, 222, 61, 0, 53, 216, 130, 222, 60, 0, 60, 0, 60, 0, 200, 36, 61, 0, 60, 216, 98, 221, 60, 0, 60, 0, 60, 0, 226, 2, 60, 0, 60, 0, 60, 0, 32, 33, 47, 0, 109, 0, 60, 0, 60, 0, 60, 0, 155, 32, 60, 0, 60, 0, 60, 0, 219, 51, 47, 0, 114, 0, 60, 0, 60, 0, 60, 0, 60, 216, 66, 221, 61, 0, 60, 216, 130, 221, 60, 0, 60, 0, 60, 0, 60, 216, 141, 221, 47, 0, 65, 0, 60, 0, 60, 0, 60, 0, 60, 216, 76, 221, 47, 0, 68, 0, 60, 0, 60, 0, 60, 0, 60, 216, 170, 221, 47, 0, 72, 0, 86, 0, 60, 0, 60, 0, 60, 0, 60, 216, 152, 221, 47, 0, 79, 0, 83, 0, 60, 0, 60, 0, 60, 0, 60, 216, 77, 221, 47, 0, 83, 0, 60, 0, 60, 0, 60, 0, 220, 51, 47, 0, 118, 0, 60, 0, 60, 0, 223, 0, 47, 0, 115, 0, 60, 0, 60, 0, 60, 0, 158, 30, 47, 0, 83, 0, 60, 0, 60, 0, 127, 1, 61, 0, 229, 29, 61, 0, 133, 167, 60, 0, 60, 0, 60, 0, 5, 251, 47, 0, 116, 0, 60, 0, 60, 0, 60, 0, 132, 167, 60, 0, 49, 167, 60, 0, 202, 167, 60, 0, 60, 0, 60, 0, 201, 167, 60, 0, 116, 29, 60, 0, 138, 29, 60, 0, 130, 2, 60, 0, 60, 0, 60, 0, 197, 167, 60, 0, 60, 0, 60, 0, 179, 29, 60, 0, 63, 2, 60, 0, 60, 0, 60, 0, 126, 44, 60, 0, 156, 30, 60, 0, 157, 30, 60, 0, 131, 2, 60, 0, 60, 0, 60, 0, 239, 29, 60, 0, 60, 0, 60, 0, 169, 1, 60, 0, 60, 0, 60, 0, 180, 29, 60, 0, 77, 171, 60, 0, 139, 29, 60, 0, 170, 1, 60, 0, 133, 2, 60, 0, 152, 29, 60, 0, 134, 2, 60, 0, 116, 0, 60, 0, 60, 0, 60, 0, 84, 255, 60, 0, 60, 0, 60, 0, 109, 3, 60, 0, 60, 0, 60, 0, 168, 2, 47, 0, 85, 2, 60, 0, 60, 0, 60, 0, 190, 1, 47, 0, 115, 0, 61, 0, 166, 2, 47, 0, 115, 0, 60, 0, 60, 0, 60, 0, 103, 171, 47, 0, 130, 2, 60, 0, 60, 0, 60, 0, 167, 2, 47, 0, 131, 2, 60, 0, 60, 0, 60, 0, 41, 167, 47, 0, 122, 0, 60, 0, 60, 0, 60, 0, 53, 216, 45, 220, 61, 0, 53, 216, 97, 220, 61, 0, 53, 216, 149, 220, 61, 0, 53, 216, 201, 220, 61, 0, 53, 216, 253, 220, 61, 0, 53, 216, 49, 221, 61, 0, 53, 216, 101, 221, 61, 0, 53, 216, 153, 221, 61, 0, 53, 216, 205, 221, 61, 0, 53, 216, 1, 222, 61, 0, 53, 216, 53, 222, 61, 0, 53, 216, 105, 222, 61, 0, 53, 216, 157, 222, 60, 0, 60, 0, 60, 0, 227, 36, 60, 0, 60, 0, 60, 0, 84, 0, 60, 0, 60, 0, 60, 0, 52, 255, 60, 0, 60, 0, 60, 0, 33, 33, 47, 0, 69, 0, 76, 0, 60, 0, 60, 0, 60, 0, 40, 167, 47, 0, 122, 0, 60, 0, 60, 0, 60, 0, 53, 216, 19, 220, 61, 0, 53, 216, 71, 220, 61, 0, 53, 216, 123, 220, 61, 0, 53, 216, 175, 220, 61, 0, 53, 216, 227, 220, 61, 0, 53, 216, 23, 221, 61, 0, 53, 216, 75, 221, 61, 0, 53, 216, 127, 221, 61, 0, 53, 216, 179, 221, 61, 0, 53, 216, 231, 221, 61, 0, 53, 216, 27, 222, 61, 0, 53, 216, 79, 222, 61, 0, 53, 216, 131, 222, 60, 0, 60, 0, 60, 0, 201, 36, 61, 0, 60, 216, 99, 221, 60, 0, 60, 0, 60, 0, 87, 29, 60, 0, 60, 0, 60, 0, 34, 33, 47, 0, 109, 0, 60, 0, 60, 0, 60, 0, 156, 32, 60, 0, 60, 0, 60, 0, 64, 29, 61, 0, 60, 216, 67, 221, 61, 0, 60, 216, 131, 221, 60, 0, 60, 0, 60, 0, 148, 51, 47, 0, 72, 0, 122, 0, 60, 0, 60, 0, 122, 29, 47, 0, 104, 0, 60, 0, 60, 0, 135, 167, 60, 0, 60, 0, 60, 0, 134, 167, 60, 0, 27, 29, 60, 0, 103, 1, 60, 0, 60, 0, 60, 0, 102, 1, 60, 0, 102, 44, 60, 0, 60, 0, 60, 0, 62, 2, 60, 0, 117, 29, 60, 0, 171, 1, 60, 0, 60, 0, 60, 0, 181, 29, 60, 0, 173, 1, 60, 0, 60, 0, 60, 0, 172, 1, 60, 0, 136, 2, 60, 0, 60, 0, 60, 0, 174, 1, 60, 0, 54, 2, 60, 0, 119, 167, 60, 0, 135, 2, 60, 0, 60, 0, 60, 0, 177, 167, 60, 0, 117, 0, 60, 0, 60, 0, 60, 0, 85, 255, 60, 0, 60, 0, 60, 0, 103, 3, 60, 0, 60, 0, 60, 0, 244, 29, 47, 0, 8, 3, 61, 0, 159, 167, 47, 0, 8, 3, 60, 0, 60, 0, 60, 0, 240, 29, 47, 0, 3, 216, 80, 223, 60, 0, 60, 0, 60, 0, 53, 216, 46, 220, 61, 0, 53, 216, 98, 220, 61, 0, 53, 216, 150, 220, 61, 0, 53, 216, 202, 220, 61, 0, 53, 216, 254, 220, 61, 0, 53, 216, 50, 221, 61, 0, 53, 216, 102, 221, 61, 0, 53, 216, 154, 221, 61, 0, 53, 216, 206, 221, 61, 0, 53, 216, 2, 222, 61, 0, 53, 216, 54, 222, 61, 0, 53, 216, 106, 222, 61, 0, 53, 216, 158, 222, 60, 0, 60, 0, 60, 0, 228, 36, 60, 0, 60, 0, 60, 0, 85, 0, 60, 0, 60, 0, 60, 0, 53, 255, 60, 0, 60, 0, 60, 0, 158, 167, 47, 0, 8, 3, 60, 0, 60, 0, 60, 0, 53, 216, 20, 220, 61, 0, 53, 216, 72, 220, 61, 0, 53, 216, 124, 220, 61, 0, 53, 216, 176, 220, 61, 0, 53, 216, 228, 220, 61, 0, 53, 216, 24, 221, 61, 0, 53, 216, 76, 221, 61, 0, 53, 216, 128, 221, 61, 0, 53, 216, 180, 221, 61, 0, 53, 216, 232, 221, 61, 0, 53, 216, 28, 222, 61, 0, 53, 216, 80, 222, 61, 0, 53, 216, 132, 222, 60, 0, 60, 0, 60, 0, 202, 36, 61, 0, 60, 216, 100, 221, 60, 0, 60, 0, 60, 0, 88, 29, 60, 0, 60, 0, 60, 0, 100, 29, 60, 0, 60, 0, 60, 0, 65, 29, 61, 0, 60, 216, 68, 221, 61, 0, 60, 216, 132, 221, 60, 0, 60, 0, 60, 0, 60, 216, 171, 221, 47, 0, 72, 0, 68, 0, 60, 0, 60, 0, 60, 0, 60, 216, 153, 221, 47, 0, 80, 0, 39, 0, 33, 0, 39, 0, 60, 0, 28, 29, 60, 0, 60, 0, 60, 0, 184, 29, 60, 0, 78, 171, 60, 0, 29, 29, 60, 0, 60, 0, 60, 0, 89, 29, 60, 0, 30, 29, 60, 0, 107, 29, 60, 0, 80, 171, 60, 0, 81, 171, 60, 0, 137, 2, 60, 0, 60, 0, 60, 0, 68, 2, 60, 0, 60, 0, 60, 0, 182, 29, 60, 0, 79, 171, 60, 0, 185, 167, 60, 0, 60, 0, 60, 0, 184, 167, 60, 0, 126, 29, 60, 0, 153, 29, 60, 0, 82, 171, 60, 0, 60, 0, 60, 0, 95, 171, 60, 0, 191, 167, 60, 0, 60, 0, 60, 0, 190, 167, 60, 0, 101, 2, 60, 0, 60, 0, 60, 0, 141, 167, 60, 0, 60, 0, 60, 0, 163, 29, 60, 0, 174, 2, 60, 0, 175, 2, 60, 0, 111, 2, 60, 0, 60, 0, 60, 0, 156, 1, 60, 0, 60, 0, 60, 0, 90, 29, 60, 0, 250, 167, 60, 0, 31, 29, 60, 0, 112, 2, 60, 0, 60, 0, 60, 0, 173, 29, 60, 0, 138, 2, 60, 0, 60, 0, 60, 0, 177, 1, 60, 0, 60, 0, 60, 0, 183, 29, 60, 0, 127, 29, 60, 0, 118, 0, 60, 0, 60, 0, 60, 0, 86, 255, 60, 0, 60, 0, 60, 0, 110, 3, 61, 0, 116, 33, 60, 0, 60, 0, 60, 0, 117, 33, 47, 0, 105, 0, 60, 0, 60, 0, 60, 0, 118, 33, 47, 0, 105, 0, 105, 0, 60, 0, 60, 0, 60, 0, 119, 33, 47, 0, 105, 0, 105, 0, 105, 0, 60, 0, 60, 0, 60, 0, 97, 167, 47, 0, 121, 0, 60, 0, 60, 0, 60, 0, 53, 216, 47, 220, 61, 0, 53, 216, 99, 220, 61, 0, 53, 216, 151, 220, 61, 0, 53, 216, 203, 220, 61, 0, 53, 216, 255, 220, 61, 0, 53, 216, 51, 221, 61, 0, 53, 216, 103, 221, 61, 0, 53, 216, 155, 221, 61, 0, 53, 216, 207, 221, 61, 0, 53, 216, 3, 222, 61, 0, 53, 216, 55, 222, 61, 0, 53, 216, 107, 222, 61, 0, 53, 216, 159, 222, 60, 0, 60, 0, 60, 0, 229, 36, 60, 0, 60, 0, 60, 0, 86, 0, 60, 0, 60, 0, 60, 0, 54, 255, 60, 0, 60, 0, 60, 0, 100, 33, 60, 0, 60, 0, 60, 0, 101, 33, 47, 0, 73, 0, 60, 0, 60, 0, 60, 0, 102, 33, 47, 0, 73, 0, 73, 0, 60, 0, 60, 0, 60, 0, 103, 33, 47, 0, 73, 0, 73, 0, 73, 0, 60, 0, 60, 0, 60, 0, 96, 167, 47, 0, 89, 0, 60, 0, 60, 0, 60, 0, 53, 216, 21, 220, 61, 0, 53, 216, 73, 220, 61, 0, 53, 216, 125, 220, 61, 0, 53, 216, 177, 220, 61, 0, 53, 216, 229, 220, 61, 0, 53, 216, 25, 221, 61, 0, 53, 216, 77, 221, 61, 0, 53, 216, 129, 221, 61, 0, 53, 216, 181, 221, 61, 0, 53, 216, 233, 221, 61, 0, 53, 216, 29, 222, 61, 0, 53, 216, 81, 222, 61, 0, 53, 216, 133, 222, 60, 0, 60, 0, 60, 0, 203, 36, 61, 0, 60, 216, 101, 221, 60, 0, 60, 0, 60, 0, 91, 29, 60, 0, 60, 0, 60, 0, 101, 29, 60, 0, 60, 0, 60, 0, 125, 44, 61, 0, 60, 216, 69, 221, 61, 0, 60, 216, 133, 221, 60, 0, 60, 0, 60, 0, 222, 51, 47, 0, 21, 34, 109, 0, 60, 0, 60, 0, 60, 0, 60, 216, 172, 221, 47, 0, 79, 0, 68, 0, 60, 0, 60, 0, 60, 0, 60, 216, 154, 221, 47, 0, 83, 0, 60, 0, 32, 29, 60, 0, 95, 167, 60, 0, 60, 0, 60, 0, 94, 167, 60, 0, 140, 29, 60, 0, 139, 2, 60, 0, 60, 0, 60, 0, 178, 1, 60, 0, 60, 0, 60, 0, 185, 29, 60, 0, 113, 44, 60, 0, 116, 44, 60, 0, 253, 30, 60, 0, 60, 0, 60, 0, 252, 30, 60, 0, 140, 2, 60, 0, 60, 0, 60, 0, 69, 2, 60, 0, 60, 0, 60, 0, 186, 29, 60, 0, 119, 0, 60, 0, 60, 0, 60, 0, 87, 255, 60, 0, 60, 0, 60, 0, 191, 26, 61, 0, 241, 29, 60, 0, 60, 0, 60, 0, 53, 216, 48, 220, 61, 0, 53, 216, 100, 220, 61, 0, 53, 216, 152, 220, 61, 0, 53, 216, 204, 220, 61, 0, 53, 216, 0, 221, 61, 0, 53, 216, 52, 221, 61, 0, 53, 216, 104, 221, 61, 0, 53, 216, 156, 221, 61, 0, 53, 216, 208, 221, 61, 0, 53, 216, 4, 222, 61, 0, 53, 216, 56, 222, 61, 0, 53, 216, 108, 222, 61, 0, 53, 216, 160, 222, 60, 0, 60, 0, 60, 0, 230, 36, 60, 0, 60, 0, 60, 0, 87, 0, 60, 0, 60, 0, 60, 0, 55, 255, 60, 0, 60, 0, 60, 0, 53, 216, 22, 220, 61, 0, 53, 216, 74, 220, 61, 0, 53, 216, 126, 220, 61, 0, 53, 216, 178, 220, 61, 0, 53, 216, 230, 220, 61, 0, 53, 216, 26, 221, 61, 0, 53, 216, 78, 221, 61, 0, 53, 216, 130, 221, 61, 0, 53, 216, 182, 221, 61, 0, 53, 216, 234, 221, 61, 0, 53, 216, 30, 222, 61, 0, 53, 216, 82, 222, 61, 0, 53, 216, 134, 222, 60, 0, 60, 0, 60, 0, 204, 36, 61, 0, 60, 216, 102, 221, 60, 0, 60, 0, 60, 0, 60, 216, 46, 221, 47, 0, 90, 0, 60, 0, 60, 0, 60, 0, 183, 2, 60, 0, 60, 0, 60, 0, 66, 29, 61, 0, 60, 216, 70, 221, 61, 0, 60, 216, 134, 221, 60, 0, 60, 0, 60, 0, 221, 51, 47, 0, 98, 0, 60, 0, 60, 0, 60, 0, 60, 216, 79, 221, 47, 0, 67, 0, 61, 0, 60, 216, 143, 221, 47, 0, 67, 0, 60, 0, 33, 29, 60, 0, 195, 167, 60, 0, 60, 0, 60, 0, 194, 167, 60, 0, 115, 44, 60, 0, 60, 0, 60, 0, 114, 44, 60, 0, 141, 2, 60, 0, 60, 0, 60, 0, 192, 26, 60, 0, 60, 0, 60, 0, 105, 171, 60, 0, 120, 0, 60, 0, 60, 0, 60, 0, 88, 255, 60, 0, 60, 0, 60, 0, 111, 3, 61, 0, 121, 33, 60, 0, 60, 0, 60, 0, 122, 33, 47, 0, 105, 0, 60, 0, 60, 0, 60, 0, 123, 33, 47, 0, 105, 0, 105, 0, 60, 0, 60, 0, 60, 0, 53, 216, 49, 220, 61, 0, 53, 216, 101, 220, 61, 0, 53, 216, 153, 220, 61, 0, 53, 216, 205, 220, 61, 0, 53, 216, 1, 221, 61, 0, 53, 216, 53, 221, 61, 0, 53, 216, 105, 221, 61, 0, 53, 216, 157, 221, 61, 0, 53, 216, 209, 221, 61, 0, 53, 216, 5, 222, 61, 0, 53, 216, 57, 222, 61, 0, 53, 216, 109, 222, 61, 0, 53, 216, 161, 222, 60, 0, 60, 0, 60, 0, 231, 36, 60, 0, 60, 0, 60, 0, 88, 0, 60, 0, 60, 0, 60, 0, 56, 255, 60, 0, 60, 0, 60, 0, 105, 33, 60, 0, 60, 0, 60, 0, 106, 33, 47, 0, 73, 0, 60, 0, 60, 0, 60, 0, 107, 33, 47, 0, 73, 0, 73, 0, 60, 0, 60, 0, 60, 0, 53, 216, 23, 220, 61, 0, 53, 216, 75, 220, 61, 0, 53, 216, 127, 220, 61, 0, 53, 216, 179, 220, 61, 0, 53, 216, 231, 220, 61, 0, 53, 216, 27, 221, 61, 0, 53, 216, 79, 221, 61, 0, 53, 216, 131, 221, 61, 0, 53, 216, 183, 221, 61, 0, 53, 216, 235, 221, 61, 0, 53, 216, 31, 222, 61, 0, 53, 216, 83, 222, 61, 0, 53, 216, 135, 222, 60, 0, 60, 0, 60, 0, 205, 36, 61, 0, 60, 216, 103, 221, 60, 0, 60, 0, 60, 0, 227, 2, 60, 0, 60, 0, 60, 0, 147, 32, 60, 0, 60, 0, 60, 0, 60, 216, 71, 221, 61, 0, 60, 216, 135, 221, 60, 0, 141, 29, 60, 0, 86, 171, 60, 0, 87, 171, 60, 0, 88, 171, 60, 0, 89, 171, 60, 0, 83, 171, 60, 0, 60, 0, 60, 0, 179, 167, 60, 0, 84, 171, 60, 0, 85, 171, 60, 0, 121, 0, 60, 0, 60, 0, 60, 0, 89, 255, 60, 0, 60, 0, 60, 0, 53, 216, 50, 220, 61, 0, 53, 216, 102, 220, 61, 0, 53, 216, 154, 220, 61, 0, 53, 216, 206, 220, 61, 0, 53, 216, 2, 221, 61, 0, 53, 216, 54, 221, 61, 0, 53, 216, 106, 221, 61, 0, 53, 216, 158, 221, 61, 0, 53, 216, 210, 221, 61, 0, 53, 216, 6, 222, 61, 0, 53, 216, 58, 222, 61, 0, 53, 216, 110, 222, 61, 0, 53, 216, 162, 222, 60, 0, 60, 0, 60, 0, 232, 36, 60, 0, 60, 0, 60, 0, 89, 0, 60, 0, 60, 0, 60, 0, 57, 255, 60, 0, 60, 0, 60, 0, 53, 216, 24, 220, 61, 0, 53, 216, 76, 220, 61, 0, 53, 216, 128, 220, 61, 0, 53, 216, 180, 220, 61, 0, 53, 216, 232, 220, 61, 0, 53, 216, 28, 221, 61, 0, 53, 216, 80, 221, 61, 0, 53, 216, 132, 221, 61, 0, 53, 216, 184, 221, 61, 0, 53, 216, 236, 221, 61, 0, 53, 216, 32, 222, 61, 0, 53, 216, 84, 222, 61, 0, 53, 216, 136, 222, 60, 0, 60, 0, 60, 0, 206, 36, 61, 0, 60, 216, 104, 221, 60, 0, 60, 0, 60, 0, 184, 2, 60, 0, 60, 0, 60, 0, 60, 216, 72, 221, 61, 0, 60, 216, 136, 221, 60, 0, 143, 2, 60, 0, 79, 2, 60, 0, 60, 0, 60, 0, 78, 2, 60, 0, 180, 1, 60, 0, 60, 0, 60, 0, 179, 1, 60, 0, 255, 30, 60, 0, 60, 0, 60, 0, 254, 30, 60, 0, 90, 171, 60, 0, 29, 2, 60, 0, 60, 0, 60, 0, 28, 2, 60, 0, 122, 0, 60, 0, 60, 0, 60, 0, 90, 255, 60, 0, 60, 0, 60, 0, 230, 29, 60, 0, 60, 0, 60, 0, 141, 1, 47, 0, 119, 0, 60, 0, 60, 0, 60, 0, 53, 216, 51, 220, 61, 0, 53, 216, 103, 220, 61, 0, 53, 216, 155, 220, 61, 0, 53, 216, 207, 220, 61, 0, 53, 216, 3, 221, 61, 0, 53, 216, 55, 221, 61, 0, 53, 216, 107, 221, 61, 0, 53, 216, 159, 221, 61, 0, 53, 216, 211, 221, 61, 0, 53, 216, 7, 222, 61, 0, 53, 216, 59, 222, 61, 0, 53, 216, 111, 222, 61, 0, 53, 216, 163, 222, 60, 0, 60, 0, 60, 0, 233, 36, 60, 0, 60, 0, 60, 0, 90, 0, 60, 0, 60, 0, 60, 0, 58, 255, 60, 0, 60, 0, 60, 0, 36, 33, 61, 0, 40, 33, 61, 0, 53, 216, 25, 220, 61, 0, 53, 216, 77, 220, 61, 0, 53, 216, 129, 220, 61, 0, 53, 216, 181, 220, 61, 0, 53, 216, 233, 220, 61, 0, 53, 216, 133, 221, 61, 0, 53, 216, 185, 221, 61, 0, 53, 216, 237, 221, 61, 0, 53, 216, 33, 222, 61, 0, 53, 216, 85, 222, 61, 0, 53, 216, 137, 222, 60, 0, 60, 0, 60, 0, 207, 36, 61, 0, 60, 216, 105, 221, 60, 0, 60, 0, 60, 0, 187, 29, 60, 0, 60, 0, 60, 0, 60, 216, 73, 221, 61, 0, 60, 216, 137, 221, 60, 0, 34, 29, 60, 0, 182, 1, 60, 0, 60, 0, 60, 0, 181, 1, 60, 0, 118, 29, 60, 0, 142, 29, 60, 0, 60, 0, 60, 0, 198, 167, 60, 0, 37, 2, 60, 0, 60, 0, 60, 0, 36, 2, 60, 0, 144, 2, 60, 0, 60, 0, 60, 0, 188, 29, 60, 0, 145, 2, 60, 0, 60, 0, 60, 0, 189, 29, 60, 0, 64, 2, 60, 0, 60, 0, 60, 0, 127, 44, 60, 0, 108, 44, 60, 0, 60, 0, 60, 0, 107, 44, 60, 0, 99, 167, 60, 0, 60, 0, 60, 0, 98, 167, 60, 0, 146, 2, 60, 0, 60, 0, 60, 0, 183, 1, 60, 0, 60, 0, 60, 0, 190, 29, 60, 0, 35, 29, 60, 0, 185, 1, 60, 0, 60, 0, 60, 0, 184, 1, 60, 0, 154, 29, 60, 0, 186, 1, 60, 0, 147, 2, 60, 0, 254, 0, 60, 0, 60, 0, 60, 0, 222, 0, 60, 0, 101, 167, 60, 0, 60, 0, 60, 0, 100, 167, 60, 0, 103, 167, 60, 0, 60, 0, 60, 0, 102, 167, 60, 0, 191, 1, 60, 0, 60, 0, 60, 0, 247, 1, 60, 0, 105, 167, 60, 0, 60, 0, 60, 0, 104, 167, 60, 0, 96, 171, 60, 0, 97, 171, 60, 0, 99, 171, 60, 0, 107, 167, 60, 0, 60, 0, 60, 0, 106, 167, 60, 0, 109, 167, 60, 0, 60, 0, 60, 0, 108, 167, 60, 0, 111, 167, 60, 0, 60, 0, 60, 0, 210, 29, 60, 0, 60, 0, 60, 0, 110, 167, 60, 0, 60, 0, 60, 0, 112, 167, 60, 0, 120, 167, 60, 0, 187, 1, 60, 0, 43, 167, 60, 0, 60, 0, 60, 0, 42, 167, 60, 0, 45, 167, 60, 0, 60, 0, 60, 0, 44, 167, 60, 0, 47, 167, 60, 0, 60, 0, 60, 0, 46, 167, 60, 0, 168, 1, 60, 0, 60, 0, 60, 0, 167, 1, 60, 0, 189, 1, 60, 0, 60, 0, 60, 0, 188, 1, 60, 0, 133, 1, 60, 0, 60, 0, 60, 0, 132, 1, 60, 0, 148, 2, 60, 0, 66, 2, 60, 0, 60, 0, 60, 0, 65, 2, 60, 0, 192, 2, 60, 0, 188, 2, 60, 0, 60, 0, 60, 0, 73, 1, 47, 0, 110, 0, 60, 0, 238, 2, 60, 0, 190, 2, 60, 0, 35, 167, 60, 0, 60, 0, 60, 0, 34, 167, 60, 0, 140, 167, 60, 0, 60, 0, 60, 0, 139, 167, 60, 0, 143, 167, 60, 0, 149, 2, 60, 0, 60, 0, 60, 0, 228, 2, 60, 0, 191, 2, 60, 0, 193, 2, 60, 0, 36, 29, 60, 0, 37, 29, 60, 0, 60, 0, 60, 0, 92, 29, 60, 0, 37, 167, 60, 0, 60, 0, 60, 0, 36, 167, 60, 0, 161, 2, 60, 0, 162, 2, 60, 0, 150, 2, 60, 0, 192, 1, 60, 0, 193, 1, 60, 0, 194, 1, 60, 0, 195, 1, 60, 0, 151, 2, 60, 0, 152, 2, 60, 0, 172, 2, 60, 0, 173, 2, 60, 0, 177, 3, 60, 0, 60, 0, 60, 0, 53, 216, 194, 222, 61, 0, 53, 216, 252, 222, 61, 0, 53, 216, 54, 223, 61, 0, 53, 216, 112, 223, 61, 0, 53, 216, 170, 223, 60, 0, 60, 0, 60, 0, 145, 3, 60, 0, 60, 0, 60, 0, 53, 216, 168, 222, 61, 0, 53, 216, 226, 222, 61, 0, 53, 216, 28, 223, 61, 0, 53, 216, 86, 223, 61, 0, 53, 216, 144, 223, 60, 0, 178, 3, 60, 0, 60, 0, 60, 0, 208, 3, 60, 0, 60, 0, 60, 0, 53, 216, 195, 222, 61, 0, 53, 216, 253, 222, 61, 0, 53, 216, 55, 223, 61, 0, 53, 216, 113, 223, 61, 0, 53, 216, 171, 223, 60, 0, 60, 0, 60, 0, 146, 3, 60, 0, 60, 0, 60, 0, 53, 216, 169, 222, 61, 0, 53, 216, 227, 222, 61, 0, 53, 216, 29, 223, 61, 0, 53, 216, 87, 223, 61, 0, 53, 216, 145, 223, 60, 0, 60, 0, 60, 0, 93, 29, 60, 0, 60, 0, 60, 0, 102, 29, 60, 0, 179, 3, 60, 0, 60, 0, 60, 0, 61, 33, 61, 0, 53, 216, 196, 222, 61, 0, 53, 216, 254, 222, 61, 0, 53, 216, 56, 223, 61, 0, 53, 216, 114, 223, 61, 0, 53, 216, 172, 223, 60, 0, 60, 0, 60, 0, 147, 3, 60, 0, 60, 0, 60, 0, 62, 33, 61, 0, 53, 216, 170, 222, 61, 0, 53, 216, 228, 222, 61, 0, 53, 216, 30, 223, 61, 0, 53, 216, 88, 223, 61, 0, 53, 216, 146, 223, 60, 0, 60, 0, 60, 0, 94, 29, 60, 0, 60, 0, 60, 0, 103, 29, 60, 0, 38, 29, 60, 0, 180, 3, 60, 0, 60, 0, 60, 0, 53, 216, 197, 222, 61, 0, 53, 216, 255, 222, 61, 0, 53, 216, 57, 223, 61, 0, 53, 216, 115, 223, 61, 0, 53, 216, 173, 223, 60, 0, 60, 0, 60, 0, 148, 3, 60, 0, 60, 0, 60, 0, 53, 216, 171, 222, 61, 0, 53, 216, 229, 222, 61, 0, 53, 216, 31, 223, 61, 0, 53, 216, 89, 223, 61, 0, 53, 216, 147, 223, 60, 0, 60, 0, 60, 0, 95, 29, 60, 0, 181, 3, 60, 0, 60, 0, 60, 0, 245, 3, 60, 0, 60, 0, 60, 0, 53, 216, 198, 222, 61, 0, 53, 216, 220, 222, 61, 0, 53, 216, 0, 223, 61, 0, 53, 216, 22, 223, 61, 0, 53, 216, 58, 223, 61, 0, 53, 216, 80, 223, 61, 0, 53, 216, 116, 223, 61, 0, 53, 216, 138, 223, 61, 0, 53, 216, 174, 223, 61, 0, 53, 216, 196, 223, 60, 0, 60, 0, 60, 0, 149, 3, 60, 0, 60, 0, 60, 0, 53, 216, 172, 222, 61, 0, 53, 216, 230, 222, 61, 0, 53, 216, 32, 223, 61, 0, 53, 216, 90, 223, 61, 0, 53, 216, 148, 223, 60, 0, 221, 3, 60, 0, 60, 0, 60, 0, 53, 216, 203, 223, 60, 0, 60, 0, 60, 0, 220, 3, 60, 0, 60, 0, 60, 0, 53, 216, 202, 223, 60, 0, 119, 3, 60, 0, 60, 0, 60, 0, 118, 3, 60, 0, 219, 3, 60, 0, 60, 0, 60, 0, 218, 3, 60, 0, 182, 3, 60, 0, 60, 0, 60, 0, 53, 216, 199, 222, 61, 0, 53, 216, 1, 223, 61, 0, 53, 216, 59, 223, 61, 0, 53, 216, 117, 223, 61, 0, 53, 216, 175, 223, 60, 0, 60, 0, 60, 0, 150, 3, 60, 0, 60, 0, 60, 0, 53, 216, 173, 222, 61, 0, 53, 216, 231, 222, 61, 0, 53, 216, 33, 223, 61, 0, 53, 216, 91, 223, 61, 0, 53, 216, 149, 223, 60, 0, 113, 3, 60, 0, 60, 0, 60, 0, 112, 3, 60, 0, 183, 3, 60, 0, 60, 0, 60, 0, 53, 216, 200, 222, 61, 0, 53, 216, 2, 223, 61, 0, 53, 216, 60, 223, 61, 0, 53, 216, 118, 223, 61, 0, 53, 216, 176, 223, 60, 0, 60, 0, 60, 0, 151, 3, 60, 0, 60, 0, 60, 0, 53, 216, 174, 222, 61, 0, 53, 216, 232, 222, 61, 0, 53, 216, 34, 223, 61, 0, 53, 216, 92, 223, 61, 0, 53, 216, 150, 223, 60, 0, 184, 3, 60, 0, 60, 0, 60, 0, 209, 3, 60, 0, 60, 0, 60, 0, 53, 216, 201, 222, 61, 0, 53, 216, 221, 222, 61, 0, 53, 216, 3, 223, 61, 0, 53, 216, 23, 223, 61, 0, 53, 216, 61, 223, 61, 0, 53, 216, 81, 223, 61, 0, 53, 216, 119, 223, 61, 0, 53, 216, 139, 223, 61, 0, 53, 216, 177, 223, 61, 0, 53, 216, 197, 223, 60, 0, 60, 0, 60, 0, 152, 3, 60, 0, 60, 0, 60, 0, 244, 3, 60, 0, 60, 0, 60, 0, 53, 216, 175, 222, 61, 0, 53, 216, 185, 222, 61, 0, 53, 216, 233, 222, 61, 0, 53, 216, 243, 222, 61, 0, 53, 216, 35, 223, 61, 0, 53, 216, 45, 223, 61, 0, 53, 216, 93, 223, 61, 0, 53, 216, 103, 223, 61, 0, 53, 216, 151, 223, 61, 0, 53, 216, 161, 223, 60, 0, 60, 0, 60, 0, 191, 29, 60, 0, 185, 3, 60, 0, 60, 0, 60, 0, 122, 3, 60, 0, 60, 0, 60, 0, 53, 216, 202, 222, 61, 0, 53, 216, 4, 223, 61, 0, 53, 216, 62, 223, 61, 0, 53, 216, 120, 223, 61, 0, 53, 216, 178, 223, 60, 0, 60, 0, 60, 0, 153, 3, 60, 0, 60, 0, 60, 0, 53, 216, 176, 222, 61, 0, 53, 216, 234, 222, 61, 0, 53, 216, 36, 223, 61, 0, 53, 216, 94, 223, 61, 0, 53, 216, 152, 223, 60, 0, 243, 3, 60, 0, 60, 0, 60, 0, 127, 3, 60, 0, 186, 3, 60, 0, 60, 0, 60, 0, 240, 3, 60, 0, 60, 0, 60, 0, 215, 3, 47, 0, 177, 3, 185, 3, 60, 0, 60, 0, 60, 0, 53, 216, 203, 222, 61, 0, 53, 216, 222, 222, 61, 0, 53, 216, 5, 223, 61, 0, 53, 216, 24, 223, 61, 0, 53, 216, 63, 223, 61, 0, 53, 216, 82, 223, 61, 0, 53, 216, 121, 223, 61, 0, 53, 216, 140, 223, 61, 0, 53, 216, 179, 223, 61, 0, 53, 216, 198, 223, 60, 0, 60, 0, 60, 0, 154, 3, 60, 0, 60, 0, 60, 0, 207, 3, 47, 0, 177, 3, 185, 3, 60, 0, 60, 0, 60, 0, 53, 216, 177, 222, 61, 0, 53, 216, 235, 222, 61, 0, 53, 216, 37, 223, 61, 0, 53, 216, 95, 223, 61, 0, 53, 216, 153, 223, 60, 0, 187, 3, 60, 0, 60, 0, 60, 0, 53, 216, 204, 222, 61, 0, 53, 216, 6, 223, 61, 0, 53, 216, 64, 223, 61, 0, 53, 216, 122, 223, 61, 0, 53, 216, 180, 223, 60, 0, 60, 0, 60, 0, 155, 3, 60, 0, 60, 0, 60, 0, 53, 216, 178, 222, 61, 0, 53, 216, 236, 222, 61, 0, 53, 216, 38, 223, 61, 0, 53, 216, 96, 223, 61, 0, 53, 216, 154, 223, 60, 0, 39, 29, 60, 0, 188, 3, 60, 0, 60, 0, 60, 0, 181, 0, 60, 0, 60, 0, 60, 0, 53, 216, 205, 222, 61, 0, 53, 216, 7, 223, 61, 0, 53, 216, 65, 223, 61, 0, 53, 216, 123, 223, 61, 0, 53, 216, 181, 223, 60, 0, 60, 0, 60, 0, 156, 3, 60, 0, 60, 0, 60, 0, 53, 216, 179, 222, 61, 0, 53, 216, 237, 222, 61, 0, 53, 216, 39, 223, 61, 0, 53, 216, 97, 223, 61, 0, 53, 216, 155, 223, 60, 0, 60, 0, 60, 0, 130, 51, 47, 0, 65, 0, 60, 0, 60, 0, 60, 0, 140, 51, 47, 0, 70, 0, 60, 0, 60, 0, 60, 0, 141, 51, 47, 0, 103, 0, 60, 0, 60, 0, 60, 0, 149, 51, 47, 0, 108, 0, 60, 0, 60, 0, 60, 0, 155, 51, 47, 0, 109, 0, 60, 0, 60, 0, 60, 0, 178, 51, 47, 0, 115, 0, 60, 0, 60, 0, 60, 0, 182, 51, 47, 0, 86, 0, 60, 0, 60, 0, 60, 0, 188, 51, 47, 0, 87, 0, 60, 0, 189, 3, 60, 0, 60, 0, 60, 0, 53, 216, 206, 222, 61, 0, 53, 216, 8, 223, 61, 0, 53, 216, 66, 223, 61, 0, 53, 216, 124, 223, 61, 0, 53, 216, 182, 223, 60, 0, 60, 0, 60, 0, 157, 3, 60, 0, 60, 0, 60, 0, 53, 216, 180, 222, 61, 0, 53, 216, 238, 222, 61, 0, 53, 216, 40, 223, 61, 0, 53, 216, 98, 223, 61, 0, 53, 216, 156, 223, 60, 0, 190, 3, 60, 0, 60, 0, 60, 0, 53, 216, 207, 222, 61, 0, 53, 216, 9, 223, 61, 0, 53, 216, 67, 223, 61, 0, 53, 216, 125, 223, 61, 0, 53, 216, 183, 223, 60, 0, 60, 0, 60, 0, 158, 3, 60, 0, 60, 0, 60, 0, 53, 216, 181, 222, 61, 0, 53, 216, 239, 222, 61, 0, 53, 216, 41, 223, 61, 0, 53, 216, 99, 223, 61, 0, 53, 216, 157, 223, 60, 0, 191, 3, 60, 0, 60, 0, 60, 0, 53, 216, 208, 222, 61, 0, 53, 216, 10, 223, 61, 0, 53, 216, 68, 223, 61, 0, 53, 216, 126, 223, 61, 0, 53, 216, 184, 223, 60, 0, 60, 0, 60, 0, 159, 3, 60, 0, 60, 0, 60, 0, 53, 216, 182, 222, 61, 0, 53, 216, 240, 222, 61, 0, 53, 216, 42, 223, 61, 0, 53, 216, 100, 223, 61, 0, 53, 216, 158, 223, 60, 0, 192, 3, 60, 0, 60, 0, 60, 0, 214, 3, 60, 0, 60, 0, 60, 0, 60, 33, 61, 0, 53, 216, 209, 222, 61, 0, 53, 216, 225, 222, 61, 0, 53, 216, 11, 223, 61, 0, 53, 216, 27, 223, 61, 0, 53, 216, 69, 223, 61, 0, 53, 216, 85, 223, 61, 0, 53, 216, 127, 223, 61, 0, 53, 216, 143, 223, 61, 0, 53, 216, 185, 223, 61, 0, 53, 216, 201, 223, 60, 0, 60, 0, 60, 0, 160, 3, 60, 0, 60, 0, 60, 0, 63, 33, 61, 0, 53, 216, 183, 222, 61, 0, 53, 216, 241, 222, 61, 0, 53, 216, 43, 223, 61, 0, 53, 216, 101, 223, 61, 0, 53, 216, 159, 223, 60, 0, 40, 29, 60, 0, 251, 3, 60, 0, 60, 0, 60, 0, 250, 3, 60, 0, 223, 3, 60, 0, 60, 0, 60, 0, 222, 3, 60, 0, 217, 3, 60, 0, 60, 0, 60, 0, 216, 3, 60, 0, 193, 3, 60, 0, 60, 0, 60, 0, 241, 3, 60, 0, 60, 0, 60, 0, 53, 216, 210, 222, 61, 0, 53, 216, 224, 222, 61, 0, 53, 216, 12, 223, 61, 0, 53, 216, 26, 223, 61, 0, 53, 216, 70, 223, 61, 0, 53, 216, 84, 223, 61, 0, 53, 216, 128, 223, 61, 0, 53, 216, 142, 223, 61, 0, 53, 216, 186, 223, 61, 0, 53, 216, 200, 223, 60, 0, 60, 0, 60, 0, 161, 3, 60, 0, 60, 0, 60, 0, 53, 216, 184, 222, 61, 0, 53, 216, 242, 222, 61, 0, 53, 216, 44, 223, 61, 0, 53, 216, 102, 223, 61, 0, 53, 216, 160, 223, 60, 0, 60, 0, 60, 0, 104, 29, 60, 0, 41, 29, 60, 0, 252, 3, 60, 0, 195, 3, 60, 0, 60, 0, 60, 0, 242, 3, 60, 0, 60, 0, 60, 0, 53, 216, 211, 222, 61, 0, 53, 216, 212, 222, 61, 0, 53, 216, 13, 223, 61, 0, 53, 216, 14, 223, 61, 0, 53, 216, 71, 223, 61, 0, 53, 216, 72, 223, 61, 0, 53, 216, 129, 223, 61, 0, 53, 216, 130, 223, 61, 0, 53, 216, 187, 223, 61, 0, 53, 216, 188, 223, 60, 0, 60, 0, 60, 0, 163, 3, 60, 0, 60, 0, 60, 0, 249, 3, 60, 0, 60, 0, 60, 0, 53, 216, 186, 222, 61, 0, 53, 216, 244, 222, 61, 0, 53, 216, 46, 223, 61, 0, 53, 216, 104, 223, 61, 0, 53, 216, 162, 223, 60, 0, 60, 0, 60, 0, 194, 3, 60, 0, 124, 3, 60, 0, 60, 0, 60, 0, 254, 3, 60, 0, 123, 3, 60, 0, 60, 0, 60, 0, 253, 3, 60, 0, 125, 3, 60, 0, 60, 0, 60, 0, 255, 3, 60, 0, 196, 3, 60, 0, 60, 0, 60, 0, 53, 216, 213, 222, 61, 0, 53, 216, 15, 223, 61, 0, 53, 216, 73, 223, 61, 0, 53, 216, 131, 223, 61, 0, 53, 216, 189, 223, 60, 0, 60, 0, 60, 0, 164, 3, 60, 0, 60, 0, 60, 0, 53, 216, 187, 222, 61, 0, 53, 216, 245, 222, 61, 0, 53, 216, 47, 223, 61, 0, 53, 216, 105, 223, 61, 0, 53, 216, 163, 223, 60, 0, 197, 3, 60, 0, 60, 0, 60, 0, 53, 216, 214, 222, 61, 0, 53, 216, 16, 223, 61, 0, 53, 216, 74, 223, 61, 0, 53, 216, 132, 223, 61, 0, 53, 216, 190, 223, 60, 0, 60, 0, 60, 0, 165, 3, 60, 0, 60, 0, 60, 0, 210, 3, 60, 0, 60, 0, 60, 0, 53, 216, 188, 222, 61, 0, 53, 216, 246, 222, 61, 0, 53, 216, 48, 223, 61, 0, 53, 216, 106, 223, 61, 0, 53, 216, 164, 223, 60, 0, 198, 3, 60, 0, 60, 0, 60, 0, 213, 3, 60, 0, 60, 0, 60, 0, 53, 216, 215, 222, 61, 0, 53, 216, 223, 222, 61, 0, 53, 216, 17, 223, 61, 0, 53, 216, 25, 223, 61, 0, 53, 216, 75, 223, 61, 0, 53, 216, 83, 223, 61, 0, 53, 216, 133, 223, 61, 0, 53, 216, 141, 223, 61, 0, 53, 216, 191, 223, 61, 0, 53, 216, 199, 223, 60, 0, 60, 0, 60, 0, 166, 3, 60, 0, 60, 0, 60, 0, 53, 216, 189, 222, 61, 0, 53, 216, 247, 222, 61, 0, 53, 216, 49, 223, 61, 0, 53, 216, 107, 223, 61, 0, 53, 216, 165, 223, 60, 0, 60, 0, 60, 0, 96, 29, 60, 0, 60, 0, 60, 0, 105, 29, 60, 0, 199, 3, 60, 0, 60, 0, 60, 0, 53, 216, 216, 222, 61, 0, 53, 216, 18, 223, 61, 0, 53, 216, 76, 223, 61, 0, 53, 216, 134, 223, 61, 0, 53, 216, 192, 223, 60, 0, 60, 0, 60, 0, 167, 3, 60, 0, 60, 0, 60, 0, 53, 216, 190, 222, 61, 0, 53, 216, 248, 222, 61, 0, 53, 216, 50, 223, 61, 0, 53, 216, 108, 223, 61, 0, 53, 216, 166, 223, 60, 0, 60, 0, 60, 0, 97, 29, 60, 0, 60, 0, 60, 0, 106, 29, 60, 0, 200, 3, 60, 0, 60, 0, 60, 0, 53, 216, 217, 222, 61, 0, 53, 216, 19, 223, 61, 0, 53, 216, 77, 223, 61, 0, 53, 216, 135, 223, 61, 0, 53, 216, 193, 223, 60, 0, 60, 0, 60, 0, 168, 3, 60, 0, 60, 0, 60, 0, 53, 216, 191, 222, 61, 0, 53, 216, 249, 222, 61, 0, 53, 216, 51, 223, 61, 0, 53, 216, 109, 223, 61, 0, 53, 216, 167, 223, 60, 0, 42, 29, 60, 0, 201, 3, 60, 0, 60, 0, 60, 0, 53, 216, 218, 222, 61, 0, 53, 216, 20, 223, 61, 0, 53, 216, 78, 223, 61, 0, 53, 216, 136, 223, 61, 0, 53, 216, 194, 223, 60, 0, 60, 0, 60, 0, 169, 3, 60, 0, 60, 0, 60, 0, 53, 216, 192, 222, 61, 0, 53, 216, 250, 222, 61, 0, 53, 216, 52, 223, 61, 0, 53, 216, 110, 223, 61, 0, 53, 216, 168, 223, 60, 0, 101, 171, 60, 0, 225, 3, 60, 0, 60, 0, 60, 0, 224, 3, 60, 0, 115, 3, 60, 0, 60, 0, 60, 0, 114, 3, 60, 0, 248, 3, 60, 0, 60, 0, 60, 0, 247, 3, 60, 0, 129, 44, 60, 0, 60, 0, 60, 0, 128, 44, 60, 0, 131, 44, 60, 0, 60, 0, 60, 0, 130, 44, 60, 0, 133, 44, 60, 0, 60, 0, 60, 0, 132, 44, 60, 0, 135, 44, 60, 0, 60, 0, 60, 0, 134, 44, 60, 0, 137, 44, 60, 0, 60, 0, 60, 0, 136, 44, 60, 0, 183, 44, 60, 0, 60, 0, 60, 0, 182, 44, 60, 0, 139, 44, 60, 0, 60, 0, 60, 0, 138, 44, 60, 0, 141, 44, 60, 0, 60, 0, 60, 0, 140, 44, 60, 0, 143, 44, 60, 0, 60, 0, 60, 0, 142, 44, 60, 0, 145, 44, 60, 0, 60, 0, 60, 0, 144, 44, 60, 0, 147, 44, 60, 0, 60, 0, 60, 0, 146, 44, 60, 0, 149, 44, 60, 0, 60, 0, 60, 0, 228, 44, 47, 0, 129, 44, 147, 44, 60, 0, 60, 0, 60, 0, 148, 44, 60, 0, 185, 44, 60, 0, 60, 0, 60, 0, 184, 44, 60, 0, 151, 44, 60, 0, 60, 0, 60, 0, 150, 44, 60, 0, 153, 44, 60, 0, 60, 0, 60, 0, 152, 44, 60, 0, 155, 44, 60, 0, 60, 0, 60, 0, 154, 44, 60, 0, 187, 44, 60, 0, 60, 0, 60, 0, 186, 44, 60, 0, 189, 44, 60, 0, 60, 0, 60, 0, 188, 44, 60, 0, 157, 44, 60, 0, 60, 0, 60, 0, 156, 44, 60, 0, 159, 44, 60, 0, 60, 0, 60, 0, 158, 44, 60, 0, 161, 44, 60, 0, 60, 0, 60, 0, 160, 44, 60, 0, 163, 44, 60, 0, 60, 0, 60, 0, 162, 44, 60, 0, 165, 44, 60, 0, 60, 0, 60, 0, 164, 44, 60, 0, 167, 44, 60, 0, 60, 0, 60, 0, 166, 44, 60, 0, 169, 44, 60, 0, 60, 0, 60, 0, 168, 44, 60, 0, 171, 44, 60, 0, 60, 0, 60, 0, 170, 44, 60, 0, 173, 44, 60, 0, 60, 0, 60, 0, 172, 44, 60, 0, 175, 44, 60, 0, 60, 0, 60, 0, 174, 44, 60, 0, 177, 44, 60, 0, 60, 0, 60, 0, 176, 44, 60, 0, 191, 44, 60, 0, 60, 0, 60, 0, 190, 44, 60, 0, 193, 44, 60, 0, 60, 0, 60, 0, 192, 44, 60, 0, 227, 3, 60, 0, 60, 0, 60, 0, 226, 3, 60, 0, 236, 44, 60, 0, 60, 0, 60, 0, 235, 44, 60, 0, 195, 44, 60, 0, 60, 0, 60, 0, 194, 44, 60, 0, 197, 44, 60, 0, 60, 0, 60, 0, 196, 44, 60, 0, 199, 44, 60, 0, 60, 0, 60, 0, 198, 44, 60, 0, 229, 3, 60, 0, 60, 0, 60, 0, 228, 3, 60, 0, 231, 3, 60, 0, 60, 0, 60, 0, 230, 3, 60, 0, 243, 44, 60, 0, 60, 0, 60, 0, 242, 44, 60, 0, 201, 44, 60, 0, 60, 0, 60, 0, 200, 44, 60, 0, 233, 3, 60, 0, 60, 0, 60, 0, 232, 3, 60, 0, 203, 44, 60, 0, 60, 0, 60, 0, 202, 44, 60, 0, 205, 44, 60, 0, 60, 0, 60, 0, 204, 44, 60, 0, 207, 44, 60, 0, 60, 0, 60, 0, 206, 44, 60, 0, 209, 44, 60, 0, 60, 0, 60, 0, 208, 44, 60, 0, 211, 44, 60, 0, 60, 0, 60, 0, 210, 44, 60, 0, 213, 44, 60, 0, 60, 0, 60, 0, 212, 44, 60, 0, 235, 3, 60, 0, 60, 0, 60, 0, 234, 3, 60, 0, 238, 44, 60, 0, 60, 0, 60, 0, 237, 44, 60, 0, 215, 44, 60, 0, 60, 0, 60, 0, 214, 44, 60, 0, 237, 3, 60, 0, 60, 0, 60, 0, 236, 3, 60, 0, 217, 44, 60, 0, 60, 0, 60, 0, 216, 44, 60, 0, 219, 44, 60, 0, 60, 0, 60, 0, 218, 44, 60, 0, 221, 44, 60, 0, 60, 0, 60, 0, 220, 44, 60, 0, 239, 3, 60, 0, 60, 0, 60, 0, 238, 3, 60, 0, 179, 44, 60, 0, 60, 0, 60, 0, 178, 44, 60, 0, 181, 44, 60, 0, 60, 0, 60, 0, 180, 44, 60, 0, 223, 44, 60, 0, 60, 0, 60, 0, 222, 44, 60, 0, 225, 44, 60, 0, 60, 0, 60, 0, 224, 44, 60, 0, 227, 44, 60, 0, 60, 0, 60, 0, 226, 44, 60, 0, 48, 4, 60, 0, 60, 0, 60, 0, 246, 45, 60, 0, 60, 0, 60, 0, 16, 4, 60, 0, 217, 4, 60, 0, 60, 0, 60, 0, 216, 4, 60, 0, 213, 4, 60, 0, 60, 0, 60, 0, 212, 4, 60, 0, 49, 4, 60, 0, 60, 0, 60, 0, 224, 45, 60, 0, 60, 0, 60, 0, 17, 4, 60, 0, 50, 4, 60, 0, 60, 0, 60, 0, 128, 28, 61, 0, 225, 45, 60, 0, 60, 0, 60, 0, 18, 4, 60, 0, 51, 4, 60, 0, 60, 0, 60, 0, 226, 45, 60, 0, 60, 0, 60, 0, 19, 4, 60, 0, 60, 0, 145, 4, 60, 0, 60, 0, 60, 0, 144, 4, 60, 0, 147, 4, 60, 0, 60, 0, 60, 0, 146, 4, 60, 0, 251, 4, 60, 0, 60, 0, 60, 0, 250, 4, 60, 0, 149, 4, 60, 0, 60, 0, 60, 0, 148, 4, 60, 0, 247, 4, 60, 0, 60, 0, 60, 0, 246, 4, 60, 0, 52, 4, 60, 0, 60, 0, 60, 0, 129, 28, 61, 0, 227, 45, 60, 0, 60, 0, 60, 0, 20, 4, 60, 0, 1, 5, 60, 0, 60, 0, 60, 0, 0, 5, 60, 0, 129, 166, 60, 0, 60, 0, 60, 0, 128, 166, 60, 0, 82, 4, 60, 0, 60, 0, 60, 0, 2, 4, 60, 0, 99, 166, 60, 0, 60, 0, 60, 0, 98, 166, 60, 0, 3, 5, 60, 0, 60, 0, 60, 0, 2, 5, 60, 0, 153, 4, 60, 0, 60, 0, 60, 0, 152, 4, 60, 0, 53, 4, 60, 0, 60, 0, 60, 0, 247, 45, 60, 0, 60, 0, 60, 0, 21, 4, 60, 0, 84, 4, 60, 0, 60, 0, 60, 0, 116, 166, 60, 0, 60, 0, 60, 0, 4, 4, 60, 0, 54, 4, 60, 0, 60, 0, 60, 0, 228, 45, 60, 0, 60, 0, 60, 0, 22, 4, 60, 0, 43, 5, 60, 0, 60, 0, 60, 0, 42, 5, 60, 0, 133, 166, 60, 0, 60, 0, 60, 0, 132, 166, 60, 0, 151, 4, 60, 0, 60, 0, 60, 0, 150, 4, 60, 0, 55, 4, 60, 0, 60, 0, 60, 0, 229, 45, 60, 0, 60, 0, 60, 0, 23, 4, 60, 0, 65, 166, 60, 0, 60, 0, 60, 0, 64, 166, 60, 0, 5, 5, 60, 0, 60, 0, 60, 0, 4, 5, 60, 0, 17, 5, 60, 0, 60, 0, 60, 0, 16, 5, 60, 0, 67, 166, 60, 0, 60, 0, 60, 0, 66, 166, 60, 0, 85, 4, 60, 0, 60, 0, 60, 0, 5, 4, 60, 0, 69, 166, 60, 0, 60, 0, 60, 0, 68, 166, 60, 0, 225, 4, 60, 0, 60, 0, 60, 0, 224, 4, 60, 0, 137, 166, 60, 0, 60, 0, 60, 0, 136, 166, 60, 0, 7, 5, 60, 0, 60, 0, 60, 0, 6, 5, 60, 0, 131, 166, 60, 0, 60, 0, 60, 0, 130, 166, 60, 0, 56, 4, 60, 0, 60, 0, 60, 0, 117, 166, 60, 0, 60, 0, 60, 0, 24, 4, 60, 0, 139, 4, 60, 0, 60, 0, 60, 0, 138, 4, 60, 0, 86, 4, 60, 0, 60, 0, 60, 0, 118, 166, 47, 0, 8, 3, 60, 0, 60, 0, 60, 0, 6, 4, 60, 0, 71, 166, 60, 0, 60, 0, 60, 0, 70, 166, 60, 0, 57, 4, 60, 0, 60, 0, 60, 0, 25, 4, 60, 0, 88, 4, 60, 0, 60, 0, 60, 0, 8, 4, 60, 0, 73, 166, 60, 0, 60, 0, 60, 0, 248, 45, 60, 0, 60, 0, 60, 0, 72, 166, 60, 0, 58, 4, 60, 0, 60, 0, 60, 0, 230, 45, 60, 0, 60, 0, 60, 0, 26, 4, 60, 0, 155, 4, 60, 0, 60, 0, 60, 0, 154, 4, 60, 0, 196, 4, 60, 0, 60, 0, 60, 0, 195, 4, 60, 0, 161, 4, 60, 0, 60, 0, 60, 0, 160, 4, 60, 0, 159, 4, 60, 0, 60, 0, 60, 0, 158, 4, 60, 0, 157, 4, 60, 0, 60, 0, 60, 0, 156, 4, 60, 0, 31, 5, 60, 0, 60, 0, 60, 0, 30, 5, 60, 0, 27, 5, 60, 0, 60, 0, 60, 0, 26, 5, 60, 0, 59, 4, 60, 0, 60, 0, 60, 0, 231, 45, 60, 0, 60, 0, 60, 0, 27, 4, 60, 0, 43, 29, 60, 0, 198, 4, 60, 0, 60, 0, 60, 0, 197, 4, 60, 0, 47, 5, 60, 0, 60, 0, 60, 0, 46, 5, 60, 0, 19, 5, 60, 0, 60, 0, 60, 0, 18, 5, 60, 0, 33, 5, 60, 0, 60, 0, 60, 0, 32, 5, 60, 0, 89, 4, 60, 0, 60, 0, 60, 0, 9, 4, 60, 0, 101, 166, 60, 0, 60, 0, 60, 0, 100, 166, 60, 0, 9, 5, 60, 0, 60, 0, 60, 0, 8, 5, 60, 0, 21, 5, 60, 0, 60, 0, 60, 0, 20, 5, 60, 0, 60, 4, 60, 0, 60, 0, 60, 0, 232, 45, 60, 0, 60, 0, 60, 0, 28, 4, 60, 0, 206, 4, 60, 0, 60, 0, 60, 0, 205, 4, 60, 0, 103, 166, 60, 0, 60, 0, 60, 0, 102, 166, 60, 0, 61, 4, 60, 0, 60, 0, 60, 0, 233, 45, 60, 0, 60, 0, 60, 0, 29, 4, 60, 0, 60, 0, 60, 0, 120, 29, 60, 0, 41, 5, 60, 0, 60, 0, 60, 0, 40, 5, 60, 0, 202, 4, 60, 0, 60, 0, 60, 0, 201, 4, 60, 0, 163, 4, 60, 0, 60, 0, 60, 0, 162, 4, 60, 0, 200, 4, 60, 0, 60, 0, 60, 0, 199, 4, 60, 0, 35, 5, 60, 0, 60, 0, 60, 0, 34, 5, 60, 0, 165, 4, 60, 0, 60, 0, 60, 0, 164, 4, 60, 0, 90, 4, 60, 0, 60, 0, 60, 0, 10, 4, 60, 0, 11, 5, 60, 0, 60, 0, 60, 0, 10, 5, 60, 0, 62, 4, 60, 0, 60, 0, 60, 0, 130, 28, 61, 0, 234, 45, 61, 0, 105, 166, 61, 0, 107, 166, 61, 0, 109, 166, 61, 0, 110, 166, 61, 0, 153, 166, 61, 0, 155, 166, 60, 0, 60, 0, 60, 0, 30, 4, 60, 0, 60, 0, 60, 0, 104, 166, 61, 0, 106, 166, 61, 0, 108, 166, 61, 0, 152, 166, 61, 0, 154, 166, 60, 0, 233, 4, 60, 0, 60, 0, 60, 0, 232, 4, 60, 0, 63, 4, 60, 0, 60, 0, 60, 0, 235, 45, 60, 0, 60, 0, 60, 0, 31, 4, 60, 0, 37, 5, 60, 0, 60, 0, 60, 0, 36, 5, 60, 0, 167, 4, 60, 0, 60, 0, 60, 0, 166, 4, 60, 0, 129, 4, 60, 0, 60, 0, 60, 0, 128, 4, 60, 0, 64, 4, 60, 0, 60, 0, 60, 0, 236, 45, 60, 0, 60, 0, 60, 0, 32, 4, 60, 0, 143, 4, 60, 0, 60, 0, 60, 0, 142, 4, 60, 0, 23, 5, 60, 0, 60, 0, 60, 0, 22, 5, 60, 0, 65, 4, 60, 0, 60, 0, 60, 0, 131, 28, 61, 0, 237, 45, 60, 0, 60, 0, 60, 0, 245, 45, 47, 0, 66, 4, 60, 0, 60, 0, 60, 0, 33, 4, 60, 0, 13, 5, 60, 0, 60, 0, 60, 0, 12, 5, 60, 0, 171, 4, 60, 0, 60, 0, 60, 0, 170, 4, 60, 0, 66, 4, 60, 0, 60, 0, 60, 0, 132, 28, 61, 0, 133, 28, 61, 0, 238, 45, 60, 0, 60, 0, 60, 0, 34, 4, 60, 0, 141, 166, 60, 0, 60, 0, 60, 0, 140, 166, 60, 0, 15, 5, 60, 0, 60, 0, 60, 0, 14, 5, 60, 0, 173, 4, 60, 0, 60, 0, 60, 0, 172, 4, 60, 0, 139, 166, 60, 0, 60, 0, 60, 0, 138, 166, 60, 0, 91, 4, 60, 0, 60, 0, 60, 0, 11, 4, 60, 0, 67, 4, 60, 0, 60, 0, 60, 0, 119, 166, 60, 0, 60, 0, 60, 0, 35, 4, 60, 0, 175, 4, 60, 0, 60, 0, 60, 0, 174, 4, 60, 0, 177, 4, 60, 0, 60, 0, 60, 0, 176, 4, 60, 0, 75, 166, 60, 0, 60, 0, 60, 0, 136, 28, 61, 0, 249, 45, 60, 0, 60, 0, 60, 0, 74, 166, 60, 0, 121, 4, 60, 0, 60, 0, 60, 0, 120, 4, 60, 0, 68, 4, 60, 0, 60, 0, 60, 0, 158, 166, 60, 0, 60, 0, 60, 0, 36, 4, 60, 0, 69, 4, 60, 0, 60, 0, 60, 0, 239, 45, 60, 0, 60, 0, 60, 0, 37, 4, 60, 0, 253, 4, 60, 0, 60, 0, 60, 0, 252, 4, 60, 0, 255, 4, 60, 0, 60, 0, 60, 0, 254, 4, 60, 0, 179, 4, 60, 0, 60, 0, 60, 0, 178, 4, 60, 0, 187, 4, 60, 0, 60, 0, 60, 0, 186, 4, 60, 0, 39, 5, 60, 0, 60, 0, 60, 0, 38, 5, 60, 0, 149, 166, 60, 0, 60, 0, 60, 0, 148, 166, 60, 0, 97, 4, 60, 0, 60, 0, 60, 0, 123, 166, 60, 0, 60, 0, 60, 0, 96, 4, 60, 0, 127, 4, 60, 0, 60, 0, 60, 0, 126, 4, 60, 0, 77, 166, 60, 0, 60, 0, 60, 0, 76, 166, 60, 0, 125, 4, 60, 0, 60, 0, 60, 0, 124, 4, 60, 0, 123, 4, 60, 0, 60, 0, 60, 0, 122, 4, 60, 0, 70, 4, 60, 0, 60, 0, 60, 0, 240, 45, 60, 0, 60, 0, 60, 0, 38, 4, 60, 0, 97, 166, 60, 0, 60, 0, 60, 0, 96, 166, 60, 0, 143, 166, 60, 0, 60, 0, 60, 0, 142, 166, 60, 0, 181, 4, 60, 0, 60, 0, 60, 0, 180, 4, 60, 0, 145, 166, 60, 0, 60, 0, 60, 0, 144, 166, 60, 0, 71, 4, 60, 0, 60, 0, 60, 0, 241, 45, 60, 0, 60, 0, 60, 0, 39, 4, 60, 0, 45, 5, 60, 0, 60, 0, 60, 0, 44, 5, 60, 0, 147, 166, 60, 0, 60, 0, 60, 0, 146, 166, 60, 0, 183, 4, 60, 0, 60, 0, 60, 0, 182, 4, 60, 0, 204, 4, 60, 0, 60, 0, 60, 0, 203, 4, 60, 0, 185, 4, 60, 0, 60, 0, 60, 0, 184, 4, 60, 0, 135, 166, 60, 0, 60, 0, 60, 0, 134, 166, 60, 0, 189, 4, 60, 0, 60, 0, 60, 0, 188, 4, 60, 0, 191, 4, 60, 0, 60, 0, 60, 0, 190, 4, 60, 0, 95, 4, 60, 0, 60, 0, 60, 0, 15, 4, 60, 0, 72, 4, 60, 0, 60, 0, 60, 0, 242, 45, 60, 0, 60, 0, 60, 0, 40, 4, 60, 0, 151, 166, 60, 0, 60, 0, 60, 0, 150, 166, 60, 0, 73, 4, 60, 0, 60, 0, 60, 0, 243, 45, 60, 0, 60, 0, 60, 0, 41, 4, 60, 0, 79, 166, 60, 0, 60, 0, 60, 0, 78, 166, 60, 0, 47, 46, 60, 0, 127, 166, 60, 0, 74, 4, 60, 0, 60, 0, 60, 0, 134, 28, 61, 0, 120, 166, 60, 0, 60, 0, 60, 0, 42, 4, 60, 0, 60, 0, 60, 0, 156, 166, 60, 0, 81, 166, 60, 0, 60, 0, 60, 0, 80, 166, 60, 0, 75, 4, 60, 0, 60, 0, 60, 0, 121, 166, 60, 0, 60, 0, 60, 0, 43, 4, 60, 0, 76, 4, 60, 0, 60, 0, 60, 0, 122, 166, 60, 0, 60, 0, 60, 0, 44, 4, 60, 0, 60, 0, 60, 0, 157, 166, 60, 0, 141, 4, 60, 0, 60, 0, 60, 0, 140, 4, 60, 0, 99, 4, 60, 0, 60, 0, 60, 0, 135, 28, 61, 0, 250, 45, 60, 0, 60, 0, 60, 0, 98, 4, 60, 0, 83, 166, 60, 0, 60, 0, 60, 0, 82, 166, 60, 0, 77, 4, 60, 0, 60, 0, 60, 0, 45, 4, 60, 0, 78, 4, 60, 0, 60, 0, 60, 0, 251, 45, 60, 0, 60, 0, 60, 0, 46, 4, 60, 0, 85, 166, 60, 0, 60, 0, 60, 0, 84, 166, 60, 0, 87, 166, 60, 0, 60, 0, 60, 0, 252, 45, 60, 0, 60, 0, 60, 0, 86, 166, 60, 0, 79, 4, 60, 0, 60, 0, 60, 0, 47, 4, 60, 0, 25, 5, 60, 0, 60, 0, 60, 0, 24, 5, 60, 0, 101, 4, 60, 0, 60, 0, 60, 0, 159, 166, 60, 0, 60, 0, 60, 0, 100, 4, 60, 0, 103, 4, 60, 0, 60, 0, 60, 0, 253, 45, 60, 0, 60, 0, 60, 0, 102, 4, 60, 0, 89, 166, 60, 0, 60, 0, 60, 0, 88, 166, 60, 0, 107, 4, 60, 0, 60, 0, 60, 0, 254, 45, 60, 0, 60, 0, 60, 0, 106, 4, 60, 0, 91, 166, 60, 0, 60, 0, 60, 0, 90, 166, 60, 0, 105, 4, 60, 0, 60, 0, 60, 0, 104, 4, 60, 0, 93, 166, 60, 0, 60, 0, 60, 0, 92, 166, 60, 0, 109, 4, 60, 0, 60, 0, 60, 0, 255, 45, 60, 0, 60, 0, 60, 0, 108, 4, 60, 0, 111, 4, 60, 0, 60, 0, 60, 0, 110, 4, 60, 0, 113, 4, 60, 0, 60, 0, 60, 0, 112, 4, 60, 0, 115, 4, 60, 0, 60, 0, 60, 0, 244, 45, 60, 0, 60, 0, 60, 0, 114, 4, 60, 0, 117, 4, 60, 0, 60, 0, 60, 0, 116, 4, 60, 0, 95, 166, 60, 0, 60, 0, 60, 0, 94, 166, 60, 0, 169, 4, 60, 0, 60, 0, 60, 0, 168, 4, 60, 0, 29, 5, 60, 0, 60, 0, 60, 0, 28, 5, 60, 0, 207, 4, 60, 0, 60, 0, 60, 0, 192, 4, 60, 0, 48, 44, 60, 0, 60, 0, 60, 0, 56, 216, 0, 220, 60, 0, 60, 0, 60, 0, 0, 44, 60, 0, 49, 44, 60, 0, 60, 0, 60, 0, 56, 216, 1, 220, 60, 0, 60, 0, 60, 0, 1, 44, 60, 0, 50, 44, 60, 0, 60, 0, 60, 0, 56, 216, 2, 220, 60, 0, 60, 0, 60, 0, 2, 44, 60, 0, 51, 44, 60, 0, 60, 0, 60, 0, 56, 216, 3, 220, 60, 0, 60, 0, 60, 0, 3, 44, 60, 0, 52, 44, 60, 0, 60, 0, 60, 0, 56, 216, 4, 220, 60, 0, 60, 0, 60, 0, 4, 44, 60, 0, 53, 44, 60, 0, 60, 0, 60, 0, 56, 216, 5, 220, 60, 0, 60, 0, 60, 0, 5, 44, 60, 0, 54, 44, 60, 0, 60, 0, 60, 0, 56, 216, 6, 220, 60, 0, 60, 0, 60, 0, 6, 44, 60, 0, 55, 44, 60, 0, 60, 0, 60, 0, 7, 44, 60, 0, 56, 44, 60, 0, 60, 0, 60, 0, 56, 216, 8, 220, 60, 0, 60, 0, 60, 0, 8, 44, 60, 0, 57, 44, 60, 0, 60, 0, 60, 0, 56, 216, 9, 220, 60, 0, 60, 0, 60, 0, 9, 44, 60, 0, 58, 44, 60, 0, 60, 0, 60, 0, 56, 216, 10, 220, 60, 0, 60, 0, 60, 0, 10, 44, 60, 0, 59, 44, 60, 0, 60, 0, 60, 0, 56, 216, 11, 220, 60, 0, 60, 0, 60, 0, 11, 44, 60, 0, 60, 44, 60, 0, 60, 0, 60, 0, 56, 216, 12, 220, 60, 0, 60, 0, 60, 0, 12, 44, 60, 0, 61, 44, 60, 0, 60, 0, 60, 0, 56, 216, 13, 220, 60, 0, 60, 0, 60, 0, 13, 44, 60, 0, 62, 44, 60, 0, 60, 0, 60, 0, 56, 216, 14, 220, 60, 0, 60, 0, 60, 0, 14, 44, 60, 0, 63, 44, 60, 0, 60, 0, 60, 0, 56, 216, 15, 220, 60, 0, 60, 0, 60, 0, 15, 44, 60, 0, 64, 44, 60, 0, 60, 0, 60, 0, 56, 216, 16, 220, 60, 0, 60, 0, 60, 0, 16, 44, 60, 0, 65, 44, 60, 0, 60, 0, 60, 0, 56, 216, 17, 220, 60, 0, 60, 0, 60, 0, 17, 44, 60, 0, 66, 44, 60, 0, 60, 0, 60, 0, 56, 216, 18, 220, 60, 0, 60, 0, 60, 0, 18, 44, 60, 0, 67, 44, 60, 0, 60, 0, 60, 0, 56, 216, 19, 220, 60, 0, 60, 0, 60, 0, 19, 44, 60, 0, 68, 44, 60, 0, 60, 0, 60, 0, 56, 216, 20, 220, 60, 0, 60, 0, 60, 0, 20, 44, 60, 0, 69, 44, 60, 0, 60, 0, 60, 0, 56, 216, 21, 220, 60, 0, 60, 0, 60, 0, 21, 44, 60, 0, 70, 44, 60, 0, 60, 0, 60, 0, 56, 216, 22, 220, 60, 0, 60, 0, 60, 0, 22, 44, 60, 0, 71, 44, 60, 0, 60, 0, 60, 0, 56, 216, 23, 220, 60, 0, 60, 0, 60, 0, 23, 44, 60, 0, 72, 44, 60, 0, 60, 0, 60, 0, 56, 216, 24, 220, 60, 0, 60, 0, 60, 0, 24, 44, 60, 0, 73, 44, 60, 0, 60, 0, 60, 0, 25, 44, 60, 0, 74, 44, 60, 0, 60, 0, 60, 0, 26, 44, 60, 0, 75, 44, 60, 0, 60, 0, 60, 0, 56, 216, 27, 220, 60, 0, 60, 0, 60, 0, 27, 44, 60, 0, 76, 44, 60, 0, 60, 0, 60, 0, 56, 216, 28, 220, 60, 0, 60, 0, 60, 0, 28, 44, 60, 0, 77, 44, 60, 0, 60, 0, 60, 0, 56, 216, 29, 220, 60, 0, 60, 0, 60, 0, 29, 44, 60, 0, 78, 44, 60, 0, 60, 0, 60, 0, 56, 216, 30, 220, 60, 0, 60, 0, 60, 0, 30, 44, 60, 0, 79, 44, 60, 0, 60, 0, 60, 0, 56, 216, 31, 220, 60, 0, 60, 0, 60, 0, 31, 44, 60, 0, 80, 44, 60, 0, 60, 0, 60, 0, 56, 216, 32, 220, 60, 0, 60, 0, 60, 0, 32, 44, 60, 0, 81, 44, 60, 0, 60, 0, 60, 0, 56, 216, 33, 220, 60, 0, 60, 0, 60, 0, 33, 44, 60, 0, 82, 44, 60, 0, 60, 0, 60, 0, 34, 44, 60, 0, 83, 44, 60, 0, 60, 0, 60, 0, 56, 216, 35, 220, 60, 0, 60, 0, 60, 0, 35, 44, 60, 0, 84, 44, 60, 0, 60, 0, 60, 0, 56, 216, 36, 220, 60, 0, 60, 0, 60, 0, 36, 44, 60, 0, 85, 44, 60, 0, 60, 0, 60, 0, 37, 44, 60, 0, 86, 44, 60, 0, 60, 0, 60, 0, 56, 216, 38, 220, 60, 0, 60, 0, 60, 0, 38, 44, 60, 0, 87, 44, 60, 0, 60, 0, 60, 0, 56, 216, 39, 220, 60, 0, 60, 0, 60, 0, 39, 44, 60, 0, 88, 44, 60, 0, 60, 0, 60, 0, 56, 216, 40, 220, 60, 0, 60, 0, 60, 0, 40, 44, 60, 0, 89, 44, 60, 0, 60, 0, 60, 0, 56, 216, 41, 220, 60, 0, 60, 0, 60, 0, 41, 44, 60, 0, 90, 44, 60, 0, 60, 0, 60, 0, 56, 216, 42, 220, 60, 0, 60, 0, 60, 0, 42, 44, 60, 0, 91, 44, 60, 0, 60, 0, 60, 0, 43, 44, 60, 0, 92, 44, 60, 0, 60, 0, 60, 0, 44, 44, 60, 0, 93, 44, 60, 0, 60, 0, 60, 0, 45, 44, 60, 0, 94, 44, 60, 0, 60, 0, 60, 0, 46, 44, 60, 0, 0, 216, 80, 223, 60, 0, 60, 0, 60, 0, 0, 216, 118, 223, 60, 0, 0, 216, 81, 223, 60, 0, 0, 216, 82, 223, 60, 0, 0, 216, 83, 223, 60, 0, 60, 0, 60, 0, 0, 216, 119, 223, 60, 0, 0, 216, 84, 223, 60, 0, 0, 216, 85, 223, 60, 0, 0, 216, 86, 223, 60, 0, 0, 216, 87, 223, 60, 0, 60, 0, 60, 0, 0, 216, 120, 223, 60, 0, 0, 216, 88, 223, 60, 0, 0, 216, 89, 223, 60, 0, 0, 216, 90, 223, 60, 0, 0, 216, 91, 223, 60, 0, 0, 216, 92, 223, 60, 0, 0, 216, 93, 223, 60, 0, 60, 0, 60, 0, 0, 216, 121, 223, 60, 0, 0, 216, 94, 223, 60, 0, 0, 216, 95, 223, 60, 0, 0, 216, 96, 223, 60, 0, 0, 216, 97, 223, 60, 0, 60, 0, 60, 0, 0, 216, 122, 223, 60, 0, 0, 216, 98, 223, 60, 0, 0, 216, 99, 223, 60, 0, 0, 216, 100, 223, 60, 0, 0, 216, 101, 223, 60, 0, 0, 216, 102, 223, 60, 0, 0, 216, 103, 223, 60, 0, 0, 216, 104, 223, 60, 0, 0, 216, 105, 223, 60, 0, 0, 216, 106, 223, 60, 0, 0, 216, 107, 223, 60, 0, 0, 216, 108, 223, 60, 0, 0, 216, 109, 223, 60, 0, 0, 216, 110, 223, 60, 0, 0, 216, 111, 223, 60, 0, 0, 216, 112, 223, 60, 0, 0, 216, 113, 223, 60, 0, 0, 216, 114, 223, 60, 0, 0, 216, 115, 223, 60, 0, 0, 216, 116, 223, 60, 0, 0, 216, 117, 223, 60, 0, 208, 16, 60, 0, 60, 0, 60, 0, 144, 28, 60, 0, 0, 45, 60, 0, 60, 0, 60, 0, 160, 16, 60, 0, 209, 16, 60, 0, 60, 0, 60, 0, 145, 28, 60, 0, 1, 45, 60, 0, 60, 0, 60, 0, 161, 16, 60, 0, 210, 16, 60, 0, 60, 0, 60, 0, 146, 28, 60, 0, 2, 45, 60, 0, 60, 0, 60, 0, 162, 16, 60, 0, 211, 16, 60, 0, 60, 0, 60, 0, 147, 28, 60, 0, 3, 45, 60, 0, 60, 0, 60, 0, 163, 16, 60, 0, 212, 16, 60, 0, 60, 0, 60, 0, 148, 28, 60, 0, 4, 45, 60, 0, 60, 0, 60, 0, 164, 16, 60, 0, 213, 16, 60, 0, 60, 0, 60, 0, 149, 28, 60, 0, 5, 45, 60, 0, 60, 0, 60, 0, 165, 16, 60, 0, 214, 16, 60, 0, 60, 0, 60, 0, 150, 28, 60, 0, 6, 45, 60, 0, 60, 0, 60, 0, 166, 16, 60, 0, 241, 16, 60, 0, 60, 0, 60, 0, 177, 28, 60, 0, 33, 45, 60, 0, 60, 0, 60, 0, 193, 16, 60, 0, 215, 16, 60, 0, 60, 0, 60, 0, 151, 28, 60, 0, 7, 45, 60, 0, 60, 0, 60, 0, 167, 16, 60, 0, 216, 16, 60, 0, 60, 0, 60, 0, 152, 28, 60, 0, 8, 45, 60, 0, 60, 0, 60, 0, 168, 16, 60, 0, 217, 16, 60, 0, 60, 0, 60, 0, 153, 28, 60, 0, 9, 45, 60, 0, 60, 0, 60, 0, 169, 16, 60, 0, 218, 16, 60, 0, 60, 0, 60, 0, 154, 28, 60, 0, 10, 45, 60, 0, 60, 0, 60, 0, 170, 16, 60, 0, 219, 16, 60, 0, 60, 0, 60, 0, 155, 28, 60, 0, 11, 45, 60, 0, 60, 0, 60, 0, 171, 16, 60, 0, 220, 16, 60, 0, 60, 0, 60, 0, 156, 28, 60, 0, 60, 0, 60, 0, 252, 16, 60, 0, 12, 45, 60, 0, 60, 0, 60, 0, 172, 16, 60, 0, 242, 16, 60, 0, 60, 0, 60, 0, 178, 28, 60, 0, 34, 45, 60, 0, 60, 0, 60, 0, 194, 16, 60, 0, 221, 16, 60, 0, 60, 0, 60, 0, 157, 28, 60, 0, 13, 45, 60, 0, 60, 0, 60, 0, 173, 16, 60, 0, 222, 16, 60, 0, 60, 0, 60, 0, 158, 28, 60, 0, 14, 45, 60, 0, 60, 0, 60, 0, 174, 16, 60, 0, 223, 16, 60, 0, 60, 0, 60, 0, 159, 28, 60, 0, 15, 45, 60, 0, 60, 0, 60, 0, 175, 16, 60, 0, 224, 16, 60, 0, 60, 0, 60, 0, 160, 28, 60, 0, 16, 45, 60, 0, 60, 0, 60, 0, 176, 16, 60, 0, 225, 16, 60, 0, 60, 0, 60, 0, 161, 28, 60, 0, 17, 45, 60, 0, 60, 0, 60, 0, 177, 16, 60, 0, 226, 16, 60, 0, 60, 0, 60, 0, 162, 28, 60, 0, 18, 45, 60, 0, 60, 0, 60, 0, 178, 16, 60, 0, 243, 16, 60, 0, 60, 0, 60, 0, 179, 28, 60, 0, 35, 45, 60, 0, 60, 0, 60, 0, 195, 16, 60, 0, 227, 16, 60, 0, 60, 0, 60, 0, 163, 28, 60, 0, 19, 45, 60, 0, 60, 0, 60, 0, 179, 16, 60, 0, 228, 16, 60, 0, 60, 0, 60, 0, 164, 28, 60, 0, 20, 45, 60, 0, 60, 0, 60, 0, 180, 16, 60, 0, 229, 16, 60, 0, 60, 0, 60, 0, 165, 28, 60, 0, 21, 45, 60, 0, 60, 0, 60, 0, 181, 16, 60, 0, 230, 16, 60, 0, 60, 0, 60, 0, 166, 28, 60, 0, 22, 45, 60, 0, 60, 0, 60, 0, 182, 16, 60, 0, 231, 16, 60, 0, 60, 0, 60, 0, 167, 28, 60, 0, 23, 45, 60, 0, 60, 0, 60, 0, 183, 16, 60, 0, 232, 16, 60, 0, 60, 0, 60, 0, 168, 28, 60, 0, 24, 45, 60, 0, 60, 0, 60, 0, 184, 16, 60, 0, 233, 16, 60, 0, 60, 0, 60, 0, 169, 28, 60, 0, 25, 45, 60, 0, 60, 0, 60, 0, 185, 16, 60, 0, 234, 16, 60, 0, 60, 0, 60, 0, 170, 28, 60, 0, 26, 45, 60, 0, 60, 0, 60, 0, 186, 16, 60, 0, 235, 16, 60, 0, 60, 0, 60, 0, 171, 28, 60, 0, 27, 45, 60, 0, 60, 0, 60, 0, 187, 16, 60, 0, 236, 16, 60, 0, 60, 0, 60, 0, 172, 28, 60, 0, 28, 45, 60, 0, 60, 0, 60, 0, 188, 16, 60, 0, 237, 16, 60, 0, 60, 0, 60, 0, 173, 28, 60, 0, 29, 45, 60, 0, 60, 0, 60, 0, 189, 16, 60, 0, 238, 16, 60, 0, 60, 0, 60, 0, 174, 28, 60, 0, 30, 45, 60, 0, 60, 0, 60, 0, 190, 16, 60, 0, 244, 16, 60, 0, 60, 0, 60, 0, 180, 28, 60, 0, 36, 45, 60, 0, 60, 0, 60, 0, 196, 16, 60, 0, 239, 16, 60, 0, 60, 0, 60, 0, 175, 28, 60, 0, 31, 45, 60, 0, 60, 0, 60, 0, 191, 16, 60, 0, 240, 16, 60, 0, 60, 0, 60, 0, 176, 28, 60, 0, 32, 45, 60, 0, 60, 0, 60, 0, 192, 16, 60, 0, 245, 16, 60, 0, 60, 0, 60, 0, 181, 28, 60, 0, 37, 45, 60, 0, 60, 0, 60, 0, 197, 16, 60, 0, 246, 16, 60, 0, 60, 0, 60, 0, 182, 28, 60, 0, 247, 16, 60, 0, 60, 0, 60, 0, 183, 28, 60, 0, 39, 45, 60, 0, 60, 0, 60, 0, 199, 16, 60, 0, 248, 16, 60, 0, 60, 0, 60, 0, 184, 28, 60, 0, 249, 16, 60, 0, 60, 0, 60, 0, 185, 28, 60, 0, 250, 16, 60, 0, 60, 0, 60, 0, 186, 28, 60, 0, 253, 16, 60, 0, 60, 0, 60, 0, 189, 28, 60, 0, 45, 45, 60, 0, 60, 0, 60, 0, 205, 16, 60, 0, 254, 16, 60, 0, 60, 0, 60, 0, 190, 28, 60, 0, 255, 16, 60, 0, 60, 0, 60, 0, 191, 28, 60, 0, 97, 5, 60, 0, 60, 0, 60, 0, 49, 5, 60, 0, 96, 5, 60, 0, 98, 5, 60, 0, 60, 0, 60, 0, 50, 5, 60, 0, 99, 5, 60, 0, 60, 0, 60, 0, 51, 5, 60, 0, 100, 5, 60, 0, 60, 0, 60, 0, 52, 5, 60, 0, 101, 5, 60, 0, 60, 0, 60, 0, 135, 5, 47, 0, 130, 5, 60, 0, 60, 0, 60, 0, 53, 5, 60, 0, 102, 5, 60, 0, 60, 0, 60, 0, 54, 5, 60, 0, 103, 5, 60, 0, 60, 0, 60, 0, 55, 5, 60, 0, 104, 5, 60, 0, 60, 0, 60, 0, 56, 5, 60, 0, 105, 5, 60, 0, 60, 0, 60, 0, 57, 5, 60, 0, 106, 5, 60, 0, 60, 0, 60, 0, 58, 5, 60, 0, 107, 5, 60, 0, 60, 0, 60, 0, 59, 5, 60, 0, 108, 5, 60, 0, 60, 0, 60, 0, 60, 5, 60, 0, 109, 5, 60, 0, 60, 0, 60, 0, 61, 5, 60, 0, 110, 5, 60, 0, 60, 0, 60, 0, 62, 5, 60, 0, 111, 5, 60, 0, 60, 0, 60, 0, 63, 5, 60, 0, 112, 5, 60, 0, 60, 0, 60, 0, 64, 5, 60, 0, 113, 5, 60, 0, 60, 0, 60, 0, 65, 5, 60, 0, 114, 5, 60, 0, 60, 0, 60, 0, 66, 5, 60, 0, 115, 5, 60, 0, 60, 0, 60, 0, 67, 5, 60, 0, 116, 5, 60, 0, 60, 0, 60, 0, 20, 251, 47, 0, 101, 5, 60, 0, 60, 0, 60, 0, 21, 251, 47, 0, 107, 5, 60, 0, 60, 0, 60, 0, 23, 251, 47, 0, 109, 5, 60, 0, 60, 0, 60, 0, 19, 251, 47, 0, 118, 5, 60, 0, 60, 0, 60, 0, 68, 5, 60, 0, 117, 5, 60, 0, 60, 0, 60, 0, 69, 5, 60, 0, 136, 5, 60, 0, 118, 5, 60, 0, 60, 0, 60, 0, 70, 5, 60, 0, 119, 5, 60, 0, 60, 0, 60, 0, 71, 5, 60, 0, 120, 5, 60, 0, 60, 0, 60, 0, 72, 5, 60, 0, 121, 5, 60, 0, 60, 0, 60, 0, 73, 5, 60, 0, 122, 5, 60, 0, 60, 0, 60, 0, 74, 5, 60, 0, 123, 5, 60, 0, 60, 0, 60, 0, 75, 5, 60, 0, 124, 5, 60, 0, 60, 0, 60, 0, 76, 5, 60, 0, 125, 5, 60, 0, 60, 0, 60, 0, 77, 5, 60, 0, 126, 5, 60, 0, 60, 0, 60, 0, 22, 251, 47, 0, 118, 5, 60, 0, 60, 0, 60, 0, 78, 5, 60, 0, 127, 5, 60, 0, 60, 0, 60, 0, 79, 5, 60, 0, 128, 5, 60, 0, 60, 0, 60, 0, 80, 5, 60, 0, 129, 5, 60, 0, 60, 0, 60, 0, 81, 5, 60, 0, 130, 5, 60, 0, 60, 0, 60, 0, 82, 5, 60, 0, 131, 5, 60, 0, 60, 0, 60, 0, 83, 5, 60, 0, 132, 5, 60, 0, 60, 0, 60, 0, 84, 5, 60, 0, 133, 5, 60, 0, 60, 0, 60, 0, 85, 5, 60, 0, 134, 5, 60, 0, 60, 0, 60, 0, 86, 5, 60, 0, 89, 5, 60, 0, 14, 32, 208, 5, 14, 32, 60, 0, 60, 0, 60, 0, 53, 33, 60, 0, 60, 0, 60, 0, 14, 32, 79, 251, 14, 32, 47, 0, 14, 32, 220, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 33, 251, 14, 32, 60, 0, 14, 32, 209, 5, 14, 32, 60, 0, 60, 0, 60, 0, 54, 33, 60, 0, 14, 32, 210, 5, 14, 32, 60, 0, 60, 0, 60, 0, 55, 33, 60, 0, 14, 32, 211, 5, 14, 32, 60, 0, 60, 0, 60, 0, 56, 33, 60, 0, 60, 0, 60, 0, 14, 32, 34, 251, 14, 32, 60, 0, 14, 32, 212, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 35, 251, 14, 32, 60, 0, 14, 32, 213, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 240, 5, 14, 32, 47, 0, 14, 32, 213, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 241, 5, 14, 32, 47, 0, 14, 32, 217, 5, 14, 32, 60, 0, 14, 32, 214, 5, 14, 32, 60, 0, 14, 32, 215, 5, 14, 32, 60, 0, 14, 32, 216, 5, 14, 32, 60, 0, 14, 32, 217, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 239, 5, 14, 32, 47, 0, 14, 32, 212, 5, 213, 5, 212, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 242, 5, 14, 32, 47, 0, 14, 32, 217, 5, 14, 32, 60, 0, 14, 32, 219, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 36, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 218, 5, 14, 32, 60, 0, 14, 32, 220, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 37, 251, 14, 32, 60, 0, 14, 32, 222, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 38, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 221, 5, 14, 32, 60, 0, 14, 32, 224, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 223, 5, 14, 32, 60, 0, 14, 32, 225, 5, 14, 32, 60, 0, 14, 32, 226, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 32, 251, 14, 32, 60, 0, 14, 32, 228, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 227, 5, 14, 32, 60, 0, 14, 32, 230, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 229, 5, 14, 32, 60, 0, 14, 32, 231, 5, 14, 32, 60, 0, 14, 32, 232, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 39, 251, 14, 32, 60, 0, 14, 32, 233, 5, 14, 32, 60, 0, 14, 32, 234, 5, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 40, 251, 14, 32, 60, 0, 14, 32, 2, 216, 0, 221, 14, 32, 60, 0, 14, 32, 2, 216, 1, 221, 14, 32, 60, 0, 14, 32, 2, 216, 2, 221, 14, 32, 60, 0, 14, 32, 2, 216, 3, 221, 14, 32, 60, 0, 14, 32, 2, 216, 4, 221, 14, 32, 60, 0, 14, 32, 2, 216, 5, 221, 14, 32, 60, 0, 14, 32, 2, 216, 6, 221, 14, 32, 60, 0, 14, 32, 2, 216, 7, 221, 14, 32, 60, 0, 14, 32, 2, 216, 8, 221, 14, 32, 60, 0, 14, 32, 2, 216, 9, 221, 14, 32, 60, 0, 14, 32, 2, 216, 10, 221, 14, 32, 60, 0, 14, 32, 2, 216, 11, 221, 14, 32, 60, 0, 14, 32, 2, 216, 12, 221, 14, 32, 60, 0, 14, 32, 2, 216, 13, 221, 14, 32, 60, 0, 14, 32, 2, 216, 14, 221, 14, 32, 60, 0, 14, 32, 2, 216, 15, 221, 14, 32, 60, 0, 14, 32, 2, 216, 16, 221, 14, 32, 60, 0, 14, 32, 2, 216, 17, 221, 14, 32, 60, 0, 14, 32, 2, 216, 18, 221, 14, 32, 60, 0, 14, 32, 2, 216, 19, 221, 14, 32, 60, 0, 14, 32, 2, 216, 20, 221, 14, 32, 60, 0, 14, 32, 2, 216, 21, 221, 14, 32, 60, 0, 14, 32, 0, 8, 14, 32, 60, 0, 14, 32, 1, 8, 14, 32, 60, 0, 14, 32, 2, 8, 14, 32, 60, 0, 14, 32, 3, 8, 14, 32, 60, 0, 14, 32, 4, 8, 14, 32, 60, 0, 14, 32, 5, 8, 14, 32, 60, 0, 14, 32, 6, 8, 14, 32, 60, 0, 14, 32, 7, 8, 14, 32, 60, 0, 14, 32, 8, 8, 14, 32, 60, 0, 14, 32, 9, 8, 14, 32, 60, 0, 14, 32, 10, 8, 14, 32, 60, 0, 14, 32, 11, 8, 14, 32, 60, 0, 14, 32, 12, 8, 14, 32, 60, 0, 14, 32, 13, 8, 14, 32, 60, 0, 14, 32, 14, 8, 14, 32, 60, 0, 14, 32, 15, 8, 14, 32, 60, 0, 14, 32, 16, 8, 14, 32, 60, 0, 14, 32, 17, 8, 14, 32, 60, 0, 14, 32, 18, 8, 14, 32, 60, 0, 14, 32, 19, 8, 14, 32, 60, 0, 14, 32, 20, 8, 14, 32, 60, 0, 14, 32, 21, 8, 14, 32, 60, 0, 22, 8, 60, 0, 23, 8, 60, 0, 14, 32, 26, 8, 14, 32, 60, 0, 27, 8, 60, 0, 14, 32, 33, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 116, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 128, 254, 14, 32, 60, 0, 60, 0, 14, 32, 253, 6, 14, 32, 60, 0, 14, 32, 34, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 130, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 129, 254, 14, 32, 60, 0, 14, 32, 35, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 132, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 131, 254, 14, 32, 60, 0, 14, 32, 114, 6, 14, 32, 60, 0, 14, 32, 113, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 81, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 80, 251, 14, 32, 60, 0, 14, 32, 36, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 134, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 133, 254, 14, 32, 60, 0, 14, 32, 37, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 136, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 135, 254, 14, 32, 60, 0, 14, 32, 115, 6, 14, 32, 60, 0, 14, 32, 115, 7, 14, 32, 60, 0, 14, 32, 116, 7, 14, 32, 60, 0, 14, 32, 38, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 139, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 151, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 152, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 153, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 154, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 155, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 251, 251, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 248, 251, 14, 32, 47, 0, 14, 32, 208, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 140, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 223, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 224, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 138, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 235, 251, 14, 32, 47, 0, 14, 32, 39, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 100, 252, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 101, 252, 14, 32, 47, 0, 14, 32, 50, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 102, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 103, 252, 14, 32, 47, 0, 14, 32, 70, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 237, 251, 14, 32, 47, 0, 14, 32, 213, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 239, 251, 14, 32, 47, 0, 14, 32, 72, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 243, 251, 14, 32, 47, 0, 14, 32, 198, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 241, 251, 14, 32, 47, 0, 14, 32, 199, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 245, 251, 14, 32, 47, 0, 14, 32, 200, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 250, 251, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 61, 0, 14, 32, 104, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 105, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 247, 251, 14, 32, 47, 0, 14, 32, 208, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 137, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 234, 251, 14, 32, 47, 0, 14, 32, 39, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 0, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 1, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 236, 251, 14, 32, 47, 0, 14, 32, 213, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 238, 251, 14, 32, 47, 0, 14, 32, 72, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 242, 251, 14, 32, 47, 0, 14, 32, 198, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 240, 251, 14, 32, 47, 0, 14, 32, 199, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 244, 251, 14, 32, 47, 0, 14, 32, 200, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 249, 251, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 61, 0, 14, 32, 3, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 4, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 246, 251, 14, 32, 47, 0, 14, 32, 208, 6, 14, 32, 60, 0, 14, 32, 168, 8, 14, 32, 60, 0, 14, 32, 169, 8, 14, 32, 60, 0, 14, 32, 172, 8, 14, 32, 60, 0, 14, 32, 39, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 173, 8, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 117, 6, 14, 32, 47, 0, 14, 32, 33, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 0, 222, 14, 32, 61, 0, 14, 32, 59, 216, 128, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 142, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 60, 253, 14, 32, 47, 0, 75, 6, 60, 0, 60, 0, 60, 0, 14, 32, 141, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 61, 253, 14, 32, 47, 0, 75, 6, 60, 0, 60, 0, 60, 0, 14, 32, 243, 253, 14, 32, 47, 0, 14, 32, 67, 6, 40, 6, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 242, 253, 14, 32, 47, 0, 14, 32, 68, 6, 68, 6, 71, 6, 14, 32, 60, 0, 14, 32, 110, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 28, 222, 14, 32, 61, 0, 14, 32, 59, 216, 124, 222, 14, 32, 60, 0, 14, 32, 40, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 1, 222, 14, 32, 61, 0, 14, 32, 59, 216, 33, 222, 14, 32, 61, 0, 14, 32, 59, 216, 97, 222, 14, 32, 61, 0, 14, 32, 59, 216, 129, 222, 14, 32, 61, 0, 14, 32, 59, 216, 161, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 145, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 156, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 157, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 158, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 159, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 160, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 146, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 225, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 226, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 144, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 194, 253, 14, 32, 47, 0, 14, 32, 45, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 158, 253, 14, 32, 47, 0, 14, 32, 46, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 106, 252, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 107, 252, 14, 32, 47, 0, 14, 32, 50, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 108, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 109, 252, 14, 32, 47, 0, 14, 32, 70, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 110, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 111, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 143, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 5, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 6, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 7, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 8, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 9, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 10, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 123, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 84, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 85, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 83, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 82, 251, 14, 32, 60, 0, 14, 32, 126, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 88, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 89, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 87, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 86, 251, 14, 32, 60, 0, 14, 32, 128, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 92, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 93, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 91, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 90, 251, 14, 32, 60, 0, 14, 32, 80, 7, 14, 32, 60, 0, 14, 32, 81, 7, 14, 32, 60, 0, 14, 32, 82, 7, 14, 32, 60, 0, 14, 32, 83, 7, 14, 32, 60, 0, 14, 32, 84, 7, 14, 32, 60, 0, 14, 32, 85, 7, 14, 32, 60, 0, 14, 32, 160, 8, 14, 32, 60, 0, 14, 32, 86, 7, 14, 32, 60, 0, 14, 32, 190, 8, 14, 32, 60, 0, 14, 32, 161, 8, 14, 32, 60, 0, 14, 32, 182, 8, 14, 32, 60, 0, 14, 32, 183, 8, 14, 32, 60, 0, 14, 32, 41, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 148, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 147, 254, 14, 32, 60, 0, 14, 32, 42, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 21, 222, 14, 32, 61, 0, 14, 32, 59, 216, 53, 222, 14, 32, 61, 0, 14, 32, 59, 216, 117, 222, 14, 32, 61, 0, 14, 32, 59, 216, 149, 222, 14, 32, 61, 0, 14, 32, 59, 216, 181, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 151, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 161, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 80, 253, 14, 32, 47, 0, 14, 32, 44, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 162, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 82, 253, 14, 32, 47, 0, 14, 32, 45, 6, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 83, 253, 14, 32, 47, 0, 14, 32, 45, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 163, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 84, 253, 14, 32, 47, 0, 14, 32, 46, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 164, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 85, 253, 14, 32, 47, 0, 14, 32, 69, 6, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 86, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 87, 253, 14, 32, 47, 0, 14, 32, 69, 6, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 165, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 152, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 227, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 228, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 150, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 160, 253, 14, 32, 47, 0, 14, 32, 44, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 159, 253, 14, 32, 47, 0, 14, 32, 44, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 81, 253, 14, 32, 47, 0, 14, 32, 45, 6, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 162, 253, 14, 32, 47, 0, 14, 32, 46, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 161, 253, 14, 32, 47, 0, 14, 32, 46, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 112, 252, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 113, 252, 14, 32, 47, 0, 14, 32, 50, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 114, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 164, 253, 14, 32, 47, 0, 14, 32, 69, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 163, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 115, 252, 14, 32, 47, 0, 14, 32, 70, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 116, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 117, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 149, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 11, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 12, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 13, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 14, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 15, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 16, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 43, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 22, 222, 14, 32, 61, 0, 14, 32, 59, 216, 54, 222, 14, 32, 61, 0, 14, 32, 59, 216, 118, 222, 14, 32, 61, 0, 14, 32, 59, 216, 150, 222, 14, 32, 61, 0, 14, 32, 59, 216, 182, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 155, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 166, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 156, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 229, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 230, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 154, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 118, 252, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 119, 252, 14, 32, 47, 0, 14, 32, 50, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 120, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 121, 252, 14, 32, 47, 0, 14, 32, 70, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 122, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 123, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 153, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 17, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 18, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 19, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 20, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 121, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 104, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 105, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 103, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 102, 251, 14, 32, 60, 0, 14, 32, 122, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 96, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 97, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 95, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 94, 251, 14, 32, 60, 0, 14, 32, 124, 6, 14, 32, 60, 0, 14, 32, 125, 6, 14, 32, 60, 0, 14, 32, 127, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 100, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 101, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 99, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 98, 251, 14, 32, 60, 0, 14, 32, 184, 8, 14, 32, 60, 0, 14, 32, 191, 8, 14, 32, 60, 0, 14, 32, 192, 8, 14, 32, 60, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 2, 222, 14, 32, 61, 0, 14, 32, 59, 216, 34, 222, 14, 32, 61, 0, 14, 32, 59, 216, 66, 222, 14, 32, 61, 0, 14, 32, 59, 216, 98, 222, 14, 32, 61, 0, 14, 32, 59, 216, 130, 222, 14, 32, 61, 0, 14, 32, 59, 216, 162, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 159, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 167, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 168, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 89, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 160, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 158, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 166, 253, 14, 32, 47, 0, 14, 32, 45, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 190, 253, 14, 32, 47, 0, 14, 32, 45, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 88, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 167, 253, 14, 32, 47, 0, 14, 32, 69, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 165, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 29, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 30, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 157, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 21, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 251, 253, 14, 32, 47, 0, 14, 32, 68, 6, 39, 0, 32, 0, 39, 0, 44, 6, 68, 6, 39, 6, 68, 6, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 22, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 1, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 131, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 120, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 121, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 119, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 118, 251, 14, 32, 60, 0, 14, 32, 132, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 116, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 117, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 115, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 114, 251, 14, 32, 60, 0, 14, 32, 134, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 124, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 125, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 123, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 122, 251, 14, 32, 60, 0, 14, 32, 191, 6, 14, 32, 60, 0, 14, 32, 135, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 128, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 129, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 127, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 126, 251, 14, 32, 60, 0, 14, 32, 162, 8, 14, 32, 60, 0, 14, 32, 193, 8, 14, 32, 60, 0, 14, 32, 197, 8, 14, 32, 60, 0, 14, 32, 198, 8, 14, 32, 60, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 7, 222, 14, 32, 61, 0, 14, 32, 59, 216, 39, 222, 14, 32, 61, 0, 14, 32, 59, 216, 71, 222, 14, 32, 61, 0, 14, 32, 59, 216, 103, 222, 14, 32, 61, 0, 14, 32, 59, 216, 135, 222, 14, 32, 61, 0, 14, 32, 59, 216, 167, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 163, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 169, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 170, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 164, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 162, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 191, 253, 14, 32, 47, 0, 14, 32, 44, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 91, 253, 14, 32, 47, 0, 14, 32, 69, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 90, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 27, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 28, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 161, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 23, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 24, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 255, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 0, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 23, 222, 14, 32, 61, 0, 14, 32, 59, 216, 55, 222, 14, 32, 61, 0, 14, 32, 59, 216, 87, 222, 14, 32, 61, 0, 14, 32, 59, 216, 119, 222, 14, 32, 61, 0, 14, 32, 59, 216, 151, 222, 14, 32, 61, 0, 14, 32, 59, 216, 183, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 167, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 171, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 172, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 168, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 166, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 31, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 32, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 165, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 25, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 26, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 27, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 4, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 129, 6, 14, 32, 60, 0, 14, 32, 130, 6, 14, 32, 60, 0, 14, 32, 133, 6, 14, 32, 60, 0, 14, 32, 87, 7, 14, 32, 60, 0, 14, 32, 88, 7, 14, 32, 60, 0, 14, 32, 110, 7, 14, 32, 60, 0, 14, 32, 111, 7, 14, 32, 60, 0, 14, 32, 114, 7, 14, 32, 60, 0, 14, 32, 124, 7, 14, 32, 60, 0, 14, 32, 47, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 3, 222, 14, 32, 61, 0, 14, 32, 59, 216, 131, 222, 14, 32, 61, 0, 14, 32, 59, 216, 163, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 170, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 169, 254, 14, 32, 60, 0, 14, 32, 48, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 24, 222, 14, 32, 61, 0, 14, 32, 59, 216, 152, 222, 14, 32, 61, 0, 14, 32, 59, 216, 184, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 172, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 171, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 91, 252, 14, 32, 47, 0, 112, 6, 60, 0, 14, 32, 136, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 137, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 136, 251, 14, 32, 60, 0, 14, 32, 137, 6, 14, 32, 60, 0, 14, 32, 138, 6, 14, 32, 60, 0, 14, 32, 139, 6, 14, 32, 60, 0, 14, 32, 140, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 133, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 132, 251, 14, 32, 60, 0, 14, 32, 141, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 131, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 130, 251, 14, 32, 60, 0, 14, 32, 174, 8, 14, 32, 60, 0, 14, 32, 142, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 135, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 134, 251, 14, 32, 60, 0, 14, 32, 143, 6, 14, 32, 60, 0, 14, 32, 144, 6, 14, 32, 60, 0, 14, 32, 238, 6, 14, 32, 60, 0, 14, 32, 89, 7, 14, 32, 60, 0, 14, 32, 90, 7, 14, 32, 60, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 19, 222, 14, 32, 61, 0, 14, 32, 59, 216, 147, 222, 14, 32, 61, 0, 14, 32, 59, 216, 179, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 174, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 173, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 92, 252, 14, 32, 47, 0, 112, 6, 60, 0, 60, 0, 60, 0, 14, 32, 246, 253, 14, 32, 47, 0, 14, 32, 51, 6, 72, 6, 68, 6, 14, 32, 60, 0, 14, 32, 50, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 6, 222, 14, 32, 61, 0, 14, 32, 59, 216, 134, 222, 14, 32, 61, 0, 14, 32, 59, 216, 166, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 176, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 175, 254, 14, 32, 60, 0, 14, 32, 145, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 141, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 140, 251, 14, 32, 60, 0, 14, 32, 146, 6, 14, 32, 60, 0, 14, 32, 147, 6, 14, 32, 60, 0, 14, 32, 148, 6, 14, 32, 60, 0, 14, 32, 149, 6, 14, 32, 60, 0, 14, 32, 150, 6, 14, 32, 60, 0, 14, 32, 151, 6, 14, 32, 60, 0, 14, 32, 152, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 139, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 138, 251, 14, 32, 60, 0, 14, 32, 153, 6, 14, 32, 60, 0, 14, 32, 239, 6, 14, 32, 60, 0, 14, 32, 91, 7, 14, 32, 60, 0, 14, 32, 107, 7, 14, 32, 60, 0, 14, 32, 108, 7, 14, 32, 60, 0, 14, 32, 113, 7, 14, 32, 60, 0, 14, 32, 170, 8, 14, 32, 60, 0, 14, 32, 178, 8, 14, 32, 60, 0, 14, 32, 185, 8, 14, 32, 60, 0, 14, 32, 51, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 14, 222, 14, 32, 61, 0, 14, 32, 59, 216, 46, 222, 14, 32, 61, 0, 14, 32, 59, 216, 78, 222, 14, 32, 61, 0, 14, 32, 59, 216, 110, 222, 14, 32, 61, 0, 14, 32, 59, 216, 142, 222, 14, 32, 61, 0, 14, 32, 59, 216, 174, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 179, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 173, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 93, 253, 14, 32, 47, 0, 14, 32, 44, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 174, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 92, 253, 14, 32, 47, 0, 14, 32, 45, 6, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 175, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 176, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 97, 253, 14, 32, 47, 0, 14, 32, 69, 6, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 96, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 99, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 49, 253, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 180, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 52, 253, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 53, 253, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 54, 253, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 231, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 232, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 178, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 94, 253, 14, 32, 47, 0, 14, 32, 44, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 168, 253, 14, 32, 47, 0, 14, 32, 46, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 198, 253, 14, 32, 47, 0, 14, 32, 46, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 42, 253, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 95, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 98, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 23, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 24, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 177, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 28, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 29, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 30, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 14, 253, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 31, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 251, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 252, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 52, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 20, 222, 14, 32, 61, 0, 14, 32, 59, 216, 52, 222, 14, 32, 61, 0, 14, 32, 59, 216, 84, 222, 14, 32, 61, 0, 14, 32, 59, 216, 116, 222, 14, 32, 61, 0, 14, 32, 59, 216, 148, 222, 14, 32, 61, 0, 14, 32, 59, 216, 180, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 183, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 45, 253, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 46, 253, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 104, 253, 14, 32, 47, 0, 14, 32, 45, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 47, 253, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 48, 253, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 107, 253, 14, 32, 47, 0, 14, 32, 69, 6, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 109, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 50, 253, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 184, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 55, 253, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 56, 253, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 57, 253, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 233, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 234, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 182, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 37, 253, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 105, 253, 14, 32, 47, 0, 14, 32, 44, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 38, 253, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 103, 253, 14, 32, 47, 0, 14, 32, 45, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 170, 253, 14, 32, 47, 0, 14, 32, 45, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 39, 253, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 41, 253, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 40, 253, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 106, 253, 14, 32, 47, 0, 14, 32, 69, 6, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 108, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 25, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 26, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 181, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 9, 253, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 10, 253, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 11, 253, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 13, 253, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 12, 253, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 253, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 254, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 154, 6, 14, 32, 60, 0, 14, 32, 155, 6, 14, 32, 60, 0, 14, 32, 156, 6, 14, 32, 60, 0, 14, 32, 250, 6, 14, 32, 60, 0, 14, 32, 92, 7, 14, 32, 60, 0, 14, 32, 109, 7, 14, 32, 60, 0, 14, 32, 112, 7, 14, 32, 60, 0, 14, 32, 125, 7, 14, 32, 60, 0, 14, 32, 126, 7, 14, 32, 60, 0, 14, 32, 53, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 17, 222, 14, 32, 61, 0, 14, 32, 59, 216, 49, 222, 14, 32, 61, 0, 14, 32, 59, 216, 81, 222, 14, 32, 61, 0, 14, 32, 59, 216, 113, 222, 14, 32, 61, 0, 14, 32, 59, 216, 145, 222, 14, 32, 61, 0, 14, 32, 59, 216, 177, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 187, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 177, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 101, 253, 14, 32, 47, 0, 14, 32, 45, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 178, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 179, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 197, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 188, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 186, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 100, 253, 14, 32, 47, 0, 14, 32, 45, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 169, 253, 14, 32, 47, 0, 14, 32, 45, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 43, 253, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 102, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 33, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 34, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 185, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 32, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 15, 253, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 245, 253, 14, 32, 47, 0, 14, 32, 68, 6, 57, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 249, 253, 14, 32, 47, 0, 14, 32, 68, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 250, 253, 14, 32, 47, 0, 14, 32, 68, 6, 73, 6, 39, 0, 32, 0, 39, 0, 39, 6, 68, 6, 68, 6, 71, 6, 39, 0, 32, 0, 39, 0, 57, 6, 68, 6, 74, 6, 71, 6, 39, 0, 32, 0, 39, 0, 72, 6, 51, 6, 68, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 240, 253, 14, 32, 47, 0, 14, 32, 68, 6, 210, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 33, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 5, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 6, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 54, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 25, 222, 14, 32, 61, 0, 14, 32, 59, 216, 57, 222, 14, 32, 61, 0, 14, 32, 59, 216, 89, 222, 14, 32, 61, 0, 14, 32, 59, 216, 121, 222, 14, 32, 61, 0, 14, 32, 59, 216, 153, 222, 14, 32, 61, 0, 14, 32, 59, 216, 185, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 191, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 180, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 181, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 182, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 112, 253, 14, 32, 47, 0, 14, 32, 46, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 183, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 192, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 190, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 110, 253, 14, 32, 47, 0, 14, 32, 45, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 171, 253, 14, 32, 47, 0, 14, 32, 45, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 111, 253, 14, 32, 47, 0, 14, 32, 46, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 44, 253, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 35, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 36, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 189, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 34, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 35, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 36, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 16, 253, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 37, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 7, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 8, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 157, 6, 14, 32, 60, 0, 14, 32, 175, 8, 14, 32, 60, 0, 14, 32, 158, 6, 14, 32, 60, 0, 14, 32, 251, 6, 14, 32, 60, 0, 14, 32, 55, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 8, 222, 14, 32, 61, 0, 14, 32, 59, 216, 104, 222, 14, 32, 61, 0, 14, 32, 59, 216, 136, 222, 14, 32, 61, 0, 14, 32, 59, 216, 168, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 195, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 184, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 51, 253, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 114, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 115, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 196, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 253, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 194, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 113, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 116, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 17, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 18, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 193, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 38, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 39, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 245, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 246, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 56, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 26, 222, 14, 32, 61, 0, 14, 32, 59, 216, 122, 222, 14, 32, 61, 0, 14, 32, 59, 216, 154, 222, 14, 32, 61, 0, 14, 32, 59, 216, 186, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 199, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 185, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 200, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 253, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 198, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 197, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 40, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 14, 32, 159, 6, 14, 32, 60, 0, 14, 32, 163, 8, 14, 32, 60, 0, 14, 32, 57, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 15, 222, 14, 32, 61, 0, 14, 32, 59, 216, 47, 222, 14, 32, 61, 0, 14, 32, 59, 216, 79, 222, 14, 32, 61, 0, 14, 32, 59, 216, 111, 222, 14, 32, 61, 0, 14, 32, 59, 216, 143, 222, 14, 32, 61, 0, 14, 32, 59, 216, 175, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 203, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 186, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 196, 253, 14, 32, 47, 0, 14, 32, 44, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 187, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 119, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 204, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 202, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 117, 253, 14, 32, 47, 0, 14, 32, 44, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 118, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 120, 253, 14, 32, 47, 0, 14, 32, 69, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 182, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 19, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 20, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 201, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 41, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 247, 253, 14, 32, 47, 0, 14, 32, 68, 6, 74, 6, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 42, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 247, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 248, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 58, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 27, 222, 14, 32, 61, 0, 14, 32, 59, 216, 59, 222, 14, 32, 61, 0, 14, 32, 59, 216, 91, 222, 14, 32, 61, 0, 14, 32, 59, 216, 123, 222, 14, 32, 61, 0, 14, 32, 59, 216, 155, 222, 14, 32, 61, 0, 14, 32, 59, 216, 187, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 207, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 188, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 189, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 208, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 206, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 121, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 123, 253, 14, 32, 47, 0, 14, 32, 69, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 122, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 21, 253, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 22, 253, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 205, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 43, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 44, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 249, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 250, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 160, 6, 14, 32, 60, 0, 14, 32, 195, 8, 14, 32, 60, 0, 14, 32, 252, 6, 14, 32, 60, 0, 14, 32, 93, 7, 14, 32, 60, 0, 14, 32, 94, 7, 14, 32, 60, 0, 14, 32, 95, 7, 14, 32, 60, 0, 14, 32, 179, 8, 14, 32, 60, 0, 14, 32, 65, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 16, 222, 14, 32, 61, 0, 14, 32, 59, 216, 48, 222, 14, 32, 61, 0, 14, 32, 59, 216, 112, 222, 14, 32, 61, 0, 14, 32, 59, 216, 144, 222, 14, 32, 61, 0, 14, 32, 59, 216, 176, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 211, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 190, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 191, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 192, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 125, 253, 14, 32, 47, 0, 14, 32, 46, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 193, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 212, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 210, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 124, 253, 14, 32, 47, 0, 14, 32, 46, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 193, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 124, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 125, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 209, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 45, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 46, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 47, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 48, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 49, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 50, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 161, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 30, 222, 14, 32, 61, 0, 14, 32, 59, 216, 126, 222, 14, 32, 60, 0, 14, 32, 162, 6, 14, 32, 60, 0, 14, 32, 187, 8, 14, 32, 60, 0, 14, 32, 163, 6, 14, 32, 60, 0, 14, 32, 164, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 108, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 109, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 107, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 106, 251, 14, 32, 60, 0, 14, 32, 164, 8, 14, 32, 60, 0, 14, 32, 165, 6, 14, 32, 60, 0, 14, 32, 166, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 112, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 113, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 111, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 110, 251, 14, 32, 60, 0, 14, 32, 96, 7, 14, 32, 60, 0, 14, 32, 97, 7, 14, 32, 60, 0, 14, 32, 111, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 31, 222, 14, 32, 61, 0, 14, 32, 59, 216, 95, 222, 14, 32, 60, 0, 14, 32, 66, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 18, 222, 14, 32, 61, 0, 14, 32, 59, 216, 50, 222, 14, 32, 61, 0, 14, 32, 59, 216, 82, 222, 14, 32, 61, 0, 14, 32, 59, 216, 114, 222, 14, 32, 61, 0, 14, 32, 59, 216, 146, 222, 14, 32, 61, 0, 14, 32, 59, 216, 178, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 215, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 194, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 195, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 180, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 216, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 214, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 126, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 127, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 178, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 126, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 127, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 213, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 51, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 241, 253, 14, 32, 47, 0, 14, 32, 68, 6, 210, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 52, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 53, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 54, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 167, 6, 14, 32, 60, 0, 14, 32, 188, 8, 14, 32, 60, 0, 14, 32, 168, 6, 14, 32, 60, 0, 14, 32, 196, 8, 14, 32, 60, 0, 14, 32, 165, 8, 14, 32, 60, 0, 14, 32, 67, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 10, 222, 14, 32, 61, 0, 14, 32, 59, 216, 42, 222, 14, 32, 61, 0, 14, 32, 59, 216, 106, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 219, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 196, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 197, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 198, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 199, 252, 14, 32, 47, 0, 14, 32, 68, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 200, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 195, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 220, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 235, 252, 14, 32, 47, 0, 14, 32, 68, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 236, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 218, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 128, 252, 14, 32, 47, 0, 14, 32, 39, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 129, 252, 14, 32, 47, 0, 14, 32, 68, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 130, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 187, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 183, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 131, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 132, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 217, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 55, 252, 14, 32, 47, 0, 14, 32, 39, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 56, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 57, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 252, 14, 32, 47, 0, 14, 32, 68, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 60, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 61, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 62, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 169, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 144, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 145, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 143, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 142, 251, 14, 32, 60, 0, 14, 32, 170, 6, 14, 32, 60, 0, 14, 32, 171, 6, 14, 32, 60, 0, 14, 32, 172, 6, 14, 32, 60, 0, 14, 32, 127, 7, 14, 32, 60, 0, 14, 32, 173, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 213, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 214, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 212, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 211, 251, 14, 32, 60, 0, 14, 32, 174, 6, 14, 32, 60, 0, 14, 32, 180, 8, 14, 32, 60, 0, 14, 32, 175, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 148, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 149, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 147, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 146, 251, 14, 32, 60, 0, 14, 32, 176, 8, 14, 32, 60, 0, 14, 32, 176, 6, 14, 32, 60, 0, 14, 32, 177, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 156, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 157, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 155, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 154, 251, 14, 32, 60, 0, 14, 32, 178, 6, 14, 32, 60, 0, 14, 32, 179, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 152, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 153, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 151, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 150, 251, 14, 32, 60, 0, 14, 32, 180, 6, 14, 32, 60, 0, 14, 32, 98, 7, 14, 32, 60, 0, 14, 32, 59, 6, 14, 32, 60, 0, 14, 32, 60, 6, 14, 32, 60, 0, 14, 32, 99, 7, 14, 32, 60, 0, 14, 32, 100, 7, 14, 32, 60, 0, 14, 32, 194, 8, 14, 32, 60, 0, 14, 32, 68, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 11, 222, 14, 32, 61, 0, 14, 32, 59, 216, 43, 222, 14, 32, 61, 0, 14, 32, 59, 216, 75, 222, 14, 32, 61, 0, 14, 32, 59, 216, 139, 222, 14, 32, 61, 0, 14, 32, 59, 216, 171, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 223, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 201, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 131, 253, 14, 32, 47, 0, 14, 32, 44, 6, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 186, 253, 14, 32, 47, 0, 14, 32, 44, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 202, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 181, 253, 14, 32, 47, 0, 14, 32, 45, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 203, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 134, 253, 14, 32, 47, 0, 14, 32, 46, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 204, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 136, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 205, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 224, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 237, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 222, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 246, 254, 14, 32, 47, 0, 14, 32, 34, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 248, 254, 14, 32, 47, 0, 14, 32, 35, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 250, 254, 14, 32, 47, 0, 14, 32, 37, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 252, 254, 14, 32, 47, 0, 14, 32, 39, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 132, 253, 14, 32, 47, 0, 14, 32, 44, 6, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 188, 253, 14, 32, 47, 0, 14, 32, 44, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 172, 253, 14, 32, 47, 0, 14, 32, 44, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 128, 253, 14, 32, 47, 0, 14, 32, 45, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 130, 253, 14, 32, 47, 0, 14, 32, 45, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 129, 253, 14, 32, 47, 0, 14, 32, 45, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 133, 253, 14, 32, 47, 0, 14, 32, 46, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 133, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 135, 253, 14, 32, 47, 0, 14, 32, 69, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 173, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 134, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 135, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 221, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 245, 254, 14, 32, 47, 0, 14, 32, 34, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 247, 254, 14, 32, 47, 0, 14, 32, 35, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 249, 254, 14, 32, 47, 0, 14, 32, 37, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 251, 254, 14, 32, 47, 0, 14, 32, 39, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 63, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 64, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 65, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 66, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 67, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 68, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 181, 6, 14, 32, 60, 0, 14, 32, 182, 6, 14, 32, 60, 0, 14, 32, 183, 6, 14, 32, 60, 0, 14, 32, 184, 6, 14, 32, 60, 0, 14, 32, 106, 7, 14, 32, 60, 0, 14, 32, 166, 8, 14, 32, 60, 0, 14, 32, 199, 8, 14, 32, 60, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 12, 222, 14, 32, 61, 0, 14, 32, 59, 216, 44, 222, 14, 32, 61, 0, 14, 32, 59, 216, 108, 222, 14, 32, 61, 0, 14, 32, 59, 216, 140, 222, 14, 32, 61, 0, 14, 32, 59, 216, 172, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 227, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 206, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 140, 253, 14, 32, 47, 0, 14, 32, 44, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 146, 253, 14, 32, 47, 0, 14, 32, 44, 6, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 141, 253, 14, 32, 47, 0, 14, 32, 44, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 207, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 137, 253, 14, 32, 47, 0, 14, 32, 45, 6, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 138, 253, 14, 32, 47, 0, 14, 32, 45, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 208, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 142, 253, 14, 32, 47, 0, 14, 32, 46, 6, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 143, 253, 14, 32, 47, 0, 14, 32, 46, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 209, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 228, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 226, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 136, 252, 14, 32, 47, 0, 14, 32, 39, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 192, 253, 14, 32, 47, 0, 14, 32, 44, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 139, 253, 14, 32, 47, 0, 14, 32, 45, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 185, 253, 14, 32, 47, 0, 14, 32, 46, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 137, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 177, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 225, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 69, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 70, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 244, 253, 14, 32, 47, 0, 14, 32, 45, 6, 69, 6, 47, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 71, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 72, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 73, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 74, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 14, 32, 254, 6, 14, 32, 60, 0, 14, 32, 101, 7, 14, 32, 60, 0, 14, 32, 102, 7, 14, 32, 60, 0, 14, 32, 167, 8, 14, 32, 60, 0, 14, 32, 70, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 13, 222, 14, 32, 61, 0, 14, 32, 59, 216, 45, 222, 14, 32, 61, 0, 14, 32, 59, 216, 77, 222, 14, 32, 61, 0, 14, 32, 59, 216, 109, 222, 14, 32, 61, 0, 14, 32, 59, 216, 141, 222, 14, 32, 61, 0, 14, 32, 59, 216, 173, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 231, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 210, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 184, 253, 14, 32, 47, 0, 14, 32, 44, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 152, 253, 14, 32, 47, 0, 14, 32, 44, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 211, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 149, 253, 14, 32, 47, 0, 14, 32, 45, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 212, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 213, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 214, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 232, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 238, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 239, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 230, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 189, 253, 14, 32, 47, 0, 14, 32, 44, 6, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 151, 253, 14, 32, 47, 0, 14, 32, 44, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 153, 253, 14, 32, 47, 0, 14, 32, 44, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 199, 253, 14, 32, 47, 0, 14, 32, 44, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 150, 253, 14, 32, 47, 0, 14, 32, 45, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 179, 253, 14, 32, 47, 0, 14, 32, 45, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 138, 252, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 139, 252, 14, 32, 47, 0, 14, 32, 50, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 140, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 155, 253, 14, 32, 47, 0, 14, 32, 69, 6, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 154, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 141, 252, 14, 32, 47, 0, 14, 32, 70, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 142, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 143, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 229, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 75, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 76, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 77, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 78, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 79, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 80, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 186, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 29, 222, 14, 32, 61, 0, 14, 32, 59, 216, 93, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 159, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 158, 251, 14, 32, 60, 0, 14, 32, 189, 8, 14, 32, 60, 0, 14, 32, 187, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 162, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 163, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 161, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 160, 251, 14, 32, 60, 0, 14, 32, 188, 6, 14, 32, 60, 0, 14, 32, 189, 6, 14, 32, 60, 0, 14, 32, 185, 6, 14, 32, 60, 0, 14, 32, 103, 7, 14, 32, 60, 0, 14, 32, 104, 7, 14, 32, 60, 0, 14, 32, 105, 7, 14, 32, 60, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 36, 222, 14, 32, 61, 0, 14, 32, 59, 216, 100, 222, 14, 32, 61, 0, 14, 32, 59, 216, 132, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 235, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 217, 252, 14, 32, 47, 0, 112, 6, 60, 0, 60, 0, 60, 0, 14, 32, 215, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 216, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 147, 253, 14, 32, 47, 0, 14, 32, 69, 6, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 148, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 236, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 234, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 233, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 81, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 82, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 83, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 84, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 190, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 172, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 173, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 171, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 170, 251, 14, 32, 60, 0, 14, 32, 193, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 168, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 169, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 167, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 166, 251, 14, 32, 60, 0, 14, 32, 195, 6, 14, 32, 60, 0, 14, 32, 255, 6, 14, 32, 60, 0, 14, 32, 213, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 165, 251, 14, 32, 47, 0, 3, 216, 171, 222, 60, 0, 60, 0, 60, 0, 14, 32, 164, 251, 14, 32, 47, 0, 3, 216, 171, 222, 60, 0, 14, 32, 72, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 229, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 118, 6, 14, 32, 47, 0, 14, 32, 33, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 5, 222, 14, 32, 61, 0, 14, 32, 59, 216, 133, 222, 14, 32, 61, 0, 14, 32, 59, 216, 165, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 238, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 237, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 248, 253, 14, 32, 47, 0, 14, 32, 51, 6, 68, 6, 69, 6, 14, 32, 60, 0, 14, 32, 196, 6, 14, 32, 60, 0, 14, 32, 197, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 225, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 224, 251, 14, 32, 60, 0, 14, 32, 198, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 218, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 217, 251, 14, 32, 60, 0, 14, 32, 199, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 119, 6, 14, 32, 47, 0, 14, 32, 33, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 216, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 215, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 221, 251, 14, 32, 47, 0, 14, 32, 33, 6, 14, 32, 60, 0, 14, 32, 200, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 220, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 219, 251, 14, 32, 60, 0, 14, 32, 201, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 227, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 226, 251, 14, 32, 60, 0, 14, 32, 202, 6, 14, 32, 60, 0, 14, 32, 203, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 223, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 222, 251, 14, 32, 60, 0, 14, 32, 177, 8, 14, 32, 60, 0, 14, 32, 207, 6, 14, 32, 60, 0, 14, 32, 120, 7, 14, 32, 60, 0, 14, 32, 121, 7, 14, 32, 60, 0, 14, 32, 171, 8, 14, 32, 60, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 232, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 233, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 240, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 144, 252, 14, 32, 47, 0, 112, 6, 60, 0, 60, 0, 60, 0, 14, 32, 239, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 93, 252, 14, 32, 47, 0, 112, 6, 60, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 230, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 120, 6, 14, 32, 47, 0, 14, 32, 33, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 59, 216, 9, 222, 14, 32, 61, 0, 14, 32, 59, 216, 41, 222, 14, 32, 61, 0, 14, 32, 59, 216, 73, 222, 14, 32, 61, 0, 14, 32, 59, 216, 105, 222, 14, 32, 61, 0, 14, 32, 59, 216, 137, 222, 14, 32, 61, 0, 14, 32, 59, 216, 169, 222, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 243, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 218, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 219, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 220, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 221, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 157, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 222, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 244, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 240, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 241, 252, 14, 32, 47, 0, 14, 32, 71, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 242, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 175, 253, 14, 32, 47, 0, 14, 32, 44, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 174, 253, 14, 32, 47, 0, 14, 32, 45, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 145, 252, 14, 32, 47, 0, 14, 32, 49, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 146, 252, 14, 32, 47, 0, 14, 32, 50, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 147, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 156, 253, 14, 32, 47, 0, 14, 32, 69, 6, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 176, 253, 14, 32, 47, 0, 14, 32, 69, 6, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 148, 252, 14, 32, 47, 0, 14, 32, 70, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 149, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 150, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 241, 254, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 85, 252, 14, 32, 47, 0, 14, 32, 44, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 86, 252, 14, 32, 47, 0, 14, 32, 45, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 87, 252, 14, 32, 47, 0, 14, 32, 46, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 88, 252, 14, 32, 47, 0, 14, 32, 69, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 89, 252, 14, 32, 47, 0, 14, 32, 73, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 90, 252, 14, 32, 47, 0, 14, 32, 74, 6, 14, 32, 60, 0, 14, 32, 204, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 254, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 255, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 253, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 252, 251, 14, 32, 60, 0, 14, 32, 205, 6, 14, 32, 60, 0, 14, 32, 206, 6, 14, 32, 60, 0, 14, 32, 208, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 230, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 231, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 229, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 228, 251, 14, 32, 60, 0, 14, 32, 209, 6, 14, 32, 60, 0, 14, 32, 61, 6, 14, 32, 60, 0, 14, 32, 62, 6, 14, 32, 60, 0, 14, 32, 63, 6, 14, 32, 60, 0, 14, 32, 32, 6, 14, 32, 60, 0, 14, 32, 117, 7, 14, 32, 60, 0, 14, 32, 118, 7, 14, 32, 60, 0, 14, 32, 119, 7, 14, 32, 60, 0, 14, 32, 186, 8, 14, 32, 60, 0, 14, 32, 210, 6, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 175, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 177, 251, 14, 32, 47, 0, 3, 216, 171, 222, 60, 0, 60, 0, 60, 0, 14, 32, 174, 251, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 176, 251, 14, 32, 47, 0, 3, 216, 171, 222, 60, 0, 14, 32, 122, 7, 14, 32, 60, 0, 14, 32, 123, 7, 14, 32, 60, 0, 14, 32, 16, 7, 14, 32, 60, 0, 14, 32, 18, 7, 14, 32, 60, 0, 60, 0, 14, 32, 45, 7, 14, 32, 60, 0, 14, 32, 19, 7, 14, 32, 60, 0, 60, 0, 14, 32, 20, 7, 14, 32, 60, 0, 60, 0, 14, 32, 46, 7, 14, 32, 60, 0, 14, 32, 22, 7, 14, 32, 60, 0, 14, 32, 21, 7, 14, 32, 60, 0, 60, 0, 14, 32, 47, 7, 14, 32, 60, 0, 14, 32, 23, 7, 14, 32, 60, 0, 14, 32, 24, 7, 14, 32, 60, 0, 14, 32, 25, 7, 14, 32, 60, 0, 14, 32, 77, 7, 14, 32, 60, 0, 14, 32, 26, 7, 14, 32, 60, 0, 14, 32, 27, 7, 14, 32, 60, 0, 60, 0, 14, 32, 28, 7, 14, 32, 60, 0, 14, 32, 29, 7, 14, 32, 60, 0, 14, 32, 30, 7, 14, 32, 60, 0, 14, 32, 31, 7, 14, 32, 60, 0, 14, 32, 78, 7, 14, 32, 60, 0, 14, 32, 32, 7, 14, 32, 60, 0, 14, 32, 33, 7, 14, 32, 60, 0, 14, 32, 34, 7, 14, 32, 60, 0, 14, 32, 35, 7, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 36, 7, 14, 32, 60, 0, 14, 32, 37, 7, 14, 32, 60, 0, 14, 32, 38, 7, 14, 32, 60, 0, 60, 0, 14, 32, 39, 7, 14, 32, 60, 0, 14, 32, 79, 7, 14, 32, 60, 0, 14, 32, 40, 7, 14, 32, 60, 0, 14, 32, 41, 7, 14, 32, 60, 0, 14, 32, 42, 7, 14, 32, 60, 0, 14, 32, 43, 7, 14, 32, 60, 0, 14, 32, 44, 7, 14, 32, 60, 0, 14, 32, 96, 8, 14, 32, 60, 0, 14, 32, 97, 8, 14, 32, 60, 0, 14, 32, 98, 8, 14, 32, 60, 0, 14, 32, 99, 8, 14, 32, 60, 0, 14, 32, 100, 8, 14, 32, 60, 0, 14, 32, 101, 8, 14, 32, 60, 0, 14, 32, 102, 8, 14, 32, 60, 0, 14, 32, 103, 8, 14, 32, 60, 0, 14, 32, 104, 8, 14, 32, 60, 0, 14, 32, 105, 8, 14, 32, 60, 0, 14, 32, 106, 8, 14, 32, 60, 0, 14, 32, 64, 8, 14, 32, 60, 0, 14, 32, 65, 8, 14, 32, 60, 0, 14, 32, 66, 8, 14, 32, 60, 0, 14, 32, 67, 8, 14, 32, 60, 0, 14, 32, 68, 8, 14, 32, 60, 0, 14, 32, 69, 8, 14, 32, 60, 0, 14, 32, 70, 8, 14, 32, 60, 0, 14, 32, 71, 8, 14, 32, 60, 0, 14, 32, 72, 8, 14, 32, 60, 0, 14, 32, 73, 8, 14, 32, 60, 0, 14, 32, 74, 8, 14, 32, 60, 0, 14, 32, 75, 8, 14, 32, 60, 0, 14, 32, 76, 8, 14, 32, 60, 0, 14, 32, 77, 8, 14, 32, 60, 0, 14, 32, 78, 8, 14, 32, 60, 0, 14, 32, 79, 8, 14, 32, 60, 0, 14, 32, 80, 8, 14, 32, 60, 0, 14, 32, 81, 8, 14, 32, 60, 0, 14, 32, 82, 8, 14, 32, 60, 0, 14, 32, 83, 8, 14, 32, 60, 0, 14, 32, 84, 8, 14, 32, 60, 0, 14, 32, 85, 8, 14, 32, 60, 0, 14, 32, 86, 8, 14, 32, 60, 0, 14, 32, 87, 8, 14, 32, 60, 0, 14, 32, 88, 8, 14, 32, 60, 0, 14, 32, 128, 7, 14, 32, 60, 0, 14, 32, 153, 7, 14, 32, 60, 0, 14, 32, 154, 7, 14, 32, 60, 0, 14, 32, 129, 7, 14, 32, 60, 0, 14, 32, 130, 7, 14, 32, 60, 0, 14, 32, 131, 7, 14, 32, 60, 0, 14, 32, 156, 7, 14, 32, 60, 0, 14, 32, 132, 7, 14, 32, 60, 0, 14, 32, 133, 7, 14, 32, 60, 0, 14, 32, 134, 7, 14, 32, 60, 0, 14, 32, 135, 7, 14, 32, 60, 0, 14, 32, 162, 7, 14, 32, 60, 0, 14, 32, 163, 7, 14, 32, 60, 0, 14, 32, 136, 7, 14, 32, 60, 0, 14, 32, 165, 7, 14, 32, 60, 0, 14, 32, 137, 7, 14, 32, 60, 0, 14, 32, 138, 7, 14, 32, 60, 0, 14, 32, 139, 7, 14, 32, 60, 0, 14, 32, 155, 7, 14, 32, 60, 0, 14, 32, 140, 7, 14, 32, 60, 0, 14, 32, 152, 7, 14, 32, 60, 0, 14, 32, 160, 7, 14, 32, 60, 0, 14, 32, 161, 7, 14, 32, 60, 0, 14, 32, 141, 7, 14, 32, 60, 0, 14, 32, 142, 7, 14, 32, 60, 0, 14, 32, 164, 7, 14, 32, 60, 0, 14, 32, 143, 7, 14, 32, 60, 0, 14, 32, 144, 7, 14, 32, 60, 0, 14, 32, 157, 7, 14, 32, 60, 0, 14, 32, 158, 7, 14, 32, 60, 0, 14, 32, 159, 7, 14, 32, 60, 0, 14, 32, 145, 7, 14, 32, 60, 0, 14, 32, 146, 7, 14, 32, 60, 0, 14, 32, 147, 7, 14, 32, 60, 0, 14, 32, 148, 7, 14, 32, 60, 0, 14, 32, 149, 7, 14, 32, 60, 0, 14, 32, 150, 7, 14, 32, 60, 0, 14, 32, 151, 7, 14, 32, 60, 0, 14, 32, 177, 7, 14, 32, 60, 0, 166, 7, 60, 0, 167, 7, 60, 0, 168, 7, 60, 0, 169, 7, 60, 0, 170, 7, 60, 0, 171, 7, 60, 0, 172, 7, 60, 0, 173, 7, 60, 0, 174, 7, 60, 0, 175, 7, 60, 0, 176, 7, 60, 0, 14, 32, 202, 7, 14, 32, 60, 0, 14, 32, 203, 7, 14, 32, 60, 0, 14, 32, 204, 7, 14, 32, 60, 0, 14, 32, 205, 7, 14, 32, 60, 0, 14, 32, 206, 7, 14, 32, 60, 0, 14, 32, 207, 7, 14, 32, 60, 0, 14, 32, 208, 7, 14, 32, 60, 0, 14, 32, 209, 7, 14, 32, 60, 0, 14, 32, 210, 7, 14, 32, 60, 0, 14, 32, 211, 7, 14, 32, 60, 0, 14, 32, 212, 7, 14, 32, 60, 0, 14, 32, 213, 7, 14, 32, 60, 0, 14, 32, 214, 7, 14, 32, 60, 0, 60, 0, 14, 32, 232, 7, 14, 32, 60, 0, 14, 32, 215, 7, 14, 32, 60, 0, 60, 0, 14, 32, 233, 7, 14, 32, 60, 0, 14, 32, 216, 7, 14, 32, 60, 0, 14, 32, 217, 7, 14, 32, 60, 0, 60, 0, 14, 32, 234, 7, 14, 32, 60, 0, 14, 32, 218, 7, 14, 32, 60, 0, 14, 32, 219, 7, 14, 32, 60, 0, 14, 32, 220, 7, 14, 32, 60, 0, 14, 32, 221, 7, 14, 32, 60, 0, 14, 32, 222, 7, 14, 32, 60, 0, 14, 32, 223, 7, 14, 32, 60, 0, 14, 32, 224, 7, 14, 32, 60, 0, 14, 32, 225, 7, 14, 32, 60, 0, 14, 32, 226, 7, 14, 32, 60, 0, 14, 32, 227, 7, 14, 32, 60, 0, 14, 32, 228, 7, 14, 32, 60, 0, 14, 32, 229, 7, 14, 32, 60, 0, 14, 32, 230, 7, 14, 32, 60, 0, 14, 32, 231, 7, 14, 32, 60, 0, 14, 32, 244, 7, 14, 32, 60, 0, 14, 32, 245, 7, 14, 32, 60, 0, 48, 45, 60, 0, 49, 45, 60, 0, 50, 45, 60, 0, 51, 45, 60, 0, 52, 45, 60, 0, 53, 45, 60, 0, 54, 45, 60, 0, 55, 45, 60, 0, 56, 45, 60, 0, 57, 45, 60, 0, 58, 45, 60, 0, 59, 45, 60, 0, 102, 45, 60, 0, 60, 45, 60, 0, 61, 45, 60, 0, 62, 45, 60, 0, 63, 45, 60, 0, 64, 45, 60, 0, 65, 45, 60, 0, 66, 45, 60, 0, 67, 45, 60, 0, 68, 45, 60, 0, 69, 45, 60, 0, 70, 45, 60, 0, 71, 45, 60, 0, 72, 45, 60, 0, 73, 45, 60, 0, 74, 45, 60, 0, 75, 45, 60, 0, 76, 45, 60, 0, 77, 45, 60, 0, 78, 45, 60, 0, 79, 45, 60, 0, 80, 45, 60, 0, 81, 45, 60, 0, 82, 45, 60, 0, 83, 45, 60, 0, 103, 45, 60, 0, 84, 45, 60, 0, 85, 45, 60, 0, 86, 45, 60, 0, 87, 45, 60, 0, 88, 45, 60, 0, 89, 45, 60, 0, 90, 45, 60, 0, 91, 45, 60, 0, 92, 45, 60, 0, 93, 45, 60, 0, 94, 45, 60, 0, 95, 45, 60, 0, 96, 45, 60, 0, 97, 45, 60, 0, 98, 45, 60, 0, 99, 45, 60, 0, 100, 45, 60, 0, 101, 45, 60, 0, 111, 45, 60, 0, 0, 18, 60, 0, 1, 18, 60, 0, 2, 18, 60, 0, 3, 18, 60, 0, 4, 18, 60, 0, 5, 18, 60, 0, 6, 18, 60, 0, 7, 18, 60, 0, 8, 18, 60, 0, 9, 18, 60, 0, 10, 18, 60, 0, 11, 18, 60, 0, 12, 18, 60, 0, 13, 18, 60, 0, 14, 18, 60, 0, 15, 18, 60, 0, 128, 45, 60, 0, 16, 18, 60, 0, 17, 18, 60, 0, 18, 18, 60, 0, 19, 18, 60, 0, 20, 18, 60, 0, 21, 18, 60, 0, 22, 18, 60, 0, 23, 18, 60, 0, 24, 18, 60, 0, 25, 18, 60, 0, 26, 18, 60, 0, 27, 18, 60, 0, 28, 18, 60, 0, 29, 18, 60, 0, 30, 18, 60, 0, 31, 18, 60, 0, 128, 19, 60, 0, 129, 19, 60, 0, 130, 19, 60, 0, 131, 19, 60, 0, 129, 45, 60, 0, 32, 18, 60, 0, 33, 18, 60, 0, 34, 18, 60, 0, 35, 18, 60, 0, 36, 18, 60, 0, 37, 18, 60, 0, 38, 18, 60, 0, 39, 18, 60, 0, 40, 18, 60, 0, 41, 18, 60, 0, 42, 18, 60, 0, 43, 18, 60, 0, 44, 18, 60, 0, 45, 18, 60, 0, 46, 18, 60, 0, 47, 18, 60, 0, 130, 45, 60, 0, 48, 18, 60, 0, 49, 18, 60, 0, 50, 18, 60, 0, 51, 18, 60, 0, 52, 18, 60, 0, 53, 18, 60, 0, 54, 18, 60, 0, 55, 18, 60, 0, 131, 45, 60, 0, 1, 171, 60, 0, 2, 171, 60, 0, 3, 171, 60, 0, 4, 171, 60, 0, 5, 171, 60, 0, 6, 171, 60, 0, 56, 18, 60, 0, 57, 18, 60, 0, 58, 18, 60, 0, 59, 18, 60, 0, 60, 18, 60, 0, 61, 18, 60, 0, 62, 18, 60, 0, 63, 18, 60, 0, 132, 45, 60, 0, 64, 18, 60, 0, 65, 18, 60, 0, 66, 18, 60, 0, 67, 18, 60, 0, 68, 18, 60, 0, 69, 18, 60, 0, 70, 18, 60, 0, 71, 18, 60, 0, 72, 18, 60, 0, 74, 18, 60, 0, 75, 18, 60, 0, 76, 18, 60, 0, 77, 18, 60, 0, 80, 18, 60, 0, 81, 18, 60, 0, 82, 18, 60, 0, 83, 18, 60, 0, 84, 18, 60, 0, 85, 18, 60, 0, 86, 18, 60, 0, 88, 18, 60, 0, 90, 18, 60, 0, 91, 18, 60, 0, 92, 18, 60, 0, 93, 18, 60, 0, 96, 18, 60, 0, 97, 18, 60, 0, 98, 18, 60, 0, 99, 18, 60, 0, 100, 18, 60, 0, 101, 18, 60, 0, 102, 18, 60, 0, 103, 18, 60, 0, 132, 19, 60, 0, 133, 19, 60, 0, 134, 19, 60, 0, 135, 19, 60, 0, 133, 45, 60, 0, 104, 18, 60, 0, 105, 18, 60, 0, 106, 18, 60, 0, 107, 18, 60, 0, 108, 18, 60, 0, 109, 18, 60, 0, 110, 18, 60, 0, 111, 18, 60, 0, 112, 18, 60, 0, 113, 18, 60, 0, 114, 18, 60, 0, 115, 18, 60, 0, 116, 18, 60, 0, 117, 18, 60, 0, 118, 18, 60, 0, 119, 18, 60, 0, 134, 45, 60, 0, 120, 18, 60, 0, 121, 18, 60, 0, 122, 18, 60, 0, 123, 18, 60, 0, 124, 18, 60, 0, 125, 18, 60, 0, 126, 18, 60, 0, 127, 18, 60, 0, 135, 45, 60, 0, 128, 18, 60, 0, 129, 18, 60, 0, 130, 18, 60, 0, 131, 18, 60, 0, 132, 18, 60, 0, 133, 18, 60, 0, 134, 18, 60, 0, 135, 18, 60, 0, 136, 18, 60, 0, 138, 18, 60, 0, 139, 18, 60, 0, 140, 18, 60, 0, 141, 18, 60, 0, 144, 18, 60, 0, 145, 18, 60, 0, 146, 18, 60, 0, 147, 18, 60, 0, 148, 18, 60, 0, 149, 18, 60, 0, 150, 18, 60, 0, 151, 18, 60, 0, 136, 45, 60, 0, 152, 18, 60, 0, 153, 18, 60, 0, 154, 18, 60, 0, 155, 18, 60, 0, 156, 18, 60, 0, 157, 18, 60, 0, 158, 18, 60, 0, 159, 18, 60, 0, 137, 45, 60, 0, 160, 18, 60, 0, 161, 18, 60, 0, 162, 18, 60, 0, 163, 18, 60, 0, 164, 18, 60, 0, 165, 18, 60, 0, 166, 18, 60, 0, 167, 18, 60, 0, 138, 45, 60, 0, 168, 18, 60, 0, 169, 18, 60, 0, 170, 18, 60, 0, 171, 18, 60, 0, 172, 18, 60, 0, 173, 18, 60, 0, 174, 18, 60, 0, 175, 18, 60, 0, 176, 18, 60, 0, 178, 18, 60, 0, 179, 18, 60, 0, 180, 18, 60, 0, 181, 18, 60, 0, 184, 18, 60, 0, 185, 18, 60, 0, 186, 18, 60, 0, 187, 18, 60, 0, 188, 18, 60, 0, 189, 18, 60, 0, 190, 18, 60, 0, 192, 18, 60, 0, 194, 18, 60, 0, 195, 18, 60, 0, 196, 18, 60, 0, 197, 18, 60, 0, 200, 18, 60, 0, 201, 18, 60, 0, 202, 18, 60, 0, 203, 18, 60, 0, 204, 18, 60, 0, 205, 18, 60, 0, 206, 18, 60, 0, 207, 18, 60, 0, 208, 18, 60, 0, 209, 18, 60, 0, 210, 18, 60, 0, 211, 18, 60, 0, 212, 18, 60, 0, 213, 18, 60, 0, 214, 18, 60, 0, 216, 18, 60, 0, 217, 18, 60, 0, 218, 18, 60, 0, 219, 18, 60, 0, 220, 18, 60, 0, 221, 18, 60, 0, 222, 18, 60, 0, 223, 18, 60, 0, 139, 45, 60, 0, 17, 171, 60, 0, 18, 171, 60, 0, 19, 171, 60, 0, 20, 171, 60, 0, 21, 171, 60, 0, 22, 171, 60, 0, 224, 18, 60, 0, 225, 18, 60, 0, 226, 18, 60, 0, 227, 18, 60, 0, 228, 18, 60, 0, 229, 18, 60, 0, 230, 18, 60, 0, 231, 18, 60, 0, 232, 18, 60, 0, 233, 18, 60, 0, 234, 18, 60, 0, 235, 18, 60, 0, 236, 18, 60, 0, 237, 18, 60, 0, 238, 18, 60, 0, 239, 18, 60, 0, 240, 18, 60, 0, 241, 18, 60, 0, 242, 18, 60, 0, 243, 18, 60, 0, 244, 18, 60, 0, 245, 18, 60, 0, 246, 18, 60, 0, 247, 18, 60, 0, 140, 45, 60, 0, 9, 171, 60, 0, 10, 171, 60, 0, 11, 171, 60, 0, 12, 171, 60, 0, 13, 171, 60, 0, 14, 171, 60, 0, 248, 18, 60, 0, 249, 18, 60, 0, 250, 18, 60, 0, 251, 18, 60, 0, 252, 18, 60, 0, 253, 18, 60, 0, 254, 18, 60, 0, 255, 18, 60, 0, 141, 45, 60, 0, 0, 19, 60, 0, 1, 19, 60, 0, 2, 19, 60, 0, 3, 19, 60, 0, 4, 19, 60, 0, 5, 19, 60, 0, 6, 19, 60, 0, 7, 19, 60, 0, 142, 45, 60, 0, 8, 19, 60, 0, 9, 19, 60, 0, 10, 19, 60, 0, 11, 19, 60, 0, 12, 19, 60, 0, 13, 19, 60, 0, 14, 19, 60, 0, 15, 19, 60, 0, 16, 19, 60, 0, 18, 19, 60, 0, 19, 19, 60, 0, 20, 19, 60, 0, 21, 19, 60, 0, 24, 19, 60, 0, 25, 19, 60, 0, 26, 19, 60, 0, 27, 19, 60, 0, 28, 19, 60, 0, 29, 19, 60, 0, 30, 19, 60, 0, 31, 19, 60, 0, 147, 45, 60, 0, 148, 45, 60, 0, 149, 45, 60, 0, 150, 45, 60, 0, 32, 19, 60, 0, 33, 19, 60, 0, 34, 19, 60, 0, 35, 19, 60, 0, 36, 19, 60, 0, 37, 19, 60, 0, 38, 19, 60, 0, 39, 19, 60, 0, 143, 45, 60, 0, 40, 19, 60, 0, 41, 19, 60, 0, 42, 19, 60, 0, 43, 19, 60, 0, 44, 19, 60, 0, 45, 19, 60, 0, 46, 19, 60, 0, 47, 19, 60, 0, 144, 45, 60, 0, 32, 171, 60, 0, 33, 171, 60, 0, 34, 171, 60, 0, 35, 171, 60, 0, 36, 171, 60, 0, 37, 171, 60, 0, 38, 171, 60, 0, 48, 19, 60, 0, 49, 19, 60, 0, 50, 19, 60, 0, 51, 19, 60, 0, 52, 19, 60, 0, 53, 19, 60, 0, 54, 19, 60, 0, 55, 19, 60, 0, 145, 45, 60, 0, 56, 19, 60, 0, 57, 19, 60, 0, 58, 19, 60, 0, 59, 19, 60, 0, 60, 19, 60, 0, 61, 19, 60, 0, 62, 19, 60, 0, 63, 19, 60, 0, 40, 171, 60, 0, 41, 171, 60, 0, 42, 171, 60, 0, 43, 171, 60, 0, 44, 171, 60, 0, 45, 171, 60, 0, 46, 171, 60, 0, 64, 19, 60, 0, 65, 19, 60, 0, 66, 19, 60, 0, 67, 19, 60, 0, 68, 19, 60, 0, 69, 19, 60, 0, 70, 19, 60, 0, 71, 19, 60, 0, 72, 19, 60, 0, 73, 19, 60, 0, 74, 19, 60, 0, 75, 19, 60, 0, 76, 19, 60, 0, 77, 19, 60, 0, 78, 19, 60, 0, 79, 19, 60, 0, 136, 19, 60, 0, 137, 19, 60, 0, 138, 19, 60, 0, 139, 19, 60, 0, 80, 19, 60, 0, 81, 19, 60, 0, 82, 19, 60, 0, 83, 19, 60, 0, 84, 19, 60, 0, 85, 19, 60, 0, 86, 19, 60, 0, 87, 19, 60, 0, 140, 19, 60, 0, 141, 19, 60, 0, 142, 19, 60, 0, 143, 19, 60, 0, 146, 45, 60, 0, 88, 19, 60, 0, 89, 19, 60, 0, 90, 19, 60, 0, 160, 45, 60, 0, 161, 45, 60, 0, 162, 45, 60, 0, 163, 45, 60, 0, 164, 45, 60, 0, 165, 45, 60, 0, 166, 45, 60, 0, 168, 45, 60, 0, 169, 45, 60, 0, 170, 45, 60, 0, 171, 45, 60, 0, 172, 45, 60, 0, 173, 45, 60, 0, 174, 45, 60, 0, 176, 45, 60, 0, 177, 45, 60, 0, 178, 45, 60, 0, 179, 45, 60, 0, 180, 45, 60, 0, 181, 45, 60, 0, 182, 45, 60, 0, 184, 45, 60, 0, 185, 45, 60, 0, 186, 45, 60, 0, 187, 45, 60, 0, 188, 45, 60, 0, 189, 45, 60, 0, 190, 45, 60, 0, 192, 45, 60, 0, 193, 45, 60, 0, 194, 45, 60, 0, 195, 45, 60, 0, 196, 45, 60, 0, 197, 45, 60, 0, 198, 45, 60, 0, 200, 45, 60, 0, 201, 45, 60, 0, 202, 45, 60, 0, 203, 45, 60, 0, 204, 45, 60, 0, 205, 45, 60, 0, 206, 45, 60, 0, 208, 45, 60, 0, 209, 45, 60, 0, 210, 45, 60, 0, 211, 45, 60, 0, 212, 45, 60, 0, 213, 45, 60, 0, 214, 45, 60, 0, 216, 45, 60, 0, 217, 45, 60, 0, 218, 45, 60, 0, 219, 45, 60, 0, 220, 45, 60, 0, 221, 45, 60, 0, 222, 45, 60, 0, 80, 9, 60, 0, 253, 168, 60, 0, 114, 9, 60, 0, 4, 9, 60, 0, 5, 9, 60, 0, 6, 9, 60, 0, 115, 9, 60, 0, 116, 9, 60, 0, 117, 9, 60, 0, 118, 9, 60, 0, 119, 9, 60, 0, 7, 9, 60, 0, 8, 9, 60, 0, 9, 9, 60, 0, 10, 9, 60, 0, 11, 9, 60, 0, 96, 9, 60, 0, 12, 9, 60, 0, 97, 9, 60, 0, 13, 9, 60, 0, 14, 9, 60, 0, 15, 9, 60, 0, 16, 9, 60, 0, 254, 168, 60, 0, 17, 9, 60, 0, 18, 9, 60, 0, 19, 9, 60, 0, 20, 9, 60, 0, 21, 9, 60, 0, 22, 9, 60, 0, 23, 9, 60, 0, 123, 9, 60, 0, 24, 9, 60, 0, 25, 9, 60, 0, 26, 9, 60, 0, 27, 9, 60, 0, 28, 9, 60, 0, 121, 9, 60, 0, 124, 9, 60, 0, 29, 9, 60, 0, 30, 9, 60, 0, 31, 9, 60, 0, 32, 9, 60, 0, 120, 9, 60, 0, 33, 9, 60, 0, 126, 9, 60, 0, 34, 9, 60, 0, 35, 9, 60, 0, 36, 9, 60, 0, 37, 9, 60, 0, 38, 9, 60, 0, 39, 9, 60, 0, 40, 9, 60, 0, 42, 9, 60, 0, 43, 9, 60, 0, 44, 9, 60, 0, 127, 9, 60, 0, 45, 9, 60, 0, 46, 9, 60, 0, 47, 9, 60, 0, 122, 9, 60, 0, 48, 9, 60, 0, 50, 9, 60, 0, 51, 9, 60, 0, 53, 9, 60, 0, 54, 9, 60, 0, 55, 9, 60, 0, 56, 9, 60, 0, 57, 9, 60, 0, 61, 9, 60, 0, 125, 9, 60, 0, 233, 28, 60, 0, 60, 0, 60, 0, 234, 28, 61, 0, 235, 28, 61, 0, 236, 28, 61, 0, 238, 28, 61, 0, 239, 28, 61, 0, 240, 28, 61, 0, 241, 28, 61, 0, 250, 28, 60, 0, 245, 28, 60, 0, 246, 28, 60, 0, 242, 168, 60, 0, 60, 0, 60, 0, 243, 168, 61, 0, 244, 168, 61, 0, 245, 168, 61, 0, 246, 168, 61, 0, 247, 168, 60, 0, 251, 168, 60, 0, 62, 9, 60, 0, 58, 9, 60, 0, 59, 9, 60, 0, 79, 9, 60, 0, 86, 9, 60, 0, 87, 9, 60, 0, 63, 9, 60, 0, 64, 9, 60, 0, 65, 9, 60, 0, 66, 9, 60, 0, 67, 9, 60, 0, 68, 9, 60, 0, 98, 9, 60, 0, 99, 9, 60, 0, 69, 9, 60, 0, 85, 9, 60, 0, 70, 9, 60, 0, 71, 9, 60, 0, 78, 9, 60, 0, 72, 9, 60, 0, 255, 168, 60, 0, 73, 9, 60, 0, 74, 9, 60, 0, 75, 9, 60, 0, 76, 9, 60, 0, 77, 9, 60, 0, 128, 9, 60, 0, 133, 9, 60, 0, 134, 9, 60, 0, 135, 9, 60, 0, 136, 9, 60, 0, 137, 9, 60, 0, 138, 9, 60, 0, 139, 9, 60, 0, 224, 9, 60, 0, 140, 9, 60, 0, 225, 9, 60, 0, 143, 9, 60, 0, 144, 9, 60, 0, 147, 9, 60, 0, 148, 9, 60, 0, 149, 9, 60, 0, 150, 9, 60, 0, 151, 9, 60, 0, 152, 9, 60, 0, 153, 9, 60, 0, 154, 9, 60, 0, 155, 9, 60, 0, 156, 9, 60, 0, 157, 9, 60, 0, 158, 9, 60, 0, 159, 9, 60, 0, 160, 9, 60, 0, 161, 9, 60, 0, 162, 9, 60, 0, 163, 9, 60, 0, 164, 9, 60, 0, 60, 0, 60, 0, 206, 9, 47, 0, 205, 9, 60, 0, 165, 9, 60, 0, 166, 9, 60, 0, 167, 9, 60, 0, 168, 9, 60, 0, 170, 9, 60, 0, 171, 9, 60, 0, 172, 9, 60, 0, 173, 9, 60, 0, 174, 9, 60, 0, 175, 9, 60, 0, 176, 9, 60, 0, 240, 9, 60, 0, 178, 9, 60, 0, 241, 9, 60, 0, 182, 9, 60, 0, 183, 9, 60, 0, 184, 9, 60, 0, 185, 9, 60, 0, 189, 9, 60, 0, 190, 9, 60, 0, 191, 9, 60, 0, 192, 9, 60, 0, 193, 9, 60, 0, 194, 9, 60, 0, 195, 9, 60, 0, 196, 9, 60, 0, 226, 9, 60, 0, 227, 9, 60, 0, 199, 9, 60, 0, 200, 9, 60, 0, 203, 9, 60, 0, 204, 9, 60, 0, 205, 9, 60, 0, 215, 9, 60, 0, 252, 9, 60, 0, 116, 10, 60, 0, 115, 10, 60, 0, 9, 10, 60, 0, 10, 10, 60, 0, 19, 10, 60, 0, 5, 10, 60, 0, 6, 10, 60, 0, 16, 10, 60, 0, 20, 10, 60, 0, 114, 10, 60, 0, 7, 10, 60, 0, 8, 10, 60, 0, 15, 10, 60, 0, 56, 10, 60, 0, 57, 10, 60, 0, 81, 10, 60, 0, 21, 10, 60, 0, 22, 10, 60, 0, 23, 10, 60, 0, 24, 10, 60, 0, 25, 10, 60, 0, 26, 10, 60, 0, 27, 10, 60, 0, 28, 10, 60, 0, 29, 10, 60, 0, 30, 10, 60, 0, 31, 10, 60, 0, 32, 10, 60, 0, 33, 10, 60, 0, 34, 10, 60, 0, 35, 10, 60, 0, 36, 10, 60, 0, 37, 10, 60, 0, 38, 10, 60, 0, 39, 10, 60, 0, 40, 10, 60, 0, 42, 10, 60, 0, 43, 10, 60, 0, 44, 10, 60, 0, 45, 10, 60, 0, 46, 10, 60, 0, 47, 10, 60, 0, 117, 10, 60, 0, 48, 10, 60, 0, 50, 10, 60, 0, 53, 10, 60, 0, 92, 10, 60, 0, 62, 10, 60, 0, 63, 10, 60, 0, 64, 10, 60, 0, 65, 10, 60, 0, 66, 10, 60, 0, 71, 10, 60, 0, 72, 10, 60, 0, 75, 10, 60, 0, 76, 10, 60, 0, 77, 10, 60, 0, 208, 10, 60, 0, 133, 10, 60, 0, 134, 10, 60, 0, 135, 10, 60, 0, 136, 10, 60, 0, 137, 10, 60, 0, 138, 10, 60, 0, 139, 10, 60, 0, 224, 10, 60, 0, 140, 10, 60, 0, 225, 10, 60, 0, 141, 10, 60, 0, 143, 10, 60, 0, 144, 10, 60, 0, 145, 10, 60, 0, 147, 10, 60, 0, 148, 10, 60, 0, 149, 10, 60, 0, 150, 10, 60, 0, 151, 10, 60, 0, 152, 10, 60, 0, 153, 10, 60, 0, 154, 10, 60, 0, 155, 10, 60, 0, 156, 10, 60, 0, 249, 10, 60, 0, 157, 10, 60, 0, 158, 10, 60, 0, 159, 10, 60, 0, 160, 10, 60, 0, 161, 10, 60, 0, 162, 10, 60, 0, 163, 10, 60, 0, 164, 10, 60, 0, 165, 10, 60, 0, 166, 10, 60, 0, 167, 10, 60, 0, 168, 10, 60, 0, 170, 10, 60, 0, 171, 10, 60, 0, 172, 10, 60, 0, 173, 10, 60, 0, 174, 10, 60, 0, 175, 10, 60, 0, 176, 10, 60, 0, 178, 10, 60, 0, 181, 10, 60, 0, 182, 10, 60, 0, 183, 10, 60, 0, 184, 10, 60, 0, 185, 10, 60, 0, 179, 10, 60, 0, 189, 10, 60, 0, 190, 10, 60, 0, 191, 10, 60, 0, 192, 10, 60, 0, 193, 10, 60, 0, 194, 10, 60, 0, 195, 10, 60, 0, 196, 10, 60, 0, 226, 10, 60, 0, 227, 10, 60, 0, 197, 10, 60, 0, 199, 10, 60, 0, 200, 10, 60, 0, 201, 10, 60, 0, 203, 10, 60, 0, 204, 10, 60, 0, 205, 10, 60, 0, 5, 11, 60, 0, 6, 11, 60, 0, 7, 11, 60, 0, 8, 11, 60, 0, 9, 11, 60, 0, 10, 11, 60, 0, 11, 11, 60, 0, 96, 11, 60, 0, 12, 11, 60, 0, 97, 11, 60, 0, 15, 11, 60, 0, 16, 11, 60, 0, 19, 11, 60, 0, 20, 11, 60, 0, 21, 11, 60, 0, 22, 11, 60, 0, 23, 11, 60, 0, 24, 11, 60, 0, 25, 11, 60, 0, 26, 11, 60, 0, 27, 11, 60, 0, 28, 11, 60, 0, 29, 11, 60, 0, 30, 11, 60, 0, 31, 11, 60, 0, 32, 11, 60, 0, 33, 11, 60, 0, 34, 11, 60, 0, 35, 11, 60, 0, 36, 11, 60, 0, 37, 11, 60, 0, 38, 11, 60, 0, 39, 11, 60, 0, 40, 11, 60, 0, 42, 11, 60, 0, 43, 11, 60, 0, 44, 11, 60, 0, 45, 11, 60, 0, 46, 11, 60, 0, 47, 11, 60, 0, 95, 11, 60, 0, 48, 11, 60, 0, 50, 11, 60, 0, 51, 11, 60, 0, 53, 11, 60, 0, 113, 11, 60, 0, 54, 11, 60, 0, 55, 11, 60, 0, 56, 11, 60, 0, 57, 11, 60, 0, 61, 11, 60, 0, 62, 11, 60, 0, 63, 11, 60, 0, 64, 11, 60, 0, 65, 11, 60, 0, 66, 11, 60, 0, 67, 11, 60, 0, 68, 11, 60, 0, 98, 11, 60, 0, 99, 11, 60, 0, 71, 11, 60, 0, 72, 11, 60, 0, 75, 11, 60, 0, 76, 11, 60, 0, 77, 11, 60, 0, 86, 11, 60, 0, 87, 11, 60, 0, 208, 11, 60, 0, 133, 11, 60, 0, 134, 11, 60, 0, 135, 11, 60, 0, 136, 11, 60, 0, 137, 11, 60, 0, 138, 11, 60, 0, 142, 11, 60, 0, 143, 11, 60, 0, 144, 11, 60, 0, 146, 11, 60, 0, 147, 11, 60, 0, 148, 11, 60, 0, 131, 11, 60, 0, 149, 11, 60, 0, 153, 11, 60, 0, 154, 11, 60, 0, 158, 11, 60, 0, 159, 11, 60, 0, 163, 11, 60, 0, 164, 11, 60, 0, 168, 11, 60, 0, 170, 11, 60, 0, 174, 11, 60, 0, 175, 11, 60, 0, 176, 11, 60, 0, 178, 11, 60, 0, 181, 11, 60, 0, 180, 11, 60, 0, 179, 11, 60, 0, 177, 11, 60, 0, 169, 11, 60, 0, 156, 11, 60, 0, 182, 11, 60, 0, 183, 11, 60, 0, 184, 11, 60, 0, 185, 11, 60, 0, 190, 11, 60, 0, 191, 11, 60, 0, 192, 11, 60, 0, 193, 11, 60, 0, 194, 11, 60, 0, 198, 11, 60, 0, 199, 11, 60, 0, 200, 11, 60, 0, 202, 11, 60, 0, 203, 11, 60, 0, 204, 11, 60, 0, 205, 11, 60, 0, 215, 11, 60, 0, 5, 12, 60, 0, 6, 12, 60, 0, 7, 12, 60, 0, 8, 12, 60, 0, 9, 12, 60, 0, 10, 12, 60, 0, 11, 12, 60, 0, 96, 12, 60, 0, 12, 12, 60, 0, 97, 12, 60, 0, 14, 12, 60, 0, 15, 12, 60, 0, 16, 12, 60, 0, 18, 12, 60, 0, 19, 12, 60, 0, 20, 12, 60, 0, 21, 12, 60, 0, 22, 12, 60, 0, 23, 12, 60, 0, 24, 12, 60, 0, 25, 12, 60, 0, 26, 12, 60, 0, 88, 12, 60, 0, 27, 12, 60, 0, 28, 12, 60, 0, 89, 12, 60, 0, 29, 12, 60, 0, 30, 12, 60, 0, 31, 12, 60, 0, 32, 12, 60, 0, 33, 12, 60, 0, 34, 12, 60, 0, 35, 12, 60, 0, 36, 12, 60, 0, 37, 12, 60, 0, 38, 12, 60, 0, 39, 12, 60, 0, 40, 12, 60, 0, 42, 12, 60, 0, 43, 12, 60, 0, 44, 12, 60, 0, 45, 12, 60, 0, 46, 12, 60, 0, 47, 12, 60, 0, 48, 12, 60, 0, 49, 12, 60, 0, 50, 12, 60, 0, 53, 12, 60, 0, 54, 12, 60, 0, 55, 12, 60, 0, 56, 12, 60, 0, 57, 12, 60, 0, 51, 12, 60, 0, 52, 12, 60, 0, 90, 12, 60, 0, 61, 12, 60, 0, 62, 12, 60, 0, 63, 12, 60, 0, 64, 12, 60, 0, 65, 12, 60, 0, 66, 12, 60, 0, 67, 12, 60, 0, 68, 12, 60, 0, 98, 12, 60, 0, 99, 12, 60, 0, 70, 12, 60, 0, 71, 12, 60, 0, 72, 12, 60, 0, 74, 12, 60, 0, 75, 12, 60, 0, 76, 12, 60, 0, 77, 12, 60, 0, 85, 12, 60, 0, 86, 12, 60, 0, 133, 12, 60, 0, 134, 12, 60, 0, 135, 12, 60, 0, 136, 12, 60, 0, 137, 12, 60, 0, 138, 12, 60, 0, 139, 12, 60, 0, 224, 12, 60, 0, 140, 12, 60, 0, 225, 12, 60, 0, 142, 12, 60, 0, 143, 12, 60, 0, 144, 12, 60, 0, 146, 12, 60, 0, 147, 12, 60, 0, 148, 12, 60, 0, 149, 12, 60, 0, 150, 12, 60, 0, 151, 12, 60, 0, 152, 12, 60, 0, 153, 12, 60, 0, 154, 12, 60, 0, 155, 12, 60, 0, 156, 12, 60, 0, 157, 12, 60, 0, 158, 12, 60, 0, 159, 12, 60, 0, 160, 12, 60, 0, 161, 12, 60, 0, 162, 12, 60, 0, 163, 12, 60, 0, 164, 12, 60, 0, 165, 12, 60, 0, 166, 12, 60, 0, 167, 12, 60, 0, 168, 12, 60, 0, 170, 12, 60, 0, 171, 12, 60, 0, 172, 12, 60, 0, 173, 12, 60, 0, 174, 12, 60, 0, 175, 12, 60, 0, 176, 12, 60, 0, 177, 12, 60, 0, 178, 12, 60, 0, 181, 12, 60, 0, 182, 12, 60, 0, 183, 12, 60, 0, 184, 12, 60, 0, 185, 12, 60, 0, 179, 12, 60, 0, 222, 12, 60, 0, 189, 12, 60, 0, 241, 12, 60, 0, 242, 12, 60, 0, 128, 12, 60, 0, 190, 12, 60, 0, 191, 12, 60, 0, 192, 12, 60, 0, 193, 12, 60, 0, 194, 12, 60, 0, 195, 12, 60, 0, 196, 12, 60, 0, 226, 12, 60, 0, 227, 12, 60, 0, 198, 12, 60, 0, 199, 12, 60, 0, 200, 12, 60, 0, 202, 12, 60, 0, 203, 12, 61, 0, 203, 12, 60, 0, 204, 12, 60, 0, 205, 12, 60, 0, 213, 12, 60, 0, 214, 12, 60, 0, 5, 13, 60, 0, 6, 13, 60, 0, 7, 13, 60, 0, 8, 13, 60, 0, 95, 13, 60, 0, 9, 13, 60, 0, 10, 13, 60, 0, 11, 13, 60, 0, 96, 13, 60, 0, 12, 13, 60, 0, 97, 13, 60, 0, 14, 13, 60, 0, 15, 13, 60, 0, 16, 13, 60, 0, 18, 13, 60, 0, 19, 13, 60, 0, 20, 13, 60, 0, 21, 13, 60, 0, 60, 0, 60, 0, 127, 13, 47, 0, 77, 13, 60, 0, 22, 13, 60, 0, 23, 13, 60, 0, 24, 13, 60, 0, 25, 13, 60, 0, 26, 13, 60, 0, 27, 13, 60, 0, 28, 13, 60, 0, 29, 13, 60, 0, 30, 13, 60, 0, 31, 13, 60, 0, 32, 13, 60, 0, 33, 13, 60, 0, 34, 13, 60, 0, 35, 13, 60, 0, 60, 0, 60, 0, 122, 13, 47, 0, 77, 13, 60, 0, 36, 13, 60, 0, 37, 13, 60, 0, 38, 13, 60, 0, 39, 13, 60, 0, 40, 13, 60, 0, 60, 0, 60, 0, 123, 13, 47, 0, 77, 13, 60, 0, 41, 13, 60, 0, 42, 13, 60, 0, 43, 13, 60, 0, 44, 13, 60, 0, 45, 13, 60, 0, 46, 13, 60, 0, 60, 0, 60, 0, 84, 13, 47, 0, 77, 13, 60, 0, 47, 13, 60, 0, 60, 0, 60, 0, 85, 13, 47, 0, 77, 13, 60, 0, 48, 13, 60, 0, 60, 0, 60, 0, 78, 13, 47, 0, 77, 13, 61, 0, 124, 13, 47, 0, 77, 13, 60, 0, 50, 13, 60, 0, 60, 0, 60, 0, 125, 13, 47, 0, 77, 13, 60, 0, 53, 13, 60, 0, 54, 13, 60, 0, 55, 13, 60, 0, 56, 13, 60, 0, 57, 13, 60, 0, 51, 13, 60, 0, 60, 0, 60, 0, 126, 13, 47, 0, 77, 13, 60, 0, 52, 13, 60, 0, 60, 0, 60, 0, 86, 13, 47, 0, 77, 13, 60, 0, 49, 13, 60, 0, 58, 13, 60, 0, 61, 13, 60, 0, 4, 13, 60, 0, 62, 13, 60, 0, 63, 13, 60, 0, 64, 13, 60, 0, 65, 13, 60, 0, 66, 13, 60, 0, 67, 13, 60, 0, 68, 13, 60, 0, 98, 13, 60, 0, 99, 13, 60, 0, 70, 13, 60, 0, 71, 13, 60, 0, 72, 13, 60, 0, 74, 13, 60, 0, 75, 13, 60, 0, 76, 13, 60, 0, 87, 13, 60, 0, 77, 13, 60, 0, 60, 0, 60, 0, 59, 13, 61, 0, 60, 13, 60, 0, 133, 13, 60, 0, 134, 13, 60, 0, 135, 13, 60, 0, 136, 13, 60, 0, 137, 13, 60, 0, 138, 13, 60, 0, 139, 13, 60, 0, 140, 13, 60, 0, 141, 13, 60, 0, 142, 13, 60, 0, 143, 13, 60, 0, 144, 13, 60, 0, 145, 13, 60, 0, 146, 13, 60, 0, 147, 13, 60, 0, 148, 13, 60, 0, 149, 13, 60, 0, 150, 13, 60, 0, 154, 13, 60, 0, 155, 13, 60, 0, 156, 13, 60, 0, 157, 13, 60, 0, 158, 13, 60, 0, 159, 13, 60, 0, 160, 13, 60, 0, 161, 13, 60, 0, 162, 13, 60, 0, 163, 13, 60, 0, 164, 13, 60, 0, 165, 13, 60, 0, 166, 13, 60, 0, 167, 13, 60, 0, 168, 13, 60, 0, 169, 13, 60, 0, 170, 13, 60, 0, 171, 13, 60, 0, 172, 13, 60, 0, 173, 13, 60, 0, 174, 13, 60, 0, 175, 13, 60, 0, 176, 13, 60, 0, 177, 13, 60, 0, 179, 13, 60, 0, 180, 13, 60, 0, 181, 13, 60, 0, 182, 13, 60, 0, 183, 13, 60, 0, 184, 13, 60, 0, 185, 13, 60, 0, 186, 13, 60, 0, 187, 13, 60, 0, 189, 13, 60, 0, 192, 13, 60, 0, 193, 13, 60, 0, 194, 13, 60, 0, 195, 13, 60, 0, 196, 13, 60, 0, 197, 13, 60, 0, 198, 13, 60, 0, 207, 13, 60, 0, 208, 13, 60, 0, 209, 13, 60, 0, 210, 13, 60, 0, 211, 13, 60, 0, 212, 13, 60, 0, 214, 13, 60, 0, 216, 13, 60, 0, 242, 13, 60, 0, 223, 13, 60, 0, 243, 13, 60, 0, 217, 13, 60, 0, 218, 13, 60, 0, 219, 13, 60, 0, 220, 13, 60, 0, 221, 13, 61, 0, 221, 13, 60, 0, 222, 13, 60, 0, 202, 13, 60, 0, 242, 170, 60, 0, 192, 171, 60, 0, 193, 171, 60, 0, 194, 171, 60, 0, 195, 171, 60, 0, 196, 171, 60, 0, 197, 171, 60, 0, 198, 171, 60, 0, 199, 171, 60, 0, 200, 171, 60, 0, 201, 171, 60, 0, 202, 171, 60, 0, 203, 171, 60, 0, 204, 171, 60, 0, 205, 171, 60, 0, 206, 171, 60, 0, 207, 171, 60, 0, 208, 171, 60, 0, 209, 171, 60, 0, 210, 171, 60, 0, 211, 171, 60, 0, 212, 171, 60, 0, 213, 171, 60, 0, 214, 171, 60, 0, 215, 171, 60, 0, 216, 171, 60, 0, 217, 171, 60, 0, 218, 171, 60, 0, 224, 170, 60, 0, 225, 170, 60, 0, 226, 170, 60, 0, 227, 170, 60, 0, 228, 170, 60, 0, 229, 170, 60, 0, 230, 170, 60, 0, 231, 170, 60, 0, 232, 170, 60, 0, 233, 170, 60, 0, 234, 170, 60, 0, 227, 171, 60, 0, 228, 171, 60, 0, 229, 171, 60, 0, 230, 171, 60, 0, 231, 171, 60, 0, 232, 171, 60, 0, 233, 171, 60, 0, 234, 171, 60, 0, 235, 170, 60, 0, 236, 170, 60, 0, 237, 170, 60, 0, 238, 170, 60, 0, 239, 170, 60, 0, 245, 170, 60, 0, 219, 171, 60, 0, 220, 171, 60, 0, 221, 171, 60, 0, 222, 171, 60, 0, 223, 171, 60, 0, 224, 171, 60, 0, 225, 171, 60, 0, 226, 171, 60, 0, 237, 171, 60, 0, 246, 170, 60, 0, 0, 168, 60, 0, 1, 168, 60, 0, 2, 168, 60, 0, 3, 168, 60, 0, 4, 168, 60, 0, 5, 168, 60, 0, 6, 168, 60, 0, 44, 168, 60, 0, 7, 168, 60, 0, 8, 168, 60, 0, 9, 168, 60, 0, 10, 168, 60, 0, 12, 168, 60, 0, 13, 168, 60, 0, 14, 168, 60, 0, 15, 168, 60, 0, 16, 168, 60, 0, 17, 168, 60, 0, 18, 168, 60, 0, 19, 168, 60, 0, 20, 168, 60, 0, 21, 168, 60, 0, 22, 168, 60, 0, 23, 168, 60, 0, 24, 168, 60, 0, 25, 168, 60, 0, 26, 168, 60, 0, 27, 168, 60, 0, 28, 168, 60, 0, 29, 168, 60, 0, 30, 168, 60, 0, 31, 168, 60, 0, 32, 168, 60, 0, 33, 168, 60, 0, 34, 168, 60, 0, 35, 168, 60, 0, 36, 168, 60, 0, 37, 168, 60, 0, 38, 168, 60, 0, 39, 168, 60, 0, 130, 168, 60, 0, 131, 168, 60, 0, 132, 168, 60, 0, 133, 168, 60, 0, 134, 168, 60, 0, 135, 168, 60, 0, 136, 168, 60, 0, 137, 168, 60, 0, 138, 168, 60, 0, 139, 168, 60, 0, 140, 168, 60, 0, 141, 168, 60, 0, 142, 168, 60, 0, 143, 168, 60, 0, 144, 168, 60, 0, 145, 168, 60, 0, 146, 168, 60, 0, 147, 168, 60, 0, 148, 168, 60, 0, 149, 168, 60, 0, 150, 168, 60, 0, 151, 168, 60, 0, 152, 168, 60, 0, 153, 168, 60, 0, 154, 168, 60, 0, 155, 168, 60, 0, 156, 168, 60, 0, 157, 168, 60, 0, 158, 168, 60, 0, 159, 168, 60, 0, 160, 168, 60, 0, 161, 168, 60, 0, 162, 168, 60, 0, 163, 168, 60, 0, 164, 168, 60, 0, 165, 168, 60, 0, 166, 168, 60, 0, 167, 168, 60, 0, 168, 168, 60, 0, 169, 168, 60, 0, 170, 168, 60, 0, 171, 168, 60, 0, 172, 168, 60, 0, 173, 168, 60, 0, 174, 168, 60, 0, 175, 168, 60, 0, 176, 168, 60, 0, 177, 168, 60, 0, 178, 168, 60, 0, 179, 168, 60, 0, 180, 168, 60, 0, 181, 168, 60, 0, 182, 168, 60, 0, 183, 168, 60, 0, 184, 168, 60, 0, 185, 168, 60, 0, 186, 168, 60, 0, 187, 168, 60, 0, 188, 168, 60, 0, 189, 168, 60, 0, 190, 168, 60, 0, 191, 168, 60, 0, 192, 168, 60, 0, 193, 168, 60, 0, 194, 168, 60, 0, 195, 168, 60, 0, 196, 168, 60, 0, 4, 216, 131, 220, 60, 0, 4, 216, 132, 220, 60, 0, 4, 216, 133, 220, 60, 0, 4, 216, 134, 220, 60, 0, 4, 216, 135, 220, 60, 0, 4, 216, 136, 220, 60, 0, 4, 216, 137, 220, 60, 0, 4, 216, 138, 220, 60, 0, 4, 216, 139, 220, 60, 0, 4, 216, 140, 220, 60, 0, 4, 216, 141, 220, 60, 0, 4, 216, 142, 220, 60, 0, 4, 216, 143, 220, 60, 0, 4, 216, 144, 220, 60, 0, 4, 216, 145, 220, 60, 0, 4, 216, 146, 220, 60, 0, 4, 216, 147, 220, 60, 0, 4, 216, 148, 220, 60, 0, 4, 216, 149, 220, 60, 0, 4, 216, 150, 220, 60, 0, 4, 216, 151, 220, 60, 0, 4, 216, 152, 220, 60, 0, 4, 216, 153, 220, 60, 0, 4, 216, 155, 220, 60, 0, 4, 216, 157, 220, 60, 0, 4, 216, 158, 220, 60, 0, 4, 216, 159, 220, 60, 0, 4, 216, 160, 220, 60, 0, 4, 216, 161, 220, 60, 0, 4, 216, 162, 220, 60, 0, 4, 216, 163, 220, 60, 0, 4, 216, 164, 220, 60, 0, 4, 216, 165, 220, 60, 0, 4, 216, 166, 220, 60, 0, 4, 216, 167, 220, 60, 0, 4, 216, 168, 220, 60, 0, 4, 216, 169, 220, 60, 0, 4, 216, 170, 220, 60, 0, 4, 216, 172, 220, 60, 0, 4, 216, 173, 220, 60, 0, 4, 216, 174, 220, 60, 0, 4, 216, 175, 220, 60, 0, 4, 216, 176, 220, 60, 0, 4, 216, 177, 220, 60, 0, 4, 216, 178, 220, 60, 0, 4, 216, 179, 220, 60, 0, 4, 216, 180, 220, 60, 0, 4, 216, 181, 220, 60, 0, 4, 216, 182, 220, 60, 0, 4, 216, 183, 220, 60, 0, 4, 216, 184, 220, 60, 0, 4, 216, 185, 220, 60, 0, 4, 216, 80, 221, 60, 0, 4, 216, 81, 221, 60, 0, 4, 216, 82, 221, 60, 0, 4, 216, 83, 221, 60, 0, 4, 216, 84, 221, 60, 0, 4, 216, 85, 221, 60, 0, 4, 216, 86, 221, 60, 0, 4, 216, 87, 221, 60, 0, 4, 216, 88, 221, 60, 0, 4, 216, 89, 221, 60, 0, 4, 216, 90, 221, 60, 0, 4, 216, 91, 221, 60, 0, 4, 216, 92, 221, 60, 0, 4, 216, 93, 221, 60, 0, 4, 216, 94, 221, 60, 0, 4, 216, 95, 221, 60, 0, 4, 216, 96, 221, 60, 0, 4, 216, 97, 221, 60, 0, 4, 216, 98, 221, 60, 0, 4, 216, 99, 221, 60, 0, 4, 216, 100, 221, 60, 0, 4, 216, 101, 221, 60, 0, 4, 216, 102, 221, 60, 0, 4, 216, 103, 221, 60, 0, 4, 216, 104, 221, 60, 0, 4, 216, 105, 221, 60, 0, 4, 216, 106, 221, 60, 0, 4, 216, 107, 221, 60, 0, 4, 216, 108, 221, 60, 0, 4, 216, 109, 221, 60, 0, 4, 216, 110, 221, 60, 0, 4, 216, 111, 221, 60, 0, 4, 216, 118, 221, 60, 0, 4, 216, 112, 221, 60, 0, 4, 216, 113, 221, 60, 0, 4, 216, 114, 221, 60, 0, 4, 216, 196, 221, 60, 0, 4, 216, 218, 221, 60, 0, 4, 216, 131, 221, 60, 0, 4, 216, 132, 221, 60, 0, 4, 216, 133, 221, 60, 0, 4, 216, 134, 221, 60, 0, 4, 216, 135, 221, 60, 0, 4, 216, 136, 221, 60, 0, 4, 216, 137, 221, 60, 0, 4, 216, 138, 221, 60, 0, 4, 216, 139, 221, 60, 0, 4, 216, 140, 221, 60, 0, 4, 216, 141, 221, 60, 0, 4, 216, 142, 221, 60, 0, 4, 216, 143, 221, 60, 0, 4, 216, 144, 221, 60, 0, 4, 216, 145, 221, 60, 0, 4, 216, 146, 221, 60, 0, 4, 216, 147, 221, 60, 0, 4, 216, 148, 221, 60, 0, 4, 216, 149, 221, 60, 0, 4, 216, 150, 221, 60, 0, 4, 216, 151, 221, 60, 0, 4, 216, 152, 221, 60, 0, 4, 216, 153, 221, 60, 0, 4, 216, 154, 221, 60, 0, 4, 216, 155, 221, 60, 0, 4, 216, 156, 221, 60, 0, 4, 216, 157, 221, 60, 0, 4, 216, 158, 221, 60, 0, 4, 216, 159, 221, 60, 0, 4, 216, 160, 221, 60, 0, 4, 216, 161, 221, 60, 0, 4, 216, 162, 221, 60, 0, 4, 216, 163, 221, 60, 0, 4, 216, 164, 221, 60, 0, 4, 216, 165, 221, 60, 0, 4, 216, 166, 221, 60, 0, 4, 216, 167, 221, 60, 0, 4, 216, 168, 221, 60, 0, 4, 216, 169, 221, 60, 0, 4, 216, 170, 221, 60, 0, 4, 216, 171, 221, 60, 0, 4, 216, 172, 221, 60, 0, 4, 216, 173, 221, 60, 0, 4, 216, 174, 221, 60, 0, 4, 216, 175, 221, 60, 0, 4, 216, 176, 221, 60, 0, 4, 216, 177, 221, 60, 0, 4, 216, 178, 221, 60, 0, 4, 216, 193, 221, 60, 0, 4, 216, 194, 221, 60, 0, 4, 216, 195, 221, 60, 0, 4, 216, 220, 221, 60, 0, 4, 216, 179, 221, 60, 0, 4, 216, 180, 221, 60, 0, 4, 216, 181, 221, 60, 0, 4, 216, 182, 221, 60, 0, 4, 216, 183, 221, 60, 0, 4, 216, 184, 221, 60, 0, 4, 216, 185, 221, 60, 0, 4, 216, 186, 221, 60, 0, 4, 216, 187, 221, 60, 0, 4, 216, 188, 221, 60, 0, 4, 216, 206, 221, 60, 0, 4, 216, 189, 221, 60, 0, 4, 216, 190, 221, 60, 0, 4, 216, 191, 221, 60, 0, 4, 216, 192, 221, 60, 0, 4, 216, 0, 222, 60, 0, 4, 216, 1, 222, 60, 0, 4, 216, 2, 222, 60, 0, 4, 216, 3, 222, 60, 0, 4, 216, 4, 222, 60, 0, 4, 216, 5, 222, 60, 0, 4, 216, 6, 222, 60, 0, 4, 216, 7, 222, 60, 0, 4, 216, 8, 222, 60, 0, 4, 216, 9, 222, 60, 0, 4, 216, 10, 222, 60, 0, 4, 216, 11, 222, 60, 0, 4, 216, 12, 222, 60, 0, 4, 216, 13, 222, 60, 0, 4, 216, 14, 222, 60, 0, 4, 216, 15, 222, 60, 0, 4, 216, 16, 222, 60, 0, 4, 216, 17, 222, 60, 0, 4, 216, 19, 222, 60, 0, 4, 216, 20, 222, 60, 0, 4, 216, 21, 222, 60, 0, 4, 216, 22, 222, 60, 0, 4, 216, 23, 222, 60, 0, 4, 216, 24, 222, 60, 0, 4, 216, 25, 222, 60, 0, 4, 216, 26, 222, 60, 0, 4, 216, 27, 222, 60, 0, 4, 216, 28, 222, 60, 0, 4, 216, 29, 222, 60, 0, 4, 216, 30, 222, 60, 0, 4, 216, 31, 222, 60, 0, 4, 216, 32, 222, 60, 0, 4, 216, 33, 222, 60, 0, 4, 216, 34, 222, 60, 0, 4, 216, 35, 222, 60, 0, 4, 216, 36, 222, 60, 0, 4, 216, 37, 222, 60, 0, 4, 216, 38, 222, 60, 0, 4, 216, 39, 222, 60, 0, 4, 216, 40, 222, 60, 0, 4, 216, 41, 222, 60, 0, 4, 216, 42, 222, 60, 0, 4, 216, 43, 222, 60, 0, 4, 216, 44, 222, 60, 0, 4, 216, 45, 222, 60, 0, 4, 216, 46, 222, 60, 0, 4, 216, 47, 222, 60, 0, 4, 216, 48, 222, 60, 0, 4, 216, 49, 222, 60, 0, 4, 216, 50, 222, 60, 0, 4, 216, 51, 222, 60, 0, 4, 216, 53, 222, 60, 0, 4, 216, 176, 222, 60, 0, 4, 216, 177, 222, 60, 0, 4, 216, 178, 222, 60, 0, 4, 216, 179, 222, 60, 0, 4, 216, 180, 222, 60, 0, 4, 216, 181, 222, 60, 0, 4, 216, 182, 222, 60, 0, 4, 216, 183, 222, 60, 0, 4, 216, 184, 222, 60, 0, 4, 216, 185, 222, 60, 0, 4, 216, 186, 222, 60, 0, 4, 216, 187, 222, 60, 0, 4, 216, 188, 222, 60, 0, 4, 216, 189, 222, 60, 0, 4, 216, 190, 222, 60, 0, 4, 216, 191, 222, 60, 0, 4, 216, 192, 222, 60, 0, 4, 216, 193, 222, 60, 0, 4, 216, 194, 222, 60, 0, 4, 216, 195, 222, 60, 0, 4, 216, 196, 222, 60, 0, 4, 216, 197, 222, 60, 0, 4, 216, 198, 222, 60, 0, 4, 216, 199, 222, 60, 0, 4, 216, 200, 222, 60, 0, 4, 216, 201, 222, 60, 0, 4, 216, 202, 222, 60, 0, 4, 216, 203, 222, 60, 0, 4, 216, 204, 222, 60, 0, 4, 216, 205, 222, 60, 0, 4, 216, 206, 222, 60, 0, 4, 216, 207, 222, 60, 0, 4, 216, 208, 222, 60, 0, 4, 216, 209, 222, 60, 0, 4, 216, 210, 222, 60, 0, 4, 216, 211, 222, 60, 0, 4, 216, 212, 222, 60, 0, 4, 216, 213, 222, 60, 0, 4, 216, 214, 222, 60, 0, 4, 216, 215, 222, 60, 0, 4, 216, 216, 222, 60, 0, 4, 216, 217, 222, 60, 0, 4, 216, 218, 222, 60, 0, 4, 216, 219, 222, 60, 0, 4, 216, 220, 222, 60, 0, 4, 216, 221, 222, 60, 0, 4, 216, 222, 222, 60, 0, 4, 216, 224, 222, 60, 0, 4, 216, 225, 222, 60, 0, 4, 216, 226, 222, 60, 0, 4, 216, 227, 222, 60, 0, 4, 216, 228, 222, 60, 0, 4, 216, 229, 222, 60, 0, 4, 216, 230, 222, 60, 0, 4, 216, 231, 222, 60, 0, 4, 216, 232, 222, 60, 0, 4, 216, 234, 222, 60, 0, 4, 216, 128, 222, 60, 0, 4, 216, 129, 222, 60, 0, 4, 216, 130, 222, 60, 0, 4, 216, 131, 222, 60, 0, 4, 216, 165, 222, 60, 0, 4, 216, 166, 222, 60, 0, 4, 216, 132, 222, 60, 0, 4, 216, 133, 222, 60, 0, 4, 216, 134, 222, 60, 0, 4, 216, 136, 222, 60, 0, 4, 216, 138, 222, 60, 0, 4, 216, 139, 222, 60, 0, 4, 216, 140, 222, 60, 0, 4, 216, 141, 222, 60, 0, 4, 216, 143, 222, 60, 0, 4, 216, 144, 222, 60, 0, 4, 216, 145, 222, 60, 0, 4, 216, 146, 222, 60, 0, 4, 216, 147, 222, 60, 0, 4, 216, 148, 222, 60, 0, 4, 216, 149, 222, 60, 0, 4, 216, 150, 222, 60, 0, 4, 216, 151, 222, 60, 0, 4, 216, 152, 222, 60, 0, 4, 216, 153, 222, 60, 0, 4, 216, 154, 222, 60, 0, 4, 216, 155, 222, 60, 0, 4, 216, 156, 222, 60, 0, 4, 216, 157, 222, 60, 0, 4, 216, 159, 222, 60, 0, 4, 216, 160, 222, 60, 0, 4, 216, 161, 222, 60, 0, 4, 216, 162, 222, 60, 0, 4, 216, 163, 222, 60, 0, 4, 216, 164, 222, 60, 0, 4, 216, 167, 222, 60, 0, 4, 216, 168, 222, 60, 0, 4, 216, 80, 223, 60, 0, 4, 216, 5, 223, 60, 0, 4, 216, 6, 223, 60, 0, 4, 216, 7, 223, 60, 0, 4, 216, 8, 223, 60, 0, 4, 216, 9, 223, 60, 0, 4, 216, 10, 223, 60, 0, 4, 216, 11, 223, 60, 0, 4, 216, 96, 223, 60, 0, 4, 216, 12, 223, 60, 0, 4, 216, 97, 223, 60, 0, 4, 216, 15, 223, 60, 0, 4, 216, 16, 223, 60, 0, 4, 216, 19, 223, 60, 0, 4, 216, 20, 223, 60, 0, 4, 216, 21, 223, 60, 0, 4, 216, 22, 223, 60, 0, 4, 216, 23, 223, 60, 0, 4, 216, 24, 223, 60, 0, 4, 216, 25, 223, 60, 0, 4, 216, 26, 223, 60, 0, 4, 216, 27, 223, 60, 0, 4, 216, 28, 223, 60, 0, 4, 216, 29, 223, 60, 0, 4, 216, 30, 223, 60, 0, 4, 216, 31, 223, 60, 0, 4, 216, 32, 223, 60, 0, 4, 216, 33, 223, 60, 0, 4, 216, 34, 223, 60, 0, 4, 216, 35, 223, 60, 0, 4, 216, 36, 223, 60, 0, 4, 216, 37, 223, 60, 0, 4, 216, 38, 223, 60, 0, 4, 216, 39, 223, 60, 0, 4, 216, 40, 223, 60, 0, 4, 216, 42, 223, 60, 0, 4, 216, 43, 223, 60, 0, 4, 216, 44, 223, 60, 0, 4, 216, 45, 223, 60, 0, 4, 216, 46, 223, 60, 0, 4, 216, 47, 223, 60, 0, 4, 216, 48, 223, 60, 0, 4, 216, 50, 223, 60, 0, 4, 216, 51, 223, 60, 0, 4, 216, 53, 223, 60, 0, 4, 216, 54, 223, 60, 0, 4, 216, 55, 223, 60, 0, 4, 216, 56, 223, 60, 0, 4, 216, 57, 223, 60, 0, 4, 216, 61, 223, 60, 0, 4, 216, 94, 223, 60, 0, 4, 216, 95, 223, 60, 0, 4, 216, 62, 223, 60, 0, 4, 216, 63, 223, 60, 0, 4, 216, 64, 223, 60, 0, 4, 216, 65, 223, 60, 0, 4, 216, 66, 223, 60, 0, 4, 216, 67, 223, 60, 0, 4, 216, 68, 223, 60, 0, 4, 216, 98, 223, 60, 0, 4, 216, 99, 223, 60, 0, 4, 216, 71, 223, 60, 0, 4, 216, 72, 223, 60, 0, 4, 216, 75, 223, 60, 0, 4, 216, 76, 223, 60, 0, 4, 216, 77, 223, 60, 0, 4, 216, 87, 223, 60, 0, 4, 216, 93, 223, 60, 0, 5, 216, 73, 220, 60, 0, 5, 216, 74, 220, 60, 0, 5, 216, 0, 220, 60, 0, 5, 216, 1, 220, 60, 0, 5, 216, 2, 220, 60, 0, 5, 216, 3, 220, 60, 0, 5, 216, 4, 220, 60, 0, 5, 216, 5, 220, 60, 0, 5, 216, 6, 220, 60, 0, 5, 216, 7, 220, 60, 0, 5, 216, 8, 220, 60, 0, 5, 216, 9, 220, 60, 0, 5, 216, 10, 220, 60, 0, 5, 216, 11, 220, 60, 0, 5, 216, 12, 220, 60, 0, 5, 216, 13, 220, 60, 0, 5, 216, 14, 220, 60, 0, 5, 216, 15, 220, 60, 0, 5, 216, 16, 220, 60, 0, 5, 216, 17, 220, 60, 0, 5, 216, 18, 220, 60, 0, 5, 216, 19, 220, 60, 0, 5, 216, 20, 220, 60, 0, 5, 216, 21, 220, 60, 0, 5, 216, 22, 220, 60, 0, 5, 216, 23, 220, 60, 0, 5, 216, 24, 220, 60, 0, 5, 216, 25, 220, 60, 0, 5, 216, 26, 220, 60, 0, 5, 216, 27, 220, 60, 0, 5, 216, 28, 220, 60, 0, 5, 216, 29, 220, 60, 0, 5, 216, 30, 220, 60, 0, 5, 216, 31, 220, 60, 0, 5, 216, 32, 220, 60, 0, 5, 216, 33, 220, 60, 0, 5, 216, 34, 220, 60, 0, 5, 216, 35, 220, 60, 0, 5, 216, 36, 220, 60, 0, 5, 216, 37, 220, 60, 0, 5, 216, 38, 220, 60, 0, 5, 216, 39, 220, 60, 0, 5, 216, 40, 220, 60, 0, 5, 216, 41, 220, 60, 0, 5, 216, 42, 220, 60, 0, 5, 216, 43, 220, 60, 0, 5, 216, 44, 220, 60, 0, 5, 216, 45, 220, 60, 0, 5, 216, 46, 220, 60, 0, 5, 216, 47, 220, 60, 0, 5, 216, 48, 220, 60, 0, 5, 216, 49, 220, 60, 0, 5, 216, 50, 220, 60, 0, 5, 216, 51, 220, 60, 0, 5, 216, 52, 220, 60, 0, 5, 216, 71, 220, 60, 0, 5, 216, 96, 220, 60, 0, 5, 216, 97, 220, 60, 0, 5, 216, 72, 220, 60, 0, 5, 216, 95, 220, 60, 0, 5, 216, 53, 220, 60, 0, 5, 216, 54, 220, 60, 0, 5, 216, 55, 220, 60, 0, 5, 216, 56, 220, 60, 0, 5, 216, 57, 220, 60, 0, 5, 216, 58, 220, 60, 0, 5, 216, 59, 220, 60, 0, 5, 216, 60, 220, 60, 0, 5, 216, 61, 220, 60, 0, 5, 216, 62, 220, 60, 0, 5, 216, 63, 220, 60, 0, 5, 216, 64, 220, 60, 0, 5, 216, 65, 220, 60, 0, 5, 216, 66, 220, 60, 0, 5, 216, 199, 220, 60, 0, 5, 216, 128, 220, 60, 0, 5, 216, 129, 220, 60, 0, 5, 216, 130, 220, 60, 0, 5, 216, 131, 220, 60, 0, 5, 216, 132, 220, 60, 0, 5, 216, 133, 220, 60, 0, 5, 216, 134, 220, 60, 0, 5, 216, 135, 220, 60, 0, 5, 216, 136, 220, 60, 0, 5, 216, 137, 220, 60, 0, 5, 216, 138, 220, 60, 0, 5, 216, 139, 220, 60, 0, 5, 216, 140, 220, 60, 0, 5, 216, 141, 220, 60, 0, 5, 216, 142, 220, 60, 0, 5, 216, 143, 220, 60, 0, 5, 216, 144, 220, 60, 0, 5, 216, 145, 220, 60, 0, 5, 216, 146, 220, 60, 0, 5, 216, 147, 220, 60, 0, 5, 216, 148, 220, 60, 0, 5, 216, 149, 220, 60, 0, 5, 216, 150, 220, 60, 0, 5, 216, 151, 220, 60, 0, 5, 216, 152, 220, 60, 0, 5, 216, 153, 220, 60, 0, 5, 216, 154, 220, 60, 0, 5, 216, 155, 220, 60, 0, 5, 216, 156, 220, 60, 0, 5, 216, 157, 220, 60, 0, 5, 216, 158, 220, 60, 0, 5, 216, 159, 220, 60, 0, 5, 216, 160, 220, 60, 0, 5, 216, 161, 220, 60, 0, 5, 216, 162, 220, 60, 0, 5, 216, 163, 220, 60, 0, 5, 216, 164, 220, 60, 0, 5, 216, 165, 220, 60, 0, 5, 216, 166, 220, 60, 0, 5, 216, 167, 220, 60, 0, 5, 216, 168, 220, 60, 0, 5, 216, 169, 220, 60, 0, 5, 216, 170, 220, 60, 0, 5, 216, 171, 220, 60, 0, 5, 216, 172, 220, 60, 0, 5, 216, 173, 220, 60, 0, 5, 216, 174, 220, 60, 0, 5, 216, 175, 220, 60, 0, 5, 216, 196, 220, 60, 0, 5, 216, 197, 220, 60, 0, 5, 216, 176, 220, 60, 0, 5, 216, 177, 220, 60, 0, 5, 216, 178, 220, 60, 0, 5, 216, 179, 220, 60, 0, 5, 216, 180, 220, 60, 0, 5, 216, 181, 220, 60, 0, 5, 216, 182, 220, 60, 0, 5, 216, 183, 220, 60, 0, 5, 216, 184, 220, 60, 0, 5, 216, 185, 220, 60, 0, 5, 216, 186, 220, 60, 0, 5, 216, 187, 220, 60, 0, 5, 216, 188, 220, 60, 0, 5, 216, 189, 220, 60, 0, 5, 216, 190, 220, 60, 0, 5, 216, 194, 220, 60, 0, 5, 216, 128, 221, 60, 0, 5, 216, 129, 221, 60, 0, 5, 216, 130, 221, 60, 0, 60, 0, 5, 216, 216, 221, 60, 0, 60, 0, 5, 216, 217, 221, 60, 0, 5, 216, 131, 221, 60, 0, 60, 0, 5, 216, 218, 221, 60, 0, 5, 216, 132, 221, 60, 0, 60, 0, 5, 216, 219, 221, 60, 0, 5, 216, 133, 221, 60, 0, 5, 216, 134, 221, 60, 0, 5, 216, 135, 221, 60, 0, 5, 216, 136, 221, 60, 0, 5, 216, 137, 221, 60, 0, 5, 216, 138, 221, 60, 0, 5, 216, 139, 221, 60, 0, 5, 216, 140, 221, 60, 0, 5, 216, 141, 221, 60, 0, 5, 216, 142, 221, 60, 0, 5, 216, 143, 221, 60, 0, 5, 216, 144, 221, 60, 0, 5, 216, 145, 221, 60, 0, 5, 216, 146, 221, 60, 0, 5, 216, 147, 221, 60, 0, 5, 216, 148, 221, 60, 0, 5, 216, 149, 221, 60, 0, 5, 216, 150, 221, 60, 0, 5, 216, 151, 221, 60, 0, 5, 216, 152, 221, 60, 0, 5, 216, 153, 221, 60, 0, 5, 216, 154, 221, 60, 0, 5, 216, 155, 221, 60, 0, 5, 216, 156, 221, 60, 0, 5, 216, 157, 221, 60, 0, 5, 216, 158, 221, 60, 0, 5, 216, 159, 221, 60, 0, 5, 216, 160, 221, 60, 0, 5, 216, 161, 221, 60, 0, 5, 216, 162, 221, 60, 0, 5, 216, 163, 221, 60, 0, 5, 216, 164, 221, 60, 0, 5, 216, 165, 221, 60, 0, 5, 216, 166, 221, 60, 0, 5, 216, 167, 221, 60, 0, 5, 216, 168, 221, 60, 0, 5, 216, 169, 221, 60, 0, 5, 216, 170, 221, 60, 0, 5, 216, 171, 221, 60, 0, 5, 216, 172, 221, 60, 0, 5, 216, 173, 221, 60, 0, 5, 216, 174, 221, 60, 0, 5, 216, 175, 221, 60, 0, 5, 216, 176, 221, 60, 0, 5, 216, 177, 221, 60, 0, 5, 216, 178, 221, 60, 0, 60, 0, 5, 216, 220, 221, 60, 0, 5, 216, 179, 221, 60, 0, 60, 0, 5, 216, 221, 221, 60, 0, 5, 216, 180, 221, 60, 0, 5, 216, 181, 221, 60, 0, 5, 216, 184, 221, 60, 0, 5, 216, 185, 221, 60, 0, 5, 216, 186, 221, 60, 0, 5, 216, 187, 221, 60, 0, 5, 216, 191, 221, 60, 0, 5, 216, 0, 222, 60, 0, 5, 216, 1, 222, 60, 0, 5, 216, 2, 222, 60, 0, 5, 216, 3, 222, 60, 0, 5, 216, 4, 222, 60, 0, 5, 216, 5, 222, 60, 0, 5, 216, 6, 222, 60, 0, 5, 216, 7, 222, 60, 0, 5, 216, 8, 222, 60, 0, 5, 216, 9, 222, 60, 0, 5, 216, 10, 222, 60, 0, 5, 216, 11, 222, 60, 0, 5, 216, 12, 222, 60, 0, 5, 216, 13, 222, 60, 0, 5, 216, 14, 222, 60, 0, 5, 216, 15, 222, 60, 0, 5, 216, 16, 222, 60, 0, 5, 216, 17, 222, 60, 0, 5, 216, 18, 222, 60, 0, 5, 216, 19, 222, 60, 0, 5, 216, 20, 222, 60, 0, 5, 216, 21, 222, 60, 0, 5, 216, 22, 222, 60, 0, 5, 216, 23, 222, 60, 0, 5, 216, 24, 222, 60, 0, 5, 216, 25, 222, 60, 0, 5, 216, 26, 222, 60, 0, 5, 216, 27, 222, 60, 0, 5, 216, 28, 222, 60, 0, 5, 216, 29, 222, 60, 0, 5, 216, 30, 222, 60, 0, 5, 216, 31, 222, 60, 0, 5, 216, 32, 222, 60, 0, 5, 216, 33, 222, 60, 0, 5, 216, 34, 222, 60, 0, 5, 216, 35, 222, 60, 0, 5, 216, 36, 222, 60, 0, 5, 216, 37, 222, 60, 0, 5, 216, 38, 222, 60, 0, 5, 216, 39, 222, 60, 0, 5, 216, 40, 222, 60, 0, 5, 216, 41, 222, 60, 0, 5, 216, 42, 222, 60, 0, 5, 216, 43, 222, 60, 0, 5, 216, 44, 222, 60, 0, 5, 216, 45, 222, 60, 0, 5, 216, 46, 222, 60, 0, 5, 216, 47, 222, 60, 0, 5, 216, 48, 222, 60, 0, 5, 216, 49, 222, 60, 0, 5, 216, 50, 222, 60, 0, 5, 216, 51, 222, 60, 0, 5, 216, 52, 222, 60, 0, 5, 216, 53, 222, 60, 0, 5, 216, 54, 222, 60, 0, 5, 216, 55, 222, 60, 0, 5, 216, 56, 222, 60, 0, 5, 216, 57, 222, 60, 0, 5, 216, 58, 222, 60, 0, 5, 216, 59, 222, 60, 0, 5, 216, 60, 222, 60, 0, 5, 216, 63, 222, 60, 0, 5, 216, 68, 222, 60, 0, 5, 216, 128, 222, 60, 0, 5, 216, 129, 222, 60, 0, 5, 216, 130, 222, 60, 0, 5, 216, 131, 222, 60, 0, 5, 216, 132, 222, 60, 0, 5, 216, 133, 222, 60, 0, 5, 216, 134, 222, 60, 0, 5, 216, 135, 222, 60, 0, 5, 216, 136, 222, 60, 0, 5, 216, 137, 222, 60, 0, 5, 216, 168, 222, 60, 0, 5, 216, 167, 222, 60, 0, 5, 216, 169, 222, 60, 0, 5, 216, 138, 222, 60, 0, 5, 216, 139, 222, 60, 0, 5, 216, 184, 222, 60, 0, 5, 216, 140, 222, 60, 0, 5, 216, 141, 222, 60, 0, 5, 216, 142, 222, 60, 0, 5, 216, 143, 222, 60, 0, 5, 216, 144, 222, 60, 0, 5, 216, 145, 222, 60, 0, 5, 216, 146, 222, 60, 0, 5, 216, 147, 222, 60, 0, 5, 216, 148, 222, 60, 0, 5, 216, 149, 222, 60, 0, 5, 216, 150, 222, 60, 0, 5, 216, 151, 222, 60, 0, 5, 216, 152, 222, 60, 0, 5, 216, 153, 222, 60, 0, 5, 216, 154, 222, 60, 0, 5, 216, 155, 222, 60, 0, 5, 216, 156, 222, 60, 0, 5, 216, 157, 222, 60, 0, 5, 216, 158, 222, 60, 0, 5, 216, 159, 222, 60, 0, 5, 216, 160, 222, 60, 0, 5, 216, 161, 222, 60, 0, 5, 216, 162, 222, 60, 0, 5, 216, 163, 222, 60, 0, 5, 216, 164, 222, 60, 0, 5, 216, 165, 222, 60, 0, 5, 216, 166, 222, 60, 0, 5, 216, 170, 222, 60, 0, 5, 216, 173, 222, 60, 0, 5, 216, 174, 222, 60, 0, 5, 216, 175, 222, 60, 0, 5, 216, 176, 222, 60, 0, 5, 216, 177, 222, 60, 0, 5, 216, 178, 222, 60, 0, 5, 216, 179, 222, 60, 0, 5, 216, 180, 222, 60, 0, 5, 216, 181, 222, 60, 0, 5, 216, 182, 222, 60, 0, 6, 216, 0, 221, 60, 0, 6, 216, 1, 221, 60, 0, 6, 216, 2, 221, 60, 0, 6, 216, 3, 221, 60, 0, 6, 216, 4, 221, 60, 0, 6, 216, 5, 221, 60, 0, 6, 216, 6, 221, 60, 0, 6, 216, 9, 221, 60, 0, 6, 216, 12, 221, 60, 0, 6, 216, 13, 221, 60, 0, 6, 216, 14, 221, 60, 0, 6, 216, 15, 221, 60, 0, 6, 216, 16, 221, 60, 0, 6, 216, 63, 221, 60, 0, 6, 216, 17, 221, 60, 0, 6, 216, 18, 221, 60, 0, 6, 216, 19, 221, 60, 0, 6, 216, 21, 221, 60, 0, 6, 216, 22, 221, 60, 0, 6, 216, 24, 221, 60, 0, 6, 216, 25, 221, 60, 0, 6, 216, 26, 221, 60, 0, 6, 216, 27, 221, 60, 0, 6, 216, 28, 221, 60, 0, 6, 216, 29, 221, 60, 0, 6, 216, 30, 221, 60, 0, 6, 216, 31, 221, 60, 0, 6, 216, 32, 221, 60, 0, 6, 216, 33, 221, 60, 0, 6, 216, 34, 221, 60, 0, 6, 216, 35, 221, 60, 0, 6, 216, 36, 221, 60, 0, 6, 216, 37, 221, 60, 0, 6, 216, 64, 221, 60, 0, 6, 216, 38, 221, 60, 0, 6, 216, 65, 221, 60, 0, 6, 216, 39, 221, 60, 0, 6, 216, 66, 221, 60, 0, 6, 216, 40, 221, 60, 0, 6, 216, 41, 221, 60, 0, 6, 216, 42, 221, 60, 0, 6, 216, 43, 221, 60, 0, 6, 216, 44, 221, 60, 0, 6, 216, 45, 221, 60, 0, 6, 216, 46, 221, 60, 0, 6, 216, 47, 221, 60, 0, 6, 216, 48, 221, 60, 0, 6, 216, 49, 221, 60, 0, 6, 216, 50, 221, 60, 0, 6, 216, 51, 221, 60, 0, 6, 216, 52, 221, 60, 0, 6, 216, 53, 221, 60, 0, 6, 216, 55, 221, 60, 0, 6, 216, 56, 221, 60, 0, 6, 216, 61, 221, 60, 0, 6, 216, 62, 221, 60, 0, 6, 216, 160, 221, 60, 0, 6, 216, 161, 221, 60, 0, 6, 216, 162, 221, 60, 0, 6, 216, 163, 221, 60, 0, 6, 216, 164, 221, 60, 0, 6, 216, 165, 221, 60, 0, 6, 216, 166, 221, 60, 0, 6, 216, 167, 221, 60, 0, 6, 216, 170, 221, 60, 0, 6, 216, 171, 221, 60, 0, 6, 216, 172, 221, 60, 0, 6, 216, 173, 221, 60, 0, 6, 216, 174, 221, 60, 0, 6, 216, 175, 221, 60, 0, 6, 216, 176, 221, 60, 0, 6, 216, 177, 221, 60, 0, 6, 216, 178, 221, 60, 0, 6, 216, 179, 221, 60, 0, 6, 216, 180, 221, 60, 0, 6, 216, 181, 221, 60, 0, 6, 216, 182, 221, 60, 0, 6, 216, 183, 221, 60, 0, 6, 216, 184, 221, 60, 0, 6, 216, 185, 221, 60, 0, 6, 216, 186, 221, 60, 0, 6, 216, 187, 221, 60, 0, 6, 216, 188, 221, 60, 0, 6, 216, 189, 221, 60, 0, 6, 216, 190, 221, 60, 0, 6, 216, 191, 221, 60, 0, 6, 216, 192, 221, 60, 0, 6, 216, 193, 221, 60, 0, 6, 216, 194, 221, 60, 0, 6, 216, 195, 221, 60, 0, 6, 216, 196, 221, 60, 0, 6, 216, 197, 221, 60, 0, 6, 216, 198, 221, 60, 0, 6, 216, 199, 221, 60, 0, 6, 216, 200, 221, 60, 0, 6, 216, 201, 221, 60, 0, 6, 216, 202, 221, 60, 0, 6, 216, 203, 221, 60, 0, 6, 216, 204, 221, 60, 0, 6, 216, 205, 221, 60, 0, 6, 216, 206, 221, 60, 0, 6, 216, 207, 221, 60, 0, 6, 216, 208, 221, 60, 0, 6, 216, 225, 221, 60, 0, 6, 216, 227, 221, 60, 0, 6, 216, 209, 221, 60, 0, 6, 216, 210, 221, 60, 0, 6, 216, 211, 221, 60, 0, 6, 216, 212, 221, 60, 0, 6, 216, 213, 221, 60, 0, 6, 216, 214, 221, 60, 0, 6, 216, 215, 221, 60, 0, 6, 216, 218, 221, 60, 0, 6, 216, 228, 221, 60, 0, 6, 216, 219, 221, 60, 0, 6, 216, 220, 221, 60, 0, 6, 216, 221, 221, 60, 0, 6, 216, 224, 221, 60, 0, 6, 216, 0, 220, 60, 0, 6, 216, 1, 220, 60, 0, 6, 216, 2, 220, 60, 0, 6, 216, 3, 220, 60, 0, 6, 216, 4, 220, 60, 0, 6, 216, 5, 220, 60, 0, 6, 216, 6, 220, 60, 0, 6, 216, 7, 220, 60, 0, 6, 216, 8, 220, 60, 0, 6, 216, 9, 220, 60, 0, 6, 216, 10, 220, 60, 0, 6, 216, 11, 220, 60, 0, 6, 216, 12, 220, 60, 0, 6, 216, 13, 220, 60, 0, 6, 216, 14, 220, 60, 0, 6, 216, 15, 220, 60, 0, 6, 216, 16, 220, 60, 0, 6, 216, 17, 220, 60, 0, 6, 216, 18, 220, 60, 0, 6, 216, 19, 220, 60, 0, 6, 216, 20, 220, 60, 0, 6, 216, 21, 220, 60, 0, 6, 216, 22, 220, 60, 0, 6, 216, 23, 220, 60, 0, 6, 216, 24, 220, 60, 0, 6, 216, 25, 220, 60, 0, 6, 216, 26, 220, 60, 0, 6, 216, 27, 220, 60, 0, 6, 216, 28, 220, 60, 0, 6, 216, 29, 220, 60, 0, 6, 216, 30, 220, 60, 0, 6, 216, 31, 220, 60, 0, 6, 216, 32, 220, 60, 0, 6, 216, 33, 220, 60, 0, 6, 216, 34, 220, 60, 0, 6, 216, 35, 220, 60, 0, 6, 216, 36, 220, 60, 0, 6, 216, 37, 220, 60, 0, 6, 216, 38, 220, 60, 0, 6, 216, 39, 220, 60, 0, 6, 216, 40, 220, 60, 0, 6, 216, 41, 220, 60, 0, 6, 216, 42, 220, 60, 0, 6, 216, 43, 220, 60, 0, 6, 216, 44, 220, 60, 0, 6, 216, 45, 220, 60, 0, 6, 216, 46, 220, 60, 0, 6, 216, 47, 220, 60, 0, 6, 216, 48, 220, 60, 0, 6, 216, 49, 220, 60, 0, 6, 216, 50, 220, 60, 0, 6, 216, 51, 220, 60, 0, 6, 216, 52, 220, 60, 0, 6, 216, 53, 220, 60, 0, 6, 216, 54, 220, 60, 0, 6, 216, 57, 220, 60, 0, 5, 216, 0, 223, 60, 0, 5, 216, 1, 223, 60, 0, 5, 216, 2, 223, 60, 0, 5, 216, 3, 223, 60, 0, 5, 216, 4, 223, 60, 0, 60, 0, 5, 216, 5, 223, 60, 0, 5, 216, 6, 223, 60, 0, 5, 216, 7, 223, 60, 0, 5, 216, 8, 223, 60, 0, 60, 0, 5, 216, 26, 223, 60, 0, 5, 216, 9, 223, 60, 0, 5, 216, 10, 223, 60, 0, 5, 216, 11, 223, 60, 0, 5, 216, 12, 223, 60, 0, 5, 216, 13, 223, 60, 0, 5, 216, 14, 223, 60, 0, 5, 216, 15, 223, 60, 0, 5, 216, 16, 223, 60, 0, 5, 216, 17, 223, 60, 0, 5, 216, 18, 223, 60, 0, 5, 216, 19, 223, 60, 0, 5, 216, 20, 223, 60, 0, 5, 216, 21, 223, 60, 0, 60, 0, 5, 216, 22, 223, 60, 0, 5, 216, 23, 223, 60, 0, 5, 216, 24, 223, 60, 0, 5, 216, 25, 223, 60, 0, 5, 216, 32, 223, 60, 0, 5, 216, 33, 223, 60, 0, 5, 216, 34, 223, 60, 0, 5, 216, 35, 223, 60, 0, 5, 216, 36, 223, 60, 0, 5, 216, 37, 223, 60, 0, 5, 216, 38, 223, 60, 0, 5, 216, 39, 223, 60, 0, 5, 216, 40, 223, 60, 0, 5, 216, 41, 223, 60, 0, 5, 216, 42, 223, 60, 0, 5, 216, 43, 223, 60, 0, 5, 216, 29, 223, 60, 0, 5, 216, 30, 223, 60, 0, 5, 216, 31, 223, 60, 0, 7, 216, 0, 221, 60, 0, 7, 216, 1, 221, 60, 0, 7, 216, 2, 221, 60, 0, 7, 216, 3, 221, 60, 0, 7, 216, 4, 221, 60, 0, 7, 216, 5, 221, 60, 0, 7, 216, 6, 221, 60, 0, 7, 216, 8, 221, 60, 0, 7, 216, 9, 221, 60, 0, 7, 216, 11, 221, 60, 0, 7, 216, 12, 221, 60, 0, 7, 216, 13, 221, 60, 0, 7, 216, 14, 221, 60, 0, 7, 216, 15, 221, 60, 0, 7, 216, 16, 221, 60, 0, 7, 216, 17, 221, 60, 0, 7, 216, 18, 221, 60, 0, 7, 216, 19, 221, 60, 0, 7, 216, 20, 221, 60, 0, 7, 216, 21, 221, 60, 0, 7, 216, 22, 221, 60, 0, 7, 216, 23, 221, 60, 0, 7, 216, 24, 221, 60, 0, 7, 216, 25, 221, 60, 0, 7, 216, 26, 221, 60, 0, 7, 216, 27, 221, 60, 0, 7, 216, 28, 221, 60, 0, 7, 216, 29, 221, 60, 0, 7, 216, 30, 221, 60, 0, 7, 216, 31, 221, 60, 0, 7, 216, 32, 221, 60, 0, 7, 216, 33, 221, 60, 0, 7, 216, 34, 221, 60, 0, 7, 216, 35, 221, 60, 0, 7, 216, 36, 221, 60, 0, 7, 216, 37, 221, 60, 0, 7, 216, 38, 221, 60, 0, 60, 0, 60, 0, 7, 216, 70, 221, 60, 0, 60, 0, 60, 0, 7, 216, 71, 221, 60, 0, 7, 216, 39, 221, 60, 0, 7, 216, 40, 221, 60, 0, 7, 216, 41, 221, 60, 0, 7, 216, 42, 221, 60, 0, 7, 216, 43, 221, 60, 0, 7, 216, 44, 221, 60, 0, 7, 216, 45, 221, 60, 0, 7, 216, 46, 221, 60, 0, 7, 216, 47, 221, 60, 0, 7, 216, 48, 221, 60, 0, 7, 216, 49, 221, 60, 0, 7, 216, 50, 221, 60, 0, 7, 216, 51, 221, 60, 0, 7, 216, 52, 221, 60, 0, 7, 216, 53, 221, 60, 0, 7, 216, 54, 221, 60, 0, 7, 216, 58, 221, 60, 0, 7, 216, 60, 221, 60, 0, 7, 216, 61, 221, 60, 0, 7, 216, 63, 221, 60, 0, 7, 216, 68, 221, 60, 0, 7, 216, 69, 221, 60, 0, 7, 216, 152, 221, 60, 0, 7, 216, 96, 221, 60, 0, 7, 216, 97, 221, 60, 0, 7, 216, 98, 221, 60, 0, 7, 216, 99, 221, 60, 0, 7, 216, 100, 221, 60, 0, 7, 216, 101, 221, 60, 0, 7, 216, 103, 221, 60, 0, 7, 216, 104, 221, 60, 0, 7, 216, 106, 221, 60, 0, 7, 216, 107, 221, 60, 0, 7, 216, 108, 221, 60, 0, 7, 216, 109, 221, 60, 0, 7, 216, 110, 221, 60, 0, 7, 216, 111, 221, 60, 0, 7, 216, 112, 221, 60, 0, 7, 216, 113, 221, 60, 0, 7, 216, 114, 221, 60, 0, 7, 216, 115, 221, 60, 0, 7, 216, 116, 221, 60, 0, 7, 216, 117, 221, 60, 0, 7, 216, 118, 221, 60, 0, 7, 216, 119, 221, 60, 0, 7, 216, 120, 221, 60, 0, 7, 216, 121, 221, 60, 0, 7, 216, 122, 221, 60, 0, 7, 216, 123, 221, 60, 0, 7, 216, 124, 221, 60, 0, 7, 216, 125, 221, 60, 0, 7, 216, 126, 221, 60, 0, 7, 216, 127, 221, 60, 0, 7, 216, 128, 221, 60, 0, 7, 216, 129, 221, 60, 0, 7, 216, 130, 221, 60, 0, 7, 216, 131, 221, 60, 0, 7, 216, 132, 221, 60, 0, 7, 216, 133, 221, 60, 0, 7, 216, 134, 221, 60, 0, 7, 216, 135, 221, 60, 0, 7, 216, 136, 221, 60, 0, 7, 216, 137, 221, 60, 0, 7, 216, 138, 221, 60, 0, 7, 216, 139, 221, 60, 0, 7, 216, 140, 221, 60, 0, 7, 216, 141, 221, 60, 0, 7, 216, 142, 221, 60, 0, 7, 216, 144, 221, 60, 0, 7, 216, 145, 221, 60, 0, 7, 216, 147, 221, 60, 0, 7, 216, 148, 221, 60, 0, 7, 216, 151, 221, 60, 0, 131, 27, 60, 0, 60, 0, 60, 0, 186, 27, 60, 0, 132, 27, 60, 0, 133, 27, 60, 0, 134, 27, 60, 0, 135, 27, 60, 0, 136, 27, 60, 0, 137, 27, 60, 0, 138, 27, 60, 0, 60, 0, 60, 0, 190, 27, 60, 0, 174, 27, 60, 0, 139, 27, 60, 0, 140, 27, 60, 0, 141, 27, 60, 0, 142, 27, 60, 0, 143, 27, 60, 0, 144, 27, 60, 0, 145, 27, 60, 0, 146, 27, 60, 0, 147, 27, 60, 0, 148, 27, 60, 0, 149, 27, 60, 0, 150, 27, 60, 0, 151, 27, 60, 0, 152, 27, 60, 0, 189, 27, 60, 0, 153, 27, 60, 0, 60, 0, 60, 0, 191, 27, 60, 0, 172, 27, 60, 0, 154, 27, 60, 0, 161, 27, 60, 0, 155, 27, 60, 0, 162, 27, 60, 0, 187, 27, 60, 0, 156, 27, 60, 0, 163, 27, 60, 0, 188, 27, 60, 0, 157, 27, 60, 0, 173, 27, 60, 0, 158, 27, 60, 0, 159, 27, 60, 0, 175, 27, 60, 0, 160, 27, 60, 0, 164, 27, 60, 0, 165, 27, 60, 0, 166, 27, 60, 0, 167, 27, 60, 0, 168, 27, 60, 0, 169, 27, 60, 0, 170, 27, 60, 0, 171, 27, 60, 0, 4, 216, 5, 220, 60, 0, 4, 216, 6, 220, 60, 0, 4, 216, 7, 220, 60, 0, 4, 216, 8, 220, 60, 0, 4, 216, 9, 220, 60, 0, 4, 216, 10, 220, 60, 0, 4, 216, 11, 220, 60, 0, 4, 216, 12, 220, 60, 0, 4, 216, 13, 220, 60, 0, 4, 216, 14, 220, 60, 0, 4, 216, 15, 220, 60, 0, 4, 216, 16, 220, 60, 0, 4, 216, 17, 220, 60, 0, 4, 216, 18, 220, 60, 0, 4, 216, 19, 220, 60, 0, 4, 216, 20, 220, 60, 0, 4, 216, 21, 220, 60, 0, 4, 216, 22, 220, 60, 0, 4, 216, 23, 220, 60, 0, 4, 216, 24, 220, 60, 0, 4, 216, 25, 220, 60, 0, 4, 216, 26, 220, 60, 0, 4, 216, 27, 220, 60, 0, 4, 216, 28, 220, 60, 0, 4, 216, 29, 220, 60, 0, 4, 216, 30, 220, 60, 0, 4, 216, 31, 220, 60, 0, 4, 216, 32, 220, 60, 0, 4, 216, 33, 220, 60, 0, 4, 216, 34, 220, 60, 0, 4, 216, 35, 220, 60, 0, 4, 216, 36, 220, 60, 0, 4, 216, 37, 220, 60, 0, 4, 216, 38, 220, 60, 0, 4, 216, 39, 220, 60, 0, 4, 216, 40, 220, 60, 0, 4, 216, 41, 220, 60, 0, 4, 216, 42, 220, 60, 0, 4, 216, 43, 220, 60, 0, 4, 216, 44, 220, 60, 0, 4, 216, 45, 220, 60, 0, 4, 216, 46, 220, 60, 0, 4, 216, 47, 220, 60, 0, 4, 216, 48, 220, 60, 0, 4, 216, 49, 220, 60, 0, 4, 216, 50, 220, 60, 0, 4, 216, 51, 220, 60, 0, 4, 216, 3, 220, 60, 0, 4, 216, 4, 220, 60, 0, 4, 216, 52, 220, 60, 0, 4, 216, 53, 220, 60, 0, 4, 216, 54, 220, 60, 0, 4, 216, 55, 220, 60, 0, 4, 216, 56, 220, 60, 0, 4, 216, 57, 220, 60, 0, 4, 216, 58, 220, 60, 0, 4, 216, 59, 220, 60, 0, 4, 216, 60, 220, 60, 0, 4, 216, 61, 220, 60, 0, 4, 216, 62, 220, 60, 0, 4, 216, 63, 220, 60, 0, 4, 216, 64, 220, 60, 0, 4, 216, 65, 220, 60, 0, 4, 216, 66, 220, 60, 0, 4, 216, 67, 220, 60, 0, 4, 216, 68, 220, 60, 0, 4, 216, 69, 220, 60, 0, 4, 216, 70, 220, 60, 0, 4, 216, 127, 220, 60, 0, 14, 32, 2, 216, 0, 222, 14, 32, 60, 0, 2, 216, 1, 222, 60, 0, 2, 216, 2, 222, 60, 0, 2, 216, 3, 222, 60, 0, 2, 216, 5, 222, 60, 0, 2, 216, 6, 222, 60, 0, 2, 216, 12, 222, 60, 0, 14, 32, 2, 216, 16, 222, 14, 32, 60, 0, 14, 32, 2, 216, 50, 222, 14, 32, 60, 0, 14, 32, 2, 216, 17, 222, 14, 32, 60, 0, 14, 32, 2, 216, 18, 222, 14, 32, 60, 0, 14, 32, 2, 216, 19, 222, 14, 32, 60, 0, 14, 32, 2, 216, 21, 222, 14, 32, 60, 0, 14, 32, 2, 216, 22, 222, 14, 32, 60, 0, 14, 32, 2, 216, 23, 222, 14, 32, 60, 0, 14, 32, 2, 216, 25, 222, 14, 32, 60, 0, 14, 32, 2, 216, 26, 222, 14, 32, 60, 0, 14, 32, 2, 216, 52, 222, 14, 32, 60, 0, 14, 32, 2, 216, 27, 222, 14, 32, 60, 0, 14, 32, 2, 216, 51, 222, 14, 32, 60, 0, 14, 32, 2, 216, 28, 222, 14, 32, 60, 0, 14, 32, 2, 216, 29, 222, 14, 32, 60, 0, 14, 32, 2, 216, 30, 222, 14, 32, 60, 0, 14, 32, 2, 216, 31, 222, 14, 32, 60, 0, 14, 32, 2, 216, 32, 222, 14, 32, 60, 0, 14, 32, 2, 216, 33, 222, 14, 32, 60, 0, 14, 32, 2, 216, 34, 222, 14, 32, 60, 0, 14, 32, 2, 216, 35, 222, 14, 32, 60, 0, 14, 32, 2, 216, 36, 222, 14, 32, 60, 0, 14, 32, 2, 216, 37, 222, 14, 32, 60, 0, 14, 32, 2, 216, 38, 222, 14, 32, 60, 0, 14, 32, 2, 216, 39, 222, 14, 32, 60, 0, 14, 32, 2, 216, 40, 222, 14, 32, 60, 0, 14, 32, 2, 216, 41, 222, 14, 32, 60, 0, 14, 32, 2, 216, 42, 222, 14, 32, 60, 0, 14, 32, 2, 216, 43, 222, 14, 32, 60, 0, 14, 32, 2, 216, 44, 222, 14, 32, 60, 0, 14, 32, 2, 216, 53, 222, 14, 32, 60, 0, 14, 32, 2, 216, 45, 222, 14, 32, 60, 0, 14, 32, 2, 216, 46, 222, 14, 32, 60, 0, 14, 32, 2, 216, 47, 222, 14, 32, 60, 0, 14, 32, 2, 216, 48, 222, 14, 32, 60, 0, 14, 32, 2, 216, 49, 222, 14, 32, 60, 0, 2, 216, 63, 222, 60, 0, 7, 216, 0, 220, 60, 0, 7, 216, 1, 220, 60, 0, 7, 216, 2, 220, 60, 0, 7, 216, 3, 220, 60, 0, 7, 216, 4, 220, 60, 0, 7, 216, 5, 220, 60, 0, 7, 216, 6, 220, 60, 0, 7, 216, 7, 220, 60, 0, 7, 216, 8, 220, 60, 0, 7, 216, 10, 220, 60, 0, 7, 216, 11, 220, 60, 0, 7, 216, 12, 220, 60, 0, 7, 216, 13, 220, 60, 0, 7, 216, 14, 220, 60, 0, 7, 216, 15, 220, 60, 0, 7, 216, 16, 220, 60, 0, 7, 216, 17, 220, 60, 0, 7, 216, 18, 220, 60, 0, 7, 216, 19, 220, 60, 0, 7, 216, 20, 220, 60, 0, 7, 216, 21, 220, 60, 0, 7, 216, 22, 220, 60, 0, 7, 216, 23, 220, 60, 0, 7, 216, 24, 220, 60, 0, 7, 216, 25, 220, 60, 0, 7, 216, 26, 220, 60, 0, 7, 216, 27, 220, 60, 0, 7, 216, 28, 220, 60, 0, 7, 216, 29, 220, 60, 0, 7, 216, 30, 220, 60, 0, 7, 216, 31, 220, 60, 0, 7, 216, 32, 220, 60, 0, 7, 216, 33, 220, 60, 0, 7, 216, 34, 220, 60, 0, 7, 216, 35, 220, 60, 0, 7, 216, 36, 220, 60, 0, 7, 216, 37, 220, 60, 0, 7, 216, 38, 220, 60, 0, 7, 216, 39, 220, 60, 0, 7, 216, 40, 220, 60, 0, 7, 216, 41, 220, 60, 0, 7, 216, 42, 220, 60, 0, 7, 216, 43, 220, 60, 0, 7, 216, 44, 220, 60, 0, 7, 216, 45, 220, 60, 0, 7, 216, 46, 220, 60, 0, 7, 216, 64, 220, 60, 0, 7, 216, 47, 220, 60, 0, 7, 216, 48, 220, 60, 0, 7, 216, 49, 220, 60, 0, 7, 216, 50, 220, 60, 0, 7, 216, 51, 220, 60, 0, 7, 216, 52, 220, 60, 0, 7, 216, 53, 220, 60, 0, 7, 216, 54, 220, 60, 0, 7, 216, 56, 220, 60, 0, 7, 216, 57, 220, 60, 0, 7, 216, 58, 220, 60, 0, 7, 216, 59, 220, 60, 0, 7, 216, 63, 220, 60, 0, 1, 14, 60, 0, 2, 14, 60, 0, 3, 14, 60, 0, 4, 14, 60, 0, 5, 14, 60, 0, 6, 14, 60, 0, 7, 14, 60, 0, 8, 14, 60, 0, 9, 14, 60, 0, 10, 14, 60, 0, 11, 14, 60, 0, 12, 14, 60, 0, 13, 14, 60, 0, 14, 14, 60, 0, 15, 14, 60, 0, 16, 14, 60, 0, 17, 14, 60, 0, 18, 14, 60, 0, 19, 14, 60, 0, 20, 14, 60, 0, 21, 14, 60, 0, 22, 14, 60, 0, 23, 14, 60, 0, 24, 14, 60, 0, 25, 14, 60, 0, 26, 14, 60, 0, 27, 14, 60, 0, 28, 14, 60, 0, 29, 14, 60, 0, 30, 14, 60, 0, 31, 14, 60, 0, 32, 14, 60, 0, 33, 14, 60, 0, 34, 14, 60, 0, 35, 14, 60, 0, 36, 14, 60, 0, 37, 14, 60, 0, 38, 14, 60, 0, 39, 14, 60, 0, 40, 14, 60, 0, 41, 14, 60, 0, 42, 14, 60, 0, 43, 14, 60, 0, 44, 14, 60, 0, 45, 14, 60, 0, 46, 14, 60, 0, 47, 14, 60, 0, 48, 14, 60, 0, 49, 14, 60, 0, 50, 14, 60, 0, 51, 14, 61, 0, 77, 14, 50, 14, 60, 0, 52, 14, 60, 0, 53, 14, 60, 0, 54, 14, 60, 0, 55, 14, 60, 0, 56, 14, 60, 0, 57, 14, 60, 0, 58, 14, 60, 0, 64, 14, 60, 0, 65, 14, 60, 0, 66, 14, 60, 0, 67, 14, 60, 0, 68, 14, 60, 0, 69, 14, 60, 0, 222, 14, 60, 0, 129, 14, 60, 0, 130, 14, 60, 0, 132, 14, 60, 0, 134, 14, 60, 0, 135, 14, 60, 0, 136, 14, 60, 0, 137, 14, 60, 0, 170, 14, 60, 0, 138, 14, 60, 0, 140, 14, 60, 0, 142, 14, 60, 0, 223, 14, 60, 0, 141, 14, 60, 0, 143, 14, 60, 0, 144, 14, 60, 0, 145, 14, 60, 0, 146, 14, 60, 0, 147, 14, 60, 0, 148, 14, 60, 0, 149, 14, 60, 0, 150, 14, 60, 0, 151, 14, 60, 0, 152, 14, 60, 0, 153, 14, 60, 0, 154, 14, 60, 0, 155, 14, 60, 0, 156, 14, 60, 0, 157, 14, 60, 0, 158, 14, 60, 0, 159, 14, 60, 0, 160, 14, 60, 0, 161, 14, 60, 0, 162, 14, 60, 0, 163, 14, 60, 0, 165, 14, 60, 0, 167, 14, 60, 0, 168, 14, 60, 0, 169, 14, 60, 0, 171, 14, 60, 0, 60, 0, 60, 0, 220, 14, 47, 0, 153, 14, 60, 0, 60, 0, 60, 0, 221, 14, 47, 0, 161, 14, 60, 0, 172, 14, 60, 0, 173, 14, 60, 0, 174, 14, 60, 0, 175, 14, 60, 0, 176, 14, 60, 0, 177, 14, 60, 0, 178, 14, 60, 0, 179, 14, 61, 0, 205, 14, 178, 14, 60, 0, 180, 14, 60, 0, 181, 14, 60, 0, 182, 14, 60, 0, 183, 14, 60, 0, 184, 14, 60, 0, 185, 14, 60, 0, 186, 14, 60, 0, 187, 14, 60, 0, 188, 14, 60, 0, 189, 14, 60, 0, 192, 14, 60, 0, 193, 14, 60, 0, 194, 14, 60, 0, 195, 14, 60, 0, 196, 14, 60, 0, 128, 170, 60, 0, 129, 170, 60, 0, 130, 170, 60, 0, 131, 170, 60, 0, 132, 170, 60, 0, 133, 170, 60, 0, 134, 170, 60, 0, 135, 170, 60, 0, 136, 170, 60, 0, 137, 170, 60, 0, 138, 170, 60, 0, 139, 170, 60, 0, 140, 170, 60, 0, 141, 170, 60, 0, 142, 170, 60, 0, 143, 170, 60, 0, 144, 170, 60, 0, 145, 170, 60, 0, 146, 170, 60, 0, 147, 170, 60, 0, 148, 170, 60, 0, 149, 170, 60, 0, 150, 170, 60, 0, 151, 170, 60, 0, 152, 170, 60, 0, 153, 170, 60, 0, 154, 170, 60, 0, 155, 170, 60, 0, 156, 170, 60, 0, 157, 170, 60, 0, 158, 170, 60, 0, 159, 170, 60, 0, 160, 170, 60, 0, 161, 170, 60, 0, 162, 170, 60, 0, 163, 170, 60, 0, 164, 170, 60, 0, 165, 170, 60, 0, 166, 170, 60, 0, 167, 170, 60, 0, 168, 170, 60, 0, 169, 170, 60, 0, 170, 170, 60, 0, 171, 170, 60, 0, 172, 170, 60, 0, 173, 170, 60, 0, 174, 170, 60, 0, 175, 170, 60, 0, 176, 170, 60, 0, 177, 170, 60, 0, 178, 170, 60, 0, 179, 170, 60, 0, 180, 170, 60, 0, 181, 170, 60, 0, 182, 170, 60, 0, 183, 170, 60, 0, 184, 170, 60, 0, 185, 170, 60, 0, 186, 170, 60, 0, 187, 170, 60, 0, 188, 170, 60, 0, 189, 170, 60, 0, 190, 170, 60, 0, 192, 170, 60, 0, 194, 170, 60, 0, 219, 170, 60, 0, 220, 170, 60, 0, 64, 15, 60, 0, 144, 15, 60, 0, 107, 15, 60, 0, 65, 15, 60, 0, 145, 15, 60, 0, 66, 15, 60, 0, 146, 15, 60, 0, 68, 15, 60, 0, 148, 15, 60, 0, 69, 15, 60, 0, 149, 15, 60, 0, 70, 15, 60, 0, 150, 15, 60, 0, 71, 15, 60, 0, 151, 15, 60, 0, 73, 15, 60, 0, 153, 15, 60, 0, 74, 15, 60, 0, 154, 15, 60, 0, 75, 15, 60, 0, 155, 15, 60, 0, 76, 15, 60, 0, 156, 15, 60, 0, 78, 15, 60, 0, 158, 15, 60, 0, 79, 15, 60, 0, 159, 15, 60, 0, 80, 15, 60, 0, 160, 15, 60, 0, 81, 15, 60, 0, 161, 15, 60, 0, 83, 15, 60, 0, 163, 15, 60, 0, 84, 15, 60, 0, 164, 15, 60, 0, 85, 15, 60, 0, 165, 15, 60, 0, 86, 15, 60, 0, 166, 15, 60, 0, 88, 15, 60, 0, 168, 15, 60, 0, 89, 15, 60, 0, 169, 15, 60, 0, 90, 15, 60, 0, 170, 15, 60, 0, 91, 15, 60, 0, 171, 15, 60, 0, 93, 15, 60, 0, 173, 15, 60, 0, 60, 0, 186, 15, 60, 0, 94, 15, 60, 0, 174, 15, 60, 0, 95, 15, 60, 0, 175, 15, 60, 0, 96, 15, 60, 0, 176, 15, 60, 0, 97, 15, 60, 0, 177, 15, 60, 0, 60, 0, 187, 15, 60, 0, 98, 15, 60, 0, 60, 0, 106, 15, 60, 0, 178, 15, 60, 0, 60, 0, 60, 0, 178, 15, 113, 15, 47, 0, 113, 15, 60, 0, 60, 0, 60, 0, 178, 15, 113, 15, 114, 15, 47, 0, 113, 15, 114, 15, 61, 0, 178, 15, 113, 15, 114, 15, 47, 0, 113, 15, 114, 15, 60, 0, 60, 0, 60, 0, 178, 15, 113, 15, 116, 15, 47, 0, 113, 15, 116, 15, 61, 0, 178, 15, 113, 15, 116, 15, 47, 0, 113, 15, 116, 15, 60, 0, 60, 0, 188, 15, 60, 0, 108, 15, 60, 0, 99, 15, 60, 0, 179, 15, 60, 0, 60, 0, 60, 0, 179, 15, 113, 15, 47, 0, 113, 15, 60, 0, 60, 0, 60, 0, 179, 15, 113, 15, 114, 15, 47, 0, 113, 15, 114, 15, 61, 0, 179, 15, 113, 15, 114, 15, 47, 0, 113, 15, 114, 15, 60, 0, 60, 0, 60, 0, 179, 15, 113, 15, 116, 15, 47, 0, 113, 15, 116, 15, 61, 0, 179, 15, 113, 15, 116, 15, 47, 0, 113, 15, 116, 15, 60, 0, 100, 15, 60, 0, 180, 15, 60, 0, 101, 15, 60, 0, 181, 15, 60, 0, 102, 15, 60, 0, 182, 15, 60, 0, 103, 15, 60, 0, 183, 15, 60, 0, 104, 15, 60, 0, 60, 0, 60, 0, 0, 15, 47, 0, 124, 15, 7, 216, 149, 221, 60, 0, 184, 15, 60, 0, 136, 15, 60, 0, 141, 15, 60, 0, 137, 15, 60, 0, 142, 15, 60, 0, 140, 15, 60, 0, 143, 15, 60, 0, 138, 15, 60, 0, 139, 15, 60, 0, 113, 15, 60, 0, 114, 15, 60, 0, 113, 15, 114, 15, 60, 0, 128, 15, 60, 0, 113, 15, 128, 15, 60, 0, 116, 15, 60, 0, 113, 15, 116, 15, 60, 0, 178, 15, 128, 15, 60, 0, 119, 15, 61, 0, 178, 15, 113, 15, 128, 15, 61, 0, 178, 15, 113, 15, 128, 15, 60, 0, 179, 15, 128, 15, 60, 0, 121, 15, 61, 0, 179, 15, 113, 15, 128, 15, 61, 0, 179, 15, 113, 15, 128, 15, 60, 0, 122, 15, 60, 0, 123, 15, 60, 0, 124, 15, 60, 0, 125, 15, 60, 0, 132, 15, 60, 0, 6, 216, 11, 222, 60, 0, 6, 216, 50, 222, 60, 0, 6, 216, 12, 222, 60, 0, 6, 216, 13, 222, 60, 0, 6, 216, 14, 222, 60, 0, 6, 216, 15, 222, 60, 0, 6, 216, 16, 222, 60, 0, 6, 216, 17, 222, 60, 0, 6, 216, 18, 222, 60, 0, 6, 216, 19, 222, 60, 0, 6, 216, 20, 222, 60, 0, 6, 216, 21, 222, 60, 0, 6, 216, 22, 222, 60, 0, 6, 216, 23, 222, 60, 0, 6, 216, 24, 222, 60, 0, 6, 216, 25, 222, 60, 0, 6, 216, 26, 222, 60, 0, 6, 216, 27, 222, 60, 0, 6, 216, 28, 222, 60, 0, 6, 216, 29, 222, 60, 0, 6, 216, 30, 222, 60, 0, 6, 216, 31, 222, 60, 0, 6, 216, 32, 222, 60, 0, 6, 216, 33, 222, 60, 0, 6, 216, 34, 222, 60, 0, 6, 216, 35, 222, 60, 0, 6, 216, 36, 222, 60, 0, 6, 216, 37, 222, 60, 0, 6, 216, 38, 222, 60, 0, 6, 216, 39, 222, 60, 0, 6, 216, 40, 222, 60, 0, 6, 216, 41, 222, 60, 0, 6, 216, 42, 222, 60, 0, 60, 0, 60, 0, 6, 216, 59, 222, 60, 0, 6, 216, 43, 222, 60, 0, 60, 0, 60, 0, 6, 216, 58, 222, 60, 0, 60, 0, 60, 0, 6, 216, 60, 222, 60, 0, 6, 216, 44, 222, 60, 0, 60, 0, 60, 0, 6, 216, 61, 222, 60, 0, 6, 216, 45, 222, 60, 0, 60, 0, 60, 0, 6, 216, 62, 222, 60, 0, 6, 216, 46, 222, 60, 0, 6, 216, 47, 222, 60, 0, 6, 216, 48, 222, 60, 0, 6, 216, 49, 222, 60, 0, 6, 216, 0, 222, 60, 0, 6, 216, 1, 222, 60, 0, 6, 216, 2, 222, 60, 0, 6, 216, 3, 222, 60, 0, 6, 216, 4, 222, 60, 0, 6, 216, 5, 222, 60, 0, 6, 216, 6, 222, 60, 0, 6, 216, 7, 222, 60, 0, 6, 216, 8, 222, 60, 0, 6, 216, 9, 222, 60, 0, 6, 216, 10, 222, 60, 0, 6, 216, 52, 222, 60, 0, 6, 216, 71, 222, 60, 0, 6, 216, 80, 222, 60, 0, 6, 216, 81, 222, 60, 0, 6, 216, 82, 222, 60, 0, 6, 216, 83, 222, 60, 0, 6, 216, 89, 222, 60, 0, 6, 216, 90, 222, 60, 0, 6, 216, 84, 222, 60, 0, 6, 216, 86, 222, 60, 0, 6, 216, 85, 222, 60, 0, 6, 216, 87, 222, 60, 0, 6, 216, 88, 222, 60, 0, 6, 216, 91, 222, 60, 0, 6, 216, 92, 222, 60, 0, 60, 0, 60, 0, 6, 216, 138, 222, 60, 0, 6, 216, 131, 222, 60, 0, 6, 216, 93, 222, 60, 0, 60, 0, 60, 0, 6, 216, 139, 222, 60, 0, 6, 216, 94, 222, 60, 0, 6, 216, 95, 222, 60, 0, 6, 216, 96, 222, 60, 0, 60, 0, 60, 0, 6, 216, 140, 222, 60, 0, 6, 216, 97, 222, 60, 0, 6, 216, 98, 222, 60, 0, 6, 216, 99, 222, 60, 0, 6, 216, 100, 222, 60, 0, 6, 216, 101, 222, 60, 0, 6, 216, 102, 222, 60, 0, 6, 216, 103, 222, 60, 0, 6, 216, 104, 222, 60, 0, 6, 216, 105, 222, 60, 0, 6, 216, 106, 222, 60, 0, 6, 216, 107, 222, 60, 0, 60, 0, 60, 0, 6, 216, 141, 222, 60, 0, 6, 216, 108, 222, 60, 0, 6, 216, 109, 222, 60, 0, 6, 216, 110, 222, 60, 0, 6, 216, 111, 222, 60, 0, 60, 0, 60, 0, 6, 216, 142, 222, 60, 0, 6, 216, 112, 222, 60, 0, 60, 0, 60, 0, 6, 216, 143, 222, 60, 0, 6, 216, 113, 222, 60, 0, 6, 216, 114, 222, 60, 0, 6, 216, 115, 222, 60, 0, 6, 216, 116, 222, 60, 0, 60, 0, 60, 0, 6, 216, 144, 222, 60, 0, 6, 216, 117, 222, 60, 0, 6, 216, 118, 222, 60, 0, 6, 216, 119, 222, 60, 0, 6, 216, 120, 222, 60, 0, 6, 216, 121, 222, 60, 0, 6, 216, 122, 222, 60, 0, 60, 0, 60, 0, 6, 216, 149, 222, 60, 0, 6, 216, 123, 222, 60, 0, 6, 216, 124, 222, 60, 0, 60, 0, 60, 0, 6, 216, 134, 222, 60, 0, 60, 0, 60, 0, 6, 216, 145, 222, 60, 0, 6, 216, 125, 222, 60, 0, 60, 0, 60, 0, 6, 216, 135, 222, 60, 0, 60, 0, 60, 0, 6, 216, 146, 222, 60, 0, 6, 216, 126, 222, 60, 0, 6, 216, 127, 222, 60, 0, 60, 0, 60, 0, 6, 216, 136, 222, 60, 0, 60, 0, 60, 0, 6, 216, 147, 222, 60, 0, 6, 216, 128, 222, 60, 0, 6, 216, 129, 222, 60, 0, 60, 0, 60, 0, 6, 216, 137, 222, 60, 0, 60, 0, 60, 0, 6, 216, 148, 222, 60, 0, 6, 216, 130, 222, 60, 0, 6, 216, 132, 222, 60, 0, 6, 216, 133, 222, 60, 0, 6, 216, 153, 222, 60, 0, 6, 216, 157, 222, 60, 0, 7, 216, 114, 220, 60, 0, 7, 216, 146, 220, 60, 0, 7, 216, 115, 220, 60, 0, 7, 216, 147, 220, 60, 0, 7, 216, 116, 220, 60, 0, 7, 216, 148, 220, 60, 0, 7, 216, 117, 220, 60, 0, 7, 216, 149, 220, 60, 0, 7, 216, 118, 220, 60, 0, 7, 216, 150, 220, 60, 0, 7, 216, 119, 220, 60, 0, 7, 216, 151, 220, 60, 0, 7, 216, 120, 220, 60, 0, 7, 216, 152, 220, 60, 0, 7, 216, 121, 220, 60, 0, 7, 216, 153, 220, 60, 0, 7, 216, 122, 220, 60, 0, 7, 216, 154, 220, 60, 0, 7, 216, 123, 220, 60, 0, 7, 216, 155, 220, 60, 0, 7, 216, 124, 220, 60, 0, 7, 216, 156, 220, 60, 0, 7, 216, 125, 220, 60, 0, 7, 216, 157, 220, 60, 0, 7, 216, 126, 220, 60, 0, 7, 216, 158, 220, 60, 0, 7, 216, 127, 220, 60, 0, 7, 216, 159, 220, 60, 0, 7, 216, 128, 220, 60, 0, 7, 216, 160, 220, 60, 0, 7, 216, 129, 220, 60, 0, 7, 216, 161, 220, 60, 0, 7, 216, 130, 220, 60, 0, 7, 216, 162, 220, 60, 0, 7, 216, 131, 220, 60, 0, 7, 216, 163, 220, 60, 0, 7, 216, 132, 220, 60, 0, 7, 216, 164, 220, 60, 0, 7, 216, 133, 220, 60, 0, 7, 216, 165, 220, 60, 0, 7, 216, 134, 220, 60, 0, 7, 216, 166, 220, 60, 0, 7, 216, 135, 220, 60, 0, 7, 216, 167, 220, 60, 0, 7, 216, 136, 220, 60, 0, 7, 216, 137, 220, 60, 0, 7, 216, 169, 220, 60, 0, 7, 216, 138, 220, 60, 0, 7, 216, 170, 220, 60, 0, 7, 216, 139, 220, 60, 0, 7, 216, 171, 220, 60, 0, 7, 216, 140, 220, 60, 0, 7, 216, 172, 220, 60, 0, 7, 216, 141, 220, 60, 0, 7, 216, 173, 220, 60, 0, 7, 216, 142, 220, 60, 0, 7, 216, 174, 220, 60, 0, 7, 216, 143, 220, 60, 0, 7, 216, 175, 220, 60, 0, 7, 216, 176, 220, 60, 0, 7, 216, 177, 220, 60, 0, 7, 216, 178, 220, 60, 0, 7, 216, 179, 220, 60, 0, 7, 216, 180, 220, 60, 0, 0, 28, 60, 0, 1, 28, 60, 0, 2, 28, 60, 0, 3, 28, 60, 0, 4, 28, 60, 0, 5, 28, 60, 0, 6, 28, 60, 0, 7, 28, 60, 0, 8, 28, 60, 0, 9, 28, 60, 0, 77, 28, 60, 0, 78, 28, 60, 0, 79, 28, 60, 0, 10, 28, 60, 0, 11, 28, 60, 0, 12, 28, 60, 0, 13, 28, 60, 0, 14, 28, 60, 0, 15, 28, 60, 0, 16, 28, 60, 0, 17, 28, 60, 0, 18, 28, 60, 0, 19, 28, 60, 0, 20, 28, 60, 0, 21, 28, 60, 0, 22, 28, 60, 0, 23, 28, 60, 0, 24, 28, 60, 0, 25, 28, 60, 0, 26, 28, 60, 0, 36, 28, 60, 0, 27, 28, 60, 0, 37, 28, 60, 0, 28, 28, 60, 0, 29, 28, 60, 0, 30, 28, 60, 0, 31, 28, 60, 0, 32, 28, 60, 0, 33, 28, 60, 0, 34, 28, 60, 0, 35, 28, 60, 0, 54, 28, 60, 0, 38, 28, 60, 0, 39, 28, 60, 0, 40, 28, 60, 0, 41, 28, 60, 0, 42, 28, 60, 0, 43, 28, 60, 0, 44, 28, 60, 0, 45, 28, 60, 0, 46, 28, 60, 0, 47, 28, 60, 0, 48, 28, 60, 0, 49, 28, 60, 0, 50, 28, 60, 0, 51, 28, 60, 0, 52, 28, 60, 0, 53, 28, 60, 0, 64, 168, 60, 0, 65, 168, 60, 0, 66, 168, 60, 0, 67, 168, 60, 0, 68, 168, 60, 0, 69, 168, 60, 0, 70, 168, 60, 0, 71, 168, 60, 0, 105, 168, 60, 0, 106, 168, 60, 0, 107, 168, 60, 0, 108, 168, 60, 0, 72, 168, 60, 0, 73, 168, 60, 0, 74, 168, 60, 0, 75, 168, 60, 0, 76, 168, 60, 0, 77, 168, 60, 0, 78, 168, 60, 0, 79, 168, 60, 0, 80, 168, 60, 0, 81, 168, 60, 0, 82, 168, 60, 0, 83, 168, 60, 0, 103, 168, 60, 0, 84, 168, 60, 0, 85, 168, 60, 0, 86, 168, 60, 0, 87, 168, 60, 0, 104, 168, 60, 0, 109, 168, 60, 0, 88, 168, 60, 0, 113, 168, 60, 0, 114, 168, 60, 0, 89, 168, 60, 0, 90, 168, 60, 0, 110, 168, 60, 0, 91, 168, 60, 0, 92, 168, 60, 0, 111, 168, 60, 0, 112, 168, 60, 0, 93, 168, 60, 0, 98, 168, 60, 0, 99, 168, 60, 0, 100, 168, 60, 0, 101, 168, 60, 0, 94, 168, 60, 0, 95, 168, 60, 0, 96, 168, 60, 0, 97, 168, 60, 0, 102, 168, 60, 0, 115, 168, 60, 0, 0, 25, 60, 0, 1, 25, 60, 0, 2, 25, 60, 0, 3, 25, 60, 0, 4, 25, 60, 0, 5, 25, 60, 0, 6, 25, 60, 0, 7, 25, 60, 0, 8, 25, 60, 0, 60, 0, 60, 0, 29, 25, 47, 0, 41, 25, 60, 0, 9, 25, 60, 0, 10, 25, 60, 0, 11, 25, 60, 0, 60, 0, 60, 0, 30, 25, 47, 0, 42, 25, 60, 0, 12, 25, 60, 0, 13, 25, 60, 0, 14, 25, 60, 0, 15, 25, 60, 0, 16, 25, 60, 0, 17, 25, 60, 0, 18, 25, 60, 0, 19, 25, 60, 0, 20, 25, 60, 0, 21, 25, 60, 0, 22, 25, 60, 0, 23, 25, 60, 0, 24, 25, 60, 0, 25, 25, 60, 0, 26, 25, 60, 0, 27, 25, 60, 0, 28, 25, 60, 0, 32, 25, 60, 0, 33, 25, 60, 0, 34, 25, 60, 0, 35, 25, 60, 0, 36, 25, 60, 0, 37, 25, 60, 0, 38, 25, 60, 0, 39, 25, 60, 0, 40, 25, 60, 0, 41, 25, 60, 0, 42, 25, 60, 0, 43, 25, 60, 0, 48, 25, 60, 0, 49, 25, 60, 0, 50, 25, 60, 0, 51, 25, 60, 0, 52, 25, 60, 0, 53, 25, 60, 0, 54, 25, 60, 0, 55, 25, 60, 0, 56, 25, 60, 0, 0, 23, 60, 0, 1, 23, 60, 0, 2, 23, 60, 0, 3, 23, 60, 0, 4, 23, 60, 0, 5, 23, 60, 0, 6, 23, 60, 0, 7, 23, 60, 0, 8, 23, 60, 0, 9, 23, 60, 0, 10, 23, 60, 0, 11, 23, 60, 0, 12, 23, 60, 0, 14, 23, 60, 0, 15, 23, 60, 0, 16, 23, 60, 0, 17, 23, 60, 0, 18, 23, 60, 0, 19, 23, 60, 0, 20, 23, 60, 0, 32, 23, 60, 0, 33, 23, 60, 0, 34, 23, 60, 0, 35, 23, 60, 0, 36, 23, 60, 0, 37, 23, 60, 0, 38, 23, 60, 0, 39, 23, 60, 0, 40, 23, 60, 0, 41, 23, 60, 0, 42, 23, 60, 0, 43, 23, 60, 0, 44, 23, 60, 0, 45, 23, 60, 0, 46, 23, 60, 0, 47, 23, 60, 0, 48, 23, 60, 0, 49, 23, 60, 0, 50, 23, 60, 0, 51, 23, 60, 0, 52, 23, 60, 0, 64, 23, 60, 0, 65, 23, 60, 0, 66, 23, 60, 0, 67, 23, 60, 0, 68, 23, 60, 0, 69, 23, 60, 0, 70, 23, 60, 0, 71, 23, 60, 0, 72, 23, 60, 0, 73, 23, 60, 0, 74, 23, 60, 0, 75, 23, 60, 0, 76, 23, 60, 0, 77, 23, 60, 0, 78, 23, 60, 0, 79, 23, 60, 0, 80, 23, 60, 0, 81, 23, 60, 0, 82, 23, 60, 0, 83, 23, 60, 0, 96, 23, 60, 0, 97, 23, 60, 0, 98, 23, 60, 0, 99, 23, 60, 0, 100, 23, 60, 0, 101, 23, 60, 0, 102, 23, 60, 0, 103, 23, 60, 0, 104, 23, 60, 0, 105, 23, 60, 0, 106, 23, 60, 0, 107, 23, 60, 0, 108, 23, 60, 0, 110, 23, 60, 0, 111, 23, 60, 0, 112, 23, 60, 0, 114, 23, 60, 0, 115, 23, 60, 0, 0, 26, 60, 0, 1, 26, 60, 0, 2, 26, 60, 0, 3, 26, 60, 0, 4, 26, 60, 0, 5, 26, 60, 0, 6, 26, 60, 0, 7, 26, 60, 0, 8, 26, 60, 0, 9, 26, 60, 0, 10, 26, 60, 0, 11, 26, 60, 0, 12, 26, 60, 0, 13, 26, 60, 0, 14, 26, 60, 0, 15, 26, 60, 0, 16, 26, 60, 0, 17, 26, 60, 0, 18, 26, 60, 0, 19, 26, 60, 0, 20, 26, 60, 0, 21, 26, 60, 0, 22, 26, 60, 0, 23, 26, 60, 0, 24, 26, 60, 0, 25, 26, 60, 0, 26, 26, 60, 0, 27, 26, 60, 0, 7, 216, 224, 222, 60, 0, 7, 216, 225, 222, 60, 0, 7, 216, 226, 222, 60, 0, 7, 216, 227, 222, 60, 0, 7, 216, 228, 222, 60, 0, 7, 216, 229, 222, 60, 0, 7, 216, 230, 222, 60, 0, 7, 216, 231, 222, 60, 0, 7, 216, 232, 222, 60, 0, 7, 216, 233, 222, 60, 0, 7, 216, 234, 222, 60, 0, 7, 216, 235, 222, 60, 0, 7, 216, 236, 222, 60, 0, 7, 216, 237, 222, 60, 0, 7, 216, 238, 222, 60, 0, 7, 216, 239, 222, 60, 0, 7, 216, 240, 222, 60, 0, 7, 216, 241, 222, 60, 0, 7, 216, 242, 222, 60, 0, 7, 216, 243, 222, 60, 0, 7, 216, 244, 222, 60, 0, 7, 216, 245, 222, 60, 0, 7, 216, 246, 222, 60, 0, 192, 27, 60, 0, 60, 0, 60, 0, 193, 27, 60, 0, 194, 27, 60, 0, 60, 0, 60, 0, 195, 27, 61, 0, 196, 27, 60, 0, 197, 27, 60, 0, 60, 0, 60, 0, 198, 27, 60, 0, 199, 27, 60, 0, 60, 0, 60, 0, 200, 27, 60, 0, 201, 27, 60, 0, 60, 0, 60, 0, 202, 27, 60, 0, 203, 27, 60, 0, 60, 0, 60, 0, 204, 27, 61, 0, 205, 27, 60, 0, 206, 27, 60, 0, 60, 0, 60, 0, 207, 27, 60, 0, 208, 27, 60, 0, 209, 27, 60, 0, 210, 27, 60, 0, 60, 0, 60, 0, 211, 27, 60, 0, 212, 27, 60, 0, 60, 0, 60, 0, 213, 27, 60, 0, 214, 27, 60, 0, 60, 0, 60, 0, 215, 27, 60, 0, 216, 27, 60, 0, 60, 0, 60, 0, 217, 27, 61, 0, 218, 27, 60, 0, 219, 27, 60, 0, 60, 0, 60, 0, 220, 27, 60, 0, 221, 27, 60, 0, 222, 27, 60, 0, 60, 0, 60, 0, 223, 27, 60, 0, 224, 27, 60, 0, 225, 27, 60, 0, 226, 27, 60, 0, 227, 27, 60, 0, 228, 27, 60, 0, 229, 27, 60, 0, 231, 27, 60, 0, 60, 0, 60, 0, 232, 27, 60, 0, 233, 27, 60, 0, 234, 27, 60, 0, 60, 0, 60, 0, 235, 27, 60, 0, 236, 27, 60, 0, 60, 0, 60, 0, 237, 27, 60, 0, 238, 27, 60, 0, 60, 0, 60, 0, 239, 27, 60, 0, 240, 27, 60, 0, 241, 27, 60, 0, 242, 27, 60, 0, 243, 27, 60, 0, 48, 169, 60, 0, 49, 169, 60, 0, 50, 169, 60, 0, 51, 169, 60, 0, 52, 169, 60, 0, 53, 169, 60, 0, 54, 169, 60, 0, 55, 169, 60, 0, 56, 169, 60, 0, 57, 169, 60, 0, 58, 169, 60, 0, 59, 169, 60, 0, 60, 169, 60, 0, 61, 169, 60, 0, 62, 169, 60, 0, 63, 169, 60, 0, 64, 169, 60, 0, 65, 169, 60, 0, 66, 169, 60, 0, 67, 169, 60, 0, 68, 169, 60, 0, 69, 169, 60, 0, 70, 169, 60, 0, 71, 169, 60, 0, 72, 169, 60, 0, 73, 169, 60, 0, 74, 169, 60, 0, 75, 169, 60, 0, 76, 169, 60, 0, 77, 169, 60, 0, 78, 169, 60, 0, 79, 169, 60, 0, 80, 169, 60, 0, 81, 169, 60, 0, 82, 169, 60, 0, 83, 169, 60, 0, 10, 169, 60, 0, 11, 169, 60, 0, 12, 169, 60, 0, 13, 169, 60, 0, 14, 169, 60, 0, 15, 169, 60, 0, 16, 169, 60, 0, 17, 169, 60, 0, 18, 169, 60, 0, 19, 169, 60, 0, 20, 169, 60, 0, 21, 169, 60, 0, 22, 169, 60, 0, 23, 169, 60, 0, 24, 169, 60, 0, 25, 169, 60, 0, 26, 169, 60, 0, 27, 169, 60, 0, 28, 169, 60, 0, 29, 169, 60, 0, 30, 169, 60, 0, 31, 169, 60, 0, 32, 169, 60, 0, 33, 169, 60, 0, 34, 169, 60, 0, 35, 169, 60, 0, 36, 169, 60, 0, 37, 169, 60, 0, 38, 169, 60, 0, 39, 169, 60, 0, 40, 169, 60, 0, 41, 169, 60, 0, 42, 169, 60, 0, 0, 16, 60, 0, 117, 16, 60, 0, 1, 16, 60, 0, 118, 16, 60, 0, 2, 16, 60, 0, 119, 16, 60, 0, 96, 170, 60, 0, 233, 169, 60, 0, 3, 16, 60, 0, 224, 169, 60, 0, 234, 169, 60, 0, 4, 16, 60, 0, 90, 16, 60, 0, 5, 16, 60, 0, 120, 16, 60, 0, 97, 170, 60, 0, 6, 16, 60, 0, 225, 169, 60, 0, 98, 170, 60, 0, 126, 170, 60, 0, 7, 16, 60, 0, 99, 170, 60, 0, 235, 169, 60, 0, 121, 16, 60, 0, 114, 170, 60, 0, 8, 16, 60, 0, 91, 16, 60, 0, 226, 169, 60, 0, 100, 170, 60, 0, 236, 169, 60, 0, 97, 16, 60, 0, 127, 170, 60, 0, 9, 16, 60, 0, 122, 16, 60, 0, 101, 170, 60, 0, 231, 169, 60, 0, 10, 16, 60, 0, 11, 16, 60, 0, 102, 170, 60, 0, 12, 16, 60, 0, 103, 170, 60, 0, 13, 16, 60, 0, 104, 170, 60, 0, 237, 169, 60, 0, 14, 16, 60, 0, 105, 170, 60, 0, 238, 169, 60, 0, 15, 16, 60, 0, 110, 16, 60, 0, 227, 169, 60, 0, 239, 169, 60, 0, 16, 16, 60, 0, 17, 16, 60, 0, 18, 16, 60, 0, 123, 16, 60, 0, 251, 169, 60, 0, 19, 16, 60, 0, 106, 170, 60, 0, 252, 169, 60, 0, 20, 16, 60, 0, 124, 16, 60, 0, 107, 170, 60, 0, 94, 16, 60, 0, 21, 16, 60, 0, 22, 16, 60, 0, 125, 16, 60, 0, 126, 16, 60, 0, 111, 170, 60, 0, 142, 16, 60, 0, 232, 169, 60, 0, 23, 16, 60, 0, 127, 16, 60, 0, 253, 169, 60, 0, 24, 16, 60, 0, 228, 169, 60, 0, 254, 169, 60, 0, 25, 16, 60, 0, 95, 16, 60, 0, 26, 16, 60, 0, 59, 16, 60, 0, 27, 16, 60, 0, 115, 170, 60, 0, 122, 170, 60, 0, 60, 16, 60, 0, 28, 16, 60, 0, 96, 16, 60, 0, 29, 16, 60, 0, 61, 16, 60, 0, 130, 16, 60, 0, 128, 16, 60, 0, 80, 16, 60, 0, 81, 16, 60, 0, 101, 16, 60, 0, 30, 16, 60, 0, 60, 0, 60, 0, 63, 16, 47, 0, 57, 16, 30, 16, 60, 0, 108, 170, 60, 0, 31, 16, 60, 0, 129, 16, 60, 0, 109, 170, 60, 0, 62, 16, 60, 0, 110, 170, 60, 0, 113, 170, 60, 0, 32, 16, 60, 0, 250, 169, 60, 0, 92, 16, 60, 0, 93, 16, 60, 0, 111, 16, 60, 0, 112, 16, 60, 0, 102, 16, 60, 0, 33, 16, 60, 0, 34, 16, 60, 0, 35, 16, 60, 0, 36, 16, 60, 0, 37, 16, 60, 0, 38, 16, 60, 0, 82, 16, 60, 0, 83, 16, 60, 0, 84, 16, 60, 0, 85, 16, 60, 0, 39, 16, 60, 0, 40, 16, 60, 0, 41, 16, 60, 0, 42, 16, 60, 0, 44, 16, 60, 0, 60, 0, 60, 0, 43, 16, 60, 0, 131, 16, 60, 0, 114, 16, 60, 0, 156, 16, 60, 0, 45, 16, 60, 0, 113, 16, 60, 0, 46, 16, 60, 0, 51, 16, 60, 0, 47, 16, 60, 0, 115, 16, 60, 0, 116, 16, 60, 0, 48, 16, 60, 0, 86, 16, 60, 0, 87, 16, 60, 0, 88, 16, 60, 0, 89, 16, 60, 0, 49, 16, 60, 0, 132, 16, 60, 0, 53, 16, 60, 0, 133, 16, 60, 0, 50, 16, 60, 0, 157, 16, 60, 0, 52, 16, 60, 0, 98, 16, 60, 0, 103, 16, 60, 0, 104, 16, 60, 0, 229, 169, 60, 0, 134, 16, 60, 0, 57, 16, 60, 0, 58, 16, 60, 0, 99, 16, 60, 0, 100, 16, 60, 0, 105, 16, 60, 0, 106, 16, 60, 0, 107, 16, 60, 0, 108, 16, 60, 0, 109, 16, 60, 0, 135, 16, 60, 0, 139, 16, 60, 0, 136, 16, 60, 0, 140, 16, 60, 0, 141, 16, 60, 0, 137, 16, 60, 0, 138, 16, 60, 0, 143, 16, 60, 0, 154, 16, 60, 0, 155, 16, 60, 0, 123, 170, 60, 0, 124, 170, 60, 0, 125, 170, 60, 0, 116, 170, 60, 0, 117, 170, 60, 0, 118, 170, 60, 0, 14, 32, 3, 216, 0, 221, 14, 32, 60, 0, 14, 32, 3, 216, 29, 221, 14, 32, 60, 0, 14, 32, 3, 216, 30, 221, 14, 32, 60, 0, 14, 32, 3, 216, 31, 221, 14, 32, 60, 0, 14, 32, 3, 216, 32, 221, 14, 32, 60, 0, 14, 32, 3, 216, 33, 221, 14, 32, 60, 0, 14, 32, 3, 216, 1, 221, 14, 32, 60, 0, 14, 32, 3, 216, 2, 221, 14, 32, 60, 0, 14, 32, 3, 216, 3, 221, 14, 32, 60, 0, 14, 32, 3, 216, 4, 221, 14, 32, 60, 0, 14, 32, 3, 216, 5, 221, 14, 32, 60, 0, 14, 32, 3, 216, 6, 221, 14, 32, 60, 0, 14, 32, 3, 216, 7, 221, 14, 32, 60, 0, 14, 32, 3, 216, 8, 221, 14, 32, 60, 0, 14, 32, 3, 216, 9, 221, 14, 32, 60, 0, 14, 32, 3, 216, 10, 221, 14, 32, 60, 0, 14, 32, 3, 216, 11, 221, 14, 32, 60, 0, 14, 32, 3, 216, 12, 221, 14, 32, 60, 0, 14, 32, 3, 216, 13, 221, 14, 32, 60, 0, 14, 32, 3, 216, 14, 221, 14, 32, 60, 0, 14, 32, 3, 216, 15, 221, 14, 32, 60, 0, 14, 32, 3, 216, 16, 221, 14, 32, 60, 0, 14, 32, 3, 216, 17, 221, 14, 32, 60, 0, 14, 32, 3, 216, 18, 221, 14, 32, 60, 0, 14, 32, 3, 216, 19, 221, 14, 32, 60, 0, 14, 32, 3, 216, 20, 221, 14, 32, 60, 0, 14, 32, 3, 216, 21, 221, 14, 32, 60, 0, 14, 32, 3, 216, 22, 221, 14, 32, 60, 0, 14, 32, 3, 216, 23, 221, 14, 32, 60, 0, 14, 32, 3, 216, 24, 221, 14, 32, 60, 0, 14, 32, 3, 216, 25, 221, 14, 32, 60, 0, 14, 32, 3, 216, 26, 221, 14, 32, 60, 0, 14, 32, 3, 216, 27, 221, 14, 32, 60, 0, 14, 32, 3, 216, 28, 221, 14, 32, 60, 0, 14, 32, 3, 216, 34, 221, 14, 32, 60, 0, 14, 32, 3, 216, 35, 221, 14, 32, 60, 0, 4, 216, 3, 221, 60, 0, 4, 216, 4, 221, 60, 0, 4, 216, 5, 221, 60, 0, 4, 216, 6, 221, 60, 0, 4, 216, 7, 221, 60, 0, 4, 216, 8, 221, 60, 0, 4, 216, 9, 221, 60, 0, 4, 216, 10, 221, 60, 0, 4, 216, 11, 221, 60, 0, 4, 216, 12, 221, 60, 0, 4, 216, 13, 221, 60, 0, 4, 216, 14, 221, 60, 0, 4, 216, 15, 221, 60, 0, 4, 216, 16, 221, 60, 0, 4, 216, 17, 221, 60, 0, 4, 216, 18, 221, 60, 0, 4, 216, 19, 221, 60, 0, 4, 216, 20, 221, 60, 0, 4, 216, 21, 221, 60, 0, 4, 216, 22, 221, 60, 0, 4, 216, 23, 221, 60, 0, 4, 216, 24, 221, 60, 0, 4, 216, 25, 221, 60, 0, 4, 216, 26, 221, 60, 0, 4, 216, 27, 221, 60, 0, 4, 216, 28, 221, 60, 0, 4, 216, 29, 221, 60, 0, 4, 216, 30, 221, 60, 0, 4, 216, 31, 221, 60, 0, 4, 216, 32, 221, 60, 0, 4, 216, 33, 221, 60, 0, 4, 216, 34, 221, 60, 0, 4, 216, 35, 221, 60, 0, 4, 216, 71, 221, 60, 0, 4, 216, 36, 221, 60, 0, 4, 216, 37, 221, 60, 0, 4, 216, 38, 221, 60, 0, 4, 216, 68, 221, 60, 0, 4, 216, 39, 221, 60, 0, 4, 216, 40, 221, 60, 0, 4, 216, 41, 221, 60, 0, 4, 216, 42, 221, 60, 0, 4, 216, 43, 221, 60, 0, 4, 216, 44, 221, 60, 0, 4, 216, 45, 221, 60, 0, 4, 216, 46, 221, 60, 0, 4, 216, 47, 221, 60, 0, 4, 216, 48, 221, 60, 0, 4, 216, 49, 221, 60, 0, 4, 216, 50, 221, 60, 0, 4, 216, 69, 221, 60, 0, 4, 216, 70, 221, 60, 0, 4, 216, 51, 221, 60, 0, 4, 216, 52, 221, 60, 0, 128, 23, 60, 0, 129, 23, 60, 0, 130, 23, 60, 0, 131, 23, 60, 0, 132, 23, 60, 0, 133, 23, 60, 0, 134, 23, 60, 0, 135, 23, 60, 0, 136, 23, 60, 0, 137, 23, 60, 0, 138, 23, 60, 0, 139, 23, 60, 0, 140, 23, 60, 0, 141, 23, 60, 0, 142, 23, 60, 0, 143, 23, 60, 0, 144, 23, 60, 0, 145, 23, 60, 0, 146, 23, 60, 0, 147, 23, 60, 0, 148, 23, 60, 0, 149, 23, 60, 0, 150, 23, 60, 0, 151, 23, 60, 0, 152, 23, 60, 0, 153, 23, 60, 0, 154, 23, 60, 0, 155, 23, 60, 0, 156, 23, 60, 0, 157, 23, 60, 0, 158, 23, 60, 0, 159, 23, 60, 0, 160, 23, 60, 0, 161, 23, 60, 0, 162, 23, 60, 0, 220, 23, 60, 0, 163, 23, 60, 0, 164, 23, 60, 0, 165, 23, 60, 0, 166, 23, 60, 0, 167, 23, 60, 0, 168, 23, 60, 0, 169, 23, 60, 0, 170, 23, 60, 0, 171, 23, 60, 0, 172, 23, 60, 0, 173, 23, 60, 0, 174, 23, 60, 0, 175, 23, 60, 0, 176, 23, 60, 0, 177, 23, 60, 0, 178, 23, 60, 0, 179, 23, 60, 0, 182, 23, 60, 0, 183, 23, 60, 0, 184, 23, 60, 0, 185, 23, 60, 0, 186, 23, 60, 0, 187, 23, 60, 0, 188, 23, 60, 0, 189, 23, 60, 0, 190, 23, 60, 0, 191, 23, 60, 0, 192, 23, 60, 0, 193, 23, 60, 0, 194, 23, 60, 0, 195, 23, 60, 0, 196, 23, 60, 0, 197, 23, 60, 0, 210, 23, 60, 0, 80, 25, 60, 0, 81, 25, 60, 0, 82, 25, 60, 0, 83, 25, 60, 0, 84, 25, 60, 0, 85, 25, 60, 0, 86, 25, 60, 0, 87, 25, 60, 0, 88, 25, 60, 0, 89, 25, 60, 0, 90, 25, 60, 0, 91, 25, 60, 0, 92, 25, 60, 0, 93, 25, 60, 0, 94, 25, 60, 0, 95, 25, 60, 0, 96, 25, 60, 0, 97, 25, 60, 0, 98, 25, 60, 0, 99, 25, 60, 0, 100, 25, 60, 0, 101, 25, 60, 0, 102, 25, 60, 0, 103, 25, 60, 0, 104, 25, 60, 0, 105, 25, 60, 0, 106, 25, 60, 0, 107, 25, 60, 0, 108, 25, 60, 0, 109, 25, 60, 0, 112, 25, 60, 0, 113, 25, 60, 0, 114, 25, 60, 0, 115, 25, 60, 0, 116, 25, 60, 0, 128, 25, 60, 0, 129, 25, 60, 0, 130, 25, 60, 0, 131, 25, 60, 0, 132, 25, 60, 0, 133, 25, 60, 0, 134, 25, 60, 0, 135, 25, 60, 0, 136, 25, 60, 0, 137, 25, 60, 0, 138, 25, 60, 0, 139, 25, 60, 0, 140, 25, 60, 0, 141, 25, 60, 0, 142, 25, 60, 0, 143, 25, 60, 0, 144, 25, 60, 0, 145, 25, 60, 0, 146, 25, 60, 0, 147, 25, 60, 0, 148, 25, 60, 0, 149, 25, 60, 0, 150, 25, 60, 0, 151, 25, 60, 0, 152, 25, 60, 0, 153, 25, 60, 0, 154, 25, 60, 0, 155, 25, 60, 0, 156, 25, 60, 0, 60, 0, 60, 0, 222, 25, 47, 0, 182, 25, 60, 0, 60, 0, 60, 0, 223, 25, 47, 0, 182, 25, 193, 25, 60, 0, 157, 25, 60, 0, 158, 25, 60, 0, 159, 25, 60, 0, 160, 25, 60, 0, 161, 25, 60, 0, 162, 25, 60, 0, 163, 25, 60, 0, 164, 25, 60, 0, 165, 25, 60, 0, 166, 25, 60, 0, 167, 25, 60, 0, 168, 25, 60, 0, 169, 25, 60, 0, 170, 25, 60, 0, 171, 25, 60, 0, 176, 25, 60, 0, 177, 25, 60, 0, 178, 25, 60, 0, 179, 25, 60, 0, 180, 25, 60, 0, 181, 25, 60, 0, 182, 25, 60, 0, 183, 25, 60, 0, 184, 25, 60, 0, 185, 25, 60, 0, 186, 25, 60, 0, 187, 25, 60, 0, 188, 25, 60, 0, 189, 25, 60, 0, 190, 25, 60, 0, 191, 25, 60, 0, 192, 25, 60, 0, 193, 25, 60, 0, 194, 25, 60, 0, 195, 25, 60, 0, 196, 25, 60, 0, 197, 25, 60, 0, 198, 25, 60, 0, 199, 25, 60, 0, 200, 25, 60, 0, 201, 25, 60, 0, 32, 26, 60, 0, 33, 26, 60, 0, 34, 26, 60, 0, 35, 26, 60, 0, 36, 26, 60, 0, 37, 26, 60, 0, 38, 26, 60, 0, 60, 0, 60, 0, 88, 26, 61, 0, 89, 26, 60, 0, 39, 26, 60, 0, 40, 26, 60, 0, 41, 26, 60, 0, 42, 26, 60, 0, 43, 26, 60, 0, 44, 26, 60, 0, 45, 26, 60, 0, 46, 26, 60, 0, 47, 26, 60, 0, 48, 26, 60, 0, 49, 26, 60, 0, 50, 26, 60, 0, 51, 26, 60, 0, 52, 26, 60, 0, 53, 26, 60, 0, 54, 26, 60, 0, 55, 26, 60, 0, 56, 26, 60, 0, 57, 26, 60, 0, 58, 26, 60, 0, 59, 26, 60, 0, 60, 0, 60, 0, 90, 26, 61, 0, 91, 26, 60, 0, 60, 26, 60, 0, 61, 26, 60, 0, 62, 26, 60, 0, 63, 26, 60, 0, 64, 26, 60, 0, 65, 26, 60, 0, 66, 26, 60, 0, 67, 26, 60, 0, 68, 26, 60, 0, 69, 26, 60, 0, 70, 26, 60, 0, 60, 0, 60, 0, 84, 26, 47, 0, 96, 26, 70, 26, 60, 0, 71, 26, 60, 0, 72, 26, 60, 0, 73, 26, 60, 0, 74, 26, 60, 0, 75, 26, 60, 0, 76, 26, 60, 0, 83, 26, 60, 0, 107, 26, 60, 0, 85, 26, 60, 0, 86, 26, 60, 0, 87, 26, 60, 0, 92, 26, 60, 0, 93, 26, 60, 0, 94, 26, 60, 0, 77, 26, 60, 0, 78, 26, 60, 0, 79, 26, 60, 0, 80, 26, 60, 0, 81, 26, 60, 0, 82, 26, 60, 0, 97, 26, 60, 0, 108, 26, 60, 0, 98, 26, 60, 0, 99, 26, 60, 0, 60, 0, 60, 0, 100, 26, 60, 0, 101, 26, 60, 0, 102, 26, 60, 0, 103, 26, 60, 0, 104, 26, 60, 0, 105, 26, 60, 0, 106, 26, 60, 0, 110, 26, 60, 0, 111, 26, 60, 0, 115, 26, 60, 0, 112, 26, 60, 0, 113, 26, 60, 0, 114, 26, 60, 0, 109, 26, 60, 0, 96, 26, 60, 0, 0, 170, 60, 0, 1, 170, 60, 0, 2, 170, 60, 0, 3, 170, 60, 0, 4, 170, 60, 0, 5, 170, 60, 0, 6, 170, 60, 0, 7, 170, 60, 0, 8, 170, 60, 0, 9, 170, 60, 0, 10, 170, 60, 0, 11, 170, 60, 0, 12, 170, 60, 0, 13, 170, 60, 0, 14, 170, 60, 0, 15, 170, 60, 0, 16, 170, 60, 0, 17, 170, 60, 0, 18, 170, 60, 0, 19, 170, 60, 0, 20, 170, 60, 0, 21, 170, 60, 0, 22, 170, 60, 0, 23, 170, 60, 0, 24, 170, 60, 0, 25, 170, 60, 0, 26, 170, 60, 0, 27, 170, 60, 0, 28, 170, 60, 0, 29, 170, 60, 0, 30, 170, 60, 0, 31, 170, 60, 0, 32, 170, 60, 0, 33, 170, 60, 0, 34, 170, 60, 0, 35, 170, 60, 0, 36, 170, 60, 0, 37, 170, 60, 0, 38, 170, 60, 0, 39, 170, 60, 0, 40, 170, 60, 0, 51, 170, 60, 0, 52, 170, 60, 0, 53, 170, 60, 0, 54, 170, 60, 0, 41, 170, 60, 0, 42, 170, 60, 0, 43, 170, 60, 0, 44, 170, 60, 0, 45, 170, 60, 0, 46, 170, 60, 0, 47, 170, 60, 0, 48, 170, 60, 0, 49, 170, 60, 0, 50, 170, 60, 0, 64, 170, 60, 0, 65, 170, 60, 0, 66, 170, 60, 0, 67, 170, 60, 0, 68, 170, 60, 0, 69, 170, 60, 0, 70, 170, 60, 0, 71, 170, 60, 0, 72, 170, 60, 0, 73, 170, 60, 0, 74, 170, 60, 0, 75, 170, 60, 0, 76, 170, 60, 0, 77, 170, 60, 0, 5, 27, 60, 0, 6, 27, 60, 0, 7, 27, 60, 0, 8, 27, 60, 0, 9, 27, 60, 0, 10, 27, 60, 0, 11, 27, 60, 0, 12, 27, 60, 0, 13, 27, 60, 0, 14, 27, 60, 0, 15, 27, 60, 0, 16, 27, 60, 0, 17, 27, 60, 0, 18, 27, 60, 0, 19, 27, 60, 0, 69, 27, 60, 0, 70, 27, 60, 0, 20, 27, 60, 0, 21, 27, 60, 0, 22, 27, 60, 0, 23, 27, 60, 0, 24, 27, 60, 0, 25, 27, 60, 0, 26, 27, 60, 0, 27, 27, 60, 0, 28, 27, 60, 0, 29, 27, 60, 0, 30, 27, 60, 0, 31, 27, 60, 0, 32, 27, 60, 0, 33, 27, 60, 0, 34, 27, 60, 0, 71, 27, 60, 0, 35, 27, 60, 0, 36, 27, 60, 0, 37, 27, 60, 0, 38, 27, 60, 0, 39, 27, 60, 0, 72, 27, 60, 0, 40, 27, 60, 0, 41, 27, 60, 0, 42, 27, 60, 0, 43, 27, 60, 0, 44, 27, 60, 0, 45, 27, 60, 0, 46, 27, 60, 0, 47, 27, 60, 0, 73, 27, 60, 0, 48, 27, 60, 0, 49, 27, 60, 0, 50, 27, 60, 0, 74, 27, 60, 0, 75, 27, 60, 0, 51, 27, 60, 0, 53, 27, 60, 0, 54, 27, 60, 0, 55, 27, 60, 0, 56, 27, 60, 0, 57, 27, 60, 0, 58, 27, 60, 0, 59, 27, 60, 0, 60, 27, 60, 0, 61, 27, 60, 0, 62, 27, 60, 0, 63, 27, 60, 0, 64, 27, 60, 0, 65, 27, 60, 0, 66, 27, 60, 0, 67, 27, 60, 0, 68, 27, 60, 0, 132, 169, 60, 0, 133, 169, 60, 0, 134, 169, 60, 0, 135, 169, 60, 0, 136, 169, 60, 0, 137, 169, 60, 0, 138, 169, 60, 0, 139, 169, 60, 0, 140, 169, 60, 0, 141, 169, 60, 0, 142, 169, 60, 0, 143, 169, 60, 0, 144, 169, 60, 0, 145, 169, 60, 0, 146, 169, 60, 0, 147, 169, 60, 0, 148, 169, 60, 0, 149, 169, 60, 0, 150, 169, 60, 0, 151, 169, 60, 0, 152, 169, 60, 0, 153, 169, 60, 0, 154, 169, 60, 0, 155, 169, 60, 0, 156, 169, 60, 0, 157, 169, 60, 0, 158, 169, 60, 0, 159, 169, 60, 0, 160, 169, 60, 0, 161, 169, 60, 0, 162, 169, 60, 0, 163, 169, 60, 0, 164, 169, 60, 0, 165, 169, 60, 0, 166, 169, 60, 0, 167, 169, 60, 0, 168, 169, 60, 0, 169, 169, 60, 0, 170, 169, 60, 0, 190, 169, 60, 0, 171, 169, 60, 0, 60, 0, 60, 0, 172, 169, 60, 0, 191, 169, 60, 0, 173, 169, 60, 0, 174, 169, 60, 0, 175, 169, 60, 0, 176, 169, 60, 0, 177, 169, 60, 0, 178, 169, 60, 0, 180, 169, 60, 0, 60, 0, 60, 0, 181, 169, 60, 0, 188, 169, 60, 0, 182, 169, 60, 0, 183, 169, 60, 0, 184, 169, 60, 0, 185, 169, 60, 0, 189, 169, 60, 0, 186, 169, 60, 0, 187, 169, 60, 0, 192, 169, 60, 0, 128, 24, 60, 0, 129, 24, 60, 0, 130, 24, 60, 0, 131, 24, 60, 0, 132, 24, 60, 0, 133, 24, 60, 0, 134, 24, 60, 0, 67, 24, 60, 0, 32, 24, 60, 0, 135, 24, 60, 0, 33, 24, 60, 0, 68, 24, 60, 0, 93, 24, 60, 0, 34, 24, 60, 0, 69, 24, 60, 0, 94, 24, 60, 0, 115, 24, 60, 0, 136, 24, 60, 0, 95, 24, 60, 0, 35, 24, 60, 0, 70, 24, 60, 0, 36, 24, 60, 0, 71, 24, 60, 0, 97, 24, 60, 0, 37, 24, 60, 0, 72, 24, 60, 0, 38, 24, 60, 0, 73, 24, 60, 0, 96, 24, 60, 0, 39, 24, 60, 0, 40, 24, 60, 0, 41, 24, 60, 0, 74, 24, 60, 0, 98, 24, 60, 0, 138, 24, 60, 0, 155, 24, 60, 0, 42, 24, 60, 0, 75, 24, 60, 0, 43, 24, 60, 0, 76, 24, 60, 0, 102, 24, 60, 0, 44, 24, 60, 0, 77, 24, 60, 0, 45, 24, 60, 0, 78, 24, 60, 0, 100, 24, 60, 0, 154, 24, 60, 0, 101, 24, 60, 0, 46, 24, 60, 0, 79, 24, 60, 0, 47, 24, 60, 0, 48, 24, 60, 0, 49, 24, 60, 0, 103, 24, 60, 0, 156, 24, 60, 0, 157, 24, 60, 0, 162, 24, 60, 0, 164, 24, 60, 0, 165, 24, 60, 0, 50, 24, 60, 0, 80, 24, 60, 0, 104, 24, 60, 0, 51, 24, 60, 0, 81, 24, 60, 0, 105, 24, 60, 0, 52, 24, 60, 0, 82, 24, 60, 0, 113, 24, 60, 0, 120, 24, 60, 0, 92, 24, 60, 0, 139, 24, 60, 0, 53, 24, 60, 0, 83, 24, 60, 0, 106, 24, 60, 0, 119, 24, 60, 0, 54, 24, 60, 0, 85, 24, 60, 0, 114, 24, 60, 0, 55, 24, 60, 0, 117, 24, 60, 0, 56, 24, 60, 0, 86, 24, 60, 0, 57, 24, 60, 0, 107, 24, 60, 0, 118, 24, 60, 0, 58, 24, 60, 0, 87, 24, 60, 0, 99, 24, 60, 0, 116, 24, 60, 0, 137, 24, 60, 0, 59, 24, 60, 0, 60, 24, 60, 0, 84, 24, 60, 0, 110, 24, 60, 0, 61, 24, 60, 0, 111, 24, 60, 0, 88, 24, 60, 0, 108, 24, 60, 0, 62, 24, 60, 0, 89, 24, 60, 0, 109, 24, 60, 0, 63, 24, 60, 0, 64, 24, 60, 0, 65, 24, 60, 0, 66, 24, 60, 0, 90, 24, 60, 0, 91, 24, 60, 0, 112, 24, 60, 0, 140, 24, 60, 0, 158, 24, 60, 0, 141, 24, 60, 0, 142, 24, 60, 0, 159, 24, 60, 0, 143, 24, 60, 0, 144, 24, 60, 0, 152, 24, 60, 0, 160, 24, 60, 0, 145, 24, 60, 0, 161, 24, 60, 0, 146, 24, 60, 0, 147, 24, 60, 0, 168, 24, 60, 0, 148, 24, 60, 0, 163, 24, 60, 0, 149, 24, 60, 0, 153, 24, 60, 0, 150, 24, 60, 0, 151, 24, 60, 0, 166, 24, 60, 0, 167, 24, 60, 0, 170, 24, 60, 0, 169, 24, 60, 0, 90, 28, 60, 0, 91, 28, 60, 0, 92, 28, 60, 0, 93, 28, 60, 0, 94, 28, 60, 0, 95, 28, 60, 0, 96, 28, 60, 0, 97, 28, 60, 0, 98, 28, 60, 0, 99, 28, 60, 0, 100, 28, 60, 0, 101, 28, 60, 0, 102, 28, 60, 0, 103, 28, 60, 0, 104, 28, 60, 0, 105, 28, 60, 0, 106, 28, 60, 0, 107, 28, 60, 0, 108, 28, 60, 0, 109, 28, 60, 0, 110, 28, 60, 0, 111, 28, 60, 0, 112, 28, 60, 0, 113, 28, 60, 0, 114, 28, 60, 0, 115, 28, 60, 0, 116, 28, 60, 0, 117, 28, 60, 0, 118, 28, 60, 0, 119, 28, 60, 0, 120, 28, 60, 0, 121, 28, 60, 0, 122, 28, 60, 0, 123, 28, 60, 0, 124, 28, 60, 0, 125, 28, 60, 0, 112, 171, 60, 0, 60, 0, 60, 0, 160, 19, 60, 0, 113, 171, 60, 0, 60, 0, 60, 0, 161, 19, 60, 0, 114, 171, 60, 0, 60, 0, 60, 0, 162, 19, 60, 0, 115, 171, 60, 0, 60, 0, 60, 0, 163, 19, 60, 0, 116, 171, 60, 0, 60, 0, 60, 0, 164, 19, 60, 0, 117, 171, 60, 0, 60, 0, 60, 0, 165, 19, 60, 0, 118, 171, 60, 0, 60, 0, 60, 0, 166, 19, 60, 0, 119, 171, 60, 0, 60, 0, 60, 0, 167, 19, 60, 0, 120, 171, 60, 0, 60, 0, 60, 0, 168, 19, 60, 0, 121, 171, 60, 0, 60, 0, 60, 0, 169, 19, 60, 0, 122, 171, 60, 0, 60, 0, 60, 0, 170, 19, 60, 0, 123, 171, 60, 0, 60, 0, 60, 0, 171, 19, 60, 0, 124, 171, 60, 0, 60, 0, 60, 0, 172, 19, 60, 0, 125, 171, 60, 0, 60, 0, 60, 0, 173, 19, 60, 0, 126, 171, 60, 0, 60, 0, 60, 0, 174, 19, 60, 0, 127, 171, 60, 0, 60, 0, 60, 0, 175, 19, 60, 0, 128, 171, 60, 0, 60, 0, 60, 0, 176, 19, 60, 0, 129, 171, 60, 0, 60, 0, 60, 0, 177, 19, 60, 0, 130, 171, 60, 0, 60, 0, 60, 0, 178, 19, 60, 0, 131, 171, 60, 0, 60, 0, 60, 0, 179, 19, 60, 0, 132, 171, 60, 0, 60, 0, 60, 0, 180, 19, 60, 0, 133, 171, 60, 0, 60, 0, 60, 0, 181, 19, 60, 0, 134, 171, 60, 0, 60, 0, 60, 0, 182, 19, 60, 0, 135, 171, 60, 0, 60, 0, 60, 0, 183, 19, 60, 0, 136, 171, 60, 0, 60, 0, 60, 0, 184, 19, 60, 0, 137, 171, 60, 0, 60, 0, 60, 0, 185, 19, 60, 0, 138, 171, 60, 0, 60, 0, 60, 0, 186, 19, 60, 0, 139, 171, 60, 0, 60, 0, 60, 0, 187, 19, 60, 0, 140, 171, 60, 0, 60, 0, 60, 0, 188, 19, 60, 0, 141, 171, 60, 0, 60, 0, 60, 0, 189, 19, 60, 0, 142, 171, 60, 0, 60, 0, 60, 0, 190, 19, 60, 0, 143, 171, 60, 0, 60, 0, 60, 0, 191, 19, 60, 0, 144, 171, 60, 0, 60, 0, 60, 0, 192, 19, 60, 0, 145, 171, 60, 0, 60, 0, 60, 0, 193, 19, 60, 0, 146, 171, 60, 0, 60, 0, 60, 0, 194, 19, 60, 0, 147, 171, 60, 0, 60, 0, 60, 0, 195, 19, 60, 0, 148, 171, 60, 0, 60, 0, 60, 0, 196, 19, 60, 0, 149, 171, 60, 0, 60, 0, 60, 0, 197, 19, 60, 0, 150, 171, 60, 0, 60, 0, 60, 0, 198, 19, 60, 0, 151, 171, 60, 0, 60, 0, 60, 0, 199, 19, 60, 0, 152, 171, 60, 0, 60, 0, 60, 0, 200, 19, 60, 0, 153, 171, 60, 0, 60, 0, 60, 0, 201, 19, 60, 0, 154, 171, 60, 0, 60, 0, 60, 0, 202, 19, 60, 0, 155, 171, 60, 0, 60, 0, 60, 0, 203, 19, 60, 0, 156, 171, 60, 0, 60, 0, 60, 0, 204, 19, 60, 0, 157, 171, 60, 0, 60, 0, 60, 0, 205, 19, 60, 0, 158, 171, 60, 0, 60, 0, 60, 0, 206, 19, 60, 0, 159, 171, 60, 0, 60, 0, 60, 0, 207, 19, 60, 0, 160, 171, 60, 0, 60, 0, 60, 0, 208, 19, 60, 0, 161, 171, 60, 0, 60, 0, 60, 0, 209, 19, 60, 0, 162, 171, 60, 0, 60, 0, 60, 0, 210, 19, 60, 0, 163, 171, 60, 0, 60, 0, 60, 0, 211, 19, 60, 0, 164, 171, 60, 0, 60, 0, 60, 0, 212, 19, 60, 0, 165, 171, 60, 0, 60, 0, 60, 0, 213, 19, 60, 0, 166, 171, 60, 0, 60, 0, 60, 0, 214, 19, 60, 0, 167, 171, 60, 0, 60, 0, 60, 0, 215, 19, 60, 0, 168, 171, 60, 0, 60, 0, 60, 0, 216, 19, 60, 0, 169, 171, 60, 0, 60, 0, 60, 0, 217, 19, 60, 0, 170, 171, 60, 0, 60, 0, 60, 0, 218, 19, 60, 0, 171, 171, 60, 0, 60, 0, 60, 0, 219, 19, 60, 0, 172, 171, 60, 0, 60, 0, 60, 0, 220, 19, 60, 0, 173, 171, 60, 0, 60, 0, 60, 0, 221, 19, 60, 0, 174, 171, 60, 0, 60, 0, 60, 0, 222, 19, 60, 0, 175, 171, 60, 0, 60, 0, 60, 0, 223, 19, 60, 0, 176, 171, 60, 0, 60, 0, 60, 0, 224, 19, 60, 0, 177, 171, 60, 0, 60, 0, 60, 0, 225, 19, 60, 0, 178, 171, 60, 0, 60, 0, 60, 0, 226, 19, 60, 0, 179, 171, 60, 0, 60, 0, 60, 0, 227, 19, 60, 0, 180, 171, 60, 0, 60, 0, 60, 0, 228, 19, 60, 0, 181, 171, 60, 0, 60, 0, 60, 0, 229, 19, 60, 0, 182, 171, 60, 0, 60, 0, 60, 0, 230, 19, 60, 0, 183, 171, 60, 0, 60, 0, 60, 0, 231, 19, 60, 0, 184, 171, 60, 0, 60, 0, 60, 0, 232, 19, 60, 0, 185, 171, 60, 0, 60, 0, 60, 0, 233, 19, 60, 0, 186, 171, 60, 0, 60, 0, 60, 0, 234, 19, 60, 0, 187, 171, 60, 0, 60, 0, 60, 0, 235, 19, 60, 0, 188, 171, 60, 0, 60, 0, 60, 0, 236, 19, 60, 0, 189, 171, 60, 0, 60, 0, 60, 0, 237, 19, 60, 0, 190, 171, 60, 0, 60, 0, 60, 0, 238, 19, 60, 0, 191, 171, 60, 0, 60, 0, 60, 0, 239, 19, 60, 0, 248, 19, 60, 0, 60, 0, 60, 0, 240, 19, 60, 0, 249, 19, 60, 0, 60, 0, 60, 0, 241, 19, 60, 0, 250, 19, 60, 0, 60, 0, 60, 0, 242, 19, 60, 0, 251, 19, 60, 0, 60, 0, 60, 0, 243, 19, 60, 0, 252, 19, 60, 0, 60, 0, 60, 0, 244, 19, 60, 0, 253, 19, 60, 0, 60, 0, 60, 0, 245, 19, 60, 0, 1, 216, 216, 220, 60, 0, 60, 0, 60, 0, 1, 216, 176, 220, 60, 0, 1, 216, 217, 220, 60, 0, 60, 0, 60, 0, 1, 216, 177, 220, 60, 0, 1, 216, 218, 220, 60, 0, 60, 0, 60, 0, 1, 216, 178, 220, 60, 0, 1, 216, 219, 220, 60, 0, 60, 0, 60, 0, 1, 216, 179, 220, 60, 0, 1, 216, 220, 220, 60, 0, 60, 0, 60, 0, 1, 216, 180, 220, 60, 0, 1, 216, 221, 220, 60, 0, 60, 0, 60, 0, 1, 216, 181, 220, 60, 0, 1, 216, 222, 220, 60, 0, 60, 0, 60, 0, 1, 216, 182, 220, 60, 0, 1, 216, 223, 220, 60, 0, 60, 0, 60, 0, 1, 216, 183, 220, 60, 0, 1, 216, 224, 220, 60, 0, 60, 0, 60, 0, 1, 216, 184, 220, 60, 0, 1, 216, 225, 220, 60, 0, 60, 0, 60, 0, 1, 216, 185, 220, 60, 0, 1, 216, 226, 220, 60, 0, 60, 0, 60, 0, 1, 216, 186, 220, 60, 0, 1, 216, 227, 220, 60, 0, 60, 0, 60, 0, 1, 216, 187, 220, 60, 0, 1, 216, 228, 220, 60, 0, 60, 0, 60, 0, 1, 216, 188, 220, 60, 0, 1, 216, 229, 220, 60, 0, 60, 0, 60, 0, 1, 216, 189, 220, 60, 0, 1, 216, 230, 220, 60, 0, 60, 0, 60, 0, 1, 216, 190, 220, 60, 0, 1, 216, 231, 220, 60, 0, 60, 0, 60, 0, 1, 216, 191, 220, 60, 0, 1, 216, 232, 220, 60, 0, 60, 0, 60, 0, 1, 216, 192, 220, 60, 0, 1, 216, 233, 220, 60, 0, 60, 0, 60, 0, 1, 216, 193, 220, 60, 0, 1, 216, 234, 220, 60, 0, 60, 0, 60, 0, 1, 216, 194, 220, 60, 0, 1, 216, 235, 220, 60, 0, 60, 0, 60, 0, 1, 216, 195, 220, 60, 0, 1, 216, 236, 220, 60, 0, 60, 0, 60, 0, 1, 216, 196, 220, 60, 0, 1, 216, 237, 220, 60, 0, 60, 0, 60, 0, 1, 216, 197, 220, 60, 0, 1, 216, 238, 220, 60, 0, 60, 0, 60, 0, 1, 216, 198, 220, 60, 0, 1, 216, 239, 220, 60, 0, 60, 0, 60, 0, 1, 216, 199, 220, 60, 0, 1, 216, 240, 220, 60, 0, 60, 0, 60, 0, 1, 216, 200, 220, 60, 0, 1, 216, 241, 220, 60, 0, 60, 0, 60, 0, 1, 216, 201, 220, 60, 0, 1, 216, 242, 220, 60, 0, 60, 0, 60, 0, 1, 216, 202, 220, 60, 0, 1, 216, 243, 220, 60, 0, 60, 0, 60, 0, 1, 216, 203, 220, 60, 0, 1, 216, 244, 220, 60, 0, 60, 0, 60, 0, 1, 216, 204, 220, 60, 0, 1, 216, 245, 220, 60, 0, 60, 0, 60, 0, 1, 216, 205, 220, 60, 0, 1, 216, 246, 220, 60, 0, 60, 0, 60, 0, 1, 216, 206, 220, 60, 0, 1, 216, 247, 220, 60, 0, 60, 0, 60, 0, 1, 216, 207, 220, 60, 0, 1, 216, 248, 220, 60, 0, 60, 0, 60, 0, 1, 216, 208, 220, 60, 0, 1, 216, 249, 220, 60, 0, 60, 0, 60, 0, 1, 216, 209, 220, 60, 0, 1, 216, 250, 220, 60, 0, 60, 0, 60, 0, 1, 216, 210, 220, 60, 0, 1, 216, 251, 220, 60, 0, 60, 0, 60, 0, 1, 216, 211, 220, 60, 0, 1, 20, 60, 0, 2, 20, 60, 0, 3, 20, 60, 0, 4, 20, 60, 0, 5, 20, 60, 0, 6, 20, 60, 0, 7, 20, 60, 0, 8, 20, 60, 0, 9, 20, 60, 0, 10, 20, 60, 0, 11, 20, 60, 0, 12, 20, 60, 0, 13, 20, 60, 0, 14, 20, 60, 0, 15, 20, 60, 0, 16, 20, 60, 0, 17, 20, 60, 0, 18, 20, 60, 0, 19, 20, 60, 0, 20, 20, 60, 0, 21, 20, 60, 0, 22, 20, 60, 0, 23, 20, 60, 0, 24, 20, 60, 0, 25, 20, 60, 0, 26, 20, 60, 0, 27, 20, 60, 0, 28, 20, 60, 0, 29, 20, 60, 0, 30, 20, 60, 0, 31, 20, 60, 0, 32, 20, 60, 0, 33, 20, 60, 0, 34, 20, 60, 0, 35, 20, 60, 0, 36, 20, 60, 0, 37, 20, 60, 0, 38, 20, 60, 0, 39, 20, 60, 0, 40, 20, 60, 0, 41, 20, 60, 0, 42, 20, 60, 0, 43, 20, 60, 0, 44, 20, 60, 0, 45, 20, 60, 0, 46, 20, 60, 0, 47, 20, 60, 0, 48, 20, 60, 0, 49, 20, 60, 0, 50, 20, 60, 0, 51, 20, 60, 0, 52, 20, 60, 0, 53, 20, 60, 0, 54, 20, 60, 0, 55, 20, 60, 0, 56, 20, 60, 0, 57, 20, 60, 0, 58, 20, 60, 0, 59, 20, 60, 0, 60, 20, 60, 0, 61, 20, 60, 0, 62, 20, 60, 0, 63, 20, 60, 0, 64, 20, 60, 0, 65, 20, 60, 0, 66, 20, 60, 0, 67, 20, 60, 0, 68, 20, 60, 0, 69, 20, 60, 0, 70, 20, 60, 0, 71, 20, 60, 0, 72, 20, 60, 0, 73, 20, 60, 0, 74, 20, 60, 0, 75, 20, 60, 0, 76, 20, 60, 0, 77, 20, 60, 0, 78, 20, 60, 0, 79, 20, 60, 0, 80, 20, 60, 0, 81, 20, 60, 0, 82, 20, 60, 0, 83, 20, 60, 0, 84, 20, 60, 0, 85, 20, 60, 0, 86, 20, 60, 0, 87, 20, 60, 0, 88, 20, 60, 0, 89, 20, 60, 0, 90, 20, 60, 0, 91, 20, 60, 0, 92, 20, 60, 0, 93, 20, 60, 0, 94, 20, 60, 0, 95, 20, 60, 0, 96, 20, 60, 0, 97, 20, 60, 0, 98, 20, 60, 0, 99, 20, 60, 0, 100, 20, 60, 0, 101, 20, 60, 0, 102, 20, 60, 0, 103, 20, 60, 0, 104, 20, 60, 0, 105, 20, 60, 0, 106, 20, 60, 0, 107, 20, 60, 0, 108, 20, 60, 0, 109, 20, 60, 0, 110, 20, 60, 0, 111, 20, 60, 0, 112, 20, 60, 0, 113, 20, 60, 0, 114, 20, 60, 0, 115, 20, 60, 0, 116, 20, 60, 0, 117, 20, 60, 0, 118, 20, 60, 0, 119, 20, 60, 0, 120, 20, 60, 0, 121, 20, 60, 0, 122, 20, 60, 0, 123, 20, 60, 0, 124, 20, 60, 0, 125, 20, 60, 0, 126, 20, 60, 0, 127, 20, 60, 0, 128, 20, 60, 0, 129, 20, 60, 0, 130, 20, 60, 0, 131, 20, 60, 0, 132, 20, 60, 0, 133, 20, 60, 0, 134, 20, 60, 0, 135, 20, 60, 0, 136, 20, 60, 0, 137, 20, 60, 0, 138, 20, 60, 0, 139, 20, 60, 0, 140, 20, 60, 0, 141, 20, 60, 0, 142, 20, 60, 0, 143, 20, 60, 0, 144, 20, 60, 0, 145, 20, 60, 0, 146, 20, 60, 0, 147, 20, 60, 0, 148, 20, 60, 0, 149, 20, 60, 0, 150, 20, 60, 0, 151, 20, 60, 0, 152, 20, 60, 0, 153, 20, 60, 0, 154, 20, 60, 0, 155, 20, 60, 0, 156, 20, 60, 0, 157, 20, 60, 0, 158, 20, 60, 0, 159, 20, 60, 0, 160, 20, 60, 0, 161, 20, 60, 0, 162, 20, 60, 0, 163, 20, 60, 0, 164, 20, 60, 0, 165, 20, 60, 0, 166, 20, 60, 0, 167, 20, 60, 0, 168, 20, 60, 0, 169, 20, 60, 0, 170, 20, 60, 0, 171, 20, 60, 0, 172, 20, 60, 0, 173, 20, 60, 0, 174, 20, 60, 0, 175, 20, 60, 0, 176, 20, 60, 0, 177, 20, 60, 0, 178, 20, 60, 0, 179, 20, 60, 0, 180, 20, 60, 0, 181, 20, 60, 0, 182, 20, 60, 0, 183, 20, 60, 0, 184, 20, 60, 0, 185, 20, 60, 0, 186, 20, 60, 0, 187, 20, 60, 0, 188, 20, 60, 0, 189, 20, 60, 0, 190, 20, 60, 0, 191, 20, 60, 0, 192, 20, 60, 0, 193, 20, 60, 0, 194, 20, 60, 0, 195, 20, 60, 0, 196, 20, 60, 0, 197, 20, 60, 0, 198, 20, 60, 0, 199, 20, 60, 0, 200, 20, 60, 0, 201, 20, 60, 0, 202, 20, 60, 0, 203, 20, 60, 0, 204, 20, 60, 0, 205, 20, 60, 0, 206, 20, 60, 0, 207, 20, 60, 0, 208, 20, 60, 0, 209, 20, 60, 0, 210, 20, 60, 0, 211, 20, 60, 0, 212, 20, 60, 0, 213, 20, 60, 0, 214, 20, 60, 0, 215, 20, 60, 0, 216, 20, 60, 0, 217, 20, 60, 0, 218, 20, 60, 0, 219, 20, 60, 0, 220, 20, 60, 0, 221, 20, 60, 0, 222, 20, 60, 0, 223, 20, 60, 0, 224, 20, 60, 0, 225, 20, 60, 0, 226, 20, 60, 0, 227, 20, 60, 0, 228, 20, 60, 0, 229, 20, 60, 0, 230, 20, 60, 0, 231, 20, 60, 0, 232, 20, 60, 0, 233, 20, 60, 0, 234, 20, 60, 0, 235, 20, 60, 0, 236, 20, 60, 0, 237, 20, 60, 0, 238, 20, 60, 0, 239, 20, 60, 0, 240, 20, 60, 0, 241, 20, 60, 0, 242, 20, 60, 0, 243, 20, 60, 0, 244, 20, 60, 0, 245, 20, 60, 0, 246, 20, 60, 0, 247, 20, 60, 0, 248, 20, 60, 0, 249, 20, 60, 0, 250, 20, 60, 0, 251, 20, 60, 0, 252, 20, 60, 0, 253, 20, 60, 0, 254, 20, 60, 0, 255, 20, 60, 0, 0, 21, 60, 0, 1, 21, 60, 0, 2, 21, 60, 0, 3, 21, 60, 0, 4, 21, 60, 0, 5, 21, 60, 0, 6, 21, 60, 0, 7, 21, 60, 0, 8, 21, 60, 0, 9, 21, 60, 0, 10, 21, 60, 0, 11, 21, 60, 0, 12, 21, 60, 0, 13, 21, 60, 0, 14, 21, 60, 0, 15, 21, 60, 0, 16, 21, 60, 0, 17, 21, 60, 0, 18, 21, 60, 0, 19, 21, 60, 0, 20, 21, 60, 0, 21, 21, 60, 0, 22, 21, 60, 0, 23, 21, 60, 0, 24, 21, 60, 0, 25, 21, 60, 0, 26, 21, 60, 0, 27, 21, 60, 0, 28, 21, 60, 0, 29, 21, 60, 0, 30, 21, 60, 0, 31, 21, 60, 0, 32, 21, 60, 0, 33, 21, 60, 0, 34, 21, 60, 0, 35, 21, 60, 0, 36, 21, 60, 0, 37, 21, 60, 0, 38, 21, 60, 0, 39, 21, 60, 0, 40, 21, 60, 0, 41, 21, 60, 0, 42, 21, 60, 0, 43, 21, 60, 0, 44, 21, 60, 0, 45, 21, 60, 0, 46, 21, 60, 0, 47, 21, 60, 0, 48, 21, 60, 0, 49, 21, 60, 0, 50, 21, 60, 0, 51, 21, 60, 0, 52, 21, 60, 0, 53, 21, 60, 0, 54, 21, 60, 0, 55, 21, 60, 0, 56, 21, 60, 0, 57, 21, 60, 0, 58, 21, 60, 0, 59, 21, 60, 0, 60, 21, 60, 0, 61, 21, 60, 0, 62, 21, 60, 0, 63, 21, 60, 0, 64, 21, 60, 0, 65, 21, 60, 0, 66, 21, 60, 0, 67, 21, 60, 0, 68, 21, 60, 0, 69, 21, 60, 0, 70, 21, 60, 0, 71, 21, 60, 0, 72, 21, 60, 0, 73, 21, 60, 0, 74, 21, 60, 0, 75, 21, 60, 0, 76, 21, 60, 0, 77, 21, 60, 0, 78, 21, 60, 0, 79, 21, 60, 0, 80, 21, 60, 0, 81, 21, 60, 0, 82, 21, 60, 0, 83, 21, 60, 0, 84, 21, 60, 0, 85, 21, 60, 0, 86, 21, 60, 0, 87, 21, 60, 0, 88, 21, 60, 0, 89, 21, 60, 0, 90, 21, 60, 0, 91, 21, 60, 0, 92, 21, 60, 0, 93, 21, 60, 0, 94, 21, 60, 0, 95, 21, 60, 0, 96, 21, 60, 0, 97, 21, 60, 0, 98, 21, 60, 0, 99, 21, 60, 0, 100, 21, 60, 0, 101, 21, 60, 0, 102, 21, 60, 0, 103, 21, 60, 0, 104, 21, 60, 0, 105, 21, 60, 0, 106, 21, 60, 0, 107, 21, 60, 0, 108, 21, 60, 0, 109, 21, 60, 0, 110, 21, 60, 0, 111, 21, 60, 0, 112, 21, 60, 0, 113, 21, 60, 0, 114, 21, 60, 0, 115, 21, 60, 0, 116, 21, 60, 0, 117, 21, 60, 0, 118, 21, 60, 0, 119, 21, 60, 0, 120, 21, 60, 0, 121, 21, 60, 0, 122, 21, 60, 0, 123, 21, 60, 0, 125, 21, 60, 0, 111, 22, 60, 0, 126, 21, 60, 0, 127, 21, 60, 0, 128, 21, 60, 0, 129, 21, 60, 0, 130, 21, 60, 0, 131, 21, 60, 0, 132, 21, 60, 0, 133, 21, 60, 0, 134, 21, 60, 0, 135, 21, 60, 0, 136, 21, 60, 0, 137, 21, 60, 0, 138, 21, 60, 0, 139, 21, 60, 0, 140, 21, 60, 0, 141, 21, 60, 0, 112, 22, 60, 0, 142, 21, 60, 0, 143, 21, 60, 0, 144, 21, 60, 0, 145, 21, 60, 0, 146, 21, 60, 0, 147, 21, 60, 0, 148, 21, 60, 0, 149, 21, 60, 0, 113, 22, 60, 0, 114, 22, 60, 0, 115, 22, 60, 0, 116, 22, 60, 0, 117, 22, 60, 0, 118, 22, 60, 0, 150, 21, 60, 0, 151, 21, 60, 0, 152, 21, 60, 0, 153, 21, 60, 0, 154, 21, 60, 0, 155, 21, 60, 0, 156, 21, 60, 0, 157, 21, 60, 0, 158, 21, 60, 0, 159, 21, 60, 0, 160, 21, 60, 0, 161, 21, 60, 0, 162, 21, 60, 0, 163, 21, 60, 0, 164, 21, 60, 0, 165, 21, 60, 0, 166, 21, 60, 0, 124, 21, 60, 0, 167, 21, 60, 0, 168, 21, 60, 0, 169, 21, 60, 0, 170, 21, 60, 0, 171, 21, 60, 0, 172, 21, 60, 0, 173, 21, 60, 0, 174, 21, 60, 0, 175, 21, 60, 0, 176, 21, 60, 0, 177, 21, 60, 0, 178, 21, 60, 0, 179, 21, 60, 0, 180, 21, 60, 0, 181, 21, 60, 0, 182, 21, 60, 0, 183, 21, 60, 0, 184, 21, 60, 0, 185, 21, 60, 0, 186, 21, 60, 0, 187, 21, 60, 0, 188, 21, 60, 0, 189, 21, 60, 0, 190, 21, 60, 0, 191, 21, 60, 0, 192, 21, 60, 0, 193, 21, 60, 0, 194, 21, 60, 0, 195, 21, 60, 0, 196, 21, 60, 0, 197, 21, 60, 0, 198, 21, 60, 0, 199, 21, 60, 0, 200, 21, 60, 0, 201, 21, 60, 0, 202, 21, 60, 0, 203, 21, 60, 0, 204, 21, 60, 0, 205, 21, 60, 0, 206, 21, 60, 0, 207, 21, 60, 0, 208, 21, 60, 0, 209, 21, 60, 0, 210, 21, 60, 0, 211, 21, 60, 0, 212, 21, 60, 0, 213, 21, 60, 0, 214, 21, 60, 0, 215, 21, 60, 0, 216, 21, 60, 0, 217, 21, 60, 0, 218, 21, 60, 0, 219, 21, 60, 0, 220, 21, 60, 0, 221, 21, 60, 0, 222, 21, 60, 0, 223, 21, 60, 0, 224, 21, 60, 0, 225, 21, 60, 0, 226, 21, 60, 0, 227, 21, 60, 0, 228, 21, 60, 0, 229, 21, 60, 0, 230, 21, 60, 0, 231, 21, 60, 0, 232, 21, 60, 0, 233, 21, 60, 0, 234, 21, 60, 0, 235, 21, 60, 0, 236, 21, 60, 0, 237, 21, 60, 0, 238, 21, 60, 0, 239, 21, 60, 0, 240, 21, 60, 0, 241, 21, 60, 0, 242, 21, 60, 0, 243, 21, 60, 0, 244, 21, 60, 0, 245, 21, 60, 0, 246, 21, 60, 0, 247, 21, 60, 0, 248, 21, 60, 0, 249, 21, 60, 0, 250, 21, 60, 0, 251, 21, 60, 0, 252, 21, 60, 0, 253, 21, 60, 0, 254, 21, 60, 0, 255, 21, 60, 0, 0, 22, 60, 0, 1, 22, 60, 0, 2, 22, 60, 0, 3, 22, 60, 0, 4, 22, 60, 0, 5, 22, 60, 0, 6, 22, 60, 0, 7, 22, 60, 0, 8, 22, 60, 0, 9, 22, 60, 0, 10, 22, 60, 0, 11, 22, 60, 0, 12, 22, 60, 0, 13, 22, 60, 0, 14, 22, 60, 0, 15, 22, 60, 0, 16, 22, 60, 0, 17, 22, 60, 0, 18, 22, 60, 0, 19, 22, 60, 0, 20, 22, 60, 0, 21, 22, 60, 0, 22, 22, 60, 0, 23, 22, 60, 0, 24, 22, 60, 0, 25, 22, 60, 0, 26, 22, 60, 0, 27, 22, 60, 0, 28, 22, 60, 0, 29, 22, 60, 0, 30, 22, 60, 0, 31, 22, 60, 0, 32, 22, 60, 0, 33, 22, 60, 0, 34, 22, 60, 0, 35, 22, 60, 0, 36, 22, 60, 0, 37, 22, 60, 0, 38, 22, 60, 0, 39, 22, 60, 0, 40, 22, 60, 0, 41, 22, 60, 0, 42, 22, 60, 0, 43, 22, 60, 0, 44, 22, 60, 0, 45, 22, 60, 0, 46, 22, 60, 0, 47, 22, 60, 0, 48, 22, 60, 0, 49, 22, 60, 0, 50, 22, 60, 0, 51, 22, 60, 0, 52, 22, 60, 0, 53, 22, 60, 0, 54, 22, 60, 0, 55, 22, 60, 0, 56, 22, 60, 0, 57, 22, 60, 0, 58, 22, 60, 0, 59, 22, 60, 0, 60, 22, 60, 0, 61, 22, 60, 0, 62, 22, 60, 0, 63, 22, 60, 0, 64, 22, 60, 0, 65, 22, 60, 0, 66, 22, 60, 0, 67, 22, 60, 0, 68, 22, 60, 0, 69, 22, 60, 0, 70, 22, 60, 0, 71, 22, 60, 0, 72, 22, 60, 0, 73, 22, 60, 0, 74, 22, 60, 0, 75, 22, 60, 0, 76, 22, 60, 0, 77, 22, 60, 0, 78, 22, 60, 0, 79, 22, 60, 0, 80, 22, 60, 0, 81, 22, 60, 0, 82, 22, 60, 0, 83, 22, 60, 0, 84, 22, 60, 0, 85, 22, 60, 0, 86, 22, 60, 0, 87, 22, 60, 0, 88, 22, 60, 0, 89, 22, 60, 0, 90, 22, 60, 0, 91, 22, 60, 0, 92, 22, 60, 0, 93, 22, 60, 0, 94, 22, 60, 0, 95, 22, 60, 0, 96, 22, 60, 0, 97, 22, 60, 0, 98, 22, 60, 0, 99, 22, 60, 0, 100, 22, 60, 0, 101, 22, 60, 0, 102, 22, 60, 0, 103, 22, 60, 0, 104, 22, 60, 0, 105, 22, 60, 0, 106, 22, 60, 0, 107, 22, 60, 0, 108, 22, 60, 0, 119, 22, 60, 0, 120, 22, 60, 0, 121, 22, 60, 0, 122, 22, 60, 0, 123, 22, 60, 0, 124, 22, 60, 0, 125, 22, 60, 0, 126, 22, 60, 0, 127, 22, 60, 0, 176, 24, 60, 0, 177, 24, 60, 0, 178, 24, 60, 0, 179, 24, 60, 0, 180, 24, 60, 0, 181, 24, 60, 0, 182, 24, 60, 0, 183, 24, 60, 0, 184, 24, 60, 0, 185, 24, 60, 0, 186, 24, 60, 0, 187, 24, 60, 0, 188, 24, 60, 0, 189, 24, 60, 0, 190, 24, 60, 0, 191, 24, 60, 0, 192, 24, 60, 0, 193, 24, 60, 0, 194, 24, 60, 0, 195, 24, 60, 0, 196, 24, 60, 0, 197, 24, 60, 0, 198, 24, 60, 0, 199, 24, 60, 0, 200, 24, 60, 0, 201, 24, 60, 0, 202, 24, 60, 0, 203, 24, 60, 0, 204, 24, 60, 0, 205, 24, 60, 0, 206, 24, 60, 0, 207, 24, 60, 0, 208, 24, 60, 0, 209, 24, 60, 0, 210, 24, 60, 0, 211, 24, 60, 0, 212, 24, 60, 0, 213, 24, 60, 0, 214, 24, 60, 0, 215, 24, 60, 0, 216, 24, 60, 0, 217, 24, 60, 0, 218, 24, 60, 0, 219, 24, 60, 0, 220, 24, 60, 0, 221, 24, 60, 0, 222, 24, 60, 0, 223, 24, 60, 0, 224, 24, 60, 0, 225, 24, 60, 0, 226, 24, 60, 0, 227, 24, 60, 0, 228, 24, 60, 0, 229, 24, 60, 0, 230, 24, 60, 0, 231, 24, 60, 0, 232, 24, 60, 0, 233, 24, 60, 0, 234, 24, 60, 0, 235, 24, 60, 0, 236, 24, 60, 0, 237, 24, 60, 0, 238, 24, 60, 0, 239, 24, 60, 0, 240, 24, 60, 0, 241, 24, 60, 0, 242, 24, 60, 0, 243, 24, 60, 0, 244, 24, 60, 0, 245, 24, 60, 0, 129, 22, 60, 0, 130, 22, 60, 0, 131, 22, 60, 0, 132, 22, 60, 0, 133, 22, 60, 0, 134, 22, 60, 0, 135, 22, 60, 0, 136, 22, 60, 0, 137, 22, 60, 0, 138, 22, 60, 0, 139, 22, 60, 0, 140, 22, 60, 0, 141, 22, 60, 0, 142, 22, 60, 0, 143, 22, 60, 0, 144, 22, 60, 0, 145, 22, 60, 0, 146, 22, 60, 0, 147, 22, 60, 0, 148, 22, 60, 0, 149, 22, 60, 0, 150, 22, 60, 0, 151, 22, 60, 0, 152, 22, 60, 0, 153, 22, 60, 0, 154, 22, 60, 0, 160, 22, 60, 0, 60, 0, 161, 22, 60, 0, 162, 22, 60, 0, 60, 0, 164, 22, 60, 0, 60, 0, 165, 22, 60, 0, 166, 22, 60, 0, 60, 0, 60, 0, 240, 22, 47, 0, 166, 22, 60, 0, 60, 0, 167, 22, 60, 0, 168, 22, 60, 0, 60, 0, 169, 22, 60, 0, 60, 0, 172, 22, 60, 0, 60, 0, 173, 22, 60, 0, 60, 0, 174, 22, 60, 0, 244, 22, 60, 0, 175, 22, 60, 0, 176, 22, 60, 0, 177, 22, 60, 0, 178, 22, 60, 0, 60, 0, 179, 22, 60, 0, 60, 0, 180, 22, 60, 0, 60, 0, 181, 22, 60, 0, 60, 0, 182, 22, 60, 0, 241, 22, 60, 0, 183, 22, 60, 0, 185, 22, 60, 0, 60, 0, 233, 22, 60, 0, 186, 22, 60, 0, 60, 0, 187, 22, 60, 0, 60, 0, 188, 22, 60, 0, 60, 0, 189, 22, 60, 0, 190, 22, 60, 0, 60, 0, 191, 22, 60, 0, 60, 0, 192, 22, 60, 0, 193, 22, 60, 0, 60, 0, 194, 22, 60, 0, 245, 22, 60, 0, 195, 22, 60, 0, 60, 0, 196, 22, 60, 0, 197, 22, 60, 0, 60, 0, 60, 0, 238, 22, 47, 0, 218, 22, 60, 0, 60, 0, 198, 22, 60, 0, 199, 22, 60, 0, 200, 22, 60, 0, 60, 0, 213, 22, 60, 0, 201, 22, 60, 0, 202, 22, 60, 0, 60, 0, 203, 22, 60, 0, 60, 0, 234, 22, 60, 0, 60, 0, 204, 22, 60, 0, 60, 0, 205, 22, 60, 0, 60, 0, 206, 22, 60, 0, 242, 22, 60, 0, 207, 22, 60, 0, 60, 0, 208, 22, 60, 0, 60, 0, 209, 22, 60, 0, 210, 22, 60, 0, 60, 0, 211, 22, 60, 0, 60, 0, 212, 22, 60, 0, 214, 22, 60, 0, 246, 22, 60, 0, 215, 22, 60, 0, 60, 0, 216, 22, 60, 0, 60, 0, 60, 0, 239, 22, 47, 0, 215, 22, 60, 0, 60, 0, 217, 22, 60, 0, 218, 22, 60, 0, 60, 0, 219, 22, 60, 0, 220, 22, 60, 0, 60, 0, 221, 22, 60, 0, 222, 22, 60, 0, 223, 22, 60, 0, 243, 22, 60, 0, 170, 22, 60, 0, 247, 22, 60, 0, 171, 22, 60, 0, 248, 22, 60, 0, 163, 22, 60, 0, 224, 22, 60, 0, 227, 22, 60, 0, 184, 22, 60, 0, 228, 22, 60, 0, 225, 22, 60, 0, 226, 22, 60, 0, 229, 22, 60, 0, 230, 22, 60, 0, 60, 0, 231, 22, 60, 0, 60, 0, 232, 22, 60, 0, 14, 32, 3, 216, 192, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 128, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 193, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 129, 220, 14, 32, 60, 0, 14, 32, 3, 216, 194, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 130, 220, 14, 32, 60, 0, 14, 32, 3, 216, 195, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 131, 220, 14, 32, 60, 0, 14, 32, 3, 216, 196, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 132, 220, 14, 32, 60, 0, 14, 32, 3, 216, 197, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 133, 220, 14, 32, 60, 0, 14, 32, 3, 216, 198, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 134, 220, 14, 32, 60, 0, 14, 32, 3, 216, 199, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 135, 220, 14, 32, 60, 0, 14, 32, 3, 216, 200, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 136, 220, 14, 32, 60, 0, 14, 32, 3, 216, 201, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 137, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 202, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 138, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 203, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 139, 220, 14, 32, 60, 0, 14, 32, 3, 216, 204, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 140, 220, 14, 32, 60, 0, 14, 32, 3, 216, 205, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 141, 220, 14, 32, 60, 0, 14, 32, 3, 216, 206, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 142, 220, 14, 32, 60, 0, 14, 32, 3, 216, 207, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 143, 220, 14, 32, 60, 0, 14, 32, 3, 216, 208, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 144, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 209, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 145, 220, 14, 32, 60, 0, 14, 32, 3, 216, 210, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 146, 220, 14, 32, 60, 0, 14, 32, 3, 216, 211, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 147, 220, 14, 32, 60, 0, 14, 32, 3, 216, 212, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 148, 220, 14, 32, 60, 0, 14, 32, 3, 216, 213, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 149, 220, 14, 32, 60, 0, 14, 32, 3, 216, 214, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 150, 220, 14, 32, 60, 0, 14, 32, 3, 216, 215, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 151, 220, 14, 32, 60, 0, 14, 32, 3, 216, 216, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 152, 220, 14, 32, 60, 0, 14, 32, 3, 216, 217, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 153, 220, 14, 32, 60, 0, 14, 32, 3, 216, 218, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 154, 220, 14, 32, 60, 0, 14, 32, 3, 216, 219, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 155, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 220, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 156, 220, 14, 32, 60, 0, 14, 32, 3, 216, 221, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 157, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 222, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 158, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 223, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 159, 220, 14, 32, 60, 0, 14, 32, 3, 216, 224, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 160, 220, 14, 32, 60, 0, 14, 32, 3, 216, 225, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 161, 220, 14, 32, 60, 0, 14, 32, 3, 216, 226, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 162, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 227, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 163, 220, 14, 32, 60, 0, 14, 32, 3, 216, 228, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 164, 220, 14, 32, 60, 0, 14, 32, 3, 216, 229, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 165, 220, 14, 32, 60, 0, 14, 32, 3, 216, 230, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 166, 220, 14, 32, 60, 0, 14, 32, 3, 216, 231, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 167, 220, 14, 32, 60, 0, 14, 32, 3, 216, 232, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 168, 220, 14, 32, 60, 0, 14, 32, 3, 216, 233, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 169, 220, 14, 32, 60, 0, 14, 32, 3, 216, 234, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 170, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 235, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 171, 220, 14, 32, 60, 0, 14, 32, 3, 216, 236, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 172, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 237, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 173, 220, 14, 32, 60, 0, 14, 32, 3, 216, 238, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 174, 220, 14, 32, 60, 0, 14, 32, 3, 216, 239, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 175, 220, 14, 32, 60, 0, 14, 32, 3, 216, 240, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 176, 220, 14, 32, 60, 0, 14, 32, 3, 216, 241, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 177, 220, 14, 32, 60, 0, 14, 32, 3, 216, 242, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 178, 220, 14, 32, 60, 0, 14, 32, 3, 216, 0, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 1, 220, 14, 32, 60, 0, 14, 32, 3, 216, 2, 220, 14, 32, 60, 0, 14, 32, 3, 216, 3, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 4, 220, 14, 32, 60, 0, 14, 32, 3, 216, 5, 220, 14, 32, 60, 0, 14, 32, 3, 216, 6, 220, 14, 32, 60, 0, 14, 32, 3, 216, 7, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 8, 220, 14, 32, 60, 0, 14, 32, 3, 216, 9, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 10, 220, 14, 32, 60, 0, 14, 32, 3, 216, 11, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 12, 220, 14, 32, 60, 0, 14, 32, 3, 216, 13, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 14, 220, 14, 32, 60, 0, 14, 32, 3, 216, 15, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 16, 220, 14, 32, 60, 0, 14, 32, 3, 216, 17, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 18, 220, 14, 32, 60, 0, 14, 32, 3, 216, 19, 220, 14, 32, 60, 0, 14, 32, 3, 216, 20, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 21, 220, 14, 32, 60, 0, 14, 32, 3, 216, 22, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 23, 220, 14, 32, 60, 0, 14, 32, 3, 216, 24, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 25, 220, 14, 32, 60, 0, 14, 32, 3, 216, 26, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 27, 220, 14, 32, 60, 0, 14, 32, 3, 216, 28, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 29, 220, 14, 32, 60, 0, 14, 32, 3, 216, 30, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 31, 220, 14, 32, 60, 0, 14, 32, 3, 216, 32, 220, 14, 32, 60, 0, 14, 32, 3, 216, 33, 220, 14, 32, 60, 0, 14, 32, 3, 216, 34, 220, 14, 32, 60, 0, 14, 32, 3, 216, 35, 220, 14, 32, 60, 0, 14, 32, 3, 216, 36, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 37, 220, 14, 32, 60, 0, 14, 32, 3, 216, 38, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 39, 220, 14, 32, 60, 0, 14, 32, 3, 216, 40, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 41, 220, 14, 32, 60, 0, 14, 32, 3, 216, 42, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 43, 220, 14, 32, 60, 0, 14, 32, 3, 216, 44, 220, 14, 32, 60, 0, 14, 32, 3, 216, 45, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 46, 220, 14, 32, 60, 0, 14, 32, 3, 216, 47, 220, 14, 32, 60, 0, 14, 32, 3, 216, 48, 220, 14, 32, 60, 0, 14, 32, 3, 216, 49, 220, 14, 32, 60, 0, 14, 32, 3, 216, 50, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 51, 220, 14, 32, 60, 0, 14, 32, 3, 216, 52, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 53, 220, 14, 32, 60, 0, 14, 32, 3, 216, 54, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 55, 220, 14, 32, 60, 0, 14, 32, 3, 216, 56, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 57, 220, 14, 32, 60, 0, 14, 32, 3, 216, 58, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 59, 220, 14, 32, 60, 0, 14, 32, 3, 216, 60, 220, 14, 32, 60, 0, 14, 32, 3, 216, 61, 220, 14, 32, 60, 0, 14, 32, 3, 216, 62, 220, 14, 32, 60, 0, 14, 32, 3, 216, 63, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 64, 220, 14, 32, 60, 0, 14, 32, 3, 216, 65, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 66, 220, 14, 32, 60, 0, 14, 32, 3, 216, 67, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 68, 220, 14, 32, 60, 0, 14, 32, 3, 216, 69, 220, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 70, 220, 14, 32, 60, 0, 14, 32, 3, 216, 71, 220, 14, 32, 60, 0, 14, 32, 3, 216, 72, 220, 14, 32, 60, 0, 0, 165, 60, 0, 1, 165, 60, 0, 2, 165, 60, 0, 3, 165, 60, 0, 4, 165, 60, 0, 5, 165, 60, 0, 6, 165, 60, 0, 7, 165, 60, 0, 8, 165, 60, 0, 9, 165, 60, 0, 10, 165, 60, 0, 11, 165, 60, 0, 12, 165, 60, 0, 60, 0, 60, 0, 19, 166, 47, 0, 11, 166, 60, 0, 13, 165, 60, 0, 14, 165, 60, 0, 15, 165, 60, 0, 16, 165, 60, 0, 17, 165, 60, 0, 18, 165, 60, 0, 19, 165, 60, 0, 20, 165, 60, 0, 21, 165, 60, 0, 22, 165, 60, 0, 23, 165, 60, 0, 24, 165, 60, 0, 25, 165, 60, 0, 26, 165, 60, 0, 27, 165, 60, 0, 28, 165, 60, 0, 29, 165, 60, 0, 30, 165, 60, 0, 60, 0, 60, 0, 20, 166, 47, 0, 11, 166, 60, 0, 31, 165, 60, 0, 32, 165, 60, 0, 33, 165, 60, 0, 34, 165, 60, 0, 35, 165, 60, 0, 36, 165, 60, 0, 37, 165, 60, 0, 38, 165, 60, 0, 39, 165, 60, 0, 40, 165, 60, 0, 41, 165, 60, 0, 42, 165, 60, 0, 43, 165, 60, 0, 44, 165, 60, 0, 45, 165, 60, 0, 46, 165, 60, 0, 47, 165, 60, 0, 48, 165, 60, 0, 49, 165, 60, 0, 50, 165, 60, 0, 51, 165, 60, 0, 60, 0, 60, 0, 21, 166, 47, 0, 11, 166, 60, 0, 52, 165, 60, 0, 53, 165, 60, 0, 54, 165, 60, 0, 55, 165, 60, 0, 56, 165, 60, 0, 57, 165, 60, 0, 58, 165, 60, 0, 59, 165, 60, 0, 60, 165, 60, 0, 61, 165, 60, 0, 62, 165, 60, 0, 63, 165, 60, 0, 64, 165, 60, 0, 65, 165, 60, 0, 66, 165, 60, 0, 67, 165, 60, 0, 68, 165, 60, 0, 69, 165, 60, 0, 70, 165, 60, 0, 71, 165, 60, 0, 60, 0, 60, 0, 22, 166, 47, 0, 12, 166, 60, 0, 72, 165, 60, 0, 73, 165, 60, 0, 74, 165, 60, 0, 75, 165, 60, 0, 76, 165, 60, 0, 77, 165, 60, 0, 78, 165, 60, 0, 79, 165, 60, 0, 80, 165, 60, 0, 81, 165, 60, 0, 82, 165, 60, 0, 60, 0, 60, 0, 23, 166, 47, 0, 11, 166, 60, 0, 83, 165, 60, 0, 84, 165, 60, 0, 85, 165, 60, 0, 86, 165, 60, 0, 87, 165, 60, 0, 88, 165, 60, 0, 60, 0, 60, 0, 16, 166, 60, 0, 60, 0, 60, 0, 24, 166, 47, 0, 12, 166, 60, 0, 89, 165, 60, 0, 90, 165, 60, 0, 60, 0, 60, 0, 25, 166, 47, 0, 12, 166, 60, 0, 91, 165, 60, 0, 92, 165, 60, 0, 93, 165, 60, 0, 94, 165, 60, 0, 95, 165, 60, 0, 96, 165, 60, 0, 60, 0, 60, 0, 26, 166, 47, 0, 11, 166, 60, 0, 97, 165, 60, 0, 98, 165, 60, 0, 99, 165, 60, 0, 100, 165, 60, 0, 101, 165, 60, 0, 102, 165, 60, 0, 103, 165, 60, 0, 104, 165, 60, 0, 105, 165, 60, 0, 106, 165, 60, 0, 60, 0, 60, 0, 17, 166, 60, 0, 107, 165, 60, 0, 108, 165, 60, 0, 109, 165, 60, 0, 110, 165, 60, 0, 60, 0, 60, 0, 42, 166, 60, 0, 111, 165, 60, 0, 112, 165, 60, 0, 113, 165, 60, 0, 114, 165, 60, 0, 115, 165, 60, 0, 116, 165, 60, 0, 117, 165, 60, 0, 118, 165, 60, 0, 119, 165, 60, 0, 120, 165, 60, 0, 121, 165, 60, 0, 122, 165, 60, 0, 123, 165, 60, 0, 124, 165, 60, 0, 125, 165, 60, 0, 126, 165, 60, 0, 127, 165, 60, 0, 128, 165, 60, 0, 129, 165, 60, 0, 130, 165, 60, 0, 131, 165, 60, 0, 132, 165, 60, 0, 133, 165, 60, 0, 60, 0, 60, 0, 27, 166, 47, 0, 11, 166, 60, 0, 134, 165, 60, 0, 135, 165, 60, 0, 60, 0, 60, 0, 18, 166, 60, 0, 136, 165, 60, 0, 137, 165, 60, 0, 138, 165, 60, 0, 139, 165, 60, 0, 140, 165, 60, 0, 141, 165, 60, 0, 142, 165, 60, 0, 143, 165, 60, 0, 144, 165, 60, 0, 145, 165, 60, 0, 146, 165, 60, 0, 147, 165, 60, 0, 148, 165, 60, 0, 149, 165, 60, 0, 150, 165, 60, 0, 151, 165, 60, 0, 152, 165, 60, 0, 153, 165, 60, 0, 154, 165, 60, 0, 155, 165, 60, 0, 156, 165, 60, 0, 157, 165, 60, 0, 158, 165, 60, 0, 159, 165, 60, 0, 160, 165, 60, 0, 161, 165, 60, 0, 162, 165, 60, 0, 163, 165, 60, 0, 164, 165, 60, 0, 165, 165, 60, 0, 166, 165, 60, 0, 167, 165, 60, 0, 168, 165, 60, 0, 169, 165, 60, 0, 170, 165, 60, 0, 171, 165, 60, 0, 172, 165, 60, 0, 173, 165, 60, 0, 174, 165, 60, 0, 175, 165, 60, 0, 176, 165, 60, 0, 177, 165, 60, 0, 178, 165, 60, 0, 179, 165, 60, 0, 180, 165, 60, 0, 60, 0, 60, 0, 28, 166, 47, 0, 11, 166, 60, 0, 181, 165, 60, 0, 182, 165, 60, 0, 183, 165, 60, 0, 184, 165, 60, 0, 185, 165, 60, 0, 186, 165, 60, 0, 187, 165, 60, 0, 188, 165, 60, 0, 189, 165, 60, 0, 190, 165, 60, 0, 191, 165, 60, 0, 192, 165, 60, 0, 193, 165, 60, 0, 194, 165, 60, 0, 195, 165, 60, 0, 196, 165, 60, 0, 197, 165, 60, 0, 198, 165, 60, 0, 199, 165, 60, 0, 200, 165, 60, 0, 201, 165, 60, 0, 202, 165, 60, 0, 203, 165, 60, 0, 60, 0, 60, 0, 29, 166, 47, 0, 11, 166, 60, 0, 204, 165, 60, 0, 205, 165, 60, 0, 206, 165, 60, 0, 207, 165, 60, 0, 208, 165, 60, 0, 209, 165, 60, 0, 60, 0, 60, 0, 43, 166, 60, 0, 60, 0, 60, 0, 30, 166, 47, 0, 12, 166, 60, 0, 210, 165, 60, 0, 211, 165, 60, 0, 212, 165, 60, 0, 213, 165, 60, 0, 214, 165, 60, 0, 215, 165, 60, 0, 216, 165, 60, 0, 60, 0, 60, 0, 31, 166, 47, 0, 11, 166, 60, 0, 217, 165, 60, 0, 218, 165, 60, 0, 219, 165, 60, 0, 220, 165, 60, 0, 221, 165, 60, 0, 222, 165, 60, 0, 223, 165, 60, 0, 224, 165, 60, 0, 225, 165, 60, 0, 226, 165, 60, 0, 227, 165, 60, 0, 228, 165, 60, 0, 229, 165, 60, 0, 230, 165, 60, 0, 231, 165, 60, 0, 232, 165, 60, 0, 233, 165, 60, 0, 234, 165, 60, 0, 235, 165, 60, 0, 236, 165, 60, 0, 237, 165, 60, 0, 238, 165, 60, 0, 239, 165, 60, 0, 240, 165, 60, 0, 241, 165, 60, 0, 242, 165, 60, 0, 243, 165, 60, 0, 244, 165, 60, 0, 245, 165, 60, 0, 246, 165, 60, 0, 247, 165, 60, 0, 248, 165, 60, 0, 249, 165, 60, 0, 250, 165, 60, 0, 251, 165, 60, 0, 252, 165, 60, 0, 253, 165, 60, 0, 254, 165, 60, 0, 255, 165, 60, 0, 0, 166, 60, 0, 1, 166, 60, 0, 2, 166, 60, 0, 3, 166, 60, 0, 4, 166, 60, 0, 5, 166, 60, 0, 6, 166, 60, 0, 7, 166, 60, 0, 8, 166, 60, 0, 9, 166, 60, 0, 10, 166, 60, 0, 11, 166, 60, 0, 12, 166, 60, 0, 160, 166, 60, 0, 161, 166, 60, 0, 162, 166, 60, 0, 163, 166, 60, 0, 164, 166, 60, 0, 165, 166, 60, 0, 166, 166, 60, 0, 167, 166, 60, 0, 168, 166, 60, 0, 169, 166, 60, 0, 170, 166, 60, 0, 171, 166, 60, 0, 172, 166, 60, 0, 173, 166, 60, 0, 174, 166, 60, 0, 175, 166, 60, 0, 176, 166, 60, 0, 177, 166, 60, 0, 178, 166, 60, 0, 179, 166, 60, 0, 180, 166, 60, 0, 181, 166, 60, 0, 182, 166, 60, 0, 183, 166, 60, 0, 184, 166, 60, 0, 185, 166, 60, 0, 186, 166, 60, 0, 187, 166, 60, 0, 188, 166, 60, 0, 189, 166, 60, 0, 190, 166, 60, 0, 191, 166, 60, 0, 192, 166, 60, 0, 193, 166, 60, 0, 194, 166, 60, 0, 195, 166, 60, 0, 196, 166, 60, 0, 197, 166, 60, 0, 198, 166, 60, 0, 199, 166, 60, 0, 200, 166, 60, 0, 201, 166, 60, 0, 202, 166, 60, 0, 203, 166, 60, 0, 204, 166, 60, 0, 205, 166, 60, 0, 206, 166, 60, 0, 207, 166, 60, 0, 208, 166, 60, 0, 209, 166, 60, 0, 210, 166, 60, 0, 211, 166, 60, 0, 212, 166, 60, 0, 213, 166, 60, 0, 214, 166, 60, 0, 215, 166, 60, 0, 216, 166, 60, 0, 217, 166, 60, 0, 218, 166, 60, 0, 219, 166, 60, 0, 220, 166, 60, 0, 221, 166, 60, 0, 222, 166, 60, 0, 223, 166, 60, 0, 224, 166, 60, 0, 225, 166, 60, 0, 226, 166, 60, 0, 227, 166, 60, 0, 228, 166, 60, 0, 229, 166, 60, 0, 230, 166, 60, 0, 231, 166, 60, 0, 232, 166, 60, 0, 233, 166, 60, 0, 234, 166, 60, 0, 235, 166, 60, 0, 236, 166, 60, 0, 237, 166, 60, 0, 238, 166, 60, 0, 239, 166, 60, 0, 26, 216, 0, 220, 60, 0, 26, 216, 1, 220, 60, 0, 26, 216, 2, 220, 60, 0, 26, 216, 3, 220, 60, 0, 26, 216, 4, 220, 60, 0, 26, 216, 5, 220, 60, 0, 26, 216, 6, 220, 60, 0, 26, 216, 7, 220, 60, 0, 26, 216, 8, 220, 60, 0, 26, 216, 9, 220, 60, 0, 26, 216, 10, 220, 60, 0, 26, 216, 11, 220, 60, 0, 26, 216, 12, 220, 60, 0, 26, 216, 13, 220, 60, 0, 26, 216, 14, 220, 60, 0, 26, 216, 15, 220, 60, 0, 26, 216, 16, 220, 60, 0, 26, 216, 17, 220, 60, 0, 26, 216, 18, 220, 60, 0, 26, 216, 19, 220, 60, 0, 26, 216, 20, 220, 60, 0, 26, 216, 21, 220, 60, 0, 26, 216, 22, 220, 60, 0, 26, 216, 23, 220, 60, 0, 26, 216, 24, 220, 60, 0, 26, 216, 25, 220, 60, 0, 26, 216, 26, 220, 60, 0, 26, 216, 27, 220, 60, 0, 26, 216, 28, 220, 60, 0, 26, 216, 29, 220, 60, 0, 26, 216, 30, 220, 60, 0, 26, 216, 31, 220, 60, 0, 26, 216, 32, 220, 60, 0, 26, 216, 33, 220, 60, 0, 26, 216, 34, 220, 60, 0, 26, 216, 35, 220, 60, 0, 26, 216, 36, 220, 60, 0, 26, 216, 37, 220, 60, 0, 26, 216, 38, 220, 60, 0, 26, 216, 39, 220, 60, 0, 26, 216, 40, 220, 60, 0, 26, 216, 41, 220, 60, 0, 26, 216, 42, 220, 60, 0, 26, 216, 43, 220, 60, 0, 26, 216, 44, 220, 60, 0, 26, 216, 45, 220, 60, 0, 26, 216, 46, 220, 60, 0, 26, 216, 47, 220, 60, 0, 26, 216, 48, 220, 60, 0, 26, 216, 49, 220, 60, 0, 26, 216, 50, 220, 60, 0, 26, 216, 51, 220, 60, 0, 26, 216, 52, 220, 60, 0, 26, 216, 53, 220, 60, 0, 26, 216, 54, 220, 60, 0, 26, 216, 55, 220, 60, 0, 26, 216, 56, 220, 60, 0, 26, 216, 57, 220, 60, 0, 26, 216, 58, 220, 60, 0, 26, 216, 59, 220, 60, 0, 26, 216, 60, 220, 60, 0, 26, 216, 61, 220, 60, 0, 26, 216, 62, 220, 60, 0, 26, 216, 63, 220, 60, 0, 26, 216, 64, 220, 60, 0, 26, 216, 65, 220, 60, 0, 26, 216, 66, 220, 60, 0, 26, 216, 67, 220, 60, 0, 26, 216, 68, 220, 60, 0, 26, 216, 69, 220, 60, 0, 26, 216, 70, 220, 60, 0, 26, 216, 71, 220, 60, 0, 26, 216, 72, 220, 60, 0, 26, 216, 73, 220, 60, 0, 26, 216, 74, 220, 60, 0, 26, 216, 75, 220, 60, 0, 26, 216, 76, 220, 60, 0, 26, 216, 77, 220, 60, 0, 26, 216, 78, 220, 60, 0, 26, 216, 79, 220, 60, 0, 26, 216, 80, 220, 60, 0, 26, 216, 81, 220, 60, 0, 26, 216, 82, 220, 60, 0, 26, 216, 83, 220, 60, 0, 26, 216, 84, 220, 60, 0, 26, 216, 85, 220, 60, 0, 26, 216, 86, 220, 60, 0, 26, 216, 87, 220, 60, 0, 26, 216, 88, 220, 60, 0, 26, 216, 89, 220, 60, 0, 26, 216, 90, 220, 60, 0, 26, 216, 91, 220, 60, 0, 26, 216, 92, 220, 60, 0, 26, 216, 93, 220, 60, 0, 26, 216, 94, 220, 60, 0, 26, 216, 95, 220, 60, 0, 26, 216, 96, 220, 60, 0, 26, 216, 97, 220, 60, 0, 26, 216, 98, 220, 60, 0, 26, 216, 99, 220, 60, 0, 26, 216, 100, 220, 60, 0, 26, 216, 101, 220, 60, 0, 26, 216, 102, 220, 60, 0, 26, 216, 103, 220, 60, 0, 26, 216, 104, 220, 60, 0, 26, 216, 105, 220, 60, 0, 26, 216, 106, 220, 60, 0, 26, 216, 107, 220, 60, 0, 26, 216, 108, 220, 60, 0, 26, 216, 109, 220, 60, 0, 26, 216, 110, 220, 60, 0, 26, 216, 111, 220, 60, 0, 26, 216, 112, 220, 60, 0, 26, 216, 113, 220, 60, 0, 26, 216, 114, 220, 60, 0, 26, 216, 115, 220, 60, 0, 26, 216, 116, 220, 60, 0, 26, 216, 117, 220, 60, 0, 26, 216, 118, 220, 60, 0, 26, 216, 119, 220, 60, 0, 26, 216, 120, 220, 60, 0, 26, 216, 121, 220, 60, 0, 26, 216, 122, 220, 60, 0, 26, 216, 123, 220, 60, 0, 26, 216, 124, 220, 60, 0, 26, 216, 125, 220, 60, 0, 26, 216, 126, 220, 60, 0, 26, 216, 127, 220, 60, 0, 26, 216, 128, 220, 60, 0, 26, 216, 129, 220, 60, 0, 26, 216, 130, 220, 60, 0, 26, 216, 131, 220, 60, 0, 26, 216, 132, 220, 60, 0, 26, 216, 133, 220, 60, 0, 26, 216, 134, 220, 60, 0, 26, 216, 135, 220, 60, 0, 26, 216, 136, 220, 60, 0, 26, 216, 137, 220, 60, 0, 26, 216, 138, 220, 60, 0, 26, 216, 139, 220, 60, 0, 26, 216, 140, 220, 60, 0, 26, 216, 141, 220, 60, 0, 26, 216, 142, 220, 60, 0, 26, 216, 143, 220, 60, 0, 26, 216, 144, 220, 60, 0, 26, 216, 145, 220, 60, 0, 26, 216, 146, 220, 60, 0, 26, 216, 147, 220, 60, 0, 26, 216, 148, 220, 60, 0, 26, 216, 149, 220, 60, 0, 26, 216, 150, 220, 60, 0, 26, 216, 151, 220, 60, 0, 26, 216, 152, 220, 60, 0, 26, 216, 153, 220, 60, 0, 26, 216, 154, 220, 60, 0, 26, 216, 155, 220, 60, 0, 26, 216, 156, 220, 60, 0, 26, 216, 157, 220, 60, 0, 26, 216, 158, 220, 60, 0, 26, 216, 159, 220, 60, 0, 26, 216, 160, 220, 60, 0, 26, 216, 161, 220, 60, 0, 26, 216, 162, 220, 60, 0, 26, 216, 163, 220, 60, 0, 26, 216, 164, 220, 60, 0, 26, 216, 165, 220, 60, 0, 26, 216, 166, 220, 60, 0, 26, 216, 167, 220, 60, 0, 26, 216, 168, 220, 60, 0, 26, 216, 169, 220, 60, 0, 26, 216, 170, 220, 60, 0, 26, 216, 171, 220, 60, 0, 26, 216, 172, 220, 60, 0, 26, 216, 173, 220, 60, 0, 26, 216, 174, 220, 60, 0, 26, 216, 175, 220, 60, 0, 26, 216, 176, 220, 60, 0, 26, 216, 177, 220, 60, 0, 26, 216, 178, 220, 60, 0, 26, 216, 179, 220, 60, 0, 26, 216, 180, 220, 60, 0, 26, 216, 181, 220, 60, 0, 26, 216, 182, 220, 60, 0, 26, 216, 183, 220, 60, 0, 26, 216, 184, 220, 60, 0, 26, 216, 185, 220, 60, 0, 26, 216, 186, 220, 60, 0, 26, 216, 187, 220, 60, 0, 26, 216, 188, 220, 60, 0, 26, 216, 189, 220, 60, 0, 26, 216, 190, 220, 60, 0, 26, 216, 191, 220, 60, 0, 26, 216, 192, 220, 60, 0, 26, 216, 193, 220, 60, 0, 26, 216, 194, 220, 60, 0, 26, 216, 195, 220, 60, 0, 26, 216, 196, 220, 60, 0, 26, 216, 197, 220, 60, 0, 26, 216, 198, 220, 60, 0, 26, 216, 199, 220, 60, 0, 26, 216, 200, 220, 60, 0, 26, 216, 201, 220, 60, 0, 26, 216, 202, 220, 60, 0, 26, 216, 203, 220, 60, 0, 26, 216, 204, 220, 60, 0, 26, 216, 205, 220, 60, 0, 26, 216, 206, 220, 60, 0, 26, 216, 207, 220, 60, 0, 26, 216, 208, 220, 60, 0, 26, 216, 209, 220, 60, 0, 26, 216, 210, 220, 60, 0, 26, 216, 211, 220, 60, 0, 26, 216, 212, 220, 60, 0, 26, 216, 213, 220, 60, 0, 26, 216, 214, 220, 60, 0, 26, 216, 215, 220, 60, 0, 26, 216, 216, 220, 60, 0, 26, 216, 217, 220, 60, 0, 26, 216, 218, 220, 60, 0, 26, 216, 219, 220, 60, 0, 26, 216, 220, 220, 60, 0, 26, 216, 221, 220, 60, 0, 26, 216, 222, 220, 60, 0, 26, 216, 223, 220, 60, 0, 26, 216, 224, 220, 60, 0, 26, 216, 225, 220, 60, 0, 26, 216, 226, 220, 60, 0, 26, 216, 227, 220, 60, 0, 26, 216, 228, 220, 60, 0, 26, 216, 229, 220, 60, 0, 26, 216, 230, 220, 60, 0, 26, 216, 231, 220, 60, 0, 26, 216, 232, 220, 60, 0, 26, 216, 233, 220, 60, 0, 26, 216, 234, 220, 60, 0, 26, 216, 235, 220, 60, 0, 26, 216, 236, 220, 60, 0, 26, 216, 237, 220, 60, 0, 26, 216, 238, 220, 60, 0, 26, 216, 239, 220, 60, 0, 26, 216, 240, 220, 60, 0, 26, 216, 241, 220, 60, 0, 26, 216, 242, 220, 60, 0, 26, 216, 243, 220, 60, 0, 26, 216, 244, 220, 60, 0, 26, 216, 245, 220, 60, 0, 26, 216, 246, 220, 60, 0, 26, 216, 247, 220, 60, 0, 26, 216, 248, 220, 60, 0, 26, 216, 249, 220, 60, 0, 26, 216, 250, 220, 60, 0, 26, 216, 251, 220, 60, 0, 26, 216, 252, 220, 60, 0, 26, 216, 253, 220, 60, 0, 26, 216, 254, 220, 60, 0, 26, 216, 255, 220, 60, 0, 26, 216, 0, 221, 60, 0, 26, 216, 1, 221, 60, 0, 26, 216, 2, 221, 60, 0, 26, 216, 3, 221, 60, 0, 26, 216, 4, 221, 60, 0, 26, 216, 5, 221, 60, 0, 26, 216, 6, 221, 60, 0, 26, 216, 7, 221, 60, 0, 26, 216, 8, 221, 60, 0, 26, 216, 9, 221, 60, 0, 26, 216, 10, 221, 60, 0, 26, 216, 11, 221, 60, 0, 26, 216, 12, 221, 60, 0, 26, 216, 13, 221, 60, 0, 26, 216, 14, 221, 60, 0, 26, 216, 15, 221, 60, 0, 26, 216, 16, 221, 60, 0, 26, 216, 17, 221, 60, 0, 26, 216, 18, 221, 60, 0, 26, 216, 19, 221, 60, 0, 26, 216, 20, 221, 60, 0, 26, 216, 21, 221, 60, 0, 26, 216, 22, 221, 60, 0, 26, 216, 23, 221, 60, 0, 26, 216, 24, 221, 60, 0, 26, 216, 25, 221, 60, 0, 26, 216, 26, 221, 60, 0, 26, 216, 27, 221, 60, 0, 26, 216, 28, 221, 60, 0, 26, 216, 29, 221, 60, 0, 26, 216, 30, 221, 60, 0, 26, 216, 31, 221, 60, 0, 26, 216, 32, 221, 60, 0, 26, 216, 33, 221, 60, 0, 26, 216, 34, 221, 60, 0, 26, 216, 35, 221, 60, 0, 26, 216, 36, 221, 60, 0, 26, 216, 37, 221, 60, 0, 26, 216, 38, 221, 60, 0, 26, 216, 39, 221, 60, 0, 26, 216, 40, 221, 60, 0, 26, 216, 41, 221, 60, 0, 26, 216, 42, 221, 60, 0, 26, 216, 43, 221, 60, 0, 26, 216, 44, 221, 60, 0, 26, 216, 45, 221, 60, 0, 26, 216, 46, 221, 60, 0, 26, 216, 47, 221, 60, 0, 26, 216, 48, 221, 60, 0, 26, 216, 49, 221, 60, 0, 26, 216, 50, 221, 60, 0, 26, 216, 51, 221, 60, 0, 26, 216, 52, 221, 60, 0, 26, 216, 53, 221, 60, 0, 26, 216, 54, 221, 60, 0, 26, 216, 55, 221, 60, 0, 26, 216, 56, 221, 60, 0, 26, 216, 57, 221, 60, 0, 26, 216, 58, 221, 60, 0, 26, 216, 59, 221, 60, 0, 26, 216, 60, 221, 60, 0, 26, 216, 61, 221, 60, 0, 26, 216, 62, 221, 60, 0, 26, 216, 63, 221, 60, 0, 26, 216, 64, 221, 60, 0, 26, 216, 65, 221, 60, 0, 26, 216, 66, 221, 60, 0, 26, 216, 67, 221, 60, 0, 26, 216, 68, 221, 60, 0, 26, 216, 69, 221, 60, 0, 26, 216, 70, 221, 60, 0, 26, 216, 71, 221, 60, 0, 26, 216, 72, 221, 60, 0, 26, 216, 73, 221, 60, 0, 26, 216, 74, 221, 60, 0, 26, 216, 75, 221, 60, 0, 26, 216, 76, 221, 60, 0, 26, 216, 77, 221, 60, 0, 26, 216, 78, 221, 60, 0, 26, 216, 79, 221, 60, 0, 26, 216, 80, 221, 60, 0, 26, 216, 81, 221, 60, 0, 26, 216, 82, 221, 60, 0, 26, 216, 83, 221, 60, 0, 26, 216, 84, 221, 60, 0, 26, 216, 85, 221, 60, 0, 26, 216, 86, 221, 60, 0, 26, 216, 87, 221, 60, 0, 26, 216, 88, 221, 60, 0, 26, 216, 89, 221, 60, 0, 26, 216, 90, 221, 60, 0, 26, 216, 91, 221, 60, 0, 26, 216, 92, 221, 60, 0, 26, 216, 93, 221, 60, 0, 26, 216, 94, 221, 60, 0, 26, 216, 95, 221, 60, 0, 26, 216, 96, 221, 60, 0, 26, 216, 97, 221, 60, 0, 26, 216, 98, 221, 60, 0, 26, 216, 99, 221, 60, 0, 26, 216, 100, 221, 60, 0, 26, 216, 101, 221, 60, 0, 26, 216, 102, 221, 60, 0, 26, 216, 103, 221, 60, 0, 26, 216, 104, 221, 60, 0, 26, 216, 105, 221, 60, 0, 26, 216, 106, 221, 60, 0, 26, 216, 107, 221, 60, 0, 26, 216, 108, 221, 60, 0, 26, 216, 109, 221, 60, 0, 26, 216, 110, 221, 60, 0, 26, 216, 111, 221, 60, 0, 26, 216, 112, 221, 60, 0, 26, 216, 113, 221, 60, 0, 26, 216, 114, 221, 60, 0, 26, 216, 115, 221, 60, 0, 26, 216, 116, 221, 60, 0, 26, 216, 117, 221, 60, 0, 26, 216, 118, 221, 60, 0, 26, 216, 119, 221, 60, 0, 26, 216, 120, 221, 60, 0, 26, 216, 121, 221, 60, 0, 26, 216, 122, 221, 60, 0, 26, 216, 123, 221, 60, 0, 26, 216, 124, 221, 60, 0, 26, 216, 125, 221, 60, 0, 26, 216, 126, 221, 60, 0, 26, 216, 127, 221, 60, 0, 26, 216, 128, 221, 60, 0, 26, 216, 129, 221, 60, 0, 26, 216, 130, 221, 60, 0, 26, 216, 131, 221, 60, 0, 26, 216, 132, 221, 60, 0, 26, 216, 133, 221, 60, 0, 26, 216, 134, 221, 60, 0, 26, 216, 135, 221, 60, 0, 26, 216, 136, 221, 60, 0, 26, 216, 137, 221, 60, 0, 26, 216, 138, 221, 60, 0, 26, 216, 139, 221, 60, 0, 26, 216, 140, 221, 60, 0, 26, 216, 141, 221, 60, 0, 26, 216, 142, 221, 60, 0, 26, 216, 143, 221, 60, 0, 26, 216, 144, 221, 60, 0, 26, 216, 145, 221, 60, 0, 26, 216, 146, 221, 60, 0, 26, 216, 147, 221, 60, 0, 26, 216, 148, 221, 60, 0, 26, 216, 149, 221, 60, 0, 26, 216, 150, 221, 60, 0, 26, 216, 151, 221, 60, 0, 26, 216, 152, 221, 60, 0, 26, 216, 153, 221, 60, 0, 26, 216, 154, 221, 60, 0, 26, 216, 155, 221, 60, 0, 26, 216, 156, 221, 60, 0, 26, 216, 157, 221, 60, 0, 26, 216, 158, 221, 60, 0, 26, 216, 159, 221, 60, 0, 26, 216, 160, 221, 60, 0, 26, 216, 161, 221, 60, 0, 26, 216, 162, 221, 60, 0, 26, 216, 163, 221, 60, 0, 26, 216, 164, 221, 60, 0, 26, 216, 165, 221, 60, 0, 26, 216, 166, 221, 60, 0, 26, 216, 167, 221, 60, 0, 26, 216, 168, 221, 60, 0, 26, 216, 169, 221, 60, 0, 26, 216, 170, 221, 60, 0, 26, 216, 171, 221, 60, 0, 26, 216, 172, 221, 60, 0, 26, 216, 173, 221, 60, 0, 26, 216, 174, 221, 60, 0, 26, 216, 175, 221, 60, 0, 26, 216, 176, 221, 60, 0, 26, 216, 177, 221, 60, 0, 26, 216, 178, 221, 60, 0, 26, 216, 179, 221, 60, 0, 26, 216, 180, 221, 60, 0, 26, 216, 181, 221, 60, 0, 26, 216, 182, 221, 60, 0, 26, 216, 183, 221, 60, 0, 26, 216, 184, 221, 60, 0, 26, 216, 185, 221, 60, 0, 26, 216, 186, 221, 60, 0, 26, 216, 187, 221, 60, 0, 26, 216, 188, 221, 60, 0, 26, 216, 189, 221, 60, 0, 26, 216, 190, 221, 60, 0, 26, 216, 191, 221, 60, 0, 26, 216, 192, 221, 60, 0, 26, 216, 193, 221, 60, 0, 26, 216, 194, 221, 60, 0, 26, 216, 195, 221, 60, 0, 26, 216, 196, 221, 60, 0, 26, 216, 197, 221, 60, 0, 26, 216, 198, 221, 60, 0, 26, 216, 199, 221, 60, 0, 26, 216, 200, 221, 60, 0, 26, 216, 201, 221, 60, 0, 26, 216, 202, 221, 60, 0, 26, 216, 203, 221, 60, 0, 26, 216, 204, 221, 60, 0, 26, 216, 205, 221, 60, 0, 26, 216, 206, 221, 60, 0, 26, 216, 207, 221, 60, 0, 26, 216, 208, 221, 60, 0, 26, 216, 209, 221, 60, 0, 26, 216, 210, 221, 60, 0, 26, 216, 211, 221, 60, 0, 26, 216, 212, 221, 60, 0, 26, 216, 213, 221, 60, 0, 26, 216, 214, 221, 60, 0, 26, 216, 215, 221, 60, 0, 26, 216, 216, 221, 60, 0, 26, 216, 217, 221, 60, 0, 26, 216, 218, 221, 60, 0, 26, 216, 219, 221, 60, 0, 26, 216, 220, 221, 60, 0, 26, 216, 221, 221, 60, 0, 26, 216, 222, 221, 60, 0, 26, 216, 223, 221, 60, 0, 26, 216, 224, 221, 60, 0, 26, 216, 225, 221, 60, 0, 26, 216, 226, 221, 60, 0, 26, 216, 227, 221, 60, 0, 26, 216, 228, 221, 60, 0, 26, 216, 229, 221, 60, 0, 26, 216, 230, 221, 60, 0, 26, 216, 231, 221, 60, 0, 26, 216, 232, 221, 60, 0, 26, 216, 233, 221, 60, 0, 26, 216, 234, 221, 60, 0, 26, 216, 235, 221, 60, 0, 26, 216, 236, 221, 60, 0, 26, 216, 237, 221, 60, 0, 26, 216, 238, 221, 60, 0, 26, 216, 239, 221, 60, 0, 26, 216, 240, 221, 60, 0, 26, 216, 241, 221, 60, 0, 26, 216, 242, 221, 60, 0, 26, 216, 243, 221, 60, 0, 26, 216, 244, 221, 60, 0, 26, 216, 245, 221, 60, 0, 26, 216, 246, 221, 60, 0, 26, 216, 247, 221, 60, 0, 26, 216, 248, 221, 60, 0, 26, 216, 249, 221, 60, 0, 26, 216, 250, 221, 60, 0, 26, 216, 251, 221, 60, 0, 26, 216, 252, 221, 60, 0, 26, 216, 253, 221, 60, 0, 26, 216, 254, 221, 60, 0, 26, 216, 255, 221, 60, 0, 26, 216, 0, 222, 60, 0, 26, 216, 1, 222, 60, 0, 26, 216, 2, 222, 60, 0, 26, 216, 3, 222, 60, 0, 26, 216, 4, 222, 60, 0, 26, 216, 5, 222, 60, 0, 26, 216, 6, 222, 60, 0, 26, 216, 7, 222, 60, 0, 26, 216, 8, 222, 60, 0, 26, 216, 9, 222, 60, 0, 26, 216, 10, 222, 60, 0, 26, 216, 11, 222, 60, 0, 26, 216, 12, 222, 60, 0, 26, 216, 13, 222, 60, 0, 26, 216, 14, 222, 60, 0, 26, 216, 15, 222, 60, 0, 26, 216, 16, 222, 60, 0, 26, 216, 17, 222, 60, 0, 26, 216, 18, 222, 60, 0, 26, 216, 19, 222, 60, 0, 26, 216, 20, 222, 60, 0, 26, 216, 21, 222, 60, 0, 26, 216, 22, 222, 60, 0, 26, 216, 23, 222, 60, 0, 26, 216, 24, 222, 60, 0, 26, 216, 25, 222, 60, 0, 26, 216, 26, 222, 60, 0, 26, 216, 27, 222, 60, 0, 26, 216, 28, 222, 60, 0, 26, 216, 29, 222, 60, 0, 26, 216, 30, 222, 60, 0, 26, 216, 31, 222, 60, 0, 26, 216, 32, 222, 60, 0, 26, 216, 33, 222, 60, 0, 26, 216, 34, 222, 60, 0, 26, 216, 35, 222, 60, 0, 26, 216, 36, 222, 60, 0, 26, 216, 37, 222, 60, 0, 26, 216, 38, 222, 60, 0, 26, 216, 39, 222, 60, 0, 26, 216, 40, 222, 60, 0, 26, 216, 41, 222, 60, 0, 26, 216, 42, 222, 60, 0, 26, 216, 43, 222, 60, 0, 26, 216, 44, 222, 60, 0, 26, 216, 45, 222, 60, 0, 26, 216, 46, 222, 60, 0, 26, 216, 47, 222, 60, 0, 26, 216, 48, 222, 60, 0, 26, 216, 49, 222, 60, 0, 26, 216, 50, 222, 60, 0, 26, 216, 51, 222, 60, 0, 26, 216, 52, 222, 60, 0, 26, 216, 53, 222, 60, 0, 26, 216, 54, 222, 60, 0, 26, 216, 55, 222, 60, 0, 26, 216, 56, 222, 60, 0, 26, 216, 208, 222, 60, 0, 26, 216, 209, 222, 60, 0, 26, 216, 210, 222, 60, 0, 26, 216, 211, 222, 60, 0, 26, 216, 212, 222, 60, 0, 26, 216, 213, 222, 60, 0, 26, 216, 214, 222, 60, 0, 26, 216, 215, 222, 60, 0, 26, 216, 216, 222, 60, 0, 26, 216, 217, 222, 60, 0, 26, 216, 218, 222, 60, 0, 26, 216, 219, 222, 60, 0, 26, 216, 220, 222, 60, 0, 26, 216, 221, 222, 60, 0, 26, 216, 222, 222, 60, 0, 26, 216, 223, 222, 60, 0, 26, 216, 224, 222, 60, 0, 26, 216, 225, 222, 60, 0, 26, 216, 226, 222, 60, 0, 26, 216, 227, 222, 60, 0, 26, 216, 228, 222, 60, 0, 26, 216, 229, 222, 60, 0, 26, 216, 230, 222, 60, 0, 26, 216, 231, 222, 60, 0, 26, 216, 232, 222, 60, 0, 26, 216, 233, 222, 60, 0, 26, 216, 234, 222, 60, 0, 26, 216, 235, 222, 60, 0, 26, 216, 236, 222, 60, 0, 26, 216, 237, 222, 60, 0, 14, 32, 58, 216, 0, 220, 14, 32, 60, 0, 14, 32, 58, 216, 1, 220, 14, 32, 60, 0, 14, 32, 58, 216, 2, 220, 14, 32, 60, 0, 14, 32, 58, 216, 3, 220, 14, 32, 60, 0, 14, 32, 58, 216, 4, 220, 14, 32, 60, 0, 14, 32, 58, 216, 5, 220, 14, 32, 60, 0, 14, 32, 58, 216, 6, 220, 14, 32, 60, 0, 14, 32, 58, 216, 7, 220, 14, 32, 60, 0, 14, 32, 58, 216, 8, 220, 14, 32, 60, 0, 14, 32, 58, 216, 9, 220, 14, 32, 60, 0, 14, 32, 58, 216, 10, 220, 14, 32, 60, 0, 14, 32, 58, 216, 11, 220, 14, 32, 60, 0, 14, 32, 58, 216, 12, 220, 14, 32, 60, 0, 14, 32, 58, 216, 13, 220, 14, 32, 60, 0, 14, 32, 58, 216, 14, 220, 14, 32, 60, 0, 14, 32, 58, 216, 15, 220, 14, 32, 60, 0, 14, 32, 58, 216, 16, 220, 14, 32, 60, 0, 14, 32, 58, 216, 17, 220, 14, 32, 60, 0, 14, 32, 58, 216, 18, 220, 14, 32, 60, 0, 14, 32, 58, 216, 19, 220, 14, 32, 60, 0, 14, 32, 58, 216, 20, 220, 14, 32, 60, 0, 14, 32, 58, 216, 21, 220, 14, 32, 60, 0, 14, 32, 58, 216, 22, 220, 14, 32, 60, 0, 14, 32, 58, 216, 23, 220, 14, 32, 60, 0, 14, 32, 58, 216, 24, 220, 14, 32, 60, 0, 14, 32, 58, 216, 25, 220, 14, 32, 60, 0, 14, 32, 58, 216, 26, 220, 14, 32, 60, 0, 14, 32, 58, 216, 27, 220, 14, 32, 60, 0, 14, 32, 58, 216, 28, 220, 14, 32, 60, 0, 14, 32, 58, 216, 29, 220, 14, 32, 60, 0, 14, 32, 58, 216, 30, 220, 14, 32, 60, 0, 14, 32, 58, 216, 31, 220, 14, 32, 60, 0, 14, 32, 58, 216, 32, 220, 14, 32, 60, 0, 14, 32, 58, 216, 33, 220, 14, 32, 60, 0, 14, 32, 58, 216, 34, 220, 14, 32, 60, 0, 14, 32, 58, 216, 35, 220, 14, 32, 60, 0, 14, 32, 58, 216, 36, 220, 14, 32, 60, 0, 14, 32, 58, 216, 37, 220, 14, 32, 60, 0, 14, 32, 58, 216, 38, 220, 14, 32, 60, 0, 14, 32, 58, 216, 39, 220, 14, 32, 60, 0, 14, 32, 58, 216, 40, 220, 14, 32, 60, 0, 14, 32, 58, 216, 41, 220, 14, 32, 60, 0, 14, 32, 58, 216, 42, 220, 14, 32, 60, 0, 14, 32, 58, 216, 43, 220, 14, 32, 60, 0, 14, 32, 58, 216, 44, 220, 14, 32, 60, 0, 14, 32, 58, 216, 45, 220, 14, 32, 60, 0, 14, 32, 58, 216, 46, 220, 14, 32, 60, 0, 14, 32, 58, 216, 47, 220, 14, 32, 60, 0, 14, 32, 58, 216, 48, 220, 14, 32, 60, 0, 14, 32, 58, 216, 49, 220, 14, 32, 60, 0, 14, 32, 58, 216, 50, 220, 14, 32, 60, 0, 14, 32, 58, 216, 51, 220, 14, 32, 60, 0, 14, 32, 58, 216, 52, 220, 14, 32, 60, 0, 14, 32, 58, 216, 53, 220, 14, 32, 60, 0, 14, 32, 58, 216, 54, 220, 14, 32, 60, 0, 14, 32, 58, 216, 55, 220, 14, 32, 60, 0, 14, 32, 58, 216, 56, 220, 14, 32, 60, 0, 14, 32, 58, 216, 57, 220, 14, 32, 60, 0, 14, 32, 58, 216, 58, 220, 14, 32, 60, 0, 14, 32, 58, 216, 59, 220, 14, 32, 60, 0, 14, 32, 58, 216, 60, 220, 14, 32, 60, 0, 14, 32, 58, 216, 61, 220, 14, 32, 60, 0, 14, 32, 58, 216, 62, 220, 14, 32, 60, 0, 14, 32, 58, 216, 63, 220, 14, 32, 60, 0, 14, 32, 58, 216, 64, 220, 14, 32, 60, 0, 14, 32, 58, 216, 65, 220, 14, 32, 60, 0, 14, 32, 58, 216, 66, 220, 14, 32, 60, 0, 14, 32, 58, 216, 67, 220, 14, 32, 60, 0, 14, 32, 58, 216, 68, 220, 14, 32, 60, 0, 14, 32, 58, 216, 69, 220, 14, 32, 60, 0, 14, 32, 58, 216, 70, 220, 14, 32, 60, 0, 14, 32, 58, 216, 71, 220, 14, 32, 60, 0, 14, 32, 58, 216, 72, 220, 14, 32, 60, 0, 14, 32, 58, 216, 73, 220, 14, 32, 60, 0, 14, 32, 58, 216, 74, 220, 14, 32, 60, 0, 14, 32, 58, 216, 75, 220, 14, 32, 60, 0, 14, 32, 58, 216, 76, 220, 14, 32, 60, 0, 14, 32, 58, 216, 77, 220, 14, 32, 60, 0, 14, 32, 58, 216, 78, 220, 14, 32, 60, 0, 14, 32, 58, 216, 79, 220, 14, 32, 60, 0, 14, 32, 58, 216, 80, 220, 14, 32, 60, 0, 14, 32, 58, 216, 81, 220, 14, 32, 60, 0, 14, 32, 58, 216, 82, 220, 14, 32, 60, 0, 14, 32, 58, 216, 83, 220, 14, 32, 60, 0, 14, 32, 58, 216, 84, 220, 14, 32, 60, 0, 14, 32, 58, 216, 85, 220, 14, 32, 60, 0, 14, 32, 58, 216, 86, 220, 14, 32, 60, 0, 14, 32, 58, 216, 87, 220, 14, 32, 60, 0, 14, 32, 58, 216, 88, 220, 14, 32, 60, 0, 14, 32, 58, 216, 89, 220, 14, 32, 60, 0, 14, 32, 58, 216, 90, 220, 14, 32, 60, 0, 14, 32, 58, 216, 91, 220, 14, 32, 60, 0, 14, 32, 58, 216, 92, 220, 14, 32, 60, 0, 14, 32, 58, 216, 93, 220, 14, 32, 60, 0, 14, 32, 58, 216, 94, 220, 14, 32, 60, 0, 14, 32, 58, 216, 95, 220, 14, 32, 60, 0, 14, 32, 58, 216, 96, 220, 14, 32, 60, 0, 14, 32, 58, 216, 97, 220, 14, 32, 60, 0, 14, 32, 58, 216, 98, 220, 14, 32, 60, 0, 14, 32, 58, 216, 99, 220, 14, 32, 60, 0, 14, 32, 58, 216, 100, 220, 14, 32, 60, 0, 14, 32, 58, 216, 101, 220, 14, 32, 60, 0, 14, 32, 58, 216, 102, 220, 14, 32, 60, 0, 14, 32, 58, 216, 103, 220, 14, 32, 60, 0, 14, 32, 58, 216, 104, 220, 14, 32, 60, 0, 14, 32, 58, 216, 105, 220, 14, 32, 60, 0, 14, 32, 58, 216, 106, 220, 14, 32, 60, 0, 14, 32, 58, 216, 107, 220, 14, 32, 60, 0, 14, 32, 58, 216, 108, 220, 14, 32, 60, 0, 14, 32, 58, 216, 109, 220, 14, 32, 60, 0, 14, 32, 58, 216, 110, 220, 14, 32, 60, 0, 14, 32, 58, 216, 111, 220, 14, 32, 60, 0, 14, 32, 58, 216, 112, 220, 14, 32, 60, 0, 14, 32, 58, 216, 113, 220, 14, 32, 60, 0, 14, 32, 58, 216, 114, 220, 14, 32, 60, 0, 14, 32, 58, 216, 115, 220, 14, 32, 60, 0, 14, 32, 58, 216, 116, 220, 14, 32, 60, 0, 14, 32, 58, 216, 117, 220, 14, 32, 60, 0, 14, 32, 58, 216, 118, 220, 14, 32, 60, 0, 14, 32, 58, 216, 119, 220, 14, 32, 60, 0, 14, 32, 58, 216, 120, 220, 14, 32, 60, 0, 14, 32, 58, 216, 121, 220, 14, 32, 60, 0, 14, 32, 58, 216, 122, 220, 14, 32, 60, 0, 14, 32, 58, 216, 123, 220, 14, 32, 60, 0, 14, 32, 58, 216, 124, 220, 14, 32, 60, 0, 14, 32, 58, 216, 125, 220, 14, 32, 60, 0, 14, 32, 58, 216, 126, 220, 14, 32, 60, 0, 14, 32, 58, 216, 127, 220, 14, 32, 60, 0, 14, 32, 58, 216, 128, 220, 14, 32, 60, 0, 14, 32, 58, 216, 129, 220, 14, 32, 60, 0, 14, 32, 58, 216, 130, 220, 14, 32, 60, 0, 14, 32, 58, 216, 131, 220, 14, 32, 60, 0, 14, 32, 58, 216, 132, 220, 14, 32, 60, 0, 14, 32, 58, 216, 133, 220, 14, 32, 60, 0, 14, 32, 58, 216, 134, 220, 14, 32, 60, 0, 14, 32, 58, 216, 135, 220, 14, 32, 60, 0, 14, 32, 58, 216, 136, 220, 14, 32, 60, 0, 14, 32, 58, 216, 137, 220, 14, 32, 60, 0, 14, 32, 58, 216, 138, 220, 14, 32, 60, 0, 14, 32, 58, 216, 139, 220, 14, 32, 60, 0, 14, 32, 58, 216, 140, 220, 14, 32, 60, 0, 14, 32, 58, 216, 141, 220, 14, 32, 60, 0, 14, 32, 58, 216, 142, 220, 14, 32, 60, 0, 14, 32, 58, 216, 143, 220, 14, 32, 60, 0, 14, 32, 58, 216, 144, 220, 14, 32, 60, 0, 14, 32, 58, 216, 145, 220, 14, 32, 60, 0, 14, 32, 58, 216, 146, 220, 14, 32, 60, 0, 14, 32, 58, 216, 147, 220, 14, 32, 60, 0, 14, 32, 58, 216, 148, 220, 14, 32, 60, 0, 14, 32, 58, 216, 149, 220, 14, 32, 60, 0, 14, 32, 58, 216, 150, 220, 14, 32, 60, 0, 14, 32, 58, 216, 151, 220, 14, 32, 60, 0, 14, 32, 58, 216, 152, 220, 14, 32, 60, 0, 14, 32, 58, 216, 153, 220, 14, 32, 60, 0, 14, 32, 58, 216, 154, 220, 14, 32, 60, 0, 14, 32, 58, 216, 155, 220, 14, 32, 60, 0, 14, 32, 58, 216, 156, 220, 14, 32, 60, 0, 14, 32, 58, 216, 157, 220, 14, 32, 60, 0, 14, 32, 58, 216, 158, 220, 14, 32, 60, 0, 14, 32, 58, 216, 159, 220, 14, 32, 60, 0, 14, 32, 58, 216, 160, 220, 14, 32, 60, 0, 14, 32, 58, 216, 161, 220, 14, 32, 60, 0, 14, 32, 58, 216, 162, 220, 14, 32, 60, 0, 14, 32, 58, 216, 163, 220, 14, 32, 60, 0, 14, 32, 58, 216, 164, 220, 14, 32, 60, 0, 14, 32, 58, 216, 165, 220, 14, 32, 60, 0, 14, 32, 58, 216, 166, 220, 14, 32, 60, 0, 14, 32, 58, 216, 167, 220, 14, 32, 60, 0, 14, 32, 58, 216, 168, 220, 14, 32, 60, 0, 14, 32, 58, 216, 169, 220, 14, 32, 60, 0, 14, 32, 58, 216, 170, 220, 14, 32, 60, 0, 14, 32, 58, 216, 171, 220, 14, 32, 60, 0, 14, 32, 58, 216, 172, 220, 14, 32, 60, 0, 14, 32, 58, 216, 173, 220, 14, 32, 60, 0, 14, 32, 58, 216, 174, 220, 14, 32, 60, 0, 14, 32, 58, 216, 175, 220, 14, 32, 60, 0, 14, 32, 58, 216, 176, 220, 14, 32, 60, 0, 14, 32, 58, 216, 177, 220, 14, 32, 60, 0, 14, 32, 58, 216, 178, 220, 14, 32, 60, 0, 14, 32, 58, 216, 179, 220, 14, 32, 60, 0, 14, 32, 58, 216, 180, 220, 14, 32, 60, 0, 14, 32, 58, 216, 181, 220, 14, 32, 60, 0, 14, 32, 58, 216, 182, 220, 14, 32, 60, 0, 14, 32, 58, 216, 183, 220, 14, 32, 60, 0, 14, 32, 58, 216, 184, 220, 14, 32, 60, 0, 14, 32, 58, 216, 185, 220, 14, 32, 60, 0, 14, 32, 58, 216, 186, 220, 14, 32, 60, 0, 14, 32, 58, 216, 187, 220, 14, 32, 60, 0, 14, 32, 58, 216, 188, 220, 14, 32, 60, 0, 14, 32, 58, 216, 189, 220, 14, 32, 60, 0, 14, 32, 58, 216, 190, 220, 14, 32, 60, 0, 14, 32, 58, 216, 191, 220, 14, 32, 60, 0, 14, 32, 58, 216, 192, 220, 14, 32, 60, 0, 14, 32, 58, 216, 193, 220, 14, 32, 60, 0, 14, 32, 58, 216, 194, 220, 14, 32, 60, 0, 14, 32, 58, 216, 195, 220, 14, 32, 60, 0, 14, 32, 58, 216, 196, 220, 14, 32, 60, 0, 27, 216, 96, 222, 60, 0, 60, 0, 60, 0, 27, 216, 64, 222, 60, 0, 27, 216, 97, 222, 60, 0, 60, 0, 60, 0, 27, 216, 65, 222, 60, 0, 27, 216, 98, 222, 60, 0, 60, 0, 60, 0, 27, 216, 66, 222, 60, 0, 27, 216, 99, 222, 60, 0, 60, 0, 60, 0, 27, 216, 67, 222, 60, 0, 27, 216, 100, 222, 60, 0, 60, 0, 60, 0, 27, 216, 68, 222, 60, 0, 27, 216, 101, 222, 60, 0, 60, 0, 60, 0, 27, 216, 69, 222, 60, 0, 27, 216, 102, 222, 60, 0, 60, 0, 60, 0, 27, 216, 70, 222, 60, 0, 27, 216, 103, 222, 60, 0, 60, 0, 60, 0, 27, 216, 71, 222, 60, 0, 27, 216, 104, 222, 60, 0, 60, 0, 60, 0, 27, 216, 72, 222, 60, 0, 27, 216, 105, 222, 60, 0, 60, 0, 60, 0, 27, 216, 73, 222, 60, 0, 27, 216, 106, 222, 60, 0, 60, 0, 60, 0, 27, 216, 74, 222, 60, 0, 27, 216, 107, 222, 60, 0, 60, 0, 60, 0, 27, 216, 75, 222, 60, 0, 27, 216, 108, 222, 60, 0, 60, 0, 60, 0, 27, 216, 76, 222, 60, 0, 27, 216, 109, 222, 60, 0, 60, 0, 60, 0, 27, 216, 77, 222, 60, 0, 27, 216, 110, 222, 60, 0, 60, 0, 60, 0, 27, 216, 78, 222, 60, 0, 27, 216, 111, 222, 60, 0, 60, 0, 60, 0, 27, 216, 79, 222, 60, 0, 27, 216, 112, 222, 60, 0, 60, 0, 60, 0, 27, 216, 80, 222, 60, 0, 27, 216, 113, 222, 60, 0, 60, 0, 60, 0, 27, 216, 81, 222, 60, 0, 27, 216, 114, 222, 60, 0, 60, 0, 60, 0, 27, 216, 82, 222, 60, 0, 27, 216, 115, 222, 60, 0, 60, 0, 60, 0, 27, 216, 83, 222, 60, 0, 27, 216, 116, 222, 60, 0, 60, 0, 60, 0, 27, 216, 84, 222, 60, 0, 27, 216, 117, 222, 60, 0, 60, 0, 60, 0, 27, 216, 85, 222, 60, 0, 27, 216, 118, 222, 60, 0, 60, 0, 60, 0, 27, 216, 86, 222, 60, 0, 27, 216, 119, 222, 60, 0, 60, 0, 60, 0, 27, 216, 87, 222, 60, 0, 27, 216, 120, 222, 60, 0, 60, 0, 60, 0, 27, 216, 88, 222, 60, 0, 27, 216, 121, 222, 60, 0, 60, 0, 60, 0, 27, 216, 89, 222, 60, 0, 27, 216, 122, 222, 60, 0, 60, 0, 60, 0, 27, 216, 90, 222, 60, 0, 27, 216, 123, 222, 60, 0, 60, 0, 60, 0, 27, 216, 91, 222, 60, 0, 27, 216, 124, 222, 60, 0, 60, 0, 60, 0, 27, 216, 92, 222, 60, 0, 27, 216, 125, 222, 60, 0, 60, 0, 60, 0, 27, 216, 93, 222, 60, 0, 27, 216, 126, 222, 60, 0, 60, 0, 60, 0, 27, 216, 94, 222, 60, 0, 27, 216, 127, 222, 60, 0, 60, 0, 60, 0, 27, 216, 95, 222, 60, 0, 14, 32, 58, 216, 34, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 0, 221, 14, 32, 60, 0, 14, 32, 58, 216, 35, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 1, 221, 14, 32, 60, 0, 14, 32, 58, 216, 36, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 2, 221, 14, 32, 60, 0, 14, 32, 58, 216, 37, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 3, 221, 14, 32, 60, 0, 14, 32, 58, 216, 38, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 4, 221, 14, 32, 60, 0, 14, 32, 58, 216, 39, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 5, 221, 14, 32, 60, 0, 14, 32, 58, 216, 40, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 6, 221, 14, 32, 60, 0, 14, 32, 58, 216, 41, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 7, 221, 14, 32, 60, 0, 14, 32, 58, 216, 42, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 8, 221, 14, 32, 60, 0, 14, 32, 58, 216, 43, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 9, 221, 14, 32, 60, 0, 14, 32, 58, 216, 44, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 10, 221, 14, 32, 60, 0, 14, 32, 58, 216, 45, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 11, 221, 14, 32, 60, 0, 14, 32, 58, 216, 46, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 12, 221, 14, 32, 60, 0, 14, 32, 58, 216, 47, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 13, 221, 14, 32, 60, 0, 14, 32, 58, 216, 48, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 14, 221, 14, 32, 60, 0, 14, 32, 58, 216, 49, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 15, 221, 14, 32, 60, 0, 14, 32, 58, 216, 50, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 16, 221, 14, 32, 60, 0, 14, 32, 58, 216, 51, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 17, 221, 14, 32, 60, 0, 14, 32, 58, 216, 52, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 18, 221, 14, 32, 60, 0, 14, 32, 58, 216, 53, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 19, 221, 14, 32, 60, 0, 14, 32, 58, 216, 54, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 20, 221, 14, 32, 60, 0, 14, 32, 58, 216, 55, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 21, 221, 14, 32, 60, 0, 14, 32, 58, 216, 56, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 22, 221, 14, 32, 60, 0, 14, 32, 58, 216, 57, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 23, 221, 14, 32, 60, 0, 14, 32, 58, 216, 58, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 24, 221, 14, 32, 60, 0, 14, 32, 58, 216, 59, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 25, 221, 14, 32, 60, 0, 14, 32, 58, 216, 60, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 26, 221, 14, 32, 60, 0, 14, 32, 58, 216, 61, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 27, 221, 14, 32, 60, 0, 14, 32, 58, 216, 62, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 28, 221, 14, 32, 60, 0, 14, 32, 58, 216, 63, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 29, 221, 14, 32, 60, 0, 14, 32, 58, 216, 64, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 30, 221, 14, 32, 60, 0, 14, 32, 58, 216, 65, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 31, 221, 14, 32, 60, 0, 14, 32, 58, 216, 66, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 32, 221, 14, 32, 60, 0, 14, 32, 58, 216, 67, 221, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 58, 216, 33, 221, 14, 32, 60, 0, 14, 32, 58, 216, 75, 221, 14, 32, 60, 0, 0, 17, 60, 0, 60, 0, 60, 0, 49, 49, 60, 0, 60, 0, 60, 0, 96, 50, 60, 0, 60, 0, 60, 0, 110, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 161, 255, 60, 0, 1, 17, 60, 0, 60, 0, 60, 0, 50, 49, 60, 0, 60, 0, 60, 0, 162, 255, 60, 0, 2, 17, 60, 0, 60, 0, 60, 0, 52, 49, 60, 0, 60, 0, 60, 0, 97, 50, 60, 0, 60, 0, 60, 0, 111, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 164, 255, 60, 0, 3, 17, 60, 0, 60, 0, 60, 0, 55, 49, 60, 0, 60, 0, 60, 0, 98, 50, 60, 0, 60, 0, 60, 0, 112, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 167, 255, 60, 0, 4, 17, 60, 0, 60, 0, 60, 0, 56, 49, 60, 0, 60, 0, 60, 0, 168, 255, 60, 0, 5, 17, 60, 0, 60, 0, 60, 0, 57, 49, 60, 0, 60, 0, 60, 0, 99, 50, 60, 0, 60, 0, 60, 0, 113, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 169, 255, 60, 0, 6, 17, 60, 0, 60, 0, 60, 0, 65, 49, 60, 0, 60, 0, 60, 0, 100, 50, 60, 0, 60, 0, 60, 0, 114, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 177, 255, 60, 0, 7, 17, 60, 0, 60, 0, 60, 0, 66, 49, 60, 0, 60, 0, 60, 0, 101, 50, 60, 0, 60, 0, 60, 0, 115, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 178, 255, 60, 0, 8, 17, 60, 0, 60, 0, 60, 0, 67, 49, 60, 0, 60, 0, 60, 0, 179, 255, 60, 0, 9, 17, 60, 0, 60, 0, 60, 0, 69, 49, 60, 0, 60, 0, 60, 0, 102, 50, 60, 0, 60, 0, 60, 0, 116, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 181, 255, 60, 0, 10, 17, 60, 0, 60, 0, 60, 0, 70, 49, 60, 0, 60, 0, 60, 0, 182, 255, 60, 0, 11, 17, 60, 0, 60, 0, 60, 0, 71, 49, 60, 0, 60, 0, 60, 0, 103, 50, 60, 0, 60, 0, 60, 0, 117, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 126, 50, 47, 0, 110, 17, 60, 0, 60, 0, 60, 0, 183, 255, 60, 0, 12, 17, 60, 0, 60, 0, 60, 0, 72, 49, 60, 0, 60, 0, 60, 0, 104, 50, 60, 0, 60, 0, 60, 0, 118, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 125, 50, 47, 0, 110, 17, 88, 199, 60, 0, 60, 0, 60, 0, 184, 255, 60, 0, 13, 17, 60, 0, 60, 0, 60, 0, 73, 49, 60, 0, 60, 0, 60, 0, 185, 255, 60, 0, 14, 17, 60, 0, 60, 0, 60, 0, 74, 49, 60, 0, 60, 0, 60, 0, 105, 50, 60, 0, 60, 0, 60, 0, 119, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 124, 50, 47, 0, 97, 17, 183, 17, 224, 172, 60, 0, 60, 0, 60, 0, 186, 255, 60, 0, 15, 17, 60, 0, 60, 0, 60, 0, 75, 49, 60, 0, 60, 0, 60, 0, 106, 50, 60, 0, 60, 0, 60, 0, 120, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 187, 255, 60, 0, 16, 17, 60, 0, 60, 0, 60, 0, 76, 49, 60, 0, 60, 0, 60, 0, 107, 50, 60, 0, 60, 0, 60, 0, 121, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 188, 255, 60, 0, 17, 17, 60, 0, 60, 0, 60, 0, 77, 49, 60, 0, 60, 0, 60, 0, 108, 50, 60, 0, 60, 0, 60, 0, 122, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 189, 255, 60, 0, 18, 17, 60, 0, 60, 0, 60, 0, 78, 49, 60, 0, 60, 0, 60, 0, 109, 50, 60, 0, 60, 0, 60, 0, 123, 50, 47, 0, 97, 17, 60, 0, 60, 0, 60, 0, 190, 255, 60, 0, 19, 17, 60, 0, 20, 17, 60, 0, 60, 0, 60, 0, 101, 49, 60, 0, 21, 17, 60, 0, 60, 0, 60, 0, 102, 49, 60, 0, 22, 17, 60, 0, 23, 17, 60, 0, 24, 17, 60, 0, 25, 17, 60, 0, 26, 17, 60, 0, 60, 0, 60, 0, 64, 49, 60, 0, 60, 0, 60, 0, 176, 255, 60, 0, 27, 17, 60, 0, 28, 17, 60, 0, 60, 0, 60, 0, 110, 49, 60, 0, 29, 17, 60, 0, 60, 0, 60, 0, 113, 49, 60, 0, 30, 17, 60, 0, 60, 0, 60, 0, 114, 49, 60, 0, 31, 17, 60, 0, 32, 17, 60, 0, 60, 0, 60, 0, 115, 49, 60, 0, 33, 17, 60, 0, 60, 0, 60, 0, 68, 49, 60, 0, 60, 0, 60, 0, 180, 255, 60, 0, 34, 17, 60, 0, 60, 0, 60, 0, 116, 49, 60, 0, 35, 17, 60, 0, 60, 0, 60, 0, 117, 49, 60, 0, 36, 17, 60, 0, 37, 17, 60, 0, 38, 17, 60, 0, 39, 17, 60, 0, 60, 0, 60, 0, 118, 49, 60, 0, 40, 17, 60, 0, 41, 17, 60, 0, 60, 0, 60, 0, 119, 49, 60, 0, 42, 17, 60, 0, 43, 17, 60, 0, 60, 0, 60, 0, 120, 49, 60, 0, 44, 17, 60, 0, 60, 0, 60, 0, 121, 49, 60, 0, 45, 17, 60, 0, 60, 0, 60, 0, 122, 49, 60, 0, 46, 17, 60, 0, 60, 0, 60, 0, 123, 49, 60, 0, 47, 17, 60, 0, 60, 0, 60, 0, 124, 49, 60, 0, 48, 17, 60, 0, 49, 17, 60, 0, 50, 17, 60, 0, 60, 0, 60, 0, 125, 49, 60, 0, 51, 17, 60, 0, 52, 17, 60, 0, 53, 17, 60, 0, 54, 17, 60, 0, 60, 0, 60, 0, 126, 49, 60, 0, 55, 17, 60, 0, 56, 17, 60, 0, 57, 17, 60, 0, 58, 17, 60, 0, 59, 17, 60, 0, 60, 17, 60, 0, 61, 17, 60, 0, 62, 17, 60, 0, 63, 17, 60, 0, 64, 17, 60, 0, 60, 0, 60, 0, 127, 49, 60, 0, 65, 17, 60, 0, 66, 17, 60, 0, 67, 17, 60, 0, 68, 17, 60, 0, 69, 17, 60, 0, 70, 17, 60, 0, 71, 17, 60, 0, 60, 0, 60, 0, 128, 49, 60, 0, 72, 17, 60, 0, 73, 17, 60, 0, 74, 17, 60, 0, 75, 17, 60, 0, 76, 17, 60, 0, 60, 0, 60, 0, 129, 49, 60, 0, 77, 17, 60, 0, 78, 17, 60, 0, 79, 17, 60, 0, 80, 17, 60, 0, 81, 17, 60, 0, 82, 17, 60, 0, 83, 17, 60, 0, 84, 17, 60, 0, 85, 17, 60, 0, 86, 17, 60, 0, 87, 17, 60, 0, 60, 0, 60, 0, 132, 49, 60, 0, 88, 17, 60, 0, 60, 0, 60, 0, 133, 49, 60, 0, 89, 17, 60, 0, 60, 0, 60, 0, 134, 49, 60, 0, 90, 17, 60, 0, 91, 17, 60, 0, 92, 17, 60, 0, 93, 17, 60, 0, 94, 17, 60, 0, 96, 169, 60, 0, 97, 169, 60, 0, 98, 169, 60, 0, 99, 169, 60, 0, 100, 169, 60, 0, 101, 169, 60, 0, 102, 169, 60, 0, 103, 169, 60, 0, 104, 169, 60, 0, 105, 169, 60, 0, 106, 169, 60, 0, 107, 169, 60, 0, 108, 169, 60, 0, 109, 169, 60, 0, 110, 169, 60, 0, 111, 169, 60, 0, 112, 169, 60, 0, 113, 169, 60, 0, 114, 169, 60, 0, 115, 169, 60, 0, 116, 169, 60, 0, 117, 169, 60, 0, 118, 169, 60, 0, 119, 169, 60, 0, 120, 169, 60, 0, 121, 169, 60, 0, 122, 169, 60, 0, 123, 169, 60, 0, 124, 169, 60, 0, 95, 17, 60, 0, 96, 17, 60, 0, 60, 0, 60, 0, 100, 49, 60, 0, 60, 0, 60, 0, 160, 255, 60, 0, 97, 17, 60, 0, 60, 0, 60, 0, 79, 49, 60, 0, 60, 0, 60, 0, 194, 255, 60, 0, 98, 17, 60, 0, 60, 0, 60, 0, 80, 49, 60, 0, 60, 0, 60, 0, 195, 255, 60, 0, 99, 17, 60, 0, 60, 0, 60, 0, 81, 49, 60, 0, 60, 0, 60, 0, 196, 255, 60, 0, 100, 17, 60, 0, 60, 0, 60, 0, 82, 49, 60, 0, 60, 0, 60, 0, 197, 255, 60, 0, 101, 17, 60, 0, 60, 0, 60, 0, 83, 49, 60, 0, 60, 0, 60, 0, 198, 255, 60, 0, 102, 17, 60, 0, 60, 0, 60, 0, 84, 49, 60, 0, 60, 0, 60, 0, 199, 255, 60, 0, 103, 17, 60, 0, 60, 0, 60, 0, 85, 49, 60, 0, 60, 0, 60, 0, 202, 255, 60, 0, 104, 17, 60, 0, 60, 0, 60, 0, 86, 49, 60, 0, 60, 0, 60, 0, 203, 255, 60, 0, 105, 17, 60, 0, 60, 0, 60, 0, 87, 49, 60, 0, 60, 0, 60, 0, 204, 255, 60, 0, 106, 17, 60, 0, 60, 0, 60, 0, 88, 49, 60, 0, 60, 0, 60, 0, 205, 255, 60, 0, 107, 17, 60, 0, 60, 0, 60, 0, 89, 49, 60, 0, 60, 0, 60, 0, 206, 255, 60, 0, 108, 17, 60, 0, 60, 0, 60, 0, 90, 49, 60, 0, 60, 0, 60, 0, 207, 255, 60, 0, 109, 17, 60, 0, 60, 0, 60, 0, 91, 49, 60, 0, 60, 0, 60, 0, 210, 255, 60, 0, 110, 17, 60, 0, 60, 0, 60, 0, 92, 49, 60, 0, 60, 0, 60, 0, 211, 255, 60, 0, 111, 17, 60, 0, 60, 0, 60, 0, 93, 49, 60, 0, 60, 0, 60, 0, 212, 255, 60, 0, 112, 17, 60, 0, 60, 0, 60, 0, 94, 49, 60, 0, 60, 0, 60, 0, 213, 255, 60, 0, 113, 17, 60, 0, 60, 0, 60, 0, 95, 49, 60, 0, 60, 0, 60, 0, 214, 255, 60, 0, 114, 17, 60, 0, 60, 0, 60, 0, 96, 49, 60, 0, 60, 0, 60, 0, 215, 255, 60, 0, 115, 17, 60, 0, 60, 0, 60, 0, 97, 49, 60, 0, 60, 0, 60, 0, 218, 255, 60, 0, 116, 17, 60, 0, 60, 0, 60, 0, 98, 49, 60, 0, 60, 0, 60, 0, 219, 255, 60, 0, 117, 17, 60, 0, 60, 0, 60, 0, 99, 49, 60, 0, 60, 0, 60, 0, 220, 255, 60, 0, 118, 17, 60, 0, 119, 17, 60, 0, 120, 17, 60, 0, 121, 17, 60, 0, 122, 17, 60, 0, 123, 17, 60, 0, 124, 17, 60, 0, 125, 17, 60, 0, 126, 17, 60, 0, 127, 17, 60, 0, 128, 17, 60, 0, 129, 17, 60, 0, 130, 17, 60, 0, 131, 17, 60, 0, 132, 17, 60, 0, 60, 0, 60, 0, 135, 49, 60, 0, 133, 17, 60, 0, 60, 0, 60, 0, 136, 49, 60, 0, 134, 17, 60, 0, 135, 17, 60, 0, 136, 17, 60, 0, 60, 0, 60, 0, 137, 49, 60, 0, 137, 17, 60, 0, 138, 17, 60, 0, 139, 17, 60, 0, 140, 17, 60, 0, 141, 17, 60, 0, 142, 17, 60, 0, 143, 17, 60, 0, 144, 17, 60, 0, 145, 17, 60, 0, 60, 0, 60, 0, 138, 49, 60, 0, 146, 17, 60, 0, 60, 0, 60, 0, 139, 49, 60, 0, 147, 17, 60, 0, 148, 17, 60, 0, 60, 0, 60, 0, 140, 49, 60, 0, 149, 17, 60, 0, 150, 17, 60, 0, 151, 17, 60, 0, 152, 17, 60, 0, 153, 17, 60, 0, 154, 17, 60, 0, 155, 17, 60, 0, 156, 17, 60, 0, 157, 17, 60, 0, 158, 17, 60, 0, 60, 0, 60, 0, 141, 49, 60, 0, 159, 17, 60, 0, 160, 17, 60, 0, 161, 17, 60, 0, 60, 0, 60, 0, 142, 49, 60, 0, 162, 17, 60, 0, 163, 17, 60, 0, 164, 17, 60, 0, 165, 17, 60, 0, 166, 17, 60, 0, 167, 17, 60, 0, 176, 215, 60, 0, 177, 215, 60, 0, 178, 215, 60, 0, 179, 215, 60, 0, 180, 215, 60, 0, 181, 215, 60, 0, 182, 215, 60, 0, 183, 215, 60, 0, 184, 215, 60, 0, 185, 215, 60, 0, 186, 215, 60, 0, 187, 215, 60, 0, 188, 215, 60, 0, 189, 215, 60, 0, 190, 215, 60, 0, 191, 215, 60, 0, 192, 215, 60, 0, 193, 215, 60, 0, 194, 215, 60, 0, 195, 215, 60, 0, 196, 215, 60, 0, 197, 215, 60, 0, 198, 215, 60, 0, 168, 17, 60, 0, 169, 17, 60, 0, 170, 17, 60, 0, 60, 0, 60, 0, 51, 49, 60, 0, 60, 0, 60, 0, 163, 255, 60, 0, 171, 17, 60, 0, 172, 17, 60, 0, 60, 0, 60, 0, 53, 49, 60, 0, 60, 0, 60, 0, 165, 255, 60, 0, 173, 17, 60, 0, 60, 0, 60, 0, 54, 49, 60, 0, 60, 0, 60, 0, 166, 255, 60, 0, 174, 17, 60, 0, 175, 17, 60, 0, 176, 17, 60, 0, 60, 0, 60, 0, 58, 49, 60, 0, 60, 0, 60, 0, 170, 255, 60, 0, 177, 17, 60, 0, 60, 0, 60, 0, 59, 49, 60, 0, 60, 0, 60, 0, 171, 255, 60, 0, 178, 17, 60, 0, 60, 0, 60, 0, 60, 49, 60, 0, 60, 0, 60, 0, 172, 255, 60, 0, 179, 17, 60, 0, 60, 0, 60, 0, 61, 49, 60, 0, 60, 0, 60, 0, 173, 255, 60, 0, 180, 17, 60, 0, 60, 0, 60, 0, 62, 49, 60, 0, 60, 0, 60, 0, 174, 255, 60, 0, 181, 17, 60, 0, 60, 0, 60, 0, 63, 49, 60, 0, 60, 0, 60, 0, 175, 255, 60, 0, 182, 17, 60, 0, 183, 17, 60, 0, 184, 17, 60, 0, 185, 17, 60, 0, 186, 17, 60, 0, 187, 17, 60, 0, 188, 17, 60, 0, 189, 17, 60, 0, 190, 17, 60, 0, 191, 17, 60, 0, 192, 17, 60, 0, 193, 17, 60, 0, 194, 17, 60, 0, 195, 17, 60, 0, 196, 17, 60, 0, 197, 17, 60, 0, 198, 17, 60, 0, 199, 17, 60, 0, 60, 0, 60, 0, 103, 49, 60, 0, 200, 17, 60, 0, 60, 0, 60, 0, 104, 49, 60, 0, 201, 17, 60, 0, 202, 17, 60, 0, 203, 17, 60, 0, 204, 17, 60, 0, 60, 0, 60, 0, 105, 49, 60, 0, 205, 17, 60, 0, 206, 17, 60, 0, 60, 0, 60, 0, 106, 49, 60, 0, 207, 17, 60, 0, 208, 17, 60, 0, 209, 17, 60, 0, 210, 17, 60, 0, 211, 17, 60, 0, 60, 0, 60, 0, 107, 49, 60, 0, 212, 17, 60, 0, 213, 17, 60, 0, 214, 17, 60, 0, 215, 17, 60, 0, 60, 0, 60, 0, 108, 49, 60, 0, 216, 17, 60, 0, 217, 17, 60, 0, 60, 0, 60, 0, 109, 49, 60, 0, 218, 17, 60, 0, 219, 17, 60, 0, 220, 17, 60, 0, 221, 17, 60, 0, 60, 0, 60, 0, 111, 49, 60, 0, 222, 17, 60, 0, 223, 17, 60, 0, 60, 0, 60, 0, 112, 49, 60, 0, 224, 17, 60, 0, 225, 17, 60, 0, 226, 17, 60, 0, 227, 17, 60, 0, 228, 17, 60, 0, 229, 17, 60, 0, 230, 17, 60, 0, 231, 17, 60, 0, 232, 17, 60, 0, 233, 17, 60, 0, 234, 17, 60, 0, 235, 17, 60, 0, 236, 17, 60, 0, 237, 17, 60, 0, 238, 17, 60, 0, 239, 17, 60, 0, 240, 17, 60, 0, 241, 17, 60, 0, 60, 0, 60, 0, 130, 49, 60, 0, 242, 17, 60, 0, 60, 0, 60, 0, 131, 49, 60, 0, 243, 17, 60, 0, 244, 17, 60, 0, 245, 17, 60, 0, 246, 17, 60, 0, 247, 17, 60, 0, 248, 17, 60, 0, 249, 17, 60, 0, 250, 17, 60, 0, 251, 17, 60, 0, 252, 17, 60, 0, 253, 17, 60, 0, 254, 17, 60, 0, 255, 17, 60, 0, 203, 215, 60, 0, 204, 215, 60, 0, 205, 215, 60, 0, 206, 215, 60, 0, 207, 215, 60, 0, 208, 215, 60, 0, 209, 215, 60, 0, 210, 215, 60, 0, 211, 215, 60, 0, 212, 215, 60, 0, 213, 215, 60, 0, 214, 215, 60, 0, 215, 215, 60, 0, 216, 215, 60, 0, 217, 215, 60, 0, 218, 215, 60, 0, 219, 215, 60, 0, 220, 215, 60, 0, 221, 215, 60, 0, 222, 215, 60, 0, 223, 215, 60, 0, 224, 215, 60, 0, 225, 215, 60, 0, 226, 215, 60, 0, 227, 215, 60, 0, 228, 215, 60, 0, 229, 215, 60, 0, 230, 215, 60, 0, 231, 215, 60, 0, 232, 215, 60, 0, 233, 215, 60, 0, 234, 215, 60, 0, 235, 215, 60, 0, 236, 215, 60, 0, 237, 215, 60, 0, 238, 215, 60, 0, 239, 215, 60, 0, 240, 215, 60, 0, 241, 215, 60, 0, 242, 215, 60, 0, 243, 215, 60, 0, 244, 215, 60, 0, 245, 215, 60, 0, 246, 215, 60, 0, 247, 215, 60, 0, 248, 215, 60, 0, 249, 215, 60, 0, 250, 215, 60, 0, 251, 215, 60, 0, 65, 48, 60, 0, 60, 0, 60, 0, 66, 48, 60, 0, 60, 0, 60, 0, 161, 48, 60, 0, 60, 0, 60, 0, 103, 255, 60, 0, 60, 0, 60, 0, 162, 48, 60, 0, 60, 0, 60, 0, 113, 255, 60, 0, 60, 0, 60, 0, 208, 50, 60, 0, 60, 0, 60, 0, 3, 51, 47, 0, 252, 48, 139, 48, 60, 0, 60, 0, 60, 0, 0, 51, 47, 0, 113, 48, 252, 48, 104, 48, 60, 0, 60, 0, 60, 0, 1, 51, 47, 0, 139, 48, 117, 48, 66, 48, 60, 0, 60, 0, 60, 0, 2, 51, 47, 0, 147, 48, 122, 48, 66, 48, 60, 0, 67, 48, 60, 0, 60, 0, 60, 0, 68, 48, 60, 0, 60, 0, 60, 0, 163, 48, 60, 0, 60, 0, 60, 0, 104, 255, 60, 0, 60, 0, 60, 0, 164, 48, 60, 0, 60, 0, 60, 0, 114, 255, 60, 0, 60, 0, 60, 0, 209, 50, 60, 0, 60, 0, 60, 0, 4, 51, 47, 0, 107, 48, 147, 48, 80, 48, 60, 0, 60, 0, 60, 0, 5, 51, 47, 0, 147, 48, 97, 48, 60, 0, 69, 48, 60, 0, 60, 0, 60, 0, 70, 48, 60, 0, 60, 0, 60, 0, 165, 48, 60, 0, 60, 0, 60, 0, 105, 255, 60, 0, 60, 0, 60, 0, 166, 48, 60, 0, 60, 0, 60, 0, 115, 255, 60, 0, 60, 0, 60, 0, 210, 50, 60, 0, 60, 0, 60, 0, 6, 51, 47, 0, 74, 48, 147, 48, 60, 0, 44, 216, 0, 220, 60, 0, 71, 48, 60, 0, 60, 0, 60, 0, 72, 48, 60, 0, 60, 0, 60, 0, 167, 48, 60, 0, 60, 0, 60, 0, 106, 255, 60, 0, 60, 0, 60, 0, 168, 48, 60, 0, 60, 0, 60, 0, 116, 255, 60, 0, 60, 0, 60, 0, 211, 50, 60, 0, 60, 0, 60, 0, 8, 51, 47, 0, 252, 48, 75, 48, 252, 48, 60, 0, 60, 0, 60, 0, 7, 51, 47, 0, 89, 48, 79, 48, 252, 48, 105, 48, 60, 0, 73, 48, 60, 0, 60, 0, 60, 0, 74, 48, 60, 0, 60, 0, 60, 0, 169, 48, 60, 0, 60, 0, 60, 0, 107, 255, 60, 0, 60, 0, 60, 0, 170, 48, 60, 0, 60, 0, 60, 0, 117, 255, 60, 0, 60, 0, 60, 0, 212, 50, 60, 0, 60, 0, 60, 0, 10, 51, 47, 0, 252, 48, 128, 48, 60, 0, 60, 0, 60, 0, 9, 51, 47, 0, 147, 48, 89, 48, 60, 0, 149, 48, 60, 0, 60, 0, 60, 0, 75, 48, 60, 0, 60, 0, 60, 0, 245, 48, 60, 0, 60, 0, 60, 0, 171, 48, 60, 0, 60, 0, 60, 0, 118, 255, 60, 0, 60, 0, 60, 0, 213, 50, 60, 0, 60, 0, 60, 0, 11, 51, 47, 0, 68, 48, 138, 48, 60, 0, 60, 0, 60, 0, 12, 51, 47, 0, 137, 48, 100, 48, 104, 48, 60, 0, 60, 0, 60, 0, 13, 51, 47, 0, 141, 48, 138, 48, 252, 48, 60, 0, 60, 0, 60, 0, 14, 51, 47, 0, 153, 48, 141, 48, 147, 48, 60, 0, 60, 0, 60, 0, 15, 51, 47, 0, 153, 48, 147, 48, 126, 48, 60, 0, 77, 48, 60, 0, 60, 0, 60, 0, 173, 48, 60, 0, 60, 0, 60, 0, 119, 255, 60, 0, 60, 0, 60, 0, 214, 50, 60, 0, 60, 0, 60, 0, 16, 51, 47, 0, 153, 48, 76, 48, 60, 0, 60, 0, 60, 0, 17, 51, 47, 0, 153, 48, 107, 48, 252, 48, 60, 0, 60, 0, 60, 0, 18, 51, 47, 0, 134, 48, 138, 48, 252, 48, 60, 0, 60, 0, 60, 0, 19, 51, 47, 0, 153, 48, 139, 48, 96, 48, 252, 48, 60, 0, 60, 0, 60, 0, 20, 51, 47, 0, 141, 48, 60, 0, 60, 0, 60, 0, 21, 51, 47, 0, 141, 48, 80, 48, 137, 48, 128, 48, 60, 0, 60, 0, 60, 0, 22, 51, 47, 0, 141, 48, 129, 48, 252, 48, 104, 48, 139, 48, 60, 0, 60, 0, 60, 0, 23, 51, 47, 0, 141, 48, 143, 48, 100, 48, 104, 48, 60, 0, 79, 48, 60, 0, 60, 0, 60, 0, 240, 49, 60, 0, 60, 0, 60, 0, 175, 48, 60, 0, 60, 0, 60, 0, 120, 255, 60, 0, 60, 0, 60, 0, 215, 50, 60, 0, 60, 0, 60, 0, 24, 51, 47, 0, 153, 48, 137, 48, 128, 48, 60, 0, 60, 0, 60, 0, 25, 51, 47, 0, 153, 48, 137, 48, 128, 48, 104, 48, 147, 48, 60, 0, 60, 0, 60, 0, 26, 51, 47, 0, 139, 48, 92, 48, 68, 48, 141, 48, 60, 0, 60, 0, 60, 0, 27, 51, 47, 0, 141, 48, 252, 48, 109, 48, 60, 0, 150, 48, 60, 0, 60, 0, 60, 0, 81, 48, 60, 0, 60, 0, 60, 0, 246, 48, 60, 0, 60, 0, 60, 0, 177, 48, 60, 0, 60, 0, 60, 0, 121, 255, 60, 0, 60, 0, 60, 0, 216, 50, 60, 0, 60, 0, 60, 0, 28, 51, 47, 0, 252, 48, 89, 48, 60, 0, 83, 48, 60, 0, 60, 0, 60, 0, 179, 48, 60, 0, 60, 0, 60, 0, 122, 255, 60, 0, 60, 0, 60, 0, 217, 50, 60, 0, 60, 0, 60, 0, 255, 48, 47, 0, 104, 48, 60, 0, 60, 0, 60, 0, 30, 51, 47, 0, 252, 48, 125, 48, 60, 0, 60, 0, 60, 0, 60, 216, 1, 222, 47, 0, 83, 48, 60, 0, 60, 0, 60, 0, 29, 51, 47, 0, 139, 48, 106, 48, 60, 0, 85, 48, 60, 0, 60, 0, 60, 0, 181, 48, 60, 0, 60, 0, 60, 0, 123, 255, 60, 0, 60, 0, 60, 0, 218, 50, 60, 0, 60, 0, 60, 0, 60, 216, 2, 222, 60, 0, 60, 0, 60, 0, 31, 51, 47, 0, 68, 48, 79, 48, 139, 48, 60, 0, 60, 0, 60, 0, 32, 51, 47, 0, 147, 48, 97, 48, 252, 48, 128, 48, 60, 0, 6, 48, 47, 0, 129, 48, 60, 0, 60, 0, 60, 0, 87, 48, 60, 0, 60, 0, 60, 0, 241, 49, 60, 0, 60, 0, 60, 0, 183, 48, 60, 0, 60, 0, 60, 0, 124, 255, 60, 0, 60, 0, 60, 0, 219, 50, 60, 0, 60, 0, 60, 0, 33, 51, 47, 0, 138, 48, 147, 48, 80, 48, 60, 0, 89, 48, 60, 0, 60, 0, 60, 0, 242, 49, 60, 0, 60, 0, 60, 0, 185, 48, 60, 0, 60, 0, 60, 0, 125, 255, 60, 0, 60, 0, 60, 0, 220, 50, 60, 0, 91, 48, 60, 0, 60, 0, 60, 0, 187, 48, 60, 0, 60, 0, 60, 0, 126, 255, 60, 0, 60, 0, 60, 0, 221, 50, 60, 0, 60, 0, 60, 0, 34, 51, 47, 0, 147, 48, 97, 48, 60, 0, 60, 0, 60, 0, 35, 51, 47, 0, 147, 48, 104, 48, 60, 0, 93, 48, 60, 0, 60, 0, 60, 0, 189, 48, 60, 0, 60, 0, 60, 0, 127, 255, 60, 0, 60, 0, 60, 0, 222, 50, 60, 0, 95, 48, 60, 0, 60, 0, 60, 0, 191, 48, 60, 0, 60, 0, 60, 0, 128, 255, 60, 0, 60, 0, 60, 0, 223, 50, 60, 0, 60, 0, 60, 0, 36, 51, 47, 0, 153, 48, 252, 48, 89, 48, 60, 0, 97, 48, 60, 0, 60, 0, 60, 0, 193, 48, 60, 0, 60, 0, 60, 0, 129, 255, 60, 0, 60, 0, 60, 0, 224, 50, 60, 0, 99, 48, 60, 0, 60, 0, 60, 0, 100, 48, 60, 0, 60, 0, 60, 0, 195, 48, 60, 0, 60, 0, 60, 0, 111, 255, 60, 0, 60, 0, 60, 0, 196, 48, 60, 0, 60, 0, 60, 0, 130, 255, 60, 0, 60, 0, 60, 0, 225, 50, 60, 0, 102, 48, 60, 0, 60, 0, 60, 0, 198, 48, 60, 0, 60, 0, 60, 0, 131, 255, 60, 0, 60, 0, 60, 0, 226, 50, 60, 0, 60, 0, 60, 0, 60, 216, 19, 222, 47, 0, 153, 48, 60, 0, 60, 0, 60, 0, 37, 51, 47, 0, 153, 48, 87, 48, 60, 0, 104, 48, 60, 0, 60, 0, 60, 0, 243, 49, 60, 0, 60, 0, 60, 0, 200, 48, 60, 0, 60, 0, 60, 0, 132, 255, 60, 0, 60, 0, 60, 0, 227, 50, 60, 0, 60, 0, 60, 0, 38, 51, 47, 0, 153, 48, 139, 48, 60, 0, 60, 0, 60, 0, 39, 51, 47, 0, 147, 48, 60, 0, 106, 48, 60, 0, 60, 0, 60, 0, 202, 48, 60, 0, 60, 0, 60, 0, 133, 255, 60, 0, 60, 0, 60, 0, 228, 50, 60, 0, 60, 0, 60, 0, 40, 51, 47, 0, 110, 48, 60, 0, 107, 48, 60, 0, 60, 0, 60, 0, 203, 48, 60, 0, 60, 0, 60, 0, 134, 255, 60, 0, 60, 0, 60, 0, 229, 50, 60, 0, 108, 48, 60, 0, 60, 0, 60, 0, 244, 49, 60, 0, 60, 0, 60, 0, 204, 48, 60, 0, 60, 0, 60, 0, 135, 255, 60, 0, 60, 0, 60, 0, 230, 50, 60, 0, 109, 48, 60, 0, 60, 0, 60, 0, 205, 48, 60, 0, 60, 0, 60, 0, 136, 255, 60, 0, 60, 0, 60, 0, 231, 50, 60, 0, 110, 48, 60, 0, 60, 0, 60, 0, 206, 48, 60, 0, 60, 0, 60, 0, 137, 255, 60, 0, 60, 0, 60, 0, 232, 50, 60, 0, 60, 0, 60, 0, 41, 51, 47, 0, 100, 48, 104, 48, 60, 0, 111, 48, 60, 0, 60, 0, 60, 0, 245, 49, 60, 0, 60, 0, 60, 0, 207, 48, 60, 0, 60, 0, 60, 0, 138, 255, 60, 0, 60, 0, 60, 0, 233, 50, 60, 0, 60, 0, 60, 0, 43, 51, 47, 0, 154, 48, 252, 48, 91, 48, 147, 48, 104, 48, 60, 0, 60, 0, 60, 0, 44, 51, 47, 0, 154, 48, 252, 48, 100, 48, 60, 0, 60, 0, 60, 0, 45, 51, 47, 0, 153, 48, 252, 48, 140, 48, 139, 48, 60, 0, 60, 0, 60, 0, 42, 51, 47, 0, 68, 48, 100, 48, 60, 0, 114, 48, 60, 0, 60, 0, 60, 0, 246, 49, 60, 0, 60, 0, 60, 0, 210, 48, 60, 0, 60, 0, 60, 0, 139, 255, 60, 0, 60, 0, 60, 0, 234, 50, 60, 0, 60, 0, 60, 0, 46, 51, 47, 0, 154, 48, 66, 48, 89, 48, 104, 48, 139, 48, 60, 0, 60, 0, 60, 0, 47, 51, 47, 0, 154, 48, 79, 48, 139, 48, 60, 0, 60, 0, 60, 0, 48, 51, 47, 0, 154, 48, 83, 48, 60, 0, 60, 0, 60, 0, 49, 51, 47, 0, 153, 48, 139, 48, 60, 0, 117, 48, 60, 0, 60, 0, 60, 0, 247, 49, 60, 0, 60, 0, 60, 0, 213, 48, 60, 0, 60, 0, 60, 0, 140, 255, 60, 0, 60, 0, 60, 0, 235, 50, 60, 0, 60, 0, 60, 0, 50, 51, 47, 0, 66, 48, 137, 48, 100, 48, 105, 48, 60, 0, 60, 0, 60, 0, 51, 51, 47, 0, 68, 48, 252, 48, 104, 48, 60, 0, 60, 0, 60, 0, 52, 51, 47, 0, 153, 48, 100, 48, 87, 48, 72, 48, 139, 48, 60, 0, 60, 0, 60, 0, 53, 51, 47, 0, 137, 48, 147, 48, 60, 0, 120, 48, 60, 0, 60, 0, 60, 0, 248, 49, 60, 0, 60, 0, 60, 0, 216, 48, 60, 0, 60, 0, 60, 0, 141, 255, 60, 0, 60, 0, 60, 0, 236, 50, 60, 0, 60, 0, 60, 0, 59, 51, 47, 0, 154, 48, 252, 48, 88, 48, 60, 0, 60, 0, 60, 0, 60, 51, 47, 0, 153, 48, 252, 48, 95, 48, 60, 0, 60, 0, 60, 0, 54, 51, 47, 0, 79, 48, 95, 48, 252, 48, 139, 48, 60, 0, 60, 0, 60, 0, 55, 51, 47, 0, 154, 48, 93, 48, 60, 0, 60, 0, 60, 0, 56, 51, 47, 0, 154, 48, 107, 48, 114, 48, 60, 0, 60, 0, 60, 0, 57, 51, 47, 0, 139, 48, 100, 48, 60, 0, 60, 0, 60, 0, 58, 51, 47, 0, 154, 48, 147, 48, 89, 48, 60, 0, 123, 48, 60, 0, 60, 0, 60, 0, 249, 49, 60, 0, 60, 0, 60, 0, 219, 48, 60, 0, 60, 0, 60, 0, 142, 255, 60, 0, 60, 0, 60, 0, 237, 50, 60, 0, 60, 0, 60, 0, 65, 51, 47, 0, 252, 48, 139, 48, 60, 0, 60, 0, 60, 0, 66, 51, 47, 0, 252, 48, 147, 48, 60, 0, 60, 0, 60, 0, 61, 51, 47, 0, 154, 48, 68, 48, 147, 48, 104, 48, 60, 0, 60, 0, 60, 0, 60, 216, 0, 222, 47, 0, 75, 48, 60, 0, 60, 0, 60, 0, 62, 51, 47, 0, 153, 48, 139, 48, 104, 48, 60, 0, 60, 0, 60, 0, 63, 51, 47, 0, 147, 48, 60, 0, 60, 0, 60, 0, 64, 51, 47, 0, 154, 48, 147, 48, 105, 48, 60, 0, 60, 48, 47, 0, 89, 48, 60, 0, 60, 0, 60, 0, 126, 48, 60, 0, 60, 0, 60, 0, 222, 48, 60, 0, 60, 0, 60, 0, 143, 255, 60, 0, 60, 0, 60, 0, 238, 50, 60, 0, 60, 0, 60, 0, 67, 51, 47, 0, 68, 48, 79, 48, 141, 48, 60, 0, 60, 0, 60, 0, 68, 51, 47, 0, 68, 48, 139, 48, 60, 0, 60, 0, 60, 0, 69, 51, 47, 0, 100, 48, 111, 48, 60, 0, 60, 0, 60, 0, 70, 51, 47, 0, 139, 48, 79, 48, 60, 0, 60, 0, 60, 0, 71, 51, 47, 0, 147, 48, 87, 48, 136, 48, 147, 48, 60, 0, 127, 48, 60, 0, 60, 0, 60, 0, 223, 48, 60, 0, 60, 0, 60, 0, 144, 255, 60, 0, 60, 0, 60, 0, 239, 50, 60, 0, 60, 0, 60, 0, 72, 51, 47, 0, 79, 48, 141, 48, 147, 48, 60, 0, 60, 0, 60, 0, 73, 51, 47, 0, 138, 48, 60, 0, 60, 0, 60, 0, 74, 51, 47, 0, 138, 48, 112, 48, 252, 48, 139, 48, 60, 0, 128, 48, 60, 0, 60, 0, 60, 0, 250, 49, 60, 0, 60, 0, 60, 0, 224, 48, 60, 0, 60, 0, 60, 0, 145, 255, 60, 0, 60, 0, 60, 0, 240, 50, 60, 0, 129, 48, 60, 0, 60, 0, 60, 0, 225, 48, 60, 0, 60, 0, 60, 0, 146, 255, 60, 0, 60, 0, 60, 0, 241, 50, 60, 0, 60, 0, 60, 0, 77, 51, 47, 0, 252, 48, 104, 48, 139, 48, 60, 0, 60, 0, 60, 0, 75, 51, 47, 0, 76, 48, 60, 0, 60, 0, 60, 0, 76, 51, 47, 0, 76, 48, 104, 48, 147, 48, 60, 0, 130, 48, 60, 0, 60, 0, 60, 0, 226, 48, 60, 0, 60, 0, 60, 0, 147, 255, 60, 0, 60, 0, 60, 0, 242, 50, 60, 0, 131, 48, 60, 0, 60, 0, 60, 0, 132, 48, 60, 0, 60, 0, 60, 0, 227, 48, 60, 0, 60, 0, 60, 0, 108, 255, 60, 0, 60, 0, 60, 0, 228, 48, 60, 0, 60, 0, 60, 0, 148, 255, 60, 0, 60, 0, 60, 0, 243, 50, 60, 0, 60, 0, 60, 0, 78, 51, 47, 0, 252, 48, 105, 48, 60, 0, 60, 0, 60, 0, 79, 51, 47, 0, 252, 48, 139, 48, 60, 0, 133, 48, 60, 0, 60, 0, 60, 0, 134, 48, 60, 0, 60, 0, 60, 0, 229, 48, 60, 0, 60, 0, 60, 0, 109, 255, 60, 0, 60, 0, 60, 0, 230, 48, 60, 0, 60, 0, 60, 0, 149, 255, 60, 0, 60, 0, 60, 0, 244, 50, 60, 0, 60, 0, 60, 0, 80, 51, 47, 0, 66, 48, 147, 48, 60, 0, 135, 48, 60, 0, 60, 0, 60, 0, 136, 48, 60, 0, 60, 0, 60, 0, 231, 48, 60, 0, 60, 0, 60, 0, 110, 255, 60, 0, 60, 0, 60, 0, 232, 48, 60, 0, 60, 0, 60, 0, 150, 255, 60, 0, 60, 0, 60, 0, 245, 50, 60, 0, 60, 0, 60, 0, 159, 48, 47, 0, 138, 48, 60, 0, 137, 48, 60, 0, 60, 0, 60, 0, 251, 49, 60, 0, 60, 0, 60, 0, 233, 48, 60, 0, 60, 0, 60, 0, 151, 255, 60, 0, 60, 0, 60, 0, 246, 50, 60, 0, 138, 48, 60, 0, 60, 0, 60, 0, 252, 49, 60, 0, 60, 0, 60, 0, 234, 48, 60, 0, 60, 0, 60, 0, 152, 255, 60, 0, 60, 0, 60, 0, 247, 50, 60, 0, 60, 0, 60, 0, 81, 51, 47, 0, 100, 48, 104, 48, 139, 48, 60, 0, 60, 0, 60, 0, 82, 51, 47, 0, 137, 48, 60, 0, 139, 48, 60, 0, 60, 0, 60, 0, 253, 49, 60, 0, 60, 0, 60, 0, 235, 48, 60, 0, 60, 0, 60, 0, 153, 255, 60, 0, 60, 0, 60, 0, 248, 50, 60, 0, 60, 0, 60, 0, 84, 51, 47, 0, 252, 48, 118, 48, 139, 48, 60, 0, 60, 0, 60, 0, 83, 51, 47, 0, 116, 48, 252, 48, 60, 0, 140, 48, 60, 0, 60, 0, 60, 0, 254, 49, 60, 0, 60, 0, 60, 0, 236, 48, 60, 0, 60, 0, 60, 0, 154, 255, 60, 0, 60, 0, 60, 0, 249, 50, 60, 0, 60, 0, 60, 0, 85, 51, 47, 0, 128, 48, 60, 0, 60, 0, 60, 0, 86, 51, 47, 0, 147, 48, 104, 48, 82, 48, 147, 48, 60, 0, 141, 48, 60, 0, 60, 0, 60, 0, 255, 49, 60, 0, 60, 0, 60, 0, 237, 48, 60, 0, 60, 0, 60, 0, 155, 255, 60, 0, 60, 0, 60, 0, 250, 50, 60, 0, 142, 48, 60, 0, 60, 0, 60, 0, 143, 48, 60, 0, 60, 0, 60, 0, 238, 48, 60, 0, 60, 0, 60, 0, 239, 48, 60, 0, 60, 0, 60, 0, 156, 255, 60, 0, 60, 0, 60, 0, 251, 50, 60, 0, 60, 0, 60, 0, 87, 51, 47, 0, 100, 48, 104, 48, 60, 0, 44, 216, 80, 221, 60, 0, 60, 0, 60, 0, 144, 48, 60, 0, 60, 0, 60, 0, 44, 216, 100, 221, 60, 0, 60, 0, 60, 0, 240, 48, 60, 0, 60, 0, 60, 0, 252, 50, 60, 0, 44, 216, 81, 221, 60, 0, 60, 0, 60, 0, 145, 48, 60, 0, 60, 0, 60, 0, 44, 216, 101, 221, 60, 0, 60, 0, 60, 0, 241, 48, 60, 0, 60, 0, 60, 0, 253, 50, 60, 0, 44, 216, 82, 221, 60, 0, 60, 0, 60, 0, 146, 48, 60, 0, 60, 0, 60, 0, 44, 216, 102, 221, 60, 0, 60, 0, 60, 0, 242, 48, 60, 0, 60, 0, 60, 0, 102, 255, 60, 0, 60, 0, 60, 0, 254, 50, 60, 0, 147, 48, 60, 0, 60, 0, 60, 0, 44, 216, 103, 221, 60, 0, 60, 0, 60, 0, 243, 48, 60, 0, 60, 0, 60, 0, 157, 255, 60, 0, 44, 216, 2, 220, 60, 0, 44, 216, 3, 220, 60, 0, 44, 216, 4, 220, 60, 0, 44, 216, 5, 220, 60, 0, 44, 216, 6, 220, 60, 0, 44, 216, 7, 220, 60, 0, 44, 216, 8, 220, 60, 0, 44, 216, 9, 220, 60, 0, 44, 216, 10, 220, 60, 0, 44, 216, 11, 220, 60, 0, 44, 216, 12, 220, 60, 0, 44, 216, 13, 220, 60, 0, 44, 216, 14, 220, 60, 0, 44, 216, 1, 220, 60, 0, 44, 216, 15, 220, 60, 0, 44, 216, 16, 220, 60, 0, 44, 216, 17, 220, 60, 0, 44, 216, 18, 220, 60, 0, 44, 216, 19, 220, 60, 0, 44, 216, 20, 220, 60, 0, 44, 216, 21, 220, 60, 0, 44, 216, 22, 220, 60, 0, 44, 216, 23, 220, 60, 0, 44, 216, 24, 220, 60, 0, 44, 216, 25, 220, 60, 0, 44, 216, 26, 220, 60, 0, 44, 216, 27, 220, 60, 0, 44, 216, 28, 220, 60, 0, 44, 216, 29, 220, 60, 0, 44, 216, 30, 220, 60, 0, 44, 216, 31, 220, 60, 0, 44, 216, 32, 220, 60, 0, 44, 216, 33, 220, 60, 0, 44, 216, 34, 220, 60, 0, 44, 216, 35, 220, 60, 0, 44, 216, 36, 220, 60, 0, 44, 216, 37, 220, 60, 0, 44, 216, 38, 220, 60, 0, 44, 216, 39, 220, 60, 0, 44, 216, 40, 220, 60, 0, 44, 216, 41, 220, 60, 0, 44, 216, 42, 220, 60, 0, 44, 216, 43, 220, 60, 0, 44, 216, 44, 220, 60, 0, 44, 216, 45, 220, 60, 0, 44, 216, 46, 220, 60, 0, 44, 216, 47, 220, 60, 0, 44, 216, 48, 220, 60, 0, 44, 216, 49, 220, 60, 0, 44, 216, 50, 220, 60, 0, 44, 216, 51, 220, 60, 0, 44, 216, 52, 220, 60, 0, 44, 216, 53, 220, 60, 0, 44, 216, 54, 220, 60, 0, 44, 216, 55, 220, 60, 0, 44, 216, 56, 220, 60, 0, 44, 216, 57, 220, 60, 0, 44, 216, 58, 220, 60, 0, 44, 216, 59, 220, 60, 0, 44, 216, 60, 220, 60, 0, 44, 216, 61, 220, 60, 0, 44, 216, 62, 220, 60, 0, 44, 216, 63, 220, 60, 0, 44, 216, 64, 220, 60, 0, 44, 216, 65, 220, 60, 0, 44, 216, 66, 220, 60, 0, 44, 216, 67, 220, 60, 0, 44, 216, 68, 220, 60, 0, 44, 216, 69, 220, 60, 0, 44, 216, 70, 220, 60, 0, 44, 216, 71, 220, 60, 0, 44, 216, 72, 220, 60, 0, 44, 216, 73, 220, 60, 0, 44, 216, 74, 220, 60, 0, 44, 216, 75, 220, 60, 0, 44, 216, 76, 220, 60, 0, 44, 216, 77, 220, 60, 0, 44, 216, 78, 220, 60, 0, 44, 216, 79, 220, 60, 0, 44, 216, 80, 220, 60, 0, 44, 216, 81, 220, 60, 0, 44, 216, 82, 220, 60, 0, 44, 216, 83, 220, 60, 0, 44, 216, 84, 220, 60, 0, 44, 216, 85, 220, 60, 0, 44, 216, 86, 220, 60, 0, 44, 216, 87, 220, 60, 0, 44, 216, 88, 220, 60, 0, 44, 216, 89, 220, 60, 0, 44, 216, 90, 220, 60, 0, 44, 216, 91, 220, 60, 0, 44, 216, 92, 220, 60, 0, 44, 216, 93, 220, 60, 0, 44, 216, 94, 220, 60, 0, 44, 216, 95, 220, 60, 0, 44, 216, 96, 220, 60, 0, 44, 216, 97, 220, 60, 0, 44, 216, 98, 220, 60, 0, 44, 216, 99, 220, 60, 0, 44, 216, 100, 220, 60, 0, 44, 216, 101, 220, 60, 0, 44, 216, 102, 220, 60, 0, 44, 216, 103, 220, 60, 0, 44, 216, 104, 220, 60, 0, 44, 216, 105, 220, 60, 0, 44, 216, 106, 220, 60, 0, 44, 216, 107, 220, 60, 0, 44, 216, 108, 220, 60, 0, 44, 216, 109, 220, 60, 0, 44, 216, 110, 220, 60, 0, 44, 216, 111, 220, 60, 0, 44, 216, 112, 220, 60, 0, 44, 216, 113, 220, 60, 0, 44, 216, 114, 220, 60, 0, 44, 216, 115, 220, 60, 0, 44, 216, 116, 220, 60, 0, 44, 216, 117, 220, 60, 0, 44, 216, 118, 220, 60, 0, 44, 216, 119, 220, 60, 0, 44, 216, 120, 220, 60, 0, 44, 216, 121, 220, 60, 0, 44, 216, 122, 220, 60, 0, 44, 216, 123, 220, 60, 0, 44, 216, 124, 220, 60, 0, 44, 216, 125, 220, 60, 0, 44, 216, 126, 220, 60, 0, 44, 216, 127, 220, 60, 0, 44, 216, 128, 220, 60, 0, 44, 216, 129, 220, 60, 0, 44, 216, 130, 220, 60, 0, 44, 216, 131, 220, 60, 0, 44, 216, 132, 220, 60, 0, 44, 216, 133, 220, 60, 0, 44, 216, 134, 220, 60, 0, 44, 216, 135, 220, 60, 0, 44, 216, 136, 220, 60, 0, 44, 216, 137, 220, 60, 0, 44, 216, 138, 220, 60, 0, 44, 216, 139, 220, 60, 0, 44, 216, 140, 220, 60, 0, 44, 216, 141, 220, 60, 0, 44, 216, 142, 220, 60, 0, 44, 216, 143, 220, 60, 0, 44, 216, 144, 220, 60, 0, 44, 216, 145, 220, 60, 0, 44, 216, 146, 220, 60, 0, 44, 216, 147, 220, 60, 0, 44, 216, 148, 220, 60, 0, 44, 216, 149, 220, 60, 0, 44, 216, 150, 220, 60, 0, 44, 216, 151, 220, 60, 0, 44, 216, 152, 220, 60, 0, 44, 216, 153, 220, 60, 0, 44, 216, 154, 220, 60, 0, 44, 216, 155, 220, 60, 0, 44, 216, 156, 220, 60, 0, 44, 216, 157, 220, 60, 0, 44, 216, 158, 220, 60, 0, 44, 216, 159, 220, 60, 0, 44, 216, 160, 220, 60, 0, 44, 216, 161, 220, 60, 0, 44, 216, 162, 220, 60, 0, 44, 216, 163, 220, 60, 0, 44, 216, 164, 220, 60, 0, 44, 216, 165, 220, 60, 0, 44, 216, 166, 220, 60, 0, 44, 216, 167, 220, 60, 0, 44, 216, 168, 220, 60, 0, 44, 216, 169, 220, 60, 0, 44, 216, 170, 220, 60, 0, 44, 216, 171, 220, 60, 0, 44, 216, 172, 220, 60, 0, 44, 216, 173, 220, 60, 0, 44, 216, 174, 220, 60, 0, 44, 216, 175, 220, 60, 0, 44, 216, 176, 220, 60, 0, 44, 216, 177, 220, 60, 0, 44, 216, 178, 220, 60, 0, 44, 216, 179, 220, 60, 0, 44, 216, 180, 220, 60, 0, 44, 216, 181, 220, 60, 0, 44, 216, 182, 220, 60, 0, 44, 216, 183, 220, 60, 0, 44, 216, 184, 220, 60, 0, 44, 216, 185, 220, 60, 0, 44, 216, 186, 220, 60, 0, 44, 216, 187, 220, 60, 0, 44, 216, 188, 220, 60, 0, 44, 216, 189, 220, 60, 0, 44, 216, 190, 220, 60, 0, 44, 216, 191, 220, 60, 0, 44, 216, 192, 220, 60, 0, 44, 216, 193, 220, 60, 0, 44, 216, 194, 220, 60, 0, 44, 216, 195, 220, 60, 0, 44, 216, 196, 220, 60, 0, 44, 216, 197, 220, 60, 0, 44, 216, 198, 220, 60, 0, 44, 216, 199, 220, 60, 0, 44, 216, 200, 220, 60, 0, 44, 216, 201, 220, 60, 0, 44, 216, 202, 220, 60, 0, 44, 216, 203, 220, 60, 0, 44, 216, 204, 220, 60, 0, 44, 216, 205, 220, 60, 0, 44, 216, 206, 220, 60, 0, 44, 216, 207, 220, 60, 0, 44, 216, 208, 220, 60, 0, 44, 216, 209, 220, 60, 0, 44, 216, 210, 220, 60, 0, 44, 216, 211, 220, 60, 0, 44, 216, 212, 220, 60, 0, 44, 216, 213, 220, 60, 0, 44, 216, 214, 220, 60, 0, 44, 216, 215, 220, 60, 0, 44, 216, 216, 220, 60, 0, 44, 216, 217, 220, 60, 0, 44, 216, 218, 220, 60, 0, 44, 216, 219, 220, 60, 0, 44, 216, 220, 220, 60, 0, 44, 216, 221, 220, 60, 0, 44, 216, 222, 220, 60, 0, 44, 216, 223, 220, 60, 0, 44, 216, 224, 220, 60, 0, 44, 216, 225, 220, 60, 0, 44, 216, 226, 220, 60, 0, 44, 216, 227, 220, 60, 0, 44, 216, 228, 220, 60, 0, 44, 216, 229, 220, 60, 0, 44, 216, 230, 220, 60, 0, 44, 216, 231, 220, 60, 0, 44, 216, 232, 220, 60, 0, 44, 216, 233, 220, 60, 0, 44, 216, 234, 220, 60, 0, 44, 216, 235, 220, 60, 0, 44, 216, 236, 220, 60, 0, 44, 216, 237, 220, 60, 0, 44, 216, 238, 220, 60, 0, 44, 216, 239, 220, 60, 0, 44, 216, 240, 220, 60, 0, 44, 216, 241, 220, 60, 0, 44, 216, 242, 220, 60, 0, 44, 216, 243, 220, 60, 0, 44, 216, 244, 220, 60, 0, 44, 216, 245, 220, 60, 0, 44, 216, 246, 220, 60, 0, 44, 216, 247, 220, 60, 0, 44, 216, 248, 220, 60, 0, 44, 216, 249, 220, 60, 0, 44, 216, 250, 220, 60, 0, 44, 216, 251, 220, 60, 0, 44, 216, 252, 220, 60, 0, 44, 216, 253, 220, 60, 0, 44, 216, 254, 220, 60, 0, 44, 216, 255, 220, 60, 0, 44, 216, 0, 221, 60, 0, 44, 216, 1, 221, 60, 0, 44, 216, 2, 221, 60, 0, 44, 216, 3, 221, 60, 0, 44, 216, 4, 221, 60, 0, 44, 216, 5, 221, 60, 0, 44, 216, 6, 221, 60, 0, 44, 216, 7, 221, 60, 0, 44, 216, 8, 221, 60, 0, 44, 216, 9, 221, 60, 0, 44, 216, 10, 221, 60, 0, 44, 216, 11, 221, 60, 0, 44, 216, 12, 221, 60, 0, 44, 216, 13, 221, 60, 0, 44, 216, 14, 221, 60, 0, 44, 216, 15, 221, 60, 0, 44, 216, 16, 221, 60, 0, 44, 216, 17, 221, 60, 0, 44, 216, 18, 221, 60, 0, 44, 216, 19, 221, 60, 0, 44, 216, 20, 221, 60, 0, 44, 216, 21, 221, 60, 0, 44, 216, 22, 221, 60, 0, 44, 216, 23, 221, 60, 0, 44, 216, 24, 221, 60, 0, 44, 216, 25, 221, 60, 0, 44, 216, 26, 221, 60, 0, 44, 216, 27, 221, 60, 0, 44, 216, 28, 221, 60, 0, 44, 216, 29, 221, 60, 0, 44, 216, 30, 221, 60, 0, 5, 49, 60, 0, 60, 0, 160, 49, 60, 0, 6, 49, 60, 0, 60, 0, 60, 0, 180, 49, 60, 0, 7, 49, 60, 0, 172, 49, 60, 0, 8, 49, 60, 0, 42, 49, 60, 0, 9, 49, 60, 0, 10, 49, 60, 0, 60, 0, 60, 0, 181, 49, 60, 0, 11, 49, 60, 0, 47, 49, 60, 0, 12, 49, 60, 0, 13, 49, 60, 0, 60, 0, 60, 0, 187, 49, 60, 0, 60, 0, 163, 49, 60, 0, 14, 49, 60, 0, 60, 0, 60, 0, 182, 49, 60, 0, 43, 49, 60, 0, 173, 49, 60, 0, 15, 49, 60, 0, 60, 0, 60, 0, 183, 49, 60, 0, 16, 49, 60, 0, 60, 0, 162, 49, 60, 0, 17, 49, 60, 0, 18, 49, 60, 0, 44, 49, 60, 0, 188, 49, 60, 0, 189, 49, 60, 0, 19, 49, 60, 0, 20, 49, 60, 0, 21, 49, 60, 0, 22, 49, 60, 0, 23, 49, 60, 0, 60, 0, 161, 49, 60, 0, 24, 49, 60, 0, 25, 49, 60, 0, 184, 49, 60, 0, 185, 49, 60, 0, 186, 49, 60, 0, 26, 49, 60, 0, 60, 0, 169, 49, 60, 0, 27, 49, 60, 0, 60, 0, 167, 49, 60, 0, 166, 49, 60, 0, 28, 49, 60, 0, 60, 0, 60, 0, 46, 49, 60, 0, 29, 49, 60, 0, 164, 49, 60, 0, 60, 0, 165, 49, 60, 0, 30, 49, 60, 0, 60, 0, 174, 49, 60, 0, 31, 49, 60, 0, 32, 49, 60, 0, 60, 0, 175, 49, 60, 0, 33, 49, 60, 0, 34, 49, 60, 0, 35, 49, 60, 0, 36, 49, 60, 0, 178, 49, 60, 0, 37, 49, 60, 0, 176, 49, 60, 0, 177, 49, 60, 0, 38, 49, 60, 0, 39, 49, 60, 0, 60, 0, 170, 49, 60, 0, 60, 0, 60, 0, 179, 49, 60, 0, 40, 49, 60, 0, 60, 0, 171, 49, 60, 0, 60, 0, 168, 49, 60, 0, 41, 49, 60, 0, 45, 49, 60, 0, 190, 49, 60, 0, 191, 49, 60, 0, 0, 160, 60, 0, 1, 160, 60, 0, 2, 160, 60, 0, 3, 160, 60, 0, 4, 160, 60, 0, 5, 160, 60, 0, 6, 160, 60, 0, 7, 160, 60, 0, 8, 160, 60, 0, 9, 160, 60, 0, 10, 160, 60, 0, 11, 160, 60, 0, 12, 160, 60, 0, 13, 160, 60, 0, 14, 160, 60, 0, 15, 160, 60, 0, 16, 160, 60, 0, 17, 160, 60, 0, 18, 160, 60, 0, 19, 160, 60, 0, 20, 160, 60, 0, 21, 160, 60, 0, 22, 160, 60, 0, 23, 160, 60, 0, 24, 160, 60, 0, 25, 160, 60, 0, 26, 160, 60, 0, 27, 160, 60, 0, 28, 160, 60, 0, 29, 160, 60, 0, 30, 160, 60, 0, 31, 160, 60, 0, 32, 160, 60, 0, 33, 160, 60, 0, 34, 160, 60, 0, 35, 160, 60, 0, 36, 160, 60, 0, 37, 160, 60, 0, 38, 160, 60, 0, 39, 160, 60, 0, 40, 160, 60, 0, 41, 160, 60, 0, 42, 160, 60, 0, 43, 160, 60, 0, 44, 160, 60, 0, 45, 160, 60, 0, 46, 160, 60, 0, 47, 160, 60, 0, 48, 160, 60, 0, 49, 160, 60, 0, 50, 160, 60, 0, 51, 160, 60, 0, 52, 160, 60, 0, 53, 160, 60, 0, 54, 160, 60, 0, 55, 160, 60, 0, 56, 160, 60, 0, 57, 160, 60, 0, 58, 160, 60, 0, 59, 160, 60, 0, 60, 160, 60, 0, 61, 160, 60, 0, 62, 160, 60, 0, 63, 160, 60, 0, 64, 160, 60, 0, 65, 160, 60, 0, 66, 160, 60, 0, 67, 160, 60, 0, 68, 160, 60, 0, 69, 160, 60, 0, 70, 160, 60, 0, 71, 160, 60, 0, 72, 160, 60, 0, 73, 160, 60, 0, 74, 160, 60, 0, 75, 160, 60, 0, 76, 160, 60, 0, 77, 160, 60, 0, 78, 160, 60, 0, 79, 160, 60, 0, 80, 160, 60, 0, 81, 160, 60, 0, 82, 160, 60, 0, 83, 160, 60, 0, 84, 160, 60, 0, 85, 160, 60, 0, 86, 160, 60, 0, 87, 160, 60, 0, 88, 160, 60, 0, 89, 160, 60, 0, 90, 160, 60, 0, 91, 160, 60, 0, 92, 160, 60, 0, 93, 160, 60, 0, 94, 160, 60, 0, 95, 160, 60, 0, 96, 160, 60, 0, 97, 160, 60, 0, 98, 160, 60, 0, 99, 160, 60, 0, 100, 160, 60, 0, 101, 160, 60, 0, 102, 160, 60, 0, 103, 160, 60, 0, 104, 160, 60, 0, 105, 160, 60, 0, 106, 160, 60, 0, 107, 160, 60, 0, 108, 160, 60, 0, 109, 160, 60, 0, 110, 160, 60, 0, 111, 160, 60, 0, 112, 160, 60, 0, 113, 160, 60, 0, 114, 160, 60, 0, 115, 160, 60, 0, 116, 160, 60, 0, 117, 160, 60, 0, 118, 160, 60, 0, 119, 160, 60, 0, 120, 160, 60, 0, 121, 160, 60, 0, 122, 160, 60, 0, 123, 160, 60, 0, 124, 160, 60, 0, 125, 160, 60, 0, 126, 160, 60, 0, 127, 160, 60, 0, 128, 160, 60, 0, 129, 160, 60, 0, 130, 160, 60, 0, 131, 160, 60, 0, 132, 160, 60, 0, 133, 160, 60, 0, 134, 160, 60, 0, 135, 160, 60, 0, 136, 160, 60, 0, 137, 160, 60, 0, 138, 160, 60, 0, 139, 160, 60, 0, 140, 160, 60, 0, 141, 160, 60, 0, 142, 160, 60, 0, 143, 160, 60, 0, 144, 160, 60, 0, 145, 160, 60, 0, 146, 160, 60, 0, 147, 160, 60, 0, 148, 160, 60, 0, 149, 160, 60, 0, 150, 160, 60, 0, 151, 160, 60, 0, 152, 160, 60, 0, 153, 160, 60, 0, 154, 160, 60, 0, 155, 160, 60, 0, 156, 160, 60, 0, 157, 160, 60, 0, 158, 160, 60, 0, 159, 160, 60, 0, 160, 160, 60, 0, 161, 160, 60, 0, 162, 160, 60, 0, 163, 160, 60, 0, 164, 160, 60, 0, 165, 160, 60, 0, 166, 160, 60, 0, 167, 160, 60, 0, 168, 160, 60, 0, 169, 160, 60, 0, 170, 160, 60, 0, 171, 160, 60, 0, 172, 160, 60, 0, 173, 160, 60, 0, 174, 160, 60, 0, 175, 160, 60, 0, 176, 160, 60, 0, 177, 160, 60, 0, 178, 160, 60, 0, 179, 160, 60, 0, 180, 160, 60, 0, 181, 160, 60, 0, 182, 160, 60, 0, 183, 160, 60, 0, 184, 160, 60, 0, 185, 160, 60, 0, 186, 160, 60, 0, 187, 160, 60, 0, 188, 160, 60, 0, 189, 160, 60, 0, 190, 160, 60, 0, 191, 160, 60, 0, 192, 160, 60, 0, 193, 160, 60, 0, 194, 160, 60, 0, 195, 160, 60, 0, 196, 160, 60, 0, 197, 160, 60, 0, 198, 160, 60, 0, 199, 160, 60, 0, 200, 160, 60, 0, 201, 160, 60, 0, 202, 160, 60, 0, 203, 160, 60, 0, 204, 160, 60, 0, 205, 160, 60, 0, 206, 160, 60, 0, 207, 160, 60, 0, 208, 160, 60, 0, 209, 160, 60, 0, 210, 160, 60, 0, 211, 160, 60, 0, 212, 160, 60, 0, 213, 160, 60, 0, 214, 160, 60, 0, 215, 160, 60, 0, 216, 160, 60, 0, 217, 160, 60, 0, 218, 160, 60, 0, 219, 160, 60, 0, 220, 160, 60, 0, 221, 160, 60, 0, 222, 160, 60, 0, 223, 160, 60, 0, 224, 160, 60, 0, 225, 160, 60, 0, 226, 160, 60, 0, 227, 160, 60, 0, 228, 160, 60, 0, 229, 160, 60, 0, 230, 160, 60, 0, 231, 160, 60, 0, 232, 160, 60, 0, 233, 160, 60, 0, 234, 160, 60, 0, 235, 160, 60, 0, 236, 160, 60, 0, 237, 160, 60, 0, 238, 160, 60, 0, 239, 160, 60, 0, 240, 160, 60, 0, 241, 160, 60, 0, 242, 160, 60, 0, 243, 160, 60, 0, 244, 160, 60, 0, 245, 160, 60, 0, 246, 160, 60, 0, 247, 160, 60, 0, 248, 160, 60, 0, 249, 160, 60, 0, 250, 160, 60, 0, 251, 160, 60, 0, 252, 160, 60, 0, 253, 160, 60, 0, 254, 160, 60, 0, 255, 160, 60, 0, 0, 161, 60, 0, 1, 161, 60, 0, 2, 161, 60, 0, 3, 161, 60, 0, 4, 161, 60, 0, 5, 161, 60, 0, 6, 161, 60, 0, 7, 161, 60, 0, 8, 161, 60, 0, 9, 161, 60, 0, 10, 161, 60, 0, 11, 161, 60, 0, 12, 161, 60, 0, 13, 161, 60, 0, 14, 161, 60, 0, 15, 161, 60, 0, 16, 161, 60, 0, 17, 161, 60, 0, 18, 161, 60, 0, 19, 161, 60, 0, 20, 161, 60, 0, 21, 161, 60, 0, 22, 161, 60, 0, 23, 161, 60, 0, 24, 161, 60, 0, 25, 161, 60, 0, 26, 161, 60, 0, 27, 161, 60, 0, 28, 161, 60, 0, 29, 161, 60, 0, 30, 161, 60, 0, 31, 161, 60, 0, 32, 161, 60, 0, 33, 161, 60, 0, 34, 161, 60, 0, 35, 161, 60, 0, 36, 161, 60, 0, 37, 161, 60, 0, 38, 161, 60, 0, 39, 161, 60, 0, 40, 161, 60, 0, 41, 161, 60, 0, 42, 161, 60, 0, 43, 161, 60, 0, 44, 161, 60, 0, 45, 161, 60, 0, 46, 161, 60, 0, 47, 161, 60, 0, 48, 161, 60, 0, 49, 161, 60, 0, 50, 161, 60, 0, 51, 161, 60, 0, 52, 161, 60, 0, 53, 161, 60, 0, 54, 161, 60, 0, 55, 161, 60, 0, 56, 161, 60, 0, 57, 161, 60, 0, 58, 161, 60, 0, 59, 161, 60, 0, 60, 161, 60, 0, 61, 161, 60, 0, 62, 161, 60, 0, 63, 161, 60, 0, 64, 161, 60, 0, 65, 161, 60, 0, 66, 161, 60, 0, 67, 161, 60, 0, 68, 161, 60, 0, 69, 161, 60, 0, 70, 161, 60, 0, 71, 161, 60, 0, 72, 161, 60, 0, 73, 161, 60, 0, 74, 161, 60, 0, 75, 161, 60, 0, 76, 161, 60, 0, 77, 161, 60, 0, 78, 161, 60, 0, 79, 161, 60, 0, 80, 161, 60, 0, 81, 161, 60, 0, 82, 161, 60, 0, 83, 161, 60, 0, 84, 161, 60, 0, 85, 161, 60, 0, 86, 161, 60, 0, 87, 161, 60, 0, 88, 161, 60, 0, 89, 161, 60, 0, 90, 161, 60, 0, 91, 161, 60, 0, 92, 161, 60, 0, 93, 161, 60, 0, 94, 161, 60, 0, 95, 161, 60, 0, 96, 161, 60, 0, 97, 161, 60, 0, 98, 161, 60, 0, 99, 161, 60, 0, 100, 161, 60, 0, 101, 161, 60, 0, 102, 161, 60, 0, 103, 161, 60, 0, 104, 161, 60, 0, 105, 161, 60, 0, 106, 161, 60, 0, 107, 161, 60, 0, 108, 161, 60, 0, 109, 161, 60, 0, 110, 161, 60, 0, 111, 161, 60, 0, 112, 161, 60, 0, 113, 161, 60, 0, 114, 161, 60, 0, 115, 161, 60, 0, 116, 161, 60, 0, 117, 161, 60, 0, 118, 161, 60, 0, 119, 161, 60, 0, 120, 161, 60, 0, 121, 161, 60, 0, 122, 161, 60, 0, 123, 161, 60, 0, 124, 161, 60, 0, 125, 161, 60, 0, 126, 161, 60, 0, 127, 161, 60, 0, 128, 161, 60, 0, 129, 161, 60, 0, 130, 161, 60, 0, 131, 161, 60, 0, 132, 161, 60, 0, 133, 161, 60, 0, 134, 161, 60, 0, 135, 161, 60, 0, 136, 161, 60, 0, 137, 161, 60, 0, 138, 161, 60, 0, 139, 161, 60, 0, 140, 161, 60, 0, 141, 161, 60, 0, 142, 161, 60, 0, 143, 161, 60, 0, 144, 161, 60, 0, 145, 161, 60, 0, 146, 161, 60, 0, 147, 161, 60, 0, 148, 161, 60, 0, 149, 161, 60, 0, 150, 161, 60, 0, 151, 161, 60, 0, 152, 161, 60, 0, 153, 161, 60, 0, 154, 161, 60, 0, 155, 161, 60, 0, 156, 161, 60, 0, 157, 161, 60, 0, 158, 161, 60, 0, 159, 161, 60, 0, 160, 161, 60, 0, 161, 161, 60, 0, 162, 161, 60, 0, 163, 161, 60, 0, 164, 161, 60, 0, 165, 161, 60, 0, 166, 161, 60, 0, 167, 161, 60, 0, 168, 161, 60, 0, 169, 161, 60, 0, 170, 161, 60, 0, 171, 161, 60, 0, 172, 161, 60, 0, 173, 161, 60, 0, 174, 161, 60, 0, 175, 161, 60, 0, 176, 161, 60, 0, 177, 161, 60, 0, 178, 161, 60, 0, 179, 161, 60, 0, 180, 161, 60, 0, 181, 161, 60, 0, 182, 161, 60, 0, 183, 161, 60, 0, 184, 161, 60, 0, 185, 161, 60, 0, 186, 161, 60, 0, 187, 161, 60, 0, 188, 161, 60, 0, 189, 161, 60, 0, 190, 161, 60, 0, 191, 161, 60, 0, 192, 161, 60, 0, 193, 161, 60, 0, 194, 161, 60, 0, 195, 161, 60, 0, 196, 161, 60, 0, 197, 161, 60, 0, 198, 161, 60, 0, 199, 161, 60, 0, 200, 161, 60, 0, 201, 161, 60, 0, 202, 161, 60, 0, 203, 161, 60, 0, 204, 161, 60, 0, 205, 161, 60, 0, 206, 161, 60, 0, 207, 161, 60, 0, 208, 161, 60, 0, 209, 161, 60, 0, 210, 161, 60, 0, 211, 161, 60, 0, 212, 161, 60, 0, 213, 161, 60, 0, 214, 161, 60, 0, 215, 161, 60, 0, 216, 161, 60, 0, 217, 161, 60, 0, 218, 161, 60, 0, 219, 161, 60, 0, 220, 161, 60, 0, 221, 161, 60, 0, 222, 161, 60, 0, 223, 161, 60, 0, 224, 161, 60, 0, 225, 161, 60, 0, 226, 161, 60, 0, 227, 161, 60, 0, 228, 161, 60, 0, 229, 161, 60, 0, 230, 161, 60, 0, 231, 161, 60, 0, 232, 161, 60, 0, 233, 161, 60, 0, 234, 161, 60, 0, 235, 161, 60, 0, 236, 161, 60, 0, 237, 161, 60, 0, 238, 161, 60, 0, 239, 161, 60, 0, 240, 161, 60, 0, 241, 161, 60, 0, 242, 161, 60, 0, 243, 161, 60, 0, 244, 161, 60, 0, 245, 161, 60, 0, 246, 161, 60, 0, 247, 161, 60, 0, 248, 161, 60, 0, 249, 161, 60, 0, 250, 161, 60, 0, 251, 161, 60, 0, 252, 161, 60, 0, 253, 161, 60, 0, 254, 161, 60, 0, 255, 161, 60, 0, 0, 162, 60, 0, 1, 162, 60, 0, 2, 162, 60, 0, 3, 162, 60, 0, 4, 162, 60, 0, 5, 162, 60, 0, 6, 162, 60, 0, 7, 162, 60, 0, 8, 162, 60, 0, 9, 162, 60, 0, 10, 162, 60, 0, 11, 162, 60, 0, 12, 162, 60, 0, 13, 162, 60, 0, 14, 162, 60, 0, 15, 162, 60, 0, 16, 162, 60, 0, 17, 162, 60, 0, 18, 162, 60, 0, 19, 162, 60, 0, 20, 162, 60, 0, 21, 162, 60, 0, 22, 162, 60, 0, 23, 162, 60, 0, 24, 162, 60, 0, 25, 162, 60, 0, 26, 162, 60, 0, 27, 162, 60, 0, 28, 162, 60, 0, 29, 162, 60, 0, 30, 162, 60, 0, 31, 162, 60, 0, 32, 162, 60, 0, 33, 162, 60, 0, 34, 162, 60, 0, 35, 162, 60, 0, 36, 162, 60, 0, 37, 162, 60, 0, 38, 162, 60, 0, 39, 162, 60, 0, 40, 162, 60, 0, 41, 162, 60, 0, 42, 162, 60, 0, 43, 162, 60, 0, 44, 162, 60, 0, 45, 162, 60, 0, 46, 162, 60, 0, 47, 162, 60, 0, 48, 162, 60, 0, 49, 162, 60, 0, 50, 162, 60, 0, 51, 162, 60, 0, 52, 162, 60, 0, 53, 162, 60, 0, 54, 162, 60, 0, 55, 162, 60, 0, 56, 162, 60, 0, 57, 162, 60, 0, 58, 162, 60, 0, 59, 162, 60, 0, 60, 162, 60, 0, 61, 162, 60, 0, 62, 162, 60, 0, 63, 162, 60, 0, 64, 162, 60, 0, 65, 162, 60, 0, 66, 162, 60, 0, 67, 162, 60, 0, 68, 162, 60, 0, 69, 162, 60, 0, 70, 162, 60, 0, 71, 162, 60, 0, 72, 162, 60, 0, 73, 162, 60, 0, 74, 162, 60, 0, 75, 162, 60, 0, 76, 162, 60, 0, 77, 162, 60, 0, 78, 162, 60, 0, 79, 162, 60, 0, 80, 162, 60, 0, 81, 162, 60, 0, 82, 162, 60, 0, 83, 162, 60, 0, 84, 162, 60, 0, 85, 162, 60, 0, 86, 162, 60, 0, 87, 162, 60, 0, 88, 162, 60, 0, 89, 162, 60, 0, 90, 162, 60, 0, 91, 162, 60, 0, 92, 162, 60, 0, 93, 162, 60, 0, 94, 162, 60, 0, 95, 162, 60, 0, 96, 162, 60, 0, 97, 162, 60, 0, 98, 162, 60, 0, 99, 162, 60, 0, 100, 162, 60, 0, 101, 162, 60, 0, 102, 162, 60, 0, 103, 162, 60, 0, 104, 162, 60, 0, 105, 162, 60, 0, 106, 162, 60, 0, 107, 162, 60, 0, 108, 162, 60, 0, 109, 162, 60, 0, 110, 162, 60, 0, 111, 162, 60, 0, 112, 162, 60, 0, 113, 162, 60, 0, 114, 162, 60, 0, 115, 162, 60, 0, 116, 162, 60, 0, 117, 162, 60, 0, 118, 162, 60, 0, 119, 162, 60, 0, 120, 162, 60, 0, 121, 162, 60, 0, 122, 162, 60, 0, 123, 162, 60, 0, 124, 162, 60, 0, 125, 162, 60, 0, 126, 162, 60, 0, 127, 162, 60, 0, 128, 162, 60, 0, 129, 162, 60, 0, 130, 162, 60, 0, 131, 162, 60, 0, 132, 162, 60, 0, 133, 162, 60, 0, 134, 162, 60, 0, 135, 162, 60, 0, 136, 162, 60, 0, 137, 162, 60, 0, 138, 162, 60, 0, 139, 162, 60, 0, 140, 162, 60, 0, 141, 162, 60, 0, 142, 162, 60, 0, 143, 162, 60, 0, 144, 162, 60, 0, 145, 162, 60, 0, 146, 162, 60, 0, 147, 162, 60, 0, 148, 162, 60, 0, 149, 162, 60, 0, 150, 162, 60, 0, 151, 162, 60, 0, 152, 162, 60, 0, 153, 162, 60, 0, 154, 162, 60, 0, 155, 162, 60, 0, 156, 162, 60, 0, 157, 162, 60, 0, 158, 162, 60, 0, 159, 162, 60, 0, 160, 162, 60, 0, 161, 162, 60, 0, 162, 162, 60, 0, 163, 162, 60, 0, 164, 162, 60, 0, 165, 162, 60, 0, 166, 162, 60, 0, 167, 162, 60, 0, 168, 162, 60, 0, 169, 162, 60, 0, 170, 162, 60, 0, 171, 162, 60, 0, 172, 162, 60, 0, 173, 162, 60, 0, 174, 162, 60, 0, 175, 162, 60, 0, 176, 162, 60, 0, 177, 162, 60, 0, 178, 162, 60, 0, 179, 162, 60, 0, 180, 162, 60, 0, 181, 162, 60, 0, 182, 162, 60, 0, 183, 162, 60, 0, 184, 162, 60, 0, 185, 162, 60, 0, 186, 162, 60, 0, 187, 162, 60, 0, 188, 162, 60, 0, 189, 162, 60, 0, 190, 162, 60, 0, 191, 162, 60, 0, 192, 162, 60, 0, 193, 162, 60, 0, 194, 162, 60, 0, 195, 162, 60, 0, 196, 162, 60, 0, 197, 162, 60, 0, 198, 162, 60, 0, 199, 162, 60, 0, 200, 162, 60, 0, 201, 162, 60, 0, 202, 162, 60, 0, 203, 162, 60, 0, 204, 162, 60, 0, 205, 162, 60, 0, 206, 162, 60, 0, 207, 162, 60, 0, 208, 162, 60, 0, 209, 162, 60, 0, 210, 162, 60, 0, 211, 162, 60, 0, 212, 162, 60, 0, 213, 162, 60, 0, 214, 162, 60, 0, 215, 162, 60, 0, 216, 162, 60, 0, 217, 162, 60, 0, 218, 162, 60, 0, 219, 162, 60, 0, 220, 162, 60, 0, 221, 162, 60, 0, 222, 162, 60, 0, 223, 162, 60, 0, 224, 162, 60, 0, 225, 162, 60, 0, 226, 162, 60, 0, 227, 162, 60, 0, 228, 162, 60, 0, 229, 162, 60, 0, 230, 162, 60, 0, 231, 162, 60, 0, 232, 162, 60, 0, 233, 162, 60, 0, 234, 162, 60, 0, 235, 162, 60, 0, 236, 162, 60, 0, 237, 162, 60, 0, 238, 162, 60, 0, 239, 162, 60, 0, 240, 162, 60, 0, 241, 162, 60, 0, 242, 162, 60, 0, 243, 162, 60, 0, 244, 162, 60, 0, 245, 162, 60, 0, 246, 162, 60, 0, 247, 162, 60, 0, 248, 162, 60, 0, 249, 162, 60, 0, 250, 162, 60, 0, 251, 162, 60, 0, 252, 162, 60, 0, 253, 162, 60, 0, 254, 162, 60, 0, 255, 162, 60, 0, 0, 163, 60, 0, 1, 163, 60, 0, 2, 163, 60, 0, 3, 163, 60, 0, 4, 163, 60, 0, 5, 163, 60, 0, 6, 163, 60, 0, 7, 163, 60, 0, 8, 163, 60, 0, 9, 163, 60, 0, 10, 163, 60, 0, 11, 163, 60, 0, 12, 163, 60, 0, 13, 163, 60, 0, 14, 163, 60, 0, 15, 163, 60, 0, 16, 163, 60, 0, 17, 163, 60, 0, 18, 163, 60, 0, 19, 163, 60, 0, 20, 163, 60, 0, 21, 163, 60, 0, 22, 163, 60, 0, 23, 163, 60, 0, 24, 163, 60, 0, 25, 163, 60, 0, 26, 163, 60, 0, 27, 163, 60, 0, 28, 163, 60, 0, 29, 163, 60, 0, 30, 163, 60, 0, 31, 163, 60, 0, 32, 163, 60, 0, 33, 163, 60, 0, 34, 163, 60, 0, 35, 163, 60, 0, 36, 163, 60, 0, 37, 163, 60, 0, 38, 163, 60, 0, 39, 163, 60, 0, 40, 163, 60, 0, 41, 163, 60, 0, 42, 163, 60, 0, 43, 163, 60, 0, 44, 163, 60, 0, 45, 163, 60, 0, 46, 163, 60, 0, 47, 163, 60, 0, 48, 163, 60, 0, 49, 163, 60, 0, 50, 163, 60, 0, 51, 163, 60, 0, 52, 163, 60, 0, 53, 163, 60, 0, 54, 163, 60, 0, 55, 163, 60, 0, 56, 163, 60, 0, 57, 163, 60, 0, 58, 163, 60, 0, 59, 163, 60, 0, 60, 163, 60, 0, 61, 163, 60, 0, 62, 163, 60, 0, 63, 163, 60, 0, 64, 163, 60, 0, 65, 163, 60, 0, 66, 163, 60, 0, 67, 163, 60, 0, 68, 163, 60, 0, 69, 163, 60, 0, 70, 163, 60, 0, 71, 163, 60, 0, 72, 163, 60, 0, 73, 163, 60, 0, 74, 163, 60, 0, 75, 163, 60, 0, 76, 163, 60, 0, 77, 163, 60, 0, 78, 163, 60, 0, 79, 163, 60, 0, 80, 163, 60, 0, 81, 163, 60, 0, 82, 163, 60, 0, 83, 163, 60, 0, 84, 163, 60, 0, 85, 163, 60, 0, 86, 163, 60, 0, 87, 163, 60, 0, 88, 163, 60, 0, 89, 163, 60, 0, 90, 163, 60, 0, 91, 163, 60, 0, 92, 163, 60, 0, 93, 163, 60, 0, 94, 163, 60, 0, 95, 163, 60, 0, 96, 163, 60, 0, 97, 163, 60, 0, 98, 163, 60, 0, 99, 163, 60, 0, 100, 163, 60, 0, 101, 163, 60, 0, 102, 163, 60, 0, 103, 163, 60, 0, 104, 163, 60, 0, 105, 163, 60, 0, 106, 163, 60, 0, 107, 163, 60, 0, 108, 163, 60, 0, 109, 163, 60, 0, 110, 163, 60, 0, 111, 163, 60, 0, 112, 163, 60, 0, 113, 163, 60, 0, 114, 163, 60, 0, 115, 163, 60, 0, 116, 163, 60, 0, 117, 163, 60, 0, 118, 163, 60, 0, 119, 163, 60, 0, 120, 163, 60, 0, 121, 163, 60, 0, 122, 163, 60, 0, 123, 163, 60, 0, 124, 163, 60, 0, 125, 163, 60, 0, 126, 163, 60, 0, 127, 163, 60, 0, 128, 163, 60, 0, 129, 163, 60, 0, 130, 163, 60, 0, 131, 163, 60, 0, 132, 163, 60, 0, 133, 163, 60, 0, 134, 163, 60, 0, 135, 163, 60, 0, 136, 163, 60, 0, 137, 163, 60, 0, 138, 163, 60, 0, 139, 163, 60, 0, 140, 163, 60, 0, 141, 163, 60, 0, 142, 163, 60, 0, 143, 163, 60, 0, 144, 163, 60, 0, 145, 163, 60, 0, 146, 163, 60, 0, 147, 163, 60, 0, 148, 163, 60, 0, 149, 163, 60, 0, 150, 163, 60, 0, 151, 163, 60, 0, 152, 163, 60, 0, 153, 163, 60, 0, 154, 163, 60, 0, 155, 163, 60, 0, 156, 163, 60, 0, 157, 163, 60, 0, 158, 163, 60, 0, 159, 163, 60, 0, 160, 163, 60, 0, 161, 163, 60, 0, 162, 163, 60, 0, 163, 163, 60, 0, 164, 163, 60, 0, 165, 163, 60, 0, 166, 163, 60, 0, 167, 163, 60, 0, 168, 163, 60, 0, 169, 163, 60, 0, 170, 163, 60, 0, 171, 163, 60, 0, 172, 163, 60, 0, 173, 163, 60, 0, 174, 163, 60, 0, 175, 163, 60, 0, 176, 163, 60, 0, 177, 163, 60, 0, 178, 163, 60, 0, 179, 163, 60, 0, 180, 163, 60, 0, 181, 163, 60, 0, 182, 163, 60, 0, 183, 163, 60, 0, 184, 163, 60, 0, 185, 163, 60, 0, 186, 163, 60, 0, 187, 163, 60, 0, 188, 163, 60, 0, 189, 163, 60, 0, 190, 163, 60, 0, 191, 163, 60, 0, 192, 163, 60, 0, 193, 163, 60, 0, 194, 163, 60, 0, 195, 163, 60, 0, 196, 163, 60, 0, 197, 163, 60, 0, 198, 163, 60, 0, 199, 163, 60, 0, 200, 163, 60, 0, 201, 163, 60, 0, 202, 163, 60, 0, 203, 163, 60, 0, 204, 163, 60, 0, 205, 163, 60, 0, 206, 163, 60, 0, 207, 163, 60, 0, 208, 163, 60, 0, 209, 163, 60, 0, 210, 163, 60, 0, 211, 163, 60, 0, 212, 163, 60, 0, 213, 163, 60, 0, 214, 163, 60, 0, 215, 163, 60, 0, 216, 163, 60, 0, 217, 163, 60, 0, 218, 163, 60, 0, 219, 163, 60, 0, 220, 163, 60, 0, 221, 163, 60, 0, 222, 163, 60, 0, 223, 163, 60, 0, 224, 163, 60, 0, 225, 163, 60, 0, 226, 163, 60, 0, 227, 163, 60, 0, 228, 163, 60, 0, 229, 163, 60, 0, 230, 163, 60, 0, 231, 163, 60, 0, 232, 163, 60, 0, 233, 163, 60, 0, 234, 163, 60, 0, 235, 163, 60, 0, 236, 163, 60, 0, 237, 163, 60, 0, 238, 163, 60, 0, 239, 163, 60, 0, 240, 163, 60, 0, 241, 163, 60, 0, 242, 163, 60, 0, 243, 163, 60, 0, 244, 163, 60, 0, 245, 163, 60, 0, 246, 163, 60, 0, 247, 163, 60, 0, 248, 163, 60, 0, 249, 163, 60, 0, 250, 163, 60, 0, 251, 163, 60, 0, 252, 163, 60, 0, 253, 163, 60, 0, 254, 163, 60, 0, 255, 163, 60, 0, 0, 164, 60, 0, 1, 164, 60, 0, 2, 164, 60, 0, 3, 164, 60, 0, 4, 164, 60, 0, 5, 164, 60, 0, 6, 164, 60, 0, 7, 164, 60, 0, 8, 164, 60, 0, 9, 164, 60, 0, 10, 164, 60, 0, 11, 164, 60, 0, 12, 164, 60, 0, 13, 164, 60, 0, 14, 164, 60, 0, 15, 164, 60, 0, 16, 164, 60, 0, 17, 164, 60, 0, 18, 164, 60, 0, 19, 164, 60, 0, 20, 164, 60, 0, 21, 164, 60, 0, 22, 164, 60, 0, 23, 164, 60, 0, 24, 164, 60, 0, 25, 164, 60, 0, 26, 164, 60, 0, 27, 164, 60, 0, 28, 164, 60, 0, 29, 164, 60, 0, 30, 164, 60, 0, 31, 164, 60, 0, 32, 164, 60, 0, 33, 164, 60, 0, 34, 164, 60, 0, 35, 164, 60, 0, 36, 164, 60, 0, 37, 164, 60, 0, 38, 164, 60, 0, 39, 164, 60, 0, 40, 164, 60, 0, 41, 164, 60, 0, 42, 164, 60, 0, 43, 164, 60, 0, 44, 164, 60, 0, 45, 164, 60, 0, 46, 164, 60, 0, 47, 164, 60, 0, 48, 164, 60, 0, 49, 164, 60, 0, 50, 164, 60, 0, 51, 164, 60, 0, 52, 164, 60, 0, 53, 164, 60, 0, 54, 164, 60, 0, 55, 164, 60, 0, 56, 164, 60, 0, 57, 164, 60, 0, 58, 164, 60, 0, 59, 164, 60, 0, 60, 164, 60, 0, 61, 164, 60, 0, 62, 164, 60, 0, 63, 164, 60, 0, 64, 164, 60, 0, 65, 164, 60, 0, 66, 164, 60, 0, 67, 164, 60, 0, 68, 164, 60, 0, 69, 164, 60, 0, 70, 164, 60, 0, 71, 164, 60, 0, 72, 164, 60, 0, 73, 164, 60, 0, 74, 164, 60, 0, 75, 164, 60, 0, 76, 164, 60, 0, 77, 164, 60, 0, 78, 164, 60, 0, 79, 164, 60, 0, 80, 164, 60, 0, 81, 164, 60, 0, 82, 164, 60, 0, 83, 164, 60, 0, 84, 164, 60, 0, 85, 164, 60, 0, 86, 164, 60, 0, 87, 164, 60, 0, 88, 164, 60, 0, 89, 164, 60, 0, 90, 164, 60, 0, 91, 164, 60, 0, 92, 164, 60, 0, 93, 164, 60, 0, 94, 164, 60, 0, 95, 164, 60, 0, 96, 164, 60, 0, 97, 164, 60, 0, 98, 164, 60, 0, 99, 164, 60, 0, 100, 164, 60, 0, 101, 164, 60, 0, 102, 164, 60, 0, 103, 164, 60, 0, 104, 164, 60, 0, 105, 164, 60, 0, 106, 164, 60, 0, 107, 164, 60, 0, 108, 164, 60, 0, 109, 164, 60, 0, 110, 164, 60, 0, 111, 164, 60, 0, 112, 164, 60, 0, 113, 164, 60, 0, 114, 164, 60, 0, 115, 164, 60, 0, 116, 164, 60, 0, 117, 164, 60, 0, 118, 164, 60, 0, 119, 164, 60, 0, 120, 164, 60, 0, 121, 164, 60, 0, 122, 164, 60, 0, 123, 164, 60, 0, 124, 164, 60, 0, 125, 164, 60, 0, 126, 164, 60, 0, 127, 164, 60, 0, 128, 164, 60, 0, 129, 164, 60, 0, 130, 164, 60, 0, 131, 164, 60, 0, 132, 164, 60, 0, 133, 164, 60, 0, 134, 164, 60, 0, 135, 164, 60, 0, 136, 164, 60, 0, 137, 164, 60, 0, 138, 164, 60, 0, 139, 164, 60, 0, 140, 164, 60, 0, 248, 164, 60, 0, 249, 164, 60, 0, 250, 164, 60, 0, 251, 164, 60, 0, 253, 164, 60, 0, 252, 164, 60, 0, 208, 164, 60, 0, 209, 164, 60, 0, 210, 164, 60, 0, 211, 164, 60, 0, 212, 164, 60, 0, 213, 164, 60, 0, 214, 164, 60, 0, 215, 164, 60, 0, 216, 164, 60, 0, 217, 164, 60, 0, 218, 164, 60, 0, 219, 164, 60, 0, 220, 164, 60, 0, 221, 164, 60, 0, 222, 164, 60, 0, 223, 164, 60, 0, 224, 164, 60, 0, 225, 164, 60, 0, 226, 164, 60, 0, 227, 164, 60, 0, 228, 164, 60, 0, 229, 164, 60, 0, 230, 164, 60, 0, 231, 164, 60, 0, 232, 164, 60, 0, 233, 164, 60, 0, 235, 164, 60, 0, 237, 164, 60, 0, 234, 164, 60, 0, 236, 164, 60, 0, 7, 216, 176, 223, 60, 0, 238, 164, 60, 0, 239, 164, 60, 0, 240, 164, 60, 0, 241, 164, 60, 0, 242, 164, 60, 0, 243, 164, 60, 0, 244, 164, 60, 0, 245, 164, 60, 0, 246, 164, 60, 0, 247, 164, 60, 0, 27, 216, 0, 223, 60, 0, 27, 216, 1, 223, 60, 0, 27, 216, 2, 223, 60, 0, 27, 216, 3, 223, 60, 0, 27, 216, 69, 223, 60, 0, 27, 216, 4, 223, 60, 0, 60, 0, 60, 0, 27, 216, 6, 223, 60, 0, 27, 216, 5, 223, 60, 0, 27, 216, 7, 223, 60, 0, 27, 216, 8, 223, 60, 0, 27, 216, 9, 223, 60, 0, 27, 216, 10, 223, 60, 0, 27, 216, 11, 223, 60, 0, 27, 216, 12, 223, 60, 0, 27, 216, 13, 223, 60, 0, 27, 216, 72, 223, 60, 0, 27, 216, 14, 223, 60, 0, 27, 216, 15, 223, 60, 0, 27, 216, 74, 223, 60, 0, 27, 216, 16, 223, 60, 0, 60, 0, 60, 0, 27, 216, 19, 223, 60, 0, 27, 216, 17, 223, 60, 0, 27, 216, 18, 223, 60, 0, 27, 216, 20, 223, 60, 0, 27, 216, 21, 223, 60, 0, 27, 216, 22, 223, 60, 0, 27, 216, 23, 223, 60, 0, 27, 216, 24, 223, 60, 0, 27, 216, 25, 223, 60, 0, 27, 216, 26, 223, 60, 0, 27, 216, 27, 223, 60, 0, 27, 216, 28, 223, 60, 0, 27, 216, 29, 223, 60, 0, 27, 216, 30, 223, 60, 0, 27, 216, 31, 223, 60, 0, 27, 216, 32, 223, 60, 0, 27, 216, 33, 223, 60, 0, 27, 216, 34, 223, 60, 0, 27, 216, 35, 223, 60, 0, 60, 0, 60, 0, 27, 216, 37, 223, 60, 0, 27, 216, 36, 223, 60, 0, 27, 216, 38, 223, 60, 0, 27, 216, 39, 223, 60, 0, 27, 216, 40, 223, 60, 0, 27, 216, 41, 223, 60, 0, 27, 216, 42, 223, 60, 0, 27, 216, 43, 223, 60, 0, 27, 216, 44, 223, 60, 0, 27, 216, 45, 223, 60, 0, 27, 216, 46, 223, 60, 0, 27, 216, 47, 223, 60, 0, 27, 216, 48, 223, 60, 0, 27, 216, 49, 223, 60, 0, 27, 216, 50, 223, 60, 0, 27, 216, 71, 223, 60, 0, 27, 216, 51, 223, 60, 0, 27, 216, 70, 223, 60, 0, 27, 216, 52, 223, 60, 0, 27, 216, 53, 223, 60, 0, 27, 216, 54, 223, 60, 0, 27, 216, 55, 223, 60, 0, 27, 216, 56, 223, 60, 0, 27, 216, 57, 223, 60, 0, 27, 216, 73, 223, 60, 0, 27, 216, 58, 223, 60, 0, 27, 216, 59, 223, 60, 0, 27, 216, 60, 223, 60, 0, 27, 216, 61, 223, 60, 0, 60, 0, 60, 0, 27, 216, 63, 223, 60, 0, 27, 216, 62, 223, 60, 0, 27, 216, 64, 223, 60, 0, 27, 216, 65, 223, 60, 0, 27, 216, 66, 223, 60, 0, 27, 216, 67, 223, 60, 0, 27, 216, 68, 223, 60, 0, 27, 216, 80, 223, 60, 0, 27, 216, 81, 223, 60, 0, 27, 216, 82, 223, 60, 0, 27, 216, 83, 223, 60, 0, 27, 216, 79, 223, 60, 0, 27, 216, 84, 223, 60, 0, 27, 216, 85, 223, 60, 0, 27, 216, 86, 223, 60, 0, 27, 216, 87, 223, 60, 0, 27, 216, 88, 223, 60, 0, 27, 216, 129, 223, 60, 0, 27, 216, 89, 223, 60, 0, 27, 216, 90, 223, 60, 0, 27, 216, 91, 223, 60, 0, 27, 216, 92, 223, 60, 0, 27, 216, 131, 223, 60, 0, 27, 216, 93, 223, 60, 0, 27, 216, 94, 223, 60, 0, 27, 216, 95, 223, 60, 0, 27, 216, 96, 223, 60, 0, 27, 216, 132, 223, 60, 0, 27, 216, 97, 223, 60, 0, 27, 216, 98, 223, 60, 0, 27, 216, 99, 223, 60, 0, 27, 216, 100, 223, 60, 0, 27, 216, 101, 223, 60, 0, 27, 216, 134, 223, 60, 0, 27, 216, 102, 223, 60, 0, 27, 216, 103, 223, 60, 0, 27, 216, 104, 223, 60, 0, 27, 216, 105, 223, 60, 0, 27, 216, 106, 223, 60, 0, 27, 216, 107, 223, 60, 0, 27, 216, 108, 223, 60, 0, 27, 216, 109, 223, 60, 0, 27, 216, 127, 223, 60, 0, 27, 216, 135, 223, 60, 0, 27, 216, 110, 223, 60, 0, 27, 216, 111, 223, 60, 0, 27, 216, 112, 223, 60, 0, 27, 216, 113, 223, 60, 0, 27, 216, 114, 223, 60, 0, 27, 216, 128, 223, 60, 0, 27, 216, 115, 223, 60, 0, 27, 216, 116, 223, 60, 0, 27, 216, 133, 223, 60, 0, 27, 216, 130, 223, 60, 0, 27, 216, 117, 223, 60, 0, 27, 216, 118, 223, 60, 0, 27, 216, 119, 223, 60, 0, 27, 216, 120, 223, 60, 0, 27, 216, 121, 223, 60, 0, 27, 216, 122, 223, 60, 0, 27, 216, 123, 223, 60, 0, 27, 216, 124, 223, 60, 0, 27, 216, 125, 223, 60, 0, 27, 216, 126, 223, 60, 0, 27, 216, 143, 223, 60, 0, 27, 216, 144, 223, 60, 0, 27, 216, 145, 223, 60, 0, 27, 216, 146, 223, 60, 0, 27, 216, 147, 223, 60, 0, 27, 216, 148, 223, 60, 0, 27, 216, 149, 223, 60, 0, 27, 216, 150, 223, 60, 0, 27, 216, 151, 223, 60, 0, 27, 216, 152, 223, 60, 0, 27, 216, 153, 223, 60, 0, 27, 216, 154, 223, 60, 0, 27, 216, 155, 223, 60, 0, 27, 216, 156, 223, 60, 0, 27, 216, 157, 223, 60, 0, 27, 216, 158, 223, 60, 0, 27, 216, 159, 223, 60, 0, 6, 216, 255, 220, 60, 0, 6, 216, 192, 220, 60, 0, 60, 0, 60, 0, 6, 216, 160, 220, 60, 0, 6, 216, 193, 220, 60, 0, 60, 0, 60, 0, 6, 216, 161, 220, 60, 0, 6, 216, 194, 220, 60, 0, 60, 0, 60, 0, 6, 216, 162, 220, 60, 0, 6, 216, 195, 220, 60, 0, 60, 0, 60, 0, 6, 216, 163, 220, 60, 0, 6, 216, 196, 220, 60, 0, 60, 0, 60, 0, 6, 216, 164, 220, 60, 0, 6, 216, 197, 220, 60, 0, 60, 0, 60, 0, 6, 216, 165, 220, 60, 0, 6, 216, 198, 220, 60, 0, 60, 0, 60, 0, 6, 216, 166, 220, 60, 0, 6, 216, 199, 220, 60, 0, 60, 0, 60, 0, 6, 216, 167, 220, 60, 0, 6, 216, 200, 220, 60, 0, 60, 0, 60, 0, 6, 216, 168, 220, 60, 0, 6, 216, 201, 220, 60, 0, 60, 0, 60, 0, 6, 216, 169, 220, 60, 0, 6, 216, 202, 220, 60, 0, 60, 0, 60, 0, 6, 216, 170, 220, 60, 0, 6, 216, 203, 220, 60, 0, 60, 0, 60, 0, 6, 216, 171, 220, 60, 0, 6, 216, 204, 220, 60, 0, 60, 0, 60, 0, 6, 216, 172, 220, 60, 0, 6, 216, 205, 220, 60, 0, 60, 0, 60, 0, 6, 216, 173, 220, 60, 0, 6, 216, 206, 220, 60, 0, 60, 0, 60, 0, 6, 216, 174, 220, 60, 0, 6, 216, 207, 220, 60, 0, 60, 0, 60, 0, 6, 216, 175, 220, 60, 0, 6, 216, 208, 220, 60, 0, 60, 0, 60, 0, 6, 216, 176, 220, 60, 0, 6, 216, 209, 220, 60, 0, 60, 0, 60, 0, 6, 216, 177, 220, 60, 0, 6, 216, 210, 220, 60, 0, 60, 0, 60, 0, 6, 216, 178, 220, 60, 0, 6, 216, 211, 220, 60, 0, 60, 0, 60, 0, 6, 216, 179, 220, 60, 0, 6, 216, 212, 220, 60, 0, 60, 0, 60, 0, 6, 216, 180, 220, 60, 0, 6, 216, 213, 220, 60, 0, 60, 0, 60, 0, 6, 216, 181, 220, 60, 0, 6, 216, 214, 220, 60, 0, 60, 0, 60, 0, 6, 216, 182, 220, 60, 0, 6, 216, 215, 220, 60, 0, 60, 0, 60, 0, 6, 216, 183, 220, 60, 0, 6, 216, 216, 220, 60, 0, 60, 0, 60, 0, 6, 216, 184, 220, 60, 0, 6, 216, 217, 220, 60, 0, 60, 0, 60, 0, 6, 216, 185, 220, 60, 0, 6, 216, 218, 220, 60, 0, 60, 0, 60, 0, 6, 216, 186, 220, 60, 0, 6, 216, 219, 220, 60, 0, 60, 0, 60, 0, 6, 216, 187, 220, 60, 0, 6, 216, 220, 220, 60, 0, 60, 0, 60, 0, 6, 216, 188, 220, 60, 0, 6, 216, 221, 220, 60, 0, 60, 0, 60, 0, 6, 216, 189, 220, 60, 0, 6, 216, 222, 220, 60, 0, 60, 0, 60, 0, 6, 216, 190, 220, 60, 0, 6, 216, 223, 220, 60, 0, 60, 0, 60, 0, 6, 216, 191, 220, 60, 0, 6, 216, 213, 222, 60, 0, 6, 216, 214, 222, 60, 0, 6, 216, 215, 222, 60, 0, 6, 216, 216, 222, 60, 0, 6, 216, 217, 222, 60, 0, 6, 216, 218, 222, 60, 0, 6, 216, 219, 222, 60, 0, 6, 216, 192, 222, 60, 0, 6, 216, 193, 222, 60, 0, 6, 216, 194, 222, 60, 0, 6, 216, 195, 222, 60, 0, 6, 216, 196, 222, 60, 0, 6, 216, 197, 222, 60, 0, 6, 216, 198, 222, 60, 0, 6, 216, 199, 222, 60, 0, 6, 216, 200, 222, 60, 0, 6, 216, 201, 222, 60, 0, 6, 216, 202, 222, 60, 0, 6, 216, 203, 222, 60, 0, 6, 216, 204, 222, 60, 0, 6, 216, 205, 222, 60, 0, 6, 216, 206, 222, 60, 0, 6, 216, 207, 222, 60, 0, 6, 216, 208, 222, 60, 0, 6, 216, 209, 222, 60, 0, 6, 216, 210, 222, 60, 0, 6, 216, 211, 222, 60, 0, 6, 216, 212, 222, 60, 0, 6, 216, 220, 222, 60, 0, 6, 216, 221, 222, 60, 0, 6, 216, 223, 222, 60, 0, 6, 216, 224, 222, 60, 0, 6, 216, 225, 222, 60, 0, 6, 216, 226, 222, 60, 0, 6, 216, 227, 222, 60, 0, 6, 216, 222, 222, 60, 0, 6, 216, 228, 222, 60, 0, 6, 216, 239, 222, 60, 0, 6, 216, 242, 222, 60, 0, 6, 216, 241, 222, 60, 0, 6, 216, 230, 222, 60, 0, 6, 216, 233, 222, 60, 0, 6, 216, 229, 222, 60, 0, 6, 216, 232, 222, 60, 0, 6, 216, 244, 222, 60, 0, 6, 216, 247, 222, 60, 0, 6, 216, 243, 222, 60, 0, 6, 216, 246, 222, 60, 0, 6, 216, 236, 222, 60, 0, 6, 216, 238, 222, 60, 0, 6, 216, 235, 222, 60, 0, 6, 216, 237, 222, 60, 0, 6, 216, 245, 222, 60, 0, 6, 216, 248, 222, 60, 0, 6, 216, 231, 222, 60, 0, 6, 216, 234, 222, 60, 0, 6, 216, 240, 222, 60, 0, 26, 216, 0, 223, 60, 0, 26, 216, 1, 223, 60, 0, 26, 216, 2, 223, 60, 0, 26, 216, 3, 223, 60, 0, 26, 216, 4, 223, 60, 0, 26, 216, 5, 223, 60, 0, 26, 216, 6, 223, 60, 0, 26, 216, 7, 223, 60, 0, 26, 216, 8, 223, 60, 0, 26, 216, 9, 223, 60, 0, 26, 216, 10, 223, 60, 0, 26, 216, 11, 223, 60, 0, 26, 216, 12, 223, 60, 0, 26, 216, 13, 223, 60, 0, 26, 216, 14, 223, 60, 0, 26, 216, 15, 223, 60, 0, 26, 216, 16, 223, 60, 0, 26, 216, 17, 223, 60, 0, 26, 216, 18, 223, 60, 0, 26, 216, 19, 223, 60, 0, 26, 216, 20, 223, 60, 0, 26, 216, 21, 223, 60, 0, 26, 216, 22, 223, 60, 0, 26, 216, 23, 223, 60, 0, 26, 216, 24, 223, 60, 0, 26, 216, 25, 223, 60, 0, 26, 216, 26, 223, 60, 0, 26, 216, 27, 223, 60, 0, 26, 216, 28, 223, 60, 0, 26, 216, 29, 223, 60, 0, 26, 216, 30, 223, 60, 0, 26, 216, 31, 223, 60, 0, 26, 216, 32, 223, 60, 0, 26, 216, 33, 223, 60, 0, 26, 216, 34, 223, 60, 0, 26, 216, 35, 223, 60, 0, 26, 216, 36, 223, 60, 0, 26, 216, 37, 223, 60, 0, 26, 216, 38, 223, 60, 0, 26, 216, 39, 223, 60, 0, 26, 216, 40, 223, 60, 0, 26, 216, 41, 223, 60, 0, 26, 216, 42, 223, 60, 0, 26, 216, 43, 223, 60, 0, 26, 216, 44, 223, 60, 0, 26, 216, 45, 223, 60, 0, 26, 216, 46, 223, 60, 0, 26, 216, 47, 223, 60, 0, 26, 216, 64, 223, 60, 0, 26, 216, 65, 223, 60, 0, 26, 216, 99, 223, 60, 0, 26, 216, 100, 223, 60, 0, 26, 216, 101, 223, 60, 0, 26, 216, 102, 223, 60, 0, 26, 216, 103, 223, 60, 0, 26, 216, 104, 223, 60, 0, 26, 216, 105, 223, 60, 0, 26, 216, 106, 223, 60, 0, 26, 216, 107, 223, 60, 0, 26, 216, 108, 223, 60, 0, 26, 216, 109, 223, 60, 0, 26, 216, 110, 223, 60, 0, 26, 216, 111, 223, 60, 0, 26, 216, 112, 223, 60, 0, 26, 216, 113, 223, 60, 0, 26, 216, 114, 223, 60, 0, 26, 216, 115, 223, 60, 0, 26, 216, 116, 223, 60, 0, 26, 216, 117, 223, 60, 0, 26, 216, 118, 223, 60, 0, 26, 216, 119, 223, 60, 0, 26, 216, 125, 223, 60, 0, 26, 216, 126, 223, 60, 0, 26, 216, 127, 223, 60, 0, 26, 216, 128, 223, 60, 0, 26, 216, 129, 223, 60, 0, 26, 216, 130, 223, 60, 0, 26, 216, 131, 223, 60, 0, 26, 216, 132, 223, 60, 0, 26, 216, 133, 223, 60, 0, 26, 216, 134, 223, 60, 0, 26, 216, 135, 223, 60, 0, 26, 216, 136, 223, 60, 0, 26, 216, 137, 223, 60, 0, 26, 216, 138, 223, 60, 0, 26, 216, 139, 223, 60, 0, 26, 216, 140, 223, 60, 0, 26, 216, 141, 223, 60, 0, 26, 216, 142, 223, 60, 0, 26, 216, 143, 223, 60, 0, 56, 216, 0, 221, 60, 0, 56, 216, 1, 221, 60, 0, 56, 216, 2, 221, 60, 0, 56, 216, 3, 221, 60, 0, 56, 216, 4, 221, 60, 0, 56, 216, 5, 221, 60, 0, 56, 216, 6, 221, 60, 0, 56, 216, 7, 221, 60, 0, 56, 216, 8, 221, 60, 0, 56, 216, 9, 221, 60, 0, 56, 216, 10, 221, 60, 0, 56, 216, 11, 221, 60, 0, 56, 216, 12, 221, 60, 0, 56, 216, 13, 221, 60, 0, 56, 216, 14, 221, 60, 0, 56, 216, 15, 221, 60, 0, 56, 216, 16, 221, 60, 0, 56, 216, 17, 221, 60, 0, 56, 216, 18, 221, 60, 0, 56, 216, 19, 221, 60, 0, 56, 216, 20, 221, 60, 0, 56, 216, 21, 221, 60, 0, 56, 216, 22, 221, 60, 0, 56, 216, 23, 221, 60, 0, 56, 216, 24, 221, 60, 0, 56, 216, 25, 221, 60, 0, 56, 216, 26, 221, 60, 0, 56, 216, 27, 221, 60, 0, 56, 216, 28, 221, 60, 0, 56, 216, 29, 221, 60, 0, 56, 216, 30, 221, 60, 0, 56, 216, 31, 221, 60, 0, 56, 216, 32, 221, 60, 0, 56, 216, 33, 221, 60, 0, 56, 216, 34, 221, 60, 0, 56, 216, 35, 221, 60, 0, 56, 216, 36, 221, 60, 0, 56, 216, 37, 221, 60, 0, 56, 216, 38, 221, 60, 0, 56, 216, 39, 221, 60, 0, 56, 216, 40, 221, 60, 0, 56, 216, 41, 221, 60, 0, 56, 216, 42, 221, 60, 0, 56, 216, 43, 221, 60, 0, 56, 216, 44, 221, 60, 0, 56, 216, 55, 221, 60, 0, 56, 216, 56, 221, 60, 0, 56, 216, 57, 221, 60, 0, 56, 216, 58, 221, 60, 0, 56, 216, 59, 221, 60, 0, 56, 216, 78, 221, 60, 0, 56, 216, 192, 222, 60, 0, 56, 216, 193, 222, 60, 0, 56, 216, 194, 222, 60, 0, 56, 216, 195, 222, 60, 0, 56, 216, 196, 222, 60, 0, 56, 216, 197, 222, 60, 0, 56, 216, 198, 222, 60, 0, 56, 216, 199, 222, 60, 0, 56, 216, 200, 222, 60, 0, 56, 216, 201, 222, 60, 0, 56, 216, 202, 222, 60, 0, 56, 216, 203, 222, 60, 0, 56, 216, 204, 222, 60, 0, 56, 216, 205, 222, 60, 0, 56, 216, 206, 222, 60, 0, 56, 216, 207, 222, 60, 0, 56, 216, 208, 222, 60, 0, 56, 216, 209, 222, 60, 0, 56, 216, 210, 222, 60, 0, 56, 216, 211, 222, 60, 0, 56, 216, 212, 222, 60, 0, 56, 216, 213, 222, 60, 0, 56, 216, 214, 222, 60, 0, 56, 216, 215, 222, 60, 0, 56, 216, 216, 222, 60, 0, 56, 216, 217, 222, 60, 0, 56, 216, 218, 222, 60, 0, 56, 216, 219, 222, 60, 0, 56, 216, 220, 222, 60, 0, 56, 216, 221, 222, 60, 0, 56, 216, 222, 222, 60, 0, 56, 216, 223, 222, 60, 0, 56, 216, 224, 222, 60, 0, 56, 216, 225, 222, 60, 0, 56, 216, 226, 222, 60, 0, 56, 216, 227, 222, 60, 0, 56, 216, 228, 222, 60, 0, 56, 216, 229, 222, 60, 0, 56, 216, 230, 222, 60, 0, 56, 216, 231, 222, 60, 0, 56, 216, 232, 222, 60, 0, 56, 216, 233, 222, 60, 0, 56, 216, 234, 222, 60, 0, 56, 216, 235, 222, 60, 0, 0, 216, 128, 222, 60, 0, 0, 216, 129, 222, 60, 0, 0, 216, 130, 222, 60, 0, 0, 216, 131, 222, 60, 0, 0, 216, 132, 222, 60, 0, 0, 216, 133, 222, 60, 0, 0, 216, 134, 222, 60, 0, 0, 216, 135, 222, 60, 0, 0, 216, 136, 222, 60, 0, 0, 216, 137, 222, 60, 0, 0, 216, 138, 222, 60, 0, 0, 216, 139, 222, 60, 0, 0, 216, 140, 222, 60, 0, 0, 216, 141, 222, 60, 0, 0, 216, 142, 222, 60, 0, 0, 216, 143, 222, 60, 0, 0, 216, 144, 222, 60, 0, 0, 216, 145, 222, 60, 0, 0, 216, 146, 222, 60, 0, 0, 216, 147, 222, 60, 0, 0, 216, 148, 222, 60, 0, 0, 216, 149, 222, 60, 0, 0, 216, 150, 222, 60, 0, 0, 216, 151, 222, 60, 0, 0, 216, 152, 222, 60, 0, 0, 216, 153, 222, 60, 0, 0, 216, 154, 222, 60, 0, 0, 216, 155, 222, 60, 0, 0, 216, 156, 222, 60, 0, 0, 216, 160, 222, 60, 0, 0, 216, 161, 222, 60, 0, 0, 216, 162, 222, 60, 0, 0, 216, 163, 222, 60, 0, 0, 216, 164, 222, 60, 0, 0, 216, 165, 222, 60, 0, 0, 216, 166, 222, 60, 0, 0, 216, 167, 222, 60, 0, 0, 216, 168, 222, 60, 0, 0, 216, 169, 222, 60, 0, 0, 216, 170, 222, 60, 0, 0, 216, 171, 222, 60, 0, 0, 216, 172, 222, 60, 0, 0, 216, 173, 222, 60, 0, 0, 216, 174, 222, 60, 0, 0, 216, 175, 222, 60, 0, 0, 216, 176, 222, 60, 0, 0, 216, 177, 222, 60, 0, 0, 216, 178, 222, 60, 0, 0, 216, 179, 222, 60, 0, 0, 216, 180, 222, 60, 0, 0, 216, 181, 222, 60, 0, 0, 216, 182, 222, 60, 0, 0, 216, 183, 222, 60, 0, 0, 216, 184, 222, 60, 0, 0, 216, 185, 222, 60, 0, 0, 216, 186, 222, 60, 0, 0, 216, 187, 222, 60, 0, 0, 216, 188, 222, 60, 0, 0, 216, 189, 222, 60, 0, 0, 216, 190, 222, 60, 0, 0, 216, 191, 222, 60, 0, 0, 216, 192, 222, 60, 0, 0, 216, 193, 222, 60, 0, 0, 216, 194, 222, 60, 0, 0, 216, 195, 222, 60, 0, 0, 216, 196, 222, 60, 0, 0, 216, 197, 222, 60, 0, 0, 216, 198, 222, 60, 0, 0, 216, 199, 222, 60, 0, 0, 216, 200, 222, 60, 0, 0, 216, 201, 222, 60, 0, 0, 216, 202, 222, 60, 0, 0, 216, 203, 222, 60, 0, 0, 216, 204, 222, 60, 0, 0, 216, 205, 222, 60, 0, 0, 216, 206, 222, 60, 0, 0, 216, 207, 222, 60, 0, 0, 216, 208, 222, 60, 0, 14, 32, 2, 216, 32, 221, 14, 32, 60, 0, 14, 32, 2, 216, 33, 221, 14, 32, 60, 0, 14, 32, 2, 216, 34, 221, 14, 32, 60, 0, 14, 32, 2, 216, 35, 221, 14, 32, 60, 0, 14, 32, 2, 216, 36, 221, 14, 32, 60, 0, 14, 32, 2, 216, 37, 221, 14, 32, 60, 0, 14, 32, 2, 216, 38, 221, 14, 32, 60, 0, 14, 32, 2, 216, 39, 221, 14, 32, 60, 0, 14, 32, 2, 216, 40, 221, 14, 32, 60, 0, 14, 32, 2, 216, 41, 221, 14, 32, 60, 0, 14, 32, 2, 216, 42, 221, 14, 32, 60, 0, 14, 32, 2, 216, 43, 221, 14, 32, 60, 0, 14, 32, 2, 216, 44, 221, 14, 32, 60, 0, 14, 32, 2, 216, 45, 221, 14, 32, 60, 0, 14, 32, 2, 216, 46, 221, 14, 32, 60, 0, 14, 32, 2, 216, 47, 221, 14, 32, 60, 0, 14, 32, 2, 216, 48, 221, 14, 32, 60, 0, 14, 32, 2, 216, 49, 221, 14, 32, 60, 0, 14, 32, 2, 216, 50, 221, 14, 32, 60, 0, 14, 32, 2, 216, 51, 221, 14, 32, 60, 0, 14, 32, 2, 216, 52, 221, 14, 32, 60, 0, 14, 32, 2, 216, 53, 221, 14, 32, 60, 0, 14, 32, 2, 216, 54, 221, 14, 32, 60, 0, 14, 32, 2, 216, 55, 221, 14, 32, 60, 0, 14, 32, 2, 216, 56, 221, 14, 32, 60, 0, 14, 32, 2, 216, 57, 221, 14, 32, 60, 0, 0, 216, 0, 223, 60, 0, 0, 216, 1, 223, 60, 0, 0, 216, 2, 223, 60, 0, 0, 216, 3, 223, 60, 0, 0, 216, 4, 223, 60, 0, 0, 216, 5, 223, 60, 0, 0, 216, 6, 223, 60, 0, 0, 216, 7, 223, 60, 0, 0, 216, 8, 223, 60, 0, 0, 216, 9, 223, 60, 0, 0, 216, 10, 223, 60, 0, 0, 216, 11, 223, 60, 0, 0, 216, 12, 223, 60, 0, 0, 216, 13, 223, 60, 0, 0, 216, 14, 223, 60, 0, 0, 216, 31, 223, 60, 0, 0, 216, 15, 223, 60, 0, 0, 216, 16, 223, 60, 0, 0, 216, 17, 223, 60, 0, 0, 216, 18, 223, 60, 0, 0, 216, 19, 223, 60, 0, 0, 216, 20, 223, 60, 0, 0, 216, 21, 223, 60, 0, 0, 216, 22, 223, 60, 0, 0, 216, 23, 223, 60, 0, 0, 216, 24, 223, 60, 0, 0, 216, 25, 223, 60, 0, 0, 216, 26, 223, 60, 0, 0, 216, 27, 223, 60, 0, 0, 216, 28, 223, 60, 0, 0, 216, 29, 223, 60, 0, 0, 216, 30, 223, 60, 0, 0, 216, 45, 223, 60, 0, 0, 216, 46, 223, 60, 0, 0, 216, 47, 223, 60, 0, 0, 216, 48, 223, 60, 0, 0, 216, 49, 223, 60, 0, 0, 216, 50, 223, 60, 0, 0, 216, 51, 223, 60, 0, 0, 216, 52, 223, 60, 0, 0, 216, 53, 223, 60, 0, 0, 216, 54, 223, 60, 0, 0, 216, 55, 223, 60, 0, 0, 216, 56, 223, 60, 0, 0, 216, 57, 223, 60, 0, 0, 216, 58, 223, 60, 0, 0, 216, 59, 223, 60, 0, 0, 216, 60, 223, 60, 0, 0, 216, 61, 223, 60, 0, 0, 216, 62, 223, 60, 0, 0, 216, 63, 223, 60, 0, 0, 216, 64, 223, 60, 0, 0, 216, 65, 223, 60, 0, 0, 216, 66, 223, 60, 0, 0, 216, 67, 223, 60, 0, 0, 216, 68, 223, 60, 0, 0, 216, 69, 223, 60, 0, 0, 216, 70, 223, 60, 0, 0, 216, 71, 223, 60, 0, 0, 216, 72, 223, 60, 0, 0, 216, 73, 223, 60, 0, 0, 216, 74, 223, 60, 0, 1, 216, 40, 220, 60, 0, 60, 0, 60, 0, 1, 216, 0, 220, 60, 0, 1, 216, 41, 220, 60, 0, 60, 0, 60, 0, 1, 216, 1, 220, 60, 0, 1, 216, 42, 220, 60, 0, 60, 0, 60, 0, 1, 216, 2, 220, 60, 0, 1, 216, 43, 220, 60, 0, 60, 0, 60, 0, 1, 216, 3, 220, 60, 0, 1, 216, 44, 220, 60, 0, 60, 0, 60, 0, 1, 216, 4, 220, 60, 0, 1, 216, 45, 220, 60, 0, 60, 0, 60, 0, 1, 216, 5, 220, 60, 0, 1, 216, 46, 220, 60, 0, 60, 0, 60, 0, 1, 216, 6, 220, 60, 0, 1, 216, 47, 220, 60, 0, 60, 0, 60, 0, 1, 216, 7, 220, 60, 0, 1, 216, 48, 220, 60, 0, 60, 0, 60, 0, 1, 216, 8, 220, 60, 0, 1, 216, 49, 220, 60, 0, 60, 0, 60, 0, 1, 216, 9, 220, 60, 0, 1, 216, 50, 220, 60, 0, 60, 0, 60, 0, 1, 216, 10, 220, 60, 0, 1, 216, 51, 220, 60, 0, 60, 0, 60, 0, 1, 216, 11, 220, 60, 0, 1, 216, 52, 220, 60, 0, 60, 0, 60, 0, 1, 216, 12, 220, 60, 0, 1, 216, 53, 220, 60, 0, 60, 0, 60, 0, 1, 216, 13, 220, 60, 0, 1, 216, 54, 220, 60, 0, 60, 0, 60, 0, 1, 216, 14, 220, 60, 0, 1, 216, 55, 220, 60, 0, 60, 0, 60, 0, 1, 216, 15, 220, 60, 0, 1, 216, 56, 220, 60, 0, 60, 0, 60, 0, 1, 216, 16, 220, 60, 0, 1, 216, 57, 220, 60, 0, 60, 0, 60, 0, 1, 216, 17, 220, 60, 0, 1, 216, 58, 220, 60, 0, 60, 0, 60, 0, 1, 216, 18, 220, 60, 0, 1, 216, 59, 220, 60, 0, 60, 0, 60, 0, 1, 216, 19, 220, 60, 0, 1, 216, 60, 220, 60, 0, 60, 0, 60, 0, 1, 216, 20, 220, 60, 0, 1, 216, 61, 220, 60, 0, 60, 0, 60, 0, 1, 216, 21, 220, 60, 0, 1, 216, 62, 220, 60, 0, 60, 0, 60, 0, 1, 216, 22, 220, 60, 0, 1, 216, 63, 220, 60, 0, 60, 0, 60, 0, 1, 216, 23, 220, 60, 0, 1, 216, 64, 220, 60, 0, 60, 0, 60, 0, 1, 216, 24, 220, 60, 0, 1, 216, 65, 220, 60, 0, 60, 0, 60, 0, 1, 216, 25, 220, 60, 0, 1, 216, 66, 220, 60, 0, 60, 0, 60, 0, 1, 216, 26, 220, 60, 0, 1, 216, 67, 220, 60, 0, 60, 0, 60, 0, 1, 216, 27, 220, 60, 0, 1, 216, 68, 220, 60, 0, 60, 0, 60, 0, 1, 216, 28, 220, 60, 0, 1, 216, 69, 220, 60, 0, 60, 0, 60, 0, 1, 216, 29, 220, 60, 0, 1, 216, 70, 220, 60, 0, 60, 0, 60, 0, 1, 216, 30, 220, 60, 0, 1, 216, 71, 220, 60, 0, 60, 0, 60, 0, 1, 216, 31, 220, 60, 0, 1, 216, 72, 220, 60, 0, 60, 0, 60, 0, 1, 216, 32, 220, 60, 0, 1, 216, 73, 220, 60, 0, 60, 0, 60, 0, 1, 216, 33, 220, 60, 0, 1, 216, 74, 220, 60, 0, 60, 0, 60, 0, 1, 216, 34, 220, 60, 0, 1, 216, 75, 220, 60, 0, 60, 0, 60, 0, 1, 216, 35, 220, 60, 0, 1, 216, 76, 220, 60, 0, 60, 0, 60, 0, 1, 216, 36, 220, 60, 0, 1, 216, 77, 220, 60, 0, 60, 0, 60, 0, 1, 216, 37, 220, 60, 0, 1, 216, 78, 220, 60, 0, 60, 0, 60, 0, 1, 216, 38, 220, 60, 0, 1, 216, 79, 220, 60, 0, 60, 0, 60, 0, 1, 216, 39, 220, 60, 0, 1, 216, 80, 220, 60, 0, 1, 216, 81, 220, 60, 0, 1, 216, 82, 220, 60, 0, 1, 216, 83, 220, 60, 0, 1, 216, 84, 220, 60, 0, 1, 216, 85, 220, 60, 0, 1, 216, 86, 220, 60, 0, 1, 216, 87, 220, 60, 0, 1, 216, 88, 220, 60, 0, 1, 216, 89, 220, 60, 0, 1, 216, 90, 220, 60, 0, 1, 216, 91, 220, 60, 0, 1, 216, 92, 220, 60, 0, 1, 216, 93, 220, 60, 0, 1, 216, 94, 220, 60, 0, 1, 216, 95, 220, 60, 0, 1, 216, 96, 220, 60, 0, 1, 216, 97, 220, 60, 0, 1, 216, 98, 220, 60, 0, 1, 216, 99, 220, 60, 0, 1, 216, 100, 220, 60, 0, 1, 216, 101, 220, 60, 0, 1, 216, 102, 220, 60, 0, 1, 216, 103, 220, 60, 0, 1, 216, 104, 220, 60, 0, 1, 216, 105, 220, 60, 0, 1, 216, 106, 220, 60, 0, 1, 216, 107, 220, 60, 0, 1, 216, 108, 220, 60, 0, 1, 216, 109, 220, 60, 0, 1, 216, 110, 220, 60, 0, 1, 216, 111, 220, 60, 0, 1, 216, 112, 220, 60, 0, 1, 216, 113, 220, 60, 0, 1, 216, 114, 220, 60, 0, 1, 216, 115, 220, 60, 0, 1, 216, 116, 220, 60, 0, 1, 216, 117, 220, 60, 0, 1, 216, 118, 220, 60, 0, 1, 216, 119, 220, 60, 0, 1, 216, 120, 220, 60, 0, 1, 216, 121, 220, 60, 0, 1, 216, 122, 220, 60, 0, 1, 216, 123, 220, 60, 0, 1, 216, 124, 220, 60, 0, 1, 216, 125, 220, 60, 0, 1, 216, 126, 220, 60, 0, 1, 216, 127, 220, 60, 0, 47, 216, 0, 220, 60, 0, 47, 216, 1, 220, 60, 0, 47, 216, 2, 220, 60, 0, 47, 216, 3, 220, 60, 0, 47, 216, 4, 220, 60, 0, 47, 216, 5, 220, 60, 0, 47, 216, 6, 220, 60, 0, 47, 216, 7, 220, 60, 0, 47, 216, 8, 220, 60, 0, 47, 216, 9, 220, 60, 0, 47, 216, 10, 220, 60, 0, 47, 216, 11, 220, 60, 0, 47, 216, 12, 220, 60, 0, 47, 216, 13, 220, 60, 0, 47, 216, 14, 220, 60, 0, 47, 216, 15, 220, 60, 0, 47, 216, 16, 220, 60, 0, 47, 216, 17, 220, 60, 0, 47, 216, 18, 220, 60, 0, 47, 216, 19, 220, 60, 0, 47, 216, 20, 220, 60, 0, 47, 216, 21, 220, 60, 0, 47, 216, 22, 220, 60, 0, 47, 216, 23, 220, 60, 0, 47, 216, 24, 220, 60, 0, 47, 216, 25, 220, 60, 0, 47, 216, 26, 220, 60, 0, 47, 216, 27, 220, 60, 0, 47, 216, 28, 220, 60, 0, 47, 216, 29, 220, 60, 0, 47, 216, 30, 220, 60, 0, 47, 216, 31, 220, 60, 0, 47, 216, 32, 220, 60, 0, 47, 216, 33, 220, 60, 0, 47, 216, 34, 220, 60, 0, 47, 216, 35, 220, 60, 0, 47, 216, 36, 220, 60, 0, 47, 216, 37, 220, 60, 0, 47, 216, 38, 220, 60, 0, 47, 216, 39, 220, 60, 0, 47, 216, 40, 220, 60, 0, 47, 216, 41, 220, 60, 0, 47, 216, 42, 220, 60, 0, 47, 216, 43, 220, 60, 0, 47, 216, 44, 220, 60, 0, 47, 216, 45, 220, 60, 0, 47, 216, 46, 220, 60, 0, 47, 216, 47, 220, 60, 0, 47, 216, 48, 220, 60, 0, 47, 216, 49, 220, 60, 0, 47, 216, 50, 220, 60, 0, 47, 216, 51, 220, 60, 0, 47, 216, 52, 220, 60, 0, 47, 216, 53, 220, 60, 0, 47, 216, 54, 220, 60, 0, 47, 216, 55, 220, 60, 0, 47, 216, 56, 220, 60, 0, 47, 216, 57, 220, 60, 0, 47, 216, 58, 220, 60, 0, 47, 216, 59, 220, 60, 0, 47, 216, 60, 220, 60, 0, 47, 216, 61, 220, 60, 0, 47, 216, 62, 220, 60, 0, 47, 216, 63, 220, 60, 0, 47, 216, 64, 220, 60, 0, 47, 216, 65, 220, 60, 0, 47, 216, 66, 220, 60, 0, 47, 216, 67, 220, 60, 0, 47, 216, 68, 220, 60, 0, 47, 216, 69, 220, 60, 0, 47, 216, 70, 220, 60, 0, 47, 216, 71, 220, 60, 0, 47, 216, 72, 220, 60, 0, 47, 216, 73, 220, 60, 0, 47, 216, 74, 220, 60, 0, 47, 216, 75, 220, 60, 0, 47, 216, 76, 220, 60, 0, 47, 216, 77, 220, 60, 0, 47, 216, 78, 220, 60, 0, 47, 216, 79, 220, 60, 0, 47, 216, 80, 220, 60, 0, 47, 216, 81, 220, 60, 0, 47, 216, 82, 220, 60, 0, 47, 216, 83, 220, 60, 0, 47, 216, 84, 220, 60, 0, 47, 216, 85, 220, 60, 0, 47, 216, 86, 220, 60, 0, 47, 216, 87, 220, 60, 0, 47, 216, 88, 220, 60, 0, 47, 216, 89, 220, 60, 0, 47, 216, 90, 220, 60, 0, 47, 216, 91, 220, 60, 0, 47, 216, 92, 220, 60, 0, 47, 216, 93, 220, 60, 0, 47, 216, 94, 220, 60, 0, 47, 216, 95, 220, 60, 0, 47, 216, 96, 220, 60, 0, 47, 216, 97, 220, 60, 0, 47, 216, 98, 220, 60, 0, 47, 216, 99, 220, 60, 0, 47, 216, 100, 220, 60, 0, 47, 216, 101, 220, 60, 0, 47, 216, 102, 220, 60, 0, 47, 216, 103, 220, 60, 0, 47, 216, 104, 220, 60, 0, 47, 216, 105, 220, 60, 0, 47, 216, 106, 220, 60, 0, 47, 216, 112, 220, 60, 0, 47, 216, 113, 220, 60, 0, 47, 216, 114, 220, 60, 0, 47, 216, 115, 220, 60, 0, 47, 216, 116, 220, 60, 0, 47, 216, 117, 220, 60, 0, 47, 216, 118, 220, 60, 0, 47, 216, 119, 220, 60, 0, 47, 216, 120, 220, 60, 0, 47, 216, 121, 220, 60, 0, 47, 216, 122, 220, 60, 0, 47, 216, 123, 220, 60, 0, 47, 216, 124, 220, 60, 0, 47, 216, 128, 220, 60, 0, 47, 216, 129, 220, 60, 0, 47, 216, 130, 220, 60, 0, 47, 216, 131, 220, 60, 0, 47, 216, 132, 220, 60, 0, 47, 216, 133, 220, 60, 0, 47, 216, 134, 220, 60, 0, 47, 216, 135, 220, 60, 0, 47, 216, 136, 220, 60, 0, 47, 216, 144, 220, 60, 0, 47, 216, 145, 220, 60, 0, 47, 216, 146, 220, 60, 0, 47, 216, 147, 220, 60, 0, 47, 216, 148, 220, 60, 0, 47, 216, 149, 220, 60, 0, 47, 216, 150, 220, 60, 0, 47, 216, 151, 220, 60, 0, 47, 216, 152, 220, 60, 0, 47, 216, 153, 220, 60, 0, 1, 216, 128, 220, 60, 0, 1, 216, 129, 220, 60, 0, 1, 216, 130, 220, 60, 0, 1, 216, 131, 220, 60, 0, 1, 216, 132, 220, 60, 0, 1, 216, 133, 220, 60, 0, 1, 216, 134, 220, 60, 0, 1, 216, 135, 220, 60, 0, 1, 216, 136, 220, 60, 0, 1, 216, 137, 220, 60, 0, 1, 216, 138, 220, 60, 0, 1, 216, 139, 220, 60, 0, 1, 216, 140, 220, 60, 0, 1, 216, 141, 220, 60, 0, 1, 216, 142, 220, 60, 0, 1, 216, 143, 220, 60, 0, 1, 216, 144, 220, 60, 0, 1, 216, 145, 220, 60, 0, 1, 216, 146, 220, 60, 0, 1, 216, 147, 220, 60, 0, 1, 216, 148, 220, 60, 0, 1, 216, 149, 220, 60, 0, 1, 216, 150, 220, 60, 0, 1, 216, 151, 220, 60, 0, 1, 216, 152, 220, 60, 0, 1, 216, 153, 220, 60, 0, 1, 216, 154, 220, 60, 0, 1, 216, 155, 220, 60, 0, 1, 216, 156, 220, 60, 0, 1, 216, 157, 220, 60, 0, 1, 216, 0, 221, 60, 0, 1, 216, 1, 221, 60, 0, 1, 216, 2, 221, 60, 0, 1, 216, 3, 221, 60, 0, 1, 216, 4, 221, 60, 0, 1, 216, 5, 221, 60, 0, 1, 216, 6, 221, 60, 0, 1, 216, 7, 221, 60, 0, 1, 216, 8, 221, 60, 0, 1, 216, 9, 221, 60, 0, 1, 216, 10, 221, 60, 0, 1, 216, 11, 221, 60, 0, 1, 216, 12, 221, 60, 0, 1, 216, 13, 221, 60, 0, 1, 216, 14, 221, 60, 0, 1, 216, 15, 221, 60, 0, 1, 216, 16, 221, 60, 0, 1, 216, 17, 221, 60, 0, 1, 216, 18, 221, 60, 0, 1, 216, 19, 221, 60, 0, 1, 216, 20, 221, 60, 0, 1, 216, 21, 221, 60, 0, 1, 216, 22, 221, 60, 0, 1, 216, 23, 221, 60, 0, 1, 216, 24, 221, 60, 0, 1, 216, 25, 221, 60, 0, 1, 216, 26, 221, 60, 0, 1, 216, 27, 221, 60, 0, 1, 216, 28, 221, 60, 0, 1, 216, 29, 221, 60, 0, 1, 216, 30, 221, 60, 0, 1, 216, 31, 221, 60, 0, 1, 216, 32, 221, 60, 0, 1, 216, 33, 221, 60, 0, 1, 216, 34, 221, 60, 0, 1, 216, 35, 221, 60, 0, 1, 216, 36, 221, 60, 0, 1, 216, 37, 221, 60, 0, 1, 216, 38, 221, 60, 0, 1, 216, 39, 221, 60, 0, 1, 216, 48, 221, 60, 0, 1, 216, 49, 221, 60, 0, 1, 216, 50, 221, 60, 0, 1, 216, 51, 221, 60, 0, 1, 216, 52, 221, 60, 0, 1, 216, 53, 221, 60, 0, 1, 216, 54, 221, 60, 0, 1, 216, 55, 221, 60, 0, 1, 216, 56, 221, 60, 0, 1, 216, 57, 221, 60, 0, 1, 216, 58, 221, 60, 0, 1, 216, 59, 221, 60, 0, 1, 216, 60, 221, 60, 0, 1, 216, 61, 221, 60, 0, 1, 216, 62, 221, 60, 0, 1, 216, 63, 221, 60, 0, 1, 216, 64, 221, 60, 0, 1, 216, 65, 221, 60, 0, 1, 216, 66, 221, 60, 0, 1, 216, 67, 221, 60, 0, 1, 216, 68, 221, 60, 0, 1, 216, 69, 221, 60, 0, 1, 216, 70, 221, 60, 0, 1, 216, 71, 221, 60, 0, 1, 216, 72, 221, 60, 0, 1, 216, 73, 221, 60, 0, 1, 216, 74, 221, 60, 0, 1, 216, 75, 221, 60, 0, 1, 216, 76, 221, 60, 0, 1, 216, 77, 221, 60, 0, 1, 216, 78, 221, 60, 0, 1, 216, 79, 221, 60, 0, 1, 216, 80, 221, 60, 0, 1, 216, 81, 221, 60, 0, 1, 216, 82, 221, 60, 0, 1, 216, 83, 221, 60, 0, 1, 216, 84, 221, 60, 0, 1, 216, 85, 221, 60, 0, 1, 216, 86, 221, 60, 0, 1, 216, 87, 221, 60, 0, 1, 216, 88, 221, 60, 0, 1, 216, 89, 221, 60, 0, 1, 216, 90, 221, 60, 0, 1, 216, 91, 221, 60, 0, 1, 216, 92, 221, 60, 0, 1, 216, 93, 221, 60, 0, 1, 216, 94, 221, 60, 0, 1, 216, 95, 221, 60, 0, 1, 216, 96, 221, 60, 0, 1, 216, 97, 221, 60, 0, 1, 216, 98, 221, 60, 0, 1, 216, 99, 221, 60, 0, 4, 216, 208, 220, 60, 0, 4, 216, 209, 220, 60, 0, 4, 216, 210, 220, 60, 0, 4, 216, 211, 220, 60, 0, 4, 216, 212, 220, 60, 0, 4, 216, 213, 220, 60, 0, 4, 216, 214, 220, 60, 0, 4, 216, 215, 220, 60, 0, 4, 216, 216, 220, 60, 0, 4, 216, 217, 220, 60, 0, 4, 216, 218, 220, 60, 0, 4, 216, 219, 220, 60, 0, 4, 216, 220, 220, 60, 0, 4, 216, 221, 220, 60, 0, 4, 216, 222, 220, 60, 0, 4, 216, 223, 220, 60, 0, 4, 216, 224, 220, 60, 0, 4, 216, 225, 220, 60, 0, 4, 216, 226, 220, 60, 0, 4, 216, 227, 220, 60, 0, 4, 216, 228, 220, 60, 0, 4, 216, 229, 220, 60, 0, 4, 216, 230, 220, 60, 0, 4, 216, 231, 220, 60, 0, 4, 216, 232, 220, 60, 0, 26, 216, 64, 222, 60, 0, 26, 216, 65, 222, 60, 0, 26, 216, 66, 222, 60, 0, 26, 216, 67, 222, 60, 0, 26, 216, 68, 222, 60, 0, 26, 216, 69, 222, 60, 0, 26, 216, 70, 222, 60, 0, 26, 216, 71, 222, 60, 0, 26, 216, 72, 222, 60, 0, 26, 216, 73, 222, 60, 0, 26, 216, 74, 222, 60, 0, 26, 216, 75, 222, 60, 0, 26, 216, 76, 222, 60, 0, 26, 216, 77, 222, 60, 0, 26, 216, 78, 222, 60, 0, 26, 216, 79, 222, 60, 0, 26, 216, 80, 222, 60, 0, 26, 216, 81, 222, 60, 0, 26, 216, 82, 222, 60, 0, 26, 216, 83, 222, 60, 0, 26, 216, 84, 222, 60, 0, 26, 216, 85, 222, 60, 0, 26, 216, 86, 222, 60, 0, 26, 216, 87, 222, 60, 0, 26, 216, 88, 222, 60, 0, 26, 216, 89, 222, 60, 0, 26, 216, 90, 222, 60, 0, 26, 216, 91, 222, 60, 0, 26, 216, 92, 222, 60, 0, 26, 216, 93, 222, 60, 0, 26, 216, 94, 222, 60, 0, 0, 216, 0, 220, 60, 0, 0, 216, 1, 220, 60, 0, 0, 216, 2, 220, 60, 0, 0, 216, 3, 220, 60, 0, 0, 216, 4, 220, 60, 0, 0, 216, 5, 220, 60, 0, 0, 216, 6, 220, 60, 0, 0, 216, 7, 220, 60, 0, 0, 216, 8, 220, 60, 0, 0, 216, 9, 220, 60, 0, 0, 216, 10, 220, 60, 0, 0, 216, 11, 220, 60, 0, 0, 216, 13, 220, 60, 0, 0, 216, 14, 220, 60, 0, 0, 216, 15, 220, 60, 0, 0, 216, 16, 220, 60, 0, 0, 216, 17, 220, 60, 0, 0, 216, 18, 220, 60, 0, 0, 216, 19, 220, 60, 0, 0, 216, 20, 220, 60, 0, 0, 216, 21, 220, 60, 0, 0, 216, 22, 220, 60, 0, 0, 216, 23, 220, 60, 0, 0, 216, 24, 220, 60, 0, 0, 216, 25, 220, 60, 0, 0, 216, 26, 220, 60, 0, 0, 216, 27, 220, 60, 0, 0, 216, 28, 220, 60, 0, 0, 216, 29, 220, 60, 0, 0, 216, 30, 220, 60, 0, 0, 216, 31, 220, 60, 0, 0, 216, 32, 220, 60, 0, 0, 216, 33, 220, 60, 0, 0, 216, 34, 220, 60, 0, 0, 216, 35, 220, 60, 0, 0, 216, 36, 220, 60, 0, 0, 216, 37, 220, 60, 0, 0, 216, 38, 220, 60, 0, 0, 216, 40, 220, 60, 0, 0, 216, 41, 220, 60, 0, 0, 216, 42, 220, 60, 0, 0, 216, 43, 220, 60, 0, 0, 216, 44, 220, 60, 0, 0, 216, 45, 220, 60, 0, 0, 216, 46, 220, 60, 0, 0, 216, 47, 220, 60, 0, 0, 216, 48, 220, 60, 0, 0, 216, 49, 220, 60, 0, 0, 216, 50, 220, 60, 0, 0, 216, 51, 220, 60, 0, 0, 216, 52, 220, 60, 0, 0, 216, 53, 220, 60, 0, 0, 216, 54, 220, 60, 0, 0, 216, 55, 220, 60, 0, 0, 216, 56, 220, 60, 0, 0, 216, 57, 220, 60, 0, 0, 216, 58, 220, 60, 0, 0, 216, 60, 220, 60, 0, 0, 216, 61, 220, 60, 0, 0, 216, 63, 220, 60, 0, 0, 216, 64, 220, 60, 0, 0, 216, 65, 220, 60, 0, 0, 216, 66, 220, 60, 0, 0, 216, 67, 220, 60, 0, 0, 216, 68, 220, 60, 0, 0, 216, 69, 220, 60, 0, 0, 216, 70, 220, 60, 0, 0, 216, 71, 220, 60, 0, 0, 216, 72, 220, 60, 0, 0, 216, 73, 220, 60, 0, 0, 216, 74, 220, 60, 0, 0, 216, 75, 220, 60, 0, 0, 216, 76, 220, 60, 0, 0, 216, 77, 220, 60, 0, 0, 216, 80, 220, 60, 0, 0, 216, 81, 220, 60, 0, 0, 216, 82, 220, 60, 0, 0, 216, 83, 220, 60, 0, 0, 216, 84, 220, 60, 0, 0, 216, 85, 220, 60, 0, 0, 216, 86, 220, 60, 0, 0, 216, 87, 220, 60, 0, 0, 216, 88, 220, 60, 0, 0, 216, 89, 220, 60, 0, 0, 216, 90, 220, 60, 0, 0, 216, 91, 220, 60, 0, 0, 216, 92, 220, 60, 0, 0, 216, 93, 220, 60, 0, 0, 216, 128, 220, 60, 0, 0, 216, 129, 220, 60, 0, 0, 216, 130, 220, 60, 0, 0, 216, 131, 220, 60, 0, 0, 216, 132, 220, 60, 0, 0, 216, 133, 220, 60, 0, 0, 216, 134, 220, 60, 0, 0, 216, 135, 220, 60, 0, 0, 216, 136, 220, 60, 0, 0, 216, 137, 220, 60, 0, 0, 216, 138, 220, 60, 0, 0, 216, 139, 220, 60, 0, 0, 216, 140, 220, 60, 0, 0, 216, 141, 220, 60, 0, 0, 216, 142, 220, 60, 0, 0, 216, 143, 220, 60, 0, 0, 216, 144, 220, 60, 0, 0, 216, 145, 220, 60, 0, 0, 216, 146, 220, 60, 0, 0, 216, 147, 220, 60, 0, 0, 216, 148, 220, 60, 0, 0, 216, 149, 220, 60, 0, 0, 216, 150, 220, 60, 0, 0, 216, 151, 220, 60, 0, 0, 216, 152, 220, 60, 0, 0, 216, 153, 220, 60, 0, 0, 216, 154, 220, 60, 0, 0, 216, 155, 220, 60, 0, 0, 216, 156, 220, 60, 0, 0, 216, 157, 220, 60, 0, 0, 216, 158, 220, 60, 0, 0, 216, 159, 220, 60, 0, 0, 216, 160, 220, 60, 0, 0, 216, 161, 220, 60, 0, 0, 216, 162, 220, 60, 0, 0, 216, 163, 220, 60, 0, 0, 216, 164, 220, 60, 0, 0, 216, 165, 220, 60, 0, 0, 216, 166, 220, 60, 0, 0, 216, 167, 220, 60, 0, 0, 216, 168, 220, 60, 0, 0, 216, 169, 220, 60, 0, 0, 216, 170, 220, 60, 0, 0, 216, 171, 220, 60, 0, 0, 216, 172, 220, 60, 0, 0, 216, 173, 220, 60, 0, 0, 216, 174, 220, 60, 0, 0, 216, 175, 220, 60, 0, 0, 216, 176, 220, 60, 0, 0, 216, 177, 220, 60, 0, 0, 216, 178, 220, 60, 0, 0, 216, 179, 220, 60, 0, 0, 216, 180, 220, 60, 0, 0, 216, 181, 220, 60, 0, 0, 216, 182, 220, 60, 0, 0, 216, 183, 220, 60, 0, 0, 216, 184, 220, 60, 0, 0, 216, 185, 220, 60, 0, 0, 216, 186, 220, 60, 0, 0, 216, 187, 220, 60, 0, 0, 216, 188, 220, 60, 0, 0, 216, 189, 220, 60, 0, 0, 216, 190, 220, 60, 0, 0, 216, 191, 220, 60, 0, 0, 216, 192, 220, 60, 0, 0, 216, 193, 220, 60, 0, 0, 216, 194, 220, 60, 0, 0, 216, 195, 220, 60, 0, 0, 216, 196, 220, 60, 0, 0, 216, 197, 220, 60, 0, 0, 216, 198, 220, 60, 0, 0, 216, 199, 220, 60, 0, 0, 216, 200, 220, 60, 0, 0, 216, 201, 220, 60, 0, 0, 216, 202, 220, 60, 0, 0, 216, 203, 220, 60, 0, 0, 216, 204, 220, 60, 0, 0, 216, 205, 220, 60, 0, 0, 216, 206, 220, 60, 0, 0, 216, 207, 220, 60, 0, 0, 216, 208, 220, 60, 0, 0, 216, 209, 220, 60, 0, 0, 216, 210, 220, 60, 0, 0, 216, 211, 220, 60, 0, 0, 216, 212, 220, 60, 0, 0, 216, 213, 220, 60, 0, 0, 216, 214, 220, 60, 0, 0, 216, 215, 220, 60, 0, 0, 216, 216, 220, 60, 0, 0, 216, 217, 220, 60, 0, 0, 216, 218, 220, 60, 0, 0, 216, 219, 220, 60, 0, 0, 216, 220, 220, 60, 0, 0, 216, 221, 220, 60, 0, 0, 216, 222, 220, 60, 0, 0, 216, 223, 220, 60, 0, 0, 216, 224, 220, 60, 0, 0, 216, 225, 220, 60, 0, 0, 216, 226, 220, 60, 0, 0, 216, 227, 220, 60, 0, 0, 216, 228, 220, 60, 0, 0, 216, 229, 220, 60, 0, 0, 216, 230, 220, 60, 0, 0, 216, 231, 220, 60, 0, 0, 216, 232, 220, 60, 0, 0, 216, 233, 220, 60, 0, 0, 216, 234, 220, 60, 0, 0, 216, 235, 220, 60, 0, 0, 216, 236, 220, 60, 0, 0, 216, 237, 220, 60, 0, 0, 216, 238, 220, 60, 0, 0, 216, 239, 220, 60, 0, 0, 216, 240, 220, 60, 0, 0, 216, 241, 220, 60, 0, 0, 216, 242, 220, 60, 0, 0, 216, 243, 220, 60, 0, 0, 216, 244, 220, 60, 0, 0, 216, 245, 220, 60, 0, 0, 216, 246, 220, 60, 0, 0, 216, 247, 220, 60, 0, 0, 216, 248, 220, 60, 0, 0, 216, 249, 220, 60, 0, 0, 216, 250, 220, 60, 0, 1, 216, 0, 222, 60, 0, 1, 216, 1, 222, 60, 0, 1, 216, 2, 222, 60, 0, 1, 216, 3, 222, 60, 0, 1, 216, 4, 222, 60, 0, 1, 216, 5, 222, 60, 0, 1, 216, 6, 222, 60, 0, 1, 216, 7, 222, 60, 0, 1, 216, 8, 222, 60, 0, 1, 216, 9, 222, 60, 0, 1, 216, 10, 222, 60, 0, 1, 216, 11, 222, 60, 0, 1, 216, 12, 222, 60, 0, 1, 216, 13, 222, 60, 0, 1, 216, 14, 222, 60, 0, 1, 216, 15, 222, 60, 0, 1, 216, 16, 222, 60, 0, 1, 216, 17, 222, 60, 0, 1, 216, 18, 222, 60, 0, 1, 216, 19, 222, 60, 0, 1, 216, 20, 222, 60, 0, 1, 216, 21, 222, 60, 0, 1, 216, 22, 222, 60, 0, 1, 216, 23, 222, 60, 0, 1, 216, 24, 222, 60, 0, 1, 216, 25, 222, 60, 0, 1, 216, 26, 222, 60, 0, 1, 216, 27, 222, 60, 0, 1, 216, 28, 222, 60, 0, 1, 216, 29, 222, 60, 0, 1, 216, 30, 222, 60, 0, 1, 216, 31, 222, 60, 0, 1, 216, 32, 222, 60, 0, 1, 216, 33, 222, 60, 0, 1, 216, 34, 222, 60, 0, 1, 216, 35, 222, 60, 0, 1, 216, 36, 222, 60, 0, 1, 216, 37, 222, 60, 0, 1, 216, 38, 222, 60, 0, 1, 216, 39, 222, 60, 0, 1, 216, 40, 222, 60, 0, 1, 216, 41, 222, 60, 0, 1, 216, 42, 222, 60, 0, 1, 216, 43, 222, 60, 0, 1, 216, 44, 222, 60, 0, 1, 216, 45, 222, 60, 0, 1, 216, 46, 222, 60, 0, 1, 216, 47, 222, 60, 0, 1, 216, 48, 222, 60, 0, 1, 216, 49, 222, 60, 0, 1, 216, 50, 222, 60, 0, 1, 216, 51, 222, 60, 0, 1, 216, 52, 222, 60, 0, 1, 216, 53, 222, 60, 0, 1, 216, 54, 222, 60, 0, 1, 216, 55, 222, 60, 0, 1, 216, 56, 222, 60, 0, 1, 216, 57, 222, 60, 0, 1, 216, 58, 222, 60, 0, 1, 216, 59, 222, 60, 0, 1, 216, 60, 222, 60, 0, 1, 216, 61, 222, 60, 0, 1, 216, 62, 222, 60, 0, 1, 216, 63, 222, 60, 0, 1, 216, 64, 222, 60, 0, 1, 216, 65, 222, 60, 0, 1, 216, 66, 222, 60, 0, 1, 216, 67, 222, 60, 0, 1, 216, 68, 222, 60, 0, 1, 216, 69, 222, 60, 0, 1, 216, 70, 222, 60, 0, 1, 216, 71, 222, 60, 0, 1, 216, 72, 222, 60, 0, 1, 216, 73, 222, 60, 0, 1, 216, 74, 222, 60, 0, 1, 216, 75, 222, 60, 0, 1, 216, 76, 222, 60, 0, 1, 216, 77, 222, 60, 0, 1, 216, 78, 222, 60, 0, 1, 216, 79, 222, 60, 0, 1, 216, 80, 222, 60, 0, 1, 216, 81, 222, 60, 0, 1, 216, 82, 222, 60, 0, 1, 216, 83, 222, 60, 0, 1, 216, 84, 222, 60, 0, 1, 216, 85, 222, 60, 0, 1, 216, 86, 222, 60, 0, 1, 216, 87, 222, 60, 0, 1, 216, 88, 222, 60, 0, 1, 216, 89, 222, 60, 0, 1, 216, 90, 222, 60, 0, 1, 216, 91, 222, 60, 0, 1, 216, 92, 222, 60, 0, 1, 216, 93, 222, 60, 0, 1, 216, 94, 222, 60, 0, 1, 216, 95, 222, 60, 0, 1, 216, 96, 222, 60, 0, 1, 216, 97, 222, 60, 0, 1, 216, 98, 222, 60, 0, 1, 216, 99, 222, 60, 0, 1, 216, 100, 222, 60, 0, 1, 216, 101, 222, 60, 0, 1, 216, 102, 222, 60, 0, 1, 216, 103, 222, 60, 0, 1, 216, 104, 222, 60, 0, 1, 216, 105, 222, 60, 0, 1, 216, 106, 222, 60, 0, 1, 216, 107, 222, 60, 0, 1, 216, 108, 222, 60, 0, 1, 216, 109, 222, 60, 0, 1, 216, 110, 222, 60, 0, 1, 216, 111, 222, 60, 0, 1, 216, 112, 222, 60, 0, 1, 216, 113, 222, 60, 0, 1, 216, 114, 222, 60, 0, 1, 216, 115, 222, 60, 0, 1, 216, 116, 222, 60, 0, 1, 216, 117, 222, 60, 0, 1, 216, 118, 222, 60, 0, 1, 216, 119, 222, 60, 0, 1, 216, 120, 222, 60, 0, 1, 216, 121, 222, 60, 0, 1, 216, 122, 222, 60, 0, 1, 216, 123, 222, 60, 0, 1, 216, 124, 222, 60, 0, 1, 216, 125, 222, 60, 0, 1, 216, 126, 222, 60, 0, 1, 216, 127, 222, 60, 0, 1, 216, 128, 222, 60, 0, 1, 216, 129, 222, 60, 0, 1, 216, 130, 222, 60, 0, 1, 216, 131, 222, 60, 0, 1, 216, 132, 222, 60, 0, 1, 216, 133, 222, 60, 0, 1, 216, 134, 222, 60, 0, 1, 216, 135, 222, 60, 0, 1, 216, 136, 222, 60, 0, 1, 216, 137, 222, 60, 0, 1, 216, 138, 222, 60, 0, 1, 216, 139, 222, 60, 0, 1, 216, 140, 222, 60, 0, 1, 216, 141, 222, 60, 0, 1, 216, 142, 222, 60, 0, 1, 216, 143, 222, 60, 0, 1, 216, 144, 222, 60, 0, 1, 216, 145, 222, 60, 0, 1, 216, 146, 222, 60, 0, 1, 216, 147, 222, 60, 0, 1, 216, 148, 222, 60, 0, 1, 216, 149, 222, 60, 0, 1, 216, 150, 222, 60, 0, 1, 216, 151, 222, 60, 0, 1, 216, 152, 222, 60, 0, 1, 216, 153, 222, 60, 0, 1, 216, 154, 222, 60, 0, 1, 216, 155, 222, 60, 0, 1, 216, 156, 222, 60, 0, 1, 216, 157, 222, 60, 0, 1, 216, 158, 222, 60, 0, 1, 216, 159, 222, 60, 0, 1, 216, 160, 222, 60, 0, 1, 216, 161, 222, 60, 0, 1, 216, 162, 222, 60, 0, 1, 216, 163, 222, 60, 0, 1, 216, 164, 222, 60, 0, 1, 216, 165, 222, 60, 0, 1, 216, 166, 222, 60, 0, 1, 216, 167, 222, 60, 0, 1, 216, 168, 222, 60, 0, 1, 216, 169, 222, 60, 0, 1, 216, 170, 222, 60, 0, 1, 216, 171, 222, 60, 0, 1, 216, 172, 222, 60, 0, 1, 216, 173, 222, 60, 0, 1, 216, 174, 222, 60, 0, 1, 216, 175, 222, 60, 0, 1, 216, 176, 222, 60, 0, 1, 216, 177, 222, 60, 0, 1, 216, 178, 222, 60, 0, 1, 216, 179, 222, 60, 0, 1, 216, 180, 222, 60, 0, 1, 216, 181, 222, 60, 0, 1, 216, 182, 222, 60, 0, 1, 216, 183, 222, 60, 0, 1, 216, 184, 222, 60, 0, 1, 216, 185, 222, 60, 0, 1, 216, 186, 222, 60, 0, 1, 216, 187, 222, 60, 0, 1, 216, 188, 222, 60, 0, 1, 216, 189, 222, 60, 0, 1, 216, 190, 222, 60, 0, 1, 216, 191, 222, 60, 0, 1, 216, 192, 222, 60, 0, 1, 216, 193, 222, 60, 0, 1, 216, 194, 222, 60, 0, 1, 216, 195, 222, 60, 0, 1, 216, 196, 222, 60, 0, 1, 216, 197, 222, 60, 0, 1, 216, 198, 222, 60, 0, 1, 216, 199, 222, 60, 0, 1, 216, 200, 222, 60, 0, 1, 216, 201, 222, 60, 0, 1, 216, 202, 222, 60, 0, 1, 216, 203, 222, 60, 0, 1, 216, 204, 222, 60, 0, 1, 216, 205, 222, 60, 0, 1, 216, 206, 222, 60, 0, 1, 216, 207, 222, 60, 0, 1, 216, 208, 222, 60, 0, 1, 216, 209, 222, 60, 0, 1, 216, 210, 222, 60, 0, 1, 216, 211, 222, 60, 0, 1, 216, 212, 222, 60, 0, 1, 216, 213, 222, 60, 0, 1, 216, 214, 222, 60, 0, 1, 216, 215, 222, 60, 0, 1, 216, 216, 222, 60, 0, 1, 216, 217, 222, 60, 0, 1, 216, 218, 222, 60, 0, 1, 216, 219, 222, 60, 0, 1, 216, 220, 222, 60, 0, 1, 216, 221, 222, 60, 0, 1, 216, 222, 222, 60, 0, 1, 216, 223, 222, 60, 0, 1, 216, 224, 222, 60, 0, 1, 216, 225, 222, 60, 0, 1, 216, 226, 222, 60, 0, 1, 216, 227, 222, 60, 0, 1, 216, 228, 222, 60, 0, 1, 216, 229, 222, 60, 0, 1, 216, 230, 222, 60, 0, 1, 216, 231, 222, 60, 0, 1, 216, 232, 222, 60, 0, 1, 216, 233, 222, 60, 0, 1, 216, 234, 222, 60, 0, 1, 216, 235, 222, 60, 0, 1, 216, 236, 222, 60, 0, 1, 216, 237, 222, 60, 0, 1, 216, 238, 222, 60, 0, 1, 216, 239, 222, 60, 0, 1, 216, 240, 222, 60, 0, 1, 216, 241, 222, 60, 0, 1, 216, 242, 222, 60, 0, 1, 216, 243, 222, 60, 0, 1, 216, 244, 222, 60, 0, 1, 216, 245, 222, 60, 0, 1, 216, 246, 222, 60, 0, 1, 216, 247, 222, 60, 0, 1, 216, 248, 222, 60, 0, 1, 216, 249, 222, 60, 0, 1, 216, 250, 222, 60, 0, 1, 216, 251, 222, 60, 0, 1, 216, 252, 222, 60, 0, 1, 216, 253, 222, 60, 0, 1, 216, 254, 222, 60, 0, 1, 216, 255, 222, 60, 0, 1, 216, 0, 223, 60, 0, 1, 216, 1, 223, 60, 0, 1, 216, 2, 223, 60, 0, 1, 216, 3, 223, 60, 0, 1, 216, 4, 223, 60, 0, 1, 216, 5, 223, 60, 0, 1, 216, 6, 223, 60, 0, 1, 216, 7, 223, 60, 0, 1, 216, 8, 223, 60, 0, 1, 216, 9, 223, 60, 0, 1, 216, 10, 223, 60, 0, 1, 216, 11, 223, 60, 0, 1, 216, 12, 223, 60, 0, 1, 216, 13, 223, 60, 0, 1, 216, 14, 223, 60, 0, 1, 216, 15, 223, 60, 0, 1, 216, 16, 223, 60, 0, 1, 216, 17, 223, 60, 0, 1, 216, 18, 223, 60, 0, 1, 216, 19, 223, 60, 0, 1, 216, 20, 223, 60, 0, 1, 216, 21, 223, 60, 0, 1, 216, 22, 223, 60, 0, 1, 216, 23, 223, 60, 0, 1, 216, 24, 223, 60, 0, 1, 216, 25, 223, 60, 0, 1, 216, 26, 223, 60, 0, 1, 216, 27, 223, 60, 0, 1, 216, 28, 223, 60, 0, 1, 216, 29, 223, 60, 0, 1, 216, 30, 223, 60, 0, 1, 216, 31, 223, 60, 0, 1, 216, 32, 223, 60, 0, 1, 216, 33, 223, 60, 0, 1, 216, 34, 223, 60, 0, 1, 216, 35, 223, 60, 0, 1, 216, 36, 223, 60, 0, 1, 216, 37, 223, 60, 0, 1, 216, 38, 223, 60, 0, 1, 216, 39, 223, 60, 0, 1, 216, 40, 223, 60, 0, 1, 216, 41, 223, 60, 0, 1, 216, 42, 223, 60, 0, 1, 216, 43, 223, 60, 0, 1, 216, 44, 223, 60, 0, 1, 216, 45, 223, 60, 0, 1, 216, 46, 223, 60, 0, 1, 216, 47, 223, 60, 0, 1, 216, 48, 223, 60, 0, 1, 216, 49, 223, 60, 0, 1, 216, 50, 223, 60, 0, 1, 216, 51, 223, 60, 0, 1, 216, 52, 223, 60, 0, 1, 216, 53, 223, 60, 0, 1, 216, 54, 223, 60, 0, 1, 216, 64, 223, 60, 0, 1, 216, 65, 223, 60, 0, 1, 216, 66, 223, 60, 0, 1, 216, 67, 223, 60, 0, 1, 216, 68, 223, 60, 0, 1, 216, 69, 223, 60, 0, 1, 216, 70, 223, 60, 0, 1, 216, 71, 223, 60, 0, 1, 216, 72, 223, 60, 0, 1, 216, 73, 223, 60, 0, 1, 216, 74, 223, 60, 0, 1, 216, 75, 223, 60, 0, 1, 216, 76, 223, 60, 0, 1, 216, 77, 223, 60, 0, 1, 216, 78, 223, 60, 0, 1, 216, 79, 223, 60, 0, 1, 216, 80, 223, 60, 0, 1, 216, 81, 223, 60, 0, 1, 216, 82, 223, 60, 0, 1, 216, 83, 223, 60, 0, 1, 216, 84, 223, 60, 0, 1, 216, 85, 223, 60, 0, 1, 216, 96, 223, 60, 0, 1, 216, 97, 223, 60, 0, 1, 216, 98, 223, 60, 0, 1, 216, 99, 223, 60, 0, 1, 216, 100, 223, 60, 0, 1, 216, 101, 223, 60, 0, 1, 216, 102, 223, 60, 0, 1, 216, 103, 223, 60, 0, 14, 32, 2, 216, 0, 220, 14, 32, 60, 0, 14, 32, 2, 216, 1, 220, 14, 32, 60, 0, 14, 32, 2, 216, 2, 220, 14, 32, 60, 0, 14, 32, 2, 216, 3, 220, 14, 32, 60, 0, 14, 32, 2, 216, 4, 220, 14, 32, 60, 0, 14, 32, 2, 216, 5, 220, 14, 32, 60, 0, 14, 32, 2, 216, 8, 220, 14, 32, 60, 0, 14, 32, 2, 216, 10, 220, 14, 32, 60, 0, 14, 32, 2, 216, 11, 220, 14, 32, 60, 0, 14, 32, 2, 216, 12, 220, 14, 32, 60, 0, 14, 32, 2, 216, 13, 220, 14, 32, 60, 0, 14, 32, 2, 216, 14, 220, 14, 32, 60, 0, 14, 32, 2, 216, 15, 220, 14, 32, 60, 0, 14, 32, 2, 216, 16, 220, 14, 32, 60, 0, 14, 32, 2, 216, 17, 220, 14, 32, 60, 0, 14, 32, 2, 216, 18, 220, 14, 32, 60, 0, 14, 32, 2, 216, 19, 220, 14, 32, 60, 0, 14, 32, 2, 216, 20, 220, 14, 32, 60, 0, 14, 32, 2, 216, 21, 220, 14, 32, 60, 0, 14, 32, 2, 216, 22, 220, 14, 32, 60, 0, 14, 32, 2, 216, 23, 220, 14, 32, 60, 0, 14, 32, 2, 216, 24, 220, 14, 32, 60, 0, 14, 32, 2, 216, 25, 220, 14, 32, 60, 0, 14, 32, 2, 216, 26, 220, 14, 32, 60, 0, 14, 32, 2, 216, 27, 220, 14, 32, 60, 0, 14, 32, 2, 216, 28, 220, 14, 32, 60, 0, 14, 32, 2, 216, 29, 220, 14, 32, 60, 0, 14, 32, 2, 216, 30, 220, 14, 32, 60, 0, 14, 32, 2, 216, 31, 220, 14, 32, 60, 0, 14, 32, 2, 216, 32, 220, 14, 32, 60, 0, 14, 32, 2, 216, 33, 220, 14, 32, 60, 0, 14, 32, 2, 216, 34, 220, 14, 32, 60, 0, 14, 32, 2, 216, 35, 220, 14, 32, 60, 0, 14, 32, 2, 216, 36, 220, 14, 32, 60, 0, 14, 32, 2, 216, 37, 220, 14, 32, 60, 0, 14, 32, 2, 216, 38, 220, 14, 32, 60, 0, 14, 32, 2, 216, 39, 220, 14, 32, 60, 0, 14, 32, 2, 216, 40, 220, 14, 32, 60, 0, 14, 32, 2, 216, 41, 220, 14, 32, 60, 0, 14, 32, 2, 216, 42, 220, 14, 32, 60, 0, 14, 32, 2, 216, 43, 220, 14, 32, 60, 0, 14, 32, 2, 216, 44, 220, 14, 32, 60, 0, 14, 32, 2, 216, 45, 220, 14, 32, 60, 0, 14, 32, 2, 216, 46, 220, 14, 32, 60, 0, 14, 32, 2, 216, 47, 220, 14, 32, 60, 0, 14, 32, 2, 216, 48, 220, 14, 32, 60, 0, 14, 32, 2, 216, 49, 220, 14, 32, 60, 0, 14, 32, 2, 216, 50, 220, 14, 32, 60, 0, 14, 32, 2, 216, 51, 220, 14, 32, 60, 0, 14, 32, 2, 216, 52, 220, 14, 32, 60, 0, 14, 32, 2, 216, 53, 220, 14, 32, 60, 0, 14, 32, 2, 216, 55, 220, 14, 32, 60, 0, 14, 32, 2, 216, 56, 220, 14, 32, 60, 0, 14, 32, 2, 216, 60, 220, 14, 32, 60, 0, 14, 32, 2, 216, 63, 220, 14, 32, 60, 0, 14, 32, 2, 216, 96, 222, 14, 32, 60, 0, 14, 32, 2, 216, 97, 222, 14, 32, 60, 0, 14, 32, 2, 216, 98, 222, 14, 32, 60, 0, 14, 32, 2, 216, 99, 222, 14, 32, 60, 0, 14, 32, 2, 216, 100, 222, 14, 32, 60, 0, 14, 32, 2, 216, 101, 222, 14, 32, 60, 0, 14, 32, 2, 216, 102, 222, 14, 32, 60, 0, 14, 32, 2, 216, 103, 222, 14, 32, 60, 0, 14, 32, 2, 216, 104, 222, 14, 32, 60, 0, 14, 32, 2, 216, 105, 222, 14, 32, 60, 0, 14, 32, 2, 216, 106, 222, 14, 32, 60, 0, 14, 32, 2, 216, 107, 222, 14, 32, 60, 0, 14, 32, 2, 216, 108, 222, 14, 32, 60, 0, 14, 32, 2, 216, 109, 222, 14, 32, 60, 0, 14, 32, 2, 216, 110, 222, 14, 32, 60, 0, 14, 32, 2, 216, 111, 222, 14, 32, 60, 0, 14, 32, 2, 216, 112, 222, 14, 32, 60, 0, 14, 32, 2, 216, 113, 222, 14, 32, 60, 0, 14, 32, 2, 216, 114, 222, 14, 32, 60, 0, 14, 32, 2, 216, 115, 222, 14, 32, 60, 0, 14, 32, 2, 216, 116, 222, 14, 32, 60, 0, 14, 32, 2, 216, 117, 222, 14, 32, 60, 0, 14, 32, 2, 216, 118, 222, 14, 32, 60, 0, 14, 32, 2, 216, 119, 222, 14, 32, 60, 0, 14, 32, 2, 216, 120, 222, 14, 32, 60, 0, 14, 32, 2, 216, 121, 222, 14, 32, 60, 0, 14, 32, 2, 216, 122, 222, 14, 32, 60, 0, 14, 32, 2, 216, 123, 222, 14, 32, 60, 0, 14, 32, 2, 216, 124, 222, 14, 32, 60, 0, 14, 32, 2, 216, 128, 222, 14, 32, 60, 0, 14, 32, 2, 216, 129, 222, 14, 32, 60, 0, 14, 32, 2, 216, 130, 222, 14, 32, 60, 0, 14, 32, 2, 216, 131, 222, 14, 32, 60, 0, 14, 32, 2, 216, 132, 222, 14, 32, 60, 0, 14, 32, 2, 216, 133, 222, 14, 32, 60, 0, 14, 32, 2, 216, 134, 222, 14, 32, 60, 0, 14, 32, 2, 216, 135, 222, 14, 32, 60, 0, 14, 32, 2, 216, 136, 222, 14, 32, 60, 0, 14, 32, 2, 216, 137, 222, 14, 32, 60, 0, 14, 32, 2, 216, 138, 222, 14, 32, 60, 0, 14, 32, 2, 216, 139, 222, 14, 32, 60, 0, 14, 32, 2, 216, 140, 222, 14, 32, 60, 0, 14, 32, 2, 216, 141, 222, 14, 32, 60, 0, 14, 32, 2, 216, 142, 222, 14, 32, 60, 0, 14, 32, 2, 216, 143, 222, 14, 32, 60, 0, 14, 32, 2, 216, 144, 222, 14, 32, 60, 0, 14, 32, 2, 216, 145, 222, 14, 32, 60, 0, 14, 32, 2, 216, 146, 222, 14, 32, 60, 0, 14, 32, 2, 216, 147, 222, 14, 32, 60, 0, 14, 32, 2, 216, 148, 222, 14, 32, 60, 0, 14, 32, 2, 216, 149, 222, 14, 32, 60, 0, 14, 32, 2, 216, 150, 222, 14, 32, 60, 0, 14, 32, 2, 216, 151, 222, 14, 32, 60, 0, 14, 32, 2, 216, 152, 222, 14, 32, 60, 0, 14, 32, 2, 216, 153, 222, 14, 32, 60, 0, 14, 32, 2, 216, 154, 222, 14, 32, 60, 0, 14, 32, 2, 216, 155, 222, 14, 32, 60, 0, 14, 32, 2, 216, 156, 222, 14, 32, 60, 0, 14, 32, 2, 216, 0, 223, 14, 32, 60, 0, 14, 32, 2, 216, 1, 223, 14, 32, 60, 0, 14, 32, 2, 216, 2, 223, 14, 32, 60, 0, 14, 32, 2, 216, 3, 223, 14, 32, 60, 0, 14, 32, 2, 216, 4, 223, 14, 32, 60, 0, 14, 32, 2, 216, 5, 223, 14, 32, 60, 0, 14, 32, 2, 216, 6, 223, 14, 32, 60, 0, 14, 32, 2, 216, 7, 223, 14, 32, 60, 0, 14, 32, 2, 216, 8, 223, 14, 32, 60, 0, 14, 32, 2, 216, 9, 223, 14, 32, 60, 0, 14, 32, 2, 216, 10, 223, 14, 32, 60, 0, 14, 32, 2, 216, 11, 223, 14, 32, 60, 0, 14, 32, 2, 216, 12, 223, 14, 32, 60, 0, 14, 32, 2, 216, 13, 223, 14, 32, 60, 0, 14, 32, 2, 216, 14, 223, 14, 32, 60, 0, 14, 32, 2, 216, 15, 223, 14, 32, 60, 0, 14, 32, 2, 216, 16, 223, 14, 32, 60, 0, 14, 32, 2, 216, 17, 223, 14, 32, 60, 0, 14, 32, 2, 216, 18, 223, 14, 32, 60, 0, 14, 32, 2, 216, 19, 223, 14, 32, 60, 0, 14, 32, 2, 216, 20, 223, 14, 32, 60, 0, 14, 32, 2, 216, 21, 223, 14, 32, 60, 0, 14, 32, 2, 216, 22, 223, 14, 32, 60, 0, 14, 32, 2, 216, 23, 223, 14, 32, 60, 0, 14, 32, 2, 216, 24, 223, 14, 32, 60, 0, 14, 32, 2, 216, 25, 223, 14, 32, 60, 0, 14, 32, 2, 216, 26, 223, 14, 32, 60, 0, 14, 32, 2, 216, 27, 223, 14, 32, 60, 0, 14, 32, 2, 216, 28, 223, 14, 32, 60, 0, 14, 32, 2, 216, 29, 223, 14, 32, 60, 0, 14, 32, 2, 216, 30, 223, 14, 32, 60, 0, 14, 32, 2, 216, 31, 223, 14, 32, 60, 0, 14, 32, 2, 216, 32, 223, 14, 32, 60, 0, 14, 32, 2, 216, 33, 223, 14, 32, 60, 0, 14, 32, 2, 216, 34, 223, 14, 32, 60, 0, 14, 32, 2, 216, 35, 223, 14, 32, 60, 0, 14, 32, 2, 216, 36, 223, 14, 32, 60, 0, 14, 32, 2, 216, 37, 223, 14, 32, 60, 0, 14, 32, 2, 216, 38, 223, 14, 32, 60, 0, 14, 32, 2, 216, 39, 223, 14, 32, 60, 0, 14, 32, 2, 216, 40, 223, 14, 32, 60, 0, 14, 32, 2, 216, 41, 223, 14, 32, 60, 0, 14, 32, 2, 216, 42, 223, 14, 32, 60, 0, 14, 32, 2, 216, 43, 223, 14, 32, 60, 0, 14, 32, 2, 216, 44, 223, 14, 32, 60, 0, 14, 32, 2, 216, 45, 223, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 46, 223, 14, 32, 60, 0, 14, 32, 2, 216, 47, 223, 14, 32, 60, 0, 14, 32, 2, 216, 48, 223, 14, 32, 60, 0, 14, 32, 2, 216, 49, 223, 14, 32, 60, 0, 14, 32, 2, 216, 50, 223, 14, 32, 60, 0, 14, 32, 2, 216, 51, 223, 14, 32, 60, 0, 14, 32, 2, 216, 52, 223, 14, 32, 60, 0, 14, 32, 2, 216, 53, 223, 14, 32, 60, 0, 14, 32, 2, 216, 96, 220, 14, 32, 60, 0, 14, 32, 2, 216, 97, 220, 14, 32, 60, 0, 14, 32, 2, 216, 98, 220, 14, 32, 60, 0, 14, 32, 2, 216, 99, 220, 14, 32, 60, 0, 14, 32, 2, 216, 100, 220, 14, 32, 60, 0, 14, 32, 2, 216, 101, 220, 14, 32, 60, 0, 14, 32, 2, 216, 102, 220, 14, 32, 60, 0, 14, 32, 2, 216, 103, 220, 14, 32, 60, 0, 14, 32, 2, 216, 104, 220, 14, 32, 60, 0, 14, 32, 2, 216, 105, 220, 14, 32, 60, 0, 14, 32, 2, 216, 106, 220, 14, 32, 60, 0, 14, 32, 2, 216, 107, 220, 14, 32, 60, 0, 14, 32, 2, 216, 108, 220, 14, 32, 60, 0, 14, 32, 2, 216, 110, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 216, 109, 220, 14, 32, 60, 0, 14, 32, 2, 216, 111, 220, 14, 32, 60, 0, 14, 32, 2, 216, 112, 220, 14, 32, 60, 0, 14, 32, 2, 216, 113, 220, 14, 32, 60, 0, 14, 32, 2, 216, 114, 220, 14, 32, 60, 0, 14, 32, 2, 216, 115, 220, 14, 32, 60, 0, 14, 32, 2, 216, 116, 220, 14, 32, 60, 0, 14, 32, 2, 216, 117, 220, 14, 32, 60, 0, 14, 32, 2, 216, 118, 220, 14, 32, 60, 0, 14, 32, 2, 216, 129, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 216, 128, 220, 14, 32, 60, 0, 14, 32, 2, 216, 131, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 216, 130, 220, 14, 32, 60, 0, 14, 32, 2, 216, 132, 220, 14, 32, 60, 0, 14, 32, 2, 216, 133, 220, 14, 32, 60, 0, 14, 32, 2, 216, 135, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 216, 134, 220, 14, 32, 60, 0, 14, 32, 2, 216, 136, 220, 14, 32, 60, 0, 14, 32, 2, 216, 137, 220, 14, 32, 60, 0, 14, 32, 2, 216, 138, 220, 14, 32, 60, 0, 14, 32, 2, 216, 139, 220, 14, 32, 60, 0, 14, 32, 2, 216, 141, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 216, 140, 220, 14, 32, 60, 0, 14, 32, 2, 216, 143, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 216, 142, 220, 14, 32, 60, 0, 14, 32, 2, 216, 145, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 216, 144, 220, 14, 32, 60, 0, 14, 32, 2, 216, 147, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 216, 146, 220, 14, 32, 60, 0, 14, 32, 2, 216, 149, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 216, 148, 220, 14, 32, 60, 0, 14, 32, 2, 216, 150, 220, 14, 32, 60, 0, 14, 32, 2, 216, 151, 220, 14, 32, 60, 0, 14, 32, 2, 216, 152, 220, 14, 32, 60, 0, 14, 32, 2, 216, 153, 220, 14, 32, 60, 0, 14, 32, 2, 216, 154, 220, 14, 32, 60, 0, 14, 32, 2, 216, 155, 220, 14, 32, 60, 0, 14, 32, 2, 216, 157, 220, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 2, 216, 156, 220, 14, 32, 60, 0, 14, 32, 2, 216, 158, 220, 14, 32, 60, 0, 14, 32, 2, 216, 224, 220, 14, 32, 60, 0, 14, 32, 2, 216, 225, 220, 14, 32, 60, 0, 14, 32, 2, 216, 226, 220, 14, 32, 60, 0, 14, 32, 2, 216, 227, 220, 14, 32, 60, 0, 14, 32, 2, 216, 228, 220, 14, 32, 60, 0, 14, 32, 2, 216, 229, 220, 14, 32, 60, 0, 14, 32, 2, 216, 230, 220, 14, 32, 60, 0, 14, 32, 2, 216, 231, 220, 14, 32, 60, 0, 14, 32, 2, 216, 232, 220, 14, 32, 60, 0, 14, 32, 2, 216, 233, 220, 14, 32, 60, 0, 14, 32, 2, 216, 234, 220, 14, 32, 60, 0, 14, 32, 2, 216, 235, 220, 14, 32, 60, 0, 14, 32, 2, 216, 236, 220, 14, 32, 60, 0, 14, 32, 2, 216, 237, 220, 14, 32, 60, 0, 14, 32, 2, 216, 238, 220, 14, 32, 60, 0, 14, 32, 2, 216, 239, 220, 14, 32, 60, 0, 14, 32, 2, 216, 240, 220, 14, 32, 60, 0, 14, 32, 2, 216, 241, 220, 14, 32, 60, 0, 14, 32, 2, 216, 242, 220, 14, 32, 60, 0, 14, 32, 2, 216, 244, 220, 14, 32, 60, 0, 14, 32, 2, 216, 245, 220, 14, 32, 60, 0, 14, 32, 2, 216, 64, 220, 14, 32, 60, 0, 14, 32, 2, 216, 65, 220, 14, 32, 60, 0, 14, 32, 2, 216, 66, 220, 14, 32, 60, 0, 14, 32, 2, 216, 67, 220, 14, 32, 60, 0, 14, 32, 2, 216, 68, 220, 14, 32, 60, 0, 14, 32, 2, 216, 69, 220, 14, 32, 60, 0, 14, 32, 2, 216, 70, 220, 14, 32, 60, 0, 14, 32, 2, 216, 71, 220, 14, 32, 60, 0, 14, 32, 2, 216, 72, 220, 14, 32, 60, 0, 14, 32, 2, 216, 73, 220, 14, 32, 60, 0, 14, 32, 2, 216, 74, 220, 14, 32, 60, 0, 14, 32, 2, 216, 75, 220, 14, 32, 60, 0, 14, 32, 2, 216, 76, 220, 14, 32, 60, 0, 14, 32, 2, 216, 77, 220, 14, 32, 60, 0, 14, 32, 2, 216, 78, 220, 14, 32, 60, 0, 14, 32, 2, 216, 79, 220, 14, 32, 60, 0, 14, 32, 2, 216, 80, 220, 14, 32, 60, 0, 14, 32, 2, 216, 81, 220, 14, 32, 60, 0, 14, 32, 2, 216, 82, 220, 14, 32, 60, 0, 14, 32, 2, 216, 83, 220, 14, 32, 60, 0, 14, 32, 2, 216, 84, 220, 14, 32, 60, 0, 14, 32, 2, 216, 85, 220, 14, 32, 60, 0, 14, 32, 2, 216, 64, 223, 14, 32, 60, 0, 14, 32, 2, 216, 65, 223, 14, 32, 60, 0, 14, 32, 2, 216, 66, 223, 14, 32, 60, 0, 14, 32, 2, 216, 67, 223, 14, 32, 60, 0, 14, 32, 2, 216, 68, 223, 14, 32, 60, 0, 14, 32, 2, 216, 69, 223, 14, 32, 60, 0, 14, 32, 2, 216, 70, 223, 14, 32, 60, 0, 14, 32, 2, 216, 71, 223, 14, 32, 60, 0, 14, 32, 2, 216, 72, 223, 14, 32, 60, 0, 14, 32, 2, 216, 73, 223, 14, 32, 60, 0, 14, 32, 2, 216, 74, 223, 14, 32, 60, 0, 14, 32, 2, 216, 75, 223, 14, 32, 60, 0, 14, 32, 2, 216, 76, 223, 14, 32, 60, 0, 14, 32, 2, 216, 77, 223, 14, 32, 60, 0, 14, 32, 2, 216, 78, 223, 14, 32, 60, 0, 14, 32, 2, 216, 79, 223, 14, 32, 60, 0, 14, 32, 2, 216, 80, 223, 14, 32, 60, 0, 14, 32, 2, 216, 81, 223, 14, 32, 60, 0, 14, 32, 2, 216, 82, 223, 14, 32, 60, 0, 14, 32, 2, 216, 83, 223, 14, 32, 60, 0, 14, 32, 2, 216, 84, 223, 14, 32, 60, 0, 14, 32, 2, 216, 85, 223, 14, 32, 60, 0, 14, 32, 2, 216, 96, 223, 14, 32, 60, 0, 14, 32, 2, 216, 97, 223, 14, 32, 60, 0, 14, 32, 2, 216, 98, 223, 14, 32, 60, 0, 14, 32, 2, 216, 99, 223, 14, 32, 60, 0, 14, 32, 2, 216, 100, 223, 14, 32, 60, 0, 14, 32, 2, 216, 101, 223, 14, 32, 60, 0, 14, 32, 2, 216, 102, 223, 14, 32, 60, 0, 14, 32, 2, 216, 103, 223, 14, 32, 60, 0, 14, 32, 2, 216, 104, 223, 14, 32, 60, 0, 14, 32, 2, 216, 105, 223, 14, 32, 60, 0, 14, 32, 2, 216, 106, 223, 14, 32, 60, 0, 14, 32, 2, 216, 107, 223, 14, 32, 60, 0, 14, 32, 2, 216, 108, 223, 14, 32, 60, 0, 14, 32, 2, 216, 109, 223, 14, 32, 60, 0, 14, 32, 2, 216, 110, 223, 14, 32, 60, 0, 14, 32, 2, 216, 111, 223, 14, 32, 60, 0, 14, 32, 2, 216, 112, 223, 14, 32, 60, 0, 14, 32, 2, 216, 113, 223, 14, 32, 60, 0, 14, 32, 2, 216, 114, 223, 14, 32, 60, 0, 14, 32, 2, 216, 128, 223, 14, 32, 60, 0, 14, 32, 2, 216, 129, 223, 14, 32, 60, 0, 14, 32, 2, 216, 130, 223, 14, 32, 60, 0, 14, 32, 2, 216, 131, 223, 14, 32, 60, 0, 14, 32, 2, 216, 132, 223, 14, 32, 60, 0, 14, 32, 2, 216, 133, 223, 14, 32, 60, 0, 14, 32, 2, 216, 134, 223, 14, 32, 60, 0, 14, 32, 2, 216, 135, 223, 14, 32, 60, 0, 14, 32, 2, 216, 136, 223, 14, 32, 60, 0, 14, 32, 2, 216, 137, 223, 14, 32, 60, 0, 14, 32, 2, 216, 138, 223, 14, 32, 60, 0, 14, 32, 2, 216, 139, 223, 14, 32, 60, 0, 14, 32, 2, 216, 140, 223, 14, 32, 60, 0, 14, 32, 2, 216, 141, 223, 14, 32, 60, 0, 14, 32, 2, 216, 142, 223, 14, 32, 60, 0, 14, 32, 2, 216, 143, 223, 14, 32, 60, 0, 14, 32, 2, 216, 144, 223, 14, 32, 60, 0, 14, 32, 2, 216, 145, 223, 14, 32, 60, 0, 14, 32, 3, 216, 224, 223, 14, 32, 60, 0, 14, 32, 3, 216, 225, 223, 14, 32, 60, 0, 14, 32, 3, 216, 226, 223, 14, 32, 60, 0, 14, 32, 3, 216, 227, 223, 14, 32, 60, 0, 14, 32, 3, 216, 228, 223, 14, 32, 60, 0, 14, 32, 3, 216, 229, 223, 14, 32, 60, 0, 14, 32, 3, 216, 230, 223, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 246, 223, 14, 32, 47, 0, 14, 32, 3, 216, 233, 223, 14, 32, 60, 0, 14, 32, 3, 216, 231, 223, 14, 32, 60, 0, 14, 32, 3, 216, 232, 223, 14, 32, 60, 0, 14, 32, 3, 216, 233, 223, 14, 32, 60, 0, 14, 32, 3, 216, 234, 223, 14, 32, 60, 0, 14, 32, 3, 216, 235, 223, 14, 32, 60, 0, 14, 32, 3, 216, 236, 223, 14, 32, 60, 0, 14, 32, 3, 216, 237, 223, 14, 32, 60, 0, 14, 32, 3, 216, 238, 223, 14, 32, 60, 0, 14, 32, 3, 216, 239, 223, 14, 32, 60, 0, 14, 32, 3, 216, 240, 223, 14, 32, 60, 0, 14, 32, 3, 216, 241, 223, 14, 32, 60, 0, 14, 32, 3, 216, 242, 223, 14, 32, 60, 0, 14, 32, 3, 216, 243, 223, 14, 32, 60, 0, 14, 32, 3, 216, 244, 223, 14, 32, 60, 0, 14, 32, 3, 216, 245, 223, 14, 32, 60, 0, 14, 32, 2, 216, 192, 222, 14, 32, 60, 0, 14, 32, 2, 216, 193, 222, 14, 32, 60, 0, 14, 32, 2, 216, 194, 222, 14, 32, 60, 0, 14, 32, 2, 216, 195, 222, 14, 32, 60, 0, 14, 32, 2, 216, 196, 222, 14, 32, 60, 0, 14, 32, 2, 216, 197, 222, 14, 32, 60, 0, 14, 32, 2, 216, 198, 222, 14, 32, 60, 0, 14, 32, 2, 216, 199, 222, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 200, 222, 14, 32, 60, 0, 14, 32, 2, 216, 201, 222, 14, 32, 60, 0, 14, 32, 2, 216, 202, 222, 14, 32, 60, 0, 14, 32, 2, 216, 203, 222, 14, 32, 60, 0, 14, 32, 2, 216, 204, 222, 14, 32, 60, 0, 14, 32, 2, 216, 205, 222, 14, 32, 60, 0, 14, 32, 2, 216, 206, 222, 14, 32, 60, 0, 14, 32, 2, 216, 207, 222, 14, 32, 60, 0, 14, 32, 2, 216, 208, 222, 14, 32, 60, 0, 14, 32, 2, 216, 209, 222, 14, 32, 60, 0, 14, 32, 2, 216, 210, 222, 14, 32, 60, 0, 14, 32, 2, 216, 211, 222, 14, 32, 60, 0, 14, 32, 2, 216, 212, 222, 14, 32, 60, 0, 14, 32, 2, 216, 213, 222, 14, 32, 60, 0, 14, 32, 2, 216, 214, 222, 14, 32, 60, 0, 14, 32, 2, 216, 215, 222, 14, 32, 60, 0, 14, 32, 2, 216, 216, 222, 14, 32, 60, 0, 14, 32, 2, 216, 217, 222, 14, 32, 60, 0, 14, 32, 2, 216, 218, 222, 14, 32, 60, 0, 14, 32, 2, 216, 219, 222, 14, 32, 60, 0, 14, 32, 2, 216, 220, 222, 14, 32, 60, 0, 14, 32, 2, 216, 221, 222, 14, 32, 60, 0, 14, 32, 2, 216, 222, 222, 14, 32, 60, 0, 14, 32, 2, 216, 223, 222, 14, 32, 60, 0, 14, 32, 2, 216, 224, 222, 14, 32, 60, 0, 14, 32, 2, 216, 225, 222, 14, 32, 60, 0, 14, 32, 2, 216, 226, 222, 14, 32, 60, 0, 14, 32, 2, 216, 227, 222, 14, 32, 60, 0, 14, 32, 2, 216, 228, 222, 14, 32, 60, 0, 14, 32, 3, 216, 0, 223, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 1, 223, 14, 32, 60, 0, 14, 32, 3, 216, 2, 223, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 3, 223, 14, 32, 60, 0, 14, 32, 3, 216, 4, 223, 14, 32, 60, 0, 14, 32, 3, 216, 5, 223, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 6, 223, 14, 32, 60, 0, 14, 32, 3, 216, 7, 223, 14, 32, 60, 0, 14, 32, 3, 216, 8, 223, 14, 32, 60, 0, 14, 32, 3, 216, 9, 223, 14, 32, 60, 0, 14, 32, 3, 216, 10, 223, 14, 32, 60, 0, 14, 32, 3, 216, 11, 223, 14, 32, 60, 0, 14, 32, 3, 216, 12, 223, 14, 32, 60, 0, 14, 32, 3, 216, 13, 223, 14, 32, 60, 0, 14, 32, 3, 216, 14, 223, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 15, 223, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 16, 223, 14, 32, 60, 0, 14, 32, 3, 216, 17, 223, 14, 32, 60, 0, 14, 32, 3, 216, 18, 223, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 39, 223, 14, 32, 47, 0, 14, 32, 3, 216, 24, 223, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 19, 223, 14, 32, 60, 0, 14, 32, 3, 216, 20, 223, 14, 32, 60, 0, 14, 32, 3, 216, 21, 223, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 22, 223, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 23, 223, 14, 32, 60, 0, 14, 32, 3, 216, 24, 223, 14, 32, 60, 0, 14, 32, 3, 216, 25, 223, 14, 32, 60, 0, 14, 32, 3, 216, 26, 223, 14, 32, 60, 0, 60, 0, 60, 0, 14, 32, 3, 216, 27, 223, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 28, 223, 14, 32, 60, 0, 14, 32, 3, 216, 48, 223, 14, 32, 60, 0, 14, 32, 3, 216, 49, 223, 14, 32, 60, 0, 14, 32, 3, 216, 50, 223, 14, 32, 60, 0, 14, 32, 3, 216, 51, 223, 14, 32, 60, 0, 14, 32, 3, 216, 52, 223, 14, 32, 60, 0, 14, 32, 3, 216, 53, 223, 14, 32, 60, 0, 14, 32, 3, 216, 54, 223, 14, 32, 60, 0, 14, 32, 3, 216, 55, 223, 14, 32, 60, 0, 14, 32, 3, 216, 56, 223, 14, 32, 60, 0, 14, 32, 3, 216, 57, 223, 14, 32, 60, 0, 14, 32, 3, 216, 58, 223, 14, 32, 60, 0, 14, 32, 3, 216, 59, 223, 14, 32, 60, 0, 14, 32, 3, 216, 60, 223, 14, 32, 60, 0, 14, 32, 3, 216, 61, 223, 14, 32, 60, 0, 14, 32, 3, 216, 62, 223, 14, 32, 60, 0, 14, 32, 3, 216, 63, 223, 14, 32, 60, 0, 14, 32, 3, 216, 64, 223, 14, 32, 60, 0, 14, 32, 3, 216, 65, 223, 14, 32, 60, 0, 60, 0, 14, 32, 3, 216, 69, 223, 14, 32, 60, 0, 14, 32, 3, 216, 66, 223, 14, 32, 60, 0, 14, 32, 3, 216, 67, 223, 14, 32, 60, 0, 14, 32, 3, 216, 68, 223, 14, 32, 60, 0, 14, 32, 3, 216, 128, 222, 14, 32, 60, 0, 14, 32, 3, 216, 129, 222, 14, 32, 60, 0, 14, 32, 3, 216, 130, 222, 14, 32, 60, 0, 14, 32, 3, 216, 131, 222, 14, 32, 60, 0, 14, 32, 3, 216, 132, 222, 14, 32, 60, 0, 14, 32, 3, 216, 133, 222, 14, 32, 60, 0, 14, 32, 3, 216, 134, 222, 14, 32, 60, 0, 14, 32, 3, 216, 135, 222, 14, 32, 60, 0, 14, 32, 3, 216, 136, 222, 14, 32, 60, 0, 14, 32, 3, 216, 137, 222, 14, 32, 60, 0, 14, 32, 3, 216, 138, 222, 14, 32, 60, 0, 14, 32, 3, 216, 139, 222, 14, 32, 60, 0, 14, 32, 3, 216, 140, 222, 14, 32, 60, 0, 14, 32, 3, 216, 141, 222, 14, 32, 60, 0, 14, 32, 3, 216, 142, 222, 14, 32, 60, 0, 14, 32, 3, 216, 143, 222, 14, 32, 60, 0, 14, 32, 3, 216, 144, 222, 14, 32, 60, 0, 14, 32, 3, 216, 145, 222, 14, 32, 60, 0, 14, 32, 3, 216, 146, 222, 14, 32, 60, 0, 14, 32, 3, 216, 147, 222, 14, 32, 60, 0, 14, 32, 3, 216, 148, 222, 14, 32, 60, 0, 14, 32, 3, 216, 149, 222, 14, 32, 60, 0, 14, 32, 3, 216, 150, 222, 14, 32, 60, 0, 14, 32, 3, 216, 151, 222, 14, 32, 60, 0, 14, 32, 3, 216, 152, 222, 14, 32, 60, 0, 14, 32, 3, 216, 153, 222, 14, 32, 60, 0, 14, 32, 3, 216, 154, 222, 14, 32, 60, 0, 14, 32, 3, 216, 155, 222, 14, 32, 60, 0, 14, 32, 3, 216, 156, 222, 14, 32, 60, 0, 14, 32, 3, 216, 157, 222, 14, 32, 60, 0, 14, 32, 3, 216, 158, 222, 14, 32, 60, 0, 14, 32, 3, 216, 159, 222, 14, 32, 60, 0, 14, 32, 3, 216, 160, 222, 14, 32, 60, 0, 14, 32, 3, 216, 176, 222, 14, 32, 60, 0, 14, 32, 3, 216, 161, 222, 14, 32, 60, 0, 14, 32, 3, 216, 162, 222, 14, 32, 60, 0, 14, 32, 3, 216, 163, 222, 14, 32, 60, 0, 14, 32, 3, 216, 164, 222, 14, 32, 60, 0, 14, 32, 3, 216, 165, 222, 14, 32, 60, 0, 14, 32, 3, 216, 166, 222, 14, 32, 60, 0, 14, 32, 3, 216, 167, 222, 14, 32, 60, 0, 14, 32, 3, 216, 168, 222, 14, 32, 60, 0, 14, 32, 3, 216, 177, 222, 14, 32, 60, 0, 14, 32, 3, 216, 169, 222, 14, 32, 60, 0, 14, 32, 3, 216, 176, 223, 14, 32, 60, 0, 14, 32, 3, 216, 177, 223, 14, 32, 60, 0, 14, 32, 3, 216, 178, 223, 14, 32, 60, 0, 14, 32, 3, 216, 179, 223, 14, 32, 60, 0, 14, 32, 3, 216, 180, 223, 14, 32, 60, 0, 14, 32, 3, 216, 181, 223, 14, 32, 60, 0, 14, 32, 3, 216, 182, 223, 14, 32, 60, 0, 14, 32, 3, 216, 183, 223, 14, 32, 60, 0, 14, 32, 3, 216, 184, 223, 14, 32, 60, 0, 14, 32, 3, 216, 185, 223, 14, 32, 60, 0, 14, 32, 3, 216, 186, 223, 14, 32, 60, 0, 14, 32, 3, 216, 187, 223, 14, 32, 60, 0, 14, 32, 3, 216, 188, 223, 14, 32, 60, 0, 14, 32, 3, 216, 189, 223, 14, 32, 60, 0, 14, 32, 3, 216, 190, 223, 14, 32, 60, 0, 14, 32, 3, 216, 191, 223, 14, 32, 60, 0, 14, 32, 3, 216, 192, 223, 14, 32, 60, 0, 14, 32, 3, 216, 193, 223, 14, 32, 60, 0, 14, 32, 3, 216, 194, 223, 14, 32, 60, 0, 14, 32, 3, 216, 195, 223, 14, 32, 60, 0, 14, 32, 3, 216, 196, 223, 14, 32, 60, 0, 0, 216, 128, 223, 60, 0, 0, 216, 129, 223, 60, 0, 0, 216, 130, 223, 60, 0, 0, 216, 131, 223, 60, 0, 0, 216, 132, 223, 60, 0, 0, 216, 133, 223, 60, 0, 0, 216, 134, 223, 60, 0, 0, 216, 135, 223, 60, 0, 0, 216, 136, 223, 60, 0, 0, 216, 137, 223, 60, 0, 0, 216, 138, 223, 60, 0, 0, 216, 139, 223, 60, 0, 0, 216, 140, 223, 60, 0, 0, 216, 141, 223, 60, 0, 0, 216, 142, 223, 60, 0, 0, 216, 143, 223, 60, 0, 0, 216, 144, 223, 60, 0, 0, 216, 145, 223, 60, 0, 0, 216, 146, 223, 60, 0, 0, 216, 147, 223, 60, 0, 0, 216, 148, 223, 60, 0, 0, 216, 149, 223, 60, 0, 0, 216, 150, 223, 60, 0, 0, 216, 151, 223, 60, 0, 0, 216, 152, 223, 60, 0, 0, 216, 153, 223, 60, 0, 0, 216, 154, 223, 60, 0, 0, 216, 155, 223, 60, 0, 0, 216, 156, 223, 60, 0, 0, 216, 157, 223, 60, 0, 0, 216, 160, 223, 60, 0, 0, 216, 161, 223, 60, 0, 0, 216, 162, 223, 60, 0, 0, 216, 163, 223, 60, 0, 0, 216, 164, 223, 60, 0, 0, 216, 165, 223, 60, 0, 0, 216, 166, 223, 60, 0, 0, 216, 167, 223, 60, 0, 0, 216, 168, 223, 60, 0, 0, 216, 169, 223, 60, 0, 0, 216, 170, 223, 60, 0, 0, 216, 171, 223, 60, 0, 0, 216, 172, 223, 60, 0, 0, 216, 173, 223, 60, 0, 0, 216, 174, 223, 60, 0, 0, 216, 175, 223, 60, 0, 0, 216, 176, 223, 60, 0, 0, 216, 177, 223, 60, 0, 0, 216, 178, 223, 60, 0, 0, 216, 179, 223, 60, 0, 0, 216, 180, 223, 60, 0, 0, 216, 181, 223, 60, 0, 0, 216, 182, 223, 60, 0, 0, 216, 183, 223, 60, 0, 0, 216, 184, 223, 60, 0, 0, 216, 185, 223, 60, 0, 0, 216, 186, 223, 60, 0, 0, 216, 187, 223, 60, 0, 0, 216, 188, 223, 60, 0, 0, 216, 189, 223, 60, 0, 0, 216, 190, 223, 60, 0, 0, 216, 191, 223, 60, 0, 0, 216, 192, 223, 60, 0, 0, 216, 193, 223, 60, 0, 0, 216, 194, 223, 60, 0, 0, 216, 195, 223, 60, 0, 0, 216, 200, 223, 60, 0, 0, 216, 201, 223, 60, 0, 0, 216, 202, 223, 60, 0, 0, 216, 203, 223, 60, 0, 0, 216, 204, 223, 60, 0, 0, 216, 205, 223, 60, 0, 0, 216, 206, 223, 60, 0, 0, 216, 207, 223, 60, 0, 8, 216, 0, 220, 60, 0, 8, 216, 1, 220, 60, 0, 8, 216, 2, 220, 60, 0, 8, 216, 3, 220, 60, 0, 8, 216, 4, 220, 60, 0, 8, 216, 5, 220, 60, 0, 8, 216, 6, 220, 60, 0, 8, 216, 7, 220, 60, 0, 8, 216, 8, 220, 60, 0, 8, 216, 9, 220, 60, 0, 8, 216, 10, 220, 60, 0, 8, 216, 11, 220, 60, 0, 8, 216, 12, 220, 60, 0, 8, 216, 13, 220, 60, 0, 8, 216, 14, 220, 60, 0, 8, 216, 15, 220, 60, 0, 8, 216, 16, 220, 60, 0, 8, 216, 17, 220, 60, 0, 8, 216, 18, 220, 60, 0, 8, 216, 19, 220, 60, 0, 8, 216, 20, 220, 60, 0, 8, 216, 21, 220, 60, 0, 8, 216, 22, 220, 60, 0, 8, 216, 23, 220, 60, 0, 8, 216, 24, 220, 60, 0, 8, 216, 25, 220, 60, 0, 8, 216, 26, 220, 60, 0, 8, 216, 27, 220, 60, 0, 8, 216, 28, 220, 60, 0, 8, 216, 29, 220, 60, 0, 8, 216, 30, 220, 60, 0, 8, 216, 31, 220, 60, 0, 8, 216, 32, 220, 60, 0, 8, 216, 33, 220, 60, 0, 8, 216, 34, 220, 60, 0, 8, 216, 35, 220, 60, 0, 8, 216, 36, 220, 60, 0, 8, 216, 37, 220, 60, 0, 8, 216, 38, 220, 60, 0, 8, 216, 39, 220, 60, 0, 8, 216, 40, 220, 60, 0, 8, 216, 41, 220, 60, 0, 8, 216, 42, 220, 60, 0, 8, 216, 43, 220, 60, 0, 8, 216, 44, 220, 60, 0, 8, 216, 45, 220, 60, 0, 8, 216, 46, 220, 60, 0, 8, 216, 47, 220, 60, 0, 8, 216, 48, 220, 60, 0, 8, 216, 49, 220, 60, 0, 8, 216, 50, 220, 60, 0, 8, 216, 51, 220, 60, 0, 8, 216, 52, 220, 60, 0, 8, 216, 53, 220, 60, 0, 8, 216, 54, 220, 60, 0, 8, 216, 55, 220, 60, 0, 8, 216, 56, 220, 60, 0, 8, 216, 57, 220, 60, 0, 8, 216, 58, 220, 60, 0, 8, 216, 59, 220, 60, 0, 8, 216, 60, 220, 60, 0, 8, 216, 61, 220, 60, 0, 8, 216, 62, 220, 60, 0, 8, 216, 63, 220, 60, 0, 8, 216, 64, 220, 60, 0, 8, 216, 65, 220, 60, 0, 8, 216, 66, 220, 60, 0, 8, 216, 67, 220, 60, 0, 8, 216, 68, 220, 60, 0, 8, 216, 69, 220, 60, 0, 8, 216, 70, 220, 60, 0, 8, 216, 71, 220, 60, 0, 8, 216, 72, 220, 60, 0, 8, 216, 73, 220, 60, 0, 8, 216, 74, 220, 60, 0, 8, 216, 75, 220, 60, 0, 8, 216, 76, 220, 60, 0, 8, 216, 77, 220, 60, 0, 8, 216, 78, 220, 60, 0, 8, 216, 79, 220, 60, 0, 8, 216, 80, 220, 60, 0, 8, 216, 81, 220, 60, 0, 8, 216, 82, 220, 60, 0, 8, 216, 83, 220, 60, 0, 8, 216, 84, 220, 60, 0, 8, 216, 85, 220, 60, 0, 8, 216, 86, 220, 60, 0, 8, 216, 87, 220, 60, 0, 8, 216, 88, 220, 60, 0, 8, 216, 89, 220, 60, 0, 8, 216, 90, 220, 60, 0, 8, 216, 91, 220, 60, 0, 8, 216, 92, 220, 60, 0, 8, 216, 93, 220, 60, 0, 8, 216, 94, 220, 60, 0, 8, 216, 95, 220, 60, 0, 8, 216, 96, 220, 60, 0, 8, 216, 97, 220, 60, 0, 8, 216, 98, 220, 60, 0, 8, 216, 99, 220, 60, 0, 8, 216, 100, 220, 60, 0, 8, 216, 101, 220, 60, 0, 8, 216, 102, 220, 60, 0, 8, 216, 103, 220, 60, 0, 8, 216, 104, 220, 60, 0, 8, 216, 105, 220, 60, 0, 8, 216, 106, 220, 60, 0, 8, 216, 107, 220, 60, 0, 8, 216, 108, 220, 60, 0, 8, 216, 109, 220, 60, 0, 8, 216, 110, 220, 60, 0, 8, 216, 111, 220, 60, 0, 8, 216, 112, 220, 60, 0, 8, 216, 113, 220, 60, 0, 8, 216, 114, 220, 60, 0, 8, 216, 115, 220, 60, 0, 8, 216, 116, 220, 60, 0, 8, 216, 117, 220, 60, 0, 8, 216, 118, 220, 60, 0, 8, 216, 119, 220, 60, 0, 8, 216, 120, 220, 60, 0, 8, 216, 121, 220, 60, 0, 8, 216, 122, 220, 60, 0, 8, 216, 123, 220, 60, 0, 8, 216, 124, 220, 60, 0, 8, 216, 125, 220, 60, 0, 8, 216, 126, 220, 60, 0, 8, 216, 127, 220, 60, 0, 8, 216, 128, 220, 60, 0, 8, 216, 129, 220, 60, 0, 8, 216, 130, 220, 60, 0, 8, 216, 131, 220, 60, 0, 8, 216, 132, 220, 60, 0, 8, 216, 133, 220, 60, 0, 8, 216, 134, 220, 60, 0, 8, 216, 135, 220, 60, 0, 8, 216, 136, 220, 60, 0, 8, 216, 137, 220, 60, 0, 8, 216, 138, 220, 60, 0, 8, 216, 139, 220, 60, 0, 8, 216, 140, 220, 60, 0, 8, 216, 141, 220, 60, 0, 8, 216, 142, 220, 60, 0, 8, 216, 143, 220, 60, 0, 8, 216, 144, 220, 60, 0, 8, 216, 145, 220, 60, 0, 8, 216, 146, 220, 60, 0, 8, 216, 147, 220, 60, 0, 8, 216, 148, 220, 60, 0, 8, 216, 149, 220, 60, 0, 8, 216, 150, 220, 60, 0, 8, 216, 151, 220, 60, 0, 8, 216, 152, 220, 60, 0, 8, 216, 153, 220, 60, 0, 8, 216, 154, 220, 60, 0, 8, 216, 155, 220, 60, 0, 8, 216, 156, 220, 60, 0, 8, 216, 157, 220, 60, 0, 8, 216, 158, 220, 60, 0, 8, 216, 159, 220, 60, 0, 8, 216, 160, 220, 60, 0, 8, 216, 161, 220, 60, 0, 8, 216, 162, 220, 60, 0, 8, 216, 163, 220, 60, 0, 8, 216, 164, 220, 60, 0, 8, 216, 165, 220, 60, 0, 8, 216, 166, 220, 60, 0, 8, 216, 167, 220, 60, 0, 8, 216, 168, 220, 60, 0, 8, 216, 169, 220, 60, 0, 8, 216, 170, 220, 60, 0, 8, 216, 171, 220, 60, 0, 8, 216, 172, 220, 60, 0, 8, 216, 173, 220, 60, 0, 8, 216, 174, 220, 60, 0, 8, 216, 175, 220, 60, 0, 8, 216, 176, 220, 60, 0, 8, 216, 177, 220, 60, 0, 8, 216, 178, 220, 60, 0, 8, 216, 179, 220, 60, 0, 8, 216, 180, 220, 60, 0, 8, 216, 181, 220, 60, 0, 8, 216, 182, 220, 60, 0, 8, 216, 183, 220, 60, 0, 8, 216, 184, 220, 60, 0, 8, 216, 185, 220, 60, 0, 8, 216, 186, 220, 60, 0, 8, 216, 187, 220, 60, 0, 8, 216, 188, 220, 60, 0, 8, 216, 189, 220, 60, 0, 8, 216, 190, 220, 60, 0, 8, 216, 191, 220, 60, 0, 8, 216, 192, 220, 60, 0, 8, 216, 193, 220, 60, 0, 8, 216, 194, 220, 60, 0, 8, 216, 195, 220, 60, 0, 8, 216, 196, 220, 60, 0, 8, 216, 197, 220, 60, 0, 8, 216, 198, 220, 60, 0, 8, 216, 199, 220, 60, 0, 8, 216, 200, 220, 60, 0, 8, 216, 201, 220, 60, 0, 8, 216, 202, 220, 60, 0, 8, 216, 203, 220, 60, 0, 8, 216, 204, 220, 60, 0, 8, 216, 205, 220, 60, 0, 8, 216, 206, 220, 60, 0, 8, 216, 207, 220, 60, 0, 8, 216, 208, 220, 60, 0, 8, 216, 209, 220, 60, 0, 8, 216, 210, 220, 60, 0, 8, 216, 211, 220, 60, 0, 8, 216, 212, 220, 60, 0, 8, 216, 213, 220, 60, 0, 8, 216, 214, 220, 60, 0, 8, 216, 215, 220, 60, 0, 8, 216, 216, 220, 60, 0, 8, 216, 217, 220, 60, 0, 8, 216, 218, 220, 60, 0, 8, 216, 219, 220, 60, 0, 8, 216, 220, 220, 60, 0, 8, 216, 221, 220, 60, 0, 8, 216, 222, 220, 60, 0, 8, 216, 223, 220, 60, 0, 8, 216, 224, 220, 60, 0, 8, 216, 225, 220, 60, 0, 8, 216, 226, 220, 60, 0, 8, 216, 227, 220, 60, 0, 8, 216, 228, 220, 60, 0, 8, 216, 229, 220, 60, 0, 8, 216, 230, 220, 60, 0, 8, 216, 231, 220, 60, 0, 8, 216, 232, 220, 60, 0, 8, 216, 233, 220, 60, 0, 8, 216, 234, 220, 60, 0, 8, 216, 235, 220, 60, 0, 8, 216, 236, 220, 60, 0, 8, 216, 237, 220, 60, 0, 8, 216, 238, 220, 60, 0, 8, 216, 239, 220, 60, 0, 8, 216, 240, 220, 60, 0, 8, 216, 241, 220, 60, 0, 8, 216, 242, 220, 60, 0, 8, 216, 243, 220, 60, 0, 8, 216, 244, 220, 60, 0, 8, 216, 245, 220, 60, 0, 8, 216, 246, 220, 60, 0, 8, 216, 247, 220, 60, 0, 8, 216, 248, 220, 60, 0, 8, 216, 249, 220, 60, 0, 8, 216, 250, 220, 60, 0, 8, 216, 251, 220, 60, 0, 8, 216, 252, 220, 60, 0, 8, 216, 253, 220, 60, 0, 8, 216, 254, 220, 60, 0, 8, 216, 255, 220, 60, 0, 8, 216, 0, 221, 60, 0, 8, 216, 1, 221, 60, 0, 8, 216, 2, 221, 60, 0, 8, 216, 3, 221, 60, 0, 8, 216, 4, 221, 60, 0, 8, 216, 5, 221, 60, 0, 8, 216, 6, 221, 60, 0, 8, 216, 7, 221, 60, 0, 8, 216, 8, 221, 60, 0, 8, 216, 9, 221, 60, 0, 8, 216, 10, 221, 60, 0, 8, 216, 11, 221, 60, 0, 8, 216, 12, 221, 60, 0, 8, 216, 13, 221, 60, 0, 8, 216, 14, 221, 60, 0, 8, 216, 15, 221, 60, 0, 8, 216, 16, 221, 60, 0, 8, 216, 17, 221, 60, 0, 8, 216, 18, 221, 60, 0, 8, 216, 19, 221, 60, 0, 8, 216, 20, 221, 60, 0, 8, 216, 21, 221, 60, 0, 8, 216, 22, 221, 60, 0, 8, 216, 23, 221, 60, 0, 8, 216, 24, 221, 60, 0, 8, 216, 25, 221, 60, 0, 8, 216, 26, 221, 60, 0, 8, 216, 27, 221, 60, 0, 8, 216, 28, 221, 60, 0, 8, 216, 29, 221, 60, 0, 8, 216, 30, 221, 60, 0, 8, 216, 31, 221, 60, 0, 8, 216, 32, 221, 60, 0, 8, 216, 33, 221, 60, 0, 8, 216, 34, 221, 60, 0, 8, 216, 35, 221, 60, 0, 8, 216, 36, 221, 60, 0, 8, 216, 37, 221, 60, 0, 8, 216, 38, 221, 60, 0, 8, 216, 39, 221, 60, 0, 8, 216, 40, 221, 60, 0, 8, 216, 41, 221, 60, 0, 8, 216, 42, 221, 60, 0, 8, 216, 43, 221, 60, 0, 8, 216, 44, 221, 60, 0, 8, 216, 45, 221, 60, 0, 8, 216, 46, 221, 60, 0, 8, 216, 47, 221, 60, 0, 8, 216, 48, 221, 60, 0, 8, 216, 49, 221, 60, 0, 8, 216, 50, 221, 60, 0, 8, 216, 51, 221, 60, 0, 8, 216, 52, 221, 60, 0, 8, 216, 53, 221, 60, 0, 8, 216, 54, 221, 60, 0, 8, 216, 55, 221, 60, 0, 8, 216, 56, 221, 60, 0, 8, 216, 57, 221, 60, 0, 8, 216, 58, 221, 60, 0, 8, 216, 59, 221, 60, 0, 8, 216, 60, 221, 60, 0, 8, 216, 61, 221, 60, 0, 8, 216, 62, 221, 60, 0, 8, 216, 63, 221, 60, 0, 8, 216, 64, 221, 60, 0, 8, 216, 65, 221, 60, 0, 8, 216, 66, 221, 60, 0, 8, 216, 67, 221, 60, 0, 8, 216, 68, 221, 60, 0, 8, 216, 69, 221, 60, 0, 8, 216, 70, 221, 60, 0, 8, 216, 71, 221, 60, 0, 8, 216, 72, 221, 60, 0, 8, 216, 73, 221, 60, 0, 8, 216, 74, 221, 60, 0, 8, 216, 75, 221, 60, 0, 8, 216, 76, 221, 60, 0, 8, 216, 77, 221, 60, 0, 8, 216, 78, 221, 60, 0, 8, 216, 79, 221, 60, 0, 8, 216, 80, 221, 60, 0, 8, 216, 81, 221, 60, 0, 8, 216, 82, 221, 60, 0, 8, 216, 83, 221, 60, 0, 8, 216, 84, 221, 60, 0, 8, 216, 85, 221, 60, 0, 8, 216, 86, 221, 60, 0, 8, 216, 87, 221, 60, 0, 8, 216, 88, 221, 60, 0, 8, 216, 89, 221, 60, 0, 8, 216, 90, 221, 60, 0, 8, 216, 91, 221, 60, 0, 8, 216, 92, 221, 60, 0, 8, 216, 93, 221, 60, 0, 8, 216, 94, 221, 60, 0, 8, 216, 95, 221, 60, 0, 8, 216, 96, 221, 60, 0, 8, 216, 97, 221, 60, 0, 8, 216, 98, 221, 60, 0, 8, 216, 99, 221, 60, 0, 8, 216, 100, 221, 60, 0, 8, 216, 101, 221, 60, 0, 8, 216, 102, 221, 60, 0, 8, 216, 103, 221, 60, 0, 8, 216, 104, 221, 60, 0, 8, 216, 105, 221, 60, 0, 8, 216, 106, 221, 60, 0, 8, 216, 107, 221, 60, 0, 8, 216, 108, 221, 60, 0, 8, 216, 109, 221, 60, 0, 8, 216, 110, 221, 60, 0, 8, 216, 111, 221, 60, 0, 8, 216, 112, 221, 60, 0, 8, 216, 113, 221, 60, 0, 8, 216, 114, 221, 60, 0, 8, 216, 115, 221, 60, 0, 8, 216, 116, 221, 60, 0, 8, 216, 117, 221, 60, 0, 8, 216, 118, 221, 60, 0, 8, 216, 119, 221, 60, 0, 8, 216, 120, 221, 60, 0, 8, 216, 121, 221, 60, 0, 8, 216, 122, 221, 60, 0, 8, 216, 123, 221, 60, 0, 8, 216, 124, 221, 60, 0, 8, 216, 125, 221, 60, 0, 8, 216, 126, 221, 60, 0, 8, 216, 127, 221, 60, 0, 8, 216, 128, 221, 60, 0, 8, 216, 129, 221, 60, 0, 8, 216, 130, 221, 60, 0, 8, 216, 131, 221, 60, 0, 8, 216, 132, 221, 60, 0, 8, 216, 133, 221, 60, 0, 8, 216, 134, 221, 60, 0, 8, 216, 135, 221, 60, 0, 8, 216, 136, 221, 60, 0, 8, 216, 137, 221, 60, 0, 8, 216, 138, 221, 60, 0, 8, 216, 139, 221, 60, 0, 8, 216, 140, 221, 60, 0, 8, 216, 141, 221, 60, 0, 8, 216, 142, 221, 60, 0, 8, 216, 143, 221, 60, 0, 8, 216, 144, 221, 60, 0, 8, 216, 145, 221, 60, 0, 8, 216, 146, 221, 60, 0, 8, 216, 147, 221, 60, 0, 8, 216, 148, 221, 60, 0, 8, 216, 149, 221, 60, 0, 8, 216, 150, 221, 60, 0, 8, 216, 151, 221, 60, 0, 8, 216, 152, 221, 60, 0, 8, 216, 153, 221, 60, 0, 8, 216, 154, 221, 60, 0, 8, 216, 155, 221, 60, 0, 8, 216, 156, 221, 60, 0, 8, 216, 157, 221, 60, 0, 8, 216, 158, 221, 60, 0, 8, 216, 159, 221, 60, 0, 8, 216, 160, 221, 60, 0, 8, 216, 161, 221, 60, 0, 8, 216, 162, 221, 60, 0, 8, 216, 163, 221, 60, 0, 8, 216, 164, 221, 60, 0, 8, 216, 165, 221, 60, 0, 8, 216, 166, 221, 60, 0, 8, 216, 167, 221, 60, 0, 8, 216, 168, 221, 60, 0, 8, 216, 169, 221, 60, 0, 8, 216, 170, 221, 60, 0, 8, 216, 171, 221, 60, 0, 8, 216, 172, 221, 60, 0, 8, 216, 173, 221, 60, 0, 8, 216, 174, 221, 60, 0, 8, 216, 175, 221, 60, 0, 8, 216, 176, 221, 60, 0, 8, 216, 177, 221, 60, 0, 8, 216, 178, 221, 60, 0, 8, 216, 179, 221, 60, 0, 8, 216, 180, 221, 60, 0, 8, 216, 181, 221, 60, 0, 8, 216, 182, 221, 60, 0, 8, 216, 183, 221, 60, 0, 8, 216, 184, 221, 60, 0, 8, 216, 185, 221, 60, 0, 8, 216, 186, 221, 60, 0, 8, 216, 187, 221, 60, 0, 8, 216, 188, 221, 60, 0, 8, 216, 189, 221, 60, 0, 8, 216, 190, 221, 60, 0, 8, 216, 191, 221, 60, 0, 8, 216, 192, 221, 60, 0, 8, 216, 193, 221, 60, 0, 8, 216, 194, 221, 60, 0, 8, 216, 195, 221, 60, 0, 8, 216, 196, 221, 60, 0, 8, 216, 197, 221, 60, 0, 8, 216, 198, 221, 60, 0, 8, 216, 199, 221, 60, 0, 8, 216, 200, 221, 60, 0, 8, 216, 201, 221, 60, 0, 8, 216, 202, 221, 60, 0, 8, 216, 203, 221, 60, 0, 8, 216, 204, 221, 60, 0, 8, 216, 205, 221, 60, 0, 8, 216, 206, 221, 60, 0, 8, 216, 207, 221, 60, 0, 8, 216, 208, 221, 60, 0, 8, 216, 209, 221, 60, 0, 8, 216, 210, 221, 60, 0, 8, 216, 211, 221, 60, 0, 8, 216, 212, 221, 60, 0, 8, 216, 213, 221, 60, 0, 8, 216, 214, 221, 60, 0, 8, 216, 215, 221, 60, 0, 8, 216, 216, 221, 60, 0, 8, 216, 217, 221, 60, 0, 8, 216, 218, 221, 60, 0, 8, 216, 219, 221, 60, 0, 8, 216, 220, 221, 60, 0, 8, 216, 221, 221, 60, 0, 8, 216, 222, 221, 60, 0, 8, 216, 223, 221, 60, 0, 8, 216, 224, 221, 60, 0, 8, 216, 225, 221, 60, 0, 8, 216, 226, 221, 60, 0, 8, 216, 227, 221, 60, 0, 8, 216, 228, 221, 60, 0, 8, 216, 229, 221, 60, 0, 8, 216, 230, 221, 60, 0, 8, 216, 231, 221, 60, 0, 8, 216, 232, 221, 60, 0, 8, 216, 233, 221, 60, 0, 8, 216, 234, 221, 60, 0, 8, 216, 235, 221, 60, 0, 8, 216, 236, 221, 60, 0, 8, 216, 237, 221, 60, 0, 8, 216, 238, 221, 60, 0, 8, 216, 239, 221, 60, 0, 8, 216, 240, 221, 60, 0, 8, 216, 241, 221, 60, 0, 8, 216, 242, 221, 60, 0, 8, 216, 243, 221, 60, 0, 8, 216, 244, 221, 60, 0, 8, 216, 245, 221, 60, 0, 8, 216, 246, 221, 60, 0, 8, 216, 247, 221, 60, 0, 8, 216, 248, 221, 60, 0, 8, 216, 249, 221, 60, 0, 8, 216, 250, 221, 60, 0, 8, 216, 251, 221, 60, 0, 8, 216, 252, 221, 60, 0, 8, 216, 253, 221, 60, 0, 8, 216, 254, 221, 60, 0, 8, 216, 255, 221, 60, 0, 8, 216, 0, 222, 60, 0, 8, 216, 1, 222, 60, 0, 8, 216, 2, 222, 60, 0, 8, 216, 3, 222, 60, 0, 8, 216, 4, 222, 60, 0, 8, 216, 5, 222, 60, 0, 8, 216, 6, 222, 60, 0, 8, 216, 7, 222, 60, 0, 8, 216, 8, 222, 60, 0, 8, 216, 9, 222, 60, 0, 8, 216, 10, 222, 60, 0, 8, 216, 11, 222, 60, 0, 8, 216, 12, 222, 60, 0, 8, 216, 13, 222, 60, 0, 8, 216, 14, 222, 60, 0, 8, 216, 15, 222, 60, 0, 8, 216, 16, 222, 60, 0, 8, 216, 17, 222, 60, 0, 8, 216, 18, 222, 60, 0, 8, 216, 19, 222, 60, 0, 8, 216, 20, 222, 60, 0, 8, 216, 21, 222, 60, 0, 8, 216, 22, 222, 60, 0, 8, 216, 23, 222, 60, 0, 8, 216, 24, 222, 60, 0, 8, 216, 25, 222, 60, 0, 8, 216, 26, 222, 60, 0, 8, 216, 27, 222, 60, 0, 8, 216, 28, 222, 60, 0, 8, 216, 29, 222, 60, 0, 8, 216, 30, 222, 60, 0, 8, 216, 31, 222, 60, 0, 8, 216, 32, 222, 60, 0, 8, 216, 33, 222, 60, 0, 8, 216, 34, 222, 60, 0, 8, 216, 35, 222, 60, 0, 8, 216, 36, 222, 60, 0, 8, 216, 37, 222, 60, 0, 8, 216, 38, 222, 60, 0, 8, 216, 39, 222, 60, 0, 8, 216, 40, 222, 60, 0, 8, 216, 41, 222, 60, 0, 8, 216, 42, 222, 60, 0, 8, 216, 43, 222, 60, 0, 8, 216, 44, 222, 60, 0, 8, 216, 45, 222, 60, 0, 8, 216, 46, 222, 60, 0, 8, 216, 47, 222, 60, 0, 8, 216, 48, 222, 60, 0, 8, 216, 49, 222, 60, 0, 8, 216, 50, 222, 60, 0, 8, 216, 51, 222, 60, 0, 8, 216, 52, 222, 60, 0, 8, 216, 53, 222, 60, 0, 8, 216, 54, 222, 60, 0, 8, 216, 55, 222, 60, 0, 8, 216, 56, 222, 60, 0, 8, 216, 57, 222, 60, 0, 8, 216, 58, 222, 60, 0, 8, 216, 59, 222, 60, 0, 8, 216, 60, 222, 60, 0, 8, 216, 61, 222, 60, 0, 8, 216, 62, 222, 60, 0, 8, 216, 63, 222, 60, 0, 8, 216, 64, 222, 60, 0, 8, 216, 65, 222, 60, 0, 8, 216, 66, 222, 60, 0, 8, 216, 67, 222, 60, 0, 8, 216, 68, 222, 60, 0, 8, 216, 69, 222, 60, 0, 8, 216, 70, 222, 60, 0, 8, 216, 71, 222, 60, 0, 8, 216, 72, 222, 60, 0, 8, 216, 73, 222, 60, 0, 8, 216, 74, 222, 60, 0, 8, 216, 75, 222, 60, 0, 8, 216, 76, 222, 60, 0, 8, 216, 77, 222, 60, 0, 8, 216, 78, 222, 60, 0, 8, 216, 79, 222, 60, 0, 8, 216, 80, 222, 60, 0, 8, 216, 81, 222, 60, 0, 8, 216, 82, 222, 60, 0, 8, 216, 83, 222, 60, 0, 8, 216, 84, 222, 60, 0, 8, 216, 85, 222, 60, 0, 8, 216, 86, 222, 60, 0, 8, 216, 87, 222, 60, 0, 8, 216, 88, 222, 60, 0, 8, 216, 89, 222, 60, 0, 8, 216, 90, 222, 60, 0, 8, 216, 91, 222, 60, 0, 8, 216, 92, 222, 60, 0, 8, 216, 93, 222, 60, 0, 8, 216, 94, 222, 60, 0, 8, 216, 95, 222, 60, 0, 8, 216, 96, 222, 60, 0, 8, 216, 97, 222, 60, 0, 8, 216, 98, 222, 60, 0, 8, 216, 212, 222, 60, 0, 8, 216, 213, 222, 60, 0, 8, 216, 99, 222, 60, 0, 8, 216, 100, 222, 60, 0, 8, 216, 101, 222, 60, 0, 8, 216, 102, 222, 60, 0, 8, 216, 103, 222, 60, 0, 8, 216, 104, 222, 60, 0, 8, 216, 105, 222, 60, 0, 8, 216, 106, 222, 60, 0, 8, 216, 107, 222, 60, 0, 8, 216, 108, 222, 60, 0, 8, 216, 109, 222, 60, 0, 8, 216, 110, 222, 60, 0, 8, 216, 111, 222, 60, 0, 8, 216, 112, 222, 60, 0, 8, 216, 113, 222, 60, 0, 8, 216, 114, 222, 60, 0, 8, 216, 115, 222, 60, 0, 8, 216, 116, 222, 60, 0, 8, 216, 117, 222, 60, 0, 8, 216, 118, 222, 60, 0, 8, 216, 119, 222, 60, 0, 8, 216, 120, 222, 60, 0, 8, 216, 121, 222, 60, 0, 8, 216, 122, 222, 60, 0, 8, 216, 123, 222, 60, 0, 8, 216, 124, 222, 60, 0, 8, 216, 125, 222, 60, 0, 8, 216, 126, 222, 60, 0, 8, 216, 127, 222, 60, 0, 8, 216, 128, 222, 60, 0, 8, 216, 129, 222, 60, 0, 8, 216, 130, 222, 60, 0, 8, 216, 131, 222, 60, 0, 8, 216, 132, 222, 60, 0, 8, 216, 133, 222, 60, 0, 8, 216, 134, 222, 60, 0, 8, 216, 135, 222, 60, 0, 8, 216, 136, 222, 60, 0, 8, 216, 137, 222, 60, 0, 8, 216, 138, 222, 60, 0, 8, 216, 139, 222, 60, 0, 8, 216, 140, 222, 60, 0, 8, 216, 141, 222, 60, 0, 8, 216, 142, 222, 60, 0, 8, 216, 143, 222, 60, 0, 8, 216, 144, 222, 60, 0, 8, 216, 145, 222, 60, 0, 8, 216, 146, 222, 60, 0, 8, 216, 147, 222, 60, 0, 8, 216, 148, 222, 60, 0, 8, 216, 149, 222, 60, 0, 8, 216, 150, 222, 60, 0, 8, 216, 151, 222, 60, 0, 8, 216, 152, 222, 60, 0, 8, 216, 153, 222, 60, 0, 8, 216, 154, 222, 60, 0, 8, 216, 155, 222, 60, 0, 8, 216, 156, 222, 60, 0, 8, 216, 157, 222, 60, 0, 8, 216, 158, 222, 60, 0, 8, 216, 159, 222, 60, 0, 8, 216, 160, 222, 60, 0, 8, 216, 161, 222, 60, 0, 8, 216, 162, 222, 60, 0, 8, 216, 163, 222, 60, 0, 8, 216, 164, 222, 60, 0, 8, 216, 165, 222, 60, 0, 8, 216, 166, 222, 60, 0, 8, 216, 167, 222, 60, 0, 8, 216, 168, 222, 60, 0, 8, 216, 169, 222, 60, 0, 8, 216, 170, 222, 60, 0, 8, 216, 171, 222, 60, 0, 8, 216, 172, 222, 60, 0, 8, 216, 173, 222, 60, 0, 8, 216, 174, 222, 60, 0, 8, 216, 175, 222, 60, 0, 8, 216, 176, 222, 60, 0, 8, 216, 177, 222, 60, 0, 8, 216, 178, 222, 60, 0, 8, 216, 179, 222, 60, 0, 8, 216, 180, 222, 60, 0, 8, 216, 181, 222, 60, 0, 8, 216, 182, 222, 60, 0, 8, 216, 183, 222, 60, 0, 8, 216, 184, 222, 60, 0, 8, 216, 185, 222, 60, 0, 8, 216, 186, 222, 60, 0, 8, 216, 187, 222, 60, 0, 8, 216, 188, 222, 60, 0, 8, 216, 189, 222, 60, 0, 8, 216, 190, 222, 60, 0, 8, 216, 191, 222, 60, 0, 8, 216, 192, 222, 60, 0, 8, 216, 193, 222, 60, 0, 8, 216, 194, 222, 60, 0, 8, 216, 195, 222, 60, 0, 8, 216, 196, 222, 60, 0, 8, 216, 197, 222, 60, 0, 8, 216, 198, 222, 60, 0, 8, 216, 199, 222, 60, 0, 8, 216, 200, 222, 60, 0, 8, 216, 201, 222, 60, 0, 8, 216, 202, 222, 60, 0, 8, 216, 203, 222, 60, 0, 8, 216, 204, 222, 60, 0, 8, 216, 205, 222, 60, 0, 8, 216, 206, 222, 60, 0, 8, 216, 207, 222, 60, 0, 8, 216, 208, 222, 60, 0, 8, 216, 209, 222, 60, 0, 8, 216, 210, 222, 60, 0, 8, 216, 211, 222, 60, 0, 8, 216, 214, 222, 60, 0, 8, 216, 215, 222, 60, 0, 8, 216, 216, 222, 60, 0, 8, 216, 217, 222, 60, 0, 8, 216, 218, 222, 60, 0, 8, 216, 219, 222, 60, 0, 8, 216, 220, 222, 60, 0, 8, 216, 221, 222, 60, 0, 8, 216, 222, 222, 60, 0, 8, 216, 223, 222, 60, 0, 8, 216, 224, 222, 60, 0, 8, 216, 225, 222, 60, 0, 8, 216, 226, 222, 60, 0, 8, 216, 227, 222, 60, 0, 8, 216, 228, 222, 60, 0, 8, 216, 229, 222, 60, 0, 8, 216, 230, 222, 60, 0, 8, 216, 231, 222, 60, 0, 8, 216, 232, 222, 60, 0, 8, 216, 233, 222, 60, 0, 8, 216, 234, 222, 60, 0, 8, 216, 235, 222, 60, 0, 8, 216, 236, 222, 60, 0, 8, 216, 237, 222, 60, 0, 8, 216, 238, 222, 60, 0, 8, 216, 239, 222, 60, 0, 8, 216, 240, 222, 60, 0, 8, 216, 241, 222, 60, 0, 8, 216, 242, 222, 60, 0, 8, 216, 243, 222, 60, 0, 8, 216, 244, 222, 60, 0, 8, 216, 245, 222, 60, 0, 8, 216, 246, 222, 60, 0, 8, 216, 247, 222, 60, 0, 8, 216, 248, 222, 60, 0, 8, 216, 249, 222, 60, 0, 8, 216, 250, 222, 60, 0, 8, 216, 251, 222, 60, 0, 8, 216, 252, 222, 60, 0, 8, 216, 253, 222, 60, 0, 8, 216, 254, 222, 60, 0, 8, 216, 255, 222, 60, 0, 8, 216, 0, 223, 60, 0, 8, 216, 1, 223, 60, 0, 8, 216, 2, 223, 60, 0, 8, 216, 3, 223, 60, 0, 8, 216, 4, 223, 60, 0, 8, 216, 5, 223, 60, 0, 8, 216, 6, 223, 60, 0, 8, 216, 7, 223, 60, 0, 8, 216, 8, 223, 60, 0, 8, 216, 9, 223, 60, 0, 8, 216, 10, 223, 60, 0, 8, 216, 11, 223, 60, 0, 8, 216, 12, 223, 60, 0, 8, 216, 13, 223, 60, 0, 8, 216, 14, 223, 60, 0, 8, 216, 15, 223, 60, 0, 8, 216, 16, 223, 60, 0, 8, 216, 17, 223, 60, 0, 8, 216, 18, 223, 60, 0, 8, 216, 19, 223, 60, 0, 8, 216, 20, 223, 60, 0, 8, 216, 21, 223, 60, 0, 8, 216, 22, 223, 60, 0, 8, 216, 23, 223, 60, 0, 8, 216, 24, 223, 60, 0, 8, 216, 25, 223, 60, 0, 8, 216, 26, 223, 60, 0, 8, 216, 27, 223, 60, 0, 8, 216, 28, 223, 60, 0, 8, 216, 29, 223, 60, 0, 8, 216, 30, 223, 60, 0, 8, 216, 31, 223, 60, 0, 8, 216, 32, 223, 60, 0, 8, 216, 33, 223, 60, 0, 8, 216, 34, 223, 60, 0, 8, 216, 35, 223, 60, 0, 8, 216, 36, 223, 60, 0, 8, 216, 37, 223, 60, 0, 8, 216, 38, 223, 60, 0, 8, 216, 39, 223, 60, 0, 8, 216, 40, 223, 60, 0, 8, 216, 41, 223, 60, 0, 8, 216, 42, 223, 60, 0, 8, 216, 43, 223, 60, 0, 8, 216, 44, 223, 60, 0, 8, 216, 45, 223, 60, 0, 8, 216, 46, 223, 60, 0, 8, 216, 47, 223, 60, 0, 8, 216, 48, 223, 60, 0, 8, 216, 49, 223, 60, 0, 8, 216, 50, 223, 60, 0, 8, 216, 51, 223, 60, 0, 8, 216, 52, 223, 60, 0, 8, 216, 53, 223, 60, 0, 8, 216, 54, 223, 60, 0, 8, 216, 55, 223, 60, 0, 8, 216, 56, 223, 60, 0, 8, 216, 57, 223, 60, 0, 8, 216, 58, 223, 60, 0, 8, 216, 59, 223, 60, 0, 8, 216, 60, 223, 60, 0, 8, 216, 61, 223, 60, 0, 8, 216, 62, 223, 60, 0, 8, 216, 63, 223, 60, 0, 8, 216, 64, 223, 60, 0, 8, 216, 65, 223, 60, 0, 8, 216, 66, 223, 60, 0, 8, 216, 67, 223, 60, 0, 8, 216, 68, 223, 60, 0, 8, 216, 69, 223, 60, 0, 8, 216, 70, 223, 60, 0, 8, 216, 71, 223, 60, 0, 8, 216, 72, 223, 60, 0, 8, 216, 73, 223, 60, 0, 8, 216, 74, 223, 60, 0, 8, 216, 75, 223, 60, 0, 8, 216, 76, 223, 60, 0, 8, 216, 77, 223, 60, 0, 8, 216, 78, 223, 60, 0, 8, 216, 79, 223, 60, 0, 8, 216, 80, 223, 60, 0, 8, 216, 81, 223, 60, 0, 8, 216, 82, 223, 60, 0, 8, 216, 83, 223, 60, 0, 8, 216, 84, 223, 60, 0, 8, 216, 85, 223, 60, 0, 8, 216, 86, 223, 60, 0, 8, 216, 87, 223, 60, 0, 8, 216, 88, 223, 60, 0, 8, 216, 89, 223, 60, 0, 8, 216, 90, 223, 60, 0, 8, 216, 91, 223, 60, 0, 8, 216, 92, 223, 60, 0, 8, 216, 93, 223, 60, 0, 8, 216, 94, 223, 60, 0, 8, 216, 95, 223, 60, 0, 8, 216, 96, 223, 60, 0, 8, 216, 97, 223, 60, 0, 8, 216, 98, 223, 60, 0, 8, 216, 99, 223, 60, 0, 8, 216, 100, 223, 60, 0, 8, 216, 101, 223, 60, 0, 8, 216, 102, 223, 60, 0, 8, 216, 103, 223, 60, 0, 8, 216, 104, 223, 60, 0, 8, 216, 105, 223, 60, 0, 8, 216, 106, 223, 60, 0, 8, 216, 107, 223, 60, 0, 8, 216, 108, 223, 60, 0, 8, 216, 109, 223, 60, 0, 8, 216, 110, 223, 60, 0, 8, 216, 111, 223, 60, 0, 8, 216, 112, 223, 60, 0, 8, 216, 113, 223, 60, 0, 8, 216, 114, 223, 60, 0, 8, 216, 115, 223, 60, 0, 8, 216, 116, 223, 60, 0, 8, 216, 117, 223, 60, 0, 8, 216, 118, 223, 60, 0, 8, 216, 119, 223, 60, 0, 8, 216, 120, 223, 60, 0, 8, 216, 121, 223, 60, 0, 8, 216, 122, 223, 60, 0, 8, 216, 123, 223, 60, 0, 8, 216, 124, 223, 60, 0, 8, 216, 125, 223, 60, 0, 8, 216, 126, 223, 60, 0, 8, 216, 127, 223, 60, 0, 8, 216, 128, 223, 60, 0, 8, 216, 129, 223, 60, 0, 8, 216, 130, 223, 60, 0, 8, 216, 131, 223, 60, 0, 8, 216, 132, 223, 60, 0, 8, 216, 133, 223, 60, 0, 8, 216, 134, 223, 60, 0, 8, 216, 135, 223, 60, 0, 8, 216, 136, 223, 60, 0, 8, 216, 137, 223, 60, 0, 8, 216, 138, 223, 60, 0, 8, 216, 139, 223, 60, 0, 8, 216, 140, 223, 60, 0, 8, 216, 141, 223, 60, 0, 8, 216, 142, 223, 60, 0, 8, 216, 143, 223, 60, 0, 8, 216, 144, 223, 60, 0, 8, 216, 145, 223, 60, 0, 8, 216, 146, 223, 60, 0, 8, 216, 147, 223, 60, 0, 8, 216, 148, 223, 60, 0, 8, 216, 149, 223, 60, 0, 8, 216, 150, 223, 60, 0, 8, 216, 151, 223, 60, 0, 8, 216, 152, 223, 60, 0, 8, 216, 153, 223, 60, 0, 9, 216, 128, 220, 60, 0, 9, 216, 129, 220, 60, 0, 9, 216, 130, 220, 60, 0, 9, 216, 131, 220, 60, 0, 9, 216, 132, 220, 60, 0, 9, 216, 133, 220, 60, 0, 9, 216, 134, 220, 60, 0, 9, 216, 135, 220, 60, 0, 9, 216, 136, 220, 60, 0, 9, 216, 137, 220, 60, 0, 9, 216, 138, 220, 60, 0, 9, 216, 139, 220, 60, 0, 9, 216, 140, 220, 60, 0, 9, 216, 141, 220, 60, 0, 9, 216, 142, 220, 60, 0, 9, 216, 143, 220, 60, 0, 9, 216, 144, 220, 60, 0, 9, 216, 145, 220, 60, 0, 9, 216, 146, 220, 60, 0, 9, 216, 147, 220, 60, 0, 9, 216, 148, 220, 60, 0, 9, 216, 149, 220, 60, 0, 9, 216, 150, 220, 60, 0, 9, 216, 151, 220, 60, 0, 9, 216, 152, 220, 60, 0, 9, 216, 153, 220, 60, 0, 9, 216, 154, 220, 60, 0, 9, 216, 155, 220, 60, 0, 9, 216, 156, 220, 60, 0, 9, 216, 157, 220, 60, 0, 9, 216, 158, 220, 60, 0, 9, 216, 159, 220, 60, 0, 9, 216, 160, 220, 60, 0, 9, 216, 161, 220, 60, 0, 9, 216, 162, 220, 60, 0, 9, 216, 163, 220, 60, 0, 9, 216, 164, 220, 60, 0, 9, 216, 165, 220, 60, 0, 9, 216, 166, 220, 60, 0, 9, 216, 167, 220, 60, 0, 9, 216, 168, 220, 60, 0, 9, 216, 169, 220, 60, 0, 9, 216, 170, 220, 60, 0, 9, 216, 171, 220, 60, 0, 9, 216, 172, 220, 60, 0, 9, 216, 173, 220, 60, 0, 9, 216, 174, 220, 60, 0, 9, 216, 175, 220, 60, 0, 9, 216, 176, 220, 60, 0, 9, 216, 177, 220, 60, 0, 9, 216, 178, 220, 60, 0, 9, 216, 179, 220, 60, 0, 9, 216, 180, 220, 60, 0, 9, 216, 181, 220, 60, 0, 9, 216, 182, 220, 60, 0, 9, 216, 183, 220, 60, 0, 9, 216, 184, 220, 60, 0, 9, 216, 185, 220, 60, 0, 9, 216, 186, 220, 60, 0, 9, 216, 187, 220, 60, 0, 9, 216, 188, 220, 60, 0, 9, 216, 189, 220, 60, 0, 9, 216, 190, 220, 60, 0, 9, 216, 191, 220, 60, 0, 9, 216, 192, 220, 60, 0, 9, 216, 193, 220, 60, 0, 9, 216, 194, 220, 60, 0, 9, 216, 195, 220, 60, 0, 9, 216, 196, 220, 60, 0, 9, 216, 197, 220, 60, 0, 9, 216, 198, 220, 60, 0, 9, 216, 199, 220, 60, 0, 9, 216, 200, 220, 60, 0, 9, 216, 201, 220, 60, 0, 9, 216, 202, 220, 60, 0, 9, 216, 203, 220, 60, 0, 9, 216, 204, 220, 60, 0, 9, 216, 205, 220, 60, 0, 9, 216, 206, 220, 60, 0, 9, 216, 207, 220, 60, 0, 9, 216, 208, 220, 60, 0, 9, 216, 209, 220, 60, 0, 9, 216, 210, 220, 60, 0, 9, 216, 211, 220, 60, 0, 9, 216, 212, 220, 60, 0, 9, 216, 213, 220, 60, 0, 9, 216, 214, 220, 60, 0, 9, 216, 215, 220, 60, 0, 9, 216, 216, 220, 60, 0, 9, 216, 217, 220, 60, 0, 9, 216, 218, 220, 60, 0, 9, 216, 219, 220, 60, 0, 9, 216, 220, 220, 60, 0, 9, 216, 221, 220, 60, 0, 9, 216, 222, 220, 60, 0, 9, 216, 223, 220, 60, 0, 9, 216, 224, 220, 60, 0, 9, 216, 225, 220, 60, 0, 9, 216, 226, 220, 60, 0, 9, 216, 227, 220, 60, 0, 9, 216, 228, 220, 60, 0, 9, 216, 229, 220, 60, 0, 9, 216, 230, 220, 60, 0, 9, 216, 231, 220, 60, 0, 9, 216, 232, 220, 60, 0, 9, 216, 233, 220, 60, 0, 9, 216, 234, 220, 60, 0, 9, 216, 235, 220, 60, 0, 9, 216, 236, 220, 60, 0, 9, 216, 237, 220, 60, 0, 9, 216, 238, 220, 60, 0, 9, 216, 239, 220, 60, 0, 9, 216, 240, 220, 60, 0, 9, 216, 241, 220, 60, 0, 9, 216, 242, 220, 60, 0, 9, 216, 243, 220, 60, 0, 9, 216, 244, 220, 60, 0, 9, 216, 245, 220, 60, 0, 9, 216, 246, 220, 60, 0, 9, 216, 247, 220, 60, 0, 9, 216, 248, 220, 60, 0, 9, 216, 249, 220, 60, 0, 9, 216, 250, 220, 60, 0, 9, 216, 251, 220, 60, 0, 9, 216, 252, 220, 60, 0, 9, 216, 253, 220, 60, 0, 9, 216, 254, 220, 60, 0, 9, 216, 255, 220, 60, 0, 9, 216, 0, 221, 60, 0, 9, 216, 1, 221, 60, 0, 9, 216, 2, 221, 60, 0, 9, 216, 3, 221, 60, 0, 9, 216, 4, 221, 60, 0, 9, 216, 5, 221, 60, 0, 9, 216, 6, 221, 60, 0, 9, 216, 7, 221, 60, 0, 9, 216, 8, 221, 60, 0, 9, 216, 9, 221, 60, 0, 9, 216, 10, 221, 60, 0, 9, 216, 11, 221, 60, 0, 9, 216, 12, 221, 60, 0, 9, 216, 13, 221, 60, 0, 9, 216, 14, 221, 60, 0, 9, 216, 15, 221, 60, 0, 9, 216, 16, 221, 60, 0, 9, 216, 17, 221, 60, 0, 9, 216, 18, 221, 60, 0, 9, 216, 19, 221, 60, 0, 9, 216, 20, 221, 60, 0, 9, 216, 21, 221, 60, 0, 9, 216, 22, 221, 60, 0, 9, 216, 23, 221, 60, 0, 9, 216, 24, 221, 60, 0, 9, 216, 25, 221, 60, 0, 9, 216, 26, 221, 60, 0, 9, 216, 27, 221, 60, 0, 9, 216, 28, 221, 60, 0, 9, 216, 29, 221, 60, 0, 9, 216, 30, 221, 60, 0, 9, 216, 31, 221, 60, 0, 9, 216, 32, 221, 60, 0, 9, 216, 33, 221, 60, 0, 9, 216, 34, 221, 60, 0, 9, 216, 35, 221, 60, 0, 9, 216, 36, 221, 60, 0, 9, 216, 37, 221, 60, 0, 9, 216, 38, 221, 60, 0, 9, 216, 39, 221, 60, 0, 9, 216, 40, 221, 60, 0, 9, 216, 41, 221, 60, 0, 9, 216, 42, 221, 60, 0, 9, 216, 43, 221, 60, 0, 9, 216, 44, 221, 60, 0, 9, 216, 45, 221, 60, 0, 9, 216, 46, 221, 60, 0, 9, 216, 47, 221, 60, 0, 9, 216, 48, 221, 60, 0, 9, 216, 49, 221, 60, 0, 9, 216, 50, 221, 60, 0, 9, 216, 51, 221, 60, 0, 9, 216, 52, 221, 60, 0, 9, 216, 53, 221, 60, 0, 9, 216, 54, 221, 60, 0, 9, 216, 55, 221, 60, 0, 9, 216, 56, 221, 60, 0, 9, 216, 57, 221, 60, 0, 9, 216, 58, 221, 60, 0, 9, 216, 59, 221, 60, 0, 9, 216, 60, 221, 60, 0, 9, 216, 61, 221, 60, 0, 9, 216, 62, 221, 60, 0, 9, 216, 63, 221, 60, 0, 9, 216, 64, 221, 60, 0, 9, 216, 65, 221, 60, 0, 9, 216, 66, 221, 60, 0, 9, 216, 67, 221, 60, 0, 12, 216, 0, 220, 60, 0, 12, 216, 1, 220, 60, 0, 12, 216, 2, 220, 60, 0, 12, 216, 3, 220, 60, 0, 12, 216, 4, 220, 60, 0, 12, 216, 5, 220, 60, 0, 12, 216, 6, 220, 60, 0, 12, 216, 7, 220, 60, 0, 12, 216, 8, 220, 60, 0, 12, 216, 9, 220, 60, 0, 12, 216, 10, 220, 60, 0, 12, 216, 11, 220, 60, 0, 12, 216, 12, 220, 60, 0, 12, 216, 13, 220, 60, 0, 12, 216, 14, 220, 60, 0, 12, 216, 15, 220, 60, 0, 12, 216, 16, 220, 60, 0, 12, 216, 17, 220, 60, 0, 12, 216, 18, 220, 60, 0, 12, 216, 19, 220, 60, 0, 12, 216, 20, 220, 60, 0, 12, 216, 21, 220, 60, 0, 12, 216, 22, 220, 60, 0, 12, 216, 23, 220, 60, 0, 12, 216, 24, 220, 60, 0, 12, 216, 25, 220, 60, 0, 12, 216, 26, 220, 60, 0, 12, 216, 27, 220, 60, 0, 12, 216, 28, 220, 60, 0, 12, 216, 29, 220, 60, 0, 12, 216, 30, 220, 60, 0, 12, 216, 31, 220, 60, 0, 12, 216, 32, 220, 60, 0, 12, 216, 33, 220, 60, 0, 12, 216, 34, 220, 60, 0, 12, 216, 35, 220, 60, 0, 12, 216, 36, 220, 60, 0, 12, 216, 37, 220, 60, 0, 12, 216, 38, 220, 60, 0, 12, 216, 39, 220, 60, 0, 12, 216, 40, 220, 60, 0, 12, 216, 41, 220, 60, 0, 12, 216, 42, 220, 60, 0, 12, 216, 43, 220, 60, 0, 12, 216, 44, 220, 60, 0, 12, 216, 45, 220, 60, 0, 12, 216, 46, 220, 60, 0, 12, 216, 47, 220, 60, 0, 12, 216, 48, 220, 60, 0, 12, 216, 49, 220, 60, 0, 12, 216, 50, 220, 60, 0, 12, 216, 51, 220, 60, 0, 12, 216, 52, 220, 60, 0, 12, 216, 53, 220, 60, 0, 12, 216, 54, 220, 60, 0, 12, 216, 55, 220, 60, 0, 12, 216, 56, 220, 60, 0, 12, 216, 57, 220, 60, 0, 12, 216, 58, 220, 60, 0, 12, 216, 59, 220, 60, 0, 12, 216, 60, 220, 60, 0, 12, 216, 61, 220, 60, 0, 12, 216, 62, 220, 60, 0, 12, 216, 63, 220, 60, 0, 12, 216, 64, 220, 60, 0, 12, 216, 65, 220, 60, 0, 12, 216, 66, 220, 60, 0, 12, 216, 67, 220, 60, 0, 12, 216, 68, 220, 60, 0, 12, 216, 69, 220, 60, 0, 12, 216, 70, 220, 60, 0, 12, 216, 71, 220, 60, 0, 12, 216, 72, 220, 60, 0, 12, 216, 73, 220, 60, 0, 12, 216, 74, 220, 60, 0, 12, 216, 75, 220, 60, 0, 12, 216, 76, 220, 60, 0, 12, 216, 77, 220, 60, 0, 12, 216, 78, 220, 60, 0, 12, 216, 79, 220, 60, 0, 12, 216, 80, 220, 60, 0, 12, 216, 81, 220, 60, 0, 12, 216, 82, 220, 60, 0, 12, 216, 83, 220, 60, 0, 12, 216, 84, 220, 60, 0, 12, 216, 85, 220, 60, 0, 12, 216, 86, 220, 60, 0, 12, 216, 87, 220, 60, 0, 12, 216, 88, 220, 60, 0, 12, 216, 89, 220, 60, 0, 12, 216, 90, 220, 60, 0, 12, 216, 91, 220, 60, 0, 12, 216, 92, 220, 60, 0, 12, 216, 93, 220, 60, 0, 12, 216, 94, 220, 60, 0, 12, 216, 95, 220, 60, 0, 12, 216, 96, 220, 60, 0, 12, 216, 97, 220, 60, 0, 12, 216, 98, 220, 60, 0, 12, 216, 99, 220, 60, 0, 12, 216, 100, 220, 60, 0, 12, 216, 101, 220, 60, 0, 12, 216, 102, 220, 60, 0, 12, 216, 103, 220, 60, 0, 12, 216, 104, 220, 60, 0, 12, 216, 105, 220, 60, 0, 12, 216, 106, 220, 60, 0, 12, 216, 107, 220, 60, 0, 12, 216, 108, 220, 60, 0, 12, 216, 109, 220, 60, 0, 12, 216, 110, 220, 60, 0, 12, 216, 111, 220, 60, 0, 12, 216, 112, 220, 60, 0, 12, 216, 113, 220, 60, 0, 12, 216, 114, 220, 60, 0, 12, 216, 115, 220, 60, 0, 12, 216, 116, 220, 60, 0, 12, 216, 117, 220, 60, 0, 12, 216, 118, 220, 60, 0, 12, 216, 119, 220, 60, 0, 12, 216, 120, 220, 60, 0, 12, 216, 121, 220, 60, 0, 12, 216, 122, 220, 60, 0, 12, 216, 123, 220, 60, 0, 12, 216, 124, 220, 60, 0, 12, 216, 125, 220, 60, 0, 12, 216, 126, 220, 60, 0, 12, 216, 127, 220, 60, 0, 12, 216, 128, 220, 60, 0, 12, 216, 129, 220, 60, 0, 12, 216, 130, 220, 60, 0, 12, 216, 131, 220, 60, 0, 12, 216, 132, 220, 60, 0, 12, 216, 133, 220, 60, 0, 12, 216, 134, 220, 60, 0, 12, 216, 135, 220, 60, 0, 12, 216, 136, 220, 60, 0, 12, 216, 137, 220, 60, 0, 12, 216, 138, 220, 60, 0, 12, 216, 139, 220, 60, 0, 12, 216, 140, 220, 60, 0, 12, 216, 141, 220, 60, 0, 12, 216, 142, 220, 60, 0, 12, 216, 143, 220, 60, 0, 12, 216, 144, 220, 60, 0, 12, 216, 145, 220, 60, 0, 12, 216, 146, 220, 60, 0, 12, 216, 147, 220, 60, 0, 12, 216, 148, 220, 60, 0, 12, 216, 149, 220, 60, 0, 12, 216, 150, 220, 60, 0, 12, 216, 151, 220, 60, 0, 12, 216, 152, 220, 60, 0, 12, 216, 153, 220, 60, 0, 12, 216, 154, 220, 60, 0, 12, 216, 155, 220, 60, 0, 12, 216, 156, 220, 60, 0, 12, 216, 157, 220, 60, 0, 12, 216, 158, 220, 60, 0, 12, 216, 159, 220, 60, 0, 12, 216, 160, 220, 60, 0, 12, 216, 161, 220, 60, 0, 12, 216, 162, 220, 60, 0, 12, 216, 163, 220, 60, 0, 12, 216, 164, 220, 60, 0, 12, 216, 165, 220, 60, 0, 12, 216, 166, 220, 60, 0, 12, 216, 167, 220, 60, 0, 12, 216, 168, 220, 60, 0, 12, 216, 169, 220, 60, 0, 12, 216, 170, 220, 60, 0, 12, 216, 171, 220, 60, 0, 12, 216, 172, 220, 60, 0, 12, 216, 173, 220, 60, 0, 12, 216, 174, 220, 60, 0, 12, 216, 175, 220, 60, 0, 12, 216, 176, 220, 60, 0, 12, 216, 177, 220, 60, 0, 12, 216, 178, 220, 60, 0, 12, 216, 179, 220, 60, 0, 12, 216, 180, 220, 60, 0, 12, 216, 181, 220, 60, 0, 12, 216, 182, 220, 60, 0, 12, 216, 183, 220, 60, 0, 12, 216, 184, 220, 60, 0, 12, 216, 185, 220, 60, 0, 12, 216, 186, 220, 60, 0, 12, 216, 187, 220, 60, 0, 12, 216, 188, 220, 60, 0, 12, 216, 189, 220, 60, 0, 12, 216, 190, 220, 60, 0, 12, 216, 191, 220, 60, 0, 12, 216, 192, 220, 60, 0, 12, 216, 193, 220, 60, 0, 12, 216, 194, 220, 60, 0, 12, 216, 195, 220, 60, 0, 12, 216, 196, 220, 60, 0, 12, 216, 197, 220, 60, 0, 12, 216, 198, 220, 60, 0, 12, 216, 199, 220, 60, 0, 12, 216, 200, 220, 60, 0, 12, 216, 201, 220, 60, 0, 12, 216, 202, 220, 60, 0, 12, 216, 203, 220, 60, 0, 12, 216, 204, 220, 60, 0, 12, 216, 205, 220, 60, 0, 12, 216, 206, 220, 60, 0, 12, 216, 207, 220, 60, 0, 12, 216, 208, 220, 60, 0, 12, 216, 209, 220, 60, 0, 12, 216, 210, 220, 60, 0, 12, 216, 211, 220, 60, 0, 12, 216, 212, 220, 60, 0, 12, 216, 213, 220, 60, 0, 12, 216, 214, 220, 60, 0, 12, 216, 215, 220, 60, 0, 12, 216, 216, 220, 60, 0, 12, 216, 217, 220, 60, 0, 12, 216, 218, 220, 60, 0, 12, 216, 219, 220, 60, 0, 12, 216, 220, 220, 60, 0, 12, 216, 221, 220, 60, 0, 12, 216, 222, 220, 60, 0, 12, 216, 223, 220, 60, 0, 12, 216, 224, 220, 60, 0, 12, 216, 225, 220, 60, 0, 12, 216, 226, 220, 60, 0, 12, 216, 227, 220, 60, 0, 12, 216, 228, 220, 60, 0, 12, 216, 229, 220, 60, 0, 12, 216, 230, 220, 60, 0, 12, 216, 231, 220, 60, 0, 12, 216, 232, 220, 60, 0, 12, 216, 233, 220, 60, 0, 12, 216, 234, 220, 60, 0, 12, 216, 235, 220, 60, 0, 12, 216, 236, 220, 60, 0, 12, 216, 237, 220, 60, 0, 12, 216, 238, 220, 60, 0, 12, 216, 239, 220, 60, 0, 12, 216, 240, 220, 60, 0, 12, 216, 241, 220, 60, 0, 12, 216, 242, 220, 60, 0, 12, 216, 243, 220, 60, 0, 12, 216, 244, 220, 60, 0, 12, 216, 245, 220, 60, 0, 12, 216, 246, 220, 60, 0, 12, 216, 247, 220, 60, 0, 12, 216, 248, 220, 60, 0, 12, 216, 249, 220, 60, 0, 12, 216, 250, 220, 60, 0, 12, 216, 251, 220, 60, 0, 12, 216, 252, 220, 60, 0, 12, 216, 253, 220, 60, 0, 12, 216, 254, 220, 60, 0, 12, 216, 255, 220, 60, 0, 12, 216, 0, 221, 60, 0, 12, 216, 1, 221, 60, 0, 12, 216, 2, 221, 60, 0, 12, 216, 3, 221, 60, 0, 12, 216, 4, 221, 60, 0, 12, 216, 5, 221, 60, 0, 12, 216, 6, 221, 60, 0, 12, 216, 7, 221, 60, 0, 12, 216, 8, 221, 60, 0, 12, 216, 9, 221, 60, 0, 12, 216, 10, 221, 60, 0, 12, 216, 11, 221, 60, 0, 12, 216, 12, 221, 60, 0, 12, 216, 13, 221, 60, 0, 12, 216, 14, 221, 60, 0, 12, 216, 15, 221, 60, 0, 12, 216, 16, 221, 60, 0, 12, 216, 17, 221, 60, 0, 12, 216, 18, 221, 60, 0, 12, 216, 19, 221, 60, 0, 12, 216, 20, 221, 60, 0, 12, 216, 21, 221, 60, 0, 12, 216, 22, 221, 60, 0, 12, 216, 23, 221, 60, 0, 12, 216, 24, 221, 60, 0, 12, 216, 25, 221, 60, 0, 12, 216, 26, 221, 60, 0, 12, 216, 27, 221, 60, 0, 12, 216, 28, 221, 60, 0, 12, 216, 29, 221, 60, 0, 12, 216, 30, 221, 60, 0, 12, 216, 31, 221, 60, 0, 12, 216, 32, 221, 60, 0, 12, 216, 33, 221, 60, 0, 12, 216, 34, 221, 60, 0, 12, 216, 35, 221, 60, 0, 12, 216, 36, 221, 60, 0, 12, 216, 37, 221, 60, 0, 12, 216, 38, 221, 60, 0, 12, 216, 39, 221, 60, 0, 12, 216, 40, 221, 60, 0, 12, 216, 41, 221, 60, 0, 12, 216, 42, 221, 60, 0, 12, 216, 43, 221, 60, 0, 12, 216, 44, 221, 60, 0, 12, 216, 45, 221, 60, 0, 12, 216, 46, 221, 60, 0, 12, 216, 47, 221, 60, 0, 12, 216, 48, 221, 60, 0, 12, 216, 49, 221, 60, 0, 12, 216, 50, 221, 60, 0, 12, 216, 51, 221, 60, 0, 12, 216, 52, 221, 60, 0, 12, 216, 53, 221, 60, 0, 12, 216, 54, 221, 60, 0, 12, 216, 55, 221, 60, 0, 12, 216, 56, 221, 60, 0, 12, 216, 57, 221, 60, 0, 12, 216, 58, 221, 60, 0, 12, 216, 59, 221, 60, 0, 12, 216, 60, 221, 60, 0, 12, 216, 61, 221, 60, 0, 12, 216, 62, 221, 60, 0, 12, 216, 63, 221, 60, 0, 12, 216, 64, 221, 60, 0, 12, 216, 65, 221, 60, 0, 12, 216, 66, 221, 60, 0, 12, 216, 67, 221, 60, 0, 12, 216, 68, 221, 60, 0, 12, 216, 69, 221, 60, 0, 12, 216, 70, 221, 60, 0, 12, 216, 71, 221, 60, 0, 12, 216, 72, 221, 60, 0, 12, 216, 73, 221, 60, 0, 12, 216, 74, 221, 60, 0, 12, 216, 75, 221, 60, 0, 12, 216, 76, 221, 60, 0, 12, 216, 77, 221, 60, 0, 12, 216, 78, 221, 60, 0, 12, 216, 79, 221, 60, 0, 12, 216, 80, 221, 60, 0, 12, 216, 81, 221, 60, 0, 12, 216, 82, 221, 60, 0, 12, 216, 83, 221, 60, 0, 12, 216, 84, 221, 60, 0, 12, 216, 85, 221, 60, 0, 12, 216, 86, 221, 60, 0, 12, 216, 87, 221, 60, 0, 12, 216, 88, 221, 60, 0, 12, 216, 89, 221, 60, 0, 12, 216, 90, 221, 60, 0, 12, 216, 91, 221, 60, 0, 12, 216, 92, 221, 60, 0, 12, 216, 93, 221, 60, 0, 12, 216, 94, 221, 60, 0, 12, 216, 95, 221, 60, 0, 12, 216, 96, 221, 60, 0, 12, 216, 97, 221, 60, 0, 12, 216, 98, 221, 60, 0, 12, 216, 99, 221, 60, 0, 12, 216, 100, 221, 60, 0, 12, 216, 101, 221, 60, 0, 12, 216, 102, 221, 60, 0, 12, 216, 103, 221, 60, 0, 12, 216, 104, 221, 60, 0, 12, 216, 105, 221, 60, 0, 12, 216, 106, 221, 60, 0, 12, 216, 107, 221, 60, 0, 12, 216, 108, 221, 60, 0, 12, 216, 109, 221, 60, 0, 12, 216, 110, 221, 60, 0, 12, 216, 111, 221, 60, 0, 12, 216, 112, 221, 60, 0, 12, 216, 113, 221, 60, 0, 12, 216, 114, 221, 60, 0, 12, 216, 115, 221, 60, 0, 12, 216, 116, 221, 60, 0, 12, 216, 117, 221, 60, 0, 12, 216, 118, 221, 60, 0, 12, 216, 119, 221, 60, 0, 12, 216, 120, 221, 60, 0, 12, 216, 121, 221, 60, 0, 12, 216, 122, 221, 60, 0, 12, 216, 123, 221, 60, 0, 12, 216, 124, 221, 60, 0, 12, 216, 125, 221, 60, 0, 12, 216, 126, 221, 60, 0, 12, 216, 127, 221, 60, 0, 12, 216, 128, 221, 60, 0, 12, 216, 129, 221, 60, 0, 12, 216, 130, 221, 60, 0, 12, 216, 131, 221, 60, 0, 12, 216, 132, 221, 60, 0, 12, 216, 133, 221, 60, 0, 12, 216, 134, 221, 60, 0, 12, 216, 135, 221, 60, 0, 12, 216, 136, 221, 60, 0, 12, 216, 137, 221, 60, 0, 12, 216, 138, 221, 60, 0, 12, 216, 139, 221, 60, 0, 12, 216, 140, 221, 60, 0, 12, 216, 141, 221, 60, 0, 12, 216, 142, 221, 60, 0, 12, 216, 143, 221, 60, 0, 12, 216, 144, 221, 60, 0, 12, 216, 145, 221, 60, 0, 12, 216, 146, 221, 60, 0, 12, 216, 147, 221, 60, 0, 12, 216, 148, 221, 60, 0, 12, 216, 149, 221, 60, 0, 12, 216, 150, 221, 60, 0, 12, 216, 151, 221, 60, 0, 12, 216, 152, 221, 60, 0, 12, 216, 153, 221, 60, 0, 12, 216, 154, 221, 60, 0, 12, 216, 155, 221, 60, 0, 12, 216, 156, 221, 60, 0, 12, 216, 157, 221, 60, 0, 12, 216, 158, 221, 60, 0, 12, 216, 159, 221, 60, 0, 12, 216, 160, 221, 60, 0, 12, 216, 161, 221, 60, 0, 12, 216, 162, 221, 60, 0, 12, 216, 163, 221, 60, 0, 12, 216, 164, 221, 60, 0, 12, 216, 165, 221, 60, 0, 12, 216, 166, 221, 60, 0, 12, 216, 167, 221, 60, 0, 12, 216, 168, 221, 60, 0, 12, 216, 169, 221, 60, 0, 12, 216, 170, 221, 60, 0, 12, 216, 171, 221, 60, 0, 12, 216, 172, 221, 60, 0, 12, 216, 173, 221, 60, 0, 12, 216, 174, 221, 60, 0, 12, 216, 175, 221, 60, 0, 12, 216, 176, 221, 60, 0, 12, 216, 177, 221, 60, 0, 12, 216, 178, 221, 60, 0, 12, 216, 179, 221, 60, 0, 12, 216, 180, 221, 60, 0, 12, 216, 181, 221, 60, 0, 12, 216, 182, 221, 60, 0, 12, 216, 183, 221, 60, 0, 12, 216, 184, 221, 60, 0, 12, 216, 185, 221, 60, 0, 12, 216, 186, 221, 60, 0, 12, 216, 187, 221, 60, 0, 12, 216, 188, 221, 60, 0, 12, 216, 189, 221, 60, 0, 12, 216, 190, 221, 60, 0, 12, 216, 191, 221, 60, 0, 12, 216, 192, 221, 60, 0, 12, 216, 193, 221, 60, 0, 12, 216, 194, 221, 60, 0, 12, 216, 195, 221, 60, 0, 12, 216, 196, 221, 60, 0, 12, 216, 197, 221, 60, 0, 12, 216, 198, 221, 60, 0, 12, 216, 199, 221, 60, 0, 12, 216, 200, 221, 60, 0, 12, 216, 201, 221, 60, 0, 12, 216, 202, 221, 60, 0, 12, 216, 203, 221, 60, 0, 12, 216, 204, 221, 60, 0, 12, 216, 205, 221, 60, 0, 12, 216, 206, 221, 60, 0, 12, 216, 207, 221, 60, 0, 12, 216, 208, 221, 60, 0, 12, 216, 209, 221, 60, 0, 12, 216, 210, 221, 60, 0, 12, 216, 211, 221, 60, 0, 12, 216, 212, 221, 60, 0, 12, 216, 213, 221, 60, 0, 12, 216, 214, 221, 60, 0, 12, 216, 215, 221, 60, 0, 12, 216, 216, 221, 60, 0, 12, 216, 217, 221, 60, 0, 12, 216, 218, 221, 60, 0, 12, 216, 219, 221, 60, 0, 12, 216, 220, 221, 60, 0, 12, 216, 221, 221, 60, 0, 12, 216, 222, 221, 60, 0, 12, 216, 223, 221, 60, 0, 12, 216, 224, 221, 60, 0, 12, 216, 225, 221, 60, 0, 12, 216, 226, 221, 60, 0, 12, 216, 227, 221, 60, 0, 12, 216, 228, 221, 60, 0, 12, 216, 229, 221, 60, 0, 12, 216, 230, 221, 60, 0, 12, 216, 231, 221, 60, 0, 12, 216, 232, 221, 60, 0, 12, 216, 233, 221, 60, 0, 12, 216, 234, 221, 60, 0, 12, 216, 235, 221, 60, 0, 12, 216, 236, 221, 60, 0, 12, 216, 237, 221, 60, 0, 12, 216, 238, 221, 60, 0, 12, 216, 239, 221, 60, 0, 12, 216, 240, 221, 60, 0, 12, 216, 241, 221, 60, 0, 12, 216, 242, 221, 60, 0, 12, 216, 243, 221, 60, 0, 12, 216, 244, 221, 60, 0, 12, 216, 245, 221, 60, 0, 12, 216, 246, 221, 60, 0, 12, 216, 247, 221, 60, 0, 12, 216, 248, 221, 60, 0, 12, 216, 249, 221, 60, 0, 12, 216, 250, 221, 60, 0, 12, 216, 251, 221, 60, 0, 12, 216, 252, 221, 60, 0, 12, 216, 253, 221, 60, 0, 12, 216, 254, 221, 60, 0, 12, 216, 255, 221, 60, 0, 12, 216, 0, 222, 60, 0, 12, 216, 1, 222, 60, 0, 12, 216, 2, 222, 60, 0, 12, 216, 3, 222, 60, 0, 12, 216, 4, 222, 60, 0, 12, 216, 5, 222, 60, 0, 12, 216, 6, 222, 60, 0, 12, 216, 7, 222, 60, 0, 12, 216, 8, 222, 60, 0, 12, 216, 9, 222, 60, 0, 12, 216, 10, 222, 60, 0, 12, 216, 11, 222, 60, 0, 12, 216, 12, 222, 60, 0, 12, 216, 13, 222, 60, 0, 12, 216, 14, 222, 60, 0, 12, 216, 15, 222, 60, 0, 12, 216, 16, 222, 60, 0, 12, 216, 17, 222, 60, 0, 12, 216, 18, 222, 60, 0, 12, 216, 19, 222, 60, 0, 12, 216, 20, 222, 60, 0, 12, 216, 21, 222, 60, 0, 12, 216, 22, 222, 60, 0, 12, 216, 23, 222, 60, 0, 12, 216, 24, 222, 60, 0, 12, 216, 25, 222, 60, 0, 12, 216, 26, 222, 60, 0, 12, 216, 27, 222, 60, 0, 12, 216, 28, 222, 60, 0, 12, 216, 29, 222, 60, 0, 12, 216, 30, 222, 60, 0, 12, 216, 31, 222, 60, 0, 12, 216, 32, 222, 60, 0, 12, 216, 33, 222, 60, 0, 12, 216, 34, 222, 60, 0, 12, 216, 35, 222, 60, 0, 12, 216, 36, 222, 60, 0, 12, 216, 37, 222, 60, 0, 12, 216, 38, 222, 60, 0, 12, 216, 39, 222, 60, 0, 12, 216, 40, 222, 60, 0, 12, 216, 41, 222, 60, 0, 12, 216, 42, 222, 60, 0, 12, 216, 43, 222, 60, 0, 12, 216, 44, 222, 60, 0, 12, 216, 45, 222, 60, 0, 12, 216, 46, 222, 60, 0, 12, 216, 47, 222, 60, 0, 12, 216, 48, 222, 60, 0, 12, 216, 49, 222, 60, 0, 12, 216, 50, 222, 60, 0, 12, 216, 51, 222, 60, 0, 12, 216, 52, 222, 60, 0, 12, 216, 53, 222, 60, 0, 12, 216, 54, 222, 60, 0, 12, 216, 55, 222, 60, 0, 12, 216, 56, 222, 60, 0, 12, 216, 57, 222, 60, 0, 12, 216, 58, 222, 60, 0, 12, 216, 59, 222, 60, 0, 12, 216, 60, 222, 60, 0, 12, 216, 61, 222, 60, 0, 12, 216, 62, 222, 60, 0, 12, 216, 63, 222, 60, 0, 12, 216, 64, 222, 60, 0, 12, 216, 65, 222, 60, 0, 12, 216, 66, 222, 60, 0, 12, 216, 67, 222, 60, 0, 12, 216, 68, 222, 60, 0, 12, 216, 69, 222, 60, 0, 12, 216, 70, 222, 60, 0, 12, 216, 71, 222, 60, 0, 12, 216, 72, 222, 60, 0, 12, 216, 73, 222, 60, 0, 12, 216, 74, 222, 60, 0, 12, 216, 75, 222, 60, 0, 12, 216, 76, 222, 60, 0, 12, 216, 77, 222, 60, 0, 12, 216, 78, 222, 60, 0, 12, 216, 79, 222, 60, 0, 12, 216, 80, 222, 60, 0, 12, 216, 81, 222, 60, 0, 12, 216, 82, 222, 60, 0, 12, 216, 83, 222, 60, 0, 12, 216, 84, 222, 60, 0, 12, 216, 85, 222, 60, 0, 12, 216, 86, 222, 60, 0, 12, 216, 87, 222, 60, 0, 12, 216, 88, 222, 60, 0, 12, 216, 89, 222, 60, 0, 12, 216, 90, 222, 60, 0, 12, 216, 91, 222, 60, 0, 12, 216, 92, 222, 60, 0, 12, 216, 93, 222, 60, 0, 12, 216, 94, 222, 60, 0, 12, 216, 95, 222, 60, 0, 12, 216, 96, 222, 60, 0, 12, 216, 97, 222, 60, 0, 12, 216, 98, 222, 60, 0, 12, 216, 99, 222, 60, 0, 12, 216, 100, 222, 60, 0, 12, 216, 101, 222, 60, 0, 12, 216, 102, 222, 60, 0, 12, 216, 103, 222, 60, 0, 12, 216, 104, 222, 60, 0, 12, 216, 105, 222, 60, 0, 12, 216, 106, 222, 60, 0, 12, 216, 107, 222, 60, 0, 12, 216, 108, 222, 60, 0, 12, 216, 109, 222, 60, 0, 12, 216, 110, 222, 60, 0, 12, 216, 111, 222, 60, 0, 12, 216, 112, 222, 60, 0, 12, 216, 113, 222, 60, 0, 12, 216, 114, 222, 60, 0, 12, 216, 115, 222, 60, 0, 12, 216, 116, 222, 60, 0, 12, 216, 117, 222, 60, 0, 12, 216, 118, 222, 60, 0, 12, 216, 119, 222, 60, 0, 12, 216, 120, 222, 60, 0, 12, 216, 121, 222, 60, 0, 12, 216, 122, 222, 60, 0, 12, 216, 123, 222, 60, 0, 12, 216, 124, 222, 60, 0, 12, 216, 125, 222, 60, 0, 12, 216, 126, 222, 60, 0, 12, 216, 127, 222, 60, 0, 12, 216, 128, 222, 60, 0, 12, 216, 129, 222, 60, 0, 12, 216, 130, 222, 60, 0, 12, 216, 131, 222, 60, 0, 12, 216, 132, 222, 60, 0, 12, 216, 133, 222, 60, 0, 12, 216, 134, 222, 60, 0, 12, 216, 135, 222, 60, 0, 12, 216, 136, 222, 60, 0, 12, 216, 137, 222, 60, 0, 12, 216, 138, 222, 60, 0, 12, 216, 139, 222, 60, 0, 12, 216, 140, 222, 60, 0, 12, 216, 141, 222, 60, 0, 12, 216, 142, 222, 60, 0, 12, 216, 143, 222, 60, 0, 12, 216, 144, 222, 60, 0, 12, 216, 145, 222, 60, 0, 12, 216, 146, 222, 60, 0, 12, 216, 147, 222, 60, 0, 12, 216, 148, 222, 60, 0, 12, 216, 149, 222, 60, 0, 12, 216, 150, 222, 60, 0, 12, 216, 151, 222, 60, 0, 12, 216, 152, 222, 60, 0, 12, 216, 153, 222, 60, 0, 12, 216, 154, 222, 60, 0, 12, 216, 155, 222, 60, 0, 12, 216, 156, 222, 60, 0, 12, 216, 157, 222, 60, 0, 12, 216, 158, 222, 60, 0, 12, 216, 159, 222, 60, 0, 12, 216, 160, 222, 60, 0, 12, 216, 161, 222, 60, 0, 12, 216, 162, 222, 60, 0, 12, 216, 163, 222, 60, 0, 12, 216, 164, 222, 60, 0, 12, 216, 165, 222, 60, 0, 12, 216, 166, 222, 60, 0, 12, 216, 167, 222, 60, 0, 12, 216, 168, 222, 60, 0, 12, 216, 169, 222, 60, 0, 12, 216, 170, 222, 60, 0, 12, 216, 171, 222, 60, 0, 12, 216, 172, 222, 60, 0, 12, 216, 173, 222, 60, 0, 12, 216, 174, 222, 60, 0, 12, 216, 175, 222, 60, 0, 12, 216, 176, 222, 60, 0, 12, 216, 177, 222, 60, 0, 12, 216, 178, 222, 60, 0, 12, 216, 179, 222, 60, 0, 12, 216, 180, 222, 60, 0, 12, 216, 181, 222, 60, 0, 12, 216, 182, 222, 60, 0, 12, 216, 183, 222, 60, 0, 12, 216, 184, 222, 60, 0, 12, 216, 185, 222, 60, 0, 12, 216, 186, 222, 60, 0, 12, 216, 187, 222, 60, 0, 12, 216, 188, 222, 60, 0, 12, 216, 189, 222, 60, 0, 12, 216, 190, 222, 60, 0, 12, 216, 191, 222, 60, 0, 12, 216, 192, 222, 60, 0, 12, 216, 193, 222, 60, 0, 12, 216, 194, 222, 60, 0, 12, 216, 195, 222, 60, 0, 12, 216, 196, 222, 60, 0, 12, 216, 197, 222, 60, 0, 12, 216, 198, 222, 60, 0, 12, 216, 199, 222, 60, 0, 12, 216, 200, 222, 60, 0, 12, 216, 201, 222, 60, 0, 12, 216, 202, 222, 60, 0, 12, 216, 203, 222, 60, 0, 12, 216, 204, 222, 60, 0, 12, 216, 205, 222, 60, 0, 12, 216, 206, 222, 60, 0, 12, 216, 207, 222, 60, 0, 12, 216, 208, 222, 60, 0, 12, 216, 209, 222, 60, 0, 12, 216, 210, 222, 60, 0, 12, 216, 211, 222, 60, 0, 12, 216, 212, 222, 60, 0, 12, 216, 213, 222, 60, 0, 12, 216, 214, 222, 60, 0, 12, 216, 215, 222, 60, 0, 12, 216, 216, 222, 60, 0, 12, 216, 217, 222, 60, 0, 12, 216, 218, 222, 60, 0, 12, 216, 219, 222, 60, 0, 12, 216, 220, 222, 60, 0, 12, 216, 221, 222, 60, 0, 12, 216, 222, 222, 60, 0, 12, 216, 223, 222, 60, 0, 12, 216, 224, 222, 60, 0, 12, 216, 225, 222, 60, 0, 12, 216, 226, 222, 60, 0, 12, 216, 227, 222, 60, 0, 12, 216, 228, 222, 60, 0, 12, 216, 229, 222, 60, 0, 12, 216, 230, 222, 60, 0, 12, 216, 231, 222, 60, 0, 12, 216, 232, 222, 60, 0, 12, 216, 233, 222, 60, 0, 12, 216, 234, 222, 60, 0, 12, 216, 235, 222, 60, 0, 12, 216, 236, 222, 60, 0, 12, 216, 237, 222, 60, 0, 12, 216, 238, 222, 60, 0, 12, 216, 239, 222, 60, 0, 12, 216, 240, 222, 60, 0, 12, 216, 241, 222, 60, 0, 12, 216, 242, 222, 60, 0, 12, 216, 243, 222, 60, 0, 12, 216, 244, 222, 60, 0, 12, 216, 245, 222, 60, 0, 12, 216, 246, 222, 60, 0, 12, 216, 247, 222, 60, 0, 12, 216, 248, 222, 60, 0, 12, 216, 249, 222, 60, 0, 12, 216, 250, 222, 60, 0, 12, 216, 251, 222, 60, 0, 12, 216, 252, 222, 60, 0, 12, 216, 253, 222, 60, 0, 12, 216, 254, 222, 60, 0, 12, 216, 255, 222, 60, 0, 12, 216, 0, 223, 60, 0, 12, 216, 1, 223, 60, 0, 12, 216, 2, 223, 60, 0, 12, 216, 3, 223, 60, 0, 12, 216, 4, 223, 60, 0, 12, 216, 5, 223, 60, 0, 12, 216, 6, 223, 60, 0, 12, 216, 7, 223, 60, 0, 12, 216, 8, 223, 60, 0, 12, 216, 9, 223, 60, 0, 12, 216, 10, 223, 60, 0, 12, 216, 11, 223, 60, 0, 12, 216, 12, 223, 60, 0, 12, 216, 13, 223, 60, 0, 12, 216, 14, 223, 60, 0, 12, 216, 15, 223, 60, 0, 12, 216, 16, 223, 60, 0, 12, 216, 17, 223, 60, 0, 12, 216, 18, 223, 60, 0, 12, 216, 19, 223, 60, 0, 12, 216, 20, 223, 60, 0, 12, 216, 21, 223, 60, 0, 12, 216, 22, 223, 60, 0, 12, 216, 23, 223, 60, 0, 12, 216, 24, 223, 60, 0, 12, 216, 25, 223, 60, 0, 12, 216, 26, 223, 60, 0, 12, 216, 27, 223, 60, 0, 12, 216, 28, 223, 60, 0, 12, 216, 29, 223, 60, 0, 12, 216, 30, 223, 60, 0, 12, 216, 31, 223, 60, 0, 12, 216, 32, 223, 60, 0, 12, 216, 33, 223, 60, 0, 12, 216, 34, 223, 60, 0, 12, 216, 35, 223, 60, 0, 12, 216, 36, 223, 60, 0, 12, 216, 37, 223, 60, 0, 12, 216, 38, 223, 60, 0, 12, 216, 39, 223, 60, 0, 12, 216, 40, 223, 60, 0, 12, 216, 41, 223, 60, 0, 12, 216, 42, 223, 60, 0, 12, 216, 43, 223, 60, 0, 12, 216, 44, 223, 60, 0, 12, 216, 45, 223, 60, 0, 12, 216, 46, 223, 60, 0, 12, 216, 47, 223, 60, 0, 12, 216, 48, 223, 60, 0, 12, 216, 49, 223, 60, 0, 12, 216, 50, 223, 60, 0, 12, 216, 51, 223, 60, 0, 12, 216, 52, 223, 60, 0, 12, 216, 53, 223, 60, 0, 12, 216, 54, 223, 60, 0, 12, 216, 55, 223, 60, 0, 12, 216, 56, 223, 60, 0, 12, 216, 57, 223, 60, 0, 12, 216, 58, 223, 60, 0, 12, 216, 59, 223, 60, 0, 12, 216, 60, 223, 60, 0, 12, 216, 61, 223, 60, 0, 12, 216, 62, 223, 60, 0, 12, 216, 63, 223, 60, 0, 12, 216, 64, 223, 60, 0, 12, 216, 65, 223, 60, 0, 12, 216, 66, 223, 60, 0, 12, 216, 67, 223, 60, 0, 12, 216, 68, 223, 60, 0, 12, 216, 69, 223, 60, 0, 12, 216, 70, 223, 60, 0, 12, 216, 71, 223, 60, 0, 12, 216, 72, 223, 60, 0, 12, 216, 73, 223, 60, 0, 12, 216, 74, 223, 60, 0, 12, 216, 75, 223, 60, 0, 12, 216, 76, 223, 60, 0, 12, 216, 77, 223, 60, 0, 12, 216, 78, 223, 60, 0, 12, 216, 79, 223, 60, 0, 12, 216, 80, 223, 60, 0, 12, 216, 81, 223, 60, 0, 12, 216, 82, 223, 60, 0, 12, 216, 83, 223, 60, 0, 12, 216, 84, 223, 60, 0, 12, 216, 85, 223, 60, 0, 12, 216, 86, 223, 60, 0, 12, 216, 87, 223, 60, 0, 12, 216, 88, 223, 60, 0, 12, 216, 89, 223, 60, 0, 12, 216, 90, 223, 60, 0, 12, 216, 91, 223, 60, 0, 12, 216, 92, 223, 60, 0, 12, 216, 93, 223, 60, 0, 12, 216, 94, 223, 60, 0, 12, 216, 95, 223, 60, 0, 12, 216, 96, 223, 60, 0, 12, 216, 97, 223, 60, 0, 12, 216, 98, 223, 60, 0, 12, 216, 99, 223, 60, 0, 12, 216, 100, 223, 60, 0, 12, 216, 101, 223, 60, 0, 12, 216, 102, 223, 60, 0, 12, 216, 103, 223, 60, 0, 12, 216, 104, 223, 60, 0, 12, 216, 105, 223, 60, 0, 12, 216, 106, 223, 60, 0, 12, 216, 107, 223, 60, 0, 12, 216, 108, 223, 60, 0, 12, 216, 109, 223, 60, 0, 12, 216, 110, 223, 60, 0, 12, 216, 111, 223, 60, 0, 12, 216, 112, 223, 60, 0, 12, 216, 113, 223, 60, 0, 12, 216, 114, 223, 60, 0, 12, 216, 115, 223, 60, 0, 12, 216, 116, 223, 60, 0, 12, 216, 117, 223, 60, 0, 12, 216, 118, 223, 60, 0, 12, 216, 119, 223, 60, 0, 12, 216, 120, 223, 60, 0, 12, 216, 121, 223, 60, 0, 12, 216, 122, 223, 60, 0, 12, 216, 123, 223, 60, 0, 12, 216, 124, 223, 60, 0, 12, 216, 125, 223, 60, 0, 12, 216, 126, 223, 60, 0, 12, 216, 127, 223, 60, 0, 12, 216, 128, 223, 60, 0, 12, 216, 129, 223, 60, 0, 12, 216, 130, 223, 60, 0, 12, 216, 131, 223, 60, 0, 12, 216, 132, 223, 60, 0, 12, 216, 133, 223, 60, 0, 12, 216, 134, 223, 60, 0, 12, 216, 135, 223, 60, 0, 12, 216, 136, 223, 60, 0, 12, 216, 137, 223, 60, 0, 12, 216, 138, 223, 60, 0, 12, 216, 139, 223, 60, 0, 12, 216, 140, 223, 60, 0, 12, 216, 141, 223, 60, 0, 12, 216, 142, 223, 60, 0, 12, 216, 143, 223, 60, 0, 12, 216, 144, 223, 60, 0, 12, 216, 145, 223, 60, 0, 12, 216, 146, 223, 60, 0, 12, 216, 147, 223, 60, 0, 12, 216, 148, 223, 60, 0, 12, 216, 149, 223, 60, 0, 12, 216, 150, 223, 60, 0, 12, 216, 151, 223, 60, 0, 12, 216, 152, 223, 60, 0, 12, 216, 153, 223, 60, 0, 12, 216, 154, 223, 60, 0, 12, 216, 155, 223, 60, 0, 12, 216, 156, 223, 60, 0, 12, 216, 157, 223, 60, 0, 12, 216, 158, 223, 60, 0, 12, 216, 159, 223, 60, 0, 12, 216, 160, 223, 60, 0, 12, 216, 161, 223, 60, 0, 12, 216, 162, 223, 60, 0, 12, 216, 163, 223, 60, 0, 12, 216, 164, 223, 60, 0, 12, 216, 165, 223, 60, 0, 12, 216, 166, 223, 60, 0, 12, 216, 167, 223, 60, 0, 12, 216, 168, 223, 60, 0, 12, 216, 169, 223, 60, 0, 12, 216, 170, 223, 60, 0, 12, 216, 171, 223, 60, 0, 12, 216, 172, 223, 60, 0, 12, 216, 173, 223, 60, 0, 12, 216, 174, 223, 60, 0, 12, 216, 175, 223, 60, 0, 12, 216, 176, 223, 60, 0, 12, 216, 177, 223, 60, 0, 12, 216, 178, 223, 60, 0, 12, 216, 179, 223, 60, 0, 12, 216, 180, 223, 60, 0, 12, 216, 181, 223, 60, 0, 12, 216, 182, 223, 60, 0, 12, 216, 183, 223, 60, 0, 12, 216, 184, 223, 60, 0, 12, 216, 185, 223, 60, 0, 12, 216, 186, 223, 60, 0, 12, 216, 187, 223, 60, 0, 12, 216, 188, 223, 60, 0, 12, 216, 189, 223, 60, 0, 12, 216, 190, 223, 60, 0, 12, 216, 191, 223, 60, 0, 12, 216, 192, 223, 60, 0, 12, 216, 193, 223, 60, 0, 12, 216, 194, 223, 60, 0, 12, 216, 195, 223, 60, 0, 12, 216, 196, 223, 60, 0, 12, 216, 197, 223, 60, 0, 12, 216, 198, 223, 60, 0, 12, 216, 199, 223, 60, 0, 12, 216, 200, 223, 60, 0, 12, 216, 201, 223, 60, 0, 12, 216, 202, 223, 60, 0, 12, 216, 203, 223, 60, 0, 12, 216, 204, 223, 60, 0, 12, 216, 205, 223, 60, 0, 12, 216, 206, 223, 60, 0, 12, 216, 207, 223, 60, 0, 12, 216, 208, 223, 60, 0, 12, 216, 209, 223, 60, 0, 12, 216, 210, 223, 60, 0, 12, 216, 211, 223, 60, 0, 12, 216, 212, 223, 60, 0, 12, 216, 213, 223, 60, 0, 12, 216, 214, 223, 60, 0, 12, 216, 215, 223, 60, 0, 12, 216, 216, 223, 60, 0, 12, 216, 217, 223, 60, 0, 12, 216, 218, 223, 60, 0, 12, 216, 219, 223, 60, 0, 12, 216, 220, 223, 60, 0, 12, 216, 221, 223, 60, 0, 12, 216, 222, 223, 60, 0, 12, 216, 223, 223, 60, 0, 12, 216, 224, 223, 60, 0, 12, 216, 225, 223, 60, 0, 12, 216, 226, 223, 60, 0, 12, 216, 227, 223, 60, 0, 12, 216, 228, 223, 60, 0, 12, 216, 229, 223, 60, 0, 12, 216, 230, 223, 60, 0, 12, 216, 231, 223, 60, 0, 12, 216, 232, 223, 60, 0, 12, 216, 233, 223, 60, 0, 12, 216, 234, 223, 60, 0, 12, 216, 235, 223, 60, 0, 12, 216, 236, 223, 60, 0, 12, 216, 237, 223, 60, 0, 12, 216, 238, 223, 60, 0, 12, 216, 239, 223, 60, 0, 12, 216, 240, 223, 60, 0, 12, 216, 241, 223, 60, 0, 12, 216, 242, 223, 60, 0, 12, 216, 243, 223, 60, 0, 12, 216, 244, 223, 60, 0, 12, 216, 245, 223, 60, 0, 12, 216, 246, 223, 60, 0, 12, 216, 247, 223, 60, 0, 12, 216, 248, 223, 60, 0, 12, 216, 249, 223, 60, 0, 12, 216, 250, 223, 60, 0, 12, 216, 251, 223, 60, 0, 12, 216, 252, 223, 60, 0, 12, 216, 253, 223, 60, 0, 12, 216, 254, 223, 60, 0, 12, 216, 255, 223, 60, 0, 13, 216, 0, 220, 60, 0, 13, 216, 1, 220, 60, 0, 13, 216, 2, 220, 60, 0, 13, 216, 3, 220, 60, 0, 13, 216, 4, 220, 60, 0, 13, 216, 5, 220, 60, 0, 13, 216, 6, 220, 60, 0, 13, 216, 7, 220, 60, 0, 13, 216, 8, 220, 60, 0, 13, 216, 9, 220, 60, 0, 13, 216, 10, 220, 60, 0, 13, 216, 11, 220, 60, 0, 13, 216, 12, 220, 60, 0, 13, 216, 13, 220, 60, 0, 13, 216, 14, 220, 60, 0, 13, 216, 15, 220, 60, 0, 13, 216, 16, 220, 60, 0, 13, 216, 17, 220, 60, 0, 13, 216, 18, 220, 60, 0, 13, 216, 19, 220, 60, 0, 13, 216, 20, 220, 60, 0, 13, 216, 21, 220, 60, 0, 13, 216, 22, 220, 60, 0, 13, 216, 23, 220, 60, 0, 13, 216, 24, 220, 60, 0, 13, 216, 25, 220, 60, 0, 13, 216, 26, 220, 60, 0, 13, 216, 27, 220, 60, 0, 13, 216, 28, 220, 60, 0, 13, 216, 29, 220, 60, 0, 13, 216, 30, 220, 60, 0, 13, 216, 31, 220, 60, 0, 13, 216, 32, 220, 60, 0, 13, 216, 33, 220, 60, 0, 13, 216, 34, 220, 60, 0, 13, 216, 35, 220, 60, 0, 13, 216, 36, 220, 60, 0, 13, 216, 37, 220, 60, 0, 13, 216, 38, 220, 60, 0, 13, 216, 39, 220, 60, 0, 13, 216, 40, 220, 60, 0, 13, 216, 41, 220, 60, 0, 13, 216, 42, 220, 60, 0, 13, 216, 43, 220, 60, 0, 13, 216, 44, 220, 60, 0, 13, 216, 45, 220, 60, 0, 13, 216, 46, 220, 60, 0, 14, 32, 2, 216, 160, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 128, 221, 14, 32, 60, 0, 14, 32, 2, 216, 161, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 129, 221, 14, 32, 60, 0, 14, 32, 2, 216, 162, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 130, 221, 14, 32, 60, 0, 14, 32, 2, 216, 163, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 131, 221, 14, 32, 60, 0, 14, 32, 2, 216, 164, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 132, 221, 14, 32, 60, 0, 14, 32, 2, 216, 165, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 133, 221, 14, 32, 60, 0, 14, 32, 2, 216, 166, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 134, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 135, 221, 14, 32, 60, 0, 14, 32, 2, 216, 167, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 136, 221, 14, 32, 60, 0, 14, 32, 2, 216, 168, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 137, 221, 14, 32, 60, 0, 14, 32, 2, 216, 169, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 138, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 139, 221, 14, 32, 60, 0, 14, 32, 2, 216, 170, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 140, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 141, 221, 14, 32, 60, 0, 14, 32, 2, 216, 171, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 142, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 143, 221, 14, 32, 60, 0, 14, 32, 2, 216, 172, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 144, 221, 14, 32, 60, 0, 14, 32, 2, 216, 173, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 145, 221, 14, 32, 60, 0, 14, 32, 2, 216, 174, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 146, 221, 14, 32, 60, 0, 14, 32, 2, 216, 175, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 176, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 147, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 148, 221, 14, 32, 60, 0, 14, 32, 2, 216, 177, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 149, 221, 14, 32, 60, 0, 14, 32, 2, 216, 178, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 150, 221, 14, 32, 60, 0, 14, 32, 2, 216, 179, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 151, 221, 14, 32, 60, 0, 14, 32, 2, 216, 180, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 152, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 153, 221, 14, 32, 60, 0, 14, 32, 2, 216, 181, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 154, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 155, 221, 14, 32, 60, 0, 14, 32, 2, 216, 182, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 156, 221, 14, 32, 60, 0, 14, 32, 2, 216, 183, 221, 14, 32, 60, 0, 60, 0, 14, 32, 2, 216, 157, 221, 14, 32, 60, 0, 14, 32, 2, 216, 190, 221, 14, 32, 60, 0, 14, 32, 2, 216, 191, 221, 14, 32, 60, 0, 14, 32, 2, 216, 158, 221, 14, 32, 60, 0, 14, 32, 2, 216, 159, 221, 14, 32, 60, 0, 17, 216, 0, 220, 60, 0, 17, 216, 1, 220, 60, 0, 17, 216, 2, 220, 60, 0, 17, 216, 3, 220, 60, 0, 17, 216, 4, 220, 60, 0, 17, 216, 5, 220, 60, 0, 17, 216, 6, 220, 60, 0, 17, 216, 7, 220, 60, 0, 17, 216, 8, 220, 60, 0, 17, 216, 9, 220, 60, 0, 17, 216, 10, 220, 60, 0, 17, 216, 11, 220, 60, 0, 17, 216, 12, 220, 60, 0, 17, 216, 13, 220, 60, 0, 17, 216, 14, 220, 60, 0, 17, 216, 15, 220, 60, 0, 17, 216, 16, 220, 60, 0, 17, 216, 17, 220, 60, 0, 17, 216, 18, 220, 60, 0, 17, 216, 19, 220, 60, 0, 17, 216, 20, 220, 60, 0, 17, 216, 21, 220, 60, 0, 17, 216, 22, 220, 60, 0, 17, 216, 23, 220, 60, 0, 17, 216, 24, 220, 60, 0, 17, 216, 25, 220, 60, 0, 17, 216, 26, 220, 60, 0, 17, 216, 27, 220, 60, 0, 17, 216, 28, 220, 60, 0, 17, 216, 29, 220, 60, 0, 17, 216, 30, 220, 60, 0, 17, 216, 31, 220, 60, 0, 17, 216, 32, 220, 60, 0, 17, 216, 33, 220, 60, 0, 17, 216, 34, 220, 60, 0, 17, 216, 35, 220, 60, 0, 17, 216, 36, 220, 60, 0, 17, 216, 37, 220, 60, 0, 17, 216, 38, 220, 60, 0, 17, 216, 39, 220, 60, 0, 17, 216, 40, 220, 60, 0, 17, 216, 41, 220, 60, 0, 17, 216, 42, 220, 60, 0, 17, 216, 43, 220, 60, 0, 17, 216, 44, 220, 60, 0, 17, 216, 45, 220, 60, 0, 17, 216, 46, 220, 60, 0, 17, 216, 47, 220, 60, 0, 17, 216, 48, 220, 60, 0, 17, 216, 49, 220, 60, 0, 17, 216, 50, 220, 60, 0, 17, 216, 51, 220, 60, 0, 17, 216, 52, 220, 60, 0, 17, 216, 53, 220, 60, 0, 17, 216, 54, 220, 60, 0, 17, 216, 55, 220, 60, 0, 17, 216, 56, 220, 60, 0, 17, 216, 57, 220, 60, 0, 17, 216, 58, 220, 60, 0, 17, 216, 59, 220, 60, 0, 17, 216, 60, 220, 60, 0, 17, 216, 61, 220, 60, 0, 17, 216, 62, 220, 60, 0, 17, 216, 63, 220, 60, 0, 17, 216, 64, 220, 60, 0, 17, 216, 65, 220, 60, 0, 17, 216, 66, 220, 60, 0, 17, 216, 67, 220, 60, 0, 17, 216, 68, 220, 60, 0, 17, 216, 69, 220, 60, 0, 17, 216, 70, 220, 60, 0, 17, 216, 71, 220, 60, 0, 17, 216, 72, 220, 60, 0, 17, 216, 73, 220, 60, 0, 17, 216, 74, 220, 60, 0, 17, 216, 75, 220, 60, 0, 17, 216, 76, 220, 60, 0, 17, 216, 77, 220, 60, 0, 17, 216, 78, 220, 60, 0, 17, 216, 79, 220, 60, 0, 17, 216, 80, 220, 60, 0, 17, 216, 81, 220, 60, 0, 17, 216, 82, 220, 60, 0, 17, 216, 83, 220, 60, 0, 17, 216, 84, 220, 60, 0, 17, 216, 85, 220, 60, 0, 17, 216, 86, 220, 60, 0, 17, 216, 87, 220, 60, 0, 17, 216, 88, 220, 60, 0, 17, 216, 89, 220, 60, 0, 17, 216, 90, 220, 60, 0, 17, 216, 91, 220, 60, 0, 17, 216, 92, 220, 60, 0, 17, 216, 93, 220, 60, 0, 17, 216, 94, 220, 60, 0, 17, 216, 95, 220, 60, 0, 17, 216, 96, 220, 60, 0, 17, 216, 97, 220, 60, 0, 17, 216, 98, 220, 60, 0, 17, 216, 99, 220, 60, 0, 17, 216, 100, 220, 60, 0, 17, 216, 101, 220, 60, 0, 17, 216, 102, 220, 60, 0, 17, 216, 103, 220, 60, 0, 17, 216, 104, 220, 60, 0, 17, 216, 105, 220, 60, 0, 17, 216, 106, 220, 60, 0, 17, 216, 107, 220, 60, 0, 17, 216, 108, 220, 60, 0, 17, 216, 109, 220, 60, 0, 17, 216, 110, 220, 60, 0, 17, 216, 111, 220, 60, 0, 17, 216, 112, 220, 60, 0, 17, 216, 113, 220, 60, 0, 17, 216, 114, 220, 60, 0, 17, 216, 115, 220, 60, 0, 17, 216, 116, 220, 60, 0, 17, 216, 117, 220, 60, 0, 17, 216, 118, 220, 60, 0, 17, 216, 119, 220, 60, 0, 17, 216, 120, 220, 60, 0, 17, 216, 121, 220, 60, 0, 17, 216, 122, 220, 60, 0, 17, 216, 123, 220, 60, 0, 17, 216, 124, 220, 60, 0, 17, 216, 125, 220, 60, 0, 17, 216, 126, 220, 60, 0, 17, 216, 127, 220, 60, 0, 17, 216, 128, 220, 60, 0, 17, 216, 129, 220, 60, 0, 17, 216, 130, 220, 60, 0, 17, 216, 131, 220, 60, 0, 17, 216, 132, 220, 60, 0, 17, 216, 133, 220, 60, 0, 17, 216, 134, 220, 60, 0, 17, 216, 135, 220, 60, 0, 17, 216, 136, 220, 60, 0, 17, 216, 137, 220, 60, 0, 17, 216, 138, 220, 60, 0, 17, 216, 139, 220, 60, 0, 17, 216, 140, 220, 60, 0, 17, 216, 141, 220, 60, 0, 17, 216, 142, 220, 60, 0, 17, 216, 143, 220, 60, 0, 17, 216, 144, 220, 60, 0, 17, 216, 145, 220, 60, 0, 17, 216, 146, 220, 60, 0, 17, 216, 147, 220, 60, 0, 17, 216, 148, 220, 60, 0, 17, 216, 149, 220, 60, 0, 17, 216, 150, 220, 60, 0, 17, 216, 151, 220, 60, 0, 17, 216, 152, 220, 60, 0, 17, 216, 153, 220, 60, 0, 17, 216, 154, 220, 60, 0, 17, 216, 155, 220, 60, 0, 17, 216, 156, 220, 60, 0, 17, 216, 157, 220, 60, 0, 17, 216, 158, 220, 60, 0, 17, 216, 159, 220, 60, 0, 17, 216, 160, 220, 60, 0, 17, 216, 161, 220, 60, 0, 17, 216, 162, 220, 60, 0, 17, 216, 163, 220, 60, 0, 17, 216, 164, 220, 60, 0, 17, 216, 165, 220, 60, 0, 17, 216, 166, 220, 60, 0, 17, 216, 167, 220, 60, 0, 17, 216, 168, 220, 60, 0, 17, 216, 169, 220, 60, 0, 17, 216, 170, 220, 60, 0, 17, 216, 171, 220, 60, 0, 17, 216, 172, 220, 60, 0, 17, 216, 173, 220, 60, 0, 17, 216, 174, 220, 60, 0, 17, 216, 175, 220, 60, 0, 17, 216, 176, 220, 60, 0, 17, 216, 177, 220, 60, 0, 17, 216, 178, 220, 60, 0, 17, 216, 179, 220, 60, 0, 17, 216, 180, 220, 60, 0, 17, 216, 181, 220, 60, 0, 17, 216, 182, 220, 60, 0, 17, 216, 183, 220, 60, 0, 17, 216, 184, 220, 60, 0, 17, 216, 185, 220, 60, 0, 17, 216, 186, 220, 60, 0, 17, 216, 187, 220, 60, 0, 17, 216, 188, 220, 60, 0, 17, 216, 189, 220, 60, 0, 17, 216, 190, 220, 60, 0, 17, 216, 191, 220, 60, 0, 17, 216, 192, 220, 60, 0, 17, 216, 193, 220, 60, 0, 17, 216, 194, 220, 60, 0, 17, 216, 195, 220, 60, 0, 17, 216, 196, 220, 60, 0, 17, 216, 197, 220, 60, 0, 17, 216, 198, 220, 60, 0, 17, 216, 199, 220, 60, 0, 17, 216, 200, 220, 60, 0, 17, 216, 201, 220, 60, 0, 17, 216, 202, 220, 60, 0, 17, 216, 203, 220, 60, 0, 17, 216, 204, 220, 60, 0, 17, 216, 205, 220, 60, 0, 17, 216, 206, 220, 60, 0, 17, 216, 207, 220, 60, 0, 17, 216, 208, 220, 60, 0, 17, 216, 209, 220, 60, 0, 17, 216, 210, 220, 60, 0, 17, 216, 211, 220, 60, 0, 17, 216, 212, 220, 60, 0, 17, 216, 213, 220, 60, 0, 17, 216, 214, 220, 60, 0, 17, 216, 215, 220, 60, 0, 17, 216, 216, 220, 60, 0, 17, 216, 217, 220, 60, 0, 17, 216, 218, 220, 60, 0, 17, 216, 219, 220, 60, 0, 17, 216, 220, 220, 60, 0, 17, 216, 221, 220, 60, 0, 17, 216, 222, 220, 60, 0, 17, 216, 223, 220, 60, 0, 17, 216, 224, 220, 60, 0, 17, 216, 225, 220, 60, 0, 17, 216, 226, 220, 60, 0, 17, 216, 227, 220, 60, 0, 17, 216, 228, 220, 60, 0, 17, 216, 229, 220, 60, 0, 17, 216, 230, 220, 60, 0, 17, 216, 231, 220, 60, 0, 17, 216, 232, 220, 60, 0, 17, 216, 233, 220, 60, 0, 17, 216, 234, 220, 60, 0, 17, 216, 235, 220, 60, 0, 17, 216, 236, 220, 60, 0, 17, 216, 237, 220, 60, 0, 17, 216, 238, 220, 60, 0, 17, 216, 239, 220, 60, 0, 17, 216, 240, 220, 60, 0, 17, 216, 241, 220, 60, 0, 17, 216, 242, 220, 60, 0, 17, 216, 243, 220, 60, 0, 17, 216, 244, 220, 60, 0, 17, 216, 245, 220, 60, 0, 17, 216, 246, 220, 60, 0, 17, 216, 247, 220, 60, 0, 17, 216, 248, 220, 60, 0, 17, 216, 249, 220, 60, 0, 17, 216, 250, 220, 60, 0, 17, 216, 251, 220, 60, 0, 17, 216, 252, 220, 60, 0, 17, 216, 253, 220, 60, 0, 17, 216, 254, 220, 60, 0, 17, 216, 255, 220, 60, 0, 17, 216, 0, 221, 60, 0, 17, 216, 1, 221, 60, 0, 17, 216, 2, 221, 60, 0, 17, 216, 3, 221, 60, 0, 17, 216, 4, 221, 60, 0, 17, 216, 5, 221, 60, 0, 17, 216, 6, 221, 60, 0, 17, 216, 7, 221, 60, 0, 17, 216, 8, 221, 60, 0, 17, 216, 9, 221, 60, 0, 17, 216, 10, 221, 60, 0, 17, 216, 11, 221, 60, 0, 17, 216, 12, 221, 60, 0, 17, 216, 13, 221, 60, 0, 17, 216, 14, 221, 60, 0, 17, 216, 15, 221, 60, 0, 17, 216, 16, 221, 60, 0, 17, 216, 17, 221, 60, 0, 17, 216, 18, 221, 60, 0, 17, 216, 19, 221, 60, 0, 17, 216, 20, 221, 60, 0, 17, 216, 21, 221, 60, 0, 17, 216, 22, 221, 60, 0, 17, 216, 23, 221, 60, 0, 17, 216, 24, 221, 60, 0, 17, 216, 25, 221, 60, 0, 17, 216, 26, 221, 60, 0, 17, 216, 27, 221, 60, 0, 17, 216, 28, 221, 60, 0, 17, 216, 29, 221, 60, 0, 17, 216, 30, 221, 60, 0, 17, 216, 31, 221, 60, 0, 17, 216, 32, 221, 60, 0, 17, 216, 33, 221, 60, 0, 17, 216, 34, 221, 60, 0, 17, 216, 35, 221, 60, 0, 17, 216, 36, 221, 60, 0, 17, 216, 37, 221, 60, 0, 17, 216, 38, 221, 60, 0, 17, 216, 39, 221, 60, 0, 17, 216, 40, 221, 60, 0, 17, 216, 41, 221, 60, 0, 17, 216, 42, 221, 60, 0, 17, 216, 43, 221, 60, 0, 17, 216, 44, 221, 60, 0, 17, 216, 45, 221, 60, 0, 17, 216, 46, 221, 60, 0, 17, 216, 47, 221, 60, 0, 17, 216, 48, 221, 60, 0, 17, 216, 49, 221, 60, 0, 17, 216, 50, 221, 60, 0, 17, 216, 51, 221, 60, 0, 17, 216, 52, 221, 60, 0, 17, 216, 53, 221, 60, 0, 17, 216, 54, 221, 60, 0, 17, 216, 55, 221, 60, 0, 17, 216, 56, 221, 60, 0, 17, 216, 57, 221, 60, 0, 17, 216, 58, 221, 60, 0, 17, 216, 59, 221, 60, 0, 17, 216, 60, 221, 60, 0, 17, 216, 61, 221, 60, 0, 17, 216, 62, 221, 60, 0, 17, 216, 63, 221, 60, 0, 17, 216, 64, 221, 60, 0, 17, 216, 65, 221, 60, 0, 17, 216, 66, 221, 60, 0, 17, 216, 67, 221, 60, 0, 17, 216, 68, 221, 60, 0, 17, 216, 69, 221, 60, 0, 17, 216, 70, 221, 60, 0, 17, 216, 71, 221, 60, 0, 17, 216, 72, 221, 60, 0, 17, 216, 73, 221, 60, 0, 17, 216, 74, 221, 60, 0, 17, 216, 75, 221, 60, 0, 17, 216, 76, 221, 60, 0, 17, 216, 77, 221, 60, 0, 17, 216, 78, 221, 60, 0, 17, 216, 79, 221, 60, 0, 17, 216, 80, 221, 60, 0, 17, 216, 81, 221, 60, 0, 17, 216, 82, 221, 60, 0, 17, 216, 83, 221, 60, 0, 17, 216, 84, 221, 60, 0, 17, 216, 85, 221, 60, 0, 17, 216, 86, 221, 60, 0, 17, 216, 87, 221, 60, 0, 17, 216, 88, 221, 60, 0, 17, 216, 89, 221, 60, 0, 17, 216, 90, 221, 60, 0, 17, 216, 91, 221, 60, 0, 17, 216, 92, 221, 60, 0, 17, 216, 93, 221, 60, 0, 17, 216, 94, 221, 60, 0, 17, 216, 95, 221, 60, 0, 17, 216, 96, 221, 60, 0, 17, 216, 97, 221, 60, 0, 17, 216, 98, 221, 60, 0, 17, 216, 99, 221, 60, 0, 17, 216, 100, 221, 60, 0, 17, 216, 101, 221, 60, 0, 17, 216, 102, 221, 60, 0, 17, 216, 103, 221, 60, 0, 17, 216, 104, 221, 60, 0, 17, 216, 105, 221, 60, 0, 17, 216, 106, 221, 60, 0, 17, 216, 107, 221, 60, 0, 17, 216, 108, 221, 60, 0, 17, 216, 109, 221, 60, 0, 17, 216, 110, 221, 60, 0, 17, 216, 111, 221, 60, 0, 17, 216, 112, 221, 60, 0, 17, 216, 113, 221, 60, 0, 17, 216, 114, 221, 60, 0, 17, 216, 115, 221, 60, 0, 17, 216, 116, 221, 60, 0, 17, 216, 117, 221, 60, 0, 17, 216, 118, 221, 60, 0, 17, 216, 119, 221, 60, 0, 17, 216, 120, 221, 60, 0, 17, 216, 121, 221, 60, 0, 17, 216, 122, 221, 60, 0, 17, 216, 123, 221, 60, 0, 17, 216, 124, 221, 60, 0, 17, 216, 125, 221, 60, 0, 17, 216, 126, 221, 60, 0, 17, 216, 127, 221, 60, 0, 17, 216, 128, 221, 60, 0, 17, 216, 129, 221, 60, 0, 17, 216, 130, 221, 60, 0, 17, 216, 131, 221, 60, 0, 17, 216, 132, 221, 60, 0, 17, 216, 133, 221, 60, 0, 17, 216, 134, 221, 60, 0, 17, 216, 135, 221, 60, 0, 17, 216, 136, 221, 60, 0, 17, 216, 137, 221, 60, 0, 17, 216, 138, 221, 60, 0, 17, 216, 139, 221, 60, 0, 17, 216, 140, 221, 60, 0, 17, 216, 141, 221, 60, 0, 17, 216, 142, 221, 60, 0, 17, 216, 143, 221, 60, 0, 17, 216, 144, 221, 60, 0, 17, 216, 145, 221, 60, 0, 17, 216, 146, 221, 60, 0, 17, 216, 147, 221, 60, 0, 17, 216, 148, 221, 60, 0, 17, 216, 149, 221, 60, 0, 17, 216, 150, 221, 60, 0, 17, 216, 151, 221, 60, 0, 17, 216, 152, 221, 60, 0, 17, 216, 153, 221, 60, 0, 17, 216, 154, 221, 60, 0, 17, 216, 155, 221, 60, 0, 17, 216, 156, 221, 60, 0, 17, 216, 157, 221, 60, 0, 17, 216, 158, 221, 60, 0, 17, 216, 159, 221, 60, 0, 17, 216, 160, 221, 60, 0, 17, 216, 161, 221, 60, 0, 17, 216, 162, 221, 60, 0, 17, 216, 163, 221, 60, 0, 17, 216, 164, 221, 60, 0, 17, 216, 165, 221, 60, 0, 17, 216, 166, 221, 60, 0, 17, 216, 167, 221, 60, 0, 17, 216, 168, 221, 60, 0, 17, 216, 169, 221, 60, 0, 17, 216, 170, 221, 60, 0, 17, 216, 171, 221, 60, 0, 17, 216, 172, 221, 60, 0, 17, 216, 173, 221, 60, 0, 17, 216, 174, 221, 60, 0, 17, 216, 175, 221, 60, 0, 17, 216, 176, 221, 60, 0, 17, 216, 177, 221, 60, 0, 17, 216, 178, 221, 60, 0, 17, 216, 179, 221, 60, 0, 17, 216, 180, 221, 60, 0, 17, 216, 181, 221, 60, 0, 17, 216, 182, 221, 60, 0, 17, 216, 183, 221, 60, 0, 17, 216, 184, 221, 60, 0, 17, 216, 185, 221, 60, 0, 17, 216, 186, 221, 60, 0, 17, 216, 187, 221, 60, 0, 17, 216, 188, 221, 60, 0, 17, 216, 189, 221, 60, 0, 17, 216, 190, 221, 60, 0, 17, 216, 191, 221, 60, 0, 17, 216, 192, 221, 60, 0, 17, 216, 193, 221, 60, 0, 17, 216, 194, 221, 60, 0, 17, 216, 195, 221, 60, 0, 17, 216, 196, 221, 60, 0, 17, 216, 197, 221, 60, 0, 17, 216, 198, 221, 60, 0, 17, 216, 199, 221, 60, 0, 17, 216, 200, 221, 60, 0, 17, 216, 201, 221, 60, 0, 17, 216, 202, 221, 60, 0, 17, 216, 203, 221, 60, 0, 17, 216, 204, 221, 60, 0, 17, 216, 205, 221, 60, 0, 17, 216, 206, 221, 60, 0, 17, 216, 207, 221, 60, 0, 17, 216, 208, 221, 60, 0, 17, 216, 209, 221, 60, 0, 17, 216, 210, 221, 60, 0, 17, 216, 211, 221, 60, 0, 17, 216, 212, 221, 60, 0, 17, 216, 213, 221, 60, 0, 17, 216, 214, 221, 60, 0, 17, 216, 215, 221, 60, 0, 17, 216, 216, 221, 60, 0, 17, 216, 217, 221, 60, 0, 17, 216, 218, 221, 60, 0, 17, 216, 219, 221, 60, 0, 17, 216, 220, 221, 60, 0, 17, 216, 221, 221, 60, 0, 17, 216, 222, 221, 60, 0, 17, 216, 223, 221, 60, 0, 17, 216, 224, 221, 60, 0, 17, 216, 225, 221, 60, 0, 17, 216, 226, 221, 60, 0, 17, 216, 227, 221, 60, 0, 17, 216, 228, 221, 60, 0, 17, 216, 229, 221, 60, 0, 17, 216, 230, 221, 60, 0, 17, 216, 231, 221, 60, 0, 17, 216, 232, 221, 60, 0, 17, 216, 233, 221, 60, 0, 17, 216, 234, 221, 60, 0, 17, 216, 235, 221, 60, 0, 17, 216, 236, 221, 60, 0, 17, 216, 237, 221, 60, 0, 17, 216, 238, 221, 60, 0, 17, 216, 239, 221, 60, 0, 17, 216, 240, 221, 60, 0, 17, 216, 241, 221, 60, 0, 17, 216, 242, 221, 60, 0, 17, 216, 243, 221, 60, 0, 17, 216, 244, 221, 60, 0, 17, 216, 245, 221, 60, 0, 17, 216, 246, 221, 60, 0, 17, 216, 247, 221, 60, 0, 17, 216, 248, 221, 60, 0, 17, 216, 249, 221, 60, 0, 17, 216, 250, 221, 60, 0, 17, 216, 251, 221, 60, 0, 17, 216, 252, 221, 60, 0, 17, 216, 253, 221, 60, 0, 17, 216, 254, 221, 60, 0, 17, 216, 255, 221, 60, 0, 17, 216, 0, 222, 60, 0, 17, 216, 1, 222, 60, 0, 17, 216, 2, 222, 60, 0, 17, 216, 3, 222, 60, 0, 17, 216, 4, 222, 60, 0, 17, 216, 5, 222, 60, 0, 17, 216, 6, 222, 60, 0, 17, 216, 7, 222, 60, 0, 17, 216, 8, 222, 60, 0, 17, 216, 9, 222, 60, 0, 17, 216, 10, 222, 60, 0, 17, 216, 11, 222, 60, 0, 17, 216, 12, 222, 60, 0, 17, 216, 13, 222, 60, 0, 17, 216, 14, 222, 60, 0, 17, 216, 15, 222, 60, 0, 17, 216, 16, 222, 60, 0, 17, 216, 17, 222, 60, 0, 17, 216, 18, 222, 60, 0, 17, 216, 19, 222, 60, 0, 17, 216, 20, 222, 60, 0, 17, 216, 21, 222, 60, 0, 17, 216, 22, 222, 60, 0, 17, 216, 23, 222, 60, 0, 17, 216, 24, 222, 60, 0, 17, 216, 25, 222, 60, 0, 17, 216, 26, 222, 60, 0, 17, 216, 27, 222, 60, 0, 17, 216, 28, 222, 60, 0, 17, 216, 29, 222, 60, 0, 17, 216, 30, 222, 60, 0, 17, 216, 31, 222, 60, 0, 17, 216, 32, 222, 60, 0, 17, 216, 33, 222, 60, 0, 17, 216, 34, 222, 60, 0, 17, 216, 35, 222, 60, 0, 17, 216, 36, 222, 60, 0, 17, 216, 37, 222, 60, 0, 17, 216, 38, 222, 60, 0, 17, 216, 39, 222, 60, 0, 17, 216, 40, 222, 60, 0, 17, 216, 41, 222, 60, 0, 17, 216, 42, 222, 60, 0, 17, 216, 43, 222, 60, 0, 17, 216, 44, 222, 60, 0, 17, 216, 45, 222, 60, 0, 17, 216, 46, 222, 60, 0, 17, 216, 47, 222, 60, 0, 17, 216, 48, 222, 60, 0, 17, 216, 49, 222, 60, 0, 17, 216, 50, 222, 60, 0, 17, 216, 51, 222, 60, 0, 17, 216, 52, 222, 60, 0, 17, 216, 53, 222, 60, 0, 17, 216, 54, 222, 60, 0, 17, 216, 55, 222, 60, 0, 17, 216, 56, 222, 60, 0, 17, 216, 57, 222, 60, 0, 17, 216, 58, 222, 60, 0, 17, 216, 59, 222, 60, 0, 17, 216, 60, 222, 60, 0, 17, 216, 61, 222, 60, 0, 17, 216, 62, 222, 60, 0, 17, 216, 63, 222, 60, 0, 17, 216, 64, 222, 60, 0, 17, 216, 65, 222, 60, 0, 17, 216, 66, 222, 60, 0, 17, 216, 67, 222, 60, 0, 17, 216, 68, 222, 60, 0, 17, 216, 69, 222, 60, 0, 17, 216, 70, 222, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 0, 47, 60, 0, 60, 0, 60, 0, 128, 50, 60, 0, 60, 0, 60, 0, 146, 49, 60, 0, 60, 0, 60, 0, 60, 216, 41, 222, 38, 0, 1, 78, 60, 0, 60, 0, 60, 0, 156, 49, 38, 0, 3, 78, 60, 0, 60, 0, 60, 0, 134, 50, 38, 0, 9, 78, 60, 0, 60, 0, 60, 0, 130, 50, 60, 0, 60, 0, 60, 0, 148, 49, 60, 0, 60, 0, 60, 0, 60, 216, 42, 222, 38, 0, 10, 78, 60, 0, 60, 0, 60, 0, 164, 50, 60, 0, 60, 0, 60, 0, 150, 49, 38, 0, 11, 78, 60, 0, 60, 0, 60, 0, 166, 50, 60, 0, 60, 0, 60, 0, 152, 49, 38, 0, 25, 78, 60, 0, 60, 0, 60, 0, 155, 49, 38, 0, 40, 78, 60, 0, 60, 0, 60, 0, 1, 47, 38, 0, 44, 78, 60, 0, 60, 0, 60, 0, 166, 46, 38, 0, 45, 78, 60, 0, 60, 0, 60, 0, 165, 50, 60, 0, 60, 0, 60, 0, 151, 49, 60, 0, 60, 0, 60, 0, 60, 216, 45, 222, 38, 0, 54, 78, 60, 0, 60, 0, 60, 0, 2, 47, 60, 0, 60, 0, 128, 46, 38, 0, 63, 78, 60, 0, 60, 0, 60, 0, 3, 47, 38, 0, 89, 78, 60, 0, 60, 0, 60, 0, 4, 47, 60, 0, 60, 0, 60, 0, 154, 49, 60, 0, 60, 0, 132, 46, 38, 0, 90, 78, 60, 0, 60, 0, 60, 0, 131, 46, 38, 0, 91, 78, 60, 0, 60, 0, 60, 0, 130, 46, 38, 0, 93, 78, 60, 0, 60, 0, 60, 0, 136, 50, 38, 0, 133, 78, 60, 0, 60, 0, 60, 0, 5, 47, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 6, 47, 60, 0, 60, 0, 60, 0, 129, 50, 60, 0, 60, 0, 60, 0, 147, 49, 60, 0, 60, 0, 60, 0, 60, 216, 20, 222, 38, 0, 148, 78, 60, 0, 60, 0, 60, 0, 132, 50, 38, 0, 160, 78, 60, 0, 60, 0, 60, 0, 7, 47, 38, 0, 164, 78, 60, 0, 60, 0, 60, 0, 60, 216, 24, 222, 38, 0, 186, 78, 60, 0, 60, 0, 60, 0, 8, 47, 60, 0, 60, 0, 60, 0, 159, 49, 38, 0, 187, 78, 60, 0, 60, 0, 60, 0, 133, 46, 38, 0, 228, 78, 60, 0, 255, 50, 47, 0, 140, 84, 38, 0, 1, 79, 60, 0, 60, 0, 60, 0, 173, 50, 38, 0, 17, 79, 60, 0, 60, 0, 60, 0, 161, 50, 38, 0, 42, 81, 60, 0, 60, 0, 60, 0, 157, 50, 38, 0, 63, 81, 60, 0, 60, 0, 60, 0, 9, 47, 38, 0, 101, 81, 60, 0, 60, 0, 60, 0, 10, 47, 38, 0, 107, 81, 60, 0, 60, 0, 60, 0, 11, 47, 60, 0, 60, 0, 60, 0, 135, 50, 38, 0, 109, 81, 60, 0, 60, 0, 60, 0, 133, 50, 38, 0, 130, 81, 60, 0, 60, 0, 60, 0, 12, 47, 60, 0, 60, 0, 134, 46, 38, 0, 141, 81, 60, 0, 60, 0, 60, 0, 60, 216, 30, 222, 38, 0, 150, 81, 60, 0, 60, 0, 60, 0, 13, 47, 38, 0, 153, 81, 60, 0, 60, 0, 60, 0, 162, 50, 38, 0, 171, 81, 60, 0, 60, 0, 60, 0, 14, 47, 38, 0, 224, 81, 60, 0, 60, 0, 60, 0, 15, 47, 60, 0, 60, 0, 135, 46, 38, 0, 245, 81, 60, 0, 60, 0, 60, 0, 16, 47, 38, 0, 0, 82, 60, 0, 60, 0, 60, 0, 17, 47, 60, 0, 60, 0, 136, 46, 38, 0, 2, 82, 60, 0, 60, 0, 60, 0, 137, 46, 38, 0, 29, 82, 60, 0, 60, 0, 60, 0, 60, 216, 32, 222, 38, 0, 77, 82, 60, 0, 60, 0, 60, 0, 60, 216, 28, 222, 38, 0, 114, 82, 60, 0, 60, 0, 60, 0, 60, 216, 57, 222, 38, 0, 155, 82, 60, 0, 60, 0, 60, 0, 18, 47, 38, 0, 180, 82, 60, 0, 60, 0, 60, 0, 152, 50, 38, 0, 249, 82, 60, 0, 60, 0, 60, 0, 19, 47, 38, 0, 21, 83, 60, 0, 60, 0, 60, 0, 20, 47, 38, 0, 26, 83, 60, 0, 60, 0, 60, 0, 21, 47, 38, 0, 56, 83, 60, 0, 60, 0, 60, 0, 22, 47, 38, 0, 59, 83, 60, 0, 60, 0, 60, 0, 169, 50, 38, 0, 65, 83, 60, 0, 60, 0, 60, 0, 23, 47, 61, 0, 56, 48, 60, 0, 60, 0, 60, 0, 137, 50, 38, 0, 68, 83, 60, 0, 60, 0, 60, 0, 57, 48, 38, 0, 69, 83, 60, 0, 60, 0, 60, 0, 58, 48, 38, 0, 84, 83, 60, 0, 60, 0, 60, 0, 175, 50, 38, 0, 92, 83, 60, 0, 60, 0, 60, 0, 24, 47, 60, 0, 60, 0, 138, 46, 38, 0, 105, 83, 60, 0, 60, 0, 60, 0, 25, 47, 60, 0, 60, 0, 139, 46, 38, 0, 112, 83, 60, 0, 60, 0, 60, 0, 158, 50, 38, 0, 130, 83, 60, 0, 60, 0, 60, 0, 26, 47, 60, 0, 60, 0, 129, 46, 38, 0, 182, 83, 60, 0, 60, 0, 60, 0, 27, 47, 38, 0, 200, 83, 60, 0, 60, 0, 60, 0, 28, 47, 38, 0, 204, 83, 60, 0, 60, 0, 60, 0, 60, 216, 18, 222, 38, 0, 227, 83, 60, 0, 60, 0, 60, 0, 29, 47, 38, 0, 239, 83, 60, 0, 60, 0, 60, 0, 60, 216, 81, 222, 38, 0, 243, 83, 60, 0, 60, 0, 60, 0, 168, 50, 60, 0, 60, 0, 60, 0, 60, 216, 46, 222, 38, 0, 8, 84, 60, 0, 60, 0, 60, 0, 60, 216, 52, 222, 38, 0, 13, 84, 60, 0, 60, 0, 60, 0, 148, 50, 38, 0, 57, 84, 60, 0, 60, 0, 60, 0, 60, 216, 37, 222, 38, 0, 79, 85, 60, 0, 60, 0, 60, 0, 68, 50, 38, 0, 182, 85, 60, 0, 60, 0, 60, 0, 60, 216, 58, 222, 38, 0, 215, 86, 60, 0, 60, 0, 60, 0, 30, 47, 38, 0, 219, 86, 60, 0, 60, 0, 60, 0, 131, 50, 60, 0, 60, 0, 60, 0, 149, 49, 38, 0, 31, 87, 60, 0, 60, 0, 60, 0, 31, 47, 60, 0, 60, 0, 60, 0, 143, 50, 38, 0, 48, 87, 60, 0, 60, 0, 60, 0, 158, 49, 38, 0, 235, 88, 60, 0, 60, 0, 60, 0, 32, 47, 38, 0, 240, 88, 60, 0, 60, 0, 60, 0, 60, 216, 36, 222, 38, 0, 2, 89, 60, 0, 60, 0, 60, 0, 33, 47, 38, 0, 10, 89, 60, 0, 60, 0, 60, 0, 34, 47, 38, 0, 21, 89, 60, 0, 60, 0, 60, 0, 35, 47, 38, 0, 26, 89, 60, 0, 60, 0, 60, 0, 60, 216, 21, 222, 38, 0, 28, 89, 60, 0, 60, 0, 60, 0, 176, 50, 38, 0, 39, 89, 60, 0, 60, 0, 60, 0, 36, 47, 60, 0, 60, 0, 60, 0, 125, 51, 47, 0, 99, 107, 38, 0, 41, 89, 60, 0, 60, 0, 60, 0, 157, 49, 60, 0, 60, 0, 60, 0, 60, 216, 23, 222, 38, 0, 115, 89, 60, 0, 60, 0, 60, 0, 37, 47, 60, 0, 60, 0, 60, 0, 155, 50, 38, 0, 80, 91, 60, 0, 60, 0, 60, 0, 38, 47, 38, 0, 87, 91, 60, 0, 60, 0, 60, 0, 60, 216, 17, 222, 38, 0, 102, 91, 60, 0, 60, 0, 60, 0, 171, 50, 38, 0, 128, 91, 60, 0, 60, 0, 60, 0, 39, 47, 38, 0, 151, 91, 60, 0, 60, 0, 60, 0, 170, 50, 38, 0, 248, 91, 60, 0, 60, 0, 60, 0, 40, 47, 38, 0, 15, 92, 60, 0, 60, 0, 60, 0, 41, 47, 60, 0, 60, 0, 140, 46, 60, 0, 60, 0, 141, 46, 38, 0, 34, 92, 60, 0, 60, 0, 60, 0, 144, 46, 61, 0, 42, 47, 60, 0, 60, 0, 142, 46, 38, 0, 35, 92, 60, 0, 60, 0, 60, 0, 143, 46, 60, 0, 60, 0, 145, 46, 38, 0, 56, 92, 60, 0, 60, 0, 60, 0, 43, 47, 38, 0, 110, 92, 60, 0, 60, 0, 60, 0, 44, 47, 38, 0, 113, 92, 60, 0, 60, 0, 60, 0, 45, 47, 38, 0, 219, 93, 60, 0, 60, 0, 60, 0, 46, 47, 38, 0, 229, 93, 60, 0, 60, 0, 60, 0, 47, 47, 38, 0, 230, 93, 60, 0, 60, 0, 60, 0, 167, 50, 60, 0, 60, 0, 60, 0, 60, 216, 44, 222, 38, 0, 241, 93, 60, 0, 60, 0, 60, 0, 48, 47, 38, 0, 243, 93, 60, 0, 60, 0, 60, 0, 146, 46, 38, 0, 254, 93, 60, 0, 60, 0, 60, 0, 49, 47, 38, 0, 114, 94, 60, 0, 60, 0, 60, 0, 50, 47, 38, 0, 115, 94, 60, 0, 123, 51, 47, 0, 16, 98, 38, 0, 122, 94, 60, 0, 60, 0, 60, 0, 147, 46, 61, 0, 51, 47, 38, 0, 124, 94, 60, 0, 60, 0, 60, 0, 69, 50, 38, 0, 127, 94, 60, 0, 60, 0, 60, 0, 52, 47, 38, 0, 244, 94, 60, 0, 60, 0, 60, 0, 53, 47, 38, 0, 254, 94, 60, 0, 60, 0, 60, 0, 54, 47, 38, 0, 11, 95, 60, 0, 60, 0, 60, 0, 55, 47, 38, 0, 19, 95, 60, 0, 60, 0, 60, 0, 56, 47, 38, 0, 80, 95, 60, 0, 60, 0, 60, 0, 57, 47, 60, 0, 60, 0, 149, 46, 38, 0, 81, 95, 60, 0, 60, 0, 60, 0, 148, 46, 38, 0, 97, 95, 60, 0, 60, 0, 60, 0, 58, 47, 38, 0, 115, 95, 60, 0, 60, 0, 60, 0, 59, 47, 38, 0, 140, 95, 60, 0, 60, 0, 60, 0, 60, 216, 29, 222, 38, 0, 151, 95, 60, 0, 60, 0, 60, 0, 60, 216, 80, 222, 38, 0, 195, 95, 60, 0, 60, 0, 60, 0, 60, 47, 60, 0, 60, 0, 151, 46, 38, 0, 196, 95, 60, 0, 60, 0, 60, 0, 150, 46, 38, 0, 8, 98, 60, 0, 60, 0, 60, 0, 61, 47, 38, 0, 54, 98, 60, 0, 60, 0, 60, 0, 62, 47, 38, 0, 75, 98, 60, 0, 60, 0, 60, 0, 63, 47, 60, 0, 60, 0, 60, 0, 60, 216, 16, 222, 38, 0, 76, 98, 60, 0, 60, 0, 60, 0, 152, 46, 38, 0, 83, 98, 60, 0, 60, 0, 60, 0, 60, 216, 49, 222, 38, 0, 149, 98, 60, 0, 60, 0, 60, 0, 60, 216, 39, 222, 38, 0, 7, 99, 60, 0, 60, 0, 60, 0, 60, 216, 47, 222, 38, 0, 85, 99, 60, 0, 60, 0, 60, 0, 60, 216, 40, 222, 38, 0, 47, 101, 60, 0, 60, 0, 60, 0, 64, 47, 38, 0, 52, 101, 60, 0, 60, 0, 60, 0, 65, 47, 38, 0, 53, 101, 60, 0, 60, 0, 60, 0, 153, 46, 38, 0, 135, 101, 60, 0, 60, 0, 60, 0, 66, 47, 60, 0, 60, 0, 60, 0, 70, 50, 38, 0, 151, 101, 60, 0, 60, 0, 60, 0, 67, 47, 38, 0, 153, 101, 60, 0, 60, 0, 60, 0, 60, 216, 27, 222, 38, 0, 164, 101, 60, 0, 60, 0, 60, 0, 68, 47, 38, 0, 176, 101, 60, 0, 60, 0, 60, 0, 60, 216, 31, 222, 38, 0, 185, 101, 60, 0, 60, 0, 60, 0, 69, 47, 38, 0, 224, 101, 60, 0, 60, 0, 60, 0, 70, 47, 38, 0, 225, 101, 60, 0, 60, 0, 60, 0, 155, 46, 38, 0, 229, 101, 60, 0, 60, 0, 60, 0, 71, 47, 60, 0, 60, 0, 60, 0, 144, 50, 60, 0, 60, 0, 156, 46, 38, 0, 14, 102, 60, 0, 126, 51, 47, 0, 187, 108, 38, 0, 32, 102, 60, 0, 60, 0, 60, 0, 60, 216, 25, 222, 38, 0, 45, 102, 60, 0, 124, 51, 47, 0, 140, 84, 38, 0, 240, 102, 60, 0, 60, 0, 60, 0, 72, 47, 38, 0, 8, 103, 60, 0, 60, 0, 60, 0, 73, 47, 60, 0, 60, 0, 60, 0, 138, 50, 60, 0, 60, 0, 60, 0, 60, 216, 55, 222, 60, 0, 60, 0, 157, 46, 38, 0, 9, 103, 60, 0, 60, 0, 60, 0, 146, 50, 60, 0, 60, 0, 60, 0, 60, 216, 54, 222, 38, 0, 40, 103, 60, 0, 60, 0, 60, 0, 74, 47, 60, 0, 60, 0, 60, 0, 141, 50, 38, 0, 42, 104, 60, 0, 60, 0, 60, 0, 145, 50, 60, 0, 60, 0, 60, 0, 127, 51, 47, 0, 15, 95, 26, 79, 62, 121, 38, 0, 32, 107, 60, 0, 60, 0, 60, 0, 75, 47, 38, 0, 98, 107, 60, 0, 60, 0, 60, 0, 76, 47, 38, 0, 99, 107, 60, 0, 60, 0, 60, 0, 163, 50, 38, 0, 121, 107, 60, 0, 60, 0, 60, 0, 77, 47, 38, 0, 122, 107, 60, 0, 60, 0, 158, 46, 38, 0, 179, 107, 60, 0, 60, 0, 60, 0, 78, 47, 38, 0, 203, 107, 60, 0, 60, 0, 60, 0, 79, 47, 38, 0, 205, 107, 60, 0, 60, 0, 60, 0, 159, 46, 38, 0, 212, 107, 60, 0, 60, 0, 60, 0, 80, 47, 38, 0, 219, 107, 60, 0, 60, 0, 60, 0, 81, 47, 38, 0, 15, 108, 60, 0, 60, 0, 60, 0, 82, 47, 38, 0, 17, 108, 60, 0, 60, 0, 60, 0, 160, 46, 38, 0, 20, 108, 60, 0, 60, 0, 60, 0, 83, 47, 38, 0, 52, 108, 60, 0, 60, 0, 60, 0, 84, 47, 60, 0, 60, 0, 60, 0, 140, 50, 38, 0, 53, 108, 60, 0, 60, 0, 60, 0, 161, 46, 38, 0, 58, 108, 60, 0, 60, 0, 60, 0, 162, 46, 38, 0, 232, 108, 60, 0, 60, 0, 60, 0, 159, 50, 38, 0, 128, 110, 60, 0, 60, 0, 60, 0, 60, 216, 53, 222, 38, 0, 20, 111, 60, 0, 60, 0, 60, 0, 60, 216, 38, 222, 38, 0, 107, 112, 60, 0, 60, 0, 60, 0, 85, 47, 60, 0, 60, 0, 60, 0, 139, 50, 38, 0, 108, 112, 60, 0, 60, 0, 60, 0, 163, 46, 38, 0, 33, 113, 60, 0, 60, 0, 60, 0, 60, 216, 26, 222, 38, 0, 42, 114, 60, 0, 60, 0, 60, 0, 86, 47, 38, 0, 43, 114, 60, 0, 60, 0, 60, 0, 164, 46, 60, 0, 60, 0, 165, 46, 38, 0, 54, 114, 60, 0, 60, 0, 60, 0, 87, 47, 38, 0, 59, 114, 60, 0, 60, 0, 60, 0, 88, 47, 38, 0, 63, 114, 60, 0, 60, 0, 60, 0, 89, 47, 38, 0, 71, 114, 60, 0, 60, 0, 60, 0, 90, 47, 38, 0, 89, 114, 60, 0, 60, 0, 60, 0, 91, 47, 38, 0, 91, 114, 60, 0, 60, 0, 60, 0, 92, 47, 60, 0, 60, 0, 167, 46, 38, 0, 121, 114, 60, 0, 60, 0, 60, 0, 149, 50, 38, 0, 172, 114, 60, 0, 60, 0, 60, 0, 93, 47, 38, 0, 173, 114, 60, 0, 60, 0, 60, 0, 168, 46, 38, 0, 132, 115, 60, 0, 60, 0, 60, 0, 94, 47, 38, 0, 137, 115, 60, 0, 60, 0, 60, 0, 95, 47, 38, 0, 139, 115, 60, 0, 60, 0, 169, 46, 38, 0, 220, 116, 60, 0, 60, 0, 60, 0, 96, 47, 38, 0, 230, 116, 60, 0, 60, 0, 60, 0, 97, 47, 38, 0, 24, 117, 60, 0, 60, 0, 60, 0, 98, 47, 38, 0, 31, 117, 60, 0, 60, 0, 60, 0, 99, 47, 60, 0, 60, 0, 60, 0, 60, 216, 34, 222, 38, 0, 40, 117, 60, 0, 60, 0, 60, 0, 100, 47, 38, 0, 48, 117, 60, 0, 60, 0, 60, 0, 101, 47, 38, 0, 50, 117, 60, 0, 60, 0, 60, 0, 153, 49, 38, 0, 51, 117, 60, 0, 60, 0, 60, 0, 60, 216, 56, 222, 38, 0, 55, 117, 60, 0, 60, 0, 60, 0, 154, 50, 38, 0, 139, 117, 60, 0, 60, 0, 60, 0, 102, 47, 60, 0, 60, 0, 170, 46, 38, 0, 146, 117, 60, 0, 60, 0, 60, 0, 103, 47, 38, 0, 118, 118, 60, 0, 60, 0, 60, 0, 104, 47, 38, 0, 125, 118, 60, 0, 60, 0, 60, 0, 105, 47, 38, 0, 174, 118, 60, 0, 60, 0, 60, 0, 106, 47, 38, 0, 191, 118, 60, 0, 60, 0, 60, 0, 107, 47, 38, 0, 227, 118, 60, 0, 60, 0, 60, 0, 172, 50, 38, 0, 238, 118, 60, 0, 60, 0, 60, 0, 108, 47, 60, 0, 60, 0, 171, 46, 38, 0, 219, 119, 60, 0, 60, 0, 60, 0, 109, 47, 38, 0, 226, 119, 60, 0, 60, 0, 60, 0, 110, 47, 38, 0, 243, 119, 60, 0, 60, 0, 60, 0, 111, 47, 38, 0, 58, 121, 60, 0, 60, 0, 60, 0, 112, 47, 60, 0, 60, 0, 172, 46, 38, 0, 59, 121, 60, 0, 60, 0, 60, 0, 173, 46, 38, 0, 62, 121, 60, 0, 60, 0, 60, 0, 147, 50, 38, 0, 93, 121, 60, 0, 60, 0, 60, 0, 151, 50, 38, 0, 129, 121, 60, 0, 60, 0, 60, 0, 60, 216, 50, 222, 38, 0, 184, 121, 60, 0, 60, 0, 60, 0, 113, 47, 38, 0, 190, 121, 60, 0, 60, 0, 60, 0, 114, 47, 38, 0, 216, 121, 60, 0, 60, 0, 60, 0, 153, 50, 38, 0, 116, 122, 60, 0, 60, 0, 60, 0, 115, 47, 38, 0, 122, 122, 60, 0, 60, 0, 60, 0, 60, 216, 51, 222, 38, 0, 203, 122, 60, 0, 60, 0, 60, 0, 116, 47, 38, 0, 249, 122, 60, 0, 60, 0, 60, 0, 117, 47, 60, 0, 60, 0, 174, 46, 38, 0, 143, 123, 60, 0, 60, 0, 60, 0, 71, 50, 38, 0, 115, 124, 60, 0, 60, 0, 60, 0, 118, 47, 38, 0, 248, 124, 60, 0, 60, 0, 60, 0, 119, 47, 38, 0, 249, 124, 60, 0, 60, 0, 60, 0, 175, 46, 38, 0, 66, 125, 60, 0, 60, 0, 60, 0, 60, 216, 33, 222, 38, 0, 159, 126, 60, 0, 60, 0, 60, 0, 176, 46, 38, 0, 54, 127, 60, 0, 60, 0, 60, 0, 120, 47, 38, 0, 81, 127, 60, 0, 60, 0, 60, 0, 121, 47, 38, 0, 82, 127, 60, 0, 60, 0, 60, 0, 178, 46, 60, 0, 60, 0, 181, 46, 38, 0, 83, 127, 60, 0, 60, 0, 60, 0, 177, 46, 60, 0, 60, 0, 179, 46, 60, 0, 60, 0, 180, 46, 38, 0, 138, 127, 60, 0, 60, 0, 60, 0, 122, 47, 60, 0, 60, 0, 182, 46, 60, 0, 60, 0, 183, 46, 38, 0, 139, 127, 60, 0, 60, 0, 60, 0, 184, 46, 38, 0, 189, 127, 60, 0, 60, 0, 60, 0, 123, 47, 38, 0, 1, 128, 60, 0, 60, 0, 60, 0, 124, 47, 38, 0, 2, 128, 60, 0, 60, 0, 60, 0, 185, 46, 38, 0, 12, 128, 60, 0, 60, 0, 60, 0, 125, 47, 38, 0, 18, 128, 60, 0, 60, 0, 60, 0, 126, 47, 38, 0, 51, 128, 60, 0, 60, 0, 60, 0, 127, 47, 38, 0, 127, 128, 60, 0, 60, 0, 60, 0, 128, 47, 60, 0, 60, 0, 187, 46, 38, 0, 128, 128, 60, 0, 60, 0, 60, 0, 186, 46, 38, 0, 137, 128, 60, 0, 60, 0, 60, 0, 129, 47, 60, 0, 60, 0, 188, 46, 38, 0, 227, 129, 60, 0, 60, 0, 60, 0, 130, 47, 38, 0, 234, 129, 60, 0, 60, 0, 60, 0, 131, 47, 38, 0, 243, 129, 60, 0, 60, 0, 60, 0, 132, 47, 38, 0, 252, 129, 60, 0, 60, 0, 60, 0, 133, 47, 60, 0, 60, 0, 189, 46, 38, 0, 12, 130, 60, 0, 60, 0, 60, 0, 134, 47, 38, 0, 27, 130, 60, 0, 60, 0, 60, 0, 135, 47, 38, 0, 31, 130, 60, 0, 60, 0, 60, 0, 136, 47, 38, 0, 110, 130, 60, 0, 60, 0, 60, 0, 137, 47, 38, 0, 114, 130, 60, 0, 60, 0, 60, 0, 138, 47, 38, 0, 120, 130, 60, 0, 60, 0, 60, 0, 139, 47, 38, 0, 121, 130, 60, 0, 60, 0, 60, 0, 190, 46, 60, 0, 60, 0, 191, 46, 60, 0, 60, 0, 192, 46, 38, 0, 77, 134, 60, 0, 60, 0, 60, 0, 140, 47, 38, 0, 78, 134, 60, 0, 60, 0, 60, 0, 193, 46, 38, 0, 107, 134, 60, 0, 60, 0, 60, 0, 141, 47, 38, 0, 64, 136, 60, 0, 60, 0, 60, 0, 142, 47, 38, 0, 76, 136, 60, 0, 60, 0, 60, 0, 143, 47, 38, 0, 99, 136, 60, 0, 60, 0, 60, 0, 144, 47, 38, 0, 100, 136, 60, 0, 60, 0, 60, 0, 194, 46, 38, 0, 126, 137, 60, 0, 60, 0, 60, 0, 145, 47, 38, 0, 127, 137, 60, 0, 60, 0, 60, 0, 196, 46, 38, 0, 128, 137, 60, 0, 60, 0, 60, 0, 195, 46, 38, 0, 139, 137, 60, 0, 60, 0, 60, 0, 146, 47, 38, 0, 193, 137, 60, 0, 60, 0, 60, 0, 197, 46, 38, 0, 210, 137, 60, 0, 60, 0, 60, 0, 198, 46, 61, 0, 147, 47, 60, 0, 60, 0, 199, 46, 38, 0, 227, 137, 60, 0, 60, 0, 60, 0, 60, 216, 22, 222, 38, 0, 0, 138, 60, 0, 60, 0, 60, 0, 148, 47, 38, 0, 160, 139, 60, 0, 60, 0, 60, 0, 200, 46, 38, 0, 55, 140, 60, 0, 60, 0, 60, 0, 149, 47, 38, 0, 70, 140, 60, 0, 60, 0, 60, 0, 150, 47, 38, 0, 85, 140, 60, 0, 60, 0, 60, 0, 151, 47, 38, 0, 120, 140, 60, 0, 60, 0, 60, 0, 152, 47, 38, 0, 157, 140, 60, 0, 60, 0, 60, 0, 153, 47, 38, 0, 161, 140, 60, 0, 60, 0, 60, 0, 150, 50, 38, 0, 169, 140, 60, 0, 60, 0, 60, 0, 60, 216, 35, 222, 38, 0, 199, 140, 60, 0, 60, 0, 60, 0, 174, 50, 38, 0, 29, 141, 60, 0, 60, 0, 60, 0, 201, 46, 38, 0, 100, 141, 60, 0, 60, 0, 60, 0, 154, 47, 38, 0, 112, 141, 60, 0, 60, 0, 60, 0, 155, 47, 60, 0, 60, 0, 60, 0, 60, 216, 48, 222, 38, 0, 179, 141, 60, 0, 60, 0, 60, 0, 156, 47, 60, 0, 60, 0, 202, 46, 38, 0, 171, 142, 60, 0, 60, 0, 60, 0, 157, 47, 38, 0, 202, 142, 60, 0, 60, 0, 60, 0, 158, 47, 38, 0, 102, 143, 60, 0, 60, 0, 60, 0, 203, 46, 38, 0, 155, 143, 60, 0, 60, 0, 60, 0, 159, 47, 38, 0, 176, 143, 60, 0, 60, 0, 60, 0, 160, 47, 38, 0, 181, 143, 60, 0, 60, 0, 60, 0, 161, 47, 38, 0, 182, 143, 60, 0, 60, 0, 60, 0, 204, 46, 60, 0, 60, 0, 205, 46, 60, 0, 60, 0, 206, 46, 38, 0, 74, 144, 60, 0, 60, 0, 60, 0, 60, 216, 43, 222, 38, 0, 105, 144, 60, 0, 60, 0, 60, 0, 156, 50, 38, 0, 145, 144, 60, 0, 60, 0, 60, 0, 162, 47, 60, 0, 60, 0, 207, 46, 38, 0, 73, 145, 60, 0, 60, 0, 60, 0, 163, 47, 38, 0, 77, 145, 60, 0, 60, 0, 60, 0, 60, 216, 59, 222, 38, 0, 198, 145, 60, 0, 60, 0, 60, 0, 164, 47, 38, 0, 204, 145, 60, 0, 60, 0, 60, 0, 165, 47, 38, 0, 209, 145, 60, 0, 60, 0, 60, 0, 166, 47, 60, 0, 60, 0, 60, 0, 142, 50, 38, 0, 133, 148, 60, 0, 60, 0, 60, 0, 208, 46, 38, 0, 119, 149, 60, 0, 60, 0, 60, 0, 209, 46, 61, 0, 167, 47, 38, 0, 120, 149, 60, 0, 60, 0, 60, 0, 210, 46, 38, 0, 127, 149, 60, 0, 60, 0, 60, 0, 211, 46, 38, 0, 128, 149, 60, 0, 60, 0, 60, 0, 168, 47, 38, 0, 232, 149, 60, 0, 60, 0, 60, 0, 212, 46, 38, 0, 28, 150, 60, 0, 60, 0, 60, 0, 169, 47, 60, 0, 60, 0, 213, 46, 38, 0, 29, 150, 60, 0, 60, 0, 60, 0, 214, 46, 38, 0, 182, 150, 60, 0, 60, 0, 60, 0, 170, 47, 38, 0, 185, 150, 60, 0, 60, 0, 60, 0, 171, 47, 38, 0, 232, 150, 60, 0, 60, 0, 60, 0, 172, 47, 60, 0, 60, 0, 215, 46, 38, 0, 81, 151, 60, 0, 60, 0, 60, 0, 173, 47, 38, 0, 82, 151, 60, 0, 60, 0, 60, 0, 216, 46, 38, 0, 94, 151, 60, 0, 60, 0, 60, 0, 174, 47, 38, 0, 98, 151, 60, 0, 60, 0, 60, 0, 175, 47, 38, 0, 105, 151, 60, 0, 60, 0, 60, 0, 176, 47, 38, 0, 203, 151, 60, 0, 60, 0, 60, 0, 177, 47, 38, 0, 230, 151, 60, 0, 60, 0, 60, 0, 217, 46, 38, 0, 237, 151, 60, 0, 60, 0, 60, 0, 178, 47, 38, 0, 243, 151, 60, 0, 60, 0, 60, 0, 179, 47, 38, 0, 1, 152, 60, 0, 60, 0, 60, 0, 180, 47, 38, 0, 5, 152, 60, 0, 60, 0, 60, 0, 160, 50, 38, 0, 117, 152, 60, 0, 60, 0, 60, 0, 218, 46, 38, 0, 168, 152, 60, 0, 60, 0, 60, 0, 181, 47, 38, 0, 206, 152, 60, 0, 60, 0, 60, 0, 219, 46, 38, 0, 219, 152, 60, 0, 60, 0, 60, 0, 182, 47, 38, 0, 222, 152, 60, 0, 60, 0, 60, 0, 220, 46, 38, 0, 223, 152, 60, 0, 60, 0, 60, 0, 221, 46, 61, 0, 183, 47, 38, 0, 224, 152, 60, 0, 60, 0, 60, 0, 223, 46, 60, 0, 60, 0, 222, 46, 38, 0, 99, 153, 60, 0, 60, 0, 60, 0, 224, 46, 38, 0, 150, 153, 60, 0, 60, 0, 60, 0, 184, 47, 60, 0, 60, 0, 225, 46, 38, 0, 153, 153, 60, 0, 60, 0, 60, 0, 185, 47, 38, 0, 172, 153, 60, 0, 60, 0, 60, 0, 186, 47, 38, 0, 108, 154, 60, 0, 60, 0, 60, 0, 226, 46, 38, 0, 168, 154, 60, 0, 60, 0, 60, 0, 187, 47, 60, 0, 60, 0, 227, 46, 38, 0, 216, 154, 60, 0, 60, 0, 60, 0, 188, 47, 38, 0, 223, 154, 60, 0, 60, 0, 60, 0, 189, 47, 38, 0, 37, 155, 60, 0, 60, 0, 60, 0, 190, 47, 38, 0, 47, 155, 60, 0, 60, 0, 60, 0, 191, 47, 38, 0, 50, 155, 60, 0, 60, 0, 60, 0, 192, 47, 38, 0, 60, 155, 60, 0, 60, 0, 60, 0, 193, 47, 60, 0, 60, 0, 228, 46, 38, 0, 90, 155, 60, 0, 60, 0, 60, 0, 194, 47, 38, 0, 124, 156, 60, 0, 60, 0, 60, 0, 229, 46, 38, 0, 229, 156, 60, 0, 60, 0, 60, 0, 195, 47, 38, 0, 31, 158, 60, 0, 60, 0, 60, 0, 230, 46, 38, 0, 117, 158, 60, 0, 60, 0, 60, 0, 196, 47, 60, 0, 60, 0, 231, 46, 38, 0, 127, 158, 60, 0, 60, 0, 60, 0, 197, 47, 38, 0, 165, 158, 60, 0, 60, 0, 60, 0, 198, 47, 38, 0, 166, 158, 60, 0, 60, 0, 60, 0, 232, 46, 38, 0, 187, 158, 60, 0, 60, 0, 60, 0, 199, 47, 38, 0, 195, 158, 60, 0, 60, 0, 60, 0, 200, 47, 38, 0, 196, 158, 60, 0, 60, 0, 60, 0, 233, 46, 38, 0, 205, 158, 60, 0, 60, 0, 60, 0, 201, 47, 38, 0, 209, 158, 60, 0, 60, 0, 60, 0, 202, 47, 38, 0, 249, 158, 60, 0, 60, 0, 60, 0, 203, 47, 38, 0, 253, 158, 60, 0, 60, 0, 60, 0, 204, 47, 38, 0, 254, 158, 60, 0, 60, 0, 60, 0, 234, 46, 38, 0, 14, 159, 60, 0, 60, 0, 60, 0, 205, 47, 38, 0, 19, 159, 60, 0, 60, 0, 60, 0, 206, 47, 38, 0, 32, 159, 60, 0, 60, 0, 60, 0, 207, 47, 38, 0, 59, 159, 60, 0, 60, 0, 60, 0, 208, 47, 38, 0, 74, 159, 60, 0, 60, 0, 60, 0, 209, 47, 60, 0, 60, 0, 235, 46, 38, 0, 80, 159, 60, 0, 60, 0, 60, 0, 236, 46, 38, 0, 82, 159, 60, 0, 60, 0, 60, 0, 210, 47, 60, 0, 60, 0, 237, 46, 38, 0, 127, 159, 60, 0, 60, 0, 60, 0, 238, 46, 38, 0, 141, 159, 60, 0, 60, 0, 60, 0, 211, 47, 60, 0, 60, 0, 239, 46, 38, 0, 153, 159, 60, 0, 60, 0, 60, 0, 240, 46, 38, 0, 156, 159, 60, 0, 60, 0, 60, 0, 212, 47, 60, 0, 60, 0, 241, 46, 60, 0, 60, 0, 242, 46, 38, 0, 159, 159, 60, 0, 60, 0, 60, 0, 243, 46, 38, 0, 160, 159, 60, 0, 60, 0, 60, 0, 213, 47, 38, 0, 14, 250, 61, 0, 14, 250, 38, 0, 15, 250, 61, 0, 15, 250, 38, 0, 17, 250, 61, 0, 17, 250, 38, 0, 19, 250, 61, 0, 19, 250, 38, 0, 20, 250, 61, 0, 20, 250, 38, 0, 31, 250, 61, 0, 31, 250, 38, 0, 33, 250, 61, 0, 33, 250, 38, 0, 35, 250, 61, 0, 35, 250, 38, 0, 36, 250, 61, 0, 36, 250, 38, 0, 39, 250, 61, 0, 39, 250, 38, 0, 40, 250, 61, 0, 40, 250, 38, 0, 41, 250, 61, 0, 41, 250, 38, 0, 1, 14, 64, 14, 61, 0, 64, 14, 1, 14, 38, 0, 1, 14, 65, 14, 61, 0, 65, 14, 1, 14, 38, 0, 1, 14, 66, 14, 61, 0, 66, 14, 1, 14, 38, 0, 1, 14, 67, 14, 61, 0, 67, 14, 1, 14, 38, 0, 1, 14, 68, 14, 61, 0, 68, 14, 1, 14, 38, 0, 2, 14, 64, 14, 61, 0, 64, 14, 2, 14, 38, 0, 2, 14, 65, 14, 61, 0, 65, 14, 2, 14, 38, 0, 2, 14, 66, 14, 61, 0, 66, 14, 2, 14, 38, 0, 2, 14, 67, 14, 61, 0, 67, 14, 2, 14, 38, 0, 2, 14, 68, 14, 61, 0, 68, 14, 2, 14, 38, 0, 3, 14, 64, 14, 61, 0, 64, 14, 3, 14, 38, 0, 3, 14, 65, 14, 61, 0, 65, 14, 3, 14, 38, 0, 3, 14, 66, 14, 61, 0, 66, 14, 3, 14, 38, 0, 3, 14, 67, 14, 61, 0, 67, 14, 3, 14, 38, 0, 3, 14, 68, 14, 61, 0, 68, 14, 3, 14, 38, 0, 4, 14, 64, 14, 61, 0, 64, 14, 4, 14, 38, 0, 4, 14, 65, 14, 61, 0, 65, 14, 4, 14, 38, 0, 4, 14, 66, 14, 61, 0, 66, 14, 4, 14, 38, 0, 4, 14, 67, 14, 61, 0, 67, 14, 4, 14, 38, 0, 4, 14, 68, 14, 61, 0, 68, 14, 4, 14, 38, 0, 5, 14, 64, 14, 61, 0, 64, 14, 5, 14, 38, 0, 5, 14, 65, 14, 61, 0, 65, 14, 5, 14, 38, 0, 5, 14, 66, 14, 61, 0, 66, 14, 5, 14, 38, 0, 5, 14, 67, 14, 61, 0, 67, 14, 5, 14, 38, 0, 5, 14, 68, 14, 61, 0, 68, 14, 5, 14, 38, 0, 6, 14, 64, 14, 61, 0, 64, 14, 6, 14, 38, 0, 6, 14, 65, 14, 61, 0, 65, 14, 6, 14, 38, 0, 6, 14, 66, 14, 61, 0, 66, 14, 6, 14, 38, 0, 6, 14, 67, 14, 61, 0, 67, 14, 6, 14, 38, 0, 6, 14, 68, 14, 61, 0, 68, 14, 6, 14, 38, 0, 7, 14, 64, 14, 61, 0, 64, 14, 7, 14, 38, 0, 7, 14, 65, 14, 61, 0, 65, 14, 7, 14, 38, 0, 7, 14, 66, 14, 61, 0, 66, 14, 7, 14, 38, 0, 7, 14, 67, 14, 61, 0, 67, 14, 7, 14, 38, 0, 7, 14, 68, 14, 61, 0, 68, 14, 7, 14, 38, 0, 8, 14, 64, 14, 61, 0, 64, 14, 8, 14, 38, 0, 8, 14, 65, 14, 61, 0, 65, 14, 8, 14, 38, 0, 8, 14, 66, 14, 61, 0, 66, 14, 8, 14, 38, 0, 8, 14, 67, 14, 61, 0, 67, 14, 8, 14, 38, 0, 8, 14, 68, 14, 61, 0, 68, 14, 8, 14, 38, 0, 9, 14, 64, 14, 61, 0, 64, 14, 9, 14, 38, 0, 9, 14, 65, 14, 61, 0, 65, 14, 9, 14, 38, 0, 9, 14, 66, 14, 61, 0, 66, 14, 9, 14, 38, 0, 9, 14, 67, 14, 61, 0, 67, 14, 9, 14, 38, 0, 9, 14, 68, 14, 61, 0, 68, 14, 9, 14, 38, 0, 10, 14, 64, 14, 61, 0, 64, 14, 10, 14, 38, 0, 10, 14, 65, 14, 61, 0, 65, 14, 10, 14, 38, 0, 10, 14, 66, 14, 61, 0, 66, 14, 10, 14, 38, 0, 10, 14, 67, 14, 61, 0, 67, 14, 10, 14, 38, 0, 10, 14, 68, 14, 61, 0, 68, 14, 10, 14, 38, 0, 11, 14, 64, 14, 61, 0, 64, 14, 11, 14, 38, 0, 11, 14, 65, 14, 61, 0, 65, 14, 11, 14, 38, 0, 11, 14, 66, 14, 61, 0, 66, 14, 11, 14, 38, 0, 11, 14, 67, 14, 61, 0, 67, 14, 11, 14, 38, 0, 11, 14, 68, 14, 61, 0, 68, 14, 11, 14, 38, 0, 12, 14, 64, 14, 61, 0, 64, 14, 12, 14, 38, 0, 12, 14, 65, 14, 61, 0, 65, 14, 12, 14, 38, 0, 12, 14, 66, 14, 61, 0, 66, 14, 12, 14, 38, 0, 12, 14, 67, 14, 61, 0, 67, 14, 12, 14, 38, 0, 12, 14, 68, 14, 61, 0, 68, 14, 12, 14, 38, 0, 13, 14, 64, 14, 61, 0, 64, 14, 13, 14, 38, 0, 13, 14, 65, 14, 61, 0, 65, 14, 13, 14, 38, 0, 13, 14, 66, 14, 61, 0, 66, 14, 13, 14, 38, 0, 13, 14, 67, 14, 61, 0, 67, 14, 13, 14, 38, 0, 13, 14, 68, 14, 61, 0, 68, 14, 13, 14, 38, 0, 14, 14, 64, 14, 61, 0, 64, 14, 14, 14, 38, 0, 14, 14, 65, 14, 61, 0, 65, 14, 14, 14, 38, 0, 14, 14, 66, 14, 61, 0, 66, 14, 14, 14, 38, 0, 14, 14, 67, 14, 61, 0, 67, 14, 14, 14, 38, 0, 14, 14, 68, 14, 61, 0, 68, 14, 14, 14, 38, 0, 15, 14, 64, 14, 61, 0, 64, 14, 15, 14, 38, 0, 15, 14, 65, 14, 61, 0, 65, 14, 15, 14, 38, 0, 15, 14, 66, 14, 61, 0, 66, 14, 15, 14, 38, 0, 15, 14, 67, 14, 61, 0, 67, 14, 15, 14, 38, 0, 15, 14, 68, 14, 61, 0, 68, 14, 15, 14, 38, 0, 16, 14, 64, 14, 61, 0, 64, 14, 16, 14, 38, 0, 16, 14, 65, 14, 61, 0, 65, 14, 16, 14, 38, 0, 16, 14, 66, 14, 61, 0, 66, 14, 16, 14, 38, 0, 16, 14, 67, 14, 61, 0, 67, 14, 16, 14, 38, 0, 16, 14, 68, 14, 61, 0, 68, 14, 16, 14, 38, 0, 17, 14, 64, 14, 61, 0, 64, 14, 17, 14, 38, 0, 17, 14, 65, 14, 61, 0, 65, 14, 17, 14, 38, 0, 17, 14, 66, 14, 61, 0, 66, 14, 17, 14, 38, 0, 17, 14, 67, 14, 61, 0, 67, 14, 17, 14, 38, 0, 17, 14, 68, 14, 61, 0, 68, 14, 17, 14, 38, 0, 18, 14, 64, 14, 61, 0, 64, 14, 18, 14, 38, 0, 18, 14, 65, 14, 61, 0, 65, 14, 18, 14, 38, 0, 18, 14, 66, 14, 61, 0, 66, 14, 18, 14, 38, 0, 18, 14, 67, 14, 61, 0, 67, 14, 18, 14, 38, 0, 18, 14, 68, 14, 61, 0, 68, 14, 18, 14, 38, 0, 19, 14, 64, 14, 61, 0, 64, 14, 19, 14, 38, 0, 19, 14, 65, 14, 61, 0, 65, 14, 19, 14, 38, 0, 19, 14, 66, 14, 61, 0, 66, 14, 19, 14, 38, 0, 19, 14, 67, 14, 61, 0, 67, 14, 19, 14, 38, 0, 19, 14, 68, 14, 61, 0, 68, 14, 19, 14, 38, 0, 20, 14, 64, 14, 61, 0, 64, 14, 20, 14, 38, 0, 20, 14, 65, 14, 61, 0, 65, 14, 20, 14, 38, 0, 20, 14, 66, 14, 61, 0, 66, 14, 20, 14, 38, 0, 20, 14, 67, 14, 61, 0, 67, 14, 20, 14, 38, 0, 20, 14, 68, 14, 61, 0, 68, 14, 20, 14, 38, 0, 21, 14, 64, 14, 61, 0, 64, 14, 21, 14, 38, 0, 21, 14, 65, 14, 61, 0, 65, 14, 21, 14, 38, 0, 21, 14, 66, 14, 61, 0, 66, 14, 21, 14, 38, 0, 21, 14, 67, 14, 61, 0, 67, 14, 21, 14, 38, 0, 21, 14, 68, 14, 61, 0, 68, 14, 21, 14, 38, 0, 22, 14, 64, 14, 61, 0, 64, 14, 22, 14, 38, 0, 22, 14, 65, 14, 61, 0, 65, 14, 22, 14, 38, 0, 22, 14, 66, 14, 61, 0, 66, 14, 22, 14, 38, 0, 22, 14, 67, 14, 61, 0, 67, 14, 22, 14, 38, 0, 22, 14, 68, 14, 61, 0, 68, 14, 22, 14, 38, 0, 23, 14, 64, 14, 61, 0, 64, 14, 23, 14, 38, 0, 23, 14, 65, 14, 61, 0, 65, 14, 23, 14, 38, 0, 23, 14, 66, 14, 61, 0, 66, 14, 23, 14, 38, 0, 23, 14, 67, 14, 61, 0, 67, 14, 23, 14, 38, 0, 23, 14, 68, 14, 61, 0, 68, 14, 23, 14, 38, 0, 24, 14, 64, 14, 61, 0, 64, 14, 24, 14, 38, 0, 24, 14, 65, 14, 61, 0, 65, 14, 24, 14, 38, 0, 24, 14, 66, 14, 61, 0, 66, 14, 24, 14, 38, 0, 24, 14, 67, 14, 61, 0, 67, 14, 24, 14, 38, 0, 24, 14, 68, 14, 61, 0, 68, 14, 24, 14, 38, 0, 25, 14, 64, 14, 61, 0, 64, 14, 25, 14, 38, 0, 25, 14, 65, 14, 61, 0, 65, 14, 25, 14, 38, 0, 25, 14, 66, 14, 61, 0, 66, 14, 25, 14, 38, 0, 25, 14, 67, 14, 61, 0, 67, 14, 25, 14, 38, 0, 25, 14, 68, 14, 61, 0, 68, 14, 25, 14, 38, 0, 26, 14, 64, 14, 61, 0, 64, 14, 26, 14, 38, 0, 26, 14, 65, 14, 61, 0, 65, 14, 26, 14, 38, 0, 26, 14, 66, 14, 61, 0, 66, 14, 26, 14, 38, 0, 26, 14, 67, 14, 61, 0, 67, 14, 26, 14, 38, 0, 26, 14, 68, 14, 61, 0, 68, 14, 26, 14, 38, 0, 27, 14, 64, 14, 61, 0, 64, 14, 27, 14, 38, 0, 27, 14, 65, 14, 61, 0, 65, 14, 27, 14, 38, 0, 27, 14, 66, 14, 61, 0, 66, 14, 27, 14, 38, 0, 27, 14, 67, 14, 61, 0, 67, 14, 27, 14, 38, 0, 27, 14, 68, 14, 61, 0, 68, 14, 27, 14, 38, 0, 28, 14, 64, 14, 61, 0, 64, 14, 28, 14, 38, 0, 28, 14, 65, 14, 61, 0, 65, 14, 28, 14, 38, 0, 28, 14, 66, 14, 61, 0, 66, 14, 28, 14, 38, 0, 28, 14, 67, 14, 61, 0, 67, 14, 28, 14, 38, 0, 28, 14, 68, 14, 61, 0, 68, 14, 28, 14, 38, 0, 29, 14, 64, 14, 61, 0, 64, 14, 29, 14, 38, 0, 29, 14, 65, 14, 61, 0, 65, 14, 29, 14, 38, 0, 29, 14, 66, 14, 61, 0, 66, 14, 29, 14, 38, 0, 29, 14, 67, 14, 61, 0, 67, 14, 29, 14, 38, 0, 29, 14, 68, 14, 61, 0, 68, 14, 29, 14, 38, 0, 30, 14, 64, 14, 61, 0, 64, 14, 30, 14, 38, 0, 30, 14, 65, 14, 61, 0, 65, 14, 30, 14, 38, 0, 30, 14, 66, 14, 61, 0, 66, 14, 30, 14, 38, 0, 30, 14, 67, 14, 61, 0, 67, 14, 30, 14, 38, 0, 30, 14, 68, 14, 61, 0, 68, 14, 30, 14, 38, 0, 31, 14, 64, 14, 61, 0, 64, 14, 31, 14, 38, 0, 31, 14, 65, 14, 61, 0, 65, 14, 31, 14, 38, 0, 31, 14, 66, 14, 61, 0, 66, 14, 31, 14, 38, 0, 31, 14, 67, 14, 61, 0, 67, 14, 31, 14, 38, 0, 31, 14, 68, 14, 61, 0, 68, 14, 31, 14, 38, 0, 32, 14, 64, 14, 61, 0, 64, 14, 32, 14, 38, 0, 32, 14, 65, 14, 61, 0, 65, 14, 32, 14, 38, 0, 32, 14, 66, 14, 61, 0, 66, 14, 32, 14, 38, 0, 32, 14, 67, 14, 61, 0, 67, 14, 32, 14, 38, 0, 32, 14, 68, 14, 61, 0, 68, 14, 32, 14, 38, 0, 33, 14, 64, 14, 61, 0, 64, 14, 33, 14, 38, 0, 33, 14, 65, 14, 61, 0, 65, 14, 33, 14, 38, 0, 33, 14, 66, 14, 61, 0, 66, 14, 33, 14, 38, 0, 33, 14, 67, 14, 61, 0, 67, 14, 33, 14, 38, 0, 33, 14, 68, 14, 61, 0, 68, 14, 33, 14, 38, 0, 34, 14, 64, 14, 61, 0, 64, 14, 34, 14, 38, 0, 34, 14, 65, 14, 61, 0, 65, 14, 34, 14, 38, 0, 34, 14, 66, 14, 61, 0, 66, 14, 34, 14, 38, 0, 34, 14, 67, 14, 61, 0, 67, 14, 34, 14, 38, 0, 34, 14, 68, 14, 61, 0, 68, 14, 34, 14, 38, 0, 35, 14, 64, 14, 61, 0, 64, 14, 35, 14, 38, 0, 35, 14, 65, 14, 61, 0, 65, 14, 35, 14, 38, 0, 35, 14, 66, 14, 61, 0, 66, 14, 35, 14, 38, 0, 35, 14, 67, 14, 61, 0, 67, 14, 35, 14, 38, 0, 35, 14, 68, 14, 61, 0, 68, 14, 35, 14, 38, 0, 36, 14, 64, 14, 61, 0, 64, 14, 36, 14, 38, 0, 36, 14, 65, 14, 61, 0, 65, 14, 36, 14, 38, 0, 36, 14, 66, 14, 61, 0, 66, 14, 36, 14, 38, 0, 36, 14, 67, 14, 61, 0, 67, 14, 36, 14, 38, 0, 36, 14, 68, 14, 61, 0, 68, 14, 36, 14, 38, 0, 37, 14, 64, 14, 61, 0, 64, 14, 37, 14, 38, 0, 37, 14, 65, 14, 61, 0, 65, 14, 37, 14, 38, 0, 37, 14, 66, 14, 61, 0, 66, 14, 37, 14, 38, 0, 37, 14, 67, 14, 61, 0, 67, 14, 37, 14, 38, 0, 37, 14, 68, 14, 61, 0, 68, 14, 37, 14, 38, 0, 38, 14, 64, 14, 61, 0, 64, 14, 38, 14, 38, 0, 38, 14, 65, 14, 61, 0, 65, 14, 38, 14, 38, 0, 38, 14, 66, 14, 61, 0, 66, 14, 38, 14, 38, 0, 38, 14, 67, 14, 61, 0, 67, 14, 38, 14, 38, 0, 38, 14, 68, 14, 61, 0, 68, 14, 38, 14, 38, 0, 39, 14, 64, 14, 61, 0, 64, 14, 39, 14, 38, 0, 39, 14, 65, 14, 61, 0, 65, 14, 39, 14, 38, 0, 39, 14, 66, 14, 61, 0, 66, 14, 39, 14, 38, 0, 39, 14, 67, 14, 61, 0, 67, 14, 39, 14, 38, 0, 39, 14, 68, 14, 61, 0, 68, 14, 39, 14, 38, 0, 40, 14, 64, 14, 61, 0, 64, 14, 40, 14, 38, 0, 40, 14, 65, 14, 61, 0, 65, 14, 40, 14, 38, 0, 40, 14, 66, 14, 61, 0, 66, 14, 40, 14, 38, 0, 40, 14, 67, 14, 61, 0, 67, 14, 40, 14, 38, 0, 40, 14, 68, 14, 61, 0, 68, 14, 40, 14, 38, 0, 41, 14, 64, 14, 61, 0, 64, 14, 41, 14, 38, 0, 41, 14, 65, 14, 61, 0, 65, 14, 41, 14, 38, 0, 41, 14, 66, 14, 61, 0, 66, 14, 41, 14, 38, 0, 41, 14, 67, 14, 61, 0, 67, 14, 41, 14, 38, 0, 41, 14, 68, 14, 61, 0, 68, 14, 41, 14, 38, 0, 42, 14, 64, 14, 61, 0, 64, 14, 42, 14, 38, 0, 42, 14, 65, 14, 61, 0, 65, 14, 42, 14, 38, 0, 42, 14, 66, 14, 61, 0, 66, 14, 42, 14, 38, 0, 42, 14, 67, 14, 61, 0, 67, 14, 42, 14, 38, 0, 42, 14, 68, 14, 61, 0, 68, 14, 42, 14, 38, 0, 43, 14, 64, 14, 61, 0, 64, 14, 43, 14, 38, 0, 43, 14, 65, 14, 61, 0, 65, 14, 43, 14, 38, 0, 43, 14, 66, 14, 61, 0, 66, 14, 43, 14, 38, 0, 43, 14, 67, 14, 61, 0, 67, 14, 43, 14, 38, 0, 43, 14, 68, 14, 61, 0, 68, 14, 43, 14, 38, 0, 44, 14, 64, 14, 61, 0, 64, 14, 44, 14, 38, 0, 44, 14, 65, 14, 61, 0, 65, 14, 44, 14, 38, 0, 44, 14, 66, 14, 61, 0, 66, 14, 44, 14, 38, 0, 44, 14, 67, 14, 61, 0, 67, 14, 44, 14, 38, 0, 44, 14, 68, 14, 61, 0, 68, 14, 44, 14, 38, 0, 45, 14, 64, 14, 61, 0, 64, 14, 45, 14, 38, 0, 45, 14, 65, 14, 61, 0, 65, 14, 45, 14, 38, 0, 45, 14, 66, 14, 61, 0, 66, 14, 45, 14, 38, 0, 45, 14, 67, 14, 61, 0, 67, 14, 45, 14, 38, 0, 45, 14, 68, 14, 61, 0, 68, 14, 45, 14, 38, 0, 46, 14, 64, 14, 61, 0, 64, 14, 46, 14, 38, 0, 46, 14, 65, 14, 61, 0, 65, 14, 46, 14, 38, 0, 46, 14, 66, 14, 61, 0, 66, 14, 46, 14, 38, 0, 46, 14, 67, 14, 61, 0, 67, 14, 46, 14, 38, 0, 46, 14, 68, 14, 61, 0, 68, 14, 46, 14, 38, 0, 222, 14, 192, 14, 61, 0, 192, 14, 222, 14, 38, 0, 222, 14, 193, 14, 61, 0, 193, 14, 222, 14, 38, 0, 222, 14, 194, 14, 61, 0, 194, 14, 222, 14, 38, 0, 222, 14, 195, 14, 61, 0, 195, 14, 222, 14, 38, 0, 222, 14, 196, 14, 61, 0, 196, 14, 222, 14, 38, 0, 129, 14, 192, 14, 61, 0, 192, 14, 129, 14, 38, 0, 129, 14, 193, 14, 61, 0, 193, 14, 129, 14, 38, 0, 129, 14, 194, 14, 61, 0, 194, 14, 129, 14, 38, 0, 129, 14, 195, 14, 61, 0, 195, 14, 129, 14, 38, 0, 129, 14, 196, 14, 61, 0, 196, 14, 129, 14, 38, 0, 130, 14, 192, 14, 61, 0, 192, 14, 130, 14, 38, 0, 130, 14, 193, 14, 61, 0, 193, 14, 130, 14, 38, 0, 130, 14, 194, 14, 61, 0, 194, 14, 130, 14, 38, 0, 130, 14, 195, 14, 61, 0, 195, 14, 130, 14, 38, 0, 130, 14, 196, 14, 61, 0, 196, 14, 130, 14, 38, 0, 132, 14, 192, 14, 61, 0, 192, 14, 132, 14, 38, 0, 132, 14, 193, 14, 61, 0, 193, 14, 132, 14, 38, 0, 132, 14, 194, 14, 61, 0, 194, 14, 132, 14, 38, 0, 132, 14, 195, 14, 61, 0, 195, 14, 132, 14, 38, 0, 132, 14, 196, 14, 61, 0, 196, 14, 132, 14, 38, 0, 134, 14, 192, 14, 61, 0, 192, 14, 134, 14, 38, 0, 134, 14, 193, 14, 61, 0, 193, 14, 134, 14, 38, 0, 134, 14, 194, 14, 61, 0, 194, 14, 134, 14, 38, 0, 134, 14, 195, 14, 61, 0, 195, 14, 134, 14, 38, 0, 134, 14, 196, 14, 61, 0, 196, 14, 134, 14, 38, 0, 135, 14, 192, 14, 61, 0, 192, 14, 135, 14, 38, 0, 135, 14, 193, 14, 61, 0, 193, 14, 135, 14, 38, 0, 135, 14, 194, 14, 61, 0, 194, 14, 135, 14, 38, 0, 135, 14, 195, 14, 61, 0, 195, 14, 135, 14, 38, 0, 135, 14, 196, 14, 61, 0, 196, 14, 135, 14, 38, 0, 136, 14, 192, 14, 61, 0, 192, 14, 136, 14, 38, 0, 136, 14, 193, 14, 61, 0, 193, 14, 136, 14, 38, 0, 136, 14, 194, 14, 61, 0, 194, 14, 136, 14, 38, 0, 136, 14, 195, 14, 61, 0, 195, 14, 136, 14, 38, 0, 136, 14, 196, 14, 61, 0, 196, 14, 136, 14, 38, 0, 137, 14, 192, 14, 61, 0, 192, 14, 137, 14, 38, 0, 137, 14, 193, 14, 61, 0, 193, 14, 137, 14, 38, 0, 137, 14, 194, 14, 61, 0, 194, 14, 137, 14, 38, 0, 137, 14, 195, 14, 61, 0, 195, 14, 137, 14, 38, 0, 137, 14, 196, 14, 61, 0, 196, 14, 137, 14, 38, 0, 170, 14, 192, 14, 61, 0, 192, 14, 170, 14, 38, 0, 170, 14, 193, 14, 61, 0, 193, 14, 170, 14, 38, 0, 170, 14, 194, 14, 61, 0, 194, 14, 170, 14, 38, 0, 170, 14, 195, 14, 61, 0, 195, 14, 170, 14, 38, 0, 170, 14, 196, 14, 61, 0, 196, 14, 170, 14, 38, 0, 138, 14, 192, 14, 61, 0, 192, 14, 138, 14, 38, 0, 138, 14, 193, 14, 61, 0, 193, 14, 138, 14, 38, 0, 138, 14, 194, 14, 61, 0, 194, 14, 138, 14, 38, 0, 138, 14, 195, 14, 61, 0, 195, 14, 138, 14, 38, 0, 138, 14, 196, 14, 61, 0, 196, 14, 138, 14, 38, 0, 140, 14, 192, 14, 61, 0, 192, 14, 140, 14, 38, 0, 140, 14, 193, 14, 61, 0, 193, 14, 140, 14, 38, 0, 140, 14, 194, 14, 61, 0, 194, 14, 140, 14, 38, 0, 140, 14, 195, 14, 61, 0, 195, 14, 140, 14, 38, 0, 140, 14, 196, 14, 61, 0, 196, 14, 140, 14, 38, 0, 142, 14, 192, 14, 61, 0, 192, 14, 142, 14, 38, 0, 142, 14, 193, 14, 61, 0, 193, 14, 142, 14, 38, 0, 142, 14, 194, 14, 61, 0, 194, 14, 142, 14, 38, 0, 142, 14, 195, 14, 61, 0, 195, 14, 142, 14, 38, 0, 142, 14, 196, 14, 61, 0, 196, 14, 142, 14, 38, 0, 223, 14, 192, 14, 61, 0, 192, 14, 223, 14, 38, 0, 223, 14, 193, 14, 61, 0, 193, 14, 223, 14, 38, 0, 223, 14, 194, 14, 61, 0, 194, 14, 223, 14, 38, 0, 223, 14, 195, 14, 61, 0, 195, 14, 223, 14, 38, 0, 223, 14, 196, 14, 61, 0, 196, 14, 223, 14, 38, 0, 141, 14, 192, 14, 61, 0, 192, 14, 141, 14, 38, 0, 141, 14, 193, 14, 61, 0, 193, 14, 141, 14, 38, 0, 141, 14, 194, 14, 61, 0, 194, 14, 141, 14, 38, 0, 141, 14, 195, 14, 61, 0, 195, 14, 141, 14, 38, 0, 141, 14, 196, 14, 61, 0, 196, 14, 141, 14, 38, 0, 143, 14, 192, 14, 61, 0, 192, 14, 143, 14, 38, 0, 143, 14, 193, 14, 61, 0, 193, 14, 143, 14, 38, 0, 143, 14, 194, 14, 61, 0, 194, 14, 143, 14, 38, 0, 143, 14, 195, 14, 61, 0, 195, 14, 143, 14, 38, 0, 143, 14, 196, 14, 61, 0, 196, 14, 143, 14, 38, 0, 144, 14, 192, 14, 61, 0, 192, 14, 144, 14, 38, 0, 144, 14, 193, 14, 61, 0, 193, 14, 144, 14, 38, 0, 144, 14, 194, 14, 61, 0, 194, 14, 144, 14, 38, 0, 144, 14, 195, 14, 61, 0, 195, 14, 144, 14, 38, 0, 144, 14, 196, 14, 61, 0, 196, 14, 144, 14, 38, 0, 145, 14, 192, 14, 61, 0, 192, 14, 145, 14, 38, 0, 145, 14, 193, 14, 61, 0, 193, 14, 145, 14, 38, 0, 145, 14, 194, 14, 61, 0, 194, 14, 145, 14, 38, 0, 145, 14, 195, 14, 61, 0, 195, 14, 145, 14, 38, 0, 145, 14, 196, 14, 61, 0, 196, 14, 145, 14, 38, 0, 146, 14, 192, 14, 61, 0, 192, 14, 146, 14, 38, 0, 146, 14, 193, 14, 61, 0, 193, 14, 146, 14, 38, 0, 146, 14, 194, 14, 61, 0, 194, 14, 146, 14, 38, 0, 146, 14, 195, 14, 61, 0, 195, 14, 146, 14, 38, 0, 146, 14, 196, 14, 61, 0, 196, 14, 146, 14, 38, 0, 147, 14, 192, 14, 61, 0, 192, 14, 147, 14, 38, 0, 147, 14, 193, 14, 61, 0, 193, 14, 147, 14, 38, 0, 147, 14, 194, 14, 61, 0, 194, 14, 147, 14, 38, 0, 147, 14, 195, 14, 61, 0, 195, 14, 147, 14, 38, 0, 147, 14, 196, 14, 61, 0, 196, 14, 147, 14, 38, 0, 148, 14, 192, 14, 61, 0, 192, 14, 148, 14, 38, 0, 148, 14, 193, 14, 61, 0, 193, 14, 148, 14, 38, 0, 148, 14, 194, 14, 61, 0, 194, 14, 148, 14, 38, 0, 148, 14, 195, 14, 61, 0, 195, 14, 148, 14, 38, 0, 148, 14, 196, 14, 61, 0, 196, 14, 148, 14, 38, 0, 149, 14, 192, 14, 61, 0, 192, 14, 149, 14, 38, 0, 149, 14, 193, 14, 61, 0, 193, 14, 149, 14, 38, 0, 149, 14, 194, 14, 61, 0, 194, 14, 149, 14, 38, 0, 149, 14, 195, 14, 61, 0, 195, 14, 149, 14, 38, 0, 149, 14, 196, 14, 61, 0, 196, 14, 149, 14, 38, 0, 150, 14, 192, 14, 61, 0, 192, 14, 150, 14, 38, 0, 150, 14, 193, 14, 61, 0, 193, 14, 150, 14, 38, 0, 150, 14, 194, 14, 61, 0, 194, 14, 150, 14, 38, 0, 150, 14, 195, 14, 61, 0, 195, 14, 150, 14, 38, 0, 150, 14, 196, 14, 61, 0, 196, 14, 150, 14, 38, 0, 151, 14, 192, 14, 61, 0, 192, 14, 151, 14, 38, 0, 151, 14, 193, 14, 61, 0, 193, 14, 151, 14, 38, 0, 151, 14, 194, 14, 61, 0, 194, 14, 151, 14, 38, 0, 151, 14, 195, 14, 61, 0, 195, 14, 151, 14, 38, 0, 151, 14, 196, 14, 61, 0, 196, 14, 151, 14, 38, 0, 152, 14, 192, 14, 61, 0, 192, 14, 152, 14, 38, 0, 152, 14, 193, 14, 61, 0, 193, 14, 152, 14, 38, 0, 152, 14, 194, 14, 61, 0, 194, 14, 152, 14, 38, 0, 152, 14, 195, 14, 61, 0, 195, 14, 152, 14, 38, 0, 152, 14, 196, 14, 61, 0, 196, 14, 152, 14, 38, 0, 153, 14, 192, 14, 61, 0, 192, 14, 153, 14, 38, 0, 153, 14, 193, 14, 61, 0, 193, 14, 153, 14, 38, 0, 153, 14, 194, 14, 61, 0, 194, 14, 153, 14, 38, 0, 153, 14, 195, 14, 61, 0, 195, 14, 153, 14, 38, 0, 153, 14, 196, 14, 61, 0, 196, 14, 153, 14, 38, 0, 154, 14, 192, 14, 61, 0, 192, 14, 154, 14, 38, 0, 154, 14, 193, 14, 61, 0, 193, 14, 154, 14, 38, 0, 154, 14, 194, 14, 61, 0, 194, 14, 154, 14, 38, 0, 154, 14, 195, 14, 61, 0, 195, 14, 154, 14, 38, 0, 154, 14, 196, 14, 61, 0, 196, 14, 154, 14, 38, 0, 155, 14, 192, 14, 61, 0, 192, 14, 155, 14, 38, 0, 155, 14, 193, 14, 61, 0, 193, 14, 155, 14, 38, 0, 155, 14, 194, 14, 61, 0, 194, 14, 155, 14, 38, 0, 155, 14, 195, 14, 61, 0, 195, 14, 155, 14, 38, 0, 155, 14, 196, 14, 61, 0, 196, 14, 155, 14, 38, 0, 156, 14, 192, 14, 61, 0, 192, 14, 156, 14, 38, 0, 156, 14, 193, 14, 61, 0, 193, 14, 156, 14, 38, 0, 156, 14, 194, 14, 61, 0, 194, 14, 156, 14, 38, 0, 156, 14, 195, 14, 61, 0, 195, 14, 156, 14, 38, 0, 156, 14, 196, 14, 61, 0, 196, 14, 156, 14, 38, 0, 157, 14, 192, 14, 61, 0, 192, 14, 157, 14, 38, 0, 157, 14, 193, 14, 61, 0, 193, 14, 157, 14, 38, 0, 157, 14, 194, 14, 61, 0, 194, 14, 157, 14, 38, 0, 157, 14, 195, 14, 61, 0, 195, 14, 157, 14, 38, 0, 157, 14, 196, 14, 61, 0, 196, 14, 157, 14, 38, 0, 158, 14, 192, 14, 61, 0, 192, 14, 158, 14, 38, 0, 158, 14, 193, 14, 61, 0, 193, 14, 158, 14, 38, 0, 158, 14, 194, 14, 61, 0, 194, 14, 158, 14, 38, 0, 158, 14, 195, 14, 61, 0, 195, 14, 158, 14, 38, 0, 158, 14, 196, 14, 61, 0, 196, 14, 158, 14, 38, 0, 159, 14, 192, 14, 61, 0, 192, 14, 159, 14, 38, 0, 159, 14, 193, 14, 61, 0, 193, 14, 159, 14, 38, 0, 159, 14, 194, 14, 61, 0, 194, 14, 159, 14, 38, 0, 159, 14, 195, 14, 61, 0, 195, 14, 159, 14, 38, 0, 159, 14, 196, 14, 61, 0, 196, 14, 159, 14, 38, 0, 160, 14, 192, 14, 61, 0, 192, 14, 160, 14, 38, 0, 160, 14, 193, 14, 61, 0, 193, 14, 160, 14, 38, 0, 160, 14, 194, 14, 61, 0, 194, 14, 160, 14, 38, 0, 160, 14, 195, 14, 61, 0, 195, 14, 160, 14, 38, 0, 160, 14, 196, 14, 61, 0, 196, 14, 160, 14, 38, 0, 161, 14, 192, 14, 61, 0, 192, 14, 161, 14, 38, 0, 161, 14, 193, 14, 61, 0, 193, 14, 161, 14, 38, 0, 161, 14, 194, 14, 61, 0, 194, 14, 161, 14, 38, 0, 161, 14, 195, 14, 61, 0, 195, 14, 161, 14, 38, 0, 161, 14, 196, 14, 61, 0, 196, 14, 161, 14, 38, 0, 162, 14, 192, 14, 61, 0, 192, 14, 162, 14, 38, 0, 162, 14, 193, 14, 61, 0, 193, 14, 162, 14, 38, 0, 162, 14, 194, 14, 61, 0, 194, 14, 162, 14, 38, 0, 162, 14, 195, 14, 61, 0, 195, 14, 162, 14, 38, 0, 162, 14, 196, 14, 61, 0, 196, 14, 162, 14, 38, 0, 163, 14, 192, 14, 61, 0, 192, 14, 163, 14, 38, 0, 163, 14, 193, 14, 61, 0, 193, 14, 163, 14, 38, 0, 163, 14, 194, 14, 61, 0, 194, 14, 163, 14, 38, 0, 163, 14, 195, 14, 61, 0, 195, 14, 163, 14, 38, 0, 163, 14, 196, 14, 61, 0, 196, 14, 163, 14, 38, 0, 165, 14, 192, 14, 61, 0, 192, 14, 165, 14, 38, 0, 165, 14, 193, 14, 61, 0, 193, 14, 165, 14, 38, 0, 165, 14, 194, 14, 61, 0, 194, 14, 165, 14, 38, 0, 165, 14, 195, 14, 61, 0, 195, 14, 165, 14, 38, 0, 165, 14, 196, 14, 61, 0, 196, 14, 165, 14, 38, 0, 167, 14, 192, 14, 61, 0, 192, 14, 167, 14, 38, 0, 167, 14, 193, 14, 61, 0, 193, 14, 167, 14, 38, 0, 167, 14, 194, 14, 61, 0, 194, 14, 167, 14, 38, 0, 167, 14, 195, 14, 61, 0, 195, 14, 167, 14, 38, 0, 167, 14, 196, 14, 61, 0, 196, 14, 167, 14, 38, 0, 168, 14, 192, 14, 61, 0, 192, 14, 168, 14, 38, 0, 168, 14, 193, 14, 61, 0, 193, 14, 168, 14, 38, 0, 168, 14, 194, 14, 61, 0, 194, 14, 168, 14, 38, 0, 168, 14, 195, 14, 61, 0, 195, 14, 168, 14, 38, 0, 168, 14, 196, 14, 61, 0, 196, 14, 168, 14, 38, 0, 169, 14, 192, 14, 61, 0, 192, 14, 169, 14, 38, 0, 169, 14, 193, 14, 61, 0, 193, 14, 169, 14, 38, 0, 169, 14, 194, 14, 61, 0, 194, 14, 169, 14, 38, 0, 169, 14, 195, 14, 61, 0, 195, 14, 169, 14, 38, 0, 169, 14, 196, 14, 61, 0, 196, 14, 169, 14, 38, 0, 171, 14, 192, 14, 61, 0, 192, 14, 171, 14, 38, 0, 171, 14, 193, 14, 61, 0, 193, 14, 171, 14, 38, 0, 171, 14, 194, 14, 61, 0, 194, 14, 171, 14, 38, 0, 171, 14, 195, 14, 61, 0, 195, 14, 171, 14, 38, 0, 171, 14, 196, 14, 61, 0, 196, 14, 171, 14, 38, 0, 220, 14, 192, 14, 61, 0, 192, 14, 220, 14, 38, 0, 220, 14, 193, 14, 61, 0, 193, 14, 220, 14, 38, 0, 220, 14, 194, 14, 61, 0, 194, 14, 220, 14, 38, 0, 220, 14, 195, 14, 61, 0, 195, 14, 220, 14, 38, 0, 220, 14, 196, 14, 61, 0, 196, 14, 220, 14, 38, 0, 221, 14, 192, 14, 61, 0, 192, 14, 221, 14, 38, 0, 221, 14, 193, 14, 61, 0, 193, 14, 221, 14, 38, 0, 221, 14, 194, 14, 61, 0, 194, 14, 221, 14, 38, 0, 221, 14, 195, 14, 61, 0, 195, 14, 221, 14, 38, 0, 221, 14, 196, 14, 61, 0, 196, 14, 221, 14, 38, 0, 172, 14, 192, 14, 61, 0, 192, 14, 172, 14, 38, 0, 172, 14, 193, 14, 61, 0, 193, 14, 172, 14, 38, 0, 172, 14, 194, 14, 61, 0, 194, 14, 172, 14, 38, 0, 172, 14, 195, 14, 61, 0, 195, 14, 172, 14, 38, 0, 172, 14, 196, 14, 61, 0, 196, 14, 172, 14, 38, 0, 173, 14, 192, 14, 61, 0, 192, 14, 173, 14, 38, 0, 173, 14, 193, 14, 61, 0, 193, 14, 173, 14, 38, 0, 173, 14, 194, 14, 61, 0, 194, 14, 173, 14, 38, 0, 173, 14, 195, 14, 61, 0, 195, 14, 173, 14, 38, 0, 173, 14, 196, 14, 61, 0, 196, 14, 173, 14, 38, 0, 174, 14, 192, 14, 61, 0, 192, 14, 174, 14, 38, 0, 174, 14, 193, 14, 61, 0, 193, 14, 174, 14, 38, 0, 174, 14, 194, 14, 61, 0, 194, 14, 174, 14, 38, 0, 174, 14, 195, 14, 61, 0, 195, 14, 174, 14, 38, 0, 174, 14, 196, 14, 61, 0, 196, 14, 174, 14, 38, 0, 128, 170, 181, 170, 61, 0, 181, 170, 128, 170, 38, 0, 128, 170, 182, 170, 61, 0, 182, 170, 128, 170, 38, 0, 128, 170, 185, 170, 61, 0, 185, 170, 128, 170, 38, 0, 128, 170, 187, 170, 61, 0, 187, 170, 128, 170, 38, 0, 128, 170, 188, 170, 61, 0, 188, 170, 128, 170, 38, 0, 129, 170, 181, 170, 61, 0, 181, 170, 129, 170, 38, 0, 129, 170, 182, 170, 61, 0, 182, 170, 129, 170, 38, 0, 129, 170, 185, 170, 61, 0, 185, 170, 129, 170, 38, 0, 129, 170, 187, 170, 61, 0, 187, 170, 129, 170, 38, 0, 129, 170, 188, 170, 61, 0, 188, 170, 129, 170, 38, 0, 130, 170, 181, 170, 61, 0, 181, 170, 130, 170, 38, 0, 130, 170, 182, 170, 61, 0, 182, 170, 130, 170, 38, 0, 130, 170, 185, 170, 61, 0, 185, 170, 130, 170, 38, 0, 130, 170, 187, 170, 61, 0, 187, 170, 130, 170, 38, 0, 130, 170, 188, 170, 61, 0, 188, 170, 130, 170, 38, 0, 131, 170, 181, 170, 61, 0, 181, 170, 131, 170, 38, 0, 131, 170, 182, 170, 61, 0, 182, 170, 131, 170, 38, 0, 131, 170, 185, 170, 61, 0, 185, 170, 131, 170, 38, 0, 131, 170, 187, 170, 61, 0, 187, 170, 131, 170, 38, 0, 131, 170, 188, 170, 61, 0, 188, 170, 131, 170, 38, 0, 132, 170, 181, 170, 61, 0, 181, 170, 132, 170, 38, 0, 132, 170, 182, 170, 61, 0, 182, 170, 132, 170, 38, 0, 132, 170, 185, 170, 61, 0, 185, 170, 132, 170, 38, 0, 132, 170, 187, 170, 61, 0, 187, 170, 132, 170, 38, 0, 132, 170, 188, 170, 61, 0, 188, 170, 132, 170, 38, 0, 133, 170, 181, 170, 61, 0, 181, 170, 133, 170, 38, 0, 133, 170, 182, 170, 61, 0, 182, 170, 133, 170, 38, 0, 133, 170, 185, 170, 61, 0, 185, 170, 133, 170, 38, 0, 133, 170, 187, 170, 61, 0, 187, 170, 133, 170, 38, 0, 133, 170, 188, 170, 61, 0, 188, 170, 133, 170, 38, 0, 134, 170, 181, 170, 61, 0, 181, 170, 134, 170, 38, 0, 134, 170, 182, 170, 61, 0, 182, 170, 134, 170, 38, 0, 134, 170, 185, 170, 61, 0, 185, 170, 134, 170, 38, 0, 134, 170, 187, 170, 61, 0, 187, 170, 134, 170, 38, 0, 134, 170, 188, 170, 61, 0, 188, 170, 134, 170, 38, 0, 135, 170, 181, 170, 61, 0, 181, 170, 135, 170, 38, 0, 135, 170, 182, 170, 61, 0, 182, 170, 135, 170, 38, 0, 135, 170, 185, 170, 61, 0, 185, 170, 135, 170, 38, 0, 135, 170, 187, 170, 61, 0, 187, 170, 135, 170, 38, 0, 135, 170, 188, 170, 61, 0, 188, 170, 135, 170, 38, 0, 136, 170, 181, 170, 61, 0, 181, 170, 136, 170, 38, 0, 136, 170, 182, 170, 61, 0, 182, 170, 136, 170, 38, 0, 136, 170, 185, 170, 61, 0, 185, 170, 136, 170, 38, 0, 136, 170, 187, 170, 61, 0, 187, 170, 136, 170, 38, 0, 136, 170, 188, 170, 61, 0, 188, 170, 136, 170, 38, 0, 137, 170, 181, 170, 61, 0, 181, 170, 137, 170, 38, 0, 137, 170, 182, 170, 61, 0, 182, 170, 137, 170, 38, 0, 137, 170, 185, 170, 61, 0, 185, 170, 137, 170, 38, 0, 137, 170, 187, 170, 61, 0, 187, 170, 137, 170, 38, 0, 137, 170, 188, 170, 61, 0, 188, 170, 137, 170, 38, 0, 138, 170, 181, 170, 61, 0, 181, 170, 138, 170, 38, 0, 138, 170, 182, 170, 61, 0, 182, 170, 138, 170, 38, 0, 138, 170, 185, 170, 61, 0, 185, 170, 138, 170, 38, 0, 138, 170, 187, 170, 61, 0, 187, 170, 138, 170, 38, 0, 138, 170, 188, 170, 61, 0, 188, 170, 138, 170, 38, 0, 139, 170, 181, 170, 61, 0, 181, 170, 139, 170, 38, 0, 139, 170, 182, 170, 61, 0, 182, 170, 139, 170, 38, 0, 139, 170, 185, 170, 61, 0, 185, 170, 139, 170, 38, 0, 139, 170, 187, 170, 61, 0, 187, 170, 139, 170, 38, 0, 139, 170, 188, 170, 61, 0, 188, 170, 139, 170, 38, 0, 140, 170, 181, 170, 61, 0, 181, 170, 140, 170, 38, 0, 140, 170, 182, 170, 61, 0, 182, 170, 140, 170, 38, 0, 140, 170, 185, 170, 61, 0, 185, 170, 140, 170, 38, 0, 140, 170, 187, 170, 61, 0, 187, 170, 140, 170, 38, 0, 140, 170, 188, 170, 61, 0, 188, 170, 140, 170, 38, 0, 141, 170, 181, 170, 61, 0, 181, 170, 141, 170, 38, 0, 141, 170, 182, 170, 61, 0, 182, 170, 141, 170, 38, 0, 141, 170, 185, 170, 61, 0, 185, 170, 141, 170, 38, 0, 141, 170, 187, 170, 61, 0, 187, 170, 141, 170, 38, 0, 141, 170, 188, 170, 61, 0, 188, 170, 141, 170, 38, 0, 142, 170, 181, 170, 61, 0, 181, 170, 142, 170, 38, 0, 142, 170, 182, 170, 61, 0, 182, 170, 142, 170, 38, 0, 142, 170, 185, 170, 61, 0, 185, 170, 142, 170, 38, 0, 142, 170, 187, 170, 61, 0, 187, 170, 142, 170, 38, 0, 142, 170, 188, 170, 61, 0, 188, 170, 142, 170, 38, 0, 143, 170, 181, 170, 61, 0, 181, 170, 143, 170, 38, 0, 143, 170, 182, 170, 61, 0, 182, 170, 143, 170, 38, 0, 143, 170, 185, 170, 61, 0, 185, 170, 143, 170, 38, 0, 143, 170, 187, 170, 61, 0, 187, 170, 143, 170, 38, 0, 143, 170, 188, 170, 61, 0, 188, 170, 143, 170, 38, 0, 144, 170, 181, 170, 61, 0, 181, 170, 144, 170, 38, 0, 144, 170, 182, 170, 61, 0, 182, 170, 144, 170, 38, 0, 144, 170, 185, 170, 61, 0, 185, 170, 144, 170, 38, 0, 144, 170, 187, 170, 61, 0, 187, 170, 144, 170, 38, 0, 144, 170, 188, 170, 61, 0, 188, 170, 144, 170, 38, 0, 145, 170, 181, 170, 61, 0, 181, 170, 145, 170, 38, 0, 145, 170, 182, 170, 61, 0, 182, 170, 145, 170, 38, 0, 145, 170, 185, 170, 61, 0, 185, 170, 145, 170, 38, 0, 145, 170, 187, 170, 61, 0, 187, 170, 145, 170, 38, 0, 145, 170, 188, 170, 61, 0, 188, 170, 145, 170, 38, 0, 146, 170, 181, 170, 61, 0, 181, 170, 146, 170, 38, 0, 146, 170, 182, 170, 61, 0, 182, 170, 146, 170, 38, 0, 146, 170, 185, 170, 61, 0, 185, 170, 146, 170, 38, 0, 146, 170, 187, 170, 61, 0, 187, 170, 146, 170, 38, 0, 146, 170, 188, 170, 61, 0, 188, 170, 146, 170, 38, 0, 147, 170, 181, 170, 61, 0, 181, 170, 147, 170, 38, 0, 147, 170, 182, 170, 61, 0, 182, 170, 147, 170, 38, 0, 147, 170, 185, 170, 61, 0, 185, 170, 147, 170, 38, 0, 147, 170, 187, 170, 61, 0, 187, 170, 147, 170, 38, 0, 147, 170, 188, 170, 61, 0, 188, 170, 147, 170, 38, 0, 148, 170, 181, 170, 61, 0, 181, 170, 148, 170, 38, 0, 148, 170, 182, 170, 61, 0, 182, 170, 148, 170, 38, 0, 148, 170, 185, 170, 61, 0, 185, 170, 148, 170, 38, 0, 148, 170, 187, 170, 61, 0, 187, 170, 148, 170, 38, 0, 148, 170, 188, 170, 61, 0, 188, 170, 148, 170, 38, 0, 149, 170, 181, 170, 61, 0, 181, 170, 149, 170, 38, 0, 149, 170, 182, 170, 61, 0, 182, 170, 149, 170, 38, 0, 149, 170, 185, 170, 61, 0, 185, 170, 149, 170, 38, 0, 149, 170, 187, 170, 61, 0, 187, 170, 149, 170, 38, 0, 149, 170, 188, 170, 61, 0, 188, 170, 149, 170, 38, 0, 150, 170, 181, 170, 61, 0, 181, 170, 150, 170, 38, 0, 150, 170, 182, 170, 61, 0, 182, 170, 150, 170, 38, 0, 150, 170, 185, 170, 61, 0, 185, 170, 150, 170, 38, 0, 150, 170, 187, 170, 61, 0, 187, 170, 150, 170, 38, 0, 150, 170, 188, 170, 61, 0, 188, 170, 150, 170, 38, 0, 151, 170, 181, 170, 61, 0, 181, 170, 151, 170, 38, 0, 151, 170, 182, 170, 61, 0, 182, 170, 151, 170, 38, 0, 151, 170, 185, 170, 61, 0, 185, 170, 151, 170, 38, 0, 151, 170, 187, 170, 61, 0, 187, 170, 151, 170, 38, 0, 151, 170, 188, 170, 61, 0, 188, 170, 151, 170, 38, 0, 152, 170, 181, 170, 61, 0, 181, 170, 152, 170, 38, 0, 152, 170, 182, 170, 61, 0, 182, 170, 152, 170, 38, 0, 152, 170, 185, 170, 61, 0, 185, 170, 152, 170, 38, 0, 152, 170, 187, 170, 61, 0, 187, 170, 152, 170, 38, 0, 152, 170, 188, 170, 61, 0, 188, 170, 152, 170, 38, 0, 153, 170, 181, 170, 61, 0, 181, 170, 153, 170, 38, 0, 153, 170, 182, 170, 61, 0, 182, 170, 153, 170, 38, 0, 153, 170, 185, 170, 61, 0, 185, 170, 153, 170, 38, 0, 153, 170, 187, 170, 61, 0, 187, 170, 153, 170, 38, 0, 153, 170, 188, 170, 61, 0, 188, 170, 153, 170, 38, 0, 154, 170, 181, 170, 61, 0, 181, 170, 154, 170, 38, 0, 154, 170, 182, 170, 61, 0, 182, 170, 154, 170, 38, 0, 154, 170, 185, 170, 61, 0, 185, 170, 154, 170, 38, 0, 154, 170, 187, 170, 61, 0, 187, 170, 154, 170, 38, 0, 154, 170, 188, 170, 61, 0, 188, 170, 154, 170, 38, 0, 155, 170, 181, 170, 61, 0, 181, 170, 155, 170, 38, 0, 155, 170, 182, 170, 61, 0, 182, 170, 155, 170, 38, 0, 155, 170, 185, 170, 61, 0, 185, 170, 155, 170, 38, 0, 155, 170, 187, 170, 61, 0, 187, 170, 155, 170, 38, 0, 155, 170, 188, 170, 61, 0, 188, 170, 155, 170, 38, 0, 156, 170, 181, 170, 61, 0, 181, 170, 156, 170, 38, 0, 156, 170, 182, 170, 61, 0, 182, 170, 156, 170, 38, 0, 156, 170, 185, 170, 61, 0, 185, 170, 156, 170, 38, 0, 156, 170, 187, 170, 61, 0, 187, 170, 156, 170, 38, 0, 156, 170, 188, 170, 61, 0, 188, 170, 156, 170, 38, 0, 157, 170, 181, 170, 61, 0, 181, 170, 157, 170, 38, 0, 157, 170, 182, 170, 61, 0, 182, 170, 157, 170, 38, 0, 157, 170, 185, 170, 61, 0, 185, 170, 157, 170, 38, 0, 157, 170, 187, 170, 61, 0, 187, 170, 157, 170, 38, 0, 157, 170, 188, 170, 61, 0, 188, 170, 157, 170, 38, 0, 158, 170, 181, 170, 61, 0, 181, 170, 158, 170, 38, 0, 158, 170, 182, 170, 61, 0, 182, 170, 158, 170, 38, 0, 158, 170, 185, 170, 61, 0, 185, 170, 158, 170, 38, 0, 158, 170, 187, 170, 61, 0, 187, 170, 158, 170, 38, 0, 158, 170, 188, 170, 61, 0, 188, 170, 158, 170, 38, 0, 159, 170, 181, 170, 61, 0, 181, 170, 159, 170, 38, 0, 159, 170, 182, 170, 61, 0, 182, 170, 159, 170, 38, 0, 159, 170, 185, 170, 61, 0, 185, 170, 159, 170, 38, 0, 159, 170, 187, 170, 61, 0, 187, 170, 159, 170, 38, 0, 159, 170, 188, 170, 61, 0, 188, 170, 159, 170, 38, 0, 160, 170, 181, 170, 61, 0, 181, 170, 160, 170, 38, 0, 160, 170, 182, 170, 61, 0, 182, 170, 160, 170, 38, 0, 160, 170, 185, 170, 61, 0, 185, 170, 160, 170, 38, 0, 160, 170, 187, 170, 61, 0, 187, 170, 160, 170, 38, 0, 160, 170, 188, 170, 61, 0, 188, 170, 160, 170, 38, 0, 161, 170, 181, 170, 61, 0, 181, 170, 161, 170, 38, 0, 161, 170, 182, 170, 61, 0, 182, 170, 161, 170, 38, 0, 161, 170, 185, 170, 61, 0, 185, 170, 161, 170, 38, 0, 161, 170, 187, 170, 61, 0, 187, 170, 161, 170, 38, 0, 161, 170, 188, 170, 61, 0, 188, 170, 161, 170, 38, 0, 162, 170, 181, 170, 61, 0, 181, 170, 162, 170, 38, 0, 162, 170, 182, 170, 61, 0, 182, 170, 162, 170, 38, 0, 162, 170, 185, 170, 61, 0, 185, 170, 162, 170, 38, 0, 162, 170, 187, 170, 61, 0, 187, 170, 162, 170, 38, 0, 162, 170, 188, 170, 61, 0, 188, 170, 162, 170, 38, 0, 163, 170, 181, 170, 61, 0, 181, 170, 163, 170, 38, 0, 163, 170, 182, 170, 61, 0, 182, 170, 163, 170, 38, 0, 163, 170, 185, 170, 61, 0, 185, 170, 163, 170, 38, 0, 163, 170, 187, 170, 61, 0, 187, 170, 163, 170, 38, 0, 163, 170, 188, 170, 61, 0, 188, 170, 163, 170, 38, 0, 164, 170, 181, 170, 61, 0, 181, 170, 164, 170, 38, 0, 164, 170, 182, 170, 61, 0, 182, 170, 164, 170, 38, 0, 164, 170, 185, 170, 61, 0, 185, 170, 164, 170, 38, 0, 164, 170, 187, 170, 61, 0, 187, 170, 164, 170, 38, 0, 164, 170, 188, 170, 61, 0, 188, 170, 164, 170, 38, 0, 165, 170, 181, 170, 61, 0, 181, 170, 165, 170, 38, 0, 165, 170, 182, 170, 61, 0, 182, 170, 165, 170, 38, 0, 165, 170, 185, 170, 61, 0, 185, 170, 165, 170, 38, 0, 165, 170, 187, 170, 61, 0, 187, 170, 165, 170, 38, 0, 165, 170, 188, 170, 61, 0, 188, 170, 165, 170, 38, 0, 166, 170, 181, 170, 61, 0, 181, 170, 166, 170, 38, 0, 166, 170, 182, 170, 61, 0, 182, 170, 166, 170, 38, 0, 166, 170, 185, 170, 61, 0, 185, 170, 166, 170, 38, 0, 166, 170, 187, 170, 61, 0, 187, 170, 166, 170, 38, 0, 166, 170, 188, 170, 61, 0, 188, 170, 166, 170, 38, 0, 167, 170, 181, 170, 61, 0, 181, 170, 167, 170, 38, 0, 167, 170, 182, 170, 61, 0, 182, 170, 167, 170, 38, 0, 167, 170, 185, 170, 61, 0, 185, 170, 167, 170, 38, 0, 167, 170, 187, 170, 61, 0, 187, 170, 167, 170, 38, 0, 167, 170, 188, 170, 61, 0, 188, 170, 167, 170, 38, 0, 168, 170, 181, 170, 61, 0, 181, 170, 168, 170, 38, 0, 168, 170, 182, 170, 61, 0, 182, 170, 168, 170, 38, 0, 168, 170, 185, 170, 61, 0, 185, 170, 168, 170, 38, 0, 168, 170, 187, 170, 61, 0, 187, 170, 168, 170, 38, 0, 168, 170, 188, 170, 61, 0, 188, 170, 168, 170, 38, 0, 169, 170, 181, 170, 61, 0, 181, 170, 169, 170, 38, 0, 169, 170, 182, 170, 61, 0, 182, 170, 169, 170, 38, 0, 169, 170, 185, 170, 61, 0, 185, 170, 169, 170, 38, 0, 169, 170, 187, 170, 61, 0, 187, 170, 169, 170, 38, 0, 169, 170, 188, 170, 61, 0, 188, 170, 169, 170, 38, 0, 170, 170, 181, 170, 61, 0, 181, 170, 170, 170, 38, 0, 170, 170, 182, 170, 61, 0, 182, 170, 170, 170, 38, 0, 170, 170, 185, 170, 61, 0, 185, 170, 170, 170, 38, 0, 170, 170, 187, 170, 61, 0, 187, 170, 170, 170, 38, 0, 170, 170, 188, 170, 61, 0, 188, 170, 170, 170, 38, 0, 171, 170, 181, 170, 61, 0, 181, 170, 171, 170, 38, 0, 171, 170, 182, 170, 61, 0, 182, 170, 171, 170, 38, 0, 171, 170, 185, 170, 61, 0, 185, 170, 171, 170, 38, 0, 171, 170, 187, 170, 61, 0, 187, 170, 171, 170, 38, 0, 171, 170, 188, 170, 61, 0, 188, 170, 171, 170, 38, 0, 172, 170, 181, 170, 61, 0, 181, 170, 172, 170, 38, 0, 172, 170, 182, 170, 61, 0, 182, 170, 172, 170, 38, 0, 172, 170, 185, 170, 61, 0, 185, 170, 172, 170, 38, 0, 172, 170, 187, 170, 61, 0, 187, 170, 172, 170, 38, 0, 172, 170, 188, 170, 61, 0, 188, 170, 172, 170, 38, 0, 173, 170, 181, 170, 61, 0, 181, 170, 173, 170, 38, 0, 173, 170, 182, 170, 61, 0, 182, 170, 173, 170, 38, 0, 173, 170, 185, 170, 61, 0, 185, 170, 173, 170, 38, 0, 173, 170, 187, 170, 61, 0, 187, 170, 173, 170, 38, 0, 173, 170, 188, 170, 61, 0, 188, 170, 173, 170, 38, 0, 174, 170, 181, 170, 61, 0, 181, 170, 174, 170, 38, 0, 174, 170, 182, 170, 61, 0, 182, 170, 174, 170, 38, 0, 174, 170, 185, 170, 61, 0, 185, 170, 174, 170, 38, 0, 174, 170, 187, 170, 61, 0, 187, 170, 174, 170, 38, 0, 174, 170, 188, 170, 61, 0, 188, 170, 174, 170, 38, 0, 175, 170, 181, 170, 61, 0, 181, 170, 175, 170, 38, 0, 175, 170, 182, 170, 61, 0, 182, 170, 175, 170, 38, 0, 175, 170, 185, 170, 61, 0, 185, 170, 175, 170, 38, 0, 175, 170, 187, 170, 61, 0, 187, 170, 175, 170, 38, 0, 175, 170, 188, 170, 61, 0, 188, 170, 175, 170, 38, 0, 128, 25, 181, 25, 61, 0, 181, 25, 128, 25, 38, 0, 128, 25, 182, 25, 61, 0, 182, 25, 128, 25, 38, 0, 128, 25, 183, 25, 61, 0, 183, 25, 128, 25, 38, 0, 128, 25, 186, 25, 61, 0, 186, 25, 128, 25, 38, 0, 129, 25, 181, 25, 61, 0, 181, 25, 129, 25, 38, 0, 129, 25, 182, 25, 61, 0, 182, 25, 129, 25, 38, 0, 129, 25, 183, 25, 61, 0, 183, 25, 129, 25, 38, 0, 129, 25, 186, 25, 61, 0, 186, 25, 129, 25, 38, 0, 130, 25, 181, 25, 61, 0, 181, 25, 130, 25, 38, 0, 130, 25, 182, 25, 61, 0, 182, 25, 130, 25, 38, 0, 130, 25, 183, 25, 61, 0, 183, 25, 130, 25, 38, 0, 130, 25, 186, 25, 61, 0, 186, 25, 130, 25, 38, 0, 131, 25, 181, 25, 61, 0, 181, 25, 131, 25, 38, 0, 131, 25, 182, 25, 61, 0, 182, 25, 131, 25, 38, 0, 131, 25, 183, 25, 61, 0, 183, 25, 131, 25, 38, 0, 131, 25, 186, 25, 61, 0, 186, 25, 131, 25, 38, 0, 132, 25, 181, 25, 61, 0, 181, 25, 132, 25, 38, 0, 132, 25, 182, 25, 61, 0, 182, 25, 132, 25, 38, 0, 132, 25, 183, 25, 61, 0, 183, 25, 132, 25, 38, 0, 132, 25, 186, 25, 61, 0, 186, 25, 132, 25, 38, 0, 133, 25, 181, 25, 61, 0, 181, 25, 133, 25, 38, 0, 133, 25, 182, 25, 61, 0, 182, 25, 133, 25, 38, 0, 133, 25, 183, 25, 61, 0, 183, 25, 133, 25, 38, 0, 133, 25, 186, 25, 61, 0, 186, 25, 133, 25, 38, 0, 134, 25, 181, 25, 61, 0, 181, 25, 134, 25, 38, 0, 134, 25, 182, 25, 61, 0, 182, 25, 134, 25, 38, 0, 134, 25, 183, 25, 61, 0, 183, 25, 134, 25, 38, 0, 134, 25, 186, 25, 61, 0, 186, 25, 134, 25, 38, 0, 135, 25, 181, 25, 61, 0, 181, 25, 135, 25, 38, 0, 135, 25, 182, 25, 61, 0, 182, 25, 135, 25, 38, 0, 135, 25, 183, 25, 61, 0, 183, 25, 135, 25, 38, 0, 135, 25, 186, 25, 61, 0, 186, 25, 135, 25, 38, 0, 136, 25, 181, 25, 61, 0, 181, 25, 136, 25, 38, 0, 136, 25, 182, 25, 61, 0, 182, 25, 136, 25, 38, 0, 136, 25, 183, 25, 61, 0, 183, 25, 136, 25, 38, 0, 136, 25, 186, 25, 61, 0, 186, 25, 136, 25, 38, 0, 137, 25, 181, 25, 61, 0, 181, 25, 137, 25, 38, 0, 137, 25, 182, 25, 61, 0, 182, 25, 137, 25, 38, 0, 137, 25, 183, 25, 61, 0, 183, 25, 137, 25, 38, 0, 137, 25, 186, 25, 61, 0, 186, 25, 137, 25, 38, 0, 138, 25, 181, 25, 61, 0, 181, 25, 138, 25, 38, 0, 138, 25, 182, 25, 61, 0, 182, 25, 138, 25, 38, 0, 138, 25, 183, 25, 61, 0, 183, 25, 138, 25, 38, 0, 138, 25, 186, 25, 61, 0, 186, 25, 138, 25, 38, 0, 139, 25, 181, 25, 61, 0, 181, 25, 139, 25, 38, 0, 139, 25, 182, 25, 61, 0, 182, 25, 139, 25, 38, 0, 139, 25, 183, 25, 61, 0, 183, 25, 139, 25, 38, 0, 139, 25, 186, 25, 61, 0, 186, 25, 139, 25, 38, 0, 140, 25, 181, 25, 61, 0, 181, 25, 140, 25, 38, 0, 140, 25, 182, 25, 61, 0, 182, 25, 140, 25, 38, 0, 140, 25, 183, 25, 61, 0, 183, 25, 140, 25, 38, 0, 140, 25, 186, 25, 61, 0, 186, 25, 140, 25, 38, 0, 141, 25, 181, 25, 61, 0, 181, 25, 141, 25, 38, 0, 141, 25, 182, 25, 61, 0, 182, 25, 141, 25, 38, 0, 141, 25, 183, 25, 61, 0, 183, 25, 141, 25, 38, 0, 141, 25, 186, 25, 61, 0, 186, 25, 141, 25, 38, 0, 142, 25, 181, 25, 61, 0, 181, 25, 142, 25, 38, 0, 142, 25, 182, 25, 61, 0, 182, 25, 142, 25, 38, 0, 142, 25, 183, 25, 61, 0, 183, 25, 142, 25, 38, 0, 142, 25, 186, 25, 61, 0, 186, 25, 142, 25, 38, 0, 143, 25, 181, 25, 61, 0, 181, 25, 143, 25, 38, 0, 143, 25, 182, 25, 61, 0, 182, 25, 143, 25, 38, 0, 143, 25, 183, 25, 61, 0, 183, 25, 143, 25, 38, 0, 143, 25, 186, 25, 61, 0, 186, 25, 143, 25, 38, 0, 144, 25, 181, 25, 61, 0, 181, 25, 144, 25, 38, 0, 144, 25, 182, 25, 61, 0, 182, 25, 144, 25, 38, 0, 144, 25, 183, 25, 61, 0, 183, 25, 144, 25, 38, 0, 144, 25, 186, 25, 61, 0, 186, 25, 144, 25, 38, 0, 145, 25, 181, 25, 61, 0, 181, 25, 145, 25, 38, 0, 145, 25, 182, 25, 61, 0, 182, 25, 145, 25, 38, 0, 145, 25, 183, 25, 61, 0, 183, 25, 145, 25, 38, 0, 145, 25, 186, 25, 61, 0, 186, 25, 145, 25, 38, 0, 146, 25, 181, 25, 61, 0, 181, 25, 146, 25, 38, 0, 146, 25, 182, 25, 61, 0, 182, 25, 146, 25, 38, 0, 146, 25, 183, 25, 61, 0, 183, 25, 146, 25, 38, 0, 146, 25, 186, 25, 61, 0, 186, 25, 146, 25, 38, 0, 147, 25, 181, 25, 61, 0, 181, 25, 147, 25, 38, 0, 147, 25, 182, 25, 61, 0, 182, 25, 147, 25, 38, 0, 147, 25, 183, 25, 61, 0, 183, 25, 147, 25, 38, 0, 147, 25, 186, 25, 61, 0, 186, 25, 147, 25, 38, 0, 148, 25, 181, 25, 61, 0, 181, 25, 148, 25, 38, 0, 148, 25, 182, 25, 61, 0, 182, 25, 148, 25, 38, 0, 148, 25, 183, 25, 61, 0, 183, 25, 148, 25, 38, 0, 148, 25, 186, 25, 61, 0, 186, 25, 148, 25, 38, 0, 149, 25, 181, 25, 61, 0, 181, 25, 149, 25, 38, 0, 149, 25, 182, 25, 61, 0, 182, 25, 149, 25, 38, 0, 149, 25, 183, 25, 61, 0, 183, 25, 149, 25, 38, 0, 149, 25, 186, 25, 61, 0, 186, 25, 149, 25, 38, 0, 150, 25, 181, 25, 61, 0, 181, 25, 150, 25, 38, 0, 150, 25, 182, 25, 61, 0, 182, 25, 150, 25, 38, 0, 150, 25, 183, 25, 61, 0, 183, 25, 150, 25, 38, 0, 150, 25, 186, 25, 61, 0, 186, 25, 150, 25, 38, 0, 151, 25, 181, 25, 61, 0, 181, 25, 151, 25, 38, 0, 151, 25, 182, 25, 61, 0, 182, 25, 151, 25, 38, 0, 151, 25, 183, 25, 61, 0, 183, 25, 151, 25, 38, 0, 151, 25, 186, 25, 61, 0, 186, 25, 151, 25, 38, 0, 152, 25, 181, 25, 61, 0, 181, 25, 152, 25, 38, 0, 152, 25, 182, 25, 61, 0, 182, 25, 152, 25, 38, 0, 152, 25, 183, 25, 61, 0, 183, 25, 152, 25, 38, 0, 152, 25, 186, 25, 61, 0, 186, 25, 152, 25, 38, 0, 153, 25, 181, 25, 61, 0, 181, 25, 153, 25, 38, 0, 153, 25, 182, 25, 61, 0, 182, 25, 153, 25, 38, 0, 153, 25, 183, 25, 61, 0, 183, 25, 153, 25, 38, 0, 153, 25, 186, 25, 61, 0, 186, 25, 153, 25, 38, 0, 154, 25, 181, 25, 61, 0, 181, 25, 154, 25, 38, 0, 154, 25, 182, 25, 61, 0, 182, 25, 154, 25, 38, 0, 154, 25, 183, 25, 61, 0, 183, 25, 154, 25, 38, 0, 154, 25, 186, 25, 61, 0, 186, 25, 154, 25, 38, 0, 155, 25, 181, 25, 61, 0, 181, 25, 155, 25, 38, 0, 155, 25, 182, 25, 61, 0, 182, 25, 155, 25, 38, 0, 155, 25, 183, 25, 61, 0, 183, 25, 155, 25, 38, 0, 155, 25, 186, 25, 61, 0, 186, 25, 155, 25, 38, 0, 156, 25, 181, 25, 61, 0, 181, 25, 156, 25, 38, 0, 156, 25, 182, 25, 61, 0, 182, 25, 156, 25, 38, 0, 156, 25, 183, 25, 61, 0, 183, 25, 156, 25, 38, 0, 156, 25, 186, 25, 61, 0, 186, 25, 156, 25, 38, 0, 157, 25, 181, 25, 61, 0, 181, 25, 157, 25, 38, 0, 157, 25, 182, 25, 61, 0, 182, 25, 157, 25, 38, 0, 157, 25, 183, 25, 61, 0, 183, 25, 157, 25, 38, 0, 157, 25, 186, 25, 61, 0, 186, 25, 157, 25, 38, 0, 158, 25, 181, 25, 61, 0, 181, 25, 158, 25, 38, 0, 158, 25, 182, 25, 61, 0, 182, 25, 158, 25, 38, 0, 158, 25, 183, 25, 61, 0, 183, 25, 158, 25, 38, 0, 158, 25, 186, 25, 61, 0, 186, 25, 158, 25, 38, 0, 159, 25, 181, 25, 61, 0, 181, 25, 159, 25, 38, 0, 159, 25, 182, 25, 61, 0, 182, 25, 159, 25, 38, 0, 159, 25, 183, 25, 61, 0, 183, 25, 159, 25, 38, 0, 159, 25, 186, 25, 61, 0, 186, 25, 159, 25, 38, 0, 160, 25, 181, 25, 61, 0, 181, 25, 160, 25, 38, 0, 160, 25, 182, 25, 61, 0, 182, 25, 160, 25, 38, 0, 160, 25, 183, 25, 61, 0, 183, 25, 160, 25, 38, 0, 160, 25, 186, 25, 61, 0, 186, 25, 160, 25, 38, 0, 161, 25, 181, 25, 61, 0, 181, 25, 161, 25, 38, 0, 161, 25, 182, 25, 61, 0, 182, 25, 161, 25, 38, 0, 161, 25, 183, 25, 61, 0, 183, 25, 161, 25, 38, 0, 161, 25, 186, 25, 61, 0, 186, 25, 161, 25, 38, 0, 162, 25, 181, 25, 61, 0, 181, 25, 162, 25, 38, 0, 162, 25, 182, 25, 61, 0, 182, 25, 162, 25, 38, 0, 162, 25, 183, 25, 61, 0, 183, 25, 162, 25, 38, 0, 162, 25, 186, 25, 61, 0, 186, 25, 162, 25, 38, 0, 163, 25, 181, 25, 61, 0, 181, 25, 163, 25, 38, 0, 163, 25, 182, 25, 61, 0, 182, 25, 163, 25, 38, 0, 163, 25, 183, 25, 61, 0, 183, 25, 163, 25, 38, 0, 163, 25, 186, 25, 61, 0, 186, 25, 163, 25, 38, 0, 164, 25, 181, 25, 61, 0, 181, 25, 164, 25, 38, 0, 164, 25, 182, 25, 61, 0, 182, 25, 164, 25, 38, 0, 164, 25, 183, 25, 61, 0, 183, 25, 164, 25, 38, 0, 164, 25, 186, 25, 61, 0, 186, 25, 164, 25, 38, 0, 165, 25, 181, 25, 61, 0, 181, 25, 165, 25, 38, 0, 165, 25, 182, 25, 61, 0, 182, 25, 165, 25, 38, 0, 165, 25, 183, 25, 61, 0, 183, 25, 165, 25, 38, 0, 165, 25, 186, 25, 61, 0, 186, 25, 165, 25, 38, 0, 166, 25, 181, 25, 61, 0, 181, 25, 166, 25, 38, 0, 166, 25, 182, 25, 61, 0, 182, 25, 166, 25, 38, 0, 166, 25, 183, 25, 61, 0, 183, 25, 166, 25, 38, 0, 166, 25, 186, 25, 61, 0, 186, 25, 166, 25, 38, 0, 167, 25, 181, 25, 61, 0, 181, 25, 167, 25, 38, 0, 167, 25, 182, 25, 61, 0, 182, 25, 167, 25, 38, 0, 167, 25, 183, 25, 61, 0, 183, 25, 167, 25, 38, 0, 167, 25, 186, 25, 61, 0, 186, 25, 167, 25, 38, 0, 168, 25, 181, 25, 61, 0, 181, 25, 168, 25, 38, 0, 168, 25, 182, 25, 61, 0, 182, 25, 168, 25, 38, 0, 168, 25, 183, 25, 61, 0, 183, 25, 168, 25, 38, 0, 168, 25, 186, 25, 61, 0, 186, 25, 168, 25, 38, 0, 169, 25, 181, 25, 61, 0, 181, 25, 169, 25, 38, 0, 169, 25, 182, 25, 61, 0, 182, 25, 169, 25, 38, 0, 169, 25, 183, 25, 61, 0, 183, 25, 169, 25, 38, 0, 169, 25, 186, 25, 61, 0, 186, 25, 169, 25, 38, 0, 170, 25, 181, 25, 61, 0, 181, 25, 170, 25, 38, 0, 170, 25, 182, 25, 61, 0, 182, 25, 170, 25, 38, 0, 170, 25, 183, 25, 61, 0, 183, 25, 170, 25, 38, 0, 170, 25, 186, 25, 61, 0, 186, 25, 170, 25, 38, 0, 171, 25, 181, 25, 61, 0, 181, 25, 171, 25, 38, 0, 171, 25, 182, 25, 61, 0, 182, 25, 171, 25, 38, 0, 171, 25, 183, 25, 61, 0, 183, 25, 171, 25, 38, 0, 171, 25, 186, 25, 61, 0, 186, 25, 171, 25, 0, 0, 1, 0, 15, 0, 2, 0, 74, 0, 41, 0, 236, 1, 1, 0, 36, 137, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 36, 64, 0, 0, 36, 64, 0, 0, 60, 90, 0, 0, 60, 90, 0, 0, 136, 90, 0, 0, 136, 90, 0, 0, 224, 131, 0, 0, 28, 133, 0, 0, 8, 137, 0, 0, 50, 105, 114, 84, 1, 0, 224, 8, 225, 2, 64, 0, 192, 0, 64, 0, 0, 0, 8, 0, 16, 0, 24, 0, 54, 0, 62, 0, 70, 0, 78, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 95, 0, 103, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 111, 0, 48, 0, 48, 0, 118, 0, 125, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 129, 0, 136, 0, 48, 0, 48, 0, 48, 0, 48, 0, 141, 0, 147, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 155, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 161, 0, 169, 0, 171, 0, 179, 0, 187, 0, 195, 0, 203, 0, 211, 0, 219, 0, 227, 0, 233, 0, 241, 0, 249, 0, 0, 1, 8, 1, 11, 1, 19, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 27, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 34, 1, 48, 0, 42, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 56, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 208, 2, 94, 0, 48, 0, 48, 0, 86, 0, 216, 2, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 86, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 216, 0, 24, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 96, 8, 160, 8, 63, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 68, 1, 48, 0, 48, 0, 48, 0, 48, 0, 72, 1, 80, 1, 48, 0, 48, 0, 87, 1, 95, 1, 102, 1, 109, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 117, 1, 125, 1, 133, 1, 141, 1, 149, 1, 157, 1, 165, 1, 173, 1, 181, 1, 189, 1, 197, 1, 205, 1, 213, 1, 221, 1, 229, 1, 237, 1, 245, 1, 253, 1, 5, 2, 13, 2, 20, 2, 27, 2, 35, 2, 43, 2, 51, 2, 59, 2, 67, 2, 48, 0, 75, 2, 83, 2, 91, 2, 99, 2, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 107, 2, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 112, 2, 120, 2, 128, 2, 136, 2, 144, 2, 152, 2, 160, 2, 168, 2, 48, 0, 48, 0, 48, 0, 176, 2, 184, 2, 192, 2, 200, 2, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 201, 0, 0, 0, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 201, 96, 1, 0, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 1, 0, 202, 65, 1, 0, 202, 98, 1, 0, 202, 131, 1, 0, 202, 164, 1, 0, 202, 197, 1, 0, 202, 230, 1, 0, 202, 7, 2, 0, 202, 40, 2, 0, 202, 73, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 167, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 168, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 140, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 141, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 169, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 194, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 71, 130, 13, 193, 70, 130, 13, 193, 69, 130, 13, 193, 63, 130, 13, 193, 65, 130, 13, 193, 67, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 72, 130, 13, 193, 73, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 134, 128, 13, 193, 132, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 123, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 125, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 112, 130, 13, 193, 116, 130, 13, 193, 119, 130, 13, 193, 121, 130, 13, 193, 113, 130, 13, 193, 117, 130, 13, 193, 114, 130, 13, 193, 135, 128, 13, 193, 136, 128, 13, 193, 137, 128, 13, 193, 133, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 122, 130, 13, 193, 123, 130, 13, 193, 124, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 244, 130, 13, 193, 245, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 111, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 115, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 241, 130, 13, 193, 240, 130, 13, 193, 243, 130, 13, 193, 242, 130, 13, 192, 0, 0, 0, 193, 176, 128, 13, 193, 184, 128, 13, 193, 199, 128, 13, 193, 204, 128, 13, 193, 206, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 114, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 157, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 200, 128, 13, 193, 226, 127, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 115, 127, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 167, 125, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 93, 125, 13, 192, 0, 0, 0, 193, 60, 130, 13, 193, 61, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 90, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 91, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 92, 130, 13, 193, 88, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 87, 130, 13, 193, 65, 125, 13, 193, 16, 125, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 132, 130, 13, 192, 0, 0, 0, 193, 133, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 95, 130, 13, 193, 96, 130, 13, 193, 97, 130, 13, 193, 98, 130, 13, 193, 99, 130, 13, 193, 100, 130, 13, 193, 101, 130, 13, 193, 102, 130, 13, 193, 103, 130, 13, 193, 104, 130, 13, 193, 105, 130, 13, 193, 106, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 29, 129, 13, 193, 25, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 28, 129, 13, 192, 0, 0, 0, 193, 26, 129, 13, 193, 27, 129, 13, 192, 0, 0, 0, 193, 54, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 150, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 139, 130, 13, 193, 39, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 226, 129, 13, 193, 109, 128, 13, 193, 229, 129, 13, 193, 149, 130, 13, 193, 240, 129, 13, 193, 248, 129, 13, 192, 0, 0, 0, 193, 238, 129, 13, 192, 0, 0, 0, 193, 84, 130, 13, 193, 151, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 49, 130, 13, 193, 202, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 134, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 230, 130, 13, 193, 229, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 31, 130, 13, 193, 33, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 237, 128, 13, 193, 238, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 205, 128, 13, 193, 185, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 186, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 107, 130, 13, 193, 225, 129, 13, 192, 0, 0, 0, 193, 80, 129, 13, 192, 0, 0, 0, 193, 243, 129, 13, 193, 51, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 42, 128, 13, 193, 38, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 3, 128, 13, 193, 201, 128, 13, 193, 44, 128, 13, 193, 2, 129, 13, 193, 114, 128, 13, 193, 110, 128, 13, 192, 0, 0, 0, 193, 141, 126, 13, 193, 3, 129, 13, 201, 128, 16, 0, 193, 45, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 103, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 213, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 156, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 117, 128, 13, 193, 174, 129, 13, 193, 170, 125, 13, 193, 152, 125, 13, 193, 157, 125, 13, 193, 180, 125, 13, 192, 0, 0, 0, 193, 187, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 188, 129, 13, 192, 0, 0, 0, 193, 158, 130, 13, 192, 0, 0, 0, 193, 135, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 89, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 86, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 215, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 164, 130, 13, 193, 165, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 203, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 166, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 159, 130, 13, 192, 0, 0, 0, 193, 160, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 130, 13, 193, 143, 130, 13, 193, 144, 130, 13, 192, 0, 0, 0, 193, 145, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 124, 125, 13, 193, 126, 125, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 136, 130, 13, 193, 137, 130, 13, 193, 138, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 64, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 161, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 162, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 74, 130, 13, 193, 75, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 68, 130, 13, 193, 62, 130, 13, 193, 66, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 238, 130, 13, 193, 239, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 180, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 155, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 146, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 163, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 218, 130, 13, 192, 0, 0, 0, 193, 219, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 31, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 30, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 97, 1, 0, 198, 129, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 1, 0, 198, 193, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 225, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 2, 0, 198, 33, 2, 0, 198, 65, 2, 0, 198, 97, 2, 0, 198, 129, 2, 0, 198, 161, 2, 0, 198, 193, 2, 0, 198, 225, 2, 0, 198, 1, 3, 0, 198, 33, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 10, 131, 13, 193, 11, 131, 13, 193, 12, 131, 13, 193, 13, 131, 13, 193, 14, 131, 13, 193, 15, 131, 13, 193, 16, 131, 13, 193, 17, 131, 13, 193, 18, 131, 13, 193, 19, 131, 13, 193, 20, 131, 13, 193, 21, 131, 13, 193, 22, 131, 13, 193, 23, 131, 13, 193, 24, 131, 13, 193, 25, 131, 13, 193, 26, 131, 13, 193, 27, 131, 13, 193, 28, 131, 13, 193, 29, 131, 13, 193, 30, 131, 13, 193, 31, 131, 13, 193, 32, 131, 13, 193, 33, 131, 13, 193, 34, 131, 13, 193, 35, 131, 13, 192, 0, 0, 0, 193, 205, 130, 13, 193, 206, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 212, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 209, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 213, 130, 13, 193, 217, 130, 13, 193, 216, 130, 13, 193, 221, 130, 13, 193, 208, 130, 13, 193, 207, 130, 13, 193, 215, 130, 13, 193, 211, 130, 13, 193, 220, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 210, 130, 13, 193, 214, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 196, 128, 13, 193, 46, 128, 13, 193, 198, 128, 13, 193, 47, 128, 13, 193, 49, 128, 13, 193, 50, 128, 13, 193, 51, 128, 13, 193, 52, 128, 13, 193, 197, 128, 13, 193, 53, 128, 13, 193, 209, 128, 13, 193, 4, 128, 13, 193, 183, 128, 13, 193, 249, 127, 13, 193, 250, 127, 13, 193, 251, 127, 13, 193, 252, 127, 13, 193, 163, 128, 13, 193, 164, 128, 13, 193, 165, 128, 13, 193, 166, 128, 13, 193, 167, 128, 13, 193, 168, 128, 13, 193, 169, 128, 13, 193, 170, 128, 13, 193, 171, 128, 13, 193, 172, 128, 13, 193, 173, 128, 13, 193, 174, 128, 13, 193, 177, 128, 13, 193, 178, 128, 13, 193, 181, 128, 13, 193, 182, 128, 13, 193, 175, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 187, 128, 13, 193, 188, 128, 13, 193, 189, 128, 13, 193, 190, 128, 13, 193, 191, 128, 13, 193, 192, 128, 13, 193, 193, 128, 13, 193, 194, 128, 13, 193, 195, 128, 13, 193, 170, 127, 13, 193, 172, 127, 13, 193, 173, 127, 13, 193, 153, 127, 13, 193, 107, 127, 13, 193, 109, 127, 13, 193, 110, 127, 13, 193, 111, 127, 13, 193, 112, 127, 13, 193, 144, 127, 13, 193, 106, 127, 13, 193, 98, 127, 13, 193, 101, 127, 13, 193, 103, 127, 13, 193, 104, 127, 13, 193, 105, 127, 13, 193, 143, 127, 13, 193, 113, 127, 13, 193, 114, 127, 13, 193, 116, 127, 13, 193, 117, 127, 13, 193, 118, 127, 13, 193, 119, 127, 13, 193, 151, 127, 13, 193, 136, 127, 13, 193, 140, 127, 13, 193, 120, 127, 13, 193, 121, 127, 13, 193, 122, 127, 13, 193, 123, 127, 13, 193, 124, 127, 13, 193, 125, 127, 13, 193, 126, 127, 13, 193, 128, 127, 13, 193, 129, 127, 13, 193, 130, 127, 13, 193, 131, 127, 13, 193, 132, 127, 13, 193, 133, 127, 13, 193, 167, 127, 13, 193, 169, 127, 13, 193, 163, 127, 13, 193, 164, 127, 13, 193, 189, 127, 13, 193, 190, 127, 13, 193, 191, 127, 13, 193, 192, 127, 13, 193, 193, 127, 13, 193, 194, 127, 13, 193, 154, 127, 13, 193, 168, 127, 13, 193, 195, 127, 13, 193, 201, 127, 13, 193, 196, 127, 13, 193, 197, 127, 13, 193, 198, 127, 13, 193, 199, 127, 13, 193, 210, 127, 13, 193, 211, 127, 13, 193, 212, 127, 13, 193, 213, 127, 13, 193, 214, 127, 13, 193, 219, 127, 13, 193, 220, 127, 13, 193, 221, 127, 13, 193, 222, 127, 13, 193, 223, 127, 13, 193, 216, 127, 13, 193, 188, 127, 13, 193, 180, 127, 13, 193, 178, 127, 13, 193, 245, 127, 13, 193, 228, 127, 13, 193, 229, 127, 13, 193, 231, 127, 13, 193, 232, 127, 13, 193, 233, 127, 13, 193, 234, 127, 13, 193, 235, 127, 13, 193, 224, 127, 13, 193, 244, 127, 13, 193, 230, 127, 13, 193, 184, 127, 13, 193, 226, 128, 13, 193, 227, 128, 13, 193, 215, 127, 13, 193, 210, 128, 13, 193, 211, 128, 13, 193, 70, 126, 13, 193, 212, 128, 13, 193, 213, 128, 13, 193, 216, 128, 13, 193, 217, 128, 13, 193, 218, 128, 13, 193, 219, 128, 13, 193, 4, 131, 13, 193, 220, 128, 13, 193, 221, 128, 13, 193, 222, 128, 13, 193, 223, 128, 13, 193, 224, 128, 13, 193, 64, 129, 13, 193, 77, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 231, 128, 13, 193, 228, 128, 13, 192, 0, 0, 0, 193, 97, 129, 13, 193, 98, 129, 13, 193, 99, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 132, 129, 13, 193, 229, 128, 13, 193, 55, 128, 13, 193, 56, 128, 13, 193, 57, 128, 13, 193, 4, 129, 13, 193, 100, 129, 13, 193, 131, 129, 13, 193, 126, 130, 13, 193, 101, 129, 13, 193, 35, 129, 13, 193, 76, 129, 13, 193, 59, 128, 13, 193, 230, 128, 13, 193, 134, 129, 13, 193, 33, 129, 13, 193, 17, 129, 13, 193, 10, 129, 13, 193, 19, 129, 13, 193, 13, 129, 13, 193, 20, 129, 13, 193, 246, 128, 13, 193, 32, 129, 13, 193, 95, 129, 13, 193, 96, 129, 13, 193, 103, 129, 13, 193, 105, 129, 13, 193, 106, 129, 13, 193, 107, 129, 13, 193, 108, 129, 13, 193, 94, 129, 13, 193, 6, 129, 13, 193, 244, 128, 13, 193, 7, 129, 13, 193, 240, 128, 13, 193, 2, 131, 13, 193, 142, 126, 13, 201, 96, 25, 0, 201, 224, 33, 0, 193, 233, 128, 13, 193, 232, 128, 13, 193, 140, 126, 13, 193, 242, 128, 13, 193, 243, 128, 13, 201, 96, 42, 0, 201, 224, 50, 0, 201, 224, 59, 0, 193, 90, 128, 13, 193, 89, 128, 13, 193, 247, 128, 13, 193, 241, 128, 13, 193, 248, 128, 13, 193, 249, 128, 13, 193, 251, 128, 13, 193, 2, 128, 13, 193, 6, 128, 13, 193, 7, 128, 13, 193, 13, 128, 13, 193, 18, 128, 13, 193, 48, 128, 13, 193, 19, 128, 13, 193, 12, 128, 13, 193, 8, 128, 13, 193, 9, 128, 13, 193, 10, 128, 13, 193, 11, 128, 13, 193, 20, 128, 13, 193, 21, 128, 13, 193, 22, 128, 13, 193, 23, 128, 13, 193, 24, 128, 13, 193, 25, 128, 13, 193, 26, 128, 13, 193, 36, 130, 13, 193, 27, 128, 13, 193, 28, 128, 13, 193, 29, 128, 13, 193, 30, 128, 13, 193, 31, 128, 13, 193, 32, 128, 13, 193, 145, 129, 13, 193, 33, 128, 13, 193, 34, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 201, 224, 68, 0, 201, 0, 71, 0, 193, 100, 127, 13, 192, 0, 0, 0, 193, 163, 129, 13, 193, 252, 128, 13, 193, 232, 129, 13, 193, 248, 127, 13, 194, 5, 224, 251, 5, 252, 0, 0, 5, 253, 0, 0, 5, 254, 0, 0, 5, 255, 0, 0, 193, 26, 127, 13, 193, 25, 127, 13, 193, 6, 127, 13, 193, 7, 127, 13, 193, 8, 127, 13, 193, 251, 126, 13, 193, 252, 126, 13, 193, 29, 127, 13, 201, 0, 77, 0, 193, 68, 127, 13, 193, 63, 127, 13, 193, 72, 127, 13, 193, 81, 127, 13, 193, 66, 127, 13, 193, 254, 126, 13, 193, 13, 127, 13, 193, 15, 127, 13, 193, 14, 127, 13, 193, 235, 126, 13, 193, 46, 127, 13, 193, 45, 127, 13, 201, 224, 77, 0, 193, 10, 127, 13, 193, 11, 127, 13, 193, 20, 127, 13, 193, 79, 127, 13, 193, 80, 127, 13, 193, 83, 127, 13, 193, 84, 127, 13, 193, 85, 127, 13, 193, 87, 127, 13, 193, 75, 127, 13, 193, 76, 127, 13, 193, 77, 127, 13, 193, 64, 127, 13, 193, 47, 127, 13, 193, 48, 127, 13, 193, 49, 127, 13, 193, 50, 127, 13, 193, 51, 127, 13, 193, 36, 127, 13, 193, 242, 126, 13, 193, 16, 127, 13, 193, 17, 127, 13, 193, 73, 127, 13, 193, 24, 127, 13, 193, 5, 127, 13, 193, 250, 126, 13, 193, 28, 127, 13, 193, 246, 126, 13, 193, 67, 127, 13, 193, 71, 127, 13, 193, 253, 126, 13, 193, 234, 126, 13, 193, 238, 126, 13, 193, 9, 127, 13, 193, 62, 127, 13, 193, 27, 127, 13, 193, 243, 126, 13, 201, 224, 78, 0, 193, 37, 127, 13, 193, 12, 127, 13, 193, 43, 127, 13, 193, 30, 127, 13, 193, 196, 125, 13, 201, 192, 79, 0, 193, 188, 125, 13, 193, 190, 125, 13, 193, 199, 125, 13, 193, 198, 125, 13, 193, 164, 125, 13, 193, 166, 125, 13, 193, 162, 125, 13, 193, 163, 125, 13, 193, 171, 125, 13, 193, 149, 125, 13, 193, 154, 125, 13, 193, 168, 125, 13, 193, 169, 125, 13, 193, 174, 125, 13, 193, 176, 125, 13, 193, 74, 129, 13, 193, 75, 129, 13, 193, 40, 129, 13, 193, 45, 129, 13, 193, 46, 129, 13, 193, 47, 129, 13, 193, 52, 129, 13, 193, 53, 129, 13, 193, 58, 129, 13, 193, 59, 129, 13, 193, 60, 129, 13, 193, 61, 129, 13, 193, 62, 129, 13, 193, 66, 129, 13, 193, 67, 129, 13, 193, 70, 129, 13, 193, 71, 129, 13, 193, 73, 129, 13, 193, 229, 126, 13, 193, 226, 126, 13, 193, 227, 126, 13, 193, 202, 125, 13, 193, 203, 125, 13, 201, 224, 81, 0, 201, 192, 160, 0, 193, 199, 126, 13, 201, 64, 240, 0, 201, 192, 242, 0, 201, 64, 245, 0, 201, 192, 247, 0, 201, 64, 0, 1, 201, 160, 1, 1, 201, 32, 10, 1, 193, 56, 126, 13, 201, 160, 18, 1, 193, 212, 125, 13, 193, 213, 125, 13, 193, 200, 125, 13, 201, 32, 27, 1, 193, 52, 126, 13, 193, 96, 125, 13, 193, 97, 125, 13, 193, 98, 125, 13, 193, 69, 126, 13, 193, 99, 125, 13, 193, 100, 125, 13, 193, 91, 125, 13, 193, 92, 125, 13, 201, 160, 35, 1, 201, 32, 44, 1, 193, 127, 126, 13, 193, 82, 129, 13, 193, 181, 125, 13, 201, 160, 52, 1, 201, 32, 61, 1, 193, 58, 128, 13, 193, 255, 129, 13, 193, 3, 130, 13, 193, 114, 125, 13, 193, 115, 125, 13, 193, 83, 129, 13, 193, 84, 129, 13, 193, 191, 126, 13, 193, 97, 127, 13, 193, 195, 126, 13, 193, 35, 128, 13, 193, 120, 125, 13, 193, 125, 125, 13, 193, 122, 125, 13, 193, 118, 125, 13, 193, 119, 125, 13, 193, 116, 125, 13, 193, 130, 125, 13, 193, 129, 125, 13, 193, 128, 125, 13, 193, 131, 125, 13, 193, 117, 125, 13, 193, 121, 125, 13, 193, 123, 125, 13, 193, 252, 130, 13, 193, 143, 129, 13, 193, 136, 125, 13, 193, 142, 125, 13, 193, 148, 125, 13, 193, 137, 125, 13, 193, 139, 125, 13, 193, 208, 128, 13, 193, 140, 125, 13, 193, 94, 125, 13, 193, 183, 125, 13, 193, 138, 125, 13, 193, 143, 125, 13, 193, 147, 125, 13, 193, 99, 127, 13, 193, 135, 125, 13, 193, 164, 129, 13, 193, 147, 130, 13, 193, 148, 130, 13, 193, 171, 129, 13, 193, 166, 129, 13, 193, 167, 129, 13, 193, 168, 129, 13, 193, 169, 129, 13, 193, 170, 129, 13, 193, 173, 129, 13, 193, 122, 128, 13, 193, 120, 129, 13, 193, 194, 129, 13, 193, 126, 129, 13, 193, 127, 129, 13, 193, 128, 129, 13, 193, 129, 129, 13, 193, 195, 129, 13, 193, 196, 129, 13, 193, 156, 129, 13, 193, 158, 129, 13, 193, 198, 129, 13, 193, 199, 129, 13, 193, 202, 129, 13, 193, 203, 129, 13, 193, 204, 129, 13, 193, 205, 129, 13, 193, 206, 129, 13, 193, 207, 129, 13, 193, 208, 129, 13, 193, 209, 129, 13, 193, 211, 129, 13, 193, 212, 129, 13, 193, 161, 129, 13, 193, 155, 129, 13, 193, 154, 129, 13, 193, 147, 129, 13, 193, 148, 129, 13, 193, 149, 129, 13, 193, 150, 129, 13, 193, 151, 129, 13, 193, 152, 129, 13, 193, 153, 129, 13, 193, 153, 130, 13, 193, 157, 129, 13, 193, 193, 129, 13, 193, 115, 129, 13, 193, 116, 129, 13, 193, 117, 129, 13, 193, 254, 129, 13, 193, 89, 129, 13, 193, 90, 129, 13, 193, 178, 129, 13, 193, 179, 129, 13, 193, 180, 129, 13, 193, 175, 129, 13, 193, 176, 129, 13, 193, 177, 129, 13, 193, 182, 129, 13, 193, 181, 129, 13, 193, 183, 129, 13, 193, 184, 129, 13, 193, 185, 129, 13, 193, 91, 129, 13, 193, 159, 129, 13, 193, 112, 129, 13, 193, 113, 129, 13, 193, 130, 130, 13, 193, 131, 130, 13, 193, 58, 130, 13, 193, 129, 130, 13, 193, 136, 129, 13, 193, 137, 129, 13, 193, 138, 129, 13, 193, 135, 129, 13, 193, 102, 129, 13, 193, 139, 129, 13, 193, 133, 129, 13, 192, 0, 0, 0, 193, 81, 129, 13, 193, 108, 130, 13, 193, 109, 130, 13, 193, 110, 130, 13, 193, 76, 130, 13, 193, 77, 130, 13, 193, 127, 130, 13, 193, 128, 130, 13, 193, 85, 129, 13, 193, 86, 129, 13, 193, 87, 129, 13, 193, 88, 129, 13, 193, 118, 129, 13, 193, 119, 129, 13, 193, 140, 129, 13, 193, 141, 129, 13, 193, 219, 129, 13, 193, 220, 129, 13, 193, 221, 129, 13, 193, 217, 129, 13, 193, 218, 129, 13, 193, 92, 129, 13, 193, 93, 129, 13, 193, 162, 129, 13, 193, 242, 129, 13, 193, 253, 130, 13, 193, 78, 130, 13, 193, 79, 130, 13, 193, 80, 130, 13, 193, 81, 130, 13, 193, 82, 130, 13, 193, 59, 130, 13, 193, 182, 130, 13, 193, 183, 130, 13, 193, 184, 130, 13, 193, 185, 130, 13, 193, 186, 130, 13, 193, 187, 130, 13, 193, 207, 128, 13, 193, 144, 129, 13, 193, 235, 129, 13, 193, 223, 129, 13, 193, 237, 129, 13, 193, 247, 127, 13, 193, 230, 129, 13, 193, 252, 129, 13, 193, 253, 129, 13, 193, 14, 129, 13, 193, 94, 130, 13, 193, 154, 130, 13, 193, 152, 130, 13, 193, 255, 130, 13, 193, 254, 130, 13, 193, 222, 130, 13, 193, 226, 130, 13, 193, 246, 130, 13, 193, 247, 130, 13, 193, 248, 130, 13, 193, 249, 130, 13, 193, 250, 130, 13, 193, 251, 130, 13, 193, 118, 130, 13, 193, 120, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 85, 130, 13, 193, 52, 127, 13, 193, 43, 128, 13, 193, 39, 128, 13, 193, 41, 128, 13, 193, 93, 130, 13, 192, 0, 0, 0, 193, 141, 128, 13, 193, 143, 128, 13, 193, 145, 128, 13, 193, 147, 128, 13, 193, 149, 128, 13, 193, 151, 128, 13, 193, 153, 128, 13, 193, 155, 128, 13, 193, 157, 128, 13, 193, 159, 128, 13, 193, 161, 128, 13, 193, 139, 128, 13, 193, 142, 128, 13, 193, 144, 128, 13, 193, 146, 128, 13, 193, 148, 128, 13, 193, 150, 128, 13, 193, 152, 128, 13, 193, 154, 128, 13, 193, 156, 128, 13, 193, 158, 128, 13, 193, 160, 128, 13, 193, 162, 128, 13, 193, 140, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 129, 13, 193, 138, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 141, 125, 13, 193, 129, 126, 13, 201, 160, 69, 1, 193, 41, 129, 13, 193, 90, 127, 13, 193, 91, 127, 13, 193, 18, 129, 13, 193, 128, 126, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 210, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 190, 129, 13, 193, 189, 129, 13, 193, 191, 129, 13, 193, 192, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 151, 125, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 165, 125, 13, 193, 153, 125, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 133, 125, 13, 193, 121, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 122, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 124, 129, 13, 193, 125, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 34, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 197, 129, 13, 193, 214, 129, 13, 193, 215, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 216, 129, 13, 193, 200, 129, 13, 193, 201, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 239, 129, 13, 193, 222, 129, 13, 193, 160, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 228, 129, 13, 192, 0, 0, 0, 193, 225, 126, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 145, 125, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 146, 125, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 186, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 253, 127, 13, 193, 5, 128, 13, 193, 36, 128, 13, 193, 37, 128, 13, 193, 254, 127, 13, 193, 34, 130, 13, 5, 5, 121, 13, 5, 5, 124, 13, 193, 5, 125, 13, 5, 5, 122, 13, 5, 5, 123, 13, 193, 3, 125, 13, 193, 2, 125, 13, 193, 10, 125, 13, 193, 90, 125, 13, 193, 8, 125, 13, 193, 9, 125, 13, 193, 20, 125, 13, 193, 40, 125, 13, 193, 12, 125, 13, 193, 59, 125, 13, 193, 35, 125, 13, 193, 32, 125, 13, 193, 33, 125, 13, 193, 36, 125, 13, 193, 82, 125, 13, 193, 41, 125, 13, 193, 62, 125, 13, 193, 79, 125, 13, 193, 15, 125, 13, 193, 14, 125, 13, 193, 18, 125, 13, 193, 17, 125, 13, 193, 21, 125, 13, 193, 22, 125, 13, 193, 24, 125, 13, 193, 81, 125, 13, 193, 63, 125, 13, 193, 88, 125, 13, 193, 87, 125, 13, 193, 76, 125, 13, 193, 80, 125, 13, 193, 86, 125, 13, 193, 75, 125, 13, 193, 71, 125, 13, 193, 72, 125, 13, 193, 73, 125, 13, 193, 83, 125, 13, 193, 42, 125, 13, 193, 84, 125, 13, 193, 38, 125, 13, 193, 77, 125, 13, 193, 66, 125, 13, 193, 67, 125, 13, 193, 74, 125, 13, 193, 78, 125, 13, 193, 68, 125, 13, 193, 69, 125, 13, 193, 44, 125, 13, 193, 54, 125, 13, 193, 34, 125, 13, 193, 45, 125, 13, 193, 103, 125, 13, 193, 104, 125, 13, 193, 102, 125, 13, 193, 105, 125, 13, 193, 106, 125, 13, 193, 107, 125, 13, 193, 110, 125, 13, 193, 109, 125, 13, 193, 108, 125, 13, 193, 64, 125, 13, 193, 6, 125, 13, 193, 7, 125, 13, 193, 37, 125, 13, 201, 160, 78, 1, 201, 32, 87, 1, 201, 160, 95, 1, 193, 111, 125, 13, 193, 112, 125, 13, 193, 113, 125, 13, 201, 32, 104, 1, 193, 175, 125, 13, 201, 160, 112, 1, 201, 32, 121, 1, 193, 179, 125, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 128, 128, 13, 193, 123, 128, 13, 193, 60, 128, 13, 193, 61, 128, 13, 193, 62, 128, 13, 193, 63, 128, 13, 193, 64, 128, 13, 193, 65, 128, 13, 193, 66, 128, 13, 193, 67, 128, 13, 193, 68, 128, 13, 193, 71, 128, 13, 193, 72, 128, 13, 193, 73, 128, 13, 193, 74, 128, 13, 193, 99, 128, 13, 193, 75, 128, 13, 193, 76, 128, 13, 193, 77, 128, 13, 193, 78, 128, 13, 193, 79, 128, 13, 193, 80, 128, 13, 193, 81, 128, 13, 193, 82, 128, 13, 193, 83, 128, 13, 193, 84, 128, 13, 193, 86, 128, 13, 193, 87, 128, 13, 193, 88, 128, 13, 193, 69, 128, 13, 193, 70, 128, 13, 193, 124, 128, 13, 193, 125, 128, 13, 193, 126, 128, 13, 193, 116, 128, 13, 201, 160, 129, 1, 193, 112, 128, 13, 193, 105, 128, 13, 193, 106, 128, 13, 193, 108, 128, 13, 193, 104, 128, 13, 193, 3, 131, 13, 193, 6, 130, 13, 193, 52, 130, 13, 193, 30, 130, 13, 193, 54, 130, 13, 193, 37, 130, 13, 193, 55, 130, 13, 193, 38, 130, 13, 193, 56, 130, 13, 193, 95, 128, 13, 193, 53, 130, 13, 201, 32, 138, 1, 201, 160, 146, 1, 201, 32, 155, 1, 193, 57, 130, 13, 193, 50, 130, 13, 193, 40, 130, 13, 193, 41, 130, 13, 193, 42, 130, 13, 193, 43, 130, 13, 193, 13, 130, 13, 193, 44, 130, 13, 193, 15, 130, 13, 193, 185, 126, 13, 193, 16, 130, 13, 193, 45, 130, 13, 193, 46, 130, 13, 193, 47, 130, 13, 193, 48, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 11, 130, 13, 193, 186, 126, 13, 193, 63, 129, 13, 193, 130, 128, 13, 193, 10, 130, 13, 193, 83, 130, 13, 193, 107, 128, 13, 193, 29, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 40, 128, 13, 193, 17, 128, 13, 193, 7, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 227, 129, 13, 193, 233, 129, 13, 193, 102, 128, 13, 193, 100, 128, 13, 193, 101, 128, 13, 193, 115, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 118, 128, 13, 192, 0, 0, 0, 193, 119, 128, 13, 193, 120, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 127, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 193, 113, 128, 13, 193, 96, 128, 13, 193, 91, 128, 13, 193, 111, 128, 13, 193, 8, 129, 13, 193, 129, 128, 13, 193, 97, 128, 13, 193, 94, 128, 13, 193, 85, 128, 13, 193, 98, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 223, 130, 13, 193, 224, 130, 13, 193, 225, 130, 13, 193, 227, 130, 13, 193, 228, 130, 13, 193, 231, 130, 13, 193, 235, 130, 13, 193, 232, 130, 13, 193, 233, 130, 13, 193, 234, 130, 13, 193, 236, 130, 13, 193, 237, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 155, 125, 13, 193, 134, 125, 13, 193, 132, 125, 13, 193, 156, 125, 13, 193, 30, 125, 13, 193, 25, 125, 13, 193, 46, 125, 13, 193, 60, 125, 13, 193, 29, 125, 13, 193, 47, 125, 13, 193, 101, 125, 13, 193, 26, 125, 13, 193, 160, 125, 13, 193, 161, 125, 13, 193, 150, 125, 13, 193, 172, 125, 13, 193, 173, 125, 13, 193, 178, 125, 13, 193, 158, 125, 13, 193, 159, 125, 13, 193, 56, 125, 13, 193, 95, 125, 13, 193, 48, 125, 13, 193, 4, 125, 13, 193, 43, 125, 13, 193, 39, 125, 13, 201, 160, 163, 1, 193, 50, 125, 13, 193, 31, 125, 13, 193, 13, 125, 13, 193, 23, 125, 13, 193, 28, 125, 13, 193, 89, 125, 13, 193, 27, 125, 13, 193, 49, 125, 13, 193, 55, 125, 13, 193, 64, 126, 13, 193, 65, 126, 13, 193, 177, 125, 13, 193, 182, 125, 13, 193, 51, 126, 13, 201, 32, 172, 1, 193, 71, 126, 13, 201, 160, 180, 1, 201, 32, 189, 1, 201, 160, 197, 1, 193, 139, 126, 13, 192, 0, 0, 0, 201, 32, 206, 1, 201, 128, 207, 1, 201, 0, 216, 1, 193, 5, 129, 13, 193, 102, 127, 13, 193, 110, 129, 13, 193, 236, 127, 13, 193, 237, 127, 13, 193, 246, 127, 13, 193, 255, 128, 13, 192, 0, 0, 0, 193, 234, 128, 13, 193, 235, 128, 13, 193, 236, 128, 13, 193, 253, 128, 13, 193, 254, 128, 13, 193, 9, 129, 13, 193, 250, 128, 13, 193, 239, 128, 13, 193, 245, 128, 13, 193, 155, 127, 13, 193, 139, 127, 13, 193, 146, 127, 13, 193, 166, 127, 13, 193, 141, 127, 13, 193, 142, 127, 13, 193, 156, 127, 13, 193, 183, 127, 13, 193, 179, 127, 13, 193, 175, 127, 13, 193, 177, 127, 13, 193, 225, 127, 13, 193, 152, 127, 13, 193, 135, 127, 13, 193, 160, 127, 13, 193, 202, 127, 13, 193, 203, 127, 13, 193, 204, 127, 13, 193, 243, 127, 13, 193, 182, 127, 13, 193, 238, 127, 13, 193, 138, 127, 13, 193, 148, 127, 13, 193, 218, 127, 13, 193, 158, 127, 13, 193, 165, 127, 13, 193, 171, 127, 13, 193, 187, 127, 13, 193, 147, 127, 13, 193, 127, 127, 13, 193, 200, 127, 13, 193, 159, 127, 13, 193, 11, 125, 13, 193, 85, 125, 13, 193, 19, 125, 13, 193, 57, 125, 13, 193, 53, 125, 13, 193, 51, 125, 13, 193, 52, 125, 13, 193, 47, 126, 13, 193, 58, 125, 13, 192, 0, 0, 0, 193, 70, 125, 13, 193, 54, 129, 13, 193, 43, 129, 13, 193, 42, 129, 13, 193, 68, 129, 13, 193, 69, 129, 13, 193, 205, 127, 13, 193, 249, 126, 13, 193, 92, 127, 13, 193, 44, 127, 13, 193, 255, 126, 13, 193, 53, 127, 13, 193, 54, 127, 13, 193, 33, 127, 13, 193, 78, 127, 13, 193, 56, 127, 13, 193, 244, 126, 13, 193, 82, 127, 13, 193, 3, 127, 13, 193, 236, 126, 13, 193, 65, 127, 13, 193, 22, 127, 13, 193, 207, 127, 13, 193, 208, 127, 13, 193, 19, 127, 13, 193, 2, 127, 13, 193, 32, 127, 13, 193, 69, 127, 13, 193, 70, 127, 13, 193, 88, 127, 13, 193, 41, 127, 13, 193, 18, 127, 13, 193, 60, 127, 13, 193, 23, 127, 13, 193, 61, 127, 13, 193, 245, 126, 13, 193, 206, 127, 13, 193, 93, 127, 13, 193, 96, 127, 13, 193, 42, 127, 13, 193, 55, 127, 13, 193, 21, 127, 13, 193, 57, 127, 13, 193, 38, 127, 13, 193, 39, 127, 13, 193, 237, 126, 13, 193, 40, 127, 13, 193, 59, 127, 13, 193, 209, 127, 13, 193, 31, 127, 13, 193, 4, 127, 13, 193, 74, 127, 13, 193, 240, 126, 13, 193, 241, 129, 13, 193, 230, 126, 13, 193, 231, 126, 13, 193, 233, 126, 13, 193, 232, 126, 13, 193, 195, 125, 13, 193, 186, 125, 13, 193, 187, 125, 13, 193, 194, 125, 13, 201, 128, 224, 1, 201, 0, 233, 1, 193, 44, 129, 13, 193, 189, 125, 13, 193, 93, 128, 13, 193, 92, 128, 13, 193, 184, 125, 13, 193, 185, 125, 13, 193, 162, 127, 13, 193, 217, 127, 13, 193, 186, 127, 13, 193, 240, 127, 13, 193, 149, 127, 13, 193, 150, 127, 13, 193, 176, 127, 13, 193, 161, 127, 13, 193, 185, 127, 13, 193, 241, 127, 13, 193, 242, 127, 13, 193, 239, 127, 13, 192, 0, 0, 0, 201, 128, 241, 1, 201, 0, 250, 1, 201, 128, 2, 2, 193, 61, 125, 13, 201, 0, 11, 2, 193, 201, 125, 13, 193, 211, 125, 13, 198, 65, 94, 0, 193, 57, 126, 13, 201, 192, 78, 2, 201, 64, 87, 2, 201, 192, 95, 2, 201, 64, 104, 2, 201, 192, 112, 2, 201, 64, 121, 2, 201, 192, 129, 2, 201, 64, 138, 2, 201, 192, 146, 2, 201, 32, 148, 2, 193, 191, 125, 13, 193, 127, 125, 13, 193, 78, 129, 13, 193, 48, 129, 13, 193, 49, 129, 13, 193, 50, 129, 13, 193, 51, 129, 13, 193, 225, 128, 13, 193, 214, 128, 13, 193, 21, 129, 13, 193, 249, 129, 13, 193, 250, 129, 13, 193, 251, 129, 13, 193, 255, 127, 13, 193, 130, 129, 13, 193, 28, 130, 13, 193, 245, 129, 13, 193, 14, 128, 13, 193, 246, 129, 13, 193, 131, 128, 13, 193, 19, 130, 13, 193, 36, 129, 13, 193, 38, 129, 13, 193, 20, 130, 13, 193, 22, 129, 13, 193, 21, 130, 13, 193, 22, 130, 13, 193, 23, 130, 13, 193, 25, 130, 13, 193, 27, 130, 13, 193, 172, 129, 13, 193, 16, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 72, 129, 13, 193, 55, 129, 13, 193, 56, 129, 13, 193, 57, 129, 13, 193, 65, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 2, 130, 13, 193, 4, 130, 13, 193, 5, 130, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 11, 129, 13, 193, 12, 129, 13, 193, 121, 128, 13, 193, 231, 129, 13, 193, 15, 129, 13, 193, 23, 129, 13, 193, 24, 129, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 179, 128, 13, 193, 12, 130, 13, 193, 18, 130, 13, 193, 224, 129, 13, 193, 146, 129, 13, 193, 109, 129, 13, 193, 79, 129, 13, 193, 104, 129, 13, 193, 111, 129, 13, 193, 165, 129, 13, 193, 234, 129, 13, 193, 236, 129, 13, 193, 247, 129, 13, 193, 244, 129, 13, 193, 8, 130, 13, 193, 9, 130, 13, 193, 14, 130, 13, 193, 37, 129, 13, 193, 39, 129, 13, 193, 24, 130, 13, 193, 17, 130, 13, 193, 26, 130, 13, 193, 32, 130, 13, 193, 35, 130, 13, 193, 15, 128, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 94, 127, 13, 193, 95, 127, 13, 193, 86, 127, 13, 193, 89, 127, 13, 193, 108, 127, 13, 193, 16, 128, 13, 193, 58, 127, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 192, 125, 13, 193, 193, 125, 13, 193, 228, 126, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 134, 127, 13, 193, 145, 127, 13, 193, 137, 127, 13, 193, 157, 127, 13, 193, 174, 127, 13, 193, 181, 127, 13, 193, 227, 127, 13, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 133, 0, 5, 0, 196, 130, 13, 0, 5, 0, 70, 0, 157, 126, 13, 0, 5, 0, 70, 0, 156, 126, 13, 0, 5, 0, 71, 0, 157, 126, 13, 0, 5, 0, 71, 0, 156, 126, 13, 0, 5, 0, 72, 0, 157, 126, 13, 0, 5, 0, 72, 0, 156, 126, 13, 0, 5, 0, 73, 0, 157, 126, 13, 0, 5, 0, 73, 0, 156, 126, 13, 0, 5, 0, 74, 0, 157, 126, 13, 0, 5, 0, 74, 0, 156, 126, 13, 0, 133, 0, 5, 0, 188, 130, 13, 0, 133, 0, 5, 0, 190, 130, 13, 0, 133, 0, 5, 0, 199, 130, 13, 0, 133, 0, 5, 0, 201, 130, 13, 0, 133, 0, 5, 0, 189, 130, 13, 0, 133, 0, 5, 0, 191, 130, 13, 0, 133, 0, 5, 0, 192, 130, 13, 0, 133, 0, 5, 0, 193, 130, 13, 0, 133, 0, 5, 0, 195, 130, 13, 0, 133, 0, 5, 0, 197, 130, 13, 0, 133, 0, 5, 0, 198, 130, 13, 0, 133, 0, 5, 0, 200, 130, 13, 0, 133, 0, 5, 0, 202, 130, 13, 0, 69, 0, 5, 0, 203, 130, 13, 0, 133, 0, 5, 0, 204, 130, 13, 0, 5, 0, 70, 0, 126, 126, 13, 0, 5, 0, 70, 0, 125, 126, 13, 0, 5, 0, 71, 0, 126, 126, 13, 0, 5, 0, 71, 0, 125, 126, 13, 0, 5, 0, 72, 0, 126, 126, 13, 0, 5, 0, 72, 0, 125, 126, 13, 0, 5, 0, 73, 0, 126, 126, 13, 0, 5, 0, 73, 0, 125, 126, 13, 0, 5, 0, 74, 0, 126, 126, 13, 0, 5, 0, 74, 0, 125, 126, 13, 0, 5, 0, 70, 0, 148, 126, 13, 0, 5, 0, 70, 0, 147, 126, 13, 0, 5, 0, 71, 0, 148, 126, 13, 0, 5, 0, 71, 0, 147, 126, 13, 0, 5, 0, 72, 0, 148, 126, 13, 0, 5, 0, 72, 0, 147, 126, 13, 0, 5, 0, 73, 0, 148, 126, 13, 0, 5, 0, 73, 0, 147, 126, 13, 0, 5, 0, 74, 0, 148, 126, 13, 0, 5, 0, 74, 0, 147, 126, 13, 0, 5, 0, 70, 0, 154, 126, 13, 0, 5, 0, 70, 0, 153, 126, 13, 0, 5, 0, 71, 0, 154, 126, 13, 0, 5, 0, 71, 0, 153, 126, 13, 0, 5, 0, 72, 0, 154, 126, 13, 0, 5, 0, 72, 0, 153, 126, 13, 0, 5, 0, 73, 0, 154, 126, 13, 0, 5, 0, 73, 0, 153, 126, 13, 0, 5, 0, 74, 0, 154, 126, 13, 0, 5, 0, 74, 0, 153, 126, 13, 0, 5, 0, 70, 0, 160, 126, 13, 0, 5, 0, 70, 0, 159, 126, 13, 0, 5, 0, 71, 0, 160, 126, 13, 0, 5, 0, 71, 0, 159, 126, 13, 0, 5, 0, 72, 0, 160, 126, 13, 0, 5, 0, 72, 0, 159, 126, 13, 0, 5, 0, 73, 0, 160, 126, 13, 0, 5, 0, 73, 0, 159, 126, 13, 0, 5, 0, 74, 0, 160, 126, 13, 0, 5, 0, 74, 0, 159, 126, 13, 0, 5, 0, 70, 0, 145, 126, 13, 0, 5, 0, 70, 0, 144, 126, 13, 0, 5, 0, 71, 0, 145, 126, 13, 0, 5, 0, 71, 0, 144, 126, 13, 0, 5, 0, 72, 0, 145, 126, 13, 0, 5, 0, 72, 0, 144, 126, 13, 0, 5, 0, 73, 0, 145, 126, 13, 0, 5, 0, 73, 0, 144, 126, 13, 0, 5, 0, 74, 0, 145, 126, 13, 0, 5, 0, 74, 0, 144, 126, 13, 0, 5, 0, 70, 0, 245, 125, 13, 0, 5, 0, 70, 0, 254, 125, 13, 0, 5, 0, 70, 0, 30, 126, 13, 0, 5, 0, 70, 0, 3, 126, 13, 0, 5, 0, 70, 0, 6, 126, 13, 0, 5, 0, 70, 0, 67, 126, 13, 0, 5, 0, 70, 0, 248, 125, 13, 0, 5, 0, 70, 0, 24, 126, 13, 0, 5, 0, 70, 0, 27, 126, 13, 0, 5, 0, 70, 0, 251, 125, 13, 0, 5, 0, 70, 0, 12, 126, 13, 0, 5, 0, 70, 0, 21, 126, 13, 0, 5, 0, 70, 0, 15, 126, 13, 0, 5, 0, 70, 0, 9, 126, 13, 0, 5, 0, 70, 0, 18, 126, 13, 0, 5, 0, 70, 0, 33, 126, 13, 0, 5, 0, 70, 0, 36, 126, 13, 0, 5, 0, 70, 0, 190, 126, 13, 0, 5, 0, 70, 0, 116, 126, 13, 0, 5, 0, 70, 0, 119, 126, 13, 0, 5, 0, 70, 0, 122, 126, 13, 0, 5, 0, 71, 0, 245, 125, 13, 0, 5, 0, 71, 0, 254, 125, 13, 0, 5, 0, 71, 0, 30, 126, 13, 0, 5, 0, 71, 0, 3, 126, 13, 0, 5, 0, 71, 0, 6, 126, 13, 0, 5, 0, 71, 0, 67, 126, 13, 0, 5, 0, 71, 0, 248, 125, 13, 0, 5, 0, 71, 0, 24, 126, 13, 0, 5, 0, 71, 0, 27, 126, 13, 0, 5, 0, 71, 0, 251, 125, 13, 0, 5, 0, 71, 0, 12, 126, 13, 0, 5, 0, 71, 0, 21, 126, 13, 0, 5, 0, 71, 0, 15, 126, 13, 0, 5, 0, 71, 0, 9, 126, 13, 0, 5, 0, 71, 0, 18, 126, 13, 0, 5, 0, 71, 0, 33, 126, 13, 0, 5, 0, 71, 0, 36, 126, 13, 0, 5, 0, 71, 0, 190, 126, 13, 0, 5, 0, 71, 0, 116, 126, 13, 0, 5, 0, 71, 0, 119, 126, 13, 0, 5, 0, 71, 0, 122, 126, 13, 0, 5, 0, 72, 0, 245, 125, 13, 0, 5, 0, 72, 0, 254, 125, 13, 0, 5, 0, 72, 0, 30, 126, 13, 0, 5, 0, 72, 0, 3, 126, 13, 0, 5, 0, 72, 0, 6, 126, 13, 0, 5, 0, 72, 0, 67, 126, 13, 0, 5, 0, 72, 0, 248, 125, 13, 0, 5, 0, 72, 0, 24, 126, 13, 0, 5, 0, 72, 0, 27, 126, 13, 0, 5, 0, 72, 0, 251, 125, 13, 0, 5, 0, 72, 0, 12, 126, 13, 0, 5, 0, 72, 0, 21, 126, 13, 0, 5, 0, 72, 0, 15, 126, 13, 0, 5, 0, 72, 0, 9, 126, 13, 0, 5, 0, 72, 0, 18, 126, 13, 0, 5, 0, 72, 0, 33, 126, 13, 0, 5, 0, 72, 0, 36, 126, 13, 0, 5, 0, 72, 0, 190, 126, 13, 0, 5, 0, 72, 0, 116, 126, 13, 0, 5, 0, 72, 0, 119, 126, 13, 0, 5, 0, 72, 0, 122, 126, 13, 0, 5, 0, 73, 0, 245, 125, 13, 0, 5, 0, 73, 0, 254, 125, 13, 0, 5, 0, 73, 0, 30, 126, 13, 0, 5, 0, 73, 0, 3, 126, 13, 0, 5, 0, 73, 0, 6, 126, 13, 0, 5, 0, 73, 0, 67, 126, 13, 0, 5, 0, 73, 0, 248, 125, 13, 0, 5, 0, 73, 0, 24, 126, 13, 0, 5, 0, 73, 0, 27, 126, 13, 0, 5, 0, 73, 0, 251, 125, 13, 0, 5, 0, 73, 0, 12, 126, 13, 0, 5, 0, 73, 0, 21, 126, 13, 0, 5, 0, 73, 0, 15, 126, 13, 0, 5, 0, 73, 0, 9, 126, 13, 0, 5, 0, 73, 0, 18, 126, 13, 0, 5, 0, 73, 0, 33, 126, 13, 0, 5, 0, 73, 0, 36, 126, 13, 0, 5, 0, 73, 0, 190, 126, 13, 0, 5, 0, 73, 0, 116, 126, 13, 0, 5, 0, 73, 0, 119, 126, 13, 0, 5, 0, 73, 0, 122, 126, 13, 0, 5, 0, 74, 0, 245, 125, 13, 0, 5, 0, 74, 0, 254, 125, 13, 0, 5, 0, 74, 0, 30, 126, 13, 0, 5, 0, 74, 0, 3, 126, 13, 0, 5, 0, 74, 0, 6, 126, 13, 0, 5, 0, 74, 0, 67, 126, 13, 0, 5, 0, 74, 0, 248, 125, 13, 0, 5, 0, 74, 0, 24, 126, 13, 0, 5, 0, 74, 0, 27, 126, 13, 0, 5, 0, 74, 0, 251, 125, 13, 0, 5, 0, 74, 0, 12, 126, 13, 0, 5, 0, 74, 0, 21, 126, 13, 0, 5, 0, 74, 0, 15, 126, 13, 0, 5, 0, 74, 0, 9, 126, 13, 0, 5, 0, 74, 0, 18, 126, 13, 0, 5, 0, 74, 0, 33, 126, 13, 0, 5, 0, 74, 0, 36, 126, 13, 0, 5, 0, 74, 0, 190, 126, 13, 0, 5, 0, 74, 0, 116, 126, 13, 0, 5, 0, 74, 0, 119, 126, 13, 0, 5, 0, 74, 0, 122, 126, 13, 0, 5, 0, 70, 0, 246, 125, 13, 0, 5, 0, 70, 0, 255, 125, 13, 0, 5, 0, 70, 0, 31, 126, 13, 0, 5, 0, 70, 0, 4, 126, 13, 0, 5, 0, 70, 0, 7, 126, 13, 0, 5, 0, 70, 0, 66, 126, 13, 0, 5, 0, 70, 0, 249, 125, 13, 0, 5, 0, 70, 0, 25, 126, 13, 0, 5, 0, 70, 0, 28, 126, 13, 0, 5, 0, 70, 0, 252, 125, 13, 0, 5, 0, 70, 0, 13, 126, 13, 0, 5, 0, 70, 0, 22, 126, 13, 0, 5, 0, 70, 0, 16, 126, 13, 0, 5, 0, 70, 0, 10, 126, 13, 0, 5, 0, 70, 0, 19, 126, 13, 0, 5, 0, 70, 0, 34, 126, 13, 0, 5, 0, 70, 0, 37, 126, 13, 0, 5, 0, 70, 0, 189, 126, 13, 0, 5, 0, 70, 0, 188, 126, 13, 0, 5, 0, 70, 0, 117, 126, 13, 0, 5, 0, 70, 0, 120, 126, 13, 0, 5, 0, 70, 0, 123, 126, 13, 0, 5, 0, 71, 0, 246, 125, 13, 0, 5, 0, 71, 0, 255, 125, 13, 0, 5, 0, 71, 0, 31, 126, 13, 0, 5, 0, 71, 0, 4, 126, 13, 0, 5, 0, 71, 0, 7, 126, 13, 0, 5, 0, 71, 0, 66, 126, 13, 0, 5, 0, 71, 0, 249, 125, 13, 0, 5, 0, 71, 0, 25, 126, 13, 0, 5, 0, 71, 0, 28, 126, 13, 0, 5, 0, 71, 0, 252, 125, 13, 0, 5, 0, 71, 0, 13, 126, 13, 0, 5, 0, 71, 0, 22, 126, 13, 0, 5, 0, 71, 0, 16, 126, 13, 0, 5, 0, 71, 0, 10, 126, 13, 0, 5, 0, 71, 0, 19, 126, 13, 0, 5, 0, 71, 0, 34, 126, 13, 0, 5, 0, 71, 0, 37, 126, 13, 0, 5, 0, 71, 0, 189, 126, 13, 0, 5, 0, 71, 0, 188, 126, 13, 0, 5, 0, 71, 0, 117, 126, 13, 0, 5, 0, 71, 0, 120, 126, 13, 0, 5, 0, 71, 0, 123, 126, 13, 0, 5, 0, 72, 0, 246, 125, 13, 0, 5, 0, 72, 0, 255, 125, 13, 0, 5, 0, 72, 0, 31, 126, 13, 0, 5, 0, 72, 0, 4, 126, 13, 0, 5, 0, 72, 0, 7, 126, 13, 0, 5, 0, 72, 0, 66, 126, 13, 0, 5, 0, 72, 0, 249, 125, 13, 0, 5, 0, 72, 0, 25, 126, 13, 0, 5, 0, 72, 0, 28, 126, 13, 0, 5, 0, 72, 0, 252, 125, 13, 0, 5, 0, 72, 0, 13, 126, 13, 0, 5, 0, 72, 0, 22, 126, 13, 0, 5, 0, 72, 0, 16, 126, 13, 0, 5, 0, 72, 0, 10, 126, 13, 0, 5, 0, 72, 0, 19, 126, 13, 0, 5, 0, 72, 0, 34, 126, 13, 0, 5, 0, 72, 0, 37, 126, 13, 0, 5, 0, 72, 0, 189, 126, 13, 0, 5, 0, 72, 0, 188, 126, 13, 0, 5, 0, 72, 0, 117, 126, 13, 0, 5, 0, 72, 0, 120, 126, 13, 0, 5, 0, 72, 0, 123, 126, 13, 0, 5, 0, 73, 0, 246, 125, 13, 0, 5, 0, 73, 0, 255, 125, 13, 0, 5, 0, 73, 0, 31, 126, 13, 0, 5, 0, 73, 0, 4, 126, 13, 0, 5, 0, 73, 0, 7, 126, 13, 0, 5, 0, 73, 0, 66, 126, 13, 0, 5, 0, 73, 0, 249, 125, 13, 0, 5, 0, 73, 0, 25, 126, 13, 0, 5, 0, 73, 0, 28, 126, 13, 0, 5, 0, 73, 0, 252, 125, 13, 0, 5, 0, 73, 0, 13, 126, 13, 0, 5, 0, 73, 0, 22, 126, 13, 0, 5, 0, 73, 0, 16, 126, 13, 0, 5, 0, 73, 0, 10, 126, 13, 0, 5, 0, 73, 0, 19, 126, 13, 0, 5, 0, 73, 0, 34, 126, 13, 0, 5, 0, 73, 0, 37, 126, 13, 0, 5, 0, 73, 0, 189, 126, 13, 0, 5, 0, 73, 0, 188, 126, 13, 0, 5, 0, 73, 0, 117, 126, 13, 0, 5, 0, 73, 0, 120, 126, 13, 0, 5, 0, 73, 0, 123, 126, 13, 0, 5, 0, 74, 0, 246, 125, 13, 0, 5, 0, 74, 0, 255, 125, 13, 0, 5, 0, 74, 0, 31, 126, 13, 0, 5, 0, 74, 0, 4, 126, 13, 0, 5, 0, 74, 0, 7, 126, 13, 0, 5, 0, 74, 0, 66, 126, 13, 0, 5, 0, 74, 0, 249, 125, 13, 0, 5, 0, 74, 0, 25, 126, 13, 0, 5, 0, 74, 0, 28, 126, 13, 0, 5, 0, 74, 0, 252, 125, 13, 0, 5, 0, 74, 0, 13, 126, 13, 0, 5, 0, 74, 0, 22, 126, 13, 0, 5, 0, 74, 0, 16, 126, 13, 0, 5, 0, 74, 0, 10, 126, 13, 0, 5, 0, 74, 0, 19, 126, 13, 0, 5, 0, 74, 0, 34, 126, 13, 0, 5, 0, 74, 0, 37, 126, 13, 0, 5, 0, 74, 0, 189, 126, 13, 0, 5, 0, 74, 0, 188, 126, 13, 0, 5, 0, 74, 0, 117, 126, 13, 0, 5, 0, 74, 0, 120, 126, 13, 0, 5, 0, 74, 0, 123, 126, 13, 0, 5, 0, 70, 0, 189, 126, 13, 0, 5, 224, 251, 0, 0, 0, 0, 0, 5, 0, 71, 0, 189, 126, 13, 0, 5, 0, 252, 0, 0, 0, 0, 0, 5, 0, 72, 0, 189, 126, 13, 0, 5, 0, 253, 0, 0, 0, 0, 0, 5, 0, 73, 0, 189, 126, 13, 0, 5, 0, 254, 0, 0, 0, 0, 0, 5, 0, 74, 0, 189, 126, 13, 0, 5, 0, 255, 0, 0, 0, 0, 0, 5, 0, 70, 0, 190, 126, 13, 0, 5, 224, 251, 0, 0, 0, 0, 0, 5, 0, 71, 0, 190, 126, 13, 0, 5, 0, 252, 0, 0, 0, 0, 0, 5, 0, 72, 0, 190, 126, 13, 0, 5, 0, 253, 0, 0, 0, 0, 0, 5, 0, 73, 0, 190, 126, 13, 0, 5, 0, 254, 0, 0, 0, 0, 0, 5, 0, 74, 0, 190, 126, 13, 0, 5, 0, 255, 0, 0, 0, 0, 0, 5, 0, 70, 0, 188, 126, 13, 0, 5, 224, 251, 0, 0, 0, 0, 0, 5, 0, 71, 0, 188, 126, 13, 0, 5, 0, 252, 0, 0, 0, 0, 0, 5, 0, 72, 0, 188, 126, 13, 0, 5, 0, 253, 0, 0, 0, 0, 0, 5, 0, 73, 0, 188, 126, 13, 0, 5, 0, 254, 0, 0, 0, 0, 0, 5, 0, 74, 0, 188, 126, 13, 0, 5, 0, 255, 0, 0, 0, 0, 0, 5, 0, 70, 0, 40, 126, 13, 0, 5, 0, 70, 0, 39, 126, 13, 0, 5, 0, 71, 0, 40, 126, 13, 0, 5, 0, 71, 0, 39, 126, 13, 0, 5, 0, 72, 0, 40, 126, 13, 0, 5, 0, 72, 0, 39, 126, 13, 0, 5, 0, 73, 0, 40, 126, 13, 0, 5, 0, 73, 0, 39, 126, 13, 0, 5, 0, 74, 0, 40, 126, 13, 0, 5, 0, 74, 0, 39, 126, 13, 0, 5, 0, 70, 0, 63, 126, 13, 0, 5, 0, 70, 0, 62, 126, 13, 0, 5, 0, 71, 0, 63, 126, 13, 0, 5, 0, 71, 0, 62, 126, 13, 0, 5, 0, 72, 0, 63, 126, 13, 0, 5, 0, 72, 0, 62, 126, 13, 0, 5, 0, 73, 0, 63, 126, 13, 0, 5, 0, 73, 0, 62, 126, 13, 0, 5, 0, 74, 0, 63, 126, 13, 0, 5, 0, 74, 0, 62, 126, 13, 0, 5, 0, 70, 0, 204, 125, 13, 0, 5, 0, 70, 0, 208, 125, 13, 0, 5, 0, 71, 0, 206, 125, 13, 0, 5, 0, 71, 0, 208, 125, 13, 0, 5, 0, 72, 0, 206, 125, 13, 0, 5, 0, 72, 0, 208, 125, 13, 0, 5, 0, 73, 0, 206, 125, 13, 0, 5, 0, 73, 0, 208, 125, 13, 0, 5, 0, 74, 0, 206, 125, 13, 0, 5, 0, 74, 0, 208, 125, 13, 0, 5, 0, 75, 0, 206, 125, 13, 0, 5, 0, 75, 0, 208, 125, 13, 0, 5, 0, 76, 0, 206, 125, 13, 0, 5, 0, 70, 0, 55, 126, 13, 0, 5, 0, 70, 0, 54, 126, 13, 0, 5, 0, 71, 0, 55, 126, 13, 0, 5, 0, 71, 0, 54, 126, 13, 0, 5, 0, 72, 0, 55, 126, 13, 0, 5, 0, 72, 0, 54, 126, 13, 0, 5, 0, 73, 0, 55, 126, 13, 0, 5, 0, 73, 0, 54, 126, 13, 0, 5, 0, 74, 0, 55, 126, 13, 0, 5, 0, 74, 0, 54, 126, 13, 0, 5, 0, 70, 0, 50, 126, 13, 0, 5, 0, 70, 0, 49, 126, 13, 0, 5, 0, 71, 0, 50, 126, 13, 0, 5, 0, 71, 0, 49, 126, 13, 0, 5, 0, 72, 0, 50, 126, 13, 0, 5, 0, 72, 0, 49, 126, 13, 0, 5, 0, 73, 0, 50, 126, 13, 0, 5, 0, 73, 0, 49, 126, 13, 0, 5, 0, 74, 0, 50, 126, 13, 0, 5, 0, 74, 0, 49, 126, 13, 0, 5, 0, 70, 0, 228, 125, 13, 0, 5, 0, 70, 0, 227, 125, 13, 0, 5, 0, 71, 0, 228, 125, 13, 0, 5, 0, 71, 0, 227, 125, 13, 0, 5, 0, 72, 0, 228, 125, 13, 0, 5, 0, 72, 0, 227, 125, 13, 0, 5, 0, 73, 0, 228, 125, 13, 0, 5, 0, 73, 0, 227, 125, 13, 0, 5, 0, 74, 0, 228, 125, 13, 0, 5, 0, 74, 0, 227, 125, 13, 0, 5, 0, 70, 0, 46, 126, 13, 0, 5, 0, 70, 0, 45, 126, 13, 0, 5, 0, 71, 0, 46, 126, 13, 0, 5, 0, 71, 0, 45, 126, 13, 0, 5, 0, 72, 0, 46, 126, 13, 0, 5, 0, 72, 0, 45, 126, 13, 0, 5, 0, 73, 0, 46, 126, 13, 0, 5, 0, 73, 0, 45, 126, 13, 0, 5, 0, 74, 0, 46, 126, 13, 0, 5, 0, 74, 0, 45, 126, 13, 0, 5, 0, 70, 0, 102, 126, 13, 0, 5, 0, 70, 0, 101, 126, 13, 0, 5, 0, 71, 0, 102, 126, 13, 0, 5, 0, 71, 0, 101, 126, 13, 0, 5, 0, 72, 0, 102, 126, 13, 0, 5, 0, 72, 0, 101, 126, 13, 0, 5, 0, 73, 0, 102, 126, 13, 0, 5, 0, 73, 0, 101, 126, 13, 0, 5, 0, 74, 0, 102, 126, 13, 0, 5, 0, 74, 0, 101, 126, 13, 0, 5, 0, 70, 0, 105, 126, 13, 0, 5, 0, 70, 0, 104, 126, 13, 0, 5, 0, 71, 0, 105, 126, 13, 0, 5, 0, 71, 0, 104, 126, 13, 0, 5, 0, 72, 0, 105, 126, 13, 0, 5, 0, 72, 0, 104, 126, 13, 0, 5, 0, 73, 0, 105, 126, 13, 0, 5, 0, 73, 0, 104, 126, 13, 0, 5, 0, 74, 0, 105, 126, 13, 0, 5, 0, 74, 0, 104, 126, 13, 0, 5, 0, 70, 0, 43, 126, 13, 0, 5, 0, 70, 0, 42, 126, 13, 0, 5, 0, 71, 0, 43, 126, 13, 0, 5, 0, 71, 0, 42, 126, 13, 0, 5, 0, 72, 0, 43, 126, 13, 0, 5, 0, 72, 0, 42, 126, 13, 0, 5, 0, 73, 0, 43, 126, 13, 0, 5, 0, 73, 0, 42, 126, 13, 0, 5, 0, 74, 0, 43, 126, 13, 0, 5, 0, 74, 0, 42, 126, 13, 0, 5, 0, 70, 0, 222, 125, 13, 0, 5, 0, 70, 0, 221, 125, 13, 0, 5, 0, 71, 0, 222, 125, 13, 0, 5, 0, 71, 0, 221, 125, 13, 0, 5, 0, 72, 0, 222, 125, 13, 0, 5, 0, 72, 0, 221, 125, 13, 0, 5, 0, 73, 0, 222, 125, 13, 0, 5, 0, 73, 0, 221, 125, 13, 0, 5, 0, 74, 0, 222, 125, 13, 0, 5, 0, 74, 0, 221, 125, 13, 0, 5, 0, 70, 0, 225, 125, 13, 0, 5, 0, 70, 0, 224, 125, 13, 0, 5, 0, 71, 0, 225, 125, 13, 0, 5, 0, 71, 0, 224, 125, 13, 0, 5, 0, 72, 0, 225, 125, 13, 0, 5, 0, 72, 0, 224, 125, 13, 0, 5, 0, 73, 0, 225, 125, 13, 0, 5, 0, 73, 0, 224, 125, 13, 0, 5, 0, 74, 0, 225, 125, 13, 0, 5, 0, 74, 0, 224, 125, 13, 0, 5, 0, 70, 0, 237, 125, 13, 0, 5, 0, 70, 0, 236, 125, 13, 0, 5, 0, 71, 0, 237, 125, 13, 0, 5, 0, 71, 0, 236, 125, 13, 0, 5, 0, 72, 0, 237, 125, 13, 0, 5, 0, 72, 0, 236, 125, 13, 0, 5, 0, 73, 0, 237, 125, 13, 0, 5, 0, 73, 0, 236, 125, 13, 0, 5, 0, 74, 0, 237, 125, 13, 0, 5, 0, 74, 0, 236, 125, 13, 0, 5, 0, 70, 0, 231, 125, 13, 0, 5, 0, 70, 0, 230, 125, 13, 0, 5, 0, 71, 0, 231, 125, 13, 0, 5, 0, 71, 0, 230, 125, 13, 0, 5, 0, 72, 0, 231, 125, 13, 0, 5, 0, 72, 0, 230, 125, 13, 0, 5, 0, 73, 0, 231, 125, 13, 0, 5, 0, 73, 0, 230, 125, 13, 0, 5, 0, 74, 0, 231, 125, 13, 0, 5, 0, 74, 0, 230, 125, 13, 0, 5, 0, 70, 0, 216, 125, 13, 0, 5, 0, 70, 0, 215, 125, 13, 0, 5, 0, 71, 0, 216, 125, 13, 0, 5, 0, 71, 0, 215, 125, 13, 0, 5, 0, 72, 0, 216, 125, 13, 0, 5, 0, 72, 0, 215, 125, 13, 0, 5, 0, 73, 0, 216, 125, 13, 0, 5, 0, 73, 0, 215, 125, 13, 0, 5, 0, 74, 0, 216, 125, 13, 0, 5, 0, 74, 0, 215, 125, 13, 0, 5, 0, 70, 0, 219, 125, 13, 0, 5, 0, 70, 0, 218, 125, 13, 0, 5, 0, 71, 0, 219, 125, 13, 0, 5, 0, 71, 0, 218, 125, 13, 0, 5, 0, 72, 0, 219, 125, 13, 0, 5, 0, 72, 0, 218, 125, 13, 0, 5, 0, 73, 0, 219, 125, 13, 0, 5, 0, 73, 0, 218, 125, 13, 0, 5, 0, 74, 0, 219, 125, 13, 0, 5, 0, 74, 0, 218, 125, 13, 0, 5, 0, 70, 0, 151, 126, 13, 0, 5, 0, 70, 0, 150, 126, 13, 0, 5, 0, 71, 0, 151, 126, 13, 0, 5, 0, 71, 0, 150, 126, 13, 0, 5, 0, 72, 0, 151, 126, 13, 0, 5, 0, 72, 0, 150, 126, 13, 0, 5, 0, 73, 0, 151, 126, 13, 0, 5, 0, 73, 0, 150, 126, 13, 0, 5, 0, 74, 0, 151, 126, 13, 0, 5, 0, 74, 0, 150, 126, 13, 0, 5, 0, 70, 0, 163, 126, 13, 0, 5, 0, 70, 0, 162, 126, 13, 0, 5, 0, 71, 0, 163, 126, 13, 0, 5, 0, 71, 0, 162, 126, 13, 0, 5, 0, 72, 0, 163, 126, 13, 0, 5, 0, 72, 0, 162, 126, 13, 0, 5, 0, 73, 0, 163, 126, 13, 0, 5, 0, 73, 0, 162, 126, 13, 0, 5, 0, 74, 0, 163, 126, 13, 0, 5, 0, 74, 0, 162, 126, 13, 0, 5, 0, 70, 0, 166, 126, 13, 0, 5, 0, 70, 0, 165, 126, 13, 0, 5, 0, 71, 0, 166, 126, 13, 0, 5, 0, 71, 0, 165, 126, 13, 0, 5, 0, 72, 0, 166, 126, 13, 0, 5, 0, 72, 0, 165, 126, 13, 0, 5, 0, 73, 0, 166, 126, 13, 0, 5, 0, 73, 0, 165, 126, 13, 0, 5, 0, 74, 0, 166, 126, 13, 0, 5, 0, 74, 0, 165, 126, 13, 0, 5, 0, 70, 0, 108, 126, 13, 0, 5, 0, 70, 0, 107, 126, 13, 0, 5, 0, 71, 0, 108, 126, 13, 0, 5, 0, 71, 0, 107, 126, 13, 0, 5, 0, 72, 0, 108, 126, 13, 0, 5, 0, 72, 0, 107, 126, 13, 0, 5, 0, 73, 0, 108, 126, 13, 0, 5, 0, 73, 0, 107, 126, 13, 0, 5, 0, 74, 0, 108, 126, 13, 0, 5, 0, 74, 0, 107, 126, 13, 0, 5, 0, 70, 0, 240, 125, 13, 0, 5, 0, 70, 0, 239, 125, 13, 0, 5, 0, 71, 0, 240, 125, 13, 0, 5, 0, 71, 0, 239, 125, 13, 0, 5, 0, 72, 0, 240, 125, 13, 0, 5, 0, 72, 0, 239, 125, 13, 0, 5, 0, 73, 0, 240, 125, 13, 0, 5, 0, 73, 0, 239, 125, 13, 0, 5, 0, 74, 0, 240, 125, 13, 0, 5, 0, 74, 0, 239, 125, 13, 0, 5, 0, 70, 0, 60, 126, 13, 0, 5, 0, 70, 0, 59, 126, 13, 0, 5, 0, 71, 0, 60, 126, 13, 0, 5, 0, 71, 0, 59, 126, 13, 0, 5, 0, 72, 0, 60, 126, 13, 0, 5, 0, 72, 0, 59, 126, 13, 0, 5, 0, 73, 0, 60, 126, 13, 0, 5, 0, 73, 0, 59, 126, 13, 0, 5, 0, 74, 0, 60, 126, 13, 0, 5, 0, 74, 0, 59, 126, 13, 0, 5, 0, 70, 0, 243, 125, 13, 0, 5, 0, 70, 0, 242, 125, 13, 0, 5, 0, 71, 0, 243, 125, 13, 0, 5, 0, 71, 0, 242, 125, 13, 0, 5, 0, 72, 0, 243, 125, 13, 0, 5, 0, 72, 0, 242, 125, 13, 0, 5, 0, 73, 0, 243, 125, 13, 0, 5, 0, 73, 0, 242, 125, 13, 0, 5, 0, 74, 0, 243, 125, 13, 0, 5, 0, 74, 0, 242, 125, 13, 0, 5, 0, 70, 0, 169, 126, 13, 0, 5, 0, 70, 0, 168, 126, 13, 0, 5, 0, 71, 0, 169, 126, 13, 0, 5, 0, 71, 0, 168, 126, 13, 0, 5, 0, 72, 0, 169, 126, 13, 0, 5, 0, 72, 0, 168, 126, 13, 0, 5, 0, 73, 0, 169, 126, 13, 0, 5, 0, 73, 0, 168, 126, 13, 0, 5, 0, 74, 0, 169, 126, 13, 0, 5, 0, 74, 0, 168, 126, 13, 0, 5, 0, 70, 0, 181, 126, 13, 0, 5, 0, 70, 0, 180, 126, 13, 0, 5, 0, 71, 0, 181, 126, 13, 0, 5, 0, 71, 0, 180, 126, 13, 0, 5, 0, 72, 0, 181, 126, 13, 0, 5, 0, 72, 0, 180, 126, 13, 0, 5, 0, 73, 0, 181, 126, 13, 0, 5, 0, 73, 0, 180, 126, 13, 0, 5, 0, 74, 0, 181, 126, 13, 0, 5, 0, 74, 0, 180, 126, 13, 0, 5, 0, 70, 0, 175, 126, 13, 0, 5, 0, 70, 0, 174, 126, 13, 0, 5, 0, 71, 0, 175, 126, 13, 0, 5, 0, 71, 0, 174, 126, 13, 0, 5, 0, 72, 0, 175, 126, 13, 0, 5, 0, 72, 0, 174, 126, 13, 0, 5, 0, 73, 0, 175, 126, 13, 0, 5, 0, 73, 0, 174, 126, 13, 0, 5, 0, 74, 0, 175, 126, 13, 0, 5, 0, 74, 0, 174, 126, 13, 0, 5, 0, 70, 0, 178, 126, 13, 0, 5, 0, 70, 0, 177, 126, 13, 0, 5, 0, 71, 0, 178, 126, 13, 0, 5, 0, 71, 0, 177, 126, 13, 0, 5, 0, 72, 0, 178, 126, 13, 0, 5, 0, 72, 0, 177, 126, 13, 0, 5, 0, 73, 0, 178, 126, 13, 0, 5, 0, 73, 0, 177, 126, 13, 0, 5, 0, 74, 0, 178, 126, 13, 0, 5, 0, 74, 0, 177, 126, 13, 0, 5, 0, 70, 0, 75, 126, 13, 0, 5, 0, 70, 0, 74, 126, 13, 0, 5, 0, 71, 0, 75, 126, 13, 0, 5, 0, 71, 0, 74, 126, 13, 0, 5, 0, 72, 0, 75, 126, 13, 0, 5, 0, 72, 0, 74, 126, 13, 0, 5, 0, 73, 0, 75, 126, 13, 0, 5, 0, 73, 0, 74, 126, 13, 0, 5, 0, 74, 0, 75, 126, 13, 0, 5, 0, 74, 0, 74, 126, 13, 0, 5, 0, 70, 0, 78, 126, 13, 0, 5, 0, 70, 0, 77, 126, 13, 0, 5, 0, 71, 0, 78, 126, 13, 0, 5, 0, 71, 0, 77, 126, 13, 0, 5, 0, 72, 0, 78, 126, 13, 0, 5, 0, 72, 0, 77, 126, 13, 0, 5, 0, 73, 0, 78, 126, 13, 0, 5, 0, 73, 0, 77, 126, 13, 0, 5, 0, 74, 0, 78, 126, 13, 0, 5, 0, 74, 0, 77, 126, 13, 0, 5, 0, 70, 0, 111, 126, 13, 0, 5, 0, 70, 0, 110, 126, 13, 0, 5, 0, 71, 0, 111, 126, 13, 0, 5, 0, 71, 0, 110, 126, 13, 0, 5, 0, 72, 0, 111, 126, 13, 0, 5, 0, 72, 0, 110, 126, 13, 0, 5, 0, 73, 0, 111, 126, 13, 0, 5, 0, 73, 0, 110, 126, 13, 0, 5, 0, 74, 0, 111, 126, 13, 0, 5, 0, 74, 0, 110, 126, 13, 0, 5, 0, 70, 0, 114, 126, 13, 0, 5, 0, 70, 0, 113, 126, 13, 0, 5, 0, 71, 0, 114, 126, 13, 0, 5, 0, 71, 0, 113, 126, 13, 0, 5, 0, 72, 0, 114, 126, 13, 0, 5, 0, 72, 0, 113, 126, 13, 0, 5, 0, 73, 0, 114, 126, 13, 0, 5, 0, 73, 0, 113, 126, 13, 0, 5, 0, 74, 0, 114, 126, 13, 0, 5, 0, 74, 0, 113, 126, 13, 0, 5, 0, 70, 0, 234, 125, 13, 0, 5, 0, 70, 0, 233, 125, 13, 0, 5, 0, 71, 0, 234, 125, 13, 0, 5, 0, 71, 0, 233, 125, 13, 0, 5, 0, 72, 0, 234, 125, 13, 0, 5, 0, 72, 0, 233, 125, 13, 0, 5, 0, 73, 0, 234, 125, 13, 0, 5, 0, 73, 0, 233, 125, 13, 0, 5, 0, 74, 0, 234, 125, 13, 0, 5, 0, 74, 0, 233, 125, 13, 0, 5, 0, 70, 0, 244, 125, 13, 0, 5, 0, 70, 0, 253, 125, 13, 0, 5, 0, 70, 0, 29, 126, 13, 0, 5, 0, 70, 0, 2, 126, 13, 0, 5, 0, 70, 0, 5, 126, 13, 0, 5, 0, 70, 0, 68, 126, 13, 0, 5, 0, 70, 0, 72, 126, 13, 0, 5, 0, 70, 0, 247, 125, 13, 0, 5, 0, 70, 0, 23, 126, 13, 0, 5, 0, 70, 0, 26, 126, 13, 0, 5, 0, 70, 0, 250, 125, 13, 0, 5, 0, 70, 0, 11, 126, 13, 0, 5, 0, 70, 0, 20, 126, 13, 0, 5, 0, 70, 0, 14, 126, 13, 0, 5, 0, 70, 0, 8, 126, 13, 0, 5, 0, 70, 0, 17, 126, 13, 0, 5, 0, 70, 0, 32, 126, 13, 0, 5, 0, 70, 0, 35, 126, 13, 0, 5, 0, 70, 0, 187, 126, 13, 0, 5, 0, 70, 0, 115, 126, 13, 0, 5, 0, 70, 0, 118, 126, 13, 0, 5, 0, 70, 0, 121, 126, 13, 0, 5, 0, 71, 0, 244, 125, 13, 0, 5, 0, 71, 0, 253, 125, 13, 0, 5, 0, 71, 0, 29, 126, 13, 0, 5, 0, 71, 0, 2, 126, 13, 0, 5, 0, 71, 0, 5, 126, 13, 0, 5, 0, 71, 0, 68, 126, 13, 0, 5, 0, 71, 0, 72, 126, 13, 0, 5, 0, 71, 0, 247, 125, 13, 0, 5, 0, 71, 0, 23, 126, 13, 0, 5, 0, 71, 0, 26, 126, 13, 0, 5, 0, 71, 0, 250, 125, 13, 0, 5, 0, 71, 0, 11, 126, 13, 0, 5, 0, 71, 0, 20, 126, 13, 0, 5, 0, 71, 0, 14, 126, 13, 0, 5, 0, 71, 0, 8, 126, 13, 0, 5, 0, 71, 0, 17, 126, 13, 0, 5, 0, 71, 0, 32, 126, 13, 0, 5, 0, 71, 0, 35, 126, 13, 0, 5, 0, 71, 0, 187, 126, 13, 0, 5, 0, 71, 0, 115, 126, 13, 0, 5, 0, 71, 0, 118, 126, 13, 0, 5, 0, 71, 0, 121, 126, 13, 0, 5, 0, 72, 0, 244, 125, 13, 0, 5, 0, 72, 0, 253, 125, 13, 0, 5, 0, 72, 0, 29, 126, 13, 0, 5, 0, 72, 0, 2, 126, 13, 0, 5, 0, 72, 0, 5, 126, 13, 0, 5, 0, 72, 0, 68, 126, 13, 0, 5, 0, 72, 0, 72, 126, 13, 0, 5, 0, 72, 0, 247, 125, 13, 0, 5, 0, 72, 0, 23, 126, 13, 0, 5, 0, 72, 0, 26, 126, 13, 0, 5, 0, 72, 0, 250, 125, 13, 0, 5, 0, 72, 0, 11, 126, 13, 0, 5, 0, 72, 0, 20, 126, 13, 0, 5, 0, 72, 0, 14, 126, 13, 0, 5, 0, 72, 0, 8, 126, 13, 0, 5, 0, 72, 0, 17, 126, 13, 0, 5, 0, 72, 0, 32, 126, 13, 0, 5, 0, 72, 0, 35, 126, 13, 0, 5, 0, 72, 0, 187, 126, 13, 0, 5, 0, 72, 0, 115, 126, 13, 0, 5, 0, 72, 0, 118, 126, 13, 0, 5, 0, 72, 0, 121, 126, 13, 0, 5, 0, 73, 0, 244, 125, 13, 0, 5, 0, 73, 0, 253, 125, 13, 0, 5, 0, 73, 0, 29, 126, 13, 0, 5, 0, 73, 0, 2, 126, 13, 0, 5, 0, 73, 0, 5, 126, 13, 0, 5, 0, 73, 0, 68, 126, 13, 0, 5, 0, 73, 0, 72, 126, 13, 0, 5, 0, 73, 0, 247, 125, 13, 0, 5, 0, 73, 0, 23, 126, 13, 0, 5, 0, 73, 0, 26, 126, 13, 0, 5, 0, 73, 0, 250, 125, 13, 0, 5, 0, 73, 0, 11, 126, 13, 0, 5, 0, 73, 0, 20, 126, 13, 0, 5, 0, 73, 0, 14, 126, 13, 0, 5, 0, 73, 0, 8, 126, 13, 0, 5, 0, 73, 0, 17, 126, 13, 0, 5, 0, 73, 0, 32, 126, 13, 0, 5, 0, 73, 0, 35, 126, 13, 0, 5, 0, 73, 0, 187, 126, 13, 0, 5, 0, 73, 0, 115, 126, 13, 0, 5, 0, 73, 0, 118, 126, 13, 0, 5, 0, 73, 0, 121, 126, 13, 0, 5, 0, 74, 0, 244, 125, 13, 0, 5, 0, 74, 0, 253, 125, 13, 0, 5, 0, 74, 0, 29, 126, 13, 0, 5, 0, 74, 0, 2, 126, 13, 0, 5, 0, 74, 0, 5, 126, 13, 0, 5, 0, 74, 0, 68, 126, 13, 0, 5, 0, 74, 0, 72, 126, 13, 0, 5, 0, 74, 0, 247, 125, 13, 0, 5, 0, 74, 0, 23, 126, 13, 0, 5, 0, 74, 0, 26, 126, 13, 0, 5, 0, 74, 0, 250, 125, 13, 0, 5, 0, 74, 0, 11, 126, 13, 0, 5, 0, 74, 0, 20, 126, 13, 0, 5, 0, 74, 0, 14, 126, 13, 0, 5, 0, 74, 0, 8, 126, 13, 0, 5, 0, 74, 0, 17, 126, 13, 0, 5, 0, 74, 0, 32, 126, 13, 0, 5, 0, 74, 0, 35, 126, 13, 0, 5, 0, 74, 0, 187, 126, 13, 0, 5, 0, 74, 0, 115, 126, 13, 0, 5, 0, 74, 0, 118, 126, 13, 0, 5, 0, 74, 0, 121, 126, 13, 0, 5, 0, 70, 0, 206, 125, 13, 0, 5, 0, 70, 0, 135, 126, 13, 0, 5, 0, 70, 0, 134, 126, 13, 0, 5, 0, 71, 0, 135, 126, 13, 0, 5, 0, 71, 0, 134, 126, 13, 0, 5, 0, 72, 0, 135, 126, 13, 0, 5, 0, 72, 0, 134, 126, 13, 0, 5, 0, 73, 0, 135, 126, 13, 0, 5, 0, 73, 0, 134, 126, 13, 0, 5, 0, 74, 0, 135, 126, 13, 0, 5, 0, 74, 0, 134, 126, 13, 0, 5, 0, 70, 0, 138, 126, 13, 0, 5, 0, 70, 0, 137, 126, 13, 0, 5, 0, 71, 0, 138, 126, 13, 0, 5, 0, 71, 0, 137, 126, 13, 0, 5, 0, 72, 0, 138, 126, 13, 0, 5, 0, 72, 0, 137, 126, 13, 0, 5, 0, 73, 0, 138, 126, 13, 0, 5, 0, 73, 0, 137, 126, 13, 0, 5, 0, 74, 0, 138, 126, 13, 0, 5, 0, 74, 0, 137, 126, 13, 0, 5, 0, 70, 0, 184, 126, 13, 0, 5, 0, 70, 0, 183, 126, 13, 0, 5, 0, 71, 0, 184, 126, 13, 0, 5, 0, 71, 0, 183, 126, 13, 0, 5, 0, 72, 0, 184, 126, 13, 0, 5, 0, 72, 0, 183, 126, 13, 0, 5, 0, 73, 0, 184, 126, 13, 0, 5, 0, 73, 0, 183, 126, 13, 0, 5, 0, 74, 0, 184, 126, 13, 0, 5, 0, 74, 0, 183, 126, 13, 0, 5, 0, 70, 0, 81, 126, 13, 0, 5, 0, 70, 0, 80, 126, 13, 0, 5, 0, 71, 0, 81, 126, 13, 0, 5, 0, 71, 0, 80, 126, 13, 0, 5, 0, 72, 0, 81, 126, 13, 0, 5, 0, 72, 0, 80, 126, 13, 0, 5, 0, 73, 0, 81, 126, 13, 0, 5, 0, 73, 0, 80, 126, 13, 0, 5, 0, 74, 0, 81, 126, 13, 0, 5, 0, 74, 0, 80, 126, 13, 0, 5, 0, 70, 0, 84, 126, 13, 0, 5, 0, 70, 0, 83, 126, 13, 0, 5, 0, 71, 0, 84, 126, 13, 0, 5, 0, 71, 0, 83, 126, 13, 0, 5, 0, 72, 0, 84, 126, 13, 0, 5, 0, 72, 0, 83, 126, 13, 0, 5, 0, 73, 0, 84, 126, 13, 0, 5, 0, 73, 0, 83, 126, 13, 0, 5, 0, 74, 0, 84, 126, 13, 0, 5, 0, 74, 0, 83, 126, 13, 0, 5, 0, 70, 0, 87, 126, 13, 0, 5, 0, 70, 0, 86, 126, 13, 0, 5, 0, 71, 0, 87, 126, 13, 0, 5, 0, 71, 0, 86, 126, 13, 0, 5, 0, 72, 0, 87, 126, 13, 0, 5, 0, 72, 0, 86, 126, 13, 0, 5, 0, 73, 0, 87, 126, 13, 0, 5, 0, 73, 0, 86, 126, 13, 0, 5, 0, 74, 0, 87, 126, 13, 0, 5, 0, 74, 0, 86, 126, 13, 0, 5, 0, 70, 0, 90, 126, 13, 0, 5, 0, 70, 0, 89, 126, 13, 0, 5, 0, 71, 0, 90, 126, 13, 0, 5, 0, 71, 0, 89, 126, 13, 0, 5, 0, 72, 0, 90, 126, 13, 0, 5, 0, 72, 0, 89, 126, 13, 0, 5, 0, 73, 0, 90, 126, 13, 0, 5, 0, 73, 0, 89, 126, 13, 0, 5, 0, 74, 0, 90, 126, 13, 0, 5, 0, 74, 0, 89, 126, 13, 0, 5, 0, 70, 0, 93, 126, 13, 0, 5, 0, 70, 0, 92, 126, 13, 0, 5, 0, 71, 0, 93, 126, 13, 0, 5, 0, 71, 0, 92, 126, 13, 0, 5, 0, 72, 0, 93, 126, 13, 0, 5, 0, 72, 0, 92, 126, 13, 0, 5, 0, 73, 0, 93, 126, 13, 0, 5, 0, 73, 0, 92, 126, 13, 0, 5, 0, 74, 0, 93, 126, 13, 0, 5, 0, 74, 0, 92, 126, 13, 0, 0, 0, 0, 193, 71, 130, 13, 5, 158, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 201, 192, 2, 0, 201, 32, 4, 0, 201, 128, 5, 0, 201, 224, 6, 0, 201, 64, 8, 0, 201, 160, 9, 0, 201, 0, 11, 0, 201, 96, 12, 0, 201, 192, 13, 0, 201, 32, 15, 0, 150, 10, 5, 5, 1, 0, 227, 32, 130, 205, 193, 170, 15, 254, 48, 0, 227, 32, 130, 205, 193, 170, 128, 10, 5, 5, 1, 0, 227, 32, 130, 205, 193, 171, 15, 254, 48, 0, 227, 32, 130, 205, 193, 171, 0, 19, 5, 5, 1, 0, 227, 32, 130, 205, 193, 172, 15, 254, 48, 0, 227, 32, 130, 205, 193, 172, 0, 21, 5, 5, 1, 0, 227, 32, 130, 205, 193, 173, 15, 254, 48, 0, 227, 32, 130, 205, 193, 173, 0, 23, 5, 5, 1, 0, 227, 32, 130, 205, 193, 174, 15, 254, 48, 0, 227, 32, 130, 205, 193, 174, 0, 25, 5, 5, 1, 0, 227, 32, 130, 205, 193, 175, 15, 254, 48, 0, 227, 32, 130, 205, 193, 175, 0, 27, 5, 5, 1, 0, 227, 32, 130, 205, 193, 176, 15, 254, 48, 0, 227, 32, 130, 205, 193, 176, 0, 29, 5, 5, 1, 0, 227, 32, 130, 205, 193, 177, 15, 254, 48, 0, 227, 32, 130, 205, 193, 177, 0, 31, 5, 5, 1, 0, 227, 32, 130, 205, 193, 178, 15, 254, 48, 0, 227, 32, 130, 205, 193, 178, 0, 33, 5, 5, 1, 0, 227, 32, 130, 205, 193, 179, 15, 254, 48, 0, 227, 32, 130, 205, 193, 179, 0, 35, 5, 5, 1, 0, 227, 32, 130, 205, 193, 180, 15, 254, 48, 0, 227, 32, 130, 205, 193, 180, 0, 37, 5, 5, 1, 0, 227, 32, 130, 205, 193, 181, 15, 254, 48, 0, 227, 32, 130, 205, 193, 181, 126, 13, 193, 155, 2, 0, 13, 32, 5, 0, 60, 216, 10, 0, 15, 254, 48, 0, 13, 32, 1, 0, 64, 38, 126, 205, 193, 157, 66, 38, 126, 205, 193, 156, 4, 0, 251, 223, 16, 0, 252, 223, 22, 0, 253, 223, 29, 0, 254, 223, 36, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 1, 192, 198, 33, 66, 38, 1, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 198, 161, 66, 38, 0, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 0, 192, 198, 97, 66, 38, 0, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 0, 192, 198, 161, 66, 38, 0, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 0, 192, 198, 225, 66, 38, 1, 192, 198, 1, 126, 13, 193, 124, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 4, 192, 198, 65, 66, 38, 4, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 3, 192, 198, 65, 66, 38, 3, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 3, 192, 198, 129, 66, 38, 3, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 3, 192, 198, 193, 66, 38, 3, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 4, 192, 198, 1, 66, 38, 4, 192, 198, 33, 1, 0, 64, 38, 126, 205, 193, 126, 66, 38, 126, 205, 193, 125, 126, 13, 193, 146, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 5, 192, 198, 129, 66, 38, 5, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 4, 192, 198, 129, 66, 38, 4, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 4, 192, 198, 193, 66, 38, 4, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 5, 192, 198, 1, 66, 38, 5, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 5, 192, 198, 65, 66, 38, 5, 192, 198, 97, 1, 0, 64, 38, 126, 205, 193, 148, 66, 38, 126, 205, 193, 147, 126, 13, 193, 152, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 6, 192, 198, 193, 66, 38, 6, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 5, 192, 198, 193, 66, 38, 5, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 6, 192, 198, 1, 66, 38, 6, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 6, 192, 198, 65, 66, 38, 6, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 6, 192, 198, 129, 66, 38, 6, 192, 198, 161, 1, 0, 64, 38, 126, 205, 193, 154, 66, 38, 126, 205, 193, 153, 126, 13, 193, 158, 2, 0, 13, 32, 5, 0, 60, 216, 10, 0, 15, 254, 48, 0, 13, 32, 1, 0, 64, 38, 126, 205, 193, 160, 66, 38, 126, 205, 193, 159, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 8, 192, 198, 1, 66, 38, 8, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 7, 192, 198, 1, 66, 38, 7, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 7, 192, 198, 65, 66, 38, 7, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 7, 192, 198, 129, 66, 38, 7, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 7, 192, 198, 193, 66, 38, 7, 192, 198, 225, 126, 13, 193, 143, 2, 0, 13, 32, 5, 0, 60, 216, 10, 0, 15, 254, 48, 0, 13, 32, 1, 0, 64, 38, 126, 205, 193, 145, 66, 38, 126, 205, 193, 144, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 9, 192, 198, 65, 66, 38, 9, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 8, 192, 198, 65, 66, 38, 8, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 8, 192, 198, 129, 66, 38, 8, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 8, 192, 198, 193, 66, 38, 8, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 9, 192, 198, 1, 66, 38, 9, 192, 198, 33, 131, 13, 193, 6, 1, 0, 13, 32, 3, 0, 15, 254, 48, 0, 13, 32, 1, 0, 167, 38, 131, 205, 193, 8, 60, 216, 48, 0, 8, 223, 131, 205, 193, 7, 131, 13, 193, 5, 1, 0, 13, 32, 39, 0, 64, 219, 51, 0, 103, 220, 64, 219, 98, 220, 64, 219, 2, 0, 101, 220, 12, 0, 115, 220, 19, 0, 119, 220, 53, 0, 64, 219, 108, 220, 64, 219, 115, 220, 64, 219, 127, 220, 131, 205, 193, 38, 53, 0, 64, 219, 110, 220, 64, 219, 103, 220, 64, 219, 127, 220, 131, 205, 193, 36, 53, 0, 64, 219, 99, 220, 64, 219, 116, 220, 64, 219, 127, 220, 131, 205, 193, 37, 48, 0, 32, 38, 131, 205, 193, 9, 126, 13, 193, 247, 49, 0, 13, 32, 27, 43, 126, 205, 193, 248, 126, 13, 193, 239, 50, 0, 13, 32, 62, 216, 186, 221, 126, 205, 193, 241, 127, 13, 193, 34, 49, 0, 13, 32, 68, 39, 127, 205, 193, 35, 125, 13, 193, 197, 1, 0, 13, 32, 7, 0, 15, 254, 50, 0, 13, 32, 61, 216, 232, 221, 125, 205, 193, 144, 49, 0, 61, 216, 232, 221, 125, 205, 193, 144, 125, 13, 193, 206, 1, 0, 13, 32, 180, 1, 60, 216, 4, 0, 251, 223, 92, 0, 252, 223, 175, 0, 253, 223, 2, 1, 254, 223, 85, 1, 255, 223, 48, 0, 13, 32, 5, 0, 60, 216, 71, 0, 60, 216, 21, 0, 61, 216, 46, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 22, 192, 198, 65, 188, 221, 22, 192, 198, 97, 189, 221, 22, 192, 198, 129, 50, 0, 13, 32, 61, 216, 104, 220, 22, 192, 198, 33, 7, 0, 164, 223, 12, 0, 164, 223, 20, 192, 198, 225, 168, 223, 21, 192, 198, 1, 235, 223, 21, 192, 198, 33, 237, 223, 21, 192, 198, 65, 62, 223, 20, 192, 198, 97, 115, 223, 20, 192, 198, 129, 124, 223, 20, 192, 198, 161, 147, 223, 20, 192, 198, 193, 5, 0, 44, 221, 9, 0, 44, 221, 21, 192, 198, 193, 128, 222, 21, 192, 198, 225, 146, 222, 22, 192, 198, 1, 187, 220, 21, 192, 198, 97, 188, 220, 21, 192, 198, 129, 39, 221, 21, 192, 198, 161, 149, 38, 20, 192, 198, 1, 150, 38, 20, 192, 198, 33, 8, 39, 20, 192, 198, 65, 48, 0, 13, 32, 5, 0, 60, 216, 71, 0, 60, 216, 21, 0, 61, 216, 46, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 11, 192, 198, 193, 188, 221, 11, 192, 198, 225, 189, 221, 12, 192, 198, 1, 50, 0, 13, 32, 61, 216, 104, 220, 11, 192, 198, 161, 7, 0, 164, 223, 12, 0, 164, 223, 10, 192, 198, 97, 168, 223, 10, 192, 198, 129, 235, 223, 10, 192, 198, 161, 237, 223, 10, 192, 198, 193, 62, 223, 9, 192, 198, 225, 115, 223, 10, 192, 198, 1, 124, 223, 10, 192, 198, 33, 147, 223, 10, 192, 198, 65, 5, 0, 44, 221, 9, 0, 44, 221, 11, 192, 198, 65, 128, 222, 11, 192, 198, 97, 146, 222, 11, 192, 198, 129, 187, 220, 10, 192, 198, 225, 188, 220, 11, 192, 198, 1, 39, 221, 11, 192, 198, 33, 149, 38, 9, 192, 198, 129, 150, 38, 9, 192, 198, 161, 8, 39, 9, 192, 198, 193, 48, 0, 13, 32, 5, 0, 60, 216, 71, 0, 60, 216, 21, 0, 61, 216, 46, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 14, 192, 198, 97, 188, 221, 14, 192, 198, 129, 189, 221, 14, 192, 198, 161, 50, 0, 13, 32, 61, 216, 104, 220, 14, 192, 198, 65, 7, 0, 164, 223, 12, 0, 164, 223, 13, 192, 198, 1, 168, 223, 13, 192, 198, 33, 235, 223, 13, 192, 198, 65, 237, 223, 13, 192, 198, 97, 62, 223, 12, 192, 198, 129, 115, 223, 12, 192, 198, 161, 124, 223, 12, 192, 198, 193, 147, 223, 12, 192, 198, 225, 5, 0, 44, 221, 9, 0, 44, 221, 13, 192, 198, 225, 128, 222, 14, 192, 198, 1, 146, 222, 14, 192, 198, 33, 187, 220, 13, 192, 198, 129, 188, 220, 13, 192, 198, 161, 39, 221, 13, 192, 198, 193, 149, 38, 12, 192, 198, 33, 150, 38, 12, 192, 198, 65, 8, 39, 12, 192, 198, 97, 48, 0, 13, 32, 5, 0, 60, 216, 71, 0, 60, 216, 21, 0, 61, 216, 46, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 17, 192, 198, 1, 188, 221, 17, 192, 198, 33, 189, 221, 17, 192, 198, 65, 50, 0, 13, 32, 61, 216, 104, 220, 16, 192, 198, 225, 7, 0, 164, 223, 12, 0, 164, 223, 15, 192, 198, 161, 168, 223, 15, 192, 198, 193, 235, 223, 15, 192, 198, 225, 237, 223, 16, 192, 198, 1, 62, 223, 15, 192, 198, 33, 115, 223, 15, 192, 198, 65, 124, 223, 15, 192, 198, 97, 147, 223, 15, 192, 198, 129, 5, 0, 44, 221, 9, 0, 44, 221, 16, 192, 198, 129, 128, 222, 16, 192, 198, 161, 146, 222, 16, 192, 198, 193, 187, 220, 16, 192, 198, 33, 188, 220, 16, 192, 198, 65, 39, 221, 16, 192, 198, 97, 149, 38, 14, 192, 198, 193, 150, 38, 14, 192, 198, 225, 8, 39, 15, 192, 198, 1, 48, 0, 13, 32, 5, 0, 60, 216, 71, 0, 60, 216, 21, 0, 61, 216, 46, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 19, 192, 198, 161, 188, 221, 19, 192, 198, 193, 189, 221, 19, 192, 198, 225, 50, 0, 13, 32, 61, 216, 104, 220, 19, 192, 198, 129, 7, 0, 164, 223, 12, 0, 164, 223, 18, 192, 198, 65, 168, 223, 18, 192, 198, 97, 235, 223, 18, 192, 198, 129, 237, 223, 18, 192, 198, 161, 62, 223, 17, 192, 198, 193, 115, 223, 17, 192, 198, 225, 124, 223, 18, 192, 198, 1, 147, 223, 18, 192, 198, 33, 5, 0, 44, 221, 9, 0, 44, 221, 19, 192, 198, 33, 128, 222, 19, 192, 198, 65, 146, 222, 19, 192, 198, 97, 187, 220, 18, 192, 198, 193, 188, 220, 18, 192, 198, 225, 39, 221, 19, 192, 198, 1, 149, 38, 17, 192, 198, 97, 150, 38, 17, 192, 198, 129, 8, 39, 17, 192, 198, 161, 6, 0, 100, 39, 178, 0, 100, 39, 15, 0, 60, 216, 44, 0, 61, 216, 69, 0, 62, 216, 2, 0, 175, 221, 126, 205, 193, 116, 188, 221, 126, 205, 193, 119, 189, 221, 126, 205, 193, 122, 1, 0, 13, 32, 15, 0, 15, 254, 49, 0, 13, 32, 61, 216, 1, 0, 104, 220, 126, 205, 193, 197, 139, 220, 50, 0, 13, 32, 61, 216, 104, 220, 126, 205, 193, 193, 48, 0, 61, 216, 1, 0, 104, 220, 126, 205, 193, 197, 139, 220, 50, 0, 13, 32, 61, 216, 104, 220, 126, 205, 193, 193, 7, 0, 164, 223, 12, 0, 164, 223, 126, 205, 193, 24, 168, 223, 126, 205, 193, 27, 235, 223, 125, 205, 193, 251, 237, 223, 126, 205, 193, 12, 62, 223, 126, 205, 193, 3, 115, 223, 126, 205, 193, 6, 124, 223, 126, 205, 193, 67, 147, 223, 125, 205, 193, 248, 9, 0, 188, 220, 15, 0, 188, 220, 126, 205, 193, 15, 39, 221, 126, 205, 193, 9, 44, 221, 126, 205, 193, 18, 128, 222, 126, 205, 193, 33, 146, 222, 126, 205, 193, 36, 102, 220, 9, 0, 103, 220, 15, 0, 104, 220, 25, 0, 105, 220, 50, 0, 187, 220, 126, 205, 193, 21, 242, 127, 126, 13, 193, 215, 13, 32, 61, 216, 102, 220, 126, 205, 193, 216, 241, 127, 126, 13, 193, 217, 13, 32, 61, 216, 1, 0, 102, 220, 126, 205, 193, 218, 103, 220, 126, 205, 193, 219, 49, 0, 13, 32, 61, 216, 1, 0, 102, 220, 13, 0, 103, 220, 241, 127, 126, 13, 193, 206, 13, 32, 61, 216, 1, 0, 102, 220, 126, 205, 193, 207, 103, 220, 126, 205, 193, 209, 242, 127, 126, 13, 193, 205, 13, 32, 61, 216, 102, 220, 126, 205, 193, 208, 49, 0, 13, 32, 61, 216, 1, 0, 102, 220, 13, 0, 103, 220, 241, 127, 126, 13, 193, 201, 13, 32, 61, 216, 1, 0, 102, 220, 126, 205, 193, 202, 103, 220, 126, 205, 193, 204, 242, 127, 126, 13, 193, 200, 13, 32, 61, 216, 102, 220, 126, 205, 193, 203, 149, 38, 125, 205, 193, 245, 150, 38, 125, 205, 193, 254, 8, 39, 126, 205, 193, 30, 125, 13, 193, 208, 1, 0, 13, 32, 200, 1, 60, 216, 4, 0, 251, 223, 96, 0, 252, 223, 183, 0, 253, 223, 14, 1, 254, 223, 101, 1, 255, 223, 48, 0, 13, 32, 5, 0, 60, 216, 75, 0, 60, 216, 25, 0, 61, 216, 50, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 36, 192, 198, 1, 188, 221, 36, 192, 198, 33, 189, 221, 36, 192, 198, 65, 49, 0, 13, 32, 61, 216, 1, 0, 104, 220, 35, 192, 198, 193, 105, 220, 35, 192, 198, 225, 7, 0, 164, 223, 12, 0, 164, 223, 34, 192, 198, 129, 168, 223, 34, 192, 198, 161, 235, 223, 34, 192, 198, 193, 237, 223, 34, 192, 198, 225, 62, 223, 34, 192, 198, 1, 115, 223, 34, 192, 198, 33, 124, 223, 34, 192, 198, 65, 147, 223, 34, 192, 198, 97, 5, 0, 44, 221, 9, 0, 44, 221, 35, 192, 198, 97, 128, 222, 35, 192, 198, 129, 146, 222, 35, 192, 198, 161, 187, 220, 35, 192, 198, 1, 188, 220, 35, 192, 198, 33, 39, 221, 35, 192, 198, 65, 149, 38, 33, 192, 198, 161, 150, 38, 33, 192, 198, 193, 8, 39, 33, 192, 198, 225, 48, 0, 13, 32, 5, 0, 60, 216, 75, 0, 60, 216, 25, 0, 61, 216, 50, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 25, 192, 198, 1, 188, 221, 25, 192, 198, 33, 189, 221, 25, 192, 198, 65, 49, 0, 13, 32, 61, 216, 1, 0, 104, 220, 24, 192, 198, 193, 105, 220, 24, 192, 198, 225, 7, 0, 164, 223, 12, 0, 164, 223, 23, 192, 198, 129, 168, 223, 23, 192, 198, 161, 235, 223, 23, 192, 198, 193, 237, 223, 23, 192, 198, 225, 62, 223, 23, 192, 198, 1, 115, 223, 23, 192, 198, 33, 124, 223, 23, 192, 198, 65, 147, 223, 23, 192, 198, 97, 5, 0, 44, 221, 9, 0, 44, 221, 24, 192, 198, 97, 128, 222, 24, 192, 198, 129, 146, 222, 24, 192, 198, 161, 187, 220, 24, 192, 198, 1, 188, 220, 24, 192, 198, 33, 39, 221, 24, 192, 198, 65, 149, 38, 22, 192, 198, 161, 150, 38, 22, 192, 198, 193, 8, 39, 22, 192, 198, 225, 48, 0, 13, 32, 5, 0, 60, 216, 75, 0, 60, 216, 25, 0, 61, 216, 50, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 27, 192, 198, 193, 188, 221, 27, 192, 198, 225, 189, 221, 28, 192, 198, 1, 49, 0, 13, 32, 61, 216, 1, 0, 104, 220, 27, 192, 198, 129, 105, 220, 27, 192, 198, 161, 7, 0, 164, 223, 12, 0, 164, 223, 26, 192, 198, 65, 168, 223, 26, 192, 198, 97, 235, 223, 26, 192, 198, 129, 237, 223, 26, 192, 198, 161, 62, 223, 25, 192, 198, 193, 115, 223, 25, 192, 198, 225, 124, 223, 26, 192, 198, 1, 147, 223, 26, 192, 198, 33, 5, 0, 44, 221, 9, 0, 44, 221, 27, 192, 198, 33, 128, 222, 27, 192, 198, 65, 146, 222, 27, 192, 198, 97, 187, 220, 26, 192, 198, 193, 188, 220, 26, 192, 198, 225, 39, 221, 27, 192, 198, 1, 149, 38, 25, 192, 198, 97, 150, 38, 25, 192, 198, 129, 8, 39, 25, 192, 198, 161, 48, 0, 13, 32, 5, 0, 60, 216, 75, 0, 60, 216, 25, 0, 61, 216, 50, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 30, 192, 198, 129, 188, 221, 30, 192, 198, 161, 189, 221, 30, 192, 198, 193, 49, 0, 13, 32, 61, 216, 1, 0, 104, 220, 30, 192, 198, 65, 105, 220, 30, 192, 198, 97, 7, 0, 164, 223, 12, 0, 164, 223, 29, 192, 198, 1, 168, 223, 29, 192, 198, 33, 235, 223, 29, 192, 198, 65, 237, 223, 29, 192, 198, 97, 62, 223, 28, 192, 198, 129, 115, 223, 28, 192, 198, 161, 124, 223, 28, 192, 198, 193, 147, 223, 28, 192, 198, 225, 5, 0, 44, 221, 9, 0, 44, 221, 29, 192, 198, 225, 128, 222, 30, 192, 198, 1, 146, 222, 30, 192, 198, 33, 187, 220, 29, 192, 198, 129, 188, 220, 29, 192, 198, 161, 39, 221, 29, 192, 198, 193, 149, 38, 28, 192, 198, 33, 150, 38, 28, 192, 198, 65, 8, 39, 28, 192, 198, 97, 48, 0, 13, 32, 5, 0, 60, 216, 75, 0, 60, 216, 25, 0, 61, 216, 50, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 33, 192, 198, 65, 188, 221, 33, 192, 198, 97, 189, 221, 33, 192, 198, 129, 49, 0, 13, 32, 61, 216, 1, 0, 104, 220, 33, 192, 198, 1, 105, 220, 33, 192, 198, 33, 7, 0, 164, 223, 12, 0, 164, 223, 31, 192, 198, 193, 168, 223, 31, 192, 198, 225, 235, 223, 32, 192, 198, 1, 237, 223, 32, 192, 198, 33, 62, 223, 31, 192, 198, 65, 115, 223, 31, 192, 198, 97, 124, 223, 31, 192, 198, 129, 147, 223, 31, 192, 198, 161, 5, 0, 44, 221, 9, 0, 44, 221, 32, 192, 198, 161, 128, 222, 32, 192, 198, 193, 146, 222, 32, 192, 198, 225, 187, 220, 32, 192, 198, 65, 188, 220, 32, 192, 198, 97, 39, 221, 32, 192, 198, 129, 149, 38, 30, 192, 198, 225, 150, 38, 31, 192, 198, 1, 8, 39, 31, 192, 198, 33, 6, 0, 100, 39, 163, 0, 100, 39, 15, 0, 60, 216, 58, 0, 61, 216, 83, 0, 62, 216, 2, 0, 175, 221, 126, 205, 193, 117, 188, 221, 126, 205, 193, 120, 189, 221, 126, 205, 193, 123, 1, 0, 13, 32, 22, 0, 15, 254, 49, 0, 13, 32, 61, 216, 2, 0, 104, 220, 126, 205, 193, 196, 105, 220, 126, 205, 193, 198, 139, 220, 49, 0, 13, 32, 61, 216, 1, 0, 104, 220, 126, 205, 193, 192, 105, 220, 126, 205, 193, 194, 48, 0, 61, 216, 2, 0, 104, 220, 126, 205, 193, 196, 105, 220, 126, 205, 193, 198, 139, 220, 49, 0, 13, 32, 61, 216, 1, 0, 104, 220, 126, 205, 193, 192, 105, 220, 126, 205, 193, 194, 7, 0, 164, 223, 12, 0, 164, 223, 126, 205, 193, 25, 168, 223, 126, 205, 193, 28, 235, 223, 125, 205, 193, 252, 237, 223, 126, 205, 193, 13, 62, 223, 126, 205, 193, 4, 115, 223, 126, 205, 193, 7, 124, 223, 126, 205, 193, 66, 147, 223, 125, 205, 193, 249, 8, 0, 188, 220, 15, 0, 188, 220, 126, 205, 193, 16, 39, 221, 126, 205, 193, 10, 44, 221, 126, 205, 193, 19, 128, 222, 126, 205, 193, 34, 146, 222, 126, 205, 193, 37, 102, 220, 7, 0, 103, 220, 13, 0, 105, 220, 23, 0, 187, 220, 126, 205, 193, 22, 242, 127, 126, 13, 193, 220, 13, 32, 61, 216, 102, 220, 126, 205, 193, 221, 241, 127, 126, 13, 193, 222, 13, 32, 61, 216, 1, 0, 102, 220, 126, 205, 193, 223, 103, 220, 126, 205, 193, 224, 49, 0, 13, 32, 61, 216, 1, 0, 102, 220, 13, 0, 103, 220, 241, 127, 126, 13, 193, 211, 13, 32, 61, 216, 1, 0, 102, 220, 126, 205, 193, 212, 103, 220, 126, 205, 193, 214, 242, 127, 126, 13, 193, 210, 13, 32, 61, 216, 102, 220, 126, 205, 193, 213, 149, 38, 125, 205, 193, 246, 150, 38, 125, 205, 193, 255, 8, 39, 126, 205, 193, 31, 126, 13, 193, 189, 48, 0, 60, 216, 4, 0, 251, 223, 36, 192, 198, 98, 252, 223, 36, 192, 198, 162, 253, 223, 36, 192, 198, 226, 254, 223, 37, 192, 198, 34, 255, 223, 37, 192, 198, 98, 126, 13, 193, 190, 48, 0, 60, 216, 4, 0, 251, 223, 37, 192, 198, 162, 252, 223, 37, 192, 198, 226, 253, 223, 38, 192, 198, 34, 254, 223, 38, 192, 198, 98, 255, 223, 38, 192, 198, 162, 126, 13, 193, 188, 48, 0, 60, 216, 4, 0, 251, 223, 38, 192, 198, 226, 252, 223, 39, 192, 198, 34, 253, 223, 39, 192, 198, 98, 254, 223, 39, 192, 198, 162, 255, 223, 39, 192, 198, 226, 126, 13, 193, 38, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 41, 192, 198, 33, 66, 38, 41, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 40, 192, 198, 33, 66, 38, 40, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 40, 192, 198, 97, 66, 38, 40, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 40, 192, 198, 161, 66, 38, 40, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 40, 192, 198, 225, 66, 38, 41, 192, 198, 1, 1, 0, 64, 38, 126, 205, 193, 40, 66, 38, 126, 205, 193, 39, 126, 13, 193, 130, 48, 0, 13, 32, 1, 0, 64, 38, 126, 205, 193, 132, 66, 38, 126, 205, 193, 131, 126, 13, 193, 61, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 42, 192, 198, 97, 66, 38, 42, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 41, 192, 198, 97, 66, 38, 41, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 41, 192, 198, 161, 66, 38, 41, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 41, 192, 198, 225, 66, 38, 42, 192, 198, 1, 48, 0, 13, 32, 1, 0, 64, 38, 42, 192, 198, 33, 66, 38, 42, 192, 198, 65, 1, 0, 64, 38, 126, 205, 193, 63, 66, 38, 126, 205, 193, 62, 42, 0, 198, 161, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 44, 192, 198, 1, 66, 38, 44, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 43, 192, 198, 1, 66, 38, 43, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 43, 192, 198, 65, 66, 38, 43, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 43, 192, 198, 129, 66, 38, 43, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 43, 192, 198, 193, 66, 38, 43, 192, 198, 225, 1, 0, 64, 38, 42, 192, 198, 193, 66, 38, 42, 192, 198, 225, 126, 13, 193, 53, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 45, 192, 198, 65, 66, 38, 45, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 44, 192, 198, 65, 66, 38, 44, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 44, 192, 198, 129, 66, 38, 44, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 44, 192, 198, 193, 66, 38, 44, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 45, 192, 198, 1, 66, 38, 45, 192, 198, 33, 1, 0, 64, 38, 126, 205, 193, 55, 66, 38, 126, 205, 193, 54, 126, 13, 193, 48, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 46, 192, 198, 129, 66, 38, 46, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 45, 192, 198, 129, 66, 38, 45, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 45, 192, 198, 193, 66, 38, 45, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 46, 192, 198, 1, 66, 38, 46, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 46, 192, 198, 65, 66, 38, 46, 192, 198, 97, 1, 0, 64, 38, 126, 205, 193, 50, 66, 38, 126, 205, 193, 49, 125, 13, 193, 226, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 47, 192, 198, 193, 66, 38, 47, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 46, 192, 198, 193, 66, 38, 46, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 47, 192, 198, 1, 66, 38, 47, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 47, 192, 198, 65, 66, 38, 47, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 47, 192, 198, 129, 66, 38, 47, 192, 198, 161, 1, 0, 64, 38, 125, 205, 193, 228, 66, 38, 125, 205, 193, 227, 126, 13, 193, 44, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 49, 192, 198, 1, 66, 38, 49, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 48, 192, 198, 1, 66, 38, 48, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 48, 192, 198, 65, 66, 38, 48, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 48, 192, 198, 129, 66, 38, 48, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 48, 192, 198, 193, 66, 38, 48, 192, 198, 225, 1, 0, 64, 38, 126, 205, 193, 46, 66, 38, 126, 205, 193, 45, 126, 13, 193, 100, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 50, 192, 198, 65, 66, 38, 50, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 49, 192, 198, 65, 66, 38, 49, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 49, 192, 198, 129, 66, 38, 49, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 49, 192, 198, 193, 66, 38, 49, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 50, 192, 198, 1, 66, 38, 50, 192, 198, 33, 1, 0, 64, 38, 126, 205, 193, 102, 66, 38, 126, 205, 193, 101, 126, 13, 193, 103, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 51, 192, 198, 129, 66, 38, 51, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 50, 192, 198, 129, 66, 38, 50, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 50, 192, 198, 193, 66, 38, 50, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 51, 192, 198, 1, 66, 38, 51, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 51, 192, 198, 65, 66, 38, 51, 192, 198, 97, 1, 0, 64, 38, 126, 205, 193, 105, 66, 38, 126, 205, 193, 104, 126, 13, 193, 41, 2, 0, 13, 32, 5, 0, 60, 216, 10, 0, 15, 254, 48, 0, 13, 32, 1, 0, 64, 38, 126, 205, 193, 43, 66, 38, 126, 205, 193, 42, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 52, 192, 198, 193, 66, 38, 52, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 51, 192, 198, 193, 66, 38, 51, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 52, 192, 198, 1, 66, 38, 52, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 52, 192, 198, 65, 66, 38, 52, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 52, 192, 198, 129, 66, 38, 52, 192, 198, 161, 125, 13, 193, 220, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 54, 192, 198, 1, 66, 38, 54, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 53, 192, 198, 1, 66, 38, 53, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 53, 192, 198, 65, 66, 38, 53, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 53, 192, 198, 129, 66, 38, 53, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 53, 192, 198, 193, 66, 38, 53, 192, 198, 225, 1, 0, 64, 38, 125, 205, 193, 222, 66, 38, 125, 205, 193, 221, 125, 13, 193, 223, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 55, 192, 198, 65, 66, 38, 55, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 54, 192, 198, 65, 66, 38, 54, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 54, 192, 198, 129, 66, 38, 54, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 54, 192, 198, 193, 66, 38, 54, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 55, 192, 198, 1, 66, 38, 55, 192, 198, 33, 1, 0, 64, 38, 125, 205, 193, 225, 66, 38, 125, 205, 193, 224, 125, 13, 193, 235, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 56, 192, 198, 129, 66, 38, 56, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 55, 192, 198, 129, 66, 38, 55, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 55, 192, 198, 193, 66, 38, 55, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 56, 192, 198, 1, 66, 38, 56, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 56, 192, 198, 65, 66, 38, 56, 192, 198, 97, 1, 0, 64, 38, 125, 205, 193, 237, 66, 38, 125, 205, 193, 236, 125, 13, 193, 229, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 57, 192, 198, 193, 66, 38, 57, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 56, 192, 198, 193, 66, 38, 56, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 57, 192, 198, 1, 66, 38, 57, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 57, 192, 198, 65, 66, 38, 57, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 57, 192, 198, 129, 66, 38, 57, 192, 198, 161, 1, 0, 64, 38, 125, 205, 193, 231, 66, 38, 125, 205, 193, 230, 125, 13, 193, 214, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 59, 192, 198, 1, 66, 38, 59, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 58, 192, 198, 1, 66, 38, 58, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 58, 192, 198, 65, 66, 38, 58, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 58, 192, 198, 129, 66, 38, 58, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 58, 192, 198, 193, 66, 38, 58, 192, 198, 225, 1, 0, 64, 38, 125, 205, 193, 216, 66, 38, 125, 205, 193, 215, 125, 13, 193, 217, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 60, 192, 198, 65, 66, 38, 60, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 59, 192, 198, 65, 66, 38, 59, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 59, 192, 198, 129, 66, 38, 59, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 59, 192, 198, 193, 66, 38, 59, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 60, 192, 198, 1, 66, 38, 60, 192, 198, 33, 1, 0, 64, 38, 125, 205, 193, 219, 66, 38, 125, 205, 193, 218, 126, 13, 193, 149, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 61, 192, 198, 129, 66, 38, 61, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 60, 192, 198, 129, 66, 38, 60, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 60, 192, 198, 193, 66, 38, 60, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 61, 192, 198, 1, 66, 38, 61, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 61, 192, 198, 65, 66, 38, 61, 192, 198, 97, 1, 0, 64, 38, 126, 205, 193, 151, 66, 38, 126, 205, 193, 150, 126, 13, 193, 161, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 62, 192, 198, 193, 66, 38, 62, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 61, 192, 198, 193, 66, 38, 61, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 62, 192, 198, 1, 66, 38, 62, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 62, 192, 198, 65, 66, 38, 62, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 62, 192, 198, 129, 66, 38, 62, 192, 198, 161, 1, 0, 64, 38, 126, 205, 193, 163, 66, 38, 126, 205, 193, 162, 126, 13, 193, 164, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 64, 192, 198, 1, 66, 38, 64, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 63, 192, 198, 1, 66, 38, 63, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 63, 192, 198, 65, 66, 38, 63, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 63, 192, 198, 129, 66, 38, 63, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 63, 192, 198, 193, 66, 38, 63, 192, 198, 225, 1, 0, 64, 38, 126, 205, 193, 166, 66, 38, 126, 205, 193, 165, 126, 13, 193, 106, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 65, 192, 198, 65, 66, 38, 65, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 64, 192, 198, 65, 66, 38, 64, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 64, 192, 198, 129, 66, 38, 64, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 64, 192, 198, 193, 66, 38, 64, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 65, 192, 198, 1, 66, 38, 65, 192, 198, 33, 1, 0, 64, 38, 126, 205, 193, 108, 66, 38, 126, 205, 193, 107, 125, 13, 193, 238, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 66, 192, 198, 129, 66, 38, 66, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 65, 192, 198, 129, 66, 38, 65, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 65, 192, 198, 193, 66, 38, 65, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 66, 192, 198, 1, 66, 38, 66, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 66, 192, 198, 65, 66, 38, 66, 192, 198, 97, 1, 0, 64, 38, 125, 205, 193, 240, 66, 38, 125, 205, 193, 239, 126, 13, 193, 58, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 67, 192, 198, 193, 66, 38, 67, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 66, 192, 198, 193, 66, 38, 66, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 67, 192, 198, 1, 66, 38, 67, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 67, 192, 198, 65, 66, 38, 67, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 67, 192, 198, 129, 66, 38, 67, 192, 198, 161, 1, 0, 64, 38, 126, 205, 193, 60, 66, 38, 126, 205, 193, 59, 125, 13, 193, 241, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 69, 192, 198, 1, 66, 38, 69, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 68, 192, 198, 1, 66, 38, 68, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 68, 192, 198, 65, 66, 38, 68, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 68, 192, 198, 129, 66, 38, 68, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 68, 192, 198, 193, 66, 38, 68, 192, 198, 225, 1, 0, 64, 38, 125, 205, 193, 243, 66, 38, 125, 205, 193, 242, 126, 13, 193, 167, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 70, 192, 198, 65, 66, 38, 70, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 69, 192, 198, 65, 66, 38, 69, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 69, 192, 198, 129, 66, 38, 69, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 69, 192, 198, 193, 66, 38, 69, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 70, 192, 198, 1, 66, 38, 70, 192, 198, 33, 1, 0, 64, 38, 126, 205, 193, 169, 66, 38, 126, 205, 193, 168, 126, 13, 193, 179, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 71, 192, 198, 129, 66, 38, 71, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 70, 192, 198, 129, 66, 38, 70, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 70, 192, 198, 193, 66, 38, 70, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 71, 192, 198, 1, 66, 38, 71, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 71, 192, 198, 65, 66, 38, 71, 192, 198, 97, 1, 0, 64, 38, 126, 205, 193, 181, 66, 38, 126, 205, 193, 180, 126, 13, 193, 170, 48, 0, 13, 32, 1, 0, 64, 38, 126, 205, 193, 172, 66, 38, 126, 205, 193, 171, 126, 13, 193, 173, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 72, 192, 198, 193, 66, 38, 72, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 71, 192, 198, 193, 66, 38, 71, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 72, 192, 198, 1, 66, 38, 72, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 72, 192, 198, 65, 66, 38, 72, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 72, 192, 198, 129, 66, 38, 72, 192, 198, 161, 1, 0, 64, 38, 126, 205, 193, 175, 66, 38, 126, 205, 193, 174, 126, 13, 193, 176, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 74, 192, 198, 1, 66, 38, 74, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 73, 192, 198, 1, 66, 38, 73, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 73, 192, 198, 65, 66, 38, 73, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 73, 192, 198, 129, 66, 38, 73, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 73, 192, 198, 193, 66, 38, 73, 192, 198, 225, 1, 0, 64, 38, 126, 205, 193, 178, 66, 38, 126, 205, 193, 177, 126, 13, 193, 73, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 75, 192, 198, 65, 66, 38, 75, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 74, 192, 198, 65, 66, 38, 74, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 74, 192, 198, 129, 66, 38, 74, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 74, 192, 198, 193, 66, 38, 74, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 75, 192, 198, 1, 66, 38, 75, 192, 198, 33, 1, 0, 64, 38, 126, 205, 193, 75, 66, 38, 126, 205, 193, 74, 126, 13, 193, 76, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 76, 192, 198, 129, 66, 38, 76, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 75, 192, 198, 129, 66, 38, 75, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 75, 192, 198, 193, 66, 38, 75, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 76, 192, 198, 1, 66, 38, 76, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 76, 192, 198, 65, 66, 38, 76, 192, 198, 97, 1, 0, 64, 38, 126, 205, 193, 78, 66, 38, 126, 205, 193, 77, 126, 13, 193, 109, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 77, 192, 198, 193, 66, 38, 77, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 76, 192, 198, 193, 66, 38, 76, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 77, 192, 198, 1, 66, 38, 77, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 77, 192, 198, 65, 66, 38, 77, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 77, 192, 198, 129, 66, 38, 77, 192, 198, 161, 1, 0, 64, 38, 126, 205, 193, 111, 66, 38, 126, 205, 193, 110, 126, 13, 193, 112, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 79, 192, 198, 1, 66, 38, 79, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 78, 192, 198, 1, 66, 38, 78, 192, 198, 33, 48, 0, 13, 32, 1, 0, 64, 38, 78, 192, 198, 65, 66, 38, 78, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 78, 192, 198, 129, 66, 38, 78, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 78, 192, 198, 193, 66, 38, 78, 192, 198, 225, 1, 0, 64, 38, 126, 205, 193, 114, 66, 38, 126, 205, 193, 113, 125, 13, 193, 232, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 80, 192, 198, 65, 66, 38, 80, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 79, 192, 198, 65, 66, 38, 79, 192, 198, 97, 48, 0, 13, 32, 1, 0, 64, 38, 79, 192, 198, 129, 66, 38, 79, 192, 198, 161, 48, 0, 13, 32, 1, 0, 64, 38, 79, 192, 198, 193, 66, 38, 79, 192, 198, 225, 48, 0, 13, 32, 1, 0, 64, 38, 80, 192, 198, 1, 66, 38, 80, 192, 198, 33, 1, 0, 64, 38, 125, 205, 193, 234, 66, 38, 125, 205, 193, 233, 125, 13, 193, 204, 1, 0, 13, 32, 195, 1, 60, 216, 4, 0, 251, 223, 95, 0, 252, 223, 181, 0, 253, 223, 11, 1, 254, 223, 97, 1, 255, 223, 48, 0, 13, 32, 5, 0, 60, 216, 74, 0, 60, 216, 21, 0, 61, 216, 49, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 93, 192, 198, 225, 188, 221, 94, 192, 198, 1, 189, 221, 94, 192, 198, 33, 50, 0, 13, 32, 62, 216, 209, 221, 93, 192, 198, 193, 8, 0, 147, 223, 15, 0, 147, 223, 92, 192, 198, 97, 164, 223, 92, 192, 198, 129, 168, 223, 92, 192, 198, 161, 235, 223, 92, 192, 198, 193, 237, 223, 92, 192, 198, 225, 62, 223, 91, 192, 198, 225, 115, 223, 92, 192, 198, 1, 124, 223, 92, 192, 198, 33, 132, 223, 92, 192, 198, 65, 5, 0, 44, 221, 9, 0, 44, 221, 93, 192, 198, 97, 128, 222, 93, 192, 198, 129, 146, 222, 93, 192, 198, 161, 187, 220, 93, 192, 198, 1, 188, 220, 93, 192, 198, 33, 39, 221, 93, 192, 198, 65, 149, 38, 91, 192, 198, 129, 150, 38, 91, 192, 198, 161, 8, 39, 91, 192, 198, 193, 48, 0, 13, 32, 5, 0, 60, 216, 74, 0, 60, 216, 21, 0, 61, 216, 49, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 82, 192, 198, 225, 188, 221, 83, 192, 198, 1, 189, 221, 83, 192, 198, 33, 50, 0, 13, 32, 62, 216, 209, 221, 82, 192, 198, 193, 8, 0, 147, 223, 15, 0, 147, 223, 81, 192, 198, 97, 164, 223, 81, 192, 198, 129, 168, 223, 81, 192, 198, 161, 235, 223, 81, 192, 198, 193, 237, 223, 81, 192, 198, 225, 62, 223, 80, 192, 198, 225, 115, 223, 81, 192, 198, 1, 124, 223, 81, 192, 198, 33, 132, 223, 81, 192, 198, 65, 5, 0, 44, 221, 9, 0, 44, 221, 82, 192, 198, 97, 128, 222, 82, 192, 198, 129, 146, 222, 82, 192, 198, 161, 187, 220, 82, 192, 198, 1, 188, 220, 82, 192, 198, 33, 39, 221, 82, 192, 198, 65, 149, 38, 80, 192, 198, 129, 150, 38, 80, 192, 198, 161, 8, 39, 80, 192, 198, 193, 48, 0, 13, 32, 5, 0, 60, 216, 74, 0, 60, 216, 21, 0, 61, 216, 49, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 85, 192, 198, 161, 188, 221, 85, 192, 198, 193, 189, 221, 85, 192, 198, 225, 50, 0, 13, 32, 62, 216, 209, 221, 85, 192, 198, 129, 8, 0, 147, 223, 15, 0, 147, 223, 84, 192, 198, 33, 164, 223, 84, 192, 198, 65, 168, 223, 84, 192, 198, 97, 235, 223, 84, 192, 198, 129, 237, 223, 84, 192, 198, 161, 62, 223, 83, 192, 198, 161, 115, 223, 83, 192, 198, 193, 124, 223, 83, 192, 198, 225, 132, 223, 84, 192, 198, 1, 5, 0, 44, 221, 9, 0, 44, 221, 85, 192, 198, 33, 128, 222, 85, 192, 198, 65, 146, 222, 85, 192, 198, 97, 187, 220, 84, 192, 198, 193, 188, 220, 84, 192, 198, 225, 39, 221, 85, 192, 198, 1, 149, 38, 83, 192, 198, 65, 150, 38, 83, 192, 198, 97, 8, 39, 83, 192, 198, 129, 48, 0, 13, 32, 5, 0, 60, 216, 74, 0, 60, 216, 21, 0, 61, 216, 49, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 88, 192, 198, 97, 188, 221, 88, 192, 198, 129, 189, 221, 88, 192, 198, 161, 50, 0, 13, 32, 62, 216, 209, 221, 88, 192, 198, 65, 8, 0, 147, 223, 15, 0, 147, 223, 86, 192, 198, 225, 164, 223, 87, 192, 198, 1, 168, 223, 87, 192, 198, 33, 235, 223, 87, 192, 198, 65, 237, 223, 87, 192, 198, 97, 62, 223, 86, 192, 198, 97, 115, 223, 86, 192, 198, 129, 124, 223, 86, 192, 198, 161, 132, 223, 86, 192, 198, 193, 5, 0, 44, 221, 9, 0, 44, 221, 87, 192, 198, 225, 128, 222, 88, 192, 198, 1, 146, 222, 88, 192, 198, 33, 187, 220, 87, 192, 198, 129, 188, 220, 87, 192, 198, 161, 39, 221, 87, 192, 198, 193, 149, 38, 86, 192, 198, 1, 150, 38, 86, 192, 198, 33, 8, 39, 86, 192, 198, 65, 48, 0, 13, 32, 5, 0, 60, 216, 74, 0, 60, 216, 21, 0, 61, 216, 49, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 91, 192, 198, 33, 188, 221, 91, 192, 198, 65, 189, 221, 91, 192, 198, 97, 50, 0, 13, 32, 62, 216, 209, 221, 91, 192, 198, 1, 8, 0, 147, 223, 15, 0, 147, 223, 89, 192, 198, 161, 164, 223, 89, 192, 198, 193, 168, 223, 89, 192, 198, 225, 235, 223, 90, 192, 198, 1, 237, 223, 90, 192, 198, 33, 62, 223, 89, 192, 198, 33, 115, 223, 89, 192, 198, 65, 124, 223, 89, 192, 198, 97, 132, 223, 89, 192, 198, 129, 5, 0, 44, 221, 9, 0, 44, 221, 90, 192, 198, 161, 128, 222, 90, 192, 198, 193, 146, 222, 90, 192, 198, 225, 187, 220, 90, 192, 198, 65, 188, 220, 90, 192, 198, 97, 39, 221, 90, 192, 198, 129, 149, 38, 88, 192, 198, 193, 150, 38, 88, 192, 198, 225, 8, 39, 89, 192, 198, 1, 5, 0, 60, 216, 74, 0, 60, 216, 21, 0, 61, 216, 49, 0, 62, 216, 3, 0, 29, 221, 9, 0, 175, 221, 126, 205, 193, 115, 188, 221, 126, 205, 193, 118, 189, 221, 126, 205, 193, 121, 50, 0, 13, 32, 62, 216, 209, 221, 126, 205, 193, 187, 8, 0, 147, 223, 15, 0, 147, 223, 125, 205, 193, 247, 164, 223, 126, 205, 193, 23, 168, 223, 126, 205, 193, 26, 235, 223, 125, 205, 193, 250, 237, 223, 126, 205, 193, 11, 62, 223, 126, 205, 193, 2, 115, 223, 126, 205, 193, 5, 124, 223, 126, 205, 193, 68, 132, 223, 126, 205, 193, 72, 5, 0, 44, 221, 9, 0, 44, 221, 126, 205, 193, 17, 128, 222, 126, 205, 193, 32, 146, 222, 126, 205, 193, 35, 187, 220, 126, 205, 193, 20, 188, 220, 126, 205, 193, 14, 39, 221, 126, 205, 193, 8, 149, 38, 125, 205, 193, 244, 150, 38, 125, 205, 193, 253, 8, 39, 126, 205, 193, 29, 126, 13, 193, 133, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 95, 192, 198, 97, 66, 38, 95, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 94, 192, 198, 97, 66, 38, 94, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 94, 192, 198, 161, 66, 38, 94, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 94, 192, 198, 225, 66, 38, 95, 192, 198, 1, 48, 0, 13, 32, 1, 0, 64, 38, 95, 192, 198, 33, 66, 38, 95, 192, 198, 65, 1, 0, 64, 38, 126, 205, 193, 135, 66, 38, 126, 205, 193, 134, 126, 13, 193, 136, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 96, 192, 198, 161, 66, 38, 96, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 95, 192, 198, 161, 66, 38, 95, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 95, 192, 198, 225, 66, 38, 96, 192, 198, 1, 48, 0, 13, 32, 1, 0, 64, 38, 96, 192, 198, 33, 66, 38, 96, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 96, 192, 198, 97, 66, 38, 96, 192, 198, 129, 1, 0, 64, 38, 126, 205, 193, 138, 66, 38, 126, 205, 193, 137, 126, 13, 193, 182, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 97, 192, 198, 225, 66, 38, 98, 192, 198, 1, 48, 0, 13, 32, 1, 0, 64, 38, 96, 192, 198, 225, 66, 38, 97, 192, 198, 1, 48, 0, 13, 32, 1, 0, 64, 38, 97, 192, 198, 33, 66, 38, 97, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 97, 192, 198, 97, 66, 38, 97, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 97, 192, 198, 161, 66, 38, 97, 192, 198, 193, 1, 0, 64, 38, 126, 205, 193, 184, 66, 38, 126, 205, 193, 183, 126, 13, 193, 79, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 99, 192, 198, 33, 66, 38, 99, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 98, 192, 198, 33, 66, 38, 98, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 98, 192, 198, 97, 66, 38, 98, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 98, 192, 198, 161, 66, 38, 98, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 98, 192, 198, 225, 66, 38, 99, 192, 198, 1, 1, 0, 64, 38, 126, 205, 193, 81, 66, 38, 126, 205, 193, 80, 126, 13, 193, 82, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 100, 192, 198, 97, 66, 38, 100, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 99, 192, 198, 97, 66, 38, 99, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 99, 192, 198, 161, 66, 38, 99, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 99, 192, 198, 225, 66, 38, 100, 192, 198, 1, 48, 0, 13, 32, 1, 0, 64, 38, 100, 192, 198, 33, 66, 38, 100, 192, 198, 65, 1, 0, 64, 38, 126, 205, 193, 84, 66, 38, 126, 205, 193, 83, 126, 13, 193, 85, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 101, 192, 198, 161, 66, 38, 101, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 100, 192, 198, 161, 66, 38, 100, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 100, 192, 198, 225, 66, 38, 101, 192, 198, 1, 48, 0, 13, 32, 1, 0, 64, 38, 101, 192, 198, 33, 66, 38, 101, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 101, 192, 198, 97, 66, 38, 101, 192, 198, 129, 1, 0, 64, 38, 126, 205, 193, 87, 66, 38, 126, 205, 193, 86, 126, 13, 193, 88, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 102, 192, 198, 225, 66, 38, 103, 192, 198, 1, 48, 0, 13, 32, 1, 0, 64, 38, 101, 192, 198, 225, 66, 38, 102, 192, 198, 1, 48, 0, 13, 32, 1, 0, 64, 38, 102, 192, 198, 33, 66, 38, 102, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 102, 192, 198, 97, 66, 38, 102, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 102, 192, 198, 161, 66, 38, 102, 192, 198, 193, 1, 0, 64, 38, 126, 205, 193, 90, 66, 38, 126, 205, 193, 89, 126, 13, 193, 91, 1, 0, 13, 32, 56, 0, 60, 216, 4, 0, 251, 223, 16, 0, 252, 223, 23, 0, 253, 223, 30, 0, 254, 223, 37, 0, 255, 223, 48, 0, 13, 32, 1, 0, 64, 38, 104, 192, 198, 33, 66, 38, 104, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 103, 192, 198, 33, 66, 38, 103, 192, 198, 65, 48, 0, 13, 32, 1, 0, 64, 38, 103, 192, 198, 97, 66, 38, 103, 192, 198, 129, 48, 0, 13, 32, 1, 0, 64, 38, 103, 192, 198, 161, 66, 38, 103, 192, 198, 193, 48, 0, 13, 32, 1, 0, 64, 38, 103, 192, 198, 225, 66, 38, 104, 192, 198, 1, 1, 0, 64, 38, 126, 205, 193, 93, 66, 38, 126, 205, 193, 92, 126, 13, 193, 94, 48, 0, 13, 32, 1, 0, 64, 38, 126, 205, 193, 96, 66, 38, 126, 205, 193, 95, 126, 13, 193, 97, 48, 0, 13, 32, 1, 0, 64, 38, 126, 205, 193, 99, 66, 38, 126, 205, 193, 98, 156, 128, 28, 0, 13, 32, 14, 32, 227, 32, 228, 32, 32, 38, 33, 38, 64, 38, 65, 38, 66, 38, 67, 38, 149, 38, 151, 38, 167, 38, 168, 38, 8, 39, 9, 39, 68, 39, 69, 39, 100, 39, 101, 39, 27, 43, 28, 43, 60, 216, 63, 216, 64, 219, 65, 219, 15, 254, 16, 254, 1, 0, 8, 243, 1, 0, 9, 243, 1, 0, 62, 243, 1, 0, 63, 243, 1, 0, 115, 243, 1, 0, 116, 243, 1, 0, 124, 243, 1, 0, 125, 243, 1, 0, 132, 243, 1, 0, 133, 243, 1, 0, 147, 243, 1, 0, 148, 243, 1, 0, 164, 243, 1, 0, 165, 243, 1, 0, 168, 243, 1, 0, 169, 243, 1, 0, 235, 243, 1, 0, 236, 243, 1, 0, 237, 243, 1, 0, 238, 243, 1, 0, 251, 243, 1, 0, 0, 244, 1, 0, 102, 244, 1, 0, 106, 244, 1, 0, 139, 244, 1, 0, 140, 244, 1, 0, 187, 244, 1, 0, 189, 244, 1, 0, 39, 245, 1, 0, 40, 245, 1, 0, 44, 245, 1, 0, 45, 245, 1, 0, 232, 245, 1, 0, 233, 245, 1, 0, 128, 246, 1, 0, 129, 246, 1, 0, 146, 246, 1, 0, 147, 246, 1, 0, 29, 249, 1, 0, 30, 249, 1, 0, 175, 249, 1, 0, 176, 249, 1, 0, 186, 249, 1, 0, 187, 249, 1, 0, 188, 249, 1, 0, 190, 249, 1, 0, 209, 249, 1, 0, 210, 249, 14, 0, 98, 0, 14, 0, 100, 0, 14, 0, 101, 0, 14, 0, 102, 0, 14, 0, 103, 0, 14, 0, 104, 0, 14, 0, 108, 0, 14, 0, 109, 0, 14, 0, 110, 0, 14, 0, 111, 0, 14, 0, 115, 0, 14, 0, 117, 0, 14, 0, 119, 0, 14, 0, 120, 0, 14, 0, 127, 0, 14, 0, 128, 0, 5, 2, 64, 12, 32, 14, 200, 14, 240, 14, 22, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 24, 4, 224, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 26, 4, 96, 14, 144, 12, 96, 12, 208, 12, 160, 13, 28, 4, 30, 4, 32, 4, 34, 4, 36, 4, 38, 4, 40, 4, 42, 4, 44, 4, 46, 4, 160, 12, 152, 12, 128, 14, 136, 14, 144, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 185, 120, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 185, 152, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 168, 120, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 168, 152, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 160, 14, 112, 13, 176, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 216, 14, 232, 14, 208, 14, 240, 14, 168, 14, 128, 13, 72, 14, 192, 14, 170, 56, 56, 13, 152, 14, 0, 0, 200, 14, 64, 14, 88, 14, 104, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 154, 114, 185, 121, 153, 121, 249, 121, 153, 122, 89, 122, 120, 14, 217, 122, 185, 153, 153, 153, 249, 153, 89, 154, 153, 169, 185, 176, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 136, 121, 232, 121, 136, 122, 72, 122, 112, 14, 200, 122, 168, 153, 136, 153, 232, 153, 72, 154, 136, 169, 168, 176, 72, 170, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 250, 114, 232, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 57, 123, 40, 123, 217, 121, 200, 121, 121, 122, 104, 122, 12, 8, 14, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 250, 142, 232, 142, 26, 142, 8, 142, 249, 146, 232, 146, 25, 146, 8, 146, 185, 148, 168, 148, 153, 154, 136, 154, 57, 155, 40, 155, 217, 153, 200, 153, 57, 154, 40, 154, 121, 154, 104, 154, 25, 155, 8, 155, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 185, 174, 168, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 184, 14, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 255, 5, 0, 0, 255, 5, 184, 13, 255, 5, 152, 13, 255, 5, 168, 16, 255, 5, 168, 20, 255, 5, 168, 24, 255, 5, 168, 28, 255, 5, 168, 32, 255, 5, 168, 36, 255, 5, 168, 40, 255, 5, 168, 44, 255, 5, 168, 48, 255, 5, 168, 52, 255, 1, 3, 0, 83, 0, 74, 0, 41, 0, 55, 230, 0, 16, 95, 13, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 56, 51, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 76, 46, 0, 0, 76, 46, 0, 0, 76, 46, 0, 0, 76, 46, 0, 0, 240, 46, 0, 0, 240, 46, 0, 0, 100, 47, 0, 0, 100, 47, 0, 0, 32, 51, 0, 0, 50, 105, 114, 84, 1, 0, 156, 8, 204, 1, 64, 0, 192, 0, 59, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 95, 0, 103, 0, 111, 0, 119, 0, 127, 0, 135, 0, 141, 0, 149, 0, 157, 0, 165, 0, 173, 0, 181, 0, 189, 0, 197, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 205, 0, 213, 0, 220, 0, 228, 0, 234, 0, 242, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 249, 0, 1, 1, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 25, 1, 33, 1, 41, 1, 49, 1, 57, 1, 65, 1, 73, 1, 81, 1, 48, 0, 48, 0, 48, 0, 48, 0, 82, 1, 48, 0, 48, 0, 90, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 98, 1, 48, 0, 48, 0, 48, 0, 105, 1, 109, 1, 114, 1, 48, 0, 122, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 130, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 138, 1, 146, 1, 154, 1, 162, 1, 168, 1, 48, 0, 171, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 187, 1, 17, 1, 48, 0, 48, 0, 9, 1, 195, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 9, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 124, 1, 188, 1, 252, 1, 52, 2, 116, 2, 180, 2, 244, 2, 192, 0, 192, 0, 192, 0, 192, 0, 52, 3, 112, 3, 168, 3, 192, 0, 192, 0, 192, 0, 228, 3, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 91, 8, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 179, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 255, 255, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 3, 0, 202, 1, 4, 0, 202, 34, 4, 0, 202, 67, 4, 0, 202, 100, 4, 0, 202, 133, 4, 0, 202, 166, 4, 0, 202, 199, 4, 0, 202, 232, 4, 0, 202, 9, 5, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 3, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 3, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 3, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 0, 58, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 77, 0, 62, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 0, 68, 134, 82, 0, 68, 5, 82, 0, 68, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 80, 5, 71, 0, 80, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 71, 0, 44, 134, 76, 0, 44, 134, 77, 0, 44, 5, 77, 0, 44, 0, 0, 0, 0, 0, 0, 0, 0, 134, 73, 0, 70, 134, 73, 0, 46, 5, 73, 0, 46, 134, 74, 0, 48, 134, 75, 0, 48, 134, 77, 0, 48, 5, 77, 0, 48, 5, 81, 0, 48, 134, 79, 0, 50, 134, 81, 0, 50, 134, 83, 0, 50, 134, 73, 0, 52, 5, 73, 0, 52, 134, 74, 0, 54, 134, 78, 0, 54, 197, 34, 0, 0, 134, 77, 0, 58, 134, 74, 0, 58, 134, 72, 0, 62, 5, 72, 0, 62, 5, 74, 0, 64, 5, 84, 0, 64, 134, 77, 0, 66, 134, 75, 0, 68, 5, 76, 0, 68, 134, 81, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 134, 80, 0, 54, 5, 80, 0, 54, 134, 75, 0, 72, 5, 75, 0, 72, 135, 70, 0, 76, 0, 0, 0, 0, 0, 0, 0, 0, 134, 77, 0, 78, 5, 79, 0, 78, 5, 74, 0, 80, 134, 75, 0, 80, 5, 75, 0, 80, 134, 76, 0, 80, 192, 0, 0, 0, 192, 0, 0, 0, 134, 76, 0, 82, 134, 74, 0, 84, 134, 72, 0, 90, 5, 72, 0, 90, 134, 71, 0, 92, 5, 71, 0, 92, 201, 6, 0, 0, 134, 86, 0, 92, 5, 86, 0, 92, 5, 88, 0, 92, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 192, 0, 0, 0, 5, 73, 0, 86, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 79, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 72, 0, 54, 5, 72, 0, 54, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 84, 0, 92, 5, 84, 0, 92, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 0, 0, 134, 73, 0, 86, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 75, 0, 90, 5, 75, 0, 90, 192, 0, 0, 0, 192, 0, 0, 0, 134, 76, 0, 68, 5, 78, 0, 48, 134, 84, 0, 70, 5, 84, 0, 70, 134, 74, 0, 92, 5, 74, 0, 92, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 81, 0, 64, 5, 80, 0, 68, 5, 77, 0, 80, 5, 70, 0, 60, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 42, 134, 71, 0, 46, 5, 71, 0, 46, 134, 74, 0, 64, 134, 72, 0, 80, 5, 74, 0, 78, 5, 78, 0, 92, 0, 0, 0, 0, 0, 0, 0, 0, 134, 71, 0, 44, 134, 73, 0, 82, 134, 78, 0, 84, 134, 76, 0, 50, 5, 76, 0, 50, 134, 72, 0, 60, 5, 72, 0, 60, 134, 73, 0, 74, 5, 73, 0, 74, 134, 73, 0, 76, 5, 73, 0, 76, 134, 71, 0, 90, 5, 71, 0, 90, 5, 73, 0, 42, 5, 74, 0, 42, 5, 76, 0, 42, 5, 76, 0, 44, 5, 73, 0, 70, 5, 74, 0, 46, 5, 74, 0, 48, 5, 75, 0, 48, 5, 85, 0, 50, 5, 81, 0, 50, 5, 86, 0, 50, 5, 83, 0, 50, 5, 87, 0, 50, 5, 90, 0, 50, 5, 91, 0, 50, 5, 74, 0, 60, 5, 74, 0, 54, 5, 70, 0, 54, 5, 71, 0, 54, 5, 78, 0, 54, 5, 79, 0, 54, 5, 76, 0, 56, 5, 71, 0, 56, 5, 75, 0, 56, 5, 74, 0, 58, 5, 77, 0, 58, 5, 71, 0, 58, 5, 76, 0, 64, 5, 77, 0, 64, 5, 79, 0, 64, 197, 162, 0, 0, 5, 77, 0, 66, 5, 80, 0, 66, 5, 73, 0, 66, 5, 75, 0, 68, 5, 79, 0, 68, 5, 71, 0, 68, 5, 81, 0, 70, 197, 226, 0, 0, 5, 83, 0, 70, 5, 79, 0, 72, 5, 75, 0, 76, 5, 77, 0, 76, 5, 79, 0, 76, 5, 81, 0, 76, 5, 82, 0, 76, 5, 83, 0, 76, 5, 85, 0, 76, 5, 70, 0, 76, 5, 86, 0, 76, 5, 73, 0, 78, 5, 77, 0, 78, 5, 75, 0, 60, 5, 80, 0, 78, 5, 82, 0, 78, 5, 79, 0, 80, 5, 76, 0, 80, 5, 73, 0, 82, 5, 76, 0, 82, 5, 74, 0, 84, 5, 78, 0, 84, 5, 72, 0, 86, 5, 74, 0, 90, 5, 70, 0, 90, 5, 75, 0, 92, 5, 76, 0, 92, 201, 230, 0, 0, 5, 89, 0, 92, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 70, 0, 44, 5, 92, 0, 50, 5, 75, 0, 54, 5, 70, 0, 56, 5, 73, 0, 60, 5, 78, 0, 62, 5, 70, 0, 64, 5, 72, 0, 74, 0, 0, 0, 0, 0, 0, 0, 0, 192, 0, 0, 0, 197, 34, 1, 0, 197, 98, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 77, 0, 56, 5, 78, 0, 56, 192, 0, 0, 0, 7, 71, 0, 56, 192, 0, 0, 0, 192, 0, 0, 0, 6, 75, 0, 76, 6, 79, 0, 76, 6, 86, 0, 76, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 78, 0, 54, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 32, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 170, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 32, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 170, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 79, 97, 5, 5, 79, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 100, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 0, 42, 197, 226, 1, 0, 197, 34, 2, 0, 5, 73, 0, 44, 5, 70, 0, 46, 5, 70, 0, 48, 5, 71, 0, 48, 5, 75, 0, 50, 5, 89, 0, 50, 5, 73, 0, 58, 5, 71, 0, 60, 5, 70, 0, 62, 5, 72, 0, 64, 5, 70, 0, 66, 5, 73, 0, 68, 5, 70, 0, 70, 5, 74, 0, 70, 5, 71, 0, 70, 5, 75, 0, 70, 5, 72, 0, 70, 197, 98, 2, 0, 5, 85, 0, 70, 5, 78, 0, 70, 5, 79, 0, 70, 5, 70, 0, 72, 5, 72, 0, 76, 5, 76, 0, 76, 5, 70, 0, 80, 5, 70, 0, 82, 5, 71, 0, 82, 5, 72, 0, 82, 5, 79, 0, 66, 5, 71, 0, 84, 5, 70, 0, 86, 5, 70, 0, 92, 5, 85, 0, 92, 0, 0, 0, 0, 0, 0, 0, 0, 5, 70, 10, 96, 5, 70, 38, 96, 5, 70, 50, 96, 5, 70, 60, 96, 5, 70, 82, 96, 5, 70, 98, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 72, 0, 44, 192, 0, 0, 0, 192, 0, 0, 0, 135, 79, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 72, 0, 68, 192, 0, 0, 0, 135, 84, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 73, 0, 42, 7, 74, 0, 42, 197, 162, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 81, 0, 50, 8, 83, 0, 50, 6, 89, 0, 50, 192, 0, 0, 0, 6, 73, 0, 58, 192, 0, 0, 0, 192, 0, 0, 0, 7, 82, 0, 68, 192, 0, 0, 0, 7, 73, 0, 70, 6, 78, 0, 70, 6, 79, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 0, 82, 7, 77, 0, 66, 192, 0, 0, 0, 0, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 2, 0, 5, 74, 0, 44, 5, 72, 0, 48, 5, 71, 0, 52, 5, 71, 0, 66, 5, 74, 0, 68, 5, 73, 0, 72, 5, 74, 0, 76, 5, 84, 0, 76, 5, 71, 0, 78, 5, 73, 0, 80, 5, 72, 0, 92, 5, 76, 0, 54, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 75, 0, 58, 5, 78, 0, 58, 5, 71, 0, 72, 5, 74, 0, 82, 5, 77, 0, 82, 5, 75, 0, 44, 5, 73, 0, 48, 5, 72, 0, 52, 5, 73, 0, 54, 5, 71, 0, 62, 5, 78, 0, 64, 5, 72, 0, 66, 5, 78, 0, 68, 5, 74, 0, 72, 5, 78, 0, 76, 5, 72, 0, 78, 5, 78, 0, 78, 5, 73, 0, 84, 5, 70, 0, 88, 5, 73, 0, 92, 5, 72, 0, 42, 5, 75, 0, 42, 5, 76, 0, 48, 5, 77, 0, 50, 5, 84, 0, 50, 5, 88, 0, 50, 5, 82, 0, 50, 5, 76, 0, 58, 5, 76, 0, 70, 5, 81, 0, 78, 5, 75, 0, 82, 5, 87, 0, 92, 7, 76, 0, 42, 192, 0, 0, 0, 6, 74, 0, 46, 192, 0, 0, 0, 6, 87, 0, 50, 192, 0, 0, 0, 6, 74, 0, 60, 6, 70, 0, 54, 7, 76, 0, 56, 7, 74, 0, 58, 7, 77, 0, 58, 6, 71, 0, 58, 6, 75, 0, 58, 6, 73, 0, 60, 6, 79, 0, 64, 6, 78, 0, 64, 7, 70, 0, 64, 7, 73, 0, 66, 6, 80, 0, 66, 7, 75, 0, 68, 6, 79, 0, 68, 7, 71, 0, 68, 7, 81, 0, 70, 6, 79, 0, 72, 6, 73, 0, 78, 7, 77, 0, 78, 6, 74, 0, 80, 7, 73, 0, 82, 7, 76, 0, 82, 6, 70, 0, 82, 7, 74, 0, 84, 7, 78, 0, 84, 192, 0, 0, 0, 6, 75, 0, 92, 6, 76, 0, 92, 7, 81, 0, 92, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 0, 54, 192, 0, 0, 0, 192, 0, 0, 0, 6, 70, 0, 64, 6, 70, 0, 66, 192, 0, 0, 0, 6, 71, 0, 68, 6, 70, 0, 76, 6, 71, 0, 76, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 75, 0, 78, 5, 76, 0, 78, 192, 0, 0, 0, 5, 80, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 77, 0, 84, 5, 77, 0, 84, 134, 73, 0, 90, 5, 73, 0, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 8, 81, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 135, 83, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 74, 0, 52, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 74, 0, 52, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 76, 0, 46, 5, 76, 0, 46, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 75, 0, 64, 5, 75, 0, 64, 134, 76, 0, 64, 134, 71, 0, 72, 134, 82, 0, 76, 5, 71, 0, 42, 5, 72, 0, 80, 134, 72, 0, 56, 5, 72, 0, 56, 134, 73, 0, 62, 5, 73, 0, 62, 134, 79, 0, 92, 5, 79, 0, 92, 134, 74, 0, 42, 134, 73, 0, 66, 134, 73, 0, 42, 134, 76, 0, 42, 5, 75, 0, 84, 134, 71, 0, 86, 5, 71, 0, 86, 5, 76, 0, 84, 134, 73, 0, 56, 5, 73, 0, 56, 5, 80, 0, 72, 5, 78, 0, 50, 5, 80, 0, 76, 5, 80, 0, 70, 5, 80, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 134, 74, 0, 78, 134, 78, 0, 92, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 134, 74, 0, 56, 5, 74, 0, 56, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 70, 0, 52, 5, 70, 0, 78, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 77, 0, 46, 5, 77, 0, 46, 134, 74, 0, 62, 5, 74, 0, 62, 134, 75, 0, 62, 5, 75, 0, 62, 134, 76, 0, 62, 5, 76, 0, 62, 134, 71, 0, 64, 5, 71, 0, 64, 134, 73, 0, 64, 5, 73, 0, 64, 134, 82, 0, 70, 5, 82, 0, 70, 134, 77, 0, 70, 5, 77, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 134, 72, 0, 72, 5, 72, 0, 72, 134, 76, 0, 72, 5, 76, 0, 72, 134, 77, 0, 72, 5, 77, 0, 72, 134, 70, 0, 74, 5, 70, 0, 74, 134, 71, 0, 74, 5, 71, 0, 74, 135, 71, 0, 76, 5, 71, 0, 76, 134, 89, 0, 76, 5, 89, 0, 76, 134, 72, 0, 84, 5, 72, 0, 84, 192, 0, 0, 0, 192, 0, 0, 0, 134, 80, 0, 92, 5, 80, 0, 92, 134, 70, 38, 93, 5, 70, 38, 93, 134, 71, 38, 93, 5, 71, 38, 93, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 79, 0, 48, 5, 82, 0, 64, 5, 76, 0, 66, 5, 81, 0, 68, 5, 87, 0, 76, 5, 88, 0, 76, 5, 78, 0, 80, 0, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 77, 0, 54, 5, 77, 0, 54, 134, 83, 0, 64, 5, 83, 0, 64, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 134, 76, 0, 56, 5, 80, 0, 64, 192, 0, 0, 0, 134, 77, 0, 68, 5, 77, 0, 68, 134, 72, 0, 46, 5, 72, 0, 46, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 56, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 78, 0, 66, 5, 75, 0, 52, 5, 78, 0, 72, 5, 74, 0, 66, 5, 72, 0, 58, 5, 75, 0, 66, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 70, 0, 58, 6, 70, 0, 60, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 56, 5, 70, 0, 84, 133, 5, 0, 56, 6, 70, 0, 84, 5, 5, 0, 64, 5, 82, 0, 92, 5, 5, 0, 70, 5, 71, 0, 50, 5, 5, 0, 48, 5, 83, 0, 92, 5, 5, 0, 48, 5, 77, 0, 92, 5, 5, 0, 80, 5, 75, 0, 46, 5, 5, 0, 42, 5, 73, 0, 50, 5, 5, 0, 42, 5, 74, 0, 50, 5, 5, 0, 70, 5, 72, 0, 50, 5, 5, 0, 42, 6, 74, 0, 50, 5, 5, 0, 82, 5, 70, 0, 50, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 92, 134, 81, 48, 0, 12, 3, 255, 255, 0, 92, 134, 84, 0, 92, 5, 81, 48, 0, 12, 3, 255, 255, 0, 92, 5, 84, 16, 97, 174, 5, 1, 0, 1, 3, 255, 255, 32, 97, 134, 70, 65, 3, 255, 255, 32, 97, 134, 70, 84, 97, 174, 5, 1, 0, 1, 3, 255, 255, 170, 97, 134, 70, 65, 3, 255, 255, 170, 97, 134, 70, 16, 97, 5, 5, 1, 0, 1, 3, 255, 255, 32, 97, 5, 70, 65, 3, 255, 255, 32, 97, 5, 70, 84, 97, 5, 5, 1, 0, 1, 3, 255, 255, 170, 97, 5, 70, 65, 3, 255, 255, 170, 97, 5, 70, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 22, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 92, 185, 96, 185, 100, 185, 104, 185, 108, 185, 112, 185, 116, 185, 120, 185, 124, 186, 128, 185, 132, 185, 136, 185, 140, 185, 144, 185, 148, 185, 152, 185, 156, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 92, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 168, 116, 168, 120, 168, 124, 168, 128, 168, 132, 168, 136, 168, 140, 168, 144, 168, 148, 168, 152, 168, 156, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 112, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 153, 110, 185, 113, 153, 113, 249, 113, 153, 114, 89, 114, 136, 14, 217, 114, 185, 137, 153, 137, 249, 137, 89, 138, 153, 153, 185, 160, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 110, 168, 113, 136, 113, 232, 113, 136, 114, 72, 114, 128, 14, 200, 114, 168, 137, 136, 137, 232, 137, 72, 138, 136, 153, 168, 160, 72, 154, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 200, 88, 6, 8, 8, 8, 250, 93, 232, 93, 249, 98, 232, 98, 200, 96, 153, 101, 136, 101, 249, 102, 232, 102, 25, 102, 8, 102, 121, 103, 104, 103, 89, 103, 72, 103, 153, 109, 136, 109, 249, 110, 232, 110, 25, 110, 8, 110, 200, 108, 249, 108, 232, 108, 57, 115, 40, 115, 217, 113, 200, 113, 121, 114, 104, 114, 10, 8, 12, 8, 153, 125, 136, 125, 249, 126, 232, 126, 25, 126, 8, 126, 154, 129, 136, 129, 250, 129, 232, 129, 250, 130, 232, 130, 26, 130, 8, 130, 249, 134, 232, 134, 25, 134, 8, 134, 217, 132, 200, 132, 153, 138, 136, 138, 57, 139, 40, 139, 217, 137, 200, 137, 57, 138, 40, 138, 121, 138, 104, 138, 25, 139, 8, 139, 249, 145, 232, 145, 249, 153, 232, 153, 89, 154, 153, 157, 136, 157, 185, 158, 168, 158, 25, 158, 8, 158, 232, 128, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 14, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 16, 8, 1, 0, 8, 14, 18, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 20, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 128, 169, 128, 200, 56, 169, 72, 187, 88, 187, 92, 169, 88, 169, 92, 217, 112, 187, 72, 200, 112, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 1, 3, 0, 83, 0, 74, 0, 41, 0, 135, 8, 1, 16, 151, 6, 0, 96, 1, 0, 0, 96, 106, 34, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 61, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 228, 31, 0, 0, 228, 31, 0, 0, 228, 31, 0, 0, 228, 31, 0, 0, 12, 34, 0, 0, 12, 34, 0, 0, 82, 34, 0, 0, 82, 34, 0, 0, 82, 34, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 241, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 94, 0, 102, 0, 110, 0, 118, 0, 48, 0, 56, 0, 117, 0, 125, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 142, 0, 150, 0, 48, 0, 48, 0, 48, 0, 158, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 48, 0, 48, 0, 174, 0, 48, 0, 180, 0, 188, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 191, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 199, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 202, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 140, 0, 48, 0, 48, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 210, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 218, 0, 48, 0, 48, 0, 224, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 120, 1, 184, 1, 192, 0, 212, 1, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 16, 0, 202, 33, 16, 0, 202, 66, 16, 0, 202, 99, 16, 0, 202, 132, 16, 0, 202, 165, 16, 0, 202, 198, 16, 0, 202, 231, 16, 0, 202, 8, 17, 0, 202, 41, 17, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 6, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 7, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 7, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 198, 0, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 2, 0, 5, 71, 238, 102, 201, 134, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 0, 0, 192, 0, 0, 0, 197, 226, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 1, 0, 192, 0, 0, 0, 197, 162, 1, 0, 192, 0, 0, 0, 197, 226, 1, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 197, 98, 2, 0, 197, 163, 2, 0, 197, 2, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 3, 0, 197, 131, 3, 0, 197, 226, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 34, 0, 0, 197, 98, 4, 0, 5, 5, 10, 121, 197, 162, 4, 0, 197, 226, 4, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 34, 0, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 98, 0, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 162, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 34, 1, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 98, 1, 0, 5, 5, 107, 121, 197, 162, 1, 0, 5, 5, 111, 121, 197, 226, 1, 0, 5, 5, 115, 121, 197, 34, 2, 0, 5, 5, 119, 121, 197, 98, 2, 0, 197, 163, 2, 0, 197, 2, 3, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 66, 3, 0, 197, 131, 3, 0, 197, 226, 3, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 34, 4, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 34, 0, 0, 197, 98, 4, 0, 5, 5, 10, 121, 197, 162, 4, 0, 197, 226, 4, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 170, 170, 3, 0, 83, 0, 74, 0, 41, 0, 91, 21, 1, 16, 27, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 2, 0, 0, 0, 16, 32, 2, 15, 3, 0, 83, 0, 74, 0, 41, 0, 255, 29, 1, 16, 0, 0, 0, 0, 1, 0, 0, 96, 6, 0, 60, 0, 68, 0, 98, 0, 102, 0, 117, 0, 124, 0, 170, 170, 6, 0, 0, 96, 129, 8, 1, 32, 86, 21, 1, 32, 27, 204, 1, 80, 247, 29, 1, 32, 8, 30, 1, 32, 5, 0, 133, 0, 32, 0, 41, 0, 49, 0, 146, 0, 25, 204, 1, 144, 22, 65, 0, 96, 1, 0, 0, 96, 13, 30, 1, 32, 15, 0, 0, 96, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 122, 0, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 126, 0, 0, 0, 126, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 60, 1, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 8, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 32, 0, 0, 0, 36, 0, 0, 0, 36, 1, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 39, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 3, 0, 69, 0, 60, 0, 32, 0, 35, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 115, 0, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 120, 0, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 105, 28, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 109, 28, 0, 0, 109, 28, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 204, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 49, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 115, 0, 101, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 0, 0, 47, 221, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 98, 0, 60, 0, 225, 0, 60, 0, 60, 0, 60, 0, 193, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 100, 0, 60, 0, 13, 1, 60, 0, 60, 0, 60, 0, 12, 1, 60, 0, 146, 2, 60, 0, 60, 0, 60, 0, 183, 1, 60, 0, 239, 1, 60, 0, 60, 0, 60, 0, 238, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 101, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 60, 0, 60, 0, 240, 0, 60, 0, 60, 0, 60, 0, 208, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 104, 0, 60, 0, 231, 1, 60, 0, 60, 0, 60, 0, 230, 1, 60, 0, 229, 1, 60, 0, 60, 0, 60, 0, 228, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 108, 0, 60, 0, 233, 1, 60, 0, 60, 0, 60, 0, 232, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 111, 0, 60, 0, 75, 1, 60, 0, 60, 0, 60, 0, 74, 1, 60, 0, 60, 0, 68, 1, 60, 0, 60, 0, 60, 0, 67, 1, 60, 0, 60, 0, 241, 0, 60, 0, 60, 0, 60, 0, 209, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 116, 0, 60, 0, 97, 1, 60, 0, 60, 0, 60, 0, 96, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 117, 0, 60, 0, 103, 1, 60, 0, 60, 0, 60, 0, 102, 1, 60, 0, 60, 0, 254, 0, 60, 0, 60, 0, 60, 0, 222, 0, 38, 0, 121, 0, 60, 0, 60, 0, 252, 0, 60, 0, 60, 0, 60, 0, 220, 0, 60, 0, 60, 0, 113, 1, 60, 0, 60, 0, 60, 0, 112, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 192, 1, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 60, 0, 60, 0, 83, 1, 60, 0, 60, 0, 60, 0, 82, 1, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 60, 0, 60, 0, 39, 2, 60, 0, 60, 0, 60, 0, 38, 2, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 60, 0, 227, 0, 60, 0, 60, 0, 60, 0, 195, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 81, 1, 60, 0, 60, 0, 60, 0, 80, 1, 60, 0, 60, 0, 245, 0, 60, 0, 60, 0, 60, 0, 213, 0, 60, 0, 60, 0, 244, 0, 60, 0, 60, 0, 60, 0, 212, 0, 60, 0, 60, 0, 235, 1, 60, 0, 60, 0, 60, 0, 234, 1, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 248, 57, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 6, 1, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 148, 40, 0, 0, 148, 40, 0, 0, 236, 45, 0, 0, 236, 45, 0, 0, 56, 51, 0, 0, 56, 51, 0, 0, 220, 57, 0, 0, 220, 57, 0, 0, 220, 57, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 124, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 126, 0, 134, 0, 142, 0, 150, 0, 48, 0, 56, 0, 154, 0, 162, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 159, 0, 167, 0, 173, 0, 181, 0, 189, 0, 197, 0, 48, 0, 56, 0, 196, 0, 204, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 221, 0, 229, 0, 48, 0, 48, 0, 48, 0, 237, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 245, 0, 48, 0, 48, 0, 253, 0, 48, 0, 3, 1, 11, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 14, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 22, 1, 26, 1, 33, 1, 41, 1, 47, 1, 55, 1, 60, 1, 68, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 74, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 82, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 85, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 212, 0, 219, 0, 48, 0, 48, 0, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 115, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 93, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 101, 1, 48, 0, 48, 0, 107, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 1, 248, 1, 56, 2, 192, 0, 104, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 124, 2, 180, 2, 244, 2, 192, 0, 16, 3, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 41, 0, 202, 65, 41, 0, 202, 98, 41, 0, 202, 131, 41, 0, 202, 164, 41, 0, 202, 197, 41, 0, 202, 230, 41, 0, 202, 7, 42, 0, 202, 40, 42, 0, 202, 73, 42, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 198, 0, 0, 156, 5, 0, 44, 201, 102, 3, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 38, 4, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 201, 230, 4, 0, 156, 5, 0, 64, 156, 5, 0, 66, 201, 166, 5, 0, 201, 134, 7, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 198, 9, 0, 156, 5, 0, 80, 201, 166, 10, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 70, 12, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 6, 13, 0, 5, 5, 0, 44, 201, 38, 16, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 230, 16, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 201, 166, 17, 0, 5, 5, 0, 64, 5, 5, 0, 66, 201, 102, 18, 0, 201, 70, 20, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 166, 22, 0, 5, 5, 0, 80, 201, 134, 23, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 70, 25, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 134, 5, 31, 43, 196, 142, 156, 42, 198, 1, 0, 0, 198, 65, 0, 0, 198, 97, 0, 0, 198, 1, 2, 0, 201, 38, 26, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 70, 29, 49, 134, 71, 33, 69, 196, 138, 156, 70, 196, 136, 156, 70, 198, 129, 0, 0, 198, 161, 0, 0, 198, 193, 0, 0, 5, 5, 127, 12, 198, 33, 2, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 134, 70, 25, 81, 197, 130, 32, 0, 196, 138, 5, 42, 5, 5, 31, 43, 196, 142, 5, 42, 198, 65, 1, 0, 193, 6, 111, 93, 193, 5, 111, 93, 193, 4, 111, 93, 201, 230, 26, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 70, 29, 49, 5, 71, 33, 69, 196, 138, 5, 70, 196, 136, 5, 70, 198, 129, 1, 0, 198, 161, 1, 0, 193, 7, 111, 93, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 5, 70, 25, 81, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 27, 0, 201, 134, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 21, 47, 5, 5, 21, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 29, 49, 5, 5, 29, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 33, 0, 201, 38, 34, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 34, 0, 201, 166, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 33, 69, 5, 70, 33, 69, 201, 102, 36, 0, 201, 230, 37, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 33, 69, 5, 5, 33, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 225, 0, 0, 198, 193, 1, 0, 198, 193, 3, 0, 198, 225, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 39, 0, 201, 38, 40, 0, 134, 5, 35, 79, 5, 5, 35, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 25, 81, 5, 5, 25, 81, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 230, 40, 0, 201, 6, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 1, 0, 193, 2, 111, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 43, 0, 201, 6, 45, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 46, 0, 201, 6, 48, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 49, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 197, 34, 7, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 98, 7, 0, 197, 162, 7, 0, 197, 226, 7, 0, 197, 34, 8, 0, 192, 0, 0, 0, 198, 194, 5, 0, 197, 98, 8, 0, 198, 2, 6, 0, 198, 66, 6, 0, 198, 130, 6, 0, 197, 162, 8, 0, 134, 5, 28, 55, 5, 5, 28, 55, 134, 5, 27, 55, 5, 5, 27, 55, 134, 5, 21, 63, 5, 5, 21, 63, 201, 230, 49, 0, 201, 198, 51, 0, 198, 130, 8, 0, 198, 194, 8, 0, 134, 5, 23, 47, 5, 5, 23, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 9, 0, 197, 34, 9, 0, 198, 66, 9, 0, 197, 98, 9, 0, 198, 130, 9, 0, 197, 162, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 53, 0, 201, 102, 54, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 198, 97, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 9, 0, 197, 98, 10, 0, 198, 2, 10, 0, 198, 66, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 55, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 230, 55, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 57, 0, 5, 71, 238, 102, 201, 166, 58, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 192, 0, 0, 0, 197, 98, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 11, 0, 192, 0, 0, 0, 197, 34, 12, 0, 192, 0, 0, 0, 197, 98, 12, 0, 192, 0, 0, 0, 197, 162, 12, 0, 192, 0, 0, 0, 197, 226, 12, 0, 197, 35, 13, 0, 197, 130, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 13, 0, 197, 3, 14, 0, 197, 98, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 162, 10, 0, 197, 226, 14, 0, 5, 5, 10, 121, 197, 34, 15, 0, 197, 98, 15, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 15, 0, 197, 226, 15, 0, 197, 34, 16, 0, 197, 98, 16, 0, 197, 162, 16, 0, 197, 226, 16, 0, 197, 34, 17, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 11, 0, 5, 5, 24, 121, 197, 98, 11, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 59, 0, 201, 102, 62, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 65, 0, 201, 6, 66, 0, 201, 198, 66, 0, 201, 134, 67, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 68, 0, 201, 198, 69, 0, 201, 70, 71, 0, 201, 198, 72, 0, 201, 70, 74, 0, 201, 198, 75, 0, 198, 2, 12, 0, 198, 66, 12, 0, 198, 130, 12, 0, 198, 194, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 77, 0, 201, 6, 78, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 22, 0, 197, 34, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 78, 0, 201, 230, 79, 0, 201, 6, 81, 0, 201, 38, 82, 0, 201, 70, 83, 0, 201, 102, 84, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 85, 0, 201, 70, 86, 0, 201, 6, 87, 0, 201, 198, 87, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 88, 0, 201, 102, 91, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 94, 0, 201, 38, 96, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 16, 0, 198, 2, 17, 0, 198, 66, 17, 0, 198, 130, 17, 0, 198, 194, 17, 0, 198, 2, 18, 0, 198, 66, 18, 0, 198, 130, 18, 0, 198, 2, 15, 0, 198, 2, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 4, 0, 198, 66, 5, 0, 201, 6, 98, 0, 201, 230, 99, 0, 201, 198, 101, 0, 201, 230, 102, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 104, 0, 201, 38, 105, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 4, 0, 198, 97, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 70, 0, 6, 111, 93, 0, 134, 0, 70, 0, 5, 111, 93, 0, 134, 0, 5, 0, 6, 111, 93, 0, 134, 0, 5, 0, 5, 111, 93, 0, 134, 0, 72, 0, 7, 111, 93, 0, 134, 0, 71, 0, 7, 111, 93, 0, 134, 0, 5, 0, 7, 111, 93, 0, 134, 0, 70, 0, 7, 111, 93, 0, 134, 0, 73, 0, 7, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 70, 0, 6, 111, 93, 0, 5, 0, 70, 0, 5, 111, 93, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 70, 0, 7, 111, 93, 0, 5, 0, 73, 0, 7, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 6, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 70, 0, 5, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 6, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 5, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 5, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 5, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 70, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 73, 0, 7, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 73, 0, 7, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 7, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 6, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 5, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 6, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 5, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 0, 70, 0, 6, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 5, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 6, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 5, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 134, 5, 21, 47, 5, 160, 0, 0, 5, 5, 21, 47, 5, 160, 0, 0, 134, 5, 31, 43, 5, 162, 0, 0, 5, 5, 31, 43, 5, 162, 0, 0, 193, 6, 111, 93, 5, 162, 0, 0, 193, 5, 111, 93, 5, 162, 0, 0, 134, 5, 27, 55, 5, 160, 0, 0, 5, 5, 27, 55, 5, 160, 0, 0, 134, 5, 21, 63, 5, 160, 0, 0, 5, 5, 21, 63, 5, 160, 0, 0, 134, 70, 33, 69, 5, 160, 0, 0, 134, 71, 33, 69, 5, 160, 0, 0, 5, 70, 33, 69, 5, 160, 0, 0, 5, 71, 33, 69, 5, 160, 0, 0, 134, 5, 35, 79, 5, 160, 0, 0, 5, 5, 35, 79, 5, 160, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 193, 7, 111, 93, 5, 190, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 6, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 164, 0, 0, 193, 7, 111, 93, 5, 162, 0, 0, 193, 5, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 134, 5, 35, 79, 5, 202, 0, 0, 5, 5, 35, 79, 5, 202, 0, 0, 193, 7, 111, 93, 5, 164, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 5, 31, 43, 5, 200, 0, 0, 5, 5, 31, 43, 5, 200, 0, 0, 193, 6, 111, 93, 5, 200, 0, 0, 193, 5, 111, 93, 5, 200, 0, 0, 134, 5, 21, 63, 5, 196, 0, 0, 5, 5, 21, 63, 5, 196, 0, 0, 134, 5, 21, 63, 5, 210, 0, 0, 5, 5, 21, 63, 5, 210, 0, 0, 134, 70, 33, 69, 5, 196, 0, 0, 134, 71, 33, 69, 5, 196, 0, 0, 5, 70, 33, 69, 5, 196, 0, 0, 5, 71, 33, 69, 5, 196, 0, 0, 134, 70, 33, 69, 5, 210, 0, 0, 134, 71, 33, 69, 5, 210, 0, 0, 5, 70, 33, 69, 5, 210, 0, 0, 5, 71, 33, 69, 5, 210, 0, 0, 134, 70, 33, 69, 5, 204, 0, 0, 134, 71, 33, 69, 5, 204, 0, 0, 5, 70, 33, 69, 5, 204, 0, 0, 5, 71, 33, 69, 5, 204, 0, 0, 134, 5, 35, 79, 5, 196, 0, 0, 5, 5, 35, 79, 5, 196, 0, 0, 134, 5, 35, 79, 5, 156, 0, 0, 5, 5, 35, 79, 5, 156, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 2, 111, 93, 5, 196, 0, 0, 193, 2, 111, 93, 5, 210, 0, 0, 134, 5, 31, 43, 5, 196, 0, 0, 5, 5, 31, 43, 5, 196, 0, 0, 193, 6, 111, 93, 5, 196, 0, 0, 193, 5, 111, 93, 5, 196, 0, 0, 193, 7, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 193, 7, 111, 93, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 6, 121, 197, 162, 10, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 226, 10, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 34, 11, 0, 5, 5, 24, 121, 197, 98, 11, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 162, 11, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 226, 11, 0, 5, 5, 107, 121, 197, 34, 12, 0, 5, 5, 111, 121, 197, 98, 12, 0, 5, 5, 115, 121, 197, 162, 12, 0, 5, 5, 119, 121, 197, 226, 12, 0, 197, 35, 13, 0, 197, 130, 13, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 194, 13, 0, 197, 3, 14, 0, 197, 98, 14, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 162, 14, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 162, 10, 0, 197, 226, 14, 0, 5, 5, 10, 121, 197, 34, 15, 0, 197, 98, 15, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 15, 0, 197, 226, 15, 0, 197, 34, 16, 0, 197, 98, 16, 0, 197, 162, 16, 0, 197, 226, 16, 0, 197, 34, 17, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 11, 0, 5, 5, 24, 121, 197, 98, 11, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 5, 0, 8, 3, 9, 0, 8, 3, 0, 192, 198, 65, 10, 3, 0, 192, 198, 97, 65, 3, 31, 235, 134, 5, 1, 3, 31, 235, 134, 5, 3, 3, 198, 129, 7, 3, 198, 161, 0, 46, 156, 5, 48, 0, 12, 3, 21, 239, 134, 5, 0, 54, 156, 5, 48, 0, 12, 3, 27, 247, 134, 5, 0, 62, 156, 5, 48, 0, 12, 3, 21, 255, 134, 5, 0, 68, 156, 5, 2, 0, 1, 3, 255, 255, 33, 69, 134, 70, 3, 3, 255, 255, 33, 69, 134, 71, 65, 3, 255, 255, 33, 69, 134, 70, 0, 70, 156, 5, 4, 0, 2, 3, 0, 192, 198, 129, 3, 3, 0, 192, 198, 161, 8, 3, 0, 192, 198, 193, 11, 3, 0, 192, 198, 225, 40, 3, 1, 192, 198, 1, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 35, 79, 134, 5, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 92, 156, 5, 48, 0, 12, 3, 1, 192, 198, 33, 0, 42, 5, 5, 5, 0, 8, 3, 11, 0, 8, 3, 255, 255, 111, 93, 193, 6, 10, 3, 255, 255, 111, 93, 193, 5, 65, 3, 31, 235, 5, 5, 1, 3, 31, 235, 5, 5, 3, 3, 1, 192, 198, 65, 7, 3, 1, 192, 198, 97, 0, 46, 5, 5, 48, 0, 12, 3, 21, 239, 5, 5, 0, 54, 5, 5, 48, 0, 12, 3, 27, 247, 5, 5, 0, 62, 5, 5, 48, 0, 12, 3, 21, 255, 5, 5, 0, 68, 5, 5, 2, 0, 1, 3, 255, 255, 33, 69, 5, 70, 3, 3, 255, 255, 33, 69, 5, 71, 65, 3, 255, 255, 33, 69, 5, 70, 0, 70, 5, 5, 4, 0, 2, 3, 1, 192, 198, 129, 3, 3, 1, 192, 198, 161, 8, 3, 255, 255, 111, 93, 193, 7, 11, 3, 1, 192, 198, 193, 40, 3, 1, 192, 198, 225, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 35, 79, 5, 5, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 111, 93, 193, 2, 156, 46, 196, 160, 48, 0, 12, 3, 0, 192, 197, 162, 5, 46, 196, 160, 48, 0, 12, 3, 0, 192, 197, 226, 156, 42, 196, 162, 5, 0, 8, 3, 9, 0, 8, 3, 2, 192, 198, 194, 10, 3, 3, 192, 198, 2, 65, 3, 1, 192, 197, 34, 1, 3, 1, 192, 197, 34, 3, 3, 2, 192, 198, 66, 7, 3, 2, 192, 198, 130, 5, 42, 196, 162, 5, 0, 8, 3, 9, 0, 8, 3, 1, 192, 197, 162, 10, 3, 1, 192, 197, 226, 65, 3, 1, 192, 197, 98, 1, 3, 1, 192, 197, 98, 3, 3, 3, 192, 198, 66, 7, 3, 3, 192, 198, 130, 156, 54, 196, 160, 48, 0, 12, 3, 2, 192, 197, 34, 5, 54, 196, 160, 48, 0, 12, 3, 2, 192, 197, 98, 156, 62, 196, 160, 48, 0, 12, 3, 2, 192, 197, 162, 5, 62, 196, 160, 48, 0, 12, 3, 2, 192, 197, 226, 156, 68, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 3, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 5, 68, 196, 160, 2, 0, 1, 3, 3, 192, 197, 162, 3, 3, 3, 192, 197, 226, 65, 3, 3, 192, 197, 162, 156, 78, 196, 160, 48, 0, 12, 3, 4, 192, 197, 34, 5, 78, 196, 160, 48, 0, 12, 3, 4, 192, 197, 98, 156, 82, 196, 162, 1, 0, 8, 3, 4, 192, 197, 162, 11, 3, 4, 192, 197, 226, 5, 82, 196, 162, 1, 0, 8, 3, 5, 192, 197, 34, 11, 3, 5, 192, 197, 98, 156, 70, 196, 190, 3, 0, 2, 3, 4, 192, 198, 2, 3, 3, 4, 192, 198, 66, 8, 3, 4, 192, 198, 130, 11, 3, 4, 192, 198, 194, 5, 70, 196, 190, 3, 0, 2, 3, 5, 192, 198, 2, 3, 3, 5, 192, 198, 66, 8, 3, 5, 192, 197, 162, 11, 3, 5, 192, 198, 130, 156, 82, 196, 190, 1, 0, 8, 3, 5, 192, 197, 226, 11, 3, 6, 192, 197, 34, 5, 82, 196, 190, 1, 0, 8, 3, 6, 192, 197, 98, 11, 3, 6, 192, 197, 162, 22, 47, 134, 5, 48, 0, 12, 3, 23, 239, 134, 5, 1, 0, 198, 1, 3, 0, 2, 3, 6, 192, 198, 194, 3, 3, 7, 192, 198, 2, 8, 3, 7, 192, 198, 66, 11, 3, 7, 192, 198, 130, 1, 0, 198, 225, 3, 0, 2, 3, 7, 192, 198, 194, 3, 3, 8, 192, 198, 2, 8, 3, 8, 192, 197, 226, 11, 3, 8, 192, 198, 66, 156, 78, 196, 202, 48, 0, 12, 3, 9, 192, 197, 226, 5, 78, 196, 202, 48, 0, 12, 3, 10, 192, 197, 34, 22, 47, 5, 5, 48, 0, 12, 3, 23, 239, 5, 5, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 5, 0, 8, 3, 9, 0, 8, 3, 11, 192, 198, 2, 10, 3, 11, 192, 198, 66, 65, 3, 17, 192, 197, 98, 1, 3, 17, 192, 197, 98, 3, 3, 10, 192, 198, 130, 7, 3, 10, 192, 198, 194, 5, 42, 196, 200, 5, 0, 8, 3, 9, 0, 8, 3, 17, 192, 197, 226, 10, 3, 18, 192, 197, 34, 65, 3, 17, 192, 197, 162, 1, 3, 17, 192, 197, 162, 3, 3, 11, 192, 198, 130, 7, 3, 11, 192, 198, 194, 156, 62, 196, 196, 48, 0, 12, 3, 18, 192, 197, 98, 5, 62, 196, 196, 48, 0, 12, 3, 18, 192, 197, 162, 156, 62, 196, 210, 48, 0, 12, 3, 18, 192, 197, 226, 5, 62, 196, 210, 48, 0, 12, 3, 19, 192, 197, 34, 156, 68, 196, 196, 2, 0, 1, 3, 19, 192, 197, 98, 3, 3, 19, 192, 197, 162, 65, 3, 19, 192, 197, 98, 5, 68, 196, 196, 2, 0, 1, 3, 19, 192, 197, 226, 3, 3, 20, 192, 197, 34, 65, 3, 19, 192, 197, 226, 156, 68, 196, 210, 2, 0, 1, 3, 20, 192, 197, 98, 3, 3, 20, 192, 197, 162, 65, 3, 20, 192, 197, 98, 5, 68, 196, 210, 2, 0, 1, 3, 20, 192, 197, 226, 3, 3, 21, 192, 197, 34, 65, 3, 20, 192, 197, 226, 156, 68, 196, 204, 2, 0, 1, 3, 21, 192, 197, 98, 3, 3, 21, 192, 197, 162, 65, 3, 21, 192, 197, 98, 5, 68, 196, 204, 2, 0, 1, 3, 21, 192, 197, 226, 3, 3, 22, 192, 197, 34, 65, 3, 21, 192, 197, 226, 156, 78, 196, 196, 48, 0, 12, 3, 22, 192, 197, 98, 5, 78, 196, 196, 48, 0, 12, 3, 22, 192, 197, 162, 156, 82, 196, 198, 1, 0, 8, 3, 23, 192, 197, 98, 11, 3, 23, 192, 197, 162, 5, 82, 196, 198, 1, 0, 8, 3, 23, 192, 197, 226, 11, 3, 24, 192, 197, 34, 156, 82, 196, 208, 1, 0, 8, 3, 24, 192, 197, 98, 11, 3, 24, 192, 197, 162, 5, 82, 196, 208, 1, 0, 8, 3, 24, 192, 197, 226, 11, 3, 25, 192, 197, 34, 156, 82, 196, 204, 1, 0, 8, 3, 25, 192, 197, 98, 11, 3, 25, 192, 197, 162, 5, 82, 196, 204, 1, 0, 8, 3, 25, 192, 197, 226, 11, 3, 26, 192, 197, 34, 156, 92, 196, 196, 48, 0, 12, 3, 13, 192, 198, 2, 5, 92, 196, 196, 48, 0, 12, 3, 26, 192, 197, 98, 156, 92, 196, 210, 48, 0, 12, 3, 13, 192, 198, 66, 5, 92, 196, 210, 48, 0, 12, 3, 26, 192, 197, 162, 156, 42, 196, 196, 5, 0, 8, 3, 9, 0, 8, 3, 14, 192, 198, 2, 10, 3, 14, 192, 198, 66, 65, 3, 26, 192, 197, 226, 1, 3, 26, 192, 197, 226, 3, 3, 13, 192, 198, 130, 7, 3, 13, 192, 198, 194, 5, 42, 196, 196, 5, 0, 8, 3, 9, 0, 8, 3, 27, 192, 197, 98, 10, 3, 27, 192, 197, 162, 65, 3, 27, 192, 197, 34, 1, 3, 27, 192, 197, 34, 3, 3, 14, 192, 198, 130, 7, 3, 14, 192, 198, 194, 156, 70, 196, 196, 3, 0, 2, 3, 15, 192, 198, 2, 3, 3, 15, 192, 198, 66, 8, 3, 15, 192, 198, 130, 11, 3, 15, 192, 198, 194, 5, 70, 196, 196, 3, 0, 2, 3, 16, 192, 198, 2, 3, 3, 16, 192, 198, 66, 8, 3, 27, 192, 197, 226, 11, 3, 16, 192, 198, 130, 28, 0, 197, 35, 3, 0, 2, 3, 18, 192, 198, 195, 3, 3, 19, 192, 198, 35, 8, 3, 19, 192, 198, 131, 11, 3, 19, 192, 198, 227, 28, 0, 197, 131, 3, 0, 2, 3, 20, 192, 198, 67, 3, 3, 20, 192, 198, 163, 8, 3, 28, 192, 197, 227, 11, 3, 21, 192, 198, 3, 156, 82, 196, 196, 1, 0, 8, 3, 29, 192, 197, 66, 11, 3, 29, 192, 197, 130, 5, 82, 196, 196, 1, 0, 8, 3, 29, 192, 197, 194, 11, 3, 30, 192, 197, 2, 30, 0, 197, 67, 1, 0, 8, 3, 30, 192, 197, 163, 11, 3, 31, 192, 197, 3, 31, 0, 197, 99, 1, 0, 8, 3, 31, 192, 197, 195, 11, 3, 32, 192, 197, 35, 3, 0, 67, 0, 58, 0, 32, 0, 207, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 46, 52, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 196, 32, 0, 0, 196, 32, 0, 0, 28, 38, 0, 0, 28, 38, 0, 0, 132, 41, 0, 0, 132, 41, 0, 0, 226, 47, 0, 0, 226, 47, 0, 0, 18, 52, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 255, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 126, 0, 134, 0, 142, 0, 150, 0, 48, 0, 56, 0, 154, 0, 162, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 186, 0, 190, 0, 197, 0, 205, 0, 211, 0, 219, 0, 224, 0, 232, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 238, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 246, 0, 178, 0, 48, 0, 48, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 170, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 1, 248, 1, 56, 2, 192, 0, 104, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 26, 0, 202, 33, 26, 0, 202, 66, 26, 0, 202, 99, 26, 0, 202, 132, 26, 0, 202, 165, 26, 0, 202, 198, 26, 0, 202, 231, 26, 0, 202, 8, 27, 0, 202, 41, 27, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 201, 166, 2, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 102, 3, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 201, 38, 4, 0, 156, 5, 0, 64, 156, 5, 0, 66, 201, 230, 4, 0, 201, 198, 6, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 6, 9, 0, 156, 5, 0, 80, 201, 230, 9, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 134, 11, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 70, 12, 0, 5, 5, 0, 44, 201, 102, 15, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 38, 16, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 201, 230, 16, 0, 5, 5, 0, 64, 5, 5, 0, 66, 201, 166, 17, 0, 201, 134, 19, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 230, 21, 0, 5, 5, 0, 80, 201, 198, 22, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 134, 24, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 134, 5, 31, 43, 196, 142, 156, 42, 198, 1, 0, 0, 198, 65, 0, 0, 198, 97, 0, 0, 198, 1, 2, 0, 201, 102, 25, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 70, 29, 49, 134, 71, 33, 69, 196, 138, 156, 70, 196, 136, 156, 70, 198, 129, 0, 0, 198, 161, 0, 0, 198, 193, 0, 0, 5, 5, 127, 12, 198, 33, 2, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 134, 70, 25, 81, 197, 194, 25, 0, 196, 138, 5, 42, 5, 5, 31, 43, 196, 142, 5, 42, 198, 65, 1, 0, 193, 6, 111, 93, 193, 5, 111, 93, 193, 4, 111, 93, 201, 38, 26, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 70, 29, 49, 5, 71, 33, 69, 196, 138, 5, 70, 196, 136, 5, 70, 198, 129, 1, 0, 198, 161, 1, 0, 193, 7, 111, 93, 5, 5, 125, 12, 193, 3, 111, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 5, 70, 25, 81, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 26, 0, 201, 198, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 21, 47, 5, 5, 21, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 29, 49, 5, 5, 29, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 32, 0, 201, 102, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 34, 0, 201, 230, 34, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 33, 69, 5, 70, 33, 69, 201, 166, 35, 0, 201, 38, 37, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 33, 69, 5, 5, 33, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 225, 0, 0, 198, 193, 1, 0, 198, 193, 3, 0, 198, 225, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 38, 0, 201, 102, 39, 0, 134, 5, 35, 79, 5, 5, 35, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 25, 81, 5, 5, 25, 81, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 38, 40, 0, 201, 70, 41, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 1, 0, 193, 2, 111, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 42, 0, 201, 70, 44, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 46, 0, 201, 70, 47, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 48, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 197, 34, 7, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 98, 7, 0, 197, 162, 7, 0, 197, 226, 7, 0, 197, 34, 8, 0, 192, 0, 0, 0, 198, 194, 5, 0, 197, 98, 8, 0, 198, 2, 6, 0, 198, 66, 6, 0, 198, 130, 6, 0, 197, 162, 8, 0, 134, 5, 28, 55, 5, 5, 28, 55, 134, 5, 27, 55, 5, 5, 27, 55, 134, 5, 21, 63, 5, 5, 21, 63, 201, 38, 49, 0, 201, 6, 51, 0, 198, 130, 8, 0, 198, 194, 8, 0, 134, 5, 23, 47, 5, 5, 23, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 2, 9, 0, 197, 34, 9, 0, 198, 66, 9, 0, 197, 98, 9, 0, 198, 130, 9, 0, 197, 162, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 52, 0, 201, 166, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 198, 97, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 9, 0, 197, 98, 10, 0, 198, 2, 10, 0, 198, 66, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 54, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 55, 0, 201, 6, 58, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 60, 0, 201, 166, 61, 0, 201, 102, 62, 0, 201, 38, 63, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 63, 0, 201, 102, 65, 0, 201, 230, 66, 0, 201, 102, 68, 0, 201, 230, 69, 0, 201, 102, 71, 0, 198, 2, 12, 0, 198, 66, 12, 0, 198, 130, 12, 0, 198, 194, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 72, 0, 201, 166, 73, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 16, 0, 197, 98, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 74, 0, 201, 134, 75, 0, 201, 166, 76, 0, 201, 198, 77, 0, 201, 230, 78, 0, 201, 6, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 81, 0, 201, 230, 81, 0, 201, 166, 82, 0, 201, 102, 83, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 84, 0, 201, 6, 87, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 89, 0, 201, 198, 91, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 194, 16, 0, 198, 2, 17, 0, 198, 66, 17, 0, 198, 130, 17, 0, 198, 194, 17, 0, 198, 2, 18, 0, 198, 66, 18, 0, 198, 130, 18, 0, 198, 2, 15, 0, 198, 2, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 66, 4, 0, 198, 66, 5, 0, 201, 166, 93, 0, 201, 134, 95, 0, 201, 102, 97, 0, 201, 134, 98, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 99, 0, 201, 198, 100, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 4, 0, 198, 97, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 70, 0, 6, 111, 93, 0, 134, 0, 70, 0, 5, 111, 93, 0, 134, 0, 5, 0, 6, 111, 93, 0, 134, 0, 5, 0, 5, 111, 93, 0, 134, 0, 72, 0, 7, 111, 93, 0, 134, 0, 71, 0, 7, 111, 93, 0, 134, 0, 5, 0, 7, 111, 93, 0, 134, 0, 70, 0, 7, 111, 93, 0, 134, 0, 73, 0, 7, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 70, 0, 6, 111, 93, 0, 5, 0, 70, 0, 5, 111, 93, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 70, 0, 7, 111, 93, 0, 5, 0, 73, 0, 7, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 70, 0, 6, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 70, 0, 5, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 6, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 5, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 70, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 70, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 5, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 5, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 70, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 70, 0, 7, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 73, 0, 7, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 73, 0, 7, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 7, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 6, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 70, 0, 5, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 6, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 0, 70, 0, 5, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 0, 70, 0, 6, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 5, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 6, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 5, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 7, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 134, 0, 72, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 70, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 72, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 70, 0, 7, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 134, 5, 21, 47, 5, 160, 0, 0, 5, 5, 21, 47, 5, 160, 0, 0, 134, 5, 31, 43, 5, 162, 0, 0, 5, 5, 31, 43, 5, 162, 0, 0, 193, 6, 111, 93, 5, 162, 0, 0, 193, 5, 111, 93, 5, 162, 0, 0, 134, 5, 27, 55, 5, 160, 0, 0, 5, 5, 27, 55, 5, 160, 0, 0, 134, 5, 21, 63, 5, 160, 0, 0, 5, 5, 21, 63, 5, 160, 0, 0, 134, 70, 33, 69, 5, 160, 0, 0, 134, 71, 33, 69, 5, 160, 0, 0, 5, 70, 33, 69, 5, 160, 0, 0, 5, 71, 33, 69, 5, 160, 0, 0, 134, 5, 35, 79, 5, 160, 0, 0, 5, 5, 35, 79, 5, 160, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 193, 7, 111, 93, 5, 190, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 6, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 164, 0, 0, 193, 7, 111, 93, 5, 162, 0, 0, 193, 5, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 136, 0, 0, 193, 3, 111, 93, 5, 136, 0, 0, 134, 5, 35, 79, 5, 202, 0, 0, 5, 5, 35, 79, 5, 202, 0, 0, 193, 7, 111, 93, 5, 164, 0, 0, 134, 5, 31, 43, 5, 200, 0, 0, 5, 5, 31, 43, 5, 200, 0, 0, 193, 6, 111, 93, 5, 200, 0, 0, 193, 5, 111, 93, 5, 200, 0, 0, 134, 5, 21, 63, 5, 196, 0, 0, 5, 5, 21, 63, 5, 196, 0, 0, 134, 5, 21, 63, 5, 210, 0, 0, 5, 5, 21, 63, 5, 210, 0, 0, 134, 70, 33, 69, 5, 196, 0, 0, 134, 71, 33, 69, 5, 196, 0, 0, 5, 70, 33, 69, 5, 196, 0, 0, 5, 71, 33, 69, 5, 196, 0, 0, 134, 70, 33, 69, 5, 210, 0, 0, 134, 71, 33, 69, 5, 210, 0, 0, 5, 70, 33, 69, 5, 210, 0, 0, 5, 71, 33, 69, 5, 210, 0, 0, 134, 70, 33, 69, 5, 204, 0, 0, 134, 71, 33, 69, 5, 204, 0, 0, 5, 70, 33, 69, 5, 204, 0, 0, 5, 71, 33, 69, 5, 204, 0, 0, 134, 5, 35, 79, 5, 196, 0, 0, 5, 5, 35, 79, 5, 196, 0, 0, 134, 5, 35, 79, 5, 156, 0, 0, 5, 5, 35, 79, 5, 156, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 2, 111, 93, 5, 196, 0, 0, 193, 2, 111, 93, 5, 210, 0, 0, 134, 5, 31, 43, 5, 196, 0, 0, 5, 5, 31, 43, 5, 196, 0, 0, 193, 6, 111, 93, 5, 196, 0, 0, 193, 5, 111, 93, 5, 196, 0, 0, 193, 7, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 193, 7, 111, 93, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 5, 0, 8, 3, 9, 0, 8, 3, 0, 192, 198, 65, 10, 3, 0, 192, 198, 97, 65, 3, 31, 235, 134, 5, 1, 3, 31, 235, 134, 5, 3, 3, 198, 129, 7, 3, 198, 161, 0, 46, 156, 5, 48, 0, 12, 3, 21, 239, 134, 5, 0, 54, 156, 5, 48, 0, 12, 3, 27, 247, 134, 5, 0, 62, 156, 5, 48, 0, 12, 3, 21, 255, 134, 5, 0, 68, 156, 5, 2, 0, 1, 3, 255, 255, 33, 69, 134, 70, 3, 3, 255, 255, 33, 69, 134, 71, 65, 3, 255, 255, 33, 69, 134, 70, 0, 70, 156, 5, 4, 0, 2, 3, 0, 192, 198, 129, 3, 3, 0, 192, 198, 161, 8, 3, 0, 192, 198, 193, 11, 3, 0, 192, 198, 225, 40, 3, 1, 192, 198, 1, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 35, 79, 134, 5, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 92, 156, 5, 48, 0, 12, 3, 1, 192, 198, 33, 0, 42, 5, 5, 5, 0, 8, 3, 11, 0, 8, 3, 255, 255, 111, 93, 193, 6, 10, 3, 255, 255, 111, 93, 193, 5, 65, 3, 31, 235, 5, 5, 1, 3, 31, 235, 5, 5, 3, 3, 1, 192, 198, 65, 7, 3, 1, 192, 198, 97, 0, 46, 5, 5, 48, 0, 12, 3, 21, 239, 5, 5, 0, 54, 5, 5, 48, 0, 12, 3, 27, 247, 5, 5, 0, 62, 5, 5, 48, 0, 12, 3, 21, 255, 5, 5, 0, 68, 5, 5, 2, 0, 1, 3, 255, 255, 33, 69, 5, 70, 3, 3, 255, 255, 33, 69, 5, 71, 65, 3, 255, 255, 33, 69, 5, 70, 0, 70, 5, 5, 4, 0, 2, 3, 1, 192, 198, 129, 3, 3, 1, 192, 198, 161, 8, 3, 255, 255, 111, 93, 193, 7, 11, 3, 1, 192, 198, 193, 40, 3, 1, 192, 198, 225, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 35, 79, 5, 5, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 111, 93, 193, 2, 156, 46, 196, 160, 48, 0, 12, 3, 0, 192, 197, 162, 5, 46, 196, 160, 48, 0, 12, 3, 0, 192, 197, 226, 156, 42, 196, 162, 5, 0, 8, 3, 9, 0, 8, 3, 2, 192, 198, 194, 10, 3, 3, 192, 198, 2, 65, 3, 1, 192, 197, 34, 1, 3, 1, 192, 197, 34, 3, 3, 2, 192, 198, 66, 7, 3, 2, 192, 198, 130, 5, 42, 196, 162, 5, 0, 8, 3, 9, 0, 8, 3, 1, 192, 197, 162, 10, 3, 1, 192, 197, 226, 65, 3, 1, 192, 197, 98, 1, 3, 1, 192, 197, 98, 3, 3, 3, 192, 198, 66, 7, 3, 3, 192, 198, 130, 156, 54, 196, 160, 48, 0, 12, 3, 2, 192, 197, 34, 5, 54, 196, 160, 48, 0, 12, 3, 2, 192, 197, 98, 156, 62, 196, 160, 48, 0, 12, 3, 2, 192, 197, 162, 5, 62, 196, 160, 48, 0, 12, 3, 2, 192, 197, 226, 156, 68, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 3, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 5, 68, 196, 160, 2, 0, 1, 3, 3, 192, 197, 162, 3, 3, 3, 192, 197, 226, 65, 3, 3, 192, 197, 162, 156, 78, 196, 160, 48, 0, 12, 3, 4, 192, 197, 34, 5, 78, 196, 160, 48, 0, 12, 3, 4, 192, 197, 98, 156, 82, 196, 162, 1, 0, 8, 3, 4, 192, 197, 162, 11, 3, 4, 192, 197, 226, 5, 82, 196, 162, 1, 0, 8, 3, 5, 192, 197, 34, 11, 3, 5, 192, 197, 98, 156, 70, 196, 190, 3, 0, 2, 3, 4, 192, 198, 2, 3, 3, 4, 192, 198, 66, 8, 3, 4, 192, 198, 130, 11, 3, 4, 192, 198, 194, 5, 70, 196, 190, 3, 0, 2, 3, 5, 192, 198, 2, 3, 3, 5, 192, 198, 66, 8, 3, 5, 192, 197, 162, 11, 3, 5, 192, 198, 130, 156, 82, 196, 190, 1, 0, 8, 3, 5, 192, 197, 226, 11, 3, 6, 192, 197, 34, 5, 82, 196, 190, 1, 0, 8, 3, 6, 192, 197, 98, 11, 3, 6, 192, 197, 162, 22, 47, 134, 5, 48, 0, 12, 3, 23, 239, 134, 5, 1, 0, 198, 1, 3, 0, 2, 3, 6, 192, 198, 194, 3, 3, 7, 192, 198, 2, 8, 3, 7, 192, 198, 66, 11, 3, 7, 192, 198, 130, 1, 0, 198, 225, 3, 0, 2, 3, 7, 192, 198, 194, 3, 3, 8, 192, 198, 2, 8, 3, 8, 192, 197, 226, 11, 3, 8, 192, 198, 66, 156, 78, 196, 202, 48, 0, 12, 3, 9, 192, 197, 226, 5, 78, 196, 202, 48, 0, 12, 3, 10, 192, 197, 34, 22, 47, 5, 5, 48, 0, 12, 3, 23, 239, 5, 5, 156, 42, 196, 200, 5, 0, 8, 3, 9, 0, 8, 3, 11, 192, 198, 2, 10, 3, 11, 192, 198, 66, 65, 3, 10, 192, 197, 162, 1, 3, 10, 192, 197, 162, 3, 3, 10, 192, 198, 130, 7, 3, 10, 192, 198, 194, 5, 42, 196, 200, 5, 0, 8, 3, 9, 0, 8, 3, 11, 192, 197, 34, 10, 3, 11, 192, 197, 98, 65, 3, 10, 192, 197, 226, 1, 3, 10, 192, 197, 226, 3, 3, 11, 192, 198, 130, 7, 3, 11, 192, 198, 194, 156, 62, 196, 196, 48, 0, 12, 3, 11, 192, 197, 162, 5, 62, 196, 196, 48, 0, 12, 3, 11, 192, 197, 226, 156, 62, 196, 210, 48, 0, 12, 3, 12, 192, 197, 34, 5, 62, 196, 210, 48, 0, 12, 3, 12, 192, 197, 98, 156, 68, 196, 196, 2, 0, 1, 3, 12, 192, 197, 162, 3, 3, 12, 192, 197, 226, 65, 3, 12, 192, 197, 162, 5, 68, 196, 196, 2, 0, 1, 3, 13, 192, 197, 34, 3, 3, 13, 192, 197, 98, 65, 3, 13, 192, 197, 34, 156, 68, 196, 210, 2, 0, 1, 3, 13, 192, 197, 162, 3, 3, 13, 192, 197, 226, 65, 3, 13, 192, 197, 162, 5, 68, 196, 210, 2, 0, 1, 3, 14, 192, 197, 34, 3, 3, 14, 192, 197, 98, 65, 3, 14, 192, 197, 34, 156, 68, 196, 204, 2, 0, 1, 3, 14, 192, 197, 162, 3, 3, 14, 192, 197, 226, 65, 3, 14, 192, 197, 162, 5, 68, 196, 204, 2, 0, 1, 3, 15, 192, 197, 34, 3, 3, 15, 192, 197, 98, 65, 3, 15, 192, 197, 34, 156, 78, 196, 196, 48, 0, 12, 3, 15, 192, 197, 162, 5, 78, 196, 196, 48, 0, 12, 3, 15, 192, 197, 226, 156, 82, 196, 198, 1, 0, 8, 3, 16, 192, 197, 162, 11, 3, 16, 192, 197, 226, 5, 82, 196, 198, 1, 0, 8, 3, 17, 192, 197, 34, 11, 3, 17, 192, 197, 98, 156, 82, 196, 208, 1, 0, 8, 3, 17, 192, 197, 162, 11, 3, 17, 192, 197, 226, 5, 82, 196, 208, 1, 0, 8, 3, 18, 192, 197, 34, 11, 3, 18, 192, 197, 98, 156, 82, 196, 204, 1, 0, 8, 3, 18, 192, 197, 162, 11, 3, 18, 192, 197, 226, 5, 82, 196, 204, 1, 0, 8, 3, 19, 192, 197, 34, 11, 3, 19, 192, 197, 98, 156, 92, 196, 196, 48, 0, 12, 3, 13, 192, 198, 2, 5, 92, 196, 196, 48, 0, 12, 3, 19, 192, 197, 162, 156, 92, 196, 210, 48, 0, 12, 3, 13, 192, 198, 66, 5, 92, 196, 210, 48, 0, 12, 3, 19, 192, 197, 226, 156, 42, 196, 196, 5, 0, 8, 3, 9, 0, 8, 3, 14, 192, 198, 2, 10, 3, 14, 192, 198, 66, 65, 3, 20, 192, 197, 34, 1, 3, 20, 192, 197, 34, 3, 3, 13, 192, 198, 130, 7, 3, 13, 192, 198, 194, 5, 42, 196, 196, 5, 0, 8, 3, 9, 0, 8, 3, 20, 192, 197, 162, 10, 3, 20, 192, 197, 226, 65, 3, 20, 192, 197, 98, 1, 3, 20, 192, 197, 98, 3, 3, 14, 192, 198, 130, 7, 3, 14, 192, 198, 194, 156, 70, 196, 196, 3, 0, 2, 3, 15, 192, 198, 2, 3, 3, 15, 192, 198, 66, 8, 3, 15, 192, 198, 130, 11, 3, 15, 192, 198, 194, 5, 70, 196, 196, 3, 0, 2, 3, 16, 192, 198, 2, 3, 3, 16, 192, 198, 66, 8, 3, 21, 192, 197, 34, 11, 3, 16, 192, 198, 130, 21, 0, 197, 99, 3, 0, 2, 3, 18, 192, 198, 195, 3, 3, 19, 192, 198, 35, 8, 3, 19, 192, 198, 131, 11, 3, 19, 192, 198, 227, 21, 0, 197, 195, 3, 0, 2, 3, 20, 192, 198, 67, 3, 3, 20, 192, 198, 163, 8, 3, 22, 192, 197, 35, 11, 3, 21, 192, 198, 3, 156, 82, 196, 196, 1, 0, 8, 3, 22, 192, 197, 130, 11, 3, 22, 192, 197, 194, 5, 82, 196, 196, 1, 0, 8, 3, 23, 192, 197, 2, 11, 3, 23, 192, 197, 66, 23, 0, 197, 131, 1, 0, 8, 3, 23, 192, 197, 227, 11, 3, 24, 192, 197, 67, 24, 0, 197, 163, 1, 0, 8, 3, 25, 192, 197, 3, 11, 3, 25, 192, 197, 99, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 16, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 18, 4, 185, 64, 20, 4, 185, 76, 185, 84, 185, 88, 22, 4, 185, 96, 186, 100, 186, 108, 24, 4, 185, 116, 185, 120, 26, 4, 28, 4, 185, 136, 185, 140, 185, 148, 30, 4, 185, 160, 32, 4, 185, 172, 185, 176, 185, 180, 185, 184, 34, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 36, 4, 168, 64, 38, 4, 168, 76, 168, 84, 168, 88, 40, 4, 168, 96, 168, 100, 168, 108, 42, 4, 168, 116, 168, 120, 44, 4, 46, 4, 168, 136, 168, 140, 168, 148, 48, 4, 168, 160, 50, 4, 168, 172, 168, 176, 168, 180, 168, 184, 52, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 132, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 185, 60, 249, 57, 217, 212, 185, 212, 185, 208, 185, 204, 54, 4, 185, 85, 153, 85, 249, 85, 89, 86, 186, 101, 154, 101, 250, 101, 90, 102, 217, 80, 249, 128, 185, 133, 153, 133, 25, 217, 249, 216, 185, 216, 136, 14, 185, 200, 185, 169, 153, 169, 249, 169, 217, 184, 153, 185, 217, 164, 0, 8, 168, 57, 168, 60, 232, 57, 200, 212, 168, 212, 168, 208, 168, 204, 56, 4, 168, 85, 136, 85, 232, 85, 72, 86, 168, 101, 136, 101, 232, 101, 72, 102, 200, 80, 232, 128, 168, 133, 136, 133, 8, 217, 232, 216, 168, 216, 128, 14, 168, 200, 168, 169, 136, 169, 232, 169, 200, 184, 136, 185, 200, 164, 72, 186, 249, 58, 232, 58, 217, 57, 200, 57, 58, 4, 60, 4, 153, 69, 136, 69, 249, 69, 232, 69, 153, 70, 136, 70, 185, 72, 168, 72, 25, 78, 8, 78, 185, 80, 168, 80, 249, 86, 232, 86, 217, 85, 200, 85, 153, 86, 136, 86, 217, 86, 200, 86, 25, 86, 8, 86, 249, 93, 232, 93, 217, 93, 200, 93, 153, 94, 136, 94, 62, 4, 64, 4, 249, 97, 232, 97, 25, 99, 8, 99, 122, 102, 104, 102, 250, 102, 232, 102, 218, 101, 200, 101, 218, 102, 200, 102, 154, 102, 168, 104, 2, 8, 4, 8, 250, 109, 232, 109, 66, 4, 68, 4, 168, 144, 153, 117, 136, 117, 185, 118, 168, 118, 25, 118, 8, 118, 57, 119, 40, 119, 25, 119, 8, 119, 217, 128, 200, 128, 70, 4, 72, 4, 26, 126, 8, 126, 6, 8, 185, 128, 168, 128, 249, 134, 232, 134, 217, 133, 200, 133, 217, 216, 200, 216, 217, 200, 200, 200, 153, 149, 136, 149, 185, 150, 168, 150, 25, 150, 8, 150, 154, 153, 136, 153, 250, 153, 232, 153, 74, 4, 76, 4, 185, 156, 168, 156, 185, 162, 168, 162, 25, 162, 8, 162, 185, 164, 168, 164, 121, 170, 104, 170, 249, 170, 232, 170, 217, 169, 200, 169, 57, 170, 40, 170, 249, 184, 232, 184, 78, 4, 80, 4, 249, 177, 232, 177, 249, 185, 232, 185, 89, 186, 153, 189, 136, 189, 153, 190, 136, 190, 185, 196, 168, 196, 232, 152, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 8, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 10, 8, 1, 0, 8, 14, 12, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 14, 8, 200, 12, 72, 12, 32, 14, 200, 152, 169, 152, 187, 100, 187, 108, 169, 100, 169, 108, 169, 192, 169, 124, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 68, 255, 5, 185, 92, 255, 5, 185, 112, 255, 5, 186, 124, 255, 5, 185, 132, 255, 5, 186, 152, 255, 5, 185, 168, 255, 5, 185, 188, 255, 5, 168, 56, 255, 5, 168, 68, 255, 5, 168, 92, 255, 5, 168, 112, 255, 5, 168, 124, 255, 5, 168, 132, 255, 5, 168, 152, 255, 5, 168, 168, 255, 5, 168, 188, 255, 5, 185, 70, 255, 5, 168, 70, 255, 5, 217, 58, 255, 5, 200, 58, 255, 5, 185, 94, 255, 5, 168, 94, 255, 5, 185, 114, 255, 5, 168, 114, 255, 5, 186, 126, 255, 5, 168, 126, 255, 5, 186, 154, 255, 5, 168, 154, 255, 5, 217, 170, 255, 5, 200, 170, 255, 1, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 83, 15, 0, 16, 57, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 78, 15, 0, 32, 96, 28, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 101, 28, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 16, 0, 0, 0, 16, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 16, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 1, 0, 32, 0, 1, 0, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 65, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 83, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 83, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 120, 13, 0, 32, 7, 0, 0, 0, 24, 0, 0, 0, 124, 13, 0, 0, 124, 13, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 118, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 100, 105, 99, 116, 105, 111, 110, 97, 114, 121, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 87, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 93, 0, 38, 0, 150, 13, 60, 0, 130, 13, 60, 0, 131, 13, 38, 0, 165, 13, 60, 0, 164, 13, 0, 0, 90, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 93, 0, 38, 0, 150, 13, 60, 0, 130, 13, 60, 0, 131, 13, 38, 0, 162, 13, 202, 13, 164, 13, 60, 0, 60, 0, 165, 13, 0, 0, 170, 170, 170, 170, 170, 170, 232, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 112, 0, 0, 0, 112, 1, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 208, 25, 0, 0, 33, 0, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 113, 182, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 41, 42, 43, 44, 45, 46, 47, 48, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 120, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 103, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 64, 1, 0, 202, 97, 1, 0, 202, 130, 1, 0, 202, 163, 1, 0, 202, 196, 1, 0, 202, 229, 1, 0, 202, 6, 2, 0, 202, 39, 2, 0, 202, 72, 2, 0, 202, 105, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 130, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 194, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 2, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 2, 41, 113, 193, 3, 41, 113, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 58, 113, 5, 5, 158, 113, 5, 70, 62, 113, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 71, 0, 62, 0, 32, 0, 119, 0, 0, 16, 95, 0, 0, 96, 1, 0, 0, 96, 220, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 112, 0, 0, 0, 112, 1, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 196, 25, 0, 0, 33, 0, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 113, 182, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 41, 42, 43, 44, 45, 46, 47, 48, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 120, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 103, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 111, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 2, 41, 113, 193, 3, 41, 113, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 113, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 71, 0, 62, 0, 32, 0, 247, 6, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 86, 0, 170, 170, 242, 6, 0, 32, 111, 13, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 116, 13, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 25, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 4, 25, 0, 0, 4, 25, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 154, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 80, 220, 38, 0, 65, 0, 60, 0, 97, 0, 8, 3, 60, 0, 60, 0, 60, 0, 65, 0, 8, 3, 38, 0, 67, 0, 60, 0, 99, 0, 12, 3, 60, 0, 60, 0, 60, 0, 67, 0, 12, 3, 38, 0, 72, 0, 60, 0, 99, 0, 104, 0, 60, 0, 60, 0, 60, 0, 99, 0, 72, 0, 60, 0, 60, 0, 60, 0, 67, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 72, 0, 38, 0, 79, 0, 60, 0, 111, 0, 2, 3, 60, 0, 60, 0, 60, 0, 79, 0, 2, 3, 38, 0, 82, 0, 60, 0, 114, 0, 12, 3, 60, 0, 60, 0, 60, 0, 82, 0, 12, 3, 38, 0, 83, 0, 60, 0, 115, 0, 12, 3, 60, 0, 60, 0, 60, 0, 83, 0, 12, 3, 38, 0, 90, 0, 60, 0, 122, 0, 12, 3, 60, 0, 60, 0, 60, 0, 90, 0, 12, 3, 0, 0, 172, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 38, 0, 65, 0, 60, 0, 225, 0, 60, 0, 60, 0, 60, 0, 193, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 38, 0, 67, 0, 60, 0, 13, 1, 60, 0, 60, 0, 60, 0, 12, 1, 38, 0, 68, 0, 60, 0, 15, 1, 60, 0, 60, 0, 60, 0, 14, 1, 38, 0, 69, 0, 60, 0, 233, 0, 60, 0, 60, 0, 60, 0, 201, 0, 60, 0, 27, 1, 60, 0, 60, 0, 60, 0, 26, 1, 38, 0, 73, 0, 60, 0, 237, 0, 60, 0, 60, 0, 60, 0, 205, 0, 38, 0, 76, 0, 60, 0, 58, 1, 60, 0, 60, 0, 60, 0, 57, 1, 60, 0, 62, 1, 60, 0, 60, 0, 60, 0, 61, 1, 38, 0, 78, 0, 60, 0, 72, 1, 60, 0, 60, 0, 60, 0, 71, 1, 38, 0, 79, 0, 60, 0, 243, 0, 60, 0, 60, 0, 60, 0, 211, 0, 60, 0, 244, 0, 60, 0, 60, 0, 60, 0, 212, 0, 38, 0, 82, 0, 60, 0, 85, 1, 60, 0, 60, 0, 60, 0, 84, 1, 60, 0, 89, 1, 60, 0, 60, 0, 60, 0, 88, 1, 38, 0, 83, 0, 60, 0, 97, 1, 60, 0, 60, 0, 60, 0, 96, 1, 38, 0, 84, 0, 60, 0, 101, 1, 60, 0, 60, 0, 60, 0, 100, 1, 38, 0, 85, 0, 60, 0, 250, 0, 60, 0, 60, 0, 60, 0, 218, 0, 60, 0, 111, 1, 60, 0, 60, 0, 60, 0, 110, 1, 38, 0, 89, 0, 60, 0, 253, 0, 60, 0, 60, 0, 60, 0, 221, 0, 38, 0, 90, 0, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 0, 0, 170, 170, 170, 170, 48, 57, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 149, 1, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 212, 39, 0, 0, 212, 39, 0, 0, 212, 39, 0, 0, 212, 39, 0, 0, 92, 47, 0, 0, 92, 47, 0, 0, 24, 57, 0, 0, 24, 57, 0, 0, 24, 57, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 112, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 125, 0, 133, 0, 136, 0, 144, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 147, 0, 155, 0, 161, 0, 169, 0, 177, 0, 185, 0, 48, 0, 56, 0, 184, 0, 192, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 209, 0, 217, 0, 48, 0, 48, 0, 48, 0, 225, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 233, 0, 48, 0, 48, 0, 241, 0, 48, 0, 247, 0, 255, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 2, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 10, 1, 17, 1, 25, 1, 33, 1, 39, 1, 47, 1, 54, 1, 62, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 70, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 73, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 200, 0, 207, 0, 48, 0, 48, 0, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 103, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 81, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 89, 1, 48, 0, 48, 0, 95, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 1, 244, 1, 32, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 76, 2, 132, 2, 196, 2, 192, 0, 224, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 59, 0, 202, 33, 59, 0, 202, 66, 59, 0, 202, 99, 59, 0, 202, 132, 59, 0, 202, 165, 59, 0, 202, 198, 59, 0, 202, 231, 59, 0, 202, 8, 60, 0, 202, 41, 60, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 198, 0, 0, 156, 5, 0, 44, 201, 70, 2, 0, 201, 6, 3, 0, 201, 198, 3, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 70, 5, 0, 156, 5, 0, 60, 156, 5, 0, 62, 201, 102, 6, 0, 156, 5, 0, 66, 201, 70, 8, 0, 201, 38, 9, 0, 156, 5, 0, 72, 160, 5, 0, 74, 201, 6, 11, 0, 201, 230, 12, 0, 201, 198, 13, 0, 201, 166, 14, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 201, 134, 16, 0, 201, 230, 17, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 198, 18, 0, 5, 5, 0, 44, 201, 70, 20, 0, 201, 6, 21, 0, 201, 198, 21, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 70, 23, 0, 5, 5, 0, 60, 5, 5, 0, 62, 201, 102, 24, 0, 5, 5, 0, 66, 201, 70, 26, 0, 201, 38, 27, 0, 5, 5, 0, 72, 5, 5, 0, 74, 201, 6, 29, 0, 201, 230, 30, 0, 201, 198, 31, 0, 201, 166, 32, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 201, 134, 34, 0, 201, 230, 35, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 134, 5, 2, 43, 196, 142, 156, 42, 196, 154, 156, 42, 134, 5, 3, 43, 196, 146, 156, 42, 197, 226, 49, 0, 201, 198, 36, 0, 196, 138, 156, 50, 134, 5, 2, 51, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 134, 5, 2, 59, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 134, 5, 2, 71, 134, 5, 3, 71, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 134, 5, 2, 83, 196, 142, 156, 82, 196, 150, 156, 82, 134, 5, 2, 91, 160, 5, 38, 93, 197, 34, 50, 0, 196, 138, 5, 42, 5, 5, 2, 43, 196, 142, 5, 42, 196, 154, 5, 42, 5, 5, 3, 43, 196, 146, 5, 42, 197, 98, 50, 0, 201, 102, 37, 0, 196, 138, 5, 50, 5, 5, 2, 51, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 5, 5, 2, 59, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 5, 5, 2, 71, 5, 5, 3, 71, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 5, 5, 2, 83, 196, 142, 5, 82, 196, 150, 5, 82, 5, 5, 2, 91, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 38, 0, 201, 166, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 134, 5, 2, 49, 5, 5, 2, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 41, 0, 201, 166, 42, 0, 134, 5, 3, 51, 5, 5, 3, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 44, 0, 201, 70, 45, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 65, 5, 5, 2, 65, 201, 102, 46, 0, 201, 230, 47, 0, 134, 5, 3, 65, 5, 5, 3, 65, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 49, 0, 201, 38, 50, 0, 134, 5, 2, 69, 5, 5, 2, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 77, 5, 5, 2, 77, 201, 230, 50, 0, 201, 102, 52, 0, 134, 5, 3, 77, 5, 5, 3, 77, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 53, 0, 201, 166, 54, 0, 134, 5, 2, 79, 5, 5, 2, 79, 201, 102, 55, 0, 201, 38, 56, 0, 134, 5, 2, 81, 5, 5, 2, 81, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 83, 5, 5, 3, 83, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 56, 0, 201, 102, 58, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 59, 0, 201, 102, 61, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 62, 0, 201, 102, 64, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 9, 0, 197, 226, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 65, 0, 201, 102, 67, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 68, 0, 201, 166, 69, 0, 201, 102, 70, 0, 201, 38, 71, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 71, 0, 201, 102, 73, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 230, 74, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 76, 0, 5, 71, 238, 102, 201, 166, 77, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 13, 0, 192, 0, 0, 0, 197, 226, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 14, 0, 192, 0, 0, 0, 197, 162, 14, 0, 192, 0, 0, 0, 197, 226, 14, 0, 192, 0, 0, 0, 197, 34, 15, 0, 192, 0, 0, 0, 197, 98, 15, 0, 197, 163, 15, 0, 197, 2, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 16, 0, 197, 131, 16, 0, 197, 226, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 34, 13, 0, 197, 98, 17, 0, 5, 5, 10, 121, 197, 162, 17, 0, 197, 226, 17, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 34, 18, 0, 197, 98, 18, 0, 197, 162, 18, 0, 197, 226, 18, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 13, 0, 5, 5, 24, 121, 197, 226, 13, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 78, 0, 201, 6, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 81, 0, 201, 70, 82, 0, 201, 6, 83, 0, 201, 198, 83, 0, 201, 134, 84, 0, 201, 70, 85, 0, 201, 6, 86, 0, 201, 198, 86, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 87, 0, 201, 6, 89, 0, 201, 134, 90, 0, 201, 6, 92, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 93, 0, 201, 166, 94, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 95, 0, 201, 70, 97, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 98, 0, 201, 70, 100, 0, 201, 198, 101, 0, 201, 70, 103, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 104, 0, 201, 134, 105, 0, 201, 70, 106, 0, 201, 6, 107, 0, 201, 198, 107, 0, 201, 134, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 109, 0, 201, 198, 110, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 112, 0, 201, 198, 113, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 115, 0, 201, 6, 116, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 32, 0, 197, 162, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 116, 0, 201, 134, 117, 0, 201, 70, 118, 0, 201, 6, 119, 0, 201, 198, 119, 0, 201, 134, 120, 0, 201, 70, 121, 0, 201, 198, 122, 0, 201, 70, 124, 0, 201, 198, 125, 0, 201, 70, 127, 0, 201, 198, 128, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 130, 0, 201, 6, 131, 0, 201, 198, 131, 0, 201, 134, 132, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 133, 0, 201, 198, 134, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 136, 0, 201, 198, 137, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 139, 0, 201, 102, 140, 0, 201, 134, 141, 0, 201, 6, 143, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 41, 0, 197, 34, 42, 0, 197, 98, 42, 0, 197, 162, 42, 0, 197, 226, 42, 0, 197, 34, 43, 0, 197, 98, 43, 0, 197, 162, 43, 0, 197, 34, 41, 0, 197, 162, 41, 0, 197, 162, 7, 0, 197, 34, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 144, 0, 201, 6, 146, 0, 201, 134, 147, 0, 201, 6, 149, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 8, 0, 197, 34, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 150, 0, 201, 6, 152, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 153, 0, 201, 166, 154, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 43, 5, 162, 0, 0, 134, 5, 3, 43, 5, 162, 0, 0, 5, 5, 2, 43, 5, 162, 0, 0, 5, 5, 3, 43, 5, 162, 0, 0, 134, 5, 2, 51, 5, 162, 0, 0, 134, 5, 3, 51, 5, 162, 0, 0, 5, 5, 2, 51, 5, 162, 0, 0, 5, 5, 3, 51, 5, 162, 0, 0, 134, 5, 2, 59, 5, 162, 0, 0, 5, 5, 2, 59, 5, 162, 0, 0, 134, 5, 2, 65, 5, 160, 0, 0, 134, 5, 3, 65, 5, 160, 0, 0, 5, 5, 2, 65, 5, 160, 0, 0, 5, 5, 3, 65, 5, 160, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 77, 5, 160, 0, 0, 134, 5, 3, 77, 5, 160, 0, 0, 5, 5, 2, 77, 5, 160, 0, 0, 5, 5, 3, 77, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 81, 5, 160, 0, 0, 5, 5, 2, 81, 5, 160, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 134, 5, 3, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 5, 5, 3, 83, 5, 162, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 134, 5, 3, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 5, 3, 71, 5, 190, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 134, 5, 3, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 5, 3, 83, 5, 190, 0, 0, 134, 5, 3, 43, 5, 164, 0, 0, 5, 5, 3, 43, 5, 164, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 134, 5, 3, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 5, 5, 3, 71, 5, 162, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 81, 5, 202, 0, 0, 5, 5, 2, 81, 5, 202, 0, 0, 134, 5, 2, 51, 5, 160, 0, 0, 134, 5, 3, 51, 5, 160, 0, 0, 5, 5, 2, 51, 5, 160, 0, 0, 5, 5, 3, 51, 5, 160, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 5, 2, 43, 5, 200, 0, 0, 134, 5, 3, 43, 5, 200, 0, 0, 5, 5, 2, 43, 5, 200, 0, 0, 5, 5, 3, 43, 5, 200, 0, 0, 134, 5, 2, 49, 5, 196, 0, 0, 5, 5, 2, 49, 5, 196, 0, 0, 134, 5, 2, 49, 5, 210, 0, 0, 5, 5, 2, 49, 5, 210, 0, 0, 134, 5, 2, 49, 5, 160, 0, 0, 5, 5, 2, 49, 5, 160, 0, 0, 134, 5, 2, 49, 5, 204, 0, 0, 5, 5, 2, 49, 5, 204, 0, 0, 134, 5, 2, 51, 5, 204, 0, 0, 134, 5, 3, 51, 5, 204, 0, 0, 5, 5, 2, 51, 5, 204, 0, 0, 5, 5, 3, 51, 5, 204, 0, 0, 134, 5, 2, 51, 5, 208, 0, 0, 134, 5, 3, 51, 5, 208, 0, 0, 5, 5, 2, 51, 5, 208, 0, 0, 5, 5, 3, 51, 5, 208, 0, 0, 134, 5, 2, 59, 5, 208, 0, 0, 5, 5, 2, 59, 5, 208, 0, 0, 134, 5, 2, 65, 5, 196, 0, 0, 134, 5, 3, 65, 5, 196, 0, 0, 5, 5, 2, 65, 5, 196, 0, 0, 5, 5, 3, 65, 5, 196, 0, 0, 134, 5, 2, 65, 5, 210, 0, 0, 134, 5, 3, 65, 5, 210, 0, 0, 5, 5, 2, 65, 5, 210, 0, 0, 5, 5, 3, 65, 5, 210, 0, 0, 134, 5, 2, 65, 5, 204, 0, 0, 134, 5, 3, 65, 5, 204, 0, 0, 5, 5, 2, 65, 5, 204, 0, 0, 5, 5, 3, 65, 5, 204, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 134, 5, 2, 77, 5, 196, 0, 0, 134, 5, 3, 77, 5, 196, 0, 0, 5, 5, 2, 77, 5, 196, 0, 0, 5, 5, 3, 77, 5, 196, 0, 0, 134, 5, 2, 77, 5, 210, 0, 0, 134, 5, 3, 77, 5, 210, 0, 0, 5, 5, 2, 77, 5, 210, 0, 0, 5, 5, 3, 77, 5, 210, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 81, 5, 196, 0, 0, 5, 5, 2, 81, 5, 196, 0, 0, 134, 5, 2, 81, 5, 210, 0, 0, 5, 5, 2, 81, 5, 210, 0, 0, 134, 5, 2, 81, 5, 204, 0, 0, 5, 5, 2, 81, 5, 204, 0, 0, 134, 5, 2, 83, 5, 198, 0, 0, 134, 5, 3, 83, 5, 198, 0, 0, 5, 5, 2, 83, 5, 198, 0, 0, 5, 5, 3, 83, 5, 198, 0, 0, 134, 5, 2, 83, 5, 208, 0, 0, 134, 5, 3, 83, 5, 208, 0, 0, 5, 5, 2, 83, 5, 208, 0, 0, 5, 5, 3, 83, 5, 208, 0, 0, 134, 5, 2, 83, 5, 204, 0, 0, 134, 5, 3, 83, 5, 204, 0, 0, 5, 5, 2, 83, 5, 204, 0, 0, 5, 5, 3, 83, 5, 204, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 134, 5, 2, 43, 5, 196, 0, 0, 134, 5, 3, 43, 5, 196, 0, 0, 5, 5, 2, 43, 5, 196, 0, 0, 5, 5, 3, 43, 5, 196, 0, 0, 134, 5, 2, 51, 5, 196, 0, 0, 134, 5, 3, 51, 5, 196, 0, 0, 5, 5, 2, 51, 5, 196, 0, 0, 5, 5, 3, 51, 5, 196, 0, 0, 134, 5, 2, 59, 5, 196, 0, 0, 5, 5, 2, 59, 5, 196, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 134, 5, 3, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 5, 5, 3, 71, 5, 196, 0, 0, 134, 5, 3, 71, 5, 136, 0, 0, 5, 5, 3, 71, 5, 136, 0, 0, 134, 5, 3, 71, 5, 138, 0, 0, 5, 5, 3, 71, 5, 138, 0, 0, 134, 5, 3, 71, 5, 182, 0, 0, 5, 5, 3, 71, 5, 182, 0, 0, 134, 5, 3, 71, 5, 154, 0, 0, 5, 5, 3, 71, 5, 154, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 3, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 3, 71, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 196, 0, 0, 134, 5, 3, 83, 5, 196, 0, 0, 5, 5, 2, 83, 5, 196, 0, 0, 5, 5, 3, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 3, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 3, 83, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 91, 5, 196, 0, 0, 5, 5, 2, 91, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 34, 13, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 98, 13, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 162, 13, 0, 5, 5, 24, 121, 197, 226, 13, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 34, 14, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 98, 14, 0, 5, 5, 107, 121, 197, 162, 14, 0, 5, 5, 111, 121, 197, 226, 14, 0, 5, 5, 115, 121, 197, 34, 15, 0, 5, 5, 119, 121, 197, 98, 15, 0, 197, 163, 15, 0, 197, 2, 16, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 66, 16, 0, 197, 131, 16, 0, 197, 226, 16, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 34, 17, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 34, 13, 0, 197, 98, 17, 0, 5, 5, 10, 121, 197, 162, 17, 0, 197, 226, 17, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 34, 18, 0, 197, 98, 18, 0, 197, 162, 18, 0, 197, 226, 18, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 13, 0, 5, 5, 24, 121, 197, 226, 13, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 2, 0, 1, 3, 2, 235, 134, 5, 8, 3, 3, 235, 134, 5, 65, 3, 2, 235, 134, 5, 0, 46, 156, 5, 48, 0, 12, 3, 2, 239, 134, 5, 0, 48, 156, 5, 48, 0, 12, 3, 2, 241, 134, 5, 0, 50, 156, 5, 2, 0, 1, 3, 2, 243, 134, 5, 12, 3, 3, 243, 134, 5, 65, 3, 2, 243, 134, 5, 0, 58, 156, 5, 1, 0, 1, 3, 2, 251, 134, 5, 65, 3, 2, 251, 134, 5, 0, 64, 156, 5, 2, 0, 1, 3, 255, 255, 2, 65, 134, 5, 12, 3, 255, 255, 3, 65, 134, 5, 65, 3, 255, 255, 2, 65, 134, 5, 0, 68, 156, 5, 48, 0, 12, 3, 255, 255, 2, 69, 134, 5, 0, 70, 156, 5, 2, 0, 1, 3, 255, 255, 2, 71, 134, 5, 2, 3, 255, 255, 3, 71, 134, 5, 65, 3, 255, 255, 2, 71, 134, 5, 0, 76, 156, 5, 2, 0, 1, 3, 255, 255, 2, 77, 134, 5, 12, 3, 255, 255, 3, 77, 134, 5, 65, 3, 255, 255, 2, 77, 134, 5, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 80, 156, 5, 48, 0, 12, 3, 255, 255, 2, 81, 134, 5, 0, 82, 156, 5, 2, 0, 1, 3, 255, 255, 2, 83, 134, 5, 10, 3, 255, 255, 3, 83, 134, 5, 65, 3, 255, 255, 2, 83, 134, 5, 0, 90, 156, 5, 1, 0, 1, 3, 255, 255, 2, 91, 134, 5, 65, 3, 255, 255, 2, 91, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 42, 5, 5, 2, 0, 1, 3, 2, 235, 5, 5, 8, 3, 3, 235, 5, 5, 65, 3, 2, 235, 5, 5, 0, 46, 5, 5, 48, 0, 12, 3, 2, 239, 5, 5, 0, 48, 5, 5, 48, 0, 12, 3, 2, 241, 5, 5, 0, 50, 5, 5, 2, 0, 1, 3, 2, 243, 5, 5, 12, 3, 3, 243, 5, 5, 65, 3, 2, 243, 5, 5, 0, 58, 5, 5, 1, 0, 1, 3, 2, 251, 5, 5, 65, 3, 2, 251, 5, 5, 0, 64, 5, 5, 2, 0, 1, 3, 255, 255, 2, 65, 5, 5, 12, 3, 255, 255, 3, 65, 5, 5, 65, 3, 255, 255, 2, 65, 5, 5, 0, 68, 5, 5, 48, 0, 12, 3, 255, 255, 2, 69, 5, 5, 0, 70, 5, 5, 2, 0, 1, 3, 255, 255, 2, 71, 5, 5, 2, 3, 255, 255, 3, 71, 5, 5, 65, 3, 255, 255, 2, 71, 5, 5, 0, 76, 5, 5, 2, 0, 1, 3, 255, 255, 2, 77, 5, 5, 12, 3, 255, 255, 3, 77, 5, 5, 65, 3, 255, 255, 2, 77, 5, 5, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 80, 5, 5, 48, 0, 12, 3, 255, 255, 2, 81, 5, 5, 0, 82, 5, 5, 2, 0, 1, 3, 255, 255, 2, 83, 5, 5, 10, 3, 255, 255, 3, 83, 5, 5, 65, 3, 255, 255, 2, 83, 5, 5, 0, 90, 5, 5, 1, 0, 1, 3, 255, 255, 2, 91, 5, 5, 65, 3, 255, 255, 2, 91, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 48, 0, 12, 3, 197, 162, 5, 46, 196, 160, 48, 0, 12, 3, 0, 192, 197, 98, 156, 42, 196, 162, 2, 0, 1, 3, 0, 192, 197, 162, 8, 3, 0, 192, 197, 226, 65, 3, 0, 192, 197, 162, 5, 42, 196, 162, 2, 0, 1, 3, 1, 192, 197, 34, 8, 3, 1, 192, 197, 98, 65, 3, 1, 192, 197, 34, 156, 50, 196, 162, 2, 0, 1, 3, 1, 192, 197, 162, 12, 3, 1, 192, 197, 226, 65, 3, 1, 192, 197, 162, 5, 50, 196, 162, 2, 0, 1, 3, 2, 192, 197, 34, 12, 3, 2, 192, 197, 98, 65, 3, 2, 192, 197, 34, 156, 58, 196, 162, 1, 0, 1, 3, 2, 192, 197, 162, 65, 3, 2, 192, 197, 162, 5, 58, 196, 162, 1, 0, 1, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 226, 156, 64, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 12, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 5, 64, 196, 160, 2, 0, 1, 3, 3, 192, 197, 162, 12, 3, 3, 192, 197, 226, 65, 3, 3, 192, 197, 162, 156, 68, 196, 160, 48, 0, 12, 3, 4, 192, 197, 34, 5, 68, 196, 160, 48, 0, 12, 3, 4, 192, 197, 98, 156, 76, 196, 160, 2, 0, 1, 3, 4, 192, 197, 162, 12, 3, 4, 192, 197, 226, 65, 3, 4, 192, 197, 162, 5, 76, 196, 160, 2, 0, 1, 3, 5, 192, 197, 34, 12, 3, 5, 192, 197, 98, 65, 3, 5, 192, 197, 34, 156, 78, 196, 160, 48, 0, 12, 3, 5, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 5, 192, 197, 226, 156, 80, 196, 160, 48, 0, 12, 3, 6, 192, 197, 34, 5, 80, 196, 160, 48, 0, 12, 3, 6, 192, 197, 98, 156, 82, 196, 162, 2, 0, 1, 3, 6, 192, 197, 162, 10, 3, 6, 192, 197, 226, 65, 3, 6, 192, 197, 162, 5, 82, 196, 162, 2, 0, 1, 3, 7, 192, 197, 34, 10, 3, 7, 192, 197, 98, 65, 3, 7, 192, 197, 34, 156, 70, 196, 190, 2, 0, 1, 3, 7, 192, 197, 162, 2, 3, 7, 192, 197, 226, 65, 3, 7, 192, 197, 162, 5, 70, 196, 190, 2, 0, 1, 3, 8, 192, 197, 34, 2, 3, 8, 192, 197, 98, 65, 3, 8, 192, 197, 34, 156, 82, 196, 190, 2, 0, 1, 3, 8, 192, 197, 162, 10, 3, 8, 192, 197, 226, 65, 3, 8, 192, 197, 162, 5, 82, 196, 190, 2, 0, 1, 3, 9, 192, 197, 34, 10, 3, 9, 192, 197, 98, 65, 3, 9, 192, 197, 34, 156, 70, 196, 162, 2, 0, 1, 3, 10, 192, 197, 34, 2, 3, 10, 192, 197, 98, 65, 3, 10, 192, 197, 34, 5, 70, 196, 162, 2, 0, 1, 3, 10, 192, 197, 162, 2, 3, 10, 192, 197, 226, 65, 3, 10, 192, 197, 162, 156, 78, 196, 202, 48, 0, 12, 3, 11, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 11, 192, 197, 98, 156, 80, 196, 202, 48, 0, 12, 3, 11, 192, 197, 162, 5, 80, 196, 202, 48, 0, 12, 3, 11, 192, 197, 226, 156, 50, 196, 160, 2, 0, 1, 3, 12, 192, 197, 34, 12, 3, 12, 192, 197, 98, 65, 3, 12, 192, 197, 34, 5, 50, 196, 160, 2, 0, 1, 3, 12, 192, 197, 162, 12, 3, 12, 192, 197, 226, 65, 3, 12, 192, 197, 162, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 2, 0, 1, 3, 19, 192, 197, 226, 8, 3, 20, 192, 197, 34, 65, 3, 19, 192, 197, 226, 5, 42, 196, 200, 2, 0, 1, 3, 20, 192, 197, 98, 8, 3, 20, 192, 197, 162, 65, 3, 20, 192, 197, 98, 156, 48, 196, 196, 48, 0, 12, 3, 20, 192, 197, 226, 5, 48, 196, 196, 48, 0, 12, 3, 21, 192, 197, 34, 156, 48, 196, 210, 48, 0, 12, 3, 21, 192, 197, 98, 5, 48, 196, 210, 48, 0, 12, 3, 21, 192, 197, 162, 156, 48, 196, 160, 48, 0, 12, 3, 21, 192, 197, 226, 5, 48, 196, 160, 48, 0, 12, 3, 22, 192, 197, 34, 156, 48, 196, 204, 48, 0, 12, 3, 22, 192, 197, 98, 5, 48, 196, 204, 48, 0, 12, 3, 22, 192, 197, 162, 156, 50, 196, 204, 2, 0, 1, 3, 22, 192, 197, 226, 12, 3, 23, 192, 197, 34, 65, 3, 22, 192, 197, 226, 5, 50, 196, 204, 2, 0, 1, 3, 23, 192, 197, 98, 12, 3, 23, 192, 197, 162, 65, 3, 23, 192, 197, 98, 156, 50, 196, 208, 2, 0, 1, 3, 23, 192, 197, 226, 12, 3, 24, 192, 197, 34, 65, 3, 23, 192, 197, 226, 5, 50, 196, 208, 2, 0, 1, 3, 24, 192, 197, 98, 12, 3, 24, 192, 197, 162, 65, 3, 24, 192, 197, 98, 156, 58, 196, 208, 1, 0, 1, 3, 24, 192, 197, 226, 65, 3, 24, 192, 197, 226, 5, 58, 196, 208, 1, 0, 1, 3, 25, 192, 197, 34, 65, 3, 25, 192, 197, 34, 156, 64, 196, 196, 2, 0, 1, 3, 25, 192, 197, 98, 12, 3, 25, 192, 197, 162, 65, 3, 25, 192, 197, 98, 5, 64, 196, 196, 2, 0, 1, 3, 25, 192, 197, 226, 12, 3, 26, 192, 197, 34, 65, 3, 25, 192, 197, 226, 156, 64, 196, 210, 2, 0, 1, 3, 26, 192, 197, 98, 12, 3, 26, 192, 197, 162, 65, 3, 26, 192, 197, 98, 5, 64, 196, 210, 2, 0, 1, 3, 26, 192, 197, 226, 12, 3, 27, 192, 197, 34, 65, 3, 26, 192, 197, 226, 156, 64, 196, 204, 2, 0, 1, 3, 27, 192, 197, 98, 12, 3, 27, 192, 197, 162, 65, 3, 27, 192, 197, 98, 5, 64, 196, 204, 2, 0, 1, 3, 27, 192, 197, 226, 12, 3, 28, 192, 197, 34, 65, 3, 27, 192, 197, 226, 156, 68, 196, 196, 48, 0, 12, 3, 28, 192, 197, 98, 5, 68, 196, 196, 48, 0, 12, 3, 28, 192, 197, 162, 156, 68, 196, 210, 48, 0, 12, 3, 28, 192, 197, 226, 5, 68, 196, 210, 48, 0, 12, 3, 29, 192, 197, 34, 156, 68, 196, 204, 48, 0, 12, 3, 29, 192, 197, 98, 5, 68, 196, 204, 48, 0, 12, 3, 29, 192, 197, 162, 156, 76, 196, 196, 2, 0, 1, 3, 29, 192, 197, 226, 12, 3, 30, 192, 197, 34, 65, 3, 29, 192, 197, 226, 5, 76, 196, 196, 2, 0, 1, 3, 30, 192, 197, 98, 12, 3, 30, 192, 197, 162, 65, 3, 30, 192, 197, 98, 156, 76, 196, 210, 2, 0, 1, 3, 30, 192, 197, 226, 12, 3, 31, 192, 197, 34, 65, 3, 30, 192, 197, 226, 5, 76, 196, 210, 2, 0, 1, 3, 31, 192, 197, 98, 12, 3, 31, 192, 197, 162, 65, 3, 31, 192, 197, 98, 156, 78, 196, 196, 48, 0, 12, 3, 31, 192, 197, 226, 5, 78, 196, 196, 48, 0, 12, 3, 32, 192, 197, 34, 156, 80, 196, 196, 48, 0, 12, 3, 32, 192, 197, 226, 5, 80, 196, 196, 48, 0, 12, 3, 33, 192, 197, 34, 156, 80, 196, 210, 48, 0, 12, 3, 33, 192, 197, 98, 5, 80, 196, 210, 48, 0, 12, 3, 33, 192, 197, 162, 156, 80, 196, 204, 48, 0, 12, 3, 33, 192, 197, 226, 5, 80, 196, 204, 48, 0, 12, 3, 34, 192, 197, 34, 156, 82, 196, 198, 2, 0, 1, 3, 34, 192, 197, 98, 10, 3, 34, 192, 197, 162, 65, 3, 34, 192, 197, 98, 5, 82, 196, 198, 2, 0, 1, 3, 34, 192, 197, 226, 10, 3, 35, 192, 197, 34, 65, 3, 34, 192, 197, 226, 156, 82, 196, 208, 2, 0, 1, 3, 35, 192, 197, 98, 10, 3, 35, 192, 197, 162, 65, 3, 35, 192, 197, 98, 5, 82, 196, 208, 2, 0, 1, 3, 35, 192, 197, 226, 10, 3, 36, 192, 197, 34, 65, 3, 35, 192, 197, 226, 156, 82, 196, 204, 2, 0, 1, 3, 36, 192, 197, 98, 10, 3, 36, 192, 197, 162, 65, 3, 36, 192, 197, 98, 5, 82, 196, 204, 2, 0, 1, 3, 36, 192, 197, 226, 10, 3, 37, 192, 197, 34, 65, 3, 36, 192, 197, 226, 156, 92, 196, 196, 48, 0, 12, 3, 37, 192, 197, 98, 5, 92, 196, 196, 48, 0, 12, 3, 37, 192, 197, 162, 156, 92, 196, 210, 48, 0, 12, 3, 37, 192, 197, 226, 5, 92, 196, 210, 48, 0, 12, 3, 38, 192, 197, 34, 156, 42, 196, 196, 2, 0, 1, 3, 38, 192, 197, 98, 8, 3, 38, 192, 197, 162, 65, 3, 38, 192, 197, 98, 5, 42, 196, 196, 2, 0, 1, 3, 38, 192, 197, 226, 8, 3, 39, 192, 197, 34, 65, 3, 38, 192, 197, 226, 156, 50, 196, 196, 2, 0, 1, 3, 39, 192, 197, 98, 12, 3, 39, 192, 197, 162, 65, 3, 39, 192, 197, 98, 5, 50, 196, 196, 2, 0, 1, 3, 39, 192, 197, 226, 12, 3, 40, 192, 197, 34, 65, 3, 39, 192, 197, 226, 156, 58, 196, 196, 1, 0, 1, 3, 40, 192, 197, 98, 65, 3, 40, 192, 197, 98, 5, 58, 196, 196, 1, 0, 1, 3, 40, 192, 197, 162, 65, 3, 40, 192, 197, 162, 156, 70, 196, 196, 2, 0, 1, 3, 40, 192, 197, 226, 2, 3, 41, 192, 197, 34, 65, 3, 40, 192, 197, 226, 5, 70, 196, 196, 2, 0, 1, 3, 41, 192, 197, 98, 2, 3, 41, 192, 197, 162, 65, 3, 41, 192, 197, 98, 43, 0, 197, 227, 2, 0, 1, 3, 44, 192, 197, 67, 2, 3, 44, 192, 197, 163, 65, 3, 44, 192, 197, 67, 45, 0, 197, 3, 2, 0, 1, 3, 45, 192, 197, 99, 2, 3, 45, 192, 197, 195, 65, 3, 45, 192, 197, 99, 156, 82, 196, 196, 2, 0, 1, 3, 46, 192, 197, 34, 10, 3, 46, 192, 197, 98, 65, 3, 46, 192, 197, 34, 5, 82, 196, 196, 2, 0, 1, 3, 46, 192, 197, 162, 10, 3, 46, 192, 197, 226, 65, 3, 46, 192, 197, 162, 47, 0, 197, 35, 2, 0, 1, 3, 47, 192, 197, 131, 10, 3, 47, 192, 197, 227, 65, 3, 47, 192, 197, 131, 48, 0, 197, 67, 2, 0, 1, 3, 48, 192, 197, 163, 10, 3, 49, 192, 197, 3, 65, 3, 48, 192, 197, 163, 156, 90, 196, 196, 1, 0, 1, 3, 49, 192, 197, 98, 65, 3, 49, 192, 197, 98, 5, 90, 196, 196, 1, 0, 1, 3, 49, 192, 197, 162, 65, 3, 49, 192, 197, 162, 3, 0, 67, 0, 58, 0, 32, 0, 155, 0, 0, 16, 88, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 28, 40, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 4, 30, 0, 0, 4, 30, 0, 0, 4, 30, 0, 0, 4, 30, 0, 0, 184, 32, 0, 0, 184, 32, 0, 0, 186, 35, 0, 0, 212, 35, 0, 0, 4, 40, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 211, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 91, 0, 99, 0, 107, 0, 115, 0, 116, 0, 124, 0, 127, 0, 135, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 159, 0, 48, 0, 161, 0, 169, 0, 173, 0, 181, 0, 186, 0, 194, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 202, 0, 151, 0, 48, 0, 48, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 143, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 108, 1, 172, 1, 208, 1, 252, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 20, 0, 202, 129, 20, 0, 202, 162, 20, 0, 202, 195, 20, 0, 202, 228, 20, 0, 202, 5, 21, 0, 202, 38, 21, 0, 202, 71, 21, 0, 202, 104, 21, 0, 202, 137, 21, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 201, 196, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 134, 8, 0, 156, 5, 0, 72, 160, 5, 0, 74, 201, 102, 9, 0, 201, 70, 10, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 38, 11, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 6, 12, 0, 5, 5, 0, 44, 201, 196, 12, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 166, 20, 0, 5, 5, 0, 72, 5, 5, 0, 74, 201, 134, 21, 0, 201, 102, 22, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 70, 23, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 134, 5, 2, 43, 196, 146, 156, 42, 197, 162, 19, 0, 201, 38, 24, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 134, 5, 2, 71, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 19, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 5, 5, 2, 43, 196, 146, 5, 42, 197, 34, 20, 0, 201, 230, 24, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 5, 5, 2, 71, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 25, 0, 201, 102, 26, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 27, 0, 201, 230, 27, 0, 134, 5, 2, 77, 5, 5, 2, 77, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 28, 0, 201, 102, 29, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 30, 0, 201, 230, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 10, 0, 197, 98, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 31, 0, 201, 102, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 33, 0, 201, 230, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 34, 0, 201, 102, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 36, 0, 201, 230, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 37, 0, 201, 102, 38, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 39, 0, 201, 230, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 13, 0, 197, 226, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 40, 0, 201, 102, 41, 0, 201, 38, 42, 0, 201, 230, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 43, 0, 201, 102, 44, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 45, 0, 201, 230, 45, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 16, 0, 197, 98, 16, 0, 197, 162, 16, 0, 197, 226, 16, 0, 197, 34, 17, 0, 197, 98, 17, 0, 197, 162, 17, 0, 197, 226, 17, 0, 197, 162, 15, 0, 197, 226, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 46, 0, 201, 102, 47, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 136, 5, 2, 57, 5, 142, 0, 0, 71, 5, 2, 57, 5, 142, 0, 0, 136, 5, 2, 57, 5, 144, 0, 0, 71, 5, 2, 57, 5, 144, 0, 0, 136, 5, 2, 57, 5, 156, 0, 0, 71, 5, 2, 57, 5, 156, 0, 0, 136, 5, 2, 57, 5, 196, 0, 0, 71, 5, 2, 57, 5, 196, 0, 0, 136, 5, 2, 57, 5, 150, 0, 0, 71, 5, 2, 57, 5, 150, 0, 0, 136, 5, 2, 57, 5, 160, 0, 0, 71, 5, 2, 57, 5, 160, 0, 0, 136, 5, 2, 57, 5, 206, 0, 0, 71, 5, 2, 57, 5, 206, 0, 0, 71, 5, 2, 57, 5, 210, 0, 0, 70, 5, 2, 57, 5, 142, 0, 0, 5, 5, 2, 57, 5, 142, 0, 0, 70, 5, 2, 57, 5, 144, 0, 0, 5, 5, 2, 57, 5, 144, 0, 0, 70, 5, 2, 57, 5, 156, 0, 0, 5, 5, 2, 57, 5, 156, 0, 0, 70, 5, 2, 57, 5, 196, 0, 0, 5, 5, 2, 57, 5, 196, 0, 0, 70, 5, 2, 57, 5, 150, 0, 0, 5, 5, 2, 57, 5, 150, 0, 0, 70, 5, 2, 57, 5, 160, 0, 0, 5, 5, 2, 57, 5, 160, 0, 0, 70, 5, 2, 57, 5, 206, 0, 0, 5, 5, 2, 57, 5, 206, 0, 0, 5, 5, 2, 57, 5, 210, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 43, 5, 162, 0, 0, 5, 5, 2, 43, 5, 162, 0, 0, 134, 5, 2, 77, 5, 160, 0, 0, 5, 5, 2, 77, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 43, 5, 164, 0, 0, 5, 5, 2, 43, 5, 164, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 43, 5, 200, 0, 0, 5, 5, 2, 43, 5, 200, 0, 0, 134, 5, 2, 77, 5, 196, 0, 0, 5, 5, 2, 77, 5, 196, 0, 0, 134, 5, 2, 77, 5, 210, 0, 0, 5, 5, 2, 77, 5, 210, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 134, 5, 2, 43, 5, 196, 0, 0, 5, 5, 2, 43, 5, 196, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 134, 5, 2, 71, 5, 136, 0, 0, 5, 5, 2, 71, 5, 136, 0, 0, 134, 5, 2, 71, 5, 138, 0, 0, 5, 5, 2, 71, 5, 138, 0, 0, 134, 5, 2, 71, 5, 182, 0, 0, 5, 5, 2, 71, 5, 182, 0, 0, 134, 5, 2, 71, 5, 154, 0, 0, 5, 5, 2, 71, 5, 154, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 48, 0, 8, 3, 2, 235, 134, 5, 0, 46, 156, 5, 17, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 2, 192, 197, 162, 41, 30, 2, 192, 197, 226, 42, 30, 3, 192, 197, 34, 43, 30, 3, 192, 197, 98, 150, 30, 3, 192, 197, 162, 36, 30, 1, 192, 197, 162, 37, 30, 1, 192, 197, 226, 38, 30, 2, 192, 197, 34, 39, 30, 2, 192, 197, 98, 30, 2, 15, 0, 30, 2, 0, 192, 197, 162, 31, 2, 0, 192, 197, 226, 12, 3, 2, 239, 134, 5, 34, 30, 1, 192, 197, 34, 35, 30, 1, 192, 197, 98, 72, 0, 2, 249, 136, 5, 104, 0, 2, 249, 71, 5, 36, 1, 197, 162, 37, 1, 0, 192, 197, 98, 0, 70, 156, 5, 48, 0, 2, 3, 255, 255, 2, 71, 134, 5, 0, 76, 156, 5, 48, 0, 12, 3, 255, 255, 2, 77, 134, 5, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 42, 5, 5, 48, 0, 8, 3, 2, 235, 5, 5, 0, 46, 5, 5, 17, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 6, 192, 197, 98, 41, 30, 6, 192, 197, 162, 42, 30, 6, 192, 197, 226, 43, 30, 7, 192, 197, 34, 150, 30, 7, 192, 197, 98, 36, 30, 5, 192, 197, 98, 37, 30, 5, 192, 197, 162, 38, 30, 5, 192, 197, 226, 39, 30, 6, 192, 197, 34, 30, 2, 15, 0, 30, 2, 4, 192, 197, 98, 31, 2, 4, 192, 197, 162, 12, 3, 2, 239, 5, 5, 34, 30, 4, 192, 197, 226, 35, 30, 5, 192, 197, 34, 72, 0, 2, 249, 70, 5, 104, 0, 2, 249, 5, 5, 36, 1, 3, 192, 197, 226, 37, 1, 4, 192, 197, 34, 0, 70, 5, 5, 48, 0, 2, 3, 255, 255, 2, 71, 5, 5, 0, 76, 5, 5, 48, 0, 12, 3, 255, 255, 2, 77, 5, 5, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 48, 0, 12, 3, 7, 192, 197, 162, 5, 46, 196, 160, 48, 0, 12, 3, 7, 192, 197, 226, 156, 42, 196, 162, 48, 0, 8, 3, 8, 192, 197, 34, 5, 42, 196, 162, 48, 0, 8, 3, 8, 192, 197, 98, 156, 76, 196, 160, 48, 0, 12, 3, 8, 192, 197, 162, 5, 76, 196, 160, 48, 0, 12, 3, 8, 192, 197, 226, 156, 78, 196, 160, 48, 0, 12, 3, 9, 192, 197, 34, 5, 78, 196, 160, 48, 0, 12, 3, 9, 192, 197, 98, 156, 70, 196, 190, 48, 0, 2, 3, 9, 192, 197, 162, 5, 70, 196, 190, 48, 0, 2, 3, 9, 192, 197, 226, 156, 70, 196, 162, 48, 0, 2, 3, 10, 192, 197, 162, 5, 70, 196, 162, 48, 0, 2, 3, 10, 192, 197, 226, 156, 78, 196, 202, 48, 0, 12, 3, 11, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 11, 192, 197, 98, 156, 42, 196, 200, 48, 0, 8, 3, 11, 192, 197, 162, 5, 42, 196, 200, 48, 0, 8, 3, 11, 192, 197, 226, 156, 76, 196, 196, 48, 0, 12, 3, 12, 192, 197, 34, 5, 76, 196, 196, 48, 0, 12, 3, 12, 192, 197, 98, 156, 76, 196, 210, 48, 0, 12, 3, 12, 192, 197, 162, 5, 76, 196, 210, 48, 0, 12, 3, 12, 192, 197, 226, 156, 78, 196, 196, 48, 0, 12, 3, 13, 192, 197, 34, 5, 78, 196, 196, 48, 0, 12, 3, 13, 192, 197, 98, 156, 92, 196, 196, 48, 0, 12, 3, 14, 192, 197, 34, 5, 92, 196, 196, 48, 0, 12, 3, 14, 192, 197, 98, 156, 92, 196, 210, 48, 0, 12, 3, 14, 192, 197, 162, 5, 92, 196, 210, 48, 0, 12, 3, 14, 192, 197, 226, 156, 42, 196, 196, 48, 0, 8, 3, 15, 192, 197, 34, 5, 42, 196, 196, 48, 0, 8, 3, 15, 192, 197, 98, 156, 70, 196, 196, 48, 0, 2, 3, 15, 192, 197, 162, 5, 70, 196, 196, 48, 0, 2, 3, 15, 192, 197, 226, 18, 0, 197, 35, 48, 0, 2, 3, 18, 192, 197, 131, 18, 0, 197, 227, 48, 0, 2, 3, 19, 192, 197, 67, 12, 0, 72, 0, 73, 0, 104, 0, 105, 0, 36, 1, 38, 1, 30, 2, 32, 2, 34, 30, 44, 30, 150, 30, 151, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 64, 28, 4, 185, 76, 186, 80, 185, 84, 185, 88, 185, 92, 186, 100, 186, 108, 185, 112, 185, 116, 185, 120, 186, 124, 38, 4, 185, 140, 185, 144, 40, 4, 42, 4, 185, 168, 185, 176, 185, 180, 185, 184, 185, 188, 185, 192, 44, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 46, 4, 168, 64, 48, 4, 168, 76, 168, 80, 168, 84, 168, 88, 168, 92, 168, 100, 168, 108, 168, 112, 168, 116, 168, 120, 168, 124, 58, 4, 168, 140, 168, 144, 60, 4, 62, 4, 168, 168, 168, 176, 168, 180, 168, 184, 168, 188, 168, 192, 64, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 132, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 185, 60, 57, 58, 0, 8, 66, 4, 186, 81, 154, 81, 250, 81, 90, 82, 186, 101, 154, 101, 250, 101, 90, 102, 217, 76, 154, 126, 185, 133, 153, 133, 185, 136, 153, 134, 89, 134, 136, 14, 217, 134, 185, 177, 153, 177, 249, 177, 89, 178, 153, 193, 185, 204, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 168, 60, 40, 58, 4, 8, 68, 4, 168, 81, 136, 81, 232, 81, 72, 82, 168, 101, 136, 101, 232, 101, 72, 102, 200, 76, 136, 126, 168, 133, 136, 133, 168, 136, 136, 134, 72, 134, 128, 14, 200, 134, 168, 177, 136, 177, 232, 177, 72, 178, 136, 193, 168, 204, 72, 194, 57, 59, 40, 59, 217, 57, 200, 57, 70, 4, 72, 4, 153, 69, 136, 69, 249, 69, 232, 69, 185, 70, 168, 70, 185, 72, 168, 72, 25, 78, 8, 78, 89, 79, 72, 79, 58, 83, 40, 83, 218, 81, 200, 81, 186, 82, 168, 82, 26, 83, 8, 83, 26, 82, 8, 82, 249, 89, 232, 89, 217, 89, 200, 89, 185, 90, 168, 90, 249, 90, 232, 90, 249, 93, 232, 93, 89, 95, 72, 95, 154, 102, 136, 102, 58, 103, 40, 103, 218, 101, 200, 101, 26, 103, 8, 103, 186, 102, 168, 104, 6, 8, 8, 8, 250, 109, 232, 109, 249, 114, 232, 114, 168, 148, 153, 117, 136, 117, 249, 118, 232, 118, 25, 118, 8, 118, 121, 119, 104, 119, 89, 119, 72, 119, 154, 125, 136, 125, 250, 126, 232, 126, 26, 126, 8, 126, 10, 8, 185, 128, 168, 128, 57, 135, 40, 135, 217, 133, 200, 133, 121, 134, 104, 134, 12, 8, 14, 8, 153, 153, 136, 153, 74, 4, 76, 4, 185, 156, 168, 156, 154, 161, 136, 161, 250, 161, 232, 161, 78, 4, 80, 4, 185, 164, 168, 164, 249, 170, 232, 170, 25, 170, 8, 170, 185, 172, 168, 172, 153, 178, 136, 178, 57, 179, 40, 179, 217, 177, 200, 177, 57, 178, 40, 178, 121, 178, 104, 178, 25, 179, 8, 179, 249, 185, 232, 185, 249, 193, 232, 193, 89, 194, 153, 197, 136, 197, 185, 198, 168, 198, 185, 200, 168, 200, 232, 160, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 80, 200, 160, 169, 160, 200, 56, 169, 80, 187, 100, 187, 108, 169, 100, 169, 108, 169, 208, 169, 124, 217, 132, 187, 80, 200, 132, 169, 80, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 68, 72, 4, 187, 96, 104, 4, 178, 96, 36, 5, 251, 97, 37, 5, 242, 97, 255, 5, 185, 132, 255, 5, 185, 152, 255, 5, 186, 160, 255, 5, 185, 196, 255, 5, 168, 56, 255, 5, 168, 68, 72, 4, 177, 96, 104, 4, 168, 96, 36, 5, 241, 97, 37, 5, 232, 97, 255, 5, 168, 132, 255, 5, 168, 152, 255, 5, 168, 160, 255, 5, 168, 196, 255, 5, 249, 70, 255, 5, 232, 70, 255, 5, 25, 59, 255, 5, 8, 59, 255, 5, 249, 154, 255, 5, 232, 154, 255, 5, 250, 162, 255, 5, 232, 162, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 239, 14, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 232, 14, 0, 32, 247, 24, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 252, 24, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 112, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 116, 8, 0, 0, 116, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 44, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 38, 0, 67, 0, 60, 0, 13, 1, 60, 0, 60, 0, 60, 0, 12, 1, 60, 0, 7, 1, 60, 0, 60, 0, 60, 0, 6, 1, 38, 0, 68, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 38, 0, 83, 0, 60, 0, 97, 1, 60, 0, 60, 0, 60, 0, 96, 1, 38, 0, 90, 0, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 228, 32, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 20, 27, 0, 0, 20, 27, 0, 0, 20, 27, 0, 0, 20, 27, 0, 0, 216, 27, 0, 0, 216, 27, 0, 0, 228, 28, 0, 0, 228, 28, 0, 0, 204, 32, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 164, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 89, 0, 97, 0, 48, 0, 56, 0, 48, 0, 56, 0, 105, 0, 113, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 135, 0, 48, 0, 48, 0, 143, 0, 147, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 155, 0, 129, 0, 48, 0, 48, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 121, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 100, 1, 192, 0, 192, 0, 164, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 4, 0, 202, 1, 5, 0, 202, 34, 5, 0, 202, 67, 5, 0, 202, 100, 5, 0, 202, 133, 5, 0, 202, 166, 5, 0, 202, 199, 5, 0, 202, 232, 5, 0, 202, 9, 6, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 134, 1, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 102, 2, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 70, 3, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 198, 4, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 166, 5, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 4, 0, 201, 134, 6, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 4, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 4, 0, 201, 198, 7, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 47, 5, 5, 3, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 49, 5, 5, 2, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 9, 0, 201, 6, 10, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 10, 0, 201, 134, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 197, 162, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 12, 0, 201, 6, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 2, 0, 197, 226, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 13, 0, 201, 134, 14, 0, 201, 70, 15, 0, 201, 6, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 3, 47, 5, 160, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 3, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 2, 0, 1, 3, 3, 239, 134, 5, 12, 3, 2, 239, 134, 5, 65, 3, 3, 239, 134, 5, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 46, 5, 5, 2, 0, 1, 3, 3, 239, 5, 5, 12, 3, 2, 239, 5, 5, 65, 3, 3, 239, 5, 5, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 2, 0, 1, 3, 197, 162, 12, 3, 0, 192, 197, 98, 65, 3, 197, 162, 5, 46, 196, 160, 2, 0, 1, 3, 0, 192, 197, 162, 12, 3, 0, 192, 197, 226, 65, 3, 0, 192, 197, 162, 156, 78, 196, 160, 48, 0, 12, 3, 1, 192, 197, 34, 5, 78, 196, 160, 48, 0, 12, 3, 1, 192, 197, 98, 156, 78, 196, 202, 48, 0, 12, 3, 1, 192, 197, 162, 5, 78, 196, 202, 48, 0, 12, 3, 1, 192, 197, 226, 156, 78, 196, 196, 48, 0, 12, 3, 2, 192, 197, 34, 5, 78, 196, 196, 48, 0, 12, 3, 2, 192, 197, 98, 156, 92, 196, 196, 48, 0, 12, 3, 3, 192, 197, 34, 5, 92, 196, 196, 48, 0, 12, 3, 3, 192, 197, 98, 156, 92, 196, 210, 48, 0, 12, 3, 3, 192, 197, 162, 5, 92, 196, 210, 48, 0, 12, 3, 3, 192, 197, 226, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 76, 186, 84, 185, 88, 185, 92, 185, 96, 186, 100, 186, 108, 185, 112, 185, 116, 185, 120, 186, 124, 185, 132, 185, 136, 185, 140, 185, 148, 28, 4, 185, 160, 185, 168, 185, 172, 185, 176, 185, 180, 185, 184, 30, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 32, 4, 168, 76, 168, 84, 168, 88, 168, 92, 168, 96, 168, 100, 168, 108, 168, 112, 168, 116, 168, 120, 168, 124, 168, 132, 168, 136, 168, 140, 168, 148, 34, 4, 168, 160, 168, 168, 168, 172, 168, 176, 168, 180, 168, 184, 36, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 132, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 38, 4, 186, 85, 154, 85, 250, 85, 90, 86, 186, 101, 154, 101, 250, 101, 90, 102, 217, 76, 154, 126, 185, 133, 153, 133, 249, 133, 153, 134, 89, 134, 136, 14, 217, 134, 185, 169, 153, 169, 249, 169, 89, 170, 153, 185, 185, 196, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 40, 4, 168, 85, 136, 85, 232, 85, 72, 86, 168, 101, 136, 101, 232, 101, 72, 102, 200, 76, 136, 126, 168, 133, 136, 133, 232, 133, 136, 134, 72, 134, 128, 14, 200, 134, 168, 169, 136, 169, 232, 169, 72, 170, 136, 185, 168, 196, 72, 186, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 185, 72, 168, 72, 249, 65, 232, 65, 185, 66, 168, 66, 185, 68, 168, 68, 25, 78, 8, 78, 185, 80, 168, 80, 58, 87, 40, 87, 218, 85, 200, 85, 186, 86, 168, 86, 26, 87, 8, 87, 26, 86, 8, 86, 249, 93, 232, 93, 217, 93, 200, 93, 185, 94, 168, 94, 249, 94, 232, 94, 249, 97, 232, 97, 89, 99, 72, 99, 154, 102, 136, 102, 58, 103, 40, 103, 218, 101, 200, 101, 26, 103, 8, 103, 186, 102, 168, 104, 6, 8, 8, 8, 250, 109, 232, 109, 249, 114, 232, 114, 168, 144, 153, 117, 136, 117, 249, 118, 232, 118, 25, 118, 8, 118, 121, 119, 104, 119, 89, 119, 72, 119, 154, 125, 136, 125, 250, 126, 232, 126, 26, 126, 8, 126, 10, 8, 185, 128, 168, 128, 57, 135, 40, 135, 217, 133, 200, 133, 121, 134, 104, 134, 12, 8, 14, 8, 153, 149, 136, 149, 249, 150, 232, 150, 25, 150, 8, 150, 154, 153, 136, 153, 250, 153, 232, 153, 42, 4, 44, 4, 185, 156, 168, 156, 249, 162, 232, 162, 25, 162, 8, 162, 185, 164, 168, 164, 153, 170, 136, 170, 57, 171, 40, 171, 217, 169, 200, 169, 57, 170, 40, 170, 121, 170, 104, 170, 25, 171, 8, 171, 249, 177, 232, 177, 249, 185, 232, 185, 89, 186, 153, 189, 136, 189, 185, 190, 168, 190, 185, 192, 168, 192, 232, 152, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 84, 200, 152, 169, 152, 200, 56, 169, 84, 187, 100, 187, 108, 169, 100, 169, 108, 169, 200, 169, 124, 217, 132, 187, 84, 200, 132, 169, 84, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 186, 152, 255, 5, 185, 188, 255, 5, 168, 64, 255, 5, 168, 152, 255, 5, 168, 188, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 250, 154, 255, 5, 232, 154, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 47, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 105, 8, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 110, 8, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 231, 23, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 235, 23, 0, 0, 235, 23, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 106, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 50, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 115, 0, 109, 0, 110, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 0, 0, 105, 220, 38, 0, 65, 0, 60, 0, 226, 0, 60, 0, 60, 0, 60, 0, 194, 0, 60, 0, 60, 0, 224, 0, 60, 0, 60, 0, 60, 0, 192, 0, 38, 0, 67, 0, 60, 0, 13, 1, 60, 0, 60, 0, 60, 0, 12, 1, 38, 0, 68, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 38, 0, 74, 1, 60, 0, 60, 0, 241, 0, 60, 0, 60, 0, 60, 0, 209, 0, 60, 0, 60, 0, 68, 1, 60, 0, 60, 0, 60, 0, 67, 1, 38, 0, 83, 0, 60, 0, 97, 1, 60, 0, 60, 0, 60, 0, 96, 1, 38, 0, 90, 0, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 60, 0, 227, 0, 60, 0, 60, 0, 60, 0, 195, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 225, 0, 60, 0, 60, 0, 60, 0, 193, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 0, 0, 170, 170, 170, 170, 170, 170, 250, 49, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 214, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 132, 39, 0, 0, 132, 39, 0, 0, 228, 40, 0, 0, 228, 40, 0, 0, 112, 45, 0, 0, 112, 45, 0, 0, 222, 49, 0, 0, 222, 49, 0, 0, 222, 49, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 107, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 121, 0, 129, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 148, 0, 156, 0, 162, 0, 170, 0, 178, 0, 186, 0, 48, 0, 56, 0, 185, 0, 193, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 210, 0, 218, 0, 48, 0, 48, 0, 48, 0, 226, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 234, 0, 48, 0, 48, 0, 242, 0, 48, 0, 248, 0, 0, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 3, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 11, 1, 48, 0, 18, 1, 26, 1, 30, 1, 38, 1, 43, 1, 51, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 57, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 65, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 68, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 201, 0, 208, 0, 48, 0, 48, 0, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 98, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 76, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 84, 1, 48, 0, 48, 0, 90, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 1, 228, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 80, 2, 136, 2, 200, 2, 192, 0, 228, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 35, 0, 202, 65, 35, 0, 202, 98, 35, 0, 202, 131, 35, 0, 202, 164, 35, 0, 202, 197, 35, 0, 202, 230, 35, 0, 202, 7, 36, 0, 202, 40, 36, 0, 202, 73, 36, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 198, 0, 0, 156, 5, 0, 44, 201, 6, 4, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 198, 4, 0, 201, 166, 6, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 102, 7, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 70, 8, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 38, 9, 0, 5, 5, 0, 44, 201, 102, 13, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 38, 14, 0, 201, 6, 16, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 230, 16, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 198, 17, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 2, 43, 198, 1, 0, 0, 134, 5, 2, 43, 198, 33, 0, 0, 198, 65, 0, 0, 198, 97, 0, 0, 134, 5, 3, 93, 201, 166, 18, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 134, 70, 30, 69, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 129, 0, 0, 5, 5, 127, 12, 134, 5, 4, 93, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 130, 26, 0, 5, 70, 2, 43, 193, 5, 5, 93, 5, 5, 2, 43, 193, 3, 5, 93, 193, 4, 5, 93, 193, 2, 5, 93, 5, 5, 3, 93, 201, 70, 19, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 5, 70, 30, 69, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 193, 6, 5, 93, 5, 5, 125, 12, 5, 5, 4, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 20, 0, 201, 166, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 49, 5, 5, 2, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 30, 69, 5, 71, 30, 69, 201, 70, 27, 0, 201, 198, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 30, 0, 201, 6, 31, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 31, 0, 201, 134, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 1, 0, 197, 98, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 33, 0, 201, 6, 34, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 2, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 34, 0, 201, 134, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 2, 0, 197, 34, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 70, 36, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 38, 0, 5, 71, 238, 102, 201, 6, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 197, 34, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 8, 0, 192, 0, 0, 0, 197, 226, 8, 0, 192, 0, 0, 0, 197, 34, 9, 0, 192, 0, 0, 0, 197, 98, 9, 0, 192, 0, 0, 0, 197, 162, 9, 0, 197, 227, 9, 0, 197, 66, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 10, 0, 197, 195, 10, 0, 197, 34, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 98, 7, 0, 197, 162, 11, 0, 5, 5, 10, 121, 197, 226, 11, 0, 197, 34, 12, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 98, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 2, 8, 0, 5, 5, 24, 121, 197, 34, 8, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 39, 0, 201, 134, 43, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 47, 0, 201, 166, 48, 0, 201, 38, 50, 0, 201, 166, 51, 0, 201, 38, 53, 0, 201, 166, 54, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 56, 0, 201, 230, 56, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 19, 0, 197, 226, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 57, 0, 201, 102, 58, 0, 201, 38, 59, 0, 201, 230, 59, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 60, 0, 201, 70, 64, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 23, 0, 197, 98, 23, 0, 197, 162, 23, 0, 197, 226, 23, 0, 197, 34, 24, 0, 197, 98, 24, 0, 197, 162, 24, 0, 197, 226, 24, 0, 197, 98, 21, 0, 197, 34, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 67, 0, 201, 166, 68, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 69, 0, 201, 38, 70, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 97, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 5, 0, 5, 5, 93, 0, 134, 0, 5, 0, 3, 5, 93, 0, 134, 0, 5, 0, 4, 5, 93, 0, 134, 0, 5, 0, 2, 5, 93, 0, 134, 0, 5, 0, 6, 5, 93, 0, 134, 0, 5, 0, 5, 5, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 5, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 5, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 5, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 5, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 5, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 5, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 5, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 5, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 5, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 5, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 5, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 5, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 5, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 70, 2, 43, 5, 162, 0, 0, 134, 5, 2, 43, 5, 162, 0, 0, 5, 70, 2, 43, 5, 162, 0, 0, 193, 5, 5, 93, 5, 162, 0, 0, 5, 5, 2, 43, 5, 162, 0, 0, 193, 3, 5, 93, 5, 162, 0, 0, 193, 4, 5, 93, 5, 162, 0, 0, 193, 2, 5, 93, 5, 162, 0, 0, 134, 71, 30, 69, 5, 160, 0, 0, 134, 70, 30, 69, 5, 160, 0, 0, 5, 71, 30, 69, 5, 160, 0, 0, 5, 70, 30, 69, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 193, 6, 5, 93, 5, 190, 0, 0, 193, 4, 5, 93, 5, 164, 0, 0, 134, 5, 3, 93, 5, 164, 0, 0, 5, 5, 3, 93, 5, 164, 0, 0, 193, 6, 5, 93, 5, 162, 0, 0, 193, 2, 5, 93, 5, 136, 0, 0, 134, 5, 3, 93, 5, 136, 0, 0, 5, 5, 3, 93, 5, 136, 0, 0, 134, 5, 4, 93, 5, 136, 0, 0, 5, 5, 4, 93, 5, 136, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 193, 6, 5, 93, 5, 164, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 70, 2, 43, 5, 200, 0, 0, 134, 5, 2, 43, 5, 200, 0, 0, 5, 70, 2, 43, 5, 200, 0, 0, 193, 5, 5, 93, 5, 200, 0, 0, 5, 5, 2, 43, 5, 200, 0, 0, 193, 3, 5, 93, 5, 200, 0, 0, 193, 4, 5, 93, 5, 200, 0, 0, 193, 2, 5, 93, 5, 200, 0, 0, 134, 71, 30, 69, 5, 196, 0, 0, 134, 70, 30, 69, 5, 196, 0, 0, 5, 71, 30, 69, 5, 196, 0, 0, 5, 70, 30, 69, 5, 196, 0, 0, 134, 71, 30, 69, 5, 210, 0, 0, 134, 70, 30, 69, 5, 210, 0, 0, 5, 71, 30, 69, 5, 210, 0, 0, 5, 70, 30, 69, 5, 210, 0, 0, 134, 71, 30, 69, 5, 204, 0, 0, 134, 70, 30, 69, 5, 204, 0, 0, 5, 71, 30, 69, 5, 204, 0, 0, 5, 70, 30, 69, 5, 204, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 134, 70, 2, 43, 5, 196, 0, 0, 134, 5, 2, 43, 5, 196, 0, 0, 5, 70, 2, 43, 5, 196, 0, 0, 193, 5, 5, 93, 5, 196, 0, 0, 5, 5, 2, 43, 5, 196, 0, 0, 193, 3, 5, 93, 5, 196, 0, 0, 193, 4, 5, 93, 5, 196, 0, 0, 193, 2, 5, 93, 5, 196, 0, 0, 134, 5, 2, 43, 5, 136, 0, 0, 5, 5, 2, 43, 5, 136, 0, 0, 134, 5, 2, 43, 5, 138, 0, 0, 5, 5, 2, 43, 5, 138, 0, 0, 134, 5, 2, 43, 5, 182, 0, 0, 5, 5, 2, 43, 5, 182, 0, 0, 134, 5, 2, 43, 5, 154, 0, 0, 5, 5, 2, 43, 5, 154, 0, 0, 193, 6, 5, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 193, 6, 5, 93, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 6, 121, 197, 98, 7, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 162, 7, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 226, 7, 0, 5, 5, 24, 121, 197, 34, 8, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 98, 8, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 162, 8, 0, 5, 5, 107, 121, 197, 226, 8, 0, 5, 5, 111, 121, 197, 34, 9, 0, 5, 5, 115, 121, 197, 98, 9, 0, 5, 5, 119, 121, 197, 162, 9, 0, 197, 227, 9, 0, 197, 66, 10, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 130, 10, 0, 197, 195, 10, 0, 197, 34, 11, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 98, 11, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 98, 7, 0, 197, 162, 11, 0, 5, 5, 10, 121, 197, 226, 11, 0, 197, 34, 12, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 98, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 2, 8, 0, 5, 5, 24, 121, 197, 34, 8, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 7, 0, 8, 3, 11, 0, 8, 3, 0, 192, 198, 65, 10, 3, 0, 192, 198, 97, 64, 3, 2, 235, 134, 70, 65, 3, 198, 129, 0, 3, 2, 235, 134, 70, 1, 3, 198, 129, 2, 3, 2, 235, 134, 5, 3, 3, 198, 161, 0, 46, 156, 5, 48, 0, 12, 3, 2, 239, 134, 5, 0, 68, 156, 5, 2, 0, 1, 3, 255, 255, 30, 69, 134, 71, 3, 3, 255, 255, 30, 69, 134, 70, 65, 3, 255, 255, 30, 69, 134, 71, 0, 70, 156, 5, 48, 0, 8, 3, 0, 192, 198, 129, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 42, 5, 5, 7, 0, 8, 3, 15, 0, 8, 3, 255, 255, 5, 93, 193, 4, 10, 3, 255, 255, 5, 93, 193, 2, 64, 3, 2, 235, 5, 70, 65, 3, 255, 255, 5, 93, 193, 5, 0, 3, 2, 235, 5, 70, 1, 3, 255, 255, 5, 93, 193, 5, 2, 3, 2, 235, 5, 5, 3, 3, 255, 255, 5, 93, 193, 3, 0, 46, 5, 5, 48, 0, 12, 3, 2, 239, 5, 5, 0, 68, 5, 5, 2, 0, 1, 3, 255, 255, 30, 69, 5, 71, 3, 3, 255, 255, 30, 69, 5, 70, 65, 3, 255, 255, 30, 69, 5, 71, 0, 70, 5, 5, 48, 0, 8, 3, 255, 255, 5, 93, 193, 6, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 48, 0, 12, 3, 197, 162, 5, 46, 196, 160, 48, 0, 12, 3, 0, 192, 197, 98, 156, 42, 196, 162, 7, 0, 8, 3, 12, 0, 8, 3, 1, 192, 198, 34, 10, 3, 1, 192, 198, 98, 64, 3, 0, 192, 197, 162, 65, 3, 0, 192, 198, 162, 0, 3, 0, 192, 197, 162, 1, 3, 0, 192, 198, 162, 2, 3, 0, 192, 197, 226, 3, 3, 0, 192, 198, 226, 5, 42, 196, 162, 7, 0, 8, 3, 12, 0, 8, 3, 2, 192, 197, 34, 10, 3, 2, 192, 197, 98, 64, 3, 1, 192, 197, 34, 65, 3, 1, 192, 197, 98, 0, 3, 1, 192, 197, 34, 1, 3, 1, 192, 197, 98, 2, 3, 1, 192, 197, 162, 3, 3, 1, 192, 197, 226, 156, 68, 196, 160, 2, 0, 1, 3, 2, 192, 197, 162, 3, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 5, 68, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 3, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 156, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 226, 156, 70, 196, 190, 48, 0, 8, 3, 1, 192, 198, 162, 5, 70, 196, 190, 48, 0, 8, 3, 4, 192, 197, 34, 156, 70, 196, 162, 48, 0, 8, 3, 2, 192, 198, 34, 5, 70, 196, 162, 48, 0, 8, 3, 5, 192, 197, 34, 156, 78, 196, 202, 48, 0, 12, 3, 6, 192, 197, 162, 5, 78, 196, 202, 48, 0, 12, 3, 6, 192, 197, 226, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 7, 0, 8, 3, 12, 0, 8, 3, 3, 192, 198, 98, 10, 3, 3, 192, 198, 162, 64, 3, 14, 192, 197, 34, 65, 3, 2, 192, 198, 226, 0, 3, 14, 192, 197, 34, 1, 3, 2, 192, 198, 226, 2, 3, 14, 192, 197, 98, 3, 3, 3, 192, 198, 34, 5, 42, 196, 200, 7, 0, 8, 3, 12, 0, 8, 3, 15, 192, 197, 162, 10, 3, 15, 192, 197, 226, 64, 3, 14, 192, 197, 162, 65, 3, 14, 192, 197, 226, 0, 3, 14, 192, 197, 162, 1, 3, 14, 192, 197, 226, 2, 3, 15, 192, 197, 34, 3, 3, 15, 192, 197, 98, 156, 68, 196, 196, 2, 0, 1, 3, 16, 192, 197, 34, 3, 3, 16, 192, 197, 98, 65, 3, 16, 192, 197, 34, 5, 68, 196, 196, 2, 0, 1, 3, 16, 192, 197, 162, 3, 3, 16, 192, 197, 226, 65, 3, 16, 192, 197, 162, 156, 68, 196, 210, 2, 0, 1, 3, 17, 192, 197, 34, 3, 3, 17, 192, 197, 98, 65, 3, 17, 192, 197, 34, 5, 68, 196, 210, 2, 0, 1, 3, 17, 192, 197, 162, 3, 3, 17, 192, 197, 226, 65, 3, 17, 192, 197, 162, 156, 68, 196, 204, 2, 0, 1, 3, 18, 192, 197, 34, 3, 3, 18, 192, 197, 98, 65, 3, 18, 192, 197, 34, 5, 68, 196, 204, 2, 0, 1, 3, 18, 192, 197, 162, 3, 3, 18, 192, 197, 226, 65, 3, 18, 192, 197, 162, 156, 78, 196, 196, 48, 0, 12, 3, 19, 192, 197, 34, 5, 78, 196, 196, 48, 0, 12, 3, 19, 192, 197, 98, 156, 92, 196, 196, 48, 0, 12, 3, 20, 192, 197, 34, 5, 92, 196, 196, 48, 0, 12, 3, 20, 192, 197, 98, 156, 92, 196, 210, 48, 0, 12, 3, 20, 192, 197, 162, 5, 92, 196, 210, 48, 0, 12, 3, 20, 192, 197, 226, 156, 42, 196, 196, 7, 0, 8, 3, 12, 0, 8, 3, 4, 192, 198, 98, 10, 3, 4, 192, 198, 162, 64, 3, 21, 192, 197, 34, 65, 3, 3, 192, 198, 226, 0, 3, 21, 192, 197, 34, 1, 3, 3, 192, 198, 226, 2, 3, 21, 192, 197, 98, 3, 3, 4, 192, 198, 34, 5, 42, 196, 196, 7, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 162, 10, 3, 22, 192, 197, 226, 64, 3, 21, 192, 197, 162, 65, 3, 21, 192, 197, 226, 0, 3, 21, 192, 197, 162, 1, 3, 21, 192, 197, 226, 2, 3, 22, 192, 197, 34, 3, 3, 22, 192, 197, 98, 156, 70, 196, 196, 48, 0, 8, 3, 4, 192, 198, 226, 5, 70, 196, 196, 48, 0, 8, 3, 25, 192, 197, 34, 25, 0, 197, 99, 48, 0, 8, 3, 5, 192, 198, 35, 25, 0, 197, 195, 48, 0, 8, 3, 26, 192, 197, 35, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 107, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 168, 43, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 84, 31, 0, 0, 84, 31, 0, 0, 180, 32, 0, 0, 180, 32, 0, 0, 92, 35, 0, 0, 92, 35, 0, 0, 132, 39, 0, 0, 132, 39, 0, 0, 140, 43, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 232, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 121, 0, 129, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 169, 0, 48, 0, 176, 0, 184, 0, 188, 0, 196, 0, 201, 0, 209, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 215, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 223, 0, 161, 0, 48, 0, 48, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 192, 1, 228, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 20, 0, 202, 33, 20, 0, 202, 66, 20, 0, 202, 99, 20, 0, 202, 132, 20, 0, 202, 165, 20, 0, 202, 198, 20, 0, 202, 231, 20, 0, 202, 8, 21, 0, 202, 41, 21, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 201, 70, 3, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 6, 4, 0, 201, 230, 5, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 166, 6, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 134, 7, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 102, 8, 0, 5, 5, 0, 44, 201, 166, 12, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 102, 13, 0, 201, 70, 15, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 38, 16, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 6, 17, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 2, 43, 198, 1, 0, 0, 134, 5, 2, 43, 198, 33, 0, 0, 198, 65, 0, 0, 198, 97, 0, 0, 134, 5, 3, 93, 201, 230, 17, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 134, 70, 30, 69, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 198, 129, 0, 0, 5, 5, 127, 12, 134, 5, 4, 93, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 194, 19, 0, 5, 70, 2, 43, 193, 5, 5, 93, 5, 5, 2, 43, 193, 3, 5, 93, 193, 4, 5, 93, 193, 2, 5, 93, 5, 5, 3, 93, 201, 134, 18, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 5, 70, 30, 69, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 193, 6, 5, 93, 5, 5, 125, 12, 5, 5, 4, 93, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 19, 0, 201, 230, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 49, 5, 5, 2, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 30, 69, 5, 71, 30, 69, 201, 134, 26, 0, 201, 6, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 29, 0, 201, 70, 30, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 31, 0, 201, 198, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 1, 0, 197, 98, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 32, 0, 201, 70, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 2, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 34, 0, 201, 198, 34, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 2, 0, 197, 34, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 35, 0, 201, 38, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 42, 0, 201, 70, 44, 0, 201, 198, 45, 0, 201, 70, 47, 0, 201, 198, 48, 0, 201, 70, 50, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 51, 0, 201, 134, 52, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 12, 0, 197, 34, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 53, 0, 201, 6, 54, 0, 201, 198, 54, 0, 201, 134, 55, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 56, 0, 201, 230, 59, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 16, 0, 197, 162, 16, 0, 197, 226, 16, 0, 197, 34, 17, 0, 197, 98, 17, 0, 197, 162, 17, 0, 197, 226, 17, 0, 197, 34, 18, 0, 197, 162, 14, 0, 197, 98, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 63, 0, 201, 70, 64, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 65, 0, 201, 198, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 97, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 5, 0, 5, 5, 93, 0, 134, 0, 5, 0, 3, 5, 93, 0, 134, 0, 5, 0, 4, 5, 93, 0, 134, 0, 5, 0, 2, 5, 93, 0, 134, 0, 5, 0, 6, 5, 93, 0, 134, 0, 5, 0, 5, 5, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 5, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 5, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 5, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 5, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 5, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 5, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 5, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 5, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 5, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 5, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 5, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 5, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 5, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 5, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 6, 5, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 70, 2, 43, 5, 162, 0, 0, 134, 5, 2, 43, 5, 162, 0, 0, 5, 70, 2, 43, 5, 162, 0, 0, 193, 5, 5, 93, 5, 162, 0, 0, 5, 5, 2, 43, 5, 162, 0, 0, 193, 3, 5, 93, 5, 162, 0, 0, 193, 4, 5, 93, 5, 162, 0, 0, 193, 2, 5, 93, 5, 162, 0, 0, 134, 71, 30, 69, 5, 160, 0, 0, 134, 70, 30, 69, 5, 160, 0, 0, 5, 71, 30, 69, 5, 160, 0, 0, 5, 70, 30, 69, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 193, 6, 5, 93, 5, 190, 0, 0, 193, 4, 5, 93, 5, 164, 0, 0, 134, 5, 3, 93, 5, 164, 0, 0, 5, 5, 3, 93, 5, 164, 0, 0, 193, 6, 5, 93, 5, 162, 0, 0, 193, 2, 5, 93, 5, 136, 0, 0, 134, 5, 3, 93, 5, 136, 0, 0, 5, 5, 3, 93, 5, 136, 0, 0, 134, 5, 4, 93, 5, 136, 0, 0, 5, 5, 4, 93, 5, 136, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 193, 6, 5, 93, 5, 164, 0, 0, 134, 70, 2, 43, 5, 200, 0, 0, 134, 5, 2, 43, 5, 200, 0, 0, 5, 70, 2, 43, 5, 200, 0, 0, 193, 5, 5, 93, 5, 200, 0, 0, 5, 5, 2, 43, 5, 200, 0, 0, 193, 3, 5, 93, 5, 200, 0, 0, 193, 4, 5, 93, 5, 200, 0, 0, 193, 2, 5, 93, 5, 200, 0, 0, 134, 71, 30, 69, 5, 196, 0, 0, 134, 70, 30, 69, 5, 196, 0, 0, 5, 71, 30, 69, 5, 196, 0, 0, 5, 70, 30, 69, 5, 196, 0, 0, 134, 71, 30, 69, 5, 210, 0, 0, 134, 70, 30, 69, 5, 210, 0, 0, 5, 71, 30, 69, 5, 210, 0, 0, 5, 70, 30, 69, 5, 210, 0, 0, 134, 71, 30, 69, 5, 204, 0, 0, 134, 70, 30, 69, 5, 204, 0, 0, 5, 71, 30, 69, 5, 204, 0, 0, 5, 70, 30, 69, 5, 204, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 134, 70, 2, 43, 5, 196, 0, 0, 134, 5, 2, 43, 5, 196, 0, 0, 5, 70, 2, 43, 5, 196, 0, 0, 193, 5, 5, 93, 5, 196, 0, 0, 5, 5, 2, 43, 5, 196, 0, 0, 193, 3, 5, 93, 5, 196, 0, 0, 193, 4, 5, 93, 5, 196, 0, 0, 193, 2, 5, 93, 5, 196, 0, 0, 134, 5, 2, 43, 5, 136, 0, 0, 5, 5, 2, 43, 5, 136, 0, 0, 134, 5, 2, 43, 5, 138, 0, 0, 5, 5, 2, 43, 5, 138, 0, 0, 134, 5, 2, 43, 5, 182, 0, 0, 5, 5, 2, 43, 5, 182, 0, 0, 134, 5, 2, 43, 5, 154, 0, 0, 5, 5, 2, 43, 5, 154, 0, 0, 193, 6, 5, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 193, 6, 5, 93, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 7, 0, 8, 3, 11, 0, 8, 3, 0, 192, 198, 65, 10, 3, 0, 192, 198, 97, 64, 3, 2, 235, 134, 70, 65, 3, 198, 129, 0, 3, 2, 235, 134, 70, 1, 3, 198, 129, 2, 3, 2, 235, 134, 5, 3, 3, 198, 161, 0, 46, 156, 5, 48, 0, 12, 3, 2, 239, 134, 5, 0, 68, 156, 5, 2, 0, 1, 3, 255, 255, 30, 69, 134, 71, 3, 3, 255, 255, 30, 69, 134, 70, 65, 3, 255, 255, 30, 69, 134, 71, 0, 70, 156, 5, 48, 0, 8, 3, 0, 192, 198, 129, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 42, 5, 5, 7, 0, 8, 3, 15, 0, 8, 3, 255, 255, 5, 93, 193, 4, 10, 3, 255, 255, 5, 93, 193, 2, 64, 3, 2, 235, 5, 70, 65, 3, 255, 255, 5, 93, 193, 5, 0, 3, 2, 235, 5, 70, 1, 3, 255, 255, 5, 93, 193, 5, 2, 3, 2, 235, 5, 5, 3, 3, 255, 255, 5, 93, 193, 3, 0, 46, 5, 5, 48, 0, 12, 3, 2, 239, 5, 5, 0, 68, 5, 5, 2, 0, 1, 3, 255, 255, 30, 69, 5, 71, 3, 3, 255, 255, 30, 69, 5, 70, 65, 3, 255, 255, 30, 69, 5, 71, 0, 70, 5, 5, 48, 0, 8, 3, 255, 255, 5, 93, 193, 6, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 48, 0, 12, 3, 197, 162, 5, 46, 196, 160, 48, 0, 12, 3, 0, 192, 197, 98, 156, 42, 196, 162, 7, 0, 8, 3, 12, 0, 8, 3, 1, 192, 198, 34, 10, 3, 1, 192, 198, 98, 64, 3, 0, 192, 197, 162, 65, 3, 0, 192, 198, 162, 0, 3, 0, 192, 197, 162, 1, 3, 0, 192, 198, 162, 2, 3, 0, 192, 197, 226, 3, 3, 0, 192, 198, 226, 5, 42, 196, 162, 7, 0, 8, 3, 12, 0, 8, 3, 2, 192, 197, 34, 10, 3, 2, 192, 197, 98, 64, 3, 1, 192, 197, 34, 65, 3, 1, 192, 197, 98, 0, 3, 1, 192, 197, 34, 1, 3, 1, 192, 197, 98, 2, 3, 1, 192, 197, 162, 3, 3, 1, 192, 197, 226, 156, 68, 196, 160, 2, 0, 1, 3, 2, 192, 197, 162, 3, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 5, 68, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 3, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 156, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 226, 156, 70, 196, 190, 48, 0, 8, 3, 1, 192, 198, 162, 5, 70, 196, 190, 48, 0, 8, 3, 4, 192, 197, 34, 156, 70, 196, 162, 48, 0, 8, 3, 2, 192, 198, 34, 5, 70, 196, 162, 48, 0, 8, 3, 5, 192, 197, 34, 156, 78, 196, 202, 48, 0, 12, 3, 6, 192, 197, 162, 5, 78, 196, 202, 48, 0, 12, 3, 6, 192, 197, 226, 156, 42, 196, 200, 7, 0, 8, 3, 12, 0, 8, 3, 3, 192, 198, 98, 10, 3, 3, 192, 198, 162, 64, 3, 7, 192, 197, 98, 65, 3, 2, 192, 198, 226, 0, 3, 7, 192, 197, 98, 1, 3, 2, 192, 198, 226, 2, 3, 7, 192, 197, 162, 3, 3, 3, 192, 198, 34, 5, 42, 196, 200, 7, 0, 8, 3, 12, 0, 8, 3, 8, 192, 197, 226, 10, 3, 9, 192, 197, 34, 64, 3, 7, 192, 197, 226, 65, 3, 8, 192, 197, 34, 0, 3, 7, 192, 197, 226, 1, 3, 8, 192, 197, 34, 2, 3, 8, 192, 197, 98, 3, 3, 8, 192, 197, 162, 156, 68, 196, 196, 2, 0, 1, 3, 9, 192, 197, 98, 3, 3, 9, 192, 197, 162, 65, 3, 9, 192, 197, 98, 5, 68, 196, 196, 2, 0, 1, 3, 9, 192, 197, 226, 3, 3, 10, 192, 197, 34, 65, 3, 9, 192, 197, 226, 156, 68, 196, 210, 2, 0, 1, 3, 10, 192, 197, 98, 3, 3, 10, 192, 197, 162, 65, 3, 10, 192, 197, 98, 5, 68, 196, 210, 2, 0, 1, 3, 10, 192, 197, 226, 3, 3, 11, 192, 197, 34, 65, 3, 10, 192, 197, 226, 156, 68, 196, 204, 2, 0, 1, 3, 11, 192, 197, 98, 3, 3, 11, 192, 197, 162, 65, 3, 11, 192, 197, 98, 5, 68, 196, 204, 2, 0, 1, 3, 11, 192, 197, 226, 3, 3, 12, 192, 197, 34, 65, 3, 11, 192, 197, 226, 156, 78, 196, 196, 48, 0, 12, 3, 12, 192, 197, 98, 5, 78, 196, 196, 48, 0, 12, 3, 12, 192, 197, 162, 156, 92, 196, 196, 48, 0, 12, 3, 13, 192, 197, 98, 5, 92, 196, 196, 48, 0, 12, 3, 13, 192, 197, 162, 156, 92, 196, 210, 48, 0, 12, 3, 13, 192, 197, 226, 5, 92, 196, 210, 48, 0, 12, 3, 14, 192, 197, 34, 156, 42, 196, 196, 7, 0, 8, 3, 12, 0, 8, 3, 4, 192, 198, 98, 10, 3, 4, 192, 198, 162, 64, 3, 14, 192, 197, 98, 65, 3, 3, 192, 198, 226, 0, 3, 14, 192, 197, 98, 1, 3, 3, 192, 198, 226, 2, 3, 14, 192, 197, 162, 3, 3, 4, 192, 198, 34, 5, 42, 196, 196, 7, 0, 8, 3, 12, 0, 8, 3, 15, 192, 197, 226, 10, 3, 16, 192, 197, 34, 64, 3, 14, 192, 197, 226, 65, 3, 15, 192, 197, 34, 0, 3, 14, 192, 197, 226, 1, 3, 15, 192, 197, 34, 2, 3, 15, 192, 197, 98, 3, 3, 15, 192, 197, 162, 156, 70, 196, 196, 48, 0, 8, 3, 4, 192, 198, 226, 5, 70, 196, 196, 48, 0, 8, 3, 18, 192, 197, 98, 18, 0, 197, 163, 48, 0, 8, 3, 5, 192, 198, 35, 19, 0, 197, 3, 48, 0, 8, 3, 19, 192, 197, 99, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 20, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 22, 4, 185, 64, 24, 4, 185, 76, 186, 84, 185, 88, 185, 92, 185, 96, 186, 100, 186, 108, 185, 112, 185, 116, 185, 120, 26, 4, 28, 4, 185, 136, 185, 140, 185, 148, 30, 4, 185, 160, 185, 168, 185, 172, 185, 176, 185, 180, 185, 184, 32, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 34, 4, 168, 64, 36, 4, 168, 76, 168, 84, 168, 88, 168, 92, 168, 96, 168, 100, 168, 108, 168, 112, 168, 116, 168, 120, 38, 4, 40, 4, 168, 136, 168, 140, 168, 148, 42, 4, 168, 160, 168, 168, 168, 172, 168, 176, 168, 180, 168, 184, 44, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 132, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 217, 60, 185, 216, 185, 60, 185, 208, 185, 212, 185, 204, 185, 196, 46, 4, 186, 85, 154, 85, 250, 85, 90, 86, 186, 101, 154, 101, 250, 101, 90, 102, 217, 76, 217, 128, 185, 133, 153, 133, 249, 133, 153, 134, 185, 220, 136, 14, 185, 200, 185, 169, 153, 169, 249, 169, 89, 170, 153, 185, 185, 224, 0, 8, 200, 60, 168, 216, 168, 60, 168, 208, 168, 212, 168, 204, 168, 196, 48, 4, 168, 85, 136, 85, 232, 85, 72, 86, 168, 101, 136, 101, 232, 101, 72, 102, 200, 76, 200, 128, 168, 133, 136, 133, 232, 133, 136, 134, 168, 220, 128, 14, 168, 200, 168, 169, 136, 169, 232, 169, 72, 170, 136, 185, 168, 224, 72, 186, 25, 59, 8, 59, 217, 57, 200, 57, 50, 4, 52, 4, 153, 69, 136, 69, 249, 69, 232, 69, 185, 70, 168, 70, 185, 72, 168, 72, 25, 78, 8, 78, 185, 80, 168, 80, 26, 87, 8, 87, 218, 85, 200, 85, 186, 86, 168, 86, 250, 86, 232, 86, 26, 86, 8, 86, 249, 93, 232, 93, 217, 93, 200, 93, 185, 94, 168, 94, 217, 94, 200, 94, 249, 97, 232, 97, 57, 99, 40, 99, 154, 102, 136, 102, 26, 103, 8, 103, 218, 101, 200, 101, 250, 102, 232, 102, 186, 102, 168, 104, 2, 8, 4, 8, 250, 109, 232, 109, 217, 114, 200, 114, 168, 144, 153, 117, 136, 117, 217, 118, 200, 118, 25, 118, 8, 118, 89, 119, 72, 119, 57, 119, 40, 119, 249, 128, 232, 128, 54, 4, 56, 4, 26, 126, 8, 126, 6, 8, 185, 128, 168, 128, 25, 135, 8, 135, 217, 133, 200, 133, 121, 134, 104, 134, 8, 8, 10, 8, 153, 149, 136, 149, 217, 150, 200, 150, 25, 150, 8, 150, 154, 153, 136, 153, 250, 153, 232, 153, 58, 4, 60, 4, 185, 156, 168, 156, 217, 162, 200, 162, 25, 162, 8, 162, 185, 164, 168, 164, 153, 170, 136, 170, 25, 171, 8, 171, 217, 169, 200, 169, 57, 170, 40, 170, 121, 170, 104, 170, 249, 170, 232, 170, 249, 177, 232, 177, 249, 185, 232, 185, 89, 186, 153, 189, 136, 189, 185, 190, 168, 190, 185, 192, 168, 192, 232, 152, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 12, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 14, 8, 1, 0, 8, 14, 16, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 18, 8, 200, 12, 72, 12, 32, 14, 200, 152, 169, 152, 187, 100, 187, 108, 169, 100, 169, 108, 169, 228, 169, 124, 217, 132, 187, 84, 200, 132, 169, 84, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 68, 255, 5, 186, 124, 255, 5, 185, 132, 255, 5, 186, 152, 255, 5, 185, 188, 255, 5, 168, 56, 255, 5, 168, 68, 255, 5, 168, 124, 255, 5, 168, 132, 255, 5, 168, 152, 255, 5, 168, 188, 255, 5, 217, 70, 255, 5, 200, 70, 255, 5, 249, 58, 255, 5, 232, 58, 255, 5, 218, 126, 255, 5, 200, 126, 255, 5, 218, 154, 255, 5, 200, 154, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 243, 12, 0, 16, 58, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 235, 12, 0, 32, 222, 23, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 227, 23, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 147, 11, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 151, 11, 0, 0, 151, 11, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 156, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 5, 221, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 68, 0, 60, 0, 231, 0, 60, 0, 60, 0, 60, 0, 199, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 69, 0, 60, 0, 100, 0, 104, 0, 60, 0, 60, 0, 60, 0, 68, 0, 104, 0, 60, 0, 60, 0, 60, 0, 68, 0, 72, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 70, 0, 60, 0, 235, 0, 60, 0, 60, 0, 60, 0, 203, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 72, 0, 60, 0, 103, 0, 106, 0, 60, 0, 60, 0, 60, 0, 71, 0, 106, 0, 60, 0, 60, 0, 60, 0, 71, 0, 74, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 77, 0, 60, 0, 108, 0, 108, 0, 60, 0, 60, 0, 60, 0, 76, 0, 108, 0, 60, 0, 60, 0, 60, 0, 76, 0, 76, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 79, 0, 60, 0, 110, 0, 106, 0, 60, 0, 60, 0, 60, 0, 78, 0, 106, 0, 60, 0, 60, 0, 60, 0, 78, 0, 74, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 83, 0, 60, 0, 114, 0, 114, 0, 60, 0, 60, 0, 60, 0, 82, 0, 114, 0, 60, 0, 60, 0, 60, 0, 82, 0, 82, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 84, 0, 60, 0, 115, 0, 104, 0, 60, 0, 60, 0, 60, 0, 83, 0, 104, 0, 60, 0, 60, 0, 60, 0, 83, 0, 72, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 85, 0, 60, 0, 116, 0, 104, 0, 60, 0, 60, 0, 60, 0, 84, 0, 104, 0, 60, 0, 60, 0, 60, 0, 84, 0, 72, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 89, 0, 60, 0, 120, 0, 104, 0, 60, 0, 60, 0, 60, 0, 88, 0, 104, 0, 60, 0, 60, 0, 60, 0, 88, 0, 72, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 183, 1, 60, 0, 122, 0, 104, 0, 60, 0, 60, 0, 60, 0, 90, 0, 104, 0, 60, 0, 60, 0, 60, 0, 90, 0, 72, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 174, 43, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 164, 25, 0, 0, 164, 25, 0, 0, 164, 25, 0, 0, 164, 25, 0, 0, 200, 31, 0, 0, 200, 31, 0, 0, 36, 38, 0, 0, 110, 38, 0, 0, 150, 43, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 141, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 87, 0, 95, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 117, 0, 48, 0, 48, 0, 48, 0, 48, 0, 124, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 111, 0, 48, 0, 48, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 103, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 192, 0, 192, 0, 192, 0, 92, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 47, 0, 202, 1, 48, 0, 202, 34, 48, 0, 202, 67, 48, 0, 202, 100, 48, 0, 202, 133, 48, 0, 202, 166, 48, 0, 202, 199, 48, 0, 202, 232, 48, 0, 202, 9, 49, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 201, 192, 0, 0, 201, 38, 8, 0, 156, 5, 0, 52, 201, 224, 8, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 32, 11, 0, 156, 5, 0, 66, 201, 128, 18, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 201, 0, 21, 0, 201, 224, 29, 0, 201, 160, 37, 0, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 201, 96, 45, 0, 156, 5, 0, 90, 201, 32, 53, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 230, 60, 0, 201, 160, 61, 0, 201, 166, 65, 0, 5, 5, 0, 52, 201, 96, 66, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 224, 67, 0, 5, 5, 0, 66, 201, 160, 71, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 201, 64, 73, 0, 201, 192, 77, 0, 201, 224, 81, 0, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 201, 0, 86, 0, 5, 5, 0, 90, 201, 32, 90, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 47, 0, 134, 5, 21, 47, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 134, 5, 49, 51, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 47, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 47, 0, 5, 5, 21, 47, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 5, 5, 49, 51, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 94, 0, 201, 6, 95, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 95, 0, 201, 134, 96, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 45, 0, 197, 98, 45, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 97, 0, 201, 6, 98, 0, 201, 198, 98, 0, 201, 134, 99, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 100, 0, 201, 6, 101, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 135, 5, 29, 49, 5, 142, 0, 0, 70, 5, 29, 49, 5, 142, 0, 0, 135, 5, 29, 49, 5, 144, 0, 0, 70, 5, 29, 49, 5, 144, 0, 0, 135, 5, 29, 49, 5, 156, 0, 0, 70, 5, 29, 49, 5, 156, 0, 0, 135, 5, 29, 49, 5, 196, 0, 0, 70, 5, 29, 49, 5, 196, 0, 0, 135, 5, 29, 49, 5, 150, 0, 0, 70, 5, 29, 49, 5, 150, 0, 0, 135, 5, 29, 49, 5, 160, 0, 0, 70, 5, 29, 49, 5, 160, 0, 0, 135, 5, 29, 49, 5, 206, 0, 0, 70, 5, 29, 49, 5, 206, 0, 0, 70, 5, 29, 49, 5, 210, 0, 0, 135, 5, 27, 55, 5, 142, 0, 0, 70, 5, 27, 55, 5, 142, 0, 0, 70, 5, 27, 55, 5, 144, 0, 0, 135, 5, 45, 65, 5, 136, 0, 0, 70, 5, 45, 65, 5, 136, 0, 0, 135, 5, 45, 65, 5, 160, 0, 0, 70, 5, 45, 65, 5, 160, 0, 0, 135, 5, 45, 65, 5, 144, 0, 0, 70, 5, 45, 65, 5, 144, 0, 0, 135, 5, 45, 65, 5, 196, 0, 0, 70, 5, 45, 65, 5, 196, 0, 0, 135, 5, 45, 65, 5, 196, 0, 0, 5, 164, 0, 0, 70, 5, 45, 65, 5, 196, 0, 0, 5, 164, 0, 0, 135, 5, 45, 65, 5, 210, 0, 0, 70, 5, 45, 65, 5, 210, 0, 0, 135, 5, 45, 65, 5, 204, 0, 0, 70, 5, 45, 65, 5, 204, 0, 0, 135, 5, 33, 69, 5, 142, 0, 0, 70, 5, 33, 69, 5, 142, 0, 0, 70, 5, 33, 69, 5, 144, 0, 0, 135, 5, 63, 77, 5, 136, 0, 0, 70, 5, 63, 77, 5, 136, 0, 0, 135, 5, 63, 77, 5, 160, 0, 0, 70, 5, 63, 77, 5, 160, 0, 0, 135, 5, 63, 77, 5, 144, 0, 0, 70, 5, 63, 77, 5, 144, 0, 0, 135, 5, 63, 77, 5, 184, 0, 0, 70, 5, 63, 77, 5, 184, 0, 0, 135, 5, 63, 77, 5, 188, 0, 0, 70, 5, 63, 77, 5, 188, 0, 0, 135, 5, 63, 77, 5, 156, 0, 0, 70, 5, 63, 77, 5, 156, 0, 0, 135, 5, 63, 77, 5, 196, 0, 0, 70, 5, 63, 77, 5, 196, 0, 0, 135, 5, 63, 77, 5, 196, 0, 0, 5, 164, 0, 0, 70, 5, 63, 77, 5, 196, 0, 0, 5, 164, 0, 0, 135, 5, 63, 77, 5, 210, 0, 0, 70, 5, 63, 77, 5, 210, 0, 0, 135, 5, 35, 79, 5, 142, 0, 0, 70, 5, 35, 79, 5, 142, 0, 0, 135, 5, 35, 79, 5, 144, 0, 0, 70, 5, 35, 79, 5, 144, 0, 0, 135, 5, 35, 79, 5, 156, 0, 0, 70, 5, 35, 79, 5, 156, 0, 0, 135, 5, 35, 79, 5, 196, 0, 0, 70, 5, 35, 79, 5, 196, 0, 0, 135, 5, 35, 79, 5, 150, 0, 0, 70, 5, 35, 79, 5, 150, 0, 0, 135, 5, 35, 79, 5, 160, 0, 0, 70, 5, 35, 79, 5, 160, 0, 0, 135, 5, 35, 79, 5, 206, 0, 0, 70, 5, 35, 79, 5, 206, 0, 0, 70, 5, 35, 79, 5, 210, 0, 0, 135, 5, 25, 81, 5, 142, 0, 0, 70, 5, 25, 81, 5, 142, 0, 0, 135, 5, 25, 81, 5, 144, 0, 0, 70, 5, 25, 81, 5, 144, 0, 0, 135, 5, 25, 81, 5, 156, 0, 0, 70, 5, 25, 81, 5, 156, 0, 0, 135, 5, 25, 81, 5, 196, 0, 0, 70, 5, 25, 81, 5, 196, 0, 0, 135, 5, 25, 81, 5, 150, 0, 0, 70, 5, 25, 81, 5, 150, 0, 0, 135, 5, 25, 81, 5, 160, 0, 0, 70, 5, 25, 81, 5, 160, 0, 0, 135, 5, 25, 81, 5, 206, 0, 0, 70, 5, 25, 81, 5, 206, 0, 0, 70, 5, 25, 81, 5, 210, 0, 0, 135, 5, 21, 89, 5, 142, 0, 0, 70, 5, 21, 89, 5, 142, 0, 0, 135, 5, 21, 89, 5, 144, 0, 0, 70, 5, 21, 89, 5, 144, 0, 0, 135, 5, 21, 89, 5, 156, 0, 0, 70, 5, 21, 89, 5, 156, 0, 0, 135, 5, 21, 89, 5, 196, 0, 0, 70, 5, 21, 89, 5, 196, 0, 0, 135, 5, 21, 89, 5, 150, 0, 0, 70, 5, 21, 89, 5, 150, 0, 0, 135, 5, 21, 89, 5, 160, 0, 0, 70, 5, 21, 89, 5, 160, 0, 0, 135, 5, 21, 89, 5, 206, 0, 0, 70, 5, 21, 89, 5, 206, 0, 0, 70, 5, 21, 89, 5, 210, 0, 0, 135, 5, 25, 93, 5, 142, 0, 0, 70, 5, 25, 93, 5, 142, 0, 0, 135, 5, 25, 93, 5, 144, 0, 0, 70, 5, 25, 93, 5, 144, 0, 0, 135, 5, 25, 93, 5, 156, 0, 0, 70, 5, 25, 93, 5, 156, 0, 0, 135, 5, 25, 93, 5, 196, 0, 0, 70, 5, 25, 93, 5, 196, 0, 0, 135, 5, 25, 93, 5, 150, 0, 0, 70, 5, 25, 93, 5, 150, 0, 0, 135, 5, 25, 93, 5, 160, 0, 0, 70, 5, 25, 93, 5, 160, 0, 0, 135, 5, 25, 93, 5, 206, 0, 0, 70, 5, 25, 93, 5, 206, 0, 0, 70, 5, 25, 93, 5, 210, 0, 0, 5, 5, 29, 49, 5, 142, 0, 0, 5, 5, 29, 49, 5, 144, 0, 0, 5, 5, 29, 49, 5, 156, 0, 0, 5, 5, 29, 49, 5, 196, 0, 0, 5, 5, 29, 49, 5, 150, 0, 0, 5, 5, 29, 49, 5, 160, 0, 0, 5, 5, 29, 49, 5, 206, 0, 0, 5, 5, 29, 49, 5, 210, 0, 0, 5, 5, 27, 55, 5, 142, 0, 0, 5, 5, 27, 55, 5, 144, 0, 0, 5, 5, 45, 65, 5, 136, 0, 0, 5, 5, 45, 65, 5, 160, 0, 0, 5, 5, 45, 65, 5, 144, 0, 0, 5, 5, 45, 65, 5, 196, 0, 0, 5, 5, 45, 65, 5, 196, 0, 0, 5, 164, 0, 0, 5, 5, 45, 65, 5, 210, 0, 0, 5, 5, 45, 65, 5, 204, 0, 0, 5, 5, 33, 69, 5, 142, 0, 0, 5, 5, 33, 69, 5, 144, 0, 0, 5, 5, 63, 77, 5, 136, 0, 0, 5, 5, 63, 77, 5, 160, 0, 0, 5, 5, 63, 77, 5, 144, 0, 0, 5, 5, 63, 77, 5, 184, 0, 0, 5, 5, 63, 77, 5, 188, 0, 0, 5, 5, 63, 77, 5, 156, 0, 0, 5, 5, 63, 77, 5, 196, 0, 0, 5, 5, 63, 77, 5, 196, 0, 0, 5, 164, 0, 0, 5, 5, 63, 77, 5, 210, 0, 0, 5, 5, 35, 79, 5, 142, 0, 0, 5, 5, 35, 79, 5, 144, 0, 0, 5, 5, 35, 79, 5, 156, 0, 0, 5, 5, 35, 79, 5, 196, 0, 0, 5, 5, 35, 79, 5, 150, 0, 0, 5, 5, 35, 79, 5, 160, 0, 0, 5, 5, 35, 79, 5, 206, 0, 0, 5, 5, 35, 79, 5, 210, 0, 0, 5, 5, 25, 81, 5, 142, 0, 0, 5, 5, 25, 81, 5, 144, 0, 0, 5, 5, 25, 81, 5, 156, 0, 0, 5, 5, 25, 81, 5, 196, 0, 0, 5, 5, 25, 81, 5, 150, 0, 0, 5, 5, 25, 81, 5, 160, 0, 0, 5, 5, 25, 81, 5, 206, 0, 0, 5, 5, 25, 81, 5, 210, 0, 0, 5, 5, 21, 89, 5, 142, 0, 0, 5, 5, 21, 89, 5, 144, 0, 0, 5, 5, 21, 89, 5, 156, 0, 0, 5, 5, 21, 89, 5, 196, 0, 0, 5, 5, 21, 89, 5, 150, 0, 0, 5, 5, 21, 89, 5, 160, 0, 0, 5, 5, 21, 89, 5, 206, 0, 0, 5, 5, 21, 89, 5, 210, 0, 0, 5, 5, 25, 93, 5, 142, 0, 0, 5, 5, 25, 93, 5, 144, 0, 0, 5, 5, 25, 93, 5, 156, 0, 0, 5, 5, 25, 93, 5, 196, 0, 0, 5, 5, 25, 93, 5, 150, 0, 0, 5, 5, 25, 93, 5, 160, 0, 0, 5, 5, 25, 93, 5, 206, 0, 0, 5, 5, 25, 93, 5, 210, 0, 0, 134, 5, 49, 51, 5, 162, 0, 0, 5, 5, 49, 51, 5, 162, 0, 0, 134, 5, 49, 51, 5, 160, 0, 0, 5, 5, 49, 51, 5, 160, 0, 0, 134, 5, 21, 47, 5, 136, 0, 0, 5, 5, 21, 47, 5, 136, 0, 0, 134, 5, 49, 51, 5, 204, 0, 0, 5, 5, 49, 51, 5, 204, 0, 0, 134, 5, 49, 51, 5, 208, 0, 0, 5, 5, 49, 51, 5, 208, 0, 0, 134, 5, 49, 51, 5, 196, 0, 0, 5, 5, 49, 51, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 48, 0, 39, 3, 21, 239, 134, 5, 0, 48, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 2, 192, 197, 162, 41, 30, 2, 192, 197, 226, 42, 30, 3, 192, 197, 34, 43, 30, 3, 192, 197, 98, 150, 30, 3, 192, 197, 162, 36, 30, 1, 192, 197, 162, 37, 30, 1, 192, 197, 226, 38, 30, 2, 192, 197, 34, 39, 30, 2, 192, 197, 98, 30, 2, 12, 0, 30, 2, 0, 192, 197, 162, 31, 2, 0, 192, 197, 226, 34, 30, 1, 192, 197, 34, 35, 30, 1, 192, 197, 98, 72, 0, 29, 241, 135, 5, 104, 0, 29, 241, 70, 5, 36, 1, 197, 162, 37, 1, 0, 192, 197, 98, 0, 50, 156, 5, 48, 0, 8, 3, 49, 243, 134, 5, 0, 54, 156, 5, 4, 0, 74, 0, 27, 247, 135, 5, 106, 0, 27, 247, 70, 5, 52, 1, 3, 192, 197, 226, 53, 1, 4, 192, 197, 34, 240, 1, 4, 192, 197, 98, 0, 64, 156, 5, 15, 0, 54, 30, 26, 0, 58, 30, 12, 0, 58, 30, 7, 192, 197, 98, 59, 30, 7, 192, 197, 162, 60, 30, 7, 192, 197, 226, 61, 30, 8, 192, 197, 34, 54, 30, 6, 192, 197, 34, 55, 30, 6, 192, 197, 98, 56, 30, 6, 192, 197, 163, 57, 30, 7, 192, 197, 3, 59, 1, 12, 0, 59, 1, 5, 192, 197, 34, 60, 1, 5, 192, 197, 98, 61, 1, 5, 192, 197, 162, 62, 1, 5, 192, 197, 226, 76, 0, 255, 255, 45, 65, 135, 5, 108, 0, 255, 255, 45, 65, 70, 5, 57, 1, 4, 192, 197, 162, 58, 1, 4, 192, 197, 226, 0, 68, 156, 5, 4, 0, 74, 0, 255, 255, 33, 69, 135, 5, 106, 0, 255, 255, 33, 69, 70, 5, 52, 1, 8, 192, 197, 98, 53, 1, 8, 192, 197, 162, 240, 1, 8, 192, 197, 226, 0, 76, 156, 5, 19, 0, 18, 2, 32, 0, 91, 30, 15, 0, 91, 30, 12, 192, 197, 98, 92, 30, 12, 192, 197, 163, 93, 30, 13, 192, 197, 3, 94, 30, 13, 192, 197, 98, 95, 30, 13, 192, 197, 162, 18, 2, 11, 192, 197, 34, 19, 2, 11, 192, 197, 98, 88, 30, 11, 192, 197, 162, 89, 30, 11, 192, 197, 226, 90, 30, 12, 192, 197, 34, 87, 1, 15, 0, 87, 1, 9, 192, 197, 226, 88, 1, 10, 192, 197, 34, 89, 1, 10, 192, 197, 98, 16, 2, 10, 192, 197, 162, 17, 2, 10, 192, 197, 226, 82, 0, 255, 255, 63, 77, 135, 5, 114, 0, 255, 255, 63, 77, 70, 5, 84, 1, 9, 192, 197, 34, 85, 1, 9, 192, 197, 98, 86, 1, 9, 192, 197, 162, 0, 78, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 16, 192, 197, 98, 41, 30, 16, 192, 197, 162, 42, 30, 16, 192, 197, 226, 43, 30, 17, 192, 197, 34, 150, 30, 17, 192, 197, 98, 36, 30, 15, 192, 197, 98, 37, 30, 15, 192, 197, 162, 38, 30, 15, 192, 197, 226, 39, 30, 16, 192, 197, 34, 30, 2, 12, 0, 30, 2, 14, 192, 197, 98, 31, 2, 14, 192, 197, 162, 34, 30, 14, 192, 197, 226, 35, 30, 15, 192, 197, 34, 72, 0, 255, 255, 35, 79, 135, 5, 104, 0, 255, 255, 35, 79, 70, 5, 36, 1, 13, 192, 197, 226, 37, 1, 14, 192, 197, 34, 0, 80, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 20, 192, 197, 34, 41, 30, 20, 192, 197, 98, 42, 30, 20, 192, 197, 162, 43, 30, 20, 192, 197, 226, 150, 30, 21, 192, 197, 34, 36, 30, 19, 192, 197, 34, 37, 30, 19, 192, 197, 98, 38, 30, 19, 192, 197, 162, 39, 30, 19, 192, 197, 226, 30, 2, 12, 0, 30, 2, 18, 192, 197, 34, 31, 2, 18, 192, 197, 98, 34, 30, 18, 192, 197, 162, 35, 30, 18, 192, 197, 226, 72, 0, 255, 255, 25, 81, 135, 5, 104, 0, 255, 255, 25, 81, 70, 5, 36, 1, 17, 192, 197, 162, 37, 1, 17, 192, 197, 226, 0, 88, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 23, 192, 197, 226, 41, 30, 24, 192, 197, 34, 42, 30, 24, 192, 197, 98, 43, 30, 24, 192, 197, 162, 150, 30, 24, 192, 197, 226, 36, 30, 22, 192, 197, 226, 37, 30, 23, 192, 197, 34, 38, 30, 23, 192, 197, 98, 39, 30, 23, 192, 197, 162, 30, 2, 12, 0, 30, 2, 21, 192, 197, 226, 31, 2, 22, 192, 197, 34, 34, 30, 22, 192, 197, 98, 35, 30, 22, 192, 197, 162, 72, 0, 255, 255, 21, 89, 135, 5, 104, 0, 255, 255, 21, 89, 70, 5, 36, 1, 21, 192, 197, 98, 37, 1, 21, 192, 197, 162, 0, 92, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 27, 192, 197, 162, 41, 30, 27, 192, 197, 226, 42, 30, 28, 192, 197, 34, 43, 30, 28, 192, 197, 98, 150, 30, 28, 192, 197, 162, 36, 30, 26, 192, 197, 162, 37, 30, 26, 192, 197, 226, 38, 30, 27, 192, 197, 34, 39, 30, 27, 192, 197, 98, 30, 2, 12, 0, 30, 2, 25, 192, 197, 162, 31, 2, 25, 192, 197, 226, 34, 30, 26, 192, 197, 34, 35, 30, 26, 192, 197, 98, 72, 0, 255, 255, 25, 93, 135, 5, 104, 0, 255, 255, 25, 93, 70, 5, 36, 1, 25, 192, 197, 34, 37, 1, 25, 192, 197, 98, 0, 46, 5, 5, 48, 0, 39, 3, 21, 239, 5, 5, 0, 48, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 29, 192, 197, 162, 39, 30, 29, 192, 197, 226, 41, 30, 30, 192, 197, 34, 43, 30, 30, 192, 197, 98, 150, 30, 30, 192, 197, 162, 104, 0, 29, 241, 5, 5, 37, 1, 28, 192, 197, 226, 31, 2, 29, 192, 197, 34, 35, 30, 29, 192, 197, 98, 0, 50, 5, 5, 48, 0, 8, 3, 49, 243, 5, 5, 0, 54, 5, 5, 2, 0, 106, 0, 27, 247, 5, 5, 53, 1, 30, 192, 197, 226, 240, 1, 31, 192, 197, 34, 0, 64, 5, 5, 7, 0, 55, 30, 12, 0, 55, 30, 32, 192, 197, 34, 57, 30, 32, 192, 197, 99, 59, 30, 32, 192, 197, 194, 61, 30, 33, 192, 197, 2, 108, 0, 255, 255, 45, 65, 5, 5, 58, 1, 31, 192, 197, 98, 60, 1, 31, 192, 197, 162, 62, 1, 31, 192, 197, 226, 0, 68, 5, 5, 2, 0, 106, 0, 255, 255, 33, 69, 5, 5, 53, 1, 33, 192, 197, 66, 240, 1, 33, 192, 197, 130, 0, 76, 5, 5, 9, 0, 19, 2, 15, 0, 19, 2, 34, 192, 197, 194, 89, 30, 35, 192, 197, 2, 91, 30, 35, 192, 197, 66, 93, 30, 35, 192, 197, 131, 95, 30, 35, 192, 197, 226, 114, 0, 255, 255, 63, 77, 5, 5, 85, 1, 33, 192, 197, 194, 87, 1, 34, 192, 197, 2, 89, 1, 34, 192, 197, 66, 17, 2, 34, 192, 197, 130, 0, 78, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 36, 192, 197, 226, 39, 30, 37, 192, 197, 34, 41, 30, 37, 192, 197, 98, 43, 30, 37, 192, 197, 162, 150, 30, 37, 192, 197, 226, 104, 0, 255, 255, 35, 79, 5, 5, 37, 1, 36, 192, 197, 34, 31, 2, 36, 192, 197, 98, 35, 30, 36, 192, 197, 162, 0, 80, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 38, 192, 197, 226, 39, 30, 39, 192, 197, 34, 41, 30, 39, 192, 197, 98, 43, 30, 39, 192, 197, 162, 150, 30, 39, 192, 197, 226, 104, 0, 255, 255, 25, 81, 5, 5, 37, 1, 38, 192, 197, 34, 31, 2, 38, 192, 197, 98, 35, 30, 38, 192, 197, 162, 0, 88, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 40, 192, 197, 226, 39, 30, 41, 192, 197, 34, 41, 30, 41, 192, 197, 98, 43, 30, 41, 192, 197, 162, 150, 30, 41, 192, 197, 226, 104, 0, 255, 255, 21, 89, 5, 5, 37, 1, 40, 192, 197, 34, 31, 2, 40, 192, 197, 98, 35, 30, 40, 192, 197, 162, 0, 92, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 42, 192, 197, 226, 39, 30, 43, 192, 197, 34, 41, 30, 43, 192, 197, 98, 43, 30, 43, 192, 197, 162, 150, 30, 43, 192, 197, 226, 104, 0, 255, 255, 25, 93, 5, 5, 37, 1, 42, 192, 197, 34, 31, 2, 42, 192, 197, 98, 35, 30, 42, 192, 197, 162, 156, 50, 196, 162, 48, 0, 8, 3, 44, 192, 197, 34, 5, 50, 196, 162, 48, 0, 8, 3, 44, 192, 197, 98, 156, 50, 196, 160, 48, 0, 8, 3, 44, 192, 197, 162, 5, 50, 196, 160, 48, 0, 8, 3, 44, 192, 197, 226, 156, 50, 196, 204, 48, 0, 8, 3, 45, 192, 197, 162, 5, 50, 196, 204, 48, 0, 8, 3, 45, 192, 197, 226, 156, 50, 196, 208, 48, 0, 8, 3, 46, 192, 197, 34, 5, 50, 196, 208, 48, 0, 8, 3, 46, 192, 197, 98, 156, 50, 196, 196, 48, 0, 8, 3, 46, 192, 197, 162, 5, 50, 196, 196, 48, 0, 8, 3, 46, 192, 197, 226, 36, 0, 72, 0, 73, 0, 74, 0, 75, 0, 82, 0, 83, 0, 104, 0, 105, 0, 106, 0, 107, 0, 108, 0, 109, 0, 114, 0, 115, 0, 36, 1, 38, 1, 52, 1, 54, 1, 57, 1, 63, 1, 84, 1, 90, 1, 240, 1, 241, 1, 16, 2, 20, 2, 30, 2, 32, 2, 34, 30, 44, 30, 54, 30, 62, 30, 88, 30, 96, 30, 150, 30, 151, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 28, 4, 38, 4, 185, 88, 40, 4, 185, 100, 186, 104, 186, 112, 185, 116, 50, 4, 185, 128, 68, 4, 185, 144, 185, 148, 185, 152, 78, 4, 96, 4, 106, 4, 185, 188, 185, 192, 185, 196, 116, 4, 185, 208, 126, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 136, 4, 138, 4, 144, 4, 168, 88, 146, 4, 168, 100, 168, 104, 168, 112, 168, 116, 152, 4, 168, 128, 162, 4, 168, 144, 168, 148, 168, 152, 168, 4, 178, 4, 184, 4, 168, 188, 168, 192, 168, 196, 190, 4, 168, 208, 196, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 144, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 185, 68, 186, 81, 154, 81, 250, 81, 185, 84, 186, 105, 154, 105, 250, 105, 90, 106, 217, 72, 154, 134, 185, 145, 153, 145, 249, 145, 153, 146, 89, 146, 136, 14, 217, 146, 185, 189, 153, 189, 249, 189, 89, 190, 153, 209, 185, 220, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 168, 68, 168, 81, 136, 81, 232, 81, 168, 84, 168, 105, 136, 105, 232, 105, 72, 106, 200, 72, 136, 134, 168, 145, 136, 145, 232, 145, 136, 146, 72, 146, 128, 14, 200, 146, 168, 189, 136, 189, 232, 189, 72, 190, 136, 209, 168, 220, 72, 210, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 74, 8, 74, 89, 75, 72, 75, 58, 83, 40, 83, 218, 81, 200, 81, 186, 82, 168, 82, 202, 4, 204, 4, 26, 82, 8, 82, 249, 93, 232, 93, 217, 93, 200, 93, 185, 94, 168, 94, 249, 94, 232, 94, 249, 101, 232, 101, 89, 103, 72, 103, 154, 106, 136, 106, 58, 107, 40, 107, 218, 105, 200, 105, 26, 107, 8, 107, 186, 106, 168, 108, 6, 8, 8, 8, 250, 113, 232, 113, 249, 118, 232, 118, 168, 156, 153, 121, 136, 121, 249, 122, 232, 122, 25, 122, 8, 122, 121, 123, 104, 123, 89, 123, 72, 123, 154, 133, 136, 133, 250, 134, 232, 134, 26, 134, 8, 134, 10, 8, 185, 136, 168, 136, 57, 147, 40, 147, 217, 145, 200, 145, 121, 146, 104, 146, 12, 8, 14, 8, 153, 161, 136, 161, 249, 162, 232, 162, 25, 162, 8, 162, 154, 169, 136, 169, 250, 169, 232, 169, 250, 170, 232, 170, 26, 170, 8, 170, 249, 178, 232, 178, 25, 178, 8, 178, 185, 180, 168, 180, 153, 190, 136, 190, 57, 191, 40, 191, 217, 189, 200, 189, 57, 190, 40, 190, 121, 190, 104, 190, 25, 191, 8, 191, 249, 197, 232, 197, 249, 209, 232, 209, 89, 210, 153, 213, 136, 213, 185, 214, 168, 214, 25, 214, 8, 214, 232, 168, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 80, 200, 168, 169, 168, 200, 56, 169, 80, 187, 104, 187, 112, 169, 104, 169, 112, 169, 224, 169, 132, 217, 144, 187, 80, 200, 144, 169, 80, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 185, 72, 72, 4, 186, 76, 104, 4, 177, 76, 36, 5, 250, 77, 37, 5, 241, 77, 255, 5, 186, 80, 255, 5, 185, 92, 74, 4, 186, 96, 106, 4, 177, 96, 52, 5, 250, 97, 53, 5, 241, 97, 255, 5, 185, 120, 76, 4, 186, 124, 108, 4, 177, 124, 57, 5, 154, 125, 58, 5, 145, 125, 59, 5, 250, 126, 60, 5, 241, 126, 61, 5, 26, 126, 62, 5, 17, 126, 255, 5, 186, 132, 74, 4, 186, 140, 106, 4, 177, 140, 52, 5, 250, 141, 53, 5, 241, 141, 255, 5, 185, 160, 82, 4, 186, 164, 114, 4, 177, 164, 84, 5, 154, 165, 85, 5, 145, 165, 86, 5, 250, 166, 87, 5, 241, 166, 88, 5, 26, 166, 89, 5, 17, 166, 255, 5, 186, 168, 72, 4, 186, 172, 104, 4, 177, 172, 36, 5, 250, 173, 37, 5, 241, 173, 255, 5, 185, 176, 72, 4, 186, 184, 104, 4, 177, 184, 36, 5, 250, 185, 37, 5, 241, 185, 255, 5, 185, 200, 72, 4, 186, 204, 104, 4, 177, 204, 36, 5, 250, 205, 37, 5, 241, 205, 255, 5, 185, 212, 72, 4, 186, 216, 104, 4, 177, 216, 36, 5, 250, 217, 37, 5, 241, 217, 255, 5, 168, 64, 255, 5, 168, 72, 104, 4, 168, 76, 37, 5, 232, 77, 255, 5, 168, 80, 255, 5, 168, 92, 106, 4, 168, 96, 53, 5, 232, 97, 255, 5, 168, 120, 108, 4, 168, 124, 58, 5, 136, 125, 60, 5, 232, 126, 62, 5, 8, 126, 255, 5, 168, 132, 106, 4, 168, 140, 53, 5, 232, 141, 255, 5, 168, 160, 114, 4, 168, 164, 85, 5, 136, 165, 87, 5, 232, 166, 89, 5, 8, 166, 255, 5, 168, 168, 104, 4, 168, 172, 37, 5, 232, 173, 255, 5, 168, 176, 104, 4, 168, 184, 37, 5, 232, 185, 255, 5, 168, 200, 104, 4, 168, 204, 37, 5, 232, 205, 255, 5, 168, 212, 104, 4, 168, 216, 37, 5, 232, 217, 255, 5, 26, 83, 255, 5, 8, 83, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 159, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 140, 11, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 145, 11, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 168, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 172, 6, 0, 0, 172, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 46, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 41, 220, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 93, 0, 91, 0, 115, 0, 117, 0, 112, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 67, 0, 111, 0, 110, 0, 116, 0, 114, 0, 97, 0, 99, 0, 116, 0, 105, 0, 111, 0, 110, 0, 115, 0, 32, 0, 91, 0, 24, 4, 56, 4, 93, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 196, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 56, 0, 0, 0, 56, 1, 0, 0, 88, 25, 0, 0, 88, 25, 0, 0, 88, 25, 0, 0, 88, 25, 0, 0, 172, 25, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 39, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 121, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 104, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 1, 0, 202, 129, 1, 0, 202, 162, 1, 0, 202, 195, 1, 0, 202, 228, 1, 0, 202, 5, 2, 0, 202, 38, 2, 0, 202, 71, 2, 0, 202, 104, 2, 0, 202, 137, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 174, 5, 70, 97, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 97, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 174, 5, 70, 97, 5, 140, 0, 0, 5, 5, 70, 97, 5, 140, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 69, 0, 60, 0, 32, 0, 47, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 161, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 166, 6, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 66, 0, 65, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 136, 19, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 140, 19, 0, 0, 140, 19, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 44, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 104, 0, 114, 0, 93, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 104, 0, 114, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 252, 40, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 113, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 180, 35, 0, 0, 180, 35, 0, 0, 180, 35, 0, 0, 180, 35, 0, 0, 172, 38, 0, 0, 172, 38, 0, 0, 198, 40, 0, 0, 228, 40, 0, 0, 228, 40, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 46, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 89, 0, 97, 0, 48, 0, 56, 0, 105, 0, 113, 0, 115, 0, 123, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 122, 0, 130, 0, 136, 0, 144, 0, 152, 0, 160, 0, 48, 0, 56, 0, 159, 0, 167, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 184, 0, 192, 0, 48, 0, 48, 0, 48, 0, 200, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 208, 0, 48, 0, 48, 0, 216, 0, 48, 0, 222, 0, 230, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 233, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 240, 0, 48, 0, 48, 0, 248, 0, 252, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 4, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 7, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 182, 0, 48, 0, 48, 0, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 37, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 15, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 23, 1, 48, 0, 48, 0, 29, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 100, 1, 192, 0, 164, 1, 204, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 232, 1, 32, 2, 96, 2, 192, 0, 124, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 128, 22, 0, 202, 161, 22, 0, 202, 194, 22, 0, 202, 227, 22, 0, 202, 4, 23, 0, 202, 37, 23, 0, 202, 70, 23, 0, 202, 103, 23, 0, 202, 136, 23, 0, 202, 169, 23, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 198, 0, 0, 201, 68, 2, 0, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 224, 4, 0, 156, 5, 0, 66, 201, 64, 7, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 198, 9, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 166, 10, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 134, 11, 0, 201, 4, 13, 0, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 128, 14, 0, 5, 5, 0, 66, 201, 32, 16, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 198, 17, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 166, 18, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 13, 0, 201, 134, 19, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 13, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 13, 0, 201, 6, 21, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 47, 5, 5, 3, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 49, 5, 5, 3, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 22, 0, 201, 70, 23, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 138, 5, 2, 49, 72, 5, 2, 49, 6, 5, 2, 49, 138, 5, 2, 65, 72, 5, 2, 65, 6, 5, 2, 65, 138, 5, 2, 69, 72, 5, 2, 69, 6, 5, 2, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 24, 0, 201, 198, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 134, 25, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 27, 0, 5, 71, 238, 102, 201, 70, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 5, 0, 192, 0, 0, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 5, 0, 192, 0, 0, 0, 197, 34, 6, 0, 192, 0, 0, 0, 197, 98, 6, 0, 192, 0, 0, 0, 197, 162, 6, 0, 192, 0, 0, 0, 197, 226, 6, 0, 197, 35, 7, 0, 197, 130, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 7, 0, 197, 3, 8, 0, 197, 98, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 162, 4, 0, 197, 226, 8, 0, 5, 5, 10, 121, 197, 34, 9, 0, 197, 98, 9, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 226, 10, 0, 197, 34, 11, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 5, 0, 5, 5, 24, 121, 197, 98, 5, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 2, 0, 197, 34, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 29, 0, 201, 230, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 11, 0, 197, 34, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 30, 0, 201, 102, 31, 0, 201, 38, 32, 0, 201, 230, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 137, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 144, 0, 0, 137, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 144, 0, 0, 5, 5, 2, 65, 5, 142, 0, 0, 5, 5, 2, 65, 5, 144, 0, 0, 5, 5, 2, 69, 5, 142, 0, 0, 5, 5, 2, 69, 5, 144, 0, 0, 134, 5, 3, 47, 5, 160, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 3, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 162, 4, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 226, 4, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 34, 5, 0, 5, 5, 24, 121, 197, 98, 5, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 162, 5, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 226, 5, 0, 5, 5, 107, 121, 197, 34, 6, 0, 5, 5, 111, 121, 197, 98, 6, 0, 5, 5, 115, 121, 197, 162, 6, 0, 5, 5, 119, 121, 197, 226, 6, 0, 197, 35, 7, 0, 197, 130, 7, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 194, 7, 0, 197, 3, 8, 0, 197, 98, 8, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 162, 8, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 162, 4, 0, 197, 226, 8, 0, 5, 5, 10, 121, 197, 34, 9, 0, 197, 98, 9, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 162, 9, 0, 197, 226, 9, 0, 197, 34, 10, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 226, 10, 0, 197, 34, 11, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 66, 5, 0, 5, 5, 24, 121, 197, 98, 5, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 46, 156, 5, 2, 0, 1, 3, 3, 239, 134, 5, 12, 3, 2, 239, 134, 5, 65, 3, 3, 239, 134, 5, 0, 48, 156, 5, 3, 0, 90, 0, 8, 0, 122, 0, 10, 0, 125, 1, 2, 241, 137, 5, 126, 1, 2, 241, 71, 5, 48, 0, 12, 3, 2, 241, 137, 5, 48, 0, 12, 3, 2, 241, 71, 5, 0, 64, 156, 5, 4, 0, 74, 0, 255, 255, 2, 65, 137, 5, 106, 0, 255, 255, 2, 65, 71, 5, 52, 1, 197, 162, 53, 1, 0, 192, 197, 98, 240, 1, 0, 192, 197, 162, 0, 68, 156, 5, 4, 0, 74, 0, 255, 255, 2, 69, 137, 5, 106, 0, 255, 255, 2, 69, 71, 5, 52, 1, 0, 192, 197, 226, 53, 1, 1, 192, 197, 34, 240, 1, 1, 192, 197, 98, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 46, 5, 5, 2, 0, 1, 3, 3, 239, 5, 5, 12, 3, 2, 239, 5, 5, 65, 3, 3, 239, 5, 5, 0, 48, 5, 5, 1, 0, 122, 0, 3, 0, 126, 1, 2, 241, 5, 5, 48, 0, 12, 3, 2, 241, 5, 5, 0, 64, 5, 5, 2, 0, 106, 0, 255, 255, 2, 65, 5, 5, 53, 1, 1, 192, 197, 162, 240, 1, 1, 192, 197, 226, 0, 68, 5, 5, 2, 0, 106, 0, 255, 255, 2, 69, 5, 5, 53, 1, 2, 192, 197, 34, 240, 1, 2, 192, 197, 98, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 2, 0, 1, 3, 2, 192, 197, 162, 12, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 5, 46, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 12, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 156, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 226, 156, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 98, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 78, 196, 196, 48, 0, 12, 3, 11, 192, 197, 98, 5, 78, 196, 196, 48, 0, 12, 3, 11, 192, 197, 162, 156, 92, 196, 196, 48, 0, 12, 3, 12, 192, 197, 98, 5, 92, 196, 196, 48, 0, 12, 3, 12, 192, 197, 162, 156, 92, 196, 210, 48, 0, 12, 3, 12, 192, 197, 226, 5, 92, 196, 210, 48, 0, 12, 3, 13, 192, 197, 34, 14, 0, 74, 0, 75, 0, 90, 0, 91, 0, 106, 0, 107, 0, 122, 0, 123, 0, 52, 1, 54, 1, 125, 1, 127, 1, 240, 1, 241, 1, 3, 0, 67, 0, 58, 0, 32, 0, 47, 0, 0, 16, 18, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 28, 36, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 76, 0, 0, 0, 76, 1, 0, 0, 188, 28, 0, 0, 188, 28, 0, 0, 188, 28, 0, 0, 188, 28, 0, 0, 208, 29, 0, 0, 208, 29, 0, 0, 164, 31, 0, 0, 194, 31, 0, 0, 4, 36, 0, 0, 25, 0, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 94, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 174, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 89, 0, 97, 0, 48, 0, 56, 0, 105, 0, 113, 0, 115, 0, 123, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 145, 0, 48, 0, 48, 0, 153, 0, 157, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 165, 0, 139, 0, 48, 0, 48, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 131, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 100, 1, 192, 0, 164, 1, 204, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 7, 0, 202, 129, 7, 0, 202, 162, 7, 0, 202, 195, 7, 0, 202, 228, 7, 0, 202, 5, 8, 0, 202, 38, 8, 0, 202, 71, 8, 0, 202, 104, 8, 0, 202, 137, 8, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 201, 132, 1, 0, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 201, 32, 4, 0, 156, 5, 0, 66, 201, 128, 6, 0, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 6, 9, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 201, 230, 9, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 198, 10, 0, 201, 68, 12, 0, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 201, 192, 13, 0, 5, 5, 0, 66, 201, 96, 15, 0, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 6, 17, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 201, 230, 17, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 6, 0, 201, 198, 18, 0, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 6, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 7, 0, 201, 70, 20, 0, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 47, 5, 5, 3, 47, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 47, 5, 5, 2, 47, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 49, 5, 5, 3, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 21, 0, 201, 134, 22, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 93, 5, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 138, 5, 2, 49, 72, 5, 2, 49, 6, 5, 2, 49, 138, 5, 2, 65, 72, 5, 2, 65, 6, 5, 2, 65, 138, 5, 2, 69, 72, 5, 2, 69, 6, 5, 2, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 23, 0, 201, 6, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 2, 0, 197, 34, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 24, 0, 201, 134, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 5, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 26, 0, 201, 6, 27, 0, 201, 198, 27, 0, 201, 134, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 137, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 142, 0, 0, 71, 5, 2, 65, 5, 144, 0, 0, 137, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 142, 0, 0, 71, 5, 2, 69, 5, 144, 0, 0, 5, 5, 2, 65, 5, 142, 0, 0, 5, 5, 2, 65, 5, 144, 0, 0, 5, 5, 2, 69, 5, 142, 0, 0, 5, 5, 2, 69, 5, 144, 0, 0, 134, 5, 3, 47, 5, 160, 0, 0, 134, 5, 2, 47, 5, 160, 0, 0, 5, 5, 3, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 79, 5, 202, 0, 0, 5, 5, 2, 79, 5, 202, 0, 0, 134, 5, 2, 79, 5, 196, 0, 0, 5, 5, 2, 79, 5, 196, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 93, 5, 196, 0, 0, 5, 5, 2, 93, 5, 196, 0, 0, 134, 5, 2, 93, 5, 210, 0, 0, 5, 5, 2, 93, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 2, 0, 1, 3, 3, 239, 134, 5, 12, 3, 2, 239, 134, 5, 65, 3, 3, 239, 134, 5, 0, 48, 156, 5, 3, 0, 90, 0, 8, 0, 122, 0, 10, 0, 125, 1, 2, 241, 137, 5, 126, 1, 2, 241, 71, 5, 48, 0, 12, 3, 2, 241, 137, 5, 48, 0, 12, 3, 2, 241, 71, 5, 0, 64, 156, 5, 4, 0, 74, 0, 255, 255, 2, 65, 137, 5, 106, 0, 255, 255, 2, 65, 71, 5, 52, 1, 197, 162, 53, 1, 0, 192, 197, 98, 240, 1, 0, 192, 197, 162, 0, 68, 156, 5, 4, 0, 74, 0, 255, 255, 2, 69, 137, 5, 106, 0, 255, 255, 2, 69, 71, 5, 52, 1, 0, 192, 197, 226, 53, 1, 1, 192, 197, 34, 240, 1, 1, 192, 197, 98, 0, 78, 156, 5, 48, 0, 12, 3, 255, 255, 2, 79, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 255, 255, 2, 93, 134, 5, 0, 46, 5, 5, 2, 0, 1, 3, 3, 239, 5, 5, 12, 3, 2, 239, 5, 5, 65, 3, 3, 239, 5, 5, 0, 48, 5, 5, 1, 0, 122, 0, 3, 0, 126, 1, 2, 241, 5, 5, 48, 0, 12, 3, 2, 241, 5, 5, 0, 64, 5, 5, 2, 0, 106, 0, 255, 255, 2, 65, 5, 5, 53, 1, 1, 192, 197, 162, 240, 1, 1, 192, 197, 226, 0, 68, 5, 5, 2, 0, 106, 0, 255, 255, 2, 69, 5, 5, 53, 1, 2, 192, 197, 34, 240, 1, 2, 192, 197, 98, 0, 78, 5, 5, 48, 0, 12, 3, 255, 255, 2, 79, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 255, 255, 2, 93, 5, 5, 156, 46, 196, 160, 2, 0, 1, 3, 2, 192, 197, 162, 12, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 162, 5, 46, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 12, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 156, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 162, 5, 78, 196, 160, 48, 0, 12, 3, 3, 192, 197, 226, 156, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 34, 5, 78, 196, 202, 48, 0, 12, 3, 4, 192, 197, 98, 156, 78, 196, 196, 48, 0, 12, 3, 4, 192, 197, 162, 5, 78, 196, 196, 48, 0, 12, 3, 4, 192, 197, 226, 156, 92, 196, 196, 48, 0, 12, 3, 5, 192, 197, 162, 5, 92, 196, 196, 48, 0, 12, 3, 5, 192, 197, 226, 156, 92, 196, 210, 48, 0, 12, 3, 6, 192, 197, 34, 5, 92, 196, 210, 48, 0, 12, 3, 6, 192, 197, 98, 14, 0, 74, 0, 75, 0, 90, 0, 91, 0, 106, 0, 107, 0, 122, 0, 123, 0, 52, 1, 54, 1, 125, 1, 127, 1, 240, 1, 241, 1, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 28, 4, 186, 88, 185, 92, 185, 96, 185, 100, 186, 104, 186, 112, 185, 116, 36, 4, 185, 128, 46, 4, 185, 144, 185, 148, 185, 152, 185, 160, 56, 4, 185, 172, 185, 180, 185, 184, 185, 188, 185, 192, 185, 196, 58, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 60, 4, 62, 4, 168, 88, 168, 92, 168, 96, 168, 100, 168, 104, 168, 112, 168, 116, 67, 4, 168, 128, 73, 4, 168, 144, 168, 148, 168, 152, 168, 160, 79, 4, 168, 172, 168, 180, 168, 184, 168, 188, 168, 192, 168, 196, 81, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 144, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 83, 4, 186, 89, 154, 89, 250, 89, 90, 90, 186, 105, 154, 105, 250, 105, 90, 106, 217, 76, 154, 134, 185, 145, 153, 145, 249, 145, 153, 146, 89, 146, 136, 14, 217, 146, 185, 181, 153, 181, 249, 181, 89, 182, 153, 197, 185, 208, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 85, 4, 168, 89, 136, 89, 232, 89, 72, 90, 168, 105, 136, 105, 232, 105, 72, 106, 200, 76, 136, 134, 168, 145, 136, 145, 232, 145, 136, 146, 72, 146, 128, 14, 200, 146, 168, 181, 136, 181, 232, 181, 72, 182, 136, 197, 168, 208, 72, 198, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 185, 72, 168, 72, 249, 65, 232, 65, 185, 66, 168, 66, 185, 68, 168, 68, 25, 78, 8, 78, 185, 84, 168, 84, 58, 91, 40, 91, 218, 89, 200, 89, 186, 90, 168, 90, 26, 91, 8, 91, 26, 90, 8, 90, 249, 97, 232, 97, 217, 97, 200, 97, 185, 98, 168, 98, 249, 98, 232, 98, 249, 101, 232, 101, 89, 103, 72, 103, 154, 106, 136, 106, 58, 107, 40, 107, 218, 105, 200, 105, 26, 107, 8, 107, 186, 106, 168, 108, 6, 8, 8, 8, 250, 113, 232, 113, 249, 118, 232, 118, 168, 156, 153, 121, 136, 121, 249, 122, 232, 122, 25, 122, 8, 122, 121, 123, 104, 123, 89, 123, 72, 123, 154, 133, 136, 133, 250, 134, 232, 134, 26, 134, 8, 134, 10, 8, 185, 140, 168, 140, 57, 147, 40, 147, 217, 145, 200, 145, 121, 146, 104, 146, 12, 8, 14, 8, 153, 161, 136, 161, 249, 162, 232, 162, 25, 162, 8, 162, 154, 165, 136, 165, 250, 165, 232, 165, 87, 4, 89, 4, 185, 168, 168, 168, 249, 174, 232, 174, 25, 174, 8, 174, 185, 176, 168, 176, 153, 182, 136, 182, 57, 183, 40, 183, 217, 181, 200, 181, 57, 182, 40, 182, 121, 182, 104, 182, 25, 183, 8, 183, 249, 189, 232, 189, 249, 197, 232, 197, 89, 198, 153, 201, 136, 201, 185, 202, 168, 202, 185, 204, 168, 204, 232, 164, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 88, 200, 164, 169, 164, 200, 56, 169, 88, 187, 104, 187, 112, 169, 104, 169, 112, 169, 212, 169, 132, 217, 144, 187, 88, 200, 144, 169, 88, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 185, 76, 90, 2, 122, 2, 125, 5, 186, 80, 126, 5, 177, 80, 255, 5, 185, 120, 74, 4, 186, 124, 106, 4, 177, 124, 52, 5, 250, 125, 53, 5, 241, 125, 255, 5, 186, 132, 74, 4, 186, 136, 106, 4, 177, 136, 52, 5, 250, 137, 53, 5, 241, 137, 255, 5, 186, 164, 255, 5, 185, 200, 255, 5, 168, 64, 255, 5, 168, 76, 122, 2, 126, 5, 168, 80, 255, 5, 168, 120, 106, 4, 168, 124, 53, 5, 232, 125, 255, 5, 168, 132, 106, 4, 168, 136, 53, 5, 232, 137, 255, 5, 168, 164, 255, 5, 168, 200, 255, 5, 249, 66, 255, 5, 232, 66, 255, 5, 250, 166, 255, 5, 232, 166, 255, 1, 3, 0, 67, 0, 58, 0, 32, 0, 119, 10, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 111, 10, 0, 32, 127, 19, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 132, 19, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 77, 0, 69, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 83, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 197, 34, 0, 32, 7, 0, 0, 0, 27, 0, 0, 0, 201, 34, 0, 0, 201, 34, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 195, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 100, 101, 102, 97, 117, 108, 116, 0, 114, 101, 102, 111, 114, 109, 101, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 101, 97, 114, 99, 104, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 114, 0, 101, 0, 102, 0, 111, 0, 114, 0, 109, 0, 101, 0, 100, 0, 0, 0, 49, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 115, 0, 118, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 0, 0, 126, 220, 38, 0, 68, 0, 60, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 60, 0, 60, 0, 240, 0, 60, 0, 60, 0, 60, 0, 208, 0, 38, 0, 116, 0, 60, 0, 60, 0, 60, 0, 254, 0, 47, 0, 104, 0, 38, 0, 84, 0, 60, 0, 60, 0, 60, 0, 222, 0, 47, 0, 72, 0, 38, 0, 89, 0, 60, 0, 60, 0, 117, 0, 8, 3, 60, 0, 60, 0, 60, 0, 85, 0, 8, 3, 60, 0, 60, 0, 117, 0, 11, 3, 60, 0, 60, 0, 60, 0, 85, 0, 11, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 192, 1, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 60, 0, 101, 0, 40, 3, 60, 0, 60, 0, 60, 0, 69, 0, 40, 3, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 60, 0, 60, 0, 81, 1, 60, 0, 60, 0, 60, 0, 80, 1, 60, 0, 60, 0, 83, 1, 60, 0, 60, 0, 60, 0, 82, 1, 60, 0, 60, 0, 244, 0, 60, 0, 60, 0, 60, 0, 212, 0, 0, 0, 139, 220, 38, 0, 68, 0, 60, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 60, 0, 60, 0, 240, 0, 60, 0, 60, 0, 60, 0, 208, 0, 38, 0, 116, 0, 60, 0, 60, 0, 60, 0, 254, 0, 47, 0, 104, 0, 38, 0, 84, 0, 60, 0, 60, 0, 60, 0, 222, 0, 47, 0, 72, 0, 38, 0, 118, 0, 60, 0, 60, 0, 60, 0, 86, 0, 60, 0, 60, 0, 119, 0, 60, 0, 60, 0, 60, 0, 87, 0, 38, 0, 89, 0, 60, 0, 60, 0, 117, 0, 8, 3, 60, 0, 60, 0, 60, 0, 85, 0, 8, 3, 60, 0, 60, 0, 117, 0, 11, 3, 60, 0, 60, 0, 60, 0, 85, 0, 11, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 192, 1, 60, 0, 229, 0, 60, 0, 60, 0, 60, 0, 197, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 60, 0, 60, 0, 230, 0, 60, 0, 60, 0, 60, 0, 198, 0, 60, 0, 60, 0, 101, 0, 40, 3, 60, 0, 60, 0, 60, 0, 69, 0, 40, 3, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 60, 0, 60, 0, 248, 0, 60, 0, 60, 0, 60, 0, 216, 0, 60, 0, 60, 0, 81, 1, 60, 0, 60, 0, 60, 0, 80, 1, 60, 0, 60, 0, 83, 1, 60, 0, 60, 0, 60, 0, 82, 1, 60, 0, 60, 0, 244, 0, 60, 0, 60, 0, 60, 0, 212, 0, 0, 0, 170, 170, 100, 42, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 52, 30, 0, 0, 52, 30, 0, 0, 132, 33, 0, 0, 132, 33, 0, 0, 148, 35, 0, 0, 148, 35, 0, 0, 96, 38, 0, 0, 96, 38, 0, 0, 72, 42, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 214, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 105, 0, 113, 0, 118, 0, 126, 0, 134, 0, 142, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 48, 0, 48, 0, 170, 0, 48, 0, 178, 0, 183, 0, 191, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 197, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 205, 0, 158, 0, 48, 0, 48, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 150, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 164, 1, 216, 1, 24, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 64, 15, 0, 202, 97, 15, 0, 202, 130, 15, 0, 202, 163, 15, 0, 202, 196, 15, 0, 202, 229, 15, 0, 202, 6, 16, 0, 202, 39, 16, 0, 202, 72, 16, 0, 202, 105, 16, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 230, 0, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 166, 1, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 38, 3, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 198, 4, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 38, 6, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 230, 6, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 134, 8, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 1, 0, 0, 198, 33, 0, 0, 198, 33, 1, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 198, 97, 0, 0, 196, 154, 156, 70, 198, 129, 0, 0, 5, 5, 127, 12, 198, 65, 1, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 2, 15, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 193, 3, 111, 93, 193, 2, 111, 93, 198, 97, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 198, 225, 0, 0, 196, 154, 5, 70, 193, 4, 111, 93, 5, 5, 125, 12, 198, 129, 1, 0, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 10, 0, 201, 102, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 65, 0, 0, 198, 193, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 0, 0, 198, 1, 1, 0, 198, 33, 2, 0, 198, 65, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 134, 12, 0, 201, 166, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 14, 0, 201, 70, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 17, 0, 201, 230, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 226, 4, 0, 197, 34, 5, 0, 192, 0, 0, 0, 198, 162, 3, 0, 197, 98, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 3, 0, 198, 34, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 20, 0, 201, 134, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 5, 0, 197, 226, 5, 0, 198, 226, 5, 0, 198, 34, 6, 0, 198, 98, 6, 0, 198, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 6, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 23, 0, 201, 38, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 25, 0, 201, 102, 26, 0, 201, 134, 27, 0, 201, 166, 28, 0, 201, 198, 29, 0, 201, 230, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 32, 0, 201, 38, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 34, 0, 201, 198, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 9, 0, 198, 162, 9, 0, 198, 226, 9, 0, 198, 34, 10, 0, 198, 98, 10, 0, 198, 162, 10, 0, 198, 226, 10, 0, 198, 34, 11, 0, 198, 34, 8, 0, 198, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 37, 0, 201, 198, 38, 0, 201, 70, 40, 0, 201, 102, 41, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 42, 0, 201, 166, 43, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 134, 0, 73, 0, 4, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 71, 0, 4, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 72, 0, 4, 111, 93, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 3, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 3, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 164, 0, 0, 193, 3, 111, 93, 5, 200, 0, 0, 193, 2, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 193, 3, 111, 93, 5, 196, 0, 0, 193, 2, 111, 93, 5, 196, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 1, 0, 8, 3, 198, 129, 10, 3, 198, 161, 0, 50, 156, 5, 48, 0, 40, 3, 0, 192, 198, 65, 0, 70, 156, 5, 2, 0, 2, 3, 0, 192, 198, 97, 8, 3, 0, 192, 198, 129, 11, 3, 0, 192, 198, 161, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 1, 0, 8, 3, 255, 255, 111, 93, 193, 3, 10, 3, 255, 255, 111, 93, 193, 2, 0, 50, 5, 5, 48, 0, 40, 3, 0, 192, 198, 193, 0, 70, 5, 5, 2, 0, 2, 3, 0, 192, 198, 225, 8, 3, 255, 255, 111, 93, 193, 4, 11, 3, 1, 192, 198, 1, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 1, 192, 198, 162, 10, 3, 1, 192, 198, 226, 5, 42, 196, 162, 1, 0, 8, 3, 1, 192, 197, 34, 10, 3, 1, 192, 197, 98, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 162, 11, 3, 1, 192, 197, 226, 5, 82, 196, 162, 1, 0, 8, 3, 2, 192, 197, 34, 11, 3, 2, 192, 197, 98, 156, 70, 196, 190, 2, 0, 2, 3, 2, 192, 198, 98, 8, 3, 2, 192, 198, 162, 11, 3, 2, 192, 198, 226, 5, 70, 196, 190, 2, 0, 2, 3, 3, 192, 198, 34, 8, 3, 2, 192, 197, 162, 11, 3, 3, 192, 198, 98, 156, 82, 196, 190, 1, 0, 8, 3, 2, 192, 197, 226, 11, 3, 3, 192, 197, 34, 5, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 98, 11, 3, 3, 192, 197, 162, 156, 70, 196, 162, 2, 0, 2, 3, 4, 192, 198, 98, 8, 3, 4, 192, 198, 162, 11, 3, 4, 192, 198, 226, 5, 70, 196, 162, 2, 0, 2, 3, 5, 192, 198, 34, 8, 3, 5, 192, 197, 162, 11, 3, 5, 192, 198, 98, 156, 42, 196, 200, 1, 0, 8, 3, 7, 192, 198, 34, 10, 3, 7, 192, 198, 98, 5, 42, 196, 200, 1, 0, 8, 3, 6, 192, 197, 98, 10, 3, 6, 192, 197, 162, 156, 82, 196, 198, 1, 0, 8, 3, 6, 192, 197, 226, 11, 3, 7, 192, 197, 34, 5, 82, 196, 198, 1, 0, 8, 3, 7, 192, 197, 98, 11, 3, 7, 192, 197, 162, 156, 82, 196, 208, 1, 0, 8, 3, 7, 192, 197, 226, 11, 3, 8, 192, 197, 34, 5, 82, 196, 208, 1, 0, 8, 3, 8, 192, 197, 98, 11, 3, 8, 192, 197, 162, 156, 82, 196, 204, 1, 0, 8, 3, 8, 192, 197, 226, 11, 3, 9, 192, 197, 34, 5, 82, 196, 204, 1, 0, 8, 3, 9, 192, 197, 98, 11, 3, 9, 192, 197, 162, 156, 42, 196, 196, 1, 0, 8, 3, 7, 192, 198, 162, 10, 3, 7, 192, 198, 226, 5, 42, 196, 196, 1, 0, 8, 3, 9, 192, 197, 226, 10, 3, 10, 192, 197, 34, 156, 70, 196, 196, 2, 0, 2, 3, 8, 192, 198, 34, 8, 3, 8, 192, 198, 98, 11, 3, 8, 192, 198, 162, 5, 70, 196, 196, 2, 0, 2, 3, 8, 192, 198, 226, 8, 3, 10, 192, 197, 98, 11, 3, 9, 192, 198, 34, 10, 0, 197, 163, 2, 0, 2, 3, 11, 192, 198, 99, 8, 3, 11, 192, 198, 195, 11, 3, 12, 192, 198, 35, 11, 0, 197, 3, 2, 0, 2, 3, 12, 192, 198, 131, 8, 3, 11, 192, 197, 99, 11, 3, 12, 192, 198, 227, 156, 82, 196, 196, 1, 0, 8, 3, 11, 192, 197, 194, 11, 3, 12, 192, 197, 2, 5, 82, 196, 196, 1, 0, 8, 3, 12, 192, 197, 66, 11, 3, 12, 192, 197, 130, 12, 0, 197, 195, 1, 0, 8, 3, 13, 192, 197, 35, 11, 3, 13, 192, 197, 131, 13, 0, 197, 227, 1, 0, 8, 3, 14, 192, 197, 67, 11, 3, 14, 192, 197, 163, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 20, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 22, 4, 185, 60, 185, 64, 185, 68, 24, 4, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 26, 4, 185, 124, 185, 128, 185, 136, 186, 140, 186, 144, 28, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 30, 4, 168, 60, 168, 64, 168, 68, 32, 4, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 34, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 36, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 121, 58, 185, 184, 185, 180, 217, 184, 185, 66, 185, 73, 153, 73, 249, 73, 89, 74, 186, 89, 154, 89, 250, 89, 90, 90, 249, 68, 122, 114, 185, 121, 153, 121, 57, 189, 121, 122, 185, 188, 136, 14, 217, 188, 185, 153, 153, 153, 249, 153, 217, 168, 153, 169, 0, 8, 2, 8, 168, 57, 136, 57, 232, 57, 104, 58, 168, 184, 168, 180, 200, 184, 168, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 232, 68, 104, 114, 168, 121, 136, 121, 40, 189, 104, 122, 168, 188, 128, 14, 200, 188, 168, 153, 136, 153, 232, 153, 200, 168, 136, 169, 4, 8, 72, 170, 249, 58, 232, 58, 217, 57, 200, 57, 38, 4, 40, 4, 153, 65, 136, 65, 249, 65, 232, 65, 153, 66, 136, 66, 25, 66, 8, 66, 25, 70, 8, 70, 217, 68, 200, 68, 249, 74, 232, 74, 217, 73, 200, 73, 153, 74, 136, 74, 249, 184, 232, 184, 25, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 153, 82, 136, 82, 185, 82, 168, 82, 249, 85, 232, 85, 25, 87, 8, 87, 122, 90, 104, 90, 250, 90, 232, 90, 218, 89, 200, 89, 218, 90, 200, 90, 154, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 185, 102, 168, 102, 168, 132, 153, 105, 136, 105, 185, 106, 168, 106, 25, 106, 8, 106, 57, 107, 40, 107, 25, 107, 8, 107, 154, 113, 136, 113, 186, 114, 168, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 249, 122, 232, 122, 217, 121, 200, 121, 249, 188, 232, 188, 25, 189, 8, 189, 153, 137, 136, 137, 185, 138, 168, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 186, 142, 168, 142, 26, 142, 8, 142, 186, 146, 168, 146, 26, 146, 8, 146, 185, 148, 168, 148, 121, 154, 104, 154, 249, 154, 232, 154, 217, 153, 200, 153, 57, 154, 40, 154, 249, 168, 232, 168, 42, 4, 44, 4, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 153, 174, 136, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 12, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 14, 8, 1, 0, 8, 14, 16, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 18, 8, 200, 12, 72, 12, 32, 14, 187, 144, 185, 84, 200, 140, 169, 140, 169, 144, 168, 84, 187, 88, 187, 96, 169, 88, 169, 96, 169, 176, 169, 112, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 72, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 255, 5, 168, 72, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 217, 58, 255, 5, 200, 58, 255, 5, 217, 154, 255, 5, 200, 154, 255, 1, 3, 0, 77, 0, 68, 0, 32, 0, 195, 0, 0, 16, 66, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 206, 49, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 202, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 244, 38, 0, 0, 244, 38, 0, 0, 68, 42, 0, 0, 68, 42, 0, 0, 160, 46, 0, 0, 160, 46, 0, 0, 178, 49, 0, 0, 178, 49, 0, 0, 178, 49, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 98, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 106, 0, 114, 0, 119, 0, 127, 0, 135, 0, 143, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 146, 0, 154, 0, 160, 0, 168, 0, 176, 0, 184, 0, 48, 0, 56, 0, 183, 0, 191, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 208, 0, 216, 0, 48, 0, 48, 0, 48, 0, 224, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 232, 0, 48, 0, 48, 0, 240, 0, 48, 0, 246, 0, 254, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 1, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 9, 1, 48, 0, 48, 0, 13, 1, 21, 1, 29, 1, 34, 1, 42, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 56, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 59, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 199, 0, 206, 0, 48, 0, 48, 0, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 89, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 67, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 75, 1, 48, 0, 48, 0, 81, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 168, 1, 220, 1, 28, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 72, 2, 128, 2, 192, 2, 192, 0, 220, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 33, 0, 202, 193, 33, 0, 202, 226, 33, 0, 202, 3, 34, 0, 202, 36, 34, 0, 202, 69, 34, 0, 202, 102, 34, 0, 202, 135, 34, 0, 202, 168, 34, 0, 202, 201, 34, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 198, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 166, 1, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 102, 2, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 230, 3, 0, 134, 5, 0, 84, 134, 70, 0, 84, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 134, 5, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 230, 6, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 166, 7, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 70, 9, 0, 5, 5, 0, 84, 5, 70, 0, 84, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 1, 0, 0, 198, 33, 0, 0, 198, 33, 1, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 198, 97, 0, 0, 196, 154, 156, 70, 198, 129, 0, 0, 5, 5, 127, 12, 198, 65, 1, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 2, 25, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 193, 3, 111, 93, 193, 2, 111, 93, 198, 97, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 198, 225, 0, 0, 196, 154, 5, 70, 193, 4, 111, 93, 5, 5, 125, 12, 198, 129, 1, 0, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 11, 0, 201, 38, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 65, 0, 0, 198, 193, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 0, 0, 198, 1, 1, 0, 198, 33, 2, 0, 198, 65, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 70, 13, 0, 201, 102, 14, 0, 197, 162, 2, 0, 197, 226, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 15, 0, 201, 6, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 18, 0, 201, 166, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 4, 0, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 192, 0, 0, 0, 198, 162, 3, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 3, 0, 198, 34, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 20, 0, 201, 70, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 5, 0, 197, 98, 6, 0, 198, 226, 5, 0, 198, 34, 6, 0, 198, 98, 6, 0, 198, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 6, 0, 197, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 198, 23, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 25, 0, 5, 71, 238, 102, 201, 134, 26, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 7, 0, 192, 0, 0, 0, 197, 162, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 8, 0, 192, 0, 0, 0, 197, 98, 8, 0, 192, 0, 0, 0, 197, 162, 8, 0, 192, 0, 0, 0, 197, 226, 8, 0, 192, 0, 0, 0, 197, 34, 9, 0, 197, 99, 9, 0, 197, 194, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 10, 0, 197, 67, 10, 0, 197, 162, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 226, 6, 0, 197, 34, 11, 0, 5, 5, 10, 121, 197, 98, 11, 0, 197, 162, 11, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 7, 0, 5, 5, 24, 121, 197, 162, 7, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 27, 0, 201, 134, 28, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 29, 0, 201, 198, 30, 0, 201, 230, 31, 0, 201, 6, 33, 0, 201, 38, 34, 0, 201, 70, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 154, 134, 84, 192, 0, 0, 0, 196, 196, 134, 84, 192, 0, 0, 0, 197, 34, 17, 0, 197, 98, 17, 0, 197, 162, 17, 0, 197, 226, 17, 0, 197, 34, 18, 0, 197, 98, 18, 0, 197, 162, 18, 0, 197, 226, 18, 0, 197, 34, 19, 0, 197, 98, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 36, 0, 201, 134, 37, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 38, 0, 201, 38, 40, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 9, 0, 198, 162, 9, 0, 198, 226, 9, 0, 198, 34, 10, 0, 198, 98, 10, 0, 198, 162, 10, 0, 198, 226, 10, 0, 198, 34, 11, 0, 198, 34, 8, 0, 198, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 41, 0, 201, 38, 43, 0, 201, 166, 44, 0, 201, 198, 45, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 46, 0, 201, 6, 48, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 134, 0, 73, 0, 4, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 71, 0, 4, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 72, 0, 4, 111, 93, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 3, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 84, 5, 142, 0, 0, 5, 70, 0, 84, 5, 142, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 3, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 164, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 193, 3, 111, 93, 5, 200, 0, 0, 193, 2, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 134, 70, 0, 84, 5, 138, 0, 0, 5, 70, 0, 84, 5, 138, 0, 0, 134, 70, 0, 84, 5, 136, 0, 0, 5, 70, 0, 84, 5, 136, 0, 0, 134, 70, 0, 84, 5, 150, 0, 0, 5, 70, 0, 84, 5, 150, 0, 0, 134, 70, 0, 84, 5, 156, 0, 0, 5, 70, 0, 84, 5, 156, 0, 0, 134, 70, 0, 84, 5, 196, 0, 0, 5, 70, 0, 84, 5, 196, 0, 0, 5, 70, 0, 84, 5, 146, 0, 0, 193, 3, 111, 93, 5, 196, 0, 0, 193, 2, 111, 93, 5, 196, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 6, 121, 197, 226, 6, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 34, 7, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 98, 7, 0, 5, 5, 24, 121, 197, 162, 7, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 226, 7, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 34, 8, 0, 5, 5, 107, 121, 197, 98, 8, 0, 5, 5, 111, 121, 197, 162, 8, 0, 5, 5, 115, 121, 197, 226, 8, 0, 5, 5, 119, 121, 197, 34, 9, 0, 197, 99, 9, 0, 197, 194, 9, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 2, 10, 0, 197, 67, 10, 0, 197, 162, 10, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 226, 10, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 226, 6, 0, 197, 34, 11, 0, 5, 5, 10, 121, 197, 98, 11, 0, 197, 162, 11, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 197, 162, 12, 0, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 7, 0, 5, 5, 24, 121, 197, 162, 7, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 42, 156, 5, 1, 0, 8, 3, 198, 129, 10, 3, 198, 161, 0, 50, 156, 5, 48, 0, 40, 3, 0, 192, 198, 65, 0, 70, 156, 5, 2, 0, 2, 3, 0, 192, 198, 97, 8, 3, 0, 192, 198, 129, 11, 3, 0, 192, 198, 161, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 1, 0, 8, 3, 255, 255, 111, 93, 193, 3, 10, 3, 255, 255, 111, 93, 193, 2, 0, 50, 5, 5, 48, 0, 40, 3, 0, 192, 198, 193, 0, 70, 5, 5, 2, 0, 2, 3, 0, 192, 198, 225, 8, 3, 255, 255, 111, 93, 193, 4, 11, 3, 1, 192, 198, 1, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 1, 192, 198, 162, 10, 3, 1, 192, 198, 226, 5, 42, 196, 162, 1, 0, 8, 3, 1, 192, 197, 34, 10, 3, 1, 192, 197, 98, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 162, 11, 3, 1, 192, 197, 226, 5, 82, 196, 162, 1, 0, 8, 3, 2, 192, 197, 34, 11, 3, 2, 192, 197, 98, 156, 70, 196, 190, 2, 0, 2, 3, 2, 192, 198, 98, 8, 3, 2, 192, 198, 162, 11, 3, 2, 192, 198, 226, 5, 70, 196, 190, 2, 0, 2, 3, 3, 192, 198, 34, 8, 3, 3, 192, 197, 34, 11, 3, 3, 192, 198, 98, 156, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 98, 11, 3, 3, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 226, 11, 3, 4, 192, 197, 34, 156, 70, 196, 162, 2, 0, 2, 3, 4, 192, 198, 98, 8, 3, 4, 192, 198, 162, 11, 3, 4, 192, 198, 226, 5, 70, 196, 162, 2, 0, 2, 3, 5, 192, 198, 34, 8, 3, 6, 192, 197, 34, 11, 3, 5, 192, 198, 98, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 156, 42, 196, 200, 1, 0, 8, 3, 7, 192, 198, 34, 10, 3, 7, 192, 198, 98, 5, 42, 196, 200, 1, 0, 8, 3, 13, 192, 197, 162, 10, 3, 13, 192, 197, 226, 156, 82, 196, 198, 1, 0, 8, 3, 14, 192, 197, 34, 11, 3, 14, 192, 197, 98, 5, 82, 196, 198, 1, 0, 8, 3, 14, 192, 197, 162, 11, 3, 14, 192, 197, 226, 156, 82, 196, 208, 1, 0, 8, 3, 15, 192, 197, 34, 11, 3, 15, 192, 197, 98, 5, 82, 196, 208, 1, 0, 8, 3, 15, 192, 197, 162, 11, 3, 15, 192, 197, 226, 156, 82, 196, 204, 1, 0, 8, 3, 16, 192, 197, 34, 11, 3, 16, 192, 197, 98, 5, 82, 196, 204, 1, 0, 8, 3, 16, 192, 197, 162, 11, 3, 16, 192, 197, 226, 156, 42, 196, 196, 1, 0, 8, 3, 7, 192, 198, 162, 10, 3, 7, 192, 198, 226, 5, 42, 196, 196, 1, 0, 8, 3, 19, 192, 197, 226, 10, 3, 20, 192, 197, 34, 156, 70, 196, 196, 2, 0, 2, 3, 8, 192, 198, 34, 8, 3, 8, 192, 198, 98, 11, 3, 8, 192, 198, 162, 5, 70, 196, 196, 2, 0, 2, 3, 8, 192, 198, 226, 8, 3, 20, 192, 197, 98, 11, 3, 9, 192, 198, 34, 20, 0, 197, 163, 2, 0, 2, 3, 11, 192, 198, 99, 8, 3, 11, 192, 198, 195, 11, 3, 12, 192, 198, 35, 21, 0, 197, 3, 2, 0, 2, 3, 12, 192, 198, 131, 8, 3, 21, 192, 197, 99, 11, 3, 12, 192, 198, 227, 156, 82, 196, 196, 1, 0, 8, 3, 21, 192, 197, 194, 11, 3, 22, 192, 197, 2, 5, 82, 196, 196, 1, 0, 8, 3, 22, 192, 197, 66, 11, 3, 22, 192, 197, 130, 22, 0, 197, 195, 1, 0, 8, 3, 23, 192, 197, 35, 11, 3, 23, 192, 197, 131, 23, 0, 197, 227, 1, 0, 8, 3, 24, 192, 197, 67, 11, 3, 24, 192, 197, 163, 170, 170, 3, 0, 77, 0, 68, 0, 32, 0, 99, 11, 0, 16, 15, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 100, 43, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 196, 30, 0, 0, 196, 30, 0, 0, 20, 34, 0, 0, 20, 34, 0, 0, 140, 36, 0, 0, 140, 36, 0, 0, 88, 39, 0, 0, 88, 39, 0, 0, 72, 43, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 223, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 92, 0, 100, 0, 106, 0, 114, 0, 119, 0, 127, 0, 135, 0, 143, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 167, 0, 48, 0, 48, 0, 171, 0, 179, 0, 187, 0, 192, 0, 200, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 206, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 214, 0, 159, 0, 48, 0, 48, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 112, 1, 168, 1, 220, 1, 28, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 128, 18, 0, 202, 161, 18, 0, 202, 194, 18, 0, 202, 227, 18, 0, 202, 4, 19, 0, 202, 37, 19, 0, 202, 70, 19, 0, 202, 103, 19, 0, 202, 136, 19, 0, 202, 169, 19, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 230, 0, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 166, 1, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 38, 3, 0, 134, 5, 0, 84, 134, 70, 0, 84, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 198, 4, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 38, 6, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 230, 6, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 134, 8, 0, 5, 5, 0, 84, 5, 70, 0, 84, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 198, 1, 0, 0, 198, 33, 0, 0, 198, 33, 1, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 134, 71, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 198, 97, 0, 0, 196, 154, 156, 70, 198, 129, 0, 0, 5, 5, 127, 12, 198, 65, 1, 0, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 70, 0, 90, 196, 136, 156, 90, 197, 162, 0, 0, 197, 66, 18, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 193, 3, 111, 93, 193, 2, 111, 93, 198, 97, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 71, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 198, 225, 0, 0, 196, 154, 5, 70, 193, 4, 111, 93, 5, 5, 125, 12, 198, 129, 1, 0, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 70, 0, 90, 196, 136, 5, 90, 197, 226, 0, 0, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 10, 0, 201, 102, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 70, 0, 48, 5, 70, 0, 48, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 65, 0, 0, 198, 193, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 161, 0, 0, 198, 1, 1, 0, 198, 33, 2, 0, 198, 65, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 90, 5, 71, 0, 90, 201, 134, 12, 0, 201, 166, 13, 0, 197, 162, 2, 0, 197, 226, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 14, 0, 201, 70, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 17, 0, 201, 230, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 4, 0, 197, 162, 4, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 226, 4, 0, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 192, 0, 0, 0, 198, 162, 3, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 3, 0, 198, 34, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 20, 0, 201, 134, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 162, 5, 0, 197, 98, 6, 0, 198, 226, 5, 0, 198, 34, 6, 0, 198, 98, 6, 0, 198, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 226, 6, 0, 197, 162, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 23, 0, 201, 38, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 25, 0, 201, 102, 26, 0, 201, 134, 27, 0, 201, 166, 28, 0, 201, 198, 29, 0, 201, 230, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 154, 134, 84, 192, 0, 0, 0, 196, 196, 134, 84, 192, 0, 0, 0, 197, 98, 10, 0, 197, 162, 10, 0, 197, 226, 10, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 197, 34, 12, 0, 197, 98, 12, 0, 197, 162, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 32, 0, 201, 38, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 34, 0, 201, 198, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 98, 9, 0, 198, 162, 9, 0, 198, 226, 9, 0, 198, 34, 10, 0, 198, 98, 10, 0, 198, 162, 10, 0, 198, 226, 10, 0, 198, 34, 11, 0, 198, 34, 8, 0, 198, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 37, 0, 201, 198, 38, 0, 201, 70, 40, 0, 201, 102, 41, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 42, 0, 201, 166, 43, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 134, 0, 5, 0, 2, 111, 93, 0, 134, 0, 71, 0, 3, 111, 93, 0, 134, 0, 73, 0, 4, 111, 93, 0, 134, 0, 5, 0, 4, 111, 93, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 71, 0, 3, 111, 93, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 71, 0, 4, 111, 93, 0, 134, 0, 70, 0, 3, 111, 93, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 70, 0, 4, 111, 93, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 72, 0, 4, 111, 93, 0, 5, 0, 72, 0, 4, 111, 93, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 3, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 70, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 164, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 0, 5, 0, 3, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 2, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 136, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 138, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 182, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 154, 0, 0, 0, 0, 0, 134, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 5, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 73, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 71, 0, 4, 111, 93, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 70, 0, 90, 5, 136, 0, 0, 5, 70, 0, 90, 5, 136, 0, 0, 157, 5, 0, 80, 156, 5, 0, 56, 6, 5, 0, 80, 5, 5, 0, 56, 193, 3, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 162, 0, 0, 134, 70, 0, 90, 5, 162, 0, 0, 134, 71, 0, 90, 5, 162, 0, 0, 5, 70, 0, 90, 5, 162, 0, 0, 5, 71, 0, 90, 5, 162, 0, 0, 134, 70, 0, 84, 5, 142, 0, 0, 5, 70, 0, 84, 5, 142, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 134, 70, 0, 90, 5, 164, 0, 0, 5, 70, 0, 90, 5, 164, 0, 0, 134, 70, 0, 90, 5, 144, 0, 0, 5, 70, 0, 90, 5, 144, 0, 0, 134, 70, 0, 90, 5, 138, 0, 0, 5, 70, 0, 90, 5, 138, 0, 0, 193, 3, 111, 93, 5, 164, 0, 0, 193, 4, 111, 93, 5, 162, 0, 0, 193, 2, 111, 93, 5, 136, 0, 0, 193, 4, 111, 93, 5, 164, 0, 0, 193, 3, 111, 93, 5, 200, 0, 0, 193, 2, 111, 93, 5, 200, 0, 0, 134, 70, 0, 90, 5, 198, 0, 0, 134, 71, 0, 90, 5, 198, 0, 0, 5, 70, 0, 90, 5, 198, 0, 0, 5, 71, 0, 90, 5, 198, 0, 0, 134, 70, 0, 90, 5, 208, 0, 0, 134, 71, 0, 90, 5, 208, 0, 0, 5, 70, 0, 90, 5, 208, 0, 0, 5, 71, 0, 90, 5, 208, 0, 0, 134, 70, 0, 90, 5, 204, 0, 0, 134, 71, 0, 90, 5, 204, 0, 0, 5, 70, 0, 90, 5, 204, 0, 0, 5, 71, 0, 90, 5, 204, 0, 0, 134, 70, 0, 84, 5, 138, 0, 0, 5, 70, 0, 84, 5, 138, 0, 0, 134, 70, 0, 84, 5, 136, 0, 0, 5, 70, 0, 84, 5, 136, 0, 0, 134, 70, 0, 84, 5, 150, 0, 0, 5, 70, 0, 84, 5, 150, 0, 0, 134, 70, 0, 84, 5, 156, 0, 0, 5, 70, 0, 84, 5, 156, 0, 0, 134, 70, 0, 84, 5, 196, 0, 0, 5, 70, 0, 84, 5, 196, 0, 0, 5, 70, 0, 84, 5, 146, 0, 0, 193, 3, 111, 93, 5, 196, 0, 0, 193, 2, 111, 93, 5, 196, 0, 0, 193, 4, 111, 93, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 193, 4, 111, 93, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 196, 0, 0, 134, 71, 0, 90, 5, 196, 0, 0, 5, 70, 0, 90, 5, 196, 0, 0, 5, 71, 0, 90, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 90, 5, 190, 0, 0, 5, 196, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 1, 0, 8, 3, 198, 129, 10, 3, 198, 161, 0, 50, 156, 5, 48, 0, 40, 3, 0, 192, 198, 65, 0, 70, 156, 5, 2, 0, 2, 3, 0, 192, 198, 97, 8, 3, 0, 192, 198, 129, 11, 3, 0, 192, 198, 161, 0, 82, 156, 5, 2, 0, 8, 3, 255, 255, 0, 90, 134, 70, 11, 3, 255, 255, 0, 90, 134, 71, 68, 3, 197, 162, 0, 42, 5, 5, 1, 0, 8, 3, 255, 255, 111, 93, 193, 3, 10, 3, 255, 255, 111, 93, 193, 2, 0, 50, 5, 5, 48, 0, 40, 3, 0, 192, 198, 193, 0, 70, 5, 5, 2, 0, 2, 3, 0, 192, 198, 225, 8, 3, 255, 255, 111, 93, 193, 4, 11, 3, 1, 192, 198, 1, 0, 82, 5, 5, 2, 0, 8, 3, 255, 255, 0, 90, 5, 70, 11, 3, 255, 255, 0, 90, 5, 71, 68, 3, 0, 192, 197, 98, 156, 42, 196, 162, 1, 0, 8, 3, 1, 192, 198, 162, 10, 3, 1, 192, 198, 226, 5, 42, 196, 162, 1, 0, 8, 3, 1, 192, 197, 34, 10, 3, 1, 192, 197, 98, 156, 82, 196, 162, 1, 0, 8, 3, 1, 192, 197, 162, 11, 3, 1, 192, 197, 226, 5, 82, 196, 162, 1, 0, 8, 3, 2, 192, 197, 34, 11, 3, 2, 192, 197, 98, 156, 70, 196, 190, 2, 0, 2, 3, 2, 192, 198, 98, 8, 3, 2, 192, 198, 162, 11, 3, 2, 192, 198, 226, 5, 70, 196, 190, 2, 0, 2, 3, 3, 192, 198, 34, 8, 3, 3, 192, 197, 34, 11, 3, 3, 192, 198, 98, 156, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 98, 11, 3, 3, 192, 197, 162, 5, 82, 196, 190, 1, 0, 8, 3, 3, 192, 197, 226, 11, 3, 4, 192, 197, 34, 156, 70, 196, 162, 2, 0, 2, 3, 4, 192, 198, 98, 8, 3, 4, 192, 198, 162, 11, 3, 4, 192, 198, 226, 5, 70, 196, 162, 2, 0, 2, 3, 5, 192, 198, 34, 8, 3, 6, 192, 197, 34, 11, 3, 5, 192, 198, 98, 156, 42, 196, 200, 1, 0, 8, 3, 7, 192, 198, 34, 10, 3, 7, 192, 198, 98, 5, 42, 196, 200, 1, 0, 8, 3, 6, 192, 197, 226, 10, 3, 7, 192, 197, 34, 156, 82, 196, 198, 1, 0, 8, 3, 7, 192, 197, 98, 11, 3, 7, 192, 197, 162, 5, 82, 196, 198, 1, 0, 8, 3, 7, 192, 197, 226, 11, 3, 8, 192, 197, 34, 156, 82, 196, 208, 1, 0, 8, 3, 8, 192, 197, 98, 11, 3, 8, 192, 197, 162, 5, 82, 196, 208, 1, 0, 8, 3, 8, 192, 197, 226, 11, 3, 9, 192, 197, 34, 156, 82, 196, 204, 1, 0, 8, 3, 9, 192, 197, 98, 11, 3, 9, 192, 197, 162, 5, 82, 196, 204, 1, 0, 8, 3, 9, 192, 197, 226, 11, 3, 10, 192, 197, 34, 156, 42, 196, 196, 1, 0, 8, 3, 7, 192, 198, 162, 10, 3, 7, 192, 198, 226, 5, 42, 196, 196, 1, 0, 8, 3, 13, 192, 197, 34, 10, 3, 13, 192, 197, 98, 156, 70, 196, 196, 2, 0, 2, 3, 8, 192, 198, 34, 8, 3, 8, 192, 198, 98, 11, 3, 8, 192, 198, 162, 5, 70, 196, 196, 2, 0, 2, 3, 8, 192, 198, 226, 8, 3, 13, 192, 197, 162, 11, 3, 9, 192, 198, 34, 13, 0, 197, 227, 2, 0, 2, 3, 11, 192, 198, 99, 8, 3, 11, 192, 198, 195, 11, 3, 12, 192, 198, 35, 14, 0, 197, 67, 2, 0, 2, 3, 12, 192, 198, 131, 8, 3, 14, 192, 197, 163, 11, 3, 12, 192, 198, 227, 156, 82, 196, 196, 1, 0, 8, 3, 15, 192, 197, 2, 11, 3, 15, 192, 197, 66, 5, 82, 196, 196, 1, 0, 8, 3, 15, 192, 197, 130, 11, 3, 15, 192, 197, 194, 16, 0, 197, 3, 1, 0, 8, 3, 16, 192, 197, 99, 11, 3, 16, 192, 197, 195, 17, 0, 197, 35, 1, 0, 8, 3, 17, 192, 197, 131, 11, 3, 17, 192, 197, 227, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 28, 4, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 30, 4, 185, 124, 185, 128, 185, 136, 186, 140, 186, 144, 32, 4, 185, 156, 217, 156, 185, 160, 185, 164, 185, 168, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 34, 4, 168, 60, 168, 64, 168, 68, 36, 4, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 38, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 40, 4, 168, 156, 200, 156, 168, 160, 168, 164, 168, 168, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 121, 58, 185, 180, 185, 176, 217, 180, 185, 66, 185, 73, 153, 73, 249, 73, 89, 74, 186, 89, 154, 89, 250, 89, 90, 90, 249, 68, 122, 114, 185, 121, 153, 121, 57, 185, 121, 122, 185, 184, 136, 14, 217, 184, 185, 153, 153, 153, 249, 153, 217, 164, 153, 165, 0, 8, 2, 8, 168, 57, 136, 57, 232, 57, 104, 58, 168, 180, 168, 176, 200, 180, 168, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 232, 68, 104, 114, 168, 121, 136, 121, 40, 185, 104, 122, 168, 184, 128, 14, 200, 184, 168, 153, 136, 153, 232, 153, 200, 164, 136, 165, 4, 8, 72, 166, 249, 58, 232, 58, 217, 57, 200, 57, 42, 4, 44, 4, 153, 65, 136, 65, 249, 65, 232, 65, 153, 66, 136, 66, 25, 66, 8, 66, 25, 70, 8, 70, 217, 68, 200, 68, 249, 74, 232, 74, 217, 73, 200, 73, 153, 74, 136, 74, 249, 180, 232, 180, 25, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 153, 82, 136, 82, 185, 82, 168, 82, 249, 85, 232, 85, 25, 87, 8, 87, 122, 90, 104, 90, 250, 90, 232, 90, 218, 89, 200, 89, 218, 90, 200, 90, 154, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 185, 102, 168, 102, 168, 132, 153, 105, 136, 105, 185, 106, 168, 106, 25, 106, 8, 106, 57, 107, 40, 107, 25, 107, 8, 107, 154, 113, 136, 113, 186, 114, 168, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 249, 122, 232, 122, 217, 121, 200, 121, 249, 184, 232, 184, 25, 185, 8, 185, 153, 137, 136, 137, 185, 138, 168, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 186, 142, 168, 142, 26, 142, 8, 142, 186, 146, 168, 146, 26, 146, 8, 146, 185, 148, 168, 148, 121, 154, 104, 154, 249, 154, 232, 154, 217, 153, 200, 153, 57, 154, 40, 154, 249, 164, 232, 164, 46, 4, 48, 4, 12, 8, 14, 8, 249, 165, 232, 165, 89, 166, 153, 169, 136, 169, 153, 170, 136, 170, 25, 170, 8, 170, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 187, 144, 185, 84, 200, 140, 169, 140, 169, 144, 168, 84, 187, 88, 187, 96, 169, 88, 169, 96, 169, 172, 169, 112, 217, 156, 232, 1, 200, 156, 232, 1, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 72, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 255, 5, 168, 72, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 217, 58, 255, 5, 200, 58, 255, 5, 217, 154, 255, 5, 200, 154, 255, 1, 3, 0, 77, 0, 68, 0, 32, 0, 223, 23, 0, 16, 194, 0, 0, 96, 1, 0, 0, 96, 4, 0, 51, 0, 59, 0, 92, 0, 99, 0, 170, 170, 6, 0, 0, 96, 93, 11, 0, 32, 216, 23, 0, 32, 185, 34, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 190, 34, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 220, 7, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 224, 7, 0, 0, 224, 7, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 252, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 197, 221, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 148, 11, 60, 0, 130, 11, 60, 0, 131, 11, 38, 0, 185, 11, 60, 0, 149, 11, 205, 11, 183, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 149, 11, 60, 0, 149, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 153, 11, 60, 0, 153, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 154, 11, 60, 0, 154, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 158, 11, 60, 0, 158, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 159, 11, 60, 0, 159, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 163, 11, 60, 0, 163, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 164, 11, 60, 0, 164, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 168, 11, 60, 0, 168, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 170, 11, 60, 0, 170, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 174, 11, 60, 0, 174, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 175, 11, 60, 0, 175, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 176, 11, 60, 0, 176, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 178, 11, 60, 0, 178, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 181, 11, 60, 0, 181, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 180, 11, 60, 0, 180, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 179, 11, 60, 0, 179, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 177, 11, 60, 0, 177, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 169, 11, 60, 0, 169, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 156, 11, 60, 0, 156, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 182, 11, 60, 0, 182, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 183, 11, 60, 0, 183, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 184, 11, 60, 0, 184, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 185, 11, 60, 0, 185, 11, 205, 11, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 149, 11, 205, 11, 183, 11, 60, 0, 149, 11, 205, 11, 183, 11, 205, 11, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 84, 27, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 128, 0, 0, 0, 128, 1, 0, 0, 160, 25, 0, 0, 160, 25, 0, 0, 160, 25, 0, 0, 160, 25, 0, 0, 228, 25, 0, 0, 228, 25, 0, 0, 54, 27, 0, 0, 60, 27, 0, 0, 35, 0, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 36, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 109, 186, 255, 0, 110, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 41, 42, 43, 44, 39, 45, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 121, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 104, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 112, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 2, 31, 109, 193, 3, 31, 109, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 1, 0, 201, 198, 2, 0, 192, 0, 0, 0, 201, 166, 3, 0, 192, 0, 0, 0, 201, 134, 4, 0, 201, 102, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 6, 0, 201, 38, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 8, 0, 201, 230, 8, 0, 201, 198, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 10, 0, 201, 134, 11, 0, 201, 102, 12, 0, 201, 70, 13, 0, 201, 38, 14, 0, 201, 6, 15, 0, 201, 230, 15, 0, 201, 198, 16, 0, 201, 166, 17, 0, 201, 134, 18, 0, 201, 102, 19, 0, 201, 70, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 34, 109, 5, 5, 48, 0, 205, 11, 240, 127, 33, 109, 5, 5, 183, 11, 240, 127, 79, 109, 193, 3, 205, 11, 255, 255, 79, 109, 193, 2, 36, 109, 5, 5, 48, 0, 205, 11, 255, 255, 35, 109, 5, 5, 38, 109, 5, 5, 48, 0, 205, 11, 255, 255, 37, 109, 5, 5, 70, 109, 5, 5, 48, 0, 205, 11, 255, 255, 69, 109, 5, 5, 40, 109, 5, 5, 48, 0, 205, 11, 255, 255, 39, 109, 5, 5, 42, 109, 5, 5, 48, 0, 205, 11, 255, 255, 41, 109, 5, 5, 44, 109, 5, 5, 48, 0, 205, 11, 255, 255, 43, 109, 5, 5, 46, 109, 5, 5, 48, 0, 205, 11, 255, 255, 45, 109, 5, 5, 48, 109, 5, 5, 48, 0, 205, 11, 255, 255, 47, 109, 5, 5, 68, 109, 5, 5, 48, 0, 205, 11, 255, 255, 67, 109, 5, 5, 50, 109, 5, 5, 48, 0, 205, 11, 255, 255, 49, 109, 5, 5, 52, 109, 5, 5, 48, 0, 205, 11, 255, 255, 51, 109, 5, 5, 54, 109, 5, 5, 48, 0, 205, 11, 255, 255, 53, 109, 5, 5, 56, 109, 5, 5, 48, 0, 205, 11, 255, 255, 55, 109, 5, 5, 66, 109, 5, 5, 48, 0, 205, 11, 255, 255, 65, 109, 5, 5, 58, 109, 5, 5, 48, 0, 205, 11, 255, 255, 57, 109, 5, 5, 64, 109, 5, 5, 48, 0, 205, 11, 255, 255, 63, 109, 5, 5, 62, 109, 5, 5, 48, 0, 205, 11, 255, 255, 61, 109, 5, 5, 60, 109, 5, 5, 48, 0, 205, 11, 255, 255, 59, 109, 5, 5, 72, 109, 5, 5, 48, 0, 205, 11, 255, 255, 71, 109, 5, 5, 74, 109, 5, 5, 48, 0, 205, 11, 255, 255, 73, 109, 5, 5, 76, 109, 5, 5, 48, 0, 205, 11, 255, 255, 75, 109, 5, 5, 78, 109, 5, 5, 48, 0, 205, 11, 255, 255, 77, 109, 5, 5, 2, 0, 183, 11, 184, 11, 3, 0, 69, 0, 60, 0, 32, 0, 255, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 213, 7, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 218, 7, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 171, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 175, 6, 0, 0, 175, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 68, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 85, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 84, 0, 101, 0, 108, 0, 117, 0, 32, 0, 68, 0, 101, 0, 118, 0, 97, 0, 32, 0, 66, 0, 101, 0, 110, 0, 103, 0, 32, 0, 71, 0, 117, 0, 114, 0, 117, 0, 32, 0, 71, 0, 117, 0, 106, 0, 114, 0, 32, 0, 79, 0, 114, 0, 121, 0, 97, 0, 32, 0, 84, 0, 97, 0, 109, 0, 108, 0, 32, 0, 75, 0, 110, 0, 100, 0, 97, 0, 32, 0, 77, 0, 108, 0, 121, 0, 109, 0, 32, 0, 83, 0, 105, 0, 110, 0, 104, 0, 93, 0, 38, 0, 20, 12, 60, 0, 1, 12, 60, 0, 2, 12, 60, 0, 3, 12, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 112, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 13, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 52, 0, 0, 0, 116, 0, 0, 0, 116, 1, 0, 0, 20, 25, 0, 0, 20, 25, 0, 0, 20, 25, 0, 0, 20, 25, 0, 0, 88, 25, 0, 0, 36, 0, 0, 0, 10, 0, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 15, 0, 0, 0, 31, 0, 0, 0, 35, 0, 0, 0, 21, 0, 0, 0, 26, 0, 0, 0, 33, 0, 0, 0, 0, 0, 0, 41, 8, 0, 0, 96, 6, 0, 0, 104, 192, 255, 0, 110, 185, 255, 0, 111, 191, 255, 160, 113, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 102, 103, 104, 105, 106, 107, 108, 109, 40, 41, 42, 43, 44, 45, 39, 46, 47, 0, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 113, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 104, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 2, 37, 110, 193, 3, 37, 110, 193, 4, 37, 110, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 3, 0, 69, 0, 60, 0, 32, 0, 71, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 164, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 169, 6, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 203, 6, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 207, 6, 0, 0, 207, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 84, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 118, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 97, 0, 108, 0, 116, 0, 101, 0, 114, 0, 110, 0, 97, 0, 116, 0, 101, 0, 32, 0, 115, 0, 104, 0, 105, 0, 102, 0, 116, 0, 101, 0, 100, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 84, 0, 104, 0, 97, 0, 105, 0, 93, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 90, 14, 60, 0, 47, 14, 38, 0, 91, 14, 60, 0, 70, 14, 38, 0, 78, 14, 60, 0, 60, 0, 76, 14, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 48, 14, 60, 0, 77, 14, 38, 0, 50, 14, 60, 0, 60, 0, 60, 0, 69, 14, 38, 0, 50, 14, 77, 14, 60, 0, 60, 0, 60, 0, 77, 14, 50, 14, 60, 0, 60, 0, 60, 0, 51, 14, 38, 0, 69, 14, 77, 14, 60, 0, 60, 0, 60, 0, 77, 14, 69, 14, 38, 0, 68, 14, 60, 0, 58, 14, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 176, 25, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 21, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 68, 0, 0, 0, 68, 1, 0, 0, 36, 25, 0, 0, 36, 25, 0, 0, 36, 25, 0, 0, 36, 25, 0, 0, 128, 25, 0, 0, 128, 25, 0, 0, 146, 25, 0, 0, 152, 25, 0, 0, 38, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 39, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 117, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 93, 0, 100, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 108, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 1, 0, 202, 193, 1, 0, 202, 226, 1, 0, 202, 3, 2, 0, 202, 36, 2, 0, 202, 69, 2, 0, 202, 102, 2, 0, 202, 135, 2, 0, 202, 168, 2, 0, 202, 201, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 1, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 1, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 11, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 104, 114, 5, 5, 55, 11, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 229, 242, 201, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 104, 114, 8, 5, 99, 114, 5, 5, 104, 114, 7, 5, 99, 114, 6, 5, 104, 114, 6, 5, 99, 114, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 99, 114, 5, 5, 1, 0, 50, 14, 0, 192, 197, 98, 69, 14, 0, 192, 197, 162, 2, 0, 69, 14, 70, 14, 3, 0, 69, 0, 60, 0, 32, 0, 87, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 196, 6, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 201, 6, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 74, 10, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 78, 10, 0, 0, 78, 10, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 57, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 64, 220, 38, 0, 67, 0, 60, 0, 231, 0, 60, 0, 60, 0, 60, 0, 199, 0, 38, 0, 69, 0, 60, 0, 228, 0, 60, 0, 60, 0, 60, 0, 196, 0, 38, 0, 74, 0, 60, 0, 126, 1, 60, 0, 60, 0, 60, 0, 125, 1, 38, 0, 78, 0, 60, 0, 72, 1, 60, 0, 60, 0, 60, 0, 71, 1, 38, 0, 79, 0, 60, 0, 246, 0, 60, 0, 60, 0, 60, 0, 214, 0, 38, 0, 83, 0, 60, 0, 95, 1, 60, 0, 60, 0, 60, 0, 94, 1, 38, 0, 85, 0, 60, 0, 252, 0, 60, 0, 60, 0, 60, 0, 220, 0, 38, 0, 89, 0, 60, 0, 253, 0, 60, 0, 60, 0, 60, 0, 221, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 26, 40, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 180, 30, 0, 0, 180, 30, 0, 0, 180, 30, 0, 0, 180, 30, 0, 0, 248, 32, 0, 0, 248, 32, 0, 0, 234, 35, 0, 0, 234, 35, 0, 0, 2, 40, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 222, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 95, 0, 103, 0, 111, 0, 119, 0, 124, 0, 132, 0, 135, 0, 143, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 167, 0, 48, 0, 174, 0, 178, 0, 184, 0, 192, 0, 197, 0, 205, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 213, 0, 159, 0, 48, 0, 48, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 151, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 124, 1, 188, 1, 240, 1, 28, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 16, 0, 202, 1, 17, 0, 202, 34, 17, 0, 202, 67, 17, 0, 202, 100, 17, 0, 202, 133, 17, 0, 202, 166, 17, 0, 202, 199, 17, 0, 202, 232, 17, 0, 202, 9, 18, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 201, 198, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 134, 1, 0, 201, 102, 2, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 70, 3, 0, 156, 5, 0, 80, 201, 38, 4, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 201, 70, 5, 0, 201, 166, 6, 0, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 102, 7, 0, 5, 5, 0, 44, 201, 38, 8, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 230, 8, 0, 201, 198, 9, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 166, 10, 0, 5, 5, 0, 80, 201, 134, 11, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 201, 198, 12, 0, 201, 38, 14, 0, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 134, 5, 2, 51, 196, 146, 156, 42, 197, 34, 16, 0, 134, 5, 2, 47, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 134, 5, 2, 71, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 5, 2, 83, 134, 5, 2, 91, 160, 5, 38, 93, 197, 98, 16, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 5, 5, 2, 51, 196, 146, 5, 42, 197, 162, 16, 0, 5, 5, 2, 47, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 5, 5, 2, 71, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 5, 2, 83, 5, 5, 2, 91, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 14, 0, 201, 166, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 16, 0, 201, 38, 17, 0, 134, 5, 2, 69, 5, 5, 2, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 17, 0, 201, 166, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 61, 5, 5, 2, 61, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 19, 0, 201, 38, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 20, 0, 201, 166, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 3, 0, 197, 98, 3, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 162, 3, 0, 197, 226, 3, 0, 197, 34, 4, 0, 197, 98, 4, 0, 192, 0, 0, 0, 197, 162, 4, 0, 197, 226, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 22, 0, 201, 38, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 5, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 23, 0, 201, 166, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 6, 0, 197, 226, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 25, 0, 201, 38, 26, 0, 201, 230, 26, 0, 201, 166, 27, 0, 201, 102, 28, 0, 201, 38, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 29, 0, 201, 166, 30, 0, 201, 102, 31, 0, 201, 38, 32, 0, 201, 230, 32, 0, 201, 166, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 34, 0, 201, 38, 35, 0, 201, 230, 35, 0, 201, 166, 36, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 37, 0, 201, 38, 38, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 38, 0, 201, 166, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 40, 0, 201, 38, 41, 0, 201, 230, 41, 0, 201, 166, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 43, 0, 201, 38, 44, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 44, 0, 201, 6, 46, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 83, 5, 136, 0, 0, 5, 5, 2, 83, 5, 136, 0, 0, 134, 5, 2, 51, 5, 162, 0, 0, 5, 5, 2, 51, 5, 162, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 134, 5, 2, 83, 5, 164, 0, 0, 5, 5, 2, 83, 5, 164, 0, 0, 134, 5, 2, 83, 5, 144, 0, 0, 5, 5, 2, 83, 5, 144, 0, 0, 134, 5, 2, 83, 5, 138, 0, 0, 5, 5, 2, 83, 5, 138, 0, 0, 134, 5, 2, 51, 5, 164, 0, 0, 5, 5, 2, 51, 5, 164, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 2, 71, 5, 164, 0, 0, 5, 5, 2, 71, 5, 164, 0, 0, 134, 5, 2, 51, 5, 200, 0, 0, 5, 5, 2, 51, 5, 200, 0, 0, 134, 5, 2, 47, 5, 136, 0, 0, 5, 5, 2, 47, 5, 136, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 134, 5, 2, 83, 5, 198, 0, 0, 5, 5, 2, 83, 5, 198, 0, 0, 134, 5, 2, 83, 5, 208, 0, 0, 5, 5, 2, 83, 5, 208, 0, 0, 134, 5, 2, 83, 5, 204, 0, 0, 5, 5, 2, 83, 5, 204, 0, 0, 134, 5, 2, 61, 5, 196, 0, 0, 5, 5, 2, 61, 5, 196, 0, 0, 134, 5, 2, 61, 5, 210, 0, 0, 5, 5, 2, 61, 5, 210, 0, 0, 134, 5, 2, 51, 5, 196, 0, 0, 5, 5, 2, 51, 5, 196, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 196, 0, 0, 5, 5, 2, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 91, 5, 196, 0, 0, 5, 5, 2, 91, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 48, 0, 8, 3, 2, 243, 134, 5, 0, 46, 156, 5, 48, 0, 39, 3, 2, 239, 134, 5, 0, 68, 156, 5, 48, 0, 12, 3, 255, 255, 2, 69, 134, 5, 0, 70, 156, 5, 48, 0, 8, 3, 255, 255, 2, 71, 134, 5, 0, 78, 156, 5, 48, 0, 39, 3, 255, 255, 2, 79, 134, 5, 0, 82, 156, 5, 1, 0, 8, 3, 255, 255, 2, 83, 134, 5, 68, 3, 197, 162, 0, 90, 156, 5, 1, 0, 1, 3, 255, 255, 2, 91, 134, 5, 65, 3, 255, 255, 2, 91, 134, 5, 0, 92, 156, 5, 48, 0, 12, 3, 2, 253, 134, 5, 0, 42, 5, 5, 48, 0, 8, 3, 2, 243, 5, 5, 0, 46, 5, 5, 48, 0, 39, 3, 2, 239, 5, 5, 0, 68, 5, 5, 48, 0, 12, 3, 255, 255, 2, 69, 5, 5, 0, 70, 5, 5, 48, 0, 8, 3, 255, 255, 2, 71, 5, 5, 0, 78, 5, 5, 48, 0, 39, 3, 255, 255, 2, 79, 5, 5, 0, 82, 5, 5, 1, 0, 8, 3, 255, 255, 2, 83, 5, 5, 68, 3, 0, 192, 197, 98, 0, 90, 5, 5, 1, 0, 1, 3, 255, 255, 2, 91, 5, 5, 65, 3, 255, 255, 2, 91, 5, 5, 0, 92, 5, 5, 48, 0, 12, 3, 2, 253, 5, 5, 156, 42, 196, 162, 48, 0, 8, 3, 0, 192, 197, 162, 5, 42, 196, 162, 48, 0, 8, 3, 0, 192, 197, 226, 156, 68, 196, 160, 48, 0, 12, 3, 1, 192, 197, 34, 5, 68, 196, 160, 48, 0, 12, 3, 1, 192, 197, 98, 156, 82, 196, 162, 48, 0, 8, 3, 1, 192, 197, 162, 5, 82, 196, 162, 48, 0, 8, 3, 1, 192, 197, 226, 156, 70, 196, 190, 48, 0, 8, 3, 2, 192, 197, 34, 5, 70, 196, 190, 48, 0, 8, 3, 2, 192, 197, 98, 156, 82, 196, 190, 48, 0, 8, 3, 2, 192, 197, 162, 5, 82, 196, 190, 48, 0, 8, 3, 2, 192, 197, 226, 156, 70, 196, 162, 48, 0, 8, 3, 5, 192, 197, 34, 5, 70, 196, 162, 48, 0, 8, 3, 5, 192, 197, 98, 156, 42, 196, 200, 48, 0, 8, 3, 6, 192, 197, 34, 5, 42, 196, 200, 48, 0, 8, 3, 6, 192, 197, 98, 156, 68, 196, 196, 48, 0, 12, 3, 7, 192, 197, 34, 5, 68, 196, 196, 48, 0, 12, 3, 7, 192, 197, 98, 156, 68, 196, 210, 48, 0, 12, 3, 7, 192, 197, 162, 5, 68, 196, 210, 48, 0, 12, 3, 7, 192, 197, 226, 156, 68, 196, 204, 48, 0, 12, 3, 8, 192, 197, 34, 5, 68, 196, 204, 48, 0, 12, 3, 8, 192, 197, 98, 156, 82, 196, 198, 48, 0, 8, 3, 8, 192, 197, 162, 5, 82, 196, 198, 48, 0, 8, 3, 8, 192, 197, 226, 156, 82, 196, 208, 48, 0, 8, 3, 9, 192, 197, 34, 5, 82, 196, 208, 48, 0, 8, 3, 9, 192, 197, 98, 156, 82, 196, 204, 48, 0, 8, 3, 9, 192, 197, 162, 5, 82, 196, 204, 48, 0, 8, 3, 9, 192, 197, 226, 156, 92, 196, 196, 48, 0, 12, 3, 10, 192, 197, 34, 5, 92, 196, 196, 48, 0, 12, 3, 10, 192, 197, 98, 156, 92, 196, 210, 48, 0, 12, 3, 10, 192, 197, 162, 5, 92, 196, 210, 48, 0, 12, 3, 10, 192, 197, 226, 156, 42, 196, 196, 48, 0, 8, 3, 11, 192, 197, 34, 5, 42, 196, 196, 48, 0, 8, 3, 11, 192, 197, 98, 156, 70, 196, 196, 48, 0, 8, 3, 11, 192, 197, 162, 5, 70, 196, 196, 48, 0, 8, 3, 11, 192, 197, 226, 12, 0, 197, 35, 48, 0, 8, 3, 12, 192, 197, 131, 12, 0, 197, 227, 48, 0, 8, 3, 13, 192, 197, 67, 156, 82, 196, 196, 48, 0, 8, 3, 13, 192, 197, 162, 5, 82, 196, 196, 48, 0, 8, 3, 13, 192, 197, 226, 14, 0, 197, 35, 48, 0, 8, 3, 14, 192, 197, 131, 14, 0, 197, 227, 48, 0, 8, 3, 15, 192, 197, 67, 156, 90, 196, 196, 1, 0, 1, 3, 15, 192, 197, 162, 65, 3, 15, 192, 197, 162, 5, 90, 196, 196, 1, 0, 1, 3, 15, 192, 197, 226, 65, 3, 15, 192, 197, 226, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 28, 4, 185, 72, 186, 76, 185, 84, 185, 88, 185, 92, 186, 96, 186, 104, 185, 112, 185, 116, 185, 120, 30, 4, 32, 4, 185, 144, 185, 148, 185, 156, 34, 4, 185, 168, 36, 4, 185, 184, 185, 188, 185, 192, 38, 4, 40, 4, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 42, 4, 168, 60, 44, 4, 168, 72, 168, 76, 168, 84, 168, 88, 168, 92, 168, 96, 168, 104, 168, 112, 168, 116, 168, 120, 46, 4, 48, 4, 168, 144, 168, 148, 168, 156, 50, 4, 168, 168, 52, 4, 168, 184, 168, 188, 168, 192, 54, 4, 56, 4, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 136, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 185, 80, 57, 58, 0, 8, 185, 68, 186, 77, 154, 77, 250, 77, 90, 78, 186, 97, 154, 97, 250, 97, 90, 98, 217, 72, 154, 126, 185, 137, 153, 137, 249, 137, 153, 138, 185, 140, 136, 14, 217, 138, 185, 177, 153, 177, 249, 177, 185, 180, 185, 200, 185, 208, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 168, 80, 40, 58, 4, 8, 168, 68, 168, 77, 136, 77, 232, 77, 72, 78, 168, 97, 136, 97, 232, 97, 72, 98, 200, 72, 136, 126, 168, 137, 136, 137, 232, 137, 136, 138, 168, 140, 128, 14, 200, 138, 168, 177, 136, 177, 232, 177, 168, 180, 168, 200, 168, 208, 72, 198, 57, 59, 40, 59, 217, 57, 200, 57, 58, 4, 60, 4, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 249, 89, 232, 89, 217, 89, 200, 89, 185, 90, 168, 90, 249, 90, 232, 90, 249, 93, 232, 93, 89, 95, 72, 95, 154, 98, 136, 98, 58, 99, 40, 99, 218, 97, 200, 97, 26, 99, 8, 99, 186, 98, 168, 100, 6, 8, 8, 8, 250, 105, 232, 105, 249, 114, 232, 114, 168, 152, 153, 117, 136, 117, 249, 118, 232, 118, 25, 118, 8, 118, 121, 119, 104, 119, 89, 119, 72, 119, 154, 125, 136, 125, 62, 4, 64, 4, 185, 128, 168, 128, 10, 8, 185, 132, 168, 132, 57, 139, 40, 139, 217, 137, 200, 137, 121, 138, 104, 138, 12, 8, 14, 8, 153, 157, 136, 157, 249, 158, 232, 158, 25, 158, 8, 158, 154, 161, 136, 161, 250, 161, 232, 161, 185, 164, 168, 164, 26, 162, 8, 162, 249, 170, 232, 170, 25, 170, 8, 170, 185, 172, 168, 172, 153, 178, 136, 178, 57, 179, 40, 179, 217, 177, 200, 177, 57, 178, 40, 178, 121, 178, 104, 178, 66, 4, 68, 4, 249, 189, 232, 189, 249, 197, 232, 197, 89, 198, 153, 205, 136, 205, 185, 206, 168, 206, 185, 108, 168, 108, 232, 160, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 160, 169, 160, 200, 56, 169, 76, 187, 96, 187, 104, 169, 96, 169, 104, 169, 212, 169, 124, 217, 136, 187, 76, 200, 136, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 64, 255, 5, 186, 124, 255, 5, 185, 136, 255, 5, 186, 160, 255, 5, 185, 176, 255, 5, 185, 196, 255, 5, 185, 204, 255, 5, 168, 56, 255, 5, 168, 64, 255, 5, 168, 124, 255, 5, 168, 136, 255, 5, 168, 160, 255, 5, 168, 176, 255, 5, 168, 196, 255, 5, 168, 204, 255, 5, 25, 59, 255, 5, 8, 59, 255, 5, 250, 126, 255, 5, 232, 126, 255, 5, 25, 179, 255, 5, 8, 179, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 59, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 67, 10, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 72, 10, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 213, 11, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 217, 11, 0, 0, 217, 11, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 81, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 111, 220, 38, 0, 110, 0, 60, 0, 110, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 71, 0, 60, 0, 60, 0, 60, 0, 75, 1, 60, 0, 60, 0, 60, 0, 74, 1, 38, 0, 122, 0, 60, 0, 187, 2, 60, 0, 60, 0, 60, 0, 189, 2, 38, 0, 97, 0, 60, 0, 60, 0, 225, 0, 60, 0, 60, 0, 60, 0, 193, 0, 60, 0, 60, 0, 1, 1, 60, 0, 60, 0, 60, 0, 0, 1, 38, 0, 101, 0, 60, 0, 60, 0, 233, 0, 60, 0, 60, 0, 60, 0, 201, 0, 60, 0, 60, 0, 19, 1, 60, 0, 60, 0, 60, 0, 18, 1, 38, 0, 105, 0, 60, 0, 60, 0, 237, 0, 60, 0, 60, 0, 60, 0, 205, 0, 60, 0, 60, 0, 43, 1, 60, 0, 60, 0, 60, 0, 42, 1, 38, 0, 111, 0, 60, 0, 60, 0, 243, 0, 60, 0, 60, 0, 60, 0, 211, 0, 60, 0, 60, 0, 77, 1, 60, 0, 60, 0, 60, 0, 76, 1, 38, 0, 117, 0, 60, 0, 60, 0, 250, 0, 60, 0, 60, 0, 60, 0, 218, 0, 60, 0, 60, 0, 107, 1, 60, 0, 60, 0, 60, 0, 106, 1, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 230, 45, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 31, 0, 0, 68, 31, 0, 0, 68, 31, 0, 0, 68, 31, 0, 0, 144, 35, 0, 0, 144, 35, 0, 0, 104, 41, 0, 0, 130, 41, 0, 0, 206, 45, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 231, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 94, 0, 102, 0, 107, 0, 115, 0, 120, 0, 128, 0, 132, 0, 140, 0, 48, 0, 56, 0, 143, 0, 151, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 175, 0, 182, 0, 186, 0, 191, 0, 48, 0, 199, 0, 206, 0, 214, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 222, 0, 167, 0, 48, 0, 48, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 159, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 120, 1, 172, 1, 224, 1, 16, 2, 192, 0, 60, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 33, 0, 202, 65, 33, 0, 202, 98, 33, 0, 202, 131, 33, 0, 202, 164, 33, 0, 202, 197, 33, 0, 202, 230, 33, 0, 202, 7, 34, 0, 202, 40, 34, 0, 202, 73, 34, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 134, 1, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 6, 3, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 128, 4, 0, 201, 198, 11, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 166, 13, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 134, 15, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 6, 17, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 134, 18, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 0, 20, 0, 201, 198, 23, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 166, 25, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 134, 70, 0, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 32, 0, 196, 160, 156, 46, 196, 138, 156, 50, 134, 70, 0, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 134, 70, 0, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 134, 70, 0, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 134, 70, 0, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 32, 0, 196, 138, 5, 42, 5, 70, 0, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 32, 0, 196, 160, 5, 46, 196, 138, 5, 50, 5, 70, 0, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 5, 70, 0, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 5, 70, 0, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 5, 70, 0, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 134, 71, 0, 42, 5, 71, 0, 42, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 27, 0, 201, 6, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 50, 5, 71, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 30, 0, 201, 6, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 58, 5, 71, 0, 58, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 33, 0, 201, 6, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 137, 5, 2, 69, 8, 5, 2, 69, 134, 71, 0, 70, 5, 71, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 82, 5, 71, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 36, 0, 201, 6, 38, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 39, 0, 201, 6, 41, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 42, 0, 201, 6, 44, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 45, 0, 201, 6, 47, 0, 197, 162, 11, 0, 197, 34, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 48, 0, 201, 6, 50, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 93, 192, 0, 0, 0, 6, 5, 2, 93, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 51, 0, 201, 6, 53, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 14, 0, 197, 162, 14, 0, 197, 226, 14, 0, 197, 34, 15, 0, 201, 134, 54, 0, 201, 6, 56, 0, 201, 134, 57, 0, 201, 6, 59, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 60, 0, 201, 6, 62, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 18, 0, 197, 162, 18, 0, 197, 226, 18, 0, 197, 34, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 63, 0, 201, 6, 65, 0, 201, 134, 66, 0, 201, 6, 68, 0, 201, 134, 69, 0, 201, 6, 71, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 22, 0, 197, 162, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 72, 0, 201, 6, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 75, 0, 201, 6, 77, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 78, 0, 201, 6, 80, 0, 201, 134, 81, 0, 201, 6, 83, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 9, 0, 197, 226, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 84, 0, 201, 6, 86, 0, 201, 134, 87, 0, 201, 6, 89, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 10, 0, 197, 226, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 90, 0, 201, 6, 92, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 135, 5, 2, 69, 5, 142, 0, 0, 70, 5, 2, 69, 5, 142, 0, 0, 135, 5, 2, 69, 5, 140, 0, 0, 70, 5, 2, 69, 5, 140, 0, 0, 135, 5, 2, 69, 5, 156, 0, 0, 70, 5, 2, 69, 5, 156, 0, 0, 135, 5, 2, 69, 5, 160, 0, 0, 70, 5, 2, 69, 5, 160, 0, 0, 135, 5, 2, 69, 5, 144, 0, 0, 70, 5, 2, 69, 5, 144, 0, 0, 135, 5, 2, 69, 5, 136, 0, 0, 70, 5, 2, 69, 5, 136, 0, 0, 135, 5, 2, 69, 5, 164, 0, 0, 70, 5, 2, 69, 5, 164, 0, 0, 5, 5, 2, 69, 5, 142, 0, 0, 5, 5, 2, 69, 5, 140, 0, 0, 5, 5, 2, 69, 5, 156, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 144, 0, 0, 5, 5, 2, 69, 5, 136, 0, 0, 5, 5, 2, 69, 5, 164, 0, 0, 134, 70, 0, 42, 5, 162, 0, 0, 134, 71, 0, 42, 5, 162, 0, 0, 5, 70, 0, 42, 5, 162, 0, 0, 5, 71, 0, 42, 5, 162, 0, 0, 134, 70, 0, 50, 5, 162, 0, 0, 134, 71, 0, 50, 5, 162, 0, 0, 5, 70, 0, 50, 5, 162, 0, 0, 5, 71, 0, 50, 5, 162, 0, 0, 134, 70, 0, 58, 5, 162, 0, 0, 134, 71, 0, 58, 5, 162, 0, 0, 5, 70, 0, 58, 5, 162, 0, 0, 5, 71, 0, 58, 5, 162, 0, 0, 134, 70, 0, 82, 5, 162, 0, 0, 134, 71, 0, 82, 5, 162, 0, 0, 5, 70, 0, 82, 5, 162, 0, 0, 5, 71, 0, 82, 5, 162, 0, 0, 134, 70, 0, 70, 5, 190, 0, 0, 134, 71, 0, 70, 5, 190, 0, 0, 5, 70, 0, 70, 5, 190, 0, 0, 5, 71, 0, 70, 5, 190, 0, 0, 134, 70, 0, 82, 5, 190, 0, 0, 134, 71, 0, 82, 5, 190, 0, 0, 5, 70, 0, 82, 5, 190, 0, 0, 5, 71, 0, 82, 5, 190, 0, 0, 134, 70, 0, 70, 5, 162, 0, 0, 134, 71, 0, 70, 5, 162, 0, 0, 5, 70, 0, 70, 5, 162, 0, 0, 5, 71, 0, 70, 5, 162, 0, 0, 134, 70, 0, 50, 5, 160, 0, 0, 134, 71, 0, 50, 5, 160, 0, 0, 5, 70, 0, 50, 5, 160, 0, 0, 5, 71, 0, 50, 5, 160, 0, 0, 134, 70, 0, 42, 5, 200, 0, 0, 134, 71, 0, 42, 5, 200, 0, 0, 5, 70, 0, 42, 5, 200, 0, 0, 5, 71, 0, 42, 5, 200, 0, 0, 134, 71, 0, 50, 5, 138, 0, 0, 5, 71, 0, 50, 5, 138, 0, 0, 134, 71, 0, 50, 5, 136, 0, 0, 5, 71, 0, 50, 5, 136, 0, 0, 134, 70, 0, 50, 5, 204, 0, 0, 134, 71, 0, 50, 5, 204, 0, 0, 5, 70, 0, 50, 5, 204, 0, 0, 5, 71, 0, 50, 5, 204, 0, 0, 134, 70, 0, 50, 5, 208, 0, 0, 134, 71, 0, 50, 5, 208, 0, 0, 5, 70, 0, 50, 5, 208, 0, 0, 5, 71, 0, 50, 5, 208, 0, 0, 134, 70, 0, 58, 5, 208, 0, 0, 134, 71, 0, 58, 5, 208, 0, 0, 5, 70, 0, 58, 5, 208, 0, 0, 5, 71, 0, 58, 5, 208, 0, 0, 134, 71, 0, 70, 5, 138, 0, 0, 5, 71, 0, 70, 5, 138, 0, 0, 134, 71, 0, 70, 5, 136, 0, 0, 5, 71, 0, 70, 5, 136, 0, 0, 134, 70, 0, 82, 5, 198, 0, 0, 134, 71, 0, 82, 5, 198, 0, 0, 5, 70, 0, 82, 5, 198, 0, 0, 5, 71, 0, 82, 5, 198, 0, 0, 134, 70, 0, 82, 5, 208, 0, 0, 134, 71, 0, 82, 5, 208, 0, 0, 5, 70, 0, 82, 5, 208, 0, 0, 5, 71, 0, 82, 5, 208, 0, 0, 134, 70, 0, 82, 5, 204, 0, 0, 134, 71, 0, 82, 5, 204, 0, 0, 5, 70, 0, 82, 5, 204, 0, 0, 5, 71, 0, 82, 5, 204, 0, 0, 134, 71, 0, 82, 5, 150, 0, 0, 5, 71, 0, 82, 5, 150, 0, 0, 134, 70, 0, 42, 5, 196, 0, 0, 134, 71, 0, 42, 5, 196, 0, 0, 5, 70, 0, 42, 5, 196, 0, 0, 5, 71, 0, 42, 5, 196, 0, 0, 134, 70, 0, 50, 5, 196, 0, 0, 134, 71, 0, 50, 5, 196, 0, 0, 5, 70, 0, 50, 5, 196, 0, 0, 5, 71, 0, 50, 5, 196, 0, 0, 134, 70, 0, 58, 5, 196, 0, 0, 134, 71, 0, 58, 5, 196, 0, 0, 5, 70, 0, 58, 5, 196, 0, 0, 5, 71, 0, 58, 5, 196, 0, 0, 134, 70, 0, 70, 5, 196, 0, 0, 134, 71, 0, 70, 5, 196, 0, 0, 5, 70, 0, 70, 5, 196, 0, 0, 5, 71, 0, 70, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 82, 5, 196, 0, 0, 134, 71, 0, 82, 5, 196, 0, 0, 5, 70, 0, 82, 5, 196, 0, 0, 5, 71, 0, 82, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 70, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 71, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 70, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 71, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 2, 0, 1, 3, 0, 234, 134, 70, 4, 3, 0, 234, 134, 71, 65, 3, 0, 234, 134, 70, 0, 50, 156, 5, 2, 0, 1, 3, 0, 242, 134, 70, 4, 3, 0, 242, 134, 71, 65, 3, 0, 242, 134, 70, 0, 58, 156, 5, 2, 0, 1, 3, 0, 250, 134, 70, 4, 3, 0, 250, 134, 71, 65, 3, 0, 250, 134, 70, 0, 68, 156, 5, 15, 0, 34, 1, 26, 0, 244, 1, 12, 0, 244, 1, 2, 192, 197, 162, 245, 1, 2, 192, 197, 226, 32, 30, 3, 192, 197, 34, 33, 30, 3, 192, 197, 98, 34, 1, 1, 192, 197, 162, 35, 1, 1, 192, 197, 226, 230, 1, 2, 192, 197, 34, 231, 1, 2, 192, 197, 98, 30, 1, 12, 0, 30, 1, 0, 192, 197, 162, 31, 1, 0, 192, 197, 226, 32, 1, 1, 192, 197, 34, 33, 1, 1, 192, 197, 98, 71, 0, 255, 255, 2, 69, 135, 5, 103, 0, 255, 255, 2, 69, 70, 5, 28, 1, 197, 162, 29, 1, 0, 192, 197, 98, 0, 70, 156, 5, 2, 0, 1, 3, 255, 255, 0, 70, 134, 70, 4, 3, 255, 255, 0, 70, 134, 71, 65, 3, 255, 255, 0, 70, 134, 70, 0, 82, 156, 5, 2, 0, 1, 3, 255, 255, 0, 82, 134, 70, 4, 3, 255, 255, 0, 82, 134, 71, 65, 3, 255, 255, 0, 82, 134, 70, 0, 42, 5, 5, 2, 0, 1, 3, 0, 234, 5, 70, 4, 3, 0, 234, 5, 71, 65, 3, 0, 234, 5, 70, 0, 50, 5, 5, 2, 0, 1, 3, 0, 242, 5, 70, 4, 3, 0, 242, 5, 71, 65, 3, 0, 242, 5, 70, 0, 58, 5, 5, 2, 0, 1, 3, 0, 250, 5, 70, 4, 3, 0, 250, 5, 71, 65, 3, 0, 250, 5, 70, 0, 68, 5, 5, 7, 0, 35, 1, 12, 0, 35, 1, 4, 192, 197, 98, 231, 1, 4, 192, 197, 162, 245, 1, 4, 192, 197, 226, 33, 30, 5, 192, 197, 34, 103, 0, 255, 255, 2, 69, 5, 5, 29, 1, 3, 192, 197, 162, 31, 1, 3, 192, 197, 226, 33, 1, 4, 192, 197, 34, 0, 70, 5, 5, 2, 0, 1, 3, 255, 255, 0, 70, 5, 70, 4, 3, 255, 255, 0, 70, 5, 71, 65, 3, 255, 255, 0, 70, 5, 70, 0, 82, 5, 5, 2, 0, 1, 3, 255, 255, 0, 82, 5, 70, 4, 3, 255, 255, 0, 82, 5, 71, 65, 3, 255, 255, 0, 82, 5, 70, 156, 42, 196, 162, 2, 0, 1, 3, 5, 192, 197, 98, 4, 3, 5, 192, 197, 162, 65, 3, 5, 192, 197, 98, 5, 42, 196, 162, 2, 0, 1, 3, 5, 192, 197, 226, 4, 3, 6, 192, 197, 34, 65, 3, 5, 192, 197, 226, 156, 50, 196, 162, 2, 0, 1, 3, 6, 192, 197, 98, 4, 3, 6, 192, 197, 162, 65, 3, 6, 192, 197, 98, 5, 50, 196, 162, 2, 0, 1, 3, 6, 192, 197, 226, 4, 3, 7, 192, 197, 34, 65, 3, 6, 192, 197, 226, 156, 58, 196, 162, 2, 0, 1, 3, 7, 192, 197, 98, 4, 3, 7, 192, 197, 162, 65, 3, 7, 192, 197, 98, 5, 58, 196, 162, 2, 0, 1, 3, 7, 192, 197, 226, 4, 3, 8, 192, 197, 34, 65, 3, 7, 192, 197, 226, 156, 82, 196, 162, 2, 0, 1, 3, 8, 192, 197, 98, 4, 3, 8, 192, 197, 162, 65, 3, 8, 192, 197, 98, 5, 82, 196, 162, 2, 0, 1, 3, 8, 192, 197, 226, 4, 3, 9, 192, 197, 34, 65, 3, 8, 192, 197, 226, 156, 70, 196, 190, 2, 0, 1, 3, 9, 192, 197, 98, 4, 3, 9, 192, 197, 162, 65, 3, 9, 192, 197, 98, 5, 70, 196, 190, 2, 0, 1, 3, 9, 192, 197, 226, 4, 3, 10, 192, 197, 34, 65, 3, 9, 192, 197, 226, 156, 82, 196, 190, 2, 0, 1, 3, 10, 192, 197, 98, 4, 3, 10, 192, 197, 162, 65, 3, 10, 192, 197, 98, 5, 82, 196, 190, 2, 0, 1, 3, 10, 192, 197, 226, 4, 3, 11, 192, 197, 34, 65, 3, 10, 192, 197, 226, 156, 70, 196, 162, 2, 0, 1, 3, 11, 192, 197, 98, 4, 3, 11, 192, 197, 162, 65, 3, 11, 192, 197, 98, 5, 70, 196, 162, 2, 0, 1, 3, 11, 192, 197, 226, 4, 3, 12, 192, 197, 34, 65, 3, 11, 192, 197, 226, 156, 50, 196, 160, 2, 0, 1, 3, 12, 192, 197, 98, 4, 3, 12, 192, 197, 162, 65, 3, 12, 192, 197, 98, 5, 50, 196, 160, 2, 0, 1, 3, 12, 192, 197, 226, 4, 3, 13, 192, 197, 34, 65, 3, 12, 192, 197, 226, 156, 42, 196, 200, 2, 0, 1, 3, 13, 192, 197, 98, 4, 3, 13, 192, 197, 162, 65, 3, 13, 192, 197, 98, 5, 42, 196, 200, 2, 0, 1, 3, 13, 192, 197, 226, 4, 3, 14, 192, 197, 34, 65, 3, 13, 192, 197, 226, 156, 50, 196, 204, 2, 0, 1, 3, 15, 192, 197, 98, 4, 3, 15, 192, 197, 162, 65, 3, 15, 192, 197, 98, 5, 50, 196, 204, 2, 0, 1, 3, 15, 192, 197, 226, 4, 3, 16, 192, 197, 34, 65, 3, 15, 192, 197, 226, 156, 50, 196, 208, 2, 0, 1, 3, 16, 192, 197, 98, 4, 3, 16, 192, 197, 162, 65, 3, 16, 192, 197, 98, 5, 50, 196, 208, 2, 0, 1, 3, 16, 192, 197, 226, 4, 3, 17, 192, 197, 34, 65, 3, 16, 192, 197, 226, 156, 58, 196, 208, 2, 0, 1, 3, 17, 192, 197, 98, 4, 3, 17, 192, 197, 162, 65, 3, 17, 192, 197, 98, 5, 58, 196, 208, 2, 0, 1, 3, 17, 192, 197, 226, 4, 3, 18, 192, 197, 34, 65, 3, 17, 192, 197, 226, 156, 82, 196, 198, 2, 0, 1, 3, 19, 192, 197, 98, 4, 3, 19, 192, 197, 162, 65, 3, 19, 192, 197, 98, 5, 82, 196, 198, 2, 0, 1, 3, 19, 192, 197, 226, 4, 3, 20, 192, 197, 34, 65, 3, 19, 192, 197, 226, 156, 82, 196, 208, 2, 0, 1, 3, 20, 192, 197, 98, 4, 3, 20, 192, 197, 162, 65, 3, 20, 192, 197, 98, 5, 82, 196, 208, 2, 0, 1, 3, 20, 192, 197, 226, 4, 3, 21, 192, 197, 34, 65, 3, 20, 192, 197, 226, 156, 82, 196, 204, 2, 0, 1, 3, 21, 192, 197, 98, 4, 3, 21, 192, 197, 162, 65, 3, 21, 192, 197, 98, 5, 82, 196, 204, 2, 0, 1, 3, 21, 192, 197, 226, 4, 3, 22, 192, 197, 34, 65, 3, 21, 192, 197, 226, 156, 42, 196, 196, 2, 0, 1, 3, 22, 192, 197, 226, 4, 3, 23, 192, 197, 34, 65, 3, 22, 192, 197, 226, 5, 42, 196, 196, 2, 0, 1, 3, 23, 192, 197, 98, 4, 3, 23, 192, 197, 162, 65, 3, 23, 192, 197, 98, 156, 50, 196, 196, 2, 0, 1, 3, 23, 192, 197, 226, 4, 3, 24, 192, 197, 34, 65, 3, 23, 192, 197, 226, 5, 50, 196, 196, 2, 0, 1, 3, 24, 192, 197, 98, 4, 3, 24, 192, 197, 162, 65, 3, 24, 192, 197, 98, 156, 58, 196, 196, 2, 0, 1, 3, 24, 192, 197, 226, 4, 3, 25, 192, 197, 34, 65, 3, 24, 192, 197, 226, 5, 58, 196, 196, 2, 0, 1, 3, 25, 192, 197, 98, 4, 3, 25, 192, 197, 162, 65, 3, 25, 192, 197, 98, 156, 70, 196, 196, 2, 0, 1, 3, 25, 192, 197, 226, 4, 3, 26, 192, 197, 34, 65, 3, 25, 192, 197, 226, 5, 70, 196, 196, 2, 0, 1, 3, 26, 192, 197, 98, 4, 3, 26, 192, 197, 162, 65, 3, 26, 192, 197, 98, 26, 0, 197, 227, 2, 0, 1, 3, 27, 192, 197, 67, 4, 3, 27, 192, 197, 163, 65, 3, 27, 192, 197, 67, 28, 0, 197, 3, 2, 0, 1, 3, 28, 192, 197, 99, 4, 3, 28, 192, 197, 195, 65, 3, 28, 192, 197, 99, 156, 82, 196, 196, 2, 0, 1, 3, 29, 192, 197, 34, 4, 3, 29, 192, 197, 98, 65, 3, 29, 192, 197, 34, 5, 82, 196, 196, 2, 0, 1, 3, 29, 192, 197, 162, 4, 3, 29, 192, 197, 226, 65, 3, 29, 192, 197, 162, 30, 0, 197, 35, 2, 0, 1, 3, 30, 192, 197, 131, 4, 3, 30, 192, 197, 227, 65, 3, 30, 192, 197, 131, 31, 0, 197, 67, 2, 0, 1, 3, 31, 192, 197, 163, 4, 3, 32, 192, 197, 3, 65, 3, 31, 192, 197, 163, 12, 0, 71, 0, 72, 0, 103, 0, 104, 0, 28, 1, 36, 1, 230, 1, 232, 1, 244, 1, 246, 1, 32, 30, 34, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 28, 4, 185, 76, 185, 80, 185, 84, 30, 4, 186, 96, 185, 100, 185, 104, 185, 108, 32, 4, 54, 4, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 56, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 58, 4, 168, 60, 168, 64, 168, 68, 60, 4, 168, 76, 168, 80, 168, 84, 62, 4, 168, 96, 168, 100, 168, 104, 168, 108, 64, 4, 76, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 78, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 217, 56, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 217, 72, 250, 73, 90, 74, 186, 89, 217, 88, 250, 89, 90, 90, 217, 68, 154, 114, 185, 121, 217, 120, 249, 121, 153, 122, 89, 122, 136, 14, 217, 122, 185, 153, 217, 152, 249, 153, 89, 154, 153, 169, 185, 176, 2, 8, 168, 57, 200, 56, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 200, 72, 232, 73, 72, 74, 168, 89, 200, 88, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 200, 120, 232, 121, 136, 122, 72, 122, 128, 14, 200, 122, 168, 153, 200, 152, 232, 153, 72, 154, 136, 169, 168, 176, 72, 170, 249, 56, 232, 56, 217, 57, 200, 57, 80, 4, 82, 4, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 57, 71, 40, 71, 249, 72, 232, 72, 218, 73, 200, 73, 186, 74, 168, 74, 84, 4, 86, 4, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 57, 87, 40, 87, 154, 90, 136, 90, 249, 88, 232, 88, 218, 89, 200, 89, 88, 4, 90, 4, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 89, 107, 72, 107, 57, 107, 40, 107, 154, 113, 136, 113, 250, 114, 232, 114, 26, 114, 8, 114, 10, 8, 188, 116, 171, 116, 249, 120, 232, 120, 217, 121, 200, 121, 121, 122, 104, 122, 12, 8, 14, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 250, 142, 232, 142, 26, 142, 8, 142, 249, 146, 232, 146, 25, 146, 8, 146, 185, 148, 168, 148, 153, 154, 136, 154, 249, 152, 232, 152, 217, 153, 200, 153, 57, 154, 40, 154, 121, 154, 104, 154, 92, 4, 94, 4, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 185, 174, 168, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 25, 57, 187, 72, 200, 140, 169, 140, 8, 57, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 25, 121, 187, 72, 8, 121, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 72, 255, 5, 186, 88, 255, 5, 186, 112, 71, 4, 186, 116, 103, 4, 177, 116, 28, 5, 250, 117, 29, 5, 241, 117, 30, 5, 218, 117, 31, 5, 209, 117, 32, 5, 186, 118, 33, 5, 177, 118, 34, 5, 250, 118, 35, 5, 241, 118, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 255, 5, 168, 72, 255, 5, 168, 88, 255, 5, 168, 112, 103, 4, 168, 116, 29, 5, 232, 117, 31, 5, 200, 117, 33, 5, 168, 118, 35, 5, 232, 118, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 25, 59, 255, 5, 8, 59, 255, 5, 26, 75, 255, 5, 8, 75, 255, 5, 26, 91, 255, 5, 8, 91, 255, 5, 25, 155, 255, 5, 8, 155, 255, 1, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 83, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 206, 11, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 211, 11, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 212, 20, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 216, 20, 0, 0, 216, 20, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 90, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 49, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 116, 0, 114, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 116, 0, 97, 0, 110, 0, 100, 0, 97, 0, 114, 0, 100, 0, 93, 0, 0, 0, 74, 220, 38, 0, 67, 0, 60, 0, 99, 0, 39, 3, 60, 0, 60, 0, 60, 0, 67, 0, 39, 3, 38, 0, 71, 0, 60, 0, 103, 0, 6, 3, 60, 0, 60, 0, 60, 0, 71, 0, 6, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 105, 0, 60, 0, 49, 1, 60, 0, 60, 0, 60, 0, 73, 0, 38, 0, 105, 0, 60, 0, 60, 0, 60, 0, 48, 1, 38, 0, 79, 0, 60, 0, 111, 0, 8, 3, 60, 0, 60, 0, 60, 0, 79, 0, 8, 3, 38, 0, 83, 0, 60, 0, 115, 0, 39, 3, 60, 0, 60, 0, 60, 0, 83, 0, 39, 3, 38, 0, 85, 0, 60, 0, 117, 0, 8, 3, 60, 0, 60, 0, 60, 0, 85, 0, 8, 3, 0, 0, 170, 170, 170, 170, 170, 170, 4, 44, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 180, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 196, 37, 0, 0, 196, 37, 0, 0, 196, 37, 0, 0, 196, 37, 0, 0, 200, 41, 0, 0, 200, 41, 0, 0, 236, 43, 0, 0, 236, 43, 0, 0, 236, 43, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 79, 1, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 93, 0, 101, 0, 106, 0, 114, 0, 119, 0, 127, 0, 133, 0, 141, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 144, 0, 152, 0, 158, 0, 166, 0, 174, 0, 182, 0, 48, 0, 56, 0, 181, 0, 189, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 206, 0, 214, 0, 48, 0, 48, 0, 48, 0, 222, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 230, 0, 48, 0, 48, 0, 238, 0, 48, 0, 244, 0, 252, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 255, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 6, 1, 11, 1, 48, 0, 15, 1, 48, 0, 48, 0, 21, 1, 29, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 37, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 40, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 197, 0, 204, 0, 48, 0, 48, 0, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 70, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 56, 1, 48, 0, 48, 0, 62, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 116, 1, 168, 1, 220, 1, 20, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 2, 120, 2, 184, 2, 192, 0, 212, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 30, 0, 202, 1, 31, 0, 202, 34, 31, 0, 202, 67, 31, 0, 202, 100, 31, 0, 202, 133, 31, 0, 202, 166, 31, 0, 202, 199, 31, 0, 202, 232, 31, 0, 202, 9, 32, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 198, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 134, 1, 0, 156, 5, 0, 56, 201, 70, 2, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 6, 3, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 230, 3, 0, 156, 5, 0, 80, 201, 198, 4, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 230, 5, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 166, 6, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 102, 7, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 70, 8, 0, 5, 5, 0, 80, 201, 38, 9, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 194, 21, 0, 134, 5, 2, 47, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 197, 162, 0, 0, 197, 226, 0, 0, 197, 34, 1, 0, 197, 98, 1, 0, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 134, 5, 2, 71, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 5, 2, 83, 196, 136, 156, 90, 160, 5, 38, 93, 197, 2, 22, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 66, 22, 0, 5, 5, 2, 47, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 5, 5, 2, 71, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 5, 2, 83, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 55, 5, 5, 2, 55, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 10, 0, 201, 38, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 197, 98, 2, 0, 192, 0, 0, 0, 197, 162, 2, 0, 192, 0, 0, 0, 201, 230, 11, 0, 192, 0, 0, 0, 134, 5, 0, 58, 5, 5, 27, 57, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 12, 0, 201, 102, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 14, 0, 201, 230, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 15, 0, 201, 102, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 4, 0, 197, 34, 5, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 17, 0, 201, 230, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 192, 0, 0, 0, 197, 34, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 7, 0, 197, 162, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 252, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 166, 18, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 20, 0, 5, 71, 238, 102, 201, 102, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 8, 0, 192, 0, 0, 0, 197, 162, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 9, 0, 192, 0, 0, 0, 197, 98, 9, 0, 192, 0, 0, 0, 197, 162, 9, 0, 192, 0, 0, 0, 197, 226, 9, 0, 192, 0, 0, 0, 197, 34, 10, 0, 197, 99, 10, 0, 197, 194, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 11, 0, 197, 67, 11, 0, 197, 162, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 226, 7, 0, 197, 34, 12, 0, 5, 5, 10, 121, 197, 98, 12, 0, 197, 162, 12, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 197, 34, 14, 0, 197, 98, 14, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 8, 0, 5, 5, 24, 121, 197, 162, 8, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 14, 0, 197, 226, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 22, 0, 192, 0, 0, 0, 197, 99, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 23, 0, 201, 198, 23, 0, 201, 134, 24, 0, 201, 70, 25, 0, 201, 6, 26, 0, 201, 198, 26, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 17, 0, 192, 0, 0, 0, 201, 134, 27, 0, 192, 0, 0, 0, 201, 70, 28, 0, 201, 6, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 29, 0, 201, 134, 30, 0, 201, 70, 31, 0, 201, 6, 32, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 32, 0, 201, 134, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 83, 5, 136, 0, 0, 5, 5, 2, 83, 5, 136, 0, 0, 134, 5, 27, 57, 5, 138, 0, 0, 134, 5, 27, 57, 5, 136, 0, 0, 134, 5, 27, 57, 5, 142, 0, 0, 134, 5, 27, 57, 5, 150, 0, 0, 134, 5, 2, 55, 5, 160, 0, 0, 5, 5, 2, 55, 5, 160, 0, 0, 134, 5, 27, 57, 5, 154, 0, 0, 134, 5, 27, 57, 5, 164, 0, 0, 134, 5, 27, 57, 5, 140, 0, 0, 134, 5, 27, 57, 5, 162, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 134, 5, 27, 57, 5, 144, 0, 0, 134, 5, 2, 83, 5, 164, 0, 0, 5, 5, 2, 83, 5, 164, 0, 0, 134, 5, 2, 83, 5, 144, 0, 0, 5, 5, 2, 83, 5, 144, 0, 0, 134, 5, 2, 83, 5, 138, 0, 0, 5, 5, 2, 83, 5, 138, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 27, 57, 5, 184, 0, 0, 134, 5, 27, 57, 5, 188, 0, 0, 134, 5, 2, 71, 5, 164, 0, 0, 5, 5, 2, 71, 5, 164, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 134, 5, 2, 47, 5, 136, 0, 0, 5, 5, 2, 47, 5, 136, 0, 0, 134, 5, 27, 57, 5, 208, 0, 0, 134, 5, 27, 57, 5, 150, 0, 0, 5, 136, 0, 0, 134, 5, 2, 83, 5, 198, 0, 0, 5, 5, 2, 83, 5, 198, 0, 0, 134, 5, 2, 83, 5, 208, 0, 0, 5, 5, 2, 83, 5, 208, 0, 0, 134, 5, 2, 83, 5, 204, 0, 0, 5, 5, 2, 83, 5, 204, 0, 0, 134, 5, 27, 57, 5, 182, 0, 0, 134, 5, 27, 57, 5, 196, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 196, 0, 0, 5, 5, 2, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 226, 7, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 34, 8, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 98, 8, 0, 5, 5, 24, 121, 197, 162, 8, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 226, 8, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 34, 9, 0, 5, 5, 107, 121, 197, 98, 9, 0, 5, 5, 111, 121, 197, 162, 9, 0, 5, 5, 115, 121, 197, 226, 9, 0, 5, 5, 119, 121, 197, 34, 10, 0, 197, 99, 10, 0, 197, 194, 10, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 2, 11, 0, 197, 67, 11, 0, 197, 162, 11, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 226, 11, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 226, 7, 0, 197, 34, 12, 0, 5, 5, 10, 121, 197, 98, 12, 0, 197, 162, 12, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 226, 12, 0, 197, 34, 13, 0, 197, 98, 13, 0, 197, 162, 13, 0, 197, 226, 13, 0, 197, 34, 14, 0, 197, 98, 14, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 130, 8, 0, 5, 5, 24, 121, 197, 162, 8, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 0, 46, 156, 5, 48, 0, 39, 3, 2, 239, 134, 5, 0, 54, 156, 5, 48, 0, 6, 3, 2, 247, 134, 5, 27, 57, 134, 5, 48, 0, 7, 3, 0, 250, 134, 5, 0, 70, 156, 5, 48, 0, 8, 3, 255, 255, 2, 71, 134, 5, 0, 78, 156, 5, 48, 0, 39, 3, 255, 255, 2, 79, 134, 5, 0, 82, 156, 5, 1, 0, 8, 3, 255, 255, 2, 83, 134, 5, 68, 3, 197, 162, 0, 46, 5, 5, 48, 0, 39, 3, 2, 239, 5, 5, 0, 54, 5, 5, 48, 0, 6, 3, 2, 247, 5, 5, 0, 70, 5, 5, 48, 0, 8, 3, 255, 255, 2, 71, 5, 5, 0, 78, 5, 5, 48, 0, 39, 3, 255, 255, 2, 79, 5, 5, 0, 82, 5, 5, 1, 0, 8, 3, 255, 255, 2, 83, 5, 5, 68, 3, 0, 192, 197, 98, 156, 54, 196, 160, 48, 0, 6, 3, 1, 192, 197, 162, 5, 54, 196, 160, 48, 0, 6, 3, 1, 192, 197, 226, 2, 0, 197, 226, 48, 0, 7, 3, 134, 250, 196, 162, 156, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 34, 5, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 98, 156, 70, 196, 190, 48, 0, 8, 3, 3, 192, 197, 162, 5, 70, 196, 190, 48, 0, 8, 3, 3, 192, 197, 226, 156, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 34, 5, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 98, 156, 70, 196, 162, 48, 0, 8, 3, 6, 192, 197, 98, 5, 70, 196, 162, 48, 0, 8, 3, 6, 192, 197, 162, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 15, 0, 197, 34, 48, 0, 7, 3, 134, 250, 196, 208, 156, 82, 196, 198, 48, 0, 8, 3, 15, 192, 197, 194, 5, 82, 196, 198, 48, 0, 8, 3, 16, 192, 197, 2, 156, 82, 196, 208, 48, 0, 8, 3, 16, 192, 197, 66, 5, 82, 196, 208, 48, 0, 8, 3, 16, 192, 197, 130, 156, 82, 196, 204, 48, 0, 8, 3, 16, 192, 197, 194, 5, 82, 196, 204, 48, 0, 8, 3, 17, 192, 197, 2, 17, 0, 197, 130, 48, 0, 7, 3, 134, 250, 196, 196, 156, 70, 196, 196, 48, 0, 8, 3, 17, 192, 197, 194, 5, 70, 196, 196, 48, 0, 8, 3, 18, 192, 197, 2, 18, 0, 197, 67, 48, 0, 8, 3, 18, 192, 197, 163, 19, 0, 197, 3, 48, 0, 8, 3, 19, 192, 197, 99, 156, 82, 196, 196, 48, 0, 8, 3, 19, 192, 197, 194, 5, 82, 196, 196, 48, 0, 8, 3, 20, 192, 197, 2, 20, 0, 197, 67, 48, 0, 8, 3, 20, 192, 197, 163, 21, 0, 197, 3, 48, 0, 8, 3, 21, 192, 197, 99, 3, 0, 67, 0, 58, 0, 32, 0, 91, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 154, 37, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 132, 29, 0, 0, 132, 29, 0, 0, 132, 29, 0, 0, 132, 29, 0, 0, 164, 31, 0, 0, 164, 31, 0, 0, 130, 33, 0, 0, 130, 33, 0, 0, 130, 37, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 203, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 93, 0, 101, 0, 106, 0, 114, 0, 119, 0, 127, 0, 133, 0, 141, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 163, 0, 168, 0, 48, 0, 172, 0, 48, 0, 48, 0, 178, 0, 186, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 194, 0, 157, 0, 48, 0, 48, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 149, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 116, 1, 168, 1, 220, 1, 20, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 192, 15, 0, 202, 225, 15, 0, 202, 2, 16, 0, 202, 35, 16, 0, 202, 68, 16, 0, 202, 101, 16, 0, 202, 134, 16, 0, 202, 167, 16, 0, 202, 200, 16, 0, 202, 233, 16, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 6, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 198, 0, 0, 156, 5, 0, 56, 201, 134, 1, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 70, 2, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 38, 3, 0, 156, 5, 0, 80, 201, 6, 4, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 38, 5, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 230, 5, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 166, 6, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 134, 7, 0, 5, 5, 0, 80, 201, 102, 8, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 2, 15, 0, 134, 5, 2, 47, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 197, 162, 0, 0, 197, 226, 0, 0, 197, 34, 1, 0, 197, 98, 1, 0, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 134, 5, 2, 71, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 134, 5, 2, 83, 196, 136, 156, 90, 160, 5, 38, 93, 197, 66, 15, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 130, 15, 0, 5, 5, 2, 47, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 5, 5, 2, 71, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 5, 5, 2, 83, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 55, 5, 5, 2, 55, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 9, 0, 201, 102, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 197, 98, 2, 0, 192, 0, 0, 0, 197, 162, 2, 0, 192, 0, 0, 0, 201, 38, 11, 0, 192, 0, 0, 0, 134, 5, 0, 58, 5, 5, 27, 57, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 11, 0, 201, 166, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 13, 0, 201, 38, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 14, 0, 201, 166, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 4, 0, 197, 34, 5, 0, 197, 34, 0, 0, 197, 98, 0, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 16, 0, 201, 38, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 6, 0, 192, 0, 0, 0, 197, 34, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 7, 0, 197, 162, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 7, 0, 197, 34, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 17, 0, 192, 0, 0, 0, 197, 163, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 18, 0, 201, 102, 19, 0, 201, 38, 20, 0, 201, 230, 20, 0, 201, 166, 21, 0, 201, 102, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 10, 0, 192, 0, 0, 0, 201, 38, 23, 0, 192, 0, 0, 0, 201, 230, 23, 0, 201, 166, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 25, 0, 201, 38, 26, 0, 201, 230, 26, 0, 201, 166, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 28, 0, 201, 38, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 83, 5, 136, 0, 0, 5, 5, 2, 83, 5, 136, 0, 0, 134, 5, 27, 57, 5, 138, 0, 0, 134, 5, 27, 57, 5, 136, 0, 0, 134, 5, 27, 57, 5, 142, 0, 0, 134, 5, 27, 57, 5, 150, 0, 0, 134, 5, 2, 55, 5, 160, 0, 0, 5, 5, 2, 55, 5, 160, 0, 0, 134, 5, 27, 57, 5, 154, 0, 0, 134, 5, 27, 57, 5, 164, 0, 0, 134, 5, 27, 57, 5, 140, 0, 0, 134, 5, 27, 57, 5, 162, 0, 0, 134, 5, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 134, 5, 27, 57, 5, 144, 0, 0, 134, 5, 2, 83, 5, 164, 0, 0, 5, 5, 2, 83, 5, 164, 0, 0, 134, 5, 2, 83, 5, 144, 0, 0, 5, 5, 2, 83, 5, 144, 0, 0, 134, 5, 2, 83, 5, 138, 0, 0, 5, 5, 2, 83, 5, 138, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 27, 57, 5, 184, 0, 0, 134, 5, 27, 57, 5, 188, 0, 0, 134, 5, 2, 71, 5, 164, 0, 0, 5, 5, 2, 71, 5, 164, 0, 0, 134, 5, 2, 47, 5, 136, 0, 0, 5, 5, 2, 47, 5, 136, 0, 0, 134, 5, 27, 57, 5, 208, 0, 0, 134, 5, 27, 57, 5, 150, 0, 0, 5, 136, 0, 0, 134, 5, 2, 83, 5, 198, 0, 0, 5, 5, 2, 83, 5, 198, 0, 0, 134, 5, 2, 83, 5, 208, 0, 0, 5, 5, 2, 83, 5, 208, 0, 0, 134, 5, 2, 83, 5, 204, 0, 0, 5, 5, 2, 83, 5, 204, 0, 0, 134, 5, 27, 57, 5, 182, 0, 0, 134, 5, 27, 57, 5, 196, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 196, 0, 0, 5, 5, 2, 83, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 83, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 48, 0, 39, 3, 2, 239, 134, 5, 0, 54, 156, 5, 48, 0, 6, 3, 2, 247, 134, 5, 27, 57, 134, 5, 48, 0, 7, 3, 0, 250, 134, 5, 0, 70, 156, 5, 48, 0, 8, 3, 255, 255, 2, 71, 134, 5, 0, 78, 156, 5, 48, 0, 39, 3, 255, 255, 2, 79, 134, 5, 0, 82, 156, 5, 1, 0, 8, 3, 255, 255, 2, 83, 134, 5, 68, 3, 197, 162, 0, 46, 5, 5, 48, 0, 39, 3, 2, 239, 5, 5, 0, 54, 5, 5, 48, 0, 6, 3, 2, 247, 5, 5, 0, 70, 5, 5, 48, 0, 8, 3, 255, 255, 2, 71, 5, 5, 0, 78, 5, 5, 48, 0, 39, 3, 255, 255, 2, 79, 5, 5, 0, 82, 5, 5, 1, 0, 8, 3, 255, 255, 2, 83, 5, 5, 68, 3, 0, 192, 197, 98, 156, 54, 196, 160, 48, 0, 6, 3, 1, 192, 197, 162, 5, 54, 196, 160, 48, 0, 6, 3, 1, 192, 197, 226, 2, 0, 197, 226, 48, 0, 7, 3, 134, 250, 196, 162, 156, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 34, 5, 82, 196, 162, 48, 0, 8, 3, 3, 192, 197, 98, 156, 70, 196, 190, 48, 0, 8, 3, 3, 192, 197, 162, 5, 70, 196, 190, 48, 0, 8, 3, 3, 192, 197, 226, 156, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 34, 5, 82, 196, 190, 48, 0, 8, 3, 4, 192, 197, 98, 156, 70, 196, 162, 48, 0, 8, 3, 6, 192, 197, 98, 5, 70, 196, 162, 48, 0, 8, 3, 6, 192, 197, 162, 8, 0, 197, 98, 48, 0, 7, 3, 134, 250, 196, 208, 156, 82, 196, 198, 48, 0, 8, 3, 9, 192, 197, 2, 5, 82, 196, 198, 48, 0, 8, 3, 9, 192, 197, 66, 156, 82, 196, 208, 48, 0, 8, 3, 9, 192, 197, 130, 5, 82, 196, 208, 48, 0, 8, 3, 9, 192, 197, 194, 156, 82, 196, 204, 48, 0, 8, 3, 10, 192, 197, 2, 5, 82, 196, 204, 48, 0, 8, 3, 10, 192, 197, 66, 10, 0, 197, 194, 48, 0, 7, 3, 134, 250, 196, 196, 156, 70, 196, 196, 48, 0, 8, 3, 11, 192, 197, 2, 5, 70, 196, 196, 48, 0, 8, 3, 11, 192, 197, 66, 11, 0, 197, 131, 48, 0, 8, 3, 11, 192, 197, 227, 12, 0, 197, 67, 48, 0, 8, 3, 12, 192, 197, 163, 156, 82, 196, 196, 48, 0, 8, 3, 13, 192, 197, 2, 5, 82, 196, 196, 48, 0, 8, 3, 13, 192, 197, 66, 13, 0, 197, 131, 48, 0, 8, 3, 13, 192, 197, 227, 14, 0, 197, 67, 48, 0, 8, 3, 14, 192, 197, 163, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 72, 186, 76, 185, 80, 28, 4, 185, 92, 30, 4, 186, 104, 185, 108, 185, 112, 185, 116, 186, 120, 32, 4, 185, 136, 185, 140, 185, 148, 34, 4, 185, 160, 36, 4, 185, 176, 185, 180, 185, 184, 185, 188, 185, 192, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 38, 4, 168, 72, 168, 76, 168, 80, 40, 4, 168, 92, 168, 100, 168, 104, 168, 108, 168, 112, 168, 116, 168, 120, 42, 4, 168, 136, 168, 140, 168, 148, 44, 4, 168, 160, 46, 4, 168, 176, 168, 180, 168, 184, 168, 188, 168, 192, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 128, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 185, 68, 186, 77, 154, 77, 250, 77, 90, 78, 185, 97, 153, 97, 249, 97, 89, 98, 217, 72, 154, 122, 185, 129, 153, 129, 249, 129, 153, 130, 185, 132, 136, 14, 217, 130, 185, 169, 153, 169, 249, 169, 185, 172, 153, 189, 185, 196, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 168, 68, 168, 77, 136, 77, 232, 77, 72, 78, 168, 101, 136, 101, 232, 101, 72, 102, 200, 72, 136, 122, 168, 129, 136, 129, 232, 129, 136, 130, 168, 132, 128, 14, 200, 130, 168, 169, 136, 169, 232, 169, 168, 172, 136, 189, 168, 196, 72, 190, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 26, 79, 8, 79, 26, 78, 8, 78, 249, 85, 232, 85, 185, 88, 168, 88, 185, 86, 168, 86, 48, 4, 50, 4, 249, 93, 232, 93, 89, 95, 72, 95, 153, 98, 136, 102, 57, 99, 40, 103, 217, 97, 200, 101, 52, 4, 8, 103, 185, 100, 168, 96, 6, 8, 8, 8, 250, 105, 232, 105, 249, 110, 232, 110, 168, 144, 153, 113, 136, 113, 249, 114, 232, 114, 25, 114, 8, 114, 121, 115, 104, 115, 89, 115, 72, 115, 154, 121, 136, 121, 250, 122, 232, 122, 26, 122, 8, 122, 10, 8, 185, 124, 168, 124, 57, 131, 40, 131, 217, 129, 200, 129, 121, 130, 104, 130, 12, 8, 14, 8, 153, 149, 136, 149, 249, 150, 232, 150, 25, 150, 8, 150, 154, 153, 136, 153, 250, 153, 232, 153, 185, 156, 168, 156, 26, 154, 8, 154, 249, 162, 232, 162, 25, 162, 8, 162, 185, 164, 168, 164, 153, 170, 136, 170, 57, 171, 40, 171, 217, 169, 200, 169, 57, 170, 40, 170, 121, 170, 104, 170, 54, 4, 56, 4, 249, 181, 232, 181, 249, 189, 232, 189, 89, 190, 153, 193, 136, 193, 185, 194, 168, 194, 25, 194, 8, 194, 232, 152, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 152, 169, 152, 200, 56, 169, 76, 187, 100, 187, 104, 170, 100, 169, 104, 169, 200, 169, 120, 217, 128, 187, 76, 200, 128, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 255, 5, 185, 84, 255, 5, 185, 96, 255, 5, 185, 128, 255, 5, 186, 152, 255, 5, 185, 168, 255, 5, 168, 64, 255, 5, 168, 84, 255, 5, 168, 128, 255, 5, 168, 152, 255, 5, 168, 168, 255, 5, 249, 86, 255, 5, 232, 86, 255, 5, 25, 99, 255, 5, 25, 171, 255, 5, 8, 171, 255, 1, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 99, 11, 0, 16, 57, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 93, 11, 0, 32, 203, 20, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 208, 20, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 20, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 106, 23, 0, 0, 80, 0, 0, 0, 80, 0, 0, 0, 80, 0, 0, 0, 24, 35, 7, 0, 24, 35, 7, 0, 80, 96, 7, 0, 80, 96, 7, 0, 124, 191, 7, 0, 184, 39, 8, 0, 182, 73, 8, 0, 90, 78, 8, 0, 26, 82, 8, 0, 254, 84, 8, 0, 254, 85, 8, 0, 254, 85, 8, 0, 50, 105, 114, 84, 1, 0, 4, 23, 75, 111, 224, 6, 192, 0, 193, 1, 0, 0, 8, 0, 16, 0, 24, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 128, 0, 136, 0, 144, 0, 152, 0, 160, 0, 168, 0, 176, 0, 184, 0, 192, 0, 200, 0, 208, 0, 216, 0, 224, 0, 232, 0, 240, 0, 248, 0, 0, 1, 8, 1, 16, 1, 24, 1, 32, 1, 40, 1, 48, 1, 56, 1, 64, 1, 72, 1, 80, 1, 88, 1, 96, 1, 104, 1, 112, 1, 120, 1, 128, 1, 136, 1, 144, 1, 152, 1, 160, 1, 168, 1, 176, 1, 184, 1, 192, 1, 200, 1, 208, 1, 216, 1, 224, 1, 232, 1, 240, 1, 248, 1, 0, 2, 8, 2, 16, 2, 24, 2, 232, 89, 240, 89, 184, 89, 169, 103, 48, 0, 25, 103, 27, 88, 35, 88, 248, 90, 72, 90, 43, 88, 72, 93, 0, 91, 80, 90, 41, 104, 8, 92, 8, 91, 88, 90, 49, 104, 16, 92, 16, 91, 96, 90, 57, 104, 16, 90, 24, 91, 104, 90, 112, 89, 116, 95, 176, 91, 73, 104, 65, 104, 123, 95, 32, 91, 89, 104, 81, 104, 175, 94, 40, 91, 112, 90, 97, 104, 59, 100, 48, 91, 105, 104, 139, 95, 147, 95, 56, 91, 113, 104, 121, 104, 183, 94, 43, 105, 97, 99, 40, 92, 48, 0, 51, 105, 59, 105, 48, 92, 48, 0, 51, 88, 59, 88, 75, 105, 184, 91, 67, 88, 163, 95, 75, 88, 48, 0, 163, 105, 192, 91, 96, 93, 171, 105, 171, 95, 209, 102, 193, 102, 160, 93, 50, 107, 90, 107, 94, 107, 110, 107, 126, 107, 134, 107, 150, 107, 154, 107, 185, 103, 201, 103, 217, 103, 225, 103, 233, 103, 241, 103, 249, 103, 1, 104, 9, 104, 17, 104, 40, 90, 8, 93, 39, 95, 32, 106, 48, 106, 64, 106, 176, 92, 86, 106, 86, 106, 86, 106, 86, 106, 86, 106, 86, 106, 86, 106, 86, 106, 86, 106, 86, 106, 87, 106, 95, 106, 102, 106, 104, 106, 104, 106, 104, 106, 104, 106, 104, 106, 64, 93, 136, 92, 123, 106, 131, 106, 32, 93, 131, 105, 88, 93, 139, 105, 147, 105, 187, 105, 83, 88, 91, 88, 67, 100, 99, 88, 8, 106, 0, 106, 16, 106, 248, 105, 24, 106, 112, 106, 115, 106, 123, 105, 80, 92, 40, 93, 195, 105, 203, 105, 211, 105, 75, 100, 187, 95, 168, 93, 219, 105, 226, 105, 107, 88, 83, 100, 120, 89, 33, 102, 48, 0, 64, 91, 120, 90, 16, 93, 115, 88, 200, 91, 91, 100, 155, 105, 128, 90, 107, 105, 136, 90, 99, 100, 152, 93, 113, 102, 217, 102, 123, 88, 129, 88, 129, 101, 89, 101, 81, 101, 169, 101, 145, 101, 225, 101, 128, 89, 136, 89, 105, 101, 241, 101, 17, 102, 25, 102, 121, 101, 97, 101, 217, 101, 9, 102, 41, 102, 65, 102, 73, 102, 57, 102, 49, 102, 23, 95, 15, 95, 7, 95, 137, 88, 145, 88, 144, 92, 153, 88, 8, 94, 105, 99, 200, 89, 144, 89, 79, 95, 203, 95, 211, 95, 177, 101, 219, 95, 226, 95, 231, 95, 239, 95, 247, 95, 87, 96, 103, 96, 79, 96, 111, 96, 127, 96, 135, 96, 119, 96, 88, 94, 135, 94, 143, 96, 143, 96, 143, 96, 143, 96, 143, 96, 143, 96, 143, 96, 150, 96, 158, 96, 16, 94, 24, 94, 40, 94, 57, 101, 171, 100, 166, 96, 167, 96, 167, 96, 167, 96, 167, 96, 175, 96, 181, 96, 184, 96, 227, 96, 231, 96, 237, 96, 242, 96, 248, 96, 252, 96, 252, 96, 252, 96, 12, 97, 20, 97, 21, 97, 127, 94, 29, 97, 34, 97, 112, 94, 119, 94, 98, 97, 98, 97, 98, 97, 98, 97, 98, 97, 98, 97, 98, 97, 98, 97, 42, 97, 42, 97, 42, 97, 42, 97, 104, 94, 50, 97, 159, 94, 96, 94, 95, 96, 58, 97, 58, 97, 0, 93, 66, 97, 66, 97, 67, 97, 75, 97, 75, 97, 75, 97, 75, 97, 78, 97, 84, 97, 90, 97, 90, 97, 90, 97, 153, 102, 137, 102, 161, 102, 137, 101, 81, 102, 89, 102, 97, 102, 104, 89, 201, 102, 225, 102, 177, 103, 161, 88, 193, 103, 25, 104, 33, 104, 105, 102, 200, 92, 184, 92, 192, 92, 48, 0, 42, 110, 26, 110, 10, 111, 58, 111, 242, 109, 138, 110, 194, 110, 234, 110, 2, 111, 42, 111, 66, 111, 50, 99, 128, 92, 112, 92, 170, 107, 210, 107, 208, 89, 208, 92, 218, 107, 216, 92, 2, 108, 58, 107, 74, 107, 82, 107, 73, 99, 10, 108, 58, 99, 65, 99, 56, 94, 64, 94, 72, 94, 81, 99, 250, 109, 235, 100, 211, 100, 226, 107, 178, 107, 202, 107, 179, 100, 219, 100, 185, 101, 193, 101, 113, 101, 203, 100, 144, 2, 96, 3, 253, 3, 87, 4, 240, 4, 120, 5, 64, 6, 48, 7, 168, 7, 8, 8, 224, 8, 136, 9, 35, 10, 151, 10, 206, 10, 26, 11, 224, 7, 150, 12, 75, 13, 246, 13, 141, 14, 84, 15, 193, 15, 169, 15, 90, 16, 136, 16, 28, 17, 172, 17, 140, 18, 243, 18, 82, 19, 174, 19, 2, 20, 80, 20, 216, 20, 45, 21, 140, 21, 18, 22, 161, 22, 9, 23, 101, 23, 198, 23, 22, 24, 83, 24, 152, 24, 13, 25, 178, 25, 31, 26, 172, 26, 12, 27, 63, 27, 170, 27, 26, 28, 130, 28, 2, 29, 154, 29, 250, 29, 105, 30, 9, 31, 134, 31, 211, 31, 13, 32, 132, 32, 140, 32, 54, 33, 172, 33, 220, 33, 60, 34, 146, 34, 14, 35, 138, 35, 209, 35, 58, 36, 135, 36, 243, 36, 113, 37, 185, 37, 45, 38, 177, 38, 84, 39, 188, 39, 131, 40, 201, 40, 25, 41, 110, 41, 213, 41, 43, 42, 128, 42, 230, 42, 70, 43, 117, 43, 213, 43, 75, 44, 105, 44, 167, 44, 53, 45, 157, 45, 37, 46, 83, 46, 146, 46, 254, 46, 131, 47, 214, 47, 11, 48, 131, 48, 17, 49, 129, 49, 176, 49, 238, 49, 62, 50, 154, 50, 7, 51, 71, 51, 137, 51, 186, 51, 231, 51, 140, 52, 244, 52, 72, 53, 159, 53, 228, 53, 86, 54, 220, 54, 68, 55, 210, 55, 26, 56, 136, 56, 238, 56, 69, 57, 134, 57, 183, 57, 94, 58, 150, 58, 246, 58, 52, 59, 147, 59, 6, 60, 56, 60, 119, 60, 48, 61, 156, 61, 244, 61, 127, 62, 247, 62, 59, 63, 194, 63, 33, 64, 121, 64, 199, 64, 47, 65, 111, 65, 204, 65, 43, 66, 135, 66, 20, 67, 131, 67, 227, 67, 106, 68, 154, 68, 248, 68, 41, 69, 141, 69, 33, 70, 120, 70, 232, 70, 120, 71, 71, 72, 175, 72, 33, 73, 176, 73, 104, 74, 207, 74, 94, 75, 34, 76, 145, 76, 6, 77, 228, 52, 213, 77, 60, 78, 138, 78, 202, 78, 25, 79, 97, 79, 149, 79, 237, 79, 75, 80, 119, 80, 207, 80, 103, 81, 179, 81, 243, 81, 99, 82, 200, 82, 61, 83, 180, 83, 233, 83, 71, 84, 214, 84, 22, 85, 91, 85, 174, 85, 106, 86, 176, 86, 248, 86, 110, 87, 208, 2, 106, 97, 106, 97, 64, 2, 152, 2, 248, 2, 56, 3, 104, 2, 160, 3, 224, 3, 229, 3, 233, 3, 238, 3, 13, 4, 245, 3, 37, 4, 42, 4, 74, 4, 79, 4, 127, 4, 127, 4, 162, 4, 162, 4, 194, 4, 200, 4, 232, 4, 32, 5, 64, 5, 80, 5, 152, 5, 184, 5, 24, 6, 80, 6, 120, 6, 184, 6, 8, 7, 40, 7, 64, 7, 104, 7, 160, 7, 216, 7, 248, 7, 32, 8, 64, 8, 104, 8, 112, 2, 176, 8, 240, 8, 24, 9, 72, 9, 144, 9, 192, 9, 216, 9, 217, 9, 239, 9, 247, 9, 19, 10, 27, 10, 80, 10, 80, 10, 86, 10, 135, 10, 143, 10, 186, 10, 190, 10, 10, 11, 18, 11, 66, 11, 74, 11, 198, 10, 138, 11, 218, 11, 225, 11, 70, 12, 134, 12, 158, 12, 206, 12, 222, 12, 230, 12, 21, 13, 45, 13, 51, 13, 122, 13, 125, 13, 164, 13, 172, 13, 201, 13, 238, 13, 46, 14, 70, 14, 176, 9, 173, 14, 229, 14, 29, 15, 76, 15, 156, 15, 161, 15, 185, 15, 217, 15, 224, 15, 0, 16, 16, 16, 24, 16, 72, 16, 77, 16, 82, 16, 128, 16, 160, 16, 197, 16, 244, 16, 20, 17, 92, 17, 116, 17, 164, 17, 180, 17, 36, 18, 16, 3, 164, 18, 211, 18, 34, 19, 66, 19, 74, 19, 98, 19, 114, 19, 166, 19, 204, 19, 226, 19, 250, 19, 24, 20, 72, 20, 136, 20, 192, 20, 232, 20, 13, 21, 37, 21, 109, 21, 132, 21, 164, 21, 218, 21, 50, 22, 106, 22, 153, 22, 217, 22, 249, 22, 33, 23, 85, 23, 125, 23, 133, 23, 157, 23, 189, 23, 190, 23, 226, 23, 234, 23, 10, 24, 14, 24, 60, 24, 67, 24, 112, 24, 75, 24, 144, 24, 204, 24, 224, 24, 5, 25, 99, 25, 130, 25, 170, 25, 218, 25, 234, 25, 241, 25, 17, 26, 23, 26, 70, 26, 76, 26, 105, 26, 112, 26, 160, 26, 164, 26, 202, 26, 204, 26, 211, 26, 4, 27, 219, 26, 48, 27, 55, 27, 98, 27, 105, 27, 146, 27, 162, 27, 234, 27, 42, 28, 90, 28, 154, 28, 234, 28, 42, 29, 106, 29, 122, 29, 146, 29, 170, 29, 210, 29, 242, 29, 49, 30, 89, 30, 218, 29, 56, 6, 233, 30, 25, 31, 41, 31, 49, 31, 54, 31, 92, 31, 92, 31, 96, 31, 120, 31, 120, 31, 126, 31, 166, 31, 166, 31, 174, 31, 198, 31, 198, 31, 203, 31, 255, 31, 255, 31, 5, 32, 76, 32, 76, 32, 84, 32, 116, 32, 89, 32, 189, 32, 124, 32, 225, 32, 233, 32, 17, 33, 46, 33, 101, 33, 156, 33, 204, 33, 244, 33, 52, 34, 106, 34, 154, 34, 218, 34, 54, 35, 216, 5, 101, 35, 108, 35, 124, 35, 130, 35, 154, 35, 161, 35, 192, 35, 198, 35, 201, 35, 241, 35, 243, 35, 250, 35, 35, 36, 35, 36, 43, 36, 113, 36, 113, 36, 50, 36, 121, 36, 176, 36, 127, 36, 184, 36, 190, 36, 227, 36, 235, 36, 43, 37, 51, 37, 105, 37, 59, 37, 137, 37, 177, 37, 209, 37, 247, 37, 29, 38, 61, 38, 101, 38, 109, 38, 138, 38, 169, 38, 217, 38, 225, 38, 230, 38, 61, 39, 68, 39, 140, 39, 180, 39, 211, 39, 19, 40, 83, 40, 171, 40, 193, 40, 249, 40, 17, 41, 62, 41, 86, 41, 102, 41, 126, 41, 190, 41, 229, 41, 245, 41, 5, 42, 29, 42, 35, 42, 91, 42, 120, 42, 158, 42, 182, 42, 206, 42, 254, 42, 38, 43, 94, 43, 141, 43, 157, 43, 205, 43, 243, 43, 43, 44, 67, 44, 97, 44, 129, 44, 145, 44, 183, 44, 238, 44, 30, 45, 77, 45, 141, 45, 197, 45, 5, 46, 29, 46, 53, 46, 106, 46, 130, 46, 180, 46, 223, 46, 22, 47, 84, 47, 115, 47, 123, 47, 158, 47, 198, 47, 206, 47, 230, 47, 251, 47, 43, 48, 75, 48, 107, 48, 155, 48, 177, 48, 241, 48, 65, 49, 89, 49, 121, 49, 152, 49, 168, 49, 6, 50, 46, 50, 92, 50, 130, 50, 194, 50, 226, 50, 255, 50, 23, 51, 47, 51, 55, 51, 101, 51, 63, 51, 163, 51, 167, 51, 223, 51, 7, 52, 61, 52, 116, 52, 132, 52, 164, 52, 236, 52, 58, 53, 64, 53, 96, 53, 128, 53, 144, 53, 182, 53, 151, 53, 214, 53, 220, 53, 3, 54, 27, 54, 59, 54, 109, 54, 140, 54, 164, 54, 188, 54, 196, 54, 212, 54, 252, 54, 44, 55, 84, 55, 139, 55, 194, 55, 234, 55, 10, 56, 50, 56, 82, 56, 98, 56, 176, 56, 200, 56, 208, 56, 224, 56, 230, 56, 22, 57, 61, 57, 93, 57, 126, 57, 249, 30, 175, 57, 230, 57, 30, 58, 86, 58, 110, 58, 158, 58, 190, 58, 222, 58, 230, 58, 14, 59, 18, 59, 238, 58, 39, 59, 44, 59, 71, 59, 76, 59, 131, 59, 131, 59, 131, 59, 49, 49, 139, 59, 195, 59, 199, 59, 76, 39, 247, 59, 254, 59, 47, 60, 48, 60, 103, 60, 111, 60, 180, 60, 186, 60, 248, 60, 40, 61, 89, 61, 0, 61, 16, 9, 204, 61, 236, 61, 252, 61, 12, 62, 60, 62, 82, 62, 88, 62, 95, 62, 156, 62, 164, 62, 211, 62, 215, 62, 172, 62, 239, 62, 75, 63, 130, 63, 170, 63, 186, 63, 210, 63, 226, 63, 250, 63, 25, 64, 81, 64, 105, 64, 137, 64, 185, 48, 207, 64, 123, 48, 39, 65, 87, 65, 103, 65, 135, 65, 143, 65, 175, 65, 183, 65, 220, 65, 227, 65, 11, 66, 35, 66, 79, 66, 103, 66, 127, 66, 158, 66, 166, 66, 198, 66, 204, 66, 228, 66, 12, 67, 52, 67, 108, 67, 155, 67, 171, 67, 211, 67, 170, 28, 34, 68, 50, 68, 58, 68, 98, 68, 132, 68, 224, 68, 240, 68, 14, 69, 33, 69, 65, 69, 125, 69, 133, 69, 219, 69, 17, 70, 49, 70, 81, 70, 97, 70, 136, 70, 200, 70, 248, 70, 8, 71, 32, 71, 48, 71, 88, 71, 104, 71, 160, 71, 176, 71, 227, 71, 42, 72, 63, 72, 87, 72, 127, 72, 159, 72, 205, 72, 251, 72, 17, 73, 49, 73, 88, 73, 136, 73, 160, 73, 168, 73, 192, 73, 195, 73, 203, 73, 207, 73, 223, 73, 226, 73, 17, 74, 23, 74, 30, 74, 51, 74, 57, 74, 96, 74, 125, 74, 129, 74, 166, 74, 172, 74, 199, 74, 227, 74, 40, 75, 46, 75, 70, 75, 86, 75, 102, 75, 118, 75, 125, 75, 165, 75, 228, 75, 244, 75, 26, 76, 81, 76, 121, 76, 137, 76, 161, 76, 201, 76, 225, 76, 254, 76, 76, 77, 122, 77, 176, 77, 205, 77, 12, 78, 52, 78, 116, 78, 146, 78, 210, 78, 1, 79, 49, 79, 89, 79, 125, 79, 165, 79, 213, 79, 229, 79, 19, 80, 83, 80, 111, 80, 135, 80, 175, 80, 239, 80, 7, 81, 47, 81, 79, 81, 95, 81, 131, 81, 171, 81, 203, 81, 235, 81, 19, 82, 59, 82, 83, 82, 147, 82, 176, 82, 246, 82, 45, 83, 132, 83, 164, 83, 172, 83, 204, 83, 209, 83, 250, 83, 2, 84, 55, 84, 111, 84, 158, 84, 174, 84, 206, 84, 230, 84, 246, 84, 14, 85, 50, 85, 71, 65, 83, 85, 126, 85, 144, 85, 63, 84, 195, 85, 2, 86, 34, 86, 50, 86, 98, 86, 153, 86, 108, 39, 8, 87, 62, 87, 126, 87, 158, 87, 197, 87, 2, 22, 48, 3, 176, 2, 180, 21, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 18, 108, 138, 97, 142, 97, 149, 97, 232, 92, 187, 106, 195, 106, 202, 106, 210, 106, 217, 106, 220, 106, 226, 106, 233, 106, 240, 92, 107, 100, 129, 102, 169, 88, 121, 102, 241, 106, 241, 106, 48, 90, 47, 95, 55, 95, 1, 102, 209, 101, 63, 95, 153, 101, 201, 101, 233, 101, 208, 91, 155, 95, 115, 105, 80, 93, 216, 91, 137, 104, 72, 91, 177, 88, 115, 100, 72, 92, 176, 93, 102, 107, 80, 91, 144, 90, 24, 93, 89, 99, 234, 105, 240, 105, 104, 93, 179, 95, 67, 105, 56, 92, 64, 92, 48, 93, 209, 103, 161, 101, 71, 95, 31, 95, 40, 106, 56, 106, 129, 104, 24, 92, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 200, 96, 48, 2, 142, 107, 162, 107, 40, 2, 158, 98, 151, 98, 143, 98, 63, 98, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 216, 97, 122, 97, 122, 97, 122, 97, 122, 97, 122, 97, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 34, 110, 82, 110, 58, 110, 10, 110, 114, 110, 66, 110, 90, 110, 74, 110, 152, 9, 20, 62, 186, 110, 18, 110, 154, 110, 226, 110, 34, 111, 48, 0, 224, 89, 71, 96, 233, 102, 105, 103, 137, 103, 100, 95, 108, 95, 249, 102, 9, 103, 153, 103, 248, 89, 8, 90, 1, 103, 81, 103, 17, 103, 0, 90, 129, 103, 151, 94, 89, 103, 145, 103, 73, 103, 97, 103, 56, 2, 92, 95, 72, 89, 80, 89, 152, 92, 88, 89, 241, 102, 121, 103, 161, 103, 96, 89, 168, 92, 160, 92, 80, 94, 224, 92, 216, 89, 66, 107, 118, 107, 32, 2, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 0, 1, 64, 1, 128, 1, 192, 1, 0, 2, 64, 2, 128, 2, 192, 2, 0, 3, 64, 3, 128, 3, 192, 3, 0, 4, 64, 4, 128, 4, 192, 4, 0, 5, 64, 5, 128, 5, 192, 5, 0, 6, 64, 6, 128, 6, 192, 6, 0, 7, 64, 7, 128, 7, 192, 7, 0, 8, 64, 8, 33, 18, 97, 20, 97, 18, 161, 20, 33, 21, 224, 6, 161, 18, 224, 6, 1, 22, 224, 6, 224, 6, 224, 6, 224, 6, 225, 18, 65, 22, 65, 22, 66, 22, 130, 22, 224, 6, 224, 6, 224, 6, 224, 6, 225, 21, 33, 19, 224, 6, 224, 6, 97, 19, 161, 19, 225, 20, 225, 19, 97, 21, 161, 21, 33, 10, 97, 11, 161, 13, 225, 12, 33, 12, 225, 13, 225, 14, 161, 12, 161, 14, 161, 11, 225, 11, 161, 16, 97, 13, 33, 15, 97, 17, 97, 14, 33, 14, 97, 15, 33, 16, 225, 15, 225, 9, 33, 13, 225, 10, 33, 11, 97, 16, 97, 10, 97, 12, 161, 15, 33, 17, 161, 17, 224, 6, 194, 22, 161, 10, 225, 16, 225, 17, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 224, 6, 33, 20, 34, 85, 58, 85, 64, 85, 67, 85, 99, 85, 102, 85, 134, 85, 136, 85, 152, 85, 157, 85, 161, 85, 166, 85, 182, 85, 187, 85, 203, 85, 219, 85, 227, 85, 234, 85, 250, 85, 82, 86, 90, 86, 114, 86, 121, 86, 137, 86, 189, 14, 161, 86, 168, 86, 184, 86, 200, 86, 216, 86, 224, 86, 232, 86, 240, 86, 0, 87, 16, 87, 24, 87, 32, 87, 38, 87, 46, 87, 70, 87, 86, 87, 118, 87, 134, 87, 142, 87, 150, 87, 166, 87, 174, 87, 181, 87, 205, 87, 212, 87, 219, 87, 235, 87, 243, 87, 3, 88, 72, 2, 48, 0, 120, 2, 168, 3, 111, 4, 16, 5, 176, 5, 128, 6, 24, 7, 24, 8, 80, 2, 160, 2, 184, 2, 232, 2, 0, 3, 24, 3, 32, 3, 64, 3, 112, 3, 136, 3, 152, 3, 176, 3, 184, 3, 216, 3, 5, 4, 21, 4, 29, 4, 50, 4, 95, 4, 103, 4, 135, 4, 142, 4, 146, 4, 170, 4, 208, 4, 216, 4, 248, 4, 0, 5, 40, 5, 8, 5, 48, 5, 72, 5, 88, 5, 104, 5, 136, 5, 144, 5, 160, 5, 192, 5, 208, 5, 224, 5, 240, 5, 8, 6, 32, 6, 48, 6, 72, 6, 96, 6, 112, 6, 136, 6, 152, 6, 176, 6, 192, 6, 216, 6, 224, 6, 0, 7, 16, 7, 56, 7, 72, 7, 88, 7, 112, 7, 120, 7, 128, 7, 136, 7, 152, 7, 176, 7, 44, 62, 231, 62, 162, 63, 57, 64, 239, 64, 127, 65, 63, 66, 182, 66, 220, 66, 244, 66, 76, 67, 235, 67, 66, 68, 208, 68, 85, 69, 235, 69, 208, 70, 72, 71, 211, 71, 2, 72, 103, 72, 213, 72, 120, 73, 1, 74, 158, 74, 32, 75, 78, 75, 141, 75, 181, 75, 113, 76, 185, 76, 44, 77, 197, 77, 84, 78, 57, 79, 221, 79, 59, 80, 223, 80, 15, 81, 55, 81, 35, 82, 139, 82, 85, 83, 47, 84, 150, 84, 222, 84, 6, 85, 110, 85, 10, 86, 58, 86, 129, 86, 78, 87, 251, 87, 88, 2, 168, 2, 200, 2, 8, 3, 72, 3, 144, 3, 208, 3, 58, 4, 154, 4, 224, 4, 24, 5, 96, 2, 240, 2, 88, 3, 200, 3, 66, 4, 186, 4, 112, 5, 0, 6, 144, 6, 32, 7, 184, 7, 96, 8, 168, 8, 40, 9, 128, 9, 72, 10, 2, 11, 130, 11, 22, 12, 37, 13, 156, 13, 38, 14, 125, 14, 53, 15, 248, 15, 12, 17, 156, 17, 252, 17, 227, 18, 158, 19, 104, 20, 61, 21, 210, 21, 177, 22, 117, 23, 2, 24, 136, 24, 67, 25, 9, 26, 97, 26, 194, 26, 90, 27, 2, 28, 18, 29, 202, 29, 57, 30, 161, 30, 158, 31, 68, 32, 181, 32, 85, 33, 12, 34, 83, 34, 6, 35, 93, 35, 233, 35, 105, 36, 168, 36, 35, 37, 169, 37, 93, 38, 209, 38, 53, 39, 172, 39, 115, 48, 1, 49, 200, 49, 186, 50, 93, 51, 255, 51, 156, 52, 112, 53, 43, 54, 172, 54, 28, 55, 178, 55, 74, 56, 45, 57, 17, 31, 166, 58, 107, 59, 171, 59, 23, 60, 164, 60, 24, 61, 180, 61, 119, 62, 43, 63, 49, 64, 223, 64, 251, 65, 174, 66, 252, 66, 243, 67, 216, 68, 179, 69, 176, 70, 64, 71, 203, 71, 119, 72, 104, 73, 249, 73, 150, 74, 24, 75, 133, 75, 4, 76, 177, 76, 160, 77, 76, 78, 117, 79, 35, 80, 255, 80, 63, 81, 27, 82, 216, 82, 225, 83, 127, 84, 182, 84, 75, 85, 18, 86, 66, 86, 102, 87, 128, 2, 14, 14, 58, 29, 201, 38, 4, 53, 212, 66, 157, 75, 136, 2, 224, 2, 40, 3, 192, 3, 119, 4, 178, 4, 56, 5, 128, 5, 200, 5, 104, 6, 208, 6, 96, 7, 208, 7, 72, 8, 160, 8, 248, 8, 56, 9, 120, 9, 11, 10, 119, 10, 170, 10, 242, 10, 106, 11, 210, 11, 46, 12, 110, 12, 5, 13, 148, 13, 22, 14, 157, 14, 45, 15, 132, 15, 240, 15, 56, 16, 152, 16, 220, 16, 4, 17, 132, 17, 196, 17, 236, 17, 20, 18, 100, 18, 195, 18, 142, 19, 96, 20, 208, 20, 101, 21, 234, 21, 66, 22, 185, 22, 61, 23, 218, 23, 104, 24, 188, 24, 75, 25, 146, 25, 54, 26, 144, 26, 244, 26, 82, 27, 242, 27, 106, 28, 194, 28, 192, 7, 200, 7, 240, 7, 0, 8, 16, 8, 40, 8, 56, 8, 80, 8, 88, 8, 192, 2, 112, 8, 120, 8, 136, 8, 144, 8, 80, 3, 192, 8, 208, 8, 232, 8, 248, 5, 8, 9, 32, 9, 48, 9, 64, 9, 200, 6, 88, 9, 112, 9, 200, 8, 16, 6, 160, 9, 184, 9, 200, 9, 225, 9, 231, 9, 255, 9, 3, 10, 43, 10, 43, 10, 48, 10, 94, 10, 101, 10, 104, 10, 111, 10, 159, 10, 159, 10, 159, 10, 162, 10, 214, 10, 222, 10, 223, 10, 226, 10, 34, 11, 42, 11, 47, 11, 47, 11, 50, 11, 82, 11, 82, 11, 82, 11, 82, 11, 90, 11, 146, 11, 153, 11, 160, 11, 186, 11, 134, 41, 158, 41, 174, 41, 182, 41, 198, 41, 205, 41, 221, 41, 237, 41, 253, 41, 13, 42, 51, 42, 67, 42, 99, 42, 104, 42, 136, 42, 142, 42, 166, 42, 190, 42, 198, 42, 214, 42, 238, 42, 6, 43, 14, 43, 22, 43, 46, 43, 62, 43, 78, 43, 102, 43, 109, 43, 125, 43, 116, 18, 193, 22, 149, 43, 165, 43, 181, 43, 104, 3, 221, 43, 227, 43, 216, 2, 3, 44, 11, 44, 35, 44, 51, 44, 83, 44, 89, 44, 113, 44, 121, 44, 137, 44, 153, 44, 159, 44, 175, 44, 191, 44, 199, 44, 223, 44, 230, 44, 246, 44, 254, 44, 14, 45, 38, 45, 45, 45, 61, 45, 93, 45, 101, 45, 117, 45, 125, 45, 133, 45, 149, 45, 165, 45, 189, 45, 168, 6, 213, 45, 237, 45, 245, 45, 13, 46, 45, 46, 61, 46, 67, 46, 91, 46, 98, 46, 10, 29, 114, 46, 138, 46, 154, 46, 157, 46, 164, 46, 188, 46, 191, 46, 231, 46, 239, 46, 246, 46, 6, 47, 14, 47, 30, 47, 38, 47, 54, 47, 60, 47, 76, 47, 92, 47, 99, 47, 139, 47, 144, 47, 150, 47, 166, 47, 222, 47, 238, 47, 243, 47, 3, 48, 19, 48, 51, 48, 67, 48, 83, 48, 91, 48, 99, 48, 139, 48, 163, 48, 169, 48, 193, 48, 209, 48, 233, 48, 249, 48, 9, 49, 33, 49, 120, 3, 81, 49, 97, 49, 137, 49, 144, 49, 160, 49, 152, 20, 160, 20, 184, 20, 200, 20, 224, 20, 240, 20, 245, 20, 5, 21, 21, 21, 29, 21, 53, 21, 69, 21, 85, 21, 128, 3, 248, 6, 117, 21, 124, 21, 52, 17, 156, 21, 172, 21, 188, 21, 194, 21, 226, 21, 250, 21, 10, 22, 26, 22, 58, 22, 232, 6, 82, 22, 98, 22, 114, 22, 121, 22, 129, 22, 169, 22, 201, 22, 168, 20, 225, 22, 233, 22, 241, 22, 17, 23, 25, 23, 41, 23, 46, 23, 53, 23, 77, 23, 93, 23, 109, 23, 141, 23, 165, 23, 173, 23, 206, 23, 210, 23, 242, 23, 242, 23, 30, 24, 30, 24, 30, 24, 36, 24, 91, 24, 91, 24, 96, 24, 120, 24, 120, 24, 160, 24, 96, 5, 168, 5, 232, 5, 40, 6, 88, 6, 160, 6, 240, 6, 80, 7, 144, 7, 232, 7, 48, 8, 128, 8, 0, 9, 80, 9, 208, 9, 64, 10, 127, 10, 178, 10, 250, 10, 58, 11, 114, 11, 122, 11, 178, 11, 246, 11, 14, 12, 54, 12, 94, 12, 174, 12, 13, 13, 114, 13, 193, 13, 30, 14, 109, 14, 205, 14, 253, 14, 100, 15, 64, 16, 189, 16, 36, 17, 108, 17, 204, 17, 12, 18, 68, 18, 203, 18, 26, 19, 150, 19, 242, 19, 48, 20, 144, 20, 253, 20, 93, 21, 148, 21, 242, 21, 74, 22, 145, 22, 1, 23, 69, 23, 181, 23, 250, 23, 52, 24, 128, 24, 196, 24, 253, 24, 83, 25, 109, 33, 116, 33, 132, 33, 13, 15, 164, 33, 180, 33, 196, 33, 212, 33, 228, 33, 236, 33, 252, 33, 152, 8, 90, 22, 88, 20, 36, 34, 68, 34, 75, 34, 91, 34, 98, 34, 114, 34, 130, 34, 162, 34, 178, 34, 194, 34, 202, 34, 210, 34, 226, 34, 242, 34, 248, 34, 254, 34, 22, 35, 30, 35, 62, 35, 69, 35, 77, 35, 85, 35, 116, 35, 146, 35, 169, 35, 176, 35, 217, 35, 2, 36, 2, 36, 5, 36, 66, 36, 66, 36, 11, 36, 74, 36, 81, 36, 143, 36, 144, 36, 198, 36, 198, 36, 198, 36, 203, 36, 251, 36, 209, 22, 3, 37, 11, 37, 67, 37, 67, 37, 73, 37, 81, 37, 121, 37, 120, 16, 112, 16, 144, 16, 168, 16, 173, 16, 205, 16, 212, 16, 228, 16, 236, 16, 252, 16, 184, 8, 44, 17, 68, 17, 84, 17, 100, 17, 124, 17, 140, 17, 148, 17, 188, 17, 212, 17, 168, 9, 228, 17, 244, 17, 4, 18, 28, 18, 44, 18, 52, 18, 76, 18, 92, 18, 108, 18, 124, 18, 132, 18, 148, 18, 156, 18, 172, 18, 180, 18, 187, 18, 219, 18, 235, 18, 251, 18, 11, 19, 18, 19, 42, 19, 50, 19, 58, 19, 90, 19, 106, 19, 122, 19, 126, 19, 182, 19, 182, 19, 188, 19, 212, 19, 218, 19, 234, 19, 10, 20, 16, 20, 32, 20, 40, 20, 56, 20, 140, 15, 64, 20, 112, 20, 128, 20, 216, 8, 104, 9, 56, 10, 234, 10, 98, 11, 38, 12, 253, 12, 106, 13, 230, 13, 93, 14, 5, 15, 116, 15, 48, 16, 181, 16, 76, 17, 220, 17, 84, 18, 3, 19, 196, 19, 120, 20, 77, 21, 202, 21, 137, 22, 149, 23, 44, 24, 180, 24, 59, 25, 1, 26, 136, 26, 32, 27, 194, 27, 202, 28, 82, 29, 186, 29, 73, 30, 201, 30, 112, 31, 231, 31, 44, 32, 165, 32, 9, 33, 124, 33, 82, 28, 38, 35, 19, 36, 152, 36, 19, 37, 129, 37, 7, 38, 130, 38, 250, 38, 100, 39, 235, 39, 115, 40, 118, 41, 75, 42, 150, 42, 222, 42, 133, 43, 251, 43, 69, 45, 253, 45, 207, 46, 174, 47, 136, 53, 167, 53, 174, 53, 190, 53, 236, 53, 243, 53, 251, 53, 11, 54, 19, 54, 35, 54, 67, 54, 70, 54, 94, 54, 101, 54, 117, 54, 124, 54, 132, 54, 236, 54, 4, 55, 12, 55, 20, 55, 134, 19, 52, 55, 42, 22, 76, 55, 100, 55, 124, 55, 131, 55, 96, 9, 147, 55, 154, 55, 162, 55, 186, 55, 218, 55, 172, 24, 242, 55, 250, 55, 2, 56, 18, 56, 42, 56, 58, 56, 66, 56, 90, 56, 106, 56, 112, 56, 120, 56, 144, 56, 152, 56, 160, 56, 184, 56, 193, 30, 216, 56, 246, 56, 254, 56, 6, 57, 30, 57, 37, 57, 225, 30, 77, 57, 101, 57, 103, 57, 142, 57, 110, 57, 150, 57, 163, 11, 170, 11, 233, 11, 233, 11, 238, 11, 254, 11, 254, 11, 6, 12, 30, 12, 62, 12, 78, 12, 86, 12, 102, 12, 126, 12, 142, 12, 166, 12, 182, 12, 198, 12, 214, 12, 238, 12, 245, 12, 29, 13, 59, 13, 67, 13, 83, 13, 90, 13, 133, 13, 140, 13, 180, 13, 185, 13, 209, 13, 214, 13, 254, 13, 54, 14, 62, 14, 78, 14, 85, 14, 101, 14, 117, 14, 133, 14, 149, 14, 165, 14, 181, 14, 197, 14, 213, 14, 237, 14, 245, 14, 21, 15, 37, 15, 61, 15, 68, 15, 92, 15, 108, 15, 148, 15, 177, 15, 201, 15, 209, 15, 232, 15, 8, 16, 32, 16, 40, 16, 98, 16, 98, 16, 104, 16, 160, 24, 164, 24, 212, 24, 212, 24, 216, 24, 232, 24, 237, 24, 21, 25, 28, 25, 35, 25, 51, 25, 91, 25, 107, 25, 114, 25, 122, 25, 138, 25, 194, 11, 154, 25, 186, 25, 202, 25, 226, 25, 249, 25, 39, 26, 46, 26, 84, 26, 89, 26, 120, 26, 120, 26, 128, 26, 180, 26, 180, 26, 186, 26, 227, 26, 228, 26, 236, 26, 20, 27, 20, 27, 24, 27, 71, 27, 71, 27, 74, 27, 113, 27, 113, 27, 114, 27, 138, 27, 138, 27, 122, 27, 154, 27, 178, 27, 186, 27, 210, 27, 218, 27, 226, 27, 250, 27, 221, 14, 10, 28, 34, 28, 50, 28, 66, 28, 74, 28, 114, 28, 98, 28, 138, 28, 146, 28, 0, 69, 6, 69, 22, 69, 25, 69, 49, 69, 73, 69, 77, 69, 101, 69, 109, 69, 149, 69, 155, 69, 163, 69, 171, 69, 187, 69, 195, 69, 211, 69, 227, 69, 243, 69, 249, 69, 1, 70, 25, 70, 41, 70, 57, 70, 73, 70, 89, 70, 105, 70, 112, 70, 128, 70, 144, 70, 168, 70, 184, 70, 216, 70, 224, 70, 210, 25, 16, 71, 40, 71, 56, 71, 80, 71, 96, 71, 112, 71, 128, 71, 136, 71, 168, 71, 184, 71, 192, 71, 195, 71, 202, 11, 235, 71, 242, 71, 192, 70, 18, 72, 26, 72, 50, 72, 55, 72, 79, 72, 95, 72, 111, 72, 135, 72, 151, 72, 167, 72, 183, 72, 190, 72, 197, 72, 229, 72, 184, 64, 191, 64, 215, 64, 231, 64, 247, 64, 255, 64, 23, 65, 31, 65, 55, 65, 63, 65, 79, 65, 95, 65, 119, 65, 151, 65, 159, 65, 191, 65, 196, 65, 212, 65, 235, 65, 243, 65, 19, 66, 51, 66, 55, 66, 71, 66, 87, 66, 95, 66, 111, 66, 143, 66, 150, 66, 118, 12, 4, 67, 222, 13, 28, 67, 44, 67, 60, 67, 84, 67, 92, 67, 100, 67, 116, 67, 123, 67, 139, 67, 147, 67, 163, 67, 179, 67, 195, 67, 219, 67, 92, 55, 251, 67, 11, 68, 18, 68, 26, 68, 74, 68, 82, 68, 90, 68, 114, 68, 119, 68, 124, 68, 140, 68, 146, 68, 162, 68, 168, 68, 184, 68, 200, 68, 232, 68, 121, 37, 176, 20, 145, 37, 153, 37, 193, 37, 201, 37, 217, 37, 223, 37, 255, 37, 15, 38, 21, 38, 37, 38, 53, 38, 77, 38, 117, 38, 122, 38, 146, 38, 153, 38, 185, 38, 193, 38, 238, 38, 242, 38, 18, 39, 22, 39, 29, 39, 92, 39, 37, 39, 116, 39, 124, 39, 148, 39, 156, 39, 124, 15, 196, 39, 203, 39, 219, 39, 251, 39, 3, 40, 11, 40, 27, 40, 35, 40, 190, 12, 59, 40, 67, 40, 75, 40, 107, 40, 123, 40, 139, 40, 147, 40, 155, 40, 179, 40, 185, 40, 209, 40, 28, 34, 233, 40, 170, 34, 241, 40, 1, 41, 9, 41, 33, 41, 38, 41, 70, 41, 78, 41, 94, 41, 194, 25, 162, 28, 178, 28, 58, 28, 218, 28, 226, 28, 250, 28, 26, 29, 50, 29, 66, 29, 90, 29, 114, 29, 130, 29, 162, 29, 178, 29, 226, 29, 234, 29, 2, 30, 9, 30, 33, 30, 34, 22, 60, 18, 97, 30, 65, 30, 113, 30, 137, 30, 145, 30, 169, 30, 185, 30, 217, 30, 1, 31, 33, 31, 62, 31, 68, 31, 104, 31, 142, 31, 98, 13, 182, 31, 182, 31, 190, 31, 219, 31, 219, 31, 223, 31, 21, 32, 28, 32, 30, 32, 36, 32, 97, 32, 97, 32, 100, 32, 148, 32, 152, 32, 157, 32, 197, 32, 197, 32, 197, 32, 201, 32, 241, 32, 249, 32, 1, 33, 25, 33, 30, 33, 62, 33, 69, 33, 93, 33, 158, 57, 159, 57, 245, 24, 167, 57, 207, 57, 214, 57, 222, 57, 246, 57, 254, 57, 6, 58, 6, 14, 22, 58, 57, 49, 227, 39, 46, 58, 62, 58, 78, 58, 102, 58, 118, 58, 134, 58, 142, 58, 221, 45, 174, 58, 182, 58, 198, 58, 254, 58, 26, 59, 31, 59, 60, 59, 63, 59, 84, 59, 88, 59, 93, 59, 99, 59, 155, 59, 155, 59, 155, 59, 155, 59, 163, 59, 207, 59, 207, 59, 211, 36, 215, 59, 223, 59, 14, 60, 14, 60, 14, 60, 15, 60, 64, 60, 72, 60, 80, 60, 80, 60, 87, 60, 127, 60, 43, 25, 135, 60, 143, 60, 151, 60, 194, 60, 156, 60, 202, 60, 208, 60, 8, 61, 216, 60, 235, 72, 3, 73, 9, 73, 25, 73, 41, 73, 57, 73, 73, 73, 80, 73, 96, 73, 112, 73, 144, 73, 184, 73, 126, 58, 215, 73, 234, 73, 241, 73, 38, 74, 43, 74, 65, 74, 65, 74, 72, 74, 112, 74, 117, 74, 137, 74, 142, 74, 180, 74, 183, 74, 215, 74, 219, 74, 235, 74, 241, 74, 1, 75, 8, 75, 16, 75, 54, 75, 189, 75, 197, 75, 213, 75, 220, 75, 236, 75, 252, 75, 12, 76, 18, 76, 42, 76, 58, 76, 65, 76, 89, 76, 97, 76, 105, 76, 153, 76, 169, 76, 60, 17, 209, 76, 233, 76, 241, 76, 246, 76, 14, 77, 20, 77, 36, 77, 52, 77, 68, 77, 84, 77, 90, 77, 98, 77, 162, 25, 62, 26, 152, 26, 252, 26, 40, 27, 130, 27, 202, 27, 18, 28, 122, 28, 242, 28, 74, 29, 138, 29, 194, 29, 25, 30, 41, 30, 81, 30, 129, 30, 177, 30, 241, 30, 84, 31, 150, 31, 247, 31, 60, 32, 173, 32, 217, 32, 77, 33, 140, 33, 4, 34, 44, 34, 122, 34, 186, 34, 46, 35, 184, 35, 27, 36, 97, 36, 160, 36, 27, 37, 97, 37, 161, 37, 239, 37, 69, 38, 161, 38, 10, 39, 45, 39, 164, 39, 243, 39, 91, 40, 225, 40, 54, 41, 142, 41, 21, 42, 112, 42, 246, 42, 86, 43, 189, 43, 235, 43, 19, 44, 207, 44, 22, 45, 85, 45, 173, 45, 21, 46, 122, 46, 215, 46, 247, 80, 186, 28, 39, 81, 71, 81, 87, 81, 111, 81, 115, 81, 139, 81, 163, 81, 147, 81, 187, 81, 195, 81, 211, 81, 219, 81, 227, 81, 3, 82, 43, 82, 67, 82, 75, 82, 91, 82, 107, 82, 115, 82, 123, 82, 131, 82, 155, 82, 160, 82, 184, 82, 208, 82, 224, 82, 230, 82, 238, 82, 254, 82, 6, 83, 14, 83, 22, 83, 29, 83, 37, 83, 53, 83, 69, 83, 77, 83, 101, 83, 108, 83, 124, 83, 148, 83, 156, 83, 140, 83, 196, 83, 217, 83, 241, 83, 242, 83, 10, 84, 18, 84, 34, 84, 39, 84, 79, 84, 87, 84, 95, 84, 119, 84, 135, 84, 142, 84, 188, 83, 238, 84, 254, 84, 30, 85, 106, 77, 130, 77, 137, 77, 144, 77, 152, 77, 168, 77, 184, 77, 189, 77, 221, 77, 250, 71, 245, 77, 253, 77, 4, 78, 20, 78, 28, 78, 44, 78, 68, 78, 92, 78, 100, 78, 108, 78, 124, 78, 130, 78, 154, 78, 162, 78, 170, 78, 186, 78, 194, 78, 218, 78, 226, 78, 242, 78, 249, 78, 9, 79, 17, 79, 33, 79, 41, 79, 65, 79, 81, 79, 105, 79, 109, 79, 133, 79, 141, 79, 157, 79, 173, 79, 181, 79, 197, 79, 210, 28, 245, 79, 251, 79, 11, 80, 27, 80, 43, 80, 51, 80, 91, 80, 97, 80, 103, 80, 127, 80, 143, 80, 151, 80, 167, 80, 183, 80, 191, 80, 199, 80, 215, 80, 231, 80, 34, 29, 98, 29, 17, 30, 153, 30, 209, 30, 76, 31, 239, 31, 52, 32, 108, 32, 209, 32, 38, 33, 188, 33, 20, 34, 138, 34, 234, 34, 225, 35, 89, 36, 219, 36, 89, 37, 231, 37, 85, 38, 2, 39, 132, 39, 43, 40, 99, 40, 217, 40, 166, 41, 59, 42, 174, 42, 30, 43, 173, 43, 59, 44, 6, 45, 181, 45, 229, 45, 172, 46, 107, 47, 59, 48, 217, 48, 73, 49, 254, 49, 210, 50, 39, 51, 155, 51, 77, 52, 180, 52, 34, 53, 198, 53, 78, 54, 180, 54, 228, 54, 60, 55, 170, 55, 128, 56, 53, 57, 238, 57, 38, 58, 206, 58, 179, 59, 31, 60, 95, 60, 224, 60, 73, 61, 140, 61, 184, 49, 192, 49, 208, 49, 224, 49, 230, 49, 246, 49, 14, 50, 22, 50, 148, 33, 38, 50, 54, 50, 70, 50, 76, 50, 100, 50, 106, 50, 114, 50, 122, 50, 138, 50, 162, 50, 178, 50, 202, 50, 218, 50, 121, 30, 234, 50, 239, 50, 15, 51, 31, 51, 79, 51, 85, 51, 109, 51, 113, 51, 145, 51, 147, 51, 175, 51, 175, 51, 178, 51, 194, 51, 194, 51, 202, 51, 239, 51, 207, 51, 247, 51, 15, 52, 22, 52, 29, 52, 45, 52, 69, 52, 85, 52, 92, 52, 108, 52, 124, 52, 148, 52, 172, 52, 196, 52, 204, 52, 220, 52, 252, 52, 12, 53, 18, 53, 26, 53, 50, 53, 80, 53, 88, 53, 104, 53, 51, 40, 163, 40, 46, 41, 150, 41, 83, 42, 54, 43, 197, 43, 27, 44, 215, 44, 109, 45, 205, 45, 75, 46, 199, 46, 68, 47, 190, 47, 35, 48, 201, 48, 113, 49, 30, 50, 170, 50, 129, 51, 53, 52, 212, 52, 120, 53, 156, 54, 204, 54, 244, 54, 202, 55, 192, 56, 14, 57, 118, 57, 199, 57, 54, 58, 214, 58, 123, 59, 239, 59, 240, 60, 196, 61, 148, 62, 51, 63, 17, 64, 15, 65, 27, 66, 190, 66, 236, 66, 36, 67, 203, 67, 42, 68, 176, 68, 93, 69, 9, 70, 160, 70, 240, 70, 0, 71, 152, 71, 34, 72, 221, 72, 152, 73, 88, 74, 62, 75, 110, 75, 149, 75, 173, 75, 50, 76, 46, 47, 182, 47, 27, 48, 147, 48, 225, 48, 25, 49, 105, 49, 216, 49, 84, 50, 146, 50, 247, 50, 121, 51, 215, 51, 37, 52, 100, 52, 188, 52, 42, 53, 206, 53, 51, 54, 148, 54, 36, 55, 116, 55, 226, 55, 34, 56, 168, 56, 85, 57, 191, 57, 14, 58, 70, 58, 6, 59, 115, 59, 187, 59, 231, 59, 39, 60, 172, 60, 232, 60, 32, 61, 81, 61, 118, 61, 172, 61, 52, 62, 203, 62, 21, 63, 138, 63, 242, 63, 73, 64, 176, 64, 7, 65, 167, 65, 3, 66, 119, 66, 68, 67, 187, 67, 3, 68, 192, 68, 57, 69, 117, 69, 203, 69, 65, 70, 152, 70, 24, 71, 144, 71, 219, 71, 10, 72, 16, 61, 56, 61, 56, 61, 58, 61, 97, 61, 97, 61, 65, 61, 105, 61, 110, 61, 126, 61, 132, 61, 148, 61, 41, 49, 164, 61, 188, 61, 212, 61, 220, 61, 228, 61, 4, 62, 28, 62, 36, 62, 68, 62, 74, 62, 103, 62, 111, 62, 135, 62, 140, 62, 180, 62, 180, 62, 188, 62, 195, 62, 223, 62, 255, 62, 108, 55, 7, 63, 8, 63, 13, 63, 29, 63, 35, 63, 67, 63, 83, 63, 91, 63, 99, 63, 106, 63, 114, 63, 122, 63, 146, 63, 154, 63, 178, 63, 202, 63, 218, 63, 234, 63, 2, 64, 9, 64, 41, 64, 65, 64, 89, 64, 97, 64, 113, 64, 129, 64, 145, 64, 152, 64, 160, 64, 168, 64, 143, 72, 243, 72, 65, 73, 128, 73, 9, 74, 80, 74, 191, 74, 249, 74, 205, 75, 73, 76, 193, 76, 28, 77, 114, 77, 229, 77, 36, 78, 178, 78, 73, 79, 189, 79, 159, 80, 31, 81, 123, 81, 251, 81, 168, 82, 93, 83, 26, 84, 190, 84, 118, 85, 242, 85, 192, 86, 54, 87, 189, 87, 19, 88, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 129, 76, 217, 76, 60, 77, 237, 77, 234, 78, 205, 79, 3, 80, 67, 80, 23, 81, 155, 81, 11, 82, 51, 82, 192, 82, 116, 83, 103, 84, 166, 84, 198, 84, 42, 85, 211, 85, 26, 86, 42, 86, 74, 86, 145, 86, 208, 86, 94, 87, 227, 87, 11, 88, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 242, 108, 249, 108, 1, 109, 48, 0, 9, 109, 9, 109, 9, 109, 11, 109, 224, 93, 157, 97, 137, 99, 165, 97, 172, 97, 180, 97, 184, 97, 120, 92, 48, 0, 48, 0, 48, 0, 48, 0, 138, 108, 146, 108, 150, 108, 185, 88, 158, 108, 121, 99, 177, 102, 185, 102, 232, 93, 105, 109, 240, 93, 48, 0, 174, 108, 166, 108, 182, 108, 186, 108, 208, 108, 123, 100, 72, 106, 78, 106, 216, 108, 222, 108, 226, 108, 215, 94, 48, 0, 48, 0, 48, 0, 48, 0, 19, 109, 19, 109, 19, 109, 19, 109, 19, 109, 19, 109, 19, 109, 19, 109, 19, 109, 22, 109, 30, 109, 38, 109, 48, 0, 48, 0, 48, 0, 48, 0, 88, 91, 27, 105, 120, 93, 209, 99, 96, 91, 193, 88, 201, 88, 131, 100, 104, 91, 139, 100, 56, 93, 152, 90, 112, 91, 145, 104, 160, 90, 217, 99, 153, 104, 24, 90, 48, 0, 48, 0, 161, 104, 200, 93, 224, 91, 168, 90, 120, 91, 176, 90, 169, 104, 209, 88, 48, 0, 48, 0, 48, 0, 48, 0, 177, 104, 180, 104, 184, 90, 188, 104, 196, 104, 128, 91, 192, 90, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 204, 104, 136, 91, 200, 90, 48, 0, 212, 104, 232, 91, 144, 91, 167, 94, 220, 104, 208, 90, 147, 100, 48, 0, 252, 104, 128, 93, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 165, 109, 217, 88, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 249, 106, 251, 106, 234, 108, 144, 93, 48, 0, 48, 0, 2, 107, 56, 90, 106, 108, 88, 92, 247, 94, 233, 99, 114, 108, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 26, 107, 18, 107, 248, 92, 48, 0, 48, 0, 48, 0, 34, 108, 50, 108, 42, 108, 66, 108, 58, 108, 48, 0, 48, 0, 225, 88, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 194, 108, 194, 108, 194, 108, 200, 108, 176, 89, 233, 88, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 208, 97, 208, 97, 208, 97, 208, 97, 208, 97, 208, 97, 208, 97, 224, 97, 232, 97, 239, 97, 246, 97, 241, 88, 249, 88, 255, 88, 254, 97, 6, 98, 14, 98, 22, 98, 7, 89, 48, 0, 48, 0, 48, 0, 48, 0, 241, 99, 114, 97, 114, 97, 130, 97, 51, 100, 48, 0, 48, 0, 48, 0, 48, 0, 243, 100, 43, 101, 246, 100, 46, 101, 254, 100, 6, 101, 65, 101, 14, 101, 22, 101, 17, 101, 30, 101, 45, 101, 15, 101, 243, 100, 43, 101, 38, 101, 46, 101, 41, 101, 49, 101, 44, 101, 14, 101, 249, 101, 255, 95, 39, 96, 7, 96, 47, 96, 15, 96, 55, 96, 23, 96, 63, 96, 31, 96, 155, 100, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 42, 99, 15, 89, 18, 89, 15, 89, 25, 89, 32, 89, 40, 89, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 10, 107, 10, 107, 10, 107, 10, 107, 10, 107, 10, 107, 48, 89, 48, 0, 42, 107, 34, 107, 64, 90, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 246, 99, 254, 99, 30, 98, 48, 0, 48, 0, 6, 100, 38, 98, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 33, 103, 57, 103, 113, 103, 49, 103, 41, 103, 65, 103, 48, 0, 87, 95, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 56, 89, 15, 89, 15, 89, 15, 89, 48, 0, 48, 0, 48, 0, 48, 0, 15, 89, 15, 89, 15, 89, 15, 89, 15, 89, 15, 89, 15, 89, 64, 89, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 109, 49, 109, 184, 93, 192, 97, 65, 109, 145, 99, 48, 0, 153, 99, 248, 93, 216, 93, 48, 0, 48, 0, 173, 109, 27, 100, 14, 100, 19, 100, 192, 89, 32, 92, 112, 93, 255, 94, 161, 99, 48, 0, 81, 109, 152, 89, 57, 109, 192, 93, 169, 99, 177, 99, 223, 94, 185, 99, 48, 0, 48, 0, 163, 106, 171, 106, 179, 106, 48, 0, 147, 106, 155, 106, 139, 106, 129, 99, 179, 105, 160, 89, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 113, 99, 89, 109, 32, 90, 48, 0, 48, 0, 187, 100, 193, 99, 168, 89, 48, 0, 48, 0, 97, 109, 201, 99, 73, 109, 244, 104, 216, 90, 48, 0, 48, 0, 48, 0, 82, 108, 74, 108, 225, 99, 228, 104, 152, 91, 224, 90, 48, 0, 48, 0, 236, 104, 240, 91, 231, 94, 83, 105, 232, 90, 191, 94, 91, 105, 248, 91, 199, 94, 90, 108, 98, 108, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 35, 105, 160, 91, 136, 93, 207, 94, 99, 105, 168, 91, 48, 0, 48, 0, 4, 105, 11, 105, 240, 90, 19, 105, 0, 92, 147, 100, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 208, 93, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 26, 108, 131, 95, 239, 94, 145, 102, 169, 102, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 122, 108, 96, 92, 200, 97, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 130, 108, 104, 92, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 113, 109, 121, 109, 132, 109, 132, 109, 124, 109, 140, 109, 140, 109, 140, 109, 140, 109, 140, 109, 142, 109, 48, 0, 48, 0, 48, 0, 195, 100, 35, 100, 43, 100, 0, 94, 150, 109, 150, 109, 150, 109, 150, 109, 150, 109, 150, 109, 157, 109, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 98, 51, 98, 55, 98, 55, 98, 71, 98, 79, 98, 87, 98, 92, 98, 32, 94, 48, 94, 73, 101, 119, 98, 227, 100, 195, 95, 48, 0, 4, 97, 194, 107, 2, 110, 143, 94, 127, 98, 48, 0, 48, 0, 48, 0, 48, 0, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 135, 98, 214, 98, 214, 98, 214, 98, 214, 98, 214, 98, 214, 98, 216, 98, 224, 98, 232, 98, 232, 98, 232, 98, 235, 98, 243, 98, 243, 98, 245, 98, 253, 98, 5, 99, 9, 99, 15, 99, 23, 99, 29, 99, 34, 99, 48, 0, 48, 0, 166, 98, 166, 98, 166, 98, 168, 98, 175, 98, 175, 98, 180, 98, 184, 98, 100, 98, 100, 98, 103, 98, 111, 98, 192, 98, 198, 98, 206, 98, 48, 0, 192, 96, 192, 96, 192, 96, 192, 96, 208, 96, 213, 96, 219, 96, 163, 100, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 186, 107, 249, 107, 249, 107, 249, 107, 249, 107, 249, 107, 249, 107, 249, 107, 250, 107, 48, 0, 234, 107, 241, 107, 222, 109, 222, 109, 222, 109, 222, 109, 222, 109, 222, 109, 222, 109, 222, 109, 222, 109, 222, 109, 222, 109, 223, 109, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 181, 109, 188, 109, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 196, 109, 198, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 206, 109, 231, 109, 231, 109, 231, 109, 231, 109, 231, 109, 231, 109, 231, 109, 231, 109, 231, 109, 231, 109, 231, 109, 231, 109, 231, 109, 231, 109, 234, 109, 48, 0, 214, 109, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 50, 110, 122, 110, 106, 110, 146, 110, 162, 110, 178, 110, 98, 110, 202, 110, 210, 110, 218, 110, 242, 110, 250, 110, 130, 110, 18, 111, 26, 111, 170, 110, 50, 111, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 255, 255, 255, 255, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 14, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 5, 0, 4, 5, 5, 92, 7, 5, 5, 136, 13, 5, 5, 140, 13, 5, 5, 134, 13, 5, 5, 142, 13, 5, 5, 139, 12, 5, 5, 114, 10, 5, 5, 16, 12, 5, 5, 91, 12, 43, 5, 0, 42, 5, 5, 138, 9, 5, 5, 135, 12, 0, 0, 0, 0, 5, 5, 95, 12, 5, 5, 12, 12, 5, 5, 83, 12, 5, 5, 123, 12, 36, 5, 0, 23, 36, 5, 0, 25, 5, 5, 6, 12, 16, 5, 42, 96, 5, 5, 118, 10, 200, 0, 0, 0, 5, 5, 20, 12, 36, 5, 0, 21, 43, 5, 0, 70, 5, 5, 140, 9, 197, 35, 69, 0, 197, 99, 68, 0, 197, 227, 83, 0, 5, 5, 100, 7, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 108, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 192, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 108, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 196, 164, 156, 42, 196, 164, 5, 42, 196, 140, 156, 42, 196, 140, 5, 42, 196, 162, 156, 42, 196, 162, 5, 42, 196, 136, 156, 46, 196, 136, 5, 46, 196, 142, 156, 46, 196, 142, 5, 46, 196, 156, 156, 46, 196, 156, 5, 46, 196, 144, 156, 46, 196, 144, 5, 46, 196, 144, 156, 48, 196, 144, 5, 48, 196, 178, 156, 48, 196, 178, 5, 48, 196, 164, 156, 50, 196, 164, 5, 50, 196, 140, 156, 50, 196, 140, 5, 50, 196, 156, 156, 50, 196, 156, 5, 50, 196, 162, 156, 50, 196, 162, 5, 50, 196, 144, 156, 50, 196, 144, 5, 50, 196, 142, 156, 54, 196, 142, 5, 54, 196, 140, 156, 54, 196, 140, 5, 54, 196, 156, 156, 54, 196, 156, 5, 54, 196, 160, 156, 54, 196, 160, 5, 54, 196, 142, 156, 56, 196, 142, 5, 56, 196, 178, 156, 56, 196, 178, 5, 56, 196, 154, 156, 58, 196, 154, 5, 58, 196, 164, 156, 58, 196, 164, 5, 58, 196, 140, 156, 58, 196, 140, 5, 58, 196, 162, 156, 58, 196, 162, 5, 58, 196, 156, 156, 58, 5, 5, 6, 59, 197, 66, 141, 0, 197, 2, 141, 0, 196, 142, 156, 60, 196, 142, 5, 60, 196, 160, 156, 62, 196, 160, 5, 62, 5, 5, 16, 75, 196, 136, 156, 64, 196, 136, 5, 64, 196, 160, 156, 64, 196, 160, 5, 64, 196, 144, 156, 64, 196, 144, 5, 64, 197, 2, 149, 0, 197, 194, 148, 0, 196, 178, 156, 64, 196, 178, 5, 64, 196, 136, 156, 68, 196, 136, 5, 68, 196, 160, 156, 68, 196, 160, 5, 68, 196, 144, 156, 68, 196, 144, 5, 68, 197, 162, 212, 0, 158, 5, 30, 69, 5, 5, 30, 69, 196, 164, 156, 70, 196, 164, 5, 70, 196, 140, 156, 70, 196, 140, 5, 70, 196, 152, 156, 70, 196, 152, 5, 70, 197, 2, 179, 0, 197, 194, 178, 0, 196, 136, 156, 76, 196, 136, 5, 76, 196, 160, 156, 76, 196, 160, 5, 76, 196, 144, 156, 76, 196, 144, 5, 76, 196, 136, 156, 78, 196, 136, 5, 78, 196, 142, 156, 78, 196, 142, 5, 78, 196, 160, 156, 78, 196, 160, 5, 78, 196, 144, 156, 78, 196, 144, 5, 78, 196, 160, 156, 80, 196, 160, 5, 80, 196, 144, 156, 80, 196, 144, 5, 80, 160, 5, 8, 81, 5, 5, 8, 81, 196, 154, 156, 82, 196, 154, 5, 82, 196, 164, 156, 82, 196, 164, 5, 82, 196, 140, 156, 82, 196, 140, 5, 82, 196, 146, 156, 82, 196, 146, 5, 82, 196, 152, 156, 82, 196, 152, 5, 82, 196, 162, 156, 82, 196, 162, 5, 82, 196, 142, 160, 86, 196, 142, 5, 86, 196, 142, 156, 90, 196, 142, 5, 90, 196, 150, 156, 90, 196, 136, 156, 92, 196, 136, 5, 92, 196, 156, 156, 92, 196, 156, 5, 92, 196, 144, 156, 92, 196, 144, 5, 92, 5, 120, 0, 78, 5, 5, 8, 45, 160, 5, 20, 45, 160, 5, 22, 45, 5, 5, 22, 45, 160, 5, 74, 93, 5, 5, 74, 93, 160, 5, 30, 71, 160, 5, 14, 47, 5, 5, 14, 47, 160, 5, 16, 49, 160, 5, 18, 49, 160, 5, 22, 49, 5, 5, 22, 49, 197, 226, 211, 0, 160, 5, 20, 51, 155, 5, 24, 51, 144, 5, 28, 51, 160, 5, 16, 53, 5, 5, 16, 53, 160, 5, 16, 55, 160, 5, 24, 55, 5, 5, 8, 57, 160, 5, 24, 59, 160, 5, 16, 59, 160, 5, 10, 63, 5, 5, 10, 63, 5, 5, 14, 65, 5, 5, 42, 65, 160, 5, 40, 83, 160, 5, 14, 69, 5, 5, 16, 69, 160, 5, 50, 71, 196, 190, 156, 70, 196, 190, 5, 70, 160, 5, 26, 55, 5, 5, 26, 55, 160, 5, 16, 73, 5, 5, 16, 73, 174, 5, 8, 77, 160, 5, 70, 93, 5, 5, 70, 93, 158, 5, 22, 79, 5, 5, 28, 79, 5, 5, 14, 81, 160, 5, 16, 81, 5, 5, 16, 81, 160, 5, 18, 81, 196, 190, 156, 82, 196, 190, 5, 82, 160, 5, 48, 83, 160, 5, 12, 85, 160, 5, 10, 91, 5, 5, 10, 91, 160, 5, 8, 93, 5, 5, 8, 93, 158, 5, 26, 93, 160, 5, 30, 93, 5, 5, 30, 93, 5, 5, 34, 93, 5, 5, 62, 93, 160, 5, 72, 93, 5, 5, 72, 93, 197, 34, 195, 0, 5, 5, 44, 93, 5, 5, 112, 93, 5, 5, 114, 93, 5, 5, 116, 93, 5, 5, 118, 93, 197, 67, 123, 0, 197, 227, 122, 0, 197, 131, 122, 0, 197, 194, 149, 0, 197, 130, 149, 0, 197, 66, 149, 0, 197, 66, 162, 0, 197, 2, 162, 0, 197, 194, 161, 0, 196, 144, 156, 42, 196, 144, 5, 42, 196, 144, 156, 58, 196, 144, 5, 58, 196, 144, 156, 70, 196, 144, 5, 70, 196, 144, 156, 82, 196, 144, 5, 82, 197, 131, 199, 0, 197, 35, 199, 0, 197, 67, 197, 0, 197, 227, 196, 0, 197, 195, 198, 0, 197, 99, 198, 0, 197, 3, 198, 0, 197, 163, 197, 0, 5, 5, 20, 51, 197, 195, 102, 0, 197, 99, 102, 0, 197, 131, 103, 0, 197, 35, 103, 0, 197, 3, 110, 0, 197, 163, 109, 0, 160, 5, 12, 55, 5, 5, 12, 55, 196, 144, 156, 54, 196, 144, 5, 54, 196, 144, 156, 62, 196, 144, 5, 62, 196, 162, 156, 70, 196, 162, 5, 70, 197, 35, 172, 0, 197, 195, 171, 0, 197, 98, 212, 0, 197, 34, 212, 0, 196, 144, 5, 60, 197, 66, 122, 0, 197, 2, 122, 0, 197, 194, 121, 0, 196, 136, 156, 54, 196, 136, 5, 54, 160, 5, 8, 57, 160, 5, 44, 93, 196, 138, 156, 68, 196, 138, 5, 68, 197, 131, 101, 0, 197, 35, 101, 0, 197, 67, 109, 0, 197, 227, 108, 0, 197, 99, 171, 0, 197, 3, 171, 0, 196, 184, 156, 42, 196, 184, 5, 42, 196, 188, 156, 42, 196, 188, 5, 42, 196, 184, 156, 50, 196, 184, 5, 50, 196, 188, 156, 50, 196, 188, 5, 50, 196, 184, 156, 58, 196, 184, 5, 58, 196, 188, 156, 58, 196, 188, 5, 58, 196, 184, 156, 70, 196, 184, 5, 70, 196, 188, 156, 70, 196, 188, 5, 70, 196, 184, 156, 76, 196, 184, 5, 76, 196, 188, 156, 76, 196, 188, 5, 76, 196, 184, 156, 82, 196, 184, 5, 82, 196, 188, 156, 82, 196, 188, 5, 82, 196, 202, 156, 78, 196, 202, 5, 78, 196, 202, 156, 80, 196, 202, 5, 80, 160, 5, 16, 91, 5, 5, 16, 91, 196, 144, 156, 56, 196, 144, 5, 56, 160, 5, 16, 69, 5, 5, 24, 49, 160, 5, 58, 71, 5, 5, 58, 71, 160, 5, 14, 93, 5, 5, 14, 93, 196, 156, 156, 42, 196, 156, 5, 42, 196, 160, 156, 50, 196, 160, 5, 50, 197, 163, 167, 0, 197, 67, 167, 0, 197, 227, 169, 0, 197, 131, 169, 0, 196, 156, 156, 70, 196, 156, 5, 70, 197, 163, 170, 0, 197, 67, 170, 0, 196, 164, 156, 90, 196, 164, 5, 90, 5, 5, 34, 65, 5, 5, 24, 69, 5, 5, 20, 81, 5, 5, 6, 61, 197, 194, 119, 0, 197, 98, 184, 0, 160, 5, 8, 43, 160, 5, 8, 47, 5, 5, 8, 47, 160, 5, 14, 65, 160, 5, 10, 81, 5, 5, 16, 79, 5, 5, 20, 93, 160, 5, 78, 93, 5, 5, 78, 93, 160, 5, 8, 45, 160, 5, 20, 83, 160, 5, 20, 85, 160, 5, 12, 51, 5, 5, 12, 51, 160, 5, 10, 61, 5, 5, 10, 61, 160, 5, 14, 75, 5, 5, 14, 75, 160, 5, 16, 77, 5, 5, 16, 77, 160, 5, 8, 91, 5, 5, 8, 91, 5, 5, 20, 43, 5, 5, 22, 43, 5, 5, 28, 43, 5, 5, 20, 45, 5, 5, 30, 71, 5, 5, 16, 47, 5, 5, 16, 49, 5, 5, 18, 49, 5, 5, 32, 51, 5, 5, 24, 51, 5, 5, 34, 51, 5, 5, 28, 51, 5, 5, 36, 51, 5, 5, 42, 51, 5, 5, 44, 51, 5, 5, 14, 61, 5, 5, 16, 55, 5, 5, 6, 55, 5, 5, 10, 55, 5, 5, 24, 55, 5, 5, 48, 51, 5, 5, 34, 83, 5, 5, 12, 57, 5, 5, 22, 57, 5, 5, 16, 59, 5, 5, 24, 59, 5, 5, 8, 59, 5, 5, 18, 65, 5, 5, 24, 65, 5, 5, 30, 65, 5, 5, 38, 65, 5, 5, 40, 83, 5, 5, 46, 83, 5, 5, 12, 67, 5, 5, 14, 69, 5, 5, 22, 69, 5, 5, 6, 69, 5, 5, 50, 71, 5, 5, 12, 71, 5, 5, 54, 71, 5, 5, 24, 73, 5, 5, 20, 77, 5, 5, 24, 77, 5, 5, 28, 77, 5, 5, 32, 77, 5, 5, 34, 77, 5, 5, 38, 77, 5, 5, 42, 77, 5, 5, 8, 77, 5, 5, 54, 77, 5, 5, 14, 79, 5, 5, 22, 79, 5, 5, 16, 61, 5, 5, 30, 79, 5, 5, 34, 79, 5, 5, 24, 81, 5, 5, 18, 81, 5, 5, 20, 83, 5, 5, 48, 83, 5, 5, 12, 85, 5, 5, 20, 85, 5, 5, 12, 87, 5, 5, 44, 65, 5, 5, 6, 91, 5, 5, 16, 93, 5, 5, 18, 93, 5, 5, 26, 93, 5, 5, 36, 93, 5, 5, 76, 93, 5, 5, 94, 93, 5, 5, 110, 93, 5, 5, 120, 93, 5, 5, 122, 93, 5, 5, 6, 45, 5, 5, 46, 51, 5, 5, 18, 55, 5, 5, 6, 57, 5, 5, 12, 61, 5, 5, 20, 63, 5, 5, 6, 65, 5, 5, 12, 75, 5, 5, 106, 93, 5, 5, 108, 93, 197, 194, 121, 0, 197, 34, 124, 0, 197, 226, 123, 0, 197, 34, 195, 0, 197, 162, 195, 0, 197, 162, 193, 0, 197, 162, 132, 0, 197, 98, 152, 0, 197, 66, 153, 0, 5, 5, 124, 93, 5, 5, 126, 93, 5, 5, 36, 83, 5, 5, 38, 83, 43, 5, 0, 56, 46, 5, 12, 57, 43, 5, 0, 60, 43, 5, 0, 76, 32, 5, 20, 77, 32, 5, 28, 77, 32, 5, 54, 77, 52, 5, 0, 86, 43, 5, 0, 90, 5, 5, 24, 12, 5, 5, 28, 12, 5, 5, 24, 57, 5, 5, 82, 93, 5, 5, 26, 57, 5, 5, 86, 93, 5, 5, 96, 93, 5, 5, 80, 93, 5, 5, 98, 93, 193, 2, 30, 12, 193, 9, 30, 12, 193, 16, 30, 12, 193, 23, 30, 12, 5, 5, 32, 12, 5, 5, 34, 12, 5, 5, 36, 12, 5, 5, 38, 12, 5, 5, 40, 12, 5, 5, 42, 12, 5, 5, 44, 12, 5, 5, 46, 12, 5, 5, 48, 12, 5, 5, 50, 12, 5, 5, 80, 13, 5, 5, 82, 13, 193, 2, 52, 12, 193, 9, 52, 12, 193, 16, 52, 12, 193, 23, 52, 12, 193, 44, 52, 12, 193, 51, 52, 12, 193, 2, 14, 12, 193, 9, 14, 12, 193, 2, 18, 12, 193, 2, 22, 12, 5, 5, 8, 12, 193, 9, 18, 12, 193, 58, 52, 12, 193, 65, 52, 12, 46, 5, 24, 55, 43, 5, 0, 64, 43, 5, 0, 78, 43, 5, 0, 88, 32, 5, 94, 93, 193, 72, 52, 12, 193, 79, 52, 12, 193, 86, 52, 12, 193, 93, 52, 12, 193, 100, 52, 12, 193, 107, 52, 12, 193, 114, 52, 12, 5, 5, 54, 12, 193, 2, 56, 12, 5, 5, 84, 93, 193, 9, 56, 12, 193, 16, 56, 12, 193, 23, 56, 12, 193, 30, 56, 12, 193, 37, 56, 12, 193, 44, 56, 12, 193, 51, 56, 12, 193, 58, 56, 12, 193, 65, 56, 12, 193, 72, 56, 12, 193, 79, 56, 12, 193, 86, 56, 12, 193, 93, 56, 12, 193, 100, 56, 12, 193, 107, 56, 12, 193, 114, 56, 12, 193, 121, 56, 12, 5, 138, 0, 0, 5, 136, 0, 0, 5, 142, 0, 0, 5, 154, 0, 0, 5, 164, 0, 0, 5, 180, 0, 0, 5, 140, 0, 0, 5, 156, 0, 0, 5, 150, 0, 0, 5, 182, 0, 0, 5, 146, 0, 0, 5, 152, 0, 0, 5, 144, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 184, 0, 0, 5, 186, 0, 0, 5, 188, 0, 0, 5, 166, 0, 0, 5, 132, 0, 0, 5, 134, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 190, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 192, 0, 0, 5, 194, 0, 0, 5, 196, 0, 0, 5, 198, 0, 0, 5, 200, 0, 0, 5, 202, 0, 0, 5, 160, 0, 0, 5, 162, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 204, 0, 0, 5, 206, 0, 0, 5, 168, 0, 0, 5, 208, 0, 0, 5, 210, 0, 0, 5, 130, 0, 0, 5, 168, 0, 0, 5, 212, 0, 0, 5, 178, 0, 0, 5, 170, 0, 0, 5, 170, 0, 0, 5, 158, 0, 0, 5, 214, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 138, 0, 0, 5, 136, 0, 0, 5, 148, 0, 0, 5, 132, 0, 0, 197, 34, 0, 0, 5, 216, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 0, 0, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 218, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 220, 0, 0, 5, 222, 0, 0, 5, 168, 0, 0, 19, 5, 0, 42, 19, 5, 0, 50, 19, 5, 0, 58, 19, 5, 0, 70, 19, 5, 0, 82, 19, 5, 0, 46, 19, 5, 0, 48, 19, 5, 0, 56, 19, 5, 0, 66, 19, 5, 0, 76, 19, 5, 0, 80, 19, 5, 0, 84, 19, 5, 0, 88, 160, 5, 26, 96, 5, 5, 26, 96, 160, 5, 92, 96, 5, 5, 92, 96, 5, 5, 24, 12, 5, 5, 26, 12, 160, 5, 20, 96, 5, 5, 20, 96, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 32, 96, 5, 5, 70, 96, 5, 5, 68, 96, 5, 5, 72, 96, 5, 5, 42, 7, 160, 5, 34, 96, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 6, 12, 197, 98, 49, 0, 197, 34, 226, 0, 200, 0, 0, 0, 197, 2, 234, 0, 197, 2, 248, 0, 197, 98, 0, 1, 255, 255, 255, 255, 197, 2, 11, 1, 255, 255, 255, 255, 197, 226, 16, 1, 197, 66, 34, 1, 197, 99, 2, 1, 166, 5, 6, 96, 158, 5, 8, 96, 152, 5, 10, 96, 155, 5, 14, 96, 166, 5, 16, 96, 158, 5, 24, 96, 158, 5, 28, 96, 158, 5, 30, 96, 166, 5, 32, 96, 160, 5, 36, 96, 158, 5, 38, 96, 160, 5, 42, 96, 158, 5, 44, 96, 158, 5, 46, 96, 158, 5, 48, 96, 166, 5, 50, 96, 160, 5, 60, 96, 255, 255, 255, 255, 158, 5, 66, 96, 158, 5, 74, 96, 155, 5, 76, 96, 158, 5, 78, 96, 152, 5, 80, 96, 158, 5, 82, 96, 155, 5, 86, 96, 197, 34, 2, 1, 197, 226, 18, 1, 197, 226, 225, 0, 197, 194, 233, 0, 197, 194, 247, 0, 197, 34, 0, 1, 197, 99, 19, 1, 5, 5, 6, 96, 5, 5, 8, 96, 5, 5, 10, 96, 5, 5, 14, 96, 5, 5, 16, 96, 5, 5, 24, 96, 5, 5, 28, 96, 5, 5, 30, 96, 5, 5, 32, 96, 5, 5, 36, 96, 5, 5, 38, 96, 5, 5, 42, 96, 5, 5, 44, 96, 5, 5, 46, 96, 5, 5, 48, 96, 5, 5, 50, 96, 5, 5, 60, 96, 51, 5, 66, 96, 5, 5, 66, 96, 5, 5, 74, 96, 5, 5, 76, 96, 5, 5, 78, 96, 5, 5, 80, 96, 5, 5, 82, 96, 5, 5, 86, 96, 197, 226, 1, 1, 197, 162, 18, 1, 197, 194, 10, 1, 197, 162, 16, 1, 197, 2, 34, 1, 197, 99, 4, 1, 16, 5, 8, 96, 16, 5, 30, 96, 166, 5, 76, 96, 197, 34, 17, 1, 197, 34, 19, 1, 16, 5, 78, 96, 16, 5, 50, 96, 197, 3, 4, 1, 160, 5, 58, 96, 5, 5, 58, 96, 160, 5, 22, 96, 5, 5, 22, 96, 158, 5, 18, 96, 5, 5, 18, 96, 160, 5, 56, 96, 5, 5, 56, 96, 160, 5, 90, 96, 5, 5, 90, 96, 160, 5, 160, 96, 5, 5, 160, 96, 160, 5, 170, 96, 5, 5, 170, 96, 160, 5, 172, 96, 5, 5, 172, 96, 160, 5, 178, 96, 5, 5, 178, 96, 160, 5, 192, 96, 5, 5, 192, 96, 160, 5, 198, 96, 5, 5, 198, 96, 160, 5, 206, 96, 5, 5, 206, 96, 16, 5, 36, 96, 16, 5, 60, 96, 16, 5, 66, 96, 5, 5, 34, 96, 165, 5, 30, 96, 16, 5, 16, 96, 193, 30, 117, 12, 160, 5, 94, 96, 5, 5, 94, 96, 165, 5, 66, 96, 160, 5, 54, 96, 5, 5, 54, 96, 5, 5, 64, 96, 160, 5, 70, 96, 160, 5, 68, 96, 160, 5, 72, 96, 197, 130, 39, 1, 197, 130, 40, 1, 160, 5, 32, 97, 197, 2, 39, 1, 174, 5, 42, 97, 160, 5, 58, 97, 174, 5, 74, 97, 197, 66, 44, 1, 160, 5, 80, 97, 160, 5, 106, 97, 160, 5, 134, 97, 160, 5, 170, 97, 197, 194, 44, 1, 197, 130, 42, 1, 197, 130, 46, 1, 160, 5, 216, 97, 174, 5, 6, 97, 174, 5, 12, 97, 174, 5, 14, 97, 174, 5, 16, 97, 174, 5, 26, 97, 174, 5, 40, 97, 174, 5, 44, 97, 174, 5, 50, 97, 201, 102, 1, 0, 160, 5, 78, 97, 174, 5, 84, 97, 174, 5, 98, 97, 174, 5, 114, 97, 158, 5, 120, 97, 158, 5, 138, 97, 174, 5, 142, 97, 174, 5, 150, 97, 174, 5, 156, 97, 174, 5, 162, 97, 174, 5, 172, 97, 174, 5, 180, 97, 174, 5, 182, 97, 174, 5, 192, 97, 174, 5, 200, 97, 174, 5, 218, 97, 174, 5, 222, 97, 158, 5, 226, 97, 174, 5, 230, 97, 158, 5, 232, 97, 160, 5, 238, 97, 174, 5, 240, 97, 160, 5, 244, 97, 5, 5, 6, 97, 5, 5, 12, 97, 5, 5, 14, 97, 5, 5, 16, 97, 5, 5, 26, 97, 5, 5, 40, 97, 5, 5, 44, 97, 5, 5, 50, 97, 201, 70, 2, 0, 5, 5, 78, 97, 5, 5, 84, 97, 5, 5, 98, 97, 5, 5, 114, 97, 5, 5, 120, 97, 5, 5, 138, 97, 5, 5, 142, 97, 5, 5, 150, 97, 5, 5, 156, 97, 5, 5, 162, 97, 5, 5, 172, 97, 5, 5, 180, 97, 5, 5, 182, 97, 5, 5, 192, 97, 5, 5, 200, 97, 5, 5, 218, 97, 5, 5, 222, 97, 5, 5, 226, 97, 5, 5, 230, 97, 5, 5, 232, 97, 5, 5, 238, 97, 5, 5, 240, 97, 5, 5, 244, 97, 197, 66, 39, 1, 197, 66, 40, 1, 5, 5, 32, 97, 197, 194, 38, 1, 5, 5, 42, 97, 5, 5, 58, 97, 5, 5, 74, 97, 197, 194, 43, 1, 5, 5, 80, 97, 5, 5, 106, 97, 5, 5, 134, 97, 5, 5, 170, 97, 197, 130, 44, 1, 197, 66, 42, 1, 197, 66, 46, 1, 5, 5, 216, 97, 198, 225, 155, 0, 193, 16, 190, 97, 198, 65, 157, 0, 193, 2, 236, 97, 198, 33, 158, 0, 193, 9, 246, 97, 198, 97, 158, 0, 193, 16, 246, 97, 198, 1, 159, 0, 193, 44, 246, 97, 198, 193, 158, 0, 193, 30, 246, 97, 198, 97, 159, 0, 193, 58, 246, 97, 198, 129, 159, 0, 193, 65, 246, 97, 198, 161, 159, 0, 193, 72, 246, 97, 198, 225, 159, 0, 193, 79, 246, 97, 198, 1, 160, 0, 193, 86, 246, 97, 198, 34, 160, 0, 197, 194, 49, 1, 198, 33, 155, 0, 193, 9, 178, 97, 198, 97, 156, 0, 193, 44, 190, 97, 198, 65, 156, 0, 193, 37, 190, 97, 198, 1, 156, 0, 193, 23, 190, 97, 160, 5, 148, 97, 5, 5, 148, 97, 193, 2, 85, 12, 5, 224, 0, 0, 5, 166, 0, 0, 5, 134, 0, 0, 5, 132, 0, 0, 5, 166, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 160, 5, 72, 97, 5, 5, 72, 97, 160, 5, 234, 97, 5, 5, 234, 97, 160, 5, 152, 97, 5, 5, 152, 97, 160, 112, 16, 97, 5, 112, 16, 97, 160, 5, 18, 97, 5, 5, 18, 97, 160, 5, 22, 97, 5, 5, 22, 97, 160, 5, 48, 97, 5, 5, 48, 97, 160, 5, 38, 97, 5, 5, 38, 97, 160, 5, 86, 97, 5, 5, 86, 97, 160, 5, 94, 97, 5, 5, 94, 97, 160, 5, 92, 97, 5, 5, 92, 97, 160, 5, 90, 97, 5, 5, 90, 97, 160, 5, 126, 97, 5, 5, 126, 97, 160, 5, 132, 97, 5, 5, 132, 97, 160, 5, 146, 97, 5, 5, 146, 97, 160, 5, 248, 97, 5, 5, 248, 97, 160, 5, 160, 97, 5, 5, 160, 97, 160, 5, 166, 97, 5, 5, 166, 97, 160, 5, 174, 97, 5, 5, 174, 97, 160, 5, 176, 97, 5, 5, 176, 97, 160, 5, 186, 97, 5, 5, 186, 97, 160, 5, 196, 97, 5, 5, 196, 97, 160, 5, 204, 97, 5, 5, 204, 97, 160, 5, 208, 97, 5, 5, 208, 97, 160, 5, 188, 97, 5, 5, 188, 97, 160, 5, 212, 97, 5, 5, 212, 97, 160, 5, 214, 97, 5, 5, 214, 97, 160, 5, 252, 97, 197, 2, 41, 1, 197, 194, 40, 1, 160, 5, 88, 97, 5, 5, 88, 97, 160, 5, 102, 97, 5, 5, 102, 97, 160, 5, 128, 97, 5, 5, 128, 97, 160, 5, 124, 97, 5, 5, 124, 97, 160, 5, 206, 97, 5, 5, 206, 97, 160, 5, 116, 97, 5, 5, 116, 97, 5, 5, 252, 97, 197, 130, 37, 1, 197, 66, 37, 1, 197, 2, 38, 1, 197, 194, 37, 1, 160, 5, 10, 97, 5, 5, 10, 97, 197, 2, 40, 1, 197, 194, 39, 1, 160, 5, 8, 97, 5, 5, 8, 97, 197, 130, 38, 1, 197, 66, 38, 1, 197, 130, 41, 1, 197, 66, 41, 1, 197, 2, 42, 1, 197, 194, 41, 1, 160, 5, 62, 97, 5, 5, 62, 97, 197, 130, 43, 1, 197, 66, 43, 1, 197, 2, 43, 1, 197, 194, 42, 1, 197, 66, 45, 1, 197, 2, 45, 1, 160, 5, 140, 97, 5, 5, 140, 97, 197, 194, 45, 1, 197, 130, 45, 1, 197, 130, 49, 1, 197, 66, 49, 1, 197, 2, 48, 1, 197, 194, 47, 1, 197, 2, 47, 1, 197, 194, 46, 1, 197, 130, 47, 1, 197, 66, 47, 1, 197, 130, 48, 1, 197, 66, 48, 1, 160, 5, 24, 97, 5, 5, 24, 97, 197, 2, 49, 1, 197, 194, 48, 1, 160, 5, 20, 97, 5, 5, 20, 97, 198, 65, 155, 0, 193, 2, 184, 97, 198, 97, 155, 0, 193, 9, 184, 97, 160, 5, 28, 97, 5, 5, 28, 97, 160, 5, 36, 97, 5, 5, 36, 97, 160, 5, 54, 97, 5, 5, 54, 97, 160, 5, 66, 97, 5, 5, 66, 97, 160, 5, 110, 97, 5, 5, 110, 97, 160, 5, 136, 97, 5, 5, 136, 97, 160, 5, 158, 97, 5, 5, 158, 97, 198, 193, 154, 0, 193, 9, 164, 97, 198, 193, 153, 0, 193, 2, 56, 97, 198, 97, 154, 0, 193, 9, 104, 97, 160, 5, 112, 97, 5, 5, 112, 97, 160, 5, 154, 97, 5, 5, 154, 97, 198, 225, 157, 0, 193, 2, 246, 97, 198, 33, 154, 0, 193, 9, 96, 97, 160, 5, 250, 97, 5, 5, 250, 97, 198, 1, 154, 0, 193, 2, 96, 97, 198, 129, 154, 0, 193, 16, 104, 97, 160, 5, 130, 97, 5, 5, 130, 97, 160, 5, 144, 97, 5, 5, 144, 97, 198, 129, 155, 0, 193, 2, 190, 97, 160, 5, 122, 97, 5, 5, 122, 97, 198, 129, 153, 0, 193, 2, 46, 97, 198, 193, 156, 0, 193, 2, 202, 97, 198, 65, 154, 0, 193, 2, 104, 97, 255, 255, 255, 255, 160, 5, 6, 100, 160, 5, 10, 100, 160, 5, 12, 100, 160, 5, 14, 100, 174, 5, 16, 100, 160, 5, 18, 100, 160, 5, 20, 100, 160, 5, 22, 100, 160, 5, 24, 100, 160, 5, 26, 100, 174, 5, 28, 100, 160, 5, 30, 100, 174, 5, 32, 100, 160, 5, 34, 100, 160, 5, 36, 100, 160, 5, 38, 100, 160, 5, 40, 100, 160, 5, 42, 100, 160, 5, 44, 100, 174, 5, 46, 100, 160, 5, 48, 100, 174, 5, 52, 100, 160, 5, 54, 100, 160, 5, 56, 100, 160, 5, 58, 100, 160, 5, 60, 100, 160, 5, 62, 100, 160, 5, 64, 100, 160, 5, 66, 100, 174, 5, 68, 100, 160, 5, 70, 100, 160, 5, 72, 100, 160, 5, 74, 100, 174, 5, 76, 100, 160, 5, 78, 100, 160, 5, 80, 100, 160, 5, 82, 100, 160, 5, 84, 100, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 86, 100, 5, 5, 8, 11, 5, 5, 10, 11, 5, 5, 94, 7, 5, 5, 18, 7, 5, 5, 104, 7, 5, 5, 12, 11, 5, 5, 8, 100, 5, 5, 6, 100, 5, 5, 10, 100, 5, 5, 12, 100, 5, 5, 14, 100, 5, 5, 16, 100, 5, 5, 18, 100, 5, 5, 20, 100, 5, 5, 22, 100, 5, 5, 24, 100, 5, 5, 26, 100, 5, 5, 28, 100, 5, 5, 30, 100, 5, 5, 32, 100, 5, 5, 34, 100, 5, 5, 36, 100, 5, 5, 38, 100, 5, 5, 40, 100, 5, 5, 42, 100, 5, 5, 44, 100, 5, 5, 46, 100, 5, 5, 48, 100, 5, 5, 52, 100, 5, 5, 54, 100, 5, 5, 56, 100, 5, 5, 58, 100, 5, 5, 60, 100, 5, 5, 62, 100, 5, 5, 64, 100, 5, 5, 66, 100, 5, 5, 68, 100, 5, 5, 70, 100, 5, 5, 72, 100, 5, 5, 74, 100, 5, 5, 76, 100, 5, 5, 78, 100, 5, 5, 80, 100, 5, 5, 82, 100, 5, 5, 84, 100, 197, 2, 50, 1, 5, 5, 50, 100, 5, 5, 56, 7, 5, 5, 16, 5, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 85, 12, 193, 16, 85, 12, 5, 5, 144, 13, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 194, 5, 35, 226, 194, 5, 68, 226, 194, 5, 101, 226, 194, 5, 134, 226, 194, 5, 167, 226, 194, 5, 200, 226, 194, 5, 233, 226, 194, 5, 12, 227, 194, 5, 45, 227, 194, 5, 78, 227, 194, 5, 78, 227, 194, 5, 111, 227, 194, 5, 210, 227, 0, 0, 0, 0, 5, 5, 14, 11, 194, 5, 243, 227, 5, 5, 16, 11, 194, 5, 177, 227, 194, 5, 144, 227, 5, 5, 18, 11, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 20, 11, 194, 5, 45, 227, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 6, 101, 5, 5, 8, 101, 5, 5, 10, 101, 5, 5, 12, 101, 5, 5, 14, 101, 5, 5, 16, 101, 5, 5, 18, 101, 5, 5, 20, 101, 5, 5, 22, 101, 5, 5, 24, 101, 46, 5, 26, 101, 5, 5, 26, 101, 5, 5, 28, 101, 46, 5, 30, 101, 5, 5, 30, 101, 32, 5, 32, 101, 5, 5, 32, 101, 5, 5, 34, 101, 5, 5, 36, 101, 32, 5, 38, 101, 5, 5, 38, 101, 32, 5, 40, 101, 5, 5, 40, 101, 5, 5, 42, 101, 5, 5, 44, 101, 5, 5, 46, 101, 5, 5, 48, 101, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 196, 55, 1, 197, 66, 54, 1, 197, 130, 54, 1, 197, 66, 56, 1, 5, 5, 22, 11, 5, 5, 24, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 44, 151, 12, 193, 58, 151, 12, 193, 23, 85, 12, 5, 5, 158, 10, 5, 5, 162, 10, 5, 5, 146, 13, 5, 5, 20, 7, 5, 5, 22, 7, 193, 44, 85, 12, 193, 51, 85, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 44, 7, 0, 0, 0, 0, 255, 255, 255, 255, 5, 5, 58, 7, 5, 5, 106, 7, 193, 30, 248, 102, 5, 5, 6, 102, 5, 5, 8, 102, 5, 5, 10, 102, 5, 5, 16, 102, 5, 5, 18, 102, 5, 5, 22, 102, 201, 38, 3, 0, 5, 5, 30, 102, 5, 5, 40, 102, 5, 5, 42, 102, 5, 5, 44, 102, 5, 5, 58, 102, 5, 5, 72, 102, 5, 5, 74, 102, 5, 5, 84, 102, 5, 5, 86, 102, 5, 5, 110, 102, 5, 5, 112, 102, 5, 5, 134, 102, 5, 5, 136, 102, 5, 5, 146, 102, 5, 5, 148, 102, 5, 5, 158, 102, 5, 5, 160, 102, 5, 5, 166, 102, 5, 5, 168, 102, 193, 51, 194, 102, 193, 58, 194, 102, 193, 9, 248, 102, 193, 16, 248, 102, 193, 23, 248, 102, 0, 0, 0, 0, 5, 5, 174, 102, 5, 5, 178, 102, 5, 5, 182, 102, 5, 5, 196, 102, 5, 5, 200, 102, 5, 5, 204, 102, 5, 5, 214, 102, 201, 6, 5, 0, 5, 5, 236, 102, 201, 230, 5, 0, 194, 5, 164, 229, 194, 5, 9, 230, 194, 5, 108, 230, 194, 5, 207, 230, 194, 5, 85, 231, 194, 5, 184, 231, 194, 5, 29, 232, 194, 5, 62, 232, 194, 5, 95, 232, 194, 5, 128, 232, 194, 5, 161, 232, 194, 5, 227, 232, 194, 5, 6, 233, 194, 5, 39, 233, 194, 5, 105, 233, 194, 5, 138, 233, 194, 5, 171, 233, 194, 5, 204, 233, 194, 5, 237, 233, 194, 5, 16, 234, 194, 5, 194, 232, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 154, 10, 5, 5, 24, 7, 5, 5, 26, 7, 5, 5, 134, 10, 5, 5, 28, 102, 193, 72, 176, 102, 194, 5, 59, 235, 5, 5, 14, 102, 5, 5, 12, 102, 193, 2, 20, 102, 32, 5, 6, 102, 197, 226, 69, 1, 197, 98, 183, 1, 197, 162, 183, 1, 197, 162, 184, 1, 5, 5, 46, 102, 5, 5, 48, 102, 5, 5, 32, 102, 5, 5, 50, 102, 5, 5, 52, 102, 5, 5, 34, 102, 5, 5, 54, 102, 5, 5, 36, 102, 5, 5, 76, 102, 5, 5, 78, 102, 5, 5, 60, 102, 5, 5, 62, 102, 5, 5, 80, 102, 5, 5, 64, 102, 5, 5, 68, 102, 5, 5, 88, 102, 5, 5, 90, 102, 5, 5, 92, 102, 5, 5, 94, 102, 5, 5, 96, 102, 5, 5, 98, 102, 5, 5, 102, 102, 5, 5, 104, 102, 5, 5, 106, 102, 5, 5, 114, 102, 5, 5, 116, 102, 5, 5, 118, 102, 5, 5, 120, 102, 5, 5, 122, 102, 5, 5, 124, 102, 5, 5, 126, 102, 5, 5, 128, 102, 5, 5, 130, 102, 5, 5, 138, 102, 5, 5, 140, 102, 5, 5, 142, 102, 5, 5, 150, 102, 5, 5, 154, 102, 5, 5, 162, 102, 5, 5, 170, 102, 193, 2, 176, 102, 193, 9, 176, 102, 193, 23, 176, 102, 193, 30, 176, 102, 193, 44, 176, 102, 193, 51, 176, 102, 193, 2, 180, 102, 193, 16, 180, 102, 5, 5, 184, 102, 5, 5, 186, 102, 5, 5, 188, 102, 193, 2, 190, 102, 193, 16, 190, 102, 193, 23, 190, 102, 5, 5, 192, 102, 193, 9, 194, 102, 193, 16, 194, 102, 193, 23, 194, 102, 193, 30, 194, 102, 193, 37, 194, 102, 193, 2, 198, 102, 193, 9, 198, 102, 193, 16, 198, 102, 193, 23, 198, 102, 193, 9, 212, 102, 5, 5, 206, 102, 193, 9, 208, 102, 5, 5, 210, 102, 193, 2, 212, 102, 5, 5, 216, 102, 5, 5, 66, 102, 197, 34, 183, 1, 5, 5, 218, 102, 197, 226, 182, 1, 193, 2, 220, 102, 5, 5, 224, 102, 193, 2, 226, 102, 193, 9, 226, 102, 5, 5, 228, 102, 5, 5, 230, 102, 5, 5, 232, 102, 193, 2, 234, 102, 193, 9, 234, 102, 5, 5, 240, 102, 5, 5, 242, 102, 5, 5, 244, 102, 193, 23, 234, 102, 5, 5, 246, 102, 193, 2, 248, 102, 5, 5, 250, 102, 197, 194, 190, 1, 5, 5, 8, 9, 193, 16, 220, 102, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 58, 85, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 5, 222, 102, 16, 5, 238, 102, 0, 0, 0, 0, 0, 0, 0, 0, 193, 65, 85, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 2, 108, 102, 193, 2, 132, 102, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 193, 2, 144, 102, 5, 5, 156, 102, 193, 9, 172, 102, 5, 112, 6, 102, 5, 112, 200, 102, 193, 9, 220, 102, 193, 28, 53, 9, 193, 2, 10, 9, 193, 9, 10, 9, 193, 2, 60, 7, 193, 9, 60, 7, 193, 16, 60, 7, 193, 23, 60, 7, 193, 30, 60, 7, 193, 37, 60, 7, 5, 5, 108, 7, 193, 2, 26, 11, 193, 9, 26, 11, 193, 16, 26, 11, 193, 23, 26, 11, 255, 255, 255, 255, 0, 0, 0, 0, 193, 9, 4, 103, 194, 5, 92, 235, 5, 5, 6, 103, 5, 5, 8, 103, 5, 112, 8, 103, 5, 5, 12, 103, 5, 5, 10, 103, 193, 2, 14, 103, 193, 9, 14, 103, 193, 16, 14, 103, 193, 30, 14, 103, 5, 5, 16, 103, 5, 112, 16, 103, 193, 2, 18, 103, 193, 9, 18, 103, 193, 16, 18, 103, 193, 30, 18, 103, 193, 37, 18, 103, 193, 44, 18, 103, 5, 5, 20, 103, 32, 5, 20, 103, 5, 5, 22, 103, 5, 5, 24, 103, 5, 112, 24, 103, 193, 9, 26, 103, 193, 16, 26, 103, 193, 23, 26, 103, 193, 30, 26, 103, 193, 37, 26, 103, 5, 112, 6, 103, 5, 120, 8, 103, 5, 112, 12, 103, 194, 5, 125, 235, 194, 5, 158, 235, 194, 5, 191, 235, 194, 5, 224, 235, 194, 5, 3, 236, 194, 5, 36, 236, 194, 5, 69, 236, 194, 5, 102, 236, 194, 5, 135, 236, 194, 5, 168, 236, 194, 5, 201, 236, 194, 5, 234, 236, 194, 5, 13, 237, 194, 5, 46, 237, 194, 5, 79, 237, 194, 5, 112, 237, 0, 0, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 193, 23, 14, 103, 193, 23, 18, 103, 193, 2, 26, 103, 193, 2, 38, 102, 193, 9, 38, 102, 193, 16, 38, 102, 193, 23, 38, 102, 193, 30, 38, 102, 193, 37, 38, 102, 193, 51, 38, 102, 193, 2, 82, 102, 193, 9, 82, 102, 193, 9, 108, 102, 193, 16, 108, 102, 193, 9, 132, 102, 193, 9, 144, 102, 193, 16, 172, 102, 193, 23, 172, 102, 193, 30, 172, 102, 193, 58, 176, 102, 193, 65, 176, 102, 193, 44, 194, 102, 193, 65, 194, 102, 193, 72, 194, 102, 193, 2, 202, 102, 193, 9, 202, 102, 193, 16, 212, 102, 193, 23, 212, 102, 193, 30, 212, 102, 193, 30, 198, 102, 193, 16, 132, 102, 193, 23, 132, 102, 193, 16, 144, 102, 193, 16, 82, 102, 193, 23, 82, 102, 193, 23, 144, 102, 193, 30, 132, 102, 193, 30, 82, 102, 193, 9, 20, 102, 193, 16, 20, 102, 193, 37, 248, 102, 193, 44, 248, 102, 193, 51, 248, 102, 193, 30, 234, 102, 193, 37, 234, 102, 193, 2, 252, 102, 193, 9, 252, 102, 193, 37, 82, 102, 193, 30, 144, 102, 193, 37, 144, 102, 193, 9, 190, 102, 5, 5, 30, 103, 5, 5, 36, 103, 5, 5, 38, 103, 5, 5, 40, 103, 5, 5, 44, 103, 5, 5, 46, 103, 5, 5, 48, 103, 5, 5, 50, 103, 5, 5, 56, 103, 5, 5, 60, 103, 5, 5, 62, 103, 5, 5, 64, 103, 5, 5, 68, 103, 5, 5, 76, 103, 5, 5, 78, 103, 5, 5, 82, 103, 5, 5, 84, 103, 5, 5, 92, 103, 5, 5, 94, 103, 5, 5, 96, 103, 5, 5, 98, 103, 5, 5, 100, 103, 5, 5, 102, 103, 5, 5, 104, 103, 5, 5, 70, 103, 5, 5, 32, 103, 5, 5, 34, 103, 5, 5, 66, 103, 5, 5, 42, 103, 5, 5, 86, 103, 5, 5, 88, 103, 5, 5, 90, 103, 5, 5, 72, 103, 5, 5, 74, 103, 5, 5, 52, 103, 5, 5, 54, 103, 5, 5, 80, 103, 5, 5, 58, 103, 5, 5, 108, 103, 5, 5, 110, 103, 5, 5, 112, 103, 5, 5, 114, 103, 5, 5, 116, 103, 5, 5, 118, 103, 5, 5, 120, 103, 5, 5, 122, 103, 5, 5, 124, 103, 5, 5, 126, 103, 5, 5, 128, 103, 5, 5, 106, 103, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 193, 9, 130, 103, 193, 16, 130, 103, 193, 23, 130, 103, 193, 30, 130, 103, 193, 37, 130, 103, 193, 44, 130, 103, 193, 51, 130, 103, 193, 58, 130, 103, 193, 65, 130, 103, 193, 72, 130, 103, 193, 79, 130, 103, 193, 86, 130, 103, 5, 5, 132, 103, 5, 5, 134, 103, 5, 5, 136, 103, 5, 5, 138, 103, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 206, 32, 227, 0, 5, 112, 132, 103, 5, 112, 134, 103, 5, 112, 138, 103, 194, 5, 145, 237, 194, 5, 178, 237, 194, 5, 211, 237, 194, 5, 244, 237, 194, 5, 23, 238, 194, 5, 56, 238, 194, 5, 89, 238, 194, 5, 122, 238, 194, 5, 155, 238, 193, 100, 140, 103, 193, 107, 140, 103, 193, 191, 85, 12, 193, 35, 53, 9, 5, 5, 28, 7, 193, 2, 96, 7, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 5, 168, 0, 0, 193, 2, 148, 13, 193, 9, 148, 13, 32, 5, 136, 13, 32, 5, 140, 13, 32, 5, 135, 12, 32, 5, 12, 12, 32, 5, 139, 12, 32, 5, 142, 13, 32, 5, 192, 13, 255, 255, 255, 255, 160, 5, 172, 12, 160, 5, 99, 12, 160, 5, 103, 12, 160, 5, 101, 12, 160, 5, 105, 12, 160, 5, 180, 12, 160, 5, 185, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 58, 78, 13, 5, 5, 253, 255, 5, 5, 0, 2, 5, 5, 255, 255, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 197, 67, 167, 1, 197, 67, 144, 1, 197, 35, 72, 1, 197, 131, 153, 1, 197, 227, 134, 1, 197, 35, 123, 1, 197, 195, 102, 1, 197, 35, 174, 1, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 201, 192, 184, 1, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 2, 2, 251, 193, 6, 2, 251, 193, 8, 2, 251, 193, 10, 2, 251, 193, 12, 2, 251, 193, 14, 2, 251, 193, 16, 2, 251, 193, 28, 2, 251, 193, 30, 2, 251, 193, 32, 2, 251, 193, 34, 2, 251, 193, 36, 2, 251, 193, 38, 2, 251, 193, 76, 2, 251, 193, 78, 2, 251, 193, 80, 2, 251, 193, 82, 2, 251, 193, 84, 2, 251, 193, 86, 2, 251, 193, 88, 2, 251, 193, 124, 2, 251, 193, 126, 2, 251, 193, 128, 2, 251, 193, 90, 2, 251, 193, 130, 2, 251, 193, 132, 2, 251, 193, 134, 2, 251, 193, 136, 2, 251, 193, 138, 2, 251, 193, 140, 2, 251, 193, 186, 2, 251, 193, 188, 2, 251, 193, 194, 224, 253, 193, 198, 224, 253, 193, 200, 224, 253, 193, 202, 224, 253, 193, 204, 224, 253, 193, 208, 224, 253, 193, 210, 224, 253, 193, 214, 224, 253, 193, 216, 224, 253, 193, 122, 224, 253, 193, 242, 224, 253, 193, 254, 224, 253, 193, 2, 225, 253, 193, 4, 225, 253, 193, 16, 225, 253, 193, 18, 225, 253, 193, 20, 225, 253, 193, 22, 225, 253, 193, 26, 225, 253, 193, 28, 225, 253, 193, 34, 225, 253, 193, 36, 225, 253, 193, 38, 225, 253, 193, 96, 57, 251, 193, 108, 83, 251, 193, 4, 2, 251, 193, 52, 2, 251, 193, 198, 131, 251, 193, 212, 147, 251, 193, 208, 33, 251, 255, 255, 255, 255, 255, 255, 255, 255, 193, 18, 2, 251, 193, 20, 2, 251, 193, 22, 2, 251, 193, 44, 2, 251, 193, 46, 2, 251, 193, 48, 2, 251, 193, 50, 2, 251, 193, 92, 2, 251, 193, 94, 2, 251, 193, 96, 2, 251, 193, 98, 2, 251, 193, 100, 2, 251, 193, 102, 2, 251, 193, 144, 2, 251, 193, 146, 2, 251, 193, 148, 2, 251, 193, 150, 2, 251, 193, 152, 2, 251, 193, 154, 2, 251, 193, 156, 2, 251, 193, 158, 2, 251, 193, 160, 2, 251, 193, 162, 2, 251, 193, 164, 2, 251, 193, 200, 2, 251, 193, 202, 2, 251, 193, 204, 2, 251, 193, 32, 103, 251, 193, 206, 2, 251, 193, 64, 49, 251, 193, 208, 2, 251, 193, 210, 2, 251, 193, 6, 225, 253, 193, 32, 225, 253, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 24, 2, 251, 193, 60, 2, 251, 193, 62, 2, 251, 193, 64, 2, 251, 193, 66, 2, 251, 193, 108, 2, 251, 193, 110, 2, 251, 193, 112, 2, 251, 193, 114, 2, 251, 193, 116, 2, 251, 193, 170, 2, 251, 193, 172, 2, 251, 193, 174, 2, 251, 193, 176, 2, 251, 193, 178, 2, 251, 193, 220, 2, 251, 193, 26, 2, 251, 193, 68, 2, 251, 193, 70, 2, 251, 193, 72, 2, 251, 193, 74, 2, 251, 193, 118, 2, 251, 193, 120, 2, 251, 193, 122, 2, 251, 193, 180, 2, 251, 193, 182, 2, 251, 193, 184, 2, 251, 193, 242, 2, 251, 193, 46, 3, 251, 193, 90, 3, 251, 193, 118, 3, 251, 193, 120, 3, 251, 193, 122, 3, 251, 193, 124, 3, 251, 193, 152, 3, 251, 193, 154, 3, 251, 193, 156, 3, 251, 193, 158, 3, 251, 193, 206, 3, 251, 193, 208, 3, 251, 193, 210, 3, 251, 193, 240, 3, 251, 193, 242, 3, 251, 193, 6, 4, 251, 193, 8, 4, 251, 193, 22, 4, 251, 193, 46, 4, 251, 193, 56, 4, 251, 193, 146, 7, 251, 193, 184, 7, 251, 193, 202, 7, 251, 193, 204, 7, 251, 193, 206, 7, 251, 193, 248, 7, 251, 193, 252, 7, 251, 193, 8, 8, 251, 193, 18, 8, 251, 193, 30, 8, 251, 193, 38, 8, 251, 193, 48, 8, 251, 193, 74, 8, 251, 193, 82, 8, 251, 193, 84, 8, 251, 193, 86, 8, 251, 193, 40, 2, 251, 193, 90, 8, 251, 193, 92, 8, 251, 193, 94, 8, 251, 193, 96, 8, 251, 193, 98, 8, 251, 193, 100, 8, 251, 193, 106, 8, 251, 193, 110, 8, 251, 193, 112, 8, 251, 193, 114, 8, 251, 193, 132, 8, 251, 193, 134, 8, 251, 193, 146, 8, 251, 193, 148, 8, 251, 193, 150, 8, 251, 193, 168, 47, 251, 193, 180, 47, 251, 193, 182, 47, 251, 193, 184, 47, 251, 193, 42, 2, 251, 193, 186, 47, 251, 193, 188, 47, 251, 193, 190, 47, 251, 193, 192, 47, 251, 193, 208, 47, 251, 193, 210, 47, 251, 193, 226, 47, 251, 193, 212, 47, 251, 193, 228, 47, 251, 193, 230, 47, 251, 193, 232, 47, 251, 193, 234, 47, 251, 193, 46, 48, 251, 193, 48, 48, 251, 193, 50, 48, 251, 193, 52, 48, 251, 193, 54, 48, 251, 193, 56, 48, 251, 193, 92, 48, 251, 193, 94, 48, 251, 193, 132, 48, 251, 193, 154, 48, 251, 193, 30, 49, 251, 193, 44, 49, 251, 193, 48, 49, 251, 193, 50, 49, 251, 193, 54, 49, 251, 193, 54, 2, 251, 193, 56, 2, 251, 193, 104, 2, 251, 193, 68, 3, 251, 193, 104, 3, 251, 193, 136, 3, 251, 193, 170, 3, 251, 193, 40, 4, 251, 193, 92, 4, 251, 193, 142, 4, 251, 193, 204, 4, 251, 193, 234, 4, 251, 193, 154, 5, 251, 193, 196, 5, 251, 193, 216, 5, 251, 193, 218, 5, 251, 193, 140, 6, 251, 193, 200, 6, 251, 193, 248, 6, 251, 193, 250, 6, 251, 193, 40, 7, 251, 193, 88, 7, 251, 193, 122, 7, 251, 193, 124, 7, 251, 193, 168, 7, 251, 193, 190, 7, 251, 193, 192, 7, 251, 193, 212, 7, 251, 193, 46, 8, 251, 193, 64, 8, 251, 193, 216, 8, 251, 193, 224, 8, 251, 193, 58, 2, 251, 193, 214, 2, 251, 193, 18, 3, 251, 193, 254, 3, 251, 193, 52, 5, 251, 193, 18, 9, 251, 193, 234, 10, 251, 193, 52, 14, 251, 193, 54, 14, 251, 193, 64, 17, 251, 193, 196, 18, 251, 193, 56, 21, 251, 193, 158, 26, 251, 193, 236, 29, 251, 193, 58, 32, 251, 193, 192, 32, 251, 193, 8, 38, 251, 193, 48, 44, 251, 193, 36, 48, 251, 193, 84, 48, 251, 193, 96, 48, 251, 193, 50, 51, 251, 193, 234, 51, 251, 193, 236, 51, 251, 193, 98, 57, 251, 193, 196, 61, 251, 193, 188, 73, 251, 193, 176, 75, 251, 193, 182, 78, 251, 193, 230, 79, 251, 193, 104, 82, 251, 193, 50, 93, 251, 193, 112, 161, 251, 193, 106, 2, 251, 193, 166, 2, 251, 193, 168, 2, 251, 193, 216, 2, 251, 193, 218, 2, 251, 193, 20, 3, 251, 193, 22, 3, 251, 193, 24, 3, 251, 193, 26, 3, 251, 193, 70, 3, 251, 193, 72, 3, 251, 193, 74, 3, 251, 193, 76, 3, 251, 193, 106, 3, 251, 193, 138, 3, 251, 193, 140, 3, 251, 193, 142, 3, 251, 193, 144, 3, 251, 193, 172, 3, 251, 193, 174, 3, 251, 193, 176, 3, 251, 193, 198, 3, 251, 193, 216, 3, 251, 193, 218, 3, 251, 193, 220, 3, 251, 193, 222, 3, 251, 193, 224, 3, 251, 193, 226, 3, 251, 193, 2, 4, 251, 193, 16, 4, 251, 193, 18, 4, 251, 193, 142, 2, 251, 193, 196, 2, 251, 193, 198, 2, 251, 193, 72, 4, 251, 193, 74, 4, 251, 193, 10, 5, 251, 193, 132, 5, 251, 193, 62, 6, 251, 193, 64, 6, 251, 193, 66, 6, 251, 193, 86, 6, 251, 193, 88, 6, 251, 193, 120, 6, 251, 193, 122, 6, 251, 193, 124, 6, 251, 193, 126, 6, 251, 193, 178, 6, 251, 193, 180, 6, 251, 193, 182, 6, 251, 193, 184, 6, 251, 193, 186, 6, 251, 193, 188, 6, 251, 193, 234, 6, 251, 193, 236, 6, 251, 193, 238, 6, 251, 193, 28, 7, 251, 193, 30, 7, 251, 193, 32, 7, 251, 193, 64, 7, 251, 193, 66, 7, 251, 193, 68, 7, 251, 193, 70, 7, 251, 193, 190, 2, 251, 193, 192, 2, 251, 193, 194, 2, 251, 193, 244, 2, 251, 193, 246, 2, 251, 193, 248, 2, 251, 193, 48, 3, 251, 193, 50, 3, 251, 193, 60, 4, 251, 193, 62, 4, 251, 193, 68, 4, 251, 193, 70, 4, 251, 193, 14, 66, 252, 193, 82, 4, 251, 193, 84, 4, 251, 193, 86, 4, 251, 193, 88, 4, 251, 193, 98, 4, 251, 193, 112, 4, 251, 193, 120, 4, 251, 193, 134, 4, 251, 193, 148, 4, 251, 193, 200, 4, 251, 193, 202, 4, 251, 193, 206, 4, 251, 193, 216, 4, 251, 193, 218, 4, 251, 193, 222, 4, 251, 193, 224, 4, 251, 193, 250, 2, 251, 193, 236, 4, 251, 193, 252, 4, 251, 193, 212, 2, 251, 193, 254, 2, 251, 193, 2, 3, 251, 193, 4, 3, 251, 193, 6, 3, 251, 193, 136, 8, 251, 193, 8, 3, 251, 193, 10, 3, 251, 193, 12, 3, 251, 193, 32, 8, 251, 193, 14, 3, 251, 193, 16, 3, 251, 193, 52, 3, 251, 193, 54, 3, 251, 193, 56, 3, 251, 193, 14, 84, 251, 193, 58, 3, 251, 193, 60, 3, 251, 193, 62, 3, 251, 193, 64, 3, 251, 193, 66, 3, 251, 193, 94, 3, 251, 193, 96, 3, 251, 193, 242, 60, 251, 193, 98, 3, 251, 193, 244, 60, 251, 193, 100, 3, 251, 193, 102, 3, 251, 193, 24, 101, 251, 193, 126, 3, 251, 193, 128, 3, 251, 193, 130, 3, 251, 193, 222, 2, 251, 193, 224, 2, 251, 193, 226, 2, 251, 193, 228, 2, 251, 193, 230, 2, 251, 193, 232, 2, 251, 193, 234, 2, 251, 193, 236, 2, 251, 193, 238, 2, 251, 193, 240, 2, 251, 193, 28, 3, 251, 193, 30, 3, 251, 193, 32, 3, 251, 193, 34, 3, 251, 193, 36, 3, 251, 193, 38, 3, 251, 193, 40, 3, 251, 193, 42, 3, 251, 193, 44, 3, 251, 193, 78, 3, 251, 193, 80, 3, 251, 193, 82, 3, 251, 193, 84, 3, 251, 193, 86, 3, 251, 193, 88, 3, 251, 193, 108, 3, 251, 193, 110, 3, 251, 193, 112, 3, 251, 193, 114, 3, 251, 193, 116, 3, 251, 193, 146, 3, 251, 193, 148, 3, 251, 193, 236, 250, 252, 193, 174, 32, 253, 193, 186, 74, 253, 193, 130, 116, 252, 193, 190, 239, 251, 193, 142, 139, 252, 193, 54, 135, 252, 193, 182, 13, 251, 193, 136, 166, 251, 193, 138, 166, 251, 193, 84, 228, 252, 193, 128, 22, 253, 193, 98, 32, 252, 193, 170, 88, 251, 193, 164, 127, 252, 193, 110, 116, 253, 193, 106, 189, 253, 193, 160, 31, 251, 193, 156, 226, 252, 193, 140, 233, 252, 193, 184, 117, 253, 193, 102, 193, 253, 193, 252, 2, 251, 193, 92, 3, 251, 193, 102, 19, 251, 193, 154, 20, 251, 193, 188, 23, 251, 193, 192, 33, 251, 193, 168, 211, 251, 193, 144, 6, 252, 193, 142, 14, 252, 193, 208, 142, 252, 193, 132, 3, 251, 193, 160, 3, 251, 193, 162, 3, 251, 193, 164, 3, 251, 193, 166, 3, 251, 193, 168, 3, 251, 193, 188, 3, 251, 193, 190, 3, 251, 193, 192, 3, 251, 193, 140, 34, 251, 193, 194, 3, 251, 193, 196, 3, 251, 193, 212, 3, 251, 193, 214, 3, 251, 193, 244, 3, 251, 193, 246, 3, 251, 193, 248, 3, 251, 193, 250, 3, 251, 193, 252, 3, 251, 193, 12, 4, 251, 193, 14, 4, 251, 193, 24, 4, 251, 193, 44, 4, 251, 193, 26, 4, 251, 193, 28, 4, 251, 193, 30, 4, 251, 193, 32, 4, 251, 193, 34, 4, 251, 193, 76, 29, 251, 193, 36, 4, 251, 193, 216, 230, 251, 193, 48, 4, 251, 193, 44, 46, 251, 193, 46, 46, 251, 193, 134, 3, 251, 193, 60, 46, 251, 193, 62, 46, 251, 193, 66, 46, 251, 193, 70, 46, 251, 193, 72, 46, 251, 193, 74, 46, 251, 193, 76, 46, 251, 193, 78, 46, 251, 193, 22, 218, 252, 193, 84, 46, 251, 193, 86, 46, 251, 193, 114, 46, 251, 193, 116, 46, 251, 193, 118, 46, 251, 193, 120, 46, 251, 193, 126, 47, 251, 193, 140, 46, 251, 193, 142, 46, 251, 193, 144, 46, 251, 193, 146, 46, 251, 193, 148, 46, 251, 193, 166, 46, 251, 193, 168, 46, 251, 193, 170, 46, 251, 193, 172, 46, 251, 193, 186, 46, 251, 193, 188, 46, 251, 193, 190, 46, 251, 193, 192, 46, 251, 193, 150, 3, 251, 193, 178, 3, 251, 193, 180, 3, 251, 193, 182, 3, 251, 193, 184, 3, 251, 193, 186, 3, 251, 193, 200, 3, 251, 193, 202, 3, 251, 193, 204, 3, 251, 193, 228, 3, 251, 193, 230, 3, 251, 193, 232, 3, 251, 193, 234, 3, 251, 193, 236, 3, 251, 193, 238, 3, 251, 193, 4, 4, 251, 193, 20, 4, 251, 193, 58, 4, 251, 193, 106, 4, 251, 193, 108, 4, 251, 193, 144, 4, 251, 193, 158, 4, 251, 193, 160, 4, 251, 193, 162, 4, 251, 193, 174, 4, 251, 193, 176, 4, 251, 193, 214, 4, 251, 193, 232, 4, 251, 193, 238, 4, 251, 193, 244, 4, 251, 193, 248, 4, 251, 193, 22, 5, 251, 193, 98, 222, 253, 193, 100, 222, 253, 193, 102, 222, 253, 193, 160, 222, 253, 193, 162, 222, 253, 193, 206, 222, 253, 193, 208, 222, 253, 193, 236, 222, 253, 193, 26, 223, 253, 193, 44, 223, 253, 193, 46, 223, 253, 193, 98, 223, 253, 193, 56, 224, 253, 193, 102, 224, 253, 193, 104, 224, 253, 193, 136, 224, 253, 193, 142, 224, 253, 193, 168, 224, 253, 193, 170, 224, 253, 193, 240, 224, 253, 193, 252, 224, 253, 193, 24, 225, 253, 193, 10, 4, 251, 193, 188, 56, 251, 193, 30, 57, 251, 193, 10, 135, 251, 193, 134, 233, 251, 193, 136, 233, 251, 193, 48, 234, 251, 193, 108, 234, 251, 193, 110, 7, 252, 193, 196, 157, 253, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 193, 76, 97, 252, 193, 78, 97, 252, 193, 80, 97, 252, 193, 82, 97, 252, 193, 84, 97, 252, 193, 86, 97, 252, 193, 88, 97, 252, 193, 38, 4, 251, 193, 90, 97, 252, 193, 92, 97, 252, 193, 130, 97, 252, 193, 132, 97, 252, 193, 134, 97, 252, 193, 136, 97, 252, 193, 138, 97, 252, 193, 140, 97, 252, 193, 142, 97, 252, 193, 144, 97, 252, 193, 146, 97, 252, 193, 76, 98, 253, 206, 128, 56, 0, 206, 128, 56, 0, 206, 128, 56, 0, 206, 128, 56, 0, 193, 42, 4, 251, 193, 50, 4, 251, 193, 52, 4, 251, 193, 76, 4, 251, 193, 78, 4, 251, 193, 94, 4, 251, 193, 118, 4, 251, 193, 130, 4, 251, 193, 154, 4, 251, 193, 156, 4, 251, 193, 180, 4, 251, 193, 184, 4, 251, 193, 186, 4, 251, 193, 194, 4, 251, 193, 198, 4, 251, 193, 54, 5, 251, 193, 116, 5, 251, 193, 156, 5, 251, 193, 158, 5, 251, 193, 160, 5, 251, 193, 168, 5, 251, 193, 180, 5, 251, 193, 182, 5, 251, 193, 184, 5, 251, 193, 186, 5, 251, 193, 198, 5, 251, 193, 200, 5, 251, 193, 220, 5, 251, 193, 222, 5, 251, 193, 240, 5, 251, 193, 248, 5, 251, 193, 10, 6, 251, 193, 54, 4, 251, 193, 64, 4, 251, 193, 66, 4, 251, 193, 90, 4, 251, 193, 98, 83, 251, 193, 100, 4, 251, 193, 104, 4, 251, 193, 114, 4, 251, 193, 116, 4, 251, 193, 122, 4, 251, 193, 124, 4, 251, 193, 126, 4, 251, 193, 128, 4, 251, 193, 136, 4, 251, 193, 138, 4, 251, 193, 140, 4, 251, 193, 150, 4, 251, 193, 152, 4, 251, 193, 2, 126, 253, 193, 168, 4, 251, 193, 170, 4, 251, 193, 172, 4, 251, 193, 178, 4, 251, 193, 132, 72, 251, 193, 182, 4, 251, 193, 188, 4, 251, 193, 192, 4, 251, 193, 196, 4, 251, 193, 210, 4, 251, 193, 212, 4, 251, 193, 90, 6, 251, 193, 220, 4, 251, 193, 80, 4, 251, 193, 96, 4, 251, 193, 102, 4, 251, 193, 110, 4, 251, 193, 132, 4, 251, 193, 146, 4, 251, 193, 164, 4, 251, 193, 166, 4, 251, 193, 190, 4, 251, 193, 242, 4, 251, 193, 30, 5, 251, 193, 32, 5, 251, 193, 122, 5, 251, 193, 250, 5, 251, 193, 252, 5, 251, 193, 254, 5, 251, 193, 76, 6, 251, 193, 116, 6, 251, 193, 152, 6, 251, 193, 154, 6, 251, 193, 156, 6, 251, 193, 158, 6, 251, 193, 216, 6, 251, 193, 218, 6, 251, 193, 220, 6, 251, 193, 222, 6, 251, 193, 224, 6, 251, 193, 226, 6, 251, 193, 228, 6, 251, 193, 8, 7, 251, 193, 10, 7, 251, 193, 12, 7, 251, 193, 254, 4, 251, 193, 2, 5, 251, 193, 4, 5, 251, 193, 6, 5, 251, 193, 8, 5, 251, 193, 34, 5, 251, 193, 36, 5, 251, 193, 38, 5, 251, 193, 40, 5, 251, 193, 208, 4, 251, 193, 42, 5, 251, 193, 58, 5, 251, 193, 60, 5, 251, 193, 80, 5, 251, 193, 82, 5, 251, 193, 84, 5, 251, 193, 86, 5, 251, 193, 124, 5, 251, 193, 126, 5, 251, 193, 128, 5, 251, 193, 130, 5, 251, 193, 164, 5, 251, 193, 172, 5, 251, 193, 190, 5, 251, 193, 202, 5, 251, 193, 26, 6, 251, 193, 28, 6, 251, 193, 30, 6, 251, 193, 46, 6, 251, 193, 48, 6, 251, 193, 56, 6, 251, 193, 58, 6, 251, 193, 226, 4, 251, 193, 228, 4, 251, 193, 230, 4, 251, 193, 240, 4, 251, 193, 246, 4, 251, 193, 250, 4, 251, 193, 12, 5, 251, 193, 14, 5, 251, 193, 16, 5, 251, 193, 18, 5, 251, 193, 20, 5, 251, 193, 44, 5, 251, 193, 46, 5, 251, 193, 48, 5, 251, 193, 50, 5, 251, 193, 64, 5, 251, 193, 66, 5, 251, 193, 68, 5, 251, 193, 70, 5, 251, 193, 72, 5, 251, 193, 88, 5, 251, 193, 90, 5, 251, 193, 92, 5, 251, 193, 94, 5, 251, 193, 96, 5, 251, 193, 98, 5, 251, 193, 100, 5, 251, 193, 102, 5, 251, 193, 104, 5, 251, 193, 106, 5, 251, 193, 108, 5, 251, 193, 110, 5, 251, 193, 24, 5, 251, 193, 26, 5, 251, 193, 28, 5, 251, 193, 56, 5, 251, 193, 74, 5, 251, 193, 76, 5, 251, 193, 78, 5, 251, 193, 118, 5, 251, 193, 120, 5, 251, 193, 162, 5, 251, 193, 170, 5, 251, 193, 188, 5, 251, 193, 224, 5, 251, 193, 226, 5, 251, 193, 228, 5, 251, 193, 230, 5, 251, 193, 242, 5, 251, 193, 6, 6, 251, 193, 14, 6, 251, 193, 16, 6, 251, 193, 18, 6, 251, 193, 20, 6, 251, 193, 74, 6, 251, 193, 112, 6, 251, 193, 114, 6, 251, 193, 144, 6, 251, 193, 146, 6, 251, 193, 148, 6, 251, 193, 150, 6, 251, 193, 204, 6, 251, 193, 206, 6, 251, 193, 208, 6, 251, 193, 18, 130, 251, 193, 36, 130, 251, 193, 144, 130, 251, 193, 146, 130, 251, 193, 152, 130, 251, 193, 172, 130, 251, 193, 174, 130, 251, 193, 176, 130, 251, 193, 190, 130, 251, 193, 192, 130, 251, 193, 194, 130, 251, 193, 196, 130, 251, 193, 198, 130, 251, 193, 226, 130, 251, 193, 254, 130, 251, 193, 2, 131, 251, 193, 72, 131, 251, 193, 74, 131, 251, 193, 102, 131, 251, 193, 104, 131, 251, 193, 106, 131, 251, 193, 146, 131, 251, 193, 172, 131, 251, 193, 174, 131, 251, 193, 192, 131, 251, 193, 62, 5, 251, 193, 196, 131, 251, 193, 202, 131, 251, 193, 204, 131, 251, 193, 222, 131, 251, 193, 252, 131, 251, 193, 254, 131, 251, 193, 112, 5, 251, 193, 114, 5, 251, 193, 134, 5, 251, 193, 136, 5, 251, 193, 138, 5, 251, 193, 140, 5, 251, 193, 142, 5, 251, 193, 144, 5, 251, 193, 130, 219, 252, 193, 146, 5, 251, 193, 148, 5, 251, 193, 150, 5, 251, 193, 232, 50, 251, 193, 152, 5, 251, 193, 234, 50, 251, 193, 166, 5, 251, 193, 26, 51, 251, 193, 184, 16, 252, 193, 174, 5, 251, 193, 176, 5, 251, 193, 192, 5, 251, 193, 224, 93, 252, 193, 178, 5, 251, 193, 194, 5, 251, 193, 74, 7, 251, 193, 204, 5, 251, 193, 206, 5, 251, 193, 208, 5, 251, 193, 210, 5, 251, 193, 212, 5, 251, 193, 214, 5, 251, 193, 232, 5, 251, 193, 234, 5, 251, 193, 236, 5, 251, 193, 238, 5, 251, 193, 244, 5, 251, 193, 246, 5, 251, 193, 4, 6, 251, 193, 2, 6, 251, 193, 8, 6, 251, 193, 22, 6, 251, 193, 32, 6, 251, 193, 34, 6, 251, 193, 36, 6, 251, 193, 38, 6, 251, 193, 40, 6, 251, 193, 42, 6, 251, 193, 52, 6, 251, 193, 54, 6, 251, 193, 44, 6, 251, 193, 68, 6, 251, 193, 70, 6, 251, 193, 72, 6, 251, 193, 92, 6, 251, 193, 94, 6, 251, 193, 96, 6, 251, 193, 98, 6, 251, 193, 100, 6, 251, 193, 102, 6, 251, 193, 104, 6, 251, 193, 106, 6, 251, 193, 108, 6, 251, 193, 128, 6, 251, 193, 130, 6, 251, 193, 12, 6, 251, 193, 24, 6, 251, 193, 110, 6, 251, 193, 142, 6, 251, 193, 202, 6, 251, 193, 252, 6, 251, 193, 254, 6, 251, 193, 2, 7, 251, 193, 42, 7, 251, 193, 90, 7, 251, 193, 126, 7, 251, 193, 128, 7, 251, 193, 130, 7, 251, 193, 170, 7, 251, 193, 172, 7, 251, 193, 214, 7, 251, 193, 216, 7, 251, 193, 128, 8, 251, 193, 144, 8, 251, 193, 164, 8, 251, 193, 196, 8, 251, 193, 208, 8, 251, 193, 210, 8, 251, 193, 218, 8, 251, 193, 228, 8, 251, 193, 230, 8, 251, 193, 42, 9, 251, 193, 114, 9, 251, 193, 176, 9, 251, 193, 212, 9, 251, 193, 214, 9, 251, 193, 228, 9, 251, 193, 236, 224, 253, 193, 238, 224, 253, 193, 246, 224, 253, 193, 250, 224, 253, 193, 12, 225, 253, 193, 14, 225, 253, 193, 74, 58, 252, 193, 12, 230, 251, 193, 218, 46, 251, 193, 176, 243, 252, 193, 102, 55, 253, 193, 214, 71, 253, 193, 86, 104, 253, 193, 62, 170, 253, 193, 182, 171, 253, 193, 212, 100, 253, 193, 192, 9, 252, 193, 94, 83, 251, 193, 84, 107, 253, 193, 78, 97, 251, 193, 50, 6, 251, 193, 162, 191, 251, 193, 192, 85, 251, 193, 14, 223, 252, 193, 60, 129, 251, 193, 110, 28, 251, 193, 58, 48, 251, 193, 66, 38, 253, 193, 252, 97, 251, 193, 76, 58, 252, 193, 202, 109, 252, 193, 208, 230, 252, 193, 60, 6, 251, 193, 90, 154, 251, 193, 130, 135, 251, 193, 78, 6, 251, 193, 80, 6, 251, 193, 82, 6, 251, 193, 84, 6, 251, 193, 118, 6, 251, 193, 160, 6, 251, 193, 162, 6, 251, 193, 164, 6, 251, 193, 166, 6, 251, 193, 168, 6, 251, 193, 170, 6, 251, 193, 172, 6, 251, 193, 174, 6, 251, 193, 176, 6, 251, 193, 230, 6, 251, 193, 232, 6, 251, 193, 16, 7, 251, 193, 18, 7, 251, 193, 20, 7, 251, 193, 22, 7, 251, 193, 24, 7, 251, 193, 26, 7, 251, 193, 56, 7, 251, 193, 58, 7, 251, 193, 60, 7, 251, 193, 62, 7, 251, 193, 110, 7, 251, 193, 142, 7, 251, 193, 144, 7, 251, 193, 132, 6, 251, 193, 134, 6, 251, 193, 136, 6, 251, 193, 138, 6, 251, 193, 190, 6, 251, 193, 192, 6, 251, 193, 194, 6, 251, 193, 196, 6, 251, 193, 198, 6, 251, 193, 240, 6, 251, 193, 242, 6, 251, 193, 72, 94, 252, 193, 244, 6, 251, 193, 34, 7, 251, 193, 36, 7, 251, 193, 38, 7, 251, 193, 226, 93, 252, 193, 238, 119, 251, 193, 76, 7, 251, 193, 78, 7, 251, 193, 80, 7, 251, 193, 82, 7, 251, 193, 84, 7, 251, 193, 86, 7, 251, 193, 112, 7, 251, 193, 114, 7, 251, 193, 116, 7, 251, 193, 118, 7, 251, 193, 154, 7, 251, 193, 156, 7, 251, 193, 158, 7, 251, 193, 160, 7, 251, 193, 210, 6, 251, 193, 212, 6, 251, 193, 214, 6, 251, 193, 4, 7, 251, 193, 6, 7, 251, 193, 44, 7, 251, 193, 46, 7, 251, 193, 48, 7, 251, 193, 92, 7, 251, 193, 94, 7, 251, 193, 96, 7, 251, 193, 98, 7, 251, 193, 132, 7, 251, 193, 134, 7, 251, 193, 174, 7, 251, 193, 176, 7, 251, 193, 178, 7, 251, 193, 180, 7, 251, 193, 194, 7, 251, 193, 196, 7, 251, 193, 198, 7, 251, 193, 218, 7, 251, 193, 12, 8, 251, 193, 14, 8, 251, 193, 16, 8, 251, 193, 36, 8, 251, 193, 54, 8, 251, 193, 66, 8, 251, 193, 72, 8, 251, 193, 88, 8, 251, 193, 166, 8, 251, 193, 176, 8, 251, 193, 214, 48, 251, 193, 216, 48, 251, 193, 218, 48, 251, 193, 220, 48, 251, 193, 222, 48, 251, 193, 238, 48, 251, 193, 240, 48, 251, 193, 242, 48, 251, 193, 244, 48, 251, 193, 2, 49, 251, 193, 4, 49, 251, 193, 6, 49, 251, 193, 8, 49, 251, 193, 10, 49, 251, 193, 22, 49, 251, 193, 32, 49, 251, 193, 36, 49, 251, 193, 40, 49, 251, 193, 46, 49, 251, 193, 62, 49, 251, 193, 66, 49, 251, 193, 68, 49, 251, 193, 70, 49, 251, 193, 72, 49, 251, 193, 74, 49, 251, 193, 82, 49, 251, 193, 84, 49, 251, 193, 86, 49, 251, 193, 88, 49, 251, 193, 246, 6, 251, 193, 104, 49, 251, 193, 106, 49, 251, 193, 14, 7, 251, 193, 50, 7, 251, 193, 52, 7, 251, 193, 54, 7, 251, 193, 100, 7, 251, 193, 102, 7, 251, 193, 104, 7, 251, 193, 106, 7, 251, 193, 108, 7, 251, 193, 136, 7, 251, 193, 138, 7, 251, 193, 140, 7, 251, 193, 182, 7, 251, 193, 200, 7, 251, 193, 230, 7, 251, 193, 232, 7, 251, 193, 236, 7, 251, 193, 240, 7, 251, 193, 28, 8, 251, 193, 58, 8, 251, 193, 60, 8, 251, 193, 104, 8, 251, 193, 108, 8, 251, 193, 130, 8, 251, 193, 180, 8, 251, 193, 212, 8, 251, 193, 220, 8, 251, 193, 226, 8, 251, 193, 22, 9, 251, 193, 150, 9, 251, 193, 182, 9, 251, 193, 184, 9, 251, 193, 72, 7, 251, 193, 148, 7, 251, 193, 150, 7, 251, 193, 152, 7, 251, 193, 186, 7, 251, 193, 234, 7, 251, 193, 242, 7, 251, 193, 20, 8, 251, 193, 50, 8, 251, 193, 152, 8, 251, 193, 4, 9, 251, 193, 6, 9, 251, 193, 30, 9, 251, 193, 54, 9, 251, 193, 190, 9, 251, 193, 34, 10, 251, 193, 202, 10, 251, 193, 204, 10, 251, 193, 206, 10, 251, 193, 208, 10, 251, 193, 210, 10, 251, 193, 212, 10, 251, 193, 214, 10, 251, 193, 216, 10, 251, 193, 130, 11, 251, 193, 132, 11, 251, 193, 134, 11, 251, 193, 136, 11, 251, 193, 138, 11, 251, 193, 140, 11, 251, 193, 142, 11, 251, 193, 144, 11, 251, 193, 112, 95, 252, 193, 120, 7, 251, 193, 114, 95, 252, 193, 116, 95, 252, 193, 118, 95, 252, 193, 178, 95, 252, 193, 180, 95, 252, 193, 182, 95, 252, 193, 184, 95, 252, 193, 186, 95, 252, 193, 188, 95, 252, 193, 190, 95, 252, 193, 192, 95, 252, 193, 194, 95, 252, 193, 196, 95, 252, 193, 198, 95, 252, 193, 200, 95, 252, 193, 202, 95, 252, 193, 204, 95, 252, 193, 206, 95, 252, 193, 208, 95, 252, 193, 210, 95, 252, 193, 212, 95, 252, 193, 214, 95, 252, 193, 216, 95, 252, 193, 218, 95, 252, 193, 220, 95, 252, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 193, 162, 7, 251, 193, 164, 7, 251, 193, 166, 7, 251, 193, 208, 7, 251, 193, 210, 7, 251, 193, 220, 7, 251, 193, 222, 7, 251, 193, 224, 7, 251, 193, 226, 7, 251, 193, 238, 7, 251, 193, 244, 7, 251, 193, 246, 7, 251, 193, 250, 7, 251, 193, 254, 7, 251, 193, 2, 8, 251, 193, 4, 8, 251, 193, 6, 8, 251, 193, 10, 8, 251, 193, 22, 8, 251, 193, 24, 8, 251, 193, 26, 8, 251, 193, 34, 8, 251, 193, 40, 8, 251, 193, 42, 8, 251, 193, 44, 8, 251, 193, 52, 8, 251, 193, 56, 8, 251, 193, 62, 8, 251, 193, 68, 8, 251, 193, 70, 8, 251, 193, 76, 8, 251, 193, 78, 8, 251, 193, 188, 7, 251, 193, 116, 140, 252, 193, 118, 140, 252, 193, 120, 140, 252, 193, 126, 140, 252, 193, 128, 140, 252, 193, 130, 140, 252, 193, 132, 140, 252, 193, 140, 140, 252, 193, 142, 140, 252, 193, 148, 140, 252, 193, 154, 140, 252, 193, 158, 140, 252, 193, 160, 140, 252, 193, 162, 140, 252, 193, 164, 140, 252, 193, 182, 140, 252, 193, 184, 140, 252, 193, 186, 140, 252, 193, 188, 140, 252, 193, 222, 140, 252, 193, 224, 140, 252, 193, 226, 140, 252, 193, 228, 140, 252, 193, 230, 140, 252, 193, 232, 140, 252, 193, 234, 140, 252, 193, 236, 140, 252, 193, 238, 140, 252, 193, 240, 140, 252, 193, 242, 140, 252, 193, 244, 140, 252, 193, 26, 154, 251, 193, 228, 7, 251, 193, 38, 154, 251, 193, 106, 102, 251, 193, 40, 154, 251, 193, 54, 154, 251, 193, 62, 154, 251, 193, 64, 154, 251, 193, 66, 154, 251, 193, 68, 154, 251, 193, 72, 154, 251, 193, 80, 154, 251, 193, 84, 154, 251, 193, 86, 154, 251, 193, 88, 154, 251, 193, 96, 154, 251, 193, 98, 154, 251, 193, 100, 154, 251, 193, 108, 154, 251, 193, 110, 154, 251, 193, 112, 154, 251, 193, 114, 154, 251, 193, 122, 154, 251, 193, 126, 154, 251, 193, 128, 154, 251, 193, 130, 154, 251, 193, 132, 154, 251, 193, 136, 154, 251, 193, 138, 154, 251, 193, 140, 154, 251, 193, 142, 154, 251, 193, 146, 154, 251, 193, 80, 8, 251, 193, 102, 8, 251, 193, 116, 8, 251, 193, 118, 8, 251, 193, 120, 8, 251, 193, 122, 8, 251, 193, 124, 8, 251, 193, 126, 8, 251, 193, 138, 8, 251, 193, 140, 8, 251, 193, 142, 8, 251, 193, 154, 8, 251, 193, 156, 8, 251, 193, 158, 8, 251, 193, 160, 8, 251, 193, 162, 8, 251, 193, 168, 8, 251, 193, 170, 8, 251, 193, 172, 8, 251, 193, 174, 8, 251, 193, 216, 162, 251, 193, 182, 8, 251, 193, 184, 8, 251, 193, 186, 8, 251, 193, 188, 8, 251, 193, 192, 8, 251, 193, 194, 8, 251, 193, 198, 8, 251, 193, 202, 8, 251, 193, 204, 8, 251, 193, 206, 8, 251, 193, 214, 8, 251, 193, 178, 8, 251, 193, 190, 8, 251, 193, 200, 8, 251, 193, 248, 8, 251, 193, 20, 9, 251, 193, 44, 9, 251, 193, 46, 9, 251, 193, 48, 9, 251, 193, 70, 9, 251, 193, 72, 9, 251, 193, 92, 9, 251, 193, 94, 9, 251, 193, 96, 9, 251, 193, 116, 9, 251, 193, 140, 9, 251, 193, 142, 9, 251, 193, 144, 9, 251, 193, 146, 9, 251, 193, 148, 9, 251, 193, 178, 9, 251, 193, 180, 9, 251, 193, 216, 9, 251, 193, 218, 9, 251, 193, 220, 9, 251, 193, 222, 9, 251, 193, 252, 9, 251, 193, 254, 9, 251, 193, 30, 10, 251, 193, 32, 10, 251, 193, 40, 10, 251, 193, 42, 10, 251, 193, 56, 10, 251, 193, 222, 8, 251, 193, 240, 8, 251, 193, 242, 8, 251, 193, 8, 9, 251, 193, 10, 9, 251, 193, 12, 9, 251, 193, 14, 9, 251, 193, 16, 9, 251, 193, 32, 9, 251, 193, 34, 9, 251, 193, 36, 9, 251, 193, 38, 9, 251, 193, 56, 9, 251, 193, 58, 9, 251, 193, 60, 9, 251, 193, 62, 9, 251, 193, 64, 9, 251, 193, 66, 9, 251, 193, 68, 9, 251, 193, 86, 9, 251, 193, 88, 9, 251, 193, 90, 9, 251, 193, 100, 9, 251, 193, 210, 101, 251, 193, 102, 9, 251, 193, 104, 9, 251, 193, 106, 9, 251, 193, 108, 9, 251, 193, 110, 9, 251, 193, 112, 9, 251, 193, 118, 9, 251, 193, 120, 9, 251, 193, 232, 8, 251, 193, 234, 8, 251, 193, 236, 8, 251, 193, 238, 8, 251, 193, 250, 8, 251, 193, 252, 8, 251, 193, 254, 8, 251, 193, 2, 9, 251, 193, 24, 9, 251, 193, 26, 9, 251, 193, 28, 9, 251, 193, 50, 9, 251, 193, 52, 9, 251, 193, 74, 9, 251, 193, 76, 9, 251, 193, 78, 9, 251, 193, 80, 9, 251, 193, 82, 9, 251, 193, 84, 9, 251, 193, 98, 9, 251, 193, 152, 9, 251, 193, 154, 9, 251, 193, 186, 9, 251, 193, 188, 9, 251, 193, 4, 10, 251, 193, 52, 10, 251, 193, 68, 10, 251, 193, 70, 10, 251, 193, 74, 10, 251, 193, 76, 10, 251, 193, 78, 10, 251, 193, 80, 10, 251, 193, 244, 8, 251, 193, 246, 8, 251, 193, 40, 9, 251, 193, 138, 9, 251, 193, 174, 9, 251, 193, 250, 9, 251, 193, 48, 10, 251, 193, 136, 10, 251, 193, 214, 11, 251, 193, 216, 11, 251, 193, 218, 11, 251, 193, 220, 11, 251, 193, 226, 12, 251, 193, 228, 12, 251, 193, 230, 12, 251, 193, 232, 12, 251, 193, 234, 12, 251, 193, 236, 12, 251, 193, 238, 12, 251, 193, 240, 12, 251, 193, 242, 12, 251, 193, 40, 14, 251, 193, 42, 14, 251, 193, 44, 14, 251, 193, 46, 14, 251, 193, 48, 14, 251, 193, 50, 14, 251, 193, 132, 15, 251, 193, 134, 15, 251, 193, 136, 15, 251, 193, 138, 15, 251, 193, 140, 15, 251, 193, 122, 9, 251, 193, 124, 9, 251, 193, 126, 9, 251, 193, 128, 9, 251, 193, 130, 9, 251, 193, 132, 9, 251, 193, 134, 9, 251, 193, 136, 9, 251, 193, 156, 9, 251, 193, 158, 9, 251, 193, 160, 9, 251, 193, 162, 9, 251, 193, 164, 9, 251, 193, 166, 9, 251, 193, 168, 9, 251, 193, 170, 9, 251, 193, 162, 124, 252, 193, 22, 102, 251, 193, 172, 9, 251, 193, 192, 9, 251, 193, 194, 9, 251, 193, 196, 9, 251, 193, 198, 9, 251, 193, 200, 9, 251, 193, 202, 9, 251, 193, 186, 154, 251, 193, 204, 9, 251, 193, 206, 9, 251, 193, 208, 9, 251, 193, 20, 130, 251, 193, 224, 9, 251, 193, 226, 9, 251, 193, 210, 9, 251, 193, 232, 9, 251, 193, 234, 9, 251, 193, 236, 9, 251, 193, 238, 9, 251, 193, 240, 9, 251, 193, 242, 9, 251, 193, 244, 9, 251, 193, 246, 9, 251, 193, 248, 9, 251, 193, 194, 74, 251, 193, 6, 10, 251, 193, 8, 10, 251, 193, 10, 10, 251, 193, 12, 10, 251, 193, 14, 10, 251, 193, 16, 10, 251, 193, 18, 10, 251, 193, 20, 10, 251, 193, 140, 102, 251, 193, 22, 10, 251, 193, 142, 102, 251, 193, 24, 10, 251, 193, 26, 10, 251, 193, 28, 10, 251, 193, 160, 102, 251, 193, 44, 10, 251, 193, 64, 231, 251, 193, 200, 220, 253, 193, 46, 10, 251, 193, 54, 10, 251, 193, 62, 10, 251, 193, 230, 9, 251, 193, 36, 10, 251, 193, 38, 10, 251, 193, 50, 10, 251, 193, 58, 10, 251, 193, 88, 10, 251, 193, 138, 10, 251, 193, 140, 10, 251, 193, 142, 10, 251, 193, 236, 10, 251, 193, 238, 10, 251, 193, 222, 11, 251, 193, 224, 11, 251, 193, 226, 11, 251, 193, 228, 11, 251, 193, 230, 11, 251, 193, 232, 11, 251, 193, 244, 12, 251, 193, 246, 12, 251, 193, 248, 12, 251, 193, 250, 12, 251, 193, 252, 12, 251, 193, 254, 12, 251, 193, 2, 13, 251, 193, 4, 13, 251, 193, 56, 14, 251, 193, 58, 14, 251, 193, 60, 14, 251, 193, 62, 14, 251, 193, 64, 14, 251, 193, 148, 15, 251, 193, 150, 15, 251, 193, 2, 10, 251, 193, 148, 10, 251, 193, 252, 10, 251, 193, 254, 10, 251, 193, 244, 11, 251, 193, 246, 11, 251, 193, 248, 11, 251, 193, 250, 11, 251, 193, 252, 11, 251, 193, 254, 11, 251, 193, 2, 12, 251, 193, 32, 13, 251, 193, 34, 13, 251, 193, 36, 13, 251, 193, 38, 13, 251, 193, 40, 13, 251, 193, 42, 13, 251, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 193, 60, 10, 251, 193, 144, 10, 251, 193, 146, 10, 251, 193, 240, 10, 251, 193, 242, 10, 251, 193, 244, 10, 251, 193, 246, 10, 251, 193, 248, 10, 251, 193, 250, 10, 251, 193, 234, 11, 251, 193, 236, 11, 251, 193, 238, 11, 251, 193, 240, 11, 251, 193, 242, 11, 251, 193, 6, 13, 251, 193, 8, 13, 251, 193, 10, 13, 251, 193, 12, 13, 251, 193, 14, 13, 251, 193, 16, 13, 251, 193, 18, 13, 251, 193, 20, 13, 251, 193, 22, 13, 251, 193, 24, 13, 251, 193, 26, 13, 251, 193, 28, 13, 251, 193, 30, 13, 251, 206, 192, 0, 0, 206, 192, 0, 0, 206, 192, 0, 0, 206, 192, 0, 0, 206, 192, 0, 0, 193, 64, 10, 251, 193, 66, 10, 251, 193, 72, 10, 251, 193, 82, 10, 251, 193, 84, 10, 251, 193, 86, 10, 251, 193, 130, 10, 251, 193, 132, 10, 251, 193, 134, 10, 251, 193, 218, 10, 251, 193, 220, 10, 251, 193, 222, 10, 251, 193, 224, 10, 251, 193, 226, 10, 251, 193, 40, 42, 251, 193, 228, 10, 251, 193, 230, 10, 251, 193, 214, 47, 251, 193, 228, 154, 251, 193, 232, 10, 251, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 193, 90, 10, 251, 193, 92, 10, 251, 193, 94, 10, 251, 193, 96, 10, 251, 193, 98, 10, 251, 193, 100, 10, 251, 193, 102, 10, 251, 193, 104, 10, 251, 193, 106, 10, 251, 193, 108, 10, 251, 193, 110, 10, 251, 193, 112, 10, 251, 193, 114, 10, 251, 193, 116, 10, 251, 193, 118, 10, 251, 193, 120, 10, 251, 193, 122, 10, 251, 193, 124, 10, 251, 193, 126, 10, 251, 193, 128, 10, 251, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 0, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 32, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 193, 128, 11, 251, 193, 114, 12, 251, 193, 116, 12, 251, 193, 118, 12, 251, 193, 120, 12, 251, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 193, 146, 11, 251, 193, 122, 12, 251, 193, 124, 12, 251, 193, 126, 12, 251, 193, 128, 12, 251, 193, 130, 12, 251, 193, 132, 12, 251, 193, 134, 12, 251, 193, 136, 12, 251, 193, 184, 13, 251, 193, 186, 13, 251, 193, 188, 13, 251, 193, 190, 13, 251, 193, 192, 13, 251, 193, 194, 13, 251, 193, 196, 13, 251, 193, 198, 13, 251, 193, 200, 13, 251, 193, 202, 13, 251, 193, 204, 13, 251, 193, 206, 13, 251, 193, 208, 13, 251, 193, 210, 13, 251, 193, 254, 14, 251, 193, 2, 15, 251, 193, 4, 15, 251, 193, 6, 15, 251, 193, 8, 15, 251, 193, 10, 15, 251, 193, 12, 15, 251, 193, 14, 15, 251, 193, 16, 15, 251, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 206, 64, 0, 0, 193, 216, 27, 251, 193, 208, 11, 251, 193, 210, 11, 251, 193, 212, 11, 251, 193, 138, 12, 251, 193, 140, 12, 251, 193, 142, 12, 251, 193, 144, 12, 251, 193, 146, 12, 251, 193, 148, 12, 251, 193, 150, 12, 251, 193, 152, 12, 251, 193, 154, 12, 251, 193, 156, 12, 251, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 206, 96, 0, 0, 193, 158, 12, 251, 193, 160, 12, 251, 193, 162, 12, 251, 193, 164, 12, 251, 193, 166, 12, 251, 193, 168, 12, 251, 193, 170, 12, 251, 193, 172, 12, 251, 193, 174, 12, 251, 193, 176, 12, 251, 193, 178, 12, 251, 193, 180, 12, 251, 193, 182, 12, 251, 193, 184, 12, 251, 193, 186, 12, 251, 193, 188, 12, 251, 193, 190, 12, 251, 193, 192, 12, 251, 193, 194, 12, 251, 193, 196, 12, 251, 193, 198, 12, 251, 193, 200, 12, 251, 193, 202, 12, 251, 193, 204, 12, 251, 193, 206, 12, 251, 193, 208, 12, 251, 193, 210, 12, 251, 193, 212, 12, 251, 193, 214, 12, 251, 193, 216, 12, 251, 193, 218, 12, 251, 193, 220, 12, 251, 193, 222, 12, 251, 193, 224, 12, 251, 193, 46, 58, 251, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 128, 0, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 206, 160, 0, 0, 193, 116, 129, 251, 193, 24, 14, 251, 193, 26, 14, 251, 193, 28, 14, 251, 193, 30, 14, 251, 193, 32, 14, 251, 193, 34, 14, 251, 193, 36, 14, 251, 193, 38, 14, 251, 193, 24, 15, 251, 193, 26, 15, 251, 193, 28, 15, 251, 193, 30, 15, 251, 193, 32, 15, 251, 193, 34, 15, 251, 193, 36, 15, 251, 193, 38, 15, 251, 193, 40, 15, 251, 193, 42, 15, 251, 193, 44, 15, 251, 193, 46, 15, 251, 193, 48, 15, 251, 193, 50, 15, 251, 193, 52, 15, 251, 193, 54, 15, 251, 193, 56, 15, 251, 193, 58, 15, 251, 193, 60, 15, 251, 206, 192, 0, 0, 206, 192, 0, 0, 206, 192, 0, 0, 206, 192, 0, 0, 206, 192, 0, 0, 193, 172, 15, 251, 193, 174, 15, 251, 193, 176, 15, 251, 193, 178, 15, 251, 193, 180, 15, 251, 193, 182, 15, 251, 193, 184, 15, 251, 193, 186, 15, 251, 193, 188, 15, 251, 193, 190, 15, 251, 193, 192, 15, 251, 193, 194, 15, 251, 193, 196, 15, 251, 193, 198, 15, 251, 193, 200, 15, 251, 193, 202, 15, 251, 193, 204, 15, 251, 193, 206, 15, 251, 193, 208, 15, 251, 193, 210, 15, 251, 193, 212, 15, 251, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 206, 224, 0, 0, 193, 214, 15, 251, 193, 216, 15, 251, 193, 218, 15, 251, 193, 220, 15, 251, 193, 222, 15, 251, 193, 224, 15, 251, 193, 226, 15, 251, 193, 228, 15, 251, 193, 230, 15, 251, 193, 232, 15, 251, 193, 136, 17, 251, 193, 138, 17, 251, 193, 140, 17, 251, 193, 142, 17, 251, 193, 144, 17, 251, 193, 146, 17, 251, 193, 148, 17, 251, 193, 150, 17, 251, 193, 250, 18, 251, 193, 252, 18, 251, 193, 254, 18, 251, 193, 2, 19, 251, 193, 4, 19, 251, 193, 6, 19, 251, 193, 8, 19, 251, 193, 10, 19, 251, 193, 12, 19, 251, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 0, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 193, 18, 15, 251, 193, 20, 15, 251, 193, 22, 15, 251, 193, 144, 16, 251, 193, 26, 18, 251, 193, 146, 16, 251, 193, 148, 16, 251, 193, 150, 16, 251, 193, 152, 16, 251, 193, 154, 16, 251, 193, 156, 16, 251, 193, 158, 16, 251, 193, 160, 16, 251, 193, 162, 16, 251, 193, 28, 18, 251, 193, 30, 18, 251, 193, 32, 18, 251, 193, 34, 18, 251, 193, 36, 18, 251, 193, 104, 19, 251, 193, 106, 19, 251, 193, 108, 19, 251, 193, 110, 19, 251, 193, 112, 19, 251, 193, 114, 19, 251, 193, 116, 19, 251, 193, 118, 19, 251, 193, 156, 20, 251, 193, 158, 20, 251, 193, 160, 20, 251, 193, 162, 20, 251, 193, 164, 20, 251, 193, 62, 15, 251, 193, 64, 15, 251, 193, 66, 15, 251, 193, 68, 15, 251, 193, 70, 15, 251, 193, 72, 15, 251, 193, 74, 15, 251, 193, 76, 15, 251, 193, 78, 15, 251, 193, 80, 15, 251, 193, 82, 15, 251, 193, 84, 15, 251, 193, 86, 15, 251, 193, 88, 15, 251, 193, 90, 15, 251, 193, 92, 15, 251, 193, 94, 15, 251, 193, 96, 15, 251, 193, 98, 15, 251, 193, 100, 15, 251, 193, 102, 15, 251, 193, 104, 15, 251, 193, 106, 15, 251, 193, 108, 15, 251, 193, 110, 15, 251, 193, 112, 15, 251, 193, 114, 15, 251, 193, 116, 15, 251, 193, 118, 15, 251, 193, 120, 15, 251, 193, 122, 15, 251, 193, 124, 15, 251, 193, 126, 15, 251, 193, 128, 15, 251, 193, 130, 15, 251, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 193, 142, 15, 251, 193, 144, 15, 251, 193, 146, 15, 251, 193, 48, 17, 251, 193, 50, 17, 251, 193, 52, 17, 251, 193, 54, 17, 251, 193, 56, 17, 251, 193, 58, 17, 251, 193, 60, 17, 251, 193, 62, 17, 251, 193, 174, 18, 251, 193, 176, 18, 251, 193, 178, 18, 251, 193, 180, 18, 251, 193, 182, 18, 251, 193, 184, 18, 251, 193, 186, 18, 251, 193, 188, 18, 251, 193, 190, 18, 251, 193, 192, 18, 251, 193, 194, 18, 251, 193, 236, 19, 251, 193, 238, 19, 251, 193, 240, 19, 251, 193, 242, 19, 251, 193, 244, 19, 251, 193, 246, 19, 251, 193, 248, 19, 251, 193, 250, 19, 251, 193, 252, 19, 251, 193, 254, 19, 251, 193, 152, 15, 251, 193, 154, 15, 251, 193, 156, 15, 251, 193, 158, 15, 251, 193, 160, 15, 251, 193, 162, 15, 251, 193, 164, 15, 251, 193, 166, 15, 251, 193, 168, 15, 251, 193, 170, 15, 251, 193, 66, 17, 251, 193, 68, 17, 251, 193, 70, 17, 251, 193, 72, 17, 251, 193, 74, 17, 251, 193, 76, 17, 251, 193, 78, 17, 251, 193, 80, 17, 251, 193, 82, 17, 251, 193, 84, 17, 251, 193, 86, 17, 251, 193, 88, 17, 251, 193, 90, 17, 251, 193, 92, 17, 251, 193, 94, 17, 251, 193, 96, 17, 251, 193, 98, 17, 251, 193, 100, 17, 251, 193, 102, 17, 251, 193, 198, 18, 251, 193, 200, 18, 251, 193, 202, 18, 251, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 32, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 64, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 206, 160, 1, 0, 193, 174, 157, 253, 193, 122, 18, 251, 193, 124, 18, 251, 193, 126, 18, 251, 193, 128, 18, 251, 193, 130, 18, 251, 193, 132, 18, 251, 193, 134, 18, 251, 193, 136, 18, 251, 193, 138, 18, 251, 193, 140, 18, 251, 193, 142, 18, 251, 193, 144, 18, 251, 193, 146, 18, 251, 193, 46, 17, 251, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 206, 96, 1, 0, 193, 212, 18, 251, 193, 214, 18, 251, 193, 216, 18, 251, 193, 218, 18, 251, 193, 220, 18, 251, 193, 222, 18, 251, 193, 224, 18, 251, 193, 226, 18, 251, 193, 228, 18, 251, 193, 230, 18, 251, 193, 232, 18, 251, 193, 234, 18, 251, 193, 236, 18, 251, 193, 238, 18, 251, 193, 240, 18, 251, 193, 242, 18, 251, 193, 244, 18, 251, 193, 246, 18, 251, 193, 248, 18, 251, 193, 28, 20, 251, 193, 30, 20, 251, 193, 32, 20, 251, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 206, 128, 1, 0, 193, 148, 18, 251, 193, 150, 18, 251, 193, 152, 18, 251, 193, 154, 18, 251, 193, 156, 18, 251, 193, 158, 18, 251, 193, 160, 18, 251, 193, 162, 18, 251, 193, 164, 18, 251, 193, 166, 18, 251, 193, 168, 18, 251, 193, 170, 18, 251, 193, 172, 18, 251, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 193, 204, 18, 251, 193, 206, 18, 251, 193, 208, 18, 251, 193, 210, 18, 251, 193, 2, 20, 251, 193, 4, 20, 251, 193, 6, 20, 251, 193, 8, 20, 251, 193, 10, 20, 251, 193, 12, 20, 251, 193, 14, 20, 251, 193, 16, 20, 251, 193, 18, 20, 251, 193, 20, 20, 251, 193, 22, 20, 251, 193, 24, 20, 251, 193, 26, 20, 251, 193, 58, 21, 251, 193, 60, 21, 251, 193, 62, 21, 251, 193, 64, 21, 251, 193, 66, 21, 251, 193, 68, 21, 251, 193, 118, 22, 251, 193, 120, 22, 251, 193, 122, 22, 251, 193, 124, 22, 251, 193, 126, 22, 251, 193, 128, 22, 251, 193, 130, 22, 251, 193, 132, 22, 251, 193, 134, 22, 251, 193, 14, 19, 251, 193, 16, 19, 251, 193, 66, 20, 251, 193, 68, 20, 251, 193, 70, 20, 251, 193, 72, 20, 251, 193, 74, 20, 251, 193, 76, 20, 251, 193, 78, 20, 251, 193, 80, 20, 251, 193, 104, 21, 251, 193, 106, 21, 251, 193, 108, 21, 251, 193, 170, 22, 251, 193, 172, 22, 251, 193, 174, 22, 251, 193, 176, 22, 251, 193, 178, 22, 251, 193, 180, 22, 251, 193, 142, 23, 251, 193, 144, 23, 251, 193, 146, 23, 251, 193, 148, 23, 251, 193, 150, 23, 251, 193, 152, 23, 251, 193, 154, 23, 251, 193, 40, 24, 251, 193, 42, 24, 251, 193, 44, 24, 251, 193, 46, 24, 251, 193, 160, 24, 251, 193, 162, 24, 251, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 192, 1, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 206, 224, 1, 0, 193, 36, 28, 251, 193, 170, 19, 251, 193, 172, 19, 251, 193, 174, 19, 251, 193, 176, 19, 251, 193, 178, 19, 251, 193, 180, 19, 251, 193, 182, 19, 251, 193, 184, 19, 251, 193, 186, 19, 251, 193, 188, 19, 251, 193, 190, 19, 251, 193, 192, 19, 251, 193, 194, 19, 251, 193, 196, 19, 251, 193, 198, 19, 251, 193, 200, 19, 251, 193, 202, 19, 251, 193, 204, 19, 251, 193, 206, 19, 251, 193, 208, 19, 251, 193, 210, 19, 251, 193, 212, 19, 251, 193, 214, 19, 251, 193, 50, 101, 251, 193, 226, 107, 252, 193, 216, 19, 251, 193, 218, 19, 251, 193, 220, 19, 251, 193, 222, 19, 251, 193, 224, 19, 251, 193, 226, 19, 251, 193, 228, 19, 251, 193, 230, 19, 251, 193, 232, 19, 251, 193, 234, 19, 251, 193, 206, 115, 252, 193, 170, 20, 251, 193, 172, 20, 251, 193, 174, 20, 251, 193, 176, 20, 251, 193, 178, 20, 251, 193, 180, 20, 251, 193, 182, 20, 251, 193, 184, 20, 251, 193, 186, 20, 251, 193, 188, 20, 251, 193, 190, 20, 251, 193, 192, 20, 251, 193, 194, 20, 251, 193, 196, 20, 251, 193, 198, 20, 251, 193, 200, 20, 251, 193, 202, 20, 251, 193, 204, 20, 251, 193, 206, 20, 251, 193, 208, 20, 251, 193, 210, 20, 251, 193, 34, 20, 251, 193, 36, 20, 251, 193, 38, 20, 251, 193, 40, 20, 251, 193, 42, 20, 251, 193, 44, 20, 251, 193, 46, 20, 251, 193, 48, 20, 251, 193, 50, 20, 251, 193, 52, 20, 251, 193, 54, 20, 251, 193, 56, 20, 251, 193, 58, 20, 251, 193, 60, 20, 251, 193, 62, 20, 251, 193, 64, 20, 251, 193, 70, 21, 251, 193, 72, 21, 251, 193, 74, 21, 251, 193, 76, 21, 251, 193, 78, 21, 251, 193, 80, 21, 251, 193, 82, 21, 251, 193, 84, 21, 251, 193, 86, 21, 251, 193, 88, 21, 251, 193, 90, 21, 251, 193, 92, 21, 251, 193, 94, 21, 251, 193, 96, 21, 251, 193, 98, 21, 251, 193, 100, 21, 251, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 0, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 193, 166, 20, 251, 193, 168, 20, 251, 193, 188, 21, 251, 193, 190, 21, 251, 193, 192, 21, 251, 193, 194, 21, 251, 193, 196, 21, 251, 193, 198, 21, 251, 193, 200, 21, 251, 193, 202, 21, 251, 193, 204, 21, 251, 193, 206, 21, 251, 193, 208, 21, 251, 193, 210, 21, 251, 193, 212, 21, 251, 193, 214, 21, 251, 193, 216, 21, 251, 193, 236, 22, 251, 193, 238, 22, 251, 193, 240, 22, 251, 193, 242, 22, 251, 193, 244, 22, 251, 193, 246, 22, 251, 193, 248, 22, 251, 193, 250, 22, 251, 193, 190, 23, 251, 193, 192, 23, 251, 193, 194, 23, 251, 193, 196, 23, 251, 193, 74, 24, 251, 193, 76, 24, 251, 193, 180, 24, 251, 193, 212, 20, 251, 193, 214, 20, 251, 193, 216, 20, 251, 193, 218, 20, 251, 193, 220, 20, 251, 193, 222, 20, 251, 193, 224, 20, 251, 193, 226, 20, 251, 193, 228, 20, 251, 193, 230, 20, 251, 193, 232, 20, 251, 193, 234, 20, 251, 193, 236, 20, 251, 193, 238, 20, 251, 193, 240, 20, 251, 193, 242, 20, 251, 193, 244, 20, 251, 193, 246, 20, 251, 193, 248, 20, 251, 193, 250, 20, 251, 193, 252, 20, 251, 193, 254, 20, 251, 193, 2, 21, 251, 193, 4, 21, 251, 193, 6, 21, 251, 193, 8, 21, 251, 193, 10, 21, 251, 193, 12, 21, 251, 193, 14, 21, 251, 193, 16, 21, 251, 193, 18, 21, 251, 193, 20, 21, 251, 193, 22, 21, 251, 193, 24, 21, 251, 193, 140, 123, 252, 193, 26, 21, 251, 193, 28, 21, 251, 193, 218, 21, 251, 193, 30, 21, 251, 193, 32, 21, 251, 193, 34, 21, 251, 193, 36, 21, 251, 193, 38, 21, 251, 193, 40, 21, 251, 193, 42, 21, 251, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 193, 62, 22, 251, 193, 64, 22, 251, 193, 66, 22, 251, 193, 68, 22, 251, 193, 70, 22, 251, 193, 72, 22, 251, 193, 74, 22, 251, 193, 76, 22, 251, 193, 78, 22, 251, 193, 80, 22, 251, 193, 82, 22, 251, 193, 84, 22, 251, 193, 86, 22, 251, 193, 88, 22, 251, 193, 90, 22, 251, 193, 92, 22, 251, 193, 94, 22, 251, 193, 96, 22, 251, 193, 98, 22, 251, 193, 100, 22, 251, 193, 102, 22, 251, 193, 44, 21, 251, 193, 210, 189, 251, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 193, 46, 21, 251, 193, 48, 21, 251, 193, 50, 21, 251, 193, 52, 21, 251, 193, 54, 21, 251, 193, 106, 22, 251, 193, 108, 22, 251, 193, 110, 22, 251, 193, 112, 22, 251, 193, 114, 22, 251, 193, 116, 22, 251, 193, 88, 23, 251, 193, 90, 23, 251, 193, 92, 23, 251, 193, 94, 23, 251, 193, 96, 23, 251, 193, 10, 24, 251, 193, 12, 24, 251, 193, 14, 24, 251, 193, 16, 24, 251, 193, 130, 24, 251, 193, 132, 24, 251, 193, 134, 24, 251, 193, 232, 24, 251, 193, 234, 24, 251, 193, 236, 24, 251, 193, 76, 25, 251, 193, 78, 25, 251, 193, 80, 25, 251, 193, 132, 25, 251, 193, 180, 25, 251, 193, 128, 26, 251, 193, 102, 21, 251, 193, 142, 22, 251, 193, 144, 22, 251, 193, 146, 22, 251, 193, 148, 22, 251, 193, 150, 22, 251, 193, 152, 22, 251, 193, 154, 22, 251, 193, 156, 22, 251, 193, 158, 22, 251, 193, 160, 22, 251, 193, 162, 22, 251, 193, 164, 22, 251, 193, 166, 22, 251, 193, 168, 22, 251, 193, 118, 23, 251, 193, 120, 23, 251, 193, 122, 23, 251, 193, 124, 23, 251, 193, 126, 23, 251, 193, 128, 23, 251, 193, 130, 23, 251, 193, 132, 23, 251, 193, 134, 23, 251, 193, 136, 23, 251, 193, 138, 23, 251, 193, 140, 23, 251, 193, 28, 24, 251, 193, 30, 24, 251, 193, 32, 24, 251, 193, 34, 24, 251, 193, 36, 24, 251, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 206, 32, 2, 0, 193, 182, 22, 251, 193, 182, 21, 251, 193, 184, 21, 251, 193, 184, 22, 251, 193, 186, 22, 251, 193, 186, 21, 251, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 206, 64, 2, 0, 193, 122, 140, 252, 193, 52, 22, 251, 193, 54, 22, 251, 193, 250, 65, 252, 193, 96, 221, 253, 193, 56, 22, 251, 193, 58, 22, 251, 193, 60, 22, 251, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 206, 128, 2, 0, 193, 104, 22, 251, 193, 48, 23, 251, 193, 198, 23, 251, 193, 196, 74, 251, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 193, 136, 22, 251, 193, 138, 22, 251, 193, 140, 22, 251, 193, 98, 23, 251, 193, 100, 23, 251, 193, 102, 23, 251, 193, 104, 23, 251, 193, 106, 23, 251, 193, 108, 23, 251, 193, 110, 23, 251, 193, 112, 23, 251, 193, 114, 23, 251, 193, 116, 23, 251, 193, 18, 24, 251, 193, 20, 24, 251, 193, 22, 24, 251, 193, 24, 24, 251, 193, 26, 24, 251, 193, 136, 24, 251, 193, 138, 24, 251, 193, 140, 24, 251, 193, 142, 24, 251, 193, 144, 24, 251, 193, 146, 24, 251, 193, 148, 24, 251, 206, 224, 2, 0, 206, 224, 2, 0, 206, 224, 2, 0, 206, 224, 2, 0, 206, 224, 2, 0, 206, 224, 2, 0, 206, 224, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 206, 96, 2, 0, 193, 156, 23, 251, 193, 158, 23, 251, 193, 160, 23, 251, 193, 162, 23, 251, 193, 164, 23, 251, 193, 166, 23, 251, 193, 168, 23, 251, 193, 170, 23, 251, 193, 172, 23, 251, 193, 174, 23, 251, 193, 176, 23, 251, 193, 178, 23, 251, 193, 180, 23, 251, 193, 182, 23, 251, 193, 184, 23, 251, 193, 48, 24, 251, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 206, 160, 2, 0, 193, 198, 74, 251, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 193, 50, 24, 251, 193, 52, 24, 251, 193, 54, 24, 251, 193, 56, 24, 251, 193, 58, 24, 251, 193, 60, 24, 251, 193, 62, 24, 251, 193, 64, 24, 251, 193, 66, 24, 251, 193, 68, 24, 251, 193, 70, 24, 251, 193, 186, 23, 251, 193, 72, 24, 251, 193, 168, 24, 251, 193, 170, 24, 251, 193, 172, 24, 251, 193, 174, 24, 251, 193, 176, 24, 251, 193, 178, 24, 251, 193, 30, 25, 251, 193, 32, 25, 251, 193, 34, 25, 251, 193, 102, 25, 251, 193, 144, 25, 251, 193, 146, 25, 251, 193, 148, 25, 251, 193, 150, 25, 251, 193, 194, 25, 251, 193, 196, 25, 251, 193, 212, 25, 251, 193, 240, 25, 251, 193, 4, 26, 251, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 206, 192, 2, 0, 193, 78, 24, 251, 193, 80, 24, 251, 193, 82, 24, 251, 193, 84, 24, 251, 193, 86, 24, 251, 193, 88, 24, 251, 193, 90, 24, 251, 193, 92, 24, 251, 193, 94, 24, 251, 193, 96, 24, 251, 193, 98, 24, 251, 193, 100, 24, 251, 193, 102, 24, 251, 193, 104, 24, 251, 193, 106, 24, 251, 193, 108, 24, 251, 193, 110, 24, 251, 193, 112, 24, 251, 193, 114, 24, 251, 193, 116, 24, 251, 193, 166, 36, 253, 193, 118, 24, 251, 193, 120, 24, 251, 193, 122, 24, 251, 193, 38, 24, 251, 193, 150, 24, 251, 193, 152, 24, 251, 193, 154, 24, 251, 193, 156, 24, 251, 193, 158, 24, 251, 193, 6, 25, 251, 193, 8, 25, 251, 193, 10, 25, 251, 193, 12, 25, 251, 193, 14, 25, 251, 193, 16, 25, 251, 193, 18, 25, 251, 193, 92, 25, 251, 193, 94, 25, 251, 193, 96, 25, 251, 193, 140, 25, 251, 193, 142, 25, 251, 193, 210, 25, 251, 193, 234, 25, 251, 193, 236, 25, 251, 193, 246, 25, 251, 193, 16, 26, 251, 193, 18, 26, 251, 193, 20, 26, 251, 193, 28, 26, 251, 193, 30, 26, 251, 193, 32, 26, 251, 193, 34, 26, 251, 193, 36, 26, 251, 193, 90, 26, 251, 193, 92, 26, 251, 193, 124, 24, 251, 193, 126, 24, 251, 193, 128, 24, 251, 193, 184, 24, 251, 193, 186, 24, 251, 193, 188, 24, 251, 193, 190, 24, 251, 193, 192, 24, 251, 193, 194, 24, 251, 193, 196, 24, 251, 193, 198, 24, 251, 193, 200, 24, 251, 193, 202, 24, 251, 193, 204, 24, 251, 193, 206, 24, 251, 193, 208, 24, 251, 193, 210, 24, 251, 193, 212, 24, 251, 193, 214, 24, 251, 193, 216, 24, 251, 193, 218, 24, 251, 193, 220, 24, 251, 193, 222, 24, 251, 193, 224, 24, 251, 193, 226, 24, 251, 193, 228, 24, 251, 193, 230, 24, 251, 193, 42, 25, 251, 193, 44, 25, 251, 193, 46, 25, 251, 193, 48, 25, 251, 193, 50, 25, 251, 193, 164, 24, 251, 193, 166, 24, 251, 193, 20, 25, 251, 193, 22, 25, 251, 193, 24, 25, 251, 193, 26, 25, 251, 193, 28, 25, 251, 193, 98, 25, 251, 193, 100, 25, 251, 193, 238, 25, 251, 193, 38, 26, 251, 193, 96, 26, 251, 193, 140, 26, 251, 193, 218, 26, 251, 193, 250, 26, 251, 193, 14, 27, 251, 193, 16, 27, 251, 193, 50, 27, 251, 193, 52, 27, 251, 193, 96, 27, 251, 193, 98, 27, 251, 193, 156, 27, 251, 193, 226, 27, 251, 193, 254, 27, 251, 193, 2, 28, 251, 193, 4, 28, 251, 193, 16, 28, 251, 193, 18, 28, 251, 193, 34, 28, 251, 193, 126, 28, 251, 193, 128, 28, 251, 193, 130, 28, 251, 193, 182, 24, 251, 193, 36, 25, 251, 193, 38, 25, 251, 193, 40, 25, 251, 193, 104, 25, 251, 193, 106, 25, 251, 193, 108, 25, 251, 193, 152, 25, 251, 193, 154, 25, 251, 193, 214, 25, 251, 193, 250, 25, 251, 193, 52, 26, 251, 193, 108, 26, 251, 193, 168, 26, 251, 193, 196, 26, 251, 193, 126, 27, 251, 193, 202, 27, 251, 193, 204, 27, 251, 193, 214, 27, 251, 193, 228, 27, 251, 193, 230, 27, 251, 193, 134, 28, 251, 193, 136, 28, 251, 193, 138, 28, 251, 193, 182, 28, 251, 193, 98, 29, 251, 193, 202, 29, 251, 193, 12, 30, 251, 193, 14, 30, 251, 193, 24, 30, 251, 193, 38, 30, 251, 193, 70, 30, 251, 206, 224, 2, 0, 206, 224, 2, 0, 206, 224, 2, 0, 206, 224, 2, 0, 193, 82, 25, 251, 193, 84, 25, 251, 193, 86, 25, 251, 193, 88, 25, 251, 193, 90, 25, 251, 193, 134, 25, 251, 193, 136, 25, 251, 193, 138, 25, 251, 193, 182, 25, 251, 193, 184, 25, 251, 193, 186, 25, 251, 193, 188, 25, 251, 193, 190, 25, 251, 193, 192, 25, 251, 193, 228, 25, 251, 193, 230, 25, 251, 193, 232, 25, 251, 193, 244, 25, 251, 193, 248, 25, 251, 193, 62, 26, 251, 193, 130, 26, 251, 193, 160, 26, 251, 193, 212, 26, 251, 193, 214, 26, 251, 193, 236, 26, 251, 193, 238, 26, 251, 193, 8, 27, 251, 193, 38, 27, 251, 193, 52, 25, 251, 193, 54, 25, 251, 193, 8, 66, 252, 193, 56, 25, 251, 193, 58, 25, 251, 193, 60, 25, 251, 193, 62, 25, 251, 193, 64, 25, 251, 193, 66, 25, 251, 193, 68, 25, 251, 193, 70, 25, 251, 193, 72, 25, 251, 193, 74, 25, 251, 193, 110, 25, 251, 193, 112, 25, 251, 193, 114, 25, 251, 193, 116, 25, 251, 193, 118, 25, 251, 193, 120, 25, 251, 193, 122, 25, 251, 193, 124, 25, 251, 193, 126, 25, 251, 193, 128, 25, 251, 193, 156, 25, 251, 193, 158, 25, 251, 193, 160, 25, 251, 193, 162, 25, 251, 193, 164, 25, 251, 193, 166, 25, 251, 193, 168, 25, 251, 193, 130, 25, 251, 193, 170, 25, 251, 193, 172, 25, 251, 193, 174, 25, 251, 193, 176, 25, 251, 193, 178, 25, 251, 193, 198, 25, 251, 193, 200, 25, 251, 193, 202, 25, 251, 193, 204, 25, 251, 193, 206, 25, 251, 193, 208, 25, 251, 193, 216, 25, 251, 193, 218, 25, 251, 193, 220, 25, 251, 193, 222, 25, 251, 193, 224, 25, 251, 193, 226, 25, 251, 193, 242, 25, 251, 193, 78, 233, 251, 193, 252, 25, 251, 193, 254, 25, 251, 193, 170, 14, 253, 193, 2, 26, 251, 193, 14, 26, 251, 193, 24, 26, 251, 193, 26, 26, 251, 193, 54, 26, 251, 193, 56, 26, 251, 193, 58, 26, 251, 193, 60, 26, 251, 193, 78, 26, 251, 193, 80, 26, 251, 193, 82, 26, 251, 193, 6, 26, 251, 193, 8, 26, 251, 193, 10, 26, 251, 193, 12, 26, 251, 193, 22, 26, 251, 193, 40, 26, 251, 193, 42, 26, 251, 193, 44, 26, 251, 193, 46, 26, 251, 193, 48, 26, 251, 193, 50, 26, 251, 193, 64, 26, 251, 193, 66, 26, 251, 193, 68, 26, 251, 193, 70, 26, 251, 193, 72, 26, 251, 193, 74, 26, 251, 193, 76, 26, 251, 193, 98, 26, 251, 193, 100, 26, 251, 193, 102, 26, 251, 193, 104, 26, 251, 193, 106, 26, 251, 193, 142, 26, 251, 193, 144, 26, 251, 193, 146, 26, 251, 193, 164, 26, 251, 193, 166, 26, 251, 193, 190, 26, 251, 193, 192, 26, 251, 193, 194, 26, 251, 193, 220, 26, 251, 193, 84, 26, 251, 193, 86, 26, 251, 193, 88, 26, 251, 193, 110, 26, 251, 193, 112, 26, 251, 193, 114, 26, 251, 193, 116, 26, 251, 193, 118, 26, 251, 193, 120, 26, 251, 193, 122, 26, 251, 193, 124, 26, 251, 193, 126, 26, 251, 193, 148, 26, 251, 193, 150, 26, 251, 193, 152, 26, 251, 193, 154, 26, 251, 193, 170, 26, 251, 193, 172, 26, 251, 193, 174, 26, 251, 193, 176, 26, 251, 193, 178, 26, 251, 193, 180, 26, 251, 193, 182, 26, 251, 193, 198, 26, 251, 193, 200, 26, 251, 193, 202, 26, 251, 193, 204, 26, 251, 193, 206, 26, 251, 193, 208, 26, 251, 193, 210, 26, 251, 193, 224, 26, 251, 193, 226, 26, 251, 193, 94, 26, 251, 193, 132, 26, 251, 193, 134, 26, 251, 193, 136, 26, 251, 193, 138, 26, 251, 193, 162, 26, 251, 193, 186, 26, 251, 193, 188, 26, 251, 193, 216, 26, 251, 193, 240, 26, 251, 193, 242, 26, 251, 193, 244, 26, 251, 193, 246, 26, 251, 193, 248, 26, 251, 193, 10, 27, 251, 193, 12, 27, 251, 193, 42, 27, 251, 193, 44, 27, 251, 193, 46, 27, 251, 193, 48, 27, 251, 193, 74, 27, 251, 193, 94, 27, 251, 193, 120, 27, 251, 193, 122, 27, 251, 193, 124, 27, 251, 193, 144, 27, 251, 193, 146, 27, 251, 193, 208, 27, 251, 193, 242, 27, 251, 193, 244, 27, 251, 193, 246, 27, 251, 193, 252, 27, 251, 193, 156, 26, 251, 193, 184, 26, 251, 193, 6, 27, 251, 193, 68, 27, 251, 193, 70, 27, 251, 193, 72, 27, 251, 193, 152, 27, 251, 193, 28, 28, 251, 193, 116, 28, 251, 193, 8, 29, 251, 193, 36, 29, 251, 193, 54, 29, 251, 193, 58, 29, 251, 193, 74, 29, 251, 193, 198, 29, 251, 193, 132, 30, 251, 193, 134, 30, 251, 193, 150, 30, 251, 193, 202, 30, 251, 193, 224, 30, 251, 193, 248, 30, 251, 193, 16, 31, 251, 193, 92, 31, 251, 193, 124, 31, 251, 193, 126, 31, 251, 193, 128, 31, 251, 193, 130, 31, 251, 193, 184, 31, 251, 193, 186, 31, 251, 193, 188, 31, 251, 193, 242, 31, 251, 193, 52, 32, 251, 193, 222, 26, 251, 193, 252, 26, 251, 193, 18, 27, 251, 193, 76, 27, 251, 193, 164, 27, 251, 193, 188, 27, 251, 193, 190, 27, 251, 193, 194, 27, 251, 193, 210, 27, 251, 193, 248, 27, 251, 193, 6, 28, 251, 193, 74, 28, 251, 193, 78, 28, 251, 193, 80, 28, 251, 193, 82, 28, 251, 193, 84, 28, 251, 193, 94, 28, 251, 193, 102, 28, 251, 193, 104, 28, 251, 193, 106, 28, 251, 193, 108, 28, 251, 193, 132, 28, 251, 193, 154, 28, 251, 193, 156, 28, 251, 193, 158, 28, 251, 193, 178, 28, 251, 193, 204, 28, 251, 193, 180, 28, 251, 193, 206, 28, 251, 193, 226, 28, 251, 193, 18, 29, 251, 193, 20, 29, 251, 193, 228, 26, 251, 193, 230, 26, 251, 193, 232, 26, 251, 193, 234, 26, 251, 193, 254, 26, 251, 193, 2, 27, 251, 193, 4, 27, 251, 193, 20, 27, 251, 193, 22, 27, 251, 193, 24, 27, 251, 193, 26, 27, 251, 193, 28, 27, 251, 193, 30, 27, 251, 193, 32, 27, 251, 193, 34, 27, 251, 193, 36, 27, 251, 193, 54, 27, 251, 193, 56, 27, 251, 193, 58, 27, 251, 193, 78, 27, 251, 193, 60, 27, 251, 193, 62, 27, 251, 193, 64, 27, 251, 193, 66, 27, 251, 193, 80, 27, 251, 193, 82, 27, 251, 193, 84, 27, 251, 193, 86, 27, 251, 193, 88, 27, 251, 193, 90, 27, 251, 193, 92, 27, 251, 193, 100, 27, 251, 193, 40, 27, 251, 193, 116, 27, 251, 193, 118, 27, 251, 193, 136, 27, 251, 193, 138, 27, 251, 193, 140, 27, 251, 193, 142, 27, 251, 193, 154, 27, 251, 193, 174, 27, 251, 193, 30, 28, 251, 193, 42, 28, 251, 193, 118, 28, 251, 193, 120, 28, 251, 193, 166, 28, 251, 193, 214, 28, 251, 193, 216, 28, 251, 193, 232, 28, 251, 193, 10, 29, 251, 193, 38, 29, 251, 193, 84, 29, 251, 193, 92, 29, 251, 193, 94, 29, 251, 193, 102, 29, 251, 193, 44, 30, 251, 193, 136, 30, 251, 193, 170, 30, 251, 193, 172, 30, 251, 193, 20, 31, 251, 193, 42, 31, 251, 193, 50, 31, 251, 193, 66, 31, 251, 193, 94, 31, 251, 193, 102, 27, 251, 193, 104, 27, 251, 193, 106, 27, 251, 193, 108, 27, 251, 193, 110, 27, 251, 193, 112, 27, 251, 193, 128, 27, 251, 193, 114, 27, 251, 193, 130, 27, 251, 193, 132, 27, 251, 193, 134, 27, 251, 193, 148, 27, 251, 193, 116, 159, 251, 193, 158, 27, 251, 193, 150, 27, 251, 193, 160, 27, 251, 193, 162, 27, 251, 193, 166, 27, 251, 193, 168, 27, 251, 193, 170, 27, 251, 193, 172, 27, 251, 193, 176, 27, 251, 193, 178, 27, 251, 193, 180, 27, 251, 193, 182, 27, 251, 193, 184, 27, 251, 193, 186, 27, 251, 193, 192, 27, 251, 193, 196, 27, 251, 193, 198, 27, 251, 193, 200, 27, 251, 193, 206, 27, 251, 193, 106, 21, 252, 193, 108, 21, 252, 193, 110, 21, 252, 193, 130, 21, 252, 193, 132, 21, 252, 193, 134, 21, 252, 193, 136, 21, 252, 193, 138, 21, 252, 193, 140, 21, 252, 193, 142, 21, 252, 193, 152, 21, 252, 193, 154, 21, 252, 193, 164, 21, 252, 193, 166, 21, 252, 193, 168, 21, 252, 193, 170, 21, 252, 193, 172, 21, 252, 193, 182, 21, 252, 193, 192, 21, 252, 193, 194, 21, 252, 193, 230, 21, 252, 193, 232, 21, 252, 193, 242, 21, 252, 193, 244, 21, 252, 193, 246, 21, 252, 193, 248, 21, 252, 193, 234, 21, 252, 193, 6, 22, 252, 193, 8, 22, 252, 193, 212, 27, 251, 193, 10, 22, 252, 193, 12, 22, 252, 193, 218, 27, 251, 193, 220, 27, 251, 193, 222, 27, 251, 193, 224, 27, 251, 193, 232, 27, 251, 193, 234, 27, 251, 193, 236, 27, 251, 193, 238, 27, 251, 193, 240, 27, 251, 193, 250, 27, 251, 193, 8, 28, 251, 193, 10, 28, 251, 193, 12, 28, 251, 193, 20, 28, 251, 193, 22, 28, 251, 193, 24, 28, 251, 193, 26, 28, 251, 193, 32, 28, 251, 193, 38, 28, 251, 193, 40, 28, 251, 193, 44, 28, 251, 193, 46, 28, 251, 193, 48, 28, 251, 193, 50, 28, 251, 193, 52, 28, 251, 193, 54, 28, 251, 193, 56, 28, 251, 193, 58, 28, 251, 193, 60, 28, 251, 193, 64, 28, 251, 193, 68, 28, 251, 193, 70, 28, 251, 193, 14, 28, 251, 193, 62, 28, 251, 193, 66, 28, 251, 193, 72, 28, 251, 193, 100, 28, 251, 193, 122, 28, 251, 193, 124, 28, 251, 193, 152, 28, 251, 193, 168, 28, 251, 193, 170, 28, 251, 193, 172, 28, 251, 193, 174, 28, 251, 193, 196, 28, 251, 193, 198, 28, 251, 193, 200, 28, 251, 193, 218, 28, 251, 193, 220, 28, 251, 193, 234, 28, 251, 193, 236, 28, 251, 193, 238, 28, 251, 193, 12, 29, 251, 193, 14, 29, 251, 193, 16, 29, 251, 193, 40, 29, 251, 193, 42, 29, 251, 193, 56, 29, 251, 193, 60, 29, 251, 193, 62, 29, 251, 193, 64, 29, 251, 193, 96, 29, 251, 193, 100, 29, 251, 193, 130, 29, 251, 193, 76, 28, 251, 193, 86, 28, 251, 193, 88, 28, 251, 193, 90, 28, 251, 193, 92, 28, 251, 193, 96, 28, 251, 193, 98, 28, 251, 193, 180, 78, 253, 193, 112, 28, 251, 193, 114, 28, 251, 193, 140, 28, 251, 193, 142, 28, 251, 193, 144, 28, 251, 193, 146, 28, 251, 193, 148, 28, 251, 193, 150, 28, 251, 193, 160, 28, 251, 193, 162, 28, 251, 193, 184, 28, 251, 193, 186, 28, 251, 193, 188, 28, 251, 193, 190, 28, 251, 193, 192, 28, 251, 193, 194, 28, 251, 193, 218, 162, 251, 193, 208, 28, 251, 193, 210, 28, 251, 193, 212, 28, 251, 193, 228, 28, 251, 193, 230, 28, 251, 193, 242, 28, 251, 193, 244, 28, 251, 193, 28, 51, 251, 193, 30, 51, 251, 193, 32, 51, 251, 193, 34, 51, 251, 193, 36, 51, 251, 193, 38, 51, 251, 193, 40, 51, 251, 193, 42, 51, 251, 193, 44, 51, 251, 193, 46, 51, 251, 193, 74, 51, 251, 193, 76, 51, 251, 193, 78, 51, 251, 193, 80, 51, 251, 193, 82, 51, 251, 193, 84, 51, 251, 193, 86, 51, 251, 193, 88, 51, 251, 193, 90, 51, 251, 193, 92, 51, 251, 193, 94, 51, 251, 193, 96, 51, 251, 193, 164, 28, 251, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 193, 176, 28, 251, 193, 202, 28, 251, 193, 222, 28, 251, 193, 224, 28, 251, 193, 240, 28, 251, 193, 44, 29, 251, 193, 108, 29, 251, 193, 180, 29, 251, 193, 182, 29, 251, 193, 6, 30, 251, 193, 8, 30, 251, 193, 36, 30, 251, 193, 54, 30, 251, 193, 56, 30, 251, 193, 102, 30, 251, 193, 138, 30, 251, 193, 140, 30, 251, 193, 218, 30, 251, 193, 6, 31, 251, 193, 58, 31, 251, 193, 60, 31, 251, 193, 76, 31, 251, 193, 140, 31, 251, 193, 142, 31, 251, 193, 144, 31, 251, 193, 204, 31, 251, 193, 206, 31, 251, 193, 4, 32, 251, 193, 6, 32, 251, 193, 8, 32, 251, 193, 70, 32, 251, 193, 148, 32, 251, 193, 246, 28, 251, 193, 248, 28, 251, 193, 250, 28, 251, 193, 252, 28, 251, 193, 254, 28, 251, 193, 2, 29, 251, 193, 4, 29, 251, 193, 6, 29, 251, 193, 22, 29, 251, 193, 24, 29, 251, 193, 26, 29, 251, 193, 28, 29, 251, 193, 30, 29, 251, 193, 32, 29, 251, 193, 34, 29, 251, 193, 46, 29, 251, 193, 48, 29, 251, 193, 50, 29, 251, 193, 80, 54, 251, 193, 68, 29, 251, 193, 70, 29, 251, 193, 72, 29, 251, 193, 80, 29, 251, 193, 82, 29, 251, 193, 86, 29, 251, 193, 88, 29, 251, 193, 90, 29, 251, 193, 104, 218, 252, 193, 106, 29, 251, 193, 122, 29, 251, 193, 124, 29, 251, 193, 126, 29, 251, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 193, 200, 55, 251, 193, 202, 55, 251, 193, 204, 55, 251, 193, 206, 55, 251, 193, 208, 55, 251, 193, 210, 55, 251, 193, 212, 55, 251, 193, 214, 55, 251, 193, 216, 55, 251, 193, 218, 55, 251, 193, 240, 55, 251, 193, 242, 55, 251, 193, 244, 55, 251, 193, 246, 55, 251, 193, 248, 55, 251, 193, 250, 55, 251, 193, 252, 55, 251, 193, 56, 101, 251, 193, 254, 55, 251, 193, 2, 56, 251, 193, 4, 56, 251, 193, 6, 56, 251, 193, 52, 29, 251, 193, 18, 56, 251, 193, 20, 56, 251, 193, 22, 56, 251, 193, 24, 56, 251, 193, 66, 29, 251, 193, 78, 29, 251, 193, 104, 29, 251, 193, 110, 29, 251, 193, 112, 29, 251, 193, 114, 29, 251, 193, 116, 29, 251, 193, 118, 29, 251, 193, 120, 29, 251, 193, 132, 29, 251, 193, 134, 29, 251, 193, 136, 29, 251, 193, 138, 29, 251, 193, 162, 29, 251, 193, 164, 29, 251, 193, 184, 29, 251, 193, 200, 29, 251, 193, 228, 29, 251, 193, 230, 29, 251, 193, 242, 29, 251, 193, 244, 29, 251, 193, 10, 30, 251, 193, 74, 30, 251, 193, 76, 30, 251, 193, 78, 30, 251, 193, 88, 30, 251, 193, 90, 30, 251, 193, 104, 30, 251, 193, 106, 30, 251, 193, 122, 30, 251, 193, 142, 30, 251, 193, 154, 30, 251, 193, 128, 29, 251, 193, 140, 29, 251, 193, 142, 29, 251, 193, 144, 29, 251, 193, 146, 29, 251, 193, 148, 29, 251, 193, 150, 29, 251, 193, 152, 29, 251, 193, 154, 29, 251, 193, 156, 29, 251, 193, 158, 29, 251, 193, 166, 29, 251, 193, 168, 29, 251, 193, 170, 29, 251, 193, 172, 29, 251, 193, 174, 29, 251, 193, 186, 29, 251, 193, 188, 29, 251, 193, 190, 29, 251, 193, 192, 29, 251, 193, 194, 29, 251, 193, 196, 29, 251, 193, 204, 29, 251, 193, 206, 29, 251, 193, 208, 29, 251, 193, 210, 29, 251, 193, 212, 29, 251, 193, 232, 29, 251, 193, 234, 29, 251, 193, 246, 29, 251, 193, 248, 29, 251, 193, 250, 29, 251, 193, 160, 29, 251, 193, 176, 29, 251, 193, 178, 29, 251, 193, 214, 29, 251, 193, 216, 29, 251, 193, 218, 29, 251, 193, 220, 29, 251, 193, 222, 29, 251, 193, 224, 29, 251, 193, 226, 29, 251, 193, 238, 29, 251, 193, 240, 29, 251, 193, 252, 29, 251, 193, 254, 29, 251, 193, 2, 30, 251, 193, 4, 30, 251, 193, 32, 30, 251, 193, 34, 30, 251, 193, 66, 30, 251, 193, 100, 30, 251, 193, 152, 30, 251, 193, 204, 30, 251, 193, 206, 30, 251, 193, 208, 30, 251, 193, 210, 30, 251, 193, 212, 30, 251, 193, 214, 30, 251, 193, 216, 30, 251, 193, 226, 30, 251, 193, 228, 30, 251, 193, 230, 30, 251, 193, 232, 30, 251, 193, 16, 30, 251, 193, 18, 30, 251, 193, 20, 30, 251, 193, 26, 30, 251, 193, 28, 30, 251, 193, 30, 30, 251, 193, 40, 30, 251, 193, 42, 30, 251, 193, 46, 30, 251, 193, 48, 30, 251, 193, 50, 30, 251, 193, 228, 181, 251, 193, 52, 30, 251, 193, 58, 30, 251, 193, 60, 30, 251, 193, 62, 30, 251, 193, 64, 30, 251, 193, 68, 30, 251, 193, 72, 30, 251, 193, 84, 30, 251, 193, 86, 30, 251, 193, 92, 30, 251, 193, 94, 30, 251, 193, 96, 30, 251, 193, 98, 30, 251, 193, 108, 30, 251, 193, 110, 30, 251, 193, 112, 30, 251, 193, 114, 30, 251, 193, 116, 30, 251, 193, 118, 30, 251, 193, 120, 30, 251, 193, 22, 30, 251, 193, 198, 232, 251, 193, 200, 232, 251, 193, 220, 232, 251, 193, 232, 232, 251, 193, 234, 232, 251, 193, 20, 233, 251, 193, 72, 233, 251, 193, 86, 233, 251, 193, 90, 233, 251, 193, 124, 233, 251, 193, 126, 233, 251, 193, 128, 233, 251, 193, 130, 233, 251, 193, 168, 233, 251, 193, 170, 233, 251, 193, 172, 233, 251, 193, 62, 205, 252, 193, 204, 233, 251, 193, 206, 233, 251, 193, 208, 233, 251, 193, 210, 233, 251, 193, 238, 233, 251, 193, 240, 233, 251, 193, 242, 233, 251, 193, 244, 233, 251, 193, 246, 233, 251, 193, 248, 233, 251, 193, 38, 234, 251, 193, 40, 234, 251, 193, 42, 234, 251, 193, 44, 234, 251, 193, 80, 30, 251, 193, 82, 30, 251, 193, 158, 30, 251, 193, 234, 30, 251, 193, 18, 31, 251, 193, 68, 31, 251, 193, 108, 31, 251, 193, 110, 31, 251, 193, 162, 31, 251, 193, 218, 31, 251, 193, 220, 31, 251, 193, 222, 31, 251, 193, 224, 31, 251, 193, 226, 31, 251, 193, 16, 32, 251, 193, 18, 32, 251, 193, 94, 32, 251, 193, 96, 32, 251, 193, 98, 32, 251, 193, 160, 32, 251, 193, 224, 32, 251, 193, 226, 32, 251, 193, 14, 33, 251, 193, 58, 33, 251, 193, 114, 34, 251, 193, 238, 34, 251, 193, 190, 35, 251, 193, 192, 35, 251, 193, 194, 35, 251, 193, 6, 36, 251, 193, 8, 36, 251, 193, 124, 36, 251, 193, 124, 30, 251, 193, 126, 30, 251, 193, 128, 30, 251, 193, 130, 30, 251, 193, 144, 30, 251, 193, 146, 30, 251, 193, 148, 30, 251, 193, 160, 30, 251, 193, 162, 30, 251, 193, 164, 30, 251, 193, 166, 30, 251, 193, 168, 30, 251, 193, 188, 30, 251, 193, 190, 30, 251, 193, 192, 30, 251, 193, 194, 30, 251, 193, 196, 30, 251, 193, 198, 30, 251, 193, 200, 30, 251, 193, 222, 30, 251, 193, 236, 30, 251, 193, 238, 30, 251, 193, 240, 30, 251, 193, 242, 30, 251, 193, 244, 30, 251, 193, 246, 30, 251, 193, 252, 30, 251, 193, 254, 30, 251, 193, 10, 31, 251, 193, 12, 31, 251, 193, 14, 31, 251, 193, 146, 138, 252, 193, 156, 30, 251, 193, 174, 30, 251, 193, 176, 30, 251, 193, 178, 30, 251, 193, 180, 30, 251, 193, 182, 30, 251, 193, 184, 30, 251, 193, 186, 30, 251, 193, 220, 30, 251, 193, 8, 31, 251, 193, 28, 31, 251, 193, 62, 31, 251, 193, 78, 31, 251, 193, 80, 31, 251, 193, 82, 31, 251, 193, 84, 31, 251, 193, 98, 31, 251, 193, 100, 31, 251, 193, 102, 31, 251, 193, 104, 31, 251, 193, 106, 31, 251, 193, 146, 31, 251, 193, 148, 31, 251, 193, 150, 31, 251, 193, 152, 31, 251, 193, 154, 31, 251, 193, 156, 31, 251, 193, 158, 31, 251, 193, 208, 31, 251, 193, 210, 31, 251, 193, 212, 31, 251, 193, 214, 31, 251, 193, 250, 30, 251, 193, 2, 31, 251, 193, 4, 31, 251, 193, 22, 31, 251, 193, 24, 31, 251, 193, 26, 31, 251, 193, 72, 31, 251, 193, 74, 31, 251, 193, 96, 31, 251, 193, 200, 31, 251, 193, 202, 31, 251, 193, 250, 31, 251, 193, 252, 31, 251, 193, 254, 31, 251, 193, 2, 32, 251, 193, 66, 32, 251, 193, 68, 32, 251, 193, 138, 32, 251, 193, 140, 32, 251, 193, 142, 32, 251, 193, 144, 32, 251, 193, 146, 32, 251, 193, 202, 32, 251, 193, 204, 32, 251, 193, 206, 32, 251, 193, 254, 32, 251, 193, 2, 33, 251, 193, 4, 33, 251, 193, 44, 33, 251, 193, 46, 33, 251, 193, 48, 33, 251, 193, 78, 33, 251, 193, 30, 31, 251, 193, 32, 31, 251, 193, 34, 31, 251, 193, 36, 31, 251, 193, 38, 31, 251, 193, 40, 31, 251, 193, 44, 31, 251, 193, 46, 31, 251, 193, 48, 31, 251, 193, 52, 31, 251, 193, 54, 31, 251, 193, 56, 31, 251, 193, 64, 31, 251, 193, 70, 31, 251, 193, 86, 31, 251, 193, 88, 31, 251, 193, 90, 31, 251, 193, 112, 31, 251, 193, 114, 31, 251, 193, 116, 31, 251, 193, 118, 31, 251, 193, 120, 31, 251, 193, 122, 31, 251, 193, 164, 31, 251, 193, 166, 31, 251, 193, 168, 31, 251, 193, 170, 31, 251, 193, 172, 31, 251, 193, 174, 31, 251, 193, 176, 31, 251, 193, 178, 31, 251, 193, 180, 31, 251, 193, 132, 31, 251, 193, 134, 31, 251, 193, 136, 31, 251, 193, 138, 31, 251, 193, 190, 31, 251, 193, 192, 31, 251, 193, 194, 31, 251, 193, 196, 31, 251, 193, 198, 31, 251, 193, 244, 31, 251, 193, 246, 31, 251, 193, 248, 31, 251, 193, 60, 32, 251, 193, 62, 32, 251, 193, 64, 32, 251, 193, 134, 32, 251, 193, 136, 32, 251, 193, 194, 32, 251, 193, 196, 32, 251, 193, 198, 32, 251, 193, 200, 32, 251, 193, 250, 32, 251, 193, 252, 32, 251, 193, 40, 33, 251, 193, 42, 33, 251, 193, 76, 33, 251, 193, 114, 33, 251, 193, 116, 33, 251, 193, 118, 33, 251, 193, 176, 33, 251, 193, 20, 34, 251, 193, 22, 34, 251, 193, 228, 31, 251, 193, 182, 31, 251, 193, 230, 31, 251, 193, 232, 31, 251, 193, 234, 31, 251, 193, 236, 31, 251, 193, 238, 31, 251, 193, 240, 31, 251, 193, 20, 32, 251, 193, 22, 32, 251, 193, 24, 32, 251, 193, 26, 32, 251, 193, 28, 32, 251, 193, 30, 32, 251, 193, 32, 32, 251, 193, 34, 32, 251, 193, 36, 32, 251, 193, 38, 32, 251, 193, 40, 32, 251, 193, 42, 32, 251, 193, 44, 32, 251, 193, 46, 32, 251, 193, 48, 32, 251, 193, 50, 32, 251, 206, 0, 3, 0, 206, 0, 3, 0, 206, 0, 3, 0, 206, 0, 3, 0, 206, 0, 3, 0, 206, 0, 3, 0, 206, 0, 3, 0, 206, 0, 3, 0, 193, 216, 31, 251, 193, 10, 32, 251, 193, 12, 32, 251, 193, 14, 32, 251, 193, 72, 32, 251, 193, 74, 32, 251, 193, 76, 32, 251, 193, 78, 32, 251, 193, 80, 32, 251, 193, 82, 32, 251, 193, 84, 32, 251, 193, 86, 32, 251, 193, 88, 32, 251, 193, 90, 32, 251, 193, 92, 32, 251, 193, 154, 32, 251, 193, 156, 32, 251, 193, 158, 32, 251, 193, 214, 32, 251, 193, 216, 32, 251, 193, 218, 32, 251, 193, 220, 32, 251, 193, 222, 32, 251, 193, 12, 33, 251, 193, 56, 33, 251, 193, 88, 33, 251, 193, 90, 33, 251, 193, 92, 33, 251, 193, 94, 33, 251, 193, 128, 33, 251, 193, 130, 33, 251, 193, 142, 33, 251, 193, 54, 32, 251, 193, 56, 32, 251, 193, 126, 32, 251, 193, 128, 32, 251, 193, 130, 32, 251, 193, 132, 32, 251, 193, 186, 32, 251, 193, 188, 32, 251, 193, 190, 32, 251, 193, 246, 32, 251, 193, 248, 32, 251, 193, 36, 33, 251, 193, 38, 33, 251, 193, 74, 33, 251, 193, 112, 33, 251, 193, 138, 33, 251, 193, 184, 33, 251, 193, 18, 34, 251, 193, 52, 34, 251, 193, 106, 34, 251, 193, 128, 34, 251, 193, 164, 34, 251, 193, 234, 34, 251, 193, 22, 35, 251, 193, 38, 35, 251, 193, 64, 35, 251, 193, 72, 35, 251, 193, 74, 35, 251, 193, 114, 35, 251, 193, 118, 35, 251, 193, 56, 36, 251, 193, 58, 36, 251, 206, 0, 3, 0, 206, 0, 3, 0, 206, 0, 3, 0, 206, 0, 3, 0, 206, 0, 3, 0, 193, 162, 32, 251, 193, 164, 32, 251, 193, 166, 32, 251, 193, 168, 32, 251, 193, 170, 32, 251, 193, 172, 32, 251, 193, 174, 32, 251, 193, 176, 32, 251, 193, 178, 32, 251, 193, 180, 32, 251, 193, 182, 32, 251, 193, 184, 32, 251, 193, 228, 32, 251, 193, 230, 32, 251, 193, 232, 32, 251, 193, 234, 32, 251, 193, 236, 32, 251, 193, 238, 32, 251, 193, 240, 32, 251, 193, 242, 32, 251, 193, 244, 32, 251, 206, 32, 3, 0, 206, 32, 3, 0, 206, 32, 3, 0, 206, 32, 3, 0, 206, 32, 3, 0, 206, 32, 3, 0, 193, 150, 32, 251, 193, 152, 32, 251, 193, 208, 32, 251, 193, 210, 32, 251, 193, 212, 32, 251, 193, 6, 33, 251, 193, 8, 33, 251, 193, 10, 33, 251, 193, 50, 33, 251, 193, 52, 33, 251, 193, 54, 33, 251, 193, 84, 33, 251, 193, 86, 33, 251, 193, 126, 33, 251, 193, 162, 33, 251, 193, 164, 33, 251, 193, 110, 34, 251, 193, 214, 34, 251, 193, 24, 35, 251, 193, 44, 35, 251, 193, 46, 35, 251, 193, 58, 35, 251, 193, 94, 35, 251, 193, 102, 35, 251, 193, 104, 35, 251, 193, 176, 35, 251, 193, 178, 35, 251, 193, 226, 35, 251, 193, 228, 35, 251, 193, 230, 35, 251, 193, 78, 36, 251, 193, 80, 36, 251, 206, 32, 3, 0, 206, 32, 3, 0, 206, 32, 3, 0, 206, 32, 3, 0, 193, 60, 33, 251, 193, 62, 33, 251, 193, 64, 33, 251, 193, 66, 33, 251, 193, 68, 33, 251, 193, 70, 33, 251, 193, 96, 33, 251, 193, 98, 33, 251, 193, 100, 33, 251, 193, 102, 33, 251, 193, 104, 33, 251, 193, 106, 33, 251, 193, 108, 33, 251, 193, 110, 33, 251, 193, 132, 33, 251, 193, 72, 33, 251, 193, 134, 33, 251, 193, 136, 33, 251, 193, 144, 33, 251, 193, 146, 33, 251, 193, 148, 33, 251, 193, 150, 33, 251, 193, 152, 33, 251, 193, 154, 33, 251, 193, 156, 33, 251, 193, 166, 33, 251, 193, 168, 33, 251, 193, 170, 33, 251, 193, 80, 33, 251, 193, 82, 33, 251, 193, 120, 33, 251, 193, 122, 33, 251, 193, 124, 33, 251, 193, 140, 33, 251, 193, 160, 33, 251, 193, 178, 33, 251, 193, 186, 33, 251, 193, 228, 33, 251, 193, 230, 33, 251, 193, 232, 33, 251, 193, 234, 33, 251, 193, 24, 34, 251, 193, 26, 34, 251, 193, 28, 34, 251, 193, 54, 34, 251, 193, 82, 34, 251, 193, 84, 34, 251, 193, 92, 34, 251, 193, 94, 34, 251, 193, 132, 34, 251, 193, 134, 34, 251, 193, 154, 34, 251, 193, 168, 34, 251, 193, 170, 34, 251, 193, 192, 34, 251, 193, 12, 35, 251, 193, 42, 35, 251, 193, 56, 35, 251, 193, 70, 35, 251, 193, 92, 35, 251, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 193, 166, 112, 252, 193, 168, 112, 252, 193, 170, 112, 252, 193, 172, 112, 252, 193, 174, 112, 252, 193, 176, 112, 252, 193, 178, 112, 252, 193, 180, 112, 252, 193, 182, 112, 252, 193, 184, 112, 252, 193, 186, 112, 252, 193, 188, 112, 252, 193, 190, 112, 252, 193, 192, 112, 252, 193, 158, 33, 251, 193, 194, 112, 252, 193, 196, 112, 252, 193, 198, 112, 252, 193, 200, 112, 252, 193, 220, 112, 252, 193, 222, 112, 252, 193, 224, 112, 252, 193, 248, 112, 252, 193, 226, 112, 252, 193, 228, 112, 252, 193, 230, 112, 252, 193, 232, 112, 252, 193, 172, 33, 251, 193, 174, 33, 251, 193, 180, 33, 251, 193, 182, 33, 251, 193, 188, 33, 251, 193, 190, 33, 251, 193, 194, 33, 251, 193, 198, 33, 251, 193, 200, 33, 251, 193, 212, 33, 251, 193, 214, 33, 251, 193, 216, 33, 251, 193, 218, 33, 251, 193, 220, 33, 251, 193, 222, 33, 251, 193, 224, 33, 251, 193, 226, 33, 251, 193, 242, 33, 251, 193, 244, 33, 251, 193, 246, 33, 251, 193, 248, 33, 251, 193, 4, 34, 251, 193, 6, 34, 251, 193, 8, 34, 251, 193, 10, 34, 251, 193, 12, 34, 251, 193, 14, 34, 251, 193, 16, 34, 251, 206, 64, 3, 0, 206, 64, 3, 0, 206, 64, 3, 0, 206, 64, 3, 0, 193, 196, 33, 251, 193, 202, 33, 251, 193, 204, 33, 251, 193, 206, 33, 251, 193, 210, 33, 251, 193, 236, 33, 251, 193, 238, 33, 251, 193, 240, 33, 251, 193, 250, 33, 251, 193, 252, 33, 251, 193, 254, 33, 251, 193, 2, 34, 251, 193, 30, 34, 251, 193, 56, 34, 251, 193, 58, 34, 251, 193, 60, 34, 251, 193, 112, 34, 251, 193, 136, 34, 251, 193, 138, 34, 251, 193, 172, 34, 251, 193, 174, 34, 251, 193, 200, 34, 251, 193, 206, 34, 251, 193, 216, 34, 251, 193, 218, 34, 251, 193, 220, 34, 251, 193, 222, 34, 251, 193, 224, 34, 251, 193, 236, 34, 251, 193, 14, 35, 251, 193, 26, 35, 251, 193, 60, 35, 251, 206, 64, 3, 0, 206, 64, 3, 0, 206, 64, 3, 0, 206, 64, 3, 0, 206, 64, 3, 0, 206, 64, 3, 0, 193, 62, 34, 251, 193, 64, 34, 251, 193, 66, 34, 251, 193, 68, 34, 251, 193, 70, 34, 251, 193, 72, 34, 251, 193, 74, 34, 251, 193, 76, 34, 251, 193, 78, 34, 251, 193, 86, 34, 251, 193, 88, 34, 251, 193, 90, 34, 251, 193, 96, 34, 251, 193, 98, 34, 251, 193, 100, 34, 251, 193, 102, 34, 251, 193, 104, 34, 251, 193, 148, 147, 251, 193, 116, 34, 251, 193, 118, 34, 251, 193, 218, 100, 251, 193, 120, 34, 251, 193, 122, 34, 251, 193, 124, 34, 251, 193, 126, 34, 251, 193, 142, 34, 251, 193, 170, 53, 251, 193, 172, 53, 251, 193, 174, 53, 251, 193, 176, 53, 251, 193, 178, 53, 251, 193, 180, 53, 251, 193, 188, 53, 251, 193, 190, 53, 251, 193, 192, 53, 251, 193, 202, 53, 251, 193, 204, 53, 251, 193, 206, 53, 251, 193, 208, 53, 251, 193, 210, 53, 251, 193, 80, 34, 251, 193, 212, 53, 251, 193, 214, 53, 251, 193, 216, 53, 251, 193, 218, 53, 251, 193, 68, 55, 251, 193, 220, 53, 251, 193, 222, 53, 251, 193, 224, 53, 251, 193, 226, 53, 251, 193, 228, 53, 251, 193, 246, 53, 251, 193, 248, 53, 251, 193, 250, 53, 251, 193, 252, 53, 251, 193, 22, 54, 251, 193, 24, 54, 251, 193, 34, 54, 251, 193, 108, 34, 251, 193, 130, 34, 251, 193, 152, 34, 251, 193, 166, 34, 251, 193, 40, 35, 251, 193, 66, 35, 251, 193, 68, 35, 251, 193, 76, 35, 251, 193, 78, 35, 251, 193, 80, 35, 251, 193, 88, 35, 251, 193, 90, 35, 251, 193, 98, 35, 251, 193, 120, 35, 251, 193, 122, 35, 251, 193, 124, 35, 251, 193, 216, 35, 251, 193, 218, 35, 251, 193, 62, 36, 251, 193, 64, 36, 251, 193, 66, 36, 251, 193, 182, 36, 251, 193, 184, 36, 251, 193, 186, 36, 251, 193, 188, 36, 251, 193, 190, 36, 251, 193, 192, 36, 251, 193, 194, 36, 251, 193, 108, 37, 251, 193, 110, 37, 251, 193, 112, 37, 251, 193, 114, 37, 251, 193, 144, 34, 251, 193, 146, 34, 251, 193, 148, 34, 251, 193, 150, 34, 251, 193, 156, 34, 251, 193, 158, 34, 251, 193, 160, 34, 251, 193, 162, 34, 251, 193, 176, 34, 251, 193, 178, 34, 251, 193, 180, 34, 251, 193, 182, 34, 251, 193, 238, 154, 253, 193, 184, 34, 251, 193, 186, 34, 251, 193, 188, 34, 251, 193, 190, 34, 251, 193, 194, 34, 251, 193, 196, 34, 251, 193, 198, 34, 251, 193, 202, 34, 251, 193, 204, 34, 251, 193, 208, 34, 251, 193, 210, 34, 251, 193, 212, 34, 251, 193, 226, 34, 251, 193, 228, 34, 251, 193, 230, 34, 251, 193, 190, 56, 251, 193, 232, 34, 251, 193, 240, 34, 251, 193, 242, 34, 251, 193, 244, 34, 251, 193, 246, 34, 251, 193, 248, 34, 251, 193, 250, 34, 251, 193, 252, 34, 251, 193, 254, 34, 251, 193, 2, 35, 251, 193, 4, 35, 251, 193, 6, 35, 251, 193, 8, 35, 251, 193, 10, 35, 251, 193, 16, 35, 251, 193, 18, 35, 251, 193, 20, 35, 251, 193, 28, 35, 251, 193, 30, 35, 251, 193, 32, 35, 251, 193, 156, 140, 253, 193, 34, 35, 251, 193, 36, 35, 251, 193, 48, 35, 251, 193, 50, 35, 251, 193, 52, 35, 251, 193, 54, 35, 251, 193, 62, 35, 251, 193, 82, 35, 251, 193, 84, 35, 251, 193, 84, 96, 251, 193, 96, 35, 251, 193, 106, 35, 251, 193, 108, 35, 251, 193, 116, 35, 251, 193, 14, 161, 251, 193, 16, 161, 251, 193, 22, 161, 251, 193, 24, 161, 251, 193, 26, 161, 251, 193, 28, 161, 251, 193, 38, 161, 251, 193, 40, 161, 251, 193, 42, 161, 251, 193, 44, 161, 251, 193, 46, 161, 251, 193, 48, 161, 251, 193, 56, 161, 251, 193, 86, 35, 251, 193, 62, 161, 251, 193, 64, 161, 251, 193, 66, 161, 251, 193, 68, 161, 251, 193, 80, 161, 251, 193, 82, 161, 251, 193, 84, 161, 251, 193, 92, 161, 251, 193, 96, 161, 251, 193, 102, 161, 251, 193, 104, 161, 251, 193, 106, 161, 251, 193, 110, 161, 251, 193, 116, 161, 251, 193, 118, 161, 251, 193, 134, 161, 251, 193, 136, 161, 251, 193, 140, 161, 251, 193, 100, 35, 251, 193, 110, 35, 251, 193, 170, 35, 251, 193, 172, 35, 251, 193, 174, 35, 251, 193, 220, 35, 251, 193, 222, 35, 251, 193, 224, 35, 251, 193, 68, 36, 251, 193, 70, 36, 251, 193, 72, 36, 251, 193, 74, 36, 251, 193, 76, 36, 251, 193, 196, 36, 251, 193, 198, 36, 251, 193, 200, 36, 251, 193, 202, 36, 251, 193, 204, 36, 251, 193, 206, 36, 251, 193, 208, 36, 251, 193, 210, 36, 251, 193, 116, 37, 251, 193, 118, 37, 251, 193, 120, 37, 251, 193, 122, 37, 251, 193, 124, 37, 251, 193, 126, 37, 251, 193, 128, 37, 251, 193, 130, 37, 251, 193, 132, 37, 251, 193, 134, 37, 251, 193, 136, 37, 251, 193, 178, 132, 251, 193, 158, 154, 251, 193, 148, 154, 251, 193, 160, 154, 251, 193, 162, 154, 251, 193, 164, 154, 251, 193, 166, 154, 251, 193, 168, 154, 251, 193, 170, 154, 251, 193, 172, 154, 251, 193, 180, 154, 251, 193, 188, 154, 251, 193, 196, 154, 251, 193, 198, 154, 251, 193, 200, 154, 251, 193, 202, 154, 251, 193, 204, 154, 251, 193, 206, 154, 251, 193, 208, 154, 251, 193, 210, 154, 251, 193, 114, 130, 251, 193, 112, 35, 251, 193, 216, 154, 251, 193, 250, 154, 251, 193, 252, 154, 251, 193, 254, 154, 251, 193, 2, 155, 251, 193, 4, 155, 251, 193, 6, 155, 251, 193, 8, 155, 251, 193, 10, 155, 251, 193, 12, 155, 251, 193, 126, 35, 251, 193, 128, 35, 251, 193, 130, 35, 251, 193, 138, 35, 251, 193, 144, 35, 251, 193, 154, 35, 251, 193, 156, 35, 251, 193, 158, 35, 251, 193, 160, 35, 251, 193, 162, 35, 251, 193, 164, 35, 251, 193, 166, 35, 251, 193, 168, 35, 251, 193, 196, 35, 251, 193, 198, 35, 251, 193, 200, 35, 251, 193, 202, 35, 251, 193, 204, 35, 251, 193, 206, 35, 251, 193, 208, 35, 251, 193, 210, 35, 251, 193, 10, 36, 251, 193, 212, 35, 251, 193, 12, 36, 251, 193, 14, 36, 251, 193, 16, 36, 251, 193, 18, 36, 251, 193, 214, 35, 251, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 193, 132, 35, 251, 193, 134, 35, 251, 193, 136, 35, 251, 193, 140, 35, 251, 193, 142, 35, 251, 193, 146, 35, 251, 193, 148, 35, 251, 193, 150, 35, 251, 193, 152, 35, 251, 193, 180, 35, 251, 193, 182, 35, 251, 193, 184, 35, 251, 193, 186, 35, 251, 193, 188, 35, 251, 193, 232, 35, 251, 193, 234, 35, 251, 193, 236, 35, 251, 193, 238, 35, 251, 193, 240, 35, 251, 193, 242, 35, 251, 193, 244, 35, 251, 193, 246, 35, 251, 193, 248, 35, 251, 193, 250, 35, 251, 193, 252, 35, 251, 193, 254, 35, 251, 193, 2, 36, 251, 193, 4, 36, 251, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 96, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 193, 60, 36, 251, 193, 178, 36, 251, 193, 180, 36, 251, 193, 102, 37, 251, 193, 104, 37, 251, 193, 106, 37, 251, 193, 4, 38, 251, 193, 6, 38, 251, 193, 186, 38, 251, 193, 188, 38, 251, 193, 72, 39, 251, 193, 74, 39, 251, 193, 76, 39, 251, 193, 78, 39, 251, 193, 218, 39, 251, 193, 220, 39, 251, 193, 222, 39, 251, 193, 74, 40, 251, 193, 76, 40, 251, 193, 78, 40, 251, 193, 52, 41, 251, 193, 110, 41, 251, 193, 112, 41, 251, 193, 150, 42, 251, 193, 152, 42, 251, 193, 154, 42, 251, 193, 220, 42, 251, 193, 222, 42, 251, 193, 38, 43, 251, 193, 112, 43, 251, 193, 114, 43, 251, 193, 116, 43, 251, 193, 82, 36, 251, 193, 84, 36, 251, 193, 86, 36, 251, 193, 212, 36, 251, 193, 214, 36, 251, 193, 216, 36, 251, 193, 218, 36, 251, 193, 142, 37, 251, 193, 144, 37, 251, 193, 146, 37, 251, 193, 148, 37, 251, 193, 150, 37, 251, 193, 28, 38, 251, 193, 30, 38, 251, 193, 32, 38, 251, 193, 34, 38, 251, 193, 36, 38, 251, 193, 38, 38, 251, 193, 40, 38, 251, 193, 220, 38, 251, 193, 222, 38, 251, 193, 224, 38, 251, 193, 226, 38, 251, 193, 116, 39, 251, 193, 118, 39, 251, 193, 120, 39, 251, 193, 122, 39, 251, 193, 106, 40, 251, 193, 108, 40, 251, 193, 110, 40, 251, 193, 230, 40, 251, 193, 232, 40, 251, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 206, 128, 3, 0, 193, 220, 36, 251, 193, 222, 36, 251, 193, 224, 36, 251, 193, 226, 36, 251, 193, 228, 36, 251, 193, 230, 36, 251, 193, 232, 36, 251, 193, 234, 36, 251, 193, 236, 36, 251, 193, 238, 36, 251, 193, 240, 36, 251, 193, 242, 36, 251, 193, 244, 36, 251, 193, 246, 36, 251, 193, 248, 36, 251, 193, 250, 36, 251, 193, 252, 36, 251, 193, 254, 36, 251, 193, 126, 36, 251, 193, 8, 37, 251, 193, 10, 37, 251, 193, 12, 37, 251, 193, 14, 37, 251, 193, 16, 37, 251, 193, 18, 37, 251, 193, 182, 37, 251, 193, 184, 37, 251, 193, 186, 37, 251, 193, 188, 37, 251, 193, 190, 37, 251, 193, 86, 38, 251, 193, 88, 38, 251, 193, 90, 38, 251, 193, 92, 38, 251, 193, 94, 38, 251, 193, 248, 38, 251, 193, 250, 38, 251, 193, 252, 38, 251, 193, 254, 38, 251, 193, 2, 39, 251, 193, 4, 39, 251, 193, 134, 39, 251, 193, 136, 39, 251, 193, 138, 39, 251, 193, 140, 39, 251, 193, 142, 39, 251, 193, 12, 40, 251, 193, 14, 40, 251, 193, 122, 40, 251, 193, 124, 40, 251, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 160, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 193, 2, 37, 251, 193, 4, 37, 251, 193, 6, 37, 251, 193, 152, 37, 251, 193, 154, 37, 251, 193, 156, 37, 251, 193, 158, 37, 251, 193, 160, 37, 251, 193, 162, 37, 251, 193, 164, 37, 251, 193, 166, 37, 251, 193, 168, 37, 251, 193, 170, 37, 251, 193, 172, 37, 251, 193, 174, 37, 251, 193, 176, 37, 251, 193, 178, 37, 251, 193, 180, 37, 251, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 192, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 193, 138, 37, 251, 193, 140, 37, 251, 193, 14, 38, 251, 193, 16, 38, 251, 193, 18, 38, 251, 193, 20, 38, 251, 193, 22, 38, 251, 193, 24, 38, 251, 193, 26, 38, 251, 193, 198, 38, 251, 193, 200, 38, 251, 193, 202, 38, 251, 193, 204, 38, 251, 193, 206, 38, 251, 193, 208, 38, 251, 193, 210, 38, 251, 193, 212, 38, 251, 193, 214, 38, 251, 193, 216, 38, 251, 193, 218, 38, 251, 193, 94, 39, 251, 193, 96, 39, 251, 193, 98, 39, 251, 193, 100, 39, 251, 193, 102, 39, 251, 193, 104, 39, 251, 193, 106, 39, 251, 193, 108, 39, 251, 193, 110, 39, 251, 193, 112, 39, 251, 193, 114, 39, 251, 193, 234, 39, 251, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 224, 3, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 193, 10, 38, 251, 193, 12, 38, 251, 193, 190, 38, 251, 193, 192, 38, 251, 193, 194, 38, 251, 193, 196, 38, 251, 193, 80, 39, 251, 193, 82, 39, 251, 193, 84, 39, 251, 193, 86, 39, 251, 193, 88, 39, 251, 193, 90, 39, 251, 193, 92, 39, 251, 193, 224, 39, 251, 193, 226, 39, 251, 193, 228, 39, 251, 193, 230, 39, 251, 193, 232, 39, 251, 193, 80, 40, 251, 193, 82, 40, 251, 193, 84, 40, 251, 193, 216, 40, 251, 193, 218, 40, 251, 193, 220, 40, 251, 193, 222, 40, 251, 193, 224, 40, 251, 193, 54, 41, 251, 193, 162, 41, 251, 193, 188, 41, 251, 193, 190, 41, 251, 193, 192, 41, 251, 193, 208, 41, 251, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 206, 0, 4, 0, 193, 228, 38, 251, 193, 124, 39, 251, 193, 230, 38, 251, 193, 232, 38, 251, 193, 234, 38, 251, 193, 236, 38, 251, 193, 238, 38, 251, 193, 240, 38, 251, 193, 242, 38, 251, 193, 244, 38, 251, 193, 126, 39, 251, 193, 128, 39, 251, 193, 130, 39, 251, 193, 132, 39, 251, 193, 246, 38, 251, 193, 248, 39, 251, 193, 250, 39, 251, 193, 252, 39, 251, 193, 254, 39, 251, 193, 2, 40, 251, 193, 4, 40, 251, 193, 6, 40, 251, 193, 8, 40, 251, 193, 10, 40, 251, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 206, 32, 4, 0, 193, 6, 39, 251, 193, 8, 39, 251, 193, 10, 39, 251, 193, 12, 39, 251, 193, 14, 39, 251, 193, 16, 39, 251, 193, 18, 39, 251, 193, 20, 39, 251, 193, 22, 39, 251, 193, 24, 39, 251, 193, 26, 39, 251, 193, 28, 39, 251, 193, 30, 39, 251, 193, 32, 39, 251, 193, 34, 39, 251, 193, 36, 39, 251, 193, 38, 39, 251, 193, 40, 39, 251, 193, 42, 39, 251, 193, 44, 39, 251, 193, 46, 39, 251, 193, 48, 39, 251, 193, 50, 39, 251, 193, 52, 39, 251, 193, 54, 39, 251, 193, 184, 38, 251, 193, 56, 39, 251, 193, 58, 39, 251, 193, 60, 39, 251, 193, 62, 39, 251, 193, 64, 39, 251, 193, 66, 39, 251, 193, 68, 39, 251, 193, 70, 39, 251, 193, 144, 39, 251, 193, 146, 39, 251, 193, 148, 39, 251, 193, 150, 39, 251, 193, 152, 39, 251, 193, 154, 39, 251, 193, 156, 39, 251, 193, 158, 39, 251, 193, 160, 39, 251, 193, 162, 39, 251, 193, 164, 39, 251, 193, 166, 39, 251, 193, 168, 39, 251, 193, 170, 39, 251, 193, 172, 39, 251, 193, 174, 39, 251, 193, 176, 39, 251, 193, 178, 39, 251, 193, 180, 39, 251, 193, 182, 39, 251, 193, 184, 39, 251, 193, 186, 39, 251, 193, 188, 39, 251, 193, 190, 39, 251, 193, 192, 39, 251, 193, 194, 39, 251, 193, 196, 39, 251, 193, 198, 39, 251, 193, 200, 39, 251, 193, 228, 142, 252, 193, 202, 39, 251, 193, 204, 39, 251, 193, 206, 39, 251, 193, 208, 39, 251, 193, 210, 39, 251, 193, 212, 39, 251, 193, 214, 39, 251, 193, 216, 39, 251, 193, 16, 40, 251, 193, 18, 40, 251, 193, 20, 40, 251, 193, 22, 40, 251, 193, 24, 40, 251, 193, 26, 40, 251, 193, 28, 40, 251, 193, 30, 40, 251, 193, 32, 40, 251, 193, 34, 40, 251, 193, 36, 40, 251, 193, 38, 40, 251, 193, 40, 40, 251, 193, 42, 40, 251, 193, 44, 40, 251, 193, 46, 40, 251, 193, 48, 40, 251, 193, 50, 40, 251, 193, 52, 40, 251, 193, 54, 40, 251, 193, 56, 40, 251, 193, 58, 40, 251, 193, 60, 40, 251, 193, 62, 40, 251, 193, 64, 40, 251, 193, 66, 40, 251, 193, 68, 40, 251, 193, 236, 39, 251, 193, 238, 39, 251, 193, 240, 39, 251, 193, 242, 39, 251, 193, 244, 39, 251, 193, 246, 39, 251, 193, 86, 40, 251, 193, 88, 40, 251, 193, 90, 40, 251, 193, 92, 40, 251, 193, 94, 40, 251, 193, 96, 40, 251, 193, 98, 40, 251, 193, 100, 40, 251, 193, 102, 40, 251, 193, 104, 40, 251, 193, 226, 40, 251, 193, 228, 40, 251, 193, 56, 41, 251, 193, 114, 41, 251, 193, 116, 41, 251, 193, 118, 41, 251, 193, 120, 41, 251, 193, 122, 41, 251, 193, 164, 41, 251, 193, 166, 41, 251, 193, 194, 41, 251, 193, 196, 41, 251, 193, 212, 41, 251, 193, 252, 41, 251, 193, 26, 42, 251, 193, 48, 42, 251, 193, 70, 40, 251, 193, 72, 40, 251, 193, 146, 40, 251, 193, 148, 40, 251, 193, 150, 40, 251, 193, 152, 40, 251, 193, 154, 40, 251, 193, 156, 40, 251, 193, 158, 40, 251, 193, 160, 40, 251, 193, 162, 40, 251, 193, 164, 40, 251, 193, 166, 40, 251, 193, 168, 40, 251, 193, 170, 40, 251, 193, 172, 40, 251, 193, 174, 40, 251, 193, 176, 40, 251, 193, 178, 40, 251, 193, 180, 40, 251, 193, 182, 40, 251, 193, 184, 40, 251, 193, 186, 40, 251, 193, 188, 40, 251, 193, 190, 40, 251, 193, 192, 40, 251, 193, 194, 40, 251, 193, 196, 40, 251, 193, 198, 40, 251, 193, 200, 40, 251, 193, 202, 40, 251, 193, 204, 40, 251, 193, 112, 40, 251, 193, 114, 40, 251, 193, 116, 40, 251, 193, 118, 40, 251, 193, 120, 40, 251, 193, 240, 40, 251, 193, 242, 40, 251, 193, 244, 40, 251, 193, 246, 40, 251, 193, 248, 40, 251, 193, 250, 40, 251, 193, 252, 40, 251, 193, 254, 40, 251, 193, 2, 41, 251, 193, 4, 41, 251, 193, 6, 41, 251, 193, 60, 41, 251, 193, 62, 41, 251, 193, 64, 41, 251, 193, 66, 41, 251, 193, 68, 41, 251, 193, 126, 41, 251, 193, 198, 41, 251, 193, 226, 41, 251, 193, 228, 41, 251, 193, 240, 41, 251, 193, 242, 41, 251, 193, 8, 42, 251, 193, 10, 42, 251, 193, 16, 42, 251, 193, 18, 42, 251, 193, 28, 42, 251, 193, 126, 40, 251, 193, 128, 40, 251, 193, 130, 40, 251, 193, 132, 40, 251, 193, 134, 40, 251, 193, 136, 40, 251, 193, 138, 40, 251, 193, 140, 40, 251, 193, 142, 40, 251, 193, 144, 40, 251, 193, 8, 41, 251, 193, 70, 41, 251, 193, 72, 41, 251, 193, 74, 41, 251, 193, 128, 41, 251, 193, 130, 41, 251, 193, 174, 41, 251, 193, 200, 41, 251, 193, 214, 41, 251, 193, 36, 42, 251, 193, 64, 42, 251, 193, 66, 42, 251, 193, 68, 42, 251, 193, 116, 42, 251, 193, 118, 42, 251, 193, 188, 42, 251, 193, 190, 42, 251, 193, 192, 42, 251, 193, 10, 43, 251, 193, 74, 43, 251, 193, 76, 43, 251, 193, 78, 43, 251, 193, 10, 41, 251, 193, 206, 40, 251, 193, 208, 40, 251, 193, 210, 40, 251, 193, 212, 40, 251, 193, 214, 40, 251, 193, 12, 41, 251, 193, 14, 41, 251, 193, 16, 41, 251, 193, 18, 41, 251, 193, 20, 41, 251, 193, 22, 41, 251, 193, 24, 41, 251, 193, 26, 41, 251, 193, 28, 41, 251, 193, 30, 41, 251, 193, 32, 41, 251, 193, 34, 41, 251, 193, 36, 41, 251, 193, 38, 41, 251, 193, 40, 41, 251, 193, 42, 41, 251, 193, 44, 41, 251, 193, 46, 41, 251, 193, 48, 41, 251, 193, 50, 41, 251, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 193, 234, 40, 251, 193, 236, 40, 251, 193, 238, 40, 251, 193, 58, 41, 251, 193, 124, 41, 251, 193, 168, 41, 251, 193, 170, 41, 251, 193, 172, 41, 251, 193, 248, 41, 251, 193, 90, 42, 251, 193, 160, 42, 251, 193, 162, 42, 251, 193, 164, 42, 251, 193, 236, 42, 251, 193, 238, 42, 251, 193, 142, 43, 251, 193, 144, 43, 251, 193, 196, 43, 251, 193, 198, 43, 251, 193, 200, 43, 251, 193, 254, 43, 251, 193, 64, 44, 251, 193, 128, 44, 251, 193, 130, 44, 251, 193, 160, 44, 251, 193, 216, 44, 251, 193, 232, 44, 251, 193, 14, 45, 251, 193, 102, 45, 251, 193, 104, 45, 251, 193, 154, 45, 251, 193, 156, 45, 251, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 206, 64, 4, 0, 193, 144, 51, 253, 193, 132, 41, 251, 193, 134, 41, 251, 193, 136, 41, 251, 193, 138, 41, 251, 193, 140, 41, 251, 193, 142, 41, 251, 193, 144, 41, 251, 193, 146, 41, 251, 193, 148, 41, 251, 193, 150, 41, 251, 193, 152, 41, 251, 193, 154, 41, 251, 193, 156, 41, 251, 193, 158, 41, 251, 193, 160, 41, 251, 193, 176, 41, 251, 193, 178, 41, 251, 193, 180, 41, 251, 193, 182, 41, 251, 193, 184, 41, 251, 193, 186, 41, 251, 193, 202, 41, 251, 193, 204, 41, 251, 193, 206, 41, 251, 193, 216, 41, 251, 193, 230, 41, 251, 193, 218, 41, 251, 193, 220, 41, 251, 193, 222, 41, 251, 193, 224, 41, 251, 193, 232, 41, 251, 193, 234, 41, 251, 193, 236, 41, 251, 193, 238, 41, 251, 193, 244, 41, 251, 193, 250, 41, 251, 193, 2, 42, 251, 193, 4, 42, 251, 193, 12, 42, 251, 193, 20, 42, 251, 193, 22, 42, 251, 193, 24, 42, 251, 193, 42, 42, 251, 193, 44, 42, 251, 193, 46, 42, 251, 193, 70, 42, 251, 193, 72, 42, 251, 193, 74, 42, 251, 193, 76, 42, 251, 193, 78, 42, 251, 193, 80, 42, 251, 193, 82, 42, 251, 193, 210, 41, 251, 193, 246, 41, 251, 193, 254, 41, 251, 193, 6, 42, 251, 193, 14, 42, 251, 193, 88, 42, 251, 193, 156, 42, 251, 193, 158, 42, 251, 193, 224, 42, 251, 193, 40, 43, 251, 193, 42, 43, 251, 193, 44, 43, 251, 193, 46, 43, 251, 193, 48, 43, 251, 193, 118, 43, 251, 193, 120, 43, 251, 193, 122, 43, 251, 193, 124, 43, 251, 193, 126, 43, 251, 193, 186, 43, 251, 193, 244, 43, 251, 193, 246, 43, 251, 193, 50, 44, 251, 193, 52, 44, 251, 193, 116, 44, 251, 193, 156, 44, 251, 193, 158, 44, 251, 193, 170, 44, 251, 193, 172, 44, 251, 193, 212, 44, 251, 193, 214, 44, 251, 193, 240, 44, 251, 193, 30, 42, 251, 193, 32, 42, 251, 193, 34, 42, 251, 193, 52, 42, 251, 193, 54, 42, 251, 193, 56, 42, 251, 193, 58, 42, 251, 193, 60, 42, 251, 193, 62, 42, 251, 193, 92, 42, 251, 193, 94, 42, 251, 193, 96, 42, 251, 193, 98, 42, 251, 193, 100, 42, 251, 193, 102, 42, 251, 193, 104, 42, 251, 193, 106, 42, 251, 193, 108, 42, 251, 193, 110, 42, 251, 193, 112, 42, 251, 193, 114, 42, 251, 193, 166, 42, 251, 193, 168, 42, 251, 193, 170, 42, 251, 193, 172, 42, 251, 193, 174, 42, 251, 193, 176, 42, 251, 193, 178, 42, 251, 193, 180, 42, 251, 193, 182, 42, 251, 193, 184, 42, 251, 193, 186, 42, 251, 206, 160, 8, 0, 206, 160, 8, 0, 206, 160, 8, 0, 206, 160, 8, 0, 206, 160, 8, 0, 193, 2, 78, 251, 193, 4, 78, 251, 193, 6, 78, 251, 193, 8, 78, 251, 193, 40, 79, 251, 193, 42, 79, 251, 193, 44, 79, 251, 193, 46, 79, 251, 193, 48, 79, 251, 193, 50, 79, 251, 193, 52, 79, 251, 193, 84, 80, 251, 193, 86, 80, 251, 193, 88, 80, 251, 193, 40, 81, 251, 193, 42, 81, 251, 193, 192, 81, 251, 193, 194, 81, 251, 193, 196, 81, 251, 193, 52, 82, 251, 193, 136, 82, 251, 193, 138, 82, 251, 193, 238, 82, 251, 193, 14, 83, 251, 193, 96, 83, 251, 193, 38, 42, 251, 193, 128, 83, 251, 193, 50, 42, 251, 193, 226, 42, 251, 193, 228, 42, 251, 193, 230, 42, 251, 193, 232, 42, 251, 193, 234, 42, 251, 193, 50, 43, 251, 193, 52, 43, 251, 193, 54, 43, 251, 193, 56, 43, 251, 193, 58, 43, 251, 193, 60, 43, 251, 193, 128, 43, 251, 193, 130, 43, 251, 193, 132, 43, 251, 193, 134, 43, 251, 193, 136, 43, 251, 193, 138, 43, 251, 193, 140, 43, 251, 193, 188, 43, 251, 193, 190, 43, 251, 193, 192, 43, 251, 193, 194, 43, 251, 193, 248, 43, 251, 193, 250, 43, 251, 193, 252, 43, 251, 193, 54, 44, 251, 193, 56, 44, 251, 193, 58, 44, 251, 193, 60, 44, 251, 193, 62, 44, 251, 193, 118, 44, 251, 193, 84, 42, 251, 193, 86, 42, 251, 193, 120, 42, 251, 193, 122, 42, 251, 193, 124, 42, 251, 193, 126, 42, 251, 193, 128, 42, 251, 193, 130, 42, 251, 193, 132, 42, 251, 193, 134, 42, 251, 193, 136, 42, 251, 193, 138, 42, 251, 193, 140, 42, 251, 193, 142, 42, 251, 193, 144, 42, 251, 193, 146, 42, 251, 193, 148, 42, 251, 193, 194, 42, 251, 193, 196, 42, 251, 193, 198, 42, 251, 193, 200, 42, 251, 193, 202, 42, 251, 193, 204, 42, 251, 193, 206, 42, 251, 193, 208, 42, 251, 193, 210, 42, 251, 193, 212, 42, 251, 193, 214, 42, 251, 193, 216, 42, 251, 193, 218, 42, 251, 193, 12, 43, 251, 193, 14, 43, 251, 193, 240, 42, 251, 193, 242, 42, 251, 193, 244, 42, 251, 193, 246, 42, 251, 193, 248, 42, 251, 193, 250, 42, 251, 193, 252, 42, 251, 193, 254, 42, 251, 193, 2, 43, 251, 193, 4, 43, 251, 193, 6, 43, 251, 193, 8, 43, 251, 193, 62, 43, 251, 193, 64, 43, 251, 193, 66, 43, 251, 193, 68, 43, 251, 193, 70, 43, 251, 193, 72, 43, 251, 193, 146, 43, 251, 193, 148, 43, 251, 193, 150, 43, 251, 193, 152, 43, 251, 193, 154, 43, 251, 193, 106, 225, 251, 193, 156, 43, 251, 193, 158, 43, 251, 193, 160, 43, 251, 193, 202, 43, 251, 193, 204, 43, 251, 193, 206, 43, 251, 193, 208, 43, 251, 193, 2, 44, 251, 193, 16, 43, 251, 193, 18, 43, 251, 193, 20, 43, 251, 193, 22, 43, 251, 193, 24, 43, 251, 193, 26, 43, 251, 193, 28, 43, 251, 193, 30, 43, 251, 193, 32, 43, 251, 193, 34, 43, 251, 193, 36, 43, 251, 193, 80, 43, 251, 193, 82, 43, 251, 193, 84, 43, 251, 193, 86, 43, 251, 193, 88, 43, 251, 193, 90, 43, 251, 193, 92, 43, 251, 193, 94, 43, 251, 193, 96, 43, 251, 193, 98, 43, 251, 193, 100, 43, 251, 193, 102, 43, 251, 193, 104, 43, 251, 193, 106, 43, 251, 193, 108, 43, 251, 193, 110, 43, 251, 206, 96, 4, 0, 206, 96, 4, 0, 206, 96, 4, 0, 206, 96, 4, 0, 206, 96, 4, 0, 193, 162, 43, 251, 193, 210, 43, 251, 193, 78, 44, 251, 193, 180, 44, 251, 193, 182, 44, 251, 193, 184, 44, 251, 193, 186, 44, 251, 193, 202, 44, 251, 193, 136, 45, 251, 193, 162, 45, 251, 193, 176, 45, 251, 193, 242, 45, 251, 193, 14, 46, 251, 193, 54, 46, 251, 193, 226, 214, 252, 193, 136, 46, 251, 193, 138, 46, 251, 193, 246, 46, 251, 193, 20, 47, 251, 193, 22, 47, 251, 193, 24, 47, 251, 193, 42, 47, 251, 193, 112, 47, 251, 193, 136, 47, 251, 193, 170, 47, 251, 193, 194, 47, 251, 193, 236, 47, 251, 193, 238, 47, 251, 193, 170, 48, 251, 193, 102, 49, 251, 193, 248, 49, 251, 193, 250, 49, 251, 206, 96, 4, 0, 206, 96, 4, 0, 206, 96, 4, 0, 206, 96, 4, 0, 206, 96, 4, 0, 193, 212, 43, 251, 193, 214, 43, 251, 193, 216, 43, 251, 193, 218, 43, 251, 193, 220, 43, 251, 193, 222, 43, 251, 193, 224, 43, 251, 193, 226, 43, 251, 193, 228, 43, 251, 193, 230, 43, 251, 193, 232, 43, 251, 193, 234, 43, 251, 193, 236, 43, 251, 193, 238, 43, 251, 193, 20, 44, 251, 193, 22, 44, 251, 193, 24, 44, 251, 193, 26, 44, 251, 193, 28, 44, 251, 193, 30, 44, 251, 193, 32, 44, 251, 193, 34, 44, 251, 193, 36, 44, 251, 193, 38, 44, 251, 193, 40, 44, 251, 193, 42, 44, 251, 193, 44, 44, 251, 193, 184, 43, 251, 193, 240, 43, 251, 193, 242, 43, 251, 193, 46, 44, 251, 193, 112, 44, 251, 193, 114, 44, 251, 193, 154, 44, 251, 193, 230, 44, 251, 193, 238, 45, 251, 193, 254, 45, 251, 193, 36, 46, 251, 193, 68, 46, 251, 193, 122, 46, 251, 193, 150, 46, 251, 193, 152, 46, 251, 193, 210, 46, 251, 193, 230, 46, 251, 193, 50, 47, 251, 193, 106, 47, 251, 193, 34, 48, 251, 193, 124, 48, 251, 193, 146, 48, 251, 193, 190, 48, 251, 193, 206, 48, 251, 193, 224, 48, 251, 193, 226, 48, 251, 193, 246, 48, 251, 193, 12, 49, 251, 193, 26, 49, 251, 193, 52, 49, 251, 193, 92, 49, 251, 193, 172, 49, 251, 193, 4, 44, 251, 193, 6, 44, 251, 193, 8, 44, 251, 193, 10, 44, 251, 193, 12, 44, 251, 193, 14, 44, 251, 193, 16, 44, 251, 193, 18, 44, 251, 193, 66, 44, 251, 193, 68, 44, 251, 193, 70, 44, 251, 193, 72, 44, 251, 193, 74, 44, 251, 193, 76, 44, 251, 193, 132, 44, 251, 193, 134, 44, 251, 193, 136, 44, 251, 193, 138, 44, 251, 193, 140, 44, 251, 193, 162, 44, 251, 193, 174, 44, 251, 193, 176, 44, 251, 193, 178, 44, 251, 193, 218, 44, 251, 193, 224, 44, 251, 193, 244, 44, 251, 193, 246, 44, 251, 193, 248, 44, 251, 193, 250, 44, 251, 193, 252, 44, 251, 193, 254, 44, 251, 193, 2, 45, 251, 193, 80, 44, 251, 193, 82, 44, 251, 193, 84, 44, 251, 193, 86, 44, 251, 193, 88, 44, 251, 193, 90, 44, 251, 193, 92, 44, 251, 193, 94, 44, 251, 193, 96, 44, 251, 193, 98, 44, 251, 193, 100, 44, 251, 193, 102, 44, 251, 193, 104, 44, 251, 193, 106, 44, 251, 193, 108, 44, 251, 193, 110, 44, 251, 193, 142, 44, 251, 193, 144, 44, 251, 193, 146, 44, 251, 193, 148, 44, 251, 193, 150, 44, 251, 193, 152, 44, 251, 193, 164, 44, 251, 193, 166, 44, 251, 193, 168, 44, 251, 193, 188, 44, 251, 193, 190, 44, 251, 193, 192, 44, 251, 193, 194, 44, 251, 193, 196, 44, 251, 193, 198, 44, 251, 193, 200, 44, 251, 193, 120, 44, 251, 193, 122, 44, 251, 193, 124, 44, 251, 193, 126, 44, 251, 193, 12, 45, 251, 193, 38, 45, 251, 193, 48, 45, 251, 193, 60, 45, 251, 193, 62, 45, 251, 193, 82, 45, 251, 193, 84, 45, 251, 193, 100, 45, 251, 193, 152, 45, 251, 193, 172, 45, 251, 193, 188, 45, 251, 193, 190, 45, 251, 193, 2, 46, 251, 193, 18, 46, 251, 193, 38, 46, 251, 193, 48, 46, 251, 193, 90, 46, 251, 193, 102, 46, 251, 193, 104, 46, 251, 193, 158, 46, 251, 193, 160, 46, 251, 193, 176, 46, 251, 193, 178, 46, 251, 193, 234, 46, 251, 193, 252, 46, 251, 193, 254, 46, 251, 193, 54, 47, 251, 193, 158, 47, 251, 193, 204, 44, 251, 193, 206, 44, 251, 193, 208, 44, 251, 193, 210, 44, 251, 193, 220, 44, 251, 193, 222, 44, 251, 193, 226, 44, 251, 193, 228, 44, 251, 193, 234, 44, 251, 193, 236, 44, 251, 193, 238, 44, 251, 193, 242, 44, 251, 193, 10, 45, 251, 193, 26, 45, 251, 193, 28, 45, 251, 193, 30, 45, 251, 193, 32, 45, 251, 193, 42, 45, 251, 193, 44, 45, 251, 193, 46, 45, 251, 193, 52, 45, 251, 193, 54, 45, 251, 193, 56, 45, 251, 193, 58, 45, 251, 193, 70, 45, 251, 193, 72, 45, 251, 193, 74, 45, 251, 193, 76, 45, 251, 193, 78, 45, 251, 193, 80, 45, 251, 193, 88, 45, 251, 193, 90, 45, 251, 193, 4, 45, 251, 193, 6, 45, 251, 193, 8, 45, 251, 193, 16, 45, 251, 193, 18, 45, 251, 193, 20, 45, 251, 193, 22, 45, 251, 193, 24, 45, 251, 193, 40, 45, 251, 193, 50, 45, 251, 193, 64, 45, 251, 193, 66, 45, 251, 193, 68, 45, 251, 193, 86, 45, 251, 193, 106, 45, 251, 193, 132, 45, 251, 193, 134, 45, 251, 193, 158, 45, 251, 193, 160, 45, 251, 193, 174, 45, 251, 193, 206, 45, 251, 193, 234, 45, 251, 193, 240, 45, 251, 193, 248, 45, 251, 193, 50, 46, 251, 193, 52, 46, 251, 193, 82, 46, 251, 193, 92, 46, 251, 193, 94, 46, 251, 193, 96, 46, 251, 193, 98, 46, 251, 193, 106, 46, 251, 193, 34, 45, 251, 193, 36, 45, 251, 193, 128, 45, 251, 193, 130, 45, 251, 193, 170, 45, 251, 193, 210, 45, 251, 193, 220, 45, 251, 193, 246, 45, 251, 193, 56, 46, 251, 193, 58, 46, 251, 193, 80, 46, 251, 193, 88, 46, 251, 193, 100, 46, 251, 193, 124, 46, 251, 193, 126, 46, 251, 193, 128, 46, 251, 193, 154, 46, 251, 193, 174, 46, 251, 193, 156, 46, 251, 193, 212, 46, 251, 193, 232, 46, 251, 193, 16, 47, 251, 193, 32, 47, 251, 193, 52, 47, 251, 193, 70, 47, 251, 193, 72, 47, 251, 193, 80, 47, 251, 193, 84, 47, 251, 193, 86, 47, 251, 193, 110, 47, 251, 193, 38, 48, 251, 193, 86, 48, 251, 193, 92, 45, 251, 193, 94, 45, 251, 193, 96, 45, 251, 193, 98, 45, 251, 193, 108, 45, 251, 193, 110, 45, 251, 193, 112, 45, 251, 193, 114, 45, 251, 193, 116, 45, 251, 193, 118, 45, 251, 193, 120, 45, 251, 193, 122, 45, 251, 193, 124, 45, 251, 193, 126, 45, 251, 193, 138, 45, 251, 193, 140, 45, 251, 193, 142, 45, 251, 193, 144, 45, 251, 193, 146, 45, 251, 193, 148, 45, 251, 193, 150, 45, 251, 193, 164, 45, 251, 193, 166, 45, 251, 193, 168, 45, 251, 193, 178, 45, 251, 193, 180, 45, 251, 193, 182, 45, 251, 193, 184, 45, 251, 193, 186, 45, 251, 193, 192, 45, 251, 193, 194, 45, 251, 193, 196, 45, 251, 193, 198, 45, 251, 193, 200, 45, 251, 193, 202, 45, 251, 193, 208, 45, 251, 193, 212, 45, 251, 193, 214, 45, 251, 193, 216, 45, 251, 193, 218, 45, 251, 193, 222, 45, 251, 193, 224, 45, 251, 193, 226, 45, 251, 193, 228, 45, 251, 193, 230, 45, 251, 193, 232, 45, 251, 193, 236, 45, 251, 193, 244, 45, 251, 193, 250, 45, 251, 193, 4, 46, 251, 193, 6, 46, 251, 193, 8, 46, 251, 193, 10, 46, 251, 193, 12, 46, 251, 193, 16, 46, 251, 193, 22, 46, 251, 193, 24, 46, 251, 193, 26, 46, 251, 193, 28, 46, 251, 193, 30, 46, 251, 193, 32, 46, 251, 193, 34, 46, 251, 193, 40, 46, 251, 193, 42, 46, 251, 193, 204, 45, 251, 193, 252, 45, 251, 193, 20, 46, 251, 193, 64, 46, 251, 193, 162, 46, 251, 193, 236, 46, 251, 193, 238, 46, 251, 193, 2, 47, 251, 193, 34, 47, 251, 193, 92, 47, 251, 193, 132, 47, 251, 193, 144, 47, 251, 193, 162, 47, 251, 193, 204, 47, 251, 193, 206, 47, 251, 193, 224, 47, 251, 193, 16, 48, 251, 193, 18, 48, 251, 193, 40, 48, 251, 193, 42, 48, 251, 193, 44, 48, 251, 193, 90, 48, 251, 193, 106, 48, 251, 193, 108, 48, 251, 193, 128, 48, 251, 193, 130, 48, 251, 193, 150, 48, 251, 193, 152, 48, 251, 193, 198, 48, 251, 193, 210, 48, 251, 193, 212, 48, 251, 193, 254, 48, 251, 193, 108, 46, 251, 193, 110, 46, 251, 193, 112, 46, 251, 193, 130, 46, 251, 193, 132, 46, 251, 193, 134, 46, 251, 193, 164, 46, 251, 193, 180, 46, 251, 193, 182, 46, 251, 193, 184, 46, 251, 193, 214, 46, 251, 193, 216, 46, 251, 193, 240, 46, 251, 193, 242, 46, 251, 193, 244, 46, 251, 193, 18, 47, 251, 193, 36, 47, 251, 193, 38, 47, 251, 193, 40, 47, 251, 193, 56, 47, 251, 193, 74, 47, 251, 193, 82, 47, 251, 193, 94, 47, 251, 193, 100, 47, 251, 193, 118, 47, 251, 193, 120, 47, 251, 193, 122, 47, 251, 193, 134, 47, 251, 193, 146, 47, 251, 193, 150, 47, 251, 193, 164, 47, 251, 193, 166, 47, 251, 193, 194, 46, 251, 193, 196, 46, 251, 193, 198, 46, 251, 193, 152, 47, 251, 193, 200, 46, 251, 193, 202, 46, 251, 193, 204, 46, 251, 193, 206, 46, 251, 193, 208, 46, 251, 193, 220, 46, 251, 193, 222, 46, 251, 193, 224, 46, 251, 193, 226, 46, 251, 193, 228, 46, 251, 193, 248, 46, 251, 193, 250, 46, 251, 193, 4, 47, 251, 193, 6, 47, 251, 193, 8, 47, 251, 193, 10, 47, 251, 193, 12, 47, 251, 193, 14, 47, 251, 193, 26, 47, 251, 193, 28, 47, 251, 193, 30, 47, 251, 193, 44, 47, 251, 193, 46, 47, 251, 193, 48, 47, 251, 193, 58, 47, 251, 193, 60, 47, 251, 193, 62, 47, 251, 193, 64, 47, 251, 193, 66, 47, 251, 193, 68, 47, 251, 193, 76, 47, 251, 193, 78, 47, 251, 193, 88, 47, 251, 193, 90, 47, 251, 193, 96, 47, 251, 193, 98, 47, 251, 193, 102, 47, 251, 193, 104, 47, 251, 193, 108, 47, 251, 193, 114, 47, 251, 193, 116, 47, 251, 193, 124, 47, 251, 193, 128, 47, 251, 193, 130, 47, 251, 193, 138, 47, 251, 193, 140, 47, 251, 193, 142, 47, 251, 193, 148, 47, 251, 193, 154, 47, 251, 193, 156, 47, 251, 193, 160, 47, 251, 193, 172, 47, 251, 193, 174, 47, 251, 193, 176, 47, 251, 193, 178, 47, 251, 193, 196, 47, 251, 193, 198, 47, 251, 193, 216, 47, 251, 193, 218, 47, 251, 193, 200, 47, 251, 193, 202, 47, 251, 193, 102, 48, 251, 193, 104, 48, 251, 193, 126, 48, 251, 193, 148, 48, 251, 193, 166, 48, 251, 193, 168, 48, 251, 193, 196, 48, 251, 193, 232, 48, 251, 193, 234, 48, 251, 193, 236, 48, 251, 193, 14, 49, 251, 193, 16, 49, 251, 193, 28, 49, 251, 193, 38, 49, 251, 193, 76, 49, 251, 193, 94, 49, 251, 193, 190, 49, 251, 193, 252, 49, 251, 193, 12, 50, 251, 193, 24, 50, 251, 193, 44, 50, 251, 193, 78, 50, 251, 193, 80, 50, 251, 193, 82, 50, 251, 193, 84, 50, 251, 193, 86, 50, 251, 193, 112, 50, 251, 193, 114, 50, 251, 193, 146, 50, 251, 193, 148, 50, 251, 193, 156, 50, 251, 193, 220, 47, 251, 193, 222, 47, 251, 193, 240, 47, 251, 193, 242, 47, 251, 193, 244, 47, 251, 193, 246, 47, 251, 193, 248, 47, 251, 193, 250, 47, 251, 193, 252, 47, 251, 193, 254, 47, 251, 193, 2, 48, 251, 193, 4, 48, 251, 193, 6, 48, 251, 193, 8, 48, 251, 193, 10, 48, 251, 193, 12, 48, 251, 193, 14, 48, 251, 193, 20, 48, 251, 193, 22, 48, 251, 193, 24, 48, 251, 193, 26, 48, 251, 193, 28, 48, 251, 193, 30, 48, 251, 193, 32, 48, 251, 193, 60, 48, 251, 193, 62, 48, 251, 193, 64, 48, 251, 193, 66, 48, 251, 193, 68, 48, 251, 193, 70, 48, 251, 193, 72, 48, 251, 193, 74, 48, 251, 193, 76, 48, 251, 193, 78, 48, 251, 193, 80, 48, 251, 193, 110, 48, 251, 193, 112, 48, 251, 193, 114, 48, 251, 193, 116, 48, 251, 193, 118, 48, 251, 193, 120, 48, 251, 193, 122, 48, 251, 193, 134, 48, 251, 193, 136, 48, 251, 193, 138, 48, 251, 193, 140, 48, 251, 193, 142, 48, 251, 193, 144, 48, 251, 193, 156, 48, 251, 193, 158, 48, 251, 193, 160, 48, 251, 193, 162, 48, 251, 193, 172, 48, 251, 193, 174, 48, 251, 193, 176, 48, 251, 193, 178, 48, 251, 193, 180, 48, 251, 193, 182, 48, 251, 193, 184, 48, 251, 193, 186, 48, 251, 193, 188, 48, 251, 193, 200, 48, 251, 193, 202, 48, 251, 193, 204, 48, 251, 193, 38, 10, 252, 193, 40, 10, 252, 193, 68, 10, 252, 193, 70, 10, 252, 193, 72, 10, 252, 193, 82, 48, 251, 193, 96, 10, 252, 193, 98, 10, 252, 193, 100, 10, 252, 193, 102, 10, 252, 193, 104, 10, 252, 193, 106, 10, 252, 193, 108, 10, 252, 193, 110, 10, 252, 193, 112, 10, 252, 193, 114, 10, 252, 193, 116, 10, 252, 193, 118, 10, 252, 193, 198, 165, 251, 193, 140, 10, 252, 193, 142, 10, 252, 193, 144, 10, 252, 193, 146, 10, 252, 193, 148, 10, 252, 193, 150, 10, 252, 206, 224, 40, 0, 206, 224, 40, 0, 206, 224, 40, 0, 206, 224, 40, 0, 206, 224, 40, 0, 206, 224, 40, 0, 206, 224, 40, 0, 193, 88, 48, 251, 193, 98, 48, 251, 193, 100, 48, 251, 193, 164, 48, 251, 193, 192, 48, 251, 193, 194, 48, 251, 193, 208, 48, 251, 193, 228, 48, 251, 193, 230, 48, 251, 193, 248, 48, 251, 193, 250, 48, 251, 193, 252, 48, 251, 193, 24, 49, 251, 193, 34, 49, 251, 193, 126, 49, 251, 193, 152, 49, 251, 193, 174, 49, 251, 193, 200, 49, 251, 193, 202, 49, 251, 193, 210, 49, 251, 193, 212, 49, 251, 193, 220, 49, 251, 193, 230, 49, 251, 193, 232, 49, 251, 193, 234, 49, 251, 193, 10, 50, 251, 193, 42, 50, 251, 193, 120, 50, 251, 193, 132, 50, 251, 193, 144, 50, 251, 193, 152, 50, 251, 193, 154, 50, 251, 193, 18, 49, 251, 193, 20, 49, 251, 193, 42, 49, 251, 193, 128, 49, 251, 193, 154, 49, 251, 193, 156, 49, 251, 193, 158, 49, 251, 193, 204, 49, 251, 193, 206, 49, 251, 193, 228, 49, 251, 193, 242, 49, 251, 193, 14, 50, 251, 193, 46, 50, 251, 193, 48, 50, 251, 193, 62, 50, 251, 193, 88, 50, 251, 193, 108, 50, 251, 193, 122, 50, 251, 193, 124, 50, 251, 193, 158, 50, 251, 193, 162, 50, 251, 193, 216, 50, 251, 193, 8, 51, 251, 193, 10, 51, 251, 193, 60, 51, 251, 193, 172, 51, 251, 193, 174, 51, 251, 193, 246, 51, 251, 193, 248, 51, 251, 193, 250, 51, 251, 193, 252, 51, 251, 193, 254, 51, 251, 193, 56, 49, 251, 193, 58, 49, 251, 193, 60, 49, 251, 193, 78, 49, 251, 193, 80, 49, 251, 193, 96, 49, 251, 193, 98, 49, 251, 193, 100, 49, 251, 193, 176, 49, 251, 193, 240, 49, 251, 193, 244, 49, 251, 193, 246, 49, 251, 193, 254, 49, 251, 193, 16, 50, 251, 193, 18, 50, 251, 193, 20, 50, 251, 193, 26, 50, 251, 193, 28, 50, 251, 193, 50, 50, 251, 193, 52, 50, 251, 193, 54, 50, 251, 193, 56, 50, 251, 193, 64, 50, 251, 193, 66, 50, 251, 193, 68, 50, 251, 193, 70, 50, 251, 193, 72, 50, 251, 193, 90, 50, 251, 193, 92, 50, 251, 193, 94, 50, 251, 193, 116, 50, 251, 193, 118, 50, 251, 193, 128, 119, 251, 193, 130, 119, 251, 193, 132, 119, 251, 193, 156, 119, 251, 193, 158, 119, 251, 193, 160, 119, 251, 193, 186, 119, 251, 193, 188, 119, 251, 193, 190, 119, 251, 193, 192, 119, 251, 193, 194, 119, 251, 193, 90, 49, 251, 193, 196, 119, 251, 193, 198, 119, 251, 193, 240, 119, 251, 193, 242, 119, 251, 193, 244, 119, 251, 193, 246, 119, 251, 193, 248, 119, 251, 193, 250, 119, 251, 193, 252, 119, 251, 193, 254, 119, 251, 193, 2, 120, 251, 193, 4, 120, 251, 193, 6, 120, 251, 193, 8, 120, 251, 193, 10, 120, 251, 206, 64, 17, 0, 206, 64, 17, 0, 206, 64, 17, 0, 206, 64, 17, 0, 206, 64, 17, 0, 193, 108, 49, 251, 193, 110, 49, 251, 193, 112, 49, 251, 193, 114, 49, 251, 193, 116, 49, 251, 193, 118, 49, 251, 193, 120, 49, 251, 193, 122, 49, 251, 193, 124, 49, 251, 193, 130, 49, 251, 193, 132, 49, 251, 193, 134, 49, 251, 193, 136, 49, 251, 193, 138, 49, 251, 193, 140, 49, 251, 193, 142, 49, 251, 193, 144, 49, 251, 193, 146, 49, 251, 193, 148, 49, 251, 193, 160, 49, 251, 193, 162, 49, 251, 193, 164, 49, 251, 193, 166, 49, 251, 193, 168, 49, 251, 193, 242, 208, 253, 193, 170, 49, 251, 193, 178, 49, 251, 193, 180, 49, 251, 193, 182, 49, 251, 193, 184, 49, 251, 193, 186, 49, 251, 193, 188, 49, 251, 193, 72, 55, 251, 193, 74, 55, 251, 193, 76, 55, 251, 193, 78, 55, 251, 193, 80, 55, 251, 193, 82, 55, 251, 193, 84, 55, 251, 193, 86, 55, 251, 193, 88, 55, 251, 193, 150, 49, 251, 193, 112, 55, 251, 193, 114, 55, 251, 193, 116, 55, 251, 193, 118, 55, 251, 193, 120, 55, 251, 193, 122, 55, 251, 193, 124, 55, 251, 193, 126, 55, 251, 193, 128, 55, 251, 193, 130, 55, 251, 193, 132, 55, 251, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 206, 192, 4, 0, 193, 192, 49, 251, 193, 194, 49, 251, 193, 196, 49, 251, 193, 24, 102, 251, 193, 198, 49, 251, 193, 208, 49, 251, 193, 214, 49, 251, 193, 216, 49, 251, 193, 222, 49, 251, 193, 224, 49, 251, 193, 226, 49, 251, 193, 236, 49, 251, 193, 238, 49, 251, 193, 2, 50, 251, 193, 4, 50, 251, 193, 6, 50, 251, 193, 8, 50, 251, 193, 22, 50, 251, 193, 30, 50, 251, 193, 32, 50, 251, 193, 34, 50, 251, 193, 36, 50, 251, 193, 38, 50, 251, 193, 40, 50, 251, 193, 60, 50, 251, 193, 96, 50, 251, 193, 98, 50, 251, 193, 100, 50, 251, 193, 102, 50, 251, 193, 104, 50, 251, 193, 110, 50, 251, 193, 126, 50, 251, 193, 218, 49, 251, 193, 76, 50, 251, 193, 106, 50, 251, 193, 210, 50, 251, 193, 48, 51, 251, 193, 98, 51, 251, 193, 100, 51, 251, 193, 158, 51, 251, 193, 160, 51, 251, 193, 162, 51, 251, 193, 164, 51, 251, 193, 226, 51, 251, 193, 228, 51, 251, 193, 230, 51, 251, 193, 232, 51, 251, 193, 40, 52, 251, 193, 42, 52, 251, 193, 92, 52, 251, 193, 94, 52, 251, 193, 158, 52, 251, 193, 160, 52, 251, 193, 162, 52, 251, 193, 232, 52, 251, 193, 234, 52, 251, 193, 26, 53, 251, 193, 28, 53, 251, 193, 60, 53, 251, 193, 86, 53, 251, 193, 230, 53, 251, 193, 254, 53, 251, 193, 26, 54, 251, 193, 50, 54, 251, 193, 58, 50, 251, 193, 74, 50, 251, 193, 226, 50, 251, 193, 228, 50, 251, 193, 230, 50, 251, 193, 20, 51, 251, 193, 22, 51, 251, 193, 24, 51, 251, 193, 70, 51, 251, 193, 72, 51, 251, 193, 124, 51, 251, 193, 186, 51, 251, 193, 188, 51, 251, 193, 190, 51, 251, 193, 70, 52, 251, 193, 138, 52, 251, 193, 188, 52, 251, 193, 190, 52, 251, 193, 254, 52, 251, 193, 56, 53, 251, 193, 104, 53, 251, 193, 136, 53, 251, 193, 244, 53, 251, 193, 18, 54, 251, 193, 20, 54, 251, 193, 92, 54, 251, 193, 138, 54, 251, 193, 140, 54, 251, 193, 190, 54, 251, 193, 230, 54, 251, 193, 24, 55, 251, 193, 66, 55, 251, 193, 128, 50, 251, 193, 130, 50, 251, 193, 138, 50, 251, 193, 140, 50, 251, 193, 142, 50, 251, 193, 150, 50, 251, 193, 164, 50, 251, 193, 166, 50, 251, 193, 168, 50, 251, 193, 170, 50, 251, 193, 172, 50, 251, 193, 176, 50, 251, 193, 188, 50, 251, 193, 196, 50, 251, 193, 198, 50, 251, 193, 200, 50, 251, 193, 202, 50, 251, 193, 204, 50, 251, 193, 206, 50, 251, 193, 208, 50, 251, 193, 236, 50, 251, 193, 238, 50, 251, 193, 240, 50, 251, 193, 242, 50, 251, 193, 244, 50, 251, 193, 246, 50, 251, 193, 248, 50, 251, 193, 250, 50, 251, 193, 252, 50, 251, 193, 254, 50, 251, 193, 2, 51, 251, 193, 4, 51, 251, 193, 134, 50, 251, 193, 136, 50, 251, 193, 178, 50, 251, 193, 180, 50, 251, 193, 182, 50, 251, 193, 184, 50, 251, 193, 186, 50, 251, 193, 190, 50, 251, 193, 192, 50, 251, 193, 194, 50, 251, 193, 218, 50, 251, 193, 220, 50, 251, 193, 222, 50, 251, 193, 224, 50, 251, 193, 12, 51, 251, 193, 14, 51, 251, 193, 16, 51, 251, 193, 18, 51, 251, 193, 62, 51, 251, 193, 64, 51, 251, 193, 66, 51, 251, 193, 68, 51, 251, 193, 112, 51, 251, 193, 114, 51, 251, 193, 116, 51, 251, 193, 118, 51, 251, 193, 120, 51, 251, 193, 122, 51, 251, 193, 176, 51, 251, 193, 178, 51, 251, 193, 180, 51, 251, 193, 182, 51, 251, 193, 160, 50, 251, 193, 174, 50, 251, 193, 212, 50, 251, 193, 52, 51, 251, 193, 54, 51, 251, 193, 102, 51, 251, 193, 104, 51, 251, 193, 106, 51, 251, 193, 108, 51, 251, 193, 166, 51, 251, 193, 168, 51, 251, 193, 238, 51, 251, 193, 44, 52, 251, 193, 46, 52, 251, 193, 48, 52, 251, 193, 50, 52, 251, 193, 52, 52, 251, 193, 96, 52, 251, 193, 98, 52, 251, 193, 100, 52, 251, 193, 102, 52, 251, 193, 104, 52, 251, 193, 106, 52, 251, 193, 108, 52, 251, 193, 164, 52, 251, 193, 166, 52, 251, 193, 168, 52, 251, 193, 170, 52, 251, 193, 236, 52, 251, 193, 238, 52, 251, 193, 30, 53, 251, 193, 32, 53, 251, 193, 214, 50, 251, 193, 6, 51, 251, 193, 56, 51, 251, 193, 58, 51, 251, 193, 110, 51, 251, 193, 170, 51, 251, 193, 240, 51, 251, 193, 242, 51, 251, 193, 244, 51, 251, 193, 54, 52, 251, 193, 56, 52, 251, 193, 58, 52, 251, 193, 110, 52, 251, 193, 112, 52, 251, 193, 114, 52, 251, 193, 116, 52, 251, 193, 118, 52, 251, 193, 172, 52, 251, 193, 174, 52, 251, 193, 240, 52, 251, 193, 242, 52, 251, 193, 244, 52, 251, 193, 34, 53, 251, 193, 52, 53, 251, 193, 110, 53, 251, 193, 182, 53, 251, 193, 184, 53, 251, 193, 236, 53, 251, 193, 238, 53, 251, 193, 240, 53, 251, 193, 4, 54, 251, 193, 6, 54, 251, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 206, 128, 4, 0, 193, 192, 51, 251, 193, 194, 51, 251, 193, 196, 51, 251, 193, 198, 51, 251, 193, 200, 51, 251, 193, 202, 51, 251, 193, 204, 51, 251, 193, 206, 51, 251, 193, 208, 51, 251, 193, 210, 51, 251, 193, 212, 51, 251, 193, 214, 51, 251, 193, 216, 51, 251, 193, 218, 51, 251, 193, 220, 51, 251, 193, 222, 51, 251, 193, 224, 51, 251, 193, 12, 52, 251, 193, 14, 52, 251, 193, 16, 52, 251, 193, 18, 52, 251, 193, 20, 52, 251, 193, 22, 52, 251, 193, 24, 52, 251, 193, 140, 52, 251, 193, 128, 251, 252, 193, 130, 251, 252, 193, 224, 251, 252, 193, 226, 251, 252, 193, 228, 251, 252, 193, 230, 251, 252, 193, 38, 252, 252, 193, 40, 252, 252, 193, 42, 252, 252, 193, 44, 252, 252, 193, 112, 252, 252, 193, 150, 252, 252, 193, 152, 252, 252, 193, 174, 252, 252, 193, 178, 252, 252, 193, 180, 252, 252, 193, 194, 252, 252, 193, 214, 252, 252, 193, 184, 51, 251, 193, 216, 252, 252, 193, 218, 252, 252, 193, 2, 253, 252, 193, 4, 253, 252, 193, 6, 253, 252, 193, 8, 253, 252, 193, 10, 253, 252, 193, 12, 253, 252, 193, 54, 253, 252, 193, 56, 253, 252, 193, 58, 253, 252, 193, 98, 253, 252, 193, 100, 253, 252, 193, 2, 52, 251, 193, 4, 52, 251, 193, 6, 52, 251, 193, 8, 52, 251, 193, 60, 52, 251, 193, 62, 52, 251, 193, 64, 52, 251, 193, 66, 52, 251, 193, 68, 52, 251, 193, 10, 52, 251, 193, 134, 52, 251, 193, 136, 52, 251, 193, 178, 52, 251, 193, 180, 52, 251, 193, 182, 52, 251, 193, 184, 52, 251, 193, 186, 52, 251, 193, 250, 52, 251, 193, 252, 52, 251, 193, 54, 53, 251, 193, 82, 53, 251, 193, 114, 53, 251, 193, 120, 53, 251, 193, 130, 53, 251, 193, 132, 53, 251, 193, 134, 53, 251, 193, 144, 53, 251, 193, 146, 53, 251, 193, 148, 53, 251, 193, 162, 53, 251, 193, 164, 53, 251, 193, 186, 53, 251, 193, 26, 52, 251, 193, 28, 52, 251, 193, 30, 52, 251, 193, 32, 52, 251, 193, 34, 52, 251, 193, 36, 52, 251, 193, 38, 52, 251, 193, 72, 52, 251, 193, 74, 52, 251, 193, 76, 52, 251, 193, 78, 52, 251, 193, 80, 52, 251, 193, 82, 52, 251, 193, 84, 52, 251, 193, 86, 52, 251, 193, 88, 52, 251, 193, 90, 52, 251, 193, 142, 52, 251, 193, 144, 52, 251, 193, 146, 52, 251, 193, 148, 52, 251, 193, 150, 52, 251, 193, 152, 52, 251, 193, 154, 52, 251, 193, 156, 52, 251, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 193, 120, 52, 251, 193, 122, 52, 251, 193, 124, 52, 251, 193, 126, 52, 251, 193, 128, 52, 251, 193, 130, 52, 251, 193, 132, 52, 251, 193, 176, 52, 251, 193, 246, 52, 251, 193, 248, 52, 251, 193, 64, 53, 251, 193, 66, 53, 251, 193, 90, 53, 251, 193, 160, 53, 251, 193, 242, 53, 251, 193, 52, 54, 251, 193, 186, 54, 251, 193, 216, 54, 251, 193, 218, 54, 251, 193, 220, 54, 251, 193, 222, 54, 251, 193, 224, 54, 251, 193, 226, 54, 251, 193, 6, 55, 251, 193, 8, 55, 251, 193, 10, 55, 251, 193, 12, 55, 251, 193, 50, 55, 251, 193, 52, 55, 251, 193, 54, 55, 251, 193, 104, 55, 251, 193, 106, 55, 251, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 206, 160, 4, 0, 193, 2, 53, 251, 193, 4, 53, 251, 193, 6, 53, 251, 193, 8, 53, 251, 193, 10, 53, 251, 193, 12, 53, 251, 193, 14, 53, 251, 193, 16, 53, 251, 193, 18, 53, 251, 193, 20, 53, 251, 193, 22, 53, 251, 193, 24, 53, 251, 193, 226, 91, 252, 193, 36, 53, 251, 193, 38, 53, 251, 193, 40, 53, 251, 193, 42, 53, 251, 193, 44, 53, 251, 193, 58, 53, 251, 193, 46, 53, 251, 193, 48, 53, 251, 193, 50, 53, 251, 193, 62, 53, 251, 193, 88, 53, 251, 193, 108, 53, 251, 193, 142, 53, 251, 193, 194, 53, 251, 193, 232, 53, 251, 193, 234, 53, 251, 193, 2, 54, 251, 193, 28, 54, 251, 193, 70, 54, 251, 193, 72, 54, 251, 193, 74, 54, 251, 193, 108, 54, 251, 193, 110, 54, 251, 193, 120, 54, 251, 193, 160, 54, 251, 193, 210, 54, 251, 193, 250, 54, 251, 193, 252, 54, 251, 193, 254, 54, 251, 193, 2, 55, 251, 193, 4, 55, 251, 193, 42, 55, 251, 193, 94, 55, 251, 193, 96, 55, 251, 193, 138, 55, 251, 193, 140, 55, 251, 193, 142, 55, 251, 193, 144, 55, 251, 193, 68, 53, 251, 193, 70, 53, 251, 193, 72, 53, 251, 193, 74, 53, 251, 193, 76, 53, 251, 193, 78, 53, 251, 193, 80, 53, 251, 193, 84, 53, 251, 193, 178, 20, 252, 193, 92, 53, 251, 193, 94, 53, 251, 193, 96, 53, 251, 193, 98, 53, 251, 193, 100, 53, 251, 193, 102, 53, 251, 193, 106, 53, 251, 193, 112, 53, 251, 193, 116, 53, 251, 193, 118, 53, 251, 193, 122, 53, 251, 193, 124, 53, 251, 193, 126, 53, 251, 193, 128, 53, 251, 193, 138, 53, 251, 193, 140, 53, 251, 193, 150, 53, 251, 193, 152, 53, 251, 193, 154, 53, 251, 193, 156, 53, 251, 193, 158, 53, 251, 193, 166, 53, 251, 193, 168, 53, 251, 193, 196, 53, 251, 193, 198, 53, 251, 193, 200, 53, 251, 193, 14, 54, 251, 193, 16, 54, 251, 193, 32, 54, 251, 193, 78, 54, 251, 193, 90, 54, 251, 193, 122, 54, 251, 193, 124, 54, 251, 193, 128, 54, 251, 193, 130, 54, 251, 193, 132, 54, 251, 193, 134, 54, 251, 193, 136, 54, 251, 193, 162, 54, 251, 193, 164, 54, 251, 193, 166, 54, 251, 193, 188, 54, 251, 193, 228, 54, 251, 193, 14, 55, 251, 193, 16, 55, 251, 193, 18, 55, 251, 193, 20, 55, 251, 193, 22, 55, 251, 193, 56, 55, 251, 193, 58, 55, 251, 193, 60, 55, 251, 193, 62, 55, 251, 193, 64, 55, 251, 193, 108, 55, 251, 193, 110, 55, 251, 193, 8, 54, 251, 193, 10, 54, 251, 193, 12, 54, 251, 193, 30, 54, 251, 193, 76, 54, 251, 193, 126, 54, 251, 193, 184, 54, 251, 193, 212, 54, 251, 193, 214, 54, 251, 193, 44, 55, 251, 193, 46, 55, 251, 193, 48, 55, 251, 193, 98, 55, 251, 193, 100, 55, 251, 193, 102, 55, 251, 193, 146, 55, 251, 193, 148, 55, 251, 193, 190, 55, 251, 193, 192, 55, 251, 193, 228, 55, 251, 193, 230, 55, 251, 193, 40, 56, 251, 193, 42, 56, 251, 193, 56, 56, 251, 193, 222, 56, 251, 193, 224, 56, 251, 193, 112, 57, 251, 193, 114, 57, 251, 193, 116, 57, 251, 193, 118, 57, 251, 193, 120, 57, 251, 193, 122, 57, 251, 193, 36, 54, 251, 193, 38, 54, 251, 193, 40, 54, 251, 193, 42, 54, 251, 193, 44, 54, 251, 193, 46, 54, 251, 193, 54, 54, 251, 193, 56, 54, 251, 193, 58, 54, 251, 193, 60, 54, 251, 193, 62, 54, 251, 193, 64, 54, 251, 193, 66, 54, 251, 193, 82, 54, 251, 193, 84, 54, 251, 193, 86, 54, 251, 193, 88, 54, 251, 193, 94, 54, 251, 193, 96, 54, 251, 193, 100, 54, 251, 193, 112, 54, 251, 193, 102, 54, 251, 193, 104, 54, 251, 193, 106, 54, 251, 193, 114, 54, 251, 193, 116, 54, 251, 193, 142, 54, 251, 193, 144, 54, 251, 193, 146, 54, 251, 193, 148, 54, 251, 193, 150, 54, 251, 193, 152, 54, 251, 206, 224, 72, 0, 206, 224, 72, 0, 206, 224, 72, 0, 206, 224, 72, 0, 206, 224, 72, 0, 193, 48, 54, 251, 193, 16, 194, 252, 193, 18, 194, 252, 193, 20, 194, 252, 193, 22, 194, 252, 193, 24, 194, 252, 193, 26, 194, 252, 193, 28, 194, 252, 193, 30, 194, 252, 193, 32, 194, 252, 193, 34, 194, 252, 193, 36, 194, 252, 193, 38, 194, 252, 193, 40, 194, 252, 193, 42, 194, 252, 193, 44, 194, 252, 193, 46, 194, 252, 193, 48, 194, 252, 193, 68, 194, 252, 193, 70, 194, 252, 193, 72, 194, 252, 193, 74, 194, 252, 193, 76, 194, 252, 193, 78, 194, 252, 193, 80, 194, 252, 193, 82, 194, 252, 193, 84, 194, 252, 193, 68, 54, 251, 193, 98, 54, 251, 193, 118, 54, 251, 193, 156, 54, 251, 193, 158, 54, 251, 193, 208, 54, 251, 193, 248, 54, 251, 193, 38, 55, 251, 193, 40, 55, 251, 193, 90, 55, 251, 193, 92, 55, 251, 193, 134, 55, 251, 193, 136, 55, 251, 193, 220, 55, 251, 193, 222, 55, 251, 193, 8, 56, 251, 193, 34, 56, 251, 193, 72, 56, 251, 193, 216, 56, 251, 193, 170, 58, 251, 193, 172, 58, 251, 193, 174, 58, 251, 193, 4, 60, 251, 193, 6, 60, 251, 193, 8, 60, 251, 193, 10, 60, 251, 206, 128, 5, 0, 206, 128, 5, 0, 206, 128, 5, 0, 206, 128, 5, 0, 206, 128, 5, 0, 206, 128, 5, 0, 193, 154, 54, 251, 193, 168, 54, 251, 193, 170, 54, 251, 193, 172, 54, 251, 193, 174, 54, 251, 193, 176, 54, 251, 193, 178, 54, 251, 193, 180, 54, 251, 193, 182, 54, 251, 193, 192, 54, 251, 193, 194, 54, 251, 193, 196, 54, 251, 193, 198, 54, 251, 193, 200, 54, 251, 193, 202, 54, 251, 193, 204, 54, 251, 193, 206, 54, 251, 193, 232, 54, 251, 193, 234, 54, 251, 193, 236, 54, 251, 193, 238, 54, 251, 193, 240, 54, 251, 193, 242, 54, 251, 193, 244, 54, 251, 193, 246, 54, 251, 193, 26, 55, 251, 193, 28, 55, 251, 193, 30, 55, 251, 193, 32, 55, 251, 193, 34, 55, 251, 193, 36, 55, 251, 193, 70, 55, 251, 193, 182, 55, 251, 193, 184, 55, 251, 193, 186, 55, 251, 193, 188, 55, 251, 193, 224, 55, 251, 193, 226, 55, 251, 193, 10, 56, 251, 193, 12, 56, 251, 193, 36, 56, 251, 193, 38, 56, 251, 193, 54, 56, 251, 193, 74, 56, 251, 193, 76, 56, 251, 193, 84, 56, 251, 193, 86, 56, 251, 193, 126, 56, 251, 193, 218, 56, 251, 193, 220, 56, 251, 193, 100, 57, 251, 193, 102, 57, 251, 193, 104, 57, 251, 193, 106, 57, 251, 193, 108, 57, 251, 193, 110, 57, 251, 193, 176, 58, 251, 193, 178, 58, 251, 193, 180, 58, 251, 193, 182, 58, 251, 193, 184, 58, 251, 193, 186, 58, 251, 193, 188, 58, 251, 193, 12, 60, 251, 193, 194, 55, 251, 193, 196, 55, 251, 193, 232, 55, 251, 193, 234, 55, 251, 193, 14, 56, 251, 193, 44, 56, 251, 193, 98, 56, 251, 193, 226, 56, 251, 193, 130, 57, 251, 193, 132, 57, 251, 193, 212, 58, 251, 193, 214, 58, 251, 193, 216, 58, 251, 193, 218, 58, 251, 193, 68, 60, 251, 193, 70, 60, 251, 193, 72, 60, 251, 193, 74, 60, 251, 193, 76, 60, 251, 193, 78, 60, 251, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 193, 198, 55, 251, 193, 16, 56, 251, 193, 144, 50, 253, 193, 182, 56, 251, 193, 184, 56, 251, 193, 186, 56, 251, 193, 24, 57, 251, 193, 26, 57, 251, 193, 28, 57, 251, 193, 18, 58, 251, 193, 20, 58, 251, 193, 22, 58, 251, 193, 24, 58, 251, 193, 26, 58, 251, 193, 28, 58, 251, 193, 30, 58, 251, 193, 32, 58, 251, 193, 34, 58, 251, 193, 36, 58, 251, 193, 38, 58, 251, 193, 40, 58, 251, 193, 42, 58, 251, 193, 44, 58, 251, 193, 96, 59, 251, 193, 98, 59, 251, 193, 100, 59, 251, 193, 102, 59, 251, 193, 104, 59, 251, 193, 106, 59, 251, 193, 108, 59, 251, 193, 110, 59, 251, 193, 112, 59, 251, 193, 236, 55, 251, 193, 58, 56, 251, 193, 88, 56, 251, 193, 112, 56, 251, 193, 128, 56, 251, 193, 130, 56, 251, 193, 132, 56, 251, 193, 134, 56, 251, 193, 136, 56, 251, 193, 138, 56, 251, 193, 140, 56, 251, 193, 142, 56, 251, 193, 144, 56, 251, 193, 146, 56, 251, 193, 148, 56, 251, 193, 150, 56, 251, 193, 152, 56, 251, 193, 154, 56, 251, 193, 156, 56, 251, 193, 158, 56, 251, 193, 160, 56, 251, 193, 162, 56, 251, 193, 164, 56, 251, 193, 166, 56, 251, 193, 168, 56, 251, 193, 170, 56, 251, 193, 172, 56, 251, 193, 174, 56, 251, 193, 176, 56, 251, 193, 178, 56, 251, 193, 180, 56, 251, 193, 228, 56, 251, 193, 150, 35, 251, 193, 212, 155, 251, 193, 206, 193, 251, 193, 236, 171, 252, 193, 200, 121, 251, 193, 114, 25, 252, 193, 60, 229, 251, 193, 206, 73, 253, 193, 118, 15, 253, 193, 174, 126, 253, 193, 42, 25, 253, 193, 252, 157, 251, 193, 6, 26, 251, 193, 160, 69, 251, 193, 238, 55, 251, 193, 192, 89, 251, 193, 182, 93, 251, 193, 118, 141, 251, 193, 96, 226, 251, 193, 232, 245, 251, 193, 110, 248, 251, 193, 130, 33, 251, 193, 54, 75, 252, 193, 12, 111, 252, 193, 216, 134, 252, 193, 160, 135, 252, 193, 36, 136, 252, 193, 246, 137, 252, 193, 234, 139, 253, 193, 186, 170, 252, 193, 136, 195, 252, 193, 178, 243, 252, 193, 26, 56, 251, 193, 28, 56, 251, 193, 30, 56, 251, 193, 46, 56, 251, 193, 48, 56, 251, 193, 50, 56, 251, 193, 52, 56, 251, 193, 60, 56, 251, 193, 62, 56, 251, 193, 64, 56, 251, 193, 66, 56, 251, 193, 68, 56, 251, 193, 70, 56, 251, 193, 78, 56, 251, 193, 80, 56, 251, 193, 82, 56, 251, 193, 90, 56, 251, 193, 94, 56, 251, 193, 96, 56, 251, 193, 100, 56, 251, 193, 102, 56, 251, 193, 104, 56, 251, 193, 106, 56, 251, 193, 108, 56, 251, 193, 110, 56, 251, 193, 114, 56, 251, 193, 116, 56, 251, 193, 118, 56, 251, 193, 120, 56, 251, 193, 122, 56, 251, 193, 124, 56, 251, 193, 192, 56, 251, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 193, 32, 56, 251, 193, 56, 126, 251, 193, 234, 98, 251, 193, 92, 56, 251, 193, 236, 98, 251, 193, 58, 99, 251, 193, 60, 99, 251, 193, 62, 99, 251, 193, 104, 99, 251, 193, 166, 99, 251, 193, 168, 99, 251, 193, 182, 99, 251, 193, 202, 99, 251, 193, 220, 99, 251, 193, 222, 99, 251, 193, 224, 99, 251, 193, 228, 99, 251, 193, 232, 99, 251, 193, 234, 99, 251, 193, 236, 99, 251, 193, 238, 99, 251, 193, 240, 99, 251, 193, 252, 99, 251, 193, 254, 99, 251, 193, 12, 100, 251, 193, 32, 100, 251, 193, 34, 100, 251, 193, 44, 100, 251, 193, 46, 100, 251, 193, 48, 100, 251, 193, 62, 100, 251, 193, 80, 100, 251, 193, 82, 100, 251, 193, 104, 100, 251, 193, 194, 56, 251, 193, 196, 56, 251, 193, 198, 56, 251, 193, 200, 56, 251, 193, 202, 56, 251, 193, 204, 56, 251, 193, 206, 56, 251, 193, 208, 56, 251, 193, 210, 56, 251, 193, 212, 56, 251, 193, 214, 56, 251, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 193, 230, 56, 251, 193, 232, 56, 251, 193, 234, 56, 251, 193, 236, 56, 251, 193, 238, 56, 251, 193, 240, 56, 251, 193, 242, 56, 251, 193, 244, 56, 251, 193, 246, 56, 251, 193, 248, 56, 251, 193, 250, 56, 251, 193, 252, 56, 251, 193, 254, 56, 251, 193, 2, 57, 251, 193, 4, 57, 251, 193, 6, 57, 251, 193, 8, 57, 251, 193, 10, 57, 251, 193, 12, 57, 251, 193, 14, 57, 251, 193, 16, 57, 251, 193, 18, 57, 251, 193, 20, 57, 251, 193, 22, 57, 251, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 224, 4, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 193, 124, 57, 251, 193, 126, 57, 251, 193, 128, 57, 251, 193, 190, 58, 251, 193, 192, 58, 251, 193, 194, 58, 251, 193, 196, 58, 251, 193, 198, 58, 251, 193, 200, 58, 251, 193, 202, 58, 251, 193, 204, 58, 251, 193, 206, 58, 251, 193, 208, 58, 251, 193, 210, 58, 251, 193, 36, 60, 251, 193, 38, 60, 251, 193, 40, 60, 251, 193, 42, 60, 251, 193, 44, 60, 251, 193, 46, 60, 251, 193, 48, 60, 251, 193, 50, 60, 251, 193, 52, 60, 251, 193, 54, 60, 251, 193, 56, 60, 251, 193, 58, 60, 251, 193, 60, 60, 251, 193, 62, 60, 251, 193, 64, 60, 251, 193, 66, 60, 251, 193, 234, 61, 251, 193, 236, 61, 251, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 206, 0, 5, 0, 193, 220, 58, 251, 193, 222, 58, 251, 193, 224, 58, 251, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 32, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 193, 226, 58, 251, 193, 228, 58, 251, 193, 230, 58, 251, 193, 232, 58, 251, 193, 234, 58, 251, 193, 236, 58, 251, 193, 238, 58, 251, 193, 240, 58, 251, 193, 242, 58, 251, 193, 244, 58, 251, 193, 246, 58, 251, 193, 248, 58, 251, 193, 250, 58, 251, 193, 252, 58, 251, 193, 254, 58, 251, 193, 2, 59, 251, 193, 4, 59, 251, 193, 6, 59, 251, 193, 8, 59, 251, 193, 10, 59, 251, 193, 12, 59, 251, 193, 14, 59, 251, 193, 16, 59, 251, 193, 18, 59, 251, 193, 20, 59, 251, 193, 22, 59, 251, 193, 24, 59, 251, 193, 26, 59, 251, 193, 28, 59, 251, 193, 30, 59, 251, 193, 32, 59, 251, 193, 34, 59, 251, 193, 36, 59, 251, 193, 38, 59, 251, 193, 40, 59, 251, 193, 42, 59, 251, 193, 44, 59, 251, 193, 46, 59, 251, 193, 48, 59, 251, 193, 50, 59, 251, 193, 52, 59, 251, 193, 54, 59, 251, 193, 56, 59, 251, 193, 58, 59, 251, 193, 60, 59, 251, 193, 62, 59, 251, 193, 64, 59, 251, 193, 66, 59, 251, 193, 68, 59, 251, 193, 70, 59, 251, 193, 72, 59, 251, 193, 74, 59, 251, 193, 76, 59, 251, 193, 78, 59, 251, 193, 80, 59, 251, 193, 82, 59, 251, 193, 84, 59, 251, 193, 86, 59, 251, 193, 88, 59, 251, 193, 90, 59, 251, 193, 92, 59, 251, 193, 94, 59, 251, 193, 80, 60, 251, 193, 82, 60, 251, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 64, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 193, 14, 60, 251, 193, 16, 60, 251, 193, 18, 60, 251, 193, 20, 60, 251, 193, 22, 60, 251, 193, 24, 60, 251, 193, 26, 60, 251, 193, 28, 60, 251, 193, 30, 60, 251, 193, 32, 60, 251, 193, 34, 60, 251, 193, 198, 61, 251, 193, 200, 61, 251, 193, 202, 61, 251, 193, 204, 61, 251, 193, 206, 61, 251, 193, 208, 61, 251, 193, 210, 61, 251, 193, 212, 61, 251, 193, 214, 61, 251, 193, 216, 61, 251, 193, 218, 61, 251, 193, 220, 61, 251, 193, 222, 61, 251, 193, 224, 61, 251, 193, 226, 61, 251, 193, 228, 61, 251, 193, 230, 61, 251, 193, 232, 61, 251, 193, 242, 63, 251, 193, 244, 63, 251, 193, 246, 63, 251, 193, 84, 60, 251, 193, 86, 60, 251, 193, 88, 60, 251, 193, 90, 60, 251, 193, 92, 60, 251, 193, 94, 60, 251, 193, 96, 60, 251, 193, 98, 60, 251, 193, 100, 60, 251, 193, 102, 60, 251, 193, 104, 60, 251, 193, 106, 60, 251, 193, 108, 60, 251, 193, 110, 60, 251, 193, 112, 60, 251, 193, 114, 60, 251, 193, 116, 60, 251, 193, 118, 60, 251, 193, 120, 60, 251, 193, 122, 60, 251, 193, 124, 60, 251, 193, 126, 60, 251, 193, 128, 60, 251, 193, 130, 60, 251, 193, 132, 60, 251, 193, 134, 60, 251, 193, 136, 60, 251, 193, 138, 60, 251, 193, 140, 60, 251, 193, 142, 60, 251, 193, 144, 60, 251, 193, 146, 60, 251, 193, 148, 60, 251, 193, 150, 60, 251, 193, 152, 60, 251, 193, 154, 60, 251, 193, 156, 60, 251, 193, 158, 60, 251, 193, 160, 60, 251, 193, 162, 60, 251, 193, 164, 60, 251, 193, 166, 60, 251, 193, 168, 60, 251, 193, 170, 60, 251, 193, 172, 60, 251, 193, 174, 60, 251, 193, 176, 60, 251, 193, 178, 60, 251, 193, 180, 60, 251, 193, 182, 60, 251, 193, 184, 60, 251, 193, 186, 60, 251, 193, 188, 60, 251, 193, 190, 60, 251, 193, 192, 60, 251, 193, 194, 60, 251, 193, 196, 60, 251, 193, 198, 60, 251, 193, 200, 60, 251, 193, 202, 60, 251, 193, 204, 60, 251, 193, 206, 60, 251, 193, 208, 60, 251, 193, 210, 60, 251, 193, 212, 60, 251, 193, 214, 60, 251, 193, 216, 60, 251, 193, 218, 60, 251, 193, 220, 60, 251, 193, 222, 60, 251, 193, 224, 60, 251, 193, 226, 60, 251, 193, 228, 60, 251, 193, 230, 60, 251, 193, 232, 60, 251, 193, 234, 60, 251, 193, 236, 60, 251, 193, 238, 60, 251, 193, 240, 60, 251, 193, 202, 62, 251, 193, 204, 62, 251, 193, 206, 62, 251, 193, 208, 62, 251, 193, 210, 62, 251, 193, 212, 62, 251, 193, 214, 62, 251, 193, 216, 62, 251, 193, 218, 62, 251, 193, 220, 62, 251, 193, 222, 62, 251, 193, 224, 62, 251, 193, 226, 62, 251, 193, 228, 62, 251, 193, 230, 62, 251, 193, 232, 62, 251, 193, 6, 65, 251, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 96, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 128, 5, 0, 206, 128, 5, 0, 206, 128, 5, 0, 206, 128, 5, 0, 193, 220, 63, 251, 193, 222, 63, 251, 193, 224, 63, 251, 193, 226, 63, 251, 193, 228, 63, 251, 193, 230, 63, 251, 193, 232, 63, 251, 193, 234, 63, 251, 193, 236, 63, 251, 193, 238, 63, 251, 193, 240, 63, 251, 193, 88, 66, 251, 193, 90, 66, 251, 193, 92, 66, 251, 193, 94, 66, 251, 193, 96, 66, 251, 193, 98, 66, 251, 193, 100, 66, 251, 193, 102, 66, 251, 193, 104, 66, 251, 193, 106, 66, 251, 193, 108, 66, 251, 193, 110, 66, 251, 193, 112, 66, 251, 206, 64, 7, 0, 206, 64, 7, 0, 206, 64, 7, 0, 206, 64, 7, 0, 193, 238, 61, 251, 193, 240, 61, 251, 193, 242, 61, 251, 193, 244, 61, 251, 193, 246, 61, 251, 193, 248, 61, 251, 193, 250, 61, 251, 193, 252, 61, 251, 193, 254, 61, 251, 193, 2, 62, 251, 193, 4, 62, 251, 193, 6, 62, 251, 193, 8, 62, 251, 193, 10, 62, 251, 193, 12, 62, 251, 193, 14, 62, 251, 193, 16, 62, 251, 193, 18, 62, 251, 193, 20, 62, 251, 193, 22, 62, 251, 193, 24, 62, 251, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 206, 160, 5, 0, 193, 76, 64, 251, 193, 78, 64, 251, 193, 80, 64, 251, 193, 82, 64, 251, 193, 84, 64, 251, 193, 86, 64, 251, 193, 88, 64, 251, 193, 90, 64, 251, 193, 92, 64, 251, 193, 94, 64, 251, 193, 96, 64, 251, 193, 98, 64, 251, 193, 100, 64, 251, 193, 246, 66, 251, 193, 248, 66, 251, 193, 250, 66, 251, 193, 252, 66, 251, 193, 254, 66, 251, 193, 2, 67, 251, 193, 4, 67, 251, 193, 6, 67, 251, 193, 8, 67, 251, 193, 10, 67, 251, 193, 12, 67, 251, 193, 14, 67, 251, 193, 16, 67, 251, 193, 18, 67, 251, 193, 120, 69, 251, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 192, 5, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 206, 224, 5, 0, 193, 52, 65, 251, 193, 94, 63, 251, 193, 96, 63, 251, 193, 98, 63, 251, 193, 100, 63, 251, 193, 102, 63, 251, 193, 248, 93, 252, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 206, 0, 6, 0, 193, 158, 81, 252, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 193, 248, 63, 251, 193, 250, 63, 251, 193, 252, 63, 251, 193, 254, 63, 251, 193, 2, 64, 251, 193, 4, 64, 251, 193, 6, 64, 251, 193, 8, 64, 251, 193, 10, 64, 251, 193, 12, 64, 251, 193, 14, 64, 251, 193, 16, 64, 251, 193, 18, 64, 251, 193, 20, 64, 251, 193, 22, 64, 251, 193, 24, 64, 251, 193, 26, 64, 251, 193, 28, 64, 251, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 32, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 206, 64, 6, 0, 193, 20, 67, 251, 193, 200, 64, 251, 193, 202, 64, 251, 193, 204, 64, 251, 193, 206, 64, 251, 193, 208, 64, 251, 193, 30, 214, 251, 193, 210, 64, 251, 193, 212, 64, 251, 193, 214, 64, 251, 193, 216, 64, 251, 193, 218, 64, 251, 193, 220, 64, 251, 193, 222, 64, 251, 193, 224, 64, 251, 193, 226, 64, 251, 193, 228, 64, 251, 193, 230, 64, 251, 193, 22, 67, 251, 193, 232, 64, 251, 193, 234, 64, 251, 193, 236, 64, 251, 193, 238, 64, 251, 193, 240, 64, 251, 193, 242, 64, 251, 193, 244, 64, 251, 193, 246, 64, 251, 193, 248, 64, 251, 193, 250, 64, 251, 193, 252, 64, 251, 193, 254, 64, 251, 193, 2, 65, 251, 193, 4, 65, 251, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 193, 8, 65, 251, 193, 10, 65, 251, 193, 12, 65, 251, 193, 14, 65, 251, 193, 16, 65, 251, 193, 18, 65, 251, 193, 20, 65, 251, 193, 22, 65, 251, 193, 24, 65, 251, 193, 26, 65, 251, 193, 28, 65, 251, 193, 30, 65, 251, 193, 32, 65, 251, 193, 34, 65, 251, 193, 36, 65, 251, 193, 38, 65, 251, 193, 40, 65, 251, 193, 42, 65, 251, 193, 44, 65, 251, 193, 46, 65, 251, 193, 48, 65, 251, 193, 50, 65, 251, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 206, 96, 6, 0, 193, 218, 67, 251, 193, 220, 67, 251, 193, 222, 67, 251, 193, 224, 67, 251, 193, 226, 67, 251, 193, 228, 67, 251, 193, 230, 67, 251, 193, 232, 67, 251, 193, 234, 67, 251, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 206, 128, 6, 0, 193, 252, 68, 251, 193, 254, 68, 251, 193, 2, 69, 251, 193, 4, 69, 251, 193, 6, 69, 251, 193, 8, 69, 251, 193, 10, 69, 251, 193, 12, 69, 251, 193, 14, 69, 251, 193, 16, 69, 251, 193, 18, 69, 251, 193, 20, 69, 251, 193, 22, 69, 251, 193, 24, 69, 251, 193, 26, 69, 251, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 160, 6, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 206, 192, 6, 0, 193, 156, 69, 251, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 206, 224, 6, 0, 193, 158, 69, 251, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 193, 160, 67, 251, 193, 130, 74, 251, 193, 132, 74, 251, 193, 134, 74, 251, 193, 136, 74, 251, 193, 138, 74, 251, 193, 140, 74, 251, 193, 142, 74, 251, 193, 144, 74, 251, 193, 146, 74, 251, 193, 148, 74, 251, 193, 150, 74, 251, 193, 98, 72, 251, 193, 152, 74, 251, 193, 100, 72, 251, 193, 154, 74, 251, 193, 22, 65, 252, 193, 156, 74, 251, 193, 158, 74, 251, 193, 160, 74, 251, 193, 162, 74, 251, 193, 164, 74, 251, 193, 166, 74, 251, 193, 168, 74, 251, 193, 170, 74, 251, 193, 30, 76, 251, 193, 32, 76, 251, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 206, 0, 7, 0, 193, 14, 70, 251, 193, 16, 70, 251, 193, 18, 70, 251, 193, 20, 70, 251, 193, 22, 70, 251, 193, 24, 70, 251, 193, 26, 70, 251, 193, 28, 70, 251, 193, 30, 70, 251, 193, 32, 70, 251, 193, 34, 70, 251, 193, 36, 70, 251, 193, 38, 70, 251, 193, 40, 70, 251, 193, 236, 67, 251, 193, 238, 67, 251, 193, 240, 67, 251, 193, 242, 67, 251, 193, 244, 67, 251, 193, 246, 67, 251, 193, 248, 67, 251, 193, 250, 67, 251, 193, 252, 67, 251, 193, 254, 67, 251, 193, 2, 68, 251, 193, 4, 68, 251, 193, 6, 68, 251, 193, 8, 68, 251, 193, 10, 68, 251, 193, 12, 68, 251, 193, 14, 68, 251, 193, 16, 68, 251, 193, 18, 68, 251, 193, 20, 68, 251, 193, 22, 68, 251, 193, 24, 68, 251, 193, 26, 68, 251, 193, 28, 68, 251, 193, 30, 68, 251, 193, 32, 68, 251, 193, 34, 68, 251, 193, 36, 68, 251, 193, 38, 68, 251, 193, 40, 68, 251, 193, 42, 68, 251, 193, 44, 68, 251, 193, 46, 68, 251, 193, 48, 68, 251, 193, 46, 70, 251, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 206, 32, 7, 0, 193, 48, 70, 251, 193, 50, 70, 251, 193, 52, 70, 251, 193, 54, 70, 251, 193, 56, 70, 251, 193, 216, 68, 251, 193, 218, 68, 251, 193, 220, 68, 251, 193, 222, 68, 251, 193, 58, 70, 251, 206, 64, 7, 0, 206, 64, 7, 0, 206, 64, 7, 0, 206, 64, 7, 0, 206, 64, 7, 0, 206, 64, 7, 0, 206, 64, 7, 0, 206, 64, 7, 0, 206, 64, 7, 0, 206, 64, 7, 0, 193, 96, 71, 251, 193, 98, 71, 251, 193, 100, 71, 251, 193, 102, 71, 251, 193, 104, 71, 251, 193, 106, 71, 251, 193, 108, 71, 251, 193, 110, 71, 251, 193, 112, 71, 251, 193, 114, 71, 251, 193, 116, 71, 251, 193, 118, 71, 251, 193, 120, 71, 251, 193, 122, 71, 251, 206, 0, 8, 0, 206, 0, 8, 0, 206, 0, 8, 0, 206, 0, 8, 0, 206, 0, 8, 0, 206, 0, 8, 0, 206, 0, 8, 0, 206, 0, 8, 0, 193, 28, 69, 251, 193, 30, 69, 251, 193, 32, 69, 251, 193, 124, 71, 251, 193, 126, 71, 251, 193, 128, 71, 251, 193, 130, 71, 251, 193, 132, 71, 251, 193, 134, 71, 251, 193, 136, 71, 251, 193, 138, 71, 251, 193, 140, 71, 251, 193, 142, 71, 251, 193, 144, 71, 251, 193, 146, 71, 251, 193, 148, 71, 251, 193, 150, 71, 251, 193, 152, 71, 251, 193, 154, 71, 251, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 96, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 193, 122, 69, 251, 193, 124, 69, 251, 193, 126, 69, 251, 193, 128, 69, 251, 193, 130, 69, 251, 193, 132, 69, 251, 193, 134, 69, 251, 193, 136, 69, 251, 193, 138, 69, 251, 193, 140, 69, 251, 193, 142, 69, 251, 193, 144, 69, 251, 193, 146, 69, 251, 193, 148, 69, 251, 193, 150, 69, 251, 193, 152, 69, 251, 193, 154, 69, 251, 193, 228, 71, 251, 193, 230, 71, 251, 193, 232, 71, 251, 193, 234, 71, 251, 193, 236, 71, 251, 193, 238, 71, 251, 193, 240, 71, 251, 193, 242, 71, 251, 193, 244, 71, 251, 193, 246, 71, 251, 193, 248, 71, 251, 193, 250, 71, 251, 193, 252, 71, 251, 193, 254, 71, 251, 193, 48, 74, 251, 193, 160, 69, 251, 193, 162, 69, 251, 193, 164, 69, 251, 193, 166, 69, 251, 193, 168, 69, 251, 193, 170, 69, 251, 193, 172, 69, 251, 193, 174, 69, 251, 193, 176, 69, 251, 193, 178, 69, 251, 193, 180, 69, 251, 193, 182, 69, 251, 193, 184, 69, 251, 193, 186, 69, 251, 193, 188, 69, 251, 193, 190, 69, 251, 193, 192, 69, 251, 193, 194, 69, 251, 193, 196, 69, 251, 193, 198, 69, 251, 193, 200, 69, 251, 193, 202, 69, 251, 193, 204, 69, 251, 193, 206, 69, 251, 193, 208, 69, 251, 193, 210, 69, 251, 193, 212, 69, 251, 193, 214, 69, 251, 193, 216, 69, 251, 193, 218, 69, 251, 193, 220, 69, 251, 193, 222, 69, 251, 193, 224, 69, 251, 193, 226, 69, 251, 193, 228, 69, 251, 193, 230, 69, 251, 193, 232, 69, 251, 193, 234, 69, 251, 193, 236, 69, 251, 193, 238, 69, 251, 193, 240, 69, 251, 193, 242, 69, 251, 193, 244, 69, 251, 193, 246, 69, 251, 193, 248, 69, 251, 193, 250, 69, 251, 193, 252, 69, 251, 193, 254, 69, 251, 193, 2, 70, 251, 193, 4, 70, 251, 193, 6, 70, 251, 193, 8, 70, 251, 193, 10, 70, 251, 193, 12, 70, 251, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 193, 42, 70, 251, 193, 44, 70, 251, 193, 102, 72, 251, 193, 104, 72, 251, 193, 106, 72, 251, 193, 108, 72, 251, 193, 110, 72, 251, 193, 112, 72, 251, 193, 114, 72, 251, 193, 116, 72, 251, 193, 118, 72, 251, 193, 120, 72, 251, 193, 122, 72, 251, 193, 124, 72, 251, 193, 126, 72, 251, 193, 128, 72, 251, 193, 130, 72, 251, 193, 174, 74, 251, 193, 176, 74, 251, 193, 178, 74, 251, 193, 180, 74, 251, 193, 182, 74, 251, 193, 184, 74, 251, 193, 186, 74, 251, 193, 188, 74, 251, 193, 190, 74, 251, 193, 192, 74, 251, 206, 160, 8, 0, 206, 160, 8, 0, 206, 160, 8, 0, 206, 160, 8, 0, 206, 160, 8, 0, 193, 60, 70, 251, 193, 62, 70, 251, 193, 64, 70, 251, 193, 66, 70, 251, 193, 68, 70, 251, 193, 70, 70, 251, 193, 72, 70, 251, 193, 74, 70, 251, 193, 76, 70, 251, 193, 78, 70, 251, 193, 80, 70, 251, 193, 82, 70, 251, 193, 84, 70, 251, 193, 86, 70, 251, 193, 88, 70, 251, 193, 90, 70, 251, 193, 92, 70, 251, 193, 94, 70, 251, 193, 96, 70, 251, 193, 98, 70, 251, 193, 100, 70, 251, 193, 102, 70, 251, 193, 104, 70, 251, 193, 106, 70, 251, 193, 108, 70, 251, 193, 110, 70, 251, 193, 112, 70, 251, 193, 114, 70, 251, 193, 116, 70, 251, 193, 118, 70, 251, 193, 120, 70, 251, 193, 122, 70, 251, 193, 124, 70, 251, 193, 126, 70, 251, 193, 128, 70, 251, 193, 130, 70, 251, 193, 132, 70, 251, 193, 134, 70, 251, 193, 136, 70, 251, 193, 138, 70, 251, 193, 140, 70, 251, 193, 142, 70, 251, 193, 144, 70, 251, 193, 146, 70, 251, 193, 148, 70, 251, 193, 150, 70, 251, 193, 152, 70, 251, 193, 154, 70, 251, 193, 200, 74, 251, 193, 202, 74, 251, 193, 204, 74, 251, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 128, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 206, 160, 7, 0, 193, 232, 73, 251, 193, 234, 73, 251, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 206, 192, 7, 0, 193, 68, 74, 251, 193, 70, 74, 251, 193, 84, 72, 251, 193, 86, 72, 251, 193, 88, 72, 251, 193, 90, 72, 251, 193, 92, 72, 251, 193, 94, 72, 251, 193, 96, 72, 251, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 64, 8, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 224, 7, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 0, 8, 0, 206, 0, 8, 0, 206, 0, 8, 0, 206, 0, 8, 0, 206, 0, 8, 0, 193, 158, 75, 251, 193, 160, 75, 251, 193, 162, 75, 251, 193, 164, 75, 251, 193, 166, 75, 251, 193, 168, 75, 251, 193, 170, 75, 251, 193, 172, 75, 251, 193, 174, 75, 251, 193, 92, 77, 251, 193, 94, 77, 251, 193, 96, 77, 251, 193, 98, 77, 251, 193, 100, 77, 251, 193, 102, 77, 251, 193, 104, 77, 251, 193, 174, 78, 251, 193, 176, 78, 251, 193, 178, 78, 251, 193, 180, 78, 251, 193, 218, 79, 251, 193, 220, 79, 251, 193, 222, 79, 251, 193, 224, 79, 251, 193, 226, 79, 251, 193, 228, 79, 251, 193, 184, 80, 251, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 206, 32, 8, 0, 193, 178, 75, 251, 193, 180, 75, 251, 193, 182, 75, 251, 193, 184, 75, 251, 193, 186, 75, 251, 193, 188, 75, 251, 193, 190, 75, 251, 193, 192, 75, 251, 193, 194, 75, 251, 193, 196, 75, 251, 193, 198, 75, 251, 193, 200, 75, 251, 193, 202, 75, 251, 193, 204, 75, 251, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 193, 236, 73, 251, 193, 238, 73, 251, 193, 240, 73, 251, 193, 242, 73, 251, 193, 244, 73, 251, 193, 246, 73, 251, 193, 248, 73, 251, 193, 250, 73, 251, 193, 252, 73, 251, 193, 254, 73, 251, 193, 2, 74, 251, 193, 4, 74, 251, 193, 6, 74, 251, 193, 8, 74, 251, 193, 10, 74, 251, 193, 12, 74, 251, 193, 14, 74, 251, 193, 16, 74, 251, 193, 18, 74, 251, 193, 20, 74, 251, 193, 22, 74, 251, 193, 24, 74, 251, 193, 26, 74, 251, 193, 28, 74, 251, 193, 30, 74, 251, 193, 32, 74, 251, 193, 34, 74, 251, 193, 36, 74, 251, 193, 38, 74, 251, 193, 40, 74, 251, 193, 42, 74, 251, 193, 44, 74, 251, 193, 46, 74, 251, 193, 206, 75, 251, 193, 208, 75, 251, 193, 210, 75, 251, 193, 212, 75, 251, 193, 214, 75, 251, 193, 216, 75, 251, 193, 218, 75, 251, 193, 220, 75, 251, 193, 222, 75, 251, 193, 224, 75, 251, 193, 226, 75, 251, 193, 228, 75, 251, 193, 230, 75, 251, 193, 232, 75, 251, 193, 234, 75, 251, 193, 236, 75, 251, 193, 238, 75, 251, 193, 240, 75, 251, 193, 242, 75, 251, 193, 244, 75, 251, 193, 246, 75, 251, 193, 248, 75, 251, 193, 250, 75, 251, 193, 252, 75, 251, 193, 254, 75, 251, 193, 2, 76, 251, 193, 4, 76, 251, 193, 6, 76, 251, 193, 8, 76, 251, 193, 10, 76, 251, 193, 12, 76, 251, 193, 50, 74, 251, 193, 52, 74, 251, 193, 54, 74, 251, 193, 56, 74, 251, 193, 58, 74, 251, 193, 60, 74, 251, 193, 62, 74, 251, 193, 64, 74, 251, 193, 66, 74, 251, 193, 18, 76, 251, 193, 20, 76, 251, 193, 22, 76, 251, 193, 24, 76, 251, 193, 26, 76, 251, 193, 28, 76, 251, 193, 192, 77, 251, 193, 194, 77, 251, 193, 196, 77, 251, 193, 198, 77, 251, 193, 200, 77, 251, 193, 202, 77, 251, 193, 204, 77, 251, 193, 206, 77, 251, 193, 246, 78, 251, 193, 248, 78, 251, 193, 250, 78, 251, 193, 252, 78, 251, 193, 254, 78, 251, 193, 2, 79, 251, 193, 4, 79, 251, 193, 50, 80, 251, 193, 52, 80, 251, 193, 34, 76, 251, 193, 36, 76, 251, 193, 38, 76, 251, 193, 40, 76, 251, 193, 42, 76, 251, 193, 44, 76, 251, 193, 46, 76, 251, 193, 48, 76, 251, 193, 50, 76, 251, 193, 52, 76, 251, 193, 54, 76, 251, 193, 56, 76, 251, 193, 58, 76, 251, 193, 60, 76, 251, 193, 62, 76, 251, 193, 64, 76, 251, 193, 66, 76, 251, 193, 68, 76, 251, 193, 70, 76, 251, 193, 72, 76, 251, 193, 172, 74, 251, 193, 74, 76, 251, 193, 76, 76, 251, 193, 78, 76, 251, 193, 80, 76, 251, 193, 82, 76, 251, 193, 84, 76, 251, 193, 86, 76, 251, 193, 88, 76, 251, 193, 208, 77, 251, 193, 210, 77, 251, 193, 212, 77, 251, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 206, 96, 8, 0, 193, 124, 140, 252, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 128, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 193, 154, 75, 251, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 206, 224, 8, 0, 193, 156, 75, 251, 193, 44, 77, 251, 193, 46, 77, 251, 193, 48, 77, 251, 193, 50, 77, 251, 193, 52, 77, 251, 193, 54, 77, 251, 193, 56, 77, 251, 193, 58, 77, 251, 193, 60, 77, 251, 193, 62, 77, 251, 193, 64, 77, 251, 193, 66, 77, 251, 193, 68, 77, 251, 193, 70, 77, 251, 193, 72, 77, 251, 193, 74, 77, 251, 193, 76, 77, 251, 193, 78, 77, 251, 193, 80, 77, 251, 193, 82, 77, 251, 193, 84, 77, 251, 193, 86, 77, 251, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 193, 14, 76, 251, 193, 16, 76, 251, 193, 140, 77, 251, 193, 142, 77, 251, 193, 144, 77, 251, 193, 146, 77, 251, 193, 148, 77, 251, 193, 150, 77, 251, 193, 152, 77, 251, 193, 154, 77, 251, 193, 156, 77, 251, 193, 158, 77, 251, 193, 160, 77, 251, 193, 162, 77, 251, 193, 164, 77, 251, 193, 166, 77, 251, 193, 168, 77, 251, 193, 170, 77, 251, 193, 172, 77, 251, 193, 174, 77, 251, 193, 176, 77, 251, 193, 178, 77, 251, 193, 180, 77, 251, 193, 182, 77, 251, 193, 184, 77, 251, 193, 186, 77, 251, 193, 188, 77, 251, 193, 190, 77, 251, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 192, 8, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 193, 70, 190, 251, 193, 72, 190, 251, 193, 74, 190, 251, 193, 76, 190, 251, 193, 78, 190, 251, 193, 80, 190, 251, 193, 82, 190, 251, 193, 84, 190, 251, 193, 86, 190, 251, 193, 88, 77, 251, 193, 88, 190, 251, 193, 90, 190, 251, 193, 120, 190, 251, 193, 122, 190, 251, 193, 124, 190, 251, 193, 126, 190, 251, 193, 128, 190, 251, 206, 64, 28, 0, 206, 64, 28, 0, 206, 64, 28, 0, 206, 64, 28, 0, 206, 64, 28, 0, 206, 64, 28, 0, 206, 64, 28, 0, 206, 64, 28, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 193, 90, 77, 251, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 206, 0, 9, 0, 193, 184, 78, 251, 193, 186, 78, 251, 193, 188, 78, 251, 193, 190, 78, 251, 193, 192, 78, 251, 193, 194, 78, 251, 193, 196, 78, 251, 193, 198, 78, 251, 193, 200, 78, 251, 193, 202, 78, 251, 193, 204, 78, 251, 193, 232, 79, 251, 193, 234, 79, 251, 193, 236, 79, 251, 193, 238, 79, 251, 193, 240, 79, 251, 193, 242, 79, 251, 193, 244, 79, 251, 193, 246, 79, 251, 193, 248, 79, 251, 193, 250, 79, 251, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 193, 214, 77, 251, 193, 216, 77, 251, 193, 218, 77, 251, 193, 220, 77, 251, 193, 222, 77, 251, 193, 224, 77, 251, 193, 226, 77, 251, 193, 228, 77, 251, 193, 230, 77, 251, 193, 232, 77, 251, 193, 234, 77, 251, 193, 236, 77, 251, 193, 238, 77, 251, 193, 240, 77, 251, 193, 242, 77, 251, 193, 244, 77, 251, 193, 246, 77, 251, 193, 248, 77, 251, 193, 250, 77, 251, 193, 252, 77, 251, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 206, 96, 9, 0, 193, 60, 80, 251, 193, 62, 80, 251, 193, 64, 80, 251, 193, 66, 80, 251, 193, 68, 80, 251, 193, 70, 80, 251, 193, 72, 80, 251, 193, 74, 80, 251, 193, 76, 80, 251, 193, 78, 80, 251, 193, 80, 80, 251, 193, 82, 80, 251, 193, 22, 81, 251, 193, 24, 81, 251, 193, 26, 81, 251, 193, 28, 81, 251, 193, 30, 81, 251, 193, 32, 81, 251, 193, 34, 81, 251, 193, 36, 81, 251, 193, 38, 81, 251, 193, 254, 77, 251, 206, 32, 10, 0, 206, 32, 10, 0, 206, 32, 10, 0, 206, 32, 10, 0, 206, 32, 10, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 32, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 64, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 206, 128, 9, 0, 193, 90, 80, 251, 193, 92, 80, 251, 193, 94, 80, 251, 193, 96, 80, 251, 193, 98, 80, 251, 193, 100, 80, 251, 193, 102, 80, 251, 193, 104, 80, 251, 193, 106, 80, 251, 193, 108, 80, 251, 193, 110, 80, 251, 193, 112, 80, 251, 193, 114, 80, 251, 193, 116, 80, 251, 193, 118, 80, 251, 193, 120, 80, 251, 193, 216, 79, 251, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 206, 160, 9, 0, 193, 222, 80, 251, 193, 224, 80, 251, 193, 226, 80, 251, 193, 228, 80, 251, 193, 230, 80, 251, 193, 232, 80, 251, 193, 234, 80, 251, 193, 236, 80, 251, 193, 238, 80, 251, 193, 240, 80, 251, 193, 242, 80, 251, 193, 244, 80, 251, 193, 246, 80, 251, 193, 248, 80, 251, 193, 250, 80, 251, 193, 252, 80, 251, 193, 254, 80, 251, 193, 2, 81, 251, 193, 4, 81, 251, 193, 6, 81, 251, 193, 8, 81, 251, 193, 10, 81, 251, 193, 54, 80, 251, 193, 56, 80, 251, 193, 58, 80, 251, 193, 16, 81, 251, 193, 18, 81, 251, 193, 20, 81, 251, 193, 162, 81, 251, 193, 164, 81, 251, 193, 166, 81, 251, 193, 168, 81, 251, 193, 32, 82, 251, 193, 34, 82, 251, 193, 128, 82, 251, 193, 242, 83, 251, 193, 244, 83, 251, 193, 42, 84, 251, 193, 44, 84, 251, 193, 206, 84, 251, 193, 38, 85, 251, 193, 90, 85, 251, 193, 116, 85, 251, 193, 202, 85, 251, 193, 246, 85, 251, 193, 248, 85, 251, 193, 70, 87, 251, 193, 72, 87, 251, 193, 74, 87, 251, 193, 76, 87, 251, 193, 78, 87, 251, 193, 80, 87, 251, 193, 82, 87, 251, 193, 70, 88, 251, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 206, 192, 9, 0, 193, 10, 66, 252, 193, 170, 80, 251, 193, 172, 80, 251, 193, 174, 80, 251, 193, 176, 80, 251, 193, 178, 80, 251, 193, 180, 80, 251, 193, 182, 80, 251, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 193, 186, 80, 251, 193, 188, 80, 251, 193, 108, 81, 251, 193, 110, 81, 251, 193, 112, 81, 251, 193, 102, 82, 251, 193, 212, 82, 251, 193, 214, 82, 251, 193, 232, 83, 251, 193, 34, 84, 251, 193, 80, 84, 251, 193, 136, 84, 251, 193, 190, 84, 251, 193, 192, 84, 251, 193, 32, 85, 251, 193, 76, 85, 251, 193, 130, 85, 251, 193, 236, 85, 251, 193, 238, 85, 251, 193, 64, 86, 251, 193, 66, 86, 251, 193, 68, 86, 251, 193, 70, 86, 251, 193, 72, 86, 251, 193, 14, 87, 251, 193, 16, 87, 251, 193, 18, 87, 251, 193, 20, 87, 251, 193, 22, 87, 251, 193, 24, 87, 251, 193, 26, 87, 251, 193, 28, 87, 251, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 206, 224, 9, 0, 193, 114, 81, 251, 193, 116, 81, 251, 193, 118, 81, 251, 193, 120, 81, 251, 193, 122, 81, 251, 193, 124, 81, 251, 193, 126, 81, 251, 193, 128, 81, 251, 193, 130, 81, 251, 193, 132, 81, 251, 193, 18, 82, 251, 193, 20, 82, 251, 193, 22, 82, 251, 193, 106, 82, 251, 193, 158, 82, 251, 193, 160, 82, 251, 193, 162, 82, 251, 193, 216, 82, 251, 193, 218, 82, 251, 193, 220, 82, 251, 193, 12, 81, 251, 193, 14, 81, 251, 193, 134, 81, 251, 193, 136, 81, 251, 193, 138, 81, 251, 193, 140, 81, 251, 193, 142, 81, 251, 193, 144, 81, 251, 193, 146, 81, 251, 193, 148, 81, 251, 193, 150, 81, 251, 193, 152, 81, 251, 193, 154, 81, 251, 193, 156, 81, 251, 193, 158, 81, 251, 193, 160, 81, 251, 193, 24, 82, 251, 193, 26, 82, 251, 193, 28, 82, 251, 193, 30, 82, 251, 193, 108, 82, 251, 193, 110, 82, 251, 193, 112, 82, 251, 193, 114, 82, 251, 193, 116, 82, 251, 193, 118, 82, 251, 193, 120, 82, 251, 193, 122, 82, 251, 193, 124, 82, 251, 193, 126, 82, 251, 193, 164, 82, 251, 193, 166, 82, 251, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 0, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 32, 10, 0, 206, 32, 10, 0, 206, 32, 10, 0, 206, 32, 10, 0, 206, 32, 10, 0, 193, 36, 82, 251, 193, 38, 82, 251, 193, 40, 82, 251, 193, 42, 82, 251, 193, 44, 82, 251, 193, 46, 82, 251, 193, 48, 82, 251, 193, 130, 82, 251, 193, 190, 81, 251, 193, 50, 82, 251, 193, 132, 82, 251, 193, 134, 82, 251, 193, 180, 82, 251, 193, 182, 82, 251, 193, 184, 82, 251, 193, 234, 82, 251, 193, 236, 82, 251, 193, 54, 83, 251, 193, 76, 83, 251, 193, 84, 83, 251, 193, 86, 83, 251, 193, 88, 83, 251, 193, 90, 83, 251, 193, 92, 83, 251, 193, 114, 83, 251, 193, 116, 83, 251, 193, 118, 83, 251, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 206, 64, 10, 0, 193, 160, 100, 251, 193, 82, 195, 252, 193, 54, 82, 251, 193, 56, 82, 251, 193, 58, 82, 251, 193, 60, 82, 251, 193, 62, 82, 251, 193, 64, 82, 251, 193, 66, 82, 251, 193, 68, 82, 251, 193, 70, 82, 251, 193, 72, 82, 251, 193, 74, 82, 251, 193, 76, 82, 251, 193, 16, 82, 251, 193, 78, 82, 251, 193, 80, 82, 251, 193, 82, 82, 251, 193, 84, 82, 251, 193, 86, 82, 251, 193, 88, 82, 251, 193, 90, 82, 251, 193, 92, 82, 251, 193, 94, 82, 251, 193, 96, 82, 251, 193, 98, 82, 251, 193, 100, 82, 251, 193, 140, 82, 251, 193, 142, 82, 251, 193, 144, 82, 251, 193, 146, 82, 251, 193, 148, 82, 251, 193, 150, 82, 251, 193, 152, 82, 251, 193, 154, 82, 251, 193, 156, 82, 251, 206, 96, 10, 0, 206, 96, 10, 0, 206, 96, 10, 0, 206, 96, 10, 0, 206, 96, 10, 0, 206, 96, 10, 0, 206, 96, 10, 0, 193, 168, 82, 251, 193, 170, 82, 251, 193, 172, 82, 251, 193, 174, 82, 251, 193, 176, 82, 251, 193, 178, 82, 251, 193, 222, 82, 251, 193, 224, 82, 251, 193, 226, 82, 251, 193, 228, 82, 251, 193, 230, 82, 251, 193, 232, 82, 251, 193, 6, 83, 251, 193, 8, 83, 251, 193, 10, 83, 251, 193, 12, 83, 251, 193, 36, 83, 251, 193, 38, 83, 251, 193, 40, 83, 251, 193, 50, 83, 251, 193, 52, 83, 251, 193, 64, 83, 251, 193, 66, 83, 251, 193, 68, 83, 251, 193, 70, 83, 251, 193, 74, 83, 251, 193, 112, 83, 251, 193, 238, 83, 251, 193, 240, 83, 251, 193, 40, 84, 251, 193, 86, 84, 251, 193, 88, 84, 251, 206, 96, 10, 0, 206, 96, 10, 0, 206, 96, 10, 0, 206, 96, 10, 0, 206, 96, 10, 0, 206, 96, 10, 0, 193, 240, 82, 251, 193, 242, 82, 251, 193, 244, 82, 251, 193, 246, 82, 251, 193, 248, 82, 251, 193, 250, 82, 251, 193, 252, 82, 251, 193, 254, 82, 251, 193, 2, 83, 251, 193, 16, 83, 251, 193, 18, 83, 251, 193, 20, 83, 251, 193, 22, 83, 251, 193, 24, 83, 251, 193, 26, 83, 251, 193, 28, 83, 251, 193, 30, 83, 251, 193, 32, 83, 251, 193, 34, 83, 251, 193, 42, 83, 251, 193, 44, 83, 251, 193, 56, 83, 251, 193, 58, 83, 251, 193, 60, 83, 251, 193, 46, 83, 251, 193, 72, 83, 251, 193, 4, 83, 251, 193, 48, 83, 251, 193, 110, 83, 251, 193, 234, 83, 251, 193, 236, 83, 251, 193, 36, 84, 251, 193, 38, 84, 251, 193, 82, 84, 251, 193, 84, 84, 251, 193, 138, 84, 251, 193, 140, 84, 251, 193, 194, 84, 251, 193, 196, 84, 251, 193, 244, 84, 251, 193, 246, 84, 251, 193, 248, 84, 251, 193, 34, 85, 251, 193, 78, 85, 251, 193, 108, 85, 251, 193, 132, 85, 251, 193, 146, 85, 251, 193, 174, 85, 251, 193, 200, 85, 251, 193, 240, 85, 251, 193, 242, 85, 251, 193, 74, 86, 251, 193, 76, 86, 251, 206, 192, 10, 0, 206, 192, 10, 0, 206, 192, 10, 0, 206, 192, 10, 0, 206, 192, 10, 0, 193, 166, 40, 253, 193, 168, 40, 253, 193, 170, 40, 253, 193, 172, 40, 253, 193, 174, 40, 253, 193, 62, 83, 251, 193, 176, 40, 253, 193, 178, 40, 253, 193, 184, 40, 253, 193, 190, 40, 253, 193, 82, 41, 253, 193, 162, 43, 253, 193, 174, 43, 253, 193, 176, 43, 253, 193, 178, 43, 253, 193, 186, 43, 253, 193, 188, 43, 253, 193, 190, 43, 253, 193, 192, 43, 253, 193, 202, 43, 253, 193, 204, 43, 253, 193, 214, 43, 253, 193, 216, 43, 253, 193, 218, 43, 253, 193, 228, 43, 253, 193, 230, 43, 253, 193, 238, 43, 253, 193, 240, 43, 253, 193, 242, 43, 253, 193, 244, 43, 253, 193, 246, 43, 253, 193, 248, 43, 253, 193, 78, 83, 251, 193, 80, 83, 251, 193, 82, 83, 251, 193, 100, 83, 251, 193, 102, 83, 251, 193, 104, 83, 251, 193, 106, 83, 251, 193, 130, 83, 251, 193, 132, 83, 251, 193, 134, 83, 251, 193, 136, 83, 251, 193, 138, 83, 251, 193, 140, 83, 251, 193, 142, 83, 251, 193, 144, 83, 251, 193, 146, 83, 251, 193, 148, 83, 251, 193, 150, 83, 251, 193, 152, 83, 251, 193, 154, 83, 251, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 193, 120, 83, 251, 193, 122, 83, 251, 193, 124, 83, 251, 193, 126, 83, 251, 193, 156, 83, 251, 193, 158, 83, 251, 193, 160, 83, 251, 193, 162, 83, 251, 193, 164, 83, 251, 193, 166, 83, 251, 193, 168, 83, 251, 193, 170, 83, 251, 193, 172, 83, 251, 193, 174, 83, 251, 193, 176, 83, 251, 193, 178, 83, 251, 193, 180, 83, 251, 193, 182, 83, 251, 193, 184, 83, 251, 193, 186, 83, 251, 193, 188, 83, 251, 193, 190, 83, 251, 193, 246, 83, 251, 193, 248, 83, 251, 193, 250, 83, 251, 193, 252, 83, 251, 193, 254, 83, 251, 193, 2, 84, 251, 193, 4, 84, 251, 193, 6, 84, 251, 193, 8, 84, 251, 193, 46, 84, 251, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 206, 128, 10, 0, 193, 16, 84, 251, 193, 18, 84, 251, 193, 20, 84, 251, 193, 22, 84, 251, 193, 24, 84, 251, 193, 120, 94, 252, 193, 26, 84, 251, 193, 28, 84, 251, 193, 30, 84, 251, 193, 32, 84, 251, 193, 52, 84, 251, 193, 54, 84, 251, 193, 56, 84, 251, 193, 58, 84, 251, 193, 60, 84, 251, 193, 62, 84, 251, 193, 64, 84, 251, 193, 66, 84, 251, 193, 68, 84, 251, 193, 70, 84, 251, 193, 72, 84, 251, 193, 74, 84, 251, 193, 76, 84, 251, 193, 78, 84, 251, 193, 10, 84, 251, 193, 12, 84, 251, 193, 50, 84, 251, 193, 102, 84, 251, 193, 48, 85, 251, 193, 160, 85, 251, 193, 216, 85, 251, 193, 160, 86, 251, 193, 162, 86, 251, 193, 164, 86, 251, 193, 166, 86, 251, 193, 168, 86, 251, 193, 170, 86, 251, 193, 172, 86, 251, 193, 174, 86, 251, 193, 176, 86, 251, 193, 178, 86, 251, 193, 168, 87, 251, 193, 170, 87, 251, 193, 172, 87, 251, 193, 174, 87, 251, 193, 176, 87, 251, 193, 172, 88, 251, 193, 174, 88, 251, 193, 176, 88, 251, 193, 178, 88, 251, 193, 180, 88, 251, 193, 182, 88, 251, 193, 184, 88, 251, 193, 186, 88, 251, 193, 188, 88, 251, 193, 194, 89, 251, 193, 48, 84, 251, 193, 90, 84, 251, 193, 92, 84, 251, 193, 94, 84, 251, 193, 96, 84, 251, 193, 98, 84, 251, 193, 100, 84, 251, 193, 152, 84, 251, 193, 154, 84, 251, 193, 156, 84, 251, 193, 158, 84, 251, 193, 160, 84, 251, 193, 208, 84, 251, 193, 210, 84, 251, 193, 212, 84, 251, 193, 162, 84, 251, 193, 214, 84, 251, 193, 254, 84, 251, 193, 2, 85, 251, 193, 4, 85, 251, 193, 6, 85, 251, 193, 8, 85, 251, 193, 40, 85, 251, 193, 42, 85, 251, 193, 44, 85, 251, 193, 46, 85, 251, 193, 92, 85, 251, 193, 118, 85, 251, 193, 120, 85, 251, 193, 178, 85, 251, 193, 180, 85, 251, 193, 186, 85, 251, 193, 104, 84, 251, 193, 106, 84, 251, 193, 108, 84, 251, 193, 110, 84, 251, 193, 112, 84, 251, 193, 114, 84, 251, 193, 116, 84, 251, 193, 118, 84, 251, 193, 120, 84, 251, 193, 122, 84, 251, 193, 124, 84, 251, 193, 126, 84, 251, 193, 128, 84, 251, 193, 130, 84, 251, 193, 132, 84, 251, 193, 134, 84, 251, 193, 164, 84, 251, 193, 166, 84, 251, 193, 168, 84, 251, 193, 170, 84, 251, 193, 172, 84, 251, 193, 174, 84, 251, 193, 176, 84, 251, 193, 178, 84, 251, 193, 180, 84, 251, 193, 182, 84, 251, 193, 184, 84, 251, 193, 186, 84, 251, 193, 216, 84, 251, 193, 218, 84, 251, 193, 220, 84, 251, 193, 222, 84, 251, 193, 142, 84, 251, 193, 144, 84, 251, 193, 146, 84, 251, 193, 148, 84, 251, 193, 150, 84, 251, 193, 198, 84, 251, 193, 200, 84, 251, 193, 202, 84, 251, 193, 204, 84, 251, 193, 250, 84, 251, 193, 252, 84, 251, 193, 36, 85, 251, 193, 80, 85, 251, 193, 82, 85, 251, 193, 84, 85, 251, 193, 86, 85, 251, 193, 88, 85, 251, 193, 110, 85, 251, 193, 112, 85, 251, 193, 114, 85, 251, 193, 134, 85, 251, 193, 148, 85, 251, 193, 152, 85, 251, 193, 158, 85, 251, 193, 176, 85, 251, 193, 184, 85, 251, 193, 244, 85, 251, 193, 78, 86, 251, 193, 80, 86, 251, 193, 82, 86, 251, 193, 56, 87, 251, 193, 58, 87, 251, 193, 188, 84, 251, 193, 224, 84, 251, 193, 226, 84, 251, 193, 228, 84, 251, 193, 230, 84, 251, 193, 232, 84, 251, 193, 234, 84, 251, 193, 236, 84, 251, 193, 238, 84, 251, 193, 240, 84, 251, 193, 10, 85, 251, 193, 12, 85, 251, 193, 14, 85, 251, 193, 16, 85, 251, 193, 18, 85, 251, 193, 20, 85, 251, 193, 22, 85, 251, 193, 24, 85, 251, 193, 26, 85, 251, 193, 28, 85, 251, 193, 30, 85, 251, 193, 50, 85, 251, 193, 52, 85, 251, 193, 54, 85, 251, 193, 56, 85, 251, 193, 58, 85, 251, 193, 60, 85, 251, 193, 62, 85, 251, 193, 64, 85, 251, 193, 66, 85, 251, 193, 68, 85, 251, 193, 70, 85, 251, 193, 110, 64, 252, 193, 112, 64, 252, 193, 114, 64, 252, 193, 116, 64, 252, 193, 118, 64, 252, 193, 120, 64, 252, 193, 122, 64, 252, 193, 124, 64, 252, 193, 126, 64, 252, 193, 128, 64, 252, 193, 130, 64, 252, 193, 132, 64, 252, 193, 144, 64, 252, 193, 146, 64, 252, 193, 148, 64, 252, 193, 150, 64, 252, 193, 170, 64, 252, 193, 172, 64, 252, 193, 174, 64, 252, 193, 176, 64, 252, 193, 190, 64, 252, 193, 192, 64, 252, 193, 194, 64, 252, 193, 242, 84, 251, 193, 196, 64, 252, 193, 198, 64, 252, 193, 200, 64, 252, 193, 202, 64, 252, 193, 226, 64, 252, 193, 228, 64, 252, 193, 230, 64, 252, 193, 232, 64, 252, 193, 72, 85, 251, 193, 74, 85, 251, 193, 94, 85, 251, 193, 96, 85, 251, 193, 98, 85, 251, 193, 122, 85, 251, 193, 100, 85, 251, 193, 102, 85, 251, 193, 104, 85, 251, 193, 106, 85, 251, 193, 124, 85, 251, 193, 126, 85, 251, 193, 128, 85, 251, 193, 136, 85, 251, 193, 138, 85, 251, 193, 140, 85, 251, 193, 142, 85, 251, 193, 144, 85, 251, 193, 150, 85, 251, 193, 154, 85, 251, 193, 156, 85, 251, 193, 162, 85, 251, 193, 164, 85, 251, 193, 166, 85, 251, 193, 168, 85, 251, 193, 170, 85, 251, 193, 172, 85, 251, 193, 182, 85, 251, 193, 194, 85, 251, 193, 196, 85, 251, 193, 198, 85, 251, 193, 218, 85, 251, 193, 188, 85, 251, 193, 190, 85, 251, 193, 204, 85, 251, 193, 206, 85, 251, 193, 208, 85, 251, 193, 210, 85, 251, 193, 212, 85, 251, 193, 214, 85, 251, 193, 250, 85, 251, 193, 252, 85, 251, 193, 254, 85, 251, 193, 2, 86, 251, 193, 4, 86, 251, 193, 6, 86, 251, 193, 8, 86, 251, 193, 10, 86, 251, 193, 12, 86, 251, 193, 14, 86, 251, 193, 16, 86, 251, 193, 18, 86, 251, 193, 20, 86, 251, 193, 22, 86, 251, 193, 24, 86, 251, 193, 26, 86, 251, 193, 28, 86, 251, 193, 30, 86, 251, 193, 32, 86, 251, 193, 84, 86, 251, 193, 86, 86, 251, 193, 88, 86, 251, 193, 90, 86, 251, 193, 92, 86, 251, 193, 220, 85, 251, 193, 222, 85, 251, 193, 224, 85, 251, 193, 226, 85, 251, 193, 228, 85, 251, 193, 230, 85, 251, 193, 232, 85, 251, 193, 234, 85, 251, 193, 34, 86, 251, 193, 36, 86, 251, 193, 38, 86, 251, 193, 40, 86, 251, 193, 42, 86, 251, 193, 44, 86, 251, 193, 46, 86, 251, 193, 48, 86, 251, 193, 50, 86, 251, 193, 52, 86, 251, 193, 54, 86, 251, 193, 56, 86, 251, 193, 58, 86, 251, 193, 60, 86, 251, 193, 62, 86, 251, 193, 154, 119, 251, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 193, 94, 86, 251, 193, 96, 86, 251, 193, 98, 86, 251, 193, 100, 86, 251, 193, 102, 86, 251, 193, 104, 86, 251, 193, 106, 86, 251, 193, 108, 86, 251, 193, 110, 86, 251, 193, 112, 86, 251, 193, 114, 86, 251, 193, 116, 86, 251, 193, 118, 86, 251, 193, 120, 86, 251, 193, 122, 86, 251, 193, 124, 86, 251, 193, 126, 86, 251, 193, 128, 86, 251, 193, 130, 86, 251, 193, 132, 86, 251, 193, 134, 86, 251, 193, 136, 86, 251, 193, 138, 86, 251, 193, 140, 86, 251, 193, 142, 86, 251, 193, 144, 86, 251, 193, 146, 86, 251, 193, 148, 86, 251, 193, 150, 86, 251, 193, 152, 86, 251, 193, 154, 86, 251, 193, 156, 86, 251, 193, 158, 86, 251, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 160, 10, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 193, 30, 87, 251, 193, 32, 87, 251, 193, 34, 87, 251, 193, 26, 88, 251, 193, 28, 88, 251, 193, 30, 88, 251, 193, 32, 88, 251, 193, 34, 88, 251, 193, 36, 88, 251, 193, 38, 88, 251, 193, 40, 88, 251, 193, 42, 89, 251, 193, 44, 89, 251, 193, 46, 89, 251, 193, 48, 89, 251, 193, 50, 89, 251, 193, 52, 89, 251, 193, 54, 89, 251, 193, 56, 89, 251, 193, 58, 89, 251, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 192, 10, 0, 206, 192, 10, 0, 206, 192, 10, 0, 206, 192, 10, 0, 206, 192, 10, 0, 193, 42, 88, 251, 193, 44, 88, 251, 193, 46, 88, 251, 193, 48, 88, 251, 193, 60, 89, 251, 193, 62, 89, 251, 193, 64, 89, 251, 193, 66, 89, 251, 193, 68, 89, 251, 193, 70, 89, 251, 193, 72, 89, 251, 193, 104, 90, 251, 193, 106, 90, 251, 193, 108, 90, 251, 193, 110, 90, 251, 193, 112, 90, 251, 193, 114, 90, 251, 193, 116, 90, 251, 193, 118, 90, 251, 193, 120, 90, 251, 193, 122, 90, 251, 193, 124, 90, 251, 193, 126, 90, 251, 206, 64, 12, 0, 206, 64, 12, 0, 206, 64, 12, 0, 206, 64, 12, 0, 193, 60, 87, 251, 193, 62, 87, 251, 193, 64, 87, 251, 193, 66, 87, 251, 193, 68, 87, 251, 193, 50, 88, 251, 193, 52, 88, 251, 193, 54, 88, 251, 193, 56, 88, 251, 193, 58, 88, 251, 193, 60, 88, 251, 193, 62, 88, 251, 193, 64, 88, 251, 193, 66, 88, 251, 193, 68, 88, 251, 193, 74, 89, 251, 193, 76, 89, 251, 193, 78, 89, 251, 193, 80, 89, 251, 193, 82, 89, 251, 193, 84, 89, 251, 193, 86, 89, 251, 193, 88, 89, 251, 193, 90, 89, 251, 193, 92, 89, 251, 193, 94, 89, 251, 193, 96, 89, 251, 193, 98, 89, 251, 206, 192, 11, 0, 206, 192, 11, 0, 206, 192, 11, 0, 206, 192, 11, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 224, 10, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 0, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 193, 72, 88, 251, 193, 74, 88, 251, 193, 76, 88, 251, 193, 78, 88, 251, 193, 80, 88, 251, 193, 100, 89, 251, 193, 102, 89, 251, 193, 104, 89, 251, 193, 106, 89, 251, 193, 108, 89, 251, 193, 110, 89, 251, 193, 112, 89, 251, 193, 114, 89, 251, 193, 116, 89, 251, 193, 154, 90, 251, 193, 156, 90, 251, 193, 158, 90, 251, 193, 160, 90, 251, 193, 162, 90, 251, 193, 164, 90, 251, 193, 166, 90, 251, 193, 168, 90, 251, 193, 170, 90, 251, 193, 244, 91, 251, 193, 246, 91, 251, 193, 248, 91, 251, 193, 250, 91, 251, 193, 252, 91, 251, 193, 254, 91, 251, 193, 2, 92, 251, 193, 4, 92, 251, 193, 6, 92, 251, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 32, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 206, 96, 11, 0, 193, 168, 88, 251, 193, 166, 89, 251, 193, 168, 89, 251, 193, 170, 89, 251, 193, 172, 89, 251, 193, 174, 89, 251, 193, 176, 89, 251, 193, 178, 89, 251, 193, 180, 89, 251, 193, 182, 89, 251, 193, 184, 89, 251, 193, 186, 89, 251, 193, 188, 89, 251, 193, 190, 89, 251, 206, 224, 11, 0, 206, 224, 11, 0, 206, 224, 11, 0, 206, 224, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 206, 64, 11, 0, 193, 168, 205, 252, 193, 10, 89, 251, 193, 12, 89, 251, 193, 14, 89, 251, 193, 16, 89, 251, 193, 18, 89, 251, 193, 20, 89, 251, 193, 22, 89, 251, 193, 24, 89, 251, 193, 26, 89, 251, 193, 28, 89, 251, 193, 30, 89, 251, 193, 32, 89, 251, 193, 34, 89, 251, 193, 36, 89, 251, 193, 38, 89, 251, 193, 40, 89, 251, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 193, 196, 89, 251, 193, 198, 89, 251, 193, 200, 89, 251, 193, 202, 89, 251, 193, 204, 89, 251, 193, 206, 89, 251, 193, 208, 89, 251, 193, 32, 91, 251, 193, 34, 91, 251, 193, 36, 91, 251, 193, 38, 91, 251, 193, 40, 91, 251, 193, 42, 91, 251, 193, 44, 91, 251, 193, 110, 92, 251, 193, 112, 92, 251, 193, 114, 92, 251, 193, 116, 92, 251, 193, 118, 92, 251, 193, 120, 92, 251, 193, 122, 92, 251, 193, 124, 92, 251, 193, 126, 92, 251, 193, 128, 92, 251, 193, 130, 92, 251, 193, 230, 93, 251, 193, 232, 93, 251, 193, 234, 93, 251, 193, 236, 93, 251, 193, 238, 93, 251, 193, 240, 93, 251, 193, 242, 93, 251, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 128, 11, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 193, 70, 90, 251, 193, 70, 240, 251, 193, 72, 240, 251, 193, 74, 240, 251, 193, 76, 240, 251, 193, 78, 240, 251, 193, 80, 240, 251, 193, 82, 240, 251, 193, 84, 240, 251, 193, 86, 240, 251, 193, 88, 240, 251, 193, 90, 240, 251, 193, 92, 240, 251, 193, 94, 240, 251, 193, 96, 240, 251, 193, 98, 240, 251, 193, 100, 240, 251, 193, 102, 240, 251, 193, 104, 240, 251, 193, 106, 240, 251, 193, 108, 240, 251, 193, 110, 240, 251, 193, 112, 240, 251, 193, 114, 240, 251, 206, 160, 36, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 206, 160, 11, 0, 193, 170, 91, 251, 193, 172, 91, 251, 193, 174, 91, 251, 193, 176, 91, 251, 193, 178, 91, 251, 193, 180, 91, 251, 193, 182, 91, 251, 193, 184, 91, 251, 193, 186, 91, 251, 193, 188, 91, 251, 193, 190, 91, 251, 193, 192, 91, 251, 193, 194, 91, 251, 193, 196, 91, 251, 193, 26, 93, 251, 193, 28, 93, 251, 193, 30, 93, 251, 193, 32, 93, 251, 193, 34, 93, 251, 193, 36, 93, 251, 193, 38, 93, 251, 193, 40, 93, 251, 193, 42, 93, 251, 193, 44, 93, 251, 193, 46, 93, 251, 193, 48, 93, 251, 193, 118, 94, 251, 193, 120, 94, 251, 206, 192, 11, 0, 206, 192, 11, 0, 206, 192, 11, 0, 206, 192, 11, 0, 206, 192, 11, 0, 206, 192, 11, 0, 206, 192, 11, 0, 206, 192, 11, 0, 206, 192, 11, 0, 193, 218, 91, 251, 193, 220, 91, 251, 193, 222, 91, 251, 193, 224, 91, 251, 193, 226, 91, 251, 193, 228, 91, 251, 193, 230, 91, 251, 193, 232, 91, 251, 193, 234, 91, 251, 193, 236, 91, 251, 193, 238, 91, 251, 193, 240, 91, 251, 193, 242, 91, 251, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 224, 11, 0, 206, 224, 11, 0, 206, 224, 11, 0, 206, 224, 11, 0, 206, 224, 11, 0, 206, 224, 11, 0, 206, 224, 11, 0, 206, 224, 11, 0, 206, 224, 11, 0, 206, 224, 11, 0, 193, 10, 92, 251, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 0, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 32, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 64, 12, 0, 206, 64, 12, 0, 206, 64, 12, 0, 206, 64, 12, 0, 206, 64, 12, 0, 206, 64, 12, 0, 193, 52, 93, 251, 193, 54, 93, 251, 193, 56, 93, 251, 193, 58, 93, 251, 193, 60, 93, 251, 193, 62, 93, 251, 193, 64, 93, 251, 193, 66, 93, 251, 193, 68, 93, 251, 193, 70, 93, 251, 193, 72, 93, 251, 193, 74, 93, 251, 193, 150, 94, 251, 193, 152, 94, 251, 193, 154, 94, 251, 193, 156, 94, 251, 193, 158, 94, 251, 193, 160, 94, 251, 193, 162, 94, 251, 193, 164, 94, 251, 193, 166, 94, 251, 193, 194, 95, 251, 193, 196, 95, 251, 193, 198, 95, 251, 193, 200, 95, 251, 193, 202, 95, 251, 193, 8, 92, 251, 193, 120, 93, 251, 193, 122, 93, 251, 193, 124, 93, 251, 193, 126, 93, 251, 193, 128, 93, 251, 193, 130, 93, 251, 193, 132, 93, 251, 193, 134, 93, 251, 193, 136, 93, 251, 193, 138, 93, 251, 193, 140, 93, 251, 193, 142, 93, 251, 193, 202, 94, 251, 193, 204, 94, 251, 193, 206, 94, 251, 193, 208, 94, 251, 193, 210, 94, 251, 193, 212, 94, 251, 193, 214, 94, 251, 193, 216, 94, 251, 193, 218, 94, 251, 193, 220, 94, 251, 193, 222, 94, 251, 193, 224, 94, 251, 206, 64, 13, 0, 206, 64, 13, 0, 206, 64, 13, 0, 206, 64, 13, 0, 206, 64, 13, 0, 206, 64, 13, 0, 206, 64, 13, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 206, 96, 12, 0, 193, 144, 93, 251, 193, 90, 92, 251, 193, 92, 92, 251, 193, 94, 92, 251, 193, 96, 92, 251, 193, 98, 92, 251, 193, 146, 93, 251, 193, 100, 92, 251, 193, 102, 92, 251, 193, 104, 92, 251, 193, 106, 92, 251, 193, 108, 92, 251, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 128, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 206, 160, 12, 0, 193, 168, 94, 251, 193, 170, 94, 251, 193, 172, 94, 251, 193, 174, 94, 251, 193, 176, 94, 251, 193, 178, 94, 251, 193, 180, 94, 251, 193, 182, 94, 251, 193, 184, 94, 251, 193, 186, 94, 251, 193, 188, 94, 251, 193, 190, 94, 251, 193, 192, 94, 251, 193, 194, 94, 251, 193, 196, 94, 251, 193, 198, 94, 251, 193, 200, 94, 251, 193, 210, 95, 251, 193, 212, 95, 251, 193, 214, 95, 251, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 192, 12, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 224, 12, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 193, 122, 44, 253, 193, 124, 44, 253, 193, 126, 44, 253, 193, 128, 44, 253, 193, 130, 44, 253, 193, 144, 44, 253, 193, 146, 44, 253, 193, 116, 94, 251, 193, 148, 44, 253, 193, 162, 44, 253, 193, 164, 44, 253, 193, 166, 44, 253, 193, 168, 44, 253, 193, 184, 44, 253, 193, 186, 44, 253, 193, 188, 44, 253, 193, 190, 44, 253, 193, 192, 44, 253, 193, 194, 44, 253, 193, 196, 44, 253, 193, 198, 44, 253, 193, 200, 44, 253, 193, 202, 44, 253, 193, 218, 44, 253, 193, 220, 44, 253, 193, 222, 44, 253, 193, 224, 44, 253, 193, 226, 44, 253, 193, 238, 44, 253, 193, 240, 44, 253, 193, 242, 44, 253, 193, 244, 44, 253, 193, 122, 94, 251, 193, 124, 94, 251, 193, 126, 94, 251, 193, 128, 94, 251, 193, 130, 94, 251, 193, 132, 94, 251, 193, 134, 94, 251, 193, 136, 94, 251, 193, 138, 94, 251, 193, 140, 94, 251, 193, 142, 94, 251, 193, 144, 94, 251, 193, 146, 94, 251, 193, 148, 94, 251, 193, 170, 95, 251, 193, 172, 95, 251, 193, 174, 95, 251, 193, 176, 95, 251, 193, 178, 95, 251, 193, 180, 95, 251, 193, 182, 95, 251, 193, 184, 95, 251, 193, 186, 95, 251, 193, 188, 95, 251, 193, 190, 95, 251, 193, 214, 96, 251, 193, 216, 96, 251, 193, 218, 96, 251, 193, 220, 96, 251, 193, 222, 96, 251, 193, 224, 96, 251, 193, 178, 97, 251, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 0, 13, 0, 206, 96, 13, 0, 206, 96, 13, 0, 206, 96, 13, 0, 206, 96, 13, 0, 193, 56, 95, 251, 193, 58, 95, 251, 193, 60, 95, 251, 193, 62, 95, 251, 193, 64, 95, 251, 193, 66, 95, 251, 193, 68, 95, 251, 193, 66, 96, 251, 193, 68, 96, 251, 193, 70, 96, 251, 193, 72, 96, 251, 193, 74, 96, 251, 193, 76, 96, 251, 193, 78, 96, 251, 193, 80, 96, 251, 193, 82, 96, 251, 193, 80, 97, 251, 193, 82, 97, 251, 193, 84, 97, 251, 193, 86, 97, 251, 193, 88, 97, 251, 193, 90, 97, 251, 193, 92, 97, 251, 193, 254, 97, 251, 193, 2, 98, 251, 193, 4, 98, 251, 193, 6, 98, 251, 193, 122, 98, 251, 193, 124, 98, 251, 193, 126, 98, 251, 193, 238, 98, 251, 193, 240, 98, 251, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 206, 32, 13, 0, 193, 86, 96, 251, 193, 88, 96, 251, 193, 200, 216, 252, 193, 210, 216, 252, 193, 216, 216, 252, 193, 218, 216, 252, 193, 220, 216, 252, 193, 228, 216, 252, 193, 230, 216, 252, 193, 232, 216, 252, 193, 168, 95, 251, 193, 236, 216, 252, 193, 238, 216, 252, 193, 240, 216, 252, 193, 242, 216, 252, 193, 244, 216, 252, 193, 250, 216, 252, 193, 252, 216, 252, 193, 254, 216, 252, 193, 12, 217, 252, 193, 14, 217, 252, 193, 16, 217, 252, 193, 18, 217, 252, 193, 24, 217, 252, 193, 26, 217, 252, 193, 34, 217, 252, 193, 36, 217, 252, 193, 28, 217, 252, 193, 38, 217, 252, 193, 40, 217, 252, 193, 42, 217, 252, 193, 56, 217, 252, 193, 58, 217, 252, 193, 60, 217, 252, 193, 192, 95, 251, 193, 192, 97, 251, 193, 166, 101, 251, 193, 178, 101, 251, 193, 126, 104, 251, 193, 4, 106, 251, 193, 140, 107, 251, 193, 122, 109, 251, 193, 124, 109, 251, 193, 164, 110, 251, 193, 166, 110, 251, 193, 168, 110, 251, 193, 128, 113, 251, 193, 144, 115, 251, 193, 160, 116, 251, 193, 62, 121, 251, 193, 250, 125, 251, 193, 128, 126, 251, 193, 250, 131, 251, 193, 32, 136, 251, 193, 130, 138, 251, 193, 150, 139, 251, 193, 8, 157, 251, 193, 10, 157, 251, 193, 180, 165, 251, 193, 142, 171, 251, 193, 138, 194, 251, 193, 14, 196, 251, 193, 6, 202, 251, 193, 226, 203, 251, 193, 208, 205, 251, 193, 64, 210, 251, 193, 204, 95, 251, 193, 206, 95, 251, 193, 208, 95, 251, 193, 226, 96, 251, 193, 228, 96, 251, 193, 230, 96, 251, 193, 232, 96, 251, 193, 234, 96, 251, 193, 236, 96, 251, 193, 238, 96, 251, 193, 240, 96, 251, 193, 242, 96, 251, 193, 244, 96, 251, 193, 194, 97, 251, 193, 196, 97, 251, 193, 198, 97, 251, 193, 96, 98, 251, 193, 98, 98, 251, 193, 100, 98, 251, 193, 102, 98, 251, 193, 104, 98, 251, 193, 106, 98, 251, 193, 196, 98, 251, 193, 198, 98, 251, 193, 200, 98, 251, 193, 36, 99, 251, 193, 38, 99, 251, 193, 40, 99, 251, 193, 42, 99, 251, 193, 90, 99, 251, 193, 92, 99, 251, 193, 94, 99, 251, 193, 216, 95, 251, 193, 218, 95, 251, 193, 220, 95, 251, 193, 222, 95, 251, 193, 224, 95, 251, 193, 226, 95, 251, 193, 228, 95, 251, 193, 230, 95, 251, 193, 232, 95, 251, 193, 234, 95, 251, 193, 246, 96, 251, 193, 248, 96, 251, 193, 250, 96, 251, 193, 252, 96, 251, 193, 254, 96, 251, 193, 2, 97, 251, 193, 4, 97, 251, 193, 6, 97, 251, 193, 8, 97, 251, 193, 10, 97, 251, 193, 12, 97, 251, 193, 14, 97, 251, 193, 200, 97, 251, 193, 202, 97, 251, 193, 204, 97, 251, 193, 206, 97, 251, 193, 208, 97, 251, 193, 210, 97, 251, 193, 212, 97, 251, 193, 214, 97, 251, 193, 108, 98, 251, 193, 110, 98, 251, 206, 64, 13, 0, 206, 64, 13, 0, 206, 64, 13, 0, 206, 64, 13, 0, 206, 64, 13, 0, 206, 64, 13, 0, 193, 16, 97, 251, 193, 18, 97, 251, 193, 20, 97, 251, 193, 22, 97, 251, 193, 24, 97, 251, 193, 26, 97, 251, 193, 28, 97, 251, 193, 30, 97, 251, 193, 32, 97, 251, 193, 34, 97, 251, 193, 36, 97, 251, 193, 38, 97, 251, 193, 216, 97, 251, 193, 218, 97, 251, 193, 220, 97, 251, 193, 222, 97, 251, 193, 224, 97, 251, 193, 226, 97, 251, 193, 228, 97, 251, 193, 114, 98, 251, 193, 116, 98, 251, 193, 210, 98, 251, 193, 212, 98, 251, 193, 214, 98, 251, 193, 216, 98, 251, 193, 218, 98, 251, 206, 96, 13, 0, 206, 96, 13, 0, 206, 96, 13, 0, 206, 96, 13, 0, 206, 96, 13, 0, 206, 96, 13, 0, 206, 96, 13, 0, 206, 96, 13, 0, 193, 16, 215, 253, 193, 32, 96, 251, 193, 34, 96, 251, 193, 36, 96, 251, 193, 38, 96, 251, 193, 40, 96, 251, 193, 42, 96, 251, 193, 44, 96, 251, 193, 46, 96, 251, 193, 48, 96, 251, 193, 50, 96, 251, 193, 52, 96, 251, 193, 54, 96, 251, 193, 56, 96, 251, 193, 58, 96, 251, 193, 60, 96, 251, 193, 62, 96, 251, 193, 64, 96, 251, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 193, 90, 96, 251, 193, 92, 96, 251, 193, 94, 96, 251, 193, 96, 96, 251, 193, 98, 96, 251, 193, 100, 96, 251, 193, 102, 96, 251, 193, 104, 96, 251, 193, 106, 96, 251, 193, 108, 96, 251, 193, 110, 96, 251, 193, 112, 96, 251, 193, 114, 96, 251, 193, 116, 96, 251, 193, 118, 96, 251, 193, 120, 96, 251, 193, 122, 96, 251, 193, 124, 96, 251, 193, 126, 96, 251, 193, 128, 96, 251, 193, 130, 96, 251, 193, 132, 96, 251, 193, 134, 96, 251, 193, 136, 96, 251, 193, 138, 96, 251, 193, 140, 96, 251, 193, 142, 96, 251, 193, 144, 96, 251, 193, 146, 96, 251, 193, 148, 96, 251, 193, 150, 96, 251, 193, 152, 96, 251, 193, 154, 96, 251, 193, 156, 96, 251, 193, 158, 96, 251, 193, 160, 96, 251, 193, 162, 96, 251, 193, 164, 96, 251, 193, 166, 96, 251, 193, 168, 96, 251, 193, 170, 96, 251, 193, 172, 96, 251, 193, 174, 96, 251, 193, 176, 96, 251, 193, 178, 96, 251, 193, 180, 96, 251, 193, 182, 96, 251, 193, 184, 96, 251, 193, 186, 96, 251, 193, 188, 96, 251, 193, 190, 96, 251, 193, 192, 96, 251, 193, 194, 96, 251, 193, 196, 96, 251, 193, 198, 96, 251, 193, 200, 96, 251, 193, 202, 96, 251, 193, 204, 96, 251, 193, 206, 96, 251, 193, 208, 96, 251, 193, 210, 96, 251, 193, 212, 96, 251, 193, 94, 97, 251, 193, 96, 97, 251, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 206, 128, 13, 0, 193, 230, 97, 251, 193, 232, 97, 251, 193, 234, 97, 251, 193, 236, 97, 251, 193, 238, 97, 251, 193, 240, 97, 251, 193, 242, 97, 251, 193, 244, 97, 251, 193, 246, 97, 251, 193, 248, 97, 251, 193, 250, 97, 251, 193, 118, 98, 251, 193, 120, 98, 251, 193, 222, 98, 251, 193, 224, 98, 251, 193, 226, 98, 251, 193, 228, 98, 251, 193, 230, 98, 251, 193, 232, 98, 251, 193, 98, 97, 251, 193, 100, 97, 251, 193, 102, 97, 251, 193, 104, 97, 251, 193, 106, 97, 251, 193, 108, 97, 251, 193, 110, 97, 251, 193, 112, 97, 251, 193, 114, 97, 251, 193, 116, 97, 251, 193, 118, 97, 251, 193, 120, 97, 251, 193, 122, 97, 251, 193, 124, 97, 251, 193, 126, 97, 251, 193, 128, 97, 251, 193, 130, 97, 251, 193, 132, 97, 251, 193, 134, 97, 251, 193, 136, 97, 251, 193, 138, 97, 251, 193, 140, 97, 251, 193, 142, 97, 251, 193, 144, 97, 251, 193, 146, 97, 251, 193, 148, 97, 251, 193, 126, 119, 253, 206, 160, 13, 0, 206, 160, 13, 0, 206, 160, 13, 0, 206, 160, 13, 0, 206, 160, 13, 0, 206, 160, 13, 0, 206, 160, 13, 0, 206, 160, 13, 0, 206, 160, 13, 0, 206, 160, 13, 0, 193, 8, 98, 251, 193, 10, 98, 251, 193, 12, 98, 251, 193, 14, 98, 251, 193, 16, 98, 251, 193, 18, 98, 251, 193, 20, 98, 251, 193, 22, 98, 251, 193, 24, 98, 251, 193, 26, 98, 251, 193, 28, 98, 251, 193, 30, 98, 251, 193, 32, 98, 251, 193, 34, 98, 251, 193, 36, 98, 251, 193, 38, 98, 251, 193, 40, 98, 251, 193, 42, 98, 251, 193, 44, 98, 251, 193, 46, 98, 251, 193, 48, 98, 251, 193, 50, 98, 251, 193, 52, 98, 251, 193, 180, 97, 251, 193, 182, 97, 251, 193, 184, 97, 251, 193, 186, 97, 251, 193, 188, 97, 251, 193, 190, 97, 251, 193, 84, 98, 251, 193, 86, 98, 251, 193, 88, 98, 251, 193, 90, 98, 251, 193, 92, 98, 251, 193, 94, 98, 251, 193, 188, 98, 251, 193, 190, 98, 251, 193, 192, 98, 251, 193, 194, 98, 251, 193, 34, 99, 251, 193, 88, 99, 251, 193, 28, 100, 251, 193, 74, 100, 251, 193, 168, 100, 251, 193, 230, 100, 251, 193, 66, 101, 251, 193, 164, 101, 251, 193, 194, 101, 251, 193, 206, 101, 251, 193, 236, 101, 251, 193, 8, 102, 251, 193, 46, 102, 251, 193, 108, 102, 251, 193, 130, 102, 251, 193, 152, 102, 251, 193, 54, 98, 251, 193, 74, 218, 253, 193, 56, 98, 251, 193, 58, 98, 251, 193, 60, 98, 251, 193, 62, 98, 251, 193, 64, 98, 251, 193, 66, 98, 251, 193, 68, 98, 251, 193, 70, 98, 251, 193, 72, 98, 251, 193, 74, 98, 251, 193, 76, 98, 251, 193, 78, 98, 251, 193, 80, 98, 251, 193, 82, 98, 251, 193, 128, 98, 251, 193, 130, 98, 251, 193, 132, 98, 251, 193, 134, 98, 251, 193, 136, 98, 251, 193, 138, 98, 251, 193, 140, 98, 251, 193, 142, 98, 251, 193, 144, 98, 251, 193, 146, 98, 251, 193, 148, 98, 251, 193, 150, 98, 251, 193, 152, 98, 251, 193, 154, 98, 251, 193, 156, 98, 251, 193, 158, 98, 251, 193, 112, 98, 251, 193, 202, 98, 251, 193, 204, 98, 251, 193, 206, 98, 251, 193, 208, 98, 251, 193, 44, 99, 251, 193, 46, 99, 251, 193, 48, 99, 251, 193, 50, 99, 251, 193, 52, 99, 251, 193, 96, 99, 251, 193, 98, 99, 251, 193, 138, 99, 251, 193, 140, 99, 251, 193, 142, 99, 251, 193, 162, 99, 251, 193, 200, 99, 251, 193, 208, 99, 251, 193, 250, 99, 251, 193, 56, 100, 251, 193, 58, 100, 251, 193, 60, 100, 251, 193, 76, 100, 251, 193, 98, 100, 251, 193, 116, 100, 251, 193, 134, 100, 251, 193, 150, 100, 251, 193, 172, 100, 251, 193, 176, 100, 251, 193, 190, 100, 251, 193, 214, 100, 251, 193, 234, 100, 251, 193, 242, 98, 251, 193, 160, 98, 251, 193, 162, 98, 251, 193, 164, 98, 251, 193, 166, 98, 251, 193, 168, 98, 251, 193, 170, 98, 251, 193, 172, 98, 251, 193, 174, 98, 251, 193, 176, 98, 251, 193, 178, 98, 251, 193, 180, 98, 251, 193, 244, 98, 251, 193, 182, 98, 251, 193, 184, 98, 251, 193, 186, 98, 251, 193, 246, 98, 251, 193, 248, 98, 251, 193, 250, 98, 251, 193, 252, 98, 251, 193, 254, 98, 251, 193, 2, 99, 251, 193, 4, 99, 251, 193, 6, 99, 251, 193, 8, 99, 251, 193, 10, 99, 251, 193, 12, 99, 251, 193, 50, 44, 253, 206, 192, 13, 0, 206, 192, 13, 0, 206, 192, 13, 0, 206, 192, 13, 0, 193, 220, 98, 251, 193, 54, 99, 251, 193, 56, 99, 251, 193, 100, 99, 251, 193, 102, 99, 251, 193, 164, 99, 251, 193, 180, 99, 251, 193, 210, 99, 251, 193, 246, 99, 251, 193, 30, 100, 251, 193, 78, 100, 251, 193, 100, 100, 251, 193, 102, 100, 251, 193, 158, 100, 251, 193, 216, 100, 251, 193, 54, 101, 251, 193, 200, 101, 251, 193, 58, 102, 251, 193, 60, 102, 251, 193, 74, 102, 251, 193, 118, 102, 251, 193, 120, 102, 251, 193, 138, 102, 251, 193, 158, 102, 251, 193, 172, 102, 251, 193, 8, 103, 251, 193, 56, 103, 251, 193, 126, 103, 251, 193, 128, 103, 251, 193, 236, 103, 251, 193, 238, 103, 251, 193, 240, 103, 251, 206, 192, 13, 0, 206, 192, 13, 0, 206, 192, 13, 0, 206, 192, 13, 0, 206, 192, 13, 0, 206, 192, 13, 0, 193, 66, 99, 251, 193, 68, 99, 251, 193, 70, 99, 251, 193, 72, 99, 251, 193, 74, 99, 251, 193, 76, 99, 251, 193, 78, 99, 251, 193, 80, 99, 251, 193, 82, 99, 251, 193, 84, 99, 251, 193, 86, 99, 251, 193, 106, 99, 251, 193, 108, 99, 251, 193, 110, 99, 251, 193, 112, 99, 251, 193, 114, 99, 251, 193, 116, 99, 251, 193, 232, 161, 253, 193, 118, 99, 251, 193, 120, 99, 251, 193, 144, 99, 251, 193, 122, 99, 251, 193, 124, 99, 251, 193, 126, 99, 251, 193, 128, 99, 251, 193, 130, 99, 251, 193, 64, 99, 251, 193, 184, 99, 251, 193, 204, 99, 251, 193, 50, 100, 251, 193, 108, 100, 251, 193, 210, 100, 251, 193, 10, 101, 251, 193, 12, 101, 251, 193, 144, 101, 251, 193, 186, 101, 251, 193, 20, 102, 251, 193, 84, 102, 251, 193, 86, 102, 251, 193, 174, 102, 251, 193, 242, 102, 251, 193, 26, 103, 251, 193, 28, 103, 251, 193, 30, 103, 251, 193, 70, 103, 251, 193, 72, 103, 251, 193, 150, 103, 251, 193, 152, 103, 251, 193, 154, 103, 251, 193, 156, 103, 251, 193, 158, 103, 251, 193, 160, 103, 251, 193, 6, 104, 251, 193, 8, 104, 251, 193, 152, 104, 251, 193, 154, 104, 251, 193, 156, 104, 251, 193, 234, 104, 251, 193, 132, 99, 251, 193, 134, 99, 251, 193, 146, 99, 251, 193, 148, 99, 251, 193, 150, 99, 251, 193, 152, 99, 251, 193, 154, 99, 251, 193, 156, 99, 251, 193, 158, 99, 251, 193, 160, 99, 251, 193, 170, 99, 251, 193, 172, 99, 251, 193, 174, 99, 251, 193, 176, 99, 251, 193, 178, 99, 251, 193, 186, 99, 251, 193, 188, 99, 251, 193, 190, 99, 251, 193, 192, 99, 251, 193, 194, 99, 251, 193, 196, 99, 251, 193, 198, 99, 251, 193, 206, 99, 251, 193, 212, 99, 251, 193, 214, 99, 251, 193, 216, 99, 251, 193, 218, 99, 251, 193, 226, 99, 251, 193, 230, 99, 251, 193, 242, 99, 251, 193, 244, 99, 251, 193, 248, 99, 251, 193, 136, 99, 251, 193, 8, 100, 251, 193, 10, 100, 251, 193, 94, 100, 251, 193, 96, 100, 251, 193, 114, 100, 251, 193, 154, 100, 251, 193, 156, 100, 251, 193, 170, 100, 251, 193, 174, 100, 251, 193, 228, 100, 251, 193, 28, 101, 251, 193, 30, 101, 251, 193, 196, 101, 251, 193, 238, 101, 251, 193, 10, 102, 251, 193, 12, 102, 251, 193, 48, 102, 251, 193, 50, 102, 251, 193, 110, 102, 251, 193, 112, 102, 251, 193, 132, 102, 251, 193, 154, 102, 251, 193, 156, 102, 251, 193, 168, 102, 251, 193, 194, 102, 251, 193, 200, 102, 251, 193, 226, 102, 251, 193, 52, 103, 251, 193, 116, 103, 251, 193, 118, 103, 251, 193, 120, 103, 251, 193, 2, 100, 251, 193, 4, 100, 251, 193, 6, 100, 251, 193, 14, 100, 251, 193, 16, 100, 251, 193, 18, 100, 251, 193, 20, 100, 251, 193, 22, 100, 251, 193, 24, 100, 251, 193, 26, 100, 251, 193, 36, 100, 251, 193, 38, 100, 251, 193, 40, 100, 251, 193, 42, 100, 251, 193, 52, 100, 251, 193, 54, 100, 251, 193, 64, 100, 251, 193, 236, 104, 251, 193, 66, 100, 251, 193, 68, 100, 251, 193, 70, 100, 251, 193, 72, 100, 251, 193, 84, 100, 251, 193, 66, 218, 253, 193, 86, 100, 251, 193, 88, 100, 251, 193, 90, 100, 251, 193, 92, 100, 251, 193, 110, 100, 251, 193, 112, 100, 251, 193, 120, 100, 251, 193, 122, 100, 251, 193, 106, 100, 251, 193, 118, 100, 251, 193, 178, 100, 251, 193, 182, 100, 251, 193, 184, 100, 251, 193, 192, 100, 251, 193, 198, 100, 251, 193, 206, 100, 251, 193, 208, 100, 251, 193, 6, 101, 251, 193, 232, 100, 251, 193, 240, 100, 251, 193, 246, 100, 251, 193, 8, 101, 251, 193, 20, 101, 251, 193, 22, 101, 251, 193, 72, 101, 251, 193, 106, 101, 251, 193, 120, 101, 251, 193, 108, 101, 251, 193, 128, 101, 251, 193, 136, 101, 251, 193, 138, 101, 251, 193, 140, 101, 251, 193, 142, 101, 251, 193, 152, 101, 251, 193, 154, 101, 251, 193, 156, 101, 251, 193, 182, 101, 251, 193, 184, 101, 251, 193, 208, 101, 251, 193, 240, 101, 251, 193, 124, 100, 251, 193, 126, 100, 251, 193, 128, 100, 251, 193, 130, 100, 251, 193, 132, 100, 251, 193, 136, 100, 251, 193, 4, 45, 253, 193, 138, 100, 251, 193, 140, 100, 251, 193, 142, 100, 251, 193, 144, 100, 251, 193, 146, 100, 251, 193, 148, 100, 251, 193, 152, 100, 251, 193, 240, 133, 252, 193, 162, 100, 251, 193, 164, 100, 251, 193, 166, 100, 251, 193, 180, 100, 251, 193, 186, 100, 251, 193, 188, 100, 251, 193, 194, 100, 251, 193, 196, 100, 251, 193, 200, 100, 251, 193, 202, 100, 251, 193, 204, 100, 251, 193, 212, 100, 251, 193, 220, 100, 251, 193, 222, 100, 251, 193, 226, 100, 251, 193, 236, 100, 251, 193, 238, 100, 251, 193, 66, 211, 253, 193, 76, 211, 253, 193, 78, 211, 253, 193, 80, 211, 253, 193, 82, 211, 253, 193, 96, 211, 253, 193, 98, 211, 253, 193, 100, 211, 253, 193, 102, 211, 253, 193, 104, 211, 253, 193, 106, 211, 253, 193, 108, 211, 253, 193, 110, 211, 253, 193, 112, 211, 253, 193, 114, 211, 253, 193, 116, 211, 253, 193, 118, 211, 253, 193, 120, 211, 253, 193, 122, 211, 253, 193, 124, 211, 253, 193, 140, 211, 253, 193, 142, 211, 253, 193, 144, 211, 253, 193, 146, 211, 253, 193, 148, 211, 253, 193, 150, 211, 253, 193, 152, 211, 253, 193, 224, 100, 251, 193, 154, 211, 253, 193, 156, 211, 253, 193, 158, 211, 253, 193, 160, 211, 253, 193, 242, 100, 251, 193, 244, 100, 251, 193, 248, 100, 251, 193, 250, 100, 251, 193, 252, 100, 251, 193, 254, 100, 251, 193, 26, 101, 251, 193, 36, 101, 251, 193, 38, 101, 251, 193, 40, 101, 251, 193, 42, 101, 251, 193, 44, 101, 251, 193, 46, 101, 251, 193, 52, 101, 251, 193, 58, 101, 251, 193, 60, 101, 251, 193, 62, 101, 251, 193, 64, 101, 251, 193, 74, 101, 251, 193, 76, 101, 251, 193, 78, 101, 251, 193, 80, 101, 251, 193, 86, 101, 251, 193, 88, 101, 251, 193, 92, 101, 251, 193, 94, 101, 251, 193, 96, 101, 251, 193, 102, 101, 251, 193, 110, 101, 251, 193, 112, 101, 251, 193, 114, 101, 251, 193, 116, 101, 251, 193, 2, 101, 251, 193, 4, 101, 251, 193, 14, 101, 251, 193, 16, 101, 251, 193, 18, 101, 251, 193, 32, 101, 251, 193, 34, 101, 251, 193, 48, 101, 251, 193, 68, 101, 251, 193, 70, 101, 251, 193, 82, 101, 251, 193, 84, 101, 251, 193, 90, 101, 251, 193, 98, 101, 251, 193, 100, 101, 251, 193, 104, 101, 251, 193, 118, 101, 251, 193, 124, 101, 251, 193, 130, 101, 251, 193, 168, 101, 251, 193, 170, 101, 251, 193, 180, 101, 251, 193, 198, 101, 251, 193, 14, 102, 251, 193, 16, 102, 251, 193, 52, 102, 251, 193, 54, 102, 251, 193, 56, 102, 251, 193, 114, 102, 251, 193, 116, 102, 251, 193, 134, 102, 251, 193, 136, 102, 251, 193, 122, 101, 251, 193, 126, 101, 251, 193, 132, 101, 251, 193, 134, 101, 251, 193, 146, 101, 251, 193, 148, 101, 251, 193, 150, 101, 251, 193, 158, 101, 251, 193, 160, 101, 251, 193, 162, 101, 251, 193, 172, 101, 251, 193, 174, 101, 251, 193, 176, 101, 251, 193, 188, 101, 251, 193, 190, 101, 251, 193, 192, 101, 251, 193, 202, 101, 251, 193, 204, 101, 251, 193, 212, 101, 251, 193, 214, 101, 251, 193, 216, 101, 251, 193, 218, 101, 251, 193, 220, 101, 251, 193, 222, 101, 251, 193, 224, 101, 251, 193, 226, 101, 251, 193, 228, 101, 251, 193, 230, 101, 251, 193, 232, 101, 251, 193, 244, 101, 251, 193, 246, 101, 251, 193, 248, 101, 251, 193, 100, 212, 251, 193, 110, 212, 251, 193, 112, 212, 251, 193, 114, 212, 251, 193, 116, 212, 251, 193, 118, 212, 251, 193, 120, 212, 251, 193, 122, 212, 251, 193, 234, 101, 251, 193, 124, 212, 251, 193, 130, 212, 251, 193, 132, 212, 251, 193, 134, 212, 251, 193, 136, 212, 251, 193, 144, 212, 251, 193, 146, 212, 251, 193, 148, 212, 251, 193, 150, 212, 251, 193, 152, 212, 251, 193, 154, 212, 251, 193, 156, 212, 251, 193, 158, 212, 251, 193, 160, 212, 251, 193, 162, 212, 251, 193, 164, 212, 251, 193, 166, 212, 251, 193, 174, 212, 251, 193, 176, 212, 251, 193, 178, 212, 251, 193, 182, 212, 251, 193, 184, 212, 251, 193, 186, 212, 251, 193, 242, 101, 251, 193, 18, 102, 251, 193, 76, 102, 251, 193, 78, 102, 251, 193, 80, 102, 251, 193, 82, 102, 251, 193, 122, 102, 251, 193, 204, 102, 251, 193, 206, 102, 251, 193, 208, 102, 251, 193, 210, 102, 251, 193, 212, 102, 251, 193, 214, 102, 251, 193, 216, 102, 251, 193, 228, 102, 251, 193, 230, 102, 251, 193, 232, 102, 251, 193, 234, 102, 251, 193, 236, 102, 251, 193, 238, 102, 251, 193, 240, 102, 251, 193, 10, 103, 251, 193, 12, 103, 251, 193, 14, 103, 251, 193, 16, 103, 251, 193, 18, 103, 251, 193, 20, 103, 251, 193, 22, 103, 251, 193, 24, 103, 251, 193, 58, 103, 251, 193, 60, 103, 251, 193, 62, 103, 251, 193, 250, 101, 251, 193, 252, 101, 251, 193, 254, 101, 251, 193, 2, 102, 251, 193, 4, 102, 251, 193, 6, 102, 251, 193, 26, 102, 251, 193, 28, 102, 251, 193, 30, 102, 251, 193, 32, 102, 251, 193, 34, 102, 251, 193, 36, 102, 251, 193, 38, 102, 251, 193, 40, 102, 251, 193, 42, 102, 251, 193, 62, 102, 251, 193, 44, 102, 251, 193, 64, 102, 251, 193, 66, 102, 251, 193, 68, 102, 251, 193, 70, 102, 251, 193, 72, 102, 251, 193, 88, 102, 251, 193, 90, 102, 251, 193, 92, 102, 251, 193, 94, 102, 251, 193, 96, 102, 251, 193, 98, 102, 251, 193, 100, 102, 251, 193, 102, 102, 251, 193, 104, 102, 251, 193, 124, 102, 251, 193, 126, 102, 251, 193, 128, 102, 251, 193, 144, 102, 251, 193, 146, 102, 251, 193, 148, 102, 251, 193, 150, 102, 251, 193, 162, 102, 251, 193, 164, 102, 251, 193, 166, 102, 251, 193, 176, 102, 251, 193, 178, 102, 251, 193, 182, 102, 251, 193, 184, 102, 251, 193, 188, 102, 251, 193, 190, 102, 251, 193, 196, 102, 251, 193, 186, 102, 251, 193, 202, 102, 251, 193, 218, 102, 251, 193, 220, 102, 251, 193, 222, 102, 251, 193, 244, 102, 251, 193, 246, 102, 251, 193, 248, 102, 251, 193, 250, 102, 251, 193, 252, 102, 251, 193, 254, 102, 251, 193, 2, 103, 251, 193, 52, 72, 252, 193, 34, 103, 251, 193, 36, 103, 251, 193, 38, 103, 251, 193, 170, 102, 251, 193, 180, 102, 251, 193, 6, 103, 251, 193, 54, 103, 251, 193, 122, 103, 251, 193, 124, 103, 251, 193, 224, 103, 251, 193, 226, 103, 251, 193, 228, 103, 251, 193, 230, 103, 251, 193, 232, 103, 251, 193, 234, 103, 251, 193, 54, 104, 251, 193, 56, 104, 251, 193, 58, 104, 251, 193, 132, 104, 251, 193, 134, 104, 251, 193, 136, 104, 251, 193, 138, 104, 251, 193, 204, 104, 251, 193, 206, 104, 251, 193, 208, 104, 251, 193, 210, 104, 251, 193, 44, 105, 251, 193, 46, 105, 251, 193, 48, 105, 251, 193, 50, 105, 251, 193, 52, 105, 251, 193, 122, 105, 251, 193, 124, 105, 251, 193, 126, 105, 251, 193, 128, 105, 251, 193, 192, 102, 251, 193, 224, 102, 251, 193, 4, 103, 251, 193, 50, 103, 251, 193, 110, 103, 251, 193, 112, 103, 251, 193, 114, 103, 251, 193, 206, 103, 251, 193, 208, 103, 251, 193, 210, 103, 251, 193, 212, 103, 251, 193, 214, 103, 251, 193, 216, 103, 251, 193, 42, 104, 251, 193, 44, 104, 251, 193, 46, 104, 251, 193, 120, 104, 251, 193, 122, 104, 251, 193, 124, 104, 251, 193, 196, 104, 251, 193, 198, 104, 251, 193, 200, 104, 251, 193, 24, 105, 251, 193, 26, 105, 251, 193, 28, 105, 251, 193, 106, 105, 251, 193, 108, 105, 251, 193, 110, 105, 251, 193, 164, 105, 251, 193, 166, 105, 251, 193, 168, 105, 251, 193, 214, 105, 251, 193, 94, 6, 252, 193, 96, 6, 252, 193, 98, 6, 252, 193, 100, 6, 252, 193, 102, 6, 252, 193, 104, 6, 252, 193, 106, 6, 252, 193, 108, 6, 252, 193, 110, 6, 252, 193, 120, 6, 252, 193, 122, 6, 252, 193, 124, 6, 252, 193, 126, 6, 252, 193, 128, 6, 252, 193, 112, 6, 252, 193, 130, 6, 252, 193, 132, 6, 252, 193, 134, 6, 252, 193, 146, 6, 252, 193, 198, 102, 251, 193, 148, 6, 252, 193, 150, 6, 252, 193, 160, 6, 252, 193, 162, 6, 252, 193, 164, 6, 252, 193, 168, 6, 252, 193, 172, 6, 252, 193, 174, 6, 252, 193, 176, 6, 252, 193, 182, 6, 252, 193, 184, 6, 252, 193, 186, 6, 252, 193, 40, 103, 251, 193, 42, 103, 251, 193, 44, 103, 251, 193, 46, 103, 251, 193, 48, 103, 251, 193, 74, 103, 251, 193, 76, 103, 251, 193, 78, 103, 251, 193, 80, 103, 251, 193, 82, 103, 251, 193, 84, 103, 251, 193, 86, 103, 251, 193, 88, 103, 251, 193, 90, 103, 251, 193, 92, 103, 251, 193, 94, 103, 251, 193, 96, 103, 251, 193, 98, 103, 251, 193, 100, 103, 251, 193, 102, 103, 251, 193, 104, 103, 251, 193, 106, 103, 251, 193, 108, 103, 251, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 193, 64, 103, 251, 193, 66, 103, 251, 193, 68, 103, 251, 193, 130, 103, 251, 193, 132, 103, 251, 193, 134, 103, 251, 193, 136, 103, 251, 193, 138, 103, 251, 193, 140, 103, 251, 193, 142, 103, 251, 193, 64, 104, 251, 193, 144, 103, 251, 193, 146, 103, 251, 193, 148, 103, 251, 193, 242, 103, 251, 193, 244, 103, 251, 193, 246, 103, 251, 193, 248, 103, 251, 193, 250, 103, 251, 193, 252, 103, 251, 193, 254, 103, 251, 193, 2, 104, 251, 193, 4, 104, 251, 193, 66, 104, 251, 193, 68, 104, 251, 193, 70, 104, 251, 193, 72, 104, 251, 193, 144, 104, 251, 193, 146, 104, 251, 193, 148, 104, 251, 193, 150, 104, 251, 193, 220, 104, 251, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 206, 224, 13, 0, 193, 10, 104, 251, 193, 12, 104, 251, 193, 14, 104, 251, 193, 16, 104, 251, 193, 18, 104, 251, 193, 20, 104, 251, 193, 22, 104, 251, 193, 24, 104, 251, 193, 26, 104, 251, 193, 28, 104, 251, 193, 30, 104, 251, 193, 32, 104, 251, 193, 34, 104, 251, 193, 36, 104, 251, 193, 38, 104, 251, 193, 40, 104, 251, 193, 74, 104, 251, 193, 76, 104, 251, 193, 78, 104, 251, 193, 218, 103, 251, 193, 220, 103, 251, 193, 222, 103, 251, 193, 48, 104, 251, 193, 50, 104, 251, 193, 52, 104, 251, 193, 128, 104, 251, 193, 130, 104, 251, 193, 202, 104, 251, 193, 30, 105, 251, 193, 32, 105, 251, 193, 34, 105, 251, 193, 36, 105, 251, 193, 38, 105, 251, 193, 40, 105, 251, 193, 42, 105, 251, 193, 112, 105, 251, 193, 114, 105, 251, 193, 116, 105, 251, 193, 118, 105, 251, 193, 120, 105, 251, 193, 170, 105, 251, 193, 172, 105, 251, 193, 174, 105, 251, 193, 220, 105, 251, 193, 222, 105, 251, 193, 224, 105, 251, 193, 226, 105, 251, 193, 228, 105, 251, 193, 6, 106, 251, 193, 8, 106, 251, 193, 10, 106, 251, 193, 60, 104, 251, 193, 62, 104, 251, 193, 140, 104, 251, 193, 142, 104, 251, 193, 212, 104, 251, 193, 214, 104, 251, 193, 216, 104, 251, 193, 218, 104, 251, 193, 54, 105, 251, 193, 56, 105, 251, 193, 58, 105, 251, 193, 60, 105, 251, 193, 62, 105, 251, 193, 130, 105, 251, 193, 132, 105, 251, 193, 178, 105, 251, 193, 180, 105, 251, 193, 230, 105, 251, 193, 20, 106, 251, 193, 22, 106, 251, 193, 72, 106, 251, 193, 106, 106, 251, 193, 142, 106, 251, 193, 234, 106, 251, 193, 236, 106, 251, 193, 170, 107, 251, 193, 172, 107, 251, 193, 174, 107, 251, 193, 158, 108, 251, 193, 162, 109, 251, 193, 164, 109, 251, 193, 166, 109, 251, 193, 80, 104, 251, 193, 82, 104, 251, 193, 84, 104, 251, 193, 86, 104, 251, 193, 88, 104, 251, 193, 90, 104, 251, 193, 92, 104, 251, 193, 94, 104, 251, 193, 96, 104, 251, 193, 98, 104, 251, 193, 100, 104, 251, 193, 102, 104, 251, 193, 104, 104, 251, 193, 106, 104, 251, 193, 108, 104, 251, 193, 110, 104, 251, 193, 112, 104, 251, 193, 114, 104, 251, 193, 116, 104, 251, 193, 118, 104, 251, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 206, 0, 14, 0, 193, 238, 104, 251, 193, 240, 104, 251, 193, 242, 104, 251, 193, 244, 104, 251, 193, 246, 104, 251, 193, 248, 104, 251, 193, 250, 104, 251, 193, 252, 104, 251, 193, 254, 104, 251, 193, 2, 105, 251, 193, 4, 105, 251, 193, 6, 105, 251, 193, 8, 105, 251, 193, 10, 105, 251, 193, 12, 105, 251, 193, 14, 105, 251, 193, 16, 105, 251, 193, 18, 105, 251, 193, 20, 105, 251, 193, 22, 105, 251, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 193, 222, 104, 251, 193, 224, 104, 251, 193, 226, 104, 251, 193, 228, 104, 251, 193, 230, 104, 251, 193, 232, 104, 251, 193, 64, 105, 251, 193, 66, 105, 251, 193, 68, 105, 251, 193, 134, 105, 251, 193, 136, 105, 251, 193, 138, 105, 251, 193, 140, 105, 251, 193, 182, 105, 251, 193, 232, 105, 251, 193, 234, 105, 251, 193, 44, 106, 251, 193, 78, 106, 251, 193, 92, 106, 251, 193, 100, 106, 251, 193, 108, 106, 251, 193, 110, 106, 251, 193, 112, 106, 251, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 193, 70, 105, 251, 193, 236, 105, 251, 193, 114, 106, 251, 193, 180, 106, 251, 193, 182, 106, 251, 193, 184, 106, 251, 193, 186, 106, 251, 193, 188, 106, 251, 193, 190, 106, 251, 193, 52, 107, 251, 193, 54, 107, 251, 193, 56, 107, 251, 193, 58, 107, 251, 193, 60, 107, 251, 193, 62, 107, 251, 193, 64, 107, 251, 193, 10, 108, 251, 193, 12, 108, 251, 193, 14, 108, 251, 193, 16, 108, 251, 193, 18, 108, 251, 193, 20, 108, 251, 193, 22, 108, 251, 193, 24, 108, 251, 193, 26, 108, 251, 193, 28, 108, 251, 193, 30, 108, 251, 193, 32, 108, 251, 193, 34, 108, 251, 193, 36, 108, 251, 193, 38, 108, 251, 193, 40, 108, 251, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 206, 32, 14, 0, 193, 142, 105, 251, 193, 144, 105, 251, 193, 146, 105, 251, 193, 148, 105, 251, 193, 150, 105, 251, 193, 152, 105, 251, 193, 154, 105, 251, 193, 156, 105, 251, 193, 158, 105, 251, 193, 160, 105, 251, 193, 162, 105, 251, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 193, 176, 105, 251, 193, 14, 106, 251, 193, 16, 106, 251, 193, 18, 106, 251, 193, 42, 106, 251, 193, 52, 106, 251, 193, 62, 106, 251, 193, 132, 106, 251, 193, 134, 106, 251, 193, 136, 106, 251, 193, 138, 106, 251, 193, 140, 106, 251, 193, 230, 106, 251, 193, 232, 106, 251, 193, 166, 107, 251, 193, 168, 107, 251, 193, 146, 108, 251, 193, 148, 108, 251, 193, 150, 108, 251, 193, 152, 108, 251, 193, 154, 108, 251, 193, 156, 108, 251, 193, 154, 109, 251, 193, 156, 109, 251, 193, 158, 109, 251, 193, 160, 109, 251, 193, 196, 110, 251, 193, 198, 110, 251, 193, 200, 110, 251, 193, 202, 110, 251, 193, 204, 110, 251, 193, 206, 110, 251, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 206, 64, 14, 0, 193, 238, 105, 251, 193, 240, 105, 251, 193, 242, 105, 251, 193, 244, 105, 251, 193, 246, 105, 251, 193, 248, 105, 251, 193, 250, 105, 251, 193, 252, 105, 251, 193, 254, 105, 251, 193, 24, 106, 251, 193, 26, 106, 251, 193, 28, 106, 251, 193, 30, 106, 251, 193, 32, 106, 251, 193, 34, 106, 251, 193, 36, 106, 251, 193, 38, 106, 251, 193, 46, 106, 251, 193, 54, 106, 251, 193, 56, 106, 251, 193, 58, 106, 251, 193, 64, 106, 251, 193, 74, 106, 251, 193, 76, 106, 251, 193, 80, 106, 251, 193, 82, 106, 251, 193, 216, 105, 251, 193, 218, 105, 251, 193, 50, 106, 251, 193, 66, 106, 251, 193, 220, 106, 251, 193, 222, 106, 251, 193, 126, 107, 251, 193, 128, 107, 251, 193, 130, 107, 251, 193, 132, 107, 251, 193, 134, 107, 251, 193, 136, 107, 251, 193, 138, 107, 251, 193, 128, 108, 251, 193, 130, 108, 251, 193, 132, 108, 251, 193, 134, 108, 251, 193, 112, 109, 251, 193, 114, 109, 251, 193, 116, 109, 251, 193, 118, 109, 251, 193, 120, 109, 251, 193, 154, 110, 251, 193, 156, 110, 251, 193, 158, 110, 251, 193, 160, 110, 251, 193, 162, 110, 251, 206, 160, 15, 0, 206, 160, 15, 0, 206, 160, 15, 0, 206, 160, 15, 0, 206, 160, 15, 0, 193, 2, 106, 251, 193, 44, 60, 252, 193, 46, 60, 252, 193, 48, 60, 252, 193, 50, 60, 252, 193, 52, 60, 252, 193, 54, 60, 252, 193, 56, 60, 252, 193, 58, 60, 252, 193, 60, 60, 252, 193, 62, 60, 252, 193, 64, 60, 252, 193, 66, 60, 252, 193, 68, 60, 252, 193, 70, 60, 252, 193, 72, 60, 252, 193, 74, 60, 252, 193, 76, 60, 252, 193, 78, 60, 252, 193, 80, 60, 252, 193, 82, 60, 252, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 193, 12, 106, 251, 193, 40, 106, 251, 193, 60, 106, 251, 193, 68, 106, 251, 193, 70, 106, 251, 193, 90, 106, 251, 193, 104, 106, 251, 193, 130, 106, 251, 193, 224, 106, 251, 193, 226, 106, 251, 193, 228, 106, 251, 193, 142, 107, 251, 193, 144, 107, 251, 193, 146, 107, 251, 193, 148, 107, 251, 193, 150, 107, 251, 193, 152, 107, 251, 193, 154, 107, 251, 193, 156, 107, 251, 193, 158, 107, 251, 193, 160, 107, 251, 193, 162, 107, 251, 193, 164, 107, 251, 193, 136, 108, 251, 193, 138, 108, 251, 193, 140, 108, 251, 193, 142, 108, 251, 193, 144, 108, 251, 206, 64, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 193, 192, 145, 251, 193, 194, 145, 251, 193, 228, 145, 251, 193, 230, 145, 251, 193, 232, 145, 251, 193, 234, 145, 251, 193, 236, 145, 251, 193, 238, 145, 251, 193, 240, 145, 251, 193, 242, 145, 251, 193, 244, 145, 251, 193, 246, 145, 251, 193, 248, 145, 251, 193, 250, 145, 251, 193, 252, 145, 251, 193, 254, 145, 251, 193, 2, 146, 251, 193, 4, 146, 251, 193, 6, 146, 251, 193, 8, 146, 251, 193, 10, 146, 251, 193, 12, 146, 251, 193, 14, 146, 251, 193, 48, 106, 251, 193, 56, 146, 251, 193, 58, 146, 251, 193, 60, 146, 251, 193, 62, 146, 251, 193, 64, 146, 251, 193, 66, 146, 251, 193, 68, 146, 251, 193, 70, 146, 251, 193, 84, 106, 251, 193, 86, 106, 251, 193, 88, 106, 251, 193, 94, 106, 251, 193, 96, 106, 251, 193, 98, 106, 251, 193, 102, 106, 251, 193, 116, 106, 251, 193, 118, 106, 251, 193, 120, 106, 251, 193, 122, 106, 251, 193, 124, 106, 251, 193, 126, 106, 251, 193, 128, 106, 251, 193, 192, 106, 251, 193, 194, 106, 251, 193, 196, 106, 251, 193, 198, 106, 251, 193, 200, 106, 251, 193, 202, 106, 251, 193, 204, 106, 251, 193, 206, 106, 251, 193, 208, 106, 251, 193, 210, 106, 251, 193, 212, 106, 251, 193, 214, 106, 251, 193, 216, 106, 251, 193, 218, 106, 251, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 206, 96, 14, 0, 193, 238, 106, 251, 193, 240, 106, 251, 193, 242, 106, 251, 193, 244, 106, 251, 193, 160, 108, 251, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 128, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 193, 66, 107, 251, 193, 50, 110, 251, 193, 52, 110, 251, 193, 54, 110, 251, 193, 56, 110, 251, 193, 90, 111, 251, 193, 92, 111, 251, 193, 94, 111, 251, 193, 96, 111, 251, 193, 98, 111, 251, 193, 100, 111, 251, 193, 102, 111, 251, 193, 104, 111, 251, 193, 106, 111, 251, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 206, 160, 14, 0, 193, 46, 108, 251, 193, 48, 108, 251, 193, 50, 108, 251, 193, 52, 108, 251, 193, 54, 108, 251, 193, 56, 108, 251, 193, 58, 108, 251, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 192, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 193, 42, 108, 251, 193, 44, 108, 251, 193, 2, 109, 251, 193, 4, 109, 251, 193, 6, 109, 251, 193, 8, 109, 251, 193, 10, 109, 251, 193, 12, 109, 251, 193, 14, 109, 251, 193, 16, 109, 251, 193, 18, 109, 251, 193, 20, 109, 251, 193, 22, 109, 251, 193, 12, 110, 251, 193, 14, 110, 251, 193, 16, 110, 251, 193, 18, 110, 251, 193, 20, 110, 251, 193, 22, 110, 251, 193, 24, 110, 251, 193, 26, 110, 251, 193, 28, 110, 251, 193, 30, 110, 251, 193, 32, 110, 251, 193, 34, 110, 251, 193, 36, 110, 251, 193, 38, 110, 251, 193, 40, 110, 251, 193, 42, 110, 251, 193, 44, 110, 251, 193, 46, 110, 251, 193, 48, 110, 251, 193, 60, 108, 251, 193, 62, 108, 251, 193, 64, 108, 251, 193, 66, 108, 251, 193, 68, 108, 251, 193, 70, 108, 251, 193, 72, 108, 251, 193, 74, 108, 251, 193, 76, 108, 251, 193, 78, 108, 251, 193, 80, 108, 251, 193, 82, 108, 251, 193, 84, 108, 251, 193, 86, 108, 251, 193, 88, 108, 251, 193, 90, 108, 251, 193, 92, 108, 251, 193, 94, 108, 251, 193, 96, 108, 251, 193, 98, 108, 251, 193, 100, 108, 251, 193, 102, 108, 251, 193, 104, 108, 251, 193, 106, 108, 251, 193, 108, 108, 251, 193, 110, 108, 251, 193, 112, 108, 251, 193, 114, 108, 251, 193, 116, 108, 251, 193, 118, 108, 251, 193, 120, 108, 251, 193, 122, 108, 251, 193, 124, 108, 251, 193, 126, 108, 251, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 206, 224, 14, 0, 193, 176, 109, 251, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 206, 0, 15, 0, 193, 114, 168, 252, 193, 248, 108, 251, 193, 250, 108, 251, 193, 252, 108, 251, 193, 254, 108, 251, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 32, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 206, 64, 15, 0, 193, 170, 110, 251, 193, 172, 110, 251, 193, 174, 110, 251, 193, 176, 110, 251, 193, 178, 110, 251, 193, 180, 110, 251, 193, 182, 110, 251, 193, 184, 110, 251, 193, 186, 110, 251, 193, 188, 110, 251, 193, 190, 110, 251, 193, 192, 110, 251, 193, 194, 110, 251, 193, 254, 111, 251, 193, 2, 112, 251, 193, 4, 112, 251, 193, 6, 112, 251, 193, 8, 112, 251, 193, 10, 112, 251, 193, 12, 112, 251, 193, 14, 112, 251, 193, 16, 112, 251, 193, 168, 109, 251, 193, 170, 109, 251, 193, 172, 109, 251, 193, 174, 109, 251, 193, 208, 110, 251, 193, 210, 110, 251, 193, 34, 112, 251, 193, 36, 112, 251, 193, 38, 112, 251, 193, 40, 112, 251, 193, 42, 112, 251, 193, 44, 112, 251, 193, 174, 113, 251, 193, 176, 113, 251, 193, 178, 113, 251, 193, 174, 114, 251, 193, 176, 114, 251, 193, 178, 114, 251, 193, 180, 114, 251, 193, 182, 114, 251, 193, 184, 114, 251, 193, 196, 115, 251, 193, 194, 116, 251, 193, 196, 116, 251, 193, 144, 117, 251, 193, 146, 117, 251, 193, 148, 117, 251, 193, 54, 119, 251, 193, 168, 119, 251, 193, 170, 119, 251, 193, 208, 119, 251, 193, 210, 119, 251, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 206, 96, 15, 0, 193, 212, 110, 251, 193, 248, 109, 251, 193, 250, 109, 251, 193, 252, 109, 251, 193, 254, 109, 251, 193, 2, 110, 251, 193, 4, 110, 251, 193, 6, 110, 251, 193, 8, 110, 251, 193, 214, 110, 251, 193, 216, 110, 251, 193, 218, 110, 251, 193, 220, 110, 251, 193, 222, 110, 251, 193, 224, 110, 251, 193, 226, 110, 251, 193, 228, 110, 251, 193, 10, 110, 251, 193, 230, 110, 251, 193, 232, 110, 251, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 206, 128, 15, 0, 193, 108, 111, 251, 193, 110, 111, 251, 193, 234, 110, 251, 193, 236, 110, 251, 193, 238, 110, 251, 193, 240, 110, 251, 193, 242, 110, 251, 193, 244, 110, 251, 193, 246, 110, 251, 193, 248, 110, 251, 193, 250, 110, 251, 193, 252, 110, 251, 193, 254, 110, 251, 193, 2, 111, 251, 193, 4, 111, 251, 193, 6, 111, 251, 193, 8, 111, 251, 193, 10, 111, 251, 193, 12, 111, 251, 193, 14, 111, 251, 193, 16, 111, 251, 193, 18, 111, 251, 193, 20, 111, 251, 193, 22, 111, 251, 193, 24, 111, 251, 193, 26, 111, 251, 193, 28, 111, 251, 193, 30, 111, 251, 193, 32, 111, 251, 193, 34, 111, 251, 193, 36, 111, 251, 193, 38, 111, 251, 193, 40, 111, 251, 193, 42, 111, 251, 193, 44, 111, 251, 193, 46, 111, 251, 193, 48, 111, 251, 193, 50, 111, 251, 193, 52, 111, 251, 193, 54, 111, 251, 193, 56, 111, 251, 193, 58, 111, 251, 193, 60, 111, 251, 193, 62, 111, 251, 193, 64, 111, 251, 193, 66, 111, 251, 193, 68, 111, 251, 193, 70, 111, 251, 193, 72, 111, 251, 193, 74, 111, 251, 193, 76, 111, 251, 193, 78, 111, 251, 193, 80, 111, 251, 193, 82, 111, 251, 193, 84, 111, 251, 193, 86, 111, 251, 193, 88, 111, 251, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 193, 112, 111, 251, 193, 114, 111, 251, 193, 116, 111, 251, 193, 118, 111, 251, 193, 120, 111, 251, 193, 122, 111, 251, 193, 124, 111, 251, 193, 126, 111, 251, 193, 128, 111, 251, 193, 130, 111, 251, 193, 132, 111, 251, 193, 134, 111, 251, 193, 136, 111, 251, 193, 138, 111, 251, 193, 140, 111, 251, 193, 142, 111, 251, 193, 144, 111, 251, 193, 146, 111, 251, 193, 148, 111, 251, 193, 150, 111, 251, 193, 152, 111, 251, 193, 154, 111, 251, 193, 156, 111, 251, 193, 158, 111, 251, 193, 160, 111, 251, 193, 162, 111, 251, 193, 164, 111, 251, 193, 166, 111, 251, 193, 168, 111, 251, 193, 170, 111, 251, 193, 172, 111, 251, 193, 174, 111, 251, 193, 176, 111, 251, 193, 178, 111, 251, 193, 180, 111, 251, 193, 182, 111, 251, 193, 184, 111, 251, 193, 186, 111, 251, 193, 188, 111, 251, 193, 190, 111, 251, 193, 192, 111, 251, 193, 194, 111, 251, 193, 196, 111, 251, 193, 198, 111, 251, 193, 200, 111, 251, 193, 202, 111, 251, 193, 204, 111, 251, 193, 206, 111, 251, 193, 208, 111, 251, 193, 210, 111, 251, 193, 212, 111, 251, 193, 214, 111, 251, 193, 216, 111, 251, 193, 218, 111, 251, 193, 220, 111, 251, 193, 222, 111, 251, 193, 224, 111, 251, 193, 226, 111, 251, 193, 228, 111, 251, 193, 230, 111, 251, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 160, 15, 0, 206, 160, 15, 0, 206, 160, 15, 0, 206, 160, 15, 0, 206, 160, 15, 0, 206, 160, 15, 0, 193, 106, 113, 251, 193, 108, 113, 251, 193, 110, 113, 251, 193, 112, 113, 251, 193, 114, 113, 251, 193, 116, 113, 251, 193, 118, 113, 251, 193, 120, 113, 251, 193, 122, 113, 251, 193, 124, 113, 251, 193, 126, 113, 251, 193, 124, 114, 251, 193, 126, 114, 251, 193, 128, 114, 251, 193, 130, 114, 251, 193, 132, 114, 251, 193, 134, 114, 251, 193, 126, 115, 251, 193, 128, 115, 251, 193, 130, 115, 251, 193, 132, 115, 251, 193, 134, 115, 251, 193, 136, 115, 251, 193, 138, 115, 251, 193, 140, 115, 251, 193, 142, 115, 251, 193, 18, 112, 251, 193, 20, 112, 251, 193, 22, 112, 251, 193, 130, 113, 251, 193, 132, 113, 251, 193, 134, 113, 251, 193, 136, 113, 251, 193, 138, 113, 251, 193, 140, 113, 251, 193, 142, 113, 251, 193, 144, 113, 251, 193, 146, 113, 251, 193, 148, 113, 251, 193, 150, 113, 251, 193, 152, 113, 251, 193, 154, 113, 251, 193, 156, 113, 251, 193, 136, 114, 251, 193, 138, 114, 251, 193, 140, 114, 251, 193, 142, 114, 251, 193, 144, 114, 251, 193, 146, 114, 251, 193, 148, 114, 251, 193, 150, 114, 251, 193, 152, 114, 251, 193, 154, 114, 251, 193, 156, 114, 251, 193, 158, 114, 251, 193, 160, 114, 251, 193, 162, 114, 251, 193, 146, 115, 251, 193, 24, 112, 251, 193, 26, 112, 251, 193, 28, 112, 251, 193, 30, 112, 251, 193, 32, 112, 251, 193, 158, 113, 251, 193, 160, 113, 251, 193, 162, 113, 251, 193, 164, 113, 251, 193, 166, 113, 251, 193, 168, 113, 251, 193, 170, 113, 251, 193, 172, 113, 251, 193, 164, 114, 251, 193, 166, 114, 251, 193, 168, 114, 251, 193, 170, 114, 251, 193, 172, 114, 251, 193, 172, 115, 251, 193, 174, 115, 251, 193, 176, 115, 251, 193, 178, 115, 251, 193, 180, 115, 251, 193, 182, 115, 251, 193, 184, 115, 251, 193, 186, 115, 251, 193, 188, 115, 251, 193, 190, 115, 251, 193, 192, 115, 251, 193, 194, 115, 251, 193, 186, 116, 251, 193, 188, 116, 251, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 206, 192, 15, 0, 193, 180, 113, 251, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 224, 15, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 206, 64, 16, 0, 193, 158, 112, 251, 193, 244, 113, 251, 193, 246, 113, 251, 193, 248, 113, 251, 193, 250, 113, 251, 193, 252, 113, 251, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 206, 0, 16, 0, 193, 254, 113, 251, 193, 2, 114, 251, 193, 4, 114, 251, 193, 6, 114, 251, 193, 8, 114, 251, 193, 10, 114, 251, 193, 12, 114, 251, 193, 14, 114, 251, 193, 16, 114, 251, 193, 18, 114, 251, 193, 20, 114, 251, 193, 22, 114, 251, 193, 24, 114, 251, 193, 26, 114, 251, 193, 28, 114, 251, 193, 30, 114, 251, 193, 246, 114, 251, 193, 248, 114, 251, 193, 250, 114, 251, 193, 252, 114, 251, 193, 254, 114, 251, 193, 2, 115, 251, 193, 4, 115, 251, 193, 6, 115, 251, 193, 8, 115, 251, 193, 10, 115, 251, 193, 12, 115, 251, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 32, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 193, 36, 115, 251, 193, 38, 115, 251, 193, 40, 115, 251, 193, 104, 113, 251, 193, 42, 115, 251, 193, 44, 115, 251, 193, 46, 115, 251, 193, 48, 115, 251, 193, 50, 115, 251, 193, 52, 115, 251, 193, 54, 115, 251, 193, 56, 115, 251, 193, 58, 115, 251, 193, 60, 115, 251, 193, 62, 115, 251, 193, 64, 115, 251, 193, 66, 115, 251, 193, 68, 115, 251, 193, 18, 116, 251, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 206, 96, 16, 0, 193, 14, 116, 251, 193, 20, 115, 251, 193, 16, 116, 251, 193, 22, 115, 251, 193, 24, 115, 251, 193, 26, 115, 251, 193, 28, 115, 251, 193, 30, 115, 251, 193, 32, 115, 251, 193, 34, 115, 251, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 128, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 193, 14, 115, 251, 193, 16, 115, 251, 193, 18, 115, 251, 193, 2, 116, 251, 193, 4, 116, 251, 193, 6, 116, 251, 193, 8, 116, 251, 193, 10, 116, 251, 193, 12, 116, 251, 193, 238, 116, 251, 193, 240, 116, 251, 193, 242, 116, 251, 193, 244, 116, 251, 193, 246, 116, 251, 193, 190, 117, 251, 193, 192, 117, 251, 193, 46, 118, 251, 193, 48, 118, 251, 193, 126, 118, 251, 193, 128, 118, 251, 193, 130, 118, 251, 193, 132, 118, 251, 193, 202, 118, 251, 193, 204, 118, 251, 193, 206, 118, 251, 193, 250, 118, 251, 193, 28, 119, 251, 193, 94, 119, 251, 193, 96, 119, 251, 193, 152, 119, 251, 193, 182, 119, 251, 193, 184, 119, 251, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 160, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 193, 148, 115, 251, 193, 150, 115, 251, 193, 152, 115, 251, 193, 154, 115, 251, 193, 156, 115, 251, 193, 158, 115, 251, 193, 160, 115, 251, 193, 162, 115, 251, 193, 164, 115, 251, 193, 166, 115, 251, 193, 168, 115, 251, 193, 170, 115, 251, 193, 162, 116, 251, 193, 164, 116, 251, 193, 166, 116, 251, 193, 168, 116, 251, 193, 170, 116, 251, 193, 172, 116, 251, 193, 174, 116, 251, 193, 176, 116, 251, 193, 178, 116, 251, 193, 180, 116, 251, 193, 182, 116, 251, 193, 184, 116, 251, 206, 32, 17, 0, 206, 32, 17, 0, 206, 32, 17, 0, 206, 32, 17, 0, 206, 32, 17, 0, 206, 32, 17, 0, 206, 32, 17, 0, 206, 32, 17, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 206, 192, 16, 0, 193, 198, 116, 251, 193, 200, 116, 251, 193, 202, 116, 251, 193, 204, 116, 251, 193, 206, 116, 251, 193, 208, 116, 251, 193, 210, 116, 251, 193, 212, 116, 251, 193, 214, 116, 251, 193, 216, 116, 251, 193, 218, 116, 251, 193, 220, 116, 251, 193, 222, 116, 251, 193, 224, 116, 251, 193, 226, 116, 251, 193, 228, 116, 251, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 206, 224, 16, 0, 193, 248, 116, 251, 193, 112, 171, 252, 193, 146, 116, 251, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 193, 148, 116, 251, 193, 150, 116, 251, 193, 152, 116, 251, 193, 154, 116, 251, 193, 156, 116, 251, 193, 158, 116, 251, 193, 100, 117, 251, 193, 102, 117, 251, 193, 104, 117, 251, 193, 106, 117, 251, 193, 108, 117, 251, 193, 12, 118, 251, 193, 88, 118, 251, 193, 90, 118, 251, 193, 92, 118, 251, 193, 94, 118, 251, 193, 170, 118, 251, 193, 172, 118, 251, 193, 228, 118, 251, 193, 14, 119, 251, 193, 16, 119, 251, 193, 64, 119, 251, 193, 200, 119, 251, 193, 202, 119, 251, 193, 204, 119, 251, 193, 206, 119, 251, 193, 12, 120, 251, 193, 68, 120, 251, 193, 144, 120, 251, 193, 146, 120, 251, 193, 148, 120, 251, 193, 204, 120, 251, 193, 190, 116, 251, 193, 192, 116, 251, 193, 136, 117, 251, 193, 138, 117, 251, 193, 140, 117, 251, 193, 142, 117, 251, 193, 30, 118, 251, 193, 32, 118, 251, 193, 34, 118, 251, 193, 110, 118, 251, 193, 112, 118, 251, 193, 114, 118, 251, 193, 116, 118, 251, 193, 118, 118, 251, 193, 186, 118, 251, 193, 234, 118, 251, 193, 236, 118, 251, 193, 238, 118, 251, 193, 240, 118, 251, 193, 242, 118, 251, 193, 84, 119, 251, 193, 92, 119, 251, 193, 136, 119, 251, 193, 138, 119, 251, 193, 140, 119, 251, 193, 142, 119, 251, 193, 162, 119, 251, 193, 164, 119, 251, 193, 166, 119, 251, 193, 20, 120, 251, 193, 22, 120, 251, 193, 24, 120, 251, 193, 230, 116, 251, 193, 232, 116, 251, 193, 234, 116, 251, 193, 150, 117, 251, 193, 152, 117, 251, 193, 154, 117, 251, 193, 156, 117, 251, 193, 158, 117, 251, 193, 160, 117, 251, 193, 162, 117, 251, 193, 164, 117, 251, 193, 166, 117, 251, 193, 168, 117, 251, 193, 170, 117, 251, 193, 172, 117, 251, 193, 174, 117, 251, 193, 176, 117, 251, 193, 178, 117, 251, 193, 180, 117, 251, 193, 182, 117, 251, 193, 236, 116, 251, 193, 184, 117, 251, 193, 186, 117, 251, 193, 188, 117, 251, 193, 36, 118, 251, 193, 120, 118, 251, 193, 38, 118, 251, 193, 40, 118, 251, 193, 42, 118, 251, 193, 44, 118, 251, 193, 122, 118, 251, 193, 124, 118, 251, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 206, 0, 17, 0, 193, 194, 117, 251, 193, 196, 117, 251, 193, 198, 117, 251, 193, 200, 117, 251, 193, 202, 117, 251, 193, 204, 117, 251, 193, 206, 117, 251, 193, 208, 117, 251, 193, 210, 117, 251, 193, 212, 117, 251, 193, 214, 117, 251, 193, 216, 117, 251, 193, 218, 117, 251, 193, 220, 117, 251, 193, 222, 117, 251, 193, 224, 117, 251, 193, 226, 117, 251, 193, 228, 117, 251, 193, 230, 117, 251, 193, 232, 117, 251, 193, 234, 117, 251, 193, 236, 117, 251, 193, 238, 117, 251, 193, 240, 117, 251, 193, 242, 117, 251, 193, 244, 117, 251, 193, 246, 117, 251, 206, 32, 17, 0, 206, 32, 17, 0, 206, 32, 17, 0, 206, 32, 17, 0, 206, 32, 17, 0, 193, 14, 118, 251, 193, 16, 118, 251, 193, 18, 118, 251, 193, 20, 118, 251, 193, 22, 118, 251, 193, 24, 118, 251, 193, 26, 118, 251, 193, 28, 118, 251, 193, 96, 118, 251, 193, 98, 118, 251, 193, 100, 118, 251, 193, 102, 118, 251, 193, 104, 118, 251, 193, 106, 118, 251, 193, 108, 118, 251, 193, 174, 118, 251, 193, 176, 118, 251, 193, 178, 118, 251, 193, 180, 118, 251, 193, 182, 118, 251, 193, 184, 118, 251, 193, 230, 118, 251, 193, 232, 118, 251, 193, 18, 119, 251, 193, 20, 119, 251, 193, 50, 119, 251, 193, 52, 119, 251, 193, 248, 117, 251, 193, 250, 117, 251, 193, 184, 98, 253, 193, 252, 117, 251, 193, 254, 117, 251, 193, 2, 118, 251, 193, 4, 118, 251, 193, 6, 118, 251, 193, 8, 118, 251, 193, 10, 118, 251, 193, 50, 118, 251, 193, 52, 118, 251, 193, 54, 118, 251, 193, 56, 118, 251, 193, 58, 118, 251, 193, 60, 118, 251, 193, 62, 118, 251, 193, 64, 118, 251, 193, 66, 118, 251, 193, 68, 118, 251, 193, 70, 118, 251, 193, 72, 118, 251, 193, 74, 118, 251, 193, 76, 118, 251, 193, 78, 118, 251, 193, 80, 118, 251, 193, 82, 118, 251, 193, 84, 118, 251, 193, 86, 118, 251, 193, 134, 118, 251, 193, 136, 118, 251, 193, 138, 118, 251, 193, 140, 118, 251, 193, 142, 118, 251, 193, 144, 118, 251, 193, 146, 118, 251, 193, 148, 118, 251, 193, 150, 118, 251, 193, 152, 118, 251, 193, 154, 118, 251, 193, 156, 118, 251, 193, 158, 118, 251, 193, 160, 118, 251, 193, 162, 118, 251, 193, 164, 118, 251, 193, 166, 118, 251, 193, 208, 118, 251, 193, 210, 118, 251, 193, 212, 118, 251, 193, 214, 118, 251, 193, 216, 118, 251, 193, 218, 118, 251, 193, 220, 118, 251, 193, 222, 118, 251, 193, 224, 118, 251, 193, 226, 118, 251, 193, 252, 118, 251, 193, 254, 118, 251, 193, 2, 119, 251, 193, 4, 119, 251, 193, 6, 119, 251, 193, 8, 119, 251, 193, 168, 118, 251, 193, 10, 119, 251, 193, 188, 118, 251, 193, 190, 118, 251, 193, 192, 118, 251, 193, 194, 118, 251, 193, 196, 118, 251, 193, 198, 118, 251, 193, 200, 118, 251, 193, 244, 118, 251, 193, 246, 118, 251, 193, 248, 118, 251, 193, 22, 119, 251, 193, 24, 119, 251, 193, 26, 119, 251, 193, 56, 119, 251, 193, 68, 119, 251, 193, 70, 119, 251, 193, 108, 119, 251, 193, 110, 119, 251, 193, 112, 119, 251, 193, 114, 119, 251, 193, 118, 119, 251, 193, 122, 119, 251, 193, 144, 119, 251, 193, 146, 119, 251, 193, 148, 119, 251, 193, 150, 119, 251, 193, 172, 119, 251, 193, 174, 119, 251, 193, 176, 119, 251, 193, 178, 119, 251, 193, 180, 119, 251, 193, 218, 119, 251, 193, 12, 119, 251, 193, 30, 119, 251, 193, 32, 119, 251, 193, 34, 119, 251, 193, 36, 119, 251, 193, 38, 119, 251, 193, 40, 119, 251, 193, 42, 119, 251, 193, 44, 119, 251, 193, 46, 119, 251, 193, 48, 119, 251, 193, 58, 119, 251, 193, 60, 119, 251, 193, 62, 119, 251, 193, 72, 119, 251, 193, 74, 119, 251, 193, 76, 119, 251, 193, 78, 119, 251, 193, 80, 119, 251, 193, 82, 119, 251, 193, 86, 119, 251, 193, 88, 119, 251, 193, 90, 119, 251, 193, 98, 119, 251, 193, 100, 119, 251, 193, 102, 119, 251, 193, 104, 119, 251, 193, 106, 119, 251, 193, 116, 119, 251, 193, 120, 119, 251, 193, 124, 119, 251, 193, 126, 119, 251, 193, 66, 119, 251, 193, 134, 119, 251, 193, 14, 120, 251, 193, 16, 120, 251, 193, 18, 120, 251, 193, 70, 120, 251, 193, 106, 120, 251, 193, 150, 120, 251, 193, 152, 120, 251, 193, 206, 120, 251, 193, 208, 120, 251, 193, 6, 121, 251, 193, 8, 121, 251, 193, 36, 121, 251, 193, 38, 121, 251, 193, 64, 121, 251, 193, 66, 121, 251, 193, 98, 121, 251, 193, 100, 121, 251, 193, 138, 121, 251, 193, 166, 121, 251, 193, 242, 121, 251, 193, 244, 121, 251, 193, 52, 122, 251, 193, 54, 122, 251, 193, 56, 122, 251, 193, 58, 122, 251, 193, 60, 122, 251, 193, 62, 122, 251, 193, 64, 122, 251, 193, 152, 122, 251, 193, 154, 122, 251, 193, 212, 119, 251, 193, 214, 119, 251, 193, 216, 119, 251, 193, 28, 120, 251, 193, 30, 120, 251, 193, 32, 120, 251, 193, 76, 120, 251, 193, 78, 120, 251, 193, 118, 120, 251, 193, 168, 120, 251, 193, 170, 120, 251, 193, 172, 120, 251, 193, 222, 120, 251, 193, 224, 120, 251, 193, 14, 121, 251, 193, 44, 121, 251, 193, 106, 121, 251, 193, 126, 121, 251, 193, 158, 121, 251, 193, 198, 121, 251, 193, 248, 121, 251, 193, 250, 121, 251, 193, 72, 122, 251, 193, 74, 122, 251, 193, 76, 122, 251, 193, 174, 122, 251, 193, 176, 122, 251, 193, 26, 123, 251, 193, 28, 123, 251, 193, 30, 123, 251, 193, 182, 123, 251, 193, 184, 123, 251, 193, 220, 119, 251, 193, 222, 119, 251, 193, 224, 119, 251, 193, 226, 119, 251, 193, 228, 119, 251, 193, 230, 119, 251, 193, 232, 119, 251, 193, 234, 119, 251, 193, 34, 120, 251, 193, 36, 120, 251, 193, 38, 120, 251, 193, 80, 120, 251, 193, 82, 120, 251, 193, 84, 120, 251, 193, 120, 120, 251, 193, 122, 120, 251, 193, 124, 120, 251, 193, 174, 120, 251, 193, 126, 120, 251, 193, 226, 120, 251, 193, 228, 120, 251, 193, 16, 121, 251, 193, 18, 121, 251, 193, 20, 121, 251, 193, 68, 121, 251, 193, 70, 121, 251, 193, 108, 121, 251, 193, 110, 121, 251, 193, 128, 121, 251, 193, 142, 121, 251, 193, 144, 121, 251, 193, 162, 121, 251, 193, 236, 119, 251, 193, 40, 120, 251, 193, 86, 120, 251, 193, 88, 120, 251, 193, 176, 120, 251, 193, 230, 120, 251, 193, 46, 121, 251, 193, 72, 121, 251, 193, 168, 121, 251, 193, 176, 121, 251, 193, 188, 121, 251, 193, 190, 121, 251, 193, 210, 121, 251, 193, 212, 121, 251, 193, 16, 122, 251, 193, 18, 122, 251, 193, 20, 122, 251, 193, 22, 122, 251, 193, 108, 122, 251, 193, 196, 122, 251, 193, 198, 122, 251, 193, 200, 122, 251, 193, 202, 122, 251, 193, 72, 123, 251, 193, 74, 123, 251, 193, 76, 123, 251, 193, 78, 123, 251, 193, 210, 123, 251, 193, 212, 123, 251, 193, 214, 123, 251, 193, 216, 123, 251, 193, 218, 123, 251, 193, 26, 120, 251, 193, 72, 120, 251, 193, 74, 120, 251, 193, 108, 120, 251, 193, 110, 120, 251, 193, 112, 120, 251, 193, 114, 120, 251, 193, 116, 120, 251, 193, 154, 120, 251, 193, 156, 120, 251, 193, 158, 120, 251, 193, 160, 120, 251, 193, 162, 120, 251, 193, 164, 120, 251, 193, 166, 120, 251, 193, 210, 120, 251, 193, 212, 120, 251, 193, 214, 120, 251, 193, 216, 120, 251, 193, 218, 120, 251, 193, 220, 120, 251, 193, 10, 121, 251, 193, 12, 121, 251, 193, 40, 121, 251, 193, 42, 121, 251, 193, 102, 121, 251, 193, 104, 121, 251, 193, 118, 121, 251, 193, 140, 121, 251, 193, 172, 121, 251, 193, 246, 121, 251, 193, 66, 122, 251, 206, 64, 17, 0, 206, 64, 17, 0, 206, 64, 17, 0, 206, 64, 17, 0, 206, 64, 17, 0, 206, 64, 17, 0, 206, 64, 17, 0, 206, 64, 17, 0, 193, 90, 120, 251, 193, 92, 120, 251, 193, 94, 120, 251, 193, 96, 120, 251, 193, 98, 120, 251, 193, 100, 120, 251, 193, 128, 120, 251, 193, 130, 120, 251, 193, 132, 120, 251, 193, 134, 120, 251, 193, 136, 120, 251, 193, 138, 120, 251, 193, 140, 120, 251, 193, 142, 120, 251, 193, 178, 120, 251, 193, 180, 120, 251, 193, 182, 120, 251, 193, 184, 120, 251, 193, 186, 120, 251, 193, 188, 120, 251, 193, 190, 120, 251, 193, 192, 120, 251, 193, 194, 120, 251, 193, 196, 120, 251, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 193, 102, 120, 251, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 193, 104, 120, 251, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 193, 198, 120, 251, 193, 200, 120, 251, 193, 202, 120, 251, 193, 232, 120, 251, 193, 234, 120, 251, 193, 236, 120, 251, 193, 238, 120, 251, 193, 240, 120, 251, 193, 242, 120, 251, 193, 244, 120, 251, 193, 246, 120, 251, 193, 248, 120, 251, 193, 250, 120, 251, 193, 252, 120, 251, 193, 254, 120, 251, 193, 2, 121, 251, 193, 4, 121, 251, 193, 22, 121, 251, 193, 24, 121, 251, 193, 26, 121, 251, 193, 28, 121, 251, 193, 30, 121, 251, 193, 32, 121, 251, 193, 48, 121, 251, 193, 50, 121, 251, 193, 52, 121, 251, 193, 54, 121, 251, 193, 56, 121, 251, 193, 58, 121, 251, 193, 60, 121, 251, 193, 74, 121, 251, 193, 76, 121, 251, 193, 34, 121, 251, 193, 96, 121, 251, 193, 114, 121, 251, 193, 116, 121, 251, 193, 160, 121, 251, 193, 46, 122, 251, 193, 48, 122, 251, 193, 50, 122, 251, 193, 148, 122, 251, 193, 150, 122, 251, 193, 250, 122, 251, 193, 252, 122, 251, 193, 254, 122, 251, 193, 148, 123, 251, 193, 150, 123, 251, 193, 152, 123, 251, 193, 154, 123, 251, 193, 24, 124, 251, 193, 26, 124, 251, 193, 28, 124, 251, 193, 182, 124, 251, 193, 184, 124, 251, 193, 186, 124, 251, 193, 188, 124, 251, 193, 190, 124, 251, 193, 192, 124, 251, 193, 194, 124, 251, 193, 72, 125, 251, 193, 74, 125, 251, 193, 76, 125, 251, 193, 78, 125, 251, 193, 80, 125, 251, 193, 78, 121, 251, 193, 80, 121, 251, 193, 82, 121, 251, 193, 84, 121, 251, 193, 86, 121, 251, 193, 88, 121, 251, 193, 90, 121, 251, 193, 92, 121, 251, 193, 94, 121, 251, 193, 112, 121, 251, 193, 120, 121, 251, 193, 122, 121, 251, 193, 124, 121, 251, 193, 130, 121, 251, 193, 132, 121, 251, 193, 134, 121, 251, 193, 136, 121, 251, 193, 146, 121, 251, 193, 148, 121, 251, 193, 150, 121, 251, 193, 152, 121, 251, 193, 154, 121, 251, 193, 156, 121, 251, 193, 164, 121, 251, 193, 170, 121, 251, 193, 178, 121, 251, 193, 192, 121, 251, 193, 194, 121, 251, 193, 196, 121, 251, 193, 214, 121, 251, 193, 216, 121, 251, 193, 218, 121, 251, 193, 174, 121, 251, 193, 180, 121, 251, 193, 182, 121, 251, 193, 184, 121, 251, 193, 186, 121, 251, 193, 200, 121, 251, 193, 202, 121, 251, 193, 204, 121, 251, 193, 206, 121, 251, 193, 208, 121, 251, 193, 252, 121, 251, 193, 254, 121, 251, 193, 2, 122, 251, 193, 4, 122, 251, 193, 6, 122, 251, 193, 8, 122, 251, 193, 10, 122, 251, 193, 12, 122, 251, 193, 14, 122, 251, 193, 78, 122, 251, 193, 80, 122, 251, 193, 82, 122, 251, 193, 84, 122, 251, 193, 86, 122, 251, 193, 88, 122, 251, 193, 90, 122, 251, 193, 92, 122, 251, 193, 94, 122, 251, 193, 96, 122, 251, 193, 98, 122, 251, 193, 100, 122, 251, 193, 102, 122, 251, 193, 220, 121, 251, 193, 222, 121, 251, 193, 224, 121, 251, 193, 226, 121, 251, 193, 228, 121, 251, 193, 230, 121, 251, 193, 232, 121, 251, 193, 234, 121, 251, 193, 236, 121, 251, 193, 238, 121, 251, 193, 240, 121, 251, 193, 24, 122, 251, 193, 26, 122, 251, 193, 28, 122, 251, 193, 30, 122, 251, 193, 32, 122, 251, 193, 34, 122, 251, 193, 36, 122, 251, 193, 38, 122, 251, 193, 40, 122, 251, 193, 42, 122, 251, 193, 44, 122, 251, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 193, 68, 122, 251, 193, 70, 122, 251, 193, 158, 122, 251, 193, 160, 122, 251, 193, 162, 122, 251, 193, 164, 122, 251, 193, 166, 122, 251, 193, 168, 122, 251, 193, 170, 122, 251, 193, 172, 122, 251, 193, 14, 123, 251, 193, 16, 123, 251, 193, 18, 123, 251, 193, 20, 123, 251, 193, 22, 123, 251, 193, 24, 123, 251, 193, 174, 123, 251, 193, 176, 123, 251, 193, 178, 123, 251, 193, 180, 123, 251, 193, 50, 124, 251, 193, 52, 124, 251, 193, 54, 124, 251, 193, 56, 124, 251, 193, 58, 124, 251, 193, 60, 124, 251, 193, 62, 124, 251, 193, 64, 124, 251, 193, 66, 124, 251, 193, 68, 124, 251, 193, 70, 124, 251, 193, 72, 124, 251, 193, 104, 122, 251, 193, 106, 122, 251, 193, 178, 122, 251, 193, 180, 122, 251, 193, 182, 122, 251, 193, 184, 122, 251, 193, 186, 122, 251, 193, 32, 123, 251, 193, 188, 122, 251, 193, 190, 122, 251, 193, 192, 122, 251, 193, 194, 122, 251, 193, 34, 123, 251, 193, 36, 123, 251, 193, 38, 123, 251, 193, 40, 123, 251, 193, 42, 123, 251, 193, 44, 123, 251, 193, 46, 123, 251, 193, 48, 123, 251, 193, 50, 123, 251, 193, 52, 123, 251, 193, 54, 123, 251, 193, 56, 123, 251, 193, 58, 123, 251, 193, 60, 123, 251, 193, 62, 123, 251, 193, 64, 123, 251, 193, 66, 123, 251, 193, 68, 123, 251, 193, 70, 123, 251, 193, 188, 123, 251, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 206, 96, 17, 0, 193, 204, 122, 251, 193, 206, 122, 251, 193, 208, 122, 251, 193, 210, 122, 251, 193, 212, 122, 251, 193, 214, 122, 251, 193, 216, 122, 251, 193, 218, 122, 251, 193, 220, 122, 251, 193, 222, 122, 251, 193, 224, 122, 251, 193, 226, 122, 251, 193, 228, 122, 251, 193, 230, 122, 251, 193, 232, 122, 251, 193, 234, 122, 251, 193, 236, 122, 251, 193, 238, 122, 251, 193, 240, 122, 251, 193, 242, 122, 251, 193, 244, 122, 251, 193, 246, 122, 251, 193, 248, 122, 251, 193, 156, 122, 251, 193, 2, 123, 251, 193, 4, 123, 251, 193, 6, 123, 251, 193, 8, 123, 251, 193, 10, 123, 251, 193, 12, 123, 251, 193, 156, 123, 251, 193, 158, 123, 251, 193, 160, 123, 251, 193, 162, 123, 251, 193, 164, 123, 251, 193, 166, 123, 251, 193, 168, 123, 251, 193, 170, 123, 251, 193, 172, 123, 251, 193, 30, 124, 251, 193, 32, 124, 251, 193, 34, 124, 251, 193, 36, 124, 251, 193, 38, 124, 251, 193, 40, 124, 251, 193, 42, 124, 251, 193, 44, 124, 251, 193, 46, 124, 251, 193, 48, 124, 251, 206, 128, 17, 0, 206, 128, 17, 0, 206, 128, 17, 0, 206, 128, 17, 0, 206, 128, 17, 0, 206, 128, 17, 0, 193, 80, 123, 251, 193, 82, 123, 251, 193, 84, 123, 251, 193, 86, 123, 251, 193, 88, 123, 251, 193, 90, 123, 251, 193, 92, 123, 251, 193, 94, 123, 251, 193, 96, 123, 251, 193, 98, 123, 251, 193, 100, 123, 251, 193, 102, 123, 251, 193, 104, 123, 251, 193, 106, 123, 251, 193, 108, 123, 251, 193, 110, 123, 251, 193, 112, 123, 251, 193, 114, 123, 251, 193, 116, 123, 251, 193, 118, 123, 251, 193, 120, 123, 251, 193, 122, 123, 251, 193, 124, 123, 251, 193, 126, 123, 251, 193, 128, 123, 251, 193, 130, 123, 251, 193, 132, 123, 251, 193, 134, 123, 251, 193, 136, 123, 251, 193, 138, 123, 251, 193, 140, 123, 251, 193, 142, 123, 251, 193, 144, 123, 251, 193, 146, 123, 251, 193, 224, 123, 251, 193, 226, 123, 251, 193, 228, 123, 251, 193, 230, 123, 251, 193, 232, 123, 251, 193, 234, 123, 251, 193, 236, 123, 251, 193, 238, 123, 251, 193, 240, 123, 251, 193, 242, 123, 251, 193, 244, 123, 251, 193, 246, 123, 251, 193, 248, 123, 251, 193, 250, 123, 251, 193, 252, 123, 251, 193, 254, 123, 251, 193, 2, 124, 251, 193, 4, 124, 251, 193, 6, 124, 251, 193, 8, 124, 251, 193, 10, 124, 251, 193, 12, 124, 251, 193, 14, 124, 251, 193, 16, 124, 251, 193, 18, 124, 251, 193, 20, 124, 251, 193, 22, 124, 251, 193, 114, 124, 251, 193, 116, 124, 251, 193, 118, 124, 251, 193, 186, 123, 251, 193, 74, 124, 251, 193, 76, 124, 251, 193, 78, 124, 251, 193, 80, 124, 251, 193, 82, 124, 251, 193, 228, 124, 251, 193, 230, 124, 251, 193, 232, 124, 251, 193, 234, 124, 251, 193, 236, 124, 251, 193, 238, 124, 251, 193, 240, 124, 251, 193, 242, 124, 251, 193, 120, 125, 251, 193, 122, 125, 251, 193, 124, 125, 251, 193, 126, 125, 251, 193, 128, 125, 251, 193, 130, 125, 251, 193, 132, 125, 251, 193, 134, 125, 251, 193, 136, 125, 251, 193, 138, 125, 251, 193, 34, 126, 251, 193, 36, 126, 251, 193, 38, 126, 251, 193, 152, 126, 251, 193, 154, 126, 251, 193, 240, 126, 251, 193, 242, 126, 251, 193, 244, 126, 251, 193, 190, 123, 251, 193, 192, 123, 251, 193, 194, 123, 251, 193, 196, 123, 251, 193, 198, 123, 251, 193, 200, 123, 251, 193, 202, 123, 251, 193, 204, 123, 251, 193, 206, 123, 251, 193, 208, 123, 251, 193, 84, 124, 251, 193, 86, 124, 251, 193, 88, 124, 251, 193, 90, 124, 251, 193, 92, 124, 251, 193, 94, 124, 251, 193, 96, 124, 251, 193, 98, 124, 251, 193, 100, 124, 251, 193, 102, 124, 251, 193, 104, 124, 251, 193, 106, 124, 251, 193, 244, 124, 251, 193, 246, 124, 251, 193, 248, 124, 251, 193, 250, 124, 251, 193, 252, 124, 251, 193, 254, 124, 251, 193, 2, 125, 251, 193, 4, 125, 251, 193, 140, 125, 251, 193, 142, 125, 251, 193, 220, 123, 251, 193, 222, 123, 251, 193, 110, 124, 251, 193, 112, 124, 251, 193, 6, 125, 251, 193, 8, 125, 251, 193, 10, 125, 251, 193, 12, 125, 251, 193, 14, 125, 251, 193, 162, 125, 251, 193, 164, 125, 251, 193, 166, 125, 251, 193, 168, 125, 251, 193, 50, 126, 251, 193, 52, 126, 251, 193, 54, 126, 251, 193, 52, 127, 251, 193, 172, 127, 251, 193, 254, 127, 251, 193, 28, 128, 251, 193, 48, 128, 251, 193, 54, 128, 251, 193, 108, 128, 251, 193, 152, 128, 251, 193, 172, 129, 251, 193, 244, 129, 251, 193, 38, 130, 251, 193, 62, 130, 251, 193, 178, 130, 251, 193, 200, 130, 251, 193, 228, 130, 251, 193, 230, 130, 251, 193, 144, 125, 251, 193, 146, 125, 251, 193, 148, 125, 251, 193, 150, 125, 251, 193, 152, 125, 251, 193, 154, 125, 251, 193, 156, 125, 251, 193, 158, 125, 251, 193, 160, 125, 251, 193, 40, 126, 251, 193, 108, 124, 251, 193, 42, 126, 251, 193, 44, 126, 251, 193, 46, 126, 251, 193, 48, 126, 251, 193, 156, 126, 251, 193, 158, 126, 251, 193, 246, 126, 251, 193, 248, 126, 251, 193, 104, 127, 251, 193, 106, 127, 251, 193, 108, 127, 251, 193, 146, 127, 251, 193, 170, 127, 251, 193, 22, 128, 251, 193, 26, 128, 251, 193, 30, 128, 251, 193, 32, 128, 251, 193, 34, 128, 251, 193, 42, 128, 251, 193, 44, 128, 251, 193, 46, 128, 251, 193, 120, 124, 251, 193, 122, 124, 251, 193, 124, 124, 251, 193, 126, 124, 251, 193, 128, 124, 251, 193, 130, 124, 251, 193, 132, 124, 251, 193, 134, 124, 251, 193, 136, 124, 251, 193, 138, 124, 251, 193, 140, 124, 251, 193, 142, 124, 251, 193, 144, 124, 251, 193, 146, 124, 251, 193, 148, 124, 251, 193, 150, 124, 251, 193, 152, 124, 251, 193, 154, 124, 251, 193, 156, 124, 251, 193, 158, 124, 251, 193, 160, 124, 251, 193, 162, 124, 251, 193, 164, 124, 251, 193, 166, 124, 251, 193, 168, 124, 251, 193, 170, 124, 251, 193, 172, 124, 251, 193, 174, 124, 251, 193, 176, 124, 251, 193, 178, 124, 251, 193, 180, 124, 251, 193, 16, 125, 251, 206, 128, 17, 0, 206, 128, 17, 0, 206, 128, 17, 0, 206, 128, 17, 0, 193, 84, 125, 251, 193, 86, 125, 251, 193, 88, 125, 251, 193, 90, 125, 251, 193, 92, 125, 251, 193, 94, 125, 251, 193, 96, 125, 251, 193, 98, 125, 251, 193, 100, 125, 251, 193, 102, 125, 251, 193, 104, 125, 251, 193, 252, 125, 251, 193, 254, 125, 251, 193, 2, 126, 251, 193, 4, 126, 251, 193, 6, 126, 251, 193, 8, 126, 251, 193, 10, 126, 251, 193, 12, 126, 251, 193, 14, 126, 251, 193, 16, 126, 251, 193, 18, 126, 251, 193, 20, 126, 251, 193, 22, 126, 251, 193, 24, 126, 251, 193, 130, 126, 251, 193, 132, 126, 251, 193, 134, 126, 251, 193, 216, 124, 251, 193, 218, 124, 251, 193, 220, 124, 251, 193, 222, 124, 251, 193, 224, 124, 251, 193, 226, 124, 251, 193, 106, 125, 251, 193, 108, 125, 251, 193, 110, 125, 251, 193, 112, 125, 251, 193, 114, 125, 251, 193, 116, 125, 251, 193, 118, 125, 251, 193, 26, 126, 251, 193, 28, 126, 251, 193, 30, 126, 251, 193, 32, 126, 251, 193, 140, 126, 251, 193, 142, 126, 251, 193, 144, 126, 251, 193, 146, 126, 251, 193, 148, 126, 251, 193, 150, 126, 251, 193, 234, 126, 251, 193, 236, 126, 251, 193, 238, 126, 251, 193, 40, 127, 251, 193, 42, 127, 251, 193, 44, 127, 251, 193, 46, 127, 251, 193, 48, 127, 251, 193, 96, 127, 251, 193, 18, 125, 251, 193, 20, 125, 251, 193, 22, 125, 251, 193, 24, 125, 251, 193, 26, 125, 251, 193, 28, 125, 251, 193, 30, 125, 251, 193, 32, 125, 251, 193, 34, 125, 251, 193, 36, 125, 251, 193, 38, 125, 251, 193, 40, 125, 251, 193, 42, 125, 251, 193, 44, 125, 251, 193, 46, 125, 251, 193, 48, 125, 251, 193, 50, 125, 251, 193, 52, 125, 251, 193, 54, 125, 251, 193, 56, 125, 251, 193, 58, 125, 251, 193, 60, 125, 251, 193, 62, 125, 251, 193, 64, 125, 251, 193, 66, 125, 251, 193, 68, 125, 251, 193, 70, 125, 251, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 206, 160, 17, 0, 193, 82, 125, 251, 193, 240, 125, 251, 193, 242, 125, 251, 193, 244, 125, 251, 193, 246, 125, 251, 193, 248, 125, 251, 193, 118, 126, 251, 193, 120, 126, 251, 193, 122, 126, 251, 193, 124, 126, 251, 193, 126, 126, 251, 193, 198, 126, 251, 193, 200, 126, 251, 193, 202, 126, 251, 193, 204, 126, 251, 193, 22, 127, 251, 193, 82, 127, 251, 193, 84, 127, 251, 193, 86, 127, 251, 193, 158, 127, 251, 193, 160, 127, 251, 193, 190, 127, 251, 193, 236, 127, 251, 193, 38, 128, 251, 193, 136, 128, 251, 193, 138, 128, 251, 193, 166, 128, 251, 193, 242, 128, 251, 193, 244, 128, 251, 193, 8, 129, 251, 193, 18, 129, 251, 193, 100, 129, 251, 193, 58, 126, 251, 193, 60, 126, 251, 193, 62, 126, 251, 193, 64, 126, 251, 193, 66, 126, 251, 193, 68, 126, 251, 193, 70, 126, 251, 193, 72, 126, 251, 193, 74, 126, 251, 193, 76, 126, 251, 193, 78, 126, 251, 193, 80, 126, 251, 193, 82, 126, 251, 193, 84, 126, 251, 193, 86, 126, 251, 193, 88, 126, 251, 193, 90, 126, 251, 193, 92, 126, 251, 193, 94, 126, 251, 193, 96, 126, 251, 193, 98, 126, 251, 193, 100, 126, 251, 193, 102, 126, 251, 193, 104, 126, 251, 193, 106, 126, 251, 193, 108, 126, 251, 193, 110, 126, 251, 193, 112, 126, 251, 193, 114, 126, 251, 193, 116, 126, 251, 193, 160, 126, 251, 193, 162, 126, 251, 193, 136, 126, 251, 193, 138, 126, 251, 193, 206, 126, 251, 193, 208, 126, 251, 193, 210, 126, 251, 193, 212, 126, 251, 193, 214, 126, 251, 193, 216, 126, 251, 193, 218, 126, 251, 193, 220, 126, 251, 193, 222, 126, 251, 193, 224, 126, 251, 193, 226, 126, 251, 193, 228, 126, 251, 193, 230, 126, 251, 193, 232, 126, 251, 193, 24, 127, 251, 193, 26, 127, 251, 193, 28, 127, 251, 193, 30, 127, 251, 193, 32, 127, 251, 193, 34, 127, 251, 193, 36, 127, 251, 193, 38, 127, 251, 193, 88, 127, 251, 193, 90, 127, 251, 193, 92, 127, 251, 193, 94, 127, 251, 193, 132, 127, 251, 193, 134, 127, 251, 193, 136, 127, 251, 193, 138, 127, 251, 193, 164, 126, 251, 193, 166, 126, 251, 193, 168, 126, 251, 193, 170, 126, 251, 193, 172, 126, 251, 193, 174, 126, 251, 193, 176, 126, 251, 193, 178, 126, 251, 193, 180, 126, 251, 193, 182, 126, 251, 193, 184, 126, 251, 193, 186, 126, 251, 193, 188, 126, 251, 193, 190, 126, 251, 193, 192, 126, 251, 193, 194, 126, 251, 193, 196, 126, 251, 193, 250, 126, 251, 193, 252, 126, 251, 193, 254, 126, 251, 193, 2, 127, 251, 193, 4, 127, 251, 193, 6, 127, 251, 193, 8, 127, 251, 193, 10, 127, 251, 193, 12, 127, 251, 193, 14, 127, 251, 193, 16, 127, 251, 193, 18, 127, 251, 193, 20, 127, 251, 193, 54, 127, 251, 193, 56, 127, 251, 193, 50, 127, 251, 193, 102, 127, 251, 193, 168, 127, 251, 193, 228, 127, 251, 193, 244, 127, 251, 193, 222, 128, 251, 193, 254, 128, 251, 193, 40, 129, 251, 193, 42, 129, 251, 193, 66, 129, 251, 193, 110, 129, 251, 193, 170, 129, 251, 193, 118, 130, 251, 193, 252, 130, 251, 193, 18, 131, 251, 193, 44, 131, 251, 193, 200, 131, 251, 193, 42, 132, 251, 193, 44, 132, 251, 193, 110, 132, 251, 193, 112, 132, 251, 193, 204, 132, 251, 193, 206, 132, 251, 193, 208, 132, 251, 193, 36, 133, 251, 193, 38, 133, 251, 193, 40, 133, 251, 193, 42, 133, 251, 193, 122, 133, 251, 193, 124, 133, 251, 193, 126, 133, 251, 193, 198, 133, 251, 193, 58, 127, 251, 193, 60, 127, 251, 193, 62, 127, 251, 193, 64, 127, 251, 193, 66, 127, 251, 193, 68, 127, 251, 193, 70, 127, 251, 193, 72, 127, 251, 193, 74, 127, 251, 193, 76, 127, 251, 193, 78, 127, 251, 193, 80, 127, 251, 193, 110, 127, 251, 193, 112, 127, 251, 193, 114, 127, 251, 193, 116, 127, 251, 193, 118, 127, 251, 193, 120, 127, 251, 193, 122, 127, 251, 193, 124, 127, 251, 193, 126, 127, 251, 193, 128, 127, 251, 193, 130, 127, 251, 193, 148, 127, 251, 193, 150, 127, 251, 193, 152, 127, 251, 193, 154, 127, 251, 193, 156, 127, 251, 206, 192, 17, 0, 206, 192, 17, 0, 206, 192, 17, 0, 206, 192, 17, 0, 193, 98, 127, 251, 193, 100, 127, 251, 193, 140, 127, 251, 193, 142, 127, 251, 193, 144, 127, 251, 193, 164, 127, 251, 193, 166, 127, 251, 193, 198, 127, 251, 193, 220, 127, 251, 193, 40, 128, 251, 193, 50, 128, 251, 193, 72, 128, 251, 193, 74, 128, 251, 193, 96, 128, 251, 193, 98, 128, 251, 193, 100, 128, 251, 193, 114, 128, 251, 193, 116, 128, 251, 193, 118, 128, 251, 193, 140, 128, 251, 193, 142, 128, 251, 193, 144, 128, 251, 193, 172, 128, 251, 193, 174, 128, 251, 193, 176, 128, 251, 193, 194, 128, 251, 193, 196, 128, 251, 193, 198, 128, 251, 193, 200, 128, 251, 193, 216, 128, 251, 193, 218, 128, 251, 193, 220, 128, 251, 193, 162, 127, 251, 193, 192, 127, 251, 193, 194, 127, 251, 193, 196, 127, 251, 193, 216, 127, 251, 193, 218, 127, 251, 193, 224, 127, 251, 193, 226, 127, 251, 193, 238, 127, 251, 193, 240, 127, 251, 193, 242, 127, 251, 193, 18, 128, 251, 193, 94, 128, 251, 193, 112, 128, 251, 193, 168, 128, 251, 193, 170, 128, 251, 193, 186, 128, 251, 193, 188, 128, 251, 193, 190, 128, 251, 193, 192, 128, 251, 193, 246, 128, 251, 193, 248, 128, 251, 193, 30, 129, 251, 193, 32, 129, 251, 193, 102, 129, 251, 193, 104, 129, 251, 193, 130, 129, 251, 193, 132, 129, 251, 193, 164, 129, 251, 193, 166, 129, 251, 193, 230, 129, 251, 193, 232, 129, 251, 206, 192, 17, 0, 206, 192, 17, 0, 206, 192, 17, 0, 206, 192, 17, 0, 193, 200, 127, 251, 193, 202, 127, 251, 193, 204, 127, 251, 193, 206, 127, 251, 193, 208, 127, 251, 193, 210, 127, 251, 193, 212, 127, 251, 193, 214, 127, 251, 193, 222, 127, 251, 193, 230, 127, 251, 193, 8, 128, 251, 193, 232, 127, 251, 193, 234, 127, 251, 193, 246, 127, 251, 193, 14, 128, 251, 193, 248, 127, 251, 193, 250, 127, 251, 193, 252, 127, 251, 193, 2, 128, 251, 193, 4, 128, 251, 193, 6, 128, 251, 193, 10, 128, 251, 193, 12, 128, 251, 193, 16, 128, 251, 193, 20, 128, 251, 193, 24, 128, 251, 193, 36, 128, 251, 193, 56, 128, 251, 193, 52, 128, 251, 193, 76, 128, 251, 193, 78, 128, 251, 193, 102, 128, 251, 193, 104, 128, 251, 193, 106, 128, 251, 193, 80, 128, 251, 193, 120, 128, 251, 193, 122, 128, 251, 193, 124, 128, 251, 193, 146, 128, 251, 193, 148, 128, 251, 193, 150, 128, 251, 193, 202, 128, 251, 193, 224, 128, 251, 193, 34, 129, 251, 193, 44, 129, 251, 193, 46, 129, 251, 193, 52, 129, 251, 193, 54, 129, 251, 193, 56, 129, 251, 193, 58, 129, 251, 193, 68, 129, 251, 193, 70, 129, 251, 193, 72, 129, 251, 193, 112, 129, 251, 193, 114, 129, 251, 193, 138, 129, 251, 193, 140, 129, 251, 193, 142, 129, 251, 193, 242, 129, 251, 193, 16, 130, 251, 193, 58, 128, 251, 193, 60, 128, 251, 193, 62, 128, 251, 193, 64, 128, 251, 193, 66, 128, 251, 193, 68, 128, 251, 193, 70, 128, 251, 193, 82, 128, 251, 193, 84, 128, 251, 193, 86, 128, 251, 193, 88, 128, 251, 193, 90, 128, 251, 193, 92, 128, 251, 193, 110, 128, 251, 193, 126, 128, 251, 193, 128, 128, 251, 193, 130, 128, 251, 193, 132, 128, 251, 193, 134, 128, 251, 193, 154, 128, 251, 193, 156, 128, 251, 193, 158, 128, 251, 193, 160, 128, 251, 193, 162, 128, 251, 193, 164, 128, 251, 193, 178, 128, 251, 193, 180, 128, 251, 193, 182, 128, 251, 193, 184, 128, 251, 193, 204, 128, 251, 193, 206, 128, 251, 193, 208, 128, 251, 193, 210, 128, 251, 193, 212, 128, 251, 193, 214, 128, 251, 193, 226, 128, 251, 193, 228, 128, 251, 193, 230, 128, 251, 193, 232, 128, 251, 193, 234, 128, 251, 193, 236, 128, 251, 193, 238, 128, 251, 193, 2, 129, 251, 193, 4, 129, 251, 193, 6, 129, 251, 193, 12, 129, 251, 193, 14, 129, 251, 193, 16, 129, 251, 193, 20, 129, 251, 193, 22, 129, 251, 193, 26, 129, 251, 193, 28, 129, 251, 193, 36, 129, 251, 193, 48, 129, 251, 193, 50, 129, 251, 193, 62, 129, 251, 193, 64, 129, 251, 193, 74, 129, 251, 193, 76, 129, 251, 193, 78, 129, 251, 193, 80, 129, 251, 193, 84, 129, 251, 193, 86, 129, 251, 193, 88, 129, 251, 206, 224, 34, 0, 206, 224, 34, 0, 206, 224, 34, 0, 206, 224, 34, 0, 206, 224, 34, 0, 206, 224, 34, 0, 206, 224, 34, 0, 206, 224, 34, 0, 206, 224, 34, 0, 193, 32, 230, 251, 193, 108, 229, 251, 193, 110, 229, 251, 193, 112, 229, 251, 193, 114, 229, 251, 193, 116, 229, 251, 193, 118, 229, 251, 193, 120, 229, 251, 193, 122, 229, 251, 193, 240, 128, 251, 193, 124, 229, 251, 193, 126, 229, 251, 193, 128, 229, 251, 193, 130, 229, 251, 193, 132, 229, 251, 193, 134, 229, 251, 193, 136, 229, 251, 193, 138, 229, 251, 193, 140, 229, 251, 193, 142, 229, 251, 193, 144, 229, 251, 193, 34, 230, 251, 193, 36, 230, 251, 193, 250, 128, 251, 193, 252, 128, 251, 193, 10, 129, 251, 193, 24, 129, 251, 193, 38, 129, 251, 193, 82, 129, 251, 193, 106, 129, 251, 193, 108, 129, 251, 193, 134, 129, 251, 193, 136, 129, 251, 193, 168, 129, 251, 193, 34, 130, 251, 193, 106, 130, 251, 193, 188, 130, 251, 193, 250, 130, 251, 193, 68, 131, 251, 193, 70, 131, 251, 193, 130, 131, 251, 193, 132, 131, 251, 193, 144, 131, 251, 193, 220, 131, 251, 193, 102, 132, 251, 193, 104, 132, 251, 193, 106, 132, 251, 193, 108, 132, 251, 193, 194, 132, 251, 193, 196, 132, 251, 193, 198, 132, 251, 193, 200, 132, 251, 193, 202, 132, 251, 193, 32, 133, 251, 193, 34, 133, 251, 193, 90, 129, 251, 193, 92, 129, 251, 193, 94, 129, 251, 193, 96, 129, 251, 193, 98, 129, 251, 193, 118, 129, 251, 193, 120, 129, 251, 193, 122, 129, 251, 193, 124, 129, 251, 193, 126, 129, 251, 193, 144, 129, 251, 193, 146, 129, 251, 193, 148, 129, 251, 193, 150, 129, 251, 193, 152, 129, 251, 193, 154, 129, 251, 193, 156, 129, 251, 193, 158, 129, 251, 193, 160, 129, 251, 193, 162, 129, 251, 193, 174, 129, 251, 193, 176, 129, 251, 193, 178, 129, 251, 193, 180, 129, 251, 193, 182, 129, 251, 193, 184, 129, 251, 193, 186, 129, 251, 193, 188, 129, 251, 193, 190, 129, 251, 193, 196, 129, 251, 193, 198, 129, 251, 193, 200, 129, 251, 193, 128, 129, 251, 193, 192, 129, 251, 193, 194, 129, 251, 193, 56, 130, 251, 193, 84, 130, 251, 193, 98, 130, 251, 193, 116, 130, 251, 193, 100, 131, 251, 193, 128, 131, 251, 193, 30, 132, 251, 193, 32, 132, 251, 193, 34, 132, 251, 193, 94, 132, 251, 193, 180, 132, 251, 193, 182, 132, 251, 193, 24, 133, 251, 193, 26, 133, 251, 193, 110, 133, 251, 193, 250, 133, 251, 193, 252, 133, 251, 193, 254, 133, 251, 193, 188, 133, 251, 193, 190, 133, 251, 193, 192, 133, 251, 193, 194, 133, 251, 193, 58, 134, 251, 193, 60, 134, 251, 193, 118, 134, 251, 193, 120, 134, 251, 193, 158, 134, 251, 193, 172, 134, 251, 193, 206, 134, 251, 193, 202, 129, 251, 193, 204, 129, 251, 193, 206, 129, 251, 193, 208, 129, 251, 193, 210, 129, 251, 193, 212, 129, 251, 193, 214, 129, 251, 193, 216, 129, 251, 193, 218, 129, 251, 193, 220, 129, 251, 193, 222, 129, 251, 193, 224, 129, 251, 193, 226, 129, 251, 193, 228, 129, 251, 193, 246, 129, 251, 193, 248, 129, 251, 193, 250, 129, 251, 193, 252, 129, 251, 193, 254, 129, 251, 193, 2, 130, 251, 193, 4, 130, 251, 193, 6, 130, 251, 193, 8, 130, 251, 193, 10, 130, 251, 193, 22, 130, 251, 193, 24, 130, 251, 193, 26, 130, 251, 193, 28, 130, 251, 193, 30, 130, 251, 193, 32, 130, 251, 193, 40, 130, 251, 193, 42, 130, 251, 193, 234, 129, 251, 193, 236, 129, 251, 193, 238, 129, 251, 193, 240, 129, 251, 193, 12, 130, 251, 193, 14, 130, 251, 193, 58, 130, 251, 193, 60, 130, 251, 193, 86, 130, 251, 193, 88, 130, 251, 193, 100, 130, 251, 193, 102, 130, 251, 193, 104, 130, 251, 193, 138, 130, 251, 193, 170, 130, 251, 193, 224, 130, 251, 193, 248, 130, 251, 193, 66, 131, 251, 193, 160, 131, 251, 193, 36, 132, 251, 193, 38, 132, 251, 193, 40, 132, 251, 193, 96, 132, 251, 193, 98, 132, 251, 193, 100, 132, 251, 193, 184, 132, 251, 193, 186, 132, 251, 193, 188, 132, 251, 193, 190, 132, 251, 193, 192, 132, 251, 193, 28, 133, 251, 193, 30, 133, 251, 193, 44, 130, 251, 193, 46, 130, 251, 193, 48, 130, 251, 193, 50, 130, 251, 193, 52, 130, 251, 193, 54, 130, 251, 193, 64, 130, 251, 193, 66, 130, 251, 193, 68, 130, 251, 193, 70, 130, 251, 193, 72, 130, 251, 193, 74, 130, 251, 193, 76, 130, 251, 193, 78, 130, 251, 193, 80, 130, 251, 193, 82, 130, 251, 193, 90, 130, 251, 193, 92, 130, 251, 193, 94, 130, 251, 193, 96, 130, 251, 193, 108, 130, 251, 193, 110, 130, 251, 193, 112, 130, 251, 193, 120, 130, 251, 193, 122, 130, 251, 193, 124, 130, 251, 193, 126, 130, 251, 193, 128, 130, 251, 193, 132, 130, 251, 193, 134, 130, 251, 193, 136, 130, 251, 193, 140, 130, 251, 193, 208, 92, 252, 193, 212, 92, 252, 193, 214, 92, 252, 193, 216, 92, 252, 193, 218, 92, 252, 193, 228, 92, 252, 193, 238, 92, 252, 193, 250, 92, 252, 193, 4, 93, 252, 193, 14, 93, 252, 193, 16, 93, 252, 193, 20, 93, 252, 193, 22, 93, 252, 193, 26, 93, 252, 193, 28, 93, 252, 193, 130, 130, 251, 193, 30, 93, 252, 193, 32, 93, 252, 193, 34, 93, 252, 193, 38, 93, 252, 193, 40, 93, 252, 193, 44, 93, 252, 193, 46, 93, 252, 193, 48, 93, 252, 193, 58, 93, 252, 193, 60, 93, 252, 193, 62, 93, 252, 193, 64, 93, 252, 193, 66, 93, 252, 193, 68, 93, 252, 193, 70, 93, 252, 193, 78, 93, 252, 193, 142, 130, 251, 193, 148, 130, 251, 193, 150, 130, 251, 193, 154, 130, 251, 193, 156, 130, 251, 193, 158, 130, 251, 193, 160, 130, 251, 193, 162, 130, 251, 193, 164, 130, 251, 193, 166, 130, 251, 193, 168, 130, 251, 193, 180, 130, 251, 193, 182, 130, 251, 193, 184, 130, 251, 193, 186, 130, 251, 193, 202, 130, 251, 193, 204, 130, 251, 193, 206, 130, 251, 193, 208, 130, 251, 193, 210, 130, 251, 193, 212, 130, 251, 193, 214, 130, 251, 193, 216, 130, 251, 193, 218, 130, 251, 193, 220, 130, 251, 193, 222, 130, 251, 193, 232, 130, 251, 193, 234, 130, 251, 193, 236, 130, 251, 193, 238, 130, 251, 193, 240, 130, 251, 193, 242, 130, 251, 193, 244, 130, 251, 193, 246, 130, 251, 193, 8, 131, 251, 193, 10, 131, 251, 193, 12, 131, 251, 193, 240, 254, 252, 193, 14, 131, 251, 193, 16, 131, 251, 193, 22, 131, 251, 193, 24, 131, 251, 193, 26, 131, 251, 193, 28, 131, 251, 193, 30, 131, 251, 193, 32, 131, 251, 193, 34, 131, 251, 193, 36, 131, 251, 193, 38, 131, 251, 193, 40, 131, 251, 193, 42, 131, 251, 193, 54, 131, 251, 193, 56, 131, 251, 193, 58, 131, 251, 193, 60, 131, 251, 193, 62, 131, 251, 193, 64, 131, 251, 193, 80, 131, 251, 193, 82, 131, 251, 193, 84, 131, 251, 193, 86, 131, 251, 193, 88, 131, 251, 193, 90, 131, 251, 193, 92, 131, 251, 193, 4, 131, 251, 193, 6, 131, 251, 193, 20, 131, 251, 193, 46, 131, 251, 193, 48, 131, 251, 193, 50, 131, 251, 193, 52, 131, 251, 193, 76, 131, 251, 193, 78, 131, 251, 193, 108, 131, 251, 193, 110, 131, 251, 193, 206, 131, 251, 193, 224, 131, 251, 193, 226, 131, 251, 193, 10, 132, 251, 193, 60, 132, 251, 193, 62, 132, 251, 193, 64, 132, 251, 193, 124, 132, 251, 193, 126, 132, 251, 193, 128, 132, 251, 193, 130, 132, 251, 193, 132, 132, 251, 193, 228, 132, 251, 193, 54, 133, 251, 193, 56, 133, 251, 193, 136, 133, 251, 193, 138, 133, 251, 193, 28, 134, 251, 193, 30, 134, 251, 193, 32, 134, 251, 193, 84, 134, 251, 193, 94, 131, 251, 193, 96, 131, 251, 193, 98, 131, 251, 193, 112, 131, 251, 193, 114, 131, 251, 193, 116, 131, 251, 193, 118, 131, 251, 193, 120, 131, 251, 193, 122, 131, 251, 193, 124, 131, 251, 193, 126, 131, 251, 193, 134, 131, 251, 193, 136, 131, 251, 193, 138, 131, 251, 193, 140, 131, 251, 193, 142, 131, 251, 193, 148, 131, 251, 193, 150, 131, 251, 193, 152, 131, 251, 193, 154, 131, 251, 193, 156, 131, 251, 193, 158, 131, 251, 193, 162, 131, 251, 193, 164, 131, 251, 193, 166, 131, 251, 193, 168, 131, 251, 193, 170, 131, 251, 193, 176, 131, 251, 193, 178, 131, 251, 193, 180, 131, 251, 193, 182, 131, 251, 193, 184, 131, 251, 193, 186, 131, 251, 193, 188, 131, 251, 193, 190, 131, 251, 193, 194, 131, 251, 193, 208, 131, 251, 193, 210, 131, 251, 193, 212, 131, 251, 193, 214, 131, 251, 193, 216, 131, 251, 193, 228, 131, 251, 193, 230, 131, 251, 193, 232, 131, 251, 193, 234, 131, 251, 193, 236, 131, 251, 193, 238, 131, 251, 193, 240, 131, 251, 193, 242, 131, 251, 193, 244, 131, 251, 193, 246, 131, 251, 193, 248, 131, 251, 193, 218, 131, 251, 193, 12, 132, 251, 193, 14, 132, 251, 193, 16, 132, 251, 193, 18, 132, 251, 193, 20, 132, 251, 193, 22, 132, 251, 193, 24, 132, 251, 193, 26, 132, 251, 193, 28, 132, 251, 193, 66, 132, 251, 193, 68, 132, 251, 193, 2, 132, 251, 193, 4, 132, 251, 193, 6, 132, 251, 193, 8, 132, 251, 193, 46, 132, 251, 193, 48, 132, 251, 193, 50, 132, 251, 193, 52, 132, 251, 193, 54, 132, 251, 193, 56, 132, 251, 193, 58, 132, 251, 193, 114, 132, 251, 193, 116, 132, 251, 193, 118, 132, 251, 193, 120, 132, 251, 193, 122, 132, 251, 193, 210, 132, 251, 193, 212, 132, 251, 193, 214, 132, 251, 193, 216, 132, 251, 193, 218, 132, 251, 193, 220, 132, 251, 193, 222, 132, 251, 193, 224, 132, 251, 193, 226, 132, 251, 193, 44, 133, 251, 193, 46, 133, 251, 193, 48, 133, 251, 193, 50, 133, 251, 193, 52, 133, 251, 193, 128, 133, 251, 193, 130, 133, 251, 193, 70, 132, 251, 193, 72, 132, 251, 193, 74, 132, 251, 193, 76, 132, 251, 193, 78, 132, 251, 193, 80, 132, 251, 193, 82, 132, 251, 193, 84, 132, 251, 193, 86, 132, 251, 193, 88, 132, 251, 193, 90, 132, 251, 193, 92, 132, 251, 193, 134, 132, 251, 193, 136, 132, 251, 193, 138, 132, 251, 193, 140, 132, 251, 193, 142, 132, 251, 193, 144, 132, 251, 193, 146, 132, 251, 193, 148, 132, 251, 193, 150, 132, 251, 193, 152, 132, 251, 193, 154, 132, 251, 193, 156, 132, 251, 193, 158, 132, 251, 193, 160, 132, 251, 193, 162, 132, 251, 193, 164, 132, 251, 193, 166, 132, 251, 193, 168, 132, 251, 193, 170, 132, 251, 193, 172, 132, 251, 193, 174, 132, 251, 193, 176, 132, 251, 193, 230, 132, 251, 193, 232, 132, 251, 193, 234, 132, 251, 193, 236, 132, 251, 193, 238, 132, 251, 193, 240, 132, 251, 193, 242, 132, 251, 193, 244, 132, 251, 193, 246, 132, 251, 193, 248, 132, 251, 193, 250, 132, 251, 193, 252, 132, 251, 193, 254, 132, 251, 193, 2, 133, 251, 193, 4, 133, 251, 193, 6, 133, 251, 193, 8, 133, 251, 193, 10, 133, 251, 193, 12, 133, 251, 193, 14, 133, 251, 193, 16, 133, 251, 193, 18, 133, 251, 193, 20, 133, 251, 193, 22, 133, 251, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 224, 17, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 193, 112, 133, 251, 193, 2, 134, 251, 193, 62, 134, 251, 193, 64, 134, 251, 193, 66, 134, 251, 193, 68, 134, 251, 193, 160, 134, 251, 193, 162, 134, 251, 193, 174, 134, 251, 193, 186, 134, 251, 193, 210, 134, 251, 193, 212, 134, 251, 193, 236, 134, 251, 193, 238, 134, 251, 193, 84, 135, 251, 193, 110, 135, 251, 193, 112, 135, 251, 193, 186, 135, 251, 193, 188, 135, 251, 193, 190, 135, 251, 193, 192, 135, 251, 193, 34, 136, 251, 193, 212, 136, 251, 193, 214, 136, 251, 193, 176, 137, 251, 193, 178, 137, 251, 193, 132, 138, 251, 193, 134, 138, 251, 193, 152, 139, 251, 193, 154, 139, 251, 193, 156, 139, 251, 193, 158, 139, 251, 193, 114, 133, 251, 193, 116, 133, 251, 193, 118, 133, 251, 193, 120, 133, 251, 193, 196, 133, 251, 193, 4, 134, 251, 193, 6, 134, 251, 193, 8, 134, 251, 193, 10, 134, 251, 193, 12, 134, 251, 193, 14, 134, 251, 193, 16, 134, 251, 193, 18, 134, 251, 193, 20, 134, 251, 193, 70, 134, 251, 193, 72, 134, 251, 193, 122, 134, 251, 193, 124, 134, 251, 193, 126, 134, 251, 193, 128, 134, 251, 193, 130, 134, 251, 193, 132, 134, 251, 193, 176, 134, 251, 193, 188, 134, 251, 193, 190, 134, 251, 193, 214, 134, 251, 193, 216, 134, 251, 193, 240, 134, 251, 193, 250, 134, 251, 193, 2, 135, 251, 193, 20, 135, 251, 193, 36, 135, 251, 193, 132, 133, 251, 193, 134, 133, 251, 193, 24, 134, 251, 193, 26, 134, 251, 193, 76, 134, 251, 193, 78, 134, 251, 193, 80, 134, 251, 193, 82, 134, 251, 193, 166, 134, 251, 193, 178, 134, 251, 193, 180, 134, 251, 193, 192, 134, 251, 193, 228, 134, 251, 193, 252, 134, 251, 193, 8, 135, 251, 193, 14, 135, 251, 193, 22, 135, 251, 193, 128, 135, 251, 193, 132, 135, 251, 193, 140, 135, 251, 193, 142, 135, 251, 193, 144, 135, 251, 193, 146, 135, 251, 193, 148, 135, 251, 206, 96, 18, 0, 206, 96, 18, 0, 206, 96, 18, 0, 206, 96, 18, 0, 206, 96, 18, 0, 206, 96, 18, 0, 206, 96, 18, 0, 206, 96, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 0, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 193, 200, 133, 251, 193, 22, 134, 251, 193, 74, 134, 251, 193, 134, 134, 251, 193, 136, 134, 251, 193, 138, 134, 251, 193, 164, 134, 251, 193, 218, 134, 251, 193, 104, 135, 251, 193, 44, 136, 251, 193, 228, 136, 251, 193, 230, 136, 251, 193, 232, 136, 251, 193, 234, 136, 251, 193, 200, 137, 251, 193, 202, 137, 251, 193, 204, 137, 251, 193, 206, 137, 251, 193, 208, 137, 251, 193, 210, 137, 251, 193, 156, 138, 251, 193, 158, 138, 251, 193, 160, 138, 251, 193, 162, 138, 251, 193, 164, 138, 251, 193, 166, 138, 251, 193, 168, 138, 251, 193, 170, 138, 251, 193, 172, 138, 251, 193, 174, 138, 251, 193, 176, 138, 251, 193, 182, 139, 251, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 206, 32, 18, 0, 193, 34, 134, 251, 193, 36, 134, 251, 193, 38, 134, 251, 193, 40, 134, 251, 193, 42, 134, 251, 193, 44, 134, 251, 193, 46, 134, 251, 193, 48, 134, 251, 193, 50, 134, 251, 193, 52, 134, 251, 193, 54, 134, 251, 193, 56, 134, 251, 193, 88, 134, 251, 193, 90, 134, 251, 193, 92, 134, 251, 193, 94, 134, 251, 193, 96, 134, 251, 193, 98, 134, 251, 193, 100, 134, 251, 193, 102, 134, 251, 193, 104, 134, 251, 193, 106, 134, 251, 193, 108, 134, 251, 193, 110, 134, 251, 193, 112, 134, 251, 193, 114, 134, 251, 193, 116, 134, 251, 193, 140, 134, 251, 193, 86, 134, 251, 193, 194, 134, 251, 193, 16, 135, 251, 193, 38, 135, 251, 193, 150, 135, 251, 193, 152, 135, 251, 193, 154, 135, 251, 193, 156, 135, 251, 193, 220, 135, 251, 193, 222, 135, 251, 193, 224, 135, 251, 193, 226, 135, 251, 193, 228, 135, 251, 193, 230, 135, 251, 193, 232, 135, 251, 193, 234, 135, 251, 193, 94, 136, 251, 193, 96, 136, 251, 193, 98, 136, 251, 193, 100, 136, 251, 193, 102, 136, 251, 193, 104, 136, 251, 193, 106, 136, 251, 193, 40, 135, 251, 193, 108, 136, 251, 206, 128, 18, 0, 206, 128, 18, 0, 206, 128, 18, 0, 206, 128, 18, 0, 206, 128, 18, 0, 206, 128, 18, 0, 206, 128, 18, 0, 193, 142, 134, 251, 193, 144, 134, 251, 193, 146, 134, 251, 193, 148, 134, 251, 193, 150, 134, 251, 193, 152, 134, 251, 193, 154, 134, 251, 193, 156, 134, 251, 193, 168, 134, 251, 193, 170, 134, 251, 193, 182, 134, 251, 193, 184, 134, 251, 193, 196, 134, 251, 193, 198, 134, 251, 193, 200, 134, 251, 193, 202, 134, 251, 193, 204, 134, 251, 193, 220, 134, 251, 193, 222, 134, 251, 193, 224, 134, 251, 193, 226, 134, 251, 193, 230, 134, 251, 193, 232, 134, 251, 193, 234, 134, 251, 193, 242, 134, 251, 193, 244, 134, 251, 193, 246, 134, 251, 193, 254, 134, 251, 193, 4, 135, 251, 193, 6, 135, 251, 193, 12, 135, 251, 193, 18, 135, 251, 193, 208, 134, 251, 193, 248, 134, 251, 193, 24, 136, 251, 193, 26, 136, 251, 193, 28, 136, 251, 193, 30, 136, 251, 193, 206, 136, 251, 193, 208, 136, 251, 193, 210, 136, 251, 193, 164, 137, 251, 193, 166, 137, 251, 193, 168, 137, 251, 193, 170, 137, 251, 193, 172, 137, 251, 193, 174, 137, 251, 193, 118, 138, 251, 193, 120, 138, 251, 193, 122, 138, 251, 193, 124, 138, 251, 193, 126, 138, 251, 193, 128, 138, 251, 193, 138, 139, 251, 193, 140, 139, 251, 193, 142, 139, 251, 193, 144, 139, 251, 193, 146, 139, 251, 193, 148, 139, 251, 193, 234, 140, 251, 193, 236, 140, 251, 193, 238, 140, 251, 193, 240, 140, 251, 193, 242, 140, 251, 193, 24, 135, 251, 193, 26, 135, 251, 193, 28, 135, 251, 193, 30, 135, 251, 193, 32, 135, 251, 193, 34, 135, 251, 193, 42, 135, 251, 193, 44, 135, 251, 193, 46, 135, 251, 193, 48, 135, 251, 193, 50, 135, 251, 193, 52, 135, 251, 193, 54, 135, 251, 193, 56, 135, 251, 193, 58, 135, 251, 193, 60, 135, 251, 193, 62, 135, 251, 193, 64, 135, 251, 193, 66, 135, 251, 193, 68, 135, 251, 193, 70, 135, 251, 193, 72, 135, 251, 193, 74, 135, 251, 193, 76, 135, 251, 193, 78, 135, 251, 193, 80, 135, 251, 193, 82, 135, 251, 206, 64, 18, 0, 206, 64, 18, 0, 206, 64, 18, 0, 206, 64, 18, 0, 206, 64, 18, 0, 193, 106, 135, 251, 193, 108, 135, 251, 193, 114, 135, 251, 193, 116, 135, 251, 193, 118, 135, 251, 193, 120, 135, 251, 193, 122, 135, 251, 193, 124, 135, 251, 193, 126, 135, 251, 193, 134, 135, 251, 193, 136, 135, 251, 193, 138, 135, 251, 193, 158, 135, 251, 193, 160, 135, 251, 193, 162, 135, 251, 193, 164, 135, 251, 193, 166, 135, 251, 193, 168, 135, 251, 193, 170, 135, 251, 193, 172, 135, 251, 193, 174, 135, 251, 193, 176, 135, 251, 193, 178, 135, 251, 193, 180, 135, 251, 193, 182, 135, 251, 193, 184, 135, 251, 193, 236, 135, 251, 193, 238, 135, 251, 193, 194, 135, 251, 193, 36, 136, 251, 193, 38, 136, 251, 193, 40, 136, 251, 193, 42, 136, 251, 193, 216, 136, 251, 193, 218, 136, 251, 193, 220, 136, 251, 193, 222, 136, 251, 193, 224, 136, 251, 193, 226, 136, 251, 193, 180, 137, 251, 193, 182, 137, 251, 193, 184, 137, 251, 193, 186, 137, 251, 193, 188, 137, 251, 193, 190, 137, 251, 193, 192, 137, 251, 193, 194, 137, 251, 193, 196, 137, 251, 193, 198, 137, 251, 193, 136, 138, 251, 193, 138, 138, 251, 193, 140, 138, 251, 193, 142, 138, 251, 193, 144, 138, 251, 193, 146, 138, 251, 193, 148, 138, 251, 193, 150, 138, 251, 193, 152, 138, 251, 193, 154, 138, 251, 193, 162, 139, 251, 206, 96, 18, 0, 206, 96, 18, 0, 206, 96, 18, 0, 206, 96, 18, 0, 193, 46, 136, 251, 193, 48, 136, 251, 193, 50, 136, 251, 193, 52, 136, 251, 193, 54, 136, 251, 193, 56, 136, 251, 193, 58, 136, 251, 193, 60, 136, 251, 193, 62, 136, 251, 193, 212, 137, 251, 193, 64, 136, 251, 193, 66, 136, 251, 193, 68, 136, 251, 193, 70, 136, 251, 193, 72, 136, 251, 193, 74, 136, 251, 193, 76, 136, 251, 193, 78, 136, 251, 193, 80, 136, 251, 193, 82, 136, 251, 193, 84, 136, 251, 193, 86, 136, 251, 193, 88, 136, 251, 193, 90, 136, 251, 193, 92, 136, 251, 193, 236, 136, 251, 193, 238, 136, 251, 193, 240, 136, 251, 193, 240, 135, 251, 193, 242, 135, 251, 193, 244, 135, 251, 193, 246, 135, 251, 193, 248, 135, 251, 193, 250, 135, 251, 193, 252, 135, 251, 193, 254, 135, 251, 193, 2, 136, 251, 193, 4, 136, 251, 193, 6, 136, 251, 193, 8, 136, 251, 193, 10, 136, 251, 193, 12, 136, 251, 193, 14, 136, 251, 193, 16, 136, 251, 193, 18, 136, 251, 193, 20, 136, 251, 193, 22, 136, 251, 193, 110, 136, 251, 193, 112, 136, 251, 193, 114, 136, 251, 193, 116, 136, 251, 193, 118, 136, 251, 193, 120, 136, 251, 193, 122, 136, 251, 193, 124, 136, 251, 193, 126, 136, 251, 193, 128, 136, 251, 193, 130, 136, 251, 193, 132, 136, 251, 193, 134, 136, 251, 193, 136, 136, 251, 193, 138, 136, 251, 193, 140, 136, 251, 193, 142, 136, 251, 193, 144, 136, 251, 193, 146, 136, 251, 193, 148, 136, 251, 193, 150, 136, 251, 193, 152, 136, 251, 193, 154, 136, 251, 193, 156, 136, 251, 193, 158, 136, 251, 193, 160, 136, 251, 193, 162, 136, 251, 193, 164, 136, 251, 193, 166, 136, 251, 193, 168, 136, 251, 193, 170, 136, 251, 193, 172, 136, 251, 193, 174, 136, 251, 193, 176, 136, 251, 193, 178, 136, 251, 193, 180, 136, 251, 193, 182, 136, 251, 193, 184, 136, 251, 193, 186, 136, 251, 193, 188, 136, 251, 193, 190, 136, 251, 193, 192, 136, 251, 193, 194, 136, 251, 193, 196, 136, 251, 193, 198, 136, 251, 193, 200, 136, 251, 193, 202, 136, 251, 193, 204, 136, 251, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 193, 242, 136, 251, 193, 244, 136, 251, 193, 246, 136, 251, 193, 248, 136, 251, 193, 250, 136, 251, 193, 252, 136, 251, 193, 254, 136, 251, 193, 2, 137, 251, 193, 4, 137, 251, 193, 6, 137, 251, 193, 8, 137, 251, 193, 10, 137, 251, 193, 12, 137, 251, 193, 14, 137, 251, 193, 16, 137, 251, 193, 18, 137, 251, 193, 20, 137, 251, 193, 22, 137, 251, 193, 24, 137, 251, 193, 26, 137, 251, 193, 28, 137, 251, 193, 30, 137, 251, 193, 32, 137, 251, 193, 34, 137, 251, 193, 36, 137, 251, 193, 38, 137, 251, 193, 40, 137, 251, 193, 42, 137, 251, 193, 44, 137, 251, 193, 46, 137, 251, 193, 48, 137, 251, 193, 50, 137, 251, 193, 52, 137, 251, 193, 54, 137, 251, 193, 56, 137, 251, 193, 58, 137, 251, 193, 60, 137, 251, 193, 62, 137, 251, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 128, 18, 0, 206, 128, 18, 0, 206, 128, 18, 0, 206, 128, 18, 0, 193, 28, 138, 251, 193, 30, 138, 251, 193, 32, 138, 251, 193, 238, 138, 251, 193, 240, 138, 251, 193, 242, 138, 251, 193, 244, 138, 251, 193, 246, 138, 251, 193, 248, 138, 251, 193, 250, 138, 251, 193, 252, 138, 251, 193, 254, 138, 251, 193, 2, 139, 251, 193, 4, 139, 251, 193, 6, 139, 251, 193, 8, 139, 251, 193, 10, 139, 251, 193, 12, 139, 251, 193, 14, 139, 251, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 206, 160, 18, 0, 193, 228, 93, 252, 193, 152, 137, 251, 193, 154, 137, 251, 193, 156, 137, 251, 193, 158, 137, 251, 193, 160, 137, 251, 193, 162, 137, 251, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 192, 18, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 224, 18, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 0, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 32, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 193, 136, 139, 251, 193, 180, 190, 252, 193, 182, 190, 252, 193, 184, 190, 252, 193, 186, 190, 252, 193, 188, 190, 252, 193, 214, 190, 252, 193, 216, 190, 252, 193, 218, 190, 252, 193, 220, 190, 252, 193, 222, 190, 252, 193, 224, 190, 252, 193, 226, 190, 252, 193, 228, 190, 252, 193, 230, 190, 252, 193, 190, 190, 252, 193, 232, 190, 252, 193, 234, 190, 252, 193, 236, 190, 252, 193, 238, 190, 252, 193, 240, 190, 252, 193, 242, 190, 252, 193, 244, 190, 252, 193, 246, 190, 252, 206, 160, 72, 0, 206, 160, 72, 0, 206, 160, 72, 0, 206, 160, 72, 0, 193, 160, 139, 251, 193, 248, 140, 251, 193, 250, 140, 251, 193, 252, 140, 251, 193, 254, 140, 251, 193, 2, 141, 251, 193, 26, 142, 251, 193, 28, 142, 251, 193, 30, 142, 251, 193, 32, 142, 251, 193, 242, 142, 251, 193, 244, 142, 251, 193, 246, 142, 251, 193, 248, 142, 251, 193, 250, 142, 251, 193, 252, 142, 251, 193, 254, 142, 251, 193, 208, 143, 251, 193, 210, 143, 251, 193, 212, 143, 251, 193, 214, 143, 251, 193, 176, 144, 251, 193, 178, 144, 251, 193, 72, 145, 251, 193, 74, 145, 251, 193, 76, 145, 251, 193, 78, 145, 251, 193, 80, 145, 251, 193, 82, 145, 251, 193, 204, 145, 251, 193, 18, 146, 251, 193, 102, 146, 251, 193, 164, 139, 251, 193, 166, 139, 251, 193, 168, 139, 251, 193, 170, 139, 251, 193, 172, 139, 251, 193, 174, 139, 251, 193, 176, 139, 251, 193, 178, 139, 251, 193, 180, 139, 251, 193, 4, 141, 251, 193, 6, 141, 251, 193, 8, 141, 251, 193, 10, 141, 251, 193, 12, 141, 251, 193, 14, 141, 251, 193, 16, 141, 251, 193, 18, 141, 251, 193, 20, 141, 251, 193, 34, 142, 251, 193, 36, 142, 251, 193, 38, 142, 251, 193, 40, 142, 251, 193, 42, 142, 251, 193, 44, 142, 251, 193, 46, 142, 251, 193, 48, 142, 251, 193, 50, 142, 251, 193, 52, 142, 251, 193, 54, 142, 251, 193, 56, 142, 251, 193, 58, 142, 251, 193, 60, 142, 251, 193, 184, 139, 251, 193, 186, 139, 251, 193, 188, 139, 251, 193, 190, 139, 251, 193, 192, 139, 251, 193, 22, 141, 251, 193, 24, 141, 251, 193, 26, 141, 251, 193, 28, 141, 251, 193, 62, 142, 251, 193, 64, 142, 251, 193, 66, 142, 251, 193, 68, 142, 251, 193, 70, 142, 251, 193, 16, 143, 251, 193, 18, 143, 251, 193, 20, 143, 251, 193, 242, 143, 251, 193, 244, 143, 251, 193, 246, 143, 251, 193, 248, 143, 251, 193, 202, 144, 251, 193, 204, 144, 251, 193, 206, 144, 251, 193, 208, 144, 251, 193, 88, 145, 251, 193, 90, 145, 251, 193, 214, 145, 251, 193, 34, 146, 251, 193, 36, 146, 251, 193, 38, 146, 251, 193, 124, 146, 251, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 64, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 206, 96, 19, 0, 193, 120, 141, 251, 193, 122, 141, 251, 193, 124, 141, 251, 193, 126, 141, 251, 193, 128, 141, 251, 193, 130, 141, 251, 193, 132, 141, 251, 193, 134, 141, 251, 193, 136, 141, 251, 193, 138, 141, 251, 193, 124, 142, 251, 193, 126, 142, 251, 193, 128, 142, 251, 193, 130, 142, 251, 193, 132, 142, 251, 193, 134, 142, 251, 193, 136, 142, 251, 193, 138, 142, 251, 193, 140, 142, 251, 206, 32, 20, 0, 206, 32, 20, 0, 206, 32, 20, 0, 206, 32, 20, 0, 206, 32, 20, 0, 206, 32, 20, 0, 206, 32, 20, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 128, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 193, 244, 140, 251, 193, 246, 140, 251, 193, 14, 142, 251, 193, 16, 142, 251, 193, 18, 142, 251, 193, 20, 142, 251, 193, 22, 142, 251, 193, 24, 142, 251, 193, 234, 142, 251, 193, 236, 142, 251, 193, 238, 142, 251, 193, 240, 142, 251, 193, 194, 143, 251, 193, 196, 143, 251, 193, 198, 143, 251, 193, 200, 143, 251, 193, 202, 143, 251, 193, 204, 143, 251, 193, 206, 143, 251, 193, 168, 144, 251, 193, 170, 144, 251, 193, 172, 144, 251, 193, 174, 144, 251, 193, 64, 145, 251, 193, 66, 145, 251, 193, 68, 145, 251, 193, 70, 145, 251, 193, 198, 145, 251, 193, 200, 145, 251, 193, 202, 145, 251, 193, 100, 146, 251, 193, 210, 146, 251, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 206, 160, 19, 0, 193, 72, 142, 251, 193, 80, 141, 251, 193, 82, 141, 251, 193, 84, 141, 251, 193, 86, 141, 251, 193, 88, 141, 251, 193, 90, 141, 251, 193, 92, 141, 251, 193, 94, 141, 251, 193, 96, 141, 251, 193, 98, 141, 251, 193, 100, 141, 251, 193, 102, 141, 251, 193, 104, 141, 251, 193, 106, 141, 251, 193, 108, 141, 251, 193, 110, 141, 251, 193, 112, 141, 251, 193, 114, 141, 251, 193, 116, 141, 251, 193, 74, 142, 251, 193, 76, 142, 251, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 192, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 193, 78, 142, 251, 193, 80, 142, 251, 193, 82, 142, 251, 193, 84, 142, 251, 193, 86, 142, 251, 193, 88, 142, 251, 193, 90, 142, 251, 193, 92, 142, 251, 193, 94, 142, 251, 193, 96, 142, 251, 193, 98, 142, 251, 193, 100, 142, 251, 193, 102, 142, 251, 193, 104, 142, 251, 193, 106, 142, 251, 193, 108, 142, 251, 193, 110, 142, 251, 193, 112, 142, 251, 193, 114, 142, 251, 193, 116, 142, 251, 193, 118, 142, 251, 193, 120, 142, 251, 193, 122, 142, 251, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 224, 19, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 193, 2, 143, 251, 193, 4, 143, 251, 193, 6, 143, 251, 193, 8, 143, 251, 193, 10, 143, 251, 193, 12, 143, 251, 193, 14, 143, 251, 193, 216, 143, 251, 193, 218, 143, 251, 193, 220, 143, 251, 193, 222, 143, 251, 193, 224, 143, 251, 193, 226, 143, 251, 193, 228, 143, 251, 193, 230, 143, 251, 193, 232, 143, 251, 193, 234, 143, 251, 193, 236, 143, 251, 193, 238, 143, 251, 193, 240, 143, 251, 193, 180, 144, 251, 193, 182, 144, 251, 193, 184, 144, 251, 193, 186, 144, 251, 193, 188, 144, 251, 193, 190, 144, 251, 193, 192, 144, 251, 193, 194, 144, 251, 193, 196, 144, 251, 193, 198, 144, 251, 193, 200, 144, 251, 193, 84, 145, 251, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 0, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 32, 20, 0, 206, 32, 20, 0, 206, 32, 20, 0, 206, 32, 20, 0, 206, 32, 20, 0, 193, 42, 144, 251, 193, 44, 144, 251, 193, 46, 144, 251, 193, 48, 144, 251, 193, 50, 144, 251, 193, 52, 144, 251, 193, 54, 144, 251, 193, 56, 144, 251, 193, 58, 144, 251, 193, 60, 144, 251, 193, 62, 144, 251, 193, 64, 144, 251, 193, 244, 144, 251, 193, 246, 144, 251, 193, 248, 144, 251, 193, 250, 144, 251, 193, 252, 144, 251, 193, 254, 144, 251, 193, 2, 145, 251, 193, 108, 145, 251, 193, 110, 145, 251, 193, 112, 145, 251, 193, 114, 145, 251, 193, 116, 145, 251, 193, 222, 145, 251, 193, 224, 145, 251, 193, 226, 145, 251, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 64, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 206, 96, 20, 0, 193, 210, 144, 251, 193, 212, 144, 251, 193, 214, 144, 251, 193, 216, 144, 251, 193, 218, 144, 251, 193, 220, 144, 251, 193, 222, 144, 251, 193, 224, 144, 251, 193, 226, 144, 251, 193, 228, 144, 251, 193, 230, 144, 251, 193, 232, 144, 251, 193, 234, 144, 251, 193, 236, 144, 251, 193, 238, 144, 251, 193, 240, 144, 251, 193, 242, 144, 251, 193, 92, 145, 251, 193, 94, 145, 251, 193, 96, 145, 251, 193, 98, 145, 251, 193, 40, 146, 251, 193, 100, 145, 251, 193, 102, 145, 251, 193, 104, 145, 251, 193, 106, 145, 251, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 206, 128, 20, 0, 193, 4, 145, 251, 193, 6, 145, 251, 193, 8, 145, 251, 193, 10, 145, 251, 193, 12, 145, 251, 193, 14, 145, 251, 193, 16, 145, 251, 193, 18, 145, 251, 193, 20, 145, 251, 193, 22, 145, 251, 193, 24, 145, 251, 193, 26, 145, 251, 193, 118, 145, 251, 193, 28, 145, 251, 193, 30, 145, 251, 193, 32, 145, 251, 193, 34, 145, 251, 193, 36, 145, 251, 193, 38, 145, 251, 193, 40, 145, 251, 193, 42, 145, 251, 193, 44, 145, 251, 193, 46, 145, 251, 193, 48, 145, 251, 193, 50, 145, 251, 193, 52, 145, 251, 193, 54, 145, 251, 193, 56, 145, 251, 193, 58, 145, 251, 193, 60, 145, 251, 193, 62, 145, 251, 193, 120, 145, 251, 193, 122, 145, 251, 193, 124, 145, 251, 193, 126, 145, 251, 193, 86, 145, 251, 193, 206, 145, 251, 193, 208, 145, 251, 193, 210, 145, 251, 193, 212, 145, 251, 193, 20, 146, 251, 193, 22, 146, 251, 193, 24, 146, 251, 193, 26, 146, 251, 193, 28, 146, 251, 193, 30, 146, 251, 193, 32, 146, 251, 193, 112, 146, 251, 193, 114, 146, 251, 193, 116, 146, 251, 193, 118, 146, 251, 193, 120, 146, 251, 193, 122, 146, 251, 193, 172, 146, 251, 193, 174, 146, 251, 193, 176, 146, 251, 193, 214, 146, 251, 193, 216, 146, 251, 193, 4, 147, 251, 193, 18, 147, 251, 193, 40, 147, 251, 193, 80, 147, 251, 193, 98, 147, 251, 193, 100, 147, 251, 193, 112, 147, 251, 193, 114, 147, 251, 193, 128, 147, 251, 193, 128, 145, 251, 193, 130, 145, 251, 193, 132, 145, 251, 193, 134, 145, 251, 193, 136, 145, 251, 193, 138, 145, 251, 193, 140, 145, 251, 193, 142, 145, 251, 193, 144, 145, 251, 193, 146, 145, 251, 193, 148, 145, 251, 193, 150, 145, 251, 193, 152, 145, 251, 193, 154, 145, 251, 193, 156, 145, 251, 193, 158, 145, 251, 193, 160, 145, 251, 193, 162, 145, 251, 193, 164, 145, 251, 193, 166, 145, 251, 193, 168, 145, 251, 193, 170, 145, 251, 193, 172, 145, 251, 193, 174, 145, 251, 193, 176, 145, 251, 193, 178, 145, 251, 193, 180, 145, 251, 193, 182, 145, 251, 193, 184, 145, 251, 193, 186, 145, 251, 193, 188, 145, 251, 193, 190, 145, 251, 193, 16, 146, 251, 193, 72, 146, 251, 193, 74, 146, 251, 193, 76, 146, 251, 193, 78, 146, 251, 193, 80, 146, 251, 193, 82, 146, 251, 193, 84, 146, 251, 193, 86, 146, 251, 193, 88, 146, 251, 193, 90, 146, 251, 193, 92, 146, 251, 193, 94, 146, 251, 193, 96, 146, 251, 193, 146, 146, 251, 193, 148, 146, 251, 193, 150, 146, 251, 193, 152, 146, 251, 193, 154, 146, 251, 193, 156, 146, 251, 193, 158, 146, 251, 193, 160, 146, 251, 193, 162, 146, 251, 193, 164, 146, 251, 193, 166, 146, 251, 193, 168, 146, 251, 193, 170, 146, 251, 193, 196, 145, 251, 206, 160, 20, 0, 206, 160, 20, 0, 206, 160, 20, 0, 206, 160, 20, 0, 193, 216, 145, 251, 193, 218, 145, 251, 193, 220, 145, 251, 193, 42, 146, 251, 193, 44, 146, 251, 193, 46, 146, 251, 193, 130, 146, 251, 193, 132, 146, 251, 193, 134, 146, 251, 193, 136, 146, 251, 193, 138, 146, 251, 193, 140, 146, 251, 193, 142, 146, 251, 193, 178, 146, 251, 193, 218, 146, 251, 193, 180, 146, 251, 193, 182, 146, 251, 193, 220, 146, 251, 193, 222, 146, 251, 193, 224, 146, 251, 193, 226, 146, 251, 193, 228, 146, 251, 193, 6, 147, 251, 193, 230, 146, 251, 193, 8, 147, 251, 193, 10, 147, 251, 193, 12, 147, 251, 193, 66, 147, 251, 193, 68, 147, 251, 193, 70, 147, 251, 193, 82, 147, 251, 193, 84, 147, 251, 193, 48, 146, 251, 193, 50, 146, 251, 193, 52, 146, 251, 193, 54, 146, 251, 193, 144, 146, 251, 193, 184, 146, 251, 193, 186, 146, 251, 193, 232, 146, 251, 193, 60, 147, 251, 193, 86, 147, 251, 193, 214, 147, 251, 193, 44, 148, 251, 193, 46, 148, 251, 193, 62, 148, 251, 193, 86, 148, 251, 193, 142, 148, 251, 193, 160, 148, 251, 193, 204, 148, 251, 193, 36, 149, 251, 193, 38, 149, 251, 193, 60, 149, 251, 193, 62, 149, 251, 193, 64, 149, 251, 193, 120, 149, 251, 193, 122, 149, 251, 193, 124, 149, 251, 193, 126, 149, 251, 193, 128, 149, 251, 193, 130, 149, 251, 193, 216, 149, 251, 193, 218, 149, 251, 193, 220, 149, 251, 193, 98, 11, 252, 193, 118, 11, 252, 193, 120, 11, 252, 193, 122, 11, 252, 193, 124, 11, 252, 193, 126, 11, 252, 193, 142, 11, 252, 193, 144, 11, 252, 193, 146, 11, 252, 193, 148, 11, 252, 193, 150, 11, 252, 193, 152, 11, 252, 193, 158, 11, 252, 193, 98, 146, 251, 193, 166, 11, 252, 193, 178, 11, 252, 193, 180, 11, 252, 193, 182, 11, 252, 193, 184, 11, 252, 193, 186, 11, 252, 193, 188, 11, 252, 193, 194, 11, 252, 193, 196, 11, 252, 193, 200, 11, 252, 193, 206, 11, 252, 193, 208, 11, 252, 193, 210, 11, 252, 193, 212, 11, 252, 193, 214, 11, 252, 193, 216, 11, 252, 193, 218, 11, 252, 193, 228, 11, 252, 193, 104, 146, 251, 193, 106, 146, 251, 193, 108, 146, 251, 193, 110, 146, 251, 193, 212, 146, 251, 193, 38, 147, 251, 193, 48, 147, 251, 193, 76, 147, 251, 193, 154, 147, 251, 193, 168, 147, 251, 193, 198, 147, 251, 193, 14, 148, 251, 193, 24, 148, 251, 193, 36, 148, 251, 193, 56, 148, 251, 193, 78, 148, 251, 193, 84, 148, 251, 193, 96, 148, 251, 193, 104, 148, 251, 193, 106, 148, 251, 193, 114, 148, 251, 193, 150, 148, 251, 193, 200, 148, 251, 193, 216, 148, 251, 193, 236, 148, 251, 193, 238, 148, 251, 193, 6, 149, 251, 193, 12, 149, 251, 193, 14, 149, 251, 193, 48, 149, 251, 193, 50, 149, 251, 193, 176, 149, 251, 193, 126, 146, 251, 193, 128, 146, 251, 193, 50, 147, 251, 193, 144, 147, 251, 193, 146, 147, 251, 193, 174, 147, 251, 193, 180, 147, 251, 193, 4, 148, 251, 193, 16, 148, 251, 193, 18, 148, 251, 193, 20, 148, 251, 193, 28, 148, 251, 193, 58, 148, 251, 193, 108, 148, 251, 193, 120, 148, 251, 193, 202, 148, 251, 193, 218, 148, 251, 193, 52, 149, 251, 193, 102, 149, 251, 193, 188, 149, 251, 193, 190, 149, 251, 193, 28, 150, 251, 193, 30, 150, 251, 193, 142, 150, 251, 193, 144, 150, 251, 193, 238, 150, 251, 193, 226, 151, 251, 193, 228, 151, 251, 193, 230, 151, 251, 193, 60, 152, 251, 193, 140, 152, 251, 193, 224, 152, 251, 206, 160, 20, 0, 206, 160, 20, 0, 206, 160, 20, 0, 206, 160, 20, 0, 206, 160, 20, 0, 206, 160, 20, 0, 206, 160, 20, 0, 193, 234, 146, 251, 193, 236, 146, 251, 193, 238, 146, 251, 193, 240, 146, 251, 193, 242, 146, 251, 193, 244, 146, 251, 193, 246, 146, 251, 193, 248, 146, 251, 193, 250, 146, 251, 193, 252, 146, 251, 193, 254, 146, 251, 193, 14, 147, 251, 193, 16, 147, 251, 193, 20, 147, 251, 193, 22, 147, 251, 193, 24, 147, 251, 193, 26, 147, 251, 193, 28, 147, 251, 193, 30, 147, 251, 193, 32, 147, 251, 193, 34, 147, 251, 193, 36, 147, 251, 193, 42, 147, 251, 193, 44, 147, 251, 193, 46, 147, 251, 193, 2, 147, 251, 193, 110, 147, 251, 193, 140, 147, 251, 193, 250, 147, 251, 193, 54, 148, 251, 193, 68, 148, 251, 193, 76, 148, 251, 193, 82, 148, 251, 193, 112, 148, 251, 193, 118, 148, 251, 193, 130, 148, 251, 193, 166, 148, 251, 193, 154, 148, 251, 193, 232, 148, 251, 193, 234, 148, 251, 193, 246, 148, 251, 193, 2, 149, 251, 193, 4, 149, 251, 193, 46, 149, 251, 193, 226, 150, 251, 193, 80, 151, 251, 193, 82, 151, 251, 193, 84, 151, 251, 193, 86, 151, 251, 193, 200, 151, 251, 193, 202, 151, 251, 193, 204, 151, 251, 193, 132, 152, 251, 193, 220, 152, 251, 193, 22, 153, 251, 193, 66, 153, 251, 193, 112, 153, 251, 193, 52, 147, 251, 193, 54, 147, 251, 193, 62, 147, 251, 193, 56, 147, 251, 193, 58, 147, 251, 193, 64, 147, 251, 193, 72, 147, 251, 193, 78, 147, 251, 193, 74, 147, 251, 193, 88, 147, 251, 193, 90, 147, 251, 193, 92, 147, 251, 193, 94, 147, 251, 193, 96, 147, 251, 193, 104, 147, 251, 193, 106, 147, 251, 193, 108, 147, 251, 193, 116, 147, 251, 193, 118, 147, 251, 193, 120, 147, 251, 193, 122, 147, 251, 193, 124, 147, 251, 193, 126, 147, 251, 193, 130, 147, 251, 193, 132, 147, 251, 193, 134, 147, 251, 193, 136, 147, 251, 193, 138, 147, 251, 193, 150, 147, 251, 193, 152, 147, 251, 193, 162, 147, 251, 193, 164, 147, 251, 193, 102, 147, 251, 193, 212, 78, 253, 193, 158, 147, 251, 193, 160, 147, 251, 193, 176, 147, 251, 193, 210, 147, 251, 193, 228, 147, 251, 193, 230, 147, 251, 193, 232, 147, 251, 193, 242, 147, 251, 193, 244, 147, 251, 193, 6, 148, 251, 193, 30, 148, 251, 193, 32, 148, 251, 193, 42, 148, 251, 193, 60, 148, 251, 193, 70, 148, 251, 193, 132, 148, 251, 193, 134, 148, 251, 193, 136, 148, 251, 193, 138, 148, 251, 193, 152, 148, 251, 193, 158, 148, 251, 193, 172, 148, 251, 193, 174, 148, 251, 193, 176, 148, 251, 193, 178, 148, 251, 193, 180, 148, 251, 193, 190, 148, 251, 193, 220, 148, 251, 193, 20, 149, 251, 193, 22, 149, 251, 193, 142, 147, 251, 193, 156, 147, 251, 193, 226, 147, 251, 193, 252, 147, 251, 193, 254, 147, 251, 193, 2, 148, 251, 193, 26, 148, 251, 193, 38, 148, 251, 193, 40, 148, 251, 193, 80, 148, 251, 193, 156, 148, 251, 193, 168, 148, 251, 193, 170, 148, 251, 193, 188, 148, 251, 193, 240, 148, 251, 193, 242, 148, 251, 193, 16, 149, 251, 193, 98, 149, 251, 193, 100, 149, 251, 193, 184, 149, 251, 193, 186, 149, 251, 193, 16, 150, 251, 193, 18, 150, 251, 193, 20, 150, 251, 193, 22, 150, 251, 193, 24, 150, 251, 193, 26, 150, 251, 193, 134, 150, 251, 193, 136, 150, 251, 193, 138, 150, 251, 193, 140, 150, 251, 193, 228, 150, 251, 193, 166, 147, 251, 193, 170, 147, 251, 193, 172, 147, 251, 193, 178, 147, 251, 193, 182, 147, 251, 193, 184, 147, 251, 193, 186, 147, 251, 193, 188, 147, 251, 193, 190, 147, 251, 193, 192, 147, 251, 193, 194, 147, 251, 193, 196, 147, 251, 193, 200, 147, 251, 193, 202, 147, 251, 193, 204, 147, 251, 193, 206, 147, 251, 193, 208, 147, 251, 193, 216, 147, 251, 193, 218, 147, 251, 193, 220, 147, 251, 193, 222, 147, 251, 193, 224, 147, 251, 193, 234, 147, 251, 193, 236, 147, 251, 193, 238, 147, 251, 193, 240, 147, 251, 193, 246, 147, 251, 193, 248, 147, 251, 193, 8, 148, 251, 193, 10, 148, 251, 193, 12, 148, 251, 193, 22, 148, 251, 193, 34, 148, 251, 193, 48, 148, 251, 193, 50, 148, 251, 193, 52, 148, 251, 193, 64, 148, 251, 193, 66, 148, 251, 193, 72, 148, 251, 193, 74, 148, 251, 193, 88, 148, 251, 193, 90, 148, 251, 193, 92, 148, 251, 193, 94, 148, 251, 193, 98, 148, 251, 193, 110, 148, 251, 193, 116, 148, 251, 193, 122, 148, 251, 193, 124, 148, 251, 193, 126, 148, 251, 193, 128, 148, 251, 193, 140, 148, 251, 193, 144, 148, 251, 193, 146, 148, 251, 193, 148, 148, 251, 193, 162, 148, 251, 193, 164, 148, 251, 193, 182, 148, 251, 193, 184, 148, 251, 193, 186, 148, 251, 193, 192, 148, 251, 193, 194, 148, 251, 193, 196, 148, 251, 193, 198, 148, 251, 193, 16, 165, 251, 193, 18, 165, 251, 193, 26, 165, 251, 193, 28, 165, 251, 193, 30, 165, 251, 193, 44, 165, 251, 193, 46, 165, 251, 193, 48, 165, 251, 193, 54, 165, 251, 193, 56, 165, 251, 193, 58, 165, 251, 193, 60, 165, 251, 193, 62, 165, 251, 193, 64, 165, 251, 193, 66, 165, 251, 193, 80, 165, 251, 193, 100, 148, 251, 193, 82, 165, 251, 193, 84, 165, 251, 193, 90, 165, 251, 193, 92, 165, 251, 193, 94, 165, 251, 193, 96, 165, 251, 193, 102, 165, 251, 193, 104, 165, 251, 193, 106, 165, 251, 193, 108, 165, 251, 193, 110, 165, 251, 193, 112, 165, 251, 193, 114, 165, 251, 193, 116, 165, 251, 193, 120, 165, 251, 206, 32, 47, 0, 206, 32, 47, 0, 193, 102, 148, 251, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 193, 206, 148, 251, 193, 208, 148, 251, 193, 210, 148, 251, 193, 212, 148, 251, 193, 214, 148, 251, 193, 222, 148, 251, 193, 224, 148, 251, 193, 226, 148, 251, 193, 228, 148, 251, 193, 230, 148, 251, 193, 244, 148, 251, 193, 248, 148, 251, 193, 250, 148, 251, 193, 252, 148, 251, 193, 8, 149, 251, 193, 254, 148, 251, 193, 10, 149, 251, 193, 18, 149, 251, 193, 40, 149, 251, 193, 42, 149, 251, 193, 44, 149, 251, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 193, 24, 149, 251, 193, 26, 149, 251, 193, 28, 149, 251, 193, 30, 149, 251, 193, 32, 149, 251, 193, 34, 149, 251, 193, 54, 149, 251, 193, 56, 149, 251, 193, 58, 149, 251, 193, 104, 149, 251, 193, 106, 149, 251, 193, 108, 149, 251, 193, 110, 149, 251, 193, 112, 149, 251, 193, 114, 149, 251, 193, 116, 149, 251, 193, 118, 149, 251, 193, 192, 149, 251, 193, 194, 149, 251, 193, 196, 149, 251, 193, 198, 149, 251, 193, 200, 149, 251, 193, 202, 149, 251, 193, 204, 149, 251, 193, 206, 149, 251, 193, 208, 149, 251, 193, 210, 149, 251, 193, 212, 149, 251, 193, 214, 149, 251, 193, 32, 150, 251, 193, 34, 150, 251, 193, 36, 150, 251, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 206, 192, 20, 0, 193, 132, 149, 251, 193, 134, 149, 251, 193, 136, 149, 251, 193, 138, 149, 251, 193, 140, 149, 251, 193, 142, 149, 251, 193, 144, 149, 251, 193, 146, 149, 251, 193, 148, 149, 251, 193, 150, 149, 251, 193, 152, 149, 251, 193, 154, 149, 251, 193, 156, 149, 251, 193, 158, 149, 251, 193, 160, 149, 251, 193, 162, 149, 251, 193, 164, 149, 251, 193, 166, 149, 251, 193, 168, 149, 251, 193, 170, 149, 251, 193, 172, 149, 251, 193, 174, 149, 251, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 193, 178, 149, 251, 193, 180, 149, 251, 193, 182, 149, 251, 193, 12, 150, 251, 193, 14, 150, 251, 193, 132, 150, 251, 193, 88, 151, 251, 193, 90, 151, 251, 193, 206, 151, 251, 193, 208, 151, 251, 193, 42, 152, 251, 193, 44, 152, 251, 193, 46, 152, 251, 193, 48, 152, 251, 193, 50, 152, 251, 193, 52, 152, 251, 193, 24, 153, 251, 193, 26, 153, 251, 193, 114, 153, 251, 193, 116, 153, 251, 193, 180, 153, 251, 193, 194, 153, 251, 193, 196, 153, 251, 193, 254, 153, 251, 193, 32, 154, 251, 193, 34, 154, 251, 193, 36, 154, 251, 193, 42, 154, 251, 193, 58, 154, 251, 193, 60, 154, 251, 193, 76, 154, 251, 193, 152, 154, 251, 193, 222, 149, 251, 193, 54, 150, 251, 193, 56, 150, 251, 193, 58, 150, 251, 193, 60, 150, 251, 193, 62, 150, 251, 193, 64, 150, 251, 193, 166, 150, 251, 193, 168, 150, 251, 193, 8, 151, 251, 193, 10, 151, 251, 193, 12, 151, 251, 193, 14, 151, 251, 193, 16, 151, 251, 193, 18, 151, 251, 193, 122, 151, 251, 193, 124, 151, 251, 193, 126, 151, 251, 193, 128, 151, 251, 193, 130, 151, 251, 193, 132, 151, 251, 193, 134, 151, 251, 193, 136, 151, 251, 193, 244, 151, 251, 193, 246, 151, 251, 193, 248, 151, 251, 193, 250, 151, 251, 193, 252, 151, 251, 193, 84, 152, 251, 193, 86, 152, 251, 193, 88, 152, 251, 193, 90, 152, 251, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 224, 20, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 193, 38, 150, 251, 193, 40, 150, 251, 193, 42, 150, 251, 193, 44, 150, 251, 193, 46, 150, 251, 193, 48, 150, 251, 193, 50, 150, 251, 193, 52, 150, 251, 193, 146, 150, 251, 193, 148, 150, 251, 193, 150, 150, 251, 193, 152, 150, 251, 193, 154, 150, 251, 193, 156, 150, 251, 193, 158, 150, 251, 193, 160, 150, 251, 193, 162, 150, 251, 193, 164, 150, 251, 193, 240, 150, 251, 193, 242, 150, 251, 193, 244, 150, 251, 193, 246, 150, 251, 193, 248, 150, 251, 193, 250, 150, 251, 193, 252, 150, 251, 193, 254, 150, 251, 193, 2, 151, 251, 193, 4, 151, 251, 193, 6, 151, 251, 193, 98, 151, 251, 193, 100, 151, 251, 193, 102, 151, 251, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 0, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 32, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 193, 230, 150, 251, 193, 232, 150, 251, 193, 234, 150, 251, 193, 236, 150, 251, 193, 92, 151, 251, 193, 94, 151, 251, 193, 96, 151, 251, 193, 210, 151, 251, 193, 212, 151, 251, 193, 214, 151, 251, 193, 216, 151, 251, 193, 218, 151, 251, 193, 220, 151, 251, 193, 222, 151, 251, 193, 224, 151, 251, 193, 54, 152, 251, 193, 56, 152, 251, 193, 58, 152, 251, 193, 134, 152, 251, 193, 136, 152, 251, 193, 138, 152, 251, 193, 222, 152, 251, 193, 28, 153, 251, 193, 30, 153, 251, 193, 32, 153, 251, 193, 82, 153, 251, 193, 84, 153, 251, 193, 86, 153, 251, 193, 118, 153, 251, 193, 156, 153, 251, 193, 182, 153, 251, 193, 214, 153, 251, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 64, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 193, 104, 151, 251, 193, 106, 151, 251, 193, 108, 151, 251, 193, 110, 151, 251, 193, 112, 151, 251, 193, 114, 151, 251, 193, 116, 151, 251, 193, 118, 151, 251, 193, 62, 152, 251, 193, 120, 151, 251, 193, 232, 151, 251, 193, 234, 151, 251, 193, 236, 151, 251, 193, 238, 151, 251, 193, 240, 151, 251, 193, 242, 151, 251, 193, 114, 209, 252, 193, 64, 152, 251, 193, 66, 152, 251, 193, 68, 152, 251, 193, 70, 152, 251, 193, 72, 152, 251, 193, 74, 152, 251, 193, 76, 152, 251, 193, 78, 152, 251, 193, 80, 152, 251, 193, 142, 152, 251, 193, 82, 152, 251, 206, 128, 21, 0, 206, 128, 21, 0, 206, 128, 21, 0, 206, 128, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 206, 96, 21, 0, 193, 254, 151, 251, 193, 2, 152, 251, 193, 4, 152, 251, 193, 6, 152, 251, 193, 8, 152, 251, 193, 10, 152, 251, 193, 12, 152, 251, 193, 14, 152, 251, 193, 16, 152, 251, 193, 18, 152, 251, 193, 20, 152, 251, 193, 22, 152, 251, 193, 24, 152, 251, 193, 26, 152, 251, 193, 28, 152, 251, 193, 30, 152, 251, 193, 32, 152, 251, 193, 34, 152, 251, 193, 36, 152, 251, 193, 38, 152, 251, 193, 40, 152, 251, 193, 92, 152, 251, 193, 94, 152, 251, 193, 96, 152, 251, 193, 98, 152, 251, 193, 100, 152, 251, 193, 102, 152, 251, 193, 104, 152, 251, 193, 106, 152, 251, 193, 108, 152, 251, 193, 110, 152, 251, 193, 112, 152, 251, 193, 114, 152, 251, 193, 116, 152, 251, 193, 118, 152, 251, 193, 120, 152, 251, 193, 122, 152, 251, 193, 124, 152, 251, 193, 126, 152, 251, 193, 128, 152, 251, 193, 130, 152, 251, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 128, 21, 0, 206, 128, 21, 0, 206, 128, 21, 0, 206, 128, 21, 0, 206, 128, 21, 0, 206, 128, 21, 0, 193, 226, 152, 251, 193, 228, 152, 251, 193, 230, 152, 251, 193, 232, 152, 251, 193, 36, 153, 251, 193, 38, 153, 251, 193, 40, 153, 251, 193, 68, 153, 251, 193, 70, 153, 251, 193, 72, 153, 251, 193, 88, 153, 251, 193, 120, 153, 251, 193, 170, 153, 251, 193, 172, 153, 251, 193, 176, 153, 251, 193, 178, 153, 251, 193, 200, 153, 251, 193, 202, 153, 251, 193, 204, 153, 251, 193, 20, 154, 251, 193, 92, 154, 251, 193, 94, 154, 251, 193, 106, 154, 251, 193, 144, 154, 251, 193, 174, 154, 251, 193, 218, 154, 251, 193, 164, 152, 251, 193, 166, 152, 251, 193, 234, 152, 251, 193, 236, 152, 251, 193, 238, 152, 251, 193, 42, 153, 251, 193, 44, 153, 251, 193, 122, 153, 251, 193, 124, 153, 251, 193, 142, 153, 251, 193, 144, 153, 251, 193, 176, 154, 251, 193, 178, 154, 251, 193, 194, 154, 251, 193, 214, 154, 251, 193, 244, 154, 251, 193, 246, 154, 251, 193, 248, 154, 251, 193, 60, 155, 251, 193, 62, 155, 251, 193, 130, 155, 251, 193, 132, 155, 251, 193, 134, 155, 251, 193, 136, 155, 251, 193, 138, 155, 251, 193, 140, 155, 251, 193, 142, 155, 251, 193, 144, 155, 251, 193, 214, 155, 251, 193, 216, 155, 251, 193, 218, 155, 251, 193, 220, 155, 251, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 206, 160, 21, 0, 193, 240, 152, 251, 193, 242, 152, 251, 193, 244, 152, 251, 193, 246, 152, 251, 193, 248, 152, 251, 193, 250, 152, 251, 193, 252, 152, 251, 193, 254, 152, 251, 193, 2, 153, 251, 193, 4, 153, 251, 193, 6, 153, 251, 193, 8, 153, 251, 193, 10, 153, 251, 193, 12, 153, 251, 193, 14, 153, 251, 193, 16, 153, 251, 193, 18, 153, 251, 193, 20, 153, 251, 193, 46, 153, 251, 193, 34, 153, 251, 193, 140, 153, 251, 193, 184, 153, 251, 193, 198, 153, 251, 193, 218, 153, 251, 193, 244, 153, 251, 193, 246, 153, 251, 193, 2, 154, 251, 193, 4, 154, 251, 193, 14, 154, 251, 193, 16, 154, 251, 193, 18, 154, 251, 193, 50, 154, 251, 193, 52, 154, 251, 193, 70, 154, 251, 193, 118, 154, 251, 193, 120, 154, 251, 193, 156, 154, 251, 193, 28, 155, 251, 193, 196, 155, 251, 193, 198, 155, 251, 193, 22, 156, 251, 193, 24, 156, 251, 193, 140, 156, 251, 193, 142, 156, 251, 193, 28, 157, 251, 193, 30, 157, 251, 193, 32, 157, 251, 193, 148, 157, 251, 193, 150, 157, 251, 193, 236, 157, 251, 193, 92, 158, 251, 193, 48, 153, 251, 193, 50, 153, 251, 193, 52, 153, 251, 193, 54, 153, 251, 193, 56, 153, 251, 193, 58, 153, 251, 193, 60, 153, 251, 193, 62, 153, 251, 193, 64, 153, 251, 193, 74, 153, 251, 193, 76, 153, 251, 193, 78, 153, 251, 193, 80, 153, 251, 193, 90, 153, 251, 193, 92, 153, 251, 193, 94, 153, 251, 193, 96, 153, 251, 193, 98, 153, 251, 193, 100, 153, 251, 193, 102, 153, 251, 193, 104, 153, 251, 193, 106, 153, 251, 193, 108, 153, 251, 193, 110, 153, 251, 193, 126, 153, 251, 193, 128, 153, 251, 193, 130, 153, 251, 193, 134, 153, 251, 193, 136, 153, 251, 193, 138, 153, 251, 193, 146, 153, 251, 193, 148, 153, 251, 193, 132, 153, 251, 193, 192, 153, 251, 193, 250, 153, 251, 193, 252, 153, 251, 193, 12, 154, 251, 193, 28, 154, 251, 193, 30, 154, 251, 193, 56, 154, 251, 193, 74, 154, 251, 193, 82, 154, 251, 193, 124, 154, 251, 193, 150, 154, 251, 193, 94, 155, 251, 193, 96, 155, 251, 193, 188, 155, 251, 193, 190, 155, 251, 193, 10, 156, 251, 193, 12, 156, 251, 193, 104, 156, 251, 193, 106, 156, 251, 193, 108, 156, 251, 193, 110, 156, 251, 193, 112, 156, 251, 193, 114, 156, 251, 193, 4, 157, 251, 193, 6, 157, 251, 193, 128, 157, 251, 193, 130, 157, 251, 193, 216, 157, 251, 193, 218, 157, 251, 193, 66, 158, 251, 193, 68, 158, 251, 193, 150, 153, 251, 193, 152, 153, 251, 193, 154, 153, 251, 193, 158, 153, 251, 193, 160, 153, 251, 193, 162, 153, 251, 193, 164, 153, 251, 193, 166, 153, 251, 193, 168, 153, 251, 193, 174, 153, 251, 193, 186, 153, 251, 193, 188, 153, 251, 193, 190, 153, 251, 193, 206, 153, 251, 193, 208, 153, 251, 193, 210, 153, 251, 193, 212, 153, 251, 193, 220, 153, 251, 193, 222, 153, 251, 193, 224, 153, 251, 193, 226, 153, 251, 193, 228, 153, 251, 193, 232, 153, 251, 193, 234, 153, 251, 193, 236, 153, 251, 193, 238, 153, 251, 193, 248, 153, 251, 193, 6, 154, 251, 193, 8, 154, 251, 193, 10, 154, 251, 193, 22, 154, 251, 193, 24, 154, 251, 193, 216, 153, 251, 193, 230, 153, 251, 193, 240, 153, 251, 193, 242, 153, 251, 193, 44, 154, 251, 193, 46, 154, 251, 193, 48, 154, 251, 193, 78, 154, 251, 193, 102, 154, 251, 193, 104, 154, 251, 193, 116, 154, 251, 193, 134, 154, 251, 193, 154, 154, 251, 193, 232, 154, 251, 193, 234, 154, 251, 193, 236, 154, 251, 193, 22, 155, 251, 193, 24, 155, 251, 193, 26, 155, 251, 193, 100, 155, 251, 193, 102, 155, 251, 193, 104, 155, 251, 193, 106, 155, 251, 193, 108, 155, 251, 193, 192, 155, 251, 193, 194, 155, 251, 193, 18, 156, 251, 193, 20, 156, 251, 193, 128, 156, 251, 193, 130, 156, 251, 193, 132, 156, 251, 193, 134, 156, 251, 193, 182, 154, 251, 193, 184, 154, 251, 193, 190, 154, 251, 193, 192, 154, 251, 193, 212, 154, 251, 193, 230, 154, 251, 193, 20, 155, 251, 193, 98, 155, 251, 193, 14, 156, 251, 193, 16, 156, 251, 193, 116, 156, 251, 193, 118, 156, 251, 193, 120, 156, 251, 193, 122, 156, 251, 193, 124, 156, 251, 193, 126, 156, 251, 193, 12, 157, 251, 193, 14, 157, 251, 193, 16, 157, 251, 193, 18, 157, 251, 193, 20, 157, 251, 193, 22, 157, 251, 193, 132, 157, 251, 193, 134, 157, 251, 193, 136, 157, 251, 193, 138, 157, 251, 193, 220, 157, 251, 193, 222, 157, 251, 193, 72, 158, 251, 193, 74, 158, 251, 193, 76, 158, 251, 193, 78, 158, 251, 193, 220, 154, 251, 193, 222, 154, 251, 193, 224, 154, 251, 193, 226, 154, 251, 193, 238, 154, 251, 193, 240, 154, 251, 193, 242, 154, 251, 193, 30, 155, 251, 193, 32, 155, 251, 193, 34, 155, 251, 193, 36, 155, 251, 193, 38, 155, 251, 193, 40, 155, 251, 193, 42, 155, 251, 193, 44, 155, 251, 193, 46, 155, 251, 193, 48, 155, 251, 193, 50, 155, 251, 193, 52, 155, 251, 193, 54, 155, 251, 193, 56, 155, 251, 193, 110, 155, 251, 193, 112, 155, 251, 193, 114, 155, 251, 193, 58, 155, 251, 193, 116, 155, 251, 193, 118, 155, 251, 193, 120, 155, 251, 193, 122, 155, 251, 193, 124, 155, 251, 193, 126, 155, 251, 193, 128, 155, 251, 193, 14, 155, 251, 193, 16, 155, 251, 193, 18, 155, 251, 193, 64, 155, 251, 193, 66, 155, 251, 193, 68, 155, 251, 193, 70, 155, 251, 193, 72, 155, 251, 193, 74, 155, 251, 193, 76, 155, 251, 193, 78, 155, 251, 193, 80, 155, 251, 193, 82, 155, 251, 193, 84, 155, 251, 193, 86, 155, 251, 193, 88, 155, 251, 193, 90, 155, 251, 193, 92, 155, 251, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 206, 192, 21, 0, 193, 228, 155, 251, 193, 230, 155, 251, 193, 232, 155, 251, 193, 234, 155, 251, 193, 236, 155, 251, 193, 238, 155, 251, 193, 240, 155, 251, 193, 242, 155, 251, 193, 244, 155, 251, 193, 246, 155, 251, 193, 248, 155, 251, 193, 250, 155, 251, 193, 252, 155, 251, 193, 254, 155, 251, 193, 2, 156, 251, 193, 4, 156, 251, 193, 6, 156, 251, 193, 8, 156, 251, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 206, 224, 21, 0, 193, 200, 155, 251, 193, 202, 155, 251, 193, 204, 155, 251, 193, 206, 155, 251, 193, 208, 155, 251, 193, 210, 155, 251, 193, 212, 155, 251, 193, 26, 156, 251, 193, 28, 156, 251, 193, 30, 156, 251, 193, 32, 156, 251, 193, 34, 156, 251, 193, 36, 156, 251, 193, 38, 156, 251, 193, 40, 156, 251, 193, 42, 156, 251, 193, 144, 156, 251, 193, 146, 156, 251, 193, 148, 156, 251, 193, 150, 156, 251, 193, 152, 156, 251, 193, 154, 156, 251, 193, 156, 156, 251, 193, 158, 156, 251, 193, 160, 156, 251, 193, 162, 156, 251, 193, 164, 156, 251, 193, 166, 156, 251, 193, 168, 156, 251, 193, 34, 157, 251, 193, 36, 157, 251, 193, 38, 157, 251, 193, 222, 155, 251, 193, 224, 155, 251, 193, 226, 155, 251, 193, 44, 156, 251, 193, 46, 156, 251, 193, 48, 156, 251, 193, 50, 156, 251, 193, 52, 156, 251, 193, 170, 156, 251, 193, 172, 156, 251, 193, 174, 156, 251, 193, 176, 156, 251, 193, 178, 156, 251, 193, 50, 157, 251, 193, 52, 157, 251, 193, 54, 157, 251, 193, 56, 157, 251, 193, 168, 157, 251, 193, 10, 158, 251, 193, 12, 158, 251, 193, 14, 158, 251, 193, 16, 158, 251, 193, 122, 158, 251, 193, 124, 158, 251, 193, 126, 158, 251, 193, 210, 158, 251, 193, 212, 158, 251, 193, 214, 158, 251, 193, 216, 158, 251, 193, 114, 159, 251, 193, 160, 159, 251, 193, 6, 160, 251, 193, 136, 156, 251, 193, 138, 156, 251, 193, 24, 157, 251, 193, 26, 157, 251, 193, 140, 157, 251, 193, 142, 157, 251, 193, 144, 157, 251, 193, 146, 157, 251, 193, 224, 157, 251, 193, 226, 157, 251, 193, 228, 157, 251, 193, 230, 157, 251, 193, 232, 157, 251, 193, 234, 157, 251, 193, 80, 158, 251, 193, 82, 158, 251, 193, 84, 158, 251, 193, 86, 158, 251, 193, 88, 158, 251, 193, 90, 158, 251, 193, 178, 158, 251, 193, 180, 158, 251, 193, 182, 158, 251, 193, 184, 158, 251, 193, 186, 158, 251, 193, 188, 158, 251, 193, 190, 158, 251, 193, 192, 158, 251, 206, 128, 22, 0, 206, 128, 22, 0, 206, 128, 22, 0, 206, 128, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 206, 0, 22, 0, 193, 180, 209, 253, 193, 250, 156, 251, 193, 252, 156, 251, 193, 254, 156, 251, 193, 2, 157, 251, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 193, 40, 157, 251, 193, 42, 157, 251, 193, 44, 157, 251, 193, 46, 157, 251, 193, 238, 157, 251, 193, 152, 157, 251, 193, 154, 157, 251, 193, 156, 157, 251, 193, 158, 157, 251, 193, 160, 157, 251, 193, 48, 157, 251, 193, 162, 157, 251, 193, 164, 157, 251, 193, 240, 157, 251, 193, 242, 157, 251, 193, 244, 157, 251, 193, 246, 157, 251, 193, 248, 157, 251, 193, 250, 157, 251, 193, 252, 157, 251, 193, 254, 157, 251, 193, 2, 158, 251, 193, 4, 158, 251, 193, 6, 158, 251, 193, 8, 158, 251, 206, 96, 22, 0, 206, 96, 22, 0, 206, 96, 22, 0, 206, 96, 22, 0, 206, 96, 22, 0, 206, 96, 22, 0, 206, 96, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 206, 32, 22, 0, 193, 170, 157, 251, 193, 124, 157, 251, 193, 126, 157, 251, 193, 172, 157, 251, 193, 168, 147, 252, 193, 12, 193, 252, 193, 186, 201, 252, 193, 26, 202, 252, 193, 228, 202, 252, 193, 138, 222, 252, 193, 238, 250, 252, 193, 176, 131, 253, 193, 160, 139, 253, 193, 254, 192, 253, 193, 220, 60, 252, 193, 214, 21, 252, 193, 208, 126, 252, 193, 86, 115, 253, 193, 166, 157, 251, 193, 242, 171, 252, 193, 98, 39, 252, 193, 20, 80, 252, 193, 154, 44, 253, 193, 78, 190, 253, 193, 178, 193, 253, 193, 72, 194, 253, 193, 74, 194, 253, 193, 244, 194, 253, 193, 110, 195, 253, 193, 112, 195, 253, 193, 102, 207, 253, 193, 36, 219, 253, 193, 38, 219, 253, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 174, 157, 251, 193, 176, 157, 251, 193, 178, 157, 251, 193, 180, 157, 251, 193, 182, 157, 251, 193, 184, 157, 251, 193, 186, 157, 251, 193, 188, 157, 251, 193, 190, 157, 251, 193, 192, 157, 251, 193, 194, 157, 251, 193, 196, 157, 251, 193, 198, 157, 251, 193, 200, 157, 251, 193, 202, 157, 251, 193, 204, 157, 251, 193, 206, 157, 251, 193, 208, 157, 251, 193, 210, 157, 251, 193, 212, 157, 251, 193, 214, 157, 251, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 206, 64, 22, 0, 193, 128, 158, 251, 193, 130, 158, 251, 193, 132, 158, 251, 193, 134, 158, 251, 193, 136, 158, 251, 193, 138, 158, 251, 193, 140, 158, 251, 193, 142, 158, 251, 193, 144, 158, 251, 193, 146, 158, 251, 193, 148, 158, 251, 193, 150, 158, 251, 193, 152, 158, 251, 193, 154, 158, 251, 193, 156, 158, 251, 193, 158, 158, 251, 193, 160, 158, 251, 193, 162, 158, 251, 193, 164, 158, 251, 193, 70, 158, 251, 193, 170, 158, 251, 193, 172, 158, 251, 193, 174, 158, 251, 193, 248, 158, 251, 193, 250, 158, 251, 193, 48, 159, 251, 193, 50, 159, 251, 193, 52, 159, 251, 193, 94, 159, 251, 193, 144, 159, 251, 193, 192, 159, 251, 193, 14, 160, 251, 193, 42, 160, 251, 193, 44, 160, 251, 193, 70, 160, 251, 193, 76, 160, 251, 193, 88, 160, 251, 193, 96, 160, 251, 193, 152, 160, 251, 193, 170, 160, 251, 193, 226, 160, 251, 193, 18, 161, 251, 193, 30, 161, 251, 193, 86, 161, 251, 193, 98, 161, 251, 193, 58, 162, 251, 193, 176, 162, 251, 193, 252, 162, 251, 193, 254, 162, 251, 193, 58, 163, 251, 193, 136, 163, 251, 193, 94, 158, 251, 193, 96, 158, 251, 193, 194, 158, 251, 193, 196, 158, 251, 193, 64, 159, 251, 193, 106, 159, 251, 193, 220, 159, 251, 193, 128, 160, 251, 193, 130, 160, 251, 193, 154, 160, 251, 193, 236, 160, 251, 193, 238, 160, 251, 193, 58, 161, 251, 193, 154, 161, 251, 193, 172, 161, 251, 193, 174, 161, 251, 193, 176, 161, 251, 193, 200, 161, 251, 193, 206, 161, 251, 193, 190, 162, 251, 193, 192, 162, 251, 193, 22, 163, 251, 193, 24, 163, 251, 193, 78, 163, 251, 193, 80, 163, 251, 193, 150, 163, 251, 193, 152, 163, 251, 193, 154, 163, 251, 193, 216, 163, 251, 193, 218, 163, 251, 193, 220, 163, 251, 193, 40, 164, 251, 206, 96, 22, 0, 206, 96, 22, 0, 206, 96, 22, 0, 206, 96, 22, 0, 206, 96, 22, 0, 193, 198, 158, 251, 193, 200, 158, 251, 193, 202, 158, 251, 193, 204, 158, 251, 193, 206, 158, 251, 193, 208, 158, 251, 193, 66, 159, 251, 193, 108, 159, 251, 193, 110, 159, 251, 193, 154, 159, 251, 193, 156, 159, 251, 193, 158, 159, 251, 193, 184, 159, 251, 193, 200, 159, 251, 193, 240, 159, 251, 193, 254, 159, 251, 193, 4, 160, 251, 193, 18, 160, 251, 193, 20, 160, 251, 193, 30, 160, 251, 193, 32, 160, 251, 193, 48, 160, 251, 193, 50, 160, 251, 193, 52, 160, 251, 193, 56, 160, 251, 193, 102, 160, 251, 193, 106, 160, 251, 193, 166, 158, 251, 193, 168, 158, 251, 193, 218, 158, 251, 193, 220, 158, 251, 193, 222, 158, 251, 193, 224, 158, 251, 193, 226, 158, 251, 193, 228, 158, 251, 193, 230, 158, 251, 193, 232, 158, 251, 193, 234, 158, 251, 193, 236, 158, 251, 193, 238, 158, 251, 193, 240, 158, 251, 193, 242, 158, 251, 193, 244, 158, 251, 193, 246, 158, 251, 193, 18, 159, 251, 193, 20, 159, 251, 193, 22, 159, 251, 193, 24, 159, 251, 193, 26, 159, 251, 193, 28, 159, 251, 193, 30, 159, 251, 193, 32, 159, 251, 193, 34, 159, 251, 193, 36, 159, 251, 193, 38, 159, 251, 193, 130, 9, 253, 193, 40, 159, 251, 193, 42, 159, 251, 193, 44, 159, 251, 193, 176, 158, 251, 193, 252, 158, 251, 193, 254, 158, 251, 193, 54, 159, 251, 193, 56, 159, 251, 193, 58, 159, 251, 193, 60, 159, 251, 193, 96, 159, 251, 193, 146, 159, 251, 193, 180, 159, 251, 193, 194, 159, 251, 193, 212, 159, 251, 193, 214, 159, 251, 193, 232, 159, 251, 193, 234, 159, 251, 193, 242, 159, 251, 193, 246, 159, 251, 193, 16, 160, 251, 193, 78, 160, 251, 193, 80, 160, 251, 193, 98, 160, 251, 193, 100, 160, 251, 193, 126, 160, 251, 193, 172, 160, 251, 193, 174, 160, 251, 193, 176, 160, 251, 193, 228, 160, 251, 193, 230, 160, 251, 193, 252, 160, 251, 193, 254, 160, 251, 193, 20, 161, 251, 193, 32, 161, 251, 206, 128, 22, 0, 206, 128, 22, 0, 206, 128, 22, 0, 206, 128, 22, 0, 193, 62, 159, 251, 193, 98, 159, 251, 193, 100, 159, 251, 193, 102, 159, 251, 193, 104, 159, 251, 193, 148, 159, 251, 193, 150, 159, 251, 193, 152, 159, 251, 193, 182, 159, 251, 193, 196, 159, 251, 193, 198, 159, 251, 193, 216, 159, 251, 193, 218, 159, 251, 193, 238, 159, 251, 193, 24, 160, 251, 193, 26, 160, 251, 193, 28, 160, 251, 193, 46, 160, 251, 193, 208, 160, 251, 193, 232, 160, 251, 193, 234, 160, 251, 193, 34, 161, 251, 193, 50, 161, 251, 193, 74, 161, 251, 193, 76, 161, 251, 193, 78, 161, 251, 193, 100, 161, 251, 193, 122, 161, 251, 193, 46, 159, 251, 193, 68, 159, 251, 193, 70, 159, 251, 193, 72, 159, 251, 193, 74, 159, 251, 193, 76, 159, 251, 193, 78, 159, 251, 193, 80, 159, 251, 193, 82, 159, 251, 193, 84, 159, 251, 193, 86, 159, 251, 193, 88, 159, 251, 193, 90, 159, 251, 193, 92, 159, 251, 193, 118, 159, 251, 193, 120, 159, 251, 193, 122, 159, 251, 193, 124, 159, 251, 193, 126, 159, 251, 193, 128, 159, 251, 193, 130, 159, 251, 193, 132, 159, 251, 193, 134, 159, 251, 193, 136, 159, 251, 193, 138, 159, 251, 193, 140, 159, 251, 193, 142, 159, 251, 206, 160, 22, 0, 206, 160, 22, 0, 206, 160, 22, 0, 206, 160, 22, 0, 206, 160, 22, 0, 193, 112, 223, 253, 193, 114, 223, 253, 193, 118, 223, 253, 193, 132, 223, 253, 193, 134, 223, 253, 193, 136, 223, 253, 193, 138, 223, 253, 193, 160, 223, 253, 193, 162, 223, 253, 193, 188, 223, 253, 193, 190, 223, 253, 193, 226, 223, 253, 193, 228, 223, 253, 193, 12, 224, 253, 193, 14, 224, 253, 193, 16, 224, 253, 193, 112, 159, 251, 193, 30, 224, 253, 193, 44, 224, 253, 193, 50, 224, 253, 193, 52, 224, 253, 193, 54, 224, 253, 193, 88, 224, 253, 193, 90, 224, 253, 193, 92, 224, 253, 193, 96, 224, 253, 193, 110, 224, 253, 193, 112, 224, 253, 193, 118, 224, 253, 193, 166, 224, 253, 193, 206, 224, 253, 193, 218, 224, 253, 206, 160, 22, 0, 206, 160, 22, 0, 206, 160, 22, 0, 206, 160, 22, 0, 193, 186, 159, 251, 193, 188, 159, 251, 193, 190, 159, 251, 193, 16, 47, 252, 193, 202, 159, 251, 193, 204, 159, 251, 193, 206, 159, 251, 193, 208, 159, 251, 193, 210, 159, 251, 193, 222, 159, 251, 193, 224, 159, 251, 193, 226, 159, 251, 193, 228, 159, 251, 193, 230, 159, 251, 193, 236, 159, 251, 193, 244, 159, 251, 193, 248, 159, 251, 193, 250, 159, 251, 193, 252, 159, 251, 193, 2, 160, 251, 193, 8, 160, 251, 193, 10, 160, 251, 193, 12, 160, 251, 193, 36, 160, 251, 193, 22, 160, 251, 193, 38, 160, 251, 193, 40, 160, 251, 193, 54, 160, 251, 193, 34, 160, 251, 193, 36, 161, 251, 193, 54, 161, 251, 193, 60, 161, 251, 193, 138, 161, 251, 193, 156, 161, 251, 193, 188, 161, 251, 193, 228, 161, 251, 193, 26, 162, 251, 193, 28, 162, 251, 193, 30, 162, 251, 193, 32, 162, 251, 193, 76, 162, 251, 193, 78, 162, 251, 193, 140, 162, 251, 193, 142, 162, 251, 193, 144, 162, 251, 193, 146, 162, 251, 193, 204, 162, 251, 193, 206, 162, 251, 193, 208, 162, 251, 193, 210, 162, 251, 193, 212, 162, 251, 193, 214, 162, 251, 193, 34, 163, 251, 193, 36, 163, 251, 193, 94, 163, 251, 193, 96, 163, 251, 193, 98, 163, 251, 193, 166, 163, 251, 193, 168, 163, 251, 193, 170, 163, 251, 193, 58, 160, 251, 193, 60, 160, 251, 193, 62, 160, 251, 193, 64, 160, 251, 193, 66, 160, 251, 193, 68, 160, 251, 193, 72, 160, 251, 193, 74, 160, 251, 193, 82, 160, 251, 193, 84, 160, 251, 193, 86, 160, 251, 193, 90, 160, 251, 193, 104, 160, 251, 193, 110, 160, 251, 193, 114, 160, 251, 193, 116, 160, 251, 193, 120, 160, 251, 193, 122, 160, 251, 193, 124, 160, 251, 193, 138, 160, 251, 193, 182, 160, 251, 193, 140, 160, 251, 193, 142, 160, 251, 193, 144, 160, 251, 193, 146, 160, 251, 193, 148, 160, 251, 193, 150, 160, 251, 193, 158, 160, 251, 193, 184, 160, 251, 193, 160, 160, 251, 193, 162, 160, 251, 193, 164, 160, 251, 193, 172, 228, 251, 193, 92, 160, 251, 193, 174, 228, 251, 193, 176, 228, 251, 193, 178, 228, 251, 193, 180, 228, 251, 193, 182, 228, 251, 193, 184, 228, 251, 193, 186, 228, 251, 193, 188, 228, 251, 193, 190, 228, 251, 193, 192, 228, 251, 193, 194, 228, 251, 193, 196, 228, 251, 193, 198, 228, 251, 193, 200, 228, 251, 193, 202, 228, 251, 193, 204, 228, 251, 193, 206, 228, 251, 193, 208, 228, 251, 193, 210, 228, 251, 193, 212, 228, 251, 193, 214, 228, 251, 193, 216, 228, 251, 193, 218, 228, 251, 193, 220, 228, 251, 193, 222, 228, 251, 206, 224, 34, 0, 206, 224, 34, 0, 206, 224, 34, 0, 206, 224, 34, 0, 206, 224, 34, 0, 193, 164, 191, 252, 193, 166, 191, 252, 193, 168, 191, 252, 193, 170, 191, 252, 193, 172, 191, 252, 193, 174, 191, 252, 193, 176, 191, 252, 193, 178, 191, 252, 193, 180, 191, 252, 193, 182, 191, 252, 193, 94, 160, 251, 193, 184, 191, 252, 193, 186, 191, 252, 193, 188, 191, 252, 193, 190, 191, 252, 193, 208, 191, 252, 193, 210, 191, 252, 193, 212, 191, 252, 193, 214, 191, 252, 193, 216, 191, 252, 193, 218, 191, 252, 193, 220, 191, 252, 193, 222, 191, 252, 193, 224, 191, 252, 193, 226, 191, 252, 193, 228, 191, 252, 193, 230, 191, 252, 193, 232, 191, 252, 193, 234, 191, 252, 193, 22, 192, 252, 193, 24, 192, 252, 193, 26, 192, 252, 193, 108, 160, 251, 193, 112, 160, 251, 193, 118, 160, 251, 193, 132, 160, 251, 193, 134, 160, 251, 193, 136, 160, 251, 193, 156, 160, 251, 193, 178, 160, 251, 193, 180, 160, 251, 193, 210, 160, 251, 193, 52, 161, 251, 193, 108, 161, 251, 193, 114, 161, 251, 193, 120, 161, 251, 193, 128, 161, 251, 193, 130, 161, 251, 193, 132, 161, 251, 193, 182, 161, 251, 193, 186, 161, 251, 193, 216, 161, 251, 193, 222, 161, 251, 193, 224, 161, 251, 193, 226, 161, 251, 193, 250, 161, 251, 193, 252, 161, 251, 193, 18, 162, 251, 193, 20, 162, 251, 193, 22, 162, 251, 193, 24, 162, 251, 193, 68, 162, 251, 193, 70, 162, 251, 193, 72, 162, 251, 193, 166, 160, 251, 193, 168, 160, 251, 193, 186, 160, 251, 193, 188, 160, 251, 193, 190, 160, 251, 193, 192, 160, 251, 193, 194, 160, 251, 193, 196, 160, 251, 193, 198, 160, 251, 193, 200, 160, 251, 193, 202, 160, 251, 193, 204, 160, 251, 193, 206, 160, 251, 193, 212, 160, 251, 193, 214, 160, 251, 193, 216, 160, 251, 193, 218, 160, 251, 193, 220, 160, 251, 193, 222, 160, 251, 193, 224, 160, 251, 193, 240, 160, 251, 193, 242, 160, 251, 193, 244, 160, 251, 193, 246, 160, 251, 193, 248, 160, 251, 193, 250, 160, 251, 193, 2, 161, 251, 193, 4, 161, 251, 193, 6, 161, 251, 193, 8, 161, 251, 193, 10, 161, 251, 193, 12, 161, 251, 193, 70, 161, 251, 193, 72, 161, 251, 193, 88, 161, 251, 193, 90, 161, 251, 193, 94, 161, 251, 193, 60, 162, 251, 193, 62, 162, 251, 193, 64, 162, 251, 193, 110, 162, 251, 193, 112, 162, 251, 193, 114, 162, 251, 193, 178, 162, 251, 193, 180, 162, 251, 193, 182, 162, 251, 193, 2, 163, 251, 193, 4, 163, 251, 193, 6, 163, 251, 193, 8, 163, 251, 193, 10, 163, 251, 193, 60, 163, 251, 193, 62, 163, 251, 193, 64, 163, 251, 193, 66, 163, 251, 193, 138, 163, 251, 193, 140, 163, 251, 193, 196, 163, 251, 193, 198, 163, 251, 193, 36, 164, 251, 193, 68, 164, 251, 193, 70, 164, 251, 193, 96, 164, 251, 193, 98, 164, 251, 193, 124, 161, 251, 193, 126, 161, 251, 193, 180, 161, 251, 193, 196, 161, 251, 193, 198, 161, 251, 193, 248, 161, 251, 193, 66, 162, 251, 193, 116, 162, 251, 193, 118, 162, 251, 193, 184, 162, 251, 193, 186, 162, 251, 193, 188, 162, 251, 193, 12, 163, 251, 193, 14, 163, 251, 193, 16, 163, 251, 193, 18, 163, 251, 193, 20, 163, 251, 193, 68, 163, 251, 193, 70, 163, 251, 193, 72, 163, 251, 193, 74, 163, 251, 193, 76, 163, 251, 193, 142, 163, 251, 193, 144, 163, 251, 193, 146, 163, 251, 193, 148, 163, 251, 193, 200, 163, 251, 193, 202, 163, 251, 193, 204, 163, 251, 193, 206, 163, 251, 193, 208, 163, 251, 193, 210, 163, 251, 193, 142, 161, 251, 193, 144, 161, 251, 193, 146, 161, 251, 193, 148, 161, 251, 193, 150, 161, 251, 193, 152, 161, 251, 193, 158, 161, 251, 193, 160, 161, 251, 193, 162, 161, 251, 193, 164, 161, 251, 193, 166, 161, 251, 193, 168, 161, 251, 193, 170, 161, 251, 193, 178, 161, 251, 193, 184, 161, 251, 193, 190, 161, 251, 193, 192, 161, 251, 193, 202, 161, 251, 193, 204, 161, 251, 193, 208, 161, 251, 193, 210, 161, 251, 193, 212, 161, 251, 193, 214, 161, 251, 193, 218, 161, 251, 193, 220, 161, 251, 193, 230, 161, 251, 193, 232, 161, 251, 193, 234, 161, 251, 193, 236, 161, 251, 193, 238, 161, 251, 193, 240, 161, 251, 193, 242, 161, 251, 206, 224, 40, 0, 206, 224, 40, 0, 206, 224, 40, 0, 206, 224, 40, 0, 206, 224, 40, 0, 206, 224, 40, 0, 206, 224, 40, 0, 193, 238, 10, 252, 193, 240, 10, 252, 193, 242, 10, 252, 193, 244, 10, 252, 193, 250, 220, 253, 193, 246, 10, 252, 193, 248, 10, 252, 193, 250, 10, 252, 193, 194, 161, 251, 193, 28, 11, 252, 193, 30, 11, 252, 193, 32, 11, 252, 193, 34, 11, 252, 193, 54, 11, 252, 193, 56, 11, 252, 193, 58, 11, 252, 193, 60, 11, 252, 193, 62, 11, 252, 193, 64, 11, 252, 193, 66, 11, 252, 193, 68, 11, 252, 193, 90, 11, 252, 193, 92, 11, 252, 193, 94, 11, 252, 193, 96, 11, 252, 193, 244, 161, 251, 193, 246, 161, 251, 193, 254, 161, 251, 193, 2, 162, 251, 193, 4, 162, 251, 193, 6, 162, 251, 193, 8, 162, 251, 193, 10, 162, 251, 193, 12, 162, 251, 193, 14, 162, 251, 193, 16, 162, 251, 193, 34, 162, 251, 193, 36, 162, 251, 193, 38, 162, 251, 193, 40, 162, 251, 193, 42, 162, 251, 193, 44, 162, 251, 193, 46, 162, 251, 193, 48, 162, 251, 193, 50, 162, 251, 193, 52, 162, 251, 193, 54, 162, 251, 193, 56, 162, 251, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 193, 74, 162, 251, 193, 120, 162, 251, 193, 122, 162, 251, 193, 124, 162, 251, 193, 126, 162, 251, 193, 128, 162, 251, 193, 130, 162, 251, 193, 132, 162, 251, 193, 134, 162, 251, 193, 136, 162, 251, 193, 138, 162, 251, 193, 194, 162, 251, 193, 196, 162, 251, 193, 198, 162, 251, 193, 200, 162, 251, 193, 202, 162, 251, 193, 26, 163, 251, 193, 28, 163, 251, 193, 30, 163, 251, 193, 32, 163, 251, 193, 82, 163, 251, 193, 84, 163, 251, 193, 86, 163, 251, 193, 88, 163, 251, 193, 90, 163, 251, 193, 92, 163, 251, 193, 156, 163, 251, 193, 158, 163, 251, 193, 160, 163, 251, 193, 162, 163, 251, 193, 164, 163, 251, 193, 222, 163, 251, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 206, 192, 22, 0, 193, 148, 162, 251, 193, 150, 162, 251, 193, 152, 162, 251, 193, 154, 162, 251, 193, 156, 162, 251, 193, 158, 162, 251, 193, 160, 162, 251, 193, 162, 162, 251, 193, 164, 162, 251, 193, 166, 162, 251, 193, 168, 162, 251, 193, 170, 162, 251, 193, 172, 162, 251, 193, 174, 162, 251, 206, 224, 22, 0, 206, 224, 22, 0, 206, 224, 22, 0, 206, 224, 22, 0, 206, 224, 22, 0, 206, 224, 22, 0, 206, 224, 22, 0, 206, 224, 22, 0, 206, 224, 22, 0, 206, 224, 22, 0, 206, 224, 22, 0, 206, 224, 22, 0, 193, 38, 163, 251, 193, 40, 163, 251, 193, 42, 163, 251, 193, 44, 163, 251, 193, 46, 163, 251, 193, 48, 163, 251, 193, 50, 163, 251, 193, 52, 163, 251, 193, 54, 163, 251, 193, 56, 163, 251, 193, 100, 163, 251, 193, 102, 163, 251, 193, 104, 163, 251, 193, 106, 163, 251, 193, 108, 163, 251, 193, 110, 163, 251, 193, 112, 163, 251, 193, 114, 163, 251, 193, 116, 163, 251, 193, 118, 163, 251, 193, 120, 163, 251, 193, 122, 163, 251, 193, 124, 163, 251, 193, 126, 163, 251, 193, 128, 163, 251, 193, 130, 163, 251, 193, 132, 163, 251, 193, 134, 163, 251, 193, 172, 163, 251, 193, 174, 163, 251, 193, 176, 163, 251, 193, 178, 163, 251, 193, 180, 163, 251, 193, 182, 163, 251, 193, 184, 163, 251, 193, 186, 163, 251, 193, 188, 163, 251, 193, 190, 163, 251, 193, 236, 163, 251, 193, 238, 163, 251, 193, 240, 163, 251, 193, 242, 163, 251, 193, 244, 163, 251, 193, 246, 163, 251, 193, 248, 163, 251, 193, 250, 163, 251, 193, 10, 164, 251, 193, 12, 164, 251, 193, 14, 164, 251, 193, 16, 164, 251, 193, 18, 164, 251, 193, 20, 164, 251, 193, 22, 164, 251, 193, 24, 164, 251, 193, 26, 164, 251, 193, 28, 164, 251, 193, 30, 164, 251, 193, 32, 164, 251, 193, 54, 164, 251, 193, 56, 164, 251, 193, 192, 163, 251, 193, 194, 163, 251, 193, 34, 164, 251, 193, 118, 164, 251, 193, 188, 164, 251, 193, 204, 164, 251, 193, 226, 164, 251, 193, 98, 165, 251, 193, 200, 165, 251, 193, 48, 166, 251, 193, 68, 166, 251, 193, 172, 166, 251, 193, 226, 166, 251, 193, 132, 167, 251, 193, 134, 167, 251, 193, 254, 167, 251, 193, 2, 168, 251, 193, 96, 168, 251, 193, 98, 168, 251, 193, 100, 168, 251, 193, 200, 168, 251, 193, 202, 168, 251, 193, 24, 169, 251, 193, 26, 169, 251, 193, 28, 169, 251, 193, 142, 169, 251, 193, 170, 169, 251, 193, 192, 169, 251, 193, 194, 169, 251, 193, 38, 170, 251, 193, 142, 170, 251, 193, 128, 171, 251, 193, 212, 163, 251, 193, 214, 163, 251, 193, 252, 163, 251, 193, 254, 163, 251, 193, 38, 164, 251, 193, 72, 164, 251, 193, 74, 164, 251, 193, 76, 164, 251, 193, 100, 164, 251, 193, 102, 164, 251, 193, 122, 164, 251, 193, 124, 164, 251, 193, 142, 164, 251, 193, 242, 164, 251, 193, 244, 164, 251, 193, 2, 165, 251, 193, 20, 165, 251, 193, 36, 165, 251, 193, 38, 165, 251, 193, 170, 165, 251, 193, 14, 166, 251, 193, 16, 166, 251, 193, 28, 166, 251, 193, 30, 166, 251, 193, 32, 166, 251, 193, 50, 166, 251, 193, 52, 166, 251, 193, 54, 166, 251, 193, 118, 166, 251, 193, 120, 166, 251, 193, 182, 166, 251, 193, 230, 166, 251, 193, 224, 163, 251, 193, 226, 163, 251, 193, 228, 163, 251, 193, 2, 164, 251, 193, 4, 164, 251, 193, 6, 164, 251, 193, 42, 164, 251, 193, 44, 164, 251, 193, 46, 164, 251, 193, 48, 164, 251, 193, 80, 164, 251, 193, 82, 164, 251, 193, 104, 164, 251, 193, 126, 164, 251, 193, 194, 164, 251, 193, 198, 164, 251, 193, 210, 164, 251, 193, 212, 164, 251, 193, 214, 164, 251, 193, 216, 164, 251, 193, 230, 164, 251, 193, 232, 164, 251, 193, 246, 164, 251, 193, 10, 165, 251, 193, 24, 165, 251, 193, 40, 165, 251, 193, 42, 165, 251, 193, 74, 165, 251, 193, 76, 165, 251, 193, 86, 165, 251, 193, 88, 165, 251, 193, 100, 165, 251, 193, 230, 163, 251, 193, 232, 163, 251, 193, 234, 163, 251, 193, 8, 164, 251, 193, 50, 164, 251, 193, 52, 164, 251, 193, 178, 164, 251, 193, 234, 164, 251, 193, 78, 165, 251, 193, 130, 165, 251, 193, 146, 165, 251, 193, 154, 165, 251, 193, 156, 165, 251, 193, 158, 165, 251, 193, 214, 165, 251, 193, 238, 165, 251, 193, 240, 165, 251, 193, 22, 166, 251, 193, 36, 166, 251, 193, 56, 166, 251, 193, 94, 166, 251, 193, 106, 166, 251, 193, 140, 166, 251, 193, 142, 166, 251, 193, 202, 166, 251, 193, 204, 166, 251, 193, 14, 167, 251, 193, 16, 167, 251, 193, 18, 167, 251, 193, 20, 167, 251, 193, 22, 167, 251, 193, 24, 167, 251, 193, 58, 164, 251, 193, 60, 164, 251, 193, 62, 164, 251, 193, 64, 164, 251, 193, 66, 164, 251, 193, 84, 164, 251, 193, 86, 164, 251, 193, 88, 164, 251, 193, 90, 164, 251, 193, 92, 164, 251, 193, 94, 164, 251, 193, 250, 171, 252, 193, 106, 164, 251, 193, 108, 164, 251, 193, 110, 164, 251, 193, 112, 164, 251, 193, 114, 164, 251, 193, 116, 164, 251, 193, 128, 164, 251, 193, 130, 164, 251, 193, 132, 164, 251, 193, 134, 164, 251, 193, 136, 164, 251, 193, 138, 164, 251, 193, 144, 164, 251, 193, 146, 164, 251, 193, 148, 164, 251, 193, 150, 164, 251, 193, 152, 164, 251, 193, 154, 164, 251, 193, 156, 164, 251, 193, 160, 164, 251, 193, 78, 164, 251, 193, 176, 164, 251, 193, 192, 164, 251, 193, 22, 165, 251, 193, 50, 165, 251, 193, 52, 165, 251, 193, 152, 165, 251, 193, 172, 165, 251, 193, 182, 165, 251, 193, 204, 165, 251, 193, 206, 165, 251, 193, 254, 165, 251, 193, 122, 166, 251, 193, 246, 166, 251, 193, 150, 167, 251, 193, 152, 167, 251, 193, 26, 168, 251, 193, 28, 168, 251, 193, 30, 168, 251, 193, 32, 168, 251, 193, 124, 168, 251, 193, 126, 168, 251, 193, 212, 168, 251, 193, 40, 169, 251, 193, 150, 169, 251, 193, 202, 169, 251, 193, 12, 170, 251, 193, 46, 170, 251, 193, 48, 170, 251, 193, 160, 170, 251, 193, 162, 170, 251, 193, 164, 170, 251, 193, 120, 164, 251, 193, 140, 164, 251, 193, 174, 164, 251, 193, 190, 164, 251, 193, 240, 164, 251, 193, 32, 165, 251, 193, 34, 165, 251, 193, 70, 165, 251, 193, 72, 165, 251, 193, 202, 165, 251, 193, 70, 166, 251, 193, 100, 166, 251, 193, 116, 166, 251, 193, 174, 166, 251, 193, 176, 166, 251, 193, 178, 166, 251, 193, 180, 166, 251, 193, 228, 166, 251, 193, 58, 167, 251, 193, 60, 167, 251, 193, 136, 167, 251, 193, 138, 167, 251, 193, 4, 168, 251, 193, 6, 168, 251, 193, 8, 168, 251, 193, 102, 168, 251, 193, 104, 168, 251, 193, 106, 168, 251, 193, 108, 168, 251, 193, 110, 168, 251, 193, 204, 168, 251, 193, 206, 168, 251, 193, 88, 93, 252, 193, 90, 93, 252, 193, 92, 93, 252, 193, 94, 93, 252, 193, 96, 93, 252, 193, 98, 93, 252, 193, 116, 93, 252, 193, 118, 93, 252, 193, 120, 93, 252, 193, 122, 93, 252, 193, 124, 93, 252, 193, 126, 93, 252, 193, 134, 93, 252, 193, 136, 93, 252, 193, 138, 93, 252, 193, 140, 93, 252, 193, 142, 93, 252, 193, 144, 93, 252, 193, 160, 93, 252, 193, 162, 93, 252, 193, 164, 93, 252, 193, 176, 93, 252, 193, 180, 93, 252, 193, 182, 93, 252, 193, 184, 93, 252, 193, 186, 93, 252, 193, 188, 93, 252, 193, 194, 93, 252, 193, 196, 93, 252, 193, 158, 164, 251, 193, 198, 93, 252, 193, 200, 93, 252, 193, 162, 164, 251, 193, 164, 164, 251, 193, 166, 164, 251, 193, 168, 164, 251, 193, 170, 164, 251, 193, 172, 164, 251, 193, 180, 164, 251, 193, 182, 164, 251, 193, 184, 164, 251, 193, 186, 164, 251, 193, 196, 164, 251, 193, 200, 164, 251, 193, 202, 164, 251, 193, 206, 164, 251, 193, 208, 164, 251, 193, 218, 164, 251, 193, 220, 164, 251, 193, 222, 164, 251, 193, 224, 164, 251, 193, 228, 164, 251, 193, 236, 164, 251, 193, 238, 164, 251, 193, 122, 94, 252, 193, 248, 164, 251, 193, 250, 164, 251, 193, 252, 164, 251, 193, 254, 164, 251, 193, 4, 165, 251, 193, 6, 165, 251, 193, 8, 165, 251, 193, 12, 165, 251, 193, 14, 165, 251, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 193, 68, 165, 251, 193, 220, 37, 252, 193, 222, 37, 252, 193, 224, 37, 252, 193, 226, 37, 252, 193, 228, 37, 252, 193, 230, 37, 252, 193, 118, 165, 251, 193, 122, 165, 251, 193, 140, 165, 251, 193, 142, 165, 251, 193, 144, 165, 251, 193, 174, 165, 251, 193, 176, 165, 251, 193, 184, 165, 251, 193, 186, 165, 251, 193, 208, 165, 251, 193, 192, 252, 252, 193, 210, 165, 251, 193, 212, 165, 251, 193, 236, 165, 251, 193, 2, 166, 251, 193, 18, 166, 251, 193, 20, 166, 251, 193, 34, 166, 251, 193, 78, 166, 251, 193, 88, 166, 251, 193, 102, 166, 251, 193, 104, 166, 251, 193, 124, 166, 251, 193, 126, 166, 251, 193, 128, 166, 251, 193, 130, 166, 251, 193, 132, 166, 251, 193, 134, 166, 251, 206, 0, 23, 0, 206, 0, 23, 0, 206, 0, 23, 0, 206, 0, 23, 0, 193, 124, 165, 251, 193, 126, 165, 251, 193, 128, 165, 251, 193, 132, 165, 251, 193, 134, 165, 251, 193, 136, 165, 251, 193, 138, 165, 251, 193, 148, 165, 251, 193, 150, 165, 251, 193, 160, 165, 251, 193, 162, 165, 251, 193, 164, 165, 251, 193, 166, 165, 251, 193, 168, 165, 251, 193, 178, 165, 251, 193, 188, 165, 251, 193, 190, 165, 251, 193, 192, 165, 251, 193, 194, 165, 251, 193, 196, 165, 251, 193, 216, 165, 251, 193, 218, 165, 251, 193, 220, 165, 251, 193, 222, 165, 251, 193, 224, 165, 251, 193, 226, 165, 251, 193, 228, 165, 251, 193, 230, 165, 251, 193, 232, 165, 251, 193, 234, 165, 251, 193, 242, 165, 251, 193, 244, 165, 251, 193, 246, 165, 251, 193, 248, 165, 251, 193, 250, 165, 251, 193, 252, 165, 251, 193, 4, 166, 251, 193, 6, 166, 251, 193, 8, 166, 251, 193, 10, 166, 251, 193, 12, 166, 251, 193, 24, 166, 251, 193, 26, 166, 251, 193, 38, 166, 251, 193, 40, 166, 251, 193, 42, 166, 251, 193, 44, 166, 251, 193, 46, 166, 251, 193, 58, 166, 251, 193, 60, 166, 251, 193, 62, 166, 251, 193, 64, 166, 251, 193, 66, 166, 251, 193, 72, 166, 251, 193, 76, 166, 251, 193, 74, 166, 251, 193, 80, 166, 251, 193, 82, 166, 251, 193, 84, 166, 251, 193, 86, 166, 251, 193, 90, 166, 251, 193, 92, 166, 251, 193, 96, 166, 251, 193, 98, 166, 251, 193, 108, 166, 251, 193, 110, 166, 251, 193, 112, 166, 251, 193, 114, 166, 251, 193, 144, 166, 251, 193, 146, 166, 251, 193, 148, 166, 251, 193, 150, 166, 251, 193, 152, 166, 251, 193, 154, 166, 251, 193, 156, 166, 251, 193, 158, 166, 251, 193, 160, 166, 251, 193, 162, 166, 251, 193, 164, 166, 251, 193, 166, 166, 251, 193, 168, 166, 251, 193, 170, 166, 251, 193, 206, 166, 251, 193, 208, 166, 251, 193, 210, 166, 251, 193, 212, 166, 251, 193, 214, 166, 251, 193, 216, 166, 251, 193, 218, 166, 251, 193, 220, 166, 251, 193, 222, 166, 251, 193, 224, 166, 251, 206, 32, 23, 0, 206, 32, 23, 0, 206, 32, 23, 0, 206, 32, 23, 0, 206, 0, 23, 0, 206, 0, 23, 0, 206, 0, 23, 0, 206, 0, 23, 0, 206, 0, 23, 0, 193, 248, 166, 251, 193, 250, 166, 251, 193, 252, 166, 251, 193, 254, 166, 251, 193, 2, 167, 251, 193, 4, 167, 251, 193, 6, 167, 251, 193, 8, 167, 251, 193, 10, 167, 251, 193, 68, 167, 251, 193, 70, 167, 251, 193, 72, 167, 251, 193, 74, 167, 251, 193, 76, 167, 251, 193, 78, 167, 251, 193, 12, 167, 251, 193, 80, 167, 251, 193, 154, 167, 251, 193, 156, 167, 251, 193, 158, 167, 251, 193, 160, 167, 251, 193, 34, 168, 251, 206, 64, 23, 0, 206, 64, 23, 0, 206, 64, 23, 0, 206, 64, 23, 0, 206, 64, 23, 0, 193, 232, 166, 251, 193, 234, 166, 251, 193, 236, 166, 251, 193, 238, 166, 251, 193, 240, 166, 251, 193, 242, 166, 251, 193, 244, 166, 251, 193, 62, 167, 251, 193, 64, 167, 251, 193, 66, 167, 251, 193, 140, 167, 251, 193, 142, 167, 251, 193, 144, 167, 251, 193, 146, 167, 251, 193, 148, 167, 251, 193, 10, 168, 251, 193, 12, 168, 251, 193, 14, 168, 251, 193, 16, 168, 251, 193, 18, 168, 251, 193, 20, 168, 251, 193, 22, 168, 251, 193, 24, 168, 251, 193, 112, 168, 251, 193, 114, 168, 251, 193, 116, 168, 251, 193, 118, 168, 251, 193, 120, 168, 251, 193, 122, 168, 251, 193, 210, 168, 251, 193, 34, 169, 251, 193, 36, 169, 251, 193, 26, 167, 251, 193, 28, 167, 251, 193, 82, 167, 251, 193, 84, 167, 251, 193, 182, 167, 251, 193, 184, 167, 251, 193, 186, 167, 251, 193, 188, 167, 251, 193, 190, 167, 251, 193, 192, 167, 251, 193, 50, 168, 251, 193, 52, 168, 251, 193, 54, 168, 251, 193, 56, 168, 251, 193, 58, 168, 251, 193, 138, 168, 251, 193, 220, 168, 251, 193, 222, 168, 251, 193, 224, 168, 251, 193, 56, 169, 251, 193, 58, 169, 251, 193, 110, 169, 251, 193, 112, 169, 251, 193, 156, 169, 251, 193, 206, 169, 251, 193, 222, 169, 251, 193, 8, 170, 251, 193, 22, 170, 251, 193, 86, 170, 251, 193, 88, 170, 251, 193, 90, 170, 251, 193, 92, 170, 251, 206, 32, 23, 0, 206, 32, 23, 0, 206, 32, 23, 0, 206, 32, 23, 0, 206, 32, 23, 0, 206, 32, 23, 0, 206, 32, 23, 0, 206, 32, 23, 0, 206, 32, 23, 0, 206, 32, 23, 0, 193, 86, 167, 251, 193, 88, 167, 251, 193, 90, 167, 251, 193, 92, 167, 251, 193, 94, 167, 251, 193, 96, 167, 251, 193, 98, 167, 251, 193, 100, 167, 251, 193, 102, 167, 251, 193, 104, 167, 251, 193, 106, 167, 251, 193, 108, 167, 251, 193, 110, 167, 251, 193, 112, 167, 251, 193, 114, 167, 251, 193, 116, 167, 251, 193, 118, 167, 251, 193, 120, 167, 251, 193, 122, 167, 251, 193, 124, 167, 251, 193, 126, 167, 251, 193, 128, 167, 251, 193, 130, 167, 251, 193, 194, 167, 251, 193, 196, 167, 251, 193, 198, 167, 251, 193, 200, 167, 251, 193, 202, 167, 251, 193, 204, 167, 251, 193, 206, 167, 251, 193, 208, 167, 251, 193, 210, 167, 251, 193, 212, 167, 251, 193, 214, 167, 251, 193, 216, 167, 251, 193, 218, 167, 251, 193, 220, 167, 251, 193, 222, 167, 251, 193, 224, 167, 251, 193, 226, 167, 251, 193, 228, 167, 251, 193, 230, 167, 251, 193, 232, 167, 251, 193, 234, 167, 251, 193, 236, 167, 251, 193, 238, 167, 251, 193, 240, 167, 251, 193, 242, 167, 251, 193, 244, 167, 251, 193, 246, 167, 251, 193, 248, 167, 251, 193, 250, 167, 251, 193, 252, 167, 251, 193, 60, 168, 251, 206, 64, 23, 0, 206, 64, 23, 0, 206, 64, 23, 0, 206, 64, 23, 0, 206, 64, 23, 0, 193, 36, 168, 251, 193, 38, 168, 251, 193, 40, 168, 251, 193, 42, 168, 251, 193, 44, 168, 251, 193, 46, 168, 251, 193, 48, 168, 251, 193, 128, 168, 251, 193, 130, 168, 251, 193, 132, 168, 251, 193, 134, 168, 251, 193, 136, 168, 251, 193, 214, 168, 251, 193, 42, 169, 251, 193, 216, 168, 251, 193, 218, 168, 251, 193, 44, 169, 251, 193, 46, 169, 251, 193, 48, 169, 251, 193, 50, 169, 251, 193, 52, 169, 251, 193, 54, 169, 251, 193, 106, 169, 251, 193, 108, 169, 251, 193, 152, 169, 251, 193, 154, 169, 251, 193, 204, 169, 251, 193, 62, 168, 251, 193, 64, 168, 251, 193, 66, 168, 251, 193, 68, 168, 251, 193, 70, 168, 251, 193, 72, 168, 251, 193, 74, 168, 251, 193, 76, 168, 251, 193, 78, 168, 251, 193, 80, 168, 251, 193, 82, 168, 251, 193, 84, 168, 251, 193, 186, 15, 253, 193, 86, 168, 251, 193, 88, 168, 251, 193, 90, 168, 251, 193, 92, 168, 251, 193, 94, 168, 251, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 96, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 206, 128, 23, 0, 193, 198, 168, 251, 193, 20, 169, 251, 193, 22, 169, 251, 193, 60, 169, 251, 193, 62, 169, 251, 193, 64, 169, 251, 193, 66, 169, 251, 193, 68, 169, 251, 193, 70, 169, 251, 193, 72, 169, 251, 193, 74, 169, 251, 193, 76, 169, 251, 193, 78, 169, 251, 193, 80, 169, 251, 193, 82, 169, 251, 193, 84, 169, 251, 193, 86, 169, 251, 193, 88, 169, 251, 206, 160, 23, 0, 206, 160, 23, 0, 206, 160, 23, 0, 206, 160, 23, 0, 206, 160, 23, 0, 206, 160, 23, 0, 206, 160, 23, 0, 193, 208, 168, 251, 193, 30, 169, 251, 193, 32, 169, 251, 193, 90, 169, 251, 193, 92, 169, 251, 193, 94, 169, 251, 193, 172, 169, 251, 193, 196, 169, 251, 193, 198, 169, 251, 193, 240, 169, 251, 193, 246, 169, 251, 193, 40, 170, 251, 193, 42, 170, 251, 193, 44, 170, 251, 193, 144, 170, 251, 193, 146, 170, 251, 193, 148, 170, 251, 193, 150, 170, 251, 193, 144, 171, 251, 193, 146, 171, 251, 193, 148, 171, 251, 193, 150, 171, 251, 193, 152, 171, 251, 193, 154, 171, 251, 193, 156, 171, 251, 193, 158, 171, 251, 193, 206, 172, 251, 193, 208, 172, 251, 193, 210, 172, 251, 193, 212, 172, 251, 193, 214, 172, 251, 193, 216, 172, 251, 193, 38, 169, 251, 193, 96, 169, 251, 193, 98, 169, 251, 193, 100, 169, 251, 193, 102, 169, 251, 193, 104, 169, 251, 193, 144, 169, 251, 193, 146, 169, 251, 193, 148, 169, 251, 193, 200, 169, 251, 193, 216, 169, 251, 193, 152, 170, 251, 193, 154, 170, 251, 193, 156, 170, 251, 193, 158, 170, 251, 193, 160, 171, 251, 193, 162, 171, 251, 193, 164, 171, 251, 193, 166, 171, 251, 193, 168, 171, 251, 193, 170, 171, 251, 193, 172, 171, 251, 193, 174, 171, 251, 193, 176, 171, 251, 193, 178, 171, 251, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 160, 23, 0, 206, 160, 23, 0, 206, 160, 23, 0, 206, 160, 23, 0, 206, 160, 23, 0, 193, 14, 16, 253, 193, 138, 169, 251, 193, 140, 169, 251, 193, 158, 169, 251, 193, 160, 169, 251, 193, 162, 169, 251, 193, 18, 16, 253, 193, 164, 169, 251, 193, 166, 169, 251, 193, 168, 169, 251, 193, 174, 169, 251, 193, 176, 169, 251, 193, 178, 169, 251, 193, 180, 169, 251, 193, 182, 169, 251, 193, 184, 169, 251, 193, 186, 169, 251, 193, 188, 169, 251, 193, 190, 169, 251, 193, 208, 169, 251, 193, 210, 169, 251, 193, 212, 169, 251, 193, 176, 215, 253, 193, 214, 169, 251, 193, 224, 169, 251, 193, 226, 169, 251, 193, 228, 169, 251, 193, 218, 169, 251, 193, 220, 169, 251, 193, 234, 169, 251, 193, 252, 169, 251, 193, 254, 169, 251, 193, 4, 170, 251, 193, 6, 170, 251, 193, 14, 170, 251, 193, 16, 170, 251, 193, 18, 170, 251, 193, 20, 170, 251, 193, 50, 170, 251, 193, 52, 170, 251, 193, 54, 170, 251, 193, 56, 170, 251, 193, 58, 170, 251, 193, 60, 170, 251, 193, 62, 170, 251, 193, 64, 170, 251, 193, 66, 170, 251, 193, 68, 170, 251, 193, 70, 170, 251, 193, 72, 170, 251, 193, 74, 170, 251, 193, 76, 170, 251, 193, 78, 170, 251, 193, 80, 170, 251, 193, 82, 170, 251, 193, 84, 170, 251, 193, 168, 170, 251, 193, 170, 170, 251, 193, 172, 170, 251, 193, 230, 169, 251, 193, 40, 16, 253, 193, 232, 169, 251, 193, 236, 169, 251, 193, 238, 169, 251, 193, 242, 169, 251, 193, 244, 169, 251, 193, 248, 169, 251, 193, 250, 169, 251, 193, 2, 170, 251, 193, 10, 170, 251, 193, 24, 170, 251, 193, 26, 170, 251, 193, 28, 170, 251, 193, 30, 170, 251, 193, 32, 170, 251, 193, 34, 170, 251, 193, 36, 170, 251, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 193, 94, 170, 251, 193, 252, 170, 251, 193, 254, 170, 251, 193, 2, 171, 251, 193, 4, 171, 251, 193, 6, 171, 251, 193, 8, 171, 251, 193, 10, 171, 251, 193, 12, 171, 251, 193, 14, 171, 251, 193, 16, 171, 251, 193, 18, 171, 251, 193, 58, 172, 251, 193, 60, 172, 251, 193, 62, 172, 251, 193, 64, 172, 251, 193, 66, 172, 251, 193, 68, 172, 251, 193, 70, 172, 251, 193, 72, 172, 251, 193, 74, 172, 251, 193, 76, 172, 251, 193, 78, 172, 251, 193, 80, 172, 251, 193, 82, 172, 251, 193, 84, 172, 251, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 192, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 193, 166, 170, 251, 193, 180, 171, 251, 193, 182, 171, 251, 193, 184, 171, 251, 193, 186, 171, 251, 193, 188, 171, 251, 193, 190, 171, 251, 193, 192, 171, 251, 193, 194, 171, 251, 193, 196, 171, 251, 193, 248, 172, 251, 193, 250, 172, 251, 193, 252, 172, 251, 193, 254, 172, 251, 193, 2, 173, 251, 193, 4, 173, 251, 193, 6, 173, 251, 193, 8, 173, 251, 193, 10, 173, 251, 193, 12, 173, 251, 193, 116, 174, 251, 193, 118, 174, 251, 193, 120, 174, 251, 193, 122, 174, 251, 193, 124, 174, 251, 193, 126, 174, 251, 193, 128, 174, 251, 193, 130, 174, 251, 206, 224, 24, 0, 206, 224, 24, 0, 206, 224, 24, 0, 206, 224, 24, 0, 193, 174, 170, 251, 193, 176, 170, 251, 193, 178, 170, 251, 193, 180, 170, 251, 193, 182, 170, 251, 193, 184, 170, 251, 193, 186, 170, 251, 193, 188, 170, 251, 193, 190, 170, 251, 193, 192, 170, 251, 193, 194, 170, 251, 193, 196, 170, 251, 193, 198, 170, 251, 193, 200, 170, 251, 193, 202, 170, 251, 193, 204, 170, 251, 193, 206, 170, 251, 193, 208, 170, 251, 193, 210, 170, 251, 193, 212, 170, 251, 193, 214, 170, 251, 193, 216, 170, 251, 193, 218, 170, 251, 193, 220, 170, 251, 193, 222, 170, 251, 193, 224, 170, 251, 193, 226, 170, 251, 193, 228, 170, 251, 193, 230, 170, 251, 193, 232, 170, 251, 193, 234, 170, 251, 193, 236, 170, 251, 193, 238, 170, 251, 193, 240, 170, 251, 193, 242, 170, 251, 193, 244, 170, 251, 193, 246, 170, 251, 193, 248, 170, 251, 193, 250, 170, 251, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 206, 224, 23, 0, 193, 86, 172, 251, 193, 130, 171, 251, 193, 132, 171, 251, 193, 134, 171, 251, 193, 136, 171, 251, 193, 138, 171, 251, 193, 140, 171, 251, 193, 192, 172, 251, 193, 194, 172, 251, 193, 196, 172, 251, 193, 198, 172, 251, 193, 200, 172, 251, 193, 202, 172, 251, 193, 204, 172, 251, 193, 48, 174, 251, 193, 50, 174, 251, 193, 52, 174, 251, 193, 54, 174, 251, 193, 56, 174, 251, 193, 58, 174, 251, 193, 60, 174, 251, 193, 62, 174, 251, 193, 166, 175, 251, 193, 168, 175, 251, 193, 170, 175, 251, 193, 172, 175, 251, 193, 174, 175, 251, 193, 176, 175, 251, 193, 178, 175, 251, 193, 180, 175, 251, 193, 182, 175, 251, 193, 184, 175, 251, 193, 156, 177, 251, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 206, 0, 24, 0, 193, 88, 172, 251, 193, 90, 172, 251, 193, 92, 172, 251, 193, 94, 172, 251, 193, 96, 172, 251, 193, 98, 172, 251, 193, 100, 172, 251, 193, 102, 172, 251, 193, 104, 172, 251, 193, 106, 172, 251, 193, 108, 172, 251, 193, 110, 172, 251, 193, 112, 172, 251, 193, 114, 172, 251, 193, 116, 172, 251, 193, 118, 172, 251, 193, 120, 172, 251, 193, 122, 172, 251, 193, 124, 172, 251, 193, 126, 172, 251, 193, 128, 172, 251, 193, 130, 172, 251, 193, 132, 172, 251, 193, 134, 172, 251, 193, 136, 172, 251, 193, 138, 172, 251, 193, 140, 172, 251, 193, 142, 172, 251, 193, 144, 172, 251, 193, 146, 172, 251, 193, 148, 172, 251, 193, 150, 172, 251, 193, 152, 172, 251, 193, 154, 172, 251, 193, 156, 172, 251, 193, 158, 172, 251, 193, 160, 172, 251, 193, 162, 172, 251, 193, 164, 172, 251, 193, 166, 172, 251, 193, 168, 172, 251, 193, 170, 172, 251, 193, 172, 172, 251, 193, 174, 172, 251, 193, 176, 172, 251, 193, 178, 172, 251, 193, 180, 172, 251, 193, 182, 172, 251, 193, 184, 172, 251, 193, 186, 172, 251, 193, 188, 172, 251, 193, 190, 172, 251, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 206, 32, 24, 0, 193, 88, 174, 251, 193, 90, 174, 251, 193, 92, 174, 251, 193, 94, 174, 251, 193, 96, 174, 251, 193, 98, 174, 251, 193, 100, 174, 251, 193, 102, 174, 251, 193, 104, 174, 251, 193, 106, 174, 251, 193, 108, 174, 251, 193, 110, 174, 251, 193, 112, 174, 251, 193, 114, 174, 251, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 206, 64, 24, 0, 193, 132, 174, 251, 193, 134, 174, 251, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 206, 96, 24, 0, 193, 232, 174, 251, 193, 234, 174, 251, 193, 236, 174, 251, 193, 238, 174, 251, 193, 240, 174, 251, 193, 242, 174, 251, 193, 244, 174, 251, 193, 246, 174, 251, 193, 248, 174, 251, 193, 250, 174, 251, 193, 252, 174, 251, 193, 254, 174, 251, 193, 178, 173, 251, 193, 2, 175, 251, 193, 4, 175, 251, 193, 6, 175, 251, 193, 146, 176, 251, 193, 148, 176, 251, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 128, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 193, 64, 174, 251, 193, 66, 174, 251, 193, 68, 174, 251, 193, 70, 174, 251, 193, 72, 174, 251, 193, 74, 174, 251, 193, 76, 174, 251, 193, 78, 174, 251, 193, 80, 174, 251, 193, 82, 174, 251, 193, 84, 174, 251, 193, 86, 174, 251, 193, 186, 175, 251, 193, 188, 175, 251, 193, 190, 175, 251, 193, 192, 175, 251, 193, 194, 175, 251, 193, 196, 175, 251, 193, 198, 175, 251, 193, 200, 175, 251, 193, 202, 175, 251, 193, 204, 175, 251, 193, 172, 177, 251, 193, 174, 177, 251, 193, 176, 177, 251, 193, 178, 177, 251, 193, 180, 177, 251, 193, 182, 177, 251, 193, 184, 177, 251, 193, 186, 177, 251, 193, 188, 177, 251, 193, 190, 177, 251, 193, 136, 174, 251, 193, 138, 174, 251, 193, 140, 174, 251, 193, 142, 174, 251, 193, 144, 174, 251, 193, 146, 174, 251, 193, 148, 174, 251, 193, 150, 174, 251, 193, 152, 174, 251, 193, 154, 174, 251, 193, 156, 174, 251, 193, 158, 174, 251, 193, 160, 174, 251, 193, 162, 174, 251, 193, 164, 174, 251, 193, 166, 174, 251, 193, 168, 174, 251, 193, 170, 174, 251, 193, 172, 174, 251, 193, 174, 174, 251, 193, 176, 174, 251, 193, 178, 174, 251, 193, 180, 174, 251, 193, 182, 174, 251, 193, 184, 174, 251, 193, 186, 174, 251, 193, 188, 174, 251, 193, 190, 174, 251, 193, 192, 174, 251, 193, 194, 174, 251, 193, 196, 174, 251, 193, 198, 174, 251, 193, 200, 174, 251, 193, 202, 174, 251, 193, 204, 174, 251, 193, 206, 174, 251, 193, 208, 174, 251, 193, 210, 174, 251, 193, 212, 174, 251, 193, 214, 174, 251, 193, 216, 174, 251, 193, 218, 174, 251, 193, 220, 174, 251, 193, 222, 174, 251, 193, 224, 174, 251, 193, 226, 174, 251, 193, 228, 174, 251, 193, 230, 174, 251, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 160, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 206, 192, 24, 0, 193, 192, 177, 251, 193, 194, 177, 251, 193, 196, 177, 251, 193, 198, 177, 251, 193, 200, 177, 251, 193, 202, 177, 251, 193, 204, 177, 251, 193, 206, 177, 251, 193, 208, 177, 251, 193, 210, 177, 251, 193, 212, 177, 251, 193, 214, 177, 251, 193, 216, 177, 251, 193, 218, 177, 251, 193, 220, 177, 251, 193, 222, 177, 251, 193, 224, 177, 251, 193, 226, 177, 251, 193, 228, 177, 251, 193, 230, 177, 251, 206, 224, 24, 0, 206, 224, 24, 0, 206, 224, 24, 0, 206, 224, 24, 0, 206, 224, 24, 0, 206, 224, 24, 0, 206, 224, 24, 0, 193, 234, 177, 251, 193, 236, 177, 251, 193, 238, 177, 251, 193, 240, 177, 251, 193, 242, 177, 251, 193, 244, 177, 251, 193, 246, 177, 251, 193, 248, 177, 251, 193, 250, 177, 251, 193, 190, 179, 251, 193, 192, 179, 251, 193, 194, 179, 251, 193, 196, 179, 251, 193, 198, 179, 251, 193, 200, 179, 251, 193, 202, 179, 251, 193, 204, 179, 251, 193, 206, 179, 251, 193, 208, 179, 251, 193, 210, 179, 251, 193, 212, 179, 251, 206, 0, 26, 0, 206, 0, 26, 0, 206, 0, 26, 0, 206, 0, 26, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 0, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 193, 150, 176, 251, 193, 152, 176, 251, 193, 154, 176, 251, 193, 156, 176, 251, 193, 158, 176, 251, 193, 160, 176, 251, 193, 162, 176, 251, 193, 164, 176, 251, 193, 166, 176, 251, 193, 168, 176, 251, 193, 170, 176, 251, 193, 172, 176, 251, 193, 174, 176, 251, 193, 176, 176, 251, 193, 178, 176, 251, 193, 180, 176, 251, 193, 182, 176, 251, 193, 184, 176, 251, 193, 186, 176, 251, 193, 188, 176, 251, 193, 190, 176, 251, 193, 192, 176, 251, 193, 194, 176, 251, 193, 196, 176, 251, 193, 198, 176, 251, 193, 200, 176, 251, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 32, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 193, 158, 177, 251, 193, 160, 177, 251, 193, 162, 177, 251, 193, 164, 177, 251, 193, 166, 177, 251, 193, 168, 177, 251, 193, 170, 177, 251, 193, 94, 179, 251, 193, 96, 179, 251, 193, 98, 179, 251, 193, 100, 179, 251, 193, 102, 179, 251, 193, 104, 179, 251, 193, 106, 179, 251, 193, 108, 179, 251, 193, 110, 179, 251, 193, 112, 179, 251, 193, 114, 179, 251, 193, 116, 179, 251, 193, 244, 180, 251, 193, 246, 180, 251, 193, 248, 180, 251, 193, 250, 180, 251, 193, 252, 180, 251, 193, 254, 180, 251, 193, 2, 181, 251, 193, 4, 181, 251, 193, 6, 181, 251, 193, 8, 181, 251, 193, 10, 181, 251, 193, 12, 181, 251, 193, 140, 182, 251, 193, 232, 177, 251, 193, 146, 179, 251, 193, 148, 179, 251, 193, 150, 179, 251, 193, 152, 179, 251, 193, 154, 179, 251, 193, 156, 179, 251, 193, 158, 179, 251, 193, 160, 179, 251, 193, 162, 179, 251, 193, 164, 179, 251, 193, 166, 179, 251, 193, 168, 179, 251, 193, 170, 179, 251, 193, 172, 179, 251, 193, 174, 179, 251, 193, 176, 179, 251, 193, 178, 179, 251, 193, 180, 179, 251, 193, 182, 179, 251, 193, 184, 179, 251, 193, 186, 179, 251, 193, 188, 179, 251, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 64, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 193, 108, 178, 251, 193, 140, 181, 251, 193, 142, 181, 251, 193, 144, 181, 251, 193, 146, 181, 251, 193, 242, 182, 251, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 206, 96, 25, 0, 193, 46, 180, 251, 193, 48, 180, 251, 193, 50, 180, 251, 193, 52, 180, 251, 193, 54, 180, 251, 193, 56, 180, 251, 193, 58, 180, 251, 193, 60, 180, 251, 193, 62, 180, 251, 193, 64, 180, 251, 193, 66, 180, 251, 193, 68, 180, 251, 193, 70, 180, 251, 193, 72, 180, 251, 193, 74, 180, 251, 193, 76, 180, 251, 206, 96, 26, 0, 206, 96, 26, 0, 206, 96, 26, 0, 206, 96, 26, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 128, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 193, 118, 179, 251, 193, 120, 179, 251, 193, 122, 179, 251, 193, 124, 179, 251, 193, 126, 179, 251, 193, 128, 179, 251, 193, 130, 179, 251, 193, 132, 179, 251, 193, 134, 179, 251, 193, 136, 179, 251, 193, 138, 179, 251, 193, 140, 179, 251, 193, 142, 179, 251, 193, 144, 179, 251, 193, 14, 181, 251, 193, 16, 181, 251, 193, 18, 181, 251, 193, 20, 181, 251, 193, 22, 181, 251, 193, 24, 181, 251, 193, 26, 181, 251, 193, 28, 181, 251, 193, 30, 181, 251, 193, 32, 181, 251, 193, 158, 182, 251, 193, 160, 182, 251, 193, 162, 182, 251, 193, 164, 182, 251, 193, 166, 182, 251, 193, 168, 182, 251, 193, 170, 182, 251, 193, 172, 182, 251, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 160, 25, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 32, 26, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 192, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 224, 25, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 0, 26, 0, 206, 0, 26, 0, 206, 0, 26, 0, 206, 0, 26, 0, 206, 0, 26, 0, 193, 230, 182, 251, 193, 232, 182, 251, 193, 234, 182, 251, 193, 236, 182, 251, 193, 238, 182, 251, 193, 240, 182, 251, 193, 110, 184, 251, 193, 112, 184, 251, 193, 114, 184, 251, 193, 116, 184, 251, 193, 118, 184, 251, 193, 120, 184, 251, 193, 122, 184, 251, 193, 124, 184, 251, 193, 126, 184, 251, 193, 118, 185, 251, 193, 120, 185, 251, 193, 122, 185, 251, 193, 44, 186, 251, 193, 46, 186, 251, 193, 48, 186, 251, 193, 50, 186, 251, 193, 214, 186, 251, 193, 216, 186, 251, 193, 218, 186, 251, 193, 220, 186, 251, 193, 84, 187, 251, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 64, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 96, 26, 0, 206, 96, 26, 0, 206, 96, 26, 0, 206, 96, 26, 0, 206, 96, 26, 0, 206, 96, 26, 0, 206, 96, 26, 0, 206, 96, 26, 0, 206, 96, 26, 0, 193, 76, 183, 251, 193, 78, 183, 251, 193, 80, 183, 251, 193, 82, 183, 251, 193, 84, 183, 251, 193, 86, 183, 251, 193, 88, 183, 251, 193, 90, 183, 251, 193, 92, 183, 251, 193, 94, 183, 251, 193, 96, 183, 251, 193, 98, 183, 251, 193, 100, 183, 251, 193, 102, 183, 251, 193, 192, 184, 251, 193, 194, 184, 251, 193, 196, 184, 251, 193, 166, 185, 251, 193, 168, 185, 251, 193, 170, 185, 251, 193, 172, 185, 251, 193, 174, 185, 251, 193, 176, 185, 251, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 128, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 32, 73, 0, 206, 32, 73, 0, 206, 32, 73, 0, 206, 32, 73, 0, 206, 32, 73, 0, 193, 232, 196, 252, 193, 194, 196, 252, 193, 234, 196, 252, 193, 236, 196, 252, 193, 238, 196, 252, 193, 240, 196, 252, 193, 242, 196, 252, 193, 244, 196, 252, 193, 246, 196, 252, 193, 248, 196, 252, 193, 250, 196, 252, 193, 252, 196, 252, 193, 138, 182, 251, 193, 254, 196, 252, 193, 2, 197, 252, 193, 4, 197, 252, 193, 6, 197, 252, 193, 8, 197, 252, 193, 10, 197, 252, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 193, 142, 182, 251, 193, 144, 182, 251, 193, 146, 182, 251, 193, 148, 182, 251, 193, 150, 182, 251, 193, 152, 182, 251, 193, 154, 182, 251, 193, 156, 182, 251, 193, 44, 184, 251, 193, 46, 184, 251, 193, 48, 184, 251, 193, 50, 184, 251, 193, 52, 184, 251, 193, 54, 184, 251, 193, 66, 185, 251, 193, 68, 185, 251, 193, 70, 185, 251, 193, 72, 185, 251, 193, 74, 185, 251, 193, 76, 185, 251, 193, 78, 185, 251, 193, 80, 185, 251, 193, 82, 185, 251, 193, 16, 186, 251, 193, 18, 186, 251, 193, 20, 186, 251, 193, 168, 186, 251, 193, 170, 186, 251, 193, 172, 186, 251, 193, 174, 186, 251, 193, 176, 186, 251, 193, 60, 187, 251, 193, 174, 182, 251, 193, 176, 182, 251, 193, 178, 182, 251, 193, 56, 184, 251, 193, 58, 184, 251, 193, 60, 184, 251, 193, 62, 184, 251, 193, 64, 184, 251, 193, 66, 184, 251, 193, 68, 184, 251, 193, 84, 185, 251, 193, 86, 185, 251, 193, 88, 185, 251, 193, 90, 185, 251, 193, 92, 185, 251, 193, 94, 185, 251, 193, 22, 186, 251, 193, 24, 186, 251, 193, 26, 186, 251, 193, 178, 186, 251, 193, 180, 186, 251, 193, 182, 186, 251, 193, 184, 186, 251, 193, 186, 186, 251, 193, 188, 186, 251, 193, 190, 186, 251, 193, 64, 187, 251, 193, 66, 187, 251, 193, 68, 187, 251, 193, 128, 187, 251, 193, 130, 187, 251, 193, 132, 187, 251, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 206, 160, 26, 0, 193, 70, 184, 251, 193, 72, 184, 251, 193, 74, 184, 251, 193, 76, 184, 251, 193, 78, 184, 251, 193, 80, 184, 251, 193, 82, 184, 251, 193, 84, 184, 251, 193, 86, 184, 251, 193, 88, 184, 251, 193, 90, 184, 251, 193, 92, 184, 251, 193, 94, 184, 251, 193, 96, 184, 251, 193, 98, 184, 251, 193, 100, 184, 251, 193, 102, 184, 251, 193, 104, 184, 251, 193, 106, 184, 251, 193, 108, 184, 251, 206, 64, 27, 0, 206, 64, 27, 0, 206, 64, 27, 0, 206, 64, 27, 0, 206, 64, 27, 0, 206, 64, 27, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 206, 192, 26, 0, 193, 128, 184, 251, 193, 130, 184, 251, 193, 132, 184, 251, 193, 134, 184, 251, 193, 136, 184, 251, 193, 138, 184, 251, 193, 140, 184, 251, 193, 124, 185, 251, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 224, 26, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 0, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 32, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 193, 118, 213, 252, 193, 120, 213, 252, 193, 122, 213, 252, 193, 124, 213, 252, 193, 8, 213, 252, 193, 126, 213, 252, 193, 128, 213, 252, 193, 130, 213, 252, 193, 132, 213, 252, 193, 134, 213, 252, 193, 136, 213, 252, 193, 138, 213, 252, 193, 140, 213, 252, 193, 142, 213, 252, 193, 250, 213, 252, 193, 144, 213, 252, 193, 146, 213, 252, 193, 148, 213, 252, 193, 150, 213, 252, 193, 152, 213, 252, 193, 154, 213, 252, 193, 156, 213, 252, 193, 30, 235, 251, 193, 64, 185, 251, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 64, 27, 0, 206, 64, 27, 0, 206, 64, 27, 0, 206, 64, 27, 0, 206, 64, 27, 0, 193, 28, 186, 251, 193, 30, 186, 251, 193, 32, 186, 251, 193, 34, 186, 251, 193, 36, 186, 251, 193, 38, 186, 251, 193, 40, 186, 251, 193, 42, 186, 251, 193, 192, 186, 251, 193, 194, 186, 251, 193, 196, 186, 251, 193, 198, 186, 251, 193, 200, 186, 251, 193, 202, 186, 251, 193, 204, 186, 251, 193, 206, 186, 251, 193, 208, 186, 251, 193, 210, 186, 251, 193, 212, 186, 251, 193, 70, 187, 251, 193, 72, 187, 251, 193, 74, 187, 251, 193, 76, 187, 251, 193, 78, 187, 251, 193, 80, 187, 251, 193, 82, 187, 251, 193, 138, 187, 251, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 206, 96, 27, 0, 193, 52, 186, 251, 193, 54, 186, 251, 193, 56, 186, 251, 193, 58, 186, 251, 193, 60, 186, 251, 193, 62, 186, 251, 193, 64, 186, 251, 193, 66, 186, 251, 193, 68, 186, 251, 193, 70, 186, 251, 193, 72, 186, 251, 193, 74, 186, 251, 193, 222, 186, 251, 193, 224, 186, 251, 193, 226, 186, 251, 193, 228, 186, 251, 193, 88, 187, 251, 193, 90, 187, 251, 193, 92, 187, 251, 193, 152, 187, 251, 193, 154, 187, 251, 193, 156, 187, 251, 193, 158, 187, 251, 193, 178, 185, 251, 193, 76, 186, 251, 193, 78, 186, 251, 193, 230, 186, 251, 193, 232, 186, 251, 193, 234, 186, 251, 193, 236, 186, 251, 193, 238, 186, 251, 193, 94, 187, 251, 193, 96, 187, 251, 193, 98, 187, 251, 193, 212, 187, 251, 193, 214, 187, 251, 193, 58, 188, 251, 193, 122, 188, 251, 193, 124, 188, 251, 193, 154, 188, 251, 193, 156, 188, 251, 193, 158, 188, 251, 193, 200, 188, 251, 193, 254, 188, 251, 193, 2, 189, 251, 193, 4, 189, 251, 193, 46, 189, 251, 193, 92, 189, 251, 193, 152, 189, 251, 193, 208, 189, 251, 193, 6, 190, 251, 193, 8, 190, 251, 193, 118, 190, 251, 193, 190, 190, 251, 193, 212, 190, 251, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 128, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 160, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 193, 166, 186, 251, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 206, 192, 27, 0, 193, 100, 187, 251, 193, 102, 187, 251, 193, 104, 187, 251, 193, 106, 187, 251, 193, 108, 187, 251, 193, 110, 187, 251, 193, 112, 187, 251, 193, 114, 187, 251, 193, 116, 187, 251, 193, 118, 187, 251, 193, 120, 187, 251, 193, 122, 187, 251, 206, 224, 27, 0, 206, 224, 27, 0, 206, 224, 27, 0, 206, 224, 27, 0, 206, 224, 27, 0, 206, 224, 27, 0, 206, 224, 27, 0, 206, 224, 27, 0, 206, 224, 27, 0, 193, 62, 187, 251, 193, 124, 187, 251, 193, 126, 187, 251, 193, 188, 187, 251, 193, 190, 187, 251, 193, 236, 187, 251, 193, 238, 187, 251, 193, 16, 188, 251, 193, 44, 188, 251, 193, 88, 188, 251, 193, 110, 188, 251, 193, 236, 188, 251, 193, 238, 188, 251, 193, 22, 189, 251, 193, 24, 189, 251, 193, 66, 189, 251, 193, 68, 189, 251, 193, 118, 189, 251, 193, 182, 189, 251, 193, 242, 189, 251, 193, 92, 190, 251, 193, 94, 190, 251, 193, 96, 190, 251, 193, 160, 190, 251, 193, 22, 191, 251, 193, 72, 191, 251, 193, 126, 191, 251, 193, 146, 191, 251, 193, 60, 192, 251, 193, 52, 193, 251, 193, 54, 193, 251, 193, 56, 193, 251, 193, 86, 187, 251, 193, 146, 187, 251, 193, 148, 187, 251, 193, 150, 187, 251, 193, 194, 188, 251, 193, 196, 188, 251, 193, 246, 188, 251, 193, 248, 188, 251, 193, 250, 188, 251, 193, 36, 189, 251, 193, 84, 189, 251, 193, 138, 189, 251, 193, 190, 189, 251, 193, 192, 189, 251, 193, 194, 189, 251, 193, 250, 189, 251, 193, 64, 190, 251, 193, 66, 190, 251, 193, 110, 190, 251, 193, 112, 190, 251, 193, 102, 191, 251, 193, 130, 191, 251, 193, 136, 191, 251, 193, 210, 191, 251, 193, 212, 191, 251, 193, 214, 191, 251, 193, 216, 191, 251, 193, 218, 191, 251, 193, 72, 192, 251, 193, 74, 192, 251, 193, 86, 193, 251, 193, 88, 193, 251, 193, 134, 187, 251, 193, 136, 187, 251, 193, 192, 187, 251, 193, 194, 187, 251, 193, 196, 187, 251, 193, 66, 188, 251, 193, 112, 188, 251, 193, 114, 188, 251, 193, 142, 188, 251, 193, 184, 188, 251, 193, 186, 188, 251, 193, 240, 188, 251, 193, 242, 188, 251, 193, 26, 189, 251, 193, 28, 189, 251, 193, 30, 189, 251, 193, 32, 189, 251, 193, 34, 189, 251, 193, 70, 189, 251, 193, 72, 189, 251, 193, 74, 189, 251, 193, 76, 189, 251, 193, 78, 189, 251, 193, 120, 189, 251, 193, 122, 189, 251, 193, 124, 189, 251, 193, 126, 189, 251, 193, 128, 189, 251, 193, 130, 189, 251, 193, 132, 189, 251, 193, 134, 189, 251, 193, 184, 189, 251, 193, 140, 187, 251, 193, 142, 187, 251, 193, 144, 187, 251, 193, 198, 187, 251, 193, 200, 187, 251, 193, 202, 187, 251, 193, 204, 187, 251, 193, 240, 187, 251, 193, 38, 188, 251, 193, 40, 188, 251, 193, 42, 188, 251, 193, 46, 188, 251, 193, 48, 188, 251, 193, 54, 188, 251, 193, 90, 188, 251, 193, 144, 188, 251, 193, 188, 188, 251, 193, 190, 188, 251, 193, 192, 188, 251, 193, 244, 188, 251, 193, 80, 189, 251, 193, 82, 189, 251, 193, 136, 189, 251, 193, 188, 189, 251, 193, 246, 189, 251, 193, 248, 189, 251, 193, 58, 190, 251, 193, 60, 190, 251, 193, 62, 190, 251, 193, 108, 190, 251, 193, 130, 190, 251, 193, 174, 190, 251, 206, 224, 27, 0, 206, 224, 27, 0, 206, 224, 27, 0, 206, 224, 27, 0, 206, 224, 27, 0, 193, 216, 187, 251, 193, 218, 187, 251, 193, 220, 187, 251, 193, 222, 187, 251, 193, 224, 187, 251, 193, 226, 187, 251, 193, 228, 187, 251, 193, 230, 187, 251, 193, 232, 187, 251, 193, 234, 187, 251, 193, 246, 187, 251, 193, 248, 187, 251, 193, 250, 187, 251, 193, 252, 187, 251, 193, 254, 187, 251, 193, 2, 188, 251, 193, 4, 188, 251, 193, 6, 188, 251, 193, 8, 188, 251, 193, 10, 188, 251, 193, 12, 188, 251, 193, 14, 188, 251, 193, 22, 188, 251, 193, 24, 188, 251, 193, 26, 188, 251, 193, 28, 188, 251, 193, 30, 188, 251, 193, 206, 187, 251, 193, 208, 187, 251, 193, 210, 187, 251, 193, 242, 187, 251, 193, 244, 187, 251, 193, 18, 188, 251, 193, 20, 188, 251, 193, 56, 188, 251, 193, 76, 188, 251, 193, 78, 188, 251, 193, 80, 188, 251, 193, 82, 188, 251, 193, 92, 188, 251, 193, 94, 188, 251, 193, 96, 188, 251, 193, 98, 188, 251, 193, 100, 188, 251, 193, 116, 188, 251, 193, 118, 188, 251, 193, 120, 188, 251, 193, 146, 188, 251, 193, 148, 188, 251, 193, 150, 188, 251, 193, 152, 188, 251, 193, 198, 188, 251, 193, 252, 188, 251, 193, 38, 189, 251, 193, 40, 189, 251, 193, 42, 189, 251, 193, 44, 189, 251, 193, 86, 189, 251, 193, 88, 189, 251, 193, 32, 188, 251, 193, 34, 188, 251, 193, 36, 188, 251, 193, 50, 188, 251, 193, 52, 188, 251, 193, 60, 188, 251, 193, 62, 188, 251, 193, 64, 188, 251, 193, 68, 188, 251, 193, 70, 188, 251, 193, 72, 188, 251, 193, 74, 188, 251, 193, 84, 188, 251, 193, 86, 188, 251, 193, 102, 188, 251, 193, 104, 188, 251, 193, 106, 188, 251, 193, 108, 188, 251, 193, 126, 188, 251, 193, 128, 188, 251, 193, 130, 188, 251, 193, 132, 188, 251, 193, 134, 188, 251, 193, 136, 188, 251, 193, 138, 188, 251, 193, 140, 188, 251, 206, 0, 28, 0, 206, 0, 28, 0, 206, 0, 28, 0, 206, 0, 28, 0, 206, 0, 28, 0, 206, 0, 28, 0, 206, 0, 28, 0, 206, 0, 28, 0, 193, 202, 188, 251, 193, 204, 188, 251, 193, 206, 188, 251, 193, 208, 188, 251, 193, 210, 188, 251, 193, 212, 188, 251, 193, 214, 188, 251, 193, 216, 188, 251, 193, 218, 188, 251, 193, 220, 188, 251, 193, 222, 188, 251, 193, 224, 188, 251, 193, 226, 188, 251, 193, 228, 188, 251, 193, 230, 188, 251, 193, 232, 188, 251, 193, 234, 188, 251, 193, 6, 189, 251, 193, 8, 189, 251, 193, 10, 189, 251, 193, 12, 189, 251, 193, 14, 189, 251, 193, 16, 189, 251, 193, 18, 189, 251, 193, 20, 189, 251, 193, 48, 189, 251, 193, 50, 189, 251, 193, 52, 189, 251, 193, 54, 189, 251, 193, 56, 189, 251, 193, 58, 189, 251, 193, 60, 189, 251, 193, 62, 189, 251, 193, 64, 189, 251, 193, 94, 189, 251, 193, 96, 189, 251, 193, 98, 189, 251, 193, 100, 189, 251, 193, 102, 189, 251, 193, 104, 189, 251, 193, 106, 189, 251, 193, 108, 189, 251, 193, 110, 189, 251, 193, 112, 189, 251, 193, 114, 189, 251, 193, 116, 189, 251, 193, 154, 189, 251, 193, 156, 189, 251, 193, 158, 189, 251, 193, 160, 189, 251, 193, 162, 189, 251, 193, 164, 189, 251, 193, 166, 189, 251, 193, 168, 189, 251, 193, 170, 189, 251, 193, 172, 189, 251, 193, 174, 189, 251, 193, 176, 189, 251, 193, 140, 189, 251, 193, 142, 189, 251, 193, 144, 189, 251, 193, 146, 189, 251, 193, 148, 189, 251, 193, 150, 189, 251, 193, 90, 189, 251, 193, 196, 189, 251, 193, 198, 189, 251, 193, 200, 189, 251, 193, 202, 189, 251, 193, 204, 189, 251, 193, 206, 189, 251, 193, 252, 189, 251, 193, 254, 189, 251, 193, 2, 190, 251, 193, 68, 190, 251, 193, 4, 190, 251, 193, 114, 190, 251, 193, 132, 190, 251, 193, 116, 190, 251, 193, 158, 190, 251, 193, 162, 190, 251, 193, 164, 190, 251, 193, 166, 190, 251, 193, 168, 190, 251, 193, 184, 190, 251, 193, 186, 190, 251, 193, 188, 190, 251, 193, 204, 190, 251, 193, 206, 190, 251, 193, 208, 190, 251, 193, 178, 189, 251, 193, 180, 189, 251, 193, 212, 189, 251, 193, 214, 189, 251, 193, 216, 189, 251, 193, 218, 189, 251, 193, 220, 189, 251, 193, 222, 189, 251, 193, 10, 190, 251, 193, 224, 189, 251, 193, 226, 189, 251, 193, 228, 189, 251, 193, 230, 189, 251, 193, 232, 189, 251, 193, 234, 189, 251, 193, 236, 189, 251, 193, 238, 189, 251, 193, 240, 189, 251, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 206, 32, 28, 0, 193, 186, 189, 251, 193, 244, 189, 251, 193, 54, 190, 251, 193, 56, 190, 251, 193, 98, 190, 251, 193, 100, 190, 251, 193, 102, 190, 251, 193, 104, 190, 251, 193, 106, 190, 251, 193, 230, 190, 251, 193, 24, 191, 251, 193, 58, 191, 251, 193, 60, 191, 251, 193, 74, 191, 251, 193, 76, 191, 251, 193, 92, 191, 251, 193, 94, 191, 251, 193, 110, 191, 251, 193, 154, 191, 251, 193, 174, 191, 251, 193, 206, 191, 251, 193, 62, 192, 251, 193, 64, 192, 251, 193, 66, 192, 251, 193, 66, 193, 251, 193, 68, 193, 251, 193, 70, 193, 251, 193, 72, 193, 251, 193, 140, 194, 251, 193, 142, 194, 251, 193, 144, 194, 251, 193, 146, 194, 251, 206, 64, 28, 0, 206, 64, 28, 0, 206, 64, 28, 0, 206, 64, 28, 0, 193, 170, 190, 251, 193, 172, 190, 251, 193, 176, 190, 251, 193, 178, 190, 251, 193, 180, 190, 251, 193, 182, 190, 251, 193, 192, 190, 251, 193, 194, 190, 251, 193, 196, 190, 251, 193, 198, 190, 251, 193, 200, 190, 251, 193, 216, 190, 251, 193, 218, 190, 251, 193, 220, 190, 251, 193, 222, 190, 251, 193, 224, 190, 251, 193, 226, 190, 251, 193, 228, 190, 251, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 193, 202, 190, 251, 193, 232, 190, 251, 193, 234, 190, 251, 193, 26, 191, 251, 193, 28, 191, 251, 193, 30, 191, 251, 193, 32, 191, 251, 193, 96, 191, 251, 193, 98, 191, 251, 193, 100, 191, 251, 193, 128, 191, 251, 193, 134, 191, 251, 193, 152, 191, 251, 193, 164, 191, 251, 193, 208, 191, 251, 193, 68, 192, 251, 193, 70, 192, 251, 193, 74, 193, 251, 193, 76, 193, 251, 193, 78, 193, 251, 193, 80, 193, 251, 193, 82, 193, 251, 193, 84, 193, 251, 193, 152, 194, 251, 193, 154, 194, 251, 193, 156, 194, 251, 193, 158, 194, 251, 193, 160, 194, 251, 193, 162, 194, 251, 193, 164, 194, 251, 193, 166, 194, 251, 193, 168, 194, 251, 193, 210, 190, 251, 193, 236, 190, 251, 193, 238, 190, 251, 193, 240, 190, 251, 193, 34, 191, 251, 193, 36, 191, 251, 193, 38, 191, 251, 193, 40, 191, 251, 193, 62, 191, 251, 193, 78, 191, 251, 193, 80, 191, 251, 193, 156, 191, 251, 193, 158, 191, 251, 193, 160, 191, 251, 193, 166, 191, 251, 193, 220, 191, 251, 193, 176, 191, 251, 193, 178, 191, 251, 193, 180, 191, 251, 193, 182, 191, 251, 193, 184, 191, 251, 193, 186, 191, 251, 193, 188, 191, 251, 193, 222, 191, 251, 193, 224, 191, 251, 193, 226, 191, 251, 193, 228, 191, 251, 193, 230, 191, 251, 193, 232, 191, 251, 193, 234, 191, 251, 193, 236, 191, 251, 193, 76, 192, 251, 193, 214, 190, 251, 193, 242, 190, 251, 193, 244, 190, 251, 193, 190, 191, 251, 193, 192, 191, 251, 193, 194, 191, 251, 193, 10, 192, 251, 193, 12, 192, 251, 193, 14, 192, 251, 193, 190, 192, 251, 193, 192, 192, 251, 193, 194, 192, 251, 193, 196, 192, 251, 193, 198, 192, 251, 193, 200, 192, 251, 193, 202, 192, 251, 193, 204, 192, 251, 193, 206, 192, 251, 193, 252, 193, 251, 193, 254, 193, 251, 193, 2, 194, 251, 193, 4, 194, 251, 193, 6, 194, 251, 193, 8, 194, 251, 193, 10, 194, 251, 193, 12, 194, 251, 193, 14, 194, 251, 193, 16, 194, 251, 193, 18, 194, 251, 193, 20, 194, 251, 193, 22, 194, 251, 193, 24, 194, 251, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 206, 96, 28, 0, 193, 42, 191, 251, 193, 44, 191, 251, 193, 46, 191, 251, 193, 48, 191, 251, 193, 50, 191, 251, 193, 52, 191, 251, 193, 54, 191, 251, 193, 56, 191, 251, 193, 64, 191, 251, 193, 66, 191, 251, 193, 68, 191, 251, 193, 70, 191, 251, 193, 82, 191, 251, 193, 84, 191, 251, 193, 86, 191, 251, 193, 88, 191, 251, 193, 90, 191, 251, 193, 104, 191, 251, 193, 106, 191, 251, 193, 112, 191, 251, 193, 114, 191, 251, 193, 116, 191, 251, 193, 118, 191, 251, 193, 120, 191, 251, 193, 122, 191, 251, 193, 124, 191, 251, 193, 132, 191, 251, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 193, 108, 191, 251, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 193, 138, 191, 251, 193, 140, 191, 251, 193, 144, 191, 251, 193, 148, 191, 251, 193, 150, 191, 251, 193, 168, 191, 251, 193, 170, 191, 251, 193, 172, 191, 251, 193, 196, 191, 251, 193, 198, 191, 251, 193, 200, 191, 251, 193, 202, 191, 251, 193, 204, 191, 251, 193, 16, 192, 251, 193, 18, 192, 251, 193, 20, 192, 251, 193, 22, 192, 251, 193, 24, 192, 251, 193, 26, 192, 251, 193, 28, 192, 251, 193, 30, 192, 251, 193, 32, 192, 251, 193, 34, 192, 251, 193, 36, 192, 251, 193, 38, 192, 251, 193, 40, 192, 251, 193, 42, 192, 251, 193, 44, 192, 251, 193, 46, 192, 251, 193, 48, 192, 251, 193, 50, 192, 251, 193, 52, 192, 251, 193, 36, 78, 253, 193, 38, 78, 253, 193, 40, 78, 253, 193, 42, 78, 253, 193, 44, 78, 253, 193, 46, 78, 253, 193, 48, 78, 253, 193, 50, 78, 253, 193, 56, 78, 253, 193, 58, 78, 253, 193, 60, 78, 253, 193, 62, 78, 253, 193, 64, 78, 253, 193, 66, 78, 253, 193, 68, 78, 253, 193, 70, 78, 253, 193, 74, 78, 253, 193, 76, 78, 253, 193, 78, 78, 253, 193, 142, 191, 251, 193, 204, 97, 252, 193, 86, 78, 253, 193, 88, 78, 253, 193, 90, 78, 253, 193, 92, 78, 253, 193, 94, 78, 253, 193, 102, 78, 253, 193, 104, 78, 253, 193, 106, 78, 253, 193, 112, 78, 253, 193, 116, 78, 253, 193, 118, 78, 253, 193, 238, 191, 251, 193, 240, 191, 251, 193, 242, 191, 251, 193, 244, 191, 251, 193, 246, 191, 251, 193, 248, 191, 251, 193, 250, 191, 251, 193, 252, 191, 251, 193, 254, 191, 251, 193, 2, 192, 251, 193, 4, 192, 251, 193, 6, 192, 251, 193, 8, 192, 251, 193, 78, 192, 251, 193, 80, 192, 251, 193, 82, 192, 251, 193, 84, 192, 251, 193, 86, 192, 251, 193, 88, 192, 251, 193, 90, 192, 251, 193, 92, 192, 251, 193, 94, 192, 251, 193, 96, 192, 251, 193, 98, 192, 251, 193, 114, 193, 251, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 193, 54, 192, 251, 193, 56, 192, 251, 193, 58, 192, 251, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 128, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 160, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 193, 58, 193, 251, 193, 60, 193, 251, 193, 62, 193, 251, 193, 64, 193, 251, 193, 124, 194, 251, 193, 126, 194, 251, 193, 128, 194, 251, 193, 130, 194, 251, 193, 132, 194, 251, 193, 134, 194, 251, 193, 136, 194, 251, 193, 250, 195, 251, 193, 252, 195, 251, 193, 254, 195, 251, 193, 2, 196, 251, 193, 4, 196, 251, 193, 6, 196, 251, 193, 8, 196, 251, 193, 10, 196, 251, 193, 12, 196, 251, 193, 166, 197, 251, 193, 168, 197, 251, 193, 170, 197, 251, 193, 172, 197, 251, 193, 174, 197, 251, 193, 176, 197, 251, 193, 178, 197, 251, 193, 180, 197, 251, 193, 182, 197, 251, 193, 184, 197, 251, 193, 186, 197, 251, 193, 188, 197, 251, 193, 90, 193, 251, 193, 92, 193, 251, 193, 94, 193, 251, 193, 96, 193, 251, 193, 98, 193, 251, 193, 100, 193, 251, 193, 102, 193, 251, 193, 104, 193, 251, 193, 106, 193, 251, 193, 108, 193, 251, 193, 110, 193, 251, 193, 112, 193, 251, 193, 174, 194, 251, 193, 176, 194, 251, 193, 178, 194, 251, 193, 180, 194, 251, 193, 182, 194, 251, 193, 184, 194, 251, 193, 186, 194, 251, 193, 188, 194, 251, 193, 190, 194, 251, 193, 192, 194, 251, 193, 194, 194, 251, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 192, 28, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 193, 26, 194, 251, 193, 66, 195, 251, 193, 68, 195, 251, 193, 70, 195, 251, 193, 72, 195, 251, 193, 74, 195, 251, 193, 76, 195, 251, 193, 78, 195, 251, 193, 80, 195, 251, 193, 82, 195, 251, 193, 84, 195, 251, 193, 86, 195, 251, 193, 88, 195, 251, 193, 90, 195, 251, 193, 92, 195, 251, 193, 94, 195, 251, 193, 96, 195, 251, 193, 98, 195, 251, 193, 242, 196, 251, 193, 244, 196, 251, 193, 246, 196, 251, 193, 248, 196, 251, 193, 250, 196, 251, 193, 252, 196, 251, 193, 254, 196, 251, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 224, 28, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 193, 148, 194, 251, 193, 150, 194, 251, 193, 16, 196, 251, 193, 18, 196, 251, 193, 20, 196, 251, 193, 22, 196, 251, 193, 24, 196, 251, 193, 26, 196, 251, 193, 28, 196, 251, 193, 30, 196, 251, 193, 32, 196, 251, 193, 34, 196, 251, 193, 36, 196, 251, 193, 38, 196, 251, 193, 40, 196, 251, 193, 42, 196, 251, 193, 44, 196, 251, 193, 192, 197, 251, 193, 194, 197, 251, 193, 196, 197, 251, 193, 198, 197, 251, 193, 200, 197, 251, 193, 202, 197, 251, 193, 204, 197, 251, 193, 206, 197, 251, 193, 208, 197, 251, 193, 210, 197, 251, 193, 212, 197, 251, 193, 214, 197, 251, 193, 216, 197, 251, 193, 218, 197, 251, 193, 220, 197, 251, 193, 170, 194, 251, 193, 172, 194, 251, 193, 46, 196, 251, 193, 48, 196, 251, 193, 50, 196, 251, 193, 52, 196, 251, 193, 54, 196, 251, 193, 56, 196, 251, 193, 58, 196, 251, 193, 60, 196, 251, 193, 62, 196, 251, 193, 64, 196, 251, 193, 66, 196, 251, 193, 68, 196, 251, 193, 70, 196, 251, 193, 72, 196, 251, 193, 74, 196, 251, 193, 76, 196, 251, 193, 78, 196, 251, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 0, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 32, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 206, 64, 29, 0, 193, 14, 198, 251, 193, 16, 198, 251, 193, 18, 198, 251, 193, 20, 198, 251, 193, 22, 198, 251, 193, 24, 198, 251, 193, 26, 198, 251, 193, 28, 198, 251, 193, 30, 198, 251, 193, 32, 198, 251, 193, 34, 198, 251, 193, 36, 198, 251, 193, 38, 198, 251, 193, 40, 198, 251, 193, 42, 198, 251, 206, 32, 30, 0, 206, 32, 30, 0, 206, 32, 30, 0, 206, 32, 30, 0, 206, 32, 30, 0, 206, 32, 30, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 96, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 128, 29, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 193, 190, 197, 251, 193, 206, 199, 251, 193, 208, 199, 251, 193, 210, 199, 251, 193, 212, 199, 251, 193, 214, 199, 251, 193, 216, 199, 251, 193, 218, 199, 251, 193, 220, 199, 251, 193, 222, 199, 251, 193, 224, 199, 251, 193, 226, 199, 251, 193, 238, 201, 251, 193, 240, 201, 251, 193, 242, 201, 251, 193, 244, 201, 251, 193, 246, 201, 251, 193, 248, 201, 251, 193, 250, 201, 251, 193, 252, 201, 251, 193, 254, 201, 251, 193, 2, 202, 251, 193, 4, 202, 251, 193, 202, 203, 251, 193, 204, 203, 251, 193, 206, 203, 251, 193, 208, 203, 251, 193, 210, 203, 251, 193, 212, 203, 251, 193, 214, 203, 251, 193, 216, 203, 251, 193, 218, 203, 251, 193, 222, 197, 251, 193, 228, 199, 251, 193, 230, 199, 251, 193, 232, 199, 251, 193, 234, 199, 251, 193, 236, 199, 251, 193, 238, 199, 251, 193, 240, 199, 251, 193, 242, 199, 251, 193, 244, 199, 251, 193, 246, 199, 251, 193, 248, 199, 251, 193, 250, 199, 251, 193, 252, 199, 251, 193, 254, 199, 251, 193, 2, 200, 251, 193, 4, 200, 251, 193, 6, 200, 251, 193, 8, 200, 251, 193, 10, 200, 251, 193, 12, 200, 251, 193, 14, 200, 251, 193, 16, 200, 251, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 206, 160, 29, 0, 193, 18, 200, 251, 193, 20, 200, 251, 193, 22, 200, 251, 193, 24, 200, 251, 193, 26, 200, 251, 193, 28, 200, 251, 193, 30, 200, 251, 193, 32, 200, 251, 193, 34, 200, 251, 193, 36, 200, 251, 193, 38, 200, 251, 193, 40, 200, 251, 193, 42, 200, 251, 193, 44, 200, 251, 193, 46, 200, 251, 193, 48, 200, 251, 193, 50, 200, 251, 193, 52, 200, 251, 193, 54, 200, 251, 193, 56, 200, 251, 193, 58, 200, 251, 193, 44, 202, 251, 193, 46, 202, 251, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 206, 192, 29, 0, 193, 82, 200, 251, 193, 198, 198, 251, 193, 200, 198, 251, 193, 202, 198, 251, 193, 204, 198, 251, 193, 206, 198, 251, 193, 208, 198, 251, 193, 210, 198, 251, 193, 212, 198, 251, 193, 214, 198, 251, 193, 216, 198, 251, 193, 218, 198, 251, 193, 84, 200, 251, 193, 86, 200, 251, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 206, 224, 29, 0, 193, 228, 200, 251, 193, 230, 200, 251, 193, 232, 200, 251, 193, 234, 200, 251, 193, 236, 200, 251, 193, 238, 200, 251, 193, 240, 200, 251, 193, 242, 200, 251, 193, 244, 200, 251, 193, 246, 200, 251, 193, 248, 200, 251, 193, 250, 200, 251, 193, 252, 200, 251, 193, 246, 202, 251, 193, 254, 200, 251, 193, 2, 201, 251, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 0, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 32, 30, 0, 206, 32, 30, 0, 206, 32, 30, 0, 206, 32, 30, 0, 206, 32, 30, 0, 193, 84, 202, 251, 193, 86, 202, 251, 193, 88, 202, 251, 193, 90, 202, 251, 193, 92, 202, 251, 193, 94, 202, 251, 193, 96, 202, 251, 193, 98, 202, 251, 193, 100, 202, 251, 193, 102, 202, 251, 193, 104, 202, 251, 193, 106, 202, 251, 193, 108, 202, 251, 193, 44, 204, 251, 193, 46, 204, 251, 193, 48, 204, 251, 193, 50, 204, 251, 193, 52, 204, 251, 193, 54, 204, 251, 193, 56, 204, 251, 193, 58, 204, 251, 193, 60, 204, 251, 193, 62, 204, 251, 193, 64, 204, 251, 193, 66, 204, 251, 193, 68, 204, 251, 193, 10, 206, 251, 193, 88, 200, 251, 193, 90, 200, 251, 193, 92, 200, 251, 193, 94, 200, 251, 193, 96, 200, 251, 193, 70, 204, 251, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 64, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 193, 72, 204, 251, 193, 74, 204, 251, 193, 76, 204, 251, 193, 78, 204, 251, 193, 80, 204, 251, 193, 82, 204, 251, 193, 84, 204, 251, 193, 226, 200, 251, 193, 86, 204, 251, 193, 88, 204, 251, 193, 90, 204, 251, 193, 26, 206, 251, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 206, 96, 30, 0, 193, 148, 117, 252, 193, 176, 201, 251, 193, 178, 201, 251, 193, 180, 201, 251, 193, 182, 201, 251, 193, 184, 201, 251, 193, 186, 201, 251, 193, 188, 201, 251, 193, 190, 201, 251, 193, 192, 201, 251, 193, 194, 201, 251, 193, 196, 201, 251, 193, 198, 201, 251, 193, 200, 201, 251, 193, 202, 201, 251, 193, 204, 201, 251, 193, 206, 201, 251, 193, 208, 201, 251, 193, 210, 201, 251, 193, 212, 201, 251, 193, 214, 201, 251, 193, 216, 201, 251, 193, 218, 201, 251, 193, 220, 201, 251, 193, 222, 201, 251, 193, 224, 201, 251, 193, 226, 201, 251, 193, 228, 201, 251, 193, 230, 201, 251, 193, 232, 201, 251, 193, 234, 201, 251, 193, 236, 201, 251, 206, 0, 31, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 206, 128, 30, 0, 193, 228, 203, 251, 193, 230, 203, 251, 193, 232, 203, 251, 193, 234, 203, 251, 193, 236, 203, 251, 193, 238, 203, 251, 193, 240, 203, 251, 193, 242, 203, 251, 193, 244, 203, 251, 193, 246, 203, 251, 193, 248, 203, 251, 193, 250, 203, 251, 193, 252, 203, 251, 193, 254, 203, 251, 193, 210, 205, 251, 193, 212, 205, 251, 193, 214, 205, 251, 193, 216, 205, 251, 193, 218, 205, 251, 193, 220, 205, 251, 193, 222, 205, 251, 193, 224, 205, 251, 193, 226, 205, 251, 193, 48, 202, 251, 193, 50, 202, 251, 193, 52, 202, 251, 193, 54, 202, 251, 193, 56, 202, 251, 193, 58, 202, 251, 193, 60, 202, 251, 193, 62, 202, 251, 193, 64, 202, 251, 193, 66, 202, 251, 193, 68, 202, 251, 193, 70, 202, 251, 193, 72, 202, 251, 193, 74, 202, 251, 193, 76, 202, 251, 193, 78, 202, 251, 193, 80, 202, 251, 193, 82, 202, 251, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 206, 160, 30, 0, 193, 24, 206, 251, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 192, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 206, 224, 30, 0, 193, 186, 204, 251, 193, 188, 204, 251, 193, 190, 204, 251, 193, 192, 204, 251, 193, 194, 204, 251, 193, 196, 204, 251, 193, 198, 204, 251, 193, 200, 204, 251, 193, 202, 204, 251, 193, 204, 204, 251, 193, 206, 204, 251, 193, 208, 204, 251, 193, 210, 204, 251, 193, 212, 204, 251, 193, 214, 204, 251, 193, 216, 204, 251, 193, 218, 204, 251, 193, 220, 204, 251, 193, 222, 204, 251, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 0, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 193, 220, 203, 251, 193, 222, 203, 251, 193, 224, 203, 251, 193, 192, 205, 251, 193, 194, 205, 251, 193, 196, 205, 251, 193, 198, 205, 251, 193, 200, 205, 251, 193, 202, 205, 251, 193, 204, 205, 251, 193, 206, 205, 251, 193, 106, 207, 251, 193, 108, 207, 251, 193, 110, 207, 251, 193, 112, 207, 251, 193, 114, 207, 251, 193, 116, 207, 251, 193, 118, 207, 251, 193, 120, 207, 251, 193, 122, 207, 251, 193, 124, 207, 251, 193, 156, 208, 251, 193, 158, 208, 251, 193, 160, 208, 251, 193, 102, 209, 251, 193, 104, 209, 251, 193, 106, 209, 251, 193, 52, 210, 251, 193, 54, 210, 251, 193, 56, 210, 251, 193, 58, 210, 251, 193, 60, 210, 251, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 206, 32, 31, 0, 193, 232, 205, 251, 193, 234, 205, 251, 193, 236, 205, 251, 193, 238, 205, 251, 193, 240, 205, 251, 193, 242, 205, 251, 193, 244, 205, 251, 193, 246, 205, 251, 193, 248, 205, 251, 193, 250, 205, 251, 193, 252, 205, 251, 193, 254, 205, 251, 193, 2, 206, 251, 193, 4, 206, 251, 193, 6, 206, 251, 193, 8, 206, 251, 193, 142, 207, 251, 193, 144, 207, 251, 193, 146, 207, 251, 193, 148, 207, 251, 193, 150, 207, 251, 193, 152, 207, 251, 193, 154, 207, 251, 193, 156, 207, 251, 193, 158, 207, 251, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 206, 64, 31, 0, 193, 28, 206, 251, 193, 30, 206, 251, 193, 184, 204, 251, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 193, 224, 204, 251, 193, 226, 204, 251, 193, 228, 204, 251, 193, 136, 206, 251, 193, 138, 206, 251, 193, 140, 206, 251, 193, 142, 206, 251, 193, 144, 206, 251, 193, 146, 206, 251, 193, 148, 206, 251, 193, 150, 206, 251, 193, 152, 206, 251, 193, 154, 206, 251, 193, 156, 206, 251, 193, 158, 206, 251, 193, 160, 206, 251, 193, 162, 206, 251, 193, 164, 206, 251, 193, 244, 207, 251, 193, 246, 207, 251, 193, 248, 207, 251, 193, 250, 207, 251, 193, 252, 207, 251, 193, 254, 207, 251, 193, 2, 208, 251, 193, 4, 208, 251, 193, 4, 209, 251, 193, 6, 209, 251, 193, 8, 209, 251, 193, 10, 209, 251, 193, 12, 209, 251, 193, 14, 209, 251, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 206, 96, 31, 0, 193, 166, 206, 251, 193, 172, 205, 251, 193, 174, 205, 251, 193, 176, 205, 251, 193, 178, 205, 251, 193, 180, 205, 251, 193, 182, 205, 251, 193, 184, 205, 251, 193, 186, 205, 251, 193, 188, 205, 251, 193, 190, 205, 251, 193, 228, 205, 251, 193, 230, 205, 251, 193, 126, 207, 251, 193, 128, 207, 251, 193, 130, 207, 251, 193, 132, 207, 251, 193, 134, 207, 251, 193, 136, 207, 251, 193, 138, 207, 251, 193, 140, 207, 251, 193, 162, 208, 251, 193, 164, 208, 251, 193, 166, 208, 251, 193, 168, 208, 251, 193, 170, 208, 251, 193, 172, 208, 251, 193, 108, 209, 251, 193, 110, 209, 251, 193, 112, 209, 251, 193, 66, 210, 251, 193, 68, 210, 251, 193, 70, 210, 251, 193, 72, 210, 251, 193, 74, 210, 251, 193, 76, 210, 251, 193, 206, 210, 251, 193, 208, 210, 251, 193, 60, 211, 251, 193, 136, 211, 251, 193, 138, 211, 251, 193, 140, 211, 251, 193, 212, 211, 251, 193, 12, 206, 251, 193, 14, 206, 251, 193, 16, 206, 251, 193, 18, 206, 251, 193, 20, 206, 251, 193, 22, 206, 251, 193, 166, 207, 251, 193, 168, 207, 251, 193, 170, 207, 251, 193, 172, 207, 251, 193, 204, 208, 251, 193, 206, 208, 251, 193, 208, 208, 251, 193, 210, 208, 251, 193, 212, 208, 251, 193, 214, 208, 251, 193, 216, 208, 251, 193, 218, 208, 251, 193, 220, 208, 251, 193, 138, 209, 251, 193, 140, 209, 251, 193, 142, 209, 251, 193, 144, 209, 251, 193, 146, 209, 251, 193, 90, 210, 251, 193, 92, 210, 251, 193, 226, 210, 251, 193, 72, 211, 251, 193, 74, 211, 251, 193, 146, 211, 251, 193, 148, 211, 251, 193, 216, 211, 251, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 128, 31, 0, 206, 192, 31, 0, 206, 192, 31, 0, 206, 192, 31, 0, 206, 192, 31, 0, 206, 192, 31, 0, 206, 192, 31, 0, 206, 192, 31, 0, 193, 134, 206, 251, 193, 196, 207, 251, 193, 198, 207, 251, 193, 200, 207, 251, 193, 202, 207, 251, 193, 204, 207, 251, 193, 206, 207, 251, 193, 208, 207, 251, 193, 210, 207, 251, 193, 212, 207, 251, 193, 214, 207, 251, 193, 216, 207, 251, 193, 218, 207, 251, 193, 220, 207, 251, 193, 222, 207, 251, 193, 224, 207, 251, 193, 226, 207, 251, 193, 228, 207, 251, 193, 230, 207, 251, 193, 232, 207, 251, 193, 234, 207, 251, 193, 236, 207, 251, 193, 238, 207, 251, 193, 240, 207, 251, 193, 242, 207, 251, 193, 222, 208, 251, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 160, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 193, 104, 207, 251, 193, 146, 208, 251, 193, 148, 208, 251, 193, 150, 208, 251, 193, 152, 208, 251, 193, 154, 208, 251, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 193, 160, 207, 251, 193, 162, 207, 251, 193, 164, 207, 251, 193, 174, 208, 251, 193, 176, 208, 251, 193, 178, 208, 251, 193, 180, 208, 251, 193, 182, 208, 251, 193, 184, 208, 251, 193, 186, 208, 251, 193, 188, 208, 251, 193, 190, 208, 251, 193, 192, 208, 251, 193, 194, 208, 251, 193, 196, 208, 251, 193, 198, 208, 251, 193, 200, 208, 251, 193, 202, 208, 251, 193, 114, 209, 251, 193, 116, 209, 251, 193, 118, 209, 251, 193, 120, 209, 251, 193, 122, 209, 251, 193, 124, 209, 251, 193, 126, 209, 251, 193, 128, 209, 251, 193, 130, 209, 251, 193, 132, 209, 251, 193, 134, 209, 251, 193, 136, 209, 251, 193, 78, 210, 251, 193, 80, 210, 251, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 206, 224, 31, 0, 193, 224, 208, 251, 193, 226, 208, 251, 193, 228, 208, 251, 193, 230, 208, 251, 193, 232, 208, 251, 193, 148, 209, 251, 193, 234, 208, 251, 193, 236, 208, 251, 193, 238, 208, 251, 193, 240, 208, 251, 193, 242, 208, 251, 193, 244, 208, 251, 193, 246, 208, 251, 193, 248, 208, 251, 193, 250, 208, 251, 193, 252, 208, 251, 193, 254, 208, 251, 193, 2, 209, 251, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 206, 0, 32, 0, 193, 216, 209, 251, 193, 218, 209, 251, 193, 220, 209, 251, 193, 222, 209, 251, 193, 224, 209, 251, 193, 226, 209, 251, 193, 228, 209, 251, 193, 230, 209, 251, 193, 232, 209, 251, 193, 234, 209, 251, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 206, 32, 32, 0, 193, 94, 210, 251, 193, 96, 210, 251, 193, 98, 210, 251, 193, 100, 210, 251, 193, 102, 210, 251, 193, 104, 210, 251, 193, 106, 210, 251, 193, 108, 210, 251, 193, 110, 210, 251, 193, 112, 210, 251, 193, 228, 210, 251, 193, 114, 210, 251, 193, 230, 210, 251, 193, 232, 210, 251, 193, 234, 210, 251, 193, 236, 210, 251, 193, 238, 210, 251, 193, 240, 210, 251, 193, 242, 210, 251, 193, 244, 210, 251, 193, 76, 211, 251, 193, 78, 211, 251, 193, 80, 211, 251, 193, 150, 211, 251, 193, 152, 211, 251, 193, 192, 209, 251, 193, 194, 209, 251, 193, 196, 209, 251, 193, 198, 209, 251, 193, 200, 209, 251, 193, 202, 209, 251, 193, 204, 209, 251, 193, 206, 209, 251, 193, 208, 209, 251, 193, 210, 209, 251, 193, 212, 209, 251, 193, 214, 209, 251, 193, 116, 210, 251, 193, 118, 210, 251, 193, 246, 210, 251, 193, 248, 210, 251, 193, 250, 210, 251, 193, 82, 211, 251, 193, 84, 211, 251, 193, 86, 211, 251, 193, 88, 211, 251, 193, 228, 211, 251, 193, 230, 211, 251, 193, 232, 211, 251, 193, 36, 212, 251, 193, 38, 212, 251, 193, 76, 212, 251, 193, 106, 212, 251, 193, 108, 212, 251, 193, 128, 212, 251, 193, 142, 212, 251, 193, 254, 212, 251, 193, 236, 209, 251, 193, 238, 209, 251, 193, 240, 209, 251, 193, 242, 209, 251, 193, 244, 209, 251, 193, 246, 209, 251, 193, 248, 209, 251, 193, 250, 209, 251, 193, 252, 209, 251, 193, 254, 209, 251, 193, 2, 210, 251, 193, 4, 210, 251, 193, 6, 210, 251, 193, 8, 210, 251, 193, 10, 210, 251, 193, 12, 210, 251, 193, 14, 210, 251, 193, 16, 210, 251, 193, 18, 210, 251, 193, 20, 210, 251, 193, 22, 210, 251, 193, 24, 210, 251, 193, 26, 210, 251, 193, 28, 210, 251, 193, 30, 210, 251, 193, 32, 210, 251, 193, 34, 210, 251, 193, 36, 210, 251, 193, 38, 210, 251, 193, 40, 210, 251, 193, 42, 210, 251, 193, 44, 210, 251, 193, 46, 210, 251, 193, 48, 210, 251, 193, 50, 210, 251, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 193, 62, 210, 251, 193, 204, 210, 251, 193, 58, 211, 251, 193, 210, 211, 251, 193, 250, 211, 251, 193, 138, 212, 251, 193, 92, 212, 251, 193, 188, 212, 251, 193, 80, 213, 251, 193, 148, 213, 251, 193, 150, 213, 251, 193, 248, 213, 251, 193, 250, 213, 251, 193, 142, 214, 251, 193, 48, 215, 251, 193, 50, 215, 251, 193, 182, 215, 251, 193, 40, 216, 251, 193, 42, 216, 251, 193, 136, 216, 251, 193, 138, 216, 251, 193, 140, 216, 251, 193, 142, 216, 251, 193, 244, 216, 251, 193, 246, 216, 251, 193, 248, 216, 251, 193, 54, 217, 251, 193, 102, 217, 251, 193, 104, 217, 251, 193, 200, 217, 251, 193, 202, 217, 251, 193, 232, 217, 251, 193, 82, 210, 251, 193, 84, 210, 251, 193, 86, 210, 251, 193, 88, 210, 251, 193, 210, 210, 251, 193, 212, 210, 251, 193, 214, 210, 251, 193, 216, 210, 251, 193, 218, 210, 251, 193, 220, 210, 251, 193, 222, 210, 251, 193, 224, 210, 251, 193, 62, 211, 251, 193, 64, 211, 251, 193, 66, 211, 251, 193, 68, 211, 251, 193, 70, 211, 251, 193, 142, 211, 251, 193, 144, 211, 251, 193, 56, 212, 251, 193, 62, 212, 251, 193, 68, 212, 251, 193, 72, 212, 251, 193, 102, 212, 251, 193, 104, 212, 251, 193, 168, 212, 251, 193, 170, 212, 251, 193, 190, 212, 251, 193, 206, 212, 251, 193, 208, 212, 251, 193, 220, 212, 251, 193, 224, 212, 251, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 206, 64, 32, 0, 193, 252, 210, 251, 193, 254, 210, 251, 193, 2, 211, 251, 193, 4, 211, 251, 193, 6, 211, 251, 193, 8, 211, 251, 193, 10, 211, 251, 193, 12, 211, 251, 193, 14, 211, 251, 193, 16, 211, 251, 193, 18, 211, 251, 193, 20, 211, 251, 193, 22, 211, 251, 193, 24, 211, 251, 193, 26, 211, 251, 193, 28, 211, 251, 193, 30, 211, 251, 193, 32, 211, 251, 193, 34, 211, 251, 193, 36, 211, 251, 193, 38, 211, 251, 193, 90, 211, 251, 193, 40, 211, 251, 193, 42, 211, 251, 193, 44, 211, 251, 193, 46, 211, 251, 193, 48, 211, 251, 193, 50, 211, 251, 193, 52, 211, 251, 193, 54, 211, 251, 193, 56, 211, 251, 193, 92, 211, 251, 193, 94, 211, 251, 193, 96, 211, 251, 193, 98, 211, 251, 193, 100, 211, 251, 193, 102, 211, 251, 193, 104, 211, 251, 193, 106, 211, 251, 193, 108, 211, 251, 193, 110, 211, 251, 193, 112, 211, 251, 193, 114, 211, 251, 193, 116, 211, 251, 193, 118, 211, 251, 193, 120, 211, 251, 193, 122, 211, 251, 193, 124, 211, 251, 193, 126, 211, 251, 193, 128, 211, 251, 193, 130, 211, 251, 193, 132, 211, 251, 193, 134, 211, 251, 193, 170, 211, 251, 193, 172, 211, 251, 193, 174, 211, 251, 193, 176, 211, 251, 193, 178, 211, 251, 193, 180, 211, 251, 193, 182, 211, 251, 193, 184, 211, 251, 193, 186, 211, 251, 193, 188, 211, 251, 193, 190, 211, 251, 193, 192, 211, 251, 193, 194, 211, 251, 193, 196, 211, 251, 193, 198, 211, 251, 193, 200, 211, 251, 193, 202, 211, 251, 193, 204, 211, 251, 193, 206, 211, 251, 193, 208, 211, 251, 193, 234, 211, 251, 193, 236, 211, 251, 193, 238, 211, 251, 193, 240, 211, 251, 193, 242, 211, 251, 193, 244, 211, 251, 193, 246, 211, 251, 193, 248, 211, 251, 193, 4, 212, 251, 193, 6, 212, 251, 193, 154, 211, 251, 193, 156, 211, 251, 193, 158, 211, 251, 193, 160, 211, 251, 193, 162, 211, 251, 193, 218, 211, 251, 193, 164, 211, 251, 193, 166, 211, 251, 193, 220, 211, 251, 193, 222, 211, 251, 193, 224, 211, 251, 193, 226, 211, 251, 193, 254, 211, 251, 193, 2, 212, 251, 193, 34, 212, 251, 193, 60, 212, 251, 193, 64, 212, 251, 193, 94, 212, 251, 193, 140, 212, 251, 193, 192, 212, 251, 193, 230, 212, 251, 193, 232, 212, 251, 193, 234, 212, 251, 193, 236, 212, 251, 193, 244, 212, 251, 193, 246, 212, 251, 193, 248, 212, 251, 193, 250, 212, 251, 193, 252, 212, 251, 193, 26, 213, 251, 193, 28, 213, 251, 193, 30, 213, 251, 193, 214, 211, 251, 193, 252, 211, 251, 193, 32, 212, 251, 193, 180, 212, 251, 193, 204, 212, 251, 193, 222, 212, 251, 193, 238, 212, 251, 193, 20, 213, 251, 193, 82, 213, 251, 193, 152, 213, 251, 193, 252, 213, 251, 193, 254, 213, 251, 193, 2, 214, 251, 193, 4, 214, 251, 193, 6, 214, 251, 193, 8, 214, 251, 193, 10, 214, 251, 193, 146, 214, 251, 193, 148, 214, 251, 193, 150, 214, 251, 193, 152, 214, 251, 193, 154, 214, 251, 193, 156, 214, 251, 193, 158, 214, 251, 193, 160, 214, 251, 193, 52, 215, 251, 193, 54, 215, 251, 193, 56, 215, 251, 193, 58, 215, 251, 193, 60, 215, 251, 193, 62, 215, 251, 193, 64, 215, 251, 193, 8, 212, 251, 193, 10, 212, 251, 193, 12, 212, 251, 193, 14, 212, 251, 193, 16, 212, 251, 193, 18, 212, 251, 193, 20, 212, 251, 193, 22, 212, 251, 193, 24, 212, 251, 193, 26, 212, 251, 193, 28, 212, 251, 193, 30, 212, 251, 193, 40, 212, 251, 193, 42, 212, 251, 193, 44, 212, 251, 193, 46, 212, 251, 193, 48, 212, 251, 193, 50, 212, 251, 193, 52, 212, 251, 193, 54, 212, 251, 193, 58, 212, 251, 193, 66, 212, 251, 193, 70, 212, 251, 193, 78, 212, 251, 193, 80, 212, 251, 193, 82, 212, 251, 193, 84, 212, 251, 193, 86, 212, 251, 193, 88, 212, 251, 193, 90, 212, 251, 193, 96, 212, 251, 193, 98, 212, 251, 193, 74, 212, 251, 193, 126, 212, 251, 193, 172, 212, 251, 193, 242, 212, 251, 193, 86, 213, 251, 193, 88, 213, 251, 193, 90, 213, 251, 193, 164, 213, 251, 193, 28, 214, 251, 193, 170, 214, 251, 193, 172, 214, 251, 193, 80, 215, 251, 193, 82, 215, 251, 193, 84, 215, 251, 193, 86, 215, 251, 193, 212, 215, 251, 193, 214, 215, 251, 193, 64, 216, 251, 193, 160, 216, 251, 193, 162, 216, 251, 193, 10, 217, 251, 193, 12, 217, 251, 193, 66, 217, 251, 193, 68, 217, 251, 193, 146, 217, 251, 193, 186, 217, 251, 193, 52, 218, 251, 193, 128, 218, 251, 193, 158, 218, 251, 193, 108, 219, 251, 193, 110, 219, 251, 193, 186, 219, 251, 193, 194, 212, 251, 193, 196, 212, 251, 193, 198, 212, 251, 193, 200, 212, 251, 193, 202, 212, 251, 193, 210, 212, 251, 193, 212, 212, 251, 193, 214, 212, 251, 193, 216, 212, 251, 193, 218, 212, 251, 193, 226, 212, 251, 193, 4, 213, 251, 193, 6, 213, 251, 193, 8, 213, 251, 193, 10, 213, 251, 193, 12, 213, 251, 193, 14, 213, 251, 193, 16, 213, 251, 193, 18, 213, 251, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 193, 228, 212, 251, 193, 240, 212, 251, 193, 22, 213, 251, 193, 24, 213, 251, 193, 84, 213, 251, 193, 154, 213, 251, 193, 156, 213, 251, 193, 158, 213, 251, 193, 160, 213, 251, 193, 162, 213, 251, 193, 12, 214, 251, 193, 14, 214, 251, 193, 16, 214, 251, 193, 18, 214, 251, 193, 20, 214, 251, 193, 22, 214, 251, 193, 24, 214, 251, 193, 26, 214, 251, 193, 162, 214, 251, 193, 164, 214, 251, 193, 166, 214, 251, 193, 168, 214, 251, 193, 68, 215, 251, 193, 70, 215, 251, 193, 72, 215, 251, 193, 74, 215, 251, 193, 76, 215, 251, 193, 78, 215, 251, 193, 198, 215, 251, 193, 200, 215, 251, 193, 202, 215, 251, 193, 204, 215, 251, 193, 2, 213, 251, 193, 32, 213, 251, 193, 34, 213, 251, 193, 104, 213, 251, 193, 106, 213, 251, 193, 108, 213, 251, 193, 110, 213, 251, 193, 180, 213, 251, 193, 182, 213, 251, 193, 184, 213, 251, 193, 62, 214, 251, 193, 64, 214, 251, 193, 66, 214, 251, 193, 68, 214, 251, 193, 70, 214, 251, 193, 206, 214, 251, 193, 208, 214, 251, 193, 210, 214, 251, 193, 212, 214, 251, 193, 214, 214, 251, 193, 216, 214, 251, 193, 218, 214, 251, 193, 220, 214, 251, 193, 222, 214, 251, 193, 224, 214, 251, 193, 226, 214, 251, 193, 228, 214, 251, 193, 98, 215, 251, 193, 100, 215, 251, 193, 102, 215, 251, 193, 104, 215, 251, 193, 106, 215, 251, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 206, 96, 32, 0, 193, 112, 213, 251, 193, 114, 213, 251, 193, 116, 213, 251, 193, 118, 213, 251, 193, 120, 213, 251, 193, 122, 213, 251, 193, 124, 213, 251, 193, 126, 213, 251, 193, 128, 213, 251, 193, 130, 213, 251, 193, 132, 213, 251, 193, 134, 213, 251, 193, 136, 213, 251, 193, 138, 213, 251, 193, 140, 213, 251, 193, 142, 213, 251, 193, 144, 213, 251, 193, 146, 213, 251, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 193, 92, 213, 251, 193, 94, 213, 251, 193, 96, 213, 251, 193, 98, 213, 251, 193, 100, 213, 251, 193, 102, 213, 251, 193, 166, 213, 251, 193, 168, 213, 251, 193, 170, 213, 251, 193, 172, 213, 251, 193, 174, 213, 251, 193, 176, 213, 251, 193, 178, 213, 251, 193, 32, 214, 251, 193, 34, 214, 251, 193, 36, 214, 251, 193, 38, 214, 251, 193, 40, 214, 251, 193, 42, 214, 251, 193, 44, 214, 251, 193, 46, 214, 251, 193, 48, 214, 251, 193, 50, 214, 251, 193, 52, 214, 251, 193, 54, 214, 251, 193, 56, 214, 251, 193, 58, 214, 251, 193, 60, 214, 251, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 128, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 193, 252, 217, 251, 193, 2, 218, 251, 193, 10, 218, 251, 193, 18, 218, 251, 193, 20, 218, 251, 193, 22, 218, 251, 193, 28, 218, 251, 193, 30, 218, 251, 193, 32, 218, 251, 193, 34, 218, 251, 193, 60, 218, 251, 193, 246, 213, 251, 193, 62, 218, 251, 193, 64, 218, 251, 193, 78, 218, 251, 193, 80, 218, 251, 193, 82, 218, 251, 193, 84, 218, 251, 193, 86, 218, 251, 193, 88, 218, 251, 193, 96, 218, 251, 193, 98, 218, 251, 193, 100, 218, 251, 193, 114, 218, 251, 193, 116, 218, 251, 193, 118, 218, 251, 193, 90, 218, 251, 193, 134, 216, 251, 193, 130, 218, 251, 193, 132, 218, 251, 193, 134, 218, 251, 193, 136, 218, 251, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 206, 160, 32, 0, 193, 230, 214, 251, 193, 232, 214, 251, 193, 138, 214, 251, 193, 234, 214, 251, 193, 236, 214, 251, 193, 238, 214, 251, 193, 240, 214, 251, 193, 242, 214, 251, 193, 244, 214, 251, 193, 246, 214, 251, 193, 248, 214, 251, 193, 250, 214, 251, 193, 252, 214, 251, 193, 254, 214, 251, 193, 2, 215, 251, 193, 4, 215, 251, 193, 6, 215, 251, 193, 8, 215, 251, 193, 10, 215, 251, 193, 12, 215, 251, 193, 14, 215, 251, 193, 16, 215, 251, 193, 18, 215, 251, 193, 20, 215, 251, 193, 22, 215, 251, 193, 24, 215, 251, 193, 26, 215, 251, 193, 28, 215, 251, 193, 30, 215, 251, 193, 140, 214, 251, 193, 32, 215, 251, 193, 34, 215, 251, 193, 36, 215, 251, 193, 38, 215, 251, 193, 40, 215, 251, 193, 42, 215, 251, 193, 44, 215, 251, 193, 108, 215, 251, 193, 126, 12, 252, 193, 128, 12, 252, 193, 190, 12, 252, 193, 192, 12, 252, 193, 22, 13, 252, 193, 122, 13, 252, 193, 230, 13, 252, 193, 104, 14, 252, 193, 222, 14, 252, 193, 224, 14, 252, 193, 144, 214, 251, 193, 24, 15, 252, 193, 64, 15, 252, 193, 66, 15, 252, 193, 68, 15, 252, 193, 120, 15, 252, 193, 186, 15, 252, 193, 14, 16, 252, 193, 16, 16, 252, 193, 94, 16, 252, 193, 164, 16, 252, 193, 252, 16, 252, 193, 26, 17, 252, 193, 108, 17, 252, 193, 110, 17, 252, 193, 206, 17, 252, 193, 208, 17, 252, 193, 38, 18, 252, 193, 40, 18, 252, 193, 222, 18, 252, 193, 224, 18, 252, 193, 72, 19, 252, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 206, 192, 32, 0, 193, 88, 215, 251, 193, 90, 215, 251, 193, 204, 214, 251, 193, 92, 215, 251, 193, 94, 215, 251, 193, 96, 215, 251, 193, 216, 215, 251, 193, 218, 215, 251, 193, 220, 215, 251, 193, 66, 216, 251, 193, 68, 216, 251, 193, 70, 216, 251, 193, 72, 216, 251, 193, 74, 216, 251, 193, 76, 216, 251, 193, 78, 216, 251, 193, 80, 216, 251, 193, 164, 216, 251, 193, 166, 216, 251, 193, 168, 216, 251, 193, 170, 216, 251, 193, 174, 215, 251, 193, 176, 215, 251, 193, 178, 215, 251, 193, 46, 215, 251, 193, 180, 215, 251, 193, 150, 117, 252, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 193, 66, 215, 251, 193, 184, 215, 251, 193, 186, 215, 251, 193, 188, 215, 251, 193, 190, 215, 251, 193, 192, 215, 251, 193, 194, 215, 251, 193, 196, 215, 251, 193, 44, 216, 251, 193, 46, 216, 251, 193, 48, 216, 251, 193, 50, 216, 251, 193, 52, 216, 251, 193, 144, 216, 251, 193, 146, 216, 251, 193, 148, 216, 251, 193, 150, 216, 251, 193, 250, 216, 251, 193, 252, 216, 251, 193, 254, 216, 251, 193, 2, 217, 251, 193, 4, 217, 251, 193, 6, 217, 251, 193, 8, 217, 251, 193, 56, 217, 251, 193, 58, 217, 251, 193, 60, 217, 251, 193, 106, 217, 251, 193, 108, 217, 251, 193, 110, 217, 251, 193, 136, 217, 251, 193, 138, 217, 251, 193, 110, 215, 251, 193, 112, 215, 251, 193, 114, 215, 251, 193, 116, 215, 251, 193, 118, 215, 251, 193, 120, 215, 251, 193, 122, 215, 251, 193, 124, 215, 251, 193, 126, 215, 251, 193, 128, 215, 251, 193, 130, 215, 251, 193, 132, 215, 251, 193, 134, 215, 251, 193, 136, 215, 251, 193, 138, 215, 251, 193, 140, 215, 251, 193, 142, 215, 251, 193, 144, 215, 251, 193, 146, 215, 251, 193, 148, 215, 251, 193, 150, 215, 251, 193, 152, 215, 251, 193, 154, 215, 251, 193, 156, 215, 251, 193, 158, 215, 251, 193, 160, 215, 251, 193, 162, 215, 251, 193, 164, 215, 251, 193, 166, 215, 251, 193, 168, 215, 251, 193, 170, 215, 251, 193, 172, 215, 251, 193, 206, 215, 251, 193, 208, 215, 251, 193, 210, 215, 251, 193, 54, 216, 251, 193, 56, 216, 251, 193, 58, 216, 251, 193, 60, 216, 251, 193, 62, 216, 251, 193, 152, 216, 251, 193, 154, 216, 251, 193, 156, 216, 251, 193, 158, 216, 251, 193, 62, 217, 251, 193, 64, 217, 251, 193, 112, 217, 251, 193, 114, 217, 251, 193, 208, 217, 251, 193, 210, 217, 251, 193, 218, 217, 251, 193, 244, 217, 251, 193, 14, 218, 251, 193, 42, 218, 251, 193, 44, 218, 251, 193, 46, 218, 251, 193, 48, 218, 251, 193, 50, 218, 251, 193, 150, 218, 251, 193, 152, 218, 251, 193, 42, 219, 251, 193, 48, 219, 251, 193, 72, 219, 251, 193, 80, 219, 251, 193, 222, 215, 251, 193, 224, 215, 251, 193, 226, 215, 251, 193, 228, 215, 251, 193, 82, 216, 251, 193, 84, 216, 251, 193, 86, 216, 251, 193, 172, 216, 251, 193, 174, 216, 251, 193, 176, 216, 251, 193, 178, 216, 251, 193, 74, 217, 251, 193, 76, 217, 251, 193, 152, 217, 251, 193, 188, 217, 251, 193, 8, 218, 251, 193, 58, 218, 251, 193, 112, 218, 251, 193, 182, 218, 251, 193, 184, 218, 251, 193, 190, 218, 251, 193, 198, 218, 251, 193, 210, 218, 251, 193, 228, 218, 251, 193, 244, 218, 251, 193, 30, 219, 251, 193, 52, 219, 251, 193, 70, 219, 251, 193, 116, 219, 251, 193, 118, 219, 251, 193, 128, 219, 251, 193, 130, 219, 251, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 206, 224, 32, 0, 193, 24, 138, 252, 193, 88, 216, 251, 193, 90, 216, 251, 193, 92, 216, 251, 193, 94, 216, 251, 193, 96, 216, 251, 193, 98, 216, 251, 193, 100, 216, 251, 193, 102, 216, 251, 193, 104, 216, 251, 193, 106, 216, 251, 193, 108, 216, 251, 193, 110, 216, 251, 193, 112, 216, 251, 193, 114, 216, 251, 193, 116, 216, 251, 193, 118, 216, 251, 193, 120, 216, 251, 193, 122, 216, 251, 193, 124, 216, 251, 193, 126, 216, 251, 193, 128, 216, 251, 193, 130, 216, 251, 193, 132, 216, 251, 193, 180, 216, 251, 193, 182, 216, 251, 193, 184, 216, 251, 193, 186, 216, 251, 193, 188, 216, 251, 193, 190, 216, 251, 193, 192, 216, 251, 193, 194, 216, 251, 193, 196, 216, 251, 193, 198, 216, 251, 193, 200, 216, 251, 193, 202, 216, 251, 193, 204, 216, 251, 193, 206, 216, 251, 193, 208, 216, 251, 193, 210, 216, 251, 193, 212, 216, 251, 193, 214, 216, 251, 193, 216, 216, 251, 193, 218, 216, 251, 193, 220, 216, 251, 193, 222, 216, 251, 193, 224, 216, 251, 193, 226, 216, 251, 193, 228, 216, 251, 193, 230, 216, 251, 193, 232, 216, 251, 193, 234, 216, 251, 193, 236, 216, 251, 193, 238, 216, 251, 193, 240, 216, 251, 193, 242, 216, 251, 193, 20, 217, 251, 193, 22, 217, 251, 193, 14, 217, 251, 193, 16, 217, 251, 193, 18, 217, 251, 193, 70, 217, 251, 193, 116, 217, 251, 193, 148, 217, 251, 193, 150, 217, 251, 193, 254, 217, 251, 193, 6, 218, 251, 193, 16, 218, 251, 193, 54, 218, 251, 193, 56, 218, 251, 193, 70, 218, 251, 193, 72, 218, 251, 193, 74, 218, 251, 193, 76, 218, 251, 193, 92, 218, 251, 193, 94, 218, 251, 193, 110, 218, 251, 193, 142, 218, 251, 193, 160, 218, 251, 193, 168, 218, 251, 193, 174, 218, 251, 193, 178, 218, 251, 193, 186, 218, 251, 193, 206, 218, 251, 193, 208, 218, 251, 193, 224, 218, 251, 193, 226, 218, 251, 193, 242, 218, 251, 193, 4, 219, 251, 193, 6, 219, 251, 193, 24, 217, 251, 193, 26, 217, 251, 193, 28, 217, 251, 193, 30, 217, 251, 193, 78, 217, 251, 193, 32, 217, 251, 193, 34, 217, 251, 193, 36, 217, 251, 193, 38, 217, 251, 193, 40, 217, 251, 193, 42, 217, 251, 193, 44, 217, 251, 193, 46, 217, 251, 193, 48, 217, 251, 193, 50, 217, 251, 193, 52, 217, 251, 193, 80, 217, 251, 193, 82, 217, 251, 193, 84, 217, 251, 193, 86, 217, 251, 193, 88, 217, 251, 193, 90, 217, 251, 193, 92, 217, 251, 193, 94, 217, 251, 193, 96, 217, 251, 193, 98, 217, 251, 193, 100, 217, 251, 193, 118, 217, 251, 193, 120, 217, 251, 193, 122, 217, 251, 193, 124, 217, 251, 193, 126, 217, 251, 193, 52, 52, 253, 193, 72, 217, 251, 193, 54, 52, 253, 193, 56, 52, 253, 193, 58, 52, 253, 193, 60, 52, 253, 193, 144, 52, 253, 193, 146, 52, 253, 193, 148, 52, 253, 193, 150, 52, 253, 193, 152, 52, 253, 193, 154, 52, 253, 193, 156, 52, 253, 193, 242, 52, 253, 193, 244, 52, 253, 193, 246, 52, 253, 193, 34, 53, 253, 193, 36, 53, 253, 193, 38, 53, 253, 193, 40, 53, 253, 193, 42, 53, 253, 193, 86, 53, 253, 193, 88, 53, 253, 193, 90, 53, 253, 193, 92, 53, 253, 193, 136, 53, 253, 193, 138, 53, 253, 193, 168, 53, 253, 193, 178, 53, 253, 193, 202, 53, 253, 193, 204, 53, 253, 193, 206, 53, 253, 193, 128, 217, 251, 193, 130, 217, 251, 193, 132, 217, 251, 193, 154, 217, 251, 193, 156, 217, 251, 193, 158, 217, 251, 193, 160, 217, 251, 193, 162, 217, 251, 193, 164, 217, 251, 193, 166, 217, 251, 193, 168, 217, 251, 193, 170, 217, 251, 193, 172, 217, 251, 193, 174, 217, 251, 193, 190, 217, 251, 193, 192, 217, 251, 193, 194, 217, 251, 193, 196, 217, 251, 193, 198, 217, 251, 193, 212, 217, 251, 193, 214, 217, 251, 193, 220, 217, 251, 193, 222, 217, 251, 193, 224, 217, 251, 193, 226, 217, 251, 193, 228, 217, 251, 193, 230, 217, 251, 193, 240, 217, 251, 193, 242, 217, 251, 193, 246, 217, 251, 193, 248, 217, 251, 193, 250, 217, 251, 193, 160, 164, 253, 193, 168, 164, 253, 193, 170, 164, 253, 193, 172, 164, 253, 193, 174, 164, 253, 193, 176, 164, 253, 193, 178, 164, 253, 193, 180, 164, 253, 193, 194, 164, 253, 193, 196, 164, 253, 193, 208, 164, 253, 193, 210, 164, 253, 193, 212, 164, 253, 193, 214, 164, 253, 193, 134, 217, 251, 193, 222, 164, 253, 193, 176, 217, 251, 193, 224, 164, 253, 193, 230, 164, 253, 193, 234, 164, 253, 193, 236, 164, 253, 193, 240, 164, 253, 193, 242, 164, 253, 193, 244, 164, 253, 193, 248, 164, 253, 193, 250, 164, 253, 193, 254, 164, 253, 193, 4, 165, 253, 193, 6, 165, 253, 193, 8, 165, 253, 193, 22, 165, 253, 193, 24, 165, 253, 193, 140, 217, 251, 193, 142, 217, 251, 193, 144, 217, 251, 193, 178, 217, 251, 193, 180, 217, 251, 193, 182, 217, 251, 193, 184, 217, 251, 193, 204, 217, 251, 193, 206, 217, 251, 193, 216, 217, 251, 193, 234, 217, 251, 193, 236, 217, 251, 193, 238, 217, 251, 193, 4, 218, 251, 193, 38, 218, 251, 193, 40, 218, 251, 193, 68, 218, 251, 193, 108, 218, 251, 193, 166, 218, 251, 193, 222, 218, 251, 193, 240, 218, 251, 193, 2, 219, 251, 193, 40, 219, 251, 193, 46, 219, 251, 193, 60, 219, 251, 193, 148, 219, 251, 193, 150, 219, 251, 193, 182, 219, 251, 193, 184, 219, 251, 193, 12, 220, 251, 193, 54, 220, 251, 193, 60, 220, 251, 193, 12, 218, 251, 193, 24, 218, 251, 193, 26, 218, 251, 193, 36, 218, 251, 193, 66, 218, 251, 193, 102, 218, 251, 193, 104, 218, 251, 193, 106, 218, 251, 193, 122, 218, 251, 193, 124, 218, 251, 193, 126, 218, 251, 193, 204, 218, 251, 193, 220, 218, 251, 193, 28, 219, 251, 193, 38, 219, 251, 193, 58, 219, 251, 193, 68, 219, 251, 193, 104, 219, 251, 193, 100, 220, 251, 193, 102, 220, 251, 193, 138, 220, 251, 193, 206, 220, 251, 193, 208, 220, 251, 193, 210, 220, 251, 193, 22, 221, 251, 193, 68, 221, 251, 193, 70, 221, 251, 193, 72, 221, 251, 193, 74, 221, 251, 193, 112, 221, 251, 193, 144, 221, 251, 193, 166, 221, 251, 193, 110, 154, 253, 193, 112, 154, 253, 193, 126, 154, 253, 193, 128, 154, 253, 193, 130, 154, 253, 193, 132, 154, 253, 193, 134, 154, 253, 193, 136, 154, 253, 193, 138, 154, 253, 193, 140, 154, 253, 193, 142, 154, 253, 193, 144, 154, 253, 193, 146, 154, 253, 193, 148, 154, 253, 193, 150, 154, 253, 193, 152, 154, 253, 193, 120, 218, 251, 193, 154, 154, 253, 193, 156, 154, 253, 193, 158, 154, 253, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 193, 138, 218, 251, 193, 140, 218, 251, 193, 144, 218, 251, 193, 146, 218, 251, 193, 148, 218, 251, 193, 154, 218, 251, 193, 156, 218, 251, 193, 162, 218, 251, 193, 164, 218, 251, 193, 170, 218, 251, 193, 172, 218, 251, 193, 176, 218, 251, 193, 180, 218, 251, 193, 188, 218, 251, 193, 192, 218, 251, 193, 194, 218, 251, 193, 196, 218, 251, 193, 200, 218, 251, 193, 202, 218, 251, 193, 212, 218, 251, 193, 214, 218, 251, 193, 216, 218, 251, 193, 218, 218, 251, 193, 230, 218, 251, 193, 232, 218, 251, 193, 234, 218, 251, 193, 236, 218, 251, 193, 238, 218, 251, 193, 246, 218, 251, 193, 248, 218, 251, 193, 250, 218, 251, 193, 252, 218, 251, 193, 254, 218, 251, 193, 8, 219, 251, 193, 20, 219, 251, 193, 10, 219, 251, 193, 12, 219, 251, 193, 14, 219, 251, 193, 22, 219, 251, 193, 24, 219, 251, 193, 26, 219, 251, 193, 32, 219, 251, 193, 34, 219, 251, 193, 36, 219, 251, 193, 54, 219, 251, 193, 56, 219, 251, 193, 62, 219, 251, 193, 64, 219, 251, 193, 66, 219, 251, 193, 76, 219, 251, 193, 82, 219, 251, 193, 84, 219, 251, 193, 94, 219, 251, 193, 96, 219, 251, 193, 98, 219, 251, 193, 100, 219, 251, 193, 102, 219, 251, 193, 120, 219, 251, 193, 122, 219, 251, 193, 132, 219, 251, 193, 134, 219, 251, 193, 136, 219, 251, 193, 142, 219, 251, 193, 144, 219, 251, 193, 16, 219, 251, 193, 18, 219, 251, 193, 44, 219, 251, 193, 50, 219, 251, 193, 74, 219, 251, 193, 78, 219, 251, 193, 86, 219, 251, 193, 88, 219, 251, 193, 90, 219, 251, 193, 92, 219, 251, 193, 112, 219, 251, 193, 114, 219, 251, 193, 138, 219, 251, 193, 140, 219, 251, 193, 152, 219, 251, 193, 154, 219, 251, 193, 188, 219, 251, 193, 190, 219, 251, 193, 216, 219, 251, 193, 234, 219, 251, 193, 250, 219, 251, 193, 18, 220, 251, 193, 20, 220, 251, 193, 34, 220, 251, 193, 64, 220, 251, 193, 68, 220, 251, 193, 86, 220, 251, 193, 88, 220, 251, 193, 90, 220, 251, 193, 110, 220, 251, 193, 112, 220, 251, 193, 114, 220, 251, 193, 106, 219, 251, 193, 124, 219, 251, 193, 126, 219, 251, 193, 210, 219, 251, 193, 212, 219, 251, 193, 230, 219, 251, 193, 32, 220, 251, 193, 70, 220, 251, 193, 76, 220, 251, 193, 78, 220, 251, 193, 80, 220, 251, 193, 106, 220, 251, 193, 108, 220, 251, 193, 140, 220, 251, 193, 142, 220, 251, 193, 144, 220, 251, 193, 146, 220, 251, 193, 148, 220, 251, 193, 150, 220, 251, 193, 152, 220, 251, 193, 154, 220, 251, 193, 224, 220, 251, 193, 226, 220, 251, 193, 228, 220, 251, 193, 230, 220, 251, 193, 232, 220, 251, 193, 30, 221, 251, 193, 80, 221, 251, 193, 82, 221, 251, 193, 84, 221, 251, 193, 86, 221, 251, 193, 122, 221, 251, 193, 146, 219, 251, 193, 158, 219, 251, 193, 160, 219, 251, 193, 162, 219, 251, 193, 164, 219, 251, 193, 166, 219, 251, 193, 168, 219, 251, 193, 170, 219, 251, 193, 172, 219, 251, 193, 174, 219, 251, 193, 176, 219, 251, 193, 208, 136, 252, 193, 178, 219, 251, 193, 180, 219, 251, 193, 192, 219, 251, 193, 194, 219, 251, 193, 196, 219, 251, 193, 198, 219, 251, 193, 200, 219, 251, 193, 202, 219, 251, 193, 204, 219, 251, 193, 206, 219, 251, 193, 220, 219, 251, 193, 222, 219, 251, 193, 224, 219, 251, 193, 226, 219, 251, 193, 228, 219, 251, 193, 238, 219, 251, 193, 240, 219, 251, 193, 242, 219, 251, 193, 244, 219, 251, 193, 246, 219, 251, 193, 156, 219, 251, 193, 218, 219, 251, 193, 236, 219, 251, 193, 72, 220, 251, 193, 92, 220, 251, 193, 172, 220, 251, 193, 118, 220, 251, 193, 174, 220, 251, 193, 176, 220, 251, 193, 178, 220, 251, 193, 252, 220, 251, 193, 254, 220, 251, 193, 2, 221, 251, 193, 42, 221, 251, 193, 156, 221, 251, 193, 158, 221, 251, 193, 116, 222, 251, 193, 118, 222, 251, 193, 164, 222, 251, 193, 166, 222, 251, 193, 168, 222, 251, 193, 170, 222, 251, 193, 172, 222, 251, 193, 174, 222, 251, 193, 46, 223, 251, 193, 48, 223, 251, 193, 50, 223, 251, 206, 128, 33, 0, 206, 128, 33, 0, 206, 128, 33, 0, 206, 128, 33, 0, 206, 128, 33, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 193, 38, 117, 252, 193, 224, 117, 252, 193, 226, 117, 252, 193, 228, 117, 252, 193, 230, 117, 252, 193, 232, 117, 252, 193, 234, 117, 252, 193, 236, 117, 252, 193, 238, 117, 252, 193, 208, 219, 251, 193, 240, 117, 252, 193, 242, 117, 252, 193, 244, 117, 252, 193, 246, 117, 252, 193, 248, 117, 252, 193, 250, 117, 252, 193, 252, 117, 252, 193, 254, 117, 252, 193, 2, 118, 252, 193, 4, 118, 252, 193, 6, 118, 252, 193, 214, 219, 251, 193, 232, 219, 251, 193, 248, 219, 251, 193, 14, 220, 251, 193, 16, 220, 251, 193, 46, 220, 251, 193, 66, 220, 251, 193, 156, 220, 251, 193, 234, 220, 251, 193, 32, 221, 251, 193, 34, 221, 251, 193, 36, 221, 251, 193, 88, 221, 251, 193, 90, 221, 251, 193, 126, 221, 251, 193, 154, 221, 251, 193, 48, 222, 251, 193, 142, 222, 251, 193, 144, 222, 251, 193, 232, 222, 251, 193, 160, 223, 251, 193, 162, 223, 251, 193, 164, 223, 251, 193, 166, 223, 251, 193, 118, 224, 251, 193, 120, 224, 251, 193, 122, 224, 251, 193, 124, 224, 251, 193, 126, 224, 251, 193, 128, 224, 251, 193, 130, 224, 251, 193, 90, 225, 251, 193, 252, 219, 251, 193, 254, 219, 251, 193, 2, 220, 251, 193, 4, 220, 251, 193, 6, 220, 251, 193, 8, 220, 251, 193, 10, 220, 251, 193, 22, 220, 251, 193, 24, 220, 251, 193, 26, 220, 251, 193, 28, 220, 251, 193, 30, 220, 251, 193, 36, 220, 251, 193, 38, 220, 251, 193, 40, 220, 251, 193, 42, 220, 251, 193, 44, 220, 251, 193, 48, 220, 251, 193, 50, 220, 251, 193, 52, 220, 251, 193, 56, 220, 251, 193, 58, 220, 251, 193, 62, 220, 251, 193, 74, 220, 251, 193, 82, 220, 251, 193, 84, 220, 251, 193, 94, 220, 251, 193, 96, 220, 251, 193, 98, 220, 251, 193, 120, 220, 251, 193, 122, 220, 251, 193, 124, 220, 251, 193, 104, 220, 251, 193, 212, 220, 251, 193, 214, 220, 251, 193, 216, 220, 251, 193, 218, 220, 251, 193, 220, 220, 251, 193, 222, 220, 251, 193, 26, 221, 251, 193, 28, 221, 251, 193, 76, 221, 251, 193, 78, 221, 251, 193, 116, 221, 251, 193, 118, 221, 251, 193, 120, 221, 251, 193, 146, 221, 251, 193, 148, 221, 251, 193, 150, 221, 251, 193, 170, 221, 251, 193, 172, 221, 251, 193, 200, 221, 251, 193, 202, 221, 251, 193, 204, 221, 251, 193, 206, 221, 251, 193, 214, 221, 251, 193, 232, 221, 251, 193, 234, 221, 251, 193, 236, 221, 251, 193, 238, 221, 251, 193, 2, 222, 251, 193, 4, 222, 251, 193, 14, 222, 251, 193, 16, 222, 251, 193, 116, 220, 251, 193, 158, 220, 251, 193, 160, 220, 251, 193, 162, 220, 251, 193, 164, 220, 251, 193, 166, 220, 251, 193, 168, 220, 251, 193, 236, 220, 251, 193, 238, 220, 251, 193, 240, 220, 251, 193, 170, 220, 251, 193, 242, 220, 251, 193, 244, 220, 251, 193, 246, 220, 251, 193, 248, 220, 251, 193, 250, 220, 251, 193, 38, 221, 251, 193, 40, 221, 251, 193, 92, 221, 251, 193, 94, 221, 251, 193, 96, 221, 251, 193, 98, 221, 251, 193, 128, 221, 251, 193, 130, 221, 251, 193, 180, 221, 251, 193, 182, 221, 251, 193, 210, 221, 251, 193, 218, 221, 251, 193, 244, 221, 251, 193, 246, 221, 251, 193, 248, 221, 251, 193, 8, 222, 251, 193, 126, 220, 251, 193, 128, 220, 251, 193, 130, 220, 251, 193, 132, 220, 251, 193, 134, 220, 251, 193, 136, 220, 251, 193, 180, 220, 251, 193, 182, 220, 251, 193, 184, 220, 251, 193, 186, 220, 251, 193, 188, 220, 251, 193, 190, 220, 251, 193, 192, 220, 251, 193, 194, 220, 251, 193, 196, 220, 251, 193, 198, 220, 251, 193, 200, 220, 251, 193, 202, 220, 251, 193, 204, 220, 251, 193, 4, 221, 251, 193, 6, 221, 251, 193, 8, 221, 251, 193, 10, 221, 251, 193, 12, 221, 251, 193, 14, 221, 251, 193, 16, 221, 251, 193, 18, 221, 251, 193, 20, 221, 251, 206, 0, 33, 0, 206, 0, 33, 0, 206, 0, 33, 0, 206, 0, 33, 0, 193, 24, 221, 251, 193, 114, 221, 251, 193, 82, 224, 251, 193, 84, 224, 251, 193, 32, 225, 251, 193, 222, 233, 251, 193, 80, 234, 251, 193, 82, 234, 251, 193, 120, 237, 251, 193, 142, 240, 251, 193, 144, 240, 251, 193, 168, 249, 251, 193, 170, 249, 251, 193, 212, 251, 251, 193, 214, 251, 251, 193, 216, 251, 251, 193, 36, 254, 251, 193, 182, 255, 251, 193, 230, 2, 252, 193, 4, 11, 252, 193, 26, 14, 252, 193, 242, 20, 252, 193, 246, 20, 252, 193, 94, 28, 252, 193, 252, 30, 252, 193, 254, 35, 252, 193, 2, 36, 252, 193, 4, 36, 252, 193, 178, 40, 252, 193, 152, 42, 252, 193, 18, 46, 252, 193, 200, 51, 252, 206, 0, 33, 0, 206, 0, 33, 0, 206, 0, 33, 0, 206, 0, 33, 0, 206, 0, 33, 0, 206, 0, 33, 0, 206, 0, 33, 0, 206, 0, 33, 0, 193, 100, 221, 251, 193, 102, 221, 251, 193, 104, 221, 251, 193, 106, 221, 251, 193, 108, 221, 251, 193, 110, 221, 251, 193, 132, 221, 251, 193, 134, 221, 251, 193, 136, 221, 251, 193, 138, 221, 251, 193, 140, 221, 251, 193, 142, 221, 251, 193, 160, 221, 251, 193, 162, 221, 251, 193, 164, 221, 251, 193, 184, 221, 251, 193, 186, 221, 251, 193, 188, 221, 251, 193, 190, 221, 251, 193, 192, 221, 251, 193, 194, 221, 251, 193, 212, 221, 251, 193, 220, 221, 251, 193, 222, 221, 251, 193, 124, 221, 251, 193, 152, 221, 251, 193, 174, 221, 251, 193, 176, 221, 251, 193, 178, 221, 251, 193, 208, 221, 251, 193, 216, 221, 251, 193, 240, 221, 251, 193, 242, 221, 251, 193, 6, 222, 251, 193, 84, 222, 251, 193, 98, 222, 251, 193, 136, 222, 251, 193, 138, 222, 251, 193, 140, 222, 251, 193, 226, 222, 251, 193, 228, 222, 251, 193, 230, 222, 251, 193, 130, 223, 251, 193, 132, 223, 251, 193, 134, 223, 251, 193, 136, 223, 251, 193, 138, 223, 251, 193, 140, 223, 251, 193, 142, 223, 251, 193, 144, 223, 251, 193, 146, 223, 251, 193, 148, 223, 251, 193, 150, 223, 251, 193, 152, 223, 251, 193, 154, 223, 251, 193, 156, 223, 251, 193, 168, 221, 251, 193, 196, 221, 251, 193, 198, 221, 251, 193, 252, 221, 251, 193, 254, 221, 251, 193, 12, 222, 251, 193, 218, 222, 251, 193, 220, 222, 251, 193, 110, 223, 251, 193, 112, 223, 251, 193, 114, 223, 251, 193, 116, 223, 251, 193, 118, 223, 251, 193, 120, 223, 251, 193, 122, 223, 251, 193, 62, 224, 251, 193, 64, 224, 251, 193, 66, 224, 251, 193, 68, 224, 251, 193, 70, 224, 251, 193, 72, 224, 251, 193, 74, 224, 251, 193, 76, 224, 251, 193, 78, 224, 251, 193, 80, 224, 251, 206, 224, 33, 0, 206, 224, 33, 0, 206, 224, 33, 0, 206, 224, 33, 0, 206, 224, 33, 0, 206, 224, 33, 0, 206, 224, 33, 0, 193, 224, 221, 251, 193, 226, 221, 251, 193, 228, 221, 251, 193, 230, 221, 251, 193, 250, 221, 251, 193, 10, 222, 251, 193, 20, 222, 251, 193, 24, 222, 251, 193, 26, 222, 251, 193, 30, 222, 251, 193, 34, 222, 251, 193, 36, 222, 251, 193, 38, 222, 251, 193, 46, 222, 251, 193, 50, 222, 251, 193, 52, 222, 251, 193, 56, 222, 251, 193, 58, 222, 251, 193, 60, 222, 251, 193, 62, 222, 251, 193, 68, 222, 251, 193, 70, 222, 251, 193, 74, 222, 251, 193, 76, 222, 251, 193, 78, 222, 251, 193, 80, 222, 251, 193, 82, 222, 251, 193, 92, 222, 251, 193, 94, 222, 251, 193, 96, 222, 251, 193, 120, 222, 251, 193, 122, 222, 251, 193, 18, 222, 251, 193, 22, 222, 251, 193, 28, 222, 251, 193, 32, 222, 251, 193, 40, 222, 251, 193, 66, 222, 251, 193, 134, 222, 251, 193, 222, 222, 251, 193, 224, 222, 251, 193, 124, 223, 251, 193, 126, 223, 251, 193, 128, 223, 251, 193, 86, 224, 251, 193, 88, 224, 251, 193, 90, 224, 251, 193, 34, 225, 251, 193, 36, 225, 251, 193, 38, 225, 251, 193, 40, 225, 251, 193, 42, 225, 251, 193, 44, 225, 251, 193, 10, 226, 251, 193, 12, 226, 251, 193, 14, 226, 251, 193, 16, 226, 251, 193, 18, 226, 251, 193, 20, 226, 251, 193, 44, 227, 251, 193, 46, 227, 251, 193, 48, 227, 251, 193, 50, 227, 251, 193, 52, 227, 251, 193, 42, 222, 251, 193, 44, 222, 251, 193, 54, 222, 251, 193, 64, 222, 251, 193, 72, 222, 251, 193, 86, 222, 251, 193, 88, 222, 251, 193, 90, 222, 251, 193, 100, 222, 251, 193, 102, 222, 251, 193, 104, 222, 251, 193, 106, 222, 251, 193, 108, 222, 251, 193, 110, 222, 251, 193, 112, 222, 251, 193, 114, 222, 251, 193, 146, 222, 251, 193, 148, 222, 251, 193, 150, 222, 251, 193, 152, 222, 251, 193, 154, 222, 251, 193, 156, 222, 251, 193, 158, 222, 251, 193, 160, 222, 251, 193, 162, 222, 251, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 193, 124, 222, 251, 193, 126, 222, 251, 193, 128, 222, 251, 193, 130, 222, 251, 193, 176, 222, 251, 193, 132, 222, 251, 193, 178, 222, 251, 193, 180, 222, 251, 193, 182, 222, 251, 193, 184, 222, 251, 193, 186, 222, 251, 193, 188, 222, 251, 193, 190, 222, 251, 193, 192, 222, 251, 193, 194, 222, 251, 193, 196, 222, 251, 193, 198, 222, 251, 193, 200, 222, 251, 193, 202, 222, 251, 193, 204, 222, 251, 193, 206, 222, 251, 193, 208, 222, 251, 193, 210, 222, 251, 193, 212, 222, 251, 193, 214, 222, 251, 193, 216, 222, 251, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 32, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 64, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 193, 158, 223, 251, 193, 92, 224, 251, 193, 94, 224, 251, 193, 96, 224, 251, 193, 98, 224, 251, 193, 100, 224, 251, 193, 102, 224, 251, 193, 104, 224, 251, 193, 106, 224, 251, 193, 108, 224, 251, 193, 110, 224, 251, 193, 112, 224, 251, 193, 114, 224, 251, 193, 116, 224, 251, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 96, 33, 0, 206, 128, 33, 0, 206, 128, 33, 0, 206, 128, 33, 0, 206, 128, 33, 0, 206, 128, 33, 0, 193, 178, 224, 251, 193, 180, 224, 251, 193, 182, 224, 251, 193, 184, 224, 251, 193, 186, 224, 251, 193, 188, 224, 251, 193, 190, 224, 251, 193, 192, 224, 251, 193, 194, 224, 251, 193, 196, 224, 251, 193, 148, 225, 251, 193, 150, 225, 251, 193, 152, 225, 251, 193, 154, 225, 251, 193, 156, 225, 251, 193, 158, 225, 251, 193, 130, 226, 251, 193, 132, 226, 251, 193, 134, 226, 251, 193, 136, 226, 251, 193, 138, 226, 251, 193, 140, 226, 251, 193, 142, 226, 251, 193, 144, 226, 251, 193, 146, 226, 251, 193, 148, 226, 251, 193, 150, 226, 251, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 206, 160, 33, 0, 193, 128, 232, 251, 193, 46, 224, 251, 193, 48, 224, 251, 193, 50, 224, 251, 193, 52, 224, 251, 193, 54, 224, 251, 193, 56, 224, 251, 193, 58, 224, 251, 193, 60, 224, 251, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 193, 132, 224, 251, 193, 134, 224, 251, 193, 136, 224, 251, 193, 138, 224, 251, 193, 140, 224, 251, 193, 142, 224, 251, 193, 144, 224, 251, 193, 146, 224, 251, 193, 148, 224, 251, 193, 150, 224, 251, 193, 152, 224, 251, 193, 154, 224, 251, 193, 156, 224, 251, 193, 158, 224, 251, 193, 160, 224, 251, 193, 162, 224, 251, 193, 164, 224, 251, 193, 166, 224, 251, 193, 168, 224, 251, 193, 170, 224, 251, 193, 172, 224, 251, 193, 174, 224, 251, 193, 176, 224, 251, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 206, 192, 33, 0, 193, 160, 225, 251, 193, 162, 225, 251, 193, 164, 225, 251, 193, 166, 225, 251, 193, 168, 225, 251, 193, 170, 225, 251, 193, 172, 225, 251, 193, 174, 225, 251, 193, 176, 225, 251, 193, 178, 225, 251, 193, 180, 225, 251, 193, 182, 225, 251, 193, 184, 225, 251, 193, 186, 225, 251, 206, 224, 33, 0, 206, 224, 33, 0, 206, 224, 33, 0, 206, 224, 33, 0, 206, 224, 33, 0, 193, 2, 226, 251, 193, 4, 226, 251, 193, 6, 226, 251, 193, 8, 226, 251, 193, 32, 227, 251, 193, 34, 227, 251, 193, 36, 227, 251, 193, 38, 227, 251, 193, 40, 227, 251, 193, 42, 227, 251, 193, 10, 228, 251, 193, 12, 228, 251, 193, 14, 228, 251, 193, 16, 228, 251, 193, 18, 228, 251, 193, 20, 228, 251, 193, 22, 228, 251, 193, 24, 228, 251, 193, 26, 228, 251, 193, 224, 228, 251, 193, 226, 228, 251, 193, 228, 228, 251, 193, 230, 228, 251, 193, 232, 228, 251, 193, 234, 228, 251, 193, 148, 229, 251, 193, 150, 229, 251, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 206, 0, 34, 0, 193, 22, 226, 251, 193, 24, 226, 251, 193, 26, 226, 251, 193, 28, 226, 251, 193, 30, 226, 251, 193, 32, 226, 251, 193, 34, 226, 251, 193, 36, 226, 251, 193, 38, 226, 251, 193, 40, 226, 251, 193, 42, 226, 251, 193, 44, 226, 251, 193, 46, 226, 251, 193, 48, 226, 251, 193, 50, 226, 251, 193, 52, 226, 251, 193, 54, 226, 251, 193, 56, 226, 251, 193, 58, 226, 251, 193, 60, 226, 251, 193, 62, 226, 251, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 193, 92, 225, 251, 193, 94, 225, 251, 193, 96, 225, 251, 193, 98, 225, 251, 193, 100, 225, 251, 193, 102, 225, 251, 193, 104, 225, 251, 193, 64, 226, 251, 193, 66, 226, 251, 193, 68, 226, 251, 193, 70, 226, 251, 193, 72, 226, 251, 193, 74, 226, 251, 193, 88, 227, 251, 193, 90, 227, 251, 193, 92, 227, 251, 193, 94, 227, 251, 193, 96, 227, 251, 193, 98, 227, 251, 193, 100, 227, 251, 193, 102, 227, 251, 193, 104, 227, 251, 193, 96, 228, 251, 193, 98, 228, 251, 193, 100, 228, 251, 193, 102, 228, 251, 193, 104, 228, 251, 193, 26, 229, 251, 193, 28, 229, 251, 193, 30, 229, 251, 193, 32, 229, 251, 193, 34, 229, 251, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 32, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 193, 150, 231, 251, 193, 152, 231, 251, 193, 154, 231, 251, 193, 198, 231, 251, 193, 200, 231, 251, 193, 202, 231, 251, 193, 204, 231, 251, 193, 206, 231, 251, 193, 208, 231, 251, 193, 242, 231, 251, 193, 38, 232, 251, 193, 40, 232, 251, 193, 42, 232, 251, 193, 52, 232, 251, 193, 54, 232, 251, 193, 70, 232, 251, 193, 90, 232, 251, 193, 92, 232, 251, 193, 94, 232, 251, 193, 96, 232, 251, 193, 102, 232, 251, 193, 104, 232, 251, 193, 114, 232, 251, 193, 126, 232, 251, 193, 146, 232, 251, 193, 176, 232, 251, 193, 148, 232, 251, 193, 178, 232, 251, 193, 180, 232, 251, 193, 146, 225, 251, 193, 194, 232, 251, 193, 196, 232, 251, 193, 188, 225, 251, 193, 190, 225, 251, 193, 192, 225, 251, 193, 194, 225, 251, 193, 196, 225, 251, 193, 198, 225, 251, 193, 200, 225, 251, 193, 202, 225, 251, 193, 204, 225, 251, 193, 206, 225, 251, 193, 208, 225, 251, 193, 210, 225, 251, 193, 212, 225, 251, 193, 214, 225, 251, 193, 216, 225, 251, 193, 218, 225, 251, 193, 220, 225, 251, 193, 222, 225, 251, 193, 224, 225, 251, 193, 226, 225, 251, 193, 228, 225, 251, 193, 230, 225, 251, 193, 232, 225, 251, 193, 234, 225, 251, 193, 236, 225, 251, 193, 238, 225, 251, 193, 240, 225, 251, 193, 242, 225, 251, 193, 244, 225, 251, 193, 246, 225, 251, 193, 248, 225, 251, 193, 250, 225, 251, 193, 252, 225, 251, 193, 254, 225, 251, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 206, 64, 34, 0, 193, 106, 227, 251, 193, 108, 227, 251, 193, 110, 227, 251, 193, 112, 227, 251, 193, 114, 227, 251, 193, 116, 227, 251, 193, 118, 227, 251, 193, 120, 227, 251, 193, 122, 227, 251, 193, 124, 227, 251, 193, 126, 227, 251, 193, 128, 227, 251, 193, 130, 227, 251, 193, 128, 226, 251, 193, 132, 227, 251, 193, 134, 227, 251, 193, 136, 227, 251, 193, 138, 227, 251, 193, 140, 227, 251, 193, 142, 227, 251, 193, 144, 227, 251, 193, 146, 227, 251, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 193, 152, 226, 251, 193, 148, 227, 251, 193, 150, 227, 251, 193, 152, 227, 251, 193, 154, 227, 251, 193, 156, 227, 251, 193, 158, 227, 251, 193, 160, 227, 251, 193, 162, 227, 251, 193, 164, 227, 251, 193, 166, 227, 251, 193, 168, 227, 251, 193, 170, 227, 251, 193, 136, 228, 251, 193, 138, 228, 251, 193, 140, 228, 251, 193, 142, 228, 251, 193, 68, 229, 251, 193, 70, 229, 251, 193, 72, 229, 251, 193, 74, 229, 251, 193, 76, 229, 251, 193, 78, 229, 251, 193, 14, 230, 251, 193, 16, 230, 251, 193, 18, 230, 251, 193, 20, 230, 251, 193, 22, 230, 251, 193, 24, 230, 251, 193, 26, 230, 251, 193, 28, 230, 251, 193, 30, 230, 251, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 96, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 193, 54, 227, 251, 193, 56, 227, 251, 193, 28, 228, 251, 193, 30, 228, 251, 193, 32, 228, 251, 193, 34, 228, 251, 193, 36, 228, 251, 193, 38, 228, 251, 193, 40, 228, 251, 193, 42, 228, 251, 193, 236, 228, 251, 193, 238, 228, 251, 193, 240, 228, 251, 193, 242, 228, 251, 193, 244, 228, 251, 193, 246, 228, 251, 193, 164, 229, 251, 193, 166, 229, 251, 193, 168, 229, 251, 193, 130, 230, 251, 193, 132, 230, 251, 193, 134, 230, 251, 193, 136, 230, 251, 193, 138, 230, 251, 193, 140, 230, 251, 193, 10, 231, 251, 193, 12, 231, 251, 193, 14, 231, 251, 193, 114, 231, 251, 193, 230, 231, 251, 193, 232, 231, 251, 193, 110, 232, 251, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 206, 128, 34, 0, 193, 44, 228, 251, 193, 46, 228, 251, 193, 48, 228, 251, 193, 50, 228, 251, 193, 52, 228, 251, 193, 54, 228, 251, 193, 56, 228, 251, 193, 58, 228, 251, 193, 60, 228, 251, 193, 62, 228, 251, 193, 64, 228, 251, 193, 66, 228, 251, 193, 68, 228, 251, 193, 70, 228, 251, 193, 72, 228, 251, 193, 74, 228, 251, 193, 76, 228, 251, 193, 78, 228, 251, 193, 80, 228, 251, 193, 82, 228, 251, 193, 84, 228, 251, 193, 86, 228, 251, 193, 88, 228, 251, 193, 90, 228, 251, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 206, 160, 34, 0, 193, 144, 228, 251, 193, 146, 228, 251, 193, 148, 228, 251, 193, 150, 228, 251, 193, 152, 228, 251, 193, 154, 228, 251, 193, 156, 228, 251, 193, 158, 228, 251, 193, 160, 228, 251, 193, 162, 228, 251, 193, 164, 228, 251, 193, 166, 228, 251, 193, 168, 228, 251, 193, 170, 228, 251, 193, 92, 228, 251, 193, 94, 228, 251, 193, 248, 228, 251, 193, 250, 228, 251, 193, 252, 228, 251, 193, 254, 228, 251, 193, 2, 229, 251, 193, 4, 229, 251, 193, 6, 229, 251, 193, 8, 229, 251, 193, 10, 229, 251, 193, 12, 229, 251, 193, 14, 229, 251, 193, 16, 229, 251, 193, 18, 229, 251, 193, 20, 229, 251, 193, 22, 229, 251, 193, 24, 229, 251, 193, 170, 229, 251, 193, 172, 229, 251, 193, 174, 229, 251, 193, 176, 229, 251, 193, 178, 229, 251, 193, 180, 229, 251, 193, 182, 229, 251, 193, 184, 229, 251, 193, 186, 229, 251, 193, 188, 229, 251, 193, 190, 229, 251, 193, 192, 229, 251, 193, 194, 229, 251, 193, 196, 229, 251, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 206, 192, 34, 0, 193, 38, 229, 251, 193, 40, 229, 251, 193, 42, 229, 251, 193, 44, 229, 251, 193, 46, 229, 251, 193, 48, 229, 251, 193, 50, 229, 251, 193, 52, 229, 251, 193, 54, 229, 251, 193, 56, 229, 251, 193, 58, 229, 251, 193, 60, 229, 251, 193, 62, 229, 251, 193, 64, 229, 251, 193, 66, 229, 251, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 193, 36, 229, 251, 193, 202, 229, 251, 193, 204, 229, 251, 193, 206, 229, 251, 193, 208, 229, 251, 193, 210, 229, 251, 193, 212, 229, 251, 193, 214, 229, 251, 193, 216, 229, 251, 193, 218, 229, 251, 193, 220, 229, 251, 193, 180, 230, 251, 193, 182, 230, 251, 193, 184, 230, 251, 193, 186, 230, 251, 193, 36, 231, 251, 193, 38, 231, 251, 193, 40, 231, 251, 193, 42, 231, 251, 193, 138, 231, 251, 193, 140, 231, 251, 193, 142, 231, 251, 193, 192, 231, 251, 193, 194, 231, 251, 193, 196, 231, 251, 193, 8, 232, 251, 193, 10, 232, 251, 193, 12, 232, 251, 193, 36, 232, 251, 193, 50, 232, 251, 193, 82, 232, 251, 193, 216, 232, 251, 193, 102, 230, 251, 193, 104, 230, 251, 193, 106, 230, 251, 193, 108, 230, 251, 193, 110, 230, 251, 193, 112, 230, 251, 193, 114, 230, 251, 193, 146, 229, 251, 193, 116, 230, 251, 193, 218, 230, 251, 193, 220, 230, 251, 193, 222, 230, 251, 193, 224, 230, 251, 193, 226, 230, 251, 193, 228, 230, 251, 193, 230, 230, 251, 193, 232, 230, 251, 193, 234, 230, 251, 193, 236, 230, 251, 193, 238, 230, 251, 193, 240, 230, 251, 193, 242, 230, 251, 193, 244, 230, 251, 193, 246, 230, 251, 193, 248, 230, 251, 193, 250, 230, 251, 193, 252, 230, 251, 193, 254, 230, 251, 193, 2, 231, 251, 193, 4, 231, 251, 193, 66, 231, 251, 193, 68, 231, 251, 193, 152, 229, 251, 193, 154, 229, 251, 193, 156, 229, 251, 193, 158, 229, 251, 193, 160, 229, 251, 193, 162, 229, 251, 193, 6, 231, 251, 193, 118, 230, 251, 193, 120, 230, 251, 193, 122, 230, 251, 193, 124, 230, 251, 193, 126, 230, 251, 193, 128, 230, 251, 193, 8, 231, 251, 193, 108, 231, 251, 193, 110, 231, 251, 193, 112, 231, 251, 193, 254, 231, 251, 193, 2, 232, 251, 193, 4, 232, 251, 193, 64, 232, 251, 193, 8, 233, 251, 193, 32, 233, 251, 193, 34, 233, 251, 193, 60, 233, 251, 193, 94, 233, 251, 193, 112, 233, 251, 193, 142, 233, 251, 193, 188, 233, 251, 193, 190, 233, 251, 193, 220, 233, 251, 193, 162, 234, 251, 193, 198, 229, 251, 193, 200, 229, 251, 193, 142, 230, 251, 193, 144, 230, 251, 193, 146, 230, 251, 193, 148, 230, 251, 193, 150, 230, 251, 193, 152, 230, 251, 193, 154, 230, 251, 193, 156, 230, 251, 193, 158, 230, 251, 193, 160, 230, 251, 193, 162, 230, 251, 193, 164, 230, 251, 193, 166, 230, 251, 193, 168, 230, 251, 193, 170, 230, 251, 193, 172, 230, 251, 193, 174, 230, 251, 193, 176, 230, 251, 193, 178, 230, 251, 193, 16, 231, 251, 193, 18, 231, 251, 193, 20, 231, 251, 193, 22, 231, 251, 193, 24, 231, 251, 193, 26, 231, 251, 193, 28, 231, 251, 193, 30, 231, 251, 193, 32, 231, 251, 193, 34, 231, 251, 193, 116, 231, 251, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 206, 0, 35, 0, 193, 188, 230, 251, 193, 190, 230, 251, 193, 192, 230, 251, 193, 194, 230, 251, 193, 196, 230, 251, 193, 198, 230, 251, 193, 200, 230, 251, 193, 202, 230, 251, 193, 204, 230, 251, 193, 206, 230, 251, 193, 44, 231, 251, 193, 46, 231, 251, 193, 208, 230, 251, 193, 48, 231, 251, 193, 50, 231, 251, 193, 144, 231, 251, 193, 146, 231, 251, 193, 148, 231, 251, 193, 38, 230, 251, 193, 40, 230, 251, 193, 42, 230, 251, 193, 44, 230, 251, 193, 46, 230, 251, 193, 48, 230, 251, 193, 50, 230, 251, 193, 52, 230, 251, 193, 54, 230, 251, 193, 56, 230, 251, 193, 58, 230, 251, 193, 60, 230, 251, 193, 62, 230, 251, 193, 64, 230, 251, 193, 66, 230, 251, 193, 68, 230, 251, 193, 70, 230, 251, 193, 72, 230, 251, 193, 74, 230, 251, 193, 76, 230, 251, 193, 78, 230, 251, 193, 80, 230, 251, 193, 82, 230, 251, 193, 84, 230, 251, 193, 86, 230, 251, 193, 88, 230, 251, 193, 90, 230, 251, 193, 92, 230, 251, 193, 94, 230, 251, 193, 96, 230, 251, 193, 98, 230, 251, 193, 100, 230, 251, 193, 210, 230, 251, 193, 212, 230, 251, 193, 214, 230, 251, 193, 52, 231, 251, 193, 54, 231, 251, 193, 56, 231, 251, 193, 58, 231, 251, 193, 60, 231, 251, 193, 62, 231, 251, 193, 156, 231, 251, 193, 158, 231, 251, 193, 210, 231, 251, 193, 14, 232, 251, 193, 16, 232, 251, 193, 44, 232, 251, 193, 72, 232, 251, 193, 98, 232, 251, 193, 22, 233, 251, 193, 36, 233, 251, 193, 132, 233, 251, 193, 212, 233, 251, 193, 214, 233, 251, 193, 250, 233, 251, 193, 252, 233, 251, 193, 46, 234, 251, 193, 112, 208, 252, 193, 148, 234, 251, 193, 150, 234, 251, 193, 192, 234, 251, 193, 194, 234, 251, 193, 196, 234, 251, 193, 234, 234, 251, 193, 70, 231, 251, 193, 72, 231, 251, 193, 74, 231, 251, 193, 76, 231, 251, 193, 78, 231, 251, 193, 80, 231, 251, 193, 82, 231, 251, 193, 84, 231, 251, 193, 86, 231, 251, 193, 88, 231, 251, 193, 90, 231, 251, 193, 92, 231, 251, 193, 94, 231, 251, 193, 96, 231, 251, 193, 98, 231, 251, 193, 100, 231, 251, 193, 102, 231, 251, 193, 104, 231, 251, 193, 160, 231, 251, 193, 162, 231, 251, 193, 164, 231, 251, 193, 166, 231, 251, 193, 168, 231, 251, 193, 170, 231, 251, 193, 172, 231, 251, 193, 174, 231, 251, 193, 176, 231, 251, 193, 178, 231, 251, 193, 180, 231, 251, 193, 182, 231, 251, 193, 184, 231, 251, 193, 212, 231, 251, 193, 86, 154, 252, 193, 88, 154, 252, 193, 106, 231, 251, 193, 90, 154, 252, 193, 92, 154, 252, 193, 98, 154, 252, 193, 100, 154, 252, 193, 102, 154, 252, 193, 112, 154, 252, 193, 114, 154, 252, 193, 116, 154, 252, 193, 118, 154, 252, 193, 120, 154, 252, 193, 122, 154, 252, 193, 126, 154, 252, 193, 128, 154, 252, 193, 134, 154, 252, 193, 136, 154, 252, 193, 146, 154, 252, 193, 140, 154, 252, 193, 142, 154, 252, 193, 148, 154, 252, 193, 150, 154, 252, 193, 156, 154, 252, 193, 158, 154, 252, 193, 164, 154, 252, 193, 166, 154, 252, 193, 168, 154, 252, 193, 170, 154, 252, 193, 194, 154, 252, 193, 196, 154, 252, 193, 198, 154, 252, 193, 118, 231, 251, 193, 120, 231, 251, 193, 122, 231, 251, 193, 124, 231, 251, 193, 126, 231, 251, 193, 128, 231, 251, 193, 130, 231, 251, 193, 132, 231, 251, 193, 134, 231, 251, 193, 136, 231, 251, 193, 186, 231, 251, 193, 188, 231, 251, 193, 190, 231, 251, 193, 234, 231, 251, 193, 236, 231, 251, 193, 238, 231, 251, 193, 240, 231, 251, 193, 6, 232, 251, 193, 48, 232, 251, 193, 66, 232, 251, 193, 68, 232, 251, 193, 78, 232, 251, 193, 80, 232, 251, 193, 84, 232, 251, 193, 100, 232, 251, 193, 142, 232, 251, 193, 144, 232, 251, 193, 172, 232, 251, 193, 174, 232, 251, 193, 214, 232, 251, 193, 248, 232, 251, 193, 12, 233, 251, 193, 214, 231, 251, 193, 216, 231, 251, 193, 218, 231, 251, 193, 220, 231, 251, 193, 222, 231, 251, 193, 224, 231, 251, 193, 226, 231, 251, 193, 228, 231, 251, 193, 244, 231, 251, 193, 246, 231, 251, 193, 248, 231, 251, 193, 250, 231, 251, 193, 252, 231, 251, 193, 18, 232, 251, 193, 20, 232, 251, 193, 22, 232, 251, 193, 24, 232, 251, 193, 26, 232, 251, 193, 28, 232, 251, 193, 30, 232, 251, 193, 32, 232, 251, 193, 34, 232, 251, 193, 46, 232, 251, 193, 56, 232, 251, 193, 58, 232, 251, 193, 60, 232, 251, 193, 82, 40, 253, 193, 62, 232, 251, 193, 74, 232, 251, 193, 76, 232, 251, 193, 128, 195, 252, 193, 86, 232, 251, 193, 88, 232, 251, 193, 106, 232, 251, 193, 108, 232, 251, 193, 116, 232, 251, 193, 118, 232, 251, 193, 120, 232, 251, 193, 122, 232, 251, 193, 130, 232, 251, 193, 132, 232, 251, 193, 134, 232, 251, 193, 136, 232, 251, 193, 138, 232, 251, 193, 140, 232, 251, 193, 150, 232, 251, 193, 152, 232, 251, 193, 154, 232, 251, 193, 156, 232, 251, 193, 158, 232, 251, 193, 160, 232, 251, 193, 162, 232, 251, 193, 164, 232, 251, 193, 166, 232, 251, 193, 168, 232, 251, 193, 182, 232, 251, 193, 184, 232, 251, 193, 186, 232, 251, 193, 188, 232, 251, 193, 202, 232, 251, 193, 204, 232, 251, 193, 206, 232, 251, 193, 222, 232, 251, 193, 224, 232, 251, 193, 112, 232, 251, 193, 124, 232, 251, 193, 170, 232, 251, 193, 190, 232, 251, 193, 192, 232, 251, 193, 208, 232, 251, 193, 210, 232, 251, 193, 212, 232, 251, 193, 242, 232, 251, 193, 244, 232, 251, 193, 246, 232, 251, 193, 10, 233, 251, 193, 42, 233, 251, 193, 66, 233, 251, 193, 68, 233, 251, 193, 70, 233, 251, 193, 96, 233, 251, 193, 114, 233, 251, 193, 116, 233, 251, 193, 118, 233, 251, 193, 144, 233, 251, 193, 146, 233, 251, 193, 148, 233, 251, 193, 150, 233, 251, 193, 152, 233, 251, 193, 192, 233, 251, 193, 224, 233, 251, 193, 226, 233, 251, 193, 228, 233, 251, 193, 18, 234, 251, 193, 20, 234, 251, 193, 22, 234, 251, 193, 218, 232, 251, 193, 250, 232, 251, 193, 54, 233, 251, 193, 80, 233, 251, 193, 166, 233, 251, 193, 202, 233, 251, 193, 234, 233, 251, 193, 236, 233, 251, 193, 102, 234, 251, 193, 104, 234, 251, 193, 144, 234, 251, 193, 188, 234, 251, 193, 226, 234, 251, 193, 228, 234, 251, 193, 12, 235, 251, 193, 72, 235, 251, 193, 206, 235, 251, 193, 208, 235, 251, 193, 210, 235, 251, 193, 106, 236, 251, 193, 144, 237, 251, 193, 146, 237, 251, 193, 148, 237, 251, 193, 150, 237, 251, 193, 4, 239, 251, 193, 6, 239, 251, 193, 8, 239, 251, 193, 10, 239, 251, 206, 96, 36, 0, 206, 96, 36, 0, 206, 96, 36, 0, 206, 96, 36, 0, 193, 226, 232, 251, 193, 228, 232, 251, 193, 230, 232, 251, 193, 236, 232, 251, 193, 238, 232, 251, 193, 240, 232, 251, 193, 252, 232, 251, 193, 254, 232, 251, 193, 2, 233, 251, 193, 4, 233, 251, 193, 6, 233, 251, 193, 24, 233, 251, 193, 26, 233, 251, 193, 28, 233, 251, 193, 30, 233, 251, 193, 38, 233, 251, 193, 40, 233, 251, 193, 44, 233, 251, 193, 46, 233, 251, 193, 48, 233, 251, 193, 50, 233, 251, 193, 52, 233, 251, 193, 56, 233, 251, 193, 58, 233, 251, 193, 62, 233, 251, 193, 64, 233, 251, 193, 76, 233, 251, 193, 82, 233, 251, 193, 84, 233, 251, 193, 88, 233, 251, 193, 92, 233, 251, 193, 102, 233, 251, 193, 14, 233, 251, 193, 16, 233, 251, 193, 18, 233, 251, 193, 74, 233, 251, 193, 98, 233, 251, 193, 100, 233, 251, 193, 120, 233, 251, 193, 122, 233, 251, 193, 154, 233, 251, 193, 156, 233, 251, 193, 158, 233, 251, 193, 160, 233, 251, 193, 162, 233, 251, 193, 164, 233, 251, 193, 194, 233, 251, 193, 196, 233, 251, 193, 198, 233, 251, 193, 200, 233, 251, 193, 230, 233, 251, 193, 232, 233, 251, 193, 28, 234, 251, 193, 30, 234, 251, 193, 32, 234, 251, 193, 34, 234, 251, 193, 36, 234, 251, 193, 92, 234, 251, 193, 94, 234, 251, 193, 96, 234, 251, 193, 98, 234, 251, 193, 100, 234, 251, 193, 138, 234, 251, 193, 140, 234, 251, 193, 104, 233, 251, 193, 106, 233, 251, 193, 108, 233, 251, 193, 110, 233, 251, 193, 138, 233, 251, 193, 174, 233, 251, 193, 176, 233, 251, 193, 178, 233, 251, 193, 170, 205, 252, 193, 180, 233, 251, 193, 182, 233, 251, 193, 184, 233, 251, 193, 186, 233, 251, 193, 216, 233, 251, 193, 218, 233, 251, 193, 160, 206, 252, 193, 162, 206, 252, 193, 254, 233, 251, 193, 2, 234, 251, 193, 4, 234, 251, 193, 6, 234, 251, 193, 8, 234, 251, 193, 10, 234, 251, 206, 32, 35, 0, 206, 32, 35, 0, 206, 32, 35, 0, 206, 32, 35, 0, 206, 32, 35, 0, 206, 32, 35, 0, 206, 32, 35, 0, 206, 32, 35, 0, 206, 32, 35, 0, 193, 196, 204, 252, 193, 198, 204, 252, 193, 200, 204, 252, 193, 202, 204, 252, 193, 204, 204, 252, 193, 206, 204, 252, 193, 208, 204, 252, 193, 210, 204, 252, 193, 212, 204, 252, 193, 214, 204, 252, 193, 216, 204, 252, 193, 218, 204, 252, 193, 220, 204, 252, 193, 222, 204, 252, 193, 224, 204, 252, 193, 226, 204, 252, 193, 228, 204, 252, 193, 230, 204, 252, 193, 232, 204, 252, 193, 234, 204, 252, 193, 236, 204, 252, 193, 238, 204, 252, 193, 240, 204, 252, 193, 140, 233, 251, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 193, 12, 234, 251, 193, 14, 234, 251, 193, 16, 234, 251, 193, 78, 234, 251, 193, 128, 234, 251, 193, 130, 234, 251, 193, 132, 234, 251, 193, 134, 234, 251, 193, 164, 234, 251, 193, 166, 234, 251, 193, 168, 234, 251, 193, 214, 234, 251, 193, 216, 234, 251, 193, 252, 234, 251, 193, 92, 235, 251, 193, 94, 235, 251, 193, 108, 235, 251, 193, 84, 236, 251, 193, 86, 236, 251, 193, 88, 236, 251, 193, 90, 236, 251, 193, 92, 236, 251, 193, 112, 237, 251, 193, 114, 237, 251, 193, 116, 237, 251, 193, 118, 237, 251, 193, 202, 238, 251, 193, 204, 238, 251, 193, 206, 238, 251, 193, 208, 238, 251, 193, 210, 238, 251, 193, 212, 238, 251, 193, 24, 234, 251, 193, 26, 234, 251, 193, 84, 234, 251, 193, 86, 234, 251, 193, 88, 234, 251, 193, 90, 234, 251, 193, 136, 234, 251, 193, 170, 234, 251, 193, 172, 234, 251, 193, 174, 234, 251, 193, 176, 234, 251, 193, 178, 234, 251, 193, 180, 234, 251, 193, 218, 234, 251, 193, 220, 234, 251, 193, 254, 234, 251, 193, 2, 235, 251, 193, 32, 235, 251, 193, 34, 235, 251, 193, 36, 235, 251, 193, 38, 235, 251, 193, 62, 235, 251, 193, 64, 235, 251, 193, 74, 235, 251, 193, 76, 235, 251, 193, 96, 235, 251, 193, 112, 235, 251, 193, 122, 235, 251, 193, 144, 235, 251, 193, 94, 236, 251, 193, 96, 236, 251, 193, 98, 236, 251, 206, 32, 35, 0, 206, 32, 35, 0, 206, 32, 35, 0, 206, 32, 35, 0, 193, 110, 234, 251, 193, 112, 234, 251, 193, 114, 234, 251, 193, 116, 234, 251, 193, 118, 234, 251, 193, 120, 234, 251, 193, 122, 234, 251, 193, 124, 234, 251, 193, 126, 234, 251, 193, 152, 234, 251, 193, 154, 234, 251, 193, 156, 234, 251, 193, 158, 234, 251, 193, 160, 234, 251, 193, 198, 234, 251, 193, 200, 234, 251, 193, 202, 234, 251, 193, 204, 234, 251, 193, 206, 234, 251, 193, 208, 234, 251, 193, 210, 234, 251, 193, 212, 234, 251, 193, 236, 234, 251, 193, 238, 234, 251, 193, 16, 235, 251, 193, 240, 234, 251, 193, 242, 234, 251, 193, 244, 234, 251, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 193, 76, 234, 251, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 193, 106, 234, 251, 193, 116, 209, 252, 193, 146, 234, 251, 193, 230, 234, 251, 193, 232, 234, 251, 193, 88, 212, 252, 193, 50, 235, 251, 193, 86, 235, 251, 193, 114, 235, 251, 193, 116, 235, 251, 193, 124, 235, 251, 193, 126, 235, 251, 193, 128, 235, 251, 193, 130, 235, 251, 193, 132, 235, 251, 193, 134, 235, 251, 193, 136, 235, 251, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 193, 142, 234, 251, 193, 182, 234, 251, 193, 184, 234, 251, 193, 186, 234, 251, 193, 222, 234, 251, 193, 224, 234, 251, 193, 4, 235, 251, 193, 6, 235, 251, 193, 8, 235, 251, 193, 10, 235, 251, 193, 40, 235, 251, 193, 42, 235, 251, 193, 44, 235, 251, 193, 46, 235, 251, 193, 48, 235, 251, 193, 66, 235, 251, 193, 68, 235, 251, 193, 70, 235, 251, 193, 78, 235, 251, 193, 80, 235, 251, 193, 82, 235, 251, 193, 84, 235, 251, 193, 104, 235, 251, 193, 110, 235, 251, 193, 146, 235, 251, 193, 100, 236, 251, 193, 102, 236, 251, 193, 104, 236, 251, 193, 132, 237, 251, 193, 134, 237, 251, 193, 136, 237, 251, 193, 138, 237, 251, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 193, 190, 234, 251, 193, 90, 212, 252, 193, 92, 212, 252, 193, 94, 212, 252, 193, 96, 212, 252, 193, 98, 212, 252, 193, 100, 212, 252, 193, 102, 212, 252, 193, 104, 212, 252, 193, 106, 212, 252, 193, 108, 212, 252, 193, 110, 212, 252, 193, 112, 212, 252, 193, 114, 212, 252, 193, 116, 212, 252, 193, 118, 212, 252, 193, 120, 212, 252, 193, 122, 212, 252, 193, 124, 212, 252, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 193, 246, 234, 251, 193, 248, 234, 251, 193, 250, 234, 251, 193, 18, 235, 251, 193, 20, 235, 251, 193, 22, 235, 251, 193, 24, 235, 251, 193, 26, 235, 251, 193, 28, 235, 251, 193, 52, 235, 251, 193, 54, 235, 251, 193, 56, 235, 251, 193, 58, 235, 251, 193, 60, 235, 251, 193, 88, 235, 251, 193, 90, 235, 251, 193, 98, 235, 251, 193, 100, 235, 251, 193, 106, 235, 251, 193, 118, 235, 251, 193, 120, 235, 251, 193, 138, 235, 251, 193, 140, 235, 251, 193, 142, 235, 251, 193, 188, 235, 251, 193, 190, 235, 251, 193, 192, 235, 251, 193, 194, 235, 251, 193, 196, 235, 251, 193, 198, 235, 251, 193, 200, 235, 251, 193, 202, 235, 251, 193, 14, 235, 251, 193, 186, 235, 251, 193, 36, 236, 251, 193, 38, 236, 251, 193, 40, 236, 251, 193, 42, 236, 251, 193, 44, 236, 251, 193, 242, 236, 251, 193, 244, 236, 251, 193, 246, 236, 251, 193, 248, 236, 251, 193, 250, 236, 251, 193, 252, 236, 251, 193, 254, 236, 251, 193, 2, 237, 251, 193, 4, 237, 251, 193, 6, 237, 251, 193, 98, 238, 251, 193, 100, 238, 251, 193, 102, 238, 251, 193, 104, 238, 251, 193, 106, 238, 251, 193, 108, 238, 251, 193, 110, 238, 251, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 193, 212, 214, 252, 193, 250, 214, 252, 193, 20, 215, 252, 193, 102, 235, 251, 193, 88, 215, 252, 193, 116, 215, 252, 193, 148, 215, 252, 193, 154, 215, 252, 193, 166, 215, 252, 193, 206, 215, 252, 193, 52, 216, 252, 193, 88, 216, 252, 193, 110, 216, 252, 193, 150, 216, 252, 193, 212, 216, 252, 193, 222, 216, 252, 193, 224, 216, 252, 193, 2, 217, 252, 193, 14, 218, 252, 193, 86, 218, 252, 193, 116, 218, 252, 193, 198, 218, 252, 193, 200, 218, 252, 193, 38, 219, 252, 193, 68, 219, 252, 193, 80, 219, 252, 193, 180, 219, 252, 193, 240, 220, 252, 193, 66, 221, 252, 193, 68, 221, 252, 193, 180, 221, 252, 193, 244, 221, 252, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 64, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 193, 204, 235, 251, 193, 46, 236, 251, 193, 48, 236, 251, 193, 50, 236, 251, 193, 52, 236, 251, 193, 54, 236, 251, 193, 56, 236, 251, 193, 58, 236, 251, 193, 60, 236, 251, 193, 62, 236, 251, 193, 64, 236, 251, 193, 66, 236, 251, 193, 68, 236, 251, 193, 70, 236, 251, 193, 72, 236, 251, 193, 74, 236, 251, 193, 76, 236, 251, 193, 78, 236, 251, 193, 80, 236, 251, 193, 82, 236, 251, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 206, 96, 35, 0, 193, 108, 236, 251, 193, 110, 236, 251, 193, 112, 236, 251, 193, 114, 236, 251, 193, 116, 236, 251, 193, 118, 236, 251, 193, 120, 236, 251, 193, 122, 236, 251, 193, 124, 236, 251, 193, 126, 236, 251, 193, 128, 236, 251, 193, 130, 236, 251, 193, 132, 236, 251, 193, 134, 236, 251, 193, 136, 236, 251, 193, 138, 236, 251, 193, 140, 236, 251, 193, 142, 236, 251, 193, 144, 236, 251, 193, 146, 236, 251, 193, 148, 236, 251, 193, 150, 236, 251, 193, 152, 236, 251, 193, 34, 236, 251, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 128, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 160, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 193, 122, 237, 251, 193, 124, 237, 251, 193, 126, 237, 251, 193, 128, 237, 251, 193, 130, 237, 251, 193, 220, 238, 251, 193, 222, 238, 251, 193, 146, 240, 251, 193, 148, 240, 251, 193, 150, 240, 251, 193, 152, 240, 251, 193, 154, 240, 251, 193, 156, 240, 251, 193, 158, 240, 251, 193, 160, 240, 251, 193, 162, 240, 251, 193, 164, 240, 251, 193, 166, 240, 251, 193, 82, 242, 251, 193, 84, 242, 251, 193, 86, 242, 251, 193, 88, 242, 251, 193, 90, 242, 251, 193, 92, 242, 251, 193, 94, 242, 251, 193, 96, 242, 251, 193, 98, 242, 251, 193, 100, 242, 251, 193, 102, 242, 251, 193, 172, 244, 251, 193, 174, 244, 251, 193, 176, 244, 251, 193, 140, 237, 251, 193, 142, 237, 251, 193, 224, 238, 251, 193, 226, 238, 251, 193, 228, 238, 251, 193, 230, 238, 251, 193, 232, 238, 251, 193, 234, 238, 251, 193, 236, 238, 251, 193, 238, 238, 251, 193, 240, 238, 251, 193, 242, 238, 251, 193, 244, 238, 251, 193, 246, 238, 251, 193, 248, 238, 251, 193, 250, 238, 251, 193, 252, 238, 251, 193, 254, 238, 251, 193, 2, 239, 251, 193, 168, 240, 251, 193, 170, 240, 251, 193, 172, 240, 251, 193, 174, 240, 251, 193, 176, 240, 251, 193, 178, 240, 251, 193, 180, 240, 251, 193, 182, 240, 251, 193, 184, 240, 251, 193, 186, 240, 251, 193, 188, 240, 251, 193, 190, 240, 251, 193, 192, 240, 251, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 192, 35, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 224, 35, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 193, 214, 238, 251, 193, 216, 238, 251, 193, 218, 238, 251, 193, 116, 240, 251, 193, 118, 240, 251, 193, 120, 240, 251, 193, 122, 240, 251, 193, 124, 240, 251, 193, 126, 240, 251, 193, 128, 240, 251, 193, 130, 240, 251, 193, 132, 240, 251, 193, 134, 240, 251, 193, 136, 240, 251, 193, 138, 240, 251, 193, 140, 240, 251, 193, 52, 242, 251, 193, 54, 242, 251, 193, 56, 242, 251, 193, 58, 242, 251, 193, 60, 242, 251, 193, 62, 242, 251, 193, 64, 242, 251, 193, 66, 242, 251, 193, 68, 242, 251, 193, 70, 242, 251, 193, 72, 242, 251, 193, 74, 242, 251, 193, 76, 242, 251, 193, 78, 242, 251, 193, 80, 242, 251, 193, 138, 244, 251, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 0, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 206, 32, 36, 0, 193, 144, 241, 251, 193, 146, 241, 251, 193, 148, 241, 251, 193, 150, 241, 251, 193, 152, 241, 251, 193, 154, 241, 251, 193, 156, 241, 251, 193, 158, 241, 251, 193, 160, 241, 251, 193, 162, 241, 251, 193, 164, 241, 251, 193, 166, 241, 251, 193, 168, 241, 251, 193, 170, 241, 251, 193, 128, 243, 251, 193, 172, 241, 251, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 206, 64, 36, 0, 193, 194, 240, 251, 193, 196, 240, 251, 193, 198, 240, 251, 193, 104, 242, 251, 193, 106, 242, 251, 193, 108, 242, 251, 193, 110, 242, 251, 193, 112, 242, 251, 193, 114, 242, 251, 193, 116, 242, 251, 193, 118, 242, 251, 193, 120, 242, 251, 193, 122, 242, 251, 193, 124, 242, 251, 193, 126, 242, 251, 193, 128, 242, 251, 193, 130, 242, 251, 193, 132, 242, 251, 193, 134, 242, 251, 193, 136, 242, 251, 193, 138, 242, 251, 193, 140, 242, 251, 193, 142, 242, 251, 193, 144, 242, 251, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 96, 36, 0, 206, 96, 36, 0, 206, 96, 36, 0, 206, 96, 36, 0, 206, 96, 36, 0, 206, 96, 36, 0, 206, 96, 36, 0, 193, 146, 242, 251, 193, 148, 242, 251, 193, 150, 242, 251, 193, 152, 242, 251, 193, 154, 242, 251, 193, 156, 242, 251, 193, 158, 242, 251, 193, 160, 242, 251, 193, 162, 242, 251, 193, 164, 242, 251, 193, 24, 245, 251, 193, 26, 245, 251, 193, 28, 245, 251, 193, 30, 245, 251, 193, 32, 245, 251, 193, 34, 245, 251, 193, 36, 245, 251, 193, 38, 245, 251, 193, 40, 245, 251, 206, 0, 38, 0, 206, 0, 38, 0, 206, 0, 38, 0, 206, 0, 38, 0, 206, 0, 38, 0, 206, 0, 38, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 128, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 160, 36, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 192, 36, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 206, 224, 36, 0, 193, 234, 245, 251, 193, 236, 245, 251, 193, 238, 245, 251, 193, 240, 245, 251, 193, 242, 245, 251, 193, 244, 245, 251, 193, 246, 245, 251, 193, 248, 245, 251, 193, 250, 245, 251, 193, 252, 245, 251, 193, 254, 245, 251, 193, 2, 246, 251, 193, 4, 246, 251, 193, 6, 246, 251, 193, 8, 246, 251, 193, 10, 246, 251, 193, 12, 246, 251, 193, 14, 246, 251, 193, 16, 246, 251, 193, 18, 246, 251, 193, 20, 246, 251, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 0, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 193, 140, 244, 251, 193, 142, 244, 251, 193, 144, 244, 251, 193, 146, 244, 251, 193, 148, 244, 251, 193, 150, 244, 251, 193, 152, 244, 251, 193, 154, 244, 251, 193, 156, 244, 251, 193, 158, 244, 251, 193, 160, 244, 251, 193, 162, 244, 251, 193, 164, 244, 251, 193, 166, 244, 251, 193, 168, 244, 251, 193, 170, 244, 251, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 193, 178, 244, 251, 193, 180, 244, 251, 193, 182, 244, 251, 193, 184, 244, 251, 193, 186, 244, 251, 193, 188, 244, 251, 193, 190, 244, 251, 193, 192, 244, 251, 193, 194, 244, 251, 193, 196, 244, 251, 193, 198, 244, 251, 193, 200, 244, 251, 193, 202, 244, 251, 193, 204, 244, 251, 193, 206, 244, 251, 193, 208, 244, 251, 193, 210, 244, 251, 193, 212, 244, 251, 193, 214, 244, 251, 193, 216, 244, 251, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 32, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 206, 64, 37, 0, 193, 192, 247, 251, 193, 224, 245, 251, 193, 226, 245, 251, 193, 194, 247, 251, 193, 196, 247, 251, 193, 198, 247, 251, 193, 200, 247, 251, 193, 202, 247, 251, 193, 204, 247, 251, 193, 206, 247, 251, 193, 228, 245, 251, 193, 208, 247, 251, 193, 210, 247, 251, 193, 212, 247, 251, 193, 230, 245, 251, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 193, 22, 246, 251, 193, 108, 55, 252, 193, 24, 246, 251, 193, 26, 246, 251, 193, 112, 248, 251, 193, 114, 248, 251, 193, 116, 248, 251, 193, 118, 248, 251, 193, 120, 248, 251, 193, 122, 248, 251, 193, 124, 248, 251, 193, 126, 248, 251, 193, 128, 248, 251, 193, 130, 248, 251, 193, 132, 248, 251, 193, 134, 248, 251, 193, 136, 248, 251, 193, 138, 248, 251, 193, 140, 248, 251, 193, 142, 248, 251, 193, 144, 248, 251, 193, 146, 248, 251, 193, 148, 248, 251, 193, 150, 248, 251, 193, 152, 248, 251, 193, 154, 248, 251, 193, 156, 248, 251, 193, 158, 248, 251, 193, 160, 248, 251, 193, 162, 248, 251, 193, 190, 250, 251, 193, 192, 250, 251, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 206, 96, 37, 0, 193, 164, 248, 251, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 128, 37, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 206, 160, 37, 0, 193, 126, 249, 251, 193, 128, 249, 251, 193, 130, 249, 251, 193, 132, 249, 251, 193, 134, 249, 251, 193, 136, 249, 251, 193, 138, 249, 251, 193, 140, 249, 251, 193, 142, 249, 251, 193, 144, 249, 251, 193, 146, 249, 251, 193, 148, 249, 251, 193, 150, 249, 251, 193, 152, 249, 251, 193, 154, 249, 251, 193, 156, 249, 251, 193, 158, 249, 251, 193, 160, 249, 251, 193, 162, 249, 251, 193, 164, 249, 251, 193, 166, 249, 251, 193, 180, 251, 251, 193, 182, 251, 251, 193, 184, 251, 251, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 192, 37, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 206, 224, 37, 0, 193, 220, 249, 251, 193, 222, 249, 251, 193, 224, 249, 251, 193, 226, 249, 251, 193, 228, 249, 251, 193, 230, 249, 251, 193, 232, 249, 251, 193, 234, 249, 251, 193, 236, 249, 251, 193, 238, 249, 251, 193, 240, 249, 251, 193, 242, 249, 251, 193, 244, 249, 251, 193, 246, 249, 251, 193, 248, 249, 251, 193, 250, 249, 251, 193, 252, 249, 251, 206, 0, 38, 0, 206, 0, 38, 0, 206, 0, 38, 0, 206, 0, 38, 0, 206, 0, 38, 0, 206, 0, 38, 0, 206, 0, 38, 0, 193, 254, 249, 251, 193, 2, 250, 251, 193, 4, 250, 251, 193, 6, 250, 251, 193, 8, 250, 251, 193, 10, 250, 251, 193, 12, 250, 251, 193, 14, 250, 251, 193, 16, 250, 251, 193, 18, 250, 251, 193, 20, 250, 251, 193, 22, 250, 251, 193, 24, 250, 251, 193, 26, 250, 251, 193, 28, 250, 251, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 206, 32, 38, 0, 193, 30, 250, 251, 193, 32, 250, 251, 193, 34, 250, 251, 193, 36, 250, 251, 193, 38, 250, 251, 193, 40, 250, 251, 193, 42, 250, 251, 193, 106, 248, 251, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 193, 108, 248, 251, 193, 180, 250, 251, 193, 182, 250, 251, 193, 184, 250, 251, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 64, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 206, 96, 38, 0, 193, 218, 251, 251, 193, 220, 251, 251, 193, 222, 251, 251, 193, 224, 251, 251, 193, 226, 251, 251, 193, 228, 251, 251, 193, 230, 251, 251, 193, 232, 251, 251, 193, 234, 251, 251, 193, 236, 251, 251, 193, 238, 251, 251, 193, 240, 251, 251, 193, 242, 251, 251, 193, 244, 251, 251, 193, 246, 251, 251, 193, 248, 251, 251, 193, 250, 251, 251, 193, 252, 251, 251, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 128, 38, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 193, 186, 250, 251, 193, 186, 252, 251, 193, 188, 252, 251, 193, 190, 252, 251, 193, 192, 252, 251, 193, 194, 252, 251, 193, 196, 252, 251, 193, 198, 252, 251, 193, 200, 252, 251, 193, 202, 252, 251, 193, 204, 252, 251, 193, 206, 252, 251, 193, 208, 252, 251, 193, 210, 252, 251, 193, 212, 252, 251, 193, 188, 250, 251, 193, 214, 252, 251, 193, 216, 252, 251, 193, 218, 252, 251, 193, 220, 252, 251, 193, 222, 252, 251, 193, 224, 252, 251, 193, 226, 252, 251, 193, 228, 252, 251, 193, 134, 254, 251, 193, 136, 254, 251, 193, 138, 254, 251, 193, 194, 250, 251, 193, 196, 250, 251, 193, 198, 250, 251, 193, 200, 250, 251, 193, 202, 250, 251, 193, 204, 250, 251, 193, 206, 250, 251, 193, 208, 250, 251, 193, 210, 250, 251, 193, 212, 250, 251, 193, 214, 250, 251, 193, 216, 250, 251, 193, 218, 250, 251, 193, 220, 250, 251, 193, 222, 250, 251, 193, 224, 250, 251, 193, 230, 252, 251, 193, 226, 250, 251, 193, 228, 250, 251, 193, 232, 252, 251, 193, 234, 252, 251, 193, 236, 252, 251, 193, 238, 252, 251, 193, 240, 252, 251, 193, 242, 252, 251, 193, 244, 252, 251, 193, 246, 252, 251, 193, 248, 252, 251, 193, 250, 252, 251, 193, 252, 252, 251, 193, 254, 252, 251, 193, 2, 253, 251, 193, 230, 250, 251, 193, 4, 253, 251, 193, 6, 253, 251, 193, 8, 253, 251, 193, 10, 253, 251, 193, 12, 253, 251, 193, 14, 253, 251, 193, 16, 253, 251, 193, 18, 253, 251, 193, 20, 253, 251, 193, 226, 254, 251, 193, 228, 254, 251, 193, 230, 254, 251, 193, 232, 254, 251, 193, 234, 254, 251, 193, 236, 254, 251, 193, 238, 254, 251, 193, 240, 254, 251, 193, 242, 254, 251, 193, 244, 254, 251, 193, 246, 254, 251, 193, 248, 254, 251, 193, 250, 254, 251, 193, 48, 2, 252, 193, 50, 2, 252, 193, 52, 2, 252, 193, 54, 2, 252, 193, 56, 2, 252, 193, 58, 2, 252, 193, 60, 2, 252, 193, 88, 3, 252, 193, 90, 3, 252, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 206, 160, 38, 0, 193, 22, 253, 251, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 192, 38, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 193, 186, 251, 251, 193, 188, 251, 251, 193, 190, 251, 251, 193, 192, 251, 251, 193, 194, 251, 251, 193, 196, 251, 251, 193, 198, 251, 251, 193, 200, 251, 251, 193, 202, 251, 251, 193, 204, 251, 251, 193, 206, 251, 251, 193, 208, 251, 251, 193, 210, 251, 251, 193, 4, 254, 251, 193, 6, 254, 251, 193, 8, 254, 251, 193, 10, 254, 251, 193, 12, 254, 251, 193, 14, 254, 251, 193, 16, 254, 251, 193, 18, 254, 251, 193, 20, 254, 251, 193, 22, 254, 251, 193, 24, 254, 251, 193, 26, 254, 251, 193, 28, 254, 251, 193, 30, 254, 251, 193, 32, 254, 251, 193, 34, 254, 251, 193, 158, 255, 251, 193, 160, 255, 251, 193, 162, 255, 251, 193, 254, 251, 251, 193, 2, 252, 251, 193, 4, 252, 251, 193, 6, 252, 251, 193, 8, 252, 251, 193, 10, 252, 251, 193, 12, 252, 251, 193, 14, 252, 251, 193, 16, 252, 251, 193, 18, 252, 251, 193, 20, 252, 251, 193, 22, 252, 251, 193, 24, 252, 251, 193, 26, 252, 251, 193, 28, 252, 251, 193, 30, 252, 251, 193, 32, 252, 251, 193, 34, 252, 251, 193, 36, 252, 251, 193, 38, 252, 251, 193, 40, 252, 251, 193, 42, 252, 251, 193, 44, 252, 251, 193, 46, 252, 251, 193, 48, 252, 251, 193, 50, 252, 251, 193, 52, 252, 251, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 206, 224, 38, 0, 193, 106, 254, 251, 193, 108, 254, 251, 193, 110, 254, 251, 193, 112, 254, 251, 193, 114, 254, 251, 193, 116, 254, 251, 193, 118, 254, 251, 193, 120, 254, 251, 193, 122, 254, 251, 193, 124, 254, 251, 193, 126, 254, 251, 193, 128, 254, 251, 193, 130, 254, 251, 193, 132, 254, 251, 193, 222, 255, 251, 193, 224, 255, 251, 193, 226, 255, 251, 193, 228, 255, 251, 193, 230, 255, 251, 193, 232, 255, 251, 193, 234, 255, 251, 193, 236, 255, 251, 193, 22, 3, 252, 193, 24, 3, 252, 193, 26, 3, 252, 193, 14, 4, 252, 193, 16, 4, 252, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 0, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 206, 32, 39, 0, 193, 252, 254, 251, 193, 254, 254, 251, 193, 2, 255, 251, 193, 4, 255, 251, 193, 6, 255, 251, 193, 8, 255, 251, 193, 10, 255, 251, 193, 12, 255, 251, 193, 14, 255, 251, 193, 16, 255, 251, 193, 18, 255, 251, 193, 20, 255, 251, 193, 22, 255, 251, 193, 24, 255, 251, 193, 26, 255, 251, 193, 28, 255, 251, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 206, 64, 39, 0, 193, 184, 255, 251, 193, 186, 255, 251, 193, 188, 255, 251, 193, 190, 255, 251, 193, 192, 255, 251, 193, 194, 255, 251, 193, 196, 255, 251, 193, 198, 255, 251, 193, 200, 255, 251, 193, 202, 255, 251, 193, 204, 255, 251, 193, 232, 2, 252, 193, 234, 2, 252, 193, 236, 2, 252, 193, 238, 2, 252, 193, 240, 2, 252, 193, 242, 2, 252, 193, 244, 2, 252, 193, 246, 2, 252, 193, 248, 2, 252, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 206, 96, 39, 0, 193, 206, 255, 251, 193, 208, 255, 251, 193, 210, 255, 251, 193, 212, 255, 251, 193, 214, 255, 251, 193, 216, 255, 251, 193, 218, 255, 251, 193, 220, 255, 251, 193, 252, 2, 252, 193, 254, 2, 252, 193, 2, 3, 252, 193, 4, 3, 252, 193, 6, 3, 252, 193, 8, 3, 252, 193, 10, 3, 252, 193, 12, 3, 252, 193, 14, 3, 252, 193, 16, 3, 252, 193, 18, 3, 252, 193, 20, 3, 252, 193, 246, 3, 252, 193, 248, 3, 252, 193, 140, 254, 251, 193, 142, 254, 251, 193, 144, 254, 251, 193, 146, 254, 251, 193, 148, 254, 251, 193, 150, 254, 251, 193, 152, 254, 251, 193, 154, 254, 251, 193, 156, 254, 251, 193, 158, 254, 251, 193, 160, 254, 251, 193, 162, 254, 251, 193, 164, 254, 251, 193, 166, 254, 251, 193, 168, 254, 251, 193, 170, 254, 251, 193, 172, 254, 251, 193, 174, 254, 251, 193, 176, 254, 251, 193, 178, 254, 251, 193, 180, 254, 251, 193, 182, 254, 251, 193, 184, 254, 251, 193, 186, 254, 251, 193, 188, 254, 251, 193, 190, 254, 251, 193, 192, 254, 251, 193, 194, 254, 251, 193, 196, 254, 251, 193, 198, 254, 251, 193, 200, 254, 251, 193, 202, 254, 251, 193, 204, 254, 251, 193, 206, 254, 251, 193, 208, 254, 251, 193, 210, 254, 251, 193, 212, 254, 251, 193, 214, 254, 251, 193, 216, 254, 251, 193, 218, 254, 251, 193, 220, 254, 251, 193, 222, 254, 251, 193, 224, 254, 251, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 193, 30, 255, 251, 193, 32, 255, 251, 193, 34, 255, 251, 193, 36, 255, 251, 193, 38, 255, 251, 193, 40, 255, 251, 193, 42, 255, 251, 193, 44, 255, 251, 193, 46, 255, 251, 193, 48, 255, 251, 193, 50, 255, 251, 193, 52, 255, 251, 193, 54, 255, 251, 193, 56, 255, 251, 193, 58, 255, 251, 193, 60, 255, 251, 193, 62, 255, 251, 193, 64, 255, 251, 193, 66, 255, 251, 193, 68, 255, 251, 193, 70, 255, 251, 193, 72, 255, 251, 193, 74, 255, 251, 193, 76, 255, 251, 193, 78, 255, 251, 193, 80, 255, 251, 193, 82, 255, 251, 193, 84, 255, 251, 193, 86, 255, 251, 193, 88, 255, 251, 193, 90, 255, 251, 193, 92, 255, 251, 193, 94, 255, 251, 193, 96, 255, 251, 193, 138, 140, 252, 193, 98, 255, 251, 193, 100, 255, 251, 193, 102, 255, 251, 193, 104, 255, 251, 193, 106, 255, 251, 193, 108, 255, 251, 193, 110, 255, 251, 193, 112, 255, 251, 193, 114, 255, 251, 193, 116, 255, 251, 193, 118, 255, 251, 193, 120, 255, 251, 193, 122, 255, 251, 193, 124, 255, 251, 193, 126, 255, 251, 193, 128, 255, 251, 193, 130, 255, 251, 193, 132, 255, 251, 193, 134, 255, 251, 193, 136, 255, 251, 193, 138, 255, 251, 193, 140, 255, 251, 193, 142, 255, 251, 193, 144, 255, 251, 193, 146, 255, 251, 193, 148, 255, 251, 193, 150, 255, 251, 193, 152, 255, 251, 193, 154, 255, 251, 193, 156, 255, 251, 193, 62, 2, 252, 193, 64, 2, 252, 193, 66, 2, 252, 193, 68, 2, 252, 193, 70, 2, 252, 193, 72, 2, 252, 193, 74, 2, 252, 193, 76, 2, 252, 193, 78, 2, 252, 193, 80, 2, 252, 193, 82, 2, 252, 193, 84, 2, 252, 193, 86, 2, 252, 193, 88, 2, 252, 193, 90, 2, 252, 193, 92, 2, 252, 193, 94, 2, 252, 193, 96, 2, 252, 193, 98, 2, 252, 193, 100, 2, 252, 193, 102, 2, 252, 193, 104, 2, 252, 193, 106, 2, 252, 193, 108, 2, 252, 193, 110, 3, 252, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 193, 164, 255, 251, 193, 166, 255, 251, 193, 168, 255, 251, 193, 170, 255, 251, 193, 172, 255, 251, 193, 174, 255, 251, 193, 176, 255, 251, 193, 178, 255, 251, 193, 180, 255, 251, 193, 216, 2, 252, 193, 218, 2, 252, 193, 220, 2, 252, 193, 222, 2, 252, 193, 224, 2, 252, 193, 226, 2, 252, 193, 228, 2, 252, 193, 218, 3, 252, 193, 220, 3, 252, 193, 222, 3, 252, 193, 224, 3, 252, 193, 226, 3, 252, 193, 228, 3, 252, 193, 164, 4, 252, 193, 166, 4, 252, 193, 168, 4, 252, 193, 170, 4, 252, 193, 172, 4, 252, 193, 62, 5, 252, 193, 64, 5, 252, 193, 66, 5, 252, 193, 68, 5, 252, 193, 180, 5, 252, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 128, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 206, 160, 39, 0, 193, 72, 133, 252, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 193, 250, 2, 252, 193, 230, 3, 252, 193, 232, 3, 252, 193, 234, 3, 252, 193, 236, 3, 252, 193, 238, 3, 252, 193, 240, 3, 252, 193, 242, 3, 252, 193, 244, 3, 252, 193, 174, 4, 252, 193, 176, 4, 252, 193, 70, 5, 252, 193, 72, 5, 252, 193, 74, 5, 252, 193, 76, 5, 252, 193, 78, 5, 252, 193, 186, 5, 252, 193, 188, 5, 252, 193, 190, 5, 252, 193, 192, 5, 252, 193, 6, 6, 252, 193, 70, 6, 252, 193, 72, 6, 252, 193, 74, 6, 252, 193, 76, 6, 252, 193, 138, 6, 252, 193, 140, 6, 252, 193, 154, 6, 252, 193, 178, 6, 252, 193, 220, 6, 252, 193, 222, 6, 252, 193, 16, 7, 252, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 206, 192, 39, 0, 193, 28, 4, 252, 193, 30, 4, 252, 193, 32, 4, 252, 193, 34, 4, 252, 193, 84, 3, 252, 193, 36, 4, 252, 193, 86, 3, 252, 193, 38, 4, 252, 193, 40, 4, 252, 193, 42, 4, 252, 193, 44, 4, 252, 193, 46, 4, 252, 193, 48, 4, 252, 193, 50, 4, 252, 193, 52, 4, 252, 193, 54, 4, 252, 193, 96, 5, 252, 206, 32, 40, 0, 206, 32, 40, 0, 206, 32, 40, 0, 206, 32, 40, 0, 206, 32, 40, 0, 206, 32, 40, 0, 206, 32, 40, 0, 206, 32, 40, 0, 193, 92, 3, 252, 193, 94, 3, 252, 193, 96, 3, 252, 193, 98, 3, 252, 193, 100, 3, 252, 193, 102, 3, 252, 193, 104, 3, 252, 193, 106, 3, 252, 193, 108, 3, 252, 193, 56, 4, 252, 193, 58, 4, 252, 193, 60, 4, 252, 193, 62, 4, 252, 193, 64, 4, 252, 193, 66, 4, 252, 193, 68, 4, 252, 193, 70, 4, 252, 193, 72, 4, 252, 193, 74, 4, 252, 193, 76, 4, 252, 193, 230, 4, 252, 193, 232, 4, 252, 193, 78, 4, 252, 193, 114, 5, 252, 193, 116, 5, 252, 193, 208, 5, 252, 193, 210, 5, 252, 193, 212, 5, 252, 193, 214, 5, 252, 193, 200, 6, 252, 193, 202, 6, 252, 193, 232, 6, 252, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 206, 224, 39, 0, 193, 80, 4, 252, 193, 82, 4, 252, 193, 84, 4, 252, 193, 86, 4, 252, 193, 88, 4, 252, 193, 90, 4, 252, 193, 92, 4, 252, 193, 94, 4, 252, 193, 96, 4, 252, 193, 98, 4, 252, 193, 100, 4, 252, 193, 102, 4, 252, 193, 104, 4, 252, 193, 216, 3, 252, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 193, 250, 3, 252, 193, 252, 3, 252, 193, 254, 3, 252, 193, 2, 4, 252, 193, 4, 4, 252, 193, 6, 4, 252, 193, 8, 4, 252, 193, 10, 4, 252, 193, 12, 4, 252, 193, 178, 4, 252, 193, 180, 4, 252, 193, 182, 4, 252, 193, 184, 4, 252, 193, 186, 4, 252, 193, 188, 4, 252, 193, 190, 4, 252, 193, 192, 4, 252, 193, 194, 4, 252, 193, 196, 4, 252, 193, 198, 4, 252, 193, 80, 5, 252, 193, 82, 5, 252, 193, 84, 5, 252, 193, 86, 5, 252, 193, 194, 5, 252, 193, 196, 5, 252, 193, 198, 5, 252, 193, 8, 6, 252, 193, 10, 6, 252, 193, 12, 6, 252, 193, 14, 6, 252, 193, 16, 6, 252, 193, 18, 4, 252, 193, 20, 4, 252, 193, 22, 4, 252, 193, 24, 4, 252, 193, 26, 4, 252, 193, 88, 5, 252, 193, 90, 5, 252, 193, 92, 5, 252, 193, 94, 5, 252, 193, 166, 6, 252, 193, 22, 7, 252, 193, 24, 7, 252, 193, 84, 7, 252, 193, 230, 7, 252, 193, 232, 7, 252, 193, 60, 8, 252, 193, 62, 8, 252, 193, 64, 8, 252, 193, 148, 8, 252, 193, 150, 8, 252, 193, 152, 8, 252, 193, 154, 8, 252, 193, 226, 8, 252, 193, 228, 8, 252, 193, 30, 9, 252, 193, 106, 9, 252, 193, 220, 9, 252, 193, 240, 9, 252, 193, 242, 9, 252, 193, 52, 10, 252, 193, 54, 10, 252, 193, 56, 10, 252, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 206, 0, 40, 0, 193, 234, 4, 252, 193, 236, 4, 252, 193, 238, 4, 252, 193, 240, 4, 252, 193, 242, 4, 252, 193, 244, 4, 252, 193, 246, 4, 252, 193, 248, 4, 252, 193, 250, 4, 252, 193, 252, 4, 252, 193, 254, 4, 252, 193, 2, 5, 252, 193, 4, 5, 252, 193, 6, 5, 252, 193, 8, 5, 252, 193, 10, 5, 252, 193, 12, 5, 252, 206, 32, 40, 0, 206, 32, 40, 0, 206, 32, 40, 0, 206, 32, 40, 0, 206, 32, 40, 0, 206, 32, 40, 0, 193, 98, 5, 252, 193, 100, 5, 252, 193, 102, 5, 252, 193, 104, 5, 252, 193, 106, 5, 252, 193, 108, 5, 252, 193, 228, 4, 252, 193, 110, 5, 252, 193, 112, 5, 252, 193, 200, 5, 252, 193, 202, 5, 252, 193, 204, 5, 252, 193, 206, 5, 252, 193, 20, 6, 252, 193, 22, 6, 252, 193, 24, 6, 252, 193, 78, 6, 252, 193, 80, 6, 252, 193, 82, 6, 252, 193, 84, 6, 252, 193, 86, 6, 252, 193, 88, 6, 252, 193, 116, 6, 252, 193, 118, 6, 252, 193, 156, 6, 252, 193, 158, 6, 252, 193, 14, 5, 252, 193, 16, 5, 252, 193, 118, 5, 252, 193, 18, 5, 252, 193, 20, 5, 252, 193, 22, 5, 252, 193, 24, 5, 252, 193, 26, 5, 252, 193, 28, 5, 252, 193, 30, 5, 252, 193, 32, 5, 252, 193, 34, 5, 252, 193, 36, 5, 252, 193, 38, 5, 252, 193, 40, 5, 252, 193, 42, 5, 252, 193, 44, 5, 252, 193, 46, 5, 252, 193, 48, 5, 252, 193, 50, 5, 252, 193, 52, 5, 252, 193, 54, 5, 252, 193, 56, 5, 252, 193, 58, 5, 252, 193, 60, 5, 252, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 64, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 193, 182, 5, 252, 193, 184, 5, 252, 193, 2, 6, 252, 193, 4, 6, 252, 193, 66, 6, 252, 193, 68, 6, 252, 193, 114, 6, 252, 193, 152, 6, 252, 193, 136, 6, 252, 193, 14, 7, 252, 193, 76, 7, 252, 193, 78, 7, 252, 193, 214, 7, 252, 193, 216, 7, 252, 193, 48, 8, 252, 193, 136, 8, 252, 193, 138, 8, 252, 193, 216, 8, 252, 193, 24, 9, 252, 193, 144, 9, 252, 193, 180, 9, 252, 193, 42, 10, 252, 193, 120, 10, 252, 193, 152, 10, 252, 193, 206, 10, 252, 193, 254, 10, 252, 193, 2, 11, 252, 193, 36, 11, 252, 193, 70, 11, 252, 193, 100, 11, 252, 193, 102, 11, 252, 193, 128, 11, 252, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 206, 96, 40, 0, 193, 26, 6, 252, 193, 28, 6, 252, 193, 30, 6, 252, 193, 32, 6, 252, 193, 34, 6, 252, 193, 36, 6, 252, 193, 38, 6, 252, 193, 40, 6, 252, 193, 42, 6, 252, 193, 44, 6, 252, 193, 46, 6, 252, 193, 48, 6, 252, 193, 50, 6, 252, 193, 52, 6, 252, 193, 54, 6, 252, 193, 56, 6, 252, 193, 58, 6, 252, 193, 60, 6, 252, 193, 62, 6, 252, 193, 90, 6, 252, 193, 92, 6, 252, 193, 18, 6, 252, 193, 142, 6, 252, 193, 170, 6, 252, 193, 180, 6, 252, 193, 196, 6, 252, 193, 18, 7, 252, 193, 20, 7, 252, 193, 154, 7, 252, 193, 156, 7, 252, 193, 158, 7, 252, 193, 160, 7, 252, 193, 162, 7, 252, 193, 228, 7, 252, 193, 54, 8, 252, 193, 56, 8, 252, 193, 58, 8, 252, 193, 144, 8, 252, 193, 146, 8, 252, 193, 224, 8, 252, 193, 26, 9, 252, 193, 28, 9, 252, 193, 70, 9, 252, 193, 72, 9, 252, 193, 146, 9, 252, 193, 214, 9, 252, 193, 216, 9, 252, 193, 218, 9, 252, 193, 46, 10, 252, 193, 48, 10, 252, 193, 50, 10, 252, 193, 78, 10, 252, 193, 80, 10, 252, 193, 178, 147, 252, 193, 180, 147, 252, 193, 182, 147, 252, 193, 184, 147, 252, 193, 186, 147, 252, 193, 188, 147, 252, 193, 208, 147, 252, 193, 210, 147, 252, 193, 212, 147, 252, 193, 214, 147, 252, 193, 220, 147, 252, 193, 222, 147, 252, 193, 224, 147, 252, 193, 226, 147, 252, 193, 228, 147, 252, 193, 230, 147, 252, 193, 232, 147, 252, 193, 234, 147, 252, 193, 236, 147, 252, 193, 242, 147, 252, 193, 64, 6, 252, 193, 244, 147, 252, 193, 246, 147, 252, 193, 252, 147, 252, 193, 254, 147, 252, 193, 2, 148, 252, 193, 4, 148, 252, 193, 6, 148, 252, 193, 8, 148, 252, 193, 10, 148, 252, 193, 12, 148, 252, 193, 14, 148, 252, 193, 188, 6, 252, 193, 190, 6, 252, 193, 192, 6, 252, 193, 224, 6, 252, 193, 198, 6, 252, 193, 226, 6, 252, 193, 228, 6, 252, 193, 230, 6, 252, 193, 26, 7, 252, 193, 28, 7, 252, 193, 30, 7, 252, 193, 86, 7, 252, 193, 88, 7, 252, 193, 90, 7, 252, 193, 92, 7, 252, 193, 94, 7, 252, 193, 96, 7, 252, 193, 98, 7, 252, 193, 164, 7, 252, 193, 166, 7, 252, 193, 168, 7, 252, 193, 170, 7, 252, 193, 172, 7, 252, 193, 174, 7, 252, 193, 176, 7, 252, 193, 234, 7, 252, 193, 236, 7, 252, 193, 238, 7, 252, 193, 240, 7, 252, 193, 242, 7, 252, 193, 244, 7, 252, 193, 246, 7, 252, 193, 194, 6, 252, 193, 204, 6, 252, 193, 206, 6, 252, 193, 208, 6, 252, 193, 210, 6, 252, 193, 212, 6, 252, 193, 214, 6, 252, 193, 216, 6, 252, 193, 218, 6, 252, 193, 240, 6, 252, 193, 242, 6, 252, 193, 244, 6, 252, 193, 246, 6, 252, 193, 248, 6, 252, 193, 250, 6, 252, 193, 252, 6, 252, 193, 254, 6, 252, 193, 2, 7, 252, 193, 4, 7, 252, 193, 6, 7, 252, 193, 8, 7, 252, 193, 10, 7, 252, 193, 12, 7, 252, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 193, 234, 6, 252, 193, 236, 6, 252, 193, 238, 6, 252, 193, 32, 7, 252, 193, 34, 7, 252, 193, 36, 7, 252, 193, 38, 7, 252, 193, 40, 7, 252, 193, 42, 7, 252, 193, 100, 7, 252, 193, 102, 7, 252, 193, 104, 7, 252, 193, 106, 7, 252, 193, 108, 7, 252, 193, 178, 7, 252, 193, 180, 7, 252, 193, 182, 7, 252, 193, 184, 7, 252, 193, 248, 7, 252, 193, 250, 7, 252, 193, 252, 7, 252, 193, 254, 7, 252, 193, 2, 8, 252, 193, 148, 12, 252, 193, 4, 8, 252, 193, 84, 8, 252, 193, 86, 8, 252, 193, 88, 8, 252, 193, 90, 8, 252, 193, 92, 8, 252, 193, 94, 8, 252, 193, 166, 8, 252, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 206, 128, 40, 0, 193, 112, 7, 252, 193, 114, 7, 252, 193, 116, 7, 252, 193, 118, 7, 252, 193, 120, 7, 252, 193, 122, 7, 252, 193, 124, 7, 252, 193, 126, 7, 252, 193, 128, 7, 252, 193, 130, 7, 252, 193, 132, 7, 252, 193, 134, 7, 252, 193, 136, 7, 252, 193, 138, 7, 252, 193, 140, 7, 252, 193, 142, 7, 252, 193, 144, 7, 252, 193, 146, 7, 252, 193, 148, 7, 252, 206, 160, 40, 0, 206, 160, 40, 0, 206, 160, 40, 0, 206, 160, 40, 0, 206, 160, 40, 0, 206, 160, 40, 0, 193, 80, 7, 252, 193, 82, 7, 252, 193, 150, 7, 252, 193, 152, 7, 252, 193, 218, 7, 252, 193, 220, 7, 252, 193, 222, 7, 252, 193, 224, 7, 252, 193, 226, 7, 252, 193, 50, 8, 252, 193, 52, 8, 252, 193, 140, 8, 252, 193, 142, 8, 252, 193, 218, 8, 252, 193, 220, 8, 252, 193, 222, 8, 252, 193, 66, 9, 252, 193, 68, 9, 252, 193, 104, 9, 252, 193, 120, 9, 252, 193, 156, 9, 252, 193, 176, 9, 252, 193, 210, 9, 252, 193, 212, 9, 252, 193, 238, 9, 252, 193, 44, 10, 252, 193, 74, 10, 252, 193, 76, 10, 252, 193, 122, 10, 252, 193, 124, 10, 252, 193, 154, 10, 252, 193, 156, 10, 252, 206, 160, 40, 0, 206, 160, 40, 0, 206, 160, 40, 0, 206, 160, 40, 0, 206, 160, 40, 0, 206, 160, 40, 0, 206, 160, 40, 0, 206, 160, 40, 0, 193, 6, 8, 252, 193, 8, 8, 252, 193, 10, 8, 252, 193, 12, 8, 252, 193, 14, 8, 252, 193, 16, 8, 252, 193, 18, 8, 252, 193, 20, 8, 252, 193, 22, 8, 252, 193, 24, 8, 252, 193, 26, 8, 252, 193, 28, 8, 252, 193, 30, 8, 252, 193, 32, 8, 252, 193, 34, 8, 252, 193, 36, 8, 252, 193, 38, 8, 252, 193, 40, 8, 252, 193, 42, 8, 252, 193, 44, 8, 252, 193, 46, 8, 252, 193, 96, 8, 252, 193, 98, 8, 252, 193, 100, 8, 252, 193, 66, 8, 252, 193, 68, 8, 252, 193, 70, 8, 252, 193, 72, 8, 252, 193, 74, 8, 252, 193, 76, 8, 252, 193, 78, 8, 252, 193, 80, 8, 252, 193, 82, 8, 252, 193, 156, 8, 252, 193, 158, 8, 252, 193, 160, 8, 252, 193, 162, 8, 252, 193, 164, 8, 252, 193, 230, 8, 252, 193, 232, 8, 252, 193, 234, 8, 252, 193, 236, 8, 252, 193, 238, 8, 252, 193, 240, 8, 252, 193, 32, 9, 252, 193, 34, 9, 252, 193, 36, 9, 252, 193, 38, 9, 252, 193, 40, 9, 252, 193, 42, 9, 252, 193, 44, 9, 252, 193, 74, 9, 252, 193, 76, 9, 252, 193, 78, 9, 252, 193, 108, 9, 252, 193, 110, 9, 252, 193, 102, 8, 252, 193, 104, 8, 252, 193, 106, 8, 252, 193, 108, 8, 252, 193, 110, 8, 252, 193, 112, 8, 252, 193, 114, 8, 252, 193, 116, 8, 252, 193, 118, 8, 252, 193, 120, 8, 252, 193, 122, 8, 252, 193, 124, 8, 252, 193, 30, 50, 253, 193, 126, 8, 252, 193, 128, 8, 252, 193, 130, 8, 252, 193, 132, 8, 252, 193, 134, 8, 252, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 193, 168, 8, 252, 193, 170, 8, 252, 193, 242, 8, 252, 193, 244, 8, 252, 193, 46, 9, 252, 193, 48, 9, 252, 193, 80, 9, 252, 193, 82, 9, 252, 193, 84, 9, 252, 193, 86, 9, 252, 193, 124, 9, 252, 193, 136, 9, 252, 193, 160, 9, 252, 193, 172, 9, 252, 193, 182, 9, 252, 193, 204, 9, 252, 193, 224, 9, 252, 193, 226, 9, 252, 193, 62, 10, 252, 193, 64, 10, 252, 193, 66, 10, 252, 193, 174, 10, 252, 193, 176, 10, 252, 193, 26, 11, 252, 193, 140, 11, 252, 193, 204, 11, 252, 193, 10, 12, 252, 193, 12, 12, 252, 193, 14, 12, 252, 193, 16, 12, 252, 193, 78, 12, 252, 193, 80, 12, 252, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 206, 192, 40, 0, 193, 246, 8, 252, 193, 248, 8, 252, 193, 250, 8, 252, 193, 252, 8, 252, 193, 254, 8, 252, 193, 2, 9, 252, 193, 4, 9, 252, 193, 6, 9, 252, 193, 8, 9, 252, 193, 10, 9, 252, 193, 12, 9, 252, 193, 14, 9, 252, 193, 16, 9, 252, 193, 18, 9, 252, 193, 20, 9, 252, 193, 22, 9, 252, 193, 50, 9, 252, 193, 52, 9, 252, 193, 54, 9, 252, 193, 56, 9, 252, 193, 58, 9, 252, 193, 60, 9, 252, 193, 62, 9, 252, 193, 64, 9, 252, 193, 88, 9, 252, 193, 90, 9, 252, 193, 92, 9, 252, 193, 94, 9, 252, 193, 96, 9, 252, 193, 98, 9, 252, 193, 100, 9, 252, 193, 102, 9, 252, 193, 112, 9, 252, 193, 114, 9, 252, 193, 116, 9, 252, 193, 118, 9, 252, 193, 126, 9, 252, 193, 128, 9, 252, 193, 130, 9, 252, 193, 132, 9, 252, 193, 134, 9, 252, 193, 138, 9, 252, 193, 140, 9, 252, 193, 142, 9, 252, 193, 148, 9, 252, 193, 150, 9, 252, 193, 162, 9, 252, 193, 152, 9, 252, 193, 154, 9, 252, 193, 164, 9, 252, 193, 166, 9, 252, 193, 168, 9, 252, 193, 170, 9, 252, 193, 174, 9, 252, 193, 178, 9, 252, 193, 184, 9, 252, 193, 122, 9, 252, 193, 158, 9, 252, 193, 190, 9, 252, 193, 196, 9, 252, 193, 202, 9, 252, 193, 222, 9, 252, 193, 244, 9, 252, 193, 246, 9, 252, 193, 248, 9, 252, 193, 250, 9, 252, 193, 58, 10, 252, 193, 60, 10, 252, 193, 92, 10, 252, 193, 94, 10, 252, 193, 170, 10, 252, 193, 172, 10, 252, 193, 230, 10, 252, 193, 232, 10, 252, 193, 234, 10, 252, 193, 236, 10, 252, 193, 24, 11, 252, 193, 50, 11, 252, 193, 52, 11, 252, 193, 88, 11, 252, 193, 138, 11, 252, 193, 192, 11, 252, 193, 198, 11, 252, 193, 202, 11, 252, 193, 226, 11, 252, 193, 254, 11, 252, 193, 2, 12, 252, 193, 4, 12, 252, 193, 186, 9, 252, 193, 188, 9, 252, 193, 194, 9, 252, 193, 198, 9, 252, 193, 200, 9, 252, 193, 206, 9, 252, 193, 208, 9, 252, 193, 228, 9, 252, 193, 230, 9, 252, 193, 252, 9, 252, 193, 232, 9, 252, 193, 234, 9, 252, 193, 236, 9, 252, 193, 254, 9, 252, 193, 2, 10, 252, 193, 4, 10, 252, 193, 6, 10, 252, 193, 8, 10, 252, 193, 10, 10, 252, 193, 12, 10, 252, 193, 14, 10, 252, 193, 16, 10, 252, 193, 18, 10, 252, 193, 20, 10, 252, 193, 22, 10, 252, 193, 24, 10, 252, 193, 26, 10, 252, 193, 28, 10, 252, 193, 30, 10, 252, 193, 32, 10, 252, 193, 34, 10, 252, 193, 36, 10, 252, 193, 82, 10, 252, 193, 126, 10, 252, 193, 160, 10, 252, 193, 162, 10, 252, 193, 164, 10, 252, 193, 166, 10, 252, 193, 214, 10, 252, 193, 216, 10, 252, 193, 218, 10, 252, 193, 220, 10, 252, 193, 10, 11, 252, 193, 12, 11, 252, 193, 42, 11, 252, 193, 44, 11, 252, 193, 76, 11, 252, 193, 78, 11, 252, 193, 80, 11, 252, 193, 82, 11, 252, 193, 108, 11, 252, 193, 110, 11, 252, 193, 112, 11, 252, 193, 134, 11, 252, 193, 136, 11, 252, 193, 156, 11, 252, 193, 168, 11, 252, 193, 170, 11, 252, 193, 172, 11, 252, 193, 174, 11, 252, 193, 190, 11, 252, 193, 220, 11, 252, 193, 222, 11, 252, 193, 224, 11, 252, 193, 84, 10, 252, 193, 86, 10, 252, 193, 88, 10, 252, 193, 90, 10, 252, 193, 128, 10, 252, 193, 130, 10, 252, 193, 132, 10, 252, 193, 134, 10, 252, 193, 136, 10, 252, 193, 138, 10, 252, 193, 168, 10, 252, 193, 222, 10, 252, 193, 224, 10, 252, 193, 226, 10, 252, 193, 228, 10, 252, 193, 14, 11, 252, 193, 16, 11, 252, 193, 18, 11, 252, 193, 20, 11, 252, 193, 22, 11, 252, 193, 46, 11, 252, 193, 48, 11, 252, 193, 84, 11, 252, 193, 86, 11, 252, 193, 114, 11, 252, 193, 116, 11, 252, 193, 164, 11, 252, 193, 176, 11, 252, 193, 252, 11, 252, 193, 218, 12, 252, 193, 220, 12, 252, 193, 38, 13, 252, 193, 158, 10, 252, 193, 208, 10, 252, 193, 210, 10, 252, 193, 212, 10, 252, 193, 6, 11, 252, 193, 8, 11, 252, 193, 38, 11, 252, 193, 40, 11, 252, 193, 72, 11, 252, 193, 74, 11, 252, 193, 104, 11, 252, 193, 106, 11, 252, 193, 130, 11, 252, 193, 132, 11, 252, 193, 154, 11, 252, 193, 160, 11, 252, 193, 162, 11, 252, 193, 130, 12, 252, 193, 132, 12, 252, 193, 194, 12, 252, 193, 196, 12, 252, 193, 198, 12, 252, 193, 24, 13, 252, 193, 26, 13, 252, 193, 124, 13, 252, 193, 126, 13, 252, 193, 128, 13, 252, 193, 186, 13, 252, 193, 140, 14, 252, 193, 154, 14, 252, 193, 178, 14, 252, 193, 190, 14, 252, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 193, 126, 70, 252, 193, 128, 70, 252, 193, 130, 70, 252, 193, 132, 70, 252, 193, 134, 70, 252, 193, 136, 70, 252, 193, 138, 70, 252, 193, 140, 70, 252, 193, 142, 70, 252, 193, 144, 70, 252, 193, 146, 70, 252, 193, 148, 70, 252, 193, 150, 70, 252, 193, 152, 70, 252, 193, 154, 70, 252, 193, 252, 10, 252, 193, 156, 70, 252, 193, 158, 70, 252, 193, 160, 70, 252, 193, 230, 11, 252, 193, 232, 11, 252, 193, 234, 11, 252, 193, 236, 11, 252, 193, 238, 11, 252, 193, 240, 11, 252, 193, 242, 11, 252, 193, 244, 11, 252, 193, 246, 11, 252, 193, 18, 12, 252, 193, 20, 12, 252, 193, 22, 12, 252, 193, 24, 12, 252, 193, 26, 12, 252, 193, 28, 12, 252, 193, 30, 12, 252, 193, 32, 12, 252, 193, 34, 12, 252, 193, 36, 12, 252, 193, 38, 12, 252, 193, 40, 12, 252, 193, 42, 12, 252, 193, 44, 12, 252, 193, 46, 12, 252, 193, 48, 12, 252, 193, 50, 12, 252, 193, 52, 12, 252, 193, 54, 12, 252, 193, 56, 12, 252, 193, 84, 12, 252, 193, 86, 12, 252, 193, 88, 12, 252, 193, 248, 11, 252, 193, 250, 11, 252, 193, 58, 12, 252, 193, 60, 12, 252, 193, 62, 12, 252, 193, 64, 12, 252, 193, 134, 12, 252, 193, 136, 12, 252, 193, 138, 12, 252, 193, 200, 12, 252, 193, 202, 12, 252, 193, 204, 12, 252, 193, 206, 12, 252, 193, 208, 12, 252, 193, 210, 12, 252, 193, 212, 12, 252, 193, 214, 12, 252, 193, 216, 12, 252, 193, 28, 13, 252, 193, 30, 13, 252, 193, 32, 13, 252, 193, 34, 13, 252, 193, 36, 13, 252, 193, 130, 13, 252, 193, 132, 13, 252, 193, 188, 13, 252, 193, 190, 13, 252, 193, 192, 13, 252, 193, 232, 13, 252, 193, 28, 14, 252, 193, 30, 14, 252, 193, 72, 14, 252, 193, 6, 12, 252, 193, 8, 12, 252, 193, 66, 12, 252, 193, 68, 12, 252, 193, 70, 12, 252, 193, 72, 12, 252, 193, 74, 12, 252, 193, 76, 12, 252, 193, 140, 12, 252, 193, 142, 12, 252, 193, 144, 12, 252, 193, 146, 12, 252, 193, 222, 12, 252, 193, 224, 12, 252, 193, 226, 12, 252, 193, 228, 12, 252, 193, 230, 12, 252, 193, 232, 12, 252, 193, 234, 12, 252, 193, 236, 12, 252, 193, 42, 13, 252, 193, 44, 13, 252, 193, 46, 13, 252, 193, 48, 13, 252, 193, 50, 13, 252, 193, 52, 13, 252, 193, 54, 13, 252, 193, 134, 13, 252, 193, 136, 13, 252, 193, 194, 13, 252, 193, 196, 13, 252, 193, 198, 13, 252, 193, 82, 12, 252, 193, 150, 12, 252, 193, 238, 12, 252, 193, 240, 12, 252, 193, 56, 13, 252, 193, 58, 13, 252, 193, 60, 13, 252, 193, 62, 13, 252, 193, 64, 13, 252, 193, 66, 13, 252, 193, 138, 13, 252, 193, 140, 13, 252, 193, 142, 13, 252, 193, 144, 13, 252, 193, 146, 13, 252, 193, 204, 13, 252, 193, 206, 13, 252, 193, 208, 13, 252, 193, 210, 13, 252, 193, 244, 13, 252, 193, 246, 13, 252, 193, 44, 14, 252, 193, 46, 14, 252, 193, 48, 14, 252, 193, 88, 14, 252, 193, 144, 14, 252, 193, 146, 14, 252, 193, 166, 14, 252, 193, 176, 14, 252, 193, 182, 14, 252, 193, 212, 14, 252, 193, 214, 14, 252, 193, 90, 12, 252, 193, 92, 12, 252, 193, 94, 12, 252, 193, 96, 12, 252, 193, 98, 12, 252, 193, 100, 12, 252, 193, 102, 12, 252, 193, 104, 12, 252, 193, 106, 12, 252, 193, 108, 12, 252, 193, 110, 12, 252, 193, 112, 12, 252, 193, 114, 12, 252, 193, 116, 12, 252, 193, 118, 12, 252, 193, 120, 12, 252, 193, 122, 12, 252, 193, 124, 12, 252, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 206, 0, 41, 0, 193, 242, 12, 252, 193, 244, 12, 252, 193, 246, 12, 252, 193, 248, 12, 252, 193, 250, 12, 252, 193, 252, 12, 252, 193, 254, 12, 252, 193, 2, 13, 252, 193, 4, 13, 252, 193, 6, 13, 252, 193, 8, 13, 252, 193, 10, 13, 252, 193, 12, 13, 252, 193, 14, 13, 252, 193, 16, 13, 252, 193, 18, 13, 252, 193, 20, 13, 252, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 193, 40, 13, 252, 193, 32, 14, 252, 193, 156, 14, 252, 193, 134, 15, 252, 193, 136, 15, 252, 193, 138, 15, 252, 193, 198, 15, 252, 193, 200, 15, 252, 193, 2, 16, 252, 193, 22, 16, 252, 193, 60, 16, 252, 193, 230, 16, 252, 193, 242, 16, 252, 193, 62, 17, 252, 193, 120, 17, 252, 193, 122, 17, 252, 193, 124, 17, 252, 193, 126, 17, 252, 193, 222, 17, 252, 193, 224, 17, 252, 193, 226, 17, 252, 193, 228, 17, 252, 193, 44, 18, 252, 193, 46, 18, 252, 193, 122, 18, 252, 193, 124, 18, 252, 193, 126, 18, 252, 193, 128, 18, 252, 193, 130, 18, 252, 193, 132, 18, 252, 193, 232, 18, 252, 193, 234, 18, 252, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 32, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 206, 64, 41, 0, 193, 212, 13, 252, 193, 214, 13, 252, 193, 216, 13, 252, 193, 218, 13, 252, 193, 220, 13, 252, 193, 222, 13, 252, 193, 224, 13, 252, 193, 226, 13, 252, 193, 228, 13, 252, 193, 248, 13, 252, 193, 250, 13, 252, 193, 252, 13, 252, 193, 254, 13, 252, 193, 2, 14, 252, 193, 4, 14, 252, 193, 6, 14, 252, 193, 8, 14, 252, 193, 10, 14, 252, 193, 12, 14, 252, 193, 14, 14, 252, 193, 16, 14, 252, 193, 18, 14, 252, 193, 20, 14, 252, 193, 22, 14, 252, 193, 24, 14, 252, 193, 50, 14, 252, 193, 52, 14, 252, 193, 54, 14, 252, 193, 200, 13, 252, 193, 202, 13, 252, 193, 234, 13, 252, 193, 236, 13, 252, 193, 238, 13, 252, 193, 240, 13, 252, 193, 242, 13, 252, 193, 34, 14, 252, 193, 36, 14, 252, 193, 38, 14, 252, 193, 40, 14, 252, 193, 42, 14, 252, 193, 82, 14, 252, 193, 84, 14, 252, 193, 86, 14, 252, 193, 106, 14, 252, 193, 132, 14, 252, 193, 134, 14, 252, 193, 158, 14, 252, 193, 174, 14, 252, 193, 180, 14, 252, 193, 196, 14, 252, 193, 198, 14, 252, 193, 210, 14, 252, 193, 240, 14, 252, 193, 242, 14, 252, 193, 244, 14, 252, 193, 246, 14, 252, 193, 40, 15, 252, 193, 42, 15, 252, 193, 82, 15, 252, 193, 84, 15, 252, 193, 56, 14, 252, 193, 58, 14, 252, 193, 60, 14, 252, 193, 62, 14, 252, 193, 64, 14, 252, 193, 66, 14, 252, 193, 68, 14, 252, 193, 70, 14, 252, 193, 90, 14, 252, 193, 92, 14, 252, 193, 94, 14, 252, 193, 96, 14, 252, 193, 98, 14, 252, 193, 100, 14, 252, 193, 102, 14, 252, 193, 108, 14, 252, 193, 110, 14, 252, 193, 112, 14, 252, 193, 114, 14, 252, 193, 116, 14, 252, 193, 118, 14, 252, 193, 120, 14, 252, 193, 122, 14, 252, 193, 136, 14, 252, 193, 138, 14, 252, 193, 148, 14, 252, 193, 150, 14, 252, 193, 152, 14, 252, 193, 160, 14, 252, 193, 162, 14, 252, 193, 164, 14, 252, 193, 168, 14, 252, 193, 74, 14, 252, 193, 76, 14, 252, 193, 78, 14, 252, 193, 80, 14, 252, 193, 124, 14, 252, 193, 126, 14, 252, 193, 128, 14, 252, 193, 130, 14, 252, 193, 194, 14, 252, 193, 230, 14, 252, 193, 232, 14, 252, 193, 234, 14, 252, 193, 236, 14, 252, 193, 238, 14, 252, 193, 32, 15, 252, 193, 34, 15, 252, 193, 36, 15, 252, 193, 38, 15, 252, 193, 72, 15, 252, 193, 74, 15, 252, 193, 76, 15, 252, 193, 78, 15, 252, 193, 80, 15, 252, 193, 130, 15, 252, 193, 132, 15, 252, 193, 164, 15, 252, 193, 166, 15, 252, 193, 168, 15, 252, 193, 194, 15, 252, 193, 196, 15, 252, 193, 248, 15, 252, 193, 250, 15, 252, 193, 170, 14, 252, 193, 172, 14, 252, 193, 184, 14, 252, 193, 186, 14, 252, 193, 188, 14, 252, 193, 200, 14, 252, 193, 202, 14, 252, 193, 204, 14, 252, 193, 216, 14, 252, 193, 218, 14, 252, 193, 220, 14, 252, 193, 250, 14, 252, 193, 252, 14, 252, 193, 254, 14, 252, 193, 2, 15, 252, 193, 4, 15, 252, 193, 6, 15, 252, 193, 8, 15, 252, 193, 10, 15, 252, 193, 12, 15, 252, 193, 14, 15, 252, 193, 16, 15, 252, 193, 18, 15, 252, 193, 20, 15, 252, 193, 22, 15, 252, 193, 50, 15, 252, 193, 52, 15, 252, 193, 54, 15, 252, 193, 56, 15, 252, 193, 58, 15, 252, 193, 60, 15, 252, 193, 62, 15, 252, 193, 192, 14, 252, 193, 206, 14, 252, 193, 208, 14, 252, 193, 226, 14, 252, 193, 228, 14, 252, 193, 26, 15, 252, 193, 28, 15, 252, 193, 30, 15, 252, 193, 70, 15, 252, 193, 122, 15, 252, 193, 124, 15, 252, 193, 126, 15, 252, 193, 128, 15, 252, 193, 162, 15, 252, 193, 188, 15, 252, 193, 190, 15, 252, 193, 192, 15, 252, 193, 246, 15, 252, 193, 18, 16, 252, 193, 20, 16, 252, 193, 72, 16, 252, 193, 102, 16, 252, 193, 136, 16, 252, 193, 152, 16, 252, 193, 174, 16, 252, 193, 232, 16, 252, 193, 240, 16, 252, 193, 12, 17, 252, 193, 60, 17, 252, 193, 112, 17, 252, 193, 114, 17, 252, 193, 116, 17, 252, 193, 248, 14, 252, 193, 44, 15, 252, 193, 46, 15, 252, 193, 48, 15, 252, 193, 92, 15, 252, 193, 144, 15, 252, 193, 206, 15, 252, 193, 208, 15, 252, 193, 210, 15, 252, 193, 4, 16, 252, 193, 24, 16, 252, 193, 204, 16, 252, 193, 28, 17, 252, 193, 68, 17, 252, 193, 70, 17, 252, 193, 72, 17, 252, 193, 74, 17, 252, 193, 76, 17, 252, 193, 132, 17, 252, 193, 248, 17, 252, 193, 56, 18, 252, 193, 58, 18, 252, 193, 60, 18, 252, 193, 62, 18, 252, 193, 64, 18, 252, 193, 66, 18, 252, 193, 68, 18, 252, 193, 150, 18, 252, 193, 152, 18, 252, 193, 154, 18, 252, 193, 156, 18, 252, 193, 4, 19, 252, 193, 86, 15, 252, 193, 88, 15, 252, 193, 90, 15, 252, 193, 140, 15, 252, 193, 142, 15, 252, 193, 170, 15, 252, 193, 172, 15, 252, 193, 202, 15, 252, 193, 204, 15, 252, 193, 44, 16, 252, 193, 74, 16, 252, 193, 104, 16, 252, 193, 106, 16, 252, 193, 110, 16, 252, 193, 112, 16, 252, 193, 118, 16, 252, 193, 120, 16, 252, 193, 124, 16, 252, 193, 126, 16, 252, 193, 154, 16, 252, 193, 180, 16, 252, 193, 182, 16, 252, 193, 198, 16, 252, 193, 200, 16, 252, 193, 202, 16, 252, 193, 208, 16, 252, 193, 6, 17, 252, 193, 22, 17, 252, 193, 40, 17, 252, 193, 42, 17, 252, 193, 64, 17, 252, 193, 66, 17, 252, 193, 94, 15, 252, 193, 96, 15, 252, 193, 98, 15, 252, 193, 100, 15, 252, 193, 102, 15, 252, 193, 104, 15, 252, 193, 106, 15, 252, 193, 108, 15, 252, 193, 110, 15, 252, 193, 112, 15, 252, 193, 114, 15, 252, 193, 116, 15, 252, 193, 118, 15, 252, 193, 152, 117, 252, 193, 146, 15, 252, 193, 148, 15, 252, 193, 150, 15, 252, 193, 152, 15, 252, 193, 154, 15, 252, 193, 156, 15, 252, 193, 158, 15, 252, 193, 160, 15, 252, 193, 174, 15, 252, 193, 176, 15, 252, 193, 178, 15, 252, 193, 180, 15, 252, 193, 182, 15, 252, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 193, 4, 71, 252, 193, 6, 71, 252, 193, 8, 71, 252, 193, 10, 71, 252, 193, 12, 71, 252, 193, 14, 71, 252, 193, 16, 71, 252, 193, 18, 71, 252, 193, 20, 71, 252, 193, 22, 71, 252, 193, 60, 71, 252, 193, 62, 71, 252, 193, 64, 71, 252, 193, 66, 71, 252, 193, 68, 71, 252, 193, 70, 71, 252, 193, 72, 71, 252, 193, 74, 71, 252, 193, 76, 71, 252, 193, 78, 71, 252, 193, 80, 71, 252, 193, 82, 71, 252, 193, 84, 71, 252, 193, 184, 15, 252, 193, 86, 71, 252, 193, 88, 71, 252, 193, 90, 71, 252, 193, 92, 71, 252, 193, 94, 71, 252, 193, 96, 71, 252, 193, 98, 71, 252, 193, 100, 71, 252, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 206, 96, 41, 0, 193, 6, 16, 252, 193, 8, 16, 252, 193, 10, 16, 252, 193, 12, 16, 252, 193, 26, 16, 252, 193, 28, 16, 252, 193, 30, 16, 252, 193, 32, 16, 252, 193, 34, 16, 252, 193, 36, 16, 252, 193, 38, 16, 252, 193, 40, 16, 252, 193, 46, 16, 252, 193, 48, 16, 252, 193, 50, 16, 252, 193, 52, 16, 252, 193, 54, 16, 252, 193, 56, 16, 252, 193, 62, 16, 252, 193, 64, 16, 252, 193, 252, 15, 252, 193, 254, 15, 252, 193, 42, 16, 252, 193, 58, 16, 252, 193, 66, 16, 252, 193, 90, 16, 252, 193, 96, 16, 252, 193, 116, 16, 252, 193, 138, 16, 252, 193, 166, 16, 252, 193, 206, 16, 252, 193, 216, 16, 252, 193, 218, 16, 252, 193, 220, 16, 252, 193, 222, 16, 252, 193, 228, 16, 252, 193, 234, 16, 252, 193, 236, 16, 252, 193, 118, 17, 252, 193, 212, 17, 252, 193, 214, 17, 252, 193, 216, 17, 252, 193, 218, 17, 252, 193, 220, 17, 252, 193, 42, 18, 252, 193, 118, 18, 252, 193, 120, 18, 252, 193, 226, 18, 252, 193, 228, 18, 252, 193, 230, 18, 252, 193, 76, 19, 252, 193, 78, 19, 252, 193, 68, 16, 252, 193, 70, 16, 252, 193, 76, 16, 252, 193, 78, 16, 252, 193, 80, 16, 252, 193, 82, 16, 252, 193, 84, 16, 252, 193, 86, 16, 252, 193, 88, 16, 252, 193, 92, 16, 252, 193, 98, 16, 252, 193, 100, 16, 252, 193, 108, 16, 252, 193, 114, 16, 252, 193, 122, 16, 252, 193, 128, 16, 252, 193, 130, 16, 252, 193, 132, 16, 252, 193, 134, 16, 252, 193, 140, 16, 252, 193, 142, 16, 252, 193, 144, 16, 252, 193, 146, 16, 252, 193, 148, 16, 252, 193, 150, 16, 252, 193, 156, 16, 252, 193, 158, 16, 252, 193, 160, 16, 252, 193, 162, 16, 252, 193, 168, 16, 252, 193, 170, 16, 252, 193, 172, 16, 252, 193, 176, 16, 252, 193, 178, 16, 252, 193, 186, 16, 252, 193, 188, 16, 252, 193, 190, 16, 252, 193, 192, 16, 252, 193, 194, 16, 252, 193, 196, 16, 252, 193, 210, 16, 252, 193, 212, 16, 252, 193, 214, 16, 252, 193, 224, 16, 252, 193, 226, 16, 252, 193, 238, 16, 252, 193, 244, 16, 252, 193, 246, 16, 252, 193, 248, 16, 252, 193, 250, 16, 252, 193, 254, 16, 252, 193, 2, 17, 252, 193, 4, 17, 252, 193, 8, 17, 252, 193, 10, 17, 252, 193, 14, 17, 252, 193, 16, 17, 252, 193, 18, 17, 252, 193, 20, 17, 252, 193, 24, 17, 252, 193, 30, 17, 252, 193, 32, 17, 252, 193, 34, 17, 252, 193, 36, 17, 252, 193, 38, 17, 252, 193, 44, 17, 252, 193, 46, 17, 252, 193, 48, 17, 252, 193, 50, 17, 252, 193, 52, 17, 252, 193, 54, 17, 252, 193, 56, 17, 252, 193, 58, 17, 252, 193, 78, 17, 252, 193, 80, 17, 252, 193, 82, 17, 252, 193, 84, 17, 252, 193, 86, 17, 252, 193, 88, 17, 252, 193, 90, 17, 252, 193, 92, 17, 252, 193, 94, 17, 252, 193, 96, 17, 252, 193, 98, 17, 252, 193, 100, 17, 252, 193, 102, 17, 252, 193, 104, 17, 252, 193, 106, 17, 252, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 193, 128, 17, 252, 193, 130, 17, 252, 193, 230, 17, 252, 193, 232, 17, 252, 193, 234, 17, 252, 193, 236, 17, 252, 193, 238, 17, 252, 193, 240, 17, 252, 193, 242, 17, 252, 193, 244, 17, 252, 193, 246, 17, 252, 193, 48, 18, 252, 193, 50, 18, 252, 193, 52, 18, 252, 193, 54, 18, 252, 193, 134, 18, 252, 193, 136, 18, 252, 193, 138, 18, 252, 193, 140, 18, 252, 193, 142, 18, 252, 193, 144, 18, 252, 193, 146, 18, 252, 193, 148, 18, 252, 193, 244, 18, 252, 193, 246, 18, 252, 193, 248, 18, 252, 193, 250, 18, 252, 193, 252, 18, 252, 193, 254, 18, 252, 193, 2, 19, 252, 193, 86, 19, 252, 193, 188, 19, 252, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 128, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 193, 210, 17, 252, 193, 116, 18, 252, 193, 74, 19, 252, 193, 162, 19, 252, 193, 164, 19, 252, 193, 240, 19, 252, 193, 50, 20, 252, 193, 98, 20, 252, 193, 150, 20, 252, 193, 250, 20, 252, 193, 4, 21, 252, 193, 6, 21, 252, 193, 18, 21, 252, 193, 24, 21, 252, 193, 42, 21, 252, 193, 94, 21, 252, 193, 124, 21, 252, 193, 146, 21, 252, 193, 200, 21, 252, 193, 220, 21, 252, 193, 162, 22, 252, 193, 164, 22, 252, 193, 166, 22, 252, 193, 168, 22, 252, 193, 172, 23, 252, 193, 174, 23, 252, 193, 176, 23, 252, 193, 244, 24, 252, 193, 246, 24, 252, 193, 248, 24, 252, 193, 250, 24, 252, 193, 252, 24, 252, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 160, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 192, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 224, 41, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 193, 236, 18, 252, 193, 238, 18, 252, 193, 240, 18, 252, 193, 242, 18, 252, 193, 80, 19, 252, 193, 82, 19, 252, 193, 84, 19, 252, 193, 172, 19, 252, 193, 174, 19, 252, 193, 176, 19, 252, 193, 178, 19, 252, 193, 180, 19, 252, 193, 182, 19, 252, 193, 184, 19, 252, 193, 186, 19, 252, 193, 242, 19, 252, 193, 244, 19, 252, 193, 246, 19, 252, 193, 58, 20, 252, 193, 154, 20, 252, 193, 156, 20, 252, 193, 176, 20, 252, 193, 214, 20, 252, 193, 248, 20, 252, 193, 20, 21, 252, 193, 26, 21, 252, 193, 48, 21, 252, 193, 68, 21, 252, 193, 98, 21, 252, 193, 100, 21, 252, 193, 126, 21, 252, 193, 128, 21, 252, 193, 6, 19, 252, 193, 8, 19, 252, 193, 10, 19, 252, 193, 12, 19, 252, 193, 88, 19, 252, 193, 90, 19, 252, 193, 92, 19, 252, 193, 94, 19, 252, 193, 96, 19, 252, 193, 98, 19, 252, 193, 4, 20, 252, 193, 6, 20, 252, 193, 8, 20, 252, 193, 10, 20, 252, 193, 12, 20, 252, 193, 110, 20, 252, 193, 112, 20, 252, 193, 160, 20, 252, 193, 226, 20, 252, 193, 22, 21, 252, 193, 84, 21, 252, 193, 156, 21, 252, 193, 202, 21, 252, 193, 210, 21, 252, 193, 250, 21, 252, 193, 34, 22, 252, 193, 36, 22, 252, 193, 38, 22, 252, 193, 40, 22, 252, 193, 120, 22, 252, 193, 122, 22, 252, 193, 124, 22, 252, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 206, 0, 42, 0, 193, 100, 19, 252, 193, 102, 19, 252, 193, 104, 19, 252, 193, 106, 19, 252, 193, 108, 19, 252, 193, 110, 19, 252, 193, 112, 19, 252, 193, 114, 19, 252, 193, 116, 19, 252, 193, 118, 19, 252, 193, 120, 19, 252, 193, 122, 19, 252, 193, 124, 19, 252, 193, 126, 19, 252, 193, 128, 19, 252, 193, 130, 19, 252, 193, 132, 19, 252, 193, 134, 19, 252, 193, 136, 19, 252, 193, 138, 19, 252, 193, 140, 19, 252, 193, 142, 19, 252, 193, 144, 19, 252, 193, 146, 19, 252, 193, 148, 19, 252, 193, 150, 19, 252, 193, 152, 19, 252, 193, 154, 19, 252, 193, 156, 19, 252, 193, 158, 19, 252, 193, 196, 19, 252, 193, 198, 19, 252, 193, 200, 19, 252, 193, 202, 19, 252, 193, 204, 19, 252, 193, 206, 19, 252, 193, 208, 19, 252, 193, 210, 19, 252, 193, 212, 19, 252, 193, 214, 19, 252, 193, 216, 19, 252, 193, 218, 19, 252, 193, 220, 19, 252, 193, 222, 19, 252, 193, 224, 19, 252, 193, 226, 19, 252, 193, 228, 19, 252, 193, 230, 19, 252, 193, 232, 19, 252, 193, 234, 19, 252, 193, 236, 19, 252, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 193, 160, 19, 252, 193, 238, 19, 252, 193, 148, 20, 252, 193, 92, 21, 252, 193, 122, 21, 252, 193, 160, 21, 252, 193, 190, 21, 252, 193, 198, 21, 252, 193, 64, 22, 252, 193, 154, 22, 252, 193, 156, 22, 252, 193, 158, 22, 252, 193, 160, 22, 252, 193, 164, 23, 252, 193, 166, 23, 252, 193, 168, 23, 252, 193, 170, 23, 252, 193, 232, 24, 252, 193, 234, 24, 252, 193, 236, 24, 252, 193, 238, 24, 252, 193, 240, 24, 252, 193, 242, 24, 252, 193, 116, 26, 252, 193, 118, 26, 252, 193, 120, 26, 252, 193, 122, 26, 252, 193, 124, 26, 252, 193, 126, 26, 252, 193, 128, 26, 252, 193, 66, 28, 252, 193, 68, 28, 252, 193, 166, 19, 252, 193, 168, 19, 252, 193, 170, 19, 252, 193, 52, 20, 252, 193, 54, 20, 252, 193, 56, 20, 252, 193, 100, 20, 252, 193, 102, 20, 252, 193, 104, 20, 252, 193, 106, 20, 252, 193, 152, 20, 252, 193, 174, 20, 252, 193, 190, 20, 252, 193, 212, 20, 252, 193, 222, 20, 252, 193, 232, 20, 252, 193, 8, 21, 252, 193, 36, 21, 252, 193, 96, 21, 252, 193, 208, 21, 252, 193, 212, 21, 252, 193, 224, 21, 252, 193, 170, 22, 252, 193, 172, 22, 252, 193, 178, 23, 252, 193, 180, 23, 252, 193, 182, 23, 252, 193, 184, 23, 252, 193, 186, 23, 252, 193, 254, 24, 252, 193, 2, 25, 252, 193, 4, 25, 252, 193, 190, 19, 252, 193, 192, 19, 252, 193, 194, 19, 252, 193, 248, 19, 252, 193, 250, 19, 252, 193, 252, 19, 252, 193, 254, 19, 252, 193, 2, 20, 252, 193, 60, 20, 252, 193, 62, 20, 252, 193, 64, 20, 252, 193, 108, 20, 252, 193, 158, 20, 252, 193, 202, 20, 252, 193, 224, 20, 252, 193, 236, 20, 252, 193, 238, 20, 252, 193, 240, 20, 252, 193, 244, 20, 252, 193, 252, 20, 252, 193, 54, 21, 252, 193, 56, 21, 252, 193, 58, 21, 252, 193, 60, 21, 252, 193, 62, 21, 252, 193, 70, 21, 252, 193, 72, 21, 252, 193, 78, 21, 252, 193, 80, 21, 252, 193, 82, 21, 252, 193, 102, 21, 252, 193, 104, 21, 252, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 206, 32, 42, 0, 193, 66, 20, 252, 193, 68, 20, 252, 193, 70, 20, 252, 193, 72, 20, 252, 193, 74, 20, 252, 193, 76, 20, 252, 193, 78, 20, 252, 193, 80, 20, 252, 193, 82, 20, 252, 193, 84, 20, 252, 193, 86, 20, 252, 193, 88, 20, 252, 193, 90, 20, 252, 193, 92, 20, 252, 193, 94, 20, 252, 193, 96, 20, 252, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 206, 64, 42, 0, 193, 162, 20, 252, 193, 164, 20, 252, 193, 166, 20, 252, 193, 168, 20, 252, 193, 170, 20, 252, 193, 172, 20, 252, 193, 180, 20, 252, 193, 182, 20, 252, 193, 184, 20, 252, 193, 186, 20, 252, 193, 188, 20, 252, 193, 192, 20, 252, 193, 194, 20, 252, 193, 196, 20, 252, 193, 198, 20, 252, 193, 200, 20, 252, 193, 204, 20, 252, 193, 206, 20, 252, 193, 208, 20, 252, 193, 210, 20, 252, 193, 216, 20, 252, 193, 218, 20, 252, 193, 220, 20, 252, 193, 228, 20, 252, 193, 230, 20, 252, 193, 234, 20, 252, 193, 254, 20, 252, 193, 2, 21, 252, 193, 10, 21, 252, 193, 12, 21, 252, 193, 14, 21, 252, 193, 16, 21, 252, 193, 28, 21, 252, 193, 30, 21, 252, 193, 32, 21, 252, 193, 34, 21, 252, 193, 38, 21, 252, 193, 40, 21, 252, 193, 44, 21, 252, 193, 46, 21, 252, 193, 50, 21, 252, 193, 52, 21, 252, 193, 64, 21, 252, 193, 66, 21, 252, 193, 74, 21, 252, 193, 76, 21, 252, 193, 86, 21, 252, 193, 88, 21, 252, 193, 90, 21, 252, 193, 112, 21, 252, 193, 114, 21, 252, 193, 116, 21, 252, 193, 118, 21, 252, 193, 120, 21, 252, 193, 144, 21, 252, 193, 174, 21, 252, 193, 158, 21, 252, 193, 176, 21, 252, 193, 178, 21, 252, 193, 184, 21, 252, 193, 186, 21, 252, 193, 188, 21, 252, 193, 196, 21, 252, 193, 204, 21, 252, 193, 206, 21, 252, 193, 216, 21, 252, 193, 218, 21, 252, 193, 222, 21, 252, 193, 226, 21, 252, 193, 228, 21, 252, 193, 236, 21, 252, 193, 252, 21, 252, 193, 254, 21, 252, 193, 2, 22, 252, 193, 42, 22, 252, 193, 44, 22, 252, 193, 46, 22, 252, 193, 48, 22, 252, 193, 50, 22, 252, 193, 52, 22, 252, 193, 54, 22, 252, 193, 56, 22, 252, 193, 58, 22, 252, 193, 60, 22, 252, 193, 62, 22, 252, 193, 148, 21, 252, 193, 150, 21, 252, 193, 162, 21, 252, 193, 180, 21, 252, 193, 238, 21, 252, 193, 240, 21, 252, 193, 4, 22, 252, 193, 174, 22, 252, 193, 176, 22, 252, 193, 178, 22, 252, 193, 180, 22, 252, 193, 182, 22, 252, 193, 188, 23, 252, 193, 190, 23, 252, 193, 192, 23, 252, 193, 194, 23, 252, 193, 196, 23, 252, 193, 198, 23, 252, 193, 30, 25, 252, 193, 32, 25, 252, 193, 34, 25, 252, 193, 36, 25, 252, 193, 38, 25, 252, 193, 40, 25, 252, 193, 42, 25, 252, 193, 44, 25, 252, 193, 46, 25, 252, 193, 48, 25, 252, 193, 50, 25, 252, 193, 52, 25, 252, 193, 166, 26, 252, 193, 168, 26, 252, 193, 14, 22, 252, 193, 16, 22, 252, 193, 18, 22, 252, 193, 20, 22, 252, 193, 22, 22, 252, 193, 24, 22, 252, 193, 26, 22, 252, 193, 28, 22, 252, 193, 30, 22, 252, 193, 32, 22, 252, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 96, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 193, 126, 22, 252, 193, 128, 22, 252, 193, 130, 22, 252, 193, 132, 22, 252, 193, 134, 22, 252, 193, 136, 22, 252, 193, 138, 22, 252, 193, 140, 22, 252, 193, 142, 22, 252, 193, 144, 22, 252, 193, 146, 22, 252, 193, 148, 22, 252, 193, 150, 22, 252, 193, 152, 22, 252, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 128, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 193, 70, 23, 252, 193, 72, 23, 252, 193, 74, 23, 252, 193, 76, 23, 252, 193, 78, 23, 252, 193, 80, 23, 252, 193, 82, 23, 252, 193, 84, 23, 252, 193, 86, 23, 252, 193, 88, 23, 252, 193, 90, 23, 252, 193, 110, 24, 252, 193, 112, 24, 252, 193, 114, 24, 252, 193, 116, 24, 252, 193, 118, 24, 252, 193, 120, 24, 252, 193, 122, 24, 252, 193, 124, 24, 252, 193, 126, 24, 252, 193, 128, 24, 252, 193, 130, 24, 252, 206, 64, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 160, 42, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 206, 192, 42, 0, 193, 54, 25, 252, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 206, 224, 42, 0, 193, 226, 125, 252, 193, 92, 24, 252, 193, 94, 24, 252, 193, 96, 24, 252, 193, 98, 24, 252, 193, 100, 24, 252, 193, 102, 24, 252, 193, 104, 24, 252, 193, 106, 24, 252, 193, 108, 24, 252, 193, 56, 25, 252, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 0, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 193, 6, 25, 252, 193, 8, 25, 252, 193, 10, 25, 252, 193, 12, 25, 252, 193, 14, 25, 252, 193, 16, 25, 252, 193, 18, 25, 252, 193, 20, 25, 252, 193, 22, 25, 252, 193, 24, 25, 252, 193, 26, 25, 252, 193, 28, 25, 252, 193, 144, 26, 252, 193, 146, 26, 252, 193, 148, 26, 252, 193, 150, 26, 252, 193, 152, 26, 252, 193, 154, 26, 252, 193, 156, 26, 252, 193, 158, 26, 252, 193, 160, 26, 252, 193, 162, 26, 252, 193, 164, 26, 252, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 193, 58, 25, 252, 193, 60, 25, 252, 193, 62, 25, 252, 193, 64, 25, 252, 193, 66, 25, 252, 193, 68, 25, 252, 193, 70, 25, 252, 193, 72, 25, 252, 193, 74, 25, 252, 193, 76, 25, 252, 193, 78, 25, 252, 193, 80, 25, 252, 193, 82, 25, 252, 193, 192, 26, 252, 193, 84, 25, 252, 193, 86, 25, 252, 193, 88, 25, 252, 193, 90, 25, 252, 193, 92, 25, 252, 193, 94, 25, 252, 193, 96, 25, 252, 193, 98, 25, 252, 193, 194, 26, 252, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 32, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 206, 64, 43, 0, 193, 108, 27, 252, 193, 110, 27, 252, 193, 112, 27, 252, 193, 114, 27, 252, 193, 116, 27, 252, 193, 118, 27, 252, 193, 120, 27, 252, 193, 122, 27, 252, 193, 124, 27, 252, 193, 126, 27, 252, 193, 128, 27, 252, 193, 130, 27, 252, 193, 132, 27, 252, 193, 134, 27, 252, 193, 136, 27, 252, 193, 138, 27, 252, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 96, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 193, 130, 26, 252, 193, 132, 26, 252, 193, 134, 26, 252, 193, 136, 26, 252, 193, 138, 26, 252, 193, 140, 26, 252, 193, 142, 26, 252, 193, 96, 28, 252, 193, 98, 28, 252, 193, 100, 28, 252, 193, 102, 28, 252, 193, 104, 28, 252, 193, 106, 28, 252, 193, 108, 28, 252, 193, 110, 28, 252, 193, 112, 28, 252, 193, 114, 28, 252, 193, 116, 28, 252, 193, 118, 28, 252, 193, 120, 28, 252, 193, 122, 28, 252, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 193, 170, 26, 252, 193, 172, 26, 252, 193, 174, 26, 252, 193, 176, 26, 252, 193, 178, 26, 252, 193, 180, 26, 252, 193, 182, 26, 252, 193, 184, 26, 252, 193, 186, 26, 252, 193, 188, 26, 252, 193, 190, 26, 252, 193, 158, 28, 252, 193, 160, 28, 252, 193, 162, 28, 252, 193, 164, 28, 252, 193, 166, 28, 252, 193, 168, 28, 252, 193, 170, 28, 252, 193, 172, 28, 252, 193, 174, 28, 252, 193, 176, 28, 252, 193, 178, 28, 252, 193, 180, 28, 252, 193, 182, 28, 252, 193, 184, 28, 252, 193, 186, 28, 252, 193, 188, 28, 252, 193, 190, 28, 252, 193, 192, 28, 252, 193, 86, 31, 252, 193, 88, 31, 252, 193, 90, 31, 252, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 206, 128, 43, 0, 193, 194, 28, 252, 206, 160, 43, 0, 206, 160, 43, 0, 206, 160, 43, 0, 206, 160, 43, 0, 206, 160, 43, 0, 206, 160, 43, 0, 206, 160, 43, 0, 206, 160, 43, 0, 206, 160, 43, 0, 206, 160, 43, 0, 206, 160, 43, 0, 206, 160, 43, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 192, 43, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 193, 64, 28, 252, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 193, 70, 28, 252, 193, 72, 28, 252, 193, 74, 28, 252, 193, 76, 28, 252, 193, 78, 28, 252, 193, 80, 28, 252, 193, 82, 28, 252, 193, 84, 28, 252, 193, 86, 28, 252, 193, 88, 28, 252, 193, 90, 28, 252, 193, 92, 28, 252, 193, 230, 30, 252, 193, 232, 30, 252, 193, 234, 30, 252, 193, 236, 30, 252, 193, 238, 30, 252, 193, 240, 30, 252, 193, 242, 30, 252, 193, 244, 30, 252, 193, 246, 30, 252, 193, 248, 30, 252, 193, 250, 30, 252, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 206, 224, 43, 0, 193, 32, 31, 252, 193, 34, 31, 252, 193, 36, 31, 252, 193, 38, 31, 252, 193, 40, 31, 252, 193, 42, 31, 252, 193, 44, 31, 252, 193, 46, 31, 252, 193, 48, 31, 252, 193, 50, 31, 252, 193, 52, 31, 252, 193, 54, 31, 252, 193, 56, 31, 252, 193, 58, 31, 252, 193, 60, 31, 252, 193, 62, 31, 252, 193, 64, 31, 252, 193, 66, 31, 252, 193, 68, 31, 252, 193, 70, 31, 252, 193, 72, 31, 252, 193, 74, 31, 252, 193, 76, 31, 252, 193, 78, 31, 252, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 206, 0, 44, 0, 193, 116, 31, 252, 193, 120, 29, 252, 193, 122, 29, 252, 193, 124, 29, 252, 193, 126, 29, 252, 193, 128, 29, 252, 193, 130, 29, 252, 193, 132, 29, 252, 193, 134, 29, 252, 193, 136, 29, 252, 193, 138, 29, 252, 193, 140, 29, 252, 193, 142, 29, 252, 193, 144, 29, 252, 193, 146, 29, 252, 193, 148, 29, 252, 193, 150, 29, 252, 193, 152, 29, 252, 193, 154, 29, 252, 193, 156, 29, 252, 193, 158, 29, 252, 193, 118, 31, 252, 193, 120, 31, 252, 193, 160, 29, 252, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 193, 162, 29, 252, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 32, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 64, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 96, 44, 0, 206, 0, 45, 0, 206, 0, 45, 0, 206, 0, 45, 0, 206, 0, 45, 0, 206, 0, 45, 0, 206, 0, 45, 0, 206, 0, 45, 0, 206, 0, 45, 0, 206, 0, 45, 0, 206, 0, 45, 0, 193, 228, 30, 252, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 206, 128, 44, 0, 193, 120, 33, 252, 193, 122, 33, 252, 193, 124, 33, 252, 193, 126, 33, 252, 193, 128, 33, 252, 193, 130, 33, 252, 193, 132, 33, 252, 193, 134, 33, 252, 193, 136, 33, 252, 193, 138, 33, 252, 193, 140, 33, 252, 193, 142, 33, 252, 193, 144, 33, 252, 193, 146, 33, 252, 193, 148, 33, 252, 193, 150, 33, 252, 193, 152, 33, 252, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 193, 80, 31, 252, 193, 82, 31, 252, 193, 84, 31, 252, 193, 154, 33, 252, 193, 156, 33, 252, 193, 158, 33, 252, 193, 160, 33, 252, 193, 162, 33, 252, 193, 164, 33, 252, 193, 166, 33, 252, 193, 168, 33, 252, 193, 170, 33, 252, 193, 172, 33, 252, 193, 174, 33, 252, 193, 176, 33, 252, 193, 178, 33, 252, 193, 180, 33, 252, 193, 182, 33, 252, 193, 184, 33, 252, 193, 186, 33, 252, 193, 188, 33, 252, 193, 190, 33, 252, 193, 192, 33, 252, 193, 194, 33, 252, 193, 196, 33, 252, 193, 198, 33, 252, 193, 200, 33, 252, 193, 202, 33, 252, 193, 204, 33, 252, 193, 206, 33, 252, 193, 208, 33, 252, 193, 210, 33, 252, 193, 92, 31, 252, 193, 94, 31, 252, 193, 96, 31, 252, 193, 98, 31, 252, 193, 100, 31, 252, 193, 102, 31, 252, 193, 104, 31, 252, 193, 106, 31, 252, 193, 108, 31, 252, 193, 110, 31, 252, 193, 112, 31, 252, 193, 114, 31, 252, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 160, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 192, 44, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 193, 94, 32, 252, 193, 100, 36, 252, 193, 136, 34, 252, 193, 138, 34, 252, 193, 140, 34, 252, 193, 142, 34, 252, 193, 144, 34, 252, 193, 146, 34, 252, 193, 148, 34, 252, 193, 150, 34, 252, 193, 152, 34, 252, 193, 154, 34, 252, 193, 156, 34, 252, 193, 158, 34, 252, 193, 160, 34, 252, 193, 102, 36, 252, 193, 162, 34, 252, 193, 164, 34, 252, 193, 166, 34, 252, 193, 96, 32, 252, 193, 168, 34, 252, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 206, 224, 44, 0, 193, 194, 34, 252, 193, 196, 34, 252, 193, 198, 34, 252, 193, 200, 34, 252, 193, 202, 34, 252, 193, 204, 34, 252, 193, 206, 34, 252, 193, 208, 34, 252, 193, 210, 34, 252, 193, 212, 34, 252, 193, 214, 34, 252, 193, 216, 34, 252, 193, 218, 34, 252, 193, 220, 34, 252, 193, 222, 34, 252, 193, 224, 34, 252, 193, 226, 34, 252, 193, 228, 34, 252, 193, 230, 34, 252, 193, 232, 34, 252, 193, 234, 34, 252, 193, 236, 34, 252, 193, 238, 34, 252, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 32, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 206, 64, 45, 0, 193, 212, 35, 252, 193, 214, 35, 252, 193, 216, 35, 252, 193, 218, 35, 252, 193, 220, 35, 252, 193, 222, 35, 252, 193, 224, 35, 252, 193, 226, 35, 252, 193, 228, 35, 252, 193, 230, 35, 252, 193, 232, 35, 252, 193, 234, 35, 252, 193, 236, 35, 252, 193, 238, 35, 252, 193, 240, 35, 252, 193, 242, 35, 252, 193, 244, 35, 252, 193, 246, 35, 252, 193, 248, 35, 252, 193, 250, 35, 252, 193, 252, 35, 252, 193, 48, 38, 252, 193, 212, 33, 252, 193, 214, 33, 252, 193, 36, 36, 252, 193, 38, 36, 252, 193, 40, 36, 252, 193, 42, 36, 252, 193, 44, 36, 252, 193, 46, 36, 252, 193, 48, 36, 252, 193, 50, 36, 252, 193, 52, 36, 252, 193, 54, 36, 252, 193, 56, 36, 252, 193, 58, 36, 252, 193, 60, 36, 252, 193, 62, 36, 252, 193, 64, 36, 252, 193, 66, 36, 252, 193, 68, 36, 252, 193, 70, 36, 252, 193, 72, 36, 252, 193, 74, 36, 252, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 206, 96, 45, 0, 193, 76, 36, 252, 193, 78, 36, 252, 193, 80, 36, 252, 193, 82, 36, 252, 193, 84, 36, 252, 193, 86, 36, 252, 193, 88, 36, 252, 193, 90, 36, 252, 193, 92, 36, 252, 193, 94, 36, 252, 193, 96, 36, 252, 193, 98, 36, 252, 193, 154, 38, 252, 193, 156, 38, 252, 193, 158, 38, 252, 193, 160, 38, 252, 193, 162, 38, 252, 193, 164, 38, 252, 193, 166, 38, 252, 193, 168, 38, 252, 193, 170, 38, 252, 193, 172, 38, 252, 193, 174, 38, 252, 193, 176, 38, 252, 193, 178, 38, 252, 193, 180, 38, 252, 193, 182, 38, 252, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 206, 128, 45, 0, 193, 170, 34, 252, 193, 172, 34, 252, 193, 174, 34, 252, 193, 176, 34, 252, 193, 178, 34, 252, 193, 180, 34, 252, 193, 182, 34, 252, 193, 184, 34, 252, 193, 186, 34, 252, 193, 188, 34, 252, 193, 190, 34, 252, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 206, 224, 45, 0, 193, 190, 38, 252, 193, 168, 36, 252, 193, 170, 36, 252, 193, 172, 36, 252, 193, 174, 36, 252, 193, 176, 36, 252, 193, 178, 36, 252, 193, 180, 36, 252, 193, 192, 34, 252, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 160, 45, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 193, 210, 35, 252, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 206, 192, 45, 0, 193, 72, 38, 252, 193, 74, 38, 252, 193, 76, 38, 252, 193, 78, 38, 252, 193, 80, 38, 252, 193, 82, 38, 252, 193, 84, 38, 252, 193, 86, 38, 252, 193, 88, 38, 252, 193, 90, 38, 252, 193, 92, 38, 252, 193, 94, 38, 252, 193, 96, 38, 252, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 206, 0, 46, 0, 193, 192, 38, 252, 193, 10, 37, 252, 193, 12, 37, 252, 193, 14, 37, 252, 193, 16, 37, 252, 193, 18, 37, 252, 193, 20, 37, 252, 193, 22, 37, 252, 193, 24, 37, 252, 193, 26, 37, 252, 193, 28, 37, 252, 193, 30, 37, 252, 193, 32, 37, 252, 193, 34, 37, 252, 193, 36, 37, 252, 193, 38, 37, 252, 193, 40, 37, 252, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 206, 32, 46, 0, 193, 100, 39, 252, 193, 102, 39, 252, 193, 104, 39, 252, 193, 106, 39, 252, 193, 108, 39, 252, 193, 110, 39, 252, 193, 112, 39, 252, 193, 114, 39, 252, 193, 116, 39, 252, 193, 118, 39, 252, 193, 120, 39, 252, 193, 122, 39, 252, 193, 124, 39, 252, 193, 126, 39, 252, 193, 128, 39, 252, 193, 130, 39, 252, 193, 132, 39, 252, 193, 134, 39, 252, 193, 136, 39, 252, 193, 160, 41, 252, 193, 162, 41, 252, 193, 164, 41, 252, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 206, 64, 46, 0, 193, 138, 39, 252, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 206, 96, 46, 0, 193, 232, 37, 252, 193, 234, 37, 252, 193, 236, 37, 252, 193, 238, 37, 252, 193, 240, 37, 252, 193, 242, 37, 252, 193, 244, 37, 252, 193, 246, 37, 252, 193, 248, 37, 252, 193, 250, 37, 252, 193, 252, 37, 252, 193, 254, 37, 252, 193, 2, 38, 252, 193, 4, 38, 252, 193, 6, 38, 252, 193, 8, 38, 252, 193, 10, 38, 252, 193, 12, 38, 252, 193, 14, 38, 252, 193, 16, 38, 252, 193, 18, 38, 252, 193, 20, 38, 252, 193, 22, 38, 252, 193, 24, 38, 252, 193, 26, 38, 252, 193, 28, 38, 252, 193, 30, 38, 252, 193, 32, 38, 252, 193, 34, 38, 252, 193, 36, 38, 252, 193, 38, 38, 252, 193, 40, 38, 252, 193, 42, 38, 252, 193, 44, 38, 252, 193, 46, 38, 252, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 193, 50, 38, 252, 193, 52, 38, 252, 193, 54, 38, 252, 193, 56, 38, 252, 193, 58, 38, 252, 193, 60, 38, 252, 193, 62, 38, 252, 193, 64, 38, 252, 193, 66, 38, 252, 193, 68, 38, 252, 193, 70, 38, 252, 193, 142, 40, 252, 193, 144, 40, 252, 193, 146, 40, 252, 193, 148, 40, 252, 193, 150, 40, 252, 193, 152, 40, 252, 193, 154, 40, 252, 193, 156, 40, 252, 193, 158, 40, 252, 193, 160, 40, 252, 193, 162, 40, 252, 193, 164, 40, 252, 193, 166, 40, 252, 193, 168, 40, 252, 193, 170, 40, 252, 193, 172, 40, 252, 193, 174, 40, 252, 193, 176, 40, 252, 193, 130, 42, 252, 193, 132, 42, 252, 193, 134, 42, 252, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 128, 46, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 193, 184, 38, 252, 193, 186, 38, 252, 193, 188, 38, 252, 193, 22, 41, 252, 193, 24, 41, 252, 193, 26, 41, 252, 193, 28, 41, 252, 193, 30, 41, 252, 193, 32, 41, 252, 193, 34, 41, 252, 193, 36, 41, 252, 193, 38, 41, 252, 193, 40, 41, 252, 193, 42, 41, 252, 193, 44, 41, 252, 193, 46, 41, 252, 193, 48, 41, 252, 193, 208, 42, 252, 193, 210, 42, 252, 193, 212, 42, 252, 193, 214, 42, 252, 193, 216, 42, 252, 193, 218, 42, 252, 193, 220, 42, 252, 193, 222, 42, 252, 193, 18, 44, 252, 193, 20, 44, 252, 193, 22, 44, 252, 193, 24, 44, 252, 193, 26, 44, 252, 193, 28, 44, 252, 193, 30, 44, 252, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 206, 160, 46, 0, 193, 224, 42, 252, 193, 40, 39, 252, 193, 42, 39, 252, 193, 44, 39, 252, 193, 46, 39, 252, 193, 48, 39, 252, 193, 50, 39, 252, 193, 52, 39, 252, 193, 54, 39, 252, 193, 56, 39, 252, 193, 58, 39, 252, 193, 60, 39, 252, 193, 62, 39, 252, 193, 64, 39, 252, 193, 66, 39, 252, 193, 68, 39, 252, 193, 70, 39, 252, 193, 72, 39, 252, 193, 74, 39, 252, 193, 76, 39, 252, 193, 78, 39, 252, 193, 80, 39, 252, 193, 82, 39, 252, 193, 84, 39, 252, 193, 86, 39, 252, 193, 50, 41, 252, 193, 88, 39, 252, 193, 90, 39, 252, 193, 92, 39, 252, 193, 94, 39, 252, 193, 52, 41, 252, 193, 54, 41, 252, 193, 120, 41, 252, 193, 122, 41, 252, 193, 124, 41, 252, 193, 126, 41, 252, 193, 128, 41, 252, 193, 130, 41, 252, 193, 132, 41, 252, 193, 134, 41, 252, 193, 136, 41, 252, 193, 138, 41, 252, 193, 140, 41, 252, 193, 142, 41, 252, 193, 144, 41, 252, 193, 146, 41, 252, 193, 148, 41, 252, 193, 150, 41, 252, 193, 96, 39, 252, 193, 152, 41, 252, 193, 154, 41, 252, 193, 156, 41, 252, 193, 226, 42, 252, 193, 228, 42, 252, 193, 158, 41, 252, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 206, 192, 46, 0, 193, 196, 41, 252, 193, 90, 40, 252, 193, 92, 40, 252, 193, 94, 40, 252, 193, 96, 40, 252, 193, 98, 40, 252, 193, 100, 40, 252, 193, 102, 40, 252, 193, 104, 40, 252, 193, 106, 40, 252, 193, 108, 40, 252, 193, 110, 40, 252, 193, 112, 40, 252, 193, 114, 40, 252, 193, 116, 40, 252, 193, 118, 40, 252, 193, 120, 40, 252, 193, 122, 40, 252, 193, 124, 40, 252, 193, 126, 40, 252, 193, 128, 40, 252, 193, 130, 40, 252, 193, 132, 40, 252, 193, 70, 43, 252, 193, 134, 40, 252, 193, 136, 40, 252, 193, 138, 40, 252, 193, 140, 40, 252, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 206, 224, 46, 0, 193, 154, 42, 252, 193, 156, 42, 252, 193, 158, 42, 252, 193, 160, 42, 252, 193, 162, 42, 252, 193, 164, 42, 252, 193, 166, 42, 252, 193, 168, 42, 252, 193, 170, 42, 252, 193, 172, 42, 252, 193, 174, 42, 252, 193, 216, 43, 252, 193, 218, 43, 252, 193, 220, 43, 252, 193, 222, 43, 252, 193, 224, 43, 252, 193, 226, 43, 252, 193, 228, 43, 252, 193, 230, 43, 252, 193, 28, 45, 252, 193, 30, 45, 252, 193, 32, 45, 252, 193, 34, 45, 252, 193, 20, 46, 252, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 206, 0, 47, 0, 193, 176, 42, 252, 193, 178, 42, 252, 193, 180, 42, 252, 193, 182, 42, 252, 193, 184, 42, 252, 193, 186, 42, 252, 193, 188, 42, 252, 193, 190, 42, 252, 193, 192, 42, 252, 193, 194, 42, 252, 193, 196, 42, 252, 193, 198, 42, 252, 193, 200, 42, 252, 193, 202, 42, 252, 193, 204, 42, 252, 193, 206, 42, 252, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 193, 56, 41, 252, 193, 58, 41, 252, 193, 60, 41, 252, 193, 62, 41, 252, 193, 64, 41, 252, 193, 66, 41, 252, 193, 68, 41, 252, 193, 70, 41, 252, 193, 72, 41, 252, 193, 74, 41, 252, 193, 76, 41, 252, 193, 78, 41, 252, 193, 80, 41, 252, 193, 82, 41, 252, 193, 84, 41, 252, 193, 86, 41, 252, 193, 88, 41, 252, 193, 90, 41, 252, 193, 92, 41, 252, 193, 94, 41, 252, 193, 96, 41, 252, 193, 98, 41, 252, 193, 100, 41, 252, 193, 102, 41, 252, 193, 104, 41, 252, 193, 106, 41, 252, 193, 108, 41, 252, 193, 110, 41, 252, 193, 112, 41, 252, 193, 114, 41, 252, 193, 116, 41, 252, 193, 118, 41, 252, 193, 166, 41, 252, 193, 168, 41, 252, 193, 170, 41, 252, 193, 172, 41, 252, 193, 174, 41, 252, 193, 176, 41, 252, 193, 178, 41, 252, 193, 180, 41, 252, 193, 182, 41, 252, 193, 184, 41, 252, 193, 186, 41, 252, 193, 188, 41, 252, 193, 190, 41, 252, 193, 192, 41, 252, 193, 194, 41, 252, 193, 42, 43, 252, 193, 44, 43, 252, 193, 46, 43, 252, 193, 48, 43, 252, 193, 50, 43, 252, 193, 52, 43, 252, 193, 54, 43, 252, 193, 56, 43, 252, 193, 58, 43, 252, 193, 60, 43, 252, 193, 62, 43, 252, 193, 64, 43, 252, 193, 66, 43, 252, 193, 68, 43, 252, 193, 92, 44, 252, 193, 94, 44, 252, 193, 96, 44, 252, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 206, 32, 47, 0, 193, 136, 42, 252, 193, 138, 42, 252, 193, 140, 42, 252, 193, 142, 42, 252, 193, 144, 42, 252, 193, 146, 42, 252, 193, 148, 42, 252, 193, 150, 42, 252, 193, 194, 43, 252, 193, 196, 43, 252, 193, 198, 43, 252, 193, 200, 43, 252, 193, 202, 43, 252, 193, 204, 43, 252, 193, 206, 43, 252, 193, 208, 43, 252, 193, 210, 43, 252, 193, 212, 43, 252, 193, 214, 43, 252, 193, 14, 45, 252, 193, 16, 45, 252, 193, 18, 45, 252, 193, 20, 45, 252, 193, 22, 45, 252, 193, 24, 45, 252, 193, 26, 45, 252, 193, 6, 46, 252, 193, 8, 46, 252, 193, 10, 46, 252, 193, 12, 46, 252, 193, 14, 46, 252, 193, 16, 46, 252, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 206, 64, 47, 0, 193, 80, 45, 252, 193, 32, 43, 252, 193, 34, 43, 252, 193, 36, 43, 252, 193, 38, 43, 252, 193, 40, 43, 252, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 206, 96, 47, 0, 193, 106, 44, 252, 193, 190, 43, 252, 193, 192, 43, 252, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 206, 128, 47, 0, 193, 36, 45, 252, 193, 38, 45, 252, 193, 40, 45, 252, 193, 42, 45, 252, 193, 44, 45, 252, 193, 46, 45, 252, 193, 48, 45, 252, 193, 50, 45, 252, 193, 52, 45, 252, 193, 54, 45, 252, 193, 56, 45, 252, 193, 58, 45, 252, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 193, 32, 44, 252, 193, 34, 44, 252, 193, 60, 45, 252, 193, 62, 45, 252, 193, 64, 45, 252, 193, 66, 45, 252, 193, 68, 45, 252, 193, 70, 45, 252, 193, 72, 45, 252, 193, 74, 45, 252, 193, 76, 45, 252, 193, 78, 45, 252, 193, 64, 46, 252, 193, 66, 46, 252, 193, 68, 46, 252, 193, 70, 46, 252, 193, 224, 46, 252, 193, 226, 46, 252, 193, 128, 47, 252, 193, 204, 47, 252, 193, 206, 47, 252, 193, 130, 48, 252, 193, 198, 48, 252, 193, 40, 49, 252, 193, 42, 49, 252, 193, 44, 49, 252, 193, 46, 49, 252, 193, 48, 49, 252, 193, 50, 49, 252, 193, 228, 49, 252, 193, 230, 49, 252, 193, 232, 49, 252, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 160, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 193, 98, 44, 252, 193, 100, 44, 252, 193, 102, 44, 252, 193, 104, 44, 252, 193, 134, 45, 252, 193, 136, 45, 252, 193, 138, 45, 252, 193, 140, 45, 252, 193, 142, 45, 252, 193, 144, 45, 252, 193, 146, 45, 252, 193, 148, 45, 252, 193, 110, 46, 252, 193, 112, 46, 252, 193, 114, 46, 252, 193, 116, 46, 252, 193, 118, 46, 252, 193, 120, 46, 252, 193, 2, 47, 252, 193, 4, 47, 252, 193, 6, 47, 252, 193, 8, 47, 252, 193, 10, 47, 252, 193, 12, 47, 252, 193, 14, 47, 252, 193, 146, 47, 252, 193, 148, 47, 252, 193, 216, 47, 252, 193, 218, 47, 252, 193, 220, 47, 252, 193, 26, 48, 252, 193, 28, 48, 252, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 206, 192, 47, 0, 193, 150, 45, 252, 193, 228, 44, 252, 193, 230, 44, 252, 193, 232, 44, 252, 193, 152, 45, 252, 193, 234, 44, 252, 193, 236, 44, 252, 193, 238, 44, 252, 193, 240, 44, 252, 193, 242, 44, 252, 193, 244, 44, 252, 193, 246, 44, 252, 193, 248, 44, 252, 193, 250, 44, 252, 193, 252, 44, 252, 193, 254, 44, 252, 193, 2, 45, 252, 193, 4, 45, 252, 193, 6, 45, 252, 193, 8, 45, 252, 193, 10, 45, 252, 193, 12, 45, 252, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 206, 224, 47, 0, 193, 72, 46, 252, 193, 74, 46, 252, 193, 76, 46, 252, 193, 78, 46, 252, 193, 80, 46, 252, 193, 82, 46, 252, 193, 84, 46, 252, 193, 86, 46, 252, 193, 88, 46, 252, 193, 90, 46, 252, 193, 92, 46, 252, 193, 94, 46, 252, 193, 96, 46, 252, 193, 98, 46, 252, 193, 100, 46, 252, 193, 102, 46, 252, 193, 104, 46, 252, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 0, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 193, 22, 46, 252, 193, 24, 46, 252, 193, 26, 46, 252, 193, 28, 46, 252, 193, 30, 46, 252, 193, 204, 46, 252, 193, 206, 46, 252, 193, 208, 46, 252, 193, 210, 46, 252, 193, 212, 46, 252, 193, 214, 46, 252, 193, 108, 47, 252, 193, 110, 47, 252, 193, 112, 47, 252, 193, 194, 47, 252, 193, 196, 47, 252, 193, 50, 48, 252, 193, 52, 48, 252, 193, 126, 48, 252, 193, 138, 48, 252, 193, 16, 49, 252, 193, 18, 49, 252, 193, 20, 49, 252, 193, 22, 49, 252, 193, 24, 49, 252, 193, 26, 49, 252, 193, 28, 49, 252, 193, 196, 49, 252, 193, 198, 49, 252, 193, 200, 49, 252, 193, 202, 49, 252, 193, 204, 49, 252, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 206, 32, 48, 0, 193, 216, 46, 252, 193, 218, 46, 252, 193, 220, 46, 252, 193, 222, 46, 252, 193, 114, 47, 252, 193, 116, 47, 252, 193, 118, 47, 252, 193, 120, 47, 252, 193, 122, 47, 252, 193, 124, 47, 252, 193, 126, 47, 252, 193, 198, 47, 252, 193, 200, 47, 252, 193, 202, 47, 252, 193, 54, 48, 252, 193, 56, 48, 252, 193, 58, 48, 252, 193, 60, 48, 252, 193, 134, 48, 252, 193, 152, 48, 252, 193, 106, 46, 252, 193, 108, 46, 252, 193, 228, 46, 252, 193, 230, 46, 252, 193, 232, 46, 252, 193, 234, 46, 252, 193, 236, 46, 252, 193, 238, 46, 252, 193, 240, 46, 252, 193, 242, 46, 252, 193, 244, 46, 252, 193, 246, 46, 252, 193, 248, 46, 252, 193, 250, 46, 252, 193, 80, 48, 252, 193, 252, 46, 252, 193, 254, 46, 252, 193, 130, 47, 252, 193, 132, 47, 252, 193, 134, 47, 252, 193, 136, 47, 252, 193, 138, 47, 252, 193, 140, 47, 252, 193, 142, 47, 252, 193, 144, 47, 252, 193, 208, 47, 252, 193, 210, 47, 252, 193, 212, 47, 252, 193, 214, 47, 252, 193, 16, 48, 252, 193, 18, 48, 252, 193, 20, 48, 252, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 64, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 193, 198, 46, 252, 193, 200, 46, 252, 193, 202, 46, 252, 193, 98, 47, 252, 193, 100, 47, 252, 193, 102, 47, 252, 193, 104, 47, 252, 193, 106, 47, 252, 193, 190, 47, 252, 193, 192, 47, 252, 193, 12, 48, 252, 193, 14, 48, 252, 193, 48, 48, 252, 193, 150, 48, 252, 193, 10, 49, 252, 193, 12, 49, 252, 193, 14, 49, 252, 193, 182, 49, 252, 193, 184, 49, 252, 193, 186, 49, 252, 193, 188, 49, 252, 193, 190, 49, 252, 193, 192, 49, 252, 193, 194, 49, 252, 193, 174, 50, 252, 193, 176, 50, 252, 193, 178, 50, 252, 193, 180, 50, 252, 193, 182, 50, 252, 193, 184, 50, 252, 193, 186, 50, 252, 193, 188, 50, 252, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 96, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 206, 128, 48, 0, 193, 222, 47, 252, 193, 224, 47, 252, 193, 226, 47, 252, 193, 228, 47, 252, 193, 230, 47, 252, 193, 232, 47, 252, 193, 234, 47, 252, 193, 236, 47, 252, 193, 238, 47, 252, 193, 240, 47, 252, 193, 242, 47, 252, 193, 244, 47, 252, 193, 246, 47, 252, 193, 248, 47, 252, 193, 250, 47, 252, 193, 252, 47, 252, 193, 254, 47, 252, 193, 2, 48, 252, 193, 4, 48, 252, 193, 6, 48, 252, 193, 8, 48, 252, 193, 10, 48, 252, 193, 22, 48, 252, 193, 24, 48, 252, 193, 62, 48, 252, 193, 64, 48, 252, 193, 82, 48, 252, 193, 84, 48, 252, 193, 86, 48, 252, 193, 104, 48, 252, 193, 106, 48, 252, 193, 128, 48, 252, 193, 132, 48, 252, 193, 142, 48, 252, 193, 144, 48, 252, 193, 146, 48, 252, 193, 154, 48, 252, 193, 156, 48, 252, 193, 158, 48, 252, 193, 160, 48, 252, 193, 162, 48, 252, 193, 164, 48, 252, 193, 200, 48, 252, 193, 202, 48, 252, 193, 204, 48, 252, 193, 206, 48, 252, 193, 208, 48, 252, 193, 210, 48, 252, 193, 212, 48, 252, 193, 214, 48, 252, 193, 216, 48, 252, 193, 218, 48, 252, 193, 220, 48, 252, 193, 222, 48, 252, 193, 30, 48, 252, 193, 32, 48, 252, 193, 34, 48, 252, 193, 36, 48, 252, 193, 38, 48, 252, 193, 40, 48, 252, 193, 42, 48, 252, 193, 44, 48, 252, 193, 46, 48, 252, 193, 66, 48, 252, 193, 68, 48, 252, 193, 70, 48, 252, 193, 72, 48, 252, 193, 74, 48, 252, 193, 76, 48, 252, 193, 78, 48, 252, 193, 90, 48, 252, 193, 92, 48, 252, 193, 94, 48, 252, 193, 96, 48, 252, 193, 98, 48, 252, 193, 100, 48, 252, 193, 102, 48, 252, 193, 108, 48, 252, 193, 110, 48, 252, 193, 112, 48, 252, 193, 114, 48, 252, 193, 116, 48, 252, 193, 118, 48, 252, 193, 120, 48, 252, 193, 122, 48, 252, 193, 124, 48, 252, 193, 88, 48, 252, 193, 168, 48, 252, 193, 170, 48, 252, 193, 226, 48, 252, 193, 228, 48, 252, 193, 230, 48, 252, 193, 110, 49, 252, 193, 112, 49, 252, 193, 114, 49, 252, 193, 116, 49, 252, 193, 118, 49, 252, 193, 56, 50, 252, 193, 58, 50, 252, 193, 60, 50, 252, 193, 62, 50, 252, 193, 64, 50, 252, 193, 66, 50, 252, 193, 68, 50, 252, 193, 70, 50, 252, 193, 64, 51, 252, 193, 66, 51, 252, 193, 68, 51, 252, 193, 70, 51, 252, 193, 72, 51, 252, 193, 74, 51, 252, 193, 110, 52, 252, 193, 112, 52, 252, 193, 114, 52, 252, 193, 116, 52, 252, 193, 118, 52, 252, 193, 120, 52, 252, 193, 122, 52, 252, 193, 136, 48, 252, 193, 140, 48, 252, 193, 148, 48, 252, 193, 172, 48, 252, 193, 174, 48, 252, 193, 176, 48, 252, 193, 178, 48, 252, 193, 180, 48, 252, 193, 182, 48, 252, 193, 184, 48, 252, 193, 186, 48, 252, 193, 188, 48, 252, 193, 190, 48, 252, 193, 232, 48, 252, 193, 234, 48, 252, 193, 236, 48, 252, 193, 238, 48, 252, 193, 240, 48, 252, 193, 242, 48, 252, 193, 244, 48, 252, 193, 246, 48, 252, 193, 248, 48, 252, 193, 250, 48, 252, 193, 252, 48, 252, 193, 254, 48, 252, 193, 2, 49, 252, 193, 4, 49, 252, 193, 6, 49, 252, 193, 8, 49, 252, 193, 120, 49, 252, 193, 122, 49, 252, 193, 124, 49, 252, 193, 224, 48, 252, 193, 52, 49, 252, 193, 54, 49, 252, 193, 56, 49, 252, 193, 58, 49, 252, 193, 60, 49, 252, 193, 62, 49, 252, 193, 166, 48, 252, 193, 64, 49, 252, 193, 66, 49, 252, 193, 68, 49, 252, 193, 70, 49, 252, 193, 72, 49, 252, 193, 74, 49, 252, 193, 76, 49, 252, 193, 78, 49, 252, 193, 80, 49, 252, 193, 82, 49, 252, 193, 84, 49, 252, 193, 86, 49, 252, 193, 88, 49, 252, 193, 90, 49, 252, 193, 92, 49, 252, 193, 94, 49, 252, 193, 96, 49, 252, 193, 98, 49, 252, 193, 100, 49, 252, 193, 102, 49, 252, 193, 104, 49, 252, 193, 106, 49, 252, 193, 108, 49, 252, 193, 236, 49, 252, 193, 192, 48, 252, 193, 194, 48, 252, 193, 196, 48, 252, 193, 30, 49, 252, 193, 32, 49, 252, 193, 34, 49, 252, 193, 36, 49, 252, 193, 38, 49, 252, 193, 208, 49, 252, 193, 210, 49, 252, 193, 212, 49, 252, 193, 214, 49, 252, 193, 216, 49, 252, 193, 218, 49, 252, 193, 220, 49, 252, 193, 222, 49, 252, 193, 224, 49, 252, 193, 226, 49, 252, 193, 200, 50, 252, 193, 202, 50, 252, 193, 204, 50, 252, 193, 206, 50, 252, 193, 208, 50, 252, 193, 210, 50, 252, 193, 212, 50, 252, 193, 214, 50, 252, 193, 216, 50, 252, 206, 0, 49, 0, 206, 0, 49, 0, 206, 0, 49, 0, 206, 0, 49, 0, 206, 0, 49, 0, 193, 126, 49, 252, 193, 128, 49, 252, 193, 130, 49, 252, 193, 132, 49, 252, 193, 134, 49, 252, 193, 136, 49, 252, 193, 138, 49, 252, 193, 140, 49, 252, 193, 142, 49, 252, 193, 144, 49, 252, 193, 146, 49, 252, 193, 148, 49, 252, 193, 150, 49, 252, 193, 152, 49, 252, 193, 154, 49, 252, 193, 156, 49, 252, 193, 158, 49, 252, 193, 160, 49, 252, 193, 162, 49, 252, 193, 164, 49, 252, 193, 166, 49, 252, 193, 168, 49, 252, 193, 170, 49, 252, 193, 172, 49, 252, 193, 174, 49, 252, 193, 176, 49, 252, 193, 178, 49, 252, 193, 180, 49, 252, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 193, 206, 49, 252, 193, 196, 50, 252, 193, 198, 50, 252, 193, 202, 51, 252, 193, 204, 51, 252, 193, 206, 51, 252, 193, 208, 51, 252, 193, 210, 51, 252, 193, 212, 51, 252, 193, 14, 53, 252, 193, 16, 53, 252, 193, 18, 53, 252, 193, 20, 53, 252, 193, 22, 53, 252, 193, 24, 53, 252, 193, 26, 53, 252, 193, 28, 53, 252, 193, 30, 53, 252, 193, 32, 53, 252, 193, 34, 53, 252, 193, 36, 53, 252, 193, 38, 53, 252, 193, 40, 53, 252, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 193, 234, 49, 252, 193, 218, 50, 252, 193, 220, 50, 252, 193, 222, 50, 252, 193, 224, 50, 252, 193, 226, 50, 252, 193, 228, 50, 252, 193, 230, 50, 252, 193, 236, 51, 252, 193, 238, 51, 252, 193, 240, 51, 252, 193, 242, 51, 252, 193, 244, 51, 252, 193, 246, 51, 252, 193, 248, 51, 252, 193, 250, 51, 252, 193, 252, 51, 252, 193, 254, 51, 252, 193, 2, 52, 252, 193, 4, 52, 252, 193, 6, 52, 252, 193, 8, 52, 252, 193, 10, 52, 252, 193, 12, 52, 252, 193, 14, 52, 252, 193, 16, 52, 252, 193, 18, 52, 252, 193, 20, 52, 252, 193, 22, 52, 252, 193, 24, 52, 252, 193, 26, 52, 252, 193, 78, 53, 252, 193, 238, 49, 252, 193, 240, 49, 252, 193, 242, 49, 252, 193, 244, 49, 252, 193, 246, 49, 252, 193, 248, 49, 252, 193, 250, 49, 252, 193, 252, 49, 252, 193, 254, 49, 252, 193, 2, 50, 252, 193, 4, 50, 252, 193, 6, 50, 252, 193, 8, 50, 252, 193, 10, 50, 252, 193, 12, 50, 252, 193, 14, 50, 252, 193, 16, 50, 252, 193, 18, 50, 252, 193, 20, 50, 252, 193, 22, 50, 252, 193, 24, 50, 252, 193, 26, 50, 252, 193, 28, 50, 252, 193, 30, 50, 252, 193, 32, 50, 252, 193, 34, 50, 252, 193, 36, 50, 252, 193, 38, 50, 252, 193, 40, 50, 252, 193, 42, 50, 252, 193, 44, 50, 252, 193, 46, 50, 252, 193, 48, 50, 252, 193, 50, 50, 252, 193, 52, 50, 252, 193, 54, 50, 252, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 160, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 193, 190, 50, 252, 193, 192, 50, 252, 193, 194, 50, 252, 193, 184, 51, 252, 193, 186, 51, 252, 193, 188, 51, 252, 193, 190, 51, 252, 193, 192, 51, 252, 193, 194, 51, 252, 193, 196, 51, 252, 193, 198, 51, 252, 193, 250, 52, 252, 193, 252, 52, 252, 193, 254, 52, 252, 193, 2, 53, 252, 193, 4, 53, 252, 193, 6, 53, 252, 193, 8, 53, 252, 193, 10, 53, 252, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 192, 48, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 206, 224, 48, 0, 193, 128, 52, 252, 193, 130, 52, 252, 193, 132, 52, 252, 193, 134, 52, 252, 193, 136, 52, 252, 193, 138, 52, 252, 193, 140, 52, 252, 193, 142, 52, 252, 193, 144, 52, 252, 193, 146, 52, 252, 193, 148, 52, 252, 193, 150, 52, 252, 193, 152, 52, 252, 193, 154, 52, 252, 193, 156, 52, 252, 193, 158, 52, 252, 193, 160, 52, 252, 193, 162, 52, 252, 193, 164, 52, 252, 193, 166, 52, 252, 193, 168, 52, 252, 193, 170, 52, 252, 193, 172, 52, 252, 193, 174, 52, 252, 193, 176, 52, 252, 193, 178, 52, 252, 193, 180, 52, 252, 206, 0, 49, 0, 206, 0, 49, 0, 206, 0, 49, 0, 206, 0, 49, 0, 206, 0, 49, 0, 206, 0, 49, 0, 193, 42, 53, 252, 193, 44, 53, 252, 193, 46, 53, 252, 193, 48, 53, 252, 193, 50, 53, 252, 193, 52, 53, 252, 193, 54, 53, 252, 193, 56, 53, 252, 193, 58, 53, 252, 193, 60, 53, 252, 193, 62, 53, 252, 193, 64, 53, 252, 193, 66, 53, 252, 193, 68, 53, 252, 193, 70, 53, 252, 193, 72, 53, 252, 193, 74, 53, 252, 193, 76, 53, 252, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 206, 32, 49, 0, 193, 106, 53, 252, 193, 102, 52, 252, 193, 104, 52, 252, 193, 106, 52, 252, 193, 108, 52, 252, 193, 108, 53, 252, 193, 110, 53, 252, 193, 112, 53, 252, 193, 114, 53, 252, 193, 116, 53, 252, 193, 118, 53, 252, 193, 120, 53, 252, 193, 124, 52, 252, 193, 126, 52, 252, 193, 190, 53, 252, 193, 192, 53, 252, 193, 194, 53, 252, 193, 196, 53, 252, 193, 198, 53, 252, 193, 200, 53, 252, 193, 202, 53, 252, 193, 204, 53, 252, 193, 206, 53, 252, 193, 208, 53, 252, 193, 210, 53, 252, 193, 212, 53, 252, 193, 110, 55, 252, 193, 112, 55, 252, 193, 114, 55, 252, 193, 116, 55, 252, 193, 118, 55, 252, 193, 120, 55, 252, 193, 122, 55, 252, 193, 124, 55, 252, 193, 126, 55, 252, 193, 128, 55, 252, 193, 130, 55, 252, 193, 132, 55, 252, 193, 134, 55, 252, 193, 136, 55, 252, 193, 138, 55, 252, 193, 140, 55, 252, 193, 142, 55, 252, 193, 250, 56, 252, 193, 182, 52, 252, 193, 184, 52, 252, 193, 186, 52, 252, 193, 188, 52, 252, 193, 190, 52, 252, 193, 192, 52, 252, 193, 194, 52, 252, 193, 196, 52, 252, 193, 198, 52, 252, 193, 200, 52, 252, 193, 202, 52, 252, 193, 204, 52, 252, 193, 206, 52, 252, 193, 208, 52, 252, 193, 210, 52, 252, 193, 212, 52, 252, 193, 214, 52, 252, 193, 216, 52, 252, 193, 218, 52, 252, 193, 220, 52, 252, 193, 222, 52, 252, 193, 224, 52, 252, 193, 226, 52, 252, 193, 228, 52, 252, 193, 230, 52, 252, 193, 232, 52, 252, 193, 234, 52, 252, 193, 236, 52, 252, 193, 238, 52, 252, 193, 240, 52, 252, 193, 242, 52, 252, 193, 244, 52, 252, 193, 246, 52, 252, 193, 248, 52, 252, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 193, 12, 53, 252, 193, 192, 57, 252, 193, 84, 69, 252, 193, 132, 73, 252, 193, 132, 75, 252, 193, 94, 80, 252, 193, 244, 81, 252, 193, 216, 83, 252, 193, 214, 85, 252, 193, 162, 86, 252, 193, 220, 92, 252, 193, 226, 95, 252, 193, 40, 100, 252, 193, 192, 103, 252, 193, 152, 107, 252, 193, 214, 110, 252, 193, 2, 111, 252, 193, 232, 111, 252, 193, 108, 115, 252, 193, 56, 116, 252, 193, 128, 136, 252, 193, 24, 137, 252, 193, 114, 139, 252, 193, 154, 142, 252, 193, 232, 143, 252, 193, 42, 148, 252, 193, 84, 152, 252, 193, 132, 152, 252, 193, 52, 153, 252, 193, 148, 160, 252, 193, 150, 160, 252, 193, 152, 160, 252, 193, 80, 53, 252, 193, 82, 53, 252, 193, 84, 53, 252, 193, 86, 53, 252, 193, 88, 53, 252, 193, 90, 53, 252, 193, 92, 53, 252, 193, 94, 53, 252, 193, 96, 53, 252, 193, 98, 53, 252, 193, 100, 53, 252, 193, 102, 53, 252, 193, 104, 53, 252, 193, 224, 54, 252, 193, 226, 54, 252, 193, 228, 54, 252, 193, 230, 54, 252, 193, 232, 54, 252, 193, 234, 54, 252, 193, 236, 54, 252, 193, 238, 54, 252, 193, 240, 54, 252, 193, 242, 54, 252, 193, 244, 54, 252, 193, 246, 54, 252, 193, 248, 54, 252, 206, 32, 50, 0, 206, 32, 50, 0, 206, 32, 50, 0, 206, 32, 50, 0, 206, 32, 50, 0, 206, 32, 50, 0, 193, 122, 53, 252, 193, 124, 53, 252, 193, 126, 53, 252, 193, 128, 53, 252, 193, 130, 53, 252, 193, 132, 53, 252, 193, 134, 53, 252, 193, 136, 53, 252, 193, 138, 53, 252, 193, 140, 53, 252, 193, 142, 53, 252, 193, 144, 53, 252, 193, 146, 53, 252, 193, 148, 53, 252, 193, 150, 53, 252, 193, 152, 53, 252, 193, 154, 53, 252, 193, 156, 53, 252, 193, 158, 53, 252, 193, 160, 53, 252, 193, 162, 53, 252, 193, 164, 53, 252, 193, 166, 53, 252, 193, 168, 53, 252, 193, 170, 53, 252, 193, 172, 53, 252, 193, 174, 53, 252, 193, 176, 53, 252, 193, 178, 53, 252, 193, 180, 53, 252, 193, 182, 53, 252, 193, 184, 53, 252, 193, 186, 53, 252, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 206, 192, 49, 0, 193, 188, 53, 252, 193, 86, 55, 252, 193, 88, 55, 252, 193, 90, 55, 252, 193, 92, 55, 252, 193, 94, 55, 252, 193, 96, 55, 252, 193, 98, 55, 252, 193, 100, 55, 252, 193, 102, 55, 252, 193, 104, 55, 252, 193, 182, 56, 252, 193, 106, 55, 252, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 64, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 206, 96, 49, 0, 193, 60, 56, 252, 193, 62, 56, 252, 193, 64, 56, 252, 193, 66, 56, 252, 193, 68, 56, 252, 193, 70, 56, 252, 193, 72, 56, 252, 193, 74, 56, 252, 193, 76, 56, 252, 193, 78, 56, 252, 193, 80, 56, 252, 193, 82, 56, 252, 193, 84, 56, 252, 193, 86, 56, 252, 193, 162, 57, 252, 193, 164, 57, 252, 193, 166, 57, 252, 193, 168, 57, 252, 193, 170, 57, 252, 193, 172, 57, 252, 193, 174, 57, 252, 193, 176, 57, 252, 193, 178, 57, 252, 193, 180, 57, 252, 193, 182, 57, 252, 193, 184, 57, 252, 193, 186, 57, 252, 193, 188, 57, 252, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 206, 128, 49, 0, 193, 88, 56, 252, 193, 90, 56, 252, 193, 92, 56, 252, 193, 94, 56, 252, 193, 96, 56, 252, 193, 98, 56, 252, 193, 100, 56, 252, 193, 102, 56, 252, 193, 104, 56, 252, 193, 194, 57, 252, 193, 196, 57, 252, 193, 198, 57, 252, 193, 200, 57, 252, 193, 202, 57, 252, 193, 204, 57, 252, 193, 206, 57, 252, 193, 208, 57, 252, 193, 210, 57, 252, 193, 212, 57, 252, 193, 214, 57, 252, 193, 216, 57, 252, 193, 218, 57, 252, 193, 8, 59, 252, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 206, 160, 49, 0, 193, 106, 56, 252, 193, 108, 56, 252, 193, 110, 56, 252, 193, 112, 56, 252, 193, 114, 56, 252, 193, 116, 56, 252, 193, 118, 56, 252, 193, 120, 56, 252, 193, 122, 56, 252, 193, 124, 56, 252, 193, 126, 56, 252, 193, 128, 56, 252, 193, 130, 56, 252, 193, 132, 56, 252, 193, 134, 56, 252, 193, 136, 56, 252, 193, 138, 56, 252, 193, 140, 56, 252, 193, 142, 56, 252, 193, 144, 56, 252, 193, 146, 56, 252, 193, 148, 56, 252, 193, 150, 56, 252, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 206, 224, 49, 0, 193, 4, 184, 253, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 0, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 193, 58, 56, 252, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 193, 152, 56, 252, 193, 154, 56, 252, 193, 220, 57, 252, 193, 222, 57, 252, 193, 224, 57, 252, 193, 226, 57, 252, 193, 228, 57, 252, 193, 230, 57, 252, 193, 232, 57, 252, 193, 234, 57, 252, 193, 236, 57, 252, 193, 238, 57, 252, 193, 240, 57, 252, 193, 242, 57, 252, 193, 244, 57, 252, 193, 246, 57, 252, 193, 248, 57, 252, 193, 250, 57, 252, 193, 252, 57, 252, 193, 20, 59, 252, 193, 22, 59, 252, 193, 24, 59, 252, 193, 26, 59, 252, 193, 28, 59, 252, 193, 30, 59, 252, 193, 32, 59, 252, 193, 34, 59, 252, 193, 36, 59, 252, 193, 38, 59, 252, 193, 120, 60, 252, 193, 122, 60, 252, 193, 124, 60, 252, 206, 32, 50, 0, 206, 32, 50, 0, 206, 32, 50, 0, 206, 32, 50, 0, 206, 32, 50, 0, 206, 32, 50, 0, 206, 32, 50, 0, 193, 254, 57, 252, 193, 2, 58, 252, 193, 4, 58, 252, 193, 6, 58, 252, 193, 8, 58, 252, 193, 10, 58, 252, 193, 12, 58, 252, 193, 14, 58, 252, 193, 16, 58, 252, 193, 40, 59, 252, 193, 42, 59, 252, 193, 44, 59, 252, 193, 46, 59, 252, 193, 48, 59, 252, 193, 50, 59, 252, 193, 52, 59, 252, 193, 54, 59, 252, 193, 56, 59, 252, 193, 150, 60, 252, 193, 152, 60, 252, 193, 154, 60, 252, 193, 156, 60, 252, 193, 158, 60, 252, 193, 160, 60, 252, 193, 162, 60, 252, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 64, 50, 0, 206, 160, 50, 0, 206, 160, 50, 0, 206, 160, 50, 0, 206, 160, 50, 0, 206, 160, 50, 0, 206, 160, 50, 0, 206, 160, 50, 0, 206, 160, 50, 0, 193, 246, 56, 252, 193, 42, 58, 252, 193, 44, 58, 252, 193, 46, 58, 252, 193, 48, 58, 252, 193, 50, 58, 252, 193, 52, 58, 252, 193, 54, 58, 252, 193, 56, 58, 252, 193, 58, 58, 252, 193, 60, 58, 252, 193, 62, 58, 252, 193, 64, 58, 252, 193, 66, 58, 252, 193, 68, 58, 252, 193, 70, 58, 252, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 193, 248, 56, 252, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 193, 252, 56, 252, 193, 254, 56, 252, 193, 2, 57, 252, 193, 4, 57, 252, 193, 6, 57, 252, 193, 8, 57, 252, 193, 10, 57, 252, 193, 12, 57, 252, 193, 14, 57, 252, 193, 16, 57, 252, 193, 18, 57, 252, 193, 78, 58, 252, 193, 80, 58, 252, 193, 82, 58, 252, 193, 84, 58, 252, 193, 86, 58, 252, 193, 88, 58, 252, 193, 90, 58, 252, 193, 92, 58, 252, 193, 140, 59, 252, 193, 142, 59, 252, 193, 144, 59, 252, 193, 146, 59, 252, 193, 148, 59, 252, 193, 150, 59, 252, 193, 152, 59, 252, 193, 154, 59, 252, 193, 156, 59, 252, 193, 158, 59, 252, 193, 160, 59, 252, 193, 222, 60, 252, 193, 224, 60, 252, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 206, 96, 50, 0, 193, 94, 58, 252, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 206, 128, 50, 0, 193, 190, 57, 252, 193, 244, 58, 252, 193, 246, 58, 252, 193, 248, 58, 252, 193, 250, 58, 252, 193, 252, 58, 252, 193, 254, 58, 252, 193, 2, 59, 252, 193, 4, 59, 252, 193, 6, 59, 252, 193, 84, 60, 252, 193, 86, 60, 252, 193, 88, 60, 252, 193, 90, 60, 252, 193, 92, 60, 252, 193, 94, 60, 252, 193, 96, 60, 252, 193, 98, 60, 252, 193, 100, 60, 252, 193, 102, 60, 252, 193, 74, 61, 252, 193, 76, 61, 252, 193, 78, 61, 252, 193, 80, 61, 252, 193, 82, 61, 252, 193, 84, 61, 252, 193, 86, 61, 252, 193, 88, 61, 252, 193, 6, 62, 252, 193, 8, 62, 252, 193, 10, 62, 252, 193, 12, 62, 252, 193, 160, 213, 253, 193, 172, 213, 253, 193, 198, 213, 253, 193, 210, 213, 253, 193, 212, 213, 253, 193, 220, 213, 253, 193, 222, 213, 253, 193, 238, 213, 253, 193, 240, 213, 253, 193, 242, 213, 253, 193, 250, 213, 253, 193, 252, 213, 253, 193, 110, 214, 253, 193, 124, 214, 253, 193, 132, 214, 253, 193, 158, 214, 253, 193, 218, 214, 253, 193, 252, 214, 253, 193, 254, 214, 253, 193, 18, 215, 253, 193, 40, 215, 253, 193, 42, 215, 253, 193, 44, 215, 253, 193, 46, 215, 253, 193, 48, 215, 253, 193, 72, 58, 252, 193, 84, 215, 253, 193, 86, 215, 253, 193, 88, 215, 253, 193, 90, 215, 253, 193, 92, 215, 253, 193, 128, 215, 253, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 206, 192, 50, 0, 193, 162, 59, 252, 193, 146, 58, 252, 193, 148, 58, 252, 193, 150, 58, 252, 193, 152, 58, 252, 193, 154, 58, 252, 193, 156, 58, 252, 193, 158, 58, 252, 193, 160, 58, 252, 193, 162, 58, 252, 193, 164, 58, 252, 193, 166, 58, 252, 193, 168, 58, 252, 193, 170, 58, 252, 193, 172, 58, 252, 193, 174, 58, 252, 193, 176, 58, 252, 193, 178, 58, 252, 193, 180, 58, 252, 193, 182, 58, 252, 193, 184, 58, 252, 193, 186, 58, 252, 193, 188, 58, 252, 193, 96, 236, 252, 193, 190, 58, 252, 193, 228, 140, 253, 193, 192, 58, 252, 193, 194, 58, 252, 193, 196, 58, 252, 193, 198, 58, 252, 193, 200, 58, 252, 193, 202, 58, 252, 193, 204, 58, 252, 193, 206, 58, 252, 193, 208, 58, 252, 193, 210, 58, 252, 193, 212, 58, 252, 193, 214, 58, 252, 193, 216, 58, 252, 193, 218, 58, 252, 193, 220, 58, 252, 193, 222, 58, 252, 193, 224, 58, 252, 193, 226, 58, 252, 193, 228, 58, 252, 193, 230, 58, 252, 193, 232, 58, 252, 193, 234, 58, 252, 193, 236, 58, 252, 193, 238, 58, 252, 193, 240, 58, 252, 193, 242, 58, 252, 193, 164, 59, 252, 193, 166, 59, 252, 193, 168, 59, 252, 193, 10, 59, 252, 193, 12, 59, 252, 193, 14, 59, 252, 193, 16, 59, 252, 193, 18, 59, 252, 193, 104, 60, 252, 193, 106, 60, 252, 193, 108, 60, 252, 193, 110, 60, 252, 193, 112, 60, 252, 193, 114, 60, 252, 193, 116, 60, 252, 193, 118, 60, 252, 193, 90, 61, 252, 193, 92, 61, 252, 193, 94, 61, 252, 193, 96, 61, 252, 193, 16, 62, 252, 193, 18, 62, 252, 193, 20, 62, 252, 193, 148, 62, 252, 193, 150, 62, 252, 193, 152, 62, 252, 193, 154, 62, 252, 193, 156, 62, 252, 193, 44, 63, 252, 193, 46, 63, 252, 193, 48, 63, 252, 193, 50, 63, 252, 193, 52, 63, 252, 193, 214, 63, 252, 193, 6, 64, 252, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 206, 224, 50, 0, 193, 170, 60, 252, 193, 170, 59, 252, 193, 172, 59, 252, 193, 174, 59, 252, 193, 176, 59, 252, 193, 178, 59, 252, 193, 180, 59, 252, 193, 182, 59, 252, 193, 184, 59, 252, 193, 186, 59, 252, 193, 188, 59, 252, 193, 190, 59, 252, 193, 192, 59, 252, 193, 194, 59, 252, 193, 196, 59, 252, 193, 198, 59, 252, 193, 200, 59, 252, 193, 202, 59, 252, 193, 204, 59, 252, 193, 206, 59, 252, 193, 208, 59, 252, 193, 210, 59, 252, 193, 212, 59, 252, 193, 214, 59, 252, 193, 216, 59, 252, 193, 218, 59, 252, 193, 220, 59, 252, 193, 222, 59, 252, 193, 224, 59, 252, 193, 226, 59, 252, 193, 228, 59, 252, 193, 230, 59, 252, 193, 232, 59, 252, 193, 234, 59, 252, 193, 236, 59, 252, 193, 238, 59, 252, 193, 240, 59, 252, 193, 242, 59, 252, 193, 244, 59, 252, 193, 246, 59, 252, 193, 248, 59, 252, 193, 250, 59, 252, 193, 252, 59, 252, 193, 254, 59, 252, 193, 2, 60, 252, 193, 4, 60, 252, 193, 6, 60, 252, 193, 8, 60, 252, 193, 10, 60, 252, 193, 12, 60, 252, 193, 14, 60, 252, 193, 16, 60, 252, 193, 18, 60, 252, 193, 20, 60, 252, 193, 22, 60, 252, 193, 24, 60, 252, 193, 26, 60, 252, 193, 28, 60, 252, 193, 30, 60, 252, 193, 32, 60, 252, 193, 34, 60, 252, 193, 36, 60, 252, 193, 38, 60, 252, 193, 40, 60, 252, 193, 42, 60, 252, 193, 126, 60, 252, 193, 128, 60, 252, 193, 130, 60, 252, 193, 132, 60, 252, 193, 134, 60, 252, 193, 136, 60, 252, 193, 138, 60, 252, 193, 140, 60, 252, 193, 142, 60, 252, 193, 144, 60, 252, 193, 146, 60, 252, 193, 148, 60, 252, 193, 98, 61, 252, 193, 100, 61, 252, 193, 102, 61, 252, 193, 104, 61, 252, 193, 106, 61, 252, 193, 108, 61, 252, 193, 110, 61, 252, 193, 112, 61, 252, 193, 114, 61, 252, 193, 116, 61, 252, 193, 118, 61, 252, 193, 120, 61, 252, 193, 122, 61, 252, 193, 124, 61, 252, 193, 22, 62, 252, 193, 24, 62, 252, 193, 26, 62, 252, 193, 28, 62, 252, 193, 30, 62, 252, 193, 32, 62, 252, 193, 164, 60, 252, 193, 166, 60, 252, 193, 168, 60, 252, 193, 126, 61, 252, 193, 128, 61, 252, 193, 130, 61, 252, 193, 132, 61, 252, 193, 134, 61, 252, 193, 136, 61, 252, 193, 138, 61, 252, 193, 38, 62, 252, 193, 40, 62, 252, 193, 42, 62, 252, 193, 44, 62, 252, 193, 46, 62, 252, 193, 48, 62, 252, 193, 162, 62, 252, 193, 164, 62, 252, 193, 166, 62, 252, 193, 168, 62, 252, 193, 60, 63, 252, 193, 62, 63, 252, 193, 64, 63, 252, 193, 66, 63, 252, 193, 68, 63, 252, 193, 70, 63, 252, 193, 126, 63, 252, 193, 234, 63, 252, 193, 96, 64, 252, 193, 140, 64, 252, 193, 164, 64, 252, 193, 166, 64, 252, 193, 172, 60, 252, 193, 174, 60, 252, 193, 176, 60, 252, 193, 178, 60, 252, 193, 180, 60, 252, 193, 182, 60, 252, 193, 184, 60, 252, 193, 186, 60, 252, 193, 188, 60, 252, 193, 190, 60, 252, 193, 192, 60, 252, 193, 194, 60, 252, 193, 196, 60, 252, 193, 198, 60, 252, 193, 200, 60, 252, 193, 202, 60, 252, 193, 204, 60, 252, 193, 206, 60, 252, 193, 208, 60, 252, 193, 210, 60, 252, 193, 212, 60, 252, 193, 214, 60, 252, 193, 216, 60, 252, 193, 218, 60, 252, 206, 32, 51, 0, 206, 32, 51, 0, 206, 32, 51, 0, 206, 32, 51, 0, 206, 32, 51, 0, 206, 32, 51, 0, 206, 32, 51, 0, 206, 32, 51, 0, 193, 226, 60, 252, 193, 228, 60, 252, 193, 230, 60, 252, 193, 232, 60, 252, 193, 164, 61, 252, 193, 166, 61, 252, 193, 72, 62, 252, 193, 74, 62, 252, 193, 190, 62, 252, 193, 76, 62, 252, 193, 192, 62, 252, 193, 194, 62, 252, 193, 196, 62, 252, 193, 78, 63, 252, 193, 138, 63, 252, 193, 176, 63, 252, 193, 178, 63, 252, 193, 10, 64, 252, 193, 72, 64, 252, 193, 74, 64, 252, 193, 168, 64, 252, 193, 2, 65, 252, 193, 170, 65, 252, 193, 178, 65, 252, 193, 188, 65, 252, 193, 204, 65, 252, 193, 222, 65, 252, 193, 52, 66, 252, 193, 120, 66, 252, 193, 10, 67, 252, 193, 12, 67, 252, 193, 66, 67, 252, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 206, 0, 51, 0, 193, 168, 61, 252, 193, 170, 61, 252, 193, 172, 61, 252, 193, 174, 61, 252, 193, 176, 61, 252, 193, 178, 61, 252, 193, 180, 61, 252, 193, 182, 61, 252, 193, 184, 61, 252, 193, 186, 61, 252, 193, 188, 61, 252, 193, 190, 61, 252, 193, 192, 61, 252, 193, 194, 61, 252, 193, 196, 61, 252, 193, 198, 61, 252, 193, 200, 61, 252, 193, 202, 61, 252, 193, 204, 61, 252, 193, 206, 61, 252, 193, 208, 61, 252, 193, 210, 61, 252, 193, 212, 61, 252, 193, 214, 61, 252, 193, 216, 61, 252, 193, 218, 61, 252, 193, 220, 61, 252, 193, 222, 61, 252, 206, 32, 51, 0, 206, 32, 51, 0, 206, 32, 51, 0, 206, 32, 51, 0, 193, 50, 62, 252, 193, 52, 62, 252, 193, 54, 62, 252, 193, 56, 62, 252, 193, 58, 62, 252, 193, 60, 62, 252, 193, 62, 62, 252, 193, 170, 62, 252, 193, 64, 62, 252, 193, 66, 62, 252, 193, 68, 62, 252, 193, 172, 62, 252, 193, 174, 62, 252, 193, 176, 62, 252, 193, 178, 62, 252, 193, 180, 62, 252, 193, 182, 62, 252, 193, 70, 62, 252, 193, 184, 62, 252, 193, 186, 62, 252, 193, 188, 62, 252, 193, 72, 63, 252, 193, 74, 63, 252, 193, 76, 63, 252, 193, 128, 63, 252, 193, 130, 63, 252, 193, 132, 63, 252, 193, 134, 63, 252, 193, 224, 61, 252, 193, 226, 61, 252, 193, 228, 61, 252, 193, 248, 213, 252, 193, 230, 61, 252, 193, 232, 61, 252, 193, 234, 61, 252, 193, 236, 61, 252, 193, 238, 61, 252, 193, 240, 61, 252, 193, 242, 61, 252, 193, 244, 61, 252, 193, 246, 61, 252, 193, 248, 61, 252, 193, 250, 61, 252, 193, 252, 61, 252, 193, 254, 61, 252, 193, 2, 62, 252, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 193, 248, 217, 252, 193, 254, 217, 252, 193, 2, 218, 252, 193, 4, 218, 252, 193, 6, 218, 252, 193, 8, 218, 252, 193, 10, 218, 252, 193, 12, 218, 252, 193, 4, 62, 252, 193, 24, 218, 252, 193, 26, 218, 252, 193, 28, 218, 252, 193, 30, 218, 252, 193, 38, 218, 252, 193, 40, 218, 252, 193, 32, 218, 252, 193, 42, 218, 252, 193, 44, 218, 252, 193, 46, 218, 252, 193, 48, 218, 252, 193, 50, 218, 252, 193, 52, 218, 252, 193, 54, 218, 252, 193, 56, 218, 252, 193, 58, 218, 252, 193, 60, 218, 252, 193, 64, 218, 252, 193, 66, 218, 252, 193, 72, 218, 252, 193, 76, 218, 252, 193, 78, 218, 252, 193, 80, 218, 252, 193, 14, 62, 252, 193, 144, 62, 252, 193, 146, 62, 252, 193, 38, 63, 252, 193, 40, 63, 252, 193, 42, 63, 252, 193, 166, 63, 252, 193, 168, 63, 252, 193, 212, 63, 252, 193, 228, 63, 252, 193, 2, 64, 252, 193, 4, 64, 252, 193, 24, 64, 252, 193, 66, 64, 252, 193, 134, 64, 252, 193, 178, 64, 252, 193, 204, 64, 252, 193, 206, 64, 252, 193, 208, 64, 252, 193, 210, 64, 252, 193, 238, 64, 252, 193, 240, 64, 252, 193, 40, 65, 252, 193, 42, 65, 252, 193, 86, 65, 252, 193, 88, 65, 252, 193, 148, 65, 252, 193, 184, 65, 252, 193, 192, 65, 252, 193, 226, 65, 252, 193, 110, 66, 252, 193, 138, 66, 252, 193, 34, 62, 252, 193, 36, 62, 252, 193, 158, 62, 252, 193, 160, 62, 252, 193, 54, 63, 252, 193, 56, 63, 252, 193, 58, 63, 252, 193, 170, 63, 252, 193, 230, 63, 252, 193, 232, 63, 252, 193, 94, 64, 252, 193, 138, 64, 252, 193, 156, 64, 252, 193, 158, 64, 252, 193, 160, 64, 252, 193, 162, 64, 252, 193, 184, 64, 252, 193, 216, 64, 252, 193, 246, 64, 252, 193, 248, 64, 252, 193, 18, 65, 252, 193, 20, 65, 252, 193, 48, 65, 252, 193, 64, 65, 252, 193, 66, 65, 252, 193, 104, 65, 252, 193, 154, 65, 252, 193, 162, 65, 252, 193, 194, 65, 252, 193, 216, 65, 252, 193, 28, 66, 252, 193, 30, 66, 252, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 206, 64, 51, 0, 193, 198, 62, 252, 193, 200, 62, 252, 193, 202, 62, 252, 193, 204, 62, 252, 193, 206, 62, 252, 193, 208, 62, 252, 193, 210, 62, 252, 193, 212, 62, 252, 193, 214, 62, 252, 193, 216, 62, 252, 193, 218, 62, 252, 193, 220, 62, 252, 193, 222, 62, 252, 193, 224, 62, 252, 193, 226, 62, 252, 193, 228, 62, 252, 193, 230, 62, 252, 193, 232, 62, 252, 193, 234, 62, 252, 193, 236, 62, 252, 193, 238, 62, 252, 193, 240, 62, 252, 193, 242, 62, 252, 193, 244, 62, 252, 193, 246, 62, 252, 193, 248, 62, 252, 193, 250, 62, 252, 193, 252, 62, 252, 193, 254, 62, 252, 193, 2, 63, 252, 193, 4, 63, 252, 193, 6, 63, 252, 193, 8, 63, 252, 193, 10, 63, 252, 193, 12, 63, 252, 193, 14, 63, 252, 193, 16, 63, 252, 193, 18, 63, 252, 193, 20, 63, 252, 193, 22, 63, 252, 193, 24, 63, 252, 193, 26, 63, 252, 193, 28, 63, 252, 193, 30, 63, 252, 193, 32, 63, 252, 193, 34, 63, 252, 193, 80, 63, 252, 193, 82, 63, 252, 193, 84, 63, 252, 193, 86, 63, 252, 193, 88, 63, 252, 193, 90, 63, 252, 193, 92, 63, 252, 193, 36, 63, 252, 193, 140, 63, 252, 193, 94, 63, 252, 193, 96, 63, 252, 193, 98, 63, 252, 193, 100, 63, 252, 193, 102, 63, 252, 193, 104, 63, 252, 193, 106, 63, 252, 193, 108, 63, 252, 193, 110, 63, 252, 193, 112, 63, 252, 193, 114, 63, 252, 193, 116, 63, 252, 193, 118, 63, 252, 193, 120, 63, 252, 193, 122, 63, 252, 193, 124, 63, 252, 193, 142, 63, 252, 193, 144, 63, 252, 193, 180, 63, 252, 193, 146, 63, 252, 193, 148, 63, 252, 193, 150, 63, 252, 193, 136, 63, 252, 193, 172, 63, 252, 193, 174, 63, 252, 193, 216, 63, 252, 193, 236, 63, 252, 193, 238, 63, 252, 193, 240, 63, 252, 193, 26, 64, 252, 193, 36, 64, 252, 193, 42, 64, 252, 193, 48, 64, 252, 193, 50, 64, 252, 193, 68, 64, 252, 193, 70, 64, 252, 193, 98, 64, 252, 193, 100, 64, 252, 193, 102, 64, 252, 193, 142, 64, 252, 193, 188, 64, 252, 193, 254, 64, 252, 193, 24, 65, 252, 193, 94, 65, 252, 193, 150, 65, 252, 193, 152, 65, 252, 193, 164, 65, 252, 193, 176, 65, 252, 193, 202, 65, 252, 193, 218, 65, 252, 193, 228, 65, 252, 193, 238, 65, 252, 193, 248, 65, 252, 193, 16, 66, 252, 193, 152, 63, 252, 193, 154, 63, 252, 193, 156, 63, 252, 193, 158, 63, 252, 193, 160, 63, 252, 193, 162, 63, 252, 193, 164, 63, 252, 193, 182, 63, 252, 193, 184, 63, 252, 193, 186, 63, 252, 193, 188, 63, 252, 193, 190, 63, 252, 193, 192, 63, 252, 193, 194, 63, 252, 193, 196, 63, 252, 193, 30, 215, 252, 193, 198, 63, 252, 193, 200, 63, 252, 193, 202, 63, 252, 193, 204, 63, 252, 193, 206, 63, 252, 193, 208, 63, 252, 193, 210, 63, 252, 193, 218, 63, 252, 193, 220, 63, 252, 193, 222, 63, 252, 193, 224, 63, 252, 193, 226, 63, 252, 193, 242, 63, 252, 193, 244, 63, 252, 193, 246, 63, 252, 193, 70, 218, 252, 193, 248, 63, 252, 193, 250, 63, 252, 193, 252, 63, 252, 193, 254, 63, 252, 193, 12, 64, 252, 193, 14, 64, 252, 193, 16, 64, 252, 193, 18, 64, 252, 193, 20, 64, 252, 193, 22, 64, 252, 193, 28, 64, 252, 193, 38, 64, 252, 193, 30, 64, 252, 193, 40, 64, 252, 193, 44, 64, 252, 193, 52, 64, 252, 193, 54, 64, 252, 193, 56, 64, 252, 193, 58, 64, 252, 193, 60, 64, 252, 193, 62, 64, 252, 193, 64, 64, 252, 193, 76, 64, 252, 193, 78, 64, 252, 193, 80, 64, 252, 193, 82, 64, 252, 193, 84, 64, 252, 193, 86, 64, 252, 193, 88, 64, 252, 193, 104, 64, 252, 193, 106, 64, 252, 193, 108, 64, 252, 193, 8, 64, 252, 193, 32, 64, 252, 193, 34, 64, 252, 193, 46, 64, 252, 193, 90, 64, 252, 193, 92, 64, 252, 193, 136, 64, 252, 193, 152, 64, 252, 193, 154, 64, 252, 193, 180, 64, 252, 193, 182, 64, 252, 193, 212, 64, 252, 193, 214, 64, 252, 193, 242, 64, 252, 193, 244, 64, 252, 193, 14, 65, 252, 193, 16, 65, 252, 193, 44, 65, 252, 193, 46, 65, 252, 193, 60, 65, 252, 193, 62, 65, 252, 193, 90, 65, 252, 193, 92, 65, 252, 193, 112, 65, 252, 193, 126, 65, 252, 193, 128, 65, 252, 193, 146, 65, 252, 193, 160, 65, 252, 193, 212, 65, 252, 193, 6, 66, 252, 193, 12, 66, 252, 193, 62, 66, 252, 193, 186, 64, 252, 193, 218, 64, 252, 193, 220, 64, 252, 193, 222, 64, 252, 193, 224, 64, 252, 193, 250, 64, 252, 193, 252, 64, 252, 193, 50, 65, 252, 193, 68, 65, 252, 193, 138, 65, 252, 193, 140, 65, 252, 193, 174, 65, 252, 193, 186, 65, 252, 193, 200, 65, 252, 193, 44, 66, 252, 193, 46, 66, 252, 193, 48, 66, 252, 193, 72, 66, 252, 193, 90, 66, 252, 193, 116, 66, 252, 193, 118, 66, 252, 193, 152, 66, 252, 193, 186, 66, 252, 193, 246, 66, 252, 193, 248, 66, 252, 193, 30, 67, 252, 193, 32, 67, 252, 193, 60, 67, 252, 193, 62, 67, 252, 193, 64, 67, 252, 193, 108, 67, 252, 193, 152, 67, 252, 193, 234, 64, 252, 193, 236, 64, 252, 193, 4, 65, 252, 193, 6, 65, 252, 193, 8, 65, 252, 193, 10, 65, 252, 193, 12, 65, 252, 193, 26, 65, 252, 193, 28, 65, 252, 193, 30, 65, 252, 193, 32, 65, 252, 193, 34, 65, 252, 193, 36, 65, 252, 193, 38, 65, 252, 193, 52, 65, 252, 193, 54, 65, 252, 193, 56, 65, 252, 193, 58, 65, 252, 193, 70, 65, 252, 193, 72, 65, 252, 193, 74, 65, 252, 193, 76, 65, 252, 193, 78, 65, 252, 193, 80, 65, 252, 193, 82, 65, 252, 193, 96, 65, 252, 193, 106, 65, 252, 193, 98, 65, 252, 193, 84, 65, 252, 193, 100, 65, 252, 193, 102, 65, 252, 193, 108, 65, 252, 193, 110, 65, 252, 193, 182, 189, 252, 193, 114, 65, 252, 193, 116, 65, 252, 193, 118, 65, 252, 193, 120, 65, 252, 193, 122, 65, 252, 193, 124, 65, 252, 193, 130, 65, 252, 193, 132, 65, 252, 193, 134, 65, 252, 193, 136, 65, 252, 193, 142, 65, 252, 193, 144, 65, 252, 193, 156, 65, 252, 193, 158, 65, 252, 193, 166, 65, 252, 193, 168, 65, 252, 193, 172, 65, 252, 193, 180, 65, 252, 193, 182, 65, 252, 193, 190, 65, 252, 193, 196, 65, 252, 193, 198, 65, 252, 193, 206, 65, 252, 193, 208, 65, 252, 193, 210, 65, 252, 193, 214, 65, 252, 193, 220, 65, 252, 193, 224, 65, 252, 193, 230, 65, 252, 193, 232, 65, 252, 193, 234, 65, 252, 193, 236, 65, 252, 193, 240, 65, 252, 193, 242, 65, 252, 193, 244, 65, 252, 193, 246, 65, 252, 193, 252, 65, 252, 193, 254, 65, 252, 193, 2, 66, 252, 193, 4, 66, 252, 193, 18, 66, 252, 193, 20, 66, 252, 193, 22, 66, 252, 193, 24, 66, 252, 193, 26, 66, 252, 193, 34, 66, 252, 193, 36, 66, 252, 193, 38, 66, 252, 193, 54, 66, 252, 193, 56, 66, 252, 193, 58, 66, 252, 193, 60, 66, 252, 193, 76, 66, 252, 193, 78, 66, 252, 193, 80, 66, 252, 193, 82, 66, 252, 193, 84, 66, 252, 206, 96, 51, 0, 206, 96, 51, 0, 206, 96, 51, 0, 206, 96, 51, 0, 206, 96, 51, 0, 193, 32, 66, 252, 193, 50, 66, 252, 193, 74, 66, 252, 193, 154, 66, 252, 193, 162, 66, 252, 193, 174, 66, 252, 193, 206, 66, 252, 193, 252, 66, 252, 193, 8, 67, 252, 193, 34, 67, 252, 193, 36, 67, 252, 193, 110, 67, 252, 193, 112, 67, 252, 193, 114, 67, 252, 193, 154, 67, 252, 193, 156, 67, 252, 193, 158, 67, 252, 193, 160, 67, 252, 193, 162, 67, 252, 193, 184, 67, 252, 193, 186, 67, 252, 193, 202, 67, 252, 193, 204, 67, 252, 193, 206, 67, 252, 193, 18, 68, 252, 193, 28, 68, 252, 193, 68, 68, 252, 193, 94, 68, 252, 193, 96, 68, 252, 193, 102, 68, 252, 193, 104, 68, 252, 193, 106, 68, 252, 193, 40, 66, 252, 193, 42, 66, 252, 193, 68, 66, 252, 193, 70, 66, 252, 193, 86, 66, 252, 193, 88, 66, 252, 193, 114, 66, 252, 193, 216, 66, 252, 193, 222, 66, 252, 193, 28, 67, 252, 193, 56, 67, 252, 193, 58, 67, 252, 193, 106, 67, 252, 193, 4, 68, 252, 193, 16, 68, 252, 193, 40, 68, 252, 193, 44, 68, 252, 193, 56, 68, 252, 193, 80, 68, 252, 193, 100, 68, 252, 193, 118, 68, 252, 193, 238, 68, 252, 193, 240, 68, 252, 193, 242, 68, 252, 193, 244, 68, 252, 193, 246, 68, 252, 193, 88, 69, 252, 193, 90, 69, 252, 193, 92, 69, 252, 193, 94, 69, 252, 193, 158, 69, 252, 193, 160, 69, 252, 193, 64, 66, 252, 193, 66, 66, 252, 193, 112, 66, 252, 193, 140, 66, 252, 193, 142, 66, 252, 193, 144, 66, 252, 193, 146, 66, 252, 193, 148, 66, 252, 193, 150, 66, 252, 193, 172, 66, 252, 193, 184, 66, 252, 193, 200, 66, 252, 193, 202, 66, 252, 193, 204, 66, 252, 193, 220, 66, 252, 193, 228, 66, 252, 193, 232, 66, 252, 193, 234, 66, 252, 193, 250, 66, 252, 193, 24, 67, 252, 193, 26, 67, 252, 193, 104, 67, 252, 193, 26, 68, 252, 193, 54, 68, 252, 193, 86, 68, 252, 193, 116, 68, 252, 193, 146, 68, 252, 193, 234, 68, 252, 193, 236, 68, 252, 193, 86, 69, 252, 193, 154, 69, 252, 193, 156, 69, 252, 206, 96, 51, 0, 206, 96, 51, 0, 206, 96, 51, 0, 206, 96, 51, 0, 193, 122, 66, 252, 193, 124, 66, 252, 193, 126, 66, 252, 193, 128, 66, 252, 193, 130, 66, 252, 193, 132, 66, 252, 193, 134, 66, 252, 193, 136, 66, 252, 193, 156, 66, 252, 193, 158, 66, 252, 193, 160, 66, 252, 193, 164, 66, 252, 193, 166, 66, 252, 193, 168, 66, 252, 193, 170, 66, 252, 193, 176, 66, 252, 193, 178, 66, 252, 193, 180, 66, 252, 193, 188, 66, 252, 193, 190, 66, 252, 193, 192, 66, 252, 193, 194, 66, 252, 193, 196, 66, 252, 193, 198, 66, 252, 193, 208, 66, 252, 193, 210, 66, 252, 193, 212, 66, 252, 193, 214, 66, 252, 193, 182, 66, 252, 193, 218, 66, 252, 193, 84, 67, 252, 193, 100, 67, 252, 193, 102, 67, 252, 193, 150, 67, 252, 193, 244, 67, 252, 193, 38, 68, 252, 193, 78, 68, 252, 193, 144, 68, 252, 193, 80, 69, 252, 193, 82, 69, 252, 193, 148, 69, 252, 193, 150, 69, 252, 193, 152, 69, 252, 193, 78, 70, 252, 193, 182, 70, 252, 193, 24, 71, 252, 193, 26, 71, 252, 193, 28, 71, 252, 193, 104, 71, 252, 193, 160, 71, 252, 193, 6, 72, 252, 193, 238, 72, 252, 193, 128, 73, 252, 193, 130, 73, 252, 193, 46, 74, 252, 193, 48, 74, 252, 193, 212, 74, 252, 193, 214, 74, 252, 193, 124, 75, 252, 193, 126, 75, 252, 193, 224, 66, 252, 193, 226, 66, 252, 193, 230, 66, 252, 193, 236, 66, 252, 193, 238, 66, 252, 193, 240, 66, 252, 193, 242, 66, 252, 193, 244, 66, 252, 193, 254, 66, 252, 193, 2, 67, 252, 193, 4, 67, 252, 193, 6, 67, 252, 193, 14, 67, 252, 193, 16, 67, 252, 193, 18, 67, 252, 193, 20, 67, 252, 193, 22, 67, 252, 193, 38, 67, 252, 193, 40, 67, 252, 193, 42, 67, 252, 193, 44, 67, 252, 193, 46, 67, 252, 193, 48, 67, 252, 193, 50, 67, 252, 193, 52, 67, 252, 193, 54, 67, 252, 193, 72, 67, 252, 193, 74, 67, 252, 193, 76, 67, 252, 193, 78, 67, 252, 193, 80, 67, 252, 193, 82, 67, 252, 193, 68, 67, 252, 193, 70, 67, 252, 193, 188, 67, 252, 193, 246, 67, 252, 193, 6, 68, 252, 193, 22, 68, 252, 193, 30, 68, 252, 193, 48, 68, 252, 193, 108, 68, 252, 193, 132, 68, 252, 193, 134, 68, 252, 193, 168, 68, 252, 193, 170, 68, 252, 193, 172, 68, 252, 193, 174, 68, 252, 193, 176, 68, 252, 193, 20, 69, 252, 193, 22, 69, 252, 193, 24, 69, 252, 193, 26, 69, 252, 193, 28, 69, 252, 193, 120, 69, 252, 193, 122, 69, 252, 193, 124, 69, 252, 193, 126, 69, 252, 193, 176, 69, 252, 193, 178, 69, 252, 193, 180, 69, 252, 193, 182, 69, 252, 193, 184, 69, 252, 193, 186, 69, 252, 193, 188, 69, 252, 193, 86, 67, 252, 193, 88, 67, 252, 193, 90, 67, 252, 193, 92, 67, 252, 193, 94, 67, 252, 193, 96, 67, 252, 193, 98, 67, 252, 193, 116, 67, 252, 193, 118, 67, 252, 193, 120, 67, 252, 193, 122, 67, 252, 193, 124, 67, 252, 193, 126, 67, 252, 193, 128, 67, 252, 193, 130, 67, 252, 193, 132, 67, 252, 193, 134, 67, 252, 193, 136, 67, 252, 193, 138, 67, 252, 193, 140, 67, 252, 193, 142, 67, 252, 193, 144, 67, 252, 193, 146, 67, 252, 193, 148, 67, 252, 193, 164, 67, 252, 193, 166, 67, 252, 193, 168, 67, 252, 193, 170, 67, 252, 193, 172, 67, 252, 193, 174, 67, 252, 193, 176, 67, 252, 193, 178, 67, 252, 193, 180, 67, 252, 193, 182, 67, 252, 193, 190, 67, 252, 193, 192, 67, 252, 193, 194, 67, 252, 193, 196, 67, 252, 193, 198, 67, 252, 193, 200, 67, 252, 193, 208, 67, 252, 193, 210, 67, 252, 193, 212, 67, 252, 193, 214, 67, 252, 193, 216, 67, 252, 193, 218, 67, 252, 193, 220, 67, 252, 193, 222, 67, 252, 193, 224, 67, 252, 193, 226, 67, 252, 193, 228, 67, 252, 193, 230, 67, 252, 193, 232, 67, 252, 193, 234, 67, 252, 193, 236, 67, 252, 193, 238, 67, 252, 193, 240, 67, 252, 193, 242, 67, 252, 193, 248, 67, 252, 193, 250, 67, 252, 193, 252, 67, 252, 193, 254, 67, 252, 193, 2, 68, 252, 193, 8, 68, 252, 193, 10, 68, 252, 193, 12, 68, 252, 193, 14, 68, 252, 193, 20, 68, 252, 193, 24, 68, 252, 193, 32, 68, 252, 193, 36, 68, 252, 193, 42, 68, 252, 193, 50, 68, 252, 193, 52, 68, 252, 193, 62, 68, 252, 193, 64, 68, 252, 193, 70, 68, 252, 193, 72, 68, 252, 193, 74, 68, 252, 193, 76, 68, 252, 193, 82, 68, 252, 193, 84, 68, 252, 193, 88, 68, 252, 193, 90, 68, 252, 193, 98, 68, 252, 193, 110, 68, 252, 193, 112, 68, 252, 193, 114, 68, 252, 193, 136, 68, 252, 193, 138, 68, 252, 193, 140, 68, 252, 193, 142, 68, 252, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 193, 34, 68, 252, 193, 46, 68, 252, 193, 58, 68, 252, 193, 60, 68, 252, 193, 66, 68, 252, 193, 92, 68, 252, 193, 120, 68, 252, 193, 148, 68, 252, 193, 150, 68, 252, 193, 248, 68, 252, 193, 250, 68, 252, 193, 252, 68, 252, 193, 254, 68, 252, 193, 96, 69, 252, 193, 98, 69, 252, 193, 100, 69, 252, 193, 102, 69, 252, 193, 104, 69, 252, 193, 106, 69, 252, 193, 108, 69, 252, 193, 162, 69, 252, 193, 242, 69, 252, 193, 102, 70, 252, 193, 104, 70, 252, 193, 192, 70, 252, 193, 38, 71, 252, 193, 40, 71, 252, 193, 116, 71, 252, 193, 166, 71, 252, 193, 126, 72, 252, 193, 128, 72, 252, 193, 250, 72, 252, 193, 122, 68, 252, 193, 124, 68, 252, 193, 126, 68, 252, 193, 128, 68, 252, 193, 130, 68, 252, 193, 152, 68, 252, 193, 154, 68, 252, 193, 156, 68, 252, 193, 158, 68, 252, 193, 160, 68, 252, 193, 162, 68, 252, 193, 164, 68, 252, 193, 166, 68, 252, 193, 2, 69, 252, 193, 4, 69, 252, 193, 6, 69, 252, 193, 8, 69, 252, 193, 10, 69, 252, 193, 12, 69, 252, 193, 14, 69, 252, 193, 16, 69, 252, 193, 18, 69, 252, 193, 110, 69, 252, 193, 112, 69, 252, 193, 114, 69, 252, 193, 116, 69, 252, 193, 118, 69, 252, 193, 164, 69, 252, 193, 166, 69, 252, 193, 168, 69, 252, 193, 170, 69, 252, 193, 172, 69, 252, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 128, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 206, 160, 51, 0, 193, 128, 69, 252, 193, 130, 69, 252, 193, 132, 69, 252, 193, 134, 69, 252, 193, 136, 69, 252, 193, 138, 69, 252, 193, 140, 69, 252, 193, 142, 69, 252, 193, 144, 69, 252, 193, 146, 69, 252, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 193, 244, 69, 252, 193, 174, 69, 252, 193, 246, 69, 252, 193, 248, 69, 252, 193, 250, 69, 252, 193, 252, 69, 252, 193, 254, 69, 252, 193, 2, 70, 252, 193, 4, 70, 252, 193, 6, 70, 252, 193, 8, 70, 252, 193, 10, 70, 252, 193, 106, 70, 252, 193, 108, 70, 252, 193, 110, 70, 252, 193, 112, 70, 252, 193, 114, 70, 252, 193, 116, 70, 252, 193, 194, 70, 252, 193, 196, 70, 252, 193, 198, 70, 252, 193, 200, 70, 252, 193, 202, 70, 252, 193, 204, 70, 252, 193, 42, 71, 252, 193, 44, 71, 252, 193, 46, 71, 252, 193, 48, 71, 252, 193, 118, 71, 252, 193, 120, 71, 252, 193, 122, 71, 252, 193, 124, 71, 252, 193, 190, 69, 252, 193, 12, 70, 252, 193, 14, 70, 252, 193, 16, 70, 252, 193, 206, 70, 252, 193, 118, 70, 252, 193, 120, 70, 252, 193, 122, 70, 252, 193, 124, 70, 252, 193, 208, 70, 252, 193, 210, 70, 252, 193, 212, 70, 252, 193, 50, 71, 252, 193, 52, 71, 252, 193, 54, 71, 252, 193, 56, 71, 252, 193, 58, 71, 252, 193, 126, 71, 252, 193, 128, 71, 252, 193, 130, 71, 252, 193, 170, 71, 252, 193, 172, 71, 252, 193, 174, 71, 252, 193, 196, 71, 252, 193, 198, 71, 252, 193, 216, 71, 252, 193, 248, 71, 252, 193, 10, 72, 252, 193, 162, 72, 252, 193, 164, 72, 252, 193, 166, 72, 252, 193, 168, 72, 252, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 192, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 206, 224, 51, 0, 193, 232, 69, 252, 193, 80, 70, 252, 193, 82, 70, 252, 193, 184, 70, 252, 193, 186, 70, 252, 193, 188, 70, 252, 193, 30, 71, 252, 193, 32, 71, 252, 193, 106, 71, 252, 193, 56, 72, 252, 193, 120, 72, 252, 193, 122, 72, 252, 193, 134, 73, 252, 193, 136, 73, 252, 193, 50, 74, 252, 193, 52, 74, 252, 193, 54, 74, 252, 193, 56, 74, 252, 193, 216, 74, 252, 193, 218, 74, 252, 193, 220, 74, 252, 193, 134, 75, 252, 193, 136, 75, 252, 193, 138, 75, 252, 193, 80, 76, 252, 193, 82, 76, 252, 193, 84, 76, 252, 193, 86, 76, 252, 193, 88, 76, 252, 193, 252, 76, 252, 193, 254, 76, 252, 193, 2, 77, 252, 193, 234, 69, 252, 193, 236, 69, 252, 193, 238, 69, 252, 193, 240, 69, 252, 193, 84, 70, 252, 193, 86, 70, 252, 193, 88, 70, 252, 193, 90, 70, 252, 193, 92, 70, 252, 193, 94, 70, 252, 193, 96, 70, 252, 193, 98, 70, 252, 193, 100, 70, 252, 193, 190, 70, 252, 193, 34, 71, 252, 193, 36, 71, 252, 193, 108, 71, 252, 193, 110, 71, 252, 193, 112, 71, 252, 193, 114, 71, 252, 193, 162, 71, 252, 193, 164, 71, 252, 193, 232, 71, 252, 193, 28, 72, 252, 193, 74, 72, 252, 193, 124, 72, 252, 193, 240, 72, 252, 193, 242, 72, 252, 193, 244, 72, 252, 193, 246, 72, 252, 193, 248, 72, 252, 193, 138, 73, 252, 193, 162, 70, 252, 193, 164, 70, 252, 193, 166, 70, 252, 193, 168, 70, 252, 193, 170, 70, 252, 193, 172, 70, 252, 193, 174, 70, 252, 193, 176, 70, 252, 193, 178, 70, 252, 193, 214, 70, 252, 193, 216, 70, 252, 193, 218, 70, 252, 193, 220, 70, 252, 193, 222, 70, 252, 193, 224, 70, 252, 193, 226, 70, 252, 193, 228, 70, 252, 193, 230, 70, 252, 193, 232, 70, 252, 193, 234, 70, 252, 193, 236, 70, 252, 193, 238, 70, 252, 193, 240, 70, 252, 193, 242, 70, 252, 193, 244, 70, 252, 193, 246, 70, 252, 193, 248, 70, 252, 193, 250, 70, 252, 193, 252, 70, 252, 193, 254, 70, 252, 193, 2, 71, 252, 193, 180, 70, 252, 193, 102, 71, 252, 193, 132, 71, 252, 193, 134, 71, 252, 193, 136, 71, 252, 193, 138, 71, 252, 193, 140, 71, 252, 193, 142, 71, 252, 193, 144, 71, 252, 193, 146, 71, 252, 193, 148, 71, 252, 193, 150, 71, 252, 193, 152, 71, 252, 193, 154, 71, 252, 193, 156, 71, 252, 193, 158, 71, 252, 193, 176, 71, 252, 193, 178, 71, 252, 193, 180, 71, 252, 193, 182, 71, 252, 193, 184, 71, 252, 193, 186, 71, 252, 193, 188, 71, 252, 193, 190, 71, 252, 193, 192, 71, 252, 193, 194, 71, 252, 193, 200, 71, 252, 193, 202, 71, 252, 193, 218, 71, 252, 193, 220, 71, 252, 193, 222, 71, 252, 193, 224, 71, 252, 193, 226, 71, 252, 193, 168, 71, 252, 193, 204, 71, 252, 193, 206, 71, 252, 193, 208, 71, 252, 193, 210, 71, 252, 193, 212, 71, 252, 193, 214, 71, 252, 193, 234, 71, 252, 193, 236, 71, 252, 193, 8, 72, 252, 193, 30, 72, 252, 193, 42, 72, 252, 193, 46, 72, 252, 193, 48, 72, 252, 193, 50, 72, 252, 193, 58, 72, 252, 193, 60, 72, 252, 193, 76, 72, 252, 193, 78, 72, 252, 193, 80, 72, 252, 193, 82, 72, 252, 193, 84, 72, 252, 193, 86, 72, 252, 193, 88, 72, 252, 193, 90, 72, 252, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 193, 228, 71, 252, 193, 230, 71, 252, 193, 238, 71, 252, 193, 240, 71, 252, 193, 242, 71, 252, 193, 244, 71, 252, 193, 246, 71, 252, 193, 250, 71, 252, 193, 252, 71, 252, 193, 254, 71, 252, 193, 2, 72, 252, 193, 4, 72, 252, 193, 12, 72, 252, 193, 14, 72, 252, 193, 16, 72, 252, 193, 18, 72, 252, 193, 20, 72, 252, 193, 22, 72, 252, 193, 24, 72, 252, 193, 26, 72, 252, 193, 32, 72, 252, 193, 34, 72, 252, 193, 36, 72, 252, 193, 38, 72, 252, 193, 40, 72, 252, 193, 44, 72, 252, 193, 54, 72, 252, 193, 62, 72, 252, 193, 64, 72, 252, 193, 66, 72, 252, 193, 68, 72, 252, 193, 70, 72, 252, 193, 72, 72, 252, 193, 92, 72, 252, 193, 94, 72, 252, 193, 96, 72, 252, 193, 98, 72, 252, 193, 100, 72, 252, 193, 102, 72, 252, 193, 104, 72, 252, 193, 106, 72, 252, 193, 108, 72, 252, 193, 110, 72, 252, 193, 112, 72, 252, 193, 114, 72, 252, 193, 116, 72, 252, 193, 118, 72, 252, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 206, 0, 52, 0, 193, 2, 73, 252, 193, 4, 73, 252, 193, 6, 73, 252, 193, 8, 73, 252, 193, 10, 73, 252, 193, 12, 73, 252, 193, 14, 73, 252, 193, 16, 73, 252, 193, 18, 73, 252, 193, 20, 73, 252, 193, 22, 73, 252, 193, 24, 73, 252, 193, 26, 73, 252, 193, 28, 73, 252, 193, 30, 73, 252, 193, 32, 73, 252, 193, 34, 73, 252, 193, 36, 73, 252, 193, 38, 73, 252, 193, 40, 73, 252, 193, 42, 73, 252, 193, 44, 73, 252, 193, 162, 73, 252, 193, 170, 72, 252, 193, 46, 73, 252, 193, 48, 73, 252, 193, 50, 73, 252, 193, 52, 73, 252, 193, 54, 73, 252, 193, 56, 73, 252, 193, 58, 73, 252, 193, 60, 73, 252, 193, 62, 73, 252, 193, 64, 73, 252, 193, 202, 73, 252, 193, 204, 73, 252, 193, 206, 73, 252, 193, 208, 73, 252, 193, 210, 73, 252, 193, 212, 73, 252, 193, 114, 74, 252, 193, 116, 74, 252, 193, 118, 74, 252, 193, 120, 74, 252, 193, 122, 74, 252, 193, 124, 74, 252, 193, 126, 74, 252, 193, 128, 74, 252, 193, 58, 75, 252, 193, 60, 75, 252, 193, 62, 75, 252, 193, 64, 75, 252, 193, 66, 75, 252, 193, 68, 75, 252, 193, 70, 75, 252, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 32, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 206, 64, 52, 0, 193, 236, 72, 252, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 193, 252, 72, 252, 193, 254, 72, 252, 193, 148, 73, 252, 193, 150, 73, 252, 193, 152, 73, 252, 193, 154, 73, 252, 193, 156, 73, 252, 193, 158, 73, 252, 193, 160, 73, 252, 193, 70, 74, 252, 193, 72, 74, 252, 193, 74, 74, 252, 193, 76, 74, 252, 193, 78, 74, 252, 193, 238, 74, 252, 193, 240, 74, 252, 193, 242, 74, 252, 193, 244, 74, 252, 193, 246, 74, 252, 193, 150, 75, 252, 193, 152, 75, 252, 193, 154, 75, 252, 193, 156, 75, 252, 193, 158, 75, 252, 193, 160, 75, 252, 193, 162, 75, 252, 193, 164, 75, 252, 193, 166, 75, 252, 193, 168, 75, 252, 193, 170, 75, 252, 193, 110, 76, 252, 193, 112, 76, 252, 193, 140, 73, 252, 193, 142, 73, 252, 193, 144, 73, 252, 193, 146, 73, 252, 193, 58, 74, 252, 193, 60, 74, 252, 193, 62, 74, 252, 193, 64, 74, 252, 193, 66, 74, 252, 193, 68, 74, 252, 193, 222, 74, 252, 193, 224, 74, 252, 193, 226, 74, 252, 193, 228, 74, 252, 193, 230, 74, 252, 193, 232, 74, 252, 193, 234, 74, 252, 193, 236, 74, 252, 193, 140, 75, 252, 193, 142, 75, 252, 193, 144, 75, 252, 193, 146, 75, 252, 193, 148, 75, 252, 193, 90, 76, 252, 193, 92, 76, 252, 193, 94, 76, 252, 193, 96, 76, 252, 193, 98, 76, 252, 193, 100, 76, 252, 193, 102, 76, 252, 193, 104, 76, 252, 193, 106, 76, 252, 193, 164, 73, 252, 193, 166, 73, 252, 193, 168, 73, 252, 193, 170, 73, 252, 193, 172, 73, 252, 193, 174, 73, 252, 193, 176, 73, 252, 193, 178, 73, 252, 193, 180, 73, 252, 193, 182, 73, 252, 193, 184, 73, 252, 193, 186, 73, 252, 193, 188, 73, 252, 193, 190, 73, 252, 193, 192, 73, 252, 193, 194, 73, 252, 193, 196, 73, 252, 193, 198, 73, 252, 193, 200, 73, 252, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 206, 96, 52, 0, 193, 130, 74, 252, 193, 132, 74, 252, 193, 134, 74, 252, 193, 136, 74, 252, 193, 138, 74, 252, 193, 140, 74, 252, 193, 142, 74, 252, 193, 144, 74, 252, 193, 146, 74, 252, 193, 148, 74, 252, 193, 44, 74, 252, 193, 150, 74, 252, 193, 152, 74, 252, 193, 154, 74, 252, 193, 156, 74, 252, 193, 158, 74, 252, 193, 160, 74, 252, 193, 162, 74, 252, 193, 164, 74, 252, 193, 166, 74, 252, 193, 168, 74, 252, 193, 170, 74, 252, 193, 172, 74, 252, 193, 174, 74, 252, 193, 176, 74, 252, 193, 178, 74, 252, 193, 180, 74, 252, 193, 182, 74, 252, 193, 184, 74, 252, 193, 186, 74, 252, 193, 188, 74, 252, 193, 190, 74, 252, 193, 192, 74, 252, 193, 194, 74, 252, 193, 196, 74, 252, 193, 198, 74, 252, 193, 200, 74, 252, 193, 202, 74, 252, 193, 204, 74, 252, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 206, 128, 52, 0, 193, 248, 74, 252, 193, 250, 74, 252, 193, 252, 74, 252, 193, 254, 74, 252, 193, 2, 75, 252, 193, 4, 75, 252, 193, 6, 75, 252, 193, 8, 75, 252, 193, 10, 75, 252, 193, 12, 75, 252, 193, 14, 75, 252, 193, 16, 75, 252, 193, 18, 75, 252, 193, 20, 75, 252, 193, 22, 75, 252, 193, 24, 75, 252, 193, 26, 75, 252, 193, 28, 75, 252, 193, 30, 75, 252, 193, 32, 75, 252, 193, 34, 75, 252, 193, 36, 75, 252, 193, 38, 75, 252, 193, 40, 75, 252, 193, 42, 75, 252, 193, 44, 75, 252, 193, 46, 75, 252, 193, 48, 75, 252, 193, 206, 74, 252, 193, 208, 74, 252, 193, 210, 74, 252, 193, 74, 75, 252, 193, 76, 75, 252, 193, 78, 75, 252, 193, 80, 75, 252, 193, 82, 75, 252, 193, 84, 75, 252, 193, 86, 75, 252, 193, 88, 75, 252, 193, 90, 75, 252, 193, 92, 75, 252, 193, 94, 75, 252, 193, 96, 75, 252, 193, 98, 75, 252, 193, 100, 75, 252, 193, 102, 75, 252, 193, 104, 75, 252, 193, 106, 75, 252, 193, 108, 75, 252, 193, 110, 75, 252, 193, 112, 75, 252, 193, 114, 75, 252, 193, 116, 75, 252, 193, 118, 75, 252, 193, 120, 75, 252, 193, 122, 75, 252, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 206, 160, 52, 0, 193, 50, 75, 252, 193, 52, 75, 252, 193, 172, 75, 252, 193, 174, 75, 252, 193, 176, 75, 252, 193, 178, 75, 252, 193, 180, 75, 252, 193, 182, 75, 252, 193, 184, 75, 252, 193, 186, 75, 252, 193, 54, 75, 252, 193, 56, 75, 252, 193, 188, 75, 252, 193, 190, 75, 252, 193, 192, 75, 252, 193, 194, 75, 252, 193, 196, 75, 252, 193, 198, 75, 252, 193, 200, 75, 252, 193, 202, 75, 252, 193, 204, 75, 252, 193, 206, 75, 252, 193, 208, 75, 252, 193, 210, 75, 252, 193, 212, 75, 252, 193, 214, 75, 252, 193, 116, 76, 252, 193, 118, 76, 252, 193, 120, 76, 252, 193, 122, 76, 252, 193, 124, 76, 252, 193, 126, 76, 252, 193, 72, 75, 252, 193, 218, 75, 252, 193, 220, 75, 252, 193, 222, 75, 252, 193, 224, 75, 252, 193, 226, 75, 252, 193, 228, 75, 252, 193, 146, 76, 252, 193, 148, 76, 252, 193, 150, 76, 252, 193, 152, 76, 252, 193, 154, 76, 252, 193, 156, 76, 252, 193, 158, 76, 252, 193, 32, 77, 252, 193, 34, 77, 252, 193, 36, 77, 252, 193, 38, 77, 252, 193, 40, 77, 252, 193, 42, 77, 252, 193, 192, 77, 252, 193, 194, 77, 252, 193, 196, 77, 252, 193, 198, 77, 252, 193, 200, 77, 252, 193, 202, 77, 252, 193, 70, 78, 252, 193, 72, 78, 252, 193, 74, 78, 252, 193, 150, 78, 252, 193, 152, 78, 252, 193, 154, 78, 252, 193, 128, 75, 252, 193, 130, 75, 252, 193, 72, 76, 252, 193, 74, 76, 252, 193, 76, 76, 252, 193, 78, 76, 252, 193, 248, 76, 252, 193, 250, 76, 252, 193, 128, 77, 252, 193, 32, 78, 252, 193, 184, 78, 252, 193, 236, 78, 252, 193, 18, 79, 252, 193, 166, 79, 252, 193, 80, 80, 252, 193, 82, 80, 252, 193, 84, 80, 252, 193, 86, 80, 252, 193, 88, 80, 252, 193, 90, 80, 252, 193, 92, 80, 252, 193, 30, 81, 252, 193, 32, 81, 252, 193, 226, 81, 252, 193, 228, 81, 252, 193, 230, 81, 252, 193, 232, 81, 252, 193, 234, 81, 252, 193, 236, 81, 252, 193, 238, 81, 252, 193, 240, 81, 252, 193, 242, 81, 252, 193, 128, 76, 252, 193, 216, 75, 252, 193, 130, 76, 252, 193, 132, 76, 252, 193, 10, 77, 252, 193, 134, 76, 252, 193, 136, 76, 252, 193, 138, 76, 252, 193, 140, 76, 252, 193, 142, 76, 252, 193, 144, 76, 252, 193, 160, 77, 252, 193, 12, 77, 252, 193, 14, 77, 252, 193, 16, 77, 252, 193, 18, 77, 252, 193, 20, 77, 252, 193, 22, 77, 252, 193, 24, 77, 252, 193, 26, 77, 252, 193, 28, 77, 252, 193, 30, 77, 252, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 192, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 193, 108, 76, 252, 193, 6, 77, 252, 193, 140, 77, 252, 193, 142, 77, 252, 193, 144, 77, 252, 193, 146, 77, 252, 193, 148, 77, 252, 193, 150, 77, 252, 193, 38, 78, 252, 193, 40, 78, 252, 193, 42, 78, 252, 193, 44, 78, 252, 193, 46, 78, 252, 193, 134, 78, 252, 193, 188, 78, 252, 193, 190, 78, 252, 193, 240, 78, 252, 193, 242, 78, 252, 193, 48, 79, 252, 193, 66, 79, 252, 193, 110, 79, 252, 193, 168, 79, 252, 193, 106, 80, 252, 193, 108, 80, 252, 193, 110, 80, 252, 193, 112, 80, 252, 193, 114, 80, 252, 193, 116, 80, 252, 206, 0, 54, 0, 206, 0, 54, 0, 206, 0, 54, 0, 206, 0, 54, 0, 193, 114, 76, 252, 193, 8, 77, 252, 193, 152, 77, 252, 193, 154, 77, 252, 193, 156, 77, 252, 193, 158, 77, 252, 193, 48, 78, 252, 193, 50, 78, 252, 193, 244, 78, 252, 193, 20, 79, 252, 193, 22, 79, 252, 193, 24, 79, 252, 193, 68, 79, 252, 193, 170, 79, 252, 193, 172, 79, 252, 193, 118, 80, 252, 193, 120, 80, 252, 193, 122, 80, 252, 193, 124, 80, 252, 193, 126, 80, 252, 193, 64, 81, 252, 193, 66, 81, 252, 193, 68, 81, 252, 193, 22, 82, 252, 193, 24, 82, 252, 193, 26, 82, 252, 193, 28, 82, 252, 193, 30, 82, 252, 193, 32, 82, 252, 193, 226, 82, 252, 193, 228, 82, 252, 193, 230, 82, 252, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 206, 224, 52, 0, 193, 44, 77, 252, 193, 46, 77, 252, 193, 48, 77, 252, 193, 50, 77, 252, 193, 52, 77, 252, 193, 54, 77, 252, 193, 56, 77, 252, 193, 4, 77, 252, 193, 130, 77, 252, 193, 132, 77, 252, 193, 134, 77, 252, 193, 136, 77, 252, 193, 138, 77, 252, 193, 34, 78, 252, 193, 36, 78, 252, 193, 128, 78, 252, 193, 130, 78, 252, 193, 132, 78, 252, 193, 186, 78, 252, 193, 238, 78, 252, 193, 46, 79, 252, 193, 104, 79, 252, 193, 130, 79, 252, 193, 226, 79, 252, 193, 228, 79, 252, 193, 96, 80, 252, 193, 98, 80, 252, 193, 100, 80, 252, 193, 102, 80, 252, 193, 104, 80, 252, 193, 34, 81, 252, 193, 36, 81, 252, 193, 38, 81, 252, 193, 40, 81, 252, 193, 42, 81, 252, 193, 44, 81, 252, 193, 46, 81, 252, 193, 246, 81, 252, 193, 248, 81, 252, 193, 58, 77, 252, 193, 60, 77, 252, 193, 62, 77, 252, 193, 64, 77, 252, 193, 66, 77, 252, 193, 68, 77, 252, 193, 70, 77, 252, 193, 72, 77, 252, 193, 74, 77, 252, 193, 76, 77, 252, 193, 78, 77, 252, 193, 80, 77, 252, 193, 82, 77, 252, 193, 84, 77, 252, 193, 86, 77, 252, 193, 88, 77, 252, 193, 90, 77, 252, 193, 92, 77, 252, 193, 94, 77, 252, 193, 96, 77, 252, 193, 98, 77, 252, 193, 100, 77, 252, 193, 102, 77, 252, 193, 104, 77, 252, 193, 106, 77, 252, 193, 108, 77, 252, 193, 110, 77, 252, 193, 112, 77, 252, 193, 114, 77, 252, 193, 116, 77, 252, 193, 118, 77, 252, 193, 120, 77, 252, 193, 122, 77, 252, 193, 124, 77, 252, 193, 126, 77, 252, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 206, 0, 53, 0, 193, 52, 78, 252, 193, 54, 78, 252, 193, 56, 78, 252, 193, 58, 78, 252, 193, 60, 78, 252, 193, 62, 78, 252, 193, 64, 78, 252, 193, 66, 78, 252, 193, 68, 78, 252, 193, 136, 78, 252, 193, 138, 78, 252, 193, 140, 78, 252, 193, 142, 78, 252, 193, 144, 78, 252, 193, 146, 78, 252, 193, 148, 78, 252, 193, 192, 78, 252, 193, 194, 78, 252, 193, 196, 78, 252, 193, 198, 78, 252, 193, 246, 78, 252, 193, 248, 78, 252, 193, 250, 78, 252, 193, 26, 79, 252, 193, 28, 79, 252, 193, 50, 79, 252, 193, 52, 79, 252, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 32, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 206, 64, 53, 0, 193, 156, 78, 252, 193, 158, 78, 252, 193, 160, 78, 252, 193, 162, 78, 252, 193, 164, 78, 252, 193, 166, 78, 252, 193, 168, 78, 252, 193, 170, 78, 252, 193, 172, 78, 252, 193, 174, 78, 252, 193, 176, 78, 252, 193, 178, 78, 252, 193, 180, 78, 252, 193, 182, 78, 252, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 193, 200, 78, 252, 193, 252, 78, 252, 193, 254, 78, 252, 193, 30, 79, 252, 193, 32, 79, 252, 193, 34, 79, 252, 193, 36, 79, 252, 193, 94, 79, 252, 193, 152, 79, 252, 193, 154, 79, 252, 193, 156, 79, 252, 193, 158, 79, 252, 193, 194, 79, 252, 193, 196, 79, 252, 193, 198, 79, 252, 193, 200, 79, 252, 193, 22, 80, 252, 193, 24, 80, 252, 193, 26, 80, 252, 193, 28, 80, 252, 193, 30, 80, 252, 193, 32, 80, 252, 193, 34, 80, 252, 193, 192, 80, 252, 193, 194, 80, 252, 193, 196, 80, 252, 193, 198, 80, 252, 193, 200, 80, 252, 193, 202, 80, 252, 193, 204, 80, 252, 193, 206, 80, 252, 193, 138, 81, 252, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 206, 96, 53, 0, 193, 2, 79, 252, 193, 4, 79, 252, 193, 6, 79, 252, 193, 8, 79, 252, 193, 10, 79, 252, 193, 12, 79, 252, 193, 14, 79, 252, 193, 16, 79, 252, 193, 38, 79, 252, 193, 40, 79, 252, 193, 42, 79, 252, 193, 44, 79, 252, 193, 54, 79, 252, 193, 56, 79, 252, 193, 58, 79, 252, 193, 60, 79, 252, 193, 62, 79, 252, 193, 64, 79, 252, 193, 72, 79, 252, 193, 74, 79, 252, 193, 76, 79, 252, 193, 78, 79, 252, 193, 86, 79, 252, 193, 88, 79, 252, 193, 90, 79, 252, 193, 92, 79, 252, 193, 96, 79, 252, 193, 70, 79, 252, 193, 80, 79, 252, 193, 82, 79, 252, 193, 84, 79, 252, 193, 112, 79, 252, 193, 116, 79, 252, 193, 120, 79, 252, 193, 122, 79, 252, 193, 124, 79, 252, 193, 132, 79, 252, 193, 138, 79, 252, 193, 134, 79, 252, 193, 140, 79, 252, 193, 142, 79, 252, 193, 144, 79, 252, 193, 146, 79, 252, 193, 148, 79, 252, 193, 150, 79, 252, 193, 174, 79, 252, 193, 176, 79, 252, 193, 178, 79, 252, 193, 180, 79, 252, 193, 182, 79, 252, 193, 184, 79, 252, 193, 186, 79, 252, 193, 188, 79, 252, 193, 190, 79, 252, 193, 192, 79, 252, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 193, 98, 79, 252, 193, 100, 79, 252, 193, 102, 79, 252, 193, 106, 79, 252, 193, 108, 79, 252, 193, 114, 79, 252, 193, 118, 79, 252, 193, 126, 79, 252, 193, 128, 79, 252, 193, 136, 79, 252, 193, 160, 79, 252, 193, 162, 79, 252, 193, 164, 79, 252, 193, 202, 79, 252, 193, 204, 79, 252, 193, 206, 79, 252, 193, 208, 79, 252, 193, 210, 79, 252, 193, 212, 79, 252, 193, 214, 79, 252, 193, 216, 79, 252, 193, 218, 79, 252, 193, 220, 79, 252, 193, 222, 79, 252, 193, 224, 79, 252, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 128, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 160, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 192, 53, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 224, 53, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 0, 54, 0, 206, 0, 54, 0, 206, 0, 54, 0, 206, 0, 54, 0, 193, 6, 82, 252, 193, 8, 82, 252, 193, 10, 82, 252, 193, 12, 82, 252, 193, 14, 82, 252, 193, 16, 82, 252, 193, 18, 82, 252, 193, 20, 82, 252, 193, 208, 82, 252, 193, 210, 82, 252, 193, 212, 82, 252, 193, 214, 82, 252, 193, 216, 82, 252, 193, 218, 82, 252, 193, 220, 82, 252, 193, 222, 82, 252, 193, 224, 82, 252, 193, 238, 83, 252, 193, 240, 83, 252, 193, 242, 83, 252, 193, 244, 83, 252, 193, 246, 83, 252, 193, 248, 83, 252, 193, 250, 83, 252, 193, 252, 83, 252, 193, 254, 83, 252, 193, 2, 84, 252, 193, 4, 84, 252, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 206, 32, 54, 0, 193, 34, 82, 252, 193, 36, 82, 252, 193, 38, 82, 252, 193, 40, 82, 252, 193, 234, 82, 252, 206, 96, 54, 0, 206, 96, 54, 0, 206, 96, 54, 0, 206, 96, 54, 0, 206, 96, 54, 0, 206, 96, 54, 0, 206, 96, 54, 0, 206, 96, 54, 0, 206, 96, 54, 0, 193, 136, 81, 252, 193, 76, 82, 252, 193, 78, 82, 252, 193, 80, 82, 252, 193, 82, 82, 252, 193, 84, 82, 252, 193, 86, 82, 252, 193, 88, 82, 252, 193, 90, 82, 252, 193, 92, 82, 252, 193, 94, 82, 252, 193, 236, 82, 252, 193, 238, 82, 252, 193, 240, 82, 252, 193, 242, 82, 252, 193, 244, 82, 252, 193, 246, 82, 252, 193, 248, 82, 252, 193, 250, 82, 252, 193, 252, 82, 252, 193, 254, 82, 252, 193, 20, 84, 252, 193, 2, 83, 252, 193, 140, 81, 252, 193, 142, 81, 252, 193, 144, 81, 252, 193, 146, 81, 252, 193, 148, 81, 252, 193, 150, 81, 252, 193, 152, 81, 252, 193, 154, 81, 252, 193, 156, 81, 252, 193, 96, 82, 252, 193, 98, 82, 252, 193, 100, 82, 252, 193, 102, 82, 252, 193, 104, 82, 252, 193, 106, 82, 252, 193, 108, 82, 252, 193, 110, 82, 252, 193, 62, 83, 252, 193, 64, 83, 252, 193, 66, 83, 252, 193, 68, 83, 252, 193, 70, 83, 252, 193, 72, 83, 252, 193, 74, 83, 252, 193, 76, 83, 252, 193, 78, 83, 252, 193, 80, 83, 252, 193, 82, 83, 252, 193, 84, 83, 252, 193, 92, 84, 252, 193, 94, 84, 252, 193, 96, 84, 252, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 64, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 193, 250, 81, 252, 193, 252, 81, 252, 193, 254, 81, 252, 193, 2, 82, 252, 193, 4, 82, 252, 193, 198, 82, 252, 193, 200, 82, 252, 193, 202, 82, 252, 193, 204, 82, 252, 193, 206, 82, 252, 193, 218, 83, 252, 193, 220, 83, 252, 193, 222, 83, 252, 193, 224, 83, 252, 193, 226, 83, 252, 193, 228, 83, 252, 193, 230, 83, 252, 193, 232, 83, 252, 193, 234, 83, 252, 193, 236, 83, 252, 193, 8, 85, 252, 193, 10, 85, 252, 193, 12, 85, 252, 193, 14, 85, 252, 193, 16, 85, 252, 193, 18, 85, 252, 193, 20, 85, 252, 206, 0, 55, 0, 206, 0, 55, 0, 206, 0, 55, 0, 206, 0, 55, 0, 206, 0, 55, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 128, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 193, 186, 82, 252, 193, 188, 82, 252, 193, 190, 82, 252, 193, 192, 82, 252, 193, 194, 82, 252, 193, 196, 82, 252, 193, 190, 83, 252, 193, 192, 83, 252, 193, 194, 83, 252, 193, 196, 83, 252, 193, 198, 83, 252, 193, 200, 83, 252, 193, 202, 83, 252, 193, 204, 83, 252, 193, 206, 83, 252, 193, 208, 83, 252, 193, 210, 83, 252, 193, 212, 83, 252, 193, 214, 83, 252, 193, 230, 84, 252, 193, 232, 84, 252, 193, 234, 84, 252, 193, 236, 84, 252, 193, 238, 84, 252, 193, 240, 84, 252, 193, 242, 84, 252, 193, 244, 84, 252, 193, 246, 84, 252, 193, 248, 84, 252, 193, 250, 84, 252, 193, 252, 84, 252, 193, 254, 84, 252, 193, 232, 82, 252, 193, 12, 84, 252, 193, 14, 84, 252, 193, 16, 84, 252, 193, 18, 84, 252, 193, 38, 85, 252, 193, 40, 85, 252, 193, 42, 85, 252, 193, 44, 85, 252, 193, 244, 85, 252, 193, 246, 85, 252, 193, 190, 86, 252, 193, 192, 86, 252, 193, 194, 86, 252, 193, 140, 87, 252, 193, 142, 87, 252, 193, 144, 87, 252, 193, 146, 87, 252, 193, 148, 87, 252, 193, 44, 88, 252, 193, 46, 88, 252, 193, 238, 88, 252, 193, 240, 88, 252, 193, 42, 89, 252, 193, 44, 89, 252, 193, 82, 89, 252, 193, 106, 89, 252, 193, 160, 89, 252, 193, 210, 89, 252, 193, 244, 89, 252, 193, 8, 90, 252, 193, 10, 90, 252, 193, 4, 83, 252, 193, 6, 83, 252, 193, 8, 83, 252, 193, 10, 83, 252, 193, 12, 83, 252, 193, 14, 83, 252, 193, 16, 83, 252, 193, 18, 83, 252, 193, 20, 83, 252, 193, 22, 83, 252, 193, 24, 83, 252, 193, 26, 83, 252, 193, 28, 83, 252, 193, 30, 83, 252, 193, 32, 83, 252, 193, 34, 83, 252, 193, 36, 83, 252, 193, 38, 83, 252, 193, 40, 83, 252, 193, 42, 83, 252, 193, 44, 83, 252, 193, 46, 83, 252, 193, 48, 83, 252, 193, 50, 83, 252, 193, 52, 83, 252, 193, 54, 83, 252, 193, 56, 83, 252, 193, 58, 83, 252, 193, 60, 83, 252, 193, 22, 84, 252, 193, 24, 84, 252, 193, 26, 84, 252, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 160, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 193, 6, 84, 252, 193, 8, 84, 252, 193, 10, 84, 252, 193, 22, 85, 252, 193, 24, 85, 252, 193, 26, 85, 252, 193, 28, 85, 252, 193, 30, 85, 252, 193, 32, 85, 252, 193, 34, 85, 252, 193, 36, 85, 252, 193, 236, 85, 252, 193, 238, 85, 252, 193, 240, 85, 252, 193, 242, 85, 252, 193, 176, 86, 252, 193, 178, 86, 252, 193, 180, 86, 252, 193, 182, 86, 252, 193, 184, 86, 252, 193, 186, 86, 252, 193, 188, 86, 252, 193, 120, 87, 252, 193, 122, 87, 252, 193, 124, 87, 252, 193, 126, 87, 252, 193, 128, 87, 252, 193, 130, 87, 252, 193, 132, 87, 252, 193, 134, 87, 252, 193, 136, 87, 252, 193, 138, 87, 252, 193, 28, 84, 252, 193, 30, 84, 252, 193, 32, 84, 252, 193, 34, 84, 252, 193, 36, 84, 252, 193, 38, 84, 252, 193, 40, 84, 252, 193, 42, 84, 252, 193, 44, 84, 252, 193, 46, 84, 252, 193, 48, 84, 252, 193, 50, 84, 252, 193, 52, 84, 252, 193, 54, 84, 252, 193, 56, 84, 252, 193, 58, 84, 252, 193, 60, 84, 252, 193, 62, 84, 252, 193, 64, 84, 252, 193, 66, 84, 252, 193, 68, 84, 252, 193, 70, 84, 252, 193, 72, 84, 252, 193, 74, 84, 252, 193, 76, 84, 252, 193, 78, 84, 252, 193, 80, 84, 252, 193, 82, 84, 252, 193, 84, 84, 252, 193, 86, 84, 252, 193, 88, 84, 252, 193, 90, 84, 252, 193, 98, 84, 252, 193, 100, 84, 252, 193, 102, 84, 252, 193, 104, 84, 252, 193, 106, 84, 252, 193, 108, 84, 252, 193, 110, 84, 252, 193, 96, 85, 252, 193, 98, 85, 252, 193, 100, 85, 252, 193, 102, 85, 252, 193, 38, 86, 252, 193, 40, 86, 252, 193, 42, 86, 252, 193, 44, 86, 252, 193, 46, 86, 252, 193, 48, 86, 252, 193, 50, 86, 252, 193, 52, 86, 252, 193, 54, 86, 252, 193, 56, 86, 252, 193, 238, 86, 252, 193, 240, 86, 252, 193, 242, 86, 252, 193, 244, 86, 252, 193, 246, 86, 252, 193, 248, 86, 252, 193, 250, 86, 252, 193, 252, 86, 252, 193, 184, 87, 252, 193, 186, 87, 252, 193, 188, 87, 252, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 192, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 193, 2, 85, 252, 193, 4, 85, 252, 193, 6, 85, 252, 193, 188, 85, 252, 193, 190, 85, 252, 193, 192, 85, 252, 193, 194, 85, 252, 193, 196, 85, 252, 193, 198, 85, 252, 193, 200, 85, 252, 193, 202, 85, 252, 193, 204, 85, 252, 193, 206, 85, 252, 193, 208, 85, 252, 193, 210, 85, 252, 193, 212, 85, 252, 193, 148, 86, 252, 193, 150, 86, 252, 193, 152, 86, 252, 193, 154, 86, 252, 193, 156, 86, 252, 193, 158, 86, 252, 193, 160, 86, 252, 193, 84, 87, 252, 193, 86, 87, 252, 193, 88, 87, 252, 193, 90, 87, 252, 193, 92, 87, 252, 193, 94, 87, 252, 193, 96, 87, 252, 193, 98, 87, 252, 193, 100, 87, 252, 193, 46, 85, 252, 193, 48, 85, 252, 193, 50, 85, 252, 193, 52, 85, 252, 193, 54, 85, 252, 193, 56, 85, 252, 193, 58, 85, 252, 193, 60, 85, 252, 193, 62, 85, 252, 193, 64, 85, 252, 193, 66, 85, 252, 193, 68, 85, 252, 193, 70, 85, 252, 193, 72, 85, 252, 193, 74, 85, 252, 193, 76, 85, 252, 193, 78, 85, 252, 193, 80, 85, 252, 193, 82, 85, 252, 193, 84, 85, 252, 193, 86, 85, 252, 193, 88, 85, 252, 193, 90, 85, 252, 193, 92, 85, 252, 193, 94, 85, 252, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 224, 54, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 0, 55, 0, 206, 0, 55, 0, 206, 0, 55, 0, 206, 0, 55, 0, 206, 0, 55, 0, 193, 164, 86, 252, 193, 166, 86, 252, 193, 168, 86, 252, 193, 170, 86, 252, 193, 172, 86, 252, 193, 174, 86, 252, 193, 104, 87, 252, 193, 106, 87, 252, 193, 108, 87, 252, 193, 110, 87, 252, 193, 112, 87, 252, 193, 114, 87, 252, 193, 116, 87, 252, 193, 118, 87, 252, 193, 18, 88, 252, 193, 20, 88, 252, 193, 22, 88, 252, 193, 24, 88, 252, 193, 142, 88, 252, 193, 144, 88, 252, 193, 146, 88, 252, 193, 228, 88, 252, 193, 34, 89, 252, 193, 36, 89, 252, 193, 38, 89, 252, 193, 134, 89, 252, 193, 146, 89, 252, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 206, 32, 55, 0, 193, 196, 86, 252, 193, 34, 86, 252, 193, 198, 86, 252, 193, 200, 86, 252, 193, 202, 86, 252, 193, 204, 86, 252, 193, 36, 86, 252, 193, 206, 86, 252, 193, 208, 86, 252, 193, 210, 86, 252, 193, 212, 86, 252, 193, 214, 86, 252, 193, 216, 86, 252, 193, 218, 86, 252, 193, 220, 86, 252, 193, 222, 86, 252, 193, 224, 86, 252, 193, 226, 86, 252, 193, 228, 86, 252, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 206, 64, 55, 0, 193, 254, 86, 252, 193, 146, 86, 252, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 193, 230, 86, 252, 193, 232, 86, 252, 193, 150, 87, 252, 193, 234, 86, 252, 193, 236, 86, 252, 193, 152, 87, 252, 193, 154, 87, 252, 193, 156, 87, 252, 193, 158, 87, 252, 193, 160, 87, 252, 193, 162, 87, 252, 193, 164, 87, 252, 193, 166, 87, 252, 193, 168, 87, 252, 193, 170, 87, 252, 193, 172, 87, 252, 193, 174, 87, 252, 193, 176, 87, 252, 193, 178, 87, 252, 193, 180, 87, 252, 193, 182, 87, 252, 193, 48, 88, 252, 193, 50, 88, 252, 193, 52, 88, 252, 193, 54, 88, 252, 193, 56, 88, 252, 193, 58, 88, 252, 193, 60, 88, 252, 193, 62, 88, 252, 193, 64, 88, 252, 193, 66, 88, 252, 193, 68, 88, 252, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 96, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 193, 102, 87, 252, 193, 6, 88, 252, 193, 8, 88, 252, 193, 10, 88, 252, 193, 12, 88, 252, 193, 14, 88, 252, 193, 16, 88, 252, 193, 138, 88, 252, 193, 140, 88, 252, 193, 222, 88, 252, 193, 224, 88, 252, 193, 226, 88, 252, 193, 26, 89, 252, 193, 28, 89, 252, 193, 30, 89, 252, 193, 32, 89, 252, 193, 76, 89, 252, 193, 126, 89, 252, 193, 132, 89, 252, 193, 166, 89, 252, 193, 182, 89, 252, 193, 6, 90, 252, 193, 136, 90, 252, 193, 176, 90, 252, 193, 178, 90, 252, 193, 226, 90, 252, 193, 136, 91, 252, 193, 138, 91, 252, 193, 206, 91, 252, 193, 252, 91, 252, 193, 254, 91, 252, 193, 160, 92, 252, 193, 190, 87, 252, 193, 192, 87, 252, 193, 194, 87, 252, 193, 174, 88, 252, 193, 176, 88, 252, 193, 178, 88, 252, 193, 180, 88, 252, 193, 2, 89, 252, 193, 110, 89, 252, 193, 168, 89, 252, 193, 170, 89, 252, 193, 190, 89, 252, 193, 192, 89, 252, 193, 214, 89, 252, 193, 216, 89, 252, 193, 228, 89, 252, 193, 12, 90, 252, 193, 14, 90, 252, 193, 16, 90, 252, 193, 28, 90, 252, 193, 30, 90, 252, 193, 60, 90, 252, 193, 96, 90, 252, 193, 112, 90, 252, 193, 124, 90, 252, 193, 126, 90, 252, 193, 158, 90, 252, 193, 160, 90, 252, 193, 194, 90, 252, 193, 196, 90, 252, 193, 198, 90, 252, 193, 200, 90, 252, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 128, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 193, 26, 88, 252, 193, 28, 88, 252, 193, 30, 88, 252, 193, 32, 88, 252, 193, 34, 88, 252, 193, 36, 88, 252, 193, 38, 88, 252, 193, 40, 88, 252, 193, 42, 88, 252, 193, 148, 88, 252, 193, 150, 88, 252, 193, 152, 88, 252, 193, 154, 88, 252, 193, 230, 88, 252, 193, 232, 88, 252, 193, 234, 88, 252, 193, 236, 88, 252, 193, 40, 89, 252, 193, 78, 89, 252, 193, 80, 89, 252, 193, 100, 89, 252, 193, 102, 89, 252, 193, 104, 89, 252, 193, 184, 89, 252, 193, 208, 89, 252, 193, 56, 90, 252, 193, 68, 90, 252, 193, 106, 90, 252, 193, 108, 90, 252, 193, 184, 90, 252, 193, 28, 91, 252, 193, 30, 91, 252, 193, 70, 88, 252, 193, 72, 88, 252, 193, 74, 88, 252, 193, 156, 88, 252, 193, 158, 88, 252, 193, 160, 88, 252, 193, 162, 88, 252, 193, 164, 88, 252, 193, 166, 88, 252, 193, 168, 88, 252, 193, 170, 88, 252, 193, 172, 88, 252, 193, 242, 88, 252, 193, 244, 88, 252, 193, 246, 88, 252, 193, 248, 88, 252, 193, 250, 88, 252, 193, 252, 88, 252, 193, 254, 88, 252, 193, 46, 89, 252, 193, 48, 89, 252, 193, 50, 89, 252, 193, 52, 89, 252, 193, 84, 89, 252, 193, 86, 89, 252, 193, 88, 89, 252, 193, 108, 89, 252, 193, 128, 89, 252, 193, 158, 89, 252, 193, 162, 89, 252, 193, 186, 89, 252, 193, 188, 89, 252, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 206, 160, 55, 0, 193, 182, 88, 252, 193, 184, 88, 252, 193, 186, 88, 252, 193, 188, 88, 252, 193, 190, 88, 252, 193, 192, 88, 252, 193, 194, 88, 252, 193, 196, 88, 252, 193, 198, 88, 252, 193, 200, 88, 252, 193, 202, 88, 252, 193, 136, 88, 252, 193, 204, 88, 252, 193, 206, 88, 252, 193, 208, 88, 252, 193, 210, 88, 252, 193, 212, 88, 252, 193, 214, 88, 252, 193, 216, 88, 252, 193, 218, 88, 252, 193, 220, 88, 252, 193, 4, 89, 252, 193, 6, 89, 252, 193, 8, 89, 252, 193, 10, 89, 252, 193, 12, 89, 252, 193, 14, 89, 252, 193, 16, 89, 252, 193, 18, 89, 252, 193, 20, 89, 252, 193, 22, 89, 252, 193, 24, 89, 252, 193, 54, 89, 252, 193, 56, 89, 252, 193, 58, 89, 252, 193, 60, 89, 252, 193, 62, 89, 252, 193, 64, 89, 252, 193, 66, 89, 252, 193, 68, 89, 252, 193, 70, 89, 252, 193, 72, 89, 252, 193, 74, 89, 252, 193, 90, 89, 252, 193, 92, 89, 252, 193, 94, 89, 252, 193, 96, 89, 252, 193, 98, 89, 252, 193, 112, 89, 252, 193, 114, 89, 252, 193, 116, 89, 252, 193, 118, 89, 252, 193, 120, 89, 252, 193, 122, 89, 252, 193, 124, 89, 252, 193, 130, 89, 252, 193, 226, 189, 252, 193, 150, 89, 252, 193, 136, 89, 252, 193, 138, 89, 252, 193, 140, 89, 252, 193, 142, 89, 252, 193, 144, 89, 252, 193, 152, 89, 252, 193, 156, 89, 252, 193, 164, 89, 252, 193, 172, 89, 252, 193, 174, 89, 252, 193, 176, 89, 252, 193, 178, 89, 252, 193, 180, 89, 252, 193, 194, 89, 252, 193, 196, 89, 252, 193, 198, 89, 252, 193, 200, 89, 252, 193, 202, 89, 252, 193, 218, 89, 252, 193, 220, 89, 252, 193, 222, 89, 252, 193, 224, 89, 252, 193, 226, 89, 252, 193, 230, 89, 252, 193, 232, 89, 252, 193, 234, 89, 252, 193, 236, 89, 252, 193, 238, 89, 252, 193, 240, 89, 252, 193, 148, 89, 252, 193, 154, 89, 252, 193, 204, 89, 252, 193, 206, 89, 252, 193, 118, 90, 252, 193, 138, 90, 252, 193, 180, 90, 252, 193, 182, 90, 252, 193, 20, 91, 252, 193, 22, 91, 252, 193, 24, 91, 252, 193, 26, 91, 252, 193, 140, 91, 252, 193, 142, 91, 252, 193, 208, 91, 252, 193, 2, 92, 252, 193, 122, 92, 252, 193, 252, 92, 252, 193, 6, 93, 252, 193, 72, 93, 252, 193, 102, 93, 252, 193, 104, 93, 252, 193, 128, 93, 252, 193, 152, 93, 252, 193, 154, 93, 252, 193, 156, 93, 252, 193, 166, 93, 252, 193, 168, 93, 252, 193, 170, 93, 252, 193, 234, 93, 252, 193, 236, 93, 252, 193, 92, 94, 252, 193, 212, 89, 252, 193, 246, 89, 252, 193, 48, 90, 252, 193, 72, 90, 252, 193, 80, 90, 252, 193, 92, 90, 252, 193, 104, 90, 252, 193, 110, 90, 252, 193, 120, 90, 252, 193, 122, 90, 252, 193, 140, 90, 252, 193, 142, 90, 252, 193, 144, 90, 252, 193, 146, 90, 252, 193, 148, 90, 252, 193, 150, 90, 252, 193, 152, 90, 252, 193, 154, 90, 252, 193, 156, 90, 252, 193, 188, 90, 252, 193, 190, 90, 252, 193, 192, 90, 252, 193, 228, 90, 252, 193, 230, 90, 252, 193, 232, 90, 252, 193, 36, 91, 252, 193, 38, 91, 252, 193, 40, 91, 252, 193, 42, 91, 252, 193, 84, 91, 252, 193, 86, 91, 252, 193, 88, 91, 252, 193, 242, 89, 252, 193, 248, 89, 252, 193, 250, 89, 252, 193, 252, 89, 252, 193, 254, 89, 252, 193, 2, 90, 252, 193, 4, 90, 252, 193, 32, 90, 252, 193, 18, 90, 252, 193, 20, 90, 252, 193, 22, 90, 252, 193, 24, 90, 252, 193, 26, 90, 252, 193, 34, 90, 252, 193, 50, 90, 252, 193, 52, 90, 252, 193, 36, 90, 252, 193, 38, 90, 252, 193, 40, 90, 252, 193, 42, 90, 252, 193, 44, 90, 252, 193, 46, 90, 252, 193, 54, 90, 252, 193, 58, 90, 252, 193, 62, 90, 252, 193, 64, 90, 252, 193, 66, 90, 252, 193, 74, 90, 252, 193, 76, 90, 252, 193, 78, 90, 252, 193, 82, 90, 252, 193, 84, 90, 252, 193, 70, 90, 252, 193, 86, 90, 252, 193, 186, 90, 252, 193, 146, 91, 252, 193, 148, 91, 252, 193, 210, 91, 252, 193, 212, 91, 252, 193, 222, 92, 252, 193, 224, 92, 252, 193, 236, 92, 252, 193, 242, 92, 252, 193, 74, 93, 252, 193, 82, 93, 252, 193, 106, 93, 252, 193, 132, 93, 252, 193, 172, 93, 252, 193, 246, 93, 252, 193, 252, 93, 252, 193, 254, 93, 252, 193, 6, 94, 252, 193, 54, 94, 252, 193, 94, 94, 252, 193, 96, 94, 252, 193, 98, 94, 252, 193, 152, 94, 252, 193, 32, 95, 252, 193, 34, 95, 252, 193, 36, 95, 252, 193, 38, 95, 252, 193, 40, 95, 252, 193, 42, 95, 252, 193, 142, 95, 252, 193, 88, 90, 252, 193, 90, 90, 252, 193, 94, 90, 252, 193, 98, 90, 252, 193, 100, 90, 252, 193, 102, 90, 252, 193, 114, 90, 252, 193, 116, 90, 252, 193, 128, 90, 252, 193, 130, 90, 252, 193, 132, 90, 252, 193, 134, 90, 252, 193, 162, 90, 252, 193, 164, 90, 252, 193, 166, 90, 252, 193, 168, 90, 252, 193, 170, 90, 252, 193, 172, 90, 252, 193, 174, 90, 252, 193, 204, 90, 252, 193, 206, 90, 252, 193, 208, 90, 252, 193, 210, 90, 252, 193, 212, 90, 252, 193, 214, 90, 252, 193, 216, 90, 252, 193, 218, 90, 252, 193, 220, 90, 252, 193, 222, 90, 252, 193, 224, 90, 252, 193, 246, 90, 252, 193, 248, 90, 252, 193, 202, 90, 252, 193, 234, 90, 252, 193, 236, 90, 252, 193, 238, 90, 252, 193, 240, 90, 252, 193, 242, 90, 252, 193, 244, 90, 252, 193, 44, 91, 252, 193, 46, 91, 252, 193, 48, 91, 252, 193, 50, 91, 252, 193, 94, 91, 252, 193, 96, 91, 252, 193, 98, 91, 252, 193, 100, 91, 252, 193, 102, 91, 252, 193, 104, 91, 252, 193, 106, 91, 252, 193, 160, 91, 252, 193, 162, 91, 252, 193, 164, 91, 252, 193, 166, 91, 252, 193, 168, 91, 252, 193, 170, 91, 252, 193, 220, 91, 252, 193, 222, 91, 252, 193, 224, 91, 252, 193, 16, 92, 252, 193, 18, 92, 252, 193, 20, 92, 252, 193, 22, 92, 252, 193, 50, 92, 252, 193, 250, 90, 252, 193, 252, 90, 252, 193, 254, 90, 252, 193, 2, 91, 252, 193, 4, 91, 252, 193, 6, 91, 252, 193, 8, 91, 252, 193, 10, 91, 252, 193, 12, 91, 252, 193, 14, 91, 252, 193, 16, 91, 252, 193, 18, 91, 252, 193, 52, 91, 252, 193, 54, 91, 252, 193, 56, 91, 252, 193, 58, 91, 252, 193, 60, 91, 252, 193, 62, 91, 252, 193, 64, 91, 252, 193, 66, 91, 252, 193, 68, 91, 252, 193, 70, 91, 252, 193, 72, 91, 252, 193, 74, 91, 252, 193, 76, 91, 252, 193, 78, 91, 252, 193, 80, 91, 252, 206, 192, 55, 0, 206, 192, 55, 0, 206, 192, 55, 0, 206, 192, 55, 0, 206, 192, 55, 0, 193, 32, 91, 252, 193, 34, 91, 252, 193, 82, 91, 252, 193, 144, 91, 252, 193, 4, 92, 252, 193, 108, 92, 252, 193, 110, 92, 252, 193, 132, 92, 252, 193, 134, 92, 252, 193, 162, 92, 252, 193, 192, 92, 252, 193, 194, 92, 252, 193, 234, 92, 252, 193, 254, 92, 252, 193, 8, 93, 252, 193, 10, 93, 252, 193, 18, 93, 252, 193, 50, 93, 252, 193, 52, 93, 252, 193, 54, 93, 252, 193, 80, 93, 252, 193, 130, 93, 252, 193, 190, 93, 252, 193, 192, 93, 252, 193, 238, 93, 252, 193, 244, 93, 252, 193, 4, 94, 252, 193, 12, 94, 252, 193, 20, 94, 252, 193, 52, 94, 252, 193, 14, 95, 252, 193, 16, 95, 252, 193, 90, 91, 252, 193, 92, 91, 252, 193, 150, 91, 252, 193, 152, 91, 252, 193, 154, 91, 252, 193, 156, 91, 252, 193, 158, 91, 252, 193, 214, 91, 252, 193, 216, 91, 252, 193, 218, 91, 252, 193, 6, 92, 252, 193, 8, 92, 252, 193, 10, 92, 252, 193, 12, 92, 252, 193, 14, 92, 252, 193, 42, 92, 252, 193, 44, 92, 252, 193, 46, 92, 252, 193, 48, 92, 252, 193, 82, 92, 252, 193, 84, 92, 252, 193, 86, 92, 252, 193, 112, 92, 252, 193, 136, 92, 252, 193, 188, 92, 252, 193, 196, 92, 252, 193, 210, 92, 252, 193, 230, 92, 252, 193, 232, 92, 252, 193, 244, 92, 252, 193, 246, 92, 252, 193, 36, 93, 252, 206, 192, 55, 0, 206, 192, 55, 0, 206, 192, 55, 0, 206, 192, 55, 0, 206, 192, 55, 0, 206, 192, 55, 0, 206, 192, 55, 0, 193, 172, 91, 252, 193, 174, 91, 252, 193, 176, 91, 252, 193, 178, 91, 252, 193, 180, 91, 252, 193, 182, 91, 252, 193, 184, 91, 252, 193, 186, 91, 252, 193, 188, 91, 252, 193, 190, 91, 252, 193, 192, 91, 252, 193, 132, 91, 252, 193, 194, 91, 252, 193, 196, 91, 252, 193, 198, 91, 252, 193, 200, 91, 252, 193, 202, 91, 252, 193, 204, 91, 252, 206, 224, 55, 0, 206, 224, 55, 0, 206, 224, 55, 0, 206, 224, 55, 0, 206, 224, 55, 0, 206, 224, 55, 0, 206, 224, 55, 0, 206, 224, 55, 0, 193, 24, 92, 252, 193, 26, 92, 252, 193, 28, 92, 252, 193, 30, 92, 252, 193, 32, 92, 252, 193, 34, 92, 252, 193, 36, 92, 252, 193, 38, 92, 252, 193, 40, 92, 252, 193, 64, 92, 252, 193, 66, 92, 252, 193, 68, 92, 252, 193, 70, 92, 252, 193, 72, 92, 252, 193, 74, 92, 252, 193, 76, 92, 252, 193, 78, 92, 252, 193, 134, 91, 252, 193, 80, 92, 252, 193, 92, 92, 252, 193, 94, 92, 252, 193, 96, 92, 252, 193, 98, 92, 252, 193, 100, 92, 252, 193, 102, 92, 252, 193, 104, 92, 252, 193, 106, 92, 252, 193, 52, 92, 252, 193, 54, 92, 252, 193, 56, 92, 252, 193, 58, 92, 252, 193, 60, 92, 252, 193, 62, 92, 252, 193, 88, 92, 252, 193, 90, 92, 252, 193, 114, 92, 252, 193, 116, 92, 252, 193, 152, 92, 252, 193, 176, 92, 252, 193, 180, 92, 252, 193, 226, 92, 252, 193, 248, 92, 252, 193, 2, 93, 252, 193, 12, 93, 252, 193, 24, 93, 252, 193, 114, 93, 252, 193, 174, 93, 252, 193, 178, 93, 252, 193, 66, 94, 252, 193, 68, 94, 252, 193, 70, 94, 252, 193, 182, 94, 252, 193, 184, 94, 252, 193, 186, 94, 252, 193, 68, 95, 252, 193, 70, 95, 252, 193, 174, 95, 252, 193, 176, 95, 252, 193, 24, 96, 252, 193, 118, 92, 252, 193, 120, 92, 252, 193, 124, 92, 252, 193, 126, 92, 252, 193, 128, 92, 252, 193, 130, 92, 252, 193, 138, 92, 252, 193, 140, 92, 252, 193, 142, 92, 252, 193, 144, 92, 252, 193, 146, 92, 252, 193, 148, 92, 252, 193, 150, 92, 252, 193, 154, 92, 252, 193, 156, 92, 252, 193, 158, 92, 252, 193, 164, 92, 252, 193, 166, 92, 252, 193, 168, 92, 252, 193, 170, 92, 252, 193, 172, 92, 252, 193, 174, 92, 252, 193, 178, 92, 252, 193, 182, 92, 252, 193, 184, 92, 252, 193, 186, 92, 252, 193, 190, 92, 252, 193, 198, 92, 252, 193, 200, 92, 252, 193, 202, 92, 252, 193, 204, 92, 252, 193, 206, 92, 252, 193, 240, 92, 252, 193, 42, 93, 252, 193, 100, 93, 252, 193, 146, 93, 252, 193, 148, 93, 252, 193, 150, 93, 252, 193, 216, 93, 252, 193, 90, 94, 252, 193, 148, 94, 252, 193, 4, 95, 252, 193, 6, 95, 252, 193, 8, 95, 252, 193, 10, 95, 252, 193, 120, 95, 252, 193, 122, 95, 252, 193, 124, 95, 252, 193, 126, 95, 252, 193, 222, 95, 252, 193, 224, 95, 252, 193, 72, 96, 252, 193, 74, 96, 252, 193, 76, 96, 252, 193, 78, 96, 252, 193, 80, 96, 252, 193, 178, 96, 252, 193, 180, 96, 252, 193, 182, 96, 252, 193, 232, 96, 252, 193, 234, 96, 252, 193, 40, 97, 252, 193, 42, 97, 252, 193, 94, 97, 252, 193, 56, 93, 252, 193, 76, 93, 252, 193, 84, 93, 252, 193, 86, 93, 252, 193, 108, 93, 252, 193, 110, 93, 252, 193, 112, 93, 252, 193, 158, 93, 252, 193, 208, 93, 252, 193, 210, 93, 252, 193, 212, 93, 252, 193, 218, 93, 252, 193, 220, 93, 252, 193, 240, 93, 252, 193, 242, 93, 252, 193, 8, 94, 252, 193, 32, 94, 252, 193, 34, 94, 252, 193, 36, 94, 252, 193, 38, 94, 252, 193, 40, 94, 252, 193, 42, 94, 252, 193, 46, 94, 252, 193, 56, 94, 252, 193, 58, 94, 252, 193, 60, 94, 252, 193, 62, 94, 252, 193, 100, 94, 252, 193, 64, 94, 252, 193, 102, 94, 252, 193, 104, 94, 252, 193, 106, 94, 252, 193, 202, 93, 252, 193, 204, 93, 252, 193, 206, 93, 252, 193, 214, 93, 252, 193, 222, 93, 252, 193, 230, 93, 252, 193, 232, 93, 252, 193, 250, 93, 252, 193, 2, 94, 252, 193, 10, 94, 252, 193, 14, 94, 252, 193, 16, 94, 252, 193, 18, 94, 252, 193, 22, 94, 252, 193, 24, 94, 252, 193, 26, 94, 252, 193, 30, 94, 252, 193, 28, 94, 252, 193, 44, 94, 252, 193, 48, 94, 252, 193, 50, 94, 252, 193, 74, 94, 252, 193, 76, 94, 252, 193, 78, 94, 252, 193, 80, 94, 252, 193, 82, 94, 252, 193, 84, 94, 252, 193, 86, 94, 252, 193, 88, 94, 252, 193, 124, 94, 252, 193, 126, 94, 252, 193, 128, 94, 252, 193, 108, 94, 252, 193, 110, 94, 252, 193, 112, 94, 252, 193, 114, 94, 252, 193, 116, 94, 252, 193, 118, 94, 252, 193, 154, 94, 252, 193, 156, 94, 252, 193, 158, 94, 252, 193, 160, 94, 252, 193, 162, 94, 252, 193, 164, 94, 252, 193, 166, 94, 252, 193, 168, 94, 252, 193, 170, 94, 252, 193, 172, 94, 252, 193, 174, 94, 252, 193, 176, 94, 252, 193, 178, 94, 252, 193, 180, 94, 252, 193, 44, 95, 252, 193, 46, 95, 252, 193, 48, 95, 252, 193, 50, 95, 252, 193, 52, 95, 252, 193, 54, 95, 252, 193, 56, 95, 252, 193, 58, 95, 252, 193, 60, 95, 252, 193, 62, 95, 252, 193, 64, 95, 252, 193, 66, 95, 252, 193, 130, 94, 252, 193, 132, 94, 252, 193, 134, 94, 252, 193, 136, 94, 252, 193, 138, 94, 252, 193, 140, 94, 252, 193, 142, 94, 252, 193, 144, 94, 252, 193, 146, 94, 252, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 193, 150, 94, 252, 193, 12, 95, 252, 193, 128, 95, 252, 193, 130, 95, 252, 193, 228, 95, 252, 193, 230, 95, 252, 193, 232, 95, 252, 193, 82, 96, 252, 193, 84, 96, 252, 193, 86, 96, 252, 193, 88, 96, 252, 193, 184, 96, 252, 193, 44, 97, 252, 193, 46, 97, 252, 193, 48, 97, 252, 193, 50, 97, 252, 193, 52, 97, 252, 193, 54, 97, 252, 193, 56, 97, 252, 193, 100, 97, 252, 193, 102, 97, 252, 193, 166, 97, 252, 193, 208, 97, 252, 193, 210, 97, 252, 193, 58, 98, 252, 193, 144, 98, 252, 193, 180, 98, 252, 193, 2, 99, 252, 193, 4, 99, 252, 193, 122, 99, 252, 193, 124, 99, 252, 193, 126, 99, 252, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 206, 0, 56, 0, 193, 72, 95, 252, 193, 74, 95, 252, 193, 76, 95, 252, 193, 78, 95, 252, 193, 80, 95, 252, 193, 82, 95, 252, 193, 84, 95, 252, 193, 86, 95, 252, 193, 88, 95, 252, 193, 90, 95, 252, 193, 92, 95, 252, 193, 94, 95, 252, 193, 96, 95, 252, 193, 98, 95, 252, 193, 100, 95, 252, 193, 102, 95, 252, 193, 104, 95, 252, 193, 106, 95, 252, 193, 108, 95, 252, 193, 110, 95, 252, 193, 18, 95, 252, 193, 20, 95, 252, 193, 22, 95, 252, 193, 24, 95, 252, 193, 26, 95, 252, 193, 28, 95, 252, 193, 30, 95, 252, 193, 132, 95, 252, 193, 134, 95, 252, 193, 136, 95, 252, 193, 138, 95, 252, 193, 140, 95, 252, 193, 234, 95, 252, 193, 236, 95, 252, 193, 238, 95, 252, 193, 240, 95, 252, 193, 242, 95, 252, 193, 244, 95, 252, 193, 246, 95, 252, 193, 248, 95, 252, 193, 90, 96, 252, 193, 92, 96, 252, 193, 94, 96, 252, 193, 96, 96, 252, 193, 98, 96, 252, 193, 100, 96, 252, 193, 102, 96, 252, 193, 186, 96, 252, 193, 188, 96, 252, 193, 190, 96, 252, 193, 192, 96, 252, 193, 194, 96, 252, 193, 144, 95, 252, 193, 146, 95, 252, 193, 148, 95, 252, 193, 150, 95, 252, 193, 152, 95, 252, 193, 154, 95, 252, 193, 250, 95, 252, 193, 252, 95, 252, 193, 254, 95, 252, 193, 2, 96, 252, 193, 4, 96, 252, 193, 104, 96, 252, 193, 196, 96, 252, 193, 246, 96, 252, 193, 248, 96, 252, 193, 62, 97, 252, 193, 64, 97, 252, 193, 66, 97, 252, 193, 120, 97, 252, 193, 122, 97, 252, 193, 174, 97, 252, 193, 176, 97, 252, 193, 178, 97, 252, 193, 180, 97, 252, 193, 182, 97, 252, 193, 184, 97, 252, 193, 186, 97, 252, 193, 244, 97, 252, 193, 16, 98, 252, 193, 36, 98, 252, 193, 88, 98, 252, 193, 92, 98, 252, 193, 110, 107, 252, 193, 156, 95, 252, 193, 158, 95, 252, 193, 160, 95, 252, 193, 162, 95, 252, 193, 164, 95, 252, 193, 166, 95, 252, 193, 168, 95, 252, 193, 16, 169, 252, 193, 170, 95, 252, 193, 6, 96, 252, 193, 8, 96, 252, 193, 10, 96, 252, 193, 12, 96, 252, 193, 14, 96, 252, 193, 16, 96, 252, 193, 172, 95, 252, 193, 198, 96, 252, 193, 18, 96, 252, 193, 20, 96, 252, 193, 22, 96, 252, 193, 106, 96, 252, 193, 108, 96, 252, 193, 110, 96, 252, 193, 112, 96, 252, 193, 114, 96, 252, 193, 116, 96, 252, 193, 200, 96, 252, 193, 202, 96, 252, 193, 204, 96, 252, 193, 250, 96, 252, 193, 252, 96, 252, 193, 26, 96, 252, 193, 118, 96, 252, 193, 120, 96, 252, 193, 122, 96, 252, 193, 124, 96, 252, 193, 206, 96, 252, 193, 254, 96, 252, 193, 2, 97, 252, 193, 4, 97, 252, 193, 74, 97, 252, 193, 128, 97, 252, 193, 190, 97, 252, 193, 192, 97, 252, 193, 220, 97, 252, 193, 248, 97, 252, 193, 50, 98, 252, 193, 138, 98, 252, 193, 192, 98, 252, 193, 194, 98, 252, 193, 232, 98, 252, 193, 234, 98, 252, 193, 236, 98, 252, 193, 238, 98, 252, 193, 38, 99, 252, 193, 40, 99, 252, 193, 42, 99, 252, 193, 44, 99, 252, 193, 46, 99, 252, 193, 48, 99, 252, 193, 198, 99, 252, 193, 200, 99, 252, 193, 202, 99, 252, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 32, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 206, 64, 56, 0, 193, 208, 96, 252, 193, 210, 96, 252, 193, 212, 96, 252, 193, 214, 96, 252, 193, 216, 96, 252, 193, 218, 96, 252, 193, 220, 96, 252, 193, 222, 96, 252, 193, 224, 96, 252, 193, 226, 96, 252, 193, 228, 96, 252, 193, 230, 96, 252, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 206, 96, 56, 0, 193, 236, 96, 252, 193, 238, 96, 252, 193, 240, 96, 252, 193, 242, 96, 252, 193, 244, 96, 252, 193, 58, 97, 252, 193, 60, 97, 252, 193, 104, 97, 252, 193, 106, 97, 252, 193, 108, 97, 252, 193, 110, 97, 252, 193, 112, 97, 252, 193, 114, 97, 252, 193, 116, 97, 252, 193, 118, 97, 252, 193, 168, 97, 252, 193, 170, 97, 252, 193, 172, 97, 252, 193, 212, 97, 252, 193, 214, 97, 252, 193, 240, 97, 252, 193, 242, 97, 252, 193, 12, 98, 252, 193, 14, 98, 252, 193, 34, 98, 252, 193, 48, 98, 252, 193, 60, 98, 252, 193, 66, 98, 252, 193, 82, 98, 252, 193, 86, 98, 252, 193, 108, 98, 252, 193, 116, 98, 252, 193, 68, 97, 252, 193, 70, 97, 252, 193, 72, 97, 252, 193, 124, 97, 252, 193, 126, 97, 252, 193, 188, 97, 252, 193, 216, 97, 252, 193, 218, 97, 252, 193, 246, 97, 252, 193, 38, 98, 252, 193, 40, 98, 252, 193, 100, 98, 252, 193, 110, 98, 252, 193, 118, 98, 252, 193, 128, 98, 252, 193, 130, 98, 252, 193, 150, 98, 252, 193, 152, 98, 252, 193, 176, 98, 252, 193, 182, 98, 252, 193, 184, 98, 252, 193, 186, 98, 252, 193, 188, 98, 252, 193, 190, 98, 252, 193, 214, 98, 252, 193, 216, 98, 252, 193, 218, 98, 252, 193, 220, 98, 252, 193, 222, 98, 252, 193, 224, 98, 252, 193, 226, 98, 252, 193, 228, 98, 252, 193, 96, 97, 252, 193, 98, 97, 252, 193, 164, 97, 252, 193, 206, 97, 252, 193, 10, 98, 252, 193, 30, 98, 252, 193, 32, 98, 252, 193, 170, 98, 252, 193, 114, 99, 252, 193, 116, 99, 252, 193, 118, 99, 252, 193, 120, 99, 252, 193, 34, 100, 252, 193, 36, 100, 252, 193, 38, 100, 252, 193, 166, 100, 252, 193, 100, 101, 252, 193, 102, 101, 252, 193, 60, 102, 252, 193, 246, 102, 252, 193, 248, 102, 252, 193, 250, 102, 252, 193, 188, 103, 252, 193, 190, 103, 252, 193, 254, 104, 252, 193, 2, 105, 252, 193, 206, 105, 252, 193, 96, 107, 252, 193, 98, 107, 252, 193, 150, 107, 252, 193, 254, 107, 252, 193, 2, 108, 252, 206, 128, 56, 0, 206, 128, 56, 0, 206, 128, 56, 0, 206, 128, 56, 0, 193, 194, 97, 252, 193, 196, 97, 252, 193, 198, 97, 252, 193, 200, 97, 252, 193, 202, 97, 252, 193, 222, 97, 252, 193, 224, 97, 252, 193, 226, 97, 252, 193, 228, 97, 252, 193, 230, 97, 252, 193, 232, 97, 252, 193, 234, 97, 252, 193, 236, 97, 252, 193, 250, 97, 252, 193, 238, 97, 252, 193, 252, 97, 252, 193, 254, 97, 252, 193, 2, 98, 252, 193, 4, 98, 252, 193, 6, 98, 252, 193, 8, 98, 252, 193, 18, 98, 252, 193, 20, 98, 252, 193, 22, 98, 252, 193, 24, 98, 252, 193, 26, 98, 252, 193, 28, 98, 252, 193, 42, 98, 252, 193, 44, 98, 252, 193, 46, 98, 252, 193, 52, 98, 252, 193, 54, 98, 252, 193, 56, 98, 252, 193, 62, 98, 252, 193, 64, 98, 252, 193, 68, 98, 252, 193, 70, 98, 252, 193, 72, 98, 252, 193, 74, 98, 252, 193, 76, 98, 252, 193, 78, 98, 252, 193, 80, 98, 252, 193, 84, 98, 252, 193, 90, 98, 252, 193, 94, 98, 252, 193, 96, 98, 252, 193, 98, 98, 252, 193, 102, 98, 252, 193, 104, 98, 252, 193, 106, 98, 252, 193, 112, 98, 252, 193, 120, 98, 252, 193, 114, 98, 252, 193, 132, 98, 252, 193, 134, 98, 252, 193, 136, 98, 252, 193, 140, 98, 252, 193, 142, 98, 252, 193, 158, 98, 252, 193, 162, 98, 252, 193, 122, 98, 252, 193, 124, 98, 252, 193, 146, 98, 252, 193, 154, 98, 252, 193, 156, 98, 252, 193, 160, 98, 252, 193, 130, 99, 252, 193, 132, 99, 252, 193, 54, 100, 252, 193, 56, 100, 252, 193, 180, 100, 252, 193, 182, 100, 252, 193, 184, 100, 252, 193, 186, 100, 252, 193, 114, 101, 252, 193, 116, 101, 252, 193, 118, 101, 252, 193, 120, 101, 252, 193, 122, 101, 252, 193, 70, 102, 252, 193, 72, 102, 252, 193, 74, 102, 252, 193, 76, 102, 252, 193, 4, 103, 252, 193, 6, 103, 252, 193, 8, 103, 252, 193, 10, 103, 252, 193, 12, 103, 252, 193, 14, 103, 252, 193, 16, 103, 252, 193, 196, 103, 252, 193, 90, 104, 252, 193, 126, 98, 252, 193, 148, 98, 252, 193, 134, 99, 252, 193, 136, 99, 252, 193, 138, 99, 252, 193, 140, 99, 252, 193, 142, 99, 252, 193, 144, 99, 252, 193, 146, 99, 252, 193, 58, 100, 252, 193, 60, 100, 252, 193, 188, 100, 252, 193, 190, 100, 252, 193, 124, 101, 252, 193, 126, 101, 252, 193, 128, 101, 252, 193, 78, 102, 252, 193, 80, 102, 252, 193, 82, 102, 252, 193, 84, 102, 252, 193, 86, 102, 252, 193, 18, 103, 252, 193, 20, 103, 252, 193, 22, 103, 252, 193, 24, 103, 252, 193, 26, 103, 252, 193, 198, 103, 252, 193, 200, 103, 252, 193, 202, 103, 252, 193, 204, 103, 252, 193, 108, 104, 252, 193, 110, 104, 252, 193, 164, 98, 252, 193, 166, 98, 252, 193, 168, 98, 252, 193, 172, 98, 252, 193, 174, 98, 252, 193, 178, 98, 252, 193, 196, 98, 252, 193, 198, 98, 252, 193, 200, 98, 252, 193, 202, 98, 252, 193, 204, 98, 252, 193, 206, 98, 252, 193, 208, 98, 252, 193, 210, 98, 252, 193, 212, 98, 252, 193, 240, 98, 252, 193, 242, 98, 252, 193, 244, 98, 252, 193, 246, 98, 252, 193, 248, 98, 252, 193, 250, 98, 252, 193, 252, 98, 252, 193, 254, 98, 252, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 193, 230, 98, 252, 193, 6, 99, 252, 193, 8, 99, 252, 193, 10, 99, 252, 193, 12, 99, 252, 193, 14, 99, 252, 193, 16, 99, 252, 193, 18, 99, 252, 193, 20, 99, 252, 193, 22, 99, 252, 193, 24, 99, 252, 193, 26, 99, 252, 193, 28, 99, 252, 193, 30, 99, 252, 193, 32, 99, 252, 193, 34, 99, 252, 193, 148, 99, 252, 193, 150, 99, 252, 193, 152, 99, 252, 193, 154, 99, 252, 193, 156, 99, 252, 193, 158, 99, 252, 193, 160, 99, 252, 193, 162, 99, 252, 193, 164, 99, 252, 193, 166, 99, 252, 193, 36, 99, 252, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 160, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 193, 128, 99, 252, 193, 42, 100, 252, 193, 44, 100, 252, 193, 46, 100, 252, 193, 48, 100, 252, 193, 50, 100, 252, 193, 52, 100, 252, 193, 168, 100, 252, 193, 170, 100, 252, 193, 172, 100, 252, 193, 174, 100, 252, 193, 176, 100, 252, 193, 178, 100, 252, 193, 104, 101, 252, 193, 106, 101, 252, 193, 108, 101, 252, 193, 110, 101, 252, 193, 112, 101, 252, 193, 62, 102, 252, 193, 64, 102, 252, 193, 66, 102, 252, 193, 68, 102, 252, 193, 252, 102, 252, 193, 254, 102, 252, 193, 2, 103, 252, 193, 194, 103, 252, 193, 86, 104, 252, 193, 88, 104, 252, 193, 4, 105, 252, 193, 6, 105, 252, 193, 126, 105, 252, 193, 54, 106, 252, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 206, 192, 56, 0, 193, 62, 100, 252, 193, 64, 100, 252, 193, 66, 100, 252, 193, 68, 100, 252, 193, 70, 100, 252, 193, 72, 100, 252, 193, 74, 100, 252, 193, 76, 100, 252, 193, 78, 100, 252, 193, 80, 100, 252, 193, 82, 100, 252, 193, 84, 100, 252, 193, 86, 100, 252, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 193, 204, 99, 252, 193, 206, 99, 252, 193, 208, 99, 252, 193, 210, 99, 252, 193, 212, 99, 252, 193, 214, 99, 252, 193, 216, 99, 252, 193, 218, 99, 252, 193, 220, 99, 252, 193, 222, 99, 252, 193, 88, 100, 252, 193, 90, 100, 252, 193, 92, 100, 252, 193, 94, 100, 252, 193, 96, 100, 252, 193, 98, 100, 252, 193, 100, 100, 252, 193, 102, 100, 252, 193, 104, 100, 252, 193, 234, 100, 252, 193, 236, 100, 252, 193, 238, 100, 252, 193, 240, 100, 252, 193, 242, 100, 252, 193, 244, 100, 252, 193, 246, 100, 252, 193, 248, 100, 252, 193, 250, 100, 252, 193, 252, 100, 252, 193, 254, 100, 252, 193, 2, 101, 252, 193, 4, 101, 252, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 224, 56, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 206, 0, 57, 0, 193, 10, 101, 252, 193, 12, 101, 252, 193, 14, 101, 252, 193, 16, 101, 252, 193, 18, 101, 252, 193, 20, 101, 252, 193, 22, 101, 252, 193, 24, 101, 252, 193, 26, 101, 252, 193, 28, 101, 252, 193, 30, 101, 252, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 32, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 193, 6, 101, 252, 193, 8, 101, 252, 193, 184, 101, 252, 193, 186, 101, 252, 193, 188, 101, 252, 193, 190, 101, 252, 193, 192, 101, 252, 193, 194, 101, 252, 193, 196, 101, 252, 193, 198, 101, 252, 193, 128, 102, 252, 193, 130, 102, 252, 193, 132, 102, 252, 193, 134, 102, 252, 193, 136, 102, 252, 193, 138, 102, 252, 193, 140, 102, 252, 193, 142, 102, 252, 193, 144, 102, 252, 193, 146, 102, 252, 193, 148, 102, 252, 193, 150, 102, 252, 193, 62, 103, 252, 193, 64, 103, 252, 193, 66, 103, 252, 193, 68, 103, 252, 193, 70, 103, 252, 193, 72, 103, 252, 193, 74, 103, 252, 193, 76, 103, 252, 193, 78, 103, 252, 193, 80, 103, 252, 193, 32, 101, 252, 193, 34, 101, 252, 193, 36, 101, 252, 193, 38, 101, 252, 193, 40, 101, 252, 193, 42, 101, 252, 193, 44, 101, 252, 193, 46, 101, 252, 193, 48, 101, 252, 193, 50, 101, 252, 193, 52, 101, 252, 193, 54, 101, 252, 193, 56, 101, 252, 193, 58, 101, 252, 193, 60, 101, 252, 193, 62, 101, 252, 193, 64, 101, 252, 193, 66, 101, 252, 193, 68, 101, 252, 193, 70, 101, 252, 193, 72, 101, 252, 193, 74, 101, 252, 193, 76, 101, 252, 193, 78, 101, 252, 193, 80, 101, 252, 193, 82, 101, 252, 193, 84, 101, 252, 193, 86, 101, 252, 193, 88, 101, 252, 193, 90, 101, 252, 193, 92, 101, 252, 193, 94, 101, 252, 193, 96, 101, 252, 193, 98, 101, 252, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 206, 64, 57, 0, 193, 88, 102, 252, 193, 90, 102, 252, 193, 92, 102, 252, 193, 94, 102, 252, 193, 96, 102, 252, 193, 98, 102, 252, 193, 100, 102, 252, 193, 102, 102, 252, 193, 104, 102, 252, 193, 106, 102, 252, 193, 108, 102, 252, 193, 110, 102, 252, 193, 112, 102, 252, 193, 114, 102, 252, 193, 116, 102, 252, 193, 118, 102, 252, 193, 120, 102, 252, 193, 122, 102, 252, 193, 28, 103, 252, 193, 30, 103, 252, 193, 32, 103, 252, 193, 34, 103, 252, 193, 36, 103, 252, 193, 38, 103, 252, 193, 124, 102, 252, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 96, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 193, 40, 103, 252, 193, 42, 103, 252, 193, 44, 103, 252, 193, 46, 103, 252, 193, 48, 103, 252, 193, 50, 103, 252, 193, 52, 103, 252, 193, 126, 102, 252, 193, 54, 103, 252, 193, 56, 103, 252, 193, 58, 103, 252, 193, 60, 103, 252, 193, 206, 103, 252, 193, 208, 103, 252, 193, 210, 103, 252, 193, 212, 103, 252, 193, 214, 103, 252, 193, 216, 103, 252, 193, 218, 103, 252, 193, 220, 103, 252, 193, 222, 103, 252, 193, 224, 103, 252, 193, 226, 103, 252, 193, 228, 103, 252, 193, 230, 103, 252, 193, 232, 103, 252, 193, 234, 103, 252, 193, 236, 103, 252, 193, 238, 103, 252, 193, 240, 103, 252, 193, 242, 103, 252, 193, 244, 103, 252, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 128, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 193, 82, 103, 252, 193, 84, 103, 252, 193, 248, 103, 252, 193, 250, 103, 252, 193, 252, 103, 252, 193, 254, 103, 252, 193, 2, 104, 252, 193, 4, 104, 252, 193, 6, 104, 252, 193, 142, 104, 252, 193, 144, 104, 252, 193, 60, 105, 252, 193, 62, 105, 252, 193, 64, 105, 252, 193, 66, 105, 252, 193, 68, 105, 252, 193, 70, 105, 252, 193, 146, 105, 252, 193, 148, 105, 252, 193, 150, 105, 252, 193, 228, 105, 252, 193, 26, 106, 252, 193, 28, 106, 252, 193, 30, 106, 252, 193, 66, 106, 252, 193, 86, 106, 252, 193, 108, 106, 252, 193, 110, 106, 252, 193, 144, 106, 252, 193, 18, 107, 252, 193, 62, 107, 252, 193, 122, 107, 252, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 206, 160, 57, 0, 193, 8, 104, 252, 193, 10, 104, 252, 193, 12, 104, 252, 193, 114, 104, 252, 193, 116, 104, 252, 193, 118, 104, 252, 193, 120, 104, 252, 193, 122, 104, 252, 193, 124, 104, 252, 193, 126, 104, 252, 193, 128, 104, 252, 193, 130, 104, 252, 193, 132, 104, 252, 193, 246, 103, 252, 193, 134, 104, 252, 193, 136, 104, 252, 193, 138, 104, 252, 193, 140, 104, 252, 193, 28, 105, 252, 193, 30, 105, 252, 193, 32, 105, 252, 193, 34, 105, 252, 193, 36, 105, 252, 193, 38, 105, 252, 193, 40, 105, 252, 193, 42, 105, 252, 193, 44, 105, 252, 193, 46, 105, 252, 193, 48, 105, 252, 193, 50, 105, 252, 193, 52, 105, 252, 193, 54, 105, 252, 193, 56, 105, 252, 193, 58, 105, 252, 193, 138, 105, 252, 193, 14, 104, 252, 193, 16, 104, 252, 193, 18, 104, 252, 193, 20, 104, 252, 193, 22, 104, 252, 193, 24, 104, 252, 193, 26, 104, 252, 193, 28, 104, 252, 193, 30, 104, 252, 193, 32, 104, 252, 193, 34, 104, 252, 193, 36, 104, 252, 193, 38, 104, 252, 193, 40, 104, 252, 193, 42, 104, 252, 193, 44, 104, 252, 193, 46, 104, 252, 193, 48, 104, 252, 193, 50, 104, 252, 193, 52, 104, 252, 193, 54, 104, 252, 193, 56, 104, 252, 193, 58, 104, 252, 193, 60, 104, 252, 193, 62, 104, 252, 193, 64, 104, 252, 193, 66, 104, 252, 193, 68, 104, 252, 193, 70, 104, 252, 193, 72, 104, 252, 193, 74, 104, 252, 193, 76, 104, 252, 193, 78, 104, 252, 193, 80, 104, 252, 193, 82, 104, 252, 193, 146, 104, 252, 193, 148, 104, 252, 193, 150, 104, 252, 193, 152, 104, 252, 193, 154, 104, 252, 193, 156, 104, 252, 193, 158, 104, 252, 193, 160, 104, 252, 193, 162, 104, 252, 193, 164, 104, 252, 193, 166, 104, 252, 193, 168, 104, 252, 193, 170, 104, 252, 193, 172, 104, 252, 193, 174, 104, 252, 193, 176, 104, 252, 193, 178, 104, 252, 193, 180, 104, 252, 193, 182, 104, 252, 193, 184, 104, 252, 193, 186, 104, 252, 193, 188, 104, 252, 193, 190, 104, 252, 193, 192, 104, 252, 193, 194, 104, 252, 193, 196, 104, 252, 193, 198, 104, 252, 193, 84, 104, 252, 193, 200, 104, 252, 193, 92, 104, 252, 193, 94, 104, 252, 193, 96, 104, 252, 193, 98, 104, 252, 193, 100, 104, 252, 193, 102, 104, 252, 193, 104, 104, 252, 193, 106, 104, 252, 193, 8, 105, 252, 193, 10, 105, 252, 193, 12, 105, 252, 193, 14, 105, 252, 193, 16, 105, 252, 193, 18, 105, 252, 193, 128, 105, 252, 193, 130, 105, 252, 193, 208, 105, 252, 193, 210, 105, 252, 193, 212, 105, 252, 193, 214, 105, 252, 193, 216, 105, 252, 193, 8, 106, 252, 193, 10, 106, 252, 193, 12, 106, 252, 193, 58, 106, 252, 193, 120, 106, 252, 193, 122, 106, 252, 193, 140, 106, 252, 193, 148, 106, 252, 193, 158, 106, 252, 193, 180, 106, 252, 193, 182, 106, 252, 193, 112, 104, 252, 193, 20, 105, 252, 193, 22, 105, 252, 193, 24, 105, 252, 193, 26, 105, 252, 193, 132, 105, 252, 193, 134, 105, 252, 193, 136, 105, 252, 193, 218, 105, 252, 193, 14, 106, 252, 193, 80, 106, 252, 193, 154, 106, 252, 193, 166, 106, 252, 193, 168, 106, 252, 193, 184, 106, 252, 193, 192, 106, 252, 193, 218, 106, 252, 193, 228, 106, 252, 193, 8, 107, 252, 193, 50, 107, 252, 193, 52, 107, 252, 193, 170, 107, 252, 193, 210, 107, 252, 193, 212, 107, 252, 193, 214, 107, 252, 193, 8, 108, 252, 193, 68, 108, 252, 193, 70, 108, 252, 193, 72, 108, 252, 193, 186, 108, 252, 193, 32, 109, 252, 193, 110, 109, 252, 193, 202, 104, 252, 193, 204, 104, 252, 193, 206, 104, 252, 193, 208, 104, 252, 193, 210, 104, 252, 193, 212, 104, 252, 193, 214, 104, 252, 193, 216, 104, 252, 193, 218, 104, 252, 193, 220, 104, 252, 193, 222, 104, 252, 193, 224, 104, 252, 193, 226, 104, 252, 193, 228, 104, 252, 193, 230, 104, 252, 193, 232, 104, 252, 193, 234, 104, 252, 193, 236, 104, 252, 193, 238, 104, 252, 193, 240, 104, 252, 193, 242, 104, 252, 193, 244, 104, 252, 193, 246, 104, 252, 193, 248, 104, 252, 193, 250, 104, 252, 193, 252, 104, 252, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 192, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 193, 140, 105, 252, 193, 142, 105, 252, 193, 220, 105, 252, 193, 144, 105, 252, 193, 222, 105, 252, 193, 224, 105, 252, 193, 226, 105, 252, 193, 16, 106, 252, 193, 18, 106, 252, 193, 20, 106, 252, 193, 22, 106, 252, 193, 24, 106, 252, 193, 60, 106, 252, 193, 62, 106, 252, 193, 64, 106, 252, 193, 82, 106, 252, 193, 84, 106, 252, 193, 104, 106, 252, 193, 106, 106, 252, 193, 130, 106, 252, 193, 142, 106, 252, 193, 152, 106, 252, 193, 156, 106, 252, 193, 164, 106, 252, 193, 170, 106, 252, 193, 172, 106, 252, 193, 174, 106, 252, 193, 202, 106, 252, 193, 204, 106, 252, 193, 250, 106, 252, 193, 252, 106, 252, 193, 254, 106, 252, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 206, 224, 57, 0, 193, 230, 105, 252, 193, 232, 105, 252, 193, 234, 105, 252, 193, 236, 105, 252, 193, 238, 105, 252, 193, 240, 105, 252, 193, 242, 105, 252, 193, 244, 105, 252, 193, 246, 105, 252, 193, 248, 105, 252, 193, 250, 105, 252, 193, 252, 105, 252, 193, 254, 105, 252, 193, 2, 106, 252, 193, 4, 106, 252, 193, 6, 106, 252, 193, 32, 106, 252, 193, 34, 106, 252, 193, 36, 106, 252, 193, 38, 106, 252, 193, 40, 106, 252, 193, 42, 106, 252, 193, 44, 106, 252, 193, 46, 106, 252, 193, 48, 106, 252, 193, 50, 106, 252, 193, 52, 106, 252, 193, 68, 106, 252, 193, 70, 106, 252, 193, 72, 106, 252, 193, 74, 106, 252, 193, 76, 106, 252, 193, 88, 106, 252, 193, 90, 106, 252, 193, 92, 106, 252, 193, 94, 106, 252, 193, 96, 106, 252, 193, 98, 106, 252, 193, 100, 106, 252, 193, 102, 106, 252, 193, 112, 106, 252, 193, 114, 106, 252, 193, 116, 106, 252, 193, 118, 106, 252, 193, 124, 106, 252, 193, 126, 106, 252, 193, 128, 106, 252, 193, 132, 106, 252, 193, 56, 106, 252, 193, 78, 106, 252, 193, 214, 106, 252, 193, 6, 107, 252, 193, 100, 107, 252, 193, 102, 107, 252, 193, 154, 107, 252, 193, 156, 107, 252, 193, 198, 107, 252, 193, 200, 107, 252, 193, 4, 108, 252, 193, 52, 108, 252, 193, 54, 108, 252, 193, 56, 108, 252, 193, 58, 108, 252, 193, 106, 108, 252, 193, 108, 108, 252, 193, 10, 109, 252, 193, 72, 109, 252, 193, 102, 109, 252, 193, 104, 109, 252, 193, 190, 109, 252, 193, 38, 110, 252, 193, 216, 110, 252, 193, 4, 111, 252, 193, 6, 111, 252, 193, 72, 111, 252, 193, 74, 111, 252, 193, 138, 111, 252, 193, 140, 111, 252, 193, 142, 111, 252, 193, 182, 111, 252, 193, 134, 106, 252, 193, 136, 106, 252, 193, 138, 106, 252, 193, 146, 106, 252, 193, 150, 106, 252, 193, 160, 106, 252, 193, 162, 106, 252, 193, 176, 106, 252, 193, 178, 106, 252, 193, 186, 106, 252, 193, 188, 106, 252, 193, 190, 106, 252, 193, 194, 106, 252, 193, 196, 106, 252, 193, 198, 106, 252, 193, 206, 106, 252, 193, 208, 106, 252, 193, 210, 106, 252, 193, 212, 106, 252, 193, 220, 106, 252, 193, 222, 106, 252, 193, 224, 106, 252, 193, 230, 106, 252, 193, 234, 106, 252, 193, 236, 106, 252, 193, 240, 106, 252, 193, 242, 106, 252, 193, 248, 106, 252, 193, 244, 106, 252, 193, 2, 107, 252, 193, 4, 107, 252, 193, 20, 107, 252, 193, 200, 106, 252, 193, 216, 106, 252, 193, 226, 106, 252, 193, 232, 106, 252, 193, 238, 106, 252, 193, 246, 106, 252, 193, 26, 107, 252, 193, 42, 107, 252, 193, 44, 107, 252, 193, 46, 107, 252, 193, 48, 107, 252, 193, 104, 107, 252, 193, 106, 107, 252, 193, 108, 107, 252, 193, 158, 107, 252, 193, 160, 107, 252, 193, 162, 107, 252, 193, 164, 107, 252, 193, 166, 107, 252, 193, 168, 107, 252, 193, 202, 107, 252, 193, 204, 107, 252, 193, 206, 107, 252, 193, 208, 107, 252, 193, 6, 108, 252, 193, 60, 108, 252, 193, 62, 108, 252, 193, 64, 108, 252, 193, 66, 108, 252, 193, 110, 108, 252, 193, 112, 108, 252, 193, 142, 108, 252, 193, 10, 107, 252, 193, 12, 107, 252, 193, 14, 107, 252, 193, 16, 107, 252, 193, 28, 107, 252, 193, 54, 107, 252, 193, 56, 107, 252, 193, 58, 107, 252, 193, 60, 107, 252, 193, 112, 107, 252, 193, 114, 107, 252, 193, 116, 107, 252, 193, 118, 107, 252, 193, 120, 107, 252, 193, 172, 107, 252, 193, 174, 107, 252, 193, 176, 107, 252, 193, 178, 107, 252, 193, 216, 107, 252, 193, 218, 107, 252, 193, 220, 107, 252, 193, 222, 107, 252, 193, 224, 107, 252, 193, 10, 108, 252, 193, 12, 108, 252, 193, 14, 108, 252, 193, 114, 108, 252, 193, 116, 108, 252, 193, 118, 108, 252, 193, 120, 108, 252, 193, 122, 108, 252, 193, 144, 108, 252, 193, 22, 107, 252, 193, 24, 107, 252, 193, 30, 107, 252, 193, 32, 107, 252, 193, 34, 107, 252, 193, 36, 107, 252, 193, 38, 107, 252, 193, 40, 107, 252, 193, 64, 107, 252, 193, 66, 107, 252, 193, 68, 107, 252, 193, 70, 107, 252, 193, 72, 107, 252, 193, 74, 107, 252, 193, 76, 107, 252, 193, 78, 107, 252, 193, 80, 107, 252, 193, 82, 107, 252, 193, 84, 107, 252, 193, 86, 107, 252, 193, 88, 107, 252, 193, 90, 107, 252, 193, 92, 107, 252, 193, 94, 107, 252, 206, 0, 58, 0, 206, 0, 58, 0, 206, 0, 58, 0, 206, 0, 58, 0, 206, 0, 58, 0, 206, 0, 58, 0, 206, 0, 58, 0, 206, 0, 58, 0, 206, 0, 58, 0, 193, 184, 107, 252, 193, 186, 107, 252, 193, 188, 107, 252, 193, 190, 107, 252, 193, 192, 107, 252, 193, 194, 107, 252, 193, 196, 107, 252, 193, 228, 107, 252, 193, 230, 107, 252, 193, 232, 107, 252, 193, 234, 107, 252, 193, 236, 107, 252, 193, 238, 107, 252, 193, 240, 107, 252, 193, 242, 107, 252, 193, 244, 107, 252, 193, 246, 107, 252, 193, 248, 107, 252, 193, 250, 107, 252, 193, 252, 107, 252, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 193, 180, 107, 252, 193, 182, 107, 252, 193, 16, 108, 252, 193, 74, 108, 252, 193, 76, 108, 252, 193, 124, 108, 252, 193, 150, 108, 252, 193, 208, 108, 252, 193, 242, 108, 252, 193, 20, 109, 252, 193, 36, 109, 252, 193, 52, 109, 252, 193, 54, 109, 252, 193, 78, 109, 252, 193, 80, 109, 252, 193, 120, 109, 252, 193, 122, 109, 252, 193, 166, 109, 252, 193, 204, 109, 252, 193, 206, 109, 252, 193, 4, 110, 252, 193, 6, 110, 252, 193, 48, 110, 252, 193, 50, 110, 252, 193, 78, 110, 252, 193, 118, 110, 252, 193, 154, 110, 252, 193, 202, 110, 252, 193, 232, 110, 252, 193, 234, 110, 252, 193, 28, 111, 252, 193, 30, 111, 252, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 206, 32, 58, 0, 193, 78, 108, 252, 193, 80, 108, 252, 193, 82, 108, 252, 193, 84, 108, 252, 193, 86, 108, 252, 193, 88, 108, 252, 193, 90, 108, 252, 193, 92, 108, 252, 193, 94, 108, 252, 193, 96, 108, 252, 193, 98, 108, 252, 193, 100, 108, 252, 193, 126, 108, 252, 193, 128, 108, 252, 193, 130, 108, 252, 193, 132, 108, 252, 193, 134, 108, 252, 193, 136, 108, 252, 193, 138, 108, 252, 193, 152, 108, 252, 193, 154, 108, 252, 193, 156, 108, 252, 193, 158, 108, 252, 193, 160, 108, 252, 193, 48, 108, 252, 193, 50, 108, 252, 193, 102, 108, 252, 193, 104, 108, 252, 193, 140, 108, 252, 193, 250, 108, 252, 193, 50, 109, 252, 193, 70, 109, 252, 193, 34, 110, 252, 193, 36, 110, 252, 193, 252, 110, 252, 193, 254, 110, 252, 193, 68, 111, 252, 193, 70, 111, 252, 193, 136, 111, 252, 193, 178, 111, 252, 193, 180, 111, 252, 193, 228, 111, 252, 193, 230, 111, 252, 193, 20, 112, 252, 193, 22, 112, 252, 193, 82, 112, 252, 193, 146, 112, 252, 193, 202, 112, 252, 193, 204, 112, 252, 193, 236, 112, 252, 193, 8, 113, 252, 193, 58, 113, 252, 193, 82, 113, 252, 193, 106, 113, 252, 193, 122, 113, 252, 193, 228, 113, 252, 193, 146, 108, 252, 193, 148, 108, 252, 193, 164, 108, 252, 193, 166, 108, 252, 193, 168, 108, 252, 193, 170, 108, 252, 193, 188, 108, 252, 193, 190, 108, 252, 193, 192, 108, 252, 193, 206, 108, 252, 193, 228, 108, 252, 193, 230, 108, 252, 193, 240, 108, 252, 193, 4, 109, 252, 193, 26, 109, 252, 193, 34, 109, 252, 193, 74, 109, 252, 193, 76, 109, 252, 193, 118, 109, 252, 193, 162, 109, 252, 193, 164, 109, 252, 193, 200, 109, 252, 193, 250, 109, 252, 193, 252, 109, 252, 193, 254, 109, 252, 193, 2, 110, 252, 193, 46, 110, 252, 193, 76, 110, 252, 193, 106, 110, 252, 193, 116, 110, 252, 193, 152, 110, 252, 193, 184, 110, 252, 193, 162, 108, 252, 193, 202, 108, 252, 193, 204, 108, 252, 193, 106, 109, 252, 193, 108, 109, 252, 193, 140, 109, 252, 193, 142, 109, 252, 193, 144, 109, 252, 193, 146, 109, 252, 193, 148, 109, 252, 193, 150, 109, 252, 193, 192, 109, 252, 193, 236, 109, 252, 193, 238, 109, 252, 193, 240, 109, 252, 193, 40, 110, 252, 193, 42, 110, 252, 193, 66, 110, 252, 193, 68, 110, 252, 193, 70, 110, 252, 193, 102, 110, 252, 193, 104, 110, 252, 193, 114, 110, 252, 193, 138, 110, 252, 193, 148, 110, 252, 193, 150, 110, 252, 193, 176, 110, 252, 193, 182, 110, 252, 193, 8, 111, 252, 193, 76, 111, 252, 193, 78, 111, 252, 193, 80, 111, 252, 193, 172, 108, 252, 193, 174, 108, 252, 193, 176, 108, 252, 193, 178, 108, 252, 193, 180, 108, 252, 193, 182, 108, 252, 193, 184, 108, 252, 193, 194, 108, 252, 193, 196, 108, 252, 193, 198, 108, 252, 193, 200, 108, 252, 193, 210, 108, 252, 193, 212, 108, 252, 193, 214, 108, 252, 193, 216, 108, 252, 193, 218, 108, 252, 193, 220, 108, 252, 193, 222, 108, 252, 193, 224, 108, 252, 193, 226, 108, 252, 193, 232, 108, 252, 193, 234, 108, 252, 193, 236, 108, 252, 193, 238, 108, 252, 193, 244, 108, 252, 193, 246, 108, 252, 193, 248, 108, 252, 193, 252, 108, 252, 193, 254, 108, 252, 193, 2, 109, 252, 193, 6, 109, 252, 193, 8, 109, 252, 193, 12, 109, 252, 193, 14, 109, 252, 193, 16, 109, 252, 193, 18, 109, 252, 193, 22, 109, 252, 193, 24, 109, 252, 193, 28, 109, 252, 193, 30, 109, 252, 193, 38, 109, 252, 193, 40, 109, 252, 193, 42, 109, 252, 193, 44, 109, 252, 193, 46, 109, 252, 193, 48, 109, 252, 193, 56, 109, 252, 193, 58, 109, 252, 193, 60, 109, 252, 193, 62, 109, 252, 193, 64, 109, 252, 193, 66, 109, 252, 193, 68, 109, 252, 193, 82, 109, 252, 193, 84, 109, 252, 193, 86, 109, 252, 193, 88, 109, 252, 193, 90, 109, 252, 193, 92, 109, 252, 193, 94, 109, 252, 193, 96, 109, 252, 193, 98, 109, 252, 193, 100, 109, 252, 193, 124, 109, 252, 193, 112, 109, 252, 193, 114, 109, 252, 193, 116, 109, 252, 193, 152, 109, 252, 193, 154, 109, 252, 193, 156, 109, 252, 193, 158, 109, 252, 193, 160, 109, 252, 193, 194, 109, 252, 193, 196, 109, 252, 193, 198, 109, 252, 193, 242, 109, 252, 193, 244, 109, 252, 193, 246, 109, 252, 193, 248, 109, 252, 193, 44, 110, 252, 193, 72, 110, 252, 193, 74, 110, 252, 193, 194, 110, 252, 193, 196, 110, 252, 193, 198, 110, 252, 193, 218, 110, 252, 193, 82, 111, 252, 193, 84, 111, 252, 193, 86, 111, 252, 193, 88, 111, 252, 193, 190, 111, 252, 193, 192, 111, 252, 193, 254, 111, 252, 193, 50, 112, 252, 193, 52, 112, 252, 193, 54, 112, 252, 193, 126, 109, 252, 193, 128, 109, 252, 193, 130, 109, 252, 193, 132, 109, 252, 193, 134, 109, 252, 193, 136, 109, 252, 193, 138, 109, 252, 193, 168, 109, 252, 193, 170, 109, 252, 193, 172, 109, 252, 193, 174, 109, 252, 193, 176, 109, 252, 193, 178, 109, 252, 193, 180, 109, 252, 193, 182, 109, 252, 193, 208, 109, 252, 193, 184, 109, 252, 193, 186, 109, 252, 193, 188, 109, 252, 193, 210, 109, 252, 193, 212, 109, 252, 193, 214, 109, 252, 193, 216, 109, 252, 193, 218, 109, 252, 193, 220, 109, 252, 193, 222, 109, 252, 193, 224, 109, 252, 193, 226, 109, 252, 193, 228, 109, 252, 193, 230, 109, 252, 193, 232, 109, 252, 193, 234, 109, 252, 193, 8, 110, 252, 193, 10, 110, 252, 193, 12, 110, 252, 193, 14, 110, 252, 193, 16, 110, 252, 193, 18, 110, 252, 193, 20, 110, 252, 193, 22, 110, 252, 193, 24, 110, 252, 193, 26, 110, 252, 193, 28, 110, 252, 193, 30, 110, 252, 193, 32, 110, 252, 193, 52, 110, 252, 193, 54, 110, 252, 193, 56, 110, 252, 193, 58, 110, 252, 193, 60, 110, 252, 193, 62, 110, 252, 193, 64, 110, 252, 193, 80, 110, 252, 193, 82, 110, 252, 193, 84, 110, 252, 193, 86, 110, 252, 193, 88, 110, 252, 193, 90, 110, 252, 193, 92, 110, 252, 193, 94, 110, 252, 193, 96, 110, 252, 193, 98, 110, 252, 193, 100, 110, 252, 193, 108, 110, 252, 193, 110, 110, 252, 193, 112, 110, 252, 193, 120, 110, 252, 193, 122, 110, 252, 193, 124, 110, 252, 193, 126, 110, 252, 193, 128, 110, 252, 193, 130, 110, 252, 193, 132, 110, 252, 193, 134, 110, 252, 193, 136, 110, 252, 193, 140, 110, 252, 193, 142, 110, 252, 193, 144, 110, 252, 193, 146, 110, 252, 193, 156, 110, 252, 193, 158, 110, 252, 193, 160, 110, 252, 193, 162, 110, 252, 193, 164, 110, 252, 193, 166, 110, 252, 193, 168, 110, 252, 193, 170, 110, 252, 193, 172, 110, 252, 193, 174, 110, 252, 193, 178, 110, 252, 193, 180, 110, 252, 193, 196, 218, 253, 193, 208, 218, 253, 193, 214, 218, 253, 193, 218, 218, 253, 193, 190, 110, 252, 193, 186, 110, 252, 193, 188, 110, 252, 193, 200, 110, 252, 193, 220, 110, 252, 193, 222, 110, 252, 193, 224, 110, 252, 193, 226, 110, 252, 193, 228, 110, 252, 193, 230, 110, 252, 193, 10, 111, 252, 193, 12, 111, 252, 193, 14, 111, 252, 193, 16, 111, 252, 193, 18, 111, 252, 193, 20, 111, 252, 193, 22, 111, 252, 193, 24, 111, 252, 193, 26, 111, 252, 193, 90, 111, 252, 193, 92, 111, 252, 193, 94, 111, 252, 193, 96, 111, 252, 193, 98, 111, 252, 193, 100, 111, 252, 193, 102, 111, 252, 193, 148, 111, 252, 193, 150, 111, 252, 193, 104, 111, 252, 193, 152, 111, 252, 193, 194, 111, 252, 193, 196, 111, 252, 193, 198, 111, 252, 193, 192, 110, 252, 193, 204, 110, 252, 193, 206, 110, 252, 193, 208, 110, 252, 193, 210, 110, 252, 193, 212, 110, 252, 193, 236, 110, 252, 193, 238, 110, 252, 193, 240, 110, 252, 193, 242, 110, 252, 193, 244, 110, 252, 193, 246, 110, 252, 193, 248, 110, 252, 193, 250, 110, 252, 193, 34, 111, 252, 193, 36, 111, 252, 193, 38, 111, 252, 193, 40, 111, 252, 193, 42, 111, 252, 193, 44, 111, 252, 193, 46, 111, 252, 193, 48, 111, 252, 193, 50, 111, 252, 193, 52, 111, 252, 193, 54, 111, 252, 193, 56, 111, 252, 193, 58, 111, 252, 193, 60, 111, 252, 193, 62, 111, 252, 193, 64, 111, 252, 193, 110, 111, 252, 193, 112, 111, 252, 193, 32, 111, 252, 193, 106, 111, 252, 193, 108, 111, 252, 193, 154, 111, 252, 193, 200, 111, 252, 193, 202, 111, 252, 193, 10, 112, 252, 193, 106, 112, 252, 193, 108, 112, 252, 193, 164, 112, 252, 193, 246, 112, 252, 193, 52, 113, 252, 193, 70, 113, 252, 193, 120, 113, 252, 193, 136, 113, 252, 193, 138, 113, 252, 193, 184, 113, 252, 193, 186, 113, 252, 193, 188, 113, 252, 193, 190, 113, 252, 193, 192, 113, 252, 193, 46, 114, 252, 193, 48, 114, 252, 193, 50, 114, 252, 193, 52, 114, 252, 193, 54, 114, 252, 193, 56, 114, 252, 193, 58, 114, 252, 193, 60, 114, 252, 193, 62, 114, 252, 193, 64, 114, 252, 193, 2, 115, 252, 193, 114, 111, 252, 193, 116, 111, 252, 193, 118, 111, 252, 193, 120, 111, 252, 193, 122, 111, 252, 193, 124, 111, 252, 193, 126, 111, 252, 193, 128, 111, 252, 193, 130, 111, 252, 193, 132, 111, 252, 193, 134, 111, 252, 193, 66, 111, 252, 193, 156, 111, 252, 193, 158, 111, 252, 193, 160, 111, 252, 193, 162, 111, 252, 193, 164, 111, 252, 193, 166, 111, 252, 193, 168, 111, 252, 193, 170, 111, 252, 193, 172, 111, 252, 193, 174, 111, 252, 193, 176, 111, 252, 193, 204, 111, 252, 193, 206, 111, 252, 193, 208, 111, 252, 193, 210, 111, 252, 193, 212, 111, 252, 193, 214, 111, 252, 193, 216, 111, 252, 193, 218, 111, 252, 193, 220, 111, 252, 193, 144, 111, 252, 193, 146, 111, 252, 193, 188, 111, 252, 193, 246, 111, 252, 193, 248, 111, 252, 193, 250, 111, 252, 193, 252, 111, 252, 193, 40, 112, 252, 193, 42, 112, 252, 193, 44, 112, 252, 193, 46, 112, 252, 193, 48, 112, 252, 193, 86, 112, 252, 193, 88, 112, 252, 193, 90, 112, 252, 193, 92, 112, 252, 193, 150, 112, 252, 193, 214, 112, 252, 193, 242, 112, 252, 193, 20, 113, 252, 193, 22, 113, 252, 193, 66, 113, 252, 193, 96, 113, 252, 193, 164, 113, 252, 193, 166, 113, 252, 193, 168, 113, 252, 193, 234, 113, 252, 193, 236, 113, 252, 193, 238, 113, 252, 193, 240, 113, 252, 193, 168, 114, 252, 193, 170, 114, 252, 193, 184, 111, 252, 193, 186, 111, 252, 193, 234, 111, 252, 193, 236, 111, 252, 193, 238, 111, 252, 193, 240, 111, 252, 193, 242, 111, 252, 193, 244, 111, 252, 193, 24, 112, 252, 193, 26, 112, 252, 193, 28, 112, 252, 193, 30, 112, 252, 193, 32, 112, 252, 193, 34, 112, 252, 193, 36, 112, 252, 193, 38, 112, 252, 193, 84, 112, 252, 193, 148, 112, 252, 193, 206, 112, 252, 193, 208, 112, 252, 193, 210, 112, 252, 193, 212, 112, 252, 193, 238, 112, 252, 193, 240, 112, 252, 193, 10, 113, 252, 193, 12, 113, 252, 193, 14, 113, 252, 193, 16, 113, 252, 193, 18, 113, 252, 193, 50, 113, 252, 193, 60, 113, 252, 193, 62, 113, 252, 193, 222, 111, 252, 193, 224, 111, 252, 193, 226, 111, 252, 193, 12, 112, 252, 193, 14, 112, 252, 193, 16, 112, 252, 193, 18, 112, 252, 193, 58, 112, 252, 193, 60, 112, 252, 193, 62, 112, 252, 193, 64, 112, 252, 193, 66, 112, 252, 193, 68, 112, 252, 193, 70, 112, 252, 193, 72, 112, 252, 193, 74, 112, 252, 193, 76, 112, 252, 193, 78, 112, 252, 193, 80, 112, 252, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 206, 64, 58, 0, 193, 2, 112, 252, 193, 4, 112, 252, 193, 6, 112, 252, 193, 8, 112, 252, 193, 56, 112, 252, 193, 100, 112, 252, 193, 102, 112, 252, 193, 104, 112, 252, 193, 158, 112, 252, 193, 160, 112, 252, 193, 162, 112, 252, 193, 216, 112, 252, 193, 218, 112, 252, 193, 26, 113, 252, 193, 128, 113, 252, 193, 134, 113, 252, 193, 172, 113, 252, 193, 174, 113, 252, 193, 176, 113, 252, 193, 178, 113, 252, 193, 180, 113, 252, 193, 182, 113, 252, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 193, 94, 112, 252, 193, 96, 112, 252, 193, 98, 112, 252, 193, 152, 112, 252, 193, 154, 112, 252, 193, 156, 112, 252, 193, 244, 112, 252, 193, 24, 113, 252, 193, 68, 113, 252, 193, 88, 113, 252, 193, 118, 113, 252, 193, 170, 113, 252, 193, 242, 113, 252, 193, 244, 113, 252, 193, 246, 113, 252, 193, 248, 113, 252, 193, 250, 113, 252, 193, 184, 114, 252, 193, 186, 114, 252, 193, 188, 114, 252, 193, 190, 114, 252, 193, 136, 115, 252, 193, 138, 115, 252, 193, 140, 115, 252, 193, 142, 115, 252, 193, 144, 115, 252, 193, 146, 115, 252, 193, 148, 115, 252, 193, 150, 115, 252, 193, 86, 116, 252, 193, 88, 116, 252, 193, 90, 116, 252, 193, 234, 112, 252, 193, 250, 112, 252, 193, 252, 112, 252, 193, 254, 112, 252, 193, 2, 113, 252, 193, 4, 113, 252, 193, 6, 113, 252, 193, 28, 113, 252, 193, 30, 113, 252, 193, 32, 113, 252, 193, 34, 113, 252, 193, 36, 113, 252, 193, 38, 113, 252, 193, 40, 113, 252, 193, 42, 113, 252, 193, 44, 113, 252, 193, 46, 113, 252, 193, 48, 113, 252, 193, 54, 113, 252, 193, 72, 113, 252, 193, 74, 113, 252, 193, 76, 113, 252, 193, 78, 113, 252, 193, 80, 113, 252, 193, 90, 113, 252, 193, 92, 113, 252, 193, 98, 113, 252, 193, 100, 113, 252, 193, 102, 113, 252, 193, 104, 113, 252, 193, 94, 113, 252, 193, 108, 113, 252, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 193, 4, 222, 252, 193, 6, 222, 252, 193, 8, 222, 252, 193, 10, 222, 252, 193, 12, 222, 252, 193, 14, 222, 252, 193, 16, 222, 252, 193, 18, 222, 252, 193, 20, 222, 252, 193, 32, 222, 252, 193, 34, 222, 252, 193, 36, 222, 252, 193, 38, 222, 252, 193, 40, 222, 252, 193, 46, 222, 252, 193, 48, 222, 252, 193, 50, 222, 252, 193, 56, 113, 252, 193, 52, 222, 252, 193, 54, 222, 252, 193, 56, 222, 252, 193, 64, 222, 252, 193, 66, 222, 252, 193, 68, 222, 252, 193, 64, 113, 252, 193, 84, 113, 252, 193, 86, 113, 252, 193, 112, 113, 252, 193, 114, 113, 252, 193, 232, 113, 252, 193, 164, 114, 252, 193, 166, 114, 252, 193, 110, 115, 252, 193, 112, 115, 252, 193, 114, 115, 252, 193, 116, 115, 252, 193, 118, 115, 252, 193, 120, 115, 252, 193, 122, 115, 252, 193, 124, 115, 252, 193, 58, 116, 252, 193, 60, 116, 252, 193, 62, 116, 252, 193, 64, 116, 252, 193, 66, 116, 252, 193, 68, 116, 252, 193, 44, 117, 252, 193, 46, 117, 252, 193, 20, 118, 252, 193, 22, 118, 252, 193, 24, 118, 252, 193, 26, 118, 252, 193, 18, 119, 252, 193, 20, 119, 252, 193, 224, 119, 252, 193, 226, 119, 252, 193, 110, 113, 252, 193, 116, 113, 252, 193, 124, 113, 252, 193, 126, 113, 252, 193, 130, 113, 252, 193, 132, 113, 252, 193, 140, 113, 252, 193, 142, 113, 252, 193, 144, 113, 252, 193, 146, 113, 252, 193, 148, 113, 252, 193, 150, 113, 252, 193, 152, 113, 252, 193, 154, 113, 252, 193, 156, 113, 252, 193, 158, 113, 252, 193, 160, 113, 252, 193, 162, 113, 252, 193, 194, 113, 252, 193, 196, 113, 252, 193, 198, 113, 252, 193, 200, 113, 252, 193, 202, 113, 252, 193, 204, 113, 252, 193, 206, 113, 252, 193, 208, 113, 252, 193, 210, 113, 252, 193, 212, 113, 252, 193, 214, 113, 252, 193, 216, 113, 252, 193, 218, 113, 252, 193, 220, 113, 252, 193, 222, 113, 252, 193, 224, 113, 252, 193, 226, 113, 252, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 193, 230, 113, 252, 193, 158, 114, 252, 193, 160, 114, 252, 193, 162, 114, 252, 193, 100, 115, 252, 193, 102, 115, 252, 193, 104, 115, 252, 193, 106, 115, 252, 193, 50, 116, 252, 193, 52, 116, 252, 193, 54, 116, 252, 193, 40, 117, 252, 193, 42, 117, 252, 193, 12, 118, 252, 193, 14, 118, 252, 193, 16, 118, 252, 193, 18, 118, 252, 193, 10, 119, 252, 193, 12, 119, 252, 193, 14, 119, 252, 193, 16, 119, 252, 193, 216, 119, 252, 193, 218, 119, 252, 193, 220, 119, 252, 193, 222, 119, 252, 193, 162, 120, 252, 193, 164, 120, 252, 193, 166, 120, 252, 193, 94, 121, 252, 193, 96, 121, 252, 193, 98, 121, 252, 193, 100, 121, 252, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 96, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 128, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 193, 172, 114, 252, 193, 174, 114, 252, 193, 176, 114, 252, 193, 178, 114, 252, 193, 180, 114, 252, 193, 182, 114, 252, 193, 126, 115, 252, 193, 128, 115, 252, 193, 130, 115, 252, 193, 132, 115, 252, 193, 134, 115, 252, 193, 70, 116, 252, 193, 72, 116, 252, 193, 74, 116, 252, 193, 76, 116, 252, 193, 78, 116, 252, 193, 80, 116, 252, 193, 82, 116, 252, 193, 84, 116, 252, 193, 48, 117, 252, 193, 50, 117, 252, 193, 52, 117, 252, 193, 54, 117, 252, 193, 56, 117, 252, 193, 58, 117, 252, 193, 28, 118, 252, 193, 30, 118, 252, 193, 32, 118, 252, 193, 34, 118, 252, 193, 36, 118, 252, 193, 38, 118, 252, 193, 40, 118, 252, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 206, 160, 58, 0, 193, 152, 115, 252, 193, 154, 115, 252, 193, 156, 115, 252, 193, 158, 115, 252, 193, 160, 115, 252, 193, 162, 115, 252, 193, 164, 115, 252, 193, 166, 115, 252, 193, 168, 115, 252, 193, 170, 115, 252, 193, 172, 115, 252, 193, 174, 115, 252, 193, 176, 115, 252, 193, 178, 115, 252, 193, 180, 115, 252, 193, 182, 115, 252, 193, 184, 115, 252, 193, 186, 115, 252, 193, 254, 114, 252, 193, 4, 115, 252, 193, 6, 115, 252, 193, 8, 115, 252, 193, 10, 115, 252, 193, 12, 115, 252, 193, 14, 115, 252, 193, 16, 115, 252, 193, 190, 115, 252, 193, 192, 115, 252, 193, 194, 115, 252, 193, 196, 115, 252, 193, 198, 115, 252, 193, 200, 115, 252, 193, 202, 115, 252, 193, 204, 115, 252, 193, 132, 116, 252, 193, 134, 116, 252, 193, 136, 116, 252, 193, 138, 116, 252, 193, 140, 116, 252, 193, 142, 116, 252, 193, 144, 116, 252, 193, 146, 116, 252, 193, 148, 116, 252, 193, 150, 116, 252, 193, 152, 116, 252, 193, 154, 116, 252, 193, 156, 116, 252, 193, 158, 116, 252, 193, 160, 116, 252, 193, 162, 116, 252, 193, 164, 116, 252, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 192, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 193, 96, 116, 252, 193, 188, 115, 252, 193, 98, 116, 252, 193, 100, 116, 252, 193, 102, 116, 252, 193, 104, 116, 252, 193, 106, 116, 252, 193, 108, 116, 252, 193, 110, 116, 252, 193, 112, 116, 252, 193, 114, 116, 252, 193, 116, 116, 252, 193, 118, 116, 252, 193, 120, 116, 252, 193, 122, 116, 252, 193, 124, 116, 252, 193, 126, 116, 252, 193, 128, 116, 252, 193, 76, 117, 252, 193, 78, 117, 252, 193, 80, 117, 252, 193, 82, 117, 252, 193, 84, 117, 252, 193, 86, 117, 252, 193, 88, 117, 252, 193, 90, 117, 252, 193, 92, 117, 252, 193, 94, 117, 252, 193, 96, 117, 252, 193, 98, 117, 252, 193, 100, 117, 252, 193, 102, 117, 252, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 224, 58, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 193, 92, 116, 252, 193, 94, 116, 252, 193, 60, 117, 252, 193, 62, 117, 252, 193, 64, 117, 252, 193, 66, 117, 252, 193, 68, 117, 252, 193, 70, 117, 252, 193, 72, 117, 252, 193, 74, 117, 252, 193, 48, 118, 252, 193, 50, 118, 252, 193, 52, 118, 252, 193, 54, 118, 252, 193, 56, 118, 252, 193, 58, 118, 252, 193, 60, 118, 252, 193, 62, 118, 252, 193, 64, 118, 252, 193, 66, 118, 252, 193, 68, 118, 252, 193, 48, 119, 252, 193, 50, 119, 252, 193, 52, 119, 252, 193, 246, 119, 252, 193, 248, 119, 252, 193, 250, 119, 252, 193, 252, 119, 252, 193, 254, 119, 252, 193, 2, 120, 252, 193, 192, 120, 252, 193, 194, 120, 252, 193, 166, 116, 252, 193, 130, 117, 252, 193, 132, 117, 252, 193, 134, 117, 252, 193, 136, 117, 252, 193, 138, 117, 252, 193, 140, 117, 252, 193, 142, 117, 252, 193, 144, 117, 252, 193, 146, 117, 252, 193, 122, 118, 252, 193, 124, 118, 252, 193, 126, 118, 252, 193, 128, 118, 252, 193, 130, 118, 252, 193, 132, 118, 252, 193, 134, 118, 252, 193, 136, 118, 252, 193, 138, 118, 252, 193, 140, 118, 252, 193, 142, 118, 252, 193, 144, 118, 252, 193, 146, 118, 252, 193, 80, 119, 252, 193, 82, 119, 252, 193, 84, 119, 252, 193, 86, 119, 252, 193, 88, 119, 252, 193, 90, 119, 252, 193, 92, 119, 252, 193, 94, 119, 252, 193, 96, 119, 252, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 206, 0, 59, 0, 193, 154, 117, 252, 193, 32, 117, 252, 193, 34, 117, 252, 193, 36, 117, 252, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 206, 32, 59, 0, 193, 104, 117, 252, 193, 70, 118, 252, 193, 106, 117, 252, 193, 108, 117, 252, 193, 110, 117, 252, 193, 112, 117, 252, 193, 114, 117, 252, 193, 116, 117, 252, 193, 118, 117, 252, 193, 120, 117, 252, 193, 122, 117, 252, 193, 124, 117, 252, 193, 126, 117, 252, 193, 128, 117, 252, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 193, 8, 118, 252, 193, 10, 118, 252, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 193, 42, 118, 252, 193, 44, 118, 252, 193, 46, 118, 252, 193, 22, 119, 252, 193, 24, 119, 252, 193, 26, 119, 252, 193, 28, 119, 252, 193, 30, 119, 252, 193, 32, 119, 252, 193, 34, 119, 252, 193, 36, 119, 252, 193, 38, 119, 252, 193, 40, 119, 252, 193, 42, 119, 252, 193, 44, 119, 252, 193, 46, 119, 252, 193, 236, 119, 252, 193, 238, 119, 252, 193, 240, 119, 252, 193, 242, 119, 252, 193, 244, 119, 252, 193, 176, 120, 252, 193, 178, 120, 252, 193, 180, 120, 252, 193, 182, 120, 252, 193, 184, 120, 252, 193, 186, 120, 252, 193, 188, 120, 252, 193, 190, 120, 252, 193, 108, 121, 252, 193, 110, 121, 252, 193, 112, 121, 252, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 206, 64, 59, 0, 193, 54, 119, 252, 193, 56, 119, 252, 193, 58, 119, 252, 193, 60, 119, 252, 193, 62, 119, 252, 193, 64, 119, 252, 193, 66, 119, 252, 193, 68, 119, 252, 193, 70, 119, 252, 193, 72, 119, 252, 193, 74, 119, 252, 193, 76, 119, 252, 193, 78, 119, 252, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 96, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 193, 98, 119, 252, 193, 100, 119, 252, 193, 36, 120, 252, 193, 38, 120, 252, 193, 40, 120, 252, 193, 42, 120, 252, 193, 44, 120, 252, 193, 46, 120, 252, 193, 48, 120, 252, 193, 50, 120, 252, 193, 52, 120, 252, 193, 240, 120, 252, 193, 242, 120, 252, 193, 244, 120, 252, 193, 246, 120, 252, 193, 248, 120, 252, 193, 250, 120, 252, 193, 252, 120, 252, 193, 254, 120, 252, 193, 2, 121, 252, 193, 144, 121, 252, 193, 146, 121, 252, 193, 148, 121, 252, 193, 150, 121, 252, 193, 152, 121, 252, 193, 154, 121, 252, 193, 156, 121, 252, 193, 98, 122, 252, 193, 100, 122, 252, 193, 102, 122, 252, 193, 104, 122, 252, 193, 156, 122, 252, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 128, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 206, 192, 59, 0, 193, 214, 119, 252, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 193, 228, 119, 252, 193, 230, 119, 252, 193, 232, 119, 252, 193, 234, 119, 252, 193, 168, 120, 252, 193, 170, 120, 252, 193, 172, 120, 252, 193, 174, 120, 252, 193, 104, 121, 252, 193, 106, 121, 252, 193, 72, 122, 252, 193, 192, 122, 252, 193, 24, 123, 252, 193, 68, 123, 252, 193, 150, 123, 252, 193, 180, 123, 252, 193, 92, 124, 252, 193, 94, 124, 252, 193, 184, 124, 252, 193, 186, 124, 252, 193, 232, 124, 252, 193, 86, 125, 252, 193, 98, 125, 252, 193, 112, 125, 252, 193, 122, 125, 252, 193, 162, 125, 252, 193, 214, 125, 252, 193, 216, 125, 252, 193, 218, 125, 252, 193, 102, 126, 252, 193, 104, 126, 252, 193, 106, 126, 252, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 206, 160, 59, 0, 193, 200, 120, 252, 193, 202, 120, 252, 193, 204, 120, 252, 193, 206, 120, 252, 193, 208, 120, 252, 193, 210, 120, 252, 193, 212, 120, 252, 193, 214, 120, 252, 193, 216, 120, 252, 193, 218, 120, 252, 193, 220, 120, 252, 193, 222, 120, 252, 193, 224, 120, 252, 193, 226, 120, 252, 193, 228, 120, 252, 193, 230, 120, 252, 193, 232, 120, 252, 193, 234, 120, 252, 193, 236, 120, 252, 193, 238, 120, 252, 193, 122, 121, 252, 193, 124, 121, 252, 193, 126, 121, 252, 193, 128, 121, 252, 193, 130, 121, 252, 193, 132, 121, 252, 193, 134, 121, 252, 193, 136, 121, 252, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 224, 59, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 206, 0, 60, 0, 193, 160, 120, 252, 193, 92, 121, 252, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 193, 196, 120, 252, 193, 198, 120, 252, 193, 118, 121, 252, 193, 120, 121, 252, 193, 250, 121, 252, 193, 252, 121, 252, 193, 254, 121, 252, 193, 84, 122, 252, 193, 86, 122, 252, 193, 146, 122, 252, 193, 202, 122, 252, 193, 232, 122, 252, 193, 52, 123, 252, 193, 62, 123, 252, 193, 80, 123, 252, 193, 100, 123, 252, 193, 114, 123, 252, 193, 152, 123, 252, 193, 18, 124, 252, 193, 28, 124, 252, 193, 64, 124, 252, 193, 66, 124, 252, 193, 98, 124, 252, 193, 188, 124, 252, 193, 238, 124, 252, 193, 240, 124, 252, 193, 10, 125, 252, 193, 40, 125, 252, 193, 42, 125, 252, 193, 44, 125, 252, 193, 58, 125, 252, 193, 60, 125, 252, 193, 102, 121, 252, 193, 70, 122, 252, 193, 144, 122, 252, 193, 188, 122, 252, 193, 190, 122, 252, 193, 48, 123, 252, 193, 78, 123, 252, 193, 136, 123, 252, 193, 56, 124, 252, 193, 58, 124, 252, 193, 150, 124, 252, 193, 230, 124, 252, 193, 6, 125, 252, 193, 30, 125, 252, 193, 32, 125, 252, 193, 116, 125, 252, 193, 120, 125, 252, 193, 96, 126, 252, 193, 98, 126, 252, 193, 100, 126, 252, 193, 66, 127, 252, 193, 68, 127, 252, 193, 70, 127, 252, 193, 72, 127, 252, 193, 74, 127, 252, 193, 36, 128, 252, 193, 38, 128, 252, 193, 40, 128, 252, 193, 42, 128, 252, 193, 44, 128, 252, 193, 214, 128, 252, 193, 216, 128, 252, 193, 114, 121, 252, 193, 116, 121, 252, 193, 236, 121, 252, 193, 238, 121, 252, 193, 240, 121, 252, 193, 242, 121, 252, 193, 244, 121, 252, 193, 246, 121, 252, 193, 248, 121, 252, 193, 74, 122, 252, 193, 76, 122, 252, 193, 78, 122, 252, 193, 80, 122, 252, 193, 82, 122, 252, 193, 194, 122, 252, 193, 196, 122, 252, 193, 198, 122, 252, 193, 200, 122, 252, 193, 230, 122, 252, 193, 12, 123, 252, 193, 14, 123, 252, 193, 56, 123, 252, 193, 60, 123, 252, 193, 96, 123, 252, 193, 98, 123, 252, 193, 110, 123, 252, 193, 112, 123, 252, 193, 198, 123, 252, 193, 208, 123, 252, 193, 26, 124, 252, 193, 60, 124, 252, 193, 62, 124, 252, 193, 138, 121, 252, 193, 140, 121, 252, 193, 142, 121, 252, 193, 2, 122, 252, 193, 4, 122, 252, 193, 6, 122, 252, 193, 8, 122, 252, 193, 10, 122, 252, 193, 12, 122, 252, 193, 14, 122, 252, 193, 16, 122, 252, 193, 88, 122, 252, 193, 90, 122, 252, 193, 92, 122, 252, 193, 94, 122, 252, 193, 96, 122, 252, 193, 148, 122, 252, 193, 150, 122, 252, 193, 152, 122, 252, 193, 154, 122, 252, 193, 234, 122, 252, 193, 254, 122, 252, 193, 26, 123, 252, 193, 2, 123, 252, 193, 16, 123, 252, 193, 18, 123, 252, 193, 28, 123, 252, 193, 54, 123, 252, 193, 70, 123, 252, 193, 82, 123, 252, 193, 116, 123, 252, 193, 118, 123, 252, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 206, 32, 60, 0, 193, 18, 122, 252, 193, 20, 122, 252, 193, 22, 122, 252, 193, 24, 122, 252, 193, 26, 122, 252, 193, 28, 122, 252, 193, 30, 122, 252, 193, 32, 122, 252, 193, 34, 122, 252, 193, 36, 122, 252, 193, 38, 122, 252, 193, 40, 122, 252, 193, 42, 122, 252, 193, 44, 122, 252, 193, 46, 122, 252, 193, 48, 122, 252, 193, 50, 122, 252, 193, 52, 122, 252, 193, 54, 122, 252, 193, 56, 122, 252, 193, 58, 122, 252, 193, 60, 122, 252, 193, 48, 144, 253, 193, 62, 122, 252, 193, 64, 122, 252, 193, 66, 122, 252, 193, 68, 122, 252, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 206, 64, 60, 0, 193, 160, 122, 252, 193, 162, 122, 252, 193, 164, 122, 252, 193, 166, 122, 252, 193, 168, 122, 252, 193, 170, 122, 252, 193, 172, 122, 252, 193, 174, 122, 252, 193, 176, 122, 252, 193, 178, 122, 252, 193, 180, 122, 252, 193, 182, 122, 252, 193, 184, 122, 252, 193, 206, 122, 252, 193, 208, 122, 252, 193, 210, 122, 252, 193, 212, 122, 252, 193, 214, 122, 252, 193, 216, 122, 252, 193, 158, 122, 252, 193, 204, 122, 252, 193, 236, 122, 252, 193, 238, 122, 252, 193, 20, 123, 252, 193, 50, 123, 252, 193, 64, 123, 252, 193, 72, 123, 252, 193, 102, 123, 252, 193, 154, 123, 252, 193, 210, 123, 252, 193, 212, 123, 252, 193, 214, 123, 252, 193, 246, 123, 252, 193, 8, 124, 252, 193, 70, 124, 252, 193, 72, 124, 252, 193, 104, 124, 252, 193, 106, 124, 252, 193, 160, 124, 252, 193, 192, 124, 252, 193, 242, 124, 252, 193, 136, 125, 252, 193, 148, 125, 252, 193, 186, 125, 252, 193, 188, 125, 252, 193, 28, 126, 252, 193, 30, 126, 252, 193, 32, 126, 252, 193, 34, 126, 252, 193, 36, 126, 252, 193, 210, 126, 252, 193, 218, 122, 252, 193, 220, 122, 252, 193, 186, 122, 252, 193, 222, 122, 252, 193, 88, 27, 253, 193, 224, 122, 252, 193, 226, 122, 252, 193, 228, 122, 252, 193, 240, 122, 252, 193, 242, 122, 252, 193, 244, 122, 252, 193, 246, 122, 252, 193, 248, 122, 252, 193, 250, 122, 252, 193, 252, 122, 252, 193, 4, 123, 252, 193, 6, 123, 252, 193, 8, 123, 252, 193, 10, 123, 252, 193, 22, 123, 252, 193, 30, 123, 252, 193, 32, 123, 252, 193, 34, 123, 252, 193, 36, 123, 252, 193, 38, 123, 252, 193, 40, 123, 252, 193, 42, 123, 252, 193, 44, 123, 252, 193, 46, 123, 252, 193, 58, 123, 252, 193, 66, 123, 252, 193, 74, 123, 252, 193, 76, 123, 252, 193, 84, 123, 252, 193, 86, 123, 252, 193, 88, 123, 252, 193, 90, 123, 252, 193, 92, 123, 252, 193, 94, 123, 252, 193, 104, 123, 252, 193, 106, 123, 252, 193, 108, 123, 252, 193, 120, 123, 252, 193, 122, 123, 252, 193, 124, 123, 252, 193, 126, 123, 252, 193, 128, 123, 252, 193, 130, 123, 252, 193, 132, 123, 252, 193, 134, 123, 252, 193, 142, 123, 252, 193, 144, 123, 252, 193, 146, 123, 252, 193, 148, 123, 252, 193, 156, 123, 252, 193, 158, 123, 252, 193, 160, 123, 252, 193, 162, 123, 252, 193, 164, 123, 252, 193, 166, 123, 252, 193, 168, 123, 252, 193, 170, 123, 252, 193, 172, 123, 252, 193, 174, 123, 252, 193, 138, 123, 252, 193, 10, 124, 252, 193, 16, 124, 252, 193, 20, 124, 252, 193, 30, 124, 252, 193, 32, 124, 252, 193, 36, 124, 252, 193, 38, 124, 252, 193, 40, 124, 252, 193, 68, 124, 252, 193, 100, 124, 252, 193, 102, 124, 252, 193, 156, 124, 252, 193, 158, 124, 252, 193, 190, 124, 252, 193, 68, 125, 252, 193, 70, 125, 252, 193, 102, 125, 252, 193, 114, 125, 252, 193, 134, 125, 252, 193, 142, 125, 252, 193, 144, 125, 252, 193, 146, 125, 252, 193, 164, 125, 252, 193, 166, 125, 252, 193, 168, 125, 252, 193, 170, 125, 252, 193, 172, 125, 252, 193, 174, 125, 252, 193, 176, 125, 252, 193, 178, 125, 252, 193, 180, 125, 252, 193, 176, 123, 252, 193, 178, 123, 252, 193, 182, 123, 252, 193, 184, 123, 252, 193, 186, 123, 252, 193, 188, 123, 252, 193, 190, 123, 252, 193, 192, 123, 252, 193, 194, 123, 252, 193, 196, 123, 252, 193, 200, 123, 252, 193, 202, 123, 252, 193, 204, 123, 252, 193, 206, 123, 252, 193, 224, 123, 252, 193, 216, 123, 252, 193, 218, 123, 252, 193, 220, 123, 252, 193, 222, 123, 252, 193, 226, 123, 252, 193, 228, 123, 252, 193, 230, 123, 252, 193, 232, 123, 252, 193, 234, 123, 252, 193, 236, 123, 252, 193, 238, 123, 252, 193, 240, 123, 252, 193, 242, 123, 252, 193, 244, 123, 252, 193, 254, 123, 252, 193, 248, 123, 252, 193, 250, 123, 252, 193, 252, 123, 252, 193, 2, 124, 252, 193, 4, 124, 252, 193, 6, 124, 252, 193, 12, 124, 252, 193, 14, 124, 252, 193, 22, 124, 252, 193, 24, 124, 252, 193, 34, 124, 252, 193, 42, 124, 252, 193, 44, 124, 252, 193, 46, 124, 252, 193, 48, 124, 252, 193, 50, 124, 252, 193, 52, 124, 252, 193, 54, 124, 252, 193, 74, 124, 252, 193, 76, 124, 252, 193, 78, 124, 252, 193, 80, 124, 252, 193, 82, 124, 252, 193, 84, 124, 252, 193, 86, 124, 252, 193, 88, 124, 252, 193, 90, 124, 252, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 193, 96, 124, 252, 193, 152, 124, 252, 193, 154, 124, 252, 193, 234, 124, 252, 193, 236, 124, 252, 193, 8, 125, 252, 193, 34, 125, 252, 193, 36, 125, 252, 193, 38, 125, 252, 193, 56, 125, 252, 193, 118, 125, 252, 193, 220, 125, 252, 193, 222, 125, 252, 193, 114, 126, 252, 193, 116, 126, 252, 193, 88, 127, 252, 193, 90, 127, 252, 193, 92, 127, 252, 193, 94, 127, 252, 193, 96, 127, 252, 193, 98, 127, 252, 193, 54, 128, 252, 193, 56, 128, 252, 193, 58, 128, 252, 193, 60, 128, 252, 193, 236, 128, 252, 193, 238, 128, 252, 193, 240, 128, 252, 193, 242, 128, 252, 193, 244, 128, 252, 193, 246, 128, 252, 193, 248, 128, 252, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 206, 96, 60, 0, 193, 164, 124, 252, 193, 166, 124, 252, 193, 168, 124, 252, 193, 170, 124, 252, 193, 172, 124, 252, 193, 174, 124, 252, 193, 176, 124, 252, 193, 178, 124, 252, 193, 180, 124, 252, 193, 182, 124, 252, 206, 128, 60, 0, 206, 128, 60, 0, 206, 128, 60, 0, 206, 128, 60, 0, 206, 128, 60, 0, 206, 128, 60, 0, 206, 128, 60, 0, 206, 128, 60, 0, 206, 128, 60, 0, 206, 128, 60, 0, 193, 244, 124, 252, 193, 246, 124, 252, 193, 248, 124, 252, 193, 250, 124, 252, 193, 252, 124, 252, 193, 254, 124, 252, 193, 2, 125, 252, 193, 4, 125, 252, 193, 12, 125, 252, 193, 14, 125, 252, 193, 16, 125, 252, 193, 18, 125, 252, 193, 20, 125, 252, 193, 22, 125, 252, 193, 24, 125, 252, 193, 26, 125, 252, 193, 28, 125, 252, 193, 46, 125, 252, 193, 48, 125, 252, 193, 50, 125, 252, 193, 52, 125, 252, 193, 54, 125, 252, 193, 62, 125, 252, 193, 64, 125, 252, 193, 66, 125, 252, 193, 100, 125, 252, 193, 224, 125, 252, 193, 118, 126, 252, 193, 120, 126, 252, 193, 122, 126, 252, 193, 124, 126, 252, 193, 126, 126, 252, 193, 128, 126, 252, 193, 130, 126, 252, 193, 132, 126, 252, 193, 100, 127, 252, 193, 102, 127, 252, 193, 104, 127, 252, 193, 106, 127, 252, 193, 108, 127, 252, 193, 110, 127, 252, 193, 62, 128, 252, 193, 64, 128, 252, 193, 66, 128, 252, 193, 68, 128, 252, 193, 70, 128, 252, 193, 72, 128, 252, 193, 74, 128, 252, 193, 76, 128, 252, 193, 78, 128, 252, 206, 64, 61, 0, 206, 64, 61, 0, 206, 64, 61, 0, 206, 64, 61, 0, 193, 72, 125, 252, 193, 74, 125, 252, 193, 76, 125, 252, 193, 78, 125, 252, 193, 80, 125, 252, 193, 82, 125, 252, 193, 84, 125, 252, 193, 88, 125, 252, 193, 90, 125, 252, 193, 92, 125, 252, 193, 94, 125, 252, 193, 96, 125, 252, 193, 104, 125, 252, 193, 106, 125, 252, 193, 108, 125, 252, 193, 110, 125, 252, 193, 128, 125, 252, 193, 124, 125, 252, 193, 126, 125, 252, 193, 130, 125, 252, 193, 132, 125, 252, 193, 138, 125, 252, 193, 140, 125, 252, 193, 150, 125, 252, 193, 152, 125, 252, 193, 154, 125, 252, 193, 156, 125, 252, 193, 158, 125, 252, 193, 160, 125, 252, 193, 190, 125, 252, 193, 192, 125, 252, 193, 194, 125, 252, 193, 182, 125, 252, 193, 184, 125, 252, 193, 228, 125, 252, 193, 230, 125, 252, 193, 232, 125, 252, 193, 234, 125, 252, 193, 236, 125, 252, 193, 238, 125, 252, 193, 240, 125, 252, 193, 242, 125, 252, 193, 244, 125, 252, 193, 246, 125, 252, 193, 248, 125, 252, 193, 250, 125, 252, 193, 252, 125, 252, 193, 254, 125, 252, 193, 2, 126, 252, 193, 4, 126, 252, 193, 6, 126, 252, 193, 8, 126, 252, 193, 10, 126, 252, 193, 12, 126, 252, 193, 14, 126, 252, 193, 16, 126, 252, 193, 18, 126, 252, 193, 20, 126, 252, 193, 22, 126, 252, 193, 24, 126, 252, 193, 26, 126, 252, 193, 134, 126, 252, 193, 136, 126, 252, 193, 138, 126, 252, 193, 196, 125, 252, 193, 198, 125, 252, 193, 200, 125, 252, 193, 202, 125, 252, 193, 204, 125, 252, 193, 206, 125, 252, 193, 208, 125, 252, 193, 210, 125, 252, 193, 212, 125, 252, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 160, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 193, 108, 126, 252, 193, 110, 126, 252, 193, 112, 126, 252, 193, 76, 127, 252, 193, 78, 127, 252, 193, 80, 127, 252, 193, 82, 127, 252, 193, 84, 127, 252, 193, 86, 127, 252, 193, 46, 128, 252, 193, 48, 128, 252, 193, 50, 128, 252, 193, 52, 128, 252, 193, 226, 128, 252, 193, 228, 128, 252, 193, 230, 128, 252, 193, 232, 128, 252, 193, 234, 128, 252, 193, 236, 129, 252, 193, 238, 129, 252, 193, 240, 129, 252, 193, 242, 129, 252, 193, 204, 130, 252, 193, 206, 130, 252, 193, 208, 130, 252, 193, 210, 130, 252, 193, 212, 130, 252, 193, 140, 131, 252, 193, 52, 132, 252, 193, 54, 132, 252, 193, 56, 132, 252, 193, 240, 132, 252, 193, 140, 126, 252, 193, 142, 126, 252, 193, 144, 126, 252, 193, 146, 126, 252, 193, 148, 126, 252, 193, 150, 126, 252, 193, 152, 126, 252, 193, 154, 126, 252, 193, 156, 126, 252, 193, 158, 126, 252, 193, 160, 126, 252, 193, 162, 126, 252, 193, 164, 126, 252, 193, 166, 126, 252, 193, 168, 126, 252, 193, 170, 126, 252, 193, 172, 126, 252, 193, 174, 126, 252, 193, 176, 126, 252, 193, 178, 126, 252, 193, 180, 126, 252, 193, 182, 126, 252, 193, 184, 126, 252, 193, 186, 126, 252, 193, 188, 126, 252, 193, 190, 126, 252, 193, 192, 126, 252, 193, 194, 126, 252, 193, 196, 126, 252, 193, 198, 126, 252, 193, 200, 126, 252, 193, 202, 126, 252, 193, 204, 126, 252, 193, 206, 126, 252, 193, 112, 127, 252, 193, 114, 127, 252, 193, 116, 127, 252, 193, 118, 127, 252, 193, 120, 127, 252, 193, 122, 127, 252, 193, 124, 127, 252, 193, 126, 127, 252, 193, 128, 127, 252, 193, 130, 127, 252, 193, 132, 127, 252, 193, 134, 127, 252, 193, 136, 127, 252, 193, 138, 127, 252, 193, 140, 127, 252, 193, 142, 127, 252, 193, 144, 127, 252, 193, 146, 127, 252, 193, 148, 127, 252, 193, 150, 127, 252, 193, 152, 127, 252, 193, 154, 127, 252, 193, 156, 127, 252, 193, 158, 127, 252, 193, 160, 127, 252, 193, 162, 127, 252, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 193, 212, 126, 252, 193, 214, 126, 252, 193, 216, 126, 252, 193, 218, 126, 252, 193, 220, 126, 252, 193, 222, 126, 252, 193, 224, 126, 252, 193, 226, 126, 252, 193, 228, 126, 252, 193, 166, 127, 252, 193, 168, 127, 252, 193, 170, 127, 252, 193, 172, 127, 252, 193, 174, 127, 252, 193, 176, 127, 252, 193, 178, 127, 252, 193, 136, 128, 252, 193, 138, 128, 252, 193, 140, 128, 252, 193, 142, 128, 252, 193, 144, 128, 252, 193, 146, 128, 252, 193, 148, 128, 252, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 192, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 224, 60, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 206, 0, 61, 0, 193, 20, 129, 252, 193, 22, 129, 252, 193, 24, 129, 252, 193, 26, 129, 252, 193, 28, 129, 252, 193, 30, 129, 252, 193, 20, 130, 252, 193, 32, 129, 252, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 32, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 193, 218, 128, 252, 193, 220, 128, 252, 193, 222, 128, 252, 193, 224, 128, 252, 193, 224, 129, 252, 193, 226, 129, 252, 193, 228, 129, 252, 193, 230, 129, 252, 193, 232, 129, 252, 193, 234, 129, 252, 193, 200, 130, 252, 193, 202, 130, 252, 193, 134, 131, 252, 193, 136, 131, 252, 193, 138, 131, 252, 193, 40, 132, 252, 193, 42, 132, 252, 193, 44, 132, 252, 193, 46, 132, 252, 193, 48, 132, 252, 193, 50, 132, 252, 193, 234, 132, 252, 193, 236, 132, 252, 193, 238, 132, 252, 193, 112, 133, 252, 193, 114, 133, 252, 193, 116, 133, 252, 193, 192, 133, 252, 193, 194, 133, 252, 193, 196, 133, 252, 193, 86, 134, 252, 193, 154, 134, 252, 193, 250, 128, 252, 193, 252, 128, 252, 193, 244, 129, 252, 193, 246, 129, 252, 193, 248, 129, 252, 193, 250, 129, 252, 193, 252, 129, 252, 193, 254, 129, 252, 193, 2, 130, 252, 193, 4, 130, 252, 193, 6, 130, 252, 193, 8, 130, 252, 193, 214, 130, 252, 193, 216, 130, 252, 193, 218, 130, 252, 193, 220, 130, 252, 193, 222, 130, 252, 193, 224, 130, 252, 193, 226, 130, 252, 193, 228, 130, 252, 193, 230, 130, 252, 193, 232, 130, 252, 193, 142, 131, 252, 193, 144, 131, 252, 193, 146, 131, 252, 193, 148, 131, 252, 193, 150, 131, 252, 206, 0, 62, 0, 206, 0, 62, 0, 206, 0, 62, 0, 206, 0, 62, 0, 206, 0, 62, 0, 206, 64, 61, 0, 206, 64, 61, 0, 206, 64, 61, 0, 206, 64, 61, 0, 206, 64, 61, 0, 206, 64, 61, 0, 193, 10, 130, 252, 193, 12, 130, 252, 193, 14, 130, 252, 193, 16, 130, 252, 193, 18, 130, 252, 193, 234, 130, 252, 193, 236, 130, 252, 193, 238, 130, 252, 193, 240, 130, 252, 193, 242, 130, 252, 193, 244, 130, 252, 193, 152, 131, 252, 193, 154, 131, 252, 193, 156, 131, 252, 193, 158, 131, 252, 193, 160, 131, 252, 193, 162, 131, 252, 193, 164, 131, 252, 193, 166, 131, 252, 193, 76, 132, 252, 193, 78, 132, 252, 193, 80, 132, 252, 193, 82, 132, 252, 193, 84, 132, 252, 193, 86, 132, 252, 193, 88, 132, 252, 193, 34, 129, 252, 193, 36, 129, 252, 193, 38, 129, 252, 193, 40, 129, 252, 193, 42, 129, 252, 193, 44, 129, 252, 193, 46, 129, 252, 193, 48, 129, 252, 193, 50, 129, 252, 193, 52, 129, 252, 193, 54, 129, 252, 193, 56, 129, 252, 193, 58, 129, 252, 193, 60, 129, 252, 193, 62, 129, 252, 193, 64, 129, 252, 193, 66, 129, 252, 193, 68, 129, 252, 193, 70, 129, 252, 193, 72, 129, 252, 193, 74, 129, 252, 193, 76, 129, 252, 193, 78, 129, 252, 193, 80, 129, 252, 193, 82, 129, 252, 193, 84, 129, 252, 193, 86, 129, 252, 193, 88, 129, 252, 193, 90, 129, 252, 193, 22, 130, 252, 193, 24, 130, 252, 193, 26, 130, 252, 193, 28, 130, 252, 193, 30, 130, 252, 193, 32, 130, 252, 193, 34, 130, 252, 193, 36, 130, 252, 193, 38, 130, 252, 193, 40, 130, 252, 193, 42, 130, 252, 193, 44, 130, 252, 193, 46, 130, 252, 193, 48, 130, 252, 193, 50, 130, 252, 193, 52, 130, 252, 193, 54, 130, 252, 193, 56, 130, 252, 193, 58, 130, 252, 193, 92, 129, 252, 193, 60, 130, 252, 193, 62, 130, 252, 193, 64, 130, 252, 193, 66, 130, 252, 193, 68, 130, 252, 193, 70, 130, 252, 193, 72, 130, 252, 193, 74, 130, 252, 193, 76, 130, 252, 193, 246, 130, 252, 193, 248, 130, 252, 193, 250, 130, 252, 193, 252, 130, 252, 193, 254, 130, 252, 193, 2, 131, 252, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 206, 96, 61, 0, 193, 80, 130, 252, 193, 82, 130, 252, 193, 84, 130, 252, 193, 86, 130, 252, 193, 88, 130, 252, 193, 90, 130, 252, 193, 92, 130, 252, 193, 94, 130, 252, 193, 96, 130, 252, 193, 58, 131, 252, 193, 60, 131, 252, 193, 62, 131, 252, 193, 64, 131, 252, 193, 66, 131, 252, 193, 68, 131, 252, 193, 70, 131, 252, 193, 206, 131, 252, 193, 208, 131, 252, 193, 210, 131, 252, 193, 212, 131, 252, 193, 214, 131, 252, 193, 216, 131, 252, 193, 218, 131, 252, 193, 138, 132, 252, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 128, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 193, 4, 131, 252, 193, 78, 130, 252, 193, 6, 131, 252, 193, 8, 131, 252, 193, 10, 131, 252, 193, 12, 131, 252, 193, 14, 131, 252, 193, 16, 131, 252, 193, 18, 131, 252, 193, 20, 131, 252, 193, 22, 131, 252, 193, 24, 131, 252, 193, 26, 131, 252, 193, 28, 131, 252, 193, 30, 131, 252, 193, 32, 131, 252, 193, 34, 131, 252, 193, 36, 131, 252, 193, 38, 131, 252, 193, 40, 131, 252, 193, 42, 131, 252, 193, 44, 131, 252, 193, 46, 131, 252, 193, 48, 131, 252, 193, 50, 131, 252, 193, 52, 131, 252, 193, 168, 131, 252, 193, 170, 131, 252, 193, 54, 131, 252, 193, 172, 131, 252, 193, 174, 131, 252, 193, 176, 131, 252, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 160, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 193, 178, 131, 252, 193, 180, 131, 252, 193, 182, 131, 252, 193, 184, 131, 252, 193, 56, 131, 252, 193, 186, 131, 252, 193, 188, 131, 252, 193, 190, 131, 252, 193, 192, 131, 252, 193, 194, 131, 252, 193, 196, 131, 252, 193, 198, 131, 252, 193, 200, 131, 252, 193, 202, 131, 252, 193, 204, 131, 252, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 192, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 193, 220, 131, 252, 193, 140, 132, 252, 193, 142, 132, 252, 193, 144, 132, 252, 193, 146, 132, 252, 193, 148, 132, 252, 193, 150, 132, 252, 193, 152, 132, 252, 193, 154, 132, 252, 193, 58, 133, 252, 193, 60, 133, 252, 193, 62, 133, 252, 193, 64, 133, 252, 193, 66, 133, 252, 193, 68, 133, 252, 193, 70, 133, 252, 193, 156, 133, 252, 193, 234, 133, 252, 193, 236, 133, 252, 193, 238, 133, 252, 193, 56, 134, 252, 193, 58, 134, 252, 193, 60, 134, 252, 193, 62, 134, 252, 193, 92, 134, 252, 193, 94, 134, 252, 193, 164, 134, 252, 193, 184, 134, 252, 193, 222, 134, 252, 193, 56, 135, 252, 193, 58, 135, 252, 193, 60, 135, 252, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 224, 61, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 0, 62, 0, 206, 0, 62, 0, 206, 0, 62, 0, 206, 0, 62, 0, 193, 250, 132, 252, 193, 252, 132, 252, 193, 254, 132, 252, 193, 2, 133, 252, 193, 4, 133, 252, 193, 6, 133, 252, 193, 122, 133, 252, 193, 124, 133, 252, 193, 126, 133, 252, 193, 128, 133, 252, 193, 130, 133, 252, 193, 132, 133, 252, 193, 202, 133, 252, 193, 204, 133, 252, 193, 206, 133, 252, 193, 208, 133, 252, 193, 36, 134, 252, 193, 88, 134, 252, 193, 156, 134, 252, 193, 158, 134, 252, 193, 160, 134, 252, 193, 176, 134, 252, 193, 178, 134, 252, 193, 194, 134, 252, 193, 236, 134, 252, 193, 238, 134, 252, 193, 20, 135, 252, 193, 120, 135, 252, 193, 90, 132, 252, 193, 8, 133, 252, 193, 10, 133, 252, 193, 12, 133, 252, 193, 14, 133, 252, 193, 16, 133, 252, 193, 18, 133, 252, 193, 20, 133, 252, 193, 22, 133, 252, 193, 210, 133, 252, 193, 38, 134, 252, 193, 128, 134, 252, 193, 130, 134, 252, 193, 22, 135, 252, 193, 24, 135, 252, 193, 26, 135, 252, 193, 28, 135, 252, 193, 126, 135, 252, 193, 128, 135, 252, 193, 26, 136, 252, 193, 28, 136, 252, 193, 30, 136, 252, 193, 156, 136, 252, 193, 158, 136, 252, 193, 160, 136, 252, 193, 162, 136, 252, 193, 164, 136, 252, 193, 166, 136, 252, 193, 168, 136, 252, 193, 170, 136, 252, 193, 172, 136, 252, 193, 174, 136, 252, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 206, 32, 62, 0, 193, 24, 133, 252, 193, 26, 133, 252, 193, 28, 133, 252, 193, 30, 133, 252, 193, 32, 133, 252, 193, 34, 133, 252, 193, 36, 133, 252, 193, 38, 133, 252, 193, 40, 133, 252, 193, 42, 133, 252, 193, 44, 133, 252, 193, 46, 133, 252, 193, 48, 133, 252, 193, 50, 133, 252, 193, 52, 133, 252, 193, 54, 133, 252, 193, 56, 133, 252, 193, 134, 133, 252, 193, 136, 133, 252, 193, 138, 133, 252, 193, 140, 133, 252, 193, 142, 133, 252, 193, 144, 133, 252, 193, 146, 133, 252, 193, 148, 133, 252, 193, 150, 133, 252, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 64, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 193, 242, 132, 252, 193, 244, 132, 252, 193, 246, 132, 252, 193, 248, 132, 252, 193, 118, 133, 252, 193, 120, 133, 252, 193, 198, 133, 252, 193, 200, 133, 252, 193, 34, 134, 252, 193, 124, 134, 252, 193, 126, 134, 252, 193, 174, 134, 252, 193, 18, 135, 252, 193, 112, 135, 252, 193, 114, 135, 252, 193, 116, 135, 252, 193, 118, 135, 252, 193, 252, 135, 252, 193, 254, 135, 252, 193, 2, 136, 252, 193, 130, 136, 252, 193, 132, 136, 252, 193, 26, 137, 252, 193, 28, 137, 252, 193, 30, 137, 252, 193, 32, 137, 252, 193, 34, 137, 252, 193, 36, 137, 252, 193, 38, 137, 252, 193, 40, 137, 252, 193, 186, 137, 252, 193, 188, 137, 252, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 206, 96, 62, 0, 193, 158, 133, 252, 193, 160, 133, 252, 193, 162, 133, 252, 193, 164, 133, 252, 193, 166, 133, 252, 193, 168, 133, 252, 193, 170, 133, 252, 193, 172, 133, 252, 193, 174, 133, 252, 193, 176, 133, 252, 193, 178, 133, 252, 193, 180, 133, 252, 193, 182, 133, 252, 193, 184, 133, 252, 193, 186, 133, 252, 193, 188, 133, 252, 193, 190, 133, 252, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 193, 152, 133, 252, 193, 154, 133, 252, 193, 212, 133, 252, 193, 214, 133, 252, 193, 216, 133, 252, 193, 218, 133, 252, 193, 220, 133, 252, 193, 222, 133, 252, 193, 224, 133, 252, 193, 226, 133, 252, 193, 228, 133, 252, 193, 230, 133, 252, 193, 232, 133, 252, 193, 40, 134, 252, 193, 42, 134, 252, 193, 44, 134, 252, 193, 46, 134, 252, 193, 48, 134, 252, 193, 50, 134, 252, 193, 52, 134, 252, 193, 54, 134, 252, 193, 90, 134, 252, 193, 132, 134, 252, 193, 134, 134, 252, 193, 162, 134, 252, 193, 180, 134, 252, 193, 210, 134, 252, 193, 212, 134, 252, 193, 216, 134, 252, 193, 220, 134, 252, 193, 240, 134, 252, 193, 242, 134, 252, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 206, 128, 62, 0, 193, 64, 134, 252, 193, 66, 134, 252, 193, 68, 134, 252, 193, 70, 134, 252, 193, 72, 134, 252, 193, 74, 134, 252, 193, 76, 134, 252, 193, 78, 134, 252, 193, 80, 134, 252, 193, 82, 134, 252, 193, 84, 134, 252, 193, 96, 134, 252, 193, 98, 134, 252, 193, 100, 134, 252, 193, 102, 134, 252, 193, 104, 134, 252, 193, 106, 134, 252, 193, 108, 134, 252, 193, 110, 134, 252, 193, 112, 134, 252, 193, 114, 134, 252, 193, 116, 134, 252, 193, 118, 134, 252, 193, 120, 134, 252, 193, 122, 134, 252, 193, 136, 134, 252, 193, 138, 134, 252, 193, 140, 134, 252, 193, 142, 134, 252, 193, 144, 134, 252, 193, 146, 134, 252, 193, 148, 134, 252, 193, 150, 134, 252, 193, 152, 134, 252, 193, 166, 134, 252, 193, 168, 134, 252, 193, 196, 165, 253, 193, 170, 134, 252, 193, 172, 134, 252, 193, 182, 134, 252, 193, 186, 134, 252, 193, 188, 134, 252, 193, 190, 134, 252, 193, 192, 134, 252, 193, 196, 134, 252, 193, 198, 134, 252, 193, 200, 134, 252, 193, 202, 134, 252, 193, 204, 134, 252, 193, 206, 134, 252, 193, 208, 134, 252, 193, 214, 134, 252, 193, 218, 134, 252, 193, 224, 134, 252, 193, 226, 134, 252, 193, 228, 134, 252, 193, 230, 134, 252, 193, 232, 134, 252, 193, 234, 134, 252, 193, 252, 134, 252, 193, 254, 134, 252, 193, 2, 135, 252, 193, 4, 135, 252, 193, 6, 135, 252, 193, 8, 135, 252, 193, 10, 135, 252, 193, 12, 135, 252, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 193, 244, 134, 252, 193, 246, 134, 252, 193, 248, 134, 252, 193, 250, 134, 252, 193, 30, 135, 252, 193, 32, 135, 252, 193, 34, 135, 252, 193, 36, 135, 252, 193, 38, 135, 252, 193, 40, 135, 252, 193, 42, 135, 252, 193, 44, 135, 252, 193, 46, 135, 252, 193, 48, 135, 252, 193, 50, 135, 252, 193, 130, 135, 252, 193, 132, 135, 252, 193, 134, 135, 252, 193, 136, 135, 252, 193, 138, 135, 252, 193, 140, 135, 252, 193, 142, 135, 252, 193, 144, 135, 252, 193, 146, 135, 252, 193, 148, 135, 252, 193, 150, 135, 252, 193, 152, 135, 252, 193, 154, 135, 252, 193, 156, 135, 252, 193, 158, 135, 252, 193, 160, 135, 252, 193, 162, 135, 252, 193, 14, 135, 252, 193, 16, 135, 252, 193, 108, 135, 252, 193, 110, 135, 252, 193, 246, 135, 252, 193, 248, 135, 252, 193, 250, 135, 252, 193, 120, 136, 252, 193, 122, 136, 252, 193, 124, 136, 252, 193, 126, 136, 252, 193, 14, 137, 252, 193, 16, 137, 252, 193, 18, 137, 252, 193, 20, 137, 252, 193, 22, 137, 252, 193, 166, 137, 252, 193, 168, 137, 252, 193, 170, 137, 252, 193, 172, 137, 252, 193, 174, 137, 252, 193, 176, 137, 252, 193, 178, 137, 252, 193, 180, 137, 252, 193, 182, 137, 252, 193, 184, 137, 252, 193, 88, 138, 252, 193, 90, 138, 252, 193, 186, 138, 252, 193, 188, 138, 252, 193, 190, 138, 252, 193, 192, 138, 252, 193, 114, 27, 253, 193, 148, 27, 253, 193, 150, 27, 253, 193, 154, 27, 253, 193, 156, 27, 253, 193, 158, 27, 253, 193, 52, 135, 252, 193, 160, 27, 253, 193, 162, 27, 253, 193, 164, 27, 253, 193, 180, 27, 253, 193, 188, 27, 253, 193, 198, 27, 253, 193, 200, 27, 253, 193, 204, 27, 253, 193, 216, 27, 253, 193, 222, 27, 253, 193, 224, 27, 253, 193, 240, 27, 253, 193, 252, 27, 253, 193, 254, 27, 253, 193, 24, 28, 253, 193, 26, 28, 253, 193, 28, 28, 253, 193, 30, 28, 253, 193, 32, 28, 253, 193, 92, 28, 253, 193, 94, 28, 253, 193, 142, 28, 253, 193, 96, 28, 253, 193, 98, 28, 253, 193, 144, 28, 253, 193, 62, 135, 252, 193, 170, 135, 252, 193, 172, 135, 252, 193, 174, 135, 252, 193, 56, 136, 252, 193, 58, 136, 252, 193, 60, 136, 252, 193, 200, 136, 252, 193, 202, 136, 252, 193, 204, 136, 252, 193, 206, 136, 252, 193, 90, 137, 252, 193, 92, 137, 252, 193, 94, 137, 252, 193, 96, 137, 252, 193, 98, 137, 252, 193, 100, 137, 252, 193, 102, 137, 252, 193, 104, 137, 252, 193, 14, 138, 252, 193, 16, 138, 252, 193, 18, 138, 252, 193, 20, 138, 252, 193, 22, 138, 252, 193, 142, 138, 252, 193, 144, 138, 252, 193, 222, 138, 252, 193, 224, 138, 252, 193, 226, 138, 252, 193, 228, 138, 252, 193, 230, 138, 252, 193, 70, 139, 252, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 160, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 193, 122, 135, 252, 193, 124, 135, 252, 193, 4, 136, 252, 193, 6, 136, 252, 193, 8, 136, 252, 193, 10, 136, 252, 193, 12, 136, 252, 193, 14, 136, 252, 193, 16, 136, 252, 193, 18, 136, 252, 193, 20, 136, 252, 193, 22, 136, 252, 193, 24, 136, 252, 193, 134, 136, 252, 193, 136, 136, 252, 193, 138, 136, 252, 193, 140, 136, 252, 193, 142, 136, 252, 193, 144, 136, 252, 193, 146, 136, 252, 193, 148, 136, 252, 193, 150, 136, 252, 193, 152, 136, 252, 193, 154, 136, 252, 193, 42, 137, 252, 193, 44, 137, 252, 193, 46, 137, 252, 193, 48, 137, 252, 193, 50, 137, 252, 193, 52, 137, 252, 193, 54, 137, 252, 193, 208, 137, 252, 193, 164, 135, 252, 193, 166, 135, 252, 193, 168, 135, 252, 193, 32, 136, 252, 193, 34, 136, 252, 193, 36, 136, 252, 193, 176, 136, 252, 193, 38, 136, 252, 193, 40, 136, 252, 193, 42, 136, 252, 193, 44, 136, 252, 193, 46, 136, 252, 193, 48, 136, 252, 193, 50, 136, 252, 193, 52, 136, 252, 193, 54, 136, 252, 193, 178, 136, 252, 193, 180, 136, 252, 193, 182, 136, 252, 193, 184, 136, 252, 193, 186, 136, 252, 193, 188, 136, 252, 193, 190, 136, 252, 193, 192, 136, 252, 193, 194, 136, 252, 206, 32, 63, 0, 206, 32, 63, 0, 206, 32, 63, 0, 206, 32, 63, 0, 206, 32, 63, 0, 206, 32, 63, 0, 206, 32, 63, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 192, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 224, 62, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 32, 63, 0, 206, 32, 63, 0, 206, 32, 63, 0, 206, 32, 63, 0, 193, 196, 136, 252, 193, 86, 137, 252, 193, 88, 137, 252, 193, 230, 137, 252, 193, 232, 137, 252, 193, 234, 137, 252, 193, 236, 137, 252, 193, 238, 137, 252, 193, 240, 137, 252, 193, 242, 137, 252, 193, 244, 137, 252, 193, 246, 137, 252, 193, 248, 137, 252, 193, 250, 137, 252, 193, 252, 137, 252, 193, 254, 137, 252, 193, 2, 138, 252, 193, 4, 138, 252, 193, 6, 138, 252, 193, 8, 138, 252, 193, 10, 138, 252, 193, 12, 138, 252, 206, 128, 63, 0, 206, 128, 63, 0, 206, 128, 63, 0, 206, 128, 63, 0, 206, 128, 63, 0, 206, 128, 63, 0, 193, 140, 24, 253, 193, 150, 24, 253, 193, 170, 24, 253, 193, 186, 24, 253, 193, 188, 24, 253, 193, 206, 24, 253, 193, 4, 25, 253, 193, 12, 25, 253, 193, 14, 25, 253, 193, 32, 25, 253, 193, 36, 25, 253, 193, 42, 25, 253, 193, 48, 25, 253, 193, 54, 25, 253, 193, 56, 25, 253, 193, 74, 25, 253, 193, 76, 25, 253, 193, 78, 25, 253, 193, 80, 25, 253, 193, 82, 25, 253, 193, 84, 25, 253, 193, 124, 25, 253, 193, 198, 136, 252, 193, 126, 25, 253, 193, 128, 25, 253, 193, 50, 25, 253, 193, 130, 25, 253, 193, 176, 25, 253, 193, 178, 25, 253, 193, 230, 25, 253, 193, 232, 25, 253, 193, 234, 25, 253, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 0, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 193, 56, 137, 252, 193, 58, 137, 252, 193, 60, 137, 252, 193, 62, 137, 252, 193, 218, 137, 252, 193, 220, 137, 252, 193, 222, 137, 252, 193, 224, 137, 252, 193, 226, 137, 252, 193, 228, 137, 252, 193, 114, 138, 252, 193, 116, 138, 252, 193, 118, 138, 252, 193, 120, 138, 252, 193, 210, 138, 252, 193, 212, 138, 252, 193, 214, 138, 252, 193, 56, 139, 252, 193, 58, 139, 252, 193, 132, 139, 252, 193, 186, 139, 252, 193, 240, 139, 252, 193, 242, 139, 252, 193, 2, 140, 252, 193, 30, 140, 252, 193, 80, 140, 252, 193, 108, 140, 252, 193, 194, 140, 252, 193, 2, 141, 252, 193, 130, 141, 252, 193, 132, 141, 252, 193, 16, 142, 252, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 64, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 193, 190, 137, 252, 193, 192, 137, 252, 193, 194, 137, 252, 193, 196, 137, 252, 193, 198, 137, 252, 193, 200, 137, 252, 193, 202, 137, 252, 193, 204, 137, 252, 193, 206, 137, 252, 193, 92, 138, 252, 193, 94, 138, 252, 193, 96, 138, 252, 193, 98, 138, 252, 193, 100, 138, 252, 193, 194, 138, 252, 193, 196, 138, 252, 193, 198, 138, 252, 193, 26, 139, 252, 193, 28, 139, 252, 193, 30, 139, 252, 193, 32, 139, 252, 193, 34, 139, 252, 193, 36, 139, 252, 193, 38, 139, 252, 193, 116, 139, 252, 193, 118, 139, 252, 193, 120, 139, 252, 193, 180, 139, 252, 193, 212, 139, 252, 193, 18, 140, 252, 193, 136, 140, 252, 193, 150, 140, 252, 193, 210, 137, 252, 193, 212, 137, 252, 193, 214, 137, 252, 193, 216, 137, 252, 193, 102, 138, 252, 193, 104, 138, 252, 193, 106, 138, 252, 193, 108, 138, 252, 193, 110, 138, 252, 193, 112, 138, 252, 193, 200, 138, 252, 193, 202, 138, 252, 193, 204, 138, 252, 193, 206, 138, 252, 193, 208, 138, 252, 193, 40, 139, 252, 193, 42, 139, 252, 193, 44, 139, 252, 193, 46, 139, 252, 193, 48, 139, 252, 193, 50, 139, 252, 193, 52, 139, 252, 193, 54, 139, 252, 193, 122, 139, 252, 193, 124, 139, 252, 193, 126, 139, 252, 193, 128, 139, 252, 193, 130, 139, 252, 193, 182, 139, 252, 193, 184, 139, 252, 193, 214, 139, 252, 193, 216, 139, 252, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 206, 96, 63, 0, 193, 148, 138, 252, 193, 150, 138, 252, 193, 152, 138, 252, 193, 154, 138, 252, 193, 156, 138, 252, 193, 158, 138, 252, 193, 160, 138, 252, 193, 162, 138, 252, 193, 164, 138, 252, 193, 166, 138, 252, 193, 168, 138, 252, 193, 170, 138, 252, 193, 172, 138, 252, 193, 174, 138, 252, 193, 176, 138, 252, 193, 178, 138, 252, 193, 180, 138, 252, 193, 182, 138, 252, 193, 184, 138, 252, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 128, 63, 0, 206, 128, 63, 0, 206, 128, 63, 0, 206, 128, 63, 0, 193, 216, 138, 252, 193, 218, 138, 252, 193, 220, 138, 252, 193, 60, 139, 252, 193, 62, 139, 252, 193, 64, 139, 252, 193, 66, 139, 252, 193, 68, 139, 252, 193, 134, 139, 252, 193, 136, 139, 252, 193, 138, 139, 252, 193, 140, 139, 252, 193, 188, 139, 252, 193, 190, 139, 252, 193, 218, 139, 252, 193, 4, 140, 252, 193, 6, 140, 252, 193, 32, 140, 252, 193, 48, 140, 252, 193, 50, 140, 252, 193, 88, 140, 252, 193, 92, 140, 252, 193, 94, 140, 252, 193, 98, 140, 252, 193, 102, 140, 252, 193, 110, 140, 252, 193, 156, 140, 252, 193, 168, 140, 252, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 206, 160, 63, 0, 193, 72, 139, 252, 193, 74, 139, 252, 193, 76, 139, 252, 193, 78, 139, 252, 193, 80, 139, 252, 193, 82, 139, 252, 193, 84, 139, 252, 193, 86, 139, 252, 193, 88, 139, 252, 193, 90, 139, 252, 193, 92, 139, 252, 193, 94, 139, 252, 193, 96, 139, 252, 193, 98, 139, 252, 193, 100, 139, 252, 193, 102, 139, 252, 193, 104, 139, 252, 193, 106, 139, 252, 193, 22, 139, 252, 193, 24, 139, 252, 193, 178, 139, 252, 193, 210, 139, 252, 193, 70, 140, 252, 193, 144, 140, 252, 193, 166, 140, 252, 193, 252, 140, 252, 193, 100, 141, 252, 193, 102, 141, 252, 193, 244, 141, 252, 193, 246, 141, 252, 193, 248, 141, 252, 193, 250, 141, 252, 193, 146, 142, 252, 193, 148, 142, 252, 193, 150, 142, 252, 193, 152, 142, 252, 193, 28, 143, 252, 193, 30, 143, 252, 193, 32, 143, 252, 193, 34, 143, 252, 193, 36, 143, 252, 193, 38, 143, 252, 193, 40, 143, 252, 193, 226, 143, 252, 193, 228, 143, 252, 193, 230, 143, 252, 193, 176, 144, 252, 193, 178, 144, 252, 193, 180, 144, 252, 193, 182, 144, 252, 193, 108, 139, 252, 193, 110, 139, 252, 193, 112, 139, 252, 193, 148, 139, 252, 193, 150, 139, 252, 193, 152, 139, 252, 193, 154, 139, 252, 193, 156, 139, 252, 193, 158, 139, 252, 193, 160, 139, 252, 193, 162, 139, 252, 193, 164, 139, 252, 193, 166, 139, 252, 193, 168, 139, 252, 193, 170, 139, 252, 193, 172, 139, 252, 193, 174, 139, 252, 193, 176, 139, 252, 193, 194, 139, 252, 193, 196, 139, 252, 193, 198, 139, 252, 193, 200, 139, 252, 193, 202, 139, 252, 193, 204, 139, 252, 193, 206, 139, 252, 193, 208, 139, 252, 193, 222, 139, 252, 193, 224, 139, 252, 193, 226, 139, 252, 193, 228, 139, 252, 193, 230, 139, 252, 193, 232, 139, 252, 193, 144, 139, 252, 193, 146, 139, 252, 193, 192, 139, 252, 193, 220, 139, 252, 193, 244, 139, 252, 193, 86, 140, 252, 193, 178, 140, 252, 193, 180, 140, 252, 193, 210, 140, 252, 193, 212, 140, 252, 193, 214, 140, 252, 193, 216, 140, 252, 193, 218, 140, 252, 193, 220, 140, 252, 193, 30, 141, 252, 193, 32, 141, 252, 193, 34, 141, 252, 193, 36, 141, 252, 193, 38, 141, 252, 193, 40, 141, 252, 193, 42, 141, 252, 193, 44, 141, 252, 193, 184, 141, 252, 193, 186, 141, 252, 193, 188, 141, 252, 193, 190, 141, 252, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 193, 234, 139, 252, 193, 236, 139, 252, 193, 238, 139, 252, 193, 20, 140, 252, 193, 22, 140, 252, 193, 24, 140, 252, 193, 26, 140, 252, 193, 28, 140, 252, 193, 46, 140, 252, 193, 60, 140, 252, 193, 62, 140, 252, 193, 64, 140, 252, 193, 106, 140, 252, 193, 146, 140, 252, 193, 152, 140, 252, 193, 192, 140, 252, 193, 254, 140, 252, 193, 112, 141, 252, 193, 114, 141, 252, 193, 116, 141, 252, 193, 118, 141, 252, 193, 120, 141, 252, 193, 122, 141, 252, 193, 124, 141, 252, 193, 126, 141, 252, 193, 128, 141, 252, 193, 2, 142, 252, 193, 4, 142, 252, 193, 6, 142, 252, 193, 8, 142, 252, 193, 10, 142, 252, 193, 12, 142, 252, 193, 246, 139, 252, 193, 248, 139, 252, 193, 250, 139, 252, 193, 252, 139, 252, 193, 254, 139, 252, 193, 8, 140, 252, 193, 10, 140, 252, 193, 12, 140, 252, 193, 14, 140, 252, 193, 16, 140, 252, 193, 34, 140, 252, 193, 36, 140, 252, 193, 38, 140, 252, 193, 40, 140, 252, 193, 42, 140, 252, 193, 44, 140, 252, 193, 52, 140, 252, 193, 54, 140, 252, 193, 56, 140, 252, 193, 58, 140, 252, 193, 66, 140, 252, 193, 68, 140, 252, 193, 72, 140, 252, 193, 74, 140, 252, 193, 76, 140, 252, 193, 82, 140, 252, 193, 84, 140, 252, 193, 90, 140, 252, 193, 96, 140, 252, 193, 100, 140, 252, 193, 104, 140, 252, 193, 114, 140, 252, 206, 64, 83, 0, 206, 64, 83, 0, 206, 64, 83, 0, 206, 64, 83, 0, 206, 64, 83, 0, 206, 64, 83, 0, 193, 78, 140, 252, 193, 120, 252, 252, 193, 122, 252, 252, 193, 124, 252, 252, 193, 126, 252, 252, 193, 128, 252, 252, 193, 130, 252, 252, 193, 132, 252, 252, 193, 134, 252, 252, 193, 142, 252, 252, 193, 144, 252, 252, 193, 146, 252, 252, 193, 148, 252, 252, 193, 170, 252, 252, 193, 154, 252, 252, 193, 156, 252, 252, 193, 160, 252, 252, 193, 164, 252, 252, 193, 172, 252, 252, 193, 168, 252, 252, 193, 162, 252, 252, 193, 184, 252, 252, 193, 198, 252, 252, 193, 200, 252, 252, 193, 202, 252, 252, 193, 228, 252, 252, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 193, 192, 232, 252, 193, 194, 232, 252, 193, 196, 232, 252, 193, 112, 140, 252, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 64, 76, 0, 206, 64, 76, 0, 206, 64, 76, 0, 206, 64, 76, 0, 206, 64, 76, 0, 193, 180, 217, 252, 193, 182, 217, 252, 193, 184, 217, 252, 193, 186, 217, 252, 193, 188, 217, 252, 193, 190, 217, 252, 193, 192, 217, 252, 193, 194, 217, 252, 193, 196, 217, 252, 193, 198, 217, 252, 193, 214, 217, 252, 193, 216, 217, 252, 193, 218, 217, 252, 193, 220, 217, 252, 193, 222, 217, 252, 193, 134, 140, 252, 193, 224, 217, 252, 193, 226, 217, 252, 193, 228, 217, 252, 193, 232, 217, 252, 193, 234, 217, 252, 193, 236, 217, 252, 193, 238, 217, 252, 193, 240, 217, 252, 193, 242, 217, 252, 193, 244, 217, 252, 193, 246, 217, 252, 193, 170, 140, 252, 193, 172, 140, 252, 193, 174, 140, 252, 193, 176, 140, 252, 193, 196, 140, 252, 193, 198, 140, 252, 193, 200, 140, 252, 193, 202, 140, 252, 193, 204, 140, 252, 193, 206, 140, 252, 193, 208, 140, 252, 193, 4, 141, 252, 193, 6, 141, 252, 193, 8, 141, 252, 193, 10, 141, 252, 193, 12, 141, 252, 193, 14, 141, 252, 193, 16, 141, 252, 193, 18, 141, 252, 193, 20, 141, 252, 193, 22, 141, 252, 193, 24, 141, 252, 193, 26, 141, 252, 193, 28, 141, 252, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 193, 190, 140, 252, 193, 104, 141, 252, 193, 106, 141, 252, 193, 108, 141, 252, 193, 110, 141, 252, 193, 252, 141, 252, 193, 254, 141, 252, 193, 156, 142, 252, 193, 158, 142, 252, 193, 160, 142, 252, 193, 42, 143, 252, 193, 44, 143, 252, 193, 46, 143, 252, 193, 234, 143, 252, 193, 236, 143, 252, 193, 238, 143, 252, 193, 240, 143, 252, 193, 186, 144, 252, 193, 188, 144, 252, 193, 190, 144, 252, 193, 192, 144, 252, 193, 106, 145, 252, 193, 108, 145, 252, 193, 110, 145, 252, 193, 112, 145, 252, 193, 114, 145, 252, 193, 20, 146, 252, 193, 22, 146, 252, 193, 24, 146, 252, 193, 26, 146, 252, 193, 170, 146, 252, 193, 172, 146, 252, 193, 246, 140, 252, 193, 248, 140, 252, 193, 250, 140, 252, 193, 46, 141, 252, 193, 48, 141, 252, 193, 50, 141, 252, 193, 52, 141, 252, 193, 54, 141, 252, 193, 56, 141, 252, 193, 58, 141, 252, 193, 60, 141, 252, 193, 62, 141, 252, 193, 64, 141, 252, 193, 66, 141, 252, 193, 68, 141, 252, 193, 70, 141, 252, 193, 72, 141, 252, 193, 74, 141, 252, 193, 76, 141, 252, 193, 78, 141, 252, 193, 80, 141, 252, 193, 82, 141, 252, 193, 84, 141, 252, 193, 86, 141, 252, 193, 88, 141, 252, 193, 90, 141, 252, 193, 92, 141, 252, 193, 94, 141, 252, 193, 96, 141, 252, 193, 98, 141, 252, 193, 192, 141, 252, 193, 194, 141, 252, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 206, 192, 63, 0, 193, 28, 142, 252, 193, 30, 142, 252, 193, 32, 142, 252, 193, 34, 142, 252, 193, 36, 142, 252, 193, 38, 142, 252, 193, 40, 142, 252, 193, 42, 142, 252, 193, 44, 142, 252, 193, 46, 142, 252, 193, 48, 142, 252, 193, 50, 142, 252, 193, 52, 142, 252, 193, 54, 142, 252, 193, 178, 142, 252, 193, 196, 141, 252, 193, 198, 141, 252, 193, 200, 141, 252, 193, 202, 141, 252, 193, 204, 141, 252, 193, 206, 141, 252, 193, 208, 141, 252, 193, 210, 141, 252, 193, 212, 141, 252, 193, 214, 141, 252, 193, 216, 141, 252, 193, 218, 141, 252, 193, 220, 141, 252, 193, 222, 141, 252, 193, 224, 141, 252, 193, 226, 141, 252, 193, 228, 141, 252, 193, 230, 141, 252, 193, 232, 141, 252, 193, 234, 141, 252, 193, 236, 141, 252, 193, 238, 141, 252, 193, 240, 141, 252, 193, 242, 141, 252, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 193, 14, 142, 252, 193, 162, 142, 252, 193, 164, 142, 252, 193, 166, 142, 252, 193, 168, 142, 252, 193, 48, 143, 252, 193, 50, 143, 252, 193, 242, 143, 252, 193, 244, 143, 252, 193, 246, 143, 252, 193, 248, 143, 252, 193, 250, 143, 252, 193, 252, 143, 252, 193, 254, 143, 252, 193, 2, 144, 252, 193, 4, 144, 252, 193, 6, 144, 252, 193, 8, 144, 252, 193, 10, 144, 252, 193, 194, 144, 252, 193, 196, 144, 252, 193, 198, 144, 252, 193, 200, 144, 252, 193, 202, 144, 252, 193, 204, 144, 252, 193, 206, 144, 252, 193, 208, 144, 252, 193, 116, 145, 252, 193, 118, 145, 252, 193, 120, 145, 252, 193, 122, 145, 252, 193, 124, 145, 252, 193, 18, 142, 252, 193, 20, 142, 252, 193, 22, 142, 252, 193, 24, 142, 252, 193, 26, 142, 252, 193, 170, 142, 252, 193, 172, 142, 252, 193, 174, 142, 252, 193, 176, 142, 252, 193, 52, 143, 252, 193, 54, 143, 252, 193, 56, 143, 252, 193, 58, 143, 252, 193, 12, 144, 252, 193, 14, 144, 252, 193, 16, 144, 252, 193, 210, 144, 252, 193, 132, 145, 252, 193, 134, 145, 252, 193, 136, 145, 252, 193, 138, 145, 252, 193, 46, 146, 252, 193, 48, 146, 252, 193, 50, 146, 252, 193, 188, 146, 252, 193, 190, 146, 252, 193, 192, 146, 252, 193, 194, 146, 252, 193, 38, 147, 252, 193, 40, 147, 252, 193, 164, 147, 252, 193, 166, 147, 252, 193, 180, 142, 252, 193, 182, 142, 252, 193, 184, 142, 252, 193, 186, 142, 252, 193, 188, 142, 252, 193, 190, 142, 252, 193, 56, 142, 252, 193, 192, 142, 252, 193, 194, 142, 252, 193, 196, 142, 252, 193, 198, 142, 252, 193, 200, 142, 252, 193, 202, 142, 252, 193, 204, 142, 252, 193, 206, 142, 252, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 206, 224, 63, 0, 193, 210, 142, 252, 193, 212, 142, 252, 193, 214, 142, 252, 193, 216, 142, 252, 193, 218, 142, 252, 193, 220, 142, 252, 193, 222, 142, 252, 193, 224, 142, 252, 193, 226, 142, 252, 193, 106, 143, 252, 193, 108, 143, 252, 193, 110, 143, 252, 193, 112, 143, 252, 193, 114, 143, 252, 193, 116, 143, 252, 193, 118, 143, 252, 193, 120, 143, 252, 193, 122, 143, 252, 193, 124, 143, 252, 193, 126, 143, 252, 193, 128, 143, 252, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 0, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 206, 32, 64, 0, 193, 144, 142, 252, 193, 14, 143, 252, 193, 16, 143, 252, 193, 18, 143, 252, 193, 20, 143, 252, 193, 22, 143, 252, 193, 24, 143, 252, 193, 26, 143, 252, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 206, 64, 64, 0, 193, 18, 144, 252, 193, 20, 144, 252, 193, 22, 144, 252, 193, 24, 144, 252, 193, 26, 144, 252, 193, 28, 144, 252, 193, 30, 144, 252, 193, 32, 144, 252, 193, 34, 144, 252, 193, 36, 144, 252, 193, 38, 144, 252, 193, 40, 144, 252, 193, 42, 144, 252, 193, 44, 144, 252, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 206, 96, 64, 0, 193, 72, 144, 252, 193, 74, 144, 252, 193, 76, 144, 252, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 206, 128, 64, 0, 193, 140, 145, 252, 193, 142, 145, 252, 193, 144, 145, 252, 193, 146, 145, 252, 193, 148, 145, 252, 193, 52, 146, 252, 193, 46, 144, 252, 193, 150, 145, 252, 193, 152, 145, 252, 193, 154, 145, 252, 193, 156, 145, 252, 193, 158, 145, 252, 193, 160, 145, 252, 193, 162, 145, 252, 193, 164, 145, 252, 193, 54, 146, 252, 193, 56, 146, 252, 193, 58, 146, 252, 193, 60, 146, 252, 193, 62, 146, 252, 193, 64, 146, 252, 193, 66, 146, 252, 193, 68, 146, 252, 193, 70, 146, 252, 193, 72, 146, 252, 193, 74, 146, 252, 193, 76, 146, 252, 193, 196, 146, 252, 193, 48, 144, 252, 193, 50, 144, 252, 193, 52, 144, 252, 193, 54, 144, 252, 193, 56, 144, 252, 193, 58, 144, 252, 193, 60, 144, 252, 193, 62, 144, 252, 193, 64, 144, 252, 193, 66, 144, 252, 193, 68, 144, 252, 193, 70, 144, 252, 193, 244, 144, 252, 193, 246, 144, 252, 193, 248, 144, 252, 193, 250, 144, 252, 193, 252, 144, 252, 193, 254, 144, 252, 193, 2, 145, 252, 193, 4, 145, 252, 193, 6, 145, 252, 193, 8, 145, 252, 193, 10, 145, 252, 193, 166, 145, 252, 193, 168, 145, 252, 193, 170, 145, 252, 193, 172, 145, 252, 193, 174, 145, 252, 193, 176, 145, 252, 193, 178, 145, 252, 193, 78, 146, 252, 193, 80, 146, 252, 193, 78, 144, 252, 193, 80, 144, 252, 193, 82, 144, 252, 193, 84, 144, 252, 193, 86, 144, 252, 193, 88, 144, 252, 193, 90, 144, 252, 193, 92, 144, 252, 193, 94, 144, 252, 193, 96, 144, 252, 193, 98, 144, 252, 193, 100, 144, 252, 193, 102, 144, 252, 193, 104, 144, 252, 193, 106, 144, 252, 193, 108, 144, 252, 193, 110, 144, 252, 193, 112, 144, 252, 193, 114, 144, 252, 193, 116, 144, 252, 193, 118, 144, 252, 193, 120, 144, 252, 193, 122, 144, 252, 193, 124, 144, 252, 193, 126, 144, 252, 193, 128, 144, 252, 193, 130, 144, 252, 193, 132, 144, 252, 193, 134, 144, 252, 193, 136, 144, 252, 193, 138, 144, 252, 193, 140, 144, 252, 193, 142, 144, 252, 193, 144, 144, 252, 193, 146, 144, 252, 193, 148, 144, 252, 193, 150, 144, 252, 193, 152, 144, 252, 193, 154, 144, 252, 193, 156, 144, 252, 193, 158, 144, 252, 193, 160, 144, 252, 193, 162, 144, 252, 193, 164, 144, 252, 193, 166, 144, 252, 193, 168, 144, 252, 193, 170, 144, 252, 193, 172, 144, 252, 193, 174, 144, 252, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 193, 184, 144, 252, 193, 98, 145, 252, 193, 100, 145, 252, 193, 102, 145, 252, 193, 104, 145, 252, 193, 18, 146, 252, 193, 168, 146, 252, 193, 30, 147, 252, 193, 32, 147, 252, 193, 160, 147, 252, 193, 146, 148, 252, 193, 234, 148, 252, 193, 70, 149, 252, 193, 72, 149, 252, 193, 182, 149, 252, 193, 184, 149, 252, 193, 186, 149, 252, 193, 42, 150, 252, 193, 44, 150, 252, 193, 46, 150, 252, 193, 134, 150, 252, 193, 136, 150, 252, 193, 226, 150, 252, 193, 80, 151, 252, 193, 82, 151, 252, 193, 84, 151, 252, 193, 86, 151, 252, 193, 184, 151, 252, 193, 226, 151, 252, 193, 92, 152, 252, 193, 146, 152, 252, 193, 186, 152, 252, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 160, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 193, 126, 145, 252, 193, 128, 145, 252, 193, 130, 145, 252, 193, 28, 146, 252, 193, 30, 146, 252, 193, 32, 146, 252, 193, 34, 146, 252, 193, 36, 146, 252, 193, 38, 146, 252, 193, 40, 146, 252, 193, 42, 146, 252, 193, 44, 146, 252, 193, 176, 146, 252, 193, 178, 146, 252, 193, 180, 146, 252, 193, 182, 146, 252, 193, 184, 146, 252, 193, 186, 146, 252, 193, 36, 147, 252, 193, 98, 147, 252, 193, 192, 147, 252, 193, 194, 147, 252, 193, 248, 147, 252, 193, 250, 147, 252, 193, 82, 148, 252, 193, 152, 148, 252, 193, 238, 148, 252, 193, 240, 148, 252, 193, 78, 149, 252, 193, 192, 149, 252, 193, 194, 149, 252, 193, 196, 149, 252, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 192, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 193, 82, 146, 252, 193, 206, 146, 252, 193, 208, 146, 252, 193, 210, 146, 252, 193, 212, 146, 252, 193, 214, 146, 252, 193, 216, 146, 252, 193, 218, 146, 252, 193, 108, 147, 252, 193, 110, 147, 252, 193, 170, 147, 252, 193, 172, 147, 252, 193, 174, 147, 252, 193, 176, 147, 252, 193, 204, 147, 252, 193, 206, 147, 252, 193, 16, 148, 252, 193, 32, 148, 252, 193, 54, 148, 252, 193, 100, 148, 252, 193, 102, 148, 252, 193, 104, 148, 252, 193, 106, 148, 252, 193, 178, 148, 252, 193, 180, 148, 252, 193, 182, 148, 252, 193, 12, 149, 252, 193, 14, 149, 252, 193, 16, 149, 252, 193, 18, 149, 252, 193, 20, 149, 252, 193, 22, 149, 252, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 224, 64, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 193, 174, 146, 252, 193, 34, 147, 252, 193, 96, 147, 252, 193, 162, 147, 252, 193, 190, 147, 252, 193, 44, 148, 252, 193, 148, 148, 252, 193, 150, 148, 252, 193, 236, 148, 252, 193, 74, 149, 252, 193, 76, 149, 252, 193, 188, 149, 252, 193, 190, 149, 252, 193, 48, 150, 252, 193, 228, 150, 252, 193, 230, 150, 252, 193, 88, 151, 252, 193, 90, 151, 252, 193, 92, 151, 252, 193, 94, 151, 252, 193, 228, 151, 252, 193, 72, 152, 252, 193, 122, 152, 252, 193, 148, 152, 252, 193, 188, 152, 252, 193, 190, 152, 252, 193, 192, 152, 252, 193, 6, 153, 252, 193, 8, 153, 252, 193, 54, 153, 252, 193, 108, 153, 252, 193, 110, 153, 252, 193, 198, 146, 252, 193, 200, 146, 252, 193, 202, 146, 252, 193, 204, 146, 252, 193, 42, 147, 252, 193, 44, 147, 252, 193, 46, 147, 252, 193, 48, 147, 252, 193, 50, 147, 252, 193, 52, 147, 252, 193, 54, 147, 252, 193, 56, 147, 252, 193, 100, 147, 252, 193, 102, 147, 252, 193, 104, 147, 252, 193, 106, 147, 252, 193, 200, 147, 252, 193, 218, 147, 252, 193, 240, 147, 252, 193, 202, 147, 252, 193, 22, 148, 252, 193, 26, 148, 252, 193, 28, 148, 252, 193, 30, 148, 252, 193, 46, 148, 252, 193, 48, 148, 252, 193, 50, 148, 252, 193, 52, 148, 252, 206, 32, 65, 0, 206, 32, 65, 0, 206, 32, 65, 0, 206, 32, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 206, 0, 65, 0, 193, 58, 147, 252, 193, 10, 147, 252, 193, 12, 147, 252, 193, 14, 147, 252, 193, 16, 147, 252, 193, 18, 147, 252, 193, 20, 147, 252, 193, 22, 147, 252, 193, 24, 147, 252, 193, 26, 147, 252, 193, 28, 147, 252, 193, 60, 147, 252, 193, 62, 147, 252, 193, 64, 147, 252, 193, 66, 147, 252, 193, 68, 147, 252, 193, 70, 147, 252, 193, 72, 147, 252, 193, 74, 147, 252, 193, 76, 147, 252, 193, 78, 147, 252, 193, 80, 147, 252, 193, 82, 147, 252, 193, 84, 147, 252, 193, 112, 147, 252, 193, 86, 147, 252, 193, 88, 147, 252, 193, 90, 147, 252, 193, 92, 147, 252, 193, 94, 147, 252, 193, 114, 147, 252, 193, 116, 147, 252, 193, 118, 147, 252, 193, 120, 147, 252, 193, 122, 147, 252, 193, 124, 147, 252, 193, 126, 147, 252, 193, 128, 147, 252, 193, 130, 147, 252, 193, 132, 147, 252, 193, 134, 147, 252, 193, 136, 147, 252, 193, 138, 147, 252, 193, 140, 147, 252, 193, 142, 147, 252, 193, 216, 98, 253, 193, 144, 147, 252, 193, 146, 147, 252, 193, 148, 147, 252, 193, 150, 147, 252, 193, 152, 147, 252, 193, 154, 147, 252, 193, 156, 147, 252, 193, 158, 147, 252, 193, 196, 147, 252, 193, 198, 147, 252, 193, 216, 147, 252, 193, 238, 147, 252, 193, 154, 148, 252, 193, 242, 148, 252, 193, 244, 148, 252, 193, 246, 148, 252, 193, 248, 148, 252, 193, 250, 148, 252, 193, 80, 149, 252, 193, 200, 149, 252, 193, 202, 149, 252, 193, 204, 149, 252, 193, 206, 149, 252, 193, 208, 149, 252, 193, 64, 150, 252, 193, 66, 150, 252, 193, 68, 150, 252, 193, 146, 150, 252, 193, 148, 150, 252, 193, 150, 150, 252, 193, 238, 150, 252, 193, 240, 150, 252, 193, 110, 151, 252, 193, 112, 151, 252, 193, 196, 151, 252, 193, 198, 151, 252, 193, 236, 151, 252, 193, 238, 151, 252, 193, 240, 151, 252, 193, 76, 152, 252, 193, 18, 148, 252, 193, 20, 148, 252, 193, 24, 148, 252, 193, 34, 148, 252, 193, 36, 148, 252, 193, 38, 148, 252, 193, 40, 148, 252, 193, 56, 148, 252, 193, 58, 148, 252, 193, 60, 148, 252, 193, 62, 148, 252, 193, 64, 148, 252, 193, 66, 148, 252, 193, 68, 148, 252, 193, 70, 148, 252, 193, 72, 148, 252, 193, 74, 148, 252, 193, 76, 148, 252, 193, 78, 148, 252, 193, 80, 148, 252, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 32, 65, 0, 206, 32, 65, 0, 206, 32, 65, 0, 206, 32, 65, 0, 193, 156, 148, 252, 193, 158, 148, 252, 193, 160, 148, 252, 193, 162, 148, 252, 193, 164, 148, 252, 193, 166, 148, 252, 193, 168, 148, 252, 193, 170, 148, 252, 193, 172, 148, 252, 193, 174, 148, 252, 193, 176, 148, 252, 193, 252, 148, 252, 193, 254, 148, 252, 193, 2, 149, 252, 193, 4, 149, 252, 193, 6, 149, 252, 193, 8, 149, 252, 193, 10, 149, 252, 193, 82, 149, 252, 193, 84, 149, 252, 193, 86, 149, 252, 193, 88, 149, 252, 193, 90, 149, 252, 193, 92, 149, 252, 193, 94, 149, 252, 193, 96, 149, 252, 193, 210, 149, 252, 193, 212, 149, 252, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 206, 64, 65, 0, 193, 184, 148, 252, 193, 186, 148, 252, 193, 188, 148, 252, 193, 190, 148, 252, 193, 192, 148, 252, 193, 194, 148, 252, 193, 196, 148, 252, 193, 198, 148, 252, 193, 200, 148, 252, 193, 202, 148, 252, 193, 204, 148, 252, 193, 206, 148, 252, 193, 208, 148, 252, 193, 210, 148, 252, 193, 212, 148, 252, 193, 214, 148, 252, 193, 216, 148, 252, 193, 218, 148, 252, 193, 220, 148, 252, 193, 222, 148, 252, 193, 224, 148, 252, 193, 226, 148, 252, 193, 228, 148, 252, 193, 230, 148, 252, 193, 232, 148, 252, 193, 24, 149, 252, 193, 98, 149, 252, 193, 100, 149, 252, 193, 102, 149, 252, 193, 104, 149, 252, 193, 106, 149, 252, 193, 230, 149, 252, 193, 232, 149, 252, 193, 234, 149, 252, 193, 236, 149, 252, 193, 82, 150, 252, 193, 84, 150, 252, 193, 166, 150, 252, 193, 168, 150, 252, 193, 170, 150, 252, 193, 6, 151, 252, 193, 8, 151, 252, 193, 10, 151, 252, 193, 12, 151, 252, 193, 14, 151, 252, 193, 16, 151, 252, 193, 126, 151, 252, 193, 128, 151, 252, 193, 130, 151, 252, 193, 132, 151, 252, 193, 134, 151, 252, 193, 136, 151, 252, 193, 138, 151, 252, 193, 204, 151, 252, 193, 206, 151, 252, 193, 244, 151, 252, 193, 246, 151, 252, 193, 26, 149, 252, 193, 28, 149, 252, 193, 30, 149, 252, 193, 32, 149, 252, 193, 34, 149, 252, 193, 36, 149, 252, 193, 38, 149, 252, 193, 40, 149, 252, 193, 42, 149, 252, 193, 44, 149, 252, 193, 46, 149, 252, 193, 48, 149, 252, 193, 50, 149, 252, 193, 52, 149, 252, 193, 54, 149, 252, 193, 56, 149, 252, 193, 58, 149, 252, 193, 60, 149, 252, 193, 62, 149, 252, 193, 64, 149, 252, 193, 66, 149, 252, 193, 68, 149, 252, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 96, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 193, 198, 149, 252, 193, 50, 150, 252, 193, 52, 150, 252, 193, 54, 150, 252, 193, 56, 150, 252, 193, 58, 150, 252, 193, 60, 150, 252, 193, 62, 150, 252, 193, 138, 150, 252, 193, 140, 150, 252, 193, 142, 150, 252, 193, 144, 150, 252, 193, 232, 150, 252, 193, 234, 150, 252, 193, 236, 150, 252, 193, 96, 151, 252, 193, 98, 151, 252, 193, 100, 151, 252, 193, 102, 151, 252, 193, 104, 151, 252, 193, 106, 151, 252, 193, 108, 151, 252, 193, 186, 151, 252, 193, 188, 151, 252, 193, 190, 151, 252, 193, 192, 151, 252, 193, 194, 151, 252, 193, 230, 151, 252, 193, 232, 151, 252, 193, 234, 151, 252, 193, 8, 152, 252, 193, 10, 152, 252, 193, 214, 149, 252, 193, 216, 149, 252, 193, 218, 149, 252, 193, 220, 149, 252, 193, 222, 149, 252, 193, 224, 149, 252, 193, 226, 149, 252, 193, 228, 149, 252, 193, 70, 150, 252, 193, 72, 150, 252, 193, 74, 150, 252, 193, 76, 150, 252, 193, 78, 150, 252, 193, 80, 150, 252, 193, 152, 150, 252, 193, 154, 150, 252, 193, 156, 150, 252, 193, 158, 150, 252, 193, 160, 150, 252, 193, 162, 150, 252, 193, 164, 150, 252, 193, 242, 150, 252, 193, 244, 150, 252, 193, 246, 150, 252, 193, 248, 150, 252, 193, 250, 150, 252, 193, 252, 150, 252, 193, 254, 150, 252, 193, 2, 151, 252, 193, 4, 151, 252, 193, 114, 151, 252, 193, 116, 151, 252, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 128, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 206, 160, 65, 0, 193, 172, 150, 252, 193, 174, 150, 252, 193, 176, 150, 252, 193, 178, 150, 252, 193, 180, 150, 252, 193, 182, 150, 252, 193, 184, 150, 252, 193, 186, 150, 252, 193, 188, 150, 252, 193, 190, 150, 252, 193, 192, 150, 252, 193, 194, 150, 252, 193, 196, 150, 252, 193, 198, 150, 252, 193, 200, 150, 252, 193, 202, 150, 252, 193, 204, 150, 252, 193, 206, 150, 252, 193, 32, 202, 252, 193, 208, 150, 252, 193, 210, 150, 252, 193, 212, 150, 252, 193, 214, 150, 252, 193, 216, 150, 252, 193, 218, 150, 252, 193, 220, 150, 252, 193, 222, 150, 252, 193, 224, 150, 252, 193, 18, 151, 252, 193, 20, 151, 252, 193, 22, 151, 252, 193, 24, 151, 252, 193, 26, 151, 252, 193, 28, 151, 252, 193, 30, 151, 252, 193, 32, 151, 252, 193, 34, 151, 252, 193, 36, 151, 252, 193, 38, 151, 252, 193, 40, 151, 252, 193, 42, 151, 252, 193, 44, 151, 252, 193, 46, 151, 252, 193, 48, 151, 252, 193, 140, 151, 252, 193, 50, 151, 252, 193, 52, 151, 252, 193, 54, 151, 252, 193, 56, 151, 252, 193, 58, 151, 252, 193, 60, 151, 252, 193, 62, 151, 252, 193, 64, 151, 252, 193, 66, 151, 252, 193, 68, 151, 252, 193, 70, 151, 252, 193, 72, 151, 252, 193, 74, 151, 252, 193, 76, 151, 252, 193, 78, 151, 252, 193, 142, 151, 252, 193, 144, 151, 252, 193, 146, 151, 252, 193, 148, 151, 252, 193, 150, 151, 252, 193, 152, 151, 252, 193, 154, 151, 252, 193, 156, 151, 252, 193, 158, 151, 252, 193, 160, 151, 252, 193, 162, 151, 252, 193, 164, 151, 252, 193, 166, 151, 252, 193, 168, 151, 252, 193, 170, 151, 252, 193, 172, 151, 252, 193, 174, 151, 252, 193, 176, 151, 252, 193, 178, 151, 252, 193, 180, 151, 252, 193, 118, 151, 252, 193, 120, 151, 252, 193, 122, 151, 252, 193, 124, 151, 252, 193, 200, 151, 252, 193, 202, 151, 252, 193, 242, 151, 252, 193, 16, 152, 252, 193, 38, 152, 252, 193, 40, 152, 252, 193, 78, 152, 252, 193, 118, 152, 252, 193, 124, 152, 252, 193, 126, 152, 252, 193, 134, 152, 252, 193, 136, 152, 252, 193, 152, 152, 252, 193, 154, 152, 252, 193, 156, 152, 252, 193, 158, 152, 252, 193, 160, 152, 252, 193, 162, 152, 252, 193, 164, 152, 252, 193, 166, 152, 252, 193, 200, 152, 252, 193, 202, 152, 252, 193, 204, 152, 252, 193, 206, 152, 252, 193, 208, 152, 252, 193, 210, 152, 252, 193, 212, 152, 252, 193, 186, 147, 253, 193, 208, 151, 252, 193, 210, 151, 252, 193, 212, 151, 252, 193, 214, 151, 252, 193, 216, 151, 252, 193, 218, 151, 252, 193, 220, 151, 252, 193, 222, 151, 252, 193, 224, 151, 252, 193, 182, 151, 252, 193, 248, 151, 252, 193, 250, 151, 252, 193, 252, 151, 252, 193, 254, 151, 252, 193, 2, 152, 252, 193, 4, 152, 252, 193, 20, 152, 252, 193, 22, 152, 252, 193, 6, 152, 252, 193, 24, 152, 252, 193, 26, 152, 252, 193, 42, 152, 252, 193, 28, 152, 252, 193, 30, 152, 252, 193, 32, 152, 252, 193, 34, 152, 252, 193, 36, 152, 252, 206, 192, 65, 0, 206, 192, 65, 0, 206, 192, 65, 0, 206, 192, 65, 0, 206, 192, 65, 0, 193, 12, 152, 252, 193, 14, 152, 252, 193, 74, 152, 252, 193, 86, 152, 252, 193, 94, 152, 252, 193, 96, 152, 252, 193, 108, 152, 252, 193, 150, 152, 252, 193, 194, 152, 252, 193, 10, 153, 252, 193, 12, 153, 252, 193, 14, 153, 252, 193, 16, 153, 252, 193, 56, 153, 252, 193, 58, 153, 252, 193, 114, 153, 252, 193, 116, 153, 252, 193, 172, 153, 252, 193, 174, 153, 252, 193, 176, 153, 252, 193, 222, 153, 252, 193, 224, 153, 252, 193, 226, 153, 252, 193, 6, 154, 252, 193, 28, 154, 252, 193, 30, 154, 252, 193, 32, 154, 252, 193, 70, 154, 252, 193, 72, 154, 252, 193, 106, 154, 252, 193, 124, 154, 252, 193, 152, 154, 252, 193, 18, 152, 252, 193, 80, 152, 252, 193, 120, 152, 252, 193, 138, 152, 252, 193, 140, 152, 252, 193, 168, 152, 252, 193, 170, 152, 252, 193, 214, 152, 252, 193, 216, 152, 252, 193, 218, 152, 252, 193, 220, 152, 252, 193, 22, 153, 252, 193, 76, 153, 252, 193, 78, 153, 252, 193, 126, 153, 252, 193, 128, 153, 252, 193, 130, 153, 252, 193, 132, 153, 252, 193, 12, 154, 252, 193, 44, 154, 252, 193, 76, 154, 252, 193, 94, 154, 252, 193, 192, 154, 252, 193, 18, 155, 252, 193, 20, 155, 252, 193, 22, 155, 252, 193, 24, 155, 252, 193, 26, 155, 252, 193, 28, 155, 252, 193, 30, 155, 252, 193, 210, 155, 252, 193, 212, 155, 252, 206, 192, 65, 0, 206, 192, 65, 0, 206, 192, 65, 0, 206, 192, 65, 0, 206, 192, 65, 0, 206, 192, 65, 0, 206, 192, 65, 0, 206, 192, 65, 0, 206, 192, 65, 0, 193, 82, 152, 252, 193, 88, 152, 252, 193, 90, 152, 252, 193, 98, 152, 252, 193, 100, 152, 252, 193, 110, 152, 252, 193, 102, 152, 252, 193, 104, 152, 252, 193, 106, 152, 252, 193, 112, 152, 252, 193, 114, 152, 252, 193, 116, 152, 252, 193, 128, 152, 252, 193, 130, 152, 252, 193, 142, 152, 252, 193, 144, 152, 252, 193, 172, 152, 252, 193, 174, 152, 252, 193, 176, 152, 252, 193, 178, 152, 252, 193, 180, 152, 252, 193, 182, 152, 252, 193, 184, 152, 252, 193, 196, 152, 252, 193, 198, 152, 252, 193, 18, 153, 252, 193, 60, 153, 252, 193, 62, 153, 252, 193, 228, 153, 252, 193, 230, 153, 252, 193, 232, 153, 252, 193, 234, 153, 252, 193, 34, 154, 252, 193, 36, 154, 252, 193, 144, 154, 252, 193, 234, 154, 252, 193, 236, 154, 252, 193, 120, 155, 252, 193, 122, 155, 252, 193, 124, 155, 252, 193, 126, 155, 252, 193, 64, 156, 252, 193, 66, 156, 252, 193, 68, 156, 252, 193, 70, 156, 252, 193, 72, 156, 252, 193, 74, 156, 252, 193, 28, 157, 252, 193, 30, 157, 252, 193, 32, 157, 252, 193, 40, 158, 252, 193, 42, 158, 252, 193, 44, 158, 252, 193, 46, 158, 252, 193, 104, 159, 252, 193, 222, 152, 252, 193, 224, 152, 252, 193, 226, 152, 252, 193, 228, 152, 252, 193, 230, 152, 252, 193, 232, 152, 252, 193, 234, 152, 252, 193, 236, 152, 252, 193, 238, 152, 252, 193, 240, 152, 252, 193, 242, 152, 252, 193, 244, 152, 252, 193, 246, 152, 252, 193, 248, 152, 252, 193, 250, 152, 252, 193, 252, 152, 252, 193, 254, 152, 252, 193, 24, 153, 252, 193, 26, 153, 252, 193, 28, 153, 252, 193, 30, 153, 252, 193, 32, 153, 252, 193, 34, 153, 252, 193, 36, 153, 252, 193, 38, 153, 252, 193, 40, 153, 252, 193, 42, 153, 252, 193, 44, 153, 252, 193, 46, 153, 252, 193, 48, 153, 252, 193, 80, 153, 252, 193, 82, 153, 252, 193, 2, 153, 252, 193, 4, 153, 252, 193, 50, 153, 252, 193, 104, 153, 252, 193, 106, 153, 252, 193, 162, 153, 252, 193, 164, 153, 252, 193, 212, 153, 252, 193, 214, 153, 252, 193, 252, 153, 252, 193, 254, 153, 252, 193, 2, 154, 252, 193, 22, 154, 252, 193, 66, 154, 252, 193, 104, 154, 252, 193, 138, 154, 252, 193, 172, 154, 252, 193, 230, 154, 252, 193, 90, 155, 252, 193, 92, 155, 252, 193, 94, 155, 252, 193, 96, 155, 252, 193, 98, 155, 252, 193, 100, 155, 252, 193, 36, 156, 252, 193, 38, 156, 252, 193, 40, 156, 252, 193, 42, 156, 252, 193, 242, 156, 252, 193, 244, 156, 252, 193, 250, 157, 252, 193, 252, 157, 252, 193, 188, 147, 253, 193, 20, 153, 252, 193, 64, 153, 252, 193, 66, 153, 252, 193, 68, 153, 252, 193, 70, 153, 252, 193, 72, 153, 252, 193, 74, 153, 252, 193, 118, 153, 252, 193, 120, 153, 252, 193, 122, 153, 252, 193, 124, 153, 252, 193, 178, 153, 252, 193, 180, 153, 252, 193, 8, 154, 252, 193, 182, 153, 252, 193, 184, 153, 252, 193, 10, 154, 252, 193, 38, 154, 252, 193, 40, 154, 252, 193, 42, 154, 252, 193, 74, 154, 252, 193, 108, 154, 252, 193, 110, 154, 252, 193, 132, 154, 252, 193, 154, 154, 252, 193, 160, 154, 252, 193, 162, 154, 252, 206, 0, 66, 0, 206, 0, 66, 0, 206, 0, 66, 0, 206, 0, 66, 0, 193, 84, 153, 252, 193, 86, 153, 252, 193, 88, 153, 252, 193, 90, 153, 252, 193, 92, 153, 252, 193, 94, 153, 252, 193, 96, 153, 252, 193, 98, 153, 252, 193, 100, 153, 252, 193, 102, 153, 252, 193, 134, 153, 252, 193, 136, 153, 252, 193, 138, 153, 252, 193, 140, 153, 252, 193, 142, 153, 252, 193, 144, 153, 252, 193, 146, 153, 252, 193, 148, 153, 252, 193, 150, 153, 252, 193, 152, 153, 252, 193, 154, 153, 252, 193, 156, 153, 252, 193, 158, 153, 252, 193, 160, 153, 252, 206, 224, 65, 0, 206, 224, 65, 0, 206, 224, 65, 0, 206, 224, 65, 0, 206, 224, 65, 0, 206, 224, 65, 0, 206, 224, 65, 0, 206, 224, 65, 0, 193, 112, 153, 252, 193, 166, 153, 252, 193, 168, 153, 252, 193, 170, 153, 252, 193, 216, 153, 252, 193, 218, 153, 252, 193, 220, 153, 252, 193, 4, 154, 252, 193, 24, 154, 252, 193, 26, 154, 252, 193, 68, 154, 252, 193, 130, 154, 252, 193, 174, 154, 252, 193, 232, 154, 252, 193, 102, 155, 252, 193, 104, 155, 252, 193, 106, 155, 252, 193, 108, 155, 252, 193, 110, 155, 252, 193, 44, 156, 252, 193, 46, 156, 252, 193, 48, 156, 252, 193, 50, 156, 252, 193, 52, 156, 252, 193, 54, 156, 252, 193, 56, 156, 252, 193, 246, 156, 252, 193, 248, 156, 252, 193, 250, 156, 252, 193, 252, 156, 252, 193, 254, 156, 252, 193, 2, 157, 252, 206, 224, 65, 0, 206, 224, 65, 0, 206, 224, 65, 0, 206, 224, 65, 0, 206, 224, 65, 0, 193, 236, 153, 252, 193, 238, 153, 252, 193, 240, 153, 252, 193, 242, 153, 252, 193, 244, 153, 252, 193, 246, 153, 252, 193, 248, 153, 252, 193, 250, 153, 252, 193, 14, 154, 252, 193, 16, 154, 252, 193, 46, 154, 252, 193, 18, 154, 252, 193, 20, 154, 252, 193, 48, 154, 252, 193, 50, 154, 252, 193, 52, 154, 252, 193, 54, 154, 252, 193, 56, 154, 252, 193, 58, 154, 252, 193, 60, 154, 252, 193, 62, 154, 252, 193, 96, 154, 252, 193, 64, 154, 252, 193, 78, 154, 252, 193, 80, 154, 252, 193, 82, 154, 252, 193, 84, 154, 252, 206, 0, 66, 0, 206, 0, 66, 0, 206, 0, 66, 0, 206, 0, 66, 0, 193, 76, 156, 252, 193, 238, 154, 252, 193, 240, 154, 252, 193, 242, 154, 252, 193, 244, 154, 252, 193, 246, 154, 252, 193, 248, 154, 252, 193, 250, 154, 252, 193, 252, 154, 252, 193, 254, 154, 252, 193, 2, 155, 252, 193, 4, 155, 252, 193, 6, 155, 252, 193, 8, 155, 252, 193, 10, 155, 252, 193, 12, 155, 252, 193, 14, 155, 252, 193, 16, 155, 252, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 193, 200, 154, 252, 193, 202, 154, 252, 193, 204, 154, 252, 193, 206, 154, 252, 193, 208, 154, 252, 193, 210, 154, 252, 193, 212, 154, 252, 193, 214, 154, 252, 193, 216, 154, 252, 193, 218, 154, 252, 193, 220, 154, 252, 193, 222, 154, 252, 193, 224, 154, 252, 193, 226, 154, 252, 193, 228, 154, 252, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 32, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 193, 112, 155, 252, 193, 114, 155, 252, 193, 116, 155, 252, 193, 118, 155, 252, 193, 58, 156, 252, 193, 60, 156, 252, 193, 62, 156, 252, 193, 8, 157, 252, 193, 10, 157, 252, 193, 12, 157, 252, 193, 14, 157, 252, 193, 16, 157, 252, 193, 18, 157, 252, 193, 20, 157, 252, 193, 22, 157, 252, 193, 24, 157, 252, 193, 26, 157, 252, 193, 32, 158, 252, 193, 34, 158, 252, 193, 36, 158, 252, 193, 38, 158, 252, 193, 80, 159, 252, 193, 82, 159, 252, 193, 84, 159, 252, 193, 86, 159, 252, 193, 88, 159, 252, 193, 90, 159, 252, 193, 92, 159, 252, 193, 94, 159, 252, 193, 96, 159, 252, 193, 98, 159, 252, 193, 100, 159, 252, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 206, 64, 66, 0, 193, 78, 156, 252, 193, 214, 155, 252, 193, 216, 155, 252, 193, 218, 155, 252, 193, 220, 155, 252, 193, 222, 155, 252, 193, 224, 155, 252, 193, 226, 155, 252, 193, 140, 156, 252, 193, 142, 156, 252, 193, 144, 156, 252, 193, 146, 156, 252, 193, 148, 156, 252, 193, 150, 156, 252, 193, 152, 156, 252, 193, 154, 156, 252, 193, 156, 156, 252, 193, 158, 156, 252, 193, 160, 156, 252, 193, 162, 156, 252, 193, 164, 156, 252, 193, 166, 156, 252, 193, 106, 157, 252, 193, 108, 157, 252, 193, 110, 157, 252, 193, 112, 157, 252, 193, 114, 157, 252, 193, 116, 157, 252, 193, 118, 157, 252, 193, 120, 157, 252, 193, 122, 157, 252, 193, 124, 157, 252, 193, 126, 157, 252, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 206, 96, 66, 0, 193, 168, 156, 252, 193, 170, 156, 252, 193, 172, 156, 252, 193, 174, 156, 252, 193, 176, 156, 252, 193, 178, 156, 252, 193, 180, 156, 252, 193, 182, 156, 252, 193, 184, 156, 252, 193, 186, 156, 252, 193, 188, 156, 252, 193, 190, 156, 252, 193, 192, 156, 252, 193, 194, 156, 252, 193, 196, 156, 252, 193, 198, 156, 252, 193, 200, 156, 252, 193, 202, 156, 252, 193, 204, 156, 252, 193, 206, 156, 252, 193, 208, 156, 252, 193, 34, 156, 252, 193, 80, 156, 252, 193, 82, 156, 252, 193, 84, 156, 252, 193, 86, 156, 252, 193, 88, 156, 252, 193, 90, 156, 252, 193, 92, 156, 252, 193, 94, 156, 252, 193, 96, 156, 252, 193, 98, 156, 252, 193, 100, 156, 252, 193, 102, 156, 252, 193, 104, 156, 252, 193, 106, 156, 252, 193, 108, 156, 252, 193, 110, 156, 252, 193, 112, 156, 252, 193, 114, 156, 252, 193, 116, 156, 252, 193, 118, 156, 252, 193, 120, 156, 252, 193, 122, 156, 252, 193, 124, 156, 252, 193, 126, 156, 252, 193, 128, 156, 252, 193, 130, 156, 252, 193, 132, 156, 252, 193, 134, 156, 252, 193, 136, 156, 252, 193, 138, 156, 252, 193, 34, 157, 252, 193, 36, 157, 252, 193, 210, 156, 252, 193, 212, 156, 252, 193, 214, 156, 252, 193, 216, 156, 252, 193, 218, 156, 252, 193, 220, 156, 252, 193, 222, 156, 252, 193, 224, 156, 252, 193, 226, 156, 252, 193, 228, 156, 252, 193, 230, 156, 252, 193, 232, 156, 252, 193, 234, 156, 252, 193, 236, 156, 252, 193, 238, 156, 252, 193, 240, 156, 252, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 193, 4, 157, 252, 193, 6, 157, 252, 193, 8, 158, 252, 193, 10, 158, 252, 193, 12, 158, 252, 193, 14, 158, 252, 193, 16, 158, 252, 193, 18, 158, 252, 193, 20, 158, 252, 193, 22, 158, 252, 193, 24, 158, 252, 193, 26, 158, 252, 193, 28, 158, 252, 193, 30, 158, 252, 193, 56, 159, 252, 193, 58, 159, 252, 193, 60, 159, 252, 193, 62, 159, 252, 193, 64, 159, 252, 193, 66, 159, 252, 193, 68, 159, 252, 193, 70, 159, 252, 193, 72, 159, 252, 193, 74, 159, 252, 193, 76, 159, 252, 193, 78, 159, 252, 193, 154, 160, 252, 193, 156, 160, 252, 193, 158, 160, 252, 193, 160, 160, 252, 193, 162, 160, 252, 193, 164, 160, 252, 193, 38, 157, 252, 193, 40, 157, 252, 193, 42, 157, 252, 193, 44, 157, 252, 193, 46, 157, 252, 193, 48, 157, 252, 193, 50, 157, 252, 193, 52, 157, 252, 193, 54, 157, 252, 193, 56, 157, 252, 193, 58, 157, 252, 193, 60, 157, 252, 193, 62, 157, 252, 193, 64, 157, 252, 193, 66, 157, 252, 193, 68, 157, 252, 193, 70, 157, 252, 193, 72, 157, 252, 193, 74, 157, 252, 193, 76, 157, 252, 193, 78, 157, 252, 193, 80, 157, 252, 193, 82, 157, 252, 193, 84, 157, 252, 193, 86, 157, 252, 193, 88, 157, 252, 193, 90, 157, 252, 193, 92, 157, 252, 193, 94, 157, 252, 193, 96, 157, 252, 193, 98, 157, 252, 193, 100, 157, 252, 193, 102, 157, 252, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 193, 104, 157, 252, 193, 156, 159, 252, 193, 158, 159, 252, 193, 160, 159, 252, 193, 162, 159, 252, 193, 164, 159, 252, 193, 166, 159, 252, 193, 168, 159, 252, 193, 170, 159, 252, 193, 218, 160, 252, 193, 172, 159, 252, 193, 174, 159, 252, 193, 176, 159, 252, 193, 178, 159, 252, 193, 180, 159, 252, 193, 182, 159, 252, 193, 184, 159, 252, 193, 186, 159, 252, 193, 188, 159, 252, 193, 190, 159, 252, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 193, 128, 157, 252, 193, 134, 158, 252, 193, 136, 158, 252, 193, 138, 158, 252, 193, 140, 158, 252, 193, 142, 158, 252, 193, 144, 158, 252, 193, 146, 158, 252, 193, 148, 158, 252, 193, 150, 158, 252, 193, 152, 158, 252, 193, 154, 158, 252, 193, 156, 158, 252, 193, 158, 158, 252, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 206, 128, 66, 0, 193, 160, 158, 252, 193, 242, 157, 252, 193, 244, 157, 252, 193, 246, 157, 252, 193, 248, 157, 252, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 193, 254, 157, 252, 193, 2, 158, 252, 193, 4, 158, 252, 193, 6, 158, 252, 193, 44, 159, 252, 193, 46, 159, 252, 193, 48, 159, 252, 193, 50, 159, 252, 193, 52, 159, 252, 193, 54, 159, 252, 193, 136, 160, 252, 193, 138, 160, 252, 193, 140, 160, 252, 193, 142, 160, 252, 193, 144, 160, 252, 193, 146, 160, 252, 193, 240, 161, 252, 193, 242, 161, 252, 193, 244, 161, 252, 193, 246, 161, 252, 193, 248, 161, 252, 193, 250, 161, 252, 193, 252, 161, 252, 193, 254, 161, 252, 193, 56, 163, 252, 193, 58, 163, 252, 193, 60, 163, 252, 193, 62, 163, 252, 193, 64, 163, 252, 193, 66, 163, 252, 193, 68, 163, 252, 193, 70, 163, 252, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 206, 160, 66, 0, 193, 116, 159, 252, 193, 118, 159, 252, 193, 120, 159, 252, 193, 122, 159, 252, 193, 124, 159, 252, 193, 126, 159, 252, 193, 128, 159, 252, 193, 130, 159, 252, 193, 132, 159, 252, 193, 134, 159, 252, 193, 136, 159, 252, 193, 138, 159, 252, 193, 140, 159, 252, 193, 142, 159, 252, 193, 144, 159, 252, 193, 146, 159, 252, 193, 148, 159, 252, 193, 150, 159, 252, 193, 152, 159, 252, 193, 154, 159, 252, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 192, 66, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 193, 102, 159, 252, 193, 170, 160, 252, 193, 172, 160, 252, 193, 174, 160, 252, 193, 176, 160, 252, 193, 178, 160, 252, 193, 180, 160, 252, 193, 182, 160, 252, 193, 184, 160, 252, 193, 186, 160, 252, 193, 188, 160, 252, 193, 190, 160, 252, 193, 192, 160, 252, 193, 194, 160, 252, 193, 196, 160, 252, 193, 16, 162, 252, 193, 18, 162, 252, 193, 20, 162, 252, 193, 22, 162, 252, 193, 24, 162, 252, 193, 26, 162, 252, 193, 28, 162, 252, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 193, 106, 159, 252, 193, 108, 159, 252, 193, 110, 159, 252, 193, 112, 159, 252, 193, 114, 159, 252, 193, 198, 160, 252, 193, 200, 160, 252, 193, 202, 160, 252, 193, 204, 160, 252, 193, 206, 160, 252, 193, 208, 160, 252, 193, 210, 160, 252, 193, 212, 160, 252, 193, 214, 160, 252, 193, 216, 160, 252, 193, 30, 162, 252, 193, 32, 162, 252, 193, 34, 162, 252, 193, 36, 162, 252, 193, 38, 162, 252, 193, 114, 163, 252, 193, 116, 163, 252, 193, 118, 163, 252, 193, 188, 164, 252, 193, 190, 164, 252, 193, 192, 164, 252, 193, 152, 165, 252, 193, 154, 165, 252, 193, 36, 166, 252, 193, 38, 166, 252, 193, 208, 166, 252, 193, 210, 166, 252, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 206, 224, 66, 0, 193, 26, 161, 252, 193, 28, 161, 252, 193, 30, 161, 252, 193, 32, 161, 252, 193, 34, 161, 252, 193, 36, 161, 252, 193, 38, 161, 252, 193, 40, 161, 252, 193, 42, 161, 252, 193, 44, 161, 252, 193, 46, 161, 252, 193, 48, 161, 252, 193, 50, 161, 252, 193, 52, 161, 252, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 0, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 193, 166, 160, 252, 193, 168, 160, 252, 193, 2, 162, 252, 193, 4, 162, 252, 193, 6, 162, 252, 193, 8, 162, 252, 193, 10, 162, 252, 193, 12, 162, 252, 193, 14, 162, 252, 193, 72, 163, 252, 193, 74, 163, 252, 193, 76, 163, 252, 193, 78, 163, 252, 193, 150, 164, 252, 193, 152, 164, 252, 193, 154, 164, 252, 193, 156, 164, 252, 193, 158, 164, 252, 193, 160, 164, 252, 193, 114, 165, 252, 193, 116, 165, 252, 193, 118, 165, 252, 193, 120, 165, 252, 193, 122, 165, 252, 193, 124, 165, 252, 193, 126, 165, 252, 193, 128, 165, 252, 193, 170, 166, 252, 193, 172, 166, 252, 193, 174, 166, 252, 193, 176, 166, 252, 193, 178, 166, 252, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 32, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 206, 96, 67, 0, 193, 22, 161, 252, 193, 106, 162, 252, 193, 108, 162, 252, 193, 110, 162, 252, 193, 24, 161, 252, 193, 112, 162, 252, 193, 114, 162, 252, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 64, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 206, 128, 67, 0, 193, 174, 163, 252, 193, 176, 163, 252, 193, 178, 163, 252, 193, 180, 163, 252, 193, 182, 163, 252, 193, 184, 163, 252, 193, 186, 163, 252, 193, 226, 164, 252, 193, 188, 163, 252, 193, 190, 163, 252, 193, 192, 163, 252, 193, 194, 163, 252, 193, 196, 163, 252, 193, 198, 163, 252, 193, 200, 163, 252, 193, 202, 163, 252, 193, 204, 163, 252, 193, 206, 163, 252, 193, 208, 163, 252, 193, 228, 164, 252, 193, 230, 164, 252, 193, 232, 164, 252, 193, 234, 164, 252, 193, 236, 164, 252, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 206, 160, 67, 0, 193, 212, 163, 252, 193, 46, 163, 252, 193, 48, 163, 252, 193, 50, 163, 252, 193, 52, 163, 252, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 193, 54, 163, 252, 193, 104, 164, 252, 193, 106, 164, 252, 193, 108, 164, 252, 193, 110, 164, 252, 193, 112, 164, 252, 193, 114, 164, 252, 193, 116, 164, 252, 193, 118, 164, 252, 193, 120, 164, 252, 193, 122, 164, 252, 193, 124, 164, 252, 193, 126, 164, 252, 193, 128, 164, 252, 193, 130, 164, 252, 193, 132, 164, 252, 193, 134, 164, 252, 193, 136, 164, 252, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 206, 192, 67, 0, 193, 162, 164, 252, 193, 164, 164, 252, 193, 166, 164, 252, 193, 168, 164, 252, 193, 170, 164, 252, 193, 172, 164, 252, 193, 174, 164, 252, 193, 176, 164, 252, 193, 178, 164, 252, 193, 180, 164, 252, 193, 182, 164, 252, 193, 184, 164, 252, 193, 186, 164, 252, 193, 130, 165, 252, 193, 132, 165, 252, 193, 134, 165, 252, 193, 136, 165, 252, 193, 138, 165, 252, 193, 140, 165, 252, 193, 142, 165, 252, 193, 144, 165, 252, 193, 146, 165, 252, 193, 148, 165, 252, 193, 150, 165, 252, 193, 18, 166, 252, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 206, 224, 67, 0, 193, 194, 164, 252, 193, 196, 164, 252, 193, 198, 164, 252, 193, 200, 164, 252, 193, 202, 164, 252, 193, 204, 164, 252, 193, 206, 164, 252, 193, 208, 164, 252, 193, 210, 164, 252, 193, 212, 164, 252, 193, 214, 164, 252, 193, 216, 164, 252, 193, 218, 164, 252, 193, 238, 164, 252, 193, 240, 164, 252, 193, 242, 164, 252, 193, 210, 163, 252, 193, 244, 164, 252, 193, 186, 165, 252, 193, 50, 166, 252, 193, 52, 166, 252, 193, 54, 166, 252, 193, 56, 166, 252, 193, 58, 166, 252, 193, 60, 166, 252, 193, 62, 166, 252, 193, 64, 166, 252, 193, 240, 166, 252, 193, 242, 166, 252, 193, 244, 166, 252, 193, 246, 166, 252, 193, 100, 167, 252, 193, 102, 167, 252, 193, 104, 167, 252, 193, 134, 162, 253, 193, 154, 167, 252, 193, 244, 167, 252, 193, 12, 168, 252, 193, 64, 168, 252, 193, 140, 168, 252, 193, 204, 168, 252, 193, 206, 168, 252, 193, 44, 169, 252, 193, 46, 169, 252, 193, 48, 169, 252, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 206, 0, 68, 0, 193, 138, 164, 252, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 193, 140, 164, 252, 193, 142, 164, 252, 193, 144, 164, 252, 193, 146, 164, 252, 193, 148, 164, 252, 193, 110, 165, 252, 193, 112, 165, 252, 193, 16, 166, 252, 193, 162, 166, 252, 193, 164, 166, 252, 193, 166, 166, 252, 193, 168, 166, 252, 193, 56, 167, 252, 193, 146, 167, 252, 193, 196, 167, 252, 193, 234, 167, 252, 193, 48, 168, 252, 193, 98, 168, 252, 193, 100, 168, 252, 193, 158, 168, 252, 193, 160, 168, 252, 193, 236, 168, 252, 193, 238, 168, 252, 193, 240, 168, 252, 193, 102, 169, 252, 193, 104, 169, 252, 193, 106, 169, 252, 193, 108, 169, 252, 193, 110, 169, 252, 193, 240, 169, 252, 193, 242, 169, 252, 193, 244, 169, 252, 193, 220, 164, 252, 193, 222, 164, 252, 193, 224, 164, 252, 193, 156, 165, 252, 193, 158, 165, 252, 193, 160, 165, 252, 193, 162, 165, 252, 193, 164, 165, 252, 193, 166, 165, 252, 193, 168, 165, 252, 193, 170, 165, 252, 193, 172, 165, 252, 193, 174, 165, 252, 193, 176, 165, 252, 193, 178, 165, 252, 193, 180, 165, 252, 193, 40, 166, 252, 193, 42, 166, 252, 193, 44, 166, 252, 193, 46, 166, 252, 193, 48, 166, 252, 193, 182, 165, 252, 193, 184, 165, 252, 193, 218, 166, 252, 193, 220, 166, 252, 193, 222, 166, 252, 193, 224, 166, 252, 193, 226, 166, 252, 193, 228, 166, 252, 193, 230, 166, 252, 193, 232, 166, 252, 193, 234, 166, 252, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 32, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 64, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 193, 20, 166, 252, 193, 22, 166, 252, 193, 24, 166, 252, 193, 26, 166, 252, 193, 28, 166, 252, 193, 30, 166, 252, 193, 32, 166, 252, 193, 34, 166, 252, 193, 180, 166, 252, 193, 182, 166, 252, 193, 184, 166, 252, 193, 186, 166, 252, 193, 188, 166, 252, 193, 190, 166, 252, 193, 192, 166, 252, 193, 194, 166, 252, 193, 196, 166, 252, 193, 198, 166, 252, 193, 200, 166, 252, 193, 202, 166, 252, 193, 204, 166, 252, 193, 206, 166, 252, 193, 70, 167, 252, 193, 72, 167, 252, 193, 74, 167, 252, 193, 76, 167, 252, 193, 78, 167, 252, 193, 80, 167, 252, 193, 82, 167, 252, 193, 84, 167, 252, 193, 236, 167, 252, 193, 238, 167, 252, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 96, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 193, 212, 166, 252, 193, 214, 166, 252, 193, 216, 166, 252, 193, 150, 167, 252, 193, 10, 168, 252, 193, 36, 168, 252, 193, 112, 168, 252, 193, 166, 168, 252, 193, 168, 168, 252, 193, 170, 168, 252, 193, 172, 168, 252, 193, 4, 169, 252, 193, 6, 169, 252, 193, 8, 169, 252, 193, 10, 169, 252, 193, 12, 169, 252, 193, 14, 169, 252, 193, 140, 169, 252, 193, 142, 169, 252, 193, 144, 169, 252, 193, 146, 169, 252, 193, 148, 169, 252, 193, 24, 170, 252, 193, 26, 170, 252, 193, 28, 170, 252, 193, 30, 170, 252, 193, 32, 170, 252, 193, 34, 170, 252, 193, 36, 170, 252, 193, 168, 170, 252, 193, 64, 171, 252, 193, 66, 171, 252, 193, 236, 166, 252, 193, 238, 166, 252, 193, 86, 167, 252, 193, 88, 167, 252, 193, 90, 167, 252, 193, 92, 167, 252, 193, 94, 167, 252, 193, 96, 167, 252, 193, 98, 167, 252, 193, 200, 167, 252, 193, 152, 167, 252, 193, 202, 167, 252, 193, 204, 167, 252, 193, 206, 167, 252, 193, 208, 167, 252, 193, 240, 167, 252, 193, 242, 167, 252, 193, 62, 168, 252, 193, 72, 168, 252, 193, 76, 168, 252, 193, 84, 168, 252, 193, 86, 168, 252, 193, 88, 168, 252, 193, 116, 168, 252, 193, 118, 168, 252, 193, 120, 168, 252, 193, 122, 168, 252, 193, 124, 168, 252, 193, 126, 168, 252, 193, 128, 168, 252, 193, 130, 168, 252, 193, 132, 168, 252, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 128, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 193, 58, 167, 252, 193, 60, 167, 252, 193, 62, 167, 252, 193, 64, 167, 252, 193, 66, 167, 252, 193, 68, 167, 252, 193, 148, 167, 252, 193, 198, 167, 252, 193, 28, 168, 252, 193, 30, 168, 252, 193, 32, 168, 252, 193, 82, 168, 252, 193, 102, 168, 252, 193, 104, 168, 252, 193, 106, 168, 252, 193, 162, 168, 252, 193, 242, 168, 252, 193, 244, 168, 252, 193, 246, 168, 252, 193, 112, 169, 252, 193, 114, 169, 252, 193, 116, 169, 252, 193, 118, 169, 252, 193, 120, 169, 252, 193, 122, 169, 252, 193, 248, 169, 252, 193, 250, 169, 252, 193, 252, 169, 252, 193, 254, 169, 252, 193, 146, 170, 252, 193, 148, 170, 252, 193, 150, 170, 252, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 206, 160, 68, 0, 193, 156, 167, 252, 193, 158, 167, 252, 193, 160, 167, 252, 193, 162, 167, 252, 193, 164, 167, 252, 193, 166, 167, 252, 193, 168, 167, 252, 193, 170, 167, 252, 193, 172, 167, 252, 193, 174, 167, 252, 193, 176, 167, 252, 193, 178, 167, 252, 193, 180, 167, 252, 193, 182, 167, 252, 193, 184, 167, 252, 193, 186, 167, 252, 193, 188, 167, 252, 193, 190, 167, 252, 193, 192, 167, 252, 193, 194, 167, 252, 206, 192, 68, 0, 206, 192, 68, 0, 206, 192, 68, 0, 206, 192, 68, 0, 206, 192, 68, 0, 206, 192, 68, 0, 206, 192, 68, 0, 206, 192, 68, 0, 193, 246, 167, 252, 193, 248, 167, 252, 193, 250, 167, 252, 193, 252, 167, 252, 193, 254, 167, 252, 193, 2, 168, 252, 193, 4, 168, 252, 193, 6, 168, 252, 193, 14, 168, 252, 193, 16, 168, 252, 193, 18, 168, 252, 193, 20, 168, 252, 193, 22, 168, 252, 193, 24, 168, 252, 193, 26, 168, 252, 193, 38, 168, 252, 193, 50, 168, 252, 193, 52, 168, 252, 193, 40, 168, 252, 193, 42, 168, 252, 193, 44, 168, 252, 193, 46, 168, 252, 193, 54, 168, 252, 193, 56, 168, 252, 193, 8, 168, 252, 193, 34, 168, 252, 193, 60, 168, 252, 193, 108, 168, 252, 193, 110, 168, 252, 193, 164, 168, 252, 193, 248, 168, 252, 193, 250, 168, 252, 193, 252, 168, 252, 193, 254, 168, 252, 193, 2, 169, 252, 193, 124, 169, 252, 193, 126, 169, 252, 193, 128, 169, 252, 193, 130, 169, 252, 193, 132, 169, 252, 193, 134, 169, 252, 193, 136, 169, 252, 193, 138, 169, 252, 193, 2, 170, 252, 193, 4, 170, 252, 193, 6, 170, 252, 193, 8, 170, 252, 193, 10, 170, 252, 193, 12, 170, 252, 193, 14, 170, 252, 193, 16, 170, 252, 193, 18, 170, 252, 193, 20, 170, 252, 193, 22, 170, 252, 193, 154, 170, 252, 193, 156, 170, 252, 193, 58, 168, 252, 193, 66, 168, 252, 193, 68, 168, 252, 193, 70, 168, 252, 193, 74, 168, 252, 193, 78, 168, 252, 193, 80, 168, 252, 193, 90, 168, 252, 193, 92, 168, 252, 193, 94, 168, 252, 193, 96, 168, 252, 193, 142, 168, 252, 193, 144, 168, 252, 193, 146, 168, 252, 193, 148, 168, 252, 193, 150, 168, 252, 193, 152, 168, 252, 193, 154, 168, 252, 193, 156, 168, 252, 193, 208, 168, 252, 193, 210, 168, 252, 193, 212, 168, 252, 193, 214, 168, 252, 193, 216, 168, 252, 193, 218, 168, 252, 193, 220, 168, 252, 193, 222, 168, 252, 193, 224, 168, 252, 193, 226, 168, 252, 193, 228, 168, 252, 193, 230, 168, 252, 193, 232, 168, 252, 193, 134, 168, 252, 193, 136, 168, 252, 193, 138, 168, 252, 193, 174, 168, 252, 193, 176, 168, 252, 193, 178, 168, 252, 193, 180, 168, 252, 193, 182, 168, 252, 193, 184, 168, 252, 193, 186, 168, 252, 193, 188, 168, 252, 193, 190, 168, 252, 193, 192, 168, 252, 193, 194, 168, 252, 193, 196, 168, 252, 193, 198, 168, 252, 193, 200, 168, 252, 193, 202, 168, 252, 193, 18, 169, 252, 193, 20, 169, 252, 193, 22, 169, 252, 193, 24, 169, 252, 193, 26, 169, 252, 193, 28, 169, 252, 193, 30, 169, 252, 193, 32, 169, 252, 193, 34, 169, 252, 193, 36, 169, 252, 193, 38, 169, 252, 193, 40, 169, 252, 193, 150, 169, 252, 193, 152, 169, 252, 193, 234, 168, 252, 193, 52, 169, 252, 193, 54, 169, 252, 193, 56, 169, 252, 193, 58, 169, 252, 193, 60, 169, 252, 193, 62, 169, 252, 193, 64, 169, 252, 193, 66, 169, 252, 193, 68, 169, 252, 193, 70, 169, 252, 193, 72, 169, 252, 193, 74, 169, 252, 193, 76, 169, 252, 193, 78, 169, 252, 193, 80, 169, 252, 193, 82, 169, 252, 193, 84, 169, 252, 193, 86, 169, 252, 193, 88, 169, 252, 193, 90, 169, 252, 193, 92, 169, 252, 193, 94, 169, 252, 193, 192, 169, 252, 193, 96, 169, 252, 193, 98, 169, 252, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 193, 154, 169, 252, 193, 156, 169, 252, 193, 158, 169, 252, 193, 42, 169, 252, 193, 160, 169, 252, 193, 162, 169, 252, 193, 164, 169, 252, 193, 166, 169, 252, 193, 168, 169, 252, 193, 170, 169, 252, 193, 172, 169, 252, 193, 174, 169, 252, 193, 176, 169, 252, 193, 178, 169, 252, 193, 38, 170, 252, 193, 40, 170, 252, 193, 42, 170, 252, 193, 44, 170, 252, 193, 46, 170, 252, 193, 48, 170, 252, 193, 50, 170, 252, 193, 52, 170, 252, 193, 170, 170, 252, 193, 172, 170, 252, 193, 174, 170, 252, 193, 176, 170, 252, 193, 178, 170, 252, 193, 180, 170, 252, 193, 182, 170, 252, 193, 184, 170, 252, 193, 186, 170, 252, 193, 188, 170, 252, 193, 50, 169, 252, 193, 180, 169, 252, 193, 182, 169, 252, 193, 184, 169, 252, 193, 186, 169, 252, 193, 188, 169, 252, 193, 190, 169, 252, 193, 56, 170, 252, 193, 58, 170, 252, 193, 60, 170, 252, 193, 62, 170, 252, 193, 64, 170, 252, 193, 66, 170, 252, 193, 192, 170, 252, 193, 194, 170, 252, 193, 196, 170, 252, 193, 198, 170, 252, 193, 200, 170, 252, 193, 202, 170, 252, 193, 98, 171, 252, 193, 100, 171, 252, 193, 102, 171, 252, 193, 104, 171, 252, 193, 106, 171, 252, 193, 108, 171, 252, 193, 110, 171, 252, 193, 244, 171, 252, 193, 246, 171, 252, 193, 122, 172, 252, 193, 124, 172, 252, 193, 126, 172, 252, 193, 128, 172, 252, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 206, 224, 68, 0, 193, 100, 169, 252, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 193, 246, 169, 252, 193, 144, 170, 252, 193, 32, 171, 252, 193, 34, 171, 252, 193, 36, 171, 252, 193, 38, 171, 252, 193, 184, 171, 252, 193, 186, 171, 252, 193, 188, 171, 252, 193, 52, 172, 252, 193, 54, 172, 252, 193, 56, 172, 252, 193, 58, 172, 252, 193, 184, 172, 252, 193, 186, 172, 252, 193, 42, 173, 252, 193, 44, 173, 252, 193, 46, 173, 252, 193, 134, 173, 252, 193, 136, 173, 252, 193, 138, 173, 252, 193, 140, 173, 252, 193, 174, 173, 252, 193, 244, 173, 252, 193, 246, 173, 252, 193, 248, 173, 252, 193, 102, 174, 252, 193, 104, 174, 252, 193, 162, 174, 252, 193, 164, 174, 252, 193, 96, 175, 252, 193, 32, 176, 252, 193, 54, 170, 252, 193, 190, 170, 252, 193, 72, 171, 252, 193, 74, 171, 252, 193, 76, 171, 252, 193, 78, 171, 252, 193, 80, 171, 252, 193, 82, 171, 252, 193, 84, 171, 252, 193, 86, 171, 252, 193, 88, 171, 252, 193, 90, 171, 252, 193, 92, 171, 252, 193, 94, 171, 252, 193, 96, 171, 252, 193, 222, 171, 252, 193, 224, 171, 252, 193, 226, 171, 252, 193, 228, 171, 252, 193, 230, 171, 252, 193, 232, 171, 252, 193, 234, 171, 252, 193, 236, 171, 252, 193, 238, 171, 252, 193, 240, 171, 252, 193, 100, 172, 252, 193, 102, 172, 252, 193, 104, 172, 252, 193, 106, 172, 252, 193, 108, 172, 252, 193, 110, 172, 252, 193, 112, 172, 252, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 206, 0, 69, 0, 193, 204, 170, 252, 193, 206, 170, 252, 193, 208, 170, 252, 193, 210, 170, 252, 193, 212, 170, 252, 193, 214, 170, 252, 193, 216, 170, 252, 193, 218, 170, 252, 193, 152, 170, 252, 193, 40, 171, 252, 193, 42, 171, 252, 193, 44, 171, 252, 193, 46, 171, 252, 193, 48, 171, 252, 193, 190, 171, 252, 193, 192, 171, 252, 193, 194, 171, 252, 193, 196, 171, 252, 193, 198, 171, 252, 193, 60, 172, 252, 193, 62, 172, 252, 193, 64, 172, 252, 193, 66, 172, 252, 193, 68, 172, 252, 193, 70, 172, 252, 193, 72, 172, 252, 193, 74, 172, 252, 193, 190, 172, 252, 193, 192, 172, 252, 193, 194, 172, 252, 193, 48, 173, 252, 193, 50, 173, 252, 193, 52, 173, 252, 193, 54, 173, 252, 193, 56, 173, 252, 193, 142, 173, 252, 193, 176, 173, 252, 193, 178, 173, 252, 193, 180, 173, 252, 193, 214, 173, 252, 193, 158, 170, 252, 193, 160, 170, 252, 193, 162, 170, 252, 193, 164, 170, 252, 193, 166, 170, 252, 193, 50, 171, 252, 193, 52, 171, 252, 193, 54, 171, 252, 193, 56, 171, 252, 193, 58, 171, 252, 193, 60, 171, 252, 193, 62, 171, 252, 193, 200, 171, 252, 193, 202, 171, 252, 193, 204, 171, 252, 193, 206, 171, 252, 193, 208, 171, 252, 193, 210, 171, 252, 193, 212, 171, 252, 193, 76, 172, 252, 193, 78, 172, 252, 193, 80, 172, 252, 193, 82, 172, 252, 193, 84, 172, 252, 193, 86, 172, 252, 193, 88, 172, 252, 193, 196, 172, 252, 193, 58, 173, 252, 193, 60, 173, 252, 193, 144, 173, 252, 193, 146, 173, 252, 193, 216, 173, 252, 193, 220, 170, 252, 193, 222, 170, 252, 193, 224, 170, 252, 193, 226, 170, 252, 193, 228, 170, 252, 193, 230, 170, 252, 193, 232, 170, 252, 193, 234, 170, 252, 193, 236, 170, 252, 193, 238, 170, 252, 193, 240, 170, 252, 193, 242, 170, 252, 193, 244, 170, 252, 193, 246, 170, 252, 193, 248, 170, 252, 193, 250, 170, 252, 193, 252, 170, 252, 193, 254, 170, 252, 193, 2, 171, 252, 193, 4, 171, 252, 193, 6, 171, 252, 193, 8, 171, 252, 193, 10, 171, 252, 193, 12, 171, 252, 193, 14, 171, 252, 193, 16, 171, 252, 193, 18, 171, 252, 193, 20, 171, 252, 193, 22, 171, 252, 193, 24, 171, 252, 193, 26, 171, 252, 193, 28, 171, 252, 193, 30, 171, 252, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 193, 68, 171, 252, 193, 70, 171, 252, 193, 214, 171, 252, 193, 216, 171, 252, 193, 218, 171, 252, 193, 220, 171, 252, 193, 90, 172, 252, 193, 92, 172, 252, 193, 94, 172, 252, 193, 96, 172, 252, 193, 98, 172, 252, 193, 198, 172, 252, 193, 200, 172, 252, 193, 202, 172, 252, 193, 204, 172, 252, 193, 206, 172, 252, 193, 208, 172, 252, 193, 62, 173, 252, 193, 64, 173, 252, 193, 66, 173, 252, 193, 182, 173, 252, 193, 184, 173, 252, 193, 254, 173, 252, 193, 106, 174, 252, 193, 174, 174, 252, 193, 176, 174, 252, 193, 108, 175, 252, 193, 110, 175, 252, 193, 112, 175, 252, 193, 114, 175, 252, 193, 56, 176, 252, 193, 58, 176, 252, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 206, 32, 69, 0, 193, 252, 171, 252, 193, 254, 171, 252, 193, 2, 172, 252, 193, 4, 172, 252, 193, 6, 172, 252, 193, 8, 172, 252, 193, 10, 172, 252, 193, 12, 172, 252, 193, 14, 172, 252, 193, 16, 172, 252, 193, 18, 172, 252, 193, 20, 172, 252, 193, 22, 172, 252, 193, 24, 172, 252, 193, 26, 172, 252, 193, 28, 172, 252, 193, 30, 172, 252, 193, 32, 172, 252, 193, 34, 172, 252, 193, 36, 172, 252, 193, 38, 172, 252, 193, 40, 172, 252, 193, 42, 172, 252, 193, 44, 172, 252, 193, 46, 172, 252, 193, 48, 172, 252, 193, 50, 172, 252, 193, 136, 172, 252, 193, 248, 171, 252, 193, 102, 133, 253, 193, 104, 133, 253, 193, 180, 133, 253, 193, 220, 133, 253, 193, 2, 134, 253, 193, 4, 134, 253, 193, 6, 134, 253, 193, 80, 134, 253, 193, 148, 134, 253, 193, 172, 134, 253, 193, 198, 134, 253, 193, 240, 134, 253, 193, 242, 134, 253, 193, 60, 135, 253, 193, 118, 135, 253, 193, 120, 135, 253, 193, 122, 135, 253, 193, 124, 135, 253, 193, 126, 135, 253, 193, 128, 135, 253, 193, 130, 135, 253, 193, 132, 135, 253, 193, 202, 135, 253, 193, 204, 135, 253, 193, 206, 135, 253, 193, 64, 136, 253, 193, 66, 136, 253, 193, 68, 136, 253, 193, 70, 136, 253, 193, 202, 136, 253, 193, 204, 136, 253, 193, 114, 172, 252, 193, 116, 172, 252, 193, 118, 172, 252, 193, 210, 172, 252, 193, 212, 172, 252, 193, 214, 172, 252, 193, 216, 172, 252, 193, 218, 172, 252, 193, 120, 172, 252, 193, 68, 173, 252, 193, 70, 173, 252, 193, 72, 173, 252, 193, 74, 173, 252, 193, 76, 173, 252, 193, 148, 173, 252, 193, 150, 173, 252, 193, 152, 173, 252, 193, 218, 173, 252, 193, 186, 173, 252, 193, 220, 173, 252, 193, 222, 173, 252, 193, 2, 174, 252, 193, 30, 174, 252, 193, 58, 174, 252, 193, 70, 174, 252, 193, 72, 174, 252, 193, 74, 174, 252, 193, 76, 174, 252, 193, 78, 174, 252, 193, 80, 174, 252, 193, 82, 174, 252, 193, 84, 174, 252, 193, 130, 172, 252, 193, 132, 172, 252, 193, 134, 172, 252, 193, 220, 172, 252, 193, 222, 172, 252, 193, 224, 172, 252, 193, 226, 172, 252, 193, 228, 172, 252, 193, 230, 172, 252, 193, 232, 172, 252, 193, 78, 173, 252, 193, 80, 173, 252, 193, 82, 173, 252, 193, 154, 173, 252, 193, 156, 173, 252, 193, 188, 173, 252, 193, 224, 173, 252, 193, 4, 174, 252, 193, 6, 174, 252, 193, 32, 174, 252, 193, 34, 174, 252, 193, 86, 174, 252, 193, 128, 174, 252, 193, 130, 174, 252, 193, 132, 174, 252, 193, 134, 174, 252, 206, 160, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 193, 138, 172, 252, 193, 140, 172, 252, 193, 142, 172, 252, 193, 144, 172, 252, 193, 146, 172, 252, 193, 148, 172, 252, 193, 150, 172, 252, 193, 152, 172, 252, 193, 154, 172, 252, 193, 156, 172, 252, 193, 158, 172, 252, 193, 160, 172, 252, 193, 162, 172, 252, 193, 164, 172, 252, 193, 166, 172, 252, 193, 168, 172, 252, 193, 170, 172, 252, 193, 172, 172, 252, 193, 174, 172, 252, 193, 176, 172, 252, 193, 178, 172, 252, 193, 180, 172, 252, 193, 182, 172, 252, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 193, 188, 172, 252, 193, 130, 180, 252, 193, 106, 183, 252, 193, 198, 186, 252, 193, 28, 187, 252, 193, 126, 187, 252, 193, 232, 187, 252, 193, 182, 188, 252, 193, 58, 193, 252, 193, 102, 200, 252, 193, 92, 211, 252, 193, 110, 217, 252, 193, 208, 225, 252, 193, 210, 225, 252, 193, 104, 227, 252, 193, 134, 229, 252, 193, 136, 229, 252, 193, 138, 229, 252, 193, 20, 235, 252, 193, 22, 235, 252, 193, 198, 237, 252, 193, 110, 240, 252, 193, 214, 242, 252, 193, 162, 12, 253, 193, 252, 15, 253, 193, 26, 16, 253, 193, 66, 19, 253, 193, 198, 24, 253, 193, 102, 27, 253, 193, 2, 40, 253, 193, 22, 41, 253, 193, 96, 41, 253, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 64, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 206, 96, 69, 0, 193, 158, 173, 252, 193, 160, 173, 252, 193, 162, 173, 252, 193, 164, 173, 252, 193, 166, 173, 252, 193, 168, 173, 252, 193, 170, 173, 252, 193, 172, 173, 252, 193, 190, 173, 252, 193, 192, 173, 252, 193, 194, 173, 252, 193, 196, 173, 252, 193, 198, 173, 252, 193, 200, 173, 252, 193, 202, 173, 252, 193, 204, 173, 252, 193, 206, 173, 252, 193, 208, 173, 252, 193, 226, 173, 252, 193, 210, 173, 252, 193, 212, 173, 252, 193, 228, 173, 252, 193, 230, 173, 252, 193, 232, 173, 252, 193, 234, 173, 252, 193, 236, 173, 252, 193, 238, 173, 252, 193, 240, 173, 252, 193, 242, 173, 252, 193, 8, 174, 252, 193, 10, 174, 252, 193, 12, 174, 252, 193, 14, 174, 252, 193, 16, 174, 252, 193, 18, 174, 252, 193, 20, 174, 252, 193, 22, 174, 252, 193, 24, 174, 252, 193, 26, 174, 252, 193, 36, 174, 252, 193, 38, 174, 252, 193, 40, 174, 252, 193, 42, 174, 252, 193, 44, 174, 252, 193, 48, 174, 252, 193, 50, 174, 252, 193, 52, 174, 252, 193, 54, 174, 252, 193, 56, 174, 252, 193, 250, 173, 252, 193, 252, 173, 252, 193, 28, 174, 252, 193, 166, 174, 252, 193, 168, 174, 252, 193, 98, 175, 252, 193, 100, 175, 252, 193, 38, 176, 252, 193, 40, 176, 252, 193, 42, 176, 252, 193, 44, 176, 252, 193, 8, 177, 252, 193, 10, 177, 252, 193, 12, 177, 252, 193, 14, 177, 252, 193, 16, 177, 252, 193, 18, 177, 252, 193, 20, 177, 252, 193, 22, 177, 252, 193, 24, 177, 252, 193, 26, 177, 252, 193, 252, 177, 252, 193, 254, 177, 252, 193, 2, 178, 252, 193, 4, 178, 252, 193, 6, 178, 252, 193, 8, 178, 252, 193, 10, 178, 252, 193, 12, 178, 252, 193, 14, 178, 252, 193, 16, 178, 252, 193, 52, 179, 252, 193, 46, 174, 252, 193, 170, 174, 252, 193, 172, 174, 252, 193, 102, 175, 252, 193, 104, 175, 252, 193, 106, 175, 252, 193, 46, 176, 252, 193, 48, 176, 252, 193, 50, 176, 252, 193, 52, 176, 252, 193, 54, 176, 252, 193, 28, 177, 252, 193, 30, 177, 252, 193, 32, 177, 252, 193, 34, 177, 252, 193, 36, 177, 252, 193, 38, 177, 252, 193, 18, 178, 252, 193, 20, 178, 252, 193, 22, 178, 252, 193, 24, 178, 252, 193, 26, 178, 252, 193, 70, 179, 252, 193, 72, 179, 252, 193, 74, 179, 252, 193, 76, 179, 252, 193, 78, 179, 252, 193, 80, 179, 252, 193, 82, 179, 252, 193, 84, 179, 252, 193, 86, 179, 252, 193, 88, 179, 252, 193, 60, 174, 252, 193, 62, 174, 252, 193, 64, 174, 252, 193, 66, 174, 252, 193, 68, 174, 252, 193, 88, 174, 252, 193, 90, 174, 252, 193, 92, 174, 252, 193, 94, 174, 252, 193, 96, 174, 252, 193, 98, 174, 252, 193, 100, 174, 252, 193, 136, 174, 252, 193, 138, 174, 252, 193, 140, 174, 252, 193, 142, 174, 252, 193, 144, 174, 252, 193, 146, 174, 252, 193, 148, 174, 252, 193, 150, 174, 252, 193, 152, 174, 252, 193, 154, 174, 252, 193, 156, 174, 252, 193, 158, 174, 252, 193, 160, 174, 252, 193, 12, 175, 252, 193, 14, 175, 252, 193, 16, 175, 252, 193, 18, 175, 252, 193, 20, 175, 252, 193, 22, 175, 252, 193, 24, 175, 252, 193, 108, 174, 252, 193, 110, 174, 252, 193, 112, 174, 252, 193, 114, 174, 252, 193, 116, 174, 252, 193, 118, 174, 252, 193, 120, 174, 252, 193, 122, 174, 252, 193, 124, 174, 252, 193, 126, 174, 252, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 128, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 206, 160, 69, 0, 193, 190, 175, 252, 193, 192, 175, 252, 193, 194, 175, 252, 193, 196, 175, 252, 193, 198, 175, 252, 193, 200, 175, 252, 193, 202, 175, 252, 193, 204, 175, 252, 193, 206, 175, 252, 193, 208, 175, 252, 193, 210, 175, 252, 193, 212, 175, 252, 193, 214, 175, 252, 193, 148, 176, 252, 193, 150, 176, 252, 193, 152, 176, 252, 193, 154, 176, 252, 193, 156, 176, 252, 193, 158, 176, 252, 193, 160, 176, 252, 193, 162, 176, 252, 193, 164, 176, 252, 193, 166, 176, 252, 193, 122, 177, 252, 193, 26, 175, 252, 193, 28, 175, 252, 193, 30, 175, 252, 193, 32, 175, 252, 193, 34, 175, 252, 193, 36, 175, 252, 193, 38, 175, 252, 193, 40, 175, 252, 193, 42, 175, 252, 193, 44, 175, 252, 193, 46, 175, 252, 193, 48, 175, 252, 193, 50, 175, 252, 193, 52, 175, 252, 193, 54, 175, 252, 193, 56, 175, 252, 193, 58, 175, 252, 193, 60, 175, 252, 193, 62, 175, 252, 193, 64, 175, 252, 193, 66, 175, 252, 193, 68, 175, 252, 193, 70, 175, 252, 193, 72, 175, 252, 193, 74, 175, 252, 193, 76, 175, 252, 193, 78, 175, 252, 193, 80, 175, 252, 193, 82, 175, 252, 193, 84, 175, 252, 193, 86, 175, 252, 193, 88, 175, 252, 193, 90, 175, 252, 193, 92, 175, 252, 193, 94, 175, 252, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 206, 192, 69, 0, 193, 64, 176, 252, 193, 66, 176, 252, 193, 68, 176, 252, 193, 70, 176, 252, 193, 72, 176, 252, 193, 74, 176, 252, 193, 76, 176, 252, 193, 78, 176, 252, 193, 80, 176, 252, 193, 82, 176, 252, 193, 84, 176, 252, 193, 86, 176, 252, 193, 88, 176, 252, 193, 90, 176, 252, 193, 46, 177, 252, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 224, 69, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 193, 34, 176, 252, 193, 36, 176, 252, 193, 2, 177, 252, 193, 4, 177, 252, 193, 6, 177, 252, 193, 242, 177, 252, 193, 244, 177, 252, 193, 246, 177, 252, 193, 248, 177, 252, 193, 250, 177, 252, 193, 34, 179, 252, 193, 36, 179, 252, 193, 38, 179, 252, 193, 40, 179, 252, 193, 42, 179, 252, 193, 44, 179, 252, 193, 46, 179, 252, 193, 48, 179, 252, 193, 50, 179, 252, 193, 120, 180, 252, 193, 122, 180, 252, 193, 124, 180, 252, 193, 126, 180, 252, 193, 128, 180, 252, 193, 146, 181, 252, 193, 148, 181, 252, 193, 150, 181, 252, 193, 152, 181, 252, 193, 154, 181, 252, 193, 122, 182, 252, 193, 124, 182, 252, 193, 126, 182, 252, 193, 60, 176, 252, 193, 62, 176, 252, 193, 40, 177, 252, 193, 42, 177, 252, 193, 44, 177, 252, 193, 28, 178, 252, 193, 30, 178, 252, 193, 32, 178, 252, 193, 34, 178, 252, 193, 96, 179, 252, 193, 98, 179, 252, 193, 100, 179, 252, 193, 102, 179, 252, 193, 104, 179, 252, 193, 174, 180, 252, 193, 176, 180, 252, 193, 180, 181, 252, 193, 168, 182, 252, 193, 170, 182, 252, 193, 134, 183, 252, 193, 136, 183, 252, 193, 4, 185, 252, 193, 6, 185, 252, 193, 8, 185, 252, 193, 80, 185, 252, 193, 82, 185, 252, 193, 120, 185, 252, 193, 210, 185, 252, 193, 228, 185, 252, 193, 60, 186, 252, 193, 62, 186, 252, 193, 64, 186, 252, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 0, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 32, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 64, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 193, 120, 177, 252, 193, 140, 178, 252, 193, 142, 178, 252, 193, 106, 179, 252, 193, 144, 178, 252, 193, 146, 178, 252, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 193, 124, 177, 252, 193, 126, 177, 252, 193, 128, 177, 252, 193, 130, 177, 252, 193, 132, 177, 252, 193, 134, 177, 252, 193, 136, 177, 252, 193, 148, 178, 252, 193, 150, 178, 252, 193, 152, 178, 252, 193, 154, 178, 252, 193, 156, 178, 252, 193, 158, 178, 252, 193, 160, 178, 252, 193, 162, 178, 252, 193, 164, 178, 252, 193, 166, 178, 252, 193, 168, 178, 252, 193, 170, 178, 252, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 96, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 128, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 206, 160, 70, 0, 193, 54, 179, 252, 193, 56, 179, 252, 193, 58, 179, 252, 193, 60, 179, 252, 193, 62, 179, 252, 193, 64, 179, 252, 193, 66, 179, 252, 193, 68, 179, 252, 193, 132, 180, 252, 193, 134, 180, 252, 193, 136, 180, 252, 193, 138, 180, 252, 193, 140, 180, 252, 193, 142, 180, 252, 193, 144, 180, 252, 193, 156, 181, 252, 193, 158, 181, 252, 193, 160, 181, 252, 193, 162, 181, 252, 193, 130, 182, 252, 193, 132, 182, 252, 193, 134, 182, 252, 193, 136, 182, 252, 193, 138, 182, 252, 193, 140, 182, 252, 193, 142, 182, 252, 193, 144, 182, 252, 193, 108, 183, 252, 193, 110, 183, 252, 193, 112, 183, 252, 193, 114, 183, 252, 193, 116, 183, 252, 193, 90, 179, 252, 193, 92, 179, 252, 193, 94, 179, 252, 193, 146, 180, 252, 193, 148, 180, 252, 193, 150, 180, 252, 193, 152, 180, 252, 193, 154, 180, 252, 193, 156, 180, 252, 193, 158, 180, 252, 193, 160, 180, 252, 193, 162, 180, 252, 193, 164, 180, 252, 193, 166, 180, 252, 193, 168, 180, 252, 193, 170, 180, 252, 193, 172, 180, 252, 193, 164, 181, 252, 193, 166, 181, 252, 193, 168, 181, 252, 193, 170, 181, 252, 193, 172, 181, 252, 193, 174, 181, 252, 193, 176, 181, 252, 193, 178, 181, 252, 206, 128, 71, 0, 206, 128, 71, 0, 206, 128, 71, 0, 206, 128, 71, 0, 206, 128, 71, 0, 206, 128, 71, 0, 206, 128, 71, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 192, 70, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 206, 224, 70, 0, 193, 244, 180, 252, 193, 246, 180, 252, 193, 248, 180, 252, 193, 250, 180, 252, 193, 252, 180, 252, 193, 254, 180, 252, 193, 2, 181, 252, 193, 252, 181, 252, 193, 254, 181, 252, 193, 2, 182, 252, 193, 4, 182, 252, 193, 6, 182, 252, 193, 8, 182, 252, 193, 10, 182, 252, 193, 12, 182, 252, 193, 14, 182, 252, 193, 16, 182, 252, 193, 226, 182, 252, 193, 228, 182, 252, 193, 230, 182, 252, 193, 232, 182, 252, 193, 234, 182, 252, 193, 236, 182, 252, 193, 178, 183, 252, 193, 180, 183, 252, 193, 182, 183, 252, 193, 184, 183, 252, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 206, 0, 71, 0, 193, 4, 181, 252, 193, 68, 180, 252, 193, 70, 180, 252, 193, 72, 180, 252, 193, 74, 180, 252, 193, 76, 180, 252, 193, 78, 180, 252, 193, 80, 180, 252, 193, 82, 180, 252, 193, 84, 180, 252, 193, 86, 180, 252, 193, 88, 180, 252, 193, 90, 180, 252, 193, 92, 180, 252, 193, 94, 180, 252, 193, 96, 180, 252, 193, 98, 180, 252, 193, 100, 180, 252, 193, 102, 180, 252, 193, 104, 180, 252, 193, 106, 180, 252, 193, 108, 180, 252, 193, 110, 180, 252, 193, 112, 180, 252, 193, 114, 180, 252, 193, 116, 180, 252, 193, 118, 180, 252, 193, 6, 181, 252, 193, 8, 181, 252, 193, 10, 181, 252, 193, 12, 181, 252, 193, 14, 181, 252, 193, 16, 181, 252, 193, 18, 181, 252, 193, 20, 181, 252, 193, 22, 181, 252, 193, 24, 181, 252, 193, 26, 181, 252, 193, 28, 181, 252, 193, 30, 181, 252, 193, 32, 181, 252, 193, 34, 181, 252, 193, 36, 181, 252, 193, 38, 181, 252, 193, 40, 181, 252, 193, 42, 181, 252, 193, 44, 181, 252, 193, 46, 181, 252, 193, 48, 181, 252, 193, 50, 181, 252, 193, 52, 181, 252, 193, 54, 181, 252, 193, 56, 181, 252, 193, 58, 181, 252, 193, 60, 181, 252, 193, 62, 181, 252, 193, 64, 181, 252, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 32, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 193, 66, 181, 252, 193, 68, 181, 252, 193, 70, 181, 252, 193, 72, 181, 252, 193, 74, 181, 252, 193, 76, 181, 252, 193, 78, 181, 252, 193, 80, 181, 252, 193, 82, 181, 252, 193, 84, 181, 252, 193, 86, 181, 252, 193, 88, 181, 252, 193, 90, 181, 252, 193, 92, 181, 252, 193, 94, 181, 252, 193, 96, 181, 252, 193, 98, 181, 252, 193, 100, 181, 252, 193, 102, 181, 252, 193, 104, 181, 252, 193, 106, 181, 252, 193, 108, 181, 252, 193, 110, 181, 252, 193, 112, 181, 252, 193, 114, 181, 252, 193, 116, 181, 252, 193, 118, 181, 252, 193, 120, 181, 252, 193, 122, 181, 252, 193, 124, 181, 252, 193, 126, 181, 252, 193, 128, 181, 252, 193, 130, 181, 252, 193, 132, 181, 252, 193, 18, 182, 252, 193, 134, 181, 252, 193, 136, 181, 252, 193, 138, 181, 252, 193, 140, 181, 252, 193, 142, 181, 252, 193, 144, 181, 252, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 206, 64, 71, 0, 193, 172, 182, 252, 193, 244, 181, 252, 193, 246, 181, 252, 193, 138, 183, 252, 193, 248, 181, 252, 193, 250, 181, 252, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 96, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 193, 128, 182, 252, 193, 98, 183, 252, 193, 100, 183, 252, 193, 102, 183, 252, 193, 104, 183, 252, 193, 18, 184, 252, 193, 20, 184, 252, 193, 22, 184, 252, 193, 24, 184, 252, 193, 26, 184, 252, 193, 28, 184, 252, 193, 30, 184, 252, 193, 128, 184, 252, 193, 130, 184, 252, 193, 132, 184, 252, 193, 236, 184, 252, 193, 238, 184, 252, 193, 240, 184, 252, 193, 62, 185, 252, 193, 64, 185, 252, 193, 66, 185, 252, 193, 112, 185, 252, 193, 174, 185, 252, 193, 184, 185, 252, 193, 202, 185, 252, 193, 222, 185, 252, 193, 42, 186, 252, 193, 44, 186, 252, 193, 46, 186, 252, 193, 48, 186, 252, 193, 112, 186, 252, 193, 114, 186, 252, 206, 128, 71, 0, 206, 128, 71, 0, 206, 128, 71, 0, 206, 128, 71, 0, 193, 122, 183, 252, 193, 124, 183, 252, 193, 126, 183, 252, 193, 128, 183, 252, 193, 130, 183, 252, 193, 132, 183, 252, 193, 34, 184, 252, 193, 36, 184, 252, 193, 38, 184, 252, 193, 40, 184, 252, 193, 42, 184, 252, 193, 44, 184, 252, 193, 146, 184, 252, 193, 148, 184, 252, 193, 150, 184, 252, 193, 250, 184, 252, 193, 252, 184, 252, 193, 254, 184, 252, 193, 2, 185, 252, 193, 74, 185, 252, 193, 76, 185, 252, 193, 78, 185, 252, 193, 118, 185, 252, 193, 152, 185, 252, 193, 154, 185, 252, 193, 176, 185, 252, 193, 178, 185, 252, 193, 204, 185, 252, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 206, 160, 71, 0, 193, 140, 183, 252, 193, 222, 182, 252, 193, 142, 183, 252, 193, 144, 183, 252, 193, 146, 183, 252, 193, 148, 183, 252, 193, 150, 183, 252, 193, 152, 183, 252, 193, 154, 183, 252, 193, 156, 183, 252, 193, 158, 183, 252, 193, 224, 182, 252, 193, 160, 183, 252, 193, 162, 183, 252, 193, 164, 183, 252, 193, 166, 183, 252, 193, 168, 183, 252, 193, 170, 183, 252, 193, 172, 183, 252, 193, 174, 183, 252, 193, 176, 183, 252, 206, 0, 72, 0, 206, 0, 72, 0, 206, 0, 72, 0, 206, 0, 72, 0, 206, 0, 72, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 192, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 193, 118, 183, 252, 193, 120, 183, 252, 193, 32, 184, 252, 193, 134, 184, 252, 193, 136, 184, 252, 193, 138, 184, 252, 193, 140, 184, 252, 193, 142, 184, 252, 193, 144, 184, 252, 193, 242, 184, 252, 193, 244, 184, 252, 193, 246, 184, 252, 193, 248, 184, 252, 193, 68, 185, 252, 193, 70, 185, 252, 193, 72, 185, 252, 193, 114, 185, 252, 193, 116, 185, 252, 193, 150, 185, 252, 193, 2, 186, 252, 193, 50, 186, 252, 193, 52, 186, 252, 193, 54, 186, 252, 193, 56, 186, 252, 193, 58, 186, 252, 193, 116, 186, 252, 193, 118, 186, 252, 193, 120, 186, 252, 193, 122, 186, 252, 193, 124, 186, 252, 193, 126, 186, 252, 193, 200, 186, 252, 193, 186, 183, 252, 193, 188, 183, 252, 193, 190, 183, 252, 193, 64, 184, 252, 193, 66, 184, 252, 193, 68, 184, 252, 193, 70, 184, 252, 193, 72, 184, 252, 193, 178, 184, 252, 193, 180, 184, 252, 193, 14, 185, 252, 193, 16, 185, 252, 193, 18, 185, 252, 193, 20, 185, 252, 193, 92, 185, 252, 193, 124, 185, 252, 193, 126, 185, 252, 193, 128, 185, 252, 193, 130, 185, 252, 193, 164, 185, 252, 193, 166, 185, 252, 193, 190, 185, 252, 193, 254, 185, 252, 193, 106, 186, 252, 193, 40, 186, 252, 193, 108, 186, 252, 193, 184, 186, 252, 193, 186, 186, 252, 193, 8, 187, 252, 193, 112, 187, 252, 193, 212, 187, 252, 193, 214, 187, 252, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 224, 71, 0, 206, 32, 72, 0, 206, 32, 72, 0, 206, 32, 72, 0, 206, 32, 72, 0, 193, 182, 184, 252, 193, 16, 184, 252, 193, 90, 184, 252, 193, 92, 184, 252, 193, 184, 184, 252, 193, 94, 184, 252, 193, 96, 184, 252, 193, 186, 184, 252, 193, 98, 184, 252, 193, 100, 184, 252, 193, 102, 184, 252, 193, 104, 184, 252, 193, 106, 184, 252, 193, 108, 184, 252, 193, 110, 184, 252, 193, 112, 184, 252, 193, 114, 184, 252, 193, 116, 184, 252, 193, 118, 184, 252, 193, 120, 184, 252, 193, 122, 184, 252, 193, 124, 184, 252, 193, 126, 184, 252, 193, 188, 184, 252, 193, 190, 184, 252, 193, 192, 184, 252, 193, 194, 184, 252, 193, 196, 184, 252, 206, 0, 72, 0, 206, 0, 72, 0, 206, 0, 72, 0, 206, 0, 72, 0, 193, 152, 184, 252, 193, 154, 184, 252, 193, 156, 184, 252, 193, 158, 184, 252, 193, 160, 184, 252, 193, 162, 184, 252, 193, 164, 184, 252, 193, 166, 184, 252, 193, 168, 184, 252, 193, 170, 184, 252, 193, 172, 184, 252, 193, 174, 184, 252, 193, 176, 184, 252, 193, 10, 185, 252, 193, 12, 185, 252, 193, 84, 185, 252, 193, 86, 185, 252, 193, 88, 185, 252, 193, 90, 185, 252, 193, 122, 185, 252, 193, 156, 185, 252, 193, 158, 185, 252, 193, 160, 185, 252, 193, 162, 185, 252, 193, 186, 185, 252, 193, 188, 185, 252, 193, 212, 185, 252, 193, 220, 185, 252, 193, 198, 184, 252, 193, 200, 184, 252, 193, 22, 185, 252, 193, 202, 184, 252, 193, 204, 184, 252, 193, 206, 184, 252, 193, 208, 184, 252, 193, 210, 184, 252, 193, 212, 184, 252, 193, 214, 184, 252, 193, 216, 184, 252, 193, 218, 184, 252, 193, 220, 184, 252, 193, 222, 184, 252, 193, 224, 184, 252, 193, 226, 184, 252, 193, 228, 184, 252, 193, 230, 184, 252, 193, 232, 184, 252, 193, 234, 184, 252, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 206, 64, 72, 0, 193, 94, 185, 252, 193, 96, 185, 252, 193, 98, 185, 252, 193, 100, 185, 252, 193, 102, 185, 252, 193, 104, 185, 252, 193, 106, 185, 252, 193, 108, 185, 252, 193, 110, 185, 252, 193, 132, 185, 252, 193, 134, 185, 252, 193, 136, 185, 252, 193, 138, 185, 252, 193, 140, 185, 252, 193, 142, 185, 252, 193, 144, 185, 252, 193, 146, 185, 252, 193, 168, 185, 252, 193, 148, 185, 252, 193, 170, 185, 252, 193, 172, 185, 252, 193, 180, 185, 252, 193, 182, 185, 252, 193, 192, 185, 252, 193, 194, 185, 252, 193, 196, 185, 252, 193, 198, 185, 252, 193, 200, 185, 252, 193, 206, 185, 252, 193, 208, 185, 252, 193, 214, 185, 252, 193, 216, 185, 252, 193, 218, 185, 252, 193, 110, 186, 252, 193, 188, 186, 252, 193, 10, 187, 252, 193, 12, 187, 252, 193, 14, 187, 252, 193, 114, 187, 252, 193, 116, 187, 252, 193, 118, 187, 252, 193, 120, 187, 252, 193, 216, 187, 252, 193, 218, 187, 252, 193, 220, 187, 252, 193, 222, 187, 252, 193, 224, 187, 252, 193, 38, 188, 252, 193, 40, 188, 252, 193, 42, 188, 252, 193, 44, 188, 252, 193, 82, 188, 252, 193, 118, 188, 252, 193, 120, 188, 252, 193, 122, 188, 252, 193, 162, 188, 252, 193, 164, 188, 252, 193, 224, 185, 252, 193, 230, 185, 252, 193, 232, 185, 252, 193, 234, 185, 252, 193, 236, 185, 252, 193, 238, 185, 252, 193, 240, 185, 252, 193, 242, 185, 252, 193, 244, 185, 252, 193, 246, 185, 252, 193, 248, 185, 252, 193, 250, 185, 252, 193, 4, 186, 252, 193, 6, 186, 252, 193, 8, 186, 252, 193, 10, 186, 252, 193, 12, 186, 252, 193, 14, 186, 252, 193, 16, 186, 252, 193, 18, 186, 252, 193, 20, 186, 252, 193, 22, 186, 252, 193, 252, 185, 252, 193, 24, 186, 252, 193, 26, 186, 252, 193, 28, 186, 252, 193, 30, 186, 252, 193, 32, 186, 252, 193, 34, 186, 252, 193, 36, 186, 252, 193, 38, 186, 252, 193, 70, 186, 252, 193, 226, 185, 252, 193, 128, 186, 252, 193, 130, 186, 252, 193, 132, 186, 252, 193, 218, 186, 252, 193, 220, 186, 252, 193, 222, 186, 252, 193, 38, 187, 252, 193, 40, 187, 252, 193, 42, 187, 252, 193, 136, 187, 252, 193, 138, 187, 252, 193, 140, 187, 252, 193, 252, 187, 252, 193, 102, 188, 252, 193, 104, 188, 252, 193, 218, 188, 252, 193, 238, 188, 252, 193, 16, 189, 252, 193, 48, 189, 252, 193, 50, 189, 252, 193, 82, 189, 252, 193, 108, 189, 252, 193, 132, 189, 252, 193, 152, 189, 252, 193, 170, 189, 252, 193, 172, 189, 252, 193, 174, 189, 252, 193, 2, 190, 252, 193, 62, 190, 252, 193, 106, 190, 252, 193, 108, 190, 252, 193, 66, 186, 252, 193, 68, 186, 252, 193, 134, 186, 252, 193, 136, 186, 252, 193, 138, 186, 252, 193, 140, 186, 252, 193, 142, 186, 252, 193, 144, 186, 252, 193, 146, 186, 252, 193, 148, 186, 252, 193, 150, 186, 252, 193, 152, 186, 252, 193, 224, 186, 252, 193, 226, 186, 252, 193, 228, 186, 252, 193, 230, 186, 252, 193, 232, 186, 252, 193, 234, 186, 252, 193, 236, 186, 252, 193, 238, 186, 252, 193, 240, 186, 252, 193, 242, 186, 252, 193, 44, 187, 252, 193, 46, 187, 252, 193, 48, 187, 252, 193, 50, 187, 252, 193, 52, 187, 252, 193, 54, 187, 252, 193, 56, 187, 252, 193, 58, 187, 252, 193, 60, 187, 252, 193, 142, 187, 252, 193, 72, 186, 252, 193, 74, 186, 252, 193, 76, 186, 252, 193, 78, 186, 252, 193, 80, 186, 252, 193, 82, 186, 252, 193, 84, 186, 252, 193, 86, 186, 252, 193, 88, 186, 252, 193, 90, 186, 252, 193, 92, 186, 252, 193, 94, 186, 252, 193, 96, 186, 252, 193, 98, 186, 252, 193, 100, 186, 252, 193, 102, 186, 252, 193, 104, 186, 252, 193, 154, 186, 252, 193, 156, 186, 252, 193, 158, 186, 252, 193, 160, 186, 252, 193, 162, 186, 252, 193, 164, 186, 252, 193, 166, 186, 252, 193, 168, 186, 252, 193, 170, 186, 252, 193, 172, 186, 252, 193, 174, 186, 252, 193, 176, 186, 252, 193, 178, 186, 252, 193, 180, 186, 252, 193, 182, 186, 252, 193, 190, 186, 252, 193, 192, 186, 252, 193, 194, 186, 252, 193, 196, 186, 252, 193, 16, 187, 252, 193, 18, 187, 252, 193, 20, 187, 252, 193, 22, 187, 252, 193, 24, 187, 252, 193, 26, 187, 252, 193, 122, 187, 252, 193, 124, 187, 252, 193, 226, 187, 252, 193, 228, 187, 252, 193, 230, 187, 252, 193, 46, 188, 252, 193, 48, 188, 252, 193, 84, 188, 252, 193, 86, 188, 252, 193, 88, 188, 252, 193, 90, 188, 252, 193, 124, 188, 252, 193, 126, 188, 252, 193, 166, 188, 252, 193, 180, 188, 252, 193, 196, 188, 252, 193, 46, 189, 252, 193, 78, 189, 252, 193, 100, 189, 252, 193, 102, 189, 252, 193, 128, 189, 252, 193, 192, 189, 252, 193, 202, 186, 252, 193, 204, 186, 252, 193, 206, 186, 252, 193, 208, 186, 252, 193, 210, 186, 252, 193, 212, 186, 252, 193, 214, 186, 252, 193, 216, 186, 252, 193, 30, 187, 252, 193, 32, 187, 252, 193, 34, 187, 252, 193, 36, 187, 252, 193, 128, 187, 252, 193, 130, 187, 252, 193, 132, 187, 252, 193, 134, 187, 252, 193, 234, 187, 252, 193, 236, 187, 252, 193, 238, 187, 252, 193, 240, 187, 252, 193, 242, 187, 252, 193, 244, 187, 252, 193, 246, 187, 252, 193, 248, 187, 252, 193, 250, 187, 252, 193, 50, 188, 252, 193, 52, 188, 252, 193, 54, 188, 252, 193, 56, 188, 252, 193, 58, 188, 252, 193, 92, 188, 252, 193, 94, 188, 252, 193, 244, 186, 252, 193, 246, 186, 252, 193, 248, 186, 252, 193, 250, 186, 252, 193, 252, 186, 252, 193, 254, 186, 252, 193, 2, 187, 252, 193, 4, 187, 252, 193, 6, 187, 252, 193, 62, 187, 252, 193, 64, 187, 252, 193, 66, 187, 252, 193, 68, 187, 252, 193, 70, 187, 252, 193, 72, 187, 252, 193, 74, 187, 252, 193, 76, 187, 252, 193, 78, 187, 252, 193, 80, 187, 252, 193, 82, 187, 252, 193, 84, 187, 252, 193, 86, 187, 252, 193, 88, 187, 252, 193, 90, 187, 252, 193, 92, 187, 252, 193, 94, 187, 252, 193, 96, 187, 252, 193, 98, 187, 252, 193, 100, 187, 252, 193, 102, 187, 252, 193, 104, 187, 252, 193, 106, 187, 252, 193, 108, 187, 252, 193, 110, 187, 252, 193, 166, 187, 252, 193, 168, 187, 252, 193, 170, 187, 252, 193, 172, 187, 252, 193, 174, 187, 252, 193, 176, 187, 252, 193, 178, 187, 252, 193, 180, 187, 252, 193, 182, 187, 252, 193, 184, 187, 252, 193, 186, 187, 252, 193, 188, 187, 252, 193, 190, 187, 252, 193, 192, 187, 252, 193, 194, 187, 252, 193, 196, 187, 252, 193, 198, 187, 252, 193, 200, 187, 252, 193, 202, 187, 252, 193, 204, 187, 252, 193, 206, 187, 252, 193, 208, 187, 252, 193, 210, 187, 252, 206, 96, 72, 0, 206, 96, 72, 0, 206, 96, 72, 0, 206, 96, 72, 0, 206, 96, 72, 0, 206, 96, 72, 0, 206, 96, 72, 0, 193, 144, 187, 252, 193, 146, 187, 252, 193, 148, 187, 252, 193, 150, 187, 252, 193, 152, 187, 252, 193, 154, 187, 252, 193, 156, 187, 252, 193, 158, 187, 252, 193, 160, 187, 252, 193, 162, 187, 252, 193, 164, 187, 252, 193, 254, 187, 252, 193, 2, 188, 252, 193, 4, 188, 252, 193, 6, 188, 252, 193, 8, 188, 252, 193, 10, 188, 252, 193, 60, 188, 252, 193, 62, 188, 252, 193, 64, 188, 252, 193, 66, 188, 252, 193, 106, 188, 252, 193, 108, 188, 252, 193, 140, 188, 252, 193, 142, 188, 252, 193, 144, 188, 252, 193, 146, 188, 252, 193, 148, 188, 252, 193, 170, 188, 252, 193, 172, 188, 252, 193, 174, 188, 252, 193, 184, 188, 252, 206, 96, 72, 0, 206, 96, 72, 0, 206, 96, 72, 0, 206, 96, 72, 0, 206, 96, 72, 0, 193, 68, 188, 252, 193, 70, 188, 252, 193, 72, 188, 252, 193, 74, 188, 252, 193, 76, 188, 252, 193, 78, 188, 252, 193, 80, 188, 252, 193, 110, 188, 252, 193, 112, 188, 252, 193, 114, 188, 252, 193, 116, 188, 252, 193, 150, 188, 252, 193, 152, 188, 252, 193, 154, 188, 252, 193, 156, 188, 252, 193, 158, 188, 252, 193, 192, 188, 252, 193, 198, 188, 252, 193, 202, 188, 252, 193, 208, 188, 252, 193, 220, 188, 252, 193, 222, 188, 252, 193, 240, 188, 252, 193, 224, 188, 252, 193, 242, 188, 252, 193, 18, 189, 252, 193, 20, 189, 252, 193, 36, 188, 252, 193, 160, 188, 252, 193, 210, 188, 252, 193, 226, 188, 252, 193, 244, 188, 252, 193, 246, 188, 252, 193, 248, 188, 252, 193, 250, 188, 252, 193, 252, 188, 252, 193, 254, 188, 252, 193, 24, 189, 252, 193, 56, 189, 252, 193, 58, 189, 252, 193, 112, 189, 252, 193, 216, 189, 252, 193, 244, 189, 252, 193, 30, 190, 252, 193, 32, 190, 252, 193, 70, 190, 252, 193, 72, 190, 252, 193, 74, 190, 252, 193, 130, 190, 252, 193, 132, 190, 252, 193, 134, 190, 252, 193, 212, 190, 252, 193, 16, 191, 252, 193, 18, 191, 252, 193, 68, 191, 252, 193, 70, 191, 252, 193, 72, 191, 252, 193, 74, 191, 252, 193, 156, 191, 252, 193, 96, 188, 252, 193, 98, 188, 252, 193, 100, 188, 252, 193, 128, 188, 252, 193, 130, 188, 252, 193, 132, 188, 252, 193, 134, 188, 252, 193, 136, 188, 252, 193, 138, 188, 252, 193, 168, 188, 252, 193, 186, 188, 252, 193, 194, 188, 252, 193, 234, 188, 252, 193, 236, 188, 252, 193, 14, 189, 252, 193, 36, 189, 252, 193, 80, 189, 252, 193, 104, 189, 252, 193, 106, 189, 252, 193, 130, 189, 252, 193, 194, 189, 252, 193, 20, 190, 252, 193, 56, 190, 252, 193, 58, 190, 252, 193, 60, 190, 252, 193, 104, 190, 252, 193, 194, 190, 252, 193, 196, 190, 252, 193, 198, 190, 252, 193, 254, 190, 252, 193, 40, 191, 252, 193, 42, 191, 252, 193, 176, 188, 252, 193, 178, 188, 252, 193, 200, 188, 252, 193, 212, 188, 252, 193, 214, 188, 252, 193, 216, 188, 252, 193, 228, 188, 252, 193, 230, 188, 252, 193, 232, 188, 252, 193, 2, 189, 252, 193, 4, 189, 252, 193, 6, 189, 252, 193, 8, 189, 252, 193, 10, 189, 252, 193, 12, 189, 252, 193, 26, 189, 252, 193, 28, 189, 252, 193, 30, 189, 252, 193, 32, 189, 252, 193, 34, 189, 252, 193, 38, 189, 252, 193, 40, 189, 252, 193, 42, 189, 252, 193, 44, 189, 252, 193, 60, 189, 252, 193, 62, 189, 252, 193, 64, 189, 252, 193, 66, 189, 252, 193, 68, 189, 252, 193, 70, 189, 252, 193, 72, 189, 252, 193, 74, 189, 252, 193, 188, 188, 252, 193, 190, 188, 252, 193, 204, 188, 252, 193, 206, 188, 252, 193, 84, 189, 252, 193, 154, 189, 252, 193, 176, 189, 252, 193, 208, 189, 252, 193, 112, 190, 252, 193, 208, 190, 252, 193, 4, 191, 252, 193, 6, 191, 252, 193, 8, 191, 252, 193, 46, 191, 252, 193, 48, 191, 252, 193, 140, 191, 252, 193, 142, 191, 252, 193, 144, 191, 252, 193, 4, 192, 252, 193, 6, 192, 252, 193, 8, 192, 252, 193, 78, 192, 252, 193, 252, 192, 252, 193, 34, 193, 252, 193, 124, 193, 252, 193, 250, 193, 252, 193, 154, 194, 252, 193, 194, 194, 252, 193, 196, 194, 252, 193, 140, 195, 252, 193, 244, 195, 252, 193, 246, 195, 252, 193, 22, 189, 252, 193, 52, 189, 252, 193, 54, 189, 252, 193, 110, 189, 252, 193, 134, 189, 252, 193, 136, 189, 252, 193, 138, 189, 252, 193, 156, 189, 252, 193, 158, 189, 252, 193, 160, 189, 252, 193, 178, 189, 252, 193, 180, 189, 252, 193, 196, 189, 252, 193, 202, 189, 252, 193, 210, 189, 252, 193, 212, 189, 252, 193, 224, 189, 252, 193, 238, 189, 252, 193, 240, 189, 252, 193, 242, 189, 252, 193, 22, 190, 252, 193, 24, 190, 252, 193, 26, 190, 252, 193, 28, 190, 252, 193, 64, 190, 252, 193, 66, 190, 252, 193, 68, 190, 252, 193, 114, 190, 252, 193, 116, 190, 252, 193, 118, 190, 252, 193, 120, 190, 252, 193, 122, 190, 252, 193, 76, 189, 252, 193, 86, 189, 252, 193, 88, 189, 252, 193, 90, 189, 252, 193, 92, 189, 252, 193, 94, 189, 252, 193, 96, 189, 252, 193, 98, 189, 252, 193, 114, 189, 252, 193, 116, 189, 252, 193, 118, 189, 252, 193, 120, 189, 252, 193, 122, 189, 252, 193, 124, 189, 252, 193, 126, 189, 252, 193, 140, 189, 252, 193, 142, 189, 252, 193, 144, 189, 252, 193, 146, 189, 252, 193, 148, 189, 252, 193, 150, 189, 252, 193, 162, 189, 252, 193, 164, 189, 252, 193, 166, 189, 252, 193, 168, 189, 252, 193, 184, 189, 252, 193, 186, 189, 252, 193, 188, 189, 252, 193, 190, 189, 252, 193, 198, 189, 252, 193, 200, 189, 252, 193, 204, 189, 252, 193, 206, 189, 252, 193, 214, 189, 252, 193, 218, 189, 252, 193, 220, 189, 252, 193, 222, 189, 252, 193, 228, 189, 252, 193, 230, 189, 252, 193, 232, 189, 252, 193, 234, 189, 252, 193, 236, 189, 252, 193, 246, 189, 252, 193, 248, 189, 252, 193, 4, 190, 252, 193, 250, 189, 252, 193, 6, 190, 252, 193, 254, 189, 252, 193, 252, 189, 252, 193, 8, 190, 252, 193, 10, 190, 252, 193, 12, 190, 252, 193, 14, 190, 252, 193, 16, 190, 252, 193, 18, 190, 252, 193, 34, 190, 252, 193, 36, 190, 252, 193, 38, 190, 252, 193, 40, 190, 252, 193, 42, 190, 252, 193, 44, 190, 252, 193, 46, 190, 252, 193, 48, 190, 252, 193, 50, 190, 252, 193, 52, 190, 252, 193, 76, 190, 252, 193, 78, 190, 252, 193, 80, 190, 252, 193, 82, 190, 252, 193, 84, 190, 252, 193, 86, 190, 252, 193, 88, 190, 252, 193, 90, 190, 252, 193, 92, 190, 252, 193, 94, 190, 252, 193, 96, 190, 252, 193, 98, 190, 252, 193, 100, 190, 252, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 206, 128, 72, 0, 193, 54, 190, 252, 193, 102, 190, 252, 193, 192, 190, 252, 193, 248, 190, 252, 193, 250, 190, 252, 193, 252, 190, 252, 193, 122, 191, 252, 193, 124, 191, 252, 193, 192, 191, 252, 193, 236, 191, 252, 193, 64, 192, 252, 193, 154, 192, 252, 193, 192, 192, 252, 193, 216, 192, 252, 193, 32, 193, 252, 193, 186, 193, 252, 193, 188, 193, 252, 193, 190, 193, 252, 193, 192, 193, 252, 193, 236, 193, 252, 193, 238, 193, 252, 193, 50, 194, 252, 193, 140, 194, 252, 193, 142, 194, 252, 193, 144, 194, 252, 193, 180, 194, 252, 193, 228, 194, 252, 193, 230, 194, 252, 193, 232, 194, 252, 193, 28, 195, 252, 193, 30, 195, 252, 193, 76, 195, 252, 193, 110, 190, 252, 193, 200, 190, 252, 193, 202, 190, 252, 193, 204, 190, 252, 193, 206, 190, 252, 193, 2, 191, 252, 193, 44, 191, 252, 193, 134, 191, 252, 193, 136, 191, 252, 193, 138, 191, 252, 193, 194, 191, 252, 193, 196, 191, 252, 193, 198, 191, 252, 193, 244, 191, 252, 193, 246, 191, 252, 193, 248, 191, 252, 193, 250, 191, 252, 193, 252, 191, 252, 193, 254, 191, 252, 193, 2, 192, 252, 193, 66, 192, 252, 193, 68, 192, 252, 193, 70, 192, 252, 193, 72, 192, 252, 193, 74, 192, 252, 193, 76, 192, 252, 193, 124, 192, 252, 193, 126, 192, 252, 193, 158, 192, 252, 193, 194, 192, 252, 193, 196, 192, 252, 193, 220, 192, 252, 193, 124, 190, 252, 193, 126, 190, 252, 193, 128, 190, 252, 193, 210, 190, 252, 193, 10, 191, 252, 193, 12, 191, 252, 193, 14, 191, 252, 193, 50, 191, 252, 193, 52, 191, 252, 193, 54, 191, 252, 193, 56, 191, 252, 193, 58, 191, 252, 193, 60, 191, 252, 193, 62, 191, 252, 193, 64, 191, 252, 193, 146, 191, 252, 193, 148, 191, 252, 193, 150, 191, 252, 193, 66, 191, 252, 193, 152, 191, 252, 193, 154, 191, 252, 193, 200, 191, 252, 193, 202, 191, 252, 193, 204, 191, 252, 193, 206, 191, 252, 193, 10, 192, 252, 193, 12, 192, 252, 193, 14, 192, 252, 193, 16, 192, 252, 193, 80, 192, 252, 193, 82, 192, 252, 193, 84, 192, 252, 206, 160, 72, 0, 206, 160, 72, 0, 206, 160, 72, 0, 206, 160, 72, 0, 206, 160, 72, 0, 193, 76, 191, 252, 193, 78, 191, 252, 193, 80, 191, 252, 193, 82, 191, 252, 193, 84, 191, 252, 193, 86, 191, 252, 193, 88, 191, 252, 193, 90, 191, 252, 193, 92, 191, 252, 193, 94, 191, 252, 193, 96, 191, 252, 193, 98, 191, 252, 193, 38, 191, 252, 193, 100, 191, 252, 193, 102, 191, 252, 193, 104, 191, 252, 193, 106, 191, 252, 193, 108, 191, 252, 193, 110, 191, 252, 193, 112, 191, 252, 193, 114, 191, 252, 193, 116, 191, 252, 193, 118, 191, 252, 193, 120, 191, 252, 193, 158, 191, 252, 193, 160, 191, 252, 193, 162, 191, 252, 193, 126, 191, 252, 193, 128, 191, 252, 193, 130, 191, 252, 193, 132, 191, 252, 193, 238, 191, 252, 193, 240, 191, 252, 193, 242, 191, 252, 193, 120, 192, 252, 193, 122, 192, 252, 193, 156, 192, 252, 193, 218, 192, 252, 193, 226, 192, 252, 193, 60, 193, 252, 193, 118, 193, 252, 193, 120, 193, 252, 193, 122, 193, 252, 193, 194, 193, 252, 193, 240, 193, 252, 193, 242, 193, 252, 193, 244, 193, 252, 193, 52, 194, 252, 193, 90, 194, 252, 193, 92, 194, 252, 193, 146, 194, 252, 193, 148, 194, 252, 193, 182, 194, 252, 193, 184, 194, 252, 193, 186, 194, 252, 193, 234, 194, 252, 193, 236, 194, 252, 193, 32, 195, 252, 193, 58, 195, 252, 193, 18, 192, 252, 193, 20, 192, 252, 193, 90, 192, 252, 193, 130, 192, 252, 193, 168, 192, 252, 193, 210, 192, 252, 193, 234, 192, 252, 193, 82, 193, 252, 193, 84, 193, 252, 193, 86, 193, 252, 193, 140, 193, 252, 193, 142, 193, 252, 193, 144, 193, 252, 193, 146, 193, 252, 193, 206, 193, 252, 193, 208, 193, 252, 193, 210, 193, 252, 193, 14, 194, 252, 193, 58, 194, 252, 193, 60, 194, 252, 193, 62, 194, 252, 193, 64, 194, 252, 193, 66, 194, 252, 193, 112, 194, 252, 193, 160, 194, 252, 193, 162, 194, 252, 193, 198, 194, 252, 193, 244, 194, 252, 193, 246, 194, 252, 193, 102, 195, 252, 193, 192, 195, 252, 193, 194, 195, 252, 193, 28, 192, 252, 193, 30, 192, 252, 193, 32, 192, 252, 193, 34, 192, 252, 193, 36, 192, 252, 193, 38, 192, 252, 193, 40, 192, 252, 193, 42, 192, 252, 193, 44, 192, 252, 193, 46, 192, 252, 193, 48, 192, 252, 193, 50, 192, 252, 193, 52, 192, 252, 193, 54, 192, 252, 193, 56, 192, 252, 193, 58, 192, 252, 193, 60, 192, 252, 193, 62, 192, 252, 193, 92, 192, 252, 193, 94, 192, 252, 193, 96, 192, 252, 193, 98, 192, 252, 193, 100, 192, 252, 193, 102, 192, 252, 193, 104, 192, 252, 193, 106, 192, 252, 193, 108, 192, 252, 193, 110, 192, 252, 193, 112, 192, 252, 193, 114, 192, 252, 193, 116, 192, 252, 193, 132, 192, 252, 193, 86, 192, 252, 193, 88, 192, 252, 193, 128, 192, 252, 193, 160, 192, 252, 193, 162, 192, 252, 193, 164, 192, 252, 193, 166, 192, 252, 193, 222, 192, 252, 193, 230, 192, 252, 193, 232, 192, 252, 193, 14, 193, 252, 193, 22, 193, 252, 193, 24, 193, 252, 193, 36, 193, 252, 193, 38, 193, 252, 193, 40, 193, 252, 193, 42, 193, 252, 193, 44, 193, 252, 193, 68, 193, 252, 193, 70, 193, 252, 193, 72, 193, 252, 193, 126, 193, 252, 193, 74, 193, 252, 193, 76, 193, 252, 193, 78, 193, 252, 193, 80, 193, 252, 193, 128, 193, 252, 193, 130, 193, 252, 193, 132, 193, 252, 193, 134, 193, 252, 193, 136, 193, 252, 193, 138, 193, 252, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 193, 118, 192, 252, 193, 186, 51, 253, 193, 188, 51, 253, 193, 190, 51, 253, 193, 192, 51, 253, 193, 194, 51, 253, 193, 196, 51, 253, 193, 242, 51, 253, 193, 244, 51, 253, 193, 246, 51, 253, 193, 248, 51, 253, 193, 250, 51, 253, 193, 252, 51, 253, 193, 254, 51, 253, 193, 2, 52, 253, 193, 4, 52, 253, 193, 6, 52, 253, 193, 8, 52, 253, 193, 10, 52, 253, 193, 12, 52, 253, 193, 14, 52, 253, 193, 16, 52, 253, 193, 18, 52, 253, 193, 20, 52, 253, 193, 22, 52, 253, 193, 64, 52, 253, 193, 66, 52, 253, 193, 134, 192, 252, 193, 136, 192, 252, 193, 138, 192, 252, 193, 140, 192, 252, 193, 142, 192, 252, 193, 144, 192, 252, 193, 146, 192, 252, 193, 148, 192, 252, 193, 150, 192, 252, 193, 248, 194, 252, 193, 170, 192, 252, 193, 172, 192, 252, 193, 174, 192, 252, 193, 176, 192, 252, 193, 178, 192, 252, 193, 180, 192, 252, 193, 182, 192, 252, 193, 184, 192, 252, 193, 186, 192, 252, 193, 188, 192, 252, 193, 190, 192, 252, 193, 198, 192, 252, 193, 200, 192, 252, 193, 202, 192, 252, 193, 204, 192, 252, 193, 206, 192, 252, 193, 208, 192, 252, 193, 212, 192, 252, 193, 214, 192, 252, 193, 224, 192, 252, 193, 236, 192, 252, 193, 238, 192, 252, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 193, 152, 192, 252, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 193, 228, 192, 252, 193, 254, 192, 252, 193, 62, 193, 252, 193, 64, 193, 252, 193, 66, 193, 252, 193, 196, 193, 252, 193, 198, 193, 252, 193, 246, 193, 252, 193, 248, 193, 252, 193, 54, 194, 252, 193, 94, 194, 252, 193, 96, 194, 252, 193, 98, 194, 252, 193, 100, 194, 252, 193, 150, 194, 252, 193, 152, 194, 252, 193, 188, 194, 252, 193, 190, 194, 252, 193, 192, 194, 252, 193, 34, 195, 252, 193, 60, 195, 252, 193, 98, 195, 252, 193, 100, 195, 252, 193, 114, 195, 252, 193, 236, 195, 252, 193, 238, 195, 252, 193, 240, 195, 252, 193, 242, 195, 252, 193, 64, 196, 252, 193, 140, 196, 252, 193, 142, 196, 252, 193, 144, 196, 252, 193, 240, 192, 252, 193, 242, 192, 252, 193, 244, 192, 252, 193, 246, 192, 252, 193, 248, 192, 252, 193, 250, 192, 252, 193, 2, 193, 252, 193, 4, 193, 252, 193, 6, 193, 252, 193, 8, 193, 252, 193, 10, 193, 252, 193, 16, 193, 252, 193, 18, 193, 252, 193, 20, 193, 252, 193, 26, 193, 252, 193, 28, 193, 252, 193, 30, 193, 252, 193, 46, 193, 252, 193, 48, 193, 252, 193, 50, 193, 252, 193, 52, 193, 252, 193, 54, 193, 252, 193, 56, 193, 252, 206, 192, 72, 0, 206, 192, 72, 0, 206, 192, 72, 0, 206, 192, 72, 0, 206, 192, 72, 0, 206, 192, 72, 0, 206, 192, 72, 0, 206, 192, 72, 0, 206, 192, 72, 0, 206, 192, 72, 0, 193, 148, 193, 252, 193, 150, 193, 252, 193, 152, 193, 252, 193, 154, 193, 252, 193, 156, 193, 252, 193, 158, 193, 252, 193, 160, 193, 252, 193, 162, 193, 252, 193, 164, 193, 252, 193, 116, 193, 252, 193, 166, 193, 252, 193, 168, 193, 252, 193, 170, 193, 252, 193, 172, 193, 252, 193, 174, 193, 252, 193, 176, 193, 252, 193, 178, 193, 252, 193, 180, 193, 252, 193, 182, 193, 252, 193, 184, 193, 252, 206, 224, 72, 0, 206, 224, 72, 0, 206, 224, 72, 0, 206, 224, 72, 0, 206, 224, 72, 0, 206, 224, 72, 0, 206, 224, 72, 0, 193, 200, 193, 252, 193, 202, 193, 252, 193, 204, 193, 252, 193, 252, 193, 252, 193, 254, 193, 252, 193, 2, 194, 252, 193, 4, 194, 252, 193, 6, 194, 252, 193, 8, 194, 252, 193, 10, 194, 252, 193, 12, 194, 252, 193, 56, 194, 252, 193, 102, 194, 252, 193, 104, 194, 252, 193, 106, 194, 252, 193, 108, 194, 252, 193, 110, 194, 252, 193, 156, 194, 252, 193, 158, 194, 252, 193, 238, 194, 252, 193, 240, 194, 252, 193, 242, 194, 252, 193, 36, 195, 252, 193, 38, 195, 252, 193, 40, 195, 252, 193, 42, 195, 252, 193, 62, 195, 252, 193, 78, 195, 252, 193, 80, 195, 252, 193, 136, 195, 252, 193, 144, 195, 252, 193, 146, 195, 252, 193, 86, 194, 252, 193, 88, 194, 252, 193, 114, 194, 252, 193, 116, 194, 252, 193, 118, 194, 252, 193, 120, 194, 252, 193, 122, 194, 252, 193, 124, 194, 252, 193, 126, 194, 252, 193, 128, 194, 252, 193, 130, 194, 252, 193, 132, 194, 252, 193, 134, 194, 252, 193, 136, 194, 252, 193, 138, 194, 252, 193, 164, 194, 252, 193, 166, 194, 252, 193, 168, 194, 252, 193, 170, 194, 252, 193, 172, 194, 252, 193, 174, 194, 252, 193, 176, 194, 252, 193, 178, 194, 252, 206, 0, 73, 0, 206, 0, 73, 0, 206, 0, 73, 0, 206, 0, 73, 0, 206, 0, 73, 0, 206, 0, 73, 0, 206, 0, 73, 0, 206, 0, 73, 0, 206, 0, 73, 0, 206, 0, 73, 0, 193, 250, 194, 252, 193, 252, 194, 252, 193, 254, 194, 252, 193, 2, 195, 252, 193, 4, 195, 252, 193, 6, 195, 252, 193, 8, 195, 252, 193, 10, 195, 252, 193, 12, 195, 252, 193, 14, 195, 252, 193, 16, 195, 252, 193, 18, 195, 252, 193, 20, 195, 252, 193, 22, 195, 252, 193, 24, 195, 252, 193, 26, 195, 252, 193, 44, 195, 252, 193, 46, 195, 252, 193, 48, 195, 252, 193, 50, 195, 252, 193, 52, 195, 252, 193, 54, 195, 252, 193, 56, 195, 252, 193, 64, 195, 252, 193, 66, 195, 252, 193, 68, 195, 252, 193, 70, 195, 252, 193, 72, 195, 252, 193, 74, 195, 252, 193, 84, 195, 252, 193, 86, 195, 252, 193, 88, 195, 252, 193, 90, 195, 252, 193, 92, 195, 252, 193, 94, 195, 252, 193, 104, 195, 252, 193, 106, 195, 252, 193, 116, 195, 252, 193, 118, 195, 252, 193, 120, 195, 252, 193, 122, 195, 252, 193, 124, 195, 252, 193, 130, 195, 252, 193, 132, 195, 252, 193, 134, 195, 252, 193, 142, 195, 252, 193, 138, 195, 252, 193, 148, 195, 252, 193, 150, 195, 252, 193, 152, 195, 252, 193, 154, 195, 252, 193, 156, 195, 252, 193, 158, 195, 252, 193, 160, 195, 252, 193, 162, 195, 252, 193, 164, 195, 252, 193, 166, 195, 252, 193, 168, 195, 252, 193, 170, 195, 252, 193, 96, 195, 252, 193, 112, 195, 252, 193, 126, 195, 252, 193, 200, 196, 252, 193, 14, 197, 252, 193, 160, 197, 252, 193, 162, 197, 252, 193, 210, 197, 252, 193, 212, 197, 252, 193, 96, 198, 252, 193, 98, 198, 252, 193, 148, 198, 252, 193, 174, 198, 252, 193, 200, 198, 252, 193, 206, 198, 252, 193, 16, 199, 252, 193, 126, 199, 252, 193, 50, 200, 252, 193, 118, 200, 252, 193, 120, 200, 252, 193, 10, 201, 252, 193, 12, 201, 252, 193, 14, 201, 252, 193, 86, 201, 252, 193, 88, 201, 252, 193, 90, 201, 252, 193, 152, 201, 252, 193, 154, 201, 252, 193, 156, 201, 252, 193, 158, 201, 252, 193, 252, 201, 252, 193, 254, 201, 252, 193, 108, 195, 252, 193, 110, 195, 252, 193, 176, 195, 252, 193, 234, 195, 252, 193, 62, 196, 252, 193, 134, 196, 252, 193, 136, 196, 252, 193, 138, 196, 252, 193, 196, 196, 252, 193, 198, 196, 252, 193, 12, 197, 252, 193, 94, 197, 252, 193, 96, 197, 252, 193, 156, 197, 252, 193, 158, 197, 252, 193, 208, 197, 252, 193, 12, 198, 252, 193, 14, 198, 252, 193, 76, 198, 252, 193, 144, 198, 252, 193, 146, 198, 252, 193, 172, 198, 252, 193, 180, 198, 252, 193, 188, 198, 252, 193, 204, 198, 252, 193, 102, 199, 252, 193, 122, 199, 252, 193, 124, 199, 252, 193, 146, 199, 252, 193, 214, 199, 252, 193, 238, 199, 252, 193, 36, 200, 252, 193, 172, 195, 252, 193, 174, 195, 252, 193, 196, 195, 252, 193, 198, 195, 252, 193, 200, 195, 252, 193, 202, 195, 252, 193, 204, 195, 252, 193, 206, 195, 252, 193, 208, 195, 252, 193, 210, 195, 252, 193, 212, 195, 252, 193, 214, 195, 252, 193, 216, 195, 252, 193, 218, 195, 252, 193, 220, 195, 252, 193, 222, 195, 252, 193, 224, 195, 252, 193, 226, 195, 252, 193, 228, 195, 252, 193, 230, 195, 252, 193, 232, 195, 252, 193, 34, 196, 252, 193, 36, 196, 252, 193, 38, 196, 252, 193, 40, 196, 252, 193, 42, 196, 252, 193, 44, 196, 252, 193, 46, 196, 252, 193, 48, 196, 252, 193, 50, 196, 252, 193, 52, 196, 252, 193, 54, 196, 252, 193, 178, 195, 252, 193, 180, 195, 252, 193, 182, 195, 252, 193, 184, 195, 252, 193, 186, 195, 252, 193, 188, 195, 252, 193, 190, 195, 252, 193, 250, 195, 252, 193, 252, 195, 252, 193, 254, 195, 252, 193, 2, 196, 252, 193, 4, 196, 252, 193, 6, 196, 252, 193, 8, 196, 252, 193, 10, 196, 252, 193, 12, 196, 252, 193, 14, 196, 252, 193, 16, 196, 252, 193, 18, 196, 252, 193, 68, 196, 252, 193, 70, 196, 252, 193, 72, 196, 252, 193, 74, 196, 252, 193, 76, 196, 252, 193, 78, 196, 252, 193, 80, 196, 252, 193, 82, 196, 252, 193, 154, 196, 252, 193, 156, 196, 252, 193, 158, 196, 252, 193, 208, 196, 252, 193, 210, 196, 252, 193, 248, 195, 252, 193, 66, 196, 252, 193, 152, 196, 252, 193, 206, 196, 252, 193, 24, 197, 252, 193, 174, 197, 252, 193, 78, 198, 252, 193, 90, 198, 252, 193, 108, 198, 252, 193, 110, 198, 252, 193, 152, 198, 252, 193, 182, 198, 252, 193, 238, 198, 252, 193, 2, 199, 252, 193, 4, 199, 252, 193, 32, 199, 252, 193, 78, 199, 252, 193, 80, 199, 252, 193, 52, 200, 252, 193, 126, 200, 252, 193, 198, 200, 252, 193, 200, 200, 252, 193, 202, 200, 252, 193, 204, 200, 252, 193, 18, 201, 252, 193, 20, 201, 252, 193, 22, 201, 252, 193, 28, 203, 252, 193, 136, 203, 252, 193, 160, 203, 252, 193, 232, 203, 252, 193, 16, 204, 252, 193, 20, 196, 252, 193, 22, 196, 252, 193, 24, 196, 252, 193, 26, 196, 252, 193, 28, 196, 252, 193, 30, 196, 252, 193, 32, 196, 252, 193, 84, 196, 252, 193, 86, 196, 252, 193, 88, 196, 252, 193, 90, 196, 252, 193, 160, 196, 252, 193, 162, 196, 252, 193, 164, 196, 252, 193, 166, 196, 252, 193, 222, 196, 252, 193, 224, 196, 252, 193, 226, 196, 252, 193, 228, 196, 252, 193, 230, 196, 252, 193, 114, 197, 252, 193, 116, 197, 252, 193, 228, 197, 252, 193, 230, 197, 252, 193, 232, 197, 252, 193, 36, 198, 252, 193, 54, 198, 252, 193, 100, 198, 252, 193, 158, 198, 252, 193, 160, 198, 252, 193, 162, 198, 252, 193, 234, 198, 252, 193, 56, 196, 252, 193, 58, 196, 252, 193, 60, 196, 252, 193, 92, 196, 252, 193, 94, 196, 252, 193, 96, 196, 252, 193, 98, 196, 252, 193, 100, 196, 252, 193, 102, 196, 252, 193, 104, 196, 252, 193, 106, 196, 252, 193, 108, 196, 252, 193, 110, 196, 252, 193, 112, 196, 252, 193, 114, 196, 252, 193, 116, 196, 252, 193, 118, 196, 252, 193, 120, 196, 252, 193, 122, 196, 252, 193, 124, 196, 252, 193, 126, 196, 252, 193, 128, 196, 252, 193, 130, 196, 252, 193, 132, 196, 252, 206, 32, 73, 0, 206, 32, 73, 0, 206, 32, 73, 0, 206, 32, 73, 0, 206, 32, 73, 0, 206, 32, 73, 0, 206, 32, 73, 0, 206, 32, 73, 0, 193, 146, 196, 252, 193, 148, 196, 252, 193, 150, 196, 252, 193, 202, 196, 252, 193, 204, 196, 252, 193, 16, 197, 252, 193, 18, 197, 252, 193, 20, 197, 252, 193, 22, 197, 252, 193, 98, 197, 252, 193, 100, 197, 252, 193, 102, 197, 252, 193, 104, 197, 252, 193, 164, 197, 252, 193, 166, 197, 252, 193, 168, 197, 252, 193, 170, 197, 252, 193, 172, 197, 252, 193, 214, 197, 252, 193, 216, 197, 252, 193, 28, 198, 252, 193, 30, 198, 252, 193, 50, 198, 252, 193, 52, 198, 252, 193, 64, 198, 252, 193, 66, 198, 252, 193, 82, 198, 252, 193, 150, 198, 252, 193, 166, 198, 252, 193, 168, 198, 252, 193, 176, 198, 252, 193, 254, 198, 252, 193, 212, 196, 252, 193, 214, 196, 252, 193, 216, 196, 252, 193, 218, 196, 252, 193, 220, 196, 252, 193, 26, 197, 252, 193, 28, 197, 252, 193, 30, 197, 252, 193, 32, 197, 252, 193, 34, 197, 252, 193, 36, 197, 252, 193, 38, 197, 252, 193, 40, 197, 252, 193, 42, 197, 252, 193, 106, 197, 252, 193, 108, 197, 252, 193, 110, 197, 252, 193, 112, 197, 252, 193, 176, 197, 252, 193, 178, 197, 252, 193, 180, 197, 252, 193, 182, 197, 252, 193, 184, 197, 252, 193, 218, 197, 252, 193, 220, 197, 252, 193, 222, 197, 252, 193, 224, 197, 252, 193, 226, 197, 252, 193, 186, 197, 252, 193, 16, 198, 252, 193, 18, 198, 252, 193, 32, 198, 252, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 206, 64, 73, 0, 193, 118, 197, 252, 193, 120, 197, 252, 193, 122, 197, 252, 193, 124, 197, 252, 193, 126, 197, 252, 193, 128, 197, 252, 193, 130, 197, 252, 193, 132, 197, 252, 193, 134, 197, 252, 193, 136, 197, 252, 193, 138, 197, 252, 193, 92, 197, 252, 193, 140, 197, 252, 193, 142, 197, 252, 193, 144, 197, 252, 193, 146, 197, 252, 193, 148, 197, 252, 193, 150, 197, 252, 193, 152, 197, 252, 193, 188, 197, 252, 193, 190, 197, 252, 193, 192, 197, 252, 193, 194, 197, 252, 193, 196, 197, 252, 193, 198, 197, 252, 193, 200, 197, 252, 193, 202, 197, 252, 193, 204, 197, 252, 193, 206, 197, 252, 193, 234, 197, 252, 193, 236, 197, 252, 193, 238, 197, 252, 193, 240, 197, 252, 193, 242, 197, 252, 193, 244, 197, 252, 193, 246, 197, 252, 193, 248, 197, 252, 193, 250, 197, 252, 193, 252, 197, 252, 193, 254, 197, 252, 193, 2, 198, 252, 193, 4, 198, 252, 193, 6, 198, 252, 193, 8, 198, 252, 193, 20, 198, 252, 193, 22, 198, 252, 193, 154, 197, 252, 193, 10, 198, 252, 193, 24, 198, 252, 193, 26, 198, 252, 193, 38, 198, 252, 193, 40, 198, 252, 193, 42, 198, 252, 193, 44, 198, 252, 193, 56, 198, 252, 193, 68, 198, 252, 193, 58, 198, 252, 193, 60, 198, 252, 193, 46, 198, 252, 193, 48, 198, 252, 193, 62, 198, 252, 193, 70, 198, 252, 193, 72, 198, 252, 193, 74, 198, 252, 193, 80, 198, 252, 193, 92, 198, 252, 193, 94, 198, 252, 193, 102, 198, 252, 193, 104, 198, 252, 193, 106, 198, 252, 193, 118, 198, 252, 193, 120, 198, 252, 193, 122, 198, 252, 193, 124, 198, 252, 193, 132, 198, 252, 193, 134, 198, 252, 193, 136, 198, 252, 193, 138, 198, 252, 193, 140, 198, 252, 193, 142, 198, 252, 193, 34, 198, 252, 193, 84, 198, 252, 193, 86, 198, 252, 193, 88, 198, 252, 193, 112, 198, 252, 193, 114, 198, 252, 193, 116, 198, 252, 193, 126, 198, 252, 193, 128, 198, 252, 193, 130, 198, 252, 193, 154, 198, 252, 193, 156, 198, 252, 193, 214, 198, 252, 193, 224, 198, 252, 193, 226, 198, 252, 193, 228, 198, 252, 193, 230, 198, 252, 193, 232, 198, 252, 193, 40, 199, 252, 193, 44, 199, 252, 193, 54, 199, 252, 193, 64, 199, 252, 193, 66, 199, 252, 193, 68, 199, 252, 193, 70, 199, 252, 193, 72, 199, 252, 193, 82, 199, 252, 193, 84, 199, 252, 193, 86, 199, 252, 193, 88, 199, 252, 193, 90, 199, 252, 193, 92, 199, 252, 193, 164, 198, 252, 193, 170, 198, 252, 193, 178, 198, 252, 193, 194, 198, 252, 193, 184, 198, 252, 193, 186, 198, 252, 193, 190, 198, 252, 193, 192, 198, 252, 193, 196, 198, 252, 193, 198, 198, 252, 193, 202, 198, 252, 193, 208, 198, 252, 193, 210, 198, 252, 193, 212, 198, 252, 193, 218, 198, 252, 193, 220, 198, 252, 193, 216, 198, 252, 193, 222, 198, 252, 193, 242, 198, 252, 193, 244, 198, 252, 193, 246, 198, 252, 193, 248, 198, 252, 193, 250, 198, 252, 193, 252, 198, 252, 193, 6, 199, 252, 193, 8, 199, 252, 193, 10, 199, 252, 193, 12, 199, 252, 193, 14, 199, 252, 193, 18, 199, 252, 193, 20, 199, 252, 193, 26, 199, 252, 193, 236, 198, 252, 193, 240, 198, 252, 193, 56, 199, 252, 193, 94, 199, 252, 193, 136, 199, 252, 193, 138, 199, 252, 193, 156, 199, 252, 193, 158, 199, 252, 193, 160, 199, 252, 193, 162, 199, 252, 193, 164, 199, 252, 193, 192, 199, 252, 193, 194, 199, 252, 193, 244, 199, 252, 193, 246, 199, 252, 193, 248, 199, 252, 193, 16, 200, 252, 193, 58, 200, 252, 193, 78, 200, 252, 193, 150, 200, 252, 193, 152, 200, 252, 193, 228, 200, 252, 193, 230, 200, 252, 193, 40, 201, 252, 193, 100, 201, 252, 193, 102, 201, 252, 193, 188, 201, 252, 193, 190, 201, 252, 193, 192, 201, 252, 193, 194, 201, 252, 193, 196, 201, 252, 193, 28, 202, 252, 193, 22, 199, 252, 193, 24, 199, 252, 193, 36, 199, 252, 193, 52, 199, 252, 193, 76, 199, 252, 193, 104, 199, 252, 193, 106, 199, 252, 193, 108, 199, 252, 193, 128, 199, 252, 193, 130, 199, 252, 193, 148, 199, 252, 193, 38, 200, 252, 193, 122, 200, 252, 193, 124, 200, 252, 193, 196, 200, 252, 193, 16, 201, 252, 193, 160, 201, 252, 193, 162, 201, 252, 193, 164, 201, 252, 193, 4, 202, 252, 193, 6, 202, 252, 193, 8, 202, 252, 193, 94, 202, 252, 193, 96, 202, 252, 193, 98, 202, 252, 193, 162, 202, 252, 193, 216, 202, 252, 193, 26, 203, 252, 193, 82, 203, 252, 193, 84, 203, 252, 193, 96, 203, 252, 193, 98, 203, 252, 193, 28, 199, 252, 193, 30, 199, 252, 193, 34, 199, 252, 193, 38, 199, 252, 193, 42, 199, 252, 193, 46, 199, 252, 193, 48, 199, 252, 193, 50, 199, 252, 193, 58, 199, 252, 193, 60, 199, 252, 193, 62, 199, 252, 193, 74, 199, 252, 193, 96, 199, 252, 193, 98, 199, 252, 193, 100, 199, 252, 193, 112, 199, 252, 193, 114, 199, 252, 193, 116, 199, 252, 193, 118, 199, 252, 193, 120, 199, 252, 193, 140, 199, 252, 193, 142, 199, 252, 193, 144, 199, 252, 193, 166, 199, 252, 193, 168, 199, 252, 193, 170, 199, 252, 193, 172, 199, 252, 193, 174, 199, 252, 193, 176, 199, 252, 193, 178, 199, 252, 193, 180, 199, 252, 193, 182, 199, 252, 193, 110, 199, 252, 193, 132, 199, 252, 193, 134, 199, 252, 193, 150, 199, 252, 193, 152, 199, 252, 193, 154, 199, 252, 193, 188, 199, 252, 193, 190, 199, 252, 193, 216, 199, 252, 193, 218, 199, 252, 193, 220, 199, 252, 193, 240, 199, 252, 193, 242, 199, 252, 193, 12, 200, 252, 193, 14, 200, 252, 193, 54, 200, 252, 193, 56, 200, 252, 193, 64, 200, 252, 193, 66, 200, 252, 193, 74, 200, 252, 193, 76, 200, 252, 193, 82, 200, 252, 193, 104, 200, 252, 193, 106, 200, 252, 193, 128, 200, 252, 193, 130, 200, 252, 193, 132, 200, 252, 193, 134, 200, 252, 193, 136, 200, 252, 193, 138, 200, 252, 193, 140, 200, 252, 193, 142, 200, 252, 193, 184, 199, 252, 193, 186, 199, 252, 193, 196, 199, 252, 193, 198, 199, 252, 193, 200, 199, 252, 193, 202, 199, 252, 193, 204, 199, 252, 193, 206, 199, 252, 193, 208, 199, 252, 193, 210, 199, 252, 193, 212, 199, 252, 193, 222, 199, 252, 193, 224, 199, 252, 193, 226, 199, 252, 193, 228, 199, 252, 193, 230, 199, 252, 193, 232, 199, 252, 193, 234, 199, 252, 193, 236, 199, 252, 193, 250, 199, 252, 193, 252, 199, 252, 193, 254, 199, 252, 193, 2, 200, 252, 193, 4, 200, 252, 193, 6, 200, 252, 193, 8, 200, 252, 193, 10, 200, 252, 193, 18, 200, 252, 193, 20, 200, 252, 193, 22, 200, 252, 193, 24, 200, 252, 193, 40, 200, 252, 193, 26, 200, 252, 193, 28, 200, 252, 193, 30, 200, 252, 193, 32, 200, 252, 193, 34, 200, 252, 193, 42, 200, 252, 193, 44, 200, 252, 193, 46, 200, 252, 193, 48, 200, 252, 193, 60, 200, 252, 193, 62, 200, 252, 193, 68, 200, 252, 193, 70, 200, 252, 193, 72, 200, 252, 193, 80, 200, 252, 193, 84, 200, 252, 193, 86, 200, 252, 193, 88, 200, 252, 193, 90, 200, 252, 193, 92, 200, 252, 193, 94, 200, 252, 193, 96, 200, 252, 193, 98, 200, 252, 193, 108, 200, 252, 193, 110, 200, 252, 193, 112, 200, 252, 193, 114, 200, 252, 193, 116, 200, 252, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 193, 100, 200, 252, 193, 194, 200, 252, 193, 82, 201, 252, 193, 84, 201, 252, 193, 146, 201, 252, 193, 148, 201, 252, 193, 150, 201, 252, 193, 248, 201, 252, 193, 250, 201, 252, 193, 88, 202, 252, 193, 148, 202, 252, 193, 150, 202, 252, 193, 152, 202, 252, 193, 154, 202, 252, 193, 212, 202, 252, 193, 214, 202, 252, 193, 18, 203, 252, 193, 20, 203, 252, 193, 54, 203, 252, 193, 124, 203, 252, 193, 180, 203, 252, 193, 196, 203, 252, 193, 24, 205, 252, 193, 26, 205, 252, 193, 28, 205, 252, 193, 20, 206, 252, 193, 22, 206, 252, 193, 24, 206, 252, 193, 26, 206, 252, 193, 20, 207, 252, 193, 22, 207, 252, 193, 254, 207, 252, 193, 144, 200, 252, 193, 146, 200, 252, 193, 148, 200, 252, 193, 206, 200, 252, 193, 208, 200, 252, 193, 210, 200, 252, 193, 212, 200, 252, 193, 214, 200, 252, 193, 216, 200, 252, 193, 218, 200, 252, 193, 220, 200, 252, 193, 222, 200, 252, 193, 224, 200, 252, 193, 226, 200, 252, 193, 24, 201, 252, 193, 26, 201, 252, 193, 28, 201, 252, 193, 30, 201, 252, 193, 32, 201, 252, 193, 34, 201, 252, 193, 36, 201, 252, 193, 92, 201, 252, 193, 94, 201, 252, 193, 96, 201, 252, 193, 98, 201, 252, 193, 166, 201, 252, 193, 168, 201, 252, 193, 170, 201, 252, 193, 172, 201, 252, 193, 174, 201, 252, 193, 176, 201, 252, 193, 178, 201, 252, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 206, 96, 73, 0, 193, 232, 200, 252, 193, 234, 200, 252, 193, 236, 200, 252, 193, 238, 200, 252, 193, 240, 200, 252, 193, 242, 200, 252, 193, 244, 200, 252, 193, 246, 200, 252, 193, 248, 200, 252, 193, 250, 200, 252, 193, 252, 200, 252, 193, 254, 200, 252, 193, 2, 201, 252, 193, 4, 201, 252, 193, 6, 201, 252, 193, 8, 201, 252, 193, 38, 201, 252, 193, 180, 201, 252, 193, 182, 201, 252, 193, 184, 201, 252, 193, 10, 202, 252, 193, 12, 202, 252, 193, 14, 202, 252, 193, 16, 202, 252, 193, 18, 202, 252, 193, 20, 202, 252, 193, 22, 202, 252, 193, 24, 202, 252, 193, 100, 202, 252, 193, 102, 202, 252, 193, 218, 202, 252, 193, 164, 202, 252, 193, 166, 202, 252, 193, 168, 202, 252, 193, 170, 202, 252, 193, 172, 202, 252, 193, 220, 202, 252, 193, 222, 202, 252, 193, 224, 202, 252, 193, 226, 202, 252, 193, 30, 203, 252, 193, 56, 203, 252, 193, 58, 203, 252, 193, 60, 203, 252, 193, 62, 203, 252, 193, 100, 203, 252, 193, 102, 203, 252, 193, 128, 203, 252, 193, 42, 201, 252, 193, 44, 201, 252, 193, 46, 201, 252, 193, 104, 201, 252, 193, 48, 201, 252, 193, 50, 201, 252, 193, 52, 201, 252, 193, 54, 201, 252, 193, 56, 201, 252, 193, 58, 201, 252, 193, 60, 201, 252, 193, 62, 201, 252, 193, 64, 201, 252, 193, 66, 201, 252, 193, 68, 201, 252, 193, 70, 201, 252, 193, 72, 201, 252, 193, 74, 201, 252, 193, 76, 201, 252, 193, 78, 201, 252, 193, 80, 201, 252, 206, 128, 73, 0, 206, 128, 73, 0, 206, 128, 73, 0, 206, 128, 73, 0, 206, 128, 73, 0, 206, 128, 73, 0, 206, 128, 73, 0, 206, 128, 73, 0, 206, 128, 73, 0, 206, 128, 73, 0, 206, 128, 73, 0, 193, 198, 201, 252, 193, 200, 201, 252, 193, 202, 201, 252, 193, 204, 201, 252, 193, 206, 201, 252, 193, 208, 201, 252, 193, 210, 201, 252, 193, 212, 201, 252, 193, 214, 201, 252, 193, 216, 201, 252, 193, 218, 201, 252, 193, 220, 201, 252, 193, 222, 201, 252, 193, 224, 201, 252, 193, 226, 201, 252, 193, 144, 201, 252, 193, 228, 201, 252, 193, 230, 201, 252, 193, 232, 201, 252, 193, 234, 201, 252, 193, 236, 201, 252, 193, 238, 201, 252, 193, 240, 201, 252, 193, 242, 201, 252, 193, 244, 201, 252, 193, 246, 201, 252, 193, 34, 202, 252, 193, 36, 202, 252, 193, 38, 202, 252, 193, 40, 202, 252, 193, 42, 202, 252, 193, 44, 202, 252, 193, 46, 202, 252, 193, 48, 202, 252, 193, 50, 202, 252, 193, 52, 202, 252, 193, 54, 202, 252, 193, 56, 202, 252, 193, 58, 202, 252, 193, 60, 202, 252, 193, 62, 202, 252, 193, 64, 202, 252, 193, 66, 202, 252, 193, 68, 202, 252, 193, 70, 202, 252, 193, 72, 202, 252, 193, 74, 202, 252, 193, 76, 202, 252, 193, 78, 202, 252, 193, 80, 202, 252, 193, 82, 202, 252, 193, 84, 202, 252, 206, 160, 73, 0, 206, 160, 73, 0, 206, 160, 73, 0, 206, 160, 73, 0, 193, 2, 202, 252, 193, 90, 202, 252, 193, 92, 202, 252, 193, 156, 202, 252, 193, 158, 202, 252, 193, 160, 202, 252, 193, 22, 203, 252, 193, 24, 203, 252, 193, 80, 203, 252, 193, 126, 203, 252, 193, 140, 203, 252, 193, 182, 203, 252, 193, 204, 203, 252, 193, 212, 203, 252, 193, 214, 203, 252, 193, 84, 204, 252, 193, 30, 205, 252, 193, 32, 205, 252, 193, 34, 205, 252, 193, 36, 205, 252, 193, 28, 206, 252, 193, 30, 206, 252, 193, 32, 206, 252, 193, 24, 207, 252, 193, 26, 207, 252, 193, 28, 207, 252, 193, 30, 207, 252, 193, 6, 208, 252, 193, 8, 208, 252, 193, 10, 208, 252, 193, 12, 208, 252, 193, 14, 208, 252, 193, 30, 202, 252, 193, 104, 202, 252, 193, 106, 202, 252, 193, 108, 202, 252, 193, 174, 202, 252, 193, 176, 202, 252, 193, 178, 202, 252, 193, 180, 202, 252, 193, 230, 202, 252, 193, 64, 203, 252, 193, 104, 203, 252, 193, 174, 203, 252, 193, 244, 203, 252, 193, 246, 203, 252, 193, 58, 204, 252, 193, 60, 204, 252, 193, 166, 204, 252, 193, 168, 204, 252, 193, 170, 204, 252, 193, 172, 204, 252, 193, 174, 204, 252, 193, 176, 204, 252, 193, 178, 204, 252, 193, 180, 204, 252, 193, 182, 204, 252, 193, 184, 204, 252, 193, 186, 204, 252, 193, 188, 204, 252, 193, 190, 204, 252, 193, 192, 204, 252, 193, 152, 205, 252, 193, 154, 205, 252, 206, 160, 73, 0, 206, 160, 73, 0, 206, 160, 73, 0, 206, 160, 73, 0, 206, 160, 73, 0, 206, 160, 73, 0, 206, 160, 73, 0, 193, 86, 202, 252, 193, 132, 202, 252, 193, 134, 202, 252, 193, 136, 202, 252, 193, 138, 202, 252, 193, 140, 202, 252, 193, 142, 202, 252, 193, 144, 202, 252, 193, 146, 202, 252, 193, 182, 202, 252, 193, 184, 202, 252, 193, 186, 202, 252, 193, 188, 202, 252, 193, 190, 202, 252, 193, 192, 202, 252, 193, 194, 202, 252, 193, 196, 202, 252, 193, 198, 202, 252, 193, 200, 202, 252, 193, 202, 202, 252, 193, 204, 202, 252, 193, 206, 202, 252, 193, 208, 202, 252, 193, 210, 202, 252, 193, 232, 202, 252, 193, 234, 202, 252, 193, 236, 202, 252, 193, 238, 202, 252, 193, 240, 202, 252, 193, 242, 202, 252, 193, 244, 202, 252, 193, 246, 202, 252, 193, 248, 202, 252, 193, 250, 202, 252, 193, 252, 202, 252, 193, 254, 202, 252, 193, 2, 203, 252, 193, 4, 203, 252, 193, 6, 203, 252, 193, 8, 203, 252, 193, 10, 203, 252, 193, 12, 203, 252, 193, 14, 203, 252, 193, 16, 203, 252, 193, 32, 203, 252, 193, 34, 203, 252, 193, 36, 203, 252, 193, 38, 203, 252, 193, 40, 203, 252, 193, 42, 203, 252, 193, 44, 203, 252, 193, 46, 203, 252, 193, 48, 203, 252, 193, 50, 203, 252, 193, 52, 203, 252, 193, 66, 203, 252, 193, 68, 203, 252, 193, 70, 203, 252, 193, 72, 203, 252, 193, 74, 203, 252, 193, 76, 203, 252, 193, 78, 203, 252, 193, 86, 203, 252, 193, 88, 203, 252, 193, 90, 203, 252, 193, 92, 203, 252, 193, 94, 203, 252, 193, 106, 203, 252, 193, 108, 203, 252, 193, 110, 203, 252, 193, 112, 203, 252, 193, 114, 203, 252, 193, 116, 203, 252, 193, 120, 203, 252, 193, 122, 203, 252, 193, 132, 203, 252, 193, 138, 203, 252, 193, 144, 203, 252, 193, 154, 203, 252, 193, 156, 203, 252, 193, 158, 203, 252, 193, 162, 203, 252, 193, 164, 203, 252, 193, 166, 203, 252, 193, 176, 203, 252, 193, 178, 203, 252, 193, 190, 203, 252, 193, 192, 203, 252, 193, 194, 203, 252, 193, 118, 203, 252, 193, 134, 203, 252, 193, 142, 203, 252, 193, 146, 203, 252, 193, 184, 203, 252, 193, 206, 203, 252, 193, 208, 203, 252, 193, 228, 203, 252, 193, 86, 204, 252, 193, 88, 204, 252, 193, 90, 204, 252, 193, 38, 205, 252, 193, 34, 206, 252, 193, 36, 206, 252, 193, 38, 206, 252, 193, 40, 206, 252, 193, 32, 207, 252, 193, 34, 207, 252, 193, 36, 207, 252, 193, 38, 207, 252, 193, 20, 208, 252, 193, 22, 208, 252, 193, 24, 208, 252, 193, 26, 208, 252, 193, 28, 208, 252, 193, 30, 208, 252, 193, 80, 209, 252, 193, 82, 209, 252, 193, 84, 209, 252, 193, 86, 209, 252, 193, 88, 209, 252, 193, 90, 209, 252, 193, 130, 203, 252, 193, 148, 203, 252, 193, 150, 203, 252, 193, 152, 203, 252, 193, 168, 203, 252, 193, 170, 203, 252, 193, 172, 203, 252, 193, 186, 203, 252, 193, 188, 203, 252, 193, 226, 203, 252, 193, 230, 203, 252, 193, 234, 203, 252, 193, 236, 203, 252, 193, 238, 203, 252, 193, 240, 203, 252, 193, 242, 203, 252, 193, 24, 204, 252, 193, 26, 204, 252, 193, 28, 204, 252, 193, 30, 204, 252, 193, 32, 204, 252, 193, 34, 204, 252, 193, 36, 204, 252, 193, 38, 204, 252, 193, 40, 204, 252, 193, 42, 204, 252, 193, 44, 204, 252, 193, 46, 204, 252, 193, 48, 204, 252, 193, 50, 204, 252, 193, 52, 204, 252, 193, 54, 204, 252, 193, 198, 203, 252, 193, 200, 203, 252, 193, 202, 203, 252, 193, 210, 203, 252, 193, 216, 203, 252, 193, 218, 203, 252, 193, 220, 203, 252, 193, 222, 203, 252, 193, 224, 203, 252, 193, 248, 203, 252, 193, 250, 203, 252, 193, 252, 203, 252, 193, 254, 203, 252, 193, 2, 204, 252, 193, 4, 204, 252, 193, 6, 204, 252, 193, 8, 204, 252, 193, 10, 204, 252, 193, 12, 204, 252, 193, 14, 204, 252, 193, 62, 204, 252, 193, 64, 204, 252, 193, 66, 204, 252, 193, 68, 204, 252, 193, 70, 204, 252, 193, 72, 204, 252, 193, 74, 204, 252, 193, 76, 204, 252, 193, 78, 204, 252, 193, 80, 204, 252, 193, 82, 204, 252, 193, 194, 204, 252, 193, 18, 204, 252, 193, 20, 204, 252, 193, 22, 204, 252, 193, 92, 204, 252, 193, 94, 204, 252, 193, 96, 204, 252, 193, 98, 204, 252, 193, 40, 205, 252, 193, 42, 205, 252, 193, 44, 205, 252, 193, 46, 205, 252, 193, 48, 205, 252, 193, 50, 205, 252, 193, 52, 205, 252, 193, 54, 205, 252, 193, 56, 205, 252, 193, 58, 205, 252, 193, 60, 205, 252, 193, 42, 206, 252, 193, 44, 206, 252, 193, 46, 206, 252, 193, 48, 206, 252, 193, 50, 206, 252, 193, 52, 206, 252, 193, 54, 206, 252, 193, 56, 206, 252, 193, 58, 206, 252, 193, 60, 206, 252, 193, 62, 206, 252, 193, 40, 207, 252, 193, 42, 207, 252, 193, 44, 207, 252, 193, 56, 204, 252, 193, 100, 204, 252, 193, 102, 204, 252, 193, 104, 204, 252, 193, 106, 204, 252, 193, 108, 204, 252, 193, 110, 204, 252, 193, 112, 204, 252, 193, 114, 204, 252, 193, 116, 204, 252, 193, 118, 204, 252, 193, 120, 204, 252, 193, 122, 204, 252, 193, 124, 204, 252, 193, 126, 204, 252, 193, 128, 204, 252, 193, 130, 204, 252, 193, 132, 204, 252, 193, 134, 204, 252, 193, 136, 204, 252, 193, 138, 204, 252, 193, 140, 204, 252, 193, 142, 204, 252, 193, 144, 204, 252, 193, 146, 204, 252, 193, 148, 204, 252, 193, 150, 204, 252, 193, 152, 204, 252, 193, 154, 204, 252, 193, 156, 204, 252, 193, 158, 204, 252, 193, 160, 204, 252, 193, 162, 204, 252, 193, 164, 204, 252, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 192, 73, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 224, 73, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 206, 32, 74, 0, 193, 150, 205, 252, 193, 120, 206, 252, 193, 122, 206, 252, 193, 124, 206, 252, 193, 126, 206, 252, 193, 128, 206, 252, 193, 130, 206, 252, 193, 132, 206, 252, 193, 134, 206, 252, 193, 136, 206, 252, 193, 138, 206, 252, 193, 140, 206, 252, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 193, 156, 205, 252, 193, 158, 205, 252, 193, 160, 205, 252, 193, 162, 205, 252, 193, 164, 205, 252, 193, 166, 205, 252, 193, 142, 206, 252, 193, 144, 206, 252, 193, 146, 206, 252, 193, 148, 206, 252, 193, 150, 206, 252, 193, 152, 206, 252, 193, 154, 206, 252, 193, 156, 206, 252, 193, 158, 206, 252, 193, 128, 207, 252, 193, 130, 207, 252, 193, 132, 207, 252, 193, 134, 207, 252, 193, 136, 207, 252, 193, 138, 207, 252, 193, 140, 207, 252, 193, 142, 207, 252, 193, 144, 207, 252, 193, 146, 207, 252, 193, 148, 207, 252, 193, 150, 207, 252, 193, 152, 207, 252, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 206, 0, 74, 0, 193, 164, 206, 252, 193, 166, 206, 252, 193, 168, 206, 252, 193, 170, 206, 252, 193, 172, 206, 252, 193, 174, 206, 252, 193, 176, 206, 252, 193, 178, 206, 252, 193, 180, 206, 252, 193, 182, 206, 252, 193, 184, 206, 252, 193, 186, 206, 252, 193, 188, 206, 252, 193, 190, 206, 252, 193, 192, 206, 252, 193, 194, 206, 252, 193, 196, 206, 252, 193, 198, 206, 252, 193, 200, 206, 252, 193, 202, 206, 252, 193, 204, 206, 252, 193, 206, 206, 252, 193, 208, 206, 252, 193, 210, 206, 252, 193, 212, 206, 252, 193, 214, 206, 252, 193, 216, 206, 252, 193, 218, 206, 252, 193, 220, 206, 252, 193, 222, 206, 252, 193, 224, 206, 252, 193, 226, 206, 252, 193, 228, 206, 252, 193, 230, 206, 252, 193, 232, 206, 252, 193, 234, 206, 252, 193, 236, 206, 252, 193, 238, 206, 252, 193, 240, 206, 252, 193, 242, 206, 252, 193, 244, 206, 252, 193, 246, 206, 252, 193, 248, 206, 252, 193, 250, 206, 252, 193, 252, 206, 252, 193, 254, 206, 252, 193, 2, 207, 252, 193, 4, 207, 252, 193, 6, 207, 252, 193, 8, 207, 252, 193, 10, 207, 252, 193, 12, 207, 252, 193, 14, 207, 252, 193, 16, 207, 252, 193, 18, 207, 252, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 193, 46, 207, 252, 193, 48, 207, 252, 193, 50, 207, 252, 193, 52, 207, 252, 193, 54, 207, 252, 193, 32, 208, 252, 193, 34, 208, 252, 193, 36, 208, 252, 193, 38, 208, 252, 193, 40, 208, 252, 193, 42, 208, 252, 193, 94, 209, 252, 193, 96, 209, 252, 193, 98, 209, 252, 193, 100, 209, 252, 193, 102, 209, 252, 193, 104, 209, 252, 193, 106, 209, 252, 193, 108, 209, 252, 193, 110, 209, 252, 193, 112, 209, 252, 193, 120, 210, 252, 193, 122, 210, 252, 193, 124, 210, 252, 193, 126, 210, 252, 193, 128, 210, 252, 193, 130, 210, 252, 193, 132, 210, 252, 193, 134, 210, 252, 193, 136, 210, 252, 193, 138, 210, 252, 193, 140, 210, 252, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 64, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 96, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 206, 192, 74, 0, 193, 2, 208, 252, 193, 4, 208, 252, 193, 60, 209, 252, 193, 62, 209, 252, 193, 64, 209, 252, 193, 66, 209, 252, 193, 96, 210, 252, 193, 98, 210, 252, 193, 100, 210, 252, 193, 80, 211, 252, 193, 82, 211, 252, 193, 84, 211, 252, 193, 86, 211, 252, 193, 88, 211, 252, 193, 90, 211, 252, 193, 32, 212, 252, 193, 34, 212, 252, 193, 36, 212, 252, 193, 38, 212, 252, 193, 40, 212, 252, 193, 42, 212, 252, 193, 12, 213, 252, 193, 14, 213, 252, 193, 16, 213, 252, 193, 18, 213, 252, 193, 204, 213, 252, 193, 206, 213, 252, 193, 208, 213, 252, 193, 50, 214, 252, 193, 52, 214, 252, 193, 142, 214, 252, 193, 144, 214, 252, 193, 16, 208, 252, 193, 18, 208, 252, 193, 68, 209, 252, 193, 70, 209, 252, 193, 72, 209, 252, 193, 74, 209, 252, 193, 76, 209, 252, 193, 78, 209, 252, 193, 102, 210, 252, 193, 104, 210, 252, 193, 106, 210, 252, 193, 108, 210, 252, 193, 94, 211, 252, 193, 96, 211, 252, 193, 98, 211, 252, 193, 44, 212, 252, 193, 46, 212, 252, 193, 48, 212, 252, 193, 50, 212, 252, 193, 52, 212, 252, 193, 20, 213, 252, 193, 22, 213, 252, 193, 24, 213, 252, 193, 26, 213, 252, 193, 28, 213, 252, 193, 210, 213, 252, 193, 212, 213, 252, 193, 214, 213, 252, 193, 54, 214, 252, 193, 56, 214, 252, 193, 214, 214, 252, 193, 252, 214, 252, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 128, 74, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 206, 160, 74, 0, 193, 190, 209, 252, 193, 192, 209, 252, 193, 194, 209, 252, 193, 196, 209, 252, 193, 198, 209, 252, 193, 200, 209, 252, 193, 202, 209, 252, 193, 204, 209, 252, 193, 206, 209, 252, 193, 208, 209, 252, 193, 210, 209, 252, 193, 212, 209, 252, 193, 214, 209, 252, 193, 216, 209, 252, 193, 218, 209, 252, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 224, 74, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 193, 92, 209, 252, 193, 110, 210, 252, 193, 112, 210, 252, 193, 114, 210, 252, 193, 116, 210, 252, 193, 118, 210, 252, 193, 100, 211, 252, 193, 102, 211, 252, 193, 104, 211, 252, 193, 106, 211, 252, 193, 108, 211, 252, 193, 110, 211, 252, 193, 112, 211, 252, 193, 114, 211, 252, 193, 116, 211, 252, 193, 118, 211, 252, 193, 120, 211, 252, 193, 54, 212, 252, 193, 56, 212, 252, 193, 58, 212, 252, 193, 60, 212, 252, 193, 62, 212, 252, 193, 64, 212, 252, 193, 66, 212, 252, 193, 68, 212, 252, 193, 70, 212, 252, 193, 30, 213, 252, 193, 32, 213, 252, 193, 34, 213, 252, 193, 36, 213, 252, 193, 38, 213, 252, 193, 40, 213, 252, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 206, 0, 75, 0, 193, 144, 210, 252, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 32, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 193, 92, 210, 252, 193, 94, 210, 252, 193, 74, 211, 252, 193, 76, 211, 252, 193, 78, 211, 252, 193, 200, 211, 252, 193, 202, 211, 252, 193, 204, 211, 252, 193, 206, 211, 252, 193, 208, 211, 252, 193, 210, 211, 252, 193, 212, 211, 252, 193, 214, 211, 252, 193, 142, 210, 252, 193, 122, 211, 252, 193, 124, 211, 252, 193, 126, 211, 252, 193, 128, 211, 252, 193, 130, 211, 252, 193, 132, 211, 252, 193, 134, 211, 252, 193, 136, 211, 252, 193, 138, 211, 252, 193, 140, 211, 252, 193, 142, 211, 252, 193, 72, 212, 252, 193, 74, 212, 252, 193, 76, 212, 252, 193, 78, 212, 252, 193, 80, 212, 252, 193, 82, 212, 252, 193, 84, 212, 252, 193, 86, 212, 252, 193, 42, 213, 252, 193, 44, 213, 252, 193, 46, 213, 252, 193, 48, 213, 252, 193, 50, 213, 252, 193, 52, 213, 252, 193, 54, 213, 252, 193, 56, 213, 252, 193, 58, 213, 252, 193, 60, 213, 252, 193, 220, 213, 252, 193, 222, 213, 252, 193, 146, 210, 252, 193, 148, 210, 252, 193, 150, 210, 252, 193, 152, 210, 252, 193, 154, 210, 252, 193, 156, 210, 252, 193, 158, 210, 252, 193, 160, 210, 252, 193, 162, 210, 252, 193, 164, 210, 252, 193, 166, 210, 252, 193, 168, 210, 252, 193, 170, 210, 252, 193, 172, 210, 252, 193, 174, 210, 252, 193, 176, 210, 252, 193, 178, 210, 252, 193, 180, 210, 252, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 128, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 206, 64, 75, 0, 193, 180, 211, 252, 193, 182, 211, 252, 193, 184, 211, 252, 193, 186, 211, 252, 193, 188, 211, 252, 193, 190, 211, 252, 193, 192, 211, 252, 193, 194, 211, 252, 193, 196, 211, 252, 193, 198, 211, 252, 193, 126, 212, 252, 193, 128, 212, 252, 193, 130, 212, 252, 193, 132, 212, 252, 193, 134, 212, 252, 193, 136, 212, 252, 193, 138, 212, 252, 193, 140, 212, 252, 193, 142, 212, 252, 193, 106, 213, 252, 193, 108, 213, 252, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 206, 96, 75, 0, 193, 216, 211, 252, 193, 218, 211, 252, 193, 220, 211, 252, 193, 222, 211, 252, 193, 224, 211, 252, 193, 226, 211, 252, 193, 228, 211, 252, 193, 230, 211, 252, 193, 232, 211, 252, 193, 234, 211, 252, 193, 236, 211, 252, 193, 238, 211, 252, 193, 240, 211, 252, 193, 242, 211, 252, 193, 244, 211, 252, 193, 246, 211, 252, 193, 248, 211, 252, 193, 250, 211, 252, 193, 252, 211, 252, 193, 254, 211, 252, 193, 2, 212, 252, 193, 4, 212, 252, 193, 6, 212, 252, 193, 8, 212, 252, 193, 10, 212, 252, 193, 12, 212, 252, 193, 14, 212, 252, 193, 16, 212, 252, 193, 18, 212, 252, 193, 20, 212, 252, 193, 22, 212, 252, 193, 24, 212, 252, 193, 26, 212, 252, 193, 28, 212, 252, 193, 30, 212, 252, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 206, 160, 75, 0, 193, 114, 213, 252, 193, 116, 213, 252, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 206, 224, 75, 0, 193, 10, 213, 252, 193, 202, 213, 252, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 206, 192, 75, 0, 193, 226, 213, 252, 193, 228, 213, 252, 193, 230, 213, 252, 193, 232, 213, 252, 193, 234, 213, 252, 193, 236, 213, 252, 193, 238, 213, 252, 193, 74, 214, 252, 193, 76, 214, 252, 193, 160, 214, 252, 193, 78, 214, 252, 193, 80, 214, 252, 193, 162, 214, 252, 193, 164, 214, 252, 193, 166, 214, 252, 193, 168, 214, 252, 193, 224, 214, 252, 193, 254, 214, 252, 193, 2, 215, 252, 193, 110, 213, 252, 193, 112, 213, 252, 193, 240, 213, 252, 193, 242, 213, 252, 193, 244, 213, 252, 193, 246, 213, 252, 193, 82, 214, 252, 193, 84, 214, 252, 193, 86, 214, 252, 193, 170, 214, 252, 193, 172, 214, 252, 193, 228, 214, 252, 193, 230, 214, 252, 193, 232, 214, 252, 193, 234, 214, 252, 193, 4, 215, 252, 193, 124, 215, 252, 193, 200, 215, 252, 193, 164, 216, 252, 193, 208, 216, 252, 193, 54, 217, 252, 193, 78, 217, 252, 193, 148, 217, 252, 193, 178, 217, 252, 193, 212, 217, 252, 193, 126, 218, 252, 193, 128, 218, 252, 193, 150, 218, 252, 193, 248, 218, 252, 193, 102, 219, 252, 193, 108, 219, 252, 193, 114, 219, 252, 193, 216, 213, 252, 193, 218, 213, 252, 193, 58, 214, 252, 193, 60, 214, 252, 193, 62, 214, 252, 193, 64, 214, 252, 193, 66, 214, 252, 193, 146, 214, 252, 193, 148, 214, 252, 193, 150, 214, 252, 193, 216, 214, 252, 193, 68, 215, 252, 193, 94, 215, 252, 193, 102, 215, 252, 193, 132, 215, 252, 193, 134, 215, 252, 193, 168, 215, 252, 193, 170, 215, 252, 193, 172, 215, 252, 193, 188, 215, 252, 193, 16, 216, 252, 193, 18, 216, 252, 193, 30, 216, 252, 193, 32, 216, 252, 193, 34, 216, 252, 193, 54, 216, 252, 193, 60, 216, 252, 193, 62, 216, 252, 193, 72, 216, 252, 193, 136, 216, 252, 193, 138, 216, 252, 193, 146, 216, 252, 193, 224, 213, 252, 193, 68, 214, 252, 193, 70, 214, 252, 193, 72, 214, 252, 193, 152, 214, 252, 193, 154, 214, 252, 193, 156, 214, 252, 193, 158, 214, 252, 193, 218, 214, 252, 193, 220, 214, 252, 193, 222, 214, 252, 193, 22, 215, 252, 193, 74, 215, 252, 193, 86, 215, 252, 193, 106, 215, 252, 193, 122, 215, 252, 193, 136, 215, 252, 193, 152, 215, 252, 193, 174, 215, 252, 193, 176, 215, 252, 193, 242, 215, 252, 193, 36, 216, 252, 193, 56, 216, 252, 193, 124, 216, 252, 193, 172, 217, 252, 193, 174, 217, 252, 193, 206, 217, 252, 193, 18, 218, 252, 193, 62, 218, 252, 193, 168, 218, 252, 193, 170, 218, 252, 193, 172, 218, 252, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 0, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 206, 32, 76, 0, 193, 174, 214, 252, 193, 176, 214, 252, 193, 178, 214, 252, 193, 180, 214, 252, 193, 182, 214, 252, 193, 184, 214, 252, 193, 186, 214, 252, 193, 188, 214, 252, 193, 190, 214, 252, 193, 192, 214, 252, 193, 194, 214, 252, 193, 196, 214, 252, 193, 198, 214, 252, 193, 200, 214, 252, 193, 202, 214, 252, 193, 204, 214, 252, 193, 206, 214, 252, 193, 208, 214, 252, 193, 210, 214, 252, 193, 236, 214, 252, 193, 238, 214, 252, 193, 240, 214, 252, 193, 242, 214, 252, 193, 244, 214, 252, 193, 246, 214, 252, 193, 248, 214, 252, 193, 6, 215, 252, 193, 8, 215, 252, 193, 10, 215, 252, 193, 12, 215, 252, 193, 14, 215, 252, 193, 16, 215, 252, 193, 18, 215, 252, 193, 32, 215, 252, 193, 34, 215, 252, 193, 36, 215, 252, 193, 38, 215, 252, 193, 40, 215, 252, 193, 42, 215, 252, 193, 44, 215, 252, 193, 46, 215, 252, 193, 48, 215, 252, 193, 50, 215, 252, 193, 52, 215, 252, 193, 56, 215, 252, 193, 58, 215, 252, 193, 60, 215, 252, 193, 62, 215, 252, 193, 64, 215, 252, 193, 70, 215, 252, 193, 72, 215, 252, 193, 76, 215, 252, 193, 78, 215, 252, 193, 80, 215, 252, 193, 82, 215, 252, 193, 84, 215, 252, 193, 96, 215, 252, 193, 92, 215, 252, 193, 104, 215, 252, 193, 24, 215, 252, 193, 26, 215, 252, 193, 28, 215, 252, 193, 90, 215, 252, 193, 98, 215, 252, 193, 100, 215, 252, 193, 138, 215, 252, 193, 158, 215, 252, 193, 190, 215, 252, 193, 192, 215, 252, 193, 222, 215, 252, 193, 226, 215, 252, 193, 244, 215, 252, 193, 246, 215, 252, 193, 38, 216, 252, 193, 40, 216, 252, 193, 42, 216, 252, 193, 74, 216, 252, 193, 90, 216, 252, 193, 152, 216, 252, 193, 162, 216, 252, 193, 190, 216, 252, 193, 192, 216, 252, 193, 194, 216, 252, 193, 196, 216, 252, 193, 206, 216, 252, 193, 214, 216, 252, 193, 234, 216, 252, 193, 10, 217, 252, 193, 20, 217, 252, 193, 22, 217, 252, 193, 32, 217, 252, 193, 54, 215, 252, 193, 66, 215, 252, 193, 118, 215, 252, 193, 120, 215, 252, 193, 130, 215, 252, 193, 150, 215, 252, 193, 156, 215, 252, 193, 182, 215, 252, 193, 184, 215, 252, 193, 198, 215, 252, 193, 212, 215, 252, 193, 216, 215, 252, 193, 28, 216, 252, 193, 70, 216, 252, 193, 106, 216, 252, 193, 122, 216, 252, 193, 134, 216, 252, 193, 160, 216, 252, 193, 176, 216, 252, 193, 202, 216, 252, 193, 246, 216, 252, 193, 248, 216, 252, 193, 4, 217, 252, 193, 8, 217, 252, 193, 46, 217, 252, 193, 48, 217, 252, 193, 140, 217, 252, 193, 200, 217, 252, 193, 230, 217, 252, 193, 250, 217, 252, 193, 16, 218, 252, 193, 34, 218, 252, 193, 108, 215, 252, 193, 110, 215, 252, 193, 112, 215, 252, 193, 114, 215, 252, 193, 126, 215, 252, 193, 128, 215, 252, 193, 140, 215, 252, 193, 142, 215, 252, 193, 144, 215, 252, 193, 146, 215, 252, 193, 186, 215, 252, 193, 160, 215, 252, 193, 162, 215, 252, 193, 178, 215, 252, 193, 164, 215, 252, 193, 180, 215, 252, 193, 194, 215, 252, 193, 196, 215, 252, 193, 202, 215, 252, 193, 204, 215, 252, 193, 208, 215, 252, 193, 210, 215, 252, 193, 214, 215, 252, 193, 218, 215, 252, 193, 220, 215, 252, 193, 224, 215, 252, 193, 228, 215, 252, 193, 230, 215, 252, 193, 232, 215, 252, 193, 234, 215, 252, 193, 236, 215, 252, 193, 238, 215, 252, 193, 240, 215, 252, 193, 248, 215, 252, 193, 250, 215, 252, 193, 252, 215, 252, 193, 254, 215, 252, 193, 2, 216, 252, 193, 4, 216, 252, 193, 6, 216, 252, 193, 8, 216, 252, 193, 10, 216, 252, 193, 12, 216, 252, 193, 14, 216, 252, 193, 20, 216, 252, 193, 22, 216, 252, 193, 24, 216, 252, 193, 26, 216, 252, 193, 44, 216, 252, 193, 46, 216, 252, 193, 48, 216, 252, 193, 50, 216, 252, 193, 58, 216, 252, 193, 64, 216, 252, 193, 66, 216, 252, 193, 68, 216, 252, 193, 76, 216, 252, 193, 78, 216, 252, 193, 80, 216, 252, 193, 82, 216, 252, 193, 84, 216, 252, 193, 86, 216, 252, 193, 92, 216, 252, 193, 94, 216, 252, 193, 96, 216, 252, 193, 98, 216, 252, 193, 100, 216, 252, 193, 102, 216, 252, 193, 104, 216, 252, 193, 108, 216, 252, 193, 112, 216, 252, 193, 114, 216, 252, 193, 116, 216, 252, 193, 118, 216, 252, 193, 120, 216, 252, 193, 126, 216, 252, 193, 128, 216, 252, 193, 130, 216, 252, 193, 132, 216, 252, 193, 140, 216, 252, 193, 142, 216, 252, 193, 144, 216, 252, 193, 148, 216, 252, 193, 154, 216, 252, 193, 156, 216, 252, 193, 158, 216, 252, 193, 166, 216, 252, 193, 168, 216, 252, 193, 170, 216, 252, 193, 172, 216, 252, 193, 174, 216, 252, 193, 182, 216, 252, 193, 184, 216, 252, 193, 186, 216, 252, 193, 188, 216, 252, 193, 198, 216, 252, 193, 178, 216, 252, 193, 180, 216, 252, 193, 204, 216, 252, 193, 226, 216, 252, 193, 6, 217, 252, 193, 30, 217, 252, 193, 72, 217, 252, 193, 74, 217, 252, 193, 112, 217, 252, 193, 114, 217, 252, 193, 116, 217, 252, 193, 118, 217, 252, 193, 120, 217, 252, 193, 142, 217, 252, 193, 144, 217, 252, 193, 170, 217, 252, 193, 202, 217, 252, 193, 204, 217, 252, 193, 82, 218, 252, 193, 94, 218, 252, 193, 120, 218, 252, 193, 122, 218, 252, 193, 164, 218, 252, 193, 166, 218, 252, 193, 204, 218, 252, 193, 240, 218, 252, 193, 16, 219, 252, 193, 18, 219, 252, 193, 20, 219, 252, 193, 22, 219, 252, 193, 24, 219, 252, 193, 40, 219, 252, 193, 62, 217, 252, 193, 64, 217, 252, 193, 66, 217, 252, 193, 68, 217, 252, 193, 70, 217, 252, 193, 80, 217, 252, 193, 82, 217, 252, 193, 44, 217, 252, 193, 84, 217, 252, 193, 86, 217, 252, 193, 88, 217, 252, 193, 90, 217, 252, 193, 92, 217, 252, 193, 98, 217, 252, 193, 100, 217, 252, 193, 102, 217, 252, 193, 104, 217, 252, 193, 106, 217, 252, 193, 108, 217, 252, 193, 124, 217, 252, 193, 126, 217, 252, 193, 128, 217, 252, 193, 130, 217, 252, 193, 132, 217, 252, 193, 134, 217, 252, 193, 136, 217, 252, 193, 138, 217, 252, 206, 64, 76, 0, 206, 64, 76, 0, 206, 64, 76, 0, 206, 64, 76, 0, 206, 64, 76, 0, 193, 50, 217, 252, 193, 52, 217, 252, 193, 76, 217, 252, 193, 94, 217, 252, 193, 96, 217, 252, 193, 122, 217, 252, 193, 208, 217, 252, 193, 146, 217, 252, 193, 176, 217, 252, 193, 210, 217, 252, 193, 252, 217, 252, 193, 20, 218, 252, 193, 106, 218, 252, 193, 110, 218, 252, 193, 112, 218, 252, 193, 114, 218, 252, 193, 124, 218, 252, 193, 148, 218, 252, 193, 174, 218, 252, 193, 216, 218, 252, 193, 218, 218, 252, 193, 220, 218, 252, 193, 244, 218, 252, 193, 246, 218, 252, 193, 28, 219, 252, 193, 52, 219, 252, 193, 64, 219, 252, 193, 90, 219, 252, 193, 96, 219, 252, 193, 98, 219, 252, 193, 100, 219, 252, 193, 128, 219, 252, 193, 36, 218, 252, 193, 68, 218, 252, 193, 74, 218, 252, 193, 88, 218, 252, 193, 96, 218, 252, 193, 160, 218, 252, 193, 162, 218, 252, 193, 202, 218, 252, 193, 238, 218, 252, 193, 8, 219, 252, 193, 10, 219, 252, 193, 12, 219, 252, 193, 14, 219, 252, 193, 48, 219, 252, 193, 60, 219, 252, 193, 70, 219, 252, 193, 72, 219, 252, 193, 150, 219, 252, 193, 182, 219, 252, 193, 104, 220, 252, 193, 162, 220, 252, 193, 242, 220, 252, 193, 244, 220, 252, 193, 70, 221, 252, 193, 128, 221, 252, 193, 130, 221, 252, 193, 132, 221, 252, 193, 182, 221, 252, 193, 58, 222, 252, 193, 176, 222, 252, 193, 200, 222, 252, 193, 4, 223, 252, 193, 84, 218, 252, 193, 90, 218, 252, 193, 92, 218, 252, 193, 98, 218, 252, 193, 100, 218, 252, 193, 102, 218, 252, 193, 108, 218, 252, 193, 118, 218, 252, 193, 130, 218, 252, 193, 132, 218, 252, 193, 134, 218, 252, 193, 136, 218, 252, 193, 138, 218, 252, 193, 140, 218, 252, 193, 142, 218, 252, 193, 144, 218, 252, 193, 146, 218, 252, 193, 152, 218, 252, 193, 154, 218, 252, 193, 156, 218, 252, 193, 158, 218, 252, 193, 176, 218, 252, 193, 178, 218, 252, 193, 180, 218, 252, 193, 182, 218, 252, 193, 184, 218, 252, 193, 186, 218, 252, 193, 188, 218, 252, 193, 190, 218, 252, 193, 194, 218, 252, 193, 196, 218, 252, 193, 222, 218, 252, 193, 192, 218, 252, 193, 206, 218, 252, 193, 208, 218, 252, 193, 210, 218, 252, 193, 212, 218, 252, 193, 214, 218, 252, 193, 242, 218, 252, 193, 26, 219, 252, 193, 50, 219, 252, 193, 62, 219, 252, 193, 94, 219, 252, 193, 242, 219, 252, 193, 58, 220, 252, 193, 174, 220, 252, 193, 250, 220, 252, 193, 72, 221, 252, 193, 74, 221, 252, 193, 76, 221, 252, 193, 134, 221, 252, 193, 246, 221, 252, 193, 114, 222, 252, 193, 146, 222, 252, 193, 170, 222, 252, 193, 172, 222, 252, 193, 184, 222, 252, 193, 202, 222, 252, 193, 234, 222, 252, 193, 240, 222, 252, 193, 72, 223, 252, 193, 74, 223, 252, 193, 76, 223, 252, 193, 78, 223, 252, 193, 224, 218, 252, 193, 226, 218, 252, 193, 228, 218, 252, 193, 230, 218, 252, 193, 232, 218, 252, 193, 234, 218, 252, 193, 236, 218, 252, 193, 250, 218, 252, 193, 252, 218, 252, 193, 254, 218, 252, 193, 2, 219, 252, 193, 4, 219, 252, 193, 6, 219, 252, 193, 30, 219, 252, 193, 32, 219, 252, 193, 34, 219, 252, 193, 36, 219, 252, 193, 42, 219, 252, 193, 44, 219, 252, 193, 46, 219, 252, 193, 54, 219, 252, 193, 56, 219, 252, 193, 58, 219, 252, 193, 66, 219, 252, 193, 74, 219, 252, 193, 76, 219, 252, 193, 78, 219, 252, 193, 82, 219, 252, 193, 84, 219, 252, 193, 88, 219, 252, 193, 104, 219, 252, 193, 106, 219, 252, 193, 86, 219, 252, 193, 92, 219, 252, 193, 124, 219, 252, 193, 238, 219, 252, 193, 240, 219, 252, 193, 46, 220, 252, 193, 48, 220, 252, 193, 50, 220, 252, 193, 52, 220, 252, 193, 54, 220, 252, 193, 56, 220, 252, 193, 106, 220, 252, 193, 108, 220, 252, 193, 164, 220, 252, 193, 166, 220, 252, 193, 168, 220, 252, 193, 170, 220, 252, 193, 172, 220, 252, 193, 246, 220, 252, 193, 248, 220, 252, 193, 184, 221, 252, 193, 60, 222, 252, 193, 82, 222, 252, 193, 84, 222, 252, 193, 124, 222, 252, 193, 126, 222, 252, 193, 134, 222, 252, 193, 142, 222, 252, 193, 144, 222, 252, 193, 216, 222, 252, 193, 250, 222, 252, 193, 152, 224, 252, 193, 110, 219, 252, 193, 112, 219, 252, 193, 118, 219, 252, 193, 116, 219, 252, 193, 120, 219, 252, 193, 122, 219, 252, 193, 126, 219, 252, 193, 132, 219, 252, 193, 138, 219, 252, 193, 140, 219, 252, 193, 142, 219, 252, 193, 144, 219, 252, 193, 146, 219, 252, 193, 148, 219, 252, 193, 166, 219, 252, 193, 168, 219, 252, 193, 170, 219, 252, 193, 172, 219, 252, 193, 174, 219, 252, 193, 176, 219, 252, 193, 178, 219, 252, 193, 214, 219, 252, 193, 216, 219, 252, 193, 218, 219, 252, 193, 220, 219, 252, 193, 222, 219, 252, 193, 224, 219, 252, 193, 226, 219, 252, 193, 228, 219, 252, 193, 230, 219, 252, 193, 232, 219, 252, 193, 234, 219, 252, 193, 134, 219, 252, 193, 152, 219, 252, 193, 154, 219, 252, 193, 156, 219, 252, 193, 158, 219, 252, 193, 184, 219, 252, 193, 186, 219, 252, 193, 188, 219, 252, 193, 190, 219, 252, 193, 192, 219, 252, 193, 194, 219, 252, 193, 196, 219, 252, 193, 198, 219, 252, 193, 200, 219, 252, 193, 202, 219, 252, 193, 204, 219, 252, 193, 206, 219, 252, 193, 208, 219, 252, 193, 244, 219, 252, 193, 246, 219, 252, 193, 248, 219, 252, 193, 250, 219, 252, 193, 252, 219, 252, 193, 254, 219, 252, 193, 2, 220, 252, 193, 4, 220, 252, 193, 6, 220, 252, 193, 8, 220, 252, 193, 60, 220, 252, 193, 62, 220, 252, 193, 64, 220, 252, 193, 66, 220, 252, 193, 136, 219, 252, 193, 160, 219, 252, 193, 162, 219, 252, 193, 164, 219, 252, 193, 210, 219, 252, 193, 212, 219, 252, 193, 10, 220, 252, 193, 12, 220, 252, 193, 14, 220, 252, 193, 16, 220, 252, 193, 68, 220, 252, 193, 70, 220, 252, 193, 72, 220, 252, 193, 74, 220, 252, 193, 76, 220, 252, 193, 130, 220, 252, 193, 132, 220, 252, 193, 184, 220, 252, 193, 186, 220, 252, 193, 188, 220, 252, 193, 190, 220, 252, 193, 192, 220, 252, 193, 194, 220, 252, 193, 196, 220, 252, 193, 198, 220, 252, 193, 12, 221, 252, 193, 14, 221, 252, 193, 16, 221, 252, 193, 88, 221, 252, 193, 90, 221, 252, 193, 92, 221, 252, 193, 144, 221, 252, 193, 236, 219, 252, 193, 18, 220, 252, 193, 20, 220, 252, 193, 22, 220, 252, 193, 24, 220, 252, 193, 26, 220, 252, 193, 28, 220, 252, 193, 30, 220, 252, 193, 32, 220, 252, 193, 34, 220, 252, 193, 36, 220, 252, 193, 38, 220, 252, 193, 40, 220, 252, 193, 42, 220, 252, 193, 44, 220, 252, 193, 78, 220, 252, 193, 80, 220, 252, 193, 82, 220, 252, 193, 84, 220, 252, 193, 134, 220, 252, 193, 86, 220, 252, 193, 88, 220, 252, 193, 90, 220, 252, 193, 92, 220, 252, 193, 94, 220, 252, 193, 96, 220, 252, 193, 98, 220, 252, 193, 100, 220, 252, 193, 102, 220, 252, 193, 136, 220, 252, 193, 138, 220, 252, 193, 140, 220, 252, 193, 110, 220, 252, 193, 112, 220, 252, 193, 114, 220, 252, 193, 116, 220, 252, 193, 118, 220, 252, 193, 120, 220, 252, 193, 122, 220, 252, 193, 124, 220, 252, 193, 126, 220, 252, 193, 128, 220, 252, 193, 176, 220, 252, 193, 178, 220, 252, 193, 180, 220, 252, 193, 182, 220, 252, 193, 252, 220, 252, 193, 254, 220, 252, 193, 2, 221, 252, 193, 4, 221, 252, 193, 6, 221, 252, 193, 8, 221, 252, 193, 10, 221, 252, 193, 78, 221, 252, 193, 80, 221, 252, 193, 82, 221, 252, 193, 84, 221, 252, 193, 86, 221, 252, 193, 136, 221, 252, 193, 138, 221, 252, 193, 140, 221, 252, 193, 142, 221, 252, 193, 186, 221, 252, 193, 188, 221, 252, 193, 142, 220, 252, 193, 144, 220, 252, 193, 146, 220, 252, 193, 148, 220, 252, 193, 150, 220, 252, 193, 152, 220, 252, 193, 154, 220, 252, 193, 156, 220, 252, 193, 158, 220, 252, 193, 160, 220, 252, 193, 200, 220, 252, 193, 202, 220, 252, 193, 204, 220, 252, 193, 206, 220, 252, 193, 208, 220, 252, 193, 210, 220, 252, 193, 212, 220, 252, 193, 214, 220, 252, 193, 216, 220, 252, 193, 218, 220, 252, 193, 220, 220, 252, 193, 222, 220, 252, 193, 224, 220, 252, 193, 226, 220, 252, 193, 228, 220, 252, 193, 230, 220, 252, 193, 232, 220, 252, 193, 234, 220, 252, 193, 236, 220, 252, 193, 18, 221, 252, 193, 20, 221, 252, 193, 22, 221, 252, 193, 20, 101, 253, 193, 22, 101, 253, 193, 194, 101, 253, 193, 196, 101, 253, 193, 198, 101, 253, 193, 200, 101, 253, 193, 202, 101, 253, 193, 204, 101, 253, 193, 206, 101, 253, 193, 208, 101, 253, 193, 210, 101, 253, 193, 212, 101, 253, 193, 214, 101, 253, 193, 216, 101, 253, 193, 218, 101, 253, 193, 220, 101, 253, 193, 222, 101, 253, 193, 224, 101, 253, 193, 226, 101, 253, 193, 228, 101, 253, 193, 230, 101, 253, 193, 232, 101, 253, 193, 234, 101, 253, 193, 236, 101, 253, 193, 238, 101, 253, 193, 240, 101, 253, 193, 242, 101, 253, 193, 244, 101, 253, 193, 246, 101, 253, 193, 248, 101, 253, 193, 238, 220, 252, 193, 250, 101, 253, 193, 24, 221, 252, 193, 26, 221, 252, 193, 28, 221, 252, 193, 30, 221, 252, 193, 32, 221, 252, 193, 34, 221, 252, 193, 36, 221, 252, 193, 38, 221, 252, 193, 40, 221, 252, 193, 42, 221, 252, 193, 44, 221, 252, 193, 46, 221, 252, 193, 48, 221, 252, 193, 50, 221, 252, 193, 52, 221, 252, 193, 54, 221, 252, 193, 56, 221, 252, 193, 58, 221, 252, 193, 60, 221, 252, 193, 62, 221, 252, 193, 94, 221, 252, 193, 96, 221, 252, 193, 98, 221, 252, 193, 100, 221, 252, 193, 102, 221, 252, 193, 104, 221, 252, 193, 106, 221, 252, 193, 108, 221, 252, 193, 110, 221, 252, 193, 112, 221, 252, 193, 114, 221, 252, 193, 64, 221, 252, 193, 116, 221, 252, 193, 118, 221, 252, 193, 120, 221, 252, 193, 122, 221, 252, 193, 124, 221, 252, 193, 126, 221, 252, 193, 158, 221, 252, 193, 160, 221, 252, 193, 162, 221, 252, 193, 164, 221, 252, 193, 166, 221, 252, 193, 168, 221, 252, 193, 170, 221, 252, 193, 172, 221, 252, 193, 174, 221, 252, 193, 176, 221, 252, 193, 178, 221, 252, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 206, 96, 76, 0, 193, 146, 221, 252, 193, 148, 221, 252, 193, 150, 221, 252, 193, 152, 221, 252, 193, 154, 221, 252, 193, 156, 221, 252, 193, 192, 221, 252, 193, 194, 221, 252, 193, 196, 221, 252, 193, 30, 222, 252, 193, 148, 222, 252, 193, 156, 222, 252, 193, 218, 222, 252, 193, 220, 222, 252, 193, 236, 222, 252, 193, 238, 222, 252, 193, 252, 222, 252, 193, 10, 223, 252, 193, 38, 223, 252, 193, 40, 223, 252, 193, 42, 223, 252, 193, 122, 223, 252, 193, 124, 223, 252, 193, 126, 223, 252, 193, 128, 223, 252, 193, 130, 223, 252, 206, 160, 76, 0, 206, 160, 76, 0, 206, 160, 76, 0, 206, 160, 76, 0, 206, 160, 76, 0, 206, 160, 76, 0, 193, 190, 221, 252, 193, 248, 221, 252, 193, 250, 221, 252, 193, 252, 221, 252, 193, 254, 221, 252, 193, 2, 222, 252, 193, 22, 222, 252, 193, 24, 222, 252, 193, 26, 222, 252, 193, 28, 222, 252, 193, 44, 222, 252, 193, 62, 222, 252, 193, 86, 222, 252, 193, 92, 222, 252, 193, 106, 222, 252, 193, 108, 222, 252, 193, 110, 222, 252, 193, 130, 222, 252, 193, 140, 222, 252, 193, 152, 222, 252, 193, 154, 222, 252, 193, 186, 222, 252, 193, 242, 222, 252, 193, 254, 222, 252, 193, 6, 223, 252, 193, 8, 223, 252, 193, 12, 223, 252, 193, 22, 223, 252, 193, 24, 223, 252, 193, 26, 223, 252, 193, 28, 223, 252, 193, 30, 223, 252, 193, 42, 222, 252, 193, 76, 222, 252, 193, 78, 222, 252, 193, 80, 222, 252, 193, 232, 222, 252, 193, 66, 223, 252, 193, 68, 223, 252, 193, 168, 223, 252, 193, 170, 223, 252, 193, 136, 224, 252, 193, 138, 224, 252, 193, 140, 224, 252, 193, 142, 224, 252, 193, 144, 224, 252, 193, 192, 225, 252, 193, 194, 225, 252, 193, 196, 225, 252, 193, 198, 225, 252, 193, 200, 225, 252, 193, 202, 225, 252, 193, 204, 225, 252, 193, 206, 225, 252, 193, 86, 227, 252, 193, 88, 227, 252, 193, 90, 227, 252, 193, 92, 227, 252, 193, 94, 227, 252, 193, 96, 227, 252, 193, 98, 227, 252, 193, 100, 227, 252, 193, 102, 227, 252, 193, 100, 229, 252, 193, 70, 222, 252, 193, 72, 222, 252, 193, 74, 222, 252, 193, 88, 222, 252, 193, 90, 222, 252, 193, 94, 222, 252, 193, 96, 222, 252, 193, 98, 222, 252, 193, 102, 222, 252, 193, 100, 222, 252, 193, 104, 222, 252, 193, 112, 222, 252, 193, 116, 222, 252, 193, 118, 222, 252, 193, 120, 222, 252, 193, 122, 222, 252, 193, 128, 222, 252, 193, 132, 222, 252, 193, 136, 222, 252, 193, 150, 222, 252, 193, 158, 222, 252, 193, 160, 222, 252, 193, 162, 222, 252, 193, 164, 222, 252, 193, 166, 222, 252, 193, 168, 222, 252, 193, 174, 222, 252, 193, 178, 222, 252, 193, 180, 222, 252, 193, 182, 222, 252, 193, 188, 222, 252, 193, 190, 222, 252, 193, 192, 222, 252, 193, 194, 222, 252, 193, 196, 222, 252, 193, 198, 222, 252, 193, 204, 222, 252, 193, 206, 222, 252, 193, 208, 222, 252, 193, 210, 222, 252, 193, 212, 222, 252, 193, 214, 222, 252, 193, 222, 222, 252, 193, 224, 222, 252, 193, 226, 222, 252, 193, 228, 222, 252, 193, 230, 222, 252, 193, 244, 222, 252, 193, 246, 222, 252, 193, 248, 222, 252, 193, 2, 223, 252, 193, 16, 223, 252, 193, 18, 223, 252, 193, 20, 223, 252, 193, 44, 223, 252, 193, 46, 223, 252, 193, 48, 223, 252, 193, 50, 223, 252, 193, 52, 223, 252, 193, 54, 223, 252, 193, 56, 223, 252, 193, 58, 223, 252, 193, 60, 223, 252, 193, 62, 223, 252, 193, 32, 223, 252, 193, 34, 223, 252, 193, 36, 223, 252, 193, 80, 223, 252, 193, 82, 223, 252, 193, 84, 223, 252, 193, 86, 223, 252, 193, 88, 223, 252, 193, 90, 223, 252, 193, 92, 223, 252, 193, 94, 223, 252, 193, 96, 223, 252, 193, 98, 223, 252, 193, 100, 223, 252, 193, 102, 223, 252, 193, 104, 223, 252, 193, 106, 223, 252, 193, 108, 223, 252, 193, 110, 223, 252, 193, 112, 223, 252, 193, 114, 223, 252, 193, 116, 223, 252, 193, 118, 223, 252, 193, 120, 223, 252, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 193, 64, 223, 252, 193, 132, 223, 252, 193, 134, 223, 252, 193, 136, 223, 252, 193, 138, 223, 252, 193, 140, 223, 252, 193, 142, 223, 252, 193, 144, 223, 252, 193, 146, 223, 252, 193, 148, 223, 252, 193, 150, 223, 252, 193, 152, 223, 252, 193, 154, 223, 252, 193, 156, 223, 252, 193, 158, 223, 252, 193, 160, 223, 252, 193, 162, 223, 252, 193, 164, 223, 252, 193, 166, 223, 252, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 193, 70, 223, 252, 193, 172, 223, 252, 193, 174, 223, 252, 193, 146, 224, 252, 193, 148, 224, 252, 193, 150, 224, 252, 193, 212, 225, 252, 193, 214, 225, 252, 193, 106, 227, 252, 193, 108, 227, 252, 193, 110, 227, 252, 193, 112, 227, 252, 193, 114, 227, 252, 193, 116, 227, 252, 193, 118, 227, 252, 193, 140, 229, 252, 193, 142, 229, 252, 193, 144, 229, 252, 193, 146, 229, 252, 193, 148, 229, 252, 193, 150, 229, 252, 193, 152, 229, 252, 193, 154, 229, 252, 193, 156, 229, 252, 193, 158, 229, 252, 193, 160, 229, 252, 193, 162, 229, 252, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 193, 176, 223, 252, 193, 178, 223, 252, 193, 180, 223, 252, 193, 162, 224, 252, 193, 164, 224, 252, 193, 166, 224, 252, 193, 168, 224, 252, 193, 170, 224, 252, 193, 172, 224, 252, 193, 174, 224, 252, 193, 176, 224, 252, 193, 222, 225, 252, 193, 224, 225, 252, 193, 226, 225, 252, 193, 228, 225, 252, 193, 230, 225, 252, 193, 232, 225, 252, 193, 234, 225, 252, 193, 156, 227, 252, 193, 158, 227, 252, 193, 160, 227, 252, 193, 162, 227, 252, 193, 164, 227, 252, 193, 166, 227, 252, 193, 168, 227, 252, 193, 170, 227, 252, 193, 172, 227, 252, 193, 174, 227, 252, 193, 176, 227, 252, 193, 178, 227, 252, 193, 210, 229, 252, 193, 212, 229, 252, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 206, 128, 76, 0, 193, 178, 224, 252, 193, 6, 224, 252, 193, 8, 224, 252, 193, 10, 224, 252, 193, 12, 224, 252, 193, 14, 224, 252, 193, 16, 224, 252, 193, 18, 224, 252, 193, 20, 224, 252, 193, 22, 224, 252, 193, 24, 224, 252, 193, 26, 224, 252, 193, 28, 224, 252, 193, 30, 224, 252, 193, 32, 224, 252, 193, 34, 224, 252, 193, 36, 224, 252, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 193, 2, 226, 252, 193, 4, 226, 252, 193, 6, 226, 252, 193, 182, 227, 252, 193, 8, 226, 252, 193, 10, 226, 252, 193, 12, 226, 252, 193, 14, 226, 252, 193, 16, 226, 252, 193, 18, 226, 252, 193, 20, 226, 252, 193, 22, 226, 252, 193, 24, 226, 252, 193, 26, 226, 252, 193, 28, 226, 252, 193, 30, 226, 252, 193, 32, 226, 252, 193, 34, 226, 252, 193, 36, 226, 252, 193, 38, 226, 252, 193, 40, 226, 252, 193, 42, 226, 252, 193, 44, 226, 252, 193, 46, 226, 252, 193, 48, 226, 252, 193, 50, 226, 252, 193, 52, 226, 252, 193, 54, 226, 252, 193, 56, 226, 252, 193, 58, 226, 252, 193, 38, 224, 252, 193, 60, 226, 252, 206, 160, 76, 0, 206, 160, 76, 0, 206, 160, 76, 0, 206, 160, 76, 0, 206, 160, 76, 0, 206, 160, 76, 0, 193, 64, 225, 252, 193, 66, 225, 252, 193, 68, 225, 252, 193, 70, 225, 252, 193, 72, 225, 252, 193, 74, 225, 252, 193, 76, 225, 252, 193, 78, 225, 252, 193, 80, 225, 252, 193, 82, 225, 252, 193, 158, 226, 252, 193, 160, 226, 252, 193, 162, 226, 252, 193, 164, 226, 252, 193, 166, 226, 252, 193, 168, 226, 252, 193, 170, 226, 252, 193, 172, 226, 252, 193, 174, 226, 252, 193, 176, 226, 252, 193, 178, 226, 252, 193, 180, 226, 252, 193, 182, 226, 252, 193, 184, 226, 252, 193, 186, 226, 252, 193, 188, 226, 252, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 192, 76, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 193, 154, 224, 252, 193, 156, 224, 252, 193, 158, 224, 252, 193, 160, 224, 252, 193, 216, 225, 252, 193, 218, 225, 252, 193, 220, 225, 252, 193, 120, 227, 252, 193, 122, 227, 252, 193, 124, 227, 252, 193, 126, 227, 252, 193, 128, 227, 252, 193, 130, 227, 252, 193, 132, 227, 252, 193, 134, 227, 252, 193, 136, 227, 252, 193, 138, 227, 252, 193, 140, 227, 252, 193, 142, 227, 252, 193, 144, 227, 252, 193, 146, 227, 252, 193, 148, 227, 252, 193, 150, 227, 252, 193, 152, 227, 252, 193, 154, 227, 252, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 206, 224, 76, 0, 193, 236, 225, 252, 193, 58, 225, 252, 193, 60, 225, 252, 193, 62, 225, 252, 193, 238, 225, 252, 193, 240, 225, 252, 193, 242, 225, 252, 193, 244, 225, 252, 193, 246, 225, 252, 193, 248, 225, 252, 193, 250, 225, 252, 193, 180, 227, 252, 193, 252, 225, 252, 193, 254, 225, 252, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 0, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 193, 62, 226, 252, 193, 64, 226, 252, 193, 66, 226, 252, 193, 68, 226, 252, 193, 70, 226, 252, 193, 72, 226, 252, 193, 74, 226, 252, 193, 76, 226, 252, 193, 78, 226, 252, 193, 80, 226, 252, 193, 82, 226, 252, 193, 84, 226, 252, 193, 86, 226, 252, 193, 88, 226, 252, 193, 90, 226, 252, 193, 92, 226, 252, 193, 94, 226, 252, 193, 96, 226, 252, 193, 98, 226, 252, 193, 244, 229, 252, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 32, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 193, 190, 226, 252, 193, 192, 226, 252, 193, 86, 228, 252, 193, 88, 228, 252, 193, 90, 228, 252, 193, 92, 228, 252, 193, 94, 228, 252, 193, 96, 228, 252, 193, 98, 228, 252, 193, 100, 228, 252, 193, 102, 228, 252, 193, 104, 228, 252, 193, 106, 228, 252, 193, 108, 228, 252, 193, 110, 228, 252, 193, 112, 228, 252, 193, 114, 228, 252, 193, 116, 228, 252, 193, 118, 228, 252, 193, 120, 228, 252, 193, 122, 228, 252, 193, 210, 230, 252, 193, 124, 228, 252, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 64, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 206, 96, 77, 0, 193, 190, 232, 252, 193, 52, 228, 252, 193, 54, 228, 252, 193, 56, 228, 252, 193, 58, 228, 252, 193, 60, 228, 252, 193, 62, 228, 252, 193, 64, 228, 252, 193, 66, 228, 252, 193, 68, 228, 252, 193, 70, 228, 252, 193, 72, 228, 252, 193, 74, 228, 252, 193, 76, 228, 252, 193, 78, 228, 252, 193, 80, 228, 252, 193, 82, 228, 252, 206, 0, 78, 0, 206, 0, 78, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 206, 128, 77, 0, 193, 4, 231, 252, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 206, 160, 77, 0, 193, 6, 231, 252, 206, 192, 77, 0, 206, 192, 77, 0, 206, 192, 77, 0, 206, 192, 77, 0, 206, 192, 77, 0, 206, 192, 77, 0, 206, 192, 77, 0, 206, 192, 77, 0, 206, 192, 77, 0, 206, 192, 77, 0, 206, 192, 77, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 193, 102, 229, 252, 193, 104, 229, 252, 193, 106, 229, 252, 193, 108, 229, 252, 193, 110, 229, 252, 193, 112, 229, 252, 193, 114, 229, 252, 193, 116, 229, 252, 193, 118, 229, 252, 193, 120, 229, 252, 193, 122, 229, 252, 193, 124, 229, 252, 193, 126, 229, 252, 193, 128, 229, 252, 193, 130, 229, 252, 193, 132, 229, 252, 193, 56, 232, 252, 193, 58, 232, 252, 193, 60, 232, 252, 193, 62, 232, 252, 193, 64, 232, 252, 193, 66, 232, 252, 193, 68, 232, 252, 193, 70, 232, 252, 193, 72, 232, 252, 193, 74, 232, 252, 193, 76, 232, 252, 193, 78, 232, 252, 193, 80, 232, 252, 193, 82, 232, 252, 193, 84, 232, 252, 193, 86, 232, 252, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 206, 224, 77, 0, 193, 124, 232, 252, 193, 126, 232, 252, 193, 128, 232, 252, 193, 130, 232, 252, 193, 132, 232, 252, 193, 134, 232, 252, 193, 136, 232, 252, 193, 138, 232, 252, 193, 140, 232, 252, 193, 142, 232, 252, 193, 144, 232, 252, 193, 146, 232, 252, 193, 148, 232, 252, 193, 150, 232, 252, 193, 152, 232, 252, 193, 154, 232, 252, 193, 214, 229, 252, 193, 216, 229, 252, 193, 218, 229, 252, 193, 220, 229, 252, 193, 222, 229, 252, 193, 224, 229, 252, 193, 226, 229, 252, 193, 228, 229, 252, 193, 230, 229, 252, 193, 232, 229, 252, 193, 234, 229, 252, 193, 236, 229, 252, 193, 238, 229, 252, 193, 240, 229, 252, 193, 242, 229, 252, 193, 162, 232, 252, 193, 164, 232, 252, 193, 166, 232, 252, 193, 168, 232, 252, 193, 170, 232, 252, 193, 172, 232, 252, 193, 174, 232, 252, 193, 176, 232, 252, 193, 178, 232, 252, 193, 180, 232, 252, 193, 182, 232, 252, 193, 184, 232, 252, 193, 186, 232, 252, 193, 188, 232, 252, 193, 116, 235, 252, 193, 118, 235, 252, 193, 120, 235, 252, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 0, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 206, 128, 78, 0, 193, 206, 230, 252, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 206, 32, 78, 0, 193, 142, 233, 252, 193, 144, 233, 252, 193, 146, 233, 252, 193, 148, 233, 252, 193, 150, 233, 252, 193, 152, 233, 252, 193, 154, 233, 252, 193, 156, 233, 252, 193, 158, 233, 252, 193, 160, 233, 252, 193, 162, 233, 252, 193, 164, 233, 252, 193, 166, 233, 252, 193, 168, 233, 252, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 64, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 193, 88, 232, 252, 193, 232, 234, 252, 193, 234, 234, 252, 193, 236, 234, 252, 193, 238, 234, 252, 193, 240, 234, 252, 193, 242, 234, 252, 193, 244, 234, 252, 193, 246, 234, 252, 193, 248, 234, 252, 193, 250, 234, 252, 193, 252, 234, 252, 193, 254, 234, 252, 193, 2, 235, 252, 193, 4, 235, 252, 193, 6, 235, 252, 193, 8, 235, 252, 193, 10, 235, 252, 193, 12, 235, 252, 193, 14, 235, 252, 193, 16, 235, 252, 193, 18, 235, 252, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 206, 96, 78, 0, 193, 24, 235, 252, 193, 26, 235, 252, 193, 28, 235, 252, 193, 30, 235, 252, 193, 32, 235, 252, 193, 34, 235, 252, 193, 36, 235, 252, 193, 38, 235, 252, 193, 40, 235, 252, 193, 42, 235, 252, 193, 44, 235, 252, 193, 46, 235, 252, 193, 48, 235, 252, 193, 50, 235, 252, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 193, 156, 232, 252, 193, 158, 232, 252, 193, 160, 232, 252, 193, 52, 235, 252, 193, 54, 235, 252, 193, 56, 235, 252, 193, 58, 235, 252, 193, 60, 235, 252, 193, 62, 235, 252, 193, 64, 235, 252, 193, 66, 235, 252, 193, 68, 235, 252, 193, 70, 235, 252, 193, 72, 235, 252, 193, 74, 235, 252, 193, 76, 235, 252, 193, 78, 235, 252, 193, 80, 235, 252, 193, 82, 235, 252, 193, 84, 235, 252, 193, 86, 235, 252, 193, 88, 235, 252, 193, 90, 235, 252, 193, 92, 235, 252, 193, 94, 235, 252, 193, 96, 235, 252, 193, 98, 235, 252, 193, 100, 235, 252, 193, 102, 235, 252, 193, 104, 235, 252, 193, 106, 235, 252, 193, 108, 235, 252, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 160, 78, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 0, 79, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 192, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 206, 224, 78, 0, 193, 98, 236, 252, 193, 182, 234, 252, 193, 184, 234, 252, 193, 186, 234, 252, 193, 188, 234, 252, 193, 190, 234, 252, 193, 192, 234, 252, 193, 194, 234, 252, 193, 196, 234, 252, 193, 198, 234, 252, 193, 200, 234, 252, 193, 202, 234, 252, 193, 204, 234, 252, 193, 206, 234, 252, 193, 208, 234, 252, 193, 210, 234, 252, 193, 212, 234, 252, 193, 214, 234, 252, 193, 216, 234, 252, 193, 218, 234, 252, 193, 220, 234, 252, 193, 222, 234, 252, 193, 224, 234, 252, 193, 226, 234, 252, 193, 228, 234, 252, 193, 230, 234, 252, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 193, 110, 235, 252, 193, 112, 235, 252, 193, 114, 235, 252, 193, 254, 237, 252, 193, 2, 238, 252, 193, 4, 238, 252, 193, 6, 238, 252, 193, 8, 238, 252, 193, 10, 238, 252, 193, 12, 238, 252, 193, 14, 238, 252, 193, 16, 238, 252, 193, 18, 238, 252, 193, 20, 238, 252, 193, 22, 238, 252, 193, 24, 238, 252, 193, 26, 238, 252, 193, 28, 238, 252, 193, 30, 238, 252, 193, 32, 238, 252, 193, 34, 238, 252, 193, 36, 238, 252, 193, 38, 238, 252, 193, 40, 238, 252, 193, 42, 238, 252, 193, 44, 238, 252, 193, 46, 238, 252, 193, 48, 238, 252, 193, 50, 238, 252, 193, 52, 238, 252, 193, 54, 238, 252, 193, 56, 238, 252, 193, 122, 235, 252, 193, 124, 235, 252, 193, 126, 235, 252, 193, 128, 235, 252, 193, 130, 235, 252, 193, 132, 235, 252, 193, 134, 235, 252, 193, 136, 235, 252, 193, 138, 235, 252, 193, 140, 235, 252, 193, 142, 235, 252, 193, 60, 238, 252, 193, 62, 238, 252, 193, 64, 238, 252, 193, 66, 238, 252, 193, 68, 238, 252, 193, 70, 238, 252, 193, 72, 238, 252, 193, 210, 240, 252, 193, 212, 240, 252, 193, 214, 240, 252, 193, 216, 240, 252, 193, 52, 243, 252, 193, 54, 243, 252, 193, 56, 243, 252, 193, 58, 243, 252, 193, 60, 243, 252, 193, 62, 243, 252, 193, 64, 243, 252, 193, 98, 245, 252, 193, 100, 245, 252, 193, 102, 245, 252, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 206, 32, 79, 0, 193, 74, 238, 252, 193, 58, 236, 252, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 64, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 206, 96, 79, 0, 193, 56, 239, 252, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 206, 128, 79, 0, 193, 68, 240, 252, 193, 70, 240, 252, 193, 72, 240, 252, 193, 74, 240, 252, 193, 76, 240, 252, 193, 78, 240, 252, 193, 80, 240, 252, 193, 82, 240, 252, 193, 84, 240, 252, 193, 86, 240, 252, 193, 88, 240, 252, 193, 90, 240, 252, 193, 92, 240, 252, 193, 94, 240, 252, 193, 96, 240, 252, 193, 98, 240, 252, 193, 100, 240, 252, 193, 102, 240, 252, 193, 104, 240, 252, 193, 106, 240, 252, 193, 108, 240, 252, 193, 198, 242, 252, 193, 200, 242, 252, 193, 202, 242, 252, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 160, 79, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 193, 58, 238, 252, 193, 162, 240, 252, 193, 164, 240, 252, 193, 166, 240, 252, 193, 168, 240, 252, 193, 170, 240, 252, 193, 172, 240, 252, 193, 174, 240, 252, 193, 176, 240, 252, 193, 178, 240, 252, 193, 180, 240, 252, 193, 182, 240, 252, 193, 184, 240, 252, 193, 186, 240, 252, 193, 188, 240, 252, 193, 190, 240, 252, 193, 192, 240, 252, 193, 194, 240, 252, 193, 196, 240, 252, 193, 198, 240, 252, 193, 200, 240, 252, 193, 202, 240, 252, 193, 204, 240, 252, 193, 206, 240, 252, 193, 208, 240, 252, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 206, 192, 79, 0, 193, 218, 240, 252, 193, 220, 240, 252, 193, 222, 240, 252, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 224, 79, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 193, 58, 239, 252, 193, 60, 239, 252, 193, 62, 239, 252, 193, 64, 239, 252, 193, 66, 239, 252, 193, 68, 239, 252, 193, 70, 239, 252, 193, 72, 239, 252, 193, 74, 239, 252, 193, 76, 239, 252, 193, 78, 239, 252, 193, 80, 239, 252, 193, 82, 239, 252, 193, 84, 239, 252, 193, 86, 239, 252, 193, 88, 239, 252, 193, 90, 239, 252, 193, 92, 239, 252, 193, 94, 239, 252, 193, 96, 239, 252, 193, 98, 239, 252, 193, 100, 239, 252, 193, 102, 239, 252, 193, 104, 239, 252, 193, 106, 239, 252, 193, 108, 239, 252, 193, 110, 239, 252, 193, 112, 239, 252, 193, 114, 239, 252, 193, 116, 239, 252, 193, 118, 239, 252, 193, 120, 239, 252, 193, 122, 239, 252, 193, 124, 239, 252, 193, 126, 239, 252, 193, 128, 239, 252, 193, 130, 239, 252, 193, 132, 239, 252, 193, 134, 239, 252, 193, 136, 239, 252, 193, 138, 239, 252, 193, 140, 239, 252, 193, 142, 239, 252, 193, 144, 239, 252, 193, 146, 239, 252, 193, 148, 239, 252, 193, 150, 239, 252, 193, 152, 239, 252, 193, 154, 239, 252, 193, 156, 239, 252, 193, 158, 239, 252, 193, 160, 239, 252, 193, 162, 239, 252, 193, 164, 239, 252, 193, 166, 239, 252, 193, 168, 239, 252, 193, 170, 239, 252, 193, 172, 239, 252, 193, 174, 239, 252, 193, 176, 239, 252, 193, 178, 239, 252, 193, 126, 241, 252, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 206, 0, 80, 0, 193, 158, 252, 252, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 206, 32, 80, 0, 193, 216, 242, 252, 193, 218, 242, 252, 193, 220, 242, 252, 193, 222, 242, 252, 193, 224, 242, 252, 193, 226, 242, 252, 193, 228, 242, 252, 193, 230, 242, 252, 193, 232, 242, 252, 193, 234, 242, 252, 193, 236, 242, 252, 193, 238, 242, 252, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 193, 224, 240, 252, 193, 226, 240, 252, 193, 228, 240, 252, 193, 230, 240, 252, 193, 232, 240, 252, 193, 234, 240, 252, 193, 236, 240, 252, 193, 238, 240, 252, 193, 240, 240, 252, 193, 242, 240, 252, 193, 244, 240, 252, 193, 246, 240, 252, 193, 248, 240, 252, 193, 250, 240, 252, 193, 252, 240, 252, 193, 254, 240, 252, 193, 2, 241, 252, 193, 4, 241, 252, 193, 6, 241, 252, 193, 8, 241, 252, 193, 10, 241, 252, 193, 12, 241, 252, 193, 14, 241, 252, 193, 16, 241, 252, 193, 18, 241, 252, 193, 20, 241, 252, 193, 22, 241, 252, 193, 24, 241, 252, 193, 26, 241, 252, 193, 28, 241, 252, 193, 30, 241, 252, 193, 32, 241, 252, 193, 34, 241, 252, 193, 36, 241, 252, 193, 38, 241, 252, 193, 40, 241, 252, 193, 42, 241, 252, 193, 44, 241, 252, 193, 46, 241, 252, 193, 48, 241, 252, 193, 50, 241, 252, 193, 52, 241, 252, 193, 54, 241, 252, 193, 56, 241, 252, 193, 58, 241, 252, 193, 60, 241, 252, 193, 62, 241, 252, 193, 64, 241, 252, 193, 66, 241, 252, 193, 68, 241, 252, 193, 70, 241, 252, 193, 72, 241, 252, 193, 74, 241, 252, 193, 76, 241, 252, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 206, 64, 80, 0, 193, 180, 243, 252, 193, 182, 243, 252, 193, 184, 243, 252, 193, 186, 243, 252, 193, 188, 243, 252, 193, 190, 243, 252, 193, 192, 243, 252, 193, 194, 243, 252, 193, 196, 243, 252, 193, 198, 243, 252, 193, 200, 243, 252, 193, 202, 243, 252, 193, 204, 243, 252, 193, 200, 245, 252, 193, 202, 245, 252, 193, 204, 245, 252, 193, 206, 245, 252, 193, 208, 245, 252, 193, 210, 245, 252, 193, 212, 245, 252, 193, 214, 245, 252, 193, 216, 245, 252, 193, 218, 245, 252, 193, 220, 245, 252, 193, 222, 245, 252, 193, 134, 247, 252, 193, 136, 247, 252, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 96, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 206, 128, 80, 0, 193, 206, 243, 252, 193, 112, 242, 252, 193, 114, 242, 252, 193, 116, 242, 252, 193, 118, 242, 252, 193, 120, 242, 252, 193, 122, 242, 252, 193, 124, 242, 252, 193, 126, 242, 252, 193, 128, 242, 252, 193, 130, 242, 252, 193, 132, 242, 252, 193, 134, 242, 252, 193, 136, 242, 252, 193, 138, 242, 252, 193, 140, 242, 252, 193, 142, 242, 252, 193, 144, 242, 252, 193, 146, 242, 252, 193, 148, 242, 252, 193, 150, 242, 252, 193, 152, 242, 252, 193, 154, 242, 252, 193, 156, 242, 252, 193, 158, 242, 252, 193, 160, 242, 252, 193, 162, 242, 252, 193, 164, 242, 252, 193, 166, 242, 252, 193, 168, 242, 252, 193, 170, 242, 252, 193, 172, 242, 252, 193, 174, 242, 252, 193, 176, 242, 252, 193, 178, 242, 252, 193, 180, 242, 252, 193, 182, 242, 252, 193, 184, 242, 252, 193, 186, 242, 252, 193, 188, 242, 252, 193, 190, 242, 252, 193, 192, 242, 252, 193, 194, 242, 252, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 193, 196, 242, 252, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 193, 204, 242, 252, 193, 206, 242, 252, 193, 208, 242, 252, 193, 210, 242, 252, 193, 212, 242, 252, 193, 214, 244, 252, 193, 216, 244, 252, 193, 218, 244, 252, 193, 220, 244, 252, 193, 222, 244, 252, 193, 224, 244, 252, 193, 226, 244, 252, 193, 228, 244, 252, 193, 230, 244, 252, 193, 232, 244, 252, 193, 234, 244, 252, 193, 236, 244, 252, 193, 238, 244, 252, 193, 240, 244, 252, 193, 242, 244, 252, 193, 244, 244, 252, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 160, 80, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 206, 192, 80, 0, 193, 120, 245, 252, 193, 168, 243, 252, 193, 170, 243, 252, 193, 172, 243, 252, 193, 122, 245, 252, 193, 174, 243, 252, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 224, 80, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 0, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 193, 72, 246, 252, 193, 210, 244, 252, 193, 74, 246, 252, 193, 76, 246, 252, 193, 78, 246, 252, 193, 80, 246, 252, 193, 82, 246, 252, 193, 84, 246, 252, 193, 86, 246, 252, 193, 88, 246, 252, 193, 90, 246, 252, 193, 92, 246, 252, 193, 94, 246, 252, 193, 96, 246, 252, 193, 98, 246, 252, 193, 100, 246, 252, 193, 102, 246, 252, 193, 104, 246, 252, 193, 106, 246, 252, 193, 108, 246, 252, 193, 110, 246, 252, 193, 112, 246, 252, 193, 114, 246, 252, 193, 116, 246, 252, 193, 212, 244, 252, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 206, 32, 81, 0, 193, 228, 246, 252, 193, 230, 246, 252, 193, 232, 246, 252, 193, 234, 246, 252, 193, 236, 246, 252, 193, 238, 246, 252, 193, 240, 246, 252, 193, 242, 246, 252, 193, 244, 246, 252, 193, 246, 246, 252, 193, 248, 246, 252, 193, 250, 246, 252, 193, 252, 246, 252, 193, 254, 246, 252, 193, 2, 247, 252, 193, 4, 247, 252, 193, 6, 247, 252, 193, 8, 247, 252, 193, 100, 248, 252, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 64, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 193, 104, 245, 252, 193, 106, 245, 252, 193, 108, 245, 252, 193, 110, 245, 252, 193, 112, 245, 252, 193, 114, 245, 252, 193, 116, 245, 252, 193, 118, 245, 252, 193, 64, 247, 252, 193, 66, 247, 252, 193, 166, 248, 252, 193, 168, 248, 252, 193, 170, 248, 252, 193, 172, 248, 252, 193, 174, 248, 252, 193, 176, 248, 252, 193, 178, 248, 252, 193, 234, 249, 252, 193, 236, 249, 252, 193, 238, 249, 252, 193, 220, 250, 252, 193, 222, 250, 252, 193, 108, 251, 252, 193, 110, 252, 252, 193, 250, 252, 252, 193, 252, 252, 252, 193, 254, 252, 252, 193, 48, 253, 252, 193, 50, 253, 252, 193, 52, 253, 252, 193, 94, 253, 252, 193, 96, 253, 252, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 96, 81, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 193, 10, 250, 252, 193, 12, 250, 252, 193, 242, 248, 252, 193, 14, 250, 252, 193, 198, 245, 252, 193, 16, 250, 252, 193, 18, 250, 252, 193, 20, 250, 252, 193, 22, 250, 252, 193, 24, 250, 252, 193, 26, 250, 252, 193, 28, 250, 252, 193, 30, 250, 252, 193, 32, 250, 252, 193, 34, 250, 252, 193, 36, 250, 252, 193, 38, 250, 252, 193, 110, 251, 252, 193, 224, 250, 252, 193, 226, 250, 252, 193, 228, 250, 252, 193, 230, 250, 252, 193, 232, 250, 252, 193, 234, 250, 252, 193, 112, 251, 252, 193, 114, 251, 252, 193, 116, 251, 252, 193, 118, 251, 252, 193, 120, 251, 252, 193, 122, 251, 252, 193, 124, 251, 252, 193, 126, 251, 252, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 206, 128, 81, 0, 193, 150, 247, 252, 193, 152, 247, 252, 193, 54, 246, 252, 193, 56, 246, 252, 193, 58, 246, 252, 193, 60, 246, 252, 193, 62, 246, 252, 193, 64, 246, 252, 193, 66, 246, 252, 193, 154, 247, 252, 193, 68, 246, 252, 193, 70, 246, 252, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 160, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 206, 192, 81, 0, 193, 86, 248, 252, 193, 88, 248, 252, 193, 90, 248, 252, 193, 92, 248, 252, 193, 94, 248, 252, 193, 96, 248, 252, 193, 98, 248, 252, 193, 180, 249, 252, 193, 182, 249, 252, 193, 184, 249, 252, 193, 186, 249, 252, 193, 176, 250, 252, 193, 178, 250, 252, 193, 180, 250, 252, 193, 182, 250, 252, 193, 74, 251, 252, 193, 76, 251, 252, 193, 78, 251, 252, 193, 202, 251, 252, 193, 28, 252, 252, 193, 30, 252, 252, 193, 32, 252, 252, 193, 104, 252, 252, 193, 106, 252, 252, 193, 204, 252, 252, 193, 206, 252, 252, 193, 236, 252, 252, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 224, 81, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 206, 0, 82, 0, 193, 180, 248, 252, 193, 120, 247, 252, 193, 122, 247, 252, 193, 124, 247, 252, 193, 126, 247, 252, 193, 128, 247, 252, 193, 130, 247, 252, 193, 132, 247, 252, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 193, 138, 247, 252, 193, 140, 247, 252, 193, 142, 247, 252, 193, 144, 247, 252, 193, 146, 247, 252, 193, 148, 247, 252, 193, 244, 248, 252, 193, 246, 248, 252, 193, 248, 248, 252, 193, 250, 248, 252, 193, 252, 248, 252, 193, 254, 248, 252, 193, 2, 249, 252, 193, 4, 249, 252, 193, 6, 249, 252, 193, 8, 249, 252, 193, 10, 249, 252, 193, 12, 249, 252, 193, 14, 249, 252, 193, 40, 250, 252, 193, 42, 250, 252, 193, 44, 250, 252, 193, 46, 250, 252, 193, 48, 250, 252, 193, 240, 250, 252, 193, 242, 250, 252, 193, 244, 250, 252, 193, 132, 251, 252, 193, 134, 251, 252, 193, 136, 251, 252, 193, 232, 251, 252, 193, 46, 252, 252, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 32, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 193, 84, 248, 252, 193, 50, 250, 252, 193, 178, 249, 252, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 193, 102, 248, 252, 193, 104, 248, 252, 193, 106, 248, 252, 193, 108, 248, 252, 193, 110, 248, 252, 193, 112, 248, 252, 193, 114, 248, 252, 193, 116, 248, 252, 193, 118, 248, 252, 193, 120, 248, 252, 193, 122, 248, 252, 193, 124, 248, 252, 193, 126, 248, 252, 193, 128, 248, 252, 193, 188, 249, 252, 193, 190, 249, 252, 193, 192, 249, 252, 193, 194, 249, 252, 193, 196, 249, 252, 193, 198, 249, 252, 193, 200, 249, 252, 193, 202, 249, 252, 193, 184, 250, 252, 193, 186, 250, 252, 193, 188, 250, 252, 193, 190, 250, 252, 193, 192, 250, 252, 193, 194, 250, 252, 193, 80, 251, 252, 193, 82, 251, 252, 193, 84, 251, 252, 193, 86, 251, 252, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 206, 64, 82, 0, 193, 204, 249, 252, 193, 206, 249, 252, 193, 208, 249, 252, 193, 210, 249, 252, 193, 212, 249, 252, 193, 214, 249, 252, 193, 216, 249, 252, 193, 218, 249, 252, 193, 220, 249, 252, 193, 222, 249, 252, 193, 224, 249, 252, 193, 226, 249, 252, 193, 228, 249, 252, 193, 230, 249, 252, 193, 232, 249, 252, 206, 192, 82, 0, 206, 192, 82, 0, 206, 192, 82, 0, 206, 192, 82, 0, 206, 192, 82, 0, 206, 192, 82, 0, 206, 192, 82, 0, 206, 192, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 206, 96, 82, 0, 193, 240, 249, 252, 193, 238, 248, 252, 193, 240, 248, 252, 193, 242, 249, 252, 193, 244, 249, 252, 193, 246, 249, 252, 193, 248, 249, 252, 193, 250, 249, 252, 193, 252, 249, 252, 193, 254, 249, 252, 193, 2, 250, 252, 193, 4, 250, 252, 193, 6, 250, 252, 193, 8, 250, 252, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 128, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 160, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 192, 82, 0, 206, 192, 82, 0, 206, 192, 82, 0, 206, 192, 82, 0, 193, 88, 251, 252, 193, 90, 251, 252, 193, 92, 251, 252, 193, 94, 251, 252, 193, 96, 251, 252, 193, 98, 251, 252, 193, 100, 251, 252, 193, 102, 251, 252, 193, 104, 251, 252, 193, 106, 251, 252, 193, 214, 251, 252, 193, 216, 251, 252, 193, 218, 251, 252, 193, 220, 251, 252, 193, 222, 251, 252, 193, 36, 252, 252, 193, 74, 252, 252, 193, 76, 252, 252, 193, 108, 252, 252, 193, 176, 252, 252, 193, 186, 252, 252, 193, 188, 252, 252, 193, 190, 252, 252, 193, 210, 252, 252, 193, 212, 252, 252, 193, 242, 252, 252, 193, 244, 252, 252, 193, 246, 252, 252, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 224, 82, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 0, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 193, 204, 251, 252, 193, 206, 251, 252, 193, 208, 251, 252, 193, 210, 251, 252, 193, 212, 251, 252, 193, 34, 252, 252, 193, 72, 252, 252, 193, 208, 252, 252, 193, 238, 252, 252, 193, 240, 252, 252, 193, 86, 253, 252, 193, 126, 253, 252, 193, 128, 253, 252, 193, 130, 253, 252, 193, 18, 254, 252, 193, 20, 254, 252, 193, 22, 254, 252, 193, 84, 254, 252, 193, 104, 254, 252, 193, 144, 254, 252, 193, 30, 255, 252, 193, 32, 255, 252, 193, 210, 255, 252, 193, 212, 255, 252, 193, 214, 255, 252, 193, 216, 255, 252, 193, 218, 255, 252, 193, 220, 255, 252, 193, 198, 2, 253, 193, 200, 2, 253, 193, 202, 2, 253, 193, 204, 2, 253, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 206, 32, 83, 0, 193, 50, 252, 252, 193, 52, 252, 252, 193, 54, 252, 252, 193, 56, 252, 252, 193, 58, 252, 252, 193, 60, 252, 252, 193, 62, 252, 252, 193, 140, 252, 252, 193, 64, 252, 252, 193, 66, 252, 252, 193, 68, 252, 252, 193, 70, 252, 252, 206, 64, 83, 0, 206, 64, 83, 0, 206, 64, 83, 0, 206, 64, 83, 0, 206, 64, 83, 0, 206, 64, 83, 0, 193, 48, 252, 252, 193, 78, 252, 252, 193, 114, 252, 252, 193, 116, 252, 252, 193, 118, 252, 252, 193, 136, 252, 252, 193, 138, 252, 252, 193, 166, 252, 252, 193, 182, 252, 252, 193, 196, 252, 252, 193, 220, 252, 252, 193, 222, 252, 252, 193, 224, 252, 252, 193, 226, 252, 252, 193, 14, 253, 252, 193, 16, 253, 252, 193, 60, 253, 252, 193, 62, 253, 252, 193, 102, 253, 252, 193, 104, 253, 252, 193, 106, 253, 252, 193, 108, 253, 252, 193, 110, 253, 252, 193, 188, 253, 252, 193, 228, 253, 252, 193, 32, 254, 252, 193, 62, 254, 252, 193, 64, 254, 252, 193, 88, 254, 252, 193, 156, 254, 252, 193, 234, 254, 252, 193, 236, 254, 252, 193, 230, 252, 252, 193, 232, 252, 252, 193, 234, 252, 252, 193, 18, 253, 252, 193, 20, 253, 252, 193, 22, 253, 252, 193, 24, 253, 252, 193, 26, 253, 252, 193, 28, 253, 252, 193, 30, 253, 252, 193, 32, 253, 252, 193, 34, 253, 252, 193, 36, 253, 252, 193, 38, 253, 252, 193, 40, 253, 252, 193, 42, 253, 252, 193, 44, 253, 252, 193, 64, 253, 252, 193, 66, 253, 252, 193, 68, 253, 252, 193, 70, 253, 252, 193, 72, 253, 252, 193, 74, 253, 252, 193, 76, 253, 252, 193, 78, 253, 252, 193, 80, 253, 252, 193, 112, 253, 252, 193, 114, 253, 252, 193, 116, 253, 252, 193, 118, 253, 252, 193, 120, 253, 252, 193, 122, 253, 252, 193, 248, 252, 252, 193, 46, 253, 252, 193, 88, 253, 252, 193, 90, 253, 252, 193, 92, 253, 252, 193, 132, 253, 252, 193, 134, 253, 252, 193, 216, 253, 252, 193, 218, 253, 252, 193, 24, 254, 252, 193, 26, 254, 252, 193, 56, 254, 252, 193, 58, 254, 252, 193, 86, 254, 252, 193, 106, 254, 252, 193, 118, 254, 252, 193, 152, 254, 252, 193, 162, 254, 252, 193, 196, 254, 252, 193, 34, 255, 252, 193, 36, 255, 252, 193, 38, 255, 252, 193, 40, 255, 252, 193, 42, 255, 252, 193, 222, 255, 252, 193, 224, 255, 252, 193, 226, 255, 252, 193, 228, 255, 252, 193, 230, 255, 252, 193, 210, 2, 253, 193, 212, 2, 253, 193, 214, 2, 253, 193, 82, 253, 252, 193, 84, 253, 252, 193, 176, 253, 252, 193, 54, 254, 252, 193, 194, 254, 252, 193, 26, 255, 252, 193, 28, 255, 252, 193, 200, 255, 252, 193, 202, 255, 252, 193, 204, 255, 252, 193, 206, 255, 252, 193, 208, 255, 252, 193, 184, 2, 253, 193, 186, 2, 253, 193, 188, 2, 253, 193, 190, 2, 253, 193, 192, 2, 253, 193, 194, 2, 253, 193, 196, 2, 253, 193, 176, 3, 253, 193, 178, 3, 253, 193, 180, 3, 253, 193, 182, 3, 253, 193, 184, 3, 253, 193, 186, 3, 253, 193, 188, 3, 253, 193, 190, 3, 253, 193, 192, 3, 253, 193, 194, 3, 253, 193, 190, 4, 253, 193, 192, 4, 253, 193, 194, 4, 253, 193, 124, 253, 252, 193, 148, 253, 252, 193, 150, 253, 252, 193, 152, 253, 252, 193, 154, 253, 252, 193, 156, 253, 252, 193, 158, 253, 252, 193, 160, 253, 252, 193, 162, 253, 252, 193, 164, 253, 252, 193, 166, 253, 252, 193, 168, 253, 252, 193, 170, 253, 252, 193, 172, 253, 252, 193, 174, 253, 252, 193, 190, 253, 252, 193, 192, 253, 252, 193, 194, 253, 252, 193, 196, 253, 252, 193, 198, 253, 252, 193, 200, 253, 252, 193, 202, 253, 252, 193, 204, 253, 252, 193, 206, 253, 252, 193, 208, 253, 252, 193, 210, 253, 252, 193, 212, 253, 252, 193, 214, 253, 252, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 193, 136, 253, 252, 193, 138, 253, 252, 193, 140, 253, 252, 193, 142, 253, 252, 193, 178, 253, 252, 193, 180, 253, 252, 193, 182, 253, 252, 193, 220, 253, 252, 193, 134, 254, 252, 193, 198, 254, 252, 193, 44, 255, 252, 193, 232, 255, 252, 193, 234, 255, 252, 193, 236, 255, 252, 193, 238, 255, 252, 193, 240, 255, 252, 193, 242, 255, 252, 193, 244, 255, 252, 193, 246, 255, 252, 193, 218, 2, 253, 193, 220, 2, 253, 193, 222, 2, 253, 193, 224, 2, 253, 193, 226, 2, 253, 193, 224, 3, 253, 193, 226, 3, 253, 193, 240, 4, 253, 193, 242, 4, 253, 193, 244, 4, 253, 193, 246, 4, 253, 193, 248, 4, 253, 193, 80, 6, 253, 193, 144, 253, 252, 193, 146, 253, 252, 193, 184, 253, 252, 193, 186, 253, 252, 193, 222, 253, 252, 193, 224, 253, 252, 193, 226, 253, 252, 193, 28, 254, 252, 193, 30, 254, 252, 193, 60, 254, 252, 193, 146, 254, 252, 193, 150, 254, 252, 193, 154, 254, 252, 193, 164, 254, 252, 193, 166, 254, 252, 193, 168, 254, 252, 193, 170, 254, 252, 193, 172, 254, 252, 193, 174, 254, 252, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 206, 96, 83, 0, 193, 34, 254, 252, 193, 36, 254, 252, 193, 38, 254, 252, 193, 40, 254, 252, 193, 42, 254, 252, 193, 44, 254, 252, 193, 46, 254, 252, 193, 48, 254, 252, 193, 50, 254, 252, 193, 52, 254, 252, 193, 66, 254, 252, 193, 68, 254, 252, 193, 70, 254, 252, 193, 72, 254, 252, 193, 74, 254, 252, 193, 76, 254, 252, 193, 78, 254, 252, 193, 90, 254, 252, 193, 80, 254, 252, 193, 82, 254, 252, 193, 92, 254, 252, 193, 94, 254, 252, 193, 96, 254, 252, 193, 98, 254, 252, 193, 100, 254, 252, 193, 102, 254, 252, 193, 108, 254, 252, 193, 110, 254, 252, 193, 112, 254, 252, 193, 114, 254, 252, 193, 116, 254, 252, 193, 110, 161, 253, 193, 120, 254, 252, 193, 122, 254, 252, 193, 124, 254, 252, 193, 126, 254, 252, 193, 128, 254, 252, 193, 130, 254, 252, 193, 132, 254, 252, 193, 136, 254, 252, 193, 138, 254, 252, 193, 140, 254, 252, 193, 142, 254, 252, 193, 148, 254, 252, 193, 158, 254, 252, 193, 160, 254, 252, 193, 176, 254, 252, 193, 178, 254, 252, 193, 180, 254, 252, 193, 182, 254, 252, 193, 184, 254, 252, 193, 186, 254, 252, 193, 188, 254, 252, 193, 190, 254, 252, 193, 192, 254, 252, 193, 242, 254, 252, 193, 244, 254, 252, 193, 246, 254, 252, 193, 248, 254, 252, 193, 250, 254, 252, 193, 252, 254, 252, 193, 254, 254, 252, 193, 2, 255, 252, 193, 4, 255, 252, 193, 6, 255, 252, 193, 8, 255, 252, 193, 10, 255, 252, 193, 12, 255, 252, 193, 14, 255, 252, 193, 16, 255, 252, 193, 18, 255, 252, 193, 20, 255, 252, 193, 22, 255, 252, 193, 24, 255, 252, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 206, 128, 83, 0, 193, 46, 255, 252, 193, 48, 255, 252, 193, 50, 255, 252, 193, 52, 255, 252, 193, 228, 2, 253, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 193, 238, 254, 252, 193, 124, 255, 252, 193, 126, 255, 252, 193, 128, 255, 252, 193, 68, 2, 253, 193, 70, 2, 253, 193, 72, 2, 253, 193, 74, 2, 253, 193, 76, 2, 253, 193, 78, 2, 253, 193, 80, 2, 253, 193, 82, 2, 253, 193, 84, 2, 253, 193, 48, 3, 253, 193, 50, 3, 253, 193, 52, 3, 253, 193, 54, 3, 253, 193, 56, 3, 253, 193, 58, 3, 253, 193, 60, 3, 253, 193, 62, 3, 253, 193, 64, 3, 253, 193, 50, 4, 253, 193, 52, 4, 253, 193, 54, 4, 253, 193, 56, 4, 253, 193, 58, 4, 253, 193, 60, 4, 253, 193, 62, 4, 253, 193, 64, 4, 253, 193, 66, 4, 253, 193, 96, 5, 253, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 206, 160, 83, 0, 193, 248, 255, 252, 193, 122, 255, 252, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 206, 192, 83, 0, 193, 86, 2, 253, 193, 88, 2, 253, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 206, 224, 83, 0, 193, 230, 2, 253, 193, 232, 2, 253, 193, 234, 2, 253, 193, 236, 2, 253, 193, 238, 2, 253, 193, 240, 2, 253, 193, 228, 3, 253, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 193, 240, 249, 252, 193, 66, 2, 253, 193, 156, 35, 253, 193, 178, 55, 253, 193, 86, 79, 253, 193, 170, 81, 253, 193, 102, 90, 253, 193, 218, 102, 253, 193, 88, 104, 253, 193, 30, 129, 253, 193, 14, 158, 253, 193, 98, 158, 253, 193, 188, 159, 253, 193, 250, 201, 253, 193, 66, 89, 253, 193, 204, 131, 253, 193, 126, 14, 251, 193, 160, 21, 251, 193, 68, 26, 251, 193, 4, 43, 251, 193, 12, 44, 251, 193, 46, 48, 251, 193, 92, 67, 251, 193, 34, 72, 251, 193, 50, 76, 251, 193, 94, 92, 251, 193, 16, 215, 253, 193, 76, 134, 251, 193, 8, 135, 251, 193, 176, 174, 251, 193, 108, 178, 251, 193, 254, 182, 251, 193, 90, 2, 253, 193, 92, 2, 253, 193, 94, 2, 253, 193, 96, 2, 253, 193, 98, 2, 253, 193, 100, 2, 253, 193, 102, 2, 253, 193, 104, 2, 253, 193, 106, 2, 253, 193, 108, 2, 253, 193, 110, 2, 253, 193, 112, 2, 253, 193, 114, 2, 253, 193, 116, 2, 253, 193, 118, 2, 253, 193, 120, 2, 253, 193, 122, 2, 253, 193, 124, 2, 253, 193, 126, 2, 253, 193, 128, 2, 253, 193, 130, 2, 253, 193, 132, 2, 253, 193, 134, 2, 253, 193, 136, 2, 253, 193, 138, 2, 253, 193, 140, 2, 253, 193, 142, 2, 253, 193, 144, 2, 253, 193, 146, 2, 253, 193, 148, 2, 253, 193, 150, 2, 253, 193, 152, 2, 253, 193, 154, 2, 253, 193, 156, 2, 253, 193, 158, 2, 253, 193, 160, 2, 253, 193, 162, 2, 253, 193, 164, 2, 253, 193, 166, 2, 253, 193, 168, 2, 253, 193, 170, 2, 253, 193, 172, 2, 253, 193, 174, 2, 253, 193, 176, 2, 253, 193, 178, 2, 253, 193, 180, 2, 253, 193, 182, 2, 253, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 193, 206, 2, 253, 193, 208, 2, 253, 193, 196, 3, 253, 193, 198, 3, 253, 193, 200, 3, 253, 193, 202, 3, 253, 193, 204, 3, 253, 193, 206, 3, 253, 193, 200, 4, 253, 193, 202, 4, 253, 193, 204, 4, 253, 193, 206, 4, 253, 193, 208, 4, 253, 193, 210, 4, 253, 193, 40, 6, 253, 193, 42, 6, 253, 193, 44, 6, 253, 193, 46, 6, 253, 193, 48, 6, 253, 193, 50, 6, 253, 193, 52, 6, 253, 193, 54, 6, 253, 193, 56, 6, 253, 193, 58, 6, 253, 193, 60, 6, 253, 193, 62, 6, 253, 193, 64, 6, 253, 193, 148, 7, 253, 193, 150, 7, 253, 193, 152, 7, 253, 193, 154, 7, 253, 193, 218, 8, 253, 193, 216, 2, 253, 193, 208, 3, 253, 193, 210, 3, 253, 193, 212, 3, 253, 193, 214, 3, 253, 193, 216, 3, 253, 193, 218, 3, 253, 193, 220, 3, 253, 193, 222, 3, 253, 193, 212, 4, 253, 193, 214, 4, 253, 193, 216, 4, 253, 193, 218, 4, 253, 193, 220, 4, 253, 193, 222, 4, 253, 193, 224, 4, 253, 193, 226, 4, 253, 193, 228, 4, 253, 193, 230, 4, 253, 193, 232, 4, 253, 193, 234, 4, 253, 193, 236, 4, 253, 193, 238, 4, 253, 193, 66, 6, 253, 193, 68, 6, 253, 193, 70, 6, 253, 193, 72, 6, 253, 193, 74, 6, 253, 193, 76, 6, 253, 193, 78, 6, 253, 193, 156, 7, 253, 193, 158, 7, 253, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 0, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 206, 32, 84, 0, 193, 68, 4, 253, 193, 122, 3, 253, 193, 124, 3, 253, 193, 70, 4, 253, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 206, 64, 84, 0, 193, 72, 4, 253, 193, 74, 4, 253, 193, 76, 4, 253, 193, 78, 4, 253, 193, 80, 4, 253, 193, 82, 4, 253, 193, 84, 4, 253, 193, 86, 4, 253, 193, 88, 4, 253, 193, 90, 4, 253, 193, 92, 4, 253, 193, 94, 4, 253, 193, 96, 4, 253, 193, 98, 4, 253, 193, 100, 4, 253, 193, 102, 4, 253, 193, 104, 4, 253, 193, 106, 4, 253, 193, 108, 4, 253, 193, 110, 4, 253, 193, 112, 4, 253, 193, 114, 4, 253, 193, 122, 5, 253, 193, 116, 4, 253, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 96, 84, 0, 206, 128, 84, 0, 206, 128, 84, 0, 206, 128, 84, 0, 206, 128, 84, 0, 206, 128, 84, 0, 206, 128, 84, 0, 206, 128, 84, 0, 206, 128, 84, 0, 206, 128, 84, 0, 206, 128, 84, 0, 206, 128, 84, 0, 193, 46, 4, 253, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 206, 160, 84, 0, 193, 48, 4, 253, 193, 86, 5, 253, 193, 88, 5, 253, 193, 90, 5, 253, 193, 92, 5, 253, 193, 94, 5, 253, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 193, 118, 4, 253, 193, 120, 4, 253, 193, 122, 4, 253, 193, 124, 4, 253, 193, 126, 4, 253, 193, 128, 4, 253, 193, 130, 4, 253, 193, 132, 4, 253, 193, 134, 4, 253, 193, 136, 4, 253, 193, 138, 4, 253, 193, 140, 4, 253, 193, 142, 4, 253, 193, 144, 4, 253, 193, 146, 4, 253, 193, 148, 4, 253, 193, 150, 4, 253, 193, 152, 4, 253, 193, 154, 4, 253, 193, 156, 4, 253, 193, 158, 4, 253, 193, 160, 4, 253, 193, 162, 4, 253, 193, 164, 4, 253, 193, 166, 4, 253, 193, 168, 4, 253, 193, 170, 4, 253, 193, 172, 4, 253, 193, 174, 4, 253, 193, 176, 4, 253, 193, 178, 4, 253, 193, 180, 4, 253, 193, 182, 4, 253, 193, 184, 4, 253, 193, 186, 4, 253, 193, 124, 5, 253, 193, 126, 5, 253, 193, 128, 5, 253, 193, 130, 5, 253, 193, 132, 5, 253, 193, 134, 5, 253, 193, 136, 5, 253, 193, 138, 5, 253, 193, 140, 5, 253, 193, 142, 5, 253, 193, 144, 5, 253, 193, 146, 5, 253, 193, 148, 5, 253, 193, 150, 5, 253, 193, 152, 5, 253, 193, 154, 5, 253, 193, 156, 5, 253, 193, 158, 5, 253, 193, 160, 5, 253, 193, 162, 5, 253, 193, 164, 5, 253, 193, 166, 5, 253, 193, 168, 5, 253, 193, 170, 5, 253, 193, 172, 5, 253, 193, 174, 5, 253, 193, 176, 5, 253, 193, 188, 4, 253, 193, 178, 5, 253, 193, 196, 4, 253, 193, 198, 4, 253, 193, 24, 6, 253, 193, 26, 6, 253, 193, 28, 6, 253, 193, 30, 6, 253, 193, 32, 6, 253, 193, 34, 6, 253, 193, 36, 6, 253, 193, 38, 6, 253, 193, 138, 7, 253, 193, 140, 7, 253, 193, 142, 7, 253, 193, 144, 7, 253, 193, 146, 7, 253, 193, 204, 8, 253, 193, 206, 8, 253, 193, 208, 8, 253, 193, 210, 8, 253, 193, 212, 8, 253, 193, 214, 8, 253, 193, 216, 8, 253, 193, 28, 10, 253, 193, 30, 10, 253, 193, 32, 10, 253, 193, 34, 10, 253, 193, 42, 11, 253, 193, 44, 11, 253, 193, 46, 11, 253, 193, 48, 11, 253, 193, 50, 11, 253, 193, 10, 12, 253, 193, 98, 5, 253, 193, 100, 5, 253, 193, 102, 5, 253, 193, 104, 5, 253, 193, 106, 5, 253, 193, 108, 5, 253, 193, 110, 5, 253, 193, 112, 5, 253, 193, 114, 5, 253, 193, 116, 5, 253, 193, 118, 5, 253, 193, 208, 6, 253, 193, 210, 6, 253, 193, 212, 6, 253, 193, 214, 6, 253, 193, 216, 6, 253, 193, 218, 6, 253, 193, 220, 6, 253, 193, 222, 6, 253, 193, 224, 6, 253, 193, 226, 6, 253, 193, 120, 5, 253, 193, 228, 6, 253, 193, 16, 8, 253, 193, 18, 8, 253, 193, 20, 8, 253, 193, 22, 8, 253, 193, 24, 8, 253, 193, 26, 8, 253, 193, 28, 8, 253, 193, 30, 8, 253, 193, 98, 9, 253, 193, 180, 5, 253, 193, 182, 5, 253, 193, 184, 5, 253, 193, 186, 5, 253, 193, 188, 5, 253, 193, 190, 5, 253, 193, 192, 5, 253, 193, 194, 5, 253, 193, 196, 5, 253, 193, 198, 5, 253, 193, 200, 5, 253, 193, 202, 5, 253, 193, 204, 5, 253, 193, 206, 5, 253, 193, 208, 5, 253, 193, 210, 5, 253, 193, 230, 6, 253, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 192, 84, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 193, 82, 6, 253, 193, 84, 6, 253, 193, 86, 6, 253, 193, 88, 6, 253, 193, 90, 6, 253, 193, 92, 6, 253, 193, 94, 6, 253, 193, 96, 6, 253, 193, 98, 6, 253, 193, 100, 6, 253, 193, 176, 7, 253, 193, 178, 7, 253, 193, 180, 7, 253, 193, 182, 7, 253, 193, 252, 8, 253, 193, 254, 8, 253, 193, 2, 9, 253, 193, 4, 9, 253, 193, 6, 9, 253, 193, 8, 9, 253, 193, 10, 9, 253, 193, 12, 9, 253, 193, 60, 10, 253, 193, 62, 10, 253, 193, 64, 10, 253, 193, 66, 10, 253, 193, 68, 10, 253, 193, 78, 11, 253, 193, 80, 11, 253, 193, 36, 12, 253, 193, 38, 12, 253, 193, 40, 12, 253, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 206, 224, 84, 0, 193, 184, 7, 253, 193, 192, 6, 253, 193, 194, 6, 253, 193, 196, 6, 253, 193, 198, 6, 253, 193, 200, 6, 253, 193, 202, 6, 253, 193, 204, 6, 253, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 193, 94, 10, 253, 193, 206, 6, 253, 193, 96, 10, 253, 193, 98, 10, 253, 193, 100, 10, 253, 193, 102, 10, 253, 193, 104, 10, 253, 193, 106, 10, 253, 193, 108, 10, 253, 193, 110, 10, 253, 193, 112, 10, 253, 193, 114, 10, 253, 193, 116, 10, 253, 193, 118, 10, 253, 193, 120, 10, 253, 193, 122, 10, 253, 193, 124, 10, 253, 193, 126, 10, 253, 193, 128, 10, 253, 193, 130, 10, 253, 193, 132, 10, 253, 193, 134, 10, 253, 193, 82, 11, 253, 193, 84, 11, 253, 193, 86, 11, 253, 193, 88, 11, 253, 193, 90, 11, 253, 193, 92, 11, 253, 193, 94, 11, 253, 193, 96, 11, 253, 193, 98, 11, 253, 193, 100, 11, 253, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 0, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 206, 64, 85, 0, 193, 136, 7, 253, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 193, 160, 7, 253, 193, 162, 7, 253, 193, 164, 7, 253, 193, 166, 7, 253, 193, 168, 7, 253, 193, 170, 7, 253, 193, 172, 7, 253, 193, 174, 7, 253, 193, 230, 8, 253, 193, 232, 8, 253, 193, 234, 8, 253, 193, 236, 8, 253, 193, 238, 8, 253, 193, 240, 8, 253, 193, 242, 8, 253, 193, 244, 8, 253, 193, 246, 8, 253, 193, 248, 8, 253, 193, 250, 8, 253, 193, 44, 10, 253, 193, 46, 10, 253, 193, 48, 10, 253, 193, 50, 10, 253, 193, 52, 10, 253, 193, 54, 10, 253, 193, 56, 10, 253, 193, 58, 10, 253, 193, 64, 11, 253, 193, 66, 11, 253, 193, 68, 11, 253, 193, 70, 11, 253, 193, 72, 11, 253, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 32, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 206, 128, 85, 0, 193, 70, 10, 253, 193, 72, 10, 253, 193, 74, 10, 253, 193, 76, 10, 253, 193, 14, 8, 253, 193, 78, 10, 253, 193, 80, 10, 253, 193, 82, 10, 253, 193, 84, 10, 253, 193, 86, 10, 253, 193, 88, 10, 253, 193, 90, 10, 253, 193, 92, 10, 253, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 96, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 193, 220, 8, 253, 193, 222, 8, 253, 193, 224, 8, 253, 193, 226, 8, 253, 193, 228, 8, 253, 193, 36, 10, 253, 193, 38, 10, 253, 193, 40, 10, 253, 193, 42, 10, 253, 193, 52, 11, 253, 193, 54, 11, 253, 193, 56, 11, 253, 193, 58, 11, 253, 193, 60, 11, 253, 193, 62, 11, 253, 193, 18, 12, 253, 193, 20, 12, 253, 193, 22, 12, 253, 193, 164, 12, 253, 193, 166, 12, 253, 193, 168, 12, 253, 193, 170, 12, 253, 193, 172, 12, 253, 193, 174, 12, 253, 193, 58, 13, 253, 193, 178, 13, 253, 193, 250, 13, 253, 193, 96, 14, 253, 193, 16, 15, 253, 193, 120, 15, 253, 193, 158, 15, 253, 193, 160, 15, 253, 193, 102, 11, 253, 193, 96, 9, 253, 193, 104, 11, 253, 193, 106, 11, 253, 193, 108, 11, 253, 193, 110, 11, 253, 193, 112, 11, 253, 193, 114, 11, 253, 193, 116, 11, 253, 193, 118, 11, 253, 193, 120, 11, 253, 193, 122, 11, 253, 193, 124, 11, 253, 193, 126, 11, 253, 193, 136, 10, 253, 193, 128, 11, 253, 193, 44, 12, 253, 193, 46, 12, 253, 193, 48, 12, 253, 193, 50, 12, 253, 193, 52, 12, 253, 193, 54, 12, 253, 193, 56, 12, 253, 193, 58, 12, 253, 193, 60, 12, 253, 193, 62, 12, 253, 206, 64, 86, 0, 206, 64, 86, 0, 206, 64, 86, 0, 206, 64, 86, 0, 206, 64, 86, 0, 206, 64, 86, 0, 193, 100, 9, 253, 193, 102, 9, 253, 193, 104, 9, 253, 193, 106, 9, 253, 193, 108, 9, 253, 193, 110, 9, 253, 193, 112, 9, 253, 193, 114, 9, 253, 193, 116, 9, 253, 193, 118, 9, 253, 193, 120, 9, 253, 193, 122, 9, 253, 193, 124, 9, 253, 193, 126, 9, 253, 193, 128, 9, 253, 193, 138, 10, 253, 193, 140, 10, 253, 193, 142, 10, 253, 193, 144, 10, 253, 193, 146, 10, 253, 193, 130, 11, 253, 193, 132, 11, 253, 193, 134, 11, 253, 193, 136, 11, 253, 193, 138, 11, 253, 193, 140, 11, 253, 193, 142, 11, 253, 193, 144, 11, 253, 193, 64, 12, 253, 193, 66, 12, 253, 193, 68, 12, 253, 193, 70, 12, 253, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 160, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 192, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 224, 85, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 193, 74, 11, 253, 193, 76, 11, 253, 193, 24, 12, 253, 193, 26, 12, 253, 193, 28, 12, 253, 193, 30, 12, 253, 193, 32, 12, 253, 193, 34, 12, 253, 193, 176, 12, 253, 193, 178, 12, 253, 193, 180, 12, 253, 193, 182, 12, 253, 193, 60, 13, 253, 193, 62, 13, 253, 193, 64, 13, 253, 193, 66, 13, 253, 193, 68, 13, 253, 193, 70, 13, 253, 193, 180, 13, 253, 193, 252, 13, 253, 193, 254, 13, 253, 193, 2, 14, 253, 193, 4, 14, 253, 193, 60, 14, 253, 193, 98, 14, 253, 193, 100, 14, 253, 193, 120, 14, 253, 193, 168, 14, 253, 193, 204, 14, 253, 193, 212, 14, 253, 193, 252, 14, 253, 193, 254, 14, 253, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 0, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 193, 12, 12, 253, 193, 14, 12, 253, 193, 16, 12, 253, 193, 156, 12, 253, 193, 158, 12, 253, 193, 160, 12, 253, 193, 54, 13, 253, 193, 56, 13, 253, 193, 176, 13, 253, 193, 180, 14, 253, 193, 238, 14, 253, 193, 250, 14, 253, 193, 78, 15, 253, 193, 174, 15, 253, 193, 202, 15, 253, 193, 250, 15, 253, 193, 32, 16, 253, 193, 48, 16, 253, 193, 20, 17, 253, 193, 22, 17, 253, 193, 24, 17, 253, 193, 204, 17, 253, 193, 206, 17, 253, 193, 208, 17, 253, 193, 136, 18, 253, 193, 138, 18, 253, 193, 140, 18, 253, 193, 142, 18, 253, 193, 144, 18, 253, 193, 64, 19, 253, 193, 40, 20, 253, 193, 42, 20, 253, 193, 42, 12, 253, 193, 184, 12, 253, 193, 186, 12, 253, 193, 188, 12, 253, 193, 72, 13, 253, 193, 182, 13, 253, 193, 62, 14, 253, 193, 2, 15, 253, 193, 20, 15, 253, 193, 80, 15, 253, 193, 204, 15, 253, 193, 222, 15, 253, 193, 224, 15, 253, 193, 226, 15, 253, 193, 234, 15, 253, 193, 34, 16, 253, 193, 84, 16, 253, 193, 46, 17, 253, 193, 240, 17, 253, 193, 242, 17, 253, 193, 244, 17, 253, 193, 246, 17, 253, 193, 248, 17, 253, 193, 250, 17, 253, 193, 252, 17, 253, 193, 178, 18, 253, 193, 180, 18, 253, 193, 106, 19, 253, 193, 108, 19, 253, 193, 110, 19, 253, 193, 112, 19, 253, 193, 114, 19, 253, 193, 72, 12, 253, 193, 74, 12, 253, 193, 214, 12, 253, 193, 216, 12, 253, 193, 218, 12, 253, 193, 90, 13, 253, 193, 92, 13, 253, 193, 94, 13, 253, 193, 96, 13, 253, 193, 98, 13, 253, 193, 198, 13, 253, 193, 200, 13, 253, 193, 202, 13, 253, 193, 66, 14, 253, 193, 152, 14, 253, 193, 218, 14, 253, 193, 226, 14, 253, 193, 242, 14, 253, 193, 22, 15, 253, 193, 34, 15, 253, 193, 50, 15, 253, 193, 148, 15, 253, 193, 184, 15, 253, 193, 10, 16, 253, 193, 12, 16, 253, 193, 38, 16, 253, 193, 62, 16, 253, 193, 64, 16, 253, 193, 100, 16, 253, 193, 102, 16, 253, 193, 104, 16, 253, 193, 204, 16, 253, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 32, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 64, 86, 0, 206, 64, 86, 0, 206, 64, 86, 0, 206, 64, 86, 0, 206, 64, 86, 0, 193, 74, 13, 253, 193, 76, 13, 253, 193, 78, 13, 253, 193, 80, 13, 253, 193, 82, 13, 253, 193, 84, 13, 253, 193, 86, 13, 253, 193, 88, 13, 253, 193, 184, 13, 253, 193, 186, 13, 253, 193, 188, 13, 253, 193, 190, 13, 253, 193, 192, 13, 253, 193, 194, 13, 253, 193, 196, 13, 253, 193, 212, 12, 253, 193, 6, 14, 253, 193, 8, 14, 253, 193, 10, 14, 253, 193, 12, 14, 253, 193, 14, 14, 253, 193, 16, 14, 253, 193, 64, 14, 253, 193, 102, 14, 253, 193, 122, 14, 253, 193, 124, 14, 253, 193, 150, 14, 253, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 206, 96, 86, 0, 193, 100, 13, 253, 193, 102, 13, 253, 193, 104, 13, 253, 193, 106, 13, 253, 193, 108, 13, 253, 193, 110, 13, 253, 193, 112, 13, 253, 193, 114, 13, 253, 193, 116, 13, 253, 193, 118, 13, 253, 193, 120, 13, 253, 193, 122, 13, 253, 193, 124, 13, 253, 193, 126, 13, 253, 193, 128, 13, 253, 193, 130, 13, 253, 193, 132, 13, 253, 193, 134, 13, 253, 193, 136, 13, 253, 193, 138, 13, 253, 193, 140, 13, 253, 193, 142, 13, 253, 193, 144, 13, 253, 193, 146, 13, 253, 193, 148, 13, 253, 193, 150, 13, 253, 193, 152, 13, 253, 193, 154, 13, 253, 193, 156, 13, 253, 193, 158, 13, 253, 193, 160, 13, 253, 193, 162, 13, 253, 193, 164, 13, 253, 193, 166, 13, 253, 193, 168, 13, 253, 193, 170, 13, 253, 193, 172, 13, 253, 193, 174, 13, 253, 193, 204, 13, 253, 193, 206, 13, 253, 193, 208, 13, 253, 193, 210, 13, 253, 193, 212, 13, 253, 193, 214, 13, 253, 193, 216, 13, 253, 193, 218, 13, 253, 193, 220, 13, 253, 193, 222, 13, 253, 193, 224, 13, 253, 193, 226, 13, 253, 193, 228, 13, 253, 193, 230, 13, 253, 193, 232, 13, 253, 193, 234, 13, 253, 193, 18, 14, 253, 193, 236, 13, 253, 193, 238, 13, 253, 193, 240, 13, 253, 193, 242, 13, 253, 193, 244, 13, 253, 193, 246, 13, 253, 193, 248, 13, 253, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 206, 128, 86, 0, 193, 68, 14, 253, 193, 70, 14, 253, 193, 72, 14, 253, 193, 74, 14, 253, 193, 76, 14, 253, 193, 78, 14, 253, 193, 80, 14, 253, 193, 82, 14, 253, 193, 84, 14, 253, 193, 86, 14, 253, 193, 88, 14, 253, 193, 90, 14, 253, 193, 92, 14, 253, 193, 94, 14, 253, 193, 104, 14, 253, 193, 106, 14, 253, 193, 108, 14, 253, 193, 110, 14, 253, 193, 112, 14, 253, 193, 114, 14, 253, 193, 116, 14, 253, 193, 126, 14, 253, 193, 128, 14, 253, 193, 130, 14, 253, 193, 132, 14, 253, 193, 134, 14, 253, 193, 136, 14, 253, 193, 118, 14, 253, 193, 138, 14, 253, 193, 140, 14, 253, 193, 142, 14, 253, 193, 144, 14, 253, 193, 146, 14, 253, 193, 148, 14, 253, 193, 154, 14, 253, 193, 156, 14, 253, 193, 158, 14, 253, 193, 160, 14, 253, 193, 162, 14, 253, 193, 164, 14, 253, 193, 166, 14, 253, 193, 172, 14, 253, 193, 174, 14, 253, 193, 176, 14, 253, 193, 178, 14, 253, 193, 182, 14, 253, 193, 184, 14, 253, 193, 186, 14, 253, 193, 188, 14, 253, 193, 18, 206, 253, 193, 190, 14, 253, 193, 192, 14, 253, 193, 194, 14, 253, 193, 196, 14, 253, 193, 198, 14, 253, 193, 200, 14, 253, 193, 206, 14, 253, 193, 208, 14, 253, 193, 210, 14, 253, 193, 220, 14, 253, 193, 228, 14, 253, 193, 230, 14, 253, 193, 232, 14, 253, 193, 234, 14, 253, 193, 236, 14, 253, 193, 244, 14, 253, 193, 246, 14, 253, 193, 248, 14, 253, 193, 12, 15, 253, 193, 14, 15, 253, 193, 24, 15, 253, 193, 26, 15, 253, 193, 28, 15, 253, 193, 30, 15, 253, 193, 32, 15, 253, 193, 36, 15, 253, 193, 38, 15, 253, 193, 40, 15, 253, 193, 42, 15, 253, 193, 44, 15, 253, 193, 46, 15, 253, 193, 202, 14, 253, 193, 214, 14, 253, 193, 216, 14, 253, 193, 222, 14, 253, 193, 224, 14, 253, 193, 240, 14, 253, 193, 4, 15, 253, 193, 6, 15, 253, 193, 8, 15, 253, 193, 10, 15, 253, 193, 102, 15, 253, 193, 114, 15, 253, 193, 118, 15, 253, 193, 130, 15, 253, 193, 132, 15, 253, 193, 146, 15, 253, 193, 162, 15, 253, 193, 164, 15, 253, 193, 166, 15, 253, 193, 168, 15, 253, 193, 170, 15, 253, 193, 176, 15, 253, 193, 178, 15, 253, 193, 180, 15, 253, 193, 182, 15, 253, 193, 206, 15, 253, 193, 236, 15, 253, 193, 238, 15, 253, 193, 240, 15, 253, 193, 242, 15, 253, 193, 2, 16, 253, 193, 4, 16, 253, 193, 18, 15, 253, 193, 106, 15, 253, 193, 144, 15, 253, 193, 220, 15, 253, 193, 58, 16, 253, 193, 78, 16, 253, 193, 80, 16, 253, 193, 82, 16, 253, 193, 148, 16, 253, 193, 32, 17, 253, 193, 34, 17, 253, 193, 36, 17, 253, 193, 38, 17, 253, 193, 40, 17, 253, 193, 42, 17, 253, 193, 44, 17, 253, 193, 222, 17, 253, 193, 224, 17, 253, 193, 226, 17, 253, 193, 228, 17, 253, 193, 230, 17, 253, 193, 232, 17, 253, 193, 234, 17, 253, 193, 236, 17, 253, 193, 238, 17, 253, 193, 162, 18, 253, 193, 164, 18, 253, 193, 166, 18, 253, 193, 168, 18, 253, 193, 170, 18, 253, 193, 172, 18, 253, 193, 174, 18, 253, 193, 52, 15, 253, 193, 54, 15, 253, 193, 56, 15, 253, 193, 58, 15, 253, 193, 48, 15, 253, 193, 60, 15, 253, 193, 62, 15, 253, 193, 64, 15, 253, 193, 66, 15, 253, 193, 68, 15, 253, 193, 70, 15, 253, 193, 72, 15, 253, 193, 74, 15, 253, 193, 76, 15, 253, 193, 82, 15, 253, 193, 84, 15, 253, 193, 86, 15, 253, 193, 88, 15, 253, 193, 90, 15, 253, 193, 92, 15, 253, 193, 94, 15, 253, 193, 96, 15, 253, 193, 98, 15, 253, 193, 100, 15, 253, 193, 104, 15, 253, 193, 108, 15, 253, 193, 110, 15, 253, 193, 112, 15, 253, 193, 116, 15, 253, 193, 122, 15, 253, 193, 124, 15, 253, 193, 126, 15, 253, 193, 128, 15, 253, 193, 134, 15, 253, 193, 136, 15, 253, 193, 138, 15, 253, 193, 140, 15, 253, 193, 142, 15, 253, 193, 150, 15, 253, 193, 152, 15, 253, 193, 154, 15, 253, 193, 156, 15, 253, 193, 172, 15, 253, 193, 188, 15, 253, 193, 190, 15, 253, 193, 192, 15, 253, 193, 194, 15, 253, 193, 196, 15, 253, 193, 198, 15, 253, 193, 200, 15, 253, 193, 208, 15, 253, 193, 210, 15, 253, 193, 212, 15, 253, 193, 214, 15, 253, 193, 216, 15, 253, 193, 218, 15, 253, 193, 228, 15, 253, 193, 244, 15, 253, 193, 230, 15, 253, 193, 232, 15, 253, 193, 246, 15, 253, 193, 248, 15, 253, 193, 16, 16, 253, 193, 20, 16, 253, 193, 254, 15, 253, 193, 44, 16, 253, 193, 76, 16, 253, 193, 138, 16, 253, 193, 140, 16, 253, 193, 142, 16, 253, 193, 144, 16, 253, 193, 146, 16, 253, 193, 26, 17, 253, 193, 28, 17, 253, 193, 30, 17, 253, 193, 210, 17, 253, 193, 212, 17, 253, 193, 214, 17, 253, 193, 216, 17, 253, 193, 218, 17, 253, 193, 220, 17, 253, 193, 146, 18, 253, 193, 148, 18, 253, 193, 150, 18, 253, 193, 152, 18, 253, 193, 154, 18, 253, 193, 156, 18, 253, 193, 158, 18, 253, 193, 160, 18, 253, 193, 68, 19, 253, 193, 70, 19, 253, 193, 72, 19, 253, 193, 74, 19, 253, 193, 76, 19, 253, 193, 54, 20, 253, 193, 56, 20, 253, 193, 6, 16, 253, 193, 8, 16, 253, 193, 50, 16, 253, 193, 52, 16, 253, 193, 54, 16, 253, 193, 60, 16, 253, 193, 86, 16, 253, 193, 88, 16, 253, 193, 90, 16, 253, 193, 92, 16, 253, 193, 94, 16, 253, 193, 96, 16, 253, 193, 98, 16, 253, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 193, 22, 16, 253, 193, 24, 16, 253, 193, 28, 16, 253, 193, 30, 16, 253, 193, 36, 16, 253, 193, 42, 16, 253, 193, 46, 16, 253, 193, 56, 16, 253, 193, 66, 16, 253, 193, 68, 16, 253, 193, 70, 16, 253, 193, 72, 16, 253, 193, 74, 16, 253, 193, 106, 16, 253, 193, 108, 16, 253, 193, 110, 16, 253, 193, 112, 16, 253, 193, 114, 16, 253, 193, 116, 16, 253, 193, 118, 16, 253, 193, 120, 16, 253, 193, 122, 16, 253, 193, 124, 16, 253, 193, 126, 16, 253, 193, 128, 16, 253, 193, 130, 16, 253, 193, 132, 16, 253, 193, 134, 16, 253, 193, 136, 16, 253, 193, 210, 16, 253, 193, 212, 16, 253, 193, 214, 16, 253, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 160, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 193, 206, 16, 253, 193, 208, 16, 253, 193, 128, 17, 253, 193, 130, 17, 253, 193, 132, 17, 253, 193, 134, 17, 253, 193, 136, 17, 253, 193, 138, 17, 253, 193, 52, 18, 253, 193, 54, 18, 253, 193, 56, 18, 253, 193, 58, 18, 253, 193, 60, 18, 253, 193, 62, 18, 253, 193, 64, 18, 253, 193, 238, 18, 253, 193, 240, 18, 253, 193, 242, 18, 253, 193, 244, 18, 253, 193, 178, 19, 253, 193, 180, 19, 253, 193, 182, 19, 253, 193, 184, 19, 253, 193, 186, 19, 253, 193, 188, 19, 253, 193, 190, 19, 253, 193, 192, 19, 253, 193, 194, 19, 253, 193, 196, 19, 253, 193, 198, 19, 253, 193, 200, 19, 253, 193, 202, 19, 253, 193, 216, 16, 253, 193, 218, 16, 253, 193, 220, 16, 253, 193, 222, 16, 253, 193, 224, 16, 253, 193, 226, 16, 253, 193, 228, 16, 253, 193, 230, 16, 253, 193, 232, 16, 253, 193, 234, 16, 253, 193, 236, 16, 253, 193, 238, 16, 253, 193, 240, 16, 253, 193, 242, 16, 253, 193, 244, 16, 253, 193, 246, 16, 253, 193, 248, 16, 253, 193, 250, 16, 253, 193, 252, 16, 253, 193, 254, 16, 253, 193, 2, 17, 253, 193, 4, 17, 253, 193, 6, 17, 253, 193, 8, 17, 253, 193, 10, 17, 253, 193, 12, 17, 253, 193, 14, 17, 253, 193, 16, 17, 253, 193, 18, 17, 253, 193, 140, 17, 253, 193, 66, 18, 253, 193, 142, 17, 253, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 206, 192, 86, 0, 193, 254, 17, 253, 193, 126, 17, 253, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 193, 144, 17, 253, 193, 146, 17, 253, 193, 148, 17, 253, 193, 150, 17, 253, 193, 152, 17, 253, 193, 154, 17, 253, 193, 156, 17, 253, 193, 158, 17, 253, 193, 160, 17, 253, 193, 162, 17, 253, 193, 164, 17, 253, 193, 166, 17, 253, 193, 168, 17, 253, 193, 170, 17, 253, 193, 172, 17, 253, 193, 174, 17, 253, 193, 176, 17, 253, 193, 178, 17, 253, 193, 180, 17, 253, 193, 182, 17, 253, 193, 184, 17, 253, 193, 186, 17, 253, 193, 188, 17, 253, 193, 190, 17, 253, 193, 192, 17, 253, 193, 194, 17, 253, 193, 196, 17, 253, 193, 198, 17, 253, 193, 200, 17, 253, 193, 202, 17, 253, 193, 68, 18, 253, 193, 70, 18, 253, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 224, 86, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 193, 72, 18, 253, 193, 74, 18, 253, 193, 76, 18, 253, 193, 78, 18, 253, 193, 80, 18, 253, 193, 82, 18, 253, 193, 84, 18, 253, 193, 86, 18, 253, 193, 88, 18, 253, 193, 90, 18, 253, 193, 92, 18, 253, 193, 94, 18, 253, 193, 96, 18, 253, 193, 98, 18, 253, 193, 100, 18, 253, 193, 102, 18, 253, 193, 104, 18, 253, 193, 106, 18, 253, 193, 108, 18, 253, 193, 110, 18, 253, 193, 112, 18, 253, 193, 114, 18, 253, 193, 116, 18, 253, 193, 118, 18, 253, 193, 120, 18, 253, 193, 122, 18, 253, 193, 124, 18, 253, 193, 126, 18, 253, 193, 128, 18, 253, 193, 130, 18, 253, 193, 132, 18, 253, 193, 134, 18, 253, 193, 176, 18, 253, 193, 78, 19, 253, 193, 80, 19, 253, 193, 82, 19, 253, 193, 84, 19, 253, 193, 86, 19, 253, 193, 88, 19, 253, 193, 90, 19, 253, 193, 92, 19, 253, 193, 94, 19, 253, 193, 96, 19, 253, 193, 98, 19, 253, 193, 100, 19, 253, 193, 102, 19, 253, 193, 104, 19, 253, 193, 68, 20, 253, 193, 70, 20, 253, 193, 72, 20, 253, 193, 74, 20, 253, 193, 76, 20, 253, 193, 248, 20, 253, 193, 250, 20, 253, 193, 252, 20, 253, 193, 254, 20, 253, 193, 2, 21, 253, 193, 4, 21, 253, 193, 6, 21, 253, 193, 8, 21, 253, 193, 10, 21, 253, 193, 12, 21, 253, 193, 14, 21, 253, 193, 16, 21, 253, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 206, 0, 87, 0, 193, 82, 20, 253, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 206, 32, 87, 0, 193, 210, 19, 253, 193, 212, 19, 253, 193, 214, 19, 253, 193, 216, 19, 253, 193, 218, 19, 253, 193, 220, 19, 253, 193, 222, 19, 253, 193, 224, 19, 253, 193, 226, 19, 253, 193, 228, 19, 253, 193, 230, 19, 253, 193, 232, 19, 253, 193, 234, 19, 253, 193, 236, 19, 253, 193, 238, 19, 253, 193, 240, 19, 253, 193, 242, 19, 253, 193, 244, 19, 253, 193, 246, 19, 253, 193, 248, 19, 253, 193, 250, 19, 253, 193, 252, 19, 253, 193, 254, 19, 253, 193, 156, 20, 253, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 193, 116, 19, 253, 193, 78, 20, 253, 193, 80, 20, 253, 193, 182, 21, 253, 193, 184, 21, 253, 193, 186, 21, 253, 193, 92, 22, 253, 193, 94, 22, 253, 193, 96, 22, 253, 193, 34, 24, 253, 193, 100, 24, 253, 193, 234, 24, 253, 193, 254, 24, 253, 193, 30, 25, 253, 193, 170, 25, 253, 193, 172, 25, 253, 193, 174, 25, 253, 193, 32, 26, 253, 193, 104, 26, 253, 193, 216, 26, 253, 193, 66, 27, 253, 193, 176, 27, 253, 193, 178, 27, 253, 193, 192, 27, 253, 193, 194, 27, 253, 193, 196, 27, 253, 193, 214, 27, 253, 193, 220, 27, 253, 193, 230, 27, 253, 193, 234, 27, 253, 193, 238, 27, 253, 193, 244, 27, 253, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 206, 64, 87, 0, 193, 84, 20, 253, 193, 86, 20, 253, 193, 88, 20, 253, 193, 90, 20, 253, 193, 92, 20, 253, 193, 94, 20, 253, 193, 96, 20, 253, 193, 98, 20, 253, 193, 100, 20, 253, 193, 102, 20, 253, 193, 104, 20, 253, 193, 106, 20, 253, 193, 108, 20, 253, 193, 110, 20, 253, 193, 112, 20, 253, 193, 114, 20, 253, 193, 116, 20, 253, 193, 118, 20, 253, 193, 120, 20, 253, 193, 122, 20, 253, 193, 124, 20, 253, 193, 126, 20, 253, 193, 128, 20, 253, 193, 130, 20, 253, 193, 132, 20, 253, 193, 20, 21, 253, 193, 22, 21, 253, 193, 204, 19, 253, 193, 206, 19, 253, 193, 208, 19, 253, 193, 134, 20, 253, 193, 136, 20, 253, 193, 138, 20, 253, 193, 140, 20, 253, 193, 142, 20, 253, 193, 144, 20, 253, 193, 146, 20, 253, 193, 148, 20, 253, 193, 150, 20, 253, 193, 152, 20, 253, 193, 154, 20, 253, 193, 64, 21, 253, 193, 66, 21, 253, 193, 68, 21, 253, 193, 70, 21, 253, 193, 72, 21, 253, 193, 74, 21, 253, 193, 226, 21, 253, 193, 228, 21, 253, 193, 230, 21, 253, 193, 232, 21, 253, 193, 234, 21, 253, 193, 236, 21, 253, 193, 238, 21, 253, 193, 240, 21, 253, 193, 242, 21, 253, 193, 244, 21, 253, 193, 130, 22, 253, 193, 132, 22, 253, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 96, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 193, 44, 20, 253, 193, 46, 20, 253, 193, 48, 20, 253, 193, 50, 20, 253, 193, 52, 20, 253, 193, 240, 20, 253, 193, 242, 20, 253, 193, 152, 21, 253, 193, 154, 21, 253, 193, 156, 21, 253, 193, 62, 22, 253, 193, 64, 22, 253, 193, 66, 22, 253, 193, 68, 22, 253, 193, 70, 22, 253, 193, 72, 22, 253, 193, 238, 22, 253, 193, 240, 22, 253, 193, 242, 22, 253, 193, 96, 23, 253, 193, 98, 23, 253, 193, 182, 23, 253, 193, 238, 23, 253, 193, 56, 24, 253, 193, 76, 24, 253, 193, 78, 24, 253, 193, 180, 24, 253, 193, 252, 24, 253, 193, 72, 25, 253, 193, 164, 25, 253, 193, 222, 25, 253, 193, 22, 26, 253, 193, 58, 20, 253, 193, 60, 20, 253, 193, 62, 20, 253, 193, 64, 20, 253, 193, 66, 20, 253, 193, 244, 20, 253, 193, 246, 20, 253, 193, 158, 21, 253, 193, 160, 21, 253, 193, 162, 21, 253, 193, 74, 22, 253, 193, 76, 22, 253, 193, 78, 22, 253, 193, 80, 22, 253, 193, 244, 22, 253, 193, 246, 22, 253, 193, 100, 23, 253, 193, 102, 23, 253, 193, 184, 23, 253, 193, 186, 23, 253, 193, 188, 23, 253, 193, 240, 23, 253, 193, 242, 23, 253, 193, 32, 24, 253, 193, 168, 24, 253, 193, 212, 24, 253, 193, 230, 24, 253, 193, 10, 25, 253, 193, 120, 25, 253, 193, 166, 25, 253, 193, 92, 26, 253, 193, 94, 26, 253, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 128, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 193, 18, 21, 253, 193, 164, 21, 253, 193, 166, 21, 253, 193, 168, 21, 253, 193, 170, 21, 253, 193, 172, 21, 253, 193, 174, 21, 253, 193, 176, 21, 253, 193, 178, 21, 253, 193, 180, 21, 253, 193, 82, 22, 253, 193, 84, 22, 253, 193, 86, 22, 253, 193, 88, 22, 253, 193, 90, 22, 253, 193, 248, 22, 253, 193, 250, 22, 253, 193, 252, 22, 253, 193, 254, 22, 253, 193, 2, 23, 253, 193, 4, 23, 253, 193, 6, 23, 253, 193, 8, 23, 253, 193, 104, 23, 253, 193, 106, 23, 253, 193, 108, 23, 253, 193, 110, 23, 253, 193, 190, 23, 253, 193, 192, 23, 253, 193, 244, 23, 253, 193, 246, 23, 253, 193, 182, 24, 253, 193, 24, 21, 253, 193, 26, 21, 253, 193, 28, 21, 253, 193, 30, 21, 253, 193, 32, 21, 253, 193, 34, 21, 253, 193, 36, 21, 253, 193, 38, 21, 253, 193, 40, 21, 253, 193, 42, 21, 253, 193, 44, 21, 253, 193, 46, 21, 253, 193, 48, 21, 253, 193, 50, 21, 253, 193, 52, 21, 253, 193, 54, 21, 253, 193, 56, 21, 253, 193, 58, 21, 253, 193, 60, 21, 253, 193, 188, 21, 253, 193, 62, 21, 253, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 206, 160, 87, 0, 193, 246, 21, 253, 193, 248, 21, 253, 193, 250, 21, 253, 193, 252, 21, 253, 193, 254, 21, 253, 193, 2, 22, 253, 193, 4, 22, 253, 193, 6, 22, 253, 193, 8, 22, 253, 193, 10, 22, 253, 193, 12, 22, 253, 193, 150, 21, 253, 193, 14, 22, 253, 193, 16, 22, 253, 193, 18, 22, 253, 193, 20, 22, 253, 193, 142, 22, 253, 193, 22, 22, 253, 193, 24, 22, 253, 193, 26, 22, 253, 193, 28, 22, 253, 193, 30, 22, 253, 193, 32, 22, 253, 193, 34, 22, 253, 193, 36, 22, 253, 193, 38, 22, 253, 193, 40, 22, 253, 193, 42, 22, 253, 193, 44, 22, 253, 193, 46, 22, 253, 193, 48, 22, 253, 193, 50, 22, 253, 193, 52, 22, 253, 193, 54, 22, 253, 193, 56, 22, 253, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 206, 192, 87, 0, 193, 98, 22, 253, 193, 100, 22, 253, 193, 102, 22, 253, 193, 104, 22, 253, 193, 106, 22, 253, 193, 108, 22, 253, 193, 110, 22, 253, 193, 112, 22, 253, 193, 114, 22, 253, 193, 116, 22, 253, 193, 118, 22, 253, 193, 120, 22, 253, 193, 122, 22, 253, 193, 124, 22, 253, 193, 224, 21, 253, 193, 126, 22, 253, 193, 10, 23, 253, 193, 12, 23, 253, 193, 14, 23, 253, 193, 16, 23, 253, 193, 18, 23, 253, 193, 20, 23, 253, 193, 22, 23, 253, 193, 24, 23, 253, 193, 26, 23, 253, 193, 28, 23, 253, 193, 58, 22, 253, 193, 60, 22, 253, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 193, 134, 22, 253, 193, 136, 22, 253, 193, 138, 22, 253, 193, 140, 22, 253, 193, 36, 23, 253, 193, 38, 23, 253, 193, 124, 23, 253, 193, 126, 23, 253, 193, 128, 23, 253, 193, 130, 23, 253, 193, 206, 23, 253, 193, 2, 24, 253, 193, 42, 24, 253, 193, 44, 24, 253, 193, 64, 24, 253, 193, 66, 24, 253, 193, 90, 24, 253, 193, 112, 24, 253, 193, 152, 24, 253, 193, 160, 24, 253, 193, 172, 24, 253, 193, 190, 24, 253, 193, 192, 24, 253, 193, 58, 25, 253, 193, 60, 25, 253, 193, 86, 25, 253, 193, 88, 25, 253, 193, 90, 25, 253, 193, 132, 25, 253, 193, 134, 25, 253, 193, 136, 25, 253, 193, 138, 25, 253, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 224, 87, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 193, 30, 23, 253, 193, 32, 23, 253, 193, 34, 23, 253, 193, 112, 23, 253, 193, 114, 23, 253, 193, 116, 23, 253, 193, 118, 23, 253, 193, 120, 23, 253, 193, 122, 23, 253, 193, 194, 23, 253, 193, 196, 23, 253, 193, 198, 23, 253, 193, 200, 23, 253, 193, 202, 23, 253, 193, 204, 23, 253, 193, 248, 23, 253, 193, 250, 23, 253, 193, 252, 23, 253, 193, 254, 23, 253, 193, 36, 24, 253, 193, 38, 24, 253, 193, 58, 24, 253, 193, 60, 24, 253, 193, 62, 24, 253, 193, 80, 24, 253, 193, 82, 24, 253, 193, 84, 24, 253, 193, 86, 24, 253, 193, 88, 24, 253, 193, 40, 24, 253, 193, 136, 24, 253, 193, 138, 24, 253, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 206, 0, 88, 0, 193, 132, 23, 253, 193, 134, 23, 253, 193, 136, 23, 253, 193, 138, 23, 253, 193, 140, 23, 253, 193, 142, 23, 253, 193, 144, 23, 253, 193, 146, 23, 253, 193, 148, 23, 253, 193, 150, 23, 253, 193, 208, 23, 253, 206, 32, 88, 0, 206, 32, 88, 0, 206, 32, 88, 0, 206, 32, 88, 0, 206, 32, 88, 0, 206, 32, 88, 0, 206, 32, 88, 0, 206, 32, 88, 0, 206, 32, 88, 0, 206, 32, 88, 0, 206, 32, 88, 0, 193, 210, 23, 253, 193, 212, 23, 253, 193, 214, 23, 253, 193, 216, 23, 253, 193, 218, 23, 253, 193, 220, 23, 253, 193, 222, 23, 253, 193, 224, 23, 253, 193, 226, 23, 253, 193, 228, 23, 253, 193, 230, 23, 253, 193, 4, 24, 253, 193, 232, 23, 253, 193, 234, 23, 253, 193, 236, 23, 253, 193, 6, 24, 253, 193, 8, 24, 253, 193, 10, 24, 253, 193, 12, 24, 253, 193, 14, 24, 253, 193, 16, 24, 253, 193, 18, 24, 253, 193, 20, 24, 253, 193, 22, 24, 253, 193, 24, 24, 253, 193, 26, 24, 253, 193, 28, 24, 253, 193, 30, 24, 253, 193, 46, 24, 253, 193, 48, 24, 253, 193, 50, 24, 253, 193, 52, 24, 253, 193, 54, 24, 253, 193, 68, 24, 253, 193, 70, 24, 253, 193, 72, 24, 253, 193, 74, 24, 253, 193, 92, 24, 253, 193, 94, 24, 253, 193, 96, 24, 253, 193, 98, 24, 253, 193, 102, 24, 253, 193, 104, 24, 253, 193, 106, 24, 253, 193, 108, 24, 253, 193, 110, 24, 253, 193, 114, 24, 253, 193, 116, 24, 253, 193, 118, 24, 253, 193, 120, 24, 253, 193, 122, 24, 253, 193, 124, 24, 253, 193, 126, 24, 253, 193, 128, 24, 253, 193, 130, 24, 253, 193, 132, 24, 253, 193, 134, 24, 253, 193, 142, 24, 253, 193, 144, 24, 253, 193, 146, 24, 253, 193, 148, 24, 253, 193, 154, 24, 253, 193, 156, 24, 253, 193, 158, 24, 253, 193, 162, 24, 253, 193, 164, 24, 253, 193, 166, 24, 253, 193, 174, 24, 253, 193, 176, 24, 253, 193, 178, 24, 253, 193, 194, 24, 253, 193, 196, 24, 253, 193, 208, 24, 253, 193, 210, 24, 253, 193, 214, 24, 253, 193, 216, 24, 253, 193, 218, 24, 253, 193, 220, 24, 253, 193, 222, 24, 253, 193, 224, 24, 253, 193, 226, 24, 253, 193, 228, 24, 253, 193, 236, 24, 253, 193, 238, 24, 253, 193, 240, 24, 253, 193, 246, 24, 253, 193, 248, 24, 253, 193, 250, 24, 253, 193, 2, 25, 253, 193, 184, 24, 253, 193, 200, 24, 253, 193, 202, 24, 253, 193, 204, 24, 253, 193, 232, 24, 253, 193, 242, 24, 253, 193, 244, 24, 253, 193, 18, 25, 253, 193, 22, 25, 253, 193, 26, 25, 253, 193, 28, 25, 253, 193, 34, 25, 253, 193, 40, 25, 253, 193, 122, 25, 253, 193, 168, 25, 253, 193, 224, 25, 253, 193, 226, 25, 253, 193, 228, 25, 253, 193, 26, 26, 253, 193, 28, 26, 253, 193, 30, 26, 253, 193, 98, 26, 253, 193, 100, 26, 253, 193, 102, 26, 253, 193, 172, 26, 253, 193, 214, 26, 253, 193, 36, 27, 253, 193, 38, 27, 253, 193, 40, 27, 253, 193, 42, 27, 253, 193, 64, 27, 253, 193, 248, 27, 253, 193, 6, 25, 253, 193, 8, 25, 253, 193, 16, 25, 253, 193, 20, 25, 253, 193, 24, 25, 253, 193, 38, 25, 253, 193, 44, 25, 253, 193, 46, 25, 253, 193, 52, 25, 253, 193, 62, 25, 253, 193, 64, 25, 253, 193, 66, 25, 253, 193, 68, 25, 253, 193, 70, 25, 253, 193, 92, 25, 253, 193, 94, 25, 253, 193, 96, 25, 253, 193, 98, 25, 253, 193, 100, 25, 253, 193, 102, 25, 253, 193, 104, 25, 253, 193, 106, 25, 253, 193, 108, 25, 253, 193, 110, 25, 253, 193, 112, 25, 253, 193, 114, 25, 253, 193, 116, 25, 253, 193, 118, 25, 253, 206, 64, 88, 0, 206, 64, 88, 0, 206, 64, 88, 0, 206, 64, 88, 0, 206, 64, 88, 0, 206, 64, 88, 0, 206, 64, 88, 0, 206, 64, 88, 0, 193, 184, 25, 253, 193, 186, 25, 253, 193, 188, 25, 253, 193, 190, 25, 253, 193, 192, 25, 253, 193, 194, 25, 253, 193, 196, 25, 253, 193, 198, 25, 253, 193, 200, 25, 253, 193, 202, 25, 253, 193, 204, 25, 253, 193, 206, 25, 253, 193, 208, 25, 253, 193, 210, 25, 253, 193, 212, 25, 253, 193, 214, 25, 253, 193, 216, 25, 253, 193, 218, 25, 253, 193, 220, 25, 253, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 193, 180, 25, 253, 193, 182, 25, 253, 193, 240, 25, 253, 193, 242, 25, 253, 193, 42, 26, 253, 193, 44, 26, 253, 193, 116, 26, 253, 193, 118, 26, 253, 193, 120, 26, 253, 193, 122, 26, 253, 193, 184, 26, 253, 193, 186, 26, 253, 193, 188, 26, 253, 193, 226, 26, 253, 193, 228, 26, 253, 193, 10, 27, 253, 193, 12, 27, 253, 193, 14, 27, 253, 193, 16, 27, 253, 193, 18, 27, 253, 193, 72, 27, 253, 193, 132, 27, 253, 193, 140, 27, 253, 193, 34, 28, 253, 193, 162, 28, 253, 193, 164, 28, 253, 193, 166, 28, 253, 193, 206, 28, 253, 193, 252, 28, 253, 193, 254, 28, 253, 193, 2, 29, 253, 193, 4, 29, 253, 193, 236, 25, 253, 193, 238, 25, 253, 193, 34, 26, 253, 193, 36, 26, 253, 193, 38, 26, 253, 193, 40, 26, 253, 193, 106, 26, 253, 193, 108, 26, 253, 193, 110, 26, 253, 193, 112, 26, 253, 193, 114, 26, 253, 193, 174, 26, 253, 193, 176, 26, 253, 193, 178, 26, 253, 193, 180, 26, 253, 193, 182, 26, 253, 193, 218, 26, 253, 193, 220, 26, 253, 193, 222, 26, 253, 193, 224, 26, 253, 193, 254, 26, 253, 193, 2, 27, 253, 193, 4, 27, 253, 193, 6, 27, 253, 193, 8, 27, 253, 193, 44, 27, 253, 193, 46, 27, 253, 193, 48, 27, 253, 193, 68, 27, 253, 193, 70, 27, 253, 193, 84, 27, 253, 193, 86, 27, 253, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 206, 96, 88, 0, 193, 46, 26, 253, 193, 48, 26, 253, 193, 50, 26, 253, 193, 52, 26, 253, 193, 54, 26, 253, 193, 56, 26, 253, 193, 58, 26, 253, 193, 60, 26, 253, 193, 62, 26, 253, 193, 64, 26, 253, 193, 66, 26, 253, 193, 68, 26, 253, 193, 70, 26, 253, 193, 72, 26, 253, 193, 74, 26, 253, 193, 76, 26, 253, 193, 78, 26, 253, 193, 80, 26, 253, 193, 82, 26, 253, 193, 84, 26, 253, 193, 86, 26, 253, 193, 24, 26, 253, 193, 90, 26, 253, 193, 168, 26, 253, 193, 170, 26, 253, 193, 250, 26, 253, 193, 62, 27, 253, 193, 128, 27, 253, 193, 130, 27, 253, 193, 166, 27, 253, 193, 168, 27, 253, 193, 182, 27, 253, 193, 206, 27, 253, 193, 226, 27, 253, 193, 236, 27, 253, 193, 182, 28, 253, 193, 184, 28, 253, 193, 104, 29, 253, 193, 106, 29, 253, 193, 154, 30, 253, 193, 156, 30, 253, 193, 90, 31, 253, 193, 92, 31, 253, 193, 54, 32, 253, 193, 56, 32, 253, 193, 58, 32, 253, 193, 60, 32, 253, 193, 40, 33, 253, 193, 42, 33, 253, 193, 34, 34, 253, 193, 36, 34, 253, 193, 38, 34, 253, 193, 60, 35, 253, 193, 88, 26, 253, 193, 124, 26, 253, 193, 126, 26, 253, 193, 128, 26, 253, 193, 130, 26, 253, 193, 132, 26, 253, 193, 134, 26, 253, 193, 136, 26, 253, 193, 138, 26, 253, 193, 140, 26, 253, 193, 142, 26, 253, 193, 144, 26, 253, 193, 146, 26, 253, 193, 148, 26, 253, 193, 150, 26, 253, 193, 152, 26, 253, 193, 154, 26, 253, 193, 156, 26, 253, 193, 158, 26, 253, 193, 160, 26, 253, 193, 162, 26, 253, 193, 164, 26, 253, 193, 166, 26, 253, 193, 190, 26, 253, 193, 192, 26, 253, 193, 194, 26, 253, 193, 196, 26, 253, 193, 198, 26, 253, 193, 200, 26, 253, 193, 202, 26, 253, 193, 204, 26, 253, 193, 206, 26, 253, 193, 96, 26, 253, 193, 212, 26, 253, 193, 252, 26, 253, 193, 32, 27, 253, 193, 34, 27, 253, 193, 82, 27, 253, 193, 104, 27, 253, 193, 152, 27, 253, 193, 170, 27, 253, 193, 172, 27, 253, 193, 174, 27, 253, 193, 184, 27, 253, 193, 190, 27, 253, 193, 186, 27, 253, 193, 202, 27, 253, 193, 208, 27, 253, 193, 210, 27, 253, 193, 212, 27, 253, 193, 218, 27, 253, 193, 228, 27, 253, 193, 232, 27, 253, 193, 132, 28, 253, 193, 186, 28, 253, 193, 232, 28, 253, 193, 54, 29, 253, 193, 238, 29, 253, 193, 76, 30, 253, 193, 78, 30, 253, 193, 158, 30, 253, 193, 160, 30, 253, 193, 162, 30, 253, 193, 94, 31, 253, 193, 208, 26, 253, 193, 210, 26, 253, 193, 230, 26, 253, 193, 232, 26, 253, 193, 234, 26, 253, 193, 236, 26, 253, 193, 238, 26, 253, 193, 240, 26, 253, 193, 242, 26, 253, 193, 244, 26, 253, 193, 246, 26, 253, 193, 248, 26, 253, 193, 20, 27, 253, 193, 22, 27, 253, 193, 24, 27, 253, 193, 26, 27, 253, 193, 28, 27, 253, 193, 30, 27, 253, 193, 50, 27, 253, 193, 52, 27, 253, 193, 54, 27, 253, 193, 56, 27, 253, 193, 58, 27, 253, 193, 60, 27, 253, 193, 74, 27, 253, 193, 76, 27, 253, 193, 78, 27, 253, 193, 80, 27, 253, 193, 90, 27, 253, 193, 92, 27, 253, 193, 94, 27, 253, 193, 96, 27, 253, 193, 98, 27, 253, 193, 100, 27, 253, 193, 106, 27, 253, 193, 108, 27, 253, 193, 110, 27, 253, 193, 112, 27, 253, 193, 116, 27, 253, 193, 118, 27, 253, 193, 120, 27, 253, 193, 122, 27, 253, 193, 124, 27, 253, 193, 126, 27, 253, 193, 134, 27, 253, 193, 136, 27, 253, 193, 138, 27, 253, 193, 142, 27, 253, 193, 144, 27, 253, 193, 146, 27, 253, 193, 242, 27, 253, 193, 246, 27, 253, 193, 2, 28, 253, 193, 4, 28, 253, 193, 6, 28, 253, 193, 10, 28, 253, 193, 12, 28, 253, 193, 14, 28, 253, 193, 16, 28, 253, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 193, 250, 27, 253, 193, 18, 28, 253, 193, 20, 28, 253, 193, 80, 28, 253, 193, 82, 28, 253, 193, 84, 28, 253, 193, 86, 28, 253, 193, 88, 28, 253, 193, 90, 28, 253, 193, 134, 28, 253, 193, 136, 28, 253, 193, 188, 28, 253, 193, 190, 28, 253, 193, 192, 28, 253, 193, 194, 28, 253, 193, 196, 28, 253, 193, 234, 28, 253, 193, 236, 28, 253, 193, 238, 28, 253, 193, 56, 29, 253, 193, 154, 29, 253, 193, 156, 29, 253, 193, 216, 29, 253, 193, 80, 30, 253, 193, 82, 30, 253, 193, 164, 30, 253, 193, 166, 30, 253, 193, 98, 31, 253, 193, 100, 31, 253, 193, 102, 31, 253, 193, 104, 31, 253, 193, 66, 32, 253, 193, 8, 28, 253, 193, 22, 28, 253, 193, 138, 28, 253, 193, 140, 28, 253, 193, 240, 28, 253, 193, 242, 28, 253, 193, 218, 29, 253, 193, 52, 30, 253, 193, 84, 30, 253, 193, 86, 30, 253, 193, 106, 31, 253, 193, 84, 32, 253, 193, 86, 32, 253, 193, 72, 33, 253, 193, 74, 33, 253, 193, 76, 33, 253, 193, 78, 33, 253, 193, 80, 33, 253, 193, 82, 33, 253, 193, 84, 33, 253, 193, 86, 33, 253, 193, 88, 33, 253, 193, 90, 33, 253, 193, 92, 33, 253, 193, 94, 33, 253, 193, 64, 34, 253, 193, 66, 34, 253, 193, 68, 34, 253, 193, 96, 35, 253, 193, 98, 35, 253, 193, 84, 36, 253, 193, 86, 36, 253, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 206, 128, 88, 0, 193, 100, 28, 253, 193, 102, 28, 253, 193, 104, 28, 253, 193, 106, 28, 253, 193, 108, 28, 253, 193, 110, 28, 253, 193, 112, 28, 253, 193, 114, 28, 253, 193, 116, 28, 253, 193, 118, 28, 253, 193, 120, 28, 253, 193, 122, 28, 253, 193, 124, 28, 253, 193, 126, 28, 253, 193, 128, 28, 253, 193, 130, 28, 253, 193, 168, 28, 253, 193, 170, 28, 253, 193, 172, 28, 253, 193, 174, 28, 253, 193, 176, 28, 253, 193, 178, 28, 253, 193, 180, 28, 253, 193, 208, 28, 253, 193, 210, 28, 253, 193, 212, 28, 253, 193, 214, 28, 253, 193, 216, 28, 253, 193, 218, 28, 253, 193, 220, 28, 253, 193, 222, 28, 253, 193, 224, 28, 253, 193, 226, 28, 253, 193, 228, 28, 253, 193, 230, 28, 253, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 193, 146, 28, 253, 193, 148, 28, 253, 193, 150, 28, 253, 193, 152, 28, 253, 193, 154, 28, 253, 193, 156, 28, 253, 193, 158, 28, 253, 193, 160, 28, 253, 193, 198, 28, 253, 193, 200, 28, 253, 193, 202, 28, 253, 193, 204, 28, 253, 193, 244, 28, 253, 193, 246, 28, 253, 193, 248, 28, 253, 193, 250, 28, 253, 193, 58, 29, 253, 193, 60, 29, 253, 193, 108, 29, 253, 193, 110, 29, 253, 193, 158, 29, 253, 193, 186, 29, 253, 193, 188, 29, 253, 193, 220, 29, 253, 193, 222, 29, 253, 193, 224, 29, 253, 193, 240, 29, 253, 193, 250, 29, 253, 193, 252, 29, 253, 193, 10, 30, 253, 193, 12, 30, 253, 193, 22, 30, 253, 193, 6, 29, 253, 193, 62, 29, 253, 193, 64, 29, 253, 193, 66, 29, 253, 193, 68, 29, 253, 193, 112, 29, 253, 193, 114, 29, 253, 193, 162, 29, 253, 193, 190, 29, 253, 193, 192, 29, 253, 193, 226, 29, 253, 193, 228, 29, 253, 193, 194, 29, 253, 193, 24, 30, 253, 193, 68, 30, 253, 193, 70, 30, 253, 193, 72, 30, 253, 193, 124, 30, 253, 193, 126, 30, 253, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 206, 160, 88, 0, 193, 70, 29, 253, 193, 72, 29, 253, 193, 74, 29, 253, 193, 76, 29, 253, 193, 78, 29, 253, 193, 80, 29, 253, 193, 82, 29, 253, 193, 84, 29, 253, 193, 86, 29, 253, 193, 88, 29, 253, 193, 90, 29, 253, 193, 92, 29, 253, 193, 94, 29, 253, 193, 96, 29, 253, 193, 98, 29, 253, 193, 100, 29, 253, 193, 116, 29, 253, 193, 118, 29, 253, 193, 120, 29, 253, 193, 122, 29, 253, 193, 124, 29, 253, 193, 126, 29, 253, 193, 128, 29, 253, 193, 102, 29, 253, 193, 130, 29, 253, 193, 132, 29, 253, 193, 134, 29, 253, 193, 136, 29, 253, 193, 138, 29, 253, 193, 140, 29, 253, 193, 142, 29, 253, 193, 144, 29, 253, 193, 146, 29, 253, 193, 148, 29, 253, 193, 150, 29, 253, 193, 152, 29, 253, 193, 164, 29, 253, 193, 166, 29, 253, 193, 168, 29, 253, 193, 170, 29, 253, 193, 172, 29, 253, 193, 174, 29, 253, 193, 176, 29, 253, 193, 178, 29, 253, 193, 180, 29, 253, 193, 182, 29, 253, 193, 184, 29, 253, 206, 192, 88, 0, 206, 192, 88, 0, 206, 192, 88, 0, 206, 192, 88, 0, 206, 192, 88, 0, 206, 192, 88, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 193, 160, 29, 253, 193, 66, 199, 253, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 192, 88, 0, 206, 192, 88, 0, 206, 192, 88, 0, 206, 192, 88, 0, 193, 230, 29, 253, 193, 232, 29, 253, 193, 234, 29, 253, 193, 236, 29, 253, 193, 242, 29, 253, 193, 244, 29, 253, 193, 246, 29, 253, 193, 248, 29, 253, 193, 254, 29, 253, 193, 2, 30, 253, 193, 4, 30, 253, 193, 6, 30, 253, 193, 8, 30, 253, 193, 14, 30, 253, 193, 16, 30, 253, 193, 18, 30, 253, 193, 20, 30, 253, 193, 26, 30, 253, 193, 28, 30, 253, 193, 30, 30, 253, 193, 32, 30, 253, 193, 34, 30, 253, 193, 36, 30, 253, 193, 42, 30, 253, 193, 44, 30, 253, 193, 46, 30, 253, 193, 48, 30, 253, 193, 50, 30, 253, 193, 38, 30, 253, 193, 40, 30, 253, 193, 54, 30, 253, 193, 56, 30, 253, 193, 58, 30, 253, 193, 60, 30, 253, 193, 62, 30, 253, 193, 64, 30, 253, 193, 66, 30, 253, 193, 88, 30, 253, 193, 90, 30, 253, 193, 92, 30, 253, 193, 94, 30, 253, 193, 96, 30, 253, 193, 98, 30, 253, 193, 100, 30, 253, 193, 102, 30, 253, 193, 104, 30, 253, 193, 106, 30, 253, 193, 108, 30, 253, 193, 110, 30, 253, 193, 112, 30, 253, 193, 114, 30, 253, 193, 116, 30, 253, 193, 118, 30, 253, 193, 120, 30, 253, 193, 122, 30, 253, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 193, 74, 30, 253, 193, 128, 30, 253, 193, 130, 30, 253, 193, 132, 30, 253, 193, 134, 30, 253, 193, 136, 30, 253, 193, 138, 30, 253, 193, 140, 30, 253, 193, 142, 30, 253, 193, 144, 30, 253, 193, 146, 30, 253, 193, 148, 30, 253, 193, 150, 30, 253, 193, 152, 30, 253, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 206, 224, 88, 0, 193, 108, 31, 253, 193, 110, 31, 253, 193, 112, 31, 253, 193, 114, 31, 253, 193, 116, 31, 253, 193, 118, 31, 253, 193, 120, 31, 253, 193, 122, 31, 253, 193, 124, 31, 253, 193, 126, 31, 253, 193, 128, 31, 253, 193, 130, 31, 253, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 206, 0, 89, 0, 193, 198, 31, 253, 193, 200, 31, 253, 193, 202, 31, 253, 193, 204, 31, 253, 193, 206, 31, 253, 193, 208, 31, 253, 193, 210, 31, 253, 193, 212, 31, 253, 193, 214, 31, 253, 193, 216, 31, 253, 193, 218, 31, 253, 193, 220, 31, 253, 193, 222, 31, 253, 193, 224, 31, 253, 193, 226, 31, 253, 193, 228, 31, 253, 193, 176, 32, 253, 193, 178, 32, 253, 193, 180, 32, 253, 193, 182, 32, 253, 193, 184, 32, 253, 193, 186, 32, 253, 193, 188, 32, 253, 193, 190, 32, 253, 193, 192, 32, 253, 193, 166, 33, 253, 193, 168, 33, 253, 193, 170, 33, 253, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 206, 32, 89, 0, 193, 230, 31, 253, 193, 232, 31, 253, 193, 234, 31, 253, 193, 236, 31, 253, 193, 96, 31, 253, 193, 62, 32, 253, 193, 64, 32, 253, 193, 44, 33, 253, 193, 46, 33, 253, 193, 48, 33, 253, 193, 50, 33, 253, 193, 52, 33, 253, 193, 54, 33, 253, 193, 56, 33, 253, 193, 40, 34, 253, 193, 42, 34, 253, 193, 44, 34, 253, 193, 46, 34, 253, 193, 64, 35, 253, 193, 66, 35, 253, 193, 68, 35, 253, 193, 70, 35, 253, 193, 72, 35, 253, 193, 74, 35, 253, 193, 52, 36, 253, 193, 54, 36, 253, 193, 56, 36, 253, 193, 58, 36, 253, 193, 60, 36, 253, 193, 62, 36, 253, 206, 64, 90, 0, 206, 64, 90, 0, 206, 64, 90, 0, 206, 64, 90, 0, 206, 64, 90, 0, 206, 64, 90, 0, 193, 132, 31, 253, 193, 134, 31, 253, 193, 136, 31, 253, 193, 138, 31, 253, 193, 140, 31, 253, 193, 142, 31, 253, 193, 144, 31, 253, 193, 146, 31, 253, 193, 148, 31, 253, 193, 150, 31, 253, 193, 152, 31, 253, 193, 154, 31, 253, 193, 156, 31, 253, 193, 158, 31, 253, 193, 160, 31, 253, 193, 162, 31, 253, 193, 164, 31, 253, 193, 166, 31, 253, 193, 168, 31, 253, 193, 170, 31, 253, 193, 172, 31, 253, 193, 174, 31, 253, 193, 176, 31, 253, 193, 178, 31, 253, 193, 180, 31, 253, 193, 182, 31, 253, 193, 184, 31, 253, 193, 186, 31, 253, 193, 188, 31, 253, 193, 190, 31, 253, 193, 192, 31, 253, 193, 194, 31, 253, 193, 196, 31, 253, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 193, 238, 31, 253, 193, 240, 31, 253, 193, 242, 31, 253, 193, 244, 31, 253, 193, 246, 31, 253, 193, 248, 31, 253, 193, 250, 31, 253, 193, 252, 31, 253, 193, 254, 31, 253, 193, 2, 32, 253, 193, 4, 32, 253, 193, 6, 32, 253, 193, 8, 32, 253, 193, 10, 32, 253, 193, 12, 32, 253, 193, 14, 32, 253, 193, 16, 32, 253, 193, 18, 32, 253, 193, 20, 32, 253, 193, 22, 32, 253, 193, 24, 32, 253, 193, 26, 32, 253, 193, 28, 32, 253, 193, 30, 32, 253, 193, 32, 32, 253, 193, 34, 32, 253, 193, 36, 32, 253, 193, 38, 32, 253, 193, 40, 32, 253, 193, 42, 32, 253, 193, 44, 32, 253, 193, 46, 32, 253, 193, 48, 32, 253, 193, 50, 32, 253, 193, 52, 32, 253, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 193, 68, 32, 253, 193, 70, 32, 253, 193, 72, 32, 253, 193, 74, 32, 253, 193, 76, 32, 253, 193, 78, 32, 253, 193, 80, 32, 253, 193, 82, 32, 253, 193, 58, 33, 253, 193, 60, 33, 253, 193, 62, 33, 253, 193, 64, 33, 253, 193, 66, 33, 253, 193, 68, 33, 253, 193, 70, 33, 253, 193, 48, 34, 253, 193, 50, 34, 253, 193, 52, 34, 253, 193, 54, 34, 253, 193, 56, 34, 253, 193, 58, 34, 253, 193, 60, 34, 253, 193, 62, 34, 253, 193, 76, 35, 253, 193, 78, 35, 253, 193, 80, 35, 253, 193, 82, 35, 253, 193, 84, 35, 253, 193, 86, 35, 253, 193, 88, 35, 253, 193, 90, 35, 253, 193, 92, 35, 253, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 206, 64, 89, 0, 193, 96, 33, 253, 193, 98, 33, 253, 193, 100, 33, 253, 193, 102, 33, 253, 193, 104, 33, 253, 193, 106, 33, 253, 193, 108, 33, 253, 193, 110, 33, 253, 193, 112, 33, 253, 193, 114, 33, 253, 193, 116, 33, 253, 193, 118, 33, 253, 193, 120, 33, 253, 193, 122, 33, 253, 193, 124, 33, 253, 193, 126, 33, 253, 193, 128, 33, 253, 193, 130, 33, 253, 193, 132, 33, 253, 193, 134, 33, 253, 193, 136, 33, 253, 193, 172, 32, 253, 193, 138, 33, 253, 193, 140, 33, 253, 193, 142, 33, 253, 193, 144, 33, 253, 193, 146, 33, 253, 193, 148, 33, 253, 193, 150, 33, 253, 193, 152, 33, 253, 193, 154, 33, 253, 193, 156, 33, 253, 193, 158, 33, 253, 193, 160, 33, 253, 193, 162, 33, 253, 193, 164, 33, 253, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 96, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 206, 128, 89, 0, 193, 38, 33, 253, 193, 16, 34, 253, 193, 18, 34, 253, 193, 20, 34, 253, 193, 172, 33, 253, 193, 174, 33, 253, 193, 176, 33, 253, 193, 178, 33, 253, 193, 180, 33, 253, 193, 182, 33, 253, 193, 184, 33, 253, 193, 186, 33, 253, 193, 188, 33, 253, 193, 160, 34, 253, 193, 162, 34, 253, 193, 164, 34, 253, 193, 166, 34, 253, 193, 168, 34, 253, 193, 170, 34, 253, 193, 172, 34, 253, 193, 174, 34, 253, 193, 176, 34, 253, 193, 178, 34, 253, 193, 180, 34, 253, 193, 182, 34, 253, 193, 184, 34, 253, 193, 186, 34, 253, 193, 188, 34, 253, 193, 190, 34, 253, 193, 180, 35, 253, 193, 182, 35, 253, 193, 184, 35, 253, 193, 186, 35, 253, 193, 188, 35, 253, 193, 156, 36, 253, 193, 158, 36, 253, 193, 22, 34, 253, 193, 24, 34, 253, 193, 26, 34, 253, 193, 28, 34, 253, 193, 30, 34, 253, 193, 32, 34, 253, 193, 192, 34, 253, 193, 194, 34, 253, 193, 196, 34, 253, 193, 198, 34, 253, 193, 200, 34, 253, 193, 202, 34, 253, 193, 204, 34, 253, 193, 206, 34, 253, 193, 208, 34, 253, 193, 210, 34, 253, 193, 212, 34, 253, 193, 214, 34, 253, 193, 216, 34, 253, 193, 218, 34, 253, 193, 220, 34, 253, 193, 222, 34, 253, 193, 224, 34, 253, 193, 226, 34, 253, 193, 228, 34, 253, 193, 230, 34, 253, 193, 232, 34, 253, 193, 234, 34, 253, 193, 236, 34, 253, 193, 238, 34, 253, 193, 240, 34, 253, 193, 242, 34, 253, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 160, 89, 0, 206, 192, 89, 0, 206, 192, 89, 0, 206, 192, 89, 0, 206, 192, 89, 0, 206, 192, 89, 0, 206, 192, 89, 0, 206, 192, 89, 0, 206, 192, 89, 0, 206, 192, 89, 0, 206, 192, 89, 0, 193, 158, 34, 253, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 193, 244, 34, 253, 193, 246, 34, 253, 193, 248, 34, 253, 193, 250, 34, 253, 193, 252, 34, 253, 193, 254, 34, 253, 193, 2, 35, 253, 193, 4, 35, 253, 193, 6, 35, 253, 193, 8, 35, 253, 193, 10, 35, 253, 193, 12, 35, 253, 193, 14, 35, 253, 193, 16, 35, 253, 193, 18, 35, 253, 193, 20, 35, 253, 193, 22, 35, 253, 193, 24, 35, 253, 193, 26, 35, 253, 193, 28, 35, 253, 193, 30, 35, 253, 193, 32, 35, 253, 193, 34, 35, 253, 193, 36, 35, 253, 193, 38, 35, 253, 193, 40, 35, 253, 193, 42, 35, 253, 193, 44, 35, 253, 193, 46, 35, 253, 193, 48, 35, 253, 193, 50, 35, 253, 193, 52, 35, 253, 193, 54, 35, 253, 193, 56, 35, 253, 193, 58, 35, 253, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 193, 62, 35, 253, 193, 42, 36, 253, 193, 44, 36, 253, 193, 46, 36, 253, 193, 48, 36, 253, 193, 50, 36, 253, 193, 30, 37, 253, 193, 32, 37, 253, 193, 34, 37, 253, 193, 232, 37, 253, 193, 234, 37, 253, 193, 236, 37, 253, 193, 238, 37, 253, 193, 164, 38, 253, 193, 166, 38, 253, 193, 168, 38, 253, 193, 54, 39, 253, 193, 56, 39, 253, 193, 156, 39, 253, 193, 64, 40, 253, 193, 108, 40, 253, 193, 110, 40, 253, 193, 112, 40, 253, 193, 154, 40, 253, 193, 180, 40, 253, 193, 230, 40, 253, 193, 18, 41, 253, 193, 20, 41, 253, 193, 84, 41, 253, 193, 86, 41, 253, 193, 88, 41, 253, 193, 90, 41, 253, 193, 94, 35, 253, 193, 64, 36, 253, 193, 66, 36, 253, 193, 68, 36, 253, 193, 70, 36, 253, 193, 72, 36, 253, 193, 74, 36, 253, 193, 76, 36, 253, 193, 78, 36, 253, 193, 80, 36, 253, 193, 82, 36, 253, 193, 58, 37, 253, 193, 60, 37, 253, 193, 62, 37, 253, 193, 64, 37, 253, 193, 66, 37, 253, 193, 242, 37, 253, 193, 244, 37, 253, 193, 246, 37, 253, 193, 248, 37, 253, 193, 250, 37, 253, 193, 252, 37, 253, 193, 254, 37, 253, 193, 2, 38, 253, 193, 4, 38, 253, 193, 178, 38, 253, 193, 180, 38, 253, 193, 182, 38, 253, 193, 184, 38, 253, 193, 186, 38, 253, 193, 62, 39, 253, 193, 64, 39, 253, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 206, 224, 89, 0, 193, 94, 36, 253, 193, 96, 36, 253, 193, 98, 36, 253, 193, 100, 36, 253, 193, 102, 36, 253, 193, 104, 36, 253, 193, 106, 36, 253, 193, 108, 36, 253, 193, 110, 36, 253, 193, 112, 36, 253, 193, 114, 36, 253, 193, 116, 36, 253, 193, 118, 36, 253, 193, 120, 36, 253, 193, 122, 36, 253, 193, 124, 36, 253, 193, 126, 36, 253, 193, 128, 36, 253, 193, 130, 36, 253, 193, 132, 36, 253, 193, 134, 36, 253, 193, 136, 36, 253, 193, 138, 36, 253, 193, 140, 36, 253, 193, 142, 36, 253, 193, 144, 36, 253, 193, 146, 36, 253, 193, 148, 36, 253, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 0, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 193, 88, 36, 253, 193, 90, 36, 253, 193, 92, 36, 253, 193, 68, 37, 253, 193, 70, 37, 253, 193, 6, 38, 253, 193, 188, 38, 253, 193, 190, 38, 253, 193, 192, 38, 253, 193, 174, 39, 253, 193, 204, 40, 253, 193, 236, 40, 253, 193, 238, 40, 253, 193, 34, 41, 253, 193, 36, 41, 253, 193, 38, 41, 253, 193, 40, 41, 253, 193, 42, 41, 253, 193, 44, 41, 253, 193, 114, 41, 253, 193, 116, 41, 253, 193, 118, 41, 253, 193, 120, 41, 253, 193, 122, 41, 253, 193, 124, 41, 253, 193, 126, 41, 253, 193, 128, 41, 253, 193, 130, 41, 253, 193, 132, 41, 253, 193, 134, 41, 253, 193, 136, 41, 253, 193, 138, 41, 253, 193, 150, 36, 253, 193, 152, 36, 253, 193, 154, 36, 253, 193, 72, 37, 253, 193, 74, 37, 253, 193, 76, 37, 253, 193, 78, 37, 253, 193, 80, 37, 253, 193, 82, 37, 253, 193, 84, 37, 253, 193, 86, 37, 253, 193, 88, 37, 253, 193, 90, 37, 253, 193, 92, 37, 253, 193, 94, 37, 253, 193, 96, 37, 253, 193, 98, 37, 253, 193, 100, 37, 253, 193, 102, 37, 253, 193, 104, 37, 253, 193, 106, 37, 253, 193, 108, 37, 253, 193, 110, 37, 253, 193, 112, 37, 253, 193, 114, 37, 253, 193, 116, 37, 253, 193, 118, 37, 253, 193, 120, 37, 253, 193, 122, 37, 253, 193, 124, 37, 253, 193, 126, 37, 253, 193, 8, 38, 253, 193, 160, 36, 253, 193, 162, 36, 253, 193, 164, 36, 253, 193, 128, 37, 253, 193, 130, 37, 253, 193, 132, 37, 253, 193, 134, 37, 253, 193, 136, 37, 253, 193, 138, 37, 253, 193, 140, 37, 253, 193, 142, 37, 253, 193, 68, 38, 253, 193, 70, 38, 253, 193, 72, 38, 253, 193, 74, 38, 253, 193, 76, 38, 253, 193, 78, 38, 253, 193, 80, 38, 253, 193, 240, 38, 253, 193, 242, 38, 253, 193, 244, 38, 253, 193, 246, 38, 253, 193, 248, 38, 253, 193, 250, 38, 253, 193, 252, 38, 253, 193, 94, 39, 253, 193, 96, 39, 253, 193, 98, 39, 253, 193, 100, 39, 253, 193, 102, 39, 253, 193, 104, 39, 253, 193, 192, 39, 253, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 32, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 64, 90, 0, 206, 64, 90, 0, 206, 64, 90, 0, 206, 64, 90, 0, 206, 64, 90, 0, 193, 240, 37, 253, 193, 170, 38, 253, 193, 172, 38, 253, 193, 174, 38, 253, 193, 176, 38, 253, 193, 58, 39, 253, 193, 60, 39, 253, 193, 158, 39, 253, 193, 160, 39, 253, 193, 162, 39, 253, 193, 164, 39, 253, 193, 166, 39, 253, 193, 168, 39, 253, 193, 4, 40, 253, 193, 6, 40, 253, 193, 8, 40, 253, 193, 66, 40, 253, 193, 114, 40, 253, 193, 186, 40, 253, 193, 188, 40, 253, 193, 232, 40, 253, 193, 234, 40, 253, 193, 24, 41, 253, 193, 26, 41, 253, 193, 28, 41, 253, 193, 30, 41, 253, 193, 32, 41, 253, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 206, 96, 90, 0, 193, 82, 38, 253, 193, 84, 38, 253, 193, 10, 38, 253, 193, 12, 38, 253, 193, 14, 38, 253, 193, 16, 38, 253, 193, 18, 38, 253, 193, 20, 38, 253, 193, 22, 38, 253, 193, 24, 38, 253, 193, 26, 38, 253, 193, 28, 38, 253, 193, 30, 38, 253, 193, 32, 38, 253, 193, 34, 38, 253, 193, 194, 38, 253, 193, 36, 38, 253, 193, 38, 38, 253, 193, 40, 38, 253, 193, 42, 38, 253, 193, 44, 38, 253, 193, 46, 38, 253, 193, 48, 38, 253, 193, 50, 38, 253, 193, 52, 38, 253, 193, 54, 38, 253, 193, 56, 38, 253, 193, 58, 38, 253, 193, 60, 38, 253, 193, 62, 38, 253, 193, 64, 38, 253, 193, 196, 38, 253, 193, 198, 38, 253, 193, 200, 38, 253, 193, 86, 38, 253, 193, 88, 38, 253, 193, 90, 38, 253, 193, 92, 38, 253, 193, 94, 38, 253, 193, 96, 38, 253, 193, 98, 38, 253, 193, 100, 38, 253, 193, 102, 38, 253, 193, 104, 38, 253, 193, 106, 38, 253, 193, 108, 38, 253, 193, 110, 38, 253, 193, 112, 38, 253, 193, 114, 38, 253, 193, 116, 38, 253, 193, 118, 38, 253, 193, 120, 38, 253, 193, 122, 38, 253, 193, 124, 38, 253, 193, 126, 38, 253, 193, 128, 38, 253, 193, 130, 38, 253, 193, 132, 38, 253, 193, 134, 38, 253, 193, 136, 38, 253, 193, 138, 38, 253, 193, 140, 38, 253, 193, 142, 38, 253, 193, 144, 38, 253, 193, 146, 38, 253, 193, 148, 38, 253, 193, 150, 38, 253, 193, 152, 38, 253, 193, 154, 38, 253, 193, 156, 38, 253, 193, 158, 38, 253, 193, 160, 38, 253, 193, 162, 38, 253, 193, 254, 38, 253, 193, 2, 39, 253, 193, 4, 39, 253, 193, 6, 39, 253, 193, 8, 39, 253, 193, 10, 39, 253, 193, 12, 39, 253, 193, 14, 39, 253, 193, 16, 39, 253, 193, 18, 39, 253, 193, 20, 39, 253, 193, 22, 39, 253, 193, 24, 39, 253, 193, 26, 39, 253, 193, 28, 39, 253, 193, 30, 39, 253, 193, 32, 39, 253, 193, 34, 39, 253, 193, 36, 39, 253, 193, 38, 39, 253, 193, 40, 39, 253, 193, 42, 39, 253, 193, 44, 39, 253, 193, 46, 39, 253, 193, 48, 39, 253, 193, 202, 38, 253, 193, 204, 38, 253, 193, 206, 38, 253, 193, 208, 38, 253, 193, 210, 38, 253, 193, 212, 38, 253, 193, 214, 38, 253, 193, 216, 38, 253, 193, 218, 38, 253, 193, 220, 38, 253, 193, 222, 38, 253, 193, 224, 38, 253, 193, 226, 38, 253, 193, 228, 38, 253, 193, 230, 38, 253, 193, 232, 38, 253, 193, 234, 38, 253, 193, 236, 38, 253, 193, 238, 38, 253, 193, 72, 39, 253, 193, 74, 39, 253, 193, 76, 39, 253, 193, 78, 39, 253, 193, 80, 39, 253, 193, 82, 39, 253, 193, 84, 39, 253, 193, 86, 39, 253, 193, 88, 39, 253, 193, 90, 39, 253, 193, 92, 39, 253, 193, 176, 39, 253, 193, 178, 39, 253, 193, 50, 39, 253, 193, 52, 39, 253, 193, 106, 39, 253, 193, 108, 39, 253, 193, 110, 39, 253, 193, 112, 39, 253, 193, 114, 39, 253, 193, 116, 39, 253, 193, 118, 39, 253, 193, 120, 39, 253, 193, 122, 39, 253, 193, 124, 39, 253, 193, 126, 39, 253, 193, 128, 39, 253, 193, 130, 39, 253, 193, 132, 39, 253, 193, 134, 39, 253, 193, 198, 39, 253, 193, 136, 39, 253, 193, 138, 39, 253, 193, 140, 39, 253, 193, 142, 39, 253, 193, 144, 39, 253, 193, 146, 39, 253, 193, 148, 39, 253, 193, 150, 39, 253, 193, 152, 39, 253, 193, 154, 39, 253, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 193, 66, 39, 253, 193, 68, 39, 253, 193, 70, 39, 253, 193, 170, 39, 253, 193, 172, 39, 253, 193, 10, 40, 253, 193, 12, 40, 253, 193, 68, 40, 253, 193, 138, 40, 253, 193, 34, 42, 253, 193, 92, 43, 253, 193, 144, 43, 253, 193, 196, 43, 253, 193, 220, 43, 253, 193, 232, 43, 253, 193, 254, 43, 253, 193, 18, 44, 253, 193, 68, 44, 253, 193, 132, 44, 253, 193, 170, 44, 253, 193, 172, 44, 253, 193, 228, 44, 253, 193, 18, 45, 253, 193, 40, 45, 253, 193, 70, 45, 253, 193, 80, 45, 253, 193, 128, 45, 253, 193, 182, 45, 253, 193, 214, 45, 253, 193, 216, 45, 253, 193, 218, 45, 253, 193, 8, 46, 253, 193, 180, 39, 253, 193, 182, 39, 253, 193, 184, 39, 253, 193, 186, 39, 253, 193, 188, 39, 253, 193, 190, 39, 253, 193, 14, 40, 253, 193, 16, 40, 253, 193, 18, 40, 253, 193, 20, 40, 253, 193, 22, 40, 253, 193, 24, 40, 253, 193, 26, 40, 253, 193, 28, 40, 253, 193, 30, 40, 253, 193, 32, 40, 253, 193, 34, 40, 253, 193, 70, 40, 253, 193, 72, 40, 253, 193, 74, 40, 253, 193, 76, 40, 253, 193, 78, 40, 253, 193, 80, 40, 253, 193, 116, 40, 253, 193, 118, 40, 253, 193, 120, 40, 253, 193, 140, 40, 253, 193, 142, 40, 253, 193, 156, 40, 253, 193, 158, 40, 253, 193, 160, 40, 253, 193, 182, 40, 253, 193, 194, 39, 253, 193, 196, 39, 253, 193, 36, 40, 253, 193, 16, 41, 253, 193, 190, 41, 253, 193, 94, 42, 253, 193, 234, 42, 253, 193, 236, 42, 253, 193, 34, 43, 253, 193, 146, 43, 253, 193, 164, 43, 253, 193, 172, 43, 253, 193, 198, 43, 253, 193, 200, 43, 253, 193, 226, 43, 253, 193, 236, 43, 253, 193, 30, 44, 253, 193, 44, 44, 253, 193, 80, 44, 253, 193, 92, 44, 253, 193, 110, 44, 253, 193, 120, 44, 253, 193, 142, 44, 253, 193, 156, 44, 253, 193, 158, 44, 253, 193, 160, 44, 253, 193, 182, 44, 253, 193, 208, 44, 253, 193, 210, 44, 253, 193, 212, 44, 253, 193, 214, 44, 253, 193, 216, 44, 253, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 128, 90, 0, 206, 160, 90, 0, 206, 160, 90, 0, 206, 160, 90, 0, 206, 160, 90, 0, 206, 160, 90, 0, 206, 160, 90, 0, 206, 160, 90, 0, 206, 160, 90, 0, 206, 160, 90, 0, 193, 84, 40, 253, 193, 86, 40, 253, 193, 88, 40, 253, 193, 90, 40, 253, 193, 92, 40, 253, 193, 94, 40, 253, 193, 96, 40, 253, 193, 98, 40, 253, 193, 100, 40, 253, 193, 102, 40, 253, 193, 104, 40, 253, 193, 106, 40, 253, 193, 122, 40, 253, 193, 124, 40, 253, 193, 126, 40, 253, 193, 128, 40, 253, 193, 130, 40, 253, 193, 132, 40, 253, 193, 134, 40, 253, 193, 136, 40, 253, 193, 144, 40, 253, 193, 146, 40, 253, 193, 148, 40, 253, 193, 150, 40, 253, 193, 152, 40, 253, 193, 162, 40, 253, 193, 164, 40, 253, 193, 192, 40, 253, 193, 194, 40, 253, 193, 196, 40, 253, 193, 198, 40, 253, 193, 200, 40, 253, 193, 202, 40, 253, 193, 206, 40, 253, 193, 208, 40, 253, 193, 210, 40, 253, 193, 212, 40, 253, 193, 214, 40, 253, 193, 216, 40, 253, 193, 218, 40, 253, 193, 220, 40, 253, 193, 222, 40, 253, 193, 224, 40, 253, 193, 226, 40, 253, 193, 228, 40, 253, 193, 240, 40, 253, 193, 242, 40, 253, 193, 244, 40, 253, 193, 246, 40, 253, 193, 248, 40, 253, 193, 250, 40, 253, 193, 252, 40, 253, 193, 254, 40, 253, 193, 2, 41, 253, 193, 4, 41, 253, 193, 6, 41, 253, 193, 8, 41, 253, 193, 10, 41, 253, 193, 12, 41, 253, 193, 14, 41, 253, 193, 46, 41, 253, 193, 48, 41, 253, 193, 50, 41, 253, 193, 52, 41, 253, 193, 54, 41, 253, 193, 56, 41, 253, 193, 58, 41, 253, 193, 60, 41, 253, 193, 62, 41, 253, 193, 64, 41, 253, 193, 66, 41, 253, 193, 68, 41, 253, 193, 70, 41, 253, 193, 72, 41, 253, 193, 74, 41, 253, 193, 76, 41, 253, 193, 78, 41, 253, 193, 80, 41, 253, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 193, 92, 41, 253, 193, 94, 41, 253, 193, 192, 41, 253, 193, 194, 41, 253, 193, 196, 41, 253, 193, 198, 41, 253, 193, 200, 41, 253, 193, 8, 42, 253, 193, 10, 42, 253, 193, 12, 42, 253, 193, 14, 42, 253, 193, 96, 42, 253, 193, 98, 42, 253, 193, 100, 42, 253, 193, 102, 42, 253, 193, 186, 42, 253, 193, 188, 42, 253, 193, 190, 42, 253, 193, 192, 42, 253, 193, 194, 42, 253, 193, 196, 42, 253, 193, 238, 42, 253, 193, 240, 42, 253, 193, 242, 42, 253, 193, 244, 42, 253, 193, 36, 43, 253, 193, 38, 43, 253, 193, 70, 43, 253, 193, 72, 43, 253, 193, 110, 43, 253, 193, 124, 43, 253, 193, 140, 43, 253, 193, 98, 41, 253, 193, 100, 41, 253, 193, 102, 41, 253, 193, 104, 41, 253, 193, 106, 41, 253, 193, 108, 41, 253, 193, 110, 41, 253, 193, 112, 41, 253, 193, 202, 41, 253, 193, 204, 41, 253, 193, 206, 41, 253, 193, 208, 41, 253, 193, 210, 41, 253, 193, 212, 41, 253, 193, 214, 41, 253, 193, 216, 41, 253, 193, 218, 41, 253, 193, 220, 41, 253, 193, 222, 41, 253, 193, 224, 41, 253, 193, 16, 42, 253, 193, 18, 42, 253, 193, 20, 42, 253, 193, 22, 42, 253, 193, 24, 42, 253, 193, 26, 42, 253, 193, 28, 42, 253, 193, 30, 42, 253, 193, 32, 42, 253, 193, 108, 42, 253, 193, 110, 42, 253, 193, 112, 42, 253, 193, 140, 41, 253, 193, 142, 41, 253, 193, 226, 41, 253, 193, 228, 41, 253, 193, 230, 41, 253, 193, 232, 41, 253, 193, 234, 41, 253, 193, 36, 42, 253, 193, 38, 42, 253, 193, 40, 42, 253, 193, 42, 42, 253, 193, 44, 42, 253, 193, 46, 42, 253, 193, 48, 42, 253, 193, 50, 42, 253, 193, 52, 42, 253, 193, 128, 42, 253, 193, 130, 42, 253, 193, 132, 42, 253, 193, 134, 42, 253, 193, 136, 42, 253, 193, 138, 42, 253, 193, 140, 42, 253, 193, 142, 42, 253, 193, 144, 42, 253, 193, 10, 43, 253, 193, 12, 43, 253, 193, 14, 43, 253, 193, 16, 43, 253, 193, 18, 43, 253, 193, 20, 43, 253, 193, 22, 43, 253, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 206, 192, 90, 0, 193, 236, 41, 253, 193, 238, 41, 253, 193, 240, 41, 253, 193, 242, 41, 253, 193, 244, 41, 253, 193, 246, 41, 253, 193, 248, 41, 253, 193, 250, 41, 253, 193, 252, 41, 253, 193, 254, 41, 253, 193, 2, 42, 253, 193, 4, 42, 253, 193, 6, 42, 253, 206, 224, 90, 0, 206, 224, 90, 0, 206, 224, 90, 0, 206, 224, 90, 0, 206, 224, 90, 0, 206, 224, 90, 0, 206, 224, 90, 0, 206, 224, 90, 0, 206, 224, 90, 0, 206, 224, 90, 0, 206, 224, 90, 0, 206, 224, 90, 0, 193, 146, 42, 253, 193, 148, 42, 253, 193, 150, 42, 253, 193, 152, 42, 253, 193, 154, 42, 253, 193, 156, 42, 253, 193, 158, 42, 253, 193, 160, 42, 253, 193, 162, 42, 253, 193, 164, 42, 253, 193, 166, 42, 253, 193, 168, 42, 253, 193, 170, 42, 253, 193, 172, 42, 253, 193, 174, 42, 253, 193, 176, 42, 253, 193, 178, 42, 253, 193, 180, 42, 253, 193, 182, 42, 253, 193, 184, 42, 253, 193, 216, 42, 253, 193, 104, 42, 253, 193, 106, 42, 253, 193, 246, 42, 253, 193, 222, 66, 253, 193, 204, 76, 253, 193, 116, 77, 253, 193, 160, 77, 253, 193, 206, 78, 253, 193, 62, 83, 253, 193, 64, 83, 253, 193, 224, 85, 253, 193, 132, 88, 253, 193, 64, 90, 253, 193, 208, 90, 253, 193, 212, 94, 253, 193, 104, 98, 253, 193, 38, 102, 253, 193, 12, 108, 253, 193, 104, 114, 253, 193, 138, 114, 253, 193, 230, 114, 253, 193, 232, 114, 253, 193, 108, 115, 253, 193, 110, 115, 253, 193, 4, 116, 253, 193, 6, 116, 253, 193, 8, 116, 253, 193, 10, 116, 253, 193, 124, 116, 253, 193, 250, 116, 253, 193, 94, 117, 253, 193, 96, 117, 253, 193, 114, 42, 253, 193, 116, 42, 253, 193, 118, 42, 253, 193, 120, 42, 253, 193, 122, 42, 253, 193, 124, 42, 253, 193, 126, 42, 253, 193, 198, 42, 253, 193, 200, 42, 253, 193, 202, 42, 253, 193, 204, 42, 253, 193, 206, 42, 253, 193, 208, 42, 253, 193, 210, 42, 253, 193, 212, 42, 253, 193, 214, 42, 253, 193, 248, 42, 253, 193, 250, 42, 253, 193, 252, 42, 253, 193, 254, 42, 253, 193, 2, 43, 253, 193, 4, 43, 253, 193, 6, 43, 253, 193, 8, 43, 253, 193, 40, 43, 253, 193, 42, 43, 253, 193, 44, 43, 253, 193, 46, 43, 253, 193, 48, 43, 253, 193, 74, 43, 253, 193, 76, 43, 253, 193, 78, 43, 253, 193, 218, 42, 253, 193, 220, 42, 253, 193, 222, 42, 253, 193, 224, 42, 253, 193, 226, 42, 253, 193, 228, 42, 253, 193, 230, 42, 253, 193, 232, 42, 253, 193, 24, 43, 253, 193, 26, 43, 253, 193, 28, 43, 253, 193, 30, 43, 253, 193, 32, 43, 253, 193, 58, 43, 253, 193, 60, 43, 253, 193, 62, 43, 253, 193, 64, 43, 253, 193, 66, 43, 253, 193, 68, 43, 253, 193, 104, 43, 253, 193, 106, 43, 253, 193, 108, 43, 253, 193, 150, 43, 253, 193, 160, 43, 253, 193, 170, 43, 253, 193, 180, 43, 253, 193, 182, 43, 253, 193, 184, 43, 253, 193, 212, 43, 253, 193, 224, 43, 253, 193, 234, 43, 253, 193, 24, 44, 253, 193, 50, 43, 253, 193, 52, 43, 253, 193, 54, 43, 253, 193, 56, 43, 253, 193, 94, 43, 253, 193, 96, 43, 253, 193, 98, 43, 253, 193, 100, 43, 253, 193, 102, 43, 253, 193, 118, 43, 253, 193, 120, 43, 253, 193, 122, 43, 253, 193, 134, 43, 253, 193, 136, 43, 253, 193, 138, 43, 253, 193, 154, 43, 253, 193, 168, 43, 253, 193, 208, 43, 253, 193, 210, 43, 253, 193, 222, 43, 253, 193, 2, 44, 253, 193, 20, 44, 253, 193, 22, 44, 253, 193, 94, 44, 253, 193, 118, 44, 253, 193, 134, 44, 253, 193, 136, 44, 253, 193, 152, 44, 253, 193, 174, 44, 253, 193, 206, 44, 253, 193, 254, 44, 253, 193, 20, 45, 253, 193, 80, 43, 253, 193, 82, 43, 253, 193, 84, 43, 253, 193, 86, 43, 253, 193, 88, 43, 253, 193, 90, 43, 253, 193, 112, 43, 253, 193, 114, 43, 253, 193, 116, 43, 253, 193, 126, 43, 253, 193, 128, 43, 253, 193, 130, 43, 253, 193, 132, 43, 253, 193, 142, 43, 253, 193, 148, 43, 253, 193, 152, 43, 253, 193, 156, 43, 253, 193, 158, 43, 253, 193, 206, 43, 253, 193, 252, 43, 253, 193, 16, 44, 253, 193, 40, 44, 253, 193, 116, 44, 253, 193, 150, 44, 253, 193, 38, 45, 253, 193, 60, 45, 253, 193, 88, 45, 253, 193, 126, 45, 253, 193, 178, 45, 253, 193, 180, 45, 253, 193, 46, 46, 253, 193, 48, 46, 253, 193, 166, 43, 253, 193, 194, 43, 253, 193, 250, 43, 253, 193, 204, 44, 253, 193, 16, 45, 253, 193, 122, 45, 253, 193, 176, 45, 253, 193, 42, 46, 253, 193, 44, 46, 253, 193, 134, 46, 253, 193, 54, 47, 253, 193, 182, 47, 253, 193, 48, 48, 253, 193, 84, 48, 253, 193, 210, 48, 253, 193, 244, 48, 253, 193, 246, 48, 253, 193, 248, 48, 253, 193, 74, 49, 253, 193, 76, 49, 253, 193, 222, 49, 253, 193, 224, 49, 253, 193, 100, 50, 253, 193, 196, 50, 253, 193, 198, 50, 253, 193, 200, 50, 253, 193, 76, 51, 253, 193, 78, 51, 253, 193, 198, 51, 253, 193, 200, 51, 253, 193, 202, 51, 253, 193, 24, 52, 253, 193, 4, 44, 253, 193, 6, 44, 253, 193, 8, 44, 253, 193, 10, 44, 253, 193, 12, 44, 253, 193, 14, 44, 253, 193, 32, 44, 253, 193, 34, 44, 253, 193, 36, 44, 253, 193, 38, 44, 253, 193, 46, 44, 253, 193, 52, 44, 253, 193, 54, 44, 253, 193, 56, 44, 253, 193, 58, 44, 253, 193, 60, 44, 253, 193, 62, 44, 253, 193, 64, 44, 253, 193, 66, 44, 253, 193, 70, 44, 253, 193, 72, 44, 253, 193, 74, 44, 253, 193, 76, 44, 253, 193, 84, 44, 253, 193, 86, 44, 253, 193, 88, 44, 253, 193, 96, 44, 253, 193, 102, 44, 253, 193, 104, 44, 253, 193, 106, 44, 253, 193, 112, 44, 253, 193, 114, 44, 253, 193, 26, 44, 253, 193, 28, 44, 253, 193, 42, 44, 253, 193, 48, 44, 253, 193, 78, 44, 253, 193, 82, 44, 253, 193, 90, 44, 253, 193, 98, 44, 253, 193, 100, 44, 253, 193, 108, 44, 253, 193, 138, 44, 253, 193, 140, 44, 253, 193, 176, 44, 253, 193, 178, 44, 253, 193, 180, 44, 253, 193, 230, 44, 253, 193, 2, 45, 253, 193, 46, 45, 253, 193, 94, 45, 253, 193, 102, 45, 253, 193, 108, 45, 253, 193, 112, 45, 253, 193, 118, 45, 253, 193, 134, 45, 253, 193, 146, 45, 253, 193, 148, 45, 253, 193, 150, 45, 253, 193, 152, 45, 253, 193, 154, 45, 253, 193, 156, 45, 253, 193, 184, 45, 253, 193, 186, 45, 253, 193, 232, 44, 253, 193, 234, 44, 253, 193, 236, 44, 253, 193, 64, 45, 253, 193, 96, 45, 253, 193, 136, 45, 253, 193, 158, 45, 253, 193, 160, 45, 253, 193, 192, 45, 253, 193, 232, 45, 253, 193, 52, 46, 253, 193, 54, 46, 253, 193, 56, 46, 253, 193, 110, 46, 253, 193, 112, 46, 253, 193, 188, 46, 253, 193, 208, 46, 253, 193, 210, 46, 253, 193, 238, 46, 253, 193, 250, 46, 253, 193, 18, 47, 253, 193, 20, 47, 253, 193, 60, 47, 253, 193, 62, 47, 253, 193, 64, 47, 253, 193, 66, 47, 253, 193, 100, 47, 253, 193, 102, 47, 253, 193, 192, 47, 253, 193, 230, 47, 253, 193, 232, 47, 253, 193, 18, 48, 253, 193, 246, 44, 253, 193, 248, 44, 253, 193, 250, 44, 253, 193, 6, 45, 253, 193, 8, 45, 253, 193, 10, 45, 253, 193, 12, 45, 253, 193, 14, 45, 253, 193, 252, 44, 253, 193, 22, 45, 253, 193, 24, 45, 253, 193, 26, 45, 253, 193, 28, 45, 253, 193, 30, 45, 253, 193, 32, 45, 253, 193, 34, 45, 253, 193, 36, 45, 253, 193, 48, 45, 253, 193, 50, 45, 253, 193, 52, 45, 253, 193, 54, 45, 253, 193, 56, 45, 253, 193, 58, 45, 253, 193, 66, 45, 253, 193, 68, 45, 253, 193, 72, 45, 253, 193, 74, 45, 253, 193, 76, 45, 253, 193, 78, 45, 253, 193, 82, 45, 253, 193, 84, 45, 253, 193, 86, 45, 253, 193, 42, 45, 253, 193, 44, 45, 253, 193, 62, 45, 253, 193, 90, 45, 253, 193, 130, 45, 253, 193, 132, 45, 253, 193, 220, 45, 253, 193, 96, 46, 253, 193, 98, 46, 253, 193, 100, 46, 253, 193, 142, 46, 253, 193, 144, 46, 253, 193, 146, 46, 253, 193, 148, 46, 253, 193, 182, 46, 253, 193, 236, 46, 253, 193, 32, 47, 253, 193, 216, 48, 253, 193, 4, 49, 253, 193, 100, 49, 253, 193, 102, 49, 253, 193, 104, 49, 253, 193, 106, 49, 253, 193, 108, 49, 253, 193, 110, 49, 253, 193, 112, 49, 253, 193, 114, 49, 253, 193, 116, 49, 253, 193, 244, 49, 253, 193, 246, 49, 253, 193, 248, 49, 253, 193, 250, 49, 253, 193, 92, 45, 253, 193, 98, 45, 253, 193, 100, 45, 253, 193, 104, 45, 253, 193, 106, 45, 253, 193, 110, 45, 253, 193, 120, 45, 253, 193, 114, 45, 253, 193, 116, 45, 253, 193, 124, 45, 253, 193, 138, 45, 253, 193, 140, 45, 253, 193, 142, 45, 253, 193, 144, 45, 253, 193, 162, 45, 253, 193, 164, 45, 253, 193, 166, 45, 253, 193, 168, 45, 253, 193, 170, 45, 253, 193, 172, 45, 253, 193, 174, 45, 253, 193, 194, 45, 253, 193, 196, 45, 253, 193, 198, 45, 253, 193, 200, 45, 253, 193, 202, 45, 253, 193, 204, 45, 253, 193, 206, 45, 253, 193, 208, 45, 253, 193, 210, 45, 253, 193, 212, 45, 253, 193, 234, 45, 253, 193, 188, 45, 253, 193, 190, 45, 253, 193, 222, 45, 253, 193, 224, 45, 253, 193, 226, 45, 253, 193, 228, 45, 253, 193, 230, 45, 253, 193, 12, 46, 253, 193, 14, 46, 253, 193, 16, 46, 253, 193, 18, 46, 253, 193, 102, 46, 253, 193, 104, 46, 253, 193, 106, 46, 253, 193, 108, 46, 253, 193, 150, 46, 253, 193, 152, 46, 253, 193, 154, 46, 253, 193, 156, 46, 253, 193, 158, 46, 253, 193, 184, 46, 253, 193, 186, 46, 253, 193, 222, 46, 253, 193, 206, 46, 253, 193, 36, 47, 253, 193, 44, 47, 253, 193, 46, 47, 253, 193, 48, 47, 253, 193, 56, 47, 253, 193, 58, 47, 253, 193, 88, 47, 253, 193, 90, 47, 253, 193, 236, 45, 253, 193, 238, 45, 253, 193, 240, 45, 253, 193, 242, 45, 253, 193, 244, 45, 253, 193, 246, 45, 253, 193, 248, 45, 253, 193, 250, 45, 253, 193, 252, 45, 253, 193, 254, 45, 253, 193, 2, 46, 253, 193, 4, 46, 253, 193, 6, 46, 253, 193, 20, 46, 253, 193, 22, 46, 253, 193, 24, 46, 253, 193, 26, 46, 253, 193, 28, 46, 253, 193, 30, 46, 253, 193, 32, 46, 253, 193, 34, 46, 253, 193, 36, 46, 253, 193, 38, 46, 253, 193, 40, 46, 253, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 193, 10, 46, 253, 193, 136, 46, 253, 193, 138, 46, 253, 193, 140, 46, 253, 193, 204, 46, 253, 193, 42, 47, 253, 193, 84, 47, 253, 193, 86, 47, 253, 193, 140, 47, 253, 193, 142, 47, 253, 193, 224, 47, 253, 193, 50, 48, 253, 193, 152, 48, 253, 193, 214, 48, 253, 193, 250, 48, 253, 193, 252, 48, 253, 193, 254, 48, 253, 193, 2, 49, 253, 193, 86, 49, 253, 193, 88, 49, 253, 193, 90, 49, 253, 193, 92, 49, 253, 193, 94, 49, 253, 193, 96, 49, 253, 193, 98, 49, 253, 193, 228, 49, 253, 193, 230, 49, 253, 193, 232, 49, 253, 193, 234, 49, 253, 193, 236, 49, 253, 193, 238, 49, 253, 193, 240, 49, 253, 193, 50, 46, 253, 193, 90, 46, 253, 193, 92, 46, 253, 193, 94, 46, 253, 193, 178, 46, 253, 193, 180, 46, 253, 193, 234, 46, 253, 193, 248, 46, 253, 193, 184, 47, 253, 193, 8, 48, 253, 193, 116, 48, 253, 193, 118, 48, 253, 193, 150, 48, 253, 193, 176, 48, 253, 193, 212, 48, 253, 193, 78, 49, 253, 193, 80, 49, 253, 193, 82, 49, 253, 193, 84, 49, 253, 193, 226, 49, 253, 193, 102, 50, 253, 193, 104, 50, 253, 193, 202, 50, 253, 193, 204, 50, 253, 193, 206, 50, 253, 193, 208, 50, 253, 193, 210, 50, 253, 193, 80, 51, 253, 193, 82, 51, 253, 193, 84, 51, 253, 193, 86, 51, 253, 193, 88, 51, 253, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 206, 0, 91, 0, 193, 114, 46, 253, 193, 116, 46, 253, 193, 118, 46, 253, 193, 120, 46, 253, 193, 122, 46, 253, 193, 124, 46, 253, 193, 126, 46, 253, 193, 128, 46, 253, 193, 130, 46, 253, 193, 132, 46, 253, 193, 160, 46, 253, 193, 162, 46, 253, 193, 164, 46, 253, 193, 166, 46, 253, 193, 168, 46, 253, 193, 170, 46, 253, 193, 172, 46, 253, 193, 174, 46, 253, 193, 176, 46, 253, 193, 190, 46, 253, 193, 192, 46, 253, 193, 194, 46, 253, 193, 196, 46, 253, 193, 198, 46, 253, 193, 200, 46, 253, 193, 202, 46, 253, 193, 212, 46, 253, 193, 214, 46, 253, 193, 216, 46, 253, 193, 218, 46, 253, 193, 220, 46, 253, 193, 224, 46, 253, 193, 226, 46, 253, 193, 228, 46, 253, 193, 230, 46, 253, 193, 232, 46, 253, 193, 240, 46, 253, 193, 242, 46, 253, 193, 244, 46, 253, 193, 246, 46, 253, 193, 252, 46, 253, 193, 254, 46, 253, 193, 2, 47, 253, 193, 4, 47, 253, 193, 8, 47, 253, 193, 10, 47, 253, 193, 12, 47, 253, 193, 6, 47, 253, 193, 14, 47, 253, 193, 16, 47, 253, 193, 22, 47, 253, 193, 24, 47, 253, 193, 26, 47, 253, 193, 28, 47, 253, 193, 30, 47, 253, 193, 34, 47, 253, 193, 38, 47, 253, 193, 40, 47, 253, 193, 50, 47, 253, 193, 52, 47, 253, 193, 68, 47, 253, 193, 70, 47, 253, 193, 72, 47, 253, 193, 74, 47, 253, 193, 76, 47, 253, 193, 78, 47, 253, 193, 80, 47, 253, 193, 82, 47, 253, 193, 104, 47, 253, 193, 106, 47, 253, 193, 108, 47, 253, 193, 110, 47, 253, 193, 112, 47, 253, 193, 114, 47, 253, 193, 116, 47, 253, 193, 118, 47, 253, 193, 120, 47, 253, 193, 122, 47, 253, 193, 124, 47, 253, 193, 126, 47, 253, 193, 128, 47, 253, 193, 130, 47, 253, 193, 132, 47, 253, 193, 134, 47, 253, 193, 136, 47, 253, 193, 138, 47, 253, 193, 158, 47, 253, 193, 160, 47, 253, 193, 92, 47, 253, 193, 94, 47, 253, 193, 96, 47, 253, 193, 98, 47, 253, 193, 144, 47, 253, 193, 146, 47, 253, 193, 148, 47, 253, 193, 150, 47, 253, 193, 152, 47, 253, 193, 154, 47, 253, 193, 156, 47, 253, 193, 186, 47, 253, 193, 188, 47, 253, 193, 190, 47, 253, 193, 226, 47, 253, 193, 228, 47, 253, 193, 10, 48, 253, 193, 12, 48, 253, 193, 14, 48, 253, 193, 16, 48, 253, 193, 52, 48, 253, 193, 54, 48, 253, 193, 56, 48, 253, 193, 86, 48, 253, 193, 88, 48, 253, 193, 90, 48, 253, 193, 120, 48, 253, 193, 184, 48, 253, 193, 186, 48, 253, 193, 190, 48, 253, 193, 192, 48, 253, 193, 194, 48, 253, 193, 162, 47, 253, 193, 164, 47, 253, 193, 166, 47, 253, 193, 168, 47, 253, 193, 170, 47, 253, 193, 172, 47, 253, 193, 174, 47, 253, 193, 176, 47, 253, 193, 178, 47, 253, 193, 180, 47, 253, 193, 194, 47, 253, 193, 196, 47, 253, 193, 198, 47, 253, 193, 200, 47, 253, 193, 202, 47, 253, 193, 204, 47, 253, 193, 206, 47, 253, 193, 208, 47, 253, 193, 210, 47, 253, 193, 212, 47, 253, 193, 214, 47, 253, 193, 216, 47, 253, 193, 218, 47, 253, 193, 220, 47, 253, 193, 222, 47, 253, 206, 32, 91, 0, 206, 32, 91, 0, 206, 32, 91, 0, 206, 32, 91, 0, 206, 32, 91, 0, 206, 32, 91, 0, 206, 32, 91, 0, 206, 32, 91, 0, 206, 32, 91, 0, 193, 20, 48, 253, 193, 22, 48, 253, 193, 24, 48, 253, 193, 26, 48, 253, 193, 28, 48, 253, 193, 30, 48, 253, 193, 32, 48, 253, 193, 34, 48, 253, 193, 36, 48, 253, 193, 38, 48, 253, 193, 40, 48, 253, 193, 6, 48, 253, 193, 42, 48, 253, 193, 44, 48, 253, 193, 46, 48, 253, 193, 60, 48, 253, 193, 62, 48, 253, 193, 64, 48, 253, 193, 66, 48, 253, 193, 68, 48, 253, 193, 70, 48, 253, 193, 72, 48, 253, 193, 74, 48, 253, 193, 76, 48, 253, 193, 78, 48, 253, 193, 80, 48, 253, 193, 58, 48, 253, 193, 92, 48, 253, 193, 94, 48, 253, 193, 122, 48, 253, 193, 124, 48, 253, 193, 126, 48, 253, 193, 128, 48, 253, 193, 26, 49, 253, 193, 28, 49, 253, 193, 156, 49, 253, 193, 158, 49, 253, 193, 160, 49, 253, 193, 162, 49, 253, 193, 164, 49, 253, 193, 166, 49, 253, 193, 168, 49, 253, 193, 24, 50, 253, 193, 26, 50, 253, 193, 28, 50, 253, 193, 146, 50, 253, 193, 148, 50, 253, 193, 150, 50, 253, 193, 12, 51, 253, 193, 14, 51, 253, 193, 16, 51, 253, 193, 18, 51, 253, 193, 20, 51, 253, 193, 22, 51, 253, 193, 24, 51, 253, 193, 26, 51, 253, 193, 28, 51, 253, 193, 30, 51, 253, 193, 82, 48, 253, 193, 96, 48, 253, 193, 98, 48, 253, 193, 100, 48, 253, 193, 102, 48, 253, 193, 104, 48, 253, 193, 130, 48, 253, 193, 106, 48, 253, 193, 108, 48, 253, 193, 110, 48, 253, 193, 112, 48, 253, 193, 114, 48, 253, 193, 132, 48, 253, 193, 134, 48, 253, 193, 136, 48, 253, 193, 138, 48, 253, 193, 154, 48, 253, 193, 140, 48, 253, 193, 142, 48, 253, 193, 144, 48, 253, 193, 146, 48, 253, 193, 148, 48, 253, 193, 156, 48, 253, 193, 158, 48, 253, 193, 160, 48, 253, 193, 162, 48, 253, 193, 164, 48, 253, 193, 166, 48, 253, 193, 168, 48, 253, 193, 170, 48, 253, 193, 172, 48, 253, 193, 174, 48, 253, 193, 178, 48, 253, 193, 180, 48, 253, 193, 182, 48, 253, 193, 188, 48, 253, 193, 198, 48, 253, 193, 200, 48, 253, 193, 202, 48, 253, 193, 204, 48, 253, 193, 206, 48, 253, 193, 208, 48, 253, 193, 226, 48, 253, 193, 228, 48, 253, 193, 230, 48, 253, 193, 232, 48, 253, 193, 234, 48, 253, 193, 236, 48, 253, 193, 238, 48, 253, 193, 240, 48, 253, 193, 242, 48, 253, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 193, 196, 48, 253, 193, 218, 48, 253, 193, 220, 48, 253, 193, 222, 48, 253, 193, 224, 48, 253, 193, 6, 49, 253, 193, 8, 49, 253, 193, 10, 49, 253, 193, 12, 49, 253, 193, 14, 49, 253, 193, 16, 49, 253, 193, 18, 49, 253, 193, 20, 49, 253, 193, 22, 49, 253, 193, 24, 49, 253, 193, 118, 49, 253, 193, 120, 49, 253, 193, 122, 49, 253, 193, 124, 49, 253, 193, 126, 49, 253, 193, 128, 49, 253, 193, 130, 49, 253, 193, 132, 49, 253, 193, 134, 49, 253, 193, 136, 49, 253, 193, 138, 49, 253, 193, 140, 49, 253, 193, 142, 49, 253, 193, 144, 49, 253, 193, 146, 49, 253, 193, 148, 49, 253, 193, 150, 49, 253, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 206, 64, 91, 0, 193, 170, 49, 253, 193, 172, 49, 253, 193, 174, 49, 253, 193, 176, 49, 253, 193, 178, 49, 253, 193, 180, 49, 253, 193, 182, 49, 253, 193, 184, 49, 253, 193, 186, 49, 253, 193, 188, 49, 253, 193, 190, 49, 253, 193, 192, 49, 253, 193, 194, 49, 253, 193, 196, 49, 253, 193, 198, 49, 253, 193, 200, 49, 253, 193, 202, 49, 253, 193, 204, 49, 253, 193, 206, 49, 253, 193, 208, 49, 253, 193, 210, 49, 253, 193, 212, 49, 253, 193, 214, 49, 253, 193, 152, 49, 253, 193, 154, 49, 253, 193, 252, 49, 253, 193, 254, 49, 253, 193, 2, 50, 253, 193, 4, 50, 253, 193, 6, 50, 253, 193, 8, 50, 253, 193, 10, 50, 253, 193, 12, 50, 253, 193, 14, 50, 253, 193, 16, 50, 253, 193, 18, 50, 253, 193, 20, 50, 253, 193, 22, 50, 253, 193, 124, 50, 253, 193, 126, 50, 253, 193, 128, 50, 253, 193, 130, 50, 253, 193, 132, 50, 253, 193, 134, 50, 253, 193, 136, 50, 253, 193, 138, 50, 253, 193, 140, 50, 253, 193, 142, 50, 253, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 193, 216, 49, 253, 193, 218, 49, 253, 193, 220, 49, 253, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 193, 242, 49, 253, 193, 106, 50, 253, 193, 108, 50, 253, 193, 110, 50, 253, 193, 112, 50, 253, 193, 212, 50, 253, 193, 214, 50, 253, 193, 216, 50, 253, 193, 218, 50, 253, 193, 220, 50, 253, 193, 222, 50, 253, 193, 90, 51, 253, 193, 92, 51, 253, 193, 94, 51, 253, 193, 96, 51, 253, 193, 98, 51, 253, 193, 100, 51, 253, 193, 102, 51, 253, 193, 104, 51, 253, 193, 106, 51, 253, 193, 108, 51, 253, 193, 210, 51, 253, 193, 212, 51, 253, 193, 214, 51, 253, 193, 216, 51, 253, 193, 218, 51, 253, 193, 34, 52, 253, 193, 36, 52, 253, 193, 38, 52, 253, 193, 40, 52, 253, 193, 130, 52, 253, 193, 132, 52, 253, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 206, 96, 91, 0, 193, 152, 50, 253, 193, 154, 50, 253, 193, 156, 50, 253, 193, 158, 50, 253, 193, 160, 50, 253, 193, 162, 50, 253, 193, 164, 50, 253, 193, 166, 50, 253, 193, 168, 50, 253, 193, 170, 50, 253, 193, 172, 50, 253, 193, 174, 50, 253, 193, 176, 50, 253, 193, 178, 50, 253, 193, 180, 50, 253, 193, 182, 50, 253, 193, 184, 50, 253, 193, 186, 50, 253, 193, 188, 50, 253, 193, 190, 50, 253, 193, 192, 50, 253, 193, 194, 50, 253, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 193, 114, 50, 253, 193, 116, 50, 253, 193, 118, 50, 253, 193, 120, 50, 253, 193, 122, 50, 253, 193, 224, 50, 253, 193, 110, 51, 253, 193, 112, 51, 253, 193, 114, 51, 253, 193, 220, 51, 253, 193, 42, 52, 253, 193, 44, 52, 253, 193, 46, 52, 253, 193, 136, 52, 253, 193, 138, 52, 253, 193, 140, 52, 253, 193, 142, 52, 253, 193, 228, 52, 253, 193, 230, 52, 253, 193, 232, 52, 253, 193, 234, 52, 253, 193, 236, 52, 253, 193, 238, 52, 253, 193, 240, 52, 253, 193, 80, 53, 253, 193, 82, 53, 253, 193, 84, 53, 253, 193, 134, 53, 253, 193, 184, 53, 253, 193, 186, 53, 253, 193, 214, 53, 253, 193, 216, 53, 253, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 206, 128, 91, 0, 193, 116, 51, 253, 193, 118, 51, 253, 193, 120, 51, 253, 193, 122, 51, 253, 193, 124, 51, 253, 193, 126, 51, 253, 193, 128, 51, 253, 193, 130, 51, 253, 193, 132, 51, 253, 193, 222, 51, 253, 193, 224, 51, 253, 193, 226, 51, 253, 193, 228, 51, 253, 193, 230, 51, 253, 193, 232, 51, 253, 193, 234, 51, 253, 193, 236, 51, 253, 193, 48, 52, 253, 193, 50, 52, 253, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 160, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 206, 192, 91, 0, 193, 134, 51, 253, 193, 136, 51, 253, 193, 138, 51, 253, 193, 140, 51, 253, 193, 142, 51, 253, 193, 238, 51, 253, 193, 240, 51, 253, 193, 62, 52, 253, 193, 158, 52, 253, 193, 248, 52, 253, 193, 94, 53, 253, 193, 140, 53, 253, 193, 58, 54, 253, 193, 98, 54, 253, 193, 100, 54, 253, 193, 122, 54, 253, 193, 124, 54, 253, 193, 198, 54, 253, 193, 226, 54, 253, 193, 242, 54, 253, 193, 244, 54, 253, 193, 60, 55, 253, 193, 124, 55, 253, 193, 126, 55, 253, 193, 128, 55, 253, 193, 180, 55, 253, 193, 182, 55, 253, 193, 184, 55, 253, 193, 186, 55, 253, 193, 22, 56, 253, 193, 24, 56, 253, 193, 26, 56, 253, 193, 204, 51, 253, 193, 206, 51, 253, 193, 208, 51, 253, 193, 28, 52, 253, 193, 30, 52, 253, 193, 32, 52, 253, 193, 126, 52, 253, 193, 128, 52, 253, 193, 214, 52, 253, 193, 28, 53, 253, 193, 66, 53, 253, 193, 68, 53, 253, 193, 70, 53, 253, 193, 72, 53, 253, 193, 126, 53, 253, 193, 128, 53, 253, 193, 130, 53, 253, 193, 164, 53, 253, 193, 200, 53, 253, 193, 248, 53, 253, 193, 28, 54, 253, 193, 30, 54, 253, 193, 32, 54, 253, 193, 60, 54, 253, 193, 74, 54, 253, 193, 110, 54, 253, 193, 128, 54, 253, 193, 130, 54, 253, 193, 132, 54, 253, 193, 134, 54, 253, 193, 148, 54, 253, 193, 180, 54, 253, 193, 26, 52, 253, 193, 124, 52, 253, 193, 212, 52, 253, 193, 122, 53, 253, 193, 124, 53, 253, 193, 162, 53, 253, 193, 24, 54, 253, 193, 26, 54, 253, 193, 72, 54, 253, 193, 108, 54, 253, 193, 126, 54, 253, 193, 168, 54, 253, 193, 178, 54, 253, 193, 218, 54, 253, 193, 234, 54, 253, 193, 10, 55, 253, 193, 48, 55, 253, 193, 238, 55, 253, 193, 90, 56, 253, 193, 198, 56, 253, 193, 200, 56, 253, 193, 54, 57, 253, 193, 56, 57, 253, 193, 58, 57, 253, 193, 152, 57, 253, 193, 154, 57, 253, 193, 156, 57, 253, 193, 248, 57, 253, 193, 250, 57, 253, 193, 62, 58, 253, 193, 144, 58, 253, 193, 146, 58, 253, 193, 68, 52, 253, 193, 70, 52, 253, 193, 72, 52, 253, 193, 74, 52, 253, 193, 76, 52, 253, 193, 78, 52, 253, 193, 80, 52, 253, 193, 82, 52, 253, 193, 84, 52, 253, 193, 86, 52, 253, 193, 88, 52, 253, 193, 90, 52, 253, 193, 92, 52, 253, 193, 94, 52, 253, 193, 96, 52, 253, 193, 98, 52, 253, 193, 100, 52, 253, 193, 102, 52, 253, 193, 104, 52, 253, 193, 106, 52, 253, 193, 108, 52, 253, 193, 110, 52, 253, 193, 112, 52, 253, 193, 114, 52, 253, 193, 116, 52, 253, 193, 118, 52, 253, 193, 120, 52, 253, 193, 122, 52, 253, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 193, 134, 52, 253, 193, 216, 52, 253, 193, 218, 52, 253, 193, 220, 52, 253, 193, 222, 52, 253, 193, 224, 52, 253, 193, 226, 52, 253, 193, 30, 53, 253, 193, 32, 53, 253, 193, 74, 53, 253, 193, 76, 53, 253, 193, 78, 53, 253, 193, 132, 53, 253, 193, 166, 53, 253, 193, 194, 53, 253, 193, 196, 53, 253, 193, 198, 53, 253, 193, 32, 55, 253, 193, 92, 55, 253, 193, 160, 55, 253, 193, 246, 55, 253, 193, 248, 55, 253, 193, 98, 56, 253, 193, 100, 56, 253, 193, 206, 56, 253, 193, 208, 56, 253, 193, 68, 57, 253, 193, 70, 57, 253, 193, 6, 58, 253, 193, 74, 58, 253, 193, 76, 58, 253, 193, 214, 58, 253, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 224, 91, 0, 206, 0, 92, 0, 206, 0, 92, 0, 206, 0, 92, 0, 206, 0, 92, 0, 206, 0, 92, 0, 206, 0, 92, 0, 206, 0, 92, 0, 206, 0, 92, 0, 206, 0, 92, 0, 206, 0, 92, 0, 206, 0, 92, 0, 206, 0, 92, 0, 193, 44, 53, 253, 193, 46, 53, 253, 193, 48, 53, 253, 193, 50, 53, 253, 193, 52, 53, 253, 193, 54, 53, 253, 193, 56, 53, 253, 193, 58, 53, 253, 193, 60, 53, 253, 193, 62, 53, 253, 193, 64, 53, 253, 193, 96, 53, 253, 193, 98, 53, 253, 193, 100, 53, 253, 193, 102, 53, 253, 193, 104, 53, 253, 193, 106, 53, 253, 193, 108, 53, 253, 193, 110, 53, 253, 193, 112, 53, 253, 193, 114, 53, 253, 193, 116, 53, 253, 193, 118, 53, 253, 193, 142, 53, 253, 193, 144, 53, 253, 193, 120, 53, 253, 193, 146, 53, 253, 193, 148, 53, 253, 193, 150, 53, 253, 193, 152, 53, 253, 193, 154, 53, 253, 193, 156, 53, 253, 193, 158, 53, 253, 193, 160, 53, 253, 193, 170, 53, 253, 193, 172, 53, 253, 193, 174, 53, 253, 193, 176, 53, 253, 193, 180, 53, 253, 193, 182, 53, 253, 193, 188, 53, 253, 193, 190, 53, 253, 193, 192, 53, 253, 193, 246, 53, 253, 193, 22, 54, 253, 193, 70, 54, 253, 193, 102, 54, 253, 193, 104, 54, 253, 193, 106, 54, 253, 193, 158, 54, 253, 193, 192, 54, 253, 193, 194, 54, 253, 193, 196, 54, 253, 193, 206, 54, 253, 193, 208, 54, 253, 193, 212, 54, 253, 193, 214, 54, 253, 193, 216, 54, 253, 193, 208, 53, 253, 193, 210, 53, 253, 193, 212, 53, 253, 193, 220, 53, 253, 193, 222, 53, 253, 193, 224, 53, 253, 193, 226, 53, 253, 193, 228, 53, 253, 193, 230, 53, 253, 193, 232, 53, 253, 193, 234, 53, 253, 193, 236, 53, 253, 193, 238, 53, 253, 193, 240, 53, 253, 193, 242, 53, 253, 193, 244, 53, 253, 193, 252, 53, 253, 193, 254, 53, 253, 193, 2, 54, 253, 193, 4, 54, 253, 193, 6, 54, 253, 193, 8, 54, 253, 193, 10, 54, 253, 193, 12, 54, 253, 193, 14, 54, 253, 193, 16, 54, 253, 193, 18, 54, 253, 193, 20, 54, 253, 206, 32, 92, 0, 206, 32, 92, 0, 206, 32, 92, 0, 206, 32, 92, 0, 193, 218, 53, 253, 193, 250, 53, 253, 193, 34, 54, 253, 193, 112, 54, 253, 193, 114, 54, 253, 193, 116, 54, 253, 193, 136, 54, 253, 193, 138, 54, 253, 193, 150, 54, 253, 193, 152, 54, 253, 193, 154, 54, 253, 193, 170, 54, 253, 193, 176, 54, 253, 193, 182, 54, 253, 193, 184, 54, 253, 193, 238, 54, 253, 193, 82, 55, 253, 193, 162, 55, 253, 193, 250, 55, 253, 193, 102, 56, 253, 193, 104, 56, 253, 193, 210, 56, 253, 193, 212, 56, 253, 193, 214, 56, 253, 193, 162, 57, 253, 193, 164, 57, 253, 193, 8, 58, 253, 193, 78, 58, 253, 193, 80, 58, 253, 193, 82, 58, 253, 193, 152, 58, 253, 193, 216, 58, 253, 206, 32, 92, 0, 206, 32, 92, 0, 206, 32, 92, 0, 206, 32, 92, 0, 206, 32, 92, 0, 206, 32, 92, 0, 206, 32, 92, 0, 193, 62, 54, 253, 193, 64, 54, 253, 193, 66, 54, 253, 193, 68, 54, 253, 193, 76, 54, 253, 193, 78, 54, 253, 193, 80, 54, 253, 193, 82, 54, 253, 193, 84, 54, 253, 193, 86, 54, 253, 193, 88, 54, 253, 193, 90, 54, 253, 193, 92, 54, 253, 193, 94, 54, 253, 193, 96, 54, 253, 193, 118, 54, 253, 193, 120, 54, 253, 193, 140, 54, 253, 193, 142, 54, 253, 193, 144, 54, 253, 193, 146, 54, 253, 193, 156, 54, 253, 193, 160, 54, 253, 193, 162, 54, 253, 193, 164, 54, 253, 193, 166, 54, 253, 193, 172, 54, 253, 193, 174, 54, 253, 193, 186, 54, 253, 193, 190, 54, 253, 193, 200, 54, 253, 193, 202, 54, 253, 193, 204, 54, 253, 193, 220, 54, 253, 193, 222, 54, 253, 193, 224, 54, 253, 193, 240, 54, 253, 193, 12, 55, 253, 193, 14, 55, 253, 193, 16, 55, 253, 193, 34, 55, 253, 193, 84, 55, 253, 193, 86, 55, 253, 193, 94, 55, 253, 193, 96, 55, 253, 193, 98, 55, 253, 193, 100, 55, 253, 193, 118, 55, 253, 193, 120, 55, 253, 193, 122, 55, 253, 193, 164, 55, 253, 193, 166, 55, 253, 193, 168, 55, 253, 193, 170, 55, 253, 193, 172, 55, 253, 193, 174, 55, 253, 193, 176, 55, 253, 193, 188, 54, 253, 193, 210, 54, 253, 193, 236, 54, 253, 193, 30, 55, 253, 193, 72, 55, 253, 193, 158, 55, 253, 193, 240, 55, 253, 193, 242, 55, 253, 193, 244, 55, 253, 193, 92, 56, 253, 193, 94, 56, 253, 193, 96, 56, 253, 193, 202, 56, 253, 193, 204, 56, 253, 193, 60, 57, 253, 193, 62, 57, 253, 193, 64, 57, 253, 193, 66, 57, 253, 193, 158, 57, 253, 193, 160, 57, 253, 193, 252, 57, 253, 193, 254, 57, 253, 193, 2, 58, 253, 193, 4, 58, 253, 193, 64, 58, 253, 193, 66, 58, 253, 193, 68, 58, 253, 193, 70, 58, 253, 193, 72, 58, 253, 193, 148, 58, 253, 193, 150, 58, 253, 193, 212, 58, 253, 193, 228, 54, 253, 193, 230, 54, 253, 193, 232, 54, 253, 193, 246, 54, 253, 193, 248, 54, 253, 193, 250, 54, 253, 193, 252, 54, 253, 193, 254, 54, 253, 193, 2, 55, 253, 193, 4, 55, 253, 193, 6, 55, 253, 193, 8, 55, 253, 193, 18, 55, 253, 193, 20, 55, 253, 193, 22, 55, 253, 193, 24, 55, 253, 193, 26, 55, 253, 193, 28, 55, 253, 193, 36, 55, 253, 193, 38, 55, 253, 193, 40, 55, 253, 193, 42, 55, 253, 193, 44, 55, 253, 193, 46, 55, 253, 193, 50, 55, 253, 193, 52, 55, 253, 193, 54, 55, 253, 193, 56, 55, 253, 193, 58, 55, 253, 193, 62, 55, 253, 193, 64, 55, 253, 193, 66, 55, 253, 193, 68, 55, 253, 193, 70, 55, 253, 193, 74, 55, 253, 193, 76, 55, 253, 193, 78, 55, 253, 193, 80, 55, 253, 193, 88, 55, 253, 193, 90, 55, 253, 193, 104, 55, 253, 193, 106, 55, 253, 193, 108, 55, 253, 193, 110, 55, 253, 193, 112, 55, 253, 193, 114, 55, 253, 193, 116, 55, 253, 193, 130, 55, 253, 193, 132, 55, 253, 193, 134, 55, 253, 193, 136, 55, 253, 193, 138, 55, 253, 193, 140, 55, 253, 193, 142, 55, 253, 193, 144, 55, 253, 193, 146, 55, 253, 193, 148, 55, 253, 193, 150, 55, 253, 193, 152, 55, 253, 193, 154, 55, 253, 193, 156, 55, 253, 193, 188, 55, 253, 193, 190, 55, 253, 193, 192, 55, 253, 193, 194, 55, 253, 193, 196, 55, 253, 193, 198, 55, 253, 193, 200, 55, 253, 193, 202, 55, 253, 193, 204, 55, 253, 193, 206, 55, 253, 193, 208, 55, 253, 193, 210, 55, 253, 193, 212, 55, 253, 193, 214, 55, 253, 193, 216, 55, 253, 193, 218, 55, 253, 193, 220, 55, 253, 193, 222, 55, 253, 193, 224, 55, 253, 193, 226, 55, 253, 193, 228, 55, 253, 193, 230, 55, 253, 193, 232, 55, 253, 193, 234, 55, 253, 193, 236, 55, 253, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 193, 252, 55, 253, 193, 254, 55, 253, 193, 2, 56, 253, 193, 4, 56, 253, 193, 6, 56, 253, 193, 8, 56, 253, 193, 10, 56, 253, 193, 12, 56, 253, 193, 14, 56, 253, 193, 16, 56, 253, 193, 18, 56, 253, 193, 20, 56, 253, 193, 106, 56, 253, 193, 108, 56, 253, 193, 110, 56, 253, 193, 112, 56, 253, 193, 114, 56, 253, 193, 116, 56, 253, 193, 118, 56, 253, 193, 120, 56, 253, 193, 122, 56, 253, 193, 216, 56, 253, 193, 218, 56, 253, 193, 220, 56, 253, 193, 222, 56, 253, 193, 224, 56, 253, 193, 226, 56, 253, 206, 160, 92, 0, 206, 160, 92, 0, 206, 160, 92, 0, 206, 160, 92, 0, 206, 160, 92, 0, 193, 28, 56, 253, 193, 30, 56, 253, 193, 32, 56, 253, 193, 34, 56, 253, 193, 36, 56, 253, 193, 124, 56, 253, 193, 126, 56, 253, 193, 128, 56, 253, 193, 130, 56, 253, 193, 132, 56, 253, 193, 134, 56, 253, 193, 228, 56, 253, 193, 230, 56, 253, 193, 232, 56, 253, 193, 234, 56, 253, 193, 236, 56, 253, 193, 238, 56, 253, 193, 240, 56, 253, 193, 242, 56, 253, 193, 92, 57, 253, 193, 94, 57, 253, 193, 96, 57, 253, 193, 98, 57, 253, 193, 100, 57, 253, 193, 102, 57, 253, 193, 172, 57, 253, 193, 174, 57, 253, 193, 176, 57, 253, 193, 178, 57, 253, 193, 12, 58, 253, 193, 14, 58, 253, 193, 16, 58, 253, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 64, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 96, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 206, 128, 92, 0, 193, 104, 57, 253, 193, 106, 57, 253, 193, 108, 57, 253, 193, 110, 57, 253, 193, 112, 57, 253, 193, 114, 57, 253, 193, 116, 57, 253, 193, 118, 57, 253, 193, 120, 57, 253, 193, 122, 57, 253, 193, 124, 57, 253, 193, 126, 57, 253, 193, 128, 57, 253, 193, 130, 57, 253, 193, 132, 57, 253, 193, 52, 57, 253, 193, 134, 57, 253, 193, 180, 57, 253, 206, 160, 92, 0, 206, 160, 92, 0, 206, 160, 92, 0, 206, 160, 92, 0, 206, 160, 92, 0, 193, 166, 57, 253, 193, 168, 57, 253, 193, 170, 57, 253, 193, 10, 58, 253, 193, 154, 58, 253, 193, 156, 58, 253, 193, 158, 58, 253, 193, 160, 58, 253, 193, 162, 58, 253, 193, 218, 58, 253, 193, 254, 58, 253, 193, 2, 59, 253, 193, 44, 59, 253, 193, 78, 59, 253, 193, 84, 59, 253, 193, 94, 59, 253, 193, 120, 59, 253, 193, 122, 59, 253, 193, 124, 59, 253, 193, 126, 59, 253, 193, 190, 59, 253, 193, 192, 59, 253, 193, 194, 59, 253, 193, 2, 61, 253, 193, 196, 59, 253, 193, 198, 59, 253, 193, 200, 59, 253, 193, 136, 57, 253, 193, 138, 57, 253, 193, 140, 57, 253, 193, 142, 57, 253, 193, 144, 57, 253, 193, 146, 57, 253, 193, 148, 57, 253, 193, 150, 57, 253, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 206, 192, 92, 0, 193, 20, 58, 253, 193, 22, 58, 253, 193, 24, 58, 253, 193, 26, 58, 253, 193, 28, 58, 253, 193, 30, 58, 253, 193, 32, 58, 253, 193, 34, 58, 253, 193, 36, 58, 253, 193, 38, 58, 253, 193, 40, 58, 253, 193, 42, 58, 253, 193, 44, 58, 253, 193, 46, 58, 253, 193, 48, 58, 253, 193, 50, 58, 253, 193, 52, 58, 253, 193, 54, 58, 253, 193, 56, 58, 253, 193, 58, 58, 253, 193, 60, 58, 253, 193, 96, 58, 253, 193, 98, 58, 253, 193, 18, 58, 253, 193, 84, 58, 253, 193, 86, 58, 253, 193, 88, 58, 253, 193, 90, 58, 253, 193, 92, 58, 253, 193, 94, 58, 253, 193, 164, 58, 253, 193, 220, 58, 253, 193, 222, 58, 253, 193, 224, 58, 253, 193, 226, 58, 253, 193, 4, 59, 253, 193, 22, 59, 253, 193, 24, 59, 253, 193, 26, 59, 253, 193, 46, 59, 253, 193, 60, 59, 253, 193, 66, 59, 253, 193, 96, 59, 253, 193, 98, 59, 253, 193, 128, 59, 253, 193, 130, 59, 253, 193, 212, 59, 253, 193, 214, 59, 253, 193, 216, 59, 253, 193, 218, 59, 253, 193, 220, 59, 253, 193, 222, 59, 253, 193, 224, 59, 253, 193, 226, 59, 253, 193, 126, 60, 253, 193, 100, 58, 253, 193, 102, 58, 253, 193, 104, 58, 253, 193, 106, 58, 253, 193, 108, 58, 253, 193, 110, 58, 253, 193, 112, 58, 253, 193, 114, 58, 253, 193, 116, 58, 253, 193, 118, 58, 253, 193, 120, 58, 253, 193, 122, 58, 253, 193, 124, 58, 253, 193, 126, 58, 253, 193, 128, 58, 253, 193, 130, 58, 253, 193, 132, 58, 253, 193, 134, 58, 253, 193, 136, 58, 253, 193, 138, 58, 253, 193, 140, 58, 253, 193, 142, 58, 253, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 206, 224, 92, 0, 193, 228, 58, 253, 193, 230, 58, 253, 193, 232, 58, 253, 193, 234, 58, 253, 193, 236, 58, 253, 193, 238, 58, 253, 193, 240, 58, 253, 193, 242, 58, 253, 193, 244, 58, 253, 193, 246, 58, 253, 193, 248, 58, 253, 193, 250, 58, 253, 193, 6, 59, 253, 193, 8, 59, 253, 193, 10, 59, 253, 193, 12, 59, 253, 193, 14, 59, 253, 193, 16, 59, 253, 193, 18, 59, 253, 193, 252, 58, 253, 193, 118, 59, 253, 193, 184, 59, 253, 193, 186, 59, 253, 193, 188, 59, 253, 193, 66, 60, 253, 193, 68, 60, 253, 193, 70, 60, 253, 193, 184, 61, 253, 193, 186, 61, 253, 193, 188, 61, 253, 193, 190, 61, 253, 193, 192, 61, 253, 193, 194, 61, 253, 193, 196, 61, 253, 193, 198, 61, 253, 193, 136, 62, 253, 193, 138, 62, 253, 193, 140, 62, 253, 193, 142, 62, 253, 193, 122, 63, 253, 193, 124, 63, 253, 193, 126, 63, 253, 193, 128, 63, 253, 193, 130, 63, 253, 193, 132, 63, 253, 206, 64, 94, 0, 206, 64, 94, 0, 206, 64, 94, 0, 206, 64, 94, 0, 206, 64, 94, 0, 206, 64, 94, 0, 193, 20, 59, 253, 193, 28, 59, 253, 193, 30, 59, 253, 193, 32, 59, 253, 193, 34, 59, 253, 193, 36, 59, 253, 193, 38, 59, 253, 193, 40, 59, 253, 193, 48, 59, 253, 193, 50, 59, 253, 193, 52, 59, 253, 193, 54, 59, 253, 193, 56, 59, 253, 193, 58, 59, 253, 193, 62, 59, 253, 193, 64, 59, 253, 193, 68, 59, 253, 193, 70, 59, 253, 193, 72, 59, 253, 193, 74, 59, 253, 193, 76, 59, 253, 193, 80, 59, 253, 193, 82, 59, 253, 193, 86, 59, 253, 193, 88, 59, 253, 193, 90, 59, 253, 193, 100, 59, 253, 193, 102, 59, 253, 193, 104, 59, 253, 193, 106, 59, 253, 193, 108, 59, 253, 193, 110, 59, 253, 193, 42, 59, 253, 193, 92, 59, 253, 193, 116, 59, 253, 193, 172, 59, 253, 193, 174, 59, 253, 193, 176, 59, 253, 193, 178, 59, 253, 193, 180, 59, 253, 193, 182, 59, 253, 193, 54, 60, 253, 193, 56, 60, 253, 193, 58, 60, 253, 193, 60, 60, 253, 193, 62, 60, 253, 193, 64, 60, 253, 193, 240, 60, 253, 193, 242, 60, 253, 193, 244, 60, 253, 193, 246, 60, 253, 193, 248, 60, 253, 193, 250, 60, 253, 193, 252, 60, 253, 193, 254, 60, 253, 193, 180, 61, 253, 193, 182, 61, 253, 193, 122, 62, 253, 193, 124, 62, 253, 193, 126, 62, 253, 193, 128, 62, 253, 193, 130, 62, 253, 193, 132, 62, 253, 193, 134, 62, 253, 193, 112, 59, 253, 193, 132, 59, 253, 193, 134, 59, 253, 193, 136, 59, 253, 193, 138, 59, 253, 193, 140, 59, 253, 193, 142, 59, 253, 193, 144, 59, 253, 193, 146, 59, 253, 193, 148, 59, 253, 193, 150, 59, 253, 193, 152, 59, 253, 193, 154, 59, 253, 193, 156, 59, 253, 193, 158, 59, 253, 193, 160, 59, 253, 193, 162, 59, 253, 193, 164, 59, 253, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 193, 114, 59, 253, 193, 168, 59, 253, 193, 170, 59, 253, 193, 44, 60, 253, 193, 46, 60, 253, 193, 48, 60, 253, 193, 50, 60, 253, 193, 52, 60, 253, 193, 232, 60, 253, 193, 234, 60, 253, 193, 236, 60, 253, 193, 238, 60, 253, 193, 168, 61, 253, 193, 170, 61, 253, 193, 172, 61, 253, 193, 174, 61, 253, 193, 176, 61, 253, 193, 178, 61, 253, 193, 114, 62, 253, 193, 116, 62, 253, 193, 118, 62, 253, 193, 120, 62, 253, 193, 80, 63, 253, 193, 82, 63, 253, 193, 84, 63, 253, 193, 86, 63, 253, 193, 88, 63, 253, 193, 90, 63, 253, 206, 32, 94, 0, 206, 32, 94, 0, 206, 32, 94, 0, 206, 32, 94, 0, 193, 166, 59, 253, 193, 38, 60, 253, 193, 40, 60, 253, 193, 42, 60, 253, 193, 226, 60, 253, 193, 228, 60, 253, 193, 230, 60, 253, 193, 158, 61, 253, 193, 160, 61, 253, 193, 162, 61, 253, 193, 164, 61, 253, 193, 166, 61, 253, 193, 108, 62, 253, 193, 110, 62, 253, 193, 112, 62, 253, 193, 70, 63, 253, 193, 72, 63, 253, 193, 74, 63, 253, 193, 76, 63, 253, 193, 78, 63, 253, 193, 50, 64, 253, 193, 52, 64, 253, 193, 54, 64, 253, 193, 56, 64, 253, 193, 58, 64, 253, 193, 24, 65, 253, 193, 26, 65, 253, 193, 28, 65, 253, 193, 6, 66, 253, 193, 8, 66, 253, 193, 10, 66, 253, 193, 12, 66, 253, 193, 202, 59, 253, 193, 204, 59, 253, 193, 206, 59, 253, 193, 208, 59, 253, 193, 210, 59, 253, 193, 72, 60, 253, 193, 74, 60, 253, 193, 76, 60, 253, 193, 78, 60, 253, 193, 80, 60, 253, 193, 82, 60, 253, 193, 84, 60, 253, 193, 86, 60, 253, 193, 88, 60, 253, 193, 90, 60, 253, 193, 92, 60, 253, 193, 94, 60, 253, 193, 96, 60, 253, 193, 98, 60, 253, 193, 100, 60, 253, 193, 102, 60, 253, 193, 104, 60, 253, 193, 106, 60, 253, 193, 108, 60, 253, 193, 110, 60, 253, 193, 112, 60, 253, 193, 114, 60, 253, 193, 116, 60, 253, 193, 118, 60, 253, 193, 120, 60, 253, 193, 122, 60, 253, 193, 4, 61, 253, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 0, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 193, 6, 61, 253, 193, 8, 61, 253, 193, 10, 61, 253, 193, 12, 61, 253, 193, 14, 61, 253, 193, 16, 61, 253, 193, 18, 61, 253, 193, 20, 61, 253, 193, 22, 61, 253, 193, 24, 61, 253, 193, 26, 61, 253, 193, 28, 61, 253, 193, 30, 61, 253, 193, 32, 61, 253, 193, 34, 61, 253, 193, 36, 61, 253, 193, 38, 61, 253, 193, 40, 61, 253, 193, 42, 61, 253, 193, 44, 61, 253, 193, 46, 61, 253, 193, 124, 60, 253, 193, 48, 61, 253, 193, 50, 61, 253, 193, 52, 61, 253, 193, 54, 61, 253, 193, 56, 61, 253, 193, 58, 61, 253, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 193, 128, 60, 253, 193, 130, 60, 253, 193, 132, 60, 253, 193, 134, 60, 253, 193, 136, 60, 253, 193, 138, 60, 253, 193, 140, 60, 253, 193, 142, 60, 253, 193, 144, 60, 253, 193, 146, 60, 253, 193, 148, 60, 253, 193, 150, 60, 253, 193, 152, 60, 253, 193, 154, 60, 253, 193, 60, 61, 253, 193, 62, 61, 253, 193, 64, 61, 253, 193, 66, 61, 253, 193, 68, 61, 253, 193, 70, 61, 253, 193, 238, 61, 253, 193, 240, 61, 253, 193, 242, 61, 253, 193, 244, 61, 253, 193, 246, 61, 253, 193, 248, 61, 253, 193, 250, 61, 253, 193, 252, 61, 253, 193, 210, 62, 253, 193, 254, 61, 253, 193, 212, 62, 253, 193, 214, 62, 253, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 32, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 64, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 96, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 128, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 206, 160, 93, 0, 193, 134, 63, 253, 193, 200, 62, 253, 193, 202, 62, 253, 193, 204, 62, 253, 193, 206, 62, 253, 193, 136, 63, 253, 193, 138, 63, 253, 193, 140, 63, 253, 193, 142, 63, 253, 193, 144, 63, 253, 193, 146, 63, 253, 193, 148, 63, 253, 193, 150, 63, 253, 193, 152, 63, 253, 193, 154, 63, 253, 193, 208, 62, 253, 206, 224, 93, 0, 206, 224, 93, 0, 206, 224, 93, 0, 206, 224, 93, 0, 206, 224, 93, 0, 193, 216, 62, 253, 193, 218, 62, 253, 193, 220, 62, 253, 193, 222, 62, 253, 193, 224, 62, 253, 193, 226, 62, 253, 193, 228, 62, 253, 193, 230, 62, 253, 193, 232, 62, 253, 193, 234, 62, 253, 193, 236, 62, 253, 193, 238, 62, 253, 193, 240, 62, 253, 193, 178, 63, 253, 193, 180, 63, 253, 193, 182, 63, 253, 193, 184, 63, 253, 193, 186, 63, 253, 193, 188, 63, 253, 193, 190, 63, 253, 193, 192, 63, 253, 193, 194, 63, 253, 193, 196, 63, 253, 193, 166, 64, 253, 193, 168, 64, 253, 193, 170, 64, 253, 193, 172, 64, 253, 193, 174, 64, 253, 193, 176, 64, 253, 193, 178, 64, 253, 193, 108, 65, 253, 193, 110, 65, 253, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 206, 192, 93, 0, 193, 198, 63, 253, 193, 200, 63, 253, 193, 68, 63, 253, 193, 92, 63, 253, 193, 94, 63, 253, 193, 96, 63, 253, 193, 98, 63, 253, 193, 100, 63, 253, 193, 102, 63, 253, 193, 104, 63, 253, 193, 106, 63, 253, 193, 108, 63, 253, 193, 110, 63, 253, 193, 112, 63, 253, 193, 114, 63, 253, 193, 116, 63, 253, 193, 118, 63, 253, 193, 120, 63, 253, 193, 76, 64, 253, 193, 78, 64, 253, 193, 80, 64, 253, 193, 82, 64, 253, 193, 84, 64, 253, 193, 86, 64, 253, 193, 88, 64, 253, 193, 90, 64, 253, 193, 92, 64, 253, 193, 94, 64, 253, 193, 96, 64, 253, 193, 98, 64, 253, 193, 100, 64, 253, 193, 102, 64, 253, 193, 104, 64, 253, 193, 106, 64, 253, 193, 108, 64, 253, 206, 224, 93, 0, 206, 224, 93, 0, 206, 224, 93, 0, 206, 224, 93, 0, 206, 224, 93, 0, 206, 224, 93, 0, 193, 138, 64, 253, 193, 140, 64, 253, 193, 142, 64, 253, 193, 144, 64, 253, 193, 146, 64, 253, 193, 148, 64, 253, 193, 150, 64, 253, 193, 152, 64, 253, 193, 154, 64, 253, 193, 156, 64, 253, 193, 158, 64, 253, 193, 160, 64, 253, 193, 162, 64, 253, 193, 164, 64, 253, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 206, 0, 94, 0, 193, 180, 64, 253, 193, 182, 64, 253, 193, 184, 64, 253, 193, 186, 64, 253, 193, 188, 64, 253, 193, 190, 64, 253, 193, 192, 64, 253, 193, 194, 64, 253, 193, 196, 64, 253, 193, 198, 64, 253, 193, 200, 64, 253, 193, 202, 64, 253, 193, 204, 64, 253, 206, 32, 94, 0, 206, 32, 94, 0, 206, 32, 94, 0, 206, 32, 94, 0, 193, 30, 65, 253, 193, 32, 65, 253, 193, 34, 65, 253, 193, 36, 65, 253, 193, 38, 65, 253, 193, 40, 65, 253, 193, 14, 66, 253, 193, 16, 66, 253, 193, 18, 66, 253, 193, 20, 66, 253, 193, 22, 66, 253, 193, 24, 66, 253, 193, 26, 66, 253, 193, 224, 66, 253, 193, 226, 66, 253, 193, 228, 66, 253, 193, 230, 66, 253, 193, 110, 67, 253, 193, 112, 67, 253, 193, 114, 67, 253, 193, 194, 67, 253, 193, 16, 68, 253, 193, 78, 68, 253, 193, 80, 68, 253, 193, 122, 68, 253, 193, 196, 68, 253, 193, 212, 68, 253, 193, 220, 68, 253, 206, 64, 94, 0, 206, 64, 94, 0, 206, 64, 94, 0, 206, 64, 94, 0, 206, 64, 94, 0, 206, 64, 94, 0, 206, 64, 94, 0, 206, 64, 94, 0, 193, 60, 65, 253, 193, 62, 65, 253, 193, 50, 66, 253, 193, 242, 66, 253, 193, 244, 66, 253, 193, 246, 66, 253, 193, 248, 66, 253, 193, 250, 66, 253, 193, 122, 67, 253, 193, 124, 67, 253, 193, 126, 67, 253, 193, 206, 67, 253, 193, 208, 67, 253, 193, 210, 67, 253, 193, 28, 68, 253, 193, 30, 68, 253, 193, 166, 68, 253, 193, 198, 68, 253, 193, 46, 69, 253, 193, 94, 69, 253, 193, 96, 69, 253, 193, 98, 69, 253, 193, 4, 70, 253, 193, 6, 70, 253, 193, 206, 64, 253, 193, 208, 64, 253, 193, 210, 64, 253, 193, 212, 64, 253, 193, 214, 64, 253, 193, 216, 64, 253, 193, 218, 64, 253, 193, 220, 64, 253, 193, 222, 64, 253, 193, 224, 64, 253, 193, 226, 64, 253, 193, 228, 64, 253, 193, 230, 64, 253, 193, 232, 64, 253, 193, 234, 64, 253, 193, 236, 64, 253, 193, 238, 64, 253, 193, 240, 64, 253, 193, 242, 64, 253, 193, 244, 64, 253, 193, 246, 64, 253, 193, 248, 64, 253, 193, 250, 64, 253, 193, 252, 64, 253, 193, 254, 64, 253, 193, 2, 65, 253, 193, 4, 65, 253, 193, 6, 65, 253, 193, 8, 65, 253, 193, 10, 65, 253, 193, 12, 65, 253, 193, 14, 65, 253, 193, 16, 65, 253, 193, 18, 65, 253, 193, 20, 65, 253, 193, 118, 65, 253, 193, 22, 65, 253, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 193, 42, 65, 253, 193, 44, 65, 253, 193, 46, 65, 253, 193, 48, 65, 253, 193, 50, 65, 253, 193, 52, 65, 253, 193, 54, 65, 253, 193, 56, 65, 253, 193, 58, 65, 253, 193, 28, 66, 253, 193, 30, 66, 253, 193, 32, 66, 253, 193, 34, 66, 253, 193, 36, 66, 253, 193, 38, 66, 253, 193, 40, 66, 253, 193, 42, 66, 253, 193, 44, 66, 253, 193, 46, 66, 253, 193, 48, 66, 253, 193, 232, 66, 253, 193, 234, 66, 253, 193, 236, 66, 253, 193, 238, 66, 253, 193, 240, 66, 253, 193, 116, 67, 253, 193, 118, 67, 253, 193, 120, 67, 253, 193, 196, 67, 253, 193, 198, 67, 253, 193, 200, 67, 253, 193, 202, 67, 253, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 206, 96, 94, 0, 193, 52, 66, 253, 193, 54, 66, 253, 193, 56, 66, 253, 193, 58, 66, 253, 193, 60, 66, 253, 193, 62, 66, 253, 193, 104, 65, 253, 193, 64, 66, 253, 193, 66, 66, 253, 193, 68, 66, 253, 193, 70, 66, 253, 193, 72, 66, 253, 193, 74, 66, 253, 193, 76, 66, 253, 193, 78, 66, 253, 193, 80, 66, 253, 193, 82, 66, 253, 193, 84, 66, 253, 193, 86, 66, 253, 193, 88, 66, 253, 193, 90, 66, 253, 193, 92, 66, 253, 193, 94, 66, 253, 193, 96, 66, 253, 193, 106, 65, 253, 193, 252, 66, 253, 193, 254, 66, 253, 193, 2, 67, 253, 193, 4, 67, 253, 193, 6, 67, 253, 193, 8, 67, 253, 193, 10, 67, 253, 193, 12, 67, 253, 193, 14, 67, 253, 193, 128, 67, 253, 193, 130, 67, 253, 193, 132, 67, 253, 193, 134, 67, 253, 193, 136, 67, 253, 193, 138, 67, 253, 193, 212, 67, 253, 193, 214, 67, 253, 193, 216, 67, 253, 193, 218, 67, 253, 193, 220, 67, 253, 193, 222, 67, 253, 193, 224, 67, 253, 193, 32, 68, 253, 193, 34, 68, 253, 193, 36, 68, 253, 193, 38, 68, 253, 193, 40, 68, 253, 193, 42, 68, 253, 193, 92, 68, 253, 193, 94, 68, 253, 193, 96, 68, 253, 193, 112, 65, 253, 193, 114, 65, 253, 193, 116, 65, 253, 193, 98, 66, 253, 193, 100, 66, 253, 193, 102, 66, 253, 193, 104, 66, 253, 193, 106, 66, 253, 193, 16, 67, 253, 193, 18, 67, 253, 193, 140, 67, 253, 193, 142, 67, 253, 193, 226, 67, 253, 193, 44, 68, 253, 193, 46, 68, 253, 193, 100, 68, 253, 193, 138, 68, 253, 193, 202, 68, 253, 193, 50, 69, 253, 193, 52, 69, 253, 193, 54, 69, 253, 193, 56, 69, 253, 193, 58, 69, 253, 193, 104, 69, 253, 193, 106, 69, 253, 193, 108, 69, 253, 193, 110, 69, 253, 193, 214, 69, 253, 193, 14, 70, 253, 193, 52, 70, 253, 193, 54, 70, 253, 193, 120, 70, 253, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 206, 128, 94, 0, 193, 108, 66, 253, 193, 110, 66, 253, 193, 112, 66, 253, 193, 114, 66, 253, 193, 116, 66, 253, 193, 118, 66, 253, 193, 120, 66, 253, 193, 122, 66, 253, 193, 124, 66, 253, 193, 126, 66, 253, 193, 128, 66, 253, 193, 130, 66, 253, 193, 132, 66, 253, 193, 134, 66, 253, 193, 136, 66, 253, 193, 138, 66, 253, 193, 4, 66, 253, 193, 140, 66, 253, 193, 142, 66, 253, 193, 144, 66, 253, 193, 146, 66, 253, 193, 148, 66, 253, 193, 150, 66, 253, 193, 152, 66, 253, 193, 154, 66, 253, 193, 156, 66, 253, 193, 158, 66, 253, 193, 160, 66, 253, 193, 162, 66, 253, 193, 164, 66, 253, 193, 166, 66, 253, 193, 168, 66, 253, 193, 170, 66, 253, 193, 172, 66, 253, 193, 174, 66, 253, 193, 176, 66, 253, 193, 178, 66, 253, 193, 180, 66, 253, 193, 182, 66, 253, 193, 184, 66, 253, 193, 186, 66, 253, 193, 188, 66, 253, 193, 190, 66, 253, 193, 192, 66, 253, 193, 194, 66, 253, 193, 196, 66, 253, 193, 198, 66, 253, 193, 200, 66, 253, 193, 202, 66, 253, 193, 204, 66, 253, 193, 206, 66, 253, 193, 208, 66, 253, 193, 210, 66, 253, 193, 212, 66, 253, 193, 214, 66, 253, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 193, 216, 66, 253, 193, 218, 66, 253, 193, 220, 66, 253, 193, 102, 67, 253, 193, 104, 67, 253, 193, 106, 67, 253, 193, 108, 67, 253, 193, 190, 67, 253, 193, 192, 67, 253, 193, 76, 68, 253, 193, 144, 69, 253, 193, 146, 69, 253, 193, 190, 69, 253, 193, 80, 70, 253, 193, 82, 70, 253, 193, 108, 70, 253, 193, 150, 70, 253, 193, 172, 70, 253, 193, 38, 71, 253, 193, 146, 71, 253, 193, 22, 72, 253, 193, 24, 72, 253, 193, 248, 72, 253, 193, 250, 72, 253, 193, 252, 72, 253, 193, 138, 73, 253, 193, 34, 74, 253, 193, 36, 74, 253, 193, 152, 74, 253, 193, 154, 74, 253, 193, 254, 74, 253, 193, 2, 75, 253, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 206, 160, 94, 0, 193, 144, 67, 253, 193, 146, 67, 253, 193, 148, 67, 253, 193, 150, 67, 253, 193, 152, 67, 253, 193, 154, 67, 253, 193, 156, 67, 253, 193, 158, 67, 253, 193, 160, 67, 253, 193, 162, 67, 253, 193, 164, 67, 253, 193, 166, 67, 253, 193, 168, 67, 253, 193, 170, 67, 253, 193, 172, 67, 253, 193, 174, 67, 253, 193, 176, 67, 253, 193, 178, 67, 253, 193, 180, 67, 253, 193, 182, 67, 253, 193, 184, 67, 253, 193, 186, 67, 253, 193, 188, 67, 253, 193, 228, 67, 253, 193, 230, 67, 253, 193, 232, 67, 253, 193, 234, 67, 253, 193, 236, 67, 253, 193, 238, 67, 253, 193, 240, 67, 253, 193, 242, 67, 253, 193, 244, 67, 253, 193, 246, 67, 253, 193, 248, 67, 253, 193, 250, 67, 253, 193, 252, 67, 253, 193, 254, 67, 253, 193, 2, 68, 253, 193, 4, 68, 253, 193, 6, 68, 253, 193, 8, 68, 253, 193, 10, 68, 253, 193, 12, 68, 253, 193, 14, 68, 253, 206, 192, 94, 0, 206, 192, 94, 0, 206, 192, 94, 0, 206, 192, 94, 0, 206, 192, 94, 0, 206, 192, 94, 0, 206, 192, 94, 0, 206, 192, 94, 0, 206, 192, 94, 0, 206, 192, 94, 0, 193, 204, 67, 253, 193, 18, 68, 253, 193, 20, 68, 253, 193, 22, 68, 253, 193, 24, 68, 253, 193, 26, 68, 253, 193, 82, 68, 253, 193, 84, 68, 253, 193, 86, 68, 253, 193, 88, 68, 253, 193, 90, 68, 253, 193, 124, 68, 253, 193, 126, 68, 253, 193, 184, 68, 253, 193, 254, 68, 253, 193, 44, 69, 253, 193, 92, 69, 253, 193, 150, 69, 253, 193, 152, 69, 253, 193, 154, 69, 253, 193, 156, 69, 253, 193, 158, 69, 253, 193, 192, 69, 253, 193, 194, 69, 253, 193, 196, 69, 253, 193, 198, 69, 253, 193, 174, 70, 253, 193, 176, 70, 253, 193, 178, 70, 253, 193, 204, 70, 253, 193, 212, 70, 253, 193, 40, 71, 253, 206, 192, 94, 0, 206, 192, 94, 0, 206, 192, 94, 0, 206, 192, 94, 0, 193, 102, 68, 253, 193, 104, 68, 253, 193, 106, 68, 253, 193, 108, 68, 253, 193, 110, 68, 253, 193, 112, 68, 253, 193, 114, 68, 253, 193, 116, 68, 253, 193, 118, 68, 253, 193, 120, 68, 253, 193, 140, 68, 253, 193, 142, 68, 253, 193, 144, 68, 253, 193, 146, 68, 253, 193, 148, 68, 253, 193, 150, 68, 253, 193, 152, 68, 253, 193, 154, 68, 253, 193, 156, 68, 253, 193, 158, 68, 253, 193, 160, 68, 253, 193, 162, 68, 253, 193, 164, 68, 253, 193, 172, 68, 253, 193, 174, 68, 253, 193, 176, 68, 253, 193, 178, 68, 253, 193, 180, 68, 253, 193, 98, 68, 253, 193, 128, 68, 253, 193, 130, 68, 253, 193, 132, 68, 253, 193, 134, 68, 253, 193, 136, 68, 253, 193, 168, 68, 253, 193, 170, 68, 253, 193, 200, 68, 253, 193, 186, 68, 253, 193, 188, 68, 253, 193, 230, 68, 253, 193, 242, 68, 253, 193, 2, 69, 253, 193, 4, 69, 253, 193, 6, 69, 253, 193, 48, 69, 253, 193, 100, 69, 253, 193, 102, 69, 253, 193, 160, 69, 253, 193, 162, 69, 253, 193, 164, 69, 253, 193, 200, 69, 253, 193, 202, 69, 253, 193, 204, 69, 253, 193, 206, 69, 253, 193, 208, 69, 253, 193, 210, 69, 253, 193, 212, 69, 253, 193, 10, 70, 253, 193, 12, 70, 253, 193, 50, 70, 253, 193, 182, 68, 253, 193, 190, 68, 253, 193, 192, 68, 253, 193, 194, 68, 253, 193, 204, 68, 253, 193, 206, 68, 253, 193, 208, 68, 253, 193, 210, 68, 253, 193, 214, 68, 253, 193, 216, 68, 253, 193, 218, 68, 253, 193, 222, 68, 253, 193, 224, 68, 253, 193, 226, 68, 253, 193, 228, 68, 253, 193, 232, 68, 253, 193, 234, 68, 253, 193, 236, 68, 253, 193, 240, 68, 253, 193, 244, 68, 253, 193, 246, 68, 253, 193, 248, 68, 253, 193, 250, 68, 253, 193, 252, 68, 253, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 193, 238, 68, 253, 193, 42, 69, 253, 193, 90, 69, 253, 193, 148, 69, 253, 193, 244, 69, 253, 193, 246, 69, 253, 193, 248, 69, 253, 193, 250, 69, 253, 193, 252, 69, 253, 193, 254, 69, 253, 193, 2, 70, 253, 193, 38, 70, 253, 193, 84, 70, 253, 193, 86, 70, 253, 193, 88, 70, 253, 193, 110, 70, 253, 193, 152, 70, 253, 193, 148, 71, 253, 193, 26, 72, 253, 193, 28, 72, 253, 193, 30, 72, 253, 193, 148, 72, 253, 193, 140, 73, 253, 193, 142, 73, 253, 193, 38, 74, 253, 193, 40, 74, 253, 193, 156, 74, 253, 193, 4, 75, 253, 193, 6, 75, 253, 193, 8, 75, 253, 193, 148, 75, 253, 193, 240, 75, 253, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 206, 224, 94, 0, 193, 60, 69, 253, 193, 62, 69, 253, 193, 64, 69, 253, 193, 66, 69, 253, 193, 68, 69, 253, 193, 70, 69, 253, 193, 72, 69, 253, 193, 74, 69, 253, 193, 76, 69, 253, 193, 78, 69, 253, 193, 80, 69, 253, 193, 82, 69, 253, 193, 84, 69, 253, 193, 86, 69, 253, 193, 88, 69, 253, 206, 0, 95, 0, 206, 0, 95, 0, 206, 0, 95, 0, 206, 0, 95, 0, 206, 0, 95, 0, 206, 0, 95, 0, 206, 0, 95, 0, 206, 0, 95, 0, 193, 166, 69, 253, 193, 168, 69, 253, 193, 170, 69, 253, 193, 172, 69, 253, 193, 174, 69, 253, 193, 176, 69, 253, 193, 178, 69, 253, 193, 180, 69, 253, 193, 182, 69, 253, 193, 184, 69, 253, 193, 186, 69, 253, 193, 188, 69, 253, 193, 216, 69, 253, 193, 218, 69, 253, 193, 220, 69, 253, 193, 222, 69, 253, 193, 224, 69, 253, 193, 226, 69, 253, 193, 228, 69, 253, 193, 230, 69, 253, 193, 232, 69, 253, 193, 234, 69, 253, 193, 236, 69, 253, 193, 238, 69, 253, 193, 240, 69, 253, 193, 242, 69, 253, 193, 16, 70, 253, 193, 18, 70, 253, 193, 20, 70, 253, 193, 22, 70, 253, 193, 56, 70, 253, 193, 24, 70, 253, 193, 26, 70, 253, 193, 28, 70, 253, 193, 30, 70, 253, 193, 32, 70, 253, 193, 34, 70, 253, 193, 36, 70, 253, 193, 58, 70, 253, 193, 60, 70, 253, 193, 62, 70, 253, 193, 64, 70, 253, 193, 66, 70, 253, 193, 68, 70, 253, 193, 70, 70, 253, 193, 72, 70, 253, 193, 74, 70, 253, 193, 76, 70, 253, 193, 78, 70, 253, 193, 96, 70, 253, 193, 98, 70, 253, 193, 100, 70, 253, 193, 102, 70, 253, 193, 104, 70, 253, 193, 106, 70, 253, 193, 122, 70, 253, 193, 8, 70, 253, 193, 40, 70, 253, 193, 42, 70, 253, 193, 44, 70, 253, 193, 46, 70, 253, 193, 48, 70, 253, 193, 90, 70, 253, 193, 154, 70, 253, 193, 222, 70, 253, 193, 224, 70, 253, 193, 42, 71, 253, 193, 44, 71, 253, 193, 162, 71, 253, 193, 46, 72, 253, 193, 48, 72, 253, 193, 50, 72, 253, 193, 52, 72, 253, 193, 160, 72, 253, 193, 4, 73, 253, 193, 162, 73, 253, 193, 164, 73, 253, 193, 166, 73, 253, 193, 168, 73, 253, 193, 170, 73, 253, 193, 172, 73, 253, 193, 174, 73, 253, 193, 176, 73, 253, 193, 48, 74, 253, 193, 50, 74, 253, 193, 52, 74, 253, 193, 54, 74, 253, 193, 164, 74, 253, 193, 92, 70, 253, 193, 94, 70, 253, 193, 112, 70, 253, 193, 114, 70, 253, 193, 116, 70, 253, 193, 118, 70, 253, 193, 156, 70, 253, 193, 180, 70, 253, 193, 206, 70, 253, 193, 216, 70, 253, 193, 218, 70, 253, 193, 220, 70, 253, 193, 226, 70, 253, 193, 228, 70, 253, 193, 250, 70, 253, 193, 252, 70, 253, 193, 254, 70, 253, 193, 2, 71, 253, 193, 4, 71, 253, 193, 6, 71, 253, 193, 8, 71, 253, 193, 10, 71, 253, 206, 32, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 193, 124, 70, 253, 193, 126, 70, 253, 193, 128, 70, 253, 193, 130, 70, 253, 193, 132, 70, 253, 193, 134, 70, 253, 193, 136, 70, 253, 193, 138, 70, 253, 193, 140, 70, 253, 193, 142, 70, 253, 193, 144, 70, 253, 193, 146, 70, 253, 193, 148, 70, 253, 193, 160, 70, 253, 193, 162, 70, 253, 193, 164, 70, 253, 193, 166, 70, 253, 193, 168, 70, 253, 193, 170, 70, 253, 193, 182, 70, 253, 193, 184, 70, 253, 193, 186, 70, 253, 193, 188, 70, 253, 193, 190, 70, 253, 193, 192, 70, 253, 193, 194, 70, 253, 193, 196, 70, 253, 193, 200, 70, 253, 193, 202, 70, 253, 193, 208, 70, 253, 193, 210, 70, 253, 193, 214, 70, 253, 193, 158, 70, 253, 193, 198, 70, 253, 193, 230, 70, 253, 193, 232, 70, 253, 193, 234, 70, 253, 193, 12, 71, 253, 193, 74, 71, 253, 193, 76, 71, 253, 193, 78, 71, 253, 193, 80, 71, 253, 193, 82, 71, 253, 193, 84, 71, 253, 193, 86, 71, 253, 193, 88, 71, 253, 193, 90, 71, 253, 193, 216, 71, 253, 193, 218, 71, 253, 193, 220, 71, 253, 193, 222, 71, 253, 193, 86, 72, 253, 193, 88, 72, 253, 193, 90, 72, 253, 193, 92, 72, 253, 193, 94, 72, 253, 193, 180, 72, 253, 193, 182, 72, 253, 193, 184, 72, 253, 193, 186, 72, 253, 193, 52, 73, 253, 193, 54, 73, 253, 193, 56, 73, 253, 193, 58, 73, 253, 193, 236, 70, 253, 193, 238, 70, 253, 193, 240, 70, 253, 193, 242, 70, 253, 193, 244, 70, 253, 193, 246, 70, 253, 193, 248, 70, 253, 193, 14, 71, 253, 193, 16, 71, 253, 193, 18, 71, 253, 193, 20, 71, 253, 193, 22, 71, 253, 193, 24, 71, 253, 193, 26, 71, 253, 193, 28, 71, 253, 193, 30, 71, 253, 193, 32, 71, 253, 193, 34, 71, 253, 193, 36, 71, 253, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 206, 32, 95, 0, 193, 164, 71, 253, 193, 166, 71, 253, 193, 168, 71, 253, 193, 170, 71, 253, 193, 172, 71, 253, 193, 174, 71, 253, 193, 176, 71, 253, 193, 178, 71, 253, 193, 180, 71, 253, 193, 54, 72, 253, 193, 182, 71, 253, 193, 184, 71, 253, 193, 186, 71, 253, 193, 188, 71, 253, 193, 190, 71, 253, 193, 192, 71, 253, 193, 194, 71, 253, 193, 196, 71, 253, 193, 198, 71, 253, 193, 200, 71, 253, 193, 202, 71, 253, 193, 204, 71, 253, 193, 206, 71, 253, 193, 208, 71, 253, 193, 210, 71, 253, 193, 212, 71, 253, 193, 56, 72, 253, 193, 58, 72, 253, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 64, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 193, 150, 71, 253, 193, 152, 71, 253, 193, 154, 71, 253, 193, 156, 71, 253, 193, 158, 71, 253, 193, 160, 71, 253, 193, 32, 72, 253, 193, 34, 72, 253, 193, 36, 72, 253, 193, 38, 72, 253, 193, 40, 72, 253, 193, 42, 72, 253, 193, 44, 72, 253, 193, 150, 72, 253, 193, 152, 72, 253, 193, 154, 72, 253, 193, 156, 72, 253, 193, 158, 72, 253, 193, 254, 72, 253, 193, 2, 73, 253, 193, 144, 73, 253, 193, 146, 73, 253, 193, 148, 73, 253, 193, 150, 73, 253, 193, 152, 73, 253, 193, 154, 73, 253, 193, 156, 73, 253, 193, 158, 73, 253, 193, 160, 73, 253, 193, 42, 74, 253, 193, 44, 74, 253, 193, 46, 74, 253, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 206, 96, 95, 0, 193, 96, 72, 253, 193, 98, 72, 253, 193, 100, 72, 253, 193, 102, 72, 253, 193, 104, 72, 253, 193, 106, 72, 253, 193, 108, 72, 253, 193, 110, 72, 253, 193, 112, 72, 253, 193, 114, 72, 253, 193, 116, 72, 253, 193, 118, 72, 253, 193, 120, 72, 253, 193, 122, 72, 253, 193, 124, 72, 253, 193, 126, 72, 253, 193, 128, 72, 253, 193, 130, 72, 253, 193, 132, 72, 253, 193, 134, 72, 253, 193, 136, 72, 253, 193, 138, 72, 253, 193, 140, 72, 253, 193, 142, 72, 253, 193, 60, 72, 253, 193, 62, 72, 253, 193, 64, 72, 253, 193, 66, 72, 253, 193, 68, 72, 253, 193, 70, 72, 253, 193, 72, 72, 253, 193, 74, 72, 253, 193, 76, 72, 253, 193, 78, 72, 253, 193, 80, 72, 253, 193, 82, 72, 253, 193, 84, 72, 253, 193, 162, 72, 253, 193, 164, 72, 253, 193, 166, 72, 253, 193, 168, 72, 253, 193, 170, 72, 253, 193, 172, 72, 253, 193, 174, 72, 253, 193, 176, 72, 253, 193, 178, 72, 253, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 193, 144, 72, 253, 193, 146, 72, 253, 193, 188, 72, 253, 193, 190, 72, 253, 193, 192, 72, 253, 193, 194, 72, 253, 193, 196, 72, 253, 193, 198, 72, 253, 193, 200, 72, 253, 193, 202, 72, 253, 193, 204, 72, 253, 193, 206, 72, 253, 193, 208, 72, 253, 193, 210, 72, 253, 193, 212, 72, 253, 193, 214, 72, 253, 193, 216, 72, 253, 193, 218, 72, 253, 193, 220, 72, 253, 193, 222, 72, 253, 193, 224, 72, 253, 193, 226, 72, 253, 193, 228, 72, 253, 193, 230, 72, 253, 193, 232, 72, 253, 193, 234, 72, 253, 193, 236, 72, 253, 193, 238, 72, 253, 193, 240, 72, 253, 193, 242, 72, 253, 193, 244, 72, 253, 193, 246, 72, 253, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 206, 128, 95, 0, 193, 178, 73, 253, 193, 180, 73, 253, 193, 182, 73, 253, 193, 184, 73, 253, 193, 186, 73, 253, 193, 188, 73, 253, 193, 190, 73, 253, 193, 192, 73, 253, 193, 194, 73, 253, 193, 196, 73, 253, 193, 198, 73, 253, 193, 200, 73, 253, 193, 202, 73, 253, 193, 204, 73, 253, 193, 206, 73, 253, 193, 208, 73, 253, 193, 56, 74, 253, 193, 58, 74, 253, 193, 60, 74, 253, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 160, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 193, 210, 73, 253, 193, 212, 73, 253, 193, 214, 73, 253, 193, 216, 73, 253, 193, 218, 73, 253, 193, 74, 74, 253, 193, 76, 74, 253, 193, 78, 74, 253, 193, 80, 74, 253, 193, 82, 74, 253, 193, 84, 74, 253, 193, 188, 74, 253, 193, 190, 74, 253, 193, 192, 74, 253, 193, 194, 74, 253, 193, 196, 74, 253, 193, 198, 74, 253, 193, 200, 74, 253, 193, 34, 75, 253, 193, 36, 75, 253, 193, 38, 75, 253, 193, 40, 75, 253, 193, 98, 75, 253, 193, 168, 75, 253, 193, 170, 75, 253, 193, 172, 75, 253, 193, 210, 75, 253, 193, 212, 75, 253, 193, 26, 76, 253, 193, 54, 76, 253, 193, 70, 76, 253, 193, 90, 76, 253, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 192, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 193, 62, 74, 253, 193, 64, 74, 253, 193, 66, 74, 253, 193, 68, 74, 253, 193, 70, 74, 253, 193, 72, 74, 253, 193, 172, 74, 253, 193, 174, 74, 253, 193, 176, 74, 253, 193, 178, 74, 253, 193, 180, 74, 253, 193, 182, 74, 253, 193, 184, 74, 253, 193, 18, 75, 253, 193, 20, 75, 253, 193, 22, 75, 253, 193, 24, 75, 253, 193, 26, 75, 253, 193, 28, 75, 253, 193, 30, 75, 253, 193, 32, 75, 253, 193, 86, 75, 253, 193, 88, 75, 253, 193, 90, 75, 253, 193, 92, 75, 253, 193, 94, 75, 253, 193, 96, 75, 253, 193, 158, 75, 253, 193, 160, 75, 253, 193, 162, 75, 253, 193, 164, 75, 253, 193, 166, 75, 253, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 224, 95, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 193, 158, 74, 253, 193, 160, 74, 253, 193, 162, 74, 253, 193, 10, 75, 253, 193, 12, 75, 253, 193, 14, 75, 253, 193, 72, 75, 253, 193, 74, 75, 253, 193, 76, 75, 253, 193, 150, 75, 253, 193, 152, 75, 253, 193, 154, 75, 253, 193, 194, 75, 253, 193, 196, 75, 253, 193, 198, 75, 253, 193, 200, 75, 253, 193, 244, 75, 253, 193, 246, 75, 253, 193, 248, 75, 253, 193, 158, 76, 253, 193, 210, 76, 253, 193, 254, 76, 253, 193, 24, 77, 253, 193, 96, 77, 253, 193, 98, 77, 253, 193, 164, 77, 253, 193, 166, 77, 253, 193, 6, 78, 253, 193, 20, 78, 253, 193, 80, 78, 253, 193, 134, 78, 253, 193, 160, 78, 253, 193, 166, 74, 253, 193, 168, 74, 253, 193, 170, 74, 253, 193, 16, 75, 253, 193, 78, 75, 253, 193, 80, 75, 253, 193, 82, 75, 253, 193, 84, 75, 253, 193, 156, 75, 253, 193, 202, 75, 253, 193, 250, 75, 253, 193, 252, 75, 253, 193, 52, 76, 253, 193, 66, 76, 253, 193, 94, 76, 253, 193, 104, 76, 253, 193, 108, 76, 253, 193, 110, 76, 253, 193, 124, 76, 253, 193, 126, 76, 253, 193, 128, 76, 253, 193, 130, 76, 253, 193, 160, 76, 253, 193, 162, 76, 253, 193, 164, 76, 253, 193, 166, 76, 253, 193, 168, 76, 253, 193, 170, 76, 253, 206, 64, 96, 0, 206, 64, 96, 0, 206, 64, 96, 0, 206, 64, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 206, 0, 96, 0, 193, 42, 75, 253, 193, 44, 75, 253, 193, 46, 75, 253, 193, 48, 75, 253, 193, 50, 75, 253, 193, 52, 75, 253, 193, 54, 75, 253, 193, 56, 75, 253, 193, 58, 75, 253, 193, 60, 75, 253, 193, 62, 75, 253, 193, 64, 75, 253, 193, 66, 75, 253, 193, 68, 75, 253, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 193, 70, 75, 253, 193, 146, 75, 253, 193, 192, 75, 253, 193, 96, 76, 253, 193, 120, 76, 253, 193, 122, 76, 253, 193, 148, 76, 253, 193, 150, 76, 253, 193, 198, 76, 253, 193, 200, 76, 253, 193, 202, 76, 253, 193, 246, 76, 253, 193, 248, 76, 253, 193, 14, 77, 253, 193, 46, 77, 253, 193, 48, 77, 253, 193, 50, 77, 253, 193, 52, 77, 253, 193, 86, 77, 253, 193, 88, 77, 253, 193, 114, 77, 253, 193, 128, 77, 253, 193, 130, 77, 253, 193, 146, 77, 253, 193, 158, 77, 253, 193, 176, 77, 253, 193, 216, 77, 253, 193, 238, 77, 253, 193, 240, 77, 253, 193, 108, 78, 253, 193, 184, 78, 253, 193, 22, 79, 253, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 206, 32, 96, 0, 193, 174, 75, 253, 193, 176, 75, 253, 193, 178, 75, 253, 193, 180, 75, 253, 193, 182, 75, 253, 193, 184, 75, 253, 193, 186, 75, 253, 193, 188, 75, 253, 193, 190, 75, 253, 193, 214, 75, 253, 193, 216, 75, 253, 193, 218, 75, 253, 193, 220, 75, 253, 193, 222, 75, 253, 193, 224, 75, 253, 193, 226, 75, 253, 193, 228, 75, 253, 193, 230, 75, 253, 193, 232, 75, 253, 193, 234, 75, 253, 193, 236, 75, 253, 193, 238, 75, 253, 193, 4, 76, 253, 193, 6, 76, 253, 193, 78, 86, 253, 193, 124, 86, 253, 193, 126, 86, 253, 193, 128, 86, 253, 193, 130, 86, 253, 193, 132, 86, 253, 193, 134, 86, 253, 193, 136, 86, 253, 193, 138, 86, 253, 193, 140, 86, 253, 193, 142, 86, 253, 193, 144, 86, 253, 193, 146, 86, 253, 193, 148, 86, 253, 193, 150, 86, 253, 193, 152, 86, 253, 193, 154, 86, 253, 193, 156, 86, 253, 193, 158, 86, 253, 193, 160, 86, 253, 193, 162, 86, 253, 193, 164, 86, 253, 193, 166, 86, 253, 193, 198, 86, 253, 193, 200, 86, 253, 193, 202, 86, 253, 193, 204, 86, 253, 193, 206, 86, 253, 193, 144, 75, 253, 193, 208, 86, 253, 193, 210, 86, 253, 193, 212, 86, 253, 193, 204, 75, 253, 193, 206, 75, 253, 193, 208, 75, 253, 193, 254, 75, 253, 193, 2, 76, 253, 193, 68, 76, 253, 193, 100, 76, 253, 193, 102, 76, 253, 193, 106, 76, 253, 193, 112, 76, 253, 193, 114, 76, 253, 193, 116, 76, 253, 193, 132, 76, 253, 193, 134, 76, 253, 193, 136, 76, 253, 193, 138, 76, 253, 193, 140, 76, 253, 193, 172, 76, 253, 193, 174, 76, 253, 193, 176, 76, 253, 193, 178, 76, 253, 193, 180, 76, 253, 193, 182, 76, 253, 193, 184, 76, 253, 193, 186, 76, 253, 193, 188, 76, 253, 193, 190, 76, 253, 193, 192, 76, 253, 206, 96, 96, 0, 206, 96, 96, 0, 206, 96, 96, 0, 206, 96, 96, 0, 193, 242, 75, 253, 193, 152, 76, 253, 193, 154, 76, 253, 193, 156, 76, 253, 193, 206, 76, 253, 193, 208, 76, 253, 193, 250, 76, 253, 193, 252, 76, 253, 193, 16, 77, 253, 193, 18, 77, 253, 193, 20, 77, 253, 193, 22, 77, 253, 193, 54, 77, 253, 193, 90, 77, 253, 193, 92, 77, 253, 193, 94, 77, 253, 193, 118, 77, 253, 193, 120, 77, 253, 193, 132, 77, 253, 193, 134, 77, 253, 193, 162, 77, 253, 193, 172, 77, 253, 193, 192, 77, 253, 193, 218, 77, 253, 193, 220, 77, 253, 193, 222, 77, 253, 193, 4, 78, 253, 193, 120, 78, 253, 193, 208, 78, 253, 193, 164, 79, 253, 193, 166, 79, 253, 193, 168, 79, 253, 193, 8, 76, 253, 193, 10, 76, 253, 193, 12, 76, 253, 193, 14, 76, 253, 193, 16, 76, 253, 193, 18, 76, 253, 193, 20, 76, 253, 193, 22, 76, 253, 193, 24, 76, 253, 193, 28, 76, 253, 193, 30, 76, 253, 193, 32, 76, 253, 193, 34, 76, 253, 193, 36, 76, 253, 193, 38, 76, 253, 193, 40, 76, 253, 193, 42, 76, 253, 193, 44, 76, 253, 193, 46, 76, 253, 193, 48, 76, 253, 193, 50, 76, 253, 193, 56, 76, 253, 193, 58, 76, 253, 193, 60, 76, 253, 193, 62, 76, 253, 193, 72, 76, 253, 193, 78, 76, 253, 193, 74, 76, 253, 193, 80, 76, 253, 193, 64, 76, 253, 193, 76, 76, 253, 193, 82, 76, 253, 193, 84, 76, 253, 193, 86, 76, 253, 193, 88, 76, 253, 193, 92, 76, 253, 193, 98, 76, 253, 193, 144, 76, 253, 193, 146, 76, 253, 193, 196, 76, 253, 193, 244, 76, 253, 193, 144, 77, 253, 193, 184, 77, 253, 193, 188, 77, 253, 193, 190, 77, 253, 193, 194, 77, 253, 193, 196, 77, 253, 193, 200, 77, 253, 193, 202, 77, 253, 193, 204, 77, 253, 193, 212, 77, 253, 193, 214, 77, 253, 193, 230, 77, 253, 193, 232, 77, 253, 193, 234, 77, 253, 193, 236, 77, 253, 193, 250, 77, 253, 193, 252, 77, 253, 193, 254, 77, 253, 193, 2, 78, 253, 193, 12, 78, 253, 193, 14, 78, 253, 193, 16, 78, 253, 193, 18, 78, 253, 193, 118, 76, 253, 193, 142, 76, 253, 193, 194, 76, 253, 193, 10, 78, 253, 193, 100, 78, 253, 193, 140, 78, 253, 193, 150, 78, 253, 193, 154, 78, 253, 193, 168, 78, 253, 193, 170, 78, 253, 193, 238, 78, 253, 193, 240, 78, 253, 193, 88, 79, 253, 193, 90, 79, 253, 193, 92, 79, 253, 193, 94, 79, 253, 193, 218, 79, 253, 193, 220, 79, 253, 193, 116, 80, 253, 193, 118, 80, 253, 193, 120, 80, 253, 193, 122, 80, 253, 193, 124, 80, 253, 193, 126, 80, 253, 193, 128, 80, 253, 193, 18, 81, 253, 193, 20, 81, 253, 193, 22, 81, 253, 193, 178, 81, 253, 193, 180, 81, 253, 193, 182, 81, 253, 193, 184, 81, 253, 206, 64, 96, 0, 206, 64, 96, 0, 206, 64, 96, 0, 206, 64, 96, 0, 193, 2, 77, 253, 193, 4, 77, 253, 193, 6, 77, 253, 193, 26, 77, 253, 193, 28, 77, 253, 193, 56, 77, 253, 193, 58, 77, 253, 193, 60, 77, 253, 193, 62, 77, 253, 193, 64, 77, 253, 193, 66, 77, 253, 193, 68, 77, 253, 193, 70, 77, 253, 193, 72, 77, 253, 193, 100, 77, 253, 193, 102, 77, 253, 193, 104, 77, 253, 193, 106, 77, 253, 193, 122, 77, 253, 193, 124, 77, 253, 193, 136, 77, 253, 193, 138, 77, 253, 193, 148, 77, 253, 193, 150, 77, 253, 193, 152, 77, 253, 193, 154, 77, 253, 193, 156, 77, 253, 193, 168, 77, 253, 206, 96, 96, 0, 206, 96, 96, 0, 206, 96, 96, 0, 206, 96, 96, 0, 193, 8, 77, 253, 193, 10, 77, 253, 193, 12, 77, 253, 193, 30, 77, 253, 193, 32, 77, 253, 193, 34, 77, 253, 193, 36, 77, 253, 193, 38, 77, 253, 193, 40, 77, 253, 193, 42, 77, 253, 193, 44, 77, 253, 193, 74, 77, 253, 193, 76, 77, 253, 193, 78, 77, 253, 193, 80, 77, 253, 193, 82, 77, 253, 193, 84, 77, 253, 193, 108, 77, 253, 193, 110, 77, 253, 193, 112, 77, 253, 193, 126, 77, 253, 193, 140, 77, 253, 193, 142, 77, 253, 193, 186, 77, 253, 193, 208, 77, 253, 193, 210, 77, 253, 193, 224, 77, 253, 193, 226, 77, 253, 193, 170, 77, 253, 193, 174, 77, 253, 193, 178, 77, 253, 193, 180, 77, 253, 193, 182, 77, 253, 193, 198, 77, 253, 193, 206, 77, 253, 193, 242, 77, 253, 193, 22, 78, 253, 193, 52, 78, 253, 193, 82, 78, 253, 193, 96, 78, 253, 193, 114, 78, 253, 193, 148, 78, 253, 193, 210, 78, 253, 193, 30, 79, 253, 193, 32, 79, 253, 193, 34, 79, 253, 193, 36, 79, 253, 193, 38, 79, 253, 193, 176, 79, 253, 193, 178, 79, 253, 193, 68, 80, 253, 193, 206, 80, 253, 193, 208, 80, 253, 193, 210, 80, 253, 193, 212, 80, 253, 193, 132, 81, 253, 193, 134, 81, 253, 193, 136, 81, 253, 193, 100, 82, 253, 193, 102, 82, 253, 193, 228, 77, 253, 193, 244, 77, 253, 193, 246, 77, 253, 193, 248, 77, 253, 193, 8, 78, 253, 193, 24, 78, 253, 193, 26, 78, 253, 193, 28, 78, 253, 193, 30, 78, 253, 193, 32, 78, 253, 193, 34, 78, 253, 193, 54, 78, 253, 193, 72, 78, 253, 193, 84, 78, 253, 193, 98, 78, 253, 193, 110, 78, 253, 193, 128, 78, 253, 193, 136, 78, 253, 193, 138, 78, 253, 193, 152, 78, 253, 193, 164, 78, 253, 193, 176, 78, 253, 193, 178, 78, 253, 193, 182, 78, 253, 193, 186, 78, 253, 193, 188, 78, 253, 193, 190, 78, 253, 193, 192, 78, 253, 193, 194, 78, 253, 193, 196, 78, 253, 193, 214, 78, 253, 193, 216, 78, 253, 193, 122, 78, 253, 193, 124, 78, 253, 193, 126, 78, 253, 193, 130, 78, 253, 193, 132, 78, 253, 193, 142, 78, 253, 193, 144, 78, 253, 193, 146, 78, 253, 193, 156, 78, 253, 193, 158, 78, 253, 193, 166, 78, 253, 193, 172, 78, 253, 193, 174, 78, 253, 193, 198, 78, 253, 193, 200, 78, 253, 193, 202, 78, 253, 193, 204, 78, 253, 193, 242, 78, 253, 193, 244, 78, 253, 193, 246, 78, 253, 193, 248, 78, 253, 193, 250, 78, 253, 193, 252, 78, 253, 193, 254, 78, 253, 193, 2, 79, 253, 193, 4, 79, 253, 193, 6, 79, 253, 193, 8, 79, 253, 193, 10, 79, 253, 193, 12, 79, 253, 193, 14, 79, 253, 193, 16, 79, 253, 193, 162, 78, 253, 193, 24, 79, 253, 193, 26, 79, 253, 193, 28, 79, 253, 193, 172, 79, 253, 193, 174, 79, 253, 193, 50, 80, 253, 193, 52, 80, 253, 193, 54, 80, 253, 193, 56, 80, 253, 193, 58, 80, 253, 193, 60, 80, 253, 193, 62, 80, 253, 193, 64, 80, 253, 193, 66, 80, 253, 193, 200, 80, 253, 193, 202, 80, 253, 193, 204, 80, 253, 193, 124, 81, 253, 193, 126, 81, 253, 193, 128, 81, 253, 193, 130, 81, 253, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 193, 218, 78, 253, 193, 220, 78, 253, 193, 222, 78, 253, 193, 224, 78, 253, 193, 226, 78, 253, 193, 228, 78, 253, 193, 230, 78, 253, 193, 232, 78, 253, 193, 234, 78, 253, 193, 236, 78, 253, 193, 40, 79, 253, 193, 42, 79, 253, 193, 44, 79, 253, 193, 46, 79, 253, 193, 48, 79, 253, 193, 50, 79, 253, 193, 52, 79, 253, 193, 54, 79, 253, 193, 56, 79, 253, 193, 58, 79, 253, 193, 60, 79, 253, 193, 62, 79, 253, 193, 64, 79, 253, 193, 66, 79, 253, 193, 68, 79, 253, 193, 70, 79, 253, 193, 72, 79, 253, 193, 74, 79, 253, 193, 76, 79, 253, 193, 78, 79, 253, 193, 80, 79, 253, 193, 82, 79, 253, 193, 18, 79, 253, 193, 20, 79, 253, 193, 96, 79, 253, 193, 98, 79, 253, 193, 100, 79, 253, 193, 102, 79, 253, 193, 104, 79, 253, 193, 106, 79, 253, 193, 108, 79, 253, 193, 110, 79, 253, 193, 112, 79, 253, 193, 114, 79, 253, 193, 116, 79, 253, 193, 118, 79, 253, 193, 120, 79, 253, 193, 122, 79, 253, 193, 124, 79, 253, 193, 126, 79, 253, 193, 128, 79, 253, 193, 130, 79, 253, 193, 132, 79, 253, 193, 134, 79, 253, 193, 136, 79, 253, 193, 138, 79, 253, 193, 140, 79, 253, 193, 142, 79, 253, 193, 144, 79, 253, 193, 146, 79, 253, 193, 148, 79, 253, 193, 150, 79, 253, 193, 152, 79, 253, 193, 154, 79, 253, 193, 180, 79, 253, 193, 182, 79, 253, 193, 184, 79, 253, 193, 186, 79, 253, 193, 188, 79, 253, 193, 190, 79, 253, 193, 192, 79, 253, 193, 194, 79, 253, 193, 196, 79, 253, 193, 198, 79, 253, 193, 200, 79, 253, 193, 202, 79, 253, 193, 84, 79, 253, 193, 204, 79, 253, 193, 206, 79, 253, 193, 208, 79, 253, 193, 210, 79, 253, 193, 212, 79, 253, 193, 214, 79, 253, 193, 216, 79, 253, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 193, 156, 79, 253, 193, 158, 79, 253, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 193, 160, 79, 253, 193, 162, 79, 253, 193, 38, 80, 253, 193, 40, 80, 253, 193, 42, 80, 253, 193, 194, 80, 253, 193, 100, 81, 253, 193, 102, 81, 253, 193, 104, 81, 253, 193, 106, 81, 253, 193, 108, 81, 253, 193, 34, 82, 253, 193, 36, 82, 253, 193, 38, 82, 253, 193, 40, 82, 253, 193, 42, 82, 253, 193, 44, 82, 253, 193, 46, 82, 253, 193, 48, 82, 253, 193, 48, 83, 253, 193, 50, 83, 253, 193, 52, 83, 253, 193, 54, 83, 253, 193, 56, 83, 253, 193, 58, 83, 253, 193, 60, 83, 253, 193, 230, 83, 253, 193, 232, 83, 253, 193, 234, 83, 253, 193, 236, 83, 253, 193, 146, 84, 253, 193, 148, 84, 253, 193, 170, 79, 253, 193, 44, 80, 253, 193, 46, 80, 253, 193, 48, 80, 253, 193, 196, 80, 253, 193, 198, 80, 253, 193, 110, 81, 253, 193, 112, 81, 253, 193, 114, 81, 253, 193, 116, 81, 253, 193, 118, 81, 253, 193, 120, 81, 253, 193, 122, 81, 253, 193, 50, 82, 253, 193, 52, 82, 253, 193, 54, 82, 253, 193, 56, 82, 253, 193, 58, 82, 253, 193, 60, 82, 253, 193, 66, 83, 253, 193, 68, 83, 253, 193, 70, 83, 253, 193, 72, 83, 253, 193, 74, 83, 253, 193, 76, 83, 253, 193, 78, 83, 253, 193, 238, 83, 253, 193, 240, 83, 253, 193, 242, 83, 253, 193, 244, 83, 253, 193, 246, 83, 253, 193, 248, 83, 253, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 128, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 160, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 192, 96, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 206, 224, 96, 0, 193, 138, 81, 253, 193, 140, 81, 253, 193, 142, 81, 253, 193, 144, 81, 253, 193, 146, 81, 253, 193, 148, 81, 253, 193, 150, 81, 253, 193, 152, 81, 253, 193, 154, 81, 253, 193, 156, 81, 253, 193, 158, 81, 253, 193, 160, 81, 253, 193, 162, 81, 253, 193, 164, 81, 253, 193, 166, 81, 253, 193, 168, 81, 253, 193, 170, 81, 253, 193, 172, 81, 253, 193, 174, 81, 253, 193, 176, 81, 253, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 0, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 193, 186, 81, 253, 193, 172, 82, 253, 193, 140, 83, 253, 193, 142, 83, 253, 193, 144, 83, 253, 193, 146, 83, 253, 193, 46, 84, 253, 193, 48, 84, 253, 193, 50, 84, 253, 193, 52, 84, 253, 193, 230, 84, 253, 193, 232, 84, 253, 193, 234, 84, 253, 193, 236, 84, 253, 193, 152, 85, 253, 193, 24, 86, 253, 193, 26, 86, 253, 193, 122, 86, 253, 193, 70, 87, 253, 193, 106, 87, 253, 193, 136, 87, 253, 193, 200, 87, 253, 193, 202, 87, 253, 193, 204, 87, 253, 193, 66, 88, 253, 193, 68, 88, 253, 193, 70, 88, 253, 193, 206, 88, 253, 193, 208, 88, 253, 193, 210, 88, 253, 193, 212, 88, 253, 193, 214, 88, 253, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 206, 32, 97, 0, 193, 174, 82, 253, 193, 4, 82, 253, 193, 6, 82, 253, 193, 8, 82, 253, 193, 10, 82, 253, 193, 12, 82, 253, 193, 14, 82, 253, 193, 16, 82, 253, 193, 18, 82, 253, 193, 20, 82, 253, 193, 22, 82, 253, 193, 24, 82, 253, 193, 26, 82, 253, 193, 28, 82, 253, 193, 30, 82, 253, 193, 176, 82, 253, 193, 178, 82, 253, 193, 180, 82, 253, 193, 32, 82, 253, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 206, 64, 97, 0, 193, 80, 83, 253, 193, 82, 83, 253, 193, 84, 83, 253, 193, 86, 83, 253, 193, 88, 83, 253, 193, 90, 83, 253, 193, 92, 83, 253, 193, 94, 83, 253, 193, 96, 83, 253, 193, 98, 83, 253, 193, 100, 83, 253, 193, 102, 83, 253, 193, 104, 83, 253, 193, 250, 83, 253, 193, 252, 83, 253, 193, 254, 83, 253, 193, 2, 84, 253, 193, 4, 84, 253, 193, 6, 84, 253, 193, 8, 84, 253, 193, 176, 84, 253, 193, 178, 84, 253, 193, 180, 84, 253, 193, 104, 82, 253, 193, 106, 82, 253, 193, 108, 82, 253, 193, 110, 82, 253, 193, 112, 82, 253, 193, 114, 82, 253, 193, 106, 83, 253, 193, 108, 83, 253, 193, 110, 83, 253, 193, 10, 84, 253, 193, 12, 84, 253, 193, 14, 84, 253, 193, 16, 84, 253, 193, 18, 84, 253, 193, 200, 84, 253, 193, 202, 84, 253, 193, 204, 84, 253, 193, 206, 84, 253, 193, 124, 85, 253, 193, 126, 85, 253, 193, 128, 85, 253, 193, 130, 85, 253, 193, 10, 86, 253, 193, 12, 86, 253, 193, 14, 86, 253, 193, 16, 86, 253, 193, 18, 86, 253, 193, 230, 86, 253, 193, 52, 87, 253, 193, 94, 87, 253, 193, 96, 87, 253, 193, 166, 87, 253, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 96, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 206, 128, 97, 0, 193, 148, 83, 253, 193, 150, 83, 253, 193, 152, 83, 253, 206, 160, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 206, 160, 97, 0, 193, 20, 84, 253, 193, 22, 84, 253, 193, 24, 84, 253, 193, 26, 84, 253, 193, 28, 84, 253, 193, 30, 84, 253, 193, 32, 84, 253, 193, 34, 84, 253, 193, 36, 84, 253, 193, 38, 84, 253, 193, 40, 84, 253, 193, 42, 84, 253, 193, 208, 84, 253, 193, 44, 84, 253, 193, 210, 84, 253, 193, 212, 84, 253, 193, 214, 84, 253, 193, 216, 84, 253, 193, 218, 84, 253, 193, 220, 84, 253, 193, 222, 84, 253, 193, 224, 84, 253, 193, 226, 84, 253, 193, 132, 85, 253, 193, 134, 85, 253, 193, 136, 85, 253, 193, 154, 83, 253, 193, 156, 83, 253, 193, 158, 83, 253, 193, 160, 83, 253, 193, 162, 83, 253, 193, 164, 83, 253, 193, 166, 83, 253, 193, 168, 83, 253, 193, 170, 83, 253, 193, 172, 83, 253, 193, 174, 83, 253, 193, 176, 83, 253, 193, 178, 83, 253, 193, 180, 83, 253, 193, 182, 83, 253, 193, 184, 83, 253, 193, 186, 83, 253, 193, 188, 83, 253, 193, 190, 83, 253, 193, 192, 83, 253, 193, 194, 83, 253, 193, 196, 83, 253, 193, 198, 83, 253, 193, 200, 83, 253, 193, 202, 83, 253, 193, 204, 83, 253, 193, 206, 83, 253, 193, 208, 83, 253, 193, 210, 83, 253, 193, 212, 83, 253, 193, 214, 83, 253, 193, 216, 83, 253, 193, 218, 83, 253, 193, 220, 83, 253, 193, 222, 83, 253, 193, 224, 83, 253, 193, 226, 83, 253, 193, 228, 83, 253, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 192, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 206, 224, 97, 0, 193, 150, 84, 253, 193, 152, 84, 253, 193, 154, 84, 253, 193, 156, 84, 253, 193, 80, 85, 253, 193, 82, 85, 253, 193, 84, 85, 253, 193, 86, 85, 253, 193, 88, 85, 253, 193, 90, 85, 253, 193, 92, 85, 253, 193, 222, 85, 253, 193, 82, 86, 253, 193, 84, 86, 253, 193, 86, 86, 253, 193, 88, 86, 253, 193, 168, 86, 253, 193, 170, 86, 253, 193, 12, 87, 253, 193, 14, 87, 253, 193, 16, 87, 253, 193, 18, 87, 253, 193, 40, 87, 253, 193, 160, 87, 253, 193, 10, 88, 253, 193, 12, 88, 253, 193, 128, 88, 253, 193, 130, 88, 253, 193, 24, 89, 253, 193, 164, 89, 253, 193, 166, 89, 253, 193, 168, 89, 253, 193, 158, 84, 253, 193, 160, 84, 253, 193, 162, 84, 253, 193, 164, 84, 253, 193, 166, 84, 253, 193, 168, 84, 253, 193, 170, 84, 253, 193, 172, 84, 253, 193, 174, 84, 253, 193, 94, 85, 253, 193, 96, 85, 253, 193, 98, 85, 253, 193, 100, 85, 253, 193, 102, 85, 253, 193, 104, 85, 253, 193, 106, 85, 253, 193, 108, 85, 253, 193, 226, 85, 253, 193, 228, 85, 253, 193, 230, 85, 253, 193, 232, 85, 253, 193, 234, 85, 253, 193, 236, 85, 253, 193, 238, 85, 253, 193, 240, 85, 253, 193, 242, 85, 253, 193, 244, 85, 253, 193, 246, 85, 253, 206, 64, 98, 0, 206, 64, 98, 0, 206, 64, 98, 0, 206, 64, 98, 0, 193, 182, 84, 253, 193, 184, 84, 253, 193, 186, 84, 253, 193, 188, 84, 253, 193, 190, 84, 253, 193, 192, 84, 253, 193, 194, 84, 253, 193, 196, 84, 253, 193, 198, 84, 253, 193, 110, 85, 253, 193, 112, 85, 253, 193, 114, 85, 253, 193, 116, 85, 253, 193, 118, 85, 253, 193, 120, 85, 253, 193, 122, 85, 253, 193, 248, 85, 253, 193, 250, 85, 253, 193, 252, 85, 253, 193, 254, 85, 253, 193, 2, 86, 253, 193, 4, 86, 253, 193, 6, 86, 253, 193, 8, 86, 253, 193, 106, 86, 253, 193, 108, 86, 253, 193, 110, 86, 253, 193, 112, 86, 253, 193, 114, 86, 253, 193, 186, 86, 253, 193, 188, 86, 253, 193, 190, 86, 253, 193, 138, 85, 253, 193, 140, 85, 253, 193, 142, 85, 253, 193, 144, 85, 253, 193, 146, 85, 253, 193, 148, 85, 253, 193, 228, 84, 253, 193, 20, 86, 253, 193, 22, 86, 253, 193, 150, 85, 253, 193, 116, 86, 253, 193, 118, 86, 253, 193, 120, 86, 253, 193, 196, 86, 253, 193, 232, 86, 253, 193, 54, 87, 253, 193, 56, 87, 253, 193, 98, 87, 253, 193, 102, 87, 253, 193, 104, 87, 253, 193, 116, 87, 253, 193, 118, 87, 253, 193, 120, 87, 253, 193, 122, 87, 253, 193, 124, 87, 253, 193, 126, 87, 253, 193, 128, 87, 253, 193, 130, 87, 253, 193, 132, 87, 253, 193, 134, 87, 253, 193, 170, 87, 253, 193, 172, 87, 253, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 0, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 206, 32, 98, 0, 193, 28, 86, 253, 193, 30, 86, 253, 193, 32, 86, 253, 193, 34, 86, 253, 193, 36, 86, 253, 193, 38, 86, 253, 193, 40, 86, 253, 193, 42, 86, 253, 193, 44, 86, 253, 193, 46, 86, 253, 193, 48, 86, 253, 193, 50, 86, 253, 193, 52, 86, 253, 193, 54, 86, 253, 193, 56, 86, 253, 193, 58, 86, 253, 193, 60, 86, 253, 193, 62, 86, 253, 193, 64, 86, 253, 193, 66, 86, 253, 193, 68, 86, 253, 193, 70, 86, 253, 193, 72, 86, 253, 193, 74, 86, 253, 193, 76, 86, 253, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 193, 80, 86, 253, 193, 132, 137, 253, 193, 134, 137, 253, 193, 136, 137, 253, 193, 138, 137, 253, 193, 140, 137, 253, 193, 142, 137, 253, 193, 144, 137, 253, 193, 146, 137, 253, 193, 148, 137, 253, 193, 150, 137, 253, 193, 152, 137, 253, 193, 154, 137, 253, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 64, 98, 0, 206, 64, 98, 0, 206, 64, 98, 0, 206, 64, 98, 0, 193, 172, 86, 253, 193, 174, 86, 253, 193, 176, 86, 253, 193, 178, 86, 253, 193, 180, 86, 253, 193, 182, 86, 253, 193, 184, 86, 253, 193, 218, 86, 253, 193, 220, 86, 253, 193, 222, 86, 253, 193, 224, 86, 253, 193, 42, 87, 253, 193, 44, 87, 253, 193, 46, 87, 253, 193, 48, 87, 253, 193, 112, 87, 253, 193, 114, 87, 253, 193, 14, 88, 253, 193, 16, 88, 253, 193, 134, 88, 253, 193, 136, 88, 253, 193, 138, 88, 253, 193, 26, 89, 253, 193, 28, 89, 253, 193, 30, 89, 253, 193, 32, 89, 253, 193, 170, 89, 253, 193, 172, 89, 253, 193, 192, 86, 253, 193, 194, 86, 253, 193, 226, 86, 253, 193, 228, 86, 253, 193, 20, 87, 253, 193, 22, 87, 253, 193, 24, 87, 253, 193, 50, 87, 253, 193, 86, 87, 253, 193, 90, 87, 253, 193, 162, 87, 253, 193, 164, 87, 253, 193, 18, 88, 253, 193, 20, 88, 253, 193, 22, 88, 253, 193, 140, 88, 253, 193, 142, 88, 253, 193, 144, 88, 253, 193, 146, 88, 253, 193, 148, 88, 253, 193, 150, 88, 253, 193, 152, 88, 253, 193, 154, 88, 253, 193, 34, 89, 253, 193, 36, 89, 253, 193, 38, 89, 253, 193, 40, 89, 253, 193, 184, 89, 253, 193, 186, 89, 253, 193, 78, 90, 253, 193, 80, 90, 253, 193, 82, 90, 253, 193, 214, 86, 253, 193, 216, 86, 253, 193, 234, 86, 253, 193, 236, 86, 253, 193, 238, 86, 253, 193, 240, 86, 253, 193, 242, 86, 253, 193, 244, 86, 253, 193, 246, 86, 253, 193, 248, 86, 253, 193, 250, 86, 253, 193, 252, 86, 253, 193, 254, 86, 253, 193, 2, 87, 253, 193, 4, 87, 253, 193, 6, 87, 253, 193, 8, 87, 253, 193, 10, 87, 253, 193, 26, 87, 253, 193, 28, 87, 253, 193, 30, 87, 253, 193, 32, 87, 253, 193, 34, 87, 253, 193, 36, 87, 253, 193, 38, 87, 253, 193, 58, 87, 253, 193, 60, 87, 253, 193, 62, 87, 253, 193, 64, 87, 253, 193, 66, 87, 253, 193, 68, 87, 253, 193, 72, 87, 253, 193, 74, 87, 253, 193, 76, 87, 253, 193, 78, 87, 253, 193, 80, 87, 253, 193, 82, 87, 253, 193, 84, 87, 253, 193, 88, 87, 253, 193, 92, 87, 253, 193, 100, 87, 253, 193, 108, 87, 253, 193, 110, 87, 253, 193, 138, 87, 253, 193, 140, 87, 253, 193, 142, 87, 253, 193, 144, 87, 253, 193, 146, 87, 253, 193, 148, 87, 253, 193, 150, 87, 253, 193, 152, 87, 253, 193, 154, 87, 253, 193, 156, 87, 253, 193, 158, 87, 253, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 193, 168, 87, 253, 193, 24, 88, 253, 193, 26, 88, 253, 193, 28, 88, 253, 193, 30, 88, 253, 193, 32, 88, 253, 193, 34, 88, 253, 193, 156, 88, 253, 193, 158, 88, 253, 193, 42, 89, 253, 193, 188, 89, 253, 193, 190, 89, 253, 193, 192, 89, 253, 193, 84, 90, 253, 193, 86, 90, 253, 193, 88, 90, 253, 193, 90, 90, 253, 193, 92, 90, 253, 193, 228, 90, 253, 193, 230, 90, 253, 193, 76, 91, 253, 193, 78, 91, 253, 193, 80, 91, 253, 193, 82, 91, 253, 193, 84, 91, 253, 193, 206, 91, 253, 193, 208, 91, 253, 193, 210, 91, 253, 193, 76, 92, 253, 193, 158, 92, 253, 193, 212, 92, 253, 193, 20, 93, 253, 193, 174, 87, 253, 193, 176, 87, 253, 193, 178, 87, 253, 193, 180, 87, 253, 193, 182, 87, 253, 193, 184, 87, 253, 193, 186, 87, 253, 193, 188, 87, 253, 193, 190, 87, 253, 193, 192, 87, 253, 193, 194, 87, 253, 193, 196, 87, 253, 193, 198, 87, 253, 193, 36, 88, 253, 193, 38, 88, 253, 193, 40, 88, 253, 193, 42, 88, 253, 193, 44, 88, 253, 193, 46, 88, 253, 193, 48, 88, 253, 193, 50, 88, 253, 193, 52, 88, 253, 193, 54, 88, 253, 193, 56, 88, 253, 193, 58, 88, 253, 193, 60, 88, 253, 193, 62, 88, 253, 193, 64, 88, 253, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 96, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 128, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 206, 160, 98, 0, 193, 44, 89, 253, 193, 204, 88, 253, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 193, 216, 88, 253, 193, 218, 88, 253, 193, 220, 88, 253, 193, 86, 89, 253, 193, 88, 89, 253, 193, 90, 89, 253, 193, 92, 89, 253, 193, 94, 89, 253, 193, 96, 89, 253, 193, 98, 89, 253, 193, 228, 89, 253, 193, 230, 89, 253, 193, 232, 89, 253, 193, 234, 89, 253, 193, 236, 89, 253, 193, 238, 89, 253, 193, 130, 90, 253, 193, 132, 90, 253, 193, 134, 90, 253, 193, 4, 91, 253, 193, 6, 91, 253, 193, 8, 91, 253, 193, 10, 91, 253, 193, 12, 91, 253, 193, 114, 91, 253, 193, 116, 91, 253, 193, 118, 91, 253, 193, 120, 91, 253, 193, 122, 91, 253, 193, 238, 91, 253, 193, 240, 91, 253, 193, 242, 91, 253, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 192, 98, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 206, 224, 98, 0, 193, 194, 89, 253, 193, 196, 89, 253, 193, 198, 89, 253, 193, 200, 89, 253, 193, 202, 89, 253, 193, 204, 89, 253, 193, 206, 89, 253, 193, 208, 89, 253, 193, 210, 89, 253, 193, 212, 89, 253, 193, 214, 89, 253, 193, 216, 89, 253, 193, 218, 89, 253, 193, 220, 89, 253, 193, 222, 89, 253, 193, 224, 89, 253, 193, 226, 89, 253, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 0, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 193, 174, 89, 253, 193, 176, 89, 253, 193, 178, 89, 253, 193, 180, 89, 253, 193, 182, 89, 253, 193, 66, 90, 253, 193, 68, 90, 253, 193, 70, 90, 253, 193, 72, 90, 253, 193, 74, 90, 253, 193, 76, 90, 253, 193, 210, 90, 253, 193, 212, 90, 253, 193, 214, 90, 253, 193, 216, 90, 253, 193, 218, 90, 253, 193, 220, 90, 253, 193, 196, 91, 253, 193, 198, 91, 253, 193, 200, 91, 253, 193, 202, 91, 253, 193, 66, 92, 253, 193, 68, 92, 253, 193, 138, 92, 253, 193, 140, 92, 253, 193, 142, 92, 253, 193, 144, 92, 253, 193, 146, 92, 253, 193, 206, 92, 253, 193, 10, 93, 253, 193, 12, 93, 253, 193, 14, 93, 253, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 206, 32, 99, 0, 193, 136, 90, 253, 193, 138, 90, 253, 193, 140, 90, 253, 193, 142, 90, 253, 193, 144, 90, 253, 193, 146, 90, 253, 193, 148, 90, 253, 193, 150, 90, 253, 193, 14, 91, 253, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 193, 58, 90, 253, 193, 60, 90, 253, 193, 62, 90, 253, 193, 204, 90, 253, 193, 206, 90, 253, 193, 66, 91, 253, 193, 68, 91, 253, 193, 190, 91, 253, 193, 192, 91, 253, 193, 194, 91, 253, 193, 134, 92, 253, 193, 136, 92, 253, 193, 204, 92, 253, 193, 8, 93, 253, 193, 52, 93, 253, 193, 54, 93, 253, 193, 56, 93, 253, 193, 106, 93, 253, 193, 144, 93, 253, 193, 190, 93, 253, 193, 208, 94, 253, 193, 210, 94, 253, 193, 48, 95, 253, 193, 50, 95, 253, 193, 52, 95, 253, 193, 164, 95, 253, 193, 146, 96, 253, 193, 234, 96, 253, 193, 68, 97, 253, 193, 70, 97, 253, 193, 166, 97, 253, 193, 222, 97, 253, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 206, 64, 99, 0, 193, 232, 90, 253, 193, 234, 90, 253, 193, 236, 90, 253, 193, 238, 90, 253, 193, 240, 90, 253, 193, 242, 90, 253, 193, 244, 90, 253, 193, 246, 90, 253, 193, 248, 90, 253, 193, 250, 90, 253, 193, 252, 90, 253, 193, 254, 90, 253, 193, 2, 91, 253, 206, 160, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 96, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 193, 222, 90, 253, 193, 224, 90, 253, 193, 226, 90, 253, 193, 70, 91, 253, 193, 72, 91, 253, 193, 74, 91, 253, 193, 204, 91, 253, 193, 70, 92, 253, 193, 72, 92, 253, 193, 74, 92, 253, 193, 148, 92, 253, 193, 150, 92, 253, 193, 152, 92, 253, 193, 154, 92, 253, 193, 156, 92, 253, 193, 208, 92, 253, 193, 210, 92, 253, 193, 16, 93, 253, 193, 18, 93, 253, 193, 90, 93, 253, 193, 108, 93, 253, 193, 150, 93, 253, 193, 160, 93, 253, 193, 162, 93, 253, 193, 192, 93, 253, 193, 194, 93, 253, 193, 196, 93, 253, 193, 30, 94, 253, 193, 32, 94, 253, 193, 34, 94, 253, 193, 36, 94, 253, 193, 38, 94, 253, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 128, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 206, 160, 99, 0, 193, 212, 91, 253, 193, 214, 91, 253, 193, 216, 91, 253, 193, 218, 91, 253, 193, 220, 91, 253, 193, 222, 91, 253, 193, 224, 91, 253, 193, 226, 91, 253, 193, 228, 91, 253, 193, 230, 91, 253, 193, 232, 91, 253, 193, 234, 91, 253, 193, 236, 91, 253, 193, 78, 92, 253, 193, 80, 92, 253, 193, 82, 92, 253, 193, 84, 92, 253, 193, 86, 92, 253, 193, 160, 92, 253, 193, 162, 92, 253, 193, 214, 92, 253, 193, 216, 92, 253, 193, 218, 92, 253, 193, 220, 92, 253, 193, 222, 92, 253, 193, 22, 93, 253, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 206, 192, 99, 0, 193, 246, 91, 253, 193, 248, 91, 253, 193, 250, 91, 253, 193, 252, 91, 253, 193, 254, 91, 253, 193, 2, 92, 253, 193, 4, 92, 253, 193, 6, 92, 253, 193, 8, 92, 253, 193, 10, 92, 253, 193, 12, 92, 253, 193, 14, 92, 253, 193, 16, 92, 253, 193, 18, 92, 253, 193, 188, 91, 253, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 193, 244, 91, 253, 193, 88, 92, 253, 193, 164, 92, 253, 193, 122, 93, 253, 193, 124, 93, 253, 193, 156, 93, 253, 193, 158, 93, 253, 193, 178, 93, 253, 193, 180, 93, 253, 193, 224, 93, 253, 193, 226, 93, 253, 193, 228, 93, 253, 193, 230, 93, 253, 193, 232, 93, 253, 193, 58, 94, 253, 193, 60, 94, 253, 193, 62, 94, 253, 193, 64, 94, 253, 193, 66, 94, 253, 193, 68, 94, 253, 193, 158, 94, 253, 193, 160, 94, 253, 193, 162, 94, 253, 193, 164, 94, 253, 193, 166, 94, 253, 193, 168, 94, 253, 193, 2, 95, 253, 193, 4, 95, 253, 193, 98, 95, 253, 193, 100, 95, 253, 193, 102, 95, 253, 193, 104, 95, 253, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 224, 99, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 206, 0, 100, 0, 193, 166, 92, 253, 193, 168, 92, 253, 193, 170, 92, 253, 193, 172, 92, 253, 193, 174, 92, 253, 193, 176, 92, 253, 193, 178, 92, 253, 193, 180, 92, 253, 193, 182, 92, 253, 193, 184, 92, 253, 193, 186, 92, 253, 193, 188, 92, 253, 193, 190, 92, 253, 193, 192, 92, 253, 193, 194, 92, 253, 193, 196, 92, 253, 193, 198, 92, 253, 193, 200, 92, 253, 193, 202, 92, 253, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 206, 32, 100, 0, 193, 28, 93, 253, 193, 30, 93, 253, 193, 32, 93, 253, 193, 34, 93, 253, 193, 36, 93, 253, 193, 38, 93, 253, 193, 40, 93, 253, 193, 42, 93, 253, 193, 44, 93, 253, 193, 46, 93, 253, 193, 48, 93, 253, 193, 50, 93, 253, 193, 72, 93, 253, 193, 74, 93, 253, 193, 76, 93, 253, 193, 78, 93, 253, 193, 80, 93, 253, 193, 82, 93, 253, 193, 84, 93, 253, 193, 24, 93, 253, 193, 68, 93, 253, 193, 26, 93, 253, 193, 70, 93, 253, 193, 92, 93, 253, 193, 94, 93, 253, 193, 96, 93, 253, 193, 110, 93, 253, 193, 112, 93, 253, 193, 148, 93, 253, 193, 152, 93, 253, 193, 154, 93, 253, 193, 164, 93, 253, 193, 166, 93, 253, 193, 168, 93, 253, 193, 170, 93, 253, 193, 172, 93, 253, 193, 174, 93, 253, 193, 176, 93, 253, 193, 200, 93, 253, 193, 202, 93, 253, 193, 204, 93, 253, 193, 206, 93, 253, 193, 208, 93, 253, 193, 210, 93, 253, 193, 212, 93, 253, 193, 214, 93, 253, 193, 216, 93, 253, 193, 218, 93, 253, 193, 220, 93, 253, 193, 222, 93, 253, 193, 44, 94, 253, 193, 58, 93, 253, 193, 60, 93, 253, 193, 62, 93, 253, 193, 86, 93, 253, 193, 88, 93, 253, 193, 140, 93, 253, 193, 26, 94, 253, 193, 28, 94, 253, 193, 120, 94, 253, 193, 214, 94, 253, 193, 54, 95, 253, 193, 56, 95, 253, 193, 58, 95, 253, 193, 28, 96, 253, 193, 30, 96, 253, 193, 32, 96, 253, 193, 34, 96, 253, 193, 36, 96, 253, 193, 148, 96, 253, 193, 150, 96, 253, 193, 152, 96, 253, 193, 154, 96, 253, 193, 236, 96, 253, 193, 130, 97, 253, 193, 132, 97, 253, 193, 134, 97, 253, 193, 36, 98, 253, 193, 56, 98, 253, 193, 136, 98, 253, 193, 138, 98, 253, 193, 146, 98, 253, 193, 192, 98, 253, 193, 64, 93, 253, 193, 66, 93, 253, 193, 146, 93, 253, 193, 198, 93, 253, 193, 42, 94, 253, 193, 130, 94, 253, 193, 132, 94, 253, 193, 66, 95, 253, 193, 68, 95, 253, 193, 70, 95, 253, 193, 180, 95, 253, 193, 182, 95, 253, 193, 184, 95, 253, 193, 52, 96, 253, 193, 54, 96, 253, 193, 158, 96, 253, 193, 246, 96, 253, 193, 248, 96, 253, 193, 250, 96, 253, 193, 80, 97, 253, 193, 82, 97, 253, 193, 176, 97, 253, 193, 202, 97, 253, 193, 48, 98, 253, 193, 100, 98, 253, 193, 122, 98, 253, 193, 176, 98, 253, 193, 178, 98, 253, 193, 180, 98, 253, 193, 182, 98, 253, 193, 196, 98, 253, 193, 120, 99, 253, 193, 98, 93, 253, 193, 100, 93, 253, 193, 102, 93, 253, 193, 104, 93, 253, 193, 114, 93, 253, 193, 116, 93, 253, 193, 118, 93, 253, 193, 120, 93, 253, 193, 126, 93, 253, 193, 128, 93, 253, 193, 130, 93, 253, 193, 132, 93, 253, 193, 134, 93, 253, 193, 136, 93, 253, 193, 138, 93, 253, 193, 142, 93, 253, 193, 182, 93, 253, 193, 184, 93, 253, 193, 186, 93, 253, 193, 188, 93, 253, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 64, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 193, 40, 94, 253, 193, 122, 94, 253, 193, 124, 94, 253, 193, 126, 94, 253, 193, 128, 94, 253, 193, 216, 94, 253, 193, 218, 94, 253, 193, 220, 94, 253, 193, 222, 94, 253, 193, 224, 94, 253, 193, 226, 94, 253, 193, 60, 95, 253, 193, 62, 95, 253, 193, 64, 95, 253, 193, 166, 95, 253, 193, 168, 95, 253, 193, 170, 95, 253, 193, 172, 95, 253, 193, 174, 95, 253, 193, 176, 95, 253, 193, 178, 95, 253, 193, 38, 96, 253, 193, 40, 96, 253, 193, 42, 96, 253, 193, 44, 96, 253, 193, 46, 96, 253, 193, 48, 96, 253, 193, 50, 96, 253, 193, 156, 96, 253, 193, 238, 96, 253, 193, 240, 96, 253, 193, 242, 96, 253, 193, 46, 94, 253, 193, 48, 94, 253, 193, 50, 94, 253, 193, 52, 94, 253, 193, 54, 94, 253, 193, 56, 94, 253, 193, 134, 94, 253, 193, 136, 94, 253, 193, 138, 94, 253, 193, 140, 94, 253, 193, 142, 94, 253, 193, 144, 94, 253, 193, 146, 94, 253, 193, 148, 94, 253, 193, 150, 94, 253, 193, 152, 94, 253, 193, 154, 94, 253, 193, 156, 94, 253, 193, 228, 94, 253, 193, 230, 94, 253, 193, 232, 94, 253, 193, 234, 94, 253, 193, 236, 94, 253, 193, 238, 94, 253, 193, 240, 94, 253, 193, 242, 94, 253, 193, 244, 94, 253, 193, 246, 94, 253, 193, 248, 94, 253, 193, 250, 94, 253, 193, 252, 94, 253, 193, 254, 94, 253, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 206, 96, 100, 0, 193, 170, 94, 253, 193, 172, 94, 253, 193, 174, 94, 253, 193, 176, 94, 253, 193, 178, 94, 253, 193, 180, 94, 253, 193, 182, 94, 253, 193, 184, 94, 253, 193, 186, 94, 253, 193, 188, 94, 253, 193, 190, 94, 253, 193, 192, 94, 253, 193, 194, 94, 253, 193, 196, 94, 253, 193, 198, 94, 253, 193, 200, 94, 253, 193, 202, 94, 253, 193, 204, 94, 253, 193, 206, 94, 253, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 128, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 193, 72, 95, 253, 193, 74, 95, 253, 193, 76, 95, 253, 193, 78, 95, 253, 193, 80, 95, 253, 193, 82, 95, 253, 193, 84, 95, 253, 193, 86, 95, 253, 193, 88, 95, 253, 193, 90, 95, 253, 193, 92, 95, 253, 193, 94, 95, 253, 193, 96, 95, 253, 193, 186, 95, 253, 193, 188, 95, 253, 193, 190, 95, 253, 193, 192, 95, 253, 193, 194, 95, 253, 193, 196, 95, 253, 193, 198, 95, 253, 193, 200, 95, 253, 193, 202, 95, 253, 193, 204, 95, 253, 193, 206, 95, 253, 206, 224, 100, 0, 206, 224, 100, 0, 206, 224, 100, 0, 206, 224, 100, 0, 206, 224, 100, 0, 206, 224, 100, 0, 206, 224, 100, 0, 206, 224, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 160, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 193, 208, 95, 253, 193, 210, 95, 253, 193, 212, 95, 253, 193, 214, 95, 253, 193, 216, 95, 253, 193, 218, 95, 253, 193, 220, 95, 253, 193, 222, 95, 253, 193, 224, 95, 253, 193, 82, 96, 253, 193, 84, 96, 253, 193, 86, 96, 253, 193, 88, 96, 253, 193, 176, 96, 253, 193, 8, 97, 253, 193, 10, 97, 253, 193, 12, 97, 253, 193, 14, 97, 253, 193, 16, 97, 253, 193, 146, 97, 253, 193, 148, 97, 253, 193, 178, 97, 253, 193, 180, 97, 253, 193, 232, 97, 253, 193, 234, 97, 253, 193, 18, 98, 253, 193, 112, 98, 253, 193, 24, 99, 253, 193, 88, 99, 253, 193, 220, 99, 253, 193, 222, 99, 253, 193, 224, 99, 253, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 206, 192, 100, 0, 193, 90, 96, 253, 193, 92, 96, 253, 193, 94, 96, 253, 193, 96, 96, 253, 193, 98, 96, 253, 193, 100, 96, 253, 193, 102, 96, 253, 193, 104, 96, 253, 193, 106, 96, 253, 193, 108, 96, 253, 193, 110, 96, 253, 193, 112, 96, 253, 193, 114, 96, 253, 193, 116, 96, 253, 193, 118, 96, 253, 193, 120, 96, 253, 193, 122, 96, 253, 193, 124, 96, 253, 193, 126, 96, 253, 193, 128, 96, 253, 193, 130, 96, 253, 193, 132, 96, 253, 193, 134, 96, 253, 193, 136, 96, 253, 193, 138, 96, 253, 193, 140, 96, 253, 193, 142, 96, 253, 206, 224, 100, 0, 206, 224, 100, 0, 206, 224, 100, 0, 206, 224, 100, 0, 206, 224, 100, 0, 193, 160, 96, 253, 193, 162, 96, 253, 193, 164, 96, 253, 193, 166, 96, 253, 193, 168, 96, 253, 193, 170, 96, 253, 193, 172, 96, 253, 193, 174, 96, 253, 193, 252, 96, 253, 193, 254, 96, 253, 193, 2, 97, 253, 193, 4, 97, 253, 193, 6, 97, 253, 193, 84, 97, 253, 193, 86, 97, 253, 193, 88, 97, 253, 193, 90, 97, 253, 193, 92, 97, 253, 193, 94, 97, 253, 193, 96, 97, 253, 193, 140, 97, 253, 193, 142, 97, 253, 193, 144, 97, 253, 193, 204, 97, 253, 193, 224, 97, 253, 193, 226, 97, 253, 193, 228, 97, 253, 193, 144, 96, 253, 193, 178, 96, 253, 193, 180, 96, 253, 193, 182, 96, 253, 193, 184, 96, 253, 193, 186, 96, 253, 193, 188, 96, 253, 193, 190, 96, 253, 193, 192, 96, 253, 193, 194, 96, 253, 193, 196, 96, 253, 193, 198, 96, 253, 193, 200, 96, 253, 193, 202, 96, 253, 193, 204, 96, 253, 193, 206, 96, 253, 193, 208, 96, 253, 193, 210, 96, 253, 193, 212, 96, 253, 193, 214, 96, 253, 193, 216, 96, 253, 193, 218, 96, 253, 193, 220, 96, 253, 193, 222, 96, 253, 193, 224, 96, 253, 193, 226, 96, 253, 193, 228, 96, 253, 193, 230, 96, 253, 193, 232, 96, 253, 193, 18, 97, 253, 193, 20, 97, 253, 193, 22, 97, 253, 193, 244, 96, 253, 193, 72, 97, 253, 193, 74, 97, 253, 193, 76, 97, 253, 193, 78, 97, 253, 193, 136, 97, 253, 193, 138, 97, 253, 193, 168, 97, 253, 193, 170, 97, 253, 193, 172, 97, 253, 193, 174, 97, 253, 193, 198, 97, 253, 193, 200, 97, 253, 193, 246, 97, 253, 193, 38, 98, 253, 193, 46, 98, 253, 193, 74, 98, 253, 193, 86, 98, 253, 193, 120, 98, 253, 193, 140, 98, 253, 193, 148, 98, 253, 193, 150, 98, 253, 193, 152, 98, 253, 193, 158, 98, 253, 193, 160, 98, 253, 193, 168, 98, 253, 193, 174, 98, 253, 193, 208, 98, 253, 193, 212, 98, 253, 193, 214, 98, 253, 193, 222, 98, 253, 193, 224, 98, 253, 193, 24, 97, 253, 193, 26, 97, 253, 193, 28, 97, 253, 193, 30, 97, 253, 193, 32, 97, 253, 193, 34, 97, 253, 193, 36, 97, 253, 193, 38, 97, 253, 193, 40, 97, 253, 193, 42, 97, 253, 193, 44, 97, 253, 193, 46, 97, 253, 193, 48, 97, 253, 193, 50, 97, 253, 193, 52, 97, 253, 193, 54, 97, 253, 193, 56, 97, 253, 193, 58, 97, 253, 193, 60, 97, 253, 193, 62, 97, 253, 193, 64, 97, 253, 193, 66, 97, 253, 206, 0, 101, 0, 206, 0, 101, 0, 206, 0, 101, 0, 206, 0, 101, 0, 206, 0, 101, 0, 206, 0, 101, 0, 206, 0, 101, 0, 206, 0, 101, 0, 206, 0, 101, 0, 206, 0, 101, 0, 206, 0, 101, 0, 206, 0, 101, 0, 193, 150, 97, 253, 193, 152, 97, 253, 193, 154, 97, 253, 193, 156, 97, 253, 193, 158, 97, 253, 193, 160, 97, 253, 193, 162, 97, 253, 193, 164, 97, 253, 193, 182, 97, 253, 193, 184, 97, 253, 193, 186, 97, 253, 193, 188, 97, 253, 193, 190, 97, 253, 193, 192, 97, 253, 193, 194, 97, 253, 193, 196, 97, 253, 193, 206, 97, 253, 193, 208, 97, 253, 193, 210, 97, 253, 193, 212, 97, 253, 193, 214, 97, 253, 193, 216, 97, 253, 193, 218, 97, 253, 193, 220, 97, 253, 193, 236, 97, 253, 193, 238, 97, 253, 193, 230, 97, 253, 193, 248, 97, 253, 193, 250, 97, 253, 193, 2, 98, 253, 193, 4, 98, 253, 193, 10, 98, 253, 193, 30, 98, 253, 193, 32, 98, 253, 193, 40, 98, 253, 193, 50, 98, 253, 193, 52, 98, 253, 193, 88, 98, 253, 193, 98, 98, 253, 193, 102, 98, 253, 193, 110, 98, 253, 193, 124, 98, 253, 193, 198, 98, 253, 193, 242, 98, 253, 193, 244, 98, 253, 193, 246, 98, 253, 193, 248, 98, 253, 193, 250, 98, 253, 193, 252, 98, 253, 193, 254, 98, 253, 193, 2, 99, 253, 193, 4, 99, 253, 193, 6, 99, 253, 193, 8, 99, 253, 193, 10, 99, 253, 193, 12, 99, 253, 193, 14, 99, 253, 193, 16, 99, 253, 193, 240, 97, 253, 193, 242, 97, 253, 193, 244, 97, 253, 193, 252, 97, 253, 193, 254, 97, 253, 193, 6, 98, 253, 193, 8, 98, 253, 193, 12, 98, 253, 193, 14, 98, 253, 193, 16, 98, 253, 193, 20, 98, 253, 193, 22, 98, 253, 193, 24, 98, 253, 193, 26, 98, 253, 193, 28, 98, 253, 193, 34, 98, 253, 193, 42, 98, 253, 193, 44, 98, 253, 193, 58, 98, 253, 193, 60, 98, 253, 193, 62, 98, 253, 193, 64, 98, 253, 193, 66, 98, 253, 193, 68, 98, 253, 193, 70, 98, 253, 193, 72, 98, 253, 193, 78, 98, 253, 193, 80, 98, 253, 193, 82, 98, 253, 193, 84, 98, 253, 193, 90, 98, 253, 193, 92, 98, 253, 193, 54, 98, 253, 193, 118, 98, 253, 193, 134, 98, 253, 193, 156, 98, 253, 193, 190, 98, 253, 193, 210, 98, 253, 193, 106, 99, 253, 193, 108, 99, 253, 193, 110, 99, 253, 193, 112, 99, 253, 193, 28, 100, 253, 193, 30, 100, 253, 193, 32, 100, 253, 193, 34, 100, 253, 193, 36, 100, 253, 193, 24, 101, 253, 193, 26, 101, 253, 193, 28, 101, 253, 193, 30, 101, 253, 193, 32, 101, 253, 193, 34, 101, 253, 193, 22, 102, 253, 193, 24, 102, 253, 193, 26, 102, 253, 193, 28, 102, 253, 193, 30, 102, 253, 193, 32, 102, 253, 193, 34, 102, 253, 193, 36, 102, 253, 193, 98, 103, 253, 193, 100, 103, 253, 193, 102, 103, 253, 193, 94, 98, 253, 193, 96, 98, 253, 193, 114, 98, 253, 193, 106, 98, 253, 193, 108, 98, 253, 193, 116, 98, 253, 193, 126, 98, 253, 193, 128, 98, 253, 193, 130, 98, 253, 193, 132, 98, 253, 193, 142, 98, 253, 193, 144, 98, 253, 193, 154, 98, 253, 193, 162, 98, 253, 193, 164, 98, 253, 193, 166, 98, 253, 193, 170, 98, 253, 193, 172, 98, 253, 193, 186, 98, 253, 193, 188, 98, 253, 193, 200, 98, 253, 193, 202, 98, 253, 193, 204, 98, 253, 193, 206, 98, 253, 193, 218, 98, 253, 193, 232, 98, 253, 193, 234, 98, 253, 193, 238, 98, 253, 193, 240, 98, 253, 193, 26, 99, 253, 193, 28, 99, 253, 193, 30, 99, 253, 193, 194, 98, 253, 193, 220, 98, 253, 193, 38, 100, 253, 193, 40, 100, 253, 193, 42, 100, 253, 193, 44, 100, 253, 193, 36, 101, 253, 193, 38, 101, 253, 193, 40, 101, 253, 193, 42, 101, 253, 193, 44, 101, 253, 193, 46, 101, 253, 193, 48, 101, 253, 193, 40, 102, 253, 193, 42, 102, 253, 193, 44, 102, 253, 193, 46, 102, 253, 193, 48, 102, 253, 193, 50, 102, 253, 193, 52, 102, 253, 193, 116, 103, 253, 193, 118, 103, 253, 193, 120, 103, 253, 193, 122, 103, 253, 193, 124, 103, 253, 193, 126, 103, 253, 193, 128, 103, 253, 193, 130, 103, 253, 193, 132, 103, 253, 193, 134, 103, 253, 193, 136, 103, 253, 193, 138, 103, 253, 193, 226, 98, 253, 193, 228, 98, 253, 193, 230, 98, 253, 193, 236, 98, 253, 193, 114, 99, 253, 193, 116, 99, 253, 193, 118, 99, 253, 193, 46, 100, 253, 193, 48, 100, 253, 193, 50, 100, 253, 193, 52, 100, 253, 193, 54, 100, 253, 193, 56, 100, 253, 193, 58, 100, 253, 193, 50, 101, 253, 193, 52, 101, 253, 193, 54, 101, 253, 193, 54, 102, 253, 193, 56, 102, 253, 193, 58, 102, 253, 193, 60, 102, 253, 193, 62, 102, 253, 193, 64, 102, 253, 193, 66, 102, 253, 193, 142, 103, 253, 193, 144, 103, 253, 193, 146, 103, 253, 193, 148, 103, 253, 193, 150, 103, 253, 193, 152, 103, 253, 193, 154, 103, 253, 193, 156, 103, 253, 193, 18, 99, 253, 193, 20, 99, 253, 193, 22, 99, 253, 193, 36, 99, 253, 193, 38, 99, 253, 193, 40, 99, 253, 193, 42, 99, 253, 193, 44, 99, 253, 193, 46, 99, 253, 193, 48, 99, 253, 193, 50, 99, 253, 193, 52, 99, 253, 193, 54, 99, 253, 193, 56, 99, 253, 193, 58, 99, 253, 193, 60, 99, 253, 193, 62, 99, 253, 193, 64, 99, 253, 193, 66, 99, 253, 193, 68, 99, 253, 193, 70, 99, 253, 193, 72, 99, 253, 193, 74, 99, 253, 193, 76, 99, 253, 193, 78, 99, 253, 193, 80, 99, 253, 193, 82, 99, 253, 193, 84, 99, 253, 193, 86, 99, 253, 193, 124, 99, 253, 193, 126, 99, 253, 193, 128, 99, 253, 193, 32, 99, 253, 193, 34, 99, 253, 193, 90, 99, 253, 193, 92, 99, 253, 193, 94, 99, 253, 193, 96, 99, 253, 193, 98, 99, 253, 193, 100, 99, 253, 193, 102, 99, 253, 193, 104, 99, 253, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 193, 122, 99, 253, 193, 60, 100, 253, 193, 62, 100, 253, 193, 64, 100, 253, 193, 56, 101, 253, 193, 58, 101, 253, 193, 60, 101, 253, 193, 68, 102, 253, 193, 70, 102, 253, 193, 72, 102, 253, 193, 74, 102, 253, 193, 76, 102, 253, 193, 78, 102, 253, 193, 162, 103, 253, 193, 164, 103, 253, 193, 166, 103, 253, 193, 168, 103, 253, 193, 170, 103, 253, 193, 172, 103, 253, 193, 174, 103, 253, 193, 176, 103, 253, 193, 80, 105, 253, 193, 82, 105, 253, 193, 84, 105, 253, 193, 86, 105, 253, 193, 88, 105, 253, 193, 90, 105, 253, 193, 214, 106, 253, 193, 216, 106, 253, 193, 218, 106, 253, 193, 220, 106, 253, 193, 222, 106, 253, 193, 130, 99, 253, 193, 132, 99, 253, 193, 134, 99, 253, 193, 136, 99, 253, 193, 138, 99, 253, 193, 140, 99, 253, 193, 142, 99, 253, 193, 144, 99, 253, 193, 146, 99, 253, 193, 148, 99, 253, 193, 150, 99, 253, 193, 152, 99, 253, 193, 154, 99, 253, 193, 156, 99, 253, 193, 158, 99, 253, 193, 160, 99, 253, 193, 162, 99, 253, 193, 164, 99, 253, 193, 166, 99, 253, 193, 168, 99, 253, 193, 170, 99, 253, 193, 172, 99, 253, 193, 174, 99, 253, 193, 176, 99, 253, 193, 178, 99, 253, 193, 180, 99, 253, 193, 182, 99, 253, 193, 184, 99, 253, 193, 186, 99, 253, 193, 188, 99, 253, 193, 190, 99, 253, 193, 192, 99, 253, 193, 194, 99, 253, 193, 196, 99, 253, 193, 198, 99, 253, 193, 200, 99, 253, 193, 202, 99, 253, 193, 204, 99, 253, 193, 206, 99, 253, 193, 208, 99, 253, 193, 210, 99, 253, 193, 212, 99, 253, 193, 214, 99, 253, 193, 216, 99, 253, 193, 218, 99, 253, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 193, 226, 99, 253, 193, 214, 100, 253, 193, 216, 100, 253, 193, 218, 100, 253, 193, 180, 101, 253, 193, 182, 101, 253, 193, 184, 101, 253, 193, 186, 101, 253, 193, 188, 101, 253, 193, 190, 101, 253, 193, 192, 101, 253, 193, 220, 102, 253, 193, 222, 102, 253, 193, 224, 102, 253, 193, 226, 102, 253, 193, 228, 102, 253, 193, 230, 102, 253, 193, 232, 102, 253, 193, 234, 102, 253, 193, 90, 104, 253, 193, 92, 104, 253, 193, 94, 104, 253, 193, 96, 104, 253, 193, 222, 105, 253, 193, 224, 105, 253, 193, 226, 105, 253, 193, 228, 105, 253, 193, 230, 105, 253, 193, 232, 105, 253, 193, 86, 107, 253, 193, 88, 107, 253, 193, 90, 107, 253, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 206, 32, 101, 0, 193, 220, 100, 253, 193, 222, 100, 253, 193, 224, 100, 253, 193, 226, 100, 253, 193, 228, 100, 253, 193, 230, 100, 253, 193, 232, 100, 253, 193, 234, 100, 253, 193, 236, 100, 253, 193, 238, 100, 253, 193, 240, 100, 253, 193, 242, 100, 253, 193, 244, 100, 253, 193, 246, 100, 253, 193, 248, 100, 253, 193, 250, 100, 253, 193, 252, 100, 253, 193, 254, 100, 253, 193, 2, 101, 253, 193, 4, 101, 253, 193, 6, 101, 253, 193, 8, 101, 253, 193, 10, 101, 253, 193, 12, 101, 253, 193, 14, 101, 253, 193, 16, 101, 253, 193, 18, 101, 253, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 206, 64, 101, 0, 193, 62, 101, 253, 193, 64, 101, 253, 193, 66, 101, 253, 193, 68, 101, 253, 193, 70, 101, 253, 193, 72, 101, 253, 193, 74, 101, 253, 193, 178, 103, 253, 193, 76, 101, 253, 193, 78, 101, 253, 193, 80, 101, 253, 193, 82, 101, 253, 193, 84, 101, 253, 193, 86, 101, 253, 193, 88, 101, 253, 193, 90, 101, 253, 193, 92, 101, 253, 193, 94, 101, 253, 193, 96, 101, 253, 193, 98, 101, 253, 193, 100, 101, 253, 193, 102, 101, 253, 193, 104, 101, 253, 193, 106, 101, 253, 193, 108, 101, 253, 193, 110, 101, 253, 193, 210, 100, 253, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 96, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 193, 252, 101, 253, 193, 254, 101, 253, 193, 2, 102, 253, 193, 4, 102, 253, 193, 6, 102, 253, 193, 8, 102, 253, 193, 10, 102, 253, 193, 12, 102, 253, 193, 14, 102, 253, 193, 16, 102, 253, 193, 18, 102, 253, 193, 20, 102, 253, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 128, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 160, 101, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 193, 104, 103, 253, 193, 106, 103, 253, 193, 108, 103, 253, 193, 110, 103, 253, 193, 112, 103, 253, 193, 114, 103, 253, 193, 244, 104, 253, 193, 246, 104, 253, 193, 248, 104, 253, 193, 250, 104, 253, 193, 252, 104, 253, 193, 254, 104, 253, 193, 2, 105, 253, 193, 4, 105, 253, 193, 6, 105, 253, 193, 8, 105, 253, 193, 156, 106, 253, 193, 158, 106, 253, 193, 160, 106, 253, 193, 162, 106, 253, 193, 164, 106, 253, 193, 166, 106, 253, 193, 168, 106, 253, 193, 170, 106, 253, 193, 252, 107, 253, 193, 254, 107, 253, 193, 2, 108, 253, 193, 4, 108, 253, 193, 6, 108, 253, 193, 8, 108, 253, 193, 10, 108, 253, 193, 66, 109, 253, 193, 140, 103, 253, 193, 10, 105, 253, 193, 12, 105, 253, 193, 14, 105, 253, 193, 16, 105, 253, 193, 18, 105, 253, 193, 20, 105, 253, 193, 22, 105, 253, 193, 24, 105, 253, 193, 26, 105, 253, 193, 28, 105, 253, 193, 30, 105, 253, 193, 32, 105, 253, 193, 172, 106, 253, 193, 174, 106, 253, 193, 176, 106, 253, 193, 178, 106, 253, 193, 180, 106, 253, 193, 182, 106, 253, 193, 14, 108, 253, 193, 16, 108, 253, 193, 18, 108, 253, 193, 20, 108, 253, 193, 22, 108, 253, 193, 24, 108, 253, 193, 26, 108, 253, 193, 28, 108, 253, 193, 30, 108, 253, 193, 32, 108, 253, 193, 34, 108, 253, 193, 36, 108, 253, 193, 76, 109, 253, 193, 158, 103, 253, 193, 160, 103, 253, 193, 34, 105, 253, 193, 36, 105, 253, 193, 38, 105, 253, 193, 40, 105, 253, 193, 42, 105, 253, 193, 44, 105, 253, 193, 46, 105, 253, 193, 48, 105, 253, 193, 50, 105, 253, 193, 52, 105, 253, 193, 54, 105, 253, 193, 56, 105, 253, 193, 58, 105, 253, 193, 60, 105, 253, 193, 62, 105, 253, 193, 64, 105, 253, 193, 66, 105, 253, 193, 68, 105, 253, 193, 70, 105, 253, 193, 72, 105, 253, 193, 74, 105, 253, 193, 76, 105, 253, 193, 78, 105, 253, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 206, 192, 101, 0, 193, 92, 105, 253, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 206, 224, 101, 0, 193, 94, 105, 253, 193, 84, 104, 253, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 0, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 32, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 64, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 206, 96, 102, 0, 193, 38, 108, 253, 193, 40, 108, 253, 193, 42, 108, 253, 193, 44, 108, 253, 193, 46, 108, 253, 193, 48, 108, 253, 193, 50, 108, 253, 193, 52, 108, 253, 193, 54, 108, 253, 193, 56, 108, 253, 193, 58, 108, 253, 193, 60, 108, 253, 193, 88, 109, 253, 193, 90, 109, 253, 193, 92, 109, 253, 193, 94, 109, 253, 193, 96, 109, 253, 193, 98, 109, 253, 193, 100, 109, 253, 193, 102, 109, 253, 193, 104, 109, 253, 193, 106, 109, 253, 193, 156, 110, 253, 193, 158, 110, 253, 193, 224, 106, 253, 193, 226, 106, 253, 193, 62, 108, 253, 193, 64, 108, 253, 193, 108, 109, 253, 193, 110, 109, 253, 193, 190, 110, 253, 193, 192, 110, 253, 193, 194, 110, 253, 193, 196, 110, 253, 193, 198, 110, 253, 193, 200, 110, 253, 193, 202, 110, 253, 193, 194, 111, 253, 193, 106, 112, 253, 193, 108, 112, 253, 193, 110, 112, 253, 193, 112, 112, 253, 193, 8, 113, 253, 193, 68, 114, 253, 193, 84, 114, 253, 193, 106, 114, 253, 193, 108, 114, 253, 193, 150, 114, 253, 193, 152, 114, 253, 193, 154, 114, 253, 193, 156, 114, 253, 193, 158, 114, 253, 206, 0, 104, 0, 206, 0, 104, 0, 206, 0, 104, 0, 206, 0, 104, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 206, 128, 102, 0, 193, 66, 108, 253, 193, 40, 107, 253, 193, 42, 107, 253, 193, 44, 107, 253, 193, 46, 107, 253, 193, 48, 107, 253, 193, 50, 107, 253, 193, 52, 107, 253, 193, 54, 107, 253, 193, 56, 107, 253, 193, 58, 107, 253, 193, 60, 107, 253, 193, 62, 107, 253, 193, 64, 107, 253, 193, 66, 107, 253, 193, 68, 107, 253, 193, 70, 107, 253, 193, 72, 107, 253, 193, 74, 107, 253, 193, 76, 107, 253, 193, 78, 107, 253, 193, 80, 107, 253, 193, 82, 107, 253, 193, 92, 107, 253, 193, 94, 107, 253, 193, 96, 107, 253, 193, 98, 107, 253, 193, 100, 107, 253, 193, 102, 107, 253, 193, 104, 107, 253, 193, 106, 107, 253, 193, 172, 108, 253, 193, 174, 108, 253, 193, 176, 108, 253, 193, 178, 108, 253, 193, 180, 108, 253, 193, 182, 108, 253, 193, 184, 108, 253, 193, 186, 108, 253, 193, 188, 108, 253, 193, 190, 108, 253, 193, 192, 108, 253, 193, 194, 108, 253, 193, 216, 109, 253, 193, 218, 109, 253, 193, 220, 109, 253, 193, 222, 109, 253, 193, 224, 109, 253, 193, 226, 109, 253, 193, 228, 109, 253, 193, 230, 109, 253, 193, 232, 109, 253, 193, 2, 111, 253, 193, 4, 111, 253, 193, 6, 111, 253, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 160, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 206, 192, 102, 0, 193, 112, 109, 253, 193, 204, 110, 253, 193, 114, 109, 253, 193, 116, 109, 253, 193, 118, 109, 253, 193, 120, 109, 253, 193, 170, 108, 253, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 206, 224, 102, 0, 193, 234, 109, 253, 193, 236, 109, 253, 193, 238, 109, 253, 193, 240, 109, 253, 193, 242, 109, 253, 193, 244, 109, 253, 193, 246, 109, 253, 193, 248, 109, 253, 193, 22, 111, 253, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 193, 68, 109, 253, 193, 70, 109, 253, 193, 72, 109, 253, 193, 74, 109, 253, 193, 122, 110, 253, 193, 124, 110, 253, 193, 126, 110, 253, 193, 128, 110, 253, 193, 130, 110, 253, 193, 132, 110, 253, 193, 134, 110, 253, 193, 136, 110, 253, 193, 138, 110, 253, 193, 140, 110, 253, 193, 142, 110, 253, 193, 146, 111, 253, 193, 148, 111, 253, 193, 150, 111, 253, 193, 152, 111, 253, 193, 154, 111, 253, 193, 156, 111, 253, 193, 158, 111, 253, 193, 74, 112, 253, 193, 76, 112, 253, 193, 78, 112, 253, 193, 80, 112, 253, 193, 82, 112, 253, 193, 236, 112, 253, 193, 238, 112, 253, 193, 84, 113, 253, 193, 86, 113, 253, 193, 88, 113, 253, 193, 78, 109, 253, 193, 80, 109, 253, 193, 82, 109, 253, 193, 84, 109, 253, 193, 86, 109, 253, 193, 144, 110, 253, 193, 146, 110, 253, 193, 148, 110, 253, 193, 150, 110, 253, 193, 152, 110, 253, 193, 154, 110, 253, 193, 160, 111, 253, 193, 162, 111, 253, 193, 164, 111, 253, 193, 166, 111, 253, 193, 168, 111, 253, 193, 170, 111, 253, 193, 172, 111, 253, 193, 84, 112, 253, 193, 86, 112, 253, 193, 88, 112, 253, 193, 90, 112, 253, 193, 92, 112, 253, 193, 94, 112, 253, 193, 240, 112, 253, 193, 242, 112, 253, 193, 244, 112, 253, 193, 246, 112, 253, 193, 90, 113, 253, 193, 92, 113, 253, 193, 94, 113, 253, 193, 160, 113, 253, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 206, 0, 103, 0, 193, 206, 110, 253, 193, 208, 110, 253, 193, 210, 110, 253, 193, 212, 110, 253, 193, 214, 110, 253, 193, 216, 110, 253, 193, 218, 110, 253, 193, 220, 110, 253, 193, 222, 110, 253, 193, 224, 110, 253, 193, 226, 110, 253, 193, 228, 110, 253, 193, 230, 110, 253, 193, 232, 110, 253, 193, 234, 110, 253, 193, 236, 110, 253, 193, 238, 110, 253, 193, 240, 110, 253, 193, 242, 110, 253, 193, 244, 110, 253, 193, 246, 110, 253, 193, 248, 110, 253, 193, 250, 110, 253, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 32, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 193, 160, 110, 253, 193, 162, 110, 253, 193, 164, 110, 253, 193, 166, 110, 253, 193, 168, 110, 253, 193, 170, 110, 253, 193, 172, 110, 253, 193, 174, 110, 253, 193, 176, 110, 253, 193, 178, 110, 253, 193, 180, 110, 253, 193, 182, 110, 253, 193, 184, 110, 253, 193, 186, 110, 253, 193, 188, 110, 253, 193, 174, 111, 253, 193, 176, 111, 253, 193, 178, 111, 253, 193, 180, 111, 253, 193, 182, 111, 253, 193, 184, 111, 253, 193, 186, 111, 253, 193, 188, 111, 253, 193, 190, 111, 253, 193, 192, 111, 253, 193, 96, 112, 253, 193, 98, 112, 253, 193, 100, 112, 253, 193, 102, 112, 253, 193, 104, 112, 253, 193, 248, 112, 253, 193, 250, 112, 253, 193, 252, 110, 253, 193, 254, 110, 253, 193, 196, 111, 253, 193, 198, 111, 253, 193, 200, 111, 253, 193, 202, 111, 253, 193, 204, 111, 253, 193, 206, 111, 253, 193, 208, 111, 253, 193, 210, 111, 253, 193, 212, 111, 253, 193, 214, 111, 253, 193, 216, 111, 253, 193, 218, 111, 253, 193, 220, 111, 253, 193, 222, 111, 253, 193, 224, 111, 253, 193, 226, 111, 253, 193, 228, 111, 253, 193, 230, 111, 253, 193, 232, 111, 253, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 193, 8, 111, 253, 193, 10, 111, 253, 193, 12, 111, 253, 193, 14, 111, 253, 193, 16, 111, 253, 193, 18, 111, 253, 193, 20, 111, 253, 193, 236, 111, 253, 193, 238, 111, 253, 193, 240, 111, 253, 193, 242, 111, 253, 193, 244, 111, 253, 193, 246, 111, 253, 193, 248, 111, 253, 193, 250, 111, 253, 193, 150, 112, 253, 193, 152, 112, 253, 193, 154, 112, 253, 193, 156, 112, 253, 193, 158, 112, 253, 193, 160, 112, 253, 193, 20, 113, 253, 193, 22, 113, 253, 193, 24, 113, 253, 193, 120, 113, 253, 193, 46, 114, 253, 193, 208, 114, 253, 193, 210, 114, 253, 193, 212, 114, 253, 193, 88, 115, 253, 193, 90, 115, 253, 193, 92, 115, 253, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 64, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 206, 128, 103, 0, 193, 234, 111, 253, 193, 10, 113, 253, 193, 12, 113, 253, 193, 14, 113, 253, 193, 16, 113, 253, 193, 18, 113, 253, 193, 106, 113, 253, 193, 108, 113, 253, 193, 110, 113, 253, 193, 112, 113, 253, 193, 114, 113, 253, 193, 116, 113, 253, 193, 118, 113, 253, 193, 168, 113, 253, 193, 170, 113, 253, 193, 172, 113, 253, 193, 174, 113, 253, 193, 176, 113, 253, 193, 178, 113, 253, 193, 180, 113, 253, 193, 234, 113, 253, 193, 236, 113, 253, 193, 238, 113, 253, 193, 240, 113, 253, 193, 242, 113, 253, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 96, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 206, 160, 103, 0, 193, 26, 113, 253, 193, 28, 113, 253, 193, 30, 113, 253, 193, 252, 112, 253, 193, 254, 112, 253, 193, 2, 113, 253, 193, 4, 113, 253, 193, 6, 113, 253, 193, 96, 113, 253, 193, 98, 113, 253, 193, 100, 113, 253, 193, 102, 113, 253, 193, 104, 113, 253, 193, 164, 113, 253, 193, 166, 113, 253, 193, 230, 113, 253, 193, 232, 113, 253, 193, 14, 114, 253, 193, 40, 114, 253, 193, 76, 114, 253, 193, 82, 114, 253, 193, 148, 114, 253, 193, 4, 115, 253, 193, 140, 116, 253, 193, 24, 117, 253, 193, 214, 117, 253, 193, 138, 118, 253, 193, 214, 118, 253, 193, 244, 118, 253, 193, 246, 118, 253, 193, 26, 119, 253, 193, 62, 119, 253, 193, 64, 119, 253, 193, 82, 119, 253, 193, 84, 119, 253, 193, 32, 113, 253, 193, 34, 113, 253, 193, 36, 113, 253, 193, 38, 113, 253, 193, 40, 113, 253, 193, 42, 113, 253, 193, 44, 113, 253, 193, 46, 113, 253, 193, 48, 113, 253, 193, 50, 113, 253, 193, 52, 113, 253, 193, 54, 113, 253, 193, 56, 113, 253, 193, 58, 113, 253, 193, 60, 113, 253, 193, 62, 113, 253, 193, 64, 113, 253, 193, 66, 113, 253, 193, 68, 113, 253, 193, 70, 113, 253, 193, 72, 113, 253, 193, 74, 113, 253, 193, 76, 113, 253, 193, 78, 113, 253, 193, 80, 113, 253, 193, 82, 113, 253, 206, 192, 103, 0, 206, 192, 103, 0, 206, 192, 103, 0, 206, 192, 103, 0, 206, 192, 103, 0, 206, 192, 103, 0, 206, 192, 103, 0, 206, 192, 103, 0, 206, 192, 103, 0, 206, 192, 103, 0, 206, 192, 103, 0, 193, 182, 113, 253, 193, 184, 113, 253, 193, 186, 113, 253, 193, 188, 113, 253, 193, 190, 113, 253, 193, 192, 113, 253, 193, 194, 113, 253, 193, 196, 113, 253, 193, 198, 113, 253, 193, 200, 113, 253, 193, 202, 113, 253, 193, 204, 113, 253, 193, 206, 113, 253, 193, 208, 113, 253, 193, 210, 113, 253, 193, 212, 113, 253, 193, 214, 113, 253, 193, 216, 113, 253, 193, 218, 113, 253, 193, 220, 113, 253, 193, 222, 113, 253, 193, 244, 113, 253, 193, 246, 113, 253, 193, 248, 113, 253, 193, 250, 113, 253, 193, 252, 113, 253, 193, 254, 113, 253, 193, 2, 114, 253, 193, 4, 114, 253, 193, 6, 114, 253, 193, 8, 114, 253, 193, 10, 114, 253, 193, 22, 114, 253, 193, 24, 114, 253, 193, 26, 114, 253, 193, 28, 114, 253, 193, 30, 114, 253, 193, 32, 114, 253, 193, 34, 114, 253, 193, 36, 114, 253, 193, 152, 113, 253, 193, 48, 114, 253, 193, 50, 114, 253, 193, 52, 114, 253, 193, 54, 114, 253, 193, 56, 114, 253, 193, 60, 114, 253, 193, 62, 114, 253, 193, 64, 114, 253, 193, 70, 114, 253, 193, 74, 114, 253, 193, 78, 114, 253, 193, 80, 114, 253, 193, 132, 114, 253, 193, 134, 114, 253, 193, 154, 113, 253, 193, 156, 113, 253, 193, 158, 113, 253, 193, 224, 113, 253, 193, 12, 114, 253, 193, 102, 114, 253, 193, 136, 114, 253, 193, 220, 114, 253, 193, 222, 114, 253, 193, 224, 114, 253, 193, 226, 114, 253, 193, 228, 114, 253, 193, 102, 115, 253, 193, 104, 115, 253, 193, 106, 115, 253, 193, 246, 115, 253, 193, 248, 115, 253, 193, 250, 115, 253, 193, 252, 115, 253, 193, 254, 115, 253, 193, 2, 116, 253, 193, 118, 116, 253, 193, 120, 116, 253, 193, 122, 116, 253, 193, 234, 116, 253, 193, 236, 116, 253, 193, 238, 116, 253, 193, 240, 116, 253, 193, 242, 116, 253, 193, 244, 116, 253, 193, 246, 116, 253, 193, 248, 116, 253, 193, 162, 113, 253, 193, 226, 113, 253, 193, 228, 113, 253, 193, 38, 114, 253, 193, 58, 114, 253, 193, 66, 114, 253, 193, 72, 114, 253, 193, 140, 114, 253, 193, 142, 114, 253, 193, 144, 114, 253, 193, 146, 114, 253, 193, 234, 114, 253, 193, 236, 114, 253, 193, 238, 114, 253, 193, 240, 114, 253, 193, 242, 114, 253, 193, 244, 114, 253, 193, 246, 114, 253, 193, 248, 114, 253, 193, 250, 114, 253, 193, 252, 114, 253, 193, 254, 114, 253, 193, 2, 115, 253, 193, 112, 115, 253, 193, 114, 115, 253, 193, 116, 115, 253, 193, 118, 115, 253, 193, 120, 115, 253, 193, 122, 115, 253, 193, 124, 115, 253, 193, 126, 115, 253, 193, 128, 115, 253, 193, 16, 114, 253, 193, 18, 114, 253, 193, 20, 114, 253, 193, 42, 114, 253, 193, 44, 114, 253, 193, 86, 114, 253, 193, 88, 114, 253, 193, 90, 114, 253, 193, 92, 114, 253, 193, 94, 114, 253, 193, 96, 114, 253, 193, 98, 114, 253, 193, 100, 114, 253, 193, 110, 114, 253, 193, 112, 114, 253, 193, 114, 114, 253, 193, 116, 114, 253, 193, 118, 114, 253, 193, 120, 114, 253, 193, 122, 114, 253, 193, 124, 114, 253, 193, 126, 114, 253, 193, 128, 114, 253, 193, 130, 114, 253, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 224, 103, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 193, 214, 114, 253, 193, 216, 114, 253, 193, 218, 114, 253, 193, 94, 115, 253, 193, 96, 115, 253, 193, 98, 115, 253, 193, 100, 115, 253, 193, 114, 116, 253, 193, 116, 116, 253, 193, 226, 116, 253, 193, 228, 116, 253, 193, 230, 116, 253, 193, 232, 116, 253, 193, 188, 117, 253, 193, 82, 117, 253, 193, 190, 117, 253, 193, 2, 118, 253, 193, 58, 118, 253, 193, 4, 118, 253, 193, 60, 118, 253, 193, 62, 118, 253, 193, 122, 118, 253, 193, 124, 118, 253, 193, 204, 118, 253, 193, 206, 118, 253, 193, 208, 118, 253, 206, 224, 104, 0, 206, 224, 104, 0, 206, 224, 104, 0, 206, 224, 104, 0, 206, 224, 104, 0, 206, 224, 104, 0, 206, 0, 104, 0, 206, 0, 104, 0, 206, 0, 104, 0, 206, 0, 104, 0, 206, 0, 104, 0, 193, 136, 115, 253, 193, 138, 115, 253, 193, 140, 115, 253, 193, 142, 115, 253, 193, 144, 115, 253, 193, 146, 115, 253, 193, 148, 115, 253, 193, 150, 115, 253, 193, 152, 115, 253, 193, 154, 115, 253, 193, 156, 115, 253, 193, 158, 115, 253, 193, 160, 115, 253, 193, 26, 116, 253, 193, 28, 116, 253, 193, 30, 116, 253, 193, 32, 116, 253, 193, 34, 116, 253, 193, 36, 116, 253, 193, 38, 116, 253, 193, 40, 116, 253, 193, 42, 116, 253, 193, 44, 116, 253, 193, 46, 116, 253, 193, 48, 116, 253, 193, 142, 116, 253, 193, 144, 116, 253, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 32, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 193, 130, 115, 253, 193, 132, 115, 253, 193, 134, 115, 253, 193, 12, 116, 253, 193, 14, 116, 253, 193, 16, 116, 253, 193, 18, 116, 253, 193, 20, 116, 253, 193, 22, 116, 253, 193, 24, 116, 253, 193, 126, 116, 253, 193, 128, 116, 253, 193, 130, 116, 253, 193, 132, 116, 253, 193, 134, 116, 253, 193, 136, 116, 253, 193, 138, 116, 253, 193, 252, 116, 253, 193, 254, 116, 253, 193, 2, 117, 253, 193, 4, 117, 253, 193, 6, 117, 253, 193, 8, 117, 253, 193, 10, 117, 253, 193, 12, 117, 253, 193, 14, 117, 253, 193, 16, 117, 253, 193, 18, 117, 253, 193, 20, 117, 253, 193, 22, 117, 253, 193, 98, 117, 253, 193, 100, 117, 253, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 64, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 193, 244, 115, 253, 193, 112, 116, 253, 193, 186, 117, 253, 193, 52, 118, 253, 193, 54, 118, 253, 193, 56, 118, 253, 193, 196, 118, 253, 193, 212, 118, 253, 193, 252, 118, 253, 193, 254, 118, 253, 193, 28, 119, 253, 193, 170, 119, 253, 193, 52, 120, 253, 193, 54, 120, 253, 193, 110, 120, 253, 193, 112, 120, 253, 193, 114, 120, 253, 193, 116, 120, 253, 193, 208, 120, 253, 193, 210, 120, 253, 193, 212, 120, 253, 193, 54, 121, 253, 193, 56, 121, 253, 193, 58, 121, 253, 193, 60, 121, 253, 193, 62, 121, 253, 193, 64, 121, 253, 193, 66, 121, 253, 193, 146, 121, 253, 193, 148, 121, 253, 193, 150, 121, 253, 193, 152, 121, 253, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 96, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 193, 146, 116, 253, 193, 148, 116, 253, 193, 150, 116, 253, 193, 152, 116, 253, 193, 154, 116, 253, 193, 156, 116, 253, 193, 158, 116, 253, 193, 160, 116, 253, 193, 162, 116, 253, 193, 164, 116, 253, 193, 166, 116, 253, 193, 168, 116, 253, 193, 26, 117, 253, 193, 28, 117, 253, 193, 30, 117, 253, 193, 32, 117, 253, 193, 34, 117, 253, 193, 36, 117, 253, 193, 38, 117, 253, 193, 40, 117, 253, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 128, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 206, 160, 104, 0, 193, 152, 117, 253, 193, 154, 117, 253, 193, 156, 117, 253, 193, 158, 117, 253, 193, 160, 117, 253, 193, 162, 117, 253, 193, 164, 117, 253, 193, 166, 117, 253, 193, 168, 117, 253, 193, 170, 117, 253, 193, 172, 117, 253, 193, 174, 117, 253, 193, 176, 117, 253, 193, 178, 117, 253, 193, 180, 117, 253, 193, 182, 117, 253, 193, 234, 117, 253, 193, 236, 117, 253, 193, 238, 117, 253, 193, 240, 117, 253, 193, 242, 117, 253, 193, 244, 117, 253, 193, 246, 117, 253, 193, 248, 117, 253, 193, 250, 117, 253, 193, 252, 117, 253, 193, 84, 117, 253, 193, 86, 117, 253, 193, 88, 117, 253, 193, 90, 117, 253, 193, 92, 117, 253, 193, 192, 117, 253, 193, 194, 117, 253, 193, 196, 117, 253, 193, 198, 117, 253, 193, 6, 118, 253, 193, 8, 118, 253, 193, 64, 118, 253, 193, 66, 118, 253, 193, 68, 118, 253, 193, 70, 118, 253, 193, 72, 118, 253, 193, 126, 118, 253, 193, 128, 118, 253, 193, 150, 118, 253, 193, 170, 118, 253, 193, 186, 118, 253, 193, 190, 118, 253, 193, 58, 119, 253, 193, 118, 119, 253, 193, 18, 120, 253, 193, 70, 120, 253, 193, 174, 120, 253, 193, 176, 120, 253, 193, 12, 121, 253, 193, 14, 121, 253, 193, 16, 121, 253, 193, 18, 121, 253, 193, 102, 117, 253, 193, 104, 117, 253, 193, 106, 117, 253, 193, 108, 117, 253, 193, 110, 117, 253, 193, 112, 117, 253, 193, 202, 117, 253, 193, 204, 117, 253, 193, 206, 117, 253, 193, 208, 117, 253, 193, 210, 117, 253, 193, 212, 117, 253, 193, 10, 118, 253, 193, 12, 118, 253, 193, 14, 118, 253, 193, 16, 118, 253, 193, 18, 118, 253, 193, 20, 118, 253, 193, 22, 118, 253, 193, 74, 118, 253, 193, 76, 118, 253, 193, 78, 118, 253, 193, 80, 118, 253, 193, 82, 118, 253, 193, 84, 118, 253, 193, 86, 118, 253, 193, 130, 118, 253, 193, 132, 118, 253, 193, 134, 118, 253, 193, 136, 118, 253, 193, 154, 118, 253, 193, 162, 118, 253, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 206, 192, 104, 0, 193, 216, 117, 253, 193, 218, 117, 253, 193, 220, 117, 253, 193, 222, 117, 253, 193, 224, 117, 253, 193, 226, 117, 253, 193, 228, 117, 253, 193, 230, 117, 253, 193, 232, 117, 253, 193, 24, 118, 253, 193, 26, 118, 253, 193, 28, 118, 253, 193, 88, 118, 253, 193, 90, 118, 253, 193, 92, 118, 253, 193, 94, 118, 253, 193, 96, 118, 253, 193, 98, 118, 253, 193, 100, 118, 253, 193, 102, 118, 253, 193, 104, 118, 253, 193, 106, 118, 253, 193, 108, 118, 253, 193, 140, 118, 253, 193, 142, 118, 253, 193, 200, 117, 253, 193, 152, 118, 253, 193, 182, 124, 253, 193, 234, 128, 253, 193, 42, 146, 253, 193, 84, 147, 253, 193, 110, 153, 253, 193, 250, 156, 253, 193, 16, 157, 253, 193, 212, 165, 253, 193, 76, 172, 253, 193, 104, 173, 253, 193, 140, 173, 253, 193, 2, 190, 253, 193, 12, 192, 253, 193, 118, 193, 253, 193, 194, 193, 253, 193, 8, 195, 253, 193, 120, 195, 253, 193, 220, 195, 253, 193, 6, 198, 253, 193, 138, 203, 253, 193, 132, 206, 253, 193, 160, 209, 253, 193, 60, 219, 253, 193, 56, 221, 253, 193, 140, 223, 253, 193, 136, 222, 253, 193, 208, 223, 253, 193, 212, 224, 253, 255, 255, 255, 255, 255, 255, 255, 255, 193, 254, 117, 253, 193, 30, 118, 253, 193, 32, 118, 253, 193, 34, 118, 253, 193, 36, 118, 253, 193, 38, 118, 253, 193, 40, 118, 253, 193, 42, 118, 253, 193, 44, 118, 253, 193, 46, 118, 253, 193, 48, 118, 253, 193, 50, 118, 253, 193, 110, 118, 253, 193, 112, 118, 253, 193, 114, 118, 253, 193, 116, 118, 253, 193, 118, 118, 253, 193, 120, 118, 253, 193, 148, 118, 253, 193, 158, 118, 253, 193, 160, 118, 253, 193, 178, 118, 253, 193, 180, 118, 253, 193, 200, 118, 253, 193, 202, 118, 253, 193, 210, 118, 253, 193, 218, 118, 253, 193, 250, 118, 253, 193, 92, 119, 253, 193, 196, 119, 253, 193, 224, 119, 253, 193, 2, 120, 253, 193, 144, 118, 253, 193, 146, 118, 253, 193, 156, 118, 253, 193, 164, 118, 253, 193, 166, 118, 253, 193, 168, 118, 253, 193, 174, 118, 253, 193, 176, 118, 253, 193, 184, 118, 253, 193, 188, 118, 253, 193, 194, 118, 253, 193, 216, 118, 253, 193, 248, 118, 253, 193, 66, 119, 253, 193, 86, 119, 253, 193, 88, 119, 253, 193, 90, 119, 253, 193, 124, 119, 253, 193, 148, 119, 253, 193, 168, 119, 253, 193, 192, 119, 253, 193, 194, 119, 253, 193, 4, 120, 253, 193, 40, 120, 253, 193, 76, 120, 253, 193, 78, 120, 253, 193, 80, 120, 253, 193, 82, 120, 253, 193, 186, 120, 253, 193, 188, 120, 253, 193, 32, 121, 253, 193, 124, 121, 253, 193, 172, 118, 253, 193, 182, 118, 253, 193, 192, 118, 253, 193, 198, 118, 253, 193, 60, 119, 253, 193, 38, 120, 253, 193, 72, 120, 253, 193, 178, 120, 253, 193, 180, 120, 253, 193, 22, 121, 253, 193, 228, 121, 253, 193, 230, 121, 253, 193, 232, 121, 253, 193, 206, 122, 253, 193, 208, 122, 253, 193, 42, 123, 253, 193, 44, 123, 253, 193, 46, 123, 253, 193, 100, 123, 253, 193, 164, 123, 253, 193, 232, 123, 253, 193, 234, 123, 253, 193, 6, 124, 253, 193, 26, 124, 253, 193, 92, 124, 253, 193, 116, 124, 253, 193, 118, 124, 253, 193, 120, 124, 253, 193, 154, 124, 253, 193, 156, 124, 253, 193, 158, 124, 253, 193, 160, 124, 253, 206, 224, 104, 0, 206, 224, 104, 0, 206, 224, 104, 0, 206, 224, 104, 0, 206, 224, 104, 0, 206, 224, 104, 0, 193, 2, 119, 253, 193, 4, 119, 253, 193, 6, 119, 253, 193, 8, 119, 253, 193, 10, 119, 253, 193, 12, 119, 253, 193, 14, 119, 253, 193, 16, 119, 253, 193, 18, 119, 253, 193, 20, 119, 253, 193, 22, 119, 253, 193, 24, 119, 253, 193, 30, 119, 253, 193, 32, 119, 253, 193, 34, 119, 253, 193, 36, 119, 253, 193, 38, 119, 253, 193, 40, 119, 253, 193, 42, 119, 253, 193, 44, 119, 253, 193, 46, 119, 253, 193, 48, 119, 253, 193, 50, 119, 253, 193, 52, 119, 253, 193, 54, 119, 253, 193, 56, 119, 253, 193, 68, 119, 253, 193, 70, 119, 253, 193, 72, 119, 253, 193, 74, 119, 253, 193, 76, 119, 253, 193, 78, 119, 253, 193, 80, 119, 253, 193, 94, 119, 253, 193, 96, 119, 253, 193, 98, 119, 253, 193, 100, 119, 253, 193, 102, 119, 253, 193, 104, 119, 253, 193, 106, 119, 253, 193, 108, 119, 253, 193, 110, 119, 253, 193, 112, 119, 253, 193, 114, 119, 253, 193, 116, 119, 253, 193, 128, 119, 253, 193, 130, 119, 253, 193, 132, 119, 253, 193, 150, 119, 253, 193, 134, 119, 253, 193, 136, 119, 253, 193, 138, 119, 253, 193, 140, 119, 253, 193, 142, 119, 253, 193, 144, 119, 253, 193, 146, 119, 253, 193, 152, 119, 253, 193, 154, 119, 253, 193, 120, 119, 253, 193, 122, 119, 253, 193, 166, 119, 253, 193, 188, 119, 253, 193, 190, 119, 253, 193, 220, 119, 253, 193, 222, 119, 253, 193, 74, 120, 253, 193, 182, 120, 253, 193, 184, 120, 253, 193, 24, 121, 253, 193, 26, 121, 253, 193, 28, 121, 253, 193, 30, 121, 253, 193, 120, 121, 253, 193, 122, 121, 253, 193, 88, 122, 253, 193, 90, 122, 253, 193, 92, 122, 253, 193, 94, 122, 253, 193, 96, 122, 253, 193, 98, 122, 253, 193, 100, 122, 253, 193, 210, 122, 253, 193, 212, 122, 253, 193, 214, 122, 253, 193, 216, 122, 253, 193, 218, 122, 253, 193, 48, 123, 253, 193, 50, 123, 253, 193, 52, 123, 253, 193, 102, 123, 253, 193, 156, 119, 253, 193, 158, 119, 253, 193, 160, 119, 253, 193, 162, 119, 253, 193, 164, 119, 253, 193, 172, 119, 253, 193, 174, 119, 253, 193, 176, 119, 253, 193, 178, 119, 253, 193, 180, 119, 253, 193, 182, 119, 253, 193, 184, 119, 253, 193, 186, 119, 253, 193, 198, 119, 253, 193, 200, 119, 253, 193, 202, 119, 253, 193, 204, 119, 253, 193, 206, 119, 253, 193, 208, 119, 253, 193, 210, 119, 253, 193, 212, 119, 253, 193, 214, 119, 253, 193, 216, 119, 253, 193, 218, 119, 253, 206, 0, 105, 0, 206, 0, 105, 0, 206, 0, 105, 0, 206, 0, 105, 0, 206, 0, 105, 0, 206, 0, 105, 0, 206, 0, 105, 0, 206, 0, 105, 0, 206, 0, 105, 0, 206, 0, 105, 0, 206, 0, 105, 0, 193, 8, 120, 253, 193, 10, 120, 253, 193, 16, 120, 253, 193, 26, 120, 253, 193, 28, 120, 253, 193, 30, 120, 253, 193, 32, 120, 253, 193, 34, 120, 253, 193, 36, 120, 253, 193, 56, 120, 253, 193, 58, 120, 253, 193, 60, 120, 253, 193, 62, 120, 253, 193, 64, 120, 253, 193, 66, 120, 253, 193, 68, 120, 253, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 193, 6, 120, 253, 193, 12, 120, 253, 193, 14, 120, 253, 193, 20, 120, 253, 193, 22, 120, 253, 193, 24, 120, 253, 193, 42, 120, 253, 193, 44, 120, 253, 193, 46, 120, 253, 193, 48, 120, 253, 193, 50, 120, 253, 193, 84, 120, 253, 193, 86, 120, 253, 193, 88, 120, 253, 193, 90, 120, 253, 193, 92, 120, 253, 193, 94, 120, 253, 193, 96, 120, 253, 193, 98, 120, 253, 193, 100, 120, 253, 193, 102, 120, 253, 193, 104, 120, 253, 193, 106, 120, 253, 193, 108, 120, 253, 193, 190, 120, 253, 193, 192, 120, 253, 193, 194, 120, 253, 193, 196, 120, 253, 193, 198, 120, 253, 193, 200, 120, 253, 193, 202, 120, 253, 193, 204, 120, 253, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 32, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 193, 206, 120, 253, 193, 34, 121, 253, 193, 36, 121, 253, 193, 38, 121, 253, 193, 40, 121, 253, 193, 42, 121, 253, 193, 44, 121, 253, 193, 46, 121, 253, 193, 48, 121, 253, 193, 50, 121, 253, 193, 52, 121, 253, 193, 126, 121, 253, 193, 128, 121, 253, 193, 130, 121, 253, 193, 132, 121, 253, 193, 134, 121, 253, 193, 136, 121, 253, 193, 138, 121, 253, 193, 140, 121, 253, 193, 142, 121, 253, 193, 144, 121, 253, 193, 234, 121, 253, 193, 236, 121, 253, 193, 108, 122, 253, 193, 238, 121, 253, 193, 240, 121, 253, 193, 242, 121, 253, 193, 244, 121, 253, 193, 246, 121, 253, 193, 248, 121, 253, 193, 250, 121, 253, 193, 252, 121, 253, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 64, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 193, 20, 121, 253, 193, 220, 121, 253, 193, 222, 121, 253, 193, 224, 121, 253, 193, 226, 121, 253, 193, 82, 122, 253, 193, 84, 122, 253, 193, 86, 122, 253, 193, 204, 122, 253, 193, 98, 123, 253, 193, 230, 123, 253, 193, 66, 124, 253, 193, 114, 124, 253, 193, 148, 124, 253, 193, 150, 124, 253, 193, 152, 124, 253, 193, 180, 124, 253, 193, 220, 124, 253, 193, 248, 124, 253, 193, 250, 124, 253, 193, 28, 125, 253, 193, 102, 125, 253, 193, 104, 125, 253, 193, 112, 125, 253, 193, 114, 125, 253, 193, 146, 125, 253, 193, 200, 125, 253, 193, 36, 126, 253, 193, 144, 126, 253, 193, 146, 126, 253, 193, 248, 126, 253, 193, 250, 126, 253, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 96, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 128, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 193, 254, 121, 253, 193, 2, 122, 253, 193, 4, 122, 253, 193, 110, 122, 253, 193, 112, 122, 253, 193, 114, 122, 253, 193, 116, 122, 253, 193, 118, 122, 253, 193, 120, 122, 253, 193, 122, 122, 253, 193, 124, 122, 253, 193, 126, 122, 253, 193, 128, 122, 253, 193, 130, 122, 253, 193, 132, 122, 253, 193, 134, 122, 253, 193, 222, 122, 253, 193, 224, 122, 253, 193, 226, 122, 253, 193, 228, 122, 253, 193, 230, 122, 253, 193, 232, 122, 253, 193, 234, 122, 253, 193, 236, 122, 253, 193, 238, 122, 253, 193, 54, 123, 253, 193, 56, 123, 253, 193, 58, 123, 253, 193, 60, 123, 253, 193, 62, 123, 253, 193, 108, 123, 253, 193, 110, 123, 253, 193, 6, 122, 253, 193, 8, 122, 253, 193, 10, 122, 253, 193, 12, 122, 253, 193, 14, 122, 253, 193, 16, 122, 253, 193, 18, 122, 253, 193, 20, 122, 253, 193, 22, 122, 253, 193, 136, 122, 253, 193, 138, 122, 253, 193, 140, 122, 253, 193, 142, 122, 253, 193, 144, 122, 253, 193, 146, 122, 253, 193, 148, 122, 253, 193, 64, 123, 253, 193, 116, 123, 253, 193, 184, 123, 253, 193, 186, 123, 253, 193, 188, 123, 253, 193, 242, 123, 253, 193, 214, 124, 253, 193, 246, 124, 253, 193, 64, 125, 253, 193, 164, 125, 253, 193, 166, 125, 253, 193, 190, 125, 253, 193, 246, 125, 253, 193, 248, 125, 253, 193, 250, 125, 253, 193, 252, 125, 253, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 206, 160, 105, 0, 193, 150, 122, 253, 193, 152, 122, 253, 193, 154, 122, 253, 193, 156, 122, 253, 193, 158, 122, 253, 193, 160, 122, 253, 193, 162, 122, 253, 193, 164, 122, 253, 193, 166, 122, 253, 193, 168, 122, 253, 193, 170, 122, 253, 193, 172, 122, 253, 193, 174, 122, 253, 193, 176, 122, 253, 193, 178, 122, 253, 193, 180, 122, 253, 193, 182, 122, 253, 193, 184, 122, 253, 193, 186, 122, 253, 193, 188, 122, 253, 193, 190, 122, 253, 193, 192, 122, 253, 193, 194, 122, 253, 193, 196, 122, 253, 193, 198, 122, 253, 193, 200, 122, 253, 193, 202, 122, 253, 193, 102, 122, 253, 193, 104, 122, 253, 193, 106, 122, 253, 193, 220, 122, 253, 193, 10, 124, 253, 193, 88, 124, 253, 193, 96, 124, 253, 193, 126, 124, 253, 193, 168, 124, 253, 193, 170, 124, 253, 193, 172, 124, 253, 193, 174, 124, 253, 193, 192, 124, 253, 193, 194, 124, 253, 193, 236, 124, 253, 193, 4, 125, 253, 193, 6, 125, 253, 193, 48, 125, 253, 193, 50, 125, 253, 193, 52, 125, 253, 193, 84, 125, 253, 193, 86, 125, 253, 193, 88, 125, 253, 193, 106, 125, 253, 193, 108, 125, 253, 193, 120, 125, 253, 193, 122, 125, 253, 193, 134, 125, 253, 193, 136, 125, 253, 193, 140, 125, 253, 193, 142, 125, 253, 193, 148, 125, 253, 193, 240, 122, 253, 193, 242, 122, 253, 193, 244, 122, 253, 193, 246, 122, 253, 193, 248, 122, 253, 193, 250, 122, 253, 193, 252, 122, 253, 193, 254, 122, 253, 193, 2, 123, 253, 193, 4, 123, 253, 193, 6, 123, 253, 193, 8, 123, 253, 193, 10, 123, 253, 193, 12, 123, 253, 193, 14, 123, 253, 193, 16, 123, 253, 193, 18, 123, 253, 193, 20, 123, 253, 193, 22, 123, 253, 193, 24, 123, 253, 193, 26, 123, 253, 193, 28, 123, 253, 193, 30, 123, 253, 193, 32, 123, 253, 193, 34, 123, 253, 193, 36, 123, 253, 193, 38, 123, 253, 193, 40, 123, 253, 206, 192, 105, 0, 206, 192, 105, 0, 206, 192, 105, 0, 206, 192, 105, 0, 206, 192, 105, 0, 206, 192, 105, 0, 206, 192, 105, 0, 206, 192, 105, 0, 206, 192, 105, 0, 206, 192, 105, 0, 206, 192, 105, 0, 206, 192, 105, 0, 193, 118, 123, 253, 193, 120, 123, 253, 193, 122, 123, 253, 193, 124, 123, 253, 193, 126, 123, 253, 193, 128, 123, 253, 193, 130, 123, 253, 193, 132, 123, 253, 193, 134, 123, 253, 193, 136, 123, 253, 193, 138, 123, 253, 193, 140, 123, 253, 193, 142, 123, 253, 193, 144, 123, 253, 193, 146, 123, 253, 193, 148, 123, 253, 193, 150, 123, 253, 193, 152, 123, 253, 193, 154, 123, 253, 193, 156, 123, 253, 193, 104, 123, 253, 193, 106, 123, 253, 193, 166, 123, 253, 193, 168, 123, 253, 193, 170, 123, 253, 193, 172, 123, 253, 193, 236, 123, 253, 193, 238, 123, 253, 193, 8, 124, 253, 193, 48, 124, 253, 193, 68, 124, 253, 193, 94, 124, 253, 193, 122, 124, 253, 193, 124, 124, 253, 193, 162, 124, 253, 193, 164, 124, 253, 193, 166, 124, 253, 193, 188, 124, 253, 193, 190, 124, 253, 193, 234, 124, 253, 193, 42, 125, 253, 193, 44, 125, 253, 193, 46, 125, 253, 193, 82, 125, 253, 193, 132, 125, 253, 193, 204, 125, 253, 193, 44, 126, 253, 193, 46, 126, 253, 193, 150, 126, 253, 193, 152, 126, 253, 193, 154, 126, 253, 193, 156, 126, 253, 193, 112, 123, 253, 193, 114, 123, 253, 193, 174, 123, 253, 193, 176, 123, 253, 193, 178, 123, 253, 193, 180, 123, 253, 193, 182, 123, 253, 193, 240, 123, 253, 193, 84, 124, 253, 193, 86, 124, 253, 193, 90, 124, 253, 193, 98, 124, 253, 193, 100, 124, 253, 193, 102, 124, 253, 193, 104, 124, 253, 193, 106, 124, 253, 193, 128, 124, 253, 193, 130, 124, 253, 193, 132, 124, 253, 193, 134, 124, 253, 193, 136, 124, 253, 193, 138, 124, 253, 193, 140, 124, 253, 193, 142, 124, 253, 193, 176, 124, 253, 193, 178, 124, 253, 193, 196, 124, 253, 193, 198, 124, 253, 193, 200, 124, 253, 193, 202, 124, 253, 193, 204, 124, 253, 193, 206, 124, 253, 193, 158, 123, 253, 193, 160, 123, 253, 193, 162, 123, 253, 193, 190, 123, 253, 193, 192, 123, 253, 193, 194, 123, 253, 193, 196, 123, 253, 193, 198, 123, 253, 193, 200, 123, 253, 193, 202, 123, 253, 193, 204, 123, 253, 193, 206, 123, 253, 193, 208, 123, 253, 193, 210, 123, 253, 193, 212, 123, 253, 193, 214, 123, 253, 193, 216, 123, 253, 193, 218, 123, 253, 193, 220, 123, 253, 193, 222, 123, 253, 193, 244, 123, 253, 193, 224, 123, 253, 193, 226, 123, 253, 193, 228, 123, 253, 193, 246, 123, 253, 193, 248, 123, 253, 193, 250, 123, 253, 193, 252, 123, 253, 193, 254, 123, 253, 193, 2, 124, 253, 193, 4, 124, 253, 193, 12, 124, 253, 193, 14, 124, 253, 193, 16, 124, 253, 193, 18, 124, 253, 193, 20, 124, 253, 193, 22, 124, 253, 193, 24, 124, 253, 193, 28, 124, 253, 193, 30, 124, 253, 193, 32, 124, 253, 193, 34, 124, 253, 193, 36, 124, 253, 193, 38, 124, 253, 193, 40, 124, 253, 193, 42, 124, 253, 193, 44, 124, 253, 193, 50, 124, 253, 193, 52, 124, 253, 193, 54, 124, 253, 193, 56, 124, 253, 193, 58, 124, 253, 193, 60, 124, 253, 193, 62, 124, 253, 193, 46, 124, 253, 193, 70, 124, 253, 193, 64, 124, 253, 193, 72, 124, 253, 193, 74, 124, 253, 193, 76, 124, 253, 193, 78, 124, 253, 193, 80, 124, 253, 193, 82, 124, 253, 193, 108, 124, 253, 193, 110, 124, 253, 193, 112, 124, 253, 193, 144, 124, 253, 193, 146, 124, 253, 193, 216, 124, 253, 193, 218, 124, 253, 193, 66, 125, 253, 193, 68, 125, 253, 193, 70, 125, 253, 193, 98, 125, 253, 193, 100, 125, 253, 193, 124, 125, 253, 193, 126, 125, 253, 193, 144, 125, 253, 193, 152, 125, 253, 193, 158, 125, 253, 193, 168, 125, 253, 193, 170, 125, 253, 193, 172, 125, 253, 193, 174, 125, 253, 193, 176, 125, 253, 193, 178, 125, 253, 193, 192, 125, 253, 193, 194, 125, 253, 193, 196, 125, 253, 193, 198, 125, 253, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 193, 184, 124, 253, 193, 186, 124, 253, 193, 222, 124, 253, 193, 224, 124, 253, 193, 226, 124, 253, 193, 228, 124, 253, 193, 230, 124, 253, 193, 232, 124, 253, 193, 252, 124, 253, 193, 254, 124, 253, 193, 2, 125, 253, 193, 30, 125, 253, 193, 32, 125, 253, 193, 34, 125, 253, 193, 36, 125, 253, 193, 38, 125, 253, 193, 40, 125, 253, 193, 72, 125, 253, 193, 74, 125, 253, 193, 76, 125, 253, 193, 78, 125, 253, 193, 80, 125, 253, 193, 116, 125, 253, 193, 118, 125, 253, 193, 128, 125, 253, 193, 130, 125, 253, 193, 202, 125, 253, 193, 38, 126, 253, 193, 40, 126, 253, 193, 42, 126, 253, 193, 148, 126, 253, 193, 254, 126, 253, 193, 208, 124, 253, 193, 210, 124, 253, 193, 212, 124, 253, 193, 238, 124, 253, 193, 240, 124, 253, 193, 242, 124, 253, 193, 244, 124, 253, 193, 8, 125, 253, 193, 10, 125, 253, 193, 12, 125, 253, 193, 14, 125, 253, 193, 16, 125, 253, 193, 18, 125, 253, 193, 20, 125, 253, 193, 22, 125, 253, 193, 24, 125, 253, 193, 26, 125, 253, 193, 54, 125, 253, 193, 56, 125, 253, 193, 58, 125, 253, 193, 60, 125, 253, 193, 62, 125, 253, 193, 90, 125, 253, 193, 92, 125, 253, 193, 94, 125, 253, 193, 96, 125, 253, 193, 110, 125, 253, 193, 138, 125, 253, 193, 154, 125, 253, 193, 156, 125, 253, 193, 160, 125, 253, 193, 162, 125, 253, 193, 150, 125, 253, 193, 206, 125, 253, 193, 208, 125, 253, 193, 48, 126, 253, 193, 50, 126, 253, 193, 52, 126, 253, 193, 158, 126, 253, 193, 160, 126, 253, 193, 162, 126, 253, 193, 164, 126, 253, 193, 166, 126, 253, 193, 16, 127, 253, 193, 18, 127, 253, 193, 20, 127, 253, 193, 22, 127, 253, 193, 24, 127, 253, 193, 26, 127, 253, 193, 160, 127, 253, 193, 162, 127, 253, 193, 164, 127, 253, 193, 166, 127, 253, 193, 168, 127, 253, 193, 170, 127, 253, 193, 80, 128, 253, 193, 82, 128, 253, 193, 84, 128, 253, 193, 86, 128, 253, 193, 88, 128, 253, 193, 90, 128, 253, 193, 92, 128, 253, 193, 94, 128, 253, 193, 96, 128, 253, 193, 180, 125, 253, 193, 182, 125, 253, 193, 184, 125, 253, 193, 186, 125, 253, 193, 188, 125, 253, 193, 210, 125, 253, 193, 212, 125, 253, 193, 214, 125, 253, 193, 216, 125, 253, 193, 218, 125, 253, 193, 220, 125, 253, 193, 222, 125, 253, 193, 224, 125, 253, 193, 226, 125, 253, 193, 228, 125, 253, 193, 230, 125, 253, 193, 232, 125, 253, 193, 234, 125, 253, 193, 236, 125, 253, 193, 238, 125, 253, 193, 240, 125, 253, 193, 242, 125, 253, 193, 244, 125, 253, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 193, 254, 125, 253, 193, 92, 126, 253, 193, 94, 126, 253, 193, 96, 126, 253, 193, 192, 126, 253, 193, 194, 126, 253, 193, 196, 126, 253, 193, 198, 126, 253, 193, 200, 126, 253, 193, 66, 127, 253, 193, 68, 127, 253, 193, 70, 127, 253, 193, 72, 127, 253, 193, 74, 127, 253, 193, 76, 127, 253, 193, 78, 127, 253, 193, 208, 127, 253, 193, 210, 127, 253, 193, 212, 127, 253, 193, 214, 127, 253, 193, 216, 127, 253, 193, 218, 127, 253, 193, 220, 127, 253, 193, 144, 128, 253, 193, 146, 128, 253, 193, 148, 128, 253, 193, 32, 129, 253, 193, 34, 129, 253, 193, 36, 129, 253, 193, 38, 129, 253, 193, 40, 129, 253, 193, 42, 129, 253, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 206, 224, 105, 0, 193, 98, 126, 253, 193, 100, 126, 253, 193, 102, 126, 253, 193, 104, 126, 253, 193, 106, 126, 253, 193, 108, 126, 253, 193, 110, 126, 253, 193, 112, 126, 253, 193, 114, 126, 253, 193, 116, 126, 253, 193, 118, 126, 253, 193, 120, 126, 253, 193, 122, 126, 253, 193, 124, 126, 253, 193, 126, 126, 253, 193, 128, 126, 253, 193, 130, 126, 253, 193, 132, 126, 253, 193, 134, 126, 253, 193, 136, 126, 253, 193, 138, 126, 253, 193, 140, 126, 253, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 206, 0, 106, 0, 193, 168, 126, 253, 193, 170, 126, 253, 193, 172, 126, 253, 193, 174, 126, 253, 193, 176, 126, 253, 193, 178, 126, 253, 193, 180, 126, 253, 193, 182, 126, 253, 193, 184, 126, 253, 193, 186, 126, 253, 193, 188, 126, 253, 193, 190, 126, 253, 193, 28, 127, 253, 193, 30, 127, 253, 193, 32, 127, 253, 193, 34, 127, 253, 193, 172, 127, 253, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 193, 142, 126, 253, 193, 202, 126, 253, 193, 204, 126, 253, 193, 206, 126, 253, 193, 208, 126, 253, 193, 210, 126, 253, 193, 212, 126, 253, 193, 214, 126, 253, 193, 216, 126, 253, 193, 218, 126, 253, 193, 220, 126, 253, 193, 222, 126, 253, 193, 224, 126, 253, 193, 226, 126, 253, 193, 228, 126, 253, 193, 230, 126, 253, 193, 232, 126, 253, 193, 234, 126, 253, 193, 236, 126, 253, 193, 238, 126, 253, 193, 240, 126, 253, 193, 242, 126, 253, 193, 244, 126, 253, 193, 246, 126, 253, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 206, 64, 106, 0, 193, 252, 126, 253, 193, 136, 127, 253, 193, 138, 127, 253, 193, 140, 127, 253, 193, 50, 128, 253, 193, 52, 128, 253, 193, 218, 128, 253, 193, 220, 128, 253, 193, 222, 128, 253, 193, 224, 128, 253, 193, 226, 128, 253, 193, 228, 128, 253, 193, 230, 128, 253, 193, 232, 128, 253, 193, 136, 129, 253, 193, 138, 129, 253, 193, 248, 129, 253, 193, 250, 129, 253, 193, 252, 129, 253, 193, 254, 129, 253, 193, 112, 130, 253, 193, 114, 130, 253, 193, 116, 130, 253, 193, 118, 130, 253, 193, 102, 131, 253, 193, 202, 131, 253, 193, 82, 132, 253, 193, 216, 132, 253, 193, 208, 133, 253, 193, 188, 134, 253, 193, 218, 134, 253, 193, 22, 135, 253, 193, 2, 127, 253, 193, 142, 127, 253, 193, 144, 127, 253, 193, 146, 127, 253, 193, 54, 128, 253, 193, 56, 128, 253, 193, 58, 128, 253, 193, 60, 128, 253, 193, 62, 128, 253, 193, 64, 128, 253, 193, 66, 128, 253, 193, 236, 128, 253, 193, 238, 128, 253, 193, 240, 128, 253, 193, 242, 128, 253, 193, 244, 128, 253, 193, 140, 129, 253, 193, 142, 129, 253, 193, 144, 129, 253, 193, 146, 129, 253, 193, 2, 130, 253, 193, 4, 130, 253, 193, 6, 130, 253, 193, 8, 130, 253, 193, 10, 130, 253, 193, 120, 130, 253, 193, 122, 130, 253, 193, 124, 130, 253, 193, 208, 130, 253, 193, 210, 130, 253, 193, 212, 130, 253, 193, 214, 130, 253, 193, 4, 127, 253, 193, 6, 127, 253, 193, 8, 127, 253, 193, 10, 127, 253, 193, 12, 127, 253, 193, 14, 127, 253, 193, 148, 127, 253, 193, 150, 127, 253, 193, 152, 127, 253, 193, 154, 127, 253, 193, 156, 127, 253, 193, 158, 127, 253, 193, 68, 128, 253, 193, 70, 128, 253, 193, 72, 128, 253, 193, 74, 128, 253, 193, 76, 128, 253, 193, 78, 128, 253, 193, 246, 128, 253, 193, 148, 129, 253, 193, 150, 129, 253, 193, 152, 129, 253, 193, 154, 129, 253, 193, 156, 129, 253, 193, 158, 129, 253, 193, 160, 129, 253, 193, 162, 129, 253, 193, 12, 130, 253, 193, 14, 130, 253, 193, 16, 130, 253, 193, 18, 130, 253, 193, 20, 130, 253, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 206, 32, 106, 0, 193, 174, 127, 253, 193, 176, 127, 253, 193, 178, 127, 253, 193, 180, 127, 253, 193, 182, 127, 253, 193, 184, 127, 253, 193, 186, 127, 253, 193, 188, 127, 253, 193, 190, 127, 253, 193, 192, 127, 253, 193, 194, 127, 253, 193, 196, 127, 253, 193, 198, 127, 253, 193, 200, 127, 253, 193, 202, 127, 253, 193, 204, 127, 253, 193, 206, 127, 253, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 206, 96, 106, 0, 193, 150, 128, 253, 193, 152, 128, 253, 193, 98, 128, 253, 193, 248, 128, 253, 193, 250, 128, 253, 193, 252, 128, 253, 193, 254, 128, 253, 193, 2, 129, 253, 193, 4, 129, 253, 193, 6, 129, 253, 193, 8, 129, 253, 193, 10, 129, 253, 193, 164, 129, 253, 193, 166, 129, 253, 193, 22, 130, 253, 193, 24, 130, 253, 193, 26, 130, 253, 193, 28, 130, 253, 193, 30, 130, 253, 193, 136, 130, 253, 193, 138, 130, 253, 193, 140, 130, 253, 193, 142, 130, 253, 193, 144, 130, 253, 193, 146, 130, 253, 193, 226, 130, 253, 193, 228, 130, 253, 193, 230, 130, 253, 193, 232, 130, 253, 193, 40, 131, 253, 193, 66, 131, 253, 193, 68, 131, 253, 193, 70, 131, 253, 193, 72, 131, 253, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 206, 128, 106, 0, 193, 12, 129, 253, 193, 14, 129, 253, 193, 16, 129, 253, 193, 18, 129, 253, 193, 20, 129, 253, 193, 22, 129, 253, 193, 24, 129, 253, 193, 26, 129, 253, 193, 28, 129, 253, 193, 168, 129, 253, 193, 170, 129, 253, 193, 172, 129, 253, 193, 174, 129, 253, 193, 176, 129, 253, 193, 178, 129, 253, 193, 154, 128, 253, 193, 156, 128, 253, 193, 158, 128, 253, 193, 160, 128, 253, 193, 162, 128, 253, 193, 164, 128, 253, 193, 166, 128, 253, 193, 168, 128, 253, 193, 170, 128, 253, 193, 172, 128, 253, 193, 174, 128, 253, 193, 176, 128, 253, 193, 178, 128, 253, 193, 180, 128, 253, 193, 182, 128, 253, 193, 184, 128, 253, 193, 186, 128, 253, 193, 188, 128, 253, 193, 190, 128, 253, 193, 192, 128, 253, 193, 194, 128, 253, 193, 196, 128, 253, 193, 198, 128, 253, 193, 200, 128, 253, 193, 202, 128, 253, 193, 204, 128, 253, 193, 206, 128, 253, 193, 208, 128, 253, 193, 210, 128, 253, 193, 212, 128, 253, 193, 214, 128, 253, 193, 216, 128, 253, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 160, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 193, 180, 129, 253, 193, 182, 129, 253, 193, 32, 130, 253, 193, 34, 130, 253, 193, 36, 130, 253, 193, 38, 130, 253, 193, 148, 130, 253, 193, 150, 130, 253, 193, 152, 130, 253, 193, 154, 130, 253, 193, 156, 130, 253, 193, 158, 130, 253, 193, 234, 130, 253, 193, 236, 130, 253, 193, 238, 130, 253, 193, 240, 130, 253, 193, 242, 130, 253, 193, 244, 130, 253, 193, 246, 130, 253, 193, 42, 131, 253, 193, 74, 131, 253, 193, 112, 131, 253, 193, 186, 131, 253, 193, 204, 131, 253, 193, 212, 131, 253, 193, 228, 131, 253, 193, 230, 131, 253, 193, 232, 131, 253, 193, 234, 131, 253, 193, 236, 131, 253, 193, 238, 131, 253, 193, 250, 131, 253, 193, 184, 129, 253, 193, 186, 129, 253, 193, 188, 129, 253, 193, 190, 129, 253, 193, 40, 130, 253, 193, 42, 130, 253, 193, 44, 130, 253, 193, 160, 130, 253, 193, 162, 130, 253, 193, 46, 130, 253, 193, 164, 130, 253, 193, 248, 130, 253, 193, 250, 130, 253, 193, 252, 130, 253, 193, 76, 131, 253, 193, 114, 131, 253, 193, 132, 131, 253, 193, 240, 131, 253, 193, 254, 131, 253, 193, 2, 132, 253, 193, 48, 132, 253, 193, 50, 132, 253, 193, 52, 132, 253, 193, 110, 132, 253, 193, 112, 132, 253, 193, 114, 132, 253, 193, 116, 132, 253, 193, 164, 132, 253, 193, 224, 132, 253, 193, 226, 132, 253, 193, 24, 133, 253, 193, 26, 133, 253, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 192, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 206, 224, 106, 0, 193, 166, 130, 253, 193, 168, 130, 253, 193, 170, 130, 253, 193, 172, 130, 253, 193, 174, 130, 253, 193, 176, 130, 253, 193, 178, 130, 253, 193, 180, 130, 253, 193, 182, 130, 253, 193, 184, 130, 253, 193, 186, 130, 253, 193, 188, 130, 253, 193, 190, 130, 253, 193, 192, 130, 253, 193, 194, 130, 253, 193, 196, 130, 253, 193, 198, 130, 253, 193, 200, 130, 253, 193, 202, 130, 253, 193, 204, 130, 253, 193, 206, 130, 253, 193, 254, 130, 253, 193, 126, 130, 253, 193, 128, 130, 253, 193, 130, 130, 253, 193, 132, 130, 253, 193, 134, 130, 253, 193, 216, 130, 253, 193, 218, 130, 253, 193, 220, 130, 253, 193, 222, 130, 253, 193, 224, 130, 253, 193, 30, 131, 253, 193, 32, 131, 253, 193, 34, 131, 253, 193, 36, 131, 253, 193, 38, 131, 253, 193, 64, 131, 253, 193, 104, 131, 253, 193, 130, 131, 253, 193, 166, 131, 253, 193, 174, 131, 253, 193, 188, 131, 253, 193, 216, 131, 253, 193, 24, 132, 253, 193, 86, 132, 253, 193, 88, 132, 253, 193, 90, 132, 253, 193, 92, 132, 253, 193, 156, 132, 253, 193, 18, 133, 253, 193, 94, 133, 253, 193, 96, 133, 253, 193, 98, 133, 253, 193, 2, 131, 253, 193, 4, 131, 253, 193, 6, 131, 253, 193, 8, 131, 253, 193, 10, 131, 253, 193, 12, 131, 253, 193, 14, 131, 253, 193, 16, 131, 253, 193, 18, 131, 253, 193, 20, 131, 253, 193, 22, 131, 253, 193, 24, 131, 253, 193, 26, 131, 253, 193, 44, 131, 253, 193, 46, 131, 253, 193, 48, 131, 253, 193, 50, 131, 253, 193, 52, 131, 253, 193, 54, 131, 253, 193, 56, 131, 253, 193, 58, 131, 253, 193, 60, 131, 253, 193, 78, 131, 253, 193, 80, 131, 253, 193, 82, 131, 253, 193, 84, 131, 253, 193, 86, 131, 253, 193, 88, 131, 253, 193, 90, 131, 253, 193, 92, 131, 253, 193, 94, 131, 253, 193, 96, 131, 253, 193, 28, 131, 253, 193, 62, 131, 253, 193, 124, 131, 253, 193, 126, 131, 253, 193, 128, 131, 253, 193, 146, 131, 253, 193, 148, 131, 253, 193, 150, 131, 253, 193, 160, 131, 253, 193, 172, 131, 253, 193, 84, 132, 253, 193, 154, 132, 253, 193, 218, 132, 253, 193, 220, 132, 253, 193, 16, 133, 253, 193, 90, 133, 253, 193, 92, 133, 253, 193, 154, 133, 253, 193, 46, 134, 253, 193, 48, 134, 253, 193, 50, 134, 253, 193, 100, 134, 253, 193, 102, 134, 253, 193, 128, 134, 253, 193, 220, 134, 253, 193, 26, 135, 253, 193, 28, 135, 253, 193, 98, 135, 253, 193, 12, 136, 253, 193, 14, 136, 253, 193, 146, 136, 253, 193, 148, 136, 253, 193, 98, 131, 253, 193, 100, 131, 253, 193, 116, 131, 253, 193, 118, 131, 253, 193, 120, 131, 253, 193, 122, 131, 253, 193, 134, 131, 253, 193, 136, 131, 253, 193, 138, 131, 253, 193, 140, 131, 253, 193, 142, 131, 253, 193, 144, 131, 253, 193, 156, 131, 253, 193, 158, 131, 253, 193, 162, 131, 253, 193, 164, 131, 253, 193, 168, 131, 253, 193, 170, 131, 253, 193, 178, 131, 253, 193, 180, 131, 253, 193, 182, 131, 253, 193, 184, 131, 253, 193, 190, 131, 253, 193, 192, 131, 253, 193, 194, 131, 253, 193, 196, 131, 253, 193, 198, 131, 253, 193, 200, 131, 253, 193, 206, 131, 253, 193, 208, 131, 253, 193, 210, 131, 253, 193, 214, 131, 253, 193, 106, 131, 253, 193, 108, 131, 253, 193, 110, 131, 253, 193, 152, 131, 253, 193, 154, 131, 253, 193, 248, 131, 253, 193, 26, 132, 253, 193, 28, 132, 253, 193, 30, 132, 253, 193, 94, 132, 253, 193, 158, 133, 253, 193, 212, 133, 253, 193, 254, 133, 253, 193, 52, 134, 253, 193, 104, 134, 253, 193, 106, 134, 253, 193, 164, 134, 253, 193, 190, 134, 253, 193, 108, 135, 253, 193, 110, 135, 253, 193, 24, 136, 253, 193, 26, 136, 253, 193, 28, 136, 253, 193, 30, 136, 253, 193, 32, 136, 253, 193, 34, 136, 253, 193, 36, 136, 253, 193, 174, 136, 253, 193, 176, 136, 253, 193, 178, 136, 253, 193, 180, 136, 253, 193, 182, 136, 253, 193, 218, 131, 253, 193, 220, 131, 253, 193, 222, 131, 253, 193, 224, 131, 253, 193, 226, 131, 253, 193, 242, 131, 253, 193, 244, 131, 253, 193, 246, 131, 253, 193, 4, 132, 253, 193, 6, 132, 253, 193, 8, 132, 253, 193, 10, 132, 253, 193, 12, 132, 253, 193, 14, 132, 253, 193, 16, 132, 253, 193, 18, 132, 253, 193, 20, 132, 253, 193, 22, 132, 253, 193, 54, 132, 253, 193, 56, 132, 253, 193, 58, 132, 253, 193, 60, 132, 253, 193, 62, 132, 253, 193, 64, 132, 253, 193, 66, 132, 253, 193, 68, 132, 253, 193, 70, 132, 253, 193, 72, 132, 253, 193, 74, 132, 253, 193, 76, 132, 253, 193, 78, 132, 253, 193, 80, 132, 253, 193, 252, 131, 253, 193, 32, 132, 253, 193, 34, 132, 253, 193, 36, 132, 253, 193, 38, 132, 253, 193, 40, 132, 253, 193, 42, 132, 253, 193, 44, 132, 253, 193, 46, 132, 253, 193, 96, 132, 253, 193, 98, 132, 253, 193, 100, 132, 253, 193, 102, 132, 253, 193, 104, 132, 253, 193, 106, 132, 253, 193, 108, 132, 253, 193, 158, 132, 253, 193, 160, 132, 253, 193, 162, 132, 253, 193, 222, 132, 253, 193, 20, 133, 253, 193, 22, 133, 253, 193, 100, 133, 253, 193, 160, 133, 253, 193, 162, 133, 253, 193, 164, 133, 253, 193, 166, 133, 253, 193, 168, 133, 253, 193, 170, 133, 253, 193, 172, 133, 253, 193, 174, 133, 253, 193, 176, 133, 253, 193, 118, 132, 253, 193, 120, 132, 253, 193, 122, 132, 253, 193, 124, 132, 253, 193, 126, 132, 253, 193, 128, 132, 253, 193, 130, 132, 253, 193, 132, 132, 253, 193, 134, 132, 253, 193, 136, 132, 253, 193, 138, 132, 253, 193, 140, 132, 253, 193, 142, 132, 253, 193, 144, 132, 253, 193, 146, 132, 253, 193, 148, 132, 253, 193, 150, 132, 253, 193, 152, 132, 253, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 206, 0, 107, 0, 193, 228, 132, 253, 193, 230, 132, 253, 193, 232, 132, 253, 193, 234, 132, 253, 193, 236, 132, 253, 193, 238, 132, 253, 193, 240, 132, 253, 193, 242, 132, 253, 193, 244, 132, 253, 193, 246, 132, 253, 193, 248, 132, 253, 193, 250, 132, 253, 193, 252, 132, 253, 193, 254, 132, 253, 193, 2, 133, 253, 193, 4, 133, 253, 193, 6, 133, 253, 193, 8, 133, 253, 193, 10, 133, 253, 193, 12, 133, 253, 193, 14, 133, 253, 193, 28, 133, 253, 193, 30, 133, 253, 193, 32, 133, 253, 193, 34, 133, 253, 193, 36, 133, 253, 193, 38, 133, 253, 193, 40, 133, 253, 193, 42, 133, 253, 193, 44, 133, 253, 193, 46, 133, 253, 193, 48, 133, 253, 193, 50, 133, 253, 193, 52, 133, 253, 193, 54, 133, 253, 193, 56, 133, 253, 193, 58, 133, 253, 193, 60, 133, 253, 193, 62, 133, 253, 193, 64, 133, 253, 193, 66, 133, 253, 193, 68, 133, 253, 193, 70, 133, 253, 193, 72, 133, 253, 193, 74, 133, 253, 193, 76, 133, 253, 193, 78, 133, 253, 193, 80, 133, 253, 193, 82, 133, 253, 193, 84, 133, 253, 193, 86, 133, 253, 193, 88, 133, 253, 193, 106, 133, 253, 193, 108, 133, 253, 193, 110, 133, 253, 193, 112, 133, 253, 193, 114, 133, 253, 193, 116, 133, 253, 193, 118, 133, 253, 193, 120, 133, 253, 193, 122, 133, 253, 193, 124, 133, 253, 193, 126, 133, 253, 193, 128, 133, 253, 193, 130, 133, 253, 193, 132, 133, 253, 193, 134, 133, 253, 193, 136, 133, 253, 193, 138, 133, 253, 193, 140, 133, 253, 193, 142, 133, 253, 193, 144, 133, 253, 193, 146, 133, 253, 193, 148, 133, 253, 193, 150, 133, 253, 193, 152, 133, 253, 206, 32, 107, 0, 206, 32, 107, 0, 206, 32, 107, 0, 206, 32, 107, 0, 206, 32, 107, 0, 206, 32, 107, 0, 206, 32, 107, 0, 206, 32, 107, 0, 206, 32, 107, 0, 193, 156, 133, 253, 193, 210, 133, 253, 193, 146, 134, 253, 193, 162, 134, 253, 193, 180, 134, 253, 193, 222, 134, 253, 193, 30, 135, 253, 193, 32, 135, 253, 193, 34, 135, 253, 193, 36, 135, 253, 193, 100, 135, 253, 193, 102, 135, 253, 193, 104, 135, 253, 193, 106, 135, 253, 193, 180, 135, 253, 193, 182, 135, 253, 193, 16, 136, 253, 193, 18, 136, 253, 193, 20, 136, 253, 193, 22, 136, 253, 193, 158, 136, 253, 193, 160, 136, 253, 193, 162, 136, 253, 193, 164, 136, 253, 193, 166, 136, 253, 193, 168, 136, 253, 193, 170, 136, 253, 193, 172, 136, 253, 206, 192, 107, 0, 206, 192, 107, 0, 206, 192, 107, 0, 206, 192, 107, 0, 193, 178, 133, 253, 193, 214, 133, 253, 193, 216, 133, 253, 193, 218, 133, 253, 193, 54, 134, 253, 193, 130, 134, 253, 193, 132, 134, 253, 193, 166, 134, 253, 193, 178, 134, 253, 193, 192, 134, 253, 193, 194, 134, 253, 193, 196, 134, 253, 193, 224, 134, 253, 193, 226, 134, 253, 193, 228, 134, 253, 193, 230, 134, 253, 193, 232, 134, 253, 193, 234, 134, 253, 193, 236, 134, 253, 193, 238, 134, 253, 193, 38, 135, 253, 193, 40, 135, 253, 193, 42, 135, 253, 193, 44, 135, 253, 193, 46, 135, 253, 193, 48, 135, 253, 193, 50, 135, 253, 193, 52, 135, 253, 193, 54, 135, 253, 193, 56, 135, 253, 193, 58, 135, 253, 193, 112, 135, 253, 206, 32, 107, 0, 206, 32, 107, 0, 206, 32, 107, 0, 206, 32, 107, 0, 193, 222, 133, 253, 193, 224, 133, 253, 193, 226, 133, 253, 193, 228, 133, 253, 193, 230, 133, 253, 193, 232, 133, 253, 193, 234, 133, 253, 193, 236, 133, 253, 193, 238, 133, 253, 193, 240, 133, 253, 193, 242, 133, 253, 193, 244, 133, 253, 193, 246, 133, 253, 193, 248, 133, 253, 193, 250, 133, 253, 193, 252, 133, 253, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 206, 64, 107, 0, 193, 56, 134, 253, 193, 58, 134, 253, 193, 60, 134, 253, 193, 62, 134, 253, 193, 64, 134, 253, 193, 66, 134, 253, 193, 68, 134, 253, 193, 70, 134, 253, 193, 72, 134, 253, 193, 74, 134, 253, 193, 76, 134, 253, 193, 78, 134, 253, 193, 82, 134, 253, 193, 84, 134, 253, 193, 86, 134, 253, 193, 88, 134, 253, 193, 90, 134, 253, 193, 92, 134, 253, 193, 94, 134, 253, 193, 96, 134, 253, 193, 98, 134, 253, 206, 96, 107, 0, 206, 96, 107, 0, 206, 96, 107, 0, 206, 96, 107, 0, 206, 96, 107, 0, 206, 96, 107, 0, 193, 134, 134, 253, 193, 136, 134, 253, 193, 138, 134, 253, 193, 140, 134, 253, 193, 142, 134, 253, 193, 144, 134, 253, 193, 150, 134, 253, 193, 152, 134, 253, 193, 154, 134, 253, 193, 156, 134, 253, 193, 158, 134, 253, 193, 160, 134, 253, 193, 168, 134, 253, 193, 170, 134, 253, 193, 174, 134, 253, 193, 176, 134, 253, 193, 182, 134, 253, 193, 184, 134, 253, 193, 186, 134, 253, 193, 200, 134, 253, 193, 202, 134, 253, 193, 204, 134, 253, 193, 206, 134, 253, 193, 208, 134, 253, 193, 210, 134, 253, 193, 212, 134, 253, 193, 214, 134, 253, 193, 216, 134, 253, 193, 244, 134, 253, 193, 246, 134, 253, 193, 248, 134, 253, 193, 250, 134, 253, 193, 252, 134, 253, 193, 254, 134, 253, 193, 2, 135, 253, 193, 4, 135, 253, 193, 6, 135, 253, 193, 8, 135, 253, 193, 10, 135, 253, 193, 12, 135, 253, 193, 14, 135, 253, 193, 16, 135, 253, 193, 18, 135, 253, 193, 20, 135, 253, 193, 62, 135, 253, 193, 64, 135, 253, 193, 66, 135, 253, 193, 68, 135, 253, 193, 70, 135, 253, 193, 72, 135, 253, 193, 74, 135, 253, 193, 76, 135, 253, 193, 78, 135, 253, 193, 80, 135, 253, 193, 82, 135, 253, 193, 84, 135, 253, 193, 86, 135, 253, 193, 88, 135, 253, 193, 24, 135, 253, 193, 94, 135, 253, 193, 96, 135, 253, 193, 176, 135, 253, 193, 178, 135, 253, 193, 8, 136, 253, 193, 10, 136, 253, 193, 142, 136, 253, 193, 144, 136, 253, 193, 28, 137, 253, 193, 30, 137, 253, 193, 156, 137, 253, 193, 16, 138, 253, 193, 18, 138, 253, 193, 108, 138, 253, 193, 110, 138, 253, 193, 112, 138, 253, 193, 114, 138, 253, 193, 190, 138, 253, 193, 6, 139, 253, 193, 8, 139, 253, 193, 96, 139, 253, 193, 98, 139, 253, 193, 186, 139, 253, 193, 222, 139, 253, 193, 244, 139, 253, 193, 246, 139, 253, 193, 22, 140, 253, 193, 32, 140, 253, 193, 102, 140, 253, 193, 142, 140, 253, 193, 148, 140, 253, 193, 90, 135, 253, 193, 92, 135, 253, 193, 134, 135, 253, 193, 136, 135, 253, 193, 138, 135, 253, 193, 140, 135, 253, 193, 142, 135, 253, 193, 144, 135, 253, 193, 146, 135, 253, 193, 148, 135, 253, 193, 150, 135, 253, 193, 152, 135, 253, 193, 154, 135, 253, 193, 156, 135, 253, 193, 158, 135, 253, 193, 160, 135, 253, 193, 162, 135, 253, 193, 164, 135, 253, 193, 166, 135, 253, 193, 168, 135, 253, 193, 170, 135, 253, 193, 172, 135, 253, 193, 174, 135, 253, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 193, 114, 135, 253, 193, 116, 135, 253, 193, 184, 135, 253, 193, 186, 135, 253, 193, 188, 135, 253, 193, 190, 135, 253, 193, 192, 135, 253, 193, 194, 135, 253, 193, 196, 135, 253, 193, 198, 135, 253, 193, 200, 135, 253, 193, 38, 136, 253, 193, 40, 136, 253, 193, 42, 136, 253, 193, 44, 136, 253, 193, 46, 136, 253, 193, 48, 136, 253, 193, 50, 136, 253, 193, 52, 136, 253, 193, 54, 136, 253, 193, 56, 136, 253, 193, 58, 136, 253, 193, 60, 136, 253, 193, 62, 136, 253, 193, 184, 136, 253, 193, 186, 136, 253, 193, 188, 136, 253, 193, 190, 136, 253, 193, 192, 136, 253, 193, 194, 136, 253, 193, 196, 136, 253, 193, 198, 136, 253, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 128, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 206, 160, 107, 0, 193, 216, 136, 253, 193, 218, 136, 253, 193, 220, 136, 253, 193, 222, 136, 253, 193, 224, 136, 253, 193, 226, 136, 253, 193, 228, 136, 253, 193, 230, 136, 253, 193, 232, 136, 253, 193, 84, 137, 253, 193, 234, 136, 253, 193, 150, 136, 253, 193, 152, 136, 253, 193, 154, 136, 253, 193, 156, 136, 253, 193, 32, 137, 253, 193, 34, 137, 253, 193, 36, 137, 253, 193, 38, 137, 253, 193, 158, 137, 253, 193, 20, 138, 253, 193, 22, 138, 253, 193, 116, 138, 253, 193, 118, 138, 253, 193, 120, 138, 253, 193, 192, 138, 253, 193, 194, 138, 253, 193, 196, 138, 253, 193, 10, 139, 253, 193, 12, 139, 253, 193, 14, 139, 253, 193, 16, 139, 253, 193, 164, 139, 253, 193, 188, 139, 253, 193, 248, 139, 253, 193, 250, 139, 253, 193, 64, 140, 253, 193, 92, 140, 253, 193, 94, 140, 253, 193, 118, 140, 253, 193, 124, 140, 253, 193, 128, 140, 253, 193, 144, 140, 253, 193, 200, 136, 253, 193, 62, 137, 253, 193, 64, 137, 253, 193, 66, 137, 253, 193, 68, 137, 253, 193, 70, 137, 253, 193, 176, 137, 253, 193, 178, 137, 253, 193, 180, 137, 253, 193, 182, 137, 253, 193, 184, 137, 253, 193, 186, 137, 253, 193, 188, 137, 253, 193, 190, 137, 253, 193, 28, 138, 253, 193, 30, 138, 253, 193, 32, 138, 253, 193, 34, 138, 253, 193, 36, 138, 253, 193, 38, 138, 253, 193, 40, 138, 253, 193, 134, 138, 253, 193, 136, 138, 253, 193, 138, 138, 253, 193, 140, 138, 253, 193, 142, 138, 253, 193, 144, 138, 253, 193, 146, 138, 253, 193, 206, 138, 253, 193, 208, 138, 253, 193, 210, 138, 253, 193, 212, 138, 253, 193, 206, 136, 253, 193, 208, 136, 253, 193, 210, 136, 253, 193, 212, 136, 253, 193, 214, 136, 253, 193, 72, 137, 253, 193, 74, 137, 253, 193, 76, 137, 253, 193, 78, 137, 253, 193, 80, 137, 253, 193, 82, 137, 253, 193, 192, 137, 253, 193, 194, 137, 253, 193, 196, 137, 253, 193, 198, 137, 253, 193, 200, 137, 253, 193, 42, 138, 253, 193, 44, 138, 253, 193, 216, 138, 253, 193, 218, 138, 253, 193, 220, 138, 253, 193, 32, 139, 253, 193, 34, 139, 253, 193, 78, 139, 253, 193, 106, 139, 253, 193, 162, 139, 253, 193, 202, 139, 253, 193, 212, 139, 253, 193, 38, 140, 253, 193, 52, 140, 253, 193, 54, 140, 253, 193, 68, 140, 253, 193, 236, 136, 253, 193, 238, 136, 253, 193, 240, 136, 253, 193, 242, 136, 253, 193, 244, 136, 253, 193, 246, 136, 253, 193, 248, 136, 253, 193, 250, 136, 253, 193, 252, 136, 253, 193, 254, 136, 253, 193, 2, 137, 253, 193, 4, 137, 253, 193, 6, 137, 253, 193, 8, 137, 253, 193, 10, 137, 253, 193, 12, 137, 253, 193, 14, 137, 253, 193, 16, 137, 253, 193, 18, 137, 253, 193, 20, 137, 253, 193, 22, 137, 253, 193, 24, 137, 253, 193, 26, 137, 253, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 224, 107, 0, 206, 192, 107, 0, 206, 192, 107, 0, 206, 192, 107, 0, 206, 192, 107, 0, 193, 160, 137, 253, 193, 162, 137, 253, 193, 164, 137, 253, 193, 166, 137, 253, 193, 168, 137, 253, 193, 170, 137, 253, 193, 172, 137, 253, 193, 174, 137, 253, 193, 24, 138, 253, 193, 122, 138, 253, 193, 124, 138, 253, 193, 126, 138, 253, 193, 128, 138, 253, 193, 130, 138, 253, 193, 132, 138, 253, 193, 198, 138, 253, 193, 200, 138, 253, 193, 202, 138, 253, 193, 204, 138, 253, 193, 18, 139, 253, 193, 20, 139, 253, 193, 22, 139, 253, 193, 24, 139, 253, 193, 26, 139, 253, 193, 54, 139, 253, 193, 100, 139, 253, 193, 128, 139, 253, 193, 150, 139, 253, 193, 56, 137, 253, 193, 58, 137, 253, 193, 60, 137, 253, 193, 26, 138, 253, 193, 28, 139, 253, 193, 56, 139, 253, 193, 58, 139, 253, 193, 60, 139, 253, 193, 62, 139, 253, 193, 102, 139, 253, 193, 194, 139, 253, 193, 200, 139, 253, 193, 214, 139, 253, 193, 10, 140, 253, 193, 12, 140, 253, 193, 140, 140, 253, 193, 68, 141, 253, 193, 198, 141, 253, 193, 22, 142, 253, 193, 120, 142, 253, 193, 208, 142, 253, 193, 210, 142, 253, 193, 42, 143, 253, 193, 142, 143, 253, 193, 16, 144, 253, 193, 18, 144, 253, 193, 20, 144, 253, 193, 126, 144, 253, 193, 128, 144, 253, 193, 20, 145, 253, 193, 146, 145, 253, 193, 210, 145, 253, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 206, 0, 108, 0, 193, 46, 138, 253, 193, 48, 138, 253, 193, 50, 138, 253, 193, 52, 138, 253, 193, 54, 138, 253, 193, 56, 138, 253, 193, 58, 138, 253, 193, 60, 138, 253, 193, 62, 138, 253, 193, 64, 138, 253, 193, 66, 138, 253, 193, 68, 138, 253, 193, 70, 138, 253, 193, 72, 138, 253, 193, 74, 138, 253, 193, 76, 138, 253, 193, 78, 138, 253, 193, 80, 138, 253, 193, 82, 138, 253, 193, 84, 138, 253, 193, 86, 138, 253, 193, 88, 138, 253, 193, 90, 138, 253, 193, 92, 138, 253, 193, 94, 138, 253, 193, 96, 138, 253, 193, 98, 138, 253, 193, 100, 138, 253, 193, 102, 138, 253, 193, 104, 138, 253, 193, 106, 138, 253, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 32, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 193, 214, 138, 253, 193, 30, 139, 253, 193, 64, 139, 253, 193, 66, 139, 253, 193, 68, 139, 253, 193, 70, 139, 253, 193, 72, 139, 253, 193, 74, 139, 253, 193, 76, 139, 253, 193, 104, 139, 253, 193, 130, 139, 253, 193, 152, 139, 253, 193, 154, 139, 253, 193, 156, 139, 253, 193, 158, 139, 253, 193, 174, 139, 253, 193, 206, 139, 253, 193, 216, 139, 253, 193, 218, 139, 253, 193, 224, 139, 253, 193, 226, 139, 253, 193, 232, 139, 253, 193, 234, 139, 253, 193, 254, 139, 253, 193, 2, 140, 253, 193, 4, 140, 253, 193, 8, 140, 253, 193, 14, 140, 253, 193, 16, 140, 253, 193, 26, 140, 253, 193, 44, 140, 253, 193, 66, 140, 253, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 206, 64, 108, 0, 193, 36, 139, 253, 193, 38, 139, 253, 193, 40, 139, 253, 193, 42, 139, 253, 193, 44, 139, 253, 193, 46, 139, 253, 193, 48, 139, 253, 193, 50, 139, 253, 193, 52, 139, 253, 193, 80, 139, 253, 193, 82, 139, 253, 193, 84, 139, 253, 193, 86, 139, 253, 193, 88, 139, 253, 193, 90, 139, 253, 193, 92, 139, 253, 193, 94, 139, 253, 193, 108, 139, 253, 193, 110, 139, 253, 193, 112, 139, 253, 193, 114, 139, 253, 193, 116, 139, 253, 193, 118, 139, 253, 193, 120, 139, 253, 193, 122, 139, 253, 193, 124, 139, 253, 193, 126, 139, 253, 193, 132, 139, 253, 193, 134, 139, 253, 193, 136, 139, 253, 193, 138, 139, 253, 193, 140, 139, 253, 193, 142, 139, 253, 193, 144, 139, 253, 193, 146, 139, 253, 193, 148, 139, 253, 193, 170, 139, 253, 193, 172, 139, 253, 193, 176, 139, 253, 193, 178, 139, 253, 193, 180, 139, 253, 193, 182, 139, 253, 193, 184, 139, 253, 193, 192, 139, 253, 193, 196, 139, 253, 193, 198, 139, 253, 193, 204, 139, 253, 193, 208, 139, 253, 193, 210, 139, 253, 193, 220, 139, 253, 193, 228, 139, 253, 193, 230, 139, 253, 193, 236, 139, 253, 193, 238, 139, 253, 193, 240, 139, 253, 193, 242, 139, 253, 193, 6, 140, 253, 193, 18, 140, 253, 193, 20, 140, 253, 193, 24, 140, 253, 193, 166, 139, 253, 193, 168, 139, 253, 193, 190, 139, 253, 193, 252, 139, 253, 193, 86, 140, 253, 193, 96, 140, 253, 193, 98, 140, 253, 193, 104, 140, 253, 193, 146, 140, 253, 193, 164, 140, 253, 193, 222, 140, 253, 193, 224, 140, 253, 193, 248, 140, 253, 193, 24, 141, 253, 193, 44, 141, 253, 193, 60, 141, 253, 193, 88, 141, 253, 193, 116, 141, 253, 193, 118, 141, 253, 193, 162, 141, 253, 193, 20, 142, 253, 193, 114, 142, 253, 193, 116, 142, 253, 193, 118, 142, 253, 193, 204, 142, 253, 193, 206, 142, 253, 193, 34, 143, 253, 193, 36, 143, 253, 193, 38, 143, 253, 193, 40, 143, 253, 193, 140, 143, 253, 193, 8, 144, 253, 193, 28, 140, 253, 193, 30, 140, 253, 193, 34, 140, 253, 193, 36, 140, 253, 193, 40, 140, 253, 193, 42, 140, 253, 193, 46, 140, 253, 193, 48, 140, 253, 193, 50, 140, 253, 193, 56, 140, 253, 193, 58, 140, 253, 193, 60, 140, 253, 193, 62, 140, 253, 193, 70, 140, 253, 193, 72, 140, 253, 193, 74, 140, 253, 193, 76, 140, 253, 193, 78, 140, 253, 193, 80, 140, 253, 193, 82, 140, 253, 193, 84, 140, 253, 193, 88, 140, 253, 193, 90, 140, 253, 193, 100, 140, 253, 193, 110, 140, 253, 193, 112, 140, 253, 193, 114, 140, 253, 193, 116, 140, 253, 193, 120, 140, 253, 193, 122, 140, 253, 193, 132, 140, 253, 193, 134, 140, 253, 193, 106, 140, 253, 193, 126, 140, 253, 193, 152, 140, 253, 193, 154, 140, 253, 193, 186, 140, 253, 193, 226, 140, 253, 193, 250, 140, 253, 193, 90, 141, 253, 193, 126, 141, 253, 193, 160, 141, 253, 193, 164, 141, 253, 193, 170, 141, 253, 193, 172, 141, 253, 193, 174, 141, 253, 193, 176, 141, 253, 193, 178, 141, 253, 193, 180, 141, 253, 193, 182, 141, 253, 193, 200, 141, 253, 193, 202, 141, 253, 193, 204, 141, 253, 193, 206, 141, 253, 193, 208, 141, 253, 193, 210, 141, 253, 193, 212, 141, 253, 193, 214, 141, 253, 206, 96, 108, 0, 206, 96, 108, 0, 206, 96, 108, 0, 206, 96, 108, 0, 206, 96, 108, 0, 206, 96, 108, 0, 193, 108, 140, 253, 193, 130, 140, 253, 193, 166, 140, 253, 193, 168, 140, 253, 193, 170, 140, 253, 193, 188, 140, 253, 193, 190, 140, 253, 193, 192, 140, 253, 193, 252, 140, 253, 193, 254, 140, 253, 193, 26, 141, 253, 193, 70, 141, 253, 193, 92, 141, 253, 193, 94, 141, 253, 193, 120, 141, 253, 193, 138, 141, 253, 193, 184, 141, 253, 193, 186, 141, 253, 193, 188, 141, 253, 193, 216, 141, 253, 193, 218, 141, 253, 193, 220, 141, 253, 193, 222, 141, 253, 193, 224, 141, 253, 193, 226, 141, 253, 193, 50, 142, 253, 193, 52, 142, 253, 193, 54, 142, 253, 193, 56, 142, 253, 193, 58, 142, 253, 193, 60, 142, 253, 193, 146, 142, 253, 193, 136, 140, 253, 193, 138, 140, 253, 193, 150, 140, 253, 193, 158, 140, 253, 193, 160, 140, 253, 193, 162, 140, 253, 193, 172, 140, 253, 193, 174, 140, 253, 193, 176, 140, 253, 193, 178, 140, 253, 193, 180, 140, 253, 193, 182, 140, 253, 193, 184, 140, 253, 193, 194, 140, 253, 193, 196, 140, 253, 193, 198, 140, 253, 193, 200, 140, 253, 193, 202, 140, 253, 193, 204, 140, 253, 193, 206, 140, 253, 193, 208, 140, 253, 193, 210, 140, 253, 193, 212, 140, 253, 193, 214, 140, 253, 193, 216, 140, 253, 193, 230, 140, 253, 193, 232, 140, 253, 193, 234, 140, 253, 193, 236, 140, 253, 193, 238, 140, 253, 193, 240, 140, 253, 193, 242, 140, 253, 193, 218, 140, 253, 193, 246, 140, 253, 193, 22, 141, 253, 193, 58, 141, 253, 193, 114, 141, 253, 193, 110, 142, 253, 193, 112, 142, 253, 193, 198, 142, 253, 193, 200, 142, 253, 193, 202, 144, 253, 193, 10, 145, 253, 193, 70, 145, 253, 193, 204, 145, 253, 193, 224, 145, 253, 193, 110, 146, 253, 193, 202, 146, 253, 193, 10, 147, 253, 193, 38, 147, 253, 193, 60, 147, 253, 193, 74, 147, 253, 193, 92, 147, 253, 193, 94, 147, 253, 193, 96, 147, 253, 193, 214, 147, 253, 193, 230, 147, 253, 193, 58, 148, 253, 193, 102, 148, 253, 193, 110, 148, 253, 193, 134, 148, 253, 193, 16, 149, 253, 193, 120, 149, 253, 193, 144, 150, 253, 193, 220, 140, 253, 193, 42, 141, 253, 193, 84, 141, 253, 193, 86, 141, 253, 193, 202, 142, 253, 193, 32, 143, 253, 193, 138, 143, 253, 193, 6, 144, 253, 193, 12, 145, 253, 193, 14, 145, 253, 193, 226, 145, 253, 193, 44, 146, 253, 193, 46, 146, 253, 193, 48, 146, 253, 193, 148, 146, 253, 193, 254, 146, 253, 193, 16, 147, 253, 193, 34, 147, 253, 193, 40, 147, 253, 193, 52, 147, 253, 193, 54, 147, 253, 193, 86, 147, 253, 193, 98, 147, 253, 193, 100, 147, 253, 193, 106, 147, 253, 193, 232, 147, 253, 193, 234, 147, 253, 193, 236, 147, 253, 193, 4, 148, 253, 193, 28, 148, 253, 193, 136, 148, 253, 193, 160, 148, 253, 193, 244, 140, 253, 193, 2, 141, 253, 193, 4, 141, 253, 193, 6, 141, 253, 193, 8, 141, 253, 193, 10, 141, 253, 193, 12, 141, 253, 193, 14, 141, 253, 193, 16, 141, 253, 193, 18, 141, 253, 193, 20, 141, 253, 193, 28, 141, 253, 193, 30, 141, 253, 193, 32, 141, 253, 193, 34, 141, 253, 193, 36, 141, 253, 193, 38, 141, 253, 193, 40, 141, 253, 193, 46, 141, 253, 193, 48, 141, 253, 193, 50, 141, 253, 193, 52, 141, 253, 193, 54, 141, 253, 193, 56, 141, 253, 193, 62, 141, 253, 193, 64, 141, 253, 193, 66, 141, 253, 193, 72, 141, 253, 193, 74, 141, 253, 193, 76, 141, 253, 193, 78, 141, 253, 193, 80, 141, 253, 193, 82, 141, 253, 193, 96, 141, 253, 193, 98, 141, 253, 193, 100, 141, 253, 193, 102, 141, 253, 193, 104, 141, 253, 193, 106, 141, 253, 193, 108, 141, 253, 193, 110, 141, 253, 193, 112, 141, 253, 193, 122, 141, 253, 193, 124, 141, 253, 193, 128, 141, 253, 193, 130, 141, 253, 193, 132, 141, 253, 193, 134, 141, 253, 193, 136, 141, 253, 193, 140, 141, 253, 193, 142, 141, 253, 193, 144, 141, 253, 193, 146, 141, 253, 193, 148, 141, 253, 193, 150, 141, 253, 193, 152, 141, 253, 193, 154, 141, 253, 193, 156, 141, 253, 193, 158, 141, 253, 193, 166, 141, 253, 193, 168, 141, 253, 193, 190, 141, 253, 193, 192, 141, 253, 193, 194, 141, 253, 193, 196, 141, 253, 193, 228, 141, 253, 193, 230, 141, 253, 193, 232, 141, 253, 193, 234, 141, 253, 193, 236, 141, 253, 193, 238, 141, 253, 193, 240, 141, 253, 193, 242, 141, 253, 193, 244, 141, 253, 193, 246, 141, 253, 193, 248, 141, 253, 193, 250, 141, 253, 193, 252, 141, 253, 193, 254, 141, 253, 193, 2, 142, 253, 193, 4, 142, 253, 193, 6, 142, 253, 193, 8, 142, 253, 193, 10, 142, 253, 193, 12, 142, 253, 193, 14, 142, 253, 193, 16, 142, 253, 193, 18, 142, 253, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 96, 108, 0, 206, 96, 108, 0, 206, 96, 108, 0, 206, 96, 108, 0, 206, 96, 108, 0, 206, 96, 108, 0, 206, 96, 108, 0, 193, 122, 142, 253, 193, 124, 142, 253, 193, 126, 142, 253, 193, 128, 142, 253, 193, 130, 142, 253, 193, 132, 142, 253, 193, 134, 142, 253, 193, 136, 142, 253, 193, 138, 142, 253, 193, 140, 142, 253, 193, 142, 142, 253, 193, 144, 142, 253, 193, 212, 142, 253, 193, 214, 142, 253, 193, 216, 142, 253, 193, 218, 142, 253, 193, 220, 142, 253, 193, 222, 142, 253, 193, 224, 142, 253, 193, 44, 143, 253, 193, 46, 143, 253, 193, 48, 143, 253, 193, 50, 143, 253, 193, 52, 143, 253, 193, 54, 143, 253, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 128, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 206, 160, 108, 0, 193, 242, 142, 253, 193, 244, 142, 253, 193, 246, 142, 253, 193, 248, 142, 253, 193, 250, 142, 253, 193, 252, 142, 253, 193, 254, 142, 253, 193, 2, 143, 253, 193, 4, 143, 253, 193, 6, 143, 253, 193, 8, 143, 253, 193, 10, 143, 253, 193, 12, 143, 253, 193, 14, 143, 253, 193, 16, 143, 253, 193, 18, 143, 253, 193, 20, 143, 253, 193, 22, 143, 253, 193, 24, 143, 253, 193, 26, 143, 253, 193, 28, 143, 253, 193, 30, 143, 253, 193, 74, 143, 253, 193, 226, 142, 253, 193, 228, 142, 253, 193, 230, 142, 253, 193, 232, 142, 253, 193, 234, 142, 253, 193, 236, 142, 253, 193, 238, 142, 253, 193, 240, 142, 253, 193, 60, 143, 253, 193, 62, 143, 253, 193, 64, 143, 253, 193, 66, 143, 253, 193, 68, 143, 253, 193, 70, 143, 253, 193, 72, 143, 253, 193, 174, 143, 253, 193, 176, 143, 253, 193, 178, 143, 253, 193, 180, 143, 253, 193, 182, 143, 253, 193, 184, 143, 253, 193, 186, 143, 253, 193, 36, 144, 253, 193, 38, 144, 253, 193, 40, 144, 253, 193, 42, 144, 253, 193, 44, 144, 253, 193, 46, 144, 253, 193, 138, 144, 253, 193, 140, 144, 253, 193, 142, 144, 253, 193, 220, 144, 253, 193, 56, 143, 253, 193, 58, 143, 253, 193, 144, 143, 253, 193, 146, 143, 253, 193, 148, 143, 253, 193, 150, 143, 253, 193, 152, 143, 253, 193, 154, 143, 253, 193, 156, 143, 253, 193, 158, 143, 253, 193, 160, 143, 253, 193, 162, 143, 253, 193, 164, 143, 253, 193, 166, 143, 253, 193, 168, 143, 253, 193, 170, 143, 253, 193, 172, 143, 253, 193, 22, 144, 253, 193, 24, 144, 253, 193, 26, 144, 253, 193, 28, 144, 253, 193, 30, 144, 253, 193, 32, 144, 253, 193, 34, 144, 253, 193, 130, 144, 253, 193, 132, 144, 253, 193, 134, 144, 253, 193, 136, 144, 253, 193, 212, 144, 253, 193, 214, 144, 253, 193, 216, 144, 253, 193, 218, 144, 253, 193, 76, 143, 253, 193, 78, 143, 253, 193, 80, 143, 253, 193, 82, 143, 253, 193, 84, 143, 253, 193, 86, 143, 253, 193, 88, 143, 253, 193, 90, 143, 253, 193, 92, 143, 253, 193, 94, 143, 253, 193, 96, 143, 253, 193, 98, 143, 253, 193, 100, 143, 253, 193, 102, 143, 253, 193, 104, 143, 253, 193, 106, 143, 253, 193, 108, 143, 253, 193, 110, 143, 253, 193, 112, 143, 253, 193, 114, 143, 253, 193, 116, 143, 253, 193, 118, 143, 253, 193, 120, 143, 253, 193, 122, 143, 253, 193, 124, 143, 253, 193, 126, 143, 253, 193, 128, 143, 253, 193, 130, 143, 253, 193, 132, 143, 253, 193, 134, 143, 253, 193, 136, 143, 253, 193, 188, 143, 253, 193, 190, 143, 253, 193, 192, 143, 253, 193, 194, 143, 253, 193, 196, 143, 253, 193, 198, 143, 253, 193, 200, 143, 253, 193, 202, 143, 253, 193, 204, 143, 253, 193, 206, 143, 253, 193, 208, 143, 253, 193, 210, 143, 253, 193, 212, 143, 253, 193, 214, 143, 253, 193, 216, 143, 253, 193, 218, 143, 253, 193, 220, 143, 253, 193, 222, 143, 253, 193, 224, 143, 253, 193, 226, 143, 253, 193, 228, 143, 253, 193, 230, 143, 253, 193, 232, 143, 253, 193, 234, 143, 253, 193, 236, 143, 253, 193, 238, 143, 253, 193, 240, 143, 253, 193, 242, 143, 253, 193, 244, 143, 253, 193, 246, 143, 253, 193, 248, 143, 253, 193, 250, 143, 253, 193, 252, 143, 253, 193, 254, 143, 253, 193, 2, 144, 253, 193, 4, 144, 253, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 193, 10, 144, 253, 193, 12, 144, 253, 193, 14, 144, 253, 193, 120, 144, 253, 193, 122, 144, 253, 193, 124, 144, 253, 193, 204, 144, 253, 193, 206, 144, 253, 193, 208, 144, 253, 193, 210, 144, 253, 193, 16, 145, 253, 193, 18, 145, 253, 193, 72, 145, 253, 193, 74, 145, 253, 193, 198, 145, 253, 193, 200, 145, 253, 193, 202, 145, 253, 193, 50, 146, 253, 193, 150, 146, 253, 193, 226, 146, 253, 193, 42, 147, 253, 193, 62, 147, 253, 193, 198, 147, 253, 193, 200, 147, 253, 193, 238, 147, 253, 193, 6, 148, 253, 193, 8, 148, 253, 193, 30, 148, 253, 193, 60, 148, 253, 193, 82, 148, 253, 193, 112, 148, 253, 193, 126, 148, 253, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 206, 192, 108, 0, 193, 144, 144, 253, 193, 146, 144, 253, 193, 148, 144, 253, 193, 150, 144, 253, 193, 152, 144, 253, 193, 154, 144, 253, 193, 156, 144, 253, 193, 158, 144, 253, 193, 160, 144, 253, 193, 162, 144, 253, 193, 164, 144, 253, 193, 166, 144, 253, 193, 168, 144, 253, 193, 170, 144, 253, 193, 172, 144, 253, 193, 174, 144, 253, 193, 176, 144, 253, 193, 178, 144, 253, 193, 180, 144, 253, 193, 182, 144, 253, 193, 184, 144, 253, 193, 186, 144, 253, 193, 188, 144, 253, 193, 190, 144, 253, 193, 192, 144, 253, 193, 194, 144, 253, 193, 196, 144, 253, 193, 198, 144, 253, 193, 200, 144, 253, 193, 232, 144, 253, 193, 234, 144, 253, 193, 236, 144, 253, 193, 238, 144, 253, 193, 240, 144, 253, 193, 242, 144, 253, 193, 244, 144, 253, 193, 246, 144, 253, 193, 248, 144, 253, 193, 250, 144, 253, 193, 252, 144, 253, 193, 254, 144, 253, 193, 2, 145, 253, 193, 4, 145, 253, 193, 6, 145, 253, 193, 8, 145, 253, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 193, 222, 144, 253, 193, 224, 144, 253, 193, 226, 144, 253, 193, 228, 144, 253, 193, 230, 144, 253, 193, 30, 145, 253, 193, 32, 145, 253, 193, 80, 145, 253, 193, 82, 145, 253, 193, 116, 145, 253, 193, 130, 145, 253, 193, 132, 145, 253, 193, 150, 145, 253, 193, 152, 145, 253, 193, 170, 145, 253, 193, 212, 145, 253, 193, 234, 145, 253, 193, 236, 145, 253, 193, 238, 145, 253, 193, 240, 145, 253, 193, 54, 146, 253, 193, 82, 146, 253, 193, 122, 146, 253, 193, 124, 146, 253, 193, 126, 146, 253, 193, 160, 146, 253, 193, 162, 146, 253, 193, 208, 146, 253, 193, 230, 146, 253, 193, 244, 146, 253, 193, 112, 147, 253, 193, 122, 147, 253, 193, 22, 145, 253, 193, 24, 145, 253, 193, 26, 145, 253, 193, 28, 145, 253, 193, 76, 145, 253, 193, 78, 145, 253, 193, 110, 145, 253, 193, 112, 145, 253, 193, 114, 145, 253, 193, 148, 145, 253, 193, 180, 145, 253, 193, 196, 145, 253, 193, 206, 145, 253, 193, 230, 145, 253, 193, 232, 145, 253, 193, 14, 146, 253, 193, 16, 146, 253, 193, 18, 146, 253, 193, 52, 146, 253, 193, 76, 146, 253, 193, 78, 146, 253, 193, 80, 146, 253, 193, 112, 146, 253, 193, 114, 146, 253, 193, 116, 146, 253, 193, 118, 146, 253, 193, 120, 146, 253, 193, 182, 146, 253, 193, 152, 146, 253, 193, 154, 146, 253, 193, 156, 146, 253, 193, 158, 146, 253, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 206, 224, 108, 0, 193, 84, 145, 253, 193, 86, 145, 253, 193, 88, 145, 253, 193, 90, 145, 253, 193, 92, 145, 253, 193, 94, 145, 253, 193, 96, 145, 253, 193, 98, 145, 253, 193, 100, 145, 253, 193, 102, 145, 253, 193, 104, 145, 253, 193, 106, 145, 253, 193, 108, 145, 253, 193, 118, 145, 253, 193, 120, 145, 253, 193, 122, 145, 253, 193, 124, 145, 253, 193, 126, 145, 253, 193, 128, 145, 253, 193, 134, 145, 253, 193, 136, 145, 253, 193, 138, 145, 253, 193, 102, 181, 253, 193, 140, 145, 253, 193, 142, 145, 253, 193, 144, 145, 253, 193, 154, 145, 253, 193, 156, 145, 253, 193, 158, 145, 253, 193, 160, 145, 253, 193, 162, 145, 253, 193, 164, 145, 253, 193, 166, 145, 253, 193, 168, 145, 253, 193, 172, 145, 253, 193, 174, 145, 253, 193, 176, 145, 253, 193, 178, 145, 253, 193, 182, 145, 253, 193, 184, 145, 253, 193, 186, 145, 253, 193, 188, 145, 253, 193, 190, 145, 253, 193, 192, 145, 253, 193, 194, 145, 253, 193, 208, 145, 253, 193, 214, 145, 253, 193, 216, 145, 253, 193, 218, 145, 253, 193, 220, 145, 253, 193, 222, 145, 253, 206, 0, 109, 0, 206, 0, 109, 0, 206, 0, 109, 0, 206, 0, 109, 0, 206, 0, 109, 0, 206, 0, 109, 0, 206, 0, 109, 0, 206, 0, 109, 0, 193, 228, 145, 253, 193, 12, 146, 253, 193, 74, 146, 253, 193, 44, 147, 253, 193, 56, 147, 253, 193, 58, 147, 253, 193, 64, 147, 253, 193, 66, 147, 253, 193, 68, 147, 253, 193, 70, 147, 253, 193, 88, 147, 253, 193, 102, 147, 253, 193, 104, 147, 253, 193, 118, 147, 253, 193, 160, 147, 253, 193, 202, 147, 253, 193, 32, 148, 253, 193, 34, 148, 253, 193, 62, 148, 253, 193, 64, 148, 253, 193, 24, 149, 253, 193, 26, 149, 253, 193, 246, 149, 253, 193, 248, 149, 253, 193, 160, 150, 253, 193, 162, 150, 253, 193, 252, 151, 253, 193, 254, 151, 253, 193, 2, 152, 253, 193, 152, 152, 253, 193, 244, 152, 253, 193, 52, 153, 253, 206, 0, 109, 0, 206, 0, 109, 0, 206, 0, 109, 0, 206, 0, 109, 0, 193, 20, 146, 253, 193, 22, 146, 253, 193, 24, 146, 253, 193, 26, 146, 253, 193, 28, 146, 253, 193, 30, 146, 253, 193, 32, 146, 253, 193, 34, 146, 253, 193, 36, 146, 253, 193, 38, 146, 253, 193, 40, 146, 253, 193, 56, 146, 253, 193, 58, 146, 253, 193, 60, 146, 253, 193, 62, 146, 253, 193, 64, 146, 253, 193, 66, 146, 253, 193, 68, 146, 253, 193, 70, 146, 253, 193, 72, 146, 253, 206, 32, 109, 0, 206, 32, 109, 0, 206, 32, 109, 0, 206, 32, 109, 0, 206, 32, 109, 0, 206, 32, 109, 0, 206, 32, 109, 0, 206, 32, 109, 0, 206, 32, 109, 0, 193, 128, 146, 253, 193, 130, 146, 253, 193, 132, 146, 253, 193, 134, 146, 253, 193, 136, 146, 253, 193, 138, 146, 253, 193, 140, 146, 253, 193, 142, 146, 253, 193, 144, 146, 253, 193, 146, 146, 253, 193, 164, 146, 253, 193, 166, 146, 253, 193, 168, 146, 253, 193, 170, 146, 253, 193, 172, 146, 253, 193, 174, 146, 253, 193, 176, 146, 253, 193, 178, 146, 253, 193, 180, 146, 253, 193, 186, 146, 253, 193, 188, 146, 253, 193, 190, 146, 253, 193, 192, 146, 253, 193, 194, 146, 253, 193, 196, 146, 253, 193, 198, 146, 253, 193, 200, 146, 253, 193, 184, 146, 253, 193, 204, 146, 253, 193, 206, 146, 253, 193, 228, 146, 253, 193, 2, 147, 253, 193, 4, 147, 253, 193, 30, 147, 253, 193, 32, 147, 253, 193, 46, 147, 253, 193, 72, 147, 253, 193, 76, 147, 253, 193, 78, 147, 253, 193, 90, 147, 253, 193, 108, 147, 253, 193, 110, 147, 253, 193, 120, 147, 253, 193, 126, 147, 253, 193, 130, 147, 253, 193, 142, 147, 253, 193, 184, 147, 253, 193, 204, 147, 253, 193, 206, 147, 253, 193, 216, 147, 253, 193, 218, 147, 253, 193, 10, 148, 253, 193, 44, 148, 253, 193, 46, 148, 253, 193, 66, 148, 253, 193, 68, 148, 253, 193, 84, 148, 253, 193, 86, 148, 253, 193, 88, 148, 253, 193, 210, 146, 253, 193, 212, 146, 253, 193, 214, 146, 253, 193, 216, 146, 253, 193, 218, 146, 253, 193, 220, 146, 253, 193, 222, 146, 253, 193, 224, 146, 253, 193, 232, 146, 253, 193, 234, 146, 253, 193, 236, 146, 253, 193, 238, 146, 253, 193, 240, 146, 253, 193, 242, 146, 253, 193, 246, 146, 253, 193, 248, 146, 253, 193, 250, 146, 253, 193, 252, 146, 253, 193, 6, 147, 253, 193, 8, 147, 253, 193, 12, 147, 253, 193, 14, 147, 253, 193, 18, 147, 253, 193, 20, 147, 253, 193, 22, 147, 253, 193, 24, 147, 253, 193, 26, 147, 253, 193, 28, 147, 253, 193, 36, 147, 253, 193, 48, 147, 253, 193, 50, 147, 253, 193, 80, 147, 253, 193, 82, 147, 253, 193, 114, 147, 253, 193, 116, 147, 253, 193, 124, 147, 253, 193, 128, 147, 253, 193, 134, 147, 253, 193, 136, 147, 253, 193, 138, 147, 253, 193, 140, 147, 253, 193, 148, 147, 253, 193, 150, 147, 253, 193, 154, 147, 253, 193, 164, 147, 253, 193, 156, 147, 253, 193, 158, 147, 253, 193, 166, 147, 253, 193, 168, 147, 253, 193, 170, 147, 253, 193, 172, 147, 253, 193, 174, 147, 253, 193, 178, 147, 253, 193, 180, 147, 253, 193, 182, 147, 253, 193, 196, 147, 253, 193, 210, 147, 253, 193, 212, 147, 253, 193, 222, 147, 253, 193, 224, 147, 253, 193, 226, 147, 253, 193, 228, 147, 253, 193, 240, 147, 253, 193, 242, 147, 253, 193, 132, 147, 253, 193, 144, 147, 253, 193, 146, 147, 253, 193, 152, 147, 253, 193, 162, 147, 253, 193, 176, 147, 253, 193, 190, 147, 253, 193, 192, 147, 253, 193, 194, 147, 253, 193, 208, 147, 253, 193, 220, 147, 253, 193, 12, 148, 253, 193, 14, 148, 253, 193, 48, 148, 253, 193, 70, 148, 253, 193, 72, 148, 253, 193, 114, 148, 253, 193, 172, 148, 253, 193, 174, 148, 253, 193, 218, 148, 253, 193, 220, 148, 253, 193, 222, 148, 253, 193, 40, 149, 253, 193, 42, 149, 253, 193, 44, 149, 253, 193, 46, 149, 253, 193, 48, 149, 253, 193, 50, 149, 253, 193, 52, 149, 253, 193, 54, 149, 253, 193, 56, 149, 253, 193, 166, 149, 253, 193, 244, 147, 253, 193, 246, 147, 253, 193, 248, 147, 253, 193, 250, 147, 253, 193, 252, 147, 253, 193, 254, 147, 253, 193, 2, 148, 253, 193, 16, 148, 253, 193, 18, 148, 253, 193, 20, 148, 253, 193, 22, 148, 253, 193, 24, 148, 253, 193, 26, 148, 253, 193, 36, 148, 253, 193, 38, 148, 253, 193, 40, 148, 253, 193, 42, 148, 253, 193, 50, 148, 253, 193, 52, 148, 253, 193, 54, 148, 253, 193, 56, 148, 253, 193, 74, 148, 253, 193, 76, 148, 253, 193, 78, 148, 253, 193, 80, 148, 253, 193, 90, 148, 253, 193, 92, 148, 253, 193, 94, 148, 253, 193, 96, 148, 253, 193, 98, 148, 253, 193, 100, 148, 253, 193, 104, 148, 253, 193, 106, 148, 253, 193, 108, 148, 253, 193, 116, 148, 253, 193, 118, 148, 253, 193, 120, 148, 253, 193, 122, 148, 253, 193, 124, 148, 253, 193, 130, 148, 253, 193, 132, 148, 253, 193, 138, 148, 253, 193, 140, 148, 253, 193, 144, 148, 253, 193, 152, 148, 253, 193, 154, 148, 253, 193, 156, 148, 253, 193, 158, 148, 253, 193, 176, 148, 253, 193, 178, 148, 253, 193, 180, 148, 253, 193, 182, 148, 253, 193, 184, 148, 253, 193, 186, 148, 253, 193, 188, 148, 253, 193, 190, 148, 253, 193, 192, 148, 253, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 193, 128, 148, 253, 193, 142, 148, 253, 193, 146, 148, 253, 193, 148, 148, 253, 193, 150, 148, 253, 193, 162, 148, 253, 193, 164, 148, 253, 193, 166, 148, 253, 193, 168, 148, 253, 193, 170, 148, 253, 193, 196, 148, 253, 193, 198, 148, 253, 193, 200, 148, 253, 193, 202, 148, 253, 193, 204, 148, 253, 193, 206, 148, 253, 193, 208, 148, 253, 193, 210, 148, 253, 193, 212, 148, 253, 193, 214, 148, 253, 193, 28, 149, 253, 193, 30, 149, 253, 193, 32, 149, 253, 193, 34, 149, 253, 193, 36, 149, 253, 193, 216, 148, 253, 193, 38, 149, 253, 206, 128, 109, 0, 206, 128, 109, 0, 206, 128, 109, 0, 206, 128, 109, 0, 206, 128, 109, 0, 193, 194, 148, 253, 193, 122, 149, 253, 193, 124, 149, 253, 193, 236, 149, 253, 193, 148, 150, 253, 193, 150, 150, 253, 193, 152, 150, 253, 193, 154, 150, 253, 193, 24, 151, 253, 193, 26, 151, 253, 193, 142, 151, 253, 193, 144, 151, 253, 193, 70, 152, 253, 193, 72, 152, 253, 193, 146, 152, 253, 193, 194, 152, 253, 193, 16, 153, 253, 193, 50, 153, 253, 193, 68, 153, 253, 193, 76, 153, 253, 193, 112, 153, 253, 193, 114, 153, 253, 193, 130, 153, 253, 193, 132, 153, 253, 193, 134, 153, 253, 193, 136, 153, 253, 193, 138, 153, 253, 193, 164, 153, 253, 193, 166, 153, 253, 193, 168, 153, 253, 193, 170, 153, 253, 193, 172, 153, 253, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 64, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 193, 18, 149, 253, 193, 20, 149, 253, 193, 22, 149, 253, 193, 126, 149, 253, 193, 128, 149, 253, 193, 130, 149, 253, 193, 132, 149, 253, 193, 238, 149, 253, 193, 240, 149, 253, 193, 242, 149, 253, 193, 244, 149, 253, 193, 156, 150, 253, 193, 158, 150, 253, 193, 28, 151, 253, 193, 30, 151, 253, 193, 32, 151, 253, 193, 34, 151, 253, 193, 36, 151, 253, 193, 38, 151, 253, 193, 40, 151, 253, 193, 42, 151, 253, 193, 146, 151, 253, 193, 148, 151, 253, 193, 242, 151, 253, 193, 244, 151, 253, 193, 246, 151, 253, 193, 248, 151, 253, 193, 250, 151, 253, 193, 74, 152, 253, 193, 76, 152, 253, 193, 148, 152, 253, 193, 150, 152, 253, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 96, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 128, 109, 0, 206, 128, 109, 0, 206, 128, 109, 0, 206, 128, 109, 0, 193, 250, 149, 253, 193, 252, 149, 253, 193, 152, 149, 253, 193, 154, 149, 253, 193, 156, 149, 253, 193, 158, 149, 253, 193, 160, 149, 253, 193, 162, 149, 253, 193, 164, 149, 253, 193, 254, 149, 253, 193, 2, 150, 253, 193, 4, 150, 253, 193, 6, 150, 253, 193, 8, 150, 253, 193, 10, 150, 253, 193, 12, 150, 253, 193, 14, 150, 253, 193, 16, 150, 253, 193, 18, 150, 253, 193, 20, 150, 253, 193, 22, 150, 253, 193, 24, 150, 253, 193, 26, 150, 253, 193, 28, 150, 253, 193, 30, 150, 253, 193, 32, 150, 253, 193, 44, 151, 253, 193, 164, 150, 253, 193, 168, 149, 253, 193, 170, 149, 253, 193, 172, 149, 253, 193, 34, 150, 253, 193, 36, 150, 253, 193, 38, 150, 253, 193, 40, 150, 253, 193, 42, 150, 253, 193, 44, 150, 253, 193, 46, 150, 253, 193, 48, 150, 253, 193, 188, 150, 253, 193, 190, 150, 253, 193, 192, 150, 253, 193, 194, 150, 253, 193, 196, 150, 253, 193, 198, 150, 253, 193, 200, 150, 253, 193, 202, 150, 253, 193, 204, 150, 253, 193, 68, 151, 253, 193, 70, 151, 253, 193, 72, 151, 253, 193, 74, 151, 253, 193, 76, 151, 253, 193, 78, 151, 253, 193, 80, 151, 253, 193, 82, 151, 253, 193, 84, 151, 253, 193, 86, 151, 253, 193, 168, 151, 253, 193, 170, 151, 253, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 160, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 206, 192, 109, 0, 193, 206, 150, 253, 193, 208, 150, 253, 193, 210, 150, 253, 193, 146, 150, 253, 193, 22, 151, 253, 193, 138, 151, 253, 193, 140, 151, 253, 193, 240, 151, 253, 193, 66, 152, 253, 193, 68, 152, 253, 193, 144, 152, 253, 193, 192, 152, 253, 193, 14, 153, 253, 193, 86, 153, 253, 193, 108, 153, 253, 193, 126, 153, 253, 193, 128, 153, 253, 193, 156, 153, 253, 193, 158, 153, 253, 193, 160, 153, 253, 193, 162, 153, 253, 193, 206, 153, 253, 193, 208, 153, 253, 193, 230, 153, 253, 193, 232, 153, 253, 193, 234, 153, 253, 193, 16, 154, 253, 193, 38, 154, 253, 193, 54, 154, 253, 193, 76, 154, 253, 193, 160, 154, 253, 193, 220, 154, 253, 193, 160, 155, 253, 193, 240, 155, 253, 193, 242, 155, 253, 193, 166, 150, 253, 193, 168, 150, 253, 193, 170, 150, 253, 193, 172, 150, 253, 193, 174, 150, 253, 193, 176, 150, 253, 193, 178, 150, 253, 193, 180, 150, 253, 193, 182, 150, 253, 193, 184, 150, 253, 193, 186, 150, 253, 193, 46, 151, 253, 193, 48, 151, 253, 193, 50, 151, 253, 193, 52, 151, 253, 193, 54, 151, 253, 193, 56, 151, 253, 193, 58, 151, 253, 193, 60, 151, 253, 193, 62, 151, 253, 193, 64, 151, 253, 193, 66, 151, 253, 193, 150, 151, 253, 193, 152, 151, 253, 193, 154, 151, 253, 193, 156, 151, 253, 193, 158, 151, 253, 193, 160, 151, 253, 193, 162, 151, 253, 193, 164, 151, 253, 193, 166, 151, 253, 193, 4, 152, 253, 193, 212, 150, 253, 193, 214, 150, 253, 193, 216, 150, 253, 193, 218, 150, 253, 193, 220, 150, 253, 193, 222, 150, 253, 193, 224, 150, 253, 193, 226, 150, 253, 193, 228, 150, 253, 193, 230, 150, 253, 193, 232, 150, 253, 193, 234, 150, 253, 193, 236, 150, 253, 193, 238, 150, 253, 193, 240, 150, 253, 193, 242, 150, 253, 193, 244, 150, 253, 193, 246, 150, 253, 193, 248, 150, 253, 193, 250, 150, 253, 193, 252, 150, 253, 193, 254, 150, 253, 193, 2, 151, 253, 193, 4, 151, 253, 193, 6, 151, 253, 193, 8, 151, 253, 193, 10, 151, 253, 193, 12, 151, 253, 193, 14, 151, 253, 193, 16, 151, 253, 193, 18, 151, 253, 193, 20, 151, 253, 193, 88, 151, 253, 193, 90, 151, 253, 193, 92, 151, 253, 193, 94, 151, 253, 193, 96, 151, 253, 193, 98, 151, 253, 193, 100, 151, 253, 193, 102, 151, 253, 193, 104, 151, 253, 193, 106, 151, 253, 193, 108, 151, 253, 193, 110, 151, 253, 193, 112, 151, 253, 193, 114, 151, 253, 193, 116, 151, 253, 193, 118, 151, 253, 193, 120, 151, 253, 193, 122, 151, 253, 193, 124, 151, 253, 193, 126, 151, 253, 193, 128, 151, 253, 193, 130, 151, 253, 193, 132, 151, 253, 193, 134, 151, 253, 193, 136, 151, 253, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 193, 172, 151, 253, 193, 174, 151, 253, 193, 176, 151, 253, 193, 178, 151, 253, 193, 180, 151, 253, 193, 182, 151, 253, 193, 184, 151, 253, 193, 186, 151, 253, 193, 14, 152, 253, 193, 16, 152, 253, 193, 18, 152, 253, 193, 20, 152, 253, 193, 88, 152, 253, 193, 90, 152, 253, 193, 92, 152, 253, 193, 94, 152, 253, 193, 96, 152, 253, 193, 98, 152, 253, 193, 160, 152, 253, 193, 162, 152, 253, 193, 164, 152, 253, 193, 248, 152, 253, 193, 250, 152, 253, 193, 26, 153, 253, 193, 84, 154, 253, 193, 96, 154, 253, 193, 118, 154, 253, 193, 120, 154, 253, 193, 122, 154, 253, 193, 124, 154, 253, 193, 176, 154, 253, 193, 178, 154, 253, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 224, 109, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 193, 6, 152, 253, 193, 8, 152, 253, 193, 10, 152, 253, 193, 12, 152, 253, 193, 78, 152, 253, 193, 80, 152, 253, 193, 82, 152, 253, 193, 84, 152, 253, 193, 86, 152, 253, 193, 154, 152, 253, 193, 156, 152, 253, 193, 158, 152, 253, 193, 200, 152, 253, 193, 202, 152, 253, 193, 204, 152, 253, 193, 206, 152, 253, 193, 246, 152, 253, 193, 22, 153, 253, 193, 24, 153, 253, 193, 54, 153, 253, 193, 56, 153, 253, 193, 70, 153, 253, 193, 72, 153, 253, 193, 74, 153, 253, 193, 78, 153, 253, 193, 80, 153, 253, 193, 82, 153, 253, 193, 84, 153, 253, 206, 32, 110, 0, 206, 32, 110, 0, 206, 32, 110, 0, 206, 32, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 206, 0, 110, 0, 193, 100, 152, 253, 193, 102, 152, 253, 193, 104, 152, 253, 193, 106, 152, 253, 193, 108, 152, 253, 193, 110, 152, 253, 193, 112, 152, 253, 193, 114, 152, 253, 193, 116, 152, 253, 193, 118, 152, 253, 193, 120, 152, 253, 193, 122, 152, 253, 193, 124, 152, 253, 193, 126, 152, 253, 193, 128, 152, 253, 193, 130, 152, 253, 193, 132, 152, 253, 193, 134, 152, 253, 193, 136, 152, 253, 193, 138, 152, 253, 193, 140, 152, 253, 193, 142, 152, 253, 193, 166, 152, 253, 193, 168, 152, 253, 193, 170, 152, 253, 193, 172, 152, 253, 193, 174, 152, 253, 193, 176, 152, 253, 193, 178, 152, 253, 193, 180, 152, 253, 193, 182, 152, 253, 193, 184, 152, 253, 193, 186, 152, 253, 193, 188, 152, 253, 193, 190, 152, 253, 193, 208, 152, 253, 193, 210, 152, 253, 193, 212, 152, 253, 193, 214, 152, 253, 193, 216, 152, 253, 193, 218, 152, 253, 193, 220, 152, 253, 193, 222, 152, 253, 193, 224, 152, 253, 193, 226, 152, 253, 193, 228, 152, 253, 193, 230, 152, 253, 193, 232, 152, 253, 193, 234, 152, 253, 193, 236, 152, 253, 193, 238, 152, 253, 193, 240, 152, 253, 193, 242, 152, 253, 193, 252, 152, 253, 193, 254, 152, 253, 193, 196, 152, 253, 193, 198, 152, 253, 193, 18, 153, 253, 193, 20, 153, 253, 193, 34, 153, 253, 193, 36, 153, 253, 193, 90, 154, 253, 193, 162, 154, 253, 193, 222, 154, 253, 193, 16, 155, 253, 193, 18, 155, 253, 193, 20, 155, 253, 193, 78, 155, 253, 193, 162, 155, 253, 193, 164, 155, 253, 193, 166, 155, 253, 193, 58, 156, 253, 193, 96, 156, 253, 193, 168, 156, 253, 193, 230, 156, 253, 193, 62, 157, 253, 193, 98, 157, 253, 193, 166, 157, 253, 193, 246, 157, 253, 193, 248, 157, 253, 193, 194, 158, 253, 193, 196, 158, 253, 193, 198, 158, 253, 193, 48, 159, 253, 193, 52, 160, 253, 193, 54, 160, 253, 193, 56, 160, 253, 193, 2, 153, 253, 193, 4, 153, 253, 193, 6, 153, 253, 193, 8, 153, 253, 193, 10, 153, 253, 193, 12, 153, 253, 193, 28, 153, 253, 193, 30, 153, 253, 193, 32, 153, 253, 193, 38, 153, 253, 193, 40, 153, 253, 193, 42, 153, 253, 193, 44, 153, 253, 193, 46, 153, 253, 193, 48, 153, 253, 193, 58, 153, 253, 193, 60, 153, 253, 193, 62, 153, 253, 193, 64, 153, 253, 193, 66, 153, 253, 193, 36, 206, 253, 193, 226, 153, 253, 193, 228, 153, 253, 193, 68, 154, 253, 193, 86, 154, 253, 193, 88, 154, 253, 193, 98, 154, 253, 193, 100, 154, 253, 193, 102, 154, 253, 193, 104, 154, 253, 193, 106, 154, 253, 193, 108, 154, 253, 193, 88, 153, 253, 193, 116, 153, 253, 193, 140, 153, 253, 193, 180, 153, 253, 193, 182, 153, 253, 193, 184, 153, 253, 193, 186, 153, 253, 193, 188, 153, 253, 193, 190, 153, 253, 193, 216, 153, 253, 193, 218, 153, 253, 193, 220, 153, 253, 193, 222, 153, 253, 193, 250, 153, 253, 193, 252, 153, 253, 193, 254, 153, 253, 193, 2, 154, 253, 193, 24, 154, 253, 193, 26, 154, 253, 193, 28, 154, 253, 193, 42, 154, 253, 193, 44, 154, 253, 193, 46, 154, 253, 193, 48, 154, 253, 193, 50, 154, 253, 193, 52, 154, 253, 193, 56, 154, 253, 193, 64, 154, 253, 193, 164, 154, 253, 193, 224, 154, 253, 193, 226, 154, 253, 193, 228, 154, 253, 206, 32, 110, 0, 206, 32, 110, 0, 206, 32, 110, 0, 206, 32, 110, 0, 206, 32, 110, 0, 193, 118, 153, 253, 193, 120, 153, 253, 193, 122, 153, 253, 193, 124, 153, 253, 193, 142, 153, 253, 193, 144, 153, 253, 193, 146, 153, 253, 193, 148, 153, 253, 193, 150, 153, 253, 193, 152, 153, 253, 193, 154, 153, 253, 193, 192, 153, 253, 193, 194, 153, 253, 193, 196, 153, 253, 193, 198, 153, 253, 193, 200, 153, 253, 193, 202, 153, 253, 193, 204, 153, 253, 193, 224, 153, 253, 193, 4, 154, 253, 193, 6, 154, 253, 193, 8, 154, 253, 193, 10, 154, 253, 193, 12, 154, 253, 193, 14, 154, 253, 193, 30, 154, 253, 193, 32, 154, 253, 193, 174, 153, 253, 193, 176, 153, 253, 193, 178, 153, 253, 193, 210, 153, 253, 193, 212, 153, 253, 193, 214, 153, 253, 193, 236, 153, 253, 193, 238, 153, 253, 193, 240, 153, 253, 193, 242, 153, 253, 193, 244, 153, 253, 193, 246, 153, 253, 193, 248, 153, 253, 193, 18, 154, 253, 193, 20, 154, 253, 193, 22, 154, 253, 193, 40, 154, 253, 193, 62, 154, 253, 193, 72, 154, 253, 193, 80, 154, 253, 193, 244, 155, 253, 193, 252, 156, 253, 193, 254, 156, 253, 193, 2, 157, 253, 193, 18, 157, 253, 193, 20, 157, 253, 193, 22, 157, 253, 193, 24, 157, 253, 193, 30, 157, 253, 193, 42, 157, 253, 193, 56, 157, 253, 193, 58, 157, 253, 193, 34, 154, 253, 193, 36, 154, 253, 193, 58, 154, 253, 193, 60, 154, 253, 193, 66, 154, 253, 193, 70, 154, 253, 193, 74, 154, 253, 193, 78, 154, 253, 193, 82, 154, 253, 193, 92, 154, 253, 193, 94, 154, 253, 193, 114, 154, 253, 193, 116, 154, 253, 193, 166, 154, 253, 193, 168, 154, 253, 193, 170, 154, 253, 193, 172, 154, 253, 193, 174, 154, 253, 193, 230, 154, 253, 193, 232, 154, 253, 193, 22, 155, 253, 193, 24, 155, 253, 193, 86, 155, 253, 193, 88, 155, 253, 193, 176, 155, 253, 193, 178, 155, 253, 193, 180, 155, 253, 193, 252, 155, 253, 193, 254, 155, 253, 193, 2, 156, 253, 193, 4, 156, 253, 193, 6, 156, 253, 193, 180, 154, 253, 193, 182, 154, 253, 193, 184, 154, 253, 193, 186, 154, 253, 193, 234, 154, 253, 193, 236, 154, 253, 193, 26, 155, 253, 193, 28, 155, 253, 193, 30, 155, 253, 193, 32, 155, 253, 193, 34, 155, 253, 193, 90, 155, 253, 193, 92, 155, 253, 193, 94, 155, 253, 193, 96, 155, 253, 193, 98, 155, 253, 193, 100, 155, 253, 193, 182, 155, 253, 193, 184, 155, 253, 193, 186, 155, 253, 193, 188, 155, 253, 193, 190, 155, 253, 193, 192, 155, 253, 193, 194, 155, 253, 193, 10, 156, 253, 193, 12, 156, 253, 193, 14, 156, 253, 193, 70, 156, 253, 193, 72, 156, 253, 193, 116, 156, 253, 193, 174, 156, 253, 193, 184, 156, 253, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 206, 64, 110, 0, 193, 240, 154, 253, 193, 242, 154, 253, 193, 244, 154, 253, 193, 246, 154, 253, 193, 248, 154, 253, 193, 250, 154, 253, 193, 252, 154, 253, 193, 254, 154, 253, 193, 2, 155, 253, 193, 4, 155, 253, 193, 6, 155, 253, 193, 8, 155, 253, 193, 10, 155, 253, 193, 12, 155, 253, 193, 14, 155, 253, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 96, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 193, 80, 155, 253, 193, 82, 155, 253, 193, 84, 155, 253, 193, 168, 155, 253, 193, 170, 155, 253, 193, 172, 155, 253, 193, 174, 155, 253, 193, 246, 155, 253, 193, 248, 155, 253, 193, 250, 155, 253, 193, 60, 156, 253, 193, 98, 156, 253, 193, 100, 156, 253, 193, 102, 156, 253, 193, 170, 156, 253, 193, 172, 156, 253, 193, 198, 156, 253, 193, 214, 156, 253, 193, 224, 156, 253, 193, 226, 156, 253, 193, 246, 156, 253, 193, 4, 157, 253, 193, 26, 157, 253, 193, 28, 157, 253, 193, 32, 157, 253, 193, 34, 157, 253, 193, 36, 157, 253, 193, 44, 157, 253, 193, 64, 157, 253, 193, 66, 157, 253, 193, 82, 157, 253, 193, 100, 157, 253, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 206, 128, 110, 0, 193, 196, 155, 253, 193, 198, 155, 253, 193, 200, 155, 253, 193, 202, 155, 253, 193, 204, 155, 253, 193, 206, 155, 253, 193, 208, 155, 253, 193, 210, 155, 253, 193, 212, 155, 253, 193, 16, 156, 253, 193, 214, 155, 253, 193, 216, 155, 253, 193, 218, 155, 253, 193, 220, 155, 253, 193, 222, 155, 253, 193, 224, 155, 253, 193, 226, 155, 253, 193, 228, 155, 253, 193, 230, 155, 253, 193, 232, 155, 253, 193, 234, 155, 253, 193, 236, 155, 253, 193, 238, 155, 253, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 193, 8, 156, 253, 193, 62, 156, 253, 193, 64, 156, 253, 193, 66, 156, 253, 193, 68, 156, 253, 193, 104, 156, 253, 193, 106, 156, 253, 193, 108, 156, 253, 193, 110, 156, 253, 193, 112, 156, 253, 193, 114, 156, 253, 193, 150, 156, 253, 193, 216, 156, 253, 193, 218, 156, 253, 193, 242, 156, 253, 193, 244, 156, 253, 193, 6, 157, 253, 193, 8, 157, 253, 193, 10, 157, 253, 193, 46, 157, 253, 193, 68, 157, 253, 193, 84, 157, 253, 193, 70, 157, 253, 193, 86, 157, 253, 193, 96, 157, 253, 193, 102, 157, 253, 193, 104, 157, 253, 193, 122, 157, 253, 193, 148, 157, 253, 193, 160, 157, 253, 193, 164, 157, 253, 193, 170, 157, 253, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 206, 160, 110, 0, 193, 74, 156, 253, 193, 76, 156, 253, 193, 78, 156, 253, 193, 80, 156, 253, 193, 82, 156, 253, 193, 84, 156, 253, 193, 86, 156, 253, 193, 88, 156, 253, 193, 90, 156, 253, 193, 92, 156, 253, 193, 94, 156, 253, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 193, 52, 156, 253, 193, 54, 156, 253, 193, 56, 156, 253, 193, 164, 156, 253, 193, 166, 156, 253, 193, 182, 156, 253, 193, 196, 156, 253, 193, 248, 156, 253, 193, 40, 157, 253, 193, 54, 157, 253, 193, 76, 157, 253, 193, 116, 157, 253, 193, 168, 157, 253, 193, 70, 158, 253, 193, 192, 158, 253, 193, 34, 159, 253, 193, 36, 159, 253, 193, 38, 159, 253, 193, 40, 159, 253, 193, 150, 159, 253, 193, 196, 160, 253, 193, 198, 160, 253, 193, 200, 160, 253, 193, 202, 160, 253, 193, 76, 161, 253, 193, 186, 161, 253, 193, 188, 161, 253, 193, 36, 162, 253, 193, 178, 162, 253, 193, 180, 162, 253, 193, 34, 163, 253, 193, 62, 163, 253, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 206, 192, 110, 0, 193, 152, 156, 253, 193, 154, 156, 253, 193, 156, 156, 253, 193, 158, 156, 253, 193, 160, 156, 253, 193, 162, 156, 253, 193, 176, 156, 253, 193, 178, 156, 253, 193, 180, 156, 253, 193, 186, 156, 253, 193, 188, 156, 253, 193, 190, 156, 253, 193, 192, 156, 253, 193, 194, 156, 253, 193, 200, 156, 253, 193, 202, 156, 253, 193, 204, 156, 253, 193, 206, 156, 253, 193, 208, 156, 253, 193, 210, 156, 253, 193, 212, 156, 253, 193, 220, 156, 253, 193, 222, 156, 253, 193, 228, 156, 253, 193, 232, 156, 253, 193, 234, 156, 253, 193, 236, 156, 253, 193, 238, 156, 253, 193, 240, 156, 253, 193, 12, 157, 253, 193, 14, 157, 253, 193, 38, 157, 253, 193, 48, 157, 253, 193, 50, 157, 253, 193, 52, 157, 253, 193, 72, 157, 253, 193, 74, 157, 253, 193, 88, 157, 253, 193, 90, 157, 253, 193, 106, 157, 253, 193, 114, 157, 253, 193, 124, 157, 253, 193, 126, 157, 253, 193, 128, 157, 253, 193, 132, 157, 253, 193, 138, 157, 253, 193, 140, 157, 253, 193, 142, 157, 253, 193, 144, 157, 253, 193, 146, 157, 253, 193, 150, 157, 253, 193, 152, 157, 253, 193, 154, 157, 253, 193, 156, 157, 253, 193, 162, 157, 253, 193, 176, 157, 253, 193, 60, 157, 253, 193, 78, 157, 253, 193, 80, 157, 253, 193, 92, 157, 253, 193, 94, 157, 253, 193, 108, 157, 253, 193, 110, 157, 253, 193, 118, 157, 253, 193, 120, 157, 253, 193, 130, 157, 253, 193, 158, 157, 253, 193, 72, 158, 253, 193, 74, 158, 253, 193, 76, 158, 253, 193, 78, 158, 253, 193, 42, 159, 253, 193, 44, 159, 253, 193, 46, 159, 253, 193, 152, 159, 253, 193, 154, 159, 253, 193, 156, 159, 253, 193, 48, 160, 253, 193, 50, 160, 253, 193, 204, 160, 253, 193, 206, 160, 253, 193, 208, 160, 253, 193, 190, 161, 253, 193, 38, 162, 253, 193, 40, 162, 253, 193, 42, 162, 253, 193, 44, 162, 253, 193, 114, 162, 253, 193, 112, 157, 253, 193, 250, 157, 253, 193, 252, 157, 253, 193, 254, 157, 253, 193, 80, 158, 253, 193, 50, 159, 253, 193, 52, 159, 253, 193, 54, 159, 253, 193, 158, 159, 253, 193, 160, 159, 253, 193, 60, 160, 253, 193, 62, 160, 253, 193, 86, 161, 253, 193, 88, 161, 253, 193, 90, 161, 253, 193, 200, 161, 253, 193, 48, 162, 253, 193, 40, 163, 253, 193, 70, 163, 253, 193, 72, 163, 253, 193, 74, 163, 253, 193, 116, 163, 253, 193, 118, 163, 253, 193, 120, 163, 253, 193, 156, 163, 253, 193, 158, 163, 253, 193, 160, 163, 253, 193, 162, 163, 253, 193, 228, 163, 253, 193, 230, 163, 253, 193, 232, 163, 253, 193, 234, 163, 253, 193, 134, 157, 253, 193, 136, 157, 253, 193, 194, 157, 253, 193, 212, 157, 253, 193, 214, 157, 253, 193, 216, 157, 253, 193, 22, 158, 253, 193, 24, 158, 253, 193, 26, 158, 253, 193, 28, 158, 253, 193, 30, 158, 253, 193, 106, 158, 253, 193, 108, 158, 253, 193, 110, 158, 253, 193, 112, 158, 253, 193, 114, 158, 253, 193, 116, 158, 253, 193, 118, 158, 253, 193, 120, 158, 253, 193, 122, 158, 253, 193, 124, 158, 253, 193, 234, 158, 253, 193, 236, 158, 253, 193, 238, 158, 253, 193, 240, 158, 253, 193, 242, 158, 253, 193, 244, 158, 253, 193, 246, 158, 253, 193, 78, 159, 253, 193, 80, 159, 253, 193, 82, 159, 253, 193, 84, 159, 253, 193, 172, 157, 253, 193, 186, 157, 253, 193, 188, 157, 253, 193, 190, 157, 253, 193, 192, 157, 253, 193, 204, 157, 253, 193, 206, 157, 253, 193, 208, 157, 253, 193, 210, 157, 253, 193, 2, 158, 253, 193, 4, 158, 253, 193, 6, 158, 253, 193, 8, 158, 253, 193, 10, 158, 253, 193, 12, 158, 253, 193, 14, 158, 253, 193, 16, 158, 253, 193, 18, 158, 253, 193, 20, 158, 253, 193, 82, 158, 253, 193, 84, 158, 253, 193, 86, 158, 253, 193, 88, 158, 253, 193, 90, 158, 253, 193, 92, 158, 253, 193, 94, 158, 253, 193, 200, 158, 253, 193, 96, 158, 253, 193, 98, 158, 253, 193, 100, 158, 253, 193, 102, 158, 253, 193, 104, 158, 253, 193, 178, 157, 253, 193, 180, 157, 253, 193, 184, 157, 253, 193, 182, 157, 253, 193, 198, 157, 253, 193, 200, 157, 253, 193, 202, 157, 253, 193, 218, 157, 253, 193, 220, 157, 253, 193, 222, 157, 253, 193, 224, 157, 253, 193, 226, 157, 253, 193, 228, 157, 253, 193, 230, 157, 253, 193, 232, 157, 253, 193, 234, 157, 253, 193, 236, 157, 253, 193, 238, 157, 253, 193, 240, 157, 253, 193, 242, 157, 253, 193, 244, 157, 253, 206, 224, 110, 0, 206, 224, 110, 0, 206, 224, 110, 0, 206, 224, 110, 0, 206, 224, 110, 0, 206, 224, 110, 0, 206, 224, 110, 0, 206, 224, 110, 0, 206, 224, 110, 0, 206, 224, 110, 0, 206, 224, 110, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 206, 0, 111, 0, 193, 248, 158, 253, 193, 250, 158, 253, 193, 252, 158, 253, 193, 254, 158, 253, 193, 2, 159, 253, 193, 4, 159, 253, 193, 6, 159, 253, 193, 8, 159, 253, 193, 10, 159, 253, 193, 12, 159, 253, 193, 14, 159, 253, 193, 16, 159, 253, 193, 18, 159, 253, 193, 20, 159, 253, 193, 22, 159, 253, 193, 24, 159, 253, 193, 26, 159, 253, 193, 28, 159, 253, 193, 30, 159, 253, 193, 32, 159, 253, 193, 92, 159, 253, 193, 94, 159, 253, 193, 96, 159, 253, 193, 202, 158, 253, 193, 204, 158, 253, 193, 206, 158, 253, 193, 208, 158, 253, 193, 210, 158, 253, 193, 212, 158, 253, 193, 214, 158, 253, 193, 216, 158, 253, 193, 218, 158, 253, 193, 220, 158, 253, 193, 222, 158, 253, 193, 224, 158, 253, 193, 226, 158, 253, 193, 228, 158, 253, 193, 230, 158, 253, 193, 232, 158, 253, 193, 56, 159, 253, 193, 58, 159, 253, 193, 60, 159, 253, 193, 62, 159, 253, 193, 64, 159, 253, 193, 66, 159, 253, 193, 68, 159, 253, 193, 70, 159, 253, 193, 72, 159, 253, 193, 74, 159, 253, 193, 162, 159, 253, 193, 164, 159, 253, 193, 166, 159, 253, 193, 76, 159, 253, 193, 168, 159, 253, 193, 170, 159, 253, 193, 86, 159, 253, 193, 88, 159, 253, 193, 90, 159, 253, 193, 192, 159, 253, 193, 194, 159, 253, 193, 196, 159, 253, 193, 198, 159, 253, 193, 200, 159, 253, 193, 90, 160, 253, 193, 92, 160, 253, 193, 94, 160, 253, 193, 96, 160, 253, 193, 98, 160, 253, 193, 100, 160, 253, 193, 102, 160, 253, 193, 104, 160, 253, 193, 250, 160, 253, 193, 252, 160, 253, 193, 254, 160, 253, 193, 2, 161, 253, 193, 4, 161, 253, 193, 104, 161, 253, 193, 106, 161, 253, 193, 108, 161, 253, 193, 222, 161, 253, 193, 224, 161, 253, 193, 226, 161, 253, 193, 228, 161, 253, 193, 230, 161, 253, 193, 62, 162, 253, 193, 64, 162, 253, 193, 66, 162, 253, 193, 98, 159, 253, 193, 100, 159, 253, 193, 102, 159, 253, 193, 104, 159, 253, 193, 106, 159, 253, 193, 108, 159, 253, 193, 110, 159, 253, 193, 112, 159, 253, 193, 114, 159, 253, 193, 116, 159, 253, 193, 118, 159, 253, 193, 120, 159, 253, 193, 122, 159, 253, 193, 124, 159, 253, 193, 126, 159, 253, 193, 128, 159, 253, 193, 130, 159, 253, 193, 132, 159, 253, 193, 134, 159, 253, 193, 136, 159, 253, 193, 138, 159, 253, 193, 140, 159, 253, 193, 142, 159, 253, 193, 144, 159, 253, 193, 146, 159, 253, 193, 148, 159, 253, 193, 202, 159, 253, 193, 204, 159, 253, 193, 206, 159, 253, 193, 208, 159, 253, 193, 210, 159, 253, 193, 212, 159, 253, 193, 172, 159, 253, 193, 174, 159, 253, 193, 176, 159, 253, 193, 178, 159, 253, 193, 180, 159, 253, 193, 182, 159, 253, 193, 184, 159, 253, 193, 186, 159, 253, 193, 188, 159, 253, 193, 190, 159, 253, 193, 64, 160, 253, 193, 66, 160, 253, 193, 68, 160, 253, 193, 70, 160, 253, 193, 72, 160, 253, 193, 74, 160, 253, 193, 76, 160, 253, 193, 78, 160, 253, 193, 80, 160, 253, 193, 82, 160, 253, 193, 84, 160, 253, 193, 86, 160, 253, 193, 224, 160, 253, 193, 88, 160, 253, 206, 64, 111, 0, 206, 64, 111, 0, 206, 64, 111, 0, 206, 64, 111, 0, 206, 64, 111, 0, 206, 64, 111, 0, 206, 64, 111, 0, 206, 64, 111, 0, 193, 214, 159, 253, 193, 216, 159, 253, 193, 218, 159, 253, 193, 220, 159, 253, 193, 222, 159, 253, 193, 224, 159, 253, 193, 226, 159, 253, 193, 228, 159, 253, 193, 230, 159, 253, 193, 232, 159, 253, 193, 234, 159, 253, 193, 236, 159, 253, 193, 238, 159, 253, 193, 240, 159, 253, 193, 242, 159, 253, 193, 244, 159, 253, 193, 246, 159, 253, 193, 248, 159, 253, 193, 250, 159, 253, 193, 252, 159, 253, 193, 254, 159, 253, 193, 2, 160, 253, 193, 4, 160, 253, 193, 6, 160, 253, 193, 8, 160, 253, 193, 10, 160, 253, 193, 12, 160, 253, 193, 14, 160, 253, 193, 16, 160, 253, 193, 18, 160, 253, 193, 20, 160, 253, 193, 22, 160, 253, 193, 24, 160, 253, 193, 26, 160, 253, 193, 28, 160, 253, 193, 106, 160, 253, 193, 30, 160, 253, 193, 32, 160, 253, 193, 34, 160, 253, 193, 36, 160, 253, 193, 38, 160, 253, 193, 40, 160, 253, 193, 42, 160, 253, 193, 44, 160, 253, 193, 46, 160, 253, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 193, 58, 160, 253, 193, 210, 160, 253, 193, 212, 160, 253, 193, 214, 160, 253, 193, 216, 160, 253, 193, 218, 160, 253, 193, 220, 160, 253, 193, 222, 160, 253, 193, 78, 161, 253, 193, 80, 161, 253, 193, 82, 161, 253, 193, 84, 161, 253, 193, 192, 161, 253, 193, 194, 161, 253, 193, 196, 161, 253, 193, 198, 161, 253, 193, 46, 162, 253, 193, 118, 162, 253, 193, 120, 162, 253, 193, 122, 162, 253, 193, 124, 162, 253, 193, 126, 162, 253, 193, 204, 162, 253, 193, 206, 162, 253, 193, 228, 162, 253, 193, 240, 162, 253, 193, 242, 162, 253, 193, 152, 163, 253, 193, 154, 163, 253, 193, 14, 164, 253, 193, 88, 164, 253, 193, 122, 164, 253, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 32, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 64, 111, 0, 206, 64, 111, 0, 206, 64, 111, 0, 206, 64, 111, 0, 193, 92, 161, 253, 193, 94, 161, 253, 193, 96, 161, 253, 193, 98, 161, 253, 193, 100, 161, 253, 193, 102, 161, 253, 193, 202, 161, 253, 193, 204, 161, 253, 193, 206, 161, 253, 193, 208, 161, 253, 193, 210, 161, 253, 193, 212, 161, 253, 193, 214, 161, 253, 193, 216, 161, 253, 193, 218, 161, 253, 193, 220, 161, 253, 193, 50, 162, 253, 193, 52, 162, 253, 193, 54, 162, 253, 193, 56, 162, 253, 193, 58, 162, 253, 193, 60, 162, 253, 193, 128, 162, 253, 193, 130, 162, 253, 193, 132, 162, 253, 193, 208, 162, 253, 193, 230, 162, 253, 193, 232, 162, 253, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 206, 96, 111, 0, 193, 112, 161, 253, 193, 60, 161, 253, 193, 62, 161, 253, 193, 64, 161, 253, 193, 66, 161, 253, 193, 68, 161, 253, 193, 70, 161, 253, 193, 72, 161, 253, 193, 74, 161, 253, 193, 114, 161, 253, 193, 116, 161, 253, 193, 118, 161, 253, 193, 120, 161, 253, 193, 122, 161, 253, 193, 124, 161, 253, 193, 126, 161, 253, 193, 128, 161, 253, 193, 130, 161, 253, 193, 132, 161, 253, 193, 134, 161, 253, 193, 136, 161, 253, 193, 138, 161, 253, 193, 140, 161, 253, 193, 142, 161, 253, 193, 144, 161, 253, 193, 146, 161, 253, 193, 148, 161, 253, 193, 150, 161, 253, 193, 152, 161, 253, 193, 154, 161, 253, 193, 156, 161, 253, 193, 158, 161, 253, 193, 160, 161, 253, 193, 162, 161, 253, 193, 164, 161, 253, 193, 166, 161, 253, 193, 168, 161, 253, 193, 170, 161, 253, 193, 172, 161, 253, 193, 174, 161, 253, 193, 176, 161, 253, 193, 178, 161, 253, 193, 180, 161, 253, 193, 182, 161, 253, 193, 184, 161, 253, 193, 234, 161, 253, 193, 236, 161, 253, 193, 238, 161, 253, 193, 240, 161, 253, 193, 242, 161, 253, 193, 244, 161, 253, 193, 246, 161, 253, 193, 248, 161, 253, 193, 250, 161, 253, 193, 252, 161, 253, 193, 254, 161, 253, 193, 2, 162, 253, 193, 4, 162, 253, 193, 6, 162, 253, 193, 8, 162, 253, 193, 10, 162, 253, 193, 12, 162, 253, 193, 14, 162, 253, 193, 16, 162, 253, 193, 18, 162, 253, 193, 20, 162, 253, 193, 70, 162, 253, 193, 22, 162, 253, 193, 24, 162, 253, 193, 26, 162, 253, 193, 28, 162, 253, 193, 30, 162, 253, 193, 32, 162, 253, 193, 34, 162, 253, 193, 72, 162, 253, 193, 74, 162, 253, 193, 68, 162, 253, 193, 136, 162, 253, 193, 138, 162, 253, 193, 140, 162, 253, 193, 142, 162, 253, 193, 182, 162, 253, 193, 234, 162, 253, 193, 236, 162, 253, 193, 244, 162, 253, 193, 8, 163, 253, 193, 76, 164, 253, 193, 144, 164, 253, 193, 146, 164, 253, 193, 206, 164, 253, 193, 220, 164, 253, 193, 20, 165, 253, 193, 66, 165, 253, 193, 68, 165, 253, 193, 70, 165, 253, 193, 202, 165, 253, 193, 220, 165, 253, 193, 254, 165, 253, 193, 2, 166, 253, 193, 66, 166, 253, 193, 68, 166, 253, 193, 70, 166, 253, 193, 72, 166, 253, 193, 74, 166, 253, 193, 76, 166, 253, 193, 78, 166, 253, 193, 80, 166, 253, 193, 192, 166, 253, 193, 76, 162, 253, 193, 78, 162, 253, 193, 80, 162, 253, 193, 82, 162, 253, 193, 84, 162, 253, 193, 86, 162, 253, 193, 88, 162, 253, 193, 90, 162, 253, 193, 92, 162, 253, 193, 94, 162, 253, 193, 96, 162, 253, 193, 98, 162, 253, 193, 100, 162, 253, 193, 102, 162, 253, 193, 104, 162, 253, 193, 106, 162, 253, 193, 108, 162, 253, 193, 110, 162, 253, 193, 112, 162, 253, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 193, 116, 162, 253, 193, 202, 162, 253, 193, 254, 162, 253, 193, 2, 163, 253, 193, 10, 163, 253, 193, 26, 163, 253, 193, 28, 163, 253, 193, 36, 163, 253, 193, 38, 163, 253, 193, 66, 163, 253, 193, 68, 163, 253, 193, 98, 163, 253, 193, 100, 163, 253, 193, 102, 163, 253, 193, 104, 163, 253, 193, 106, 163, 253, 193, 108, 163, 253, 193, 110, 163, 253, 193, 112, 163, 253, 193, 114, 163, 253, 193, 144, 163, 253, 193, 146, 163, 253, 193, 148, 163, 253, 193, 150, 163, 253, 193, 188, 163, 253, 193, 190, 163, 253, 193, 192, 163, 253, 193, 194, 163, 253, 193, 214, 163, 253, 193, 216, 163, 253, 193, 218, 163, 253, 193, 220, 163, 253, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 206, 128, 111, 0, 193, 184, 162, 253, 193, 186, 162, 253, 193, 188, 162, 253, 193, 190, 162, 253, 193, 192, 162, 253, 193, 194, 162, 253, 193, 196, 162, 253, 193, 198, 162, 253, 193, 200, 162, 253, 193, 210, 162, 253, 193, 212, 162, 253, 193, 214, 162, 253, 193, 216, 162, 253, 193, 218, 162, 253, 193, 220, 162, 253, 193, 222, 162, 253, 193, 224, 162, 253, 193, 226, 162, 253, 193, 238, 162, 253, 193, 246, 162, 253, 193, 248, 162, 253, 193, 250, 162, 253, 193, 252, 162, 253, 193, 12, 163, 253, 193, 16, 163, 253, 193, 18, 163, 253, 193, 56, 163, 253, 193, 58, 163, 253, 193, 4, 163, 253, 193, 6, 163, 253, 193, 14, 163, 253, 193, 20, 163, 253, 193, 22, 163, 253, 193, 24, 163, 253, 193, 30, 163, 253, 193, 32, 163, 253, 193, 42, 163, 253, 193, 44, 163, 253, 193, 46, 163, 253, 193, 48, 163, 253, 193, 50, 163, 253, 193, 52, 163, 253, 193, 54, 163, 253, 193, 76, 163, 253, 193, 78, 163, 253, 193, 80, 163, 253, 193, 82, 163, 253, 193, 84, 163, 253, 193, 86, 163, 253, 193, 122, 163, 253, 193, 124, 163, 253, 193, 126, 163, 253, 193, 128, 163, 253, 193, 130, 163, 253, 193, 132, 163, 253, 193, 134, 163, 253, 193, 136, 163, 253, 193, 164, 163, 253, 193, 166, 163, 253, 193, 168, 163, 253, 193, 60, 163, 253, 193, 88, 163, 253, 193, 90, 163, 253, 193, 138, 163, 253, 193, 176, 163, 253, 193, 178, 163, 253, 193, 204, 163, 253, 193, 180, 163, 253, 193, 206, 163, 253, 193, 208, 163, 253, 193, 244, 163, 253, 193, 246, 163, 253, 193, 38, 164, 253, 193, 62, 164, 253, 193, 64, 164, 253, 193, 78, 164, 253, 193, 112, 164, 253, 193, 116, 164, 253, 193, 118, 164, 253, 193, 120, 164, 253, 193, 124, 164, 253, 193, 128, 164, 253, 193, 130, 164, 253, 193, 132, 164, 253, 193, 134, 164, 253, 193, 136, 164, 253, 193, 148, 164, 253, 193, 150, 164, 253, 193, 152, 164, 253, 193, 154, 164, 253, 193, 156, 164, 253, 193, 158, 164, 253, 193, 64, 163, 253, 193, 92, 163, 253, 193, 94, 163, 253, 193, 96, 163, 253, 193, 140, 163, 253, 193, 142, 163, 253, 193, 182, 163, 253, 193, 184, 163, 253, 193, 186, 163, 253, 193, 210, 163, 253, 193, 212, 163, 253, 193, 248, 163, 253, 193, 250, 163, 253, 193, 252, 163, 253, 193, 254, 163, 253, 193, 2, 164, 253, 193, 40, 164, 253, 193, 42, 164, 253, 193, 80, 164, 253, 193, 82, 164, 253, 193, 94, 164, 253, 193, 106, 164, 253, 193, 138, 164, 253, 193, 182, 164, 253, 193, 198, 164, 253, 193, 226, 164, 253, 193, 232, 164, 253, 193, 246, 164, 253, 193, 40, 165, 253, 193, 42, 165, 253, 193, 80, 165, 253, 193, 106, 165, 253, 193, 170, 163, 253, 193, 172, 163, 253, 193, 174, 163, 253, 193, 196, 163, 253, 193, 198, 163, 253, 193, 200, 163, 253, 193, 202, 163, 253, 193, 236, 163, 253, 193, 238, 163, 253, 193, 26, 164, 253, 193, 240, 163, 253, 193, 242, 163, 253, 193, 28, 164, 253, 193, 30, 164, 253, 193, 32, 164, 253, 193, 34, 164, 253, 193, 36, 164, 253, 193, 58, 164, 253, 193, 60, 164, 253, 193, 72, 164, 253, 193, 74, 164, 253, 193, 108, 164, 253, 193, 110, 164, 253, 193, 114, 164, 253, 193, 192, 164, 253, 193, 252, 164, 253, 193, 2, 165, 253, 193, 14, 165, 253, 193, 16, 165, 253, 193, 18, 165, 253, 193, 48, 165, 253, 193, 50, 165, 253, 193, 222, 163, 253, 193, 224, 163, 253, 193, 226, 163, 253, 193, 4, 164, 253, 193, 6, 164, 253, 193, 8, 164, 253, 193, 10, 164, 253, 193, 12, 164, 253, 193, 44, 164, 253, 193, 46, 164, 253, 193, 48, 164, 253, 193, 50, 164, 253, 193, 52, 164, 253, 193, 54, 164, 253, 193, 66, 164, 253, 193, 68, 164, 253, 193, 70, 164, 253, 193, 84, 164, 253, 193, 86, 164, 253, 193, 96, 164, 253, 193, 98, 164, 253, 193, 162, 164, 253, 193, 164, 164, 253, 193, 184, 164, 253, 193, 186, 164, 253, 193, 200, 164, 253, 193, 228, 164, 253, 193, 238, 164, 253, 193, 10, 165, 253, 193, 82, 165, 253, 193, 84, 165, 253, 193, 108, 165, 253, 193, 16, 164, 253, 193, 18, 164, 253, 193, 20, 164, 253, 193, 22, 164, 253, 193, 24, 164, 253, 193, 56, 164, 253, 193, 90, 164, 253, 193, 92, 164, 253, 193, 100, 164, 253, 193, 102, 164, 253, 193, 104, 164, 253, 193, 126, 164, 253, 193, 142, 164, 253, 193, 204, 164, 253, 193, 12, 165, 253, 193, 58, 165, 253, 193, 160, 165, 253, 193, 238, 165, 253, 193, 140, 166, 253, 193, 30, 167, 253, 193, 32, 167, 253, 193, 34, 167, 253, 193, 36, 167, 253, 193, 38, 167, 253, 193, 190, 167, 253, 193, 192, 167, 253, 193, 194, 167, 253, 193, 196, 167, 253, 193, 198, 167, 253, 193, 44, 168, 253, 193, 46, 168, 253, 193, 48, 168, 253, 193, 140, 164, 253, 193, 166, 164, 253, 193, 188, 164, 253, 193, 190, 164, 253, 193, 202, 164, 253, 193, 216, 164, 253, 193, 218, 164, 253, 193, 44, 165, 253, 193, 46, 165, 253, 193, 86, 165, 253, 193, 88, 165, 253, 193, 134, 165, 253, 193, 136, 165, 253, 193, 158, 165, 253, 193, 178, 165, 253, 193, 186, 165, 253, 193, 188, 165, 253, 193, 194, 165, 253, 193, 236, 165, 253, 193, 22, 166, 253, 193, 24, 166, 253, 193, 26, 166, 253, 193, 28, 166, 253, 193, 134, 166, 253, 193, 136, 166, 253, 193, 138, 166, 253, 193, 22, 167, 253, 193, 24, 167, 253, 193, 26, 167, 253, 193, 28, 167, 253, 193, 176, 167, 253, 193, 178, 167, 253, 193, 26, 165, 253, 193, 28, 165, 253, 193, 34, 165, 253, 193, 30, 165, 253, 193, 32, 165, 253, 193, 36, 165, 253, 193, 38, 165, 253, 193, 54, 165, 253, 193, 56, 165, 253, 193, 72, 165, 253, 193, 74, 165, 253, 193, 76, 165, 253, 193, 94, 165, 253, 193, 96, 165, 253, 193, 98, 165, 253, 193, 78, 165, 253, 193, 100, 165, 253, 193, 102, 165, 253, 193, 104, 165, 253, 193, 116, 165, 253, 193, 118, 165, 253, 193, 120, 165, 253, 193, 122, 165, 253, 193, 124, 165, 253, 193, 126, 165, 253, 193, 140, 165, 253, 193, 142, 165, 253, 193, 144, 165, 253, 193, 146, 165, 253, 193, 164, 165, 253, 193, 166, 165, 253, 193, 168, 165, 253, 193, 52, 165, 253, 193, 60, 165, 253, 193, 62, 165, 253, 193, 64, 165, 253, 193, 90, 165, 253, 193, 92, 165, 253, 193, 112, 165, 253, 193, 114, 165, 253, 193, 138, 165, 253, 193, 162, 165, 253, 193, 190, 165, 253, 193, 200, 165, 253, 193, 208, 165, 253, 193, 216, 165, 253, 193, 218, 165, 253, 193, 240, 165, 253, 193, 242, 165, 253, 193, 244, 165, 253, 193, 246, 165, 253, 193, 248, 165, 253, 193, 250, 165, 253, 193, 252, 165, 253, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 193, 110, 165, 253, 193, 128, 165, 253, 193, 130, 165, 253, 193, 132, 165, 253, 193, 152, 165, 253, 193, 154, 165, 253, 193, 156, 165, 253, 193, 176, 165, 253, 193, 20, 166, 253, 193, 130, 166, 253, 193, 132, 166, 253, 193, 174, 167, 253, 193, 36, 168, 253, 193, 202, 168, 253, 193, 204, 168, 253, 193, 206, 168, 253, 193, 126, 170, 253, 193, 240, 170, 253, 193, 238, 171, 253, 193, 24, 172, 253, 193, 78, 172, 253, 193, 80, 172, 253, 193, 82, 172, 253, 193, 84, 172, 253, 193, 86, 172, 253, 193, 140, 172, 253, 193, 142, 172, 253, 193, 144, 172, 253, 193, 146, 172, 253, 193, 148, 172, 253, 193, 150, 172, 253, 193, 200, 172, 253, 193, 148, 165, 253, 193, 150, 165, 253, 193, 174, 165, 253, 193, 184, 165, 253, 193, 206, 165, 253, 193, 18, 166, 253, 193, 122, 166, 253, 193, 124, 166, 253, 193, 126, 166, 253, 193, 128, 166, 253, 193, 172, 167, 253, 193, 194, 168, 253, 193, 196, 168, 253, 193, 198, 168, 253, 193, 200, 168, 253, 193, 94, 169, 253, 193, 96, 169, 253, 193, 98, 169, 253, 193, 10, 170, 253, 193, 12, 170, 253, 193, 14, 170, 253, 193, 16, 170, 253, 193, 18, 170, 253, 193, 238, 170, 253, 193, 62, 171, 253, 193, 106, 171, 253, 193, 194, 171, 253, 193, 254, 171, 253, 193, 20, 172, 253, 193, 22, 172, 253, 193, 72, 172, 253, 193, 74, 172, 253, 193, 170, 165, 253, 193, 172, 165, 253, 193, 180, 165, 253, 193, 182, 165, 253, 193, 192, 165, 253, 193, 198, 165, 253, 193, 204, 165, 253, 193, 210, 165, 253, 193, 214, 165, 253, 193, 222, 165, 253, 193, 224, 165, 253, 193, 226, 165, 253, 193, 228, 165, 253, 193, 230, 165, 253, 193, 232, 165, 253, 193, 234, 165, 253, 193, 4, 166, 253, 193, 6, 166, 253, 193, 8, 166, 253, 193, 10, 166, 253, 193, 12, 166, 253, 193, 14, 166, 253, 193, 16, 166, 253, 193, 82, 166, 253, 193, 84, 166, 253, 193, 86, 166, 253, 193, 88, 166, 253, 193, 90, 166, 253, 193, 92, 166, 253, 193, 94, 166, 253, 193, 96, 166, 253, 193, 206, 166, 253, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 206, 160, 111, 0, 193, 142, 166, 253, 193, 144, 166, 253, 193, 146, 166, 253, 193, 148, 166, 253, 193, 150, 166, 253, 193, 152, 166, 253, 193, 154, 166, 253, 193, 156, 166, 253, 193, 158, 166, 253, 193, 160, 166, 253, 193, 162, 166, 253, 193, 164, 166, 253, 193, 166, 166, 253, 193, 168, 166, 253, 193, 170, 166, 253, 193, 172, 166, 253, 193, 174, 166, 253, 193, 176, 166, 253, 193, 178, 166, 253, 193, 180, 166, 253, 193, 182, 166, 253, 193, 184, 166, 253, 193, 186, 166, 253, 193, 188, 166, 253, 193, 98, 166, 253, 193, 100, 166, 253, 193, 102, 166, 253, 193, 104, 166, 253, 193, 106, 166, 253, 193, 108, 166, 253, 193, 208, 166, 253, 193, 110, 166, 253, 193, 112, 166, 253, 193, 114, 166, 253, 193, 116, 166, 253, 193, 118, 166, 253, 193, 120, 166, 253, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 193, 190, 166, 253, 193, 40, 167, 253, 193, 42, 167, 253, 193, 44, 167, 253, 193, 46, 167, 253, 193, 48, 167, 253, 193, 50, 167, 253, 193, 52, 167, 253, 193, 54, 167, 253, 193, 56, 167, 253, 193, 58, 167, 253, 193, 60, 167, 253, 193, 62, 167, 253, 193, 64, 167, 253, 193, 66, 167, 253, 193, 68, 167, 253, 193, 70, 167, 253, 193, 72, 167, 253, 193, 74, 167, 253, 193, 56, 168, 253, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 193, 194, 166, 253, 193, 196, 166, 253, 193, 198, 166, 253, 193, 200, 166, 253, 193, 202, 166, 253, 193, 204, 166, 253, 193, 76, 167, 253, 193, 78, 167, 253, 193, 80, 167, 253, 193, 82, 167, 253, 193, 84, 167, 253, 193, 86, 167, 253, 193, 88, 167, 253, 193, 90, 167, 253, 193, 232, 167, 253, 193, 234, 167, 253, 193, 236, 167, 253, 193, 238, 167, 253, 193, 240, 167, 253, 193, 90, 168, 253, 193, 92, 168, 253, 193, 94, 168, 253, 193, 96, 168, 253, 193, 98, 168, 253, 193, 100, 168, 253, 193, 102, 168, 253, 193, 104, 168, 253, 193, 106, 168, 253, 193, 8, 169, 253, 193, 10, 169, 253, 193, 12, 169, 253, 193, 14, 169, 253, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 192, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 224, 111, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 193, 180, 167, 253, 193, 182, 167, 253, 193, 184, 167, 253, 193, 186, 167, 253, 193, 188, 167, 253, 193, 38, 168, 253, 193, 40, 168, 253, 193, 42, 168, 253, 193, 208, 168, 253, 193, 210, 168, 253, 193, 212, 168, 253, 193, 214, 168, 253, 193, 216, 168, 253, 193, 100, 169, 253, 193, 102, 169, 253, 193, 104, 169, 253, 193, 106, 169, 253, 193, 108, 169, 253, 193, 110, 169, 253, 193, 112, 169, 253, 193, 114, 169, 253, 193, 20, 170, 253, 193, 128, 170, 253, 193, 130, 170, 253, 193, 132, 170, 253, 193, 134, 170, 253, 193, 64, 171, 253, 193, 108, 171, 253, 193, 142, 171, 253, 193, 170, 171, 253, 193, 172, 171, 253, 193, 196, 171, 253, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 206, 0, 112, 0, 193, 58, 168, 253, 193, 60, 168, 253, 193, 62, 168, 253, 193, 64, 168, 253, 193, 66, 168, 253, 193, 68, 168, 253, 193, 70, 168, 253, 193, 72, 168, 253, 193, 74, 168, 253, 193, 76, 168, 253, 193, 78, 168, 253, 193, 80, 168, 253, 193, 82, 168, 253, 193, 84, 168, 253, 193, 86, 168, 253, 193, 88, 168, 253, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 32, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 193, 38, 169, 253, 193, 40, 169, 253, 193, 42, 169, 253, 193, 44, 169, 253, 193, 46, 169, 253, 193, 48, 169, 253, 193, 50, 169, 253, 193, 52, 169, 253, 193, 54, 169, 253, 193, 56, 169, 253, 193, 58, 169, 253, 193, 60, 169, 253, 193, 62, 169, 253, 193, 64, 169, 253, 193, 66, 169, 253, 193, 68, 169, 253, 193, 70, 169, 253, 193, 72, 169, 253, 193, 74, 169, 253, 193, 76, 169, 253, 193, 78, 169, 253, 193, 80, 169, 253, 193, 82, 169, 253, 193, 34, 168, 253, 193, 84, 169, 253, 193, 86, 169, 253, 193, 88, 169, 253, 193, 90, 169, 253, 193, 92, 169, 253, 193, 172, 169, 253, 193, 174, 169, 253, 193, 176, 169, 253, 193, 50, 168, 253, 193, 52, 168, 253, 193, 54, 168, 253, 193, 218, 168, 253, 193, 116, 169, 253, 193, 118, 169, 253, 193, 120, 169, 253, 193, 122, 169, 253, 193, 124, 169, 253, 193, 22, 170, 253, 193, 24, 170, 253, 193, 26, 170, 253, 193, 28, 170, 253, 193, 30, 170, 253, 193, 136, 170, 253, 193, 138, 170, 253, 193, 140, 170, 253, 193, 142, 170, 253, 193, 242, 170, 253, 193, 244, 170, 253, 193, 246, 170, 253, 193, 66, 171, 253, 193, 68, 171, 253, 193, 110, 171, 253, 193, 112, 171, 253, 193, 144, 171, 253, 193, 146, 171, 253, 193, 148, 171, 253, 193, 2, 172, 253, 193, 4, 172, 253, 193, 6, 172, 253, 193, 8, 172, 253, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 206, 64, 112, 0, 193, 22, 169, 253, 193, 24, 169, 253, 193, 26, 169, 253, 193, 28, 169, 253, 193, 30, 169, 253, 193, 32, 169, 253, 193, 34, 169, 253, 193, 192, 168, 253, 193, 36, 169, 253, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 206, 96, 112, 0, 193, 126, 169, 253, 193, 128, 169, 253, 193, 130, 169, 253, 193, 132, 169, 253, 193, 134, 169, 253, 193, 136, 169, 253, 193, 138, 169, 253, 193, 140, 169, 253, 193, 142, 169, 253, 193, 144, 169, 253, 193, 146, 169, 253, 193, 148, 169, 253, 193, 150, 169, 253, 193, 152, 169, 253, 193, 154, 169, 253, 193, 156, 169, 253, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 193, 16, 169, 253, 193, 18, 169, 253, 193, 20, 169, 253, 193, 158, 169, 253, 193, 160, 169, 253, 193, 162, 169, 253, 193, 164, 169, 253, 193, 166, 169, 253, 193, 168, 169, 253, 193, 170, 169, 253, 193, 64, 170, 253, 193, 66, 170, 253, 193, 68, 170, 253, 193, 70, 170, 253, 193, 72, 170, 253, 193, 74, 170, 253, 193, 76, 170, 253, 193, 78, 170, 253, 193, 172, 170, 253, 193, 174, 170, 253, 193, 176, 170, 253, 193, 178, 170, 253, 193, 180, 170, 253, 193, 8, 171, 253, 193, 10, 171, 253, 193, 12, 171, 253, 193, 76, 171, 253, 193, 78, 171, 253, 193, 80, 171, 253, 193, 114, 171, 253, 193, 158, 171, 253, 193, 184, 171, 253, 193, 178, 169, 253, 193, 180, 169, 253, 193, 182, 169, 253, 193, 184, 169, 253, 193, 186, 169, 253, 193, 188, 169, 253, 193, 190, 169, 253, 193, 192, 169, 253, 193, 194, 169, 253, 193, 196, 169, 253, 193, 198, 169, 253, 193, 200, 169, 253, 193, 202, 169, 253, 193, 204, 169, 253, 193, 206, 169, 253, 193, 208, 169, 253, 193, 210, 169, 253, 193, 212, 169, 253, 193, 214, 169, 253, 193, 216, 169, 253, 193, 218, 169, 253, 193, 220, 169, 253, 193, 222, 169, 253, 193, 224, 169, 253, 193, 226, 169, 253, 193, 228, 169, 253, 193, 230, 169, 253, 193, 232, 169, 253, 193, 234, 169, 253, 193, 236, 169, 253, 193, 238, 169, 253, 193, 240, 169, 253, 193, 242, 169, 253, 193, 244, 169, 253, 193, 246, 169, 253, 193, 248, 169, 253, 193, 250, 169, 253, 193, 252, 169, 253, 193, 254, 169, 253, 193, 2, 170, 253, 193, 4, 170, 253, 193, 6, 170, 253, 193, 8, 170, 253, 193, 80, 170, 253, 193, 82, 170, 253, 193, 84, 170, 253, 193, 86, 170, 253, 193, 88, 170, 253, 193, 90, 170, 253, 193, 92, 170, 253, 193, 94, 170, 253, 193, 96, 170, 253, 193, 98, 170, 253, 193, 100, 170, 253, 193, 102, 170, 253, 193, 104, 170, 253, 193, 106, 170, 253, 193, 108, 170, 253, 193, 110, 170, 253, 193, 112, 170, 253, 193, 114, 170, 253, 193, 116, 170, 253, 193, 118, 170, 253, 193, 120, 170, 253, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 206, 128, 112, 0, 193, 144, 170, 253, 193, 146, 170, 253, 193, 148, 170, 253, 193, 150, 170, 253, 193, 152, 170, 253, 193, 154, 170, 253, 193, 156, 170, 253, 193, 158, 170, 253, 193, 160, 170, 253, 193, 162, 170, 253, 193, 164, 170, 253, 193, 166, 170, 253, 193, 168, 170, 253, 193, 170, 170, 253, 193, 248, 170, 253, 193, 250, 170, 253, 193, 252, 170, 253, 193, 254, 170, 253, 193, 2, 171, 253, 193, 4, 171, 253, 193, 6, 171, 253, 193, 70, 171, 253, 193, 72, 171, 253, 193, 74, 171, 253, 193, 122, 170, 253, 193, 124, 170, 253, 193, 182, 170, 253, 193, 184, 170, 253, 193, 186, 170, 253, 193, 188, 170, 253, 193, 190, 170, 253, 193, 192, 170, 253, 193, 194, 170, 253, 193, 196, 170, 253, 193, 198, 170, 253, 193, 200, 170, 253, 193, 202, 170, 253, 193, 204, 170, 253, 193, 206, 170, 253, 193, 208, 170, 253, 193, 210, 170, 253, 193, 212, 170, 253, 193, 214, 170, 253, 193, 216, 170, 253, 193, 218, 170, 253, 193, 220, 170, 253, 193, 222, 170, 253, 193, 224, 170, 253, 193, 226, 170, 253, 193, 228, 170, 253, 193, 230, 170, 253, 193, 232, 170, 253, 193, 234, 170, 253, 193, 236, 170, 253, 193, 14, 171, 253, 193, 16, 171, 253, 193, 18, 171, 253, 193, 20, 171, 253, 193, 22, 171, 253, 193, 24, 171, 253, 193, 26, 171, 253, 193, 28, 171, 253, 193, 30, 171, 253, 193, 32, 171, 253, 193, 34, 171, 253, 193, 36, 171, 253, 193, 38, 171, 253, 193, 40, 171, 253, 193, 42, 171, 253, 193, 44, 171, 253, 193, 46, 171, 253, 193, 48, 171, 253, 193, 50, 171, 253, 193, 52, 171, 253, 193, 54, 171, 253, 193, 56, 171, 253, 193, 58, 171, 253, 193, 60, 171, 253, 193, 82, 171, 253, 193, 84, 171, 253, 193, 86, 171, 253, 193, 88, 171, 253, 193, 90, 171, 253, 193, 92, 171, 253, 193, 94, 171, 253, 193, 96, 171, 253, 193, 98, 171, 253, 193, 100, 171, 253, 193, 102, 171, 253, 193, 104, 171, 253, 193, 116, 171, 253, 193, 118, 171, 253, 193, 120, 171, 253, 193, 122, 171, 253, 193, 124, 171, 253, 193, 126, 171, 253, 193, 128, 171, 253, 193, 130, 171, 253, 193, 132, 171, 253, 193, 134, 171, 253, 193, 136, 171, 253, 193, 138, 171, 253, 193, 140, 171, 253, 193, 160, 171, 253, 193, 162, 171, 253, 193, 164, 171, 253, 193, 166, 171, 253, 193, 168, 171, 253, 193, 186, 171, 253, 193, 188, 171, 253, 193, 190, 171, 253, 193, 192, 171, 253, 193, 204, 171, 253, 193, 206, 171, 253, 193, 208, 171, 253, 193, 210, 171, 253, 193, 212, 171, 253, 193, 214, 171, 253, 193, 216, 171, 253, 193, 224, 171, 253, 193, 150, 171, 253, 193, 152, 171, 253, 193, 154, 171, 253, 193, 156, 171, 253, 193, 174, 171, 253, 193, 176, 171, 253, 193, 178, 171, 253, 193, 180, 171, 253, 193, 198, 171, 253, 193, 200, 171, 253, 193, 220, 171, 253, 193, 228, 171, 253, 193, 244, 171, 253, 193, 246, 171, 253, 193, 248, 171, 253, 193, 250, 171, 253, 193, 252, 171, 253, 193, 10, 172, 253, 193, 12, 172, 253, 193, 14, 172, 253, 193, 16, 172, 253, 193, 32, 172, 253, 193, 34, 172, 253, 193, 36, 172, 253, 193, 38, 172, 253, 193, 40, 172, 253, 193, 42, 172, 253, 193, 44, 172, 253, 193, 46, 172, 253, 193, 48, 172, 253, 193, 50, 172, 253, 193, 52, 172, 253, 193, 202, 171, 253, 193, 222, 171, 253, 193, 240, 171, 253, 193, 118, 172, 253, 193, 184, 172, 253, 193, 62, 173, 253, 193, 170, 173, 253, 193, 174, 173, 253, 193, 216, 173, 253, 193, 218, 173, 253, 193, 10, 174, 253, 193, 12, 174, 253, 193, 14, 174, 253, 193, 16, 174, 253, 193, 18, 174, 253, 193, 68, 174, 253, 193, 70, 174, 253, 193, 72, 174, 253, 193, 74, 174, 253, 193, 76, 174, 253, 193, 78, 174, 253, 193, 118, 174, 253, 193, 120, 174, 253, 193, 122, 174, 253, 193, 124, 174, 253, 193, 126, 174, 253, 193, 128, 174, 253, 193, 174, 174, 253, 193, 176, 174, 253, 193, 242, 174, 253, 193, 244, 174, 253, 193, 246, 174, 253, 193, 218, 171, 253, 193, 204, 172, 253, 193, 206, 172, 253, 193, 44, 173, 253, 193, 80, 173, 253, 193, 132, 173, 253, 193, 182, 173, 253, 193, 54, 174, 253, 193, 56, 174, 253, 193, 106, 174, 253, 193, 162, 174, 253, 193, 164, 174, 253, 193, 166, 174, 253, 193, 222, 174, 253, 193, 224, 174, 253, 193, 226, 174, 253, 193, 38, 175, 253, 193, 40, 175, 253, 193, 76, 175, 253, 193, 114, 175, 253, 193, 116, 175, 253, 193, 176, 175, 253, 193, 208, 175, 253, 193, 222, 175, 253, 193, 26, 176, 253, 193, 40, 176, 253, 193, 98, 176, 253, 193, 118, 176, 253, 193, 120, 176, 253, 193, 40, 177, 253, 193, 42, 177, 253, 193, 44, 177, 253, 193, 226, 171, 253, 193, 230, 171, 253, 193, 232, 171, 253, 193, 234, 171, 253, 193, 236, 171, 253, 193, 242, 171, 253, 193, 18, 172, 253, 193, 56, 172, 253, 193, 58, 172, 253, 193, 60, 172, 253, 193, 62, 172, 253, 193, 64, 172, 253, 193, 66, 172, 253, 193, 68, 172, 253, 193, 120, 172, 253, 193, 70, 172, 253, 193, 122, 172, 253, 193, 124, 172, 253, 193, 126, 172, 253, 193, 128, 172, 253, 193, 130, 172, 253, 193, 132, 172, 253, 193, 134, 172, 253, 193, 186, 172, 253, 193, 188, 172, 253, 193, 190, 172, 253, 193, 192, 172, 253, 193, 236, 172, 253, 193, 238, 172, 253, 193, 240, 172, 253, 193, 12, 173, 253, 193, 14, 173, 253, 193, 26, 172, 253, 193, 28, 172, 253, 193, 30, 172, 253, 193, 88, 172, 253, 193, 90, 172, 253, 193, 92, 172, 253, 193, 94, 172, 253, 193, 96, 172, 253, 193, 98, 172, 253, 193, 152, 172, 253, 193, 154, 172, 253, 193, 156, 172, 253, 193, 158, 172, 253, 193, 160, 172, 253, 193, 162, 172, 253, 193, 164, 172, 253, 193, 166, 172, 253, 193, 168, 172, 253, 193, 170, 172, 253, 193, 208, 172, 253, 193, 210, 172, 253, 193, 212, 172, 253, 193, 214, 172, 253, 193, 216, 172, 253, 193, 218, 172, 253, 193, 220, 172, 253, 193, 252, 172, 253, 193, 254, 172, 253, 193, 46, 173, 253, 193, 48, 173, 253, 193, 50, 173, 253, 193, 52, 173, 253, 193, 54, 172, 253, 193, 100, 172, 253, 193, 102, 172, 253, 193, 104, 172, 253, 193, 106, 172, 253, 193, 108, 172, 253, 193, 110, 172, 253, 193, 112, 172, 253, 193, 114, 172, 253, 193, 116, 172, 253, 193, 172, 172, 253, 193, 174, 172, 253, 193, 176, 172, 253, 193, 178, 172, 253, 193, 180, 172, 253, 193, 182, 172, 253, 193, 222, 172, 253, 193, 224, 172, 253, 193, 226, 172, 253, 193, 228, 172, 253, 193, 230, 172, 253, 193, 232, 172, 253, 193, 234, 172, 253, 193, 2, 173, 253, 193, 4, 173, 253, 193, 6, 173, 253, 193, 8, 173, 253, 193, 10, 173, 253, 193, 54, 173, 253, 193, 56, 173, 253, 193, 58, 173, 253, 193, 60, 173, 253, 193, 136, 172, 253, 193, 138, 172, 253, 193, 194, 172, 253, 193, 196, 172, 253, 193, 198, 172, 253, 193, 242, 172, 253, 193, 244, 172, 253, 193, 246, 172, 253, 193, 36, 173, 253, 193, 38, 173, 253, 193, 72, 173, 253, 193, 74, 173, 253, 193, 76, 173, 253, 193, 100, 173, 253, 193, 102, 173, 253, 193, 126, 173, 253, 193, 138, 173, 253, 193, 164, 173, 253, 193, 104, 174, 253, 193, 34, 175, 253, 193, 36, 175, 253, 193, 238, 175, 253, 193, 38, 176, 253, 193, 62, 176, 253, 193, 188, 176, 253, 193, 222, 176, 253, 193, 224, 176, 253, 193, 36, 177, 253, 193, 132, 177, 253, 193, 208, 177, 253, 193, 34, 178, 253, 193, 36, 178, 253, 193, 202, 172, 253, 193, 248, 172, 253, 193, 250, 172, 253, 193, 40, 173, 253, 193, 42, 173, 253, 193, 78, 173, 253, 193, 106, 173, 253, 193, 108, 173, 253, 193, 110, 173, 253, 193, 128, 173, 253, 193, 130, 173, 253, 193, 142, 173, 253, 193, 146, 173, 253, 193, 150, 173, 253, 193, 154, 173, 253, 193, 162, 173, 253, 193, 166, 173, 253, 193, 158, 174, 253, 193, 160, 174, 253, 193, 220, 175, 253, 193, 240, 175, 253, 193, 24, 176, 253, 193, 64, 176, 253, 193, 66, 176, 253, 193, 82, 176, 253, 193, 94, 176, 253, 193, 96, 176, 253, 193, 116, 176, 253, 193, 126, 176, 253, 193, 134, 176, 253, 193, 136, 176, 253, 193, 138, 176, 253, 193, 16, 173, 253, 193, 18, 173, 253, 193, 20, 173, 253, 193, 22, 173, 253, 193, 64, 173, 253, 193, 24, 173, 253, 193, 26, 173, 253, 193, 28, 173, 253, 193, 30, 173, 253, 193, 32, 173, 253, 193, 34, 173, 253, 193, 66, 173, 253, 193, 68, 173, 253, 193, 70, 173, 253, 193, 122, 173, 253, 193, 124, 173, 253, 193, 148, 173, 253, 193, 176, 173, 253, 193, 178, 173, 253, 193, 180, 173, 253, 193, 194, 173, 253, 193, 196, 173, 253, 193, 198, 173, 253, 193, 200, 173, 253, 193, 202, 173, 253, 193, 204, 173, 253, 193, 206, 173, 253, 193, 208, 173, 253, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 193, 82, 173, 253, 193, 84, 173, 253, 193, 86, 173, 253, 193, 88, 173, 253, 193, 90, 173, 253, 193, 92, 173, 253, 193, 112, 173, 253, 193, 114, 173, 253, 193, 116, 173, 253, 193, 118, 173, 253, 193, 134, 173, 253, 193, 136, 173, 253, 193, 156, 173, 253, 193, 108, 174, 253, 193, 110, 174, 253, 193, 112, 174, 253, 193, 228, 174, 253, 193, 230, 174, 253, 193, 232, 174, 253, 193, 140, 175, 253, 193, 242, 175, 253, 193, 244, 175, 253, 193, 246, 175, 253, 193, 254, 175, 253, 193, 42, 176, 253, 193, 44, 176, 253, 193, 46, 176, 253, 193, 68, 176, 253, 193, 70, 176, 253, 193, 150, 176, 253, 193, 152, 176, 253, 193, 140, 177, 253, 193, 94, 173, 253, 193, 96, 173, 253, 193, 98, 173, 253, 193, 120, 173, 253, 193, 144, 173, 253, 193, 152, 173, 253, 193, 158, 173, 253, 193, 160, 173, 253, 193, 168, 173, 253, 193, 172, 173, 253, 193, 184, 173, 253, 193, 186, 173, 253, 193, 188, 173, 253, 193, 190, 173, 253, 193, 192, 173, 253, 193, 210, 173, 253, 193, 212, 173, 253, 193, 214, 173, 253, 193, 252, 173, 253, 193, 254, 173, 253, 193, 2, 174, 253, 193, 4, 174, 253, 193, 6, 174, 253, 193, 8, 174, 253, 193, 58, 174, 253, 193, 60, 174, 253, 193, 62, 174, 253, 193, 64, 174, 253, 193, 66, 174, 253, 193, 114, 174, 253, 193, 116, 174, 253, 193, 168, 174, 253, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 206, 160, 112, 0, 193, 20, 174, 253, 193, 22, 174, 253, 193, 24, 174, 253, 193, 26, 174, 253, 193, 28, 174, 253, 193, 30, 174, 253, 193, 32, 174, 253, 193, 34, 174, 253, 193, 36, 174, 253, 193, 38, 174, 253, 193, 40, 174, 253, 193, 42, 174, 253, 193, 44, 174, 253, 193, 46, 174, 253, 193, 48, 174, 253, 193, 50, 174, 253, 193, 52, 174, 253, 193, 80, 174, 253, 193, 82, 174, 253, 193, 84, 174, 253, 193, 86, 174, 253, 193, 88, 174, 253, 193, 90, 174, 253, 193, 92, 174, 253, 193, 94, 174, 253, 193, 96, 174, 253, 193, 98, 174, 253, 193, 100, 174, 253, 193, 102, 174, 253, 193, 130, 174, 253, 193, 132, 174, 253, 193, 134, 174, 253, 193, 136, 174, 253, 193, 138, 174, 253, 193, 140, 174, 253, 193, 142, 174, 253, 193, 144, 174, 253, 193, 146, 174, 253, 193, 148, 174, 253, 193, 150, 174, 253, 193, 152, 174, 253, 193, 154, 174, 253, 193, 156, 174, 253, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 193, 170, 174, 253, 193, 172, 174, 253, 193, 234, 174, 253, 193, 236, 174, 253, 193, 238, 174, 253, 193, 240, 174, 253, 193, 42, 175, 253, 193, 44, 175, 253, 193, 46, 175, 253, 193, 48, 175, 253, 193, 50, 175, 253, 193, 52, 175, 253, 193, 54, 175, 253, 193, 78, 175, 253, 193, 80, 175, 253, 193, 82, 175, 253, 193, 118, 175, 253, 193, 142, 175, 253, 193, 144, 175, 253, 193, 146, 175, 253, 193, 148, 175, 253, 193, 178, 175, 253, 193, 194, 175, 253, 193, 210, 175, 253, 193, 232, 175, 253, 193, 234, 175, 253, 193, 2, 176, 253, 193, 28, 176, 253, 193, 84, 176, 253, 193, 128, 176, 253, 193, 140, 176, 253, 193, 186, 176, 253, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 206, 192, 112, 0, 193, 250, 174, 253, 193, 252, 174, 253, 193, 254, 174, 253, 193, 2, 175, 253, 193, 4, 175, 253, 193, 6, 175, 253, 193, 8, 175, 253, 193, 10, 175, 253, 193, 12, 175, 253, 193, 14, 175, 253, 193, 16, 175, 253, 193, 18, 175, 253, 193, 20, 175, 253, 193, 22, 175, 253, 193, 24, 175, 253, 193, 26, 175, 253, 193, 28, 175, 253, 193, 30, 175, 253, 193, 32, 175, 253, 193, 248, 174, 253, 193, 56, 175, 253, 193, 84, 175, 253, 193, 120, 175, 253, 193, 150, 175, 253, 193, 180, 175, 253, 193, 196, 175, 253, 193, 236, 175, 253, 193, 248, 175, 253, 193, 4, 176, 253, 193, 100, 176, 253, 193, 130, 176, 253, 193, 160, 176, 253, 193, 192, 176, 253, 193, 194, 176, 253, 193, 242, 176, 253, 193, 244, 176, 253, 193, 246, 176, 253, 193, 248, 176, 253, 193, 250, 176, 253, 193, 252, 176, 253, 193, 66, 177, 253, 193, 68, 177, 253, 193, 158, 177, 253, 193, 160, 177, 253, 193, 234, 177, 253, 193, 236, 177, 253, 193, 238, 177, 253, 193, 240, 177, 253, 193, 60, 178, 253, 193, 62, 178, 253, 193, 64, 178, 253, 193, 58, 175, 253, 193, 60, 175, 253, 193, 62, 175, 253, 193, 64, 175, 253, 193, 66, 175, 253, 193, 68, 175, 253, 193, 70, 175, 253, 193, 72, 175, 253, 193, 74, 175, 253, 193, 86, 175, 253, 193, 88, 175, 253, 193, 90, 175, 253, 193, 92, 175, 253, 193, 94, 175, 253, 193, 96, 175, 253, 193, 98, 175, 253, 193, 100, 175, 253, 193, 102, 175, 253, 193, 104, 175, 253, 193, 106, 175, 253, 193, 108, 175, 253, 193, 110, 175, 253, 193, 112, 175, 253, 193, 122, 175, 253, 193, 124, 175, 253, 193, 126, 175, 253, 193, 128, 175, 253, 193, 130, 175, 253, 193, 132, 175, 253, 193, 134, 175, 253, 193, 136, 175, 253, 193, 138, 175, 253, 193, 152, 175, 253, 193, 154, 175, 253, 193, 156, 175, 253, 193, 158, 175, 253, 193, 160, 175, 253, 193, 162, 175, 253, 193, 164, 175, 253, 193, 166, 175, 253, 193, 168, 175, 253, 193, 170, 175, 253, 193, 172, 175, 253, 193, 174, 175, 253, 193, 182, 175, 253, 193, 184, 175, 253, 193, 186, 175, 253, 193, 188, 175, 253, 193, 190, 175, 253, 193, 192, 175, 253, 193, 198, 175, 253, 193, 200, 175, 253, 193, 202, 175, 253, 193, 204, 175, 253, 193, 206, 175, 253, 193, 212, 175, 253, 193, 214, 175, 253, 193, 216, 175, 253, 193, 218, 175, 253, 193, 224, 175, 253, 193, 226, 175, 253, 193, 228, 175, 253, 193, 230, 175, 253, 193, 250, 175, 253, 193, 252, 175, 253, 193, 6, 176, 253, 193, 8, 176, 253, 193, 10, 176, 253, 193, 12, 176, 253, 193, 14, 176, 253, 193, 16, 176, 253, 193, 18, 176, 253, 193, 20, 176, 253, 193, 22, 176, 253, 193, 30, 176, 253, 193, 32, 176, 253, 193, 34, 176, 253, 193, 36, 176, 253, 193, 48, 176, 253, 193, 50, 176, 253, 193, 52, 176, 253, 193, 54, 176, 253, 193, 56, 176, 253, 193, 58, 176, 253, 193, 60, 176, 253, 193, 72, 176, 253, 193, 74, 176, 253, 193, 76, 176, 253, 193, 78, 176, 253, 193, 80, 176, 253, 193, 86, 176, 253, 193, 88, 176, 253, 193, 90, 176, 253, 193, 92, 176, 253, 193, 102, 176, 253, 193, 104, 176, 253, 193, 106, 176, 253, 193, 108, 176, 253, 193, 110, 176, 253, 193, 112, 176, 253, 193, 114, 176, 253, 193, 122, 176, 253, 193, 124, 176, 253, 193, 132, 176, 253, 193, 142, 176, 253, 193, 144, 176, 253, 193, 146, 176, 253, 193, 154, 176, 253, 193, 156, 176, 253, 193, 162, 176, 253, 193, 166, 176, 253, 193, 168, 176, 253, 193, 172, 176, 253, 193, 174, 176, 253, 193, 176, 176, 253, 193, 184, 176, 253, 193, 196, 176, 253, 193, 198, 176, 253, 193, 200, 176, 253, 193, 202, 176, 253, 193, 204, 176, 253, 193, 210, 176, 253, 193, 212, 176, 253, 193, 214, 176, 253, 193, 216, 176, 253, 193, 218, 176, 253, 193, 220, 176, 253, 193, 254, 176, 253, 193, 148, 176, 253, 193, 158, 176, 253, 193, 164, 176, 253, 193, 170, 176, 253, 193, 178, 176, 253, 193, 180, 176, 253, 193, 182, 176, 253, 193, 38, 177, 253, 193, 210, 177, 253, 193, 212, 177, 253, 193, 38, 178, 253, 193, 134, 178, 253, 193, 136, 178, 253, 193, 226, 178, 253, 193, 228, 178, 253, 193, 208, 179, 253, 193, 90, 180, 253, 193, 130, 180, 253, 193, 176, 180, 253, 193, 182, 180, 253, 193, 202, 180, 253, 193, 230, 180, 253, 193, 244, 180, 253, 193, 18, 181, 253, 193, 66, 181, 253, 193, 96, 181, 253, 193, 108, 181, 253, 193, 150, 181, 253, 193, 152, 181, 253, 193, 178, 181, 253, 193, 210, 181, 253, 193, 132, 182, 253, 193, 190, 176, 253, 193, 206, 176, 253, 193, 208, 176, 253, 193, 226, 176, 253, 193, 228, 176, 253, 193, 230, 176, 253, 193, 232, 176, 253, 193, 234, 176, 253, 193, 236, 176, 253, 193, 238, 176, 253, 193, 240, 176, 253, 193, 46, 177, 253, 193, 48, 177, 253, 193, 50, 177, 253, 193, 52, 177, 253, 193, 54, 177, 253, 193, 56, 177, 253, 193, 58, 177, 253, 193, 60, 177, 253, 193, 62, 177, 253, 193, 64, 177, 253, 193, 142, 177, 253, 193, 144, 177, 253, 193, 146, 177, 253, 193, 148, 177, 253, 193, 150, 177, 253, 193, 152, 177, 253, 193, 154, 177, 253, 193, 220, 177, 253, 193, 222, 177, 253, 193, 224, 177, 253, 193, 226, 177, 253, 193, 2, 177, 253, 193, 4, 177, 253, 193, 6, 177, 253, 193, 8, 177, 253, 193, 10, 177, 253, 193, 12, 177, 253, 193, 14, 177, 253, 193, 16, 177, 253, 193, 18, 177, 253, 193, 20, 177, 253, 193, 22, 177, 253, 193, 24, 177, 253, 193, 26, 177, 253, 193, 28, 177, 253, 193, 30, 177, 253, 193, 32, 177, 253, 193, 34, 177, 253, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 224, 112, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 193, 134, 177, 253, 193, 136, 177, 253, 193, 138, 177, 253, 193, 214, 177, 253, 193, 216, 177, 253, 193, 218, 177, 253, 193, 40, 178, 253, 193, 42, 178, 253, 193, 44, 178, 253, 193, 138, 178, 253, 193, 140, 178, 253, 193, 142, 178, 253, 193, 144, 178, 253, 193, 230, 178, 253, 193, 232, 178, 253, 193, 52, 179, 253, 193, 174, 179, 253, 193, 176, 179, 253, 193, 210, 179, 253, 193, 32, 180, 253, 193, 92, 180, 253, 193, 100, 180, 253, 193, 138, 180, 253, 193, 156, 180, 253, 193, 162, 180, 253, 193, 174, 180, 253, 193, 110, 181, 253, 193, 132, 181, 253, 193, 18, 182, 253, 193, 20, 182, 253, 193, 134, 182, 253, 193, 136, 182, 253, 193, 228, 177, 253, 193, 230, 177, 253, 193, 232, 177, 253, 193, 50, 178, 253, 193, 52, 178, 253, 193, 54, 178, 253, 193, 56, 178, 253, 193, 156, 177, 253, 193, 58, 178, 253, 193, 146, 178, 253, 193, 148, 178, 253, 193, 150, 178, 253, 193, 152, 178, 253, 193, 154, 178, 253, 193, 156, 178, 253, 193, 158, 178, 253, 193, 240, 178, 253, 193, 242, 178, 253, 193, 244, 178, 253, 193, 246, 178, 253, 193, 56, 179, 253, 193, 58, 179, 253, 193, 60, 179, 253, 193, 62, 179, 253, 193, 64, 179, 253, 193, 116, 179, 253, 193, 118, 179, 253, 193, 120, 179, 253, 193, 122, 179, 253, 193, 66, 179, 253, 193, 180, 179, 253, 193, 182, 179, 253, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 206, 0, 113, 0, 193, 242, 177, 253, 193, 244, 177, 253, 193, 246, 177, 253, 193, 248, 177, 253, 193, 250, 177, 253, 193, 252, 177, 253, 193, 254, 177, 253, 193, 2, 178, 253, 193, 4, 178, 253, 193, 6, 178, 253, 193, 8, 178, 253, 193, 10, 178, 253, 193, 12, 178, 253, 193, 14, 178, 253, 193, 16, 178, 253, 193, 18, 178, 253, 193, 20, 178, 253, 193, 22, 178, 253, 193, 24, 178, 253, 193, 26, 178, 253, 193, 28, 178, 253, 193, 30, 178, 253, 193, 32, 178, 253, 193, 74, 178, 253, 193, 76, 178, 253, 193, 46, 178, 253, 193, 48, 178, 253, 193, 234, 178, 253, 193, 236, 178, 253, 193, 238, 178, 253, 193, 54, 179, 253, 193, 114, 179, 253, 193, 178, 179, 253, 193, 34, 180, 253, 193, 128, 180, 253, 193, 140, 180, 253, 193, 204, 180, 253, 193, 206, 180, 253, 193, 208, 180, 253, 193, 246, 180, 253, 193, 50, 181, 253, 193, 52, 181, 253, 193, 98, 181, 253, 193, 112, 181, 253, 193, 124, 181, 253, 193, 126, 181, 253, 193, 160, 181, 253, 193, 180, 181, 253, 193, 22, 182, 253, 193, 24, 182, 253, 193, 138, 182, 253, 193, 140, 182, 253, 193, 196, 182, 253, 193, 12, 183, 253, 193, 14, 183, 253, 193, 56, 183, 253, 193, 58, 183, 253, 193, 66, 178, 253, 193, 68, 178, 253, 193, 70, 178, 253, 193, 72, 178, 253, 193, 160, 178, 253, 193, 162, 178, 253, 193, 164, 178, 253, 193, 166, 178, 253, 193, 248, 178, 253, 193, 250, 178, 253, 193, 252, 178, 253, 193, 254, 178, 253, 193, 68, 179, 253, 193, 70, 179, 253, 193, 124, 179, 253, 193, 216, 179, 253, 193, 218, 179, 253, 193, 220, 179, 253, 193, 222, 179, 253, 193, 246, 179, 253, 193, 248, 179, 253, 193, 16, 180, 253, 193, 52, 180, 253, 193, 96, 180, 253, 193, 102, 180, 253, 193, 184, 180, 253, 193, 186, 180, 253, 193, 194, 180, 253, 193, 212, 180, 253, 193, 56, 181, 253, 193, 100, 181, 253, 193, 222, 181, 253, 193, 78, 178, 253, 193, 80, 178, 253, 193, 82, 178, 253, 193, 84, 178, 253, 193, 86, 178, 253, 193, 88, 178, 253, 193, 90, 178, 253, 193, 92, 178, 253, 193, 94, 178, 253, 193, 96, 178, 253, 193, 98, 178, 253, 193, 100, 178, 253, 193, 102, 178, 253, 193, 104, 178, 253, 193, 106, 178, 253, 193, 108, 178, 253, 193, 110, 178, 253, 193, 112, 178, 253, 193, 114, 178, 253, 193, 116, 178, 253, 193, 118, 178, 253, 193, 120, 178, 253, 193, 122, 178, 253, 193, 124, 178, 253, 193, 126, 178, 253, 193, 128, 178, 253, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 193, 130, 178, 253, 193, 132, 178, 253, 193, 224, 178, 253, 193, 112, 179, 253, 193, 192, 180, 253, 193, 200, 180, 253, 193, 48, 181, 253, 193, 138, 181, 253, 193, 176, 181, 253, 193, 16, 182, 253, 193, 130, 182, 253, 193, 190, 182, 253, 193, 2, 183, 253, 193, 4, 183, 253, 193, 168, 183, 253, 193, 44, 184, 253, 193, 92, 184, 253, 193, 230, 184, 253, 193, 232, 184, 253, 193, 148, 185, 253, 193, 150, 185, 253, 193, 152, 185, 253, 193, 154, 185, 253, 193, 156, 185, 253, 193, 78, 186, 253, 193, 80, 186, 253, 193, 82, 186, 253, 193, 84, 186, 253, 193, 86, 186, 253, 193, 88, 186, 253, 193, 90, 186, 253, 193, 92, 186, 253, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 32, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 206, 64, 113, 0, 193, 72, 179, 253, 193, 74, 179, 253, 193, 76, 179, 253, 193, 78, 179, 253, 193, 80, 179, 253, 193, 82, 179, 253, 193, 84, 179, 253, 193, 86, 179, 253, 193, 88, 179, 253, 193, 90, 179, 253, 193, 92, 179, 253, 193, 94, 179, 253, 193, 96, 179, 253, 193, 98, 179, 253, 193, 100, 179, 253, 193, 102, 179, 253, 193, 104, 179, 253, 193, 106, 179, 253, 193, 108, 179, 253, 193, 110, 179, 253, 193, 126, 179, 253, 193, 128, 179, 253, 193, 130, 179, 253, 193, 132, 179, 253, 193, 134, 179, 253, 193, 136, 179, 253, 193, 138, 179, 253, 193, 140, 179, 253, 193, 142, 179, 253, 193, 144, 179, 253, 193, 146, 179, 253, 193, 148, 179, 253, 193, 150, 179, 253, 193, 152, 179, 253, 193, 154, 179, 253, 193, 156, 179, 253, 193, 158, 179, 253, 193, 160, 179, 253, 193, 162, 179, 253, 193, 164, 179, 253, 193, 166, 179, 253, 193, 168, 179, 253, 193, 170, 179, 253, 193, 172, 179, 253, 206, 96, 113, 0, 206, 96, 113, 0, 206, 96, 113, 0, 206, 96, 113, 0, 206, 96, 113, 0, 193, 184, 179, 253, 193, 212, 179, 253, 193, 214, 179, 253, 193, 244, 179, 253, 193, 36, 180, 253, 193, 58, 180, 253, 193, 66, 180, 253, 193, 74, 180, 253, 193, 78, 180, 253, 193, 84, 180, 253, 193, 94, 180, 253, 193, 110, 180, 253, 193, 118, 180, 253, 193, 120, 180, 253, 193, 132, 180, 253, 193, 136, 180, 253, 193, 164, 180, 253, 193, 170, 180, 253, 193, 172, 180, 253, 193, 210, 180, 253, 193, 232, 180, 253, 193, 248, 180, 253, 193, 250, 180, 253, 193, 8, 181, 253, 193, 20, 181, 253, 193, 32, 181, 253, 193, 34, 181, 253, 193, 54, 181, 253, 193, 166, 181, 253, 193, 182, 181, 253, 193, 212, 181, 253, 193, 214, 181, 253, 206, 96, 113, 0, 206, 96, 113, 0, 206, 96, 113, 0, 206, 96, 113, 0, 206, 96, 113, 0, 206, 96, 113, 0, 193, 224, 179, 253, 193, 226, 179, 253, 193, 228, 179, 253, 193, 230, 179, 253, 193, 232, 179, 253, 193, 234, 179, 253, 193, 236, 179, 253, 193, 238, 179, 253, 193, 240, 179, 253, 193, 242, 179, 253, 193, 250, 179, 253, 193, 252, 179, 253, 193, 254, 179, 253, 193, 2, 180, 253, 193, 4, 180, 253, 193, 6, 180, 253, 193, 8, 180, 253, 193, 10, 180, 253, 193, 12, 180, 253, 193, 14, 180, 253, 193, 18, 180, 253, 193, 20, 180, 253, 193, 22, 180, 253, 193, 24, 180, 253, 193, 26, 180, 253, 193, 28, 180, 253, 193, 30, 180, 253, 193, 38, 180, 253, 193, 40, 180, 253, 193, 42, 180, 253, 193, 44, 180, 253, 193, 46, 180, 253, 193, 48, 180, 253, 193, 50, 180, 253, 193, 54, 180, 253, 193, 56, 180, 253, 193, 60, 180, 253, 193, 62, 180, 253, 193, 64, 180, 253, 193, 68, 180, 253, 193, 70, 180, 253, 193, 72, 180, 253, 193, 76, 180, 253, 193, 80, 180, 253, 193, 82, 180, 253, 193, 86, 180, 253, 193, 88, 180, 253, 193, 98, 180, 253, 193, 104, 180, 253, 193, 106, 180, 253, 193, 108, 180, 253, 193, 112, 180, 253, 193, 114, 180, 253, 193, 116, 180, 253, 193, 122, 180, 253, 193, 124, 180, 253, 193, 126, 180, 253, 193, 134, 180, 253, 193, 142, 180, 253, 193, 144, 180, 253, 193, 146, 180, 253, 193, 148, 180, 253, 193, 150, 180, 253, 193, 152, 180, 253, 193, 154, 180, 253, 193, 158, 180, 253, 193, 160, 180, 253, 193, 166, 180, 253, 193, 168, 180, 253, 193, 178, 180, 253, 193, 180, 180, 253, 193, 188, 180, 253, 193, 190, 180, 253, 193, 196, 180, 253, 193, 198, 180, 253, 193, 214, 180, 253, 193, 216, 180, 253, 193, 218, 180, 253, 193, 220, 180, 253, 193, 222, 180, 253, 193, 224, 180, 253, 193, 226, 180, 253, 193, 228, 180, 253, 193, 234, 180, 253, 193, 236, 180, 253, 193, 238, 180, 253, 193, 240, 180, 253, 193, 242, 180, 253, 193, 252, 180, 253, 193, 254, 180, 253, 193, 2, 181, 253, 193, 4, 181, 253, 193, 6, 181, 253, 193, 10, 181, 253, 193, 12, 181, 253, 193, 14, 181, 253, 193, 16, 181, 253, 193, 22, 181, 253, 193, 24, 181, 253, 193, 26, 181, 253, 193, 28, 181, 253, 193, 30, 181, 253, 193, 36, 181, 253, 193, 38, 181, 253, 193, 40, 181, 253, 193, 42, 181, 253, 193, 44, 181, 253, 193, 46, 181, 253, 193, 58, 181, 253, 193, 60, 181, 253, 193, 62, 181, 253, 193, 64, 181, 253, 206, 128, 113, 0, 206, 128, 113, 0, 206, 128, 113, 0, 206, 128, 113, 0, 206, 128, 113, 0, 206, 128, 113, 0, 206, 128, 113, 0, 206, 128, 113, 0, 206, 128, 113, 0, 206, 128, 113, 0, 193, 104, 181, 253, 193, 106, 181, 253, 193, 114, 181, 253, 193, 116, 181, 253, 193, 118, 181, 253, 193, 120, 181, 253, 193, 122, 181, 253, 193, 128, 181, 253, 193, 130, 181, 253, 193, 134, 181, 253, 193, 136, 181, 253, 193, 140, 181, 253, 193, 142, 181, 253, 193, 144, 181, 253, 193, 146, 181, 253, 193, 148, 181, 253, 193, 154, 181, 253, 193, 156, 181, 253, 193, 158, 181, 253, 193, 162, 181, 253, 193, 164, 181, 253, 193, 168, 181, 253, 193, 170, 181, 253, 193, 172, 181, 253, 193, 174, 181, 253, 193, 184, 181, 253, 193, 186, 181, 253, 193, 188, 181, 253, 193, 190, 181, 253, 193, 192, 181, 253, 193, 194, 181, 253, 193, 196, 181, 253, 193, 198, 181, 253, 193, 200, 181, 253, 193, 202, 181, 253, 193, 204, 181, 253, 193, 206, 181, 253, 193, 208, 181, 253, 193, 224, 181, 253, 193, 226, 181, 253, 193, 228, 181, 253, 193, 230, 181, 253, 193, 232, 181, 253, 193, 234, 181, 253, 193, 236, 181, 253, 193, 238, 181, 253, 193, 240, 181, 253, 193, 242, 181, 253, 193, 244, 181, 253, 193, 246, 181, 253, 193, 248, 181, 253, 193, 250, 181, 253, 193, 252, 181, 253, 193, 254, 181, 253, 193, 2, 182, 253, 193, 4, 182, 253, 193, 6, 182, 253, 193, 8, 182, 253, 193, 10, 182, 253, 193, 12, 182, 253, 193, 216, 181, 253, 193, 218, 181, 253, 193, 220, 181, 253, 193, 26, 182, 253, 193, 28, 182, 253, 193, 30, 182, 253, 193, 32, 182, 253, 193, 88, 182, 253, 193, 142, 182, 253, 193, 144, 182, 253, 193, 198, 182, 253, 193, 200, 182, 253, 193, 202, 182, 253, 193, 204, 182, 253, 193, 206, 182, 253, 193, 208, 182, 253, 193, 60, 183, 253, 193, 88, 183, 253, 193, 90, 183, 253, 193, 92, 183, 253, 193, 94, 183, 253, 193, 122, 183, 253, 193, 124, 183, 253, 193, 180, 183, 253, 193, 182, 183, 253, 193, 184, 183, 253, 193, 2, 184, 253, 193, 18, 184, 253, 193, 20, 184, 253, 193, 22, 184, 253, 193, 24, 184, 253, 193, 52, 184, 253, 193, 14, 182, 253, 193, 38, 182, 253, 193, 40, 182, 253, 193, 42, 182, 253, 193, 44, 182, 253, 193, 46, 182, 253, 193, 48, 182, 253, 193, 50, 182, 253, 193, 52, 182, 253, 193, 54, 182, 253, 193, 56, 182, 253, 193, 58, 182, 253, 193, 60, 182, 253, 193, 62, 182, 253, 193, 64, 182, 253, 193, 66, 182, 253, 193, 68, 182, 253, 193, 70, 182, 253, 193, 72, 182, 253, 193, 74, 182, 253, 193, 76, 182, 253, 193, 78, 182, 253, 193, 80, 182, 253, 193, 82, 182, 253, 193, 84, 182, 253, 193, 86, 182, 253, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 193, 34, 182, 253, 193, 36, 182, 253, 193, 90, 182, 253, 193, 92, 182, 253, 193, 210, 182, 253, 193, 212, 182, 253, 193, 214, 182, 253, 193, 216, 182, 253, 193, 16, 183, 253, 193, 18, 183, 253, 193, 62, 183, 253, 193, 126, 183, 253, 193, 128, 183, 253, 193, 130, 183, 253, 193, 170, 183, 253, 193, 186, 183, 253, 193, 188, 183, 253, 193, 252, 183, 253, 193, 16, 184, 253, 193, 26, 184, 253, 193, 60, 184, 253, 193, 62, 184, 253, 193, 64, 184, 253, 193, 148, 184, 253, 193, 150, 184, 253, 193, 152, 184, 253, 193, 154, 184, 253, 193, 156, 184, 253, 193, 158, 184, 253, 193, 160, 184, 253, 193, 162, 184, 253, 193, 60, 185, 253, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 206, 160, 113, 0, 193, 146, 182, 253, 193, 148, 182, 253, 193, 150, 182, 253, 193, 152, 182, 253, 193, 154, 182, 253, 193, 156, 182, 253, 193, 158, 182, 253, 193, 160, 182, 253, 193, 162, 182, 253, 193, 164, 182, 253, 193, 166, 182, 253, 193, 168, 182, 253, 193, 170, 182, 253, 193, 172, 182, 253, 193, 174, 182, 253, 193, 176, 182, 253, 193, 178, 182, 253, 193, 180, 182, 253, 193, 182, 182, 253, 193, 184, 182, 253, 193, 186, 182, 253, 193, 188, 182, 253, 193, 218, 182, 253, 193, 220, 182, 253, 193, 222, 182, 253, 193, 224, 182, 253, 193, 226, 182, 253, 193, 228, 182, 253, 193, 230, 182, 253, 193, 232, 182, 253, 193, 234, 182, 253, 193, 236, 182, 253, 193, 238, 182, 253, 193, 240, 182, 253, 193, 242, 182, 253, 193, 244, 182, 253, 193, 246, 182, 253, 193, 248, 182, 253, 193, 250, 182, 253, 193, 252, 182, 253, 193, 254, 182, 253, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 193, 192, 182, 253, 193, 46, 184, 253, 193, 48, 184, 253, 193, 94, 184, 253, 193, 96, 184, 253, 193, 234, 184, 253, 193, 236, 184, 253, 193, 158, 185, 253, 193, 160, 185, 253, 193, 162, 185, 253, 193, 164, 185, 253, 193, 166, 185, 253, 193, 168, 185, 253, 193, 94, 186, 253, 193, 96, 186, 253, 193, 98, 186, 253, 193, 100, 186, 253, 193, 102, 186, 253, 193, 36, 187, 253, 193, 38, 187, 253, 193, 40, 187, 253, 193, 42, 187, 253, 193, 44, 187, 253, 193, 70, 188, 253, 193, 72, 188, 253, 193, 74, 188, 253, 193, 76, 188, 253, 193, 78, 188, 253, 193, 36, 189, 253, 193, 38, 189, 253, 193, 40, 189, 253, 193, 42, 189, 253, 193, 194, 182, 253, 193, 6, 183, 253, 193, 8, 183, 253, 193, 10, 183, 253, 193, 54, 183, 253, 193, 84, 183, 253, 193, 86, 183, 253, 193, 120, 183, 253, 193, 178, 183, 253, 193, 224, 183, 253, 193, 234, 183, 253, 193, 246, 183, 253, 193, 8, 184, 253, 193, 10, 184, 253, 193, 12, 184, 253, 193, 14, 184, 253, 193, 98, 184, 253, 193, 238, 184, 253, 193, 240, 184, 253, 193, 170, 185, 253, 193, 172, 185, 253, 193, 104, 186, 253, 193, 106, 186, 253, 193, 108, 186, 253, 193, 110, 186, 253, 193, 46, 187, 253, 193, 48, 187, 253, 193, 50, 187, 253, 193, 52, 187, 253, 193, 54, 187, 253, 193, 80, 188, 253, 193, 82, 188, 253, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 206, 192, 113, 0, 193, 64, 183, 253, 193, 66, 183, 253, 193, 68, 183, 253, 193, 70, 183, 253, 193, 72, 183, 253, 193, 74, 183, 253, 193, 76, 183, 253, 193, 78, 183, 253, 193, 80, 183, 253, 193, 82, 183, 253, 193, 96, 183, 253, 193, 98, 183, 253, 193, 100, 183, 253, 193, 102, 183, 253, 193, 104, 183, 253, 193, 106, 183, 253, 193, 108, 183, 253, 193, 110, 183, 253, 193, 112, 183, 253, 193, 114, 183, 253, 193, 116, 183, 253, 193, 118, 183, 253, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 206, 224, 113, 0, 193, 172, 183, 253, 193, 174, 183, 253, 193, 176, 183, 253, 193, 190, 183, 253, 193, 192, 183, 253, 193, 194, 183, 253, 193, 196, 183, 253, 193, 198, 183, 253, 193, 200, 183, 253, 193, 202, 183, 253, 193, 204, 183, 253, 193, 206, 183, 253, 193, 212, 183, 253, 193, 214, 183, 253, 193, 216, 183, 253, 193, 218, 183, 253, 193, 220, 183, 253, 193, 222, 183, 253, 193, 208, 183, 253, 193, 210, 183, 253, 193, 254, 183, 253, 193, 50, 184, 253, 193, 100, 184, 253, 193, 242, 184, 253, 193, 244, 184, 253, 193, 246, 184, 253, 193, 248, 184, 253, 193, 174, 185, 253, 193, 176, 185, 253, 193, 178, 185, 253, 193, 112, 186, 253, 193, 56, 187, 253, 193, 58, 187, 253, 193, 60, 187, 253, 193, 62, 187, 253, 193, 64, 187, 253, 193, 66, 187, 253, 193, 84, 188, 253, 193, 86, 188, 253, 193, 56, 189, 253, 193, 58, 189, 253, 193, 60, 189, 253, 193, 62, 189, 253, 193, 64, 189, 253, 193, 22, 190, 253, 193, 24, 190, 253, 193, 26, 190, 253, 193, 220, 190, 253, 193, 222, 190, 253, 193, 136, 191, 253, 193, 226, 183, 253, 193, 228, 183, 253, 193, 230, 183, 253, 193, 232, 183, 253, 193, 236, 183, 253, 193, 238, 183, 253, 193, 240, 183, 253, 193, 242, 183, 253, 193, 244, 183, 253, 193, 248, 183, 253, 193, 250, 183, 253, 193, 6, 184, 253, 193, 28, 184, 253, 193, 30, 184, 253, 193, 32, 184, 253, 193, 34, 184, 253, 193, 36, 184, 253, 193, 38, 184, 253, 193, 40, 184, 253, 193, 42, 184, 253, 193, 66, 184, 253, 193, 68, 184, 253, 193, 70, 184, 253, 193, 72, 184, 253, 193, 74, 184, 253, 193, 76, 184, 253, 193, 78, 184, 253, 193, 80, 184, 253, 193, 82, 184, 253, 193, 84, 184, 253, 193, 86, 184, 253, 193, 88, 184, 253, 193, 54, 184, 253, 193, 56, 184, 253, 193, 58, 184, 253, 193, 102, 184, 253, 193, 104, 184, 253, 193, 106, 184, 253, 193, 108, 184, 253, 193, 110, 184, 253, 193, 112, 184, 253, 193, 114, 184, 253, 193, 116, 184, 253, 193, 118, 184, 253, 193, 120, 184, 253, 193, 122, 184, 253, 193, 124, 184, 253, 193, 126, 184, 253, 193, 128, 184, 253, 193, 130, 184, 253, 193, 132, 184, 253, 193, 134, 184, 253, 193, 136, 184, 253, 193, 138, 184, 253, 193, 140, 184, 253, 193, 142, 184, 253, 193, 144, 184, 253, 193, 146, 184, 253, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 193, 144, 185, 253, 193, 246, 185, 253, 193, 248, 185, 253, 193, 250, 185, 253, 193, 252, 185, 253, 193, 254, 185, 253, 193, 2, 186, 253, 193, 4, 186, 253, 193, 6, 186, 253, 193, 8, 186, 253, 193, 10, 186, 253, 193, 12, 186, 253, 193, 14, 186, 253, 193, 16, 186, 253, 193, 18, 186, 253, 193, 20, 186, 253, 193, 22, 186, 253, 193, 24, 186, 253, 193, 26, 186, 253, 193, 28, 186, 253, 193, 30, 186, 253, 193, 32, 186, 253, 193, 34, 186, 253, 193, 90, 184, 253, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 193, 164, 184, 253, 193, 166, 184, 253, 193, 168, 184, 253, 193, 170, 184, 253, 193, 172, 184, 253, 193, 174, 184, 253, 193, 176, 184, 253, 193, 178, 184, 253, 193, 180, 184, 253, 193, 182, 184, 253, 193, 184, 184, 253, 193, 186, 184, 253, 193, 188, 184, 253, 193, 190, 184, 253, 193, 192, 184, 253, 193, 194, 184, 253, 193, 196, 184, 253, 193, 198, 184, 253, 193, 200, 184, 253, 193, 202, 184, 253, 193, 204, 184, 253, 193, 206, 184, 253, 193, 208, 184, 253, 193, 210, 184, 253, 193, 212, 184, 253, 193, 214, 184, 253, 193, 216, 184, 253, 193, 218, 184, 253, 193, 220, 184, 253, 193, 222, 184, 253, 193, 224, 184, 253, 193, 226, 184, 253, 193, 228, 184, 253, 193, 82, 185, 253, 193, 84, 185, 253, 193, 86, 185, 253, 193, 88, 185, 253, 193, 90, 185, 253, 193, 92, 185, 253, 193, 94, 185, 253, 193, 96, 185, 253, 193, 98, 185, 253, 193, 100, 185, 253, 193, 102, 185, 253, 193, 104, 185, 253, 193, 106, 185, 253, 193, 108, 185, 253, 193, 110, 185, 253, 193, 112, 185, 253, 193, 114, 185, 253, 193, 116, 185, 253, 193, 118, 185, 253, 193, 120, 185, 253, 193, 122, 185, 253, 193, 124, 185, 253, 193, 126, 185, 253, 193, 128, 185, 253, 193, 130, 185, 253, 193, 132, 185, 253, 193, 134, 185, 253, 193, 136, 185, 253, 193, 138, 185, 253, 193, 140, 185, 253, 193, 142, 185, 253, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 206, 0, 114, 0, 193, 180, 185, 253, 193, 182, 185, 253, 193, 184, 185, 253, 193, 186, 185, 253, 193, 188, 185, 253, 193, 190, 185, 253, 193, 192, 185, 253, 193, 194, 185, 253, 193, 196, 185, 253, 193, 198, 185, 253, 193, 58, 185, 253, 193, 200, 185, 253, 193, 202, 185, 253, 193, 204, 185, 253, 193, 206, 185, 253, 193, 208, 185, 253, 193, 210, 185, 253, 193, 212, 185, 253, 193, 214, 185, 253, 193, 216, 185, 253, 193, 218, 185, 253, 193, 220, 185, 253, 193, 222, 185, 253, 193, 224, 185, 253, 193, 226, 185, 253, 193, 228, 185, 253, 193, 230, 185, 253, 193, 232, 185, 253, 193, 114, 186, 253, 193, 116, 186, 253, 193, 118, 186, 253, 193, 120, 186, 253, 193, 122, 186, 253, 193, 234, 185, 253, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 193, 62, 185, 253, 193, 64, 185, 253, 193, 66, 185, 253, 193, 68, 185, 253, 193, 70, 185, 253, 193, 72, 185, 253, 193, 74, 185, 253, 193, 76, 185, 253, 193, 78, 185, 253, 193, 80, 185, 253, 193, 236, 185, 253, 193, 238, 185, 253, 193, 240, 185, 253, 193, 242, 185, 253, 193, 244, 185, 253, 193, 176, 186, 253, 193, 178, 186, 253, 193, 180, 186, 253, 193, 182, 186, 253, 193, 184, 186, 253, 193, 186, 186, 253, 193, 188, 186, 253, 193, 190, 186, 253, 193, 192, 186, 253, 193, 194, 186, 253, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 193, 146, 185, 253, 193, 76, 194, 253, 193, 78, 194, 253, 193, 80, 194, 253, 193, 82, 194, 253, 193, 150, 194, 253, 193, 152, 194, 253, 193, 154, 194, 253, 193, 156, 194, 253, 193, 26, 187, 253, 193, 250, 194, 253, 193, 252, 194, 253, 193, 114, 195, 253, 193, 254, 191, 253, 193, 196, 195, 253, 193, 224, 195, 253, 193, 226, 195, 253, 193, 228, 195, 253, 193, 254, 195, 253, 193, 2, 196, 253, 193, 4, 196, 253, 193, 6, 196, 253, 193, 8, 196, 253, 193, 10, 196, 253, 193, 12, 196, 253, 193, 14, 196, 253, 193, 16, 196, 253, 193, 18, 196, 253, 193, 20, 196, 253, 193, 52, 196, 253, 193, 54, 196, 253, 193, 56, 196, 253, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 32, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 64, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 206, 128, 114, 0, 193, 88, 188, 253, 193, 90, 188, 253, 193, 92, 188, 253, 193, 94, 188, 253, 193, 96, 188, 253, 193, 136, 187, 253, 193, 98, 188, 253, 193, 174, 186, 253, 193, 100, 188, 253, 193, 102, 188, 253, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 96, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 193, 28, 187, 253, 193, 30, 187, 253, 193, 32, 187, 253, 193, 34, 187, 253, 193, 62, 188, 253, 193, 64, 188, 253, 193, 66, 188, 253, 193, 68, 188, 253, 193, 30, 189, 253, 193, 32, 189, 253, 193, 34, 189, 253, 193, 246, 189, 253, 193, 248, 189, 253, 193, 250, 189, 253, 193, 252, 189, 253, 193, 254, 189, 253, 193, 188, 190, 253, 193, 190, 190, 253, 193, 192, 190, 253, 193, 194, 190, 253, 193, 196, 190, 253, 193, 198, 190, 253, 193, 200, 190, 253, 193, 106, 191, 253, 193, 108, 191, 253, 193, 110, 191, 253, 193, 112, 191, 253, 193, 114, 191, 253, 193, 116, 191, 253, 193, 2, 192, 253, 193, 4, 192, 253, 193, 6, 192, 253, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 206, 160, 114, 0, 193, 162, 188, 253, 193, 164, 188, 253, 193, 166, 188, 253, 193, 168, 188, 253, 193, 170, 188, 253, 193, 172, 188, 253, 193, 174, 188, 253, 193, 176, 188, 253, 193, 178, 188, 253, 193, 180, 188, 253, 193, 108, 189, 253, 193, 110, 189, 253, 193, 112, 189, 253, 193, 114, 189, 253, 193, 116, 189, 253, 193, 118, 189, 253, 193, 120, 189, 253, 193, 122, 189, 253, 193, 124, 189, 253, 193, 80, 190, 253, 193, 82, 190, 253, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 206, 192, 114, 0, 193, 182, 188, 253, 193, 184, 188, 253, 193, 104, 188, 253, 193, 106, 188, 253, 193, 108, 188, 253, 193, 110, 188, 253, 193, 112, 188, 253, 193, 114, 188, 253, 193, 116, 188, 253, 193, 118, 188, 253, 193, 120, 188, 253, 193, 122, 188, 253, 193, 124, 188, 253, 193, 126, 188, 253, 193, 128, 188, 253, 193, 130, 188, 253, 193, 132, 188, 253, 193, 134, 188, 253, 193, 136, 188, 253, 193, 138, 188, 253, 193, 140, 188, 253, 193, 142, 188, 253, 193, 144, 188, 253, 193, 146, 188, 253, 193, 148, 188, 253, 193, 150, 188, 253, 193, 152, 188, 253, 193, 154, 188, 253, 193, 156, 188, 253, 193, 158, 188, 253, 193, 66, 189, 253, 193, 68, 189, 253, 193, 70, 189, 253, 193, 72, 189, 253, 193, 160, 188, 253, 193, 74, 189, 253, 193, 76, 189, 253, 193, 78, 189, 253, 193, 80, 189, 253, 193, 82, 189, 253, 193, 84, 189, 253, 193, 86, 189, 253, 193, 88, 189, 253, 193, 90, 189, 253, 193, 92, 189, 253, 193, 94, 189, 253, 193, 96, 189, 253, 193, 98, 189, 253, 193, 100, 189, 253, 193, 102, 189, 253, 193, 104, 189, 253, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 193, 186, 188, 253, 193, 188, 188, 253, 193, 190, 188, 253, 193, 192, 188, 253, 193, 194, 188, 253, 193, 196, 188, 253, 193, 198, 188, 253, 193, 200, 188, 253, 193, 202, 188, 253, 193, 204, 188, 253, 193, 206, 188, 253, 193, 208, 188, 253, 193, 210, 188, 253, 193, 212, 188, 253, 193, 214, 188, 253, 193, 216, 188, 253, 193, 218, 188, 253, 193, 220, 188, 253, 193, 222, 188, 253, 193, 224, 188, 253, 193, 226, 188, 253, 193, 228, 188, 253, 193, 230, 188, 253, 193, 232, 188, 253, 193, 234, 188, 253, 193, 236, 188, 253, 193, 238, 188, 253, 193, 240, 188, 253, 193, 242, 188, 253, 193, 244, 188, 253, 193, 246, 188, 253, 193, 248, 188, 253, 193, 250, 188, 253, 193, 252, 188, 253, 193, 254, 188, 253, 193, 2, 189, 253, 193, 4, 189, 253, 193, 6, 189, 253, 193, 8, 189, 253, 193, 10, 189, 253, 193, 12, 189, 253, 193, 14, 189, 253, 193, 16, 189, 253, 193, 18, 189, 253, 193, 20, 189, 253, 193, 22, 189, 253, 193, 24, 189, 253, 193, 26, 189, 253, 193, 28, 189, 253, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 193, 44, 189, 253, 193, 46, 189, 253, 193, 48, 189, 253, 193, 50, 189, 253, 193, 52, 189, 253, 193, 54, 189, 253, 193, 12, 190, 253, 193, 14, 190, 253, 193, 16, 190, 253, 193, 18, 190, 253, 193, 20, 190, 253, 193, 214, 190, 253, 193, 216, 190, 253, 193, 218, 190, 253, 193, 126, 191, 253, 193, 128, 191, 253, 193, 130, 191, 253, 193, 132, 191, 253, 193, 134, 191, 253, 193, 22, 192, 253, 193, 24, 192, 253, 193, 102, 192, 253, 193, 160, 192, 253, 193, 210, 192, 253, 193, 212, 192, 253, 193, 246, 192, 253, 193, 248, 192, 253, 193, 130, 193, 253, 193, 28, 194, 253, 193, 30, 194, 253, 193, 100, 194, 253, 193, 184, 194, 253, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 224, 114, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 193, 4, 190, 253, 193, 6, 190, 253, 193, 8, 190, 253, 193, 10, 190, 253, 193, 202, 190, 253, 193, 204, 190, 253, 193, 206, 190, 253, 193, 208, 190, 253, 193, 210, 190, 253, 193, 212, 190, 253, 193, 118, 191, 253, 193, 120, 191, 253, 193, 122, 191, 253, 193, 124, 191, 253, 193, 14, 192, 253, 193, 16, 192, 253, 193, 18, 192, 253, 193, 20, 192, 253, 193, 94, 192, 253, 193, 96, 192, 253, 193, 98, 192, 253, 193, 100, 192, 253, 193, 158, 192, 253, 193, 204, 192, 253, 193, 206, 192, 253, 193, 208, 192, 253, 193, 244, 192, 253, 193, 20, 193, 253, 193, 22, 193, 253, 193, 34, 193, 253, 193, 36, 193, 253, 193, 54, 193, 253, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 206, 0, 115, 0, 193, 224, 190, 253, 193, 226, 190, 253, 193, 228, 190, 253, 193, 230, 190, 253, 193, 232, 190, 253, 193, 234, 190, 253, 193, 236, 190, 253, 193, 238, 190, 253, 193, 240, 190, 253, 193, 242, 190, 253, 193, 244, 190, 253, 193, 246, 190, 253, 193, 248, 190, 253, 193, 250, 190, 253, 193, 252, 190, 253, 193, 254, 190, 253, 193, 2, 191, 253, 193, 4, 191, 253, 193, 6, 191, 253, 206, 64, 115, 0, 206, 64, 115, 0, 206, 64, 115, 0, 206, 64, 115, 0, 193, 76, 190, 253, 193, 160, 203, 253, 193, 162, 203, 253, 193, 164, 203, 253, 193, 166, 203, 253, 193, 168, 203, 253, 193, 170, 203, 253, 193, 172, 203, 253, 193, 174, 203, 253, 193, 176, 203, 253, 193, 178, 203, 253, 193, 180, 203, 253, 193, 182, 203, 253, 193, 184, 203, 253, 193, 186, 203, 253, 193, 188, 203, 253, 193, 190, 203, 253, 193, 192, 203, 253, 193, 194, 203, 253, 193, 196, 203, 253, 193, 198, 203, 253, 193, 200, 203, 253, 193, 202, 203, 253, 193, 204, 203, 253, 193, 206, 203, 253, 193, 128, 204, 253, 193, 208, 203, 253, 193, 210, 203, 253, 193, 212, 203, 253, 193, 130, 204, 253, 193, 132, 204, 253, 193, 134, 204, 253, 193, 84, 190, 253, 193, 86, 190, 253, 193, 88, 190, 253, 193, 90, 190, 253, 193, 92, 190, 253, 193, 94, 190, 253, 193, 96, 190, 253, 193, 98, 190, 253, 193, 100, 190, 253, 193, 8, 191, 253, 193, 10, 191, 253, 193, 12, 191, 253, 193, 14, 191, 253, 193, 16, 191, 253, 193, 18, 191, 253, 193, 20, 191, 253, 193, 172, 191, 253, 193, 174, 191, 253, 193, 176, 191, 253, 193, 178, 191, 253, 193, 180, 191, 253, 193, 182, 191, 253, 193, 40, 192, 253, 193, 42, 192, 253, 193, 44, 192, 253, 193, 118, 192, 253, 193, 174, 192, 253, 193, 176, 192, 253, 193, 62, 193, 253, 193, 104, 193, 253, 193, 180, 193, 253, 193, 182, 193, 253, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 206, 32, 115, 0, 193, 22, 191, 253, 193, 24, 191, 253, 193, 26, 191, 253, 193, 28, 191, 253, 193, 30, 191, 253, 193, 32, 191, 253, 193, 34, 191, 253, 193, 36, 191, 253, 193, 38, 191, 253, 193, 40, 191, 253, 193, 42, 191, 253, 193, 44, 191, 253, 193, 46, 191, 253, 193, 48, 191, 253, 193, 50, 191, 253, 193, 52, 191, 253, 193, 54, 191, 253, 193, 56, 191, 253, 193, 58, 191, 253, 193, 60, 191, 253, 193, 62, 191, 253, 193, 64, 191, 253, 193, 66, 191, 253, 193, 68, 191, 253, 193, 70, 191, 253, 193, 72, 191, 253, 193, 74, 191, 253, 193, 76, 191, 253, 193, 78, 191, 253, 193, 80, 191, 253, 193, 82, 191, 253, 193, 84, 191, 253, 193, 86, 191, 253, 193, 88, 191, 253, 193, 90, 191, 253, 193, 92, 191, 253, 193, 94, 191, 253, 193, 96, 191, 253, 193, 98, 191, 253, 193, 100, 191, 253, 193, 102, 191, 253, 193, 104, 191, 253, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 193, 138, 191, 253, 193, 140, 191, 253, 193, 26, 192, 253, 193, 104, 192, 253, 193, 162, 192, 253, 193, 164, 192, 253, 193, 166, 192, 253, 193, 168, 192, 253, 193, 214, 192, 253, 193, 250, 192, 253, 193, 38, 193, 253, 193, 58, 193, 253, 193, 60, 193, 253, 193, 66, 193, 253, 193, 74, 193, 253, 193, 90, 193, 253, 193, 132, 193, 253, 193, 134, 193, 253, 193, 136, 193, 253, 193, 138, 193, 253, 193, 140, 193, 253, 193, 142, 193, 253, 193, 144, 193, 253, 193, 146, 193, 253, 193, 148, 193, 253, 193, 150, 193, 253, 206, 192, 115, 0, 206, 192, 115, 0, 206, 192, 115, 0, 206, 192, 115, 0, 206, 192, 115, 0, 206, 192, 115, 0, 206, 64, 115, 0, 206, 64, 115, 0, 206, 64, 115, 0, 206, 64, 115, 0, 206, 64, 115, 0, 206, 64, 115, 0, 206, 64, 115, 0, 206, 64, 115, 0, 193, 28, 192, 253, 193, 166, 191, 253, 193, 168, 191, 253, 193, 170, 191, 253, 193, 30, 192, 253, 193, 32, 192, 253, 193, 34, 192, 253, 193, 36, 192, 253, 193, 38, 192, 253, 193, 106, 192, 253, 193, 108, 192, 253, 193, 110, 192, 253, 193, 112, 192, 253, 193, 114, 192, 253, 193, 116, 192, 253, 193, 170, 192, 253, 193, 172, 192, 253, 193, 252, 192, 253, 193, 24, 193, 253, 193, 56, 193, 253, 193, 64, 193, 253, 193, 68, 193, 253, 193, 76, 193, 253, 193, 92, 193, 253, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 206, 96, 115, 0, 193, 46, 192, 253, 193, 48, 192, 253, 193, 50, 192, 253, 193, 52, 192, 253, 193, 54, 192, 253, 193, 56, 192, 253, 193, 58, 192, 253, 193, 60, 192, 253, 193, 62, 192, 253, 193, 64, 192, 253, 193, 66, 192, 253, 193, 68, 192, 253, 193, 70, 192, 253, 193, 72, 192, 253, 193, 74, 192, 253, 193, 76, 192, 253, 193, 78, 192, 253, 193, 80, 192, 253, 193, 82, 192, 253, 193, 84, 192, 253, 193, 86, 192, 253, 193, 88, 192, 253, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 193, 8, 192, 253, 193, 10, 192, 253, 193, 90, 192, 253, 193, 92, 192, 253, 193, 156, 192, 253, 193, 242, 192, 253, 193, 32, 193, 253, 193, 52, 193, 253, 193, 78, 193, 253, 193, 80, 193, 253, 193, 82, 193, 253, 193, 112, 193, 253, 193, 114, 193, 253, 193, 116, 193, 253, 193, 184, 193, 253, 193, 186, 193, 253, 193, 188, 193, 253, 193, 190, 193, 253, 193, 192, 193, 253, 193, 10, 194, 253, 193, 12, 194, 253, 193, 14, 194, 253, 193, 16, 194, 253, 193, 18, 194, 253, 193, 20, 194, 253, 193, 84, 194, 253, 193, 86, 194, 253, 193, 88, 194, 253, 193, 158, 194, 253, 193, 160, 194, 253, 193, 162, 194, 253, 193, 164, 194, 253, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 206, 128, 115, 0, 193, 178, 192, 253, 193, 180, 192, 253, 193, 182, 192, 253, 193, 184, 192, 253, 193, 186, 192, 253, 193, 188, 192, 253, 193, 190, 192, 253, 193, 192, 192, 253, 193, 194, 192, 253, 193, 196, 192, 253, 193, 198, 192, 253, 193, 200, 192, 253, 193, 202, 192, 253, 206, 160, 115, 0, 206, 160, 115, 0, 206, 160, 115, 0, 206, 160, 115, 0, 206, 160, 115, 0, 206, 160, 115, 0, 206, 160, 115, 0, 206, 160, 115, 0, 206, 160, 115, 0, 193, 2, 193, 253, 193, 4, 193, 253, 193, 6, 193, 253, 193, 8, 193, 253, 193, 10, 193, 253, 193, 12, 193, 253, 193, 14, 193, 253, 193, 16, 193, 253, 193, 18, 193, 253, 193, 26, 193, 253, 193, 28, 193, 253, 193, 30, 193, 253, 193, 40, 193, 253, 193, 42, 193, 253, 193, 44, 193, 253, 193, 46, 193, 253, 193, 48, 193, 253, 193, 50, 193, 253, 193, 106, 193, 253, 193, 108, 193, 253, 193, 110, 193, 253, 193, 2, 194, 253, 193, 4, 194, 253, 193, 6, 194, 253, 193, 8, 194, 253, 193, 70, 193, 253, 193, 72, 193, 253, 193, 84, 193, 253, 193, 86, 193, 253, 193, 88, 193, 253, 193, 120, 193, 253, 193, 122, 193, 253, 193, 124, 193, 253, 193, 126, 193, 253, 193, 128, 193, 253, 193, 196, 193, 253, 193, 198, 193, 253, 193, 200, 193, 253, 193, 202, 193, 253, 193, 22, 194, 253, 193, 24, 194, 253, 193, 26, 194, 253, 193, 90, 194, 253, 193, 92, 194, 253, 193, 94, 194, 253, 193, 96, 194, 253, 193, 98, 194, 253, 193, 170, 194, 253, 193, 172, 194, 253, 193, 10, 195, 253, 193, 174, 194, 253, 193, 176, 194, 253, 193, 178, 194, 253, 193, 180, 194, 253, 193, 182, 194, 253, 193, 12, 195, 253, 193, 14, 195, 253, 193, 94, 193, 253, 193, 96, 193, 253, 193, 98, 193, 253, 193, 100, 193, 253, 193, 152, 193, 253, 193, 154, 193, 253, 193, 156, 193, 253, 193, 158, 193, 253, 193, 160, 193, 253, 193, 162, 193, 253, 193, 164, 193, 253, 193, 166, 193, 253, 193, 168, 193, 253, 193, 170, 193, 253, 193, 172, 193, 253, 193, 174, 193, 253, 193, 176, 193, 253, 193, 226, 193, 253, 193, 228, 193, 253, 193, 230, 193, 253, 193, 232, 193, 253, 193, 234, 193, 253, 193, 236, 193, 253, 193, 238, 193, 253, 193, 240, 193, 253, 193, 242, 193, 253, 193, 244, 193, 253, 193, 246, 193, 253, 193, 248, 193, 253, 193, 250, 193, 253, 193, 252, 193, 253, 193, 254, 193, 253, 206, 192, 115, 0, 206, 192, 115, 0, 206, 192, 115, 0, 206, 192, 115, 0, 206, 192, 115, 0, 193, 32, 194, 253, 193, 34, 194, 253, 193, 36, 194, 253, 193, 38, 194, 253, 193, 40, 194, 253, 193, 42, 194, 253, 193, 44, 194, 253, 193, 102, 194, 253, 193, 104, 194, 253, 193, 106, 194, 253, 193, 108, 194, 253, 193, 110, 194, 253, 193, 112, 194, 253, 193, 114, 194, 253, 193, 116, 194, 253, 193, 118, 194, 253, 193, 188, 194, 253, 193, 190, 194, 253, 193, 192, 194, 253, 193, 194, 194, 253, 193, 196, 194, 253, 193, 198, 194, 253, 193, 200, 194, 253, 193, 202, 194, 253, 193, 204, 194, 253, 193, 206, 194, 253, 193, 208, 194, 253, 193, 46, 194, 253, 193, 48, 194, 253, 193, 50, 194, 253, 193, 52, 194, 253, 193, 54, 194, 253, 193, 56, 194, 253, 193, 58, 194, 253, 193, 60, 194, 253, 193, 62, 194, 253, 193, 64, 194, 253, 193, 66, 194, 253, 193, 68, 194, 253, 193, 70, 194, 253, 193, 120, 194, 253, 193, 122, 194, 253, 193, 124, 194, 253, 193, 126, 194, 253, 193, 128, 194, 253, 193, 130, 194, 253, 193, 132, 194, 253, 193, 134, 194, 253, 193, 136, 194, 253, 193, 138, 194, 253, 193, 140, 194, 253, 193, 142, 194, 253, 193, 144, 194, 253, 193, 146, 194, 253, 193, 148, 194, 253, 193, 214, 194, 253, 193, 216, 194, 253, 193, 48, 195, 253, 193, 218, 194, 253, 193, 166, 194, 253, 193, 168, 194, 253, 193, 254, 194, 253, 193, 2, 195, 253, 193, 4, 195, 253, 193, 6, 195, 253, 193, 70, 195, 253, 193, 116, 195, 253, 193, 118, 195, 253, 193, 144, 195, 253, 193, 146, 195, 253, 193, 178, 195, 253, 193, 188, 195, 253, 193, 198, 195, 253, 193, 230, 195, 253, 193, 12, 197, 253, 193, 14, 197, 253, 193, 16, 197, 253, 193, 18, 197, 253, 193, 250, 197, 253, 193, 252, 197, 253, 193, 254, 197, 253, 193, 2, 198, 253, 193, 4, 198, 253, 193, 232, 198, 253, 193, 234, 198, 253, 193, 236, 198, 253, 193, 238, 198, 253, 193, 174, 199, 253, 193, 176, 199, 253, 193, 178, 199, 253, 193, 180, 199, 253, 193, 186, 194, 253, 193, 76, 195, 253, 193, 78, 195, 253, 193, 128, 195, 253, 193, 202, 195, 253, 193, 24, 196, 253, 193, 96, 196, 253, 193, 98, 196, 253, 193, 100, 196, 253, 193, 102, 196, 253, 193, 104, 196, 253, 193, 30, 197, 253, 193, 32, 197, 253, 193, 34, 197, 253, 193, 36, 197, 253, 193, 38, 197, 253, 193, 24, 198, 253, 193, 26, 198, 253, 193, 28, 198, 253, 193, 252, 198, 253, 193, 254, 198, 253, 193, 2, 199, 253, 193, 4, 199, 253, 193, 6, 199, 253, 193, 8, 199, 253, 193, 200, 199, 253, 193, 202, 199, 253, 193, 204, 199, 253, 206, 64, 117, 0, 206, 64, 117, 0, 206, 64, 117, 0, 206, 64, 117, 0, 193, 210, 194, 253, 193, 212, 194, 253, 193, 34, 195, 253, 193, 36, 195, 253, 193, 38, 195, 253, 193, 40, 195, 253, 193, 42, 195, 253, 193, 44, 195, 253, 193, 46, 195, 253, 193, 80, 195, 253, 193, 82, 195, 253, 193, 84, 195, 253, 193, 86, 195, 253, 193, 88, 195, 253, 193, 90, 195, 253, 193, 130, 195, 253, 193, 132, 195, 253, 193, 134, 195, 253, 193, 162, 195, 253, 193, 164, 195, 253, 193, 166, 195, 253, 193, 168, 195, 253, 193, 184, 195, 253, 193, 192, 195, 253, 193, 194, 195, 253, 193, 204, 195, 253, 193, 208, 195, 253, 193, 210, 195, 253, 193, 212, 195, 253, 193, 26, 196, 253, 193, 40, 197, 253, 193, 42, 197, 253, 193, 220, 194, 253, 193, 222, 194, 253, 193, 224, 194, 253, 193, 226, 194, 253, 193, 228, 194, 253, 193, 230, 194, 253, 193, 232, 194, 253, 193, 234, 194, 253, 193, 236, 194, 253, 193, 238, 194, 253, 193, 240, 194, 253, 193, 242, 194, 253, 193, 50, 195, 253, 193, 52, 195, 253, 193, 54, 195, 253, 193, 56, 195, 253, 193, 58, 195, 253, 193, 60, 195, 253, 193, 62, 195, 253, 193, 92, 195, 253, 193, 94, 195, 253, 193, 96, 195, 253, 193, 98, 195, 253, 193, 100, 195, 253, 193, 102, 195, 253, 193, 104, 195, 253, 193, 106, 195, 253, 193, 108, 195, 253, 193, 136, 195, 253, 193, 138, 195, 253, 193, 140, 195, 253, 193, 142, 195, 253, 193, 246, 194, 253, 193, 248, 194, 253, 193, 64, 195, 253, 193, 66, 195, 253, 193, 68, 195, 253, 193, 252, 195, 253, 193, 40, 196, 253, 193, 42, 196, 253, 193, 44, 196, 253, 193, 46, 196, 253, 193, 48, 196, 253, 193, 50, 196, 253, 193, 152, 196, 253, 193, 154, 196, 253, 193, 156, 196, 253, 193, 158, 196, 253, 193, 160, 196, 253, 193, 162, 196, 253, 193, 164, 196, 253, 193, 166, 196, 253, 193, 168, 196, 253, 193, 170, 196, 253, 193, 172, 196, 253, 193, 174, 196, 253, 193, 176, 196, 253, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 193, 16, 195, 253, 193, 18, 195, 253, 193, 20, 195, 253, 193, 22, 195, 253, 193, 24, 195, 253, 193, 26, 195, 253, 193, 28, 195, 253, 193, 30, 195, 253, 193, 32, 195, 253, 193, 72, 195, 253, 193, 74, 195, 253, 193, 122, 195, 253, 193, 124, 195, 253, 193, 126, 195, 253, 193, 148, 195, 253, 193, 150, 195, 253, 193, 152, 195, 253, 193, 154, 195, 253, 193, 156, 195, 253, 193, 158, 195, 253, 193, 160, 195, 253, 193, 180, 195, 253, 193, 182, 195, 253, 193, 190, 195, 253, 193, 200, 195, 253, 193, 206, 195, 253, 193, 214, 195, 253, 193, 216, 195, 253, 193, 22, 196, 253, 193, 90, 196, 253, 193, 92, 196, 253, 193, 94, 196, 253, 193, 170, 195, 253, 193, 172, 195, 253, 193, 174, 195, 253, 193, 176, 195, 253, 193, 186, 195, 253, 193, 218, 195, 253, 193, 222, 195, 253, 193, 232, 195, 253, 193, 234, 195, 253, 193, 236, 195, 253, 193, 238, 195, 253, 193, 240, 195, 253, 193, 242, 195, 253, 193, 244, 195, 253, 193, 246, 195, 253, 193, 248, 195, 253, 193, 250, 195, 253, 193, 28, 196, 253, 193, 30, 196, 253, 193, 32, 196, 253, 193, 34, 196, 253, 193, 36, 196, 253, 193, 38, 196, 253, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 193, 58, 196, 253, 193, 60, 196, 253, 193, 62, 196, 253, 193, 64, 196, 253, 193, 66, 196, 253, 193, 68, 196, 253, 193, 70, 196, 253, 193, 72, 196, 253, 193, 74, 196, 253, 193, 76, 196, 253, 193, 78, 196, 253, 193, 80, 196, 253, 193, 82, 196, 253, 193, 84, 196, 253, 193, 86, 196, 253, 193, 88, 196, 253, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 206, 224, 115, 0, 193, 44, 197, 253, 193, 144, 196, 253, 193, 146, 196, 253, 193, 148, 196, 253, 193, 150, 196, 253, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 0, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 193, 20, 197, 253, 193, 22, 197, 253, 193, 24, 197, 253, 193, 26, 197, 253, 193, 28, 197, 253, 193, 8, 198, 253, 193, 10, 198, 253, 193, 12, 198, 253, 193, 14, 198, 253, 193, 16, 198, 253, 193, 18, 198, 253, 193, 20, 198, 253, 193, 22, 198, 253, 193, 240, 198, 253, 193, 242, 198, 253, 193, 244, 198, 253, 193, 246, 198, 253, 193, 248, 198, 253, 193, 250, 198, 253, 193, 184, 199, 253, 193, 186, 199, 253, 193, 118, 207, 253, 193, 188, 199, 253, 193, 190, 199, 253, 193, 192, 199, 253, 193, 194, 199, 253, 193, 196, 199, 253, 193, 198, 199, 253, 193, 208, 200, 253, 193, 210, 200, 253, 193, 212, 200, 253, 193, 214, 200, 253, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 32, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 206, 64, 116, 0, 193, 100, 198, 253, 193, 102, 198, 253, 193, 104, 198, 253, 193, 106, 198, 253, 193, 108, 198, 253, 193, 110, 198, 253, 193, 112, 198, 253, 193, 114, 198, 253, 193, 116, 198, 253, 193, 118, 198, 253, 193, 68, 199, 253, 193, 70, 199, 253, 193, 72, 199, 253, 193, 74, 199, 253, 193, 76, 199, 253, 193, 78, 199, 253, 193, 80, 199, 253, 193, 82, 199, 253, 193, 48, 200, 253, 193, 50, 200, 253, 193, 52, 200, 253, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 96, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 193, 30, 198, 253, 193, 32, 198, 253, 193, 34, 198, 253, 193, 36, 198, 253, 193, 38, 198, 253, 193, 40, 198, 253, 193, 10, 199, 253, 193, 12, 199, 253, 193, 206, 199, 253, 193, 208, 199, 253, 193, 210, 199, 253, 193, 212, 199, 253, 193, 214, 199, 253, 193, 236, 200, 253, 193, 238, 200, 253, 193, 240, 200, 253, 193, 242, 200, 253, 193, 228, 201, 253, 193, 230, 201, 253, 193, 232, 201, 253, 193, 200, 202, 253, 193, 202, 202, 253, 193, 204, 202, 253, 193, 206, 202, 253, 193, 208, 202, 253, 193, 158, 203, 253, 193, 120, 204, 253, 193, 122, 204, 253, 193, 124, 204, 253, 193, 126, 204, 253, 193, 252, 204, 253, 193, 254, 204, 253, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 128, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 192, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 160, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 224, 116, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 193, 182, 199, 253, 193, 192, 200, 253, 193, 194, 200, 253, 193, 196, 200, 253, 193, 198, 200, 253, 193, 200, 200, 253, 193, 202, 200, 253, 193, 204, 200, 253, 193, 206, 200, 253, 193, 202, 201, 253, 193, 204, 201, 253, 193, 206, 201, 253, 193, 164, 202, 253, 193, 166, 202, 253, 193, 168, 202, 253, 193, 170, 202, 253, 193, 136, 203, 253, 193, 102, 204, 253, 193, 104, 204, 253, 193, 106, 204, 253, 193, 108, 204, 253, 193, 110, 204, 253, 193, 240, 204, 253, 193, 82, 205, 253, 193, 142, 205, 253, 193, 32, 206, 253, 193, 122, 206, 253, 193, 124, 206, 253, 193, 126, 206, 253, 193, 128, 206, 253, 193, 130, 206, 253, 193, 188, 206, 253, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 206, 0, 117, 0, 193, 244, 200, 253, 193, 46, 200, 253, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 193, 54, 200, 253, 193, 56, 200, 253, 193, 58, 200, 253, 193, 60, 200, 253, 193, 62, 200, 253, 193, 64, 200, 253, 193, 44, 201, 253, 193, 46, 201, 253, 193, 48, 201, 253, 193, 50, 201, 253, 193, 52, 201, 253, 193, 54, 201, 253, 193, 56, 201, 253, 193, 58, 201, 253, 193, 60, 201, 253, 193, 62, 201, 253, 193, 64, 201, 253, 193, 50, 202, 253, 193, 52, 202, 253, 193, 54, 202, 253, 193, 56, 202, 253, 193, 58, 202, 253, 193, 60, 202, 253, 193, 244, 202, 253, 193, 246, 202, 253, 193, 248, 202, 253, 193, 250, 202, 253, 193, 252, 202, 253, 193, 254, 202, 253, 193, 2, 203, 253, 193, 214, 203, 253, 193, 216, 203, 253, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 32, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 193, 216, 200, 253, 193, 218, 200, 253, 193, 208, 201, 253, 193, 210, 201, 253, 193, 212, 201, 253, 193, 214, 201, 253, 193, 216, 201, 253, 193, 218, 201, 253, 193, 172, 202, 253, 193, 174, 202, 253, 193, 176, 202, 253, 193, 178, 202, 253, 193, 180, 202, 253, 193, 182, 202, 253, 193, 140, 203, 253, 193, 142, 203, 253, 193, 144, 203, 253, 193, 112, 204, 253, 193, 114, 204, 253, 193, 116, 204, 253, 193, 242, 204, 253, 193, 244, 204, 253, 193, 246, 204, 253, 193, 84, 205, 253, 193, 144, 205, 253, 193, 202, 205, 253, 193, 46, 206, 253, 193, 74, 206, 253, 193, 82, 206, 253, 193, 84, 206, 253, 193, 96, 206, 253, 193, 134, 206, 253, 206, 64, 117, 0, 206, 64, 117, 0, 206, 64, 117, 0, 206, 64, 117, 0, 193, 220, 201, 253, 193, 222, 201, 253, 193, 224, 201, 253, 193, 226, 201, 253, 193, 184, 202, 253, 193, 186, 202, 253, 193, 188, 202, 253, 193, 190, 202, 253, 193, 192, 202, 253, 193, 194, 202, 253, 193, 196, 202, 253, 193, 198, 202, 253, 193, 146, 203, 253, 193, 148, 203, 253, 193, 150, 203, 253, 193, 152, 203, 253, 193, 154, 203, 253, 193, 156, 203, 253, 193, 118, 204, 253, 193, 248, 204, 253, 193, 250, 204, 253, 193, 86, 205, 253, 193, 88, 205, 253, 193, 146, 205, 253, 193, 148, 205, 253, 193, 204, 205, 253, 193, 206, 205, 253, 193, 12, 206, 253, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 96, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 128, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 206, 160, 117, 0, 193, 210, 202, 253, 193, 42, 202, 253, 193, 44, 202, 253, 193, 46, 202, 253, 193, 48, 202, 253, 193, 212, 202, 253, 193, 214, 202, 253, 193, 216, 202, 253, 193, 218, 202, 253, 193, 220, 202, 253, 193, 222, 202, 253, 193, 224, 202, 253, 193, 226, 202, 253, 193, 228, 202, 253, 193, 230, 202, 253, 193, 232, 202, 253, 193, 234, 202, 253, 193, 236, 202, 253, 193, 238, 202, 253, 193, 240, 202, 253, 193, 242, 202, 253, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 192, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 224, 117, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 193, 218, 203, 253, 193, 220, 203, 253, 193, 222, 203, 253, 193, 224, 203, 253, 193, 226, 203, 253, 193, 228, 203, 253, 193, 230, 203, 253, 193, 158, 204, 253, 193, 160, 204, 253, 193, 162, 204, 253, 193, 164, 204, 253, 193, 166, 204, 253, 193, 24, 205, 253, 193, 26, 205, 253, 193, 96, 205, 253, 193, 158, 205, 253, 193, 34, 206, 253, 193, 62, 206, 253, 193, 64, 206, 253, 193, 114, 206, 253, 193, 8, 207, 253, 193, 10, 207, 253, 193, 182, 207, 253, 193, 254, 207, 253, 193, 74, 208, 253, 193, 188, 208, 253, 193, 250, 208, 253, 193, 34, 209, 253, 193, 96, 209, 253, 193, 118, 209, 253, 193, 140, 209, 253, 193, 198, 209, 253, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 206, 0, 118, 0, 193, 168, 204, 253, 193, 100, 204, 253, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 193, 136, 204, 253, 193, 138, 204, 253, 193, 140, 204, 253, 193, 142, 204, 253, 193, 144, 204, 253, 193, 146, 204, 253, 193, 148, 204, 253, 193, 150, 204, 253, 193, 152, 204, 253, 193, 154, 204, 253, 193, 156, 204, 253, 193, 8, 205, 253, 193, 10, 205, 253, 193, 12, 205, 253, 193, 14, 205, 253, 193, 16, 205, 253, 193, 18, 205, 253, 193, 20, 205, 253, 193, 22, 205, 253, 193, 92, 205, 253, 193, 94, 205, 253, 193, 152, 205, 253, 193, 154, 205, 253, 193, 156, 205, 253, 193, 212, 205, 253, 193, 214, 205, 253, 193, 216, 205, 253, 193, 244, 205, 253, 193, 246, 205, 253, 193, 14, 206, 253, 193, 16, 206, 253, 193, 72, 206, 253, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 206, 32, 118, 0, 193, 28, 205, 253, 193, 30, 205, 253, 193, 32, 205, 253, 193, 34, 205, 253, 193, 36, 205, 253, 193, 38, 205, 253, 193, 40, 205, 253, 193, 42, 205, 253, 193, 44, 205, 253, 193, 46, 205, 253, 193, 48, 205, 253, 193, 50, 205, 253, 193, 52, 205, 253, 193, 198, 218, 253, 193, 54, 205, 253, 193, 2, 205, 253, 193, 4, 205, 253, 193, 6, 205, 253, 193, 90, 205, 253, 193, 150, 205, 253, 193, 208, 205, 253, 193, 210, 205, 253, 193, 76, 206, 253, 193, 86, 206, 253, 193, 88, 206, 253, 193, 98, 206, 253, 193, 100, 206, 253, 193, 102, 206, 253, 193, 144, 206, 253, 193, 146, 206, 253, 193, 148, 206, 253, 193, 150, 206, 253, 193, 152, 206, 253, 193, 154, 206, 253, 193, 156, 206, 253, 193, 158, 206, 253, 193, 216, 206, 253, 193, 218, 206, 253, 193, 220, 206, 253, 193, 222, 206, 253, 193, 224, 206, 253, 193, 226, 206, 253, 193, 228, 206, 253, 193, 230, 206, 253, 193, 232, 206, 253, 193, 234, 206, 253, 193, 236, 206, 253, 193, 56, 205, 253, 193, 58, 205, 253, 193, 60, 205, 253, 193, 62, 205, 253, 193, 64, 205, 253, 193, 66, 205, 253, 193, 68, 205, 253, 193, 70, 205, 253, 193, 72, 205, 253, 193, 74, 205, 253, 193, 76, 205, 253, 193, 78, 205, 253, 193, 80, 205, 253, 193, 98, 205, 253, 193, 100, 205, 253, 193, 102, 205, 253, 193, 104, 205, 253, 193, 106, 205, 253, 193, 108, 205, 253, 193, 110, 205, 253, 193, 112, 205, 253, 193, 114, 205, 253, 193, 116, 205, 253, 193, 118, 205, 253, 193, 120, 205, 253, 193, 122, 205, 253, 193, 124, 205, 253, 193, 126, 205, 253, 193, 128, 205, 253, 193, 130, 205, 253, 193, 132, 205, 253, 193, 134, 205, 253, 193, 136, 205, 253, 193, 138, 205, 253, 193, 140, 205, 253, 193, 160, 205, 253, 193, 162, 205, 253, 193, 164, 205, 253, 193, 166, 205, 253, 193, 168, 205, 253, 193, 170, 205, 253, 193, 172, 205, 253, 193, 174, 205, 253, 193, 176, 205, 253, 193, 178, 205, 253, 193, 180, 205, 253, 193, 182, 205, 253, 193, 184, 205, 253, 193, 186, 205, 253, 193, 188, 205, 253, 193, 190, 205, 253, 193, 192, 205, 253, 193, 194, 205, 253, 193, 196, 205, 253, 193, 198, 205, 253, 193, 200, 205, 253, 206, 64, 118, 0, 206, 64, 118, 0, 206, 64, 118, 0, 206, 64, 118, 0, 206, 64, 118, 0, 206, 64, 118, 0, 206, 64, 118, 0, 206, 64, 118, 0, 206, 64, 118, 0, 193, 248, 205, 253, 193, 250, 205, 253, 193, 252, 205, 253, 193, 254, 205, 253, 193, 2, 206, 253, 193, 4, 206, 253, 193, 6, 206, 253, 193, 8, 206, 253, 193, 10, 206, 253, 193, 20, 206, 253, 193, 22, 206, 253, 193, 24, 206, 253, 193, 26, 206, 253, 193, 28, 206, 253, 193, 30, 206, 253, 193, 38, 206, 253, 193, 40, 206, 253, 193, 42, 206, 253, 193, 44, 206, 253, 193, 48, 206, 253, 193, 50, 206, 253, 193, 52, 206, 253, 193, 54, 206, 253, 193, 56, 206, 253, 193, 58, 206, 253, 193, 66, 206, 253, 193, 68, 206, 253, 193, 60, 206, 253, 193, 138, 206, 253, 193, 140, 206, 253, 193, 142, 206, 253, 193, 214, 206, 253, 193, 48, 207, 253, 193, 130, 207, 253, 193, 132, 207, 253, 193, 96, 208, 253, 193, 98, 208, 253, 193, 136, 208, 253, 193, 24, 209, 253, 193, 46, 209, 253, 193, 66, 209, 253, 193, 68, 209, 253, 193, 90, 209, 253, 193, 154, 209, 253, 193, 182, 209, 253, 193, 204, 209, 253, 193, 26, 210, 253, 193, 70, 210, 253, 193, 118, 210, 253, 193, 162, 210, 253, 193, 164, 210, 253, 193, 192, 210, 253, 193, 212, 210, 253, 193, 214, 210, 253, 193, 166, 211, 253, 193, 34, 212, 253, 193, 36, 212, 253, 193, 72, 212, 253, 193, 96, 212, 253, 193, 70, 206, 253, 193, 116, 206, 253, 193, 118, 206, 253, 193, 120, 206, 253, 193, 186, 206, 253, 193, 12, 207, 253, 193, 14, 207, 253, 193, 16, 207, 253, 193, 18, 207, 253, 193, 20, 207, 253, 193, 22, 207, 253, 193, 24, 207, 253, 193, 104, 207, 253, 193, 106, 207, 253, 193, 184, 207, 253, 193, 2, 208, 253, 193, 4, 208, 253, 193, 76, 208, 253, 193, 78, 208, 253, 193, 80, 208, 253, 193, 82, 208, 253, 193, 222, 208, 253, 193, 244, 208, 253, 193, 246, 208, 253, 193, 252, 208, 253, 193, 254, 208, 253, 193, 2, 209, 253, 193, 4, 209, 253, 193, 12, 209, 253, 193, 14, 209, 253, 193, 16, 209, 253, 193, 18, 209, 253, 193, 78, 206, 253, 193, 80, 206, 253, 193, 90, 206, 253, 193, 92, 206, 253, 193, 94, 206, 253, 193, 104, 206, 253, 193, 106, 206, 253, 193, 108, 206, 253, 193, 110, 206, 253, 193, 112, 206, 253, 193, 160, 206, 253, 193, 162, 206, 253, 193, 164, 206, 253, 193, 166, 206, 253, 193, 168, 206, 253, 193, 170, 206, 253, 193, 172, 206, 253, 193, 174, 206, 253, 193, 176, 206, 253, 193, 178, 206, 253, 193, 180, 206, 253, 193, 182, 206, 253, 193, 184, 206, 253, 193, 244, 206, 253, 193, 246, 206, 253, 193, 248, 206, 253, 193, 250, 206, 253, 193, 252, 206, 253, 193, 254, 206, 253, 193, 2, 207, 253, 193, 4, 207, 253, 193, 6, 207, 253, 193, 136, 206, 253, 193, 200, 206, 253, 193, 202, 206, 253, 193, 204, 206, 253, 193, 206, 206, 253, 193, 208, 206, 253, 193, 210, 206, 253, 193, 212, 206, 253, 193, 38, 207, 253, 193, 40, 207, 253, 193, 42, 207, 253, 193, 44, 207, 253, 193, 46, 207, 253, 193, 120, 207, 253, 193, 122, 207, 253, 193, 124, 207, 253, 193, 126, 207, 253, 193, 128, 207, 253, 193, 198, 207, 253, 193, 200, 207, 253, 193, 202, 207, 253, 193, 204, 207, 253, 193, 206, 207, 253, 193, 208, 207, 253, 193, 20, 208, 253, 193, 22, 208, 253, 193, 24, 208, 253, 193, 26, 208, 253, 193, 28, 208, 253, 193, 30, 208, 253, 193, 32, 208, 253, 193, 34, 208, 253, 193, 190, 206, 253, 193, 192, 206, 253, 193, 194, 206, 253, 193, 196, 206, 253, 193, 198, 206, 253, 193, 26, 207, 253, 193, 28, 207, 253, 193, 30, 207, 253, 193, 32, 207, 253, 193, 34, 207, 253, 193, 36, 207, 253, 193, 108, 207, 253, 193, 110, 207, 253, 193, 112, 207, 253, 193, 114, 207, 253, 193, 116, 207, 253, 193, 186, 207, 253, 193, 188, 207, 253, 193, 190, 207, 253, 193, 192, 207, 253, 193, 194, 207, 253, 193, 196, 207, 253, 193, 6, 208, 253, 193, 8, 208, 253, 193, 10, 208, 253, 193, 12, 208, 253, 193, 14, 208, 253, 193, 16, 208, 253, 193, 18, 208, 253, 193, 84, 208, 253, 193, 86, 208, 253, 193, 122, 208, 253, 193, 238, 206, 253, 193, 240, 206, 253, 193, 242, 206, 253, 193, 50, 207, 253, 193, 52, 207, 253, 193, 54, 207, 253, 193, 56, 207, 253, 193, 58, 207, 253, 193, 60, 207, 253, 193, 62, 207, 253, 193, 64, 207, 253, 193, 66, 207, 253, 193, 68, 207, 253, 193, 70, 207, 253, 193, 72, 207, 253, 193, 74, 207, 253, 193, 76, 207, 253, 193, 78, 207, 253, 193, 80, 207, 253, 193, 82, 207, 253, 193, 134, 207, 253, 193, 136, 207, 253, 193, 138, 207, 253, 193, 140, 207, 253, 193, 142, 207, 253, 193, 144, 207, 253, 193, 146, 207, 253, 193, 148, 207, 253, 193, 150, 207, 253, 193, 152, 207, 253, 193, 154, 207, 253, 193, 156, 207, 253, 193, 84, 207, 253, 193, 86, 207, 253, 193, 88, 207, 253, 193, 90, 207, 253, 193, 92, 207, 253, 193, 94, 207, 253, 193, 96, 207, 253, 193, 98, 207, 253, 193, 100, 207, 253, 193, 158, 207, 253, 193, 160, 207, 253, 193, 162, 207, 253, 193, 164, 207, 253, 193, 166, 207, 253, 193, 168, 207, 253, 193, 170, 207, 253, 193, 172, 207, 253, 193, 174, 207, 253, 193, 176, 207, 253, 193, 178, 207, 253, 193, 180, 207, 253, 193, 240, 207, 253, 193, 242, 207, 253, 193, 244, 207, 253, 193, 56, 208, 253, 193, 246, 207, 253, 193, 248, 207, 253, 193, 250, 207, 253, 193, 252, 207, 253, 193, 58, 208, 253, 193, 60, 208, 253, 193, 62, 208, 253, 193, 210, 207, 253, 193, 212, 207, 253, 193, 214, 207, 253, 193, 216, 207, 253, 193, 218, 207, 253, 193, 220, 207, 253, 193, 222, 207, 253, 193, 224, 207, 253, 193, 226, 207, 253, 193, 228, 207, 253, 193, 230, 207, 253, 193, 232, 207, 253, 193, 234, 207, 253, 193, 236, 207, 253, 193, 238, 207, 253, 193, 36, 208, 253, 193, 38, 208, 253, 193, 40, 208, 253, 193, 42, 208, 253, 193, 44, 208, 253, 193, 46, 208, 253, 193, 48, 208, 253, 193, 50, 208, 253, 193, 52, 208, 253, 193, 54, 208, 253, 193, 100, 208, 253, 193, 102, 208, 253, 193, 104, 208, 253, 193, 106, 208, 253, 193, 108, 208, 253, 193, 110, 208, 253, 193, 112, 208, 253, 193, 64, 208, 253, 193, 66, 208, 253, 193, 68, 208, 253, 193, 70, 208, 253, 193, 72, 208, 253, 193, 114, 208, 253, 193, 116, 208, 253, 193, 118, 208, 253, 193, 120, 208, 253, 193, 150, 208, 253, 193, 152, 208, 253, 193, 154, 208, 253, 193, 156, 208, 253, 193, 180, 208, 253, 193, 182, 208, 253, 193, 184, 208, 253, 193, 186, 208, 253, 193, 214, 208, 253, 193, 216, 208, 253, 193, 232, 208, 253, 193, 234, 208, 253, 193, 240, 208, 253, 193, 248, 208, 253, 193, 10, 209, 253, 193, 50, 209, 253, 193, 70, 209, 253, 193, 92, 209, 253, 193, 94, 209, 253, 193, 128, 209, 253, 193, 130, 209, 253, 193, 164, 209, 253, 193, 178, 209, 253, 193, 88, 208, 253, 193, 90, 208, 253, 193, 92, 208, 253, 193, 94, 208, 253, 193, 130, 208, 253, 193, 132, 208, 253, 193, 134, 208, 253, 193, 162, 208, 253, 193, 164, 208, 253, 193, 166, 208, 253, 193, 168, 208, 253, 193, 170, 208, 253, 193, 192, 208, 253, 193, 194, 208, 253, 193, 196, 208, 253, 193, 198, 208, 253, 193, 228, 208, 253, 193, 238, 208, 253, 193, 6, 209, 253, 193, 44, 209, 253, 193, 138, 209, 253, 193, 156, 209, 253, 193, 162, 209, 253, 193, 168, 209, 253, 193, 170, 209, 253, 193, 172, 209, 253, 193, 184, 209, 253, 193, 242, 209, 253, 193, 24, 210, 253, 193, 66, 210, 253, 193, 68, 210, 253, 193, 160, 210, 253, 193, 124, 208, 253, 193, 126, 208, 253, 193, 128, 208, 253, 193, 158, 208, 253, 193, 160, 208, 253, 193, 190, 208, 253, 193, 218, 208, 253, 193, 30, 209, 253, 193, 42, 209, 253, 193, 88, 209, 253, 193, 166, 209, 253, 193, 196, 209, 253, 193, 240, 209, 253, 193, 64, 210, 253, 193, 116, 210, 253, 193, 2, 211, 253, 193, 30, 212, 253, 193, 222, 212, 253, 193, 232, 212, 253, 193, 12, 213, 253, 193, 32, 213, 253, 193, 68, 213, 253, 193, 146, 213, 253, 193, 168, 213, 253, 193, 236, 213, 253, 193, 6, 216, 253, 193, 8, 216, 253, 193, 84, 216, 253, 193, 42, 217, 253, 193, 134, 217, 253, 193, 208, 217, 253, 193, 248, 217, 253, 193, 138, 208, 253, 193, 140, 208, 253, 193, 142, 208, 253, 193, 144, 208, 253, 193, 146, 208, 253, 193, 148, 208, 253, 193, 172, 208, 253, 193, 174, 208, 253, 193, 176, 208, 253, 193, 178, 208, 253, 193, 200, 208, 253, 193, 202, 208, 253, 193, 204, 208, 253, 193, 206, 208, 253, 193, 208, 208, 253, 193, 210, 208, 253, 193, 212, 208, 253, 193, 220, 208, 253, 193, 224, 208, 253, 193, 226, 208, 253, 193, 230, 208, 253, 193, 236, 208, 253, 193, 8, 209, 253, 193, 26, 209, 253, 193, 32, 209, 253, 193, 48, 209, 253, 193, 152, 209, 253, 193, 158, 209, 253, 193, 174, 209, 253, 193, 176, 209, 253, 193, 120, 210, 253, 193, 122, 210, 253, 193, 20, 209, 253, 193, 22, 209, 253, 193, 28, 209, 253, 193, 36, 209, 253, 193, 38, 209, 253, 193, 40, 209, 253, 193, 52, 209, 253, 193, 54, 209, 253, 193, 56, 209, 253, 193, 58, 209, 253, 193, 60, 209, 253, 193, 62, 209, 253, 193, 64, 209, 253, 193, 72, 209, 253, 193, 74, 209, 253, 193, 76, 209, 253, 193, 78, 209, 253, 193, 80, 209, 253, 193, 82, 209, 253, 193, 84, 209, 253, 193, 86, 209, 253, 193, 98, 209, 253, 193, 100, 209, 253, 193, 102, 209, 253, 193, 104, 209, 253, 193, 106, 209, 253, 193, 108, 209, 253, 193, 110, 209, 253, 193, 112, 209, 253, 193, 114, 209, 253, 193, 116, 209, 253, 193, 120, 209, 253, 193, 122, 209, 253, 193, 124, 209, 253, 193, 126, 209, 253, 193, 132, 209, 253, 193, 134, 209, 253, 193, 136, 209, 253, 193, 142, 209, 253, 193, 144, 209, 253, 193, 146, 209, 253, 193, 148, 209, 253, 193, 150, 209, 253, 193, 192, 209, 253, 193, 194, 209, 253, 193, 200, 209, 253, 193, 202, 209, 253, 193, 216, 209, 253, 193, 218, 209, 253, 193, 220, 209, 253, 193, 222, 209, 253, 193, 224, 209, 253, 193, 226, 209, 253, 193, 228, 209, 253, 193, 230, 209, 253, 193, 232, 209, 253, 193, 234, 209, 253, 193, 236, 209, 253, 193, 238, 209, 253, 206, 96, 118, 0, 206, 96, 118, 0, 206, 96, 118, 0, 206, 96, 118, 0, 206, 96, 118, 0, 193, 186, 209, 253, 193, 188, 209, 253, 193, 190, 209, 253, 193, 206, 209, 253, 193, 208, 209, 253, 193, 244, 209, 253, 193, 246, 209, 253, 193, 248, 209, 253, 193, 250, 209, 253, 193, 28, 210, 253, 193, 30, 210, 253, 193, 32, 210, 253, 193, 72, 210, 253, 193, 74, 210, 253, 193, 76, 210, 253, 193, 78, 210, 253, 193, 80, 210, 253, 193, 124, 210, 253, 193, 126, 210, 253, 193, 128, 210, 253, 193, 130, 210, 253, 193, 132, 210, 253, 193, 134, 210, 253, 193, 136, 210, 253, 193, 166, 210, 253, 193, 168, 210, 253, 193, 170, 210, 253, 193, 172, 210, 253, 193, 194, 210, 253, 193, 196, 210, 253, 193, 216, 210, 253, 193, 236, 210, 253, 193, 210, 209, 253, 193, 212, 209, 253, 193, 214, 209, 253, 193, 252, 209, 253, 193, 254, 209, 253, 193, 2, 210, 253, 193, 82, 210, 253, 193, 138, 210, 253, 193, 198, 210, 253, 193, 218, 210, 253, 193, 6, 211, 253, 193, 8, 211, 253, 193, 70, 211, 253, 193, 72, 211, 253, 193, 74, 211, 253, 193, 92, 211, 253, 193, 94, 211, 253, 193, 132, 211, 253, 193, 134, 211, 253, 193, 136, 211, 253, 193, 138, 211, 253, 193, 172, 211, 253, 193, 202, 211, 253, 193, 204, 211, 253, 193, 206, 211, 253, 193, 242, 211, 253, 193, 244, 211, 253, 193, 246, 211, 253, 193, 248, 211, 253, 193, 250, 211, 253, 193, 74, 212, 253, 193, 76, 212, 253, 206, 96, 118, 0, 206, 96, 118, 0, 206, 96, 118, 0, 206, 96, 118, 0, 206, 96, 118, 0, 193, 34, 210, 253, 193, 36, 210, 253, 193, 38, 210, 253, 193, 40, 210, 253, 193, 42, 210, 253, 193, 44, 210, 253, 193, 46, 210, 253, 193, 48, 210, 253, 193, 50, 210, 253, 193, 52, 210, 253, 193, 54, 210, 253, 193, 56, 210, 253, 193, 58, 210, 253, 193, 60, 210, 253, 193, 62, 210, 253, 206, 128, 118, 0, 206, 128, 118, 0, 206, 128, 118, 0, 206, 128, 118, 0, 206, 128, 118, 0, 206, 128, 118, 0, 206, 128, 118, 0, 206, 128, 118, 0, 206, 128, 118, 0, 206, 128, 118, 0, 206, 128, 118, 0, 206, 128, 118, 0, 193, 140, 210, 253, 193, 142, 210, 253, 193, 144, 210, 253, 193, 146, 210, 253, 193, 148, 210, 253, 193, 150, 210, 253, 193, 152, 210, 253, 193, 154, 210, 253, 193, 156, 210, 253, 193, 158, 210, 253, 193, 174, 210, 253, 193, 176, 210, 253, 193, 178, 210, 253, 193, 180, 210, 253, 193, 182, 210, 253, 193, 184, 210, 253, 193, 186, 210, 253, 193, 188, 210, 253, 193, 190, 210, 253, 193, 200, 210, 253, 193, 202, 210, 253, 193, 204, 210, 253, 193, 206, 210, 253, 193, 208, 210, 253, 193, 220, 210, 253, 193, 222, 210, 253, 193, 224, 210, 253, 193, 226, 210, 253, 193, 210, 210, 253, 193, 230, 210, 253, 193, 196, 211, 253, 193, 32, 212, 253, 193, 154, 212, 253, 193, 208, 212, 253, 193, 248, 212, 253, 193, 14, 213, 253, 193, 16, 213, 253, 193, 18, 213, 253, 193, 34, 213, 253, 193, 36, 213, 253, 193, 38, 213, 253, 193, 76, 213, 253, 193, 84, 213, 253, 193, 154, 213, 253, 193, 218, 213, 253, 193, 26, 214, 253, 193, 86, 214, 253, 193, 100, 214, 253, 193, 156, 214, 253, 193, 160, 215, 253, 193, 162, 215, 253, 193, 198, 215, 253, 193, 86, 216, 253, 193, 128, 216, 253, 193, 130, 216, 253, 193, 6, 217, 253, 193, 60, 217, 253, 193, 70, 217, 253, 193, 74, 217, 253, 193, 114, 217, 253, 193, 228, 210, 253, 193, 238, 210, 253, 193, 240, 210, 253, 193, 242, 210, 253, 193, 244, 210, 253, 193, 246, 210, 253, 193, 248, 210, 253, 193, 250, 210, 253, 193, 252, 210, 253, 193, 254, 210, 253, 193, 10, 211, 253, 193, 12, 211, 253, 193, 14, 211, 253, 193, 16, 211, 253, 193, 20, 211, 253, 193, 22, 211, 253, 193, 24, 211, 253, 193, 26, 211, 253, 193, 28, 211, 253, 193, 30, 211, 253, 193, 32, 211, 253, 193, 34, 211, 253, 193, 38, 211, 253, 193, 40, 211, 253, 193, 44, 211, 253, 193, 48, 211, 253, 193, 52, 211, 253, 193, 54, 211, 253, 193, 58, 211, 253, 193, 60, 211, 253, 193, 62, 211, 253, 193, 64, 211, 253, 193, 232, 210, 253, 193, 234, 210, 253, 193, 42, 211, 253, 193, 38, 212, 253, 193, 202, 212, 253, 193, 204, 212, 253, 193, 210, 212, 253, 193, 212, 212, 253, 193, 214, 212, 253, 193, 228, 212, 253, 193, 236, 212, 253, 193, 238, 212, 253, 193, 240, 212, 253, 193, 242, 212, 253, 193, 244, 212, 253, 193, 246, 212, 253, 193, 252, 212, 253, 193, 254, 212, 253, 193, 2, 213, 253, 193, 4, 213, 253, 193, 6, 213, 253, 193, 8, 213, 253, 193, 20, 213, 253, 193, 22, 213, 253, 193, 24, 213, 253, 193, 40, 213, 253, 193, 42, 213, 253, 193, 44, 213, 253, 193, 46, 213, 253, 193, 48, 213, 253, 193, 52, 213, 253, 193, 54, 213, 253, 193, 4, 211, 253, 193, 18, 211, 253, 193, 36, 211, 253, 193, 46, 211, 253, 193, 50, 211, 253, 193, 56, 211, 253, 193, 196, 212, 253, 193, 68, 211, 253, 193, 84, 211, 253, 193, 86, 211, 253, 193, 88, 211, 253, 193, 90, 211, 253, 193, 126, 211, 253, 193, 128, 211, 253, 193, 130, 211, 253, 193, 168, 211, 253, 193, 170, 211, 253, 193, 198, 211, 253, 193, 200, 211, 253, 193, 238, 211, 253, 193, 240, 211, 253, 193, 40, 212, 253, 193, 100, 212, 253, 193, 190, 212, 253, 193, 216, 212, 253, 193, 26, 213, 253, 193, 50, 213, 253, 193, 94, 213, 253, 193, 98, 213, 253, 193, 100, 213, 253, 193, 108, 213, 253, 193, 148, 213, 253, 193, 162, 211, 253, 193, 164, 211, 253, 193, 174, 211, 253, 193, 176, 211, 253, 193, 178, 211, 253, 193, 180, 211, 253, 193, 182, 211, 253, 193, 184, 211, 253, 193, 186, 211, 253, 193, 188, 211, 253, 193, 190, 211, 253, 193, 192, 211, 253, 193, 194, 211, 253, 193, 208, 211, 253, 193, 210, 211, 253, 193, 212, 211, 253, 193, 214, 211, 253, 193, 216, 211, 253, 193, 218, 211, 253, 193, 220, 211, 253, 193, 222, 211, 253, 193, 224, 211, 253, 193, 226, 211, 253, 193, 228, 211, 253, 193, 230, 211, 253, 193, 232, 211, 253, 193, 234, 211, 253, 193, 236, 211, 253, 206, 160, 118, 0, 206, 160, 118, 0, 206, 160, 118, 0, 206, 160, 118, 0, 206, 160, 118, 0, 206, 160, 118, 0, 206, 160, 118, 0, 206, 160, 118, 0, 206, 160, 118, 0, 206, 160, 118, 0, 206, 160, 118, 0, 206, 160, 118, 0, 193, 42, 212, 253, 193, 44, 212, 253, 193, 46, 212, 253, 193, 48, 212, 253, 193, 50, 212, 253, 193, 52, 212, 253, 193, 54, 212, 253, 193, 56, 212, 253, 193, 58, 212, 253, 193, 60, 212, 253, 193, 62, 212, 253, 193, 64, 212, 253, 193, 66, 212, 253, 193, 68, 212, 253, 193, 70, 212, 253, 193, 80, 212, 253, 193, 82, 212, 253, 193, 84, 212, 253, 193, 86, 212, 253, 193, 88, 212, 253, 193, 78, 212, 253, 193, 102, 212, 253, 193, 104, 212, 253, 193, 140, 212, 253, 193, 156, 212, 253, 193, 106, 212, 253, 193, 172, 212, 253, 193, 110, 213, 253, 193, 174, 213, 253, 193, 176, 213, 253, 193, 224, 213, 253, 193, 254, 213, 253, 193, 16, 214, 253, 193, 30, 214, 253, 193, 60, 214, 253, 193, 88, 214, 253, 193, 102, 214, 253, 193, 134, 214, 253, 193, 140, 214, 253, 193, 142, 214, 253, 193, 186, 214, 253, 193, 188, 214, 253, 193, 190, 214, 253, 193, 204, 214, 253, 193, 206, 214, 253, 193, 208, 214, 253, 193, 220, 214, 253, 193, 224, 214, 253, 193, 236, 214, 253, 193, 4, 215, 253, 193, 8, 215, 253, 193, 20, 215, 253, 193, 90, 212, 253, 193, 92, 212, 253, 193, 94, 212, 253, 193, 108, 212, 253, 193, 110, 212, 253, 193, 112, 212, 253, 193, 114, 212, 253, 193, 116, 212, 253, 193, 118, 212, 253, 193, 120, 212, 253, 193, 122, 212, 253, 193, 124, 212, 253, 193, 126, 212, 253, 193, 128, 212, 253, 193, 130, 212, 253, 193, 132, 212, 253, 193, 134, 212, 253, 193, 136, 212, 253, 193, 142, 212, 253, 193, 144, 212, 253, 193, 146, 212, 253, 193, 148, 212, 253, 193, 150, 212, 253, 193, 152, 212, 253, 193, 158, 212, 253, 193, 160, 212, 253, 193, 162, 212, 253, 193, 164, 212, 253, 193, 166, 212, 253, 193, 168, 212, 253, 193, 170, 212, 253, 193, 174, 212, 253, 193, 98, 212, 253, 193, 138, 212, 253, 193, 188, 212, 253, 193, 200, 212, 253, 193, 224, 212, 253, 193, 226, 212, 253, 193, 234, 212, 253, 193, 250, 212, 253, 193, 124, 213, 253, 193, 126, 213, 253, 193, 128, 213, 253, 193, 156, 213, 253, 193, 196, 213, 253, 193, 204, 213, 253, 193, 208, 213, 253, 193, 246, 213, 253, 193, 126, 214, 253, 193, 128, 214, 253, 193, 182, 214, 253, 193, 240, 214, 253, 193, 80, 215, 253, 193, 82, 215, 253, 193, 124, 215, 253, 193, 164, 215, 253, 193, 166, 215, 253, 193, 10, 216, 253, 193, 12, 216, 253, 193, 14, 216, 253, 193, 88, 216, 253, 193, 158, 216, 253, 193, 244, 216, 253, 193, 8, 217, 253, 193, 176, 212, 253, 193, 178, 212, 253, 193, 180, 212, 253, 193, 182, 212, 253, 193, 184, 212, 253, 193, 186, 212, 253, 193, 192, 212, 253, 193, 194, 212, 253, 193, 206, 212, 253, 193, 218, 212, 253, 193, 220, 212, 253, 193, 230, 212, 253, 193, 10, 213, 253, 193, 198, 212, 253, 193, 28, 213, 253, 193, 30, 213, 253, 193, 74, 213, 253, 193, 96, 213, 253, 193, 102, 213, 253, 193, 104, 213, 253, 193, 106, 213, 253, 193, 112, 213, 253, 193, 114, 213, 253, 193, 116, 213, 253, 193, 118, 213, 253, 193, 120, 213, 253, 193, 122, 213, 253, 193, 130, 213, 253, 193, 132, 213, 253, 193, 134, 213, 253, 193, 136, 213, 253, 193, 138, 213, 253, 193, 56, 213, 253, 193, 58, 213, 253, 193, 60, 213, 253, 193, 62, 213, 253, 193, 64, 213, 253, 193, 66, 213, 253, 193, 70, 213, 253, 193, 72, 213, 253, 193, 78, 213, 253, 193, 80, 213, 253, 193, 82, 213, 253, 193, 86, 213, 253, 193, 88, 213, 253, 193, 90, 213, 253, 193, 92, 213, 253, 193, 158, 213, 253, 193, 170, 213, 253, 193, 182, 213, 253, 193, 248, 213, 253, 193, 28, 214, 253, 193, 58, 214, 253, 193, 202, 214, 253, 193, 126, 215, 253, 193, 200, 215, 253, 193, 202, 215, 253, 193, 160, 216, 253, 193, 198, 216, 253, 193, 154, 217, 253, 193, 156, 217, 253, 193, 186, 217, 253, 193, 228, 217, 253, 193, 230, 217, 253, 193, 140, 213, 253, 193, 142, 213, 253, 193, 144, 213, 253, 193, 150, 213, 253, 193, 152, 213, 253, 193, 162, 213, 253, 193, 164, 213, 253, 193, 166, 213, 253, 193, 178, 213, 253, 193, 180, 213, 253, 193, 184, 213, 253, 193, 186, 213, 253, 193, 188, 213, 253, 193, 190, 213, 253, 193, 192, 213, 253, 193, 194, 213, 253, 193, 200, 213, 253, 193, 202, 213, 253, 193, 206, 213, 253, 193, 214, 213, 253, 193, 216, 213, 253, 193, 226, 213, 253, 193, 228, 213, 253, 193, 230, 213, 253, 193, 232, 213, 253, 193, 234, 213, 253, 193, 244, 213, 253, 193, 2, 214, 253, 193, 4, 214, 253, 193, 6, 214, 253, 193, 8, 214, 253, 193, 10, 214, 253, 193, 12, 214, 253, 193, 14, 214, 253, 193, 18, 214, 253, 193, 20, 214, 253, 193, 22, 214, 253, 193, 24, 214, 253, 193, 32, 214, 253, 193, 34, 214, 253, 193, 36, 214, 253, 193, 38, 214, 253, 193, 40, 214, 253, 193, 42, 214, 253, 193, 44, 214, 253, 193, 46, 214, 253, 193, 48, 214, 253, 193, 50, 214, 253, 193, 52, 214, 253, 193, 54, 214, 253, 193, 56, 214, 253, 193, 62, 214, 253, 193, 64, 214, 253, 193, 66, 214, 253, 193, 68, 214, 253, 193, 70, 214, 253, 193, 72, 214, 253, 193, 74, 214, 253, 193, 76, 214, 253, 193, 78, 214, 253, 193, 80, 214, 253, 193, 82, 214, 253, 193, 84, 214, 253, 193, 90, 214, 253, 193, 92, 214, 253, 193, 94, 214, 253, 193, 96, 214, 253, 193, 98, 214, 253, 193, 104, 214, 253, 193, 106, 214, 253, 193, 108, 214, 253, 193, 112, 214, 253, 193, 114, 214, 253, 193, 116, 214, 253, 193, 118, 214, 253, 193, 120, 214, 253, 193, 122, 214, 253, 193, 130, 214, 253, 193, 136, 214, 253, 193, 138, 214, 253, 193, 144, 214, 253, 193, 146, 214, 253, 193, 148, 214, 253, 193, 150, 214, 253, 193, 152, 214, 253, 193, 154, 214, 253, 193, 160, 214, 253, 193, 162, 214, 253, 193, 164, 214, 253, 193, 166, 214, 253, 193, 168, 214, 253, 193, 170, 214, 253, 193, 172, 214, 253, 193, 174, 214, 253, 193, 176, 214, 253, 193, 178, 214, 253, 193, 180, 214, 253, 193, 184, 214, 253, 193, 192, 214, 253, 193, 194, 214, 253, 193, 196, 214, 253, 193, 198, 214, 253, 193, 200, 214, 253, 193, 210, 214, 253, 193, 212, 214, 253, 193, 214, 214, 253, 193, 216, 214, 253, 193, 222, 214, 253, 193, 226, 214, 253, 193, 228, 214, 253, 193, 230, 214, 253, 193, 232, 214, 253, 193, 234, 214, 253, 193, 242, 214, 253, 193, 238, 214, 253, 193, 244, 214, 253, 193, 246, 214, 253, 193, 248, 214, 253, 193, 250, 214, 253, 193, 2, 215, 253, 193, 6, 215, 253, 193, 10, 215, 253, 193, 12, 215, 253, 193, 14, 215, 253, 206, 192, 118, 0, 206, 192, 118, 0, 206, 192, 118, 0, 206, 192, 118, 0, 193, 22, 215, 253, 193, 50, 215, 253, 193, 94, 215, 253, 193, 96, 215, 253, 193, 134, 215, 253, 193, 136, 215, 253, 193, 138, 215, 253, 193, 140, 215, 253, 193, 172, 215, 253, 193, 174, 215, 253, 193, 218, 215, 253, 193, 220, 215, 253, 193, 222, 215, 253, 193, 26, 216, 253, 193, 28, 216, 253, 193, 30, 216, 253, 193, 138, 216, 253, 193, 162, 216, 253, 193, 202, 216, 253, 193, 204, 216, 253, 193, 206, 216, 253, 193, 232, 216, 253, 193, 98, 217, 253, 193, 120, 217, 253, 193, 166, 217, 253, 193, 196, 217, 253, 193, 20, 218, 253, 193, 52, 218, 253, 193, 54, 218, 253, 193, 88, 218, 253, 193, 90, 218, 253, 193, 92, 218, 253, 206, 192, 118, 0, 206, 192, 118, 0, 206, 192, 118, 0, 206, 192, 118, 0, 193, 52, 215, 253, 193, 54, 215, 253, 193, 56, 215, 253, 193, 58, 215, 253, 193, 60, 215, 253, 193, 62, 215, 253, 193, 64, 215, 253, 193, 66, 215, 253, 193, 68, 215, 253, 193, 70, 215, 253, 193, 72, 215, 253, 193, 74, 215, 253, 193, 76, 215, 253, 193, 78, 215, 253, 193, 98, 215, 253, 193, 100, 215, 253, 193, 102, 215, 253, 193, 104, 215, 253, 193, 106, 215, 253, 193, 108, 215, 253, 193, 110, 215, 253, 193, 112, 215, 253, 193, 114, 215, 253, 193, 116, 215, 253, 193, 118, 215, 253, 193, 120, 215, 253, 193, 122, 215, 253, 193, 142, 215, 253, 193, 130, 215, 253, 193, 132, 215, 253, 193, 168, 215, 253, 193, 170, 215, 253, 193, 204, 215, 253, 193, 206, 215, 253, 193, 208, 215, 253, 193, 210, 215, 253, 193, 212, 215, 253, 193, 214, 215, 253, 193, 216, 215, 253, 193, 16, 216, 253, 193, 18, 216, 253, 193, 20, 216, 253, 193, 22, 216, 253, 193, 24, 216, 253, 193, 90, 216, 253, 193, 92, 216, 253, 193, 132, 216, 253, 193, 134, 216, 253, 193, 136, 216, 253, 193, 200, 216, 253, 193, 230, 216, 253, 193, 246, 216, 253, 193, 10, 217, 253, 193, 36, 217, 253, 193, 40, 217, 253, 193, 44, 217, 253, 193, 48, 217, 253, 193, 62, 217, 253, 193, 222, 217, 253, 193, 76, 217, 253, 193, 144, 215, 253, 193, 146, 215, 253, 193, 148, 215, 253, 193, 150, 215, 253, 193, 152, 215, 253, 193, 154, 215, 253, 193, 156, 215, 253, 193, 158, 215, 253, 193, 178, 215, 253, 193, 180, 215, 253, 193, 182, 215, 253, 193, 184, 215, 253, 193, 186, 215, 253, 193, 188, 215, 253, 193, 190, 215, 253, 193, 192, 215, 253, 193, 194, 215, 253, 193, 196, 215, 253, 193, 224, 215, 253, 193, 226, 215, 253, 193, 228, 215, 253, 193, 230, 215, 253, 193, 232, 215, 253, 193, 234, 215, 253, 193, 236, 215, 253, 193, 238, 215, 253, 193, 240, 215, 253, 193, 242, 215, 253, 193, 244, 215, 253, 193, 246, 215, 253, 193, 248, 215, 253, 193, 250, 215, 253, 193, 252, 215, 253, 193, 254, 215, 253, 193, 2, 216, 253, 193, 32, 216, 253, 193, 4, 216, 253, 193, 34, 216, 253, 193, 36, 216, 253, 193, 38, 216, 253, 193, 40, 216, 253, 193, 42, 216, 253, 193, 44, 216, 253, 193, 46, 216, 253, 193, 48, 216, 253, 193, 50, 216, 253, 193, 52, 216, 253, 193, 54, 216, 253, 193, 56, 216, 253, 193, 58, 216, 253, 193, 60, 216, 253, 193, 62, 216, 253, 193, 64, 216, 253, 193, 66, 216, 253, 193, 68, 216, 253, 193, 70, 216, 253, 193, 72, 216, 253, 193, 74, 216, 253, 193, 76, 216, 253, 193, 78, 216, 253, 193, 80, 216, 253, 193, 82, 216, 253, 193, 94, 216, 253, 193, 96, 216, 253, 193, 98, 216, 253, 193, 100, 216, 253, 193, 102, 216, 253, 193, 104, 216, 253, 193, 106, 216, 253, 193, 108, 216, 253, 193, 110, 216, 253, 193, 112, 216, 253, 193, 114, 216, 253, 193, 116, 216, 253, 193, 118, 216, 253, 193, 120, 216, 253, 193, 122, 216, 253, 193, 124, 216, 253, 193, 126, 216, 253, 193, 140, 216, 253, 193, 142, 216, 253, 193, 144, 216, 253, 193, 146, 216, 253, 193, 148, 216, 253, 193, 150, 216, 253, 193, 152, 216, 253, 193, 154, 216, 253, 193, 156, 216, 253, 206, 224, 118, 0, 206, 224, 118, 0, 206, 224, 118, 0, 206, 224, 118, 0, 206, 224, 118, 0, 206, 224, 118, 0, 206, 224, 118, 0, 206, 224, 118, 0, 206, 224, 118, 0, 193, 208, 216, 253, 193, 210, 216, 253, 193, 212, 216, 253, 193, 214, 216, 253, 193, 216, 216, 253, 193, 218, 216, 253, 193, 220, 216, 253, 193, 222, 216, 253, 193, 224, 216, 253, 193, 226, 216, 253, 193, 228, 216, 253, 193, 234, 216, 253, 193, 236, 216, 253, 193, 238, 216, 253, 193, 240, 216, 253, 193, 242, 216, 253, 193, 248, 216, 253, 193, 250, 216, 253, 193, 252, 216, 253, 193, 254, 216, 253, 193, 2, 217, 253, 193, 4, 217, 253, 193, 12, 217, 253, 193, 14, 217, 253, 193, 16, 217, 253, 193, 18, 217, 253, 193, 20, 217, 253, 193, 22, 217, 253, 193, 24, 217, 253, 193, 26, 217, 253, 193, 28, 217, 253, 193, 30, 217, 253, 193, 32, 217, 253, 193, 34, 217, 253, 193, 46, 217, 253, 193, 50, 217, 253, 193, 52, 217, 253, 193, 54, 217, 253, 193, 56, 217, 253, 193, 58, 217, 253, 193, 64, 217, 253, 193, 66, 217, 253, 193, 68, 217, 253, 193, 78, 217, 253, 193, 80, 217, 253, 193, 82, 217, 253, 193, 84, 217, 253, 193, 86, 217, 253, 193, 88, 217, 253, 193, 100, 217, 253, 193, 102, 217, 253, 193, 104, 217, 253, 193, 106, 217, 253, 193, 108, 217, 253, 193, 110, 217, 253, 193, 38, 217, 253, 193, 72, 217, 253, 193, 90, 217, 253, 193, 136, 217, 253, 193, 152, 217, 253, 193, 32, 218, 253, 193, 68, 218, 253, 193, 102, 218, 253, 193, 126, 218, 253, 193, 140, 218, 253, 193, 204, 218, 253, 193, 206, 218, 253, 193, 18, 219, 253, 193, 64, 219, 253, 193, 136, 219, 253, 193, 222, 219, 253, 193, 242, 219, 253, 193, 244, 219, 253, 193, 246, 219, 253, 193, 248, 219, 253, 193, 250, 219, 253, 193, 252, 219, 253, 193, 22, 220, 253, 193, 78, 220, 253, 193, 152, 220, 253, 193, 170, 220, 253, 193, 182, 220, 253, 193, 184, 220, 253, 193, 186, 220, 253, 193, 252, 220, 253, 193, 12, 221, 253, 193, 24, 221, 253, 193, 92, 217, 253, 193, 94, 217, 253, 193, 96, 217, 253, 193, 116, 217, 253, 193, 118, 217, 253, 193, 158, 217, 253, 193, 188, 217, 253, 193, 190, 217, 253, 193, 200, 217, 253, 193, 202, 217, 253, 193, 210, 217, 253, 193, 224, 217, 253, 193, 226, 217, 253, 193, 2, 218, 253, 193, 6, 218, 253, 193, 10, 218, 253, 193, 12, 218, 253, 193, 14, 218, 253, 193, 18, 218, 253, 193, 62, 218, 253, 193, 64, 218, 253, 193, 84, 218, 253, 193, 86, 218, 253, 193, 106, 218, 253, 193, 142, 218, 253, 193, 144, 218, 253, 193, 146, 218, 253, 193, 148, 218, 253, 193, 176, 218, 253, 193, 180, 218, 253, 193, 182, 218, 253, 193, 194, 218, 253, 193, 112, 217, 253, 193, 122, 217, 253, 193, 124, 217, 253, 193, 126, 217, 253, 193, 128, 217, 253, 193, 130, 217, 253, 193, 132, 217, 253, 193, 138, 217, 253, 193, 140, 217, 253, 193, 142, 217, 253, 193, 144, 217, 253, 193, 146, 217, 253, 193, 160, 217, 253, 193, 162, 217, 253, 193, 168, 217, 253, 193, 170, 217, 253, 193, 172, 217, 253, 193, 174, 217, 253, 193, 176, 217, 253, 193, 178, 217, 253, 193, 180, 217, 253, 193, 182, 217, 253, 193, 184, 217, 253, 193, 192, 217, 253, 193, 194, 217, 253, 193, 198, 217, 253, 193, 204, 217, 253, 193, 196, 224, 253, 193, 212, 217, 253, 193, 206, 217, 253, 193, 214, 217, 253, 193, 216, 217, 253, 193, 148, 217, 253, 193, 150, 217, 253, 193, 164, 217, 253, 193, 238, 217, 253, 193, 242, 217, 253, 193, 30, 218, 253, 193, 36, 218, 253, 193, 38, 218, 253, 193, 40, 218, 253, 193, 44, 218, 253, 193, 50, 218, 253, 193, 56, 218, 253, 193, 60, 218, 253, 193, 62, 219, 253, 193, 134, 219, 253, 193, 76, 220, 253, 193, 108, 220, 253, 193, 180, 220, 253, 193, 212, 220, 253, 193, 238, 220, 253, 193, 134, 221, 253, 193, 206, 221, 253, 193, 208, 221, 253, 193, 210, 221, 253, 193, 12, 223, 253, 193, 72, 223, 253, 193, 122, 223, 253, 193, 124, 223, 253, 193, 126, 223, 253, 193, 128, 223, 253, 193, 142, 223, 253, 193, 144, 223, 253, 193, 218, 217, 253, 193, 220, 217, 253, 193, 8, 218, 253, 193, 16, 218, 253, 193, 22, 218, 253, 193, 24, 218, 253, 193, 26, 218, 253, 193, 34, 218, 253, 193, 42, 218, 253, 193, 46, 218, 253, 193, 58, 218, 253, 193, 70, 218, 253, 193, 72, 218, 253, 193, 76, 218, 253, 193, 78, 218, 253, 193, 80, 218, 253, 193, 94, 218, 253, 193, 96, 218, 253, 193, 98, 218, 253, 193, 100, 218, 253, 193, 110, 218, 253, 193, 112, 218, 253, 193, 114, 218, 253, 193, 116, 218, 253, 193, 118, 218, 253, 193, 120, 218, 253, 193, 122, 218, 253, 193, 124, 218, 253, 193, 128, 218, 253, 193, 130, 218, 253, 193, 132, 218, 253, 193, 134, 218, 253, 193, 232, 217, 253, 193, 234, 217, 253, 193, 236, 217, 253, 193, 240, 217, 253, 193, 244, 217, 253, 193, 246, 217, 253, 193, 250, 217, 253, 193, 252, 217, 253, 193, 254, 217, 253, 193, 4, 218, 253, 193, 82, 218, 253, 193, 104, 218, 253, 193, 192, 218, 253, 193, 238, 218, 253, 193, 88, 219, 253, 193, 90, 219, 253, 193, 224, 219, 253, 193, 10, 220, 253, 193, 12, 220, 253, 193, 40, 220, 253, 193, 62, 220, 253, 193, 80, 220, 253, 193, 82, 220, 253, 193, 240, 220, 253, 193, 242, 220, 253, 193, 244, 220, 253, 193, 62, 221, 253, 193, 64, 221, 253, 193, 66, 221, 253, 193, 140, 221, 253, 193, 24, 222, 253, 193, 146, 222, 253, 193, 28, 218, 253, 193, 48, 218, 253, 193, 224, 218, 253, 193, 220, 219, 253, 193, 38, 220, 253, 193, 204, 221, 253, 193, 20, 222, 253, 193, 10, 223, 253, 193, 120, 223, 253, 193, 166, 223, 253, 193, 168, 223, 253, 193, 192, 223, 253, 193, 204, 223, 253, 193, 206, 223, 253, 193, 230, 223, 253, 193, 238, 223, 253, 193, 2, 224, 253, 193, 74, 224, 253, 193, 114, 224, 253, 193, 230, 224, 253, 193, 30, 225, 253, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 108, 218, 253, 193, 150, 218, 253, 193, 230, 218, 253, 193, 246, 218, 253, 193, 40, 219, 253, 193, 68, 219, 253, 193, 126, 219, 253, 193, 146, 219, 253, 193, 176, 219, 253, 193, 178, 219, 253, 193, 24, 220, 253, 193, 26, 220, 253, 193, 46, 220, 253, 193, 64, 220, 253, 193, 66, 220, 253, 193, 84, 220, 253, 193, 156, 220, 253, 193, 164, 220, 253, 193, 192, 220, 253, 193, 194, 220, 253, 193, 16, 221, 253, 193, 32, 221, 253, 193, 34, 221, 253, 193, 90, 221, 253, 193, 92, 221, 253, 193, 94, 221, 253, 193, 154, 221, 253, 193, 156, 221, 253, 193, 224, 221, 253, 193, 226, 221, 253, 193, 36, 222, 253, 193, 38, 222, 253, 193, 136, 218, 253, 193, 138, 218, 253, 193, 152, 218, 253, 193, 154, 218, 253, 193, 156, 218, 253, 193, 158, 218, 253, 193, 160, 218, 253, 193, 162, 218, 253, 193, 164, 218, 253, 193, 166, 218, 253, 193, 168, 218, 253, 193, 170, 218, 253, 193, 172, 218, 253, 193, 174, 218, 253, 193, 178, 218, 253, 193, 184, 218, 253, 193, 186, 218, 253, 193, 188, 218, 253, 193, 190, 218, 253, 193, 200, 218, 253, 193, 202, 218, 253, 193, 210, 218, 253, 193, 212, 218, 253, 193, 216, 218, 253, 193, 226, 218, 253, 193, 228, 218, 253, 193, 232, 218, 253, 193, 234, 218, 253, 193, 236, 218, 253, 193, 248, 218, 253, 193, 250, 218, 253, 193, 252, 218, 253, 193, 220, 218, 253, 193, 222, 218, 253, 193, 240, 218, 253, 193, 242, 218, 253, 193, 244, 218, 253, 193, 20, 219, 253, 193, 22, 219, 253, 193, 24, 219, 253, 193, 26, 219, 253, 193, 28, 219, 253, 193, 30, 219, 253, 193, 32, 219, 253, 193, 34, 219, 253, 193, 66, 219, 253, 193, 92, 219, 253, 193, 94, 219, 253, 193, 96, 219, 253, 193, 124, 219, 253, 193, 138, 219, 253, 193, 140, 219, 253, 193, 142, 219, 253, 193, 144, 219, 253, 193, 168, 219, 253, 193, 170, 219, 253, 193, 172, 219, 253, 193, 174, 219, 253, 193, 230, 219, 253, 193, 240, 219, 253, 193, 2, 220, 253, 193, 4, 220, 253, 193, 14, 220, 253, 193, 16, 220, 253, 193, 254, 218, 253, 193, 2, 219, 253, 193, 4, 219, 253, 193, 6, 219, 253, 193, 8, 219, 253, 193, 10, 219, 253, 193, 12, 219, 253, 193, 14, 219, 253, 193, 16, 219, 253, 193, 42, 219, 253, 193, 44, 219, 253, 193, 46, 219, 253, 193, 48, 219, 253, 193, 50, 219, 253, 193, 52, 219, 253, 193, 54, 219, 253, 193, 56, 219, 253, 193, 58, 219, 253, 193, 70, 219, 253, 193, 72, 219, 253, 193, 74, 219, 253, 193, 76, 219, 253, 193, 78, 219, 253, 193, 80, 219, 253, 193, 82, 219, 253, 193, 84, 219, 253, 193, 98, 219, 253, 193, 100, 219, 253, 193, 102, 219, 253, 193, 104, 219, 253, 193, 106, 219, 253, 193, 86, 219, 253, 193, 108, 219, 253, 193, 110, 219, 253, 193, 112, 219, 253, 193, 114, 219, 253, 193, 116, 219, 253, 193, 118, 219, 253, 193, 120, 219, 253, 193, 122, 219, 253, 193, 128, 219, 253, 193, 130, 219, 253, 193, 132, 219, 253, 193, 148, 219, 253, 193, 150, 219, 253, 193, 152, 219, 253, 193, 154, 219, 253, 193, 156, 219, 253, 193, 158, 219, 253, 193, 160, 219, 253, 193, 162, 219, 253, 193, 164, 219, 253, 193, 166, 219, 253, 193, 180, 219, 253, 193, 182, 219, 253, 193, 184, 219, 253, 193, 186, 219, 253, 193, 188, 219, 253, 193, 190, 219, 253, 193, 192, 219, 253, 193, 194, 219, 253, 193, 196, 219, 253, 193, 198, 219, 253, 193, 126, 220, 253, 193, 200, 219, 253, 193, 202, 219, 253, 193, 204, 219, 253, 193, 206, 219, 253, 193, 208, 219, 253, 193, 210, 219, 253, 193, 212, 219, 253, 193, 214, 219, 253, 193, 216, 219, 253, 193, 218, 219, 253, 193, 226, 219, 253, 193, 228, 219, 253, 193, 232, 219, 253, 193, 234, 219, 253, 193, 236, 219, 253, 193, 238, 219, 253, 193, 254, 219, 253, 193, 6, 220, 253, 193, 8, 220, 253, 193, 18, 220, 253, 193, 20, 220, 253, 193, 28, 220, 253, 193, 30, 220, 253, 193, 32, 220, 253, 193, 34, 220, 253, 193, 36, 220, 253, 193, 48, 220, 253, 193, 50, 220, 253, 193, 52, 220, 253, 193, 54, 220, 253, 193, 56, 220, 253, 193, 58, 220, 253, 193, 42, 220, 253, 193, 44, 220, 253, 193, 102, 220, 253, 193, 110, 220, 253, 193, 112, 220, 253, 193, 122, 220, 253, 193, 124, 220, 253, 193, 134, 220, 253, 193, 154, 220, 253, 193, 174, 220, 253, 193, 176, 220, 253, 193, 190, 220, 253, 193, 198, 220, 253, 193, 206, 220, 253, 193, 218, 220, 253, 193, 224, 220, 253, 193, 236, 220, 253, 193, 246, 220, 253, 193, 248, 220, 253, 193, 254, 220, 253, 193, 2, 221, 253, 193, 14, 221, 253, 193, 26, 221, 253, 193, 28, 221, 253, 193, 30, 221, 253, 206, 32, 119, 0, 206, 32, 119, 0, 206, 32, 119, 0, 206, 32, 119, 0, 206, 32, 119, 0, 206, 32, 119, 0, 206, 32, 119, 0, 193, 60, 220, 253, 193, 68, 220, 253, 193, 70, 220, 253, 193, 72, 220, 253, 193, 74, 220, 253, 193, 86, 220, 253, 193, 88, 220, 253, 193, 90, 220, 253, 193, 92, 220, 253, 193, 94, 220, 253, 193, 96, 220, 253, 193, 98, 220, 253, 193, 100, 220, 253, 193, 104, 220, 253, 193, 106, 220, 253, 193, 114, 220, 253, 193, 116, 220, 253, 193, 118, 220, 253, 193, 120, 220, 253, 193, 128, 220, 253, 193, 130, 220, 253, 193, 132, 220, 253, 193, 136, 220, 253, 193, 138, 220, 253, 193, 140, 220, 253, 193, 142, 220, 253, 193, 144, 220, 253, 193, 146, 220, 253, 193, 148, 220, 253, 193, 150, 220, 253, 193, 158, 220, 253, 193, 160, 220, 253, 193, 162, 220, 253, 193, 166, 220, 253, 193, 168, 220, 253, 193, 172, 220, 253, 193, 178, 220, 253, 193, 188, 220, 253, 193, 196, 220, 253, 193, 202, 220, 253, 193, 204, 220, 253, 193, 208, 220, 253, 193, 210, 220, 253, 193, 214, 220, 253, 193, 216, 220, 253, 193, 222, 220, 253, 193, 220, 220, 253, 193, 226, 220, 253, 193, 228, 220, 253, 193, 230, 220, 253, 193, 232, 220, 253, 193, 234, 220, 253, 193, 4, 221, 253, 193, 6, 221, 253, 193, 8, 221, 253, 193, 10, 221, 253, 193, 18, 221, 253, 193, 20, 221, 253, 193, 22, 221, 253, 206, 0, 119, 0, 206, 0, 119, 0, 206, 0, 119, 0, 206, 0, 119, 0, 206, 0, 119, 0, 206, 0, 119, 0, 193, 98, 221, 253, 193, 100, 221, 253, 193, 102, 221, 253, 193, 104, 221, 253, 193, 106, 221, 253, 193, 108, 221, 253, 193, 110, 221, 253, 193, 112, 221, 253, 193, 114, 221, 253, 193, 116, 221, 253, 193, 118, 221, 253, 193, 120, 221, 253, 193, 122, 221, 253, 193, 124, 221, 253, 193, 126, 221, 253, 193, 128, 221, 253, 193, 130, 221, 253, 193, 132, 221, 253, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 193, 58, 221, 253, 193, 60, 221, 253, 193, 136, 221, 253, 193, 138, 221, 253, 193, 212, 221, 253, 193, 214, 221, 253, 193, 216, 221, 253, 193, 22, 222, 253, 193, 80, 222, 253, 193, 82, 222, 253, 193, 84, 222, 253, 193, 138, 222, 253, 193, 140, 222, 253, 193, 142, 222, 253, 193, 144, 222, 253, 193, 198, 222, 253, 193, 200, 222, 253, 193, 228, 222, 253, 193, 230, 222, 253, 193, 232, 222, 253, 193, 14, 223, 253, 193, 16, 223, 253, 193, 18, 223, 253, 193, 40, 223, 253, 193, 84, 223, 253, 193, 148, 223, 253, 193, 66, 224, 253, 193, 124, 224, 253, 193, 126, 224, 253, 193, 128, 224, 253, 193, 130, 224, 253, 193, 132, 224, 253, 206, 32, 119, 0, 206, 32, 119, 0, 206, 32, 119, 0, 206, 32, 119, 0, 193, 142, 221, 253, 193, 144, 221, 253, 193, 146, 221, 253, 193, 148, 221, 253, 193, 150, 221, 253, 193, 152, 221, 253, 193, 218, 221, 253, 193, 220, 221, 253, 193, 222, 221, 253, 193, 26, 222, 253, 193, 28, 222, 253, 193, 30, 222, 253, 193, 32, 222, 253, 193, 34, 222, 253, 193, 86, 222, 253, 193, 88, 222, 253, 193, 90, 222, 253, 193, 92, 222, 253, 193, 94, 222, 253, 193, 96, 222, 253, 193, 152, 222, 253, 193, 154, 222, 253, 193, 156, 222, 253, 193, 158, 222, 253, 193, 22, 223, 253, 193, 24, 223, 253, 193, 96, 223, 253, 193, 110, 223, 253, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 64, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 206, 96, 119, 0, 193, 40, 222, 253, 193, 42, 222, 253, 193, 44, 222, 253, 193, 46, 222, 253, 193, 48, 222, 253, 193, 50, 222, 253, 193, 52, 222, 253, 193, 54, 222, 253, 193, 56, 222, 253, 193, 58, 222, 253, 193, 60, 222, 253, 193, 62, 222, 253, 193, 64, 222, 253, 193, 66, 222, 253, 193, 68, 222, 253, 193, 70, 222, 253, 193, 72, 222, 253, 193, 74, 222, 253, 193, 76, 222, 253, 193, 78, 222, 253, 206, 128, 119, 0, 206, 128, 119, 0, 206, 128, 119, 0, 206, 128, 119, 0, 206, 128, 119, 0, 206, 128, 119, 0, 206, 128, 119, 0, 206, 128, 119, 0, 206, 128, 119, 0, 206, 128, 119, 0, 206, 128, 119, 0, 206, 128, 119, 0, 193, 164, 222, 253, 193, 166, 222, 253, 193, 168, 222, 253, 193, 170, 222, 253, 193, 172, 222, 253, 193, 174, 222, 253, 193, 176, 222, 253, 193, 178, 222, 253, 193, 180, 222, 253, 193, 182, 222, 253, 193, 184, 222, 253, 193, 186, 222, 253, 193, 188, 222, 253, 193, 190, 222, 253, 193, 192, 222, 253, 193, 194, 222, 253, 193, 196, 222, 253, 193, 210, 222, 253, 193, 212, 222, 253, 193, 214, 222, 253, 193, 216, 222, 253, 193, 218, 222, 253, 193, 220, 222, 253, 193, 148, 222, 253, 193, 150, 222, 253, 193, 202, 222, 253, 193, 204, 222, 253, 193, 234, 222, 253, 193, 20, 223, 253, 193, 42, 223, 253, 193, 50, 223, 253, 193, 74, 223, 253, 193, 116, 223, 253, 193, 130, 223, 253, 193, 150, 223, 253, 193, 152, 223, 253, 193, 154, 223, 253, 193, 156, 223, 253, 193, 158, 223, 253, 193, 180, 223, 253, 193, 182, 223, 253, 193, 184, 223, 253, 193, 186, 223, 253, 193, 198, 223, 253, 193, 200, 223, 253, 193, 220, 223, 253, 193, 222, 223, 253, 193, 224, 223, 253, 193, 236, 223, 253, 193, 246, 223, 253, 193, 248, 223, 253, 193, 250, 223, 253, 193, 252, 223, 253, 193, 48, 224, 253, 193, 68, 224, 253, 193, 222, 222, 253, 193, 224, 222, 253, 193, 226, 222, 253, 193, 238, 222, 253, 193, 240, 222, 253, 193, 242, 222, 253, 193, 244, 222, 253, 193, 246, 222, 253, 193, 248, 222, 253, 193, 250, 222, 253, 193, 252, 222, 253, 193, 254, 222, 253, 193, 2, 223, 253, 193, 4, 223, 253, 193, 6, 223, 253, 193, 8, 223, 253, 193, 28, 223, 253, 193, 30, 223, 253, 193, 32, 223, 253, 193, 34, 223, 253, 193, 36, 223, 253, 193, 38, 223, 253, 193, 48, 223, 253, 193, 52, 223, 253, 193, 54, 223, 253, 193, 56, 223, 253, 193, 58, 223, 253, 193, 60, 223, 253, 193, 62, 223, 253, 193, 64, 223, 253, 193, 66, 223, 253, 193, 68, 223, 253, 193, 70, 223, 253, 193, 76, 223, 253, 193, 78, 223, 253, 193, 80, 223, 253, 193, 82, 223, 253, 193, 86, 223, 253, 193, 88, 223, 253, 193, 90, 223, 253, 193, 92, 223, 253, 193, 94, 223, 253, 193, 100, 223, 253, 193, 102, 223, 253, 193, 104, 223, 253, 193, 106, 223, 253, 193, 108, 223, 253, 193, 164, 223, 253, 193, 202, 223, 253, 193, 18, 224, 253, 193, 20, 224, 253, 193, 22, 224, 253, 193, 24, 224, 253, 193, 32, 224, 253, 193, 34, 224, 253, 193, 36, 224, 253, 193, 38, 224, 253, 193, 46, 224, 253, 193, 40, 224, 253, 193, 42, 224, 253, 193, 58, 224, 253, 193, 60, 224, 253, 193, 62, 224, 253, 193, 64, 224, 253, 193, 146, 223, 253, 193, 170, 223, 253, 193, 172, 223, 253, 193, 174, 223, 253, 193, 176, 223, 253, 193, 178, 223, 253, 193, 194, 223, 253, 193, 196, 223, 253, 193, 210, 223, 253, 193, 212, 223, 253, 193, 214, 223, 253, 193, 216, 223, 253, 193, 218, 223, 253, 193, 232, 223, 253, 193, 234, 223, 253, 193, 240, 223, 253, 193, 242, 223, 253, 193, 244, 223, 253, 193, 254, 223, 253, 193, 4, 224, 253, 193, 6, 224, 253, 193, 8, 224, 253, 193, 10, 224, 253, 193, 26, 224, 253, 193, 28, 224, 253, 193, 76, 224, 253, 193, 78, 224, 253, 193, 98, 224, 253, 193, 106, 224, 253, 193, 116, 224, 253, 193, 220, 224, 253, 193, 244, 224, 253, 193, 70, 224, 253, 193, 72, 224, 253, 193, 80, 224, 253, 193, 84, 224, 253, 193, 86, 224, 253, 193, 94, 224, 253, 193, 120, 224, 253, 193, 172, 224, 253, 193, 134, 224, 253, 193, 138, 224, 253, 193, 140, 224, 253, 193, 144, 224, 253, 193, 146, 224, 253, 193, 148, 224, 253, 193, 150, 224, 253, 193, 152, 224, 253, 193, 154, 224, 253, 193, 156, 224, 253, 193, 158, 224, 253, 193, 160, 224, 253, 193, 162, 224, 253, 193, 164, 224, 253, 193, 174, 224, 253, 193, 176, 224, 253, 193, 178, 224, 253, 193, 180, 224, 253, 193, 182, 224, 253, 193, 184, 224, 253, 193, 186, 224, 253, 193, 188, 224, 253, 193, 190, 224, 253, 193, 192, 224, 253, 193, 82, 224, 253, 193, 100, 224, 253, 193, 108, 224, 253, 193, 222, 224, 253, 193, 224, 224, 253, 193, 226, 224, 253, 193, 228, 224, 253, 193, 232, 224, 253, 193, 234, 224, 253, 193, 248, 224, 253, 193, 10, 225, 253, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 8, 225, 253, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 16, 56, 102, 193, 9, 70, 102, 193, 79, 194, 102, 193, 2, 172, 102, 193, 23, 180, 102, 193, 16, 70, 102, 193, 23, 70, 102, 193, 44, 198, 102, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 194, 5, 49, 234, 194, 5, 240, 230, 194, 5, 118, 231, 194, 5, 217, 231, 194, 5, 230, 229, 194, 5, 75, 230, 194, 5, 174, 230, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 194, 5, 197, 229, 194, 5, 42, 230, 194, 5, 141, 230, 0, 0, 0, 0, 194, 5, 19, 231, 194, 5, 52, 231, 194, 5, 250, 231, 194, 5, 82, 234, 194, 5, 115, 234, 194, 5, 247, 234, 194, 5, 26, 235, 194, 5, 181, 234, 194, 5, 214, 234, 194, 5, 148, 234, 194, 5, 151, 231, 194, 5, 72, 233, 5, 5, 172, 104, 5, 5, 174, 104, 5, 5, 176, 104, 5, 5, 178, 104, 5, 5, 180, 104, 5, 5, 186, 104, 5, 5, 190, 104, 5, 5, 192, 104, 5, 5, 196, 104, 5, 5, 200, 104, 5, 5, 202, 104, 5, 5, 204, 104, 5, 5, 206, 104, 5, 5, 208, 104, 5, 5, 194, 104, 5, 5, 164, 104, 5, 5, 6, 104, 0, 0, 0, 0, 0, 0, 0, 0, 5, 138, 0, 0, 5, 136, 0, 0, 5, 5, 188, 104, 5, 5, 166, 104, 5, 5, 168, 104, 197, 130, 191, 1, 197, 194, 191, 1, 197, 2, 192, 1, 197, 66, 192, 1, 197, 130, 192, 1, 197, 194, 192, 1, 197, 66, 193, 1, 197, 130, 193, 1, 197, 67, 124, 2, 193, 252, 86, 12, 193, 5, 87, 12, 193, 12, 87, 12, 5, 5, 58, 11, 5, 5, 60, 11, 5, 5, 62, 11, 5, 5, 64, 11, 5, 5, 66, 11, 5, 5, 68, 11, 5, 5, 70, 11, 5, 5, 76, 11, 32, 5, 76, 11, 5, 5, 78, 11, 5, 5, 80, 11, 5, 5, 82, 11, 5, 5, 84, 11, 5, 5, 86, 11, 5, 5, 88, 11, 193, 19, 87, 12, 5, 5, 76, 7, 193, 26, 87, 12, 193, 33, 87, 12, 193, 40, 87, 12, 0, 0, 0, 0, 0, 0, 0, 0, 193, 47, 87, 12, 193, 54, 87, 12, 193, 61, 87, 12, 193, 68, 87, 12, 193, 75, 87, 12, 193, 82, 87, 12, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 21, 5, 0, 21, 21, 5, 0, 23, 21, 5, 0, 25, 21, 5, 0, 27, 21, 5, 0, 29, 21, 5, 0, 31, 21, 5, 0, 33, 21, 5, 0, 35, 22, 5, 0, 37, 21, 5, 0, 19, 193, 89, 87, 12, 0, 0, 0, 0, 193, 96, 87, 12, 0, 0, 0, 0, 193, 103, 87, 12, 194, 5, 248, 244, 5, 5, 154, 9, 5, 5, 156, 9, 5, 5, 158, 9, 5, 5, 160, 9, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 172, 116, 5, 5, 174, 116, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 196, 116, 5, 5, 90, 11, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 150, 116, 5, 5, 154, 116, 5, 5, 162, 116, 5, 5, 164, 116, 5, 5, 158, 116, 5, 5, 152, 116, 5, 5, 156, 116, 5, 5, 160, 116, 5, 5, 8, 116, 5, 5, 14, 116, 5, 5, 18, 116, 197, 130, 120, 2, 5, 5, 22, 116, 5, 5, 26, 116, 5, 5, 30, 116, 5, 5, 34, 116, 255, 255, 255, 255, 5, 5, 38, 116, 5, 5, 42, 116, 5, 5, 46, 116, 5, 5, 50, 116, 197, 2, 121, 2, 5, 5, 54, 116, 5, 5, 58, 116, 193, 124, 87, 12, 193, 131, 87, 12, 193, 138, 87, 12, 193, 145, 87, 12, 193, 152, 87, 12, 193, 159, 87, 12, 0, 0, 0, 0, 193, 166, 87, 12, 193, 173, 87, 12, 193, 180, 87, 12, 193, 187, 87, 12, 193, 194, 87, 12, 193, 201, 87, 12, 255, 255, 255, 255, 193, 208, 87, 12, 193, 215, 87, 12, 5, 5, 72, 11, 5, 5, 74, 11, 5, 5, 92, 11, 5, 5, 94, 11, 5, 5, 96, 11, 193, 222, 87, 12, 193, 229, 87, 12, 193, 236, 87, 12, 193, 243, 87, 12, 5, 5, 98, 11, 5, 5, 100, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 164, 117, 5, 5, 166, 117, 5, 5, 168, 117, 5, 5, 172, 117, 5, 5, 174, 117, 5, 5, 176, 117, 5, 5, 178, 117, 5, 5, 180, 117, 5, 5, 182, 117, 5, 5, 184, 117, 5, 5, 186, 117, 5, 5, 188, 117, 5, 5, 190, 117, 5, 5, 192, 117, 5, 5, 194, 117, 5, 5, 196, 117, 5, 5, 198, 117, 5, 5, 200, 117, 5, 5, 202, 117, 5, 5, 204, 117, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 206, 117, 5, 5, 208, 117, 5, 5, 210, 117, 5, 5, 212, 117, 5, 5, 214, 117, 5, 5, 216, 117, 5, 5, 218, 117, 5, 5, 220, 117, 5, 5, 222, 117, 5, 5, 224, 117, 5, 5, 226, 117, 5, 5, 228, 117, 5, 5, 230, 117, 5, 5, 232, 117, 5, 5, 234, 117, 5, 5, 236, 117, 194, 5, 241, 240, 194, 5, 20, 241, 194, 5, 159, 245, 194, 5, 192, 245, 194, 5, 225, 245, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 5, 238, 117, 0, 0, 0, 0, 5, 5, 48, 9, 5, 5, 50, 9, 5, 5, 78, 7, 5, 5, 90, 13, 5, 5, 112, 11, 5, 5, 114, 11, 5, 5, 116, 11, 5, 5, 172, 13, 5, 5, 170, 117, 5, 166, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 193, 37, 26, 11, 5, 5, 6, 9, 5, 5, 30, 7, 5, 5, 14, 9, 5, 5, 72, 7, 5, 5, 74, 7, 193, 16, 18, 5, 193, 23, 18, 5, 193, 2, 32, 7, 193, 2, 16, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 36, 118, 193, 142, 30, 118, 193, 156, 30, 118, 5, 5, 32, 118, 32, 5, 32, 118, 193, 2, 34, 118, 193, 9, 34, 118, 193, 16, 34, 118, 193, 23, 34, 118, 193, 30, 34, 118, 193, 37, 34, 118, 193, 51, 30, 118, 193, 149, 30, 118, 193, 86, 34, 118, 193, 44, 34, 118, 193, 51, 34, 118, 193, 65, 34, 118, 193, 72, 34, 118, 193, 79, 34, 118, 193, 58, 34, 118, 194, 5, 241, 240, 194, 5, 4, 246, 194, 5, 37, 246, 194, 5, 70, 246, 194, 5, 103, 246, 194, 5, 136, 246, 194, 5, 169, 246, 194, 5, 202, 246, 194, 5, 235, 246, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 193, 9, 18, 5, 193, 22, 89, 12, 193, 29, 89, 12, 193, 36, 89, 12, 193, 43, 89, 12, 193, 50, 89, 12, 193, 57, 89, 12, 193, 64, 89, 12, 193, 71, 89, 12, 193, 78, 89, 12, 193, 85, 89, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 92, 89, 12, 193, 99, 89, 12, 193, 106, 89, 12, 193, 113, 89, 12, 193, 120, 89, 12, 193, 127, 89, 12, 193, 134, 89, 12, 193, 141, 89, 12, 193, 148, 89, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 65, 118, 11, 193, 72, 118, 11, 193, 79, 118, 11, 193, 86, 118, 11, 193, 93, 118, 11, 193, 100, 118, 11, 193, 107, 118, 11, 193, 114, 118, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 148, 104, 32, 5, 148, 104, 32, 5, 148, 104, 32, 5, 148, 104, 194, 5, 241, 240, 32, 5, 148, 104, 32, 5, 148, 104, 32, 5, 148, 104, 32, 5, 148, 104, 194, 5, 20, 241, 194, 5, 20, 241, 0, 0, 0, 0, 5, 5, 150, 104, 5, 5, 152, 104, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 5, 148, 104, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 24, 5, 0, 4, 24, 5, 0, 4, 24, 5, 0, 4, 24, 5, 0, 4, 24, 5, 0, 4, 24, 5, 0, 4, 24, 5, 0, 4, 40, 5, 0, 4, 24, 5, 0, 4, 24, 5, 0, 4, 24, 5, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 20, 5, 32, 5, 20, 5, 5, 5, 22, 5, 5, 5, 24, 5, 5, 5, 26, 5, 5, 5, 28, 5, 5, 5, 102, 10, 5, 5, 12, 5, 5, 5, 108, 9, 5, 5, 110, 9, 5, 5, 112, 9, 5, 5, 114, 9, 5, 5, 122, 9, 5, 5, 124, 9, 5, 5, 126, 9, 5, 5, 128, 9, 5, 5, 164, 10, 5, 5, 166, 10, 5, 5, 176, 10, 5, 5, 178, 10, 24, 5, 0, 8, 197, 194, 3, 0, 197, 67, 4, 0, 5, 5, 180, 10, 5, 5, 16, 3, 5, 5, 18, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 5, 0, 4, 5, 5, 156, 10, 5, 5, 160, 10, 5, 5, 188, 10, 197, 34, 48, 0, 197, 99, 48, 0, 5, 5, 190, 10, 197, 194, 48, 0, 197, 3, 49, 0, 5, 5, 196, 10, 5, 5, 116, 9, 5, 5, 118, 9, 5, 5, 198, 10, 197, 2, 3, 0, 5, 5, 114, 7, 5, 5, 8, 5, 5, 5, 200, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 36, 5, 0, 19, 43, 5, 0, 58, 255, 255, 255, 255, 255, 255, 255, 255, 36, 5, 0, 27, 36, 5, 0, 29, 36, 5, 0, 31, 36, 5, 0, 33, 36, 5, 0, 35, 40, 5, 0, 37, 40, 5, 121, 12, 32, 5, 143, 12, 40, 5, 131, 12, 37, 5, 142, 9, 37, 5, 144, 9, 43, 5, 0, 68, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 193, 93, 141, 103, 193, 14, 142, 103, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 147, 142, 103, 193, 121, 118, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 198, 129, 156, 0, 193, 2, 194, 97, 160, 5, 34, 97, 5, 5, 34, 97, 160, 5, 108, 97, 5, 5, 108, 97, 160, 5, 118, 97, 5, 5, 118, 97, 172, 5, 138, 97, 16, 5, 138, 97, 172, 5, 138, 97, 16, 5, 138, 97, 172, 5, 138, 97, 16, 5, 138, 97, 16, 5, 138, 97, 194, 5, 2, 226, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 136, 10, 32, 5, 42, 97, 32, 5, 70, 97, 197, 2, 44, 1, 32, 5, 172, 97, 16, 5, 226, 97, 32, 5, 230, 97, 16, 5, 232, 97, 198, 193, 155, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 5, 6, 11, 193, 16, 224, 97, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 154, 104, 32, 5, 154, 104, 32, 5, 154, 104, 32, 5, 154, 104, 32, 5, 154, 104, 32, 5, 154, 104, 5, 5, 30, 11, 5, 5, 32, 11, 5, 5, 34, 11, 5, 5, 156, 104, 5, 5, 36, 11, 5, 5, 8, 104, 5, 5, 52, 104, 5, 5, 198, 104, 0, 0, 0, 0, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 110, 17, 5, 5, 112, 17, 5, 5, 114, 17, 5, 5, 116, 17, 5, 5, 118, 17, 5, 5, 120, 17, 5, 5, 122, 17, 5, 5, 124, 17, 5, 5, 126, 17, 5, 5, 128, 17, 5, 5, 130, 17, 5, 5, 132, 17, 5, 5, 134, 17, 5, 5, 136, 17, 5, 5, 138, 17, 5, 5, 140, 17, 5, 5, 142, 17, 5, 5, 144, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 198, 166, 113, 193, 205, 166, 113, 193, 212, 166, 113, 193, 219, 166, 113, 193, 226, 166, 113, 193, 233, 166, 113, 193, 240, 166, 113, 193, 247, 166, 113, 193, 254, 166, 113, 193, 7, 167, 113, 193, 14, 167, 113, 197, 2, 200, 1, 193, 21, 167, 113, 193, 28, 167, 113, 193, 35, 167, 113, 193, 42, 167, 113, 193, 49, 167, 113, 193, 56, 167, 113, 193, 63, 167, 113, 193, 70, 167, 113, 193, 77, 167, 113, 193, 84, 167, 113, 193, 91, 167, 113, 193, 98, 167, 113, 193, 105, 167, 113, 193, 112, 167, 113, 194, 5, 175, 240, 193, 219, 122, 11, 193, 226, 122, 11, 0, 0, 0, 0, 193, 198, 61, 9, 193, 205, 61, 9, 193, 121, 52, 9, 193, 128, 52, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 193, 65, 178, 113, 193, 79, 178, 113, 193, 168, 179, 113, 193, 175, 179, 113, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 2, 106, 13, 193, 9, 106, 13, 5, 5, 40, 7, 5, 5, 108, 13, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 166, 249, 194, 5, 199, 249, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 54, 252, 12, 193, 54, 252, 12, 193, 54, 252, 12, 193, 54, 252, 12, 193, 54, 252, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 96, 252, 12, 193, 103, 252, 12, 193, 110, 252, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 117, 252, 12, 193, 124, 252, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 206, 224, 241, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 87, 253, 12, 193, 94, 253, 12, 193, 101, 253, 12, 193, 108, 253, 12, 193, 115, 253, 12, 193, 122, 253, 12, 193, 129, 253, 12, 193, 136, 253, 12, 193, 143, 253, 12, 193, 150, 253, 12, 193, 157, 253, 12, 193, 164, 253, 12, 193, 171, 253, 12, 193, 164, 253, 12, 193, 171, 253, 12, 193, 164, 253, 12, 193, 171, 253, 12, 193, 164, 253, 12, 193, 144, 2, 13, 193, 151, 2, 13, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 158, 2, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 158, 72, 13, 193, 165, 72, 13, 193, 172, 72, 13, 193, 179, 72, 13, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 193, 186, 72, 13, 193, 193, 72, 13, 193, 200, 72, 13, 193, 207, 72, 13, 193, 214, 72, 13, 193, 221, 72, 13, 193, 228, 72, 13, 193, 235, 72, 13, 0, 0, 0, 0, 206, 64, 242, 0, 206, 64, 242, 0, 206, 64, 242, 0, 206, 64, 242, 0, 206, 64, 242, 0, 206, 64, 242, 0, 206, 64, 242, 0, 206, 64, 242, 0, 206, 64, 242, 0, 206, 64, 242, 0, 0, 0, 0, 0, 193, 86, 73, 13, 193, 93, 73, 13, 193, 58, 128, 11, 193, 65, 128, 11, 193, 72, 128, 11, 193, 79, 128, 11, 193, 86, 128, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 49, 5, 0, 6, 44, 5, 34, 7, 46, 5, 24, 9, 49, 5, 54, 7, 44, 5, 42, 7, 40, 5, 90, 7, 49, 5, 98, 7, 32, 5, 86, 10, 32, 5, 88, 10, 197, 163, 4, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 222, 0, 0, 0, 0, 0, 0, 5, 220, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 168, 0, 0, 0, 0, 0, 0, 5, 220, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 224, 0, 0, 0, 0, 0, 0, 197, 2, 4, 0, 46, 5, 26, 5, 32, 5, 24, 5, 44, 5, 10, 5, 44, 5, 10, 5, 51, 5, 142, 9, 51, 5, 144, 9, 44, 5, 150, 9, 44, 5, 152, 9, 44, 5, 82, 10, 44, 5, 84, 10, 32, 5, 78, 10, 32, 5, 80, 10, 32, 5, 66, 10, 32, 5, 68, 10, 32, 5, 62, 10, 46, 5, 144, 10, 46, 5, 128, 10, 32, 5, 121, 12, 46, 5, 14, 5, 46, 5, 129, 12, 46, 5, 133, 12, 32, 5, 131, 12, 255, 255, 255, 255, 46, 5, 140, 10, 46, 5, 138, 13, 46, 5, 152, 10, 46, 5, 126, 10, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 194, 44, 164, 229, 194, 16, 164, 229, 194, 32, 9, 230, 0, 0, 0, 0, 194, 32, 108, 230, 255, 255, 255, 255, 194, 46, 207, 230, 194, 32, 207, 230, 194, 46, 85, 231, 194, 32, 85, 231, 194, 46, 184, 231, 194, 32, 184, 231, 194, 46, 29, 232, 194, 32, 29, 232, 194, 46, 62, 232, 194, 32, 62, 232, 32, 5, 196, 102, 48, 5, 200, 102, 40, 5, 200, 102, 24, 5, 200, 102, 32, 5, 200, 102, 48, 5, 204, 102, 40, 5, 204, 102, 24, 5, 204, 102, 32, 5, 204, 102, 48, 5, 214, 102, 40, 5, 214, 102, 24, 5, 214, 102, 32, 5, 214, 102, 49, 5, 222, 102, 38, 5, 222, 102, 49, 5, 236, 102, 38, 5, 236, 102, 51, 5, 238, 102, 44, 5, 238, 102, 30, 5, 238, 102, 37, 5, 238, 102, 197, 226, 155, 1, 197, 162, 155, 1, 197, 98, 156, 1, 197, 34, 156, 1, 197, 226, 156, 1, 197, 162, 156, 1, 197, 162, 91, 1, 197, 34, 157, 1, 255, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 160, 5, 214, 96, 5, 5, 214, 96, 160, 5, 216, 96, 5, 5, 216, 96, 197, 227, 36, 1, 193, 13, 227, 12, 193, 20, 227, 12, 193, 27, 227, 12, 193, 34, 227, 12, 193, 41, 227, 12, 193, 48, 227, 12, 160, 5, 162, 96, 5, 5, 162, 96, 160, 5, 194, 96, 5, 5, 194, 96, 5, 166, 0, 0, 5, 134, 0, 0, 5, 132, 0, 0, 160, 5, 174, 96, 5, 5, 174, 96, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 23, 16, 9, 193, 9, 112, 7, 193, 16, 112, 7, 193, 2, 92, 9, 193, 2, 200, 16, 193, 30, 16, 9, 193, 9, 92, 9, 5, 5, 112, 108, 5, 5, 114, 108, 5, 5, 116, 108, 5, 5, 118, 108, 5, 5, 120, 108, 255, 255, 255, 255, 255, 255, 255, 255, 201, 32, 8, 0, 5, 5, 128, 108, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 130, 108, 5, 5, 132, 108, 5, 5, 134, 108, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 166, 0, 0, 5, 5, 136, 108, 5, 5, 138, 108, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 194, 196, 1, 197, 2, 197, 1, 255, 255, 255, 255, 5, 5, 86, 108, 193, 2, 118, 11, 193, 9, 118, 11, 193, 16, 118, 11, 193, 23, 118, 11, 193, 30, 118, 11, 193, 37, 118, 11, 193, 44, 118, 11, 5, 5, 92, 13, 193, 2, 52, 9, 193, 9, 52, 9, 193, 16, 52, 9, 193, 23, 52, 9, 193, 51, 118, 11, 193, 58, 118, 11, 255, 255, 255, 255, 255, 255, 255, 255, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 172, 0, 0, 20, 5, 0, 86, 5, 166, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 19, 5, 0, 76, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 16, 5, 58, 93, 5, 112, 0, 42, 197, 34, 108, 0, 197, 98, 110, 0, 197, 162, 111, 0, 197, 2, 114, 0, 5, 120, 0, 48, 5, 112, 0, 48, 19, 5, 0, 54, 32, 5, 10, 55, 19, 5, 0, 62, 19, 5, 0, 64, 32, 5, 6, 65, 32, 5, 6, 67, 19, 5, 0, 68, 32, 5, 6, 69, 32, 5, 8, 77, 32, 5, 12, 77, 19, 5, 0, 78, 5, 120, 0, 78, 19, 5, 0, 92, 16, 5, 22, 43, 19, 5, 0, 44, 32, 5, 24, 45, 16, 5, 24, 51, 19, 5, 0, 52, 32, 5, 20, 65, 197, 2, 174, 0, 19, 5, 0, 72, 16, 5, 22, 79, 197, 98, 201, 0, 20, 5, 0, 86, 197, 226, 101, 0, 197, 194, 166, 0, 197, 98, 196, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 255, 255, 255, 255, 5, 166, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 172, 0, 0, 194, 5, 21, 251, 5, 172, 0, 0, 5, 172, 0, 0, 5, 172, 0, 0, 5, 170, 0, 0, 194, 5, 54, 251, 194, 5, 87, 251, 194, 5, 120, 251, 194, 5, 153, 251, 5, 170, 0, 0, 5, 170, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 5, 166, 0, 0, 5, 168, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 29, 193, 156, 146, 17, 193, 163, 146, 17, 193, 170, 146, 17, 193, 142, 122, 11, 193, 149, 122, 11, 193, 156, 122, 11, 193, 163, 122, 11, 193, 170, 122, 11, 193, 177, 122, 11, 193, 184, 122, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 37, 94, 117, 193, 44, 94, 117, 193, 247, 94, 117, 193, 254, 94, 117, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 121, 196, 124, 5, 5, 198, 124, 193, 2, 200, 124, 193, 9, 200, 124, 193, 16, 200, 124, 5, 112, 198, 124, 5, 168, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 166, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 5, 5, 0, 21, 193, 35, 147, 17, 193, 42, 147, 17, 193, 49, 147, 17, 193, 163, 61, 9, 193, 170, 61, 9, 193, 177, 61, 9, 193, 184, 61, 9, 193, 191, 61, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 87, 91, 124, 193, 94, 91, 124, 193, 101, 91, 124, 193, 108, 91, 124, 193, 115, 91, 124, 193, 122, 91, 124, 193, 129, 91, 124, 193, 136, 91, 124, 193, 143, 91, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 150, 91, 124, 193, 157, 91, 124, 193, 164, 91, 124, 193, 171, 91, 124, 193, 178, 91, 124, 193, 185, 91, 124, 193, 192, 91, 124, 193, 199, 91, 124, 193, 206, 91, 124, 193, 213, 91, 124, 255, 255, 255, 255, 255, 255, 255, 255, 193, 165, 2, 13, 5, 166, 0, 0, 5, 170, 0, 0, 193, 37, 22, 9, 193, 9, 27, 103, 193, 16, 27, 103, 193, 23, 27, 103, 193, 30, 27, 103, 193, 37, 27, 103, 193, 44, 27, 103, 193, 51, 27, 103, 193, 58, 27, 103, 193, 65, 27, 103, 193, 72, 27, 103, 193, 79, 27, 103, 193, 86, 27, 103, 193, 93, 27, 103, 193, 100, 27, 103, 193, 107, 27, 103, 193, 114, 27, 103, 193, 121, 27, 103, 193, 128, 27, 103, 193, 135, 27, 103, 193, 142, 27, 103, 193, 149, 27, 103, 193, 156, 27, 103, 193, 163, 27, 103, 193, 170, 27, 103, 193, 177, 27, 103, 5, 168, 0, 0, 5, 168, 0, 0, 5, 168, 0, 0, 255, 255, 255, 255, 255, 255, 255, 255, 193, 30, 26, 11, 255, 255, 255, 255, 193, 9, 246, 113, 193, 16, 246, 113, 193, 23, 246, 113, 193, 30, 246, 113, 255, 255, 255, 255, 193, 37, 246, 113, 193, 44, 246, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 51, 246, 113, 5, 168, 0, 0, 194, 5, 241, 240, 194, 5, 20, 241, 193, 58, 246, 113, 193, 72, 246, 113, 193, 79, 246, 113, 193, 86, 246, 113, 255, 255, 255, 255, 193, 93, 246, 113, 193, 100, 246, 113, 193, 107, 246, 113, 255, 255, 255, 255, 193, 114, 246, 113, 193, 121, 246, 113, 193, 135, 246, 113, 193, 149, 246, 113, 193, 156, 246, 113, 193, 163, 246, 113, 193, 170, 246, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 232, 249, 194, 5, 11, 250, 194, 5, 44, 250, 194, 5, 44, 250, 194, 5, 77, 250, 194, 5, 110, 250, 194, 5, 143, 250, 194, 5, 176, 250, 5, 170, 0, 0, 5, 170, 0, 0, 5, 170, 0, 0, 194, 5, 209, 250, 194, 5, 242, 250, 5, 172, 0, 0, 5, 172, 0, 0, 5, 172, 0, 0, 133, 5, 72, 122, 133, 5, 74, 122, 133, 5, 76, 122, 3, 5, 78, 122, 133, 5, 78, 122, 3, 5, 80, 122, 133, 5, 80, 122, 3, 5, 82, 122, 133, 5, 82, 122, 133, 5, 84, 122, 133, 5, 86, 122, 133, 5, 88, 122, 133, 5, 90, 122, 133, 5, 92, 122, 3, 5, 94, 122, 133, 5, 94, 122, 133, 5, 96, 122, 133, 5, 98, 122, 133, 5, 100, 122, 133, 5, 102, 122, 197, 130, 182, 2, 3, 5, 18, 122, 3, 5, 24, 122, 255, 255, 255, 255, 255, 255, 255, 255, 5, 174, 0, 0, 5, 176, 0, 0, 193, 30, 22, 12, 193, 37, 22, 12, 5, 5, 116, 13, 197, 194, 65, 0, 197, 226, 232, 2, 155, 5, 38, 122, 155, 5, 40, 122, 165, 5, 42, 122, 155, 5, 44, 122, 158, 5, 46, 122, 155, 5, 48, 122, 155, 5, 50, 122, 166, 5, 52, 122, 155, 5, 54, 122, 155, 5, 56, 122, 160, 5, 58, 122, 160, 5, 60, 122, 160, 5, 62, 122, 160, 5, 64, 122, 160, 5, 66, 122, 155, 5, 68, 122, 155, 5, 70, 122, 160, 5, 72, 122, 155, 5, 74, 122, 158, 5, 76, 122, 165, 5, 78, 122, 165, 5, 80, 122, 162, 5, 82, 122, 160, 5, 84, 122, 158, 5, 86, 122, 160, 5, 88, 122, 160, 5, 90, 122, 160, 5, 92, 122, 160, 5, 94, 122, 166, 5, 102, 122, 160, 174, 0, 0, 160, 176, 0, 0, 197, 98, 131, 0, 197, 194, 131, 0, 197, 34, 132, 0, 197, 163, 131, 0, 197, 3, 132, 0, 197, 34, 193, 0, 197, 226, 192, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 2, 51, 1, 197, 66, 50, 1, 197, 130, 50, 1, 197, 66, 51, 1, 197, 194, 50, 1, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 66, 55, 1, 194, 5, 22, 228, 197, 131, 56, 1, 193, 9, 51, 101, 193, 16, 51, 101, 193, 23, 51, 101, 193, 30, 51, 101, 193, 37, 51, 101, 193, 44, 51, 101, 193, 51, 51, 101, 193, 58, 51, 101, 193, 65, 51, 101, 193, 72, 51, 101, 193, 79, 51, 101, 193, 86, 51, 101, 193, 93, 51, 101, 193, 100, 51, 101, 193, 107, 51, 101, 193, 114, 51, 101, 193, 121, 51, 101, 193, 128, 51, 101, 193, 135, 51, 101, 193, 142, 51, 101, 193, 149, 51, 101, 193, 156, 51, 101, 193, 163, 51, 101, 193, 170, 51, 101, 194, 5, 65, 229, 194, 5, 98, 229, 193, 177, 51, 101, 193, 184, 51, 101, 194, 5, 55, 228, 194, 5, 55, 228, 194, 5, 88, 228, 194, 5, 88, 228, 194, 5, 88, 228, 194, 5, 121, 228, 194, 5, 121, 228, 194, 5, 121, 228, 194, 5, 154, 228, 194, 5, 154, 228, 194, 5, 187, 228, 194, 5, 187, 228, 194, 5, 220, 228, 194, 5, 220, 228, 194, 5, 220, 228, 194, 5, 253, 228, 194, 5, 32, 229, 194, 5, 131, 229, 255, 255, 255, 255, 255, 255, 255, 255, 193, 44, 60, 7, 193, 51, 60, 7, 193, 58, 60, 7, 193, 65, 60, 7, 193, 72, 60, 7, 193, 79, 60, 7, 193, 86, 60, 7, 193, 93, 60, 7, 193, 100, 60, 7, 193, 107, 60, 7, 193, 114, 60, 7, 193, 121, 60, 7, 193, 128, 60, 7, 193, 135, 60, 7, 193, 142, 60, 7, 255, 255, 255, 255, 197, 2, 160, 1, 197, 2, 162, 1, 197, 2, 122, 1, 197, 194, 119, 1, 197, 66, 121, 1, 197, 226, 165, 1, 197, 162, 167, 1, 197, 98, 97, 1, 197, 162, 170, 1, 197, 66, 171, 1, 197, 130, 171, 1, 197, 2, 172, 1, 197, 194, 174, 1, 197, 98, 176, 1, 197, 226, 177, 1, 197, 226, 179, 1, 197, 98, 180, 1, 197, 2, 92, 1, 197, 98, 181, 1, 197, 98, 182, 1, 197, 162, 182, 1, 197, 34, 185, 1, 197, 66, 94, 1, 197, 162, 186, 1, 197, 226, 187, 1, 197, 2, 190, 1, 197, 130, 190, 1, 197, 2, 98, 1, 197, 66, 98, 1, 197, 98, 184, 1, 197, 98, 0, 0, 197, 162, 0, 0, 197, 162, 65, 1, 197, 34, 74, 1, 197, 98, 75, 1, 197, 66, 82, 1, 197, 98, 85, 1, 197, 162, 87, 1, 197, 162, 88, 1, 197, 226, 103, 1, 197, 130, 106, 1, 197, 2, 113, 1, 197, 130, 115, 1, 197, 194, 151, 1, 197, 194, 152, 1, 197, 66, 163, 1, 197, 98, 177, 1, 197, 98, 179, 1, 197, 162, 187, 1, 197, 130, 189, 1, 197, 226, 0, 0, 197, 98, 1, 0, 197, 226, 1, 0, 197, 226, 132, 1, 197, 98, 133, 1, 197, 162, 137, 1, 197, 34, 138, 1, 197, 194, 140, 1, 197, 66, 141, 1, 197, 2, 107, 1, 197, 130, 107, 1, 197, 2, 116, 1, 197, 130, 116, 1, 197, 2, 96, 1, 197, 34, 1, 0, 197, 162, 1, 0, 197, 34, 2, 0, 197, 98, 2, 0, 197, 162, 63, 1, 197, 226, 63, 1, 197, 162, 64, 1, 197, 34, 65, 1, 197, 162, 67, 1, 197, 34, 68, 1, 197, 98, 73, 1, 197, 162, 73, 1, 197, 98, 74, 1, 197, 226, 74, 1, 197, 162, 75, 1, 197, 34, 76, 1, 197, 130, 81, 1, 197, 194, 81, 1, 197, 130, 82, 1, 197, 226, 84, 1, 197, 162, 85, 1, 197, 34, 86, 1, 197, 226, 86, 1, 197, 34, 87, 1, 197, 226, 87, 1, 197, 98, 88, 1, 197, 226, 88, 1, 197, 98, 89, 1, 197, 162, 144, 1, 197, 34, 145, 1, 197, 130, 148, 1, 197, 2, 149, 1, 5, 5, 22, 107, 5, 5, 26, 107, 5, 5, 126, 107, 5, 5, 128, 107, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 42, 11, 5, 5, 156, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 56, 107, 194, 5, 62, 232, 194, 5, 29, 232, 194, 5, 95, 232, 194, 5, 175, 240, 194, 5, 175, 240, 194, 5, 175, 240, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 194, 5, 241, 240, 193, 112, 173, 113, 194, 5, 175, 240, 194, 5, 29, 232, 193, 163, 52, 9, 193, 170, 52, 9, 193, 21, 123, 11, 193, 28, 123, 11, 193, 35, 123, 11, 193, 42, 123, 11, 194, 5, 62, 232, 255, 255, 255, 255, 193, 233, 201, 124, 193, 247, 201, 124, 193, 254, 201, 124, 193, 7, 202, 124, 193, 14, 202, 124, 193, 21, 202, 124, 193, 28, 202, 124, 193, 35, 202, 124, 193, 42, 202, 124, 193, 56, 202, 124, 255, 255, 255, 255, 194, 5, 128, 232, 194, 5, 95, 232, 5, 5, 50, 5, 255, 255, 255, 255, 255, 255, 255, 255, 193, 240, 201, 124, 193, 49, 202, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 250, 152, 103, 193, 3, 153, 103, 193, 10, 153, 103, 193, 17, 153, 103, 193, 24, 153, 103, 193, 31, 153, 103, 193, 38, 153, 103, 193, 45, 153, 103, 193, 52, 153, 103, 193, 59, 153, 103, 193, 66, 153, 103, 193, 73, 153, 103, 193, 80, 153, 103, 193, 87, 153, 103, 193, 94, 153, 103, 193, 101, 153, 103, 193, 136, 153, 103, 193, 143, 153, 103, 193, 150, 153, 103, 193, 157, 153, 103, 193, 164, 153, 103, 193, 171, 153, 103, 193, 178, 153, 103, 193, 185, 153, 103, 193, 227, 153, 103, 193, 234, 153, 103, 193, 241, 153, 103, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 254, 238, 194, 5, 221, 238, 194, 5, 188, 238, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 194, 5, 33, 239, 194, 5, 66, 239, 193, 100, 61, 9, 193, 16, 22, 9, 5, 5, 88, 7, 193, 23, 32, 7, 5, 5, 50, 7, 193, 30, 112, 7, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 99, 239, 194, 5, 132, 239, 194, 5, 165, 239, 194, 5, 198, 239, 194, 5, 231, 239, 193, 23, 22, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 229, 120, 5, 5, 231, 120, 5, 5, 233, 120, 5, 5, 235, 120, 194, 5, 10, 240, 194, 5, 10, 240, 194, 5, 10, 240, 194, 5, 76, 240, 194, 5, 109, 240, 194, 5, 142, 240, 194, 5, 43, 240, 193, 2, 237, 120, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 16, 96, 7, 193, 51, 112, 7, 5, 5, 90, 104, 5, 5, 94, 104, 5, 5, 98, 104, 5, 5, 100, 104, 5, 5, 102, 104, 5, 5, 104, 104, 5, 5, 106, 104, 5, 5, 108, 104, 5, 5, 110, 104, 197, 2, 193, 1, 5, 5, 112, 104, 5, 5, 114, 104, 5, 5, 116, 104, 5, 5, 120, 104, 5, 5, 122, 104, 5, 5, 124, 104, 5, 5, 128, 104, 197, 194, 193, 1, 5, 5, 130, 104, 5, 5, 132, 104, 197, 2, 194, 1, 5, 5, 134, 104, 5, 5, 136, 104, 5, 5, 138, 104, 5, 5, 140, 104, 5, 5, 142, 104, 5, 5, 160, 104, 5, 5, 162, 104, 194, 5, 175, 240, 5, 5, 144, 104, 5, 5, 158, 104, 5, 5, 170, 104, 5, 5, 58, 105, 5, 5, 60, 105, 5, 5, 62, 105, 5, 5, 64, 105, 5, 5, 66, 105, 5, 5, 68, 105, 5, 5, 70, 105, 5, 5, 72, 105, 5, 5, 74, 105, 255, 255, 255, 255, 5, 5, 76, 105, 5, 5, 78, 105, 5, 5, 80, 105, 5, 5, 82, 105, 5, 5, 84, 105, 5, 5, 86, 105, 5, 5, 88, 105, 255, 255, 255, 255, 5, 5, 92, 105, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 96, 105, 5, 5, 98, 105, 5, 5, 100, 105, 5, 5, 102, 105, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 175, 240, 5, 5, 104, 105, 5, 5, 106, 105, 5, 5, 108, 105, 5, 5, 60, 106, 5, 5, 62, 106, 5, 5, 64, 106, 5, 5, 66, 106, 5, 5, 68, 106, 5, 5, 70, 106, 5, 5, 72, 106, 5, 5, 74, 106, 5, 5, 76, 106, 255, 255, 255, 255, 5, 5, 78, 106, 5, 5, 80, 106, 5, 5, 82, 106, 5, 5, 84, 106, 5, 5, 86, 106, 5, 5, 88, 106, 5, 5, 92, 106, 255, 255, 255, 255, 5, 5, 94, 106, 197, 130, 196, 1, 255, 255, 255, 255, 5, 5, 96, 106, 197, 66, 195, 1, 255, 255, 255, 255, 5, 5, 32, 106, 5, 5, 34, 106, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 175, 240, 255, 255, 255, 255, 5, 5, 100, 106, 5, 5, 102, 106, 5, 5, 64, 107, 5, 5, 66, 107, 5, 5, 68, 107, 5, 5, 70, 107, 5, 5, 72, 107, 5, 5, 74, 107, 5, 5, 76, 107, 5, 5, 78, 107, 5, 5, 80, 107, 255, 255, 255, 255, 5, 5, 82, 107, 5, 5, 84, 107, 5, 5, 86, 107, 5, 5, 88, 107, 5, 5, 90, 107, 5, 5, 92, 107, 5, 5, 94, 107, 255, 255, 255, 255, 5, 5, 96, 107, 5, 5, 108, 107, 255, 255, 255, 255, 5, 5, 98, 107, 5, 5, 100, 107, 5, 5, 102, 107, 5, 5, 104, 107, 5, 5, 106, 107, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 175, 240, 5, 5, 110, 107, 5, 5, 112, 107, 5, 5, 114, 107, 5, 5, 56, 108, 5, 5, 58, 108, 5, 5, 60, 108, 5, 5, 62, 108, 5, 5, 64, 108, 5, 5, 66, 108, 5, 5, 68, 108, 5, 5, 70, 108, 5, 5, 72, 108, 255, 255, 255, 255, 5, 5, 74, 108, 5, 5, 76, 108, 5, 5, 78, 108, 5, 5, 80, 108, 5, 5, 82, 108, 5, 5, 84, 108, 5, 5, 88, 108, 255, 255, 255, 255, 5, 5, 90, 108, 5, 5, 92, 108, 255, 255, 255, 255, 5, 5, 94, 108, 5, 5, 98, 108, 5, 5, 100, 108, 5, 5, 102, 108, 5, 5, 104, 108, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 175, 240, 5, 5, 106, 108, 5, 5, 108, 108, 5, 5, 110, 108, 5, 5, 60, 111, 5, 5, 62, 111, 5, 5, 64, 111, 5, 5, 66, 111, 5, 5, 68, 111, 5, 5, 70, 111, 5, 5, 72, 111, 5, 5, 74, 111, 5, 5, 76, 111, 255, 255, 255, 255, 5, 5, 78, 111, 5, 5, 80, 111, 5, 5, 82, 111, 5, 5, 84, 111, 5, 5, 86, 111, 5, 5, 88, 111, 5, 5, 90, 111, 5, 5, 92, 111, 5, 5, 94, 111, 5, 5, 106, 111, 255, 255, 255, 255, 5, 5, 96, 111, 5, 5, 98, 111, 5, 5, 100, 111, 5, 5, 102, 111, 5, 5, 104, 111, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 175, 240, 5, 5, 110, 111, 5, 5, 118, 111, 201, 0, 14, 0, 193, 212, 40, 118, 193, 219, 40, 118, 193, 226, 40, 118, 193, 240, 40, 118, 193, 247, 40, 118, 193, 254, 40, 118, 193, 7, 41, 118, 193, 14, 41, 118, 193, 28, 41, 118, 193, 35, 41, 118, 193, 42, 41, 118, 193, 49, 41, 118, 193, 56, 41, 118, 193, 63, 41, 118, 193, 70, 41, 118, 193, 77, 41, 118, 193, 91, 41, 118, 193, 98, 41, 118, 193, 105, 41, 118, 193, 126, 41, 118, 194, 5, 175, 240, 193, 133, 41, 118, 193, 140, 41, 118, 193, 147, 41, 118, 193, 154, 41, 118, 193, 161, 41, 118, 201, 192, 69, 1, 193, 175, 41, 118, 201, 160, 70, 1, 193, 189, 41, 118, 201, 128, 71, 1, 201, 96, 72, 1, 193, 2, 63, 117, 193, 9, 63, 117, 193, 16, 63, 117, 193, 23, 63, 117, 193, 30, 63, 117, 193, 37, 63, 117, 194, 5, 175, 240, 5, 5, 65, 117, 32, 5, 65, 117, 5, 5, 67, 117, 5, 5, 69, 117, 32, 5, 69, 117, 5, 5, 71, 117, 32, 5, 71, 117, 5, 5, 73, 117, 32, 5, 73, 117, 193, 2, 75, 117, 193, 9, 75, 117, 193, 16, 75, 117, 193, 23, 75, 117, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 142, 118, 11, 193, 149, 118, 11, 193, 156, 118, 11, 193, 163, 118, 11, 193, 14, 7, 117, 193, 21, 7, 117, 193, 28, 7, 117, 193, 35, 7, 117, 193, 219, 6, 117, 193, 233, 6, 117, 193, 49, 7, 117, 193, 56, 7, 117, 193, 63, 7, 117, 193, 70, 7, 117, 193, 77, 7, 117, 193, 84, 7, 117, 193, 91, 7, 117, 193, 98, 7, 117, 193, 105, 7, 117, 193, 112, 7, 117, 193, 119, 7, 117, 193, 126, 7, 117, 193, 133, 7, 117, 193, 140, 7, 117, 193, 147, 7, 117, 193, 154, 7, 117, 193, 42, 7, 117, 194, 5, 175, 240, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 16, 36, 9, 193, 23, 36, 9, 193, 128, 102, 11, 193, 135, 102, 11, 193, 142, 102, 11, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 5, 5, 45, 118, 32, 5, 45, 118, 193, 9, 47, 118, 193, 16, 47, 118, 193, 23, 47, 118, 193, 30, 47, 118, 193, 37, 47, 118, 193, 44, 47, 118, 194, 5, 175, 240, 5, 5, 49, 118, 32, 5, 49, 118, 193, 9, 51, 118, 193, 16, 51, 118, 193, 23, 51, 118, 193, 30, 51, 118, 193, 44, 51, 118, 193, 51, 51, 118, 193, 2, 51, 118, 193, 37, 51, 118, 193, 28, 43, 118, 193, 2, 47, 118, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 193, 240, 168, 113, 193, 247, 168, 113, 193, 254, 168, 113, 194, 5, 175, 240, 193, 233, 122, 11, 193, 240, 122, 11, 193, 233, 168, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 231, 171, 113, 193, 105, 171, 113, 193, 112, 171, 113, 193, 119, 171, 113, 193, 9, 170, 113, 193, 149, 52, 9, 193, 156, 52, 9, 193, 254, 122, 11, 193, 219, 61, 9, 194, 5, 53, 241, 194, 5, 175, 240, 194, 5, 129, 242, 194, 5, 162, 242, 193, 247, 122, 11, 193, 203, 171, 113, 194, 5, 208, 240, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 193, 16, 170, 113, 193, 7, 123, 11, 193, 126, 171, 113, 193, 14, 123, 11, 193, 226, 61, 9, 193, 233, 61, 9, 193, 84, 175, 113, 193, 91, 175, 113, 193, 98, 175, 113, 193, 105, 175, 113, 193, 112, 175, 113, 193, 119, 175, 113, 193, 126, 175, 113, 193, 133, 175, 113, 193, 140, 175, 113, 194, 5, 175, 240, 193, 147, 175, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 255, 255, 255, 255, 193, 254, 178, 113, 193, 7, 179, 113, 193, 14, 179, 113, 193, 21, 179, 113, 193, 28, 179, 113, 193, 35, 179, 113, 193, 42, 179, 113, 255, 255, 255, 255, 193, 49, 179, 113, 193, 56, 179, 113, 255, 255, 255, 255, 193, 63, 179, 113, 193, 70, 179, 113, 193, 77, 179, 113, 193, 84, 179, 113, 193, 91, 179, 113, 255, 255, 255, 255, 194, 5, 175, 240, 194, 5, 175, 240, 193, 98, 179, 113, 193, 119, 179, 113, 193, 126, 179, 113, 193, 252, 181, 113, 193, 5, 182, 113, 193, 12, 182, 113, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 194, 5, 175, 240, 193, 140, 181, 113, 193, 161, 181, 113, 193, 9, 180, 113, 193, 16, 180, 113, 193, 177, 52, 9, 193, 184, 52, 9, 193, 49, 123, 11, 193, 63, 123, 11, 193, 70, 123, 11, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 193, 56, 123, 11, 193, 77, 123, 11, 255, 255, 255, 255, 193, 84, 123, 11, 194, 5, 53, 241, 193, 168, 181, 113, 194, 5, 241, 240, 194, 5, 20, 241, 193, 217, 184, 113, 194, 5, 175, 240, 193, 98, 184, 113, 193, 105, 184, 113, 193, 91, 123, 11, 193, 9, 183, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 175, 240, 193, 98, 123, 11, 193, 191, 52, 9, 193, 198, 52, 9, 193, 105, 123, 11, 193, 112, 123, 11, 193, 119, 123, 11, 193, 126, 123, 11, 193, 133, 123, 11, 193, 140, 123, 11, 193, 147, 123, 11, 193, 154, 123, 11, 193, 161, 123, 11, 193, 168, 123, 11, 193, 175, 123, 11, 193, 182, 123, 11, 193, 189, 123, 11, 193, 196, 123, 11, 193, 203, 123, 11, 193, 210, 123, 11, 193, 217, 123, 11, 193, 224, 123, 11, 193, 231, 123, 11, 193, 238, 123, 11, 5, 112, 187, 113, 5, 120, 187, 113, 5, 112, 189, 113, 5, 112, 191, 113, 5, 112, 196, 113, 5, 112, 198, 113, 255, 255, 255, 255, 255, 255, 255, 255, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 193, 86, 203, 113, 193, 79, 203, 113, 193, 93, 203, 113, 193, 56, 204, 113, 194, 5, 241, 240, 194, 5, 20, 241, 193, 63, 204, 113, 193, 70, 204, 113, 193, 77, 204, 113, 193, 84, 204, 113, 193, 91, 204, 113, 193, 98, 204, 113, 193, 105, 204, 113, 193, 112, 204, 113, 193, 119, 204, 113, 193, 126, 204, 113, 194, 5, 175, 240, 193, 114, 203, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 194, 5, 241, 240, 194, 5, 20, 241, 193, 140, 210, 113, 194, 5, 175, 240, 193, 252, 123, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 240, 205, 113, 193, 254, 205, 113, 193, 14, 206, 113, 194, 5, 175, 240, 193, 233, 52, 9, 193, 5, 124, 11, 193, 254, 61, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 5, 5, 200, 116, 5, 5, 202, 116, 5, 5, 204, 116, 5, 5, 206, 116, 193, 2, 208, 116, 193, 9, 208, 116, 193, 16, 208, 116, 193, 23, 208, 116, 193, 16, 198, 116, 194, 5, 175, 240, 193, 107, 208, 116, 194, 5, 208, 240, 194, 5, 208, 240, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 32, 5, 202, 116, 32, 5, 200, 116, 46, 5, 202, 116, 32, 5, 204, 116, 32, 5, 206, 116, 193, 2, 102, 11, 194, 5, 241, 240, 194, 5, 20, 241, 194, 5, 175, 240, 194, 5, 208, 240, 193, 142, 228, 113, 193, 149, 228, 113, 32, 5, 226, 113, 46, 5, 226, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 208, 240, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 5, 5, 12, 104, 5, 5, 14, 104, 5, 5, 16, 104, 5, 5, 28, 104, 5, 5, 30, 104, 5, 5, 32, 104, 5, 5, 34, 104, 5, 5, 36, 104, 5, 5, 40, 104, 5, 5, 44, 104, 5, 5, 46, 104, 5, 5, 48, 104, 5, 5, 50, 104, 5, 5, 54, 104, 5, 5, 56, 104, 5, 5, 58, 104, 5, 5, 60, 104, 5, 5, 62, 104, 5, 5, 64, 104, 5, 5, 66, 104, 5, 5, 70, 104, 5, 5, 72, 104, 5, 5, 74, 104, 5, 5, 76, 104, 5, 5, 78, 104, 5, 5, 84, 104, 5, 5, 86, 104, 5, 5, 88, 104, 5, 5, 6, 105, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 255, 255, 255, 255, 5, 5, 8, 105, 5, 5, 10, 105, 5, 5, 12, 105, 5, 5, 14, 105, 5, 5, 16, 105, 5, 5, 18, 105, 5, 5, 20, 105, 5, 5, 24, 105, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 28, 105, 5, 5, 30, 105, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 32, 105, 5, 5, 34, 105, 5, 5, 36, 105, 5, 5, 38, 105, 5, 5, 40, 105, 5, 5, 42, 105, 5, 5, 44, 105, 5, 5, 46, 105, 5, 5, 48, 105, 5, 5, 50, 105, 5, 5, 52, 105, 5, 5, 54, 105, 5, 5, 56, 105, 255, 255, 255, 255, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 255, 255, 255, 255, 5, 5, 16, 106, 5, 5, 18, 106, 5, 5, 26, 106, 5, 5, 28, 106, 5, 5, 10, 106, 5, 5, 12, 106, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 30, 106, 5, 5, 20, 106, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 14, 106, 5, 5, 22, 106, 5, 5, 38, 106, 5, 5, 40, 106, 5, 5, 42, 106, 5, 5, 44, 106, 5, 5, 46, 106, 5, 5, 48, 106, 5, 5, 50, 106, 5, 5, 52, 106, 5, 5, 54, 106, 5, 5, 56, 106, 5, 5, 58, 106, 255, 255, 255, 255, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 255, 255, 255, 255, 5, 5, 8, 107, 5, 5, 10, 107, 5, 5, 12, 107, 5, 5, 14, 107, 5, 5, 16, 107, 5, 5, 18, 107, 5, 5, 20, 107, 5, 5, 24, 107, 5, 5, 28, 107, 255, 255, 255, 255, 5, 5, 30, 107, 5, 5, 32, 107, 5, 5, 34, 107, 255, 255, 255, 255, 5, 5, 36, 107, 5, 5, 38, 107, 5, 5, 40, 107, 5, 5, 42, 107, 5, 5, 44, 107, 5, 5, 46, 107, 5, 5, 48, 107, 5, 5, 50, 107, 5, 5, 52, 107, 5, 5, 54, 107, 5, 5, 58, 107, 5, 5, 60, 107, 5, 5, 62, 107, 255, 255, 255, 255, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 255, 255, 255, 255, 5, 5, 6, 108, 5, 5, 8, 108, 5, 5, 10, 108, 5, 5, 12, 108, 5, 5, 14, 108, 5, 5, 16, 108, 5, 5, 18, 108, 5, 5, 22, 108, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 26, 108, 5, 5, 28, 108, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 30, 108, 5, 5, 32, 108, 5, 5, 34, 108, 5, 5, 36, 108, 5, 5, 38, 108, 5, 5, 40, 108, 5, 5, 42, 108, 5, 5, 44, 108, 5, 5, 46, 108, 5, 5, 48, 108, 5, 5, 50, 108, 5, 5, 52, 108, 5, 5, 54, 108, 194, 5, 208, 240, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 194, 5, 241, 240, 5, 5, 6, 110, 5, 5, 8, 110, 5, 5, 10, 110, 5, 5, 12, 110, 5, 5, 14, 110, 5, 5, 16, 110, 5, 5, 18, 110, 5, 5, 22, 110, 255, 255, 255, 255, 5, 5, 26, 110, 5, 5, 28, 110, 5, 5, 30, 110, 255, 255, 255, 255, 5, 5, 32, 110, 5, 5, 34, 110, 5, 5, 36, 110, 5, 5, 38, 110, 5, 5, 40, 110, 5, 5, 42, 110, 5, 5, 44, 110, 5, 5, 46, 110, 5, 5, 48, 110, 5, 5, 52, 110, 5, 5, 54, 110, 5, 5, 58, 110, 5, 5, 60, 110, 5, 5, 62, 110, 5, 5, 116, 111, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 5, 5, 46, 11, 5, 5, 6, 111, 5, 5, 8, 111, 5, 5, 10, 111, 5, 5, 12, 111, 5, 5, 14, 111, 5, 5, 16, 111, 5, 5, 18, 111, 5, 5, 22, 111, 255, 255, 255, 255, 5, 5, 26, 111, 5, 5, 28, 111, 5, 5, 30, 111, 255, 255, 255, 255, 5, 5, 32, 111, 5, 5, 34, 111, 5, 5, 36, 111, 5, 5, 38, 111, 5, 5, 40, 111, 5, 5, 42, 111, 5, 5, 44, 111, 5, 5, 46, 111, 5, 5, 48, 111, 5, 5, 50, 111, 5, 5, 52, 111, 5, 5, 54, 111, 5, 5, 56, 111, 5, 5, 58, 111, 194, 5, 241, 240, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 5, 5, 118, 112, 5, 5, 6, 112, 5, 5, 8, 112, 5, 5, 10, 112, 5, 5, 12, 112, 5, 5, 16, 112, 5, 5, 18, 112, 5, 5, 20, 112, 5, 5, 24, 112, 255, 255, 255, 255, 5, 5, 28, 112, 5, 5, 30, 112, 5, 5, 32, 112, 255, 255, 255, 255, 5, 5, 34, 112, 5, 5, 36, 112, 5, 5, 38, 112, 5, 5, 40, 112, 5, 5, 42, 112, 5, 5, 44, 112, 5, 5, 46, 112, 5, 5, 48, 112, 5, 5, 50, 112, 5, 5, 52, 112, 5, 5, 54, 112, 5, 5, 56, 112, 5, 5, 58, 112, 5, 5, 60, 112, 255, 255, 255, 255, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 255, 255, 255, 255, 5, 5, 6, 113, 5, 5, 8, 113, 5, 5, 10, 113, 5, 5, 12, 113, 5, 5, 14, 113, 5, 5, 16, 113, 5, 5, 18, 113, 5, 5, 20, 113, 5, 5, 22, 113, 5, 5, 24, 113, 5, 5, 26, 113, 5, 5, 28, 113, 5, 5, 30, 113, 5, 5, 32, 113, 5, 5, 34, 113, 5, 5, 36, 113, 5, 5, 38, 113, 5, 5, 40, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 42, 113, 5, 5, 44, 113, 5, 5, 46, 113, 5, 5, 48, 113, 5, 5, 50, 113, 5, 5, 52, 113, 194, 5, 208, 240, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 152, 241, 194, 5, 20, 241, 201, 128, 64, 1, 193, 16, 40, 118, 201, 96, 65, 1, 193, 30, 40, 118, 201, 64, 66, 1, 193, 44, 40, 118, 201, 32, 67, 1, 193, 58, 40, 118, 201, 0, 68, 1, 193, 72, 40, 118, 193, 79, 40, 118, 193, 86, 40, 118, 201, 224, 68, 1, 193, 100, 40, 118, 193, 107, 40, 118, 193, 128, 40, 118, 193, 135, 40, 118, 193, 142, 40, 118, 193, 149, 40, 118, 193, 156, 40, 118, 193, 163, 40, 118, 193, 170, 40, 118, 193, 177, 40, 118, 193, 184, 40, 118, 193, 191, 40, 118, 193, 198, 40, 118, 193, 205, 40, 118, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 194, 5, 208, 240, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 2, 36, 9, 193, 9, 36, 9, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 185, 241, 194, 5, 20, 241, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 206, 192, 233, 0, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 193, 84, 245, 113, 193, 91, 245, 113, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 193, 9, 166, 113, 193, 16, 166, 113, 193, 23, 166, 113, 193, 30, 166, 113, 193, 37, 166, 113, 193, 44, 166, 113, 193, 51, 166, 113, 193, 58, 166, 113, 193, 65, 166, 113, 193, 72, 166, 113, 193, 79, 166, 113, 193, 86, 166, 113, 193, 93, 166, 113, 193, 100, 166, 113, 193, 107, 166, 113, 193, 114, 166, 113, 193, 121, 166, 113, 193, 128, 166, 113, 193, 135, 166, 113, 193, 142, 166, 113, 193, 149, 166, 113, 193, 156, 166, 113, 193, 163, 166, 113, 197, 130, 199, 1, 193, 170, 166, 113, 197, 194, 199, 1, 193, 177, 166, 113, 193, 184, 166, 113, 193, 191, 166, 113, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 194, 5, 241, 240, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 255, 255, 255, 255, 193, 16, 178, 113, 193, 23, 178, 113, 193, 30, 178, 113, 193, 37, 178, 113, 193, 44, 178, 113, 193, 51, 178, 113, 193, 58, 178, 113, 193, 72, 178, 113, 255, 255, 255, 255, 255, 255, 255, 255, 193, 86, 178, 113, 193, 93, 178, 113, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 160, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 193, 112, 184, 113, 193, 119, 184, 113, 193, 126, 184, 113, 193, 133, 184, 113, 193, 140, 184, 113, 193, 147, 184, 113, 193, 154, 184, 113, 193, 161, 184, 113, 193, 168, 184, 113, 201, 192, 26, 2, 193, 182, 184, 113, 193, 189, 184, 113, 193, 196, 184, 113, 193, 203, 184, 113, 193, 210, 184, 113, 194, 5, 208, 240, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 5, 5, 196, 113, 5, 5, 198, 113, 193, 2, 200, 113, 193, 9, 200, 113, 255, 255, 255, 255, 255, 255, 255, 255, 201, 224, 28, 2, 201, 224, 29, 2, 193, 30, 200, 113, 193, 37, 200, 113, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 193, 44, 200, 113, 194, 5, 208, 240, 193, 205, 52, 9, 193, 212, 52, 9, 193, 245, 123, 11, 193, 189, 202, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 128, 238, 0, 206, 128, 238, 0, 206, 128, 238, 0, 206, 128, 238, 0, 206, 128, 238, 0, 206, 128, 238, 0, 193, 247, 205, 113, 193, 7, 206, 113, 193, 21, 206, 113, 193, 28, 206, 113, 193, 35, 206, 113, 193, 42, 206, 113, 193, 49, 206, 113, 193, 56, 206, 113, 193, 63, 206, 113, 193, 70, 206, 113, 193, 77, 206, 113, 193, 84, 206, 113, 193, 91, 206, 113, 193, 98, 206, 113, 193, 105, 206, 113, 201, 224, 30, 2, 255, 255, 255, 255, 193, 119, 206, 113, 193, 126, 206, 113, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 241, 240, 194, 5, 208, 240, 193, 133, 206, 113, 193, 140, 206, 113, 193, 100, 205, 113, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 193, 84, 249, 113, 193, 91, 249, 113, 193, 98, 249, 113, 193, 105, 249, 113, 193, 112, 249, 113, 193, 119, 249, 113, 193, 126, 249, 113, 193, 133, 249, 113, 255, 255, 255, 255, 193, 140, 249, 113, 193, 147, 249, 113, 193, 154, 249, 113, 193, 161, 249, 113, 194, 5, 208, 240, 194, 5, 241, 240, 194, 5, 20, 241, 193, 168, 249, 113, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 255, 255, 255, 255, 193, 77, 5, 117, 193, 91, 5, 117, 193, 105, 5, 117, 193, 119, 5, 117, 193, 133, 5, 117, 193, 147, 5, 117, 193, 161, 5, 117, 193, 168, 5, 117, 193, 175, 5, 117, 193, 182, 5, 117, 193, 189, 5, 117, 193, 196, 5, 117, 194, 5, 241, 240, 194, 5, 208, 240, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 241, 240, 5, 5, 32, 109, 255, 255, 255, 255, 5, 5, 8, 109, 5, 5, 10, 109, 5, 5, 12, 109, 5, 5, 14, 109, 5, 5, 16, 109, 5, 5, 18, 109, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 20, 109, 5, 5, 22, 109, 5, 5, 24, 109, 255, 255, 255, 255, 201, 0, 10, 0, 5, 5, 28, 109, 5, 5, 30, 109, 5, 5, 34, 109, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 36, 109, 5, 5, 38, 109, 255, 255, 255, 255, 5, 5, 70, 109, 255, 255, 255, 255, 5, 5, 40, 109, 5, 5, 42, 109, 5, 5, 112, 116, 5, 5, 116, 116, 5, 5, 120, 116, 5, 5, 126, 116, 5, 5, 130, 116, 5, 5, 134, 116, 5, 5, 138, 116, 5, 5, 142, 116, 5, 5, 146, 116, 197, 194, 119, 2, 5, 112, 120, 116, 5, 5, 10, 116, 5, 5, 124, 116, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 201, 70, 231, 0, 5, 5, 168, 116, 5, 5, 170, 116, 5, 5, 176, 116, 5, 5, 178, 116, 5, 5, 180, 116, 5, 5, 182, 116, 5, 5, 184, 116, 5, 5, 186, 116, 5, 5, 188, 116, 5, 5, 190, 116, 5, 5, 192, 116, 5, 5, 194, 116, 194, 5, 241, 240, 194, 5, 20, 241, 193, 51, 85, 117, 193, 100, 85, 117, 193, 107, 85, 117, 193, 114, 85, 117, 193, 121, 85, 117, 201, 32, 241, 0, 193, 135, 85, 117, 193, 170, 85, 117, 193, 177, 85, 117, 193, 184, 85, 117, 193, 191, 85, 117, 32, 5, 87, 117, 5, 5, 87, 117, 193, 23, 89, 117, 193, 37, 89, 117, 193, 51, 89, 117, 193, 72, 89, 117, 193, 107, 89, 117, 193, 135, 89, 117, 193, 44, 89, 117, 193, 149, 89, 117, 193, 121, 89, 117, 194, 5, 241, 240, 194, 5, 126, 245, 194, 5, 20, 241, 5, 5, 91, 117, 193, 2, 93, 117, 193, 54, 81, 117, 193, 82, 81, 117, 193, 110, 81, 117, 193, 30, 85, 117, 197, 35, 125, 2, 194, 5, 241, 240, 194, 5, 218, 241, 194, 5, 20, 241, 5, 5, 233, 113, 193, 2, 235, 113, 193, 9, 235, 113, 193, 16, 235, 113, 193, 23, 235, 113, 193, 30, 235, 113, 193, 37, 235, 113, 5, 5, 237, 113, 193, 9, 239, 113, 193, 16, 239, 113, 193, 23, 239, 113, 193, 30, 239, 113, 193, 37, 239, 113, 193, 44, 239, 113, 193, 51, 239, 113, 193, 58, 239, 113, 193, 65, 239, 113, 193, 72, 239, 113, 193, 79, 239, 113, 193, 86, 239, 113, 193, 93, 239, 113, 193, 100, 239, 113, 5, 5, 241, 113, 193, 9, 243, 113, 193, 23, 243, 113, 193, 44, 243, 113, 193, 65, 243, 113, 193, 79, 243, 113, 193, 86, 243, 113, 193, 9, 162, 113, 193, 16, 162, 113, 193, 23, 162, 113, 193, 30, 162, 113, 193, 37, 162, 113, 193, 44, 162, 113, 193, 51, 162, 113, 193, 65, 162, 113, 193, 72, 162, 113, 193, 79, 162, 113, 193, 86, 162, 113, 194, 5, 241, 240, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 194, 5, 241, 240, 194, 5, 20, 241, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 194, 5, 241, 240, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 194, 5, 241, 240, 194, 5, 20, 241, 193, 182, 202, 113, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 193, 98, 208, 113, 193, 105, 208, 113, 193, 112, 208, 113, 193, 119, 208, 113, 193, 126, 208, 113, 193, 133, 208, 113, 193, 140, 208, 113, 255, 255, 255, 255, 255, 255, 255, 255, 193, 147, 208, 113, 193, 161, 208, 113, 193, 168, 208, 113, 193, 175, 208, 113, 194, 5, 241, 240, 194, 5, 20, 241, 5, 5, 249, 116, 5, 5, 251, 116, 193, 2, 253, 116, 5, 5, 213, 116, 193, 9, 253, 116, 193, 16, 253, 116, 32, 5, 241, 116, 32, 5, 243, 116, 32, 5, 247, 116, 32, 5, 251, 116, 32, 5, 211, 116, 32, 5, 215, 116, 32, 5, 219, 116, 32, 5, 223, 116, 32, 5, 227, 116, 32, 5, 229, 116, 32, 5, 233, 116, 46, 5, 241, 116, 46, 5, 243, 116, 46, 5, 247, 116, 46, 5, 251, 116, 32, 5, 237, 116, 194, 5, 241, 240, 194, 5, 20, 241, 194, 5, 195, 242, 193, 23, 253, 116, 193, 58, 102, 11, 193, 65, 102, 11, 193, 72, 102, 11, 193, 30, 253, 116, 193, 79, 102, 11, 193, 86, 102, 11, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 255, 255, 255, 255, 193, 77, 230, 113, 193, 84, 230, 113, 255, 255, 255, 255, 193, 91, 230, 113, 193, 98, 230, 113, 194, 5, 241, 240, 194, 5, 20, 241, 193, 105, 230, 113, 193, 9, 229, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 22, 105, 5, 5, 26, 105, 5, 5, 120, 105, 5, 5, 122, 105, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 90, 105, 5, 5, 94, 105, 5, 5, 150, 13, 5, 5, 152, 13, 5, 5, 6, 16, 5, 5, 8, 16, 5, 5, 10, 16, 5, 5, 12, 16, 5, 5, 14, 16, 5, 5, 16, 16, 193, 198, 85, 12, 5, 5, 154, 13, 5, 5, 136, 105, 5, 5, 38, 11, 194, 5, 53, 241, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 194, 5, 86, 241, 194, 5, 119, 241, 5, 5, 24, 106, 5, 5, 8, 106, 5, 5, 6, 106, 5, 5, 90, 106, 5, 5, 40, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 161, 161, 113, 193, 168, 161, 113, 193, 175, 161, 113, 193, 28, 161, 113, 193, 35, 161, 113, 193, 42, 161, 113, 193, 49, 161, 113, 193, 56, 161, 113, 193, 63, 161, 113, 193, 70, 161, 113, 193, 77, 161, 113, 193, 86, 52, 9, 194, 5, 251, 241, 193, 182, 161, 113, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 177, 246, 113, 193, 184, 246, 113, 193, 191, 246, 113, 193, 198, 246, 113, 193, 205, 246, 113, 193, 212, 246, 113, 193, 219, 246, 113, 193, 226, 246, 113, 193, 233, 246, 113, 193, 240, 246, 113, 193, 247, 246, 113, 193, 254, 246, 113, 193, 7, 247, 113, 193, 21, 247, 113, 193, 28, 247, 113, 193, 35, 247, 113, 193, 42, 247, 113, 193, 49, 247, 113, 193, 65, 246, 113, 193, 142, 246, 113, 193, 128, 246, 113, 193, 14, 247, 113, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 30, 242, 194, 5, 63, 242, 194, 5, 96, 242, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 56, 247, 113, 201, 224, 23, 0, 201, 192, 44, 0, 201, 160, 65, 0, 201, 128, 86, 0, 201, 96, 107, 0, 5, 5, 132, 114, 5, 5, 86, 13, 194, 5, 7, 243, 194, 5, 40, 243, 194, 5, 73, 243, 194, 5, 106, 243, 194, 5, 139, 243, 194, 5, 172, 243, 201, 64, 128, 0, 194, 5, 228, 242, 5, 5, 50, 11, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 52, 11, 5, 5, 54, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 201, 32, 129, 0, 201, 96, 149, 0, 201, 160, 169, 0, 201, 224, 189, 0, 201, 32, 210, 0, 255, 255, 255, 255, 5, 5, 88, 13, 255, 255, 255, 255, 194, 5, 238, 243, 194, 5, 17, 244, 194, 5, 50, 244, 194, 5, 83, 244, 194, 5, 116, 244, 201, 96, 230, 0, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 197, 130, 50, 2, 197, 162, 52, 2, 5, 5, 6, 115, 5, 5, 30, 115, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 201, 32, 74, 1, 201, 64, 96, 1, 193, 140, 133, 115, 193, 147, 133, 115, 201, 96, 118, 1, 193, 161, 133, 115, 201, 128, 140, 1, 201, 160, 162, 1, 193, 182, 133, 115, 193, 189, 133, 115, 194, 5, 182, 244, 193, 196, 133, 115, 194, 5, 215, 244, 193, 203, 133, 115, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 210, 133, 115, 193, 217, 133, 115, 193, 2, 100, 13, 193, 2, 56, 11, 193, 9, 56, 11, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 194, 5, 27, 245, 194, 5, 60, 245, 194, 5, 93, 245, 5, 5, 120, 11, 5, 5, 42, 9, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 193, 121, 18, 117, 193, 128, 18, 117, 193, 135, 18, 117, 193, 142, 18, 117, 193, 149, 18, 117, 193, 156, 18, 117, 193, 163, 18, 117, 193, 170, 18, 117, 193, 177, 18, 117, 5, 5, 20, 117, 5, 5, 22, 117, 193, 2, 24, 117, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 24, 117, 193, 16, 24, 117, 193, 23, 24, 117, 193, 30, 24, 117, 193, 37, 24, 117, 193, 44, 24, 117, 193, 51, 24, 117, 193, 58, 24, 117, 193, 65, 24, 117, 194, 5, 14, 247, 194, 5, 47, 247, 194, 5, 80, 247, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 194, 5, 113, 247, 194, 5, 146, 247, 194, 5, 179, 247, 194, 5, 212, 247, 194, 5, 245, 247, 194, 5, 24, 248, 194, 5, 57, 248, 193, 2, 128, 11, 193, 9, 128, 11, 193, 16, 128, 11, 193, 23, 128, 11, 193, 30, 128, 11, 193, 207, 242, 12, 193, 214, 242, 12, 193, 221, 242, 12, 193, 228, 242, 12, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 194, 5, 245, 247, 194, 5, 113, 247, 194, 5, 179, 247, 194, 5, 212, 247, 194, 5, 24, 248, 194, 5, 57, 248, 194, 5, 146, 247, 193, 70, 244, 123, 193, 77, 244, 123, 193, 84, 244, 123, 193, 91, 244, 123, 193, 98, 244, 123, 193, 37, 100, 13, 193, 44, 100, 13, 255, 255, 255, 255, 255, 255, 255, 255, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 194, 5, 90, 248, 194, 5, 123, 248, 194, 5, 156, 248, 194, 5, 189, 248, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 174, 13, 193, 9, 78, 13, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 194, 5, 222, 248, 194, 5, 255, 248, 194, 5, 34, 249, 194, 5, 67, 249, 194, 5, 100, 249, 194, 5, 133, 249, 5, 5, 44, 5, 5, 5, 110, 13, 197, 66, 65, 0, 5, 5, 112, 13, 197, 130, 65, 0, 5, 5, 114, 13, 32, 5, 76, 13, 193, 16, 78, 13, 198, 65, 124, 0, 198, 225, 180, 0, 198, 1, 181, 0, 5, 5, 104, 13, 197, 130, 227, 2, 5, 5, 194, 10, 193, 23, 78, 13, 193, 30, 78, 13, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 194, 5, 186, 251, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 0, 4, 5, 5, 34, 7, 5, 5, 24, 9, 5, 5, 192, 10, 193, 182, 74, 13, 5, 5, 102, 13, 197, 194, 200, 2, 5, 5, 0, 19, 5, 5, 62, 10, 5, 5, 64, 10, 5, 5, 66, 10, 5, 5, 68, 10, 5, 5, 70, 10, 5, 5, 72, 10, 5, 5, 74, 10, 5, 5, 76, 10, 5, 5, 78, 10, 5, 5, 80, 10, 5, 5, 76, 13, 193, 2, 78, 13, 5, 5, 82, 10, 5, 5, 84, 10, 5, 5, 86, 10, 5, 5, 88, 10, 5, 5, 90, 10, 5, 5, 92, 10, 5, 5, 94, 10, 5, 5, 96, 10, 5, 5, 42, 5, 5, 5, 132, 9, 5, 5, 134, 9, 5, 5, 136, 9, 24, 5, 0, 4, 193, 9, 254, 118, 193, 16, 254, 118, 193, 23, 254, 118, 193, 30, 254, 118, 193, 37, 254, 118, 193, 44, 254, 118, 193, 51, 254, 118, 193, 58, 254, 118, 193, 65, 254, 118, 193, 72, 254, 118, 193, 79, 254, 118, 193, 86, 254, 118, 193, 93, 254, 118, 193, 100, 254, 118, 193, 107, 254, 118, 193, 114, 254, 118, 193, 121, 254, 118, 193, 128, 254, 118, 193, 135, 254, 118, 193, 142, 254, 118, 193, 149, 254, 118, 193, 156, 254, 118, 193, 163, 254, 118, 193, 170, 254, 118, 193, 177, 254, 118, 193, 184, 254, 118, 193, 2, 162, 9, 193, 9, 162, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 204, 10, 5, 5, 208, 10, 5, 5, 210, 10, 5, 5, 182, 10, 5, 5, 149, 12, 5, 5, 164, 9, 5, 5, 166, 9, 197, 98, 3, 0, 197, 130, 3, 0, 197, 66, 3, 0, 5, 5, 146, 10, 5, 5, 120, 10, 5, 5, 184, 10, 5, 5, 186, 10, 5, 5, 130, 10, 5, 5, 46, 7, 5, 5, 206, 10, 5, 5, 132, 10, 193, 2, 145, 12, 5, 5, 34, 5, 5, 5, 202, 10, 5, 5, 64, 9, 5, 5, 66, 9, 197, 36, 48, 0, 5, 5, 68, 9, 5, 5, 70, 9, 5, 5, 72, 9, 5, 5, 74, 9, 5, 5, 76, 9, 5, 5, 78, 9, 5, 5, 80, 9, 24, 5, 0, 4, 32, 5, 64, 10, 46, 5, 70, 10, 46, 5, 72, 10, 32, 5, 74, 10, 32, 5, 76, 10, 5, 5, 36, 7, 5, 5, 38, 7, 46, 5, 146, 9, 46, 5, 148, 9, 32, 5, 8, 5, 32, 5, 8, 5, 32, 5, 8, 5, 32, 5, 8, 5, 30, 5, 10, 5, 30, 5, 10, 5, 30, 5, 10, 5, 38, 5, 0, 6, 16, 5, 34, 7, 32, 5, 0, 8, 255, 255, 255, 255, 30, 5, 42, 7, 38, 5, 54, 7, 38, 5, 98, 7, 32, 5, 90, 7, 32, 5, 26, 5, 30, 5, 142, 9, 30, 5, 144, 9, 30, 5, 150, 9, 30, 5, 152, 9, 30, 5, 82, 10, 30, 5, 84, 10, 46, 5, 150, 10, 32, 5, 126, 10, 159, 5, 0, 42, 159, 5, 0, 44, 159, 5, 0, 46, 159, 5, 0, 48, 159, 5, 0, 50, 159, 5, 0, 52, 159, 5, 0, 54, 159, 5, 0, 56, 159, 5, 0, 58, 159, 5, 0, 60, 159, 5, 0, 62, 159, 5, 0, 64, 159, 5, 0, 66, 159, 5, 0, 68, 159, 5, 0, 70, 159, 5, 0, 72, 164, 5, 0, 74, 159, 5, 0, 76, 159, 5, 0, 78, 159, 5, 0, 80, 159, 5, 0, 82, 159, 5, 0, 84, 164, 5, 0, 86, 159, 5, 0, 88, 159, 5, 0, 90, 159, 5, 0, 92, 32, 5, 146, 9, 32, 5, 140, 10, 32, 5, 148, 9, 32, 5, 10, 12, 16, 5, 10, 5, 255, 255, 255, 255, 16, 5, 90, 7, 32, 5, 120, 9, 32, 5, 150, 10, 32, 5, 138, 13, 32, 5, 152, 10, 32, 5, 144, 10, 32, 5, 106, 9, 16, 5, 142, 9, 16, 5, 144, 9, 32, 5, 128, 10, 16, 5, 121, 12, 16, 5, 0, 6, 32, 5, 14, 5, 16, 5, 0, 8, 32, 5, 138, 10, 202, 224, 246, 2, 202, 1, 247, 2, 202, 34, 247, 2, 202, 67, 247, 2, 202, 100, 247, 2, 202, 133, 247, 2, 202, 166, 247, 2, 202, 199, 247, 2, 202, 232, 247, 2, 202, 9, 248, 2, 16, 5, 54, 7, 16, 5, 42, 7, 32, 5, 129, 12, 16, 5, 131, 12, 32, 5, 133, 12, 16, 5, 98, 7, 193, 2, 18, 5, 193, 9, 234, 118, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 5, 5, 42, 10, 5, 5, 44, 10, 5, 5, 46, 10, 5, 5, 48, 10, 5, 5, 50, 10, 5, 5, 52, 10, 5, 5, 54, 10, 5, 5, 56, 10, 5, 5, 58, 10, 5, 5, 60, 10, 5, 5, 82, 9, 5, 5, 84, 9, 5, 5, 86, 9, 5, 5, 88, 9, 5, 5, 102, 7, 193, 9, 224, 97, 5, 5, 18, 9, 5, 5, 28, 9, 5, 5, 8, 7, 5, 5, 30, 9, 5, 5, 6, 7, 5, 5, 48, 7, 5, 5, 168, 10, 5, 5, 170, 10, 5, 5, 172, 10, 5, 5, 116, 10, 5, 5, 30, 5, 5, 5, 32, 5, 5, 5, 20, 9, 5, 5, 90, 9, 5, 5, 104, 10, 5, 5, 124, 10, 5, 5, 40, 5, 5, 5, 10, 7, 5, 5, 130, 9, 5, 5, 36, 5, 5, 5, 250, 10, 5, 5, 252, 10, 5, 5, 254, 10, 5, 5, 2, 11, 5, 5, 4, 11, 5, 5, 52, 7, 5, 5, 142, 10, 5, 5, 174, 10, 5, 5, 12, 7, 5, 5, 122, 10, 5, 5, 14, 7, 5, 5, 16, 7, 193, 56, 202, 12, 193, 63, 202, 12, 5, 5, 148, 10, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 212, 10, 5, 5, 214, 10, 5, 5, 22, 10, 5, 5, 24, 10, 5, 5, 26, 10, 5, 5, 28, 10, 5, 5, 216, 10, 5, 5, 218, 10, 5, 5, 220, 10, 5, 5, 30, 10, 5, 5, 32, 10, 5, 5, 222, 10, 5, 5, 34, 10, 5, 5, 36, 10, 5, 5, 224, 10, 5, 5, 226, 10, 5, 5, 228, 10, 5, 5, 230, 10, 5, 5, 232, 10, 5, 5, 234, 10, 5, 5, 236, 10, 5, 5, 238, 10, 5, 5, 240, 10, 5, 5, 38, 5, 5, 5, 116, 7, 5, 5, 94, 9, 5, 5, 242, 10, 5, 5, 244, 10, 5, 5, 38, 10, 5, 5, 40, 10, 5, 5, 246, 10, 5, 5, 248, 10, 5, 5, 46, 5, 16, 5, 6, 122, 158, 5, 6, 122, 16, 5, 8, 122, 158, 5, 8, 122, 16, 5, 10, 122, 158, 5, 10, 122, 16, 5, 14, 122, 158, 5, 14, 122, 16, 5, 16, 122, 158, 5, 16, 122, 152, 5, 18, 122, 197, 162, 185, 2, 144, 5, 20, 122, 197, 130, 188, 2, 152, 5, 22, 122, 197, 162, 193, 2, 152, 5, 24, 122, 197, 34, 196, 2, 144, 5, 26, 122, 197, 2, 197, 2, 144, 5, 28, 122, 197, 226, 198, 2, 152, 5, 30, 122, 197, 130, 200, 2, 152, 5, 32, 122, 197, 226, 201, 2, 144, 5, 34, 122, 197, 98, 202, 2, 144, 5, 36, 122, 197, 162, 203, 2, 144, 5, 38, 122, 152, 5, 72, 122, 144, 5, 74, 122, 144, 5, 76, 122, 16, 5, 78, 122, 158, 5, 78, 122, 16, 5, 80, 122, 158, 5, 80, 122, 16, 5, 82, 122, 156, 5, 82, 122, 152, 5, 84, 122, 151, 5, 86, 122, 152, 5, 88, 122, 152, 5, 90, 122, 152, 5, 92, 122, 16, 5, 94, 122, 152, 5, 94, 122, 158, 5, 96, 122, 158, 5, 98, 122, 155, 5, 100, 122, 155, 5, 102, 122, 197, 194, 182, 2, 16, 5, 18, 122, 16, 5, 24, 122, 197, 2, 236, 2, 197, 66, 236, 2, 197, 130, 236, 2, 197, 194, 236, 2, 5, 5, 48, 5, 5, 5, 118, 13, 5, 5, 120, 13, 197, 2, 66, 0, 197, 2, 198, 2, 32, 5, 186, 9, 160, 5, 24, 9, 160, 5, 70, 10, 160, 5, 72, 10, 158, 5, 34, 7, 160, 5, 48, 5, 166, 5, 100, 122, 23, 5, 6, 122, 23, 5, 8, 122, 23, 5, 10, 122, 23, 5, 14, 122, 23, 5, 16, 122, 23, 5, 78, 122, 23, 5, 80, 122, 22, 5, 82, 122, 23, 5, 42, 122, 160, 5, 118, 13, 165, 5, 6, 122, 165, 5, 8, 122, 165, 5, 10, 122, 165, 5, 14, 122, 165, 5, 16, 122, 160, 5, 18, 122, 155, 5, 20, 122, 160, 5, 22, 122, 160, 5, 24, 122, 152, 5, 26, 122, 155, 5, 28, 122, 160, 5, 30, 122, 160, 5, 32, 122, 155, 5, 34, 122, 155, 5, 36, 122, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 193, 247, 151, 123, 193, 233, 151, 123, 193, 254, 151, 123, 193, 240, 151, 123, 193, 14, 152, 123, 193, 21, 152, 123, 193, 28, 152, 123, 193, 35, 152, 123, 193, 42, 152, 123, 193, 49, 152, 123, 193, 56, 152, 123, 193, 63, 152, 123, 193, 70, 152, 123, 193, 77, 152, 123, 193, 9, 151, 123, 193, 16, 151, 123, 193, 23, 151, 123, 193, 30, 151, 123, 193, 44, 151, 123, 193, 37, 151, 123, 193, 9, 32, 7, 193, 2, 22, 9, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 5, 5, 72, 120, 5, 5, 74, 120, 193, 16, 32, 7, 193, 9, 22, 9, 193, 23, 112, 7, 32, 5, 26, 120, 32, 5, 38, 120, 32, 5, 50, 120, 197, 66, 171, 2, 197, 130, 171, 2, 197, 194, 171, 2, 197, 2, 172, 2, 197, 66, 172, 2, 197, 130, 172, 2, 197, 194, 172, 2, 197, 2, 173, 2, 197, 66, 173, 2, 197, 130, 173, 2, 197, 194, 173, 2, 197, 2, 174, 2, 197, 66, 174, 2, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 193, 114, 172, 17, 193, 121, 172, 17, 193, 128, 172, 17, 193, 135, 172, 17, 193, 142, 172, 17, 193, 149, 172, 17, 193, 156, 172, 17, 193, 163, 172, 17, 193, 170, 172, 17, 193, 177, 172, 17, 5, 112, 0, 21, 5, 112, 0, 23, 5, 112, 0, 25, 193, 30, 32, 7, 193, 30, 22, 9, 193, 44, 128, 11, 193, 51, 128, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 197, 163, 2, 0, 197, 226, 54, 0, 197, 35, 55, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 5, 5, 57, 9, 5, 5, 62, 7, 5, 5, 12, 9, 5, 5, 64, 7, 5, 5, 66, 7, 5, 5, 68, 7, 5, 5, 70, 7, 5, 5, 110, 7, 5, 5, 59, 9, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 122, 16, 5, 5, 124, 16, 5, 5, 126, 16, 5, 5, 128, 16, 5, 5, 130, 16, 5, 5, 132, 16, 5, 5, 134, 16, 5, 5, 136, 16, 5, 5, 138, 16, 5, 5, 140, 16, 5, 5, 142, 16, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 210, 41, 118, 193, 217, 41, 118, 201, 64, 73, 1, 193, 231, 41, 118, 193, 238, 41, 118, 193, 114, 40, 118, 193, 121, 40, 118, 193, 233, 40, 118, 193, 21, 41, 118, 193, 84, 41, 118, 193, 112, 41, 118, 193, 119, 41, 118, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 193, 16, 61, 9, 193, 23, 61, 9, 193, 16, 16, 9, 193, 2, 80, 7, 193, 30, 52, 9, 193, 37, 52, 9, 193, 58, 51, 118, 193, 30, 61, 9, 193, 37, 61, 9, 193, 44, 61, 9, 193, 51, 61, 9, 193, 58, 61, 9, 193, 65, 61, 9, 193, 9, 80, 7, 193, 44, 52, 9, 193, 51, 52, 9, 193, 72, 61, 9, 193, 79, 61, 9, 193, 86, 61, 9, 193, 93, 61, 9, 255, 255, 255, 255, 5, 5, 94, 13, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 2, 122, 11, 193, 9, 122, 11, 193, 58, 56, 119, 193, 86, 56, 119, 193, 93, 56, 119, 193, 65, 56, 119, 193, 79, 56, 119, 193, 100, 56, 119, 5, 5, 58, 119, 5, 112, 58, 119, 5, 120, 58, 119, 5, 112, 20, 119, 5, 115, 40, 119, 5, 5, 82, 7, 5, 5, 84, 7, 5, 5, 86, 7, 197, 194, 170, 2, 197, 2, 171, 2, 197, 130, 170, 2, 193, 2, 18, 119, 5, 5, 42, 119, 193, 16, 56, 119, 193, 2, 14, 119, 5, 5, 28, 119, 193, 9, 48, 119, 193, 30, 56, 119, 193, 44, 56, 119, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 3, 88, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 96, 7, 193, 2, 112, 7, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 193, 205, 160, 113, 193, 212, 160, 113, 193, 219, 160, 113, 193, 226, 160, 113, 193, 233, 160, 113, 193, 240, 160, 113, 193, 247, 160, 113, 193, 254, 160, 113, 193, 7, 161, 113, 193, 14, 161, 113, 193, 21, 161, 113, 193, 84, 161, 113, 193, 91, 161, 113, 193, 98, 161, 113, 193, 105, 161, 113, 193, 112, 161, 113, 193, 79, 52, 9, 193, 37, 112, 7, 193, 9, 160, 113, 193, 9, 100, 13, 193, 16, 100, 13, 193, 119, 161, 113, 193, 189, 161, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 212, 61, 9, 193, 135, 52, 9, 193, 142, 52, 9, 193, 44, 112, 7, 193, 14, 97, 117, 193, 105, 97, 117, 193, 112, 97, 117, 193, 240, 96, 117, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 32, 237, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 193, 250, 87, 12, 193, 9, 16, 9, 193, 129, 244, 118, 193, 248, 244, 118, 193, 57, 245, 118, 193, 64, 245, 118, 193, 71, 245, 118, 193, 78, 245, 118, 193, 85, 245, 118, 193, 92, 245, 118, 193, 87, 251, 118, 193, 94, 251, 118, 193, 101, 251, 118, 193, 108, 251, 118, 193, 115, 251, 118, 193, 122, 251, 118, 193, 129, 251, 118, 193, 136, 251, 118, 193, 143, 251, 118, 5, 5, 38, 104, 5, 5, 42, 104, 5, 5, 182, 104, 5, 5, 184, 104, 5, 5, 32, 9, 5, 5, 34, 9, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 28, 11, 5, 5, 84, 13, 5, 5, 10, 104, 5, 5, 18, 104, 5, 5, 20, 104, 5, 5, 22, 104, 5, 5, 24, 104, 5, 5, 26, 104, 5, 5, 92, 104, 5, 5, 80, 104, 5, 5, 126, 104, 5, 5, 68, 104, 5, 5, 82, 104, 5, 5, 146, 104, 5, 5, 96, 104, 5, 5, 118, 104, 193, 91, 9, 117, 193, 98, 9, 117, 193, 49, 9, 117, 193, 56, 9, 117, 193, 63, 9, 117, 193, 70, 9, 117, 193, 105, 9, 117, 193, 177, 8, 117, 193, 212, 8, 117, 193, 65, 8, 117, 193, 72, 8, 117, 193, 79, 8, 117, 193, 86, 8, 117, 193, 219, 8, 117, 193, 7, 9, 117, 193, 28, 9, 117, 193, 35, 9, 117, 193, 233, 8, 117, 193, 240, 8, 117, 193, 112, 9, 117, 193, 128, 118, 11, 193, 135, 118, 11, 193, 30, 36, 9, 193, 37, 36, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 26, 117, 193, 16, 26, 117, 193, 23, 26, 117, 193, 30, 26, 117, 193, 37, 26, 117, 193, 44, 26, 117, 193, 51, 26, 117, 193, 58, 26, 117, 193, 65, 26, 117, 193, 72, 26, 117, 193, 79, 26, 117, 193, 86, 26, 117, 193, 93, 26, 117, 193, 100, 26, 117, 193, 107, 26, 117, 193, 114, 26, 117, 193, 121, 26, 117, 193, 128, 26, 117, 193, 135, 26, 117, 193, 142, 26, 117, 193, 149, 26, 117, 5, 5, 38, 9, 5, 5, 40, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 44, 9, 5, 5, 46, 9, 5, 5, 104, 11, 5, 5, 106, 11, 5, 5, 108, 11, 5, 5, 110, 11, 193, 131, 81, 117, 193, 138, 81, 117, 193, 142, 85, 117, 193, 149, 85, 117, 193, 156, 85, 117, 193, 163, 85, 117, 193, 79, 89, 117, 193, 86, 89, 117, 193, 93, 89, 117, 193, 100, 89, 117, 193, 93, 79, 117, 193, 191, 79, 117, 193, 65, 85, 117, 193, 72, 85, 117, 193, 189, 80, 117, 193, 40, 81, 117, 193, 140, 39, 118, 193, 147, 39, 118, 193, 154, 39, 118, 193, 161, 39, 118, 193, 168, 39, 118, 193, 175, 39, 118, 193, 182, 39, 118, 193, 189, 39, 118, 193, 196, 39, 118, 193, 203, 39, 118, 193, 210, 39, 118, 193, 217, 39, 118, 193, 224, 39, 118, 193, 231, 39, 118, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 193, 16, 122, 11, 193, 58, 52, 9, 193, 65, 52, 9, 193, 72, 52, 9, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 193, 168, 147, 17, 193, 175, 147, 17, 193, 182, 147, 17, 193, 189, 147, 17, 193, 196, 147, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 65, 122, 11, 193, 72, 122, 11, 193, 79, 122, 11, 193, 86, 122, 11, 193, 93, 122, 11, 193, 100, 122, 11, 193, 93, 52, 9, 193, 100, 52, 9, 193, 107, 122, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 193, 107, 52, 9, 193, 114, 52, 9, 193, 30, 122, 11, 193, 37, 122, 11, 193, 44, 122, 11, 193, 51, 122, 11, 193, 58, 122, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 206, 192, 236, 0, 206, 192, 236, 0, 206, 192, 236, 0, 206, 192, 236, 0, 206, 192, 236, 0, 193, 23, 223, 113, 193, 30, 223, 113, 193, 37, 223, 113, 193, 44, 223, 113, 193, 51, 223, 113, 193, 58, 223, 113, 193, 65, 223, 113, 193, 72, 223, 113, 193, 79, 223, 113, 193, 86, 223, 113, 193, 93, 223, 113, 193, 100, 223, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 193, 77, 147, 17, 193, 84, 147, 17, 193, 219, 52, 9, 193, 226, 52, 9, 193, 247, 61, 9, 193, 45, 88, 12, 193, 77, 249, 113, 193, 240, 52, 9, 193, 247, 52, 9, 193, 114, 122, 11, 193, 121, 122, 11, 193, 128, 122, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 254, 52, 9, 193, 7, 53, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 193, 14, 53, 9, 193, 21, 53, 9, 5, 5, 78, 99, 5, 5, 82, 99, 5, 5, 86, 99, 5, 5, 94, 99, 5, 5, 98, 99, 5, 5, 102, 99, 5, 5, 106, 99, 5, 5, 110, 99, 5, 5, 114, 99, 5, 5, 118, 99, 5, 5, 122, 99, 5, 5, 126, 99, 5, 5, 130, 99, 5, 5, 134, 99, 5, 5, 138, 99, 5, 5, 146, 99, 5, 5, 150, 99, 5, 5, 34, 99, 5, 5, 62, 99, 5, 5, 90, 99, 5, 5, 142, 99, 5, 5, 154, 99, 5, 5, 158, 99, 5, 5, 160, 99, 5, 5, 164, 99, 5, 5, 166, 99, 5, 5, 168, 99, 5, 5, 55, 9, 46, 5, 58, 99, 5, 5, 170, 99, 5, 5, 174, 99, 5, 5, 176, 99, 193, 9, 29, 117, 193, 16, 29, 117, 193, 23, 29, 117, 193, 30, 29, 117, 193, 37, 29, 117, 193, 44, 29, 117, 193, 51, 29, 117, 193, 58, 29, 117, 193, 65, 29, 117, 193, 72, 29, 117, 193, 79, 29, 117, 193, 86, 29, 117, 193, 93, 29, 117, 193, 100, 29, 117, 193, 107, 29, 117, 193, 114, 29, 117, 193, 121, 29, 117, 193, 128, 29, 117, 193, 135, 29, 117, 193, 142, 29, 117, 193, 149, 29, 117, 193, 156, 29, 117, 193, 163, 29, 117, 193, 170, 29, 117, 193, 177, 29, 117, 193, 184, 29, 117, 193, 191, 29, 117, 193, 198, 29, 117, 255, 255, 255, 255, 255, 255, 255, 255, 193, 2, 61, 9, 193, 9, 61, 9, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 206, 160, 233, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 107, 61, 9, 193, 9, 158, 124, 193, 16, 158, 124, 193, 23, 158, 124, 193, 30, 158, 124, 193, 37, 158, 124, 193, 44, 158, 124, 193, 51, 158, 124, 193, 58, 158, 124, 193, 65, 158, 124, 193, 72, 158, 124, 193, 79, 158, 124, 193, 86, 158, 124, 193, 93, 158, 124, 193, 100, 158, 124, 193, 107, 158, 124, 193, 114, 158, 124, 193, 121, 158, 124, 193, 128, 158, 124, 193, 135, 158, 124, 193, 142, 158, 124, 193, 149, 158, 124, 193, 156, 158, 124, 255, 255, 255, 255, 193, 114, 61, 9, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 193, 121, 146, 17, 193, 128, 146, 17, 193, 135, 146, 17, 193, 142, 146, 17, 193, 149, 146, 17, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 5, 5, 122, 124, 5, 112, 122, 124, 193, 2, 124, 124, 193, 9, 124, 124, 193, 16, 124, 124, 193, 23, 124, 124, 193, 30, 124, 124, 193, 37, 124, 124, 193, 44, 124, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 135, 122, 11, 193, 121, 61, 9, 193, 128, 61, 9, 193, 135, 61, 9, 193, 142, 61, 9, 193, 149, 61, 9, 193, 156, 61, 9, 193, 219, 176, 113, 193, 226, 176, 113, 193, 233, 176, 113, 193, 240, 176, 113, 193, 247, 176, 113, 193, 37, 176, 113, 193, 44, 176, 113, 193, 254, 176, 113, 193, 7, 177, 113, 193, 240, 61, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 206, 128, 237, 0, 193, 9, 30, 117, 193, 16, 30, 117, 193, 23, 30, 117, 193, 30, 30, 117, 193, 37, 30, 117, 193, 44, 30, 117, 193, 51, 30, 117, 193, 58, 30, 117, 193, 65, 30, 117, 193, 72, 30, 117, 193, 79, 30, 117, 193, 86, 30, 117, 193, 93, 30, 117, 193, 100, 30, 117, 193, 107, 30, 117, 193, 114, 30, 117, 193, 121, 30, 117, 193, 128, 30, 117, 193, 135, 30, 117, 193, 142, 30, 117, 193, 149, 30, 117, 193, 156, 30, 117, 193, 163, 30, 117, 193, 7, 62, 9, 193, 14, 62, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 250, 123, 193, 16, 250, 123, 193, 23, 250, 123, 193, 30, 250, 123, 193, 37, 250, 123, 193, 44, 250, 123, 193, 51, 250, 123, 193, 58, 250, 123, 193, 65, 250, 123, 193, 72, 250, 123, 193, 79, 250, 123, 193, 86, 250, 123, 193, 93, 250, 123, 193, 100, 250, 123, 193, 107, 250, 123, 193, 114, 250, 123, 193, 121, 250, 123, 193, 128, 250, 123, 193, 135, 250, 123, 193, 142, 250, 123, 193, 149, 250, 123, 193, 156, 250, 123, 193, 163, 250, 123, 193, 170, 250, 123, 193, 177, 250, 123, 193, 184, 250, 123, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 96, 9, 5, 5, 98, 9, 5, 5, 100, 9, 5, 5, 102, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 202, 16, 5, 5, 204, 16, 5, 5, 206, 16, 5, 5, 208, 16, 5, 5, 210, 16, 5, 5, 212, 16, 5, 5, 214, 16, 5, 5, 216, 16, 5, 5, 218, 16, 5, 5, 220, 16, 5, 5, 222, 16, 5, 5, 224, 16, 5, 5, 226, 16, 5, 5, 228, 16, 5, 5, 230, 16, 5, 5, 232, 16, 193, 9, 204, 124, 193, 16, 204, 124, 193, 23, 204, 124, 193, 30, 204, 124, 193, 37, 204, 124, 193, 44, 204, 124, 193, 51, 204, 124, 193, 58, 204, 124, 193, 65, 204, 124, 193, 72, 204, 124, 193, 79, 204, 124, 193, 86, 204, 124, 193, 93, 204, 124, 193, 100, 204, 124, 193, 107, 204, 124, 193, 114, 204, 124, 193, 121, 204, 124, 193, 128, 204, 124, 193, 135, 204, 124, 193, 142, 204, 124, 193, 149, 204, 124, 193, 156, 204, 124, 193, 163, 204, 124, 193, 170, 204, 124, 193, 177, 204, 124, 193, 184, 204, 124, 193, 191, 204, 124, 193, 198, 204, 124, 193, 205, 204, 124, 193, 212, 204, 124, 255, 255, 255, 255, 193, 2, 104, 9, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 7, 206, 124, 193, 14, 206, 124, 193, 21, 206, 124, 193, 28, 206, 124, 193, 35, 206, 124, 193, 42, 206, 124, 193, 49, 206, 124, 193, 56, 206, 124, 193, 9, 104, 9, 5, 5, 0, 21, 5, 5, 0, 23, 193, 2, 146, 17, 193, 9, 146, 17, 193, 16, 146, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 50, 101, 193, 16, 50, 101, 193, 23, 50, 101, 193, 30, 50, 101, 193, 37, 50, 101, 193, 44, 50, 101, 193, 51, 50, 101, 193, 58, 50, 101, 193, 65, 50, 101, 193, 72, 50, 101, 193, 79, 50, 101, 193, 86, 50, 101, 193, 93, 50, 101, 193, 100, 50, 101, 193, 107, 50, 101, 193, 114, 50, 101, 193, 121, 50, 101, 193, 128, 50, 101, 193, 135, 50, 101, 193, 142, 50, 101, 193, 149, 50, 101, 193, 156, 50, 101, 5, 5, 0, 21, 193, 100, 146, 17, 193, 107, 146, 17, 193, 114, 146, 17, 5, 5, 0, 23, 5, 5, 0, 25, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 16, 104, 9, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 255, 255, 255, 255, 193, 23, 104, 9, 193, 30, 104, 9, 193, 37, 104, 9, 193, 44, 104, 9, 193, 51, 104, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 41, 5, 0, 19, 41, 5, 0, 21, 41, 5, 0, 23, 41, 5, 0, 25, 41, 5, 0, 27, 41, 5, 0, 29, 41, 5, 0, 31, 41, 5, 0, 33, 41, 5, 0, 35, 46, 5, 0, 37, 48, 5, 121, 12, 46, 5, 143, 12, 48, 5, 131, 12, 44, 5, 142, 9, 44, 5, 144, 9, 255, 255, 255, 255, 46, 5, 0, 42, 46, 5, 0, 50, 46, 5, 0, 70, 46, 5, 0, 88, 49, 5, 24, 51, 46, 5, 0, 56, 46, 5, 0, 62, 46, 5, 0, 64, 46, 5, 0, 66, 46, 5, 0, 68, 46, 5, 0, 72, 46, 5, 0, 78, 46, 5, 0, 80, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 31, 5, 0, 21, 31, 5, 0, 23, 31, 5, 0, 25, 31, 5, 0, 27, 31, 5, 0, 29, 31, 5, 0, 31, 31, 5, 0, 33, 31, 5, 0, 35, 34, 5, 0, 37, 197, 98, 71, 0, 197, 2, 72, 0, 197, 162, 72, 0, 197, 194, 73, 0, 197, 98, 74, 0, 197, 2, 75, 0, 197, 162, 75, 0, 197, 66, 76, 0, 197, 226, 76, 0, 197, 130, 77, 0, 197, 34, 79, 0, 197, 3, 5, 0, 197, 99, 10, 0, 197, 67, 11, 0, 197, 163, 11, 0, 197, 3, 12, 0, 197, 99, 12, 0, 197, 195, 12, 0, 197, 35, 13, 0, 197, 131, 13, 0, 197, 100, 5, 0, 197, 228, 5, 0, 197, 100, 6, 0, 197, 228, 6, 0, 197, 100, 7, 0, 197, 228, 7, 0, 197, 100, 8, 0, 197, 228, 8, 0, 197, 100, 9, 0, 197, 228, 9, 0, 197, 196, 10, 0, 197, 98, 67, 0, 197, 2, 73, 0, 197, 34, 74, 0, 197, 194, 74, 0, 197, 98, 75, 0, 197, 2, 76, 0, 197, 162, 76, 0, 197, 66, 77, 0, 197, 226, 77, 0, 197, 163, 71, 0, 197, 67, 72, 0, 197, 227, 72, 0, 197, 3, 74, 0, 197, 163, 74, 0, 197, 67, 75, 0, 197, 227, 75, 0, 197, 131, 76, 0, 197, 35, 77, 0, 197, 195, 77, 0, 197, 99, 79, 0, 197, 227, 13, 0, 197, 163, 14, 0, 197, 99, 15, 0, 197, 35, 16, 0, 197, 130, 66, 0, 197, 66, 66, 0, 197, 34, 67, 0, 197, 34, 78, 0, 197, 162, 83, 0, 197, 130, 87, 0, 197, 98, 90, 0, 197, 2, 92, 0, 197, 226, 92, 0, 197, 34, 94, 0, 197, 226, 94, 0, 31, 5, 0, 19, 31, 5, 0, 19, 193, 168, 11, 13, 193, 175, 11, 13, 193, 182, 11, 13, 197, 67, 14, 0, 197, 3, 15, 0, 197, 195, 15, 0, 197, 131, 16, 0, 197, 67, 17, 0, 197, 3, 18, 0, 197, 195, 18, 0, 197, 131, 19, 0, 197, 67, 20, 0, 197, 3, 21, 0, 197, 195, 21, 0, 197, 131, 22, 0, 197, 67, 23, 0, 197, 3, 24, 0, 197, 195, 24, 0, 197, 131, 25, 0, 197, 227, 16, 0, 197, 163, 17, 0, 197, 99, 18, 0, 197, 35, 19, 0, 197, 227, 19, 0, 197, 163, 20, 0, 197, 99, 21, 0, 197, 35, 22, 0, 197, 227, 22, 0, 197, 163, 23, 0, 197, 99, 24, 0, 197, 35, 25, 0, 197, 227, 25, 0, 197, 163, 26, 0, 197, 99, 27, 0, 197, 35, 28, 0, 197, 227, 28, 0, 197, 163, 29, 0, 197, 99, 30, 0, 197, 35, 31, 0, 197, 227, 31, 0, 197, 163, 32, 0, 168, 5, 0, 42, 168, 5, 0, 44, 168, 5, 0, 46, 168, 5, 0, 48, 168, 5, 0, 50, 168, 5, 0, 52, 168, 5, 0, 54, 168, 5, 0, 56, 168, 5, 0, 58, 168, 5, 0, 60, 197, 67, 26, 0, 197, 3, 27, 0, 197, 195, 27, 0, 197, 131, 28, 0, 197, 67, 29, 0, 197, 3, 30, 0, 197, 195, 30, 0, 197, 131, 31, 0, 197, 67, 32, 0, 197, 3, 33, 0, 197, 195, 47, 0, 168, 5, 0, 46, 168, 5, 0, 76, 197, 226, 116, 0, 197, 98, 210, 0, 5, 5, 93, 12, 180, 5, 0, 42, 177, 5, 0, 44, 177, 5, 0, 46, 177, 5, 0, 48, 180, 5, 0, 50, 177, 5, 0, 52, 177, 5, 0, 54, 180, 5, 0, 56, 180, 5, 0, 58, 177, 5, 0, 60, 180, 5, 0, 62, 180, 5, 0, 64, 180, 5, 0, 66, 180, 5, 0, 68, 180, 5, 0, 70, 180, 5, 0, 72, 197, 99, 33, 0, 197, 67, 34, 0, 197, 35, 35, 0, 197, 3, 36, 0, 197, 227, 36, 0, 197, 195, 37, 0, 197, 163, 38, 0, 197, 131, 39, 0, 197, 3, 42, 0, 197, 99, 43, 0, 197, 67, 44, 0, 197, 35, 45, 0, 197, 3, 46, 0, 197, 227, 46, 0, 197, 196, 33, 0, 197, 164, 34, 0, 197, 132, 35, 0, 197, 100, 36, 0, 197, 68, 37, 0, 197, 36, 38, 0, 197, 4, 39, 0, 197, 228, 39, 0, 197, 100, 42, 0, 197, 196, 43, 0, 197, 164, 44, 0, 197, 132, 45, 0, 197, 100, 46, 0, 197, 68, 47, 0, 197, 228, 42, 0, 197, 103, 40, 0, 197, 70, 41, 0, 255, 255, 255, 255, 198, 163, 119, 0, 198, 35, 121, 0, 198, 99, 120, 0, 198, 163, 125, 0, 198, 131, 121, 0, 198, 99, 123, 0, 198, 3, 120, 0, 198, 3, 123, 0, 198, 195, 120, 0, 198, 35, 124, 0, 198, 35, 127, 0, 198, 3, 129, 0, 198, 163, 128, 0, 198, 227, 127, 0, 198, 195, 132, 0, 198, 3, 126, 0, 198, 195, 126, 0, 198, 67, 128, 0, 198, 131, 127, 0, 198, 35, 130, 0, 198, 227, 124, 0, 198, 99, 129, 0, 198, 3, 132, 0, 198, 131, 130, 0, 198, 195, 123, 0, 198, 227, 121, 0, 198, 67, 125, 0, 198, 99, 126, 0, 198, 195, 129, 0, 198, 67, 122, 0, 198, 99, 132, 0, 198, 131, 124, 0, 198, 227, 130, 0, 198, 163, 122, 0, 198, 67, 131, 0, 198, 163, 131, 0, 198, 97, 183, 0, 198, 97, 189, 0, 198, 1, 193, 0, 198, 193, 205, 0, 197, 98, 71, 0, 197, 34, 79, 0, 197, 2, 85, 0, 197, 34, 88, 0, 197, 194, 91, 0, 197, 66, 92, 0, 197, 226, 93, 0, 197, 98, 94, 0, 197, 131, 183, 0, 197, 194, 79, 0, 197, 2, 80, 0, 197, 194, 80, 0, 197, 2, 81, 0, 197, 66, 81, 0, 197, 130, 81, 0, 197, 194, 81, 0, 197, 2, 82, 0, 197, 66, 82, 0, 197, 2, 85, 0, 197, 66, 85, 0, 197, 226, 80, 0, 197, 130, 85, 0, 197, 194, 85, 0, 197, 2, 86, 0, 32, 5, 4, 12, 16, 5, 0, 42, 16, 5, 0, 44, 16, 5, 0, 46, 16, 5, 0, 48, 16, 5, 0, 50, 16, 5, 0, 52, 16, 5, 0, 54, 16, 5, 0, 56, 16, 5, 0, 58, 16, 5, 0, 60, 16, 5, 0, 62, 16, 5, 0, 64, 16, 5, 0, 66, 16, 5, 0, 68, 16, 5, 0, 70, 16, 5, 0, 72, 16, 5, 0, 74, 16, 5, 0, 76, 16, 5, 0, 78, 16, 5, 0, 80, 16, 5, 0, 82, 16, 5, 0, 84, 16, 5, 0, 86, 16, 5, 0, 88, 16, 5, 0, 90, 16, 5, 0, 92, 16, 5, 150, 9, 32, 5, 137, 12, 16, 5, 152, 9, 32, 5, 141, 12, 32, 5, 184, 9, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 5, 5, 168, 9, 5, 5, 170, 9, 5, 5, 172, 9, 5, 5, 174, 9, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 5, 5, 176, 9, 5, 5, 178, 9, 193, 27, 213, 12, 193, 34, 213, 12, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 5, 5, 180, 9, 5, 5, 182, 9, 5, 5, 184, 9, 5, 5, 186, 9, 5, 5, 188, 9, 5, 5, 190, 9, 5, 5, 192, 9, 5, 5, 194, 9, 5, 5, 196, 9, 5, 5, 198, 9, 5, 5, 200, 9, 5, 5, 202, 9, 5, 5, 204, 9, 5, 5, 206, 9, 5, 5, 208, 9, 5, 5, 210, 9, 5, 5, 212, 9, 5, 5, 214, 9, 5, 5, 216, 9, 5, 5, 218, 9, 5, 5, 220, 9, 5, 5, 222, 9, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 5, 5, 224, 9, 5, 5, 226, 9, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 206, 96, 230, 0, 5, 5, 228, 9, 5, 5, 230, 9, 5, 5, 232, 9, 5, 5, 234, 9, 5, 5, 236, 9, 5, 5, 238, 9, 5, 5, 240, 9, 5, 5, 242, 9, 5, 5, 244, 9, 5, 5, 246, 9, 193, 111, 206, 12, 193, 118, 206, 12, 193, 125, 206, 12, 193, 132, 206, 12, 193, 139, 206, 12, 193, 146, 206, 12, 193, 153, 206, 12, 193, 160, 206, 12, 193, 167, 206, 12, 193, 174, 206, 12, 193, 181, 206, 12, 193, 188, 206, 12, 193, 195, 206, 12, 193, 202, 206, 12, 193, 209, 206, 12, 193, 216, 206, 12, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 5, 5, 248, 9, 5, 5, 250, 9, 5, 5, 252, 9, 5, 5, 254, 9, 5, 5, 2, 10, 5, 5, 4, 10, 5, 5, 6, 10, 5, 5, 8, 10, 5, 5, 10, 10, 5, 5, 12, 10, 5, 5, 14, 10, 5, 5, 16, 10, 5, 5, 18, 10, 5, 5, 20, 10, 31, 5, 0, 21, 31, 5, 0, 23, 31, 5, 0, 25, 31, 5, 0, 27, 31, 5, 0, 29, 31, 5, 0, 31, 31, 5, 0, 33, 31, 5, 0, 35, 34, 5, 0, 37, 197, 98, 71, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 206, 0, 229, 0, 5, 5, 62, 10, 5, 5, 64, 10, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 198, 99, 135, 0, 198, 35, 133, 0, 198, 131, 133, 0, 198, 67, 134, 0, 198, 195, 135, 0, 198, 163, 134, 0, 198, 35, 136, 0, 198, 227, 133, 0, 198, 3, 135, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 198, 225, 190, 0, 198, 161, 182, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 34, 73, 1, 197, 226, 123, 1, 197, 98, 124, 1, 197, 226, 128, 1, 197, 98, 129, 1, 197, 66, 108, 1, 197, 162, 109, 1, 197, 194, 111, 1, 197, 66, 113, 1, 197, 66, 112, 1, 197, 34, 103, 1, 197, 162, 118, 1, 197, 226, 127, 1, 197, 194, 107, 1, 197, 34, 109, 1, 197, 66, 111, 1, 197, 194, 112, 1, 197, 66, 106, 1, 197, 66, 115, 1, 197, 98, 130, 1, 197, 66, 99, 1, 197, 194, 100, 1, 197, 226, 101, 1, 197, 2, 108, 1, 197, 98, 109, 1, 197, 130, 111, 1, 197, 162, 130, 1, 197, 226, 133, 1, 197, 98, 69, 1, 197, 162, 69, 1, 5, 5, 98, 10, 5, 5, 100, 10, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 5, 5, 106, 10, 5, 5, 108, 10, 5, 5, 110, 10, 5, 5, 112, 10, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 206, 224, 230, 0, 193, 44, 26, 11, 193, 51, 26, 11, 193, 58, 26, 11, 193, 65, 26, 11, 193, 72, 26, 11, 193, 79, 26, 11, 193, 86, 26, 11, 193, 93, 26, 11, 193, 100, 26, 11, 193, 107, 26, 11, 193, 114, 26, 11, 193, 121, 26, 11, 193, 128, 26, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 20, 110, 5, 5, 24, 110, 5, 5, 132, 110, 5, 5, 134, 110, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 44, 11, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 193, 182, 86, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 140, 113, 5, 5, 144, 113, 5, 5, 48, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 102, 11, 193, 16, 102, 11, 193, 23, 102, 11, 193, 30, 102, 11, 193, 37, 102, 11, 193, 44, 102, 11, 193, 51, 102, 11, 193, 114, 208, 116, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 209, 116, 193, 16, 209, 116, 193, 23, 209, 116, 193, 30, 209, 116, 193, 51, 209, 116, 193, 65, 209, 116, 193, 58, 209, 116, 193, 72, 209, 116, 193, 79, 209, 116, 193, 37, 209, 116, 193, 44, 209, 116, 193, 86, 209, 116, 5, 5, 211, 116, 5, 5, 215, 116, 193, 2, 217, 116, 193, 9, 217, 116, 193, 93, 102, 11, 193, 100, 102, 11, 193, 107, 102, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 193, 203, 147, 17, 193, 210, 147, 17, 193, 217, 147, 17, 193, 224, 147, 17, 193, 231, 147, 17, 193, 238, 147, 17, 193, 245, 147, 17, 193, 252, 147, 17, 193, 5, 148, 17, 193, 12, 148, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 114, 102, 11, 193, 121, 102, 11, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 23, 122, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 161, 124, 193, 16, 161, 124, 193, 23, 161, 124, 193, 30, 161, 124, 193, 37, 161, 124, 193, 44, 161, 124, 193, 51, 161, 124, 193, 58, 161, 124, 193, 65, 161, 124, 193, 72, 161, 124, 193, 79, 161, 124, 193, 86, 161, 124, 193, 93, 161, 124, 193, 100, 161, 124, 193, 107, 161, 124, 193, 114, 161, 124, 193, 121, 161, 124, 193, 128, 161, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 191, 122, 11, 193, 198, 122, 11, 193, 205, 122, 11, 193, 212, 122, 11, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 182, 208, 113, 193, 84, 208, 113, 193, 12, 124, 11, 193, 91, 208, 113, 193, 154, 208, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 168, 13, 193, 63, 86, 12, 193, 70, 86, 12, 193, 77, 86, 12, 193, 84, 86, 12, 193, 91, 86, 12, 193, 98, 86, 12, 193, 105, 86, 12, 193, 112, 86, 12, 193, 119, 86, 12, 193, 126, 86, 12, 193, 133, 86, 12, 193, 140, 86, 12, 193, 147, 86, 12, 193, 154, 86, 12, 193, 161, 86, 12, 193, 168, 86, 12, 193, 175, 86, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 126, 11, 193, 91, 241, 123, 193, 98, 241, 123, 193, 23, 100, 13, 193, 30, 100, 13, 193, 37, 128, 11, 193, 235, 242, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 193, 65, 172, 17, 193, 72, 172, 17, 193, 79, 172, 17, 193, 86, 172, 17, 193, 93, 172, 17, 193, 9, 117, 124, 193, 16, 117, 124, 193, 23, 117, 124, 193, 30, 117, 124, 193, 37, 117, 124, 193, 44, 117, 124, 193, 51, 117, 124, 193, 58, 117, 124, 193, 65, 117, 124, 193, 72, 117, 124, 193, 79, 117, 124, 193, 86, 117, 124, 193, 93, 117, 124, 193, 100, 117, 124, 193, 107, 117, 124, 193, 114, 117, 124, 193, 121, 117, 124, 193, 128, 117, 124, 193, 135, 117, 124, 193, 142, 117, 124, 193, 149, 117, 124, 193, 156, 117, 124, 193, 163, 117, 124, 193, 170, 117, 124, 193, 177, 117, 124, 193, 184, 117, 124, 193, 191, 117, 124, 193, 198, 117, 124, 193, 205, 117, 124, 5, 5, 0, 21, 193, 79, 146, 17, 193, 93, 128, 11, 197, 226, 17, 1, 197, 130, 20, 1, 197, 195, 19, 1, 197, 99, 19, 1, 197, 194, 11, 1, 197, 2, 12, 1, 197, 98, 18, 1, 197, 35, 20, 1, 197, 34, 18, 1, 197, 194, 20, 1, 197, 162, 17, 1, 197, 226, 16, 1, 197, 66, 12, 1, 197, 162, 49, 0, 197, 98, 49, 0, 5, 5, 4, 12, 255, 255, 255, 255, 255, 255, 255, 255, 197, 99, 35, 1, 197, 98, 36, 1, 197, 131, 34, 1, 255, 255, 255, 255, 197, 194, 35, 1, 197, 3, 36, 1, 197, 130, 11, 1, 197, 2, 11, 1, 197, 34, 35, 1, 197, 66, 34, 1, 197, 162, 36, 1, 5, 5, 6, 12, 193, 16, 22, 12, 255, 255, 255, 255, 193, 23, 22, 12, 197, 226, 49, 0, 197, 35, 249, 0, 197, 34, 250, 0, 197, 67, 248, 0, 255, 255, 255, 255, 197, 130, 249, 0, 197, 195, 249, 0, 197, 130, 234, 0, 197, 2, 234, 0, 197, 226, 248, 0, 197, 2, 248, 0, 197, 98, 250, 0, 197, 98, 50, 0, 197, 34, 50, 0, 197, 162, 50, 0, 197, 34, 1, 1, 197, 130, 3, 1, 197, 195, 2, 1, 197, 99, 2, 1, 255, 255, 255, 255, 255, 255, 255, 255, 197, 162, 1, 1, 197, 35, 3, 1, 197, 98, 1, 1, 197, 194, 3, 1, 197, 226, 0, 1, 197, 98, 0, 1, 255, 255, 255, 255, 197, 34, 51, 0, 197, 226, 50, 0, 197, 98, 51, 0, 197, 195, 26, 1, 197, 67, 33, 1, 197, 4, 24, 1, 197, 132, 30, 1, 197, 68, 22, 1, 197, 196, 28, 1, 197, 196, 25, 1, 197, 68, 32, 1, 197, 35, 27, 1, 197, 163, 33, 1, 197, 132, 24, 1, 197, 4, 31, 1, 197, 196, 22, 1, 197, 68, 29, 1, 197, 68, 26, 1, 197, 196, 32, 1, 197, 162, 227, 0, 197, 194, 228, 0, 197, 67, 227, 0, 197, 66, 229, 0, 197, 99, 226, 0, 255, 255, 255, 255, 197, 34, 228, 0, 197, 99, 228, 0, 197, 226, 227, 0, 197, 2, 229, 0, 197, 2, 227, 0, 197, 34, 226, 0, 197, 130, 229, 0, 193, 9, 22, 12, 5, 5, 32, 96, 193, 9, 22, 12, 5, 5, 48, 93, 5, 5, 50, 93, 5, 5, 40, 71, 5, 5, 52, 93, 5, 5, 30, 43, 5, 5, 88, 96, 197, 162, 123, 0, 197, 98, 195, 0, 5, 5, 56, 77, 46, 5, 12, 87, 193, 30, 52, 12, 193, 37, 52, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 60, 118, 5, 5, 62, 118, 5, 5, 64, 118, 5, 5, 66, 118, 5, 5, 68, 118, 5, 5, 70, 118, 5, 5, 72, 118, 5, 5, 74, 118, 5, 5, 76, 118, 5, 5, 78, 118, 5, 5, 80, 118, 5, 5, 82, 118, 5, 5, 84, 118, 5, 5, 86, 118, 5, 5, 88, 118, 5, 5, 90, 118, 193, 240, 143, 103, 193, 247, 143, 103, 193, 254, 143, 103, 193, 7, 144, 103, 193, 31, 146, 103, 193, 38, 146, 103, 193, 45, 146, 103, 193, 52, 146, 103, 193, 108, 153, 103, 193, 115, 153, 103, 193, 122, 153, 103, 193, 129, 153, 103, 193, 192, 153, 103, 193, 199, 153, 103, 193, 206, 153, 103, 193, 213, 153, 103, 193, 128, 56, 12, 193, 135, 56, 12, 193, 142, 56, 12, 193, 149, 56, 12, 193, 156, 56, 12, 193, 163, 56, 12, 193, 170, 56, 12, 193, 177, 56, 12, 193, 184, 56, 12, 193, 191, 56, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 198, 56, 12, 193, 205, 56, 12, 193, 212, 56, 12, 193, 219, 56, 12, 193, 226, 56, 12, 193, 233, 56, 12, 193, 240, 56, 12, 193, 247, 56, 12, 193, 254, 56, 12, 193, 7, 57, 12, 193, 14, 57, 12, 193, 21, 57, 12, 193, 28, 57, 12, 193, 35, 57, 12, 193, 42, 57, 12, 193, 49, 57, 12, 193, 56, 57, 12, 193, 63, 57, 12, 193, 70, 57, 12, 193, 77, 57, 12, 193, 84, 57, 12, 193, 91, 57, 12, 193, 98, 57, 12, 5, 5, 59, 12, 5, 5, 61, 12, 5, 5, 63, 12, 5, 5, 65, 12, 5, 5, 67, 12, 5, 5, 69, 12, 5, 5, 71, 12, 5, 5, 73, 12, 5, 5, 75, 12, 193, 2, 77, 12, 193, 9, 77, 12, 160, 5, 88, 93, 5, 5, 88, 93, 160, 5, 104, 93, 5, 5, 104, 93, 160, 5, 20, 57, 5, 5, 20, 57, 197, 34, 196, 0, 197, 226, 195, 0, 160, 5, 64, 93, 5, 5, 64, 93, 160, 5, 66, 93, 5, 5, 66, 93, 160, 5, 68, 93, 5, 5, 68, 93, 5, 5, 6, 53, 5, 5, 6, 79, 197, 162, 107, 0, 197, 98, 107, 0, 197, 162, 110, 0, 197, 98, 110, 0, 197, 34, 111, 0, 197, 226, 110, 0, 197, 226, 111, 0, 197, 162, 111, 0, 197, 98, 112, 0, 197, 34, 112, 0, 197, 226, 112, 0, 197, 162, 112, 0, 160, 5, 20, 47, 5, 5, 20, 47, 160, 5, 40, 65, 5, 5, 40, 65, 160, 112, 0, 76, 5, 112, 0, 76, 160, 120, 0, 78, 5, 120, 0, 78, 160, 120, 0, 80, 5, 120, 0, 80, 5, 5, 79, 12, 193, 2, 81, 12, 193, 9, 81, 12, 160, 5, 90, 93, 5, 5, 90, 93, 160, 5, 34, 83, 5, 5, 32, 65, 5, 5, 92, 93, 160, 5, 18, 69, 5, 5, 18, 69, 160, 5, 10, 47, 5, 5, 10, 47, 5, 5, 12, 47, 5, 5, 10, 57, 160, 5, 16, 45, 5, 5, 16, 45, 160, 5, 10, 53, 5, 5, 10, 53, 197, 34, 102, 0, 197, 226, 101, 0, 197, 2, 167, 0, 197, 194, 166, 0, 197, 162, 196, 0, 197, 98, 196, 0, 5, 5, 20, 71, 5, 5, 16, 71, 5, 5, 18, 71, 5, 5, 22, 71, 5, 5, 24, 71, 5, 5, 6, 77, 5, 5, 10, 77, 5, 5, 44, 77, 5, 5, 46, 77, 5, 5, 36, 77, 5, 5, 48, 77, 5, 5, 50, 77, 5, 5, 52, 77, 5, 5, 24, 79, 5, 5, 8, 83, 5, 5, 22, 83, 5, 5, 16, 83, 5, 5, 18, 83, 5, 5, 30, 83, 5, 5, 16, 89, 5, 5, 18, 89, 5, 5, 20, 89, 5, 5, 8, 89, 5, 5, 10, 89, 5, 5, 12, 89, 5, 5, 14, 89, 5, 5, 14, 91, 193, 16, 81, 12, 46, 5, 20, 57, 32, 5, 26, 65, 46, 5, 18, 65, 32, 5, 30, 83, 197, 227, 105, 0, 197, 67, 106, 0, 165, 5, 0, 46, 197, 162, 51, 0, 193, 9, 97, 12, 197, 3, 115, 0, 197, 99, 115, 0, 158, 5, 28, 51, 193, 16, 97, 12, 197, 226, 51, 0, 22, 5, 0, 54, 165, 5, 0, 56, 165, 5, 0, 56, 165, 5, 0, 56, 22, 5, 0, 56, 196, 178, 5, 56, 165, 5, 0, 58, 165, 5, 0, 58, 165, 5, 0, 64, 22, 5, 0, 64, 193, 23, 97, 12, 165, 5, 0, 68, 197, 194, 162, 0, 193, 30, 97, 12, 193, 37, 97, 12, 165, 5, 0, 72, 172, 5, 0, 74, 165, 5, 0, 76, 165, 5, 0, 76, 165, 5, 0, 76, 193, 44, 97, 12, 193, 51, 97, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 30, 85, 12, 193, 37, 85, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 67, 122, 1, 197, 35, 146, 1, 197, 164, 70, 1, 197, 36, 70, 1, 197, 164, 168, 1, 197, 36, 119, 1, 197, 132, 98, 1, 197, 36, 121, 1, 197, 196, 121, 1, 197, 163, 119, 1, 197, 18, 120, 1, 197, 40, 91, 1, 5, 5, 242, 13, 193, 72, 85, 12, 255, 255, 255, 255, 255, 255, 255, 255, 198, 161, 164, 0, 198, 129, 164, 0, 198, 65, 164, 0, 198, 97, 164, 0, 198, 2, 165, 0, 198, 194, 164, 0, 49, 5, 218, 102, 38, 5, 218, 102, 16, 5, 218, 102, 27, 5, 218, 102, 49, 5, 216, 102, 38, 5, 216, 102, 16, 5, 216, 102, 27, 5, 216, 102, 46, 5, 250, 102, 32, 5, 250, 102, 197, 66, 191, 1, 197, 2, 191, 1, 193, 79, 85, 12, 193, 86, 85, 12, 193, 93, 85, 12, 193, 100, 85, 12, 193, 107, 85, 12, 193, 114, 85, 12, 193, 121, 85, 12, 193, 128, 85, 12, 193, 135, 85, 12, 193, 142, 85, 12, 193, 149, 85, 12, 193, 156, 85, 12, 193, 163, 85, 12, 193, 170, 85, 12, 193, 177, 85, 12, 193, 184, 85, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 198, 33, 163, 0, 198, 1, 163, 0, 198, 193, 162, 0, 198, 225, 162, 0, 44, 5, 228, 102, 30, 5, 228, 102, 198, 97, 161, 0, 198, 33, 161, 0, 46, 5, 230, 102, 32, 5, 230, 102, 197, 226, 183, 1, 198, 161, 165, 0, 198, 129, 165, 0, 5, 5, 20, 108, 5, 5, 24, 108, 5, 5, 122, 108, 5, 5, 124, 108, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 193, 205, 85, 12, 5, 5, 96, 108, 5, 5, 18, 16, 5, 5, 20, 16, 5, 5, 22, 16, 5, 5, 24, 16, 5, 5, 26, 16, 5, 5, 28, 16, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 42, 16, 5, 5, 44, 16, 5, 5, 46, 16, 193, 212, 85, 12, 193, 219, 85, 12, 193, 226, 85, 12, 193, 233, 85, 12, 193, 240, 85, 12, 193, 247, 85, 12, 5, 5, 160, 13, 193, 254, 85, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 48, 16, 5, 5, 50, 16, 5, 5, 52, 16, 5, 5, 54, 16, 5, 5, 56, 16, 5, 5, 58, 16, 5, 5, 60, 16, 5, 5, 62, 16, 5, 5, 64, 16, 5, 5, 66, 16, 5, 5, 68, 16, 5, 5, 70, 16, 5, 5, 72, 16, 5, 5, 74, 16, 5, 5, 76, 16, 5, 5, 78, 16, 5, 5, 80, 16, 5, 5, 82, 16, 5, 5, 84, 16, 5, 5, 86, 16, 5, 5, 88, 16, 193, 7, 86, 12, 193, 14, 86, 12, 193, 21, 86, 12, 193, 28, 86, 12, 193, 35, 86, 12, 193, 42, 86, 12, 193, 49, 86, 12, 193, 56, 86, 12, 5, 5, 162, 13, 5, 5, 164, 13, 5, 5, 166, 13, 5, 5, 124, 112, 5, 5, 126, 112, 5, 5, 128, 112, 5, 5, 130, 112, 5, 5, 132, 112, 255, 255, 255, 255, 201, 64, 18, 0, 201, 160, 19, 0, 5, 5, 142, 112, 255, 255, 255, 255, 5, 5, 144, 112, 5, 5, 146, 112, 5, 5, 148, 112, 5, 5, 152, 112, 197, 130, 198, 1, 193, 189, 86, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 2, 198, 1, 197, 66, 198, 1, 197, 66, 199, 1, 5, 5, 150, 112, 5, 5, 90, 16, 5, 5, 92, 16, 5, 5, 94, 16, 5, 5, 96, 16, 5, 5, 98, 16, 5, 5, 100, 16, 5, 5, 102, 16, 5, 5, 14, 112, 5, 5, 22, 112, 5, 5, 26, 112, 5, 5, 134, 112, 5, 5, 136, 112, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 104, 16, 5, 5, 106, 16, 5, 5, 108, 16, 5, 5, 110, 16, 5, 5, 112, 16, 5, 5, 114, 16, 5, 5, 116, 16, 5, 5, 118, 16, 5, 5, 120, 16, 193, 196, 86, 12, 197, 130, 197, 1, 197, 194, 197, 1, 197, 130, 198, 1, 197, 194, 198, 1, 197, 2, 199, 1, 197, 66, 197, 1, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 206, 64, 233, 0, 193, 203, 86, 12, 193, 210, 86, 12, 193, 217, 86, 12, 193, 224, 86, 12, 193, 58, 162, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 30, 16, 5, 5, 32, 16, 5, 5, 34, 16, 5, 5, 36, 16, 5, 5, 38, 16, 5, 5, 40, 16, 193, 231, 86, 12, 193, 238, 86, 12, 5, 5, 158, 13, 193, 245, 86, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 62, 116, 5, 5, 66, 116, 197, 130, 121, 2, 5, 5, 70, 116, 5, 5, 74, 116, 5, 5, 78, 116, 5, 5, 82, 116, 197, 2, 122, 2, 5, 5, 86, 116, 5, 5, 90, 116, 5, 5, 94, 116, 5, 5, 98, 116, 197, 130, 122, 2, 5, 5, 102, 116, 5, 5, 106, 116, 5, 5, 110, 116, 5, 5, 114, 116, 5, 5, 118, 116, 201, 38, 233, 0, 201, 38, 237, 0, 5, 5, 132, 116, 5, 5, 136, 116, 5, 5, 140, 116, 5, 5, 144, 116, 5, 5, 148, 116, 197, 2, 120, 2, 5, 112, 102, 116, 5, 112, 118, 116, 5, 112, 122, 116, 255, 255, 255, 255, 193, 110, 87, 12, 193, 117, 87, 12, 193, 124, 81, 117, 193, 16, 85, 117, 193, 117, 81, 117, 193, 2, 89, 117, 193, 114, 89, 117, 193, 128, 89, 117, 193, 184, 89, 117, 193, 58, 93, 117, 193, 72, 93, 117, 193, 93, 93, 117, 193, 100, 93, 117, 193, 65, 93, 117, 193, 79, 93, 117, 193, 86, 93, 117, 193, 231, 80, 117, 193, 107, 93, 117, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 193, 114, 93, 117, 193, 121, 93, 117, 193, 16, 89, 117, 193, 142, 89, 117, 193, 10, 88, 12, 193, 17, 88, 12, 193, 51, 79, 117, 193, 114, 79, 117, 193, 135, 79, 117, 193, 156, 79, 117, 193, 205, 79, 117, 193, 247, 79, 117, 193, 21, 80, 117, 193, 35, 80, 117, 193, 49, 80, 117, 193, 70, 80, 117, 193, 154, 80, 117, 193, 182, 80, 117, 193, 2, 85, 117, 193, 23, 85, 117, 193, 37, 85, 117, 193, 224, 80, 117, 5, 5, 98, 13, 193, 44, 85, 117, 193, 177, 79, 117, 193, 68, 81, 117, 193, 149, 93, 117, 193, 156, 93, 117, 193, 163, 93, 117, 193, 24, 88, 12, 193, 31, 88, 12, 193, 38, 88, 12, 193, 75, 81, 117, 193, 128, 93, 117, 193, 135, 93, 117, 193, 142, 93, 117, 193, 142, 79, 117, 193, 226, 79, 117, 193, 52, 88, 12, 193, 59, 88, 12, 193, 66, 88, 12, 193, 73, 88, 12, 193, 80, 88, 12, 193, 87, 88, 12, 193, 94, 88, 12, 193, 101, 88, 12, 193, 108, 88, 12, 193, 115, 88, 12, 193, 122, 88, 12, 193, 129, 88, 12, 193, 136, 88, 12, 193, 143, 88, 12, 193, 150, 88, 12, 193, 157, 88, 12, 193, 164, 88, 12, 193, 171, 88, 12, 193, 178, 88, 12, 193, 185, 88, 12, 193, 192, 88, 12, 193, 199, 88, 12, 193, 206, 88, 12, 193, 213, 88, 12, 193, 220, 88, 12, 193, 227, 88, 12, 193, 234, 88, 12, 193, 241, 88, 12, 193, 248, 88, 12, 193, 255, 88, 12, 193, 8, 89, 12, 193, 15, 89, 12, 197, 163, 90, 0, 197, 35, 93, 0, 197, 68, 80, 0, 197, 131, 92, 0, 197, 68, 73, 0, 46, 5, 0, 48, 197, 2, 136, 0, 197, 67, 136, 0, 197, 230, 136, 0, 197, 104, 151, 0, 197, 35, 191, 0, 197, 99, 205, 0, 197, 195, 208, 0, 193, 2, 97, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 130, 191, 0, 197, 227, 193, 0, 197, 226, 194, 0, 193, 58, 97, 12, 165, 5, 0, 92, 193, 65, 97, 12, 155, 5, 86, 96, 193, 72, 97, 12, 165, 5, 0, 92, 193, 79, 97, 12, 156, 5, 0, 62, 196, 146, 156, 42, 165, 5, 0, 44, 165, 5, 0, 46, 193, 86, 97, 12, 22, 5, 0, 50, 165, 5, 0, 50, 165, 5, 0, 52, 160, 5, 18, 53, 165, 5, 0, 66, 22, 5, 0, 70, 32, 5, 6, 101, 32, 5, 8, 101, 32, 5, 10, 101, 32, 5, 12, 101, 22, 5, 0, 58, 193, 93, 97, 12, 197, 3, 131, 0, 27, 5, 50, 96, 16, 5, 10, 96, 160, 5, 10, 96, 177, 5, 50, 96, 32, 5, 119, 12, 193, 100, 97, 12, 193, 107, 97, 12, 193, 114, 97, 12, 193, 121, 97, 12, 165, 5, 0, 48, 22, 5, 0, 48, 22, 5, 0, 50, 22, 5, 0, 58, 22, 5, 0, 60, 193, 128, 97, 12, 193, 45, 160, 12, 193, 135, 97, 12, 197, 163, 106, 0, 5, 5, 18, 53, 193, 142, 97, 12, 197, 67, 70, 0, 197, 3, 71, 0, 197, 228, 67, 0, 197, 195, 68, 0, 197, 99, 78, 0, 197, 131, 69, 0, 197, 195, 78, 0, 197, 67, 84, 0, 197, 195, 87, 0, 197, 227, 69, 0, 197, 3, 91, 0, 197, 163, 70, 0, 197, 163, 84, 0, 197, 99, 91, 0, 197, 131, 93, 0, 197, 162, 67, 0, 5, 5, 144, 16, 5, 5, 146, 16, 5, 5, 148, 16, 160, 5, 18, 47, 5, 5, 18, 47, 5, 5, 0, 31, 5, 5, 150, 16, 5, 5, 152, 16, 5, 5, 154, 16, 197, 195, 66, 0, 193, 149, 97, 12, 193, 156, 97, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 99, 12, 5, 5, 103, 12, 5, 5, 101, 12, 5, 5, 105, 12, 193, 2, 107, 12, 193, 9, 107, 12, 193, 16, 107, 12, 193, 23, 107, 12, 193, 30, 107, 12, 193, 37, 107, 12, 197, 34, 52, 0, 197, 98, 52, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 206, 128, 228, 0, 197, 162, 52, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 206, 160, 228, 0, 197, 226, 52, 0, 197, 98, 53, 0, 197, 34, 53, 0, 193, 126, 108, 12, 193, 133, 108, 12, 193, 140, 108, 12, 193, 147, 108, 12, 193, 154, 108, 12, 193, 161, 108, 12, 193, 168, 108, 12, 193, 175, 108, 12, 193, 182, 108, 12, 193, 189, 108, 12, 193, 196, 108, 12, 193, 203, 108, 12, 193, 210, 108, 12, 193, 217, 108, 12, 193, 224, 108, 12, 193, 231, 108, 12, 193, 238, 108, 12, 193, 245, 108, 12, 193, 252, 108, 12, 193, 5, 109, 12, 193, 12, 109, 12, 193, 19, 109, 12, 193, 26, 109, 12, 193, 33, 109, 12, 193, 40, 109, 12, 193, 47, 109, 12, 193, 54, 109, 12, 193, 61, 109, 12, 193, 68, 109, 12, 193, 75, 109, 12, 193, 82, 109, 12, 193, 89, 109, 12, 193, 96, 109, 12, 193, 103, 109, 12, 193, 110, 109, 12, 193, 117, 109, 12, 193, 124, 109, 12, 193, 131, 109, 12, 193, 138, 109, 12, 193, 145, 109, 12, 193, 152, 109, 12, 193, 159, 109, 12, 193, 166, 109, 12, 193, 173, 109, 12, 193, 180, 109, 12, 193, 187, 109, 12, 193, 194, 109, 12, 193, 201, 109, 12, 193, 208, 109, 12, 193, 215, 109, 12, 5, 5, 111, 12, 193, 2, 113, 12, 197, 162, 53, 0, 193, 9, 113, 12, 193, 16, 113, 12, 5, 5, 115, 12, 193, 2, 117, 12, 197, 226, 53, 0, 193, 9, 117, 12, 193, 16, 117, 12, 197, 34, 54, 0, 193, 23, 117, 12, 193, 37, 117, 12, 193, 44, 117, 12, 193, 51, 117, 12, 5, 5, 119, 12, 5, 5, 143, 12, 193, 9, 145, 12, 193, 16, 145, 12, 5, 5, 147, 12, 193, 2, 151, 12, 193, 9, 151, 12, 193, 16, 151, 12, 193, 23, 151, 12, 193, 30, 151, 12, 193, 37, 151, 12, 193, 51, 151, 12, 193, 65, 151, 12, 193, 72, 151, 12, 193, 79, 151, 12, 166, 5, 86, 96, 32, 5, 115, 12, 27, 5, 6, 96, 23, 5, 8, 96, 16, 5, 10, 96, 16, 5, 14, 96, 27, 5, 16, 96, 16, 5, 24, 96, 16, 5, 28, 96, 23, 5, 30, 96, 27, 5, 32, 96, 24, 5, 36, 96, 16, 5, 38, 96, 24, 5, 42, 96, 16, 5, 44, 96, 16, 5, 46, 96, 16, 5, 48, 96, 27, 5, 50, 96, 24, 5, 60, 96, 23, 5, 66, 96, 23, 5, 66, 96, 16, 5, 74, 96, 16, 5, 76, 96, 23, 5, 78, 96, 16, 5, 80, 96, 16, 5, 82, 96, 16, 5, 86, 96, 32, 5, 111, 12, 27, 5, 16, 96, 23, 5, 30, 96, 24, 5, 36, 96, 23, 5, 78, 96, 27, 5, 16, 96, 16, 5, 24, 96, 16, 5, 28, 96, 23, 5, 30, 96, 27, 5, 32, 96, 24, 5, 36, 96, 16, 5, 38, 96, 24, 5, 42, 96, 16, 5, 44, 96, 16, 5, 46, 96, 16, 5, 48, 96, 27, 5, 50, 96, 24, 5, 60, 96, 23, 5, 66, 96, 23, 5, 66, 96, 16, 5, 74, 96, 16, 5, 76, 96, 23, 5, 78, 96, 16, 5, 80, 96, 16, 5, 82, 96, 16, 5, 86, 96, 32, 5, 111, 12, 27, 5, 16, 96, 23, 5, 30, 96, 24, 5, 36, 96, 23, 5, 78, 96, 24, 5, 60, 96, 27, 5, 50, 96, 177, 5, 6, 96, 165, 5, 8, 96, 160, 5, 10, 96, 166, 5, 14, 96, 16, 5, 38, 96, 24, 5, 42, 96, 16, 5, 44, 96, 16, 5, 46, 96, 16, 5, 48, 96, 27, 5, 50, 96, 24, 5, 60, 96, 23, 5, 66, 96, 23, 5, 66, 96, 16, 5, 74, 96, 16, 5, 76, 96, 23, 5, 78, 96, 16, 5, 80, 96, 16, 5, 82, 96, 16, 5, 86, 96, 32, 5, 111, 12, 27, 5, 16, 96, 23, 5, 30, 96, 24, 5, 36, 96, 23, 5, 78, 96, 24, 5, 60, 96, 27, 5, 50, 96, 177, 5, 6, 96, 165, 5, 8, 96, 160, 5, 10, 96, 166, 5, 14, 96, 177, 5, 16, 96, 172, 5, 24, 96, 172, 5, 28, 96, 172, 5, 30, 96, 177, 5, 32, 96, 176, 5, 36, 96, 24, 5, 60, 96, 23, 5, 66, 96, 23, 5, 66, 96, 16, 5, 74, 96, 16, 5, 76, 96, 23, 5, 78, 96, 16, 5, 80, 96, 16, 5, 82, 96, 16, 5, 86, 96, 32, 5, 111, 12, 27, 5, 16, 96, 23, 5, 30, 96, 24, 5, 36, 96, 23, 5, 78, 96, 24, 5, 60, 96, 27, 5, 50, 96, 177, 5, 6, 96, 165, 5, 8, 96, 160, 5, 10, 96, 166, 5, 14, 96, 177, 5, 16, 96, 172, 5, 24, 96, 172, 5, 28, 96, 172, 5, 30, 96, 177, 5, 32, 96, 176, 5, 36, 96, 172, 5, 38, 96, 168, 5, 42, 96, 172, 5, 44, 96, 172, 5, 46, 96, 172, 5, 48, 96, 177, 5, 50, 96, 16, 5, 80, 96, 16, 5, 82, 96, 16, 5, 86, 96, 32, 5, 111, 12, 27, 5, 16, 96, 23, 5, 30, 96, 24, 5, 36, 96, 23, 5, 78, 96, 24, 5, 60, 96, 27, 5, 50, 96, 172, 5, 18, 96, 16, 5, 18, 96, 255, 255, 255, 255, 255, 255, 255, 255, 202, 32, 248, 2, 202, 65, 248, 2, 202, 98, 248, 2, 202, 131, 248, 2, 202, 164, 248, 2, 202, 197, 248, 2, 202, 230, 248, 2, 202, 7, 249, 2, 202, 40, 249, 2, 202, 73, 249, 2, 202, 32, 248, 2, 202, 65, 248, 2, 202, 98, 248, 2, 202, 131, 248, 2, 202, 164, 248, 2, 202, 197, 248, 2, 202, 230, 248, 2, 202, 7, 249, 2, 24, 5, 60, 96, 27, 5, 50, 96, 177, 5, 6, 96, 165, 5, 8, 96, 160, 5, 10, 96, 166, 5, 14, 96, 177, 5, 16, 96, 172, 5, 24, 96, 172, 5, 28, 96, 172, 5, 30, 96, 177, 5, 32, 96, 176, 5, 36, 96, 172, 5, 38, 96, 168, 5, 42, 96, 172, 5, 44, 96, 172, 5, 46, 96, 172, 5, 48, 96, 177, 5, 50, 96, 168, 5, 60, 96, 172, 5, 30, 96, 172, 5, 66, 96, 172, 5, 74, 96, 177, 5, 76, 96, 165, 5, 78, 96, 160, 5, 80, 96, 172, 5, 82, 96, 166, 5, 86, 96, 32, 5, 115, 12, 27, 5, 6, 96, 23, 5, 8, 96, 16, 5, 10, 96, 16, 5, 14, 96, 177, 5, 16, 96, 172, 5, 24, 96, 172, 5, 28, 96, 172, 5, 30, 96, 177, 5, 32, 96, 176, 5, 36, 96, 172, 5, 38, 96, 168, 5, 42, 96, 172, 5, 44, 96, 172, 5, 46, 96, 172, 5, 48, 96, 177, 5, 50, 96, 168, 5, 60, 96, 172, 5, 30, 96, 172, 5, 66, 96, 172, 5, 74, 96, 177, 5, 76, 96, 165, 5, 78, 96, 160, 5, 80, 96, 172, 5, 82, 96, 166, 5, 86, 96, 32, 5, 115, 12, 27, 5, 6, 96, 23, 5, 8, 96, 16, 5, 10, 96, 16, 5, 14, 96, 27, 5, 16, 96, 16, 5, 24, 96, 16, 5, 28, 96, 23, 5, 30, 96, 27, 5, 32, 96, 24, 5, 36, 96, 172, 5, 38, 96, 168, 5, 42, 96, 172, 5, 44, 96, 172, 5, 46, 96, 172, 5, 48, 96, 177, 5, 50, 96, 168, 5, 60, 96, 172, 5, 30, 96, 172, 5, 66, 96, 172, 5, 74, 96, 177, 5, 76, 96, 165, 5, 78, 96, 160, 5, 80, 96, 172, 5, 82, 96, 166, 5, 86, 96, 32, 5, 115, 12, 27, 5, 6, 96, 23, 5, 8, 96, 16, 5, 10, 96, 16, 5, 14, 96, 27, 5, 16, 96, 16, 5, 24, 96, 16, 5, 28, 96, 23, 5, 30, 96, 27, 5, 32, 96, 24, 5, 36, 96, 16, 5, 38, 96, 24, 5, 42, 96, 16, 5, 44, 96, 16, 5, 46, 96, 16, 5, 48, 96, 27, 5, 50, 96, 168, 5, 60, 96, 172, 5, 30, 96, 172, 5, 66, 96, 172, 5, 74, 96, 177, 5, 76, 96, 165, 5, 78, 96, 160, 5, 80, 96, 172, 5, 82, 96, 166, 5, 86, 96, 32, 5, 115, 12, 27, 5, 6, 96, 23, 5, 8, 96, 16, 5, 10, 96, 16, 5, 14, 96, 27, 5, 16, 96, 16, 5, 24, 96, 16, 5, 28, 96, 23, 5, 30, 96, 27, 5, 32, 96, 24, 5, 36, 96, 16, 5, 38, 96, 24, 5, 42, 96, 16, 5, 44, 96, 16, 5, 46, 96, 16, 5, 48, 96, 27, 5, 50, 96, 24, 5, 60, 96, 23, 5, 66, 96, 23, 5, 66, 96, 16, 5, 74, 96, 16, 5, 76, 96, 23, 5, 78, 96, 32, 5, 36, 101, 46, 5, 6, 101, 46, 5, 12, 101, 46, 5, 14, 101, 32, 5, 26, 101, 46, 5, 28, 101, 32, 5, 30, 101, 32, 5, 44, 101, 32, 5, 48, 101, 24, 5, 121, 12, 197, 98, 60, 1, 197, 34, 60, 1, 197, 67, 61, 1, 197, 227, 60, 1, 197, 130, 51, 1, 197, 194, 51, 1, 197, 2, 52, 1, 197, 130, 52, 1, 197, 2, 53, 1, 197, 66, 53, 1, 197, 130, 53, 1, 197, 2, 54, 1, 197, 194, 54, 1, 255, 255, 255, 255, 197, 2, 55, 1, 197, 130, 55, 1, 197, 34, 57, 1, 197, 226, 56, 1, 197, 162, 57, 1, 255, 255, 255, 255, 197, 226, 57, 1, 255, 255, 255, 255, 197, 162, 54, 0, 193, 49, 158, 12, 197, 194, 58, 0, 193, 56, 158, 12, 193, 63, 158, 12, 193, 70, 158, 12, 193, 77, 158, 12, 193, 84, 158, 12, 193, 91, 158, 12, 193, 98, 158, 12, 193, 105, 158, 12, 193, 112, 158, 12, 193, 119, 158, 12, 197, 130, 58, 0, 197, 98, 54, 0, 197, 130, 55, 0, 197, 2, 59, 0, 197, 66, 59, 0, 193, 126, 158, 12, 193, 133, 158, 12, 197, 130, 59, 0, 197, 194, 59, 0, 193, 140, 158, 12, 193, 147, 158, 12, 197, 2, 60, 0, 197, 66, 60, 0, 193, 154, 158, 12, 193, 161, 158, 12, 193, 168, 158, 12, 193, 175, 158, 12, 193, 182, 158, 12, 193, 189, 158, 12, 193, 86, 151, 12, 193, 93, 151, 12, 193, 100, 151, 12, 193, 107, 151, 12, 197, 194, 55, 0, 193, 114, 151, 12, 197, 2, 56, 0, 193, 121, 151, 12, 193, 128, 151, 12, 193, 135, 151, 12, 193, 142, 151, 12, 5, 5, 153, 12, 197, 66, 56, 0, 197, 131, 56, 0, 5, 5, 155, 12, 197, 226, 56, 0, 197, 35, 57, 0, 193, 2, 157, 12, 193, 9, 157, 12, 193, 16, 157, 12, 193, 23, 157, 12, 193, 30, 157, 12, 193, 37, 157, 12, 193, 44, 157, 12, 193, 51, 157, 12, 193, 58, 157, 12, 193, 65, 157, 12, 193, 72, 157, 12, 193, 79, 157, 12, 193, 86, 157, 12, 193, 93, 157, 12, 193, 100, 157, 12, 193, 41, 213, 12, 193, 48, 213, 12, 193, 55, 213, 12, 193, 62, 213, 12, 193, 69, 213, 12, 193, 76, 213, 12, 193, 83, 213, 12, 193, 90, 213, 12, 193, 97, 213, 12, 193, 104, 213, 12, 193, 111, 213, 12, 193, 118, 213, 12, 197, 68, 56, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 193, 107, 157, 12, 197, 130, 57, 0, 193, 114, 157, 12, 193, 121, 157, 12, 197, 194, 57, 0, 193, 128, 157, 12, 193, 135, 157, 12, 197, 2, 58, 0, 193, 142, 157, 12, 197, 66, 58, 0, 193, 149, 157, 12, 193, 156, 157, 12, 193, 163, 157, 12, 193, 170, 157, 12, 193, 177, 157, 12, 193, 184, 157, 12, 193, 191, 157, 12, 193, 198, 157, 12, 193, 205, 157, 12, 193, 212, 157, 12, 193, 219, 157, 12, 193, 226, 157, 12, 193, 233, 157, 12, 193, 240, 157, 12, 193, 247, 157, 12, 193, 254, 157, 12, 193, 7, 158, 12, 193, 14, 158, 12, 193, 21, 158, 12, 193, 28, 158, 12, 193, 35, 158, 12, 193, 42, 158, 12, 197, 130, 60, 0, 197, 194, 60, 0, 193, 196, 158, 12, 193, 203, 158, 12, 197, 130, 61, 0, 197, 194, 61, 0, 193, 210, 158, 12, 193, 217, 158, 12, 197, 2, 62, 0, 197, 66, 62, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 197, 2, 61, 0, 197, 66, 61, 0, 197, 130, 62, 0, 197, 194, 62, 0, 193, 43, 161, 12, 193, 50, 161, 12, 193, 57, 161, 12, 193, 64, 161, 12, 193, 71, 161, 12, 193, 78, 161, 12, 197, 2, 64, 0, 197, 66, 64, 0, 197, 130, 64, 0, 197, 194, 64, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 224, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 206, 192, 228, 0, 197, 2, 63, 0, 197, 66, 63, 0, 197, 130, 63, 0, 197, 194, 63, 0, 193, 208, 159, 12, 193, 215, 159, 12, 193, 222, 159, 12, 193, 229, 159, 12, 193, 236, 159, 12, 193, 243, 159, 12, 193, 250, 159, 12, 193, 3, 160, 12, 193, 10, 160, 12, 193, 17, 160, 12, 193, 24, 160, 12, 193, 31, 160, 12, 193, 38, 160, 12, 193, 52, 160, 12, 193, 59, 160, 12, 193, 66, 160, 12, 193, 73, 160, 12, 193, 80, 160, 12, 193, 87, 160, 12, 193, 94, 160, 12, 193, 101, 160, 12, 193, 108, 160, 12, 193, 115, 160, 12, 193, 122, 160, 12, 193, 129, 160, 12, 193, 136, 160, 12, 193, 143, 160, 12, 193, 150, 160, 12, 193, 157, 160, 12, 193, 164, 160, 12, 193, 171, 160, 12, 193, 178, 160, 12, 193, 185, 160, 12, 193, 192, 160, 12, 193, 199, 160, 12, 193, 206, 160, 12, 193, 213, 160, 12, 193, 220, 160, 12, 193, 227, 160, 12, 193, 234, 160, 12, 193, 241, 160, 12, 193, 248, 160, 12, 193, 255, 160, 12, 193, 8, 161, 12, 193, 15, 161, 12, 193, 22, 161, 12, 193, 29, 161, 12, 193, 36, 161, 12, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 206, 32, 229, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 202, 169, 12, 193, 209, 169, 12, 193, 216, 169, 12, 193, 223, 169, 12, 193, 230, 169, 12, 193, 237, 169, 12, 193, 244, 169, 12, 193, 251, 169, 12, 193, 4, 170, 12, 193, 11, 170, 12, 193, 18, 170, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 25, 170, 12, 193, 32, 170, 12, 5, 5, 172, 12, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 206, 64, 229, 0, 5, 5, 180, 12, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 206, 96, 229, 0, 5, 5, 185, 12, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 128, 229, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 206, 160, 244, 0, 255, 255, 255, 255, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 206, 192, 244, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 206, 160, 229, 0, 193, 111, 234, 12, 193, 118, 234, 12, 193, 125, 234, 12, 193, 132, 234, 12, 193, 139, 234, 12, 193, 146, 234, 12, 193, 153, 234, 12, 193, 160, 234, 12, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 206, 192, 229, 0, 193, 212, 250, 12, 193, 219, 250, 12, 193, 226, 250, 12, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 206, 224, 229, 0, 193, 69, 234, 12, 193, 76, 234, 12, 193, 83, 234, 12, 193, 90, 234, 12, 193, 97, 234, 12, 193, 104, 234, 12, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 206, 0, 230, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 151, 200, 12, 193, 158, 200, 12, 193, 165, 200, 12, 193, 172, 200, 12, 193, 179, 200, 12, 193, 186, 200, 12, 193, 193, 200, 12, 193, 200, 200, 12, 193, 207, 200, 12, 193, 214, 200, 12, 193, 221, 200, 12, 193, 228, 200, 12, 193, 235, 200, 12, 193, 242, 200, 12, 193, 249, 200, 12, 193, 2, 201, 12, 193, 9, 201, 12, 193, 16, 201, 12, 193, 23, 201, 12, 193, 30, 201, 12, 193, 37, 201, 12, 193, 44, 201, 12, 193, 51, 201, 12, 193, 58, 201, 12, 193, 65, 201, 12, 193, 72, 201, 12, 193, 79, 201, 12, 193, 86, 201, 12, 193, 93, 201, 12, 193, 100, 201, 12, 193, 107, 201, 12, 193, 114, 201, 12, 193, 121, 201, 12, 193, 128, 201, 12, 193, 135, 201, 12, 193, 142, 201, 12, 193, 149, 201, 12, 193, 156, 201, 12, 193, 163, 201, 12, 193, 170, 201, 12, 193, 177, 201, 12, 193, 184, 201, 12, 193, 191, 201, 12, 193, 198, 201, 12, 193, 205, 201, 12, 193, 212, 201, 12, 193, 219, 201, 12, 193, 226, 201, 12, 193, 233, 201, 12, 193, 240, 201, 12, 193, 247, 201, 12, 193, 254, 201, 12, 193, 7, 202, 12, 193, 14, 202, 12, 193, 21, 202, 12, 193, 28, 202, 12, 193, 35, 202, 12, 193, 42, 202, 12, 193, 49, 202, 12, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 206, 32, 230, 0, 31, 5, 0, 21, 31, 5, 0, 23, 31, 5, 0, 25, 31, 5, 0, 27, 31, 5, 0, 29, 31, 5, 0, 31, 31, 5, 0, 33, 31, 5, 0, 35, 34, 5, 0, 37, 197, 98, 71, 0, 31, 5, 0, 21, 31, 5, 0, 23, 31, 5, 0, 25, 31, 5, 0, 27, 31, 5, 0, 29, 31, 5, 0, 31, 31, 5, 0, 33, 31, 5, 0, 35, 34, 5, 0, 37, 197, 98, 71, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 64, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 160, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 192, 230, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 0, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 206, 32, 231, 0, 197, 2, 65, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 206, 64, 231, 0, 255, 255, 255, 255, 255, 255, 255, 255, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 206, 96, 231, 0, 255, 255, 255, 255, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 231, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 128, 230, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 232, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 206, 0, 242, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 206, 64, 232, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 206, 96, 232, 0, 5, 5, 234, 16, 5, 5, 236, 16, 5, 5, 238, 16, 5, 5, 240, 16, 5, 5, 242, 16, 5, 5, 244, 16, 5, 5, 246, 16, 5, 5, 248, 16, 5, 5, 250, 16, 5, 5, 252, 16, 5, 5, 254, 16, 5, 5, 2, 17, 5, 5, 4, 17, 5, 5, 6, 17, 5, 5, 8, 17, 5, 5, 10, 17, 5, 5, 12, 17, 5, 5, 14, 17, 5, 5, 16, 17, 5, 5, 18, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 85, 240, 12, 193, 92, 240, 12, 193, 99, 240, 12, 193, 106, 240, 12, 193, 113, 240, 12, 193, 120, 240, 12, 193, 127, 240, 12, 193, 134, 240, 12, 193, 141, 240, 12, 5, 5, 60, 17, 5, 5, 62, 17, 5, 5, 64, 17, 5, 5, 66, 17, 5, 5, 68, 17, 5, 5, 70, 17, 5, 5, 72, 17, 5, 5, 74, 17, 5, 5, 76, 17, 5, 5, 78, 17, 5, 5, 80, 17, 5, 5, 82, 17, 5, 5, 84, 17, 5, 5, 86, 17, 5, 5, 88, 17, 5, 5, 90, 17, 5, 5, 92, 17, 5, 5, 94, 17, 5, 5, 96, 17, 5, 5, 0, 29, 5, 5, 98, 17, 5, 5, 100, 17, 5, 5, 102, 17, 5, 5, 104, 17, 5, 5, 106, 17, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 206, 160, 234, 0, 5, 5, 0, 19, 5, 5, 108, 17, 193, 13, 241, 12, 193, 20, 241, 12, 193, 27, 241, 12, 255, 255, 255, 255, 193, 34, 241, 12, 193, 41, 241, 12, 193, 48, 241, 12, 193, 55, 241, 12, 193, 62, 241, 12, 193, 69, 241, 12, 193, 76, 241, 12, 193, 83, 241, 12, 193, 90, 241, 12, 193, 97, 241, 12, 193, 104, 241, 12, 193, 111, 241, 12, 193, 118, 241, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 125, 241, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 206, 192, 234, 0, 193, 9, 125, 124, 193, 16, 125, 124, 193, 23, 125, 124, 193, 30, 125, 124, 193, 37, 125, 124, 193, 44, 125, 124, 193, 51, 125, 124, 193, 58, 125, 124, 193, 65, 125, 124, 193, 72, 125, 124, 193, 79, 125, 124, 193, 86, 125, 124, 193, 93, 125, 124, 32, 5, 127, 124, 5, 5, 127, 124, 193, 2, 129, 124, 193, 9, 129, 124, 193, 16, 129, 124, 193, 23, 129, 124, 193, 30, 129, 124, 193, 37, 129, 124, 193, 44, 129, 124, 193, 51, 129, 124, 193, 193, 242, 12, 193, 200, 242, 12, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 193, 23, 146, 17, 193, 30, 146, 17, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 105, 244, 123, 193, 242, 242, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 206, 128, 241, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 206, 160, 241, 0, 255, 255, 255, 255, 255, 255, 255, 255, 193, 91, 251, 12, 193, 233, 250, 12, 193, 240, 250, 12, 193, 247, 250, 12, 193, 254, 250, 12, 193, 7, 251, 12, 193, 14, 251, 12, 193, 21, 251, 12, 193, 28, 251, 12, 193, 35, 251, 12, 193, 42, 251, 12, 193, 49, 251, 12, 193, 56, 251, 12, 193, 63, 251, 12, 193, 70, 251, 12, 193, 77, 251, 12, 193, 84, 251, 12, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 206, 192, 241, 0, 193, 47, 252, 12, 193, 54, 252, 12, 193, 171, 253, 12, 193, 178, 253, 12, 193, 185, 253, 12, 193, 192, 253, 12, 193, 199, 253, 12, 193, 206, 253, 12, 193, 213, 253, 12, 193, 220, 253, 12, 193, 227, 253, 12, 193, 234, 253, 12, 193, 241, 253, 12, 193, 248, 253, 12, 193, 255, 253, 12, 193, 8, 254, 12, 193, 15, 254, 12, 193, 22, 254, 12, 193, 29, 254, 12, 193, 36, 254, 12, 193, 43, 254, 12, 193, 50, 254, 12, 193, 57, 254, 12, 193, 64, 254, 12, 193, 71, 254, 12, 193, 78, 254, 12, 193, 85, 254, 12, 193, 92, 254, 12, 193, 99, 254, 12, 193, 106, 254, 12, 193, 113, 254, 12, 193, 120, 254, 12, 193, 127, 254, 12, 193, 134, 254, 12, 193, 141, 254, 12, 193, 148, 254, 12, 193, 155, 254, 12, 193, 162, 254, 12, 193, 169, 254, 12, 193, 176, 254, 12, 193, 183, 254, 12, 193, 190, 254, 12, 193, 197, 254, 12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 204, 254, 12, 193, 211, 254, 12, 193, 218, 254, 12, 193, 225, 254, 12, 193, 232, 254, 12, 193, 239, 254, 12, 193, 246, 254, 12, 193, 253, 254, 12, 193, 6, 255, 12, 193, 13, 255, 12, 193, 20, 255, 12, 193, 27, 255, 12, 193, 34, 255, 12, 193, 41, 255, 12, 193, 48, 255, 12, 193, 55, 255, 12, 193, 62, 255, 12, 193, 69, 255, 12, 193, 76, 255, 12, 193, 83, 255, 12, 193, 90, 255, 12, 193, 97, 255, 12, 193, 104, 255, 12, 193, 111, 255, 12, 193, 118, 255, 12, 193, 125, 255, 12, 193, 132, 255, 12, 193, 139, 255, 12, 193, 146, 255, 12, 193, 153, 255, 12, 193, 160, 255, 12, 193, 167, 255, 12, 193, 174, 255, 12, 193, 181, 255, 12, 193, 188, 255, 12, 193, 195, 255, 12, 193, 202, 255, 12, 193, 209, 255, 12, 193, 216, 255, 12, 193, 223, 255, 12, 193, 230, 255, 12, 193, 237, 255, 12, 193, 244, 255, 12, 193, 251, 255, 12, 193, 4, 2, 13, 193, 11, 2, 13, 193, 18, 2, 13, 193, 25, 2, 13, 193, 32, 2, 13, 193, 39, 2, 13, 193, 46, 2, 13, 193, 53, 2, 13, 193, 60, 2, 13, 193, 67, 2, 13, 193, 74, 2, 13, 193, 81, 2, 13, 193, 88, 2, 13, 193, 95, 2, 13, 193, 102, 2, 13, 193, 109, 2, 13, 193, 116, 2, 13, 193, 123, 2, 13, 193, 130, 2, 13, 193, 137, 2, 13, 5, 5, 16, 18, 5, 5, 18, 18, 5, 5, 20, 18, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 193, 172, 2, 13, 5, 5, 22, 18, 5, 5, 24, 18, 5, 5, 26, 18, 5, 5, 226, 13, 5, 112, 0, 21, 5, 112, 0, 23, 5, 5, 28, 18, 5, 5, 30, 18, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 76, 18, 5, 5, 78, 18, 5, 5, 80, 18, 5, 5, 82, 18, 5, 5, 84, 18, 5, 5, 86, 18, 5, 5, 88, 18, 5, 5, 90, 18, 5, 5, 92, 18, 5, 5, 94, 18, 5, 5, 96, 18, 5, 5, 98, 18, 5, 5, 100, 18, 5, 5, 102, 18, 193, 179, 2, 13, 5, 112, 0, 23, 32, 112, 0, 25, 5, 112, 0, 27, 5, 112, 0, 29, 5, 112, 0, 31, 5, 112, 0, 33, 5, 112, 0, 35, 5, 112, 0, 37, 5, 5, 104, 18, 5, 5, 106, 18, 5, 5, 108, 18, 5, 5, 110, 18, 5, 5, 112, 18, 5, 5, 114, 18, 5, 5, 116, 18, 255, 255, 255, 255, 255, 255, 255, 255, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 206, 192, 242, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 206, 224, 242, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 178, 6, 13, 193, 185, 6, 13, 193, 192, 6, 13, 193, 199, 6, 13, 193, 206, 6, 13, 193, 213, 6, 13, 193, 220, 6, 13, 193, 227, 6, 13, 193, 234, 6, 13, 193, 241, 6, 13, 193, 248, 6, 13, 193, 255, 6, 13, 193, 8, 7, 13, 193, 15, 7, 13, 193, 22, 7, 13, 255, 255, 255, 255, 255, 255, 255, 255, 193, 29, 7, 13, 193, 36, 7, 13, 193, 43, 7, 13, 193, 50, 7, 13, 193, 57, 7, 13, 193, 64, 7, 13, 193, 71, 7, 13, 193, 78, 7, 13, 193, 85, 7, 13, 193, 92, 7, 13, 193, 99, 7, 13, 193, 106, 7, 13, 193, 113, 7, 13, 193, 120, 7, 13, 193, 127, 7, 13, 255, 255, 255, 255, 193, 134, 7, 13, 193, 141, 7, 13, 193, 148, 7, 13, 193, 155, 7, 13, 193, 162, 7, 13, 193, 169, 7, 13, 193, 176, 7, 13, 193, 183, 7, 13, 193, 190, 7, 13, 193, 197, 7, 13, 193, 204, 7, 13, 193, 211, 7, 13, 193, 218, 7, 13, 193, 225, 7, 13, 193, 232, 7, 13, 255, 255, 255, 255, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 206, 0, 243, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 206, 96, 244, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 70, 11, 13, 193, 77, 11, 13, 193, 84, 11, 13, 193, 91, 11, 13, 193, 98, 11, 13, 193, 105, 11, 13, 193, 112, 11, 13, 193, 119, 11, 13, 193, 126, 11, 13, 193, 133, 11, 13, 193, 140, 11, 13, 193, 147, 11, 13, 193, 154, 11, 13, 193, 161, 11, 13, 255, 255, 255, 255, 255, 255, 255, 255, 193, 40, 40, 13, 193, 47, 40, 13, 193, 54, 40, 13, 193, 61, 40, 13, 193, 68, 40, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 75, 40, 13, 193, 82, 40, 13, 193, 89, 40, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 176, 5, 0, 74, 168, 5, 0, 76, 168, 5, 0, 78, 168, 5, 0, 80, 168, 5, 0, 82, 168, 5, 0, 84, 176, 5, 0, 86, 168, 5, 0, 88, 168, 5, 0, 90, 168, 5, 0, 92, 197, 34, 155, 0, 197, 98, 155, 0, 197, 98, 158, 0, 193, 189, 11, 13, 193, 196, 11, 13, 193, 203, 11, 13, 180, 5, 0, 42, 177, 5, 0, 44, 177, 5, 0, 46, 177, 5, 0, 48, 180, 5, 0, 50, 177, 5, 0, 52, 177, 5, 0, 54, 180, 5, 0, 56, 180, 5, 0, 58, 177, 5, 0, 60, 180, 5, 0, 62, 180, 5, 0, 64, 180, 5, 0, 66, 180, 5, 0, 68, 180, 5, 0, 70, 180, 5, 0, 72, 193, 210, 11, 13, 193, 217, 11, 13, 193, 224, 11, 13, 193, 231, 11, 13, 193, 238, 11, 13, 193, 245, 11, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 206, 32, 243, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 0, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 0, 0, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 206, 0, 244, 0, 255, 255, 255, 255, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 206, 32, 244, 0, 255, 255, 255, 255, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 206, 64, 244, 0, 193, 96, 40, 13, 193, 103, 40, 13, 193, 110, 40, 13, 193, 117, 40, 13, 193, 124, 40, 13, 193, 131, 40, 13, 193, 138, 40, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 206, 128, 244, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 66, 41, 13, 193, 73, 41, 13, 193, 80, 41, 13, 193, 87, 41, 13, 193, 94, 41, 13, 193, 101, 41, 13, 193, 108, 41, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 115, 41, 13, 193, 122, 41, 13, 193, 129, 41, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 136, 41, 13, 193, 143, 41, 13, 193, 150, 41, 13, 193, 157, 41, 13, 193, 164, 41, 13, 193, 171, 41, 13, 193, 178, 41, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 206, 64, 243, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 173, 47, 13, 193, 180, 47, 13, 193, 187, 47, 13, 193, 194, 47, 13, 193, 201, 47, 13, 193, 208, 47, 13, 193, 215, 47, 13, 193, 222, 47, 13, 193, 229, 47, 13, 193, 236, 47, 13, 193, 243, 47, 13, 193, 250, 47, 13, 193, 3, 48, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 10, 48, 13, 193, 17, 48, 13, 193, 24, 48, 13, 193, 31, 48, 13, 193, 38, 48, 13, 193, 45, 48, 13, 193, 52, 48, 13, 193, 59, 48, 13, 193, 66, 48, 13, 193, 73, 48, 13, 193, 80, 48, 13, 193, 87, 48, 13, 193, 94, 48, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 206, 96, 243, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 206, 128, 243, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 12, 54, 13, 193, 19, 54, 13, 193, 26, 54, 13, 193, 33, 54, 13, 193, 40, 54, 13, 193, 47, 54, 13, 193, 54, 54, 13, 193, 61, 54, 13, 193, 68, 54, 13, 193, 75, 54, 13, 193, 82, 54, 13, 193, 89, 54, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 96, 54, 13, 193, 103, 54, 13, 193, 110, 54, 13, 193, 117, 54, 13, 193, 124, 54, 13, 193, 131, 54, 13, 193, 138, 54, 13, 193, 145, 54, 13, 193, 152, 54, 13, 193, 159, 54, 13, 193, 166, 54, 13, 193, 173, 54, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 206, 160, 243, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 64, 56, 13, 193, 71, 56, 13, 193, 78, 56, 13, 193, 85, 56, 13, 193, 92, 56, 13, 193, 99, 56, 13, 193, 106, 56, 13, 193, 113, 56, 13, 193, 120, 56, 13, 193, 127, 56, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 206, 192, 243, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 206, 224, 243, 0, 255, 255, 255, 255, 255, 255, 255, 255, 193, 116, 58, 13, 193, 123, 58, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 206, 32, 242, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 100, 73, 13, 193, 107, 73, 13, 193, 114, 73, 13, 193, 121, 73, 13, 193, 128, 73, 13, 193, 135, 73, 13, 193, 142, 73, 13, 193, 149, 73, 13, 193, 156, 73, 13, 193, 163, 73, 13, 193, 170, 73, 13, 193, 177, 73, 13, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 206, 224, 231, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 22, 122, 16, 5, 30, 122, 16, 5, 32, 122, 16, 5, 46, 122, 16, 5, 52, 122, 16, 5, 58, 122, 16, 5, 60, 122, 16, 5, 62, 122, 16, 5, 64, 122, 16, 5, 66, 122, 16, 5, 72, 122, 16, 5, 84, 122, 16, 5, 86, 122, 16, 5, 88, 122, 16, 5, 90, 122, 16, 5, 92, 122, 198, 161, 170, 0, 198, 193, 170, 0, 198, 193, 171, 0, 198, 225, 171, 0, 32, 5, 95, 121, 32, 5, 97, 121, 32, 5, 99, 121, 32, 5, 147, 121, 32, 5, 149, 121, 32, 5, 155, 121, 32, 5, 159, 121, 32, 5, 161, 121, 32, 5, 165, 121, 32, 5, 169, 121, 32, 5, 175, 121, 255, 255, 255, 255, 193, 37, 78, 13, 193, 44, 78, 13, 198, 97, 172, 0, 198, 225, 175, 0, 198, 1, 173, 0, 198, 225, 183, 0, 198, 97, 173, 0, 198, 65, 174, 0, 198, 161, 173, 0, 198, 1, 202, 0, 198, 1, 175, 0, 198, 193, 173, 0, 198, 161, 172, 0, 198, 129, 185, 0, 198, 33, 184, 0, 198, 161, 176, 0, 30, 5, 6, 121, 30, 5, 10, 121, 30, 5, 12, 121, 30, 5, 16, 121, 30, 5, 18, 121, 30, 5, 20, 121, 30, 5, 24, 121, 30, 5, 28, 121, 30, 5, 30, 121, 30, 5, 34, 121, 30, 5, 36, 121, 30, 5, 38, 121, 30, 5, 40, 121, 30, 5, 42, 121, 197, 130, 174, 2, 197, 194, 174, 2, 197, 2, 175, 2, 197, 66, 175, 2, 197, 130, 175, 2, 197, 194, 175, 2, 197, 2, 176, 2, 197, 66, 176, 2, 197, 194, 176, 2, 197, 130, 177, 2, 197, 98, 178, 2, 197, 162, 178, 2, 197, 226, 178, 2, 197, 34, 179, 2, 197, 197, 177, 2, 197, 4, 177, 2, 197, 130, 176, 2, 193, 51, 78, 13, 193, 72, 79, 117, 193, 128, 79, 117, 193, 198, 79, 117, 193, 98, 80, 117, 193, 19, 81, 117, 193, 177, 89, 117, 5, 5, 96, 13, 193, 254, 79, 117, 193, 238, 80, 117, 193, 58, 79, 117, 193, 79, 79, 117, 193, 163, 79, 117, 193, 212, 79, 117, 193, 56, 80, 117, 193, 77, 80, 117, 193, 105, 80, 117, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 193, 58, 85, 117, 193, 140, 80, 117, 193, 161, 80, 117, 193, 5, 81, 117, 193, 26, 81, 117, 255, 255, 255, 255, 5, 5, 68, 114, 5, 5, 70, 114, 5, 5, 72, 114, 5, 5, 74, 114, 5, 5, 76, 114, 5, 5, 78, 114, 5, 5, 80, 114, 5, 5, 82, 114, 5, 5, 84, 114, 5, 5, 86, 114, 5, 5, 88, 114, 5, 5, 90, 114, 5, 5, 92, 114, 5, 5, 94, 114, 5, 5, 96, 114, 5, 5, 98, 114, 5, 5, 100, 114, 5, 5, 102, 114, 5, 5, 104, 114, 5, 5, 106, 114, 5, 5, 108, 114, 5, 5, 110, 114, 5, 5, 112, 114, 5, 5, 114, 114, 5, 5, 116, 114, 5, 5, 118, 114, 5, 5, 120, 114, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 170, 13, 5, 5, 176, 13, 5, 5, 178, 13, 5, 5, 180, 13, 5, 5, 182, 13, 5, 5, 184, 13, 5, 5, 186, 13, 5, 5, 188, 13, 5, 5, 190, 13, 5, 5, 240, 13, 5, 5, 192, 13, 5, 5, 194, 13, 5, 5, 196, 13, 5, 5, 198, 13, 5, 5, 200, 13, 5, 5, 202, 13, 5, 5, 204, 13, 5, 5, 206, 13, 5, 5, 208, 13, 5, 5, 210, 13, 5, 5, 212, 13, 5, 5, 214, 13, 5, 5, 216, 13, 5, 5, 218, 13, 5, 5, 220, 13, 5, 5, 222, 13, 5, 5, 224, 13, 5, 5, 228, 13, 5, 5, 230, 13, 5, 5, 232, 13, 5, 5, 234, 13, 5, 5, 236, 13, 5, 5, 238, 13, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 156, 16, 5, 5, 158, 16, 5, 5, 160, 16, 5, 5, 162, 16, 5, 5, 164, 16, 5, 5, 166, 16, 5, 5, 168, 16, 5, 5, 170, 16, 5, 5, 172, 16, 5, 5, 174, 16, 5, 5, 176, 16, 5, 5, 178, 16, 5, 5, 180, 16, 5, 5, 182, 16, 5, 5, 184, 16, 5, 5, 186, 16, 5, 5, 188, 16, 5, 5, 190, 16, 5, 5, 192, 16, 5, 5, 194, 16, 5, 5, 196, 16, 5, 5, 198, 16, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 29, 193, 9, 200, 16, 193, 16, 200, 16, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 233, 251, 123, 193, 240, 251, 123, 193, 247, 251, 123, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 5, 5, 112, 119, 5, 5, 114, 119, 5, 5, 116, 119, 5, 112, 116, 119, 5, 5, 118, 119, 5, 5, 120, 119, 5, 5, 122, 119, 5, 5, 124, 119, 5, 5, 126, 119, 5, 5, 128, 119, 5, 5, 130, 119, 5, 112, 130, 119, 5, 5, 132, 119, 5, 112, 132, 119, 5, 5, 134, 119, 5, 5, 136, 119, 5, 5, 138, 119, 5, 5, 140, 119, 5, 5, 142, 119, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 29, 193, 23, 200, 16, 193, 30, 200, 16, 193, 37, 200, 16, 193, 44, 200, 16, 5, 5, 20, 17, 5, 5, 22, 17, 5, 5, 0, 21, 5, 5, 0, 29, 5, 5, 24, 17, 5, 5, 26, 17, 5, 5, 28, 17, 5, 5, 30, 17, 5, 5, 0, 29, 5, 5, 32, 17, 5, 5, 34, 17, 5, 5, 36, 17, 5, 5, 38, 17, 5, 5, 40, 17, 5, 5, 42, 17, 5, 5, 0, 29, 5, 5, 44, 17, 5, 5, 46, 17, 5, 5, 48, 17, 5, 5, 50, 17, 5, 5, 52, 17, 5, 5, 54, 17, 5, 5, 56, 17, 5, 5, 58, 17, 5, 5, 0, 21, 5, 5, 0, 21, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 23, 5, 5, 0, 23, 5, 5, 0, 23, 5, 5, 0, 29, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 27, 5, 5, 0, 29, 193, 37, 146, 17, 193, 44, 146, 17, 193, 51, 146, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 157, 124, 193, 16, 157, 124, 193, 23, 157, 124, 193, 30, 157, 124, 193, 37, 157, 124, 193, 44, 157, 124, 193, 51, 157, 124, 193, 58, 157, 124, 193, 65, 157, 124, 193, 72, 157, 124, 193, 79, 157, 124, 193, 86, 157, 124, 193, 93, 157, 124, 193, 100, 157, 124, 193, 107, 157, 124, 193, 114, 157, 124, 193, 121, 157, 124, 193, 128, 157, 124, 193, 135, 157, 124, 255, 255, 255, 255, 193, 142, 157, 124, 193, 149, 157, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 29, 193, 58, 146, 17, 193, 65, 146, 17, 193, 72, 146, 17, 193, 9, 118, 124, 193, 16, 118, 124, 193, 23, 118, 124, 193, 30, 118, 124, 193, 37, 118, 124, 193, 44, 118, 124, 193, 51, 118, 124, 193, 58, 118, 124, 193, 65, 118, 124, 193, 72, 118, 124, 193, 79, 118, 124, 193, 86, 118, 124, 193, 93, 118, 124, 193, 100, 118, 124, 193, 107, 118, 124, 193, 114, 118, 124, 193, 121, 118, 124, 193, 128, 118, 124, 193, 135, 118, 124, 193, 142, 118, 124, 193, 149, 118, 124, 193, 156, 118, 124, 193, 163, 118, 124, 193, 170, 118, 124, 193, 177, 118, 124, 193, 184, 118, 124, 193, 191, 118, 124, 193, 198, 118, 124, 193, 205, 118, 124, 5, 5, 0, 21, 193, 86, 146, 17, 193, 93, 146, 17, 193, 9, 159, 124, 193, 16, 159, 124, 193, 23, 159, 124, 193, 30, 159, 124, 193, 37, 159, 124, 193, 44, 159, 124, 193, 51, 159, 124, 193, 58, 159, 124, 193, 65, 159, 124, 193, 72, 159, 124, 193, 79, 159, 124, 193, 86, 159, 124, 193, 93, 159, 124, 193, 100, 159, 124, 193, 107, 159, 124, 193, 114, 159, 124, 193, 121, 159, 124, 193, 128, 159, 124, 193, 135, 159, 124, 193, 142, 159, 124, 193, 149, 159, 124, 193, 156, 159, 124, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 193, 177, 146, 17, 193, 184, 146, 17, 193, 191, 146, 17, 193, 198, 146, 17, 193, 9, 160, 124, 193, 16, 160, 124, 193, 23, 160, 124, 193, 30, 160, 124, 193, 37, 160, 124, 193, 44, 160, 124, 193, 51, 160, 124, 193, 58, 160, 124, 193, 65, 160, 124, 193, 72, 160, 124, 193, 79, 160, 124, 193, 86, 160, 124, 193, 93, 160, 124, 193, 100, 160, 124, 193, 107, 160, 124, 193, 114, 160, 124, 193, 121, 160, 124, 193, 128, 160, 124, 193, 135, 160, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 193, 205, 146, 17, 193, 212, 146, 17, 193, 219, 146, 17, 193, 226, 146, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 193, 233, 146, 17, 193, 240, 146, 17, 193, 247, 146, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 27, 5, 5, 0, 29, 193, 254, 146, 17, 193, 7, 147, 17, 193, 14, 147, 17, 193, 21, 147, 17, 193, 28, 147, 17, 197, 2, 237, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 196, 124, 193, 16, 196, 124, 193, 23, 196, 124, 193, 30, 196, 124, 193, 37, 196, 124, 193, 44, 196, 124, 193, 51, 196, 124, 193, 58, 196, 124, 193, 65, 196, 124, 193, 72, 196, 124, 193, 79, 196, 124, 193, 86, 196, 124, 193, 93, 196, 124, 193, 100, 196, 124, 193, 107, 196, 124, 193, 114, 196, 124, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 193, 56, 147, 17, 193, 63, 147, 17, 193, 70, 147, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 192, 236, 0, 206, 192, 236, 0, 206, 192, 236, 0, 206, 192, 236, 0, 206, 192, 236, 0, 206, 192, 236, 0, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 231, 245, 113, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 150, 17, 5, 5, 152, 17, 5, 5, 154, 17, 5, 5, 156, 17, 5, 5, 158, 17, 5, 5, 160, 17, 5, 5, 162, 17, 5, 5, 164, 17, 5, 5, 166, 17, 5, 5, 168, 17, 5, 5, 170, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 193, 2, 172, 17, 193, 9, 172, 17, 193, 16, 172, 17, 193, 23, 172, 17, 193, 30, 172, 17, 193, 37, 172, 17, 193, 44, 172, 17, 193, 51, 172, 17, 193, 58, 172, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 172, 123, 193, 100, 172, 17, 193, 107, 172, 17, 255, 255, 255, 255, 193, 105, 241, 123, 193, 112, 241, 123, 193, 119, 241, 123, 193, 126, 241, 123, 193, 133, 241, 123, 193, 140, 241, 123, 193, 147, 241, 123, 193, 154, 241, 123, 193, 161, 241, 123, 193, 168, 241, 123, 193, 175, 241, 123, 193, 182, 241, 123, 193, 189, 241, 123, 193, 196, 241, 123, 193, 203, 241, 123, 193, 210, 241, 123, 193, 217, 241, 123, 193, 224, 241, 123, 193, 231, 241, 123, 193, 238, 241, 123, 193, 245, 241, 123, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 252, 241, 123, 193, 5, 242, 123, 193, 12, 242, 123, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 174, 17, 5, 5, 176, 17, 5, 5, 178, 17, 5, 5, 180, 17, 5, 5, 182, 17, 5, 5, 184, 17, 5, 5, 186, 17, 5, 5, 188, 17, 5, 5, 190, 17, 5, 5, 192, 17, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 194, 17, 5, 5, 196, 17, 5, 5, 198, 17, 5, 5, 200, 17, 5, 5, 202, 17, 5, 5, 204, 17, 5, 5, 206, 17, 5, 5, 208, 17, 5, 5, 210, 17, 5, 5, 212, 17, 5, 5, 214, 17, 5, 5, 216, 17, 5, 5, 218, 17, 5, 5, 220, 17, 5, 5, 222, 17, 5, 5, 224, 17, 5, 5, 226, 17, 5, 5, 228, 17, 5, 5, 230, 17, 5, 5, 232, 17, 5, 5, 234, 17, 5, 5, 236, 17, 5, 5, 238, 17, 5, 5, 240, 17, 5, 5, 242, 17, 5, 5, 244, 17, 5, 5, 246, 17, 5, 5, 248, 17, 5, 5, 250, 17, 5, 5, 252, 17, 5, 5, 254, 17, 5, 5, 2, 18, 5, 5, 4, 18, 5, 5, 6, 18, 5, 5, 8, 18, 5, 5, 10, 18, 5, 5, 12, 18, 5, 5, 14, 18, 255, 255, 255, 255, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 32, 18, 5, 5, 34, 18, 5, 5, 36, 18, 5, 5, 38, 18, 5, 5, 40, 18, 5, 5, 42, 18, 5, 5, 44, 18, 5, 5, 46, 18, 5, 5, 48, 18, 5, 5, 50, 18, 5, 5, 52, 18, 5, 5, 54, 18, 5, 5, 56, 18, 5, 5, 58, 18, 5, 5, 60, 18, 5, 5, 62, 18, 5, 5, 64, 18, 5, 5, 66, 18, 5, 5, 68, 18, 5, 5, 70, 18, 5, 5, 72, 18, 5, 5, 74, 18, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 193, 2, 118, 18, 193, 9, 118, 18, 193, 16, 118, 18, 193, 23, 118, 18, 193, 30, 118, 18, 193, 37, 118, 18, 193, 44, 118, 18, 255, 255, 255, 255, 255, 255, 255, 255, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 206, 0, 236, 0, 193, 56, 119, 18, 193, 63, 119, 18, 193, 70, 119, 18, 193, 77, 119, 18, 193, 84, 119, 18, 193, 91, 119, 18, 193, 98, 119, 18, 193, 105, 119, 18, 193, 112, 119, 18, 193, 119, 119, 18, 5, 5, 36, 125, 5, 5, 38, 125, 5, 5, 40, 125, 5, 5, 42, 125, 5, 5, 44, 125, 5, 5, 46, 125, 5, 5, 48, 125, 5, 5, 50, 125, 5, 5, 52, 125, 5, 5, 54, 125, 5, 5, 56, 125, 5, 5, 58, 125, 5, 5, 60, 125, 5, 5, 62, 125, 5, 5, 64, 125, 5, 5, 66, 125, 5, 112, 66, 125, 5, 5, 68, 125, 5, 5, 70, 125, 5, 5, 72, 125, 5, 5, 74, 125, 5, 5, 76, 125, 5, 5, 78, 125, 5, 5, 80, 125, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 126, 119, 18, 193, 49, 119, 18, 193, 2, 82, 125, 193, 9, 82, 125, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 193, 133, 119, 18, 193, 140, 119, 18, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 25, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 27, 5, 5, 0, 27, 5, 5, 0, 27, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 33, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 0, 37, 5, 5, 0, 37, 5, 5, 0, 37, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 29, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 21, 5, 5, 0, 23, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 206, 224, 239, 0, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 121, 18, 5, 5, 123, 18, 5, 5, 125, 18, 5, 5, 127, 18, 5, 5, 129, 18, 5, 5, 131, 18, 5, 5, 133, 18, 5, 5, 135, 18, 5, 5, 137, 18, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 21, 5, 5, 0, 29, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 20, 111, 5, 5, 24, 111, 5, 5, 132, 111, 5, 5, 134, 111, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 5, 5, 112, 111, 5, 5, 114, 111, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 65, 13, 118, 5, 5, 15, 118, 193, 2, 17, 118, 193, 9, 17, 118, 193, 16, 17, 118, 193, 23, 17, 118, 193, 30, 17, 118, 193, 37, 17, 118, 193, 44, 17, 118, 193, 51, 17, 118, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 5, 5, 0, 21, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 2, 154, 2, 197, 67, 154, 2, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 100, 243, 113, 193, 16, 243, 113, 193, 30, 243, 113, 193, 51, 243, 113, 193, 107, 243, 113, 193, 114, 243, 113, 193, 121, 243, 113, 193, 128, 243, 113, 193, 135, 243, 113, 193, 142, 243, 113, 193, 149, 243, 113, 193, 156, 243, 113, 193, 2, 243, 113, 193, 72, 243, 113, 193, 2, 239, 113, 193, 93, 243, 113, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 32, 5, 233, 113, 193, 37, 243, 113, 193, 58, 243, 113, 193, 107, 239, 113, 32, 5, 237, 113, 32, 5, 241, 113, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 79, 6, 117, 193, 86, 6, 117, 193, 93, 6, 117, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 206, 96, 228, 0, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 32, 5, 42, 120, 32, 5, 63, 120, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 206, 128, 233, 0, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 198, 193, 165, 0, 198, 225, 165, 0, 198, 1, 166, 0, 198, 33, 166, 0, 198, 65, 166, 0, 198, 97, 166, 0, 198, 129, 166, 0, 198, 161, 166, 0, 198, 193, 166, 0, 198, 225, 166, 0, 198, 1, 167, 0, 198, 33, 167, 0, 198, 65, 167, 0, 198, 97, 167, 0, 198, 129, 167, 0, 198, 161, 167, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 206, 224, 236, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 206, 0, 237, 0, 193, 247, 96, 117, 193, 254, 96, 117, 193, 7, 97, 117, 193, 21, 97, 117, 193, 28, 97, 117, 193, 35, 97, 117, 193, 42, 97, 117, 193, 49, 97, 117, 193, 56, 97, 117, 193, 63, 97, 117, 193, 70, 97, 117, 193, 77, 97, 117, 193, 84, 97, 117, 201, 32, 23, 2, 201, 32, 24, 2, 193, 119, 97, 117, 193, 126, 97, 117, 255, 255, 255, 255, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 202, 160, 245, 2, 202, 193, 245, 2, 202, 226, 245, 2, 202, 3, 246, 2, 202, 36, 246, 2, 202, 69, 246, 2, 202, 102, 246, 2, 202, 135, 246, 2, 202, 168, 246, 2, 202, 201, 246, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 40, 249, 2, 202, 73, 249, 2, 202, 32, 248, 2, 202, 65, 248, 2, 202, 98, 248, 2, 202, 131, 248, 2, 202, 164, 248, 2, 202, 197, 248, 2, 202, 230, 248, 2, 202, 7, 249, 2, 202, 40, 249, 2, 202, 73, 249, 2, 202, 32, 248, 2, 202, 65, 248, 2, 202, 98, 248, 2, 202, 131, 248, 2, 202, 164, 248, 2, 202, 197, 248, 2, 202, 230, 248, 2, 202, 7, 249, 2, 202, 40, 249, 2, 202, 73, 249, 2, 202, 32, 248, 2, 202, 65, 248, 2, 202, 98, 248, 2, 202, 131, 248, 2, 202, 164, 248, 2, 202, 197, 248, 2, 202, 230, 248, 2, 202, 7, 249, 2, 202, 40, 249, 2, 202, 73, 249, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 202, 32, 248, 2, 202, 65, 248, 2, 202, 98, 248, 2, 202, 131, 248, 2, 202, 164, 248, 2, 202, 197, 248, 2, 202, 230, 248, 2, 202, 7, 249, 2, 202, 40, 249, 2, 202, 73, 249, 2, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 28, 5, 0, 74, 25, 5, 0, 76, 25, 5, 0, 78, 25, 5, 0, 80, 25, 5, 0, 82, 25, 5, 0, 84, 28, 5, 0, 86, 25, 5, 0, 88, 25, 5, 0, 90, 25, 5, 0, 92, 31, 5, 0, 19, 197, 2, 72, 0, 197, 162, 72, 0, 197, 194, 73, 0, 197, 98, 74, 0, 197, 2, 75, 0, 197, 162, 75, 0, 197, 66, 76, 0, 197, 226, 76, 0, 197, 130, 77, 0, 197, 34, 79, 0, 31, 5, 0, 21, 31, 5, 0, 23, 31, 5, 0, 25, 31, 5, 0, 27, 31, 5, 0, 29, 31, 5, 0, 31, 31, 5, 0, 33, 31, 5, 0, 35, 34, 5, 0, 37, 197, 98, 71, 0, 31, 5, 0, 19, 197, 5, 226, 2, 197, 163, 223, 2, 197, 3, 224, 2, 197, 164, 226, 2, 197, 35, 227, 2, 197, 195, 227, 2, 197, 35, 228, 2, 197, 133, 228, 2, 197, 36, 229, 2, 197, 162, 229, 2, 197, 230, 229, 2, 197, 163, 230, 2, 197, 5, 231, 2, 197, 68, 192, 2, 197, 164, 231, 2, 197, 35, 232, 2, 197, 131, 232, 2, 197, 36, 233, 2, 197, 162, 233, 2, 197, 132, 234, 2, 197, 229, 233, 2, 197, 2, 235, 2, 197, 70, 235, 2, 197, 3, 193, 2, 198, 130, 136, 0, 198, 194, 138, 0, 198, 226, 139, 0, 198, 162, 140, 0, 198, 98, 141, 0, 198, 34, 142, 0, 198, 226, 142, 0, 198, 162, 143, 0, 5, 5, 170, 124, 32, 5, 170, 124, 5, 5, 172, 124, 32, 5, 172, 124, 5, 5, 174, 124, 5, 5, 176, 124, 32, 5, 176, 124, 193, 2, 178, 124, 193, 9, 178, 124, 193, 16, 178, 124, 193, 23, 178, 124, 193, 30, 178, 124, 193, 37, 178, 124, 193, 44, 178, 124, 5, 5, 180, 124, 32, 5, 180, 124, 5, 112, 180, 124, 5, 5, 182, 124, 5, 5, 184, 124, 5, 112, 184, 124, 5, 5, 186, 124, 5, 5, 188, 124, 32, 5, 188, 124, 5, 112, 188, 124, 5, 5, 190, 124, 5, 5, 192, 124, 5, 5, 194, 124, 32, 5, 194, 124, 5, 112, 194, 124, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 5, 5, 0, 21, 5, 5, 0, 23, 198, 2, 138, 0, 198, 34, 139, 0, 198, 66, 140, 0, 198, 2, 141, 0, 198, 194, 141, 0, 198, 130, 142, 0, 198, 66, 143, 0, 198, 2, 144, 0, 198, 194, 144, 0, 198, 195, 136, 0, 198, 227, 137, 0, 198, 3, 139, 0, 198, 35, 140, 0, 198, 227, 140, 0, 198, 163, 141, 0, 198, 99, 142, 0, 198, 35, 143, 0, 198, 227, 143, 0, 198, 163, 144, 0, 198, 99, 145, 0, 198, 35, 146, 0, 198, 227, 146, 0, 198, 163, 147, 0, 198, 99, 148, 0, 198, 35, 149, 0, 198, 131, 149, 0, 198, 227, 149, 0, 198, 67, 150, 0, 198, 163, 150, 0, 198, 3, 151, 0, 198, 99, 151, 0, 197, 227, 133, 0, 198, 98, 138, 0, 198, 130, 139, 0, 198, 194, 151, 0, 198, 2, 152, 0, 198, 66, 152, 0, 198, 130, 152, 0, 198, 194, 152, 0, 198, 2, 153, 0, 198, 66, 153, 0, 198, 35, 137, 0, 198, 67, 138, 0, 198, 99, 139, 0, 197, 162, 136, 0, 197, 99, 130, 0, 197, 194, 130, 0, 197, 163, 152, 0, 44, 5, 6, 122, 44, 5, 8, 122, 44, 5, 10, 122, 44, 5, 14, 122, 44, 5, 16, 122, 40, 5, 18, 122, 38, 5, 20, 122, 40, 5, 22, 122, 40, 5, 24, 122, 32, 5, 26, 122, 38, 5, 28, 122, 40, 5, 30, 122, 40, 5, 32, 122, 38, 5, 34, 122, 38, 5, 36, 122, 38, 5, 38, 122, 198, 98, 144, 0, 198, 34, 145, 0, 198, 131, 137, 0, 198, 163, 138, 0, 198, 195, 139, 0, 198, 131, 140, 0, 198, 67, 141, 0, 198, 3, 142, 0, 198, 195, 142, 0, 198, 131, 143, 0, 198, 67, 144, 0, 198, 3, 145, 0, 198, 195, 145, 0, 198, 131, 146, 0, 198, 67, 147, 0, 198, 3, 148, 0, 198, 195, 148, 0, 197, 227, 137, 0, 197, 130, 119, 0, 197, 98, 111, 0, 197, 99, 113, 0, 197, 66, 180, 0, 197, 130, 181, 0, 197, 194, 120, 0, 197, 3, 121, 0, 197, 99, 121, 0, 197, 194, 141, 0, 198, 2, 189, 0, 198, 162, 194, 0, 198, 66, 185, 0, 198, 66, 194, 0, 198, 228, 195, 0, 184, 5, 0, 74, 180, 5, 0, 76, 180, 5, 0, 78, 180, 5, 0, 80, 177, 5, 0, 82, 180, 5, 0, 84, 184, 5, 0, 86, 180, 5, 0, 88, 177, 5, 0, 90, 177, 5, 0, 92, 180, 5, 0, 72, 197, 66, 139, 0, 197, 66, 181, 0, 197, 162, 190, 0, 197, 226, 107, 0, 197, 34, 210, 0, 197, 66, 120, 0, 197, 98, 117, 0, 197, 4, 119, 0, 197, 228, 132, 0, 197, 130, 139, 0, 197, 227, 160, 0, 197, 130, 161, 0, 197, 130, 179, 0, 197, 195, 191, 0, 197, 195, 205, 0, 197, 34, 209, 0, 197, 66, 87, 0, 197, 199, 82, 0, 197, 130, 82, 0, 197, 34, 90, 0, 197, 162, 94, 0, 198, 193, 219, 0, 198, 65, 177, 0, 198, 129, 178, 0, 198, 161, 196, 0, 198, 65, 173, 0, 198, 33, 174, 0, 198, 129, 173, 0, 198, 65, 188, 0, 198, 193, 182, 0, 198, 161, 180, 0, 198, 129, 186, 0, 198, 65, 186, 0, 198, 65, 203, 0, 198, 33, 177, 0, 198, 193, 213, 0, 198, 33, 181, 0, 198, 1, 185, 0, 197, 66, 86, 0, 197, 130, 86, 0, 197, 194, 86, 0, 197, 2, 87, 0, 197, 34, 88, 0, 197, 98, 88, 0, 197, 34, 81, 0, 197, 226, 85, 0, 197, 162, 88, 0, 197, 226, 88, 0, 197, 34, 89, 0, 197, 98, 89, 0, 197, 162, 89, 0, 197, 226, 89, 0, 197, 194, 91, 0, 165, 5, 0, 42, 165, 5, 0, 44, 165, 5, 0, 46, 165, 5, 0, 48, 165, 5, 0, 50, 165, 5, 0, 52, 165, 5, 0, 54, 165, 5, 0, 56, 165, 5, 0, 58, 165, 5, 0, 60, 165, 5, 0, 62, 165, 5, 0, 64, 165, 5, 0, 66, 165, 5, 0, 68, 165, 5, 0, 70, 165, 5, 0, 72, 172, 5, 0, 74, 165, 5, 0, 76, 165, 5, 0, 78, 165, 5, 0, 80, 165, 5, 0, 82, 165, 5, 0, 84, 172, 5, 0, 86, 165, 5, 0, 88, 165, 5, 0, 90, 165, 5, 0, 92, 22, 5, 0, 42, 22, 5, 0, 44, 22, 5, 0, 46, 22, 5, 0, 48, 22, 5, 0, 50, 22, 5, 0, 52, 22, 5, 0, 54, 255, 255, 255, 255, 22, 5, 0, 58, 22, 5, 0, 60, 22, 5, 0, 62, 22, 5, 0, 64, 22, 5, 0, 66, 22, 5, 0, 68, 22, 5, 0, 70, 22, 5, 0, 72, 24, 5, 0, 74, 22, 5, 0, 76, 165, 5, 0, 90, 165, 5, 0, 92, 22, 5, 0, 42, 22, 5, 0, 44, 22, 5, 0, 46, 22, 5, 0, 48, 22, 5, 0, 50, 22, 5, 0, 52, 22, 5, 0, 54, 22, 5, 0, 56, 22, 5, 0, 58, 22, 5, 0, 60, 22, 5, 0, 62, 22, 5, 0, 64, 22, 5, 0, 66, 22, 5, 0, 68, 22, 5, 0, 70, 22, 5, 0, 72, 24, 5, 0, 74, 22, 5, 0, 76, 22, 5, 0, 78, 22, 5, 0, 80, 22, 5, 0, 82, 22, 5, 0, 84, 24, 5, 0, 86, 22, 5, 0, 88, 22, 5, 0, 90, 22, 5, 0, 92, 165, 5, 0, 42, 255, 255, 255, 255, 165, 5, 0, 46, 165, 5, 0, 48, 255, 255, 255, 255, 255, 255, 255, 255, 165, 5, 0, 54, 255, 255, 255, 255, 255, 255, 255, 255, 165, 5, 0, 60, 165, 5, 0, 62, 255, 255, 255, 255, 255, 255, 255, 255, 165, 5, 0, 68, 165, 5, 0, 70, 165, 5, 0, 72, 172, 5, 0, 74, 255, 255, 255, 255, 165, 5, 0, 78, 165, 5, 0, 80, 165, 5, 0, 82, 165, 5, 0, 84, 172, 5, 0, 86, 165, 5, 0, 88, 165, 5, 0, 90, 165, 5, 0, 92, 22, 5, 0, 42, 22, 5, 0, 44, 22, 5, 0, 46, 22, 5, 0, 48, 255, 255, 255, 255, 22, 5, 0, 52, 255, 255, 255, 255, 22, 5, 0, 56, 22, 5, 0, 58, 22, 5, 0, 60, 172, 5, 0, 74, 165, 5, 0, 76, 165, 5, 0, 78, 165, 5, 0, 80, 165, 5, 0, 82, 165, 5, 0, 84, 172, 5, 0, 86, 165, 5, 0, 88, 165, 5, 0, 90, 165, 5, 0, 92, 22, 5, 0, 42, 22, 5, 0, 44, 22, 5, 0, 46, 22, 5, 0, 48, 22, 5, 0, 50, 22, 5, 0, 52, 22, 5, 0, 54, 22, 5, 0, 56, 22, 5, 0, 58, 22, 5, 0, 60, 22, 5, 0, 62, 22, 5, 0, 64, 22, 5, 0, 66, 22, 5, 0, 68, 22, 5, 0, 70, 22, 5, 0, 72, 24, 5, 0, 74, 22, 5, 0, 76, 22, 5, 0, 78, 22, 5, 0, 80, 22, 5, 0, 82, 22, 5, 0, 84, 24, 5, 0, 86, 22, 5, 0, 88, 22, 5, 0, 90, 22, 5, 0, 92, 165, 5, 0, 42, 165, 5, 0, 44, 255, 255, 255, 255, 165, 5, 0, 48, 165, 5, 0, 50, 165, 5, 0, 52, 165, 5, 0, 54, 255, 255, 255, 255, 255, 255, 255, 255, 165, 5, 0, 60, 165, 5, 0, 62, 165, 5, 0, 64, 165, 5, 0, 66, 165, 5, 0, 68, 165, 5, 0, 70, 165, 5, 0, 72, 172, 5, 0, 74, 255, 255, 255, 255, 165, 5, 0, 78, 165, 5, 0, 80, 165, 5, 0, 82, 165, 5, 0, 84, 172, 5, 0, 86, 165, 5, 0, 88, 165, 5, 0, 90, 255, 255, 255, 255, 22, 5, 0, 42, 22, 5, 0, 44, 165, 5, 0, 58, 165, 5, 0, 60, 165, 5, 0, 62, 165, 5, 0, 64, 165, 5, 0, 66, 255, 255, 255, 255, 165, 5, 0, 70, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 165, 5, 0, 78, 165, 5, 0, 80, 165, 5, 0, 82, 165, 5, 0, 84, 172, 5, 0, 86, 165, 5, 0, 88, 165, 5, 0, 90, 255, 255, 255, 255, 22, 5, 0, 42, 22, 5, 0, 44, 22, 5, 0, 46, 22, 5, 0, 48, 22, 5, 0, 50, 22, 5, 0, 52, 22, 5, 0, 54, 22, 5, 0, 56, 22, 5, 0, 58, 22, 5, 0, 60, 22, 5, 0, 62, 22, 5, 0, 64, 22, 5, 0, 66, 22, 5, 0, 68, 165, 5, 0, 66, 165, 5, 0, 68, 165, 5, 0, 70, 165, 5, 0, 72, 172, 5, 0, 74, 165, 5, 0, 76, 165, 5, 0, 78, 165, 5, 0, 80, 165, 5, 0, 82, 165, 5, 0, 84, 172, 5, 0, 86, 165, 5, 0, 88, 165, 5, 0, 90, 165, 5, 0, 92, 22, 5, 0, 42, 22, 5, 0, 44, 22, 5, 0, 46, 22, 5, 0, 48, 22, 5, 0, 50, 22, 5, 0, 52, 22, 5, 0, 54, 22, 5, 0, 56, 22, 5, 0, 58, 22, 5, 0, 60, 22, 5, 0, 62, 22, 5, 0, 64, 22, 5, 0, 66, 22, 5, 0, 68, 22, 5, 0, 70, 22, 5, 0, 72, 24, 5, 0, 74, 22, 5, 0, 76, 22, 5, 0, 78, 22, 5, 0, 80, 22, 5, 0, 82, 22, 5, 0, 84, 24, 5, 0, 86, 22, 5, 0, 88, 22, 5, 0, 90, 22, 5, 0, 92, 165, 5, 0, 42, 165, 5, 0, 44, 165, 5, 0, 46, 165, 5, 0, 48, 165, 5, 0, 50, 165, 5, 0, 52, 165, 5, 0, 54, 165, 5, 0, 56, 165, 5, 0, 58, 165, 5, 0, 60, 165, 5, 0, 62, 165, 5, 0, 64, 165, 5, 0, 66, 165, 5, 0, 68, 165, 5, 0, 70, 165, 5, 0, 72, 172, 5, 0, 74, 165, 5, 0, 76, 165, 5, 0, 78, 165, 5, 0, 80, 165, 5, 0, 82, 165, 5, 0, 84, 172, 5, 0, 86, 165, 5, 0, 88, 165, 5, 0, 90, 165, 5, 0, 92, 22, 5, 0, 42, 22, 5, 0, 44, 22, 5, 0, 46, 22, 5, 0, 48, 22, 5, 0, 50, 22, 5, 0, 52, 22, 5, 0, 54, 22, 5, 0, 56, 22, 5, 0, 58, 22, 5, 0, 60, 168, 5, 0, 62, 168, 5, 0, 64, 168, 5, 0, 66, 168, 5, 0, 68, 168, 5, 0, 70, 168, 5, 0, 72, 176, 5, 0, 74, 168, 5, 0, 76, 168, 5, 0, 78, 168, 5, 0, 80, 168, 5, 0, 82, 168, 5, 0, 84, 176, 5, 0, 86, 168, 5, 0, 88, 168, 5, 0, 90, 168, 5, 0, 92, 25, 5, 0, 42, 25, 5, 0, 44, 25, 5, 0, 46, 25, 5, 0, 48, 25, 5, 0, 50, 25, 5, 0, 52, 25, 5, 0, 54, 25, 5, 0, 56, 25, 5, 0, 58, 25, 5, 0, 60, 25, 5, 0, 62, 25, 5, 0, 64, 25, 5, 0, 66, 25, 5, 0, 68, 25, 5, 0, 70, 25, 5, 0, 72, 22, 5, 0, 62, 22, 5, 0, 64, 22, 5, 0, 66, 22, 5, 0, 68, 255, 255, 255, 255, 22, 5, 0, 72, 24, 5, 0, 74, 22, 5, 0, 76, 22, 5, 0, 78, 22, 5, 0, 80, 22, 5, 0, 82, 22, 5, 0, 84, 24, 5, 0, 86, 22, 5, 0, 88, 22, 5, 0, 90, 22, 5, 0, 92, 165, 5, 0, 42, 165, 5, 0, 44, 165, 5, 0, 46, 165, 5, 0, 48, 165, 5, 0, 50, 165, 5, 0, 52, 165, 5, 0, 54, 165, 5, 0, 56, 165, 5, 0, 58, 165, 5, 0, 60, 165, 5, 0, 62, 165, 5, 0, 64, 165, 5, 0, 66, 165, 5, 0, 68, 165, 5, 0, 70, 165, 5, 0, 72, 184, 5, 0, 74, 180, 5, 0, 76, 180, 5, 0, 78, 180, 5, 0, 80, 177, 5, 0, 82, 180, 5, 0, 84, 184, 5, 0, 86, 180, 5, 0, 88, 177, 5, 0, 90, 177, 5, 0, 92, 197, 66, 138, 0, 197, 34, 159, 0, 197, 226, 190, 0, 197, 34, 192, 0, 197, 163, 182, 0, 197, 34, 210, 0, 168, 5, 0, 42, 168, 5, 0, 44, 168, 5, 0, 46, 168, 5, 0, 48, 168, 5, 0, 50, 168, 5, 0, 52, 168, 5, 0, 54, 168, 5, 0, 56, 168, 5, 0, 58, 168, 5, 0, 60, 168, 5, 0, 62, 168, 5, 0, 64, 168, 5, 0, 66, 168, 5, 0, 68, 168, 5, 0, 70, 168, 5, 0, 72, 180, 5, 0, 80, 177, 5, 0, 82, 184, 5, 0, 86, 43, 5, 0, 42, 46, 5, 20, 43, 44, 5, 22, 43, 32, 5, 16, 43, 43, 5, 0, 44, 43, 5, 0, 48, 43, 5, 0, 50, 38, 5, 24, 51, 44, 5, 28, 51, 32, 5, 40, 51, 43, 5, 0, 54, 32, 5, 14, 59, 43, 5, 0, 62, 43, 5, 0, 66, 44, 5, 30, 69, 43, 5, 0, 70, 46, 5, 30, 71, 32, 5, 44, 71, 32, 5, 46, 71, 43, 5, 0, 72, 43, 5, 0, 80, 43, 5, 0, 82, 32, 5, 10, 83, 46, 5, 40, 83, 43, 5, 0, 84, 32, 5, 102, 93, 44, 5, 8, 96, 40, 5, 10, 96, 49, 5, 14, 96, 5, 5, 6, 85, 5, 5, 6, 87, 5, 5, 6, 93, 5, 5, 28, 93, 5, 5, 100, 93, 5, 5, 102, 93, 5, 5, 12, 96, 5, 5, 40, 96, 5, 5, 52, 96, 5, 5, 62, 96, 5, 5, 84, 96, 5, 5, 100, 97, 180, 5, 0, 42, 197, 162, 108, 0, 177, 5, 0, 44, 5, 5, 10, 45, 177, 5, 0, 48, 180, 5, 0, 50, 174, 5, 20, 51, 177, 5, 0, 54, 180, 5, 0, 56, 180, 5, 0, 58, 177, 5, 0, 60, 180, 5, 0, 62, 180, 5, 0, 64, 180, 5, 0, 66, 180, 5, 0, 68, 5, 5, 8, 69, 180, 5, 0, 70, 174, 5, 58, 71, 180, 5, 0, 72, 180, 5, 0, 76, 196, 196, 156, 42, 196, 196, 5, 42, 196, 182, 156, 42, 196, 182, 5, 42, 197, 131, 98, 0, 197, 35, 98, 0, 197, 67, 99, 0, 197, 227, 98, 0, 197, 195, 100, 0, 197, 99, 100, 0, 197, 3, 100, 0, 197, 163, 99, 0, 197, 3, 105, 0, 197, 163, 104, 0, 197, 131, 95, 0, 197, 35, 95, 0, 197, 67, 96, 0, 197, 227, 95, 0, 197, 195, 97, 0, 197, 99, 97, 0, 197, 3, 97, 0, 197, 163, 96, 0, 197, 67, 104, 0, 197, 227, 103, 0, 196, 196, 156, 50, 196, 196, 5, 50, 196, 182, 156, 50, 196, 182, 5, 50, 196, 154, 156, 50, 196, 154, 5, 50, 197, 195, 124, 0, 197, 99, 124, 0, 196, 200, 156, 42, 196, 200, 5, 42, 196, 156, 156, 44, 196, 156, 5, 44, 196, 196, 156, 44, 196, 196, 5, 44, 196, 210, 156, 44, 196, 210, 5, 44, 197, 163, 114, 0, 197, 67, 114, 0, 196, 156, 156, 48, 196, 156, 5, 48, 196, 196, 156, 48, 196, 196, 5, 48, 196, 210, 156, 48, 196, 210, 5, 48, 196, 160, 156, 48, 196, 160, 5, 48, 196, 204, 156, 48, 196, 204, 5, 48, 197, 67, 129, 0, 197, 227, 128, 0, 197, 131, 128, 0, 197, 35, 128, 0, 196, 204, 156, 50, 196, 204, 5, 50, 196, 208, 156, 50, 196, 208, 5, 50, 197, 195, 127, 0, 197, 99, 127, 0, 196, 156, 156, 52, 196, 156, 5, 52, 197, 194, 147, 0, 197, 226, 159, 0, 197, 100, 105, 0, 197, 194, 113, 0, 197, 162, 116, 0, 197, 34, 117, 0, 197, 196, 115, 0, 197, 163, 118, 0, 197, 2, 120, 0, 197, 66, 135, 0, 197, 194, 135, 0, 197, 162, 137, 0, 197, 130, 141, 0, 197, 226, 144, 0, 197, 162, 145, 0, 197, 2, 147, 0, 197, 130, 150, 0, 197, 194, 150, 0, 197, 3, 151, 0, 197, 2, 153, 0, 197, 162, 154, 0, 197, 67, 156, 0, 197, 163, 157, 0, 197, 2, 182, 0, 197, 132, 180, 0, 197, 67, 182, 0, 197, 2, 183, 0, 197, 194, 186, 0, 197, 98, 193, 0, 197, 226, 209, 0, 197, 35, 206, 0, 197, 3, 107, 0, 196, 138, 160, 86, 196, 138, 5, 86, 196, 136, 160, 86, 196, 136, 5, 86, 196, 150, 160, 86, 196, 150, 5, 86, 196, 156, 160, 86, 196, 156, 5, 86, 196, 196, 160, 86, 196, 196, 5, 86, 196, 156, 156, 88, 196, 156, 5, 88, 196, 150, 156, 88, 196, 150, 5, 88, 196, 156, 156, 90, 196, 156, 5, 90, 196, 142, 156, 92, 196, 142, 5, 92, 196, 196, 156, 92, 196, 196, 5, 92, 196, 210, 156, 92, 196, 210, 5, 92, 196, 210, 5, 56, 196, 150, 5, 80, 196, 146, 5, 86, 196, 146, 5, 90, 197, 34, 113, 0, 197, 98, 190, 0, 5, 5, 18, 79, 5, 5, 20, 79, 197, 162, 192, 0, 5, 5, 28, 49, 5, 5, 6, 43, 5, 5, 14, 43, 5, 5, 16, 43, 5, 5, 12, 45, 5, 5, 6, 47, 5, 5, 6, 49, 5, 5, 8, 49, 5, 5, 6, 51, 5, 5, 40, 51, 5, 5, 14, 59, 5, 5, 8, 61, 5, 5, 6, 63, 5, 5, 10, 65, 5, 5, 6, 67, 5, 5, 10, 69, 5, 5, 6, 71, 5, 5, 32, 71, 5, 5, 8, 71, 5, 5, 34, 71, 5, 5, 26, 71, 5, 5, 14, 71, 5, 5, 60, 71, 5, 5, 44, 71, 5, 5, 46, 71, 5, 5, 6, 73, 5, 5, 14, 77, 5, 5, 22, 77, 5, 5, 6, 81, 5, 5, 6, 83, 5, 5, 10, 83, 5, 5, 12, 83, 5, 5, 44, 83, 160, 5, 16, 65, 5, 5, 16, 65, 160, 5, 18, 65, 160, 5, 8, 73, 160, 5, 34, 77, 5, 5, 8, 43, 5, 5, 10, 81, 160, 5, 14, 57, 5, 5, 14, 57, 160, 5, 12, 63, 5, 5, 12, 63, 160, 5, 22, 93, 5, 5, 22, 93, 158, 5, 22, 43, 160, 5, 12, 67, 160, 5, 20, 43, 160, 5, 28, 43, 5, 5, 14, 85, 160, 5, 10, 87, 5, 5, 10, 87, 5, 5, 16, 85, 160, 5, 16, 57, 5, 5, 16, 57, 5, 5, 26, 73, 5, 5, 18, 51, 5, 5, 30, 77, 5, 5, 48, 71, 5, 5, 22, 51, 46, 5, 0, 60, 180, 5, 0, 84, 160, 5, 16, 79, 160, 5, 20, 93, 5, 5, 18, 45, 5, 5, 14, 49, 5, 5, 14, 53, 5, 5, 14, 55, 5, 5, 8, 63, 5, 5, 28, 65, 5, 5, 10, 67, 5, 5, 20, 69, 5, 5, 14, 73, 5, 5, 26, 77, 5, 5, 12, 79, 5, 5, 26, 79, 5, 5, 10, 85, 5, 5, 6, 89, 5, 5, 12, 93, 5, 5, 10, 43, 5, 5, 26, 43, 5, 5, 20, 49, 5, 5, 14, 51, 5, 5, 30, 51, 5, 5, 38, 51, 5, 5, 26, 51, 5, 5, 20, 59, 5, 5, 38, 71, 5, 5, 32, 79, 5, 5, 28, 83, 5, 5, 32, 93, 46, 5, 28, 43, 43, 5, 0, 46, 46, 5, 16, 47, 46, 112, 0, 48, 46, 5, 36, 51, 197, 162, 133, 0, 197, 98, 133, 0, 197, 66, 143, 0, 197, 2, 143, 0, 197, 98, 160, 0, 197, 34, 160, 0, 197, 226, 184, 0, 197, 162, 184, 0, 197, 98, 189, 0, 197, 34, 189, 0, 160, 5, 12, 57, 160, 5, 36, 51, 160, 5, 6, 55, 160, 5, 24, 65, 160, 5, 8, 59, 5, 5, 6, 75, 160, 5, 20, 63, 160, 5, 24, 81, 160, 5, 12, 61, 160, 5, 16, 89, 174, 5, 24, 45, 5, 5, 24, 45, 160, 5, 56, 71, 5, 5, 56, 71, 160, 5, 24, 83, 5, 5, 24, 83, 160, 5, 12, 43, 5, 5, 12, 43, 160, 5, 22, 59, 5, 5, 22, 59, 160, 5, 32, 83, 5, 5, 32, 83, 193, 33, 152, 103, 193, 40, 152, 103, 193, 47, 152, 103, 193, 54, 152, 103, 193, 61, 152, 103, 193, 68, 152, 103, 193, 75, 152, 103, 255, 255, 255, 255, 193, 201, 152, 103, 193, 208, 152, 103, 193, 215, 152, 103, 193, 222, 152, 103, 193, 229, 152, 103, 193, 236, 152, 103, 193, 243, 152, 103, 255, 255, 255, 255, 5, 5, 24, 43, 5, 5, 18, 43, 5, 5, 8, 51, 5, 5, 10, 51, 5, 5, 16, 51, 5, 5, 8, 53, 5, 5, 8, 55, 5, 5, 26, 65, 5, 5, 20, 65, 5, 5, 22, 65, 5, 5, 14, 67, 5, 5, 26, 69, 5, 5, 32, 69, 5, 5, 10, 71, 5, 5, 28, 71, 5, 5, 36, 71, 44, 5, 78, 96, 40, 5, 80, 96, 46, 5, 0, 58, 46, 5, 0, 76, 46, 5, 0, 82, 46, 5, 0, 84, 51, 5, 8, 96, 48, 5, 10, 96, 48, 5, 60, 96, 51, 5, 78, 96, 48, 5, 80, 96, 5, 5, 14, 83, 5, 5, 14, 45, 5, 5, 12, 49, 5, 5, 12, 53, 5, 5, 8, 67, 5, 5, 12, 69, 5, 5, 12, 73, 5, 5, 18, 77, 5, 5, 40, 77, 5, 5, 10, 79, 5, 5, 12, 81, 5, 5, 10, 93, 5, 5, 20, 55, 44, 5, 120, 97, 5, 112, 0, 54, 197, 66, 194, 0, 5, 5, 18, 59, 5, 5, 26, 59, 5, 5, 8, 73, 5, 5, 26, 83, 5, 5, 50, 83, 162, 5, 0, 58, 197, 2, 140, 0, 197, 163, 140, 0, 197, 66, 142, 0, 162, 5, 0, 84, 197, 194, 206, 0, 197, 99, 207, 0, 197, 68, 208, 0, 197, 194, 142, 0, 162, 5, 0, 88, 197, 226, 210, 0, 197, 131, 211, 0, 162, 5, 0, 64, 162, 5, 0, 46, 162, 5, 0, 48, 162, 5, 0, 66, 19, 5, 0, 58, 197, 194, 139, 0, 197, 67, 140, 0, 197, 2, 142, 0, 19, 5, 0, 84, 197, 130, 206, 0, 197, 3, 207, 0, 197, 196, 207, 0, 197, 130, 142, 0, 19, 5, 0, 88, 197, 162, 210, 0, 197, 35, 211, 0, 19, 5, 0, 64, 19, 5, 0, 46, 19, 5, 0, 48, 19, 5, 0, 66, 197, 2, 181, 0, 197, 162, 160, 0, 197, 194, 4, 1, 197, 98, 154, 0, 197, 130, 143, 0, 197, 194, 143, 0, 197, 226, 154, 0, 197, 66, 134, 0, 197, 67, 116, 0, 197, 4, 144, 0, 197, 194, 181, 0, 197, 66, 161, 0, 197, 2, 5, 1, 197, 66, 5, 1, 197, 162, 155, 0, 197, 2, 116, 0, 197, 162, 134, 0, 197, 131, 144, 0, 197, 227, 155, 0, 197, 131, 134, 0, 197, 131, 194, 0, 197, 130, 5, 1, 197, 162, 150, 0, 197, 130, 120, 0, 197, 34, 145, 0, 197, 98, 132, 0, 197, 130, 162, 0, 197, 194, 5, 1, 197, 162, 156, 0, 197, 162, 117, 0, 197, 98, 145, 0, 197, 227, 156, 0, 197, 227, 117, 0, 197, 2, 118, 0, 197, 227, 145, 0, 197, 67, 157, 0, 197, 67, 118, 0, 197, 98, 118, 0, 197, 67, 146, 0, 197, 131, 153, 0, 197, 228, 153, 0, 197, 2, 135, 0, 197, 163, 146, 0, 197, 3, 158, 0, 197, 227, 134, 0, 197, 227, 185, 0, 197, 69, 186, 0, 197, 230, 186, 0, 197, 66, 183, 0, 197, 2, 163, 0, 197, 2, 6, 1, 197, 162, 158, 0, 197, 226, 183, 0, 197, 66, 163, 0, 197, 66, 6, 1, 197, 226, 158, 0, 197, 66, 147, 0, 197, 34, 159, 0, 197, 34, 184, 0, 197, 130, 163, 0, 197, 130, 6, 1, 197, 98, 159, 0, 197, 130, 147, 0, 197, 162, 159, 0, 255, 255, 255, 255, 255, 255, 255, 255, 160, 5, 8, 87, 5, 5, 8, 87, 160, 5, 12, 47, 158, 5, 14, 79, 160, 5, 12, 93, 160, 5, 10, 49, 5, 5, 10, 49, 160, 5, 8, 79, 5, 5, 8, 79, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 162, 209, 0, 197, 98, 209, 0, 160, 5, 24, 93, 5, 5, 24, 93, 160, 5, 40, 93, 5, 5, 40, 93, 160, 5, 42, 93, 5, 5, 42, 93, 160, 5, 46, 93, 5, 5, 46, 93, 160, 5, 54, 93, 5, 5, 54, 93, 160, 5, 56, 93, 5, 5, 56, 93, 158, 5, 58, 93, 5, 5, 58, 93, 44, 5, 58, 93, 5, 5, 26, 49, 5, 5, 36, 65, 5, 5, 20, 67, 5, 5, 28, 69, 5, 5, 58, 77, 5, 5, 60, 77, 5, 5, 22, 81, 5, 5, 60, 93, 160, 120, 0, 48, 5, 120, 0, 48, 160, 112, 0, 52, 5, 112, 0, 52, 160, 112, 0, 54, 160, 5, 22, 55, 5, 5, 22, 55, 197, 131, 125, 0, 197, 35, 125, 0, 197, 3, 127, 0, 197, 163, 126, 0, 197, 67, 126, 0, 197, 227, 125, 0, 197, 3, 130, 0, 197, 163, 129, 0, 196, 182, 156, 58, 196, 182, 5, 58, 196, 196, 156, 58, 196, 196, 5, 58, 196, 196, 156, 70, 196, 196, 5, 70, 196, 182, 156, 70, 196, 182, 5, 70, 197, 35, 164, 0, 197, 195, 163, 0, 197, 227, 164, 0, 197, 131, 164, 0, 197, 99, 166, 0, 197, 3, 166, 0, 197, 163, 165, 0, 197, 67, 165, 0, 197, 99, 178, 0, 197, 3, 178, 0, 197, 163, 174, 0, 197, 67, 174, 0, 197, 99, 175, 0, 197, 3, 175, 0, 197, 227, 176, 0, 197, 131, 176, 0, 43, 5, 0, 52, 32, 5, 14, 61, 46, 5, 6, 55, 46, 5, 34, 83, 46, 5, 16, 59, 46, 5, 24, 59, 46, 5, 8, 59, 32, 5, 18, 59, 46, 5, 12, 61, 32, 5, 30, 65, 32, 5, 28, 65, 46, 5, 6, 65, 46, 5, 12, 67, 32, 5, 46, 83, 46, 5, 14, 69, 32, 5, 22, 69, 46, 5, 6, 69, 46, 5, 50, 71, 32, 5, 24, 73, 44, 5, 14, 79, 44, 5, 22, 79, 32, 5, 14, 81, 46, 5, 20, 83, 46, 5, 48, 83, 32, 5, 6, 83, 46, 5, 12, 85, 46, 5, 20, 85, 43, 5, 0, 92, 46, 5, 16, 93, 46, 5, 18, 93, 44, 5, 26, 93, 51, 5, 30, 96, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 160, 5, 18, 57, 5, 5, 18, 57, 5, 5, 12, 59, 197, 130, 135, 0, 197, 66, 179, 0, 5, 5, 42, 83, 5, 5, 20, 53, 5, 5, 22, 73, 5, 5, 16, 67, 5, 5, 10, 59, 5, 5, 18, 67, 196, 164, 156, 54, 196, 164, 5, 54, 196, 156, 156, 56, 196, 156, 5, 56, 196, 196, 156, 56, 196, 196, 5, 56, 196, 150, 156, 56, 196, 150, 5, 56, 196, 160, 156, 56, 196, 160, 5, 56, 196, 206, 156, 56, 196, 206, 5, 56, 196, 208, 156, 58, 196, 208, 5, 58, 197, 227, 138, 0, 197, 131, 138, 0, 196, 136, 156, 62, 196, 136, 5, 62, 196, 196, 156, 62, 196, 196, 5, 62, 196, 210, 156, 62, 196, 210, 5, 62, 196, 196, 156, 64, 196, 196, 5, 64, 197, 99, 148, 0, 197, 3, 148, 0, 196, 210, 156, 64, 196, 210, 5, 64, 196, 204, 156, 64, 196, 204, 5, 64, 196, 136, 156, 66, 196, 136, 5, 66, 24, 5, 0, 86, 22, 5, 0, 88, 22, 5, 0, 90, 22, 5, 0, 92, 32, 5, 6, 59, 32, 5, 6, 61, 255, 255, 255, 255, 255, 255, 255, 255, 177, 5, 6, 96, 165, 5, 8, 96, 160, 5, 10, 96, 166, 5, 14, 96, 177, 5, 16, 96, 172, 5, 24, 96, 172, 5, 28, 96, 172, 5, 30, 96, 177, 5, 32, 96, 176, 5, 36, 96, 172, 5, 38, 96, 168, 5, 42, 96, 172, 5, 44, 96, 172, 5, 46, 96, 172, 5, 48, 96, 177, 5, 50, 96, 168, 5, 60, 96, 172, 5, 30, 96, 172, 5, 66, 96, 172, 5, 74, 96, 177, 5, 76, 96, 165, 5, 78, 96, 160, 5, 80, 96, 172, 5, 82, 96, 160, 5, 14, 63, 5, 5, 14, 63, 160, 5, 16, 63, 5, 5, 16, 63, 160, 5, 18, 63, 5, 5, 18, 63, 160, 5, 8, 65, 5, 5, 8, 65, 160, 5, 12, 65, 5, 5, 12, 65, 160, 5, 52, 71, 5, 5, 52, 71, 160, 5, 42, 71, 5, 5, 42, 71, 197, 2, 180, 0, 197, 194, 179, 0, 160, 5, 10, 73, 5, 5, 10, 73, 160, 5, 18, 73, 5, 5, 18, 73, 160, 5, 20, 73, 5, 5, 20, 73, 160, 5, 8, 75, 5, 5, 8, 75, 160, 5, 10, 75, 5, 5, 10, 75, 174, 5, 12, 77, 5, 5, 12, 77, 160, 5, 62, 77, 5, 5, 62, 77, 160, 5, 8, 85, 5, 5, 8, 85, 197, 35, 176, 0, 197, 195, 175, 0, 197, 163, 177, 0, 197, 67, 177, 0, 196, 196, 156, 82, 196, 196, 5, 82, 196, 182, 156, 82, 196, 182, 5, 82, 197, 3, 202, 0, 197, 163, 201, 0, 197, 195, 202, 0, 197, 99, 202, 0, 197, 67, 204, 0, 197, 227, 203, 0, 197, 131, 203, 0, 197, 35, 203, 0, 197, 3, 205, 0, 197, 163, 204, 0, 196, 138, 156, 90, 196, 138, 5, 90, 196, 196, 156, 90, 196, 196, 5, 90, 196, 182, 156, 90, 196, 182, 5, 90, 196, 154, 156, 90, 196, 154, 5, 90, 197, 66, 150, 0, 197, 2, 150, 0, 160, 5, 18, 85, 5, 5, 18, 85, 160, 5, 12, 91, 5, 5, 12, 91, 196, 156, 156, 66, 196, 156, 5, 66, 196, 196, 156, 66, 196, 196, 5, 66, 196, 156, 156, 68, 196, 156, 5, 68, 196, 196, 156, 68, 196, 196, 5, 68, 196, 210, 156, 68, 196, 210, 5, 68, 196, 204, 156, 68, 196, 204, 5, 68, 197, 99, 168, 0, 197, 3, 168, 0, 197, 35, 169, 0, 197, 195, 168, 0, 197, 163, 173, 0, 197, 67, 173, 0, 197, 227, 172, 0, 197, 131, 172, 0, 196, 136, 156, 72, 196, 136, 5, 72, 196, 156, 156, 72, 196, 156, 5, 72, 196, 156, 156, 76, 196, 156, 5, 76, 196, 196, 156, 76, 196, 196, 5, 76, 197, 131, 185, 0, 197, 35, 185, 0, 196, 210, 156, 76, 196, 210, 5, 76, 196, 156, 156, 78, 196, 156, 5, 78, 196, 196, 156, 78, 196, 196, 5, 78, 197, 3, 188, 0, 197, 163, 187, 0, 197, 195, 188, 0, 197, 99, 188, 0, 197, 3, 190, 0, 197, 163, 189, 0, 196, 156, 156, 80, 196, 156, 5, 80, 196, 196, 156, 80, 196, 196, 5, 80, 196, 210, 156, 80, 196, 210, 5, 80, 196, 204, 156, 80, 196, 204, 5, 80, 196, 198, 156, 82, 196, 198, 5, 82, 196, 208, 156, 82, 196, 208, 5, 82, 196, 204, 156, 82, 196, 204, 5, 82, 197, 67, 200, 0, 197, 227, 199, 0, 197, 3, 201, 0, 197, 163, 200, 0, 196, 154, 156, 84, 196, 154, 5, 84, 196, 196, 156, 84, 196, 196, 5, 84, 32, 5, 12, 87, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 226, 212, 0, 197, 98, 219, 0, 197, 35, 215, 0, 197, 163, 221, 0, 197, 99, 213, 0, 197, 227, 219, 0, 197, 227, 216, 0, 197, 99, 223, 0, 197, 34, 213, 0, 197, 162, 219, 0, 197, 131, 215, 0, 197, 3, 222, 0, 197, 195, 213, 0, 197, 67, 220, 0, 197, 67, 217, 0, 197, 195, 223, 0, 197, 194, 229, 0, 197, 194, 231, 0, 197, 3, 231, 0, 197, 3, 233, 0, 197, 67, 230, 0, 197, 67, 232, 0, 255, 255, 255, 255, 255, 255, 255, 255, 197, 2, 230, 0, 197, 2, 232, 0, 197, 99, 231, 0, 197, 99, 233, 0, 197, 163, 230, 0, 197, 163, 232, 0, 255, 255, 255, 255, 255, 255, 255, 255, 197, 163, 218, 0, 197, 35, 225, 0, 197, 228, 215, 0, 197, 100, 222, 0, 197, 36, 214, 0, 197, 164, 220, 0, 197, 164, 217, 0, 197, 36, 224, 0, 197, 3, 219, 0, 197, 131, 225, 0, 197, 100, 216, 0, 197, 228, 222, 0, 197, 164, 214, 0, 197, 36, 221, 0, 197, 36, 218, 0, 197, 164, 224, 0, 197, 131, 240, 0, 197, 3, 247, 0, 197, 196, 237, 0, 197, 68, 244, 0, 197, 4, 236, 0, 197, 132, 242, 0, 197, 132, 239, 0, 197, 4, 246, 0, 197, 227, 240, 0, 197, 99, 247, 0, 197, 68, 238, 0, 197, 196, 244, 0, 197, 132, 236, 0, 197, 4, 243, 0, 197, 4, 240, 0, 197, 132, 246, 0, 197, 2, 21, 1, 197, 130, 27, 1, 197, 67, 23, 1, 197, 195, 29, 1, 197, 131, 21, 1, 197, 3, 28, 1, 197, 3, 25, 1, 197, 131, 31, 1, 197, 66, 21, 1, 197, 194, 27, 1, 197, 163, 23, 1, 197, 35, 30, 1, 197, 227, 21, 1, 197, 99, 28, 1, 197, 99, 25, 1, 197, 227, 31, 1, 197, 194, 226, 0, 197, 226, 225, 0, 197, 66, 234, 0, 197, 194, 233, 0, 197, 162, 248, 0, 197, 194, 247, 0, 197, 162, 0, 1, 197, 34, 0, 1, 197, 66, 11, 1, 197, 194, 10, 1, 197, 98, 17, 1, 197, 162, 16, 1, 197, 226, 34, 1, 197, 2, 34, 1, 255, 255, 255, 255, 255, 255, 255, 255, 197, 194, 234, 0, 197, 66, 241, 0, 197, 3, 237, 0, 197, 131, 243, 0, 197, 67, 235, 0, 197, 195, 241, 0, 197, 195, 238, 0, 197, 67, 245, 0, 197, 2, 235, 0, 197, 130, 241, 0, 197, 99, 237, 0, 197, 227, 243, 0, 197, 163, 235, 0, 197, 35, 242, 0, 197, 35, 239, 0, 197, 163, 245, 0, 197, 162, 250, 0, 197, 98, 253, 0, 197, 227, 251, 0, 197, 163, 254, 0, 197, 35, 251, 0, 197, 227, 253, 0, 197, 163, 252, 0, 197, 99, 255, 0, 197, 226, 250, 0, 197, 162, 253, 0, 197, 67, 252, 0, 197, 3, 255, 0, 197, 131, 251, 0, 197, 67, 254, 0, 197, 3, 253, 0, 197, 195, 255, 0, 197, 194, 6, 1, 197, 194, 8, 1, 197, 3, 8, 1, 197, 3, 10, 1, 197, 67, 7, 1, 197, 67, 9, 1, 255, 255, 255, 255, 255, 255, 255, 255, 197, 2, 7, 1, 197, 2, 9, 1, 197, 99, 8, 1, 197, 99, 10, 1, 197, 163, 7, 1, 197, 163, 9, 1, 255, 255, 255, 255, 255, 255, 255, 255, 197, 130, 12, 1, 197, 226, 13, 1, 197, 35, 13, 1, 197, 35, 15, 1, 197, 195, 12, 1, 197, 99, 14, 1, 197, 131, 13, 1, 197, 227, 15, 1, 255, 255, 255, 255, 197, 34, 14, 1, 255, 255, 255, 255, 197, 131, 15, 1, 255, 255, 255, 255, 197, 195, 14, 1, 255, 255, 255, 255, 197, 67, 16, 1, 174, 5, 98, 96, 5, 5, 98, 96, 160, 5, 100, 96, 5, 5, 100, 96, 160, 5, 102, 96, 5, 5, 102, 96, 160, 5, 104, 96, 5, 5, 104, 96, 160, 5, 106, 96, 5, 5, 106, 96, 160, 5, 110, 96, 5, 5, 110, 96, 160, 5, 112, 96, 5, 5, 112, 96, 160, 5, 114, 96, 5, 5, 114, 96, 160, 5, 116, 96, 5, 5, 116, 96, 174, 5, 118, 96, 5, 5, 118, 96, 174, 5, 120, 96, 5, 5, 120, 96, 160, 5, 124, 96, 5, 5, 124, 96, 160, 5, 126, 96, 5, 5, 126, 96, 160, 5, 128, 96, 5, 5, 128, 96, 160, 5, 134, 96, 5, 5, 134, 96, 160, 5, 136, 96, 5, 5, 136, 96, 160, 5, 138, 96, 5, 5, 138, 96, 160, 5, 140, 96, 5, 5, 140, 96, 160, 5, 142, 96, 5, 5, 142, 96, 160, 5, 144, 96, 5, 5, 144, 96, 160, 5, 146, 96, 5, 5, 146, 96, 160, 5, 148, 96, 5, 5, 148, 96, 160, 5, 150, 96, 5, 5, 150, 96, 160, 5, 152, 96, 5, 5, 152, 96, 160, 5, 154, 96, 5, 5, 154, 96, 160, 5, 208, 96, 5, 5, 208, 96, 160, 5, 210, 96, 5, 5, 210, 96, 160, 5, 108, 96, 5, 5, 108, 96, 160, 5, 122, 96, 5, 5, 122, 96, 160, 5, 130, 96, 5, 5, 130, 96, 160, 5, 132, 96, 5, 5, 132, 96, 160, 5, 156, 96, 5, 5, 156, 96, 160, 5, 158, 96, 5, 5, 158, 96, 160, 5, 164, 96, 5, 5, 164, 96, 160, 5, 166, 96, 5, 5, 166, 96, 160, 5, 168, 96, 5, 5, 168, 96, 160, 5, 176, 96, 5, 5, 176, 96, 160, 5, 180, 96, 5, 5, 180, 96, 160, 5, 182, 96, 5, 5, 182, 96, 160, 5, 184, 96, 5, 5, 184, 96, 160, 5, 186, 96, 5, 5, 186, 96, 160, 5, 188, 96, 5, 5, 188, 96, 160, 5, 190, 96, 5, 5, 190, 96, 160, 5, 196, 96, 5, 5, 196, 96, 160, 5, 200, 96, 5, 5, 200, 96, 160, 5, 202, 96, 5, 5, 202, 96, 160, 5, 204, 96, 5, 5, 204, 96, 160, 5, 212, 96, 5, 5, 212, 96, 32, 5, 12, 97, 32, 5, 14, 97, 32, 5, 16, 97, 32, 5, 26, 97, 32, 5, 44, 97, 32, 5, 50, 97, 32, 5, 84, 97, 32, 5, 98, 97, 32, 5, 114, 97, 16, 5, 120, 97, 16, 5, 138, 97, 32, 5, 142, 97, 32, 5, 150, 97, 32, 5, 156, 97, 32, 5, 162, 97, 32, 5, 182, 97, 32, 5, 192, 97, 32, 5, 200, 97, 32, 5, 218, 97, 32, 5, 222, 97, 198, 193, 159, 0, 197, 2, 46, 1, 32, 5, 6, 97, 32, 5, 40, 97, 32, 5, 82, 97, 198, 225, 154, 0, 198, 33, 157, 0, 32, 5, 240, 97, 198, 161, 157, 0, 198, 65, 158, 0, 198, 161, 158, 0, 198, 65, 159, 0, 32, 5, 14, 97, 32, 5, 26, 97, 16, 5, 138, 97, 32, 5, 156, 97, 32, 5, 162, 97, 32, 5, 162, 97, 16, 5, 226, 97, 198, 33, 157, 0, 198, 225, 154, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 160, 5, 6, 99, 160, 5, 10, 99, 160, 5, 14, 99, 160, 5, 18, 99, 160, 5, 22, 99, 160, 5, 26, 99, 160, 5, 30, 99, 160, 5, 38, 99, 160, 5, 42, 99, 160, 5, 46, 99, 160, 5, 50, 99, 160, 5, 54, 99, 160, 5, 58, 99, 160, 5, 66, 99, 160, 5, 70, 99, 160, 5, 74, 99, 160, 5, 30, 97, 5, 5, 30, 97, 160, 5, 68, 97, 5, 5, 68, 97, 198, 161, 153, 0, 193, 9, 46, 97, 160, 5, 210, 97, 5, 5, 210, 97, 160, 5, 64, 97, 5, 5, 64, 97, 160, 5, 168, 97, 5, 5, 168, 97, 198, 161, 154, 0, 193, 2, 164, 97, 198, 161, 156, 0, 193, 9, 194, 97, 160, 5, 198, 97, 5, 5, 198, 97, 198, 225, 156, 0, 193, 9, 202, 97, 198, 161, 155, 0, 193, 9, 190, 97, 160, 5, 220, 97, 5, 5, 220, 97, 172, 5, 138, 97, 16, 5, 138, 97, 172, 5, 138, 97, 16, 5, 138, 97, 44, 5, 226, 97, 44, 5, 232, 97, 32, 5, 180, 97, 198, 1, 158, 0, 160, 5, 52, 97, 5, 5, 52, 97, 198, 225, 153, 0, 193, 9, 56, 97, 160, 5, 60, 97, 5, 5, 60, 97, 160, 5, 76, 97, 5, 5, 76, 97, 174, 5, 82, 97, 5, 5, 82, 97, 198, 1, 155, 0, 193, 2, 178, 97, 198, 33, 156, 0, 193, 30, 190, 97, 198, 1, 157, 0, 193, 2, 224, 97, 160, 5, 228, 97, 5, 5, 228, 97, 198, 97, 157, 0, 193, 9, 236, 97, 198, 129, 157, 0, 193, 2, 242, 97, 198, 193, 157, 0, 193, 9, 242, 97, 198, 129, 158, 0, 193, 23, 246, 97, 198, 225, 158, 0, 193, 37, 246, 97, 198, 33, 159, 0, 193, 51, 246, 97, 198, 97, 160, 0, 193, 93, 246, 97, 174, 5, 70, 98, 174, 5, 72, 98, 160, 5, 74, 98, 174, 5, 76, 98, 174, 5, 78, 98, 160, 5, 80, 98, 174, 5, 82, 98, 174, 5, 84, 98, 174, 5, 86, 98, 174, 5, 88, 98, 174, 5, 90, 98, 160, 5, 92, 98, 160, 5, 94, 98, 160, 5, 96, 98, 160, 5, 98, 98, 255, 255, 255, 255, 5, 5, 6, 98, 5, 5, 8, 98, 5, 5, 10, 98, 5, 5, 12, 98, 5, 5, 14, 98, 5, 5, 16, 98, 5, 5, 18, 98, 5, 5, 20, 98, 5, 5, 22, 98, 5, 5, 24, 98, 5, 5, 26, 98, 5, 5, 28, 98, 5, 5, 30, 98, 5, 5, 32, 98, 5, 5, 34, 98, 5, 5, 36, 98, 32, 5, 6, 98, 32, 5, 8, 98, 32, 5, 10, 98, 32, 5, 12, 98, 32, 5, 14, 98, 32, 5, 16, 98, 32, 5, 18, 98, 255, 255, 255, 255, 32, 5, 22, 98, 32, 5, 24, 98, 32, 5, 26, 98, 32, 5, 28, 98, 32, 5, 30, 98, 32, 5, 32, 98, 32, 5, 34, 98, 32, 5, 36, 98, 32, 5, 38, 98, 32, 5, 40, 98, 32, 5, 42, 98, 32, 5, 44, 98, 32, 5, 46, 98, 32, 5, 48, 98, 32, 5, 50, 98, 32, 5, 52, 98, 32, 5, 54, 98, 255, 255, 255, 255, 255, 255, 255, 255, 32, 5, 60, 98, 32, 5, 62, 98, 32, 5, 64, 98, 32, 5, 66, 98, 32, 5, 68, 98, 174, 5, 6, 98, 174, 5, 8, 98, 174, 5, 10, 98, 174, 5, 12, 98, 174, 5, 14, 98, 174, 5, 16, 98, 174, 5, 18, 98, 160, 5, 20, 98, 174, 5, 22, 98, 174, 5, 24, 98, 174, 5, 26, 98, 174, 5, 28, 98, 174, 5, 30, 98, 174, 5, 32, 98, 174, 5, 34, 98, 174, 5, 36, 98, 174, 5, 38, 98, 174, 5, 40, 98, 174, 5, 42, 98, 174, 5, 44, 98, 174, 5, 46, 98, 174, 5, 48, 98, 174, 5, 50, 98, 174, 5, 52, 98, 174, 5, 54, 98, 160, 5, 56, 98, 160, 5, 58, 98, 174, 5, 60, 98, 174, 5, 62, 98, 174, 5, 64, 98, 174, 5, 66, 98, 174, 5, 68, 98, 5, 5, 38, 98, 5, 5, 40, 98, 5, 5, 42, 98, 5, 5, 44, 98, 5, 5, 46, 98, 5, 5, 48, 98, 5, 5, 50, 98, 5, 5, 52, 98, 5, 5, 54, 98, 5, 5, 56, 98, 5, 5, 58, 98, 5, 5, 60, 98, 5, 5, 62, 98, 5, 5, 64, 98, 5, 5, 66, 98, 5, 5, 68, 98, 5, 5, 70, 98, 5, 5, 72, 98, 5, 5, 74, 98, 5, 5, 76, 98, 5, 5, 78, 98, 5, 5, 80, 98, 5, 5, 82, 98, 5, 5, 84, 98, 5, 5, 86, 98, 5, 5, 88, 98, 5, 5, 90, 98, 5, 5, 92, 98, 5, 5, 94, 98, 5, 5, 96, 98, 5, 5, 98, 98, 255, 255, 255, 255, 32, 5, 70, 98, 32, 5, 72, 98, 255, 255, 255, 255, 32, 5, 76, 98, 32, 5, 78, 98, 255, 255, 255, 255, 32, 5, 82, 98, 32, 5, 84, 98, 32, 5, 86, 98, 32, 5, 88, 98, 32, 5, 90, 98, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 206, 0, 235, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 102, 98, 193, 2, 104, 98, 193, 9, 104, 98, 5, 5, 106, 98, 193, 2, 108, 98, 193, 9, 108, 98, 193, 16, 108, 98, 5, 5, 110, 98, 193, 2, 112, 98, 193, 9, 112, 98, 193, 16, 112, 98, 193, 23, 112, 98, 193, 30, 112, 98, 5, 5, 114, 98, 193, 2, 116, 98, 193, 9, 116, 98, 193, 16, 116, 98, 5, 5, 118, 98, 193, 2, 120, 98, 193, 9, 120, 98, 193, 16, 120, 98, 193, 23, 120, 98, 193, 30, 120, 98, 193, 37, 120, 98, 193, 44, 120, 98, 193, 51, 120, 98, 193, 58, 120, 98, 193, 65, 120, 98, 193, 72, 120, 98, 193, 79, 120, 98, 193, 86, 120, 98, 193, 93, 120, 98, 193, 100, 120, 98, 193, 107, 120, 98, 193, 114, 120, 98, 193, 121, 120, 98, 193, 128, 120, 98, 193, 135, 120, 98, 32, 5, 102, 98, 32, 5, 106, 98, 32, 5, 110, 98, 32, 5, 114, 98, 32, 5, 118, 98, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 160, 5, 152, 99, 160, 5, 36, 99, 160, 5, 64, 99, 160, 5, 92, 99, 160, 5, 144, 99, 160, 5, 156, 99, 255, 255, 255, 255, 160, 5, 162, 99, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 160, 5, 172, 99, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 6, 99, 5, 5, 10, 99, 5, 5, 14, 99, 5, 5, 18, 99, 5, 5, 22, 99, 5, 5, 26, 99, 5, 5, 30, 99, 5, 5, 38, 99, 5, 5, 42, 99, 5, 5, 46, 99, 5, 5, 50, 99, 5, 5, 54, 99, 5, 5, 58, 99, 5, 5, 66, 99, 5, 5, 70, 99, 5, 5, 74, 99, 5, 5, 8, 99, 5, 5, 12, 99, 5, 5, 16, 99, 5, 5, 20, 99, 5, 5, 24, 99, 5, 5, 28, 99, 5, 5, 32, 99, 5, 5, 40, 99, 5, 5, 44, 99, 5, 5, 48, 99, 5, 5, 52, 99, 5, 5, 56, 99, 5, 5, 60, 99, 5, 5, 68, 99, 5, 5, 72, 99, 5, 5, 76, 99, 5, 5, 80, 99, 5, 5, 84, 99, 5, 5, 88, 99, 5, 5, 96, 99, 5, 5, 100, 99, 5, 5, 104, 99, 5, 5, 108, 99, 5, 5, 112, 99, 5, 5, 116, 99, 5, 5, 120, 99, 5, 5, 124, 99, 5, 5, 128, 99, 5, 5, 132, 99, 5, 5, 136, 99, 5, 5, 140, 99, 5, 5, 148, 99, 160, 5, 8, 99, 160, 5, 12, 99, 160, 5, 16, 99, 160, 5, 20, 99, 160, 5, 24, 99, 160, 5, 28, 99, 160, 5, 32, 99, 160, 5, 40, 99, 160, 5, 44, 99, 160, 5, 48, 99, 160, 5, 52, 99, 160, 5, 56, 99, 160, 5, 60, 99, 160, 5, 68, 99, 160, 5, 72, 99, 160, 5, 76, 99, 160, 5, 80, 99, 160, 5, 84, 99, 160, 5, 88, 99, 160, 5, 96, 99, 160, 5, 100, 99, 160, 5, 104, 99, 160, 5, 108, 99, 160, 5, 112, 99, 160, 5, 116, 99, 160, 5, 120, 99, 160, 5, 124, 99, 160, 5, 128, 99, 160, 5, 132, 99, 160, 5, 136, 99, 160, 5, 140, 99, 160, 5, 148, 99, 160, 5, 78, 99, 160, 5, 82, 99, 160, 5, 86, 99, 160, 5, 94, 99, 160, 5, 98, 99, 160, 5, 102, 99, 160, 5, 106, 99, 160, 5, 110, 99, 160, 5, 114, 99, 160, 5, 118, 99, 160, 5, 122, 99, 160, 5, 126, 99, 160, 5, 130, 99, 160, 5, 134, 99, 160, 5, 138, 99, 160, 5, 146, 99, 160, 5, 150, 99, 160, 5, 34, 99, 160, 5, 62, 99, 160, 5, 90, 99, 160, 5, 142, 99, 160, 5, 154, 99, 160, 5, 158, 99, 160, 5, 160, 99, 160, 5, 164, 99, 160, 5, 166, 99, 160, 5, 168, 99, 255, 255, 255, 255, 255, 255, 255, 255, 160, 5, 170, 99, 160, 5, 174, 99, 160, 5, 176, 99, 5, 5, 152, 99, 5, 5, 36, 99, 5, 5, 64, 99, 5, 5, 92, 99, 5, 5, 144, 99, 5, 5, 156, 99, 255, 255, 255, 255, 5, 5, 162, 99, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 172, 99, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 141, 103, 193, 16, 141, 103, 193, 23, 141, 103, 193, 30, 141, 103, 193, 37, 141, 103, 193, 44, 141, 103, 193, 51, 141, 103, 193, 58, 141, 103, 193, 65, 141, 103, 193, 72, 141, 103, 193, 79, 141, 103, 193, 86, 141, 103, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 197, 34, 58, 1, 197, 98, 58, 1, 255, 255, 255, 255, 197, 226, 58, 1, 197, 162, 58, 1, 255, 255, 255, 255, 197, 98, 59, 1, 197, 162, 59, 1, 197, 226, 59, 1, 197, 162, 60, 1, 197, 162, 61, 1, 197, 194, 53, 1, 197, 194, 52, 1, 197, 98, 57, 1, 197, 34, 59, 1, 197, 66, 52, 1, 46, 5, 14, 102, 32, 5, 14, 102, 49, 5, 32, 102, 38, 5, 32, 102, 16, 5, 32, 102, 27, 5, 32, 102, 49, 5, 34, 102, 38, 5, 34, 102, 16, 5, 34, 102, 27, 5, 34, 102, 49, 5, 36, 102, 38, 5, 36, 102, 16, 5, 36, 102, 27, 5, 36, 102, 49, 5, 48, 102, 38, 5, 48, 102, 46, 5, 6, 102, 46, 5, 8, 102, 32, 5, 8, 102, 46, 5, 10, 102, 32, 5, 10, 102, 46, 5, 16, 102, 32, 5, 16, 102, 46, 5, 18, 102, 32, 5, 18, 102, 49, 5, 22, 102, 38, 5, 22, 102, 16, 5, 22, 102, 27, 5, 22, 102, 49, 5, 26, 102, 38, 5, 26, 102, 48, 5, 30, 102, 40, 5, 30, 102, 24, 5, 30, 102, 32, 5, 30, 102, 46, 5, 40, 102, 32, 5, 40, 102, 48, 5, 42, 102, 40, 5, 42, 102, 24, 5, 42, 102, 32, 5, 42, 102, 48, 5, 44, 102, 40, 5, 44, 102, 24, 5, 44, 102, 32, 5, 44, 102, 48, 5, 58, 102, 40, 5, 58, 102, 24, 5, 58, 102, 198, 97, 165, 0, 198, 65, 165, 0, 46, 5, 232, 102, 32, 5, 232, 102, 49, 5, 246, 102, 38, 5, 246, 102, 16, 5, 246, 102, 27, 5, 246, 102, 16, 5, 236, 102, 27, 5, 236, 102, 197, 34, 62, 1, 197, 226, 61, 1, 198, 194, 160, 0, 198, 130, 160, 0, 197, 34, 66, 1, 197, 226, 65, 1, 197, 162, 66, 1, 197, 98, 66, 1, 198, 66, 161, 0, 198, 2, 161, 0, 197, 34, 67, 1, 197, 226, 66, 1, 197, 34, 69, 1, 197, 226, 68, 1, 197, 162, 68, 1, 197, 226, 67, 1, 197, 162, 67, 1, 197, 98, 67, 1, 49, 5, 240, 102, 38, 5, 240, 102, 16, 5, 240, 102, 27, 5, 240, 102, 197, 130, 149, 1, 197, 2, 152, 1, 197, 2, 153, 1, 197, 162, 154, 1, 197, 34, 155, 1, 197, 130, 163, 1, 197, 34, 165, 1, 197, 98, 165, 1, 197, 162, 165, 1, 197, 2, 106, 1, 197, 162, 176, 1, 197, 226, 176, 1, 197, 162, 177, 1, 197, 226, 178, 1, 197, 162, 179, 1, 197, 34, 180, 1, 197, 34, 184, 1, 197, 226, 186, 1, 197, 34, 187, 1, 197, 130, 165, 1, 197, 194, 178, 1, 197, 194, 189, 1, 197, 66, 190, 1, 197, 98, 62, 1, 197, 226, 62, 1, 197, 98, 63, 1, 197, 34, 64, 1, 197, 98, 65, 1, 197, 34, 71, 1, 197, 162, 71, 1, 197, 130, 72, 1, 197, 226, 73, 1, 197, 162, 62, 1, 197, 34, 63, 1, 197, 226, 64, 1, 197, 226, 67, 1, 197, 98, 68, 1, 197, 98, 71, 1, 197, 226, 71, 1, 197, 194, 72, 1, 197, 162, 74, 1, 197, 226, 75, 1, 197, 98, 76, 1, 197, 226, 76, 1, 197, 130, 78, 1, 197, 34, 80, 1, 197, 194, 82, 1, 197, 226, 85, 1, 197, 98, 86, 1, 197, 162, 86, 1, 197, 34, 88, 1, 197, 34, 89, 1, 197, 162, 89, 1, 197, 34, 90, 1, 197, 34, 92, 1, 197, 98, 94, 1, 197, 2, 95, 1, 197, 194, 96, 1, 197, 2, 97, 1, 197, 66, 97, 1, 197, 130, 99, 1, 197, 2, 101, 1, 197, 34, 102, 1, 197, 34, 104, 1, 197, 130, 142, 1, 197, 194, 143, 1, 197, 162, 145, 1, 197, 130, 146, 1, 197, 2, 150, 1, 197, 130, 150, 1, 197, 2, 151, 1, 197, 130, 151, 1, 197, 130, 152, 1, 197, 98, 157, 1, 197, 194, 159, 1, 197, 194, 161, 1, 197, 2, 163, 1, 197, 226, 164, 1, 197, 34, 83, 1, 197, 130, 83, 1, 197, 226, 83, 1, 197, 162, 105, 1, 197, 194, 171, 1, 197, 130, 174, 1, 197, 34, 176, 1, 197, 34, 177, 1, 197, 34, 179, 1, 197, 226, 180, 1, 197, 34, 181, 1, 197, 162, 180, 1, 197, 226, 184, 1, 197, 194, 185, 1, 197, 98, 186, 1, 197, 98, 187, 1, 197, 66, 189, 1, 197, 98, 64, 1, 193, 44, 38, 102, 193, 65, 38, 102, 193, 2, 70, 102, 5, 5, 164, 102, 193, 37, 176, 102, 193, 30, 180, 102, 193, 37, 198, 102, 193, 16, 202, 102, 193, 2, 24, 102, 193, 9, 24, 102, 193, 37, 132, 102, 193, 44, 234, 102, 193, 16, 24, 102, 16, 5, 26, 102, 5, 5, 100, 102, 5, 5, 152, 102, 193, 2, 194, 102, 193, 16, 234, 102, 193, 44, 132, 102, 193, 37, 172, 102, 193, 30, 190, 102, 255, 255, 255, 255, 193, 72, 38, 102, 193, 79, 38, 102, 193, 2, 56, 102, 193, 51, 132, 102, 193, 58, 248, 102, 193, 16, 176, 102, 193, 9, 180, 102, 193, 2, 208, 102, 193, 58, 38, 102, 193, 9, 56, 102, 27, 5, 26, 102, 16, 5, 30, 102, 16, 5, 58, 102, 16, 5, 84, 102, 255, 255, 255, 255, 27, 5, 222, 102, 16, 5, 112, 102, 16, 5, 72, 102, 16, 5, 158, 102, 23, 5, 238, 102, 16, 5, 182, 102, 16, 5, 196, 102, 16, 5, 200, 102, 16, 5, 204, 102, 16, 5, 134, 102, 16, 5, 166, 102, 16, 5, 174, 102, 16, 5, 146, 102, 16, 5, 178, 102, 16, 5, 110, 102, 16, 5, 136, 102, 16, 5, 42, 102, 16, 5, 44, 102, 16, 5, 74, 102, 16, 5, 86, 102, 16, 5, 148, 102, 16, 5, 160, 102, 16, 5, 168, 102, 32, 5, 28, 102, 16, 5, 206, 102, 198, 129, 161, 0, 198, 161, 162, 0, 27, 5, 26, 102, 16, 5, 30, 102, 16, 5, 58, 102, 16, 5, 84, 102, 16, 5, 214, 102, 27, 5, 222, 102, 16, 5, 112, 102, 16, 5, 72, 102, 16, 5, 158, 102, 23, 5, 238, 102, 255, 255, 255, 255, 16, 5, 196, 102, 16, 5, 200, 102, 16, 5, 204, 102, 16, 5, 134, 102, 16, 5, 166, 102, 16, 5, 174, 102, 16, 5, 146, 102, 16, 5, 178, 102, 16, 5, 110, 102, 16, 5, 136, 102, 16, 5, 42, 102, 16, 5, 44, 102, 16, 5, 74, 102, 16, 5, 86, 102, 16, 5, 148, 102, 16, 5, 160, 102, 16, 5, 168, 102, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 30, 102, 16, 5, 58, 102, 255, 255, 255, 255, 16, 5, 214, 102, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 72, 102, 16, 5, 158, 102, 23, 5, 238, 102, 16, 5, 182, 102, 255, 255, 255, 255, 16, 5, 200, 102, 16, 5, 204, 102, 16, 5, 134, 102, 16, 5, 166, 102, 16, 5, 174, 102, 16, 5, 146, 102, 16, 5, 178, 102, 255, 255, 255, 255, 16, 5, 136, 102, 16, 5, 42, 102, 16, 5, 44, 102, 16, 5, 74, 102, 255, 255, 255, 255, 16, 5, 148, 102, 16, 5, 160, 102, 16, 5, 168, 102, 32, 5, 28, 102, 255, 255, 255, 255, 198, 129, 161, 0, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 30, 102, 16, 5, 58, 102, 255, 255, 255, 255, 16, 5, 214, 102, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 72, 102, 255, 255, 255, 255, 23, 5, 238, 102, 16, 5, 182, 102, 16, 5, 196, 102, 16, 5, 200, 102, 16, 5, 204, 102, 16, 5, 134, 102, 16, 5, 166, 102, 16, 5, 174, 102, 16, 5, 146, 102, 16, 5, 178, 102, 255, 255, 255, 255, 16, 5, 136, 102, 16, 5, 42, 102, 16, 5, 44, 102, 16, 5, 74, 102, 255, 255, 255, 255, 16, 5, 148, 102, 255, 255, 255, 255, 16, 5, 168, 102, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 30, 102, 16, 5, 58, 102, 16, 5, 84, 102, 255, 255, 255, 255, 27, 5, 222, 102, 16, 5, 112, 102, 16, 5, 72, 102, 16, 5, 158, 102, 23, 5, 238, 102, 255, 255, 255, 255, 16, 5, 196, 102, 16, 5, 200, 102, 16, 5, 204, 102, 16, 5, 134, 102, 16, 5, 166, 102, 16, 5, 174, 102, 16, 5, 146, 102, 16, 5, 178, 102, 16, 5, 110, 102, 16, 5, 136, 102, 16, 5, 42, 102, 16, 5, 44, 102, 16, 5, 74, 102, 16, 5, 86, 102, 16, 5, 148, 102, 16, 5, 160, 102, 16, 5, 168, 102, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 163, 160, 1, 197, 99, 161, 1, 197, 3, 161, 1, 197, 227, 157, 1, 197, 67, 158, 1, 197, 163, 162, 1, 197, 67, 162, 1, 197, 35, 164, 1, 197, 195, 163, 1, 197, 227, 167, 1, 197, 67, 168, 1, 197, 35, 169, 1, 197, 35, 166, 1, 197, 227, 166, 1, 197, 131, 169, 1, 197, 227, 169, 1, 255, 255, 255, 255, 255, 255, 255, 255, 197, 131, 166, 1, 197, 163, 181, 1, 197, 3, 182, 1, 197, 3, 175, 1, 197, 99, 175, 1, 197, 99, 173, 1, 197, 3, 173, 1, 197, 195, 173, 1, 197, 131, 178, 1, 197, 35, 178, 1, 197, 131, 188, 1, 197, 35, 188, 1, 197, 3, 73, 1, 197, 227, 77, 1, 197, 34, 75, 1, 197, 162, 76, 1, 197, 66, 78, 1, 197, 226, 79, 1, 197, 2, 82, 1, 197, 34, 85, 1, 197, 98, 87, 1, 197, 226, 89, 1, 197, 66, 77, 1, 197, 226, 78, 1, 197, 162, 79, 1, 197, 130, 96, 1, 197, 130, 80, 1, 197, 2, 99, 1, 197, 130, 100, 1, 197, 162, 101, 1, 197, 162, 103, 1, 197, 194, 116, 1, 197, 98, 118, 1, 197, 162, 122, 1, 197, 226, 124, 1, 197, 98, 125, 1, 197, 162, 126, 1, 197, 98, 128, 1, 197, 226, 129, 1, 197, 162, 133, 1, 197, 98, 134, 1, 197, 162, 135, 1, 197, 98, 138, 1, 197, 226, 138, 1, 197, 130, 141, 1, 197, 2, 142, 1, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 35, 77, 1, 197, 35, 79, 1, 197, 195, 78, 1, 197, 131, 79, 1, 197, 99, 80, 1, 197, 3, 83, 1, 197, 99, 83, 1, 197, 195, 83, 1, 197, 195, 92, 1, 197, 99, 92, 1, 197, 163, 95, 1, 197, 67, 95, 1, 197, 67, 101, 1, 197, 195, 99, 1, 197, 35, 100, 1, 197, 35, 105, 1, 197, 131, 77, 1, 197, 35, 81, 1, 197, 195, 80, 1, 197, 131, 84, 1, 197, 35, 84, 1, 197, 131, 93, 1, 197, 99, 90, 1, 197, 35, 93, 1, 197, 99, 102, 1, 197, 3, 118, 1, 197, 227, 110, 1, 197, 67, 126, 1, 197, 99, 159, 1, 197, 131, 164, 1, 197, 3, 186, 1, 197, 99, 185, 1, 197, 227, 188, 1, 197, 227, 170, 1, 197, 35, 148, 1, 197, 195, 175, 1, 197, 3, 147, 1, 197, 67, 160, 1, 197, 3, 137, 1, 197, 67, 154, 1, 197, 67, 172, 1, 197, 67, 170, 1, 197, 163, 158, 1, 197, 227, 153, 1, 197, 3, 159, 1, 197, 163, 172, 1, 197, 195, 90, 1, 197, 163, 94, 1, 16, 5, 48, 102, 27, 5, 48, 102, 49, 5, 54, 102, 38, 5, 54, 102, 16, 5, 54, 102, 27, 5, 54, 102, 49, 5, 46, 102, 38, 5, 46, 102, 16, 5, 46, 102, 27, 5, 46, 102, 198, 1, 162, 0, 198, 225, 161, 0, 198, 161, 161, 0, 198, 193, 161, 0, 198, 129, 162, 0, 198, 97, 162, 0, 198, 33, 162, 0, 198, 65, 162, 0, 49, 5, 62, 102, 38, 5, 62, 102, 16, 5, 62, 102, 27, 5, 62, 102, 49, 5, 60, 102, 38, 5, 60, 102, 16, 5, 60, 102, 27, 5, 60, 102, 49, 5, 64, 102, 38, 5, 64, 102, 16, 5, 64, 102, 27, 5, 64, 102, 49, 5, 68, 102, 38, 5, 68, 102, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 58, 102, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 72, 102, 255, 255, 255, 255, 23, 5, 238, 102, 255, 255, 255, 255, 16, 5, 196, 102, 255, 255, 255, 255, 16, 5, 204, 102, 16, 5, 134, 102, 16, 5, 166, 102, 255, 255, 255, 255, 16, 5, 146, 102, 16, 5, 178, 102, 255, 255, 255, 255, 16, 5, 136, 102, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 74, 102, 255, 255, 255, 255, 16, 5, 148, 102, 255, 255, 255, 255, 16, 5, 168, 102, 255, 255, 255, 255, 16, 5, 206, 102, 255, 255, 255, 255, 198, 161, 162, 0, 32, 5, 58, 102, 48, 5, 72, 102, 40, 5, 72, 102, 24, 5, 72, 102, 32, 5, 72, 102, 48, 5, 74, 102, 40, 5, 74, 102, 24, 5, 74, 102, 32, 5, 74, 102, 44, 5, 84, 102, 30, 5, 84, 102, 44, 5, 86, 102, 30, 5, 86, 102, 44, 5, 110, 102, 30, 5, 110, 102, 44, 5, 112, 102, 30, 5, 112, 102, 48, 5, 134, 102, 40, 5, 134, 102, 24, 5, 134, 102, 32, 5, 134, 102, 48, 5, 136, 102, 40, 5, 136, 102, 24, 5, 136, 102, 32, 5, 136, 102, 48, 5, 146, 102, 40, 5, 146, 102, 24, 5, 146, 102, 32, 5, 146, 102, 48, 5, 148, 102, 40, 5, 148, 102, 24, 5, 148, 102, 197, 66, 96, 1, 197, 226, 93, 1, 197, 34, 94, 1, 197, 130, 97, 1, 197, 194, 97, 1, 197, 34, 124, 1, 197, 162, 124, 1, 197, 34, 129, 1, 197, 162, 129, 1, 197, 130, 108, 1, 197, 226, 109, 1, 197, 2, 112, 1, 197, 130, 113, 1, 197, 130, 112, 1, 197, 98, 103, 1, 197, 226, 118, 1, 197, 34, 128, 1, 197, 162, 132, 1, 197, 34, 133, 1, 197, 98, 137, 1, 197, 226, 137, 1, 197, 130, 140, 1, 197, 2, 141, 1, 197, 194, 106, 1, 197, 66, 107, 1, 197, 194, 115, 1, 197, 66, 116, 1, 197, 130, 90, 1, 197, 66, 72, 1, 197, 162, 77, 1, 197, 2, 78, 1, 197, 226, 80, 1, 16, 5, 68, 102, 27, 5, 68, 102, 46, 5, 98, 102, 32, 5, 98, 102, 46, 5, 96, 102, 32, 5, 96, 102, 46, 5, 102, 102, 32, 5, 102, 102, 46, 5, 88, 102, 32, 5, 88, 102, 46, 5, 128, 102, 32, 5, 128, 102, 46, 5, 114, 102, 32, 5, 114, 102, 49, 5, 184, 102, 38, 5, 184, 102, 16, 5, 184, 102, 27, 5, 184, 102, 49, 5, 192, 102, 38, 5, 192, 102, 16, 5, 192, 102, 27, 5, 192, 102, 198, 33, 164, 0, 198, 1, 164, 0, 198, 193, 163, 0, 198, 225, 163, 0, 198, 161, 163, 0, 198, 129, 163, 0, 198, 65, 163, 0, 198, 97, 163, 0, 44, 5, 206, 102, 30, 5, 206, 102, 197, 195, 104, 1, 197, 99, 104, 1, 197, 227, 105, 1, 197, 131, 105, 1, 197, 163, 117, 1, 197, 67, 117, 1, 197, 131, 123, 1, 197, 131, 110, 1, 197, 35, 110, 1, 197, 195, 108, 1, 197, 35, 114, 1, 197, 195, 113, 1, 197, 227, 114, 1, 197, 131, 114, 1, 197, 227, 125, 1, 197, 131, 127, 1, 197, 35, 127, 1, 197, 131, 131, 1, 197, 35, 131, 1, 197, 227, 131, 1, 197, 67, 132, 1, 197, 67, 135, 1, 197, 67, 136, 1, 197, 227, 135, 1, 197, 163, 136, 1, 197, 99, 139, 1, 197, 35, 140, 1, 197, 195, 139, 1, 197, 99, 143, 1, 197, 3, 143, 1, 197, 99, 147, 1, 197, 195, 147, 1, 197, 2, 117, 1, 197, 226, 122, 1, 197, 34, 125, 1, 197, 162, 125, 1, 197, 226, 126, 1, 197, 162, 128, 1, 197, 34, 130, 1, 197, 226, 130, 1, 197, 34, 134, 1, 197, 162, 134, 1, 197, 98, 119, 1, 197, 162, 138, 1, 197, 34, 139, 1, 197, 194, 141, 1, 197, 66, 142, 1, 197, 194, 142, 1, 197, 2, 144, 1, 197, 226, 144, 1, 197, 98, 145, 1, 197, 226, 145, 1, 197, 194, 146, 1, 197, 194, 148, 1, 197, 66, 149, 1, 197, 194, 149, 1, 197, 66, 150, 1, 197, 194, 150, 1, 197, 66, 151, 1, 197, 66, 152, 1, 197, 66, 153, 1, 197, 226, 154, 1, 197, 98, 155, 1, 197, 162, 157, 1, 32, 5, 148, 102, 48, 5, 158, 102, 40, 5, 158, 102, 24, 5, 158, 102, 32, 5, 158, 102, 48, 5, 160, 102, 40, 5, 160, 102, 24, 5, 160, 102, 32, 5, 160, 102, 48, 5, 166, 102, 40, 5, 166, 102, 24, 5, 166, 102, 32, 5, 166, 102, 48, 5, 168, 102, 40, 5, 168, 102, 24, 5, 168, 102, 32, 5, 168, 102, 48, 5, 174, 102, 40, 5, 174, 102, 24, 5, 174, 102, 32, 5, 174, 102, 48, 5, 178, 102, 40, 5, 178, 102, 24, 5, 178, 102, 32, 5, 178, 102, 48, 5, 182, 102, 40, 5, 182, 102, 24, 5, 182, 102, 32, 5, 182, 102, 48, 5, 196, 102, 40, 5, 196, 102, 24, 5, 196, 102, 193, 44, 26, 103, 193, 51, 26, 103, 193, 58, 26, 103, 193, 65, 26, 103, 193, 72, 26, 103, 193, 79, 26, 103, 193, 86, 26, 103, 193, 93, 26, 103, 193, 100, 26, 103, 193, 107, 26, 103, 193, 114, 26, 103, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 160, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 206, 192, 231, 0, 193, 9, 143, 103, 193, 16, 143, 103, 193, 23, 143, 103, 193, 30, 143, 103, 193, 37, 143, 103, 193, 44, 143, 103, 193, 51, 143, 103, 193, 58, 143, 103, 193, 65, 143, 103, 193, 72, 143, 103, 193, 79, 143, 103, 193, 86, 143, 103, 193, 93, 143, 103, 193, 100, 143, 103, 193, 107, 143, 103, 193, 114, 143, 103, 193, 128, 143, 103, 193, 135, 143, 103, 193, 142, 143, 103, 193, 149, 143, 103, 193, 156, 143, 103, 193, 163, 143, 103, 193, 170, 143, 103, 193, 177, 143, 103, 193, 184, 143, 103, 193, 191, 143, 103, 193, 198, 143, 103, 193, 205, 143, 103, 193, 212, 143, 103, 193, 219, 143, 103, 193, 226, 143, 103, 193, 233, 143, 103, 193, 121, 143, 103, 193, 14, 144, 103, 193, 133, 144, 103, 193, 196, 144, 103, 193, 47, 145, 103, 193, 59, 146, 103, 193, 178, 146, 103, 193, 241, 146, 103, 193, 141, 147, 103, 193, 204, 147, 103, 193, 13, 148, 103, 193, 102, 149, 103, 193, 65, 150, 103, 193, 170, 150, 103, 193, 233, 150, 103, 193, 217, 151, 103, 193, 26, 152, 103, 193, 138, 152, 103, 193, 220, 153, 103, 193, 133, 151, 103, 193, 140, 151, 103, 193, 147, 151, 103, 193, 154, 151, 103, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 21, 144, 103, 193, 28, 144, 103, 193, 35, 144, 103, 193, 42, 144, 103, 193, 49, 144, 103, 193, 56, 144, 103, 193, 63, 144, 103, 193, 70, 144, 103, 193, 77, 144, 103, 193, 84, 144, 103, 193, 91, 144, 103, 193, 98, 144, 103, 193, 105, 144, 103, 193, 112, 144, 103, 193, 119, 144, 103, 193, 126, 144, 103, 193, 140, 144, 103, 193, 147, 144, 103, 193, 154, 144, 103, 193, 161, 144, 103, 193, 168, 144, 103, 193, 175, 144, 103, 193, 182, 144, 103, 193, 189, 144, 103, 193, 245, 144, 103, 193, 252, 144, 103, 193, 5, 145, 103, 193, 12, 145, 103, 193, 19, 145, 103, 193, 26, 145, 103, 193, 33, 145, 103, 193, 40, 145, 103, 255, 255, 255, 255, 193, 203, 144, 103, 193, 210, 144, 103, 193, 217, 144, 103, 193, 224, 144, 103, 193, 231, 144, 103, 193, 238, 144, 103, 255, 255, 255, 255, 255, 255, 255, 255, 193, 72, 150, 103, 193, 79, 150, 103, 193, 86, 150, 103, 193, 93, 150, 103, 193, 100, 150, 103, 193, 107, 150, 103, 255, 255, 255, 255, 255, 255, 255, 255, 193, 109, 149, 103, 193, 116, 149, 103, 193, 123, 149, 103, 193, 130, 149, 103, 193, 137, 149, 103, 193, 144, 149, 103, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 54, 145, 103, 193, 61, 145, 103, 193, 68, 145, 103, 193, 75, 145, 103, 193, 82, 145, 103, 193, 89, 145, 103, 193, 96, 145, 103, 193, 103, 145, 103, 193, 110, 145, 103, 255, 255, 255, 255, 193, 117, 145, 103, 193, 124, 145, 103, 193, 131, 145, 103, 193, 138, 145, 103, 255, 255, 255, 255, 255, 255, 255, 255, 193, 145, 145, 103, 193, 152, 145, 103, 193, 159, 145, 103, 193, 166, 145, 103, 193, 173, 145, 103, 193, 180, 145, 103, 193, 187, 145, 103, 255, 255, 255, 255, 193, 194, 145, 103, 255, 255, 255, 255, 193, 201, 145, 103, 193, 208, 145, 103, 193, 215, 145, 103, 193, 222, 145, 103, 255, 255, 255, 255, 255, 255, 255, 255, 193, 229, 145, 103, 193, 236, 145, 103, 193, 243, 145, 103, 193, 250, 145, 103, 193, 3, 146, 103, 193, 10, 146, 103, 193, 17, 146, 103, 193, 24, 146, 103, 193, 66, 146, 103, 193, 73, 146, 103, 193, 80, 146, 103, 193, 87, 146, 103, 193, 94, 146, 103, 193, 101, 146, 103, 193, 108, 146, 103, 193, 115, 146, 103, 193, 122, 146, 103, 193, 129, 146, 103, 193, 136, 146, 103, 193, 143, 146, 103, 193, 150, 146, 103, 193, 157, 146, 103, 193, 164, 146, 103, 193, 171, 146, 103, 193, 185, 146, 103, 193, 192, 146, 103, 193, 199, 146, 103, 193, 206, 146, 103, 193, 213, 146, 103, 193, 220, 146, 103, 193, 227, 146, 103, 193, 234, 146, 103, 193, 248, 146, 103, 193, 255, 146, 103, 193, 8, 147, 103, 193, 15, 147, 103, 193, 22, 147, 103, 193, 29, 147, 103, 193, 36, 147, 103, 193, 43, 147, 103, 193, 50, 147, 103, 255, 255, 255, 255, 193, 57, 147, 103, 193, 64, 147, 103, 193, 71, 147, 103, 193, 78, 147, 103, 255, 255, 255, 255, 255, 255, 255, 255, 193, 85, 147, 103, 193, 92, 147, 103, 193, 99, 147, 103, 193, 106, 147, 103, 193, 113, 147, 103, 193, 120, 147, 103, 193, 127, 147, 103, 193, 134, 147, 103, 193, 148, 147, 103, 193, 155, 147, 103, 193, 162, 147, 103, 193, 169, 147, 103, 193, 176, 147, 103, 193, 183, 147, 103, 193, 190, 147, 103, 193, 197, 147, 103, 193, 211, 147, 103, 193, 218, 147, 103, 193, 225, 147, 103, 193, 232, 147, 103, 193, 239, 147, 103, 193, 246, 147, 103, 193, 253, 147, 103, 193, 6, 148, 103, 193, 20, 148, 103, 193, 27, 148, 103, 193, 34, 148, 103, 193, 41, 148, 103, 193, 48, 148, 103, 193, 55, 148, 103, 193, 62, 148, 103, 193, 69, 148, 103, 193, 76, 148, 103, 255, 255, 255, 255, 193, 83, 148, 103, 193, 90, 148, 103, 193, 97, 148, 103, 193, 104, 148, 103, 255, 255, 255, 255, 255, 255, 255, 255, 193, 111, 148, 103, 193, 118, 148, 103, 193, 125, 148, 103, 193, 132, 148, 103, 193, 139, 148, 103, 193, 146, 148, 103, 193, 153, 148, 103, 255, 255, 255, 255, 193, 160, 148, 103, 255, 255, 255, 255, 193, 167, 148, 103, 193, 174, 148, 103, 193, 181, 148, 103, 193, 188, 148, 103, 255, 255, 255, 255, 255, 255, 255, 255, 193, 195, 148, 103, 193, 202, 148, 103, 193, 209, 148, 103, 193, 216, 148, 103, 193, 223, 148, 103, 193, 230, 148, 103, 193, 237, 148, 103, 193, 244, 148, 103, 193, 251, 148, 103, 193, 4, 149, 103, 193, 11, 149, 103, 193, 18, 149, 103, 193, 25, 149, 103, 193, 32, 149, 103, 193, 39, 149, 103, 255, 255, 255, 255, 193, 46, 149, 103, 193, 53, 149, 103, 193, 60, 149, 103, 193, 67, 149, 103, 193, 74, 149, 103, 193, 81, 149, 103, 193, 88, 149, 103, 193, 95, 149, 103, 193, 151, 149, 103, 193, 158, 149, 103, 193, 165, 149, 103, 193, 172, 149, 103, 193, 179, 149, 103, 193, 186, 149, 103, 193, 193, 149, 103, 193, 200, 149, 103, 193, 207, 149, 103, 193, 214, 149, 103, 193, 221, 149, 103, 193, 228, 149, 103, 193, 235, 149, 103, 193, 242, 149, 103, 193, 249, 149, 103, 193, 2, 150, 103, 193, 9, 150, 103, 193, 16, 150, 103, 193, 23, 150, 103, 193, 30, 150, 103, 193, 37, 150, 103, 193, 44, 150, 103, 193, 51, 150, 103, 193, 58, 150, 103, 193, 114, 150, 103, 193, 121, 150, 103, 193, 128, 150, 103, 193, 135, 150, 103, 193, 142, 150, 103, 193, 149, 150, 103, 193, 156, 150, 103, 193, 163, 150, 103, 193, 177, 150, 103, 193, 184, 150, 103, 193, 191, 150, 103, 193, 198, 150, 103, 193, 205, 150, 103, 193, 212, 150, 103, 193, 219, 150, 103, 193, 226, 150, 103, 193, 240, 150, 103, 193, 247, 150, 103, 193, 254, 150, 103, 193, 7, 151, 103, 193, 14, 151, 103, 193, 21, 151, 103, 193, 28, 151, 103, 193, 35, 151, 103, 193, 42, 151, 103, 255, 255, 255, 255, 193, 49, 151, 103, 193, 56, 151, 103, 193, 63, 151, 103, 193, 70, 151, 103, 255, 255, 255, 255, 255, 255, 255, 255, 193, 77, 151, 103, 193, 84, 151, 103, 193, 91, 151, 103, 193, 98, 151, 103, 193, 105, 151, 103, 193, 112, 151, 103, 193, 119, 151, 103, 193, 126, 151, 103, 193, 161, 151, 103, 193, 168, 151, 103, 193, 175, 151, 103, 193, 182, 151, 103, 193, 189, 151, 103, 193, 196, 151, 103, 193, 203, 151, 103, 193, 210, 151, 103, 193, 224, 151, 103, 193, 231, 151, 103, 193, 238, 151, 103, 193, 245, 151, 103, 193, 252, 151, 103, 193, 5, 152, 103, 193, 12, 152, 103, 193, 19, 152, 103, 193, 82, 152, 103, 193, 89, 152, 103, 193, 96, 152, 103, 193, 103, 152, 103, 193, 110, 152, 103, 193, 117, 152, 103, 193, 124, 152, 103, 193, 131, 152, 103, 193, 145, 152, 103, 193, 152, 152, 103, 193, 159, 152, 103, 193, 166, 152, 103, 193, 173, 152, 103, 193, 180, 152, 103, 193, 187, 152, 103, 193, 194, 152, 103, 193, 248, 153, 103, 193, 255, 153, 103, 193, 8, 154, 103, 193, 15, 154, 103, 193, 22, 154, 103, 193, 29, 154, 103, 193, 36, 154, 103, 255, 255, 255, 255, 193, 43, 154, 103, 193, 50, 154, 103, 193, 57, 154, 103, 193, 64, 154, 103, 193, 71, 154, 103, 193, 78, 154, 103, 193, 85, 154, 103, 255, 255, 255, 255, 193, 92, 154, 103, 193, 99, 154, 103, 193, 106, 154, 103, 193, 113, 154, 103, 193, 120, 154, 103, 193, 127, 154, 103, 193, 134, 154, 103, 255, 255, 255, 255, 193, 141, 154, 103, 193, 148, 154, 103, 193, 155, 154, 103, 193, 162, 154, 103, 193, 169, 154, 103, 193, 176, 154, 103, 193, 183, 154, 103, 255, 255, 255, 255, 193, 190, 154, 103, 193, 197, 154, 103, 193, 204, 154, 103, 193, 211, 154, 103, 193, 218, 154, 103, 193, 225, 154, 103, 193, 232, 154, 103, 255, 255, 255, 255, 193, 239, 154, 103, 193, 246, 154, 103, 193, 253, 154, 103, 193, 6, 155, 103, 193, 13, 155, 103, 193, 20, 155, 103, 193, 27, 155, 103, 255, 255, 255, 255, 193, 34, 155, 103, 193, 41, 155, 103, 193, 48, 155, 103, 193, 55, 155, 103, 193, 62, 155, 103, 193, 69, 155, 103, 193, 76, 155, 103, 255, 255, 255, 255, 193, 83, 155, 103, 193, 90, 155, 103, 193, 97, 155, 103, 193, 104, 155, 103, 193, 111, 155, 103, 193, 118, 155, 103, 193, 125, 155, 103, 255, 255, 255, 255, 5, 5, 110, 105, 5, 5, 112, 105, 5, 5, 114, 105, 5, 5, 116, 105, 5, 5, 118, 105, 255, 255, 255, 255, 255, 255, 255, 255, 201, 192, 6, 0, 5, 5, 126, 105, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 128, 105, 5, 5, 130, 105, 5, 5, 132, 105, 197, 194, 194, 1, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 134, 105, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 66, 194, 1, 197, 130, 194, 1, 255, 255, 255, 255, 197, 2, 195, 1, 5, 5, 104, 106, 5, 5, 106, 106, 5, 5, 108, 106, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 110, 106, 5, 5, 112, 106, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 114, 106, 5, 5, 116, 106, 5, 5, 118, 106, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 36, 106, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 197, 130, 195, 1, 197, 194, 195, 1, 197, 2, 196, 1, 5, 5, 98, 106, 255, 255, 255, 255, 197, 66, 196, 1, 255, 255, 255, 255, 5, 5, 116, 107, 5, 5, 118, 107, 5, 5, 120, 107, 5, 5, 122, 107, 5, 5, 124, 107, 5, 5, 130, 107, 255, 255, 255, 255, 5, 5, 132, 107, 5, 5, 134, 107, 5, 5, 136, 107, 255, 255, 255, 255, 5, 5, 138, 107, 5, 5, 140, 107, 5, 5, 142, 107, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 6, 107, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 84, 109, 5, 5, 86, 109, 5, 5, 88, 109, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 201, 224, 10, 0, 201, 64, 12, 0, 5, 5, 94, 109, 255, 255, 255, 255, 5, 5, 96, 109, 5, 5, 98, 109, 5, 5, 100, 109, 5, 5, 102, 109, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 6, 109, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 104, 109, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 44, 109, 5, 5, 46, 109, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 48, 109, 5, 5, 68, 109, 5, 5, 50, 109, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 52, 109, 5, 5, 54, 109, 5, 5, 56, 109, 5, 5, 66, 109, 5, 5, 58, 109, 5, 5, 64, 109, 5, 5, 62, 109, 5, 5, 60, 109, 5, 5, 72, 109, 5, 5, 74, 109, 5, 5, 76, 109, 5, 5, 78, 109, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 80, 109, 5, 5, 82, 109, 5, 5, 122, 110, 5, 5, 124, 110, 5, 5, 126, 110, 5, 5, 128, 110, 5, 5, 130, 110, 255, 255, 255, 255, 201, 38, 13, 0, 5, 5, 138, 110, 5, 5, 140, 110, 255, 255, 255, 255, 5, 5, 142, 110, 5, 5, 144, 110, 5, 5, 146, 110, 5, 5, 148, 110, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 150, 110, 5, 5, 152, 110, 255, 255, 255, 255, 5, 5, 50, 110, 5, 5, 56, 110, 5, 5, 114, 110, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 64, 110, 5, 5, 66, 110, 5, 5, 68, 110, 5, 5, 70, 110, 5, 5, 72, 110, 5, 5, 74, 110, 5, 5, 76, 110, 5, 5, 78, 110, 5, 5, 80, 110, 255, 255, 255, 255, 5, 5, 82, 110, 5, 5, 84, 110, 5, 5, 86, 110, 5, 5, 88, 110, 5, 5, 90, 110, 5, 5, 92, 110, 5, 5, 94, 110, 5, 5, 96, 110, 5, 5, 98, 110, 5, 5, 110, 110, 5, 5, 112, 110, 5, 5, 100, 110, 5, 5, 102, 110, 5, 5, 104, 110, 5, 5, 106, 110, 5, 5, 108, 110, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 116, 110, 5, 5, 118, 110, 5, 5, 120, 110, 5, 5, 122, 111, 5, 5, 124, 111, 5, 5, 126, 111, 5, 5, 128, 111, 5, 5, 130, 111, 255, 255, 255, 255, 201, 224, 14, 0, 5, 5, 138, 111, 5, 5, 140, 111, 255, 255, 255, 255, 201, 96, 17, 0, 5, 5, 144, 111, 5, 5, 146, 111, 5, 5, 148, 111, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 150, 111, 5, 5, 152, 111, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 108, 111, 255, 255, 255, 255, 5, 5, 62, 112, 5, 5, 64, 112, 5, 5, 66, 112, 5, 5, 68, 112, 5, 5, 70, 112, 5, 5, 72, 112, 5, 5, 74, 112, 5, 5, 76, 112, 5, 5, 78, 112, 5, 5, 80, 112, 5, 5, 82, 112, 5, 5, 84, 112, 5, 5, 86, 112, 5, 5, 88, 112, 5, 5, 90, 112, 5, 5, 92, 112, 5, 5, 94, 112, 5, 5, 112, 112, 5, 5, 96, 112, 5, 5, 108, 112, 5, 5, 110, 112, 5, 5, 98, 112, 5, 5, 100, 112, 5, 5, 102, 112, 5, 5, 104, 112, 5, 5, 106, 112, 5, 5, 114, 112, 32, 5, 152, 112, 32, 5, 152, 112, 5, 5, 116, 112, 5, 5, 120, 112, 5, 5, 122, 112, 5, 5, 54, 113, 5, 5, 56, 113, 5, 5, 58, 113, 5, 5, 60, 113, 5, 5, 62, 113, 5, 5, 64, 113, 5, 5, 66, 113, 5, 5, 68, 113, 5, 5, 70, 113, 5, 5, 72, 113, 5, 5, 74, 113, 5, 5, 76, 113, 5, 5, 78, 113, 5, 5, 80, 113, 5, 5, 82, 113, 5, 5, 84, 113, 5, 5, 86, 113, 5, 5, 88, 113, 255, 255, 255, 255, 5, 5, 90, 113, 5, 5, 92, 113, 5, 5, 94, 113, 5, 5, 96, 113, 5, 5, 98, 113, 5, 5, 100, 113, 5, 5, 102, 113, 5, 5, 104, 113, 5, 5, 106, 113, 255, 255, 255, 255, 5, 5, 108, 113, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 110, 113, 5, 5, 112, 113, 5, 5, 114, 113, 5, 5, 116, 113, 5, 5, 118, 113, 5, 5, 120, 113, 5, 5, 122, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 158, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 124, 113, 5, 5, 126, 113, 5, 5, 128, 113, 5, 5, 130, 113, 5, 5, 132, 113, 5, 5, 134, 113, 255, 255, 255, 255, 5, 5, 136, 113, 255, 255, 255, 255, 5, 5, 138, 113, 201, 132, 20, 0, 5, 5, 148, 113, 5, 5, 150, 113, 201, 6, 23, 0, 5, 5, 154, 113, 5, 5, 156, 113, 5, 5, 142, 113, 193, 16, 160, 113, 193, 23, 160, 113, 193, 30, 160, 113, 193, 37, 160, 113, 193, 44, 160, 113, 193, 51, 160, 113, 193, 58, 160, 113, 193, 65, 160, 113, 193, 72, 160, 113, 193, 79, 160, 113, 193, 86, 160, 113, 193, 93, 160, 113, 193, 100, 160, 113, 193, 107, 160, 113, 193, 114, 160, 113, 193, 121, 160, 113, 193, 128, 160, 113, 193, 135, 160, 113, 193, 142, 160, 113, 193, 149, 160, 113, 193, 156, 160, 113, 193, 163, 160, 113, 193, 170, 160, 113, 193, 177, 160, 113, 193, 184, 160, 113, 193, 191, 160, 113, 193, 198, 160, 113, 193, 126, 161, 113, 193, 133, 161, 113, 193, 140, 161, 113, 193, 147, 161, 113, 193, 154, 161, 113, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 96, 233, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 206, 64, 237, 0, 193, 133, 171, 113, 193, 140, 171, 113, 193, 147, 171, 113, 193, 154, 171, 113, 193, 161, 171, 113, 193, 168, 171, 113, 193, 175, 171, 113, 193, 182, 171, 113, 193, 189, 171, 113, 193, 196, 171, 113, 193, 210, 171, 113, 193, 217, 171, 113, 193, 224, 171, 113, 193, 9, 172, 113, 193, 16, 172, 113, 193, 23, 172, 113, 193, 30, 172, 113, 193, 37, 172, 113, 193, 44, 172, 113, 193, 51, 172, 113, 193, 58, 172, 113, 193, 65, 172, 113, 193, 72, 172, 113, 193, 79, 172, 113, 193, 86, 172, 113, 193, 93, 172, 113, 193, 100, 172, 113, 193, 107, 172, 113, 193, 114, 172, 113, 193, 121, 172, 113, 193, 128, 172, 113, 255, 255, 255, 255, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 206, 96, 237, 0, 193, 9, 176, 113, 193, 16, 176, 113, 193, 23, 176, 113, 193, 30, 176, 113, 193, 51, 176, 113, 193, 58, 176, 113, 193, 65, 176, 113, 255, 255, 255, 255, 193, 72, 176, 113, 255, 255, 255, 255, 193, 79, 176, 113, 193, 86, 176, 113, 193, 93, 176, 113, 193, 100, 176, 113, 255, 255, 255, 255, 193, 107, 176, 113, 193, 114, 176, 113, 193, 121, 176, 113, 193, 128, 176, 113, 193, 135, 176, 113, 193, 142, 176, 113, 193, 149, 176, 113, 193, 156, 176, 113, 193, 163, 176, 113, 193, 170, 176, 113, 193, 177, 176, 113, 193, 184, 176, 113, 193, 191, 176, 113, 193, 198, 176, 113, 193, 205, 176, 113, 255, 255, 255, 255, 193, 212, 176, 113, 193, 133, 179, 113, 193, 140, 179, 113, 193, 147, 179, 113, 193, 154, 179, 113, 193, 161, 179, 113, 255, 255, 255, 255, 255, 255, 255, 255, 201, 32, 25, 2, 193, 189, 179, 113, 255, 255, 255, 255, 255, 255, 255, 255, 193, 196, 179, 113, 193, 203, 179, 113, 193, 210, 179, 113, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 178, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 217, 179, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 224, 179, 113, 193, 105, 179, 113, 193, 112, 179, 113, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 206, 192, 237, 0, 193, 175, 181, 113, 193, 182, 181, 113, 193, 189, 181, 113, 193, 196, 181, 113, 193, 203, 181, 113, 193, 210, 181, 113, 193, 217, 181, 113, 193, 224, 181, 113, 193, 231, 181, 113, 193, 238, 181, 113, 193, 245, 181, 113, 193, 147, 181, 113, 193, 154, 181, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 206, 224, 237, 0, 193, 9, 185, 113, 193, 16, 185, 113, 5, 5, 187, 113, 5, 5, 189, 113, 5, 5, 191, 113, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 0, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 206, 32, 238, 0, 193, 9, 203, 113, 193, 16, 203, 113, 193, 23, 203, 113, 193, 30, 203, 113, 193, 37, 203, 113, 193, 44, 203, 113, 193, 51, 203, 113, 193, 58, 203, 113, 193, 65, 203, 113, 193, 72, 203, 113, 193, 100, 203, 113, 193, 107, 203, 113, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 206, 64, 238, 0, 193, 9, 205, 113, 193, 16, 205, 113, 193, 23, 205, 113, 193, 30, 205, 113, 193, 37, 205, 113, 193, 44, 205, 113, 193, 51, 205, 113, 255, 255, 255, 255, 255, 255, 255, 255, 193, 58, 205, 113, 255, 255, 255, 255, 255, 255, 255, 255, 193, 65, 205, 113, 193, 72, 205, 113, 193, 79, 205, 113, 193, 86, 205, 113, 193, 93, 205, 113, 193, 107, 205, 113, 193, 114, 205, 113, 193, 121, 205, 113, 255, 255, 255, 255, 193, 128, 205, 113, 193, 135, 205, 113, 255, 255, 255, 255, 206, 128, 238, 0, 206, 128, 238, 0, 206, 128, 238, 0, 206, 128, 238, 0, 206, 128, 238, 0, 206, 128, 238, 0, 206, 128, 238, 0, 206, 128, 238, 0, 193, 9, 207, 113, 193, 16, 207, 113, 193, 23, 207, 113, 193, 30, 207, 113, 193, 37, 207, 113, 193, 44, 207, 113, 193, 51, 207, 113, 193, 58, 207, 113, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 160, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 206, 96, 238, 0, 193, 9, 211, 113, 193, 16, 211, 113, 193, 23, 211, 113, 193, 30, 211, 113, 5, 5, 213, 113, 5, 112, 213, 113, 193, 2, 215, 113, 193, 9, 215, 113, 5, 5, 217, 113, 193, 2, 219, 113, 193, 9, 219, 113, 193, 16, 219, 113, 193, 23, 219, 113, 193, 30, 219, 113, 193, 37, 219, 113, 193, 44, 219, 113, 193, 51, 219, 113, 193, 58, 219, 113, 193, 65, 219, 113, 193, 72, 219, 113, 193, 79, 219, 113, 5, 5, 221, 113, 5, 112, 221, 113, 193, 2, 223, 113, 193, 9, 223, 113, 193, 16, 223, 113, 5, 112, 217, 113, 255, 255, 255, 255, 255, 255, 255, 255, 193, 107, 223, 113, 193, 114, 223, 113, 193, 121, 223, 113, 193, 9, 224, 113, 193, 16, 224, 113, 193, 23, 224, 113, 193, 30, 224, 113, 193, 37, 224, 113, 193, 44, 224, 113, 193, 51, 224, 113, 255, 255, 255, 255, 193, 58, 224, 113, 193, 65, 224, 113, 255, 255, 255, 255, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 206, 64, 239, 0, 5, 5, 226, 113, 193, 2, 228, 113, 193, 9, 228, 113, 193, 16, 228, 113, 193, 23, 228, 113, 193, 30, 228, 113, 193, 37, 228, 113, 193, 44, 228, 113, 193, 51, 228, 113, 193, 58, 228, 113, 193, 65, 228, 113, 193, 72, 228, 113, 193, 79, 228, 113, 193, 86, 228, 113, 193, 93, 228, 113, 193, 100, 228, 113, 193, 107, 228, 113, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 114, 228, 113, 255, 255, 255, 255, 193, 121, 228, 113, 193, 128, 228, 113, 255, 255, 255, 255, 193, 135, 228, 113, 193, 16, 229, 113, 193, 23, 229, 113, 193, 30, 229, 113, 193, 37, 229, 113, 193, 44, 229, 113, 193, 51, 229, 113, 255, 255, 255, 255, 193, 58, 229, 113, 193, 65, 229, 113, 255, 255, 255, 255, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 96, 239, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 128, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 206, 160, 236, 0, 193, 9, 248, 113, 193, 16, 248, 113, 193, 23, 248, 113, 193, 30, 248, 113, 193, 37, 248, 113, 193, 44, 248, 113, 193, 51, 248, 113, 193, 58, 248, 113, 193, 65, 248, 113, 255, 255, 255, 255, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 206, 224, 238, 0, 255, 255, 255, 255, 5, 5, 6, 114, 5, 5, 8, 114, 5, 5, 10, 114, 5, 5, 12, 114, 5, 5, 14, 114, 5, 5, 16, 114, 5, 5, 18, 114, 5, 5, 20, 114, 5, 5, 22, 114, 5, 5, 24, 114, 5, 5, 26, 114, 5, 5, 28, 114, 5, 5, 30, 114, 5, 5, 32, 114, 5, 5, 34, 114, 5, 5, 36, 114, 5, 5, 38, 114, 5, 5, 40, 114, 5, 5, 42, 114, 5, 5, 44, 114, 5, 5, 46, 114, 5, 5, 48, 114, 5, 5, 50, 114, 5, 5, 52, 114, 5, 5, 54, 114, 5, 5, 56, 114, 5, 5, 58, 114, 5, 5, 60, 114, 5, 5, 62, 114, 5, 5, 64, 114, 5, 5, 66, 114, 255, 255, 255, 255, 5, 5, 8, 115, 5, 5, 10, 115, 255, 255, 255, 255, 5, 5, 12, 115, 255, 255, 255, 255, 5, 5, 14, 115, 5, 5, 16, 115, 5, 5, 18, 115, 5, 5, 20, 115, 5, 5, 24, 115, 255, 255, 255, 255, 5, 5, 26, 115, 5, 5, 32, 115, 5, 5, 28, 115, 5, 5, 34, 115, 5, 5, 36, 115, 5, 5, 38, 115, 5, 5, 40, 115, 5, 5, 42, 115, 5, 5, 44, 115, 5, 5, 46, 115, 5, 5, 48, 115, 5, 5, 50, 115, 5, 5, 52, 115, 5, 5, 54, 115, 5, 5, 56, 115, 5, 5, 58, 115, 5, 5, 60, 115, 5, 5, 62, 115, 5, 5, 64, 115, 5, 5, 66, 115, 5, 5, 68, 115, 5, 5, 70, 115, 5, 5, 72, 115, 5, 5, 74, 115, 255, 255, 255, 255, 5, 5, 76, 115, 255, 255, 255, 255, 5, 5, 78, 115, 5, 5, 80, 115, 5, 5, 82, 115, 5, 5, 22, 115, 5, 5, 84, 115, 5, 5, 86, 115, 5, 5, 88, 115, 5, 5, 90, 115, 5, 5, 92, 115, 5, 5, 94, 115, 5, 5, 96, 115, 5, 5, 98, 115, 5, 5, 100, 115, 5, 5, 102, 115, 5, 5, 104, 115, 5, 5, 106, 115, 5, 5, 108, 115, 5, 5, 110, 115, 5, 5, 112, 115, 5, 5, 114, 115, 5, 5, 116, 115, 5, 5, 118, 115, 5, 5, 120, 115, 255, 255, 255, 255, 255, 255, 255, 255, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 206, 0, 234, 0, 5, 5, 6, 116, 5, 5, 12, 116, 5, 5, 16, 116, 197, 66, 120, 2, 5, 5, 20, 116, 5, 5, 24, 116, 5, 5, 28, 116, 5, 5, 32, 116, 255, 255, 255, 255, 5, 5, 36, 116, 5, 5, 40, 116, 5, 5, 44, 116, 5, 5, 48, 116, 197, 194, 120, 2, 5, 5, 52, 116, 5, 5, 56, 116, 5, 5, 60, 116, 5, 5, 64, 116, 197, 66, 121, 2, 5, 5, 68, 116, 5, 5, 72, 116, 5, 5, 76, 116, 5, 5, 80, 116, 197, 194, 121, 2, 5, 5, 84, 116, 5, 5, 88, 116, 5, 5, 92, 116, 5, 5, 96, 116, 197, 66, 122, 2, 5, 5, 100, 116, 5, 5, 104, 116, 5, 5, 108, 116, 193, 30, 208, 116, 193, 37, 208, 116, 193, 44, 208, 116, 193, 51, 208, 116, 193, 58, 208, 116, 193, 65, 208, 116, 193, 72, 208, 116, 193, 79, 208, 116, 193, 86, 208, 116, 193, 93, 208, 116, 193, 100, 208, 116, 193, 9, 198, 116, 193, 23, 198, 116, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 206, 192, 238, 0, 5, 5, 219, 116, 193, 2, 221, 116, 193, 9, 221, 116, 193, 16, 221, 116, 193, 23, 221, 116, 193, 30, 221, 116, 193, 37, 221, 116, 193, 44, 221, 116, 193, 51, 221, 116, 193, 58, 221, 116, 193, 65, 221, 116, 5, 5, 223, 116, 193, 2, 225, 116, 193, 9, 225, 116, 193, 16, 225, 116, 5, 5, 227, 116, 5, 5, 229, 116, 193, 2, 231, 116, 193, 9, 231, 116, 193, 16, 231, 116, 5, 5, 233, 116, 193, 2, 235, 116, 193, 9, 235, 116, 193, 16, 235, 116, 193, 23, 235, 116, 193, 30, 235, 116, 5, 5, 237, 116, 5, 5, 239, 116, 5, 5, 241, 116, 5, 5, 243, 116, 5, 5, 245, 116, 5, 5, 247, 116, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 206, 0, 239, 0, 193, 70, 5, 117, 193, 84, 5, 117, 193, 98, 5, 117, 193, 112, 5, 117, 193, 126, 5, 117, 193, 140, 5, 117, 193, 154, 5, 117, 255, 255, 255, 255, 255, 255, 255, 255, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 206, 32, 239, 0, 193, 9, 6, 117, 193, 16, 6, 117, 193, 23, 6, 117, 193, 30, 6, 117, 193, 37, 6, 117, 193, 44, 6, 117, 193, 51, 6, 117, 193, 58, 6, 117, 193, 65, 6, 117, 193, 72, 6, 117, 193, 100, 6, 117, 193, 107, 6, 117, 193, 114, 6, 117, 193, 121, 6, 117, 193, 128, 6, 117, 193, 135, 6, 117, 193, 142, 6, 117, 193, 149, 6, 117, 193, 156, 6, 117, 193, 163, 6, 117, 193, 170, 6, 117, 193, 177, 6, 117, 193, 184, 6, 117, 193, 191, 6, 117, 193, 198, 6, 117, 193, 205, 6, 117, 193, 212, 6, 117, 193, 226, 6, 117, 193, 240, 6, 117, 193, 247, 6, 117, 193, 254, 6, 117, 193, 7, 7, 117, 193, 9, 8, 117, 193, 16, 8, 117, 193, 23, 8, 117, 193, 30, 8, 117, 193, 37, 8, 117, 193, 44, 8, 117, 193, 51, 8, 117, 193, 58, 8, 117, 193, 93, 8, 117, 193, 100, 8, 117, 193, 107, 8, 117, 193, 114, 8, 117, 193, 121, 8, 117, 193, 128, 8, 117, 193, 135, 8, 117, 193, 142, 8, 117, 193, 149, 8, 117, 193, 156, 8, 117, 193, 163, 8, 117, 193, 170, 8, 117, 193, 184, 8, 117, 193, 191, 8, 117, 193, 198, 8, 117, 193, 205, 8, 117, 193, 226, 8, 117, 193, 247, 8, 117, 193, 254, 8, 117, 193, 14, 9, 117, 193, 21, 9, 117, 193, 42, 9, 117, 193, 77, 9, 117, 193, 84, 9, 117, 193, 9, 10, 117, 193, 16, 10, 117, 193, 23, 10, 117, 193, 30, 10, 117, 193, 37, 10, 117, 193, 44, 10, 117, 193, 51, 10, 117, 193, 58, 10, 117, 5, 5, 12, 117, 193, 2, 14, 117, 193, 9, 14, 117, 5, 5, 16, 117, 193, 2, 18, 117, 193, 9, 18, 117, 193, 16, 18, 117, 193, 23, 18, 117, 193, 30, 18, 117, 193, 37, 18, 117, 193, 44, 18, 117, 193, 51, 18, 117, 193, 58, 18, 117, 193, 65, 18, 117, 193, 72, 18, 117, 193, 79, 18, 117, 193, 86, 18, 117, 193, 93, 18, 117, 193, 100, 18, 117, 193, 107, 18, 117, 193, 114, 18, 117, 197, 162, 124, 2, 197, 226, 124, 2, 255, 255, 255, 255, 193, 9, 25, 117, 193, 16, 25, 117, 193, 23, 25, 117, 193, 30, 25, 117, 193, 37, 25, 117, 193, 44, 25, 117, 193, 51, 25, 117, 193, 58, 25, 117, 193, 65, 25, 117, 193, 72, 25, 117, 193, 79, 25, 117, 193, 86, 25, 117, 193, 93, 25, 117, 255, 255, 255, 255, 193, 100, 25, 117, 193, 107, 25, 117, 193, 114, 25, 117, 193, 121, 25, 117, 193, 128, 25, 117, 193, 135, 25, 117, 193, 142, 25, 117, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 27, 117, 193, 16, 27, 117, 193, 23, 27, 117, 193, 30, 27, 117, 193, 37, 27, 117, 193, 44, 27, 117, 193, 51, 27, 117, 193, 58, 27, 117, 193, 65, 27, 117, 193, 72, 27, 117, 193, 79, 27, 117, 193, 86, 27, 117, 193, 93, 27, 117, 193, 100, 27, 117, 193, 107, 27, 117, 193, 114, 27, 117, 193, 121, 27, 117, 193, 128, 27, 117, 193, 135, 27, 117, 193, 142, 27, 117, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 28, 117, 193, 16, 28, 117, 193, 23, 28, 117, 193, 30, 28, 117, 193, 37, 28, 117, 193, 44, 28, 117, 193, 51, 28, 117, 193, 58, 28, 117, 193, 65, 28, 117, 193, 72, 28, 117, 193, 79, 28, 117, 193, 86, 28, 117, 193, 93, 28, 117, 255, 255, 255, 255, 193, 100, 28, 117, 193, 107, 28, 117, 193, 114, 28, 117, 255, 255, 255, 255, 193, 121, 28, 117, 193, 128, 28, 117, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 33, 117, 32, 5, 33, 117, 5, 5, 35, 117, 32, 5, 35, 117, 32, 5, 35, 117, 5, 5, 37, 117, 32, 5, 37, 117, 5, 5, 39, 117, 32, 5, 39, 117, 5, 5, 41, 117, 32, 5, 41, 117, 5, 5, 43, 117, 32, 5, 43, 117, 32, 5, 43, 117, 5, 5, 45, 117, 32, 5, 45, 117, 193, 2, 47, 117, 193, 9, 47, 117, 5, 5, 49, 117, 32, 5, 49, 117, 5, 5, 51, 117, 32, 5, 51, 117, 5, 5, 53, 117, 32, 5, 53, 117, 5, 5, 55, 117, 32, 5, 55, 117, 32, 5, 55, 117, 5, 5, 57, 117, 32, 5, 57, 117, 5, 5, 59, 117, 5, 5, 61, 117, 32, 5, 61, 117, 193, 9, 79, 117, 193, 23, 79, 117, 193, 37, 79, 117, 193, 65, 79, 117, 193, 86, 79, 117, 193, 100, 79, 117, 193, 121, 79, 117, 193, 149, 79, 117, 193, 184, 79, 117, 193, 233, 79, 117, 193, 7, 80, 117, 193, 14, 80, 117, 193, 28, 80, 117, 193, 42, 80, 117, 193, 63, 80, 117, 193, 84, 80, 117, 193, 112, 80, 117, 193, 119, 80, 117, 193, 126, 80, 117, 193, 147, 80, 117, 193, 168, 80, 117, 193, 196, 80, 117, 193, 203, 80, 117, 193, 245, 80, 117, 193, 12, 81, 117, 193, 33, 81, 117, 193, 47, 81, 117, 193, 61, 81, 117, 193, 89, 81, 117, 193, 103, 81, 117, 5, 5, 83, 117, 193, 9, 85, 117, 193, 96, 81, 117, 193, 219, 79, 117, 193, 156, 89, 117, 193, 9, 93, 117, 193, 16, 93, 117, 193, 145, 81, 117, 193, 93, 85, 117, 193, 163, 89, 117, 193, 170, 89, 117, 193, 23, 93, 117, 193, 30, 93, 117, 193, 37, 93, 117, 193, 44, 93, 117, 193, 51, 93, 117, 193, 91, 80, 117, 193, 79, 85, 117, 193, 86, 85, 117, 193, 30, 89, 117, 193, 9, 89, 117, 193, 58, 89, 117, 193, 65, 89, 117, 193, 16, 79, 117, 193, 30, 79, 117, 193, 44, 79, 117, 193, 107, 79, 117, 193, 170, 79, 117, 193, 240, 79, 117, 193, 133, 80, 117, 193, 175, 80, 117, 193, 210, 80, 117, 193, 217, 80, 117, 193, 252, 80, 117, 193, 9, 94, 117, 193, 51, 94, 117, 193, 58, 94, 117, 193, 65, 94, 117, 193, 72, 94, 117, 193, 79, 94, 117, 193, 86, 94, 117, 193, 93, 94, 117, 193, 100, 94, 117, 193, 107, 94, 117, 193, 114, 94, 117, 193, 121, 94, 117, 193, 128, 94, 117, 193, 135, 94, 117, 193, 142, 94, 117, 193, 149, 94, 117, 193, 156, 94, 117, 193, 163, 94, 117, 193, 170, 94, 117, 193, 177, 94, 117, 193, 184, 94, 117, 193, 191, 94, 117, 193, 198, 94, 117, 193, 205, 94, 117, 193, 212, 94, 117, 193, 219, 94, 117, 193, 226, 94, 117, 193, 233, 94, 117, 193, 240, 94, 117, 193, 16, 94, 117, 193, 23, 94, 117, 193, 30, 94, 117, 5, 5, 100, 117, 5, 5, 102, 117, 5, 5, 104, 117, 5, 5, 106, 117, 5, 5, 108, 117, 5, 5, 110, 117, 5, 5, 112, 117, 5, 5, 114, 117, 5, 5, 116, 117, 5, 5, 118, 117, 5, 5, 120, 117, 5, 5, 122, 117, 5, 5, 124, 117, 5, 5, 126, 117, 5, 5, 128, 117, 5, 5, 130, 117, 5, 5, 132, 117, 5, 5, 134, 117, 5, 5, 136, 117, 5, 5, 138, 117, 5, 5, 140, 117, 5, 5, 142, 117, 5, 5, 144, 117, 5, 5, 146, 117, 5, 5, 148, 117, 5, 5, 150, 117, 5, 5, 152, 117, 5, 5, 154, 117, 5, 5, 156, 117, 5, 5, 158, 117, 5, 5, 160, 117, 5, 5, 162, 117, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 206, 32, 228, 0, 255, 255, 255, 255, 255, 255, 255, 255, 193, 219, 4, 118, 193, 226, 4, 118, 193, 233, 4, 118, 193, 240, 4, 118, 193, 247, 4, 118, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 5, 118, 193, 16, 5, 118, 193, 23, 5, 118, 193, 30, 5, 118, 193, 37, 5, 118, 193, 44, 5, 118, 193, 51, 5, 118, 193, 58, 5, 118, 193, 65, 5, 118, 193, 72, 5, 118, 193, 79, 5, 118, 193, 86, 5, 118, 193, 93, 5, 118, 193, 100, 5, 118, 193, 107, 5, 118, 193, 114, 5, 118, 193, 121, 5, 118, 193, 128, 5, 118, 193, 135, 5, 118, 193, 142, 5, 118, 193, 149, 5, 118, 193, 156, 5, 118, 193, 163, 5, 118, 193, 170, 5, 118, 193, 177, 5, 118, 193, 184, 5, 118, 193, 191, 5, 118, 193, 198, 5, 118, 5, 5, 7, 118, 193, 2, 9, 118, 193, 9, 9, 118, 193, 16, 9, 118, 193, 23, 9, 118, 193, 30, 9, 118, 193, 37, 9, 118, 193, 44, 9, 118, 193, 51, 9, 118, 193, 58, 9, 118, 193, 65, 9, 118, 193, 72, 9, 118, 193, 79, 9, 118, 193, 86, 9, 118, 193, 93, 9, 118, 193, 100, 9, 118, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 107, 9, 118, 193, 114, 9, 118, 193, 121, 9, 118, 193, 128, 9, 118, 193, 135, 9, 118, 201, 0, 242, 0, 201, 160, 5, 1, 201, 64, 25, 1, 193, 9, 13, 118, 193, 16, 13, 118, 201, 224, 44, 1, 193, 30, 13, 118, 193, 37, 13, 118, 193, 44, 13, 118, 193, 51, 13, 118, 193, 58, 13, 118, 193, 9, 18, 118, 193, 16, 18, 118, 193, 23, 18, 118, 193, 30, 18, 118, 193, 37, 18, 118, 193, 44, 18, 118, 5, 5, 20, 118, 193, 2, 22, 118, 193, 9, 22, 118, 193, 16, 22, 118, 193, 23, 22, 118, 193, 30, 22, 118, 193, 37, 22, 118, 193, 44, 22, 118, 193, 51, 22, 118, 193, 58, 22, 118, 193, 65, 22, 118, 193, 72, 22, 118, 193, 79, 22, 118, 193, 86, 22, 118, 193, 93, 22, 118, 193, 100, 22, 118, 193, 107, 22, 118, 193, 114, 22, 118, 193, 121, 22, 118, 193, 128, 22, 118, 193, 135, 22, 118, 5, 5, 24, 118, 206, 64, 228, 0, 206, 64, 228, 0, 206, 64, 228, 0, 206, 64, 228, 0, 206, 64, 228, 0, 206, 64, 228, 0, 5, 5, 28, 118, 193, 2, 30, 118, 193, 9, 30, 118, 193, 16, 30, 118, 193, 23, 30, 118, 193, 30, 30, 118, 193, 37, 30, 118, 193, 100, 30, 118, 193, 107, 30, 118, 193, 114, 30, 118, 193, 121, 30, 118, 193, 128, 30, 118, 193, 135, 30, 118, 193, 44, 30, 118, 197, 35, 170, 2, 193, 58, 30, 118, 193, 65, 30, 118, 193, 72, 30, 118, 32, 5, 20, 118, 32, 5, 20, 118, 32, 5, 24, 118, 32, 5, 24, 118, 193, 79, 30, 118, 193, 86, 30, 118, 193, 93, 30, 118, 255, 255, 255, 255, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 206, 224, 233, 0, 193, 70, 39, 118, 193, 77, 39, 118, 193, 84, 39, 118, 193, 91, 39, 118, 193, 98, 39, 118, 193, 105, 39, 118, 193, 112, 39, 118, 193, 119, 39, 118, 193, 126, 39, 118, 193, 133, 39, 118, 193, 42, 39, 118, 193, 49, 39, 118, 193, 56, 39, 118, 193, 63, 39, 118, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 52, 118, 193, 16, 52, 118, 193, 23, 52, 118, 193, 30, 52, 118, 193, 37, 52, 118, 193, 44, 52, 118, 193, 51, 52, 118, 193, 72, 52, 118, 193, 128, 52, 118, 193, 124, 54, 118, 193, 247, 52, 118, 193, 245, 53, 118, 193, 3, 55, 118, 193, 17, 55, 118, 193, 24, 55, 118, 193, 38, 55, 118, 193, 45, 55, 118, 193, 66, 55, 118, 193, 80, 55, 118, 193, 87, 55, 118, 193, 101, 55, 118, 193, 115, 55, 118, 193, 129, 55, 118, 193, 136, 55, 118, 193, 52, 55, 118, 193, 122, 55, 118, 193, 77, 53, 118, 193, 254, 52, 118, 193, 133, 53, 118, 193, 140, 53, 118, 193, 10, 55, 118, 193, 31, 55, 118, 193, 215, 54, 118, 193, 222, 54, 118, 193, 229, 54, 118, 193, 58, 52, 118, 193, 86, 52, 118, 193, 107, 52, 118, 193, 149, 52, 118, 193, 163, 52, 118, 193, 184, 52, 118, 193, 198, 52, 118, 193, 233, 52, 118, 193, 14, 53, 118, 193, 28, 53, 118, 193, 49, 53, 118, 193, 63, 53, 118, 193, 98, 53, 118, 193, 175, 53, 118, 193, 196, 53, 118, 193, 217, 53, 118, 193, 5, 54, 118, 193, 145, 54, 118, 193, 33, 54, 118, 193, 68, 54, 118, 193, 103, 54, 118, 193, 173, 54, 118, 193, 194, 54, 118, 193, 236, 54, 118, 193, 243, 54, 118, 193, 238, 53, 118, 193, 93, 52, 118, 193, 114, 52, 118, 193, 135, 52, 118, 193, 65, 52, 118, 193, 79, 52, 118, 193, 100, 52, 118, 193, 142, 52, 118, 193, 156, 52, 118, 193, 177, 52, 118, 193, 191, 52, 118, 193, 212, 52, 118, 193, 219, 52, 118, 193, 226, 52, 118, 193, 7, 53, 118, 193, 21, 53, 118, 193, 42, 53, 118, 193, 56, 53, 118, 193, 91, 53, 118, 193, 105, 53, 118, 193, 112, 53, 118, 193, 119, 53, 118, 193, 168, 53, 118, 193, 189, 53, 118, 193, 210, 53, 118, 193, 252, 53, 118, 193, 26, 54, 118, 193, 47, 54, 118, 193, 61, 54, 118, 193, 75, 54, 118, 193, 96, 54, 118, 193, 131, 54, 118, 193, 138, 54, 118, 193, 159, 54, 118, 193, 187, 54, 118, 193, 208, 54, 118, 193, 205, 52, 118, 193, 170, 52, 118, 193, 240, 52, 118, 193, 110, 54, 118, 193, 70, 53, 118, 193, 84, 53, 118, 193, 35, 53, 118, 193, 126, 53, 118, 193, 182, 53, 118, 193, 203, 53, 118, 193, 12, 54, 118, 193, 82, 54, 118, 193, 180, 54, 118, 193, 201, 54, 118, 193, 152, 54, 118, 193, 166, 54, 118, 193, 250, 54, 118, 193, 224, 53, 118, 193, 40, 54, 118, 193, 121, 52, 118, 193, 117, 54, 118, 193, 54, 54, 118, 193, 89, 54, 118, 193, 19, 54, 118, 193, 231, 53, 118, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 59, 55, 118, 193, 73, 55, 118, 193, 147, 53, 118, 193, 108, 55, 118, 193, 154, 53, 118, 193, 161, 53, 118, 193, 143, 55, 118, 193, 150, 55, 118, 193, 94, 55, 118, 193, 164, 55, 118, 193, 157, 55, 118, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 160, 5, 60, 118, 160, 5, 62, 118, 160, 5, 64, 118, 160, 5, 66, 118, 160, 5, 68, 118, 160, 5, 70, 118, 160, 5, 72, 118, 160, 5, 74, 118, 160, 5, 76, 118, 160, 5, 78, 118, 160, 5, 80, 118, 160, 5, 82, 118, 160, 5, 84, 118, 160, 5, 86, 118, 160, 5, 88, 118, 160, 5, 90, 118, 160, 5, 92, 118, 160, 5, 94, 118, 160, 5, 96, 118, 160, 5, 98, 118, 160, 5, 100, 118, 160, 5, 102, 118, 160, 5, 104, 118, 160, 5, 106, 118, 160, 5, 108, 118, 160, 5, 110, 118, 160, 5, 112, 118, 160, 5, 114, 118, 160, 5, 116, 118, 160, 5, 118, 118, 160, 5, 120, 118, 160, 5, 122, 118, 5, 5, 92, 118, 5, 5, 94, 118, 5, 5, 96, 118, 5, 5, 98, 118, 5, 5, 100, 118, 5, 5, 102, 118, 5, 5, 104, 118, 5, 5, 106, 118, 5, 5, 108, 118, 5, 5, 110, 118, 5, 5, 112, 118, 5, 5, 114, 118, 5, 5, 116, 118, 5, 5, 118, 118, 5, 5, 120, 118, 5, 5, 122, 118, 5, 5, 124, 118, 5, 5, 126, 118, 5, 5, 128, 118, 5, 5, 130, 118, 5, 5, 132, 118, 5, 5, 134, 118, 5, 5, 136, 118, 5, 5, 138, 118, 5, 5, 140, 118, 5, 5, 142, 118, 5, 5, 144, 118, 5, 5, 146, 118, 5, 5, 148, 118, 5, 5, 150, 118, 5, 5, 152, 118, 5, 5, 154, 118, 160, 5, 124, 118, 160, 5, 126, 118, 160, 5, 128, 118, 160, 5, 130, 118, 160, 5, 132, 118, 160, 5, 134, 118, 160, 5, 136, 118, 160, 5, 138, 118, 160, 5, 140, 118, 160, 5, 142, 118, 160, 5, 144, 118, 160, 5, 146, 118, 160, 5, 148, 118, 160, 5, 150, 118, 160, 5, 152, 118, 160, 5, 154, 118, 160, 5, 156, 118, 160, 5, 158, 118, 160, 5, 160, 118, 160, 5, 162, 118, 160, 5, 164, 118, 160, 5, 166, 118, 160, 5, 168, 118, 160, 5, 170, 118, 160, 5, 172, 118, 160, 5, 174, 118, 160, 5, 176, 118, 160, 5, 178, 118, 160, 5, 180, 118, 160, 5, 182, 118, 160, 5, 184, 118, 160, 5, 186, 118, 5, 5, 156, 118, 5, 5, 158, 118, 5, 5, 160, 118, 5, 5, 162, 118, 5, 5, 164, 118, 5, 5, 166, 118, 5, 5, 168, 118, 5, 5, 170, 118, 5, 5, 172, 118, 5, 5, 174, 118, 5, 5, 176, 118, 5, 5, 178, 118, 5, 5, 180, 118, 5, 5, 182, 118, 5, 5, 184, 118, 5, 5, 186, 118, 5, 5, 188, 118, 5, 5, 190, 118, 5, 5, 192, 118, 5, 5, 194, 118, 5, 5, 196, 118, 5, 5, 198, 118, 5, 5, 200, 118, 5, 5, 202, 118, 5, 5, 204, 118, 5, 5, 206, 118, 5, 5, 208, 118, 5, 5, 210, 118, 5, 5, 212, 118, 5, 5, 214, 118, 5, 5, 216, 118, 5, 5, 218, 118, 160, 5, 188, 118, 160, 5, 190, 118, 160, 5, 192, 118, 160, 5, 194, 118, 160, 5, 196, 118, 160, 5, 198, 118, 160, 5, 200, 118, 160, 5, 202, 118, 160, 5, 204, 118, 160, 5, 206, 118, 160, 5, 208, 118, 160, 5, 210, 118, 160, 5, 212, 118, 160, 5, 214, 118, 160, 5, 216, 118, 160, 5, 218, 118, 160, 5, 220, 118, 160, 5, 222, 118, 160, 5, 224, 118, 160, 5, 226, 118, 160, 5, 228, 118, 160, 5, 230, 118, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 220, 118, 5, 5, 222, 118, 5, 5, 224, 118, 5, 5, 226, 118, 5, 5, 228, 118, 5, 5, 230, 118, 255, 255, 255, 255, 255, 255, 255, 255, 198, 193, 167, 0, 198, 225, 167, 0, 198, 1, 168, 0, 198, 33, 168, 0, 198, 65, 168, 0, 198, 97, 168, 0, 198, 129, 168, 0, 198, 161, 168, 0, 198, 193, 168, 0, 198, 225, 168, 0, 198, 1, 169, 0, 198, 33, 169, 0, 198, 65, 169, 0, 198, 97, 169, 0, 198, 129, 169, 0, 198, 161, 169, 0, 198, 193, 169, 0, 198, 225, 169, 0, 198, 1, 170, 0, 198, 33, 170, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 206, 96, 235, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 206, 160, 227, 0, 193, 218, 245, 118, 193, 122, 244, 118, 193, 136, 244, 118, 193, 143, 244, 118, 193, 150, 244, 118, 193, 157, 244, 118, 193, 164, 244, 118, 193, 171, 244, 118, 193, 178, 244, 118, 193, 185, 244, 118, 193, 192, 244, 118, 193, 199, 244, 118, 193, 206, 244, 118, 193, 213, 244, 118, 193, 220, 244, 118, 193, 227, 244, 118, 193, 234, 244, 118, 193, 241, 244, 118, 193, 255, 244, 118, 193, 8, 245, 118, 193, 15, 245, 118, 193, 22, 245, 118, 193, 29, 245, 118, 193, 36, 245, 118, 193, 43, 245, 118, 193, 50, 245, 118, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 192, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 224, 227, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 206, 0, 228, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 6, 119, 5, 112, 6, 119, 5, 5, 8, 119, 193, 51, 56, 119, 5, 112, 8, 119, 5, 120, 8, 119, 5, 5, 10, 119, 5, 112, 10, 119, 5, 5, 12, 119, 5, 112, 12, 119, 193, 23, 56, 119, 193, 37, 56, 119, 5, 116, 12, 119, 5, 120, 12, 119, 5, 124, 12, 119, 193, 9, 14, 119, 193, 16, 14, 119, 193, 23, 14, 119, 5, 5, 16, 119, 5, 112, 16, 119, 5, 116, 16, 119, 5, 120, 16, 119, 5, 124, 16, 119, 193, 9, 18, 119, 193, 72, 56, 119, 5, 5, 20, 119, 5, 5, 22, 119, 5, 112, 22, 119, 5, 117, 22, 119, 5, 122, 22, 119, 5, 5, 24, 119, 5, 112, 24, 119, 5, 120, 24, 119, 5, 5, 26, 119, 5, 112, 26, 119, 5, 5, 30, 119, 5, 112, 30, 119, 5, 5, 32, 119, 5, 112, 32, 119, 5, 5, 34, 119, 5, 5, 36, 119, 5, 5, 38, 119, 5, 5, 40, 119, 5, 112, 40, 119, 5, 118, 40, 119, 5, 121, 40, 119, 5, 124, 40, 119, 5, 5, 44, 119, 5, 112, 44, 119, 5, 120, 44, 119, 5, 5, 46, 119, 5, 112, 46, 119, 5, 120, 46, 119, 5, 112, 36, 119, 193, 2, 48, 119, 5, 5, 50, 119, 5, 112, 50, 119, 5, 120, 50, 119, 5, 5, 52, 119, 5, 112, 52, 119, 5, 5, 54, 119, 5, 112, 54, 119, 193, 2, 56, 119, 193, 9, 56, 119, 5, 5, 62, 119, 5, 112, 62, 119, 5, 5, 64, 119, 5, 5, 66, 119, 5, 5, 68, 119, 5, 5, 70, 119, 5, 5, 72, 119, 5, 5, 74, 119, 5, 5, 76, 119, 5, 5, 78, 119, 5, 112, 78, 119, 5, 120, 78, 119, 5, 5, 80, 119, 5, 5, 82, 119, 5, 5, 84, 119, 5, 5, 86, 119, 5, 5, 88, 119, 5, 112, 88, 119, 5, 5, 90, 119, 5, 5, 92, 119, 5, 5, 94, 119, 5, 5, 96, 119, 5, 5, 98, 119, 5, 5, 100, 119, 5, 5, 102, 119, 5, 5, 104, 119, 5, 5, 106, 119, 5, 5, 108, 119, 5, 112, 108, 119, 5, 5, 110, 119, 5, 112, 110, 119, 5, 120, 110, 119, 160, 5, 62, 119, 160, 112, 62, 119, 160, 5, 64, 119, 160, 5, 66, 119, 160, 5, 68, 119, 160, 5, 70, 119, 160, 5, 72, 119, 160, 5, 74, 119, 160, 5, 76, 119, 160, 5, 78, 119, 160, 112, 78, 119, 160, 120, 78, 119, 160, 5, 80, 119, 160, 5, 82, 119, 160, 5, 84, 119, 160, 5, 86, 119, 160, 5, 88, 119, 160, 112, 88, 119, 160, 5, 90, 119, 160, 5, 92, 119, 160, 5, 94, 119, 160, 5, 96, 119, 160, 5, 98, 119, 160, 5, 100, 119, 160, 5, 102, 119, 160, 5, 104, 119, 160, 5, 106, 119, 160, 5, 108, 119, 160, 112, 108, 119, 160, 5, 110, 119, 160, 112, 110, 119, 160, 120, 110, 119, 160, 5, 112, 119, 160, 5, 114, 119, 160, 5, 116, 119, 160, 112, 116, 119, 160, 5, 118, 119, 160, 5, 120, 119, 160, 5, 122, 119, 160, 5, 124, 119, 160, 5, 126, 119, 160, 5, 128, 119, 160, 5, 130, 119, 160, 112, 130, 119, 160, 5, 132, 119, 160, 112, 132, 119, 160, 5, 134, 119, 160, 5, 136, 119, 160, 5, 138, 119, 160, 5, 140, 119, 160, 5, 142, 119, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 146, 119, 5, 112, 146, 119, 5, 5, 148, 119, 5, 5, 150, 119, 5, 112, 150, 119, 193, 2, 152, 119, 193, 9, 152, 119, 5, 5, 154, 119, 5, 112, 154, 119, 5, 5, 156, 119, 5, 112, 156, 119, 5, 5, 158, 119, 5, 112, 158, 119, 5, 5, 160, 119, 5, 112, 160, 119, 5, 5, 162, 119, 5, 112, 162, 119, 5, 5, 164, 119, 5, 112, 164, 119, 5, 5, 166, 119, 5, 5, 168, 119, 5, 112, 168, 119, 5, 5, 170, 119, 5, 112, 170, 119, 5, 5, 172, 119, 5, 112, 172, 119, 5, 5, 174, 119, 5, 112, 174, 119, 5, 5, 176, 119, 5, 112, 176, 119, 5, 5, 178, 119, 5, 112, 178, 119, 193, 2, 180, 119, 193, 9, 180, 119, 193, 16, 180, 119, 193, 23, 180, 119, 5, 5, 182, 119, 5, 112, 182, 119, 5, 5, 184, 119, 5, 112, 184, 119, 5, 5, 186, 119, 5, 112, 186, 119, 5, 5, 188, 119, 5, 112, 188, 119, 5, 5, 190, 119, 5, 5, 192, 119, 5, 112, 192, 119, 193, 2, 194, 119, 193, 9, 194, 119, 193, 16, 194, 119, 5, 5, 196, 119, 5, 112, 196, 119, 5, 5, 198, 119, 5, 112, 198, 119, 5, 5, 200, 119, 5, 112, 200, 119, 5, 5, 202, 119, 5, 112, 202, 119, 5, 5, 204, 119, 5, 112, 204, 119, 193, 2, 206, 119, 193, 9, 206, 119, 193, 16, 206, 119, 5, 5, 208, 119, 5, 112, 208, 119, 5, 5, 210, 119, 5, 112, 210, 119, 5, 5, 212, 119, 5, 112, 212, 119, 5, 5, 214, 119, 5, 112, 214, 119, 193, 2, 216, 119, 193, 9, 216, 119, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 4, 120, 193, 16, 4, 120, 193, 23, 4, 120, 193, 30, 4, 120, 193, 37, 4, 120, 193, 44, 4, 120, 193, 51, 4, 120, 193, 58, 4, 120, 193, 65, 4, 120, 193, 72, 4, 120, 193, 79, 4, 120, 193, 86, 4, 120, 5, 5, 6, 120, 193, 2, 8, 120, 193, 9, 8, 120, 193, 16, 8, 120, 193, 23, 8, 120, 193, 30, 8, 120, 193, 37, 8, 120, 193, 44, 8, 120, 193, 51, 8, 120, 193, 58, 8, 120, 193, 65, 8, 120, 193, 72, 8, 120, 193, 79, 8, 120, 193, 86, 8, 120, 193, 93, 8, 120, 193, 100, 8, 120, 193, 107, 8, 120, 193, 114, 8, 120, 5, 5, 10, 120, 193, 2, 12, 120, 193, 9, 12, 120, 193, 16, 12, 120, 193, 23, 12, 120, 193, 30, 12, 120, 193, 37, 12, 120, 193, 44, 12, 120, 193, 51, 12, 120, 193, 58, 12, 120, 193, 65, 12, 120, 193, 72, 12, 120, 193, 79, 12, 120, 193, 86, 12, 120, 193, 93, 12, 120, 193, 100, 12, 120, 193, 107, 12, 120, 193, 114, 12, 120, 193, 121, 12, 120, 193, 128, 12, 120, 193, 135, 12, 120, 5, 5, 14, 120, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 206, 128, 232, 0, 5, 5, 18, 120, 193, 2, 20, 120, 193, 9, 20, 120, 193, 16, 20, 120, 193, 23, 20, 120, 193, 30, 20, 120, 193, 37, 20, 120, 193, 44, 20, 120, 193, 51, 20, 120, 193, 58, 20, 120, 193, 65, 20, 120, 5, 5, 22, 120, 193, 2, 24, 120, 193, 9, 24, 120, 193, 16, 24, 120, 193, 23, 24, 120, 193, 30, 24, 120, 5, 5, 26, 120, 5, 5, 28, 120, 5, 5, 30, 120, 193, 2, 32, 120, 193, 9, 32, 120, 193, 16, 32, 120, 193, 23, 32, 120, 193, 30, 32, 120, 5, 5, 34, 120, 193, 2, 36, 120, 193, 9, 36, 120, 193, 16, 36, 120, 193, 23, 36, 120, 193, 30, 36, 120, 193, 37, 36, 120, 193, 44, 36, 120, 193, 51, 36, 120, 193, 58, 36, 120, 5, 5, 38, 120, 193, 2, 40, 120, 193, 9, 40, 120, 193, 16, 40, 120, 5, 5, 42, 120, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 206, 160, 232, 0, 5, 5, 46, 120, 5, 5, 48, 120, 5, 5, 50, 120, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 206, 192, 232, 0, 5, 5, 55, 120, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 206, 224, 232, 0, 5, 5, 59, 120, 193, 2, 61, 120, 193, 9, 61, 120, 193, 16, 61, 120, 193, 23, 61, 120, 193, 30, 61, 120, 5, 5, 63, 120, 193, 2, 65, 120, 193, 9, 65, 120, 193, 16, 65, 120, 193, 23, 65, 120, 193, 30, 65, 120, 193, 37, 65, 120, 5, 5, 67, 120, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 0, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 32, 233, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 206, 96, 240, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 128, 240, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 206, 160, 242, 0, 5, 5, 103, 120, 5, 5, 105, 120, 5, 5, 107, 120, 5, 5, 109, 120, 5, 5, 111, 120, 5, 5, 113, 120, 5, 5, 115, 120, 5, 5, 117, 120, 5, 5, 119, 120, 5, 5, 121, 120, 5, 5, 123, 120, 5, 5, 125, 120, 5, 5, 127, 120, 5, 5, 129, 120, 5, 5, 131, 120, 5, 5, 133, 120, 5, 5, 135, 120, 5, 5, 137, 120, 5, 5, 139, 120, 5, 5, 141, 120, 5, 5, 143, 120, 5, 5, 145, 120, 5, 5, 147, 120, 5, 5, 149, 120, 5, 5, 151, 120, 5, 5, 153, 120, 5, 5, 155, 120, 5, 5, 157, 120, 5, 5, 159, 120, 5, 5, 161, 120, 5, 5, 163, 120, 5, 5, 165, 120, 160, 5, 103, 120, 160, 5, 105, 120, 160, 5, 107, 120, 160, 5, 109, 120, 160, 5, 111, 120, 160, 5, 113, 120, 160, 5, 115, 120, 160, 5, 117, 120, 160, 5, 119, 120, 160, 5, 121, 120, 160, 5, 123, 120, 160, 5, 125, 120, 160, 5, 127, 120, 160, 5, 129, 120, 160, 5, 131, 120, 160, 5, 133, 120, 160, 5, 135, 120, 160, 5, 137, 120, 160, 5, 139, 120, 160, 5, 141, 120, 160, 5, 143, 120, 160, 5, 145, 120, 160, 5, 147, 120, 160, 5, 149, 120, 160, 5, 151, 120, 160, 5, 153, 120, 160, 5, 155, 120, 160, 5, 157, 120, 160, 5, 159, 120, 160, 5, 161, 120, 160, 5, 163, 120, 160, 5, 165, 120, 160, 5, 233, 120, 160, 5, 235, 120, 5, 5, 169, 120, 5, 5, 171, 120, 5, 5, 173, 120, 5, 5, 175, 120, 5, 5, 177, 120, 5, 5, 179, 120, 5, 5, 181, 120, 5, 5, 183, 120, 5, 5, 185, 120, 5, 5, 187, 120, 5, 5, 189, 120, 5, 5, 191, 120, 5, 5, 193, 120, 5, 5, 195, 120, 5, 5, 197, 120, 5, 5, 199, 120, 5, 5, 201, 120, 5, 5, 203, 120, 5, 5, 205, 120, 5, 5, 207, 120, 5, 5, 209, 120, 5, 5, 211, 120, 5, 5, 213, 120, 5, 5, 215, 120, 5, 5, 217, 120, 5, 5, 219, 120, 5, 5, 221, 120, 5, 5, 223, 120, 5, 5, 225, 120, 5, 5, 227, 120, 160, 5, 169, 120, 160, 5, 171, 120, 160, 5, 173, 120, 160, 5, 175, 120, 160, 5, 177, 120, 160, 5, 179, 120, 160, 5, 181, 120, 160, 5, 183, 120, 160, 5, 185, 120, 160, 5, 187, 120, 160, 5, 189, 120, 160, 5, 191, 120, 160, 5, 193, 120, 160, 5, 195, 120, 160, 5, 197, 120, 160, 5, 199, 120, 160, 5, 201, 120, 160, 5, 203, 120, 160, 5, 205, 120, 160, 5, 207, 120, 160, 5, 209, 120, 160, 5, 211, 120, 160, 5, 213, 120, 160, 5, 215, 120, 160, 5, 217, 120, 160, 5, 219, 120, 160, 5, 221, 120, 160, 5, 223, 120, 160, 5, 225, 120, 160, 5, 227, 120, 160, 5, 229, 120, 160, 5, 231, 120, 5, 5, 6, 121, 5, 5, 8, 121, 5, 5, 10, 121, 5, 5, 12, 121, 5, 5, 14, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 5, 5, 22, 121, 5, 5, 24, 121, 5, 5, 26, 121, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 32, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 44, 121, 5, 5, 46, 121, 5, 5, 48, 121, 193, 2, 50, 121, 193, 9, 50, 121, 193, 16, 50, 121, 193, 23, 50, 121, 5, 5, 52, 121, 5, 5, 54, 121, 5, 5, 56, 121, 5, 5, 58, 121, 5, 5, 60, 121, 5, 5, 62, 121, 5, 5, 86, 123, 5, 5, 88, 123, 5, 5, 90, 123, 5, 5, 92, 123, 5, 5, 94, 123, 5, 5, 98, 123, 5, 5, 104, 123, 5, 5, 106, 123, 5, 5, 108, 123, 5, 5, 110, 123, 5, 5, 16, 123, 5, 5, 32, 123, 5, 5, 44, 123, 5, 5, 112, 123, 32, 5, 76, 123, 5, 5, 24, 123, 255, 255, 255, 255, 16, 5, 6, 121, 32, 5, 8, 121, 32, 5, 183, 121, 16, 5, 10, 121, 32, 5, 187, 121, 32, 5, 189, 121, 16, 5, 12, 121, 32, 5, 14, 121, 16, 5, 16, 121, 32, 5, 195, 121, 32, 5, 197, 121, 32, 5, 199, 121, 32, 5, 201, 121, 32, 5, 203, 121, 32, 5, 205, 121, 198, 1, 171, 0, 172, 5, 6, 121, 174, 5, 8, 121, 174, 5, 183, 121, 172, 5, 10, 121, 174, 5, 187, 121, 174, 5, 189, 121, 172, 5, 12, 121, 174, 5, 14, 121, 172, 5, 16, 121, 174, 5, 195, 121, 174, 5, 197, 121, 174, 5, 199, 121, 174, 5, 201, 121, 174, 5, 203, 121, 174, 5, 205, 121, 174, 5, 52, 121, 172, 5, 18, 121, 172, 5, 20, 121, 174, 5, 22, 121, 174, 5, 66, 121, 172, 5, 24, 121, 174, 5, 26, 121, 172, 5, 28, 121, 172, 5, 30, 121, 174, 5, 32, 121, 172, 5, 34, 121, 172, 5, 36, 121, 172, 5, 38, 121, 172, 5, 40, 121, 172, 5, 42, 121, 255, 255, 255, 255, 32, 5, 52, 121, 16, 5, 18, 121, 16, 5, 20, 121, 32, 5, 22, 121, 32, 5, 66, 121, 16, 5, 24, 121, 32, 5, 26, 121, 16, 5, 28, 121, 16, 5, 30, 121, 32, 5, 32, 121, 16, 5, 34, 121, 16, 5, 36, 121, 16, 5, 38, 121, 16, 5, 40, 121, 16, 5, 42, 121, 16, 5, 103, 121, 32, 5, 105, 121, 32, 5, 107, 121, 32, 5, 109, 121, 32, 5, 111, 121, 32, 5, 113, 121, 32, 5, 115, 121, 32, 5, 117, 121, 16, 5, 119, 121, 32, 5, 121, 121, 32, 5, 123, 121, 32, 5, 125, 121, 32, 5, 127, 121, 16, 5, 129, 121, 32, 5, 131, 121, 32, 5, 133, 121, 32, 5, 135, 121, 32, 5, 137, 121, 32, 5, 139, 121, 16, 5, 141, 121, 32, 5, 143, 121, 198, 225, 170, 0, 32, 5, 46, 121, 32, 5, 48, 121, 32, 5, 235, 121, 32, 5, 237, 121, 32, 5, 241, 121, 32, 5, 245, 121, 198, 33, 171, 0, 198, 65, 171, 0, 198, 97, 171, 0, 32, 5, 56, 121, 198, 129, 171, 0, 198, 161, 171, 0, 32, 5, 58, 121, 32, 5, 60, 121, 32, 5, 64, 121, 32, 5, 68, 121, 32, 5, 70, 121, 32, 5, 74, 121, 32, 5, 78, 121, 32, 5, 82, 121, 32, 5, 84, 121, 32, 5, 86, 121, 32, 5, 88, 121, 32, 5, 90, 121, 198, 65, 170, 0, 198, 97, 170, 0, 198, 129, 170, 0, 5, 5, 64, 121, 5, 5, 66, 121, 5, 5, 68, 121, 5, 5, 70, 121, 193, 2, 72, 121, 193, 9, 72, 121, 193, 16, 72, 121, 5, 5, 74, 121, 5, 5, 76, 121, 5, 5, 78, 121, 5, 5, 80, 121, 5, 5, 82, 121, 5, 5, 84, 121, 5, 5, 86, 121, 5, 5, 88, 121, 5, 5, 90, 121, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 206, 64, 227, 0, 5, 5, 95, 121, 5, 5, 97, 121, 5, 5, 99, 121, 193, 2, 101, 121, 193, 9, 101, 121, 193, 16, 101, 121, 193, 23, 101, 121, 193, 30, 101, 121, 193, 240, 101, 121, 193, 37, 101, 121, 193, 44, 101, 121, 193, 51, 101, 121, 193, 58, 101, 121, 193, 65, 101, 121, 193, 72, 101, 121, 193, 79, 101, 121, 193, 86, 101, 121, 193, 93, 101, 121, 193, 100, 101, 121, 193, 107, 101, 121, 193, 114, 101, 121, 193, 121, 101, 121, 193, 128, 101, 121, 193, 135, 101, 121, 193, 142, 101, 121, 193, 149, 101, 121, 193, 156, 101, 121, 193, 163, 101, 121, 193, 170, 101, 121, 193, 177, 101, 121, 193, 184, 101, 121, 193, 191, 101, 121, 193, 198, 101, 121, 193, 205, 101, 121, 193, 212, 101, 121, 193, 219, 101, 121, 193, 226, 101, 121, 193, 233, 101, 121, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 247, 101, 121, 5, 5, 103, 121, 5, 5, 105, 121, 5, 5, 107, 121, 5, 5, 109, 121, 5, 5, 111, 121, 5, 5, 113, 121, 5, 5, 115, 121, 5, 5, 117, 121, 5, 5, 119, 121, 5, 5, 121, 121, 5, 5, 123, 121, 5, 5, 125, 121, 5, 5, 127, 121, 5, 5, 129, 121, 5, 5, 131, 121, 5, 5, 133, 121, 5, 5, 135, 121, 5, 5, 137, 121, 5, 5, 139, 121, 5, 5, 141, 121, 5, 5, 143, 121, 206, 96, 227, 0, 206, 96, 227, 0, 206, 96, 227, 0, 206, 96, 227, 0, 206, 96, 227, 0, 206, 96, 227, 0, 206, 96, 227, 0, 206, 96, 227, 0, 206, 96, 227, 0, 206, 96, 227, 0, 255, 255, 255, 255, 255, 255, 255, 255, 172, 5, 103, 121, 174, 5, 105, 121, 174, 5, 107, 121, 174, 5, 109, 121, 174, 5, 111, 121, 174, 5, 113, 121, 255, 255, 255, 255, 255, 255, 255, 255, 174, 5, 115, 121, 174, 5, 117, 121, 172, 5, 119, 121, 174, 5, 121, 121, 174, 5, 123, 121, 174, 5, 125, 121, 255, 255, 255, 255, 255, 255, 255, 255, 174, 5, 127, 121, 172, 5, 129, 121, 174, 5, 131, 121, 174, 5, 133, 121, 174, 5, 135, 121, 174, 5, 137, 121, 255, 255, 255, 255, 255, 255, 255, 255, 174, 5, 139, 121, 172, 5, 141, 121, 174, 5, 143, 121, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 96, 227, 0, 206, 96, 227, 0, 206, 96, 227, 0, 206, 96, 227, 0, 5, 5, 147, 121, 5, 5, 149, 121, 5, 5, 151, 121, 5, 5, 153, 121, 5, 5, 155, 121, 193, 2, 157, 121, 193, 9, 157, 121, 193, 16, 157, 121, 193, 23, 157, 121, 193, 30, 157, 121, 193, 37, 157, 121, 193, 44, 157, 121, 193, 51, 157, 121, 5, 5, 159, 121, 5, 5, 161, 121, 5, 5, 163, 121, 5, 5, 165, 121, 193, 2, 167, 121, 193, 9, 167, 121, 193, 16, 167, 121, 193, 23, 167, 121, 193, 30, 167, 121, 193, 37, 167, 121, 193, 44, 167, 121, 193, 51, 167, 121, 193, 58, 167, 121, 5, 5, 169, 121, 5, 5, 171, 121, 5, 5, 173, 121, 5, 5, 175, 121, 193, 2, 177, 121, 193, 9, 177, 121, 193, 16, 177, 121, 193, 23, 177, 121, 193, 30, 177, 121, 193, 37, 177, 121, 5, 5, 179, 121, 5, 5, 181, 121, 5, 5, 183, 121, 5, 5, 185, 121, 5, 5, 187, 121, 5, 5, 189, 121, 5, 5, 191, 121, 5, 5, 193, 121, 5, 5, 195, 121, 5, 5, 197, 121, 5, 5, 199, 121, 5, 5, 201, 121, 5, 5, 203, 121, 5, 5, 205, 121, 5, 5, 207, 121, 5, 5, 209, 121, 5, 5, 211, 121, 5, 5, 213, 121, 5, 5, 215, 121, 5, 5, 217, 121, 5, 5, 219, 121, 5, 5, 221, 121, 5, 5, 223, 121, 5, 5, 225, 121, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 206, 32, 234, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 5, 5, 227, 121, 5, 5, 229, 121, 5, 5, 231, 121, 193, 2, 233, 121, 193, 9, 233, 121, 193, 16, 233, 121, 193, 23, 233, 121, 5, 5, 235, 121, 5, 5, 237, 121, 193, 2, 239, 121, 193, 9, 239, 121, 193, 16, 239, 121, 5, 5, 241, 121, 5, 5, 243, 121, 5, 5, 245, 121, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 128, 227, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 206, 64, 234, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 3, 5, 6, 122, 133, 5, 6, 122, 3, 5, 8, 122, 133, 5, 8, 122, 3, 5, 10, 122, 133, 5, 10, 122, 3, 5, 14, 122, 133, 5, 14, 122, 3, 5, 16, 122, 133, 5, 16, 122, 133, 5, 18, 122, 197, 98, 185, 2, 133, 5, 20, 122, 197, 66, 188, 2, 133, 5, 22, 122, 197, 98, 193, 2, 133, 5, 24, 122, 197, 226, 195, 2, 133, 5, 26, 122, 197, 194, 196, 2, 133, 5, 28, 122, 197, 162, 198, 2, 133, 5, 30, 122, 197, 66, 200, 2, 133, 5, 32, 122, 197, 162, 201, 2, 133, 5, 34, 122, 197, 34, 202, 2, 133, 5, 36, 122, 197, 98, 203, 2, 133, 5, 38, 122, 197, 197, 179, 2, 197, 100, 180, 2, 197, 229, 180, 2, 197, 99, 179, 2, 197, 133, 181, 2, 197, 35, 182, 2, 197, 3, 183, 2, 197, 230, 183, 2, 197, 100, 183, 2, 197, 3, 185, 2, 197, 163, 184, 2, 197, 227, 185, 2, 197, 68, 186, 2, 197, 196, 186, 2, 197, 68, 187, 2, 197, 196, 187, 2, 197, 196, 188, 2, 197, 68, 189, 2, 197, 196, 189, 2, 197, 70, 190, 2, 197, 2, 191, 2, 197, 70, 191, 2, 197, 6, 192, 2, 197, 197, 192, 2, 197, 132, 191, 2, 197, 230, 193, 2, 197, 166, 194, 2, 197, 100, 195, 2, 197, 99, 196, 2, 197, 67, 198, 2, 197, 68, 197, 2, 197, 36, 199, 2, 133, 5, 12, 122, 193, 93, 104, 122, 193, 2, 104, 122, 193, 9, 104, 122, 193, 16, 104, 122, 193, 23, 104, 122, 193, 30, 104, 122, 193, 37, 104, 122, 193, 44, 104, 122, 193, 51, 104, 122, 193, 58, 104, 122, 193, 65, 104, 122, 193, 72, 104, 122, 193, 79, 104, 122, 193, 86, 104, 122, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 197, 2, 225, 2, 197, 194, 197, 2, 49, 5, 28, 122, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 198, 193, 191, 0, 198, 33, 186, 0, 198, 97, 182, 0, 197, 98, 206, 2, 198, 1, 176, 0, 198, 225, 184, 0, 198, 129, 212, 0, 198, 161, 185, 0, 198, 97, 176, 0, 198, 129, 194, 0, 198, 1, 199, 0, 198, 65, 193, 0, 198, 161, 179, 0, 198, 193, 190, 0, 198, 65, 178, 0, 198, 129, 193, 0, 197, 165, 199, 2, 197, 5, 201, 2, 197, 163, 202, 2, 197, 3, 203, 2, 197, 100, 204, 2, 197, 163, 206, 2, 197, 131, 207, 2, 197, 98, 194, 2, 197, 226, 207, 2, 197, 35, 208, 2, 197, 99, 211, 2, 197, 134, 209, 2, 197, 68, 210, 2, 197, 197, 210, 2, 197, 198, 212, 2, 197, 132, 213, 2, 197, 3, 214, 2, 197, 99, 214, 2, 197, 198, 215, 2, 197, 132, 216, 2, 197, 6, 217, 2, 197, 195, 217, 2, 197, 69, 220, 2, 197, 227, 220, 2, 197, 68, 221, 2, 197, 195, 221, 2, 197, 36, 222, 2, 197, 37, 219, 2, 197, 196, 219, 2, 197, 101, 224, 2, 197, 68, 225, 2, 197, 194, 225, 2, 197, 226, 203, 2, 133, 5, 40, 122, 197, 226, 204, 2, 3, 5, 42, 122, 133, 5, 42, 122, 197, 98, 205, 2, 133, 5, 44, 122, 197, 226, 205, 2, 133, 5, 46, 122, 197, 2, 207, 2, 133, 5, 48, 122, 133, 5, 50, 122, 133, 5, 52, 122, 133, 5, 54, 122, 133, 5, 56, 122, 133, 5, 58, 122, 197, 130, 208, 2, 197, 2, 209, 2, 133, 5, 60, 122, 197, 194, 211, 2, 197, 66, 212, 2, 133, 5, 62, 122, 197, 194, 214, 2, 197, 66, 215, 2, 133, 5, 64, 122, 197, 34, 218, 2, 197, 162, 218, 2, 133, 5, 66, 122, 197, 162, 222, 2, 197, 34, 223, 2, 133, 5, 68, 122, 133, 5, 70, 122, 197, 34, 204, 2, 144, 5, 40, 122, 197, 34, 205, 2, 16, 5, 42, 122, 158, 5, 42, 122, 197, 162, 205, 2, 144, 5, 44, 122, 197, 34, 206, 2, 151, 5, 46, 122, 197, 66, 207, 2, 144, 5, 48, 122, 144, 5, 50, 122, 155, 5, 52, 122, 144, 5, 54, 122, 144, 5, 56, 122, 152, 5, 58, 122, 197, 194, 208, 2, 197, 66, 209, 2, 152, 5, 60, 122, 197, 2, 212, 2, 197, 130, 212, 2, 152, 5, 62, 122, 197, 2, 215, 2, 197, 130, 215, 2, 152, 5, 64, 122, 197, 98, 218, 2, 197, 226, 218, 2, 152, 5, 66, 122, 197, 226, 222, 2, 197, 98, 223, 2, 144, 5, 68, 122, 144, 5, 70, 122, 38, 5, 40, 122, 44, 5, 42, 122, 38, 5, 44, 122, 37, 5, 46, 122, 38, 5, 48, 122, 38, 5, 50, 122, 49, 5, 52, 122, 38, 5, 54, 122, 38, 5, 56, 122, 40, 5, 58, 122, 40, 5, 60, 122, 40, 5, 62, 122, 40, 5, 64, 122, 40, 5, 66, 122, 38, 5, 68, 122, 38, 5, 70, 122, 40, 5, 72, 122, 38, 5, 74, 122, 44, 5, 76, 122, 44, 5, 78, 122, 44, 5, 80, 122, 40, 5, 82, 122, 40, 5, 84, 122, 37, 5, 86, 122, 40, 5, 88, 122, 40, 5, 90, 122, 40, 5, 92, 122, 40, 5, 94, 122, 44, 5, 96, 122, 44, 5, 98, 122, 49, 5, 100, 122, 198, 226, 176, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 3, 5, 96, 122, 3, 5, 98, 122, 3, 5, 100, 122, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 16, 5, 96, 122, 16, 5, 98, 122, 16, 5, 100, 122, 16, 5, 102, 122, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 206, 32, 241, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 5, 5, 6, 123, 5, 5, 8, 123, 5, 5, 10, 123, 5, 5, 14, 123, 5, 5, 18, 123, 5, 5, 20, 123, 5, 5, 22, 123, 5, 5, 26, 123, 5, 5, 28, 123, 5, 5, 30, 123, 5, 5, 36, 123, 5, 5, 38, 123, 5, 5, 40, 123, 5, 5, 42, 123, 5, 5, 50, 123, 5, 5, 52, 123, 5, 5, 54, 123, 5, 5, 56, 123, 5, 5, 58, 123, 5, 5, 60, 123, 5, 5, 62, 123, 5, 5, 70, 123, 5, 5, 72, 123, 5, 5, 76, 123, 5, 5, 78, 123, 5, 5, 82, 123, 5, 5, 84, 123, 5, 112, 6, 123, 5, 112, 58, 123, 5, 112, 38, 123, 5, 112, 28, 123, 5, 5, 80, 123, 5, 112, 80, 123, 5, 5, 74, 123, 5, 112, 72, 123, 5, 120, 108, 123, 5, 112, 70, 123, 5, 112, 106, 123, 5, 112, 108, 123, 5, 5, 12, 123, 5, 5, 34, 123, 5, 112, 82, 123, 5, 112, 86, 123, 5, 5, 100, 123, 5, 5, 102, 123, 5, 5, 96, 123, 32, 112, 106, 123, 32, 5, 8, 123, 32, 5, 20, 123, 32, 5, 30, 123, 32, 5, 36, 123, 5, 5, 64, 123, 5, 5, 66, 123, 5, 5, 68, 123, 32, 5, 28, 123, 5, 5, 46, 123, 5, 5, 48, 123, 5, 5, 114, 123, 5, 5, 116, 123, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 206, 32, 232, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 7, 152, 123, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 153, 123, 193, 16, 153, 123, 193, 23, 153, 123, 193, 30, 153, 123, 5, 5, 155, 123, 193, 2, 157, 123, 32, 5, 155, 123, 193, 9, 157, 123, 193, 16, 157, 123, 193, 23, 157, 123, 193, 30, 157, 123, 193, 37, 157, 123, 193, 44, 157, 123, 193, 51, 157, 123, 193, 65, 157, 123, 193, 72, 157, 123, 5, 5, 159, 123, 193, 2, 161, 123, 193, 9, 161, 123, 32, 5, 159, 123, 193, 16, 161, 123, 193, 23, 161, 123, 193, 30, 161, 123, 193, 37, 161, 123, 193, 44, 161, 123, 193, 51, 161, 123, 193, 58, 161, 123, 193, 65, 161, 123, 193, 72, 161, 123, 193, 79, 161, 123, 193, 86, 161, 123, 193, 93, 161, 123, 193, 9, 169, 123, 193, 16, 169, 123, 193, 23, 169, 123, 193, 30, 169, 123, 193, 37, 169, 123, 193, 37, 153, 123, 193, 114, 165, 123, 193, 100, 165, 123, 193, 58, 157, 123, 193, 163, 165, 123, 193, 79, 157, 123, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 72, 169, 123, 193, 44, 169, 123, 193, 51, 169, 123, 193, 58, 169, 123, 193, 65, 169, 123, 193, 79, 169, 123, 193, 86, 169, 123, 193, 93, 169, 123, 193, 100, 169, 123, 193, 107, 169, 123, 193, 121, 169, 123, 193, 128, 169, 123, 193, 135, 169, 123, 193, 142, 169, 123, 193, 156, 169, 123, 193, 163, 169, 123, 193, 170, 169, 123, 193, 100, 161, 123, 193, 107, 161, 123, 193, 114, 161, 123, 5, 5, 163, 123, 193, 2, 165, 123, 32, 5, 163, 123, 193, 9, 165, 123, 193, 16, 165, 123, 193, 23, 165, 123, 193, 30, 165, 123, 193, 37, 165, 123, 193, 44, 165, 123, 193, 51, 165, 123, 193, 58, 165, 123, 193, 65, 165, 123, 193, 72, 165, 123, 193, 79, 165, 123, 193, 86, 165, 123, 193, 93, 165, 123, 193, 107, 165, 123, 193, 121, 165, 123, 193, 128, 165, 123, 193, 135, 165, 123, 193, 142, 165, 123, 193, 149, 165, 123, 193, 156, 165, 123, 193, 170, 165, 123, 193, 177, 165, 123, 193, 184, 165, 123, 5, 5, 167, 123, 193, 2, 169, 123, 32, 5, 167, 123, 193, 84, 170, 123, 193, 114, 169, 123, 193, 112, 170, 123, 193, 149, 169, 123, 193, 184, 169, 123, 193, 105, 170, 123, 193, 226, 169, 123, 193, 42, 170, 123, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 189, 170, 123, 193, 196, 170, 123, 193, 203, 170, 123, 193, 210, 170, 123, 193, 217, 170, 123, 193, 224, 170, 123, 193, 231, 170, 123, 193, 238, 170, 123, 193, 245, 170, 123, 193, 252, 170, 123, 193, 5, 171, 123, 193, 12, 171, 123, 193, 19, 171, 123, 193, 26, 171, 123, 193, 33, 171, 123, 193, 40, 171, 123, 193, 47, 171, 123, 193, 177, 169, 123, 193, 191, 169, 123, 193, 198, 169, 123, 193, 205, 169, 123, 193, 212, 169, 123, 193, 219, 169, 123, 193, 233, 169, 123, 193, 240, 169, 123, 193, 247, 169, 123, 193, 254, 169, 123, 193, 7, 170, 123, 193, 14, 170, 123, 193, 21, 170, 123, 193, 28, 170, 123, 193, 49, 170, 123, 193, 56, 170, 123, 193, 63, 170, 123, 193, 70, 170, 123, 193, 77, 170, 123, 193, 91, 170, 123, 193, 98, 170, 123, 193, 119, 170, 123, 193, 126, 170, 123, 193, 133, 170, 123, 193, 140, 170, 123, 193, 147, 170, 123, 193, 154, 170, 123, 193, 161, 170, 123, 193, 168, 170, 123, 193, 175, 170, 123, 193, 182, 170, 123, 193, 35, 170, 123, 5, 5, 174, 123, 5, 5, 176, 123, 5, 5, 178, 123, 5, 5, 180, 123, 5, 5, 182, 123, 5, 5, 184, 123, 5, 5, 186, 123, 5, 5, 188, 123, 5, 5, 190, 123, 5, 5, 192, 123, 5, 5, 194, 123, 5, 5, 196, 123, 5, 5, 198, 123, 5, 5, 200, 123, 5, 5, 202, 123, 5, 5, 204, 123, 5, 5, 206, 123, 5, 5, 208, 123, 5, 5, 210, 123, 5, 5, 212, 123, 5, 5, 214, 123, 5, 5, 216, 123, 5, 5, 218, 123, 5, 5, 220, 123, 5, 5, 222, 123, 5, 5, 224, 123, 5, 5, 226, 123, 5, 5, 228, 123, 5, 5, 230, 123, 5, 5, 232, 123, 5, 5, 234, 123, 5, 5, 236, 123, 160, 5, 174, 123, 160, 5, 176, 123, 160, 5, 178, 123, 160, 5, 180, 123, 160, 5, 182, 123, 160, 5, 184, 123, 160, 5, 186, 123, 160, 5, 188, 123, 160, 5, 190, 123, 160, 5, 192, 123, 160, 5, 194, 123, 160, 5, 196, 123, 160, 5, 198, 123, 160, 5, 200, 123, 160, 5, 202, 123, 160, 5, 204, 123, 160, 5, 206, 123, 160, 5, 208, 123, 160, 5, 210, 123, 160, 5, 212, 123, 160, 5, 214, 123, 160, 5, 216, 123, 160, 5, 218, 123, 160, 5, 220, 123, 160, 5, 222, 123, 160, 5, 224, 123, 160, 5, 226, 123, 160, 5, 228, 123, 160, 5, 230, 123, 160, 5, 232, 123, 160, 5, 234, 123, 160, 5, 236, 123, 193, 58, 238, 123, 193, 65, 238, 123, 193, 72, 238, 123, 193, 79, 238, 123, 193, 86, 238, 123, 193, 93, 238, 123, 193, 100, 238, 123, 193, 107, 238, 123, 193, 114, 238, 123, 193, 121, 238, 123, 193, 128, 238, 123, 193, 135, 238, 123, 193, 142, 238, 123, 193, 149, 238, 123, 193, 156, 238, 123, 193, 163, 238, 123, 193, 170, 238, 123, 193, 177, 238, 123, 193, 184, 238, 123, 193, 191, 238, 123, 193, 198, 238, 123, 193, 9, 238, 123, 193, 16, 238, 123, 193, 23, 238, 123, 193, 30, 238, 123, 193, 37, 238, 123, 193, 44, 238, 123, 193, 51, 238, 123, 193, 205, 238, 123, 193, 212, 238, 123, 193, 254, 238, 123, 193, 219, 238, 123, 193, 226, 238, 123, 193, 233, 238, 123, 193, 240, 238, 123, 193, 247, 238, 123, 193, 7, 239, 123, 193, 49, 239, 123, 193, 35, 239, 123, 193, 133, 239, 123, 193, 56, 239, 123, 193, 42, 239, 123, 193, 140, 239, 123, 193, 105, 239, 123, 193, 91, 239, 123, 193, 112, 239, 123, 193, 98, 239, 123, 193, 14, 239, 123, 193, 147, 239, 123, 193, 28, 239, 123, 193, 21, 239, 123, 193, 77, 239, 123, 193, 63, 239, 123, 193, 119, 239, 123, 193, 84, 239, 123, 193, 70, 239, 123, 193, 126, 239, 123, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 206, 160, 240, 0, 193, 19, 242, 123, 193, 26, 242, 123, 193, 33, 242, 123, 193, 40, 242, 123, 193, 47, 242, 123, 193, 54, 242, 123, 193, 61, 242, 123, 193, 68, 242, 123, 193, 75, 242, 123, 193, 82, 242, 123, 193, 89, 242, 123, 193, 96, 242, 123, 193, 103, 242, 123, 193, 110, 242, 123, 193, 117, 242, 123, 193, 124, 242, 123, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 96, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 206, 128, 242, 0, 193, 9, 247, 123, 193, 16, 247, 123, 193, 23, 247, 123, 193, 30, 247, 123, 193, 37, 247, 123, 193, 44, 247, 123, 193, 51, 247, 123, 193, 58, 247, 123, 193, 65, 247, 123, 193, 72, 247, 123, 193, 79, 247, 123, 193, 86, 247, 123, 193, 93, 247, 123, 193, 100, 247, 123, 193, 107, 247, 123, 193, 114, 247, 123, 193, 121, 247, 123, 193, 128, 247, 123, 193, 135, 247, 123, 193, 142, 247, 123, 193, 149, 247, 123, 193, 156, 247, 123, 193, 163, 247, 123, 193, 170, 247, 123, 193, 177, 247, 123, 193, 184, 247, 123, 193, 191, 247, 123, 193, 198, 247, 123, 193, 205, 247, 123, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 206, 224, 234, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 251, 123, 193, 16, 251, 123, 193, 23, 251, 123, 193, 30, 251, 123, 193, 37, 251, 123, 193, 44, 251, 123, 193, 51, 251, 123, 193, 58, 251, 123, 193, 65, 251, 123, 193, 72, 251, 123, 193, 79, 251, 123, 193, 86, 251, 123, 193, 93, 251, 123, 193, 100, 251, 123, 193, 107, 251, 123, 193, 121, 251, 123, 193, 128, 251, 123, 193, 135, 251, 123, 193, 142, 251, 123, 193, 149, 251, 123, 193, 156, 251, 123, 193, 163, 251, 123, 193, 170, 251, 123, 193, 177, 251, 123, 193, 184, 251, 123, 193, 191, 251, 123, 193, 198, 251, 123, 193, 205, 251, 123, 193, 212, 251, 123, 193, 219, 251, 123, 193, 226, 251, 123, 193, 114, 251, 123, 160, 5, 70, 124, 160, 5, 72, 124, 160, 5, 74, 124, 160, 5, 76, 124, 160, 5, 78, 124, 160, 5, 80, 124, 160, 5, 82, 124, 160, 5, 84, 124, 5, 5, 6, 124, 5, 5, 8, 124, 5, 5, 10, 124, 5, 5, 12, 124, 5, 5, 14, 124, 5, 5, 16, 124, 5, 5, 18, 124, 5, 5, 20, 124, 5, 5, 22, 124, 5, 5, 24, 124, 5, 5, 26, 124, 5, 5, 28, 124, 5, 5, 30, 124, 5, 5, 32, 124, 5, 5, 34, 124, 5, 5, 36, 124, 5, 5, 38, 124, 5, 5, 40, 124, 5, 5, 42, 124, 5, 5, 44, 124, 5, 5, 46, 124, 5, 5, 48, 124, 5, 5, 50, 124, 5, 5, 52, 124, 160, 5, 6, 124, 160, 5, 8, 124, 160, 5, 10, 124, 160, 5, 12, 124, 160, 5, 14, 124, 160, 5, 16, 124, 160, 5, 18, 124, 160, 5, 20, 124, 160, 5, 22, 124, 160, 5, 24, 124, 160, 5, 26, 124, 160, 5, 28, 124, 160, 5, 30, 124, 160, 5, 32, 124, 160, 5, 34, 124, 160, 5, 36, 124, 160, 5, 38, 124, 160, 5, 40, 124, 160, 5, 42, 124, 160, 5, 44, 124, 160, 5, 46, 124, 160, 5, 48, 124, 160, 5, 50, 124, 160, 5, 52, 124, 160, 5, 54, 124, 160, 5, 56, 124, 160, 5, 58, 124, 160, 5, 60, 124, 160, 5, 62, 124, 160, 5, 64, 124, 160, 5, 66, 124, 160, 5, 68, 124, 5, 5, 54, 124, 5, 5, 56, 124, 5, 5, 58, 124, 5, 5, 60, 124, 5, 5, 62, 124, 5, 5, 64, 124, 5, 5, 66, 124, 5, 5, 68, 124, 5, 5, 70, 124, 5, 5, 72, 124, 5, 5, 74, 124, 5, 5, 76, 124, 5, 5, 78, 124, 5, 5, 80, 124, 5, 5, 82, 124, 5, 5, 84, 124, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 64, 235, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 206, 96, 241, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 250, 90, 124, 193, 3, 91, 124, 193, 10, 91, 124, 193, 17, 91, 124, 193, 24, 91, 124, 193, 31, 91, 124, 193, 38, 91, 124, 193, 45, 91, 124, 193, 52, 91, 124, 193, 59, 91, 124, 193, 66, 91, 124, 193, 73, 91, 124, 193, 80, 91, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 92, 124, 193, 16, 92, 124, 193, 23, 92, 124, 193, 30, 92, 124, 193, 37, 92, 124, 193, 44, 92, 124, 193, 51, 92, 124, 193, 58, 92, 124, 193, 65, 92, 124, 193, 72, 92, 124, 193, 79, 92, 124, 193, 86, 92, 124, 193, 93, 92, 124, 193, 100, 92, 124, 193, 107, 92, 124, 193, 114, 92, 124, 193, 121, 92, 124, 193, 128, 92, 124, 193, 135, 92, 124, 193, 142, 92, 124, 193, 149, 92, 124, 193, 156, 92, 124, 193, 163, 92, 124, 193, 170, 92, 124, 193, 177, 92, 124, 193, 184, 92, 124, 193, 191, 92, 124, 193, 198, 92, 124, 193, 205, 92, 124, 193, 212, 92, 124, 255, 255, 255, 255, 255, 255, 255, 255, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 206, 128, 235, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 206, 160, 235, 0, 193, 9, 98, 124, 193, 16, 98, 124, 193, 23, 98, 124, 193, 30, 98, 124, 193, 37, 98, 124, 193, 44, 98, 124, 193, 51, 98, 124, 193, 58, 98, 124, 193, 65, 98, 124, 193, 72, 98, 124, 193, 79, 98, 124, 193, 86, 98, 124, 193, 93, 98, 124, 193, 100, 98, 124, 193, 107, 98, 124, 193, 114, 98, 124, 193, 121, 98, 124, 193, 128, 98, 124, 193, 135, 98, 124, 193, 142, 98, 124, 193, 149, 98, 124, 193, 156, 98, 124, 193, 163, 98, 124, 193, 170, 98, 124, 193, 177, 98, 124, 193, 184, 98, 124, 193, 191, 98, 124, 193, 198, 98, 124, 193, 205, 98, 124, 193, 212, 98, 124, 193, 219, 98, 124, 255, 255, 255, 255, 193, 9, 99, 124, 193, 16, 99, 124, 193, 23, 99, 124, 193, 30, 99, 124, 193, 37, 99, 124, 193, 44, 99, 124, 193, 51, 99, 124, 193, 58, 99, 124, 193, 65, 99, 124, 193, 72, 99, 124, 193, 79, 99, 124, 193, 86, 99, 124, 255, 255, 255, 255, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 206, 96, 234, 0, 255, 255, 255, 255, 193, 21, 100, 124, 193, 28, 100, 124, 193, 35, 100, 124, 193, 42, 100, 124, 193, 49, 100, 124, 193, 56, 100, 124, 193, 63, 100, 124, 193, 70, 100, 124, 193, 77, 100, 124, 193, 84, 100, 124, 193, 91, 100, 124, 193, 98, 100, 124, 193, 105, 100, 124, 193, 112, 100, 124, 193, 119, 100, 124, 193, 126, 100, 124, 193, 133, 100, 124, 193, 140, 100, 124, 193, 147, 100, 124, 255, 255, 255, 255, 193, 154, 100, 124, 193, 161, 100, 124, 255, 255, 255, 255, 193, 168, 100, 124, 193, 175, 100, 124, 193, 182, 100, 124, 193, 189, 100, 124, 193, 196, 100, 124, 193, 203, 100, 124, 193, 210, 100, 124, 193, 217, 100, 124, 193, 224, 100, 124, 193, 231, 100, 124, 193, 238, 100, 124, 193, 245, 100, 124, 193, 252, 100, 124, 193, 5, 101, 124, 193, 12, 101, 124, 255, 255, 255, 255, 255, 255, 255, 255, 193, 19, 101, 124, 193, 26, 101, 124, 193, 33, 101, 124, 193, 40, 101, 124, 193, 47, 101, 124, 193, 54, 101, 124, 193, 61, 101, 124, 193, 68, 101, 124, 193, 75, 101, 124, 193, 82, 101, 124, 193, 89, 101, 124, 193, 96, 101, 124, 193, 103, 101, 124, 193, 110, 101, 124, 255, 255, 255, 255, 255, 255, 255, 255, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 206, 128, 234, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 206, 192, 235, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 154, 113, 124, 193, 161, 113, 124, 193, 168, 113, 124, 193, 175, 113, 124, 193, 182, 113, 124, 193, 189, 113, 124, 193, 196, 113, 124, 193, 203, 113, 124, 193, 210, 113, 124, 193, 217, 113, 124, 193, 224, 113, 124, 193, 231, 113, 124, 193, 238, 113, 124, 193, 245, 113, 124, 193, 252, 113, 124, 193, 5, 114, 124, 193, 12, 114, 124, 193, 19, 114, 124, 193, 26, 114, 124, 193, 33, 114, 124, 193, 40, 114, 124, 193, 47, 114, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 54, 114, 124, 193, 61, 114, 124, 193, 68, 114, 124, 193, 75, 114, 124, 193, 82, 114, 124, 193, 89, 114, 124, 193, 96, 114, 124, 193, 103, 114, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 115, 124, 193, 16, 115, 124, 193, 23, 115, 124, 193, 30, 115, 124, 193, 37, 115, 124, 193, 44, 115, 124, 255, 255, 255, 255, 255, 255, 255, 255, 193, 51, 115, 124, 255, 255, 255, 255, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 206, 224, 235, 0, 255, 255, 255, 255, 193, 112, 116, 124, 193, 119, 116, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 126, 116, 124, 255, 255, 255, 255, 255, 255, 255, 255, 193, 133, 116, 124, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 206, 64, 236, 0, 32, 5, 132, 124, 5, 5, 132, 124, 32, 5, 134, 124, 5, 5, 134, 124, 193, 2, 136, 124, 193, 9, 136, 124, 32, 5, 138, 124, 5, 5, 138, 124, 193, 2, 140, 124, 193, 9, 140, 124, 193, 16, 140, 124, 193, 23, 140, 124, 32, 5, 142, 124, 5, 5, 142, 124, 32, 5, 144, 124, 5, 5, 144, 124, 32, 5, 146, 124, 5, 5, 146, 124, 32, 5, 148, 124, 5, 5, 148, 124, 32, 5, 150, 124, 5, 5, 150, 124, 193, 2, 152, 124, 193, 9, 152, 124, 193, 16, 152, 124, 193, 23, 152, 124, 193, 30, 152, 124, 193, 37, 152, 124, 32, 5, 154, 124, 5, 5, 154, 124, 193, 2, 156, 124, 255, 255, 255, 255, 193, 9, 162, 124, 193, 16, 162, 124, 193, 23, 162, 124, 193, 30, 162, 124, 193, 37, 162, 124, 193, 44, 162, 124, 193, 51, 162, 124, 193, 58, 162, 124, 193, 65, 162, 124, 193, 72, 162, 124, 193, 79, 162, 124, 193, 86, 162, 124, 193, 93, 162, 124, 193, 100, 162, 124, 193, 107, 162, 124, 193, 114, 162, 124, 193, 121, 162, 124, 193, 128, 162, 124, 193, 135, 162, 124, 193, 142, 162, 124, 193, 149, 162, 124, 193, 156, 162, 124, 198, 2, 172, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 9, 163, 124, 193, 16, 163, 124, 193, 23, 163, 124, 193, 30, 163, 124, 193, 37, 163, 124, 193, 44, 163, 124, 193, 51, 163, 124, 5, 5, 165, 124, 5, 112, 165, 124, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 206, 32, 236, 0, 193, 9, 201, 124, 193, 16, 201, 124, 193, 23, 201, 124, 193, 30, 201, 124, 193, 37, 201, 124, 193, 44, 201, 124, 193, 51, 201, 124, 193, 58, 201, 124, 193, 65, 201, 124, 193, 72, 201, 124, 193, 79, 201, 124, 193, 86, 201, 124, 193, 93, 201, 124, 193, 100, 201, 124, 193, 107, 201, 124, 193, 114, 201, 124, 193, 121, 201, 124, 193, 128, 201, 124, 193, 135, 201, 124, 193, 142, 201, 124, 193, 149, 201, 124, 193, 156, 201, 124, 193, 163, 201, 124, 193, 170, 201, 124, 193, 177, 201, 124, 193, 184, 201, 124, 193, 191, 201, 124, 193, 198, 201, 124, 193, 205, 201, 124, 193, 212, 201, 124, 193, 219, 201, 124, 193, 226, 201, 124, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 206, 96, 236, 0, 193, 9, 205, 124, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 32, 235, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 128, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 193, 222, 223, 124, 193, 229, 223, 124, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 160, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 206, 192, 239, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 206, 0, 240, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 206, 32, 240, 0, 5, 112, 36, 125, 5, 112, 38, 125, 5, 112, 40, 125, 5, 112, 42, 125, 5, 112, 44, 125, 5, 112, 46, 125, 5, 112, 48, 125, 5, 120, 48, 125, 5, 112, 50, 125, 5, 112, 52, 125, 5, 112, 54, 125, 5, 120, 54, 125, 5, 112, 56, 125, 5, 120, 56, 125, 5, 112, 58, 125, 5, 120, 58, 125, 5, 112, 60, 125, 5, 112, 62, 125, 5, 112, 64, 125, 5, 117, 66, 125, 5, 122, 66, 125, 5, 112, 68, 125, 5, 112, 70, 125, 5, 112, 72, 125, 5, 112, 74, 125, 5, 120, 74, 125, 5, 112, 76, 125, 5, 120, 76, 125, 5, 112, 78, 125, 5, 112, 80, 125, 193, 16, 82, 125, 193, 23, 82, 125, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 206, 64, 240, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 206, 192, 240, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 206, 224, 240, 0, 193, 96, 154, 125, 193, 98, 154, 125, 193, 100, 154, 125, 193, 102, 154, 125, 193, 104, 154, 125, 193, 106, 154, 125, 193, 108, 154, 125, 193, 110, 154, 125, 193, 112, 154, 125, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 206, 64, 241, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 206, 0, 241, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 198, 193, 119, 0, 198, 225, 173, 0, 198, 129, 174, 0, 198, 193, 174, 0, 198, 225, 174, 0, 198, 161, 175, 0, 198, 65, 121, 0, 198, 65, 176, 0, 198, 129, 176, 0, 198, 129, 177, 0, 198, 161, 177, 0, 198, 33, 123, 0, 198, 1, 178, 0, 198, 97, 178, 0, 198, 161, 178, 0, 198, 193, 178, 0, 198, 1, 179, 0, 198, 33, 179, 0, 198, 225, 179, 0, 198, 33, 180, 0, 198, 65, 180, 0, 198, 97, 180, 0, 198, 129, 180, 0, 198, 65, 124, 0, 198, 65, 181, 0, 198, 129, 181, 0, 198, 225, 181, 0, 198, 33, 182, 0, 198, 65, 182, 0, 198, 129, 182, 0, 198, 161, 183, 0, 198, 33, 126, 0, 198, 65, 172, 0, 198, 193, 175, 0, 198, 225, 172, 0, 198, 193, 183, 0, 198, 33, 176, 0, 198, 225, 177, 0, 198, 193, 172, 0, 198, 193, 177, 0, 198, 129, 175, 0, 198, 193, 180, 0, 198, 1, 195, 0, 198, 193, 198, 0, 198, 1, 198, 0, 198, 161, 195, 0, 198, 193, 216, 0, 198, 1, 184, 0, 198, 1, 194, 0, 198, 193, 195, 0, 198, 97, 195, 0, 198, 97, 204, 0, 198, 33, 183, 0, 198, 97, 200, 0, 198, 129, 213, 0, 198, 129, 204, 0, 198, 1, 180, 0, 198, 1, 205, 0, 198, 65, 202, 0, 198, 225, 185, 0, 198, 225, 215, 0, 198, 97, 177, 0, 198, 193, 181, 0, 198, 97, 198, 0, 198, 129, 179, 0, 198, 65, 206, 0, 198, 161, 201, 0, 198, 161, 213, 0, 198, 97, 184, 0, 198, 65, 183, 0, 198, 161, 198, 0, 198, 33, 192, 0, 198, 97, 192, 0, 198, 129, 172, 0, 198, 33, 173, 0, 198, 193, 215, 0, 198, 97, 188, 0, 198, 97, 174, 0, 198, 225, 182, 0, 198, 65, 192, 0, 198, 65, 214, 0, 198, 1, 192, 0, 198, 161, 204, 0, 198, 65, 205, 0, 198, 1, 183, 0, 198, 129, 198, 0, 198, 129, 195, 0, 198, 33, 195, 0, 198, 33, 202, 0, 198, 193, 179, 0, 198, 129, 183, 0, 198, 97, 216, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 193, 220, 171, 251, 193, 122, 79, 252, 193, 172, 95, 252, 193, 248, 45, 251, 193, 116, 132, 252, 193, 160, 14, 251, 193, 44, 168, 251, 193, 76, 2, 251, 193, 12, 24, 252, 193, 74, 216, 251, 193, 226, 174, 252, 193, 14, 54, 251, 193, 190, 93, 251, 193, 20, 114, 252, 193, 254, 232, 252, 193, 156, 33, 253, 193, 244, 14, 252, 193, 128, 78, 253, 193, 254, 22, 252, 193, 236, 194, 251, 193, 222, 224, 252, 193, 162, 198, 251, 193, 164, 95, 252, 193, 76, 9, 251, 193, 248, 27, 251, 193, 82, 32, 251, 193, 12, 241, 251, 193, 218, 172, 252, 193, 108, 222, 252, 193, 140, 34, 253, 193, 124, 98, 253, 193, 176, 44, 251, 193, 106, 20, 253, 193, 198, 136, 252, 193, 174, 35, 253, 193, 116, 37, 253, 193, 132, 50, 253, 193, 148, 52, 253, 193, 178, 78, 253, 193, 170, 81, 253, 193, 218, 133, 253, 193, 88, 148, 253, 193, 28, 150, 253, 193, 120, 173, 251, 193, 10, 56, 252, 193, 28, 219, 252, 255, 255, 255, 255, 255, 255, 255, 255, 193, 48, 3, 251, 193, 146, 31, 251, 193, 210, 27, 251, 193, 90, 13, 251, 193, 40, 26, 251, 193, 66, 29, 251, 193, 254, 42, 251, 193, 246, 44, 251, 193, 92, 67, 251, 193, 192, 64, 251, 193, 84, 67, 251, 193, 228, 69, 251, 193, 182, 93, 251, 193, 52, 96, 251, 193, 132, 103, 251, 193, 254, 103, 251, 198, 193, 197, 0, 198, 33, 198, 0, 198, 65, 198, 0, 198, 225, 198, 0, 198, 65, 199, 0, 198, 97, 199, 0, 198, 1, 174, 0, 198, 65, 200, 0, 198, 161, 200, 0, 198, 1, 201, 0, 198, 129, 202, 0, 198, 129, 203, 0, 198, 33, 204, 0, 198, 65, 204, 0, 198, 161, 205, 0, 198, 33, 206, 0, 198, 97, 206, 0, 198, 1, 207, 0, 198, 193, 206, 0, 198, 33, 207, 0, 198, 65, 207, 0, 198, 225, 206, 0, 198, 129, 207, 0, 198, 161, 207, 0, 198, 193, 207, 0, 198, 33, 208, 0, 198, 225, 208, 0, 198, 193, 208, 0, 198, 33, 209, 0, 198, 129, 209, 0, 198, 97, 210, 0, 198, 129, 210, 0, 193, 100, 44, 253, 193, 102, 232, 251, 193, 218, 70, 253, 193, 10, 50, 253, 193, 144, 34, 252, 193, 112, 4, 251, 193, 130, 56, 251, 193, 118, 224, 253, 193, 118, 224, 253, 193, 248, 103, 251, 193, 242, 98, 253, 193, 50, 67, 251, 193, 140, 103, 251, 193, 224, 186, 251, 193, 20, 106, 252, 193, 164, 192, 252, 193, 126, 251, 252, 193, 46, 9, 253, 193, 152, 19, 253, 193, 54, 87, 253, 193, 92, 250, 251, 193, 108, 25, 252, 193, 20, 51, 252, 193, 94, 81, 252, 193, 230, 232, 252, 193, 142, 94, 253, 193, 72, 167, 253, 193, 202, 7, 251, 193, 56, 50, 251, 193, 224, 4, 252, 193, 76, 63, 252, 193, 32, 250, 252, 198, 161, 174, 0, 198, 1, 182, 0, 198, 97, 175, 0, 198, 65, 175, 0, 198, 33, 175, 0, 198, 193, 176, 0, 198, 33, 178, 0, 198, 225, 178, 0, 198, 65, 179, 0, 198, 97, 179, 0, 198, 97, 181, 0, 198, 161, 181, 0, 198, 225, 186, 0, 198, 1, 187, 0, 198, 65, 187, 0, 198, 97, 187, 0, 198, 33, 187, 0, 198, 129, 187, 0, 198, 161, 188, 0, 198, 65, 189, 0, 198, 97, 190, 0, 198, 65, 190, 0, 198, 65, 191, 0, 198, 33, 191, 0, 198, 225, 191, 0, 198, 193, 192, 0, 255, 255, 255, 255, 198, 225, 193, 0, 198, 33, 194, 0, 198, 65, 195, 0, 198, 225, 196, 0, 198, 65, 197, 0, 193, 250, 2, 251, 193, 206, 4, 251, 193, 2, 5, 251, 193, 116, 8, 251, 193, 106, 12, 251, 193, 126, 14, 251, 193, 152, 14, 251, 193, 20, 16, 251, 193, 14, 18, 251, 193, 68, 19, 251, 193, 160, 21, 251, 193, 112, 21, 251, 193, 76, 24, 251, 193, 14, 34, 251, 193, 68, 26, 251, 193, 102, 26, 251, 193, 164, 27, 251, 193, 156, 28, 251, 193, 228, 28, 251, 193, 98, 29, 251, 193, 194, 27, 251, 193, 162, 29, 251, 193, 166, 29, 251, 193, 76, 30, 251, 193, 180, 30, 251, 193, 114, 10, 251, 193, 78, 31, 251, 193, 146, 31, 251, 193, 24, 140, 253, 193, 200, 34, 251, 193, 140, 35, 251, 193, 124, 36, 251, 193, 178, 209, 253, 193, 148, 34, 253, 193, 232, 98, 251, 193, 134, 160, 251, 193, 236, 166, 252, 193, 102, 203, 252, 193, 126, 68, 252, 193, 22, 131, 252, 193, 252, 49, 253, 193, 44, 135, 253, 193, 118, 98, 251, 193, 24, 134, 251, 193, 126, 250, 251, 193, 38, 29, 252, 193, 174, 36, 252, 193, 130, 175, 252, 193, 210, 181, 252, 193, 170, 126, 253, 193, 146, 43, 251, 193, 234, 203, 252, 193, 94, 33, 251, 193, 88, 32, 251, 193, 86, 143, 252, 193, 104, 178, 252, 193, 94, 230, 252, 193, 196, 127, 253, 193, 180, 39, 253, 193, 198, 193, 251, 193, 92, 250, 251, 193, 168, 35, 253, 193, 216, 4, 251, 193, 158, 125, 251, 193, 36, 18, 253, 193, 156, 33, 253, 193, 160, 157, 251, 193, 216, 170, 251, 193, 88, 198, 251, 193, 86, 14, 252, 193, 216, 164, 252, 193, 192, 78, 252, 193, 182, 10, 251, 193, 252, 83, 251, 193, 158, 125, 251, 193, 98, 145, 251, 193, 240, 171, 251, 193, 128, 80, 252, 193, 64, 85, 252, 193, 128, 193, 252, 193, 212, 200, 252, 193, 80, 100, 253, 193, 42, 135, 253, 193, 76, 139, 253, 193, 36, 149, 253, 193, 112, 13, 251, 193, 138, 139, 252, 193, 88, 97, 253, 193, 212, 131, 253, 193, 116, 176, 251, 193, 252, 7, 251, 193, 134, 21, 251, 193, 48, 126, 251, 193, 254, 131, 251, 193, 206, 218, 251, 193, 92, 250, 251, 193, 32, 223, 251, 193, 232, 235, 251, 193, 90, 241, 251, 193, 60, 24, 252, 193, 70, 82, 252, 193, 214, 100, 252, 193, 10, 192, 252, 193, 192, 18, 253, 193, 228, 18, 253, 193, 98, 98, 253, 193, 216, 133, 253, 193, 166, 47, 251, 193, 102, 34, 252, 193, 144, 57, 251, 193, 110, 59, 252, 193, 214, 86, 252, 193, 190, 248, 252, 193, 34, 130, 253, 193, 252, 190, 253, 193, 236, 210, 253, 193, 196, 236, 251, 193, 8, 29, 252, 193, 190, 215, 252, 193, 118, 152, 252, 193, 148, 155, 252, 193, 18, 169, 252, 193, 144, 72, 252, 193, 98, 49, 252, 193, 56, 38, 253, 193, 90, 10, 251, 193, 44, 226, 252, 193, 244, 36, 251, 193, 16, 206, 253, 193, 90, 141, 251, 193, 16, 43, 252, 193, 184, 245, 252, 193, 114, 23, 253, 193, 186, 193, 251, 193, 80, 214, 252, 193, 200, 12, 253, 193, 48, 157, 251, 193, 240, 233, 251, 193, 236, 26, 252, 193, 98, 74, 252, 193, 196, 88, 253, 193, 102, 13, 251, 193, 150, 31, 251, 193, 208, 43, 251, 193, 192, 207, 251, 193, 44, 3, 252, 193, 174, 62, 252, 193, 104, 112, 252, 193, 84, 198, 252, 193, 214, 248, 252, 193, 56, 253, 252, 193, 36, 61, 253, 193, 36, 138, 253, 193, 126, 184, 253, 193, 208, 203, 253, 193, 58, 129, 252, 193, 76, 137, 252, 193, 44, 178, 252, 193, 14, 230, 252, 193, 206, 103, 253, 193, 106, 59, 252, 193, 118, 104, 252, 193, 118, 238, 252, 193, 226, 84, 253, 193, 12, 224, 253, 193, 114, 227, 251, 193, 228, 125, 253, 193, 248, 40, 251, 193, 142, 236, 251, 193, 36, 238, 251, 193, 184, 25, 252, 193, 42, 34, 252, 193, 136, 81, 252, 193, 54, 95, 252, 193, 110, 128, 252, 193, 190, 174, 252, 193, 166, 151, 253, 193, 80, 28, 251, 193, 254, 189, 251, 193, 202, 127, 253, 193, 102, 16, 251, 193, 232, 139, 251, 193, 70, 29, 252, 193, 46, 73, 253, 193, 6, 167, 251, 193, 18, 180, 251, 193, 22, 239, 251, 193, 122, 79, 252, 193, 122, 128, 253, 193, 114, 36, 251, 193, 6, 57, 251, 193, 78, 134, 251, 193, 70, 204, 251, 193, 180, 198, 251, 193, 188, 204, 251, 193, 128, 204, 251, 193, 170, 204, 251, 193, 74, 206, 251, 193, 156, 204, 251, 193, 116, 210, 251, 193, 36, 214, 251, 193, 90, 215, 251, 193, 42, 217, 251, 193, 64, 222, 251, 193, 146, 232, 251, 193, 150, 224, 251, 193, 18, 230, 251, 193, 128, 226, 251, 193, 162, 227, 251, 193, 8, 224, 251, 193, 230, 29, 251, 193, 10, 30, 251, 193, 22, 30, 251, 193, 110, 228, 251, 193, 124, 204, 252, 193, 184, 204, 252, 193, 240, 233, 251, 193, 248, 233, 251, 193, 116, 209, 252, 193, 10, 236, 251, 193, 242, 235, 251, 193, 50, 236, 251, 193, 246, 236, 251, 193, 50, 238, 251, 193, 88, 60, 251, 193, 102, 62, 251, 193, 154, 62, 251, 193, 188, 64, 251, 193, 216, 64, 251, 193, 44, 67, 251, 193, 44, 67, 251, 193, 84, 67, 251, 193, 120, 67, 251, 193, 136, 67, 251, 193, 164, 69, 251, 193, 40, 85, 251, 193, 34, 72, 251, 193, 42, 85, 251, 193, 98, 72, 251, 193, 172, 74, 251, 193, 150, 35, 251, 193, 228, 99, 251, 193, 168, 88, 251, 193, 218, 90, 251, 193, 14, 91, 251, 193, 82, 88, 251, 193, 68, 92, 251, 193, 66, 92, 251, 193, 38, 96, 251, 193, 30, 99, 251, 193, 238, 99, 251, 193, 32, 100, 251, 193, 198, 100, 251, 193, 154, 101, 251, 193, 76, 102, 251, 193, 226, 104, 251, 193, 218, 57, 251, 193, 100, 106, 251, 193, 108, 159, 251, 193, 166, 220, 251, 193, 46, 44, 252, 193, 228, 133, 252, 193, 130, 121, 253, 193, 220, 171, 253, 193, 136, 210, 253, 193, 132, 214, 253, 193, 8, 42, 251, 193, 250, 229, 251, 193, 88, 11, 252, 193, 208, 75, 253, 193, 176, 153, 251, 193, 4, 183, 251, 193, 206, 187, 251, 193, 68, 206, 251, 193, 212, 36, 252, 193, 12, 55, 252, 193, 26, 86, 252, 193, 208, 140, 252, 193, 168, 179, 252, 193, 164, 202, 252, 193, 38, 73, 253, 193, 90, 238, 252, 193, 8, 81, 253, 193, 98, 105, 253, 193, 250, 35, 251, 193, 52, 42, 251, 193, 142, 60, 251, 193, 240, 50, 252, 193, 246, 36, 251, 193, 158, 37, 251, 193, 126, 39, 251, 193, 248, 39, 251, 193, 66, 42, 251, 193, 254, 42, 251, 193, 4, 43, 251, 193, 12, 44, 251, 193, 246, 44, 251, 193, 20, 45, 251, 193, 22, 45, 251, 193, 248, 45, 251, 193, 208, 47, 251, 193, 46, 48, 251, 193, 154, 48, 251, 193, 52, 50, 251, 193, 94, 50, 251, 193, 118, 50, 251, 193, 118, 50, 251, 193, 118, 50, 251, 193, 188, 50, 251, 193, 158, 48, 252, 193, 128, 54, 251, 193, 110, 55, 251, 193, 178, 55, 251, 193, 142, 56, 251, 193, 154, 56, 251, 193, 242, 56, 251, 193, 80, 60, 251, 193, 198, 57, 251, 193, 230, 57, 251, 193, 242, 58, 251, 193, 102, 233, 251, 193, 176, 204, 252, 193, 134, 204, 252, 193, 108, 206, 252, 193, 144, 208, 252, 193, 54, 208, 252, 193, 192, 113, 251, 193, 240, 210, 252, 193, 14, 211, 252, 193, 200, 234, 251, 193, 16, 235, 251, 193, 52, 217, 252, 193, 96, 217, 252, 193, 224, 77, 253, 193, 70, 220, 252, 193, 108, 223, 252, 193, 96, 223, 252, 193, 192, 223, 252, 193, 112, 42, 251, 193, 232, 223, 252, 193, 236, 223, 252, 193, 2, 224, 252, 193, 224, 224, 252, 193, 94, 225, 252, 193, 222, 224, 252, 193, 180, 227, 252, 193, 130, 226, 252, 193, 190, 232, 252, 193, 182, 227, 252, 193, 82, 228, 252, 193, 74, 230, 252, 193, 206, 230, 252, 198, 65, 184, 0, 198, 129, 184, 0, 198, 161, 184, 0, 198, 193, 184, 0, 198, 33, 185, 0, 198, 193, 185, 0, 198, 1, 186, 0, 198, 97, 186, 0, 198, 161, 186, 0, 198, 193, 186, 0, 198, 33, 187, 0, 198, 161, 187, 0, 198, 193, 187, 0, 198, 225, 187, 0, 198, 1, 188, 0, 198, 33, 188, 0, 198, 129, 188, 0, 198, 193, 188, 0, 198, 225, 188, 0, 198, 65, 189, 0, 198, 129, 189, 0, 198, 161, 189, 0, 198, 193, 189, 0, 198, 225, 189, 0, 198, 1, 190, 0, 198, 33, 190, 0, 198, 129, 190, 0, 198, 161, 190, 0, 198, 1, 191, 0, 198, 97, 191, 0, 198, 129, 191, 0, 198, 161, 191, 0, 193, 118, 106, 251, 193, 80, 108, 251, 193, 176, 109, 251, 193, 214, 109, 251, 193, 238, 109, 251, 193, 162, 108, 251, 193, 56, 111, 251, 193, 160, 112, 251, 193, 188, 112, 251, 193, 240, 115, 251, 193, 122, 118, 251, 193, 122, 118, 251, 193, 236, 121, 251, 193, 196, 123, 251, 193, 248, 124, 251, 193, 158, 125, 251, 193, 104, 127, 251, 193, 6, 128, 251, 193, 46, 128, 251, 193, 80, 128, 251, 193, 216, 164, 251, 193, 144, 130, 251, 193, 6, 131, 251, 193, 132, 133, 251, 193, 8, 135, 251, 193, 52, 137, 251, 193, 212, 137, 251, 193, 46, 139, 251, 193, 66, 141, 251, 193, 50, 139, 251, 193, 106, 142, 251, 193, 100, 142, 251, 193, 48, 111, 251, 193, 160, 117, 251, 193, 250, 157, 251, 193, 98, 158, 251, 193, 208, 165, 251, 193, 130, 168, 251, 193, 98, 176, 251, 193, 38, 180, 251, 193, 46, 178, 251, 193, 254, 182, 251, 193, 124, 181, 251, 193, 70, 186, 251, 193, 116, 190, 251, 193, 96, 200, 251, 193, 154, 202, 251, 193, 226, 200, 251, 193, 50, 214, 251, 193, 96, 226, 251, 193, 240, 233, 251, 193, 248, 233, 251, 193, 248, 235, 251, 193, 192, 11, 252, 193, 244, 14, 252, 193, 184, 25, 252, 193, 162, 34, 252, 193, 94, 32, 252, 193, 210, 36, 252, 193, 100, 45, 252, 193, 188, 53, 252, 193, 206, 120, 252, 193, 94, 65, 252, 193, 58, 72, 252, 193, 32, 143, 251, 193, 212, 78, 253, 193, 158, 147, 251, 193, 142, 148, 251, 193, 220, 148, 251, 193, 146, 150, 251, 193, 98, 151, 251, 193, 232, 152, 251, 193, 234, 152, 251, 193, 158, 153, 251, 193, 214, 155, 251, 193, 144, 156, 251, 193, 150, 156, 251, 193, 156, 156, 251, 193, 48, 157, 251, 193, 102, 213, 253, 193, 102, 160, 251, 193, 122, 160, 251, 193, 122, 160, 251, 193, 52, 217, 252, 193, 122, 162, 251, 193, 122, 162, 251, 193, 234, 164, 251, 193, 86, 230, 251, 193, 36, 185, 252, 193, 140, 165, 251, 193, 210, 165, 251, 193, 84, 167, 251, 193, 34, 168, 251, 193, 54, 170, 251, 193, 74, 170, 251, 193, 224, 170, 251, 193, 44, 85, 253, 193, 170, 85, 253, 193, 116, 87, 253, 193, 212, 89, 253, 193, 244, 90, 253, 193, 196, 90, 253, 193, 92, 91, 253, 193, 88, 100, 253, 193, 154, 102, 253, 193, 156, 102, 253, 193, 178, 103, 253, 193, 170, 108, 253, 193, 176, 109, 253, 193, 122, 113, 253, 193, 84, 120, 253, 193, 54, 121, 253, 193, 108, 122, 253, 193, 186, 122, 253, 193, 44, 130, 253, 193, 36, 132, 253, 193, 234, 144, 251, 193, 66, 137, 253, 193, 222, 137, 253, 193, 232, 140, 253, 193, 72, 143, 253, 193, 36, 144, 253, 193, 184, 146, 253, 193, 150, 147, 253, 193, 174, 148, 253, 193, 188, 149, 253, 193, 198, 148, 253, 193, 198, 148, 253, 193, 138, 174, 251, 193, 178, 173, 251, 193, 142, 173, 251, 193, 176, 174, 251, 193, 242, 176, 251, 193, 64, 176, 251, 193, 26, 180, 251, 193, 34, 180, 251, 193, 38, 180, 251, 193, 34, 180, 251, 193, 172, 181, 251, 193, 254, 182, 251, 193, 72, 183, 251, 193, 44, 183, 251, 193, 66, 183, 251, 193, 144, 185, 251, 193, 70, 186, 251, 193, 224, 186, 251, 193, 100, 188, 251, 193, 40, 189, 251, 193, 234, 191, 251, 193, 138, 193, 251, 193, 208, 193, 251, 193, 200, 196, 251, 193, 180, 195, 251, 193, 162, 196, 251, 193, 232, 194, 251, 193, 50, 198, 251, 193, 104, 198, 251, 193, 158, 200, 251, 193, 20, 201, 251, 193, 166, 202, 251, 193, 70, 186, 251, 193, 36, 214, 251, 193, 54, 222, 251, 193, 128, 226, 251, 193, 20, 241, 251, 193, 8, 27, 252, 193, 160, 29, 252, 193, 210, 36, 252, 193, 188, 53, 252, 193, 50, 64, 252, 193, 8, 83, 252, 193, 68, 129, 252, 193, 240, 134, 252, 193, 40, 135, 252, 193, 38, 135, 252, 193, 132, 135, 252, 193, 144, 135, 252, 193, 158, 135, 252, 193, 242, 137, 252, 193, 244, 137, 252, 193, 236, 144, 252, 193, 94, 148, 252, 193, 104, 157, 252, 193, 168, 179, 252, 193, 180, 180, 252, 193, 230, 181, 252, 193, 66, 191, 252, 193, 114, 198, 252, 193, 246, 215, 252, 193, 8, 223, 252, 193, 8, 223, 252, 193, 206, 230, 252, 198, 129, 192, 0, 198, 161, 192, 0, 198, 225, 192, 0, 198, 33, 193, 0, 198, 97, 193, 0, 198, 161, 193, 0, 198, 193, 193, 0, 198, 225, 126, 0, 198, 225, 194, 0, 198, 65, 127, 0, 198, 1, 128, 0, 198, 97, 196, 0, 198, 129, 196, 0, 198, 193, 196, 0, 198, 1, 197, 0, 198, 33, 197, 0, 198, 97, 197, 0, 198, 129, 197, 0, 198, 161, 197, 0, 198, 225, 197, 0, 198, 193, 128, 0, 198, 33, 129, 0, 198, 33, 199, 0, 198, 129, 199, 0, 198, 161, 199, 0, 198, 193, 199, 0, 198, 225, 199, 0, 198, 1, 200, 0, 198, 33, 200, 0, 198, 129, 200, 0, 198, 193, 200, 0, 198, 225, 200, 0, 193, 160, 236, 251, 193, 140, 239, 251, 193, 20, 241, 251, 193, 80, 240, 251, 193, 38, 241, 251, 193, 38, 239, 251, 193, 74, 243, 251, 193, 16, 246, 251, 193, 80, 245, 251, 193, 234, 247, 251, 193, 44, 250, 251, 193, 220, 254, 251, 193, 104, 253, 251, 193, 60, 3, 252, 193, 116, 5, 252, 193, 190, 6, 252, 193, 112, 7, 252, 193, 32, 9, 252, 193, 182, 9, 252, 193, 234, 10, 252, 193, 198, 13, 252, 193, 244, 14, 252, 193, 246, 14, 252, 193, 254, 14, 252, 193, 52, 135, 251, 193, 158, 16, 252, 193, 74, 22, 252, 193, 152, 24, 252, 193, 240, 23, 252, 193, 14, 24, 252, 193, 188, 22, 252, 193, 194, 26, 252, 193, 178, 25, 252, 193, 8, 27, 252, 193, 184, 25, 252, 193, 234, 26, 252, 193, 10, 27, 252, 193, 36, 27, 252, 193, 140, 27, 252, 193, 158, 25, 252, 193, 162, 31, 252, 193, 34, 32, 252, 193, 222, 34, 252, 193, 94, 32, 252, 193, 128, 34, 252, 193, 78, 37, 252, 193, 100, 29, 252, 193, 4, 39, 252, 193, 194, 39, 252, 193, 36, 40, 252, 193, 132, 41, 252, 193, 92, 46, 252, 193, 100, 45, 252, 193, 94, 45, 252, 193, 10, 47, 252, 193, 244, 46, 252, 193, 218, 48, 252, 193, 206, 48, 252, 193, 10, 50, 252, 193, 2, 29, 251, 193, 4, 55, 252, 193, 216, 53, 252, 193, 22, 58, 252, 193, 192, 59, 252, 193, 36, 64, 252, 193, 94, 65, 252, 193, 158, 67, 252, 193, 50, 68, 252, 193, 244, 69, 252, 193, 200, 70, 252, 193, 184, 72, 252, 193, 174, 76, 252, 193, 248, 78, 252, 193, 134, 79, 252, 193, 194, 79, 252, 193, 248, 79, 252, 193, 194, 80, 252, 193, 194, 80, 252, 193, 42, 84, 252, 193, 84, 84, 252, 193, 80, 85, 252, 193, 18, 86, 252, 193, 170, 88, 252, 193, 160, 90, 252, 193, 108, 93, 252, 193, 214, 94, 252, 193, 104, 94, 252, 193, 128, 96, 252, 193, 172, 95, 252, 193, 24, 154, 251, 193, 106, 102, 252, 193, 174, 108, 252, 193, 244, 108, 252, 193, 210, 110, 252, 193, 232, 110, 252, 193, 108, 112, 252, 193, 54, 75, 252, 193, 80, 85, 252, 193, 158, 91, 252, 193, 100, 94, 252, 193, 36, 103, 252, 193, 124, 102, 252, 193, 12, 111, 252, 193, 104, 111, 252, 193, 180, 113, 252, 193, 114, 116, 252, 193, 96, 116, 252, 193, 26, 131, 252, 193, 158, 150, 252, 193, 104, 157, 252, 193, 124, 168, 252, 193, 46, 177, 252, 193, 168, 179, 252, 193, 18, 189, 252, 193, 114, 198, 252, 193, 98, 226, 252, 193, 90, 230, 252, 193, 184, 7, 253, 193, 214, 21, 253, 193, 4, 25, 253, 193, 198, 136, 252, 193, 102, 34, 253, 193, 156, 35, 253, 193, 126, 34, 253, 193, 174, 35, 253, 193, 168, 35, 253, 193, 134, 35, 253, 193, 116, 37, 253, 198, 33, 201, 0, 198, 65, 201, 0, 198, 97, 201, 0, 198, 129, 201, 0, 198, 193, 201, 0, 198, 225, 201, 0, 198, 97, 202, 0, 198, 161, 202, 0, 198, 193, 202, 0, 198, 225, 202, 0, 198, 1, 203, 0, 198, 33, 203, 0, 198, 97, 203, 0, 198, 161, 203, 0, 198, 193, 203, 0, 198, 225, 203, 0, 198, 1, 204, 0, 198, 193, 204, 0, 198, 225, 204, 0, 198, 33, 205, 0, 198, 97, 205, 0, 198, 129, 205, 0, 198, 225, 205, 0, 198, 1, 206, 0, 198, 129, 206, 0, 198, 161, 206, 0, 198, 97, 207, 0, 198, 225, 207, 0, 198, 1, 208, 0, 198, 65, 208, 0, 198, 97, 208, 0, 198, 129, 208, 0, 193, 180, 113, 252, 193, 196, 113, 252, 193, 194, 113, 252, 193, 110, 114, 252, 193, 22, 115, 252, 193, 224, 114, 252, 193, 226, 114, 252, 193, 226, 114, 252, 193, 114, 116, 252, 193, 152, 116, 252, 193, 62, 119, 252, 193, 140, 117, 252, 193, 136, 125, 252, 193, 190, 125, 252, 193, 136, 127, 252, 193, 58, 129, 252, 193, 26, 131, 252, 193, 144, 132, 252, 193, 98, 135, 252, 193, 144, 135, 252, 193, 142, 137, 252, 193, 148, 138, 252, 193, 246, 137, 252, 193, 172, 141, 252, 193, 32, 141, 252, 193, 236, 144, 252, 193, 46, 144, 252, 193, 158, 145, 252, 193, 120, 149, 252, 193, 240, 153, 252, 193, 240, 153, 252, 193, 8, 154, 252, 193, 136, 158, 252, 193, 222, 158, 252, 193, 166, 159, 252, 193, 218, 160, 252, 193, 242, 159, 252, 193, 170, 162, 252, 193, 228, 171, 252, 193, 130, 172, 252, 193, 120, 172, 252, 193, 210, 172, 252, 193, 108, 174, 252, 193, 90, 174, 252, 193, 106, 176, 252, 193, 236, 179, 252, 193, 122, 178, 252, 193, 196, 179, 252, 193, 238, 181, 252, 193, 166, 185, 252, 193, 232, 188, 252, 193, 98, 189, 252, 193, 16, 191, 252, 193, 114, 191, 252, 193, 12, 192, 252, 193, 236, 192, 252, 193, 126, 193, 252, 193, 224, 197, 252, 193, 114, 198, 252, 193, 10, 199, 252, 193, 32, 200, 252, 193, 38, 201, 252, 193, 46, 202, 252, 193, 166, 202, 252, 198, 161, 208, 0, 198, 1, 209, 0, 198, 65, 209, 0, 198, 97, 131, 0, 198, 193, 131, 0, 198, 97, 209, 0, 198, 161, 209, 0, 198, 193, 209, 0, 198, 225, 209, 0, 198, 1, 210, 0, 198, 33, 210, 0, 198, 65, 210, 0, 198, 193, 210, 0, 198, 1, 211, 0, 198, 33, 211, 0, 198, 65, 211, 0, 198, 97, 211, 0, 198, 161, 211, 0, 198, 1, 212, 0, 198, 65, 212, 0, 198, 161, 212, 0, 198, 225, 212, 0, 198, 1, 213, 0, 198, 33, 213, 0, 198, 65, 213, 0, 198, 97, 213, 0, 198, 1, 214, 0, 198, 33, 214, 0, 198, 97, 214, 0, 198, 161, 214, 0, 198, 193, 214, 0, 198, 1, 215, 0, 198, 161, 210, 0, 198, 225, 210, 0, 198, 129, 211, 0, 198, 225, 211, 0, 198, 193, 211, 0, 198, 33, 212, 0, 198, 65, 212, 0, 198, 97, 212, 0, 198, 193, 212, 0, 198, 225, 213, 0, 198, 129, 214, 0, 198, 225, 214, 0, 198, 97, 215, 0, 198, 129, 215, 0, 198, 161, 215, 0, 198, 33, 216, 0, 198, 225, 216, 0, 198, 1, 217, 0, 198, 33, 217, 0, 198, 65, 217, 0, 198, 129, 217, 0, 198, 193, 217, 0, 198, 225, 217, 0, 198, 97, 218, 0, 198, 161, 218, 0, 198, 65, 219, 0, 198, 225, 219, 0, 198, 33, 220, 0, 198, 97, 220, 0, 198, 129, 220, 0, 198, 193, 220, 0, 198, 161, 220, 0, 193, 244, 229, 252, 193, 16, 230, 252, 193, 20, 230, 252, 193, 52, 230, 252, 193, 32, 231, 252, 193, 120, 236, 252, 193, 210, 233, 252, 193, 52, 239, 252, 193, 96, 238, 252, 193, 100, 238, 252, 193, 170, 238, 252, 193, 112, 8, 253, 193, 42, 241, 252, 193, 214, 243, 252, 193, 222, 245, 252, 193, 140, 247, 252, 193, 238, 245, 252, 193, 58, 250, 252, 193, 254, 248, 252, 193, 194, 252, 252, 193, 56, 253, 252, 193, 28, 254, 252, 193, 60, 254, 252, 193, 118, 255, 252, 193, 228, 2, 253, 193, 26, 4, 253, 193, 10, 3, 253, 193, 184, 7, 253, 193, 28, 5, 253, 193, 164, 6, 253, 193, 196, 7, 253, 193, 16, 8, 253, 193, 206, 6, 253, 193, 96, 9, 253, 193, 140, 11, 253, 193, 6, 16, 253, 193, 52, 16, 253, 193, 86, 18, 253, 193, 208, 18, 253, 193, 222, 18, 253, 193, 182, 19, 253, 193, 156, 19, 253, 193, 12, 30, 251, 193, 138, 27, 253, 193, 138, 30, 253, 193, 240, 30, 253, 193, 204, 31, 253, 193, 172, 32, 253, 193, 134, 35, 253, 193, 22, 40, 253, 193, 112, 45, 253, 193, 76, 47, 253, 193, 18, 49, 253, 193, 154, 49, 253, 193, 168, 53, 253, 193, 120, 55, 253, 193, 196, 57, 253, 193, 96, 41, 251, 193, 84, 60, 253, 193, 2, 61, 253, 193, 38, 61, 253, 193, 88, 45, 251, 193, 8, 71, 253, 193, 200, 73, 253, 193, 22, 40, 253, 193, 148, 52, 253, 193, 200, 73, 253, 193, 208, 84, 253, 193, 58, 96, 253, 193, 64, 101, 253, 193, 102, 128, 253, 193, 218, 133, 253, 193, 234, 139, 253, 193, 182, 146, 253, 193, 88, 148, 253, 193, 198, 148, 253, 193, 28, 150, 253, 193, 244, 178, 253, 193, 118, 224, 253, 193, 154, 183, 251, 193, 142, 183, 251, 193, 10, 237, 251, 193, 16, 246, 251, 193, 52, 114, 252, 193, 152, 116, 252, 193, 128, 119, 252, 193, 234, 163, 252, 193, 142, 56, 253, 193, 110, 220, 253, 193, 14, 224, 253, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 198, 33, 215, 0, 198, 65, 215, 0, 198, 1, 216, 0, 198, 65, 216, 0, 198, 129, 216, 0, 198, 161, 216, 0, 198, 225, 132, 0, 198, 1, 217, 0, 198, 97, 217, 0, 198, 161, 217, 0, 198, 1, 218, 0, 198, 33, 218, 0, 198, 65, 218, 0, 198, 129, 218, 0, 198, 193, 218, 0, 198, 225, 218, 0, 198, 1, 219, 0, 198, 33, 219, 0, 198, 97, 219, 0, 198, 129, 219, 0, 198, 161, 219, 0, 198, 1, 220, 0, 198, 65, 220, 0, 198, 129, 220, 0, 198, 1, 221, 0, 198, 65, 221, 0, 198, 97, 221, 0, 198, 161, 221, 0, 198, 225, 221, 0, 198, 1, 222, 0, 198, 33, 222, 0, 198, 65, 222, 0, 193, 158, 149, 253, 193, 192, 154, 253, 193, 188, 157, 253, 193, 122, 158, 253, 193, 190, 159, 253, 193, 114, 165, 253, 193, 54, 166, 253, 193, 220, 167, 253, 193, 18, 174, 253, 193, 162, 177, 253, 193, 244, 178, 253, 193, 58, 190, 253, 193, 118, 196, 253, 193, 106, 198, 253, 193, 58, 201, 253, 193, 218, 199, 253, 193, 114, 201, 253, 193, 246, 202, 253, 193, 84, 202, 253, 193, 182, 205, 253, 193, 220, 209, 253, 193, 94, 213, 253, 193, 190, 214, 253, 193, 36, 217, 253, 193, 222, 217, 253, 193, 158, 217, 253, 193, 10, 218, 253, 193, 86, 218, 253, 193, 240, 219, 253, 193, 176, 221, 253, 255, 255, 255, 255, 255, 255, 255, 255, 198, 225, 220, 0, 198, 33, 221, 0, 198, 129, 221, 0, 198, 193, 221, 0, 198, 161, 222, 0, 198, 225, 222, 0, 198, 33, 223, 0, 198, 97, 223, 0, 198, 193, 223, 0, 198, 33, 224, 0, 198, 193, 224, 0, 198, 129, 225, 0, 198, 161, 225, 0, 198, 225, 225, 0, 198, 1, 226, 0, 198, 65, 226, 0, 198, 97, 226, 0, 198, 161, 226, 0, 198, 193, 226, 0, 198, 225, 226, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 198, 97, 222, 0, 198, 129, 222, 0, 198, 193, 222, 0, 198, 1, 223, 0, 198, 65, 223, 0, 198, 129, 223, 0, 198, 161, 223, 0, 198, 225, 223, 0, 198, 1, 224, 0, 198, 65, 224, 0, 198, 97, 224, 0, 198, 129, 224, 0, 198, 161, 224, 0, 198, 225, 224, 0, 198, 1, 225, 0, 198, 33, 225, 0, 198, 65, 225, 0, 198, 97, 225, 0, 198, 193, 225, 0, 198, 33, 226, 0, 198, 129, 226, 0, 198, 1, 227, 0, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 2, 212, 78, 0, 0, 150, 10, 251, 2, 238, 78, 0, 0, 2, 11, 251, 2, 180, 1, 2, 0, 148, 11, 251, 2, 45, 79, 0, 0, 4, 12, 251, 2, 105, 79, 0, 0, 44, 13, 251, 2, 3, 2, 2, 0, 212, 13, 251, 2, 123, 207, 2, 0, 66, 14, 251, 2, 113, 0, 3, 0, 86, 14, 251, 2, 174, 79, 0, 0, 126, 14, 251, 2, 238, 79, 0, 0, 234, 15, 251, 2, 99, 2, 2, 0, 164, 16, 251, 2, 154, 207, 2, 0, 104, 17, 251, 2, 64, 80, 0, 0, 152, 17, 251, 2, 167, 2, 2, 0, 38, 18, 251, 2, 128, 80, 0, 0, 18, 19, 251, 2, 237, 2, 2, 0, 120, 19, 251, 2, 170, 80, 0, 0, 82, 20, 251, 2, 206, 80, 0, 0, 110, 21, 251, 2, 109, 3, 2, 0, 220, 21, 251, 2, 248, 80, 0, 0, 188, 22, 251, 2, 183, 3, 2, 0, 252, 22, 251, 2, 212, 3, 2, 0, 50, 23, 251, 2, 232, 3, 2, 0, 200, 23, 251, 2, 249, 184, 2, 0, 238, 24, 251, 2, 216, 5, 2, 0, 100, 32, 251, 2, 250, 5, 2, 0, 16, 33, 251, 2, 60, 6, 2, 0, 32, 34, 251, 2, 188, 6, 2, 0, 20, 36, 251, 2, 28, 82, 0, 0, 88, 36, 251, 2, 206, 6, 2, 0, 128, 36, 251, 2, 231, 6, 2, 0, 20, 37, 251, 2, 16, 7, 2, 0, 192, 37, 251, 2, 82, 82, 0, 0, 42, 38, 251, 2, 49, 7, 2, 0, 96, 38, 251, 2, 250, 7, 2, 0, 76, 41, 251, 2, 123, 8, 2, 0, 164, 43, 251, 2, 87, 10, 2, 0, 126, 51, 251, 2, 153, 10, 2, 0, 192, 52, 251, 2, 85, 11, 2, 0, 150, 55, 251, 2, 171, 11, 2, 0, 32, 57, 251, 2, 24, 84, 0, 0, 134, 57, 251, 2, 203, 11, 2, 0, 48, 58, 251, 2, 8, 12, 2, 0, 114, 59, 251, 2, 80, 12, 2, 0, 246, 60, 251, 2, 58, 168, 2, 0, 176, 61, 251, 2, 212, 1, 3, 0, 26, 62, 251, 2, 224, 84, 0, 0, 58, 62, 251, 2, 172, 12, 2, 0, 234, 62, 251, 2, 236, 12, 2, 0, 104, 63, 251, 2, 245, 209, 2, 0, 30, 64, 251, 2, 40, 85, 0, 0, 102, 64, 251, 2, 39, 13, 2, 0, 54, 65, 251, 2, 146, 186, 2, 0, 114, 66, 251, 2, 12, 210, 2, 0, 176, 66, 251, 2, 122, 85, 0, 0, 24, 67, 251, 2, 142, 85, 0, 0, 62, 67, 251, 2, 182, 53, 0, 0, 162, 67, 251, 2, 227, 13, 2, 0, 50, 68, 251, 2, 92, 168, 2, 0, 224, 68, 251, 2, 47, 210, 2, 0, 34, 69, 251, 2, 115, 14, 2, 0, 156, 70, 251, 2, 90, 210, 2, 0, 156, 71, 251, 2, 246, 85, 0, 0, 2, 72, 251, 2, 212, 14, 2, 0, 134, 72, 251, 2, 120, 168, 2, 0, 162, 73, 251, 2, 211, 186, 2, 0, 190, 73, 251, 2, 40, 86, 0, 0, 72, 74, 251, 2, 97, 15, 2, 0, 206, 74, 251, 2, 168, 15, 2, 0, 92, 75, 251, 2, 251, 53, 0, 0, 90, 76, 251, 2, 199, 15, 2, 0, 110, 76, 251, 2, 21, 16, 2, 0, 10, 77, 251, 2, 246, 186, 2, 0, 106, 77, 251, 2, 61, 16, 2, 0, 10, 78, 251, 2, 220, 210, 2, 0, 206, 78, 251, 2, 148, 86, 0, 0, 6, 79, 251, 2, 143, 16, 2, 0, 54, 79, 251, 2, 240, 210, 2, 0, 252, 79, 251, 2, 241, 16, 2, 0, 122, 80, 251, 2, 28, 187, 2, 0, 190, 80, 251, 2, 17, 17, 2, 0, 44, 81, 251, 2, 187, 86, 0, 0, 170, 81, 251, 2, 49, 17, 2, 0, 198, 81, 251, 2, 121, 17, 2, 0, 186, 82, 251, 2, 180, 17, 2, 0, 192, 83, 251, 2, 88, 18, 2, 0, 180, 86, 251, 2, 91, 187, 2, 0, 36, 87, 251, 2, 97, 87, 0, 0, 84, 87, 251, 2, 132, 18, 2, 0, 178, 87, 251, 2, 139, 87, 0, 0, 82, 88, 251, 2, 183, 18, 2, 0, 190, 88, 251, 2, 182, 87, 0, 0, 118, 89, 251, 2, 237, 18, 2, 0, 210, 89, 251, 2, 212, 168, 2, 0, 72, 90, 251, 2, 156, 211, 2, 0, 128, 90, 251, 2, 220, 87, 0, 0, 172, 90, 251, 2, 235, 87, 0, 0, 200, 90, 251, 2, 38, 19, 2, 0, 46, 91, 251, 2, 124, 187, 2, 0, 198, 91, 251, 2, 22, 88, 0, 0, 12, 92, 251, 2, 100, 19, 2, 0, 132, 92, 251, 2, 182, 211, 2, 0, 76, 93, 251, 2, 73, 88, 0, 0, 148, 93, 251, 2, 174, 19, 2, 0, 244, 93, 251, 2, 114, 88, 0, 0, 226, 94, 251, 2, 237, 19, 2, 0, 70, 95, 251, 2, 153, 2, 3, 0, 236, 95, 251, 2, 156, 88, 0, 0, 8, 96, 251, 2, 186, 88, 0, 0, 40, 97, 251, 2, 123, 20, 2, 0, 150, 97, 251, 2, 220, 20, 2, 0, 14, 99, 251, 2, 247, 21, 2, 0, 162, 103, 251, 2, 52, 22, 2, 0, 158, 104, 251, 2, 91, 22, 2, 0, 72, 105, 251, 2, 119, 22, 2, 0, 184, 105, 251, 2, 119, 89, 0, 0, 144, 106, 251, 2, 142, 89, 0, 0, 246, 106, 251, 2, 188, 22, 2, 0, 68, 107, 251, 2, 172, 89, 0, 0, 176, 107, 251, 2, 216, 89, 0, 0, 162, 108, 251, 2, 237, 89, 0, 0, 202, 108, 251, 2, 2, 23, 2, 0, 24, 109, 251, 2, 28, 188, 2, 0, 126, 109, 251, 2, 9, 90, 0, 0, 178, 109, 251, 2, 46, 23, 2, 0, 58, 110, 251, 2, 123, 169, 2, 0, 232, 111, 251, 2, 119, 90, 0, 0, 46, 112, 251, 2, 145, 90, 0, 0, 96, 112, 251, 2, 238, 54, 0, 0, 160, 112, 251, 2, 156, 23, 2, 0, 206, 112, 251, 2, 176, 90, 0, 0, 182, 113, 251, 2, 232, 23, 2, 0, 32, 114, 251, 2, 213, 90, 0, 0, 186, 114, 251, 2, 51, 24, 2, 0, 70, 115, 251, 2, 243, 90, 0, 0, 198, 115, 251, 2, 79, 24, 2, 0, 20, 116, 251, 2, 145, 24, 2, 0, 250, 116, 251, 2, 120, 188, 2, 0, 110, 117, 251, 2, 91, 25, 2, 0, 42, 120, 251, 2, 214, 25, 2, 0, 110, 122, 251, 2, 218, 188, 2, 0, 196, 124, 251, 2, 123, 26, 2, 0, 170, 125, 251, 2, 252, 26, 2, 0, 174, 127, 251, 2, 122, 28, 2, 0, 58, 133, 251, 2, 148, 28, 2, 0, 140, 133, 251, 2, 172, 28, 2, 0, 202, 133, 251, 2, 27, 29, 2, 0, 86, 135, 251, 2, 120, 92, 0, 0, 196, 135, 251, 2, 185, 55, 0, 0, 64, 137, 251, 2, 131, 29, 2, 0, 86, 137, 251, 2, 198, 92, 0, 0, 214, 137, 251, 2, 171, 29, 2, 0, 34, 138, 251, 2, 232, 92, 0, 0, 178, 138, 251, 2, 213, 29, 2, 0, 16, 139, 251, 2, 6, 93, 0, 0, 194, 139, 251, 2, 215, 55, 0, 0, 26, 140, 251, 2, 17, 30, 2, 0, 56, 140, 251, 2, 49, 93, 0, 0, 30, 141, 251, 2, 106, 30, 2, 0, 140, 141, 251, 2, 170, 30, 2, 0, 142, 142, 251, 2, 119, 93, 0, 0, 22, 143, 251, 2, 249, 55, 0, 0, 70, 143, 251, 2, 216, 30, 2, 0, 94, 143, 251, 2, 143, 93, 0, 0, 250, 143, 251, 2, 10, 31, 2, 0, 66, 144, 251, 2, 188, 31, 2, 0, 188, 146, 251, 2, 85, 32, 2, 0, 66, 149, 251, 2, 123, 32, 2, 0, 224, 149, 251, 2, 144, 32, 2, 0, 66, 150, 251, 2, 177, 32, 2, 0, 170, 150, 251, 2, 205, 32, 2, 0, 20, 151, 251, 2, 235, 32, 2, 0, 138, 151, 251, 2, 92, 94, 0, 0, 144, 152, 251, 2, 51, 33, 2, 0, 168, 152, 251, 2, 242, 33, 2, 0, 146, 155, 251, 2, 25, 34, 2, 0, 54, 156, 251, 2, 51, 34, 2, 0, 180, 156, 251, 2, 91, 34, 2, 0, 58, 157, 251, 2, 149, 34, 2, 0, 18, 158, 251, 2, 217, 94, 0, 0, 98, 158, 251, 2, 124, 214, 2, 0, 2, 159, 251, 2, 251, 34, 2, 0, 162, 159, 251, 2, 183, 35, 2, 0, 80, 162, 251, 2, 212, 35, 2, 0, 220, 162, 251, 2, 124, 95, 0, 0, 184, 166, 251, 2, 220, 36, 2, 0, 30, 167, 251, 2, 155, 95, 0, 0, 162, 167, 251, 2, 50, 37, 2, 0, 140, 168, 251, 2, 79, 37, 2, 0, 226, 168, 251, 2, 121, 37, 2, 0, 114, 169, 251, 2, 178, 37, 2, 0, 96, 170, 251, 2, 201, 37, 2, 0, 20, 171, 251, 2, 7, 96, 0, 0, 198, 171, 251, 2, 25, 215, 2, 0, 218, 172, 251, 2, 64, 96, 0, 0, 14, 173, 251, 2, 26, 57, 0, 0, 138, 173, 251, 2, 52, 38, 2, 0, 180, 173, 251, 2, 113, 38, 2, 0, 8, 175, 251, 2, 54, 215, 2, 0, 206, 175, 251, 2, 92, 4, 3, 0, 250, 175, 251, 2, 176, 96, 0, 0, 18, 176, 251, 2, 192, 38, 2, 0, 202, 176, 251, 2, 240, 96, 0, 0, 252, 177, 251, 2, 90, 57, 0, 0, 110, 178, 251, 2, 40, 39, 2, 0, 146, 178, 251, 2, 39, 97, 0, 0, 214, 179, 251, 2, 141, 39, 2, 0, 78, 180, 251, 2, 119, 215, 2, 0, 34, 181, 251, 2, 124, 4, 3, 0, 78, 181, 251, 2, 82, 97, 0, 0, 96, 181, 251, 2, 110, 97, 0, 0, 148, 181, 251, 2, 124, 57, 0, 0, 202, 181, 251, 2, 224, 39, 2, 0, 230, 181, 251, 2, 141, 215, 2, 0, 180, 182, 251, 2, 137, 97, 0, 0, 244, 182, 251, 2, 49, 40, 2, 0, 104, 183, 251, 2, 188, 97, 0, 0, 142, 184, 251, 2, 146, 40, 2, 0, 198, 184, 251, 2, 186, 215, 2, 0, 96, 185, 251, 2, 213, 97, 0, 0, 126, 185, 251, 2, 206, 40, 2, 0, 180, 185, 251, 2, 251, 40, 2, 0, 80, 186, 251, 2, 38, 41, 2, 0, 240, 186, 251, 2, 87, 41, 2, 0, 160, 187, 251, 2, 154, 41, 2, 0, 160, 188, 251, 2, 242, 41, 2, 0, 12, 190, 251, 2, 24, 42, 2, 0, 134, 190, 251, 2, 54, 42, 2, 0, 246, 190, 251, 2, 121, 98, 0, 0, 100, 192, 251, 2, 131, 42, 2, 0, 208, 192, 251, 2, 166, 98, 0, 0, 116, 193, 251, 2, 180, 42, 2, 0, 28, 194, 251, 2, 234, 98, 0, 0, 196, 194, 251, 2, 228, 42, 2, 0, 100, 195, 251, 2, 215, 4, 3, 0, 80, 196, 251, 2, 40, 99, 0, 0, 120, 196, 251, 2, 47, 43, 2, 0, 2, 197, 251, 2, 51, 216, 2, 0, 224, 197, 251, 2, 101, 99, 0, 0, 44, 198, 251, 2, 249, 57, 0, 0, 220, 198, 251, 2, 129, 43, 2, 0, 252, 198, 251, 2, 250, 4, 3, 0, 60, 200, 251, 2, 198, 99, 0, 0, 98, 200, 251, 2, 233, 43, 2, 0, 4, 201, 251, 2, 87, 191, 2, 0, 8, 202, 251, 2, 6, 100, 0, 0, 110, 202, 251, 2, 50, 100, 0, 0, 196, 202, 251, 2, 25, 58, 0, 0, 248, 202, 251, 2, 95, 44, 2, 0, 34, 203, 251, 2, 114, 216, 2, 0, 2, 204, 251, 2, 87, 100, 0, 0, 92, 204, 251, 2, 179, 44, 2, 0, 230, 204, 251, 2, 136, 100, 0, 0, 32, 206, 251, 2, 32, 45, 2, 0, 168, 206, 251, 2, 187, 100, 0, 0, 174, 207, 251, 2, 127, 45, 2, 0, 6, 208, 251, 2, 203, 45, 2, 0, 16, 209, 251, 2, 242, 100, 0, 0, 150, 209, 251, 2, 35, 46, 2, 0, 120, 210, 251, 2, 243, 46, 2, 0, 36, 213, 251, 2, 27, 47, 2, 0, 186, 213, 251, 2, 57, 47, 2, 0, 72, 214, 251, 2, 92, 101, 0, 0, 174, 214, 251, 2, 166, 47, 2, 0, 230, 215, 251, 2, 252, 48, 2, 0, 44, 221, 251, 2, 249, 101, 0, 0, 234, 222, 251, 2, 90, 49, 2, 0, 52, 223, 251, 2, 26, 102, 0, 0, 168, 223, 251, 2, 219, 58, 0, 0, 244, 223, 251, 2, 119, 49, 2, 0, 10, 224, 251, 2, 146, 49, 2, 0, 198, 224, 251, 2, 25, 172, 2, 0, 8, 225, 251, 2, 110, 217, 2, 0, 46, 225, 251, 2, 87, 102, 0, 0, 108, 225, 251, 2, 106, 102, 0, 0, 76, 226, 251, 2, 226, 49, 2, 0, 154, 226, 251, 2, 153, 217, 2, 0, 58, 227, 251, 2, 36, 50, 2, 0, 172, 227, 251, 2, 154, 102, 0, 0, 106, 228, 251, 2, 123, 50, 2, 0, 80, 229, 251, 2, 184, 102, 0, 0, 222, 229, 251, 2, 119, 51, 2, 0, 50, 234, 251, 2, 49, 103, 0, 0, 148, 235, 251, 2, 68, 103, 0, 0, 212, 235, 251, 2, 130, 103, 0, 0, 154, 236, 251, 2, 212, 51, 2, 0, 8, 237, 251, 2, 174, 103, 0, 0, 152, 237, 251, 2, 8, 52, 2, 0, 112, 238, 251, 2, 18, 104, 0, 0, 12, 239, 251, 2, 88, 59, 0, 0, 192, 239, 251, 2, 53, 52, 2, 0, 224, 239, 251, 2, 220, 5, 3, 0, 200, 240, 251, 2, 107, 104, 0, 0, 222, 240, 251, 2, 127, 52, 2, 0, 174, 241, 251, 2, 195, 104, 0, 0, 166, 242, 251, 2, 120, 59, 0, 0, 130, 243, 251, 2, 193, 52, 2, 0, 168, 243, 251, 2, 152, 218, 2, 0, 218, 244, 251, 2, 47, 105, 0, 0, 42, 245, 251, 2, 49, 53, 2, 0, 28, 246, 251, 2, 101, 53, 2, 0, 130, 246, 251, 2, 176, 172, 2, 0, 14, 247, 251, 2, 212, 192, 2, 0, 62, 247, 251, 2, 182, 218, 2, 0, 116, 247, 251, 2, 250, 5, 3, 0, 166, 247, 251, 2, 153, 105, 0, 0, 214, 247, 251, 2, 170, 53, 2, 0, 166, 248, 251, 2, 239, 192, 2, 0, 172, 249, 251, 2, 234, 105, 0, 0, 44, 250, 251, 2, 21, 54, 2, 0, 232, 250, 251, 2, 83, 54, 2, 0, 100, 251, 251, 2, 22, 6, 3, 0, 54, 252, 251, 2, 50, 106, 0, 0, 84, 252, 251, 2, 123, 54, 2, 0, 24, 253, 251, 2, 25, 193, 2, 0, 38, 254, 251, 2, 251, 218, 2, 0, 76, 254, 251, 2, 171, 106, 0, 0, 238, 255, 251, 2, 90, 55, 2, 0, 110, 2, 252, 2, 203, 106, 0, 0, 28, 3, 252, 2, 144, 55, 2, 0, 112, 3, 252, 2, 210, 55, 2, 0, 106, 4, 252, 2, 248, 106, 0, 0, 200, 4, 252, 2, 25, 56, 2, 0, 120, 5, 252, 2, 55, 56, 2, 0, 216, 5, 252, 2, 151, 56, 2, 0, 44, 7, 252, 2, 186, 56, 2, 0, 186, 7, 252, 2, 242, 56, 2, 0, 172, 8, 252, 2, 121, 57, 2, 0, 178, 10, 252, 2, 242, 57, 2, 0, 152, 12, 252, 2, 22, 58, 2, 0, 68, 13, 252, 2, 49, 58, 2, 0, 148, 13, 252, 2, 187, 58, 2, 0, 212, 15, 252, 2, 56, 59, 2, 0, 134, 17, 252, 2, 92, 59, 2, 0, 250, 17, 252, 2, 113, 59, 2, 0, 70, 18, 252, 2, 136, 59, 2, 0, 158, 18, 252, 2, 168, 59, 2, 0, 14, 19, 252, 2, 248, 59, 2, 0, 14, 20, 252, 2, 26, 60, 2, 0, 114, 20, 252, 2, 74, 108, 0, 0, 66, 22, 252, 2, 101, 108, 0, 0, 184, 22, 252, 2, 142, 60, 2, 0, 92, 23, 252, 2, 171, 108, 0, 0, 200, 23, 252, 2, 219, 108, 0, 0, 40, 24, 252, 2, 178, 60, 2, 0, 132, 24, 252, 2, 23, 109, 0, 0, 100, 25, 252, 2, 214, 60, 0, 0, 226, 25, 252, 2, 228, 60, 2, 0, 254, 25, 252, 2, 86, 109, 0, 0, 196, 26, 252, 2, 154, 109, 0, 0, 76, 27, 252, 2, 30, 61, 2, 0, 140, 27, 252, 2, 23, 220, 2, 0, 124, 28, 252, 2, 170, 109, 0, 0, 196, 28, 252, 2, 244, 60, 0, 0, 164, 29, 252, 2, 119, 61, 2, 0, 218, 29, 252, 2, 219, 61, 2, 0, 162, 30, 252, 2, 245, 193, 2, 0, 254, 30, 252, 2, 27, 110, 0, 0, 122, 31, 252, 2, 117, 110, 0, 0, 46, 32, 252, 2, 15, 61, 0, 0, 100, 32, 252, 2, 252, 61, 2, 0, 144, 32, 252, 2, 11, 62, 2, 0, 172, 32, 252, 2, 151, 173, 2, 0, 82, 33, 252, 2, 12, 7, 3, 0, 216, 33, 252, 2, 141, 110, 0, 0, 12, 34, 252, 2, 93, 62, 2, 0, 240, 34, 252, 2, 22, 194, 2, 0, 6, 36, 252, 2, 235, 110, 0, 0, 104, 36, 252, 2, 20, 111, 0, 0, 182, 36, 252, 2, 60, 61, 0, 0, 42, 37, 252, 2, 205, 62, 2, 0, 70, 37, 252, 2, 245, 62, 2, 0, 148, 37, 252, 2, 118, 220, 2, 0, 98, 38, 252, 2, 78, 111, 0, 0, 194, 38, 252, 2, 67, 63, 2, 0, 140, 39, 252, 2, 50, 194, 2, 0, 180, 40, 252, 2, 146, 220, 2, 0, 224, 40, 252, 2, 197, 63, 2, 0, 198, 41, 252, 2, 215, 111, 0, 0, 230, 42, 252, 2, 35, 64, 2, 0, 72, 43, 252, 2, 188, 220, 2, 0, 232, 43, 252, 2, 249, 111, 0, 0, 36, 44, 252, 2, 97, 64, 2, 0, 108, 44, 252, 2, 21, 112, 0, 0, 82, 45, 252, 2, 179, 64, 2, 0, 154, 45, 252, 2, 220, 220, 2, 0, 32, 46, 252, 2, 232, 64, 2, 0, 122, 46, 252, 2, 14, 65, 2, 0, 18, 47, 252, 2, 54, 65, 2, 0, 150, 47, 252, 2, 188, 65, 2, 0, 72, 50, 252, 2, 196, 112, 0, 0, 232, 50, 252, 2, 239, 65, 2, 0, 76, 51, 252, 2, 27, 221, 2, 0, 214, 51, 252, 2, 239, 112, 0, 0, 28, 52, 252, 2, 98, 66, 2, 0, 214, 53, 252, 2, 51, 174, 2, 0, 120, 54, 252, 2, 151, 194, 2, 0, 154, 54, 252, 2, 56, 221, 2, 0, 190, 54, 252, 2, 65, 113, 0, 0, 250, 54, 252, 2, 178, 66, 2, 0, 144, 55, 252, 2, 244, 66, 2, 0, 20, 56, 252, 2, 186, 7, 3, 0, 156, 56, 252, 2, 123, 113, 0, 0, 184, 56, 252, 2, 7, 67, 2, 0, 20, 57, 252, 2, 51, 67, 2, 0, 106, 57, 252, 2, 154, 113, 0, 0, 18, 58, 252, 2, 80, 67, 2, 0, 96, 58, 252, 2, 182, 113, 0, 0, 58, 59, 252, 2, 245, 67, 2, 0, 234, 60, 252, 2, 248, 113, 0, 0, 140, 61, 252, 2, 82, 68, 2, 0, 78, 62, 252, 2, 123, 69, 2, 0, 92, 66, 252, 2, 28, 70, 2, 0, 178, 68, 252, 2, 56, 70, 2, 0, 30, 69, 252, 2, 91, 70, 2, 0, 192, 69, 252, 2, 111, 70, 2, 0, 18, 70, 252, 2, 185, 114, 0, 0, 130, 72, 252, 2, 47, 71, 2, 0, 172, 72, 252, 2, 79, 71, 2, 0, 66, 73, 252, 2, 111, 71, 2, 0, 214, 73, 252, 2, 243, 114, 0, 0, 80, 74, 252, 2, 220, 71, 2, 0, 230, 75, 252, 2, 246, 71, 2, 0, 26, 76, 252, 2, 13, 72, 2, 0, 160, 76, 252, 2, 86, 115, 0, 0, 162, 77, 252, 2, 99, 72, 2, 0, 204, 77, 252, 2, 140, 72, 2, 0, 76, 78, 252, 2, 180, 72, 2, 0, 202, 78, 252, 2, 156, 115, 0, 0, 230, 79, 252, 2, 249, 72, 2, 0, 36, 80, 252, 2, 178, 115, 0, 0, 128, 80, 252, 2, 15, 73, 2, 0, 208, 80, 252, 2, 60, 222, 2, 0, 48, 81, 252, 2, 210, 115, 0, 0, 70, 81, 252, 2, 53, 73, 2, 0, 160, 81, 252, 2, 248, 115, 0, 0, 42, 82, 252, 2, 86, 73, 2, 0, 112, 82, 252, 2, 123, 73, 2, 0, 86, 83, 252, 2, 175, 73, 2, 0, 112, 84, 252, 2, 234, 73, 2, 0, 104, 85, 252, 2, 123, 195, 2, 0, 216, 85, 252, 2, 121, 116, 0, 0, 248, 85, 252, 2, 20, 74, 2, 0, 58, 86, 252, 2, 66, 74, 2, 0, 2, 87, 252, 2, 107, 74, 2, 0, 196, 87, 252, 2, 139, 74, 2, 0, 76, 88, 252, 2, 91, 75, 2, 0, 108, 91, 252, 2, 121, 75, 2, 0, 228, 91, 252, 2, 41, 76, 2, 0, 188, 94, 252, 2, 123, 76, 2, 0, 28, 96, 252, 2, 145, 76, 2, 0, 126, 96, 252, 2, 183, 76, 2, 0, 6, 97, 252, 2, 220, 76, 2, 0, 148, 97, 252, 2, 55, 77, 2, 0, 50, 99, 252, 2, 187, 117, 0, 0, 168, 99, 252, 2, 87, 77, 2, 0, 224, 99, 252, 2, 119, 77, 2, 0, 106, 100, 252, 2, 215, 117, 0, 0, 192, 100, 252, 2, 236, 117, 0, 0, 130, 101, 252, 2, 194, 77, 2, 0, 200, 101, 252, 2, 251, 77, 2, 0, 152, 102, 252, 2, 42, 78, 2, 0, 86, 103, 252, 2, 186, 78, 2, 0, 72, 105, 252, 2, 213, 78, 2, 0, 152, 105, 252, 2, 88, 79, 2, 0, 124, 107, 252, 2, 121, 79, 2, 0, 18, 108, 252, 2, 147, 80, 2, 0, 110, 112, 252, 2, 246, 118, 0, 0, 252, 113, 252, 2, 3, 81, 2, 0, 66, 114, 252, 2, 14, 119, 0, 0, 192, 114, 252, 2, 49, 81, 2, 0, 18, 115, 252, 2, 90, 81, 2, 0, 208, 115, 252, 2, 138, 81, 2, 0, 168, 116, 252, 2, 201, 81, 2, 0, 156, 117, 252, 2, 110, 119, 0, 0, 72, 118, 252, 2, 2, 82, 2, 0, 148, 118, 252, 2, 60, 82, 2, 0, 102, 119, 252, 2, 148, 119, 0, 0, 4, 120, 252, 2, 116, 82, 2, 0, 54, 120, 252, 2, 136, 82, 2, 0, 92, 120, 252, 2, 170, 82, 2, 0, 4, 121, 252, 2, 216, 82, 2, 0, 158, 121, 252, 2, 26, 83, 2, 0, 106, 122, 252, 2, 185, 83, 2, 0, 108, 124, 252, 2, 216, 83, 2, 0, 194, 124, 252, 2, 41, 84, 2, 0, 38, 126, 252, 2, 70, 84, 2, 0, 230, 126, 252, 2, 115, 84, 2, 0, 180, 127, 252, 2, 92, 120, 0, 0, 80, 128, 252, 2, 170, 84, 2, 0, 150, 128, 252, 2, 188, 9, 3, 0, 254, 128, 252, 2, 183, 64, 0, 0, 94, 129, 252, 2, 202, 84, 2, 0, 128, 129, 252, 2, 250, 84, 2, 0, 98, 130, 252, 2, 45, 85, 2, 0, 72, 131, 252, 2, 76, 85, 2, 0, 222, 131, 252, 2, 59, 224, 2, 0, 58, 132, 252, 2, 239, 120, 0, 0, 92, 132, 252, 2, 112, 85, 2, 0, 156, 132, 252, 2, 151, 85, 2, 0, 74, 133, 252, 2, 187, 85, 2, 0, 242, 133, 252, 2, 21, 86, 2, 0, 64, 135, 252, 2, 43, 86, 2, 0, 176, 135, 252, 2, 78, 86, 2, 0, 62, 136, 252, 2, 107, 86, 2, 0, 210, 136, 252, 2, 121, 121, 0, 0, 64, 137, 252, 2, 136, 86, 2, 0, 106, 137, 252, 2, 166, 86, 2, 0, 26, 138, 252, 2, 154, 121, 0, 0, 122, 138, 252, 2, 216, 86, 2, 0, 232, 138, 252, 2, 216, 121, 0, 0, 134, 141, 252, 2, 58, 65, 0, 0, 58, 142, 252, 2, 152, 87, 2, 0, 92, 142, 252, 2, 178, 87, 2, 0, 230, 142, 252, 2, 15, 122, 0, 0, 60, 143, 252, 2, 205, 87, 2, 0, 130, 143, 252, 2, 52, 122, 0, 0, 212, 144, 252, 2, 49, 88, 2, 0, 12, 145, 252, 2, 92, 88, 2, 0, 180, 145, 252, 2, 138, 88, 2, 0, 84, 146, 252, 2, 180, 88, 2, 0, 220, 146, 252, 2, 124, 122, 0, 0, 84, 148, 252, 2, 52, 89, 2, 0, 108, 148, 252, 2, 118, 89, 2, 0, 108, 149, 252, 2, 155, 89, 2, 0, 238, 149, 252, 2, 184, 89, 2, 0, 86, 150, 252, 2, 59, 90, 2, 0, 44, 152, 252, 2, 152, 90, 2, 0, 186, 153, 252, 2, 252, 122, 0, 0, 176, 154, 252, 2, 239, 90, 2, 0, 32, 155, 252, 2, 22, 123, 0, 0, 128, 155, 252, 2, 12, 91, 2, 0, 228, 155, 252, 2, 80, 91, 2, 0, 130, 157, 252, 2, 129, 123, 0, 0, 48, 158, 252, 2, 141, 91, 2, 0, 162, 158, 252, 2, 14, 66, 0, 0, 192, 159, 252, 2, 209, 91, 2, 0, 248, 159, 252, 2, 212, 123, 0, 0, 220, 160, 252, 2, 24, 92, 2, 0, 54, 161, 252, 2, 240, 123, 0, 0, 40, 162, 252, 2, 56, 66, 0, 0, 116, 162, 252, 2, 117, 92, 2, 0, 148, 162, 252, 2, 118, 225, 2, 0, 80, 163, 252, 2, 24, 124, 0, 0, 120, 163, 252, 2, 198, 92, 2, 0, 214, 163, 252, 2, 33, 93, 2, 0, 246, 164, 252, 2, 92, 93, 2, 0, 188, 165, 252, 2, 133, 93, 2, 0, 66, 166, 252, 2, 181, 93, 2, 0, 248, 166, 252, 2, 212, 93, 2, 0, 106, 167, 252, 2, 252, 93, 2, 0, 210, 167, 252, 2, 90, 94, 2, 0, 194, 169, 252, 2, 114, 94, 2, 0, 68, 170, 252, 2, 193, 94, 2, 0, 114, 171, 252, 2, 23, 95, 2, 0, 234, 172, 252, 2, 54, 95, 2, 0, 84, 173, 252, 2, 10, 125, 0, 0, 178, 174, 252, 2, 186, 66, 0, 0, 238, 174, 252, 2, 40, 125, 0, 0, 116, 175, 252, 2, 195, 95, 2, 0, 216, 175, 252, 2, 92, 125, 0, 0, 92, 176, 252, 2, 230, 95, 2, 0, 168, 176, 252, 2, 120, 125, 0, 0, 48, 177, 252, 2, 18, 96, 2, 0, 138, 177, 252, 2, 156, 125, 0, 0, 36, 178, 252, 2, 70, 96, 2, 0, 172, 178, 252, 2, 214, 125, 0, 0, 108, 179, 252, 2, 243, 66, 0, 0, 208, 179, 252, 2, 128, 96, 2, 0, 244, 179, 252, 2, 8, 126, 0, 0, 178, 180, 252, 2, 41, 126, 0, 0, 182, 181, 252, 2, 9, 97, 2, 0, 20, 182, 252, 2, 57, 226, 2, 0, 146, 182, 252, 2, 78, 126, 0, 0, 174, 182, 252, 2, 60, 97, 2, 0, 238, 182, 252, 2, 117, 97, 2, 0, 192, 183, 252, 2, 123, 126, 0, 0, 46, 184, 252, 2, 156, 97, 2, 0, 74, 184, 252, 2, 212, 97, 2, 0, 24, 185, 252, 2, 25, 127, 0, 0, 12, 188, 252, 2, 142, 98, 2, 0, 136, 190, 252, 2, 188, 98, 2, 0, 20, 191, 252, 2, 87, 99, 2, 0, 88, 193, 252, 2, 121, 99, 2, 0, 212, 193, 252, 2, 183, 99, 2, 0, 200, 194, 252, 2, 56, 100, 2, 0, 168, 196, 252, 2, 88, 100, 2, 0, 44, 197, 252, 2, 60, 101, 2, 0, 154, 200, 252, 2, 117, 101, 2, 0, 106, 201, 252, 2, 188, 101, 2, 0, 110, 202, 252, 2, 88, 102, 2, 0, 242, 204, 252, 2, 194, 128, 0, 0, 64, 205, 252, 2, 106, 102, 2, 0, 172, 205, 252, 2, 237, 128, 0, 0, 64, 206, 252, 2, 21, 129, 0, 0, 56, 207, 252, 2, 212, 102, 2, 0, 154, 207, 252, 2, 57, 129, 0, 0, 44, 208, 252, 2, 252, 67, 0, 0, 114, 208, 252, 2, 6, 103, 2, 0, 148, 208, 252, 2, 52, 103, 2, 0, 238, 208, 252, 2, 91, 129, 0, 0, 118, 209, 252, 2, 90, 103, 2, 0, 220, 209, 252, 2, 28, 68, 0, 0, 182, 210, 252, 2, 153, 103, 2, 0, 212, 210, 252, 2, 146, 129, 0, 0, 144, 211, 252, 2, 3, 104, 2, 0, 144, 212, 252, 2, 183, 129, 0, 0, 62, 213, 252, 2, 88, 104, 2, 0, 158, 213, 252, 2, 112, 104, 2, 0, 252, 213, 252, 2, 138, 104, 2, 0, 88, 214, 252, 2, 123, 105, 2, 0, 150, 217, 252, 2, 145, 106, 2, 0, 198, 221, 252, 2, 152, 130, 0, 0, 182, 223, 252, 2, 154, 68, 0, 0, 40, 224, 252, 2, 19, 107, 2, 0, 64, 224, 252, 2, 208, 130, 0, 0, 180, 224, 252, 2, 55, 107, 2, 0, 84, 225, 252, 2, 84, 131, 0, 0, 100, 226, 252, 2, 109, 107, 2, 0, 194, 226, 252, 2, 112, 131, 0, 0, 184, 227, 252, 2, 182, 107, 2, 0, 126, 228, 252, 2, 242, 107, 2, 0, 244, 228, 252, 2, 23, 108, 2, 0, 62, 229, 252, 2, 185, 227, 2, 0, 164, 229, 252, 2, 190, 131, 0, 0, 246, 229, 252, 2, 215, 68, 0, 0, 212, 230, 252, 2, 42, 108, 2, 0, 8, 231, 252, 2, 59, 199, 2, 0, 90, 232, 252, 2, 45, 132, 0, 0, 198, 232, 252, 2, 88, 132, 0, 0, 28, 233, 252, 2, 193, 108, 2, 0, 170, 233, 252, 2, 55, 109, 2, 0, 150, 234, 252, 2, 144, 132, 0, 0, 144, 235, 252, 2, 185, 132, 0, 0, 224, 235, 252, 2, 252, 68, 0, 0, 60, 236, 252, 2, 97, 109, 2, 0, 100, 236, 252, 2, 54, 178, 2, 0, 162, 237, 252, 2, 90, 199, 2, 0, 200, 237, 252, 2, 231, 132, 0, 0, 76, 238, 252, 2, 14, 69, 0, 0, 248, 238, 252, 2, 62, 110, 2, 0, 180, 239, 252, 2, 117, 199, 2, 0, 112, 240, 252, 2, 45, 69, 0, 0, 78, 241, 252, 2, 134, 110, 2, 0, 128, 241, 252, 2, 200, 110, 2, 0, 4, 242, 252, 2, 57, 228, 2, 0, 240, 242, 252, 2, 118, 133, 0, 0, 66, 243, 252, 2, 41, 111, 2, 0, 208, 243, 252, 2, 110, 111, 2, 0, 90, 244, 252, 2, 154, 199, 2, 0, 246, 244, 252, 2, 90, 228, 2, 0, 30, 245, 252, 2, 175, 133, 0, 0, 124, 245, 252, 2, 170, 111, 2, 0, 224, 245, 252, 2, 249, 111, 2, 0, 118, 246, 252, 2, 117, 178, 2, 0, 196, 246, 252, 2, 124, 228, 2, 0, 10, 247, 252, 2, 212, 133, 0, 0, 68, 247, 252, 2, 32, 112, 2, 0, 156, 247, 252, 2, 151, 228, 2, 0, 130, 248, 252, 2, 246, 133, 0, 0, 182, 248, 252, 2, 123, 112, 2, 0, 16, 249, 252, 2, 207, 112, 2, 0, 52, 250, 252, 2, 184, 228, 2, 0, 196, 250, 252, 2, 13, 113, 2, 0, 246, 250, 252, 2, 54, 113, 2, 0, 138, 251, 252, 2, 86, 113, 2, 0, 234, 251, 252, 2, 122, 113, 2, 0, 80, 252, 252, 2, 220, 113, 2, 0, 230, 253, 252, 2, 115, 134, 0, 0, 200, 254, 252, 2, 137, 134, 0, 0, 54, 255, 252, 2, 59, 114, 2, 0, 130, 255, 252, 2, 173, 134, 0, 0, 250, 255, 252, 2, 215, 134, 0, 0, 242, 2, 253, 2, 143, 114, 2, 0, 66, 3, 253, 2, 175, 114, 2, 0, 126, 3, 253, 2, 245, 134, 0, 0, 230, 3, 253, 2, 24, 135, 0, 0, 250, 4, 253, 2, 44, 135, 0, 0, 34, 5, 253, 2, 49, 115, 2, 0, 212, 5, 253, 2, 76, 135, 0, 0, 102, 6, 253, 2, 82, 115, 2, 0, 232, 6, 253, 2, 129, 135, 0, 0, 186, 7, 253, 2, 161, 115, 2, 0, 32, 8, 253, 2, 200, 115, 2, 0, 108, 8, 253, 2, 170, 135, 0, 0, 14, 9, 253, 2, 248, 115, 2, 0, 132, 9, 253, 2, 49, 116, 2, 0, 244, 9, 253, 2, 68, 116, 2, 0, 148, 10, 253, 2, 142, 116, 2, 0, 146, 11, 253, 2, 201, 116, 2, 0, 76, 12, 253, 2, 26, 136, 0, 0, 190, 12, 253, 2, 241, 116, 2, 0, 220, 12, 253, 2, 91, 117, 2, 0, 20, 14, 253, 2, 109, 136, 0, 0, 150, 16, 253, 2, 136, 136, 0, 0, 48, 17, 253, 2, 177, 136, 0, 0, 2, 18, 253, 2, 202, 136, 0, 0, 182, 18, 253, 2, 128, 118, 2, 0, 246, 18, 253, 2, 231, 136, 0, 0, 118, 19, 253, 2, 188, 118, 2, 0, 2, 20, 253, 2, 207, 118, 2, 0, 158, 20, 253, 2, 248, 118, 2, 0, 76, 21, 253, 2, 53, 137, 0, 0, 190, 21, 253, 2, 66, 119, 2, 0, 144, 22, 253, 2, 113, 119, 2, 0, 40, 23, 253, 2, 152, 119, 2, 0, 152, 23, 253, 2, 28, 120, 2, 0, 140, 25, 253, 2, 59, 120, 2, 0, 244, 25, 253, 2, 187, 120, 2, 0, 36, 28, 253, 2, 244, 120, 2, 0, 8, 29, 253, 2, 58, 121, 2, 0, 196, 29, 253, 2, 27, 138, 0, 0, 168, 30, 253, 2, 179, 70, 0, 0, 218, 30, 253, 2, 110, 121, 2, 0, 252, 30, 253, 2, 97, 138, 0, 0, 88, 32, 253, 2, 195, 121, 2, 0, 194, 32, 253, 2, 244, 121, 2, 0, 190, 33, 253, 2, 175, 138, 0, 0, 70, 34, 253, 2, 219, 138, 0, 0, 100, 35, 253, 2, 234, 138, 0, 0, 128, 35, 253, 2, 99, 122, 2, 0, 190, 35, 253, 2, 152, 122, 2, 0, 168, 36, 253, 2, 186, 200, 2, 0, 36, 37, 253, 2, 210, 122, 2, 0, 144, 37, 253, 2, 92, 123, 2, 0, 200, 39, 253, 2, 120, 123, 2, 0, 38, 40, 253, 2, 211, 139, 0, 0, 144, 41, 253, 2, 247, 139, 0, 0, 54, 42, 253, 2, 88, 124, 2, 0, 58, 46, 253, 2, 217, 124, 2, 0, 234, 47, 253, 2, 51, 125, 2, 0, 30, 49, 253, 2, 99, 125, 2, 0, 32, 50, 253, 2, 217, 140, 0, 0, 226, 50, 253, 2, 155, 125, 2, 0, 32, 51, 253, 2, 177, 125, 2, 0, 146, 51, 253, 2, 252, 125, 2, 0, 160, 52, 253, 2, 22, 126, 2, 0, 250, 52, 253, 2, 60, 141, 0, 0, 36, 54, 253, 2, 182, 126, 2, 0, 38, 56, 253, 2, 208, 126, 2, 0, 136, 56, 253, 2, 239, 126, 2, 0, 244, 56, 253, 2, 155, 141, 0, 0, 72, 57, 253, 2, 40, 127, 2, 0, 182, 57, 253, 2, 118, 127, 2, 0, 166, 58, 253, 2, 210, 127, 2, 0, 228, 59, 253, 2, 242, 127, 2, 0, 156, 60, 253, 2, 21, 128, 2, 0, 72, 61, 253, 2, 252, 141, 0, 0, 200, 61, 253, 2, 64, 128, 2, 0, 2, 62, 253, 2, 15, 142, 0, 0, 144, 62, 253, 2, 117, 128, 2, 0, 242, 62, 253, 2, 59, 142, 0, 0, 156, 63, 253, 2, 160, 128, 2, 0, 202, 63, 253, 2, 188, 201, 2, 0, 60, 64, 253, 2, 26, 14, 3, 0, 110, 64, 253, 2, 84, 142, 0, 0, 64, 65, 253, 2, 5, 129, 2, 0, 120, 65, 253, 2, 129, 129, 2, 0, 20, 67, 253, 2, 214, 129, 2, 0, 48, 68, 253, 2, 24, 130, 2, 0, 8, 69, 253, 2, 56, 130, 2, 0, 112, 69, 253, 2, 214, 142, 0, 0, 46, 71, 253, 2, 179, 130, 2, 0, 92, 71, 253, 2, 206, 130, 2, 0, 224, 71, 253, 2, 22, 143, 0, 0, 6, 73, 253, 2, 32, 131, 2, 0, 60, 73, 253, 2, 71, 131, 2, 0, 220, 73, 253, 2, 105, 131, 2, 0, 86, 74, 253, 2, 138, 131, 2, 0, 202, 74, 253, 2, 178, 131, 2, 0, 100, 75, 253, 2, 124, 14, 3, 0, 212, 76, 253, 2, 124, 143, 0, 0, 228, 76, 253, 2, 130, 132, 2, 0, 222, 79, 253, 2, 244, 143, 0, 0, 70, 80, 253, 2, 165, 132, 2, 0, 130, 80, 253, 2, 11, 144, 0, 0, 214, 80, 253, 2, 197, 132, 2, 0, 24, 81, 253, 2, 235, 132, 2, 0, 188, 81, 253, 2, 150, 231, 2, 0, 62, 82, 253, 2, 60, 144, 0, 0, 116, 82, 253, 2, 33, 133, 2, 0, 182, 82, 253, 2, 88, 144, 0, 0, 112, 83, 253, 2, 134, 133, 2, 0, 54, 84, 253, 2, 180, 133, 2, 0, 238, 84, 253, 2, 212, 133, 2, 0, 46, 85, 253, 2, 229, 133, 2, 0, 154, 85, 253, 2, 92, 202, 2, 0, 90, 86, 253, 2, 118, 134, 2, 0, 206, 87, 253, 2, 147, 134, 2, 0, 72, 88, 253, 2, 188, 144, 0, 0, 160, 88, 253, 2, 175, 134, 2, 0, 222, 88, 253, 2, 212, 144, 0, 0, 46, 89, 253, 2, 203, 134, 2, 0, 100, 89, 253, 2, 235, 134, 2, 0, 240, 89, 253, 2, 249, 144, 0, 0, 94, 90, 253, 2, 24, 135, 2, 0, 152, 90, 253, 2, 50, 135, 2, 0, 16, 91, 253, 2, 24, 145, 0, 0, 86, 91, 253, 2, 75, 135, 2, 0, 124, 91, 253, 2, 122, 135, 2, 0, 20, 92, 253, 2, 145, 135, 2, 0, 90, 92, 253, 2, 186, 135, 2, 0, 224, 92, 253, 2, 244, 135, 2, 0, 234, 93, 253, 2, 11, 136, 2, 0, 70, 94, 253, 2, 55, 136, 2, 0, 6, 95, 253, 2, 76, 136, 2, 0, 106, 95, 253, 2, 105, 136, 2, 0, 226, 95, 253, 2, 152, 145, 0, 0, 56, 96, 253, 2, 214, 136, 2, 0, 98, 97, 253, 2, 74, 137, 2, 0, 228, 99, 253, 2, 45, 146, 0, 0, 66, 100, 253, 2, 144, 146, 0, 0, 112, 101, 253, 2, 178, 146, 0, 0, 80, 102, 253, 2, 172, 137, 2, 0, 236, 102, 253, 2, 247, 146, 0, 0, 180, 103, 253, 2, 41, 147, 0, 0, 24, 104, 253, 2, 230, 137, 2, 0, 98, 104, 253, 2, 73, 147, 0, 0, 96, 105, 253, 2, 47, 138, 2, 0, 234, 105, 253, 2, 153, 232, 2, 0, 184, 106, 253, 2, 136, 147, 0, 0, 228, 106, 253, 2, 135, 138, 2, 0, 108, 107, 253, 2, 192, 147, 0, 0, 68, 108, 253, 2, 207, 138, 2, 0, 196, 108, 253, 2, 250, 147, 0, 0, 122, 109, 253, 2, 22, 139, 2, 0, 250, 109, 253, 2, 85, 139, 2, 0, 24, 111, 253, 2, 146, 139, 2, 0, 252, 111, 253, 2, 85, 148, 0, 0, 114, 112, 253, 2, 184, 139, 2, 0, 162, 112, 253, 2, 250, 139, 2, 0, 122, 113, 253, 2, 152, 148, 0, 0, 160, 114, 253, 2, 92, 15, 3, 0, 6, 115, 253, 2, 176, 148, 0, 0, 24, 115, 253, 2, 207, 148, 0, 0, 162, 115, 253, 2, 248, 148, 0, 0, 50, 116, 253, 2, 22, 149, 0, 0, 170, 116, 253, 2, 50, 149, 0, 0, 42, 117, 253, 2, 148, 15, 3, 0, 114, 117, 253, 2, 90, 140, 2, 0, 220, 118, 253, 2, 184, 140, 2, 0, 226, 119, 253, 2, 215, 140, 2, 0, 118, 120, 253, 2, 243, 140, 2, 0, 214, 120, 253, 2, 13, 141, 2, 0, 68, 121, 253, 2, 39, 141, 2, 0, 154, 121, 253, 2, 72, 141, 2, 0, 24, 122, 253, 2, 156, 141, 2, 0, 66, 123, 253, 2, 26, 142, 2, 0, 4, 126, 253, 2, 55, 150, 0, 0, 54, 126, 253, 2, 91, 150, 0, 0, 36, 127, 253, 2, 88, 142, 2, 0, 80, 127, 253, 2, 117, 142, 2, 0, 222, 127, 253, 2, 123, 150, 0, 0, 100, 128, 253, 2, 192, 142, 2, 0, 44, 129, 253, 2, 238, 142, 2, 0, 192, 129, 253, 2, 10, 143, 2, 0, 48, 130, 253, 2, 178, 143, 2, 0, 166, 132, 253, 2, 23, 144, 2, 0, 182, 133, 253, 2, 52, 144, 2, 0, 8, 134, 253, 2, 92, 144, 2, 0, 108, 134, 253, 2, 183, 144, 2, 0, 208, 135, 253, 2, 210, 144, 2, 0, 72, 136, 253, 2, 156, 233, 2, 0, 40, 137, 253, 2, 23, 145, 2, 0, 86, 137, 253, 2, 59, 145, 2, 0, 202, 137, 253, 2, 124, 145, 2, 0, 148, 138, 253, 2, 145, 145, 2, 0, 222, 138, 253, 2, 122, 151, 0, 0, 24, 142, 253, 2, 120, 146, 2, 0, 62, 142, 253, 2, 144, 146, 2, 0, 148, 142, 253, 2, 3, 147, 2, 0, 50, 144, 253, 2, 83, 147, 2, 0, 34, 145, 253, 2, 152, 147, 2, 0, 242, 145, 253, 2, 184, 147, 2, 0, 84, 146, 253, 2, 89, 148, 2, 0, 224, 148, 253, 2, 112, 148, 2, 0, 58, 149, 253, 2, 27, 152, 0, 0, 134, 149, 253, 2, 143, 148, 2, 0, 174, 149, 253, 2, 174, 148, 2, 0, 50, 150, 253, 2, 25, 149, 2, 0, 188, 151, 253, 2, 51, 149, 2, 0, 22, 152, 253, 2, 124, 152, 0, 0, 90, 153, 253, 2, 180, 149, 2, 0, 188, 154, 253, 2, 211, 149, 2, 0, 36, 155, 253, 2, 232, 149, 2, 0, 102, 155, 253, 2, 28, 150, 2, 0, 18, 156, 253, 2, 56, 150, 2, 0, 118, 156, 253, 2, 149, 150, 2, 0, 32, 158, 253, 2, 168, 150, 2, 0, 126, 158, 253, 2, 45, 151, 2, 0, 108, 160, 253, 2, 56, 153, 0, 0, 226, 160, 253, 2, 89, 151, 2, 0, 6, 161, 253, 2, 211, 151, 2, 0, 144, 162, 253, 2, 182, 153, 0, 0, 30, 166, 253, 2, 141, 152, 2, 0, 210, 166, 253, 2, 174, 152, 2, 0, 92, 167, 253, 2, 244, 153, 0, 0, 200, 167, 253, 2, 214, 152, 2, 0, 242, 167, 253, 2, 237, 152, 2, 0, 108, 168, 253, 2, 20, 154, 0, 0, 220, 168, 253, 2, 57, 154, 0, 0, 32, 170, 253, 2, 28, 154, 2, 0, 220, 173, 253, 2, 87, 154, 2, 0, 178, 174, 253, 2, 17, 155, 2, 0, 70, 177, 253, 2, 48, 155, 2, 0, 162, 177, 253, 2, 122, 155, 2, 0, 168, 178, 253, 2, 150, 155, 2, 0, 2, 179, 253, 2, 219, 155, 2, 0, 186, 179, 253, 2, 86, 156, 2, 0, 68, 181, 253, 2, 186, 156, 2, 0, 94, 182, 253, 2, 245, 156, 2, 0, 20, 183, 253, 2, 28, 157, 2, 0, 132, 183, 253, 2, 122, 155, 0, 0, 250, 184, 253, 2, 184, 157, 2, 0, 36, 186, 253, 2, 187, 155, 0, 0, 124, 186, 253, 2, 205, 157, 2, 0, 196, 186, 253, 2, 212, 155, 0, 0, 68, 187, 253, 2, 89, 76, 0, 0, 138, 187, 253, 2, 247, 157, 2, 0, 174, 187, 253, 2, 113, 158, 2, 0, 126, 189, 253, 2, 49, 156, 0, 0, 28, 190, 253, 2, 173, 158, 2, 0, 102, 190, 253, 2, 92, 156, 0, 0, 142, 191, 253, 2, 2, 159, 2, 0, 184, 191, 253, 2, 59, 159, 2, 0, 120, 192, 253, 2, 90, 159, 2, 0, 216, 192, 253, 2, 218, 17, 3, 0, 204, 193, 253, 2, 247, 156, 0, 0, 106, 196, 253, 2, 176, 159, 2, 0, 178, 196, 253, 2, 15, 157, 0, 0, 46, 197, 253, 2, 185, 76, 0, 0, 106, 197, 253, 2, 220, 159, 2, 0, 142, 197, 253, 2, 45, 157, 0, 0, 42, 198, 253, 2, 18, 160, 2, 0, 120, 198, 253, 2, 74, 157, 0, 0, 14, 199, 253, 2, 74, 160, 2, 0, 84, 199, 253, 2, 102, 157, 0, 0, 216, 199, 253, 2, 119, 160, 2, 0, 66, 200, 253, 2, 60, 235, 2, 0, 220, 200, 253, 2, 146, 157, 0, 0, 246, 200, 253, 2, 182, 160, 2, 0, 66, 201, 253, 2, 172, 157, 0, 0, 234, 201, 253, 2, 250, 160, 2, 0, 62, 202, 253, 2, 45, 161, 2, 0, 4, 203, 253, 2, 111, 161, 2, 0, 232, 203, 253, 2, 174, 161, 2, 0, 170, 204, 253, 2, 24, 162, 2, 0, 218, 205, 253, 2, 155, 162, 2, 0, 4, 210, 253, 2, 180, 162, 2, 0, 84, 210, 253, 2, 60, 163, 2, 0, 252, 211, 253, 2, 28, 164, 2, 0, 24, 215, 253, 2, 152, 164, 2, 0, 164, 216, 253, 2, 219, 165, 2, 0, 36, 221, 253, 2, 89, 159, 0, 0, 68, 221, 253, 2, 247, 165, 2, 0, 158, 221, 253, 2, 14, 166, 2, 0, 228, 221, 253, 2, 57, 166, 2, 0, 104, 222, 253, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 2, 2, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 10, 2, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 32, 2, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 48, 6, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 74, 8, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 96, 8, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 180, 10, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 40, 11, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 72, 11, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 74, 28, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 80, 28, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 114, 42, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 180, 47, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 52, 48, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 2, 57, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 28, 59, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 90, 83, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 186, 85, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 232, 119, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 86, 222, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 86, 233, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 90, 233, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 114, 235, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 60, 239, 251, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 230, 21, 252, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 142, 48, 252, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 116, 69, 252, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 8, 112, 252, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 240, 134, 252, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 158, 135, 252, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 50, 136, 252, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 222, 215, 252, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 152, 216, 252, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 218, 48, 253, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 8, 50, 253, 0, 23, 0, 5, 0, 0, 144, 9, 0, 23, 0, 5, 0, 0, 142, 9, 0, 16, 0, 5, 0, 242, 98, 253, 0, 23, 0, 5, 0, 0, 144, 9, 0, 16, 0, 5, 0, 0, 82, 10, 0, 16, 0, 5, 0, 32, 2, 251, 0, 16, 0, 5, 0, 0, 84, 10, 0, 16, 0, 5, 0, 0, 82, 10, 0, 16, 0, 5, 0, 74, 8, 251, 0, 16, 0, 5, 0, 0, 84, 10, 0, 16, 0, 5, 0, 0, 82, 10, 0, 16, 0, 5, 0, 206, 43, 251, 0, 16, 0, 5, 0, 0, 84, 10, 0, 16, 0, 5, 0, 0, 82, 10, 0, 16, 0, 5, 0, 208, 121, 251, 0, 16, 0, 5, 0, 0, 84, 10, 0, 16, 0, 5, 0, 0, 82, 10, 0, 16, 0, 5, 0, 182, 191, 251, 0, 16, 0, 5, 0, 0, 84, 10, 0, 16, 0, 5, 0, 0, 82, 10, 0, 16, 0, 5, 0, 52, 214, 251, 0, 16, 0, 5, 0, 0, 84, 10, 0, 16, 0, 5, 0, 0, 82, 10, 0, 16, 0, 5, 0, 128, 235, 251, 0, 16, 0, 5, 0, 0, 84, 10, 0, 16, 0, 5, 0, 0, 82, 10, 0, 16, 0, 5, 0, 34, 50, 252, 0, 16, 0, 5, 0, 0, 84, 10, 0, 16, 0, 5, 0, 0, 82, 10, 0, 16, 0, 5, 0, 100, 111, 252, 0, 16, 0, 5, 0, 0, 84, 10, 0, 21, 0, 5, 0, 0, 0, 19, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 19, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 19, 0, 16, 0, 5, 0, 86, 233, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 19, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 21, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 21, 0, 16, 0, 5, 0, 86, 233, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 21, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 23, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 23, 0, 16, 0, 5, 0, 86, 233, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 23, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 25, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 25, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 27, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 27, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 29, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 29, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 31, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 31, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 33, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 33, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 35, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 21, 0, 5, 0, 0, 0, 35, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 21, 0, 22, 0, 5, 0, 0, 0, 37, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 21, 0, 22, 0, 5, 0, 0, 0, 37, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 19, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 19, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 21, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 21, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 23, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 23, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 25, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 25, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 27, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 27, 0, 16, 0, 5, 0, 34, 50, 252, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 29, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 31, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 33, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 23, 0, 21, 0, 5, 0, 0, 0, 35, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 23, 0, 22, 0, 5, 0, 0, 0, 37, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 25, 0, 21, 0, 5, 0, 0, 0, 19, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 25, 0, 21, 0, 5, 0, 0, 0, 21, 0, 16, 0, 5, 0, 86, 222, 251, 0, 21, 0, 5, 0, 0, 0, 25, 0, 16, 0, 5, 0, 86, 233, 251, 0, 21, 0, 5, 0, 0, 0, 27, 0, 16, 0, 5, 0, 86, 233, 251, 0, 21, 0, 5, 0, 0, 0, 29, 0, 16, 0, 5, 0, 86, 233, 251, 0, 21, 0, 5, 0, 0, 0, 31, 0, 16, 0, 5, 0, 86, 233, 251, 0, 21, 0, 5, 0, 0, 0, 33, 0, 16, 0, 5, 0, 86, 233, 251, 0, 21, 0, 5, 0, 0, 0, 35, 0, 16, 0, 5, 0, 86, 233, 251, 0, 22, 0, 5, 0, 0, 0, 37, 0, 16, 0, 5, 0, 86, 233, 251, 0, 160, 0, 5, 0, 2, 46, 97, 0, 160, 0, 5, 0, 9, 46, 97, 0, 160, 0, 5, 0, 2, 56, 97, 0, 160, 0, 5, 0, 9, 56, 97, 0, 160, 0, 5, 0, 2, 96, 97, 0, 160, 0, 5, 0, 9, 96, 97, 0, 160, 0, 5, 0, 2, 104, 97, 0, 160, 0, 5, 0, 9, 104, 97, 0, 160, 0, 5, 0, 16, 104, 97, 0, 160, 0, 5, 0, 2, 164, 97, 0, 160, 0, 5, 0, 9, 164, 97, 0, 32, 0, 5, 0, 2, 178, 97, 0, 174, 0, 5, 0, 2, 178, 97, 0, 160, 0, 5, 0, 9, 178, 97, 0, 160, 0, 5, 0, 2, 184, 97, 0, 160, 0, 5, 0, 9, 184, 97, 0, 160, 0, 5, 0, 2, 190, 97, 0, 160, 0, 5, 0, 9, 190, 97, 0, 32, 0, 5, 0, 16, 190, 97, 0, 174, 0, 5, 0, 16, 190, 97, 0, 160, 0, 5, 0, 23, 190, 97, 0, 160, 0, 5, 0, 30, 190, 97, 0, 160, 0, 5, 0, 37, 190, 97, 0, 160, 0, 5, 0, 44, 190, 97, 0, 160, 0, 5, 0, 2, 194, 97, 0, 160, 0, 5, 0, 9, 194, 97, 0, 160, 0, 5, 0, 2, 202, 97, 0, 160, 0, 5, 0, 9, 202, 97, 0, 160, 0, 5, 0, 2, 224, 97, 0, 32, 0, 5, 0, 2, 236, 97, 0, 174, 0, 5, 0, 2, 236, 97, 0, 160, 0, 5, 0, 9, 236, 97, 0, 160, 0, 5, 0, 2, 242, 97, 0, 32, 0, 5, 0, 9, 242, 97, 0, 174, 0, 5, 0, 9, 242, 97, 0, 160, 0, 5, 0, 2, 246, 97, 0, 32, 0, 5, 0, 9, 246, 97, 0, 174, 0, 5, 0, 9, 246, 97, 0, 32, 0, 5, 0, 16, 246, 97, 0, 174, 0, 5, 0, 16, 246, 97, 0, 160, 0, 5, 0, 23, 246, 97, 0, 32, 0, 5, 0, 30, 246, 97, 0, 174, 0, 5, 0, 30, 246, 97, 0, 160, 0, 5, 0, 37, 246, 97, 0, 160, 0, 5, 0, 44, 246, 97, 0, 160, 0, 5, 0, 51, 246, 97, 0, 32, 0, 5, 0, 58, 246, 97, 0, 174, 0, 5, 0, 58, 246, 97, 0, 160, 0, 5, 0, 65, 246, 97, 0, 160, 0, 5, 0, 72, 246, 97, 0, 32, 0, 5, 0, 79, 246, 97, 0, 174, 0, 5, 0, 79, 246, 97, 0, 160, 0, 5, 0, 86, 246, 97, 0, 160, 0, 5, 0, 86, 246, 97, 0, 5, 0, 184, 0, 0, 0, 0, 0, 160, 0, 5, 0, 93, 246, 97, 0, 38, 0, 5, 0, 0, 22, 102, 0, 32, 0, 5, 0, 16, 220, 102, 0, 49, 0, 5, 0, 0, 22, 102, 0, 46, 0, 5, 0, 16, 220, 102, 0, 38, 0, 5, 0, 0, 22, 102, 0, 32, 0, 5, 0, 9, 226, 102, 0, 49, 0, 5, 0, 0, 22, 102, 0, 46, 0, 5, 0, 9, 226, 102, 0, 32, 0, 5, 0, 2, 176, 102, 0, 16, 0, 5, 0, 30, 176, 102, 0, 27, 0, 5, 0, 30, 176, 102, 0, 38, 0, 5, 0, 30, 176, 102, 0, 49, 0, 5, 0, 30, 176, 102, 0, 16, 0, 5, 0, 51, 176, 102, 0, 27, 0, 5, 0, 51, 176, 102, 0, 38, 0, 5, 0, 51, 176, 102, 0, 49, 0, 5, 0, 51, 176, 102, 0, 32, 0, 5, 0, 72, 176, 102, 0, 16, 0, 5, 0, 16, 190, 102, 0, 27, 0, 5, 0, 16, 190, 102, 0, 38, 0, 5, 0, 16, 190, 102, 0, 49, 0, 5, 0, 16, 190, 102, 0, 16, 0, 5, 0, 16, 194, 102, 0, 27, 0, 5, 0, 16, 194, 102, 0, 38, 0, 5, 0, 16, 194, 102, 0, 49, 0, 5, 0, 16, 194, 102, 0, 16, 0, 5, 0, 30, 194, 102, 0, 27, 0, 5, 0, 30, 194, 102, 0, 38, 0, 5, 0, 30, 194, 102, 0, 49, 0, 5, 0, 30, 194, 102, 0, 16, 0, 5, 0, 9, 208, 102, 0, 27, 0, 5, 0, 9, 208, 102, 0, 38, 0, 5, 0, 9, 208, 102, 0, 49, 0, 5, 0, 9, 208, 102, 0, 32, 0, 5, 0, 16, 220, 102, 0, 32, 128, 232, 0, 0, 0, 0, 0, 46, 0, 5, 0, 16, 220, 102, 0, 46, 128, 232, 0, 0, 0, 0, 0, 32, 0, 5, 0, 2, 226, 102, 0, 46, 0, 5, 0, 2, 226, 102, 0, 32, 0, 5, 0, 9, 234, 102, 0, 46, 0, 5, 0, 9, 234, 102, 0, 160, 0, 5, 0, 9, 232, 118, 0, 160, 0, 5, 0, 16, 232, 118, 0, 160, 0, 5, 0, 23, 232, 118, 0, 160, 0, 5, 0, 30, 232, 118, 0, 160, 0, 5, 0, 37, 232, 118, 0, 160, 0, 5, 0, 44, 232, 118, 0, 160, 0, 5, 0, 51, 232, 118, 0, 160, 0, 5, 0, 58, 232, 118, 0, 160, 0, 5, 0, 65, 232, 118, 0, 160, 0, 5, 0, 72, 232, 118, 0, 160, 0, 5, 0, 79, 232, 118, 0, 160, 0, 5, 0, 86, 232, 118, 0, 160, 0, 5, 0, 93, 232, 118, 0, 160, 0, 5, 0, 100, 232, 118, 0, 160, 0, 5, 0, 107, 232, 118, 0, 160, 0, 5, 0, 114, 232, 118, 0, 160, 0, 5, 0, 121, 232, 118, 0, 160, 0, 5, 0, 128, 232, 118, 0, 160, 0, 5, 0, 135, 232, 118, 0, 160, 0, 5, 0, 142, 232, 118, 0, 160, 0, 5, 0, 149, 232, 118, 0, 160, 0, 5, 0, 156, 232, 118, 0, 160, 0, 5, 0, 163, 232, 118, 0, 160, 0, 5, 0, 170, 232, 118, 0, 160, 0, 5, 0, 177, 232, 118, 0, 160, 0, 5, 0, 184, 232, 118, 0, 160, 0, 5, 0, 191, 232, 118, 0, 160, 0, 5, 0, 198, 232, 118, 0, 160, 0, 5, 0, 205, 232, 118, 0, 160, 0, 5, 0, 212, 232, 118, 0, 160, 0, 5, 0, 219, 232, 118, 0, 160, 0, 5, 0, 226, 232, 118, 0, 160, 0, 5, 0, 233, 232, 118, 0, 160, 0, 5, 0, 240, 232, 118, 0, 160, 0, 5, 0, 247, 232, 118, 0, 160, 0, 5, 0, 254, 232, 118, 0, 32, 0, 5, 0, 16, 92, 121, 0, 32, 0, 5, 0, 44, 92, 121, 0, 32, 0, 5, 0, 114, 92, 121, 0, 32, 0, 5, 0, 163, 92, 121, 0, 32, 0, 5, 0, 198, 92, 121, 0, 32, 0, 5, 0, 247, 101, 121, 0, 174, 0, 5, 0, 247, 101, 121, 0, 32, 0, 5, 0, 30, 247, 121, 0, 32, 0, 5, 0, 58, 247, 121, 0, 32, 0, 5, 0, 72, 247, 121, 0, 32, 0, 5, 0, 100, 247, 121, 0, 32, 0, 5, 0, 114, 247, 121, 0, 32, 0, 5, 0, 240, 247, 121, 0, 32, 0, 5, 0, 247, 247, 121, 0, 32, 0, 5, 0, 51, 162, 124, 0, 32, 0, 5, 0, 72, 162, 124, 0, 27, 0, 5, 0, 2, 2, 251, 0, 38, 0, 5, 0, 2, 2, 251, 0, 49, 0, 5, 0, 2, 2, 251, 0, 38, 0, 5, 0, 6, 2, 251, 0, 27, 0, 5, 0, 10, 2, 251, 0, 27, 0, 5, 0, 32, 2, 251, 0, 38, 0, 5, 0, 32, 2, 251, 0, 49, 0, 5, 0, 32, 2, 251, 0, 27, 0, 5, 0, 34, 2, 251, 0, 38, 0, 5, 0, 34, 2, 251, 0, 27, 0, 5, 0, 36, 2, 251, 0, 38, 0, 5, 0, 36, 2, 251, 0, 38, 0, 5, 0, 132, 2, 251, 0, 16, 0, 5, 0, 60, 4, 251, 0, 16, 0, 5, 0, 14, 66, 252, 0, 27, 0, 5, 0, 82, 4, 251, 0, 38, 0, 5, 0, 82, 4, 251, 0, 49, 0, 5, 0, 82, 4, 251, 0, 16, 0, 5, 0, 200, 4, 251, 0, 5, 0, 112, 0, 200, 4, 251, 0, 16, 0, 5, 0, 252, 4, 251, 0, 16, 0, 5, 0, 26, 6, 251, 0, 38, 0, 5, 0, 26, 6, 251, 0, 5, 0, 112, 0, 26, 6, 251, 0, 16, 0, 5, 0, 28, 6, 251, 0, 16, 0, 5, 0, 30, 6, 251, 0, 27, 0, 5, 0, 48, 6, 251, 0, 16, 0, 5, 0, 248, 7, 251, 0, 27, 0, 5, 0, 74, 8, 251, 0, 38, 0, 5, 0, 74, 8, 251, 0, 49, 0, 5, 0, 74, 8, 251, 0, 27, 0, 5, 0, 96, 8, 251, 0, 16, 0, 5, 0, 232, 8, 251, 0, 49, 0, 5, 0, 250, 8, 251, 0, 16, 0, 5, 0, 68, 10, 251, 0, 38, 0, 5, 0, 68, 10, 251, 0, 16, 0, 5, 0, 70, 10, 251, 0, 49, 0, 5, 0, 182, 10, 251, 0, 49, 0, 5, 0, 60, 59, 251, 0, 27, 0, 5, 0, 40, 11, 251, 0, 27, 0, 5, 0, 72, 11, 251, 0, 27, 0, 5, 0, 70, 24, 251, 0, 16, 0, 5, 0, 4, 26, 251, 0, 16, 0, 5, 0, 188, 27, 251, 0, 27, 0, 5, 0, 74, 28, 251, 0, 27, 0, 5, 0, 80, 28, 251, 0, 16, 0, 5, 0, 104, 29, 251, 0, 5, 0, 112, 0, 104, 29, 251, 0, 49, 0, 5, 0, 162, 29, 251, 0, 16, 0, 5, 0, 74, 30, 251, 0, 27, 0, 5, 0, 88, 30, 251, 0, 16, 0, 5, 0, 62, 31, 251, 0, 16, 0, 5, 0, 196, 33, 251, 0, 5, 0, 112, 0, 196, 33, 251, 0, 16, 0, 5, 0, 200, 34, 251, 0, 16, 0, 5, 0, 132, 35, 251, 0, 5, 0, 112, 0, 132, 35, 251, 0, 16, 0, 5, 0, 136, 35, 251, 0, 49, 0, 5, 0, 90, 36, 251, 0, 49, 0, 5, 0, 172, 37, 251, 0, 49, 0, 5, 0, 126, 39, 251, 0, 16, 0, 5, 0, 8, 42, 251, 0, 27, 0, 5, 0, 114, 42, 251, 0, 16, 0, 5, 0, 244, 44, 251, 0, 16, 0, 5, 0, 234, 45, 251, 0, 16, 0, 5, 0, 82, 46, 251, 0, 16, 0, 5, 0, 118, 47, 251, 0, 27, 0, 5, 0, 134, 47, 251, 0, 27, 0, 5, 0, 180, 47, 251, 0, 16, 0, 5, 0, 42, 2, 251, 0, 16, 0, 5, 0, 186, 47, 251, 0, 27, 0, 5, 0, 52, 48, 251, 0, 16, 0, 5, 0, 44, 49, 251, 0, 5, 0, 112, 0, 44, 49, 251, 0, 16, 0, 5, 0, 240, 49, 251, 0, 5, 0, 112, 0, 240, 49, 251, 0, 27, 0, 5, 0, 26, 50, 251, 0, 16, 0, 5, 0, 178, 50, 251, 0, 5, 0, 112, 0, 178, 50, 251, 0, 16, 0, 5, 0, 120, 53, 251, 0, 16, 0, 5, 0, 122, 54, 251, 0, 49, 0, 5, 0, 132, 54, 251, 0, 16, 0, 5, 0, 112, 56, 251, 0, 27, 0, 5, 0, 150, 56, 251, 0, 27, 0, 5, 0, 158, 56, 251, 0, 49, 0, 5, 0, 158, 56, 251, 0, 49, 0, 5, 0, 246, 56, 251, 0, 27, 0, 5, 0, 2, 57, 251, 0, 49, 0, 5, 0, 200, 57, 251, 0, 27, 0, 5, 0, 180, 64, 251, 0, 49, 0, 5, 0, 142, 67, 251, 0, 16, 0, 5, 0, 76, 83, 251, 0, 27, 0, 5, 0, 90, 83, 251, 0, 38, 0, 5, 0, 90, 83, 251, 0, 27, 0, 5, 0, 186, 85, 251, 0, 38, 0, 5, 0, 12, 86, 251, 0, 16, 0, 5, 0, 220, 99, 251, 0, 49, 0, 5, 0, 234, 99, 251, 0, 16, 0, 5, 0, 178, 100, 251, 0, 16, 0, 5, 0, 232, 100, 251, 0, 16, 0, 5, 0, 136, 101, 251, 0, 49, 0, 5, 0, 154, 101, 251, 0, 27, 0, 5, 0, 182, 101, 251, 0, 16, 0, 5, 0, 204, 102, 251, 0, 49, 0, 5, 0, 204, 102, 251, 0, 49, 0, 5, 0, 196, 9, 252, 0, 38, 0, 5, 0, 208, 102, 251, 0, 49, 0, 5, 0, 208, 102, 251, 0, 16, 0, 5, 0, 100, 106, 251, 0, 27, 0, 5, 0, 100, 106, 251, 0, 16, 0, 5, 0, 108, 119, 251, 0, 49, 0, 5, 0, 146, 119, 251, 0, 27, 0, 5, 0, 232, 119, 251, 0, 16, 0, 5, 0, 174, 121, 251, 0, 27, 0, 5, 0, 86, 122, 251, 0, 16, 0, 5, 0, 22, 128, 251, 0, 16, 0, 5, 0, 34, 129, 251, 0, 5, 0, 112, 0, 34, 129, 251, 0, 5, 0, 120, 0, 34, 129, 251, 0, 16, 0, 5, 0, 144, 130, 251, 0, 5, 0, 112, 0, 144, 130, 251, 0, 16, 0, 5, 0, 146, 130, 251, 0, 5, 0, 112, 0, 146, 130, 251, 0, 16, 0, 5, 0, 192, 131, 251, 0, 16, 0, 5, 0, 8, 135, 251, 0, 16, 0, 5, 0, 128, 135, 251, 0, 16, 0, 5, 0, 66, 147, 251, 0, 16, 0, 5, 0, 210, 147, 251, 0, 27, 0, 5, 0, 228, 147, 251, 0, 49, 0, 5, 0, 228, 147, 251, 0, 16, 0, 5, 0, 132, 148, 251, 0, 16, 0, 5, 0, 136, 148, 251, 0, 16, 0, 5, 0, 20, 149, 251, 0, 16, 0, 5, 0, 170, 153, 251, 0, 49, 0, 5, 0, 172, 153, 251, 0, 49, 0, 5, 0, 100, 188, 251, 0, 16, 0, 5, 0, 92, 154, 251, 0, 27, 0, 5, 0, 106, 154, 251, 0, 16, 0, 5, 0, 218, 154, 251, 0, 16, 0, 5, 0, 254, 159, 251, 0, 16, 0, 5, 0, 102, 160, 251, 0, 16, 0, 5, 0, 108, 161, 251, 0, 16, 0, 5, 0, 216, 161, 251, 0, 16, 0, 5, 0, 210, 164, 251, 0, 5, 0, 112, 0, 210, 164, 251, 0, 16, 0, 5, 0, 212, 164, 251, 0, 16, 0, 5, 0, 122, 165, 251, 0, 16, 0, 5, 0, 88, 166, 251, 0, 49, 0, 5, 0, 8, 167, 251, 0, 27, 0, 5, 0, 156, 167, 251, 0, 16, 0, 5, 0, 252, 169, 251, 0, 5, 0, 112, 0, 252, 169, 251, 0, 16, 0, 5, 0, 254, 169, 251, 0, 16, 0, 5, 0, 76, 188, 251, 0, 16, 0, 5, 0, 162, 190, 251, 0, 16, 0, 5, 0, 156, 191, 251, 0, 49, 0, 5, 0, 156, 191, 251, 0, 16, 0, 5, 0, 158, 191, 251, 0, 49, 0, 5, 0, 182, 191, 251, 0, 49, 0, 5, 0, 156, 192, 251, 0, 49, 0, 5, 0, 254, 194, 251, 0, 49, 0, 5, 0, 210, 196, 251, 0, 16, 0, 5, 0, 60, 212, 251, 0, 16, 0, 5, 0, 230, 212, 251, 0, 16, 0, 5, 0, 232, 212, 251, 0, 16, 0, 5, 0, 254, 217, 251, 0, 27, 0, 5, 0, 254, 217, 251, 0, 16, 0, 5, 0, 178, 218, 251, 0, 49, 0, 5, 0, 206, 218, 251, 0, 16, 0, 5, 0, 74, 219, 251, 0, 49, 0, 5, 0, 188, 219, 251, 0, 16, 0, 5, 0, 68, 220, 251, 0, 16, 0, 5, 0, 42, 222, 251, 0, 16, 0, 5, 0, 44, 222, 251, 0, 27, 0, 5, 0, 86, 222, 251, 0, 5, 0, 112, 0, 86, 222, 251, 0, 49, 0, 5, 0, 22, 223, 251, 0, 49, 0, 5, 0, 232, 23, 252, 0, 49, 0, 5, 0, 180, 223, 251, 0, 49, 0, 5, 0, 206, 223, 251, 0, 49, 0, 5, 0, 60, 59, 251, 0, 16, 0, 5, 0, 90, 232, 251, 0, 27, 0, 5, 0, 86, 233, 251, 0, 49, 0, 5, 0, 86, 233, 251, 0, 5, 0, 112, 0, 86, 233, 251, 0, 27, 0, 5, 0, 90, 233, 251, 0, 49, 0, 5, 0, 90, 233, 251, 0, 27, 0, 5, 0, 114, 235, 251, 0, 27, 0, 5, 0, 60, 239, 251, 0, 49, 0, 5, 0, 60, 239, 251, 0, 49, 0, 5, 0, 130, 161, 251, 0, 49, 0, 5, 0, 90, 11, 251, 0, 49, 0, 5, 0, 240, 134, 252, 0, 16, 0, 5, 0, 188, 6, 252, 0, 16, 0, 5, 0, 190, 9, 252, 0, 27, 0, 5, 0, 196, 9, 252, 0, 16, 0, 5, 0, 192, 11, 252, 0, 5, 0, 112, 0, 198, 11, 252, 0, 16, 0, 5, 0, 174, 14, 252, 0, 16, 0, 5, 0, 104, 16, 252, 0, 16, 0, 5, 0, 110, 16, 252, 0, 16, 0, 5, 0, 180, 16, 252, 0, 16, 0, 5, 0, 22, 17, 252, 0, 16, 0, 5, 0, 236, 20, 252, 0, 16, 0, 5, 0, 240, 20, 252, 0, 16, 0, 5, 0, 54, 21, 252, 0, 27, 0, 5, 0, 230, 21, 252, 0, 16, 0, 5, 0, 232, 21, 252, 0, 16, 0, 5, 0, 234, 21, 252, 0, 27, 0, 5, 0, 66, 24, 252, 0, 49, 0, 5, 0, 68, 32, 252, 0, 49, 0, 5, 0, 182, 36, 252, 0, 27, 0, 5, 0, 142, 48, 252, 0, 16, 0, 5, 0, 144, 48, 252, 0, 49, 0, 5, 0, 124, 53, 252, 0, 16, 0, 5, 0, 48, 64, 252, 0, 16, 0, 5, 0, 50, 64, 252, 0, 5, 0, 112, 0, 50, 64, 252, 0, 16, 0, 5, 0, 150, 65, 252, 0, 16, 0, 5, 0, 218, 65, 252, 0, 16, 0, 5, 0, 16, 66, 252, 0, 16, 0, 5, 0, 252, 66, 252, 0, 16, 0, 5, 0, 28, 68, 252, 0, 16, 0, 5, 0, 94, 68, 252, 0, 5, 0, 112, 0, 94, 68, 252, 0, 27, 0, 5, 0, 116, 69, 252, 0, 16, 0, 5, 0, 46, 72, 252, 0, 16, 0, 5, 0, 48, 72, 252, 0, 16, 0, 5, 0, 112, 79, 252, 0, 16, 0, 5, 0, 132, 79, 252, 0, 5, 0, 112, 0, 134, 79, 252, 0, 16, 0, 5, 0, 158, 89, 252, 0, 16, 0, 5, 0, 104, 90, 252, 0, 16, 0, 5, 0, 188, 92, 252, 0, 16, 0, 5, 0, 36, 93, 252, 0, 49, 0, 5, 0, 36, 93, 252, 0, 16, 0, 5, 0, 208, 93, 252, 0, 16, 0, 5, 0, 32, 94, 252, 0, 38, 0, 5, 0, 36, 94, 252, 0, 49, 0, 5, 0, 38, 94, 252, 0, 27, 0, 5, 0, 56, 94, 252, 0, 16, 0, 5, 0, 100, 98, 252, 0, 5, 0, 112, 0, 100, 98, 252, 0, 16, 0, 5, 0, 176, 98, 252, 0, 16, 0, 5, 0, 156, 106, 252, 0, 16, 0, 5, 0, 250, 106, 252, 0, 16, 0, 5, 0, 26, 109, 252, 0, 16, 0, 5, 0, 184, 110, 252, 0, 27, 0, 5, 0, 8, 112, 252, 0, 16, 0, 5, 0, 128, 113, 252, 0, 5, 0, 112, 0, 128, 113, 252, 0, 16, 0, 5, 0, 54, 123, 252, 0, 16, 0, 5, 0, 16, 124, 252, 0, 16, 0, 5, 0, 134, 125, 252, 0, 16, 0, 5, 0, 210, 134, 252, 0, 5, 0, 112, 0, 210, 134, 252, 0, 16, 0, 5, 0, 212, 134, 252, 0, 27, 0, 5, 0, 240, 134, 252, 0, 27, 0, 5, 0, 158, 135, 252, 0, 49, 0, 5, 0, 80, 137, 252, 0, 16, 0, 5, 0, 88, 140, 252, 0, 16, 0, 5, 0, 156, 140, 252, 0, 27, 0, 5, 0, 134, 141, 252, 0, 16, 0, 5, 0, 22, 148, 252, 0, 49, 0, 5, 0, 50, 148, 252, 0, 16, 0, 5, 0, 118, 152, 252, 0, 16, 0, 5, 0, 154, 154, 252, 0, 5, 0, 112, 0, 154, 154, 252, 0, 27, 0, 5, 0, 76, 158, 252, 0, 16, 0, 5, 0, 76, 168, 252, 0, 16, 0, 5, 0, 70, 174, 252, 0, 16, 0, 5, 0, 72, 174, 252, 0, 49, 0, 5, 0, 168, 175, 252, 0, 16, 0, 5, 0, 220, 185, 252, 0, 16, 0, 5, 0, 198, 188, 252, 0, 16, 0, 5, 0, 238, 189, 252, 0, 16, 0, 5, 0, 240, 189, 252, 0, 5, 0, 112, 0, 240, 189, 252, 0, 16, 0, 5, 0, 242, 189, 252, 0, 5, 0, 112, 0, 242, 189, 252, 0, 5, 0, 120, 0, 242, 189, 252, 0, 16, 0, 5, 0, 14, 193, 252, 0, 5, 0, 112, 0, 14, 193, 252, 0, 5, 0, 120, 0, 14, 193, 252, 0, 16, 0, 5, 0, 22, 193, 252, 0, 16, 0, 5, 0, 136, 195, 252, 0, 16, 0, 5, 0, 84, 198, 252, 0, 16, 0, 5, 0, 86, 198, 252, 0, 16, 0, 5, 0, 214, 198, 252, 0, 16, 0, 5, 0, 40, 199, 252, 0, 16, 0, 5, 0, 74, 200, 252, 0, 16, 0, 5, 0, 128, 203, 252, 0, 5, 0, 112, 0, 128, 203, 252, 0, 16, 0, 5, 0, 130, 203, 252, 0, 16, 0, 5, 0, 226, 203, 252, 0, 5, 0, 112, 0, 226, 203, 252, 0, 16, 0, 5, 0, 90, 215, 252, 0, 16, 0, 5, 0, 10, 217, 252, 0, 5, 0, 112, 0, 10, 217, 252, 0, 16, 0, 5, 0, 106, 218, 252, 0, 16, 0, 5, 0, 90, 219, 252, 0, 16, 0, 5, 0, 128, 219, 252, 0, 16, 0, 5, 0, 106, 222, 252, 0, 16, 0, 5, 0, 140, 222, 252, 0, 16, 0, 5, 0, 6, 223, 252, 0, 16, 0, 5, 0, 8, 223, 252, 0, 5, 0, 112, 0, 8, 223, 252, 0, 5, 0, 120, 0, 8, 223, 252, 0, 16, 0, 5, 0, 174, 252, 252, 0, 16, 0, 5, 0, 178, 252, 252, 0, 16, 0, 5, 0, 150, 254, 252, 0, 16, 0, 5, 0, 202, 14, 253, 0, 16, 0, 5, 0, 118, 15, 253, 0, 16, 0, 5, 0, 52, 16, 253, 0, 16, 0, 5, 0, 54, 16, 253, 0, 16, 0, 5, 0, 136, 24, 253, 0, 16, 0, 5, 0, 138, 24, 253, 0, 16, 0, 5, 0, 140, 24, 253, 0, 16, 0, 5, 0, 42, 25, 253, 0, 16, 0, 5, 0, 148, 27, 253, 0, 16, 0, 5, 0, 240, 27, 253, 0, 5, 0, 112, 0, 240, 27, 253, 0, 49, 0, 5, 0, 152, 28, 253, 0, 16, 0, 5, 0, 38, 30, 253, 0, 16, 0, 5, 0, 192, 40, 253, 0, 16, 0, 5, 0, 160, 43, 253, 0, 16, 0, 5, 0, 90, 44, 253, 0, 16, 0, 5, 0, 112, 45, 253, 0, 16, 0, 5, 0, 36, 47, 253, 0, 16, 0, 5, 0, 190, 48, 253, 0, 27, 0, 5, 0, 218, 48, 253, 0, 49, 0, 5, 0, 14, 49, 253, 0, 27, 0, 5, 0, 8, 50, 253, 0, 16, 0, 5, 0, 202, 53, 253, 0, 16, 0, 5, 0, 190, 54, 253, 0, 16, 0, 5, 0, 84, 55, 253, 0, 49, 0, 5, 0, 84, 55, 253, 0, 16, 0, 5, 0, 84, 59, 253, 0, 5, 0, 112, 0, 84, 59, 253, 0, 16, 0, 5, 0, 230, 68, 253, 0, 16, 0, 5, 0, 218, 70, 253, 0, 16, 0, 5, 0, 100, 76, 253, 0, 16, 0, 5, 0, 186, 77, 253, 0, 16, 0, 5, 0, 128, 78, 253, 0, 16, 0, 5, 0, 176, 78, 253, 0, 16, 0, 5, 0, 178, 78, 253, 0, 5, 0, 112, 0, 178, 78, 253, 0, 5, 0, 120, 0, 178, 78, 253, 0, 49, 0, 5, 0, 144, 82, 253, 0, 27, 0, 5, 0, 26, 84, 253, 0, 16, 0, 5, 0, 98, 87, 253, 0, 5, 0, 112, 0, 98, 87, 253, 0, 16, 0, 5, 0, 148, 93, 253, 0, 49, 0, 5, 0, 166, 93, 253, 0, 16, 0, 5, 0, 30, 98, 253, 0, 16, 0, 5, 0, 98, 98, 253, 0, 27, 0, 5, 0, 242, 98, 253, 0, 16, 0, 5, 0, 86, 114, 253, 0, 16, 0, 5, 0, 200, 118, 253, 0, 16, 0, 5, 0, 202, 118, 253, 0, 16, 0, 5, 0, 2, 120, 253, 0, 16, 0, 5, 0, 6, 120, 253, 0, 16, 0, 5, 0, 84, 124, 253, 0, 16, 0, 5, 0, 154, 125, 253, 0, 5, 0, 112, 0, 154, 125, 253, 0, 16, 0, 5, 0, 156, 125, 253, 0, 16, 0, 5, 0, 186, 131, 253, 0, 16, 0, 5, 0, 228, 131, 253, 0, 16, 0, 5, 0, 178, 134, 253, 0, 5, 0, 112, 0, 178, 134, 253, 0, 16, 0, 5, 0, 216, 139, 253, 0, 16, 0, 5, 0, 218, 139, 253, 0, 16, 0, 5, 0, 44, 140, 253, 0, 16, 0, 5, 0, 152, 140, 253, 0, 16, 0, 5, 0, 160, 141, 253, 0, 16, 0, 5, 0, 196, 145, 253, 0, 16, 0, 5, 0, 30, 147, 253, 0, 16, 0, 5, 0, 108, 147, 253, 0, 16, 0, 5, 0, 184, 147, 253, 0, 16, 0, 5, 0, 142, 148, 253, 0, 27, 0, 5, 0, 162, 148, 253, 0, 16, 0, 5, 0, 70, 153, 253, 0, 16, 0, 5, 0, 82, 154, 253, 0, 16, 0, 5, 0, 242, 156, 253, 0, 16, 0, 5, 0, 122, 157, 253, 0, 16, 0, 5, 0, 164, 157, 253, 0, 16, 0, 5, 0, 170, 157, 253, 0, 16, 0, 5, 0, 172, 157, 253, 0, 5, 0, 112, 0, 172, 157, 253, 0, 16, 0, 5, 0, 20, 163, 253, 0, 16, 0, 5, 0, 110, 164, 253, 0, 5, 0, 112, 0, 110, 164, 253, 0, 16, 0, 5, 0, 252, 164, 253, 0, 16, 0, 5, 0, 208, 165, 253, 0, 16, 0, 5, 0, 244, 171, 253, 0, 16, 0, 5, 0, 168, 173, 253, 0, 5, 0, 112, 0, 168, 173, 253, 0, 16, 0, 5, 0, 232, 175, 253, 0, 16, 0, 5, 0, 186, 176, 253, 0, 16, 0, 5, 0, 58, 180, 253, 0, 16, 0, 5, 0, 136, 180, 253, 0, 16, 0, 5, 0, 172, 180, 253, 0, 16, 0, 5, 0, 166, 181, 253, 0, 5, 0, 112, 0, 166, 181, 253, 0, 16, 0, 5, 0, 2, 184, 253, 0, 16, 0, 5, 0, 64, 193, 253, 0, 16, 0, 5, 0, 218, 195, 253, 0, 16, 0, 5, 0, 72, 206, 253, 0, 16, 0, 5, 0, 240, 208, 253, 0, 5, 0, 112, 0, 240, 208, 253, 0, 16, 0, 5, 0, 178, 209, 253, 0, 16, 0, 5, 0, 56, 211, 253, 0, 16, 0, 5, 0, 196, 212, 253, 0, 16, 0, 5, 0, 94, 213, 253, 0, 16, 0, 5, 0, 210, 213, 253, 0, 16, 0, 5, 0, 212, 213, 253, 0, 16, 0, 5, 0, 124, 214, 253, 0, 16, 0, 5, 0, 252, 214, 253, 0, 16, 0, 5, 0, 36, 217, 253, 0, 16, 0, 5, 0, 62, 217, 253, 0, 16, 0, 5, 0, 222, 217, 253, 0, 16, 0, 5, 0, 6, 218, 253, 0, 16, 0, 5, 0, 62, 218, 253, 0, 16, 0, 5, 0, 220, 218, 253, 0, 16, 0, 5, 0, 240, 219, 253, 0, 16, 0, 5, 0, 176, 220, 253, 0, 5, 0, 112, 0, 176, 220, 253, 0, 16, 0, 5, 0, 236, 220, 253, 0, 16, 0, 5, 0, 248, 220, 253, 0, 5, 0, 112, 0, 248, 220, 253, 0, 16, 0, 5, 0, 110, 223, 253, 0, 16, 0, 5, 0, 12, 224, 253, 0, 5, 0, 112, 0, 12, 224, 253, 0, 16, 0, 5, 0, 96, 224, 253, 0, 16, 0, 5, 0, 118, 224, 253, 0, 5, 0, 112, 0, 118, 224, 253, 0, 5, 0, 120, 0, 118, 224, 253, 0, 16, 0, 5, 0, 218, 224, 253, 0, 16, 0, 5, 0, 236, 224, 253, 135, 218, 7, 0, 0, 2, 140, 103, 135, 48, 17, 0, 0, 2, 92, 121, 135, 118, 17, 0, 0, 2, 145, 121, 135, 207, 17, 0, 0, 2, 247, 121, 135, 2, 20, 0, 0, 16, 234, 118, 135, 150, 21, 0, 0, 99, 245, 118, 135, 167, 21, 0, 0, 225, 245, 118, 135, 176, 24, 0, 0, 150, 251, 118, 135, 80, 25, 0, 0, 9, 4, 118, 135, 60, 26, 0, 0, 2, 26, 118, 135, 90, 28, 0, 0, 9, 56, 118, 7, 156, 33, 0, 0, 44, 107, 12, 7, 175, 33, 0, 0, 170, 107, 12, 7, 138, 34, 0, 0, 224, 158, 12, 7, 238, 34, 0, 0, 85, 161, 12, 7, 12, 35, 0, 0, 13, 162, 12, 7, 43, 35, 0, 0, 216, 162, 12, 7, 3, 37, 0, 0, 2, 174, 12, 7, 161, 37, 0, 0, 2, 182, 12, 7, 204, 37, 0, 0, 2, 187, 12, 7, 0, 38, 0, 0, 2, 194, 12, 7, 56, 38, 0, 0, 84, 195, 12, 7, 112, 38, 0, 0, 201, 196, 12, 7, 144, 38, 0, 0, 129, 197, 12, 7, 33, 39, 0, 0, 70, 202, 12, 7, 148, 39, 0, 0, 59, 204, 12, 7, 199, 39, 0, 0, 148, 205, 12, 7, 0, 40, 0, 0, 55, 227, 12, 7, 0, 41, 0, 0, 223, 206, 12, 7, 153, 41, 0, 0, 124, 210, 12, 7, 220, 41, 0, 0, 57, 212, 12, 7, 13, 42, 0, 0, 125, 213, 12, 7, 119, 42, 0, 0, 84, 216, 12, 7, 221, 42, 0, 0, 29, 219, 12, 7, 118, 43, 0, 0, 70, 223, 12, 7, 151, 43, 0, 0, 40, 224, 12, 135, 60, 45, 0, 0, 100, 141, 103, 135, 84, 45, 0, 0, 21, 142, 103, 7, 192, 49, 0, 0, 184, 73, 13, 7, 192, 77, 0, 0, 167, 234, 12, 135, 0, 160, 0, 0, 9, 118, 123, 7, 144, 164, 0, 0, 208, 238, 12, 135, 208, 164, 0, 0, 51, 151, 123, 135, 52, 165, 0, 0, 2, 16, 120, 135, 111, 165, 0, 0, 2, 44, 120, 135, 136, 165, 0, 0, 2, 52, 120, 135, 181, 165, 0, 0, 2, 57, 120, 135, 217, 165, 0, 0, 2, 69, 120, 135, 160, 166, 0, 0, 9, 76, 120, 135, 12, 168, 0, 0, 93, 162, 113, 135, 130, 168, 0, 0, 9, 164, 113, 135, 10, 169, 0, 0, 9, 78, 117, 135, 48, 169, 0, 0, 9, 76, 117, 135, 132, 169, 0, 0, 9, 42, 118, 135, 0, 170, 0, 0, 9, 38, 118, 135, 128, 170, 0, 0, 9, 132, 115, 135, 176, 215, 0, 0, 44, 177, 121, 135, 203, 215, 0, 0, 91, 248, 121, 135, 13, 0, 1, 0, 93, 99, 124, 135, 128, 0, 1, 0, 117, 101, 124, 7, 121, 1, 1, 0, 148, 240, 12, 7, 208, 1, 1, 0, 132, 241, 12, 135, 160, 2, 1, 0, 9, 248, 123, 135, 48, 3, 1, 0, 9, 252, 123, 135, 161, 3, 1, 0, 16, 205, 124, 135, 80, 4, 1, 0, 9, 86, 124, 135, 216, 4, 1, 0, 9, 232, 118, 135, 0, 5, 1, 0, 9, 93, 124, 135, 48, 5, 1, 0, 9, 95, 124, 135, 0, 6, 1, 0, 9, 105, 124, 135, 10, 8, 1, 0, 58, 115, 124, 7, 210, 9, 1, 0, 51, 118, 18, 135, 201, 10, 1, 0, 2, 167, 124, 135, 0, 11, 1, 0, 9, 119, 124, 135, 176, 15, 1, 0, 9, 203, 124, 135, 5, 16, 1, 0, 9, 244, 113, 135, 52, 16, 1, 0, 98, 245, 113, 7, 91, 16, 1, 0, 91, 147, 17, 135, 208, 16, 1, 0, 9, 97, 124, 135, 3, 17, 1, 0, 9, 96, 117, 135, 80, 17, 1, 0, 9, 168, 113, 135, 131, 17, 1, 0, 23, 170, 113, 135, 19, 18, 1, 0, 135, 172, 113, 135, 176, 18, 1, 0, 9, 174, 113, 135, 19, 19, 1, 0, 100, 178, 113, 135, 0, 20, 1, 0, 23, 180, 113, 135, 128, 20, 1, 0, 16, 183, 113, 135, 133, 21, 1, 0, 2, 193, 113, 135, 0, 22, 1, 0, 9, 201, 113, 135, 140, 22, 1, 0, 121, 203, 113, 135, 0, 24, 1, 0, 9, 209, 113, 135, 24, 25, 1, 0, 142, 205, 113, 135, 170, 25, 1, 0, 65, 207, 113, 135, 13, 26, 1, 0, 30, 198, 116, 135, 10, 28, 1, 0, 72, 248, 113, 142, 114, 28, 1, 0, 9, 4, 117, 142, 146, 28, 1, 0, 16, 4, 117, 135, 11, 29, 1, 0, 72, 224, 113, 135, 106, 29, 1, 0, 72, 229, 113, 135, 0, 32, 1, 0, 9, 207, 124, 135, 99, 34, 1, 0, 236, 223, 124, 135, 214, 34, 1, 0, 11, 227, 124, 7, 90, 36, 1, 0, 147, 119, 18, 135, 128, 36, 1, 0, 113, 232, 124, 135, 0, 48, 1, 0, 9, 4, 125, 135, 0, 68, 1, 0, 9, 83, 125, 135, 0, 104, 1, 0, 61, 78, 120, 135, 208, 106, 1, 0, 9, 94, 120, 135, 0, 107, 1, 0, 9, 240, 123, 130, 0, 112, 1, 0, 4, 100, 125, 130, 0, 136, 1, 0, 84, 148, 125, 135, 0, 139, 1, 0, 9, 166, 125, 135, 15, 176, 1, 0, 100, 104, 122, 135, 112, 177, 1, 0, 9, 155, 125, 135, 0, 188, 1, 0, 9, 88, 124, 7, 0, 208, 1, 0, 249, 242, 12, 7, 0, 209, 1, 0, 193, 249, 12, 7, 58, 209, 1, 0, 98, 251, 12, 7, 140, 209, 1, 0, 131, 252, 12, 7, 0, 211, 1, 0, 107, 236, 12, 7, 0, 216, 1, 0, 130, 58, 13, 7, 118, 218, 1, 0, 242, 72, 13, 135, 0, 225, 1, 0, 9, 243, 123, 135, 192, 226, 1, 0, 9, 245, 123, 135, 0, 232, 1, 0, 9, 95, 120, 7, 0, 240, 1, 0, 186, 2, 13, 7, 48, 240, 1, 0, 240, 3, 13, 7, 209, 240, 1, 0, 239, 7, 13, 7, 0, 243, 1, 0, 252, 11, 13, 7, 0, 246, 1, 0, 185, 41, 13, 7, 0, 247, 1, 0, 101, 48, 13, 7, 128, 247, 1, 0, 151, 51, 13, 7, 16, 248, 1, 0, 180, 54, 13, 7, 96, 248, 1, 0, 134, 56, 13, 7, 144, 248, 1, 0, 160, 57, 13, 7, 0, 249, 1, 0, 40, 33, 13, 7, 122, 249, 1, 0, 125, 36, 13, 7, 205, 249, 1, 0, 191, 38, 13, 7, 0, 250, 1, 0, 244, 8, 13, 7, 144, 250, 1, 0, 145, 40, 13, 7, 0, 251, 1, 0, 112, 188, 12, 7, 148, 251, 1, 0, 125, 192, 12, 0, 0, 0, 0, 5, 150, 0, 0, 5, 136, 0, 0, 194, 32, 9, 230, 194, 46, 29, 232, 194, 32, 108, 230, 194, 46, 29, 232, 194, 32, 207, 230, 194, 32, 29, 232, 194, 46, 207, 230, 194, 46, 29, 232, 194, 32, 85, 231, 194, 32, 29, 232, 194, 46, 85, 231, 194, 46, 29, 232, 194, 32, 184, 231, 194, 32, 29, 232, 194, 46, 184, 231, 194, 46, 29, 232, 194, 46, 29, 232, 194, 44, 59, 235, 27, 5, 54, 7, 27, 5, 54, 7, 24, 5, 131, 12, 24, 5, 90, 7, 24, 5, 90, 7, 24, 5, 90, 7, 27, 5, 98, 7, 27, 5, 98, 7, 24, 5, 90, 7, 24, 5, 0, 8, 24, 5, 0, 8, 40, 5, 0, 8, 40, 5, 0, 8, 24, 5, 0, 8, 24, 5, 0, 8, 24, 5, 0, 8, 40, 5, 0, 8, 40, 5, 0, 8, 40, 5, 0, 8, 23, 5, 142, 9, 21, 5, 0, 21, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 21, 21, 5, 0, 19, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 21, 21, 5, 0, 21, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 21, 21, 5, 0, 23, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 21, 21, 5, 0, 25, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 21, 21, 5, 0, 27, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 21, 21, 5, 0, 29, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 21, 21, 5, 0, 31, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 21, 21, 5, 0, 33, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 21, 21, 5, 0, 35, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 21, 22, 5, 0, 37, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 23, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 23, 21, 5, 0, 19, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 25, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 27, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 29, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 31, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 33, 23, 5, 144, 9, 23, 5, 142, 9, 21, 5, 0, 35, 23, 5, 144, 9, 23, 5, 142, 9, 22, 5, 0, 37, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 42, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 42, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 44, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 44, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 46, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 46, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 48, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 48, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 50, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 50, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 52, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 52, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 54, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 54, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 56, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 56, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 58, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 58, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 60, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 60, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 62, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 62, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 64, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 64, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 66, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 66, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 68, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 68, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 70, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 70, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 72, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 72, 23, 5, 144, 9, 23, 5, 142, 9, 20, 5, 0, 74, 23, 5, 144, 9, 23, 5, 142, 9, 168, 5, 0, 74, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 76, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 76, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 78, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 78, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 80, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 80, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 82, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 82, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 84, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 84, 23, 5, 144, 9, 23, 5, 142, 9, 20, 5, 0, 86, 23, 5, 144, 9, 23, 5, 142, 9, 168, 5, 0, 86, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 88, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 88, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 90, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 90, 23, 5, 144, 9, 23, 5, 142, 9, 19, 5, 0, 92, 23, 5, 144, 9, 23, 5, 142, 9, 162, 5, 0, 92, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 6, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 6, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 10, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 10, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 12, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 12, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 16, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 16, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 18, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 18, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 20, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 20, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 24, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 24, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 28, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 28, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 28, 121, 16, 5, 119, 121, 16, 5, 30, 121, 32, 5, 111, 121, 32, 5, 185, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 28, 121, 16, 5, 119, 121, 16, 5, 42, 121, 16, 5, 129, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 30, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 30, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 30, 121, 16, 5, 129, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 34, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 34, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 36, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 36, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 38, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 38, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 40, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 40, 121, 16, 5, 103, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 42, 121, 23, 5, 144, 9, 23, 5, 142, 9, 16, 5, 42, 121, 16, 5, 103, 121, 23, 5, 144, 9, 16, 5, 82, 10, 162, 5, 0, 78, 16, 5, 84, 10, 32, 5, 188, 10, 32, 5, 188, 10, 32, 5, 188, 10, 32, 5, 188, 10, 32, 5, 188, 10, 32, 5, 190, 10, 32, 5, 190, 10, 32, 5, 190, 10, 32, 5, 190, 10, 32, 5, 190, 10, 5, 5, 16, 12, 5, 136, 0, 0, 5, 5, 16, 12, 5, 138, 0, 0, 5, 5, 16, 12, 5, 148, 0, 0, 193, 9, 22, 12, 5, 136, 0, 0, 193, 9, 22, 12, 5, 138, 0, 0, 193, 9, 22, 12, 5, 148, 0, 0, 193, 16, 22, 12, 5, 136, 0, 0, 193, 16, 22, 12, 5, 138, 0, 0, 193, 16, 22, 12, 5, 148, 0, 0, 32, 5, 83, 12, 162, 5, 0, 46, 32, 5, 83, 12, 162, 5, 0, 52, 5, 5, 99, 12, 5, 158, 0, 0, 5, 5, 101, 12, 5, 158, 0, 0, 193, 2, 107, 12, 5, 158, 0, 0, 193, 126, 108, 12, 5, 158, 0, 0, 193, 140, 108, 12, 5, 158, 0, 0, 193, 154, 108, 12, 5, 158, 0, 0, 193, 2, 113, 12, 5, 158, 0, 0, 193, 2, 117, 12, 5, 158, 0, 0, 193, 16, 117, 12, 5, 158, 0, 0, 5, 5, 129, 12, 5, 158, 0, 0, 5, 5, 131, 12, 5, 158, 0, 0, 24, 5, 131, 12, 24, 5, 131, 12, 24, 5, 131, 12, 24, 5, 131, 12, 24, 5, 131, 12, 5, 5, 133, 12, 5, 158, 0, 0, 193, 107, 151, 12, 5, 158, 0, 0, 193, 114, 151, 12, 5, 158, 0, 0, 32, 5, 153, 12, 32, 5, 153, 12, 32, 5, 153, 12, 32, 5, 153, 12, 32, 5, 153, 12, 32, 5, 155, 12, 32, 5, 155, 12, 32, 5, 155, 12, 32, 5, 155, 12, 32, 5, 155, 12, 193, 79, 157, 12, 5, 158, 0, 0, 193, 121, 157, 12, 5, 158, 0, 0, 193, 128, 157, 12, 5, 158, 0, 0, 193, 142, 157, 12, 5, 158, 0, 0, 193, 170, 157, 12, 5, 158, 0, 0, 193, 49, 158, 12, 5, 158, 0, 0, 193, 63, 158, 12, 5, 158, 0, 0, 193, 70, 158, 12, 5, 158, 0, 0, 193, 126, 158, 12, 5, 158, 0, 0, 193, 133, 158, 12, 5, 158, 0, 0, 193, 140, 158, 12, 5, 158, 0, 0, 193, 147, 158, 12, 5, 158, 0, 0, 193, 154, 158, 12, 5, 158, 0, 0, 193, 161, 158, 12, 5, 158, 0, 0, 193, 168, 158, 12, 5, 158, 0, 0, 193, 175, 158, 12, 5, 158, 0, 0, 193, 196, 158, 12, 5, 158, 0, 0, 193, 203, 158, 12, 5, 158, 0, 0, 193, 210, 158, 12, 5, 158, 0, 0, 193, 217, 158, 12, 5, 158, 0, 0, 193, 19, 159, 12, 5, 158, 0, 0, 193, 26, 159, 12, 5, 158, 0, 0, 193, 138, 159, 12, 5, 158, 0, 0, 193, 180, 159, 12, 5, 158, 0, 0, 193, 187, 159, 12, 5, 158, 0, 0, 193, 201, 159, 12, 5, 158, 0, 0, 193, 222, 159, 12, 5, 158, 0, 0, 193, 229, 159, 12, 5, 158, 0, 0, 193, 236, 159, 12, 5, 158, 0, 0, 193, 243, 159, 12, 5, 158, 0, 0, 193, 29, 219, 12, 5, 158, 0, 0, 5, 5, 110, 13, 5, 174, 0, 0, 5, 5, 112, 13, 5, 174, 0, 0, 5, 5, 116, 13, 5, 174, 0, 0, 5, 5, 120, 13, 5, 174, 0, 0, 21, 5, 0, 19, 27, 5, 0, 6, 21, 5, 0, 19, 24, 5, 0, 8, 51, 5, 0, 19, 32, 5, 149, 12, 51, 5, 0, 25, 21, 5, 0, 21, 27, 5, 0, 6, 21, 5, 0, 21, 24, 5, 0, 8, 51, 5, 0, 21, 32, 5, 149, 12, 51, 5, 0, 21, 32, 5, 149, 12, 51, 5, 0, 21, 51, 5, 0, 19, 51, 5, 0, 21, 32, 5, 149, 12, 51, 5, 0, 23, 51, 5, 0, 21, 32, 5, 149, 12, 51, 5, 0, 25, 51, 5, 0, 21, 32, 5, 149, 12, 51, 5, 0, 27, 51, 5, 0, 21, 32, 5, 149, 12, 51, 5, 0, 29, 51, 5, 0, 21, 32, 5, 149, 12, 51, 5, 0, 31, 51, 5, 0, 21, 32, 5, 149, 12, 51, 5, 0, 33, 51, 5, 0, 21, 32, 5, 149, 12, 51, 5, 0, 35, 51, 5, 0, 21, 32, 5, 149, 12, 52, 5, 0, 37, 31, 5, 0, 21, 31, 5, 0, 19, 21, 5, 0, 21, 21, 5, 0, 19, 24, 5, 0, 8, 31, 5, 0, 21, 31, 5, 0, 21, 21, 5, 0, 21, 21, 5, 0, 21, 24, 5, 0, 8, 31, 5, 0, 21, 31, 5, 0, 23, 21, 5, 0, 21, 21, 5, 0, 23, 24, 5, 0, 8, 46, 5, 0, 21, 46, 5, 0, 23, 46, 5, 0, 19, 180, 5, 0, 72, 31, 5, 0, 21, 31, 5, 0, 25, 21, 5, 0, 21, 21, 5, 0, 25, 24, 5, 0, 8, 31, 5, 0, 21, 31, 5, 0, 27, 21, 5, 0, 21, 21, 5, 0, 27, 24, 5, 0, 8, 31, 5, 0, 21, 31, 5, 0, 29, 21, 5, 0, 21, 21, 5, 0, 29, 24, 5, 0, 8, 31, 5, 0, 21, 31, 5, 0, 31, 21, 5, 0, 21, 21, 5, 0, 31, 24, 5, 0, 8, 31, 5, 0, 21, 31, 5, 0, 33, 21, 5, 0, 21, 21, 5, 0, 33, 24, 5, 0, 8, 31, 5, 0, 21, 31, 5, 0, 35, 21, 5, 0, 21, 21, 5, 0, 35, 24, 5, 0, 8, 31, 5, 0, 21, 34, 5, 0, 37, 21, 5, 0, 21, 22, 5, 0, 37, 24, 5, 0, 8, 21, 5, 0, 23, 27, 5, 0, 6, 51, 5, 0, 23, 32, 5, 149, 12, 51, 5, 0, 25, 51, 5, 0, 23, 32, 5, 149, 12, 51, 5, 0, 29, 31, 5, 0, 23, 31, 5, 0, 19, 21, 5, 0, 23, 21, 5, 0, 19, 24, 5, 0, 8, 31, 5, 0, 23, 31, 5, 0, 21, 31, 5, 0, 23, 31, 5, 0, 23, 46, 5, 0, 23, 46, 5, 0, 23, 48, 5, 0, 8, 46, 5, 0, 23, 31, 5, 0, 23, 31, 5, 0, 25, 31, 5, 0, 23, 31, 5, 0, 27, 31, 5, 0, 23, 31, 5, 0, 29, 31, 5, 0, 23, 31, 5, 0, 31, 31, 5, 0, 23, 31, 5, 0, 33, 31, 5, 0, 23, 31, 5, 0, 35, 31, 5, 0, 23, 34, 5, 0, 37, 46, 5, 0, 23, 180, 5, 0, 62, 46, 5, 0, 23, 49, 5, 0, 68, 46, 5, 0, 48, 48, 5, 0, 4, 180, 5, 0, 78, 46, 5, 0, 46, 49, 5, 0, 76, 21, 5, 0, 25, 27, 5, 0, 6, 51, 5, 0, 25, 32, 5, 149, 12, 51, 5, 0, 27, 51, 5, 0, 25, 32, 5, 149, 12, 51, 5, 0, 29, 51, 5, 0, 25, 32, 5, 149, 12, 51, 5, 0, 35, 31, 5, 0, 25, 31, 5, 0, 19, 31, 5, 0, 25, 31, 5, 0, 21, 31, 5, 0, 25, 31, 5, 0, 25, 31, 5, 0, 25, 31, 5, 0, 27, 31, 5, 0, 25, 31, 5, 0, 29, 31, 5, 0, 25, 31, 5, 0, 31, 31, 5, 0, 25, 31, 5, 0, 33, 31, 5, 0, 25, 31, 5, 0, 35, 31, 5, 0, 25, 34, 5, 0, 37, 46, 5, 0, 25, 177, 5, 0, 48, 21, 5, 0, 27, 27, 5, 0, 6, 51, 5, 0, 27, 32, 5, 149, 12, 51, 5, 0, 29, 31, 5, 0, 27, 31, 5, 0, 19, 31, 5, 0, 27, 31, 5, 0, 21, 31, 5, 0, 27, 31, 5, 0, 27, 31, 5, 0, 27, 31, 5, 0, 29, 31, 5, 0, 27, 31, 5, 0, 31, 31, 5, 0, 27, 31, 5, 0, 33, 31, 5, 0, 27, 31, 5, 0, 35, 31, 5, 0, 27, 34, 5, 0, 37, 46, 5, 0, 27, 180, 5, 0, 62, 21, 5, 0, 29, 27, 5, 0, 6, 46, 5, 0, 29, 48, 5, 0, 8, 46, 5, 0, 21, 51, 5, 0, 29, 32, 5, 149, 12, 51, 5, 0, 31, 51, 5, 0, 29, 32, 5, 149, 12, 51, 5, 0, 35, 31, 5, 0, 29, 31, 5, 0, 19, 21, 5, 0, 31, 27, 5, 0, 6, 31, 5, 0, 31, 31, 5, 0, 19, 46, 5, 0, 31, 46, 5, 0, 19, 180, 5, 0, 72, 21, 5, 0, 33, 27, 5, 0, 6, 46, 5, 0, 33, 48, 5, 0, 8, 46, 5, 0, 21, 51, 5, 0, 33, 32, 5, 149, 12, 51, 5, 0, 35, 31, 5, 0, 33, 31, 5, 0, 19, 21, 5, 0, 35, 27, 5, 0, 6, 31, 5, 0, 35, 31, 5, 0, 19, 46, 5, 0, 35, 180, 5, 0, 62, 22, 5, 0, 37, 27, 5, 0, 6, 5, 5, 0, 42, 5, 140, 0, 0, 5, 136, 0, 0, 156, 5, 0, 42, 5, 140, 0, 0, 5, 136, 0, 0, 5, 5, 0, 42, 5, 140, 0, 0, 5, 138, 0, 0, 156, 5, 0, 42, 5, 140, 0, 0, 5, 138, 0, 0, 5, 5, 0, 42, 5, 140, 0, 0, 5, 154, 0, 0, 156, 5, 0, 42, 5, 140, 0, 0, 5, 154, 0, 0, 5, 5, 0, 42, 5, 140, 0, 0, 5, 182, 0, 0, 156, 5, 0, 42, 5, 140, 0, 0, 5, 182, 0, 0, 5, 5, 0, 42, 5, 142, 0, 0, 5, 136, 0, 0, 156, 5, 0, 42, 5, 142, 0, 0, 5, 136, 0, 0, 5, 5, 0, 42, 5, 142, 0, 0, 5, 138, 0, 0, 156, 5, 0, 42, 5, 142, 0, 0, 5, 138, 0, 0, 5, 5, 0, 42, 5, 142, 0, 0, 5, 154, 0, 0, 156, 5, 0, 42, 5, 142, 0, 0, 5, 154, 0, 0, 5, 5, 0, 42, 5, 142, 0, 0, 5, 182, 0, 0, 156, 5, 0, 42, 5, 142, 0, 0, 5, 182, 0, 0, 5, 5, 0, 42, 5, 146, 0, 0, 5, 136, 0, 0, 156, 5, 0, 42, 5, 146, 0, 0, 5, 136, 0, 0, 19, 5, 0, 42, 32, 150, 0, 0, 162, 5, 0, 42, 32, 150, 0, 0, 5, 5, 0, 42, 5, 150, 0, 0, 5, 164, 0, 0, 156, 5, 0, 42, 5, 150, 0, 0, 5, 164, 0, 0, 5, 5, 0, 42, 5, 156, 0, 0, 5, 164, 0, 0, 156, 5, 0, 42, 5, 156, 0, 0, 5, 164, 0, 0, 5, 5, 0, 42, 5, 196, 0, 0, 5, 140, 0, 0, 156, 5, 0, 42, 5, 196, 0, 0, 5, 140, 0, 0, 5, 5, 0, 42, 5, 196, 0, 0, 5, 142, 0, 0, 156, 5, 0, 42, 5, 196, 0, 0, 5, 142, 0, 0, 49, 5, 0, 42, 48, 5, 0, 8, 49, 5, 0, 66, 48, 5, 0, 8, 19, 5, 0, 42, 46, 5, 138, 10, 19, 5, 0, 46, 19, 5, 0, 42, 46, 5, 138, 10, 19, 5, 0, 78, 162, 5, 0, 42, 46, 5, 138, 10, 162, 5, 0, 78, 180, 5, 0, 42, 32, 5, 147, 12, 49, 5, 0, 66, 19, 5, 0, 42, 19, 5, 0, 42, 162, 5, 0, 42, 162, 5, 0, 42, 180, 5, 0, 42, 177, 5, 0, 44, 5, 112, 0, 42, 19, 5, 0, 50, 160, 112, 0, 42, 162, 5, 0, 50, 46, 112, 0, 42, 43, 5, 0, 50, 5, 112, 0, 42, 19, 5, 0, 50, 5, 136, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 136, 0, 0, 5, 112, 0, 42, 19, 5, 0, 50, 5, 164, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 164, 0, 0, 19, 5, 0, 42, 19, 5, 0, 70, 162, 5, 0, 42, 162, 5, 0, 70, 19, 5, 0, 42, 19, 5, 0, 82, 162, 5, 0, 42, 162, 5, 0, 82, 180, 5, 0, 42, 177, 5, 0, 82, 19, 5, 0, 42, 19, 5, 0, 84, 162, 5, 0, 42, 162, 5, 0, 84, 5, 112, 0, 42, 19, 5, 0, 84, 160, 112, 0, 42, 162, 5, 0, 84, 19, 5, 0, 42, 19, 5, 0, 90, 162, 5, 0, 42, 162, 5, 0, 90, 19, 5, 0, 42, 32, 5, 86, 93, 46, 5, 0, 44, 49, 5, 0, 42, 49, 5, 0, 76, 177, 5, 0, 44, 52, 5, 0, 74, 19, 5, 0, 46, 32, 160, 0, 0, 5, 5, 0, 46, 5, 160, 0, 0, 5, 136, 0, 0, 156, 5, 0, 46, 5, 160, 0, 0, 5, 136, 0, 0, 19, 5, 0, 46, 46, 5, 138, 10, 19, 5, 0, 70, 19, 5, 0, 46, 46, 5, 138, 10, 19, 5, 0, 82, 177, 5, 0, 46, 32, 5, 147, 12, 49, 5, 0, 62, 46, 5, 0, 54, 46, 5, 0, 46, 49, 5, 0, 42, 49, 5, 0, 64, 46, 5, 0, 46, 46, 5, 0, 46, 168, 5, 0, 46, 168, 5, 0, 48, 46, 5, 0, 46, 46, 5, 0, 48, 177, 5, 0, 46, 180, 5, 0, 64, 46, 5, 0, 46, 49, 5, 0, 66, 46, 5, 0, 46, 49, 5, 0, 66, 46, 5, 0, 23, 46, 5, 0, 46, 49, 5, 0, 66, 46, 5, 0, 25, 177, 5, 0, 46, 49, 5, 0, 70, 48, 5, 0, 8, 177, 5, 0, 46, 180, 5, 0, 70, 180, 5, 0, 70, 180, 5, 0, 64, 46, 5, 0, 48, 49, 5, 0, 42, 19, 5, 0, 48, 19, 5, 0, 44, 46, 5, 0, 48, 177, 5, 0, 44, 177, 5, 0, 48, 177, 5, 0, 60, 46, 5, 0, 48, 49, 5, 0, 64, 46, 5, 0, 48, 49, 5, 0, 66, 46, 5, 0, 48, 49, 5, 0, 66, 46, 5, 0, 23, 46, 5, 0, 48, 49, 5, 0, 66, 46, 5, 0, 25, 19, 5, 0, 48, 19, 5, 0, 92, 162, 5, 0, 48, 19, 5, 0, 92, 162, 5, 0, 48, 162, 5, 0, 92, 19, 5, 0, 48, 19, 5, 0, 92, 32, 144, 0, 0, 162, 5, 0, 48, 19, 5, 0, 92, 32, 144, 0, 0, 162, 5, 0, 48, 162, 5, 0, 92, 32, 144, 0, 0, 19, 5, 0, 48, 32, 5, 16, 93, 19, 5, 0, 48, 32, 5, 18, 93, 19, 5, 0, 48, 16, 5, 26, 93, 5, 5, 0, 50, 5, 142, 0, 0, 5, 136, 0, 0, 156, 5, 0, 50, 5, 142, 0, 0, 5, 136, 0, 0, 5, 5, 0, 50, 5, 142, 0, 0, 5, 138, 0, 0, 156, 5, 0, 50, 5, 142, 0, 0, 5, 138, 0, 0, 5, 5, 0, 50, 5, 142, 0, 0, 5, 154, 0, 0, 156, 5, 0, 50, 5, 142, 0, 0, 5, 154, 0, 0, 5, 5, 0, 50, 5, 142, 0, 0, 5, 182, 0, 0, 156, 5, 0, 50, 5, 142, 0, 0, 5, 182, 0, 0, 5, 5, 0, 50, 5, 160, 0, 0, 5, 140, 0, 0, 156, 5, 0, 50, 5, 160, 0, 0, 5, 140, 0, 0, 5, 5, 0, 50, 5, 164, 0, 0, 5, 136, 0, 0, 156, 5, 0, 50, 5, 164, 0, 0, 5, 136, 0, 0, 5, 5, 0, 50, 5, 164, 0, 0, 5, 138, 0, 0, 156, 5, 0, 50, 5, 164, 0, 0, 5, 138, 0, 0, 5, 5, 0, 50, 5, 196, 0, 0, 5, 142, 0, 0, 156, 5, 0, 50, 5, 196, 0, 0, 5, 142, 0, 0, 49, 5, 0, 50, 49, 5, 0, 76, 46, 5, 0, 54, 49, 5, 0, 50, 180, 5, 0, 84, 162, 5, 0, 52, 162, 5, 0, 42, 162, 5, 0, 88, 19, 5, 0, 52, 19, 5, 0, 52, 19, 5, 0, 52, 19, 5, 0, 52, 19, 5, 0, 58, 19, 5, 0, 52, 19, 5, 0, 52, 19, 5, 0, 64, 46, 5, 0, 52, 49, 5, 0, 66, 19, 5, 0, 52, 16, 5, 30, 69, 177, 5, 0, 52, 180, 5, 0, 76, 180, 5, 0, 50, 180, 5, 0, 50, 19, 5, 0, 54, 32, 170, 0, 0, 162, 5, 0, 54, 32, 170, 0, 0, 46, 5, 0, 54, 49, 5, 0, 42, 49, 5, 0, 64, 177, 5, 0, 54, 177, 5, 0, 44, 177, 5, 0, 54, 180, 5, 0, 56, 46, 5, 0, 92, 177, 5, 0, 54, 180, 5, 0, 72, 49, 5, 0, 42, 177, 5, 0, 54, 46, 5, 0, 90, 43, 5, 0, 56, 32, 178, 0, 0, 49, 5, 0, 56, 49, 5, 0, 42, 180, 5, 0, 56, 46, 5, 0, 46, 180, 5, 0, 56, 46, 5, 0, 48, 180, 5, 0, 76, 180, 5, 0, 56, 46, 5, 0, 54, 180, 5, 0, 56, 49, 5, 0, 58, 48, 5, 0, 4, 180, 5, 0, 76, 49, 5, 0, 50, 49, 5, 0, 78, 180, 5, 0, 56, 180, 5, 0, 72, 49, 5, 0, 56, 180, 5, 0, 72, 49, 5, 0, 42, 180, 5, 0, 56, 180, 5, 0, 84, 5, 5, 0, 58, 5, 150, 0, 0, 5, 136, 0, 0, 156, 5, 0, 58, 5, 150, 0, 0, 5, 136, 0, 0, 180, 5, 0, 58, 177, 5, 0, 46, 180, 5, 0, 58, 177, 5, 0, 48, 19, 5, 0, 58, 19, 5, 0, 58, 162, 5, 0, 58, 162, 5, 0, 58, 19, 5, 0, 58, 19, 5, 0, 58, 19, 5, 0, 58, 162, 5, 0, 58, 162, 5, 0, 58, 162, 5, 0, 58, 19, 5, 0, 58, 19, 5, 0, 60, 162, 5, 0, 58, 162, 5, 0, 60, 49, 5, 0, 58, 49, 5, 0, 68, 180, 5, 0, 58, 177, 5, 0, 82, 19, 5, 0, 58, 19, 5, 0, 84, 162, 5, 0, 58, 162, 5, 0, 84, 19, 5, 0, 58, 19, 5, 0, 88, 162, 5, 0, 58, 162, 5, 0, 88, 19, 5, 0, 62, 32, 170, 0, 0, 162, 5, 0, 62, 32, 170, 0, 0, 49, 5, 0, 62, 180, 5, 0, 42, 180, 5, 0, 62, 177, 5, 0, 44, 49, 5, 0, 62, 46, 5, 0, 46, 49, 5, 0, 42, 49, 5, 0, 64, 49, 5, 0, 62, 180, 5, 0, 56, 46, 5, 0, 92, 180, 5, 0, 62, 180, 5, 0, 62, 49, 5, 0, 62, 49, 5, 0, 64, 49, 5, 0, 62, 49, 5, 0, 66, 180, 5, 0, 62, 180, 5, 0, 66, 49, 5, 0, 62, 49, 5, 0, 66, 46, 5, 0, 23, 49, 5, 0, 62, 49, 5, 0, 66, 46, 5, 0, 25, 49, 5, 0, 62, 180, 5, 0, 72, 49, 5, 0, 42, 49, 5, 0, 62, 49, 5, 0, 80, 49, 5, 0, 62, 180, 5, 0, 84, 49, 5, 0, 62, 184, 5, 0, 86, 49, 5, 0, 62, 177, 5, 86, 96, 5, 5, 0, 64, 5, 196, 0, 0, 5, 164, 0, 0, 156, 5, 0, 64, 5, 196, 0, 0, 5, 164, 0, 0, 5, 5, 0, 64, 194, 5, 219, 251, 156, 5, 0, 64, 194, 5, 219, 251, 19, 5, 0, 64, 19, 5, 0, 60, 162, 5, 0, 64, 19, 5, 0, 60, 162, 5, 0, 64, 162, 5, 0, 60, 19, 5, 0, 64, 19, 5, 0, 64, 162, 5, 0, 64, 162, 5, 0, 64, 49, 5, 0, 64, 49, 5, 0, 66, 49, 5, 0, 64, 49, 5, 0, 68, 49, 5, 0, 64, 49, 5, 0, 70, 46, 5, 0, 54, 180, 5, 0, 64, 49, 5, 0, 70, 49, 5, 0, 78, 49, 5, 0, 78, 49, 5, 0, 64, 49, 5, 0, 50, 49, 5, 0, 78, 49, 5, 0, 78, 19, 5, 0, 64, 19, 5, 0, 78, 180, 5, 0, 64, 180, 5, 0, 80, 177, 5, 0, 48, 49, 5, 0, 64, 49, 5, 0, 88, 19, 5, 0, 64, 19, 5, 0, 92, 49, 5, 0, 66, 32, 5, 147, 12, 49, 5, 0, 78, 49, 5, 0, 66, 32, 5, 147, 12, 49, 5, 0, 78, 46, 5, 0, 23, 49, 5, 0, 66, 180, 5, 0, 42, 49, 5, 0, 66, 46, 5, 0, 44, 180, 5, 0, 66, 177, 5, 0, 44, 43, 5, 0, 66, 43, 5, 0, 46, 43, 5, 0, 66, 43, 5, 0, 48, 49, 5, 0, 66, 46, 5, 0, 54, 180, 5, 0, 66, 180, 5, 0, 56, 46, 5, 0, 92, 49, 5, 0, 66, 49, 5, 0, 58, 49, 5, 0, 64, 49, 5, 0, 66, 49, 5, 0, 66, 49, 5, 0, 66, 49, 5, 0, 66, 46, 5, 0, 23, 49, 5, 0, 66, 49, 5, 0, 66, 46, 5, 0, 25, 49, 5, 0, 66, 49, 5, 0, 70, 49, 5, 0, 64, 180, 5, 0, 66, 180, 5, 0, 72, 49, 5, 0, 42, 43, 5, 0, 66, 43, 5, 0, 76, 49, 5, 0, 66, 49, 5, 0, 78, 49, 5, 0, 66, 180, 5, 0, 84, 180, 5, 0, 66, 180, 5, 0, 84, 49, 5, 0, 66, 184, 5, 0, 86, 180, 5, 0, 66, 184, 5, 0, 86, 180, 5, 0, 66, 177, 5, 86, 96, 19, 5, 0, 68, 32, 170, 0, 0, 162, 5, 0, 68, 32, 170, 0, 0, 49, 5, 0, 68, 180, 5, 0, 42, 180, 5, 0, 68, 180, 5, 0, 50, 184, 5, 0, 86, 49, 5, 0, 68, 177, 5, 0, 52, 180, 5, 0, 68, 177, 5, 0, 54, 19, 5, 0, 68, 19, 5, 0, 60, 162, 5, 0, 68, 19, 5, 0, 60, 162, 5, 0, 68, 162, 5, 0, 60, 49, 5, 0, 68, 49, 5, 0, 66, 162, 5, 0, 68, 19, 5, 0, 70, 49, 5, 0, 68, 49, 5, 0, 78, 49, 5, 0, 68, 180, 5, 0, 84, 49, 5, 0, 68, 184, 5, 0, 86, 5, 5, 0, 70, 5, 142, 0, 0, 5, 136, 0, 0, 156, 5, 0, 70, 5, 142, 0, 0, 5, 136, 0, 0, 5, 5, 0, 70, 5, 142, 0, 0, 5, 138, 0, 0, 156, 5, 0, 70, 5, 142, 0, 0, 5, 138, 0, 0, 5, 5, 0, 70, 5, 142, 0, 0, 5, 154, 0, 0, 156, 5, 0, 70, 5, 142, 0, 0, 5, 154, 0, 0, 5, 5, 0, 70, 5, 142, 0, 0, 5, 182, 0, 0, 156, 5, 0, 70, 5, 142, 0, 0, 5, 182, 0, 0, 19, 5, 0, 70, 32, 150, 0, 0, 162, 5, 0, 70, 32, 150, 0, 0, 5, 5, 0, 70, 5, 150, 0, 0, 5, 164, 0, 0, 156, 5, 0, 70, 5, 150, 0, 0, 5, 164, 0, 0, 5, 5, 0, 70, 5, 154, 0, 0, 5, 136, 0, 0, 156, 5, 0, 70, 5, 154, 0, 0, 5, 136, 0, 0, 5, 5, 0, 70, 5, 154, 0, 0, 5, 150, 0, 0, 156, 5, 0, 70, 5, 154, 0, 0, 5, 150, 0, 0, 5, 5, 0, 70, 5, 154, 0, 0, 5, 164, 0, 0, 156, 5, 0, 70, 5, 154, 0, 0, 5, 164, 0, 0, 5, 5, 0, 70, 5, 156, 0, 0, 5, 164, 0, 0, 156, 5, 0, 70, 5, 156, 0, 0, 5, 164, 0, 0, 5, 5, 0, 70, 5, 158, 0, 0, 5, 136, 0, 0, 156, 5, 0, 70, 5, 158, 0, 0, 5, 136, 0, 0, 5, 5, 0, 70, 5, 162, 0, 0, 5, 164, 0, 0, 156, 5, 0, 70, 5, 162, 0, 0, 5, 164, 0, 0, 5, 5, 0, 70, 5, 164, 0, 0, 5, 136, 0, 0, 156, 5, 0, 70, 5, 164, 0, 0, 5, 136, 0, 0, 5, 5, 0, 70, 5, 164, 0, 0, 5, 138, 0, 0, 156, 5, 0, 70, 5, 164, 0, 0, 5, 138, 0, 0, 19, 5, 0, 70, 32, 168, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 136, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 136, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 138, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 138, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 154, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 154, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 182, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 182, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 196, 0, 0, 5, 142, 0, 0, 156, 5, 0, 70, 5, 196, 0, 0, 5, 142, 0, 0, 5, 112, 0, 70, 19, 5, 0, 50, 160, 112, 0, 70, 162, 5, 0, 50, 46, 112, 0, 70, 43, 5, 0, 50, 180, 5, 0, 70, 180, 5, 0, 62, 19, 5, 0, 70, 19, 5, 0, 70, 162, 5, 0, 70, 162, 5, 0, 70, 49, 5, 0, 70, 180, 5, 0, 84, 49, 5, 0, 72, 48, 5, 0, 8, 49, 5, 0, 66, 48, 5, 0, 8, 49, 5, 0, 72, 180, 5, 0, 42, 180, 5, 0, 72, 180, 5, 0, 42, 49, 5, 0, 72, 46, 5, 0, 46, 49, 5, 0, 72, 177, 5, 0, 52, 180, 5, 0, 72, 180, 5, 0, 56, 180, 5, 0, 72, 180, 5, 0, 72, 180, 5, 0, 66, 180, 5, 0, 72, 180, 5, 0, 72, 180, 5, 0, 84, 180, 5, 0, 72, 180, 5, 0, 76, 49, 5, 0, 72, 49, 5, 0, 78, 180, 5, 0, 72, 180, 5, 0, 80, 180, 5, 0, 50, 49, 5, 0, 72, 180, 5, 0, 84, 49, 5, 0, 72, 184, 5, 0, 86, 20, 5, 0, 74, 19, 5, 0, 72, 19, 5, 0, 76, 32, 170, 0, 0, 162, 5, 0, 76, 32, 170, 0, 0, 5, 5, 0, 76, 5, 196, 0, 0, 5, 164, 0, 0, 156, 5, 0, 76, 5, 196, 0, 0, 5, 164, 0, 0, 49, 5, 0, 76, 49, 5, 0, 42, 46, 5, 0, 48, 49, 5, 0, 76, 49, 5, 0, 42, 46, 5, 0, 48, 32, 5, 147, 12, 49, 5, 0, 78, 49, 5, 0, 76, 49, 5, 0, 42, 46, 5, 0, 48, 32, 5, 147, 12, 49, 5, 0, 78, 46, 5, 0, 23, 5, 5, 0, 78, 5, 136, 0, 0, 5, 156, 0, 0, 156, 5, 0, 78, 5, 136, 0, 0, 5, 156, 0, 0, 5, 5, 0, 78, 5, 144, 0, 0, 5, 156, 0, 0, 156, 5, 0, 78, 5, 144, 0, 0, 5, 156, 0, 0, 19, 5, 0, 78, 32, 170, 0, 0, 162, 5, 0, 78, 32, 170, 0, 0, 5, 5, 0, 78, 5, 196, 0, 0, 5, 156, 0, 0, 156, 5, 0, 78, 5, 196, 0, 0, 5, 156, 0, 0, 5, 120, 0, 78, 5, 156, 0, 0, 180, 5, 0, 78, 180, 5, 0, 42, 180, 5, 0, 78, 177, 5, 0, 48, 180, 5, 0, 78, 180, 5, 0, 56, 180, 5, 0, 84, 43, 5, 0, 78, 43, 5, 0, 66, 180, 5, 0, 78, 180, 5, 0, 70, 180, 5, 0, 78, 180, 5, 0, 78, 180, 5, 0, 78, 5, 112, 0, 78, 19, 5, 0, 78, 160, 112, 0, 78, 162, 5, 0, 78, 19, 5, 0, 78, 19, 5, 0, 80, 5, 120, 0, 78, 19, 5, 0, 80, 180, 5, 0, 78, 49, 5, 0, 84, 19, 5, 0, 80, 32, 5, 16, 47, 162, 5, 0, 80, 162, 5, 0, 50, 162, 5, 0, 64, 5, 112, 0, 80, 19, 5, 0, 56, 180, 5, 0, 80, 180, 5, 0, 56, 46, 5, 0, 92, 43, 5, 0, 80, 43, 5, 0, 66, 19, 5, 0, 80, 19, 5, 0, 78, 19, 5, 0, 80, 16, 5, 14, 79, 19, 5, 0, 80, 16, 5, 22, 79, 19, 5, 0, 80, 19, 5, 0, 92, 162, 5, 0, 80, 19, 5, 0, 92, 19, 5, 0, 82, 32, 150, 0, 0, 162, 5, 0, 82, 32, 150, 0, 0, 5, 5, 0, 82, 5, 150, 0, 0, 5, 136, 0, 0, 156, 5, 0, 82, 5, 150, 0, 0, 5, 136, 0, 0, 5, 5, 0, 82, 5, 150, 0, 0, 5, 138, 0, 0, 156, 5, 0, 82, 5, 150, 0, 0, 5, 138, 0, 0, 5, 5, 0, 82, 5, 150, 0, 0, 5, 144, 0, 0, 156, 5, 0, 82, 5, 150, 0, 0, 5, 144, 0, 0, 5, 5, 0, 82, 5, 150, 0, 0, 5, 164, 0, 0, 156, 5, 0, 82, 5, 150, 0, 0, 5, 164, 0, 0, 5, 5, 0, 82, 5, 154, 0, 0, 5, 136, 0, 0, 156, 5, 0, 82, 5, 154, 0, 0, 5, 136, 0, 0, 5, 5, 0, 82, 5, 164, 0, 0, 5, 150, 0, 0, 156, 5, 0, 82, 5, 164, 0, 0, 5, 150, 0, 0, 19, 5, 0, 82, 32, 168, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 136, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 136, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 138, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 138, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 154, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 154, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 182, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 182, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 177, 5, 0, 82, 180, 5, 0, 56, 177, 5, 0, 48, 177, 5, 0, 82, 180, 5, 0, 72, 48, 5, 90, 7, 180, 5, 0, 84, 32, 5, 147, 12, 49, 5, 0, 66, 19, 5, 0, 84, 19, 5, 0, 58, 162, 5, 0, 84, 162, 5, 0, 58, 19, 5, 0, 84, 19, 5, 0, 58, 19, 5, 0, 58, 162, 5, 0, 84, 162, 5, 0, 58, 162, 5, 0, 58, 19, 5, 0, 84, 19, 5, 0, 58, 19, 5, 0, 58, 19, 5, 0, 58, 162, 5, 0, 84, 162, 5, 0, 58, 162, 5, 0, 58, 162, 5, 0, 58, 180, 5, 0, 84, 180, 5, 0, 70, 177, 5, 0, 48, 180, 5, 0, 84, 180, 5, 0, 78, 19, 5, 0, 84, 19, 5, 0, 90, 162, 5, 0, 84, 162, 5, 0, 90, 184, 5, 0, 86, 46, 5, 0, 44, 184, 5, 0, 86, 177, 5, 0, 46, 176, 5, 0, 86, 168, 5, 0, 92, 19, 5, 0, 88, 19, 5, 0, 58, 162, 5, 0, 88, 162, 5, 0, 58, 19, 5, 0, 88, 19, 5, 0, 58, 19, 5, 0, 58, 162, 5, 0, 88, 162, 5, 0, 58, 162, 5, 0, 58, 19, 5, 0, 92, 20, 5, 0, 86, 5, 5, 26, 93, 5, 144, 0, 0, 158, 5, 26, 93, 5, 144, 0, 0, 32, 5, 82, 93, 19, 5, 0, 68, 5, 5, 6, 96, 5, 132, 0, 0, 166, 5, 6, 96, 5, 132, 0, 0, 5, 5, 6, 96, 5, 132, 0, 0, 5, 136, 0, 0, 166, 5, 6, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 5, 6, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 166, 5, 6, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 132, 0, 0, 5, 138, 0, 0, 166, 5, 6, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 5, 6, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 166, 5, 6, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 132, 0, 0, 5, 148, 0, 0, 166, 5, 6, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 5, 6, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 166, 5, 6, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 132, 0, 0, 5, 216, 0, 0, 166, 5, 6, 96, 5, 132, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 134, 0, 0, 166, 5, 6, 96, 5, 134, 0, 0, 5, 5, 6, 96, 5, 134, 0, 0, 5, 136, 0, 0, 166, 5, 6, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 5, 6, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 166, 5, 6, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 134, 0, 0, 5, 138, 0, 0, 166, 5, 6, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 5, 6, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 166, 5, 6, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 134, 0, 0, 5, 148, 0, 0, 166, 5, 6, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 5, 6, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 166, 5, 6, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 134, 0, 0, 5, 216, 0, 0, 166, 5, 6, 96, 5, 134, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 136, 0, 0, 166, 5, 6, 96, 5, 136, 0, 0, 5, 5, 6, 96, 5, 136, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 138, 0, 0, 166, 5, 6, 96, 5, 138, 0, 0, 5, 5, 6, 96, 5, 138, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 140, 0, 0, 166, 5, 6, 96, 5, 140, 0, 0, 5, 5, 6, 96, 5, 148, 0, 0, 5, 5, 6, 96, 5, 148, 0, 0, 5, 216, 0, 0, 5, 5, 6, 96, 5, 164, 0, 0, 166, 5, 6, 96, 5, 164, 0, 0, 5, 5, 6, 96, 5, 216, 0, 0, 166, 5, 6, 96, 5, 216, 0, 0, 5, 5, 16, 96, 5, 132, 0, 0, 166, 5, 16, 96, 5, 132, 0, 0, 5, 5, 16, 96, 5, 132, 0, 0, 5, 136, 0, 0, 166, 5, 16, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 5, 16, 96, 5, 132, 0, 0, 5, 138, 0, 0, 166, 5, 16, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 5, 16, 96, 5, 134, 0, 0, 166, 5, 16, 96, 5, 134, 0, 0, 5, 5, 16, 96, 5, 134, 0, 0, 5, 136, 0, 0, 166, 5, 16, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 5, 16, 96, 5, 134, 0, 0, 5, 138, 0, 0, 166, 5, 16, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 5, 16, 96, 5, 136, 0, 0, 166, 5, 16, 96, 5, 136, 0, 0, 5, 5, 16, 96, 5, 138, 0, 0, 166, 5, 16, 96, 5, 138, 0, 0, 5, 5, 28, 96, 5, 132, 0, 0, 158, 5, 28, 96, 5, 132, 0, 0, 5, 5, 28, 96, 5, 132, 0, 0, 5, 136, 0, 0, 158, 5, 28, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 5, 28, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 158, 5, 28, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 132, 0, 0, 5, 138, 0, 0, 158, 5, 28, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 5, 28, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 158, 5, 28, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 132, 0, 0, 5, 148, 0, 0, 158, 5, 28, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 5, 28, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 158, 5, 28, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 132, 0, 0, 5, 216, 0, 0, 158, 5, 28, 96, 5, 132, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 134, 0, 0, 158, 5, 28, 96, 5, 134, 0, 0, 5, 5, 28, 96, 5, 134, 0, 0, 5, 136, 0, 0, 158, 5, 28, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 5, 28, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 158, 5, 28, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 134, 0, 0, 5, 138, 0, 0, 158, 5, 28, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 5, 28, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 158, 5, 28, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 134, 0, 0, 5, 148, 0, 0, 158, 5, 28, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 5, 28, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 158, 5, 28, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 134, 0, 0, 5, 216, 0, 0, 158, 5, 28, 96, 5, 134, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 136, 0, 0, 158, 5, 28, 96, 5, 136, 0, 0, 5, 5, 28, 96, 5, 136, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 138, 0, 0, 158, 5, 28, 96, 5, 138, 0, 0, 5, 5, 28, 96, 5, 138, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 148, 0, 0, 5, 5, 28, 96, 5, 148, 0, 0, 5, 216, 0, 0, 5, 5, 28, 96, 5, 216, 0, 0, 158, 5, 28, 96, 5, 216, 0, 0, 5, 5, 32, 96, 5, 132, 0, 0, 166, 5, 32, 96, 5, 132, 0, 0, 5, 5, 32, 96, 5, 132, 0, 0, 5, 136, 0, 0, 166, 5, 32, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 5, 32, 96, 5, 132, 0, 0, 5, 138, 0, 0, 166, 5, 32, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 5, 32, 96, 5, 132, 0, 0, 5, 148, 0, 0, 166, 5, 32, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 5, 32, 96, 5, 134, 0, 0, 166, 5, 32, 96, 5, 134, 0, 0, 5, 5, 32, 96, 5, 134, 0, 0, 5, 136, 0, 0, 166, 5, 32, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 5, 32, 96, 5, 134, 0, 0, 5, 138, 0, 0, 166, 5, 32, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 5, 32, 96, 5, 134, 0, 0, 5, 148, 0, 0, 166, 5, 32, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 5, 32, 96, 5, 136, 0, 0, 166, 5, 32, 96, 5, 136, 0, 0, 5, 5, 32, 96, 5, 138, 0, 0, 166, 5, 32, 96, 5, 138, 0, 0, 5, 5, 32, 96, 5, 140, 0, 0, 166, 5, 32, 96, 5, 140, 0, 0, 5, 5, 32, 96, 5, 148, 0, 0, 5, 5, 32, 96, 5, 150, 0, 0, 166, 5, 32, 96, 5, 150, 0, 0, 5, 5, 32, 96, 5, 150, 0, 0, 5, 136, 0, 0, 5, 5, 32, 96, 5, 150, 0, 0, 5, 138, 0, 0, 5, 5, 32, 96, 5, 150, 0, 0, 5, 148, 0, 0, 5, 5, 32, 96, 5, 164, 0, 0, 166, 5, 32, 96, 5, 164, 0, 0, 16, 5, 36, 96, 16, 5, 6, 96, 16, 5, 32, 96, 168, 5, 36, 96, 16, 5, 6, 96, 16, 5, 32, 96, 48, 5, 42, 96, 180, 5, 0, 42, 48, 5, 42, 96, 177, 5, 0, 52, 48, 5, 42, 96, 46, 5, 0, 54, 48, 5, 42, 96, 49, 5, 0, 64, 48, 5, 42, 96, 49, 5, 0, 66, 48, 5, 42, 96, 49, 5, 0, 78, 48, 5, 42, 96, 180, 5, 0, 84, 48, 5, 42, 96, 184, 5, 0, 86, 5, 5, 48, 96, 5, 132, 0, 0, 158, 5, 48, 96, 5, 132, 0, 0, 5, 5, 48, 96, 5, 132, 0, 0, 5, 136, 0, 0, 158, 5, 48, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 5, 48, 96, 5, 132, 0, 0, 5, 138, 0, 0, 158, 5, 48, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 5, 48, 96, 5, 134, 0, 0, 158, 5, 48, 96, 5, 134, 0, 0, 5, 5, 48, 96, 5, 134, 0, 0, 5, 136, 0, 0, 158, 5, 48, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 5, 48, 96, 5, 134, 0, 0, 5, 138, 0, 0, 158, 5, 48, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 5, 48, 96, 5, 136, 0, 0, 158, 5, 48, 96, 5, 136, 0, 0, 5, 5, 48, 96, 5, 138, 0, 0, 158, 5, 48, 96, 5, 138, 0, 0, 5, 5, 60, 96, 5, 132, 0, 0, 5, 5, 60, 96, 5, 134, 0, 0, 160, 5, 60, 96, 5, 134, 0, 0, 5, 5, 76, 96, 5, 132, 0, 0, 5, 5, 76, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 5, 76, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 5, 76, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 5, 76, 96, 5, 134, 0, 0, 155, 5, 76, 96, 5, 134, 0, 0, 5, 5, 76, 96, 5, 134, 0, 0, 5, 136, 0, 0, 155, 5, 76, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 5, 76, 96, 5, 134, 0, 0, 5, 138, 0, 0, 155, 5, 76, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 5, 76, 96, 5, 134, 0, 0, 5, 148, 0, 0, 155, 5, 76, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 5, 76, 96, 5, 136, 0, 0, 155, 5, 76, 96, 5, 136, 0, 0, 166, 5, 76, 96, 5, 136, 0, 0, 5, 5, 76, 96, 5, 138, 0, 0, 155, 5, 76, 96, 5, 138, 0, 0, 5, 5, 76, 96, 5, 140, 0, 0, 155, 5, 76, 96, 5, 140, 0, 0, 5, 5, 76, 96, 5, 148, 0, 0, 5, 5, 76, 96, 5, 150, 0, 0, 155, 5, 76, 96, 5, 150, 0, 0, 166, 5, 76, 96, 5, 150, 0, 0, 5, 5, 76, 96, 5, 150, 0, 0, 5, 136, 0, 0, 5, 5, 76, 96, 5, 150, 0, 0, 5, 138, 0, 0, 5, 5, 76, 96, 5, 150, 0, 0, 5, 148, 0, 0, 5, 5, 76, 96, 5, 164, 0, 0, 155, 5, 76, 96, 5, 164, 0, 0, 5, 5, 86, 96, 5, 132, 0, 0, 155, 5, 86, 96, 5, 132, 0, 0, 5, 5, 86, 96, 5, 132, 0, 0, 5, 136, 0, 0, 155, 5, 86, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 5, 86, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 155, 5, 86, 96, 5, 132, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 132, 0, 0, 5, 138, 0, 0, 155, 5, 86, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 5, 86, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 155, 5, 86, 96, 5, 132, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 132, 0, 0, 5, 148, 0, 0, 155, 5, 86, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 5, 86, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 155, 5, 86, 96, 5, 132, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 132, 0, 0, 5, 216, 0, 0, 155, 5, 86, 96, 5, 132, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 134, 0, 0, 155, 5, 86, 96, 5, 134, 0, 0, 5, 5, 86, 96, 5, 134, 0, 0, 5, 136, 0, 0, 155, 5, 86, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 5, 86, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 155, 5, 86, 96, 5, 134, 0, 0, 5, 136, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 134, 0, 0, 5, 138, 0, 0, 155, 5, 86, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 5, 86, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 155, 5, 86, 96, 5, 134, 0, 0, 5, 138, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 134, 0, 0, 5, 148, 0, 0, 155, 5, 86, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 5, 86, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 155, 5, 86, 96, 5, 134, 0, 0, 5, 148, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 134, 0, 0, 5, 216, 0, 0, 155, 5, 86, 96, 5, 134, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 136, 0, 0, 155, 5, 86, 96, 5, 136, 0, 0, 5, 5, 86, 96, 5, 136, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 138, 0, 0, 155, 5, 86, 96, 5, 138, 0, 0, 5, 5, 86, 96, 5, 138, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 148, 0, 0, 5, 5, 86, 96, 5, 148, 0, 0, 5, 216, 0, 0, 5, 5, 86, 96, 5, 216, 0, 0, 155, 5, 86, 96, 5, 216, 0, 0, 32, 5, 120, 96, 32, 5, 98, 96, 32, 5, 118, 96, 5, 5, 6, 97, 5, 140, 0, 0, 174, 5, 6, 97, 5, 140, 0, 0, 5, 5, 6, 97, 5, 150, 0, 0, 174, 5, 6, 97, 5, 150, 0, 0, 5, 5, 8, 97, 5, 150, 0, 0, 160, 5, 8, 97, 5, 150, 0, 0, 5, 5, 16, 97, 5, 136, 0, 0, 174, 5, 16, 97, 5, 136, 0, 0, 5, 5, 40, 97, 5, 138, 0, 0, 174, 5, 40, 97, 5, 138, 0, 0, 5, 5, 40, 97, 5, 140, 0, 0, 174, 5, 40, 97, 5, 140, 0, 0, 5, 5, 40, 97, 5, 150, 0, 0, 174, 5, 40, 97, 5, 150, 0, 0, 5, 5, 44, 97, 5, 140, 0, 0, 174, 5, 44, 97, 5, 140, 0, 0, 5, 5, 44, 97, 5, 150, 0, 0, 174, 5, 44, 97, 5, 150, 0, 0, 5, 5, 50, 97, 5, 150, 0, 0, 174, 5, 50, 97, 5, 150, 0, 0, 5, 5, 70, 97, 5, 138, 0, 0, 174, 5, 70, 97, 5, 138, 0, 0, 5, 5, 70, 97, 5, 150, 0, 0, 174, 5, 70, 97, 5, 150, 0, 0, 5, 5, 70, 97, 5, 164, 0, 0, 174, 5, 70, 97, 5, 164, 0, 0, 5, 5, 74, 97, 5, 150, 0, 0, 32, 5, 74, 97, 32, 150, 0, 0, 174, 5, 74, 97, 5, 150, 0, 0, 5, 5, 84, 97, 5, 136, 0, 0, 174, 5, 84, 97, 5, 136, 0, 0, 5, 5, 138, 97, 5, 150, 0, 0, 158, 5, 138, 97, 5, 150, 0, 0, 5, 5, 140, 97, 5, 150, 0, 0, 160, 5, 140, 97, 5, 150, 0, 0, 32, 5, 156, 97, 32, 5, 162, 97, 5, 5, 172, 97, 5, 140, 0, 0, 174, 5, 172, 97, 5, 140, 0, 0, 5, 5, 172, 97, 5, 150, 0, 0, 174, 5, 172, 97, 5, 150, 0, 0, 5, 5, 172, 97, 5, 152, 0, 0, 174, 5, 172, 97, 5, 152, 0, 0, 5, 5, 172, 97, 5, 164, 0, 0, 174, 5, 172, 97, 5, 164, 0, 0, 5, 5, 200, 97, 5, 150, 0, 0, 174, 5, 200, 97, 5, 150, 0, 0, 5, 5, 230, 97, 5, 150, 0, 0, 174, 5, 230, 97, 5, 150, 0, 0, 5, 5, 238, 97, 5, 150, 0, 0, 160, 5, 238, 97, 5, 150, 0, 0, 193, 86, 246, 97, 5, 184, 0, 0, 32, 5, 16, 100, 32, 5, 76, 100, 32, 5, 46, 100, 32, 5, 16, 100, 32, 5, 46, 100, 32, 5, 28, 100, 32, 5, 46, 100, 32, 5, 32, 100, 32, 5, 46, 100, 32, 5, 52, 100, 32, 5, 68, 100, 32, 5, 52, 100, 5, 5, 6, 101, 194, 5, 12, 227, 5, 5, 6, 101, 194, 5, 45, 227, 5, 5, 6, 101, 194, 5, 210, 227, 32, 5, 6, 101, 32, 5, 28, 101, 5, 5, 8, 101, 194, 5, 210, 227, 5, 5, 8, 101, 194, 5, 243, 227, 5, 5, 10, 101, 194, 5, 210, 227, 5, 5, 12, 101, 194, 5, 210, 227, 5, 5, 14, 101, 194, 5, 210, 227, 5, 5, 16, 101, 194, 5, 78, 227, 5, 5, 16, 101, 194, 5, 210, 227, 32, 5, 16, 101, 32, 5, 16, 101, 32, 5, 16, 101, 32, 5, 24, 101, 5, 5, 18, 101, 194, 5, 210, 227, 5, 5, 22, 101, 194, 5, 210, 227, 5, 5, 24, 101, 194, 5, 167, 226, 5, 5, 24, 101, 194, 5, 210, 227, 32, 5, 24, 101, 32, 5, 14, 101, 32, 5, 16, 101, 32, 5, 14, 101, 32, 5, 24, 101, 32, 5, 24, 101, 32, 5, 24, 101, 32, 5, 24, 101, 194, 5, 12, 227, 5, 5, 26, 101, 194, 5, 210, 227, 46, 5, 26, 101, 194, 5, 210, 227, 5, 5, 26, 101, 194, 5, 243, 227, 5, 5, 28, 101, 194, 5, 210, 227, 5, 5, 30, 101, 194, 5, 210, 227, 5, 5, 32, 101, 194, 5, 210, 227, 5, 5, 34, 101, 194, 5, 210, 227, 5, 5, 38, 101, 194, 5, 210, 227, 32, 5, 38, 101, 194, 5, 210, 227, 5, 5, 38, 101, 194, 5, 243, 227, 5, 5, 40, 101, 194, 5, 210, 227, 5, 5, 42, 101, 194, 5, 210, 227, 5, 5, 44, 101, 194, 5, 210, 227, 5, 5, 46, 101, 194, 5, 144, 227, 5, 5, 46, 101, 194, 5, 177, 227, 5, 5, 46, 101, 194, 5, 210, 227, 5, 5, 46, 101, 194, 5, 210, 227, 194, 5, 144, 227, 5, 5, 46, 101, 194, 5, 210, 227, 194, 5, 177, 227, 5, 5, 48, 101, 194, 5, 210, 227, 38, 5, 22, 102, 38, 5, 26, 102, 49, 5, 22, 102, 49, 5, 26, 102, 16, 5, 22, 102, 24, 5, 58, 102, 49, 5, 22, 102, 48, 5, 58, 102, 16, 5, 22, 102, 24, 5, 72, 102, 49, 5, 22, 102, 48, 5, 72, 102, 16, 5, 22, 102, 24, 5, 74, 102, 38, 5, 22, 102, 30, 5, 110, 102, 38, 5, 22, 102, 30, 5, 112, 102, 16, 5, 22, 102, 24, 5, 200, 102, 27, 5, 22, 102, 32, 5, 200, 102, 38, 5, 22, 102, 40, 5, 200, 102, 49, 5, 22, 102, 48, 5, 200, 102, 38, 5, 22, 102, 40, 5, 204, 102, 16, 5, 22, 102, 24, 5, 214, 102, 27, 5, 22, 102, 32, 5, 214, 102, 38, 5, 22, 102, 38, 5, 222, 102, 49, 5, 22, 102, 49, 5, 222, 102, 38, 5, 22, 102, 30, 5, 228, 102, 49, 5, 22, 102, 44, 5, 228, 102, 38, 5, 22, 102, 32, 5, 230, 102, 49, 5, 22, 102, 46, 5, 230, 102, 16, 5, 22, 102, 16, 5, 236, 102, 38, 5, 22, 102, 38, 5, 236, 102, 49, 5, 22, 102, 49, 5, 236, 102, 38, 5, 22, 102, 44, 5, 238, 102, 49, 5, 22, 102, 51, 5, 238, 102, 16, 5, 22, 102, 16, 5, 246, 102, 38, 5, 22, 102, 38, 5, 246, 102, 49, 5, 22, 102, 49, 5, 246, 102, 38, 5, 26, 102, 194, 30, 164, 229, 49, 5, 26, 102, 194, 44, 164, 229, 16, 5, 26, 102, 32, 5, 6, 102, 49, 5, 26, 102, 48, 5, 182, 102, 48, 5, 30, 102, 44, 5, 110, 102, 49, 5, 26, 102, 48, 5, 196, 102, 48, 5, 196, 102, 48, 5, 214, 102, 24, 5, 30, 102, 24, 5, 58, 102, 48, 5, 30, 102, 48, 5, 58, 102, 24, 5, 30, 102, 24, 5, 72, 102, 48, 5, 30, 102, 48, 5, 72, 102, 40, 5, 30, 102, 40, 5, 72, 102, 44, 5, 238, 102, 24, 5, 30, 102, 24, 5, 74, 102, 48, 5, 30, 102, 48, 5, 74, 102, 40, 5, 30, 102, 40, 5, 74, 102, 44, 5, 238, 102, 40, 5, 30, 102, 30, 5, 110, 102, 40, 5, 30, 102, 30, 5, 112, 102, 24, 5, 30, 102, 24, 5, 200, 102, 32, 5, 30, 102, 32, 5, 200, 102, 40, 5, 30, 102, 40, 5, 200, 102, 48, 5, 30, 102, 48, 5, 200, 102, 40, 5, 30, 102, 40, 5, 204, 102, 24, 5, 30, 102, 24, 5, 214, 102, 32, 5, 30, 102, 32, 5, 214, 102, 40, 5, 30, 102, 38, 5, 236, 102, 48, 5, 30, 102, 49, 5, 236, 102, 40, 5, 30, 102, 44, 5, 238, 102, 48, 5, 30, 102, 51, 5, 238, 102, 24, 5, 42, 102, 24, 5, 58, 102, 48, 5, 42, 102, 48, 5, 58, 102, 24, 5, 42, 102, 24, 5, 58, 102, 24, 5, 200, 102, 40, 5, 42, 102, 40, 5, 58, 102, 38, 5, 236, 102, 40, 5, 42, 102, 40, 5, 58, 102, 44, 5, 238, 102, 24, 5, 42, 102, 24, 5, 72, 102, 48, 5, 42, 102, 48, 5, 72, 102, 24, 5, 42, 102, 24, 5, 72, 102, 24, 5, 58, 102, 40, 5, 42, 102, 40, 5, 72, 102, 40, 5, 58, 102, 24, 5, 42, 102, 24, 5, 72, 102, 24, 5, 200, 102, 24, 5, 42, 102, 24, 5, 74, 102, 48, 5, 42, 102, 48, 5, 74, 102, 24, 5, 42, 102, 24, 5, 74, 102, 24, 5, 200, 102, 40, 5, 42, 102, 40, 5, 74, 102, 38, 5, 236, 102, 40, 5, 42, 102, 40, 5, 74, 102, 44, 5, 238, 102, 40, 5, 42, 102, 30, 5, 110, 102, 40, 5, 42, 102, 30, 5, 112, 102, 24, 5, 42, 102, 24, 5, 200, 102, 32, 5, 42, 102, 32, 5, 200, 102, 40, 5, 42, 102, 40, 5, 200, 102, 48, 5, 42, 102, 48, 5, 200, 102, 24, 5, 42, 102, 24, 5, 200, 102, 24, 5, 58, 102, 24, 5, 42, 102, 24, 5, 200, 102, 24, 5, 72, 102, 24, 5, 42, 102, 24, 5, 200, 102, 24, 5, 74, 102, 40, 5, 42, 102, 40, 5, 200, 102, 38, 5, 236, 102, 40, 5, 42, 102, 40, 5, 200, 102, 44, 5, 238, 102, 40, 5, 42, 102, 40, 5, 204, 102, 24, 5, 42, 102, 24, 5, 214, 102, 32, 5, 42, 102, 32, 5, 214, 102, 40, 5, 42, 102, 38, 5, 236, 102, 48, 5, 42, 102, 49, 5, 236, 102, 40, 5, 42, 102, 44, 5, 238, 102, 48, 5, 42, 102, 51, 5, 238, 102, 48, 5, 44, 102, 48, 5, 58, 102, 40, 5, 44, 102, 30, 5, 110, 102, 40, 5, 44, 102, 30, 5, 112, 102, 24, 5, 44, 102, 24, 5, 200, 102, 32, 5, 44, 102, 32, 5, 200, 102, 40, 5, 44, 102, 40, 5, 200, 102, 48, 5, 44, 102, 48, 5, 200, 102, 40, 5, 44, 102, 40, 5, 204, 102, 32, 5, 44, 102, 32, 5, 214, 102, 40, 5, 44, 102, 38, 5, 236, 102, 48, 5, 44, 102, 49, 5, 236, 102, 40, 5, 44, 102, 44, 5, 238, 102, 48, 5, 44, 102, 51, 5, 238, 102, 24, 5, 58, 102, 24, 5, 72, 102, 48, 5, 58, 102, 48, 5, 72, 102, 40, 5, 58, 102, 40, 5, 72, 102, 38, 5, 236, 102, 40, 5, 58, 102, 40, 5, 72, 102, 44, 5, 238, 102, 48, 5, 58, 102, 48, 5, 196, 102, 32, 5, 0, 4, 48, 5, 58, 102, 48, 5, 196, 102, 49, 5, 26, 102, 48, 5, 196, 102, 48, 5, 214, 102, 48, 5, 58, 102, 48, 5, 200, 102, 24, 5, 58, 102, 24, 5, 200, 102, 24, 5, 72, 102, 40, 5, 58, 102, 40, 5, 200, 102, 40, 5, 72, 102, 40, 5, 58, 102, 40, 5, 200, 102, 38, 5, 236, 102, 40, 5, 58, 102, 40, 5, 200, 102, 44, 5, 238, 102, 48, 5, 58, 102, 49, 5, 236, 102, 48, 5, 58, 102, 51, 5, 238, 102, 48, 5, 72, 102, 48, 5, 58, 102, 40, 5, 72, 102, 40, 5, 58, 102, 44, 5, 238, 102, 48, 5, 72, 102, 48, 5, 200, 102, 40, 5, 72, 102, 40, 5, 200, 102, 38, 5, 236, 102, 40, 5, 72, 102, 40, 5, 200, 102, 44, 5, 238, 102, 48, 5, 72, 102, 49, 5, 236, 102, 48, 5, 72, 102, 51, 5, 238, 102, 24, 5, 74, 102, 24, 5, 58, 102, 48, 5, 74, 102, 48, 5, 58, 102, 48, 5, 74, 102, 48, 5, 72, 102, 48, 5, 74, 102, 48, 5, 200, 102, 48, 5, 74, 102, 49, 5, 236, 102, 48, 5, 74, 102, 51, 5, 238, 102, 44, 5, 86, 102, 194, 44, 59, 235, 44, 5, 110, 102, 194, 44, 59, 235, 44, 5, 110, 102, 48, 5, 134, 102, 49, 5, 222, 102, 48, 5, 196, 102, 24, 5, 134, 102, 24, 5, 58, 102, 32, 5, 134, 102, 32, 5, 58, 102, 48, 5, 134, 102, 48, 5, 58, 102, 24, 5, 134, 102, 24, 5, 58, 102, 24, 5, 72, 102, 40, 5, 134, 102, 40, 5, 58, 102, 38, 5, 236, 102, 24, 5, 134, 102, 24, 5, 72, 102, 32, 5, 134, 102, 32, 5, 72, 102, 48, 5, 134, 102, 48, 5, 72, 102, 24, 5, 134, 102, 24, 5, 72, 102, 24, 5, 58, 102, 24, 5, 134, 102, 24, 5, 74, 102, 32, 5, 134, 102, 32, 5, 74, 102, 48, 5, 134, 102, 48, 5, 74, 102, 40, 5, 134, 102, 40, 5, 74, 102, 38, 5, 236, 102, 40, 5, 134, 102, 40, 5, 74, 102, 44, 5, 238, 102, 40, 5, 134, 102, 30, 5, 110, 102, 48, 5, 134, 102, 44, 5, 110, 102, 24, 5, 134, 102, 24, 5, 200, 102, 32, 5, 134, 102, 32, 5, 200, 102, 48, 5, 134, 102, 48, 5, 200, 102, 24, 5, 134, 102, 24, 5, 200, 102, 24, 5, 58, 102, 24, 5, 134, 102, 24, 5, 200, 102, 24, 5, 72, 102, 40, 5, 134, 102, 40, 5, 200, 102, 40, 5, 72, 102, 24, 5, 134, 102, 24, 5, 200, 102, 24, 5, 200, 102, 40, 5, 134, 102, 40, 5, 200, 102, 40, 5, 200, 102, 24, 5, 134, 102, 24, 5, 214, 102, 32, 5, 134, 102, 32, 5, 214, 102, 40, 5, 134, 102, 38, 5, 236, 102, 48, 5, 134, 102, 49, 5, 236, 102, 40, 5, 134, 102, 44, 5, 238, 102, 48, 5, 134, 102, 51, 5, 238, 102, 24, 5, 136, 102, 24, 5, 58, 102, 32, 5, 136, 102, 32, 5, 58, 102, 40, 5, 136, 102, 40, 5, 58, 102, 48, 5, 136, 102, 48, 5, 58, 102, 40, 5, 136, 102, 40, 5, 58, 102, 44, 5, 238, 102, 24, 5, 136, 102, 24, 5, 72, 102, 32, 5, 136, 102, 32, 5, 72, 102, 40, 5, 136, 102, 40, 5, 72, 102, 48, 5, 136, 102, 48, 5, 72, 102, 24, 5, 136, 102, 24, 5, 72, 102, 24, 5, 200, 102, 40, 5, 136, 102, 40, 5, 72, 102, 40, 5, 200, 102, 40, 5, 136, 102, 40, 5, 72, 102, 44, 5, 238, 102, 24, 5, 136, 102, 24, 5, 74, 102, 32, 5, 136, 102, 32, 5, 74, 102, 40, 5, 136, 102, 40, 5, 74, 102, 48, 5, 136, 102, 48, 5, 74, 102, 40, 5, 136, 102, 30, 5, 110, 102, 48, 5, 136, 102, 44, 5, 110, 102, 24, 5, 136, 102, 24, 5, 200, 102, 32, 5, 136, 102, 32, 5, 200, 102, 40, 5, 136, 102, 40, 5, 200, 102, 48, 5, 136, 102, 48, 5, 200, 102, 24, 5, 136, 102, 24, 5, 200, 102, 24, 5, 74, 102, 40, 5, 136, 102, 40, 5, 200, 102, 40, 5, 74, 102, 24, 5, 136, 102, 24, 5, 200, 102, 24, 5, 200, 102, 40, 5, 136, 102, 40, 5, 200, 102, 40, 5, 200, 102, 24, 5, 136, 102, 24, 5, 214, 102, 32, 5, 136, 102, 32, 5, 214, 102, 40, 5, 136, 102, 38, 5, 236, 102, 48, 5, 136, 102, 49, 5, 236, 102, 40, 5, 136, 102, 44, 5, 238, 102, 48, 5, 136, 102, 51, 5, 238, 102, 24, 5, 146, 102, 24, 5, 72, 102, 48, 5, 146, 102, 48, 5, 72, 102, 24, 5, 146, 102, 24, 5, 72, 102, 24, 5, 72, 102, 40, 5, 146, 102, 40, 5, 72, 102, 40, 5, 72, 102, 40, 5, 146, 102, 40, 5, 72, 102, 44, 5, 238, 102, 24, 5, 146, 102, 24, 5, 74, 102, 40, 5, 146, 102, 30, 5, 110, 102, 48, 5, 146, 102, 44, 5, 110, 102, 48, 5, 146, 102, 48, 5, 196, 102, 48, 5, 166, 102, 48, 5, 200, 102, 48, 5, 146, 102, 48, 5, 196, 102, 49, 5, 236, 102, 48, 5, 146, 102, 48, 5, 196, 102, 49, 5, 236, 102, 32, 5, 0, 4, 49, 5, 26, 102, 48, 5, 196, 102, 48, 5, 196, 102, 48, 5, 214, 102, 32, 5, 0, 4, 48, 5, 166, 102, 48, 5, 196, 102, 51, 5, 238, 102, 48, 5, 214, 102, 32, 5, 0, 4, 49, 5, 222, 102, 48, 5, 134, 102, 48, 5, 196, 102, 48, 5, 200, 102, 48, 5, 146, 102, 48, 5, 196, 102, 46, 5, 250, 102, 24, 5, 146, 102, 24, 5, 200, 102, 48, 5, 146, 102, 48, 5, 200, 102, 24, 5, 146, 102, 24, 5, 200, 102, 24, 5, 200, 102, 40, 5, 146, 102, 40, 5, 200, 102, 40, 5, 200, 102, 40, 5, 146, 102, 38, 5, 236, 102, 48, 5, 146, 102, 49, 5, 236, 102, 40, 5, 146, 102, 44, 5, 238, 102, 48, 5, 146, 102, 51, 5, 238, 102, 24, 5, 148, 102, 24, 5, 58, 102, 48, 5, 148, 102, 48, 5, 58, 102, 24, 5, 148, 102, 24, 5, 72, 102, 48, 5, 148, 102, 48, 5, 72, 102, 40, 5, 148, 102, 40, 5, 72, 102, 38, 5, 236, 102, 40, 5, 148, 102, 40, 5, 72, 102, 44, 5, 238, 102, 24, 5, 148, 102, 24, 5, 74, 102, 48, 5, 148, 102, 48, 5, 74, 102, 24, 5, 148, 102, 24, 5, 74, 102, 24, 5, 200, 102, 40, 5, 148, 102, 40, 5, 74, 102, 40, 5, 200, 102, 40, 5, 148, 102, 30, 5, 110, 102, 48, 5, 148, 102, 44, 5, 110, 102, 24, 5, 148, 102, 24, 5, 200, 102, 48, 5, 148, 102, 48, 5, 200, 102, 40, 5, 148, 102, 38, 5, 236, 102, 48, 5, 148, 102, 49, 5, 236, 102, 40, 5, 148, 102, 44, 5, 238, 102, 48, 5, 148, 102, 51, 5, 238, 102, 24, 5, 158, 102, 24, 5, 72, 102, 48, 5, 158, 102, 48, 5, 72, 102, 24, 5, 158, 102, 24, 5, 200, 102, 32, 5, 158, 102, 32, 5, 200, 102, 48, 5, 158, 102, 48, 5, 200, 102, 24, 5, 158, 102, 24, 5, 200, 102, 24, 5, 72, 102, 40, 5, 158, 102, 40, 5, 200, 102, 40, 5, 72, 102, 24, 5, 158, 102, 24, 5, 200, 102, 24, 5, 200, 102, 40, 5, 158, 102, 40, 5, 200, 102, 44, 5, 238, 102, 40, 5, 158, 102, 38, 5, 236, 102, 48, 5, 158, 102, 49, 5, 236, 102, 40, 5, 158, 102, 44, 5, 238, 102, 48, 5, 158, 102, 51, 5, 238, 102, 24, 5, 160, 102, 24, 5, 200, 102, 32, 5, 160, 102, 32, 5, 200, 102, 48, 5, 160, 102, 48, 5, 200, 102, 24, 5, 166, 102, 24, 5, 58, 102, 48, 5, 166, 102, 48, 5, 58, 102, 24, 5, 166, 102, 24, 5, 58, 102, 24, 5, 200, 102, 40, 5, 166, 102, 40, 5, 58, 102, 40, 5, 200, 102, 24, 5, 166, 102, 24, 5, 200, 102, 24, 5, 166, 102, 24, 5, 200, 102, 24, 5, 200, 102, 40, 5, 166, 102, 40, 5, 200, 102, 40, 5, 200, 102, 40, 5, 166, 102, 40, 5, 200, 102, 38, 5, 236, 102, 40, 5, 166, 102, 40, 5, 200, 102, 44, 5, 238, 102, 40, 5, 166, 102, 38, 5, 236, 102, 48, 5, 166, 102, 49, 5, 236, 102, 40, 5, 166, 102, 44, 5, 238, 102, 48, 5, 166, 102, 51, 5, 238, 102, 24, 5, 168, 102, 24, 5, 58, 102, 48, 5, 168, 102, 48, 5, 58, 102, 24, 5, 168, 102, 24, 5, 200, 102, 48, 5, 168, 102, 48, 5, 200, 102, 40, 5, 168, 102, 40, 5, 200, 102, 40, 5, 200, 102, 40, 5, 168, 102, 40, 5, 200, 102, 38, 5, 236, 102, 40, 5, 168, 102, 40, 5, 200, 102, 44, 5, 238, 102, 40, 5, 168, 102, 38, 5, 236, 102, 48, 5, 168, 102, 49, 5, 236, 102, 40, 5, 168, 102, 44, 5, 238, 102, 48, 5, 168, 102, 51, 5, 238, 102, 24, 5, 174, 102, 24, 5, 58, 102, 48, 5, 174, 102, 48, 5, 58, 102, 24, 5, 174, 102, 24, 5, 72, 102, 48, 5, 174, 102, 48, 5, 72, 102, 24, 5, 174, 102, 24, 5, 74, 102, 48, 5, 174, 102, 48, 5, 74, 102, 24, 5, 174, 102, 24, 5, 74, 102, 24, 5, 200, 102, 40, 5, 174, 102, 40, 5, 74, 102, 40, 5, 200, 102, 24, 5, 174, 102, 24, 5, 200, 102, 48, 5, 174, 102, 48, 5, 200, 102, 40, 5, 174, 102, 40, 5, 200, 102, 44, 5, 238, 102, 40, 5, 174, 102, 38, 5, 236, 102, 48, 5, 174, 102, 49, 5, 236, 102, 40, 5, 174, 102, 44, 5, 238, 102, 48, 5, 174, 102, 51, 5, 238, 102, 24, 5, 178, 102, 24, 5, 72, 102, 48, 5, 178, 102, 48, 5, 72, 102, 48, 5, 178, 102, 48, 5, 196, 102, 46, 5, 250, 102, 24, 5, 178, 102, 24, 5, 200, 102, 48, 5, 178, 102, 48, 5, 200, 102, 24, 5, 178, 102, 24, 5, 200, 102, 24, 5, 72, 102, 40, 5, 178, 102, 40, 5, 200, 102, 40, 5, 72, 102, 40, 5, 178, 102, 40, 5, 200, 102, 40, 5, 200, 102, 40, 5, 178, 102, 40, 5, 200, 102, 44, 5, 238, 102, 40, 5, 178, 102, 38, 5, 236, 102, 48, 5, 178, 102, 49, 5, 236, 102, 40, 5, 178, 102, 44, 5, 238, 102, 48, 5, 178, 102, 51, 5, 238, 102, 40, 5, 182, 102, 38, 5, 26, 102, 48, 5, 182, 102, 49, 5, 26, 102, 24, 5, 182, 102, 24, 5, 58, 102, 48, 5, 182, 102, 48, 5, 58, 102, 24, 5, 182, 102, 24, 5, 72, 102, 48, 5, 182, 102, 48, 5, 72, 102, 24, 5, 182, 102, 24, 5, 74, 102, 48, 5, 182, 102, 48, 5, 74, 102, 24, 5, 182, 102, 24, 5, 196, 102, 32, 5, 182, 102, 32, 5, 196, 102, 40, 5, 182, 102, 40, 5, 196, 102, 48, 5, 182, 102, 48, 5, 196, 102, 24, 5, 182, 102, 24, 5, 200, 102, 32, 5, 182, 102, 32, 5, 200, 102, 40, 5, 182, 102, 40, 5, 200, 102, 48, 5, 182, 102, 48, 5, 200, 102, 24, 5, 182, 102, 24, 5, 200, 102, 24, 5, 200, 102, 40, 5, 182, 102, 40, 5, 200, 102, 40, 5, 200, 102, 40, 5, 182, 102, 40, 5, 200, 102, 44, 5, 238, 102, 40, 5, 182, 102, 38, 5, 236, 102, 48, 5, 182, 102, 49, 5, 236, 102, 40, 5, 182, 102, 44, 5, 238, 102, 48, 5, 182, 102, 51, 5, 238, 102, 40, 5, 196, 102, 32, 5, 8, 102, 48, 5, 196, 102, 46, 5, 8, 102, 40, 5, 196, 102, 32, 5, 10, 102, 48, 5, 196, 102, 46, 5, 10, 102, 40, 5, 196, 102, 32, 5, 18, 102, 48, 5, 196, 102, 46, 5, 18, 102, 40, 5, 196, 102, 38, 5, 26, 102, 24, 5, 196, 102, 24, 5, 58, 102, 48, 5, 196, 102, 48, 5, 58, 102, 24, 5, 196, 102, 24, 5, 58, 102, 24, 5, 58, 102, 40, 5, 196, 102, 40, 5, 58, 102, 40, 5, 58, 102, 24, 5, 196, 102, 24, 5, 58, 102, 24, 5, 200, 102, 40, 5, 196, 102, 40, 5, 58, 102, 40, 5, 200, 102, 40, 5, 196, 102, 40, 5, 58, 102, 44, 5, 238, 102, 24, 5, 196, 102, 24, 5, 72, 102, 48, 5, 196, 102, 48, 5, 72, 102, 24, 5, 196, 102, 24, 5, 72, 102, 24, 5, 200, 102, 40, 5, 196, 102, 40, 5, 72, 102, 40, 5, 200, 102, 40, 5, 196, 102, 40, 5, 72, 102, 38, 5, 236, 102, 40, 5, 196, 102, 40, 5, 72, 102, 44, 5, 238, 102, 24, 5, 196, 102, 24, 5, 74, 102, 48, 5, 196, 102, 48, 5, 74, 102, 24, 5, 196, 102, 24, 5, 74, 102, 24, 5, 200, 102, 40, 5, 196, 102, 40, 5, 74, 102, 40, 5, 200, 102, 24, 5, 196, 102, 24, 5, 200, 102, 32, 5, 196, 102, 32, 5, 200, 102, 40, 5, 196, 102, 40, 5, 200, 102, 24, 5, 196, 102, 24, 5, 200, 102, 24, 5, 72, 102, 40, 5, 196, 102, 40, 5, 200, 102, 40, 5, 72, 102, 40, 5, 196, 102, 40, 5, 200, 102, 44, 5, 238, 102, 24, 5, 196, 102, 24, 5, 214, 102, 40, 5, 196, 102, 38, 5, 236, 102, 40, 5, 196, 102, 44, 5, 238, 102, 40, 5, 200, 102, 38, 5, 26, 102, 48, 5, 200, 102, 48, 5, 58, 102, 24, 5, 200, 102, 24, 5, 58, 102, 24, 5, 72, 102, 24, 5, 200, 102, 24, 5, 58, 102, 24, 5, 74, 102, 24, 5, 200, 102, 24, 5, 58, 102, 24, 5, 200, 102, 40, 5, 200, 102, 40, 5, 58, 102, 44, 5, 238, 102, 48, 5, 200, 102, 48, 5, 72, 102, 24, 5, 200, 102, 24, 5, 72, 102, 24, 5, 58, 102, 24, 5, 200, 102, 24, 5, 72, 102, 24, 5, 200, 102, 48, 5, 200, 102, 48, 5, 72, 102, 48, 5, 200, 102, 44, 5, 84, 102, 40, 5, 200, 102, 40, 5, 72, 102, 44, 5, 238, 102, 24, 5, 200, 102, 24, 5, 74, 102, 24, 5, 58, 102, 24, 5, 200, 102, 24, 5, 74, 102, 24, 5, 200, 102, 40, 5, 200, 102, 40, 5, 74, 102, 44, 5, 238, 102, 48, 5, 200, 102, 48, 5, 200, 102, 40, 5, 200, 102, 40, 5, 200, 102, 44, 5, 238, 102, 48, 5, 200, 102, 49, 5, 236, 102, 48, 5, 200, 102, 51, 5, 238, 102, 24, 5, 204, 102, 24, 5, 58, 102, 48, 5, 204, 102, 48, 5, 58, 102, 24, 5, 204, 102, 24, 5, 58, 102, 24, 5, 72, 102, 40, 5, 204, 102, 40, 5, 58, 102, 40, 5, 72, 102, 24, 5, 204, 102, 24, 5, 58, 102, 24, 5, 200, 102, 40, 5, 204, 102, 40, 5, 58, 102, 40, 5, 200, 102, 40, 5, 204, 102, 40, 5, 58, 102, 38, 5, 236, 102, 40, 5, 204, 102, 40, 5, 58, 102, 44, 5, 238, 102, 24, 5, 204, 102, 24, 5, 72, 102, 48, 5, 204, 102, 48, 5, 72, 102, 24, 5, 204, 102, 24, 5, 72, 102, 24, 5, 200, 102, 40, 5, 204, 102, 40, 5, 72, 102, 38, 5, 236, 102, 40, 5, 204, 102, 40, 5, 72, 102, 44, 5, 238, 102, 24, 5, 204, 102, 24, 5, 74, 102, 48, 5, 204, 102, 48, 5, 74, 102, 40, 5, 204, 102, 30, 5, 110, 102, 40, 5, 204, 102, 30, 5, 112, 102, 24, 5, 204, 102, 24, 5, 200, 102, 32, 5, 204, 102, 32, 5, 200, 102, 40, 5, 204, 102, 40, 5, 200, 102, 48, 5, 204, 102, 48, 5, 200, 102, 40, 5, 204, 102, 40, 5, 200, 102, 38, 5, 236, 102, 40, 5, 204, 102, 40, 5, 200, 102, 44, 5, 238, 102, 40, 5, 204, 102, 40, 5, 204, 102, 24, 5, 204, 102, 24, 5, 214, 102, 32, 5, 204, 102, 32, 5, 214, 102, 40, 5, 204, 102, 38, 5, 236, 102, 48, 5, 204, 102, 49, 5, 236, 102, 40, 5, 204, 102, 44, 5, 238, 102, 48, 5, 204, 102, 51, 5, 238, 102, 24, 5, 214, 102, 194, 16, 59, 235, 24, 5, 214, 102, 24, 5, 58, 102, 24, 5, 214, 102, 24, 5, 200, 102, 48, 5, 214, 102, 48, 5, 200, 102, 24, 5, 214, 102, 24, 5, 200, 102, 24, 5, 58, 102, 24, 5, 214, 102, 24, 5, 200, 102, 24, 5, 200, 102, 48, 5, 214, 102, 49, 5, 236, 102, 48, 5, 214, 102, 51, 5, 238, 102, 5, 5, 218, 102, 194, 5, 128, 232, 193, 16, 220, 102, 194, 5, 128, 232, 16, 5, 222, 102, 32, 5, 6, 102, 16, 5, 228, 102, 32, 5, 6, 102, 44, 5, 228, 102, 46, 5, 6, 102, 38, 5, 236, 102, 194, 30, 59, 235, 49, 5, 236, 102, 194, 44, 59, 235, 16, 5, 238, 102, 32, 5, 6, 102, 30, 5, 238, 102, 24, 5, 58, 102, 51, 5, 238, 102, 48, 5, 58, 102, 44, 5, 238, 102, 40, 5, 58, 102, 44, 5, 238, 102, 30, 5, 238, 102, 24, 5, 72, 102, 44, 5, 238, 102, 40, 5, 72, 102, 44, 5, 238, 102, 30, 5, 238, 102, 24, 5, 74, 102, 51, 5, 238, 102, 48, 5, 74, 102, 44, 5, 238, 102, 30, 5, 110, 102, 44, 5, 238, 102, 30, 5, 112, 102, 30, 5, 238, 102, 24, 5, 200, 102, 37, 5, 238, 102, 32, 5, 200, 102, 51, 5, 238, 102, 48, 5, 200, 102, 30, 5, 238, 102, 24, 5, 200, 102, 24, 5, 200, 102, 44, 5, 238, 102, 40, 5, 200, 102, 40, 5, 200, 102, 44, 5, 238, 102, 40, 5, 200, 102, 44, 5, 238, 102, 30, 5, 238, 102, 24, 5, 214, 102, 37, 5, 238, 102, 32, 5, 214, 102, 44, 5, 238, 102, 38, 5, 236, 102, 51, 5, 238, 102, 49, 5, 236, 102, 44, 5, 238, 102, 44, 5, 238, 102, 51, 5, 238, 102, 51, 5, 238, 102, 5, 5, 250, 102, 194, 5, 128, 232, 32, 5, 250, 102, 194, 32, 128, 232, 46, 5, 250, 102, 194, 46, 128, 232, 5, 5, 62, 104, 194, 5, 175, 240, 5, 5, 64, 104, 194, 5, 175, 240, 5, 5, 66, 104, 194, 5, 175, 240, 5, 5, 78, 104, 194, 5, 175, 240, 5, 5, 94, 104, 194, 5, 175, 240, 5, 5, 98, 104, 194, 5, 175, 240, 5, 5, 110, 104, 194, 5, 175, 240, 5, 5, 114, 104, 194, 5, 175, 240, 5, 5, 124, 104, 194, 5, 175, 240, 5, 5, 128, 104, 194, 5, 175, 240, 5, 5, 132, 104, 194, 5, 175, 240, 5, 5, 60, 105, 194, 5, 175, 240, 5, 5, 62, 105, 194, 5, 175, 240, 32, 5, 66, 105, 32, 5, 132, 105, 5, 5, 86, 105, 194, 5, 175, 240, 5, 5, 32, 106, 194, 5, 175, 240, 5, 5, 40, 106, 194, 5, 175, 240, 5, 5, 42, 106, 194, 5, 175, 240, 5, 5, 52, 106, 194, 5, 175, 240, 5, 5, 80, 106, 194, 5, 175, 240, 5, 5, 94, 106, 194, 5, 175, 240, 5, 5, 58, 108, 194, 5, 175, 240, 5, 5, 60, 108, 194, 5, 175, 240, 32, 5, 40, 112, 32, 5, 152, 112, 32, 5, 68, 112, 32, 5, 152, 112, 32, 5, 78, 112, 32, 5, 152, 112, 32, 5, 90, 112, 32, 5, 152, 112, 32, 5, 92, 112, 32, 5, 152, 112, 32, 5, 94, 112, 32, 5, 152, 112, 32, 5, 96, 112, 32, 5, 152, 112, 32, 5, 108, 112, 32, 5, 152, 112, 32, 5, 110, 112, 32, 5, 152, 112, 193, 163, 166, 113, 194, 5, 175, 240, 193, 170, 166, 113, 194, 5, 175, 240, 193, 233, 166, 113, 194, 5, 175, 240, 5, 5, 6, 114, 5, 5, 122, 114, 5, 5, 6, 114, 5, 5, 124, 114, 5, 5, 6, 114, 5, 5, 126, 114, 5, 5, 6, 114, 5, 5, 128, 114, 5, 5, 6, 114, 5, 5, 130, 114, 5, 5, 8, 114, 5, 5, 122, 114, 5, 5, 8, 114, 5, 5, 124, 114, 5, 5, 8, 114, 5, 5, 126, 114, 5, 5, 8, 114, 5, 5, 128, 114, 5, 5, 8, 114, 5, 5, 130, 114, 5, 5, 10, 114, 5, 5, 122, 114, 5, 5, 10, 114, 5, 5, 124, 114, 5, 5, 10, 114, 5, 5, 126, 114, 5, 5, 10, 114, 5, 5, 128, 114, 5, 5, 10, 114, 5, 5, 130, 114, 5, 5, 12, 114, 5, 5, 122, 114, 5, 5, 12, 114, 5, 5, 124, 114, 5, 5, 12, 114, 5, 5, 126, 114, 5, 5, 12, 114, 5, 5, 128, 114, 5, 5, 12, 114, 5, 5, 130, 114, 5, 5, 14, 114, 5, 5, 122, 114, 5, 5, 14, 114, 5, 5, 124, 114, 5, 5, 14, 114, 5, 5, 126, 114, 5, 5, 14, 114, 5, 5, 128, 114, 5, 5, 14, 114, 5, 5, 130, 114, 5, 5, 16, 114, 5, 5, 122, 114, 5, 5, 16, 114, 5, 5, 124, 114, 5, 5, 16, 114, 5, 5, 126, 114, 5, 5, 16, 114, 5, 5, 128, 114, 5, 5, 16, 114, 5, 5, 130, 114, 5, 5, 18, 114, 5, 5, 122, 114, 5, 5, 18, 114, 5, 5, 124, 114, 5, 5, 18, 114, 5, 5, 126, 114, 5, 5, 18, 114, 5, 5, 128, 114, 5, 5, 18, 114, 5, 5, 130, 114, 5, 5, 20, 114, 5, 5, 122, 114, 5, 5, 20, 114, 5, 5, 124, 114, 5, 5, 20, 114, 5, 5, 126, 114, 5, 5, 20, 114, 5, 5, 128, 114, 5, 5, 20, 114, 5, 5, 130, 114, 5, 5, 22, 114, 5, 5, 122, 114, 5, 5, 22, 114, 5, 5, 124, 114, 5, 5, 22, 114, 5, 5, 126, 114, 5, 5, 22, 114, 5, 5, 128, 114, 5, 5, 22, 114, 5, 5, 130, 114, 5, 5, 24, 114, 5, 5, 122, 114, 5, 5, 24, 114, 5, 5, 124, 114, 5, 5, 24, 114, 5, 5, 126, 114, 5, 5, 24, 114, 5, 5, 128, 114, 5, 5, 24, 114, 5, 5, 130, 114, 5, 5, 26, 114, 5, 5, 122, 114, 5, 5, 26, 114, 5, 5, 124, 114, 5, 5, 26, 114, 5, 5, 126, 114, 5, 5, 26, 114, 5, 5, 128, 114, 5, 5, 26, 114, 5, 5, 130, 114, 5, 5, 28, 114, 5, 5, 122, 114, 5, 5, 28, 114, 5, 5, 124, 114, 5, 5, 28, 114, 5, 5, 126, 114, 5, 5, 28, 114, 5, 5, 128, 114, 5, 5, 28, 114, 5, 5, 130, 114, 5, 5, 30, 114, 5, 5, 122, 114, 5, 5, 30, 114, 5, 5, 124, 114, 5, 5, 30, 114, 5, 5, 126, 114, 5, 5, 30, 114, 5, 5, 128, 114, 5, 5, 30, 114, 5, 5, 130, 114, 5, 5, 32, 114, 5, 5, 122, 114, 5, 5, 32, 114, 5, 5, 124, 114, 5, 5, 32, 114, 5, 5, 126, 114, 5, 5, 32, 114, 5, 5, 128, 114, 5, 5, 32, 114, 5, 5, 130, 114, 5, 5, 34, 114, 5, 5, 122, 114, 5, 5, 34, 114, 5, 5, 124, 114, 5, 5, 34, 114, 5, 5, 126, 114, 5, 5, 34, 114, 5, 5, 128, 114, 5, 5, 34, 114, 5, 5, 130, 114, 5, 5, 36, 114, 5, 5, 122, 114, 5, 5, 36, 114, 5, 5, 124, 114, 5, 5, 36, 114, 5, 5, 126, 114, 5, 5, 36, 114, 5, 5, 128, 114, 5, 5, 36, 114, 5, 5, 130, 114, 5, 5, 38, 114, 5, 5, 122, 114, 5, 5, 38, 114, 5, 5, 124, 114, 5, 5, 38, 114, 5, 5, 126, 114, 5, 5, 38, 114, 5, 5, 128, 114, 5, 5, 38, 114, 5, 5, 130, 114, 5, 5, 40, 114, 5, 5, 122, 114, 5, 5, 40, 114, 5, 5, 124, 114, 5, 5, 40, 114, 5, 5, 126, 114, 5, 5, 40, 114, 5, 5, 128, 114, 5, 5, 40, 114, 5, 5, 130, 114, 5, 5, 42, 114, 5, 5, 122, 114, 5, 5, 42, 114, 5, 5, 124, 114, 5, 5, 42, 114, 5, 5, 126, 114, 5, 5, 42, 114, 5, 5, 128, 114, 5, 5, 42, 114, 5, 5, 130, 114, 5, 5, 44, 114, 5, 5, 122, 114, 5, 5, 44, 114, 5, 5, 124, 114, 5, 5, 44, 114, 5, 5, 126, 114, 5, 5, 44, 114, 5, 5, 128, 114, 5, 5, 44, 114, 5, 5, 130, 114, 5, 5, 46, 114, 5, 5, 122, 114, 5, 5, 46, 114, 5, 5, 124, 114, 5, 5, 46, 114, 5, 5, 126, 114, 5, 5, 46, 114, 5, 5, 128, 114, 5, 5, 46, 114, 5, 5, 130, 114, 5, 5, 48, 114, 5, 5, 122, 114, 5, 5, 48, 114, 5, 5, 124, 114, 5, 5, 48, 114, 5, 5, 126, 114, 5, 5, 48, 114, 5, 5, 128, 114, 5, 5, 48, 114, 5, 5, 130, 114, 5, 5, 50, 114, 5, 5, 122, 114, 5, 5, 50, 114, 5, 5, 124, 114, 5, 5, 50, 114, 5, 5, 126, 114, 5, 5, 50, 114, 5, 5, 128, 114, 5, 5, 50, 114, 5, 5, 130, 114, 5, 5, 52, 114, 5, 5, 122, 114, 5, 5, 52, 114, 5, 5, 124, 114, 5, 5, 52, 114, 5, 5, 126, 114, 5, 5, 52, 114, 5, 5, 128, 114, 5, 5, 52, 114, 5, 5, 130, 114, 5, 5, 54, 114, 5, 5, 122, 114, 5, 5, 54, 114, 5, 5, 124, 114, 5, 5, 54, 114, 5, 5, 126, 114, 5, 5, 54, 114, 5, 5, 128, 114, 5, 5, 54, 114, 5, 5, 130, 114, 5, 5, 56, 114, 5, 5, 122, 114, 5, 5, 56, 114, 5, 5, 124, 114, 5, 5, 56, 114, 5, 5, 126, 114, 5, 5, 56, 114, 5, 5, 128, 114, 5, 5, 56, 114, 5, 5, 130, 114, 5, 5, 58, 114, 5, 5, 122, 114, 5, 5, 58, 114, 5, 5, 124, 114, 5, 5, 58, 114, 5, 5, 126, 114, 5, 5, 58, 114, 5, 5, 128, 114, 5, 5, 58, 114, 5, 5, 130, 114, 5, 5, 60, 114, 5, 5, 122, 114, 5, 5, 60, 114, 5, 5, 124, 114, 5, 5, 60, 114, 5, 5, 126, 114, 5, 5, 60, 114, 5, 5, 128, 114, 5, 5, 60, 114, 5, 5, 130, 114, 5, 5, 62, 114, 5, 5, 122, 114, 5, 5, 62, 114, 5, 5, 124, 114, 5, 5, 62, 114, 5, 5, 126, 114, 5, 5, 62, 114, 5, 5, 128, 114, 5, 5, 62, 114, 5, 5, 130, 114, 5, 5, 64, 114, 5, 5, 122, 114, 5, 5, 64, 114, 5, 5, 124, 114, 5, 5, 64, 114, 5, 5, 126, 114, 5, 5, 64, 114, 5, 5, 128, 114, 5, 5, 64, 114, 5, 5, 130, 114, 5, 5, 66, 114, 5, 5, 122, 114, 5, 5, 66, 114, 5, 5, 124, 114, 5, 5, 66, 114, 5, 5, 126, 114, 5, 5, 66, 114, 5, 5, 128, 114, 5, 5, 66, 114, 5, 5, 130, 114, 5, 5, 68, 114, 5, 5, 122, 114, 5, 5, 68, 114, 5, 5, 124, 114, 5, 5, 68, 114, 5, 5, 126, 114, 5, 5, 68, 114, 5, 5, 128, 114, 5, 5, 68, 114, 5, 5, 130, 114, 5, 5, 70, 114, 5, 5, 122, 114, 5, 5, 70, 114, 5, 5, 124, 114, 5, 5, 70, 114, 5, 5, 126, 114, 5, 5, 70, 114, 5, 5, 128, 114, 5, 5, 70, 114, 5, 5, 130, 114, 5, 5, 72, 114, 5, 5, 122, 114, 5, 5, 72, 114, 5, 5, 124, 114, 5, 5, 72, 114, 5, 5, 126, 114, 5, 5, 72, 114, 5, 5, 128, 114, 5, 5, 72, 114, 5, 5, 130, 114, 5, 5, 74, 114, 5, 5, 122, 114, 5, 5, 74, 114, 5, 5, 124, 114, 5, 5, 74, 114, 5, 5, 126, 114, 5, 5, 74, 114, 5, 5, 128, 114, 5, 5, 74, 114, 5, 5, 130, 114, 5, 5, 76, 114, 5, 5, 122, 114, 5, 5, 76, 114, 5, 5, 124, 114, 5, 5, 76, 114, 5, 5, 126, 114, 5, 5, 76, 114, 5, 5, 128, 114, 5, 5, 76, 114, 5, 5, 130, 114, 5, 5, 78, 114, 5, 5, 122, 114, 5, 5, 78, 114, 5, 5, 124, 114, 5, 5, 78, 114, 5, 5, 126, 114, 5, 5, 78, 114, 5, 5, 128, 114, 5, 5, 78, 114, 5, 5, 130, 114, 5, 5, 80, 114, 5, 5, 122, 114, 5, 5, 80, 114, 5, 5, 124, 114, 5, 5, 80, 114, 5, 5, 126, 114, 5, 5, 80, 114, 5, 5, 128, 114, 5, 5, 80, 114, 5, 5, 130, 114, 5, 5, 82, 114, 5, 5, 122, 114, 5, 5, 82, 114, 5, 5, 124, 114, 5, 5, 82, 114, 5, 5, 126, 114, 5, 5, 82, 114, 5, 5, 128, 114, 5, 5, 82, 114, 5, 5, 130, 114, 5, 5, 84, 114, 5, 5, 122, 114, 5, 5, 84, 114, 5, 5, 124, 114, 5, 5, 84, 114, 5, 5, 126, 114, 5, 5, 84, 114, 5, 5, 128, 114, 5, 5, 84, 114, 5, 5, 130, 114, 5, 5, 86, 114, 5, 5, 122, 114, 5, 5, 86, 114, 5, 5, 124, 114, 5, 5, 86, 114, 5, 5, 126, 114, 5, 5, 86, 114, 5, 5, 128, 114, 5, 5, 86, 114, 5, 5, 130, 114, 5, 5, 88, 114, 5, 5, 122, 114, 5, 5, 88, 114, 5, 5, 124, 114, 5, 5, 88, 114, 5, 5, 126, 114, 5, 5, 88, 114, 5, 5, 128, 114, 5, 5, 88, 114, 5, 5, 130, 114, 5, 5, 90, 114, 5, 5, 122, 114, 5, 5, 90, 114, 5, 5, 124, 114, 5, 5, 90, 114, 5, 5, 126, 114, 5, 5, 90, 114, 5, 5, 128, 114, 5, 5, 90, 114, 5, 5, 130, 114, 5, 5, 92, 114, 5, 5, 122, 114, 5, 5, 92, 114, 5, 5, 124, 114, 5, 5, 92, 114, 5, 5, 126, 114, 5, 5, 92, 114, 5, 5, 128, 114, 5, 5, 92, 114, 5, 5, 130, 114, 5, 5, 94, 114, 5, 5, 122, 114, 5, 5, 94, 114, 5, 5, 124, 114, 5, 5, 94, 114, 5, 5, 126, 114, 5, 5, 94, 114, 5, 5, 128, 114, 5, 5, 94, 114, 5, 5, 130, 114, 5, 5, 96, 114, 5, 5, 122, 114, 5, 5, 96, 114, 5, 5, 124, 114, 5, 5, 96, 114, 5, 5, 126, 114, 5, 5, 96, 114, 5, 5, 128, 114, 5, 5, 96, 114, 5, 5, 130, 114, 5, 5, 6, 115, 5, 5, 122, 115, 5, 5, 6, 115, 5, 5, 124, 115, 5, 5, 6, 115, 5, 5, 126, 115, 5, 5, 6, 115, 5, 5, 128, 115, 5, 5, 6, 115, 5, 5, 130, 115, 5, 5, 8, 115, 5, 5, 122, 115, 5, 5, 8, 115, 5, 5, 124, 115, 5, 5, 8, 115, 5, 5, 126, 115, 5, 5, 8, 115, 5, 5, 128, 115, 5, 5, 8, 115, 5, 5, 130, 115, 5, 5, 10, 115, 5, 5, 122, 115, 5, 5, 10, 115, 5, 5, 124, 115, 5, 5, 10, 115, 5, 5, 126, 115, 5, 5, 10, 115, 5, 5, 128, 115, 5, 5, 10, 115, 5, 5, 130, 115, 5, 5, 12, 115, 5, 5, 122, 115, 5, 5, 12, 115, 5, 5, 124, 115, 5, 5, 12, 115, 5, 5, 126, 115, 5, 5, 12, 115, 5, 5, 128, 115, 5, 5, 12, 115, 5, 5, 130, 115, 5, 5, 14, 115, 5, 5, 122, 115, 5, 5, 14, 115, 5, 5, 124, 115, 5, 5, 14, 115, 5, 5, 126, 115, 5, 5, 14, 115, 5, 5, 128, 115, 5, 5, 14, 115, 5, 5, 130, 115, 5, 5, 16, 115, 5, 5, 122, 115, 5, 5, 16, 115, 5, 5, 124, 115, 5, 5, 16, 115, 5, 5, 126, 115, 5, 5, 16, 115, 5, 5, 128, 115, 5, 5, 16, 115, 5, 5, 130, 115, 5, 5, 18, 115, 5, 5, 122, 115, 5, 5, 18, 115, 5, 5, 124, 115, 5, 5, 18, 115, 5, 5, 126, 115, 5, 5, 18, 115, 5, 5, 128, 115, 5, 5, 18, 115, 5, 5, 130, 115, 5, 5, 20, 115, 5, 5, 122, 115, 5, 5, 20, 115, 5, 5, 124, 115, 5, 5, 20, 115, 5, 5, 126, 115, 5, 5, 20, 115, 5, 5, 128, 115, 5, 5, 20, 115, 5, 5, 130, 115, 5, 5, 22, 115, 5, 5, 122, 115, 5, 5, 22, 115, 5, 5, 124, 115, 5, 5, 22, 115, 5, 5, 126, 115, 5, 5, 22, 115, 5, 5, 128, 115, 5, 5, 22, 115, 5, 5, 130, 115, 5, 5, 24, 115, 5, 5, 122, 115, 5, 5, 24, 115, 5, 5, 124, 115, 5, 5, 24, 115, 5, 5, 126, 115, 5, 5, 24, 115, 5, 5, 128, 115, 5, 5, 24, 115, 5, 5, 130, 115, 5, 5, 26, 115, 5, 5, 122, 115, 5, 5, 26, 115, 5, 5, 124, 115, 5, 5, 26, 115, 5, 5, 126, 115, 5, 5, 26, 115, 5, 5, 128, 115, 5, 5, 26, 115, 5, 5, 130, 115, 5, 5, 28, 115, 5, 5, 122, 115, 5, 5, 28, 115, 5, 5, 124, 115, 5, 5, 28, 115, 5, 5, 126, 115, 5, 5, 28, 115, 5, 5, 128, 115, 5, 5, 28, 115, 5, 5, 130, 115, 5, 5, 30, 115, 5, 5, 122, 115, 5, 5, 30, 115, 5, 5, 124, 115, 5, 5, 30, 115, 5, 5, 126, 115, 5, 5, 30, 115, 5, 5, 128, 115, 5, 5, 30, 115, 5, 5, 130, 115, 5, 5, 32, 115, 5, 5, 122, 115, 5, 5, 32, 115, 5, 5, 124, 115, 5, 5, 32, 115, 5, 5, 126, 115, 5, 5, 32, 115, 5, 5, 128, 115, 5, 5, 32, 115, 5, 5, 130, 115, 5, 5, 34, 115, 5, 5, 122, 115, 5, 5, 34, 115, 5, 5, 124, 115, 5, 5, 34, 115, 5, 5, 126, 115, 5, 5, 34, 115, 5, 5, 128, 115, 5, 5, 34, 115, 5, 5, 130, 115, 5, 5, 36, 115, 5, 5, 122, 115, 5, 5, 36, 115, 5, 5, 124, 115, 5, 5, 36, 115, 5, 5, 126, 115, 5, 5, 36, 115, 5, 5, 128, 115, 5, 5, 36, 115, 5, 5, 130, 115, 5, 5, 38, 115, 5, 5, 122, 115, 5, 5, 38, 115, 5, 5, 124, 115, 5, 5, 38, 115, 5, 5, 126, 115, 5, 5, 38, 115, 5, 5, 128, 115, 5, 5, 38, 115, 5, 5, 130, 115, 5, 5, 40, 115, 5, 5, 122, 115, 5, 5, 40, 115, 5, 5, 124, 115, 5, 5, 40, 115, 5, 5, 126, 115, 5, 5, 40, 115, 5, 5, 128, 115, 5, 5, 40, 115, 5, 5, 130, 115, 5, 5, 42, 115, 5, 5, 122, 115, 5, 5, 42, 115, 5, 5, 124, 115, 5, 5, 42, 115, 5, 5, 126, 115, 5, 5, 42, 115, 5, 5, 128, 115, 5, 5, 42, 115, 5, 5, 130, 115, 5, 5, 44, 115, 5, 5, 122, 115, 5, 5, 44, 115, 5, 5, 124, 115, 5, 5, 44, 115, 5, 5, 126, 115, 5, 5, 44, 115, 5, 5, 128, 115, 5, 5, 44, 115, 5, 5, 130, 115, 5, 5, 46, 115, 5, 5, 122, 115, 5, 5, 46, 115, 5, 5, 124, 115, 5, 5, 46, 115, 5, 5, 126, 115, 5, 5, 46, 115, 5, 5, 128, 115, 5, 5, 46, 115, 5, 5, 130, 115, 5, 5, 48, 115, 5, 5, 122, 115, 5, 5, 48, 115, 5, 5, 124, 115, 5, 5, 48, 115, 5, 5, 126, 115, 5, 5, 48, 115, 5, 5, 128, 115, 5, 5, 48, 115, 5, 5, 130, 115, 5, 5, 50, 115, 5, 5, 122, 115, 5, 5, 50, 115, 5, 5, 124, 115, 5, 5, 50, 115, 5, 5, 126, 115, 5, 5, 50, 115, 5, 5, 128, 115, 5, 5, 50, 115, 5, 5, 130, 115, 5, 5, 52, 115, 5, 5, 122, 115, 5, 5, 52, 115, 5, 5, 124, 115, 5, 5, 52, 115, 5, 5, 126, 115, 5, 5, 52, 115, 5, 5, 128, 115, 5, 5, 52, 115, 5, 5, 130, 115, 5, 5, 54, 115, 5, 5, 122, 115, 5, 5, 54, 115, 5, 5, 124, 115, 5, 5, 54, 115, 5, 5, 126, 115, 5, 5, 54, 115, 5, 5, 128, 115, 5, 5, 54, 115, 5, 5, 130, 115, 5, 5, 56, 115, 5, 5, 122, 115, 5, 5, 56, 115, 5, 5, 124, 115, 5, 5, 56, 115, 5, 5, 126, 115, 5, 5, 56, 115, 5, 5, 128, 115, 5, 5, 56, 115, 5, 5, 130, 115, 5, 5, 58, 115, 5, 5, 122, 115, 5, 5, 58, 115, 5, 5, 124, 115, 5, 5, 58, 115, 5, 5, 126, 115, 5, 5, 58, 115, 5, 5, 128, 115, 5, 5, 58, 115, 5, 5, 130, 115, 5, 5, 60, 115, 5, 5, 122, 115, 5, 5, 60, 115, 5, 5, 124, 115, 5, 5, 60, 115, 5, 5, 126, 115, 5, 5, 60, 115, 5, 5, 128, 115, 5, 5, 60, 115, 5, 5, 130, 115, 5, 5, 62, 115, 5, 5, 122, 115, 5, 5, 62, 115, 5, 5, 124, 115, 5, 5, 62, 115, 5, 5, 126, 115, 5, 5, 62, 115, 5, 5, 128, 115, 5, 5, 62, 115, 5, 5, 130, 115, 5, 5, 64, 115, 5, 5, 122, 115, 5, 5, 64, 115, 5, 5, 124, 115, 5, 5, 64, 115, 5, 5, 126, 115, 5, 5, 64, 115, 5, 5, 128, 115, 5, 5, 64, 115, 5, 5, 130, 115, 5, 5, 66, 115, 5, 5, 122, 115, 5, 5, 66, 115, 5, 5, 124, 115, 5, 5, 66, 115, 5, 5, 126, 115, 5, 5, 66, 115, 5, 5, 128, 115, 5, 5, 66, 115, 5, 5, 130, 115, 5, 5, 68, 115, 5, 5, 122, 115, 5, 5, 68, 115, 5, 5, 124, 115, 5, 5, 68, 115, 5, 5, 126, 115, 5, 5, 68, 115, 5, 5, 128, 115, 5, 5, 68, 115, 5, 5, 130, 115, 5, 5, 70, 115, 5, 5, 122, 115, 5, 5, 70, 115, 5, 5, 124, 115, 5, 5, 70, 115, 5, 5, 126, 115, 5, 5, 70, 115, 5, 5, 128, 115, 5, 5, 70, 115, 5, 5, 130, 115, 5, 5, 72, 115, 5, 5, 122, 115, 5, 5, 72, 115, 5, 5, 124, 115, 5, 5, 72, 115, 5, 5, 126, 115, 5, 5, 72, 115, 5, 5, 128, 115, 5, 5, 72, 115, 5, 5, 130, 115, 5, 5, 74, 115, 5, 5, 122, 115, 5, 5, 74, 115, 5, 5, 124, 115, 5, 5, 74, 115, 5, 5, 126, 115, 5, 5, 74, 115, 5, 5, 128, 115, 5, 5, 74, 115, 5, 5, 130, 115, 5, 5, 76, 115, 5, 5, 122, 115, 5, 5, 76, 115, 5, 5, 124, 115, 5, 5, 76, 115, 5, 5, 126, 115, 5, 5, 76, 115, 5, 5, 128, 115, 5, 5, 76, 115, 5, 5, 130, 115, 5, 5, 78, 115, 5, 5, 122, 115, 5, 5, 78, 115, 5, 5, 124, 115, 5, 5, 78, 115, 5, 5, 126, 115, 5, 5, 78, 115, 5, 5, 128, 115, 5, 5, 78, 115, 5, 5, 130, 115, 5, 5, 80, 115, 5, 5, 122, 115, 5, 5, 80, 115, 5, 5, 124, 115, 5, 5, 80, 115, 5, 5, 126, 115, 5, 5, 80, 115, 5, 5, 128, 115, 5, 5, 80, 115, 5, 5, 130, 115, 5, 5, 82, 115, 5, 5, 122, 115, 5, 5, 82, 115, 5, 5, 124, 115, 5, 5, 82, 115, 5, 5, 126, 115, 5, 5, 82, 115, 5, 5, 128, 115, 5, 5, 82, 115, 5, 5, 130, 115, 32, 5, 84, 115, 32, 5, 54, 115, 32, 5, 84, 115, 32, 5, 54, 115, 5, 5, 122, 115, 32, 5, 84, 115, 32, 5, 54, 115, 5, 5, 124, 115, 32, 5, 84, 115, 32, 5, 54, 115, 5, 5, 126, 115, 32, 5, 84, 115, 32, 5, 54, 115, 5, 5, 128, 115, 32, 5, 84, 115, 32, 5, 54, 115, 5, 5, 130, 115, 32, 5, 84, 115, 32, 5, 70, 115, 32, 5, 84, 115, 32, 5, 70, 115, 5, 5, 122, 115, 32, 5, 84, 115, 32, 5, 70, 115, 5, 5, 124, 115, 32, 5, 84, 115, 32, 5, 70, 115, 5, 5, 126, 115, 32, 5, 84, 115, 32, 5, 70, 115, 5, 5, 128, 115, 32, 5, 84, 115, 32, 5, 70, 115, 5, 5, 130, 115, 5, 5, 84, 115, 5, 5, 122, 115, 5, 5, 84, 115, 5, 5, 124, 115, 5, 5, 84, 115, 5, 5, 126, 115, 5, 5, 84, 115, 5, 5, 128, 115, 5, 5, 84, 115, 5, 5, 130, 115, 5, 5, 86, 115, 5, 5, 122, 115, 5, 5, 86, 115, 5, 5, 124, 115, 5, 5, 86, 115, 5, 5, 126, 115, 5, 5, 86, 115, 5, 5, 128, 115, 5, 5, 86, 115, 5, 5, 130, 115, 5, 5, 88, 115, 5, 5, 122, 115, 5, 5, 88, 115, 5, 5, 124, 115, 5, 5, 88, 115, 5, 5, 126, 115, 5, 5, 88, 115, 5, 5, 128, 115, 5, 5, 88, 115, 5, 5, 130, 115, 5, 5, 90, 115, 5, 5, 122, 115, 5, 5, 90, 115, 5, 5, 124, 115, 5, 5, 90, 115, 5, 5, 126, 115, 5, 5, 90, 115, 5, 5, 128, 115, 5, 5, 90, 115, 5, 5, 130, 115, 193, 9, 132, 115, 193, 126, 133, 115, 193, 9, 132, 115, 193, 133, 133, 115, 193, 9, 132, 115, 193, 154, 133, 115, 193, 9, 132, 115, 193, 168, 133, 115, 193, 9, 132, 115, 193, 175, 133, 115, 193, 16, 132, 115, 193, 126, 133, 115, 193, 16, 132, 115, 193, 133, 133, 115, 193, 16, 132, 115, 193, 154, 133, 115, 193, 16, 132, 115, 193, 168, 133, 115, 193, 16, 132, 115, 193, 175, 133, 115, 193, 23, 132, 115, 193, 126, 133, 115, 193, 23, 132, 115, 193, 133, 133, 115, 193, 23, 132, 115, 193, 154, 133, 115, 193, 23, 132, 115, 193, 168, 133, 115, 193, 23, 132, 115, 193, 175, 133, 115, 193, 30, 132, 115, 193, 126, 133, 115, 193, 30, 132, 115, 193, 133, 133, 115, 193, 30, 132, 115, 193, 154, 133, 115, 193, 30, 132, 115, 193, 168, 133, 115, 193, 30, 132, 115, 193, 175, 133, 115, 193, 37, 132, 115, 193, 126, 133, 115, 193, 37, 132, 115, 193, 133, 133, 115, 193, 37, 132, 115, 193, 154, 133, 115, 193, 37, 132, 115, 193, 168, 133, 115, 193, 37, 132, 115, 193, 175, 133, 115, 193, 44, 132, 115, 193, 126, 133, 115, 193, 44, 132, 115, 193, 133, 133, 115, 193, 44, 132, 115, 193, 154, 133, 115, 193, 44, 132, 115, 193, 168, 133, 115, 193, 44, 132, 115, 193, 175, 133, 115, 193, 51, 132, 115, 193, 126, 133, 115, 193, 51, 132, 115, 193, 133, 133, 115, 193, 51, 132, 115, 193, 154, 133, 115, 193, 51, 132, 115, 193, 168, 133, 115, 193, 51, 132, 115, 193, 175, 133, 115, 193, 58, 132, 115, 193, 126, 133, 115, 193, 58, 132, 115, 193, 133, 133, 115, 193, 58, 132, 115, 193, 154, 133, 115, 193, 58, 132, 115, 193, 168, 133, 115, 193, 58, 132, 115, 193, 175, 133, 115, 193, 65, 132, 115, 193, 126, 133, 115, 193, 65, 132, 115, 193, 133, 133, 115, 193, 65, 132, 115, 193, 154, 133, 115, 193, 65, 132, 115, 193, 168, 133, 115, 193, 65, 132, 115, 193, 175, 133, 115, 193, 72, 132, 115, 193, 126, 133, 115, 193, 72, 132, 115, 193, 133, 133, 115, 193, 72, 132, 115, 193, 154, 133, 115, 193, 72, 132, 115, 193, 168, 133, 115, 193, 72, 132, 115, 193, 175, 133, 115, 193, 79, 132, 115, 193, 126, 133, 115, 193, 79, 132, 115, 193, 133, 133, 115, 193, 79, 132, 115, 193, 154, 133, 115, 193, 79, 132, 115, 193, 168, 133, 115, 193, 79, 132, 115, 193, 175, 133, 115, 193, 86, 132, 115, 193, 126, 133, 115, 193, 86, 132, 115, 193, 133, 133, 115, 193, 86, 132, 115, 193, 154, 133, 115, 193, 86, 132, 115, 193, 168, 133, 115, 193, 86, 132, 115, 193, 175, 133, 115, 193, 93, 132, 115, 193, 126, 133, 115, 193, 93, 132, 115, 193, 133, 133, 115, 193, 93, 132, 115, 193, 154, 133, 115, 193, 93, 132, 115, 193, 168, 133, 115, 193, 93, 132, 115, 193, 175, 133, 115, 193, 100, 132, 115, 193, 126, 133, 115, 193, 100, 132, 115, 193, 133, 133, 115, 193, 100, 132, 115, 193, 154, 133, 115, 193, 100, 132, 115, 193, 168, 133, 115, 193, 100, 132, 115, 193, 175, 133, 115, 193, 107, 132, 115, 193, 126, 133, 115, 193, 107, 132, 115, 193, 133, 133, 115, 193, 107, 132, 115, 193, 154, 133, 115, 193, 107, 132, 115, 193, 168, 133, 115, 193, 107, 132, 115, 193, 175, 133, 115, 193, 114, 132, 115, 193, 126, 133, 115, 193, 114, 132, 115, 193, 133, 133, 115, 193, 114, 132, 115, 193, 154, 133, 115, 193, 114, 132, 115, 193, 168, 133, 115, 193, 114, 132, 115, 193, 175, 133, 115, 193, 121, 132, 115, 193, 126, 133, 115, 193, 121, 132, 115, 193, 133, 133, 115, 193, 121, 132, 115, 193, 154, 133, 115, 193, 121, 132, 115, 193, 168, 133, 115, 193, 121, 132, 115, 193, 175, 133, 115, 193, 128, 132, 115, 193, 126, 133, 115, 193, 128, 132, 115, 193, 133, 133, 115, 193, 128, 132, 115, 193, 154, 133, 115, 193, 128, 132, 115, 193, 168, 133, 115, 193, 128, 132, 115, 193, 175, 133, 115, 193, 135, 132, 115, 193, 126, 133, 115, 193, 135, 132, 115, 193, 133, 133, 115, 193, 135, 132, 115, 193, 154, 133, 115, 193, 135, 132, 115, 193, 168, 133, 115, 193, 135, 132, 115, 193, 175, 133, 115, 193, 142, 132, 115, 193, 126, 133, 115, 193, 142, 132, 115, 193, 133, 133, 115, 193, 142, 132, 115, 193, 154, 133, 115, 193, 142, 132, 115, 193, 168, 133, 115, 193, 142, 132, 115, 193, 175, 133, 115, 193, 149, 132, 115, 193, 126, 133, 115, 193, 149, 132, 115, 193, 133, 133, 115, 193, 149, 132, 115, 193, 154, 133, 115, 193, 149, 132, 115, 193, 168, 133, 115, 193, 149, 132, 115, 193, 175, 133, 115, 193, 156, 132, 115, 193, 126, 133, 115, 193, 156, 132, 115, 193, 133, 133, 115, 193, 156, 132, 115, 193, 154, 133, 115, 193, 156, 132, 115, 193, 168, 133, 115, 193, 156, 132, 115, 193, 175, 133, 115, 193, 163, 132, 115, 193, 126, 133, 115, 193, 163, 132, 115, 193, 133, 133, 115, 193, 163, 132, 115, 193, 154, 133, 115, 193, 163, 132, 115, 193, 168, 133, 115, 193, 163, 132, 115, 193, 175, 133, 115, 193, 170, 132, 115, 193, 126, 133, 115, 193, 170, 132, 115, 193, 133, 133, 115, 193, 170, 132, 115, 193, 154, 133, 115, 193, 170, 132, 115, 193, 168, 133, 115, 193, 170, 132, 115, 193, 175, 133, 115, 193, 177, 132, 115, 193, 126, 133, 115, 193, 177, 132, 115, 193, 133, 133, 115, 193, 177, 132, 115, 193, 154, 133, 115, 193, 177, 132, 115, 193, 168, 133, 115, 193, 177, 132, 115, 193, 175, 133, 115, 193, 184, 132, 115, 193, 126, 133, 115, 193, 184, 132, 115, 193, 133, 133, 115, 193, 184, 132, 115, 193, 154, 133, 115, 193, 184, 132, 115, 193, 168, 133, 115, 193, 184, 132, 115, 193, 175, 133, 115, 193, 191, 132, 115, 193, 126, 133, 115, 193, 191, 132, 115, 193, 133, 133, 115, 193, 191, 132, 115, 193, 154, 133, 115, 193, 191, 132, 115, 193, 168, 133, 115, 193, 191, 132, 115, 193, 175, 133, 115, 193, 198, 132, 115, 193, 126, 133, 115, 193, 198, 132, 115, 193, 133, 133, 115, 193, 198, 132, 115, 193, 154, 133, 115, 193, 198, 132, 115, 193, 168, 133, 115, 193, 198, 132, 115, 193, 175, 133, 115, 193, 205, 132, 115, 193, 126, 133, 115, 193, 205, 132, 115, 193, 133, 133, 115, 193, 205, 132, 115, 193, 154, 133, 115, 193, 205, 132, 115, 193, 168, 133, 115, 193, 205, 132, 115, 193, 175, 133, 115, 193, 212, 132, 115, 193, 126, 133, 115, 193, 212, 132, 115, 193, 133, 133, 115, 193, 212, 132, 115, 193, 154, 133, 115, 193, 212, 132, 115, 193, 168, 133, 115, 193, 212, 132, 115, 193, 175, 133, 115, 193, 219, 132, 115, 193, 126, 133, 115, 193, 219, 132, 115, 193, 133, 133, 115, 193, 219, 132, 115, 193, 154, 133, 115, 193, 219, 132, 115, 193, 168, 133, 115, 193, 219, 132, 115, 193, 175, 133, 115, 193, 226, 132, 115, 193, 126, 133, 115, 193, 226, 132, 115, 193, 133, 133, 115, 193, 226, 132, 115, 193, 154, 133, 115, 193, 226, 132, 115, 193, 168, 133, 115, 193, 226, 132, 115, 193, 175, 133, 115, 193, 233, 132, 115, 193, 126, 133, 115, 193, 233, 132, 115, 193, 133, 133, 115, 193, 233, 132, 115, 193, 154, 133, 115, 193, 233, 132, 115, 193, 168, 133, 115, 193, 233, 132, 115, 193, 175, 133, 115, 193, 240, 132, 115, 193, 126, 133, 115, 193, 240, 132, 115, 193, 133, 133, 115, 193, 240, 132, 115, 193, 154, 133, 115, 193, 240, 132, 115, 193, 168, 133, 115, 193, 240, 132, 115, 193, 175, 133, 115, 193, 247, 132, 115, 193, 126, 133, 115, 193, 247, 132, 115, 193, 133, 133, 115, 193, 247, 132, 115, 193, 154, 133, 115, 193, 247, 132, 115, 193, 168, 133, 115, 193, 247, 132, 115, 193, 175, 133, 115, 193, 254, 132, 115, 193, 126, 133, 115, 193, 254, 132, 115, 193, 133, 133, 115, 193, 254, 132, 115, 193, 154, 133, 115, 193, 254, 132, 115, 193, 168, 133, 115, 193, 254, 132, 115, 193, 175, 133, 115, 193, 7, 133, 115, 193, 126, 133, 115, 193, 7, 133, 115, 193, 133, 133, 115, 193, 7, 133, 115, 193, 154, 133, 115, 193, 7, 133, 115, 193, 168, 133, 115, 193, 7, 133, 115, 193, 175, 133, 115, 193, 14, 133, 115, 193, 126, 133, 115, 193, 14, 133, 115, 193, 133, 133, 115, 193, 14, 133, 115, 193, 154, 133, 115, 193, 14, 133, 115, 193, 168, 133, 115, 193, 14, 133, 115, 193, 175, 133, 115, 193, 21, 133, 115, 193, 126, 133, 115, 193, 21, 133, 115, 193, 133, 133, 115, 193, 21, 133, 115, 193, 154, 133, 115, 193, 21, 133, 115, 193, 168, 133, 115, 193, 21, 133, 115, 193, 175, 133, 115, 193, 28, 133, 115, 193, 126, 133, 115, 193, 28, 133, 115, 193, 133, 133, 115, 193, 28, 133, 115, 193, 154, 133, 115, 193, 28, 133, 115, 193, 168, 133, 115, 193, 28, 133, 115, 193, 175, 133, 115, 193, 35, 133, 115, 193, 126, 133, 115, 193, 35, 133, 115, 193, 133, 133, 115, 193, 35, 133, 115, 193, 154, 133, 115, 193, 35, 133, 115, 193, 168, 133, 115, 193, 35, 133, 115, 193, 175, 133, 115, 193, 42, 133, 115, 193, 126, 133, 115, 193, 42, 133, 115, 193, 133, 133, 115, 193, 42, 133, 115, 193, 154, 133, 115, 193, 42, 133, 115, 193, 168, 133, 115, 193, 42, 133, 115, 193, 175, 133, 115, 193, 49, 133, 115, 193, 126, 133, 115, 193, 49, 133, 115, 193, 133, 133, 115, 193, 49, 133, 115, 193, 154, 133, 115, 193, 49, 133, 115, 193, 168, 133, 115, 193, 49, 133, 115, 193, 175, 133, 115, 193, 56, 133, 115, 193, 126, 133, 115, 193, 56, 133, 115, 193, 133, 133, 115, 193, 56, 133, 115, 193, 154, 133, 115, 193, 56, 133, 115, 193, 168, 133, 115, 193, 56, 133, 115, 193, 175, 133, 115, 193, 63, 133, 115, 193, 126, 133, 115, 193, 63, 133, 115, 193, 133, 133, 115, 193, 63, 133, 115, 193, 154, 133, 115, 193, 63, 133, 115, 193, 168, 133, 115, 193, 63, 133, 115, 193, 175, 133, 115, 193, 70, 133, 115, 193, 126, 133, 115, 193, 70, 133, 115, 193, 133, 133, 115, 193, 70, 133, 115, 193, 154, 133, 115, 193, 70, 133, 115, 193, 168, 133, 115, 193, 70, 133, 115, 193, 175, 133, 115, 193, 77, 133, 115, 193, 126, 133, 115, 193, 77, 133, 115, 193, 133, 133, 115, 193, 77, 133, 115, 193, 154, 133, 115, 193, 77, 133, 115, 193, 168, 133, 115, 193, 77, 133, 115, 193, 175, 133, 115, 193, 84, 133, 115, 193, 126, 133, 115, 193, 84, 133, 115, 193, 133, 133, 115, 193, 84, 133, 115, 193, 154, 133, 115, 193, 84, 133, 115, 193, 168, 133, 115, 193, 84, 133, 115, 193, 175, 133, 115, 5, 5, 6, 116, 5, 5, 136, 116, 5, 5, 8, 116, 5, 5, 136, 116, 5, 5, 16, 116, 5, 5, 144, 116, 5, 5, 18, 116, 5, 5, 144, 116, 5, 5, 48, 116, 5, 5, 144, 116, 5, 5, 50, 116, 5, 5, 144, 116, 5, 5, 64, 116, 5, 5, 144, 116, 5, 5, 66, 116, 5, 5, 144, 116, 5, 5, 80, 116, 5, 5, 144, 116, 5, 5, 82, 116, 5, 5, 144, 116, 5, 5, 96, 116, 5, 5, 144, 116, 5, 5, 98, 116, 5, 5, 144, 116, 5, 5, 122, 116, 5, 5, 166, 116, 5, 5, 122, 116, 5, 5, 170, 116, 5, 5, 122, 116, 5, 5, 178, 116, 5, 5, 128, 116, 5, 5, 166, 116, 5, 5, 128, 116, 5, 5, 170, 116, 5, 5, 128, 116, 5, 5, 178, 116, 32, 5, 146, 116, 32, 5, 192, 116, 194, 32, 241, 240, 32, 5, 12, 117, 32, 5, 20, 117, 32, 5, 16, 117, 32, 5, 22, 117, 32, 5, 83, 117, 32, 5, 91, 117, 32, 5, 83, 117, 193, 9, 5, 118, 193, 142, 9, 118, 193, 9, 5, 118, 5, 5, 11, 118, 193, 9, 5, 118, 193, 2, 13, 118, 193, 9, 5, 118, 193, 23, 13, 118, 193, 16, 5, 118, 193, 142, 9, 118, 193, 16, 5, 118, 5, 5, 11, 118, 193, 16, 5, 118, 193, 2, 13, 118, 193, 16, 5, 118, 193, 23, 13, 118, 193, 23, 5, 118, 193, 142, 9, 118, 193, 23, 5, 118, 5, 5, 11, 118, 193, 23, 5, 118, 193, 2, 13, 118, 193, 23, 5, 118, 193, 23, 13, 118, 193, 30, 5, 118, 193, 142, 9, 118, 193, 30, 5, 118, 5, 5, 11, 118, 193, 30, 5, 118, 193, 2, 13, 118, 193, 30, 5, 118, 193, 23, 13, 118, 193, 37, 5, 118, 193, 142, 9, 118, 193, 37, 5, 118, 5, 5, 11, 118, 193, 37, 5, 118, 193, 2, 13, 118, 193, 37, 5, 118, 193, 23, 13, 118, 193, 44, 5, 118, 193, 142, 9, 118, 193, 44, 5, 118, 5, 5, 11, 118, 193, 44, 5, 118, 193, 2, 13, 118, 193, 44, 5, 118, 193, 23, 13, 118, 193, 51, 5, 118, 193, 142, 9, 118, 193, 51, 5, 118, 5, 5, 11, 118, 193, 51, 5, 118, 193, 2, 13, 118, 193, 51, 5, 118, 193, 23, 13, 118, 193, 58, 5, 118, 193, 142, 9, 118, 193, 58, 5, 118, 5, 5, 11, 118, 193, 58, 5, 118, 193, 2, 13, 118, 193, 58, 5, 118, 193, 23, 13, 118, 193, 65, 5, 118, 193, 142, 9, 118, 193, 65, 5, 118, 5, 5, 11, 118, 193, 65, 5, 118, 193, 2, 13, 118, 193, 65, 5, 118, 193, 23, 13, 118, 193, 72, 5, 118, 193, 142, 9, 118, 193, 72, 5, 118, 5, 5, 11, 118, 193, 72, 5, 118, 193, 2, 13, 118, 193, 72, 5, 118, 193, 23, 13, 118, 193, 79, 5, 118, 193, 142, 9, 118, 193, 79, 5, 118, 5, 5, 11, 118, 193, 79, 5, 118, 193, 2, 13, 118, 193, 79, 5, 118, 193, 23, 13, 118, 193, 86, 5, 118, 193, 142, 9, 118, 193, 86, 5, 118, 5, 5, 11, 118, 193, 86, 5, 118, 193, 2, 13, 118, 193, 86, 5, 118, 193, 23, 13, 118, 193, 93, 5, 118, 193, 142, 9, 118, 193, 93, 5, 118, 5, 5, 11, 118, 193, 93, 5, 118, 193, 2, 13, 118, 193, 93, 5, 118, 193, 23, 13, 118, 193, 100, 5, 118, 193, 142, 9, 118, 193, 100, 5, 118, 5, 5, 11, 118, 193, 100, 5, 118, 193, 2, 13, 118, 193, 100, 5, 118, 193, 23, 13, 118, 193, 107, 5, 118, 193, 142, 9, 118, 193, 107, 5, 118, 5, 5, 11, 118, 193, 107, 5, 118, 193, 2, 13, 118, 193, 107, 5, 118, 193, 23, 13, 118, 193, 114, 5, 118, 193, 142, 9, 118, 193, 114, 5, 118, 5, 5, 11, 118, 193, 114, 5, 118, 193, 2, 13, 118, 193, 114, 5, 118, 193, 23, 13, 118, 193, 121, 5, 118, 193, 142, 9, 118, 193, 121, 5, 118, 5, 5, 11, 118, 193, 121, 5, 118, 193, 2, 13, 118, 193, 121, 5, 118, 193, 23, 13, 118, 193, 128, 5, 118, 193, 142, 9, 118, 193, 128, 5, 118, 5, 5, 11, 118, 193, 128, 5, 118, 193, 2, 13, 118, 193, 128, 5, 118, 193, 23, 13, 118, 193, 135, 5, 118, 193, 142, 9, 118, 193, 135, 5, 118, 5, 5, 11, 118, 193, 135, 5, 118, 193, 2, 13, 118, 193, 135, 5, 118, 193, 23, 13, 118, 193, 142, 5, 118, 193, 142, 9, 118, 193, 142, 5, 118, 5, 5, 11, 118, 193, 142, 5, 118, 193, 2, 13, 118, 193, 142, 5, 118, 193, 23, 13, 118, 193, 149, 5, 118, 193, 142, 9, 118, 193, 149, 5, 118, 5, 5, 11, 118, 193, 149, 5, 118, 193, 2, 13, 118, 193, 149, 5, 118, 193, 23, 13, 118, 193, 156, 5, 118, 193, 142, 9, 118, 193, 156, 5, 118, 5, 5, 11, 118, 193, 156, 5, 118, 193, 2, 13, 118, 193, 156, 5, 118, 193, 23, 13, 118, 193, 163, 5, 118, 193, 142, 9, 118, 193, 163, 5, 118, 5, 5, 11, 118, 193, 163, 5, 118, 193, 2, 13, 118, 193, 163, 5, 118, 193, 23, 13, 118, 193, 170, 5, 118, 193, 142, 9, 118, 193, 170, 5, 118, 5, 5, 11, 118, 193, 170, 5, 118, 193, 2, 13, 118, 193, 170, 5, 118, 193, 23, 13, 118, 193, 177, 5, 118, 193, 142, 9, 118, 193, 177, 5, 118, 5, 5, 11, 118, 193, 177, 5, 118, 193, 2, 13, 118, 193, 177, 5, 118, 193, 23, 13, 118, 193, 184, 5, 118, 193, 142, 9, 118, 193, 184, 5, 118, 5, 5, 11, 118, 193, 184, 5, 118, 193, 2, 13, 118, 193, 184, 5, 118, 193, 23, 13, 118, 193, 191, 5, 118, 193, 142, 9, 118, 193, 191, 5, 118, 5, 5, 11, 118, 193, 191, 5, 118, 193, 2, 13, 118, 193, 191, 5, 118, 193, 23, 13, 118, 193, 198, 5, 118, 193, 142, 9, 118, 193, 198, 5, 118, 5, 5, 11, 118, 193, 198, 5, 118, 193, 2, 13, 118, 193, 198, 5, 118, 193, 23, 13, 118, 5, 5, 7, 118, 193, 142, 9, 118, 5, 5, 7, 118, 5, 5, 11, 118, 32, 5, 7, 118, 32, 5, 11, 118, 32, 5, 7, 118, 32, 5, 11, 118, 32, 5, 15, 118, 5, 5, 7, 118, 193, 2, 13, 118, 5, 5, 7, 118, 193, 23, 13, 118, 193, 2, 9, 118, 193, 142, 9, 118, 193, 2, 9, 118, 5, 5, 11, 118, 193, 2, 9, 118, 193, 2, 13, 118, 193, 2, 9, 118, 193, 23, 13, 118, 193, 9, 9, 118, 193, 142, 9, 118, 193, 9, 9, 118, 5, 5, 11, 118, 193, 9, 9, 118, 193, 2, 13, 118, 193, 9, 9, 118, 193, 23, 13, 118, 193, 16, 9, 118, 193, 142, 9, 118, 193, 16, 9, 118, 5, 5, 11, 118, 193, 16, 9, 118, 193, 2, 13, 118, 193, 16, 9, 118, 193, 23, 13, 118, 193, 23, 9, 118, 193, 142, 9, 118, 193, 23, 9, 118, 5, 5, 11, 118, 193, 23, 9, 118, 193, 2, 13, 118, 193, 23, 9, 118, 193, 23, 13, 118, 193, 30, 9, 118, 193, 142, 9, 118, 193, 30, 9, 118, 5, 5, 11, 118, 193, 30, 9, 118, 193, 2, 13, 118, 193, 30, 9, 118, 193, 23, 13, 118, 193, 37, 9, 118, 193, 142, 9, 118, 193, 37, 9, 118, 5, 5, 11, 118, 193, 37, 9, 118, 193, 2, 13, 118, 193, 37, 9, 118, 193, 23, 13, 118, 193, 44, 9, 118, 193, 142, 9, 118, 193, 44, 9, 118, 5, 5, 11, 118, 193, 44, 9, 118, 193, 2, 13, 118, 193, 44, 9, 118, 193, 23, 13, 118, 193, 51, 9, 118, 193, 142, 9, 118, 193, 51, 9, 118, 5, 5, 11, 118, 193, 51, 9, 118, 193, 2, 13, 118, 193, 51, 9, 118, 193, 23, 13, 118, 193, 58, 9, 118, 193, 142, 9, 118, 193, 58, 9, 118, 5, 5, 11, 118, 193, 58, 9, 118, 193, 2, 13, 118, 193, 58, 9, 118, 193, 23, 13, 118, 193, 65, 9, 118, 193, 142, 9, 118, 193, 65, 9, 118, 5, 5, 11, 118, 193, 65, 9, 118, 193, 2, 13, 118, 193, 65, 9, 118, 193, 23, 13, 118, 193, 72, 9, 118, 193, 142, 9, 118, 193, 72, 9, 118, 5, 5, 11, 118, 193, 72, 9, 118, 193, 2, 13, 118, 193, 72, 9, 118, 193, 23, 13, 118, 193, 79, 9, 118, 193, 142, 9, 118, 193, 79, 9, 118, 5, 5, 11, 118, 193, 79, 9, 118, 193, 2, 13, 118, 193, 79, 9, 118, 193, 23, 13, 118, 193, 86, 9, 118, 193, 142, 9, 118, 193, 86, 9, 118, 5, 5, 11, 118, 193, 86, 9, 118, 193, 2, 13, 118, 193, 86, 9, 118, 193, 23, 13, 118, 193, 93, 9, 118, 193, 142, 9, 118, 193, 93, 9, 118, 5, 5, 11, 118, 193, 93, 9, 118, 193, 2, 13, 118, 193, 93, 9, 118, 193, 23, 13, 118, 193, 100, 9, 118, 193, 142, 9, 118, 193, 100, 9, 118, 5, 5, 11, 118, 193, 100, 9, 118, 193, 2, 13, 118, 193, 100, 9, 118, 193, 23, 13, 118, 32, 5, 28, 118, 32, 5, 36, 118, 32, 5, 28, 118, 32, 5, 10, 119, 32, 5, 10, 119, 32, 5, 32, 119, 32, 5, 52, 119, 5, 112, 50, 119, 5, 112, 50, 119, 32, 5, 6, 120, 32, 5, 72, 120, 32, 5, 10, 120, 32, 5, 72, 120, 32, 5, 14, 120, 32, 5, 72, 120, 32, 5, 18, 120, 32, 5, 74, 120, 32, 5, 22, 120, 32, 5, 72, 120, 32, 5, 26, 120, 32, 5, 74, 120, 32, 5, 30, 120, 32, 5, 74, 120, 32, 5, 34, 120, 32, 5, 72, 120, 32, 5, 46, 120, 32, 5, 72, 120, 32, 5, 55, 120, 32, 5, 72, 120, 32, 5, 59, 120, 32, 5, 72, 120, 32, 5, 63, 120, 32, 5, 74, 120, 32, 5, 67, 120, 32, 5, 72, 120, 30, 5, 6, 121, 30, 5, 103, 121, 30, 5, 10, 121, 30, 5, 103, 121, 30, 5, 12, 121, 30, 5, 103, 121, 30, 5, 16, 121, 30, 5, 103, 121, 30, 5, 18, 121, 30, 5, 103, 121, 30, 5, 20, 121, 30, 5, 103, 121, 30, 5, 24, 121, 30, 5, 103, 121, 30, 5, 28, 121, 30, 5, 103, 121, 30, 5, 28, 121, 30, 5, 129, 121, 30, 5, 30, 121, 30, 5, 103, 121, 30, 5, 30, 121, 30, 5, 129, 121, 30, 5, 28, 121, 30, 5, 141, 121, 30, 5, 34, 121, 30, 5, 103, 121, 30, 5, 34, 121, 30, 5, 103, 121, 32, 5, 209, 121, 30, 5, 6, 121, 30, 5, 119, 121, 30, 5, 36, 121, 30, 5, 103, 121, 30, 5, 38, 121, 30, 5, 103, 121, 30, 5, 40, 121, 30, 5, 103, 121, 30, 5, 42, 121, 30, 5, 103, 121, 51, 5, 6, 122, 46, 5, 118, 13, 48, 5, 88, 122, 51, 5, 6, 122, 48, 5, 58, 122, 46, 176, 0, 0, 46, 5, 118, 13, 51, 5, 46, 122, 51, 5, 6, 122, 48, 5, 88, 122, 48, 5, 62, 122, 51, 5, 6, 122, 51, 5, 6, 122, 49, 5, 102, 122, 48, 5, 64, 122, 46, 176, 0, 0, 51, 5, 6, 122, 51, 5, 8, 122, 49, 5, 50, 122, 49, 5, 102, 122, 48, 5, 22, 122, 46, 174, 0, 0, 51, 5, 8, 122, 49, 5, 102, 122, 49, 5, 40, 122, 133, 5, 10, 122, 5, 174, 0, 0, 158, 5, 10, 122, 5, 174, 0, 0, 51, 5, 10, 122, 51, 5, 16, 122, 49, 5, 102, 122, 51, 5, 14, 122, 46, 5, 118, 13, 48, 5, 18, 122, 46, 5, 118, 13, 51, 5, 14, 122, 48, 5, 32, 122, 48, 5, 22, 122, 46, 5, 118, 13, 51, 5, 46, 122, 46, 174, 0, 0, 51, 5, 16, 122, 46, 5, 118, 13, 48, 5, 72, 122, 51, 5, 16, 122, 49, 5, 102, 122, 48, 5, 32, 122, 133, 5, 18, 122, 5, 174, 0, 0, 152, 5, 18, 122, 5, 174, 0, 0, 48, 5, 18, 122, 51, 5, 8, 122, 51, 5, 86, 122, 48, 5, 18, 122, 48, 5, 84, 122, 51, 5, 42, 122, 51, 5, 46, 122, 48, 5, 18, 122, 48, 5, 92, 122, 51, 5, 86, 122, 46, 5, 118, 13, 48, 5, 18, 122, 46, 174, 0, 0, 48, 5, 92, 122, 49, 5, 102, 122, 48, 5, 18, 122, 46, 174, 0, 0, 49, 5, 102, 122, 49, 5, 68, 122, 133, 5, 20, 122, 5, 174, 0, 0, 144, 5, 20, 122, 5, 174, 0, 0, 49, 5, 20, 122, 46, 174, 0, 0, 48, 5, 18, 122, 46, 174, 0, 0, 49, 5, 20, 122, 46, 174, 0, 0, 49, 5, 50, 122, 46, 5, 118, 13, 49, 5, 20, 122, 51, 5, 80, 122, 51, 5, 86, 122, 46, 5, 118, 13, 49, 5, 20, 122, 46, 174, 0, 0, 48, 5, 88, 122, 49, 5, 38, 122, 46, 174, 0, 0, 46, 5, 118, 13, 49, 5, 20, 122, 48, 5, 92, 122, 49, 5, 20, 122, 48, 5, 92, 122, 48, 5, 22, 122, 46, 174, 0, 0, 48, 5, 84, 122, 48, 5, 72, 122, 49, 5, 20, 122, 48, 5, 92, 122, 49, 5, 74, 122, 46, 5, 118, 13, 51, 5, 46, 122, 48, 5, 88, 122, 49, 5, 20, 122, 48, 5, 92, 122, 48, 5, 94, 122, 51, 5, 42, 122, 51, 5, 46, 122, 133, 5, 22, 122, 5, 174, 0, 0, 152, 5, 22, 122, 5, 174, 0, 0, 48, 5, 22, 122, 46, 174, 0, 0, 48, 5, 84, 122, 48, 5, 72, 122, 51, 5, 46, 122, 49, 5, 102, 122, 48, 5, 22, 122, 48, 5, 88, 122, 49, 5, 34, 122, 46, 174, 0, 0, 51, 5, 8, 122, 48, 5, 92, 122, 48, 5, 22, 122, 48, 5, 92, 122, 46, 5, 118, 13, 49, 5, 54, 122, 133, 5, 24, 122, 5, 174, 0, 0, 152, 5, 24, 122, 5, 174, 0, 0, 48, 5, 24, 122, 46, 5, 118, 13, 48, 5, 32, 122, 133, 5, 26, 122, 5, 174, 0, 0, 144, 5, 26, 122, 5, 174, 0, 0, 48, 5, 26, 122, 46, 5, 118, 13, 48, 5, 66, 122, 46, 176, 0, 0, 48, 5, 26, 122, 48, 5, 26, 122, 40, 5, 26, 122, 44, 5, 46, 122, 48, 5, 26, 122, 48, 5, 88, 122, 49, 5, 48, 122, 133, 5, 28, 122, 5, 174, 0, 0, 144, 5, 28, 122, 5, 174, 0, 0, 49, 5, 28, 122, 51, 5, 8, 122, 48, 5, 22, 122, 48, 5, 88, 122, 49, 5, 28, 122, 49, 5, 102, 122, 49, 5, 40, 122, 46, 5, 118, 13, 48, 5, 72, 122, 133, 5, 30, 122, 5, 174, 0, 0, 152, 5, 30, 122, 5, 174, 0, 0, 3, 5, 30, 122, 3, 5, 74, 122, 48, 5, 30, 122, 51, 5, 86, 122, 49, 5, 102, 122, 48, 5, 22, 122, 46, 174, 0, 0, 133, 5, 32, 122, 5, 174, 0, 0, 152, 5, 32, 122, 5, 174, 0, 0, 133, 5, 34, 122, 5, 174, 0, 0, 144, 5, 34, 122, 5, 174, 0, 0, 49, 5, 34, 122, 49, 5, 102, 122, 49, 5, 40, 122, 49, 5, 34, 122, 49, 5, 102, 122, 51, 5, 46, 122, 133, 5, 36, 122, 5, 174, 0, 0, 144, 5, 36, 122, 5, 174, 0, 0, 133, 5, 38, 122, 5, 174, 0, 0, 144, 5, 38, 122, 5, 174, 0, 0, 49, 5, 38, 122, 46, 174, 0, 0, 46, 5, 118, 13, 48, 5, 32, 122, 133, 5, 40, 122, 5, 174, 0, 0, 144, 5, 40, 122, 5, 174, 0, 0, 133, 5, 42, 122, 5, 174, 0, 0, 158, 5, 42, 122, 5, 174, 0, 0, 133, 5, 44, 122, 5, 174, 0, 0, 144, 5, 44, 122, 5, 174, 0, 0, 49, 5, 44, 122, 46, 174, 0, 0, 49, 5, 44, 122, 46, 174, 0, 0, 48, 5, 30, 122, 133, 5, 46, 122, 5, 174, 0, 0, 151, 5, 46, 122, 5, 174, 0, 0, 51, 5, 46, 122, 46, 174, 0, 0, 48, 5, 88, 122, 49, 5, 48, 122, 49, 5, 56, 122, 49, 5, 56, 122, 51, 5, 42, 122, 51, 5, 46, 122, 133, 5, 58, 122, 5, 174, 0, 0, 152, 5, 58, 122, 5, 174, 0, 0, 133, 5, 58, 122, 5, 176, 0, 0, 152, 5, 58, 122, 5, 176, 0, 0, 48, 5, 58, 122, 46, 176, 0, 0, 46, 5, 118, 13, 49, 5, 34, 122, 49, 5, 102, 122, 51, 5, 46, 122, 48, 5, 58, 122, 46, 176, 0, 0, 46, 5, 118, 13, 51, 5, 42, 122, 48, 5, 58, 122, 46, 174, 0, 0, 46, 5, 118, 13, 48, 5, 90, 122, 48, 5, 88, 122, 48, 5, 58, 122, 51, 5, 8, 122, 51, 5, 42, 122, 133, 5, 60, 122, 5, 174, 0, 0, 152, 5, 60, 122, 5, 174, 0, 0, 133, 5, 60, 122, 5, 176, 0, 0, 152, 5, 60, 122, 5, 176, 0, 0, 48, 5, 60, 122, 46, 176, 0, 0, 51, 5, 6, 122, 48, 5, 32, 122, 51, 5, 46, 122, 48, 5, 88, 122, 48, 5, 60, 122, 46, 176, 0, 0, 48, 5, 22, 122, 48, 5, 88, 122, 48, 5, 60, 122, 46, 176, 0, 0, 48, 5, 26, 122, 48, 5, 60, 122, 46, 174, 0, 0, 48, 5, 88, 122, 133, 5, 62, 122, 5, 174, 0, 0, 152, 5, 62, 122, 5, 174, 0, 0, 133, 5, 62, 122, 5, 176, 0, 0, 152, 5, 62, 122, 5, 176, 0, 0, 48, 5, 62, 122, 51, 5, 6, 122, 48, 5, 84, 122, 51, 5, 42, 122, 51, 5, 46, 122, 46, 174, 0, 0, 48, 5, 62, 122, 51, 5, 8, 122, 46, 5, 118, 13, 51, 5, 46, 122, 48, 5, 62, 122, 46, 174, 0, 0, 51, 5, 42, 122, 48, 5, 30, 122, 51, 5, 14, 122, 48, 5, 88, 122, 48, 5, 62, 122, 48, 5, 84, 122, 49, 5, 102, 122, 133, 5, 64, 122, 5, 174, 0, 0, 152, 5, 64, 122, 5, 174, 0, 0, 133, 5, 64, 122, 5, 176, 0, 0, 152, 5, 64, 122, 5, 176, 0, 0, 48, 5, 64, 122, 46, 176, 0, 0, 46, 5, 118, 13, 48, 5, 30, 122, 46, 174, 0, 0, 48, 5, 64, 122, 46, 174, 0, 0, 46, 5, 118, 13, 49, 5, 38, 122, 48, 5, 64, 122, 48, 5, 22, 122, 49, 5, 38, 122, 46, 5, 118, 13, 48, 5, 88, 122, 48, 5, 64, 122, 46, 176, 0, 0, 49, 5, 36, 122, 48, 5, 64, 122, 46, 176, 0, 0, 49, 5, 50, 122, 48, 5, 60, 122, 48, 5, 64, 122, 48, 5, 88, 122, 51, 5, 42, 122, 48, 5, 64, 122, 46, 176, 0, 0, 49, 5, 102, 122, 48, 5, 32, 122, 133, 5, 66, 122, 5, 174, 0, 0, 152, 5, 66, 122, 5, 174, 0, 0, 133, 5, 66, 122, 5, 176, 0, 0, 152, 5, 66, 122, 5, 176, 0, 0, 48, 5, 66, 122, 46, 5, 118, 13, 48, 5, 88, 122, 48, 5, 66, 122, 46, 5, 118, 13, 49, 5, 102, 122, 48, 5, 66, 122, 46, 176, 0, 0, 51, 5, 8, 122, 49, 5, 102, 122, 51, 5, 46, 122, 48, 5, 66, 122, 48, 5, 18, 122, 48, 5, 66, 122, 46, 174, 0, 0, 48, 5, 88, 122, 51, 5, 46, 122, 48, 5, 66, 122, 49, 5, 102, 122, 48, 5, 66, 122, 46, 176, 0, 0, 49, 5, 102, 122, 51, 5, 46, 122, 46, 174, 0, 0, 49, 5, 68, 122, 51, 5, 8, 122, 48, 5, 22, 122, 48, 5, 92, 122, 49, 5, 68, 122, 51, 5, 8, 122, 48, 5, 88, 122, 3, 5, 68, 122, 3, 5, 32, 122, 49, 5, 68, 122, 51, 5, 42, 122, 48, 5, 58, 122, 49, 5, 68, 122, 48, 5, 88, 122, 48, 5, 22, 122, 49, 5, 68, 122, 49, 5, 102, 122, 48, 5, 30, 122, 52, 5, 82, 122, 49, 5, 102, 122, 49, 5, 70, 122, 48, 5, 22, 122, 48, 5, 92, 122, 49, 5, 102, 122, 49, 5, 70, 122, 51, 5, 86, 122, 49, 5, 70, 122, 51, 5, 86, 122, 48, 5, 58, 122, 46, 174, 0, 0, 46, 5, 118, 13, 48, 5, 88, 122, 49, 5, 74, 122, 48, 5, 18, 122, 46, 174, 0, 0, 49, 5, 74, 122, 48, 5, 18, 122, 46, 174, 0, 0, 51, 5, 46, 122, 49, 5, 102, 122, 51, 5, 78, 122, 46, 5, 118, 13, 51, 5, 46, 122, 46, 174, 0, 0, 51, 5, 78, 122, 46, 5, 118, 13, 48, 5, 88, 122, 51, 5, 80, 122, 51, 5, 6, 122, 49, 5, 102, 122, 46, 5, 82, 122, 44, 5, 86, 122, 51, 5, 86, 122, 51, 5, 42, 122, 51, 5, 46, 122, 48, 5, 88, 122, 51, 5, 86, 122, 48, 5, 84, 122, 48, 5, 88, 122, 46, 5, 118, 13, 48, 5, 62, 122, 46, 174, 0, 0, 48, 5, 88, 122, 48, 5, 88, 122, 48, 5, 60, 122, 46, 176, 0, 0, 46, 5, 118, 13, 48, 5, 90, 122, 48, 5, 72, 122, 48, 5, 90, 122, 49, 5, 102, 122, 51, 5, 46, 122, 48, 5, 24, 122, 46, 174, 0, 0, 49, 5, 102, 122, 152, 5, 94, 122, 5, 174, 0, 0, 158, 5, 96, 122, 5, 174, 0, 0, 158, 5, 98, 122, 5, 174, 0, 0, 155, 5, 100, 122, 5, 174, 0, 0, 32, 5, 184, 124, 32, 5, 190, 124, 5, 5, 6, 121, 5, 5, 8, 121, 5, 5, 10, 121, 5, 5, 12, 121, 5, 5, 14, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 5, 5, 22, 121, 5, 5, 24, 121, 5, 5, 26, 121, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 32, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 5, 5, 105, 121, 5, 5, 107, 121, 5, 5, 109, 121, 5, 5, 111, 121, 5, 5, 113, 121, 5, 5, 115, 121, 5, 5, 117, 121, 5, 5, 119, 121, 5, 5, 121, 121, 5, 5, 123, 121, 5, 5, 125, 121, 5, 5, 127, 121, 5, 5, 129, 121, 5, 5, 131, 121, 5, 5, 133, 121, 5, 5, 135, 121, 5, 5, 137, 121, 5, 5, 139, 121, 5, 5, 141, 121, 5, 5, 143, 121, 5, 5, 179, 121, 5, 5, 181, 121, 5, 5, 183, 121, 5, 5, 185, 121, 5, 5, 187, 121, 5, 5, 189, 121, 5, 5, 191, 121, 5, 5, 193, 121, 5, 5, 195, 121, 5, 5, 197, 121, 5, 5, 199, 121, 5, 5, 201, 121, 5, 5, 203, 121, 5, 5, 205, 121, 5, 5, 207, 121, 5, 5, 209, 121, 5, 5, 211, 121, 5, 5, 213, 121, 5, 5, 215, 121, 5, 5, 217, 121, 5, 5, 219, 121, 5, 5, 221, 121, 5, 5, 223, 121, 5, 5, 225, 121, 5, 5, 227, 121, 5, 5, 229, 121, 5, 5, 231, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 16, 5, 0, 19, 16, 5, 0, 21, 16, 5, 0, 23, 16, 5, 0, 25, 16, 5, 0, 27, 16, 5, 0, 29, 16, 5, 0, 31, 16, 5, 0, 33, 16, 5, 0, 35, 16, 5, 0, 37, 26, 5, 0, 19, 26, 5, 0, 21, 26, 5, 0, 23, 26, 5, 0, 25, 26, 5, 0, 27, 26, 5, 0, 29, 26, 5, 0, 31, 26, 5, 0, 33, 26, 5, 0, 35, 28, 5, 0, 37, 5, 0, 0, 0, 6, 0, 0, 0, 29, 1, 0, 0, 0, 5, 0, 5, 60, 0, 128, 69, 130, 61, 0, 0, 128, 5, 0, 130, 128, 5, 0, 132, 128, 5, 0, 134, 128, 5, 0, 136, 128, 5, 0, 138, 128, 5, 0, 140, 128, 5, 0, 142, 128, 5, 0, 144, 128, 32, 0, 144, 128, 5, 0, 146, 128, 5, 0, 148, 128, 5, 0, 150, 128, 32, 0, 150, 128, 5, 0, 152, 128, 5, 0, 154, 128, 5, 0, 156, 128, 5, 0, 158, 128, 5, 0, 160, 128, 32, 0, 160, 128, 5, 0, 162, 128, 5, 0, 164, 128, 5, 0, 166, 128, 5, 0, 168, 128, 32, 0, 168, 128, 5, 0, 170, 128, 32, 0, 170, 128, 5, 0, 172, 128, 5, 0, 174, 128, 32, 0, 174, 128, 46, 0, 174, 128, 5, 0, 176, 128, 32, 0, 176, 128, 46, 0, 176, 128, 5, 0, 178, 128, 32, 0, 178, 128, 5, 0, 180, 128, 5, 0, 182, 128, 5, 0, 184, 128, 5, 0, 186, 128, 5, 0, 188, 128, 5, 0, 190, 128, 5, 0, 192, 128, 5, 0, 194, 128, 5, 0, 196, 128, 5, 0, 198, 128, 5, 0, 200, 128, 5, 0, 202, 128, 5, 0, 204, 128, 5, 0, 206, 128, 5, 0, 208, 128, 5, 0, 210, 128, 5, 0, 212, 128, 5, 0, 214, 128, 5, 0, 216, 128, 5, 0, 218, 128, 5, 0, 220, 128, 5, 0, 222, 128, 5, 0, 224, 128, 5, 2, 226, 128, 5, 35, 226, 128, 5, 68, 226, 128, 5, 101, 226, 128, 5, 134, 226, 128, 5, 167, 226, 128, 5, 200, 226, 128, 5, 233, 226, 128, 5, 12, 227, 128, 5, 45, 227, 128, 5, 78, 227, 128, 5, 111, 227, 128, 5, 144, 227, 128, 5, 177, 227, 128, 5, 210, 227, 128, 5, 243, 227, 128, 5, 22, 228, 128, 5, 55, 228, 128, 5, 88, 228, 128, 5, 121, 228, 128, 5, 154, 228, 128, 5, 187, 228, 128, 5, 220, 228, 128, 5, 253, 228, 128, 5, 32, 229, 128, 5, 65, 229, 128, 5, 98, 229, 128, 5, 131, 229, 128, 5, 164, 229, 128, 16, 164, 229, 128, 30, 164, 229, 128, 44, 164, 229, 128, 5, 197, 229, 128, 5, 230, 229, 128, 5, 9, 230, 128, 32, 9, 230, 128, 5, 42, 230, 128, 5, 75, 230, 128, 5, 108, 230, 128, 32, 108, 230, 128, 5, 141, 230, 128, 5, 174, 230, 128, 5, 207, 230, 128, 32, 207, 230, 128, 46, 207, 230, 128, 5, 240, 230, 128, 5, 19, 231, 128, 5, 52, 231, 128, 5, 85, 231, 128, 32, 85, 231, 128, 46, 85, 231, 128, 5, 118, 231, 128, 5, 151, 231, 128, 5, 184, 231, 128, 32, 184, 231, 128, 46, 184, 231, 128, 5, 217, 231, 128, 5, 250, 231, 128, 5, 29, 232, 128, 32, 29, 232, 128, 46, 29, 232, 128, 5, 62, 232, 128, 32, 62, 232, 128, 46, 62, 232, 128, 5, 95, 232, 128, 5, 128, 232, 128, 32, 128, 232, 128, 46, 128, 232, 128, 5, 161, 232, 128, 5, 194, 232, 128, 5, 227, 232, 128, 5, 6, 233, 128, 5, 39, 233, 128, 5, 72, 233, 128, 5, 105, 233, 128, 5, 138, 233, 128, 5, 171, 233, 128, 5, 204, 233, 128, 5, 237, 233, 128, 5, 16, 234, 128, 5, 49, 234, 128, 5, 82, 234, 128, 5, 115, 234, 128, 5, 148, 234, 128, 5, 181, 234, 128, 5, 214, 234, 128, 5, 247, 234, 128, 5, 26, 235, 128, 5, 59, 235, 128, 16, 59, 235, 128, 30, 59, 235, 128, 44, 59, 235, 128, 5, 92, 235, 128, 5, 125, 235, 128, 5, 158, 235, 128, 5, 191, 235, 128, 5, 224, 235, 128, 5, 3, 236, 128, 5, 36, 236, 128, 5, 69, 236, 128, 5, 102, 236, 128, 5, 135, 236, 128, 5, 168, 236, 128, 5, 201, 236, 128, 5, 234, 236, 128, 5, 13, 237, 128, 5, 46, 237, 128, 5, 79, 237, 128, 5, 112, 237, 128, 5, 145, 237, 128, 5, 178, 237, 128, 5, 211, 237, 128, 5, 244, 237, 128, 5, 23, 238, 128, 5, 56, 238, 128, 5, 89, 238, 128, 5, 122, 238, 128, 5, 155, 238, 128, 5, 188, 238, 128, 5, 221, 238, 128, 5, 254, 238, 128, 5, 33, 239, 128, 5, 66, 239, 128, 5, 99, 239, 128, 5, 132, 239, 128, 5, 165, 239, 128, 5, 198, 239, 128, 5, 231, 239, 128, 5, 10, 240, 128, 5, 43, 240, 128, 5, 76, 240, 128, 5, 109, 240, 128, 5, 142, 240, 128, 5, 175, 240, 128, 5, 208, 240, 128, 5, 241, 240, 128, 32, 241, 240, 128, 5, 20, 241, 128, 5, 53, 241, 128, 5, 86, 241, 128, 5, 119, 241, 128, 5, 152, 241, 128, 5, 185, 241, 128, 5, 218, 241, 128, 5, 251, 241, 128, 5, 30, 242, 128, 5, 63, 242, 128, 5, 96, 242, 128, 5, 129, 242, 128, 5, 162, 242, 128, 5, 195, 242, 128, 5, 228, 242, 128, 5, 7, 243, 128, 5, 40, 243, 128, 5, 73, 243, 128, 5, 106, 243, 128, 5, 139, 243, 128, 5, 172, 243, 128, 5, 205, 243, 128, 5, 238, 243, 128, 5, 17, 244, 128, 5, 50, 244, 128, 5, 83, 244, 128, 5, 116, 244, 128, 5, 149, 244, 128, 5, 182, 244, 128, 5, 215, 244, 128, 5, 248, 244, 128, 5, 27, 245, 128, 5, 60, 245, 128, 5, 93, 245, 128, 5, 126, 245, 128, 5, 159, 245, 128, 5, 192, 245, 128, 5, 225, 245, 128, 5, 4, 246, 128, 5, 37, 246, 128, 5, 70, 246, 128, 5, 103, 246, 128, 5, 136, 246, 128, 5, 169, 246, 128, 5, 202, 246, 128, 5, 235, 246, 128, 5, 14, 247, 128, 5, 47, 247, 128, 5, 80, 247, 128, 5, 113, 247, 128, 5, 146, 247, 128, 5, 179, 247, 128, 5, 212, 247, 128, 5, 245, 247, 128, 5, 24, 248, 128, 5, 57, 248, 128, 5, 90, 248, 128, 5, 123, 248, 128, 5, 156, 248, 128, 5, 189, 248, 128, 5, 222, 248, 128, 5, 255, 248, 128, 5, 34, 249, 128, 5, 67, 249, 128, 5, 100, 249, 128, 5, 133, 249, 128, 5, 166, 249, 128, 5, 199, 249, 128, 5, 232, 249, 128, 5, 11, 250, 128, 5, 44, 250, 128, 5, 77, 250, 128, 5, 110, 250, 128, 5, 143, 250, 128, 5, 176, 250, 128, 5, 209, 250, 128, 5, 242, 250, 128, 5, 21, 251, 128, 5, 54, 251, 128, 5, 87, 251, 128, 5, 120, 251, 128, 5, 153, 251, 128, 5, 186, 251, 128, 5, 219, 251, 0, 2, 2, 3, 0, 0, 4, 3, 2, 0, 18, 3, 0, 0, 0, 4, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 2, 6, 5, 0, 0, 8, 5, 128, 32, 0, 5, 0, 0, 10, 5, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 12, 5, 0, 0, 14, 5, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 16, 5, 0, 2, 18, 5, 7, 23, 18, 5, 0, 0, 20, 5, 128, 32, 0, 5, 0, 0, 22, 5, 0, 0, 24, 5, 128, 32, 0, 5, 0, 0, 26, 5, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 28, 5, 2, 0, 48, 5, 128, 32, 0, 5, 0, 0, 50, 5, 0, 0, 0, 6, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 6, 7, 0, 0, 8, 7, 2, 0, 30, 7, 0, 2, 32, 7, 7, 30, 32, 7, 0, 0, 34, 7, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 36, 7, 2, 0, 42, 7, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 44, 7, 2, 0, 54, 7, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 56, 7, 0, 0, 58, 7, 0, 2, 60, 7, 7, 142, 60, 7, 0, 0, 62, 7, 0, 0, 64, 7, 2, 0, 78, 7, 0, 2, 80, 7, 0, 9, 80, 7, 0, 0, 82, 7, 2, 0, 90, 7, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 92, 7, 0, 0, 94, 7, 0, 2, 96, 7, 7, 16, 96, 7, 0, 0, 98, 7, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 100, 7, 2, 0, 110, 7, 0, 2, 112, 7, 7, 51, 112, 7, 0, 0, 114, 7, 0, 0, 116, 7, 0, 0, 0, 8, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 6, 9, 0, 0, 8, 9, 0, 2, 10, 9, 0, 9, 10, 9, 0, 0, 12, 9, 0, 0, 14, 9, 0, 2, 16, 9, 7, 30, 16, 9, 0, 0, 18, 9, 0, 0, 20, 9, 0, 2, 22, 9, 7, 37, 22, 9, 0, 0, 24, 9, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 26, 9, 2, 0, 34, 9, 0, 2, 36, 9, 7, 37, 36, 9, 0, 0, 38, 9, 2, 0, 50, 9, 0, 2, 52, 9, 7, 35, 53, 9, 0, 0, 55, 9, 2, 0, 59, 9, 0, 2, 61, 9, 7, 14, 62, 9, 0, 0, 64, 9, 2, 0, 90, 9, 0, 2, 92, 9, 0, 9, 92, 9, 0, 0, 94, 9, 2, 0, 102, 9, 0, 2, 104, 9, 7, 51, 104, 9, 0, 0, 106, 9, 128, 32, 0, 5, 0, 0, 108, 9, 2, 0, 120, 9, 128, 32, 0, 5, 0, 0, 122, 9, 2, 0, 142, 9, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 144, 9, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 146, 9, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 148, 9, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 150, 9, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 152, 9, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 154, 9, 2, 0, 160, 9, 0, 2, 162, 9, 0, 9, 162, 9, 0, 0, 164, 9, 2, 0, 184, 9, 128, 32, 0, 5, 0, 0, 186, 9, 128, 32, 0, 5, 0, 0, 188, 9, 2, 0, 62, 10, 128, 32, 0, 5, 0, 0, 64, 10, 128, 32, 0, 5, 0, 0, 66, 10, 128, 32, 0, 5, 0, 0, 68, 10, 128, 32, 0, 5, 0, 0, 70, 10, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 72, 10, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 74, 10, 128, 32, 0, 5, 0, 0, 76, 10, 128, 32, 0, 5, 0, 0, 78, 10, 128, 32, 0, 5, 0, 0, 80, 10, 128, 32, 0, 5, 0, 0, 82, 10, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 84, 10, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 86, 10, 128, 32, 0, 5, 0, 0, 88, 10, 128, 32, 0, 5, 0, 0, 90, 10, 2, 0, 126, 10, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 128, 10, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 130, 10, 2, 0, 138, 10, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 140, 10, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 142, 10, 0, 0, 144, 10, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 146, 10, 2, 0, 150, 10, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 152, 10, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 154, 10, 2, 0, 188, 10, 128, 32, 0, 5, 0, 0, 190, 10, 128, 32, 0, 5, 0, 0, 192, 10, 2, 0, 24, 11, 0, 2, 26, 11, 7, 128, 26, 11, 0, 0, 28, 11, 2, 0, 54, 11, 0, 2, 56, 11, 0, 9, 56, 11, 0, 0, 58, 11, 2, 0, 76, 11, 128, 32, 0, 5, 0, 0, 78, 11, 2, 0, 100, 11, 0, 2, 102, 11, 7, 142, 102, 11, 0, 0, 104, 11, 2, 0, 116, 11, 0, 2, 118, 11, 7, 163, 118, 11, 0, 0, 120, 11, 0, 2, 122, 11, 7, 12, 124, 11, 0, 0, 126, 11, 0, 2, 128, 11, 7, 93, 128, 11, 0, 2, 2, 12, 0, 0, 4, 12, 128, 32, 0, 5, 0, 0, 6, 12, 0, 0, 8, 12, 0, 0, 10, 12, 128, 32, 0, 5, 0, 0, 12, 12, 128, 32, 0, 5, 0, 2, 14, 12, 0, 9, 14, 12, 0, 0, 16, 12, 0, 2, 18, 12, 0, 9, 18, 12, 0, 0, 20, 12, 0, 2, 22, 12, 7, 37, 22, 12, 0, 0, 24, 12, 2, 0, 28, 12, 0, 2, 30, 12, 7, 23, 30, 12, 0, 0, 32, 12, 2, 0, 50, 12, 0, 2, 52, 12, 7, 114, 52, 12, 0, 0, 54, 12, 0, 2, 56, 12, 7, 98, 57, 12, 0, 0, 59, 12, 2, 0, 75, 12, 0, 2, 77, 12, 0, 9, 77, 12, 0, 0, 79, 12, 0, 2, 81, 12, 7, 16, 81, 12, 0, 0, 83, 12, 128, 32, 0, 5, 0, 2, 85, 12, 7, 148, 89, 12, 0, 0, 91, 12, 2, 0, 95, 12, 0, 2, 97, 12, 7, 156, 97, 12, 0, 0, 99, 12, 128, 32, 0, 5, 0, 0, 101, 12, 128, 32, 0, 5, 0, 0, 103, 12, 128, 32, 0, 5, 0, 0, 105, 12, 128, 32, 0, 5, 0, 2, 107, 12, 7, 215, 109, 12, 0, 0, 111, 12, 128, 32, 0, 5, 0, 2, 113, 12, 7, 16, 113, 12, 0, 0, 115, 12, 128, 32, 0, 5, 0, 2, 117, 12, 7, 51, 117, 12, 0, 0, 119, 12, 128, 32, 0, 5, 0, 0, 121, 12, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 123, 12, 2, 0, 129, 12, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 131, 12, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 133, 12, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 135, 12, 128, 32, 0, 5, 0, 0, 137, 12, 128, 32, 0, 5, 0, 0, 139, 12, 128, 32, 0, 5, 0, 0, 141, 12, 128, 32, 0, 5, 0, 0, 143, 12, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 145, 12, 7, 16, 145, 12, 0, 0, 147, 12, 128, 32, 0, 5, 0, 0, 149, 12, 128, 32, 0, 5, 0, 2, 151, 12, 7, 142, 151, 12, 0, 0, 153, 12, 128, 32, 0, 5, 0, 0, 155, 12, 128, 32, 0, 5, 0, 2, 157, 12, 7, 32, 170, 12, 0, 0, 172, 12, 128, 32, 0, 5, 0, 2, 174, 12, 7, 78, 178, 12, 0, 0, 180, 12, 128, 32, 0, 5, 0, 2, 182, 12, 7, 35, 183, 12, 0, 0, 185, 12, 128, 32, 0, 5, 0, 2, 187, 12, 7, 182, 74, 13, 0, 0, 76, 13, 128, 32, 0, 5, 0, 2, 78, 13, 7, 58, 78, 13, 0, 0, 80, 13, 2, 0, 98, 13, 0, 2, 100, 13, 7, 44, 100, 13, 0, 0, 102, 13, 0, 0, 104, 13, 0, 2, 106, 13, 0, 9, 106, 13, 0, 0, 108, 13, 2, 0, 118, 13, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 120, 13, 0, 2, 132, 13, 0, 0, 134, 13, 0, 0, 136, 13, 128, 32, 0, 5, 0, 0, 138, 13, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 140, 13, 128, 32, 0, 5, 0, 0, 142, 13, 128, 32, 0, 5, 0, 0, 144, 13, 0, 0, 146, 13, 0, 2, 148, 13, 0, 9, 148, 13, 0, 0, 150, 13, 2, 0, 192, 13, 128, 32, 0, 5, 0, 0, 194, 13, 2, 0, 242, 13, 0, 2, 2, 14, 0, 0, 0, 15, 0, 0, 6, 16, 0, 0, 8, 16, 2, 0, 198, 16, 0, 2, 200, 16, 7, 44, 200, 16, 0, 0, 202, 16, 2, 0, 144, 17, 0, 2, 146, 17, 7, 12, 148, 17, 0, 0, 150, 17, 2, 0, 170, 17, 0, 2, 172, 17, 7, 177, 172, 17, 0, 0, 174, 17, 2, 0, 116, 18, 0, 2, 118, 18, 7, 245, 119, 18, 0, 0, 121, 18, 2, 0, 137, 18, 0, 0, 0, 19, 128, 16, 0, 5, 128, 21, 0, 5, 128, 26, 0, 5, 128, 31, 0, 5, 128, 36, 0, 5, 128, 41, 0, 5, 128, 46, 0, 5, 128, 51, 0, 5, 0, 0, 0, 21, 128, 16, 0, 5, 128, 21, 0, 5, 128, 26, 0, 5, 128, 31, 0, 5, 128, 36, 0, 5, 128, 41, 0, 5, 128, 46, 0, 5, 128, 51, 0, 5, 128, 5, 0, 112, 0, 0, 0, 23, 128, 16, 0, 5, 128, 21, 0, 5, 128, 26, 0, 5, 128, 31, 0, 5, 128, 36, 0, 5, 128, 41, 0, 5, 128, 46, 0, 5, 128, 51, 0, 5, 128, 5, 0, 112, 0, 0, 0, 25, 128, 16, 0, 5, 128, 21, 0, 5, 128, 26, 0, 5, 128, 31, 0, 5, 128, 36, 0, 5, 128, 41, 0, 5, 128, 46, 0, 5, 128, 51, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 0, 27, 128, 16, 0, 5, 128, 21, 0, 5, 128, 26, 0, 5, 128, 31, 0, 5, 128, 36, 0, 5, 128, 41, 0, 5, 128, 46, 0, 5, 128, 51, 0, 5, 128, 5, 0, 112, 0, 0, 0, 29, 128, 16, 0, 5, 128, 21, 0, 5, 128, 26, 0, 5, 128, 31, 0, 5, 128, 36, 0, 5, 128, 41, 0, 5, 128, 46, 0, 5, 128, 51, 0, 5, 128, 5, 0, 112, 0, 0, 0, 31, 128, 16, 0, 5, 128, 21, 0, 5, 128, 26, 0, 5, 128, 31, 0, 5, 128, 36, 0, 5, 128, 41, 0, 5, 128, 46, 0, 5, 128, 51, 0, 5, 128, 5, 0, 112, 0, 0, 0, 33, 128, 16, 0, 5, 128, 21, 0, 5, 128, 26, 0, 5, 128, 31, 0, 5, 128, 36, 0, 5, 128, 41, 0, 5, 128, 46, 0, 5, 128, 51, 0, 5, 128, 5, 0, 112, 0, 0, 0, 35, 128, 16, 0, 5, 128, 21, 0, 5, 128, 26, 0, 5, 128, 31, 0, 5, 128, 36, 0, 5, 128, 41, 0, 5, 128, 46, 0, 5, 128, 51, 0, 5, 128, 5, 0, 112, 0, 0, 0, 37, 128, 16, 0, 5, 128, 22, 0, 5, 128, 28, 0, 5, 128, 34, 0, 5, 128, 40, 0, 5, 128, 46, 0, 5, 128, 52, 0, 5, 128, 5, 0, 112, 0, 2, 2, 39, 0, 2, 2, 41, 0, 0, 0, 42, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 128, 46, 0, 112, 0, 0, 6, 43, 0, 0, 8, 43, 128, 32, 0, 5, 0, 0, 10, 43, 0, 0, 12, 43, 128, 32, 0, 5, 0, 0, 14, 43, 0, 0, 16, 43, 128, 32, 0, 5, 0, 0, 18, 43, 0, 0, 20, 43, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 22, 43, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 24, 43, 2, 0, 28, 43, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 30, 43, 0, 0, 0, 44, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 0, 0, 6, 45, 0, 0, 8, 45, 128, 32, 0, 5, 0, 0, 10, 45, 2, 0, 16, 45, 128, 32, 0, 5, 0, 0, 18, 45, 0, 0, 20, 45, 128, 32, 0, 5, 0, 0, 22, 45, 128, 32, 0, 5, 0, 0, 24, 45, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 0, 46, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 0, 0, 6, 47, 0, 0, 8, 47, 128, 32, 0, 5, 0, 0, 10, 47, 128, 32, 0, 5, 0, 0, 12, 47, 128, 32, 0, 5, 0, 0, 14, 47, 128, 32, 0, 5, 0, 0, 16, 47, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 18, 47, 128, 32, 0, 5, 0, 0, 20, 47, 128, 32, 0, 5, 0, 0, 0, 48, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 128, 46, 0, 112, 128, 5, 0, 120, 128, 32, 0, 120, 0, 0, 6, 49, 0, 0, 8, 49, 0, 0, 10, 49, 128, 32, 0, 5, 0, 0, 12, 49, 2, 0, 16, 49, 128, 32, 0, 5, 0, 0, 18, 49, 128, 32, 0, 5, 0, 0, 20, 49, 0, 0, 22, 49, 128, 32, 0, 5, 0, 0, 24, 49, 2, 0, 28, 49, 0, 0, 0, 50, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 0, 0, 6, 51, 0, 0, 8, 51, 2, 0, 12, 51, 128, 32, 0, 5, 0, 0, 14, 51, 0, 0, 16, 51, 2, 0, 20, 51, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 22, 51, 0, 0, 24, 51, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 26, 51, 0, 0, 28, 51, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 30, 51, 0, 0, 32, 51, 2, 0, 36, 51, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 38, 51, 0, 0, 40, 51, 128, 32, 0, 5, 0, 0, 42, 51, 2, 0, 48, 51, 0, 0, 0, 52, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 6, 53, 0, 0, 8, 53, 0, 0, 10, 53, 128, 32, 0, 5, 0, 0, 12, 53, 2, 0, 16, 53, 128, 32, 0, 5, 0, 0, 18, 53, 128, 32, 0, 5, 0, 0, 20, 53, 0, 0, 0, 54, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 6, 55, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 8, 55, 0, 0, 10, 55, 128, 32, 0, 5, 0, 0, 12, 55, 128, 32, 0, 5, 0, 0, 14, 55, 0, 0, 16, 55, 128, 32, 0, 5, 0, 0, 18, 55, 2, 0, 22, 55, 128, 32, 0, 5, 0, 0, 24, 55, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 26, 55, 128, 32, 0, 5, 0, 0, 0, 56, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 0, 0, 6, 57, 0, 0, 8, 57, 128, 32, 0, 5, 0, 0, 10, 57, 0, 0, 12, 57, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 14, 57, 128, 32, 0, 5, 0, 0, 16, 57, 128, 32, 0, 5, 0, 0, 18, 57, 128, 32, 0, 5, 0, 0, 20, 57, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 22, 57, 2, 0, 26, 57, 0, 0, 0, 58, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 0, 0, 6, 59, 128, 32, 0, 5, 0, 0, 8, 59, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 10, 59, 2, 0, 14, 59, 128, 32, 0, 5, 0, 0, 16, 59, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 18, 59, 128, 32, 0, 5, 0, 0, 20, 59, 0, 0, 22, 59, 128, 32, 0, 5, 0, 0, 24, 59, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 26, 59, 0, 0, 0, 60, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 0, 0, 6, 61, 128, 32, 0, 5, 0, 0, 8, 61, 0, 0, 10, 61, 128, 32, 0, 5, 0, 0, 12, 61, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 14, 61, 128, 32, 0, 5, 0, 0, 16, 61, 0, 0, 0, 62, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 0, 0, 6, 63, 0, 0, 8, 63, 0, 0, 10, 63, 128, 32, 0, 5, 0, 0, 12, 63, 128, 32, 0, 5, 0, 0, 14, 63, 128, 32, 0, 5, 0, 0, 16, 63, 128, 32, 0, 5, 0, 0, 18, 63, 128, 32, 0, 5, 0, 0, 20, 63, 128, 32, 0, 5, 0, 0, 0, 64, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 0, 0, 6, 65, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 8, 65, 128, 32, 0, 5, 0, 0, 10, 65, 0, 0, 12, 65, 128, 32, 0, 5, 0, 0, 14, 65, 128, 32, 0, 5, 0, 0, 16, 65, 128, 32, 0, 5, 0, 0, 18, 65, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 20, 65, 128, 32, 0, 5, 0, 0, 22, 65, 0, 0, 24, 65, 128, 32, 0, 5, 0, 0, 26, 65, 128, 32, 0, 5, 0, 0, 28, 65, 128, 32, 0, 5, 0, 0, 30, 65, 128, 32, 0, 5, 0, 0, 32, 65, 2, 0, 40, 65, 128, 32, 0, 5, 0, 0, 42, 65, 0, 0, 44, 65, 0, 0, 0, 66, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 0, 0, 6, 67, 128, 32, 0, 5, 0, 0, 8, 67, 2, 0, 12, 67, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 14, 67, 0, 0, 16, 67, 2, 0, 20, 67, 0, 0, 0, 68, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 0, 0, 6, 69, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 8, 69, 2, 0, 14, 69, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 16, 69, 128, 32, 0, 5, 0, 0, 18, 69, 128, 32, 0, 5, 0, 0, 20, 69, 0, 0, 22, 69, 128, 32, 0, 5, 0, 0, 24, 69, 2, 0, 30, 69, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 32, 69, 0, 0, 0, 70, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 128, 46, 0, 112, 0, 0, 6, 71, 0, 0, 8, 71, 2, 0, 30, 71, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 32, 71, 2, 0, 42, 71, 128, 32, 0, 5, 0, 0, 44, 71, 128, 32, 0, 5, 0, 0, 46, 71, 128, 32, 0, 5, 0, 0, 48, 71, 0, 0, 50, 71, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 52, 71, 128, 32, 0, 5, 0, 0, 54, 71, 0, 0, 56, 71, 128, 32, 0, 5, 0, 0, 58, 71, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 60, 71, 0, 0, 0, 72, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 0, 0, 6, 73, 0, 0, 8, 73, 128, 32, 0, 5, 0, 0, 10, 73, 128, 32, 0, 5, 0, 0, 12, 73, 2, 0, 16, 73, 128, 32, 0, 5, 0, 0, 18, 73, 128, 32, 0, 5, 0, 0, 20, 73, 128, 32, 0, 5, 0, 0, 22, 73, 0, 0, 24, 73, 128, 32, 0, 5, 0, 0, 26, 73, 0, 0, 0, 74, 128, 16, 0, 5, 128, 20, 0, 5, 128, 24, 0, 5, 128, 28, 0, 5, 128, 32, 0, 5, 128, 36, 0, 5, 128, 40, 0, 5, 128, 44, 0, 5, 128, 48, 0, 5, 128, 52, 0, 5, 128, 56, 0, 5, 0, 0, 6, 75, 0, 0, 8, 75, 128, 32, 0, 5, 0, 0, 10, 75, 128, 32, 0, 5, 0, 0, 12, 75, 0, 0, 14, 75, 128, 32, 0, 5, 0, 0, 16, 75, 0, 0, 0, 76, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 6, 77, 0, 0, 8, 77, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 10, 77, 0, 0, 12, 77, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 14, 77, 0, 0, 16, 77, 128, 32, 0, 5, 0, 0, 18, 77, 0, 0, 20, 77, 128, 32, 0, 5, 0, 0, 22, 77, 2, 0, 28, 77, 128, 32, 0, 5, 0, 0, 30, 77, 0, 0, 32, 77, 0, 0, 34, 77, 128, 32, 0, 5, 0, 0, 36, 77, 2, 0, 54, 77, 128, 32, 0, 5, 0, 0, 56, 77, 2, 0, 62, 77, 128, 32, 0, 5, 0, 0, 0, 78, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 128, 5, 0, 120, 128, 32, 0, 120, 0, 0, 6, 79, 0, 0, 8, 79, 128, 32, 0, 5, 0, 0, 10, 79, 2, 0, 14, 79, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 16, 79, 128, 32, 0, 5, 0, 0, 18, 79, 2, 0, 22, 79, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 24, 79, 2, 0, 34, 79, 0, 0, 0, 80, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 128, 5, 0, 112, 128, 5, 0, 120, 128, 32, 0, 120, 0, 0, 6, 81, 0, 0, 8, 81, 128, 32, 0, 5, 0, 0, 10, 81, 128, 32, 0, 5, 0, 0, 12, 81, 0, 0, 14, 81, 128, 32, 0, 5, 0, 0, 16, 81, 128, 32, 0, 5, 0, 0, 18, 81, 128, 32, 0, 5, 0, 0, 20, 81, 2, 0, 24, 81, 128, 32, 0, 5, 0, 0, 0, 82, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 0, 0, 6, 83, 128, 32, 0, 5, 0, 0, 8, 83, 0, 0, 10, 83, 128, 32, 0, 5, 0, 0, 12, 83, 2, 0, 20, 83, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 22, 83, 0, 0, 24, 83, 128, 32, 0, 5, 0, 0, 26, 83, 2, 0, 30, 83, 128, 32, 0, 5, 0, 0, 32, 83, 128, 32, 0, 5, 0, 0, 34, 83, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 36, 83, 2, 0, 40, 83, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 42, 83, 2, 0, 46, 83, 128, 32, 0, 5, 0, 0, 48, 83, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 50, 83, 0, 0, 0, 84, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 0, 0, 6, 85, 0, 0, 8, 85, 128, 32, 0, 5, 0, 0, 10, 85, 0, 0, 12, 85, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 14, 85, 0, 0, 16, 85, 0, 0, 18, 85, 128, 32, 0, 5, 0, 0, 20, 85, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 0, 86, 128, 16, 0, 5, 128, 20, 0, 5, 128, 24, 0, 5, 128, 28, 0, 5, 128, 32, 0, 5, 128, 36, 0, 5, 128, 40, 0, 5, 128, 44, 0, 5, 128, 48, 0, 5, 128, 52, 0, 5, 128, 56, 0, 5, 0, 0, 6, 87, 0, 0, 8, 87, 128, 32, 0, 5, 0, 0, 10, 87, 128, 32, 0, 5, 0, 0, 12, 87, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 0, 88, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 128, 52, 0, 5, 0, 0, 6, 89, 0, 0, 8, 89, 2, 0, 16, 89, 128, 32, 0, 5, 0, 0, 18, 89, 0, 0, 20, 89, 0, 0, 0, 90, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 0, 0, 6, 91, 0, 0, 8, 91, 128, 32, 0, 5, 0, 0, 10, 91, 128, 32, 0, 5, 0, 0, 12, 91, 128, 32, 0, 5, 0, 0, 14, 91, 0, 0, 16, 91, 128, 32, 0, 5, 0, 0, 0, 92, 128, 16, 0, 5, 128, 19, 0, 5, 128, 22, 0, 5, 128, 25, 0, 5, 128, 28, 0, 5, 128, 31, 0, 5, 128, 34, 0, 5, 128, 37, 0, 5, 128, 40, 0, 5, 128, 43, 0, 5, 128, 46, 0, 5, 128, 49, 0, 5, 0, 0, 6, 93, 0, 0, 8, 93, 128, 32, 0, 5, 0, 0, 10, 93, 0, 0, 12, 93, 128, 32, 0, 5, 0, 0, 14, 93, 128, 32, 0, 5, 0, 0, 16, 93, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 18, 93, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 20, 93, 128, 32, 0, 5, 0, 0, 22, 93, 128, 32, 0, 5, 0, 0, 24, 93, 128, 32, 0, 5, 0, 0, 26, 93, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 28, 93, 0, 0, 30, 93, 128, 32, 0, 5, 0, 0, 32, 93, 2, 0, 38, 93, 128, 32, 0, 5, 0, 0, 40, 93, 128, 32, 0, 5, 0, 0, 42, 93, 128, 32, 0, 5, 0, 0, 44, 93, 128, 32, 0, 5, 0, 0, 46, 93, 128, 32, 0, 5, 0, 0, 48, 93, 2, 0, 54, 93, 128, 32, 0, 5, 0, 0, 56, 93, 128, 32, 0, 5, 0, 0, 58, 93, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 60, 93, 2, 0, 64, 93, 128, 32, 0, 5, 0, 0, 66, 93, 128, 32, 0, 5, 0, 0, 68, 93, 128, 32, 0, 5, 0, 0, 70, 93, 128, 32, 0, 5, 0, 0, 72, 93, 128, 32, 0, 5, 0, 0, 74, 93, 128, 32, 0, 5, 0, 0, 76, 93, 0, 0, 78, 93, 128, 32, 0, 5, 0, 0, 80, 93, 0, 0, 82, 93, 128, 32, 0, 5, 0, 0, 84, 93, 0, 0, 86, 93, 128, 32, 0, 5, 0, 0, 88, 93, 128, 32, 0, 5, 0, 0, 90, 93, 128, 32, 0, 5, 0, 0, 92, 93, 0, 0, 94, 93, 128, 32, 0, 5, 0, 0, 96, 93, 2, 0, 102, 93, 128, 32, 0, 5, 0, 0, 104, 93, 128, 32, 0, 5, 0, 0, 106, 93, 2, 0, 126, 93, 0, 2, 2, 94, 0, 2, 4, 96, 0, 0, 6, 96, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 8, 96, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 10, 96, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 12, 96, 0, 0, 14, 96, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 16, 96, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 18, 96, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 20, 96, 128, 32, 0, 5, 0, 0, 22, 96, 128, 32, 0, 5, 0, 0, 24, 96, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 26, 96, 128, 32, 0, 5, 0, 0, 28, 96, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 30, 96, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 32, 96, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 34, 96, 128, 32, 0, 5, 0, 0, 36, 96, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 38, 96, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 40, 96, 0, 0, 42, 96, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 44, 96, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 46, 96, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 48, 96, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 50, 96, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 52, 96, 0, 0, 54, 96, 128, 32, 0, 5, 0, 0, 56, 96, 128, 32, 0, 5, 0, 0, 58, 96, 128, 32, 0, 5, 0, 0, 60, 96, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 62, 96, 0, 0, 64, 96, 0, 0, 66, 96, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 68, 96, 128, 32, 0, 5, 0, 0, 70, 96, 128, 32, 0, 5, 0, 0, 72, 96, 128, 32, 0, 5, 0, 0, 74, 96, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 76, 96, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 78, 96, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 80, 96, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 82, 96, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 84, 96, 0, 0, 86, 96, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 88, 96, 0, 0, 90, 96, 128, 32, 0, 5, 0, 0, 92, 96, 128, 32, 0, 5, 0, 0, 94, 96, 128, 32, 0, 5, 0, 2, 96, 96, 0, 0, 98, 96, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 100, 96, 128, 32, 0, 5, 0, 0, 102, 96, 128, 32, 0, 5, 0, 0, 104, 96, 128, 32, 0, 5, 0, 0, 106, 96, 128, 32, 0, 5, 0, 0, 108, 96, 128, 32, 0, 5, 0, 0, 110, 96, 128, 32, 0, 5, 0, 0, 112, 96, 128, 32, 0, 5, 0, 0, 114, 96, 128, 32, 0, 5, 0, 0, 116, 96, 128, 32, 0, 5, 0, 0, 118, 96, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 120, 96, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 122, 96, 128, 32, 0, 5, 0, 0, 124, 96, 128, 32, 0, 5, 0, 0, 126, 96, 128, 32, 0, 5, 0, 0, 128, 96, 128, 32, 0, 5, 0, 0, 130, 96, 128, 32, 0, 5, 0, 0, 132, 96, 128, 32, 0, 5, 0, 0, 134, 96, 128, 32, 0, 5, 0, 0, 136, 96, 128, 32, 0, 5, 0, 0, 138, 96, 128, 32, 0, 5, 0, 0, 140, 96, 128, 32, 0, 5, 0, 0, 142, 96, 128, 32, 0, 5, 0, 0, 144, 96, 128, 32, 0, 5, 0, 0, 146, 96, 128, 32, 0, 5, 0, 0, 148, 96, 128, 32, 0, 5, 0, 0, 150, 96, 128, 32, 0, 5, 0, 0, 152, 96, 128, 32, 0, 5, 0, 0, 154, 96, 128, 32, 0, 5, 0, 0, 156, 96, 128, 32, 0, 5, 0, 0, 158, 96, 128, 32, 0, 5, 0, 0, 160, 96, 128, 32, 0, 5, 0, 0, 162, 96, 128, 32, 0, 5, 0, 0, 164, 96, 128, 32, 0, 5, 0, 0, 166, 96, 128, 32, 0, 5, 0, 0, 168, 96, 128, 32, 0, 5, 0, 0, 170, 96, 128, 32, 0, 5, 0, 0, 172, 96, 128, 32, 0, 5, 0, 0, 174, 96, 128, 32, 0, 5, 0, 0, 176, 96, 128, 32, 0, 5, 0, 0, 178, 96, 128, 32, 0, 5, 0, 0, 180, 96, 128, 32, 0, 5, 0, 0, 182, 96, 128, 32, 0, 5, 0, 0, 184, 96, 128, 32, 0, 5, 0, 0, 186, 96, 128, 32, 0, 5, 0, 0, 188, 96, 128, 32, 0, 5, 0, 0, 190, 96, 128, 32, 0, 5, 0, 0, 192, 96, 128, 32, 0, 5, 0, 0, 194, 96, 128, 32, 0, 5, 0, 0, 196, 96, 128, 32, 0, 5, 0, 0, 198, 96, 128, 32, 0, 5, 0, 0, 200, 96, 128, 32, 0, 5, 0, 0, 202, 96, 128, 32, 0, 5, 0, 0, 204, 96, 128, 32, 0, 5, 0, 0, 206, 96, 128, 32, 0, 5, 0, 0, 208, 96, 128, 32, 0, 5, 0, 0, 210, 96, 128, 32, 0, 5, 0, 0, 212, 96, 128, 32, 0, 5, 0, 0, 214, 96, 128, 32, 0, 5, 0, 0, 216, 96, 128, 32, 0, 5, 0, 2, 4, 97, 0, 0, 6, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 8, 97, 128, 32, 0, 5, 0, 0, 10, 97, 128, 32, 0, 5, 0, 0, 12, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 14, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 16, 97, 128, 32, 0, 5, 128, 46, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 18, 97, 128, 32, 0, 5, 0, 0, 20, 97, 128, 32, 0, 5, 0, 0, 22, 97, 128, 32, 0, 5, 0, 0, 24, 97, 128, 32, 0, 5, 0, 0, 26, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 28, 97, 128, 32, 0, 5, 0, 0, 30, 97, 128, 32, 0, 5, 0, 0, 32, 97, 128, 32, 0, 5, 0, 0, 34, 97, 128, 32, 0, 5, 0, 0, 36, 97, 128, 32, 0, 5, 0, 0, 38, 97, 128, 32, 0, 5, 0, 0, 40, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 42, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 44, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 46, 97, 128, 32, 0, 5, 0, 9, 46, 97, 128, 32, 0, 5, 0, 0, 48, 97, 128, 32, 0, 5, 0, 0, 50, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 52, 97, 128, 32, 0, 5, 0, 0, 54, 97, 128, 32, 0, 5, 0, 2, 56, 97, 128, 32, 0, 5, 0, 9, 56, 97, 128, 32, 0, 5, 0, 0, 58, 97, 128, 32, 0, 5, 0, 0, 60, 97, 128, 32, 0, 5, 0, 0, 62, 97, 128, 32, 0, 5, 0, 0, 64, 97, 128, 32, 0, 5, 0, 0, 66, 97, 128, 32, 0, 5, 0, 0, 68, 97, 128, 32, 0, 5, 0, 0, 70, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 72, 97, 128, 32, 0, 5, 0, 0, 74, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 76, 97, 128, 32, 0, 5, 0, 0, 78, 97, 128, 32, 0, 5, 0, 0, 80, 97, 128, 32, 0, 5, 0, 0, 82, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 84, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 86, 97, 128, 32, 0, 5, 0, 0, 88, 97, 128, 32, 0, 5, 0, 0, 90, 97, 128, 32, 0, 5, 0, 0, 92, 97, 128, 32, 0, 5, 0, 0, 94, 97, 128, 32, 0, 5, 0, 2, 96, 97, 128, 32, 0, 5, 0, 9, 96, 97, 128, 32, 0, 5, 0, 0, 98, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 100, 97, 0, 0, 102, 97, 128, 32, 0, 5, 0, 2, 104, 97, 128, 32, 0, 5, 0, 9, 104, 97, 128, 32, 0, 5, 0, 16, 104, 97, 128, 32, 0, 5, 0, 0, 106, 97, 128, 32, 0, 5, 0, 0, 108, 97, 128, 32, 0, 5, 0, 0, 110, 97, 128, 32, 0, 5, 0, 0, 112, 97, 128, 32, 0, 5, 0, 0, 114, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 116, 97, 128, 32, 0, 5, 0, 0, 118, 97, 128, 32, 0, 5, 0, 0, 120, 97, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 122, 97, 128, 32, 0, 5, 0, 0, 124, 97, 128, 32, 0, 5, 0, 0, 126, 97, 128, 32, 0, 5, 0, 0, 128, 97, 128, 32, 0, 5, 0, 0, 130, 97, 128, 32, 0, 5, 0, 0, 132, 97, 128, 32, 0, 5, 0, 0, 134, 97, 128, 32, 0, 5, 0, 0, 136, 97, 128, 32, 0, 5, 0, 0, 138, 97, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 140, 97, 128, 32, 0, 5, 0, 0, 142, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 144, 97, 128, 32, 0, 5, 0, 0, 146, 97, 128, 32, 0, 5, 0, 0, 148, 97, 128, 32, 0, 5, 0, 0, 150, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 152, 97, 128, 32, 0, 5, 0, 0, 154, 97, 128, 32, 0, 5, 0, 0, 156, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 158, 97, 128, 32, 0, 5, 0, 0, 160, 97, 128, 32, 0, 5, 0, 0, 162, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 164, 97, 128, 32, 0, 5, 0, 9, 164, 97, 128, 32, 0, 5, 0, 0, 166, 97, 128, 32, 0, 5, 0, 0, 168, 97, 128, 32, 0, 5, 0, 0, 170, 97, 128, 32, 0, 5, 0, 0, 172, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 174, 97, 128, 32, 0, 5, 0, 0, 176, 97, 128, 32, 0, 5, 0, 2, 178, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 9, 178, 97, 128, 32, 0, 5, 0, 0, 180, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 182, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 184, 97, 128, 32, 0, 5, 0, 9, 184, 97, 128, 32, 0, 5, 0, 0, 186, 97, 128, 32, 0, 5, 0, 0, 188, 97, 128, 32, 0, 5, 0, 2, 190, 97, 128, 32, 0, 5, 0, 9, 190, 97, 128, 32, 0, 5, 0, 16, 190, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 23, 190, 97, 128, 32, 0, 5, 0, 30, 190, 97, 128, 32, 0, 5, 0, 37, 190, 97, 128, 32, 0, 5, 0, 44, 190, 97, 128, 32, 0, 5, 0, 0, 192, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 194, 97, 128, 32, 0, 5, 0, 9, 194, 97, 128, 32, 0, 5, 0, 0, 196, 97, 128, 32, 0, 5, 0, 0, 198, 97, 128, 32, 0, 5, 0, 0, 200, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 202, 97, 128, 32, 0, 5, 0, 9, 202, 97, 128, 32, 0, 5, 0, 0, 204, 97, 128, 32, 0, 5, 0, 0, 206, 97, 128, 32, 0, 5, 0, 0, 208, 97, 128, 32, 0, 5, 0, 0, 210, 97, 128, 32, 0, 5, 0, 0, 212, 97, 128, 32, 0, 5, 0, 0, 214, 97, 128, 32, 0, 5, 0, 0, 216, 97, 128, 32, 0, 5, 0, 0, 218, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 220, 97, 128, 32, 0, 5, 0, 0, 222, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 224, 97, 128, 32, 0, 5, 0, 9, 224, 97, 0, 16, 224, 97, 0, 0, 226, 97, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 228, 97, 128, 32, 0, 5, 0, 0, 230, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 232, 97, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 234, 97, 128, 32, 0, 5, 0, 2, 236, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 9, 236, 97, 128, 32, 0, 5, 0, 0, 238, 97, 128, 32, 0, 5, 0, 0, 240, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 242, 97, 128, 32, 0, 5, 0, 9, 242, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 244, 97, 128, 32, 0, 5, 0, 2, 246, 97, 128, 32, 0, 5, 0, 9, 246, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 16, 246, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 23, 246, 97, 128, 32, 0, 5, 0, 30, 246, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 37, 246, 97, 128, 32, 0, 5, 0, 44, 246, 97, 128, 32, 0, 5, 0, 51, 246, 97, 128, 32, 0, 5, 0, 58, 246, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 65, 246, 97, 128, 32, 0, 5, 0, 72, 246, 97, 128, 32, 0, 5, 0, 79, 246, 97, 128, 32, 0, 5, 128, 46, 0, 5, 0, 86, 246, 97, 128, 32, 0, 5, 0, 93, 246, 97, 128, 32, 0, 5, 0, 0, 248, 97, 128, 32, 0, 5, 0, 0, 250, 97, 128, 32, 0, 5, 0, 0, 252, 97, 128, 32, 0, 5, 0, 2, 4, 98, 0, 0, 6, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 8, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 10, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 12, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 14, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 16, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 18, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 20, 98, 128, 32, 0, 5, 0, 0, 22, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 24, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 26, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 28, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 30, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 32, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 34, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 36, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 38, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 40, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 42, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 44, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 46, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 48, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 50, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 52, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 54, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 56, 98, 128, 32, 0, 5, 0, 0, 58, 98, 128, 32, 0, 5, 0, 0, 60, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 62, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 64, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 66, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 68, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 70, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 72, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 74, 98, 128, 32, 0, 5, 0, 0, 76, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 78, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 80, 98, 128, 32, 0, 5, 0, 0, 82, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 84, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 86, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 88, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 90, 98, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 92, 98, 128, 32, 0, 5, 0, 0, 94, 98, 128, 32, 0, 5, 0, 0, 96, 98, 128, 32, 0, 5, 0, 0, 98, 98, 128, 32, 0, 5, 0, 2, 100, 98, 0, 0, 102, 98, 128, 32, 0, 5, 0, 2, 104, 98, 0, 9, 104, 98, 0, 0, 106, 98, 128, 32, 0, 5, 0, 2, 108, 98, 7, 16, 108, 98, 0, 0, 110, 98, 128, 32, 0, 5, 0, 2, 112, 98, 7, 30, 112, 98, 0, 0, 114, 98, 128, 32, 0, 5, 0, 2, 116, 98, 7, 16, 116, 98, 0, 0, 118, 98, 128, 32, 0, 5, 0, 2, 120, 98, 7, 135, 120, 98, 0, 2, 4, 99, 0, 0, 6, 99, 128, 32, 0, 5, 0, 0, 8, 99, 128, 32, 0, 5, 0, 0, 10, 99, 128, 32, 0, 5, 0, 0, 12, 99, 128, 32, 0, 5, 0, 0, 14, 99, 128, 32, 0, 5, 0, 0, 16, 99, 128, 32, 0, 5, 0, 0, 18, 99, 128, 32, 0, 5, 0, 0, 20, 99, 128, 32, 0, 5, 0, 0, 22, 99, 128, 32, 0, 5, 0, 0, 24, 99, 128, 32, 0, 5, 0, 0, 26, 99, 128, 32, 0, 5, 0, 0, 28, 99, 128, 32, 0, 5, 0, 0, 30, 99, 128, 32, 0, 5, 0, 0, 32, 99, 128, 32, 0, 5, 0, 0, 34, 99, 128, 32, 0, 5, 0, 0, 36, 99, 128, 32, 0, 5, 0, 0, 38, 99, 128, 32, 0, 5, 0, 0, 40, 99, 128, 32, 0, 5, 0, 0, 42, 99, 128, 32, 0, 5, 0, 0, 44, 99, 128, 32, 0, 5, 0, 0, 46, 99, 128, 32, 0, 5, 0, 0, 48, 99, 128, 32, 0, 5, 0, 0, 50, 99, 128, 32, 0, 5, 0, 0, 52, 99, 128, 32, 0, 5, 0, 0, 54, 99, 128, 32, 0, 5, 0, 0, 56, 99, 128, 32, 0, 5, 0, 0, 58, 99, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 60, 99, 128, 32, 0, 5, 0, 0, 62, 99, 128, 32, 0, 5, 0, 0, 64, 99, 128, 32, 0, 5, 0, 0, 66, 99, 128, 32, 0, 5, 0, 0, 68, 99, 128, 32, 0, 5, 0, 0, 70, 99, 128, 32, 0, 5, 0, 0, 72, 99, 128, 32, 0, 5, 0, 0, 74, 99, 128, 32, 0, 5, 0, 0, 76, 99, 128, 32, 0, 5, 0, 0, 78, 99, 128, 32, 0, 5, 0, 0, 80, 99, 128, 32, 0, 5, 0, 0, 82, 99, 128, 32, 0, 5, 0, 0, 84, 99, 128, 32, 0, 5, 0, 0, 86, 99, 128, 32, 0, 5, 0, 0, 88, 99, 128, 32, 0, 5, 0, 0, 90, 99, 128, 32, 0, 5, 0, 0, 92, 99, 128, 32, 0, 5, 0, 0, 94, 99, 128, 32, 0, 5, 0, 0, 96, 99, 128, 32, 0, 5, 0, 0, 98, 99, 128, 32, 0, 5, 0, 0, 100, 99, 128, 32, 0, 5, 0, 0, 102, 99, 128, 32, 0, 5, 0, 0, 104, 99, 128, 32, 0, 5, 0, 0, 106, 99, 128, 32, 0, 5, 0, 0, 108, 99, 128, 32, 0, 5, 0, 0, 110, 99, 128, 32, 0, 5, 0, 0, 112, 99, 128, 32, 0, 5, 0, 0, 114, 99, 128, 32, 0, 5, 0, 0, 116, 99, 128, 32, 0, 5, 0, 0, 118, 99, 128, 32, 0, 5, 0, 0, 120, 99, 128, 32, 0, 5, 0, 0, 122, 99, 128, 32, 0, 5, 0, 0, 124, 99, 128, 32, 0, 5, 0, 0, 126, 99, 128, 32, 0, 5, 0, 0, 128, 99, 128, 32, 0, 5, 0, 0, 130, 99, 128, 32, 0, 5, 0, 0, 132, 99, 128, 32, 0, 5, 0, 0, 134, 99, 128, 32, 0, 5, 0, 0, 136, 99, 128, 32, 0, 5, 0, 0, 138, 99, 128, 32, 0, 5, 0, 0, 140, 99, 128, 32, 0, 5, 0, 0, 142, 99, 128, 32, 0, 5, 0, 0, 144, 99, 128, 32, 0, 5, 0, 0, 146, 99, 128, 32, 0, 5, 0, 0, 148, 99, 128, 32, 0, 5, 0, 0, 150, 99, 128, 32, 0, 5, 0, 0, 152, 99, 128, 32, 0, 5, 0, 0, 154, 99, 128, 32, 0, 5, 0, 0, 156, 99, 128, 32, 0, 5, 0, 0, 158, 99, 128, 32, 0, 5, 0, 0, 160, 99, 128, 32, 0, 5, 0, 0, 162, 99, 128, 32, 0, 5, 0, 0, 164, 99, 128, 32, 0, 5, 0, 0, 166, 99, 128, 32, 0, 5, 0, 0, 168, 99, 128, 32, 0, 5, 0, 0, 170, 99, 128, 32, 0, 5, 0, 0, 172, 99, 128, 32, 0, 5, 0, 0, 174, 99, 128, 32, 0, 5, 0, 0, 176, 99, 128, 32, 0, 5, 0, 2, 4, 100, 0, 0, 6, 100, 128, 32, 0, 5, 0, 0, 8, 100, 0, 0, 10, 100, 128, 32, 0, 5, 0, 0, 12, 100, 128, 32, 0, 5, 0, 0, 14, 100, 128, 32, 0, 5, 0, 0, 16, 100, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 18, 100, 128, 32, 0, 5, 0, 0, 20, 100, 128, 32, 0, 5, 0, 0, 22, 100, 128, 32, 0, 5, 0, 0, 24, 100, 128, 32, 0, 5, 0, 0, 26, 100, 128, 32, 0, 5, 0, 0, 28, 100, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 30, 100, 128, 32, 0, 5, 0, 0, 32, 100, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 34, 100, 128, 32, 0, 5, 0, 0, 36, 100, 128, 32, 0, 5, 0, 0, 38, 100, 128, 32, 0, 5, 0, 0, 40, 100, 128, 32, 0, 5, 0, 0, 42, 100, 128, 32, 0, 5, 0, 0, 44, 100, 128, 32, 0, 5, 0, 0, 46, 100, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 48, 100, 128, 32, 0, 5, 0, 0, 50, 100, 0, 0, 52, 100, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 54, 100, 128, 32, 0, 5, 0, 0, 56, 100, 128, 32, 0, 5, 0, 0, 58, 100, 128, 32, 0, 5, 0, 0, 60, 100, 128, 32, 0, 5, 0, 0, 62, 100, 128, 32, 0, 5, 0, 0, 64, 100, 128, 32, 0, 5, 0, 0, 66, 100, 128, 32, 0, 5, 0, 0, 68, 100, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 70, 100, 128, 32, 0, 5, 0, 0, 72, 100, 128, 32, 0, 5, 0, 0, 74, 100, 128, 32, 0, 5, 0, 0, 76, 100, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 78, 100, 128, 32, 0, 5, 0, 0, 80, 100, 128, 32, 0, 5, 0, 0, 82, 100, 128, 32, 0, 5, 0, 0, 84, 100, 128, 32, 0, 5, 0, 0, 86, 100, 0, 2, 4, 101, 0, 0, 6, 101, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 8, 101, 128, 32, 0, 5, 0, 0, 10, 101, 128, 32, 0, 5, 0, 0, 12, 101, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 14, 101, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 16, 101, 128, 32, 0, 5, 0, 0, 18, 101, 2, 0, 24, 101, 128, 32, 0, 5, 0, 0, 26, 101, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 28, 101, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 30, 101, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 32, 101, 128, 32, 0, 5, 0, 0, 34, 101, 0, 0, 36, 101, 128, 32, 0, 5, 0, 0, 38, 101, 128, 32, 0, 5, 0, 0, 40, 101, 128, 32, 0, 5, 0, 0, 42, 101, 0, 0, 44, 101, 128, 32, 0, 5, 0, 0, 46, 101, 0, 0, 48, 101, 128, 32, 0, 5, 0, 2, 50, 101, 7, 156, 50, 101, 0, 2, 51, 101, 7, 184, 51, 101, 0, 2, 4, 102, 0, 0, 6, 102, 128, 32, 0, 5, 128, 46, 0, 5, 128, 5, 0, 112, 0, 0, 8, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 10, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 12, 102, 0, 0, 14, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 16, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 18, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 20, 102, 7, 16, 20, 102, 0, 0, 22, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 2, 24, 102, 7, 16, 24, 102, 0, 0, 26, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 28, 102, 128, 32, 0, 5, 0, 0, 30, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 32, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 34, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 36, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 2, 38, 102, 7, 79, 38, 102, 0, 0, 40, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 42, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 44, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 46, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 48, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 50, 102, 2, 0, 54, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 2, 56, 102, 7, 16, 56, 102, 0, 0, 58, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 60, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 62, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 64, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 66, 102, 0, 0, 68, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 2, 70, 102, 7, 23, 70, 102, 0, 0, 72, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 74, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 76, 102, 2, 0, 80, 102, 0, 2, 82, 102, 7, 37, 82, 102, 0, 0, 84, 102, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 86, 102, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 88, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 90, 102, 2, 0, 96, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 98, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 100, 102, 0, 0, 102, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 104, 102, 0, 0, 106, 102, 0, 2, 108, 102, 7, 16, 108, 102, 0, 0, 110, 102, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 112, 102, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 114, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 116, 102, 2, 0, 128, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 130, 102, 0, 2, 132, 102, 7, 51, 132, 102, 0, 0, 134, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 136, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 138, 102, 2, 0, 142, 102, 0, 2, 144, 102, 7, 37, 144, 102, 0, 0, 146, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 148, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 150, 102, 2, 0, 158, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 160, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 162, 102, 2, 0, 166, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 168, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 170, 102, 0, 2, 172, 102, 7, 37, 172, 102, 0, 0, 174, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 2, 176, 102, 128, 32, 0, 5, 0, 9, 176, 102, 7, 30, 176, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 37, 176, 102, 7, 51, 176, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 58, 176, 102, 7, 72, 176, 102, 128, 32, 0, 5, 0, 0, 178, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 2, 180, 102, 7, 30, 180, 102, 0, 0, 182, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 184, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 186, 102, 0, 0, 188, 102, 0, 2, 190, 102, 7, 16, 190, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 23, 190, 102, 0, 30, 190, 102, 0, 0, 192, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 2, 194, 102, 7, 16, 194, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 23, 194, 102, 0, 30, 194, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 37, 194, 102, 7, 79, 194, 102, 0, 0, 196, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 2, 198, 102, 7, 44, 198, 102, 0, 0, 200, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 128, 5, 0, 112, 0, 2, 202, 102, 7, 16, 202, 102, 0, 0, 204, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 206, 102, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 2, 208, 102, 0, 9, 208, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 210, 102, 0, 2, 212, 102, 7, 30, 212, 102, 0, 0, 214, 102, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 216, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 218, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 2, 220, 102, 7, 16, 220, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 222, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 224, 102, 0, 2, 226, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 9, 226, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 228, 102, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 230, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 232, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 234, 102, 0, 9, 234, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 16, 234, 102, 7, 44, 234, 102, 0, 0, 236, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 238, 102, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 240, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 242, 102, 2, 0, 246, 102, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 2, 248, 102, 7, 58, 248, 102, 0, 0, 250, 102, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 252, 102, 0, 9, 252, 102, 0, 2, 4, 103, 0, 9, 4, 103, 0, 0, 6, 103, 128, 5, 0, 112, 0, 0, 8, 103, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 10, 103, 0, 0, 12, 103, 128, 5, 0, 112, 0, 2, 14, 103, 7, 30, 14, 103, 0, 0, 16, 103, 128, 5, 0, 112, 0, 2, 18, 103, 7, 44, 18, 103, 0, 0, 20, 103, 128, 32, 0, 5, 0, 0, 22, 103, 0, 0, 24, 103, 128, 5, 0, 112, 0, 2, 26, 103, 7, 114, 26, 103, 0, 2, 27, 103, 7, 177, 27, 103, 0, 2, 28, 103, 0, 0, 30, 103, 0, 0, 32, 103, 2, 0, 128, 103, 0, 2, 130, 103, 7, 86, 130, 103, 0, 0, 132, 103, 128, 5, 0, 112, 0, 0, 134, 103, 128, 5, 0, 112, 0, 0, 136, 103, 0, 0, 138, 103, 128, 5, 0, 112, 0, 2, 140, 103, 7, 107, 140, 103, 0, 2, 141, 103, 7, 147, 142, 103, 0, 2, 143, 103, 7, 125, 155, 103, 0, 2, 4, 104, 0, 0, 6, 104, 0, 0, 8, 104, 2, 0, 148, 104, 128, 32, 0, 5, 0, 0, 150, 104, 2, 0, 154, 104, 128, 32, 0, 5, 0, 0, 156, 104, 2, 0, 208, 104, 0, 2, 4, 105, 0, 0, 6, 105, 0, 0, 8, 105, 2, 0, 66, 105, 128, 32, 0, 5, 0, 0, 68, 105, 2, 0, 132, 105, 128, 32, 0, 5, 0, 0, 134, 105, 0, 0, 136, 105, 0, 2, 4, 106, 0, 0, 6, 106, 0, 0, 8, 106, 2, 0, 118, 106, 0, 2, 4, 107, 0, 0, 6, 107, 0, 0, 8, 107, 2, 0, 142, 107, 0, 2, 4, 108, 0, 0, 6, 108, 0, 0, 8, 108, 2, 0, 138, 108, 0, 2, 4, 109, 0, 0, 6, 109, 0, 0, 8, 109, 2, 0, 104, 109, 0, 2, 4, 110, 0, 0, 6, 110, 0, 0, 8, 110, 2, 0, 152, 110, 0, 2, 4, 111, 0, 0, 6, 111, 0, 0, 8, 111, 2, 0, 152, 111, 0, 2, 4, 112, 0, 0, 6, 112, 0, 0, 8, 112, 2, 0, 40, 112, 128, 32, 0, 5, 0, 0, 42, 112, 2, 0, 68, 112, 128, 32, 0, 5, 0, 0, 70, 112, 2, 0, 78, 112, 128, 32, 0, 5, 0, 0, 80, 112, 2, 0, 90, 112, 128, 32, 0, 5, 0, 0, 92, 112, 128, 32, 0, 5, 0, 0, 94, 112, 128, 32, 0, 5, 0, 0, 96, 112, 128, 32, 0, 5, 0, 0, 98, 112, 2, 0, 108, 112, 128, 32, 0, 5, 0, 0, 110, 112, 128, 32, 0, 5, 0, 0, 112, 112, 2, 0, 152, 112, 128, 32, 0, 5, 0, 2, 4, 113, 0, 0, 6, 113, 0, 0, 8, 113, 2, 0, 158, 113, 0, 2, 160, 113, 7, 189, 161, 113, 0, 2, 162, 113, 7, 28, 163, 113, 0, 2, 164, 113, 7, 217, 165, 113, 0, 2, 166, 113, 7, 112, 167, 113, 0, 2, 168, 113, 7, 254, 168, 113, 0, 2, 170, 113, 7, 231, 171, 113, 0, 2, 172, 113, 7, 112, 173, 113, 0, 2, 174, 113, 7, 147, 175, 113, 0, 2, 176, 113, 7, 7, 177, 113, 0, 2, 178, 113, 7, 224, 179, 113, 0, 2, 180, 113, 7, 12, 182, 113, 0, 2, 183, 113, 7, 217, 184, 113, 0, 2, 185, 113, 7, 16, 185, 113, 0, 0, 187, 113, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 189, 113, 128, 5, 0, 112, 0, 0, 191, 113, 128, 5, 0, 112, 0, 2, 193, 113, 7, 56, 194, 113, 0, 0, 196, 113, 128, 5, 0, 112, 0, 0, 198, 113, 128, 5, 0, 112, 0, 2, 200, 113, 7, 44, 200, 113, 0, 2, 201, 113, 7, 189, 202, 113, 0, 2, 203, 113, 7, 126, 204, 113, 0, 2, 205, 113, 7, 140, 206, 113, 0, 2, 207, 113, 7, 182, 208, 113, 0, 2, 209, 113, 7, 140, 210, 113, 0, 2, 211, 113, 7, 30, 211, 113, 0, 0, 213, 113, 128, 5, 0, 112, 0, 2, 215, 113, 0, 9, 215, 113, 0, 0, 217, 113, 128, 5, 0, 112, 0, 2, 219, 113, 7, 79, 219, 113, 0, 0, 221, 113, 128, 5, 0, 112, 0, 2, 223, 113, 7, 121, 223, 113, 0, 2, 224, 113, 7, 254, 224, 113, 0, 0, 226, 113, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 228, 113, 7, 149, 228, 113, 0, 2, 229, 113, 7, 105, 230, 113, 0, 2, 231, 113, 0, 0, 233, 113, 128, 32, 0, 5, 0, 2, 235, 113, 7, 37, 235, 113, 0, 0, 237, 113, 128, 32, 0, 5, 0, 2, 239, 113, 7, 107, 239, 113, 0, 0, 241, 113, 128, 32, 0, 5, 0, 2, 243, 113, 7, 156, 243, 113, 0, 2, 244, 113, 7, 231, 245, 113, 0, 2, 246, 113, 7, 56, 247, 113, 0, 2, 248, 113, 7, 168, 249, 113, 0, 2, 4, 114, 0, 0, 6, 114, 0, 0, 8, 114, 2, 0, 132, 114, 0, 2, 4, 115, 0, 0, 6, 115, 0, 0, 8, 115, 2, 0, 54, 115, 128, 32, 0, 5, 0, 0, 56, 115, 2, 0, 70, 115, 128, 32, 0, 5, 0, 0, 72, 115, 2, 0, 84, 115, 128, 32, 0, 5, 0, 0, 86, 115, 2, 0, 130, 115, 0, 2, 132, 115, 7, 217, 133, 115, 0, 2, 4, 116, 0, 0, 6, 116, 0, 0, 8, 116, 2, 0, 102, 116, 128, 5, 0, 112, 0, 0, 104, 116, 2, 0, 118, 116, 128, 5, 0, 112, 0, 0, 120, 116, 128, 5, 0, 112, 0, 0, 122, 116, 128, 5, 0, 112, 0, 0, 124, 116, 2, 0, 146, 116, 128, 32, 0, 5, 0, 0, 148, 116, 2, 0, 192, 116, 128, 32, 0, 5, 0, 0, 194, 116, 0, 0, 196, 116, 0, 2, 198, 116, 7, 226, 198, 116, 0, 0, 200, 116, 128, 32, 0, 5, 0, 0, 202, 116, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 204, 116, 128, 32, 0, 5, 0, 0, 206, 116, 128, 32, 0, 5, 0, 2, 208, 116, 7, 114, 208, 116, 0, 2, 209, 116, 7, 86, 209, 116, 0, 0, 211, 116, 128, 32, 0, 5, 0, 0, 213, 116, 0, 0, 215, 116, 128, 32, 0, 5, 0, 2, 217, 116, 0, 9, 217, 116, 0, 0, 219, 116, 128, 32, 0, 5, 0, 2, 221, 116, 7, 65, 221, 116, 0, 0, 223, 116, 128, 32, 0, 5, 0, 2, 225, 116, 7, 16, 225, 116, 0, 0, 227, 116, 128, 32, 0, 5, 0, 0, 229, 116, 128, 32, 0, 5, 0, 2, 231, 116, 7, 16, 231, 116, 0, 0, 233, 116, 128, 32, 0, 5, 0, 2, 235, 116, 7, 30, 235, 116, 0, 0, 237, 116, 128, 32, 0, 5, 0, 0, 239, 116, 0, 0, 241, 116, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 243, 116, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 245, 116, 0, 0, 247, 116, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 249, 116, 0, 0, 251, 116, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 253, 116, 7, 30, 253, 116, 0, 2, 4, 117, 7, 196, 5, 117, 0, 2, 6, 117, 7, 154, 7, 117, 0, 2, 8, 117, 7, 112, 9, 117, 0, 2, 10, 117, 7, 58, 10, 117, 0, 0, 12, 117, 128, 32, 0, 5, 0, 2, 14, 117, 0, 9, 14, 117, 0, 0, 16, 117, 128, 32, 0, 5, 0, 2, 18, 117, 7, 177, 18, 117, 0, 0, 20, 117, 128, 32, 0, 5, 0, 0, 22, 117, 128, 32, 0, 5, 0, 2, 24, 117, 7, 65, 24, 117, 0, 2, 25, 117, 7, 142, 25, 117, 0, 2, 26, 117, 7, 149, 26, 117, 0, 2, 27, 117, 7, 142, 27, 117, 0, 2, 28, 117, 7, 128, 28, 117, 0, 2, 29, 117, 7, 198, 29, 117, 0, 2, 30, 117, 7, 163, 30, 117, 0, 2, 31, 117, 0, 0, 33, 117, 128, 32, 0, 5, 0, 0, 35, 117, 128, 32, 0, 5, 0, 0, 37, 117, 128, 32, 0, 5, 0, 0, 39, 117, 128, 32, 0, 5, 0, 0, 41, 117, 128, 32, 0, 5, 0, 0, 43, 117, 128, 32, 0, 5, 0, 0, 45, 117, 128, 32, 0, 5, 0, 2, 47, 117, 0, 9, 47, 117, 0, 0, 49, 117, 128, 32, 0, 5, 0, 0, 51, 117, 128, 32, 0, 5, 0, 0, 53, 117, 128, 32, 0, 5, 0, 0, 55, 117, 128, 32, 0, 5, 0, 0, 57, 117, 128, 32, 0, 5, 0, 0, 59, 117, 0, 0, 61, 117, 128, 32, 0, 5, 0, 2, 63, 117, 7, 37, 63, 117, 0, 0, 65, 117, 128, 32, 0, 5, 0, 0, 67, 117, 0, 0, 69, 117, 128, 32, 0, 5, 0, 0, 71, 117, 128, 32, 0, 5, 0, 0, 73, 117, 128, 32, 0, 5, 0, 2, 75, 117, 7, 23, 75, 117, 0, 2, 76, 117, 7, 254, 76, 117, 0, 2, 78, 117, 7, 233, 78, 117, 0, 2, 79, 117, 7, 145, 81, 117, 0, 0, 83, 117, 128, 32, 0, 5, 0, 2, 85, 117, 7, 191, 85, 117, 0, 0, 87, 117, 128, 32, 0, 5, 0, 2, 89, 117, 7, 184, 89, 117, 0, 0, 91, 117, 128, 32, 0, 5, 0, 2, 93, 117, 7, 163, 93, 117, 0, 2, 94, 117, 7, 254, 94, 117, 0, 2, 96, 117, 7, 126, 97, 117, 0, 2, 98, 117, 0, 0, 100, 117, 2, 0, 238, 117, 0, 2, 4, 118, 7, 247, 4, 118, 0, 2, 5, 118, 7, 198, 5, 118, 0, 0, 7, 118, 128, 32, 0, 5, 0, 2, 9, 118, 7, 142, 9, 118, 0, 0, 11, 118, 128, 32, 0, 5, 0, 2, 13, 118, 7, 65, 13, 118, 0, 0, 15, 118, 128, 32, 0, 5, 0, 2, 17, 118, 7, 51, 17, 118, 0, 2, 18, 118, 7, 44, 18, 118, 0, 0, 20, 118, 128, 32, 0, 5, 0, 2, 22, 118, 7, 135, 22, 118, 0, 0, 24, 118, 128, 32, 0, 5, 0, 2, 26, 118, 7, 65, 26, 118, 0, 0, 28, 118, 128, 32, 0, 5, 0, 2, 30, 118, 7, 156, 30, 118, 0, 0, 32, 118, 128, 32, 0, 5, 0, 2, 34, 118, 7, 86, 34, 118, 0, 0, 36, 118, 128, 32, 0, 5, 0, 2, 38, 118, 7, 231, 39, 118, 0, 2, 40, 118, 7, 238, 41, 118, 0, 2, 42, 118, 7, 28, 43, 118, 0, 0, 45, 118, 128, 32, 0, 5, 0, 2, 47, 118, 7, 44, 47, 118, 0, 0, 49, 118, 128, 32, 0, 5, 0, 2, 51, 118, 7, 58, 51, 118, 0, 2, 52, 118, 7, 164, 55, 118, 0, 2, 56, 118, 7, 254, 56, 118, 0, 2, 58, 118, 0, 0, 60, 118, 128, 32, 0, 5, 0, 0, 62, 118, 128, 32, 0, 5, 0, 0, 64, 118, 128, 32, 0, 5, 0, 0, 66, 118, 128, 32, 0, 5, 0, 0, 68, 118, 128, 32, 0, 5, 0, 0, 70, 118, 128, 32, 0, 5, 0, 0, 72, 118, 128, 32, 0, 5, 0, 0, 74, 118, 128, 32, 0, 5, 0, 0, 76, 118, 128, 32, 0, 5, 0, 0, 78, 118, 128, 32, 0, 5, 0, 0, 80, 118, 128, 32, 0, 5, 0, 0, 82, 118, 128, 32, 0, 5, 0, 0, 84, 118, 128, 32, 0, 5, 0, 0, 86, 118, 128, 32, 0, 5, 0, 0, 88, 118, 128, 32, 0, 5, 0, 0, 90, 118, 128, 32, 0, 5, 0, 0, 92, 118, 128, 32, 0, 5, 0, 0, 94, 118, 128, 32, 0, 5, 0, 0, 96, 118, 128, 32, 0, 5, 0, 0, 98, 118, 128, 32, 0, 5, 0, 0, 100, 118, 128, 32, 0, 5, 0, 0, 102, 118, 128, 32, 0, 5, 0, 0, 104, 118, 128, 32, 0, 5, 0, 0, 106, 118, 128, 32, 0, 5, 0, 0, 108, 118, 128, 32, 0, 5, 0, 0, 110, 118, 128, 32, 0, 5, 0, 0, 112, 118, 128, 32, 0, 5, 0, 0, 114, 118, 128, 32, 0, 5, 0, 0, 116, 118, 128, 32, 0, 5, 0, 0, 118, 118, 128, 32, 0, 5, 0, 0, 120, 118, 128, 32, 0, 5, 0, 0, 122, 118, 128, 32, 0, 5, 0, 0, 124, 118, 128, 32, 0, 5, 0, 0, 126, 118, 128, 32, 0, 5, 0, 0, 128, 118, 128, 32, 0, 5, 0, 0, 130, 118, 128, 32, 0, 5, 0, 0, 132, 118, 128, 32, 0, 5, 0, 0, 134, 118, 128, 32, 0, 5, 0, 0, 136, 118, 128, 32, 0, 5, 0, 0, 138, 118, 128, 32, 0, 5, 0, 0, 140, 118, 128, 32, 0, 5, 0, 0, 142, 118, 128, 32, 0, 5, 0, 0, 144, 118, 128, 32, 0, 5, 0, 0, 146, 118, 128, 32, 0, 5, 0, 0, 148, 118, 128, 32, 0, 5, 0, 0, 150, 118, 128, 32, 0, 5, 0, 0, 152, 118, 128, 32, 0, 5, 0, 0, 154, 118, 128, 32, 0, 5, 0, 0, 156, 118, 128, 32, 0, 5, 0, 0, 158, 118, 128, 32, 0, 5, 0, 0, 160, 118, 128, 32, 0, 5, 0, 0, 162, 118, 128, 32, 0, 5, 0, 0, 164, 118, 128, 32, 0, 5, 0, 0, 166, 118, 128, 32, 0, 5, 0, 0, 168, 118, 128, 32, 0, 5, 0, 0, 170, 118, 128, 32, 0, 5, 0, 0, 172, 118, 128, 32, 0, 5, 0, 0, 174, 118, 128, 32, 0, 5, 0, 0, 176, 118, 128, 32, 0, 5, 0, 0, 178, 118, 128, 32, 0, 5, 0, 0, 180, 118, 128, 32, 0, 5, 0, 0, 182, 118, 128, 32, 0, 5, 0, 0, 184, 118, 128, 32, 0, 5, 0, 0, 186, 118, 128, 32, 0, 5, 0, 0, 188, 118, 128, 32, 0, 5, 0, 0, 190, 118, 128, 32, 0, 5, 0, 0, 192, 118, 128, 32, 0, 5, 0, 0, 194, 118, 128, 32, 0, 5, 0, 0, 196, 118, 128, 32, 0, 5, 0, 0, 198, 118, 128, 32, 0, 5, 0, 0, 200, 118, 128, 32, 0, 5, 0, 0, 202, 118, 128, 32, 0, 5, 0, 0, 204, 118, 128, 32, 0, 5, 0, 0, 206, 118, 128, 32, 0, 5, 0, 0, 208, 118, 128, 32, 0, 5, 0, 0, 210, 118, 128, 32, 0, 5, 0, 0, 212, 118, 128, 32, 0, 5, 0, 0, 214, 118, 128, 32, 0, 5, 0, 0, 216, 118, 128, 32, 0, 5, 0, 0, 218, 118, 128, 32, 0, 5, 0, 0, 220, 118, 128, 32, 0, 5, 0, 0, 222, 118, 128, 32, 0, 5, 0, 0, 224, 118, 128, 32, 0, 5, 0, 0, 226, 118, 128, 32, 0, 5, 0, 0, 228, 118, 128, 32, 0, 5, 0, 0, 230, 118, 128, 32, 0, 5, 0, 2, 232, 118, 0, 9, 232, 118, 128, 32, 0, 5, 0, 16, 232, 118, 128, 32, 0, 5, 0, 23, 232, 118, 128, 32, 0, 5, 0, 30, 232, 118, 128, 32, 0, 5, 0, 37, 232, 118, 128, 32, 0, 5, 0, 44, 232, 118, 128, 32, 0, 5, 0, 51, 232, 118, 128, 32, 0, 5, 0, 58, 232, 118, 128, 32, 0, 5, 0, 65, 232, 118, 128, 32, 0, 5, 0, 72, 232, 118, 128, 32, 0, 5, 0, 79, 232, 118, 128, 32, 0, 5, 0, 86, 232, 118, 128, 32, 0, 5, 0, 93, 232, 118, 128, 32, 0, 5, 0, 100, 232, 118, 128, 32, 0, 5, 0, 107, 232, 118, 128, 32, 0, 5, 0, 114, 232, 118, 128, 32, 0, 5, 0, 121, 232, 118, 128, 32, 0, 5, 0, 128, 232, 118, 128, 32, 0, 5, 0, 135, 232, 118, 128, 32, 0, 5, 0, 142, 232, 118, 128, 32, 0, 5, 0, 149, 232, 118, 128, 32, 0, 5, 0, 156, 232, 118, 128, 32, 0, 5, 0, 163, 232, 118, 128, 32, 0, 5, 0, 170, 232, 118, 128, 32, 0, 5, 0, 177, 232, 118, 128, 32, 0, 5, 0, 184, 232, 118, 128, 32, 0, 5, 0, 191, 232, 118, 128, 32, 0, 5, 0, 198, 232, 118, 128, 32, 0, 5, 0, 205, 232, 118, 128, 32, 0, 5, 0, 212, 232, 118, 128, 32, 0, 5, 0, 219, 232, 118, 128, 32, 0, 5, 0, 226, 232, 118, 128, 32, 0, 5, 0, 233, 232, 118, 128, 32, 0, 5, 0, 240, 232, 118, 128, 32, 0, 5, 0, 247, 232, 118, 128, 32, 0, 5, 0, 254, 232, 118, 128, 32, 0, 5, 0, 2, 234, 118, 7, 125, 253, 118, 0, 2, 254, 118, 7, 184, 254, 118, 0, 2, 4, 119, 0, 0, 6, 119, 128, 5, 0, 112, 0, 0, 8, 119, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 10, 119, 128, 32, 0, 5, 128, 5, 0, 112, 0, 0, 12, 119, 128, 5, 0, 112, 128, 5, 0, 116, 128, 5, 0, 120, 128, 5, 0, 124, 0, 2, 14, 119, 7, 23, 14, 119, 0, 0, 16, 119, 128, 5, 0, 112, 128, 5, 0, 116, 128, 5, 0, 120, 128, 5, 0, 124, 0, 2, 18, 119, 0, 9, 18, 119, 0, 0, 20, 119, 128, 5, 0, 112, 0, 0, 22, 119, 128, 5, 0, 112, 128, 5, 0, 117, 128, 5, 0, 122, 0, 0, 24, 119, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 26, 119, 128, 5, 0, 112, 0, 0, 28, 119, 0, 0, 30, 119, 128, 5, 0, 112, 0, 0, 32, 119, 128, 32, 0, 5, 128, 5, 0, 112, 0, 0, 34, 119, 0, 0, 36, 119, 128, 5, 0, 112, 0, 0, 38, 119, 0, 0, 40, 119, 128, 5, 0, 112, 128, 5, 0, 115, 128, 5, 0, 118, 128, 5, 0, 121, 128, 5, 0, 124, 0, 0, 42, 119, 0, 0, 44, 119, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 46, 119, 128, 5, 0, 112, 128, 5, 0, 120, 0, 2, 48, 119, 0, 9, 48, 119, 0, 0, 50, 119, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 52, 119, 128, 32, 0, 5, 128, 5, 0, 112, 0, 0, 54, 119, 128, 5, 0, 112, 0, 2, 56, 119, 7, 100, 56, 119, 0, 0, 58, 119, 128, 5, 0, 112, 128, 5, 0, 120, 0, 2, 60, 119, 0, 0, 62, 119, 128, 32, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 64, 119, 128, 32, 0, 5, 0, 0, 66, 119, 128, 32, 0, 5, 0, 0, 68, 119, 128, 32, 0, 5, 0, 0, 70, 119, 128, 32, 0, 5, 0, 0, 72, 119, 128, 32, 0, 5, 0, 0, 74, 119, 128, 32, 0, 5, 0, 0, 76, 119, 128, 32, 0, 5, 0, 0, 78, 119, 128, 32, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 128, 5, 0, 120, 128, 32, 0, 120, 0, 0, 80, 119, 128, 32, 0, 5, 0, 0, 82, 119, 128, 32, 0, 5, 0, 0, 84, 119, 128, 32, 0, 5, 0, 0, 86, 119, 128, 32, 0, 5, 0, 0, 88, 119, 128, 32, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 90, 119, 128, 32, 0, 5, 0, 0, 92, 119, 128, 32, 0, 5, 0, 0, 94, 119, 128, 32, 0, 5, 0, 0, 96, 119, 128, 32, 0, 5, 0, 0, 98, 119, 128, 32, 0, 5, 0, 0, 100, 119, 128, 32, 0, 5, 0, 0, 102, 119, 128, 32, 0, 5, 0, 0, 104, 119, 128, 32, 0, 5, 0, 0, 106, 119, 128, 32, 0, 5, 0, 0, 108, 119, 128, 32, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 110, 119, 128, 32, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 128, 5, 0, 120, 128, 32, 0, 120, 0, 0, 112, 119, 128, 32, 0, 5, 0, 0, 114, 119, 128, 32, 0, 5, 0, 0, 116, 119, 128, 32, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 118, 119, 128, 32, 0, 5, 0, 0, 120, 119, 128, 32, 0, 5, 0, 0, 122, 119, 128, 32, 0, 5, 0, 0, 124, 119, 128, 32, 0, 5, 0, 0, 126, 119, 128, 32, 0, 5, 0, 0, 128, 119, 128, 32, 0, 5, 0, 0, 130, 119, 128, 32, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 132, 119, 128, 32, 0, 5, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 134, 119, 128, 32, 0, 5, 0, 0, 136, 119, 128, 32, 0, 5, 0, 0, 138, 119, 128, 32, 0, 5, 0, 0, 140, 119, 128, 32, 0, 5, 0, 0, 142, 119, 128, 32, 0, 5, 0, 2, 144, 119, 0, 0, 146, 119, 128, 5, 0, 112, 0, 0, 148, 119, 0, 0, 150, 119, 128, 5, 0, 112, 0, 2, 152, 119, 0, 9, 152, 119, 0, 0, 154, 119, 128, 5, 0, 112, 0, 0, 156, 119, 128, 5, 0, 112, 0, 0, 158, 119, 128, 5, 0, 112, 0, 0, 160, 119, 128, 5, 0, 112, 0, 0, 162, 119, 128, 5, 0, 112, 0, 0, 164, 119, 128, 5, 0, 112, 0, 0, 166, 119, 0, 0, 168, 119, 128, 5, 0, 112, 0, 0, 170, 119, 128, 5, 0, 112, 0, 0, 172, 119, 128, 5, 0, 112, 0, 0, 174, 119, 128, 5, 0, 112, 0, 0, 176, 119, 128, 5, 0, 112, 0, 0, 178, 119, 128, 5, 0, 112, 0, 2, 180, 119, 7, 23, 180, 119, 0, 0, 182, 119, 128, 5, 0, 112, 0, 0, 184, 119, 128, 5, 0, 112, 0, 0, 186, 119, 128, 5, 0, 112, 0, 0, 188, 119, 128, 5, 0, 112, 0, 0, 190, 119, 0, 0, 192, 119, 128, 5, 0, 112, 0, 2, 194, 119, 7, 16, 194, 119, 0, 0, 196, 119, 128, 5, 0, 112, 0, 0, 198, 119, 128, 5, 0, 112, 0, 0, 200, 119, 128, 5, 0, 112, 0, 0, 202, 119, 128, 5, 0, 112, 0, 0, 204, 119, 128, 5, 0, 112, 0, 2, 206, 119, 7, 16, 206, 119, 0, 0, 208, 119, 128, 5, 0, 112, 0, 0, 210, 119, 128, 5, 0, 112, 0, 0, 212, 119, 128, 5, 0, 112, 0, 0, 214, 119, 128, 5, 0, 112, 0, 2, 216, 119, 0, 9, 216, 119, 0, 2, 4, 120, 7, 86, 4, 120, 0, 0, 6, 120, 128, 32, 0, 5, 0, 2, 8, 120, 7, 114, 8, 120, 0, 0, 10, 120, 128, 32, 0, 5, 0, 2, 12, 120, 7, 135, 12, 120, 0, 0, 14, 120, 128, 32, 0, 5, 0, 2, 16, 120, 7, 128, 16, 120, 0, 0, 18, 120, 128, 32, 0, 5, 0, 2, 20, 120, 7, 65, 20, 120, 0, 0, 22, 120, 128, 32, 0, 5, 0, 2, 24, 120, 7, 30, 24, 120, 0, 0, 26, 120, 128, 32, 0, 5, 0, 0, 28, 120, 0, 0, 30, 120, 128, 32, 0, 5, 0, 2, 32, 120, 7, 30, 32, 120, 0, 0, 34, 120, 128, 32, 0, 5, 0, 2, 36, 120, 7, 58, 36, 120, 0, 0, 38, 120, 128, 32, 0, 5, 0, 2, 40, 120, 7, 16, 40, 120, 0, 0, 42, 120, 128, 32, 0, 5, 0, 2, 44, 120, 7, 149, 44, 120, 0, 0, 46, 120, 128, 32, 0, 5, 0, 0, 48, 120, 0, 0, 50, 120, 128, 32, 0, 5, 0, 2, 52, 120, 7, 49, 53, 120, 0, 0, 55, 120, 128, 32, 0, 5, 0, 2, 57, 120, 7, 149, 57, 120, 0, 0, 59, 120, 128, 32, 0, 5, 0, 2, 61, 120, 7, 30, 61, 120, 0, 0, 63, 120, 128, 32, 0, 5, 0, 2, 65, 120, 7, 37, 65, 120, 0, 0, 67, 120, 128, 32, 0, 5, 0, 2, 69, 120, 7, 91, 70, 120, 0, 0, 72, 120, 128, 32, 0, 5, 0, 0, 74, 120, 128, 32, 0, 5, 0, 2, 76, 120, 7, 227, 93, 120, 0, 2, 94, 120, 7, 212, 94, 120, 0, 2, 95, 120, 7, 111, 100, 120, 0, 2, 101, 120, 0, 0, 103, 120, 128, 32, 0, 5, 0, 0, 105, 120, 128, 32, 0, 5, 0, 0, 107, 120, 128, 32, 0, 5, 0, 0, 109, 120, 128, 32, 0, 5, 0, 0, 111, 120, 128, 32, 0, 5, 0, 0, 113, 120, 128, 32, 0, 5, 0, 0, 115, 120, 128, 32, 0, 5, 0, 0, 117, 120, 128, 32, 0, 5, 0, 0, 119, 120, 128, 32, 0, 5, 0, 0, 121, 120, 128, 32, 0, 5, 0, 0, 123, 120, 128, 32, 0, 5, 0, 0, 125, 120, 128, 32, 0, 5, 0, 0, 127, 120, 128, 32, 0, 5, 0, 0, 129, 120, 128, 32, 0, 5, 0, 0, 131, 120, 128, 32, 0, 5, 0, 0, 133, 120, 128, 32, 0, 5, 0, 0, 135, 120, 128, 32, 0, 5, 0, 0, 137, 120, 128, 32, 0, 5, 0, 0, 139, 120, 128, 32, 0, 5, 0, 0, 141, 120, 128, 32, 0, 5, 0, 0, 143, 120, 128, 32, 0, 5, 0, 0, 145, 120, 128, 32, 0, 5, 0, 0, 147, 120, 128, 32, 0, 5, 0, 0, 149, 120, 128, 32, 0, 5, 0, 0, 151, 120, 128, 32, 0, 5, 0, 0, 153, 120, 128, 32, 0, 5, 0, 0, 155, 120, 128, 32, 0, 5, 0, 0, 157, 120, 128, 32, 0, 5, 0, 0, 159, 120, 128, 32, 0, 5, 0, 0, 161, 120, 128, 32, 0, 5, 0, 0, 163, 120, 128, 32, 0, 5, 0, 0, 165, 120, 128, 32, 0, 5, 0, 2, 167, 120, 0, 0, 169, 120, 128, 32, 0, 5, 0, 0, 171, 120, 128, 32, 0, 5, 0, 0, 173, 120, 128, 32, 0, 5, 0, 0, 175, 120, 128, 32, 0, 5, 0, 0, 177, 120, 128, 32, 0, 5, 0, 0, 179, 120, 128, 32, 0, 5, 0, 0, 181, 120, 128, 32, 0, 5, 0, 0, 183, 120, 128, 32, 0, 5, 0, 0, 185, 120, 128, 32, 0, 5, 0, 0, 187, 120, 128, 32, 0, 5, 0, 0, 189, 120, 128, 32, 0, 5, 0, 0, 191, 120, 128, 32, 0, 5, 0, 0, 193, 120, 128, 32, 0, 5, 0, 0, 195, 120, 128, 32, 0, 5, 0, 0, 197, 120, 128, 32, 0, 5, 0, 0, 199, 120, 128, 32, 0, 5, 0, 0, 201, 120, 128, 32, 0, 5, 0, 0, 203, 120, 128, 32, 0, 5, 0, 0, 205, 120, 128, 32, 0, 5, 0, 0, 207, 120, 128, 32, 0, 5, 0, 0, 209, 120, 128, 32, 0, 5, 0, 0, 211, 120, 128, 32, 0, 5, 0, 0, 213, 120, 128, 32, 0, 5, 0, 0, 215, 120, 128, 32, 0, 5, 0, 0, 217, 120, 128, 32, 0, 5, 0, 0, 219, 120, 128, 32, 0, 5, 0, 0, 221, 120, 128, 32, 0, 5, 0, 0, 223, 120, 128, 32, 0, 5, 0, 0, 225, 120, 128, 32, 0, 5, 0, 0, 227, 120, 128, 32, 0, 5, 0, 0, 229, 120, 128, 32, 0, 5, 0, 0, 231, 120, 128, 32, 0, 5, 0, 0, 233, 120, 128, 32, 0, 5, 0, 0, 235, 120, 128, 32, 0, 5, 0, 2, 237, 120, 0, 2, 4, 121, 0, 0, 6, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 8, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 10, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 12, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 14, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 16, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 18, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 20, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 22, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 24, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 26, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 28, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 30, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 32, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 34, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 36, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 38, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 40, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 42, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 44, 121, 0, 0, 46, 121, 128, 32, 0, 5, 0, 0, 48, 121, 128, 32, 0, 5, 0, 2, 50, 121, 7, 23, 50, 121, 0, 0, 52, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 54, 121, 0, 0, 56, 121, 128, 32, 0, 5, 0, 0, 58, 121, 128, 32, 0, 5, 0, 0, 60, 121, 128, 32, 0, 5, 0, 0, 62, 121, 0, 0, 64, 121, 128, 32, 0, 5, 0, 0, 66, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 68, 121, 128, 32, 0, 5, 0, 0, 70, 121, 128, 32, 0, 5, 0, 2, 72, 121, 7, 16, 72, 121, 0, 0, 74, 121, 128, 32, 0, 5, 0, 0, 76, 121, 0, 0, 78, 121, 128, 32, 0, 5, 0, 0, 80, 121, 0, 0, 82, 121, 128, 32, 0, 5, 0, 0, 84, 121, 128, 32, 0, 5, 0, 0, 86, 121, 128, 32, 0, 5, 0, 0, 88, 121, 128, 32, 0, 5, 0, 0, 90, 121, 128, 32, 0, 5, 0, 2, 92, 121, 7, 16, 92, 121, 128, 32, 0, 5, 0, 23, 92, 121, 7, 44, 92, 121, 128, 32, 0, 5, 0, 51, 92, 121, 7, 114, 92, 121, 128, 32, 0, 5, 0, 121, 92, 121, 7, 163, 92, 121, 128, 32, 0, 5, 0, 170, 92, 121, 7, 198, 92, 121, 128, 32, 0, 5, 0, 205, 92, 121, 7, 14, 93, 121, 0, 0, 95, 121, 128, 32, 0, 5, 0, 0, 97, 121, 128, 32, 0, 5, 0, 0, 99, 121, 128, 32, 0, 5, 0, 2, 101, 121, 7, 247, 101, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 103, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 105, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 107, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 109, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 111, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 113, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 115, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 117, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 119, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 121, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 123, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 125, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 127, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 129, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 131, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 133, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 135, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 137, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 139, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 141, 121, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 143, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 2, 145, 121, 7, 93, 145, 121, 0, 0, 147, 121, 128, 32, 0, 5, 0, 0, 149, 121, 128, 32, 0, 5, 0, 0, 151, 121, 2, 0, 155, 121, 128, 32, 0, 5, 0, 2, 157, 121, 7, 51, 157, 121, 0, 0, 159, 121, 128, 32, 0, 5, 0, 0, 161, 121, 128, 32, 0, 5, 0, 0, 163, 121, 0, 0, 165, 121, 128, 32, 0, 5, 0, 2, 167, 121, 7, 58, 167, 121, 0, 0, 169, 121, 128, 32, 0, 5, 0, 0, 171, 121, 2, 0, 175, 121, 128, 32, 0, 5, 0, 2, 177, 121, 7, 198, 177, 121, 0, 0, 179, 121, 2, 0, 183, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 185, 121, 128, 32, 0, 5, 0, 0, 187, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 189, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 191, 121, 2, 0, 195, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 197, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 199, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 201, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 203, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 205, 121, 128, 32, 0, 5, 128, 46, 0, 5, 0, 0, 207, 121, 0, 0, 209, 121, 128, 32, 0, 5, 0, 0, 211, 121, 2, 0, 231, 121, 0, 2, 233, 121, 7, 23, 233, 121, 0, 0, 235, 121, 128, 32, 0, 5, 0, 0, 237, 121, 128, 32, 0, 5, 0, 2, 239, 121, 7, 16, 239, 121, 0, 0, 241, 121, 128, 32, 0, 5, 0, 0, 243, 121, 0, 0, 245, 121, 128, 32, 0, 5, 0, 2, 247, 121, 7, 30, 247, 121, 128, 32, 0, 5, 0, 37, 247, 121, 7, 58, 247, 121, 128, 32, 0, 5, 0, 65, 247, 121, 0, 72, 247, 121, 128, 32, 0, 5, 0, 79, 247, 121, 7, 100, 247, 121, 128, 32, 0, 5, 0, 107, 247, 121, 0, 114, 247, 121, 128, 32, 0, 5, 0, 121, 247, 121, 7, 240, 247, 121, 128, 32, 0, 5, 0, 247, 247, 121, 128, 32, 0, 5, 0, 254, 247, 121, 7, 173, 249, 121, 0, 2, 4, 122, 0, 0, 6, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 8, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 10, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 12, 122, 0, 0, 14, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 16, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 18, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 20, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 22, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 24, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 26, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 28, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 30, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 32, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 34, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 36, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 38, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 40, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 42, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 44, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 46, 122, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 48, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 50, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 52, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 54, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 56, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 58, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 60, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 62, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 64, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 66, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 68, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 70, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 72, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 74, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 76, 122, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 78, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 80, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 82, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 22, 0, 5, 128, 28, 0, 5, 128, 34, 0, 5, 128, 40, 0, 5, 128, 46, 0, 5, 128, 52, 0, 5, 0, 0, 84, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 86, 122, 128, 16, 0, 5, 128, 23, 0, 5, 128, 30, 0, 5, 128, 37, 0, 5, 128, 44, 0, 5, 128, 51, 0, 5, 0, 0, 88, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 90, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 92, 122, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 94, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 24, 0, 5, 128, 32, 0, 5, 128, 40, 0, 5, 128, 48, 0, 5, 0, 0, 96, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 98, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 30, 0, 5, 128, 44, 0, 5, 0, 0, 100, 122, 128, 3, 0, 5, 128, 5, 0, 5, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 0, 102, 122, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 2, 104, 122, 7, 219, 111, 122, 0, 2, 4, 123, 0, 0, 6, 123, 128, 5, 0, 112, 0, 0, 8, 123, 128, 32, 0, 5, 0, 0, 10, 123, 2, 0, 20, 123, 128, 32, 0, 5, 0, 0, 22, 123, 2, 0, 28, 123, 128, 32, 0, 5, 128, 5, 0, 112, 0, 0, 30, 123, 128, 32, 0, 5, 0, 0, 32, 123, 2, 0, 36, 123, 128, 32, 0, 5, 0, 0, 38, 123, 128, 5, 0, 112, 0, 0, 40, 123, 2, 0, 58, 123, 128, 5, 0, 112, 0, 0, 60, 123, 2, 0, 70, 123, 128, 5, 0, 112, 0, 0, 72, 123, 128, 5, 0, 112, 0, 0, 74, 123, 0, 0, 76, 123, 128, 32, 0, 5, 0, 0, 78, 123, 0, 0, 80, 123, 128, 5, 0, 112, 0, 0, 82, 123, 128, 5, 0, 112, 0, 0, 84, 123, 0, 0, 86, 123, 128, 5, 0, 112, 0, 0, 88, 123, 2, 0, 106, 123, 128, 5, 0, 112, 128, 32, 0, 112, 0, 0, 108, 123, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 110, 123, 2, 0, 116, 123, 0, 2, 118, 123, 7, 29, 150, 123, 0, 2, 151, 123, 7, 77, 152, 123, 0, 2, 153, 123, 7, 37, 153, 123, 0, 0, 155, 123, 128, 32, 0, 5, 0, 2, 157, 123, 7, 79, 157, 123, 0, 0, 159, 123, 128, 32, 0, 5, 0, 2, 161, 123, 7, 114, 161, 123, 0, 0, 163, 123, 128, 32, 0, 5, 0, 2, 165, 123, 7, 184, 165, 123, 0, 0, 167, 123, 128, 32, 0, 5, 0, 2, 169, 123, 7, 47, 171, 123, 0, 2, 172, 123, 0, 9, 172, 123, 0, 0, 174, 123, 128, 32, 0, 5, 0, 0, 176, 123, 128, 32, 0, 5, 0, 0, 178, 123, 128, 32, 0, 5, 0, 0, 180, 123, 128, 32, 0, 5, 0, 0, 182, 123, 128, 32, 0, 5, 0, 0, 184, 123, 128, 32, 0, 5, 0, 0, 186, 123, 128, 32, 0, 5, 0, 0, 188, 123, 128, 32, 0, 5, 0, 0, 190, 123, 128, 32, 0, 5, 0, 0, 192, 123, 128, 32, 0, 5, 0, 0, 194, 123, 128, 32, 0, 5, 0, 0, 196, 123, 128, 32, 0, 5, 0, 0, 198, 123, 128, 32, 0, 5, 0, 0, 200, 123, 128, 32, 0, 5, 0, 0, 202, 123, 128, 32, 0, 5, 0, 0, 204, 123, 128, 32, 0, 5, 0, 0, 206, 123, 128, 32, 0, 5, 0, 0, 208, 123, 128, 32, 0, 5, 0, 0, 210, 123, 128, 32, 0, 5, 0, 0, 212, 123, 128, 32, 0, 5, 0, 0, 214, 123, 128, 32, 0, 5, 0, 0, 216, 123, 128, 32, 0, 5, 0, 0, 218, 123, 128, 32, 0, 5, 0, 0, 220, 123, 128, 32, 0, 5, 0, 0, 222, 123, 128, 32, 0, 5, 0, 0, 224, 123, 128, 32, 0, 5, 0, 0, 226, 123, 128, 32, 0, 5, 0, 0, 228, 123, 128, 32, 0, 5, 0, 0, 230, 123, 128, 32, 0, 5, 0, 0, 232, 123, 128, 32, 0, 5, 0, 0, 234, 123, 128, 32, 0, 5, 0, 0, 236, 123, 128, 32, 0, 5, 0, 2, 238, 123, 7, 147, 239, 123, 0, 2, 240, 123, 7, 124, 242, 123, 0, 2, 243, 123, 7, 105, 244, 123, 0, 2, 245, 123, 7, 56, 246, 123, 0, 2, 247, 123, 7, 205, 247, 123, 0, 2, 248, 123, 7, 91, 249, 123, 0, 2, 250, 123, 7, 184, 250, 123, 0, 2, 251, 123, 7, 247, 251, 123, 0, 2, 252, 123, 7, 191, 252, 123, 0, 2, 4, 124, 0, 0, 6, 124, 128, 32, 0, 5, 0, 0, 8, 124, 128, 32, 0, 5, 0, 0, 10, 124, 128, 32, 0, 5, 0, 0, 12, 124, 128, 32, 0, 5, 0, 0, 14, 124, 128, 32, 0, 5, 0, 0, 16, 124, 128, 32, 0, 5, 0, 0, 18, 124, 128, 32, 0, 5, 0, 0, 20, 124, 128, 32, 0, 5, 0, 0, 22, 124, 128, 32, 0, 5, 0, 0, 24, 124, 128, 32, 0, 5, 0, 0, 26, 124, 128, 32, 0, 5, 0, 0, 28, 124, 128, 32, 0, 5, 0, 0, 30, 124, 128, 32, 0, 5, 0, 0, 32, 124, 128, 32, 0, 5, 0, 0, 34, 124, 128, 32, 0, 5, 0, 0, 36, 124, 128, 32, 0, 5, 0, 0, 38, 124, 128, 32, 0, 5, 0, 0, 40, 124, 128, 32, 0, 5, 0, 0, 42, 124, 128, 32, 0, 5, 0, 0, 44, 124, 128, 32, 0, 5, 0, 0, 46, 124, 128, 32, 0, 5, 0, 0, 48, 124, 128, 32, 0, 5, 0, 0, 50, 124, 128, 32, 0, 5, 0, 0, 52, 124, 128, 32, 0, 5, 0, 0, 54, 124, 128, 32, 0, 5, 0, 0, 56, 124, 128, 32, 0, 5, 0, 0, 58, 124, 128, 32, 0, 5, 0, 0, 60, 124, 128, 32, 0, 5, 0, 0, 62, 124, 128, 32, 0, 5, 0, 0, 64, 124, 128, 32, 0, 5, 0, 0, 66, 124, 128, 32, 0, 5, 0, 0, 68, 124, 128, 32, 0, 5, 0, 0, 70, 124, 128, 32, 0, 5, 0, 0, 72, 124, 128, 32, 0, 5, 0, 0, 74, 124, 128, 32, 0, 5, 0, 0, 76, 124, 128, 32, 0, 5, 0, 0, 78, 124, 128, 32, 0, 5, 0, 0, 80, 124, 128, 32, 0, 5, 0, 0, 82, 124, 128, 32, 0, 5, 0, 0, 84, 124, 128, 32, 0, 5, 0, 2, 86, 124, 7, 84, 87, 124, 0, 2, 88, 124, 7, 213, 91, 124, 0, 2, 92, 124, 7, 212, 92, 124, 0, 2, 93, 124, 7, 28, 94, 124, 0, 2, 95, 124, 7, 112, 96, 124, 0, 2, 97, 124, 7, 177, 97, 124, 0, 2, 98, 124, 7, 219, 98, 124, 0, 2, 99, 124, 7, 209, 104, 124, 0, 2, 105, 124, 7, 103, 114, 124, 0, 2, 115, 124, 7, 133, 116, 124, 0, 2, 117, 124, 7, 205, 117, 124, 0, 2, 118, 124, 7, 205, 118, 124, 0, 2, 119, 124, 7, 63, 120, 124, 0, 0, 122, 124, 128, 5, 0, 112, 0, 2, 124, 124, 7, 44, 124, 124, 0, 2, 125, 124, 7, 93, 125, 124, 0, 0, 127, 124, 128, 32, 0, 5, 0, 2, 129, 124, 7, 51, 129, 124, 0, 2, 130, 124, 0, 0, 132, 124, 128, 32, 0, 5, 0, 0, 134, 124, 128, 32, 0, 5, 0, 2, 136, 124, 0, 9, 136, 124, 0, 0, 138, 124, 128, 32, 0, 5, 0, 2, 140, 124, 7, 23, 140, 124, 0, 0, 142, 124, 128, 32, 0, 5, 0, 0, 144, 124, 128, 32, 0, 5, 0, 0, 146, 124, 128, 32, 0, 5, 0, 0, 148, 124, 128, 32, 0, 5, 0, 0, 150, 124, 128, 32, 0, 5, 0, 2, 152, 124, 7, 37, 152, 124, 0, 0, 154, 124, 128, 32, 0, 5, 0, 2, 156, 124, 0, 2, 157, 124, 7, 149, 157, 124, 0, 2, 158, 124, 7, 156, 158, 124, 0, 2, 159, 124, 7, 156, 159, 124, 0, 2, 160, 124, 7, 135, 160, 124, 0, 2, 161, 124, 7, 128, 161, 124, 0, 2, 162, 124, 7, 51, 162, 124, 128, 32, 0, 5, 0, 58, 162, 124, 7, 72, 162, 124, 128, 32, 0, 5, 0, 79, 162, 124, 7, 156, 162, 124, 0, 2, 163, 124, 7, 51, 163, 124, 0, 0, 165, 124, 128, 5, 0, 112, 0, 2, 167, 124, 7, 191, 167, 124, 0, 2, 168, 124, 0, 0, 170, 124, 128, 32, 0, 5, 0, 0, 172, 124, 128, 32, 0, 5, 0, 0, 174, 124, 0, 0, 176, 124, 128, 32, 0, 5, 0, 2, 178, 124, 7, 44, 178, 124, 0, 0, 180, 124, 128, 32, 0, 5, 128, 5, 0, 112, 0, 0, 182, 124, 0, 0, 184, 124, 128, 32, 0, 5, 128, 5, 0, 112, 0, 0, 186, 124, 0, 0, 188, 124, 128, 32, 0, 5, 128, 5, 0, 112, 0, 0, 190, 124, 128, 32, 0, 5, 0, 0, 192, 124, 0, 0, 194, 124, 128, 32, 0, 5, 128, 5, 0, 112, 0, 2, 196, 124, 7, 121, 196, 124, 0, 0, 198, 124, 128, 5, 0, 112, 0, 2, 200, 124, 7, 16, 200, 124, 0, 2, 201, 124, 7, 56, 202, 124, 0, 2, 203, 124, 7, 149, 203, 124, 0, 2, 204, 124, 7, 212, 204, 124, 0, 2, 205, 124, 7, 56, 206, 124, 0, 2, 207, 124, 7, 208, 237, 124, 0, 2, 4, 125, 7, 133, 33, 125, 0, 2, 34, 125, 0, 0, 36, 125, 128, 5, 0, 112, 0, 0, 38, 125, 128, 5, 0, 112, 0, 0, 40, 125, 128, 5, 0, 112, 0, 0, 42, 125, 128, 5, 0, 112, 0, 0, 44, 125, 128, 5, 0, 112, 0, 0, 46, 125, 128, 5, 0, 112, 0, 0, 48, 125, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 50, 125, 128, 5, 0, 112, 0, 0, 52, 125, 128, 5, 0, 112, 0, 0, 54, 125, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 56, 125, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 58, 125, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 60, 125, 128, 5, 0, 112, 0, 0, 62, 125, 128, 5, 0, 112, 0, 0, 64, 125, 128, 5, 0, 112, 0, 0, 66, 125, 128, 5, 0, 112, 128, 5, 0, 117, 128, 5, 0, 122, 0, 0, 68, 125, 128, 5, 0, 112, 0, 0, 70, 125, 128, 5, 0, 112, 0, 0, 72, 125, 128, 5, 0, 112, 0, 0, 74, 125, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 76, 125, 128, 5, 0, 112, 128, 5, 0, 120, 0, 0, 78, 125, 128, 5, 0, 112, 0, 0, 80, 125, 128, 5, 0, 112, 0, 2, 82, 125, 7, 23, 82, 125, 0, 2, 83, 125, 7, 19, 99, 125, 0, 2, 100, 125, 2, 112, 154, 125, 0, 2, 155, 125, 7, 234, 165, 125, 0, 2, 166, 125, 7, 244, 178, 125, 0, 2, 2, 126, 0, 2, 2, 251, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 4, 2, 251, 2, 6, 2, 251, 128, 38, 0, 5, 0, 8, 2, 251, 2, 10, 2, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 12, 2, 251, 2, 32, 2, 251, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 34, 2, 251, 128, 27, 0, 5, 128, 38, 0, 5, 0, 36, 2, 251, 128, 27, 0, 5, 128, 38, 0, 5, 0, 38, 2, 251, 2, 42, 2, 251, 128, 16, 0, 5, 0, 44, 2, 251, 2, 132, 2, 251, 128, 38, 0, 5, 0, 134, 2, 251, 2, 60, 4, 251, 128, 16, 0, 5, 0, 62, 4, 251, 2, 82, 4, 251, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 84, 4, 251, 2, 200, 4, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 202, 4, 251, 2, 252, 4, 251, 128, 16, 0, 5, 0, 254, 4, 251, 2, 26, 6, 251, 128, 16, 0, 5, 128, 38, 0, 5, 128, 5, 0, 112, 0, 28, 6, 251, 128, 16, 0, 5, 0, 30, 6, 251, 128, 16, 0, 5, 0, 32, 6, 251, 2, 48, 6, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 50, 6, 251, 2, 248, 7, 251, 128, 16, 0, 5, 0, 250, 7, 251, 2, 74, 8, 251, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 128, 49, 0, 5, 0, 76, 8, 251, 2, 96, 8, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 98, 8, 251, 2, 232, 8, 251, 128, 16, 0, 5, 0, 234, 8, 251, 2, 250, 8, 251, 128, 49, 0, 5, 0, 252, 8, 251, 2, 68, 10, 251, 128, 16, 0, 5, 128, 38, 0, 5, 0, 70, 10, 251, 128, 16, 0, 5, 0, 72, 10, 251, 2, 180, 10, 251, 128, 16, 0, 5, 0, 182, 10, 251, 128, 49, 0, 5, 0, 184, 10, 251, 2, 40, 11, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 42, 11, 251, 2, 72, 11, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 74, 11, 251, 2, 90, 11, 251, 128, 49, 0, 5, 0, 92, 11, 251, 2, 70, 24, 251, 128, 27, 0, 5, 0, 72, 24, 251, 2, 4, 26, 251, 128, 16, 0, 5, 0, 6, 26, 251, 2, 188, 27, 251, 128, 16, 0, 5, 0, 190, 27, 251, 2, 74, 28, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 76, 28, 251, 2, 80, 28, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 82, 28, 251, 2, 104, 29, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 106, 29, 251, 2, 162, 29, 251, 128, 49, 0, 5, 0, 164, 29, 251, 2, 74, 30, 251, 128, 16, 0, 5, 0, 76, 30, 251, 2, 88, 30, 251, 128, 27, 0, 5, 0, 90, 30, 251, 2, 62, 31, 251, 128, 16, 0, 5, 0, 64, 31, 251, 2, 196, 33, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 198, 33, 251, 2, 200, 34, 251, 128, 16, 0, 5, 0, 202, 34, 251, 2, 132, 35, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 134, 35, 251, 2, 136, 35, 251, 128, 16, 0, 5, 0, 138, 35, 251, 2, 90, 36, 251, 128, 49, 0, 5, 0, 92, 36, 251, 2, 172, 37, 251, 128, 49, 0, 5, 0, 174, 37, 251, 2, 126, 39, 251, 128, 49, 0, 5, 0, 128, 39, 251, 2, 8, 42, 251, 128, 16, 0, 5, 0, 10, 42, 251, 2, 114, 42, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 116, 42, 251, 2, 206, 43, 251, 128, 16, 0, 5, 0, 208, 43, 251, 2, 244, 44, 251, 128, 16, 0, 5, 0, 246, 44, 251, 2, 234, 45, 251, 128, 16, 0, 5, 0, 236, 45, 251, 2, 82, 46, 251, 128, 16, 0, 5, 0, 84, 46, 251, 2, 118, 47, 251, 128, 16, 0, 5, 0, 120, 47, 251, 2, 134, 47, 251, 128, 27, 0, 5, 0, 136, 47, 251, 2, 180, 47, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 182, 47, 251, 2, 186, 47, 251, 128, 16, 0, 5, 0, 188, 47, 251, 2, 52, 48, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 54, 48, 251, 2, 44, 49, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 46, 49, 251, 2, 240, 49, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 242, 49, 251, 2, 26, 50, 251, 128, 27, 0, 5, 0, 28, 50, 251, 2, 178, 50, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 180, 50, 251, 2, 120, 53, 251, 128, 16, 0, 5, 0, 122, 53, 251, 2, 122, 54, 251, 128, 16, 0, 5, 0, 124, 54, 251, 2, 132, 54, 251, 128, 49, 0, 5, 0, 134, 54, 251, 2, 112, 56, 251, 128, 16, 0, 5, 0, 114, 56, 251, 2, 150, 56, 251, 128, 27, 0, 5, 0, 152, 56, 251, 2, 158, 56, 251, 128, 27, 0, 5, 128, 49, 0, 5, 0, 160, 56, 251, 2, 246, 56, 251, 128, 49, 0, 5, 0, 248, 56, 251, 2, 2, 57, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 4, 57, 251, 2, 200, 57, 251, 128, 49, 0, 5, 0, 202, 57, 251, 2, 28, 59, 251, 128, 16, 0, 5, 0, 30, 59, 251, 2, 60, 59, 251, 128, 49, 0, 5, 0, 62, 59, 251, 2, 180, 64, 251, 128, 27, 0, 5, 0, 182, 64, 251, 2, 142, 67, 251, 128, 49, 0, 5, 0, 144, 67, 251, 2, 76, 83, 251, 128, 16, 0, 5, 0, 78, 83, 251, 2, 90, 83, 251, 128, 16, 0, 5, 128, 27, 0, 5, 128, 38, 0, 5, 0, 92, 83, 251, 2, 186, 85, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 188, 85, 251, 2, 12, 86, 251, 128, 38, 0, 5, 0, 14, 86, 251, 2, 220, 99, 251, 128, 16, 0, 5, 0, 222, 99, 251, 2, 234, 99, 251, 128, 49, 0, 5, 0, 236, 99, 251, 2, 178, 100, 251, 128, 16, 0, 5, 0, 180, 100, 251, 2, 232, 100, 251, 128, 16, 0, 5, 0, 234, 100, 251, 2, 136, 101, 251, 128, 16, 0, 5, 0, 138, 101, 251, 2, 154, 101, 251, 128, 49, 0, 5, 0, 156, 101, 251, 2, 182, 101, 251, 128, 27, 0, 5, 0, 184, 101, 251, 2, 204, 102, 251, 128, 16, 0, 5, 128, 49, 0, 5, 0, 206, 102, 251, 2, 208, 102, 251, 128, 38, 0, 5, 128, 49, 0, 5, 0, 210, 102, 251, 2, 100, 106, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 102, 106, 251, 2, 108, 119, 251, 128, 16, 0, 5, 0, 110, 119, 251, 2, 146, 119, 251, 128, 49, 0, 5, 0, 148, 119, 251, 2, 232, 119, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 234, 119, 251, 2, 174, 121, 251, 128, 16, 0, 5, 0, 176, 121, 251, 2, 208, 121, 251, 128, 16, 0, 5, 0, 210, 121, 251, 2, 86, 122, 251, 128, 27, 0, 5, 0, 88, 122, 251, 2, 22, 128, 251, 128, 16, 0, 5, 0, 24, 128, 251, 2, 34, 129, 251, 128, 16, 0, 5, 128, 5, 0, 112, 128, 5, 0, 120, 0, 36, 129, 251, 2, 144, 130, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 146, 130, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 148, 130, 251, 2, 192, 131, 251, 128, 16, 0, 5, 0, 194, 131, 251, 2, 8, 135, 251, 128, 16, 0, 5, 0, 10, 135, 251, 2, 128, 135, 251, 128, 16, 0, 5, 0, 130, 135, 251, 2, 66, 147, 251, 128, 16, 0, 5, 0, 68, 147, 251, 2, 210, 147, 251, 128, 16, 0, 5, 0, 212, 147, 251, 2, 228, 147, 251, 128, 27, 0, 5, 128, 49, 0, 5, 0, 230, 147, 251, 2, 132, 148, 251, 128, 16, 0, 5, 0, 134, 148, 251, 2, 136, 148, 251, 128, 16, 0, 5, 0, 138, 148, 251, 2, 20, 149, 251, 128, 16, 0, 5, 0, 22, 149, 251, 2, 170, 153, 251, 128, 16, 0, 5, 0, 172, 153, 251, 128, 49, 0, 5, 0, 174, 153, 251, 2, 92, 154, 251, 128, 16, 0, 5, 0, 94, 154, 251, 2, 106, 154, 251, 128, 27, 0, 5, 0, 108, 154, 251, 2, 218, 154, 251, 128, 16, 0, 5, 0, 220, 154, 251, 2, 254, 159, 251, 128, 16, 0, 5, 0, 2, 160, 251, 2, 102, 160, 251, 128, 16, 0, 5, 0, 104, 160, 251, 2, 108, 161, 251, 128, 16, 0, 5, 0, 110, 161, 251, 2, 130, 161, 251, 128, 49, 0, 5, 0, 132, 161, 251, 2, 216, 161, 251, 128, 16, 0, 5, 0, 218, 161, 251, 2, 210, 164, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 212, 164, 251, 128, 16, 0, 5, 0, 214, 164, 251, 2, 122, 165, 251, 128, 16, 0, 5, 0, 124, 165, 251, 2, 88, 166, 251, 128, 16, 0, 5, 0, 90, 166, 251, 2, 8, 167, 251, 128, 49, 0, 5, 0, 10, 167, 251, 2, 156, 167, 251, 128, 27, 0, 5, 0, 158, 167, 251, 2, 252, 169, 251, 128, 16, 0, 5, 128, 5, 0, 112, 0, 254, 169, 251, 128, 16, 0, 5, 0, 2, 170, 251, 2, 76, 188, 251, 128, 16, 0, 5, 0, 78, 188, 251, 2, 100, 188, 251, 128, 49, 0, 5, 0, 102, 188, 251, 2, 162, 190, 251, 128, 16, 0, 5, 0, 164, 190, 251, 2, 156, 191, 251, 128, 16, 0, 5, 128, 49, 0, 5, 0, 158, 191, 251, 128, 16, 0, 5, 0, 160, 191, 251, 2, 182, 191, 251, 128, 16, 0, 5, 128, 49, 0, 5, 0, 184, 191, 251, 2, 156, 192, 251, 128, 49, 0, 5, 0, 158, 192, 251, 2, 254, 194, 251, 128, 49, 0, 5, 0, 2, 195, 251, 2, 210, 196, 251, 128, 49, 0, 5, 0, 212, 196, 251, 2, 60, 212, 251, 128, 16, 0, 5, 0, 62, 212, 251, 2, 230, 212, 251, 128, 16, 0, 5, 0, 232, 212, 251, 128, 16, 0, 5, 0, 234, 212, 251, 2, 52, 214, 251, 128, 16, 0, 5, 0, 54, 214, 251, 2, 254, 217, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 2, 218, 251, 2, 178, 218, 251, 128, 16, 0, 5, 0, 180, 218, 251, 2, 206, 218, 251, 128, 49, 0, 5, 0, 208, 218, 251, 2, 74, 219, 251, 128, 16, 0, 5, 0, 76, 219, 251, 2, 188, 219, 251, 128, 49, 0, 5, 0, 190, 219, 251, 2, 68, 220, 251, 128, 16, 0, 5, 0, 70, 220, 251, 2, 42, 222, 251, 128, 16, 0, 5, 0, 44, 222, 251, 128, 16, 0, 5, 0, 46, 222, 251, 2, 86, 222, 251, 128, 16, 0, 5, 128, 27, 0, 5, 128, 5, 0, 112, 0, 88, 222, 251, 2, 22, 223, 251, 128, 49, 0, 5, 0, 24, 223, 251, 2, 180, 223, 251, 128, 49, 0, 5, 0, 182, 223, 251, 2, 206, 223, 251, 128, 49, 0, 5, 0, 208, 223, 251, 2, 90, 232, 251, 128, 16, 0, 5, 0, 92, 232, 251, 2, 86, 233, 251, 128, 16, 0, 5, 128, 27, 0, 5, 128, 49, 0, 5, 128, 5, 0, 112, 0, 88, 233, 251, 2, 90, 233, 251, 128, 16, 0, 5, 128, 27, 0, 5, 128, 49, 0, 5, 0, 92, 233, 251, 2, 114, 235, 251, 128, 16, 0, 5, 128, 27, 0, 5, 0, 116, 235, 251, 2, 128, 235, 251, 128, 16, 0, 5, 0, 130, 235, 251, 2, 60, 239, 251, 128, 16, 0, 5, 128, 27, 0, 5, 128, 49, 0, 5, 0, 62, 239, 251, 2, 188, 6, 252, 128, 16, 0, 5, 0, 190, 6, 252, 2, 190, 9, 252, 128, 16, 0, 5, 0, 192, 9, 252, 2, 196, 9, 252, 128, 27, 0, 5, 128, 49, 0, 5, 0, 198, 9, 252, 2, 192, 11, 252, 128, 16, 0, 5, 0, 194, 11, 252, 2, 198, 11, 252, 128, 5, 0, 112, 0, 200, 11, 252, 2, 174, 14, 252, 128, 16, 0, 5, 0, 176, 14, 252, 2, 104, 16, 252, 128, 16, 0, 5, 0, 106, 16, 252, 2, 110, 16, 252, 128, 16, 0, 5, 0, 112, 16, 252, 2, 180, 16, 252, 128, 16, 0, 5, 0, 182, 16, 252, 2, 22, 17, 252, 128, 16, 0, 5, 0, 24, 17, 252, 2, 236, 20, 252, 128, 16, 0, 5, 0, 238, 20, 252, 2, 240, 20, 252, 128, 16, 0, 5, 0, 242, 20, 252, 2, 54, 21, 252, 128, 16, 0, 5, 0, 56, 21, 252, 2, 230, 21, 252, 128, 16, 0, 5, 128, 27, 0, 5, 0, 232, 21, 252, 128, 16, 0, 5, 0, 234, 21, 252, 128, 16, 0, 5, 0, 236, 21, 252, 2, 232, 23, 252, 128, 49, 0, 5, 0, 234, 23, 252, 2, 66, 24, 252, 128, 27, 0, 5, 0, 68, 24, 252, 2, 68, 32, 252, 128, 49, 0, 5, 0, 70, 32, 252, 2, 182, 36, 252, 128, 49, 0, 5, 0, 184, 36, 252, 2, 142, 48, 252, 128, 16, 0, 5, 128, 27, 0, 5, 0, 144, 48, 252, 128, 16, 0, 5, 0, 146, 48, 252, 2, 34, 50, 252, 128, 16, 0, 5, 0, 36, 50, 252, 2, 124, 53, 252, 128, 49, 0, 5, 0, 126, 53, 252, 2, 48, 64, 252, 128, 16, 0, 5, 0, 50, 64, 252, 128, 16, 0, 5, 128, 5, 0, 112, 0, 52, 64, 252, 2, 150, 65, 252, 128, 16, 0, 5, 0, 152, 65, 252, 2, 218, 65, 252, 128, 16, 0, 5, 0, 220, 65, 252, 2, 14, 66, 252, 128, 16, 0, 5, 0, 16, 66, 252, 128, 16, 0, 5, 0, 18, 66, 252, 2, 252, 66, 252, 128, 16, 0, 5, 0, 254, 66, 252, 2, 28, 68, 252, 128, 16, 0, 5, 0, 30, 68, 252, 2, 94, 68, 252, 128, 16, 0, 5, 128, 5, 0, 112, 0, 96, 68, 252, 2, 116, 69, 252, 128, 16, 0, 5, 128, 27, 0, 5, 0, 118, 69, 252, 2, 46, 72, 252, 128, 16, 0, 5, 0, 48, 72, 252, 128, 16, 0, 5, 0, 50, 72, 252, 2, 112, 79, 252, 128, 16, 0, 5, 0, 114, 79, 252, 2, 132, 79, 252, 128, 16, 0, 5, 0, 134, 79, 252, 128, 5, 0, 112, 0, 136, 79, 252, 2, 158, 89, 252, 128, 16, 0, 5, 0, 160, 89, 252, 2, 104, 90, 252, 128, 16, 0, 5, 0, 106, 90, 252, 2, 188, 92, 252, 128, 16, 0, 5, 0, 190, 92, 252, 2, 36, 93, 252, 128, 16, 0, 5, 128, 49, 0, 5, 0, 38, 93, 252, 2, 208, 93, 252, 128, 16, 0, 5, 0, 210, 93, 252, 2, 32, 94, 252, 128, 16, 0, 5, 0, 34, 94, 252, 2, 36, 94, 252, 128, 38, 0, 5, 0, 38, 94, 252, 128, 49, 0, 5, 0, 40, 94, 252, 2, 56, 94, 252, 128, 27, 0, 5, 0, 58, 94, 252, 2, 100, 98, 252, 128, 16, 0, 5, 128, 5, 0, 112, 0, 102, 98, 252, 2, 176, 98, 252, 128, 16, 0, 5, 0, 178, 98, 252, 2, 156, 106, 252, 128, 16, 0, 5, 0, 158, 106, 252, 2, 250, 106, 252, 128, 16, 0, 5, 0, 252, 106, 252, 2, 26, 109, 252, 128, 16, 0, 5, 0, 28, 109, 252, 2, 184, 110, 252, 128, 16, 0, 5, 0, 186, 110, 252, 2, 100, 111, 252, 128, 16, 0, 5, 0, 102, 111, 252, 2, 8, 112, 252, 128, 16, 0, 5, 128, 27, 0, 5, 0, 10, 112, 252, 2, 128, 113, 252, 128, 16, 0, 5, 128, 5, 0, 112, 0, 130, 113, 252, 2, 54, 123, 252, 128, 16, 0, 5, 0, 56, 123, 252, 2, 16, 124, 252, 128, 16, 0, 5, 0, 18, 124, 252, 2, 134, 125, 252, 128, 16, 0, 5, 0, 136, 125, 252, 2, 210, 134, 252, 128, 16, 0, 5, 128, 5, 0, 112, 0, 212, 134, 252, 128, 16, 0, 5, 0, 214, 134, 252, 2, 240, 134, 252, 128, 16, 0, 5, 128, 27, 0, 5, 128, 49, 0, 5, 0, 242, 134, 252, 2, 158, 135, 252, 128, 16, 0, 5, 128, 27, 0, 5, 0, 160, 135, 252, 2, 50, 136, 252, 128, 16, 0, 5, 0, 52, 136, 252, 2, 80, 137, 252, 128, 49, 0, 5, 0, 82, 137, 252, 2, 88, 140, 252, 128, 16, 0, 5, 0, 90, 140, 252, 2, 156, 140, 252, 128, 16, 0, 5, 0, 158, 140, 252, 2, 134, 141, 252, 128, 27, 0, 5, 0, 136, 141, 252, 2, 22, 148, 252, 128, 16, 0, 5, 0, 24, 148, 252, 2, 50, 148, 252, 128, 49, 0, 5, 0, 52, 148, 252, 2, 118, 152, 252, 128, 16, 0, 5, 0, 120, 152, 252, 2, 154, 154, 252, 128, 16, 0, 5, 128, 5, 0, 112, 0, 156, 154, 252, 2, 76, 158, 252, 128, 27, 0, 5, 0, 78, 158, 252, 2, 76, 168, 252, 128, 16, 0, 5, 0, 78, 168, 252, 2, 70, 174, 252, 128, 16, 0, 5, 0, 72, 174, 252, 128, 16, 0, 5, 0, 74, 174, 252, 2, 168, 175, 252, 128, 49, 0, 5, 0, 170, 175, 252, 2, 220, 185, 252, 128, 16, 0, 5, 0, 222, 185, 252, 2, 198, 188, 252, 128, 16, 0, 5, 0, 200, 188, 252, 2, 238, 189, 252, 128, 16, 0, 5, 0, 240, 189, 252, 128, 16, 0, 5, 128, 5, 0, 112, 0, 242, 189, 252, 128, 16, 0, 5, 128, 5, 0, 112, 128, 5, 0, 120, 0, 244, 189, 252, 2, 14, 193, 252, 128, 16, 0, 5, 128, 5, 0, 112, 128, 5, 0, 120, 0, 16, 193, 252, 2, 22, 193, 252, 128, 16, 0, 5, 0, 24, 193, 252, 2, 136, 195, 252, 128, 16, 0, 5, 0, 138, 195, 252, 2, 84, 198, 252, 128, 16, 0, 5, 0, 86, 198, 252, 128, 16, 0, 5, 0, 88, 198, 252, 2, 214, 198, 252, 128, 16, 0, 5, 0, 216, 198, 252, 2, 40, 199, 252, 128, 16, 0, 5, 0, 42, 199, 252, 2, 74, 200, 252, 128, 16, 0, 5, 0, 76, 200, 252, 2, 128, 203, 252, 128, 16, 0, 5, 128, 5, 0, 112, 0, 130, 203, 252, 128, 16, 0, 5, 0, 132, 203, 252, 2, 226, 203, 252, 128, 16, 0, 5, 128, 5, 0, 112, 0, 228, 203, 252, 2, 90, 215, 252, 128, 16, 0, 5, 0, 92, 215, 252, 2, 222, 215, 252, 128, 16, 0, 5, 0, 224, 215, 252, 2, 152, 216, 252, 128, 16, 0, 5, 0, 154, 216, 252, 2, 10, 217, 252, 128, 16, 0, 5, 128, 5, 0, 112, 0, 12, 217, 252, 2, 106, 218, 252, 128, 16, 0, 5, 0, 108, 218, 252, 2, 90, 219, 252, 128, 16, 0, 5, 0, 92, 219, 252, 2, 128, 219, 252, 128, 16, 0, 5, 0, 130, 219, 252, 2, 106, 222, 252, 128, 16, 0, 5, 0, 108, 222, 252, 2, 140, 222, 252, 128, 16, 0, 5, 0, 142, 222, 252, 2, 6, 223, 252, 128, 16, 0, 5, 0, 8, 223, 252, 128, 16, 0, 5, 128, 5, 0, 112, 128, 5, 0, 120, 0, 10, 223, 252, 2, 174, 252, 252, 128, 16, 0, 5, 0, 176, 252, 252, 2, 178, 252, 252, 128, 16, 0, 5, 0, 180, 252, 252, 2, 150, 254, 252, 128, 16, 0, 5, 0, 152, 254, 252, 2, 202, 14, 253, 128, 16, 0, 5, 0, 204, 14, 253, 2, 118, 15, 253, 128, 16, 0, 5, 0, 120, 15, 253, 2, 52, 16, 253, 128, 16, 0, 5, 0, 54, 16, 253, 128, 16, 0, 5, 0, 56, 16, 253, 2, 136, 24, 253, 128, 16, 0, 5, 0, 138, 24, 253, 128, 16, 0, 5, 0, 140, 24, 253, 128, 16, 0, 5, 0, 142, 24, 253, 2, 42, 25, 253, 128, 16, 0, 5, 0, 44, 25, 253, 2, 148, 27, 253, 128, 16, 0, 5, 0, 150, 27, 253, 2, 240, 27, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 242, 27, 253, 2, 152, 28, 253, 128, 49, 0, 5, 0, 154, 28, 253, 2, 38, 30, 253, 128, 16, 0, 5, 0, 40, 30, 253, 2, 192, 40, 253, 128, 16, 0, 5, 0, 194, 40, 253, 2, 160, 43, 253, 128, 16, 0, 5, 0, 162, 43, 253, 2, 90, 44, 253, 128, 16, 0, 5, 0, 92, 44, 253, 2, 112, 45, 253, 128, 16, 0, 5, 0, 114, 45, 253, 2, 36, 47, 253, 128, 16, 0, 5, 0, 38, 47, 253, 2, 190, 48, 253, 128, 16, 0, 5, 0, 192, 48, 253, 2, 218, 48, 253, 128, 16, 0, 5, 128, 27, 0, 5, 0, 220, 48, 253, 2, 14, 49, 253, 128, 49, 0, 5, 0, 16, 49, 253, 2, 8, 50, 253, 128, 16, 0, 5, 128, 27, 0, 5, 0, 10, 50, 253, 2, 202, 53, 253, 128, 16, 0, 5, 0, 204, 53, 253, 2, 190, 54, 253, 128, 16, 0, 5, 0, 192, 54, 253, 2, 84, 55, 253, 128, 16, 0, 5, 128, 49, 0, 5, 0, 86, 55, 253, 2, 84, 59, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 86, 59, 253, 2, 230, 68, 253, 128, 16, 0, 5, 0, 232, 68, 253, 2, 218, 70, 253, 128, 16, 0, 5, 0, 220, 70, 253, 2, 100, 76, 253, 128, 16, 0, 5, 0, 102, 76, 253, 2, 186, 77, 253, 128, 16, 0, 5, 0, 188, 77, 253, 2, 128, 78, 253, 128, 16, 0, 5, 0, 130, 78, 253, 2, 176, 78, 253, 128, 16, 0, 5, 0, 178, 78, 253, 128, 16, 0, 5, 128, 5, 0, 112, 128, 5, 0, 120, 0, 180, 78, 253, 2, 144, 82, 253, 128, 49, 0, 5, 0, 146, 82, 253, 2, 26, 84, 253, 128, 27, 0, 5, 0, 28, 84, 253, 2, 98, 87, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 100, 87, 253, 2, 148, 93, 253, 128, 16, 0, 5, 0, 150, 93, 253, 2, 166, 93, 253, 128, 49, 0, 5, 0, 168, 93, 253, 2, 30, 98, 253, 128, 16, 0, 5, 0, 32, 98, 253, 2, 98, 98, 253, 128, 16, 0, 5, 0, 100, 98, 253, 2, 242, 98, 253, 128, 16, 0, 5, 128, 27, 0, 5, 0, 244, 98, 253, 2, 86, 114, 253, 128, 16, 0, 5, 0, 88, 114, 253, 2, 200, 118, 253, 128, 16, 0, 5, 0, 202, 118, 253, 128, 16, 0, 5, 0, 204, 118, 253, 2, 2, 120, 253, 128, 16, 0, 5, 0, 4, 120, 253, 2, 6, 120, 253, 128, 16, 0, 5, 0, 8, 120, 253, 2, 84, 124, 253, 128, 16, 0, 5, 0, 86, 124, 253, 2, 154, 125, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 156, 125, 253, 128, 16, 0, 5, 0, 158, 125, 253, 2, 186, 131, 253, 128, 16, 0, 5, 0, 188, 131, 253, 2, 228, 131, 253, 128, 16, 0, 5, 0, 230, 131, 253, 2, 178, 134, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 180, 134, 253, 2, 216, 139, 253, 128, 16, 0, 5, 0, 218, 139, 253, 128, 16, 0, 5, 0, 220, 139, 253, 2, 44, 140, 253, 128, 16, 0, 5, 0, 46, 140, 253, 2, 152, 140, 253, 128, 16, 0, 5, 0, 154, 140, 253, 2, 160, 141, 253, 128, 16, 0, 5, 0, 162, 141, 253, 2, 196, 145, 253, 128, 16, 0, 5, 0, 198, 145, 253, 2, 30, 147, 253, 128, 16, 0, 5, 0, 32, 147, 253, 2, 108, 147, 253, 128, 16, 0, 5, 0, 110, 147, 253, 2, 184, 147, 253, 128, 16, 0, 5, 0, 186, 147, 253, 2, 142, 148, 253, 128, 16, 0, 5, 0, 144, 148, 253, 2, 162, 148, 253, 128, 27, 0, 5, 0, 164, 148, 253, 2, 70, 153, 253, 128, 16, 0, 5, 0, 72, 153, 253, 2, 82, 154, 253, 128, 16, 0, 5, 0, 84, 154, 253, 2, 242, 156, 253, 128, 16, 0, 5, 0, 244, 156, 253, 2, 122, 157, 253, 128, 16, 0, 5, 0, 124, 157, 253, 2, 164, 157, 253, 128, 16, 0, 5, 0, 166, 157, 253, 2, 170, 157, 253, 128, 16, 0, 5, 0, 172, 157, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 174, 157, 253, 2, 20, 163, 253, 128, 16, 0, 5, 0, 22, 163, 253, 2, 110, 164, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 112, 164, 253, 2, 252, 164, 253, 128, 16, 0, 5, 0, 254, 164, 253, 2, 208, 165, 253, 128, 16, 0, 5, 0, 210, 165, 253, 2, 244, 171, 253, 128, 16, 0, 5, 0, 246, 171, 253, 2, 168, 173, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 170, 173, 253, 2, 232, 175, 253, 128, 16, 0, 5, 0, 234, 175, 253, 2, 186, 176, 253, 128, 16, 0, 5, 0, 188, 176, 253, 2, 58, 180, 253, 128, 16, 0, 5, 0, 60, 180, 253, 2, 136, 180, 253, 128, 16, 0, 5, 0, 138, 180, 253, 2, 172, 180, 253, 128, 16, 0, 5, 0, 174, 180, 253, 2, 166, 181, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 168, 181, 253, 2, 2, 184, 253, 128, 16, 0, 5, 0, 4, 184, 253, 2, 64, 193, 253, 128, 16, 0, 5, 0, 66, 193, 253, 2, 218, 195, 253, 128, 16, 0, 5, 0, 220, 195, 253, 2, 72, 206, 253, 128, 16, 0, 5, 0, 74, 206, 253, 2, 240, 208, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 242, 208, 253, 2, 178, 209, 253, 128, 16, 0, 5, 0, 180, 209, 253, 2, 56, 211, 253, 128, 16, 0, 5, 0, 58, 211, 253, 2, 196, 212, 253, 128, 16, 0, 5, 0, 198, 212, 253, 2, 94, 213, 253, 128, 16, 0, 5, 0, 96, 213, 253, 2, 210, 213, 253, 128, 16, 0, 5, 0, 212, 213, 253, 128, 16, 0, 5, 0, 214, 213, 253, 2, 124, 214, 253, 128, 16, 0, 5, 0, 126, 214, 253, 2, 252, 214, 253, 128, 16, 0, 5, 0, 254, 214, 253, 2, 36, 217, 253, 128, 16, 0, 5, 0, 38, 217, 253, 2, 62, 217, 253, 128, 16, 0, 5, 0, 64, 217, 253, 2, 222, 217, 253, 128, 16, 0, 5, 0, 224, 217, 253, 2, 6, 218, 253, 128, 16, 0, 5, 0, 8, 218, 253, 2, 62, 218, 253, 128, 16, 0, 5, 0, 64, 218, 253, 2, 220, 218, 253, 128, 16, 0, 5, 0, 222, 218, 253, 2, 240, 219, 253, 128, 16, 0, 5, 0, 242, 219, 253, 2, 176, 220, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 178, 220, 253, 2, 236, 220, 253, 128, 16, 0, 5, 0, 238, 220, 253, 2, 248, 220, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 250, 220, 253, 2, 110, 223, 253, 128, 16, 0, 5, 0, 112, 223, 253, 2, 12, 224, 253, 128, 16, 0, 5, 128, 5, 0, 112, 0, 14, 224, 253, 2, 96, 224, 253, 128, 16, 0, 5, 0, 98, 224, 253, 2, 118, 224, 253, 128, 16, 0, 5, 128, 5, 0, 112, 128, 5, 0, 120, 0, 120, 224, 253, 2, 218, 224, 253, 128, 16, 0, 5, 0, 220, 224, 253, 2, 236, 224, 253, 128, 16, 0, 5, 0, 238, 224, 253, 2, 40, 225, 253, 0, 2, 4, 254, 0, 2, 2, 255, 0, 0, 253, 255, 0, 0, 255, 255, 0, 255, 255, 255, 26, 9, 5, 5, 1, 0, 76, 0, 255, 255, 219, 251, 194, 5, 108, 0, 255, 255, 219, 251, 194, 5, 70, 97, 174, 5, 48, 0, 6, 3, 255, 255, 78, 97, 160, 5, 70, 97, 5, 5, 48, 0, 6, 3, 255, 255, 78, 97, 5, 5, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 8, 102, 5, 5, 84, 6, 255, 255, 10, 102, 5, 5, 85, 6, 255, 255, 18, 102, 5, 5, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 16, 102, 5, 5, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 22, 102, 5, 5, 124, 105, 5, 5, 1, 0, 190, 9, 255, 255, 128, 105, 5, 5, 215, 9, 255, 255, 130, 105, 5, 5, 126, 108, 5, 5, 2, 0, 62, 11, 255, 255, 130, 108, 5, 5, 86, 11, 255, 255, 128, 108, 5, 5, 87, 11, 255, 255, 132, 108, 5, 5, 26, 109, 5, 5, 48, 0, 215, 11, 255, 255, 30, 109, 5, 5, 90, 109, 5, 5, 1, 0, 190, 11, 255, 255, 96, 109, 5, 5, 215, 11, 255, 255, 100, 109, 5, 5, 92, 109, 5, 5, 48, 0, 190, 11, 255, 255, 98, 109, 5, 5, 136, 110, 5, 5, 48, 0, 86, 12, 255, 255, 140, 110, 5, 5, 120, 111, 5, 5, 48, 0, 213, 12, 255, 255, 122, 111, 5, 5, 136, 111, 5, 5, 2, 0, 194, 12, 8, 0, 213, 12, 255, 255, 138, 111, 5, 5, 214, 12, 255, 255, 140, 111, 5, 5, 240, 127, 142, 111, 5, 5, 213, 12, 255, 255, 144, 111, 5, 5, 142, 111, 5, 5, 48, 0, 213, 12, 255, 255, 144, 111, 5, 5, 138, 112, 5, 5, 1, 0, 62, 13, 255, 255, 144, 112, 5, 5, 87, 13, 255, 255, 148, 112, 5, 5, 140, 112, 5, 5, 48, 0, 62, 13, 255, 255, 146, 112, 5, 5, 146, 113, 5, 5, 2, 0, 202, 13, 255, 255, 148, 113, 5, 5, 207, 13, 4, 0, 223, 13, 255, 255, 156, 113, 5, 5, 240, 127, 152, 113, 5, 5, 202, 13, 255, 255, 154, 113, 5, 5, 152, 113, 5, 5, 48, 0, 202, 13, 255, 255, 154, 113, 5, 5, 122, 114, 5, 5, 45, 0, 24, 14, 81, 0, 35, 14, 42, 0, 41, 14, 20, 0, 44, 14, 9, 0, 44, 14, 254, 193, 197, 2, 45, 14, 255, 193, 197, 66, 46, 14, 0, 194, 197, 130, 41, 14, 250, 193, 197, 66, 42, 14, 251, 193, 197, 130, 43, 14, 252, 193, 197, 194, 38, 14, 9, 0, 38, 14, 246, 193, 197, 130, 39, 14, 247, 193, 197, 194, 40, 14, 249, 193, 197, 2, 35, 14, 242, 193, 197, 194, 36, 14, 244, 193, 197, 2, 37, 14, 245, 193, 197, 66, 29, 14, 20, 0, 32, 14, 9, 0, 32, 14, 239, 193, 197, 2, 33, 14, 240, 193, 197, 66, 34, 14, 241, 193, 197, 130, 29, 14, 235, 193, 197, 66, 30, 14, 236, 193, 197, 130, 31, 14, 237, 193, 197, 194, 24, 14, 229, 193, 197, 2, 25, 14, 230, 193, 197, 66, 26, 14, 231, 193, 197, 130, 27, 14, 232, 193, 197, 194, 28, 14, 234, 193, 197, 2, 12, 14, 42, 0, 18, 14, 20, 0, 21, 14, 9, 0, 21, 14, 225, 193, 197, 66, 22, 14, 226, 193, 197, 130, 23, 14, 227, 193, 197, 194, 18, 14, 221, 193, 197, 130, 19, 14, 222, 193, 197, 194, 20, 14, 224, 193, 197, 2, 15, 14, 9, 0, 15, 14, 217, 193, 197, 194, 16, 14, 219, 193, 197, 2, 17, 14, 220, 193, 197, 66, 12, 14, 214, 193, 197, 2, 13, 14, 215, 193, 197, 66, 14, 14, 216, 193, 197, 130, 6, 14, 20, 0, 9, 14, 9, 0, 9, 14, 210, 193, 197, 66, 10, 14, 211, 193, 197, 130, 11, 14, 212, 193, 197, 194, 6, 14, 206, 193, 197, 130, 7, 14, 207, 193, 197, 194, 8, 14, 209, 193, 197, 2, 1, 14, 200, 193, 197, 66, 2, 14, 201, 193, 197, 130, 3, 14, 202, 193, 197, 194, 4, 14, 204, 193, 197, 2, 5, 14, 205, 193, 197, 66, 124, 114, 5, 5, 45, 0, 24, 14, 81, 0, 35, 14, 42, 0, 41, 14, 20, 0, 44, 14, 9, 0, 44, 14, 254, 193, 197, 66, 45, 14, 255, 193, 197, 130, 46, 14, 0, 194, 197, 194, 41, 14, 250, 193, 197, 130, 42, 14, 251, 193, 197, 194, 43, 14, 253, 193, 197, 2, 38, 14, 9, 0, 38, 14, 246, 193, 197, 194, 39, 14, 248, 193, 197, 2, 40, 14, 249, 193, 197, 66, 35, 14, 243, 193, 197, 2, 36, 14, 244, 193, 197, 66, 37, 14, 245, 193, 197, 130, 29, 14, 20, 0, 32, 14, 9, 0, 32, 14, 239, 193, 197, 66, 33, 14, 240, 193, 197, 130, 34, 14, 241, 193, 197, 194, 29, 14, 235, 193, 197, 130, 30, 14, 236, 193, 197, 194, 31, 14, 238, 193, 197, 2, 24, 14, 229, 193, 197, 66, 25, 14, 230, 193, 197, 130, 26, 14, 231, 193, 197, 194, 27, 14, 233, 193, 197, 2, 28, 14, 234, 193, 197, 66, 12, 14, 42, 0, 18, 14, 20, 0, 21, 14, 9, 0, 21, 14, 225, 193, 197, 130, 22, 14, 226, 193, 197, 194, 23, 14, 228, 193, 197, 2, 18, 14, 221, 193, 197, 194, 19, 14, 223, 193, 197, 2, 20, 14, 224, 193, 197, 66, 15, 14, 9, 0, 15, 14, 218, 193, 197, 2, 16, 14, 219, 193, 197, 66, 17, 14, 220, 193, 197, 130, 12, 14, 214, 193, 197, 66, 13, 14, 215, 193, 197, 130, 14, 14, 216, 193, 197, 194, 6, 14, 20, 0, 9, 14, 9, 0, 9, 14, 210, 193, 197, 130, 10, 14, 211, 193, 197, 194, 11, 14, 213, 193, 197, 2, 6, 14, 206, 193, 197, 194, 7, 14, 208, 193, 197, 2, 8, 14, 209, 193, 197, 66, 1, 14, 200, 193, 197, 130, 2, 14, 201, 193, 197, 194, 3, 14, 203, 193, 197, 2, 4, 14, 204, 193, 197, 66, 5, 14, 205, 193, 197, 130, 126, 114, 5, 5, 45, 0, 24, 14, 81, 0, 35, 14, 42, 0, 41, 14, 20, 0, 44, 14, 9, 0, 44, 14, 254, 193, 197, 130, 45, 14, 255, 193, 197, 194, 46, 14, 1, 194, 197, 2, 41, 14, 250, 193, 197, 194, 42, 14, 252, 193, 197, 2, 43, 14, 253, 193, 197, 66, 38, 14, 9, 0, 38, 14, 247, 193, 197, 2, 39, 14, 248, 193, 197, 66, 40, 14, 249, 193, 197, 130, 35, 14, 243, 193, 197, 66, 36, 14, 244, 193, 197, 130, 37, 14, 245, 193, 197, 194, 29, 14, 20, 0, 32, 14, 9, 0, 32, 14, 239, 193, 197, 130, 33, 14, 240, 193, 197, 194, 34, 14, 242, 193, 197, 2, 29, 14, 235, 193, 197, 194, 30, 14, 237, 193, 197, 2, 31, 14, 238, 193, 197, 66, 24, 14, 229, 193, 197, 130, 25, 14, 230, 193, 197, 194, 26, 14, 232, 193, 197, 2, 27, 14, 233, 193, 197, 66, 28, 14, 234, 193, 197, 130, 12, 14, 42, 0, 18, 14, 20, 0, 21, 14, 9, 0, 21, 14, 225, 193, 197, 194, 22, 14, 227, 193, 197, 2, 23, 14, 228, 193, 197, 66, 18, 14, 222, 193, 197, 2, 19, 14, 223, 193, 197, 66, 20, 14, 224, 193, 197, 130, 15, 14, 9, 0, 15, 14, 218, 193, 197, 66, 16, 14, 219, 193, 197, 130, 17, 14, 220, 193, 197, 194, 12, 14, 214, 193, 197, 130, 13, 14, 215, 193, 197, 194, 14, 14, 217, 193, 197, 2, 6, 14, 20, 0, 9, 14, 9, 0, 9, 14, 210, 193, 197, 194, 10, 14, 212, 193, 197, 2, 11, 14, 213, 193, 197, 66, 6, 14, 207, 193, 197, 2, 7, 14, 208, 193, 197, 66, 8, 14, 209, 193, 197, 130, 1, 14, 200, 193, 197, 194, 2, 14, 202, 193, 197, 2, 3, 14, 203, 193, 197, 66, 4, 14, 204, 193, 197, 130, 5, 14, 205, 193, 197, 194, 128, 114, 5, 5, 45, 0, 24, 14, 81, 0, 35, 14, 42, 0, 41, 14, 20, 0, 44, 14, 9, 0, 44, 14, 254, 193, 197, 194, 45, 14, 0, 194, 197, 2, 46, 14, 1, 194, 197, 66, 41, 14, 251, 193, 197, 2, 42, 14, 252, 193, 197, 66, 43, 14, 253, 193, 197, 130, 38, 14, 9, 0, 38, 14, 247, 193, 197, 66, 39, 14, 248, 193, 197, 130, 40, 14, 249, 193, 197, 194, 35, 14, 243, 193, 197, 130, 36, 14, 244, 193, 197, 194, 37, 14, 246, 193, 197, 2, 29, 14, 20, 0, 32, 14, 9, 0, 32, 14, 239, 193, 197, 194, 33, 14, 241, 193, 197, 2, 34, 14, 242, 193, 197, 66, 29, 14, 236, 193, 197, 2, 30, 14, 237, 193, 197, 66, 31, 14, 238, 193, 197, 130, 24, 14, 229, 193, 197, 194, 25, 14, 231, 193, 197, 2, 26, 14, 232, 193, 197, 66, 27, 14, 233, 193, 197, 130, 28, 14, 234, 193, 197, 194, 12, 14, 42, 0, 18, 14, 20, 0, 21, 14, 9, 0, 21, 14, 226, 193, 197, 2, 22, 14, 227, 193, 197, 66, 23, 14, 228, 193, 197, 130, 18, 14, 222, 193, 197, 66, 19, 14, 223, 193, 197, 130, 20, 14, 224, 193, 197, 194, 15, 14, 9, 0, 15, 14, 218, 193, 197, 130, 16, 14, 219, 193, 197, 194, 17, 14, 221, 193, 197, 2, 12, 14, 214, 193, 197, 194, 13, 14, 216, 193, 197, 2, 14, 14, 217, 193, 197, 66, 6, 14, 20, 0, 9, 14, 9, 0, 9, 14, 211, 193, 197, 2, 10, 14, 212, 193, 197, 66, 11, 14, 213, 193, 197, 130, 6, 14, 207, 193, 197, 66, 7, 14, 208, 193, 197, 130, 8, 14, 209, 193, 197, 194, 1, 14, 201, 193, 197, 2, 2, 14, 202, 193, 197, 66, 3, 14, 203, 193, 197, 130, 4, 14, 204, 193, 197, 194, 5, 14, 206, 193, 197, 2, 130, 114, 5, 5, 45, 0, 24, 14, 81, 0, 35, 14, 42, 0, 41, 14, 20, 0, 44, 14, 9, 0, 44, 14, 255, 193, 197, 2, 45, 14, 0, 194, 197, 66, 46, 14, 1, 194, 197, 130, 41, 14, 251, 193, 197, 66, 42, 14, 252, 193, 197, 130, 43, 14, 253, 193, 197, 194, 38, 14, 9, 0, 38, 14, 247, 193, 197, 130, 39, 14, 248, 193, 197, 194, 40, 14, 250, 193, 197, 2, 35, 14, 243, 193, 197, 194, 36, 14, 245, 193, 197, 2, 37, 14, 246, 193, 197, 66, 29, 14, 20, 0, 32, 14, 9, 0, 32, 14, 240, 193, 197, 2, 33, 14, 241, 193, 197, 66, 34, 14, 242, 193, 197, 130, 29, 14, 236, 193, 197, 66, 30, 14, 237, 193, 197, 130, 31, 14, 238, 193, 197, 194, 24, 14, 230, 193, 197, 2, 25, 14, 231, 193, 197, 66, 26, 14, 232, 193, 197, 130, 27, 14, 233, 193, 197, 194, 28, 14, 235, 193, 197, 2, 12, 14, 42, 0, 18, 14, 20, 0, 21, 14, 9, 0, 21, 14, 226, 193, 197, 66, 22, 14, 227, 193, 197, 130, 23, 14, 228, 193, 197, 194, 18, 14, 222, 193, 197, 130, 19, 14, 223, 193, 197, 194, 20, 14, 225, 193, 197, 2, 15, 14, 9, 0, 15, 14, 218, 193, 197, 194, 16, 14, 220, 193, 197, 2, 17, 14, 221, 193, 197, 66, 12, 14, 215, 193, 197, 2, 13, 14, 216, 193, 197, 66, 14, 14, 217, 193, 197, 130, 6, 14, 20, 0, 9, 14, 9, 0, 9, 14, 211, 193, 197, 66, 10, 14, 212, 193, 197, 130, 11, 14, 213, 193, 197, 194, 6, 14, 207, 193, 197, 130, 7, 14, 208, 193, 197, 194, 8, 14, 210, 193, 197, 2, 1, 14, 201, 193, 197, 66, 2, 14, 202, 193, 197, 130, 3, 14, 203, 193, 197, 194, 4, 14, 205, 193, 197, 2, 5, 14, 206, 193, 197, 66, 205, 243, 194, 5, 48, 0, 50, 14, 255, 255, 106, 114, 5, 5, 122, 115, 5, 5, 44, 0, 154, 14, 81, 0, 167, 14, 42, 0, 173, 14, 20, 0, 221, 14, 9, 0, 221, 14, 52, 194, 197, 227, 222, 14, 1, 194, 197, 194, 223, 14, 16, 194, 197, 194, 173, 14, 57, 194, 197, 66, 174, 14, 58, 194, 197, 130, 220, 14, 50, 194, 197, 195, 170, 14, 9, 0, 170, 14, 11, 194, 197, 194, 171, 14, 54, 194, 197, 194, 172, 14, 56, 194, 197, 2, 167, 14, 46, 194, 197, 194, 168, 14, 48, 194, 197, 2, 169, 14, 49, 194, 197, 66, 159, 14, 20, 0, 162, 14, 9, 0, 162, 14, 43, 194, 197, 2, 163, 14, 44, 194, 197, 66, 165, 14, 45, 194, 197, 130, 159, 14, 39, 194, 197, 66, 160, 14, 40, 194, 197, 130, 161, 14, 41, 194, 197, 194, 154, 14, 33, 194, 197, 2, 155, 14, 34, 194, 197, 66, 156, 14, 35, 194, 197, 130, 157, 14, 36, 194, 197, 194, 158, 14, 38, 194, 197, 2, 143, 14, 37, 0, 148, 14, 20, 0, 151, 14, 9, 0, 151, 14, 29, 194, 197, 66, 152, 14, 30, 194, 197, 130, 153, 14, 31, 194, 197, 194, 148, 14, 25, 194, 197, 130, 149, 14, 26, 194, 197, 194, 150, 14, 28, 194, 197, 2, 143, 14, 19, 194, 197, 66, 144, 14, 20, 194, 197, 130, 145, 14, 21, 194, 197, 194, 146, 14, 23, 194, 197, 2, 147, 14, 24, 194, 197, 66, 136, 14, 20, 0, 140, 14, 9, 0, 140, 14, 14, 194, 197, 66, 141, 14, 18, 194, 197, 2, 142, 14, 15, 194, 197, 130, 136, 14, 9, 194, 197, 66, 137, 14, 10, 194, 197, 130, 138, 14, 13, 194, 197, 2, 129, 14, 3, 194, 197, 2, 130, 14, 4, 194, 197, 66, 132, 14, 5, 194, 197, 130, 134, 14, 6, 194, 197, 194, 135, 14, 8, 194, 197, 2, 124, 115, 5, 5, 44, 0, 154, 14, 81, 0, 167, 14, 42, 0, 173, 14, 20, 0, 221, 14, 9, 0, 221, 14, 53, 194, 197, 67, 222, 14, 2, 194, 197, 2, 223, 14, 17, 194, 197, 2, 173, 14, 57, 194, 197, 130, 174, 14, 58, 194, 197, 194, 220, 14, 51, 194, 197, 35, 170, 14, 9, 0, 170, 14, 12, 194, 197, 2, 171, 14, 55, 194, 197, 2, 172, 14, 56, 194, 197, 66, 167, 14, 47, 194, 197, 2, 168, 14, 48, 194, 197, 66, 169, 14, 49, 194, 197, 130, 159, 14, 20, 0, 162, 14, 9, 0, 162, 14, 43, 194, 197, 66, 163, 14, 44, 194, 197, 130, 165, 14, 45, 194, 197, 194, 159, 14, 39, 194, 197, 130, 160, 14, 40, 194, 197, 194, 161, 14, 42, 194, 197, 2, 154, 14, 33, 194, 197, 66, 155, 14, 34, 194, 197, 130, 156, 14, 35, 194, 197, 194, 157, 14, 37, 194, 197, 2, 158, 14, 38, 194, 197, 66, 143, 14, 37, 0, 148, 14, 20, 0, 151, 14, 9, 0, 151, 14, 29, 194, 197, 130, 152, 14, 30, 194, 197, 194, 153, 14, 32, 194, 197, 2, 148, 14, 25, 194, 197, 194, 149, 14, 27, 194, 197, 2, 150, 14, 28, 194, 197, 66, 143, 14, 19, 194, 197, 130, 144, 14, 20, 194, 197, 194, 145, 14, 22, 194, 197, 2, 146, 14, 23, 194, 197, 66, 147, 14, 24, 194, 197, 130, 136, 14, 20, 0, 140, 14, 9, 0, 140, 14, 14, 194, 197, 130, 141, 14, 18, 194, 197, 66, 142, 14, 15, 194, 197, 194, 136, 14, 9, 194, 197, 130, 137, 14, 10, 194, 197, 194, 138, 14, 13, 194, 197, 66, 129, 14, 3, 194, 197, 66, 130, 14, 4, 194, 197, 130, 132, 14, 5, 194, 197, 194, 134, 14, 7, 194, 197, 2, 135, 14, 8, 194, 197, 66, 126, 115, 5, 5, 44, 0, 154, 14, 81, 0, 167, 14, 42, 0, 173, 14, 20, 0, 221, 14, 9, 0, 221, 14, 53, 194, 197, 163, 222, 14, 2, 194, 197, 66, 223, 14, 17, 194, 197, 66, 173, 14, 57, 194, 197, 194, 174, 14, 59, 194, 197, 2, 220, 14, 51, 194, 197, 131, 170, 14, 9, 0, 170, 14, 12, 194, 197, 66, 171, 14, 55, 194, 197, 66, 172, 14, 56, 194, 197, 130, 167, 14, 47, 194, 197, 66, 168, 14, 48, 194, 197, 130, 169, 14, 49, 194, 197, 194, 159, 14, 20, 0, 162, 14, 9, 0, 162, 14, 43, 194, 197, 130, 163, 14, 44, 194, 197, 194, 165, 14, 46, 194, 197, 2, 159, 14, 39, 194, 197, 194, 160, 14, 41, 194, 197, 2, 161, 14, 42, 194, 197, 66, 154, 14, 33, 194, 197, 130, 155, 14, 34, 194, 197, 194, 156, 14, 36, 194, 197, 2, 157, 14, 37, 194, 197, 66, 158, 14, 38, 194, 197, 130, 143, 14, 37, 0, 148, 14, 20, 0, 151, 14, 9, 0, 151, 14, 29, 194, 197, 194, 152, 14, 31, 194, 197, 2, 153, 14, 32, 194, 197, 66, 148, 14, 26, 194, 197, 2, 149, 14, 27, 194, 197, 66, 150, 14, 28, 194, 197, 130, 143, 14, 19, 194, 197, 194, 144, 14, 21, 194, 197, 2, 145, 14, 22, 194, 197, 66, 146, 14, 23, 194, 197, 130, 147, 14, 24, 194, 197, 194, 136, 14, 20, 0, 140, 14, 9, 0, 140, 14, 14, 194, 197, 194, 141, 14, 18, 194, 197, 130, 142, 14, 16, 194, 197, 2, 136, 14, 9, 194, 197, 194, 137, 14, 11, 194, 197, 2, 138, 14, 13, 194, 197, 130, 129, 14, 3, 194, 197, 130, 130, 14, 4, 194, 197, 194, 132, 14, 6, 194, 197, 2, 134, 14, 7, 194, 197, 66, 135, 14, 8, 194, 197, 130, 128, 115, 5, 5, 44, 0, 154, 14, 81, 0, 167, 14, 42, 0, 173, 14, 20, 0, 221, 14, 9, 0, 221, 14, 54, 194, 197, 3, 222, 14, 2, 194, 197, 130, 223, 14, 17, 194, 197, 130, 173, 14, 58, 194, 197, 2, 174, 14, 59, 194, 197, 66, 220, 14, 51, 194, 197, 227, 170, 14, 9, 0, 170, 14, 12, 194, 197, 130, 171, 14, 55, 194, 197, 130, 172, 14, 56, 194, 197, 194, 167, 14, 47, 194, 197, 130, 168, 14, 48, 194, 197, 194, 169, 14, 50, 194, 197, 2, 159, 14, 20, 0, 162, 14, 9, 0, 162, 14, 43, 194, 197, 194, 163, 14, 45, 194, 197, 2, 165, 14, 46, 194, 197, 66, 159, 14, 40, 194, 197, 2, 160, 14, 41, 194, 197, 66, 161, 14, 42, 194, 197, 130, 154, 14, 33, 194, 197, 194, 155, 14, 35, 194, 197, 2, 156, 14, 36, 194, 197, 66, 157, 14, 37, 194, 197, 130, 158, 14, 38, 194, 197, 194, 143, 14, 37, 0, 148, 14, 20, 0, 151, 14, 9, 0, 151, 14, 30, 194, 197, 2, 152, 14, 31, 194, 197, 66, 153, 14, 32, 194, 197, 130, 148, 14, 26, 194, 197, 66, 149, 14, 27, 194, 197, 130, 150, 14, 28, 194, 197, 194, 143, 14, 20, 194, 197, 2, 144, 14, 21, 194, 197, 66, 145, 14, 22, 194, 197, 130, 146, 14, 23, 194, 197, 194, 147, 14, 25, 194, 197, 2, 136, 14, 20, 0, 140, 14, 9, 0, 140, 14, 15, 194, 197, 2, 141, 14, 18, 194, 197, 194, 142, 14, 16, 194, 197, 66, 136, 14, 10, 194, 197, 2, 137, 14, 11, 194, 197, 66, 138, 14, 13, 194, 197, 194, 129, 14, 3, 194, 197, 194, 130, 14, 5, 194, 197, 2, 132, 14, 6, 194, 197, 66, 134, 14, 7, 194, 197, 130, 135, 14, 8, 194, 197, 194, 130, 115, 5, 5, 44, 0, 154, 14, 81, 0, 167, 14, 42, 0, 173, 14, 20, 0, 221, 14, 9, 0, 221, 14, 54, 194, 197, 99, 222, 14, 2, 194, 197, 194, 223, 14, 17, 194, 197, 194, 173, 14, 58, 194, 197, 66, 174, 14, 59, 194, 197, 130, 220, 14, 52, 194, 197, 67, 170, 14, 9, 0, 170, 14, 12, 194, 197, 194, 171, 14, 55, 194, 197, 194, 172, 14, 57, 194, 197, 2, 167, 14, 47, 194, 197, 194, 168, 14, 49, 194, 197, 2, 169, 14, 50, 194, 197, 66, 159, 14, 20, 0, 162, 14, 9, 0, 162, 14, 44, 194, 197, 2, 163, 14, 45, 194, 197, 66, 165, 14, 46, 194, 197, 130, 159, 14, 40, 194, 197, 66, 160, 14, 41, 194, 197, 130, 161, 14, 42, 194, 197, 194, 154, 14, 34, 194, 197, 2, 155, 14, 35, 194, 197, 66, 156, 14, 36, 194, 197, 130, 157, 14, 37, 194, 197, 194, 158, 14, 39, 194, 197, 2, 143, 14, 37, 0, 148, 14, 20, 0, 151, 14, 9, 0, 151, 14, 30, 194, 197, 66, 152, 14, 31, 194, 197, 130, 153, 14, 32, 194, 197, 194, 148, 14, 26, 194, 197, 130, 149, 14, 27, 194, 197, 194, 150, 14, 29, 194, 197, 2, 143, 14, 20, 194, 197, 66, 144, 14, 21, 194, 197, 130, 145, 14, 22, 194, 197, 194, 146, 14, 24, 194, 197, 2, 147, 14, 25, 194, 197, 66, 136, 14, 20, 0, 140, 14, 9, 0, 140, 14, 15, 194, 197, 66, 141, 14, 19, 194, 197, 2, 142, 14, 16, 194, 197, 130, 136, 14, 10, 194, 197, 66, 137, 14, 11, 194, 197, 130, 138, 14, 14, 194, 197, 2, 129, 14, 4, 194, 197, 2, 130, 14, 5, 194, 197, 66, 132, 14, 6, 194, 197, 130, 134, 14, 7, 194, 197, 194, 135, 14, 9, 194, 197, 2, 149, 244, 194, 5, 48, 0, 178, 14, 255, 255, 100, 115, 5, 5, 166, 116, 5, 5, 2, 0, 114, 15, 255, 255, 170, 116, 5, 5, 116, 15, 255, 255, 178, 116, 5, 5, 128, 15, 255, 255, 174, 116, 5, 5, 122, 116, 5, 5, 4, 0, 113, 15, 14, 0, 115, 15, 123, 194, 197, 2, 117, 15, 123, 194, 197, 66, 128, 15, 255, 255, 180, 116, 5, 5, 129, 15, 255, 255, 182, 116, 5, 5, 194, 127, 122, 2, 197, 194, 114, 15, 123, 194, 197, 2, 116, 15, 123, 194, 197, 66, 128, 15, 255, 255, 182, 116, 5, 5, 128, 116, 5, 5, 4, 0, 113, 15, 14, 0, 115, 15, 123, 194, 197, 194, 117, 15, 124, 194, 197, 2, 128, 15, 255, 255, 184, 116, 5, 5, 129, 15, 255, 255, 186, 116, 5, 5, 194, 127, 123, 2, 197, 130, 114, 15, 123, 194, 197, 194, 116, 15, 124, 194, 197, 2, 128, 15, 255, 255, 186, 116, 5, 5, 85, 117, 193, 128, 48, 0, 46, 16, 255, 255, 85, 117, 193, 135, 9, 118, 193, 142, 43, 0, 150, 25, 76, 0, 161, 25, 37, 0, 166, 25, 20, 0, 169, 25, 9, 0, 169, 25, 167, 194, 197, 34, 170, 25, 168, 194, 197, 34, 171, 25, 169, 194, 197, 34, 166, 25, 164, 194, 197, 34, 167, 25, 165, 194, 197, 34, 168, 25, 166, 194, 197, 34, 161, 25, 159, 194, 197, 34, 162, 25, 160, 194, 197, 34, 163, 25, 161, 194, 197, 34, 164, 25, 162, 194, 197, 34, 165, 25, 163, 194, 197, 34, 155, 25, 20, 0, 158, 25, 9, 0, 158, 25, 156, 194, 197, 34, 159, 25, 157, 194, 197, 34, 160, 25, 158, 194, 197, 34, 155, 25, 152, 194, 197, 130, 156, 25, 153, 194, 197, 130, 157, 25, 155, 194, 197, 34, 150, 25, 147, 194, 197, 130, 151, 25, 148, 194, 197, 130, 152, 25, 149, 194, 197, 130, 153, 25, 150, 194, 197, 130, 154, 25, 151, 194, 197, 130, 139, 25, 37, 0, 144, 25, 20, 0, 147, 25, 9, 0, 147, 25, 144, 194, 197, 130, 148, 25, 145, 194, 197, 130, 149, 25, 146, 194, 197, 130, 144, 25, 141, 194, 197, 130, 145, 25, 142, 194, 197, 130, 146, 25, 143, 194, 197, 130, 139, 25, 136, 194, 197, 130, 140, 25, 137, 194, 197, 130, 141, 25, 138, 194, 197, 130, 142, 25, 139, 194, 197, 130, 143, 25, 140, 194, 197, 130, 133, 25, 20, 0, 136, 25, 9, 0, 136, 25, 133, 194, 197, 130, 137, 25, 134, 194, 197, 130, 138, 25, 135, 194, 197, 130, 133, 25, 130, 194, 197, 130, 134, 25, 131, 194, 197, 130, 135, 25, 132, 194, 197, 130, 128, 25, 125, 194, 197, 130, 129, 25, 126, 194, 197, 130, 130, 25, 127, 194, 197, 130, 131, 25, 128, 194, 197, 130, 132, 25, 129, 194, 197, 130, 11, 118, 5, 5, 43, 0, 150, 25, 76, 0, 161, 25, 37, 0, 166, 25, 20, 0, 169, 25, 9, 0, 169, 25, 167, 194, 197, 98, 170, 25, 168, 194, 197, 98, 171, 25, 169, 194, 197, 98, 166, 25, 164, 194, 197, 98, 167, 25, 165, 194, 197, 98, 168, 25, 166, 194, 197, 98, 161, 25, 159, 194, 197, 98, 162, 25, 160, 194, 197, 98, 163, 25, 161, 194, 197, 98, 164, 25, 162, 194, 197, 98, 165, 25, 163, 194, 197, 98, 155, 25, 20, 0, 158, 25, 9, 0, 158, 25, 156, 194, 197, 98, 159, 25, 157, 194, 197, 98, 160, 25, 158, 194, 197, 98, 155, 25, 152, 194, 197, 194, 156, 25, 153, 194, 197, 194, 157, 25, 155, 194, 197, 98, 150, 25, 147, 194, 197, 194, 151, 25, 148, 194, 197, 194, 152, 25, 149, 194, 197, 194, 153, 25, 150, 194, 197, 194, 154, 25, 151, 194, 197, 194, 139, 25, 37, 0, 144, 25, 20, 0, 147, 25, 9, 0, 147, 25, 144, 194, 197, 194, 148, 25, 145, 194, 197, 194, 149, 25, 146, 194, 197, 194, 144, 25, 141, 194, 197, 194, 145, 25, 142, 194, 197, 194, 146, 25, 143, 194, 197, 194, 139, 25, 136, 194, 197, 194, 140, 25, 137, 194, 197, 194, 141, 25, 138, 194, 197, 194, 142, 25, 139, 194, 197, 194, 143, 25, 140, 194, 197, 194, 133, 25, 20, 0, 136, 25, 9, 0, 136, 25, 133, 194, 197, 194, 137, 25, 134, 194, 197, 194, 138, 25, 135, 194, 197, 194, 133, 25, 130, 194, 197, 194, 134, 25, 131, 194, 197, 194, 135, 25, 132, 194, 197, 194, 128, 25, 125, 194, 197, 194, 129, 25, 126, 194, 197, 194, 130, 25, 127, 194, 197, 194, 131, 25, 128, 194, 197, 194, 132, 25, 129, 194, 197, 194, 13, 118, 193, 2, 43, 0, 150, 25, 76, 0, 161, 25, 37, 0, 166, 25, 20, 0, 169, 25, 9, 0, 169, 25, 167, 194, 197, 162, 170, 25, 168, 194, 197, 162, 171, 25, 169, 194, 197, 162, 166, 25, 164, 194, 197, 162, 167, 25, 165, 194, 197, 162, 168, 25, 166, 194, 197, 162, 161, 25, 159, 194, 197, 162, 162, 25, 160, 194, 197, 162, 163, 25, 161, 194, 197, 162, 164, 25, 162, 194, 197, 162, 165, 25, 163, 194, 197, 162, 155, 25, 20, 0, 158, 25, 9, 0, 158, 25, 156, 194, 197, 162, 159, 25, 157, 194, 197, 162, 160, 25, 158, 194, 197, 162, 155, 25, 153, 194, 197, 2, 156, 25, 154, 194, 197, 162, 157, 25, 155, 194, 197, 162, 150, 25, 148, 194, 197, 2, 151, 25, 149, 194, 197, 2, 152, 25, 150, 194, 197, 2, 153, 25, 151, 194, 197, 2, 154, 25, 152, 194, 197, 2, 139, 25, 37, 0, 144, 25, 20, 0, 147, 25, 9, 0, 147, 25, 145, 194, 197, 2, 148, 25, 146, 194, 197, 2, 149, 25, 147, 194, 197, 2, 144, 25, 142, 194, 197, 2, 145, 25, 143, 194, 197, 2, 146, 25, 144, 194, 197, 2, 139, 25, 137, 194, 197, 2, 140, 25, 138, 194, 197, 2, 141, 25, 139, 194, 197, 2, 142, 25, 140, 194, 197, 2, 143, 25, 141, 194, 197, 2, 133, 25, 20, 0, 136, 25, 9, 0, 136, 25, 134, 194, 197, 2, 137, 25, 135, 194, 197, 2, 138, 25, 136, 194, 197, 2, 133, 25, 131, 194, 197, 2, 134, 25, 132, 194, 197, 2, 135, 25, 133, 194, 197, 2, 128, 25, 126, 194, 197, 2, 129, 25, 127, 194, 197, 2, 130, 25, 128, 194, 197, 2, 131, 25, 129, 194, 197, 2, 132, 25, 130, 194, 197, 2, 13, 118, 193, 23, 43, 0, 150, 25, 76, 0, 161, 25, 37, 0, 166, 25, 20, 0, 169, 25, 9, 0, 169, 25, 167, 194, 197, 226, 170, 25, 168, 194, 197, 226, 171, 25, 169, 194, 197, 226, 166, 25, 164, 194, 197, 226, 167, 25, 165, 194, 197, 226, 168, 25, 166, 194, 197, 226, 161, 25, 159, 194, 197, 226, 162, 25, 160, 194, 197, 226, 163, 25, 161, 194, 197, 226, 164, 25, 162, 194, 197, 226, 165, 25, 163, 194, 197, 226, 155, 25, 20, 0, 158, 25, 9, 0, 158, 25, 156, 194, 197, 226, 159, 25, 157, 194, 197, 226, 160, 25, 158, 194, 197, 226, 155, 25, 153, 194, 197, 66, 156, 25, 154, 194, 197, 226, 157, 25, 155, 194, 197, 226, 150, 25, 148, 194, 197, 66, 151, 25, 149, 194, 197, 66, 152, 25, 150, 194, 197, 66, 153, 25, 151, 194, 197, 66, 154, 25, 152, 194, 197, 66, 139, 25, 37, 0, 144, 25, 20, 0, 147, 25, 9, 0, 147, 25, 145, 194, 197, 66, 148, 25, 146, 194, 197, 66, 149, 25, 147, 194, 197, 66, 144, 25, 142, 194, 197, 66, 145, 25, 143, 194, 197, 66, 146, 25, 144, 194, 197, 66, 139, 25, 137, 194, 197, 66, 140, 25, 138, 194, 197, 66, 141, 25, 139, 194, 197, 66, 142, 25, 140, 194, 197, 66, 143, 25, 141, 194, 197, 66, 133, 25, 20, 0, 136, 25, 9, 0, 136, 25, 134, 194, 197, 66, 137, 25, 135, 194, 197, 66, 138, 25, 136, 194, 197, 66, 133, 25, 131, 194, 197, 66, 134, 25, 132, 194, 197, 66, 135, 25, 133, 194, 197, 66, 128, 25, 126, 194, 197, 66, 129, 25, 127, 194, 197, 66, 130, 25, 128, 194, 197, 66, 131, 25, 129, 194, 197, 66, 132, 25, 130, 194, 197, 66, 40, 118, 193, 9, 48, 0, 53, 27, 255, 255, 40, 118, 193, 16, 40, 118, 193, 23, 48, 0, 53, 27, 255, 255, 40, 118, 193, 30, 40, 118, 193, 37, 48, 0, 53, 27, 255, 255, 40, 118, 193, 44, 40, 118, 193, 51, 48, 0, 53, 27, 255, 255, 40, 118, 193, 58, 40, 118, 193, 65, 48, 0, 53, 27, 255, 255, 40, 118, 193, 72, 40, 118, 193, 93, 48, 0, 53, 27, 255, 255, 40, 118, 193, 100, 41, 118, 193, 168, 48, 0, 53, 27, 255, 255, 41, 118, 193, 175, 41, 118, 193, 182, 48, 0, 53, 27, 255, 255, 41, 118, 193, 189, 41, 118, 193, 196, 48, 0, 53, 27, 255, 255, 41, 118, 193, 210, 41, 118, 193, 203, 48, 0, 53, 27, 255, 255, 41, 118, 193, 217, 41, 118, 193, 224, 48, 0, 53, 27, 255, 255, 41, 118, 193, 231, 133, 115, 193, 126, 47, 0, 152, 170, 86, 0, 164, 170, 42, 0, 170, 170, 20, 0, 173, 170, 9, 0, 173, 170, 116, 194, 197, 2, 174, 170, 117, 194, 197, 66, 175, 170, 118, 194, 197, 130, 170, 170, 112, 194, 197, 66, 171, 170, 113, 194, 197, 130, 172, 170, 114, 194, 197, 194, 167, 170, 9, 0, 167, 170, 108, 194, 197, 130, 168, 170, 109, 194, 197, 194, 169, 170, 111, 194, 197, 2, 164, 170, 104, 194, 197, 194, 165, 170, 106, 194, 197, 2, 166, 170, 107, 194, 197, 66, 158, 170, 20, 0, 161, 170, 9, 0, 161, 170, 101, 194, 197, 2, 162, 170, 102, 194, 197, 66, 163, 170, 103, 194, 197, 130, 158, 170, 97, 194, 197, 66, 159, 170, 98, 194, 197, 130, 160, 170, 99, 194, 197, 194, 155, 170, 9, 0, 155, 170, 93, 194, 197, 130, 156, 170, 94, 194, 197, 194, 157, 170, 96, 194, 197, 2, 152, 170, 89, 194, 197, 194, 153, 170, 91, 194, 197, 2, 154, 170, 92, 194, 197, 66, 140, 170, 42, 0, 146, 170, 20, 0, 149, 170, 9, 0, 149, 170, 86, 194, 197, 2, 150, 170, 87, 194, 197, 66, 151, 170, 88, 194, 197, 130, 146, 170, 82, 194, 197, 66, 147, 170, 83, 194, 197, 130, 148, 170, 84, 194, 197, 194, 143, 170, 9, 0, 143, 170, 78, 194, 197, 130, 144, 170, 79, 194, 197, 194, 145, 170, 81, 194, 197, 2, 140, 170, 74, 194, 197, 194, 141, 170, 76, 194, 197, 2, 142, 170, 77, 194, 197, 66, 134, 170, 20, 0, 137, 170, 9, 0, 137, 170, 71, 194, 197, 2, 138, 170, 72, 194, 197, 66, 139, 170, 73, 194, 197, 130, 134, 170, 67, 194, 197, 66, 135, 170, 68, 194, 197, 130, 136, 170, 69, 194, 197, 194, 131, 170, 9, 0, 131, 170, 63, 194, 197, 130, 132, 170, 64, 194, 197, 194, 133, 170, 66, 194, 197, 2, 128, 170, 59, 194, 197, 194, 129, 170, 61, 194, 197, 2, 130, 170, 62, 194, 197, 66, 133, 115, 193, 133, 47, 0, 152, 170, 86, 0, 164, 170, 42, 0, 170, 170, 20, 0, 173, 170, 9, 0, 173, 170, 116, 194, 197, 66, 174, 170, 117, 194, 197, 130, 175, 170, 118, 194, 197, 194, 170, 170, 112, 194, 197, 130, 171, 170, 113, 194, 197, 194, 172, 170, 115, 194, 197, 2, 167, 170, 9, 0, 167, 170, 108, 194, 197, 194, 168, 170, 110, 194, 197, 2, 169, 170, 111, 194, 197, 66, 164, 170, 105, 194, 197, 2, 165, 170, 106, 194, 197, 66, 166, 170, 107, 194, 197, 130, 158, 170, 20, 0, 161, 170, 9, 0, 161, 170, 101, 194, 197, 66, 162, 170, 102, 194, 197, 130, 163, 170, 103, 194, 197, 194, 158, 170, 97, 194, 197, 130, 159, 170, 98, 194, 197, 194, 160, 170, 100, 194, 197, 2, 155, 170, 9, 0, 155, 170, 93, 194, 197, 194, 156, 170, 95, 194, 197, 2, 157, 170, 96, 194, 197, 66, 152, 170, 90, 194, 197, 2, 153, 170, 91, 194, 197, 66, 154, 170, 92, 194, 197, 130, 140, 170, 42, 0, 146, 170, 20, 0, 149, 170, 9, 0, 149, 170, 86, 194, 197, 66, 150, 170, 87, 194, 197, 130, 151, 170, 88, 194, 197, 194, 146, 170, 82, 194, 197, 130, 147, 170, 83, 194, 197, 194, 148, 170, 85, 194, 197, 2, 143, 170, 9, 0, 143, 170, 78, 194, 197, 194, 144, 170, 80, 194, 197, 2, 145, 170, 81, 194, 197, 66, 140, 170, 75, 194, 197, 2, 141, 170, 76, 194, 197, 66, 142, 170, 77, 194, 197, 130, 134, 170, 20, 0, 137, 170, 9, 0, 137, 170, 71, 194, 197, 66, 138, 170, 72, 194, 197, 130, 139, 170, 73, 194, 197, 194, 134, 170, 67, 194, 197, 130, 135, 170, 68, 194, 197, 194, 136, 170, 70, 194, 197, 2, 131, 170, 9, 0, 131, 170, 63, 194, 197, 194, 132, 170, 65, 194, 197, 2, 133, 170, 66, 194, 197, 66, 128, 170, 60, 194, 197, 2, 129, 170, 61, 194, 197, 66, 130, 170, 62, 194, 197, 130, 133, 115, 193, 154, 47, 0, 152, 170, 86, 0, 164, 170, 42, 0, 170, 170, 20, 0, 173, 170, 9, 0, 173, 170, 116, 194, 197, 130, 174, 170, 117, 194, 197, 194, 175, 170, 119, 194, 197, 2, 170, 170, 112, 194, 197, 194, 171, 170, 114, 194, 197, 2, 172, 170, 115, 194, 197, 66, 167, 170, 9, 0, 167, 170, 109, 194, 197, 2, 168, 170, 110, 194, 197, 66, 169, 170, 111, 194, 197, 130, 164, 170, 105, 194, 197, 66, 165, 170, 106, 194, 197, 130, 166, 170, 107, 194, 197, 194, 158, 170, 20, 0, 161, 170, 9, 0, 161, 170, 101, 194, 197, 130, 162, 170, 102, 194, 197, 194, 163, 170, 104, 194, 197, 2, 158, 170, 97, 194, 197, 194, 159, 170, 99, 194, 197, 2, 160, 170, 100, 194, 197, 66, 155, 170, 9, 0, 155, 170, 94, 194, 197, 2, 156, 170, 95, 194, 197, 66, 157, 170, 96, 194, 197, 130, 152, 170, 90, 194, 197, 66, 153, 170, 91, 194, 197, 130, 154, 170, 92, 194, 197, 194, 140, 170, 42, 0, 146, 170, 20, 0, 149, 170, 9, 0, 149, 170, 86, 194, 197, 130, 150, 170, 87, 194, 197, 194, 151, 170, 89, 194, 197, 2, 146, 170, 82, 194, 197, 194, 147, 170, 84, 194, 197, 2, 148, 170, 85, 194, 197, 66, 143, 170, 9, 0, 143, 170, 79, 194, 197, 2, 144, 170, 80, 194, 197, 66, 145, 170, 81, 194, 197, 130, 140, 170, 75, 194, 197, 66, 141, 170, 76, 194, 197, 130, 142, 170, 77, 194, 197, 194, 134, 170, 20, 0, 137, 170, 9, 0, 137, 170, 71, 194, 197, 130, 138, 170, 72, 194, 197, 194, 139, 170, 74, 194, 197, 2, 134, 170, 67, 194, 197, 194, 135, 170, 69, 194, 197, 2, 136, 170, 70, 194, 197, 66, 131, 170, 9, 0, 131, 170, 64, 194, 197, 2, 132, 170, 65, 194, 197, 66, 133, 170, 66, 194, 197, 130, 128, 170, 60, 194, 197, 66, 129, 170, 61, 194, 197, 130, 130, 170, 62, 194, 197, 194, 133, 115, 193, 168, 47, 0, 152, 170, 86, 0, 164, 170, 42, 0, 170, 170, 20, 0, 173, 170, 9, 0, 173, 170, 116, 194, 197, 194, 174, 170, 118, 194, 197, 2, 175, 170, 119, 194, 197, 66, 170, 170, 113, 194, 197, 2, 171, 170, 114, 194, 197, 66, 172, 170, 115, 194, 197, 130, 167, 170, 9, 0, 167, 170, 109, 194, 197, 66, 168, 170, 110, 194, 197, 130, 169, 170, 111, 194, 197, 194, 164, 170, 105, 194, 197, 130, 165, 170, 106, 194, 197, 194, 166, 170, 108, 194, 197, 2, 158, 170, 20, 0, 161, 170, 9, 0, 161, 170, 101, 194, 197, 194, 162, 170, 103, 194, 197, 2, 163, 170, 104, 194, 197, 66, 158, 170, 98, 194, 197, 2, 159, 170, 99, 194, 197, 66, 160, 170, 100, 194, 197, 130, 155, 170, 9, 0, 155, 170, 94, 194, 197, 66, 156, 170, 95, 194, 197, 130, 157, 170, 96, 194, 197, 194, 152, 170, 90, 194, 197, 130, 153, 170, 91, 194, 197, 194, 154, 170, 93, 194, 197, 2, 140, 170, 42, 0, 146, 170, 20, 0, 149, 170, 9, 0, 149, 170, 86, 194, 197, 194, 150, 170, 88, 194, 197, 2, 151, 170, 89, 194, 197, 66, 146, 170, 83, 194, 197, 2, 147, 170, 84, 194, 197, 66, 148, 170, 85, 194, 197, 130, 143, 170, 9, 0, 143, 170, 79, 194, 197, 66, 144, 170, 80, 194, 197, 130, 145, 170, 81, 194, 197, 194, 140, 170, 75, 194, 197, 130, 141, 170, 76, 194, 197, 194, 142, 170, 78, 194, 197, 2, 134, 170, 20, 0, 137, 170, 9, 0, 137, 170, 71, 194, 197, 194, 138, 170, 73, 194, 197, 2, 139, 170, 74, 194, 197, 66, 134, 170, 68, 194, 197, 2, 135, 170, 69, 194, 197, 66, 136, 170, 70, 194, 197, 130, 131, 170, 9, 0, 131, 170, 64, 194, 197, 66, 132, 170, 65, 194, 197, 130, 133, 170, 66, 194, 197, 194, 128, 170, 60, 194, 197, 130, 129, 170, 61, 194, 197, 194, 130, 170, 63, 194, 197, 2, 133, 115, 193, 175, 47, 0, 152, 170, 86, 0, 164, 170, 42, 0, 170, 170, 20, 0, 173, 170, 9, 0, 173, 170, 117, 194, 197, 2, 174, 170, 118, 194, 197, 66, 175, 170, 119, 194, 197, 130, 170, 170, 113, 194, 197, 66, 171, 170, 114, 194, 197, 130, 172, 170, 115, 194, 197, 194, 167, 170, 9, 0, 167, 170, 109, 194, 197, 130, 168, 170, 110, 194, 197, 194, 169, 170, 112, 194, 197, 2, 164, 170, 105, 194, 197, 194, 165, 170, 107, 194, 197, 2, 166, 170, 108, 194, 197, 66, 158, 170, 20, 0, 161, 170, 9, 0, 161, 170, 102, 194, 197, 2, 162, 170, 103, 194, 197, 66, 163, 170, 104, 194, 197, 130, 158, 170, 98, 194, 197, 66, 159, 170, 99, 194, 197, 130, 160, 170, 100, 194, 197, 194, 155, 170, 9, 0, 155, 170, 94, 194, 197, 130, 156, 170, 95, 194, 197, 194, 157, 170, 97, 194, 197, 2, 152, 170, 90, 194, 197, 194, 153, 170, 92, 194, 197, 2, 154, 170, 93, 194, 197, 66, 140, 170, 42, 0, 146, 170, 20, 0, 149, 170, 9, 0, 149, 170, 87, 194, 197, 2, 150, 170, 88, 194, 197, 66, 151, 170, 89, 194, 197, 130, 146, 170, 83, 194, 197, 66, 147, 170, 84, 194, 197, 130, 148, 170, 85, 194, 197, 194, 143, 170, 9, 0, 143, 170, 79, 194, 197, 130, 144, 170, 80, 194, 197, 194, 145, 170, 82, 194, 197, 2, 140, 170, 75, 194, 197, 194, 141, 170, 77, 194, 197, 2, 142, 170, 78, 194, 197, 66, 134, 170, 20, 0, 137, 170, 9, 0, 137, 170, 72, 194, 197, 2, 138, 170, 73, 194, 197, 66, 139, 170, 74, 194, 197, 130, 134, 170, 68, 194, 197, 66, 135, 170, 69, 194, 197, 130, 136, 170, 70, 194, 197, 194, 131, 170, 9, 0, 131, 170, 64, 194, 197, 130, 132, 170, 65, 194, 197, 194, 133, 170, 67, 194, 197, 2, 128, 170, 60, 194, 197, 194, 129, 170, 62, 194, 197, 2, 130, 170, 63, 194, 197, 66, 255, 255, 255, 255, 0, 0, 92, 0, 131, 27, 30, 2, 0, 170, 183, 1, 8, 216, 104, 0, 35, 216, 48, 0, 56, 216, 26, 0, 56, 216, 6, 0, 58, 216, 13, 0, 208, 253, 255, 255, 4, 254, 193, 2, 1, 0, 8, 221, 255, 255, 243, 123, 193, 2, 225, 222, 255, 255, 245, 123, 193, 2, 1, 0, 2, 220, 255, 255, 95, 120, 193, 2, 9, 221, 255, 255, 167, 120, 193, 2, 35, 216, 8, 0, 44, 216, 11, 0, 47, 216, 48, 0, 32, 220, 255, 255, 88, 124, 193, 2, 48, 0, 101, 220, 255, 255, 166, 125, 193, 2, 48, 0, 196, 221, 255, 255, 155, 125, 193, 2, 26, 216, 32, 0, 26, 216, 8, 0, 27, 216, 19, 0, 32, 216, 48, 0, 41, 222, 255, 255, 100, 125, 193, 2, 2, 0, 79, 222, 255, 255, 98, 124, 193, 2, 230, 222, 255, 255, 94, 120, 193, 2, 28, 223, 255, 255, 240, 123, 193, 2, 1, 0, 64, 222, 255, 255, 101, 120, 193, 2, 0, 223, 255, 255, 153, 123, 193, 2, 8, 216, 8, 0, 12, 216, 11, 0, 17, 216, 48, 0, 0, 220, 255, 255, 83, 125, 193, 2, 48, 0, 0, 220, 255, 255, 207, 124, 193, 2, 48, 0, 83, 221, 255, 255, 4, 125, 193, 2, 2, 216, 244, 0, 5, 216, 84, 0, 5, 216, 24, 0, 6, 216, 49, 0, 7, 216, 4, 0, 14, 220, 255, 255, 248, 113, 193, 2, 114, 220, 255, 255, 4, 117, 193, 2, 16, 221, 255, 255, 224, 113, 193, 2, 113, 221, 255, 255, 229, 113, 193, 2, 229, 222, 255, 255, 30, 117, 193, 2, 5, 0, 14, 222, 12, 0, 14, 222, 255, 255, 201, 113, 193, 2, 128, 222, 255, 255, 203, 113, 193, 2, 23, 223, 255, 255, 211, 113, 193, 2, 18, 220, 255, 255, 180, 113, 193, 2, 132, 220, 255, 255, 183, 113, 193, 2, 142, 221, 255, 255, 185, 113, 193, 2, 6, 0, 206, 221, 16, 0, 206, 221, 255, 255, 207, 113, 193, 2, 11, 222, 255, 255, 198, 116, 193, 2, 92, 222, 255, 255, 209, 116, 193, 2, 192, 222, 255, 255, 238, 123, 193, 2, 11, 220, 255, 255, 209, 113, 193, 2, 180, 220, 255, 255, 172, 123, 193, 2, 12, 221, 255, 255, 205, 113, 193, 2, 2, 216, 46, 0, 3, 216, 119, 0, 4, 216, 9, 0, 131, 221, 20, 0, 131, 221, 255, 255, 170, 113, 193, 2, 8, 222, 255, 255, 172, 113, 193, 2, 143, 222, 255, 255, 176, 113, 193, 2, 190, 222, 255, 255, 174, 113, 193, 2, 21, 223, 255, 255, 178, 113, 193, 2, 5, 220, 255, 255, 244, 113, 193, 2, 131, 220, 255, 255, 166, 113, 193, 2, 208, 220, 255, 255, 97, 124, 193, 2, 3, 221, 255, 255, 96, 117, 193, 2, 82, 221, 255, 255, 168, 113, 193, 2, 16, 0, 160, 221, 38, 0, 216, 222, 20, 0, 216, 222, 255, 255, 163, 124, 193, 2, 0, 223, 255, 255, 119, 124, 193, 2, 64, 223, 255, 255, 159, 124, 193, 2, 96, 223, 255, 255, 160, 124, 193, 2, 143, 223, 255, 255, 161, 124, 193, 2, 160, 221, 255, 255, 34, 125, 193, 2, 0, 222, 255, 255, 246, 113, 193, 2, 96, 222, 255, 255, 117, 124, 193, 2, 149, 222, 255, 255, 118, 124, 193, 2, 244, 220, 16, 0, 244, 220, 255, 255, 157, 124, 193, 2, 0, 221, 255, 255, 50, 101, 193, 2, 32, 221, 255, 255, 250, 123, 193, 2, 128, 221, 255, 255, 34, 125, 193, 2, 0, 220, 255, 255, 115, 124, 193, 2, 64, 220, 255, 255, 158, 124, 193, 2, 115, 220, 255, 255, 125, 124, 193, 2, 150, 220, 255, 255, 130, 124, 193, 2, 7, 0, 25, 223, 16, 0, 25, 223, 255, 255, 168, 124, 193, 2, 66, 223, 255, 255, 196, 124, 193, 2, 191, 223, 255, 255, 203, 124, 193, 2, 241, 223, 255, 255, 162, 124, 193, 2, 0, 220, 255, 255, 144, 119, 193, 2, 161, 220, 255, 255, 60, 119, 193, 2, 18, 221, 255, 255, 94, 117, 193, 2, 136, 222, 255, 255, 201, 124, 193, 2, 0, 172, 73, 0, 0, 172, 255, 255, 4, 121, 193, 2, 0, 216, 32, 0, 1, 216, 6, 0, 181, 220, 16, 0, 181, 220, 255, 255, 232, 118, 193, 2, 0, 221, 255, 255, 93, 124, 193, 2, 55, 221, 255, 255, 95, 124, 193, 2, 71, 222, 255, 255, 105, 124, 193, 2, 20, 220, 255, 255, 4, 124, 193, 2, 80, 220, 255, 255, 86, 124, 193, 2, 128, 220, 255, 255, 92, 124, 193, 2, 7, 0, 48, 223, 16, 0, 48, 223, 255, 255, 252, 123, 193, 2, 107, 223, 255, 255, 100, 98, 193, 2, 128, 223, 255, 255, 204, 124, 193, 2, 160, 223, 255, 255, 205, 124, 193, 2, 0, 220, 255, 255, 99, 124, 193, 2, 128, 222, 255, 255, 247, 123, 193, 2, 160, 222, 255, 255, 248, 123, 193, 2, 0, 223, 255, 255, 251, 123, 193, 2, 0, 170, 255, 255, 38, 118, 193, 2, 128, 170, 255, 255, 132, 115, 193, 2, 192, 171, 255, 255, 160, 113, 193, 2, 5, 49, 54, 0, 0, 168, 26, 0, 10, 169, 12, 0, 10, 169, 255, 255, 78, 117, 193, 2, 48, 169, 255, 255, 76, 117, 193, 2, 132, 169, 255, 255, 42, 118, 193, 2, 0, 168, 255, 255, 162, 113, 193, 2, 64, 168, 255, 255, 8, 117, 193, 2, 130, 168, 255, 255, 164, 113, 193, 2, 208, 164, 12, 0, 208, 164, 255, 255, 151, 123, 193, 2, 73, 165, 255, 255, 4, 120, 193, 2, 160, 166, 255, 255, 76, 120, 193, 2, 5, 49, 255, 255, 4, 123, 193, 2, 87, 91, 255, 255, 2, 126, 193, 2, 136, 162, 255, 255, 118, 123, 193, 2, 172, 32, 24, 0, 48, 45, 12, 0, 48, 45, 255, 255, 141, 103, 193, 2, 75, 48, 255, 255, 4, 122, 193, 2, 171, 48, 255, 255, 4, 122, 193, 2, 172, 32, 132, 205, 193, 2, 58, 38, 2, 204, 193, 2, 0, 44, 255, 255, 4, 98, 193, 2, 131, 27, 255, 255, 231, 113, 193, 2, 192, 27, 255, 255, 31, 117, 193, 2, 0, 28, 255, 255, 6, 117, 193, 2, 90, 28, 255, 255, 56, 118, 193, 2, 28, 32, 6, 197, 193, 2, 133, 13, 104, 0, 3, 23, 54, 0, 0, 25, 26, 0, 0, 26, 12, 0, 0, 26, 255, 255, 29, 117, 193, 2, 32, 26, 255, 255, 18, 118, 193, 2, 5, 27, 255, 255, 40, 118, 193, 2, 0, 25, 255, 255, 10, 117, 193, 2, 80, 25, 255, 255, 4, 118, 193, 2, 128, 25, 255, 255, 5, 118, 193, 2, 99, 23, 12, 0, 99, 23, 255, 255, 28, 117, 193, 2, 128, 23, 255, 255, 98, 117, 193, 2, 38, 24, 255, 255, 52, 118, 193, 2, 3, 23, 255, 255, 25, 117, 193, 2, 35, 23, 255, 255, 26, 117, 193, 2, 67, 23, 255, 255, 27, 117, 193, 2, 211, 16, 26, 0, 192, 20, 12, 0, 192, 20, 255, 255, 234, 118, 193, 2, 143, 22, 255, 255, 254, 118, 193, 2, 160, 22, 255, 255, 4, 119, 193, 2, 211, 16, 255, 255, 4, 99, 193, 2, 160, 18, 255, 255, 143, 103, 193, 2, 196, 19, 255, 255, 58, 118, 193, 2, 133, 13, 255, 255, 4, 113, 193, 2, 23, 14, 255, 255, 4, 114, 193, 2, 165, 14, 255, 255, 4, 115, 193, 2, 64, 15, 255, 255, 4, 116, 193, 2, 0, 16, 255, 255, 79, 117, 193, 2, 202, 7, 54, 0, 149, 10, 26, 0, 21, 12, 12, 0, 21, 12, 255, 255, 4, 110, 193, 2, 149, 12, 255, 255, 4, 111, 193, 2, 21, 13, 255, 255, 4, 112, 193, 2, 149, 10, 255, 255, 4, 107, 193, 2, 21, 11, 255, 255, 4, 108, 193, 2, 149, 11, 255, 255, 4, 109, 193, 2, 5, 9, 12, 0, 5, 9, 255, 255, 4, 104, 193, 2, 149, 9, 255, 255, 4, 105, 193, 2, 21, 10, 255, 255, 4, 106, 193, 2, 202, 7, 255, 255, 130, 103, 193, 2, 0, 8, 255, 255, 51, 101, 193, 2, 64, 8, 255, 255, 27, 103, 193, 2, 47, 4, 26, 0, 40, 6, 12, 0, 40, 6, 255, 255, 4, 102, 193, 2, 16, 7, 255, 255, 4, 103, 193, 2, 140, 7, 255, 255, 28, 103, 193, 2, 47, 4, 255, 255, 4, 97, 193, 2, 49, 5, 255, 255, 4, 100, 193, 2, 208, 5, 255, 255, 4, 101, 193, 2, 52, 0, 2, 206, 193, 2, 76, 0, 2, 233, 193, 2, 160, 0, 2, 195, 193, 2, 169, 3, 255, 255, 4, 96, 193, 2, 226, 3, 255, 255, 96, 96, 193, 2, 97, 117, 193, 91, 49, 0, 4, 216, 39, 221, 255, 255, 97, 117, 193, 70, 97, 117, 193, 98, 49, 0, 4, 216, 39, 221, 255, 255, 97, 117, 193, 77, 179, 113, 193, 182, 48, 0, 4, 216, 1, 0, 62, 223, 255, 255, 179, 113, 193, 196, 87, 223, 255, 255, 179, 113, 193, 203, 184, 113, 193, 175, 48, 0, 5, 216, 2, 0, 176, 220, 255, 255, 184, 113, 193, 196, 186, 220, 255, 255, 184, 113, 193, 189, 189, 220, 255, 255, 184, 113, 193, 210, 200, 113, 193, 16, 49, 0, 5, 216, 175, 221, 255, 255, 200, 113, 193, 30, 200, 113, 193, 23, 49, 0, 5, 216, 175, 221, 255, 255, 200, 113, 193, 37, 206, 113, 193, 112, 49, 0, 6, 216, 48, 221, 255, 255, 206, 113, 193, 126, 80, 130, 224, 0, 52, 0, 53, 0, 76, 0, 77, 0, 160, 0, 161, 0, 6, 3, 7, 3, 169, 3, 170, 3, 226, 3, 227, 3, 47, 4, 48, 4, 49, 5, 50, 5, 208, 5, 209, 5, 40, 6, 41, 6, 83, 6, 86, 6, 16, 7, 17, 7, 140, 7, 141, 7, 202, 7, 203, 7, 0, 8, 1, 8, 64, 8, 65, 8, 5, 9, 6, 9, 149, 9, 150, 9, 190, 9, 191, 9, 215, 9, 216, 9, 21, 10, 22, 10, 149, 10, 150, 10, 21, 11, 22, 11, 62, 11, 63, 11, 86, 11, 88, 11, 149, 11, 150, 11, 190, 11, 191, 11, 215, 11, 216, 11, 21, 12, 22, 12, 86, 12, 87, 12, 149, 12, 150, 12, 194, 12, 195, 12, 213, 12, 215, 12, 21, 13, 22, 13, 62, 13, 63, 13, 87, 13, 88, 13, 133, 13, 134, 13, 202, 13, 203, 13, 207, 13, 208, 13, 223, 13, 224, 13, 1, 14, 47, 14, 50, 14, 51, 14, 129, 14, 131, 14, 132, 14, 133, 14, 134, 14, 139, 14, 140, 14, 164, 14, 165, 14, 166, 14, 167, 14, 175, 14, 178, 14, 179, 14, 220, 14, 224, 14, 64, 15, 65, 15, 113, 15, 118, 15, 128, 15, 130, 15, 0, 16, 1, 16, 46, 16, 47, 16, 211, 16, 212, 16, 160, 18, 161, 18, 196, 19, 197, 19, 192, 20, 193, 20, 143, 22, 144, 22, 160, 22, 161, 22, 3, 23, 4, 23, 35, 23, 36, 23, 67, 23, 68, 23, 99, 23, 100, 23, 128, 23, 129, 23, 38, 24, 39, 24, 0, 25, 1, 25, 80, 25, 81, 25, 128, 25, 172, 25, 0, 26, 1, 26, 32, 26, 33, 26, 5, 27, 6, 27, 53, 27, 54, 27, 131, 27, 132, 27, 192, 27, 193, 27, 0, 28, 1, 28, 90, 28, 91, 28, 28, 32, 29, 32, 172, 32, 173, 32, 58, 38, 59, 38, 0, 44, 1, 44, 48, 45, 49, 45, 75, 48, 76, 48, 171, 48, 172, 48, 5, 49, 6, 49, 87, 91, 88, 91, 136, 162, 137, 162, 208, 164, 209, 164, 73, 165, 74, 165, 160, 166, 161, 166, 0, 168, 1, 168, 64, 168, 65, 168, 130, 168, 131, 168, 10, 169, 11, 169, 48, 169, 49, 169, 132, 169, 133, 169, 0, 170, 1, 170, 128, 170, 176, 170, 192, 171, 193, 171, 0, 172, 1, 172, 0, 216, 9, 216, 12, 216, 13, 216, 17, 216, 18, 216, 26, 216, 28, 216, 32, 216, 33, 216, 35, 216, 36, 216, 44, 216, 45, 216, 47, 216, 48, 216, 56, 216, 57, 216, 58, 216, 59, 216, 208, 253, 209, 253, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 128, 2, 1, 0, 129, 2, 1, 0, 160, 2, 1, 0, 161, 2, 1, 0, 0, 3, 1, 0, 1, 3, 1, 0, 48, 3, 1, 0, 49, 3, 1, 0, 107, 3, 1, 0, 108, 3, 1, 0, 128, 3, 1, 0, 129, 3, 1, 0, 160, 3, 1, 0, 161, 3, 1, 0, 20, 4, 1, 0, 21, 4, 1, 0, 80, 4, 1, 0, 81, 4, 1, 0, 128, 4, 1, 0, 129, 4, 1, 0, 181, 4, 1, 0, 182, 4, 1, 0, 0, 5, 1, 0, 1, 5, 1, 0, 55, 5, 1, 0, 56, 5, 1, 0, 71, 6, 1, 0, 72, 6, 1, 0, 0, 8, 1, 0, 1, 8, 1, 0, 64, 8, 1, 0, 65, 8, 1, 0, 115, 8, 1, 0, 116, 8, 1, 0, 150, 8, 1, 0, 151, 8, 1, 0, 244, 8, 1, 0, 245, 8, 1, 0, 0, 9, 1, 0, 1, 9, 1, 0, 32, 9, 1, 0, 33, 9, 1, 0, 128, 9, 1, 0, 129, 9, 1, 0, 160, 9, 1, 0, 161, 9, 1, 0, 0, 10, 1, 0, 1, 10, 1, 0, 96, 10, 1, 0, 97, 10, 1, 0, 149, 10, 1, 0, 150, 10, 1, 0, 216, 10, 1, 0, 217, 10, 1, 0, 0, 11, 1, 0, 1, 11, 1, 0, 64, 11, 1, 0, 65, 11, 1, 0, 96, 11, 1, 0, 97, 11, 1, 0, 143, 11, 1, 0, 144, 11, 1, 0, 0, 12, 1, 0, 1, 12, 1, 0, 161, 12, 1, 0, 162, 12, 1, 0, 18, 13, 1, 0, 19, 13, 1, 0, 136, 14, 1, 0, 137, 14, 1, 0, 25, 15, 1, 0, 26, 15, 1, 0, 66, 15, 1, 0, 67, 15, 1, 0, 191, 15, 1, 0, 192, 15, 1, 0, 241, 15, 1, 0, 242, 15, 1, 0, 5, 16, 1, 0, 6, 16, 1, 0, 131, 16, 1, 0, 132, 16, 1, 0, 208, 16, 1, 0, 209, 16, 1, 0, 3, 17, 1, 0, 4, 17, 1, 0, 39, 17, 1, 0, 40, 17, 1, 0, 82, 17, 1, 0, 83, 17, 1, 0, 131, 17, 1, 0, 132, 17, 1, 0, 8, 18, 1, 0, 9, 18, 1, 0, 143, 18, 1, 0, 144, 18, 1, 0, 190, 18, 1, 0, 191, 18, 1, 0, 21, 19, 1, 0, 22, 19, 1, 0, 62, 19, 1, 0, 63, 19, 1, 0, 87, 19, 1, 0, 88, 19, 1, 0, 18, 20, 1, 0, 19, 20, 1, 0, 132, 20, 1, 0, 133, 20, 1, 0, 176, 20, 1, 0, 177, 20, 1, 0, 186, 20, 1, 0, 187, 20, 1, 0, 189, 20, 1, 0, 190, 20, 1, 0, 142, 21, 1, 0, 143, 21, 1, 0, 175, 21, 1, 0, 176, 21, 1, 0, 14, 22, 1, 0, 15, 22, 1, 0, 128, 22, 1, 0, 129, 22, 1, 0, 23, 23, 1, 0, 24, 23, 1, 0, 11, 24, 1, 0, 12, 24, 1, 0, 180, 24, 1, 0, 181, 24, 1, 0, 12, 25, 1, 0, 13, 25, 1, 0, 48, 25, 1, 0, 49, 25, 1, 0, 206, 25, 1, 0, 207, 25, 1, 0, 11, 26, 1, 0, 12, 26, 1, 0, 92, 26, 1, 0, 93, 26, 1, 0, 192, 26, 1, 0, 193, 26, 1, 0, 14, 28, 1, 0, 15, 28, 1, 0, 114, 28, 1, 0, 115, 28, 1, 0, 16, 29, 1, 0, 17, 29, 1, 0, 113, 29, 1, 0, 114, 29, 1, 0, 229, 30, 1, 0, 230, 30, 1, 0, 0, 32, 1, 0, 1, 32, 1, 0, 83, 49, 1, 0, 84, 49, 1, 0, 0, 68, 1, 0, 1, 68, 1, 0, 79, 106, 1, 0, 80, 106, 1, 0, 230, 106, 1, 0, 231, 106, 1, 0, 28, 107, 1, 0, 29, 107, 1, 0, 64, 110, 1, 0, 65, 110, 1, 0, 0, 111, 1, 0, 1, 111, 1, 0, 41, 130, 1, 0, 42, 130, 1, 0, 101, 140, 1, 0, 102, 140, 1, 0, 196, 177, 1, 0, 197, 177, 1, 0, 32, 188, 1, 0, 33, 188, 1, 0, 8, 225, 1, 0, 9, 225, 1, 0, 225, 226, 1, 0, 226, 226, 1, 0, 2, 232, 1, 0, 3, 232, 1, 0, 9, 233, 1, 0, 10, 233, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 186, 72, 185, 76, 185, 80, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 185, 120, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 185, 152, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 168, 72, 168, 76, 168, 80, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 168, 120, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 168, 152, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 154, 114, 185, 121, 153, 121, 249, 121, 153, 122, 89, 122, 136, 14, 217, 122, 185, 153, 153, 153, 249, 153, 89, 154, 153, 169, 185, 176, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 136, 121, 232, 121, 136, 122, 72, 122, 128, 14, 200, 122, 168, 153, 136, 153, 232, 153, 72, 154, 136, 169, 168, 176, 72, 170, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 250, 114, 232, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 57, 123, 40, 123, 217, 121, 200, 121, 121, 122, 104, 122, 12, 8, 14, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 250, 142, 232, 142, 26, 142, 8, 142, 249, 146, 232, 146, 25, 146, 8, 146, 185, 148, 168, 148, 153, 154, 136, 154, 57, 155, 40, 155, 217, 153, 200, 153, 57, 154, 40, 154, 121, 154, 104, 154, 25, 155, 8, 155, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 185, 174, 168, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 1, 193, 0, 0, 0, 0, 0, 19, 0, 15, 0, 27, 0, 107, 0, 92, 0, 10, 0, 11, 0, 121, 0, 26, 0, 25, 0, 14, 0, 120, 0, 9, 0, 29, 0, 28, 0, 157, 0, 105, 0, 16, 0, 106, 0, 33, 0, 106, 0, 83, 0, 62, 0, 7, 0, 34, 0, 90, 0, 80, 0, 95, 0, 119, 0, 30, 0, 96, 0, 35, 0, 20, 0, 31, 0, 32, 0, 22, 0, 61, 0, 64, 0, 94, 0, 108, 0, 71, 0, 72, 0, 73, 0, 74, 0, 0, 0, 131, 0, 70, 0, 129, 0, 124, 0, 122, 0, 84, 0, 148, 0, 106, 0, 75, 0, 12, 0, 59, 0, 37, 0, 85, 0, 24, 0, 149, 0, 88, 0, 77, 0, 0, 0, 58, 0, 87, 0, 0, 0, 0, 0, 0, 0, 0, 0, 151, 0, 0, 0, 157, 0, 157, 0, 113, 0, 97, 0, 0, 0, 89, 0, 79, 0, 0, 0, 0, 0, 68, 0, 130, 0, 21, 0, 0, 0, 152, 0, 23, 0, 98, 0, 13, 0, 69, 0, 17, 0, 110, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 99, 0, 0, 0, 150, 0, 0, 0, 158, 0, 117, 0, 0, 0, 86, 0, 116, 0, 118, 0, 91, 0, 78, 0, 38, 0, 0, 0, 57, 0, 0, 0, 36, 0, 138, 0, 134, 0, 82, 0, 0, 0, 39, 0, 143, 0, 140, 0, 141, 0, 0, 0, 139, 0, 18, 0, 63, 0, 0, 0, 0, 0, 100, 0, 109, 0, 0, 0, 132, 0, 101, 0, 123, 0, 125, 0, 45, 0, 0, 0, 0, 0, 102, 0, 152, 0, 133, 0, 136, 0, 135, 0, 43, 0, 111, 0, 0, 0, 0, 0, 128, 0, 155, 0, 41, 0, 127, 0, 50, 0, 154, 0, 0, 0, 153, 0, 42, 0, 47, 0, 126, 0, 40, 0, 54, 0, 137, 0, 49, 0, 44, 0, 112, 0, 48, 0, 104, 0, 60, 0, 67, 0, 46, 0, 93, 0, 0, 0, 0, 0, 0, 0, 55, 0, 66, 0, 65, 0, 53, 0, 56, 0, 76, 0, 103, 0, 81, 0, 145, 0, 144, 0, 142, 0, 114, 0, 52, 0, 115, 0, 147, 0, 51, 0, 156, 0, 146, 0, 1, 0, 2, 0, 3, 0, 4, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 8, 0, 0, 0, 0, 3, 6, 5, 0, 12, 132, 13, 0, 14, 0, 39, 0, 41, 0, 94, 0, 96, 96, 96, 0, 97, 0, 98, 100, 98, 0, 99, 0, 100, 0, 101, 50, 101, 51, 101, 0, 102, 0, 103, 27, 103, 28, 103, 130, 103, 141, 103, 143, 103, 0, 104, 0, 105, 0, 106, 0, 107, 0, 108, 0, 109, 0, 110, 0, 111, 0, 112, 0, 113, 160, 113, 162, 113, 164, 113, 166, 113, 168, 113, 170, 113, 172, 113, 174, 113, 176, 113, 178, 113, 180, 113, 183, 113, 185, 113, 201, 113, 203, 113, 205, 113, 207, 113, 209, 113, 211, 113, 224, 113, 229, 113, 231, 113, 244, 113, 246, 113, 248, 113, 0, 114, 0, 115, 132, 115, 0, 116, 198, 116, 209, 116, 0, 117, 6, 117, 8, 117, 10, 117, 25, 117, 26, 117, 27, 117, 28, 117, 29, 117, 30, 117, 31, 117, 76, 117, 78, 117, 79, 117, 94, 117, 96, 117, 98, 117, 0, 118, 5, 118, 18, 118, 38, 118, 40, 118, 42, 118, 52, 118, 56, 118, 58, 118, 232, 118, 234, 118, 254, 118, 0, 119, 60, 119, 144, 119, 0, 120, 76, 120, 94, 120, 95, 120, 101, 120, 167, 120, 0, 121, 0, 122, 0, 123, 118, 123, 151, 123, 153, 123, 172, 123, 238, 123, 240, 123, 243, 123, 245, 123, 247, 123, 248, 123, 250, 123, 251, 123, 252, 123, 0, 124, 86, 124, 88, 124, 92, 124, 93, 124, 95, 124, 97, 124, 98, 124, 99, 124, 105, 124, 115, 124, 117, 124, 118, 124, 119, 124, 125, 124, 130, 124, 157, 124, 158, 124, 159, 124, 160, 124, 161, 124, 162, 124, 163, 124, 168, 124, 196, 124, 201, 124, 203, 124, 204, 124, 205, 124, 207, 124, 0, 125, 34, 125, 83, 125, 100, 125, 155, 125, 166, 125, 0, 126, 0, 254, 0, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 67, 7, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 71, 7, 0, 0, 71, 7, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 48, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 46, 220, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 65, 0, 114, 0, 97, 0, 98, 0, 93, 0, 38, 0, 39, 6, 60, 0, 213, 6, 60, 0, 40, 6, 38, 0, 67, 6, 60, 0, 175, 6, 60, 0, 173, 6, 60, 0, 68, 6, 38, 0, 190, 6, 60, 0, 72, 6, 60, 0, 199, 6, 60, 0, 198, 6, 60, 0, 200, 6, 60, 0, 203, 6, 60, 0, 208, 6, 60, 0, 73, 6, 60, 0, 74, 6, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 28, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 15, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 60, 0, 0, 0, 64, 0, 0, 0, 64, 1, 0, 0, 160, 27, 0, 0, 160, 27, 0, 0, 160, 27, 0, 0, 160, 27, 0, 0, 236, 27, 0, 0, 236, 27, 0, 0, 8, 28, 0, 0, 2, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 39, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 157, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 95, 0, 103, 0, 100, 0, 108, 0, 116, 0, 124, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 148, 0, 140, 0, 48, 0, 48, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 132, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 124, 1, 144, 1, 208, 1, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 1, 0, 202, 65, 1, 0, 202, 98, 1, 0, 202, 131, 1, 0, 202, 164, 1, 0, 202, 197, 1, 0, 202, 230, 1, 0, 202, 7, 2, 0, 202, 40, 2, 0, 202, 73, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 3, 27, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 4, 183, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 0, 0, 193, 8, 217, 102, 201, 230, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 3, 183, 102, 192, 0, 0, 0, 193, 2, 183, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 4, 217, 102, 193, 3, 217, 102, 193, 5, 217, 102, 192, 0, 0, 0, 192, 0, 0, 0, 193, 6, 217, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 7, 217, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 2, 27, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 193, 2, 27, 102, 194, 5, 128, 232, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 217, 102, 193, 2, 48, 0, 84, 6, 255, 255, 16, 102, 5, 5, 217, 102, 193, 9, 48, 0, 84, 6, 255, 255, 22, 102, 5, 5, 3, 0, 69, 0, 60, 0, 32, 0, 51, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 60, 7, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 65, 7, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17, 7, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 21, 7, 0, 0, 21, 7, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 42, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 67, 0, 121, 0, 114, 0, 108, 0, 93, 0, 38, 0, 19, 4, 60, 0, 145, 4, 60, 0, 60, 0, 60, 0, 144, 4, 38, 0, 71, 166, 60, 0, 87, 4, 60, 0, 60, 0, 60, 0, 118, 166, 60, 0, 60, 0, 60, 0, 7, 4, 0, 0, 170, 170, 170, 170, 170, 170, 120, 27, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 15, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 60, 0, 0, 0, 64, 0, 0, 0, 64, 1, 0, 0, 0, 27, 0, 0, 0, 27, 0, 0, 0, 27, 0, 0, 0, 27, 0, 0, 68, 27, 0, 0, 68, 27, 0, 0, 96, 27, 0, 0, 8, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 39, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 147, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 91, 0, 99, 0, 103, 0, 111, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 130, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 138, 0, 127, 0, 48, 0, 48, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 119, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 108, 1, 156, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 0, 0, 202, 1, 1, 0, 202, 34, 1, 0, 202, 67, 1, 0, 202, 100, 1, 0, 202, 133, 1, 0, 202, 166, 1, 0, 202, 199, 1, 0, 202, 232, 1, 0, 202, 9, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 0, 0, 135, 5, 77, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 0, 0, 5, 5, 77, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 17, 97, 5, 5, 17, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 77, 97, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 74, 97, 174, 5, 48, 0, 8, 3, 255, 255, 77, 97, 135, 5, 74, 97, 5, 5, 48, 0, 8, 3, 255, 255, 77, 97, 5, 5, 3, 0, 69, 0, 60, 0, 32, 0, 43, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 10, 7, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 15, 7, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 55, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 59, 8, 0, 0, 59, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 151, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 252, 220, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 65, 0, 114, 0, 97, 0, 98, 0, 93, 0, 38, 0, 39, 6, 60, 0, 60, 0, 35, 6, 60, 0, 34, 6, 60, 0, 40, 6, 60, 0, 40, 6, 190, 6, 60, 0, 126, 6, 60, 0, 126, 6, 190, 6, 60, 0, 42, 6, 60, 0, 42, 6, 190, 6, 60, 0, 121, 6, 60, 0, 121, 6, 190, 6, 60, 0, 43, 6, 60, 0, 44, 6, 60, 0, 44, 6, 190, 6, 60, 0, 134, 6, 60, 0, 134, 6, 190, 6, 60, 0, 45, 6, 60, 0, 46, 6, 60, 0, 47, 6, 60, 0, 47, 6, 190, 6, 60, 0, 136, 6, 60, 0, 136, 6, 190, 6, 60, 0, 48, 6, 60, 0, 49, 6, 60, 0, 49, 6, 190, 6, 60, 0, 145, 6, 60, 0, 145, 6, 190, 6, 60, 0, 50, 6, 60, 0, 152, 6, 60, 0, 51, 6, 60, 0, 52, 6, 60, 0, 53, 6, 60, 0, 54, 6, 60, 0, 55, 6, 60, 0, 56, 6, 60, 0, 57, 6, 60, 0, 58, 6, 60, 0, 65, 6, 60, 0, 66, 6, 60, 0, 169, 6, 60, 0, 169, 6, 190, 6, 60, 0, 175, 6, 60, 0, 175, 6, 190, 6, 60, 0, 68, 6, 60, 0, 68, 6, 190, 6, 60, 0, 69, 6, 60, 0, 69, 6, 190, 6, 60, 0, 70, 6, 60, 0, 70, 6, 190, 6, 60, 0, 186, 6, 60, 0, 186, 6, 190, 6, 60, 0, 72, 6, 60, 0, 60, 0, 36, 6, 60, 0, 72, 6, 190, 6, 60, 0, 193, 6, 60, 0, 60, 0, 194, 6, 60, 0, 190, 6, 60, 0, 195, 6, 60, 0, 33, 6, 60, 0, 204, 6, 60, 0, 60, 0, 38, 6, 60, 0, 204, 6, 190, 6, 60, 0, 210, 6, 60, 0, 60, 0, 211, 6, 38, 0, 82, 6, 60, 0, 60, 0, 78, 6, 60, 0, 60, 0, 80, 6, 60, 0, 60, 0, 79, 6, 60, 0, 60, 0, 112, 6, 60, 0, 60, 0, 86, 6, 60, 0, 60, 0, 87, 6, 60, 0, 60, 0, 75, 6, 60, 0, 60, 0, 77, 6, 60, 0, 60, 0, 76, 6, 60, 0, 60, 0, 84, 6, 60, 0, 60, 0, 81, 6, 60, 0, 60, 0, 88, 6, 60, 0, 60, 0, 83, 6, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 116, 0, 101, 0, 114, 0, 116, 0, 105, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 60, 0, 60, 0, 60, 0, 16, 6, 60, 0, 60, 0, 60, 0, 17, 6, 60, 0, 60, 0, 60, 0, 19, 6, 60, 0, 60, 0, 60, 0, 18, 6, 60, 0, 60, 0, 60, 0, 20, 6, 0, 0, 94, 30, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 16, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 68, 0, 0, 0, 68, 1, 0, 0, 100, 28, 0, 0, 100, 28, 0, 0, 172, 28, 0, 0, 172, 28, 0, 0, 248, 28, 0, 0, 248, 28, 0, 0, 60, 30, 0, 0, 66, 30, 0, 0, 2, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 39, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 169, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 96, 0, 104, 0, 112, 0, 120, 0, 128, 0, 136, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 160, 0, 152, 0, 48, 0, 48, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 144, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 128, 1, 192, 1, 0, 2, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 1, 0, 202, 65, 1, 0, 202, 98, 1, 0, 202, 131, 1, 0, 202, 164, 1, 0, 202, 197, 1, 0, 202, 230, 1, 0, 202, 7, 2, 0, 202, 40, 2, 0, 202, 73, 2, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 0, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 0, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 0, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 198, 33, 0, 0, 198, 65, 0, 0, 198, 97, 0, 0, 198, 129, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 56, 27, 102, 193, 2, 27, 102, 5, 70, 26, 102, 198, 161, 0, 0, 192, 0, 0, 0, 198, 193, 0, 0, 201, 6, 0, 0, 201, 224, 1, 0, 192, 0, 0, 0, 201, 192, 2, 0, 193, 11, 27, 102, 201, 160, 3, 0, 193, 16, 27, 102, 193, 17, 27, 102, 201, 128, 4, 0, 193, 22, 27, 102, 201, 96, 5, 0, 193, 27, 27, 102, 193, 29, 27, 102, 193, 30, 27, 102, 193, 31, 27, 102, 193, 32, 27, 102, 193, 33, 27, 102, 193, 34, 27, 102, 193, 35, 27, 102, 193, 36, 27, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 37, 27, 102, 193, 38, 27, 102, 192, 0, 0, 0, 201, 64, 6, 0, 201, 32, 7, 0, 201, 0, 8, 0, 192, 0, 0, 0, 201, 228, 8, 0, 192, 0, 0, 0, 201, 38, 10, 0, 194, 5, 69, 232, 194, 5, 71, 232, 194, 5, 70, 232, 194, 5, 63, 232, 194, 5, 65, 232, 194, 5, 64, 232, 194, 5, 73, 232, 192, 0, 0, 0, 194, 5, 75, 232, 194, 5, 72, 232, 192, 0, 0, 0, 194, 5, 67, 232, 194, 5, 68, 232, 194, 5, 74, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 66, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 224, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 192, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 160, 12, 0, 192, 0, 0, 0, 201, 128, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 96, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 28, 27, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 64, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 32, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 0, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 54, 27, 102, 192, 0, 0, 0, 197, 34, 0, 0, 201, 230, 17, 0, 198, 225, 0, 0, 193, 55, 27, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 160, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 19, 0, 198, 1, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 2, 188, 0, 0, 0, 0, 0, 0, 3, 188, 0, 0, 0, 0, 0, 0, 5, 188, 0, 0, 0, 0, 0, 0, 4, 188, 0, 0, 0, 0, 0, 0, 6, 188, 0, 0, 0, 0, 0, 0, 0, 5, 0, 70, 0, 51, 27, 102, 0, 5, 0, 70, 0, 57, 27, 102, 0, 5, 0, 70, 0, 53, 27, 102, 0, 5, 0, 70, 0, 59, 27, 102, 0, 0, 0, 0, 193, 16, 220, 102, 194, 5, 72, 232, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 27, 102, 193, 2, 84, 6, 255, 255, 26, 102, 5, 70, 85, 6, 255, 255, 18, 102, 5, 5, 27, 102, 193, 3, 48, 0, 190, 6, 255, 255, 27, 102, 193, 4, 27, 102, 193, 7, 48, 0, 190, 6, 255, 255, 27, 102, 193, 8, 27, 102, 193, 12, 48, 0, 190, 6, 255, 255, 27, 102, 193, 13, 27, 102, 193, 18, 48, 0, 190, 6, 255, 255, 27, 102, 193, 19, 27, 102, 193, 23, 48, 0, 190, 6, 255, 255, 27, 102, 193, 24, 27, 102, 193, 43, 48, 0, 190, 6, 255, 255, 27, 102, 193, 44, 27, 102, 193, 45, 48, 0, 190, 6, 255, 255, 27, 102, 193, 46, 27, 102, 193, 47, 48, 0, 190, 6, 255, 255, 27, 102, 193, 48, 27, 102, 193, 51, 1, 0, 84, 6, 0, 192, 198, 161, 190, 6, 255, 255, 27, 102, 193, 52, 238, 102, 5, 5, 48, 0, 84, 6, 0, 192, 198, 193, 27, 102, 193, 9, 48, 0, 190, 6, 255, 255, 27, 102, 193, 10, 27, 102, 193, 5, 48, 0, 190, 6, 255, 255, 27, 102, 193, 6, 27, 102, 193, 14, 48, 0, 190, 6, 255, 255, 27, 102, 193, 15, 27, 102, 193, 20, 48, 0, 190, 6, 255, 255, 27, 102, 193, 21, 27, 102, 193, 25, 48, 0, 190, 6, 255, 255, 27, 102, 193, 26, 27, 102, 193, 39, 48, 0, 190, 6, 255, 255, 27, 102, 193, 40, 27, 102, 193, 41, 48, 0, 190, 6, 255, 255, 27, 102, 193, 42, 27, 102, 193, 49, 48, 0, 190, 6, 255, 255, 27, 102, 193, 50, 27, 102, 193, 53, 48, 0, 84, 6, 0, 192, 198, 225, 27, 102, 193, 57, 48, 0, 190, 6, 255, 255, 27, 102, 193, 58, 27, 102, 193, 59, 48, 0, 84, 6, 1, 192, 198, 1, 2, 0, 190, 6, 191, 6, 170, 170, 3, 0, 69, 0, 60, 0, 32, 0, 151, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 48, 8, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 53, 8, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 108, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 112, 8, 0, 0, 112, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 66, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 82, 220, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 146, 2, 60, 0, 111, 0, 187, 2, 61, 0, 111, 0, 24, 32, 61, 0, 111, 0, 39, 0, 39, 0, 60, 0, 60, 0, 60, 0, 79, 0, 187, 2, 61, 0, 79, 0, 24, 32, 61, 0, 79, 0, 39, 0, 39, 0, 60, 0, 103, 0, 187, 2, 61, 0, 103, 0, 24, 32, 61, 0, 103, 0, 39, 0, 39, 0, 60, 0, 60, 0, 60, 0, 71, 0, 187, 2, 61, 0, 71, 0, 24, 32, 61, 0, 71, 0, 39, 0, 39, 0, 60, 0, 115, 0, 104, 0, 60, 0, 60, 0, 60, 0, 83, 0, 104, 0, 60, 0, 60, 0, 60, 0, 83, 0, 72, 0, 60, 0, 99, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 72, 0, 0, 0, 170, 170, 170, 170, 132, 32, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 100, 23, 0, 0, 100, 23, 0, 0, 116, 25, 0, 0, 116, 25, 0, 0, 56, 26, 0, 0, 56, 26, 0, 0, 18, 28, 0, 0, 56, 28, 0, 0, 104, 32, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 105, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 88, 0, 48, 0, 48, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 80, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 224, 4, 0, 202, 1, 5, 0, 202, 34, 5, 0, 202, 67, 5, 0, 202, 100, 5, 0, 202, 133, 5, 0, 202, 166, 5, 0, 202, 199, 5, 0, 202, 232, 5, 0, 202, 9, 6, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 201, 0, 0, 0, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 201, 64, 7, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 192, 8, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 64, 10, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 201, 192, 17, 0, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 201, 192, 21, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 160, 23, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 128, 25, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 4, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 4, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 4, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 135, 0, 5, 0, 5, 25, 93, 0, 70, 0, 5, 0, 5, 25, 93, 0, 135, 0, 5, 0, 5, 25, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 25, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 5, 0, 5, 25, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 25, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 135, 0, 5, 0, 5, 25, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 25, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 135, 0, 5, 0, 5, 25, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 25, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 135, 0, 5, 0, 5, 25, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 25, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 135, 0, 5, 0, 5, 25, 93, 0, 5, 0, 160, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 25, 93, 0, 5, 0, 160, 0, 0, 0, 0, 0, 135, 0, 5, 0, 5, 25, 93, 0, 5, 0, 206, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 25, 93, 0, 5, 0, 206, 0, 0, 0, 0, 0, 70, 0, 5, 0, 5, 25, 93, 0, 5, 0, 210, 0, 0, 0, 0, 0, 70, 0, 5, 0, 3, 25, 93, 0, 70, 0, 5, 0, 2, 25, 93, 0, 135, 0, 5, 0, 4, 25, 93, 0, 70, 0, 5, 0, 4, 25, 93, 0, 135, 0, 5, 0, 4, 25, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 25, 93, 0, 5, 0, 142, 0, 0, 0, 0, 0, 135, 0, 5, 0, 4, 25, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 25, 93, 0, 5, 0, 144, 0, 0, 0, 0, 0, 135, 0, 5, 0, 4, 25, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 25, 93, 0, 5, 0, 156, 0, 0, 0, 0, 0, 135, 0, 5, 0, 4, 25, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 25, 93, 0, 5, 0, 196, 0, 0, 0, 0, 0, 135, 0, 5, 0, 4, 25, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 25, 93, 0, 5, 0, 150, 0, 0, 0, 0, 0, 135, 0, 5, 0, 4, 25, 93, 0, 5, 0, 160, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 25, 93, 0, 5, 0, 160, 0, 0, 0, 0, 0, 135, 0, 5, 0, 4, 25, 93, 0, 5, 0, 206, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 25, 93, 0, 5, 0, 206, 0, 0, 0, 0, 0, 70, 0, 5, 0, 4, 25, 93, 0, 5, 0, 210, 0, 0, 0, 0, 0, 0, 0, 0, 193, 5, 25, 93, 5, 142, 0, 0, 193, 5, 25, 93, 5, 144, 0, 0, 193, 5, 25, 93, 5, 156, 0, 0, 193, 5, 25, 93, 5, 196, 0, 0, 193, 5, 25, 93, 5, 150, 0, 0, 193, 5, 25, 93, 5, 160, 0, 0, 193, 5, 25, 93, 5, 206, 0, 0, 193, 5, 25, 93, 5, 210, 0, 0, 193, 4, 25, 93, 5, 142, 0, 0, 193, 4, 25, 93, 5, 144, 0, 0, 193, 4, 25, 93, 5, 156, 0, 0, 193, 4, 25, 93, 5, 196, 0, 0, 193, 4, 25, 93, 5, 150, 0, 0, 193, 4, 25, 93, 5, 160, 0, 0, 193, 4, 25, 93, 5, 206, 0, 0, 193, 4, 25, 93, 5, 210, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 46, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 2, 192, 198, 194, 41, 30, 3, 192, 198, 2, 42, 30, 3, 192, 198, 66, 43, 30, 3, 192, 198, 130, 150, 30, 3, 192, 198, 194, 36, 30, 1, 192, 198, 194, 37, 30, 2, 192, 198, 2, 38, 30, 2, 192, 198, 66, 39, 30, 2, 192, 198, 130, 30, 2, 12, 0, 30, 2, 0, 192, 198, 194, 31, 2, 1, 192, 198, 2, 34, 30, 1, 192, 198, 66, 35, 30, 1, 192, 198, 130, 72, 0, 198, 129, 104, 0, 198, 161, 36, 1, 0, 192, 198, 66, 37, 1, 0, 192, 198, 130, 0, 54, 156, 5, 2, 0, 39, 0, 4, 192, 198, 1, 187, 2, 4, 192, 198, 1, 24, 32, 4, 192, 198, 1, 0, 70, 156, 5, 2, 0, 39, 0, 4, 192, 198, 33, 187, 2, 4, 192, 198, 33, 24, 32, 4, 192, 198, 33, 0, 78, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 7, 192, 198, 2, 41, 30, 7, 192, 198, 66, 42, 30, 7, 192, 198, 130, 43, 30, 7, 192, 198, 194, 150, 30, 8, 192, 198, 2, 36, 30, 6, 192, 198, 2, 37, 30, 6, 192, 198, 66, 38, 30, 6, 192, 198, 130, 39, 30, 6, 192, 198, 194, 30, 2, 12, 0, 30, 2, 5, 192, 198, 2, 31, 2, 5, 192, 198, 66, 34, 30, 5, 192, 198, 130, 35, 30, 5, 192, 198, 194, 72, 0, 4, 192, 198, 65, 104, 0, 4, 192, 198, 97, 36, 1, 4, 192, 198, 130, 37, 1, 4, 192, 198, 194, 0, 46, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 0, 192, 197, 226, 39, 30, 1, 192, 197, 34, 41, 30, 1, 192, 197, 98, 43, 30, 1, 192, 197, 162, 150, 30, 1, 192, 197, 226, 104, 0, 255, 255, 25, 93, 193, 5, 37, 1, 197, 162, 31, 2, 0, 192, 197, 98, 35, 30, 0, 192, 197, 162, 0, 54, 5, 5, 2, 0, 39, 0, 255, 255, 25, 93, 193, 3, 187, 2, 255, 255, 25, 93, 193, 3, 24, 32, 255, 255, 25, 93, 193, 3, 0, 70, 5, 5, 2, 0, 39, 0, 255, 255, 25, 93, 193, 2, 187, 2, 255, 255, 25, 93, 193, 2, 24, 32, 255, 255, 25, 93, 193, 2, 0, 78, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 2, 192, 197, 226, 39, 30, 3, 192, 197, 34, 41, 30, 3, 192, 197, 98, 43, 30, 3, 192, 197, 162, 150, 30, 3, 192, 197, 226, 104, 0, 255, 255, 25, 93, 193, 4, 37, 1, 2, 192, 197, 34, 31, 2, 2, 192, 197, 98, 35, 30, 2, 192, 197, 162, 18, 0, 39, 0, 40, 0, 72, 0, 73, 0, 104, 0, 105, 0, 36, 1, 38, 1, 30, 2, 32, 2, 187, 2, 188, 2, 34, 30, 44, 30, 150, 30, 151, 30, 24, 32, 25, 32, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 26, 4, 185, 68, 186, 72, 185, 76, 36, 4, 185, 84, 186, 88, 186, 96, 185, 100, 185, 104, 185, 108, 186, 112, 42, 4, 185, 124, 185, 128, 185, 136, 48, 4, 185, 144, 185, 152, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 58, 4, 168, 68, 168, 72, 168, 76, 64, 4, 168, 84, 168, 88, 168, 96, 168, 100, 168, 104, 168, 108, 168, 112, 70, 4, 168, 124, 168, 128, 168, 136, 76, 4, 168, 144, 168, 152, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 89, 154, 89, 250, 89, 90, 90, 217, 68, 154, 114, 185, 121, 153, 121, 249, 121, 153, 122, 89, 122, 136, 14, 217, 122, 185, 153, 153, 153, 249, 153, 89, 154, 153, 169, 185, 192, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 89, 136, 89, 232, 89, 72, 90, 200, 68, 136, 114, 168, 121, 136, 121, 232, 121, 136, 122, 72, 122, 128, 14, 200, 122, 168, 153, 136, 153, 232, 153, 72, 154, 136, 169, 168, 192, 72, 170, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 26, 75, 8, 75, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 85, 232, 85, 89, 87, 72, 87, 154, 90, 136, 90, 58, 91, 40, 91, 218, 89, 200, 89, 26, 91, 8, 91, 186, 90, 168, 92, 6, 8, 8, 8, 250, 97, 232, 97, 249, 102, 232, 102, 168, 132, 153, 105, 136, 105, 249, 106, 232, 106, 25, 106, 8, 106, 121, 107, 104, 107, 89, 107, 72, 107, 154, 113, 136, 113, 250, 114, 232, 114, 26, 114, 8, 114, 10, 8, 185, 116, 168, 116, 57, 123, 40, 123, 217, 121, 200, 121, 121, 122, 104, 122, 12, 8, 14, 8, 153, 137, 136, 137, 249, 138, 232, 138, 25, 138, 8, 138, 154, 141, 136, 141, 250, 141, 232, 141, 250, 142, 232, 142, 26, 142, 8, 142, 249, 146, 232, 146, 25, 146, 8, 146, 185, 148, 168, 148, 153, 154, 136, 154, 57, 155, 40, 155, 217, 153, 200, 153, 57, 154, 40, 154, 121, 154, 104, 154, 25, 155, 8, 155, 249, 161, 232, 161, 249, 169, 232, 169, 89, 170, 153, 173, 136, 173, 185, 174, 168, 174, 25, 174, 8, 174, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 196, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 64, 72, 4, 186, 188, 104, 4, 177, 188, 36, 5, 250, 189, 37, 5, 241, 189, 255, 5, 185, 80, 39, 4, 177, 180, 152, 5, 177, 180, 255, 5, 185, 120, 39, 4, 177, 176, 152, 5, 177, 176, 255, 5, 186, 140, 72, 4, 186, 184, 104, 4, 177, 184, 36, 5, 250, 185, 37, 5, 241, 185, 255, 5, 168, 64, 104, 4, 168, 188, 37, 5, 232, 189, 255, 5, 168, 80, 39, 4, 168, 180, 152, 5, 168, 180, 255, 5, 168, 120, 39, 4, 168, 176, 152, 5, 168, 176, 255, 5, 168, 140, 104, 4, 168, 184, 37, 5, 232, 185, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 67, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 101, 8, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 106, 8, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 48, 36, 0, 32, 7, 0, 0, 0, 24, 0, 0, 0, 52, 36, 0, 0, 52, 36, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 196, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 116, 114, 97, 100, 105, 116, 105, 111, 110, 97, 108, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 84, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 38, 0, 0, 3, 60, 0, 60, 0, 9, 3, 60, 0, 60, 0, 3, 3, 60, 0, 60, 0, 1, 3, 60, 0, 60, 0, 35, 3, 38, 0, 97, 0, 60, 0, 3, 1, 60, 0, 60, 0, 60, 0, 2, 1, 60, 0, 226, 0, 60, 0, 60, 0, 60, 0, 194, 0, 38, 0, 100, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 38, 0, 101, 0, 60, 0, 234, 0, 60, 0, 60, 0, 60, 0, 202, 0, 38, 0, 111, 0, 60, 0, 244, 0, 60, 0, 60, 0, 60, 0, 212, 0, 60, 0, 161, 1, 60, 0, 60, 0, 60, 0, 160, 1, 38, 0, 117, 0, 60, 0, 176, 1, 60, 0, 60, 0, 60, 0, 175, 1, 0, 0, 250, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 38, 0, 0, 3, 60, 0, 60, 0, 9, 3, 60, 0, 60, 0, 3, 3, 60, 0, 60, 0, 1, 3, 60, 0, 60, 0, 35, 3, 38, 0, 97, 0, 60, 0, 3, 1, 60, 0, 60, 0, 60, 0, 2, 1, 60, 0, 226, 0, 60, 0, 60, 0, 60, 0, 194, 0, 38, 0, 67, 0, 60, 0, 99, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 104, 0, 60, 0, 60, 0, 60, 0, 67, 0, 72, 0, 38, 0, 100, 0, 60, 0, 17, 1, 60, 0, 60, 0, 60, 0, 16, 1, 38, 0, 101, 0, 60, 0, 234, 0, 60, 0, 60, 0, 60, 0, 202, 0, 38, 0, 71, 0, 60, 0, 103, 0, 104, 0, 60, 0, 60, 0, 60, 0, 71, 0, 104, 0, 60, 0, 60, 0, 60, 0, 71, 0, 72, 0, 60, 0, 103, 0, 105, 0, 60, 0, 60, 0, 60, 0, 71, 0, 105, 0, 60, 0, 60, 0, 60, 0, 71, 0, 73, 0, 38, 0, 75, 0, 60, 0, 107, 0, 104, 0, 60, 0, 60, 0, 60, 0, 75, 0, 104, 0, 60, 0, 60, 0, 60, 0, 75, 0, 72, 0, 38, 0, 78, 0, 60, 0, 110, 0, 104, 0, 60, 0, 60, 0, 60, 0, 78, 0, 104, 0, 60, 0, 60, 0, 60, 0, 78, 0, 72, 0, 60, 0, 110, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 103, 0, 60, 0, 60, 0, 60, 0, 78, 0, 71, 0, 60, 0, 110, 0, 103, 0, 104, 0, 60, 0, 60, 0, 60, 0, 78, 0, 103, 0, 104, 0, 60, 0, 60, 0, 60, 0, 78, 0, 71, 0, 104, 0, 60, 0, 60, 0, 60, 0, 78, 0, 71, 0, 72, 0, 38, 0, 111, 0, 60, 0, 244, 0, 60, 0, 60, 0, 60, 0, 212, 0, 60, 0, 161, 1, 60, 0, 60, 0, 60, 0, 160, 1, 38, 0, 80, 0, 60, 0, 112, 0, 104, 0, 60, 0, 60, 0, 60, 0, 80, 0, 104, 0, 60, 0, 60, 0, 60, 0, 80, 0, 72, 0, 38, 0, 81, 0, 60, 0, 113, 0, 117, 0, 60, 0, 60, 0, 60, 0, 81, 0, 117, 0, 60, 0, 60, 0, 60, 0, 81, 0, 85, 0, 38, 0, 84, 0, 60, 0, 116, 0, 104, 0, 60, 0, 60, 0, 60, 0, 84, 0, 104, 0, 60, 0, 60, 0, 60, 0, 84, 0, 72, 0, 60, 0, 116, 0, 114, 0, 60, 0, 60, 0, 60, 0, 84, 0, 114, 0, 60, 0, 60, 0, 60, 0, 84, 0, 82, 0, 38, 0, 117, 0, 60, 0, 176, 1, 60, 0, 60, 0, 60, 0, 175, 1, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 236, 57, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 116, 41, 0, 0, 116, 41, 0, 0, 116, 41, 0, 0, 116, 41, 0, 0, 240, 51, 0, 0, 240, 51, 0, 0, 220, 53, 0, 0, 220, 53, 0, 0, 212, 57, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 138, 1, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 111, 0, 119, 0, 124, 0, 132, 0, 140, 0, 148, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 156, 0, 164, 0, 172, 0, 180, 0, 187, 0, 195, 0, 200, 0, 208, 0, 216, 0, 224, 0, 228, 0, 236, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 4, 1, 11, 1, 19, 1, 27, 1, 35, 1, 43, 1, 51, 1, 59, 1, 66, 1, 74, 1, 82, 1, 90, 1, 98, 1, 106, 1, 113, 1, 121, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 129, 1, 252, 0, 48, 0, 48, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 188, 1, 240, 1, 48, 2, 192, 0, 192, 0, 192, 0, 112, 2, 176, 2, 236, 2, 32, 3, 96, 3, 144, 3, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 82, 0, 202, 193, 82, 0, 202, 226, 82, 0, 202, 3, 83, 0, 202, 36, 83, 0, 202, 69, 83, 0, 202, 102, 83, 0, 202, 135, 83, 0, 202, 168, 83, 0, 202, 201, 83, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 38, 1, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 230, 1, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 70, 3, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 38, 4, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 70, 5, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 6, 6, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 102, 7, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 197, 34, 0, 0, 134, 5, 3, 43, 197, 98, 0, 0, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 14, 0, 196, 160, 156, 46, 196, 138, 156, 50, 197, 162, 0, 0, 134, 5, 2, 51, 196, 150, 156, 50, 196, 138, 156, 58, 197, 226, 0, 0, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 197, 34, 1, 0, 196, 138, 156, 70, 197, 98, 1, 0, 134, 5, 2, 71, 197, 162, 1, 0, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 197, 226, 1, 0, 196, 142, 156, 82, 196, 150, 156, 82, 197, 34, 2, 0, 160, 5, 38, 93, 197, 98, 82, 0, 196, 138, 5, 42, 197, 98, 2, 0, 5, 5, 3, 43, 197, 162, 2, 0, 196, 150, 5, 42, 196, 146, 5, 42, 197, 130, 14, 0, 196, 160, 5, 46, 196, 138, 5, 50, 197, 226, 2, 0, 5, 5, 2, 51, 196, 150, 5, 50, 196, 138, 5, 58, 197, 34, 3, 0, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 197, 98, 3, 0, 196, 138, 5, 70, 197, 162, 3, 0, 5, 5, 2, 71, 197, 226, 3, 0, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 197, 34, 4, 0, 196, 142, 5, 82, 196, 150, 5, 82, 197, 98, 4, 0, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 43, 5, 5, 2, 43, 201, 70, 8, 0, 201, 102, 9, 0, 197, 162, 5, 0, 197, 226, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 49, 5, 5, 2, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 10, 0, 201, 70, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 6, 0, 197, 226, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 7, 0, 197, 98, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 7, 0, 197, 226, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 8, 0, 197, 98, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 8, 0, 197, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 9, 0, 197, 98, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 12, 0, 201, 198, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 10, 0, 197, 98, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 13, 0, 201, 70, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 83, 5, 5, 2, 83, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 11, 0, 197, 131, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 15, 0, 201, 38, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 12, 0, 197, 34, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 13, 0, 197, 195, 13, 0, 197, 35, 14, 0, 197, 131, 14, 0, 197, 227, 14, 0, 197, 67, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 17, 0, 201, 6, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 16, 0, 197, 131, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 4, 139, 192, 0, 0, 0, 194, 5, 3, 139, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 2, 139, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 5, 139, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 4, 139, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 16, 0, 197, 34, 17, 0, 192, 0, 0, 0, 197, 98, 17, 0, 197, 162, 17, 0, 197, 226, 17, 0, 192, 0, 0, 0, 197, 34, 18, 0, 192, 0, 0, 0, 197, 98, 18, 0, 197, 162, 18, 0, 197, 227, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 19, 0, 197, 130, 19, 0, 197, 194, 19, 0, 197, 2, 20, 0, 197, 67, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 20, 0, 197, 226, 20, 0, 197, 34, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 18, 0, 201, 230, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 23, 0, 197, 226, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 24, 0, 197, 131, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 24, 0, 197, 34, 25, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 25, 0, 197, 195, 25, 0, 201, 6, 21, 0, 201, 198, 21, 0, 201, 134, 22, 0, 201, 70, 23, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 27, 0, 197, 98, 27, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 27, 0, 197, 3, 28, 0, 197, 98, 28, 0, 197, 162, 28, 0, 197, 226, 28, 0, 197, 34, 29, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 29, 0, 197, 162, 29, 0, 197, 227, 29, 0, 197, 67, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 30, 0, 197, 226, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 31, 0, 197, 98, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 31, 0, 197, 226, 31, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 32, 0, 197, 131, 32, 0, 197, 227, 32, 0, 197, 67, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 33, 0, 197, 3, 34, 0, 197, 98, 34, 0, 197, 162, 34, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 34, 0, 197, 34, 35, 0, 197, 99, 35, 0, 197, 195, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 36, 0, 197, 98, 36, 0, 197, 163, 36, 0, 197, 3, 37, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 37, 0, 197, 195, 37, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 38, 0, 197, 98, 38, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 38, 0, 197, 3, 39, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 39, 0, 197, 162, 39, 0, 197, 226, 39, 0, 197, 34, 40, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 40, 0, 197, 162, 40, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 40, 0, 197, 34, 41, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 41, 0, 197, 162, 41, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 24, 0, 201, 38, 25, 0, 197, 98, 43, 0, 197, 162, 43, 0, 197, 226, 43, 0, 197, 34, 44, 0, 197, 98, 44, 0, 197, 162, 44, 0, 197, 226, 44, 0, 197, 34, 45, 0, 197, 98, 45, 0, 197, 162, 45, 0, 197, 34, 42, 0, 197, 226, 42, 0, 197, 226, 45, 0, 197, 34, 46, 0, 197, 98, 46, 0, 197, 162, 46, 0, 197, 226, 46, 0, 197, 34, 47, 0, 197, 98, 47, 0, 197, 162, 47, 0, 197, 98, 42, 0, 197, 34, 43, 0, 201, 70, 26, 0, 201, 6, 27, 0, 197, 226, 48, 0, 197, 34, 49, 0, 197, 98, 49, 0, 197, 162, 49, 0, 197, 226, 49, 0, 197, 34, 50, 0, 197, 98, 50, 0, 197, 162, 50, 0, 197, 226, 50, 0, 197, 34, 51, 0, 197, 98, 51, 0, 197, 162, 51, 0, 197, 34, 48, 0, 197, 162, 48, 0, 197, 226, 51, 0, 197, 34, 52, 0, 197, 98, 52, 0, 197, 162, 52, 0, 201, 198, 27, 0, 201, 134, 28, 0, 197, 226, 53, 0, 197, 34, 54, 0, 197, 98, 54, 0, 197, 162, 54, 0, 197, 226, 54, 0, 197, 34, 55, 0, 197, 98, 55, 0, 197, 162, 55, 0, 197, 226, 55, 0, 197, 34, 56, 0, 197, 34, 53, 0, 197, 162, 53, 0, 197, 98, 56, 0, 197, 162, 56, 0, 197, 226, 56, 0, 197, 34, 57, 0, 197, 98, 57, 0, 197, 162, 57, 0, 197, 226, 57, 0, 197, 34, 58, 0, 201, 70, 29, 0, 201, 6, 30, 0, 197, 162, 59, 0, 197, 226, 59, 0, 197, 34, 60, 0, 197, 98, 60, 0, 197, 162, 60, 0, 197, 226, 60, 0, 197, 34, 61, 0, 197, 98, 61, 0, 197, 162, 61, 0, 197, 226, 61, 0, 197, 34, 62, 0, 197, 98, 62, 0, 197, 162, 62, 0, 197, 226, 62, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 63, 0, 197, 98, 63, 0, 197, 162, 63, 0, 197, 226, 63, 0, 197, 34, 64, 0, 197, 98, 64, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 64, 0, 197, 3, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 65, 0, 197, 195, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 66, 0, 197, 131, 66, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 66, 0, 197, 67, 67, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 67, 0, 197, 3, 68, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 68, 0, 197, 195, 68, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 69, 0, 197, 131, 69, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 69, 0, 197, 67, 70, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 70, 0, 197, 3, 71, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 71, 0, 197, 195, 71, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 72, 0, 197, 131, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 73, 0, 197, 163, 73, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 74, 0, 197, 99, 74, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 19, 0, 192, 0, 0, 0, 197, 130, 19, 0, 192, 0, 0, 0, 197, 194, 19, 0, 192, 0, 0, 0, 197, 2, 20, 0, 192, 0, 0, 0, 197, 162, 20, 0, 192, 0, 0, 0, 197, 226, 20, 0, 192, 0, 0, 0, 197, 34, 21, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 196, 74, 0, 197, 68, 75, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 196, 75, 0, 197, 68, 76, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 196, 76, 0, 197, 68, 77, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 196, 77, 0, 197, 68, 78, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 196, 78, 0, 197, 68, 79, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 196, 79, 0, 197, 68, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 80, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 35, 81, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 17, 0, 192, 0, 0, 0, 197, 162, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 81, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 81, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 20, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 16, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 82, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 18, 0, 192, 0, 0, 0, 197, 162, 18, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 156, 5, 0, 42, 194, 5, 4, 139, 156, 5, 0, 42, 194, 5, 3, 139, 156, 5, 0, 50, 194, 5, 4, 139, 156, 5, 0, 58, 194, 5, 4, 139, 156, 5, 0, 68, 194, 5, 3, 139, 156, 5, 0, 70, 194, 5, 4, 139, 156, 5, 0, 70, 194, 5, 3, 139, 156, 5, 0, 82, 194, 5, 4, 139, 156, 5, 0, 90, 194, 5, 4, 139, 5, 5, 0, 42, 194, 5, 4, 139, 5, 5, 0, 42, 194, 5, 3, 139, 5, 5, 0, 50, 194, 5, 4, 139, 5, 5, 0, 58, 194, 5, 4, 139, 5, 5, 0, 68, 194, 5, 3, 139, 5, 5, 0, 70, 194, 5, 4, 139, 5, 5, 0, 70, 194, 5, 3, 139, 5, 5, 0, 82, 194, 5, 4, 139, 5, 5, 0, 90, 194, 5, 4, 139, 134, 5, 3, 43, 5, 162, 0, 0, 134, 5, 2, 43, 5, 162, 0, 0, 5, 5, 3, 43, 5, 162, 0, 0, 5, 5, 2, 43, 5, 162, 0, 0, 156, 5, 0, 46, 194, 5, 4, 139, 5, 5, 0, 46, 194, 5, 4, 139, 134, 5, 2, 51, 5, 162, 0, 0, 5, 5, 2, 51, 5, 162, 0, 0, 156, 5, 0, 58, 194, 5, 3, 139, 5, 5, 0, 58, 194, 5, 3, 139, 156, 5, 0, 64, 194, 5, 4, 139, 5, 5, 0, 64, 194, 5, 4, 139, 156, 5, 0, 68, 194, 5, 4, 139, 5, 5, 0, 68, 194, 5, 4, 139, 156, 5, 0, 76, 194, 5, 4, 139, 5, 5, 0, 76, 194, 5, 4, 139, 156, 5, 0, 78, 194, 5, 4, 139, 5, 5, 0, 78, 194, 5, 4, 139, 156, 5, 0, 82, 194, 5, 3, 139, 5, 5, 0, 82, 194, 5, 3, 139, 134, 5, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 156, 5, 0, 92, 194, 5, 4, 139, 5, 5, 0, 92, 194, 5, 4, 139, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 156, 5, 0, 82, 5, 150, 0, 0, 194, 5, 4, 139, 5, 5, 0, 82, 5, 150, 0, 0, 194, 5, 4, 139, 134, 5, 2, 71, 5, 162, 0, 0, 134, 5, 3, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 5, 5, 3, 71, 5, 162, 0, 0, 156, 5, 0, 54, 194, 5, 4, 139, 5, 5, 0, 54, 194, 5, 4, 139, 156, 5, 0, 42, 5, 146, 0, 0, 194, 5, 4, 139, 5, 5, 0, 42, 5, 146, 0, 0, 194, 5, 4, 139, 160, 112, 0, 42, 162, 5, 0, 50, 194, 5, 4, 139, 5, 112, 0, 42, 19, 5, 0, 50, 194, 5, 4, 139, 156, 5, 0, 70, 5, 158, 0, 0, 194, 5, 4, 139, 5, 5, 0, 70, 5, 158, 0, 0, 194, 5, 4, 139, 134, 5, 2, 51, 5, 160, 0, 0, 5, 5, 2, 51, 5, 160, 0, 0, 156, 5, 0, 70, 194, 5, 3, 139, 5, 164, 0, 0, 5, 5, 0, 70, 194, 5, 3, 139, 5, 164, 0, 0, 5, 5, 16, 12, 194, 5, 4, 139, 166, 5, 6, 96, 194, 5, 4, 139, 166, 5, 16, 96, 194, 5, 4, 139, 158, 5, 28, 96, 194, 5, 4, 139, 166, 5, 32, 96, 194, 5, 4, 139, 158, 5, 48, 96, 194, 5, 4, 139, 155, 5, 76, 96, 194, 5, 4, 139, 155, 5, 86, 96, 194, 5, 4, 139, 5, 5, 32, 96, 5, 150, 0, 0, 194, 5, 4, 139, 5, 5, 6, 96, 194, 5, 4, 139, 5, 5, 16, 96, 194, 5, 4, 139, 5, 5, 28, 96, 194, 5, 4, 139, 5, 5, 32, 96, 194, 5, 4, 139, 5, 5, 76, 96, 5, 150, 0, 0, 194, 5, 4, 139, 5, 5, 48, 96, 194, 5, 4, 139, 5, 5, 76, 96, 194, 5, 4, 139, 5, 5, 86, 96, 194, 5, 4, 139, 166, 5, 76, 96, 194, 5, 4, 139, 174, 5, 16, 97, 194, 5, 4, 139, 174, 5, 84, 97, 194, 5, 4, 139, 5, 5, 16, 97, 194, 5, 4, 139, 5, 5, 84, 97, 194, 5, 4, 139, 134, 5, 3, 43, 5, 200, 0, 0, 134, 5, 2, 43, 5, 200, 0, 0, 5, 5, 3, 43, 5, 200, 0, 0, 5, 5, 2, 43, 5, 200, 0, 0, 156, 5, 0, 44, 194, 5, 5, 139, 5, 5, 0, 44, 194, 5, 5, 139, 156, 5, 0, 46, 5, 160, 0, 0, 194, 5, 4, 139, 5, 5, 0, 46, 5, 160, 0, 0, 194, 5, 4, 139, 156, 5, 0, 48, 194, 5, 5, 139, 5, 5, 0, 48, 194, 5, 5, 139, 156, 5, 0, 50, 5, 164, 0, 0, 194, 5, 4, 139, 5, 5, 0, 50, 5, 164, 0, 0, 194, 5, 4, 139, 134, 5, 2, 51, 5, 204, 0, 0, 5, 5, 2, 51, 5, 204, 0, 0, 134, 5, 2, 51, 5, 208, 0, 0, 5, 5, 2, 51, 5, 208, 0, 0, 156, 5, 0, 56, 194, 5, 5, 139, 5, 5, 0, 56, 194, 5, 5, 139, 156, 5, 0, 58, 5, 150, 0, 0, 194, 5, 4, 139, 5, 5, 0, 58, 5, 150, 0, 0, 194, 5, 4, 139, 156, 5, 0, 62, 194, 5, 4, 139, 5, 5, 0, 62, 194, 5, 4, 139, 156, 5, 0, 62, 194, 5, 5, 139, 5, 5, 0, 62, 194, 5, 5, 139, 156, 5, 0, 64, 194, 5, 5, 139, 5, 5, 0, 64, 194, 5, 5, 139, 156, 5, 0, 64, 194, 5, 5, 139, 5, 164, 0, 0, 5, 5, 0, 64, 194, 5, 5, 139, 5, 164, 0, 0, 156, 5, 0, 66, 194, 5, 4, 139, 5, 5, 0, 66, 194, 5, 4, 139, 156, 5, 0, 66, 194, 5, 5, 139, 5, 5, 0, 66, 194, 5, 5, 139, 156, 5, 0, 68, 194, 5, 5, 139, 5, 5, 0, 68, 194, 5, 5, 139, 156, 5, 0, 70, 194, 5, 3, 139, 194, 5, 4, 139, 5, 5, 0, 70, 194, 5, 3, 139, 194, 5, 4, 139, 156, 5, 0, 70, 194, 5, 3, 139, 5, 150, 0, 0, 5, 5, 0, 70, 194, 5, 3, 139, 5, 150, 0, 0, 156, 5, 0, 70, 5, 164, 0, 0, 194, 5, 4, 139, 5, 5, 0, 70, 5, 164, 0, 0, 194, 5, 4, 139, 156, 5, 0, 72, 194, 5, 4, 139, 5, 5, 0, 72, 194, 5, 4, 139, 156, 5, 0, 76, 194, 5, 5, 139, 5, 5, 0, 76, 194, 5, 5, 139, 156, 5, 0, 76, 194, 5, 5, 139, 5, 164, 0, 0, 5, 5, 0, 76, 194, 5, 5, 139, 5, 164, 0, 0, 156, 5, 0, 78, 194, 5, 5, 139, 5, 5, 0, 78, 194, 5, 5, 139, 156, 5, 0, 78, 194, 5, 4, 139, 5, 156, 0, 0, 5, 5, 0, 78, 194, 5, 4, 139, 5, 156, 0, 0, 156, 5, 0, 78, 194, 5, 5, 139, 5, 156, 0, 0, 5, 5, 0, 78, 194, 5, 5, 139, 5, 156, 0, 0, 156, 5, 0, 80, 194, 5, 5, 139, 5, 5, 0, 80, 194, 5, 5, 139, 156, 5, 0, 82, 194, 5, 3, 139, 194, 5, 4, 139, 5, 5, 0, 82, 194, 5, 3, 139, 194, 5, 4, 139, 156, 5, 0, 84, 194, 5, 3, 139, 5, 5, 0, 84, 194, 5, 3, 139, 156, 5, 0, 84, 194, 5, 5, 139, 5, 5, 0, 84, 194, 5, 5, 139, 160, 5, 0, 86, 194, 5, 4, 139, 5, 5, 0, 86, 194, 5, 4, 139, 160, 5, 0, 86, 194, 5, 5, 139, 5, 5, 0, 86, 194, 5, 5, 139, 156, 5, 0, 92, 194, 5, 5, 139, 5, 5, 0, 92, 194, 5, 5, 139, 156, 5, 0, 42, 194, 5, 5, 139, 134, 5, 3, 43, 194, 5, 5, 139, 134, 5, 2, 43, 194, 5, 5, 139, 5, 5, 0, 42, 194, 5, 5, 139, 5, 5, 3, 43, 194, 5, 5, 139, 5, 5, 2, 43, 194, 5, 5, 139, 156, 5, 0, 42, 194, 5, 2, 139, 5, 5, 0, 42, 194, 5, 2, 139, 134, 5, 3, 43, 194, 5, 4, 139, 5, 5, 3, 43, 194, 5, 4, 139, 134, 5, 3, 43, 5, 138, 0, 0, 5, 5, 3, 43, 5, 138, 0, 0, 134, 5, 3, 43, 194, 5, 2, 139, 5, 5, 3, 43, 194, 5, 2, 139, 134, 5, 3, 43, 194, 5, 3, 139, 5, 5, 3, 43, 194, 5, 3, 139, 134, 5, 2, 43, 194, 5, 4, 139, 5, 5, 2, 43, 194, 5, 4, 139, 134, 5, 2, 43, 5, 138, 0, 0, 5, 5, 2, 43, 5, 138, 0, 0, 134, 5, 2, 43, 194, 5, 2, 139, 5, 5, 2, 43, 194, 5, 2, 139, 134, 5, 2, 43, 194, 5, 3, 139, 5, 5, 2, 43, 194, 5, 3, 139, 156, 5, 0, 50, 194, 5, 5, 139, 134, 5, 2, 51, 194, 5, 5, 139, 5, 5, 0, 50, 194, 5, 5, 139, 5, 5, 2, 51, 194, 5, 5, 139, 156, 5, 0, 50, 194, 5, 2, 139, 5, 5, 0, 50, 194, 5, 2, 139, 156, 5, 0, 50, 194, 5, 3, 139, 5, 5, 0, 50, 194, 5, 3, 139, 134, 5, 2, 51, 194, 5, 4, 139, 5, 5, 2, 51, 194, 5, 4, 139, 134, 5, 2, 51, 5, 138, 0, 0, 5, 5, 2, 51, 5, 138, 0, 0, 134, 5, 2, 51, 194, 5, 2, 139, 5, 5, 2, 51, 194, 5, 2, 139, 134, 5, 2, 51, 194, 5, 3, 139, 5, 5, 2, 51, 194, 5, 3, 139, 156, 5, 0, 58, 194, 5, 2, 139, 5, 5, 0, 58, 194, 5, 2, 139, 156, 5, 0, 58, 194, 5, 5, 139, 5, 5, 0, 58, 194, 5, 5, 139, 156, 5, 0, 70, 194, 5, 5, 139, 134, 5, 2, 71, 194, 5, 5, 139, 5, 5, 0, 70, 194, 5, 5, 139, 5, 5, 2, 71, 194, 5, 5, 139, 156, 5, 0, 70, 194, 5, 2, 139, 5, 5, 0, 70, 194, 5, 2, 139, 134, 5, 2, 71, 194, 5, 4, 139, 5, 5, 2, 71, 194, 5, 4, 139, 134, 5, 2, 71, 5, 138, 0, 0, 5, 5, 2, 71, 5, 138, 0, 0, 134, 5, 2, 71, 194, 5, 2, 139, 5, 5, 2, 71, 194, 5, 2, 139, 134, 5, 2, 71, 194, 5, 3, 139, 5, 5, 2, 71, 194, 5, 3, 139, 134, 5, 3, 71, 194, 5, 4, 139, 5, 5, 3, 71, 194, 5, 4, 139, 134, 5, 3, 71, 5, 138, 0, 0, 5, 5, 3, 71, 5, 138, 0, 0, 134, 5, 3, 71, 194, 5, 2, 139, 5, 5, 3, 71, 194, 5, 2, 139, 134, 5, 3, 71, 194, 5, 3, 139, 5, 5, 3, 71, 194, 5, 3, 139, 134, 5, 3, 71, 194, 5, 5, 139, 134, 5, 2, 71, 5, 190, 0, 0, 194, 5, 5, 139, 5, 5, 3, 71, 194, 5, 5, 139, 5, 5, 2, 71, 5, 190, 0, 0, 194, 5, 5, 139, 156, 5, 0, 82, 194, 5, 5, 139, 5, 5, 0, 82, 194, 5, 5, 139, 156, 5, 0, 82, 194, 5, 2, 139, 5, 5, 0, 82, 194, 5, 2, 139, 134, 5, 2, 83, 194, 5, 4, 139, 5, 5, 2, 83, 194, 5, 4, 139, 134, 5, 2, 83, 5, 138, 0, 0, 5, 5, 2, 83, 5, 138, 0, 0, 134, 5, 2, 83, 194, 5, 2, 139, 5, 5, 2, 83, 194, 5, 2, 139, 134, 5, 2, 83, 194, 5, 3, 139, 5, 5, 2, 83, 194, 5, 3, 139, 134, 5, 2, 83, 194, 5, 5, 139, 5, 5, 2, 83, 194, 5, 5, 139, 156, 5, 0, 90, 194, 5, 5, 139, 5, 5, 0, 90, 194, 5, 5, 139, 156, 5, 0, 90, 194, 5, 2, 139, 5, 5, 0, 90, 194, 5, 2, 139, 156, 5, 0, 90, 194, 5, 3, 139, 5, 5, 0, 90, 194, 5, 3, 139, 5, 5, 6, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 6, 96, 5, 134, 0, 0, 194, 5, 4, 139, 166, 5, 6, 96, 5, 132, 0, 0, 194, 5, 4, 139, 166, 5, 6, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 16, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 16, 96, 5, 134, 0, 0, 194, 5, 4, 139, 166, 5, 16, 96, 5, 132, 0, 0, 194, 5, 4, 139, 166, 5, 16, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 28, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 28, 96, 5, 134, 0, 0, 194, 5, 4, 139, 158, 5, 28, 96, 5, 132, 0, 0, 194, 5, 4, 139, 158, 5, 28, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 32, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 32, 96, 5, 134, 0, 0, 194, 5, 4, 139, 166, 5, 32, 96, 5, 132, 0, 0, 194, 5, 4, 139, 166, 5, 32, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 48, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 48, 96, 5, 134, 0, 0, 194, 5, 4, 139, 158, 5, 48, 96, 5, 132, 0, 0, 194, 5, 4, 139, 158, 5, 48, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 76, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 76, 96, 5, 134, 0, 0, 194, 5, 4, 139, 155, 5, 76, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 86, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 86, 96, 5, 134, 0, 0, 194, 5, 4, 139, 155, 5, 86, 96, 5, 132, 0, 0, 194, 5, 4, 139, 155, 5, 86, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 6, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 6, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 166, 5, 6, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 166, 5, 6, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 28, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 28, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 158, 5, 28, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 158, 5, 28, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 86, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 86, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 155, 5, 86, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 155, 5, 86, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 6, 96, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 28, 96, 194, 5, 4, 139, 5, 216, 0, 0, 193, 9, 22, 12, 194, 5, 4, 139, 193, 16, 22, 12, 194, 5, 4, 139, 5, 5, 86, 96, 194, 5, 4, 139, 5, 216, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 1, 0, 2, 3, 3, 235, 134, 5, 6, 3, 2, 235, 134, 5, 0, 50, 156, 5, 48, 0, 2, 3, 2, 243, 134, 5, 0, 70, 156, 5, 1, 0, 2, 3, 255, 255, 2, 71, 134, 5, 27, 3, 255, 255, 3, 71, 134, 5, 0, 82, 156, 5, 48, 0, 27, 3, 255, 255, 2, 83, 134, 5, 0, 42, 5, 5, 1, 0, 2, 3, 3, 235, 5, 5, 6, 3, 2, 235, 5, 5, 0, 50, 5, 5, 48, 0, 2, 3, 2, 243, 5, 5, 0, 70, 5, 5, 1, 0, 2, 3, 255, 255, 2, 71, 5, 5, 27, 3, 255, 255, 3, 71, 5, 5, 0, 82, 5, 5, 48, 0, 27, 3, 255, 255, 2, 83, 5, 5, 156, 42, 196, 162, 1, 0, 2, 3, 4, 192, 197, 162, 6, 3, 4, 192, 197, 226, 5, 42, 196, 162, 1, 0, 2, 3, 5, 192, 197, 34, 6, 3, 5, 192, 197, 98, 156, 50, 196, 162, 48, 0, 2, 3, 6, 192, 197, 34, 5, 50, 196, 162, 48, 0, 2, 3, 6, 192, 197, 98, 156, 82, 196, 162, 48, 0, 27, 3, 9, 192, 197, 162, 5, 82, 196, 162, 48, 0, 27, 3, 9, 192, 197, 226, 3, 71, 134, 5, 48, 0, 2, 3, 10, 192, 197, 162, 3, 71, 5, 5, 48, 0, 2, 3, 10, 192, 197, 226, 156, 70, 196, 162, 1, 0, 2, 3, 11, 192, 197, 226, 27, 3, 12, 192, 197, 34, 5, 70, 196, 162, 1, 0, 2, 3, 12, 192, 197, 98, 27, 3, 12, 192, 197, 162, 156, 50, 196, 160, 48, 0, 2, 3, 15, 192, 197, 162, 5, 50, 196, 160, 48, 0, 2, 3, 15, 192, 197, 226, 156, 42, 196, 200, 1, 0, 2, 3, 22, 192, 197, 162, 6, 3, 22, 192, 197, 226, 5, 42, 196, 200, 1, 0, 2, 3, 23, 192, 197, 34, 6, 3, 23, 192, 197, 98, 156, 50, 196, 204, 48, 0, 2, 3, 26, 192, 197, 34, 5, 50, 196, 204, 48, 0, 2, 3, 26, 192, 197, 98, 156, 50, 196, 208, 48, 0, 2, 3, 26, 192, 197, 162, 5, 50, 196, 208, 48, 0, 2, 3, 26, 192, 197, 226, 41, 0, 197, 226, 1, 0, 2, 3, 42, 192, 197, 34, 6, 3, 42, 192, 197, 98, 42, 0, 197, 162, 1, 0, 2, 3, 42, 192, 197, 226, 6, 3, 43, 192, 197, 34, 47, 0, 197, 226, 48, 0, 2, 3, 48, 192, 197, 34, 48, 0, 197, 98, 48, 0, 2, 3, 48, 192, 197, 162, 52, 0, 197, 226, 48, 0, 2, 3, 53, 192, 197, 34, 53, 0, 197, 98, 48, 0, 2, 3, 53, 192, 197, 162, 58, 0, 197, 98, 48, 0, 2, 3, 58, 192, 197, 163, 59, 0, 197, 2, 48, 0, 2, 3, 59, 192, 197, 67, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 68, 185, 72, 185, 76, 28, 4, 185, 92, 185, 96, 185, 100, 186, 104, 186, 112, 185, 116, 185, 120, 185, 124, 186, 128, 30, 4, 185, 144, 185, 148, 185, 156, 186, 160, 185, 164, 32, 4, 185, 176, 185, 180, 185, 184, 185, 188, 185, 192, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 34, 4, 168, 68, 168, 72, 168, 76, 36, 4, 168, 92, 168, 96, 168, 100, 168, 104, 168, 112, 168, 116, 168, 120, 168, 124, 168, 128, 38, 4, 168, 144, 168, 148, 168, 156, 168, 160, 168, 164, 40, 4, 168, 176, 168, 180, 168, 184, 168, 188, 168, 192, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 136, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 153, 57, 217, 57, 185, 64, 185, 57, 121, 58, 89, 58, 0, 8, 249, 74, 154, 85, 218, 85, 185, 88, 122, 86, 154, 105, 218, 105, 26, 106, 122, 106, 217, 76, 186, 129, 153, 137, 217, 137, 185, 140, 185, 137, 121, 138, 136, 14, 217, 138, 153, 173, 217, 173, 25, 174, 121, 174, 217, 189, 185, 196, 2, 8, 136, 57, 200, 57, 168, 64, 168, 57, 104, 58, 72, 58, 4, 8, 232, 74, 136, 85, 200, 85, 168, 88, 104, 86, 136, 105, 200, 105, 8, 106, 104, 106, 200, 76, 168, 129, 136, 137, 200, 137, 168, 140, 168, 137, 104, 138, 128, 14, 200, 138, 136, 173, 200, 173, 8, 174, 104, 174, 200, 189, 168, 196, 104, 190, 57, 59, 40, 59, 185, 60, 168, 60, 42, 4, 44, 4, 217, 73, 200, 73, 25, 74, 8, 74, 185, 74, 168, 74, 57, 74, 40, 74, 57, 78, 40, 78, 185, 80, 168, 80, 58, 87, 40, 87, 250, 85, 232, 85, 186, 86, 168, 86, 46, 4, 48, 4, 58, 86, 40, 86, 25, 98, 8, 98, 249, 97, 232, 97, 185, 98, 168, 98, 249, 98, 232, 98, 25, 102, 8, 102, 89, 103, 72, 103, 186, 105, 168, 105, 58, 107, 40, 107, 250, 105, 232, 105, 26, 107, 8, 107, 186, 106, 168, 108, 6, 8, 8, 8, 26, 114, 8, 114, 249, 118, 232, 118, 168, 152, 217, 121, 200, 121, 249, 122, 232, 122, 57, 122, 40, 122, 121, 123, 104, 123, 89, 123, 72, 123, 218, 129, 200, 129, 250, 130, 232, 130, 58, 130, 40, 130, 10, 8, 185, 132, 168, 132, 57, 139, 40, 139, 249, 137, 232, 137, 153, 138, 136, 138, 12, 8, 14, 8, 217, 157, 200, 157, 249, 158, 232, 158, 57, 158, 40, 158, 218, 161, 200, 161, 26, 162, 8, 162, 250, 162, 232, 162, 58, 162, 40, 162, 249, 166, 232, 166, 57, 166, 40, 166, 185, 168, 168, 168, 185, 173, 168, 173, 57, 175, 40, 175, 249, 173, 232, 173, 89, 174, 72, 174, 153, 174, 136, 174, 50, 4, 52, 4, 25, 182, 8, 182, 25, 190, 8, 190, 121, 190, 217, 193, 200, 193, 185, 194, 168, 194, 57, 194, 40, 194, 232, 160, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 84, 200, 160, 169, 160, 200, 56, 169, 84, 187, 104, 187, 112, 169, 104, 169, 112, 169, 200, 169, 128, 217, 136, 187, 84, 200, 136, 169, 84, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 84, 255, 5, 185, 136, 255, 5, 185, 172, 255, 5, 168, 56, 255, 5, 168, 84, 255, 5, 168, 136, 255, 5, 168, 172, 255, 5, 25, 59, 255, 5, 8, 59, 255, 5, 26, 87, 255, 5, 8, 87, 255, 5, 25, 175, 255, 5, 8, 175, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 199, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 108, 83, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 228, 41, 0, 0, 228, 41, 0, 0, 228, 41, 0, 0, 228, 41, 0, 0, 212, 63, 0, 0, 212, 63, 0, 0, 212, 76, 0, 0, 70, 77, 0, 0, 84, 83, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 145, 1, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 111, 0, 119, 0, 124, 0, 132, 0, 140, 0, 148, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 156, 0, 164, 0, 172, 0, 180, 0, 187, 0, 195, 0, 200, 0, 208, 0, 216, 0, 224, 0, 228, 0, 236, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 4, 1, 11, 1, 19, 1, 27, 1, 35, 1, 43, 1, 51, 1, 59, 1, 66, 1, 74, 1, 82, 1, 90, 1, 98, 1, 106, 1, 113, 1, 121, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 136, 1, 252, 0, 48, 0, 48, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 244, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 188, 1, 240, 1, 48, 2, 192, 0, 192, 0, 192, 0, 112, 2, 176, 2, 236, 2, 32, 3, 96, 3, 144, 3, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 64, 174, 0, 202, 97, 174, 0, 202, 130, 174, 0, 202, 163, 174, 0, 202, 196, 174, 0, 202, 229, 174, 0, 202, 6, 175, 0, 202, 39, 175, 0, 202, 72, 175, 0, 202, 105, 175, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 201, 32, 1, 0, 156, 5, 0, 48, 201, 134, 8, 0, 156, 5, 0, 52, 201, 68, 9, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 201, 128, 33, 0, 156, 5, 0, 64, 156, 5, 0, 66, 201, 0, 41, 0, 201, 102, 67, 0, 201, 192, 68, 0, 201, 132, 76, 0, 156, 5, 0, 76, 156, 5, 0, 78, 201, 64, 105, 0, 201, 198, 121, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 166, 122, 0, 5, 5, 0, 44, 201, 192, 123, 0, 5, 5, 0, 48, 201, 198, 127, 0, 5, 5, 0, 52, 201, 132, 128, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 201, 160, 140, 0, 5, 5, 0, 64, 5, 5, 0, 66, 201, 160, 144, 0, 201, 70, 156, 0, 201, 160, 157, 0, 201, 196, 161, 0, 5, 5, 0, 76, 5, 5, 0, 78, 201, 32, 176, 0, 201, 166, 184, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 197, 194, 91, 0, 134, 5, 3, 43, 197, 2, 92, 0, 196, 150, 156, 42, 196, 146, 156, 42, 197, 194, 105, 0, 196, 160, 156, 46, 196, 138, 156, 50, 197, 66, 92, 0, 134, 5, 2, 51, 196, 150, 156, 50, 196, 138, 156, 58, 197, 130, 92, 0, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 197, 194, 92, 0, 196, 138, 156, 70, 197, 2, 93, 0, 134, 5, 2, 71, 197, 66, 93, 0, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 197, 130, 93, 0, 196, 142, 156, 82, 196, 150, 156, 82, 197, 194, 93, 0, 160, 5, 38, 93, 197, 2, 174, 0, 196, 138, 5, 42, 197, 2, 94, 0, 5, 5, 3, 43, 197, 66, 94, 0, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 106, 0, 196, 160, 5, 46, 196, 138, 5, 50, 197, 130, 94, 0, 5, 5, 2, 51, 196, 150, 5, 50, 196, 138, 5, 58, 197, 194, 94, 0, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 197, 2, 95, 0, 196, 138, 5, 70, 197, 66, 95, 0, 5, 5, 2, 71, 197, 130, 95, 0, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 197, 194, 95, 0, 196, 142, 5, 82, 196, 150, 5, 82, 197, 2, 96, 0, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 43, 5, 5, 2, 43, 201, 134, 185, 0, 201, 166, 186, 0, 197, 66, 97, 0, 197, 130, 97, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 49, 5, 5, 2, 49, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 187, 0, 201, 134, 188, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 98, 0, 197, 130, 98, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 98, 0, 197, 2, 99, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 99, 0, 197, 130, 99, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 99, 0, 197, 2, 100, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 100, 0, 197, 130, 100, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 100, 0, 197, 2, 101, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 189, 0, 201, 6, 190, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 101, 0, 197, 2, 102, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 190, 0, 201, 134, 191, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 83, 5, 5, 2, 83, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 102, 0, 197, 35, 103, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 192, 0, 201, 102, 193, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 104, 0, 197, 194, 104, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 105, 0, 197, 99, 105, 0, 197, 195, 105, 0, 197, 35, 106, 0, 197, 131, 106, 0, 197, 227, 106, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 194, 0, 201, 70, 195, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 107, 0, 197, 35, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 4, 139, 192, 0, 0, 0, 194, 5, 3, 139, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 2, 139, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 5, 139, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 4, 139, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 108, 0, 197, 194, 108, 0, 192, 0, 0, 0, 197, 2, 109, 0, 197, 66, 109, 0, 197, 130, 109, 0, 192, 0, 0, 0, 197, 194, 109, 0, 192, 0, 0, 0, 197, 2, 110, 0, 197, 66, 110, 0, 197, 131, 110, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 110, 0, 197, 34, 111, 0, 197, 98, 111, 0, 197, 162, 111, 0, 197, 227, 111, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 112, 0, 197, 130, 112, 0, 197, 194, 112, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 113, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 113, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 113, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 113, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 114, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 196, 0, 201, 38, 197, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 115, 0, 197, 130, 115, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 115, 0, 197, 35, 116, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 116, 0, 197, 194, 116, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 117, 0, 197, 99, 117, 0, 201, 70, 198, 0, 201, 6, 199, 0, 201, 198, 199, 0, 201, 134, 200, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 118, 0, 197, 2, 119, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 119, 0, 197, 163, 119, 0, 197, 2, 120, 0, 197, 66, 120, 0, 197, 130, 120, 0, 197, 194, 120, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 121, 0, 197, 66, 121, 0, 197, 131, 121, 0, 197, 227, 121, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 122, 0, 197, 130, 122, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 122, 0, 197, 2, 123, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 123, 0, 197, 130, 123, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 123, 0, 197, 35, 124, 0, 197, 131, 124, 0, 197, 227, 124, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 125, 0, 197, 163, 125, 0, 197, 2, 126, 0, 197, 66, 126, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 126, 0, 197, 194, 126, 0, 197, 3, 127, 0, 197, 99, 127, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 127, 0, 197, 2, 128, 0, 197, 67, 128, 0, 197, 163, 128, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 129, 0, 197, 99, 129, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 129, 0, 197, 2, 130, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 130, 0, 197, 163, 130, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 131, 0, 197, 66, 131, 0, 197, 130, 131, 0, 197, 194, 131, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 132, 0, 197, 66, 132, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 132, 0, 197, 194, 132, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 133, 0, 197, 66, 133, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 201, 0, 201, 102, 202, 0, 197, 2, 135, 0, 197, 66, 135, 0, 197, 130, 135, 0, 197, 194, 135, 0, 197, 2, 136, 0, 197, 66, 136, 0, 197, 130, 136, 0, 197, 194, 136, 0, 197, 2, 137, 0, 197, 66, 137, 0, 197, 194, 133, 0, 197, 130, 134, 0, 197, 130, 137, 0, 197, 194, 137, 0, 197, 2, 138, 0, 197, 66, 138, 0, 197, 130, 138, 0, 197, 194, 138, 0, 197, 2, 139, 0, 197, 66, 139, 0, 197, 2, 134, 0, 197, 194, 134, 0, 201, 134, 203, 0, 201, 70, 204, 0, 197, 130, 140, 0, 197, 194, 140, 0, 197, 2, 141, 0, 197, 66, 141, 0, 197, 130, 141, 0, 197, 194, 141, 0, 197, 2, 142, 0, 197, 66, 142, 0, 197, 130, 142, 0, 197, 194, 142, 0, 197, 2, 143, 0, 197, 66, 143, 0, 197, 194, 139, 0, 197, 66, 140, 0, 197, 130, 143, 0, 197, 194, 143, 0, 197, 2, 144, 0, 197, 66, 144, 0, 201, 6, 205, 0, 201, 198, 205, 0, 197, 130, 145, 0, 197, 194, 145, 0, 197, 2, 146, 0, 197, 66, 146, 0, 197, 130, 146, 0, 197, 194, 146, 0, 197, 2, 147, 0, 197, 66, 147, 0, 197, 130, 147, 0, 197, 194, 147, 0, 197, 194, 144, 0, 197, 66, 145, 0, 197, 2, 148, 0, 197, 66, 148, 0, 197, 130, 148, 0, 197, 194, 148, 0, 197, 2, 149, 0, 197, 66, 149, 0, 197, 130, 149, 0, 197, 194, 149, 0, 201, 134, 206, 0, 201, 70, 207, 0, 197, 66, 151, 0, 197, 130, 151, 0, 197, 194, 151, 0, 197, 2, 152, 0, 197, 66, 152, 0, 197, 130, 152, 0, 197, 194, 152, 0, 197, 2, 153, 0, 197, 66, 153, 0, 197, 130, 153, 0, 197, 194, 153, 0, 197, 2, 154, 0, 197, 66, 154, 0, 197, 130, 154, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 154, 0, 197, 2, 155, 0, 197, 66, 155, 0, 197, 130, 155, 0, 197, 194, 155, 0, 197, 2, 156, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 156, 0, 197, 163, 156, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 157, 0, 197, 99, 157, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 157, 0, 197, 35, 158, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 158, 0, 197, 227, 158, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 159, 0, 197, 163, 159, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 160, 0, 197, 99, 160, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 160, 0, 197, 35, 161, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 161, 0, 197, 227, 161, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 67, 162, 0, 197, 163, 162, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 163, 0, 197, 99, 163, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 163, 0, 197, 35, 164, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 164, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 164, 0, 197, 67, 165, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 165, 0, 197, 3, 166, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 110, 0, 192, 0, 0, 0, 197, 34, 111, 0, 192, 0, 0, 0, 197, 98, 111, 0, 192, 0, 0, 0, 197, 162, 111, 0, 192, 0, 0, 0, 197, 66, 112, 0, 192, 0, 0, 0, 197, 130, 112, 0, 192, 0, 0, 0, 197, 194, 112, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 100, 166, 0, 197, 228, 166, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 100, 167, 0, 197, 228, 167, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 100, 168, 0, 197, 228, 168, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 100, 169, 0, 197, 228, 169, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 100, 170, 0, 197, 228, 170, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 100, 171, 0, 197, 228, 171, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 99, 172, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 172, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 109, 0, 192, 0, 0, 0, 197, 66, 109, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 173, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 131, 110, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 109, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 173, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 111, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 110, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 130, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 163, 173, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 109, 0, 192, 0, 0, 0, 197, 66, 110, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 128, 33, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 135, 5, 2, 47, 5, 142, 0, 0, 70, 5, 2, 47, 5, 142, 0, 0, 135, 5, 2, 47, 5, 144, 0, 0, 70, 5, 2, 47, 5, 144, 0, 0, 135, 5, 2, 47, 5, 156, 0, 0, 70, 5, 2, 47, 5, 156, 0, 0, 135, 5, 2, 47, 194, 5, 5, 139, 70, 5, 2, 47, 194, 5, 5, 139, 135, 5, 2, 47, 5, 150, 0, 0, 70, 5, 2, 47, 5, 150, 0, 0, 135, 5, 2, 47, 5, 160, 0, 0, 70, 5, 2, 47, 5, 160, 0, 0, 135, 5, 2, 47, 5, 206, 0, 0, 70, 5, 2, 47, 5, 206, 0, 0, 70, 5, 2, 47, 5, 210, 0, 0, 135, 5, 3, 55, 5, 150, 0, 0, 194, 5, 4, 139, 70, 5, 3, 55, 5, 150, 0, 0, 194, 5, 4, 139, 135, 5, 3, 55, 5, 138, 0, 0, 135, 5, 3, 55, 194, 5, 4, 139, 135, 5, 3, 55, 5, 142, 0, 0, 70, 5, 3, 55, 5, 138, 0, 0, 70, 5, 3, 55, 194, 5, 4, 139, 70, 5, 3, 55, 5, 142, 0, 0, 135, 5, 2, 55, 5, 142, 0, 0, 70, 5, 2, 55, 5, 142, 0, 0, 135, 5, 3, 55, 194, 5, 3, 139, 70, 5, 3, 55, 194, 5, 3, 139, 135, 5, 3, 55, 5, 164, 0, 0, 70, 5, 3, 55, 5, 164, 0, 0, 135, 5, 3, 55, 5, 140, 0, 0, 70, 5, 3, 55, 5, 140, 0, 0, 135, 5, 3, 55, 5, 162, 0, 0, 70, 5, 3, 55, 5, 162, 0, 0, 135, 5, 3, 55, 5, 156, 0, 0, 135, 5, 3, 55, 5, 144, 0, 0, 70, 5, 3, 55, 5, 144, 0, 0, 135, 5, 3, 55, 5, 184, 0, 0, 70, 5, 3, 55, 5, 184, 0, 0, 135, 5, 3, 55, 5, 188, 0, 0, 70, 5, 3, 55, 5, 188, 0, 0, 135, 5, 2, 55, 5, 144, 0, 0, 70, 5, 2, 55, 5, 144, 0, 0, 135, 5, 2, 55, 5, 156, 0, 0, 70, 5, 2, 55, 5, 156, 0, 0, 135, 5, 2, 55, 194, 5, 5, 139, 70, 5, 2, 55, 194, 5, 5, 139, 135, 5, 2, 55, 5, 150, 0, 0, 70, 5, 2, 55, 5, 150, 0, 0, 135, 5, 2, 55, 5, 160, 0, 0, 70, 5, 2, 55, 5, 160, 0, 0, 135, 5, 2, 55, 5, 206, 0, 0, 70, 5, 2, 55, 5, 206, 0, 0, 135, 5, 3, 55, 5, 208, 0, 0, 70, 5, 3, 55, 5, 208, 0, 0, 70, 5, 2, 55, 5, 210, 0, 0, 135, 5, 3, 55, 194, 5, 2, 139, 70, 5, 3, 55, 194, 5, 2, 139, 135, 5, 3, 55, 194, 5, 5, 139, 70, 5, 3, 55, 194, 5, 5, 139, 135, 5, 2, 63, 5, 142, 0, 0, 70, 5, 2, 63, 5, 142, 0, 0, 135, 5, 2, 63, 5, 144, 0, 0, 70, 5, 2, 63, 5, 144, 0, 0, 135, 5, 2, 63, 5, 156, 0, 0, 70, 5, 2, 63, 5, 156, 0, 0, 135, 5, 2, 63, 194, 5, 5, 139, 70, 5, 2, 63, 194, 5, 5, 139, 135, 5, 2, 63, 5, 150, 0, 0, 70, 5, 2, 63, 5, 150, 0, 0, 135, 5, 2, 63, 5, 160, 0, 0, 70, 5, 2, 63, 5, 160, 0, 0, 135, 5, 2, 63, 5, 206, 0, 0, 70, 5, 2, 63, 5, 206, 0, 0, 70, 5, 2, 63, 5, 210, 0, 0, 136, 5, 4, 69, 5, 142, 0, 0, 71, 5, 4, 69, 5, 142, 0, 0, 136, 5, 4, 69, 5, 144, 0, 0, 71, 5, 4, 69, 5, 144, 0, 0, 136, 5, 4, 69, 5, 156, 0, 0, 71, 5, 4, 69, 5, 156, 0, 0, 136, 5, 4, 69, 194, 5, 5, 139, 71, 5, 4, 69, 194, 5, 5, 139, 136, 5, 4, 69, 5, 150, 0, 0, 71, 5, 4, 69, 5, 150, 0, 0, 136, 5, 4, 69, 5, 160, 0, 0, 71, 5, 4, 69, 5, 160, 0, 0, 136, 5, 4, 69, 5, 206, 0, 0, 71, 5, 4, 69, 5, 206, 0, 0, 71, 5, 4, 69, 5, 210, 0, 0, 70, 5, 4, 69, 5, 142, 0, 0, 70, 5, 4, 69, 5, 144, 0, 0, 70, 5, 4, 69, 5, 156, 0, 0, 70, 5, 4, 69, 194, 5, 5, 139, 70, 5, 4, 69, 5, 150, 0, 0, 70, 5, 4, 69, 5, 160, 0, 0, 70, 5, 4, 69, 5, 206, 0, 0, 70, 5, 4, 69, 5, 210, 0, 0, 135, 5, 3, 69, 5, 142, 0, 0, 70, 5, 3, 69, 5, 142, 0, 0, 135, 5, 3, 69, 5, 140, 0, 0, 70, 5, 3, 69, 5, 140, 0, 0, 135, 5, 3, 69, 5, 156, 0, 0, 70, 5, 3, 69, 5, 156, 0, 0, 135, 5, 3, 69, 5, 160, 0, 0, 70, 5, 3, 69, 5, 160, 0, 0, 135, 5, 2, 69, 5, 142, 0, 0, 70, 5, 2, 69, 5, 142, 0, 0, 135, 5, 3, 69, 5, 144, 0, 0, 70, 5, 3, 69, 5, 144, 0, 0, 135, 5, 3, 69, 194, 5, 4, 139, 70, 5, 3, 69, 194, 5, 4, 139, 135, 5, 2, 69, 5, 144, 0, 0, 70, 5, 2, 69, 5, 144, 0, 0, 135, 5, 3, 69, 5, 164, 0, 0, 70, 5, 3, 69, 5, 164, 0, 0, 135, 5, 2, 69, 5, 156, 0, 0, 70, 5, 2, 69, 5, 156, 0, 0, 135, 5, 2, 69, 194, 5, 5, 139, 70, 5, 2, 69, 194, 5, 5, 139, 135, 5, 2, 69, 5, 150, 0, 0, 70, 5, 2, 69, 5, 150, 0, 0, 135, 5, 2, 69, 5, 160, 0, 0, 70, 5, 2, 69, 5, 160, 0, 0, 135, 5, 2, 69, 5, 206, 0, 0, 70, 5, 2, 69, 5, 206, 0, 0, 70, 5, 2, 69, 5, 210, 0, 0, 135, 5, 2, 73, 5, 142, 0, 0, 70, 5, 2, 73, 5, 142, 0, 0, 135, 5, 2, 73, 5, 144, 0, 0, 70, 5, 2, 73, 5, 144, 0, 0, 135, 5, 2, 73, 5, 156, 0, 0, 70, 5, 2, 73, 5, 156, 0, 0, 135, 5, 2, 73, 194, 5, 5, 139, 70, 5, 2, 73, 194, 5, 5, 139, 135, 5, 2, 73, 5, 150, 0, 0, 70, 5, 2, 73, 5, 150, 0, 0, 135, 5, 2, 73, 5, 160, 0, 0, 70, 5, 2, 73, 5, 160, 0, 0, 135, 5, 2, 73, 5, 206, 0, 0, 70, 5, 2, 73, 5, 206, 0, 0, 70, 5, 2, 73, 5, 210, 0, 0, 135, 5, 2, 75, 5, 150, 0, 0, 194, 5, 4, 139, 70, 5, 2, 75, 5, 150, 0, 0, 194, 5, 4, 139, 135, 5, 2, 75, 5, 138, 0, 0, 135, 5, 2, 75, 194, 5, 4, 139, 135, 5, 2, 75, 5, 142, 0, 0, 70, 5, 2, 75, 5, 138, 0, 0, 70, 5, 2, 75, 194, 5, 4, 139, 70, 5, 2, 75, 5, 142, 0, 0, 135, 5, 2, 75, 194, 5, 3, 139, 70, 5, 2, 75, 194, 5, 3, 139, 135, 5, 2, 75, 5, 164, 0, 0, 70, 5, 2, 75, 5, 164, 0, 0, 135, 5, 2, 75, 5, 140, 0, 0, 70, 5, 2, 75, 5, 140, 0, 0, 135, 5, 2, 75, 5, 146, 0, 0, 70, 5, 2, 75, 5, 146, 0, 0, 135, 5, 2, 75, 5, 152, 0, 0, 70, 5, 2, 75, 5, 152, 0, 0, 135, 5, 2, 75, 5, 162, 0, 0, 70, 5, 2, 75, 5, 162, 0, 0, 135, 5, 2, 75, 5, 190, 0, 0, 70, 5, 2, 75, 5, 190, 0, 0, 135, 5, 2, 75, 5, 144, 0, 0, 70, 5, 2, 75, 5, 144, 0, 0, 135, 5, 2, 75, 5, 150, 0, 0, 5, 164, 0, 0, 70, 5, 2, 75, 5, 150, 0, 0, 5, 164, 0, 0, 135, 5, 2, 75, 5, 150, 0, 0, 5, 144, 0, 0, 70, 5, 2, 75, 5, 150, 0, 0, 5, 144, 0, 0, 135, 5, 2, 75, 5, 150, 0, 0, 5, 138, 0, 0, 70, 5, 2, 75, 5, 150, 0, 0, 5, 138, 0, 0, 135, 5, 2, 75, 5, 184, 0, 0, 70, 5, 2, 75, 5, 184, 0, 0, 135, 5, 2, 75, 5, 188, 0, 0, 70, 5, 2, 75, 5, 188, 0, 0, 135, 5, 2, 75, 5, 198, 0, 0, 70, 5, 2, 75, 5, 198, 0, 0, 135, 5, 2, 75, 5, 208, 0, 0, 70, 5, 2, 75, 5, 208, 0, 0, 135, 5, 2, 75, 5, 204, 0, 0, 70, 5, 2, 75, 5, 204, 0, 0, 135, 5, 2, 75, 194, 5, 3, 139, 194, 5, 4, 139, 70, 5, 2, 75, 194, 5, 3, 139, 194, 5, 4, 139, 135, 5, 2, 75, 5, 164, 0, 0, 5, 150, 0, 0, 70, 5, 2, 75, 5, 164, 0, 0, 5, 150, 0, 0, 135, 5, 2, 75, 194, 5, 5, 139, 70, 5, 2, 75, 194, 5, 5, 139, 135, 5, 2, 75, 194, 5, 2, 139, 70, 5, 2, 75, 194, 5, 2, 139, 135, 5, 2, 75, 5, 190, 0, 0, 194, 5, 4, 139, 70, 5, 2, 75, 5, 190, 0, 0, 194, 5, 4, 139, 135, 5, 2, 75, 5, 190, 0, 0, 5, 138, 0, 0, 70, 5, 2, 75, 5, 190, 0, 0, 5, 138, 0, 0, 135, 5, 2, 75, 5, 190, 0, 0, 194, 5, 2, 139, 70, 5, 2, 75, 5, 190, 0, 0, 194, 5, 2, 139, 135, 5, 2, 75, 5, 190, 0, 0, 194, 5, 3, 139, 70, 5, 2, 75, 5, 190, 0, 0, 194, 5, 3, 139, 135, 5, 2, 75, 5, 190, 0, 0, 194, 5, 5, 139, 70, 5, 2, 75, 5, 190, 0, 0, 194, 5, 5, 139, 135, 5, 2, 81, 5, 142, 0, 0, 70, 5, 2, 81, 5, 142, 0, 0, 135, 5, 3, 81, 194, 5, 4, 139, 70, 5, 3, 81, 194, 5, 4, 139, 135, 5, 3, 81, 5, 160, 0, 0, 70, 5, 3, 81, 5, 160, 0, 0, 135, 5, 3, 81, 5, 144, 0, 0, 70, 5, 3, 81, 5, 144, 0, 0, 135, 5, 3, 81, 5, 184, 0, 0, 70, 5, 3, 81, 5, 184, 0, 0, 135, 5, 3, 81, 5, 188, 0, 0, 70, 5, 3, 81, 5, 188, 0, 0, 135, 5, 2, 81, 5, 144, 0, 0, 70, 5, 2, 81, 5, 144, 0, 0, 135, 5, 2, 81, 5, 156, 0, 0, 70, 5, 2, 81, 5, 156, 0, 0, 135, 5, 2, 81, 194, 5, 5, 139, 70, 5, 2, 81, 194, 5, 5, 139, 135, 5, 2, 81, 5, 150, 0, 0, 70, 5, 2, 81, 5, 150, 0, 0, 135, 5, 2, 81, 5, 160, 0, 0, 70, 5, 2, 81, 5, 160, 0, 0, 135, 5, 2, 81, 5, 206, 0, 0, 70, 5, 2, 81, 5, 206, 0, 0, 135, 5, 3, 81, 5, 156, 0, 0, 70, 5, 3, 81, 5, 156, 0, 0, 135, 5, 3, 81, 194, 5, 5, 139, 70, 5, 3, 81, 194, 5, 5, 139, 135, 5, 3, 81, 194, 5, 5, 139, 5, 164, 0, 0, 70, 5, 3, 81, 194, 5, 5, 139, 5, 164, 0, 0, 135, 5, 3, 81, 5, 210, 0, 0, 70, 5, 3, 81, 5, 210, 0, 0, 70, 5, 2, 81, 5, 210, 0, 0, 5, 5, 2, 47, 5, 142, 0, 0, 5, 5, 2, 47, 5, 144, 0, 0, 5, 5, 2, 47, 5, 156, 0, 0, 5, 5, 2, 47, 194, 5, 5, 139, 5, 5, 2, 47, 5, 150, 0, 0, 5, 5, 2, 47, 5, 160, 0, 0, 5, 5, 2, 47, 5, 206, 0, 0, 5, 5, 2, 47, 5, 210, 0, 0, 5, 5, 3, 55, 5, 150, 0, 0, 194, 5, 4, 139, 5, 5, 3, 55, 5, 138, 0, 0, 5, 5, 3, 55, 194, 5, 4, 139, 5, 5, 3, 55, 5, 142, 0, 0, 5, 5, 2, 55, 5, 142, 0, 0, 5, 5, 3, 55, 194, 5, 3, 139, 5, 5, 3, 55, 5, 164, 0, 0, 5, 5, 3, 55, 5, 140, 0, 0, 5, 5, 3, 55, 5, 162, 0, 0, 5, 5, 3, 55, 5, 144, 0, 0, 5, 5, 3, 55, 5, 184, 0, 0, 5, 5, 3, 55, 5, 188, 0, 0, 5, 5, 2, 55, 5, 144, 0, 0, 5, 5, 2, 55, 5, 156, 0, 0, 5, 5, 2, 55, 194, 5, 5, 139, 5, 5, 2, 55, 5, 150, 0, 0, 5, 5, 2, 55, 5, 160, 0, 0, 5, 5, 2, 55, 5, 206, 0, 0, 5, 5, 3, 55, 5, 208, 0, 0, 5, 5, 2, 55, 5, 210, 0, 0, 5, 5, 3, 55, 194, 5, 2, 139, 5, 5, 3, 55, 194, 5, 5, 139, 5, 5, 2, 63, 5, 142, 0, 0, 5, 5, 2, 63, 5, 144, 0, 0, 5, 5, 2, 63, 5, 156, 0, 0, 5, 5, 2, 63, 194, 5, 5, 139, 5, 5, 2, 63, 5, 150, 0, 0, 5, 5, 2, 63, 5, 160, 0, 0, 5, 5, 2, 63, 5, 206, 0, 0, 5, 5, 2, 63, 5, 210, 0, 0, 5, 5, 4, 69, 5, 142, 0, 0, 5, 5, 4, 69, 5, 144, 0, 0, 5, 5, 4, 69, 5, 156, 0, 0, 5, 5, 4, 69, 194, 5, 5, 139, 5, 5, 4, 69, 5, 150, 0, 0, 5, 5, 4, 69, 5, 160, 0, 0, 5, 5, 4, 69, 5, 206, 0, 0, 5, 5, 4, 69, 5, 210, 0, 0, 5, 5, 3, 69, 5, 142, 0, 0, 5, 5, 3, 69, 5, 140, 0, 0, 5, 5, 3, 69, 5, 156, 0, 0, 5, 5, 3, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 142, 0, 0, 5, 5, 3, 69, 5, 144, 0, 0, 5, 5, 3, 69, 194, 5, 4, 139, 5, 5, 2, 69, 5, 144, 0, 0, 5, 5, 3, 69, 5, 164, 0, 0, 5, 5, 2, 69, 5, 156, 0, 0, 5, 5, 2, 69, 194, 5, 5, 139, 5, 5, 2, 69, 5, 150, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 206, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 73, 5, 142, 0, 0, 5, 5, 2, 73, 5, 144, 0, 0, 5, 5, 2, 73, 5, 156, 0, 0, 5, 5, 2, 73, 194, 5, 5, 139, 5, 5, 2, 73, 5, 150, 0, 0, 5, 5, 2, 73, 5, 160, 0, 0, 5, 5, 2, 73, 5, 206, 0, 0, 5, 5, 2, 73, 5, 210, 0, 0, 5, 5, 2, 75, 5, 150, 0, 0, 194, 5, 4, 139, 5, 5, 2, 75, 5, 138, 0, 0, 5, 5, 2, 75, 194, 5, 4, 139, 5, 5, 2, 75, 5, 142, 0, 0, 5, 5, 2, 75, 194, 5, 3, 139, 5, 5, 2, 75, 5, 164, 0, 0, 5, 5, 2, 75, 5, 140, 0, 0, 5, 5, 2, 75, 5, 146, 0, 0, 5, 5, 2, 75, 5, 152, 0, 0, 5, 5, 2, 75, 5, 162, 0, 0, 5, 5, 2, 75, 5, 190, 0, 0, 5, 5, 2, 75, 5, 144, 0, 0, 5, 5, 2, 75, 5, 150, 0, 0, 5, 164, 0, 0, 5, 5, 2, 75, 5, 150, 0, 0, 5, 144, 0, 0, 5, 5, 2, 75, 5, 150, 0, 0, 5, 138, 0, 0, 5, 5, 2, 75, 5, 184, 0, 0, 5, 5, 2, 75, 5, 188, 0, 0, 5, 5, 2, 75, 5, 198, 0, 0, 5, 5, 2, 75, 5, 208, 0, 0, 5, 5, 2, 75, 5, 204, 0, 0, 5, 5, 2, 75, 194, 5, 3, 139, 194, 5, 4, 139, 5, 5, 2, 75, 5, 164, 0, 0, 5, 150, 0, 0, 5, 5, 2, 75, 194, 5, 5, 139, 5, 5, 2, 75, 194, 5, 2, 139, 5, 5, 2, 75, 5, 190, 0, 0, 194, 5, 4, 139, 5, 5, 2, 75, 5, 190, 0, 0, 5, 138, 0, 0, 5, 5, 2, 75, 5, 190, 0, 0, 194, 5, 2, 139, 5, 5, 2, 75, 5, 190, 0, 0, 194, 5, 3, 139, 5, 5, 2, 75, 5, 190, 0, 0, 194, 5, 5, 139, 5, 5, 2, 81, 5, 142, 0, 0, 5, 5, 3, 81, 194, 5, 4, 139, 5, 5, 3, 81, 5, 160, 0, 0, 5, 5, 3, 81, 5, 144, 0, 0, 5, 5, 3, 81, 5, 184, 0, 0, 5, 5, 3, 81, 5, 188, 0, 0, 5, 5, 2, 81, 5, 144, 0, 0, 5, 5, 2, 81, 5, 156, 0, 0, 5, 5, 2, 81, 194, 5, 5, 139, 5, 5, 2, 81, 5, 150, 0, 0, 5, 5, 2, 81, 5, 160, 0, 0, 5, 5, 2, 81, 5, 206, 0, 0, 5, 5, 3, 81, 5, 156, 0, 0, 5, 5, 3, 81, 194, 5, 5, 139, 5, 5, 3, 81, 194, 5, 5, 139, 5, 164, 0, 0, 5, 5, 3, 81, 5, 210, 0, 0, 5, 5, 2, 81, 5, 210, 0, 0, 156, 5, 0, 42, 194, 5, 4, 139, 156, 5, 0, 42, 194, 5, 3, 139, 156, 5, 0, 50, 194, 5, 4, 139, 156, 5, 0, 58, 194, 5, 4, 139, 156, 5, 0, 68, 194, 5, 3, 139, 156, 5, 0, 70, 194, 5, 4, 139, 156, 5, 0, 70, 194, 5, 3, 139, 156, 5, 0, 82, 194, 5, 4, 139, 156, 5, 0, 90, 194, 5, 4, 139, 5, 5, 0, 42, 194, 5, 4, 139, 5, 5, 0, 42, 194, 5, 3, 139, 5, 5, 0, 50, 194, 5, 4, 139, 5, 5, 0, 58, 194, 5, 4, 139, 5, 5, 0, 68, 194, 5, 3, 139, 5, 5, 0, 70, 194, 5, 4, 139, 5, 5, 0, 70, 194, 5, 3, 139, 5, 5, 0, 82, 194, 5, 4, 139, 5, 5, 0, 90, 194, 5, 4, 139, 134, 5, 3, 43, 5, 162, 0, 0, 134, 5, 2, 43, 5, 162, 0, 0, 5, 5, 3, 43, 5, 162, 0, 0, 5, 5, 2, 43, 5, 162, 0, 0, 156, 5, 0, 46, 194, 5, 4, 139, 5, 5, 0, 46, 194, 5, 4, 139, 134, 5, 2, 51, 5, 162, 0, 0, 5, 5, 2, 51, 5, 162, 0, 0, 156, 5, 0, 58, 194, 5, 3, 139, 5, 5, 0, 58, 194, 5, 3, 139, 156, 5, 0, 64, 194, 5, 4, 139, 5, 5, 0, 64, 194, 5, 4, 139, 156, 5, 0, 68, 194, 5, 4, 139, 5, 5, 0, 68, 194, 5, 4, 139, 156, 5, 0, 76, 194, 5, 4, 139, 5, 5, 0, 76, 194, 5, 4, 139, 156, 5, 0, 78, 194, 5, 4, 139, 5, 5, 0, 78, 194, 5, 4, 139, 156, 5, 0, 82, 194, 5, 3, 139, 5, 5, 0, 82, 194, 5, 3, 139, 134, 5, 2, 83, 5, 162, 0, 0, 5, 5, 2, 83, 5, 162, 0, 0, 156, 5, 0, 92, 194, 5, 4, 139, 5, 5, 0, 92, 194, 5, 4, 139, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 156, 5, 0, 82, 5, 150, 0, 0, 194, 5, 4, 139, 5, 5, 0, 82, 5, 150, 0, 0, 194, 5, 4, 139, 134, 5, 2, 71, 5, 162, 0, 0, 134, 5, 3, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 5, 5, 3, 71, 5, 162, 0, 0, 156, 5, 0, 54, 194, 5, 4, 139, 5, 5, 0, 54, 194, 5, 4, 139, 156, 5, 0, 42, 5, 146, 0, 0, 194, 5, 4, 139, 5, 5, 0, 42, 5, 146, 0, 0, 194, 5, 4, 139, 160, 112, 0, 42, 162, 5, 0, 50, 194, 5, 4, 139, 5, 112, 0, 42, 19, 5, 0, 50, 194, 5, 4, 139, 156, 5, 0, 70, 5, 158, 0, 0, 194, 5, 4, 139, 5, 5, 0, 70, 5, 158, 0, 0, 194, 5, 4, 139, 134, 5, 2, 51, 5, 160, 0, 0, 5, 5, 2, 51, 5, 160, 0, 0, 156, 5, 0, 70, 194, 5, 3, 139, 5, 164, 0, 0, 5, 5, 0, 70, 194, 5, 3, 139, 5, 164, 0, 0, 5, 5, 16, 12, 194, 5, 4, 139, 166, 5, 6, 96, 194, 5, 4, 139, 166, 5, 16, 96, 194, 5, 4, 139, 158, 5, 28, 96, 194, 5, 4, 139, 166, 5, 32, 96, 194, 5, 4, 139, 158, 5, 48, 96, 194, 5, 4, 139, 155, 5, 76, 96, 194, 5, 4, 139, 155, 5, 86, 96, 194, 5, 4, 139, 5, 5, 32, 96, 5, 150, 0, 0, 194, 5, 4, 139, 5, 5, 6, 96, 194, 5, 4, 139, 5, 5, 16, 96, 194, 5, 4, 139, 5, 5, 28, 96, 194, 5, 4, 139, 5, 5, 32, 96, 194, 5, 4, 139, 5, 5, 76, 96, 5, 150, 0, 0, 194, 5, 4, 139, 5, 5, 48, 96, 194, 5, 4, 139, 5, 5, 76, 96, 194, 5, 4, 139, 5, 5, 86, 96, 194, 5, 4, 139, 166, 5, 76, 96, 194, 5, 4, 139, 174, 5, 16, 97, 194, 5, 4, 139, 174, 5, 84, 97, 194, 5, 4, 139, 5, 5, 16, 97, 194, 5, 4, 139, 5, 5, 84, 97, 194, 5, 4, 139, 134, 5, 3, 43, 5, 200, 0, 0, 134, 5, 2, 43, 5, 200, 0, 0, 5, 5, 3, 43, 5, 200, 0, 0, 5, 5, 2, 43, 5, 200, 0, 0, 156, 5, 0, 44, 194, 5, 5, 139, 5, 5, 0, 44, 194, 5, 5, 139, 156, 5, 0, 46, 5, 160, 0, 0, 194, 5, 4, 139, 5, 5, 0, 46, 5, 160, 0, 0, 194, 5, 4, 139, 156, 5, 0, 48, 194, 5, 5, 139, 5, 5, 0, 48, 194, 5, 5, 139, 156, 5, 0, 50, 5, 164, 0, 0, 194, 5, 4, 139, 5, 5, 0, 50, 5, 164, 0, 0, 194, 5, 4, 139, 134, 5, 2, 51, 5, 204, 0, 0, 5, 5, 2, 51, 5, 204, 0, 0, 134, 5, 2, 51, 5, 208, 0, 0, 5, 5, 2, 51, 5, 208, 0, 0, 156, 5, 0, 56, 194, 5, 5, 139, 5, 5, 0, 56, 194, 5, 5, 139, 156, 5, 0, 58, 5, 150, 0, 0, 194, 5, 4, 139, 5, 5, 0, 58, 5, 150, 0, 0, 194, 5, 4, 139, 156, 5, 0, 62, 194, 5, 4, 139, 5, 5, 0, 62, 194, 5, 4, 139, 156, 5, 0, 62, 194, 5, 5, 139, 5, 5, 0, 62, 194, 5, 5, 139, 156, 5, 0, 64, 194, 5, 5, 139, 5, 5, 0, 64, 194, 5, 5, 139, 156, 5, 0, 64, 194, 5, 5, 139, 5, 164, 0, 0, 5, 5, 0, 64, 194, 5, 5, 139, 5, 164, 0, 0, 156, 5, 0, 66, 194, 5, 4, 139, 5, 5, 0, 66, 194, 5, 4, 139, 156, 5, 0, 66, 194, 5, 5, 139, 5, 5, 0, 66, 194, 5, 5, 139, 156, 5, 0, 68, 194, 5, 5, 139, 5, 5, 0, 68, 194, 5, 5, 139, 156, 5, 0, 70, 194, 5, 3, 139, 194, 5, 4, 139, 5, 5, 0, 70, 194, 5, 3, 139, 194, 5, 4, 139, 156, 5, 0, 70, 194, 5, 3, 139, 5, 150, 0, 0, 5, 5, 0, 70, 194, 5, 3, 139, 5, 150, 0, 0, 156, 5, 0, 70, 5, 164, 0, 0, 194, 5, 4, 139, 5, 5, 0, 70, 5, 164, 0, 0, 194, 5, 4, 139, 156, 5, 0, 72, 194, 5, 4, 139, 5, 5, 0, 72, 194, 5, 4, 139, 156, 5, 0, 76, 194, 5, 5, 139, 5, 5, 0, 76, 194, 5, 5, 139, 156, 5, 0, 76, 194, 5, 5, 139, 5, 164, 0, 0, 5, 5, 0, 76, 194, 5, 5, 139, 5, 164, 0, 0, 156, 5, 0, 78, 194, 5, 5, 139, 5, 5, 0, 78, 194, 5, 5, 139, 156, 5, 0, 78, 194, 5, 4, 139, 5, 156, 0, 0, 5, 5, 0, 78, 194, 5, 4, 139, 5, 156, 0, 0, 156, 5, 0, 78, 194, 5, 5, 139, 5, 156, 0, 0, 5, 5, 0, 78, 194, 5, 5, 139, 5, 156, 0, 0, 156, 5, 0, 80, 194, 5, 5, 139, 5, 5, 0, 80, 194, 5, 5, 139, 156, 5, 0, 82, 194, 5, 3, 139, 194, 5, 4, 139, 5, 5, 0, 82, 194, 5, 3, 139, 194, 5, 4, 139, 156, 5, 0, 84, 194, 5, 3, 139, 5, 5, 0, 84, 194, 5, 3, 139, 156, 5, 0, 84, 194, 5, 5, 139, 5, 5, 0, 84, 194, 5, 5, 139, 160, 5, 0, 86, 194, 5, 4, 139, 5, 5, 0, 86, 194, 5, 4, 139, 160, 5, 0, 86, 194, 5, 5, 139, 5, 5, 0, 86, 194, 5, 5, 139, 156, 5, 0, 92, 194, 5, 5, 139, 5, 5, 0, 92, 194, 5, 5, 139, 156, 5, 0, 42, 194, 5, 5, 139, 134, 5, 3, 43, 194, 5, 5, 139, 134, 5, 2, 43, 194, 5, 5, 139, 5, 5, 0, 42, 194, 5, 5, 139, 5, 5, 3, 43, 194, 5, 5, 139, 5, 5, 2, 43, 194, 5, 5, 139, 156, 5, 0, 42, 194, 5, 2, 139, 5, 5, 0, 42, 194, 5, 2, 139, 134, 5, 3, 43, 194, 5, 4, 139, 5, 5, 3, 43, 194, 5, 4, 139, 134, 5, 3, 43, 5, 138, 0, 0, 5, 5, 3, 43, 5, 138, 0, 0, 134, 5, 3, 43, 194, 5, 2, 139, 5, 5, 3, 43, 194, 5, 2, 139, 134, 5, 3, 43, 194, 5, 3, 139, 5, 5, 3, 43, 194, 5, 3, 139, 134, 5, 2, 43, 194, 5, 4, 139, 5, 5, 2, 43, 194, 5, 4, 139, 134, 5, 2, 43, 5, 138, 0, 0, 5, 5, 2, 43, 5, 138, 0, 0, 134, 5, 2, 43, 194, 5, 2, 139, 5, 5, 2, 43, 194, 5, 2, 139, 134, 5, 2, 43, 194, 5, 3, 139, 5, 5, 2, 43, 194, 5, 3, 139, 156, 5, 0, 50, 194, 5, 5, 139, 134, 5, 2, 51, 194, 5, 5, 139, 5, 5, 0, 50, 194, 5, 5, 139, 5, 5, 2, 51, 194, 5, 5, 139, 156, 5, 0, 50, 194, 5, 2, 139, 5, 5, 0, 50, 194, 5, 2, 139, 156, 5, 0, 50, 194, 5, 3, 139, 5, 5, 0, 50, 194, 5, 3, 139, 134, 5, 2, 51, 194, 5, 4, 139, 5, 5, 2, 51, 194, 5, 4, 139, 134, 5, 2, 51, 5, 138, 0, 0, 5, 5, 2, 51, 5, 138, 0, 0, 134, 5, 2, 51, 194, 5, 2, 139, 5, 5, 2, 51, 194, 5, 2, 139, 134, 5, 2, 51, 194, 5, 3, 139, 5, 5, 2, 51, 194, 5, 3, 139, 156, 5, 0, 58, 194, 5, 2, 139, 5, 5, 0, 58, 194, 5, 2, 139, 156, 5, 0, 58, 194, 5, 5, 139, 5, 5, 0, 58, 194, 5, 5, 139, 156, 5, 0, 70, 194, 5, 5, 139, 134, 5, 2, 71, 194, 5, 5, 139, 5, 5, 0, 70, 194, 5, 5, 139, 5, 5, 2, 71, 194, 5, 5, 139, 156, 5, 0, 70, 194, 5, 2, 139, 5, 5, 0, 70, 194, 5, 2, 139, 134, 5, 2, 71, 194, 5, 4, 139, 5, 5, 2, 71, 194, 5, 4, 139, 134, 5, 2, 71, 5, 138, 0, 0, 5, 5, 2, 71, 5, 138, 0, 0, 134, 5, 2, 71, 194, 5, 2, 139, 5, 5, 2, 71, 194, 5, 2, 139, 134, 5, 2, 71, 194, 5, 3, 139, 5, 5, 2, 71, 194, 5, 3, 139, 134, 5, 3, 71, 194, 5, 4, 139, 5, 5, 3, 71, 194, 5, 4, 139, 134, 5, 3, 71, 5, 138, 0, 0, 5, 5, 3, 71, 5, 138, 0, 0, 134, 5, 3, 71, 194, 5, 2, 139, 5, 5, 3, 71, 194, 5, 2, 139, 134, 5, 3, 71, 194, 5, 3, 139, 5, 5, 3, 71, 194, 5, 3, 139, 134, 5, 3, 71, 194, 5, 5, 139, 134, 5, 2, 71, 5, 190, 0, 0, 194, 5, 5, 139, 5, 5, 3, 71, 194, 5, 5, 139, 5, 5, 2, 71, 5, 190, 0, 0, 194, 5, 5, 139, 156, 5, 0, 82, 194, 5, 5, 139, 5, 5, 0, 82, 194, 5, 5, 139, 156, 5, 0, 82, 194, 5, 2, 139, 5, 5, 0, 82, 194, 5, 2, 139, 134, 5, 2, 83, 194, 5, 4, 139, 5, 5, 2, 83, 194, 5, 4, 139, 134, 5, 2, 83, 5, 138, 0, 0, 5, 5, 2, 83, 5, 138, 0, 0, 134, 5, 2, 83, 194, 5, 2, 139, 5, 5, 2, 83, 194, 5, 2, 139, 134, 5, 2, 83, 194, 5, 3, 139, 5, 5, 2, 83, 194, 5, 3, 139, 134, 5, 2, 83, 194, 5, 5, 139, 5, 5, 2, 83, 194, 5, 5, 139, 156, 5, 0, 90, 194, 5, 5, 139, 5, 5, 0, 90, 194, 5, 5, 139, 156, 5, 0, 90, 194, 5, 2, 139, 5, 5, 0, 90, 194, 5, 2, 139, 156, 5, 0, 90, 194, 5, 3, 139, 5, 5, 0, 90, 194, 5, 3, 139, 5, 5, 6, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 6, 96, 5, 134, 0, 0, 194, 5, 4, 139, 166, 5, 6, 96, 5, 132, 0, 0, 194, 5, 4, 139, 166, 5, 6, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 16, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 16, 96, 5, 134, 0, 0, 194, 5, 4, 139, 166, 5, 16, 96, 5, 132, 0, 0, 194, 5, 4, 139, 166, 5, 16, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 28, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 28, 96, 5, 134, 0, 0, 194, 5, 4, 139, 158, 5, 28, 96, 5, 132, 0, 0, 194, 5, 4, 139, 158, 5, 28, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 32, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 32, 96, 5, 134, 0, 0, 194, 5, 4, 139, 166, 5, 32, 96, 5, 132, 0, 0, 194, 5, 4, 139, 166, 5, 32, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 48, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 48, 96, 5, 134, 0, 0, 194, 5, 4, 139, 158, 5, 48, 96, 5, 132, 0, 0, 194, 5, 4, 139, 158, 5, 48, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 76, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 76, 96, 5, 134, 0, 0, 194, 5, 4, 139, 155, 5, 76, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 86, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 5, 86, 96, 5, 134, 0, 0, 194, 5, 4, 139, 155, 5, 86, 96, 5, 132, 0, 0, 194, 5, 4, 139, 155, 5, 86, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 5, 6, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 6, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 166, 5, 6, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 166, 5, 6, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 28, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 28, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 158, 5, 28, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 158, 5, 28, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 86, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 86, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 155, 5, 86, 96, 5, 132, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 155, 5, 86, 96, 5, 134, 0, 0, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 6, 96, 194, 5, 4, 139, 5, 216, 0, 0, 5, 5, 28, 96, 194, 5, 4, 139, 5, 216, 0, 0, 193, 9, 22, 12, 194, 5, 4, 139, 193, 16, 22, 12, 194, 5, 4, 139, 5, 5, 86, 96, 194, 5, 4, 139, 5, 216, 0, 0, 5, 112, 0, 78, 19, 5, 0, 78, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 1, 0, 2, 3, 3, 235, 134, 5, 6, 3, 2, 235, 134, 5, 0, 46, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 2, 192, 197, 162, 41, 30, 2, 192, 197, 226, 42, 30, 3, 192, 197, 34, 43, 30, 3, 192, 197, 98, 150, 30, 3, 192, 197, 162, 36, 30, 1, 192, 197, 162, 37, 30, 1, 192, 197, 226, 38, 30, 2, 192, 197, 34, 39, 30, 2, 192, 197, 98, 30, 2, 12, 0, 30, 2, 0, 192, 197, 162, 31, 2, 0, 192, 197, 226, 34, 30, 1, 192, 197, 34, 35, 30, 1, 192, 197, 98, 72, 0, 2, 239, 135, 5, 104, 0, 2, 239, 70, 5, 36, 1, 197, 162, 37, 1, 0, 192, 197, 98, 0, 50, 156, 5, 48, 0, 2, 3, 2, 243, 134, 5, 0, 54, 156, 5, 0, 0, 49, 0, 8, 2, 89, 0, 40, 30, 45, 0, 46, 30, 23, 0, 200, 30, 12, 0, 200, 30, 14, 192, 197, 34, 201, 30, 14, 192, 197, 98, 202, 30, 14, 192, 197, 162, 203, 30, 14, 192, 197, 226, 46, 30, 3, 192, 197, 227, 47, 30, 4, 192, 197, 67, 150, 30, 13, 192, 197, 226, 43, 30, 9, 0, 43, 30, 13, 192, 197, 34, 44, 30, 13, 192, 197, 98, 45, 30, 13, 192, 197, 162, 40, 30, 12, 192, 197, 98, 41, 30, 12, 192, 197, 162, 42, 30, 12, 192, 197, 226, 34, 30, 20, 0, 37, 30, 9, 0, 37, 30, 11, 192, 197, 162, 38, 30, 11, 192, 197, 226, 39, 30, 12, 192, 197, 34, 34, 30, 10, 192, 197, 226, 35, 30, 11, 192, 197, 34, 36, 30, 11, 192, 197, 98, 11, 2, 9, 0, 11, 2, 10, 192, 197, 34, 30, 2, 10, 192, 197, 98, 31, 2, 10, 192, 197, 162, 8, 2, 9, 192, 197, 98, 9, 2, 9, 192, 197, 162, 10, 2, 9, 192, 197, 226, 36, 1, 45, 0, 44, 1, 23, 0, 47, 1, 12, 0, 47, 1, 8, 192, 197, 98, 48, 1, 8, 192, 197, 162, 207, 1, 8, 192, 197, 226, 208, 1, 9, 192, 197, 34, 44, 1, 7, 192, 197, 162, 45, 1, 7, 192, 197, 226, 46, 1, 8, 192, 197, 34, 41, 1, 9, 0, 41, 1, 6, 192, 197, 226, 42, 1, 7, 192, 197, 34, 43, 1, 7, 192, 197, 98, 36, 1, 6, 192, 197, 34, 37, 1, 6, 192, 197, 98, 40, 1, 6, 192, 197, 162, 206, 0, 20, 0, 237, 0, 9, 0, 237, 0, 5, 192, 197, 162, 238, 0, 5, 192, 197, 226, 239, 0, 4, 192, 197, 66, 206, 0, 5, 192, 197, 34, 207, 0, 3, 192, 197, 226, 236, 0, 5, 192, 197, 98, 105, 0, 14, 0, 105, 0, 6, 0, 204, 0, 4, 192, 197, 162, 205, 0, 4, 192, 197, 226, 240, 127, 3, 55, 70, 5, 68, 3, 4, 192, 197, 67, 72, 0, 2, 247, 135, 5, 73, 0, 3, 0, 104, 0, 2, 247, 70, 5, 240, 127, 3, 55, 135, 5, 68, 3, 3, 192, 197, 227, 0, 62, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 17, 192, 197, 162, 41, 30, 17, 192, 197, 226, 42, 30, 18, 192, 197, 34, 43, 30, 18, 192, 197, 98, 150, 30, 18, 192, 197, 162, 36, 30, 16, 192, 197, 162, 37, 30, 16, 192, 197, 226, 38, 30, 17, 192, 197, 34, 39, 30, 17, 192, 197, 98, 30, 2, 12, 0, 30, 2, 15, 192, 197, 162, 31, 2, 15, 192, 197, 226, 34, 30, 16, 192, 197, 34, 35, 30, 16, 192, 197, 98, 72, 0, 2, 255, 135, 5, 104, 0, 2, 255, 70, 5, 36, 1, 15, 192, 197, 34, 37, 1, 15, 192, 197, 98, 0, 68, 156, 5, 32, 0, 244, 1, 57, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 30, 192, 197, 162, 41, 30, 30, 192, 197, 226, 42, 30, 31, 192, 197, 34, 43, 30, 31, 192, 197, 98, 150, 30, 31, 192, 197, 162, 36, 30, 29, 192, 197, 162, 37, 30, 29, 192, 197, 226, 38, 30, 30, 192, 197, 34, 39, 30, 30, 192, 197, 98, 32, 30, 12, 0, 32, 30, 28, 192, 197, 162, 33, 30, 28, 192, 197, 226, 34, 30, 29, 192, 197, 34, 35, 30, 29, 192, 197, 98, 244, 1, 27, 192, 197, 162, 245, 1, 27, 192, 197, 226, 30, 2, 28, 192, 197, 34, 31, 2, 28, 192, 197, 98, 32, 1, 26, 0, 36, 1, 12, 0, 36, 1, 26, 192, 197, 162, 37, 1, 26, 192, 197, 226, 230, 1, 27, 192, 197, 34, 231, 1, 27, 192, 197, 98, 32, 1, 25, 192, 197, 162, 33, 1, 25, 192, 197, 226, 34, 1, 26, 192, 197, 34, 35, 1, 26, 192, 197, 98, 28, 1, 12, 0, 28, 1, 24, 192, 197, 162, 29, 1, 24, 192, 197, 226, 30, 1, 25, 192, 197, 34, 31, 1, 25, 192, 197, 98, 71, 0, 10, 0, 72, 0, 255, 255, 2, 69, 135, 5, 103, 0, 66, 0, 104, 0, 255, 255, 2, 69, 70, 5, 208, 127, 3, 69, 135, 5, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 21, 192, 197, 98, 41, 30, 21, 192, 197, 162, 42, 30, 21, 192, 197, 226, 43, 30, 22, 192, 197, 34, 150, 30, 22, 192, 197, 98, 36, 30, 20, 192, 197, 98, 37, 30, 20, 192, 197, 162, 38, 30, 20, 192, 197, 226, 39, 30, 21, 192, 197, 34, 30, 2, 12, 0, 30, 2, 19, 192, 197, 98, 31, 2, 19, 192, 197, 162, 34, 30, 19, 192, 197, 226, 35, 30, 20, 192, 197, 34, 72, 0, 255, 255, 4, 69, 136, 5, 104, 0, 255, 255, 4, 69, 71, 5, 36, 1, 18, 192, 197, 226, 37, 1, 19, 192, 197, 34, 200, 127, 3, 69, 70, 5, 37, 30, 15, 0, 37, 30, 23, 192, 197, 98, 39, 30, 23, 192, 197, 162, 41, 30, 23, 192, 197, 226, 43, 30, 24, 192, 197, 34, 150, 30, 24, 192, 197, 98, 104, 0, 255, 255, 4, 69, 70, 5, 37, 1, 22, 192, 197, 162, 31, 2, 22, 192, 197, 226, 35, 30, 23, 192, 197, 34, 0, 70, 156, 5, 1, 0, 2, 3, 255, 255, 2, 71, 134, 5, 27, 3, 255, 255, 3, 71, 134, 5, 0, 72, 156, 5, 16, 0, 36, 30, 29, 0, 40, 30, 15, 0, 40, 30, 34, 192, 197, 98, 41, 30, 34, 192, 197, 162, 42, 30, 34, 192, 197, 226, 43, 30, 35, 192, 197, 34, 150, 30, 35, 192, 197, 98, 36, 30, 33, 192, 197, 98, 37, 30, 33, 192, 197, 162, 38, 30, 33, 192, 197, 226, 39, 30, 34, 192, 197, 34, 30, 2, 12, 0, 30, 2, 32, 192, 197, 98, 31, 2, 32, 192, 197, 162, 34, 30, 32, 192, 197, 226, 35, 30, 33, 192, 197, 34, 72, 0, 255, 255, 2, 73, 135, 5, 104, 0, 255, 255, 2, 73, 70, 5, 36, 1, 31, 192, 197, 226, 37, 1, 32, 192, 197, 34, 0, 74, 160, 5, 0, 0, 61, 0, 218, 1, 107, 0, 122, 30, 54, 0, 234, 30, 26, 0, 238, 30, 12, 0, 238, 30, 51, 192, 197, 99, 239, 30, 51, 192, 197, 195, 240, 30, 52, 192, 197, 35, 241, 30, 52, 192, 197, 131, 234, 30, 49, 192, 197, 227, 235, 30, 50, 192, 197, 67, 236, 30, 50, 192, 197, 163, 237, 30, 51, 192, 197, 3, 230, 30, 12, 0, 230, 30, 48, 192, 197, 162, 231, 30, 48, 192, 197, 226, 232, 30, 49, 192, 197, 35, 233, 30, 49, 192, 197, 131, 122, 30, 47, 192, 197, 99, 123, 30, 47, 192, 197, 195, 228, 30, 48, 192, 197, 34, 229, 30, 48, 192, 197, 98, 114, 30, 26, 0, 118, 30, 12, 0, 118, 30, 46, 192, 197, 34, 119, 30, 46, 192, 197, 98, 120, 30, 46, 192, 197, 163, 121, 30, 47, 192, 197, 3, 114, 30, 45, 192, 197, 34, 115, 30, 45, 192, 197, 98, 116, 30, 45, 192, 197, 162, 117, 30, 45, 192, 197, 226, 20, 2, 12, 0, 20, 2, 44, 192, 197, 34, 21, 2, 44, 192, 197, 98, 22, 2, 44, 192, 197, 162, 23, 2, 44, 192, 197, 226, 218, 1, 43, 192, 197, 3, 219, 1, 43, 192, 197, 99, 220, 1, 43, 192, 197, 195, 109, 1, 54, 0, 176, 1, 26, 0, 214, 1, 12, 0, 214, 1, 42, 192, 197, 67, 215, 1, 35, 192, 197, 163, 216, 1, 36, 192, 197, 3, 217, 1, 42, 192, 197, 163, 176, 1, 41, 192, 197, 34, 211, 1, 41, 192, 197, 98, 212, 1, 41, 192, 197, 162, 213, 1, 41, 192, 197, 227, 113, 1, 12, 0, 113, 1, 40, 192, 197, 34, 114, 1, 40, 192, 197, 98, 115, 1, 40, 192, 197, 162, 175, 1, 40, 192, 197, 226, 109, 1, 39, 192, 197, 34, 110, 1, 39, 192, 197, 98, 111, 1, 39, 192, 197, 162, 112, 1, 39, 192, 197, 226, 250, 0, 26, 0, 105, 1, 12, 0, 105, 1, 38, 192, 197, 34, 106, 1, 38, 192, 197, 98, 107, 1, 38, 192, 197, 162, 108, 1, 38, 192, 197, 226, 250, 0, 37, 192, 197, 98, 251, 0, 37, 192, 197, 162, 252, 0, 36, 192, 197, 2, 104, 1, 37, 192, 197, 226, 218, 0, 12, 0, 218, 0, 36, 192, 197, 162, 219, 0, 36, 192, 197, 226, 220, 0, 35, 192, 197, 162, 249, 0, 37, 192, 197, 34, 85, 0, 5, 0, 117, 0, 9, 0, 217, 0, 36, 192, 197, 98, 240, 127, 2, 75, 135, 5, 68, 3, 35, 192, 197, 163, 240, 127, 2, 75, 70, 5, 68, 3, 36, 192, 197, 3, 0, 80, 156, 5, 36, 0, 34, 30, 63, 0, 43, 30, 32, 0, 92, 30, 15, 0, 92, 30, 59, 192, 197, 227, 93, 30, 60, 192, 197, 67, 94, 30, 60, 192, 197, 162, 95, 30, 60, 192, 197, 226, 150, 30, 61, 192, 197, 34, 43, 30, 58, 192, 197, 162, 88, 30, 58, 192, 197, 226, 89, 30, 59, 192, 197, 34, 90, 30, 59, 192, 197, 98, 91, 30, 59, 192, 197, 162, 38, 30, 15, 0, 38, 30, 57, 192, 197, 98, 39, 30, 57, 192, 197, 162, 40, 30, 57, 192, 197, 226, 41, 30, 58, 192, 197, 34, 42, 30, 58, 192, 197, 98, 34, 30, 56, 192, 197, 98, 35, 30, 56, 192, 197, 162, 36, 30, 56, 192, 197, 226, 37, 30, 57, 192, 197, 34, 87, 1, 29, 0, 17, 2, 15, 0, 17, 2, 55, 192, 197, 34, 18, 2, 55, 192, 197, 98, 19, 2, 55, 192, 197, 162, 30, 2, 55, 192, 197, 226, 31, 2, 56, 192, 197, 34, 87, 1, 54, 192, 197, 34, 88, 1, 54, 192, 197, 98, 89, 1, 54, 192, 197, 162, 16, 2, 54, 192, 197, 226, 36, 1, 15, 0, 36, 1, 52, 192, 197, 226, 37, 1, 53, 192, 197, 34, 84, 1, 53, 192, 197, 98, 85, 1, 53, 192, 197, 162, 86, 1, 53, 192, 197, 226, 72, 0, 255, 255, 2, 81, 135, 5, 82, 0, 255, 255, 3, 81, 135, 5, 104, 0, 255, 255, 2, 81, 70, 5, 114, 0, 255, 255, 3, 81, 70, 5, 0, 82, 156, 5, 48, 0, 27, 3, 255, 255, 2, 83, 134, 5, 0, 42, 5, 5, 1, 0, 2, 3, 3, 235, 5, 5, 6, 3, 2, 235, 5, 5, 0, 46, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 62, 192, 197, 34, 39, 30, 62, 192, 197, 98, 41, 30, 62, 192, 197, 162, 43, 30, 62, 192, 197, 226, 150, 30, 63, 192, 197, 34, 104, 0, 2, 239, 5, 5, 37, 1, 61, 192, 197, 98, 31, 2, 61, 192, 197, 162, 35, 30, 61, 192, 197, 226, 0, 50, 5, 5, 48, 0, 2, 3, 2, 243, 5, 5, 0, 54, 5, 5, 24, 0, 9, 2, 45, 0, 41, 30, 23, 0, 47, 30, 12, 0, 47, 30, 63, 192, 197, 99, 150, 30, 68, 192, 197, 66, 201, 30, 68, 192, 197, 130, 203, 30, 68, 192, 197, 194, 41, 30, 67, 192, 197, 130, 43, 30, 67, 192, 197, 194, 45, 30, 68, 192, 197, 2, 35, 30, 9, 0, 35, 30, 66, 192, 197, 194, 37, 30, 67, 192, 197, 2, 39, 30, 67, 192, 197, 66, 9, 2, 66, 192, 197, 2, 11, 2, 66, 192, 197, 66, 31, 2, 66, 192, 197, 130, 37, 1, 20, 0, 45, 1, 9, 0, 45, 1, 65, 192, 197, 66, 47, 1, 65, 192, 197, 130, 208, 1, 65, 192, 197, 194, 37, 1, 64, 192, 197, 130, 41, 1, 64, 192, 197, 194, 43, 1, 65, 192, 197, 2, 237, 0, 9, 0, 237, 0, 64, 192, 197, 2, 238, 0, 64, 192, 197, 66, 239, 0, 63, 192, 197, 98, 104, 0, 2, 247, 5, 5, 105, 0, 3, 0, 236, 0, 63, 192, 197, 194, 240, 127, 3, 55, 5, 5, 68, 3, 63, 192, 197, 99, 0, 62, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 69, 192, 197, 194, 39, 30, 70, 192, 197, 2, 41, 30, 70, 192, 197, 66, 43, 30, 70, 192, 197, 130, 150, 30, 70, 192, 197, 194, 104, 0, 2, 255, 5, 5, 37, 1, 69, 192, 197, 2, 31, 2, 69, 192, 197, 66, 35, 30, 69, 192, 197, 130, 0, 68, 5, 5, 16, 0, 245, 1, 29, 0, 37, 30, 15, 0, 37, 30, 75, 192, 197, 130, 39, 30, 75, 192, 197, 194, 41, 30, 76, 192, 197, 2, 43, 30, 76, 192, 197, 66, 150, 30, 76, 192, 197, 130, 245, 1, 74, 192, 197, 130, 31, 2, 74, 192, 197, 194, 33, 30, 75, 192, 197, 2, 35, 30, 75, 192, 197, 66, 33, 1, 12, 0, 33, 1, 73, 192, 197, 130, 35, 1, 73, 192, 197, 194, 37, 1, 74, 192, 197, 2, 231, 1, 74, 192, 197, 66, 103, 0, 10, 0, 104, 0, 255, 255, 2, 69, 5, 5, 29, 1, 73, 192, 197, 2, 31, 1, 73, 192, 197, 66, 200, 127, 3, 69, 5, 5, 37, 30, 15, 0, 37, 30, 71, 192, 197, 194, 39, 30, 72, 192, 197, 2, 41, 30, 72, 192, 197, 66, 43, 30, 72, 192, 197, 130, 150, 30, 72, 192, 197, 194, 104, 0, 255, 255, 4, 69, 5, 5, 37, 1, 71, 192, 197, 2, 31, 2, 71, 192, 197, 66, 35, 30, 71, 192, 197, 130, 0, 70, 5, 5, 1, 0, 2, 3, 255, 255, 2, 71, 5, 5, 27, 3, 255, 255, 3, 71, 5, 5, 0, 72, 5, 5, 8, 0, 37, 30, 15, 0, 37, 30, 77, 192, 197, 130, 39, 30, 77, 192, 197, 194, 41, 30, 78, 192, 197, 2, 43, 30, 78, 192, 197, 66, 150, 30, 78, 192, 197, 130, 104, 0, 255, 255, 2, 73, 5, 5, 37, 1, 76, 192, 197, 194, 31, 2, 77, 192, 197, 2, 35, 30, 77, 192, 197, 66, 0, 74, 5, 5, 30, 0, 218, 1, 54, 0, 123, 30, 26, 0, 235, 30, 12, 0, 235, 30, 85, 192, 197, 227, 237, 30, 86, 192, 197, 67, 239, 30, 86, 192, 197, 163, 241, 30, 87, 192, 197, 3, 123, 30, 84, 192, 197, 163, 229, 30, 85, 192, 197, 2, 231, 30, 85, 192, 197, 66, 233, 30, 85, 192, 197, 131, 115, 30, 12, 0, 115, 30, 83, 192, 197, 130, 117, 30, 83, 192, 197, 194, 119, 30, 84, 192, 197, 2, 121, 30, 84, 192, 197, 67, 218, 1, 82, 192, 197, 67, 220, 1, 82, 192, 197, 163, 21, 2, 83, 192, 197, 2, 23, 2, 83, 192, 197, 66, 109, 1, 26, 0, 176, 1, 12, 0, 176, 1, 81, 192, 197, 98, 212, 1, 81, 192, 197, 162, 214, 1, 81, 192, 197, 227, 216, 1, 78, 192, 197, 195, 109, 1, 80, 192, 197, 98, 111, 1, 80, 192, 197, 162, 113, 1, 80, 192, 197, 226, 115, 1, 81, 192, 197, 34, 251, 0, 12, 0, 251, 0, 79, 192, 197, 162, 252, 0, 78, 192, 197, 194, 105, 1, 79, 192, 197, 226, 107, 1, 80, 192, 197, 34, 117, 0, 6, 0, 249, 0, 79, 192, 197, 34, 250, 0, 79, 192, 197, 98, 240, 127, 2, 75, 5, 5, 68, 3, 78, 192, 197, 195, 0, 80, 5, 5, 18, 0, 35, 30, 32, 0, 89, 30, 15, 0, 89, 30, 90, 192, 197, 98, 91, 30, 90, 192, 197, 162, 93, 30, 90, 192, 197, 227, 95, 30, 91, 192, 197, 66, 150, 30, 91, 192, 197, 130, 35, 30, 89, 192, 197, 34, 37, 30, 89, 192, 197, 98, 39, 30, 89, 192, 197, 162, 41, 30, 89, 192, 197, 226, 43, 30, 90, 192, 197, 34, 87, 1, 15, 0, 87, 1, 87, 192, 197, 226, 89, 1, 88, 192, 197, 34, 17, 2, 88, 192, 197, 98, 19, 2, 88, 192, 197, 162, 31, 2, 88, 192, 197, 226, 104, 0, 255, 255, 2, 81, 5, 5, 114, 0, 255, 255, 3, 81, 5, 5, 37, 1, 87, 192, 197, 98, 85, 1, 87, 192, 197, 162, 0, 82, 5, 5, 48, 0, 27, 3, 255, 255, 2, 83, 5, 5, 156, 42, 196, 162, 1, 0, 2, 3, 96, 192, 197, 66, 6, 3, 96, 192, 197, 130, 5, 42, 196, 162, 1, 0, 2, 3, 96, 192, 197, 194, 6, 3, 97, 192, 197, 2, 156, 50, 196, 162, 48, 0, 2, 3, 97, 192, 197, 194, 5, 50, 196, 162, 48, 0, 2, 3, 98, 192, 197, 2, 156, 82, 196, 162, 48, 0, 27, 3, 101, 192, 197, 66, 5, 82, 196, 162, 48, 0, 27, 3, 101, 192, 197, 130, 3, 71, 134, 5, 48, 0, 2, 3, 102, 192, 197, 66, 3, 71, 5, 5, 48, 0, 2, 3, 102, 192, 197, 130, 156, 70, 196, 162, 1, 0, 2, 3, 103, 192, 197, 130, 27, 3, 103, 192, 197, 194, 5, 70, 196, 162, 1, 0, 2, 3, 104, 192, 197, 2, 27, 3, 104, 192, 197, 66, 156, 50, 196, 160, 48, 0, 2, 3, 107, 192, 197, 66, 5, 50, 196, 160, 48, 0, 2, 3, 107, 192, 197, 130, 156, 42, 196, 200, 1, 0, 2, 3, 114, 192, 197, 66, 6, 3, 114, 192, 197, 130, 5, 42, 196, 200, 1, 0, 2, 3, 114, 192, 197, 194, 6, 3, 115, 192, 197, 2, 156, 50, 196, 204, 48, 0, 2, 3, 117, 192, 197, 194, 5, 50, 196, 204, 48, 0, 2, 3, 118, 192, 197, 2, 156, 50, 196, 208, 48, 0, 2, 3, 118, 192, 197, 66, 5, 50, 196, 208, 48, 0, 2, 3, 118, 192, 197, 130, 133, 0, 197, 130, 1, 0, 2, 3, 133, 192, 197, 194, 6, 3, 134, 192, 197, 2, 134, 0, 197, 66, 1, 0, 2, 3, 134, 192, 197, 130, 6, 3, 134, 192, 197, 194, 139, 0, 197, 130, 48, 0, 2, 3, 139, 192, 197, 194, 140, 0, 197, 2, 48, 0, 2, 3, 140, 192, 197, 66, 144, 0, 197, 130, 48, 0, 2, 3, 144, 192, 197, 194, 145, 0, 197, 2, 48, 0, 2, 3, 145, 192, 197, 66, 150, 0, 197, 2, 48, 0, 2, 3, 150, 192, 197, 67, 150, 0, 197, 162, 48, 0, 2, 3, 150, 192, 197, 227, 56, 0, 71, 0, 74, 0, 82, 0, 83, 0, 85, 0, 86, 0, 103, 0, 106, 0, 114, 0, 115, 0, 117, 0, 118, 0, 204, 0, 208, 0, 217, 0, 221, 0, 236, 0, 240, 0, 249, 0, 253, 0, 28, 1, 38, 1, 40, 1, 49, 1, 84, 1, 90, 1, 104, 1, 116, 1, 175, 1, 177, 1, 207, 1, 209, 1, 211, 1, 221, 1, 230, 1, 232, 1, 244, 1, 246, 1, 8, 2, 12, 2, 16, 2, 24, 2, 30, 2, 32, 2, 32, 30, 48, 30, 88, 30, 96, 30, 114, 30, 124, 30, 150, 30, 151, 30, 200, 30, 204, 30, 228, 30, 242, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 68, 28, 4, 185, 80, 38, 4, 185, 96, 40, 4, 185, 112, 186, 116, 186, 124, 86, 4, 185, 136, 185, 140, 96, 4, 124, 4, 126, 4, 136, 4, 185, 188, 186, 192, 180, 4, 206, 4, 185, 216, 185, 220, 185, 224, 185, 228, 185, 232, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 208, 4, 168, 68, 210, 4, 168, 80, 216, 4, 168, 96, 218, 4, 168, 112, 168, 116, 168, 124, 241, 4, 168, 136, 168, 140, 247, 4, 6, 5, 8, 5, 14, 5, 168, 188, 168, 192, 37, 5, 51, 5, 168, 216, 168, 220, 168, 224, 168, 228, 168, 232, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 160, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 153, 57, 217, 57, 185, 64, 185, 57, 121, 58, 89, 58, 0, 8, 249, 74, 154, 89, 218, 89, 185, 92, 122, 90, 154, 117, 218, 117, 26, 118, 122, 118, 217, 80, 186, 145, 153, 161, 217, 161, 185, 164, 185, 161, 121, 162, 136, 14, 217, 162, 153, 213, 217, 213, 25, 214, 121, 214, 217, 229, 185, 236, 2, 8, 136, 57, 200, 57, 168, 64, 168, 57, 104, 58, 72, 58, 4, 8, 232, 74, 136, 89, 200, 89, 168, 92, 104, 90, 136, 117, 200, 117, 8, 118, 104, 118, 200, 80, 168, 145, 136, 161, 200, 161, 168, 164, 168, 161, 104, 162, 128, 14, 200, 162, 136, 213, 200, 213, 8, 214, 104, 214, 200, 229, 168, 236, 104, 230, 57, 59, 40, 59, 185, 60, 168, 60, 53, 5, 55, 5, 217, 73, 200, 73, 25, 74, 8, 74, 185, 74, 168, 74, 57, 74, 40, 74, 57, 82, 40, 82, 185, 84, 168, 84, 58, 91, 40, 91, 250, 89, 232, 89, 186, 90, 168, 90, 57, 5, 59, 5, 58, 90, 40, 90, 25, 102, 8, 102, 249, 101, 232, 101, 185, 102, 168, 102, 249, 102, 232, 102, 25, 114, 8, 114, 89, 115, 72, 115, 186, 117, 168, 117, 58, 119, 40, 119, 250, 117, 232, 117, 26, 119, 8, 119, 186, 118, 168, 120, 6, 8, 8, 8, 26, 126, 8, 126, 249, 130, 232, 130, 168, 184, 217, 137, 200, 137, 249, 138, 232, 138, 57, 138, 40, 138, 121, 139, 104, 139, 89, 139, 72, 139, 218, 145, 200, 145, 250, 146, 232, 146, 58, 146, 40, 146, 10, 8, 185, 156, 168, 156, 57, 163, 40, 163, 249, 161, 232, 161, 153, 162, 136, 162, 12, 8, 14, 8, 217, 189, 200, 189, 249, 190, 232, 190, 57, 190, 40, 190, 218, 193, 200, 193, 26, 194, 8, 194, 250, 194, 232, 194, 58, 194, 40, 194, 249, 198, 232, 198, 57, 198, 40, 198, 185, 208, 168, 208, 185, 213, 168, 213, 57, 215, 40, 215, 249, 213, 232, 213, 89, 214, 72, 214, 153, 214, 136, 214, 61, 5, 63, 5, 25, 222, 8, 222, 25, 230, 8, 230, 121, 230, 217, 233, 200, 233, 185, 234, 168, 234, 57, 234, 40, 234, 232, 192, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 88, 200, 192, 169, 192, 200, 56, 169, 88, 187, 116, 187, 124, 169, 116, 169, 124, 169, 240, 169, 144, 217, 160, 187, 88, 200, 160, 169, 88, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 185, 72, 72, 4, 186, 76, 104, 4, 177, 76, 36, 5, 26, 78, 37, 5, 17, 78, 255, 5, 186, 88, 255, 5, 185, 100, 72, 4, 186, 104, 73, 2, 104, 4, 177, 104, 105, 2, 204, 4, 154, 109, 205, 4, 218, 109, 206, 4, 26, 110, 207, 4, 122, 110, 236, 4, 145, 109, 237, 4, 209, 109, 238, 4, 17, 110, 239, 4, 113, 110, 36, 5, 26, 106, 37, 5, 17, 106, 40, 5, 186, 109, 41, 5, 177, 109, 42, 5, 58, 111, 43, 5, 49, 111, 44, 5, 250, 109, 45, 5, 241, 109, 46, 5, 26, 111, 47, 5, 17, 111, 48, 5, 186, 110, 255, 5, 185, 128, 72, 4, 186, 132, 104, 4, 177, 132, 36, 5, 26, 134, 37, 5, 17, 134, 255, 5, 186, 144, 71, 2, 72, 4, 186, 148, 103, 2, 104, 4, 177, 148, 28, 5, 26, 154, 29, 5, 17, 154, 30, 5, 250, 153, 31, 5, 241, 153, 32, 5, 186, 154, 33, 5, 177, 154, 34, 5, 250, 154, 35, 5, 241, 154, 36, 5, 26, 150, 37, 5, 17, 150, 255, 5, 185, 160, 255, 5, 185, 168, 72, 4, 186, 172, 104, 4, 177, 172, 36, 5, 26, 174, 37, 5, 17, 174, 255, 5, 185, 176, 85, 2, 117, 2, 217, 4, 154, 181, 218, 4, 218, 181, 219, 4, 26, 182, 220, 4, 122, 182, 249, 4, 145, 181, 250, 4, 209, 181, 251, 4, 17, 182, 252, 4, 113, 182, 104, 5, 186, 181, 105, 5, 177, 181, 106, 5, 58, 183, 107, 5, 49, 183, 108, 5, 250, 181, 109, 5, 241, 181, 110, 5, 90, 182, 111, 5, 81, 182, 112, 5, 154, 182, 113, 5, 145, 182, 114, 5, 26, 183, 115, 5, 17, 183, 255, 5, 185, 196, 72, 4, 186, 200, 82, 4, 186, 204, 104, 4, 177, 200, 114, 4, 177, 204, 36, 5, 26, 202, 37, 5, 17, 202, 84, 5, 218, 205, 85, 5, 209, 205, 86, 5, 250, 206, 87, 5, 241, 206, 88, 5, 58, 206, 89, 5, 49, 206, 255, 5, 185, 212, 255, 5, 168, 56, 255, 5, 168, 72, 104, 4, 168, 76, 37, 5, 8, 78, 255, 5, 168, 88, 255, 5, 168, 100, 104, 4, 168, 104, 105, 2, 236, 4, 136, 109, 237, 4, 200, 109, 238, 4, 8, 110, 239, 4, 104, 110, 37, 5, 8, 106, 41, 5, 168, 109, 43, 5, 40, 111, 45, 5, 232, 109, 47, 5, 8, 111, 255, 5, 168, 128, 104, 4, 168, 132, 37, 5, 8, 134, 255, 5, 168, 144, 103, 2, 104, 4, 168, 148, 29, 5, 8, 154, 31, 5, 232, 153, 33, 5, 168, 154, 35, 5, 232, 154, 37, 5, 8, 150, 255, 5, 168, 160, 255, 5, 168, 168, 104, 4, 168, 172, 37, 5, 8, 174, 255, 5, 168, 176, 117, 2, 249, 4, 136, 181, 250, 4, 200, 181, 251, 4, 8, 182, 252, 4, 104, 182, 105, 5, 168, 181, 107, 5, 40, 183, 109, 5, 232, 181, 111, 5, 72, 182, 113, 5, 136, 182, 115, 5, 8, 183, 255, 5, 168, 196, 104, 4, 168, 200, 114, 4, 168, 204, 37, 5, 8, 202, 85, 5, 200, 205, 87, 5, 232, 206, 89, 5, 40, 206, 255, 5, 168, 212, 255, 5, 25, 59, 255, 5, 8, 59, 255, 5, 26, 91, 255, 5, 8, 91, 255, 5, 25, 215, 255, 5, 8, 215, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 75, 15, 0, 16, 92, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 84, 0, 170, 170, 67, 15, 0, 32, 39, 36, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 44, 36, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 202, 9, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 206, 9, 0, 0, 206, 9, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 56, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 62, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 38, 0, 65, 0, 60, 0, 224, 0, 60, 0, 60, 0, 60, 0, 192, 0, 38, 0, 69, 0, 60, 0, 233, 0, 60, 0, 60, 0, 60, 0, 201, 0, 60, 0, 235, 0, 60, 0, 60, 0, 60, 0, 203, 0, 38, 0, 78, 0, 60, 0, 241, 0, 60, 0, 60, 0, 60, 0, 209, 0, 60, 0, 75, 1, 60, 0, 60, 0, 60, 0, 74, 1, 38, 0, 79, 0, 60, 0, 243, 0, 60, 0, 60, 0, 60, 0, 211, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 28, 38, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 116, 29, 0, 0, 116, 29, 0, 0, 116, 29, 0, 0, 116, 29, 0, 0, 40, 31, 0, 0, 40, 31, 0, 0, 12, 34, 0, 0, 12, 34, 0, 0, 4, 38, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 202, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 95, 0, 103, 0, 103, 0, 111, 0, 112, 0, 120, 0, 123, 0, 131, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 155, 0, 48, 0, 162, 0, 48, 0, 48, 0, 170, 0, 177, 0, 185, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 193, 0, 147, 0, 48, 0, 48, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 139, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 124, 1, 156, 1, 192, 1, 236, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 96, 12, 0, 202, 129, 12, 0, 202, 162, 12, 0, 202, 195, 12, 0, 202, 228, 12, 0, 202, 5, 13, 0, 202, 38, 13, 0, 202, 71, 13, 0, 202, 104, 13, 0, 202, 137, 13, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 38, 1, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 201, 166, 2, 0, 201, 134, 3, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 230, 4, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 6, 6, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 201, 134, 7, 0, 201, 102, 8, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 43, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 162, 11, 0, 196, 160, 156, 46, 196, 138, 156, 50, 134, 5, 2, 51, 196, 142, 156, 50, 134, 5, 3, 51, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 134, 5, 2, 69, 196, 138, 156, 70, 134, 5, 2, 71, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 226, 11, 0, 5, 5, 2, 43, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 34, 12, 0, 196, 160, 5, 46, 196, 138, 5, 50, 5, 5, 2, 51, 196, 142, 5, 50, 5, 5, 3, 51, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 5, 5, 2, 69, 196, 138, 5, 70, 5, 5, 2, 71, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 9, 0, 201, 166, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 11, 0, 201, 70, 13, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 14, 0, 201, 134, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 3, 69, 5, 5, 3, 69, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 16, 0, 201, 102, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 18, 0, 201, 166, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 20, 0, 201, 70, 22, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 23, 0, 201, 230, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 26, 0, 201, 134, 27, 0, 201, 6, 29, 0, 201, 134, 30, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 32, 0, 201, 198, 32, 0, 201, 134, 33, 0, 201, 70, 34, 0, 201, 6, 35, 0, 201, 198, 35, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 134, 36, 0, 201, 166, 37, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 38, 0, 201, 70, 40, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 198, 41, 0, 201, 230, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 2, 0, 197, 98, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 44, 0, 201, 38, 45, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 134, 5, 2, 43, 5, 162, 0, 0, 5, 5, 2, 43, 5, 162, 0, 0, 134, 5, 2, 51, 5, 162, 0, 0, 134, 5, 3, 51, 5, 162, 0, 0, 5, 5, 2, 51, 5, 162, 0, 0, 5, 5, 3, 51, 5, 162, 0, 0, 134, 5, 2, 69, 5, 160, 0, 0, 5, 5, 2, 69, 5, 160, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 2, 51, 5, 160, 0, 0, 134, 5, 3, 51, 5, 160, 0, 0, 5, 5, 2, 51, 5, 160, 0, 0, 5, 5, 3, 51, 5, 160, 0, 0, 134, 5, 2, 43, 5, 200, 0, 0, 5, 5, 2, 43, 5, 200, 0, 0, 134, 5, 2, 51, 5, 204, 0, 0, 134, 5, 3, 51, 5, 204, 0, 0, 5, 5, 2, 51, 5, 204, 0, 0, 5, 5, 3, 51, 5, 204, 0, 0, 134, 5, 2, 51, 5, 208, 0, 0, 134, 5, 3, 51, 5, 208, 0, 0, 5, 5, 2, 51, 5, 208, 0, 0, 5, 5, 3, 51, 5, 208, 0, 0, 134, 5, 2, 69, 5, 196, 0, 0, 5, 5, 2, 69, 5, 196, 0, 0, 134, 5, 2, 69, 5, 210, 0, 0, 5, 5, 2, 69, 5, 210, 0, 0, 134, 5, 2, 69, 5, 204, 0, 0, 5, 5, 2, 69, 5, 204, 0, 0, 134, 5, 2, 43, 5, 196, 0, 0, 5, 5, 2, 43, 5, 196, 0, 0, 134, 5, 2, 51, 5, 196, 0, 0, 134, 5, 3, 51, 5, 196, 0, 0, 5, 5, 2, 51, 5, 196, 0, 0, 5, 5, 3, 51, 5, 196, 0, 0, 134, 5, 2, 71, 5, 196, 0, 0, 5, 5, 2, 71, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 1, 0, 0, 3, 2, 235, 134, 5, 64, 3, 2, 235, 134, 5, 0, 50, 156, 5, 2, 0, 1, 3, 2, 243, 134, 5, 8, 3, 3, 243, 134, 5, 65, 3, 2, 243, 134, 5, 0, 68, 156, 5, 48, 0, 3, 3, 255, 255, 2, 69, 134, 5, 0, 70, 156, 5, 1, 0, 1, 3, 255, 255, 2, 71, 134, 5, 65, 3, 255, 255, 2, 71, 134, 5, 0, 42, 5, 5, 1, 0, 0, 3, 2, 235, 5, 5, 64, 3, 2, 235, 5, 5, 0, 50, 5, 5, 2, 0, 1, 3, 2, 243, 5, 5, 8, 3, 3, 243, 5, 5, 65, 3, 2, 243, 5, 5, 0, 68, 5, 5, 48, 0, 3, 3, 255, 255, 2, 69, 5, 5, 0, 70, 5, 5, 1, 0, 1, 3, 255, 255, 2, 71, 5, 5, 65, 3, 255, 255, 2, 71, 5, 5, 156, 42, 196, 162, 1, 0, 0, 3, 197, 162, 64, 3, 197, 162, 5, 42, 196, 162, 1, 0, 0, 3, 0, 192, 197, 98, 64, 3, 0, 192, 197, 98, 156, 50, 196, 162, 2, 0, 1, 3, 0, 192, 197, 162, 8, 3, 0, 192, 197, 226, 65, 3, 0, 192, 197, 162, 5, 50, 196, 162, 2, 0, 1, 3, 1, 192, 197, 34, 8, 3, 1, 192, 197, 98, 65, 3, 1, 192, 197, 34, 156, 68, 196, 160, 48, 0, 3, 3, 1, 192, 197, 162, 5, 68, 196, 160, 48, 0, 3, 3, 1, 192, 197, 226, 156, 70, 196, 190, 1, 0, 1, 3, 2, 192, 197, 34, 65, 3, 2, 192, 197, 34, 5, 70, 196, 190, 1, 0, 1, 3, 2, 192, 197, 98, 65, 3, 2, 192, 197, 98, 156, 70, 196, 162, 1, 0, 1, 3, 2, 192, 197, 162, 65, 3, 2, 192, 197, 162, 5, 70, 196, 162, 1, 0, 1, 3, 2, 192, 197, 226, 65, 3, 2, 192, 197, 226, 156, 50, 196, 160, 2, 0, 1, 3, 3, 192, 197, 34, 8, 3, 3, 192, 197, 98, 65, 3, 3, 192, 197, 34, 5, 50, 196, 160, 2, 0, 1, 3, 3, 192, 197, 162, 8, 3, 3, 192, 197, 226, 65, 3, 3, 192, 197, 162, 156, 42, 196, 200, 1, 0, 0, 3, 4, 192, 197, 34, 64, 3, 4, 192, 197, 34, 5, 42, 196, 200, 1, 0, 0, 3, 4, 192, 197, 98, 64, 3, 4, 192, 197, 98, 156, 50, 196, 204, 2, 0, 1, 3, 4, 192, 197, 162, 8, 3, 4, 192, 197, 226, 65, 3, 4, 192, 197, 162, 5, 50, 196, 204, 2, 0, 1, 3, 5, 192, 197, 34, 8, 3, 5, 192, 197, 98, 65, 3, 5, 192, 197, 34, 156, 50, 196, 208, 2, 0, 1, 3, 5, 192, 197, 162, 8, 3, 5, 192, 197, 226, 65, 3, 5, 192, 197, 162, 5, 50, 196, 208, 2, 0, 1, 3, 6, 192, 197, 34, 8, 3, 6, 192, 197, 98, 65, 3, 6, 192, 197, 34, 156, 68, 196, 196, 48, 0, 3, 3, 6, 192, 197, 162, 5, 68, 196, 196, 48, 0, 3, 3, 6, 192, 197, 226, 156, 68, 196, 210, 48, 0, 3, 3, 7, 192, 197, 34, 5, 68, 196, 210, 48, 0, 3, 3, 7, 192, 197, 98, 156, 68, 196, 204, 48, 0, 3, 3, 7, 192, 197, 162, 5, 68, 196, 204, 48, 0, 3, 3, 7, 192, 197, 226, 156, 42, 196, 196, 1, 0, 0, 3, 8, 192, 197, 34, 64, 3, 8, 192, 197, 34, 5, 42, 196, 196, 1, 0, 0, 3, 8, 192, 197, 98, 64, 3, 8, 192, 197, 98, 156, 50, 196, 196, 2, 0, 1, 3, 8, 192, 197, 162, 8, 3, 8, 192, 197, 226, 65, 3, 8, 192, 197, 162, 5, 50, 196, 196, 2, 0, 1, 3, 9, 192, 197, 34, 8, 3, 9, 192, 197, 98, 65, 3, 9, 192, 197, 34, 156, 70, 196, 196, 1, 0, 1, 3, 9, 192, 197, 162, 65, 3, 9, 192, 197, 162, 5, 70, 196, 196, 1, 0, 1, 3, 9, 192, 197, 226, 65, 3, 9, 192, 197, 226, 10, 0, 197, 35, 1, 0, 1, 3, 10, 192, 197, 131, 65, 3, 10, 192, 197, 131, 10, 0, 197, 227, 1, 0, 1, 3, 11, 192, 197, 67, 65, 3, 11, 192, 197, 67, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 64, 185, 68, 185, 72, 28, 4, 185, 88, 185, 92, 185, 96, 186, 100, 186, 108, 185, 112, 185, 116, 185, 120, 30, 4, 32, 4, 185, 144, 185, 148, 185, 156, 186, 160, 185, 164, 185, 172, 185, 176, 185, 180, 185, 184, 185, 188, 185, 192, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 34, 4, 168, 64, 168, 68, 168, 72, 36, 4, 168, 88, 168, 92, 168, 96, 168, 100, 168, 108, 168, 112, 168, 116, 168, 120, 38, 4, 40, 4, 168, 144, 168, 148, 168, 156, 168, 160, 168, 164, 168, 172, 168, 176, 168, 180, 168, 184, 168, 188, 168, 192, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 136, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 60, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 70, 186, 77, 185, 80, 250, 77, 185, 84, 186, 101, 154, 101, 250, 101, 90, 102, 217, 72, 185, 128, 185, 137, 185, 140, 249, 137, 153, 138, 89, 138, 136, 14, 217, 138, 185, 173, 153, 173, 249, 173, 89, 174, 153, 189, 185, 196, 2, 8, 168, 60, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 70, 168, 77, 168, 80, 232, 77, 168, 84, 168, 101, 136, 101, 232, 101, 72, 102, 200, 72, 168, 128, 168, 137, 168, 140, 232, 137, 136, 138, 72, 138, 128, 14, 200, 138, 168, 173, 136, 173, 232, 173, 72, 174, 136, 189, 168, 196, 72, 190, 57, 59, 40, 59, 217, 57, 200, 57, 42, 4, 44, 4, 153, 69, 136, 69, 249, 69, 232, 69, 185, 70, 168, 70, 25, 70, 8, 70, 25, 74, 8, 74, 89, 75, 72, 75, 58, 79, 40, 79, 218, 77, 200, 77, 186, 78, 168, 78, 46, 4, 48, 4, 26, 78, 8, 78, 249, 93, 232, 93, 217, 93, 200, 93, 185, 94, 168, 94, 249, 94, 232, 94, 249, 97, 232, 97, 89, 99, 72, 99, 154, 102, 136, 102, 58, 103, 40, 103, 218, 101, 200, 101, 26, 103, 8, 103, 186, 102, 168, 104, 6, 8, 8, 8, 250, 109, 232, 109, 249, 114, 232, 114, 168, 152, 153, 117, 136, 117, 249, 118, 232, 118, 25, 118, 8, 118, 121, 119, 104, 119, 89, 119, 72, 119, 154, 125, 136, 125, 50, 4, 52, 4, 26, 126, 8, 126, 10, 8, 185, 132, 168, 132, 57, 139, 40, 139, 217, 137, 200, 137, 121, 138, 104, 138, 12, 8, 14, 8, 153, 157, 136, 157, 249, 158, 232, 158, 25, 158, 8, 158, 154, 161, 136, 161, 250, 161, 232, 161, 250, 162, 232, 162, 26, 162, 8, 162, 249, 166, 232, 166, 25, 166, 8, 166, 185, 168, 168, 168, 153, 174, 136, 174, 57, 175, 40, 175, 217, 173, 200, 173, 57, 174, 40, 174, 121, 174, 104, 174, 25, 175, 8, 175, 249, 181, 232, 181, 249, 189, 232, 189, 89, 190, 153, 193, 136, 193, 185, 194, 168, 194, 25, 194, 8, 194, 232, 160, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 76, 200, 160, 169, 160, 200, 56, 169, 76, 187, 100, 187, 108, 169, 100, 169, 108, 169, 200, 169, 124, 217, 136, 187, 76, 200, 136, 169, 76, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 76, 255, 5, 186, 124, 255, 5, 185, 136, 255, 5, 168, 56, 255, 5, 168, 76, 255, 5, 168, 124, 255, 5, 168, 136, 255, 5, 25, 59, 255, 5, 8, 59, 255, 5, 26, 79, 255, 5, 8, 79, 255, 5, 250, 126, 255, 5, 232, 126, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 59, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 195, 9, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 200, 9, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 76, 16, 0, 32, 7, 0, 0, 0, 23, 0, 0, 0, 80, 16, 0, 0, 80, 16, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 125, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 101, 97, 114, 99, 104, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 115, 0, 101, 0, 97, 0, 114, 0, 99, 0, 104, 0, 93, 0, 14, 32, 38, 0, 243, 5, 60, 0, 60, 0, 39, 0, 39, 0, 14, 32, 38, 0, 244, 5, 60, 0, 60, 0, 39, 0, 34, 0, 39, 0, 0, 0, 155, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 72, 0, 101, 0, 98, 0, 114, 0, 93, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 39, 0, 39, 0, 60, 0, 60, 0, 243, 5, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 39, 0, 34, 0, 39, 0, 60, 0, 60, 0, 244, 5, 38, 0, 209, 5, 60, 0, 209, 5, 191, 5, 38, 0, 213, 5, 60, 0, 60, 0, 213, 5, 188, 5, 60, 0, 60, 0, 213, 5, 213, 5, 60, 0, 60, 0, 213, 5, 217, 5, 38, 0, 217, 5, 60, 0, 60, 0, 217, 5, 180, 5, 60, 0, 60, 0, 217, 5, 217, 5, 60, 0, 60, 0, 242, 5, 183, 5, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 219, 5, 60, 0, 219, 5, 188, 5, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 228, 5, 60, 0, 228, 5, 188, 5, 38, 0, 228, 5, 191, 5, 60, 0, 60, 0, 60, 0, 227, 5, 38, 0, 233, 5, 60, 0, 233, 5, 194, 5, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 49, 0, 93, 0, 234, 5, 60, 0, 234, 5, 188, 5, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 106, 34, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 0, 15, 0, 0, 0, 0, 0, 0, 0, 0, 61, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 228, 31, 0, 0, 228, 31, 0, 0, 228, 31, 0, 0, 228, 31, 0, 0, 12, 34, 0, 0, 12, 34, 0, 0, 82, 34, 0, 0, 82, 34, 0, 0, 82, 34, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 241, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 94, 0, 102, 0, 110, 0, 118, 0, 48, 0, 56, 0, 117, 0, 125, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 142, 0, 150, 0, 48, 0, 48, 0, 48, 0, 158, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 0, 48, 0, 48, 0, 174, 0, 48, 0, 180, 0, 188, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 191, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 199, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 202, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 140, 0, 48, 0, 48, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 232, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 210, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 218, 0, 48, 0, 48, 0, 224, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 120, 1, 184, 1, 192, 0, 212, 1, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 253, 0, 0, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 194, 5, 221, 251, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 0, 16, 0, 202, 33, 16, 0, 202, 66, 16, 0, 202, 99, 16, 0, 202, 132, 16, 0, 202, 165, 16, 0, 202, 198, 16, 0, 202, 231, 16, 0, 202, 8, 17, 0, 202, 41, 17, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 201, 6, 0, 0, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 6, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 7, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 7, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 220, 251, 5, 252, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 26, 102, 5, 71, 26, 102, 5, 70, 222, 102, 5, 72, 26, 102, 5, 70, 238, 102, 201, 198, 0, 0, 192, 0, 0, 0, 5, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 2, 0, 5, 71, 238, 102, 201, 134, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 222, 102, 6, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 204, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 114, 5, 5, 124, 114, 5, 5, 126, 114, 5, 5, 128, 114, 5, 5, 130, 114, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 115, 5, 5, 124, 115, 5, 5, 126, 115, 5, 5, 128, 115, 5, 5, 130, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 0, 0, 192, 0, 0, 0, 197, 226, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 1, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 1, 0, 192, 0, 0, 0, 197, 162, 1, 0, 192, 0, 0, 0, 197, 226, 1, 0, 192, 0, 0, 0, 197, 34, 2, 0, 192, 0, 0, 0, 197, 98, 2, 0, 197, 163, 2, 0, 197, 2, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 3, 0, 197, 131, 3, 0, 197, 226, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 4, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 121, 197, 34, 0, 0, 197, 98, 4, 0, 5, 5, 10, 121, 197, 162, 4, 0, 197, 226, 4, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 142, 9, 118, 5, 5, 11, 118, 193, 2, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 193, 23, 13, 118, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 12, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 193, 126, 133, 115, 193, 133, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 193, 154, 133, 115, 192, 0, 0, 0, 193, 168, 133, 115, 193, 175, 133, 115, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 71, 238, 102, 7, 71, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 26, 102, 6, 70, 26, 102, 7, 71, 26, 102, 6, 71, 26, 102, 7, 70, 222, 102, 6, 70, 222, 102, 7, 72, 26, 102, 6, 72, 26, 102, 9, 70, 238, 102, 8, 70, 238, 102, 6, 70, 238, 102, 7, 70, 238, 102, 7, 5, 26, 102, 6, 5, 26, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 70, 214, 102, 6, 70, 214, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 214, 102, 8, 5, 214, 102, 6, 5, 214, 102, 7, 5, 214, 102, 8, 5, 222, 102, 7, 5, 222, 102, 9, 71, 238, 102, 8, 71, 238, 102, 10, 5, 238, 102, 9, 5, 238, 102, 7, 5, 238, 102, 8, 5, 238, 102, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 5, 6, 121, 5, 5, 6, 121, 5, 5, 12, 121, 5, 5, 12, 121, 5, 5, 20, 121, 5, 5, 20, 121, 5, 5, 24, 121, 5, 5, 24, 121, 5, 5, 30, 121, 5, 5, 30, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 107, 121, 5, 5, 143, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 115, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 119, 121, 5, 5, 103, 121, 5, 5, 143, 121, 5, 5, 119, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 129, 121, 5, 5, 111, 121, 5, 5, 143, 121, 5, 5, 129, 121, 5, 5, 143, 121, 5, 5, 139, 121, 5, 5, 143, 121, 5, 5, 6, 121, 5, 5, 24, 121, 5, 5, 10, 121, 5, 5, 30, 121, 5, 5, 10, 121, 5, 5, 42, 121, 5, 5, 16, 121, 5, 5, 6, 121, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 16, 121, 5, 5, 20, 121, 5, 5, 16, 121, 5, 5, 24, 121, 5, 5, 16, 121, 5, 5, 38, 121, 5, 5, 16, 121, 5, 5, 40, 121, 5, 5, 16, 121, 5, 5, 42, 121, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 6, 121, 197, 34, 0, 0, 5, 5, 10, 121, 5, 5, 12, 121, 197, 98, 0, 0, 5, 5, 16, 121, 5, 5, 18, 121, 5, 5, 20, 121, 197, 162, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 197, 34, 1, 0, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 103, 121, 197, 98, 1, 0, 5, 5, 107, 121, 197, 162, 1, 0, 5, 5, 111, 121, 197, 226, 1, 0, 5, 5, 115, 121, 197, 34, 2, 0, 5, 5, 119, 121, 197, 98, 2, 0, 197, 163, 2, 0, 197, 2, 3, 0, 5, 5, 127, 121, 5, 5, 129, 121, 197, 66, 3, 0, 197, 131, 3, 0, 197, 226, 3, 0, 5, 5, 137, 121, 5, 5, 139, 121, 197, 34, 4, 0, 5, 5, 143, 121, 5, 5, 6, 121, 197, 34, 0, 0, 197, 98, 4, 0, 5, 5, 10, 121, 197, 162, 4, 0, 197, 226, 4, 0, 5, 5, 12, 121, 5, 5, 16, 121, 197, 34, 5, 0, 197, 98, 5, 0, 197, 162, 5, 0, 197, 226, 5, 0, 197, 34, 6, 0, 197, 98, 6, 0, 197, 162, 6, 0, 5, 5, 18, 121, 5, 5, 20, 121, 197, 194, 0, 0, 5, 5, 24, 121, 197, 226, 0, 0, 5, 5, 28, 121, 5, 5, 30, 121, 5, 5, 34, 121, 5, 5, 36, 121, 5, 5, 38, 121, 5, 5, 40, 121, 5, 5, 42, 121, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 131, 12, 5, 5, 48, 0, 56, 3, 132, 204, 5, 5, 26, 102, 5, 5, 2, 0, 83, 6, 255, 255, 26, 102, 5, 70, 84, 6, 255, 255, 26, 102, 5, 71, 85, 6, 255, 255, 26, 102, 5, 72, 222, 102, 5, 5, 48, 0, 84, 6, 255, 255, 222, 102, 5, 70, 238, 102, 5, 5, 48, 0, 84, 6, 255, 255, 238, 102, 5, 70, 170, 170, 3, 0, 67, 0, 58, 0, 32, 0, 127, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 124, 28, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 16, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 64, 0, 0, 0, 76, 0, 0, 0, 76, 1, 0, 0, 124, 26, 0, 0, 124, 26, 0, 0, 124, 26, 0, 0, 124, 26, 0, 0, 84, 27, 0, 0, 84, 27, 0, 0, 74, 28, 0, 0, 100, 28, 0, 0, 19, 0, 0, 0, 0, 0, 0, 101, 194, 255, 50, 101, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 0, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 138, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 80, 0, 88, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 129, 0, 104, 0, 48, 0, 48, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 96, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 105, 0, 113, 0, 121, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 128, 5, 0, 202, 161, 5, 0, 202, 194, 5, 0, 202, 227, 5, 0, 202, 4, 6, 0, 202, 37, 6, 0, 202, 70, 6, 0, 202, 103, 6, 0, 202, 136, 6, 0, 202, 169, 6, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 156, 5, 0, 50, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 156, 5, 0, 70, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 5, 5, 0, 50, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 5, 5, 0, 70, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 194, 4, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 2, 5, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 66, 5, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 228, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 100, 4, 0, 192, 0, 0, 0, 201, 134, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 1, 0, 201, 102, 7, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 70, 8, 0, 201, 38, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 6, 10, 0, 5, 2, 106, 9, 5, 2, 120, 9, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 24, 101, 192, 0, 0, 0, 5, 72, 24, 101, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 10, 0, 193, 2, 47, 101, 201, 166, 11, 0, 197, 2, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 70, 16, 101, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 101, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 3, 0, 5, 5, 37, 101, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 13, 0, 193, 3, 47, 101, 201, 166, 14, 0, 5, 5, 9, 101, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 5, 72, 16, 101, 194, 5, 167, 226, 5, 71, 16, 101, 194, 5, 210, 227, 5, 72, 16, 101, 194, 5, 210, 227, 5, 71, 16, 101, 194, 5, 78, 227, 5, 71, 24, 101, 194, 5, 167, 226, 5, 71, 24, 101, 194, 5, 210, 227, 5, 5, 38, 101, 194, 6, 243, 227, 32, 5, 24, 101, 32, 5, 24, 101, 5, 5, 46, 101, 194, 5, 177, 227, 193, 2, 47, 101, 194, 5, 177, 227, 5, 5, 46, 101, 194, 5, 210, 227, 194, 5, 177, 227, 193, 2, 47, 101, 194, 5, 210, 227, 194, 5, 177, 227, 5, 5, 8, 101, 194, 5, 210, 227, 5, 5, 9, 101, 194, 5, 210, 227, 5, 5, 38, 101, 194, 6, 243, 227, 194, 5, 210, 227, 5, 5, 16, 101, 194, 5, 78, 227, 5, 70, 16, 101, 194, 5, 78, 227, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 8, 101, 5, 5, 48, 0, 191, 5, 255, 255, 9, 101, 5, 5, 16, 101, 5, 5, 6, 0, 29, 251, 11, 0, 29, 251, 197, 162, 53, 251, 0, 192, 197, 98, 57, 251, 0, 192, 197, 162, 75, 251, 0, 192, 197, 226, 188, 5, 255, 255, 16, 101, 5, 70, 213, 5, 255, 255, 16, 101, 5, 71, 217, 5, 255, 255, 16, 101, 5, 72, 24, 101, 5, 5, 3, 0, 180, 5, 255, 255, 24, 101, 5, 70, 217, 5, 255, 255, 24, 101, 5, 71, 29, 251, 1, 192, 197, 34, 57, 251, 1, 192, 197, 98, 26, 101, 5, 5, 48, 0, 188, 5, 255, 255, 25, 101, 5, 5, 38, 101, 5, 5, 48, 0, 188, 5, 255, 255, 37, 101, 5, 5, 46, 101, 5, 5, 48, 0, 194, 5, 255, 255, 47, 101, 193, 2, 48, 101, 5, 5, 48, 0, 188, 5, 255, 255, 47, 101, 193, 3, 1, 0, 197, 226, 48, 0, 183, 5, 255, 255, 24, 101, 5, 72, 2, 0, 197, 34, 48, 0, 194, 5, 2, 192, 197, 98, 2, 0, 197, 163, 48, 0, 194, 5, 3, 192, 197, 3, 3, 0, 197, 98, 48, 0, 191, 5, 3, 192, 197, 162, 2, 0, 197, 162, 1, 0, 193, 5, 3, 0, 194, 5, 3, 192, 197, 2, 48, 0, 194, 5, 3, 192, 197, 3, 4, 0, 197, 66, 48, 0, 188, 5, 4, 192, 197, 130, 12, 0, 213, 5, 214, 5, 217, 5, 218, 5, 29, 251, 30, 251, 53, 251, 54, 251, 57, 251, 58, 251, 75, 251, 76, 251, 3, 0, 67, 0, 58, 0, 32, 0, 35, 9, 0, 16, 46, 0, 0, 96, 1, 0, 0, 96, 2, 0, 51, 0, 82, 0, 170, 170, 27, 9, 0, 32, 67, 16, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 72, 16, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 231, 8, 0, 32, 7, 0, 0, 0, 21, 0, 0, 0, 235, 8, 0, 0, 235, 8, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 54, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 57, 220, 91, 0, 110, 0, 111, 0, 114, 0, 109, 0, 97, 0, 108, 0, 105, 0, 122, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 32, 0, 111, 0, 110, 0, 93, 0, 38, 0, 69, 0, 60, 0, 185, 30, 60, 0, 60, 0, 60, 0, 184, 30, 38, 0, 71, 0, 60, 0, 103, 0, 98, 0, 60, 0, 60, 0, 60, 0, 71, 0, 98, 0, 60, 0, 60, 0, 60, 0, 71, 0, 66, 0, 38, 0, 79, 0, 60, 0, 205, 30, 60, 0, 60, 0, 60, 0, 204, 30, 38, 0, 83, 0, 60, 0, 99, 30, 60, 0, 60, 0, 60, 0, 98, 30, 0, 0, 170, 170, 170, 170, 170, 170, 160, 34, 0, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 17, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 68, 0, 0, 0, 68, 0, 0, 0, 84, 28, 0, 0, 84, 28, 0, 0, 84, 28, 0, 0, 84, 28, 0, 0, 96, 29, 0, 0, 96, 29, 0, 0, 126, 30, 0, 0, 140, 30, 0, 0, 136, 34, 0, 0, 50, 105, 114, 84, 1, 0, 64, 8, 184, 0, 255, 255, 192, 0, 28, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 89, 0, 97, 0, 97, 0, 105, 0, 106, 0, 114, 0, 117, 0, 125, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 149, 0, 48, 0, 152, 0, 159, 0, 167, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 175, 0, 141, 0, 48, 0, 48, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 133, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 100, 1, 132, 1, 168, 1, 212, 1, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 32, 7, 0, 202, 65, 7, 0, 202, 98, 7, 0, 202, 131, 7, 0, 202, 164, 7, 0, 202, 197, 7, 0, 202, 230, 7, 0, 202, 7, 8, 0, 202, 40, 8, 0, 202, 73, 8, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 156, 5, 0, 42, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 6, 0, 0, 156, 5, 0, 52, 201, 192, 0, 0, 156, 5, 0, 56, 156, 5, 0, 58, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 156, 5, 0, 66, 156, 5, 0, 68, 201, 70, 4, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 201, 38, 5, 0, 156, 5, 0, 80, 156, 5, 0, 82, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 5, 5, 0, 42, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 6, 6, 0, 5, 5, 0, 52, 201, 192, 6, 0, 5, 5, 0, 56, 5, 5, 0, 58, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 5, 5, 0, 66, 5, 5, 0, 68, 201, 166, 8, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 201, 134, 9, 0, 5, 5, 0, 80, 5, 5, 0, 82, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 196, 138, 156, 42, 196, 136, 156, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 6, 0, 196, 160, 156, 46, 196, 138, 156, 50, 196, 136, 156, 50, 196, 142, 156, 50, 196, 150, 156, 50, 196, 138, 156, 58, 196, 136, 156, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 196, 138, 156, 70, 196, 136, 156, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 196, 138, 156, 82, 196, 136, 156, 82, 196, 142, 156, 82, 196, 150, 156, 82, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 6, 0, 196, 138, 5, 42, 196, 136, 5, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 6, 0, 196, 160, 5, 46, 196, 138, 5, 50, 196, 136, 5, 50, 196, 142, 5, 50, 196, 150, 5, 50, 196, 138, 5, 58, 196, 136, 5, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 196, 138, 5, 70, 196, 136, 5, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 196, 138, 5, 82, 196, 136, 5, 82, 196, 142, 5, 82, 196, 150, 5, 82, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 10, 0, 201, 38, 11, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 11, 0, 201, 166, 12, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 13, 0, 201, 38, 14, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 14, 0, 201, 166, 15, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 16, 0, 201, 38, 17, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 79, 5, 5, 2, 79, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 4, 0, 197, 34, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 51, 5, 5, 2, 51, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 5, 0, 197, 162, 5, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 5, 2, 71, 5, 5, 2, 71, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 5, 0, 197, 34, 6, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 98, 3, 0, 197, 162, 3, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 0, 0, 0, 135, 5, 2, 55, 5, 156, 0, 0, 70, 5, 2, 55, 5, 156, 0, 0, 135, 5, 2, 55, 5, 196, 0, 0, 70, 5, 2, 55, 5, 196, 0, 0, 135, 5, 2, 55, 5, 210, 0, 0, 70, 5, 2, 55, 5, 210, 0, 0, 5, 5, 2, 55, 5, 156, 0, 0, 5, 5, 2, 55, 5, 196, 0, 0, 5, 5, 2, 55, 5, 210, 0, 0, 134, 5, 2, 51, 5, 162, 0, 0, 5, 5, 2, 51, 5, 162, 0, 0, 134, 5, 2, 79, 5, 160, 0, 0, 5, 5, 2, 79, 5, 160, 0, 0, 134, 5, 2, 71, 5, 190, 0, 0, 5, 5, 2, 71, 5, 190, 0, 0, 134, 5, 2, 71, 5, 162, 0, 0, 5, 5, 2, 71, 5, 162, 0, 0, 134, 5, 2, 51, 5, 160, 0, 0, 5, 5, 2, 51, 5, 160, 0, 0, 134, 5, 2, 79, 5, 156, 0, 0, 5, 5, 2, 79, 5, 156, 0, 0, 134, 5, 2, 51, 5, 142, 0, 0, 5, 5, 2, 51, 5, 142, 0, 0, 134, 5, 2, 71, 5, 142, 0, 0, 5, 5, 2, 71, 5, 142, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 50, 156, 5, 48, 0, 35, 3, 2, 243, 134, 5, 0, 54, 156, 5, 7, 0, 4, 30, 12, 0, 4, 30, 0, 192, 197, 162, 5, 30, 0, 192, 197, 226, 6, 30, 1, 192, 197, 34, 7, 30, 1, 192, 197, 98, 66, 0, 2, 247, 135, 5, 98, 0, 2, 247, 70, 5, 2, 30, 197, 162, 3, 30, 0, 192, 197, 98, 0, 70, 156, 5, 48, 0, 35, 3, 255, 255, 2, 71, 134, 5, 0, 78, 156, 5, 48, 0, 35, 3, 255, 255, 2, 79, 134, 5, 0, 50, 5, 5, 48, 0, 35, 3, 2, 243, 5, 5, 0, 54, 5, 5, 3, 0, 98, 0, 2, 247, 5, 5, 3, 30, 1, 192, 197, 162, 5, 30, 1, 192, 197, 226, 7, 30, 2, 192, 197, 34, 0, 70, 5, 5, 48, 0, 35, 3, 255, 255, 2, 71, 5, 5, 0, 78, 5, 5, 48, 0, 35, 3, 255, 255, 2, 79, 5, 5, 156, 50, 196, 162, 48, 0, 35, 3, 2, 192, 197, 98, 5, 50, 196, 162, 48, 0, 35, 3, 2, 192, 197, 162, 156, 78, 196, 160, 48, 0, 35, 3, 2, 192, 197, 226, 5, 78, 196, 160, 48, 0, 35, 3, 3, 192, 197, 34, 156, 70, 196, 190, 48, 0, 35, 3, 3, 192, 197, 98, 5, 70, 196, 190, 48, 0, 35, 3, 3, 192, 197, 162, 156, 70, 196, 162, 48, 0, 35, 3, 3, 192, 197, 226, 5, 70, 196, 162, 48, 0, 35, 3, 4, 192, 197, 34, 156, 50, 196, 160, 48, 0, 35, 3, 4, 192, 197, 98, 5, 50, 196, 160, 48, 0, 35, 3, 4, 192, 197, 162, 6, 0, 66, 0, 67, 0, 98, 0, 99, 0, 2, 30, 8, 30, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 185, 56, 185, 60, 185, 64, 185, 68, 26, 4, 185, 76, 28, 4, 185, 88, 186, 92, 186, 100, 185, 104, 185, 108, 185, 112, 186, 116, 34, 4, 185, 128, 185, 132, 185, 140, 36, 4, 185, 148, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 185, 176, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 168, 56, 168, 60, 168, 64, 168, 68, 38, 4, 168, 76, 40, 4, 168, 88, 168, 92, 168, 100, 168, 104, 168, 108, 168, 112, 168, 116, 44, 4, 168, 128, 168, 132, 168, 140, 46, 4, 168, 148, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 168, 176, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 124, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 185, 57, 153, 57, 249, 57, 153, 58, 89, 58, 57, 58, 0, 8, 249, 66, 186, 73, 154, 73, 250, 73, 90, 74, 186, 93, 154, 93, 250, 93, 90, 94, 217, 68, 154, 118, 185, 125, 153, 125, 249, 125, 153, 126, 89, 126, 136, 14, 217, 126, 185, 157, 153, 157, 249, 157, 89, 158, 153, 173, 185, 180, 2, 8, 168, 57, 136, 57, 232, 57, 136, 58, 72, 58, 40, 58, 4, 8, 232, 66, 168, 73, 136, 73, 232, 73, 72, 74, 168, 93, 136, 93, 232, 93, 72, 94, 200, 68, 136, 118, 168, 125, 136, 125, 232, 125, 136, 126, 72, 126, 128, 14, 200, 126, 168, 157, 136, 157, 232, 157, 72, 158, 136, 173, 168, 180, 72, 174, 57, 59, 40, 59, 217, 57, 200, 57, 25, 59, 8, 59, 153, 65, 136, 65, 249, 65, 232, 65, 185, 66, 168, 66, 25, 66, 8, 66, 25, 70, 8, 70, 89, 71, 72, 71, 58, 75, 40, 75, 218, 73, 200, 73, 186, 74, 168, 74, 48, 4, 50, 4, 26, 74, 8, 74, 249, 81, 232, 81, 217, 81, 200, 81, 185, 82, 168, 82, 249, 82, 232, 82, 249, 89, 232, 89, 89, 91, 72, 91, 154, 94, 136, 94, 58, 95, 40, 95, 218, 93, 200, 93, 26, 95, 8, 95, 186, 94, 168, 96, 6, 8, 8, 8, 250, 101, 232, 101, 249, 106, 232, 106, 168, 136, 153, 109, 136, 109, 249, 110, 232, 110, 25, 110, 8, 110, 121, 111, 104, 111, 89, 111, 72, 111, 154, 117, 136, 117, 250, 118, 232, 118, 26, 118, 8, 118, 10, 8, 185, 120, 168, 120, 57, 127, 40, 127, 217, 125, 200, 125, 121, 126, 104, 126, 12, 8, 14, 8, 153, 141, 136, 141, 249, 142, 232, 142, 25, 142, 8, 142, 154, 145, 136, 145, 250, 145, 232, 145, 52, 4, 54, 4, 26, 146, 8, 146, 249, 150, 232, 150, 25, 150, 8, 150, 185, 152, 168, 152, 153, 158, 136, 158, 57, 159, 40, 159, 217, 157, 200, 157, 57, 158, 40, 158, 121, 158, 104, 158, 25, 159, 8, 159, 249, 165, 232, 165, 249, 173, 232, 173, 89, 174, 153, 177, 136, 177, 185, 178, 168, 178, 25, 178, 8, 178, 232, 144, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 144, 169, 144, 200, 56, 169, 72, 187, 92, 187, 100, 169, 92, 169, 100, 169, 184, 169, 116, 217, 124, 187, 72, 200, 124, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 186, 72, 255, 5, 185, 80, 66, 4, 186, 84, 98, 4, 177, 84, 255, 5, 185, 124, 255, 5, 186, 144, 255, 5, 168, 72, 255, 5, 168, 80, 98, 4, 168, 84, 255, 5, 168, 124, 255, 5, 168, 144, 255, 5, 26, 75, 255, 5, 8, 75, 255, 5, 250, 146, 255, 5, 232, 146, 255, 1, 3, 0, 69, 0, 60, 0, 32, 0, 55, 0, 0, 16, 6, 0, 0, 96, 1, 0, 0, 96, 1, 0, 51, 0, 224, 8, 0, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 229, 8, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 16, 0, 0, 0, 16, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 16, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 0, 0, 1, 0, 32, 0, 1, 0, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 16, 0, 0, 0, 16, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 16, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 0, 0, 1, 0, 32, 0, 1, 0, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 16, 0, 0, 0, 16, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 16, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 0, 0, 1, 0, 32, 0, 1, 0, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 203, 128, 3, 32, 7, 0, 0, 0, 35, 0, 0, 0, 207, 128, 3, 0, 207, 128, 3, 0, 8, 0, 0, 0, 0, 0, 0, 0, 26, 206, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 98, 105, 103, 53, 104, 97, 110, 0, 83, 101, 113, 117, 101, 110, 99, 101, 0, 37, 37, 67, 111, 108, 108, 97, 116, 105, 111, 110, 66, 105, 110, 0, 100, 101, 102, 97, 117, 108, 116, 0, 103, 98, 50, 51, 49, 50, 104, 97, 110, 0, 112, 114, 105, 118, 97, 116, 101, 45, 112, 105, 110, 121, 105, 110, 0, 115, 116, 114, 111, 107, 101, 0, 117, 110, 105, 104, 97, 110, 0, 122, 104, 117, 121, 105, 110, 0, 170, 170, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 112, 0, 105, 0, 110, 0, 121, 0, 105, 0, 110, 0, 0, 0, 82, 220, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 117, 0, 110, 0, 100, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 117, 0, 110, 0, 105, 0, 104, 0, 97, 0, 110, 0, 93, 0, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 122, 0, 104, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 112, 0, 105, 0, 110, 0, 121, 0, 105, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 72, 0, 97, 0, 110, 0, 105, 0, 32, 0, 66, 0, 111, 0, 112, 0, 111, 0, 93, 0, 0, 0, 115, 221, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 97, 0, 60, 0, 60, 0, 1, 1, 60, 0, 60, 0, 60, 0, 0, 1, 60, 0, 60, 0, 225, 0, 60, 0, 60, 0, 60, 0, 193, 0, 60, 0, 60, 0, 206, 1, 60, 0, 60, 0, 60, 0, 205, 1, 60, 0, 60, 0, 224, 0, 60, 0, 60, 0, 60, 0, 192, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 101, 0, 60, 0, 60, 0, 19, 1, 60, 0, 60, 0, 60, 0, 18, 1, 60, 0, 60, 0, 233, 0, 60, 0, 60, 0, 60, 0, 201, 0, 60, 0, 60, 0, 27, 1, 60, 0, 60, 0, 60, 0, 26, 1, 60, 0, 60, 0, 232, 0, 60, 0, 60, 0, 60, 0, 200, 0, 38, 0, 101, 0, 60, 0, 60, 0, 101, 0, 2, 3, 4, 3, 60, 0, 60, 0, 60, 0, 69, 0, 2, 3, 4, 3, 60, 0, 60, 0, 101, 0, 2, 3, 1, 3, 60, 0, 60, 0, 60, 0, 69, 0, 2, 3, 1, 3, 60, 0, 60, 0, 101, 0, 2, 3, 12, 3, 60, 0, 60, 0, 60, 0, 69, 0, 2, 3, 12, 3, 60, 0, 60, 0, 101, 0, 2, 3, 0, 3, 60, 0, 60, 0, 60, 0, 69, 0, 2, 3, 0, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 105, 0, 60, 0, 60, 0, 43, 1, 60, 0, 60, 0, 60, 0, 42, 1, 60, 0, 60, 0, 237, 0, 60, 0, 60, 0, 60, 0, 205, 0, 60, 0, 60, 0, 208, 1, 60, 0, 60, 0, 60, 0, 207, 1, 60, 0, 60, 0, 236, 0, 60, 0, 60, 0, 60, 0, 204, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 109, 0, 60, 0, 60, 0, 109, 0, 4, 3, 60, 0, 60, 0, 60, 0, 77, 0, 4, 3, 60, 0, 60, 0, 63, 30, 60, 0, 60, 0, 60, 0, 62, 30, 60, 0, 60, 0, 109, 0, 12, 3, 60, 0, 60, 0, 60, 0, 77, 0, 12, 3, 60, 0, 60, 0, 109, 0, 0, 3, 60, 0, 60, 0, 60, 0, 77, 0, 0, 3, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 110, 0, 60, 0, 60, 0, 110, 0, 4, 3, 60, 0, 60, 0, 60, 0, 78, 0, 4, 3, 60, 0, 60, 0, 68, 1, 60, 0, 60, 0, 60, 0, 67, 1, 60, 0, 60, 0, 72, 1, 60, 0, 60, 0, 60, 0, 71, 1, 60, 0, 60, 0, 249, 1, 60, 0, 60, 0, 60, 0, 248, 1, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 111, 0, 60, 0, 60, 0, 77, 1, 60, 0, 60, 0, 60, 0, 76, 1, 60, 0, 60, 0, 243, 0, 60, 0, 60, 0, 60, 0, 211, 0, 60, 0, 60, 0, 210, 1, 60, 0, 60, 0, 60, 0, 209, 1, 60, 0, 60, 0, 242, 0, 60, 0, 60, 0, 60, 0, 210, 0, 38, 0, 91, 0, 98, 0, 101, 0, 102, 0, 111, 0, 114, 0, 101, 0, 32, 0, 50, 0, 93, 0, 117, 0, 60, 0, 60, 0, 107, 1, 60, 0, 60, 0, 60, 0, 106, 1, 60, 0, 60, 0, 250, 0, 60, 0, 60, 0, 60, 0, 218, 0, 60, 0, 60, 0, 212, 1, 60, 0, 60, 0, 60, 0, 211, 1, 60, 0, 60, 0, 249, 0, 60, 0, 60, 0, 60, 0, 217, 0, 38, 0, 85, 0, 60, 0, 60, 0, 214, 1, 60, 0, 60, 0, 60, 0, 213, 1, 60, 0, 60, 0, 216, 1, 60, 0, 60, 0, 60, 0, 215, 1, 60, 0, 60, 0, 218, 1, 60, 0, 60, 0, 60, 0, 217, 1, 60, 0, 60, 0, 220, 1, 60, 0, 60, 0, 60, 0, 219, 1, 60, 0, 60, 0, 252, 0, 60, 0, 60, 0, 60, 0, 220, 0, 0, 0, 239, 223, 179, 26, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 122, 0, 104, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 112, 0, 105, 0, 110, 0, 121, 0, 105, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 76, 0, 97, 0, 116, 0, 110, 0, 32, 0, 72, 0, 97, 0, 110, 0, 105, 0, 93, 0, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 114, 0, 101, 0, 103, 0, 117, 0, 108, 0, 97, 0, 114, 0, 93, 0, 60, 0, 42, 0, 74, 85, 63, 150, 195, 87, 40, 99, 206, 84, 9, 85, 192, 84, 145, 118, 76, 118, 60, 133, 238, 119, 126, 130, 141, 120, 49, 114, 152, 150, 141, 151, 40, 108, 137, 91, 250, 79, 9, 99, 151, 102, 184, 92, 250, 128, 72, 104, 174, 128, 2, 102, 206, 118, 249, 81, 86, 101, 172, 113, 241, 127, 132, 136, 178, 80, 101, 89, 202, 97, 179, 111, 173, 130, 76, 99, 82, 98, 237, 83, 39, 84, 6, 123, 107, 81, 164, 117, 244, 93, 212, 98, 203, 141, 118, 151, 138, 98, 25, 128, 93, 87, 56, 151, 98, 127, 56, 114, 125, 118, 207, 103, 126, 118, 70, 100, 112, 79, 37, 141, 220, 98, 23, 122, 145, 101, 237, 115, 44, 100, 115, 98, 44, 130, 129, 152, 127, 103, 72, 114, 110, 98, 204, 98, 52, 79, 227, 116, 74, 83, 158, 82, 202, 126, 166, 144, 46, 94, 134, 104, 156, 105, 128, 129, 209, 126, 210, 104, 197, 120, 140, 134, 81, 149, 141, 80, 36, 140, 222, 130, 222, 128, 5, 83, 18, 137, 101, 82, 132, 133, 249, 150, 221, 79, 33, 88, 113, 153, 157, 91, 177, 98, 165, 98, 180, 102, 121, 140, 141, 156, 6, 114, 111, 103, 145, 120, 178, 96, 81, 83, 23, 83, 136, 143, 204, 128, 29, 141, 161, 148, 13, 80, 200, 114, 7, 89, 235, 96, 25, 113, 171, 136, 84, 89, 239, 130, 44, 103, 40, 123, 41, 93, 247, 126, 45, 117, 245, 108, 102, 142, 248, 143, 60, 144, 59, 159, 212, 107, 25, 145, 20, 123, 124, 95, 167, 120, 214, 132, 61, 133, 213, 107, 217, 107, 214, 107, 1, 94, 135, 94, 249, 117, 237, 149, 93, 101, 10, 95, 197, 95, 159, 143, 193, 88, 194, 129, 127, 144, 91, 150, 173, 151, 185, 143, 22, 127, 44, 141, 65, 98, 191, 79, 216, 83, 94, 83, 168, 143, 169, 143, 171, 143, 77, 144, 7, 104, 106, 95, 152, 129, 104, 136, 214, 156, 139, 97, 43, 82, 42, 118, 108, 95, 140, 101, 210, 111, 232, 110, 190, 91, 72, 100, 117, 81, 176, 81, 196, 103, 25, 78, 201, 121, 124, 153, 179, 112, 197, 117, 118, 94, 187, 115, 224, 131, 173, 100, 232, 98, 181, 148, 226, 108, 90, 83, 195, 82, 15, 100, 194, 148, 148, 123, 47, 79, 27, 94, 54, 130, 22, 129, 138, 129, 36, 110, 202, 108, 115, 154, 85, 99, 92, 83, 250, 84, 101, 136, 224, 87, 13, 78, 3, 94, 101, 107, 63, 124, 232, 144, 22, 96, 230, 100, 28, 115, 193, 136, 80, 103, 77, 98, 34, 141, 108, 119, 41, 142, 199, 145, 105, 95, 220, 131, 33, 133, 16, 153, 194, 83, 149, 134, 139, 107, 237, 96, 232, 96, 127, 112, 205, 130, 49, 130, 211, 78, 167, 108, 207, 133, 205, 100, 217, 124, 253, 105, 249, 102, 73, 131, 149, 83, 86, 123, 167, 79, 140, 81, 75, 109, 66, 92, 109, 142, 210, 99, 201, 83, 44, 131, 54, 131, 229, 103, 180, 120, 61, 100, 223, 91, 148, 92, 238, 93, 231, 139, 198, 98, 244, 103, 122, 140, 0, 100, 186, 99, 73, 135, 139, 153, 23, 140, 32, 127, 242, 148, 167, 78, 16, 150, 164, 152, 12, 102, 22, 115, 58, 87, 29, 92, 56, 94, 127, 149, 127, 80, 160, 128, 130, 83, 94, 101, 69, 117, 49, 85, 33, 80, 133, 141, 132, 98, 158, 148, 29, 103, 50, 86, 110, 111, 226, 93, 53, 84, 146, 112, 102, 143, 111, 98, 164, 100, 163, 99, 123, 95, 136, 111, 244, 144, 227, 129, 176, 143, 24, 92, 104, 102, 241, 95, 137, 108, 72, 150, 129, 141, 108, 136, 145, 100, 240, 121, 206, 87, 89, 106, 16, 98, 72, 84, 88, 78, 11, 122, 233, 96, 132, 111, 218, 139, 127, 98, 30, 144, 139, 154, 228, 121, 3, 84, 244, 117, 1, 99, 25, 83, 96, 108, 223, 143, 27, 95, 112, 154, 59, 128, 127, 159, 136, 79, 58, 92, 100, 141, 197, 127, 165, 101, 189, 112, 69, 81, 178, 81, 107, 134, 7, 93, 160, 91, 189, 98, 108, 145, 116, 117, 12, 142, 32, 122, 1, 97, 121, 123, 199, 78, 248, 126, 133, 119, 17, 78, 237, 129, 29, 82, 250, 81, 113, 106, 168, 83, 135, 142, 4, 149, 207, 150, 193, 110, 100, 150, 90, 105, 64, 120, 168, 80, 215, 119, 16, 100, 230, 137, 4, 89, 227, 99, 221, 93, 127, 122, 61, 105, 32, 79, 57, 130, 152, 85, 50, 78, 174, 117, 151, 122, 98, 94, 138, 94, 239, 149, 27, 82, 57, 84, 138, 112, 118, 99, 36, 149, 130, 87, 37, 102, 63, 105, 135, 145, 7, 85, 243, 109, 175, 126, 34, 136, 51, 98, 240, 126, 181, 117, 40, 131, 193, 120, 204, 150, 158, 143, 72, 97, 247, 116, 205, 139, 100, 107, 58, 82, 80, 141, 33, 107, 106, 128, 113, 132, 241, 86, 6, 83, 206, 78, 27, 78, 209, 81, 151, 124, 139, 145, 7, 124, 195, 79, 127, 142, 225, 123, 156, 122, 103, 100, 20, 93, 172, 80, 6, 129, 1, 118, 185, 124, 236, 109, 224, 127, 81, 103, 88, 91, 248, 91, 203, 120, 174, 100, 19, 100, 170, 99, 43, 99, 25, 149, 45, 100, 190, 143, 84, 123, 41, 118, 83, 98, 39, 89, 70, 84, 121, 107, 163, 80, 52, 98, 38, 94, 134, 107, 227, 78, 55, 141, 139, 136, 133, 95, 46, 144, 32, 96, 61, 128, 197, 98, 57, 78, 85, 83, 248, 144, 184, 99, 198, 128, 230, 101, 46, 108, 70, 79, 238, 96, 225, 109, 222, 139, 57, 95, 203, 134, 83, 95, 33, 99, 90, 81, 97, 131, 99, 104, 0, 82, 99, 99, 72, 142, 18, 80, 155, 92, 119, 121, 252, 91, 48, 82, 59, 122, 188, 96, 83, 144, 215, 118, 183, 95, 151, 95, 132, 118, 108, 142, 111, 112, 123, 118, 73, 123, 170, 119, 243, 81, 147, 144, 36, 88, 78, 79, 244, 110, 234, 143, 76, 101, 27, 123, 196, 114, 164, 109, 223, 127, 225, 90, 181, 98, 149, 94, 48, 87, 130, 132, 44, 123, 29, 94, 31, 95, 18, 144, 20, 127, 160, 152, 130, 99, 199, 110, 152, 120, 185, 112, 120, 81, 91, 151, 171, 87, 53, 117, 67, 79, 56, 117, 151, 94, 230, 96, 96, 89, 192, 109, 191, 107, 137, 120, 252, 83, 213, 150, 203, 81, 1, 82, 137, 99, 10, 84, 147, 148, 3, 140, 204, 141, 57, 114, 159, 120, 118, 135, 237, 143, 13, 140, 224, 83, 1, 78, 239, 118, 238, 83, 137, 148, 118, 152, 14, 159, 45, 149, 154, 91, 162, 139, 34, 78, 28, 78, 172, 81, 99, 132, 194, 97, 168, 82, 11, 104, 151, 79, 107, 96, 187, 81, 30, 109, 92, 81, 150, 98, 151, 101, 97, 150, 70, 140, 23, 144, 216, 117, 253, 144, 99, 119, 210, 107, 138, 114, 236, 114, 251, 139, 53, 88, 121, 119, 76, 141, 92, 103, 64, 149, 154, 128, 166, 94, 33, 110, 146, 89, 239, 122, 237, 119, 59, 149, 181, 107, 173, 101, 14, 127, 6, 88, 81, 81, 31, 150, 249, 91, 169, 88, 40, 84, 114, 142, 102, 101, 127, 152, 228, 86, 157, 148, 254, 118, 65, 144, 135, 99, 198, 84, 26, 89, 58, 89, 155, 87, 178, 142, 53, 103, 250, 141, 53, 130, 65, 82, 240, 96, 21, 88, 254, 134, 232, 92, 69, 158, 196, 79, 157, 152, 185, 139, 37, 90, 118, 96, 132, 83, 124, 98, 79, 144, 2, 145, 127, 153, 105, 96, 12, 128, 63, 81, 51, 128, 20, 92, 117, 153, 49, 109, 140, 78, 48, 141, 209, 83, 90, 127, 79, 123, 16, 79, 79, 78, 0, 150, 213, 108, 208, 115, 233, 133, 6, 94, 106, 117, 251, 127, 10, 106, 254, 119, 146, 148, 65, 126, 225, 81, 230, 112, 205, 83, 212, 143, 3, 131, 41, 141, 175, 114, 109, 153, 219, 108, 74, 87, 179, 130, 185, 101, 170, 128, 63, 98, 50, 150, 168, 89, 255, 78, 191, 139, 186, 126, 62, 101, 242, 131, 94, 151, 97, 85, 222, 152, 165, 128, 42, 83, 253, 139, 32, 84, 186, 128, 159, 94, 184, 108, 57, 141, 172, 130, 90, 145, 41, 84, 27, 108, 6, 82, 183, 126, 95, 87, 26, 113, 126, 108, 137, 124, 75, 89, 253, 78, 255, 95, 36, 97, 170, 124, 48, 78, 1, 92, 171, 103, 2, 135, 240, 92, 11, 149, 206, 152, 175, 117, 253, 112, 34, 144, 175, 81, 29, 127, 189, 139, 73, 89, 228, 81, 91, 79, 38, 84, 43, 89, 119, 101, 164, 128, 117, 91, 118, 98, 194, 98, 144, 143, 69, 94, 31, 108, 38, 123, 15, 79, 216, 79, 13, 103, 110, 109, 170, 109, 143, 121, 177, 136, 23, 95, 43, 117, 154, 98, 133, 143, 239, 79, 220, 145, 167, 101, 47, 129, 81, 129, 156, 94, 80, 129, 116, 141, 111, 82, 134, 137, 75, 141, 13, 89, 133, 80, 216, 78, 28, 150, 54, 114, 121, 129, 31, 141, 204, 91, 163, 139, 68, 150, 135, 89, 26, 127, 144, 84, 118, 86, 14, 86, 229, 139, 57, 101, 130, 105, 153, 148, 214, 118, 137, 110, 114, 94, 24, 117, 70, 103, 209, 103, 255, 122, 157, 128, 118, 141, 31, 97, 198, 121, 98, 101, 99, 141, 136, 81, 26, 82, 162, 148, 56, 127, 155, 128, 178, 126, 151, 92, 47, 110, 96, 103, 217, 123, 139, 118, 216, 154, 143, 129, 148, 127, 213, 124, 30, 100, 80, 149, 63, 122, 74, 84, 229, 84, 76, 107, 1, 100, 8, 98, 61, 158, 243, 128, 153, 117, 114, 82, 105, 151, 91, 132, 60, 104, 228, 134, 1, 150, 148, 150, 236, 148, 42, 78, 4, 84, 217, 126, 57, 104, 223, 141, 21, 128, 244, 102, 154, 94, 185, 127, 194, 87, 63, 128, 151, 104, 229, 93, 59, 101, 159, 82, 109, 96, 154, 159, 155, 79, 172, 142, 108, 81, 171, 91, 19, 95, 233, 93, 94, 108, 241, 98, 33, 141, 113, 81, 169, 148, 254, 82, 159, 108, 223, 130, 215, 114, 162, 87, 132, 103, 45, 141, 31, 89, 156, 143, 199, 131, 149, 84, 141, 123, 48, 79, 189, 108, 100, 91, 209, 89, 19, 159, 228, 83, 202, 134, 168, 154, 55, 140, 161, 128, 69, 101, 126, 152, 250, 86, 199, 150, 46, 82, 220, 116, 80, 82, 225, 91, 2, 99, 2, 137, 86, 78, 208, 98, 42, 96, 250, 104, 115, 81, 152, 91, 160, 81, 194, 137, 161, 123, 134, 153, 80, 127, 239, 96, 76, 112, 47, 141, 73, 81, 127, 94, 27, 144, 112, 116, 196, 137, 45, 87, 69, 120, 82, 95, 159, 159, 250, 149, 104, 143, 60, 155, 225, 139, 120, 118, 66, 104, 220, 103, 234, 141, 53, 141, 61, 82, 138, 143, 218, 110, 205, 104, 5, 149, 237, 144, 253, 86, 156, 103, 249, 136, 199, 143, 200, 84, 184, 154, 105, 91, 119, 109, 38, 108, 165, 78, 179, 91, 135, 154, 99, 145, 168, 97, 175, 144, 233, 151, 43, 84, 181, 109, 210, 91, 253, 81, 138, 85, 85, 127, 240, 127, 188, 100, 77, 99, 241, 101, 190, 97, 141, 96, 10, 113, 87, 108, 73, 108, 47, 89, 109, 103, 42, 130, 213, 88, 142, 86, 106, 140, 235, 107, 221, 144, 125, 89, 23, 128, 247, 83, 105, 109, 117, 84, 157, 85, 119, 131, 207, 131, 56, 104, 190, 121, 140, 84, 85, 79, 8, 84, 210, 118, 137, 140, 2, 150, 179, 108, 184, 109, 107, 141, 16, 137, 100, 158, 58, 141, 63, 86, 209, 158, 213, 117, 136, 95, 224, 114, 104, 96, 252, 84, 168, 78, 42, 106, 97, 136, 82, 96, 112, 143, 196, 84, 216, 112, 121, 134, 63, 158, 42, 109, 143, 91, 24, 95, 162, 126, 137, 85, 175, 79, 52, 115, 60, 84, 154, 83, 25, 80, 14, 84, 124, 84, 78, 78, 253, 95, 90, 116, 246, 88, 107, 132, 225, 128, 116, 135, 208, 114, 202, 124, 86, 110, 39, 95, 78, 134, 44, 85, 164, 98, 146, 78, 170, 108, 55, 98, 177, 130, 215, 84, 78, 83, 62, 115, 209, 110, 59, 117, 18, 82, 22, 83, 221, 139, 208, 105, 138, 95, 0, 96, 238, 109, 79, 87, 34, 107, 175, 115, 83, 104, 216, 143, 19, 127, 98, 99, 163, 96, 36, 85, 234, 117, 98, 140, 21, 113, 163, 109, 166, 91, 123, 94, 82, 131, 76, 97, 196, 158, 250, 120, 87, 135, 39, 124, 135, 118, 240, 81, 246, 96, 76, 113, 67, 102, 76, 94, 77, 96, 14, 140, 112, 112, 37, 99, 137, 143, 189, 95, 98, 96, 212, 134, 222, 86, 193, 107, 148, 96, 103, 97, 73, 83, 224, 96, 102, 102, 63, 141, 253, 121, 26, 79, 233, 112, 71, 108, 179, 139, 242, 139, 216, 126, 100, 131, 15, 102, 90, 90, 66, 155, 81, 109, 247, 109, 65, 140, 59, 109, 25, 79, 107, 112, 183, 131, 22, 98, 209, 96, 13, 151, 39, 141, 120, 121, 251, 81, 62, 87, 250, 87, 58, 103, 120, 117, 61, 122, 239, 121, 149, 123, 140, 128, 101, 153, 249, 143, 192, 111, 165, 139, 33, 158, 236, 89, 233, 126, 9, 127, 9, 84, 129, 103, 216, 104, 145, 143, 77, 124, 198, 150, 202, 83, 37, 96, 190, 117, 114, 108, 115, 83, 201, 90, 167, 126, 36, 99, 224, 81, 10, 129, 241, 93, 223, 132, 128, 98, 128, 81, 99, 91, 14, 79, 109, 121, 66, 82, 184, 96, 78, 109, 196, 91, 194, 91, 161, 139, 176, 139, 226, 101, 204, 95, 69, 150, 147, 89, 231, 126, 170, 126, 9, 86, 183, 103, 57, 89, 115, 79, 182, 91, 160, 82, 90, 131, 138, 152, 62, 141, 50, 117, 190, 148, 71, 80, 60, 122, 247, 78, 182, 103, 126, 154, 193, 90, 124, 107, 209, 118, 90, 87, 22, 92, 58, 123, 244, 149, 78, 113, 124, 81, 169, 128, 112, 130, 120, 89, 4, 127, 39, 131, 192, 104, 236, 103, 177, 120, 119, 120, 227, 98, 97, 99, 128, 123, 237, 79, 106, 82, 207, 81, 80, 131, 219, 105, 116, 146, 245, 141, 49, 141, 193, 137, 46, 149, 173, 123, 246, 78, 101, 80, 48, 130, 81, 82, 111, 153, 16, 110, 133, 110, 167, 109, 250, 94, 245, 80, 220, 89, 6, 92, 70, 109, 95, 108, 134, 117, 139, 132, 104, 104, 86, 89, 178, 139, 32, 83, 113, 145, 77, 150, 73, 133, 18, 105, 1, 121, 38, 113, 246, 128, 164, 78, 202, 144, 71, 109, 132, 154, 7, 90, 188, 86, 5, 100, 240, 148, 235, 119, 165, 79, 26, 129, 225, 114, 210, 137, 122, 153, 52, 127, 222, 126, 127, 82, 89, 101, 117, 145, 127, 143, 131, 143, 235, 83, 150, 122, 237, 99, 165, 99, 134, 118, 248, 121, 87, 136, 54, 150, 42, 98, 171, 82, 130, 130, 84, 104, 112, 103, 119, 99, 107, 119, 237, 122, 1, 109, 211, 126, 227, 137, 208, 89, 18, 98, 201, 133, 165, 130, 76, 117, 31, 80, 203, 78, 165, 117, 235, 139, 74, 92, 254, 93, 75, 123, 164, 101, 209, 145, 202, 78, 37, 109, 95, 137, 39, 125, 38, 149, 197, 78, 40, 140, 219, 143, 115, 151, 75, 102, 129, 121, 209, 143, 236, 112, 120, 109, 61, 92, 178, 82, 70, 131, 98, 81, 14, 131, 91, 119, 118, 102, 184, 156, 172, 78, 202, 96, 190, 124, 179, 124, 207, 126, 149, 78, 102, 139, 111, 102, 136, 152, 89, 151, 131, 88, 108, 101, 92, 149, 132, 95, 201, 117, 86, 151, 223, 122, 222, 122, 192, 81, 175, 112, 152, 122, 234, 99, 118, 122, 160, 126, 150, 115, 237, 151, 69, 78, 120, 112, 93, 78, 82, 145, 169, 83, 81, 101, 231, 101, 252, 129, 5, 130, 142, 84, 49, 92, 154, 117, 160, 151, 216, 98, 217, 114, 189, 117, 69, 92, 121, 154, 202, 131, 64, 92, 128, 84, 233, 119, 62, 78, 174, 108, 90, 128, 210, 98, 110, 99, 232, 93, 119, 81, 221, 141, 30, 142, 47, 149, 241, 79, 229, 83, 231, 96, 172, 112, 103, 82, 80, 99, 67, 158, 31, 90, 38, 80, 55, 119, 119, 83, 226, 126, 133, 100, 43, 101, 137, 98, 152, 99, 20, 80, 53, 114, 201, 137, 179, 81, 192, 139, 221, 126, 71, 87, 204, 131, 167, 148, 155, 81, 27, 84, 251, 92, 202, 79, 227, 122, 90, 109, 225, 144, 143, 154, 128, 85, 150, 84, 97, 83, 175, 84, 0, 95, 233, 99, 119, 105, 239, 81, 104, 97, 10, 82, 42, 88, 216, 82, 78, 87, 13, 120, 11, 119, 183, 94, 119, 97, 224, 124, 91, 98, 151, 98, 162, 78, 149, 112, 3, 128, 247, 98, 228, 112, 96, 151, 119, 87, 219, 130, 239, 103, 245, 104, 213, 120, 151, 152, 209, 121, 243, 88, 179, 84, 239, 83, 52, 110, 75, 81, 59, 82, 162, 91, 254, 139, 175, 128, 67, 85, 166, 87, 115, 96, 81, 87, 45, 84, 122, 122, 80, 96, 84, 91, 167, 99, 160, 98, 227, 83, 99, 98, 199, 91, 175, 103, 237, 84, 159, 122, 230, 130, 119, 145, 147, 94, 228, 136, 56, 89, 174, 87, 14, 99, 232, 141, 239, 128, 87, 87, 119, 123, 169, 79, 235, 95, 189, 91, 62, 107, 33, 83, 80, 123, 194, 114, 70, 104, 255, 119, 54, 119, 247, 101, 181, 81, 143, 78, 212, 118, 191, 92, 165, 122, 117, 132, 78, 89, 65, 155, 128, 80, 136, 153, 39, 97, 131, 110, 100, 87, 6, 102, 70, 99, 240, 86, 236, 98, 105, 98, 211, 94, 20, 150, 131, 87, 201, 98, 135, 85, 33, 135, 74, 129, 163, 143, 102, 85, 177, 131, 101, 103, 86, 141, 221, 132, 106, 90, 15, 104, 230, 98, 238, 123, 17, 150, 112, 81, 156, 111, 48, 140, 253, 99, 200, 137, 210, 97, 6, 127, 194, 112, 229, 110, 5, 116, 148, 105, 252, 114, 202, 94, 206, 144, 23, 103, 106, 109, 94, 99, 179, 82, 98, 114, 1, 128, 108, 79, 229, 89, 106, 145, 217, 112, 157, 109, 210, 82, 80, 78, 247, 150, 109, 149, 126, 133, 202, 120, 47, 125, 33, 81, 146, 87, 194, 100, 139, 128, 123, 124, 234, 108, 241, 104, 94, 105, 183, 81, 152, 83, 168, 104, 129, 114, 206, 158, 241, 123, 248, 114, 187, 121, 19, 111, 6, 116, 78, 103, 204, 145, 164, 156, 60, 121, 137, 131, 84, 131, 15, 84, 23, 104, 61, 78, 137, 83, 177, 82, 62, 120, 134, 83, 41, 82, 136, 80, 139, 79, 208, 79, 226, 117, 203, 122, 146, 124, 165, 108, 182, 150, 155, 82, 131, 116, 233, 84, 233, 79, 84, 128, 178, 131, 222, 143, 112, 149, 201, 94, 28, 96, 159, 109, 24, 94, 91, 101, 56, 129, 254, 148, 75, 96, 188, 112, 195, 126, 174, 124, 201, 81, 129, 104, 177, 124, 111, 130, 36, 78, 134, 143, 207, 145, 126, 102, 174, 78, 5, 140, 169, 100, 74, 128, 218, 80, 151, 117, 206, 113, 229, 91, 189, 143, 102, 111, 134, 78, 130, 100, 99, 149, 214, 94, 153, 101, 23, 82, 194, 136, 200, 112, 163, 82, 14, 115, 51, 116, 151, 103, 247, 120, 22, 151, 52, 78, 187, 144, 222, 156, 203, 109, 219, 81, 65, 141, 29, 84, 206, 98, 178, 115, 241, 131, 246, 150, 132, 159, 195, 148, 54, 79, 154, 127, 204, 81, 117, 112, 117, 150, 173, 92, 134, 152, 230, 83, 228, 78, 156, 110, 9, 116, 180, 105, 107, 120, 143, 153, 89, 117, 24, 82, 36, 118, 65, 109, 243, 103, 109, 81, 153, 159, 75, 128, 153, 84, 60, 123, 191, 122, 134, 150, 132, 87, 226, 98, 71, 150, 124, 105, 4, 90, 2, 100, 211, 123, 15, 111, 75, 150, 166, 130, 98, 83, 133, 152, 144, 94, 137, 112, 179, 99, 100, 83, 79, 134, 129, 156, 147, 158, 140, 120, 50, 151, 239, 141, 66, 141, 127, 158, 94, 111, 132, 121, 85, 95, 70, 150, 46, 98, 116, 154, 21, 84, 221, 148, 163, 79, 197, 101, 101, 92, 97, 92, 21, 127, 81, 134, 47, 108, 139, 95, 135, 115, 228, 110, 255, 126, 230, 92, 27, 99, 106, 91, 230, 110, 117, 83, 113, 78, 160, 99, 101, 117, 161, 98, 110, 143, 38, 79, 209, 78, 166, 108, 182, 126, 186, 139, 29, 132, 186, 135, 87, 127, 59, 144, 35, 149, 169, 123, 161, 154, 248, 136, 61, 132, 27, 109, 134, 154, 220, 126, 136, 89, 187, 158, 155, 115, 1, 120, 130, 134, 108, 154, 130, 154, 27, 86, 23, 84, 203, 87, 112, 78, 166, 158, 86, 83, 200, 143, 9, 129, 146, 119, 146, 153, 238, 134, 225, 110, 19, 133, 252, 102, 98, 97, 43, 111, 41, 140, 146, 130, 43, 131, 242, 118, 19, 108, 217, 95, 189, 131, 43, 115, 5, 131, 26, 149, 219, 107, 219, 119, 198, 148, 111, 83, 2, 131, 146, 81, 61, 94, 140, 140, 56, 141, 72, 78, 171, 115, 154, 103, 133, 104, 118, 145, 9, 151, 100, 113, 161, 108, 9, 119, 146, 90, 65, 149, 207, 107, 142, 127, 39, 102, 208, 91, 185, 89, 154, 90, 232, 149, 247, 149, 236, 78, 12, 132, 153, 132, 172, 106, 223, 118, 48, 149, 27, 115, 166, 104, 95, 91, 47, 119, 154, 145, 97, 151, 220, 124, 247, 143, 28, 140, 37, 95, 115, 124, 216, 121, 197, 137, 204, 108, 28, 135, 198, 91, 66, 94, 201, 104, 32, 119, 245, 126, 149, 81, 77, 81, 201, 82, 41, 90, 5, 127, 98, 151, 215, 130, 207, 99, 132, 119, 208, 133, 210, 121, 58, 110, 153, 94, 153, 89, 17, 133, 109, 112, 17, 108, 191, 98, 191, 118, 79, 101, 175, 96, 253, 149, 14, 102, 159, 135, 35, 158, 237, 148, 13, 84, 125, 84, 44, 140, 120, 100, 121, 100, 17, 134, 33, 106, 156, 129, 232, 120, 105, 100, 84, 155, 185, 98, 43, 103, 171, 131, 168, 88, 216, 158, 171, 108, 32, 111, 222, 91, 76, 150, 11, 140, 95, 114, 208, 103, 199, 98, 97, 114, 169, 78, 198, 89, 205, 107, 147, 88, 174, 102, 85, 94, 223, 82, 85, 97, 40, 103, 238, 118, 102, 119, 103, 114, 70, 122, 255, 98, 234, 84, 80, 84, 160, 148, 163, 144, 28, 90, 179, 126, 22, 108, 67, 78, 118, 89, 16, 128, 72, 89, 87, 83, 55, 117, 190, 150, 202, 86, 32, 99, 17, 129, 124, 96, 249, 149, 214, 109, 98, 84, 129, 153, 133, 81, 233, 90, 253, 128, 174, 89, 19, 151, 42, 80, 229, 108, 60, 92, 223, 98, 96, 79, 63, 83, 123, 129, 6, 144, 186, 110, 43, 133, 200, 98, 116, 94, 190, 120, 181, 100, 123, 99, 245, 95, 24, 90, 127, 145, 31, 158, 63, 92, 79, 99, 66, 128, 125, 91, 110, 85, 74, 149, 77, 149, 133, 109, 168, 96, 224, 103, 222, 114, 221, 81, 129, 91, 231, 98, 222, 108, 91, 114, 109, 98, 174, 148, 189, 126, 19, 129, 83, 109, 156, 81, 4, 95, 116, 89, 170, 82, 18, 96, 115, 89, 150, 102, 80, 134, 159, 117, 42, 99, 230, 97, 239, 124, 250, 139, 230, 84, 39, 107, 37, 158, 180, 107, 213, 133, 85, 84, 118, 80, 164, 108, 106, 85, 180, 141, 44, 114, 21, 94, 21, 96, 54, 116, 205, 98, 146, 99, 76, 114, 152, 95, 67, 110, 62, 109, 0, 101, 88, 111, 216, 118, 208, 120, 252, 118, 84, 117, 36, 82, 219, 83, 83, 78, 158, 94, 193, 101, 42, 128, 214, 128, 155, 98, 134, 84, 40, 82, 174, 112, 141, 136, 209, 141, 225, 108, 120, 84, 218, 128, 249, 87, 244, 136, 84, 141, 106, 150, 77, 145, 105, 79, 155, 108, 183, 85, 198, 118, 48, 120, 168, 98, 249, 112, 142, 111, 109, 95, 236, 132, 218, 104, 124, 120, 247, 123, 168, 129, 11, 103, 79, 158, 103, 99, 176, 120, 111, 87, 18, 120, 57, 151, 121, 98, 171, 98, 136, 82, 53, 116, 215, 107, 100, 85, 62, 129, 178, 117, 174, 118, 57, 83, 222, 117, 251, 80, 65, 92, 108, 139, 199, 123, 79, 80, 71, 114, 151, 154, 216, 152, 2, 111, 226, 116, 104, 121, 135, 100, 165, 119, 252, 98, 145, 152, 43, 141, 193, 84, 88, 128, 82, 78, 106, 87, 249, 130, 13, 132, 115, 94, 237, 81, 246, 116, 196, 139, 79, 92, 97, 87, 252, 108, 135, 152, 70, 90, 52, 120, 68, 155, 235, 143, 149, 124, 86, 82, 81, 98, 250, 148, 198, 78, 134, 131, 97, 132, 233, 131, 178, 132, 212, 87, 52, 103, 3, 87, 110, 102, 102, 109, 49, 140, 221, 102, 17, 112, 31, 103, 58, 107, 22, 104, 26, 98, 187, 89, 3, 78, 196, 81, 6, 111, 210, 103, 143, 108, 118, 81, 203, 104, 71, 89, 103, 107, 102, 117, 14, 93, 16, 129, 80, 159, 215, 101, 72, 121, 65, 121, 145, 154, 119, 141, 130, 92, 94, 78, 1, 79, 47, 84, 81, 89, 12, 120, 104, 86, 20, 108, 196, 143, 3, 95, 125, 108, 227, 108, 171, 139, 144, 99, 112, 96, 61, 109, 117, 114, 102, 98, 142, 148, 197, 148, 67, 83, 193, 143, 126, 123, 223, 78, 38, 140, 126, 78, 212, 158, 177, 148, 179, 148, 77, 82, 92, 111, 99, 144, 69, 109, 52, 140, 17, 88, 76, 93, 32, 107, 73, 107, 170, 103, 91, 84, 84, 129, 140, 127, 153, 88, 55, 133, 58, 95, 162, 98, 71, 106, 57, 149, 114, 101, 132, 96, 101, 104, 167, 119, 84, 78, 168, 79, 231, 93, 152, 151, 172, 100, 216, 127, 237, 92, 207, 79, 141, 122, 7, 82, 4, 131, 20, 78, 47, 96, 131, 122, 166, 148, 181, 79, 178, 78, 230, 121, 52, 116, 228, 82, 185, 130, 210, 100, 189, 121, 221, 91, 129, 108, 82, 151, 123, 143, 34, 108, 62, 80, 127, 83, 5, 110, 206, 100, 116, 102, 48, 108, 197, 96, 119, 152, 247, 139, 134, 94, 60, 116, 119, 122, 203, 121, 24, 78, 177, 144, 3, 116, 66, 108, 218, 86, 75, 145, 197, 108, 139, 141, 58, 83, 198, 134, 242, 102, 175, 142, 72, 92, 113, 154, 32, 110, 214, 83, 54, 90, 139, 159, 163, 141, 187, 83, 8, 87, 167, 152, 67, 103, 155, 145, 201, 108, 104, 81, 202, 117, 243, 98, 172, 114, 56, 82, 157, 82, 58, 127, 148, 112, 56, 118, 116, 83, 74, 158, 183, 105, 110, 120, 192, 150, 217, 136, 164, 127, 54, 113, 195, 113, 137, 81, 211, 103, 228, 116, 228, 88, 24, 101, 183, 86, 169, 139, 118, 153, 112, 98, 213, 126, 249, 96, 237, 112, 236, 88, 193, 78, 186, 78, 205, 95, 231, 151, 251, 78, 164, 139, 3, 82, 138, 89, 171, 126, 84, 98, 205, 78, 229, 101, 14, 98, 56, 131, 201, 132, 99, 131, 141, 135, 148, 113, 182, 110, 185, 91, 210, 126, 151, 81, 201, 99, 212, 103, 137, 128, 57, 131, 21, 136, 18, 81, 122, 91, 130, 89, 177, 143, 115, 78, 93, 108, 101, 81, 37, 137, 111, 143, 46, 150, 74, 133, 94, 116, 16, 149, 240, 149, 166, 109, 229, 130, 49, 95, 146, 100, 18, 109, 40, 132, 110, 129, 195, 156, 94, 88, 91, 141, 9, 78, 193, 83, 30, 79, 99, 101, 81, 104, 211, 85, 39, 78, 20, 100, 154, 154, 107, 98, 194, 90, 95, 116, 114, 130, 169, 109, 238, 104, 231, 80, 142, 131, 2, 120, 64, 103, 57, 82, 153, 108, 177, 126, 187, 80, 101, 85, 94, 113, 91, 123, 82, 102, 202, 115, 235, 130, 73, 103, 113, 92, 32, 82, 125, 113, 107, 136, 234, 149, 85, 150, 197, 100, 97, 141, 179, 129, 132, 85, 85, 108, 71, 98, 46, 127, 146, 88, 36, 79, 70, 85, 79, 141, 76, 102, 10, 78, 26, 92, 243, 136, 162, 104, 78, 99, 13, 122, 231, 112, 141, 130, 250, 82, 246, 151, 17, 92, 232, 84, 181, 144, 205, 126, 98, 89, 74, 141, 199, 134, 12, 130, 13, 130, 102, 141, 68, 100, 4, 92, 81, 97, 137, 109, 62, 121, 190, 139, 55, 120, 51, 117, 123, 84, 56, 79, 171, 142, 241, 109, 32, 90, 197, 126, 94, 121, 136, 108, 161, 91, 118, 90, 26, 117, 190, 128, 78, 97, 23, 110, 240, 88, 31, 117, 37, 117, 114, 114, 71, 83, 243, 126, 1, 119, 219, 118, 105, 82, 220, 128, 35, 87, 8, 94, 49, 89, 238, 114, 189, 101, 127, 110, 215, 139, 56, 92, 113, 134, 65, 83, 243, 119, 254, 98, 246, 101, 192, 78, 223, 152, 128, 134, 158, 91, 198, 139, 242, 83, 226, 119, 127, 79, 78, 92, 118, 154, 203, 89, 15, 95, 58, 121, 235, 88, 22, 78, 255, 103, 139, 78, 237, 98, 147, 138, 29, 144, 191, 82, 47, 102, 220, 85, 108, 86, 2, 144, 213, 78, 141, 79, 202, 145, 112, 153, 15, 108, 2, 94, 67, 96, 164, 91, 198, 137, 213, 139, 54, 101, 75, 98, 150, 153, 136, 91, 255, 91, 136, 99, 46, 85, 215, 83, 38, 118, 125, 81, 44, 133, 162, 103, 179, 104, 138, 107, 146, 98, 147, 143, 212, 83, 18, 130, 209, 109, 143, 117, 102, 78, 78, 141, 112, 91, 159, 113, 175, 133, 145, 102, 217, 102, 114, 127, 0, 135, 205, 158, 32, 159, 94, 92, 47, 103, 240, 143, 17, 104, 95, 103, 13, 98, 214, 122, 133, 88, 182, 94, 112, 101, 49, 111, 85, 96, 55, 82, 13, 128, 84, 100, 112, 136, 41, 117, 5, 94, 19, 104, 244, 98, 28, 151, 204, 83, 61, 114, 1, 140, 52, 108, 97, 119, 14, 122, 46, 84, 172, 119, 122, 152, 28, 130, 244, 139, 85, 120, 20, 103, 193, 112, 175, 101, 149, 100, 54, 86, 29, 96, 193, 121, 248, 83, 29, 78, 123, 107, 134, 128, 250, 91, 227, 85, 219, 86, 58, 79, 60, 79, 114, 153, 243, 93, 126, 103, 56, 128, 2, 96, 130, 152, 1, 144, 139, 91, 188, 139, 245, 139, 28, 100, 88, 130, 222, 100, 253, 85, 207, 130, 101, 145, 215, 79, 32, 125, 31, 144, 159, 124, 243, 80, 81, 88, 175, 110, 191, 91, 201, 139, 131, 128, 120, 145, 156, 132, 151, 123, 125, 134, 139, 150, 143, 150, 229, 126, 211, 154, 142, 120, 129, 92, 87, 122, 66, 144, 167, 150, 95, 121, 89, 91, 95, 99, 11, 123, 209, 132, 173, 104, 6, 85, 41, 127, 16, 116, 34, 125, 1, 149, 64, 98, 76, 88, 214, 78, 131, 91, 121, 89, 84, 88, 109, 115, 30, 99, 75, 142, 15, 142, 206, 128, 212, 130, 172, 98, 240, 83, 240, 108, 94, 145, 42, 89, 1, 96, 112, 108, 77, 87, 74, 100, 42, 141, 43, 118, 233, 110, 91, 87, 128, 106, 240, 117, 109, 111, 45, 140, 8, 140, 102, 87, 239, 107, 146, 136, 179, 120, 162, 99, 249, 83, 173, 112, 100, 108, 88, 88, 42, 100, 2, 88, 224, 104, 155, 129, 16, 85, 214, 124, 24, 80, 186, 142, 204, 109, 159, 141, 235, 112, 143, 99, 155, 109, 212, 110, 230, 126, 4, 132, 67, 104, 3, 144, 216, 109, 118, 150, 168, 139, 87, 89, 121, 114, 228, 133, 126, 129, 188, 117, 138, 138, 175, 104, 84, 82, 34, 142, 17, 149, 208, 99, 152, 152, 68, 142, 124, 85, 83, 79, 255, 102, 143, 86, 213, 96, 149, 109, 67, 82, 73, 92, 41, 89, 251, 109, 107, 88, 48, 117, 28, 117, 108, 96, 20, 130, 70, 129, 17, 99, 97, 103, 226, 143, 58, 119, 243, 141, 52, 141, 193, 148, 22, 94, 133, 83, 44, 84, 195, 112, 64, 108, 247, 94, 92, 80, 173, 78, 173, 94, 58, 99, 71, 130, 26, 144, 80, 104, 110, 145, 179, 119, 12, 84, 220, 148, 100, 95, 229, 122, 118, 104, 69, 99, 82, 123, 223, 126, 219, 117, 119, 80, 149, 98, 52, 89, 15, 144, 248, 81, 195, 121, 129, 122, 254, 86, 146, 95, 20, 144, 130, 109, 96, 92, 31, 87, 16, 84, 84, 81, 77, 110, 226, 86, 168, 99, 147, 152, 127, 129, 21, 135, 42, 137, 0, 144, 30, 84, 111, 92, 192, 129, 214, 98, 88, 98, 49, 129, 53, 158, 64, 150, 110, 154, 124, 154, 45, 105, 165, 89, 211, 98, 62, 85, 22, 99, 199, 84, 217, 134, 60, 109, 3, 90, 230, 116, 156, 136, 106, 107, 22, 89, 76, 140, 47, 95, 126, 110, 169, 115, 125, 152, 56, 78, 247, 112, 140, 91, 151, 120, 61, 99, 90, 102, 150, 118, 203, 96, 155, 91, 73, 90, 7, 78, 85, 129, 106, 108, 139, 115, 161, 78, 137, 103, 81, 127, 128, 95, 250, 101, 27, 103, 216, 95, 132, 89, 1, 90, 205, 93, 174, 95, 113, 83, 230, 151, 221, 143, 69, 104, 244, 86, 47, 85, 223, 96, 58, 78, 77, 111, 244, 126, 199, 130, 14, 132, 212, 89, 31, 79, 42, 79, 62, 92, 172, 126, 42, 103, 26, 133, 115, 84, 79, 117, 195, 128, 130, 85, 79, 155, 77, 79, 45, 110, 19, 140, 9, 92, 112, 97, 107, 83, 31, 118, 41, 110, 138, 134, 135, 101, 251, 149, 185, 126, 59, 84, 51, 122, 10, 125, 238, 149, 225, 85, 193, 127, 238, 116, 29, 99, 23, 135, 161, 109, 157, 122, 17, 98, 161, 101, 103, 83, 225, 99, 131, 108, 235, 93, 92, 84, 168, 148, 76, 78, 97, 108, 236, 139, 75, 92, 224, 101, 156, 130, 167, 104, 62, 84, 52, 84, 203, 107, 102, 107, 148, 78, 66, 99, 72, 83, 30, 130, 13, 79, 174, 79, 94, 87, 10, 98, 254, 150, 100, 102, 105, 114, 255, 82, 161, 82, 159, 96, 239, 139, 20, 102, 153, 113, 144, 103, 127, 137, 82, 120, 253, 119, 112, 102, 59, 86, 56, 84, 33, 149, 122, 114, 0, 122, 111, 96, 12, 94, 137, 96, 157, 129, 21, 89, 220, 96, 132, 113, 239, 112, 170, 110, 80, 108, 128, 114, 132, 106, 173, 136, 45, 94, 96, 78, 179, 90, 156, 85, 227, 148, 23, 109, 251, 124, 153, 150, 15, 98, 198, 126, 142, 119, 126, 134, 35, 83, 30, 151, 150, 143, 135, 102, 225, 92, 160, 79, 237, 114, 11, 78, 166, 83, 15, 89, 19, 84, 128, 99, 40, 149, 72, 81, 217, 78, 156, 156, 164, 126, 184, 84, 36, 141, 84, 136, 55, 130, 242, 149, 142, 109, 38, 95, 204, 90, 62, 102, 105, 150, 176, 115, 46, 115, 191, 83, 122, 129, 133, 153, 161, 127, 170, 91, 119, 150, 80, 150, 191, 126, 248, 118, 162, 83, 118, 149, 153, 153, 177, 123, 68, 137, 88, 110, 97, 78, 212, 127, 101, 121, 230, 139, 243, 96, 205, 84, 171, 78, 121, 152, 247, 93, 97, 106, 207, 80, 17, 84, 97, 140, 39, 132, 93, 120, 4, 151, 74, 82, 238, 84, 163, 86, 0, 149, 136, 109, 181, 91, 198, 109, 83, 102, 15, 92, 93, 91, 33, 104, 150, 128, 120, 85, 17, 123, 72, 101, 84, 105, 155, 78, 71, 107, 78, 135, 139, 151, 79, 83, 31, 99, 58, 100, 170, 144, 156, 101, 193, 128, 16, 140, 153, 81, 176, 104, 120, 83, 249, 135, 200, 97, 196, 108, 251, 108, 34, 140, 81, 92, 170, 133, 175, 130, 12, 149, 35, 107, 155, 143, 176, 101, 251, 95, 195, 95, 225, 79, 69, 136, 31, 102, 101, 129, 41, 115, 250, 96, 116, 81, 17, 82, 139, 87, 98, 95, 162, 144, 76, 136, 146, 145, 120, 94, 79, 103, 39, 96, 211, 89, 68, 81, 246, 81, 248, 128, 8, 83, 121, 108, 196, 150, 138, 113, 17, 79, 238, 79, 158, 127, 61, 103, 197, 85, 8, 149, 192, 121, 150, 136, 227, 126, 159, 88, 12, 98, 0, 151, 90, 134, 24, 86, 123, 152, 144, 95, 184, 139, 196, 132, 87, 145, 217, 83, 237, 101, 143, 94, 92, 117, 100, 96, 110, 125, 127, 90, 234, 126, 237, 126, 105, 143, 167, 85, 163, 91, 172, 96, 203, 101, 132, 115, 9, 144, 99, 118, 41, 119, 218, 126, 116, 151, 155, 133, 102, 91, 116, 122, 234, 150, 64, 136, 203, 82, 143, 113, 170, 95, 236, 101, 226, 139, 251, 91, 111, 154, 225, 93, 137, 107, 91, 108, 173, 139, 175, 139, 10, 144, 197, 143, 139, 83, 188, 98, 38, 158, 45, 158, 64, 84, 43, 78, 189, 130, 89, 114, 156, 134, 22, 93, 89, 136, 175, 109, 197, 150, 209, 84, 154, 78, 182, 139, 9, 113, 189, 84, 9, 150, 223, 112, 249, 109, 208, 118, 37, 78, 20, 120, 18, 135, 169, 92, 246, 94, 0, 138, 156, 152, 14, 150, 142, 112, 191, 108, 68, 89, 169, 99, 60, 119, 77, 136, 20, 111, 115, 130, 48, 88, 213, 113, 140, 83, 26, 120, 193, 150, 1, 85, 102, 95, 48, 113, 180, 91, 26, 140, 140, 154, 131, 107, 46, 89, 47, 158, 231, 121, 104, 103, 108, 98, 111, 79, 161, 117, 138, 127, 11, 109, 51, 150, 39, 108, 240, 78, 210, 117, 123, 81, 55, 104, 62, 111, 128, 144, 112, 129, 150, 89, 118, 116, 71, 100, 39, 92, 101, 144, 145, 122, 35, 140, 218, 89, 172, 84, 0, 130, 111, 131, 129, 137, 0, 128, 48, 105, 78, 86, 54, 128, 55, 114, 206, 145, 182, 81, 95, 78, 117, 152, 150, 99, 26, 78, 246, 83, 243, 102, 75, 129, 28, 89, 178, 109, 0, 78, 249, 88, 59, 83, 214, 99, 241, 148, 157, 79, 10, 79, 99, 136, 144, 152, 55, 89, 87, 144, 251, 121, 234, 78, 240, 128, 145, 117, 130, 108, 156, 91, 232, 89, 93, 95, 5, 105, 129, 134, 26, 80, 242, 93, 89, 78, 227, 119, 229, 78, 122, 130, 145, 98, 19, 102, 145, 144, 121, 92, 191, 78, 121, 95, 198, 129, 56, 144, 132, 128, 171, 117, 166, 78, 212, 136, 15, 97, 197, 107, 198, 95, 73, 78, 202, 118, 162, 110, 227, 139, 174, 139, 10, 140, 209, 139, 2, 95, 252, 127, 204, 127, 206, 126, 53, 131, 107, 131, 224, 86, 183, 107, 243, 151, 52, 150, 251, 89, 31, 84, 246, 148, 235, 109, 197, 91, 110, 153, 57, 92, 21, 95, 144, 150, 112, 83, 241, 130, 49, 106, 116, 90, 112, 158, 148, 94, 40, 127, 185, 131, 36, 132, 37, 132, 103, 131, 71, 135, 206, 143, 98, 141, 200, 118, 113, 95, 150, 152, 108, 120, 32, 102, 223, 84, 229, 98, 99, 79, 195, 129, 200, 117, 184, 94, 205, 150, 10, 142, 249, 134, 143, 84, 243, 108, 140, 109, 56, 108, 127, 96, 199, 82, 40, 117, 125, 94, 24, 79, 160, 96, 231, 95, 36, 92, 49, 117, 174, 144, 192, 148, 185, 114, 185, 108, 56, 110, 73, 145, 9, 103, 203, 83, 243, 83, 81, 79, 201, 145, 241, 139, 200, 83, 124, 94, 194, 143, 228, 109, 142, 78, 194, 118, 134, 105, 94, 134, 26, 97, 6, 130, 89, 79, 222, 79, 62, 144, 124, 156, 9, 97, 29, 110, 20, 110, 133, 150, 136, 78, 49, 90, 232, 150, 14, 78, 127, 92, 185, 121, 135, 91, 237, 139, 189, 127, 137, 115, 223, 87, 139, 130, 193, 144, 1, 84, 71, 144, 187, 85, 234, 92, 161, 95, 8, 97, 50, 107, 241, 114, 178, 128, 137, 138, 116, 109, 211, 91, 213, 136, 132, 152, 107, 140, 109, 154, 51, 158, 10, 110, 164, 81, 67, 81, 163, 87, 129, 136, 159, 83, 244, 99, 149, 143, 237, 86, 88, 84, 6, 87, 63, 115, 144, 110, 24, 127, 220, 143, 209, 130, 63, 97, 40, 96, 98, 150, 240, 102, 166, 126, 138, 141, 195, 141, 165, 148, 179, 92, 164, 124, 8, 103, 166, 96, 5, 150, 24, 128, 145, 78, 231, 144, 0, 83, 104, 150, 65, 81, 208, 143, 116, 133, 93, 145, 85, 102, 245, 151, 85, 91, 29, 83, 56, 120, 66, 103, 61, 104, 201, 84, 126, 112, 176, 91, 125, 143, 141, 81, 40, 87, 177, 84, 18, 101, 130, 102, 94, 141, 67, 141, 15, 129, 108, 132, 109, 144, 223, 124, 255, 81, 251, 133, 163, 103, 233, 101, 161, 111, 164, 134, 129, 142, 106, 86, 32, 144, 130, 118, 118, 112, 229, 113, 35, 141, 233, 98, 25, 82, 253, 108, 60, 141, 14, 96, 158, 88, 142, 97, 254, 102, 96, 141, 78, 98, 179, 85, 35, 110, 45, 103, 103, 143, 225, 148, 248, 149, 40, 119, 5, 104, 168, 105, 139, 84, 77, 78, 184, 112, 200, 139, 88, 100, 139, 101, 133, 91, 132, 122, 58, 80, 232, 91, 187, 119, 225, 107, 121, 138, 152, 124, 190, 108, 207, 118, 169, 101, 151, 143, 45, 93, 85, 92, 56, 134, 8, 104, 96, 83, 24, 98, 217, 122, 91, 110, 253, 126, 31, 106, 224, 122, 112, 95, 51, 111, 32, 95, 140, 99, 168, 109, 86, 103, 8, 78, 16, 94, 38, 141, 215, 78, 192, 128, 52, 118, 156, 150, 219, 98, 45, 102, 126, 98, 188, 108, 117, 141, 103, 113, 105, 127, 70, 81, 135, 128, 236, 83, 110, 144, 152, 98, 242, 84, 240, 134, 153, 143, 5, 128, 23, 149, 23, 133, 217, 143, 89, 109, 205, 115, 159, 101, 31, 119, 4, 117, 39, 120, 251, 129, 30, 141, 136, 148, 166, 79, 149, 103, 185, 117, 202, 139, 7, 151, 47, 99, 71, 149, 53, 150, 184, 132, 35, 99, 65, 119, 129, 95, 240, 114, 137, 78, 20, 96, 116, 101, 239, 98, 99, 107, 63, 101, 39, 94, 199, 117, 209, 144, 193, 139, 157, 130, 157, 103, 47, 101, 49, 84, 24, 135, 229, 119, 162, 128, 2, 129, 65, 108, 75, 78, 199, 126, 76, 128, 244, 118, 13, 105, 150, 107, 103, 98, 60, 80, 132, 79, 64, 87, 7, 99, 98, 107, 190, 141, 234, 83, 232, 101, 184, 126, 215, 95, 26, 99, 183, 99, 243, 129, 244, 129, 110, 127, 28, 94, 217, 92, 54, 82, 122, 102, 233, 121, 26, 122, 40, 141, 153, 112, 212, 117, 222, 110, 187, 108, 146, 122, 45, 78, 197, 118, 224, 95, 159, 148, 119, 136, 200, 126, 205, 121, 191, 128, 205, 145, 242, 78, 23, 79, 31, 130, 104, 84, 222, 93, 50, 109, 204, 139, 165, 124, 116, 143, 152, 128, 26, 94, 146, 84, 177, 118, 153, 91, 60, 102, 164, 154, 224, 115, 42, 104, 219, 134, 49, 103, 42, 115, 248, 139, 219, 139, 16, 144, 249, 122, 219, 112, 110, 113, 196, 98, 169, 119, 49, 86, 59, 78, 87, 132, 241, 103, 169, 82, 192, 134, 46, 141, 248, 148, 81, 123, 79, 79, 232, 108, 93, 121, 123, 154, 147, 98, 42, 114, 253, 98, 19, 78, 22, 120, 108, 143, 176, 100, 90, 141, 198, 123, 105, 104, 132, 94, 197, 136, 134, 89, 158, 100, 238, 88, 182, 114, 14, 105, 37, 149, 253, 143, 88, 141, 96, 87, 0, 127, 6, 140, 198, 81, 73, 99, 217, 98, 83, 83, 76, 104, 34, 116, 1, 131, 76, 145, 68, 85, 64, 119, 124, 112, 74, 109, 121, 81, 168, 84, 68, 141, 255, 89, 203, 110, 196, 109, 92, 91, 43, 125, 212, 78, 125, 124, 211, 110, 80, 91, 234, 129, 13, 110, 87, 91, 3, 155, 213, 104, 42, 142, 151, 91, 252, 126, 59, 96, 181, 126, 185, 144, 112, 141, 79, 89, 205, 99, 223, 121, 179, 141, 82, 83, 207, 101, 86, 121, 197, 139, 59, 150, 196, 126, 187, 148, 130, 126, 52, 86, 137, 145, 0, 103, 106, 127, 10, 92, 117, 144, 40, 102, 230, 93, 80, 79, 222, 103, 90, 80, 92, 79, 80, 87, 167, 94, 216, 226, 217, 226, 218, 226, 219, 226, 220, 226, 141, 78, 12, 78, 64, 81, 16, 78, 255, 94, 69, 83, 21, 78, 152, 78, 30, 78, 50, 155, 108, 91, 105, 86, 40, 78, 186, 121, 63, 78, 21, 83, 71, 78, 45, 89, 59, 114, 110, 83, 16, 108, 223, 86, 228, 128, 151, 153, 211, 107, 126, 119, 23, 159, 54, 78, 159, 78, 16, 159, 92, 78, 105, 78, 147, 78, 136, 130, 91, 91, 108, 85, 15, 86, 196, 78, 141, 83, 157, 83, 163, 83, 165, 83, 174, 83, 101, 151, 93, 141, 26, 83, 245, 83, 38, 83, 46, 83, 62, 83, 92, 141, 102, 83, 99, 83, 2, 82, 8, 82, 14, 82, 45, 82, 51, 82, 63, 82, 64, 82, 76, 82, 94, 82, 97, 82, 92, 82, 175, 132, 125, 82, 130, 82, 129, 82, 144, 82, 147, 82, 130, 81, 84, 127, 187, 78, 195, 78, 201, 78, 194, 78, 232, 78, 225, 78, 235, 78, 222, 78, 27, 79, 243, 78, 34, 79, 100, 79, 245, 78, 37, 79, 39, 79, 9, 79, 43, 79, 94, 79, 103, 79, 56, 101, 90, 79, 93, 79, 95, 79, 87, 79, 50, 79, 61, 79, 118, 79, 116, 79, 145, 79, 137, 79, 131, 79, 143, 79, 126, 79, 123, 79, 170, 79, 124, 79, 172, 79, 148, 79, 230, 79, 232, 79, 234, 79, 197, 79, 218, 79, 227, 79, 220, 79, 209, 79, 223, 79, 248, 79, 41, 80, 76, 80, 243, 79, 44, 80, 15, 80, 46, 80, 45, 80, 254, 79, 28, 80, 12, 80, 37, 80, 40, 80, 126, 80, 67, 80, 85, 80, 72, 80, 78, 80, 108, 80, 123, 80, 165, 80, 167, 80, 169, 80, 186, 80, 214, 80, 6, 81, 237, 80, 236, 80, 230, 80, 238, 80, 7, 81, 11, 81, 221, 78, 61, 108, 88, 79, 101, 79, 206, 79, 160, 159, 70, 108, 116, 124, 110, 81, 253, 93, 201, 158, 152, 153, 129, 81, 20, 89, 249, 82, 13, 83, 7, 138, 16, 83, 235, 81, 25, 89, 85, 81, 160, 78, 86, 81, 179, 78, 110, 136, 164, 136, 181, 78, 20, 129, 210, 136, 128, 121, 52, 91, 3, 136, 184, 127, 171, 81, 177, 81, 189, 81, 188, 81, 199, 81, 150, 81, 162, 81, 165, 81, 160, 139, 166, 139, 167, 139, 170, 139, 180, 139, 181, 139, 183, 139, 194, 139, 195, 139, 203, 139, 207, 139, 206, 139, 210, 139, 211, 139, 212, 139, 214, 139, 216, 139, 217, 139, 220, 139, 223, 139, 224, 139, 228, 139, 232, 139, 233, 139, 238, 139, 240, 139, 243, 139, 246, 139, 249, 139, 252, 139, 255, 139, 0, 140, 2, 140, 4, 140, 7, 140, 12, 140, 15, 140, 17, 140, 18, 140, 20, 140, 21, 140, 22, 140, 25, 140, 27, 140, 24, 140, 29, 140, 31, 140, 32, 140, 33, 140, 37, 140, 39, 140, 42, 140, 43, 140, 46, 140, 47, 140, 50, 140, 51, 140, 53, 140, 54, 140, 105, 83, 122, 83, 29, 150, 34, 150, 33, 150, 49, 150, 42, 150, 61, 150, 60, 150, 66, 150, 73, 150, 84, 150, 95, 150, 103, 150, 108, 150, 114, 150, 116, 150, 136, 150, 141, 150, 151, 150, 176, 150, 151, 144, 155, 144, 157, 144, 153, 144, 172, 144, 161, 144, 180, 144, 179, 144, 182, 144, 186, 144, 184, 144, 176, 144, 207, 144, 197, 144, 190, 144, 208, 144, 196, 144, 199, 144, 211, 144, 230, 144, 226, 144, 220, 144, 215, 144, 219, 144, 235, 144, 239, 144, 254, 144, 4, 145, 34, 145, 30, 145, 35, 145, 49, 145, 47, 145, 57, 145, 67, 145, 70, 145, 13, 82, 66, 89, 162, 82, 172, 82, 173, 82, 190, 82, 255, 84, 208, 82, 214, 82, 240, 82, 223, 83, 238, 113, 205, 119, 244, 94, 245, 81, 252, 81, 47, 155, 182, 83, 1, 95, 90, 117, 239, 93, 76, 87, 169, 87, 161, 87, 126, 88, 188, 88, 197, 88, 209, 88, 41, 87, 44, 87, 42, 87, 51, 87, 57, 87, 46, 87, 47, 87, 92, 87, 59, 87, 66, 87, 105, 87, 133, 87, 107, 87, 134, 87, 124, 87, 123, 87, 104, 87, 109, 87, 118, 87, 115, 87, 173, 87, 164, 87, 140, 87, 178, 87, 207, 87, 167, 87, 180, 87, 147, 87, 160, 87, 213, 87, 216, 87, 218, 87, 217, 87, 210, 87, 184, 87, 244, 87, 239, 87, 248, 87, 228, 87, 221, 87, 11, 88, 13, 88, 253, 87, 237, 87, 0, 88, 30, 88, 25, 88, 68, 88, 32, 88, 101, 88, 108, 88, 129, 88, 137, 88, 154, 88, 128, 88, 168, 153, 25, 159, 255, 97, 121, 130, 125, 130, 127, 130, 143, 130, 138, 130, 168, 130, 132, 130, 142, 130, 145, 130, 151, 130, 153, 130, 171, 130, 184, 130, 190, 130, 176, 130, 200, 130, 202, 130, 227, 130, 152, 130, 183, 130, 174, 130, 203, 130, 204, 130, 193, 130, 169, 130, 180, 130, 161, 130, 170, 130, 159, 130, 196, 130, 206, 130, 164, 130, 225, 130, 9, 131, 247, 130, 228, 130, 15, 131, 7, 131, 220, 130, 244, 130, 210, 130, 216, 130, 12, 131, 251, 130, 211, 130, 17, 131, 26, 131, 6, 131, 20, 131, 21, 131, 224, 130, 213, 130, 28, 131, 81, 131, 91, 131, 92, 131, 8, 131, 146, 131, 60, 131, 52, 131, 49, 131, 155, 131, 94, 131, 47, 131, 79, 131, 71, 131, 67, 131, 95, 131, 64, 131, 23, 131, 96, 131, 45, 131, 58, 131, 51, 131, 102, 131, 101, 131, 104, 131, 27, 131, 105, 131, 108, 131, 106, 131, 109, 131, 110, 131, 176, 131, 120, 131, 179, 131, 180, 131, 160, 131, 170, 131, 147, 131, 156, 131, 133, 131, 124, 131, 182, 131, 169, 131, 125, 131, 184, 131, 123, 131, 152, 131, 158, 131, 168, 131, 186, 131, 188, 131, 193, 131, 1, 132, 229, 131, 216, 131, 7, 88, 24, 132, 11, 132, 221, 131, 253, 131, 214, 131, 28, 132, 56, 132, 17, 132, 6, 132, 212, 131, 223, 131, 15, 132, 3, 132, 248, 131, 249, 131, 234, 131, 197, 131, 192, 131, 38, 132, 240, 131, 225, 131, 92, 132, 81, 132, 90, 132, 89, 132, 115, 132, 135, 132, 136, 132, 122, 132, 137, 132, 120, 132, 60, 132, 70, 132, 105, 132, 118, 132, 140, 132, 142, 132, 49, 132, 109, 132, 193, 132, 205, 132, 208, 132, 230, 132, 189, 132, 211, 132, 202, 132, 191, 132, 186, 132, 224, 132, 161, 132, 185, 132, 180, 132, 151, 132, 229, 132, 227, 132, 12, 133, 13, 117, 56, 133, 240, 132, 57, 133, 31, 133, 58, 133, 86, 133, 59, 133, 255, 132, 252, 132, 89, 133, 72, 133, 104, 133, 100, 133, 94, 133, 122, 133, 162, 119, 67, 133, 114, 133, 123, 133, 164, 133, 168, 133, 135, 133, 143, 133, 121, 133, 174, 133, 156, 133, 133, 133, 185, 133, 183, 133, 176, 133, 211, 133, 193, 133, 220, 133, 255, 133, 39, 134, 5, 134, 41, 134, 22, 134, 60, 134, 254, 94, 8, 95, 60, 89, 65, 89, 55, 128, 85, 89, 90, 89, 88, 89, 15, 83, 34, 92, 37, 92, 44, 92, 52, 92, 76, 98, 106, 98, 159, 98, 187, 98, 202, 98, 218, 98, 215, 98, 238, 98, 34, 99, 246, 98, 57, 99, 75, 99, 67, 99, 173, 99, 246, 99, 113, 99, 122, 99, 142, 99, 180, 99, 109, 99, 172, 99, 138, 99, 105, 99, 174, 99, 188, 99, 242, 99, 248, 99, 224, 99, 255, 99, 196, 99, 222, 99, 206, 99, 82, 100, 198, 99, 190, 99, 69, 100, 65, 100, 11, 100, 27, 100, 32, 100, 12, 100, 38, 100, 33, 100, 94, 100, 132, 100, 109, 100, 150, 100, 122, 100, 183, 100, 184, 100, 153, 100, 186, 100, 192, 100, 208, 100, 215, 100, 228, 100, 226, 100, 9, 101, 37, 101, 46, 101, 11, 95, 210, 95, 25, 117, 17, 95, 95, 83, 241, 83, 253, 83, 233, 83, 232, 83, 251, 83, 18, 84, 22, 84, 6, 84, 75, 84, 82, 84, 83, 84, 84, 84, 86, 84, 67, 84, 33, 84, 87, 84, 89, 84, 35, 84, 50, 84, 130, 84, 148, 84, 119, 84, 113, 84, 100, 84, 154, 84, 155, 84, 132, 84, 118, 84, 102, 84, 157, 84, 208, 84, 173, 84, 194, 84, 180, 84, 210, 84, 167, 84, 166, 84, 211, 84, 212, 84, 114, 84, 163, 84, 213, 84, 187, 84, 191, 84, 204, 84, 217, 84, 218, 84, 220, 84, 169, 84, 170, 84, 164, 84, 221, 84, 207, 84, 222, 84, 27, 85, 231, 84, 32, 85, 253, 84, 20, 85, 243, 84, 34, 85, 35, 85, 15, 85, 17, 85, 39, 85, 42, 85, 103, 85, 143, 85, 181, 85, 73, 85, 109, 85, 65, 85, 85, 85, 63, 85, 80, 85, 60, 85, 55, 85, 86, 85, 117, 85, 118, 85, 119, 85, 51, 85, 48, 85, 92, 85, 139, 85, 210, 85, 131, 85, 177, 85, 185, 85, 136, 85, 129, 85, 159, 85, 126, 85, 214, 85, 145, 85, 123, 85, 223, 85, 189, 85, 190, 85, 148, 85, 153, 85, 234, 85, 247, 85, 201, 85, 31, 86, 209, 85, 235, 85, 236, 85, 212, 85, 230, 85, 221, 85, 196, 85, 239, 85, 229, 85, 242, 85, 243, 85, 204, 85, 205, 85, 232, 85, 245, 85, 228, 85, 148, 143, 30, 86, 8, 86, 12, 86, 1, 86, 36, 86, 35, 86, 254, 85, 0, 86, 39, 86, 45, 86, 88, 86, 57, 86, 87, 86, 44, 86, 77, 86, 98, 86, 89, 86, 92, 86, 76, 86, 84, 86, 134, 86, 100, 86, 113, 86, 107, 86, 123, 86, 124, 86, 133, 86, 147, 86, 175, 86, 212, 86, 215, 86, 221, 86, 225, 86, 245, 86, 235, 86, 249, 86, 255, 86, 4, 87, 10, 87, 9, 87, 28, 87, 15, 94, 25, 94, 20, 94, 17, 94, 49, 94, 59, 94, 60, 94, 55, 94, 68, 94, 84, 94, 91, 94, 94, 94, 97, 94, 140, 92, 122, 92, 141, 92, 144, 92, 150, 92, 136, 92, 152, 92, 153, 92, 145, 92, 154, 92, 156, 92, 181, 92, 162, 92, 189, 92, 172, 92, 171, 92, 177, 92, 163, 92, 193, 92, 183, 92, 196, 92, 210, 92, 228, 92, 203, 92, 229, 92, 2, 93, 3, 93, 39, 93, 38, 93, 46, 93, 36, 93, 30, 93, 6, 93, 27, 93, 88, 93, 62, 93, 52, 93, 61, 93, 108, 93, 91, 93, 111, 93, 93, 93, 107, 93, 75, 93, 74, 93, 105, 93, 116, 93, 130, 93, 153, 93, 157, 93, 115, 140, 183, 93, 197, 93, 115, 95, 119, 95, 130, 95, 135, 95, 137, 95, 140, 95, 149, 95, 153, 95, 156, 95, 168, 95, 173, 95, 181, 95, 188, 95, 98, 136, 97, 95, 173, 114, 176, 114, 180, 114, 183, 114, 184, 114, 195, 114, 193, 114, 206, 114, 205, 114, 210, 114, 232, 114, 239, 114, 233, 114, 242, 114, 244, 114, 247, 114, 1, 115, 243, 114, 3, 115, 250, 114, 251, 114, 23, 115, 19, 115, 33, 115, 10, 115, 30, 115, 29, 115, 21, 115, 34, 115, 57, 115, 37, 115, 44, 115, 56, 115, 49, 115, 80, 115, 77, 115, 87, 115, 96, 115, 108, 115, 111, 115, 126, 115, 27, 130, 37, 89, 231, 152, 36, 89, 2, 89, 99, 153, 103, 153, 104, 153, 105, 153, 106, 153, 107, 153, 108, 153, 116, 153, 119, 153, 125, 153, 128, 153, 132, 153, 135, 153, 138, 153, 141, 153, 144, 153, 145, 153, 147, 153, 148, 153, 149, 153, 128, 94, 145, 94, 139, 94, 150, 94, 165, 94, 160, 94, 185, 94, 181, 94, 190, 94, 179, 94, 83, 141, 210, 94, 209, 94, 219, 94, 232, 94, 234, 94, 186, 129, 196, 95, 201, 95, 214, 95, 207, 95, 3, 96, 238, 95, 4, 96, 225, 95, 228, 95, 254, 95, 5, 96, 6, 96, 234, 95, 237, 95, 248, 95, 25, 96, 53, 96, 38, 96, 27, 96, 15, 96, 13, 96, 41, 96, 43, 96, 10, 96, 63, 96, 33, 96, 120, 96, 121, 96, 123, 96, 122, 96, 66, 96, 106, 96, 125, 96, 150, 96, 154, 96, 173, 96, 157, 96, 131, 96, 146, 96, 140, 96, 155, 96, 236, 96, 187, 96, 177, 96, 221, 96, 216, 96, 198, 96, 218, 96, 180, 96, 32, 97, 38, 97, 21, 97, 35, 97, 244, 96, 0, 97, 14, 97, 43, 97, 74, 97, 117, 97, 172, 97, 148, 97, 167, 97, 183, 97, 212, 97, 245, 97, 221, 95, 179, 150, 233, 149, 235, 149, 241, 149, 243, 149, 245, 149, 246, 149, 252, 149, 254, 149, 3, 150, 4, 150, 6, 150, 8, 150, 10, 150, 11, 150, 12, 150, 13, 150, 15, 150, 18, 150, 21, 150, 22, 150, 23, 150, 25, 150, 26, 150, 44, 78, 63, 114, 21, 98, 53, 108, 84, 108, 92, 108, 74, 108, 163, 108, 133, 108, 144, 108, 148, 108, 140, 108, 104, 108, 105, 108, 116, 108, 118, 108, 134, 108, 169, 108, 208, 108, 212, 108, 173, 108, 247, 108, 248, 108, 241, 108, 215, 108, 178, 108, 224, 108, 214, 108, 250, 108, 235, 108, 238, 108, 177, 108, 211, 108, 239, 108, 254, 108, 57, 109, 39, 109, 12, 109, 67, 109, 72, 109, 7, 109, 4, 109, 25, 109, 14, 109, 43, 109, 77, 109, 46, 109, 53, 109, 26, 109, 79, 109, 82, 109, 84, 109, 51, 109, 145, 109, 111, 109, 158, 109, 160, 109, 94, 109, 147, 109, 148, 109, 92, 109, 96, 109, 124, 109, 99, 109, 26, 110, 199, 109, 197, 109, 222, 109, 14, 110, 191, 109, 224, 109, 17, 110, 230, 109, 221, 109, 217, 109, 22, 110, 171, 109, 12, 110, 174, 109, 43, 110, 110, 110, 78, 110, 107, 110, 178, 110, 95, 110, 134, 110, 83, 110, 84, 110, 50, 110, 37, 110, 68, 110, 223, 110, 177, 110, 152, 110, 224, 110, 45, 111, 226, 110, 165, 110, 167, 110, 189, 110, 187, 110, 183, 110, 215, 110, 180, 110, 207, 110, 143, 110, 194, 110, 159, 110, 98, 111, 70, 111, 71, 111, 36, 111, 21, 111, 249, 110, 47, 111, 54, 111, 75, 111, 116, 111, 42, 111, 9, 111, 41, 111, 137, 111, 141, 111, 140, 111, 120, 111, 114, 111, 124, 111, 122, 111, 209, 111, 201, 111, 167, 111, 185, 111, 182, 111, 194, 111, 225, 111, 238, 111, 222, 111, 224, 111, 239, 111, 26, 112, 35, 112, 27, 112, 57, 112, 53, 112, 79, 112, 94, 112, 128, 91, 132, 91, 149, 91, 147, 91, 165, 91, 184, 91, 47, 117, 158, 154, 52, 100, 228, 91, 238, 91, 48, 137, 240, 91, 71, 142, 7, 139, 182, 143, 211, 143, 213, 143, 229, 143, 238, 143, 228, 143, 233, 143, 230, 143, 243, 143, 232, 143, 5, 144, 4, 144, 11, 144, 38, 144, 17, 144, 13, 144, 22, 144, 33, 144, 53, 144, 54, 144, 45, 144, 47, 144, 68, 144, 81, 144, 82, 144, 80, 144, 104, 144, 88, 144, 98, 144, 91, 144, 185, 102, 116, 144, 125, 144, 130, 144, 136, 144, 131, 144, 139, 144, 80, 95, 87, 95, 86, 95, 88, 95, 59, 92, 171, 84, 80, 92, 89, 92, 113, 91, 99, 92, 102, 92, 188, 127, 42, 95, 41, 95, 45, 95, 116, 130, 60, 95, 59, 155, 110, 92, 129, 89, 131, 89, 141, 89, 169, 89, 170, 89, 163, 89, 151, 89, 202, 89, 171, 89, 158, 89, 164, 89, 210, 89, 178, 89, 175, 89, 215, 89, 190, 89, 5, 90, 6, 90, 221, 89, 8, 90, 227, 89, 216, 89, 249, 89, 12, 90, 9, 90, 50, 90, 52, 90, 17, 90, 35, 90, 19, 90, 64, 90, 103, 90, 74, 90, 85, 90, 60, 90, 98, 90, 117, 90, 236, 128, 170, 90, 155, 90, 119, 90, 122, 90, 190, 90, 235, 90, 178, 90, 210, 90, 212, 90, 184, 90, 224, 90, 227, 90, 241, 90, 214, 90, 230, 90, 216, 90, 220, 90, 9, 91, 23, 91, 22, 91, 50, 91, 55, 91, 64, 91, 21, 92, 28, 92, 90, 91, 101, 91, 115, 91, 81, 91, 83, 91, 98, 91, 117, 154, 119, 154, 120, 154, 122, 154, 127, 154, 125, 154, 128, 154, 129, 154, 133, 154, 136, 154, 138, 154, 144, 154, 146, 154, 147, 154, 150, 154, 152, 154, 155, 154, 156, 154, 157, 154, 159, 154, 160, 154, 162, 154, 163, 154, 165, 154, 167, 154, 159, 126, 161, 126, 163, 126, 165, 126, 168, 126, 169, 126, 173, 126, 176, 126, 190, 126, 192, 126, 193, 126, 194, 126, 201, 126, 203, 126, 204, 126, 208, 126, 212, 126, 215, 126, 219, 126, 224, 126, 225, 126, 232, 126, 235, 126, 238, 126, 239, 126, 241, 126, 242, 126, 13, 127, 246, 126, 250, 126, 251, 126, 254, 126, 1, 127, 2, 127, 3, 127, 7, 127, 8, 127, 11, 127, 12, 127, 15, 127, 17, 127, 18, 127, 23, 127, 25, 127, 28, 127, 27, 127, 31, 127, 33, 127, 34, 127, 35, 127, 36, 127, 37, 127, 38, 127, 39, 127, 42, 127, 43, 127, 44, 127, 45, 127, 47, 127, 48, 127, 49, 127, 50, 127, 51, 127, 53, 127, 122, 94, 127, 117, 219, 93, 62, 117, 149, 144, 142, 115, 145, 115, 174, 115, 162, 115, 159, 115, 207, 115, 194, 115, 209, 115, 183, 115, 179, 115, 192, 115, 201, 115, 200, 115, 229, 115, 217, 115, 124, 152, 10, 116, 233, 115, 231, 115, 222, 115, 186, 115, 242, 115, 15, 116, 42, 116, 91, 116, 38, 116, 37, 116, 40, 116, 48, 116, 46, 116, 44, 116, 27, 116, 26, 116, 65, 116, 92, 116, 87, 116, 85, 116, 89, 116, 119, 116, 109, 116, 126, 116, 156, 116, 142, 116, 128, 116, 129, 116, 135, 116, 139, 116, 158, 116, 168, 116, 169, 116, 144, 116, 167, 116, 210, 116, 186, 116, 234, 151, 235, 151, 236, 151, 76, 103, 83, 103, 94, 103, 72, 103, 105, 103, 165, 103, 135, 103, 106, 103, 115, 103, 152, 103, 167, 103, 117, 103, 168, 103, 158, 103, 173, 103, 139, 103, 119, 103, 124, 103, 240, 103, 9, 104, 216, 103, 10, 104, 233, 103, 176, 103, 12, 104, 217, 103, 181, 103, 218, 103, 179, 103, 221, 103, 0, 104, 195, 103, 184, 103, 226, 103, 14, 104, 193, 103, 253, 103, 50, 104, 51, 104, 96, 104, 97, 104, 78, 104, 98, 104, 68, 104, 100, 104, 131, 104, 29, 104, 85, 104, 102, 104, 65, 104, 103, 104, 64, 104, 62, 104, 74, 104, 73, 104, 41, 104, 181, 104, 143, 104, 116, 104, 119, 104, 147, 104, 107, 104, 194, 104, 110, 105, 252, 104, 31, 105, 32, 105, 249, 104, 36, 105, 240, 104, 11, 105, 1, 105, 87, 105, 227, 104, 16, 105, 113, 105, 57, 105, 96, 105, 66, 105, 93, 105, 132, 105, 107, 105, 128, 105, 152, 105, 120, 105, 52, 105, 204, 105, 135, 105, 136, 105, 206, 105, 137, 105, 102, 105, 99, 105, 121, 105, 155, 105, 167, 105, 187, 105, 171, 105, 173, 105, 212, 105, 177, 105, 193, 105, 202, 105, 223, 105, 149, 105, 224, 105, 141, 105, 255, 105, 47, 106, 237, 105, 23, 106, 24, 106, 101, 106, 242, 105, 68, 106, 62, 106, 160, 106, 80, 106, 91, 106, 53, 106, 142, 106, 121, 106, 61, 106, 40, 106, 88, 106, 124, 106, 145, 106, 144, 106, 169, 106, 151, 106, 171, 106, 55, 115, 82, 115, 129, 107, 130, 107, 135, 107, 132, 107, 146, 107, 147, 107, 141, 107, 154, 107, 155, 107, 161, 107, 170, 107, 107, 143, 109, 143, 113, 143, 114, 143, 115, 143, 117, 143, 118, 143, 120, 143, 119, 143, 121, 143, 122, 143, 124, 143, 126, 143, 129, 143, 130, 143, 132, 143, 135, 143, 139, 143, 141, 143, 142, 143, 143, 143, 152, 143, 154, 143, 206, 142, 11, 98, 23, 98, 27, 98, 31, 98, 34, 98, 33, 98, 37, 98, 36, 98, 44, 98, 231, 129, 239, 116, 244, 116, 255, 116, 15, 117, 17, 117, 19, 117, 52, 101, 238, 101, 239, 101, 240, 101, 10, 102, 25, 102, 114, 103, 3, 102, 21, 102, 0, 102, 133, 112, 247, 102, 29, 102, 52, 102, 49, 102, 54, 102, 53, 102, 6, 128, 95, 102, 84, 102, 65, 102, 79, 102, 86, 102, 97, 102, 87, 102, 119, 102, 132, 102, 140, 102, 167, 102, 157, 102, 190, 102, 219, 102, 220, 102, 230, 102, 233, 102, 50, 141, 51, 141, 54, 141, 59, 141, 61, 141, 64, 141, 69, 141, 70, 141, 72, 141, 73, 141, 71, 141, 77, 141, 85, 141, 89, 141, 199, 137, 202, 137, 203, 137, 204, 137, 206, 137, 207, 137, 208, 137, 209, 137, 110, 114, 159, 114, 93, 114, 102, 114, 111, 114, 126, 114, 127, 114, 132, 114, 139, 114, 141, 114, 143, 114, 146, 114, 8, 99, 50, 99, 176, 99, 63, 100, 216, 100, 4, 128, 234, 107, 243, 107, 253, 107, 245, 107, 249, 107, 5, 108, 7, 108, 6, 108, 13, 108, 21, 108, 24, 108, 25, 108, 26, 108, 33, 108, 41, 108, 36, 108, 42, 108, 50, 108, 53, 101, 85, 101, 107, 101, 77, 114, 82, 114, 86, 114, 48, 114, 98, 134, 22, 82, 159, 128, 156, 128, 147, 128, 188, 128, 10, 103, 189, 128, 177, 128, 171, 128, 173, 128, 180, 128, 183, 128, 231, 128, 232, 128, 233, 128, 234, 128, 219, 128, 194, 128, 196, 128, 217, 128, 205, 128, 215, 128, 16, 103, 221, 128, 235, 128, 241, 128, 244, 128, 237, 128, 13, 129, 14, 129, 242, 128, 252, 128, 21, 103, 18, 129, 90, 140, 54, 129, 30, 129, 44, 129, 24, 129, 50, 129, 72, 129, 76, 129, 83, 129, 116, 129, 89, 129, 90, 129, 113, 129, 96, 129, 105, 129, 124, 129, 125, 129, 109, 129, 103, 129, 77, 88, 181, 90, 136, 129, 130, 129, 145, 129, 213, 110, 163, 129, 170, 129, 204, 129, 38, 103, 202, 129, 187, 129, 193, 129, 166, 129, 36, 107, 55, 107, 57, 107, 67, 107, 70, 107, 89, 107, 209, 152, 210, 152, 211, 152, 213, 152, 217, 152, 218, 152, 179, 107, 64, 95, 194, 107, 243, 137, 144, 101, 81, 159, 147, 101, 188, 101, 198, 101, 196, 101, 195, 101, 204, 101, 206, 101, 210, 101, 214, 101, 128, 112, 156, 112, 150, 112, 157, 112, 187, 112, 192, 112, 183, 112, 171, 112, 177, 112, 232, 112, 202, 112, 16, 113, 19, 113, 22, 113, 47, 113, 49, 113, 115, 113, 92, 113, 104, 113, 69, 113, 114, 113, 74, 113, 120, 113, 122, 113, 152, 113, 179, 113, 181, 113, 168, 113, 160, 113, 224, 113, 212, 113, 231, 113, 249, 113, 29, 114, 40, 114, 108, 112, 24, 113, 102, 113, 185, 113, 62, 98, 61, 98, 67, 98, 72, 98, 73, 98, 59, 121, 64, 121, 70, 121, 73, 121, 91, 121, 92, 121, 83, 121, 90, 121, 98, 121, 87, 121, 96, 121, 111, 121, 103, 121, 122, 121, 133, 121, 138, 121, 154, 121, 167, 121, 179, 121, 209, 95, 208, 95, 60, 96, 93, 96, 90, 96, 103, 96, 65, 96, 89, 96, 99, 96, 171, 96, 6, 97, 13, 97, 93, 97, 169, 97, 157, 97, 203, 97, 209, 97, 6, 98, 128, 128, 127, 128, 147, 108, 246, 108, 252, 109, 246, 119, 248, 119, 0, 120, 9, 120, 23, 120, 24, 120, 17, 120, 171, 101, 45, 120, 28, 120, 29, 120, 57, 120, 58, 120, 59, 120, 31, 120, 60, 120, 37, 120, 44, 120, 35, 120, 41, 120, 78, 120, 109, 120, 86, 120, 87, 120, 38, 120, 80, 120, 71, 120, 76, 120, 106, 120, 155, 120, 147, 120, 154, 120, 135, 120, 156, 120, 161, 120, 163, 120, 178, 120, 185, 120, 165, 120, 212, 120, 217, 120, 201, 120, 236, 120, 242, 120, 5, 121, 244, 120, 19, 121, 36, 121, 30, 121, 52, 121, 155, 159, 249, 158, 251, 158, 252, 158, 241, 118, 4, 119, 13, 119, 249, 118, 7, 119, 8, 119, 26, 119, 34, 119, 25, 119, 45, 119, 38, 119, 53, 119, 56, 119, 80, 119, 81, 119, 71, 119, 67, 119, 90, 119, 104, 119, 98, 119, 101, 119, 127, 119, 141, 119, 125, 119, 128, 119, 140, 119, 145, 119, 159, 119, 160, 119, 176, 119, 181, 119, 189, 119, 58, 117, 64, 117, 78, 117, 75, 117, 72, 117, 91, 117, 114, 117, 121, 117, 131, 117, 88, 127, 97, 127, 95, 127, 72, 138, 104, 127, 116, 127, 113, 127, 121, 127, 129, 127, 126, 127, 205, 118, 229, 118, 50, 136, 133, 148, 134, 148, 135, 148, 139, 148, 138, 148, 140, 148, 141, 148, 143, 148, 144, 148, 148, 148, 151, 148, 149, 148, 154, 148, 155, 148, 156, 148, 163, 148, 164, 148, 171, 148, 170, 148, 173, 148, 172, 148, 175, 148, 176, 148, 178, 148, 180, 148, 182, 148, 183, 148, 184, 148, 185, 148, 186, 148, 188, 148, 189, 148, 191, 148, 196, 148, 200, 148, 201, 148, 202, 148, 203, 148, 204, 148, 205, 148, 206, 148, 208, 148, 209, 148, 210, 148, 213, 148, 214, 148, 215, 148, 217, 148, 216, 148, 219, 148, 222, 148, 223, 148, 224, 148, 226, 148, 228, 148, 229, 148, 231, 148, 232, 148, 234, 148, 233, 148, 235, 148, 238, 148, 239, 148, 243, 148, 244, 148, 245, 148, 247, 148, 249, 148, 252, 148, 253, 148, 255, 148, 3, 149, 2, 149, 6, 149, 7, 149, 9, 149, 10, 149, 13, 149, 14, 149, 15, 149, 18, 149, 19, 149, 20, 149, 21, 149, 22, 149, 24, 149, 27, 149, 29, 149, 30, 149, 31, 149, 34, 149, 42, 149, 43, 149, 41, 149, 44, 149, 49, 149, 50, 149, 52, 149, 54, 149, 55, 149, 56, 149, 60, 149, 62, 149, 63, 149, 66, 149, 53, 149, 68, 149, 69, 149, 70, 149, 73, 149, 76, 149, 78, 149, 79, 149, 82, 149, 83, 149, 84, 149, 86, 149, 87, 149, 88, 149, 89, 149, 91, 149, 94, 149, 95, 149, 93, 149, 97, 149, 98, 149, 100, 149, 101, 149, 102, 149, 103, 149, 104, 149, 105, 149, 106, 149, 107, 149, 108, 149, 111, 149, 113, 149, 114, 149, 115, 149, 58, 149, 231, 119, 236, 119, 201, 150, 213, 121, 237, 121, 227, 121, 235, 121, 6, 122, 71, 93, 3, 122, 2, 122, 30, 122, 20, 122, 57, 122, 55, 122, 81, 122, 207, 158, 165, 153, 112, 122, 136, 118, 142, 118, 147, 118, 153, 118, 164, 118, 222, 116, 224, 116, 44, 117, 32, 158, 34, 158, 40, 158, 41, 158, 42, 158, 43, 158, 44, 158, 50, 158, 49, 158, 54, 158, 56, 158, 55, 158, 57, 158, 58, 158, 62, 158, 65, 158, 66, 158, 68, 158, 70, 158, 71, 158, 72, 158, 73, 158, 75, 158, 76, 158, 78, 158, 81, 158, 85, 158, 87, 158, 90, 158, 91, 158, 92, 158, 94, 158, 99, 158, 102, 158, 103, 158, 104, 158, 105, 158, 106, 158, 107, 158, 108, 158, 113, 158, 109, 158, 115, 158, 146, 117, 148, 117, 150, 117, 160, 117, 157, 117, 172, 117, 163, 117, 179, 117, 180, 117, 184, 117, 196, 117, 177, 117, 176, 117, 195, 117, 194, 117, 214, 117, 205, 117, 227, 117, 232, 117, 230, 117, 228, 117, 235, 117, 231, 117, 3, 118, 241, 117, 252, 117, 255, 117, 16, 118, 0, 118, 5, 118, 12, 118, 23, 118, 10, 118, 37, 118, 24, 118, 21, 118, 25, 118, 27, 118, 60, 118, 34, 118, 32, 118, 64, 118, 45, 118, 48, 118, 63, 118, 53, 118, 67, 118, 62, 118, 51, 118, 77, 118, 94, 118, 84, 118, 92, 118, 86, 118, 107, 118, 111, 118, 202, 127, 230, 122, 120, 122, 121, 122, 128, 122, 134, 122, 136, 122, 149, 122, 166, 122, 160, 122, 172, 122, 168, 122, 173, 122, 179, 122, 100, 136, 105, 136, 114, 136, 125, 136, 127, 136, 130, 136, 162, 136, 198, 136, 183, 136, 188, 136, 201, 136, 226, 136, 206, 136, 227, 136, 229, 136, 241, 136, 26, 137, 252, 136, 232, 136, 254, 136, 240, 136, 33, 137, 25, 137, 19, 137, 27, 137, 10, 137, 52, 137, 43, 137, 54, 137, 65, 137, 102, 137, 123, 137, 139, 117, 229, 128, 178, 118, 180, 118, 220, 119, 18, 128, 20, 128, 22, 128, 28, 128, 32, 128, 34, 128, 37, 128, 38, 128, 39, 128, 41, 128, 40, 128, 49, 128, 11, 128, 53, 128, 67, 128, 70, 128, 77, 128, 82, 128, 105, 128, 113, 128, 131, 137, 120, 152, 128, 152, 131, 152, 137, 152, 140, 152, 141, 152, 143, 152, 148, 152, 154, 152, 155, 152, 158, 152, 159, 152, 161, 152, 162, 152, 165, 152, 166, 152, 77, 134, 84, 134, 108, 134, 110, 134, 127, 134, 122, 134, 124, 134, 123, 134, 168, 134, 141, 134, 139, 134, 172, 134, 157, 134, 167, 134, 163, 134, 170, 134, 147, 134, 169, 134, 182, 134, 196, 134, 181, 134, 206, 134, 176, 134, 186, 134, 177, 134, 175, 134, 201, 134, 207, 134, 180, 134, 233, 134, 241, 134, 242, 134, 237, 134, 243, 134, 208, 134, 19, 135, 222, 134, 244, 134, 223, 134, 216, 134, 209, 134, 3, 135, 7, 135, 248, 134, 8, 135, 10, 135, 13, 135, 9, 135, 35, 135, 59, 135, 30, 135, 37, 135, 46, 135, 26, 135, 62, 135, 72, 135, 52, 135, 49, 135, 41, 135, 55, 135, 63, 135, 130, 135, 34, 135, 125, 135, 126, 135, 123, 135, 96, 135, 112, 135, 76, 135, 110, 135, 139, 135, 83, 135, 99, 135, 124, 135, 100, 135, 89, 135, 101, 135, 147, 135, 175, 135, 168, 135, 210, 135, 198, 135, 136, 135, 133, 135, 173, 135, 151, 135, 131, 135, 171, 135, 229, 135, 172, 135, 181, 135, 179, 135, 203, 135, 211, 135, 189, 135, 209, 135, 192, 135, 202, 135, 219, 135, 234, 135, 224, 135, 238, 135, 22, 136, 19, 136, 254, 135, 10, 136, 27, 136, 33, 136, 57, 136, 60, 136, 54, 127, 66, 127, 68, 127, 69, 127, 16, 130, 250, 122, 253, 122, 8, 123, 3, 123, 4, 123, 21, 123, 10, 123, 43, 123, 15, 123, 71, 123, 56, 123, 42, 123, 25, 123, 46, 123, 49, 123, 32, 123, 37, 123, 36, 123, 51, 123, 62, 123, 30, 123, 88, 123, 90, 123, 69, 123, 117, 123, 76, 123, 93, 123, 96, 123, 110, 123, 123, 123, 98, 123, 114, 123, 113, 123, 144, 123, 166, 123, 167, 123, 184, 123, 172, 123, 157, 123, 168, 123, 133, 123, 170, 123, 156, 123, 162, 123, 171, 123, 180, 123, 209, 123, 193, 123, 204, 123, 221, 123, 218, 123, 229, 123, 230, 123, 234, 123, 12, 124, 254, 123, 252, 123, 15, 124, 22, 124, 11, 124, 31, 124, 42, 124, 38, 124, 56, 124, 65, 124, 64, 124, 254, 129, 1, 130, 2, 130, 4, 130, 236, 129, 68, 136, 33, 130, 34, 130, 35, 130, 45, 130, 47, 130, 40, 130, 43, 130, 56, 130, 59, 130, 51, 130, 52, 130, 62, 130, 68, 130, 73, 130, 75, 130, 79, 130, 90, 130, 95, 130, 104, 130, 126, 136, 133, 136, 136, 136, 216, 136, 223, 136, 94, 137, 157, 127, 159, 127, 167, 127, 175, 127, 176, 127, 178, 127, 124, 124, 73, 101, 145, 124, 157, 124, 156, 124, 158, 124, 162, 124, 178, 124, 188, 124, 189, 124, 193, 124, 199, 124, 204, 124, 205, 124, 200, 124, 197, 124, 215, 124, 232, 124, 110, 130, 168, 102, 191, 127, 206, 127, 213, 127, 229, 127, 225, 127, 230, 127, 233, 127, 238, 127, 243, 127, 248, 124, 119, 125, 166, 125, 174, 125, 71, 126, 155, 126, 184, 158, 180, 158, 115, 141, 132, 141, 148, 141, 145, 141, 177, 141, 103, 141, 109, 141, 71, 140, 73, 140, 74, 145, 80, 145, 78, 145, 79, 145, 100, 145, 98, 145, 97, 145, 112, 145, 105, 145, 111, 145, 125, 145, 126, 145, 114, 145, 116, 145, 121, 145, 140, 145, 133, 145, 144, 145, 141, 145, 145, 145, 162, 145, 163, 145, 170, 145, 173, 145, 174, 145, 175, 145, 181, 145, 180, 145, 186, 145, 85, 140, 126, 158, 184, 141, 235, 141, 5, 142, 89, 142, 105, 142, 181, 141, 191, 141, 188, 141, 186, 141, 196, 141, 214, 141, 215, 141, 218, 141, 222, 141, 206, 141, 207, 141, 219, 141, 198, 141, 236, 141, 247, 141, 248, 141, 227, 141, 249, 141, 251, 141, 228, 141, 9, 142, 253, 141, 20, 142, 29, 142, 31, 142, 44, 142, 46, 142, 35, 142, 47, 142, 58, 142, 64, 142, 57, 142, 53, 142, 61, 142, 49, 142, 73, 142, 65, 142, 66, 142, 81, 142, 82, 142, 74, 142, 112, 142, 118, 142, 124, 142, 111, 142, 116, 142, 133, 142, 143, 142, 148, 142, 144, 142, 156, 142, 158, 142, 120, 140, 130, 140, 138, 140, 133, 140, 152, 140, 148, 140, 155, 101, 214, 137, 222, 137, 218, 137, 220, 137, 229, 137, 235, 137, 239, 137, 62, 138, 38, 139, 83, 151, 233, 150, 243, 150, 239, 150, 6, 151, 1, 151, 8, 151, 15, 151, 14, 151, 42, 151, 45, 151, 48, 151, 62, 151, 128, 159, 131, 159, 133, 159, 134, 159, 135, 159, 136, 159, 137, 159, 138, 159, 140, 159, 254, 158, 11, 159, 13, 159, 185, 150, 188, 150, 189, 150, 206, 150, 210, 150, 191, 119, 224, 150, 142, 146, 174, 146, 200, 146, 62, 147, 106, 147, 202, 147, 143, 147, 62, 148, 107, 148, 127, 156, 130, 156, 133, 156, 134, 156, 135, 156, 136, 156, 35, 122, 139, 156, 142, 156, 144, 156, 145, 156, 146, 156, 148, 156, 149, 156, 154, 156, 155, 156, 158, 156, 159, 156, 160, 156, 161, 156, 162, 156, 163, 156, 165, 156, 166, 156, 167, 156, 168, 156, 169, 156, 171, 156, 173, 156, 174, 156, 176, 156, 177, 156, 178, 156, 179, 156, 180, 156, 181, 156, 182, 156, 183, 156, 186, 156, 187, 156, 188, 156, 189, 156, 196, 156, 197, 156, 198, 156, 199, 156, 202, 156, 203, 156, 204, 156, 205, 156, 206, 156, 207, 156, 208, 156, 211, 156, 212, 156, 213, 156, 215, 156, 216, 156, 217, 156, 220, 156, 221, 156, 223, 156, 226, 156, 124, 151, 133, 151, 145, 151, 146, 151, 148, 151, 175, 151, 171, 151, 163, 151, 178, 151, 180, 151, 177, 154, 176, 154, 183, 154, 88, 158, 182, 154, 186, 154, 188, 154, 193, 154, 192, 154, 197, 154, 194, 154, 203, 154, 204, 154, 209, 154, 69, 155, 67, 155, 71, 155, 73, 155, 72, 155, 77, 155, 81, 155, 232, 152, 13, 153, 46, 153, 85, 153, 84, 153, 223, 154, 225, 154, 230, 154, 239, 154, 235, 154, 251, 154, 237, 154, 249, 154, 8, 155, 15, 155, 19, 155, 31, 155, 35, 155, 189, 158, 190, 158, 59, 126, 130, 158, 135, 158, 136, 158, 139, 158, 146, 158, 214, 147, 157, 158, 159, 158, 219, 158, 220, 158, 221, 158, 224, 158, 223, 158, 226, 158, 233, 158, 231, 158, 229, 158, 234, 158, 239, 158, 34, 159, 44, 159, 47, 159, 57, 159, 55, 159, 61, 159, 62, 159, 68, 159, 0, 0, 239, 223, 76, 51, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 122, 0, 104, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 112, 0, 105, 0, 110, 0, 121, 0, 105, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 76, 0, 97, 0, 116, 0, 110, 0, 32, 0, 72, 0, 97, 0, 110, 0, 105, 0, 32, 0, 66, 0, 111, 0, 112, 0, 111, 0, 93, 0, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 114, 0, 101, 0, 103, 0, 117, 0, 108, 0, 97, 0, 114, 0, 93, 0, 60, 0, 42, 0, 89, 81, 91, 81, 94, 81, 93, 81, 97, 81, 99, 81, 231, 85, 233, 116, 206, 124, 0, 78, 89, 78, 1, 78, 3, 78, 67, 78, 93, 78, 134, 78, 140, 78, 186, 78, 63, 81, 101, 81, 107, 81, 224, 81, 0, 82, 1, 82, 155, 82, 21, 83, 65, 83, 92, 83, 200, 83, 9, 78, 11, 78, 8, 78, 10, 78, 43, 78, 56, 78, 225, 81, 69, 78, 72, 78, 95, 78, 94, 78, 142, 78, 161, 78, 64, 81, 3, 82, 250, 82, 67, 83, 201, 83, 227, 83, 31, 87, 235, 88, 21, 89, 39, 89, 115, 89, 80, 91, 81, 91, 83, 91, 248, 91, 15, 92, 34, 92, 56, 92, 113, 92, 221, 93, 229, 93, 241, 93, 242, 93, 243, 93, 254, 93, 114, 94, 254, 94, 11, 95, 19, 95, 77, 98, 17, 78, 16, 78, 13, 78, 45, 78, 48, 78, 57, 78, 75, 78, 57, 92, 136, 78, 145, 78, 149, 78, 146, 78, 148, 78, 162, 78, 193, 78, 192, 78, 195, 78, 198, 78, 199, 78, 205, 78, 202, 78, 203, 78, 196, 78, 67, 81, 65, 81, 103, 81, 109, 81, 110, 81, 108, 81, 151, 81, 246, 81, 6, 82, 7, 82, 8, 82, 251, 82, 254, 82, 255, 82, 22, 83, 57, 83, 72, 83, 71, 83, 69, 83, 94, 83, 132, 83, 203, 83, 202, 83, 205, 83, 236, 88, 41, 89, 43, 89, 42, 89, 45, 89, 84, 91, 17, 92, 36, 92, 58, 92, 111, 92, 244, 93, 123, 94, 255, 94, 20, 95, 21, 95, 195, 95, 8, 98, 54, 98, 75, 98, 78, 98, 47, 101, 135, 101, 151, 101, 164, 101, 185, 101, 229, 101, 240, 102, 8, 103, 40, 103, 32, 107, 98, 107, 121, 107, 203, 107, 212, 107, 219, 107, 15, 108, 52, 108, 107, 112, 42, 114, 54, 114, 59, 114, 71, 114, 89, 114, 91, 114, 172, 114, 139, 115, 25, 78, 22, 78, 21, 78, 20, 78, 24, 78, 59, 78, 77, 78, 79, 78, 78, 78, 229, 78, 216, 78, 212, 78, 213, 78, 214, 78, 215, 78, 227, 78, 228, 78, 217, 78, 222, 78, 69, 81, 68, 81, 137, 81, 138, 81, 172, 81, 249, 81, 250, 81, 248, 81, 10, 82, 160, 82, 159, 82, 5, 83, 6, 83, 23, 83, 29, 83, 223, 78, 74, 83, 73, 83, 97, 83, 96, 83, 111, 83, 110, 83, 187, 83, 239, 83, 228, 83, 243, 83, 236, 83, 238, 83, 233, 83, 232, 83, 252, 83, 248, 83, 245, 83, 235, 83, 230, 83, 234, 83, 242, 83, 241, 83, 240, 83, 229, 83, 237, 83, 251, 83, 219, 86, 218, 86, 22, 89, 46, 89, 49, 89, 116, 89, 118, 89, 85, 91, 131, 91, 60, 92, 232, 93, 231, 93, 230, 93, 2, 94, 3, 94, 115, 94, 124, 94, 1, 95, 24, 95, 23, 95, 197, 95, 10, 98, 83, 98, 84, 98, 82, 98, 81, 98, 165, 101, 230, 101, 46, 103, 44, 103, 42, 103, 43, 103, 45, 103, 99, 107, 205, 107, 17, 108, 16, 108, 56, 108, 65, 108, 64, 108, 62, 108, 175, 114, 132, 115, 137, 115, 220, 116, 230, 116, 24, 117, 31, 117, 40, 117, 41, 117, 48, 117, 49, 117, 50, 117, 51, 117, 139, 117, 125, 118, 174, 118, 191, 118, 238, 118, 219, 119, 226, 119, 243, 119, 58, 121, 190, 121, 116, 122, 203, 122, 30, 78, 31, 78, 82, 78, 83, 78, 105, 78, 153, 78, 164, 78, 166, 78, 165, 78, 255, 78, 9, 79, 25, 79, 10, 79, 21, 79, 13, 79, 16, 79, 17, 79, 15, 79, 242, 78, 246, 78, 251, 78, 240, 78, 243, 78, 253, 78, 1, 79, 11, 79, 73, 81, 71, 81, 70, 81, 72, 81, 104, 81, 113, 81, 141, 81, 176, 81, 23, 82, 17, 82, 18, 82, 14, 82, 22, 82, 163, 82, 8, 83, 33, 83, 32, 83, 112, 83, 113, 83, 9, 84, 15, 84, 12, 84, 10, 84, 16, 84, 1, 84, 11, 84, 4, 84, 17, 84, 13, 84, 8, 84, 3, 84, 14, 84, 6, 84, 18, 84, 224, 86, 222, 86, 221, 86, 51, 87, 48, 87, 40, 87, 45, 87, 44, 87, 47, 87, 41, 87, 25, 89, 26, 89, 55, 89, 56, 89, 132, 89, 120, 89, 131, 89, 125, 89, 121, 89, 130, 89, 129, 89, 87, 91, 88, 91, 135, 91, 136, 91, 133, 91, 137, 91, 250, 91, 22, 92, 121, 92, 222, 93, 6, 94, 118, 94, 116, 94, 15, 95, 27, 95, 217, 95, 214, 95, 14, 98, 12, 98, 13, 98, 16, 98, 99, 98, 91, 98, 88, 98, 54, 101, 233, 101, 232, 101, 236, 101, 237, 101, 242, 102, 243, 102, 9, 103, 61, 103, 52, 103, 49, 103, 53, 103, 33, 107, 100, 107, 123, 107, 22, 108, 93, 108, 87, 108, 89, 108, 95, 108, 96, 108, 80, 108, 85, 108, 97, 108, 91, 108, 77, 108, 78, 108, 112, 112, 95, 114, 93, 114, 126, 118, 249, 122, 115, 124, 248, 124, 54, 127, 138, 127, 189, 127, 1, 128, 3, 128, 12, 128, 18, 128, 51, 128, 127, 128, 137, 128, 139, 128, 140, 128, 227, 129, 234, 129, 243, 129, 252, 129, 12, 130, 27, 130, 31, 130, 110, 130, 114, 130, 126, 130, 107, 134, 64, 136, 76, 136, 99, 136, 127, 137, 33, 150, 50, 78, 168, 78, 77, 79, 79, 79, 71, 79, 87, 79, 94, 79, 52, 79, 91, 79, 85, 79, 48, 79, 80, 79, 81, 79, 61, 79, 58, 79, 56, 79, 67, 79, 84, 79, 60, 79, 70, 79, 99, 79, 92, 79, 96, 79, 47, 79, 78, 79, 54, 79, 89, 79, 93, 79, 72, 79, 90, 79, 76, 81, 75, 81, 77, 81, 117, 81, 182, 81, 183, 81, 37, 82, 36, 82, 41, 82, 42, 82, 40, 82, 171, 82, 169, 82, 170, 82, 172, 82, 35, 83, 115, 83, 117, 83, 29, 84, 45, 84, 30, 84, 62, 84, 38, 84, 78, 84, 39, 84, 70, 84, 67, 84, 51, 84, 72, 84, 66, 84, 27, 84, 41, 84, 74, 84, 57, 84, 59, 84, 56, 84, 46, 84, 53, 84, 54, 84, 32, 84, 60, 84, 64, 84, 49, 84, 43, 84, 31, 84, 44, 84, 234, 86, 240, 86, 228, 86, 235, 86, 74, 87, 81, 87, 64, 87, 77, 87, 71, 87, 78, 87, 62, 87, 80, 87, 79, 87, 59, 87, 239, 88, 62, 89, 157, 89, 146, 89, 168, 89, 158, 89, 163, 89, 153, 89, 150, 89, 141, 89, 164, 89, 147, 89, 138, 89, 165, 89, 93, 91, 92, 91, 90, 91, 91, 91, 140, 91, 139, 91, 143, 91, 44, 92, 64, 92, 65, 92, 63, 92, 62, 92, 144, 92, 145, 92, 148, 92, 140, 92, 235, 93, 12, 94, 143, 94, 135, 94, 138, 94, 247, 94, 4, 95, 31, 95, 100, 95, 98, 95, 119, 95, 121, 95, 216, 95, 204, 95, 215, 95, 205, 95, 241, 95, 235, 95, 248, 95, 234, 95, 18, 98, 17, 98, 132, 98, 151, 98, 150, 98, 128, 98, 118, 98, 137, 98, 109, 98, 138, 98, 124, 98, 126, 98, 121, 98, 115, 98, 146, 98, 111, 98, 152, 98, 110, 98, 149, 98, 147, 98, 145, 98, 134, 98, 57, 101, 59, 101, 56, 101, 241, 101, 244, 102, 95, 103, 78, 103, 79, 103, 80, 103, 81, 103, 92, 103, 86, 103, 94, 103, 73, 103, 70, 103, 96, 103, 83, 103, 87, 103, 101, 107, 207, 107, 66, 108, 94, 108, 153, 108, 129, 108, 136, 108, 137, 108, 133, 108, 155, 108, 106, 108, 122, 108, 144, 108, 112, 108, 140, 108, 104, 108, 150, 108, 146, 108, 125, 108, 131, 108, 114, 108, 126, 108, 116, 108, 134, 108, 118, 108, 141, 108, 148, 108, 152, 108, 130, 108, 118, 112, 124, 112, 125, 112, 120, 112, 98, 114, 97, 114, 96, 114, 196, 114, 194, 114, 150, 115, 44, 117, 43, 117, 55, 117, 56, 117, 130, 118, 239, 118, 227, 119, 193, 121, 192, 121, 191, 121, 118, 122, 251, 124, 85, 127, 150, 128, 147, 128, 157, 128, 152, 128, 155, 128, 154, 128, 178, 128, 111, 130, 146, 130, 139, 130, 141, 130, 139, 137, 210, 137, 0, 138, 55, 140, 70, 140, 85, 140, 157, 140, 100, 141, 112, 141, 179, 141, 171, 142, 202, 142, 155, 143, 176, 143, 194, 143, 198, 143, 197, 143, 196, 143, 225, 93, 145, 144, 162, 144, 170, 144, 166, 144, 163, 144, 73, 145, 198, 145, 204, 145, 50, 150, 46, 150, 49, 150, 42, 150, 44, 150, 38, 78, 86, 78, 115, 78, 139, 78, 155, 78, 158, 78, 171, 78, 172, 78, 111, 79, 157, 79, 141, 79, 115, 79, 127, 79, 108, 79, 155, 79, 139, 79, 134, 79, 131, 79, 112, 79, 117, 79, 136, 79, 105, 79, 123, 79, 150, 79, 126, 79, 143, 79, 145, 79, 122, 79, 84, 81, 82, 81, 85, 81, 105, 81, 119, 81, 118, 81, 120, 81, 189, 81, 253, 81, 59, 82, 56, 82, 55, 82, 58, 82, 48, 82, 46, 82, 54, 82, 65, 82, 190, 82, 187, 82, 82, 83, 84, 83, 83, 83, 81, 83, 102, 83, 119, 83, 120, 83, 121, 83, 214, 83, 212, 83, 215, 83, 115, 84, 117, 84, 150, 84, 120, 84, 149, 84, 128, 84, 123, 84, 119, 84, 132, 84, 146, 84, 134, 84, 124, 84, 144, 84, 113, 84, 118, 84, 140, 84, 154, 84, 98, 84, 104, 84, 139, 84, 125, 84, 142, 84, 250, 86, 131, 87, 119, 87, 106, 87, 105, 87, 97, 87, 102, 87, 100, 87, 124, 87, 28, 89, 73, 89, 71, 89, 72, 89, 68, 89, 84, 89, 190, 89, 187, 89, 212, 89, 185, 89, 174, 89, 209, 89, 198, 89, 208, 89, 205, 89, 203, 89, 211, 89, 202, 89, 175, 89, 179, 89, 210, 89, 197, 89, 95, 91, 100, 91, 99, 91, 151, 91, 154, 91, 152, 91, 156, 91, 153, 91, 155, 91, 26, 92, 72, 92, 69, 92, 70, 92, 183, 92, 161, 92, 184, 92, 169, 92, 171, 92, 177, 92, 179, 92, 24, 94, 26, 94, 22, 94, 21, 94, 27, 94, 17, 94, 120, 94, 154, 94, 151, 94, 156, 94, 149, 94, 150, 94, 246, 94, 38, 95, 39, 95, 41, 95, 128, 95, 129, 95, 127, 95, 124, 95, 221, 95, 224, 95, 253, 95, 245, 95, 255, 95, 15, 96, 20, 96, 47, 96, 53, 96, 22, 96, 42, 96, 21, 96, 33, 96, 39, 96, 41, 96, 43, 96, 27, 96, 22, 98, 21, 98, 63, 98, 62, 98, 64, 98, 127, 98, 201, 98, 204, 98, 196, 98, 191, 98, 194, 98, 185, 98, 210, 98, 219, 98, 171, 98, 211, 98, 212, 98, 203, 98, 200, 98, 168, 98, 189, 98, 188, 98, 208, 98, 217, 98, 199, 98, 205, 98, 181, 98, 218, 98, 177, 98, 216, 98, 214, 98, 215, 98, 198, 98, 172, 98, 206, 98, 62, 101, 167, 101, 188, 101, 250, 101, 20, 102, 19, 102, 12, 102, 6, 102, 2, 102, 14, 102, 0, 102, 15, 102, 21, 102, 10, 102, 7, 102, 13, 103, 11, 103, 109, 103, 139, 103, 149, 103, 113, 103, 156, 103, 115, 103, 119, 103, 135, 103, 157, 103, 151, 103, 111, 103, 112, 103, 127, 103, 137, 103, 126, 103, 144, 103, 117, 103, 154, 103, 147, 103, 124, 103, 106, 103, 114, 103, 35, 107, 102, 107, 103, 107, 127, 107, 19, 108, 27, 108, 227, 108, 232, 108, 243, 108, 177, 108, 204, 108, 229, 108, 179, 108, 189, 108, 190, 108, 188, 108, 226, 108, 171, 108, 213, 108, 211, 108, 184, 108, 196, 108, 185, 108, 193, 108, 174, 108, 215, 108, 197, 108, 241, 108, 191, 108, 187, 108, 225, 108, 219, 108, 202, 108, 172, 108, 239, 108, 220, 108, 214, 108, 224, 108, 149, 112, 142, 112, 146, 112, 138, 112, 153, 112, 44, 114, 45, 114, 56, 114, 72, 114, 103, 114, 105, 114, 192, 114, 206, 114, 217, 114, 215, 114, 208, 114, 169, 115, 168, 115, 159, 115, 171, 115, 165, 115, 61, 117, 157, 117, 153, 117, 154, 117, 132, 118, 194, 118, 242, 118, 244, 118, 229, 119, 253, 119, 62, 121, 64, 121, 65, 121, 201, 121, 200, 121, 122, 122, 121, 122, 250, 122, 254, 124, 84, 127, 140, 127, 139, 127, 5, 128, 186, 128, 165, 128, 162, 128, 177, 128, 161, 128, 171, 128, 169, 128, 180, 128, 170, 128, 175, 128, 229, 129, 254, 129, 13, 130, 179, 130, 157, 130, 153, 130, 173, 130, 189, 130, 159, 130, 185, 130, 177, 130, 172, 130, 165, 130, 175, 130, 184, 130, 163, 130, 176, 130, 190, 130, 183, 130, 78, 134, 113, 134, 29, 82, 104, 136, 203, 142, 206, 143, 212, 143, 209, 143, 181, 144, 184, 144, 177, 144, 182, 144, 199, 145, 209, 145, 119, 149, 128, 149, 28, 150, 64, 150, 63, 150, 59, 150, 68, 150, 66, 150, 185, 150, 232, 150, 82, 151, 94, 151, 159, 78, 173, 78, 174, 78, 225, 79, 181, 79, 175, 79, 191, 79, 224, 79, 209, 79, 207, 79, 221, 79, 195, 79, 182, 79, 216, 79, 223, 79, 202, 79, 215, 79, 174, 79, 208, 79, 196, 79, 194, 79, 218, 79, 206, 79, 222, 79, 183, 79, 87, 81, 146, 81, 145, 81, 160, 81, 78, 82, 67, 82, 74, 82, 77, 82, 76, 82, 75, 82, 71, 82, 199, 82, 201, 82, 195, 82, 193, 82, 13, 83, 87, 83, 123, 83, 154, 83, 219, 83, 172, 84, 192, 84, 168, 84, 206, 84, 201, 84, 184, 84, 166, 84, 179, 84, 199, 84, 194, 84, 189, 84, 170, 84, 193, 84, 196, 84, 200, 84, 175, 84, 171, 84, 177, 84, 187, 84, 169, 84, 167, 84, 191, 84, 255, 86, 130, 87, 139, 87, 160, 87, 163, 87, 162, 87, 206, 87, 174, 87, 147, 87, 85, 89, 81, 89, 79, 89, 78, 89, 80, 89, 220, 89, 216, 89, 255, 89, 227, 89, 232, 89, 3, 90, 229, 89, 234, 89, 218, 89, 230, 89, 1, 90, 251, 89, 105, 91, 163, 91, 166, 91, 164, 91, 162, 91, 165, 91, 1, 92, 78, 92, 79, 92, 77, 92, 75, 92, 217, 92, 210, 92, 247, 93, 29, 94, 37, 94, 31, 94, 125, 94, 160, 94, 166, 94, 250, 94, 8, 95, 45, 95, 101, 95, 136, 95, 133, 95, 138, 95, 139, 95, 135, 95, 140, 95, 137, 95, 18, 96, 29, 96, 32, 96, 37, 96, 14, 96, 40, 96, 77, 96, 112, 96, 104, 96, 98, 96, 70, 96, 67, 96, 108, 96, 107, 96, 106, 96, 100, 96, 65, 98, 220, 98, 22, 99, 9, 99, 252, 98, 237, 98, 1, 99, 238, 98, 253, 98, 7, 99, 241, 98, 247, 98, 239, 98, 236, 98, 254, 98, 244, 98, 17, 99, 2, 99, 63, 101, 69, 101, 171, 101, 189, 101, 226, 101, 37, 102, 45, 102, 32, 102, 39, 102, 47, 102, 31, 102, 40, 102, 49, 102, 36, 102, 247, 102, 255, 103, 211, 103, 241, 103, 212, 103, 208, 103, 236, 103, 182, 103, 175, 103, 245, 103, 233, 103, 239, 103, 196, 103, 209, 103, 180, 103, 218, 103, 229, 103, 184, 103, 207, 103, 222, 103, 243, 103, 176, 103, 217, 103, 226, 103, 221, 103, 210, 103, 106, 107, 131, 107, 134, 107, 181, 107, 210, 107, 215, 107, 31, 108, 201, 108, 11, 109, 50, 109, 42, 109, 65, 109, 37, 109, 12, 109, 49, 109, 30, 109, 23, 109, 59, 109, 61, 109, 62, 109, 54, 109, 27, 109, 245, 108, 57, 109, 39, 109, 56, 109, 41, 109, 46, 109, 53, 109, 14, 109, 43, 109, 171, 112, 186, 112, 179, 112, 172, 112, 175, 112, 173, 112, 184, 112, 174, 112, 164, 112, 48, 114, 114, 114, 111, 114, 116, 114, 233, 114, 224, 114, 225, 114, 183, 115, 202, 115, 187, 115, 178, 115, 205, 115, 192, 115, 179, 115, 26, 117, 45, 117, 79, 117, 76, 117, 78, 117, 75, 117, 171, 117, 164, 117, 165, 117, 162, 117, 163, 117, 120, 118, 134, 118, 135, 118, 136, 118, 200, 118, 198, 118, 195, 118, 197, 118, 1, 119, 249, 118, 248, 118, 9, 119, 11, 119, 254, 118, 252, 118, 7, 119, 220, 119, 2, 120, 20, 120, 12, 120, 13, 120, 70, 121, 73, 121, 72, 121, 71, 121, 185, 121, 186, 121, 209, 121, 210, 121, 203, 121, 127, 122, 129, 122, 255, 122, 253, 122, 125, 124, 2, 125, 5, 125, 0, 125, 9, 125, 7, 125, 4, 125, 6, 125, 56, 127, 142, 127, 191, 127, 4, 128, 16, 128, 13, 128, 17, 128, 54, 128, 214, 128, 229, 128, 218, 128, 195, 128, 196, 128, 204, 128, 225, 128, 219, 128, 206, 128, 222, 128, 228, 128, 221, 128, 244, 129, 34, 130, 231, 130, 3, 131, 5, 131, 227, 130, 219, 130, 230, 130, 4, 131, 229, 130, 2, 131, 9, 131, 210, 130, 215, 130, 241, 130, 1, 131, 220, 130, 212, 130, 209, 130, 222, 130, 211, 130, 223, 130, 239, 130, 6, 131, 80, 134, 121, 134, 123, 134, 122, 134, 77, 136, 107, 136, 129, 137, 212, 137, 8, 138, 2, 138, 3, 138, 158, 140, 160, 140, 116, 141, 115, 141, 180, 141, 205, 142, 204, 142, 240, 143, 230, 143, 226, 143, 234, 143, 229, 143, 237, 143, 235, 143, 228, 143, 232, 143, 202, 144, 206, 144, 193, 144, 195, 144, 75, 145, 74, 145, 205, 145, 130, 149, 80, 150, 75, 150, 76, 150, 77, 150, 98, 151, 105, 151, 203, 151, 237, 151, 243, 151, 1, 152, 168, 152, 219, 152, 223, 152, 150, 153, 153, 153, 88, 78, 179, 78, 12, 80, 13, 80, 35, 80, 239, 79, 38, 80, 37, 80, 248, 79, 41, 80, 22, 80, 6, 80, 60, 80, 31, 80, 26, 80, 18, 80, 17, 80, 250, 79, 0, 80, 20, 80, 40, 80, 241, 79, 33, 80, 11, 80, 25, 80, 24, 80, 243, 79, 238, 79, 45, 80, 42, 80, 254, 79, 43, 80, 9, 80, 124, 81, 164, 81, 165, 81, 162, 81, 205, 81, 204, 81, 198, 81, 203, 81, 86, 82, 92, 82, 84, 82, 91, 82, 93, 82, 42, 83, 127, 83, 159, 83, 157, 83, 223, 83, 232, 84, 16, 85, 1, 85, 55, 85, 252, 84, 229, 84, 242, 84, 6, 85, 250, 84, 20, 85, 233, 84, 237, 84, 225, 84, 9, 85, 238, 84, 234, 84, 230, 84, 39, 85, 7, 85, 253, 84, 15, 85, 3, 87, 4, 87, 194, 87, 212, 87, 203, 87, 195, 87, 9, 88, 15, 89, 87, 89, 88, 89, 90, 89, 17, 90, 24, 90, 28, 90, 31, 90, 27, 90, 19, 90, 236, 89, 32, 90, 35, 90, 41, 90, 37, 90, 12, 90, 9, 90, 107, 91, 88, 92, 176, 91, 179, 91, 182, 91, 180, 91, 174, 91, 181, 91, 185, 91, 184, 91, 4, 92, 81, 92, 85, 92, 80, 92, 237, 92, 253, 92, 251, 92, 234, 92, 232, 92, 240, 92, 246, 92, 1, 93, 244, 92, 238, 93, 45, 94, 43, 94, 171, 94, 173, 94, 167, 94, 49, 95, 146, 95, 145, 95, 144, 95, 89, 96, 99, 96, 101, 96, 80, 96, 85, 96, 109, 96, 105, 96, 111, 96, 132, 96, 159, 96, 154, 96, 141, 96, 148, 96, 140, 96, 133, 96, 150, 96, 71, 98, 243, 98, 8, 99, 255, 98, 78, 99, 62, 99, 47, 99, 85, 99, 66, 99, 70, 99, 79, 99, 73, 99, 58, 99, 80, 99, 61, 99, 42, 99, 43, 99, 40, 99, 77, 99, 76, 99, 72, 101, 73, 101, 153, 101, 193, 101, 197, 101, 66, 102, 73, 102, 79, 102, 67, 102, 82, 102, 76, 102, 69, 102, 65, 102, 248, 102, 20, 103, 21, 103, 23, 103, 33, 104, 56, 104, 72, 104, 70, 104, 83, 104, 57, 104, 66, 104, 84, 104, 41, 104, 179, 104, 23, 104, 76, 104, 81, 104, 61, 104, 244, 103, 80, 104, 64, 104, 60, 104, 67, 104, 42, 104, 69, 104, 19, 104, 24, 104, 65, 104, 138, 107, 137, 107, 183, 107, 35, 108, 39, 108, 40, 108, 38, 108, 36, 108, 240, 108, 106, 109, 149, 109, 136, 109, 135, 109, 102, 109, 120, 109, 119, 109, 89, 109, 147, 109, 108, 109, 137, 109, 110, 109, 90, 109, 116, 109, 105, 109, 140, 109, 138, 109, 121, 109, 133, 109, 101, 109, 148, 109, 202, 112, 216, 112, 228, 112, 217, 112, 200, 112, 207, 112, 57, 114, 121, 114, 252, 114, 249, 114, 253, 114, 248, 114, 247, 114, 134, 115, 237, 115, 9, 116, 238, 115, 224, 115, 234, 115, 222, 115, 84, 117, 93, 117, 92, 117, 90, 117, 89, 117, 190, 117, 197, 117, 199, 117, 178, 117, 179, 117, 189, 117, 188, 117, 185, 117, 194, 117, 184, 117, 139, 118, 176, 118, 202, 118, 205, 118, 206, 118, 41, 119, 31, 119, 32, 119, 40, 119, 233, 119, 48, 120, 39, 120, 56, 120, 29, 120, 52, 120, 55, 120, 37, 120, 45, 120, 32, 120, 31, 120, 50, 120, 85, 121, 80, 121, 96, 121, 95, 121, 86, 121, 94, 121, 93, 121, 87, 121, 90, 121, 228, 121, 227, 121, 231, 121, 223, 121, 230, 121, 233, 121, 216, 121, 132, 122, 136, 122, 217, 122, 6, 123, 17, 123, 137, 124, 33, 125, 23, 125, 11, 125, 10, 125, 32, 125, 34, 125, 20, 125, 16, 125, 21, 125, 26, 125, 28, 125, 13, 125, 25, 125, 27, 125, 58, 127, 95, 127, 148, 127, 197, 127, 193, 127, 6, 128, 24, 128, 21, 128, 25, 128, 23, 128, 61, 128, 63, 128, 241, 128, 2, 129, 240, 128, 5, 129, 237, 128, 244, 128, 6, 129, 248, 128, 243, 128, 8, 129, 253, 128, 10, 129, 252, 128, 239, 128, 237, 129, 236, 129, 0, 130, 16, 130, 42, 130, 43, 130, 40, 130, 44, 130, 187, 130, 43, 131, 82, 131, 84, 131, 74, 131, 56, 131, 80, 131, 73, 131, 53, 131, 52, 131, 79, 131, 50, 131, 57, 131, 54, 131, 23, 131, 64, 131, 49, 131, 40, 131, 67, 131, 84, 134, 138, 134, 170, 134, 147, 134, 164, 134, 169, 134, 140, 134, 163, 134, 156, 134, 112, 136, 119, 136, 129, 136, 130, 136, 125, 136, 121, 136, 24, 138, 16, 138, 14, 138, 12, 138, 21, 138, 10, 138, 23, 138, 19, 138, 22, 138, 15, 138, 17, 138, 72, 140, 122, 140, 121, 140, 161, 140, 162, 140, 119, 141, 172, 142, 210, 142, 212, 142, 207, 142, 177, 143, 1, 144, 6, 144, 247, 143, 0, 144, 250, 143, 244, 143, 3, 144, 253, 143, 5, 144, 248, 143, 149, 144, 225, 144, 221, 144, 226, 144, 82, 145, 77, 145, 76, 145, 216, 145, 221, 145, 215, 145, 220, 145, 217, 145, 131, 149, 98, 150, 99, 150, 97, 150, 91, 150, 93, 150, 100, 150, 88, 150, 94, 150, 187, 150, 226, 152, 172, 153, 168, 154, 216, 154, 37, 155, 50, 155, 60, 155, 126, 78, 122, 80, 125, 80, 92, 80, 71, 80, 67, 80, 76, 80, 90, 80, 73, 80, 101, 80, 118, 80, 78, 80, 85, 80, 117, 80, 116, 80, 119, 80, 79, 80, 15, 80, 111, 80, 109, 80, 92, 81, 149, 81, 240, 81, 106, 82, 111, 82, 210, 82, 217, 82, 216, 82, 213, 82, 16, 83, 15, 83, 25, 83, 63, 83, 64, 83, 62, 83, 195, 83, 252, 102, 70, 85, 106, 85, 102, 85, 68, 85, 94, 85, 97, 85, 67, 85, 74, 85, 49, 85, 86, 85, 79, 85, 85, 85, 47, 85, 100, 85, 56, 85, 46, 85, 92, 85, 44, 85, 99, 85, 51, 85, 65, 85, 87, 85, 8, 87, 11, 87, 9, 87, 223, 87, 5, 88, 10, 88, 6, 88, 224, 87, 228, 87, 250, 87, 2, 88, 53, 88, 247, 87, 249, 87, 32, 89, 98, 89, 54, 90, 65, 90, 73, 90, 102, 90, 106, 90, 64, 90, 60, 90, 98, 90, 90, 90, 70, 90, 74, 90, 112, 91, 199, 91, 197, 91, 196, 91, 194, 91, 191, 91, 198, 91, 9, 92, 8, 92, 7, 92, 96, 92, 92, 92, 93, 92, 7, 93, 6, 93, 14, 93, 27, 93, 22, 93, 34, 93, 17, 93, 41, 93, 20, 93, 25, 93, 36, 93, 39, 93, 23, 93, 226, 93, 56, 94, 54, 94, 51, 94, 55, 94, 183, 94, 184, 94, 182, 94, 181, 94, 190, 94, 53, 95, 55, 95, 87, 95, 108, 95, 105, 95, 107, 95, 151, 95, 153, 95, 158, 95, 152, 95, 161, 95, 160, 95, 156, 95, 127, 96, 163, 96, 137, 96, 160, 96, 168, 96, 203, 96, 180, 96, 230, 96, 189, 96, 197, 96, 187, 96, 181, 96, 220, 96, 188, 96, 216, 96, 213, 96, 198, 96, 223, 96, 184, 96, 218, 96, 199, 96, 26, 98, 27, 98, 72, 98, 160, 99, 167, 99, 114, 99, 150, 99, 162, 99, 165, 99, 119, 99, 103, 99, 152, 99, 170, 99, 113, 99, 169, 99, 137, 99, 131, 99, 155, 99, 107, 99, 168, 99, 132, 99, 136, 99, 153, 99, 161, 99, 172, 99, 146, 99, 143, 99, 128, 99, 123, 99, 105, 99, 104, 99, 122, 99, 93, 101, 86, 101, 81, 101, 89, 101, 87, 101, 95, 85, 79, 101, 88, 101, 85, 101, 84, 101, 156, 101, 155, 101, 172, 101, 207, 101, 203, 101, 204, 101, 206, 101, 93, 102, 90, 102, 100, 102, 104, 102, 102, 102, 94, 102, 249, 102, 215, 82, 27, 103, 129, 104, 175, 104, 162, 104, 147, 104, 181, 104, 127, 104, 118, 104, 177, 104, 167, 104, 151, 104, 176, 104, 131, 104, 196, 104, 173, 104, 134, 104, 133, 104, 148, 104, 157, 104, 168, 104, 159, 104, 161, 104, 130, 104, 50, 107, 186, 107, 235, 107, 236, 107, 43, 108, 142, 109, 188, 109, 243, 109, 217, 109, 178, 109, 225, 109, 204, 109, 228, 109, 251, 109, 250, 109, 5, 110, 199, 109, 203, 109, 175, 109, 209, 109, 174, 109, 222, 109, 249, 109, 184, 109, 247, 109, 245, 109, 197, 109, 210, 109, 26, 110, 181, 109, 218, 109, 235, 109, 216, 109, 234, 109, 241, 109, 238, 109, 232, 109, 198, 109, 196, 109, 170, 109, 236, 109, 191, 109, 230, 109, 249, 112, 9, 113, 10, 113, 253, 112, 239, 112, 61, 114, 125, 114, 129, 114, 28, 115, 27, 115, 22, 115, 19, 115, 25, 115, 135, 115, 5, 116, 10, 116, 3, 116, 6, 116, 254, 115, 13, 116, 224, 116, 246, 116, 247, 116, 28, 117, 34, 117, 101, 117, 102, 117, 98, 117, 112, 117, 143, 117, 212, 117, 213, 117, 181, 117, 202, 117, 205, 117, 142, 118, 212, 118, 210, 118, 219, 118, 55, 119, 62, 119, 60, 119, 54, 119, 56, 119, 58, 119, 107, 120, 67, 120, 78, 120, 101, 121, 104, 121, 109, 121, 251, 121, 146, 122, 149, 122, 32, 123, 40, 123, 27, 123, 44, 123, 38, 123, 25, 123, 30, 123, 46, 123, 146, 124, 151, 124, 149, 124, 70, 125, 67, 125, 113, 125, 46, 125, 57, 125, 60, 125, 64, 125, 48, 125, 51, 125, 68, 125, 47, 125, 66, 125, 50, 125, 49, 125, 61, 127, 158, 127, 154, 127, 204, 127, 206, 127, 210, 127, 28, 128, 74, 128, 70, 128, 47, 129, 22, 129, 35, 129, 43, 129, 41, 129, 48, 129, 36, 129, 2, 130, 53, 130, 55, 130, 54, 130, 57, 130, 142, 131, 158, 131, 152, 131, 120, 131, 162, 131, 150, 131, 189, 131, 171, 131, 146, 131, 138, 131, 147, 131, 137, 131, 160, 131, 119, 131, 123, 131, 124, 131, 134, 131, 167, 131, 85, 134, 106, 95, 199, 134, 192, 134, 182, 134, 196, 134, 181, 134, 198, 134, 203, 134, 177, 134, 175, 134, 201, 134, 83, 136, 158, 136, 136, 136, 171, 136, 146, 136, 150, 136, 141, 136, 139, 136, 147, 137, 143, 137, 42, 138, 29, 138, 35, 138, 37, 138, 49, 138, 45, 138, 31, 138, 27, 138, 34, 138, 73, 140, 90, 140, 169, 140, 172, 140, 171, 140, 168, 140, 170, 140, 167, 140, 103, 141, 102, 141, 190, 141, 186, 141, 219, 142, 223, 142, 25, 144, 13, 144, 26, 144, 23, 144, 35, 144, 31, 144, 29, 144, 16, 144, 21, 144, 30, 144, 32, 144, 15, 144, 34, 144, 22, 144, 27, 144, 20, 144, 232, 144, 237, 144, 253, 144, 87, 145, 206, 145, 245, 145, 230, 145, 227, 145, 231, 145, 237, 145, 233, 145, 137, 149, 106, 150, 117, 150, 115, 150, 120, 150, 112, 150, 116, 150, 118, 150, 119, 150, 108, 150, 192, 150, 234, 150, 233, 150, 224, 122, 223, 122, 2, 152, 3, 152, 90, 155, 229, 156, 117, 158, 127, 158, 165, 158, 187, 158, 162, 80, 141, 80, 133, 80, 153, 80, 145, 80, 128, 80, 150, 80, 152, 80, 154, 80, 0, 103, 241, 81, 114, 82, 116, 82, 117, 82, 105, 82, 222, 82, 221, 82, 219, 82, 90, 83, 165, 83, 123, 85, 128, 85, 167, 85, 124, 85, 138, 85, 157, 85, 152, 85, 130, 85, 156, 85, 170, 85, 148, 85, 135, 85, 139, 85, 131, 85, 179, 85, 174, 85, 159, 85, 62, 85, 178, 85, 154, 85, 187, 85, 172, 85, 177, 85, 126, 85, 137, 85, 171, 85, 153, 85, 13, 87, 47, 88, 42, 88, 52, 88, 36, 88, 48, 88, 49, 88, 33, 88, 29, 88, 32, 88, 249, 88, 250, 88, 96, 89, 119, 90, 154, 90, 127, 90, 146, 90, 155, 90, 167, 90, 115, 91, 113, 91, 210, 91, 204, 91, 211, 91, 208, 91, 10, 92, 11, 92, 49, 92, 76, 93, 80, 93, 52, 93, 71, 93, 253, 93, 69, 94, 61, 94, 64, 94, 67, 94, 126, 94, 202, 94, 193, 94, 194, 94, 196, 94, 60, 95, 109, 95, 169, 95, 170, 95, 168, 95, 209, 96, 225, 96, 178, 96, 182, 96, 224, 96, 28, 97, 35, 97, 250, 96, 21, 97, 240, 96, 251, 96, 244, 96, 104, 97, 241, 96, 14, 97, 246, 96, 9, 97, 0, 97, 18, 97, 31, 98, 73, 98, 163, 99, 140, 99, 207, 99, 192, 99, 233, 99, 201, 99, 198, 99, 205, 99, 210, 99, 227, 99, 208, 99, 225, 99, 214, 99, 237, 99, 238, 99, 118, 99, 244, 99, 234, 99, 219, 99, 82, 100, 218, 99, 249, 99, 94, 101, 102, 101, 98, 101, 99, 101, 145, 101, 144, 101, 175, 101, 110, 102, 112, 102, 116, 102, 118, 102, 111, 102, 145, 102, 122, 102, 126, 102, 119, 102, 254, 102, 255, 102, 31, 103, 29, 103, 250, 104, 213, 104, 224, 104, 216, 104, 215, 104, 5, 105, 223, 104, 245, 104, 238, 104, 231, 104, 249, 104, 210, 104, 242, 104, 227, 104, 203, 104, 205, 104, 13, 105, 18, 105, 14, 105, 201, 104, 218, 104, 110, 105, 251, 104, 62, 107, 58, 107, 61, 107, 152, 107, 150, 107, 188, 107, 239, 107, 46, 108, 47, 108, 44, 108, 47, 110, 56, 110, 84, 110, 33, 110, 50, 110, 103, 110, 74, 110, 32, 110, 37, 110, 35, 110, 27, 110, 91, 110, 88, 110, 36, 110, 86, 110, 110, 110, 45, 110, 38, 110, 111, 110, 52, 110, 77, 110, 58, 110, 44, 110, 67, 110, 29, 110, 62, 110, 203, 110, 137, 110, 25, 110, 78, 110, 99, 110, 68, 110, 114, 110, 105, 110, 95, 110, 25, 113, 26, 113, 38, 113, 48, 113, 33, 113, 54, 113, 110, 113, 28, 113, 76, 114, 132, 114, 128, 114, 54, 115, 37, 115, 52, 115, 41, 115, 58, 116, 42, 116, 51, 116, 34, 116, 37, 116, 53, 116, 54, 116, 52, 116, 47, 116, 27, 116, 38, 116, 40, 116, 37, 117, 38, 117, 107, 117, 106, 117, 226, 117, 219, 117, 227, 117, 217, 117, 216, 117, 222, 117, 224, 117, 123, 118, 124, 118, 150, 118, 147, 118, 180, 118, 220, 118, 79, 119, 237, 119, 93, 120, 108, 120, 111, 120, 13, 122, 8, 122, 11, 122, 5, 122, 0, 122, 152, 122, 151, 122, 150, 122, 229, 122, 227, 122, 73, 123, 86, 123, 70, 123, 80, 123, 82, 123, 84, 123, 77, 123, 75, 123, 79, 123, 81, 123, 159, 124, 165, 124, 94, 125, 80, 125, 104, 125, 85, 125, 43, 125, 110, 125, 114, 125, 97, 125, 102, 125, 98, 125, 112, 125, 115, 125, 132, 85, 212, 127, 213, 127, 11, 128, 82, 128, 133, 128, 85, 129, 84, 129, 75, 129, 81, 129, 78, 129, 57, 129, 70, 129, 62, 129, 76, 129, 83, 129, 116, 129, 18, 130, 28, 130, 233, 131, 3, 132, 248, 131, 13, 132, 224, 131, 197, 131, 11, 132, 193, 131, 239, 131, 241, 131, 244, 131, 87, 132, 10, 132, 240, 131, 12, 132, 204, 131, 253, 131, 242, 131, 202, 131, 56, 132, 14, 132, 4, 132, 220, 131, 7, 132, 212, 131, 223, 131, 91, 134, 223, 134, 217, 134, 237, 134, 212, 134, 219, 134, 228, 134, 208, 134, 222, 134, 87, 136, 193, 136, 194, 136, 177, 136, 131, 137, 150, 137, 59, 138, 96, 138, 85, 138, 94, 138, 60, 138, 65, 138, 84, 138, 91, 138, 80, 138, 70, 138, 52, 138, 58, 138, 54, 138, 86, 138, 97, 140, 130, 140, 175, 140, 188, 140, 179, 140, 189, 140, 193, 140, 187, 140, 192, 140, 180, 140, 183, 140, 182, 140, 191, 140, 184, 140, 138, 141, 133, 141, 129, 141, 206, 141, 221, 141, 203, 141, 218, 141, 209, 141, 204, 141, 219, 141, 198, 141, 251, 142, 248, 142, 252, 142, 156, 143, 46, 144, 53, 144, 49, 144, 56, 144, 50, 144, 54, 144, 2, 145, 245, 144, 9, 145, 254, 144, 99, 145, 101, 145, 207, 145, 20, 146, 21, 146, 35, 146, 9, 146, 30, 146, 13, 146, 16, 146, 7, 146, 17, 146, 148, 149, 143, 149, 139, 149, 145, 149, 147, 149, 146, 149, 142, 149, 138, 150, 142, 150, 139, 150, 125, 150, 133, 150, 134, 150, 141, 150, 114, 150, 132, 150, 193, 150, 197, 150, 196, 150, 198, 150, 199, 150, 239, 150, 242, 150, 204, 151, 5, 152, 6, 152, 8, 152, 231, 152, 234, 152, 239, 152, 233, 152, 242, 152, 237, 152, 174, 153, 173, 153, 195, 158, 205, 158, 209, 158, 130, 78, 173, 80, 181, 80, 178, 80, 179, 80, 197, 80, 190, 80, 172, 80, 183, 80, 187, 80, 175, 80, 199, 80, 127, 82, 119, 82, 125, 82, 223, 82, 230, 82, 228, 82, 226, 82, 227, 82, 47, 83, 223, 85, 232, 85, 211, 85, 230, 85, 206, 85, 220, 85, 199, 85, 209, 85, 227, 85, 228, 85, 239, 85, 218, 85, 225, 85, 197, 85, 198, 85, 229, 85, 201, 85, 18, 87, 19, 87, 94, 88, 81, 88, 88, 88, 87, 88, 90, 88, 84, 88, 107, 88, 76, 88, 109, 88, 74, 88, 98, 88, 82, 88, 75, 88, 103, 89, 193, 90, 201, 90, 204, 90, 190, 90, 189, 90, 188, 90, 179, 90, 194, 90, 178, 90, 105, 93, 111, 93, 76, 94, 121, 94, 201, 94, 200, 94, 18, 95, 89, 95, 172, 95, 174, 95, 26, 97, 15, 97, 72, 97, 31, 97, 243, 96, 27, 97, 249, 96, 1, 97, 8, 97, 78, 97, 76, 97, 68, 97, 77, 97, 62, 97, 52, 97, 39, 97, 13, 97, 6, 97, 55, 97, 33, 98, 34, 98, 19, 100, 62, 100, 30, 100, 42, 100, 45, 100, 61, 100, 44, 100, 15, 100, 28, 100, 20, 100, 13, 100, 54, 100, 22, 100, 23, 100, 6, 100, 108, 101, 159, 101, 176, 101, 151, 102, 137, 102, 135, 102, 136, 102, 150, 102, 132, 102, 152, 102, 141, 102, 3, 103, 148, 105, 109, 105, 90, 105, 119, 105, 96, 105, 84, 105, 117, 105, 48, 105, 130, 105, 74, 105, 104, 105, 107, 105, 94, 105, 83, 105, 121, 105, 134, 105, 93, 105, 99, 105, 91, 105, 71, 107, 114, 107, 192, 107, 191, 107, 211, 107, 253, 107, 162, 110, 175, 110, 211, 110, 182, 110, 194, 110, 144, 110, 157, 110, 199, 110, 197, 110, 165, 110, 152, 110, 188, 110, 186, 110, 171, 110, 209, 110, 150, 110, 156, 110, 196, 110, 212, 110, 170, 110, 167, 110, 180, 110, 78, 113, 89, 113, 105, 113, 100, 113, 73, 113, 103, 113, 92, 113, 108, 113, 102, 113, 76, 113, 101, 113, 94, 113, 70, 113, 104, 113, 86, 113, 58, 114, 82, 114, 55, 115, 69, 115, 63, 115, 62, 115, 111, 116, 90, 116, 85, 116, 95, 116, 94, 116, 65, 116, 63, 116, 89, 116, 91, 116, 92, 116, 118, 117, 120, 117, 0, 118, 240, 117, 1, 118, 242, 117, 241, 117, 250, 117, 255, 117, 244, 117, 243, 117, 222, 118, 223, 118, 91, 119, 107, 119, 102, 119, 94, 119, 99, 119, 121, 119, 106, 119, 108, 119, 92, 119, 101, 119, 104, 119, 98, 119, 238, 119, 142, 120, 176, 120, 151, 120, 152, 120, 140, 120, 137, 120, 124, 120, 145, 120, 147, 120, 127, 120, 122, 121, 127, 121, 129, 121, 44, 132, 189, 121, 28, 122, 26, 122, 32, 122, 20, 122, 31, 122, 30, 122, 159, 122, 160, 122, 119, 123, 192, 123, 96, 123, 110, 123, 103, 123, 177, 124, 179, 124, 181, 124, 147, 125, 121, 125, 145, 125, 129, 125, 143, 125, 91, 125, 110, 127, 105, 127, 106, 127, 114, 127, 169, 127, 168, 127, 164, 127, 86, 128, 88, 128, 134, 128, 132, 128, 113, 129, 112, 129, 120, 129, 101, 129, 110, 129, 115, 129, 107, 129, 121, 129, 122, 129, 102, 129, 5, 130, 71, 130, 130, 132, 119, 132, 61, 132, 49, 132, 117, 132, 102, 132, 107, 132, 73, 132, 108, 132, 91, 132, 60, 132, 53, 132, 97, 132, 99, 132, 105, 132, 109, 132, 70, 132, 94, 134, 92, 134, 95, 134, 249, 134, 19, 135, 8, 135, 7, 135, 0, 135, 254, 134, 251, 134, 2, 135, 3, 135, 6, 135, 10, 135, 89, 136, 223, 136, 212, 136, 217, 136, 220, 136, 216, 136, 221, 136, 225, 136, 202, 136, 213, 136, 210, 136, 156, 137, 227, 137, 107, 138, 114, 138, 115, 138, 102, 138, 105, 138, 112, 138, 135, 138, 124, 138, 99, 138, 160, 138, 113, 138, 133, 138, 109, 138, 98, 138, 110, 138, 108, 138, 121, 138, 123, 138, 62, 138, 104, 138, 98, 140, 138, 140, 137, 140, 202, 140, 199, 140, 200, 140, 196, 140, 178, 140, 195, 140, 194, 140, 197, 140, 225, 141, 223, 141, 232, 141, 239, 141, 243, 141, 250, 141, 234, 141, 228, 141, 230, 141, 178, 142, 3, 143, 9, 143, 254, 142, 10, 143, 159, 143, 178, 143, 75, 144, 74, 144, 83, 144, 66, 144, 84, 144, 60, 144, 85, 144, 80, 144, 71, 144, 79, 144, 78, 144, 77, 144, 81, 144, 62, 144, 65, 144, 18, 145, 23, 145, 108, 145, 106, 145, 105, 145, 201, 145, 55, 146, 87, 146, 56, 146, 61, 146, 64, 146, 62, 146, 91, 146, 75, 146, 100, 146, 81, 146, 52, 146, 73, 146, 77, 146, 69, 146, 57, 146, 63, 146, 90, 146, 152, 149, 152, 150, 148, 150, 149, 150, 205, 150, 203, 150, 201, 150, 202, 150, 247, 150, 251, 150, 249, 150, 246, 150, 86, 151, 116, 151, 118, 151, 16, 152, 17, 152, 19, 152, 10, 152, 18, 152, 12, 152, 252, 152, 244, 152, 253, 152, 254, 152, 179, 153, 177, 153, 180, 153, 225, 154, 233, 156, 130, 158, 14, 159, 19, 159, 32, 159, 231, 80, 238, 80, 229, 80, 214, 80, 237, 80, 218, 80, 213, 80, 207, 80, 209, 80, 241, 80, 206, 80, 233, 80, 98, 81, 243, 81, 131, 82, 130, 82, 49, 83, 173, 83, 254, 85, 0, 86, 27, 86, 23, 86, 253, 85, 20, 86, 6, 86, 9, 86, 13, 86, 14, 86, 247, 85, 22, 86, 31, 86, 8, 86, 16, 86, 246, 85, 24, 87, 22, 87, 117, 88, 126, 88, 131, 88, 147, 88, 138, 88, 121, 88, 133, 88, 125, 88, 253, 88, 37, 89, 34, 89, 36, 89, 106, 89, 105, 89, 225, 90, 230, 90, 233, 90, 215, 90, 214, 90, 216, 90, 227, 90, 117, 91, 222, 91, 231, 91, 225, 91, 229, 91, 230, 91, 232, 91, 226, 91, 228, 91, 223, 91, 13, 92, 98, 92, 132, 93, 135, 93, 91, 94, 99, 94, 85, 94, 87, 94, 84, 94, 211, 94, 214, 94, 10, 95, 70, 95, 112, 95, 185, 95, 71, 97, 63, 97, 75, 97, 119, 97, 98, 97, 99, 97, 95, 97, 90, 97, 88, 97, 117, 97, 42, 98, 135, 100, 88, 100, 84, 100, 164, 100, 120, 100, 95, 100, 122, 100, 81, 100, 103, 100, 52, 100, 109, 100, 123, 100, 114, 101, 161, 101, 215, 101, 214, 101, 162, 102, 168, 102, 157, 102, 156, 105, 168, 105, 149, 105, 193, 105, 174, 105, 211, 105, 203, 105, 155, 105, 183, 105, 187, 105, 171, 105, 180, 105, 208, 105, 205, 105, 173, 105, 204, 105, 166, 105, 195, 105, 163, 105, 73, 107, 76, 107, 51, 108, 51, 111, 20, 111, 254, 110, 19, 111, 244, 110, 41, 111, 62, 111, 32, 111, 44, 111, 15, 111, 2, 111, 34, 111, 255, 110, 239, 110, 6, 111, 49, 111, 56, 111, 50, 111, 35, 111, 21, 111, 43, 111, 47, 111, 136, 111, 42, 111, 236, 110, 1, 111, 242, 110, 204, 110, 247, 110, 148, 113, 153, 113, 125, 113, 138, 113, 132, 113, 146, 113, 62, 114, 146, 114, 150, 114, 68, 115, 80, 115, 100, 116, 99, 116, 106, 116, 112, 116, 109, 116, 4, 117, 145, 117, 39, 118, 13, 118, 11, 118, 9, 118, 19, 118, 225, 118, 227, 118, 132, 119, 125, 119, 127, 119, 97, 119, 193, 120, 159, 120, 167, 120, 179, 120, 169, 120, 163, 120, 142, 121, 143, 121, 141, 121, 46, 122, 49, 122, 170, 122, 169, 122, 237, 122, 239, 122, 161, 123, 149, 123, 139, 123, 117, 123, 151, 123, 157, 123, 148, 123, 143, 123, 184, 123, 135, 123, 132, 123, 185, 124, 189, 124, 190, 124, 187, 125, 176, 125, 156, 125, 189, 125, 190, 125, 160, 125, 202, 125, 180, 125, 178, 125, 177, 125, 186, 125, 162, 125, 191, 125, 181, 125, 184, 125, 173, 125, 210, 125, 199, 125, 172, 125, 112, 127, 224, 127, 225, 127, 223, 127, 94, 128, 90, 128, 135, 128, 80, 129, 128, 129, 143, 129, 136, 129, 138, 129, 127, 129, 130, 129, 231, 129, 250, 129, 7, 130, 20, 130, 30, 130, 75, 130, 201, 132, 191, 132, 198, 132, 196, 132, 153, 132, 158, 132, 178, 132, 156, 132, 203, 132, 184, 132, 192, 132, 211, 132, 144, 132, 188, 132, 209, 132, 202, 132, 63, 135, 28, 135, 59, 135, 34, 135, 37, 135, 52, 135, 24, 135, 85, 135, 55, 135, 41, 135, 243, 136, 2, 137, 244, 136, 249, 136, 248, 136, 253, 136, 232, 136, 26, 137, 239, 136, 166, 138, 140, 138, 158, 138, 163, 138, 141, 138, 161, 138, 147, 138, 164, 138, 170, 138, 165, 138, 168, 138, 152, 138, 145, 138, 154, 138, 167, 138, 106, 140, 141, 140, 140, 140, 211, 140, 209, 140, 210, 140, 107, 141, 153, 141, 149, 141, 252, 141, 20, 143, 18, 143, 21, 143, 19, 143, 163, 143, 96, 144, 88, 144, 92, 144, 99, 144, 89, 144, 94, 144, 98, 144, 93, 144, 91, 144, 25, 145, 24, 145, 30, 145, 117, 145, 120, 145, 119, 145, 116, 145, 120, 146, 128, 146, 133, 146, 152, 146, 150, 146, 123, 146, 147, 146, 156, 146, 168, 146, 124, 146, 145, 146, 161, 149, 168, 149, 169, 149, 163, 149, 165, 149, 164, 149, 153, 150, 156, 150, 155, 150, 204, 150, 210, 150, 0, 151, 124, 151, 133, 151, 246, 151, 23, 152, 24, 152, 175, 152, 177, 152, 3, 153, 5, 153, 12, 153, 9, 153, 193, 153, 175, 154, 176, 154, 230, 154, 65, 155, 66, 155, 244, 156, 246, 156, 243, 156, 188, 158, 59, 159, 74, 159, 4, 81, 0, 81, 251, 80, 245, 80, 249, 80, 2, 81, 8, 81, 9, 81, 5, 81, 220, 81, 135, 82, 136, 82, 137, 82, 141, 82, 138, 82, 240, 82, 178, 83, 46, 86, 59, 86, 57, 86, 50, 86, 63, 86, 52, 86, 41, 86, 83, 86, 78, 86, 87, 86, 116, 86, 54, 86, 47, 86, 48, 86, 128, 88, 159, 88, 158, 88, 179, 88, 156, 88, 174, 88, 169, 88, 166, 88, 109, 89, 9, 91, 251, 90, 11, 91, 245, 90, 12, 91, 8, 91, 238, 91, 236, 91, 233, 91, 235, 91, 100, 92, 101, 92, 157, 93, 148, 93, 98, 94, 95, 94, 97, 94, 226, 94, 218, 94, 223, 94, 221, 94, 227, 94, 224, 94, 72, 95, 113, 95, 183, 95, 181, 95, 118, 97, 103, 97, 110, 97, 93, 97, 85, 97, 130, 97, 124, 97, 112, 97, 107, 97, 126, 97, 167, 97, 144, 97, 171, 97, 142, 97, 172, 97, 154, 97, 164, 97, 148, 97, 174, 97, 46, 98, 105, 100, 111, 100, 121, 100, 158, 100, 178, 100, 136, 100, 144, 100, 176, 100, 165, 100, 147, 100, 149, 100, 169, 100, 146, 100, 174, 100, 173, 100, 171, 100, 154, 100, 172, 100, 153, 100, 162, 100, 179, 100, 117, 101, 119, 101, 120, 101, 174, 102, 171, 102, 180, 102, 177, 102, 35, 106, 31, 106, 232, 105, 1, 106, 30, 106, 25, 106, 253, 105, 33, 106, 19, 106, 10, 106, 243, 105, 2, 106, 5, 106, 237, 105, 17, 106, 80, 107, 78, 107, 164, 107, 197, 107, 198, 107, 63, 111, 124, 111, 132, 111, 81, 111, 102, 111, 84, 111, 134, 111, 109, 111, 91, 111, 120, 111, 110, 111, 142, 111, 122, 111, 112, 111, 100, 111, 151, 111, 88, 111, 213, 110, 111, 111, 96, 111, 95, 111, 159, 113, 172, 113, 177, 113, 168, 113, 86, 114, 155, 114, 78, 115, 87, 115, 105, 116, 139, 116, 131, 116, 126, 116, 128, 116, 127, 117, 32, 118, 41, 118, 31, 118, 36, 118, 38, 118, 33, 118, 34, 118, 154, 118, 186, 118, 228, 118, 142, 119, 135, 119, 140, 119, 145, 119, 139, 119, 203, 120, 197, 120, 186, 120, 202, 120, 190, 120, 213, 120, 188, 120, 208, 120, 63, 122, 60, 122, 64, 122, 61, 122, 55, 122, 59, 122, 175, 122, 174, 122, 173, 123, 177, 123, 196, 123, 180, 123, 198, 123, 199, 123, 193, 123, 160, 123, 204, 123, 202, 124, 224, 125, 244, 125, 239, 125, 251, 125, 216, 125, 236, 125, 221, 125, 232, 125, 227, 125, 218, 125, 222, 125, 233, 125, 158, 125, 217, 125, 242, 125, 249, 125, 117, 127, 119, 127, 175, 127, 233, 127, 38, 128, 155, 129, 156, 129, 157, 129, 160, 129, 154, 129, 152, 129, 23, 133, 61, 133, 26, 133, 238, 132, 44, 133, 45, 133, 19, 133, 17, 133, 35, 133, 33, 133, 20, 133, 236, 132, 37, 133, 255, 132, 6, 133, 130, 135, 116, 135, 118, 135, 96, 135, 102, 135, 120, 135, 104, 135, 89, 135, 87, 135, 76, 135, 83, 135, 91, 136, 93, 136, 16, 137, 7, 137, 18, 137, 19, 137, 21, 137, 10, 137, 188, 138, 210, 138, 199, 138, 196, 138, 149, 138, 203, 138, 248, 138, 178, 138, 201, 138, 194, 138, 191, 138, 176, 138, 214, 138, 205, 138, 182, 138, 185, 138, 219, 138, 76, 140, 78, 140, 108, 140, 224, 140, 222, 140, 230, 140, 228, 140, 236, 140, 237, 140, 226, 140, 227, 140, 220, 140, 234, 140, 225, 140, 109, 141, 159, 141, 163, 141, 43, 142, 16, 142, 29, 142, 34, 142, 15, 142, 41, 142, 31, 142, 33, 142, 30, 142, 186, 142, 29, 143, 27, 143, 31, 143, 41, 143, 38, 143, 42, 143, 28, 143, 30, 143, 37, 143, 105, 144, 110, 144, 104, 144, 109, 144, 119, 144, 48, 145, 45, 145, 39, 145, 49, 145, 135, 145, 137, 145, 139, 145, 131, 145, 197, 146, 187, 146, 183, 146, 234, 146, 172, 146, 228, 146, 193, 146, 179, 146, 188, 146, 210, 146, 199, 146, 240, 146, 178, 146, 173, 149, 177, 149, 4, 151, 6, 151, 7, 151, 9, 151, 96, 151, 141, 151, 139, 151, 143, 151, 33, 152, 43, 152, 28, 152, 179, 152, 10, 153, 19, 153, 18, 153, 24, 153, 221, 153, 208, 153, 223, 153, 219, 153, 209, 153, 213, 153, 210, 153, 217, 153, 183, 154, 238, 154, 239, 154, 39, 155, 69, 155, 68, 155, 119, 155, 111, 155, 6, 157, 9, 157, 3, 157, 169, 158, 190, 158, 206, 158, 168, 88, 82, 159, 18, 81, 24, 81, 20, 81, 16, 81, 21, 81, 128, 81, 170, 81, 221, 81, 145, 82, 147, 82, 243, 82, 89, 86, 107, 86, 121, 86, 105, 86, 100, 86, 120, 86, 106, 86, 104, 86, 101, 86, 113, 86, 111, 86, 108, 86, 98, 86, 118, 86, 193, 88, 190, 88, 199, 88, 197, 88, 110, 89, 29, 91, 52, 91, 120, 91, 240, 91, 14, 92, 74, 95, 178, 97, 145, 97, 169, 97, 138, 97, 205, 97, 182, 97, 190, 97, 202, 97, 200, 97, 48, 98, 197, 100, 193, 100, 203, 100, 187, 100, 188, 100, 218, 100, 196, 100, 199, 100, 194, 100, 205, 100, 191, 100, 210, 100, 212, 100, 190, 100, 116, 101, 198, 102, 201, 102, 185, 102, 196, 102, 199, 102, 184, 102, 61, 106, 56, 106, 58, 106, 89, 106, 107, 106, 88, 106, 57, 106, 68, 106, 98, 106, 97, 106, 75, 106, 71, 106, 53, 106, 95, 106, 72, 106, 89, 107, 119, 107, 5, 108, 194, 111, 177, 111, 161, 111, 195, 111, 164, 111, 193, 111, 167, 111, 179, 111, 192, 111, 185, 111, 182, 111, 166, 111, 160, 111, 180, 111, 190, 113, 201, 113, 208, 113, 210, 113, 200, 113, 213, 113, 185, 113, 206, 113, 217, 113, 220, 113, 195, 113, 196, 113, 104, 115, 156, 116, 163, 116, 152, 116, 159, 116, 158, 116, 226, 116, 12, 117, 13, 117, 52, 118, 56, 118, 58, 118, 231, 118, 229, 118, 160, 119, 158, 119, 159, 119, 165, 119, 232, 120, 218, 120, 236, 120, 231, 120, 166, 121, 77, 122, 78, 122, 70, 122, 76, 122, 75, 122, 186, 122, 217, 123, 17, 124, 201, 123, 228, 123, 219, 123, 225, 123, 233, 123, 230, 123, 213, 124, 214, 124, 10, 126, 17, 126, 8, 126, 27, 126, 35, 126, 30, 126, 29, 126, 9, 126, 16, 126, 121, 127, 178, 127, 240, 127, 241, 127, 238, 127, 40, 128, 179, 129, 169, 129, 168, 129, 251, 129, 8, 130, 88, 130, 89, 130, 74, 133, 89, 133, 72, 133, 104, 133, 105, 133, 67, 133, 73, 133, 109, 133, 106, 133, 94, 133, 131, 135, 159, 135, 158, 135, 162, 135, 141, 135, 97, 136, 42, 137, 50, 137, 37, 137, 43, 137, 33, 137, 170, 137, 166, 137, 230, 138, 250, 138, 235, 138, 241, 138, 0, 139, 220, 138, 231, 138, 238, 138, 254, 138, 1, 139, 2, 139, 247, 138, 237, 138, 243, 138, 246, 138, 252, 138, 107, 140, 109, 140, 147, 140, 244, 140, 68, 142, 49, 142, 52, 142, 66, 142, 57, 142, 53, 142, 59, 143, 47, 143, 56, 143, 51, 143, 168, 143, 166, 143, 117, 144, 116, 144, 120, 144, 114, 144, 124, 144, 122, 144, 52, 145, 146, 145, 32, 147, 54, 147, 248, 146, 51, 147, 47, 147, 34, 147, 252, 146, 43, 147, 4, 147, 26, 147, 16, 147, 38, 147, 33, 147, 21, 147, 46, 147, 25, 147, 187, 149, 167, 150, 168, 150, 170, 150, 213, 150, 14, 151, 17, 151, 22, 151, 13, 151, 19, 151, 15, 151, 91, 151, 92, 151, 102, 151, 152, 151, 48, 152, 56, 152, 59, 152, 55, 152, 45, 152, 57, 152, 36, 152, 16, 153, 40, 153, 30, 153, 27, 153, 33, 153, 26, 153, 237, 153, 226, 153, 241, 153, 184, 154, 188, 154, 251, 154, 237, 154, 40, 155, 145, 155, 21, 157, 35, 157, 38, 157, 40, 157, 18, 157, 27, 157, 216, 158, 212, 158, 141, 159, 156, 159, 42, 81, 31, 81, 33, 81, 50, 81, 245, 82, 142, 86, 128, 86, 144, 86, 133, 86, 135, 86, 143, 86, 213, 88, 211, 88, 209, 88, 206, 88, 48, 91, 42, 91, 36, 91, 122, 91, 55, 92, 104, 92, 188, 93, 186, 93, 189, 93, 184, 93, 107, 94, 76, 95, 189, 95, 201, 97, 194, 97, 199, 97, 230, 97, 203, 97, 50, 98, 52, 98, 206, 100, 202, 100, 216, 100, 224, 100, 240, 100, 230, 100, 236, 100, 241, 100, 226, 100, 237, 100, 130, 101, 131, 101, 217, 102, 214, 102, 128, 106, 148, 106, 132, 106, 162, 106, 156, 106, 219, 106, 163, 106, 126, 106, 151, 106, 144, 106, 160, 106, 92, 107, 174, 107, 218, 107, 8, 108, 216, 111, 241, 111, 223, 111, 224, 111, 219, 111, 228, 111, 235, 111, 239, 111, 128, 111, 236, 111, 225, 111, 233, 111, 213, 111, 238, 111, 240, 111, 231, 113, 223, 113, 238, 113, 230, 113, 229, 113, 237, 113, 236, 113, 244, 113, 224, 113, 53, 114, 70, 114, 112, 115, 114, 115, 169, 116, 176, 116, 166, 116, 168, 116, 70, 118, 66, 118, 76, 118, 234, 118, 179, 119, 170, 119, 176, 119, 172, 119, 167, 119, 173, 119, 239, 119, 247, 120, 250, 120, 244, 120, 239, 120, 1, 121, 167, 121, 170, 121, 87, 122, 191, 122, 7, 124, 13, 124, 254, 123, 247, 123, 12, 124, 224, 123, 224, 124, 220, 124, 222, 124, 226, 124, 223, 124, 217, 124, 221, 124, 46, 126, 62, 126, 70, 126, 55, 126, 50, 126, 67, 126, 43, 126, 61, 126, 49, 126, 69, 126, 65, 126, 52, 126, 57, 126, 72, 126, 53, 126, 63, 126, 47, 126, 68, 127, 243, 127, 252, 127, 113, 128, 114, 128, 112, 128, 111, 128, 115, 128, 198, 129, 195, 129, 186, 129, 194, 129, 192, 129, 191, 129, 189, 129, 201, 129, 190, 129, 232, 129, 9, 130, 113, 130, 170, 133, 132, 133, 126, 133, 156, 133, 145, 133, 148, 133, 175, 133, 155, 133, 135, 133, 168, 133, 138, 133, 103, 134, 192, 135, 209, 135, 179, 135, 210, 135, 198, 135, 171, 135, 187, 135, 186, 135, 200, 135, 203, 135, 59, 137, 54, 137, 68, 137, 56, 137, 61, 137, 172, 137, 14, 139, 23, 139, 25, 139, 27, 139, 10, 139, 32, 139, 29, 139, 4, 139, 16, 139, 65, 140, 63, 140, 115, 140, 250, 140, 253, 140, 252, 140, 248, 140, 251, 140, 168, 141, 73, 142, 75, 142, 72, 142, 74, 142, 68, 143, 62, 143, 66, 143, 69, 143, 63, 143, 127, 144, 125, 144, 132, 144, 129, 144, 130, 144, 128, 144, 57, 145, 163, 145, 158, 145, 156, 145, 77, 147, 130, 147, 40, 147, 117, 147, 74, 147, 101, 147, 75, 147, 24, 147, 126, 147, 108, 147, 91, 147, 112, 147, 90, 147, 84, 147, 202, 149, 203, 149, 204, 149, 200, 149, 198, 149, 177, 150, 184, 150, 214, 150, 28, 151, 30, 151, 160, 151, 211, 151, 70, 152, 182, 152, 53, 153, 1, 154, 255, 153, 174, 155, 171, 155, 170, 155, 173, 155, 59, 157, 63, 157, 139, 158, 207, 158, 222, 158, 220, 158, 221, 158, 219, 158, 62, 159, 75, 159, 226, 83, 149, 86, 174, 86, 217, 88, 216, 88, 56, 91, 93, 95, 227, 97, 51, 98, 244, 100, 242, 100, 254, 100, 6, 101, 250, 100, 251, 100, 247, 100, 183, 101, 220, 102, 38, 103, 179, 106, 172, 106, 195, 106, 187, 106, 184, 106, 194, 106, 174, 106, 175, 106, 95, 107, 120, 107, 175, 107, 9, 112, 11, 112, 254, 111, 6, 112, 250, 111, 17, 112, 15, 112, 251, 113, 252, 113, 254, 113, 248, 113, 119, 115, 117, 115, 167, 116, 191, 116, 21, 117, 86, 118, 88, 118, 82, 118, 189, 119, 191, 119, 187, 119, 188, 119, 14, 121, 174, 121, 97, 122, 98, 122, 96, 122, 196, 122, 197, 122, 43, 124, 39, 124, 42, 124, 30, 124, 35, 124, 33, 124, 231, 124, 84, 126, 85, 126, 94, 126, 90, 126, 97, 126, 82, 126, 89, 126, 72, 127, 249, 127, 251, 127, 119, 128, 118, 128, 205, 129, 207, 129, 10, 130, 207, 133, 169, 133, 205, 133, 208, 133, 201, 133, 176, 133, 186, 133, 185, 133, 166, 133, 239, 135, 236, 135, 242, 135, 224, 135, 134, 137, 178, 137, 244, 137, 40, 139, 57, 139, 44, 139, 43, 139, 80, 140, 5, 141, 89, 142, 99, 142, 102, 142, 100, 142, 95, 142, 85, 142, 192, 142, 73, 143, 77, 143, 135, 144, 131, 144, 136, 144, 171, 145, 172, 145, 208, 145, 148, 147, 138, 147, 150, 147, 162, 147, 179, 147, 174, 147, 172, 147, 176, 147, 152, 147, 154, 147, 151, 147, 212, 149, 214, 149, 208, 149, 213, 149, 226, 150, 220, 150, 217, 150, 219, 150, 222, 150, 36, 151, 163, 151, 166, 151, 173, 151, 249, 151, 77, 152, 79, 152, 76, 152, 78, 152, 83, 152, 186, 152, 62, 153, 63, 153, 61, 153, 46, 153, 165, 153, 14, 154, 193, 154, 3, 155, 6, 155, 79, 155, 78, 155, 77, 155, 202, 155, 201, 155, 253, 155, 200, 155, 192, 155, 81, 157, 93, 157, 96, 157, 224, 158, 21, 159, 44, 159, 51, 81, 165, 86, 222, 88, 223, 88, 226, 88, 245, 91, 144, 159, 236, 94, 242, 97, 247, 97, 246, 97, 245, 97, 0, 101, 15, 101, 224, 102, 221, 102, 229, 106, 221, 106, 218, 106, 211, 106, 27, 112, 31, 112, 40, 112, 26, 112, 29, 112, 21, 112, 24, 112, 6, 114, 13, 114, 88, 114, 162, 114, 120, 115, 122, 115, 189, 116, 202, 116, 227, 116, 135, 117, 134, 117, 95, 118, 97, 118, 199, 119, 25, 121, 177, 121, 107, 122, 105, 122, 62, 124, 63, 124, 56, 124, 61, 124, 55, 124, 64, 124, 107, 126, 109, 126, 121, 126, 105, 126, 106, 126, 133, 127, 115, 126, 182, 127, 185, 127, 184, 127, 216, 129, 233, 133, 221, 133, 234, 133, 213, 133, 228, 133, 229, 133, 247, 133, 251, 135, 5, 136, 13, 136, 249, 135, 254, 135, 96, 137, 95, 137, 86, 137, 94, 137, 65, 139, 92, 139, 88, 139, 73, 139, 90, 139, 78, 139, 79, 139, 70, 139, 89, 139, 8, 141, 10, 141, 124, 142, 114, 142, 135, 142, 118, 142, 108, 142, 122, 142, 116, 142, 84, 143, 78, 143, 173, 143, 138, 144, 139, 144, 177, 145, 174, 145, 225, 147, 209, 147, 223, 147, 195, 147, 200, 147, 220, 147, 221, 147, 214, 147, 226, 147, 205, 147, 216, 147, 228, 147, 215, 147, 232, 147, 220, 149, 180, 150, 227, 150, 42, 151, 39, 151, 97, 151, 220, 151, 251, 151, 94, 152, 88, 152, 91, 152, 188, 152, 69, 153, 73, 153, 22, 154, 25, 154, 13, 155, 232, 155, 231, 155, 214, 155, 219, 155, 137, 157, 97, 157, 114, 157, 106, 157, 108, 157, 146, 158, 151, 158, 147, 158, 180, 158, 248, 82, 168, 86, 183, 86, 182, 86, 180, 86, 188, 86, 228, 88, 64, 91, 67, 91, 125, 91, 246, 91, 201, 93, 248, 97, 250, 97, 24, 101, 20, 101, 25, 101, 230, 102, 39, 103, 236, 106, 62, 112, 48, 112, 50, 112, 16, 114, 123, 115, 207, 116, 98, 118, 101, 118, 38, 121, 42, 121, 44, 121, 43, 121, 199, 122, 246, 122, 76, 124, 67, 124, 77, 124, 239, 124, 240, 124, 174, 143, 125, 126, 124, 126, 130, 126, 76, 127, 0, 128, 218, 129, 102, 130, 251, 133, 249, 133, 17, 134, 250, 133, 6, 134, 11, 134, 7, 134, 10, 134, 20, 136, 21, 136, 100, 137, 186, 137, 248, 137, 112, 139, 108, 139, 102, 139, 111, 139, 95, 139, 107, 139, 15, 141, 13, 141, 137, 142, 129, 142, 133, 142, 130, 142, 180, 145, 203, 145, 24, 148, 3, 148, 253, 147, 225, 149, 48, 151, 196, 152, 82, 153, 81, 153, 168, 153, 43, 154, 48, 154, 55, 154, 53, 154, 19, 156, 13, 156, 121, 158, 181, 158, 232, 158, 47, 159, 95, 159, 99, 159, 97, 159, 55, 81, 56, 81, 193, 86, 192, 86, 194, 86, 20, 89, 108, 92, 205, 93, 252, 97, 254, 97, 29, 101, 28, 101, 149, 101, 233, 102, 251, 106, 4, 107, 250, 106, 178, 107, 76, 112, 27, 114, 167, 114, 214, 116, 212, 116, 105, 118, 211, 119, 80, 124, 143, 126, 140, 126, 188, 127, 23, 134, 45, 134, 26, 134, 35, 136, 34, 136, 33, 136, 31, 136, 106, 137, 108, 137, 189, 137, 116, 139, 119, 139, 125, 139, 19, 141, 138, 142, 141, 142, 139, 142, 95, 143, 175, 143, 186, 145, 46, 148, 51, 148, 53, 148, 58, 148, 56, 148, 50, 148, 43, 148, 226, 149, 56, 151, 57, 151, 50, 151, 255, 151, 103, 152, 101, 152, 87, 153, 69, 154, 67, 154, 64, 154, 62, 154, 207, 154, 84, 155, 81, 155, 45, 156, 37, 156, 175, 157, 180, 157, 194, 157, 184, 157, 157, 158, 239, 158, 25, 159, 92, 159, 102, 159, 103, 159, 60, 81, 59, 81, 200, 86, 202, 86, 201, 86, 127, 91, 212, 93, 210, 93, 78, 95, 255, 97, 36, 101, 10, 107, 97, 107, 81, 112, 88, 112, 128, 115, 228, 116, 138, 117, 110, 118, 108, 118, 179, 121, 96, 124, 95, 124, 126, 128, 125, 128, 223, 129, 114, 137, 111, 137, 252, 137, 128, 139, 22, 141, 23, 141, 145, 142, 147, 142, 97, 143, 72, 145, 68, 148, 81, 148, 82, 148, 61, 151, 62, 151, 195, 151, 193, 151, 107, 152, 85, 153, 85, 154, 77, 154, 210, 154, 26, 155, 73, 156, 49, 156, 62, 156, 59, 156, 211, 157, 215, 157, 52, 159, 108, 159, 106, 159, 148, 159, 204, 86, 214, 93, 0, 98, 35, 101, 43, 101, 42, 101, 236, 102, 16, 107, 218, 116, 202, 122, 100, 124, 99, 124, 101, 124, 147, 126, 150, 126, 148, 126, 226, 129, 56, 134, 63, 134, 49, 136, 138, 139, 144, 144, 143, 144, 99, 148, 96, 148, 100, 148, 104, 151, 111, 152, 92, 153, 90, 154, 91, 154, 87, 154, 211, 154, 212, 154, 209, 154, 84, 156, 87, 156, 86, 156, 229, 157, 159, 158, 244, 158, 209, 86, 233, 88, 44, 101, 94, 112, 113, 118, 114, 118, 215, 119, 80, 127, 136, 127, 54, 136, 57, 136, 98, 136, 147, 139, 146, 139, 150, 139, 119, 130, 27, 141, 192, 145, 106, 148, 66, 151, 72, 151, 68, 151, 198, 151, 112, 152, 95, 154, 34, 155, 88, 155, 95, 156, 249, 157, 250, 157, 124, 158, 125, 158, 7, 159, 119, 159, 114, 159, 243, 94, 22, 107, 99, 112, 108, 124, 110, 124, 59, 136, 192, 137, 161, 142, 193, 145, 114, 148, 112, 148, 113, 152, 94, 153, 214, 154, 35, 155, 204, 158, 100, 112, 218, 119, 154, 139, 119, 148, 201, 151, 98, 154, 101, 154, 156, 126, 156, 139, 170, 142, 197, 145, 125, 148, 126, 148, 124, 148, 119, 156, 120, 156, 247, 158, 84, 140, 127, 148, 26, 158, 40, 114, 106, 154, 49, 155, 27, 158, 30, 158, 114, 124, 66, 78, 92, 78, 245, 81, 26, 83, 130, 83, 7, 78, 12, 78, 71, 78, 141, 78, 215, 86, 110, 92, 115, 95, 15, 78, 135, 81, 14, 78, 46, 78, 147, 78, 194, 78, 201, 78, 200, 78, 152, 81, 252, 82, 108, 83, 185, 83, 32, 87, 3, 89, 44, 89, 16, 92, 255, 93, 225, 101, 179, 107, 204, 107, 20, 108, 63, 114, 49, 78, 60, 78, 232, 78, 220, 78, 233, 78, 225, 78, 221, 78, 218, 78, 12, 82, 28, 83, 76, 83, 34, 87, 35, 87, 23, 89, 47, 89, 129, 91, 132, 91, 18, 92, 59, 92, 116, 92, 115, 92, 4, 94, 128, 94, 130, 94, 201, 95, 9, 98, 80, 98, 21, 108, 54, 108, 67, 108, 63, 108, 59, 108, 174, 114, 176, 114, 138, 115, 184, 121, 138, 128, 30, 150, 14, 79, 24, 79, 44, 79, 245, 78, 20, 79, 241, 78, 0, 79, 247, 78, 8, 79, 29, 79, 2, 79, 5, 79, 34, 79, 19, 79, 4, 79, 244, 78, 18, 79, 177, 81, 19, 82, 9, 82, 16, 82, 166, 82, 34, 83, 31, 83, 77, 83, 138, 83, 7, 84, 225, 86, 223, 86, 46, 87, 42, 87, 52, 87, 60, 89, 128, 89, 124, 89, 133, 89, 123, 89, 126, 89, 119, 89, 127, 89, 86, 91, 21, 92, 37, 92, 124, 92, 122, 92, 123, 92, 126, 92, 223, 93, 117, 94, 132, 94, 2, 95, 26, 95, 116, 95, 213, 95, 212, 95, 207, 95, 92, 98, 94, 98, 100, 98, 97, 98, 102, 98, 98, 98, 89, 98, 96, 98, 90, 98, 101, 98, 239, 101, 238, 101, 62, 103, 57, 103, 56, 103, 59, 103, 58, 103, 63, 103, 60, 103, 51, 103, 24, 108, 70, 108, 82, 108, 92, 108, 79, 108, 74, 108, 84, 108, 75, 108, 76, 108, 113, 112, 94, 114, 180, 114, 181, 114, 142, 115, 42, 117, 127, 118, 117, 122, 81, 127, 120, 130, 124, 130, 128, 130, 125, 130, 127, 130, 77, 134, 126, 137, 153, 144, 151, 144, 152, 144, 155, 144, 148, 144, 34, 150, 36, 150, 32, 150, 35, 150, 86, 79, 59, 79, 98, 79, 73, 79, 83, 79, 100, 79, 62, 79, 103, 79, 82, 79, 95, 79, 65, 79, 88, 79, 45, 79, 51, 79, 63, 79, 97, 79, 143, 81, 185, 81, 28, 82, 30, 82, 33, 82, 173, 82, 174, 82, 9, 83, 99, 83, 114, 83, 142, 83, 143, 83, 48, 84, 55, 84, 42, 84, 84, 84, 69, 84, 25, 84, 28, 84, 37, 84, 24, 84, 61, 84, 79, 84, 65, 84, 40, 84, 36, 84, 71, 84, 238, 86, 231, 86, 229, 86, 65, 87, 69, 87, 76, 87, 73, 87, 75, 87, 82, 87, 6, 89, 64, 89, 166, 89, 152, 89, 160, 89, 151, 89, 142, 89, 162, 89, 144, 89, 143, 89, 167, 89, 161, 89, 142, 91, 146, 91, 40, 92, 42, 92, 141, 92, 143, 92, 136, 92, 139, 92, 137, 92, 146, 92, 138, 92, 134, 92, 147, 92, 149, 92, 224, 93, 10, 94, 14, 94, 139, 94, 137, 94, 140, 94, 136, 94, 141, 94, 5, 95, 29, 95, 120, 95, 118, 95, 210, 95, 209, 95, 208, 95, 237, 95, 232, 95, 238, 95, 243, 95, 225, 95, 228, 95, 227, 95, 250, 95, 239, 95, 247, 95, 251, 95, 0, 96, 244, 95, 58, 98, 131, 98, 140, 98, 142, 98, 143, 98, 148, 98, 135, 98, 113, 98, 123, 98, 122, 98, 112, 98, 129, 98, 136, 98, 119, 98, 125, 98, 114, 98, 116, 98, 55, 101, 240, 101, 244, 101, 243, 101, 242, 101, 245, 101, 69, 103, 71, 103, 89, 103, 85, 103, 76, 103, 72, 103, 93, 103, 77, 103, 90, 103, 75, 103, 208, 107, 25, 108, 26, 108, 120, 108, 103, 108, 107, 108, 132, 108, 139, 108, 143, 108, 113, 108, 111, 108, 105, 108, 154, 108, 109, 108, 135, 108, 149, 108, 156, 108, 102, 108, 115, 108, 101, 108, 123, 108, 142, 108, 116, 112, 122, 112, 99, 114, 191, 114, 189, 114, 195, 114, 198, 114, 193, 114, 186, 114, 197, 114, 149, 115, 151, 115, 147, 115, 148, 115, 146, 115, 58, 117, 57, 117, 148, 117, 149, 117, 129, 118, 61, 121, 52, 128, 149, 128, 153, 128, 144, 128, 146, 128, 156, 128, 144, 130, 143, 130, 133, 130, 142, 130, 145, 130, 147, 130, 138, 130, 131, 130, 132, 130, 120, 140, 201, 143, 191, 143, 159, 144, 161, 144, 165, 144, 158, 144, 167, 144, 160, 144, 48, 150, 40, 150, 47, 150, 45, 150, 51, 78, 152, 79, 124, 79, 133, 79, 125, 79, 128, 79, 135, 79, 118, 79, 116, 79, 137, 79, 132, 79, 119, 79, 76, 79, 151, 79, 106, 79, 154, 79, 121, 79, 129, 79, 120, 79, 144, 79, 156, 79, 148, 79, 158, 79, 146, 79, 130, 79, 149, 79, 107, 79, 110, 79, 158, 81, 188, 81, 190, 81, 53, 82, 50, 82, 51, 82, 70, 82, 49, 82, 188, 82, 10, 83, 11, 83, 60, 83, 146, 83, 148, 83, 135, 84, 127, 84, 129, 84, 145, 84, 130, 84, 136, 84, 107, 84, 122, 84, 126, 84, 101, 84, 108, 84, 116, 84, 102, 84, 141, 84, 111, 84, 97, 84, 96, 84, 152, 84, 99, 84, 103, 84, 100, 84, 247, 86, 249, 86, 111, 87, 114, 87, 109, 87, 107, 87, 113, 87, 112, 87, 118, 87, 128, 87, 117, 87, 123, 87, 115, 87, 116, 87, 98, 87, 104, 87, 125, 87, 12, 89, 69, 89, 181, 89, 186, 89, 207, 89, 206, 89, 178, 89, 204, 89, 193, 89, 182, 89, 188, 89, 195, 89, 214, 89, 177, 89, 189, 89, 192, 89, 200, 89, 180, 89, 199, 89, 98, 91, 101, 91, 147, 91, 149, 91, 68, 92, 71, 92, 174, 92, 164, 92, 160, 92, 181, 92, 175, 92, 168, 92, 172, 92, 159, 92, 163, 92, 173, 92, 162, 92, 170, 92, 167, 92, 157, 92, 165, 92, 182, 92, 176, 92, 166, 92, 23, 94, 20, 94, 25, 94, 40, 95, 34, 95, 35, 95, 36, 95, 84, 95, 130, 95, 126, 95, 125, 95, 222, 95, 229, 95, 45, 96, 38, 96, 25, 96, 50, 96, 11, 96, 52, 96, 10, 96, 23, 96, 51, 96, 26, 96, 30, 96, 44, 96, 34, 96, 13, 96, 16, 96, 46, 96, 19, 96, 17, 96, 12, 96, 9, 96, 28, 96, 20, 98, 61, 98, 173, 98, 180, 98, 209, 98, 190, 98, 170, 98, 182, 98, 202, 98, 174, 98, 179, 98, 175, 98, 187, 98, 169, 98, 176, 98, 184, 98, 61, 101, 168, 101, 187, 101, 9, 102, 252, 101, 4, 102, 18, 102, 8, 102, 251, 101, 3, 102, 11, 102, 13, 102, 5, 102, 253, 101, 17, 102, 16, 102, 246, 102, 10, 103, 133, 103, 108, 103, 142, 103, 146, 103, 118, 103, 123, 103, 152, 103, 134, 103, 132, 103, 116, 103, 141, 103, 140, 103, 122, 103, 159, 103, 145, 103, 153, 103, 131, 103, 125, 103, 129, 103, 120, 103, 121, 103, 148, 103, 37, 107, 128, 107, 126, 107, 222, 107, 29, 108, 147, 108, 236, 108, 235, 108, 238, 108, 217, 108, 182, 108, 212, 108, 173, 108, 231, 108, 183, 108, 208, 108, 194, 108, 186, 108, 195, 108, 198, 108, 237, 108, 242, 108, 210, 108, 221, 108, 180, 108, 138, 108, 157, 108, 128, 108, 222, 108, 192, 108, 48, 109, 205, 108, 199, 108, 176, 108, 249, 108, 207, 108, 233, 108, 209, 108, 148, 112, 152, 112, 133, 112, 147, 112, 134, 112, 132, 112, 145, 112, 150, 112, 130, 112, 154, 112, 131, 112, 106, 114, 214, 114, 203, 114, 216, 114, 201, 114, 220, 114, 210, 114, 212, 114, 218, 114, 204, 114, 209, 114, 164, 115, 161, 115, 173, 115, 166, 115, 162, 115, 160, 115, 172, 115, 157, 115, 221, 116, 232, 116, 63, 117, 64, 117, 62, 117, 140, 117, 152, 117, 175, 118, 243, 118, 241, 118, 240, 118, 245, 118, 248, 119, 252, 119, 249, 119, 251, 119, 250, 119, 247, 119, 66, 121, 63, 121, 197, 121, 120, 122, 123, 122, 251, 122, 117, 124, 253, 124, 53, 128, 143, 128, 174, 128, 163, 128, 184, 128, 181, 128, 173, 128, 32, 130, 160, 130, 192, 130, 171, 130, 154, 130, 152, 130, 155, 130, 181, 130, 167, 130, 174, 130, 188, 130, 158, 130, 186, 130, 180, 130, 168, 130, 161, 130, 169, 130, 194, 130, 164, 130, 195, 130, 182, 130, 162, 130, 112, 134, 111, 134, 109, 134, 110, 134, 86, 140, 210, 143, 203, 143, 211, 143, 205, 143, 214, 143, 213, 143, 215, 143, 178, 144, 180, 144, 175, 144, 179, 144, 176, 144, 57, 150, 61, 150, 60, 150, 58, 150, 67, 150, 205, 79, 197, 79, 211, 79, 178, 79, 201, 79, 203, 79, 193, 79, 212, 79, 220, 79, 217, 79, 187, 79, 179, 79, 219, 79, 199, 79, 214, 79, 186, 79, 192, 79, 185, 79, 236, 79, 68, 82, 73, 82, 192, 82, 194, 82, 61, 83, 124, 83, 151, 83, 150, 83, 153, 83, 152, 83, 186, 84, 161, 84, 173, 84, 165, 84, 207, 84, 195, 84, 13, 131, 183, 84, 174, 84, 214, 84, 182, 84, 197, 84, 198, 84, 160, 84, 112, 84, 188, 84, 162, 84, 190, 84, 114, 84, 222, 84, 176, 84, 181, 87, 158, 87, 159, 87, 164, 87, 140, 87, 151, 87, 157, 87, 155, 87, 148, 87, 152, 87, 143, 87, 153, 87, 165, 87, 154, 87, 149, 87, 244, 88, 13, 89, 83, 89, 225, 89, 222, 89, 238, 89, 0, 90, 241, 89, 221, 89, 250, 89, 253, 89, 252, 89, 246, 89, 228, 89, 242, 89, 247, 89, 219, 89, 233, 89, 243, 89, 245, 89, 224, 89, 254, 89, 244, 89, 237, 89, 168, 91, 76, 92, 208, 92, 216, 92, 204, 92, 215, 92, 203, 92, 219, 92, 222, 92, 218, 92, 201, 92, 199, 92, 202, 92, 214, 92, 211, 92, 212, 92, 207, 92, 200, 92, 198, 92, 206, 92, 223, 92, 248, 92, 249, 93, 33, 94, 34, 94, 35, 94, 32, 94, 36, 94, 176, 94, 164, 94, 162, 94, 155, 94, 163, 94, 165, 94, 7, 95, 46, 95, 86, 95, 134, 95, 55, 96, 57, 96, 84, 96, 114, 96, 94, 96, 69, 96, 83, 96, 71, 96, 73, 96, 91, 96, 76, 96, 64, 96, 66, 96, 95, 96, 36, 96, 68, 96, 88, 96, 102, 96, 110, 96, 66, 98, 67, 98, 207, 98, 13, 99, 11, 99, 245, 98, 14, 99, 3, 99, 235, 98, 249, 98, 15, 99, 12, 99, 248, 98, 246, 98, 0, 99, 19, 99, 20, 99, 250, 98, 21, 99, 251, 98, 240, 98, 65, 101, 67, 101, 170, 101, 191, 101, 54, 102, 33, 102, 50, 102, 53, 102, 28, 102, 38, 102, 34, 102, 51, 102, 43, 102, 58, 102, 29, 102, 52, 102, 57, 102, 46, 102, 15, 103, 16, 103, 193, 103, 242, 103, 200, 103, 186, 103, 220, 103, 187, 103, 248, 103, 216, 103, 192, 103, 183, 103, 197, 103, 235, 103, 228, 103, 223, 103, 181, 103, 205, 103, 179, 103, 247, 103, 246, 103, 238, 103, 227, 103, 194, 103, 185, 103, 206, 103, 231, 103, 240, 103, 178, 103, 252, 103, 198, 103, 237, 103, 204, 103, 174, 103, 230, 103, 219, 103, 250, 103, 201, 103, 202, 103, 195, 103, 234, 103, 203, 103, 40, 107, 130, 107, 132, 107, 182, 107, 214, 107, 216, 107, 224, 107, 32, 108, 33, 108, 40, 109, 52, 109, 45, 109, 31, 109, 60, 109, 63, 109, 18, 109, 10, 109, 218, 108, 51, 109, 4, 109, 25, 109, 58, 109, 26, 109, 17, 109, 0, 109, 29, 109, 66, 109, 1, 109, 24, 109, 55, 109, 3, 109, 15, 109, 64, 109, 7, 109, 32, 109, 44, 109, 8, 109, 34, 109, 9, 109, 16, 109, 183, 112, 159, 112, 190, 112, 177, 112, 176, 112, 161, 112, 180, 112, 181, 112, 169, 112, 65, 114, 73, 114, 74, 114, 108, 114, 112, 114, 115, 114, 110, 114, 202, 114, 228, 114, 232, 114, 235, 114, 223, 114, 234, 114, 230, 114, 227, 114, 133, 115, 204, 115, 194, 115, 200, 115, 197, 115, 185, 115, 182, 115, 181, 115, 180, 115, 235, 115, 191, 115, 199, 115, 190, 115, 195, 115, 198, 115, 184, 115, 203, 115, 236, 116, 238, 116, 46, 117, 71, 117, 72, 117, 167, 117, 170, 117, 121, 118, 196, 118, 8, 119, 3, 119, 4, 119, 5, 119, 10, 119, 247, 118, 251, 118, 250, 118, 231, 119, 232, 119, 6, 120, 17, 120, 18, 120, 5, 120, 16, 120, 15, 120, 14, 120, 9, 120, 3, 120, 19, 120, 74, 121, 76, 121, 75, 121, 69, 121, 68, 121, 213, 121, 205, 121, 207, 121, 214, 121, 206, 121, 128, 122, 126, 122, 209, 122, 0, 123, 1, 123, 122, 124, 120, 124, 121, 124, 127, 124, 128, 124, 129, 124, 3, 125, 8, 125, 1, 125, 88, 127, 145, 127, 141, 127, 190, 127, 7, 128, 14, 128, 15, 128, 20, 128, 55, 128, 216, 128, 199, 128, 224, 128, 209, 128, 200, 128, 194, 128, 208, 128, 197, 128, 227, 128, 217, 128, 220, 128, 202, 128, 213, 128, 201, 128, 207, 128, 215, 128, 230, 128, 205, 128, 255, 129, 33, 130, 148, 130, 217, 130, 254, 130, 249, 130, 7, 131, 232, 130, 0, 131, 213, 130, 58, 131, 235, 130, 214, 130, 244, 130, 236, 130, 225, 130, 242, 130, 245, 130, 12, 131, 251, 130, 246, 130, 240, 130, 234, 130, 228, 130, 224, 130, 250, 130, 243, 130, 237, 130, 119, 134, 116, 134, 124, 134, 115, 134, 65, 136, 78, 136, 103, 136, 106, 136, 105, 136, 211, 137, 4, 138, 7, 138, 114, 141, 227, 143, 225, 143, 238, 143, 224, 143, 241, 144, 189, 144, 191, 144, 213, 144, 197, 144, 190, 144, 199, 144, 203, 144, 200, 144, 212, 145, 211, 145, 84, 150, 79, 150, 81, 150, 83, 150, 74, 150, 78, 150, 30, 80, 5, 80, 7, 80, 19, 80, 34, 80, 48, 80, 27, 80, 245, 79, 244, 79, 51, 80, 55, 80, 44, 80, 246, 79, 247, 79, 23, 80, 28, 80, 32, 80, 39, 80, 53, 80, 47, 80, 49, 80, 14, 80, 90, 81, 148, 81, 147, 81, 202, 81, 196, 81, 197, 81, 200, 81, 206, 81, 97, 82, 90, 82, 82, 82, 94, 82, 95, 82, 85, 82, 98, 82, 205, 82, 14, 83, 158, 83, 38, 85, 226, 84, 23, 85, 18, 85, 231, 84, 243, 84, 228, 84, 26, 85, 255, 84, 4, 85, 8, 85, 235, 84, 17, 85, 5, 85, 241, 84, 10, 85, 251, 84, 247, 84, 248, 84, 224, 84, 14, 85, 3, 85, 11, 85, 1, 87, 2, 87, 204, 87, 50, 88, 213, 87, 210, 87, 186, 87, 198, 87, 189, 87, 188, 87, 184, 87, 182, 87, 191, 87, 199, 87, 208, 87, 185, 87, 193, 87, 14, 89, 74, 89, 25, 90, 22, 90, 45, 90, 46, 90, 21, 90, 15, 90, 23, 90, 10, 90, 30, 90, 51, 90, 108, 91, 167, 91, 173, 91, 172, 91, 3, 92, 86, 92, 84, 92, 236, 92, 255, 92, 238, 92, 241, 92, 247, 92, 0, 93, 249, 92, 41, 94, 40, 94, 168, 94, 174, 94, 170, 94, 172, 94, 51, 95, 48, 95, 103, 95, 93, 96, 90, 96, 103, 96, 65, 96, 162, 96, 136, 96, 128, 96, 146, 96, 129, 96, 157, 96, 131, 96, 149, 96, 155, 96, 151, 96, 135, 96, 156, 96, 142, 96, 25, 98, 70, 98, 242, 98, 16, 99, 86, 99, 44, 99, 68, 99, 69, 99, 54, 99, 67, 99, 228, 99, 57, 99, 75, 99, 74, 99, 60, 99, 41, 99, 65, 99, 52, 99, 88, 99, 84, 99, 89, 99, 45, 99, 71, 99, 51, 99, 90, 99, 81, 99, 56, 99, 87, 99, 64, 99, 72, 99, 74, 101, 70, 101, 198, 101, 195, 101, 196, 101, 194, 101, 74, 102, 95, 102, 71, 102, 81, 102, 18, 103, 19, 103, 31, 104, 26, 104, 73, 104, 50, 104, 51, 104, 59, 104, 75, 104, 79, 104, 22, 104, 49, 104, 28, 104, 53, 104, 43, 104, 45, 104, 47, 104, 78, 104, 68, 104, 52, 104, 29, 104, 18, 104, 20, 104, 38, 104, 40, 104, 46, 104, 77, 104, 58, 104, 37, 104, 32, 104, 44, 107, 47, 107, 45, 107, 49, 107, 52, 107, 109, 107, 130, 128, 136, 107, 230, 107, 228, 107, 232, 107, 227, 107, 226, 107, 231, 107, 37, 108, 122, 109, 99, 109, 100, 109, 118, 109, 13, 109, 97, 109, 146, 109, 88, 109, 98, 109, 109, 109, 111, 109, 145, 109, 141, 109, 239, 109, 127, 109, 134, 109, 94, 109, 103, 109, 96, 109, 151, 109, 112, 109, 124, 109, 95, 109, 130, 109, 152, 109, 47, 109, 104, 109, 139, 109, 126, 109, 128, 109, 132, 109, 22, 109, 131, 109, 123, 109, 125, 109, 117, 109, 144, 109, 220, 112, 211, 112, 209, 112, 221, 112, 203, 112, 57, 127, 226, 112, 215, 112, 210, 112, 222, 112, 224, 112, 212, 112, 205, 112, 197, 112, 198, 112, 199, 112, 218, 112, 206, 112, 225, 112, 66, 114, 120, 114, 119, 114, 118, 114, 0, 115, 250, 114, 244, 114, 254, 114, 246, 114, 243, 114, 251, 114, 1, 115, 211, 115, 217, 115, 229, 115, 214, 115, 188, 115, 231, 115, 227, 115, 233, 115, 220, 115, 210, 115, 219, 115, 212, 115, 221, 115, 218, 115, 215, 115, 216, 115, 232, 115, 222, 116, 223, 116, 244, 116, 245, 116, 33, 117, 91, 117, 95, 117, 176, 117, 193, 117, 187, 117, 196, 117, 192, 117, 191, 117, 182, 117, 186, 117, 138, 118, 201, 118, 29, 119, 27, 119, 16, 119, 19, 119, 18, 119, 35, 119, 17, 119, 21, 119, 25, 119, 26, 119, 34, 119, 39, 119, 35, 120, 44, 120, 34, 120, 53, 120, 47, 120, 40, 120, 46, 120, 43, 120, 33, 120, 41, 120, 51, 120, 42, 120, 49, 120, 84, 121, 91, 121, 79, 121, 92, 121, 83, 121, 82, 121, 81, 121, 235, 121, 236, 121, 224, 121, 238, 121, 237, 121, 234, 121, 220, 121, 222, 121, 221, 121, 134, 122, 137, 122, 133, 122, 139, 122, 140, 122, 138, 122, 135, 122, 216, 122, 16, 123, 4, 123, 19, 123, 5, 123, 15, 123, 8, 123, 10, 123, 14, 123, 9, 123, 18, 123, 132, 124, 145, 124, 138, 124, 140, 124, 136, 124, 141, 124, 133, 124, 30, 125, 29, 125, 17, 125, 14, 125, 24, 125, 22, 125, 19, 125, 31, 125, 18, 125, 15, 125, 12, 125, 92, 127, 97, 127, 94, 127, 96, 127, 93, 127, 91, 127, 150, 127, 146, 127, 195, 127, 194, 127, 192, 127, 22, 128, 62, 128, 57, 128, 250, 128, 242, 128, 249, 128, 245, 128, 1, 129, 251, 128, 0, 129, 1, 130, 47, 130, 37, 130, 51, 131, 45, 131, 68, 131, 25, 131, 81, 131, 37, 131, 86, 131, 63, 131, 65, 131, 38, 131, 28, 131, 34, 131, 66, 131, 78, 131, 27, 131, 42, 131, 8, 131, 60, 131, 77, 131, 22, 131, 36, 131, 32, 131, 55, 131, 47, 131, 41, 131, 71, 131, 69, 131, 76, 131, 83, 131, 30, 131, 44, 131, 75, 131, 39, 131, 72, 131, 83, 134, 82, 134, 162, 134, 168, 134, 150, 134, 141, 134, 145, 134, 158, 134, 135, 134, 151, 134, 134, 134, 139, 134, 154, 134, 133, 134, 165, 134, 153, 134, 161, 134, 167, 134, 149, 134, 152, 134, 142, 134, 157, 134, 144, 134, 148, 134, 67, 136, 68, 136, 109, 136, 117, 136, 118, 136, 114, 136, 128, 136, 113, 136, 127, 136, 111, 136, 131, 136, 126, 136, 116, 136, 124, 136, 18, 138, 71, 140, 87, 140, 123, 140, 164, 140, 163, 140, 118, 141, 120, 141, 181, 141, 183, 141, 182, 141, 209, 142, 211, 142, 254, 143, 245, 143, 2, 144, 255, 143, 251, 143, 4, 144, 252, 143, 246, 143, 214, 144, 224, 144, 217, 144, 218, 144, 227, 144, 223, 144, 229, 144, 216, 144, 219, 144, 215, 144, 220, 144, 228, 144, 80, 145, 78, 145, 79, 145, 213, 145, 226, 145, 218, 145, 92, 150, 95, 150, 188, 150, 227, 152, 223, 154, 47, 155, 127, 78, 112, 80, 106, 80, 97, 80, 94, 80, 96, 80, 83, 80, 75, 80, 93, 80, 114, 80, 72, 80, 77, 80, 65, 80, 91, 80, 74, 80, 98, 80, 21, 80, 69, 80, 95, 80, 105, 80, 107, 80, 99, 80, 100, 80, 70, 80, 64, 80, 110, 80, 115, 80, 87, 80, 81, 80, 208, 81, 107, 82, 109, 82, 108, 82, 110, 82, 214, 82, 211, 82, 45, 83, 156, 83, 117, 85, 118, 85, 60, 85, 77, 85, 80, 85, 52, 85, 42, 85, 81, 85, 98, 85, 54, 85, 53, 85, 48, 85, 82, 85, 69, 85, 12, 85, 50, 85, 101, 85, 78, 85, 57, 85, 72, 85, 45, 85, 59, 85, 64, 85, 75, 85, 10, 87, 7, 87, 251, 87, 20, 88, 226, 87, 246, 87, 220, 87, 244, 87, 0, 88, 237, 87, 253, 87, 8, 88, 248, 87, 11, 88, 243, 87, 207, 87, 7, 88, 238, 87, 227, 87, 242, 87, 229, 87, 236, 87, 225, 87, 14, 88, 252, 87, 16, 88, 231, 87, 1, 88, 12, 88, 241, 87, 233, 87, 240, 87, 13, 88, 4, 88, 92, 89, 96, 90, 88, 90, 85, 90, 103, 90, 94, 90, 56, 90, 53, 90, 109, 90, 80, 90, 95, 90, 101, 90, 108, 90, 83, 90, 100, 90, 87, 90, 67, 90, 93, 90, 82, 90, 68, 90, 91, 90, 72, 90, 142, 90, 62, 90, 77, 90, 57, 90, 76, 90, 112, 90, 105, 90, 71, 90, 81, 90, 86, 90, 66, 90, 92, 90, 114, 91, 110, 91, 193, 91, 192, 91, 89, 92, 30, 93, 11, 93, 29, 93, 26, 93, 32, 93, 12, 93, 40, 93, 13, 93, 38, 93, 37, 93, 15, 93, 48, 93, 18, 93, 35, 93, 31, 93, 46, 93, 62, 94, 52, 94, 177, 94, 180, 94, 185, 94, 178, 94, 179, 94, 54, 95, 56, 95, 155, 95, 150, 95, 159, 95, 138, 96, 144, 96, 134, 96, 190, 96, 176, 96, 186, 96, 211, 96, 212, 96, 207, 96, 228, 96, 217, 96, 221, 96, 200, 96, 177, 96, 219, 96, 183, 96, 202, 96, 191, 96, 195, 96, 205, 96, 192, 96, 50, 99, 101, 99, 138, 99, 130, 99, 125, 99, 189, 99, 158, 99, 173, 99, 157, 99, 151, 99, 171, 99, 142, 99, 111, 99, 135, 99, 144, 99, 110, 99, 175, 99, 117, 99, 156, 99, 109, 99, 174, 99, 124, 99, 164, 99, 59, 99, 159, 99, 120, 99, 133, 99, 129, 99, 145, 99, 141, 99, 112, 99, 83, 101, 205, 101, 101, 102, 97, 102, 91, 102, 89, 102, 92, 102, 98, 102, 24, 103, 121, 104, 135, 104, 144, 104, 156, 104, 109, 104, 110, 104, 174, 104, 171, 104, 86, 105, 111, 104, 163, 104, 172, 104, 169, 104, 117, 104, 116, 104, 178, 104, 143, 104, 119, 104, 146, 104, 124, 104, 107, 104, 114, 104, 170, 104, 128, 104, 113, 104, 126, 104, 155, 104, 150, 104, 139, 104, 160, 104, 137, 104, 164, 104, 120, 104, 123, 104, 145, 104, 140, 104, 138, 104, 125, 104, 54, 107, 51, 107, 55, 107, 56, 107, 145, 107, 143, 107, 141, 107, 142, 107, 140, 107, 42, 108, 192, 109, 171, 109, 180, 109, 179, 109, 116, 110, 172, 109, 233, 109, 226, 109, 183, 109, 246, 109, 212, 109, 0, 110, 200, 109, 224, 109, 223, 109, 214, 109, 190, 109, 229, 109, 220, 109, 221, 109, 219, 109, 244, 109, 202, 109, 189, 109, 237, 109, 240, 109, 186, 109, 213, 109, 194, 109, 207, 109, 201, 109, 208, 109, 242, 109, 211, 109, 253, 109, 215, 109, 205, 109, 227, 109, 187, 109, 250, 112, 13, 113, 247, 112, 23, 113, 244, 112, 12, 113, 240, 112, 4, 113, 243, 112, 16, 113, 252, 112, 255, 112, 6, 113, 19, 113, 0, 113, 248, 112, 246, 112, 11, 113, 2, 113, 14, 113, 126, 114, 123, 114, 124, 114, 127, 114, 29, 115, 23, 115, 7, 115, 17, 115, 24, 115, 10, 115, 8, 115, 255, 114, 15, 115, 30, 115, 136, 115, 246, 115, 248, 115, 245, 115, 4, 116, 1, 116, 253, 115, 7, 116, 0, 116, 250, 115, 252, 115, 255, 115, 12, 116, 11, 116, 244, 115, 8, 116, 100, 117, 99, 117, 206, 117, 210, 117, 207, 117, 203, 117, 204, 117, 209, 117, 208, 117, 143, 118, 137, 118, 211, 118, 57, 119, 47, 119, 45, 119, 49, 119, 50, 119, 52, 119, 51, 119, 61, 119, 37, 119, 59, 119, 53, 119, 72, 120, 82, 120, 73, 120, 77, 120, 74, 120, 76, 120, 38, 120, 69, 120, 80, 120, 100, 121, 103, 121, 105, 121, 106, 121, 99, 121, 107, 121, 97, 121, 187, 121, 250, 121, 248, 121, 246, 121, 247, 121, 143, 122, 148, 122, 144, 122, 53, 123, 71, 123, 52, 123, 37, 123, 48, 123, 34, 123, 36, 123, 51, 123, 24, 123, 42, 123, 29, 123, 49, 123, 43, 123, 45, 123, 47, 123, 50, 123, 56, 123, 26, 123, 35, 123, 148, 124, 152, 124, 150, 124, 163, 124, 53, 125, 61, 125, 56, 125, 54, 125, 58, 125, 69, 125, 44, 125, 41, 125, 65, 125, 71, 125, 62, 125, 63, 125, 74, 125, 59, 125, 40, 125, 99, 127, 149, 127, 156, 127, 157, 127, 155, 127, 202, 127, 203, 127, 205, 127, 208, 127, 209, 127, 199, 127, 207, 127, 201, 127, 31, 128, 30, 128, 27, 128, 71, 128, 67, 128, 72, 128, 24, 129, 37, 129, 25, 129, 27, 129, 45, 129, 31, 129, 44, 129, 30, 129, 33, 129, 21, 129, 39, 129, 29, 129, 34, 129, 17, 130, 56, 130, 51, 130, 58, 130, 52, 130, 50, 130, 116, 130, 144, 131, 163, 131, 168, 131, 141, 131, 122, 131, 115, 131, 164, 131, 116, 131, 143, 131, 129, 131, 149, 131, 153, 131, 117, 131, 148, 131, 169, 131, 125, 131, 131, 131, 140, 131, 157, 131, 155, 131, 170, 131, 139, 131, 126, 131, 165, 131, 175, 131, 136, 131, 151, 131, 176, 131, 127, 131, 166, 131, 135, 131, 174, 131, 118, 131, 154, 131, 89, 134, 86, 134, 191, 134, 183, 134, 194, 134, 193, 134, 197, 134, 186, 134, 176, 134, 200, 134, 185, 134, 179, 134, 184, 134, 204, 134, 180, 134, 187, 134, 188, 134, 195, 134, 189, 134, 190, 134, 82, 136, 137, 136, 149, 136, 168, 136, 162, 136, 170, 136, 154, 136, 145, 136, 161, 136, 159, 136, 152, 136, 167, 136, 153, 136, 155, 136, 151, 136, 164, 136, 172, 136, 140, 136, 147, 136, 142, 136, 130, 137, 214, 137, 217, 137, 213, 137, 48, 138, 39, 138, 44, 138, 30, 138, 57, 140, 59, 140, 92, 140, 93, 140, 125, 140, 165, 140, 125, 141, 123, 141, 121, 141, 188, 141, 194, 141, 185, 141, 191, 141, 193, 141, 216, 142, 222, 142, 221, 142, 220, 142, 215, 142, 224, 142, 225, 142, 36, 144, 11, 144, 17, 144, 28, 144, 12, 144, 33, 144, 239, 144, 234, 144, 240, 144, 244, 144, 242, 144, 243, 144, 212, 144, 235, 144, 236, 144, 233, 144, 86, 145, 88, 145, 90, 145, 83, 145, 85, 145, 236, 145, 244, 145, 241, 145, 243, 145, 248, 145, 228, 145, 249, 145, 234, 145, 235, 145, 247, 145, 232, 145, 238, 145, 122, 149, 134, 149, 136, 149, 124, 150, 109, 150, 107, 150, 113, 150, 111, 150, 191, 150, 106, 151, 4, 152, 229, 152, 151, 153, 155, 80, 149, 80, 148, 80, 158, 80, 139, 80, 163, 80, 131, 80, 140, 80, 142, 80, 157, 80, 104, 80, 156, 80, 146, 80, 130, 80, 135, 80, 95, 81, 212, 81, 18, 83, 17, 83, 164, 83, 167, 83, 145, 85, 168, 85, 165, 85, 173, 85, 119, 85, 69, 86, 162, 85, 147, 85, 136, 85, 143, 85, 181, 85, 129, 85, 163, 85, 146, 85, 164, 85, 125, 85, 140, 85, 166, 85, 127, 85, 149, 85, 161, 85, 142, 85, 12, 87, 41, 88, 55, 88, 25, 88, 30, 88, 39, 88, 35, 88, 40, 88, 245, 87, 72, 88, 37, 88, 28, 88, 27, 88, 51, 88, 63, 88, 54, 88, 46, 88, 57, 88, 56, 88, 45, 88, 44, 88, 59, 88, 97, 89, 175, 90, 148, 90, 159, 90, 122, 90, 162, 90, 158, 90, 120, 90, 166, 90, 124, 90, 165, 90, 172, 90, 149, 90, 174, 90, 55, 90, 132, 90, 138, 90, 151, 90, 131, 90, 139, 90, 169, 90, 123, 90, 125, 90, 140, 90, 156, 90, 143, 90, 147, 90, 157, 90, 234, 91, 205, 91, 203, 91, 212, 91, 209, 91, 202, 91, 206, 91, 12, 92, 48, 92, 55, 93, 67, 93, 107, 93, 65, 93, 75, 93, 63, 93, 53, 93, 81, 93, 78, 93, 85, 93, 51, 93, 58, 93, 82, 93, 61, 93, 49, 93, 89, 93, 66, 93, 57, 93, 73, 93, 56, 93, 60, 93, 50, 93, 54, 93, 64, 93, 69, 93, 68, 94, 65, 94, 88, 95, 166, 95, 165, 95, 171, 95, 201, 96, 185, 96, 204, 96, 226, 96, 206, 96, 196, 96, 20, 97, 242, 96, 10, 97, 22, 97, 5, 97, 245, 96, 19, 97, 248, 96, 252, 96, 254, 96, 193, 96, 3, 97, 24, 97, 29, 97, 16, 97, 255, 96, 4, 97, 11, 97, 74, 98, 148, 99, 177, 99, 176, 99, 206, 99, 229, 99, 232, 99, 239, 99, 195, 99, 157, 100, 243, 99, 202, 99, 224, 99, 246, 99, 213, 99, 242, 99, 245, 99, 97, 100, 223, 99, 190, 99, 221, 99, 220, 99, 196, 99, 216, 99, 211, 99, 194, 99, 199, 99, 204, 99, 203, 99, 200, 99, 240, 99, 215, 99, 217, 99, 50, 101, 103, 101, 106, 101, 100, 101, 92, 101, 104, 101, 101, 101, 140, 101, 157, 101, 158, 101, 174, 101, 208, 101, 210, 101, 124, 102, 108, 102, 123, 102, 128, 102, 113, 102, 121, 102, 106, 102, 114, 102, 1, 103, 12, 105, 211, 104, 4, 105, 220, 104, 42, 105, 236, 104, 234, 104, 241, 104, 15, 105, 214, 104, 247, 104, 235, 104, 228, 104, 246, 104, 19, 105, 16, 105, 243, 104, 225, 104, 7, 105, 204, 104, 8, 105, 112, 105, 180, 104, 17, 105, 239, 104, 198, 104, 20, 105, 248, 104, 208, 104, 253, 104, 252, 104, 232, 104, 11, 105, 10, 105, 23, 105, 206, 104, 200, 104, 221, 104, 222, 104, 230, 104, 244, 104, 209, 104, 6, 105, 212, 104, 233, 104, 21, 105, 37, 105, 199, 104, 57, 107, 59, 107, 63, 107, 60, 107, 148, 107, 151, 107, 153, 107, 149, 107, 189, 107, 240, 107, 242, 107, 243, 107, 48, 108, 252, 109, 70, 110, 71, 110, 31, 110, 73, 110, 136, 110, 60, 110, 61, 110, 69, 110, 98, 110, 43, 110, 63, 110, 65, 110, 93, 110, 115, 110, 28, 110, 51, 110, 75, 110, 64, 110, 81, 110, 59, 110, 3, 110, 46, 110, 94, 110, 104, 110, 92, 110, 97, 110, 49, 110, 40, 110, 96, 110, 113, 110, 107, 110, 57, 110, 34, 110, 48, 110, 83, 110, 101, 110, 39, 110, 120, 110, 100, 110, 119, 110, 85, 110, 121, 110, 82, 110, 102, 110, 53, 110, 54, 110, 90, 110, 32, 113, 30, 113, 47, 113, 251, 112, 46, 113, 49, 113, 35, 113, 37, 113, 34, 113, 50, 113, 31, 113, 40, 113, 58, 113, 27, 113, 75, 114, 90, 114, 136, 114, 137, 114, 134, 114, 133, 114, 139, 114, 18, 115, 11, 115, 48, 115, 34, 115, 49, 115, 51, 115, 39, 115, 50, 115, 45, 115, 38, 115, 35, 115, 53, 115, 12, 115, 46, 116, 44, 116, 48, 116, 43, 116, 22, 116, 26, 116, 33, 116, 45, 116, 49, 116, 36, 116, 35, 116, 29, 116, 41, 116, 32, 116, 50, 116, 251, 116, 47, 117, 111, 117, 108, 117, 231, 117, 218, 117, 225, 117, 230, 117, 221, 117, 223, 117, 228, 117, 215, 117, 149, 118, 146, 118, 218, 118, 70, 119, 71, 119, 68, 119, 77, 119, 69, 119, 74, 119, 78, 119, 75, 119, 76, 119, 222, 119, 236, 119, 96, 120, 100, 120, 101, 120, 92, 120, 109, 120, 113, 120, 106, 120, 110, 120, 112, 120, 105, 120, 104, 120, 94, 120, 98, 120, 116, 121, 115, 121, 114, 121, 112, 121, 2, 122, 10, 122, 3, 122, 12, 122, 4, 122, 153, 122, 230, 122, 228, 122, 74, 123, 59, 123, 68, 123, 72, 123, 76, 123, 78, 123, 64, 123, 88, 123, 69, 123, 162, 124, 158, 124, 168, 124, 161, 124, 88, 125, 111, 125, 99, 125, 83, 125, 86, 125, 103, 125, 106, 125, 79, 125, 109, 125, 92, 125, 107, 125, 82, 125, 84, 125, 105, 125, 81, 125, 95, 125, 78, 125, 62, 127, 63, 127, 101, 127, 102, 127, 162, 127, 160, 127, 161, 127, 215, 127, 81, 128, 79, 128, 80, 128, 254, 128, 212, 128, 67, 129, 74, 129, 82, 129, 79, 129, 71, 129, 61, 129, 77, 129, 58, 129, 230, 129, 238, 129, 247, 129, 248, 129, 249, 129, 4, 130, 60, 130, 61, 130, 63, 130, 117, 130, 59, 131, 207, 131, 249, 131, 35, 132, 192, 131, 232, 131, 18, 132, 231, 131, 228, 131, 252, 131, 246, 131, 16, 132, 198, 131, 200, 131, 235, 131, 227, 131, 191, 131, 1, 132, 221, 131, 229, 131, 216, 131, 255, 131, 225, 131, 203, 131, 206, 131, 214, 131, 245, 131, 201, 131, 9, 132, 15, 132, 222, 131, 17, 132, 6, 132, 194, 131, 243, 131, 213, 131, 250, 131, 199, 131, 209, 131, 234, 131, 19, 132, 195, 131, 236, 131, 238, 131, 196, 131, 251, 131, 215, 131, 226, 131, 27, 132, 219, 131, 254, 131, 216, 134, 226, 134, 230, 134, 211, 134, 227, 134, 218, 134, 234, 134, 221, 134, 235, 134, 220, 134, 236, 134, 233, 134, 215, 134, 232, 134, 209, 134, 72, 136, 86, 136, 85, 136, 186, 136, 215, 136, 185, 136, 184, 136, 192, 136, 190, 136, 182, 136, 188, 136, 183, 136, 189, 136, 178, 136, 1, 137, 201, 136, 149, 137, 152, 137, 151, 137, 221, 137, 218, 137, 219, 137, 78, 138, 77, 138, 57, 138, 89, 138, 64, 138, 87, 138, 88, 138, 68, 138, 69, 138, 82, 138, 72, 138, 81, 138, 74, 138, 76, 138, 79, 138, 95, 140, 129, 140, 128, 140, 186, 140, 190, 140, 176, 140, 185, 140, 181, 140, 132, 141, 128, 141, 137, 141, 216, 141, 211, 141, 205, 141, 199, 141, 214, 141, 220, 141, 207, 141, 213, 141, 217, 141, 200, 141, 215, 141, 197, 141, 239, 142, 247, 142, 250, 142, 249, 142, 230, 142, 238, 142, 229, 142, 245, 142, 231, 142, 232, 142, 246, 142, 235, 142, 241, 142, 236, 142, 244, 142, 233, 142, 45, 144, 52, 144, 47, 144, 6, 145, 44, 145, 4, 145, 255, 144, 252, 144, 8, 145, 249, 144, 251, 144, 1, 145, 0, 145, 7, 145, 5, 145, 3, 145, 97, 145, 100, 145, 95, 145, 98, 145, 96, 145, 1, 146, 10, 146, 37, 146, 3, 146, 26, 146, 38, 146, 15, 146, 12, 146, 0, 146, 18, 146, 255, 145, 253, 145, 6, 146, 4, 146, 39, 146, 2, 146, 28, 146, 36, 146, 25, 146, 23, 146, 5, 146, 22, 146, 123, 149, 141, 149, 140, 149, 144, 149, 135, 150, 126, 150, 136, 150, 137, 150, 131, 150, 128, 150, 194, 150, 200, 150, 195, 150, 241, 150, 240, 150, 108, 151, 112, 151, 110, 151, 7, 152, 169, 152, 235, 152, 230, 156, 249, 158, 131, 78, 132, 78, 182, 78, 189, 80, 191, 80, 198, 80, 174, 80, 196, 80, 202, 80, 180, 80, 200, 80, 194, 80, 176, 80, 193, 80, 186, 80, 177, 80, 203, 80, 201, 80, 182, 80, 184, 80, 215, 81, 122, 82, 120, 82, 123, 82, 124, 82, 195, 85, 219, 85, 204, 85, 208, 85, 203, 85, 202, 85, 221, 85, 192, 85, 212, 85, 196, 85, 233, 85, 191, 85, 210, 85, 141, 85, 207, 85, 213, 85, 226, 85, 214, 85, 200, 85, 242, 85, 205, 85, 217, 85, 194, 85, 20, 87, 83, 88, 104, 88, 100, 88, 79, 88, 77, 88, 73, 88, 111, 88, 85, 88, 78, 88, 93, 88, 89, 88, 101, 88, 91, 88, 61, 88, 99, 88, 113, 88, 252, 88, 199, 90, 196, 90, 203, 90, 186, 90, 184, 90, 177, 90, 181, 90, 176, 90, 191, 90, 200, 90, 187, 90, 198, 90, 183, 90, 192, 90, 202, 90, 180, 90, 182, 90, 205, 90, 185, 90, 144, 90, 214, 91, 216, 91, 217, 91, 31, 92, 51, 92, 113, 93, 99, 93, 74, 93, 101, 93, 114, 93, 108, 93, 94, 93, 104, 93, 103, 93, 98, 93, 240, 93, 79, 94, 78, 94, 74, 94, 77, 94, 75, 94, 197, 94, 204, 94, 198, 94, 203, 94, 199, 94, 64, 95, 175, 95, 173, 95, 247, 96, 73, 97, 74, 97, 43, 97, 69, 97, 54, 97, 50, 97, 46, 97, 70, 97, 47, 97, 79, 97, 41, 97, 64, 97, 32, 98, 104, 145, 35, 98, 37, 98, 36, 98, 197, 99, 241, 99, 235, 99, 16, 100, 18, 100, 9, 100, 32, 100, 36, 100, 51, 100, 67, 100, 31, 100, 21, 100, 24, 100, 57, 100, 55, 100, 34, 100, 35, 100, 12, 100, 38, 100, 48, 100, 40, 100, 65, 100, 53, 100, 47, 100, 10, 100, 26, 100, 64, 100, 37, 100, 39, 100, 11, 100, 231, 99, 27, 100, 46, 100, 33, 100, 14, 100, 111, 101, 146, 101, 211, 101, 134, 102, 140, 102, 149, 102, 144, 102, 139, 102, 138, 102, 153, 102, 148, 102, 120, 102, 32, 103, 102, 105, 95, 105, 56, 105, 78, 105, 98, 105, 113, 105, 63, 105, 69, 105, 106, 105, 57, 105, 66, 105, 87, 105, 89, 105, 122, 105, 72, 105, 73, 105, 53, 105, 108, 105, 51, 105, 61, 105, 101, 105, 240, 104, 120, 105, 52, 105, 105, 105, 64, 105, 111, 105, 68, 105, 118, 105, 88, 105, 65, 105, 116, 105, 76, 105, 59, 105, 75, 105, 55, 105, 92, 105, 79, 105, 81, 105, 50, 105, 82, 105, 47, 105, 123, 105, 60, 105, 70, 107, 69, 107, 67, 107, 66, 107, 72, 107, 65, 107, 155, 107, 251, 107, 252, 107, 249, 107, 247, 107, 248, 107, 155, 110, 214, 110, 200, 110, 143, 110, 192, 110, 159, 110, 147, 110, 148, 110, 160, 110, 177, 110, 185, 110, 198, 110, 210, 110, 189, 110, 193, 110, 158, 110, 201, 110, 183, 110, 176, 110, 205, 110, 166, 110, 207, 110, 178, 110, 190, 110, 195, 110, 220, 110, 216, 110, 153, 110, 146, 110, 142, 110, 141, 110, 164, 110, 161, 110, 191, 110, 179, 110, 208, 110, 202, 110, 151, 110, 174, 110, 163, 110, 71, 113, 84, 113, 82, 113, 99, 113, 96, 113, 65, 113, 93, 113, 98, 113, 114, 113, 120, 113, 106, 113, 97, 113, 66, 113, 88, 113, 67, 113, 75, 113, 112, 113, 95, 113, 80, 113, 83, 113, 68, 113, 77, 113, 90, 113, 79, 114, 141, 114, 140, 114, 145, 114, 144, 114, 142, 114, 60, 115, 66, 115, 59, 115, 58, 115, 64, 115, 74, 115, 73, 115, 68, 116, 74, 116, 75, 116, 82, 116, 81, 116, 87, 116, 64, 116, 79, 116, 80, 116, 78, 116, 66, 116, 70, 116, 77, 116, 84, 116, 225, 116, 255, 116, 254, 116, 253, 116, 29, 117, 121, 117, 119, 117, 131, 105, 239, 117, 15, 118, 3, 118, 247, 117, 254, 117, 252, 117, 249, 117, 248, 117, 16, 118, 251, 117, 246, 117, 237, 117, 245, 117, 253, 117, 153, 118, 181, 118, 221, 118, 85, 119, 95, 119, 96, 119, 82, 119, 86, 119, 90, 119, 105, 119, 103, 119, 84, 119, 89, 119, 109, 119, 224, 119, 135, 120, 154, 120, 148, 120, 143, 120, 132, 120, 149, 120, 133, 120, 134, 120, 161, 120, 131, 120, 121, 120, 153, 120, 128, 120, 150, 120, 123, 120, 124, 121, 130, 121, 125, 121, 121, 121, 17, 122, 24, 122, 25, 122, 18, 122, 23, 122, 21, 122, 34, 122, 19, 122, 27, 122, 16, 122, 163, 122, 162, 122, 158, 122, 235, 122, 102, 123, 100, 123, 109, 123, 116, 123, 105, 123, 114, 123, 101, 123, 115, 123, 113, 123, 112, 123, 97, 123, 120, 123, 118, 123, 99, 123, 178, 124, 180, 124, 175, 124, 136, 125, 134, 125, 128, 125, 141, 125, 127, 125, 133, 125, 122, 125, 142, 125, 123, 125, 131, 125, 124, 125, 140, 125, 148, 125, 132, 125, 125, 125, 146, 125, 109, 127, 107, 127, 103, 127, 104, 127, 108, 127, 166, 127, 165, 127, 167, 127, 219, 127, 220, 127, 33, 128, 100, 129, 96, 129, 119, 129, 92, 129, 105, 129, 91, 129, 98, 129, 114, 129, 33, 103, 94, 129, 118, 129, 103, 129, 111, 129, 68, 129, 97, 129, 29, 130, 73, 130, 68, 130, 64, 130, 66, 130, 69, 130, 241, 132, 63, 132, 86, 132, 118, 132, 121, 132, 143, 132, 141, 132, 101, 132, 81, 132, 64, 132, 134, 132, 103, 132, 48, 132, 77, 132, 125, 132, 90, 132, 89, 132, 116, 132, 115, 132, 93, 132, 7, 133, 94, 132, 55, 132, 58, 132, 52, 132, 122, 132, 67, 132, 120, 132, 50, 132, 69, 132, 41, 132, 217, 131, 75, 132, 47, 132, 66, 132, 45, 132, 95, 132, 112, 132, 57, 132, 78, 132, 76, 132, 82, 132, 111, 132, 197, 132, 142, 132, 59, 132, 71, 132, 54, 132, 51, 132, 104, 132, 126, 132, 68, 132, 43, 132, 96, 132, 84, 132, 110, 132, 80, 132, 11, 135, 4, 135, 247, 134, 12, 135, 250, 134, 214, 134, 245, 134, 77, 135, 248, 134, 14, 135, 9, 135, 1, 135, 246, 134, 13, 135, 5, 135, 214, 136, 203, 136, 205, 136, 206, 136, 222, 136, 219, 136, 218, 136, 204, 136, 208, 136, 133, 137, 155, 137, 223, 137, 229, 137, 228, 137, 225, 137, 224, 137, 226, 137, 220, 137, 230, 137, 118, 138, 134, 138, 127, 138, 97, 138, 63, 138, 119, 138, 130, 138, 132, 138, 117, 138, 131, 138, 129, 138, 116, 138, 122, 138, 60, 140, 75, 140, 74, 140, 101, 140, 100, 140, 102, 140, 134, 140, 132, 140, 133, 140, 204, 140, 104, 141, 105, 141, 145, 141, 140, 141, 142, 141, 143, 141, 141, 141, 147, 141, 148, 141, 144, 141, 146, 141, 240, 141, 224, 141, 236, 141, 241, 141, 238, 141, 208, 141, 233, 141, 227, 141, 226, 141, 231, 141, 242, 141, 235, 141, 244, 141, 6, 143, 255, 142, 1, 143, 0, 143, 5, 143, 7, 143, 8, 143, 2, 143, 11, 143, 82, 144, 63, 144, 68, 144, 73, 144, 61, 144, 16, 145, 13, 145, 15, 145, 17, 145, 22, 145, 20, 145, 11, 145, 14, 145, 110, 145, 111, 145, 72, 146, 82, 146, 48, 146, 58, 146, 102, 146, 51, 146, 101, 146, 94, 146, 131, 146, 46, 146, 74, 146, 70, 146, 109, 146, 108, 146, 79, 146, 96, 146, 103, 146, 111, 146, 54, 146, 97, 146, 112, 146, 49, 146, 84, 146, 99, 146, 80, 146, 114, 146, 78, 146, 83, 146, 76, 146, 86, 146, 50, 146, 159, 149, 156, 149, 158, 149, 155, 149, 146, 150, 147, 150, 145, 150, 151, 150, 206, 150, 250, 150, 253, 150, 248, 150, 245, 150, 115, 151, 119, 151, 120, 151, 114, 151, 15, 152, 13, 152, 14, 152, 172, 152, 246, 152, 249, 152, 175, 153, 178, 153, 176, 153, 181, 153, 173, 154, 171, 154, 91, 155, 234, 156, 237, 156, 231, 156, 128, 158, 253, 158, 230, 80, 212, 80, 215, 80, 232, 80, 243, 80, 219, 80, 234, 80, 221, 80, 228, 80, 211, 80, 236, 80, 240, 80, 239, 80, 227, 80, 224, 80, 216, 81, 128, 82, 129, 82, 233, 82, 235, 82, 48, 83, 172, 83, 39, 86, 21, 86, 12, 86, 18, 86, 252, 85, 15, 86, 28, 86, 1, 86, 19, 86, 2, 86, 250, 85, 29, 86, 4, 86, 255, 85, 249, 85, 137, 88, 124, 88, 144, 88, 152, 88, 134, 88, 129, 88, 127, 88, 116, 88, 139, 88, 122, 88, 135, 88, 145, 88, 142, 88, 118, 88, 130, 88, 136, 88, 123, 88, 148, 88, 143, 88, 254, 88, 107, 89, 220, 90, 238, 90, 229, 90, 213, 90, 234, 90, 218, 90, 237, 90, 235, 90, 243, 90, 226, 90, 224, 90, 219, 90, 236, 90, 222, 90, 221, 90, 217, 90, 232, 90, 223, 90, 119, 91, 224, 91, 227, 91, 99, 92, 130, 93, 128, 93, 125, 93, 134, 93, 122, 93, 129, 93, 119, 93, 138, 93, 137, 93, 136, 93, 126, 93, 124, 93, 141, 93, 121, 93, 127, 93, 88, 94, 89, 94, 83, 94, 216, 94, 209, 94, 215, 94, 206, 94, 220, 94, 213, 94, 217, 94, 210, 94, 212, 94, 68, 95, 67, 95, 111, 95, 182, 95, 44, 97, 40, 97, 65, 97, 94, 97, 113, 97, 115, 97, 82, 97, 83, 97, 114, 97, 108, 97, 128, 97, 116, 97, 84, 97, 122, 97, 91, 97, 101, 97, 59, 97, 106, 97, 97, 97, 86, 97, 41, 98, 39, 98, 43, 98, 43, 100, 77, 100, 91, 100, 93, 100, 116, 100, 118, 100, 114, 100, 115, 100, 125, 100, 117, 100, 102, 100, 166, 100, 78, 100, 130, 100, 94, 100, 92, 100, 75, 100, 83, 100, 96, 100, 80, 100, 127, 100, 63, 100, 108, 100, 107, 100, 89, 100, 101, 100, 119, 100, 115, 101, 160, 101, 161, 102, 160, 102, 159, 102, 5, 103, 4, 103, 34, 103, 177, 105, 182, 105, 201, 105, 160, 105, 206, 105, 150, 105, 176, 105, 172, 105, 188, 105, 145, 105, 153, 105, 142, 105, 167, 105, 141, 105, 169, 105, 190, 105, 175, 105, 191, 105, 196, 105, 189, 105, 164, 105, 212, 105, 185, 105, 202, 105, 154, 105, 207, 105, 179, 105, 147, 105, 170, 105, 161, 105, 158, 105, 217, 105, 151, 105, 144, 105, 194, 105, 181, 105, 165, 105, 198, 105, 74, 107, 77, 107, 75, 107, 158, 107, 159, 107, 160, 107, 195, 107, 196, 107, 254, 107, 206, 110, 245, 110, 241, 110, 3, 111, 37, 111, 248, 110, 55, 111, 251, 110, 46, 111, 9, 111, 78, 111, 25, 111, 26, 111, 39, 111, 24, 111, 59, 111, 18, 111, 237, 110, 10, 111, 54, 111, 115, 111, 249, 110, 238, 110, 45, 111, 64, 111, 48, 111, 60, 111, 53, 111, 235, 110, 7, 111, 14, 111, 67, 111, 5, 111, 253, 110, 246, 110, 57, 111, 28, 111, 252, 110, 58, 111, 31, 111, 13, 111, 30, 111, 8, 111, 33, 111, 135, 113, 144, 113, 137, 113, 128, 113, 133, 113, 130, 113, 143, 113, 123, 113, 134, 113, 129, 113, 151, 113, 68, 114, 83, 114, 151, 114, 149, 114, 147, 114, 67, 115, 77, 115, 81, 115, 76, 115, 98, 116, 115, 116, 113, 116, 117, 116, 114, 116, 103, 116, 110, 116, 0, 117, 2, 117, 3, 117, 125, 117, 144, 117, 22, 118, 8, 118, 12, 118, 21, 118, 17, 118, 10, 118, 20, 118, 184, 118, 129, 119, 124, 119, 133, 119, 130, 119, 110, 119, 128, 119, 111, 119, 126, 119, 131, 119, 178, 120, 170, 120, 180, 120, 173, 120, 168, 120, 126, 120, 171, 120, 158, 120, 165, 120, 160, 120, 172, 120, 162, 120, 164, 120, 152, 121, 138, 121, 139, 121, 150, 121, 149, 121, 148, 121, 147, 121, 151, 121, 136, 121, 146, 121, 144, 121, 43, 122, 74, 122, 48, 122, 47, 122, 40, 122, 38, 122, 168, 122, 171, 122, 172, 122, 238, 122, 136, 123, 156, 123, 138, 123, 145, 123, 144, 123, 150, 123, 141, 123, 140, 123, 155, 123, 142, 123, 133, 123, 152, 123, 132, 82, 153, 123, 164, 123, 130, 123, 187, 124, 191, 124, 188, 124, 186, 124, 167, 125, 183, 125, 194, 125, 163, 125, 170, 125, 193, 125, 192, 125, 197, 125, 157, 125, 206, 125, 196, 125, 198, 125, 203, 125, 204, 125, 175, 125, 185, 125, 150, 125, 188, 125, 159, 125, 166, 125, 174, 125, 169, 125, 161, 125, 201, 125, 115, 127, 226, 127, 227, 127, 229, 127, 222, 127, 36, 128, 93, 128, 92, 128, 137, 129, 134, 129, 131, 129, 135, 129, 141, 129, 140, 129, 139, 129, 21, 130, 151, 132, 164, 132, 161, 132, 159, 132, 186, 132, 206, 132, 194, 132, 172, 132, 174, 132, 171, 132, 185, 132, 180, 132, 193, 132, 205, 132, 170, 132, 154, 132, 177, 132, 208, 132, 157, 132, 167, 132, 187, 132, 162, 132, 148, 132, 199, 132, 204, 132, 155, 132, 169, 132, 175, 132, 168, 132, 214, 132, 152, 132, 182, 132, 207, 132, 160, 132, 215, 132, 212, 132, 210, 132, 219, 132, 176, 132, 145, 132, 97, 134, 51, 135, 35, 135, 40, 135, 107, 135, 64, 135, 46, 135, 30, 135, 33, 135, 25, 135, 27, 135, 67, 135, 44, 135, 65, 135, 62, 135, 70, 135, 32, 135, 50, 135, 42, 135, 45, 135, 60, 135, 18, 135, 58, 135, 49, 135, 53, 135, 66, 135, 38, 135, 39, 135, 56, 135, 36, 135, 26, 135, 48, 135, 17, 135, 247, 136, 231, 136, 241, 136, 242, 136, 250, 136, 254, 136, 238, 136, 252, 136, 246, 136, 251, 136, 240, 136, 236, 136, 235, 136, 157, 137, 161, 137, 159, 137, 158, 137, 233, 137, 235, 137, 232, 137, 171, 138, 153, 138, 139, 138, 146, 138, 143, 138, 150, 138, 61, 140, 104, 140, 105, 140, 213, 140, 207, 140, 215, 140, 150, 141, 9, 142, 2, 142, 255, 141, 13, 142, 253, 141, 10, 142, 3, 142, 7, 142, 6, 142, 5, 142, 254, 141, 0, 142, 4, 142, 16, 143, 17, 143, 14, 143, 13, 143, 35, 145, 28, 145, 32, 145, 34, 145, 31, 145, 29, 145, 26, 145, 36, 145, 33, 145, 27, 145, 122, 145, 114, 145, 121, 145, 115, 145, 165, 146, 164, 146, 118, 146, 155, 146, 122, 146, 160, 146, 148, 146, 170, 146, 141, 146, 166, 146, 154, 146, 171, 146, 121, 146, 151, 146, 127, 146, 163, 146, 238, 146, 142, 146, 130, 146, 149, 146, 162, 146, 125, 146, 136, 146, 161, 146, 138, 146, 134, 146, 140, 146, 153, 146, 167, 146, 126, 146, 135, 146, 169, 146, 157, 146, 139, 146, 45, 146, 158, 150, 161, 150, 255, 150, 88, 151, 125, 151, 122, 151, 126, 151, 131, 151, 128, 151, 130, 151, 123, 151, 132, 151, 129, 151, 127, 151, 206, 151, 205, 151, 22, 152, 173, 152, 174, 152, 2, 153, 0, 153, 7, 153, 157, 153, 156, 153, 195, 153, 185, 153, 187, 153, 186, 153, 194, 153, 189, 153, 199, 153, 177, 154, 227, 154, 231, 154, 62, 155, 63, 155, 96, 155, 97, 155, 95, 155, 241, 156, 242, 156, 245, 156, 167, 158, 255, 80, 3, 81, 48, 81, 248, 80, 6, 81, 7, 81, 246, 80, 254, 80, 11, 81, 12, 81, 253, 80, 10, 81, 139, 82, 140, 82, 241, 82, 239, 82, 72, 86, 66, 86, 76, 86, 53, 86, 65, 86, 74, 86, 73, 86, 70, 86, 88, 86, 90, 86, 64, 86, 51, 86, 61, 86, 44, 86, 62, 86, 56, 86, 42, 86, 58, 86, 26, 87, 171, 88, 157, 88, 177, 88, 160, 88, 163, 88, 175, 88, 172, 88, 165, 88, 161, 88, 255, 88, 255, 90, 244, 90, 253, 90, 247, 90, 246, 90, 3, 91, 248, 90, 2, 91, 249, 90, 1, 91, 7, 91, 5, 91, 15, 91, 103, 92, 153, 93, 151, 93, 159, 93, 146, 93, 162, 93, 147, 93, 149, 93, 160, 93, 156, 93, 161, 93, 154, 93, 158, 93, 105, 94, 93, 94, 96, 94, 92, 94, 243, 125, 219, 94, 222, 94, 225, 94, 73, 95, 178, 95, 139, 97, 131, 97, 121, 97, 177, 97, 176, 97, 162, 97, 137, 97, 155, 97, 147, 97, 175, 97, 173, 97, 159, 97, 146, 97, 170, 97, 161, 97, 141, 97, 102, 97, 179, 97, 45, 98, 110, 100, 112, 100, 150, 100, 160, 100, 133, 100, 151, 100, 156, 100, 143, 100, 139, 100, 138, 100, 140, 100, 163, 100, 159, 100, 104, 100, 177, 100, 152, 100, 118, 101, 122, 101, 121, 101, 123, 101, 178, 101, 179, 101, 181, 102, 176, 102, 169, 102, 178, 102, 183, 102, 170, 102, 175, 102, 0, 106, 6, 106, 23, 106, 229, 105, 248, 105, 21, 106, 241, 105, 228, 105, 32, 106, 255, 105, 236, 105, 226, 105, 27, 106, 29, 106, 254, 105, 39, 106, 242, 105, 238, 105, 20, 106, 247, 105, 231, 105, 64, 106, 8, 106, 230, 105, 251, 105, 13, 106, 252, 105, 235, 105, 9, 106, 4, 106, 24, 106, 37, 106, 15, 106, 246, 105, 38, 106, 7, 106, 244, 105, 22, 106, 81, 107, 165, 107, 163, 107, 162, 107, 166, 107, 1, 108, 0, 108, 255, 107, 2, 108, 65, 111, 38, 111, 126, 111, 135, 111, 198, 111, 146, 111, 141, 111, 137, 111, 140, 111, 98, 111, 79, 111, 133, 111, 90, 111, 150, 111, 118, 111, 108, 111, 130, 111, 85, 111, 114, 111, 82, 111, 80, 111, 87, 111, 148, 111, 147, 111, 93, 111, 0, 111, 97, 111, 107, 111, 125, 111, 103, 111, 144, 111, 83, 111, 139, 111, 105, 111, 127, 111, 149, 111, 99, 111, 119, 111, 106, 111, 123, 111, 178, 113, 175, 113, 155, 113, 176, 113, 160, 113, 154, 113, 169, 113, 181, 113, 157, 113, 165, 113, 158, 113, 164, 113, 161, 113, 170, 113, 156, 113, 167, 113, 179, 113, 152, 114, 154, 114, 88, 115, 82, 115, 94, 115, 95, 115, 96, 115, 93, 115, 91, 115, 97, 115, 90, 115, 89, 115, 98, 115, 135, 116, 137, 116, 138, 116, 134, 116, 129, 116, 125, 116, 133, 116, 136, 116, 124, 116, 121, 116, 8, 117, 7, 117, 126, 117, 37, 118, 30, 118, 25, 118, 29, 118, 28, 118, 35, 118, 26, 118, 40, 118, 27, 118, 156, 118, 157, 118, 158, 118, 155, 118, 141, 119, 143, 119, 137, 119, 136, 119, 205, 120, 187, 120, 207, 120, 204, 120, 209, 120, 206, 120, 212, 120, 200, 120, 195, 120, 196, 120, 201, 120, 154, 121, 161, 121, 160, 121, 156, 121, 162, 121, 155, 121, 118, 107, 57, 122, 178, 122, 180, 122, 179, 122, 183, 123, 203, 123, 190, 123, 172, 123, 206, 123, 175, 123, 185, 123, 202, 123, 181, 123, 197, 124, 200, 124, 204, 124, 203, 124, 247, 125, 219, 125, 234, 125, 231, 125, 215, 125, 225, 125, 3, 126, 250, 125, 230, 125, 246, 125, 241, 125, 240, 125, 238, 125, 223, 125, 118, 127, 172, 127, 176, 127, 173, 127, 237, 127, 235, 127, 234, 127, 236, 127, 230, 127, 232, 127, 100, 128, 103, 128, 163, 129, 159, 129, 158, 129, 149, 129, 162, 129, 153, 129, 151, 129, 22, 130, 79, 130, 83, 130, 82, 130, 80, 130, 78, 130, 81, 130, 36, 133, 59, 133, 15, 133, 0, 133, 41, 133, 14, 133, 9, 133, 13, 133, 31, 133, 10, 133, 39, 133, 28, 133, 251, 132, 43, 133, 250, 132, 8, 133, 12, 133, 244, 132, 42, 133, 242, 132, 21, 133, 247, 132, 235, 132, 243, 132, 252, 132, 18, 133, 234, 132, 233, 132, 22, 133, 254, 132, 40, 133, 29, 133, 46, 133, 2, 133, 253, 132, 30, 133, 246, 132, 49, 133, 38, 133, 231, 132, 232, 132, 240, 132, 239, 132, 249, 132, 24, 133, 32, 133, 48, 133, 11, 133, 25, 133, 47, 133, 98, 134, 86, 135, 99, 135, 100, 135, 119, 135, 225, 135, 115, 135, 88, 135, 84, 135, 91, 135, 82, 135, 97, 135, 90, 135, 81, 135, 94, 135, 109, 135, 106, 135, 80, 135, 78, 135, 95, 135, 93, 135, 111, 135, 108, 135, 122, 135, 110, 135, 92, 135, 101, 135, 79, 135, 123, 135, 117, 135, 98, 135, 103, 135, 105, 135, 90, 136, 5, 137, 12, 137, 20, 137, 11, 137, 23, 137, 24, 137, 25, 137, 6, 137, 22, 137, 17, 137, 14, 137, 9, 137, 162, 137, 164, 137, 163, 137, 237, 137, 240, 137, 236, 137, 207, 138, 198, 138, 184, 138, 211, 138, 209, 138, 212, 138, 213, 138, 187, 138, 215, 138, 190, 138, 192, 138, 197, 138, 216, 138, 195, 138, 186, 138, 189, 138, 217, 138, 62, 140, 77, 140, 143, 140, 229, 140, 223, 140, 217, 140, 232, 140, 218, 140, 221, 140, 231, 140, 160, 141, 156, 141, 161, 141, 155, 141, 32, 142, 35, 142, 37, 142, 36, 142, 46, 142, 21, 142, 27, 142, 22, 142, 17, 142, 25, 142, 38, 142, 39, 142, 20, 142, 18, 142, 24, 142, 19, 142, 28, 142, 23, 142, 26, 142, 44, 143, 36, 143, 24, 143, 26, 143, 32, 143, 35, 143, 22, 143, 23, 143, 115, 144, 112, 144, 111, 144, 103, 144, 107, 144, 47, 145, 43, 145, 41, 145, 42, 145, 50, 145, 38, 145, 46, 145, 133, 145, 134, 145, 138, 145, 129, 145, 130, 145, 132, 145, 128, 145, 208, 146, 195, 146, 196, 146, 192, 146, 217, 146, 182, 146, 207, 146, 241, 146, 223, 146, 216, 146, 233, 146, 215, 146, 221, 146, 204, 146, 239, 146, 194, 146, 232, 146, 202, 146, 200, 146, 206, 146, 230, 146, 205, 146, 213, 146, 201, 146, 224, 146, 222, 146, 231, 146, 209, 146, 211, 146, 181, 146, 225, 146, 198, 146, 180, 146, 124, 149, 172, 149, 171, 149, 174, 149, 176, 149, 164, 150, 162, 150, 211, 150, 5, 151, 8, 151, 2, 151, 90, 151, 138, 151, 142, 151, 136, 151, 208, 151, 207, 151, 30, 152, 29, 152, 38, 152, 41, 152, 40, 152, 32, 152, 27, 152, 39, 152, 178, 152, 8, 153, 250, 152, 17, 153, 20, 153, 22, 153, 23, 153, 21, 153, 220, 153, 205, 153, 207, 153, 211, 153, 212, 153, 206, 153, 201, 153, 214, 153, 216, 153, 203, 153, 215, 153, 204, 153, 179, 154, 236, 154, 235, 154, 243, 154, 242, 154, 241, 154, 70, 155, 67, 155, 103, 155, 116, 155, 113, 155, 102, 155, 118, 155, 117, 155, 112, 155, 104, 155, 100, 155, 108, 155, 252, 156, 250, 156, 253, 156, 255, 156, 247, 156, 7, 157, 0, 157, 249, 156, 251, 156, 8, 157, 5, 157, 4, 157, 131, 158, 211, 158, 15, 159, 16, 159, 28, 81, 19, 81, 23, 81, 26, 81, 17, 81, 222, 81, 52, 83, 225, 83, 112, 86, 96, 86, 110, 86, 115, 86, 102, 86, 99, 86, 109, 86, 114, 86, 94, 86, 119, 86, 28, 87, 27, 87, 200, 88, 189, 88, 201, 88, 191, 88, 186, 88, 194, 88, 188, 88, 198, 88, 23, 91, 25, 91, 27, 91, 33, 91, 20, 91, 19, 91, 16, 91, 22, 91, 40, 91, 26, 91, 32, 91, 30, 91, 239, 91, 172, 93, 177, 93, 169, 93, 167, 93, 181, 93, 176, 93, 174, 93, 170, 93, 168, 93, 178, 93, 173, 93, 175, 93, 180, 93, 103, 94, 104, 94, 102, 94, 111, 94, 233, 94, 231, 94, 230, 94, 232, 94, 229, 94, 75, 95, 188, 95, 157, 97, 168, 97, 150, 97, 197, 97, 180, 97, 198, 97, 193, 97, 204, 97, 186, 97, 191, 97, 184, 97, 140, 97, 215, 100, 214, 100, 208, 100, 207, 100, 201, 100, 189, 100, 137, 100, 195, 100, 219, 100, 243, 100, 217, 100, 51, 101, 127, 101, 124, 101, 162, 101, 200, 102, 190, 102, 192, 102, 202, 102, 203, 102, 207, 102, 189, 102, 187, 102, 186, 102, 204, 102, 35, 103, 52, 106, 102, 106, 73, 106, 103, 106, 50, 106, 104, 106, 62, 106, 93, 106, 109, 106, 118, 106, 91, 106, 81, 106, 40, 106, 90, 106, 59, 106, 63, 106, 65, 106, 106, 106, 100, 106, 80, 106, 79, 106, 84, 106, 111, 106, 105, 106, 96, 106, 60, 106, 94, 106, 86, 106, 85, 106, 77, 106, 78, 106, 70, 106, 85, 107, 84, 107, 86, 107, 167, 107, 170, 107, 171, 107, 200, 107, 199, 107, 4, 108, 3, 108, 6, 108, 173, 111, 203, 111, 163, 111, 199, 111, 188, 111, 206, 111, 200, 111, 94, 111, 196, 111, 189, 111, 158, 111, 202, 111, 168, 111, 4, 112, 165, 111, 174, 111, 186, 111, 172, 111, 170, 111, 207, 111, 191, 111, 184, 111, 162, 111, 201, 111, 171, 111, 205, 111, 175, 111, 178, 111, 176, 111, 197, 113, 194, 113, 191, 113, 184, 113, 214, 113, 192, 113, 193, 113, 203, 113, 212, 113, 202, 113, 199, 113, 207, 113, 189, 113, 216, 113, 188, 113, 198, 113, 218, 113, 219, 113, 157, 114, 158, 114, 105, 115, 102, 115, 103, 115, 108, 115, 101, 115, 107, 115, 106, 115, 127, 116, 154, 116, 160, 116, 148, 116, 146, 116, 149, 116, 161, 116, 11, 117, 128, 117, 47, 118, 45, 118, 49, 118, 61, 118, 51, 118, 60, 118, 53, 118, 50, 118, 48, 118, 187, 118, 230, 118, 154, 119, 157, 119, 161, 119, 156, 119, 155, 119, 162, 119, 163, 119, 149, 119, 153, 119, 151, 119, 221, 120, 233, 120, 229, 120, 234, 120, 222, 120, 227, 120, 219, 120, 225, 120, 226, 120, 237, 120, 223, 120, 224, 120, 164, 121, 68, 122, 72, 122, 71, 122, 182, 122, 184, 122, 181, 122, 177, 122, 183, 122, 222, 123, 227, 123, 231, 123, 221, 123, 213, 123, 229, 123, 218, 123, 232, 123, 249, 123, 212, 123, 234, 123, 226, 123, 220, 123, 235, 123, 216, 123, 223, 123, 210, 124, 212, 124, 215, 124, 208, 124, 209, 124, 18, 126, 33, 126, 23, 126, 12, 126, 31, 126, 32, 126, 19, 126, 14, 126, 28, 126, 21, 126, 26, 126, 34, 126, 11, 126, 15, 126, 22, 126, 13, 126, 20, 126, 37, 126, 36, 126, 67, 127, 123, 127, 124, 127, 122, 127, 177, 127, 239, 127, 42, 128, 41, 128, 108, 128, 177, 129, 166, 129, 174, 129, 185, 129, 181, 129, 171, 129, 176, 129, 172, 129, 180, 129, 178, 129, 183, 129, 167, 129, 242, 129, 85, 130, 86, 130, 87, 130, 86, 133, 69, 133, 107, 133, 77, 133, 83, 133, 97, 133, 88, 133, 64, 133, 70, 133, 100, 133, 65, 133, 98, 133, 68, 133, 81, 133, 71, 133, 99, 133, 62, 133, 91, 133, 113, 133, 78, 133, 110, 133, 117, 133, 85, 133, 103, 133, 96, 133, 140, 133, 102, 133, 93, 133, 84, 133, 101, 133, 108, 133, 99, 134, 101, 134, 100, 134, 155, 135, 143, 135, 151, 135, 147, 135, 146, 135, 136, 135, 129, 135, 150, 135, 152, 135, 121, 135, 135, 135, 163, 135, 133, 135, 144, 135, 145, 135, 157, 135, 132, 135, 148, 135, 156, 135, 154, 135, 137, 135, 30, 137, 38, 137, 48, 137, 45, 137, 46, 137, 39, 137, 49, 137, 34, 137, 41, 137, 35, 137, 47, 137, 44, 137, 31, 137, 241, 137, 224, 138, 226, 138, 242, 138, 244, 138, 245, 138, 221, 138, 20, 139, 228, 138, 223, 138, 240, 138, 200, 138, 222, 138, 225, 138, 232, 138, 255, 138, 239, 138, 251, 138, 145, 140, 146, 140, 144, 140, 245, 140, 238, 140, 241, 140, 240, 140, 243, 140, 108, 141, 110, 141, 165, 141, 167, 141, 51, 142, 62, 142, 56, 142, 64, 142, 69, 142, 54, 142, 60, 142, 61, 142, 65, 142, 48, 142, 63, 142, 189, 142, 54, 143, 46, 143, 53, 143, 50, 143, 57, 143, 55, 143, 52, 143, 118, 144, 121, 144, 123, 144, 134, 144, 250, 144, 51, 145, 53, 145, 54, 145, 147, 145, 144, 145, 145, 145, 141, 145, 143, 145, 39, 147, 30, 147, 8, 147, 31, 147, 6, 147, 15, 147, 122, 147, 56, 147, 60, 147, 27, 147, 35, 147, 18, 147, 1, 147, 70, 147, 45, 147, 14, 147, 13, 147, 203, 146, 29, 147, 250, 146, 37, 147, 19, 147, 249, 146, 247, 146, 52, 147, 2, 147, 36, 147, 255, 146, 41, 147, 57, 147, 53, 147, 42, 147, 20, 147, 12, 147, 11, 147, 254, 146, 9, 147, 0, 147, 251, 146, 22, 147, 188, 149, 205, 149, 190, 149, 185, 149, 186, 149, 182, 149, 191, 149, 181, 149, 189, 149, 169, 150, 212, 150, 11, 151, 18, 151, 16, 151, 153, 151, 151, 151, 148, 151, 240, 151, 248, 151, 53, 152, 47, 152, 50, 152, 36, 153, 31, 153, 39, 153, 41, 153, 158, 153, 238, 153, 236, 153, 229, 153, 228, 153, 240, 153, 227, 153, 234, 153, 233, 153, 231, 153, 185, 154, 191, 154, 180, 154, 187, 154, 246, 154, 250, 154, 249, 154, 247, 154, 51, 155, 128, 155, 133, 155, 135, 155, 124, 155, 126, 155, 123, 155, 130, 155, 147, 155, 146, 155, 144, 155, 122, 155, 149, 155, 125, 155, 136, 155, 37, 157, 23, 157, 32, 157, 30, 157, 20, 157, 41, 157, 29, 157, 24, 157, 34, 157, 16, 157, 25, 157, 31, 157, 136, 158, 134, 158, 135, 158, 174, 158, 173, 158, 213, 158, 214, 158, 250, 158, 18, 159, 61, 159, 38, 81, 37, 81, 34, 81, 36, 81, 32, 81, 41, 81, 244, 82, 147, 86, 140, 86, 141, 86, 134, 86, 132, 86, 131, 86, 126, 86, 130, 86, 127, 86, 129, 86, 214, 88, 212, 88, 207, 88, 210, 88, 45, 91, 37, 91, 50, 91, 35, 91, 44, 91, 39, 91, 38, 91, 47, 91, 46, 91, 123, 91, 241, 91, 242, 91, 183, 93, 108, 94, 106, 94, 190, 95, 187, 95, 195, 97, 181, 97, 188, 97, 231, 97, 224, 97, 229, 97, 228, 97, 232, 97, 222, 97, 239, 100, 233, 100, 227, 100, 235, 100, 228, 100, 232, 100, 129, 101, 128, 101, 182, 101, 218, 101, 210, 102, 141, 106, 150, 106, 129, 106, 165, 106, 137, 106, 159, 106, 155, 106, 161, 106, 158, 106, 135, 106, 147, 106, 142, 106, 149, 106, 131, 106, 168, 106, 164, 106, 145, 106, 127, 106, 166, 106, 154, 106, 133, 106, 140, 106, 146, 106, 91, 107, 173, 107, 9, 108, 204, 111, 169, 111, 244, 111, 212, 111, 227, 111, 220, 111, 237, 111, 231, 111, 230, 111, 222, 111, 242, 111, 221, 111, 226, 111, 232, 111, 225, 113, 241, 113, 232, 113, 242, 113, 228, 113, 240, 113, 226, 113, 115, 115, 110, 115, 111, 115, 151, 116, 178, 116, 171, 116, 144, 116, 170, 116, 173, 116, 177, 116, 165, 116, 175, 116, 16, 117, 17, 117, 18, 117, 15, 117, 132, 117, 67, 118, 72, 118, 73, 118, 71, 118, 164, 118, 233, 118, 181, 119, 171, 119, 178, 119, 183, 119, 182, 119, 180, 119, 177, 119, 168, 119, 240, 119, 243, 120, 253, 120, 2, 121, 251, 120, 252, 120, 242, 120, 5, 121, 249, 120, 254, 120, 4, 121, 171, 121, 168, 121, 92, 122, 91, 122, 86, 122, 88, 122, 84, 122, 90, 122, 190, 122, 192, 122, 193, 122, 5, 124, 15, 124, 242, 123, 0, 124, 255, 123, 251, 123, 14, 124, 244, 123, 11, 124, 243, 123, 2, 124, 9, 124, 3, 124, 1, 124, 248, 123, 253, 123, 6, 124, 240, 123, 241, 123, 16, 124, 10, 124, 232, 124, 45, 126, 60, 126, 66, 126, 51, 126, 72, 152, 56, 126, 42, 126, 73, 126, 64, 126, 71, 126, 41, 126, 76, 126, 48, 126, 59, 126, 54, 126, 68, 126, 58, 126, 69, 127, 127, 127, 126, 127, 125, 127, 244, 127, 242, 127, 44, 128, 187, 129, 196, 129, 204, 129, 202, 129, 197, 129, 199, 129, 188, 129, 233, 129, 91, 130, 90, 130, 92, 130, 131, 133, 128, 133, 143, 133, 167, 133, 149, 133, 160, 133, 139, 133, 163, 133, 123, 133, 164, 133, 154, 133, 158, 133, 119, 133, 124, 133, 137, 133, 161, 133, 122, 133, 120, 133, 87, 133, 142, 133, 150, 133, 134, 133, 141, 133, 153, 133, 157, 133, 129, 133, 162, 133, 130, 133, 136, 133, 133, 133, 121, 133, 118, 133, 152, 133, 144, 133, 159, 133, 104, 134, 190, 135, 170, 135, 173, 135, 197, 135, 176, 135, 172, 135, 185, 135, 181, 135, 188, 135, 174, 135, 201, 135, 195, 135, 194, 135, 204, 135, 183, 135, 175, 135, 196, 135, 202, 135, 180, 135, 182, 135, 191, 135, 184, 135, 189, 135, 222, 135, 178, 135, 53, 137, 51, 137, 60, 137, 62, 137, 65, 137, 82, 137, 55, 137, 66, 137, 173, 137, 175, 137, 174, 137, 242, 137, 243, 137, 30, 139, 24, 139, 22, 139, 17, 139, 5, 139, 11, 139, 34, 139, 15, 139, 18, 139, 21, 139, 7, 139, 13, 139, 8, 139, 6, 139, 28, 139, 19, 139, 26, 139, 79, 140, 112, 140, 114, 140, 113, 140, 111, 140, 149, 140, 148, 140, 249, 140, 111, 141, 78, 142, 77, 142, 83, 142, 80, 142, 76, 142, 71, 142, 67, 143, 64, 143, 133, 144, 126, 144, 56, 145, 154, 145, 162, 145, 155, 145, 153, 145, 159, 145, 161, 145, 157, 145, 160, 145, 161, 147, 131, 147, 175, 147, 100, 147, 86, 147, 71, 147, 124, 147, 88, 147, 92, 147, 118, 147, 73, 147, 80, 147, 81, 147, 96, 147, 109, 147, 143, 147, 76, 147, 106, 147, 121, 147, 87, 147, 85, 147, 82, 147, 79, 147, 113, 147, 119, 147, 123, 147, 97, 147, 94, 147, 99, 147, 103, 147, 128, 147, 78, 147, 89, 147, 199, 149, 192, 149, 201, 149, 195, 149, 197, 149, 183, 149, 174, 150, 176, 150, 172, 150, 32, 151, 31, 151, 24, 151, 29, 151, 25, 151, 154, 151, 161, 151, 156, 151, 158, 151, 157, 151, 213, 151, 212, 151, 241, 151, 65, 152, 68, 152, 74, 152, 73, 152, 69, 152, 67, 152, 37, 153, 43, 153, 44, 153, 42, 153, 51, 153, 50, 153, 47, 153, 45, 153, 49, 153, 48, 153, 152, 153, 163, 153, 161, 153, 2, 154, 250, 153, 244, 153, 247, 153, 249, 153, 248, 153, 246, 153, 251, 153, 253, 153, 254, 153, 252, 153, 3, 154, 190, 154, 254, 154, 253, 154, 1, 155, 252, 154, 72, 155, 154, 155, 168, 155, 158, 155, 155, 155, 166, 155, 161, 155, 165, 155, 164, 155, 134, 155, 162, 155, 160, 155, 175, 155, 51, 157, 65, 157, 103, 157, 54, 157, 46, 157, 47, 157, 49, 157, 56, 157, 48, 157, 69, 157, 66, 157, 67, 157, 62, 157, 55, 157, 64, 157, 61, 157, 245, 127, 45, 157, 138, 158, 137, 158, 141, 158, 176, 158, 200, 158, 218, 158, 251, 158, 255, 158, 36, 159, 35, 159, 34, 159, 84, 159, 160, 159, 49, 81, 45, 81, 46, 81, 152, 86, 156, 86, 151, 86, 154, 86, 157, 86, 153, 86, 112, 89, 60, 91, 105, 92, 106, 92, 192, 93, 109, 94, 110, 94, 216, 97, 223, 97, 237, 97, 238, 97, 241, 97, 234, 97, 240, 97, 235, 97, 214, 97, 233, 97, 255, 100, 4, 101, 253, 100, 248, 100, 1, 101, 3, 101, 252, 100, 148, 101, 219, 101, 218, 102, 219, 102, 216, 102, 197, 106, 185, 106, 189, 106, 225, 106, 198, 106, 186, 106, 182, 106, 183, 106, 199, 106, 180, 106, 173, 106, 94, 107, 201, 107, 11, 108, 7, 112, 12, 112, 13, 112, 1, 112, 5, 112, 20, 112, 14, 112, 255, 111, 0, 112, 251, 111, 38, 112, 252, 111, 247, 111, 10, 112, 1, 114, 255, 113, 249, 113, 3, 114, 253, 113, 118, 115, 184, 116, 192, 116, 181, 116, 193, 116, 190, 116, 182, 116, 187, 116, 194, 116, 20, 117, 19, 117, 92, 118, 100, 118, 89, 118, 80, 118, 83, 118, 87, 118, 90, 118, 166, 118, 189, 118, 236, 118, 194, 119, 186, 119, 255, 120, 12, 121, 19, 121, 20, 121, 9, 121, 16, 121, 18, 121, 17, 121, 173, 121, 172, 121, 95, 122, 28, 124, 41, 124, 25, 124, 32, 124, 31, 124, 45, 124, 29, 124, 38, 124, 40, 124, 34, 124, 37, 124, 48, 124, 92, 126, 80, 126, 86, 126, 99, 126, 88, 126, 98, 126, 95, 126, 81, 126, 96, 126, 87, 126, 83, 126, 181, 127, 179, 127, 247, 127, 248, 127, 117, 128, 209, 129, 210, 129, 208, 129, 95, 130, 94, 130, 180, 133, 198, 133, 192, 133, 195, 133, 194, 133, 179, 133, 181, 133, 189, 133, 199, 133, 196, 133, 191, 133, 203, 133, 206, 133, 200, 133, 197, 133, 177, 133, 182, 133, 210, 133, 36, 134, 184, 133, 183, 133, 190, 133, 105, 134, 231, 135, 230, 135, 226, 135, 219, 135, 235, 135, 234, 135, 229, 135, 223, 135, 243, 135, 228, 135, 212, 135, 220, 135, 211, 135, 237, 135, 216, 135, 227, 135, 164, 135, 215, 135, 217, 135, 1, 136, 244, 135, 232, 135, 221, 135, 83, 137, 75, 137, 79, 137, 76, 137, 70, 137, 80, 137, 81, 137, 73, 137, 42, 139, 39, 139, 35, 139, 51, 139, 48, 139, 53, 139, 71, 139, 47, 139, 60, 139, 62, 139, 49, 139, 37, 139, 55, 139, 38, 139, 54, 139, 46, 139, 36, 139, 59, 139, 61, 139, 58, 139, 66, 140, 117, 140, 153, 140, 152, 140, 151, 140, 254, 140, 4, 141, 2, 141, 0, 141, 92, 142, 98, 142, 96, 142, 87, 142, 86, 142, 94, 142, 101, 142, 103, 142, 91, 142, 90, 142, 97, 142, 93, 142, 105, 142, 84, 142, 70, 143, 71, 143, 72, 143, 75, 143, 40, 145, 58, 145, 59, 145, 62, 145, 168, 145, 165, 145, 167, 145, 175, 145, 170, 145, 181, 147, 140, 147, 146, 147, 183, 147, 155, 147, 157, 147, 137, 147, 167, 147, 142, 147, 170, 147, 158, 147, 166, 147, 149, 147, 136, 147, 153, 147, 159, 147, 141, 147, 177, 147, 145, 147, 178, 147, 164, 147, 168, 147, 180, 147, 163, 147, 165, 147, 210, 149, 211, 149, 209, 149, 179, 150, 215, 150, 218, 150, 194, 93, 223, 150, 216, 150, 221, 150, 35, 151, 34, 151, 37, 151, 172, 151, 174, 151, 168, 151, 171, 151, 164, 151, 170, 151, 162, 151, 165, 151, 215, 151, 217, 151, 214, 151, 216, 151, 250, 151, 80, 152, 81, 152, 82, 152, 184, 152, 65, 153, 60, 153, 58, 153, 15, 154, 11, 154, 9, 154, 13, 154, 4, 154, 17, 154, 10, 154, 5, 154, 7, 154, 6, 154, 192, 154, 220, 154, 8, 155, 4, 155, 5, 155, 41, 155, 53, 155, 74, 155, 76, 155, 75, 155, 199, 155, 198, 155, 195, 155, 191, 155, 193, 155, 181, 155, 184, 155, 211, 155, 182, 155, 196, 155, 185, 155, 189, 155, 92, 157, 83, 157, 79, 157, 74, 157, 91, 157, 75, 157, 89, 157, 86, 157, 76, 157, 87, 157, 82, 157, 84, 157, 95, 157, 88, 157, 90, 157, 142, 158, 140, 158, 223, 158, 1, 159, 0, 159, 22, 159, 37, 159, 43, 159, 42, 159, 41, 159, 40, 159, 76, 159, 85, 159, 52, 81, 53, 81, 150, 82, 247, 82, 180, 83, 171, 86, 173, 86, 166, 86, 167, 86, 170, 86, 172, 86, 218, 88, 221, 88, 219, 88, 18, 89, 61, 91, 62, 91, 63, 91, 195, 93, 112, 94, 191, 95, 251, 97, 7, 101, 16, 101, 13, 101, 9, 101, 12, 101, 14, 101, 132, 101, 222, 101, 221, 101, 222, 102, 231, 106, 224, 106, 204, 106, 209, 106, 217, 106, 203, 106, 223, 106, 220, 106, 208, 106, 235, 106, 207, 106, 205, 106, 222, 106, 96, 107, 176, 107, 12, 108, 25, 112, 39, 112, 32, 112, 22, 112, 43, 112, 33, 112, 34, 112, 35, 112, 41, 112, 23, 112, 36, 112, 28, 112, 42, 112, 12, 114, 10, 114, 7, 114, 2, 114, 5, 114, 165, 114, 166, 114, 164, 114, 163, 114, 161, 114, 203, 116, 197, 116, 183, 116, 195, 116, 22, 117, 96, 118, 201, 119, 202, 119, 196, 119, 241, 119, 29, 121, 27, 121, 33, 121, 28, 121, 23, 121, 30, 121, 176, 121, 103, 122, 104, 122, 51, 124, 60, 124, 57, 124, 44, 124, 59, 124, 236, 124, 234, 124, 118, 126, 117, 126, 120, 126, 112, 126, 119, 126, 111, 126, 122, 126, 114, 126, 116, 126, 104, 126, 75, 127, 74, 127, 131, 127, 134, 127, 183, 127, 253, 127, 254, 127, 120, 128, 215, 129, 213, 129, 100, 130, 97, 130, 99, 130, 235, 133, 241, 133, 237, 133, 217, 133, 225, 133, 232, 133, 218, 133, 215, 133, 236, 133, 242, 133, 248, 133, 216, 133, 223, 133, 227, 133, 220, 133, 209, 133, 240, 133, 230, 133, 239, 133, 222, 133, 226, 133, 0, 136, 250, 135, 3, 136, 246, 135, 247, 135, 9, 136, 12, 136, 11, 136, 6, 136, 252, 135, 8, 136, 255, 135, 10, 136, 2, 136, 98, 137, 90, 137, 91, 137, 87, 137, 97, 137, 92, 137, 88, 137, 93, 137, 89, 137, 136, 137, 183, 137, 182, 137, 246, 137, 80, 139, 72, 139, 74, 139, 64, 139, 83, 139, 86, 139, 84, 139, 75, 139, 85, 139, 81, 139, 66, 139, 82, 139, 87, 139, 67, 140, 119, 140, 118, 140, 154, 140, 6, 141, 7, 141, 9, 141, 172, 141, 170, 141, 173, 141, 171, 141, 109, 142, 120, 142, 115, 142, 106, 142, 111, 142, 123, 142, 194, 142, 82, 143, 81, 143, 79, 143, 80, 143, 83, 143, 180, 143, 64, 145, 63, 145, 176, 145, 173, 145, 222, 147, 199, 147, 207, 147, 194, 147, 218, 147, 208, 147, 249, 147, 236, 147, 204, 147, 217, 147, 169, 147, 230, 147, 202, 147, 212, 147, 238, 147, 227, 147, 213, 147, 196, 147, 206, 147, 192, 147, 210, 147, 231, 147, 125, 149, 218, 149, 219, 149, 225, 150, 41, 151, 43, 151, 44, 151, 40, 151, 38, 151, 179, 151, 183, 151, 182, 151, 221, 151, 222, 151, 223, 151, 92, 152, 89, 152, 93, 152, 87, 152, 191, 152, 189, 152, 187, 152, 190, 152, 72, 153, 71, 153, 67, 153, 166, 153, 167, 153, 26, 154, 21, 154, 37, 154, 29, 154, 36, 154, 27, 154, 34, 154, 32, 154, 39, 154, 35, 154, 30, 154, 28, 154, 20, 154, 194, 154, 11, 155, 10, 155, 14, 155, 12, 155, 55, 155, 234, 155, 235, 155, 224, 155, 222, 155, 228, 155, 230, 155, 226, 155, 240, 155, 212, 155, 215, 155, 236, 155, 220, 155, 217, 155, 229, 155, 213, 155, 225, 155, 218, 155, 119, 157, 129, 157, 138, 157, 132, 157, 136, 157, 113, 157, 128, 157, 120, 157, 134, 157, 139, 157, 140, 157, 125, 157, 107, 157, 116, 157, 117, 157, 112, 157, 105, 157, 133, 157, 115, 157, 123, 157, 130, 157, 111, 157, 121, 157, 127, 157, 135, 157, 104, 157, 148, 158, 145, 158, 192, 158, 252, 158, 45, 159, 64, 159, 65, 159, 77, 159, 86, 159, 87, 159, 88, 159, 55, 83, 178, 86, 181, 86, 179, 86, 227, 88, 69, 91, 198, 93, 199, 93, 238, 94, 239, 94, 192, 95, 193, 95, 249, 97, 23, 101, 22, 101, 21, 101, 19, 101, 223, 101, 232, 102, 227, 102, 228, 102, 243, 106, 240, 106, 234, 106, 232, 106, 249, 106, 241, 106, 238, 106, 239, 106, 60, 112, 53, 112, 47, 112, 55, 112, 52, 112, 49, 112, 66, 112, 56, 112, 63, 112, 58, 112, 57, 112, 64, 112, 59, 112, 51, 112, 65, 112, 19, 114, 20, 114, 168, 114, 125, 115, 124, 115, 186, 116, 171, 118, 170, 118, 190, 118, 237, 118, 204, 119, 206, 119, 207, 119, 205, 119, 242, 119, 37, 121, 35, 121, 39, 121, 40, 121, 36, 121, 41, 121, 178, 121, 110, 122, 108, 122, 109, 122, 247, 122, 73, 124, 72, 124, 74, 124, 71, 124, 69, 124, 238, 124, 123, 126, 126, 126, 129, 126, 128, 126, 186, 127, 255, 127, 121, 128, 219, 129, 217, 129, 11, 130, 104, 130, 105, 130, 34, 134, 255, 133, 1, 134, 254, 133, 27, 134, 0, 134, 246, 133, 4, 134, 9, 134, 5, 134, 12, 134, 253, 133, 25, 136, 16, 136, 17, 136, 23, 136, 19, 136, 22, 136, 99, 137, 102, 137, 185, 137, 247, 137, 96, 139, 106, 139, 93, 139, 104, 139, 99, 139, 101, 139, 103, 139, 109, 139, 174, 141, 134, 142, 136, 142, 132, 142, 89, 143, 86, 143, 87, 143, 85, 143, 88, 143, 90, 143, 141, 144, 67, 145, 65, 145, 183, 145, 181, 145, 178, 145, 179, 145, 11, 148, 19, 148, 251, 147, 32, 148, 15, 148, 20, 148, 254, 147, 21, 148, 16, 148, 40, 148, 25, 148, 13, 148, 245, 147, 0, 148, 247, 147, 7, 148, 14, 148, 22, 148, 18, 148, 250, 147, 9, 148, 248, 147, 10, 148, 255, 147, 252, 147, 12, 148, 246, 147, 17, 148, 6, 148, 222, 149, 224, 149, 223, 149, 46, 151, 47, 151, 185, 151, 187, 151, 253, 151, 254, 151, 96, 152, 98, 152, 99, 152, 95, 152, 193, 152, 194, 152, 80, 153, 78, 153, 89, 153, 76, 153, 75, 153, 83, 153, 50, 154, 52, 154, 49, 154, 44, 154, 42, 154, 54, 154, 41, 154, 46, 154, 56, 154, 45, 154, 199, 154, 202, 154, 198, 154, 16, 155, 18, 155, 17, 155, 11, 156, 8, 156, 247, 155, 5, 156, 18, 156, 248, 155, 64, 156, 7, 156, 14, 156, 6, 156, 23, 156, 20, 156, 9, 156, 159, 157, 153, 157, 164, 157, 157, 157, 146, 157, 152, 157, 144, 157, 155, 157, 160, 157, 148, 157, 156, 157, 170, 157, 151, 157, 161, 157, 154, 157, 162, 157, 168, 157, 158, 157, 163, 157, 191, 157, 169, 157, 150, 157, 166, 157, 167, 157, 153, 158, 155, 158, 154, 158, 229, 158, 228, 158, 231, 158, 230, 158, 48, 159, 46, 159, 91, 159, 96, 159, 94, 159, 93, 159, 89, 159, 145, 159, 58, 81, 57, 81, 152, 82, 151, 82, 195, 86, 189, 86, 190, 86, 72, 91, 71, 91, 203, 93, 207, 93, 241, 94, 253, 97, 27, 101, 2, 107, 252, 106, 3, 107, 248, 106, 0, 107, 67, 112, 68, 112, 74, 112, 72, 112, 73, 112, 69, 112, 70, 112, 29, 114, 26, 114, 25, 114, 126, 115, 23, 117, 106, 118, 208, 119, 45, 121, 49, 121, 47, 121, 84, 124, 83, 124, 242, 124, 138, 126, 135, 126, 136, 126, 139, 126, 134, 126, 141, 126, 77, 127, 187, 127, 48, 128, 221, 129, 24, 134, 42, 134, 38, 134, 31, 134, 35, 134, 28, 134, 25, 134, 39, 134, 46, 134, 33, 134, 32, 134, 41, 134, 30, 134, 37, 134, 41, 136, 29, 136, 27, 136, 32, 136, 36, 136, 28, 136, 43, 136, 74, 136, 109, 137, 105, 137, 110, 137, 107, 137, 250, 137, 121, 139, 120, 139, 69, 139, 122, 139, 123, 139, 16, 141, 20, 141, 175, 141, 142, 142, 140, 142, 94, 143, 91, 143, 93, 143, 70, 145, 68, 145, 69, 145, 185, 145, 63, 148, 59, 148, 54, 148, 41, 148, 61, 148, 60, 148, 48, 148, 57, 148, 42, 148, 55, 148, 44, 148, 64, 148, 49, 148, 229, 149, 228, 149, 227, 149, 53, 151, 58, 151, 191, 151, 225, 151, 100, 152, 201, 152, 198, 152, 192, 152, 88, 153, 86, 153, 57, 154, 61, 154, 70, 154, 68, 154, 66, 154, 65, 154, 58, 154, 63, 154, 205, 154, 21, 155, 23, 155, 24, 155, 22, 155, 58, 155, 82, 155, 43, 156, 29, 156, 28, 156, 44, 156, 35, 156, 40, 156, 41, 156, 36, 156, 33, 156, 183, 157, 182, 157, 188, 157, 193, 157, 199, 157, 202, 157, 207, 157, 190, 157, 197, 157, 195, 157, 187, 157, 181, 157, 206, 157, 185, 157, 186, 157, 172, 157, 200, 157, 177, 157, 173, 157, 204, 157, 179, 157, 205, 157, 178, 157, 122, 158, 156, 158, 235, 158, 238, 158, 237, 158, 27, 159, 24, 159, 26, 159, 49, 159, 78, 159, 101, 159, 100, 159, 146, 159, 185, 78, 198, 86, 197, 86, 203, 86, 113, 89, 75, 91, 76, 91, 213, 93, 209, 93, 242, 94, 33, 101, 32, 101, 38, 101, 34, 101, 11, 107, 8, 107, 9, 107, 13, 108, 85, 112, 86, 112, 87, 112, 82, 112, 30, 114, 31, 114, 169, 114, 127, 115, 216, 116, 213, 116, 217, 116, 215, 116, 109, 118, 173, 118, 53, 121, 180, 121, 112, 122, 113, 122, 87, 124, 92, 124, 89, 124, 91, 124, 90, 124, 244, 124, 241, 124, 145, 126, 79, 127, 135, 127, 222, 129, 107, 130, 52, 134, 53, 134, 51, 134, 44, 134, 50, 134, 54, 134, 44, 136, 40, 136, 38, 136, 42, 136, 37, 136, 113, 137, 191, 137, 190, 137, 251, 137, 126, 139, 132, 139, 130, 139, 134, 139, 133, 139, 127, 139, 21, 141, 149, 142, 148, 142, 154, 142, 146, 142, 144, 142, 150, 142, 151, 142, 96, 143, 98, 143, 71, 145, 76, 148, 80, 148, 74, 148, 75, 148, 79, 148, 71, 148, 69, 148, 72, 148, 73, 148, 70, 148, 63, 151, 227, 151, 106, 152, 105, 152, 203, 152, 84, 153, 91, 153, 78, 154, 83, 154, 84, 154, 76, 154, 79, 154, 72, 154, 74, 154, 73, 154, 82, 154, 80, 154, 208, 154, 25, 155, 43, 155, 59, 155, 86, 155, 85, 155, 70, 156, 72, 156, 63, 156, 68, 156, 57, 156, 51, 156, 65, 156, 60, 156, 55, 156, 52, 156, 50, 156, 61, 156, 54, 156, 219, 157, 210, 157, 222, 157, 218, 157, 203, 157, 208, 157, 220, 157, 209, 157, 223, 157, 233, 157, 217, 157, 216, 157, 214, 157, 245, 157, 213, 157, 221, 157, 182, 158, 240, 158, 53, 159, 51, 159, 50, 159, 66, 159, 107, 159, 149, 159, 162, 159, 61, 81, 153, 82, 232, 88, 231, 88, 114, 89, 77, 91, 216, 93, 47, 136, 79, 95, 1, 98, 3, 98, 4, 98, 41, 101, 37, 101, 150, 101, 235, 102, 17, 107, 18, 107, 15, 107, 202, 107, 91, 112, 90, 112, 34, 114, 130, 115, 129, 115, 131, 115, 112, 118, 212, 119, 103, 124, 102, 124, 149, 126, 108, 130, 58, 134, 64, 134, 57, 134, 60, 134, 49, 134, 59, 134, 62, 134, 48, 136, 50, 136, 46, 136, 51, 136, 118, 137, 116, 137, 115, 137, 254, 137, 140, 139, 142, 139, 139, 139, 136, 139, 69, 140, 25, 141, 152, 142, 100, 143, 99, 143, 188, 145, 98, 148, 85, 148, 93, 148, 87, 148, 94, 148, 196, 151, 197, 151, 0, 152, 86, 154, 89, 154, 30, 155, 31, 155, 32, 155, 82, 156, 88, 156, 80, 156, 74, 156, 77, 156, 75, 156, 85, 156, 89, 156, 76, 156, 78, 156, 251, 157, 247, 157, 239, 157, 227, 157, 235, 157, 248, 157, 228, 157, 246, 157, 225, 157, 238, 157, 230, 157, 242, 157, 240, 157, 226, 157, 236, 157, 244, 157, 243, 157, 232, 157, 237, 157, 194, 158, 208, 158, 242, 158, 243, 158, 6, 159, 28, 159, 56, 159, 55, 159, 54, 159, 67, 159, 79, 159, 113, 159, 112, 159, 110, 159, 111, 159, 211, 86, 205, 86, 78, 91, 109, 92, 45, 101, 237, 102, 238, 102, 19, 107, 95, 112, 97, 112, 93, 112, 96, 112, 35, 114, 219, 116, 229, 116, 213, 119, 56, 121, 183, 121, 182, 121, 106, 124, 151, 126, 137, 127, 109, 130, 67, 134, 56, 136, 55, 136, 53, 136, 75, 136, 148, 139, 149, 139, 158, 142, 159, 142, 160, 142, 157, 142, 190, 145, 189, 145, 194, 145, 107, 148, 104, 148, 105, 148, 229, 150, 70, 151, 67, 151, 71, 151, 199, 151, 229, 151, 94, 154, 213, 154, 89, 155, 99, 156, 103, 156, 102, 156, 98, 156, 94, 156, 96, 156, 2, 158, 254, 157, 7, 158, 3, 158, 6, 158, 5, 158, 0, 158, 1, 158, 9, 158, 255, 157, 253, 157, 4, 158, 160, 158, 30, 159, 70, 159, 116, 159, 117, 159, 118, 159, 212, 86, 46, 101, 184, 101, 24, 107, 25, 107, 23, 107, 26, 107, 98, 112, 38, 114, 170, 114, 216, 119, 217, 119, 57, 121, 105, 124, 107, 124, 246, 124, 154, 126, 152, 126, 155, 126, 153, 126, 224, 129, 225, 129, 70, 134, 71, 134, 72, 134, 121, 137, 122, 137, 124, 137, 123, 137, 255, 137, 152, 139, 153, 139, 165, 142, 164, 142, 163, 142, 110, 148, 109, 148, 111, 148, 113, 148, 115, 148, 73, 151, 114, 152, 95, 153, 104, 156, 110, 156, 109, 156, 11, 158, 13, 158, 16, 158, 15, 158, 18, 158, 17, 158, 161, 158, 245, 158, 9, 159, 71, 159, 120, 159, 123, 159, 122, 159, 121, 159, 30, 87, 102, 112, 111, 124, 60, 136, 178, 141, 166, 142, 195, 145, 116, 148, 120, 148, 118, 148, 117, 148, 96, 154, 116, 156, 115, 156, 113, 156, 117, 156, 20, 158, 19, 158, 246, 158, 10, 159, 164, 159, 104, 112, 101, 112, 247, 124, 106, 134, 62, 136, 61, 136, 63, 136, 158, 139, 156, 140, 169, 142, 201, 142, 75, 151, 115, 152, 116, 152, 204, 152, 97, 153, 171, 153, 100, 154, 102, 154, 103, 154, 36, 155, 21, 158, 23, 158, 72, 159, 7, 98, 30, 107, 39, 114, 76, 134, 168, 142, 130, 148, 128, 148, 129, 148, 105, 154, 104, 154, 46, 155, 25, 158, 41, 114, 75, 134, 159, 139, 131, 148, 121, 156, 183, 158, 117, 118, 107, 154, 122, 156, 29, 158, 105, 112, 106, 112, 164, 158, 126, 159, 73, 159, 152, 159, 0, 0, 239, 223, 29, 105, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 122, 0, 104, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 112, 0, 105, 0, 110, 0, 121, 0, 105, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 72, 0, 97, 0, 110, 0, 105, 0, 93, 0, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 114, 0, 101, 0, 103, 0, 117, 0, 108, 0, 97, 0, 114, 0, 93, 0, 60, 0, 39, 0, 208, 253, 65, 0, 39, 0, 60, 0, 42, 0, 63, 150, 21, 149, 60, 0, 42, 0, 196, 85, 60, 0, 42, 0, 74, 85, 60, 0, 42, 0, 206, 84, 192, 84, 9, 85, 195, 87, 45, 90, 40, 99, 56, 107, 190, 110, 243, 85, 176, 146, 63, 149, 111, 86, 132, 147, 60, 0, 42, 0, 64, 85, 113, 99, 145, 118, 210, 81, 176, 110, 10, 86, 113, 101, 115, 101, 154, 118, 76, 118, 3, 154, 60, 0, 42, 0, 208, 107, 57, 102, 62, 90, 238, 119, 60, 133, 183, 142, 237, 111, 249, 133, 106, 139, 45, 151, 68, 151, 60, 0, 42, 0, 126, 130, 12, 79, 49, 114, 57, 120, 75, 120, 152, 150, 204, 85, 103, 88, 210, 90, 27, 97, 141, 120, 198, 83, 167, 102, 119, 116, 161, 149, 254, 80, 210, 88, 33, 91, 211, 97, 134, 133, 49, 157, 221, 97, 214, 102, 166, 116, 50, 153, 167, 118, 185, 119, 164, 153, 25, 121, 122, 139, 64, 148, 107, 156, 73, 151, 60, 0, 42, 0, 137, 91, 146, 79, 214, 92, 73, 104, 40, 108, 181, 94, 244, 131, 25, 140, 149, 90, 59, 132, 74, 132, 247, 117, 100, 129, 76, 158, 237, 132, 157, 138, 140, 151, 141, 151, 230, 118, 243, 138, 163, 153, 235, 118, 106, 157, 253, 151, 149, 157, 60, 0, 42, 0, 181, 115, 125, 85, 248, 150, 17, 81, 60, 0, 42, 0, 181, 87, 250, 79, 53, 85, 239, 87, 245, 148, 140, 150, 222, 99, 111, 127, 168, 146, 60, 0, 42, 0, 180, 114, 184, 92, 9, 99, 29, 109, 76, 131, 72, 104, 250, 128, 123, 140, 19, 88, 105, 90, 123, 102, 151, 102, 139, 140, 12, 147, 199, 149, 159, 155, 239, 158, 60, 0, 42, 0, 174, 128, 175, 154, 60, 0, 42, 0, 108, 83, 135, 92, 2, 102, 59, 102, 60, 0, 42, 0, 138, 103, 206, 118, 160, 145, 60, 0, 42, 0, 249, 81, 234, 103, 142, 104, 234, 142, 10, 114, 60, 0, 42, 0, 86, 101, 171, 83, 158, 150, 247, 85, 248, 85, 133, 93, 210, 94, 246, 110, 83, 115, 28, 133, 104, 144, 110, 100, 172, 113, 82, 115, 136, 116, 221, 120, 241, 127, 113, 128, 175, 135, 246, 127, 55, 139, 56, 139, 250, 127, 204, 156, 214, 147, 50, 156, 212, 157, 7, 159, 60, 0, 42, 0, 157, 98, 186, 130, 215, 98, 132, 136, 122, 149, 170, 90, 188, 90, 86, 137, 60, 0, 42, 0, 153, 92, 119, 98, 115, 87, 135, 87, 176, 92, 178, 80, 97, 89, 101, 89, 103, 89, 239, 90, 96, 97, 156, 154, 169, 150, 186, 88, 180, 93, 202, 97, 179, 111, 217, 100, 202, 147, 65, 154, 60, 0, 39, 0, 208, 253, 66, 0, 39, 0, 60, 0, 42, 0, 55, 78, 107, 81, 200, 78, 244, 93, 237, 83, 82, 98, 51, 103, 144, 115, 63, 89, 156, 92, 173, 130, 199, 92, 237, 103, 164, 117, 245, 84, 252, 93, 76, 99, 6, 123, 145, 124, 147, 127, 134, 134, 219, 145, 223, 145, 93, 140, 131, 156, 94, 155, 60, 0, 42, 0, 208, 83, 174, 114, 156, 98, 122, 87, 173, 89, 212, 98, 7, 131, 166, 112, 121, 118, 200, 128, 221, 131, 89, 138, 203, 141, 247, 142, 176, 152, 67, 155, 162, 88, 37, 159, 60, 0, 42, 0, 138, 98, 175, 148, 0, 146, 118, 151, 60, 0, 42, 0, 93, 87, 29, 95, 56, 114, 187, 87, 98, 127, 25, 128, 193, 141, 133, 156, 140, 156, 119, 127, 138, 155, 135, 137, 242, 119, 56, 151, 233, 88, 94, 112, 27, 107, 60, 0, 42, 0, 39, 84, 38, 125, 60, 0, 42, 0, 0, 99, 176, 99, 216, 100, 60, 0, 42, 0, 125, 118, 60, 0, 42, 0, 126, 118, 112, 79, 207, 103, 34, 104, 109, 99, 248, 116, 225, 122, 168, 124, 84, 125, 70, 100, 250, 100, 108, 137, 60, 0, 42, 0, 141, 94, 221, 98, 37, 141, 220, 98, 87, 101, 8, 115, 23, 122, 186, 124, 173, 133, 1, 141, 219, 151, 60, 0, 42, 0, 115, 98, 61, 101, 237, 115, 44, 130, 129, 152, 145, 101, 44, 100, 146, 101, 18, 152, 34, 118, 251, 156, 140, 135, 41, 137, 77, 118, 172, 143, 60, 0, 42, 0, 42, 150, 66, 87, 133, 92, 4, 102, 127, 103, 72, 114, 234, 116, 163, 148, 132, 124, 40, 130, 17, 146, 66, 135, 108, 155, 198, 149, 60, 0, 42, 0, 158, 82, 74, 83, 52, 79, 110, 98, 98, 87, 197, 89, 17, 96, 204, 98, 202, 126, 200, 103, 218, 121, 116, 110, 70, 125, 97, 146, 125, 151, 166, 143, 227, 116, 60, 0, 42, 0, 129, 135, 60, 0, 42, 0, 166, 144, 185, 87, 46, 94, 96, 99, 134, 104, 92, 109, 171, 144, 71, 94, 90, 94, 13, 126, 107, 94, 164, 151, 60, 0, 42, 0, 209, 126, 129, 125, 156, 105, 83, 114, 128, 129, 200, 154, 60, 0, 42, 0, 164, 115, 140, 134, 141, 80, 210, 104, 211, 104, 36, 140, 93, 88, 18, 100, 22, 122, 161, 132, 47, 135, 197, 120, 81, 149, 85, 130, 23, 139, 138, 147, 60, 0, 42, 0, 249, 82, 5, 83, 104, 79, 98, 91, 222, 130, 185, 103, 222, 128, 35, 123, 114, 113, 133, 159, 84, 133, 18, 137, 67, 137, 193, 149, 89, 159, 60, 0, 42, 0, 135, 122, 209, 90, 249, 150, 132, 133, 60, 0, 42, 0, 157, 91, 9, 96, 113, 153, 221, 79, 40, 158, 178, 91, 228, 115, 33, 88, 34, 88, 172, 90, 70, 132, 218, 91, 249, 152, 253, 152, 19, 137, 194, 153, 245, 156, 229, 125, 7, 157, 242, 140, 245, 133, 243, 91, 246, 91, 76, 151, 60, 0, 42, 0, 253, 82, 165, 98, 177, 98, 121, 140, 181, 141, 199, 148, 226, 131, 171, 134, 140, 136, 49, 88, 75, 146, 141, 156, 100, 151, 178, 154, 180, 102, 241, 154, 99, 134, 145, 155, 36, 81, 211, 102, 6, 114, 193, 95, 100, 148, 60, 0, 42, 0, 66, 150, 81, 83, 111, 103, 195, 118, 110, 104, 178, 96, 249, 99, 17, 105, 145, 120, 78, 158, 227, 133, 111, 157, 60, 0, 42, 0, 23, 83, 115, 146, 60, 0, 42, 0, 29, 141, 91, 91, 200, 114, 157, 140, 182, 144, 7, 89, 1, 102, 108, 114, 221, 130, 204, 128, 229, 144, 161, 148, 251, 79, 13, 80, 150, 96, 253, 114, 171, 136, 93, 80, 121, 80, 150, 104, 252, 115, 1, 145, 153, 80, 195, 80, 235, 96, 25, 113, 50, 116, 240, 142, 136, 143, 2, 97, 154, 120, 153, 121, 211, 132, 253, 134, 149, 114, 25, 137, 150, 138, 129, 151, 179, 154, 41, 143, 199, 146, 138, 97, 210, 124, 180, 151, 62, 148, 60, 0, 42, 0, 87, 84, 4, 85, 60, 0, 42, 0, 84, 89, 205, 108, 50, 141, 31, 104, 135, 114, 27, 149, 27, 147, 60, 0, 42, 0, 44, 103, 239, 130, 89, 89, 90, 117, 201, 127, 77, 105, 60, 0, 42, 0, 75, 87, 76, 87, 52, 80, 121, 99, 115, 104, 0, 110, 40, 123, 41, 144, 170, 100, 86, 115, 61, 143, 60, 0, 42, 0, 59, 79, 74, 121, 95, 89, 41, 93, 247, 126, 99, 125, 141, 149, 176, 80, 109, 93, 237, 117, 35, 86, 179, 125, 60, 0, 42, 0, 45, 117, 60, 0, 42, 0, 196, 87, 242, 87, 246, 131, 35, 116, 43, 116, 67, 126, 155, 151, 60, 0, 42, 0, 245, 108, 248, 143, 44, 144, 116, 88, 15, 117, 90, 149, 102, 142, 240, 147, 60, 0, 42, 0, 252, 99, 60, 0, 42, 0, 68, 92, 106, 80, 244, 107, 60, 144, 69, 105, 140, 105, 77, 140, 149, 135, 86, 157, 190, 156, 158, 147, 15, 156, 60, 0, 42, 0, 120, 131, 59, 159, 54, 91, 60, 0, 42, 0, 21, 83, 212, 107, 54, 89, 60, 103, 74, 79, 33, 84, 163, 89, 152, 108, 149, 117, 124, 95, 192, 103, 213, 121, 254, 79, 20, 123, 131, 124, 45, 130, 90, 85, 70, 123, 25, 145, 132, 123, 91, 128, 143, 140, 60, 0, 42, 0, 1, 94, 197, 95, 213, 107, 237, 149, 86, 79, 82, 87, 135, 94, 208, 139, 178, 144, 188, 89, 45, 96, 54, 96, 136, 103, 64, 117, 65, 117, 254, 130, 212, 84, 242, 103, 214, 107, 204, 115, 170, 117, 92, 131, 91, 150, 217, 107, 244, 114, 98, 117, 19, 123, 138, 124, 144, 136, 203, 148, 98, 90, 179, 94, 93, 101, 144, 104, 6, 132, 30, 132, 135, 149, 137, 149, 27, 88, 59, 95, 60, 95, 10, 97, 14, 97, 98, 110, 149, 118, 134, 121, 90, 123, 86, 138, 177, 140, 193, 140, 81, 141, 246, 85, 67, 95, 215, 110, 237, 110, 79, 113, 249, 117, 250, 117, 100, 119, 119, 129, 214, 132, 253, 132, 12, 135, 232, 136, 248, 141, 77, 146, 159, 149, 246, 152, 99, 94, 10, 95, 154, 113, 89, 115, 167, 120, 133, 123, 134, 123, 188, 125, 61, 133, 42, 145, 157, 153, 100, 94, 119, 111, 88, 115, 124, 127, 69, 137, 220, 153, 242, 154, 193, 88, 22, 91, 230, 94, 230, 123, 243, 123, 42, 126, 156, 133, 241, 137, 127, 144, 133, 155, 131, 101, 222, 111, 194, 129, 85, 142, 192, 154, 112, 89, 167, 116, 40, 145, 206, 147, 70, 153, 116, 126, 94, 137, 99, 137, 184, 151, 224, 151, 83, 155, 131, 142, 132, 142, 70, 154, 20, 141, 52, 148, 221, 157, 233, 157, 10, 159, 60, 0, 42, 0, 185, 143, 186, 143, 45, 120, 62, 123, 217, 99, 53, 115, 22, 127, 120, 113, 81, 114, 2, 117, 175, 123, 232, 125, 89, 135, 113, 115, 137, 144, 125, 147, 202, 156, 138, 144, 173, 151, 254, 155, 255, 155, 105, 124, 60, 0, 42, 0, 44, 141, 65, 98, 134, 122, 62, 83, 182, 140, 252, 96, 57, 132, 165, 120, 40, 122, 10, 137, 196, 124, 24, 157, 202, 133, 60, 0, 42, 0, 94, 83, 1, 95, 237, 95, 131, 98, 115, 108, 116, 108, 196, 130, 198, 145, 216, 83, 197, 92, 163, 115, 191, 79, 9, 89, 42, 102, 141, 137, 167, 95, 15, 127, 77, 144, 158, 149, 161, 143, 246, 125, 81, 130, 167, 143, 168, 143, 169, 143, 171, 143, 174, 143, 175, 143, 138, 139, 60, 0, 42, 0, 158, 112, 60, 0, 42, 0, 108, 112, 83, 103, 7, 104, 209, 152, 137, 154, 223, 154, 242, 109, 106, 95, 11, 115, 63, 129, 169, 152, 130, 88, 86, 94, 125, 100, 238, 110, 8, 133, 174, 152, 160, 154, 25, 106, 155, 113, 152, 129, 45, 118, 230, 120, 86, 149, 217, 152, 218, 152, 38, 81, 183, 152, 12, 112, 232, 133, 36, 139, 2, 114, 213, 129, 6, 141, 226, 147, 110, 122, 115, 149, 198, 152, 199, 152, 200, 152, 202, 152, 67, 154, 99, 148, 107, 154, 60, 0, 42, 0, 104, 136, 74, 90, 241, 136, 216, 138, 62, 137, 54, 147, 166, 106, 60, 0, 42, 0, 245, 79, 212, 156, 62, 156, 60, 0, 42, 0, 139, 97, 222, 135, 214, 156, 73, 156, 8, 159, 76, 134, 158, 159, 60, 0, 42, 0, 37, 82, 43, 82, 135, 84, 130, 131, 194, 134, 182, 95, 82, 137, 105, 142, 60, 0, 42, 0, 42, 118, 95, 118, 60, 0, 42, 0, 70, 95, 60, 0, 42, 0, 67, 108, 160, 144, 162, 115, 15, 120, 190, 91, 108, 95, 185, 104, 167, 80, 140, 101, 21, 105, 232, 110, 36, 127, 223, 105, 120, 116, 105, 140, 211, 140, 212, 140, 84, 149, 16, 81, 210, 111, 241, 111, 245, 111, 104, 134, 115, 140, 179, 106, 184, 116, 21, 112, 38, 151, 125, 126, 76, 148, 110, 152, 60, 0, 42, 0, 72, 100, 161, 107, 145, 129, 233, 154, 239, 100, 2, 155, 175, 107, 207, 129, 204, 154, 19, 155, 213, 154, 34, 155, 60, 0, 42, 0, 30, 108, 60, 0, 42, 0, 171, 81, 204, 78, 210, 78, 55, 108, 176, 81, 117, 81, 164, 99, 242, 146, 60, 0, 42, 0, 25, 78, 180, 144, 67, 150, 50, 96, 166, 98, 201, 121, 234, 130, 30, 102, 58, 102, 196, 103, 179, 112, 124, 153, 42, 119, 137, 122, 195, 134, 82, 100, 197, 104, 128, 121, 31, 122, 53, 146, 124, 146, 134, 151, 5, 153, 32, 153, 158, 151, 60, 0, 42, 0, 118, 94, 38, 78, 117, 79, 119, 94, 170, 87, 176, 94, 2, 80, 36, 104, 197, 117, 221, 122, 75, 80, 161, 80, 206, 91, 129, 138, 169, 155, 80, 151, 60, 0, 42, 0, 118, 118, 23, 94, 232, 98, 226, 108, 119, 118, 187, 115, 93, 82, 101, 82, 241, 84, 203, 118, 53, 120, 154, 136, 181, 148, 125, 153, 117, 85, 52, 125, 61, 127, 224, 131, 176, 136, 138, 110, 134, 120, 98, 146, 224, 80, 147, 93, 165, 100, 173, 100, 17, 153, 129, 155, 115, 142, 75, 154, 77, 156, 60, 0, 42, 0, 226, 78, 47, 79, 187, 114, 145, 128, 115, 154, 27, 94, 219, 114, 221, 116, 233, 130, 188, 79, 195, 82, 201, 128, 227, 144, 179, 78, 44, 99, 97, 109, 223, 116, 225, 121, 175, 136, 185, 148, 194, 148, 255, 109, 22, 129, 54, 130, 185, 136, 90, 83, 36, 110, 103, 132, 65, 158, 61, 97, 15, 100, 60, 115, 56, 146, 81, 146, 142, 153, 240, 80, 127, 113, 84, 114, 148, 123, 138, 129, 74, 130, 20, 133, 155, 153, 193, 153, 35, 142, 205, 146, 72, 149, 158, 153, 238, 153, 79, 137, 112, 140, 151, 86, 234, 97, 33, 121, 25, 124, 155, 147, 58, 153, 83, 157, 166, 114, 198, 154, 201, 154, 2, 107, 110, 137, 52, 121, 110, 148, 60, 0, 42, 0, 219, 141, 165, 123, 56, 124, 60, 0, 42, 0, 121, 91, 151, 106, 234, 124, 82, 139, 23, 134, 60, 0, 42, 0, 92, 83, 33, 132, 60, 0, 42, 0, 236, 92, 175, 94, 11, 144, 184, 148, 97, 102, 61, 146, 167, 138, 60, 0, 42, 0, 234, 156, 80, 143, 173, 145, 60, 0, 42, 0, 95, 83, 101, 136, 250, 84, 85, 99, 184, 85, 220, 136, 79, 157, 20, 158, 60, 0, 42, 0, 13, 78, 3, 94, 72, 79, 37, 84, 101, 107, 152, 84, 22, 96, 170, 98, 104, 107, 105, 107, 232, 103, 154, 148, 207, 82, 212, 87, 215, 87, 145, 96, 87, 99, 121, 131, 232, 144, 224, 87, 255, 116, 205, 94, 0, 133, 4, 142, 246, 144, 20, 153, 240, 123, 34, 153, 63, 124, 60, 0, 39, 0, 208, 253, 67, 0, 39, 0, 60, 0, 42, 0, 147, 86, 230, 100, 3, 101, 60, 0, 42, 0, 36, 121, 56, 121, 60, 0, 42, 0, 106, 144, 195, 86, 60, 0, 42, 0, 114, 80, 28, 115, 60, 0, 42, 0, 77, 98, 80, 103, 34, 141, 161, 140, 193, 136, 148, 126, 60, 0, 42, 0, 221, 107, 199, 145, 56, 80, 75, 85, 71, 90, 192, 91, 105, 95, 161, 99, 108, 119, 244, 141, 181, 125, 41, 142, 60, 0, 42, 0, 240, 87, 220, 131, 204, 104, 33, 133, 41, 126, 60, 0, 42, 0, 194, 83, 195, 83, 196, 83, 225, 152, 150, 154, 197, 83, 176, 85, 76, 110, 170, 80, 32, 91, 16, 153, 24, 114, 66, 154, 60, 0, 42, 0, 139, 107, 149, 134, 237, 96, 152, 107, 90, 97, 69, 135, 89, 97, 54, 136, 58, 136, 60, 0, 42, 0, 232, 96, 1, 103, 88, 97, 175, 97, 71, 122, 248, 123, 234, 158, 242, 158, 60, 0, 42, 0, 127, 112, 186, 99, 113, 91, 178, 124, 123, 100, 15, 81, 175, 111, 146, 133, 230, 113, 168, 116, 50, 139, 60, 0, 42, 0, 211, 78, 250, 78, 39, 79, 167, 108, 205, 130, 39, 158, 9, 80, 49, 130, 150, 80, 98, 93, 196, 110, 74, 115, 188, 132, 248, 111, 89, 130, 165, 135, 172, 157, 60, 0, 42, 0, 207, 133, 12, 107, 118, 148, 60, 0, 42, 0, 246, 140, 60, 0, 42, 0, 161, 100, 205, 100, 217, 124, 60, 0, 42, 0, 250, 102, 249, 102, 8, 86, 134, 93, 21, 111, 248, 132, 253, 105, 63, 137, 90, 130, 172, 135, 234, 147, 60, 0, 42, 0, 120, 130, 73, 131, 58, 97, 198, 97, 50, 154, 60, 0, 42, 0, 143, 128, 53, 145, 89, 137, 60, 0, 42, 0, 121, 130, 60, 0, 42, 0, 138, 81, 140, 81, 167, 79, 149, 83, 123, 96, 250, 98, 75, 109, 93, 131, 71, 101, 95, 117, 116, 80, 160, 83, 39, 123, 163, 124, 23, 132, 193, 94, 251, 96, 44, 110, 86, 123, 52, 132, 94, 123, 116, 123, 219, 132, 132, 88, 163, 123, 161, 97, 14, 124, 60, 0, 42, 0, 126, 93, 60, 0, 42, 0, 145, 92, 148, 109, 18, 123, 163, 104, 60, 0, 42, 0, 253, 102, 76, 86, 60, 0, 42, 0, 66, 92, 254, 102, 100, 92, 146, 93, 242, 122, 83, 154, 60, 0, 42, 0, 109, 142, 60, 0, 42, 0, 201, 83, 96, 98, 72, 103, 158, 128, 255, 129, 63, 99, 91, 80, 207, 85, 210, 99, 247, 99, 135, 153, 159, 146, 56, 149, 86, 130, 128, 117, 100, 147, 55, 153, 60, 0, 42, 0, 197, 121, 158, 87, 229, 103, 44, 131, 54, 131, 86, 93, 61, 100, 57, 115, 107, 151, 206, 105, 103, 138, 223, 91, 180, 120, 171, 106, 60, 0, 42, 0, 105, 136, 69, 142, 114, 149, 84, 148, 60, 0, 42, 0, 124, 89, 74, 108, 148, 92, 152, 79, 231, 139, 249, 89, 238, 93, 1, 125, 107, 138, 60, 0, 42, 0, 134, 130, 198, 98, 151, 148, 245, 145, 60, 0, 42, 0, 178, 114, 170, 79, 244, 103, 122, 140, 97, 121, 141, 85, 21, 81, 92, 159, 60, 0, 42, 0, 29, 131, 60, 0, 42, 0, 127, 134, 131, 136, 13, 138, 37, 118, 6, 136, 198, 86, 60, 0, 42, 0, 191, 143, 199, 137, 180, 104, 0, 100, 152, 137, 231, 136, 70, 146, 211, 146, 104, 94, 92, 137, 25, 101, 60, 0, 42, 0, 117, 90, 23, 140, 206, 104, 121, 110, 133, 121, 139, 153, 88, 113, 32, 127, 221, 80, 81, 115, 73, 135, 151, 138, 203, 146, 3, 81, 11, 91, 219, 94, 121, 111, 122, 111, 254, 125, 182, 111, 219, 120, 170, 121, 218, 107, 61, 145, 97, 149, 13, 112, 236, 135, 51, 81, 150, 82, 254, 135, 65, 145, 181, 86, 229, 88, 201, 93, 58, 112, 3, 107, 143, 126, 146, 126, 148, 142, 117, 149, 108, 130, 146, 139, 113, 148, 94, 153, 60, 0, 42, 0, 167, 78, 44, 82, 245, 101, 51, 78, 186, 101, 80, 109, 87, 82, 4, 140, 116, 85, 34, 117, 35, 117, 242, 148, 16, 150, 135, 132, 119, 82, 124, 93, 76, 100, 251, 110, 61, 86, 93, 94, 70, 133, 194, 138, 179, 149, 163, 154, 192, 113, 5, 124, 129, 81, 95, 126, 66, 139, 180, 143, 223, 147, 225, 149, 197, 86, 91, 112, 135, 139, 60, 0, 42, 0, 207, 95, 95, 120, 114, 100, 244, 97, 164, 152, 250, 97, 188, 127, 194, 151, 107, 152, 60, 0, 42, 0, 37, 79, 12, 102, 0, 80, 60, 90, 208, 109, 22, 115, 214, 131, 10, 150, 127, 102, 25, 105, 41, 116, 238, 136, 32, 149, 41, 147, 182, 149, 179, 156, 231, 155, 26, 159, 60, 0, 42, 0, 231, 78, 233, 78, 79, 81, 160, 128, 204, 130, 120, 149, 29, 92, 127, 80, 56, 94, 156, 95, 250, 116, 7, 132, 30, 117, 120, 129, 23, 86, 114, 88, 230, 90, 122, 116, 147, 129, 255, 146, 31, 81, 144, 86, 191, 156, 219, 147, 104, 156, 60, 0, 42, 0, 130, 83, 58, 87, 54, 102, 221, 96, 52, 88, 94, 101, 216, 80, 176, 83, 224, 94, 5, 108, 249, 146, 60, 0, 42, 0, 5, 96, 154, 115, 69, 117, 33, 80, 47, 155, 49, 85, 181, 96, 59, 113, 82, 116, 162, 102, 124, 117, 175, 138, 212, 151, 60, 0, 42, 0, 208, 135, 60, 0, 42, 0, 132, 98, 40, 95, 10, 96, 41, 107, 158, 148, 44, 138, 47, 113, 133, 141, 20, 146, 230, 82, 60, 0, 42, 0, 74, 114, 65, 102, 226, 93, 227, 93, 29, 103, 27, 145, 12, 159, 5, 111, 50, 86, 20, 106, 110, 111, 178, 122, 122, 127, 72, 143, 2, 159, 63, 139, 60, 0, 42, 0, 53, 84, 146, 112, 39, 119, 35, 113, 124, 113, 168, 158, 208, 93, 60, 0, 42, 0, 230, 78, 239, 78, 22, 128, 216, 137, 60, 0, 42, 0, 102, 143, 33, 79, 202, 142, 229, 79, 23, 120, 19, 85, 151, 131, 104, 120, 252, 134, 60, 0, 42, 0, 111, 98, 86, 80, 166, 100, 60, 0, 42, 0, 110, 92, 123, 95, 124, 87, 224, 143, 226, 112, 69, 128, 163, 99, 105, 120, 25, 152, 185, 95, 164, 100, 136, 111, 246, 82, 174, 119, 33, 114, 60, 0, 42, 0, 187, 98, 244, 144, 117, 99, 27, 116, 212, 85, 157, 125, 139, 119, 195, 138, 221, 140, 29, 126, 19, 139, 60, 0, 42, 0, 24, 92, 227, 129, 241, 95, 136, 108, 137, 108, 176, 143, 72, 150, 231, 143, 30, 131, 184, 91, 128, 131, 144, 131, 115, 150, 80, 101, 104, 102, 38, 138, 12, 140, 217, 142, 22, 97, 232, 99, 2, 146, 65, 113, 47, 133, 117, 88, 4, 106, 14, 118, 3, 151, 180, 135, 246, 138, 188, 133, 142, 158, 223, 102, 208, 157, 60, 0, 42, 0, 187, 141, 118, 120, 156, 120, 139, 88, 38, 89, 227, 120, 56, 142, 86, 147, 2, 141, 166, 145, 60, 0, 42, 0, 108, 136, 162, 117, 128, 159, 129, 141, 130, 141, 135, 105, 83, 159, 84, 159, 45, 81, 171, 86, 54, 140, 236, 106, 111, 137, 150, 139, 60, 0, 42, 0, 55, 150, 223, 108, 253, 103, 47, 114, 230, 104, 126, 109, 36, 116, 240, 121, 65, 80, 207, 134, 94, 110, 90, 114, 106, 141, 220, 80, 134, 97, 90, 100, 49, 122, 87, 151, 144, 100, 145, 100, 253, 125, 85, 106, 160, 119, 108, 141, 51, 152, 137, 106, 192, 122, 73, 127, 106, 122, 246, 135, 243, 147, 255, 147, 35, 148, 83, 153, 60, 0, 42, 0, 30, 78, 16, 98, 62, 103, 72, 84, 127, 98, 168, 103, 218, 139, 213, 144, 87, 78, 206, 87, 13, 90, 172, 91, 248, 92, 6, 109, 127, 131, 88, 78, 213, 87, 48, 99, 95, 102, 249, 115, 0, 129, 129, 99, 245, 115, 128, 120, 154, 122, 45, 129, 214, 148, 24, 88, 233, 96, 214, 104, 9, 105, 11, 122, 108, 123, 126, 125, 206, 136, 77, 88, 86, 88, 151, 110, 160, 138, 123, 117, 114, 145, 238, 146, 149, 97, 130, 111, 132, 111, 89, 106, 153, 106, 206, 155, 19, 112, 242, 97, 44, 154, 60, 0, 42, 0, 177, 79, 142, 95, 156, 96, 30, 144, 139, 154, 177, 94, 72, 119, 1, 154, 60, 0, 42, 0, 228, 121, 60, 0, 42, 0, 3, 84, 155, 89, 153, 79, 231, 84, 104, 95, 245, 128, 169, 134, 49, 158, 251, 116, 53, 119, 30, 123, 154, 124, 171, 85, 53, 138, 228, 85, 184, 90, 91, 100, 244, 117, 122, 125, 68, 86, 157, 119, 186, 138, 173, 135, 31, 157, 68, 157, 97, 118, 81, 155, 93, 159, 33, 101, 114, 95, 208, 158, 60, 0, 42, 0, 27, 95, 96, 108, 112, 154, 223, 143, 123, 87, 187, 92, 12, 131, 1, 99, 254, 122, 78, 131, 109, 107, 179, 134, 127, 141, 66, 123, 190, 140, 69, 144, 141, 141, 95, 144, 179, 153, 142, 123, 128, 88, 38, 111, 31, 142, 114, 144, 234, 123, 24, 139, 60, 0, 42, 0, 58, 92, 250, 83, 78, 84, 148, 128, 136, 79, 118, 83, 127, 159, 145, 87, 227, 128, 101, 96, 59, 128, 135, 134, 179, 136, 60, 107, 111, 107, 178, 136, 237, 136, 121, 146, 43, 137, 82, 159, 60, 0, 42, 0, 115, 95, 241, 83, 165, 101, 88, 103, 123, 112, 100, 141, 108, 153, 182, 98, 197, 82, 92, 96, 189, 112, 209, 82, 196, 127, 197, 127, 85, 101, 254, 112, 211, 117, 123, 85, 65, 110, 115, 120, 237, 152, 186, 80, 248, 117, 95, 129, 238, 141, 83, 146, 244, 150, 143, 97, 8, 118, 228, 127, 107, 144, 144, 146, 87, 97, 27, 118, 232, 127, 190, 113, 216, 97, 169, 141, 78, 153, 146, 157, 216, 157, 60, 0, 42, 0, 182, 158, 60, 0, 42, 0, 69, 81, 178, 81, 225, 95, 150, 108, 58, 131, 122, 109, 235, 115, 192, 127, 2, 130, 3, 86, 79, 100, 184, 95, 131, 97, 167, 97, 93, 136, 127, 127, 95, 130, 86, 142, 60, 0, 42, 0, 107, 134, 7, 93, 8, 93, 128, 150, 8, 137, 223, 125, 105, 135, 242, 135, 30, 114, 60, 0, 42, 0, 160, 91, 235, 87, 245, 91, 60, 0, 42, 0, 243, 148, 240, 99, 131, 146, 60, 0, 42, 0, 189, 98, 100, 90, 10, 100, 51, 118, 216, 123, 168, 114, 171, 114, 60, 0, 42, 0, 199, 78, 30, 96, 230, 79, 49, 94, 38, 104, 198, 96, 44, 125, 248, 126, 215, 131, 6, 105, 116, 117, 82, 125, 1, 97, 151, 118, 32, 122, 121, 123, 239, 136, 103, 145, 108, 145, 162, 125, 12, 142, 20, 81, 212, 150, 139, 86, 38, 91, 108, 94, 228, 97, 181, 133, 253, 113, 224, 150, 135, 117, 76, 124, 138, 142, 187, 145, 142, 139, 144, 139, 60, 0, 42, 0, 17, 78, 18, 78, 28, 84, 123, 103, 125, 103, 180, 79, 98, 80, 133, 119, 156, 145, 193, 119, 87, 155, 60, 0, 42, 0, 237, 129, 240, 129, 90, 144, 160, 107, 60, 0, 42, 0, 250, 81, 128, 92, 29, 82, 116, 100, 23, 106, 153, 140, 99, 159, 60, 0, 42, 0, 13, 82, 100, 150, 187, 130, 168, 83, 193, 110, 162, 132, 96, 140, 4, 149, 176, 90, 33, 128, 173, 132, 13, 135, 142, 141, 79, 146, 207, 150, 147, 114, 79, 133, 218, 94, 232, 123, 228, 146, 113, 106, 110, 94, 201, 106, 248, 133, 245, 135, 135, 142, 219, 150, 229, 106, 112, 142, 181, 157, 149, 142, 60, 0, 42, 0, 230, 81, 117, 103, 64, 120, 24, 105, 168, 80, 110, 105, 90, 105, 26, 137, 203, 111, 50, 81, 154, 106, 180, 116, 14, 121, 109, 159, 124, 159, 60, 0, 42, 0, 141, 78, 4, 89, 204, 122, 53, 96, 192, 98, 204, 126, 86, 140, 247, 103, 42, 107, 208, 122, 246, 79, 74, 101, 92, 117, 241, 87, 255, 115, 64, 125, 85, 134, 151, 80, 33, 116, 16, 145, 16, 100, 192, 110, 235, 132, 230, 137, 0, 142, 166, 149, 10, 81, 60, 86, 212, 138, 183, 97, 182, 101, 92, 107, 197, 129, 220, 158, 248, 137, 215, 119, 60, 0, 42, 0, 139, 105, 123, 106, 60, 0, 42, 0, 59, 107, 88, 107, 60, 0, 42, 0, 227, 99, 11, 100, 60, 0, 42, 0, 151, 129, 60, 0, 42, 0, 92, 85, 44, 86, 170, 129, 57, 142, 60, 0, 42, 0, 219, 93, 221, 93, 26, 108, 127, 122, 118, 82, 45, 115, 79, 116, 60, 0, 42, 0, 32, 79, 33, 130, 41, 130, 57, 130, 12, 87, 68, 144, 179, 80, 61, 105, 183, 102, 197, 123, 50, 143, 60, 0, 42, 0, 27, 130, 72, 131, 152, 85, 66, 107, 226, 80, 51, 142, 60, 0, 42, 0, 76, 108, 50, 78, 148, 115, 143, 148, 231, 145, 215, 140, 168, 157, 60, 0, 42, 0, 5, 82, 174, 117, 147, 122, 151, 122, 78, 114, 80, 100, 85, 114, 33, 118, 187, 122, 60, 0, 42, 0, 138, 94, 64, 114, 71, 86, 98, 94, 60, 0, 42, 0, 239, 149, 184, 80, 100, 100, 226, 120, 214, 149, 60, 0, 42, 0, 27, 82, 6, 96, 49, 82, 79, 82, 89, 82, 212, 81, 117, 82, 52, 97, 60, 0, 42, 0, 57, 84, 138, 112, 161, 159, 60, 0, 42, 0, 130, 87, 21, 80, 192, 87, 114, 150, 118, 99, 217, 131, 37, 100, 240, 104, 14, 105, 68, 129, 204, 105, 36, 149, 160, 123, 24, 147, 154, 147, 64, 152, 60, 0, 42, 0, 254, 101, 118, 103, 37, 102, 5, 132, 62, 88, 139, 90, 153, 102, 63, 105, 198, 105, 67, 116, 186, 123, 125, 135, 65, 106, 52, 143, 196, 106, 6, 156, 158, 157, 60, 0, 42, 0, 175, 126, 89, 150, 7, 85, 113, 109, 20, 125, 188, 131, 243, 109, 35, 129, 123, 110, 137, 114, 227, 110, 147, 132, 81, 158, 24, 111, 244, 132, 135, 145, 149, 145, 30, 147, 217, 155, 137, 157, 60, 0, 42, 0, 70, 80, 54, 132, 247, 96, 118, 119, 240, 140, 34, 136, 60, 0, 42, 0, 52, 144, 20, 142, 51, 98, 60, 0, 42, 0, 182, 143, 181, 143, 21, 90, 22, 90, 124, 90, 217, 96, 176, 109, 240, 126, 79, 129, 141, 143, 107, 145, 189, 125, 160, 141, 31, 143, 138, 159, 201, 100, 237, 120, 91, 126, 96, 107, 189, 86, 106, 159, 97, 148, 60, 0, 42, 0, 114, 84, 181, 117, 128, 141, 104, 80, 208, 141, 18, 126, 180, 154, 202, 154, 0, 136, 121, 159, 60, 0, 42, 0, 205, 139, 193, 115, 144, 87, 204, 103, 96, 121, 8, 131, 40, 131, 50, 88, 247, 116, 94, 138, 157, 143, 72, 97, 6, 117, 158, 143, 193, 120, 204, 150, 90, 158, 205, 124, 164, 143, 250, 152, 8, 153, 40, 91, 232, 111, 139, 133, 28, 157, 32, 121, 173, 143, 191, 157, 192, 157, 60, 0, 42, 0, 100, 107, 76, 79, 218, 108, 188, 115, 137, 118, 134, 155, 60, 0, 42, 0, 63, 103, 33, 107, 58, 79, 125, 79, 58, 82, 62, 82, 155, 94, 38, 131, 40, 104, 191, 131, 88, 125, 211, 134, 80, 141, 134, 135, 220, 140, 60, 0, 42, 0, 6, 83, 234, 86, 241, 86, 193, 130, 233, 95, 158, 103, 16, 131, 49, 96, 164, 96, 199, 104, 39, 113, 113, 132, 23, 111, 97, 128, 239, 132, 37, 133, 162, 154, 176, 102, 5, 106, 44, 106, 156, 113, 125, 116, 129, 116, 235, 125, 102, 128, 106, 128, 155, 119, 245, 123, 112, 128, 204, 135, 111, 147, 113, 126, 230, 147, 24, 154, 68, 154, 60, 0, 42, 0, 206, 78, 27, 78, 147, 95, 67, 90, 110, 91, 150, 95, 158, 95, 176, 96, 217, 109, 46, 116, 82, 97, 14, 111, 64, 111, 104, 111, 180, 138, 232, 140, 233, 140, 55, 106, 194, 133, 226, 83, 71, 112, 9, 107, 28, 114, 60, 0, 42, 0, 129, 97, 37, 139, 60, 0, 42, 0, 209, 81, 74, 110, 96, 129, 143, 143, 51, 143, 60, 0, 42, 0, 151, 124, 213, 137, 129, 158, 132, 158, 164, 158, 60, 0, 42, 0, 130, 95, 130, 107, 60, 0, 42, 0, 195, 79, 29, 115, 40, 129, 168, 90, 98, 145, 4, 118, 31, 133, 142, 138, 151, 141, 72, 86, 177, 97, 39, 142, 139, 145, 47, 118, 7, 124, 44, 126, 89, 142, 0, 159, 116, 142, 117, 142, 99, 152, 60, 0, 42, 0, 70, 108, 186, 100, 209, 146, 105, 149, 127, 142, 27, 101, 165, 142, 121, 148, 60, 0, 42, 0, 213, 106, 209, 93, 17, 107, 115, 122, 60, 0, 42, 0, 156, 122, 169, 107, 182, 113, 225, 123, 18, 124, 196, 122, 40, 114, 60, 0, 42, 0, 20, 93, 172, 80, 215, 81, 30, 127, 148, 88, 137, 93, 91, 97, 103, 100, 177, 105, 85, 115, 239, 105, 234, 120, 23, 126, 217, 147, 60, 0, 42, 0, 60, 111, 128, 116, 161, 141, 160, 118, 60, 0, 42, 0, 28, 79, 240, 95, 169, 117, 5, 80, 139, 124, 35, 125, 198, 127, 3, 129, 6, 129, 80, 85, 91, 85, 180, 96, 236, 109, 3, 132, 243, 107, 32, 113, 58, 129, 1, 118, 185, 124, 183, 125, 224, 127, 181, 129, 172, 129, 226, 111, 193, 122, 74, 137, 71, 152, 206, 129, 60, 0, 42, 0, 124, 78, 60, 0, 42, 0, 168, 144, 81, 103, 180, 118, 6, 142, 138, 111, 244, 122, 165, 129, 60, 0, 42, 0, 88, 91, 159, 79, 245, 98, 60, 0, 42, 0, 12, 82, 214, 95, 60, 0, 42, 0, 248, 91, 11, 84, 127, 124, 60, 0, 42, 0, 19, 100, 115, 116, 115, 144, 203, 120, 174, 100, 73, 142, 157, 145, 60, 0, 42, 0, 88, 134, 111, 93, 115, 93, 228, 117, 73, 119, 236, 119, 171, 132, 22, 133, 126, 158, 66, 145, 122, 158, 166, 142, 60, 0, 42, 0, 30, 129, 60, 0, 42, 0, 73, 82, 82, 82, 157, 83, 14, 89, 43, 99, 157, 131, 161, 131, 170, 99, 42, 144, 174, 101, 228, 104, 9, 149, 204, 132, 25, 149, 188, 146, 47, 147, 60, 0, 39, 0, 208, 253, 68, 0, 39, 0, 60, 0, 42, 0, 145, 84, 55, 128, 69, 131, 26, 123, 210, 85, 45, 100, 33, 137, 96, 86, 152, 100, 157, 147, 60, 0, 42, 0, 190, 143, 207, 143, 214, 143, 218, 143, 126, 84, 178, 89, 27, 96, 147, 108, 210, 84, 159, 112, 141, 127, 89, 131, 87, 117, 115, 82, 18, 83, 99, 117, 42, 123, 57, 144, 84, 123, 90, 138, 84, 144, 246, 141, 24, 150, 124, 151, 152, 133, 145, 151, 245, 113, 253, 135, 137, 147, 130, 142, 61, 148, 195, 151, 150, 159, 152, 159, 60, 0, 42, 0, 83, 98, 60, 0, 42, 0, 39, 89, 163, 78, 79, 108, 20, 119, 60, 0, 42, 0, 175, 87, 41, 118, 182, 88, 104, 126, 60, 0, 42, 0, 70, 84, 84, 84, 67, 115, 219, 97, 60, 0, 42, 0, 121, 107, 46, 144, 163, 80, 60, 0, 42, 0, 227, 78, 106, 143, 162, 79, 136, 87, 177, 92, 18, 94, 25, 117, 208, 126, 232, 143, 128, 154, 38, 94, 133, 95, 32, 96, 203, 103, 134, 107, 179, 115, 55, 141, 47, 94, 209, 142, 237, 87, 54, 94, 63, 125, 174, 134, 139, 136, 218, 142, 184, 140, 233, 142, 71, 116, 215, 94, 199, 83, 195, 102, 255, 125, 15, 157, 52, 98, 92, 130, 219, 158, 36, 124, 91, 142, 59, 112, 52, 151, 118, 137, 241, 158, 70, 151, 60, 0, 42, 0, 152, 155, 60, 0, 42, 0, 57, 78, 137, 89, 85, 83, 197, 98, 88, 83, 8, 119, 3, 120, 60, 128, 61, 128, 248, 144, 67, 128, 173, 142, 174, 85, 133, 90, 154, 107, 5, 118, 48, 83, 170, 123, 29, 137, 50, 145, 21, 152, 11, 81, 239, 82, 212, 100, 171, 107, 20, 117, 73, 118, 76, 137, 30, 124, 120, 128, 60, 0, 42, 0, 20, 79, 16, 82, 140, 98, 172, 115, 237, 116, 198, 128, 116, 136, 184, 117, 30, 125, 184, 99, 85, 141, 182, 78, 162, 100, 163, 100, 184, 111, 213, 158, 189, 129, 238, 158, 60, 0, 42, 0, 230, 101, 70, 79, 14, 94, 138, 108, 249, 108, 218, 114, 222, 139, 230, 103, 141, 117, 86, 85, 87, 85, 57, 95, 238, 96, 225, 109, 15, 132, 203, 134, 127, 85, 62, 95, 46, 108, 69, 129, 17, 135, 219, 137, 158, 122, 149, 138, 228, 80, 73, 86, 190, 153, 231, 154, 62, 86, 72, 95, 154, 97, 186, 97, 185, 111, 171, 121, 222, 132, 243, 153, 32, 157, 90, 118, 170, 86, 117, 126, 9, 141, 46, 151, 79, 153, 60, 0, 42, 0, 83, 95, 240, 115, 198, 136, 219, 148, 92, 123, 118, 117, 121, 86, 162, 111, 171, 116, 96, 137, 57, 124, 97, 130, 247, 135, 58, 148, 60, 0, 42, 0, 33, 99, 90, 81, 32, 140, 203, 100, 97, 139, 232, 158, 41, 101, 89, 112, 19, 107, 156, 139, 60, 0, 42, 0, 57, 108, 252, 81, 53, 87, 149, 91, 0, 120, 177, 87, 97, 131, 99, 104, 234, 131, 120, 90, 19, 97, 253, 116, 63, 144, 99, 93, 252, 150, 82, 111, 173, 120, 5, 81, 138, 119, 105, 133, 164, 141, 203, 88, 148, 106, 151, 116, 234, 118, 17, 121, 28, 124, 47, 134, 227, 149, 60, 0, 42, 0, 0, 82, 2, 82, 232, 83, 201, 95, 55, 103, 24, 108, 32, 130, 214, 145, 125, 156, 91, 155, 60, 0, 42, 0, 111, 99, 60, 0, 42, 0, 252, 91, 155, 92, 102, 150, 246, 92, 99, 99, 119, 121, 130, 121, 23, 100, 157, 150, 139, 93, 140, 93, 221, 105, 14, 92, 175, 150, 212, 88, 185, 93, 227, 100, 72, 142, 177, 121, 60, 0, 42, 0, 48, 82, 18, 80, 188, 96, 24, 113, 215, 118, 255, 131, 33, 105, 220, 118, 83, 144, 50, 122, 140, 123, 226, 127, 117, 86, 59, 122, 84, 130, 92, 136, 164, 106, 95, 136, 254, 113, 255, 127, 199, 142, 217, 116, 155, 126, 60, 0, 42, 0, 26, 86, 60, 0, 42, 0, 116, 96, 151, 95, 194, 109, 179, 96, 234, 96, 207, 104, 29, 149, 179, 95, 183, 95, 64, 147, 60, 0, 42, 0, 48, 87, 132, 118, 38, 129, 60, 0, 42, 0, 101, 98, 125, 98, 60, 0, 42, 0, 111, 112, 123, 118, 75, 140, 84, 86, 1, 91, 200, 113, 146, 116, 243, 122, 38, 124, 96, 130, 180, 137, 108, 142, 60, 0, 42, 0, 41, 103, 73, 123, 37, 98, 60, 0, 42, 0, 147, 144, 243, 81, 39, 145, 165, 150, 177, 88, 157, 93, 170, 119, 244, 120, 107, 149, 200, 106, 25, 148, 60, 0, 42, 0, 16, 108, 254, 78, 78, 79, 67, 89, 125, 95, 155, 136, 114, 85, 222, 87, 157, 127, 132, 150, 36, 88, 134, 141, 244, 110, 93, 149, 254, 120, 73, 147, 174, 151, 60, 0, 42, 0, 248, 94, 196, 114, 116, 124, 214, 130, 234, 143, 25, 85, 76, 101, 164, 109, 123, 131, 145, 104, 27, 123, 204, 137, 110, 151, 204, 110, 176, 153, 226, 154, 0, 86, 225, 90, 223, 127, 11, 133, 16, 133, 20, 152, 117, 101, 0, 106, 244, 123, 129, 86, 225, 133, 116, 140, 98, 142, 4, 155, 209, 147, 244, 124, 191, 137, 16, 158, 60, 0, 42, 0, 142, 83, 88, 87, 203, 139, 184, 144, 58, 150, 103, 84, 149, 94, 36, 95, 181, 98, 222, 98, 226, 103, 116, 114, 37, 120, 139, 99, 231, 131, 221, 137, 70, 138, 231, 142, 92, 128, 182, 154, 243, 155, 60, 0, 42, 0, 84, 87, 31, 95, 243, 101, 85, 103, 147, 115, 31, 96, 164, 103, 228, 79, 29, 94, 202, 87, 35, 90, 18, 144, 19, 144, 89, 80, 71, 85, 138, 104, 13, 113, 246, 115, 49, 119, 118, 121, 44, 123, 194, 131, 27, 140, 241, 145, 130, 90, 227, 104, 39, 110, 71, 119, 20, 127, 130, 132, 192, 80, 152, 121, 99, 129, 94, 144, 106, 146, 145, 88, 172, 88, 85, 100, 178, 120, 21, 133, 67, 135, 112, 144, 120, 97, 11, 117, 224, 125, 179, 93, 230, 138, 54, 142, 174, 135, 60, 0, 42, 0, 56, 117, 65, 101, 130, 99, 142, 80, 167, 83, 110, 93, 199, 110, 199, 105, 217, 105, 40, 118, 160, 152, 78, 142, 197, 93, 90, 152, 91, 152, 107, 118, 211, 93, 212, 93, 39, 101, 114, 118, 123, 159, 60, 0, 42, 0, 120, 81, 76, 89, 185, 112, 112, 90, 32, 115, 95, 101, 35, 105, 213, 141, 152, 120, 167, 132, 71, 133, 46, 142, 222, 158, 184, 86, 60, 0, 42, 0, 53, 117, 67, 79, 61, 150, 107, 87, 151, 94, 171, 87, 66, 98, 183, 115, 191, 148, 93, 90, 230, 96, 192, 109, 96, 89, 20, 116, 191, 107, 20, 135, 251, 150, 138, 88, 194, 88, 66, 106, 93, 106, 177, 111, 91, 151, 92, 118, 31, 124, 84, 154, 60, 0, 42, 0, 1, 82, 252, 83, 72, 108, 31, 82, 109, 134, 203, 81, 93, 89, 52, 95, 107, 95, 193, 134, 49, 116, 130, 140, 137, 120, 237, 156, 166, 107, 151, 119, 213, 150, 137, 155, 183, 156, 19, 124, 38, 159, 219, 155, 112, 157, 60, 0, 42, 0, 90, 98, 76, 92, 60, 0, 42, 0, 20, 95, 4, 79, 10, 84, 147, 148, 142, 122, 11, 138, 3, 140, 137, 99, 227, 145, 222, 148, 235, 148, 31, 146, 232, 122, 231, 132, 177, 146, 255, 150, 97, 155, 191, 138, 57, 118, 181, 122, 253, 146, 203, 133, 67, 148, 60, 0, 42, 0, 57, 114, 204, 141, 242, 85, 58, 137, 60, 0, 42, 0, 245, 130, 237, 143, 164, 87, 204, 92, 78, 96, 21, 99, 51, 102, 214, 126, 197, 128, 222, 116, 35, 119, 10, 128, 28, 98, 13, 140, 139, 85, 30, 88, 73, 94, 245, 96, 242, 99, 115, 117, 112, 125, 11, 128, 247, 129, 68, 138, 131, 141, 123, 149, 224, 83, 156, 107, 67, 114, 82, 114, 125, 93, 159, 120, 40, 135, 11, 137, 83, 130, 118, 135, 130, 117, 220, 138, 64, 142, 189, 156, 225, 102, 137, 117, 8, 156, 138, 117, 14, 108, 60, 0, 42, 0, 203, 84, 48, 119, 60, 0, 42, 0, 1, 78, 195, 78, 238, 83, 4, 94, 142, 115, 148, 117, 239, 118, 137, 148, 53, 128, 112, 134, 74, 145, 216, 145, 106, 151, 60, 0, 42, 0, 117, 89, 118, 152, 2, 152, 14, 159, 127, 93, 17, 159, 206, 111, 161, 133, 36, 148, 60, 0, 42, 0, 162, 139, 202, 95, 100, 153, 244, 119, 154, 91, 2, 138, 227, 152, 118, 85, 35, 132, 228, 148, 23, 105, 90, 129, 135, 120, 45, 149, 160, 120, 98, 128, 74, 135, 204, 146, 32, 147, 248, 120, 65, 152, 60, 0, 42, 0, 31, 78, 34, 78, 229, 148, 169, 146, 60, 0, 42, 0, 28, 78, 172, 81, 154, 84, 189, 92, 113, 103, 243, 130, 56, 102, 33, 108, 50, 80, 43, 158, 236, 87, 59, 90, 32, 93, 44, 93, 183, 109, 23, 123, 196, 131, 154, 95, 45, 108, 64, 135, 151, 155, 21, 159, 223, 155, 135, 157, 171, 157, 60, 0, 42, 0, 99, 132, 165, 88, 30, 91, 194, 97, 189, 123, 107, 133, 204, 138, 60, 0, 42, 0, 168, 82, 187, 81, 151, 79, 140, 87, 219, 89, 210, 92, 107, 96, 15, 99, 11, 104, 30, 109, 232, 128, 245, 143, 205, 81, 25, 98, 244, 128, 213, 82, 80, 120, 223, 104, 105, 110, 103, 125, 86, 129, 205, 80, 231, 153, 24, 151, 60, 0, 42, 0, 58, 84, 23, 85, 253, 144, 92, 81, 96, 81, 56, 133, 119, 106, 252, 123, 60, 0, 42, 0, 103, 78, 39, 150, 150, 98, 147, 103, 97, 150, 30, 85, 170, 134, 4, 146, 60, 0, 42, 0, 151, 101, 70, 140, 214, 144, 98, 109, 115, 131, 23, 144, 126, 153, 37, 155, 170, 104, 237, 107, 48, 129, 88, 145, 216, 117, 151, 149, 166, 122, 38, 155, 22, 153, 163, 101, 216, 149, 199, 122, 42, 155, 45, 155, 44, 155, 60, 0, 42, 0, 190, 83, 98, 82, 7, 150, 31, 86, 99, 119, 143, 145, 205, 149, 60, 0, 42, 0, 210, 107, 236, 114, 156, 109, 251, 139, 14, 110, 31, 105, 77, 114, 138, 114, 161, 120, 251, 136, 173, 138, 115, 135, 104, 115, 22, 147, 223, 81, 53, 83, 59, 91, 6, 112, 221, 106, 176, 107, 88, 114, 162, 114, 196, 116, 190, 118, 51, 154, 233, 158, 128, 139, 68, 140, 21, 141, 227, 151, 209, 154, 95, 148, 199, 151, 229, 151, 247, 158, 159, 139, 60, 0, 42, 0, 3, 123, 53, 88, 62, 94, 61, 116, 76, 141, 121, 119, 169, 137, 237, 140, 228, 123, 60, 0, 42, 0, 143, 130, 146, 89, 92, 103, 154, 128, 172, 89, 166, 94, 112, 131, 250, 121, 33, 110, 111, 151, 64, 149, 153, 135, 172, 107, 77, 147, 39, 136, 57, 136, 60, 0, 42, 0, 17, 128, 115, 80, 108, 82, 143, 90, 239, 122, 13, 137, 116, 147, 60, 0, 42, 0, 237, 119, 60, 0, 42, 0, 181, 107, 173, 101, 69, 88, 14, 127, 110, 132, 52, 105, 69, 113, 86, 116, 118, 129, 171, 120, 59, 149, 222, 125, 200, 107, 22, 124, 91, 147, 183, 101, 150, 142, 106, 124, 60, 0, 42, 0, 150, 87, 6, 88, 96, 88, 95, 93, 253, 117, 211, 120, 45, 157, 28, 148, 60, 0, 42, 0, 39, 152, 60, 0, 42, 0, 31, 150, 249, 91, 74, 81, 76, 81, 81, 81, 254, 91, 75, 121, 60, 96, 110, 150, 138, 150, 147, 120, 144, 125, 13, 92, 158, 97, 157, 97, 231, 111, 177, 133, 102, 149, 223, 97, 41, 112, 72, 139, 104, 137, 19, 148, 117, 139, 60, 0, 42, 0, 40, 84, 199, 96, 102, 101, 51, 135, 169, 88, 170, 88, 180, 100, 100, 115, 120, 86, 137, 100, 84, 106, 156, 114, 5, 121, 114, 142, 126, 142, 80, 154, 60, 0, 42, 0, 249, 118, 184, 141, 137, 142, 60, 0, 42, 0, 5, 79, 228, 86, 137, 94, 140, 108, 150, 112, 254, 118, 24, 120, 7, 144, 157, 148, 127, 152, 65, 144, 13, 146, 111, 105, 19, 152, 183, 120, 111, 144, 97, 111, 201, 113, 50, 142, 60, 0, 42, 0, 26, 89, 27, 89, 132, 84, 198, 84, 83, 117, 95, 82, 28, 93, 135, 99, 96, 101, 106, 101, 242, 107, 240, 136, 137, 86, 60, 0, 42, 0, 58, 89, 206, 148, 107, 82, 83, 101, 90, 101, 165, 85, 229, 117, 44, 146, 106, 89, 217, 81, 49, 142, 181, 155, 56, 148, 60, 0, 42, 0, 53, 103, 54, 103, 218, 84, 155, 87, 156, 87, 5, 99, 6, 99, 245, 87, 13, 127, 47, 105, 147, 141, 177, 142, 178, 142, 158, 125, 184, 78, 122, 147, 195, 142, 178, 86, 114, 89, 60, 0, 42, 0, 52, 82, 65, 82, 74, 150, 79, 150, 115, 153, 46, 92, 193, 103, 238, 103, 87, 104, 21, 88, 53, 130, 240, 96, 226, 141, 229, 141, 250, 141, 255, 152, 174, 88, 158, 93, 156, 97, 175, 88, 125, 157, 60, 0, 39, 0, 208, 253, 69, 0, 39, 0, 60, 0, 42, 0, 184, 89, 191, 89, 63, 90, 64, 90, 89, 92, 254, 117, 60, 0, 42, 0, 185, 139, 42, 84, 238, 86, 215, 143, 196, 79, 37, 90, 232, 92, 233, 92, 144, 109, 170, 131, 244, 115, 27, 138, 146, 118, 75, 119, 11, 146, 7, 149, 69, 158, 254, 134, 192, 120, 144, 138, 31, 152, 157, 152, 100, 155, 77, 152, 93, 157, 94, 157, 76, 139, 42, 156, 60, 0, 42, 0, 153, 103, 8, 120, 11, 152, 65, 86, 0, 154, 72, 157, 60, 0, 42, 0, 132, 83, 117, 92, 57, 98, 122, 107, 139, 92, 40, 150, 67, 84, 124, 98, 202, 130, 56, 150, 93, 84, 16, 120, 109, 143, 162, 84, 185, 84, 169, 87, 246, 89, 201, 92, 14, 83, 118, 96, 40, 120, 133, 134, 127, 153, 84, 80, 126, 83, 10, 88, 170, 96, 70, 120, 20, 140, 219, 142, 2, 145, 15, 150, 46, 88, 63, 93, 225, 96, 21, 97, 66, 110, 60, 132, 95, 140, 246, 142, 76, 144, 79, 144, 42, 146, 197, 94, 36, 100, 57, 100, 39, 116, 109, 129, 123, 138, 235, 80, 65, 135, 55, 149, 101, 155, 87, 158, 90, 133, 123, 144, 30, 152, 154, 152, 19, 153, 105, 86, 220, 100, 168, 137, 228, 138, 188, 149, 41, 153, 84, 147, 196, 156, 94, 107, 78, 152, 24, 121, 238, 106, 16, 156, 154, 157, 141, 139, 67, 159, 105, 148, 118, 159, 119, 156, 60, 0, 42, 0, 246, 139, 146, 138, 60, 0, 42, 0, 64, 89, 105, 96, 189, 132, 126, 113, 60, 0, 42, 0, 206, 92, 60, 0, 42, 0, 65, 100, 60, 0, 42, 0, 165, 151, 60, 0, 42, 0, 63, 81, 12, 128, 80, 81, 149, 79, 82, 81, 81, 150, 207, 92, 15, 109, 75, 131, 45, 104, 249, 128, 50, 85, 187, 136, 56, 158, 171, 124, 79, 128, 0, 143, 149, 156, 173, 150, 245, 154, 158, 155, 47, 157, 92, 143, 60, 0, 42, 0, 188, 83, 18, 92, 19, 92, 20, 92, 51, 128, 233, 143, 49, 109, 117, 153, 46, 104, 230, 107, 229, 115, 210, 148, 62, 114, 12, 153, 236, 153, 190, 133, 135, 144, 176, 141, 60, 0, 42, 0, 140, 78, 13, 95, 16, 95, 116, 79, 53, 82, 161, 84, 48, 141, 174, 140, 72, 136, 179, 140, 128, 138, 122, 146, 50, 106, 60, 0, 39, 0, 208, 253, 70, 0, 39, 0, 60, 0, 42, 0, 209, 83, 183, 108, 122, 118, 124, 118, 66, 95, 177, 145, 60, 0, 42, 0, 79, 78, 16, 79, 194, 89, 161, 87, 76, 109, 186, 117, 90, 127, 55, 131, 0, 150, 48, 104, 29, 120, 41, 123, 160, 80, 79, 123, 130, 119, 112, 127, 165, 149, 120, 127, 67, 106, 197, 133, 60, 0, 42, 0, 113, 79, 213, 108, 69, 147, 75, 112, 60, 0, 42, 0, 208, 115, 58, 116, 234, 154, 95, 133, 238, 154, 60, 0, 42, 0, 6, 94, 219, 95, 106, 117, 235, 82, 67, 86, 15, 91, 97, 94, 163, 97, 67, 133, 217, 101, 219, 101, 89, 126, 251, 127, 233, 133, 83, 143, 191, 152, 83, 124, 220, 152, 85, 156, 60, 0, 42, 0, 225, 81, 226, 81, 227, 81, 37, 83, 75, 103, 201, 103, 254, 119, 117, 124, 146, 148, 36, 130, 230, 112, 39, 130, 50, 123, 229, 104, 34, 110, 105, 113, 208, 125, 166, 88, 10, 106, 78, 106, 212, 113, 160, 116, 176, 129, 160, 133, 65, 126, 78, 137, 179, 127, 111, 142, 42, 112, 63, 112, 44, 121, 41, 134, 7, 148, 34, 148, 28, 136, 237, 157, 60, 0, 42, 0, 205, 83, 85, 98, 238, 78, 212, 143, 233, 145, 60, 0, 42, 0, 62, 108, 175, 114, 127, 89, 78, 108, 219, 108, 109, 153, 3, 131, 41, 141, 72, 117, 9, 138, 211, 142, 79, 90, 181, 104, 213, 118, 53, 123, 169, 140, 236, 142, 239, 152, 240, 152, 252, 110, 14, 91, 196, 123, 60, 0, 42, 0, 26, 83, 185, 101, 161, 144, 74, 87, 120, 108, 179, 130, 139, 103, 101, 114, 171, 148, 211, 109, 132, 134, 1, 146, 58, 147, 11, 157, 60, 0, 42, 0, 50, 150, 168, 89, 63, 98, 170, 128, 197, 87, 130, 156, 116, 155, 31, 156, 60, 0, 42, 0, 255, 78, 191, 139, 119, 95, 186, 126, 9, 102, 24, 102, 236, 116, 6, 119, 35, 80, 202, 101, 33, 125, 43, 130, 42, 138, 227, 154, 173, 157, 60, 0, 42, 0, 62, 101, 189, 141, 60, 0, 42, 0, 15, 88, 60, 0, 42, 0, 222, 152, 131, 89, 94, 151, 219, 152, 97, 85, 83, 90, 84, 90, 4, 110, 239, 126, 242, 131, 73, 98, 131, 102, 6, 115, 95, 151, 246, 136, 203, 125, 26, 135, 15, 151, 177, 156, 37, 153, 161, 153, 17, 154, 27, 154, 221, 152, 60, 0, 42, 0, 165, 128, 221, 109, 83, 129, 48, 135, 230, 135, 60, 0, 42, 0, 15, 103, 208, 128, 42, 83, 253, 139, 92, 89, 177, 96, 144, 101, 208, 104, 167, 105, 225, 127, 92, 133, 185, 138, 218, 123, 60, 0, 42, 0, 32, 84, 190, 130, 159, 94, 110, 103, 184, 108, 210, 114, 186, 128, 50, 102, 199, 128, 57, 141, 247, 79, 85, 82, 158, 83, 191, 117, 107, 150, 93, 92, 9, 132, 195, 94, 187, 140, 241, 117, 68, 149, 226, 94, 202, 102, 72, 118, 35, 159, 247, 111, 224, 106, 225, 155, 40, 148, 69, 151, 60, 0, 42, 0, 6, 82, 41, 84, 9, 94, 183, 126, 172, 130, 16, 102, 27, 108, 219, 84, 213, 122, 111, 136, 122, 81, 27, 125, 194, 127, 93, 81, 251, 104, 28, 138, 174, 142, 90, 145, 22, 146, 240, 150, 6, 103, 52, 153, 89, 153, 60, 0, 42, 0, 95, 87, 162, 89, 142, 92, 126, 108, 12, 103, 140, 103, 131, 112, 166, 128, 146, 127, 160, 134, 161, 134, 164, 104, 252, 104, 26, 113, 182, 132, 154, 153, 171, 150, 179, 88, 105, 94, 198, 111, 97, 133, 117, 155, 104, 106, 204, 113, 211, 113, 110, 140, 34, 159, 181, 127, 22, 159, 118, 140, 82, 143, 60, 148, 169, 153, 194, 158, 60, 0, 42, 0, 137, 124, 250, 158, 60, 0, 42, 0, 253, 78, 5, 95, 75, 89, 255, 95, 206, 121, 126, 80, 36, 97, 170, 124, 232, 80, 164, 97, 110, 89, 185, 129, 222, 124, 188, 156, 53, 112, 93, 156, 60, 0, 42, 0, 48, 78, 206, 152, 249, 78, 232, 81, 236, 81, 166, 89, 163, 108, 168, 108, 238, 81, 171, 103, 1, 92, 175, 117, 253, 118, 28, 120, 168, 152, 239, 92, 240, 92, 81, 80, 123, 104, 253, 112, 18, 116, 54, 93, 38, 115, 81, 132, 11, 149, 83, 105, 142, 114, 2, 135, 11, 118, 184, 120, 252, 80, 200, 123, 55, 145, 210, 146, 146, 106, 207, 149, 80, 140, 189, 147, 224, 147, 70, 145, 247, 91, 67, 112, 52, 134, 59, 151, 45, 136, 74, 151, 204, 152, 183, 158, 60, 0, 42, 0, 175, 81, 6, 89, 64, 99, 114, 109, 34, 144, 56, 88, 132, 110, 174, 153, 83, 100, 40, 111, 152, 125, 66, 130, 60, 0, 42, 0, 189, 139, 130, 137, 42, 85, 247, 138, 60, 0, 42, 0, 228, 81, 73, 89, 46, 117, 248, 79, 87, 110, 40, 113, 72, 113, 29, 127, 87, 141, 239, 156, 243, 156, 12, 157, 43, 126, 245, 140, 60, 0, 42, 0, 133, 137, 60, 0, 42, 0, 207, 78, 114, 87, 187, 104, 60, 0, 42, 0, 17, 125, 230, 136, 60, 0, 42, 0, 54, 127, 38, 84, 154, 89, 57, 127, 59, 127, 149, 107, 236, 150, 0, 157, 60, 0, 42, 0, 43, 89, 21, 79, 158, 144, 75, 84, 139, 89, 199, 89, 158, 115, 164, 128, 36, 96, 206, 103, 6, 120, 66, 131, 109, 136, 186, 87, 16, 90, 3, 92, 116, 131, 201, 101, 40, 125, 186, 141, 184, 158, 225, 117, 3, 122, 215, 141, 7, 146, 95, 123, 146, 125, 28, 145, 117, 91, 103, 140, 119, 101, 154, 129, 250, 156, 169, 158, 208, 124, 172, 158, 177, 158, 239, 97, 60, 0, 42, 0, 64, 78, 255, 93, 23, 95, 15, 79, 235, 81, 54, 117, 91, 79, 185, 81, 28, 82, 90, 91, 118, 98, 153, 130, 163, 130, 136, 84, 170, 92, 127, 95, 43, 96, 194, 98, 13, 103, 142, 103, 237, 108, 194, 126, 203, 126, 251, 130, 0, 131, 216, 79, 152, 87, 235, 103, 31, 108, 17, 109, 165, 112, 184, 115, 73, 117, 80, 117, 83, 121, 88, 127, 47, 131, 219, 144, 232, 151, 249, 84, 63, 104, 110, 109, 41, 120, 169, 131, 168, 134, 16, 83, 116, 104, 170, 109, 240, 112, 8, 116, 38, 123, 48, 123, 49, 125, 60, 125, 199, 127, 116, 130, 212, 131, 89, 134, 177, 136, 69, 94, 244, 104, 101, 125, 102, 127, 77, 132, 143, 121, 176, 124, 141, 125, 64, 130, 9, 135, 144, 143, 88, 146, 92, 146, 171, 152, 231, 156, 145, 105, 42, 122, 153, 123, 205, 151, 94, 94, 147, 111, 96, 135, 244, 154, 20, 157, 232, 138, 62, 142, 59, 143, 132, 155, 65, 118, 70, 137, 178, 155, 251, 158, 101, 137, 105, 157, 157, 157, 60, 0, 42, 0, 82, 84, 154, 98, 43, 117, 118, 78, 156, 94, 35, 95, 202, 98, 167, 101, 204, 79, 219, 79, 213, 128, 217, 144, 236, 156, 239, 79, 220, 145, 225, 145, 108, 99, 133, 143, 40, 105, 36, 113, 217, 118, 81, 129, 207, 110, 5, 135, 80, 129, 20, 143, 56, 86, 168, 100, 171, 100, 43, 152, 52, 155, 32, 124, 252, 158, 60, 0, 42, 0, 29, 150, 54, 114, 163, 139, 216, 78, 135, 89, 31, 141, 68, 150, 144, 84, 127, 87, 206, 122, 28, 150, 120, 154, 13, 89, 202, 92, 84, 121, 3, 138, 160, 140, 116, 141, 165, 134, 157, 136, 90, 150, 105, 80, 168, 81, 111, 82, 102, 90, 185, 134, 133, 80, 141, 90, 204, 91, 169, 95, 255, 121, 47, 132, 215, 134, 132, 137, 66, 138, 75, 141, 49, 105, 26, 127, 121, 129, 139, 156, 163, 121, 7, 137, 20, 137, 89, 141, 238, 125, 103, 133, 92, 135, 110, 135, 230, 140, 217, 153, 20, 91, 27, 126, 57, 143, 146, 155, 251, 140, 81, 147, 98, 147, 198, 156, 134, 137, 165, 153, 18, 156, 60, 0, 42, 0, 92, 145, 60, 0, 39, 0, 208, 253, 71, 0, 39, 0, 60, 0, 42, 0, 238, 101, 119, 84, 14, 86, 32, 86, 60, 0, 42, 0, 134, 148, 28, 92, 118, 86, 55, 147, 60, 0, 42, 0, 21, 92, 141, 115, 60, 0, 42, 0, 44, 92, 64, 155, 60, 0, 42, 0, 133, 79, 229, 139, 194, 144, 84, 150, 147, 87, 223, 89, 208, 92, 68, 131, 80, 102, 69, 141, 97, 117, 116, 121, 111, 125, 114, 138, 101, 140, 197, 140, 204, 140, 60, 0, 42, 0, 203, 95, 57, 101, 96, 125, 60, 0, 42, 0, 16, 78, 98, 78, 3, 83, 4, 83, 35, 150, 90, 103, 153, 148, 214, 118, 97, 100, 137, 110, 98, 132, 35, 146, 145, 150, 36, 98, 130, 105, 233, 105, 203, 132, 17, 111, 234, 105, 194, 116, 60, 0, 42, 0, 24, 117, 211, 95, 137, 130, 192, 143, 60, 101, 70, 103, 149, 115, 157, 128, 105, 87, 212, 108, 248, 119, 247, 130, 121, 78, 209, 103, 255, 122, 179, 117, 80, 145, 126, 78, 147, 124, 129, 78, 242, 81, 50, 92, 52, 92, 120, 123, 39, 111, 241, 156, 54, 92, 55, 92, 80, 155, 60, 0, 42, 0, 224, 78, 94, 98, 175, 118, 198, 121, 102, 136, 118, 141, 98, 101, 127, 104, 52, 123, 8, 122, 31, 97, 137, 111, 149, 141, 68, 106, 192, 100, 51, 124, 20, 156, 225, 156, 100, 156, 60, 0, 42, 0, 114, 94, 240, 101, 117, 108, 240, 118, 192, 126, 29, 80, 206, 81, 230, 109, 58, 125, 76, 138, 173, 154, 121, 94, 166, 105, 138, 106, 17, 141, 99, 141, 27, 141, 104, 112, 60, 0, 42, 0, 136, 81, 83, 127, 174, 81, 26, 82, 96, 103, 178, 126, 155, 128, 161, 92, 104, 114, 152, 117, 252, 119, 56, 127, 162, 148, 91, 82, 97, 127, 8, 88, 134, 99, 237, 145, 225, 104, 133, 114, 61, 88, 177, 125, 65, 127, 252, 146, 160, 147, 60, 0, 42, 0, 151, 92, 23, 93, 47, 110, 60, 0, 42, 0, 53, 113, 57, 113, 123, 123, 211, 105, 5, 98, 6, 98, 60, 0, 42, 0, 139, 118, 148, 127, 153, 127, 216, 154, 144, 118, 217, 154, 239, 129, 220, 110, 212, 105, 126, 119, 143, 129, 249, 105, 112, 106, 217, 123, 213, 124, 59, 153, 220, 106, 223, 151, 206, 157, 27, 159, 241, 157, 60, 0, 42, 0, 48, 89, 114, 103, 210, 131, 1, 122, 30, 100, 31, 127, 160, 102, 192, 105, 193, 105, 62, 122, 63, 122, 80, 149, 30, 126, 193, 133, 186, 106, 243, 133, 60, 0, 42, 0, 63, 84, 74, 84, 194, 82, 240, 139, 220, 144, 252, 92, 110, 121, 112, 121, 6, 149, 118, 123, 158, 121, 165, 138, 239, 146, 60, 0, 42, 0, 8, 98, 225, 78, 42, 87, 181, 114, 165, 126, 19, 98, 144, 128, 107, 114, 153, 117, 175, 84, 113, 114, 229, 84, 243, 128, 188, 136, 61, 158, 114, 82, 1, 100, 65, 95, 210, 110, 40, 98, 76, 107, 16, 157, 26, 157, 241, 100, 12, 139, 63, 157, 182, 147, 60, 0, 42, 0, 68, 84, 110, 79, 12, 83, 12, 99, 22, 131, 1, 150, 105, 151, 75, 101, 60, 104, 50, 155, 5, 97, 245, 129, 91, 132, 210, 134, 211, 136, 148, 150, 221, 85, 101, 88, 198, 110, 225, 137, 63, 100, 197, 105, 136, 129, 163, 149, 164, 149, 102, 115, 73, 149, 136, 151, 208, 151, 188, 154, 253, 138, 53, 143, 175, 155, 202, 106, 218, 151, 85, 143, 183, 151, 20, 154, 60, 0, 42, 0, 255, 84, 56, 130, 240, 85, 60, 0, 42, 0, 42, 78, 4, 84, 124, 134, 11, 80, 76, 120, 236, 148, 135, 123, 60, 0, 42, 0, 217, 126, 102, 125, 60, 0, 42, 0, 57, 104, 223, 141, 60, 0, 42, 0, 207, 84, 60, 0, 42, 0, 110, 130, 60, 0, 42, 0, 152, 78, 153, 78, 27, 131, 239, 99, 4, 100, 60, 0, 42, 0, 47, 82, 154, 94, 74, 117, 109, 109, 21, 128, 238, 131, 41, 105, 63, 113, 90, 125, 83, 141, 82, 158, 234, 125, 6, 126, 174, 127, 225, 140, 185, 127, 138, 157, 60, 0, 42, 0, 224, 144, 253, 84, 194, 87, 250, 92, 45, 99, 224, 126, 63, 128, 132, 131, 151, 104, 134, 125, 160, 156, 190, 154, 193, 155, 60, 0, 42, 0, 244, 102, 41, 88, 133, 102, 60, 0, 42, 0, 229, 93, 19, 95, 108, 81, 183, 83, 159, 82, 59, 101, 91, 103, 155, 79, 252, 124, 177, 128, 171, 91, 174, 91, 109, 96, 163, 134, 172, 142, 154, 159, 17, 83, 104, 88, 74, 94, 41, 97, 229, 137, 179, 142, 149, 113, 20, 83, 189, 120, 248, 154, 245, 137, 143, 159, 148, 159, 60, 0, 42, 0, 254, 94, 233, 93, 94, 108, 241, 98, 242, 98, 49, 104, 217, 115, 1, 143, 143, 151, 60, 0, 42, 0, 113, 81, 33, 141, 190, 127, 29, 85, 162, 140, 187, 131, 60, 0, 42, 0, 80, 97, 60, 0, 42, 0, 254, 82, 93, 79, 159, 108, 169, 148, 167, 136, 17, 127, 14, 146, 157, 110, 100, 146, 241, 125, 32, 137, 221, 123, 60, 124, 178, 151, 221, 151, 60, 0, 42, 0, 182, 130, 163, 92, 215, 114, 223, 130, 184, 103, 189, 115, 7, 128, 9, 128, 49, 123, 8, 128, 188, 134, 127, 140, 60, 0, 42, 0, 120, 87, 132, 103, 223, 139, 45, 141, 162, 87, 228, 89, 41, 131, 147, 81, 31, 89, 32, 89, 61, 138, 190, 90, 64, 95, 6, 100, 108, 138, 88, 144, 202, 150, 203, 105, 121, 113, 207, 137, 128, 100, 175, 137, 252, 140, 60, 0, 42, 0, 48, 79, 113, 84, 149, 84, 209, 89, 100, 91, 189, 108, 210, 108, 253, 130, 231, 103, 113, 143, 2, 85, 91, 127, 42, 158, 31, 123, 199, 131, 240, 131, 196, 134, 218, 137, 241, 142, 242, 142, 156, 143, 100, 145, 50, 146, 141, 123, 155, 123, 244, 90, 208, 123, 109, 106, 149, 155, 35, 157, 60, 0, 42, 0, 187, 157, 60, 0, 42, 0, 3, 89, 228, 83, 98, 98, 105, 108, 194, 139, 55, 140, 161, 128, 111, 114, 168, 154, 3, 85, 95, 127, 150, 127, 39, 144, 180, 148, 166, 80, 82, 85, 200, 109, 53, 129, 202, 134, 204, 134, 51, 92, 50, 97, 199, 132, 65, 138, 137, 153, 68, 158, 190, 105, 194, 107, 55, 146, 19, 159, 20, 159, 15, 86, 150, 105, 183, 118, 88, 158, 64, 122, 14, 126, 211, 124, 163, 133, 242, 111, 188, 118, 204, 129, 66, 143, 54, 153, 20, 112, 236, 118, 189, 119, 49, 136, 60, 0, 42, 0, 250, 86, 69, 101, 197, 81, 126, 152, 12, 88, 19, 93, 46, 93, 143, 104, 127, 114, 221, 104, 123, 121, 199, 150, 252, 117, 18, 122, 34, 149, 241, 80, 46, 147, 180, 156, 221, 155, 103, 152, 60, 0, 42, 0, 220, 116, 46, 82, 205, 128, 29, 104, 57, 158, 68, 107, 113, 113, 82, 128, 170, 152, 143, 141, 128, 82, 250, 125, 59, 142, 189, 146, 179, 152, 48, 157, 39, 154, 60, 0, 42, 0, 142, 81, 231, 83, 80, 82, 110, 82, 225, 91, 60, 0, 42, 0, 102, 83, 108, 87, 214, 139, 2, 99, 105, 85, 155, 99, 99, 127, 83, 125, 107, 127, 2, 137, 127, 138, 60, 0, 42, 0, 86, 78, 180, 99, 81, 100, 60, 0, 42, 0, 208, 98, 180, 103, 250, 103, 137, 123, 60, 0, 42, 0, 44, 89, 207, 83, 42, 96, 96, 96, 60, 0, 42, 0, 115, 81, 194, 137, 152, 91, 160, 81, 140, 137, 12, 80, 250, 104, 132, 132, 164, 122, 162, 149, 29, 118, 79, 118, 179, 137, 215, 149, 207, 156, 220, 149, 37, 156, 192, 137, 94, 156, 60, 0, 42, 0, 158, 131, 134, 153, 47, 116, 239, 117, 102, 123, 161, 123, 40, 143, 24, 130, 39, 147, 40, 153, 228, 156, 60, 0, 42, 0, 204, 107, 49, 78, 47, 141, 244, 108, 186, 96, 239, 96, 188, 99, 171, 109, 171, 140, 185, 96, 124, 121, 99, 97, 92, 100, 69, 111, 102, 144, 12, 106, 229, 118, 70, 127, 218, 150, 128, 142, 198, 147, 76, 112, 31, 114, 216, 116, 212, 119, 54, 121, 115, 158, 80, 127, 117, 148, 121, 156, 27, 158, 60, 0, 42, 0, 73, 81, 110, 112, 138, 79, 151, 112, 154, 112, 155, 112, 163, 84, 153, 87, 239, 89, 56, 109, 42, 131, 68, 104, 225, 112, 214, 115, 241, 128, 217, 80, 4, 143, 167, 146, 198, 158, 60, 0, 42, 0, 127, 94, 131, 94, 183, 114, 227, 94, 119, 115, 233, 129, 60, 0, 42, 0, 199, 79, 27, 144, 230, 129, 151, 100, 60, 0, 42, 0, 31, 107, 60, 0, 42, 0, 82, 95, 45, 87, 171, 89, 159, 159, 196, 137, 189, 144, 136, 118, 37, 131, 250, 149, 48, 94, 234, 115, 255, 128, 128, 78, 128, 80, 69, 120, 144, 122, 191, 136, 143, 137, 175, 90, 198, 94, 29, 105, 112, 116, 204, 144, 226, 90, 107, 100, 168, 149, 145, 156, 0, 91, 251, 105, 252, 105, 157, 135, 157, 116, 161, 119, 173, 129, 173, 155, 156, 159, 194, 93, 120, 107, 54, 155, 41, 154, 204, 116, 57, 155, 247, 106, 60, 0, 42, 0, 132, 91, 63, 108, 57, 103, 104, 143, 139, 94, 121, 79, 38, 83, 225, 139, 82, 150, 157, 87, 253, 89, 81, 96, 49, 101, 120, 118, 204, 142, 60, 155, 170, 94, 106, 121, 45, 83, 119, 102, 64, 110, 235, 134, 228, 137, 109, 138, 172, 83, 11, 124, 225, 135, 60, 0, 42, 0, 48, 101, 61, 82, 63, 82, 11, 102, 220, 103, 148, 112, 53, 141, 66, 104, 103, 104, 34, 105, 36, 115, 64, 123, 180, 140, 213, 132, 234, 141, 49, 83, 138, 82, 140, 82, 161, 93, 140, 100, 246, 105, 156, 106, 182, 119, 172, 121, 2, 124, 195, 106, 80, 118, 88, 137, 220, 156, 188, 151, 86, 156, 101, 156, 60, 0, 42, 0, 40, 78, 110, 136, 195, 96, 242, 126, 158, 136, 138, 143, 218, 110, 216, 132, 254, 110, 196, 125, 9, 133, 217, 120, 37, 143, 167, 156, 140, 155, 192, 155, 60, 0, 42, 0, 205, 104, 84, 119, 116, 119, 173, 116, 52, 139, 60, 0, 42, 0, 89, 84, 188, 84, 218, 87, 237, 144, 111, 85, 29, 88, 30, 93, 27, 146, 5, 149, 142, 88, 17, 118, 19, 86, 73, 95, 196, 111, 72, 135, 75, 147, 77, 95, 200, 135, 60, 0, 42, 0, 239, 86, 246, 86, 251, 86, 253, 86, 0, 87, 11, 87, 60, 94, 88, 129, 87, 94, 86, 97, 13, 111, 93, 128, 46, 133, 149, 129, 98, 134, 152, 153, 60, 0, 42, 0, 156, 103, 200, 96, 201, 109, 19, 115, 211, 131, 131, 153, 1, 105, 1, 137, 232, 105, 191, 124, 182, 125, 62, 135, 249, 136, 32, 143, 1, 147, 28, 153, 57, 148, 60, 0, 42, 0, 199, 143, 78, 144, 60, 0, 39, 0, 208, 253, 72, 0, 39, 0, 60, 0, 42, 0, 200, 84, 234, 148, 60, 0, 42, 0, 228, 134, 60, 0, 42, 0, 100, 89, 60, 0, 42, 0, 141, 84, 179, 84, 232, 85, 60, 0, 42, 0, 216, 143, 105, 91, 38, 152, 184, 154, 132, 144, 60, 0, 42, 0, 119, 109, 242, 128, 248, 112, 112, 88, 124, 145, 162, 145, 60, 0, 42, 0, 165, 78, 142, 89, 135, 154, 179, 91, 38, 108, 208, 85, 0, 153, 237, 153, 244, 153, 90, 153, 60, 0, 42, 0, 161, 86, 60, 0, 42, 0, 68, 79, 120, 152, 251, 84, 182, 134, 99, 145, 7, 152, 232, 90, 61, 140, 168, 97, 160, 153, 91, 107, 62, 159, 60, 0, 42, 0, 151, 144, 43, 84, 175, 144, 253, 81, 129, 84, 163, 128, 254, 81, 119, 134, 5, 85, 5, 87, 34, 90, 91, 109, 33, 93, 87, 102, 146, 104, 181, 109, 19, 113, 0, 116, 210, 91, 69, 93, 233, 151, 29, 117, 104, 123, 44, 135, 143, 111, 225, 146, 125, 155, 211, 151, 60, 0, 42, 0, 6, 78, 136, 83, 85, 127, 107, 109, 138, 85, 10, 133, 26, 150, 67, 140, 43, 155, 60, 0, 42, 0, 73, 108, 125, 92, 87, 108, 236, 149, 241, 101, 190, 92, 190, 87, 141, 96, 77, 99, 134, 109, 2, 115, 159, 131, 88, 102, 10, 113, 225, 131, 236, 145, 136, 149, 148, 118, 69, 119, 188, 80, 255, 134, 148, 152, 175, 153, 150, 100, 34, 111, 45, 135, 181, 102, 175, 113, 178, 146, 206, 146, 190, 97, 188, 100, 240, 127, 146, 135, 55, 152, 68, 152, 251, 153, 64, 139, 215, 150, 26, 112, 43, 134, 190, 157, 60, 0, 42, 0, 111, 81, 51, 114, 60, 0, 42, 0, 47, 89, 60, 0, 42, 0, 192, 130, 210, 143, 187, 101, 109, 103, 179, 87, 215, 126, 233, 115, 16, 123, 42, 130, 162, 134, 131, 152, 165, 140, 85, 123, 78, 125, 15, 152, 103, 155, 60, 0, 42, 0, 134, 108, 60, 0, 42, 0, 32, 131, 191, 132, 134, 86, 133, 133, 167, 133, 60, 0, 42, 0, 220, 107, 211, 122, 157, 134, 235, 107, 3, 105, 229, 85, 70, 115, 137, 140, 81, 86, 84, 115, 106, 140, 55, 86, 75, 115, 213, 138, 43, 81, 142, 86, 213, 88, 224, 111, 71, 124, 20, 136, 121, 139, 60, 0, 42, 0, 125, 89, 221, 144, 60, 0, 42, 0, 247, 83, 10, 102, 38, 102, 207, 121, 224, 84, 79, 96, 142, 96, 105, 109, 23, 128, 103, 102, 207, 109, 144, 80, 147, 118, 23, 145, 200, 110, 85, 128, 95, 134, 164, 102, 173, 102, 148, 111, 156, 118, 158, 118, 205, 102, 161, 118, 131, 133, 165, 118, 172, 147, 162, 152, 79, 112, 101, 152, 29, 156, 93, 112, 60, 0, 42, 0, 195, 139, 117, 84, 178, 98, 49, 107, 157, 85, 54, 138, 236, 85, 26, 136, 60, 0, 42, 0, 190, 121, 8, 84, 85, 79, 190, 82, 138, 84, 140, 84, 192, 89, 179, 108, 195, 144, 198, 92, 247, 102, 199, 103, 226, 114, 199, 118, 122, 124, 7, 125, 2, 150, 120, 153, 236, 84, 70, 101, 56, 104, 201, 118, 205, 118, 119, 131, 93, 85, 184, 109, 46, 110, 210, 118, 244, 121, 207, 131, 2, 132, 181, 134, 129, 159, 210, 96, 173, 124, 56, 138, 140, 152, 65, 105, 252, 107, 149, 111, 101, 138, 136, 140, 5, 143, 76, 146, 22, 150, 132, 156, 134, 113, 86, 158, 167, 158, 28, 152, 213, 123, 238, 127, 155, 135, 122, 155, 9, 121, 212, 149, 168, 151, 85, 159, 136, 137, 161, 157, 172, 118, 73, 148, 162, 159, 60, 0, 42, 0, 107, 79, 142, 87, 58, 141, 148, 136, 201, 91, 3, 113, 192, 140, 195, 85, 66, 113, 139, 120, 135, 113, 16, 137, 107, 141, 100, 158, 82, 122, 239, 127, 209, 88, 75, 118, 30, 139, 250, 113, 0, 114, 174, 157, 180, 157, 77, 151, 78, 151, 22, 158, 79, 151, 60, 0, 42, 0, 210, 158, 209, 158, 63, 86, 118, 111, 60, 0, 42, 0, 235, 98, 213, 117, 142, 151, 60, 0, 42, 0, 119, 79, 136, 95, 224, 114, 106, 138, 60, 0, 42, 0, 104, 96, 60, 0, 42, 0, 168, 78, 252, 84, 153, 96, 165, 109, 72, 85, 29, 129, 60, 0, 42, 0, 238, 89, 70, 96, 82, 96, 65, 104, 198, 112, 251, 128, 59, 158, 42, 106, 107, 106, 97, 136, 52, 157, 70, 157, 5, 134, 69, 148, 60, 0, 42, 0, 60, 88, 60, 0, 42, 0, 119, 86, 60, 0, 42, 0, 255, 83, 61, 84, 77, 84, 116, 112, 112, 143, 196, 84, 7, 138, 216, 112, 227, 142, 200, 99, 57, 110, 34, 113, 97, 120, 62, 140, 168, 133, 55, 143, 157, 86, 103, 147, 95, 143, 60, 0, 42, 0, 220, 78, 24, 95, 133, 89, 162, 126, 48, 84, 143, 91, 111, 108, 146, 115, 174, 126, 243, 149, 150, 91, 211, 108, 156, 115, 240, 130, 172, 87, 2, 90, 42, 109, 209, 122, 5, 125, 109, 131, 121, 134, 100, 109, 24, 125, 195, 127, 62, 128, 84, 120, 45, 125, 57, 140, 63, 158, 49, 110, 228, 122, 160, 124, 82, 132, 83, 132, 28, 146, 142, 149, 139, 125, 221, 127, 60, 140, 66, 111, 119, 146, 131, 151, 95, 155, 208, 146, 75, 95, 123, 133, 16, 151, 201, 158, 31, 151, 59, 157, 204, 158, 60, 0, 42, 0, 78, 102, 202, 85, 60, 0, 42, 0, 167, 139, 12, 138, 167, 149, 148, 100, 139, 111, 146, 111, 190, 146, 194, 149, 40, 155, 192, 149, 60, 0, 42, 0, 65, 159, 60, 0, 42, 0, 175, 79, 230, 119, 7, 145, 137, 85, 63, 94, 52, 115, 84, 132, 10, 118, 122, 119, 204, 123, 199, 124, 237, 127, 186, 154, 245, 127, 109, 147, 49, 153, 248, 155, 60, 0, 42, 0, 60, 84, 188, 114, 60, 0, 42, 0, 14, 84, 200, 144, 154, 83, 149, 87, 140, 95, 9, 109, 5, 144, 25, 80, 32, 88, 94, 140, 142, 156, 152, 156, 156, 155, 95, 156, 60, 0, 42, 0, 78, 78, 111, 78, 34, 83, 77, 134, 124, 84, 128, 87, 253, 95, 18, 102, 246, 102, 216, 108, 248, 130, 87, 96, 192, 112, 119, 143, 43, 83, 63, 85, 218, 96, 244, 109, 86, 134, 228, 142, 17, 86, 227, 91, 249, 110, 208, 150, 96, 94, 47, 98, 81, 107, 180, 129, 60, 139, 60, 0, 42, 0, 235, 86, 135, 98, 39, 95, 208, 114, 243, 116, 225, 128, 246, 88, 186, 150, 247, 88, 155, 101, 0, 113, 150, 85, 250, 88, 169, 90, 48, 100, 86, 110, 34, 115, 87, 125, 107, 132, 92, 105, 115, 113, 90, 116, 29, 86, 27, 133, 85, 158, 242, 105, 182, 123, 202, 124, 116, 135, 90, 136, 113, 155, 32, 126, 156, 135, 144, 145, 54, 152, 243, 137, 120, 147, 44, 153, 96, 157, 43, 112, 13, 155, 23, 156, 152, 157, 166, 157, 60, 0, 42, 0, 85, 78, 123, 108, 78, 134, 82, 109, 255, 79, 44, 85, 0, 132, 37, 116, 93, 134, 248, 110, 63, 147, 241, 155, 60, 0, 42, 0, 101, 78, 146, 78, 22, 95, 54, 98, 55, 98, 56, 98, 177, 81, 180, 81, 144, 130, 13, 94, 164, 98, 141, 108, 170, 108, 181, 92, 25, 96, 61, 98, 8, 102, 145, 103, 24, 96, 92, 121, 15, 123, 144, 124, 95, 90, 72, 98, 224, 116, 91, 105, 192, 85, 148, 125, 32, 145, 253, 150, 237, 90, 238, 90, 98, 100, 236, 110, 48, 133, 244, 105, 169, 113, 248, 156, 4, 124, 89, 147, 155, 86, 113, 158, 119, 139, 224, 156, 196, 151, 0, 152, 111, 156, 12, 158, 60, 0, 42, 0, 177, 130, 178, 130, 215, 84, 214, 87, 114, 90, 27, 105, 116, 120, 41, 86, 192, 124, 174, 138, 138, 132, 53, 147, 36, 134, 60, 0, 42, 0, 78, 83, 225, 89, 133, 154, 239, 131, 234, 145, 235, 145, 231, 148, 209, 110, 62, 115, 51, 100, 182, 100, 198, 120, 80, 133, 150, 135, 216, 146, 65, 139, 245, 147, 74, 154, 232, 157, 60, 0, 42, 0, 22, 83, 18, 82, 59, 89, 121, 103, 59, 117, 221, 139, 11, 93, 102, 104, 115, 90, 107, 117, 5, 91, 117, 117, 223, 137, 113, 138, 131, 82, 102, 100, 58, 106, 255, 90, 236, 105, 133, 111, 217, 138, 227, 138, 202, 158, 99, 126, 25, 130, 110, 139, 60, 0, 42, 0, 0, 96, 138, 95, 238, 109, 208, 105, 34, 137, 29, 142, 208, 97, 49, 137, 247, 97, 36, 112, 240, 106, 50, 128, 57, 134, 60, 0, 42, 0, 79, 87, 182, 84, 202, 88, 222, 88, 62, 134, 60, 0, 42, 0, 34, 107, 191, 114, 83, 107, 5, 157, 77, 157, 68, 145, 190, 86, 253, 97, 126, 115, 97, 107, 153, 139, 155, 140, 105, 154, 60, 0, 42, 0, 175, 115, 199, 144, 216, 92, 57, 109, 223, 114, 65, 131, 83, 104, 8, 132, 17, 132, 207, 91, 89, 125, 200, 150, 132, 125, 166, 127, 134, 140, 110, 146, 62, 149, 28, 87, 27, 91, 240, 91, 180, 111, 51, 127, 27, 150, 176, 116, 114, 140, 112, 147, 110, 149, 110, 158, 235, 124, 111, 126, 88, 143, 54, 148, 228, 149, 31, 155, 219, 116, 60, 0, 42, 0, 19, 127, 233, 125, 12, 101, 60, 0, 42, 0, 123, 94, 66, 89, 146, 128, 80, 89, 166, 91, 36, 85, 98, 99, 99, 109, 163, 109, 201, 112, 163, 96, 153, 104, 21, 113, 45, 144, 154, 85, 155, 85, 72, 93, 12, 97, 219, 99, 25, 110, 234, 117, 70, 119, 101, 113, 77, 116, 98, 140, 54, 111, 19, 118, 245, 105, 169, 156, 208, 100, 163, 111, 231, 133, 199, 155, 246, 155, 0, 156, 60, 0, 42, 0, 223, 93, 147, 128, 82, 131, 65, 136, 26, 103, 67, 88, 76, 97, 60, 0, 42, 0, 135, 118, 95, 80, 240, 81, 141, 150, 196, 158, 164, 85, 45, 88, 147, 90, 50, 93, 168, 95, 246, 96, 95, 110, 95, 132, 81, 144, 195, 158, 123, 105, 76, 113, 93, 116, 180, 88, 98, 111, 90, 115, 61, 149, 191, 113, 156, 116, 193, 123, 202, 123, 78, 130, 87, 135, 64, 118, 250, 120, 84, 122, 251, 138, 39, 124, 229, 135, 96, 147, 45, 153, 199, 156, 170, 141, 249, 151, 4, 148, 28, 154, 9, 156, 81, 156, 236, 157, 60, 0, 42, 0, 51, 96, 77, 96, 190, 112, 186, 91, 67, 102, 68, 102, 91, 89, 14, 140, 76, 94, 100, 138, 128, 113, 40, 126, 10, 139, 206, 106, 100, 81, 60, 0, 42, 0, 48, 97, 201, 110, 165, 105, 194, 102, 157, 118, 164, 147, 169, 118, 60, 0, 42, 0, 112, 112, 115, 112, 217, 139, 180, 84, 98, 96, 251, 98, 37, 99, 3, 109, 122, 134, 134, 136, 86, 102, 227, 112, 242, 115, 87, 140, 78, 90, 136, 90, 238, 99, 218, 127, 137, 143, 147, 150, 137, 102, 78, 105, 71, 113, 136, 121, 124, 138, 81, 94, 115, 119, 24, 137, 69, 86, 157, 100, 85, 86, 236, 127, 29, 143, 190, 158, 189, 95, 179, 150, 8, 112, 51, 134, 52, 156, 60, 0, 42, 0, 216, 86, 222, 86, 236, 86, 106, 79, 251, 94, 253, 94, 91, 96, 4, 109, 52, 131, 244, 143, 224, 112, 152, 134, 37, 144, 208, 117, 212, 134, 213, 134, 22, 135, 176, 155, 60, 0, 42, 0, 148, 96, 192, 107, 193, 107, 199, 107, 147, 106, 236, 113, 109, 139, 60, 0, 42, 0, 73, 83, 118, 92, 119, 92, 71, 108, 26, 79, 179, 139, 203, 108, 213, 84, 77, 109, 216, 126, 148, 130, 95, 131, 242, 139, 90, 96, 117, 96, 233, 112, 63, 141, 87, 95, 102, 102, 253, 121, 153, 85, 224, 96, 79, 110, 117, 125, 11, 127, 217, 127, 19, 150, 47, 83, 89, 95, 90, 95, 3, 103, 217, 110, 111, 138, 196, 140, 146, 152, 225, 80, 18, 86, 35, 118, 39, 133, 168, 138, 26, 87, 237, 91, 103, 97, 147, 97, 179, 102, 229, 105, 83, 111, 89, 133, 102, 86, 18, 91, 187, 95, 94, 106, 168, 107, 174, 111, 202, 111, 105, 115, 164, 116, 136, 133, 137, 133, 241, 138, 46, 152, 133, 106, 244, 113, 175, 116, 242, 123, 241, 133, 47, 153, 150, 86, 186, 119, 98, 122, 98, 126, 234, 135, 216, 106, 106, 126, 253, 127, 83, 139, 54, 81, 248, 147, 224, 149, 72, 91, 44, 148, 103, 151, 127, 139, 106, 152, 60, 0, 42, 0, 243, 97, 60, 0, 42, 0, 15, 102, 44, 102, 100, 131, 90, 90, 219, 96, 189, 109, 13, 150, 212, 104, 153, 107, 119, 132, 103, 119, 111, 119, 189, 149, 60, 0, 42, 0, 246, 95, 81, 109, 161, 104, 132, 153, 26, 88, 62, 110, 63, 116, 66, 155, 27, 153, 73, 126, 75, 143, 50, 159, 60, 0, 42, 0, 232, 139, 210, 79, 49, 80, 2, 87, 141, 99, 247, 109, 29, 113, 183, 110, 65, 97, 232, 137, 226, 138, 60, 0, 42, 0, 25, 84, 104, 82, 32, 128, 42, 149, 144, 82, 132, 86, 67, 147, 65, 140, 9, 101, 30, 154, 60, 0, 42, 0, 120, 79, 59, 109, 238, 121, 243, 121, 60, 0, 42, 0, 107, 112, 25, 79, 169, 144, 172, 148, 37, 146, 55, 111, 37, 89, 60, 0, 42, 0, 142, 108, 22, 98, 39, 141, 159, 84, 9, 120, 240, 79, 71, 99, 19, 119, 183, 131, 132, 149, 157, 99, 120, 121, 168, 140, 209, 96, 228, 101, 71, 105, 113, 110, 141, 121, 166, 132, 111, 89, 233, 111, 114, 115, 13, 151, 180, 106, 11, 139, 198, 119, 107, 122, 108, 149, 175, 86, 22, 112, 47, 128, 103, 130, 255, 133, 22, 136, 191, 86, 228, 102, 219, 129, 104, 118, 208, 119, 74, 148, 67, 151, 60, 0, 39, 0, 208, 253, 74, 0, 39, 0, 60, 0, 42, 0, 12, 78, 165, 139, 251, 81, 9, 82, 253, 83, 101, 153, 105, 78, 15, 82, 62, 87, 58, 103, 145, 115, 140, 128, 168, 130, 246, 119, 33, 158, 133, 103, 173, 84, 235, 89, 249, 143, 94, 82, 39, 85, 236, 89, 80, 92, 239, 121, 4, 123, 226, 152, 250, 87, 233, 126, 158, 85, 70, 93, 71, 93, 103, 101, 30, 103, 132, 114, 83, 123, 9, 127, 77, 141, 227, 82, 216, 85, 120, 117, 24, 122, 225, 141, 251, 141, 238, 156, 223, 80, 196, 107, 149, 123, 136, 146, 48, 86, 131, 100, 227, 105, 45, 106, 127, 117, 61, 122, 221, 125, 237, 137, 235, 140, 184, 142, 81, 159, 188, 88, 95, 106, 192, 111, 163, 116, 168, 121, 77, 122, 64, 137, 36, 147, 174, 150, 202, 100, 239, 120, 10, 124, 62, 126, 129, 127, 247, 140, 63, 145, 197, 106, 45, 128, 95, 142, 222, 150, 79, 139, 242, 151, 143, 157, 100, 139, 22, 148, 81, 153, 106, 118, 139, 142, 191, 151, 196, 157, 78, 159, 135, 127, 64, 134, 71, 148, 137, 137, 89, 148, 79, 159, 136, 127, 4, 158, 138, 137, 60, 0, 42, 0, 188, 78, 189, 78, 202, 83, 11, 79, 9, 84, 140, 92, 118, 95, 227, 95, 114, 108, 167, 126, 115, 83, 129, 103, 128, 118, 159, 78, 118, 79, 216, 139, 198, 144, 125, 83, 221, 83, 222, 89, 37, 96, 228, 114, 141, 118, 8, 123, 26, 125, 228, 99, 190, 117, 10, 129, 217, 137, 110, 80, 89, 83, 180, 94, 15, 113, 59, 140, 34, 98, 216, 104, 117, 105, 155, 107, 82, 110, 198, 150, 73, 88, 201, 90, 49, 97, 107, 105, 186, 132, 140, 141, 145, 143, 201, 105, 36, 128, 140, 129, 161, 146, 175, 93, 87, 111, 200, 111, 32, 118, 191, 123, 64, 133, 122, 133, 22, 142, 97, 158, 118, 106, 157, 106, 143, 135, 47, 143, 201, 133, 75, 137, 80, 142, 83, 147, 101, 130, 77, 124, 90, 143, 246, 147, 53, 151, 186, 157, 209, 157, 164, 142, 230, 150, 231, 150, 60, 0, 42, 0, 224, 81, 241, 93, 46, 78, 128, 89, 177, 114, 242, 108, 110, 134, 36, 99, 142, 99, 126, 156, 126, 94, 31, 98, 116, 93, 130, 158, 98, 155, 160, 100, 224, 100, 86, 122, 227, 135, 85, 155, 60, 0, 42, 0, 80, 95, 81, 95, 225, 101, 161, 139, 176, 139, 14, 79, 170, 126, 86, 87, 147, 89, 204, 95, 128, 98, 176, 130, 69, 150, 66, 82, 99, 91, 220, 84, 141, 87, 220, 92, 226, 101, 14, 109, 78, 109, 0, 125, 13, 131, 36, 131, 96, 131, 8, 138, 100, 82, 18, 125, 231, 126, 202, 137, 24, 138, 72, 80, 194, 91, 196, 91, 155, 95, 184, 96, 227, 101, 158, 104, 8, 110, 109, 121, 72, 88, 206, 96, 238, 129, 106, 132, 7, 133, 126, 81, 245, 117, 153, 125, 223, 132, 218, 136, 155, 150, 62, 155, 168, 102, 3, 111, 8, 111, 157, 121, 41, 122, 74, 122, 139, 138, 253, 141, 1, 151, 154, 156, 169, 102, 55, 122, 197, 138, 171, 156, 128, 81, 145, 82, 193, 102, 68, 122, 138, 133, 251, 154, 140, 86, 149, 106, 223, 111, 75, 126, 125, 127, 186, 133, 172, 137, 181, 106, 75, 157, 76, 159, 237, 94, 251, 97, 96, 118, 103, 122, 14, 134, 165, 154, 218, 155, 49, 112, 124, 126, 46, 134, 64, 156, 59, 134, 61, 151, 54, 156, 63, 156, 109, 156, 101, 154, 60, 0, 42, 0, 160, 82, 107, 78, 57, 89, 61, 79, 62, 89, 184, 98, 115, 79, 193, 98, 199, 108, 4, 131, 230, 143, 183, 103, 224, 107, 67, 109, 200, 115, 201, 87, 182, 91, 121, 109, 194, 117, 156, 104, 51, 123, 30, 128, 136, 136, 162, 80, 51, 115, 109, 132, 207, 141, 140, 114, 117, 129, 107, 146, 9, 86, 127, 146, 83, 149, 216, 124, 109, 140, 145, 140, 181, 147, 154, 158, 60, 0, 42, 0, 63, 87, 230, 95, 116, 98, 207, 144, 90, 131, 223, 144, 10, 85, 93, 96, 162, 131, 27, 98, 183, 136, 215, 148, 30, 98, 241, 134, 204, 136, 138, 152, 250, 134, 242, 141, 4, 153, 207, 146, 44, 152, 48, 152, 54, 157, 74, 157, 60, 0, 42, 0, 50, 117, 172, 92, 218, 83, 190, 115, 219, 128, 154, 101, 62, 141, 190, 148, 71, 80, 125, 90, 166, 95, 157, 101, 53, 105, 200, 140, 64, 146, 142, 105, 218, 105, 21, 118, 159, 106, 60, 0, 42, 0, 247, 78, 126, 154, 182, 103, 193, 90, 79, 94, 162, 105, 249, 80, 60, 122, 213, 153, 60, 0, 42, 0, 11, 98, 120, 89, 22, 92, 117, 94, 90, 87, 124, 107, 244, 149, 191, 81, 20, 98, 170, 115, 169, 128, 112, 130, 230, 89, 231, 89, 124, 81, 209, 118, 5, 88, 228, 96, 15, 115, 58, 123, 197, 131, 250, 131, 92, 140, 84, 110, 75, 114, 141, 114, 4, 127, 76, 132, 147, 149, 27, 100, 55, 105, 62, 105, 78, 113, 74, 116, 119, 119, 138, 120, 35, 127, 185, 132, 99, 140, 227, 118, 139, 123, 43, 106, 158, 113, 216, 125, 81, 133, 115, 133, 163, 156, 253, 156, 99, 158, 184, 113, 239, 123, 17, 126, 113, 130, 172, 151, 48, 153, 162, 153, 137, 158, 16, 112, 175, 151, 210, 156, 177, 107, 27, 121, 184, 137, 115, 157, 56, 112, 39, 148, 252, 106, 178, 107, 188, 157, 192, 151, 57, 156, 207, 86, 67, 134, 111, 148, 201, 151, 60, 0, 42, 0, 221, 86, 227, 98, 167, 103, 237, 79, 236, 103, 39, 131, 57, 80, 56, 99, 97, 99, 21, 123, 207, 81, 106, 82, 152, 104, 192, 104, 85, 110, 188, 141, 63, 88, 192, 99, 195, 99, 28, 105, 27, 110, 81, 119, 119, 120, 229, 136, 67, 138, 15, 149, 63, 95, 149, 102, 80, 116, 103, 123, 128, 123, 120, 125, 43, 140, 69, 95, 41, 98, 44, 98, 177, 120, 9, 81, 230, 127, 191, 100, 162, 106, 198, 133, 71, 137, 73, 137, 7, 139, 71, 142, 188, 119, 6, 121, 33, 124, 109, 126, 45, 139, 11, 155, 14, 156, 120, 158, 61, 112, 18, 136, 23, 148, 151, 82, 123, 158, 91, 124, 126, 139, 122, 137, 124, 158, 60, 0, 42, 0, 193, 137, 246, 78, 139, 137, 250, 94, 111, 153, 81, 82, 10, 109, 110, 114, 80, 131, 49, 141, 244, 79, 101, 80, 99, 82, 43, 104, 167, 109, 212, 115, 48, 130, 113, 82, 164, 95, 16, 110, 184, 136, 15, 140, 252, 145, 203, 91, 212, 101, 87, 105, 253, 107, 133, 110, 113, 129, 246, 129, 101, 132, 245, 141, 206, 140, 116, 146, 46, 149, 237, 80, 151, 105, 56, 111, 42, 133, 141, 82, 142, 82, 185, 88, 151, 111, 173, 123, 203, 124, 211, 138, 228, 140, 157, 141, 16, 142, 58, 142, 146, 82, 148, 82, 166, 133, 235, 138, 251, 146, 117, 147, 30, 153, 183, 119, 245, 120, 0, 121, 185, 135, 115, 147, 246, 100, 250, 111, 93, 126, 51, 112, 181, 137, 233, 147, 102, 130, 124, 139, 94, 143, 49, 148, 81, 148, 82, 148, 108, 148, 115, 148, 60, 0, 42, 0, 122, 106, 60, 0, 42, 0, 95, 108, 220, 89, 6, 92, 51, 131, 70, 109, 85, 117, 71, 140, 7, 92, 65, 132, 122, 117, 106, 100, 222, 127, 245, 80, 63, 111, 128, 135, 195, 88, 48, 127, 145, 133, 127, 106, 173, 107, 191, 135, 201, 156, 133, 117, 19, 121, 134, 117, 110, 126, 193, 151, 66, 156, 60, 0, 42, 0, 178, 139, 86, 89, 104, 104, 139, 80, 139, 132, 104, 89, 108, 89, 35, 133, 243, 105, 78, 115, 41, 128, 153, 129, 27, 139, 92, 152, 60, 0, 42, 0, 30, 83, 32, 83, 5, 89, 28, 95, 77, 150, 26, 109, 219, 126, 54, 95, 182, 136, 115, 125, 113, 145, 229, 82, 240, 110, 121, 93, 126, 100, 74, 95, 159, 114, 225, 124, 164, 145, 232, 124, 172, 145, 228, 106, 61, 139, 60, 0, 42, 0, 98, 103, 60, 0, 42, 0, 125, 130, 129, 130, 164, 78, 202, 144, 227, 89, 7, 90, 231, 92, 71, 109, 45, 131, 46, 131, 132, 154, 246, 128, 18, 105, 38, 113, 223, 134, 228, 141, 236, 80, 4, 86, 96, 134, 155, 156, 12, 91, 149, 93, 163, 93, 141, 97, 134, 111, 160, 129, 73, 133, 203, 113, 178, 129, 1, 121, 90, 122, 171, 155, 65, 157, 106, 158, 37, 124, 237, 135, 71, 143, 14, 148, 85, 154, 230, 157, 238, 157, 60, 0, 42, 0, 235, 129, 210, 137, 124, 79, 165, 79, 84, 96, 34, 99, 225, 114, 222, 126, 122, 153, 65, 99, 72, 102, 196, 112, 5, 123, 142, 118, 235, 119, 26, 129, 240, 148, 5, 100, 107, 110, 94, 125, 127, 82, 107, 101, 108, 110, 77, 113, 115, 129, 203, 140, 229, 80, 119, 100, 158, 102, 11, 142, 120, 146, 3, 153, 12, 81, 139, 82, 186, 95, 159, 100, 185, 100, 166, 150, 188, 95, 191, 97, 125, 101, 127, 101, 222, 113, 52, 127, 210, 102, 172, 116, 239, 119, 166, 118, 220, 135, 115, 126, 81, 139, 66, 91, 144, 126, 42, 101, 90, 112, 78, 156, 60, 0, 42, 0, 235, 83, 76, 84, 228, 92, 13, 99, 6, 138, 211, 115, 140, 122, 127, 143, 131, 143, 78, 101, 89, 101, 150, 122, 216, 110, 3, 143, 2, 86, 38, 86, 160, 101, 22, 111, 117, 145, 77, 86, 160, 93, 80, 111, 109, 86, 19, 91, 101, 115, 224, 133, 173, 141, 78, 143, 174, 145, 101, 139, 173, 118, 194, 145, 60, 0, 42, 0, 100, 157, 245, 106, 60, 0, 42, 0, 54, 150, 150, 117, 134, 118, 165, 99, 178, 99, 206, 117, 248, 121, 232, 131, 142, 150, 136, 85, 188, 85, 223, 85, 38, 88, 152, 90, 197, 90, 237, 99, 4, 105, 93, 110, 59, 129, 87, 136, 111, 113, 45, 122, 130, 151, 209, 100, 84, 135, 100, 118, 47, 139, 155, 157, 60, 0, 42, 0, 105, 83, 106, 83, 81, 91, 16, 92, 130, 130, 166, 139, 38, 82, 39, 82, 171, 82, 138, 92, 5, 102, 60, 82, 188, 82, 112, 103, 140, 117, 113, 136, 238, 98, 1, 109, 211, 126, 252, 143, 34, 80, 64, 104, 93, 104, 173, 131, 16, 138, 124, 80, 85, 90, 40, 93, 119, 99, 182, 99, 186, 136, 145, 80, 171, 90, 80, 125, 92, 125, 137, 152, 101, 93, 108, 105, 118, 105, 208, 110, 107, 119, 192, 123, 16, 135, 77, 135, 112, 138, 99, 146, 93, 155, 42, 98, 164, 105, 163, 120, 237, 122, 245, 132, 146, 156, 84, 111, 175, 127, 177, 138, 21, 142, 138, 151, 111, 94, 187, 93, 238, 100, 13, 121, 123, 147, 154, 155, 192, 93, 237, 106, 30, 136, 24, 136, 61, 136, 60, 0, 42, 0, 208, 89, 209, 107, 142, 90, 227, 137, 231, 137, 247, 152, 158, 106, 60, 0, 42, 0, 47, 78, 203, 78, 36, 84, 149, 92, 142, 94, 18, 98, 165, 130, 70, 92, 74, 92, 160, 115, 76, 117, 77, 117, 165, 117, 14, 120, 120, 136, 235, 139, 31, 80, 136, 96, 167, 134, 163, 95, 58, 88, 80, 105, 62, 116, 246, 134, 177, 154, 151, 114, 161, 138, 47, 137, 106, 155, 133, 147, 60, 0, 42, 0, 254, 93, 202, 78, 164, 101, 133, 148, 83, 81, 209, 145, 210, 145, 37, 109, 220, 119, 27, 120, 85, 131, 127, 136, 212, 137, 208, 87, 210, 115, 221, 119, 31, 125, 205, 96, 59, 88, 59, 116, 75, 123, 255, 145, 156, 93, 118, 158, 197, 158, 95, 137, 60, 0, 42, 0, 197, 78, 61, 92, 173, 79, 122, 83, 249, 93, 39, 125, 7, 88, 235, 131, 197, 80, 170, 83, 40, 140, 38, 149, 228, 90, 209, 94, 12, 111, 225, 118, 202, 125, 243, 132, 145, 153, 255, 105, 126, 116, 24, 81, 38, 147, 57, 139, 73, 153, 60, 0, 42, 0, 18, 79, 164, 82, 178, 82, 151, 89, 209, 143, 219, 143, 131, 103, 193, 82, 85, 109, 105, 131, 73, 102, 75, 102, 120, 109, 236, 112, 70, 141, 43, 85, 14, 116, 114, 121, 50, 144, 97, 113, 214, 91, 34, 100, 141, 110, 129, 121, 25, 127, 115, 151, 144, 88, 104, 116, 248, 80, 218, 81, 79, 107, 163, 107, 161, 116, 208, 137, 100, 86, 197, 111, 9, 126, 238, 140, 141, 86, 215, 88, 39, 91, 220, 111, 206, 133, 252, 113, 182, 116, 178, 137, 16, 141, 125, 159, 60, 0, 42, 0, 85, 87, 89, 87, 224, 93, 172, 78, 254, 108, 207, 126, 14, 131, 176, 78, 212, 121, 70, 131, 74, 131, 135, 109, 150, 131, 91, 90, 202, 96, 204, 101, 205, 101, 4, 115, 76, 125, 193, 131, 118, 102, 9, 122, 72, 129, 79, 132, 91, 119, 179, 124, 147, 125, 98, 81, 190, 124, 89, 128, 120, 106, 184, 156, 91, 157, 232, 155, 129, 157, 132, 157, 150, 158, 49, 159, 90, 154, 160, 158, 60, 0, 42, 0, 149, 78, 60, 78, 49, 150, 45, 82, 83, 87, 145, 91, 107, 108, 108, 108, 188, 128, 68, 82, 125, 122, 136, 152, 111, 102, 6, 81, 26, 152, 92, 94, 172, 97, 132, 116, 188, 97, 187, 102, 159, 116, 165, 116, 56, 152, 252, 135, 102, 139, 60, 0, 42, 0, 140, 89, 192, 81, 42, 95, 132, 95, 243, 143, 211, 79, 89, 90, 68, 109, 235, 128, 30, 80, 200, 81, 51, 95, 145, 95, 201, 117, 222, 122, 21, 144, 103, 90, 113, 104, 183, 104, 232, 109, 235, 122, 27, 129, 223, 122, 108, 101, 217, 117, 231, 122, 83, 151, 185, 80, 86, 151, 131, 88, 77, 115, 169, 138, 1, 142, 89, 151, 90, 151, 212, 102, 92, 149, 92, 151, 30, 112, 225, 147, 246, 122, 248, 122, 60, 0, 42, 0, 221, 113, 60, 0, 42, 0, 130, 81, 139, 81, 112, 87, 67, 98, 219, 87, 69, 125, 201, 153, 235, 153, 15, 134, 20, 134, 60, 0, 42, 0, 143, 81, 231, 86, 194, 108, 133, 112, 229, 143, 176, 79, 175, 112, 8, 144, 123, 109, 241, 112, 90, 113, 152, 122, 142, 152, 151, 125, 210, 80, 91, 113, 178, 113, 131, 111, 219, 113, 39, 137, 60, 0, 42, 0, 41, 78, 252, 82, 160, 126, 59, 103, 94, 114, 118, 122, 250, 124, 32, 158, 254, 124, 115, 141, 4, 150, 27, 132, 126, 85, 194, 99, 234, 99, 235, 99, 233, 156, 78, 100, 27, 106, 15, 155, 46, 155, 60, 0, 42, 0, 93, 78, 69, 78, 70, 78, 99, 78, 122, 89, 99, 108, 102, 103, 120, 112, 150, 115, 15, 130, 237, 151, 36, 125, 82, 145, 121, 149, 238, 151, 60, 0, 42, 0, 27, 83, 231, 101, 252, 129, 142, 84, 154, 117, 233, 103, 254, 103, 3, 80, 68, 99, 85, 104, 19, 83, 169, 83, 81, 101, 49, 92, 196, 94, 208, 94, 5, 130, 230, 80, 207, 94, 102, 97, 167, 107, 10, 130, 107, 158, 54, 83, 230, 155, 148, 158, 13, 107, 104, 159, 242, 157, 60, 0, 42, 0, 229, 81, 133, 98, 10, 83, 69, 92, 216, 98, 195, 108, 217, 114, 244, 130, 121, 154, 54, 99, 189, 117, 192, 117, 23, 119, 32, 120, 93, 127, 113, 150, 53, 90, 110, 90, 12, 93, 172, 99, 174, 104, 186, 109, 249, 131, 16, 105, 26, 116, 82, 129, 132, 141, 212, 141, 20, 149, 254, 136, 206, 150, 77, 130, 27, 135, 202, 138, 24, 142, 230, 146, 210, 153, 136, 155, 33, 157, 160, 151, 171, 151, 139, 157, 60, 0, 42, 0, 64, 92, 230, 108, 183, 79, 202, 114, 84, 104, 233, 107, 121, 85, 69, 90, 215, 109, 23, 113, 202, 131, 249, 144, 8, 105, 241, 107, 104, 110, 145, 114, 2, 143, 234, 80, 183, 124, 252, 141, 176, 149, 156, 141, 185, 142, 88, 106, 139, 106, 246, 153, 89, 157, 107, 142, 116, 157, 200, 93, 28, 134, 170, 157, 48, 159, 51, 159, 103, 154, 60, 0, 42, 0, 128, 84, 6, 95, 174, 108, 62, 78, 233, 119, 146, 131, 25, 99, 7, 105, 101, 123, 137, 105, 152, 105, 159, 132, 131, 159, 101, 128, 9, 130, 61, 142, 231, 100, 248, 106, 95, 159, 5, 107, 119, 137, 60, 0, 42, 0, 232, 93, 229, 83, 108, 78, 234, 93, 181, 139, 214, 89, 160, 92, 7, 96, 210, 98, 48, 109, 227, 130, 173, 144, 119, 81, 16, 96, 26, 96, 224, 98, 27, 102, 107, 107, 172, 112, 236, 121, 156, 148, 241, 79, 40, 80, 54, 80, 163, 81, 103, 82, 229, 112, 148, 124, 31, 128, 183, 134, 147, 136, 231, 87, 254, 87, 231, 96, 110, 99, 78, 138, 221, 141, 139, 114, 217, 141, 69, 146, 211, 152, 97, 134, 102, 140, 47, 149, 224, 91, 51, 97, 173, 122, 90, 128, 207, 153, 135, 82, 238, 82, 102, 92, 30, 142, 148, 155, 201, 88, 197, 97, 218, 100, 189, 111, 182, 122, 125, 144, 248, 146, 104, 92, 182, 152, 151, 140, 52, 124, 134, 142, 181, 145, 252, 97, 59, 148, 60, 0, 42, 0, 32, 114, 60, 0, 42, 0, 226, 89, 31, 90, 80, 99, 147, 109, 6, 113, 249, 116, 39, 129, 208, 136, 67, 158, 236, 82, 76, 149, 184, 147, 81, 157, 43, 148, 50, 136, 60, 0, 42, 0, 119, 83, 95, 84, 35, 94, 205, 87, 114, 99, 228, 131, 41, 149, 199, 129, 8, 147, 60, 0, 42, 0, 70, 89, 181, 82, 251, 93, 38, 80, 204, 82, 74, 104, 247, 114, 226, 126, 189, 150, 195, 109, 55, 119, 4, 145, 74, 119, 109, 125, 101, 127, 203, 150, 96, 119, 121, 125, 236, 152, 123, 97, 40, 133, 11, 153, 103, 115, 51, 126, 130, 127, 60, 0, 42, 0, 88, 86, 133, 100, 167, 100, 105, 92, 123, 142, 60, 0, 42, 0, 133, 78, 82, 91, 83, 91, 179, 81, 20, 82, 18, 108, 192, 139, 33, 95, 137, 98, 122, 108, 181, 130, 236, 108, 166, 115, 168, 115, 23, 99, 207, 115, 166, 117, 4, 120, 221, 126, 115, 134, 201, 137, 20, 80, 84, 99, 46, 107, 151, 134, 27, 93, 152, 99, 141, 101, 119, 104, 140, 107, 144, 137, 214, 137, 35, 138, 125, 141, 185, 141, 43, 144, 149, 80, 165, 83, 51, 113, 85, 125, 118, 125, 154, 137, 137, 141, 12, 146, 130, 82, 234, 82, 116, 116, 50, 140, 195, 153, 165, 93, 176, 97, 166, 113, 52, 114, 87, 115, 26, 118, 93, 133, 104, 133, 2, 157, 3, 157, 113, 86, 160, 97, 91, 106, 92, 106, 53, 114, 196, 129, 98, 149, 232, 135, 233, 135, 107, 92, 17, 114, 78, 139, 118, 142, 119, 142, 140, 157, 55, 83, 188, 86, 205, 119, 186, 137, 13, 148, 29, 148, 77, 112, 29, 114, 252, 137, 79, 95, 4, 98, 43, 101, 131, 115, 226, 157, 20, 107, 225, 119, 163, 159, 156, 140, 169, 142, 129, 148, 60, 0, 42, 0, 155, 81, 27, 84, 71, 87, 110, 108, 240, 89, 128, 136, 205, 142, 167, 148, 153, 131, 144, 134, 126, 104, 178, 118, 204, 131, 30, 146, 133, 120, 184, 118, 185, 118, 160, 137, 129, 146, 158, 146, 170, 156, 135, 158, 85, 147, 182, 155, 143, 158, 149, 158, 60, 0, 42, 0, 65, 84, 202, 79, 225, 144, 86, 150, 200, 87, 251, 92, 67, 99, 90, 109, 130, 153, 143, 154, 89, 102, 12, 113, 250, 115, 222, 104, 111, 117, 227, 122, 1, 81, 152, 123, 159, 123, 32, 135, 239, 91, 207, 97, 21, 153, 199, 113, 236, 111, 255, 153, 84, 157, 85, 157, 88, 157, 8, 101, 31, 101, 60, 0, 39, 0, 208, 253, 75, 0, 39, 0, 60, 0, 42, 0, 148, 84, 150, 84, 128, 85, 73, 136, 214, 100, 60, 0, 42, 0, 97, 83, 103, 79, 176, 87, 233, 128, 195, 136, 114, 146, 60, 0, 42, 0, 0, 95, 82, 89, 233, 99, 14, 149, 139, 149, 38, 148, 60, 0, 42, 0, 239, 81, 64, 82, 178, 87, 122, 96, 255, 149, 224, 148, 241, 81, 116, 82, 5, 86, 104, 97, 136, 132, 79, 88, 102, 93, 55, 97, 119, 105, 6, 143, 159, 102, 52, 149, 71, 147, 167, 147, 211, 149, 189, 152, 60, 0, 42, 0, 254, 95, 140, 112, 143, 112, 44, 107, 215, 112, 211, 82, 18, 97, 62, 97, 142, 147, 60, 0, 42, 0, 10, 82, 30, 104, 216, 82, 155, 159, 42, 88, 65, 93, 33, 98, 149, 159, 60, 0, 42, 0, 154, 81, 78, 87, 131, 79, 13, 120, 176, 131, 88, 80, 243, 87, 194, 96, 63, 107, 106, 88, 65, 107, 219, 105, 33, 143, 187, 106, 81, 152, 247, 122, 87, 143, 60, 0, 42, 0, 11, 119, 78, 136, 1, 93, 136, 88, 176, 119, 225, 120, 222, 149, 217, 119, 60, 0, 42, 0, 252, 95, 246, 149, 10, 120, 135, 124, 183, 94, 221, 90, 123, 93, 119, 97, 46, 111, 250, 105, 69, 122, 224, 124, 191, 142, 238, 147, 71, 156, 60, 0, 42, 0, 91, 98, 67, 100, 60, 0, 42, 0, 162, 78, 9, 79, 31, 83, 159, 144, 229, 86, 151, 98, 186, 114, 149, 112, 170, 148, 39, 146, 140, 149, 60, 0, 42, 0, 59, 92, 219, 154, 60, 0, 42, 0, 2, 78, 55, 101, 3, 128, 247, 98, 24, 109, 50, 104, 228, 112, 60, 0, 42, 0, 208, 148, 146, 114, 172, 146, 147, 156, 96, 151, 179, 155, 204, 155, 60, 0, 42, 0, 60, 83, 219, 130, 239, 103, 65, 114, 194, 115, 209, 121, 226, 128, 114, 143, 180, 117, 34, 120, 183, 141, 182, 148, 89, 93, 245, 104, 42, 132, 251, 142, 143, 152, 209, 85, 21, 100, 144, 114, 30, 122, 160, 122, 51, 146, 188, 105, 150, 133, 151, 152, 22, 106, 140, 119, 213, 120, 76, 135, 18, 147, 152, 145, 70, 152, 193, 154, 26, 121, 60, 0, 42, 0, 243, 88, 226, 99, 188, 107, 215, 127, 60, 0, 42, 0, 239, 83, 119, 87, 162, 92, 163, 112, 7, 110, 81, 93, 100, 101, 52, 110, 177, 93, 60, 0, 42, 0, 75, 81, 59, 82, 75, 82, 192, 82, 202, 82, 162, 91, 106, 96, 20, 90, 5, 92, 254, 139, 1, 88, 42, 108, 146, 154, 2, 127, 25, 97, 152, 110, 30, 149, 166, 120, 217, 125, 80, 130, 178, 138, 10, 121, 13, 154, 60, 0, 42, 0, 142, 128, 175, 128, 187, 128, 166, 87, 115, 96, 67, 85, 100, 140, 136, 159, 190, 88, 57, 147, 199, 97, 102, 159, 60, 0, 42, 0, 175, 99, 201, 136, 3, 137, 60, 0, 42, 0, 165, 82, 44, 150, 45, 84, 81, 87, 148, 89, 51, 99, 65, 120, 124, 114, 92, 120, 255, 148, 123, 120, 124, 100, 153, 138, 181, 146, 94, 147, 215, 147, 60, 0, 42, 0, 122, 122, 37, 80, 234, 87, 6, 93, 190, 96, 179, 109, 127, 120, 156, 123, 187, 142, 19, 147, 124, 157, 60, 0, 42, 0, 84, 91, 80, 96, 60, 0, 42, 0, 167, 99, 154, 151, 60, 0, 42, 0, 160, 98, 164, 130, 13, 119, 126, 82, 68, 95, 115, 100, 152, 119, 60, 0, 42, 0, 227, 83, 182, 82, 60, 0, 42, 0, 233, 83, 99, 98, 66, 101, 166, 81, 188, 91, 199, 91, 230, 145, 155, 122, 88, 123, 241, 110, 50, 133, 59, 133, 137, 119, 6, 124, 199, 157, 60, 0, 42, 0, 93, 98, 51, 82, 251, 119, 192, 144, 175, 103, 237, 84, 77, 104, 0, 88, 43, 93, 16, 87, 205, 141, 159, 122, 183, 154, 172, 155, 60, 0, 42, 0, 220, 114, 230, 130, 60, 0, 42, 0, 147, 94, 200, 79, 212, 126, 171, 94, 217, 121, 182, 141, 5, 113, 180, 136, 190, 85, 93, 125, 228, 136, 20, 118, 119, 145, 228, 94, 50, 137, 179, 86, 60, 0, 42, 0, 56, 89, 241, 89, 63, 130, 135, 138, 60, 0, 42, 0, 137, 79, 181, 84, 174, 87, 153, 146, 60, 0, 42, 0, 14, 99, 239, 128, 232, 141, 187, 154, 60, 0, 42, 0, 175, 132, 211, 100, 60, 0, 42, 0, 220, 93, 247, 81, 38, 87, 87, 87, 235, 95, 169, 79, 208, 144, 217, 84, 239, 114, 13, 129, 74, 88, 119, 123, 153, 156, 8, 81, 164, 88, 54, 145, 114, 86, 229, 94, 106, 115, 190, 129, 221, 101, 233, 124, 96, 156, 60, 0, 42, 0, 189, 91, 219, 91, 236, 91, 215, 129, 203, 154, 103, 148, 214, 154, 60, 0, 42, 0, 53, 107, 62, 107, 64, 107, 189, 122, 190, 122, 60, 0, 42, 0, 33, 83, 187, 82, 211, 139, 188, 144, 41, 83, 208, 84, 71, 96, 45, 109, 70, 104, 68, 120, 80, 123, 122, 123, 134, 138, 237, 142, 60, 0, 42, 0, 249, 95, 130, 98, 194, 114, 197, 114, 243, 139, 214, 142, 224, 142, 145, 138, 95, 157, 60, 0, 42, 0, 60, 89, 35, 81, 237, 97, 60, 0, 42, 0, 93, 83, 157, 144, 57, 87, 169, 126, 181, 81, 247, 101, 178, 92, 193, 108, 255, 119, 63, 102, 54, 141, 22, 119, 63, 120, 54, 119, 75, 125, 86, 125, 186, 140, 230, 142, 113, 146, 219, 146, 58, 145, 217, 88, 203, 158, 236, 97, 224, 102, 12, 114, 204, 119, 38, 121, 108, 122, 138, 126, 91, 148, 60, 0, 42, 0, 143, 78, 50, 82, 191, 92, 157, 96, 212, 118, 165, 122, 103, 128, 186, 122, 103, 134, 93, 152, 218, 149, 203, 93, 44, 134, 60, 0, 42, 0, 78, 89, 70, 102, 53, 144, 8, 145, 151, 150, 4, 152, 151, 153, 185, 85, 198, 99, 117, 132, 153, 154, 35, 98, 140, 102, 79, 105, 81, 105, 65, 155, 125, 119, 112, 135, 47, 152, 198, 106, 200, 133, 104, 147, 119, 147, 36, 154, 20, 89, 55, 134, 217, 93, 65, 134, 170, 114, 168, 142, 60, 0, 42, 0, 67, 113, 236, 141, 13, 152, 94, 142, 60, 0, 42, 0, 47, 92, 46, 83, 51, 107, 159, 85, 191, 90, 38, 97, 39, 97, 131, 110, 67, 129, 137, 132, 136, 153, 134, 119, 51, 86, 7, 91, 146, 97, 112, 111, 209, 123, 105, 128, 109, 128, 98, 133, 59, 106, 9, 139, 61, 153, 35, 124, 117, 128, 68, 124, 0, 148, 75, 153, 78, 148, 60, 0, 42, 0, 100, 87, 6, 102, 3, 88, 18, 88, 107, 90, 16, 93, 17, 93, 92, 102, 17, 115, 206, 131, 200, 136, 28, 113, 40, 116, 224, 154, 233, 136, 135, 140, 31, 149, 225, 154, 77, 158, 73, 111, 43, 135, 12, 137, 232, 154, 180, 113, 123, 116, 140, 145, 21, 147, 178, 156, 9, 154, 228, 155, 126, 157, 164, 157, 60, 0, 42, 0, 131, 96, 70, 99, 3, 150, 248, 88, 177, 104, 117, 121, 113, 120, 7, 122, 205, 136, 252, 88, 27, 122, 145, 125, 171, 149, 184, 149, 60, 0, 42, 0, 240, 86, 131, 109, 79, 119, 60, 0, 42, 0, 33, 92, 60, 0, 42, 0, 105, 98, 225, 98, 236, 98, 4, 99, 112, 104, 72, 123, 63, 132, 64, 132, 222, 134, 20, 150, 211, 94, 34, 152, 250, 154, 244, 100, 246, 111, 202, 149, 159, 151, 213, 151, 214, 97, 41, 151, 185, 151, 32, 155, 60, 0, 39, 0, 208, 253, 76, 0, 39, 0, 60, 0, 42, 0, 131, 87, 201, 98, 198, 103, 203, 127, 200, 131, 26, 100, 139, 144, 60, 0, 42, 0, 239, 101, 76, 82, 44, 120, 230, 99, 214, 120, 185, 86, 60, 0, 42, 0, 135, 85, 222, 133, 60, 0, 42, 0, 130, 110, 74, 129, 231, 99, 75, 105, 12, 118, 33, 135, 75, 135, 162, 143, 163, 143, 114, 135, 200, 129, 11, 101, 9, 114, 216, 129, 14, 155, 206, 116, 116, 149, 251, 155, 31, 136, 94, 148, 60, 0, 42, 0, 102, 85, 161, 151, 60, 0, 42, 0, 101, 103, 134, 79, 235, 79, 8, 80, 3, 93, 149, 95, 158, 109, 177, 131, 242, 144, 97, 90, 13, 93, 178, 94, 160, 95, 190, 104, 246, 109, 13, 115, 10, 132, 40, 144, 246, 104, 28, 116, 89, 123, 252, 148, 130, 123, 56, 147, 11, 154, 224, 155, 134, 157, 179, 158, 60, 0, 42, 0, 59, 85, 73, 141, 80, 119, 94, 119, 86, 141, 218, 140, 209, 111, 244, 140, 60, 152, 66, 152, 94, 118, 99, 157, 40, 112, 44, 112, 65, 124, 254, 133, 244, 106, 105, 118, 112, 137, 95, 124, 60, 0, 42, 0, 112, 81, 154, 92, 230, 98, 15, 104, 106, 90, 207, 96, 80, 93, 123, 132, 17, 150, 221, 132, 48, 140, 177, 83, 156, 111, 52, 137, 22, 81, 147, 101, 238, 123, 226, 97, 227, 113, 247, 113, 205, 133, 85, 137, 103, 149, 204, 149, 188, 116, 100, 137, 75, 139, 113, 94, 20, 101, 62, 112, 70, 112, 67, 124, 127, 126, 45, 134, 149, 101, 4, 107, 55, 121, 116, 137, 210, 86, 97, 112, 99, 124, 23, 107, 149, 139, 157, 142, 125, 137, 132, 148, 202, 151, 60, 0, 42, 0, 200, 137, 104, 109, 253, 99, 6, 127, 132, 105, 36, 111, 113, 127, 130, 145, 200, 88, 210, 97, 167, 137, 229, 100, 62, 91, 246, 97, 68, 91, 189, 137, 79, 91, 44, 101, 96, 112, 213, 86, 22, 107, 38, 114, 114, 152, 156, 126, 60, 0, 42, 0, 194, 112, 229, 110, 215, 113, 130, 86, 235, 111, 1, 114, 27, 114, 211, 116, 36, 114, 109, 148, 247, 124, 60, 0, 42, 0, 119, 85, 60, 0, 42, 0, 198, 82, 206, 144, 222, 144, 52, 107, 252, 114, 6, 150, 207, 90, 202, 94, 143, 101, 121, 104, 5, 116, 200, 132, 148, 105, 111, 116, 96, 120, 2, 122, 18, 149, 100, 123, 70, 130, 11, 135, 210, 144, 130, 135, 180, 142, 195, 146, 175, 147, 250, 153, 60, 0, 42, 0, 23, 103, 22, 103, 250, 112, 113, 88, 226, 132, 3, 106, 143, 138, 36, 103, 60, 0, 42, 0, 204, 87, 0, 93, 106, 109, 168, 131, 151, 132, 172, 149, 60, 0, 42, 0, 37, 85, 60, 0, 42, 0, 94, 99, 169, 124, 136, 100, 60, 0, 42, 0, 179, 82, 180, 82, 98, 114, 130, 122, 240, 84, 32, 85, 2, 93, 118, 109, 222, 82, 232, 117, 249, 148, 215, 80, 46, 86, 151, 93, 165, 97, 37, 103, 70, 118, 241, 120, 41, 124, 231, 135, 170, 145, 18, 148, 95, 152, 221, 154, 60, 0, 42, 0, 2, 128, 1, 128, 108, 79, 190, 84, 229, 89, 69, 96, 235, 114, 86, 131, 51, 104, 239, 115, 83, 120, 209, 148, 239, 134, 160, 146, 102, 111, 81, 106, 177, 155, 81, 143, 60, 0, 42, 0, 157, 109, 217, 112, 224, 85, 34, 128, 106, 145, 234, 90, 166, 97, 135, 111, 188, 142, 111, 106, 46, 128, 194, 142, 60, 0, 42, 0, 139, 128, 60, 0, 42, 0, 194, 78, 30, 150, 80, 78, 251, 83, 199, 95, 80, 98, 59, 108, 123, 130, 143, 115, 208, 108, 251, 122, 51, 120, 125, 105, 247, 151, 2, 106, 21, 124, 211, 156, 51, 156, 60, 0, 42, 0, 134, 78, 121, 153, 14, 153, 60, 0, 42, 0, 210, 82, 60, 0, 42, 0, 247, 150, 216, 90, 39, 127, 2, 133, 126, 117, 194, 100, 145, 106, 50, 126, 12, 121, 109, 149, 209, 106, 195, 116, 184, 127, 39, 121, 141, 126, 77, 127, 50, 134, 51, 148, 96, 143, 61, 81, 232, 88, 88, 148, 65, 151, 70, 134, 105, 156, 25, 107, 157, 126, 58, 159, 60, 0, 42, 0, 189, 83, 18, 128, 212, 139, 146, 87, 65, 88, 107, 125, 66, 129, 171, 80, 132, 138, 15, 106, 202, 120, 76, 133, 229, 120, 126, 133, 33, 81, 216, 88, 87, 118, 223, 133, 208, 106, 203, 119, 40, 121, 69, 112, 29, 136, 61, 134, 132, 139, 120, 148, 19, 158, 60, 0, 42, 0, 234, 108, 33, 109, 123, 124, 153, 109, 218, 109, 47, 125, 121, 145, 135, 146, 27, 152, 42, 152, 17, 147, 2, 101, 163, 152, 94, 152, 135, 126, 49, 134, 183, 121, 60, 0, 42, 0, 30, 86, 60, 0, 42, 0, 26, 93, 68, 88, 241, 104, 94, 105, 144, 120, 28, 122, 24, 143, 144, 133, 60, 0, 42, 0, 183, 81, 60, 0, 42, 0, 48, 80, 14, 88, 35, 97, 86, 119, 28, 142, 60, 0, 42, 0, 233, 84, 60, 0, 42, 0, 21, 82, 93, 103, 152, 83, 83, 82, 248, 114, 187, 121, 114, 131, 138, 154, 161, 96, 168, 104, 169, 104, 184, 104, 129, 114, 13, 116, 222, 131, 177, 85, 195, 104, 130, 114, 66, 158, 122, 82, 19, 111, 93, 119, 99, 123, 33, 127, 67, 130, 224, 132, 10, 135, 224, 90, 119, 91, 6, 106, 131, 116, 224, 118, 240, 122, 141, 140, 206, 124, 62, 133, 53, 137, 235, 146, 161, 156, 206, 158, 241, 123, 45, 126, 121, 127, 5, 147, 205, 135, 39, 139, 168, 145, 159, 86, 220, 133, 140, 144, 208, 145, 226, 150, 207, 155, 132, 101, 200, 116, 235, 147, 236, 155, 121, 157, 231, 158, 196, 86, 85, 112, 58, 134, 33, 136, 43, 136, 75, 91, 242, 94, 153, 82, 87, 148, 114, 122, 108, 124, 154, 126, 106, 154, 122, 156, 29, 158, 60, 0, 42, 0, 60, 121, 78, 103, 204, 145, 218, 79, 219, 92, 226, 92, 12, 90, 242, 92, 108, 109, 38, 144, 6, 116, 225, 136, 2, 149, 180, 124, 207, 136, 74, 140, 240, 146, 164, 156, 99, 81, 167, 111, 174, 121, 201, 155, 248, 135, 180, 145, 226, 156, 144, 144, 103, 156, 26, 107, 60, 0, 42, 0, 155, 82, 134, 83, 137, 83, 116, 92, 203, 122, 15, 84, 56, 103, 61, 78, 41, 82, 177, 82, 86, 84, 92, 87, 165, 108, 200, 130, 139, 79, 166, 92, 62, 98, 165, 103, 180, 108, 160, 117, 217, 130, 182, 150, 208, 79, 234, 79, 3, 104, 14, 104, 172, 117, 5, 120, 24, 131, 84, 131, 114, 141, 121, 143, 230, 144, 14, 85, 51, 90, 167, 96, 23, 104, 27, 104, 150, 109, 1, 115, 213, 115, 58, 120, 62, 120, 221, 121, 133, 131, 137, 131, 51, 85, 111, 90, 183, 96, 32, 123, 146, 124, 157, 124, 55, 129, 184, 134, 206, 134, 136, 80, 211, 81, 164, 83, 217, 104, 226, 117, 224, 134, 72, 138, 222, 141, 243, 150, 175, 83, 91, 88, 68, 97, 46, 100, 167, 110, 154, 132, 158, 132, 93, 146, 232, 156, 178, 83, 166, 102, 116, 107, 110, 116, 159, 125, 39, 135, 119, 135, 245, 82, 198, 102, 119, 107, 229, 123, 183, 150, 23, 157, 193, 93, 170, 106, 255, 111, 88, 118, 255, 120, 184, 150, 1, 155, 46, 81, 222, 102, 212, 106, 4, 114, 161, 114, 178, 121, 7, 136, 152, 147, 166, 86, 226, 88, 10, 101, 223, 106, 29, 112, 197, 116, 42, 121, 246, 133, 151, 158, 234, 106, 15, 114, 209, 116, 170, 118, 237, 118, 43, 121, 242, 124, 35, 136, 55, 81, 103, 118, 48, 121, 72, 145, 197, 157, 156, 158, 199, 86, 38, 101, 251, 137, 146, 142, 60, 0, 42, 0, 98, 143, 16, 107, 136, 139, 99, 143, 45, 101, 229, 116, 66, 151, 113, 156, 115, 156, 75, 151, 60, 0, 42, 0, 233, 79, 6, 80, 60, 0, 42, 0, 65, 89, 222, 143, 24, 94, 28, 96, 159, 109, 178, 131, 35, 144, 191, 104, 84, 128, 226, 136, 183, 78, 249, 85, 201, 94, 105, 97, 147, 110, 35, 111, 238, 132, 50, 83, 105, 89, 228, 105, 145, 113, 157, 137, 134, 82, 51, 83, 82, 86, 254, 90, 144, 97, 207, 120, 104, 128, 107, 128, 51, 137, 162, 156, 194, 111, 211, 111, 58, 126, 244, 127, 110, 128, 149, 133, 138, 135, 227, 106, 235, 113, 111, 128, 193, 129, 48, 139, 101, 142, 140, 147, 112, 149, 62, 124, 10, 136, 17, 155, 46, 148, 49, 156, 98, 124, 104, 124, 60, 0, 42, 0, 91, 101, 15, 116, 56, 129, 227, 136, 89, 100, 137, 116, 57, 133, 26, 91, 130, 101, 201, 129, 59, 145, 93, 137, 183, 127, 29, 134, 30, 134, 60, 0, 42, 0, 195, 126, 188, 112, 75, 96, 112, 109, 147, 107, 198, 80, 28, 88, 161, 90, 69, 110, 48, 132, 254, 148, 93, 105, 73, 113, 83, 116, 75, 111, 244, 125, 176, 111, 44, 147, 174, 107, 74, 147, 200, 147, 50, 112, 10, 156, 0, 98, 158, 126, 60, 0, 42, 0, 111, 130, 205, 79, 201, 81, 129, 104, 188, 109, 11, 105, 140, 143, 174, 124, 177, 124, 154, 88, 161, 125, 9, 142, 17, 106, 44, 143, 231, 124, 60, 0, 42, 0, 33, 78, 36, 78, 105, 81, 33, 85, 98, 85, 154, 99, 60, 129, 242, 136, 201, 125, 61, 135, 73, 155, 78, 155, 60, 0, 42, 0, 174, 78, 244, 84, 162, 96, 5, 140, 134, 143, 168, 85, 126, 102, 120, 110, 207, 145, 12, 143, 210, 138, 27, 143, 68, 147, 60, 0, 42, 0, 119, 113, 23, 124, 60, 0, 42, 0, 169, 100, 125, 142, 60, 0, 42, 0, 189, 143, 151, 117, 74, 128, 218, 80, 229, 91, 122, 93, 128, 97, 59, 111, 139, 129, 57, 86, 253, 90, 238, 91, 154, 93, 155, 93, 121, 101, 96, 115, 45, 127, 124, 144, 184, 102, 206, 113, 153, 116, 171, 129, 66, 118, 194, 122, 105, 158, 106, 92, 235, 94, 29, 124, 90, 126, 223, 135, 66, 140, 255, 140, 88, 142, 14, 114, 16, 148, 206, 154, 212, 133, 201, 152, 239, 157, 60, 0, 42, 0, 254, 83, 140, 148, 213, 145, 29, 145, 252, 132, 173, 97, 173, 119, 226, 102, 125, 149, 18, 114, 60, 0, 42, 0, 37, 92, 38, 92, 147, 112, 153, 101, 30, 92, 214, 94, 130, 100, 183, 122, 99, 149, 60, 0, 42, 0, 223, 107, 167, 84, 24, 99, 60, 0, 42, 0, 23, 82, 163, 82, 189, 81, 189, 82, 244, 89, 18, 99, 12, 109, 34, 131, 254, 143, 247, 84, 210, 87, 211, 87, 53, 104, 86, 109, 200, 112, 238, 112, 105, 99, 14, 115, 31, 115, 31, 129, 218, 134, 194, 136, 109, 113, 89, 119, 87, 128, 148, 141, 228, 93, 178, 152, 32, 81, 164, 155, 55, 157, 248, 100, 117, 115, 163, 114, 144, 142, 27, 155, 35, 155, 114, 156, 60, 0, 42, 0, 206, 98, 60, 0, 42, 0, 184, 83, 187, 144, 151, 103, 52, 78, 167, 81, 73, 85, 10, 93, 203, 109, 125, 102, 51, 116, 166, 124, 243, 117, 132, 120, 150, 123, 188, 124, 48, 145, 163, 150, 153, 93, 126, 111, 92, 115, 116, 144, 180, 101, 189, 102, 208, 113, 152, 116, 154, 143, 22, 151, 181, 119, 247, 120, 232, 129, 87, 126, 247, 127, 144, 158, 84, 143, 227, 88, 54, 112, 251, 147, 222, 156, 78, 154, 87, 156, 159, 158, 60, 0, 42, 0, 251, 131, 131, 78, 219, 81, 220, 81, 155, 100, 233, 94, 234, 94, 205, 97, 212, 97, 159, 111, 129, 106, 169, 106, 91, 118, 93, 118, 60, 0, 42, 0, 29, 84, 97, 96, 139, 96, 65, 141, 27, 113, 195, 140, 239, 80, 58, 133, 73, 106, 16, 117, 166, 129, 181, 149, 132, 117, 250, 133, 120, 142, 143, 142, 153, 142, 170, 142, 101, 143, 60, 0, 42, 0, 7, 48, 54, 79, 34, 82, 117, 112, 249, 86, 125, 87, 12, 89, 200, 89, 186, 92, 126, 95, 224, 108, 209, 114, 211, 130, 36, 102, 14, 103, 195, 103, 178, 115, 244, 116, 204, 81, 138, 118, 49, 120, 226, 121, 219, 122, 195, 148, 117, 150, 48, 158, 72, 90, 149, 99, 194, 104, 233, 109, 12, 116, 45, 123, 55, 125, 235, 126, 154, 127, 206, 127, 70, 128, 50, 130, 241, 131, 201, 134, 81, 136, 126, 121, 69, 138, 201, 141, 232, 142, 228, 132, 236, 136, 52, 146, 157, 149, 246, 150, 132, 159, 190, 125, 6, 133, 10, 151, 214, 153, 170, 111, 118, 133, 2, 147, 23, 151, 127, 155, 174, 156, 18, 157, 119, 158, 239, 113, 27, 151, 29, 151, 98, 159, 67, 145, 234, 155, 65, 91, 38, 134, 97, 159, 250, 106, 189, 145, 72, 151, 30, 107, 39, 114, 162, 158, 151, 159, 60, 0, 42, 0, 62, 150, 173, 92, 138, 136, 134, 152, 24, 152, 186, 93, 60, 0, 42, 0, 228, 78, 230, 83, 100, 84, 169, 112, 60, 0, 42, 0, 46, 112, 60, 0, 42, 0, 156, 110, 152, 113, 83, 142, 60, 0, 42, 0, 24, 82, 160, 108, 68, 117, 79, 109, 65, 109, 89, 117, 200, 101, 9, 116, 113, 117, 107, 120, 215, 136, 185, 90, 103, 93, 210, 101, 165, 132, 197, 132, 91, 144, 143, 153, 157, 154, 180, 105, 96, 116, 215, 152, 137, 82, 108, 116, 36, 118, 194, 120, 79, 149, 224, 153, 96, 158, 74, 106, 162, 116, 129, 117, 96, 149, 69, 118, 201, 135, 245, 153, 160, 86, 240, 97, 15, 112, 240, 133, 143, 147, 166, 147, 141, 158, 208, 147, 192, 152, 2, 148, 46, 154, 197, 152, 33, 156, 185, 157, 81, 154, 60, 0, 42, 0, 243, 103, 1, 104, 94, 104, 203, 115, 122, 104, 250, 126, 13, 149, 90, 146, 185, 125, 174, 113, 118, 127, 246, 146, 110, 106, 60, 91, 128, 127, 60, 0, 42, 0, 109, 81, 66, 117, 207, 127, 111, 88, 199, 94, 145, 111, 223, 120, 104, 158, 36, 151, 62, 153, 225, 150, 194, 152, 56, 155, 218, 157, 60, 0, 42, 0, 214, 86, 60, 0, 42, 0, 153, 159, 120, 92, 153, 84, 247, 108, 15, 131, 61, 102, 10, 104, 209, 115, 231, 128, 44, 119, 59, 120, 220, 122, 60, 123, 75, 128, 134, 150, 112, 110, 221, 110, 144, 93, 222, 105, 11, 111, 111, 133, 67, 118, 191, 122, 237, 123, 141, 159, 168, 86, 195, 93, 196, 93, 39, 112, 34, 134, 231, 147, 51, 151, 232, 102, 39, 103, 243, 106, 22, 114, 207, 116, 211, 119, 49, 121, 50, 121, 113, 137, 146, 159, 96, 124, 126, 128, 42, 136, 44, 136, 69, 140, 152, 142, 104, 148, 71, 151, 97, 154, 23, 158, 60, 0, 42, 0, 71, 150, 132, 87, 133, 87, 226, 98, 226, 123, 49, 81, 180, 150, 223, 88, 224, 88, 15, 101, 201, 122, 147, 159, 60, 0, 42, 0, 226, 84, 135, 104, 191, 95, 26, 141, 60, 0, 42, 0, 156, 119, 60, 0, 42, 0, 69, 82, 4, 90, 123, 80, 65, 90, 189, 85, 135, 110, 140, 132, 194, 80, 124, 105, 13, 86, 212, 94, 122, 97, 10, 111, 30, 133, 113, 144, 19, 106, 161, 113, 39, 128, 124, 135, 44, 128, 91, 130, 187, 135, 49, 139, 193, 142, 197, 154, 187, 151, 207, 154, 60, 0, 42, 0, 93, 93, 2, 100, 127, 88, 129, 93, 95, 100, 10, 117, 211, 123, 13, 124, 60, 0, 42, 0, 75, 150, 90, 92, 15, 111, 24, 118, 66, 149, 58, 118, 59, 118, 228, 147, 60, 0, 42, 0, 92, 86, 184, 100, 34, 139, 60, 0, 42, 0, 98, 83, 144, 94, 166, 130, 134, 87, 166, 103, 248, 108, 137, 112, 12, 104, 234, 128, 115, 143, 46, 130, 44, 158, 136, 115, 59, 130, 133, 152, 41, 146, 136, 156, 114, 155, 231, 118, 167, 86, 218, 88, 236, 94, 14, 101, 24, 112, 121, 115, 183, 116, 6, 134, 229, 102, 232, 106, 16, 114, 208, 116, 218, 129, 209, 119, 90, 124, 145, 126, 79, 127, 107, 130, 38, 136, 100, 143, 106, 148, 113, 152, 215, 154, 120, 156, 21, 158, 248, 158, 60, 0, 42, 0, 100, 83, 79, 134, 179, 99, 117, 158, 117, 120, 129, 156, 92, 134, 119, 88, 247, 110, 254, 132, 16, 106, 155, 111, 111, 155, 196, 100, 121, 106, 224, 120, 101, 149, 149, 86, 252, 100, 2, 112, 211, 106, 12, 108, 99, 130, 192, 147, 106, 130, 42, 148, 101, 148, 60, 0, 42, 0, 37, 87, 42, 117, 70, 150, 147, 79, 116, 87, 84, 95, 85, 95, 205, 92, 206, 82, 66, 141, 130, 143, 120, 150, 61, 90, 213, 109, 229, 109, 12, 110, 73, 120, 201, 131, 47, 144, 127, 158, 2, 105, 45, 116, 132, 121, 127, 121, 199, 80, 121, 82, 224, 82, 221, 118, 105, 119, 140, 120, 17, 122, 194, 140, 239, 141, 118, 88, 216, 94, 93, 100, 9, 111, 147, 123, 182, 124, 13, 133, 46, 98, 26, 106, 157, 113, 148, 129, 238, 137, 162, 141, 27, 142, 152, 143, 129, 145, 94, 111, 75, 122, 87, 133, 4, 147, 50, 147, 52, 147, 144, 116, 15, 124, 176, 135, 54, 124, 87, 142, 70, 143, 4, 154, 109, 158, 44, 124, 53, 124, 213, 147, 229, 155, 102, 157, 113, 157, 147, 158, 244, 147, 50, 151, 60, 154, 89, 124, 66, 134, 250, 157, 60, 0, 42, 0, 7, 108, 60, 0, 42, 0, 116, 154, 254, 149, 136, 105, 173, 149, 191, 153, 0, 108, 162, 129, 218, 106, 216, 133, 220, 157, 98, 154, 60, 0, 42, 0, 21, 84, 66, 84, 163, 79, 216, 144, 182, 79, 20, 99, 91, 99, 75, 99, 197, 101, 160, 104, 18, 113, 99, 121, 6, 122, 221, 148, 97, 92, 125, 125, 21, 127, 98, 92, 130, 129, 27, 137, 193, 146, 101, 92, 144, 129, 56, 137, 34, 81, 94, 122, 55, 126, 109, 122, 60, 0, 42, 0, 253, 91, 143, 87, 139, 95, 81, 134, 135, 115, 255, 126, 66, 93, 47, 108, 78, 132, 228, 110, 160, 125, 209, 125, 110, 97, 187, 123, 159, 129, 244, 82, 66, 126, 254, 111, 214, 106, 8, 114, 98, 148, 60, 0, 42, 0, 8, 90, 106, 91, 230, 92, 27, 99, 62, 104, 62, 158, 20, 129, 230, 110, 174, 146, 73, 157, 29, 87, 113, 89, 76, 91, 127, 91, 210, 93, 35, 101, 235, 102, 18, 107, 83, 112, 137, 127, 224, 129, 30, 87, 100, 112, 74, 134, 126, 148, 116, 118, 117, 118, 30, 158, 60, 0, 42, 0, 117, 83, 60, 0, 42, 0, 113, 78, 224, 145, 130, 78, 60, 0, 42, 0, 160, 99, 101, 117, 103, 117, 10, 149, 36, 122, 25, 87, 221, 146, 226, 146, 253, 100, 60, 0, 42, 0, 161, 98, 132, 99, 60, 0, 42, 0, 209, 78, 38, 79, 245, 86, 166, 108, 182, 126, 150, 79, 110, 143, 43, 80, 111, 150, 7, 87, 104, 90, 24, 93, 25, 93, 192, 96, 234, 109, 213, 131, 198, 104, 64, 129, 184, 125, 38, 135, 26, 142, 42, 143, 238, 120, 0, 147, 233, 155, 60, 0, 42, 0, 232, 87, 150, 120, 16, 122, 35, 128, 60, 0, 42, 0, 186, 139, 163, 110, 214, 138, 60, 0, 42, 0, 87, 127, 112, 85, 49, 152, 201, 86, 60, 0, 42, 0, 86, 127, 33, 115, 54, 129, 29, 132, 59, 144, 36, 105, 97, 129, 153, 137, 35, 149, 169, 123, 161, 154, 89, 149, 186, 135, 133, 127, 182, 137, 205, 147, 56, 81, 188, 137, 62, 154, 30, 101, 128, 115, 63, 134, 143, 144, 15, 107, 88, 154, 1, 158, 110, 124, 124, 148, 96, 153, 60, 0, 42, 0, 70, 82, 46, 80, 207, 132, 248, 136, 182, 142, 48, 118, 3, 136, 221, 129, 234, 102, 115, 118, 60, 0, 42, 0, 250, 108, 200, 92, 27, 109, 220, 126, 102, 131, 134, 154, 28, 109, 222, 115, 102, 120, 63, 123, 97, 125, 61, 132, 94, 100, 47, 111, 150, 114, 123, 146, 210, 150, 241, 153, 165, 155, 60, 157, 69, 157, 252, 111, 153, 126, 60, 0, 39, 0, 208, 253, 77, 0, 39, 0, 60, 0, 42, 0, 99, 84, 60, 0, 42, 0, 136, 89, 86, 91, 189, 90, 36, 91, 55, 91, 60, 0, 42, 0, 187, 158, 242, 117, 52, 133, 152, 114, 198, 135, 199, 135, 60, 0, 42, 0, 108, 154, 155, 115, 1, 120, 130, 134, 172, 153, 164, 110, 100, 144, 106, 116, 188, 120, 158, 135, 183, 147, 34, 156, 204, 157, 60, 0, 42, 0, 184, 114, 105, 103, 67, 121, 129, 149, 130, 154, 27, 85, 140, 80, 65, 115, 112, 119, 28, 86, 170, 105, 161, 121, 117, 127, 225, 153, 35, 121, 21, 155, 60, 0, 42, 0, 135, 78, 23, 84, 206, 85, 27, 86, 242, 90, 60, 0, 42, 0, 203, 87, 182, 133, 62, 151, 60, 0, 42, 0, 112, 78, 108, 131, 183, 140, 42, 86, 82, 133, 246, 157, 60, 0, 42, 0, 162, 82, 200, 143, 69, 79, 242, 88, 166, 158, 86, 83, 9, 129, 8, 129, 165, 158, 71, 136, 241, 82, 227, 140, 129, 144, 33, 151, 34, 151, 60, 0, 42, 0, 218, 90, 159, 152, 60, 0, 42, 0, 207, 89, 151, 96, 238, 134, 200, 80, 41, 140, 114, 97, 146, 153, 32, 106, 146, 119, 158, 119, 148, 151, 82, 142, 62, 139, 69, 153, 215, 156, 98, 152, 23, 155, 24, 155, 59, 156, 59, 136, 60, 0, 42, 0, 88, 92, 128, 110, 76, 119, 225, 110, 255, 110, 168, 135, 84, 137, 206, 135, 203, 147, 213, 119, 60, 0, 42, 0, 252, 102, 36, 145, 129, 88, 84, 94, 98, 97, 113, 100, 43, 111, 76, 115, 38, 127, 4, 133, 19, 133, 254, 105, 179, 113, 183, 111, 88, 149, 53, 126, 221, 147, 48, 134, 60, 0, 42, 0, 100, 114, 60, 0, 42, 0, 153, 144, 2, 84, 217, 95, 82, 108, 146, 130, 40, 92, 87, 103, 103, 103, 19, 108, 242, 118, 126, 96, 0, 123, 43, 131, 228, 84, 15, 90, 172, 94, 93, 109, 245, 114, 123, 114, 109, 120, 239, 145, 211, 148, 221, 117, 214, 134, 233, 146, 249, 153, 60, 0, 42, 0, 189, 131, 190, 131, 101, 120, 59, 131, 254, 88, 45, 111, 210, 135, 14, 136, 60, 0, 42, 0, 43, 115, 147, 140, 60, 0, 42, 0, 219, 107, 219, 119, 134, 103, 102, 114, 5, 131, 6, 131, 196, 101, 94, 127, 94, 81, 53, 110, 222, 142, 85, 145, 37, 88, 26, 149, 230, 154, 2, 108, 155, 114, 101, 135, 243, 154, 40, 147, 202, 135, 156, 157, 60, 0, 42, 0, 135, 81, 111, 83, 24, 89, 110, 78, 60, 98, 193, 92, 214, 108, 52, 102, 198, 148, 55, 123, 233, 132, 60, 0, 42, 0, 131, 81, 131, 118, 188, 130, 144, 81, 2, 131, 146, 81, 213, 103, 10, 119, 56, 141, 4, 128, 164, 136, 146, 137, 162, 90, 61, 94, 58, 132, 191, 140, 26, 145, 23, 97, 147, 102, 89, 105, 247, 107, 65, 116, 128, 119, 140, 140, 46, 145, 80, 135, 203, 97, 60, 0, 42, 0, 146, 86, 60, 0, 42, 0, 72, 78, 188, 158, 249, 111, 156, 86, 102, 118, 60, 0, 42, 0, 69, 84, 70, 87, 146, 108, 161, 108, 154, 103, 171, 115, 250, 130, 2, 104, 9, 119, 18, 90, 4, 129, 147, 131, 133, 104, 251, 115, 34, 129, 255, 144, 51, 88, 146, 90, 75, 93, 68, 110, 72, 110, 56, 115, 66, 119, 127, 132, 99, 105, 115, 105, 100, 113, 66, 116, 150, 121, 92, 129, 122, 88, 209, 105, 118, 145, 69, 149, 91, 158, 194, 146, 9, 151, 72, 122, 190, 95, 135, 147, 192, 119, 23, 101, 42, 134, 165, 157, 244, 158, 60, 0, 42, 0, 206, 107, 207, 107, 194, 81, 142, 127, 52, 99, 124, 109, 132, 90, 68, 93, 60, 110, 186, 90, 65, 149, 13, 91, 216, 113, 190, 142, 130, 147, 227, 158, 60, 0, 42, 0, 185, 89, 186, 98, 172, 108, 192, 101, 39, 102, 89, 121, 130, 136, 27, 119, 154, 90, 208, 91, 215, 117, 202, 141, 61, 155, 93, 113, 120, 119, 206, 151, 69, 155, 195, 123, 94, 135, 60, 0, 42, 0, 26, 105, 60, 0, 42, 0, 232, 149, 106, 98, 167, 115, 148, 148, 128, 149, 133, 149, 107, 99, 219, 131, 138, 116, 70, 147, 185, 78, 75, 134, 60, 0, 42, 0, 247, 149, 22, 113, 182, 96, 170, 102, 220, 113, 209, 97, 227, 97, 60, 0, 42, 0, 236, 78, 17, 80, 60, 0, 42, 0, 221, 100, 60, 0, 42, 0, 63, 117, 123, 134, 161, 81, 148, 131, 12, 132, 149, 110, 223, 118, 153, 132, 13, 117, 26, 81, 87, 106, 162, 119, 68, 133, 113, 135, 51, 145, 56, 145, 106, 94, 222, 97, 219, 111, 218, 102, 38, 103, 172, 106, 11, 108, 199, 119, 30, 121, 205, 155, 114, 158, 104, 130, 9, 134, 210, 119, 63, 151, 64, 151, 91, 153, 109, 152, 6, 159, 15, 158, 60, 0, 42, 0, 208, 82, 27, 115, 254, 116, 48, 149, 75, 130, 34, 135, 220, 97, 116, 115, 51, 147, 245, 97, 19, 136, 237, 155, 60, 0, 42, 0, 95, 91, 166, 104, 34, 89, 35, 89, 37, 151, 60, 0, 42, 0, 185, 99, 60, 0, 42, 0, 170, 84, 47, 119, 135, 119, 60, 0, 42, 0, 158, 81, 37, 95, 98, 121, 247, 143, 174, 136, 21, 115, 28, 140, 190, 132, 120, 138, 14, 139, 154, 145, 76, 95, 223, 100, 220, 124, 59, 126, 138, 158, 139, 158, 176, 121, 97, 151, 48, 112, 124, 115, 155, 158, 126, 149, 2, 98, 32, 101, 213, 116, 60, 134, 34, 114, 190, 145, 191, 145, 13, 158, 196, 145, 60, 0, 42, 0, 115, 124, 136, 130, 142, 79, 181, 108, 139, 127, 45, 95, 35, 109, 73, 101, 43, 119, 142, 124, 18, 129, 51, 110, 94, 132, 29, 133, 164, 146, 212, 111, 74, 91, 86, 112, 60, 0, 42, 0, 150, 81, 248, 124, 104, 108, 149, 108, 147, 91, 204, 108, 197, 137, 218, 92, 85, 121, 187, 91, 216, 121, 198, 91, 231, 109, 147, 137, 148, 137, 66, 94, 39, 140, 83, 88, 78, 94, 155, 137, 39, 86, 147, 105, 245, 110, 30, 111, 144, 113, 36, 133, 28, 135, 15, 159, 170, 81, 18, 106, 102, 94, 215, 111, 204, 133, 16, 139, 193, 106, 26, 124, 131, 127, 60, 0, 42, 0, 128, 91, 135, 130, 99, 103, 32, 119, 66, 90, 245, 126, 148, 90, 201, 104, 191, 125, 220, 125, 241, 129, 82, 135, 53, 91, 176, 106, 203, 106, 200, 119, 202, 119, 207, 119, 60, 0, 42, 0, 15, 78, 69, 108, 77, 81, 148, 108, 254, 158, 201, 82, 4, 119, 41, 90, 109, 80, 149, 81, 212, 82, 17, 110, 149, 85, 16, 97, 78, 110, 5, 127, 66, 132, 123, 125, 124, 129, 253, 158, 236, 125, 160, 111, 184, 155, 60, 0, 42, 0, 99, 151, 98, 151, 198, 124, 170, 158, 171, 158, 186, 158, 181, 158, 60, 0, 42, 0, 181, 85, 60, 0, 42, 0, 215, 130, 140, 90, 207, 99, 132, 119, 75, 158, 249, 90, 226, 125, 147, 157, 89, 156, 60, 0, 42, 0, 106, 103, 7, 119, 210, 121, 252, 109, 58, 110, 8, 127, 206, 123, 242, 125, 208, 133, 136, 144, 60, 0, 42, 0, 153, 89, 153, 94, 133, 115, 215, 122, 191, 94, 223, 94, 60, 0, 42, 0, 92, 78, 0, 84, 169, 84, 246, 84, 109, 91, 60, 0, 42, 0, 109, 112, 213, 112, 149, 137, 35, 100, 197, 110, 17, 133, 142, 133, 19, 157, 109, 94, 241, 97, 254, 123, 215, 106, 27, 136, 74, 136, 86, 148, 116, 156, 60, 0, 42, 0, 17, 108, 196, 89, 183, 92, 222, 95, 11, 96, 251, 101, 252, 101, 224, 130, 201, 115, 255, 118, 7, 120, 96, 127, 15, 93, 106, 99, 24, 116, 29, 116, 23, 127, 73, 116, 251, 117, 136, 120, 49, 146, 205, 125, 225, 125, 9, 147, 22, 157, 114, 147, 60, 0, 42, 0, 191, 118, 186, 81, 33, 82, 245, 149, 191, 98, 239, 108, 196, 82, 67, 101, 253, 149, 175, 96, 79, 101, 34, 123, 61, 123, 253, 96, 99, 110, 148, 149, 13, 97, 111, 101, 139, 102, 169, 149, 246, 80, 92, 97, 171, 97, 99, 111, 34, 124, 216, 156, 32, 136, 53, 156, 60, 0, 42, 0, 138, 87, 60, 0, 42, 0, 13, 84, 14, 102, 35, 158, 58, 109, 0, 119, 23, 131, 165, 81, 25, 103, 51, 119, 237, 148, 13, 145, 199, 90, 159, 110, 61, 115, 194, 132, 157, 102, 160, 105, 152, 146, 244, 156, 145, 119, 159, 135, 173, 137, 60, 0, 42, 0, 114, 79, 243, 89, 213, 81, 79, 97, 105, 145, 60, 0, 42, 0, 125, 84, 39, 105, 122, 138, 60, 0, 42, 0, 181, 99, 60, 0, 42, 0, 44, 140, 44, 139, 60, 0, 42, 0, 120, 100, 60, 0, 42, 0, 133, 94, 27, 92, 31, 140, 235, 90, 141, 153, 121, 100, 33, 106, 156, 129, 189, 158, 105, 100, 121, 155, 69, 106, 232, 120, 226, 124, 40, 139, 41, 139, 164, 86, 245, 100, 67, 153, 169, 86, 176, 86, 17, 134, 205, 154, 84, 155, 152, 82, 93, 153, 60, 0, 42, 0, 185, 98, 225, 97, 60, 0, 42, 0, 43, 103, 176, 82, 61, 87, 186, 89, 19, 94, 126, 107, 127, 107, 129, 107, 171, 108, 9, 131, 76, 150, 30, 94, 41, 102, 186, 103, 28, 85, 140, 118, 28, 119, 63, 119, 30, 120, 227, 121, 136, 131, 171, 131, 61, 119, 150, 124, 72, 125, 80, 110, 232, 134, 131, 140, 252, 85, 123, 88, 222, 91, 32, 111, 79, 115, 230, 132, 138, 140, 175, 102, 134, 146, 122, 151, 252, 90, 217, 158, 60, 118, 144, 119, 153, 119, 70, 149, 105, 155, 168, 88, 216, 158, 14, 112, 152, 140, 230, 133, 212, 135, 204, 147, 5, 114, 64, 154, 51, 121, 134, 126, 49, 128, 60, 0, 42, 0, 61, 96, 191, 158, 60, 0, 42, 0, 222, 84, 60, 0, 42, 0, 95, 114, 148, 79, 186, 82, 72, 96, 32, 109, 56, 119, 11, 140, 209, 134, 42, 127, 14, 142, 126, 146, 0, 139, 180, 119, 70, 126, 106, 147, 62, 157, 176, 158, 60, 0, 42, 0, 208, 103, 60, 0, 42, 0, 234, 107, 1, 108, 178, 88, 60, 0, 42, 0, 205, 107, 169, 78, 97, 114, 118, 87, 198, 89, 199, 98, 212, 92, 115, 114, 70, 117, 82, 117, 223, 128, 93, 117, 94, 117, 42, 120, 110, 117, 103, 146, 7, 142, 60, 0, 42, 0, 40, 103, 235, 78, 48, 103, 238, 118, 233, 81, 144, 108, 199, 114, 145, 112, 103, 114, 220, 130, 227, 107, 175, 131, 158, 134, 188, 148, 223, 82, 238, 150, 147, 88, 85, 94, 89, 94, 84, 97, 88, 105, 102, 119, 108, 146, 85, 97, 174, 102, 82, 130, 2, 151, 70, 122, 56, 126, 170, 151, 60, 0, 39, 0, 208, 253, 78, 0, 39, 0, 60, 0, 42, 0, 239, 85, 60, 0, 42, 0, 207, 98, 255, 98, 16, 99, 241, 85, 78, 149, 191, 147, 60, 0, 42, 0, 120, 78, 234, 84, 235, 150, 60, 0, 42, 0, 163, 144, 54, 84, 80, 84, 160, 89, 179, 126, 173, 128, 28, 90, 114, 136, 160, 148, 13, 125, 166, 136, 122, 99, 29, 123, 125, 140, 220, 142, 128, 140, 9, 146, 179, 132, 121, 151, 118, 155, 60, 0, 42, 0, 73, 129, 139, 113, 104, 100, 123, 91, 60, 0, 42, 0, 67, 78, 118, 89, 127, 130, 22, 108, 147, 117, 179, 89, 252, 94, 250, 143, 55, 80, 226, 145, 45, 91, 60, 0, 42, 0, 72, 89, 240, 103, 15, 128, 16, 128, 24, 132, 63, 110, 16, 159, 38, 137, 154, 135, 60, 147, 60, 0, 42, 0, 225, 86, 60, 0, 42, 0, 55, 117, 169, 98, 143, 103, 189, 79, 87, 83, 223, 103, 26, 90, 88, 117, 174, 131, 190, 150, 131, 85, 86, 144, 148, 102, 96, 105, 245, 138, 227, 150, 60, 0, 42, 0, 103, 141, 199, 99, 115, 110, 51, 132, 117, 113, 105, 129, 123, 135, 1, 98, 60, 0, 42, 0, 123, 90, 60, 0, 42, 0, 212, 86, 60, 0, 42, 0, 106, 78, 162, 86, 104, 139, 202, 86, 48, 136, 30, 155, 149, 153, 28, 107, 98, 153, 60, 0, 42, 0, 195, 100, 233, 102, 46, 101, 98, 112, 60, 0, 42, 0, 62, 81, 73, 159, 60, 0, 42, 0, 108, 91, 60, 0, 42, 0, 118, 84, 19, 96, 32, 99, 241, 92, 71, 120, 217, 148, 49, 115, 242, 134, 73, 138, 153, 120, 147, 100, 169, 93, 185, 97, 239, 135, 18, 89, 74, 139, 3, 148, 206, 93, 60, 0, 42, 0, 180, 87, 124, 96, 169, 96, 17, 129, 24, 83, 51, 129, 22, 88, 241, 96, 208, 90, 89, 116, 102, 129, 175, 120, 118, 115, 127, 115, 60, 0, 42, 0, 249, 149, 101, 90, 214, 109, 153, 149, 39, 155, 209, 129, 60, 0, 42, 0, 146, 117, 183, 139, 144, 98, 50, 119, 37, 138, 60, 0, 42, 0, 98, 84, 60, 0, 42, 0, 30, 90, 129, 153, 46, 129, 71, 129, 18, 153, 190, 155, 216, 155, 60, 0, 42, 0, 103, 81, 133, 81, 29, 108, 23, 147, 60, 0, 42, 0, 65, 96, 233, 90, 240, 90, 60, 0, 42, 0, 253, 128, 60, 0, 42, 0, 174, 89, 60, 0, 42, 0, 60, 92, 109, 87, 41, 96, 229, 108, 126, 124, 42, 80, 84, 92, 220, 121, 243, 144, 204, 148, 255, 87, 87, 90, 227, 109, 10, 115, 173, 134, 255, 104, 220, 141, 93, 129, 99, 128, 58, 135, 236, 137, 142, 140, 23, 143, 19, 151, 181, 156, 226, 155, 145, 158, 111, 159, 225, 129, 60, 0, 42, 0, 49, 79, 96, 79, 223, 98, 179, 98, 212, 114, 232, 130, 197, 103, 206, 101, 114, 102, 116, 91, 46, 146, 156, 153, 23, 81, 30, 81, 172, 150, 236, 100, 191, 133, 183, 106, 123, 128, 60, 0, 42, 0, 112, 92, 60, 108, 50, 79, 225, 143, 53, 102, 210, 128, 6, 144, 63, 83, 36, 119, 4, 88, 196, 96, 223, 90, 53, 97, 186, 110, 104, 119, 123, 129, 177, 102, 12, 126, 189, 138, 169, 129, 58, 91, 60, 0, 42, 0, 200, 98, 43, 133, 60, 0, 42, 0, 116, 94, 202, 121, 214, 84, 229, 121, 135, 156, 142, 155, 182, 156, 71, 157, 207, 158, 240, 155, 60, 0, 42, 0, 138, 109, 123, 99, 240, 109, 62, 113, 200, 141, 135, 143, 151, 143, 154, 100, 181, 100, 190, 120, 38, 143, 16, 124, 77, 142, 6, 101, 104, 142, 142, 142, 60, 0, 42, 0, 68, 83, 255, 94, 245, 95, 233, 89, 56, 85, 221, 87, 76, 130, 60, 0, 42, 0, 24, 90, 34, 91, 67, 91, 60, 0, 42, 0, 127, 145, 184, 145, 192, 145, 60, 0, 42, 0, 31, 158, 17, 131, 133, 136, 229, 156, 203, 90, 202, 136, 38, 133, 34, 106, 29, 91, 45, 137, 50, 91, 60, 0, 42, 0, 63, 92, 50, 129, 60, 0, 42, 0, 79, 99, 209, 99, 60, 0, 42, 0, 246, 130, 60, 0, 42, 0, 7, 94, 60, 87, 191, 103, 103, 150, 133, 109, 198, 117, 66, 128, 236, 129, 110, 85, 215, 96, 205, 131, 137, 150, 166, 85, 92, 101, 124, 110, 235, 85, 114, 93, 2, 142, 91, 86, 112, 100, 247, 105, 23, 142, 25, 142, 74, 149, 77, 149, 173, 93, 222, 123, 242, 129, 28, 147, 158, 152, 81, 142, 153, 86, 118, 128, 179, 147, 209, 149, 124, 91, 125, 91, 241, 106, 75, 124, 22, 134, 193, 86, 103, 159, 213, 93, 241, 124, 245, 124, 37, 136, 72, 148, 211, 86, 152, 139, 161, 142, 119, 148, 115, 152, 128, 148, 60, 0, 42, 0, 220, 86, 168, 96, 60, 0, 42, 0, 240, 98, 60, 0, 42, 0, 12, 129, 60, 0, 42, 0, 129, 91, 155, 84, 231, 98, 222, 114, 231, 130, 224, 103, 77, 128, 205, 91, 213, 91, 47, 117, 215, 91, 220, 91, 231, 91, 28, 81, 221, 81, 128, 86, 35, 91, 240, 100, 112, 115, 180, 133, 184, 106, 121, 128, 79, 148, 33, 155, 11, 158, 60, 0, 42, 0, 99, 106, 195, 119, 60, 0, 42, 0, 94, 79, 171, 79, 222, 108, 63, 80, 157, 111, 216, 111, 60, 0, 42, 0, 158, 89, 60, 0, 42, 0, 91, 114, 92, 114, 124, 108, 60, 0, 42, 0, 248, 95, 109, 98, 195, 114, 189, 126, 132, 112, 174, 148, 16, 125, 165, 131, 21, 146, 117, 151, 60, 0, 42, 0, 156, 81, 172, 79, 221, 84, 83, 109, 19, 129, 254, 121, 178, 143, 2, 81, 179, 143, 101, 86, 195, 111, 125, 133, 130, 106, 246, 113, 175, 121, 191, 129, 96, 122, 91, 137, 178, 145, 1, 107, 60, 0, 42, 0, 119, 126, 60, 0, 42, 0, 4, 95, 10, 99, 53, 99, 81, 118, 72, 159, 60, 0, 42, 0, 186, 127, 60, 0, 42, 0, 66, 85, 60, 0, 42, 0, 200, 105, 40, 128, 115, 115, 189, 106, 146, 147, 30, 148, 115, 139, 60, 0, 42, 0, 116, 89, 101, 91, 125, 154, 47, 123, 209, 153, 60, 0, 42, 0, 46, 79, 170, 82, 41, 95, 46, 120, 236, 128, 60, 0, 42, 0, 18, 96, 137, 80, 25, 100, 60, 0, 42, 0, 115, 89, 149, 148, 121, 124, 249, 145, 60, 0, 42, 0, 145, 108, 66, 136, 103, 96, 18, 103, 68, 136, 60, 0, 42, 0, 123, 89, 60, 0, 42, 0, 28, 110, 150, 102, 86, 113, 87, 113, 42, 153, 60, 0, 42, 0, 159, 117, 80, 134, 120, 120, 39, 118, 60, 0, 42, 0, 193, 158, 60, 0, 42, 0, 205, 144, 42, 99, 155, 104, 169, 80, 58, 81, 60, 0, 42, 0, 96, 106, 60, 0, 42, 0, 250, 139, 143, 85, 191, 99, 61, 144, 30, 97, 38, 100, 24, 149, 59, 100, 146, 105, 44, 122, 254, 138, 67, 142, 209, 124, 230, 97, 231, 97, 229, 124, 100, 122, 239, 124, 60, 0, 39, 0, 208, 253, 79, 0, 39, 0, 60, 0, 42, 0, 148, 85, 98, 86, 60, 0, 42, 0, 230, 84, 60, 0, 42, 0, 125, 123, 60, 0, 42, 0, 180, 139, 164, 108, 39, 107, 180, 107, 239, 116, 37, 158, 120, 88, 26, 111, 80, 107, 198, 107, 176, 113, 12, 117, 146, 129, 14, 157, 217, 106, 242, 133, 51, 139, 194, 147, 215, 157, 60, 0, 42, 0, 117, 159, 60, 0, 42, 0, 24, 84, 85, 84, 118, 80, 98, 129, 20, 86, 38, 128, 69, 133, 213, 133, 60, 0, 42, 0, 4, 96, 106, 97, 60, 0, 39, 0, 208, 253, 80, 0, 39, 0, 60, 0, 42, 0, 145, 89, 133, 118, 180, 141, 37, 130, 106, 85, 105, 132, 60, 0, 42, 0, 119, 103, 44, 114, 177, 99, 54, 116, 98, 123, 86, 111, 60, 0, 42, 0, 10, 94, 21, 94, 21, 96, 153, 136, 60, 0, 42, 0, 205, 98, 60, 0, 42, 0, 243, 79, 152, 95, 146, 99, 5, 115, 209, 104, 76, 114, 43, 143, 48, 124, 50, 124, 164, 114, 60, 0, 42, 0, 249, 94, 60, 0, 42, 0, 204, 84, 62, 109, 18, 110, 67, 110, 142, 132, 131, 147, 60, 0, 42, 0, 5, 119, 104, 117, 32, 132, 88, 111, 0, 101, 60, 0, 42, 0, 63, 114, 0, 109, 216, 118, 216, 141, 187, 90, 75, 94, 176, 132, 43, 100, 195, 105, 228, 118, 208, 120, 15, 126, 251, 120, 10, 112, 224, 135, 99, 142, 156, 147, 182, 151, 60, 0, 42, 0, 184, 81, 36, 82, 156, 108, 218, 98, 238, 108, 141, 112, 219, 83, 73, 114, 252, 118, 84, 117, 65, 128, 162, 136, 74, 138, 191, 110, 22, 152, 236, 146, 101, 157, 123, 137, 123, 148, 60, 0, 42, 0, 83, 78, 151, 108, 238, 128, 241, 150, 194, 110, 150, 129, 54, 151, 60, 0, 42, 0, 144, 83, 158, 94, 150, 83, 4, 144, 193, 101, 61, 130, 206, 90, 172, 95, 131, 135, 209, 156, 142, 159, 144, 159, 60, 0, 42, 0, 217, 85, 42, 128, 171, 137, 60, 0, 42, 0, 144, 112, 168, 128, 214, 128, 60, 0, 42, 0, 155, 98, 203, 98, 44, 129, 34, 132, 60, 0, 42, 0, 40, 82, 134, 84, 137, 87, 150, 94, 205, 114, 176, 112, 46, 114, 141, 136, 15, 83, 243, 142, 132, 151, 28, 137, 131, 158, 133, 158, 60, 0, 42, 0, 209, 141, 60, 0, 42, 0, 69, 89, 225, 108, 174, 112, 177, 117, 176, 118, 50, 120, 173, 158, 31, 121, 46, 121, 60, 0, 42, 0, 120, 84, 12, 96, 167, 128, 248, 103, 218, 128, 67, 136, 133, 145, 60, 0, 42, 0, 43, 150, 106, 150, 249, 87, 240, 107, 84, 141, 43, 149, 244, 136, 245, 136, 224, 140, 205, 153, 60, 0, 42, 0, 214, 79, 60, 0, 42, 0, 2, 79, 155, 108, 105, 79, 20, 94, 245, 89, 190, 101, 198, 101, 127, 109, 238, 115, 77, 145, 220, 132, 148, 143, 183, 153, 143, 93, 8, 151, 97, 143, 60, 0, 42, 0, 183, 85, 116, 86, 85, 107, 60, 0, 42, 0, 235, 116, 198, 118, 83, 110, 80, 132, 60, 0, 42, 0, 96, 84, 248, 127, 60, 0, 42, 0, 175, 85, 60, 0, 42, 0, 9, 83, 38, 96, 168, 98, 114, 96, 48, 120, 136, 104, 249, 112, 81, 120, 239, 142, 155, 149, 48, 111, 45, 86, 142, 111, 222, 120, 60, 0, 42, 0, 131, 130, 11, 103, 55, 99, 252, 122, 23, 80, 145, 131, 11, 88, 56, 95, 109, 95, 218, 104, 22, 105, 115, 88, 124, 120, 29, 122, 236, 132, 79, 158, 240, 105, 37, 106, 162, 113, 137, 97, 35, 143, 227, 123, 247, 123, 168, 129, 11, 147, 248, 151, 252, 154, 218, 135, 219, 135, 5, 155, 132, 126, 21, 134, 252, 151, 108, 157, 47, 154, 20, 155, 93, 148, 60, 0, 42, 0, 103, 99, 206, 109, 143, 118, 123, 82, 60, 0, 42, 0, 189, 99, 42, 105, 176, 120, 43, 142, 60, 0, 42, 0, 21, 78, 19, 79, 62, 79, 121, 98, 176, 126, 179, 144, 111, 87, 171, 98, 183, 98, 139, 112, 201, 114, 18, 120, 130, 96, 219, 121, 224, 121, 21, 125, 205, 148, 199, 101, 205, 127, 26, 128, 126, 140, 8, 146, 26, 146, 57, 146, 95, 146, 148, 146, 136, 82, 199, 120, 211, 153, 236, 154, 124, 86, 3, 147, 13, 147, 126, 155, 141, 155, 181, 97, 20, 121, 21, 121, 57, 151, 60, 0, 42, 0, 174, 118, 48, 150, 152, 130, 175, 92, 135, 103, 222, 107, 211, 114, 182, 128, 215, 107, 216, 107, 178, 117, 141, 134, 235, 144, 116, 150, 100, 85, 228, 87, 37, 93, 189, 134, 190, 134, 124, 140, 55, 113, 53, 116, 62, 129, 87, 129, 143, 156, 116, 127, 141, 129, 49, 135, 110, 155, 192, 88, 250, 123, 183, 135, 148, 140, 103, 157, 134, 127, 7, 103, 25, 159, 47, 136, 60, 0, 42, 0, 57, 83, 128, 94, 139, 117, 243, 78, 46, 87, 201, 130, 52, 129, 222, 117, 162, 146, 192, 138, 4, 157, 215, 100, 125, 86, 86, 118, 173, 86, 60, 0, 42, 0, 65, 92, 224, 109, 202, 99, 253, 145, 178, 90, 211, 90, 101, 119, 159, 143, 78, 111, 43, 122, 251, 80, 188, 111, 138, 86, 19, 117, 136, 117, 108, 139, 226, 149, 255, 157, 10, 158, 60, 0, 42, 0, 232, 86, 79, 80, 165, 90, 143, 114, 199, 123, 233, 127, 66, 147, 163, 157, 60, 0, 42, 0, 136, 154, 252, 128, 29, 140, 65, 129, 68, 105, 105, 105, 198, 140, 240, 141, 218, 138, 191, 154, 65, 142, 226, 153, 8, 154, 60, 0, 42, 0, 145, 137, 181, 140, 222, 138, 60, 0, 42, 0, 71, 114, 151, 154, 120, 155, 23, 154, 25, 154, 60, 0, 42, 0, 125, 82, 111, 95, 83, 97, 37, 127, 216, 152, 218, 101, 242, 127, 181, 135, 165, 114, 195, 152, 196, 152, 82, 155, 60, 0, 42, 0, 214, 90, 226, 116, 184, 133, 221, 149, 60, 0, 42, 0, 141, 107, 159, 119, 251, 123, 57, 126, 165, 145, 171, 118, 96, 152, 60, 0, 42, 0, 104, 121, 196, 80, 225, 82, 12, 86, 177, 95, 2, 111, 60, 0, 42, 0, 21, 108, 135, 100, 134, 100, 188, 102, 165, 119, 60, 0, 42, 0, 63, 78, 228, 130, 5, 148, 60, 0, 42, 0, 243, 90, 60, 0, 42, 0, 216, 89, 252, 98, 23, 121, 102, 122, 170, 153, 94, 154, 60, 0, 42, 0, 173, 115, 43, 141, 38, 90, 167, 140, 21, 116, 212, 90, 145, 152, 59, 152, 42, 91, 178, 133, 172, 86, 201, 119, 25, 136, 166, 152, 112, 152, 60, 0, 42, 0, 193, 84, 128, 105, 60, 0, 42, 0, 93, 114, 86, 108, 88, 128, 60, 0, 42, 0, 82, 78, 57, 117, 220, 79, 9, 90, 132, 109, 47, 120, 96, 128, 117, 130, 238, 122, 41, 152, 60, 0, 42, 0, 115, 94, 196, 139, 237, 81, 111, 84, 106, 87, 188, 92, 217, 108, 249, 130, 241, 144, 79, 92, 33, 94, 176, 103, 52, 109, 182, 115, 211, 128, 83, 131, 246, 116, 91, 92, 50, 94, 220, 109, 13, 132, 178, 134, 64, 88, 72, 94, 41, 113, 1, 117, 62, 127, 241, 132, 226, 134, 85, 138, 255, 142, 134, 156, 244, 81, 127, 97, 179, 123, 39, 143, 145, 97, 131, 155, 152, 106, 8, 124, 11, 134, 60, 0, 42, 0, 139, 148, 97, 87, 165, 92, 202, 108, 252, 108, 140, 110, 85, 146, 23, 152, 81, 111, 250, 147, 60, 0, 42, 0, 70, 90, 25, 86, 34, 133, 49, 145, 164, 118, 8, 139, 199, 106, 60, 0, 42, 0, 245, 83, 0, 92, 183, 148, 56, 123, 135, 152, 202, 153, 60, 0, 42, 0, 182, 92, 235, 143, 64, 101, 34, 102, 38, 109, 192, 115, 222, 112, 52, 120, 54, 120, 217, 145, 149, 124, 102, 145, 170, 132, 68, 155, 151, 145, 60, 0, 42, 0, 114, 104, 60, 0, 42, 0, 86, 82, 29, 90, 60, 0, 42, 0, 148, 98, 153, 98, 74, 99, 138, 99, 210, 136, 129, 123, 7, 147, 60, 0, 42, 0, 133, 84, 227, 84, 68, 90, 131, 114, 60, 0, 42, 0, 198, 78, 52, 101, 53, 101, 81, 98, 135, 112, 96, 150, 87, 86, 178, 100, 125, 111, 200, 100, 198, 155, 60, 0, 42, 0, 36, 87, 13, 83, 134, 131, 47, 129, 233, 131, 208, 131, 97, 132, 177, 132, 178, 132, 213, 80, 122, 145, 163, 88, 91, 115, 158, 116, 238, 111, 168, 119, 89, 122, 100, 149, 12, 141, 128, 126, 247, 147, 60, 0, 42, 0, 52, 103, 3, 87, 102, 109, 243, 112, 110, 102, 17, 87, 165, 110, 156, 102, 49, 140, 233, 138, 56, 106, 6, 108, 143, 106, 104, 149, 92, 139, 124, 142, 32, 148, 60, 0, 42, 0, 250, 148, 22, 130, 23, 130, 234, 146, 17, 112, 221, 102, 60, 0, 42, 0, 236, 93, 237, 93, 60, 0, 39, 0, 208, 253, 81, 0, 39, 0, 60, 0, 42, 0, 3, 78, 201, 143, 143, 108, 187, 89, 210, 103, 27, 80, 196, 81, 22, 104, 100, 104, 234, 144, 56, 90, 189, 96, 26, 98, 127, 99, 124, 104, 210, 109, 11, 132, 50, 101, 31, 103, 242, 104, 58, 107, 42, 125, 227, 134, 4, 137, 219, 80, 1, 86, 125, 97, 191, 105, 6, 111, 192, 125, 124, 97, 237, 105, 198, 138, 255, 138, 11, 151, 74, 142, 76, 155, 218, 147, 136, 157, 60, 0, 42, 0, 147, 78, 65, 121, 80, 159, 59, 87, 144, 92, 147, 92, 239, 95, 170, 130, 157, 78, 118, 81, 71, 89, 137, 101, 103, 107, 71, 121, 72, 121, 181, 128, 223, 79, 167, 117, 210, 122, 88, 82, 138, 101, 194, 101, 6, 128, 16, 129, 145, 134, 148, 134, 154, 134, 128, 152, 252, 87, 14, 93, 58, 94, 145, 99, 199, 109, 9, 115, 102, 117, 1, 132, 21, 132, 194, 141, 221, 142, 238, 145, 144, 154, 145, 154, 92, 93, 202, 104, 203, 104, 38, 116, 42, 116, 122, 121, 244, 134, 45, 97, 129, 120, 149, 120, 0, 137, 28, 149, 14, 152, 63, 155, 215, 101, 184, 124, 165, 125, 166, 125, 168, 125, 29, 135, 30, 135, 74, 159, 130, 116, 165, 121, 114, 133, 17, 142, 167, 135, 33, 147, 175, 156, 224, 97, 221, 111, 196, 133, 177, 106, 192, 106, 49, 124, 205, 129, 14, 154, 15, 154, 205, 156, 4, 134, 213, 155, 120, 157, 128, 157, 146, 158, 79, 124, 131, 126, 105, 130, 16, 136, 16, 155, 57, 154, 45, 156, 130, 115, 161, 158, 60, 0, 42, 0, 94, 78, 148, 144, 1, 79, 122, 92, 130, 92, 145, 130, 47, 84, 71, 84, 94, 103, 152, 115, 192, 118, 24, 85, 72, 140, 119, 141, 83, 85, 84, 85, 77, 90, 95, 85, 238, 126, 117, 102, 232, 104, 174, 125, 186, 125, 236, 138, 217, 149, 60, 0, 42, 0, 20, 108, 171, 139, 212, 95, 23, 108, 84, 108, 196, 143, 3, 95, 125, 108, 245, 119, 158, 130, 110, 84, 227, 108, 129, 112, 245, 118, 160, 84, 81, 89, 12, 120, 20, 104, 35, 108, 22, 138, 45, 85, 43, 107, 33, 89, 196, 104, 70, 110, 71, 110, 122, 132, 155, 120, 86, 100, 163, 102, 8, 117, 182, 120, 80, 86, 135, 97, 104, 86, 169, 97, 220, 120, 231, 120, 233, 120, 74, 127, 255, 135, 28, 159, 60, 0, 42, 0, 60, 127, 213, 125, 47, 124, 60, 0, 42, 0, 144, 99, 92, 132, 60, 0, 42, 0, 228, 98, 60, 0, 42, 0, 224, 92, 210, 141, 96, 145, 144, 151, 60, 0, 42, 0, 54, 87, 190, 81, 34, 94, 112, 96, 61, 109, 142, 107, 72, 120, 24, 97, 194, 154, 60, 0, 42, 0, 67, 83, 223, 78, 33, 150, 49, 87, 50, 87, 119, 89, 102, 98, 88, 108, 138, 130, 193, 143, 101, 79, 141, 92, 68, 103, 103, 108, 233, 116, 62, 131, 38, 107, 207, 122, 228, 129, 142, 148, 234, 98, 117, 114, 129, 124, 91, 81, 173, 96, 136, 134, 56, 140, 197, 148, 92, 90, 111, 91, 125, 114, 250, 145, 148, 99, 38, 140, 6, 146, 195, 150, 201, 80, 6, 97, 126, 123, 91, 146, 158, 154, 80, 158, 115, 97, 52, 100, 129, 100, 158, 123, 208, 138, 119, 144, 48, 137, 25, 139, 69, 152, 182, 106, 16, 101, 17, 101, 207, 106, 61, 124, 110, 157, 69, 91, 19, 101, 43, 154, 29, 155, 28, 155, 100, 124, 198, 151, 60, 0, 42, 0, 241, 78, 146, 92, 244, 95, 114, 98, 209, 98, 77, 82, 164, 148, 108, 107, 84, 134, 177, 148, 179, 148, 66, 80, 174, 99, 245, 99, 225, 142, 138, 90, 16, 146, 108, 151, 87, 146, 152, 88, 169, 105, 157, 123, 173, 146, 91, 111, 92, 111, 172, 127, 65, 133, 108, 106, 34, 147, 212, 158, 218, 158, 29, 154, 243, 111, 26, 154, 74, 112, 44, 156, 60, 0, 42, 0, 245, 81, 69, 109, 183, 128, 250, 109, 37, 129, 219, 85, 112, 93, 99, 144, 207, 105, 129, 129, 56, 135, 52, 140, 49, 127, 126, 126, 116, 139, 83, 148, 60, 0, 42, 0, 32, 107, 11, 82, 161, 130, 212, 79, 28, 131, 41, 80, 147, 96, 17, 88, 148, 80, 76, 93, 200, 104, 32, 105, 74, 97, 152, 118, 168, 132, 121, 88, 73, 107, 170, 125, 51, 133, 25, 81, 231, 105, 207, 123, 36, 143, 223, 123, 205, 88, 49, 91, 52, 126, 28, 156, 60, 0, 42, 0, 134, 147, 242, 147, 86, 124, 60, 0, 42, 0, 91, 84, 140, 127, 21, 98, 23, 98, 168, 101, 170, 103, 177, 115, 151, 127, 16, 115, 196, 141, 244, 85, 12, 105, 172, 110, 71, 115, 84, 129, 198, 85, 35, 135, 22, 149, 136, 93, 39, 98, 205, 105, 68, 114, 114, 116, 171, 127, 53, 149, 236, 123, 6, 147, 18, 139, 76, 142, 106, 149, 97, 142, 151, 147, 216, 147, 60, 0, 42, 0, 44, 78, 55, 95, 58, 95, 153, 88, 241, 90, 55, 133, 47, 106, 18, 111, 3, 133, 187, 88, 25, 91, 231, 94, 148, 133, 163, 106, 70, 114, 98, 130, 32, 134, 60, 0, 42, 0, 162, 98, 159, 127, 54, 100, 165, 127, 143, 88, 72, 126, 65, 137, 102, 126, 249, 147, 60, 0, 42, 0, 157, 112, 52, 85, 151, 113, 187, 127, 60, 0, 42, 0, 132, 96, 87, 120, 251, 144, 106, 93, 247, 141, 33, 145, 37, 145, 129, 82, 114, 101, 195, 107, 13, 142, 57, 149, 157, 88, 29, 152, 185, 154, 189, 88, 103, 94, 71, 106, 198, 113, 50, 127, 253, 120, 107, 147, 108, 147, 81, 126, 172, 141, 122, 142, 48, 148, 60, 0, 42, 0, 84, 78, 168, 79, 77, 131, 94, 131, 101, 104, 90, 120, 236, 131, 172, 85, 209, 80, 215, 105, 47, 140, 58, 86, 246, 90, 148, 97, 78, 133, 146, 151, 53, 106, 75, 106, 158, 114, 68, 118, 167, 119, 4, 121, 238, 133, 171, 141, 8, 148, 189, 151, 102, 152, 60, 0, 42, 0, 231, 93, 229, 145, 0, 97, 220, 154, 60, 0, 42, 0, 207, 79, 238, 139, 87, 150, 237, 92, 41, 94, 141, 122, 187, 107, 216, 127, 154, 138, 218, 154, 250, 80, 172, 100, 189, 100, 152, 151, 210, 151, 197, 122, 249, 127, 169, 151, 89, 139, 136, 142, 60, 0, 42, 0, 198, 130, 60, 0, 42, 0, 127, 118, 122, 128, 60, 0, 42, 0, 20, 78, 60, 0, 42, 0, 7, 82, 146, 83, 190, 89, 47, 96, 196, 144, 39, 83, 131, 122, 143, 96, 8, 99, 47, 109, 236, 96, 193, 109, 33, 123, 28, 97, 234, 134, 5, 103, 167, 123, 193, 125, 50, 149, 203, 123, 37, 142, 85, 122, 210, 133, 101, 147, 220, 155, 17, 148, 202, 122, 97, 124, 60, 0, 42, 0, 178, 78, 181, 79, 166, 148, 126, 136, 142, 154, 135, 90, 90, 93, 61, 107, 133, 125, 155, 138, 148, 93, 170, 137, 73, 152, 248, 153, 188, 155, 244, 91, 60, 0, 42, 0, 136, 94, 169, 130, 185, 130, 193, 87, 225, 115, 230, 121, 57, 128, 230, 131, 153, 134, 102, 99, 243, 131, 52, 116, 57, 116, 189, 121, 25, 146, 43, 146, 194, 150, 228, 82, 234, 85, 192, 90, 177, 110, 114, 151, 108, 97, 89, 86, 210, 100, 179, 101, 249, 156, 196, 97, 142, 106, 191, 111, 61, 118, 147, 135, 195, 97, 4, 136, 53, 155, 109, 157, 60, 0, 42, 0, 69, 87, 17, 102, 9, 123, 171, 104, 126, 141, 209, 91, 19, 149, 221, 91, 226, 91, 223, 146, 188, 135, 60, 0, 42, 0, 34, 84, 35, 84, 139, 98, 129, 108, 26, 85, 227, 131, 255, 99, 7, 100, 179, 100, 25, 112, 253, 133, 60, 0, 42, 0, 81, 151, 82, 151, 34, 108, 123, 143, 62, 80, 127, 83, 236, 144, 10, 87, 229, 87, 200, 91, 43, 108, 248, 109, 5, 110, 190, 80, 59, 135, 21, 143, 173, 156, 75, 148, 60, 0, 42, 0, 29, 89, 32, 117, 96, 82, 205, 82, 197, 96, 145, 107, 116, 102, 254, 104, 48, 108, 93, 132, 146, 102, 207, 100, 8, 106, 206, 100, 160, 106, 229, 158, 60, 0, 42, 0, 216, 130, 119, 152, 247, 139, 188, 94, 3, 152, 206, 94, 0, 111, 203, 138, 190, 106, 60, 0, 42, 0, 134, 94, 202, 81, 133, 99, 184, 107, 131, 120, 144, 123, 88, 151, 118, 97, 236, 120, 44, 81, 234, 111, 68, 127, 38, 139, 230, 106, 60, 0, 42, 0, 88, 120, 60, 0, 42, 0, 142, 130, 60, 0, 42, 0, 109, 83, 155, 144, 134, 91, 119, 122, 121, 122, 21, 131, 79, 104, 59, 123, 71, 123, 121, 141, 248, 96, 42, 113, 45, 113, 60, 116, 60, 130, 233, 134, 236, 134, 98, 113, 141, 113, 88, 119, 235, 141, 142, 146, 143, 119, 174, 122, 29, 81, 140, 97, 105, 106, 154, 116, 209, 133, 202, 116, 198, 122, 237, 133, 215, 116, 60, 0, 42, 0, 24, 78, 32, 78, 177, 144, 117, 87, 88, 96, 203, 121, 204, 121, 175, 134, 157, 90, 41, 132, 120, 105, 242, 132, 89, 158, 205, 123, 231, 125, 117, 135, 80, 122, 165, 141, 197, 156, 215, 135, 166, 151, 167, 151, 18, 134, 12, 156, 13, 156, 150, 157, 36, 136, 157, 159, 60, 0, 42, 0, 244, 83, 218, 86, 79, 98, 176, 114, 140, 115, 83, 108, 141, 128, 66, 108, 108, 134, 197, 108, 111, 134, 197, 79, 211, 137, 4, 138, 5, 138, 75, 145, 211, 145, 18, 85, 87, 109, 12, 125, 141, 131, 14, 144, 17, 144, 218, 145, 130, 104, 143, 107, 236, 107, 3, 116, 71, 141, 251, 145, 55, 93, 239, 93, 30, 110, 109, 110, 179, 118, 218, 118, 82, 144, 106, 113, 127, 125, 247, 134, 216, 136, 240, 93, 233, 137, 213, 140, 134, 116, 100, 135, 182, 146, 148, 145, 130, 155, 61, 159, 196, 155, 61, 156, 60, 0, 42, 0, 29, 100, 215, 124, 60, 0, 42, 0, 58, 83, 242, 102, 57, 79, 73, 79, 36, 83, 150, 92, 206, 139, 57, 150, 113, 154, 101, 87, 72, 92, 168, 92, 180, 92, 190, 98, 64, 109, 91, 121, 224, 128, 170, 136, 64, 83, 54, 125, 198, 134, 175, 142, 65, 123, 172, 124, 208, 134, 88, 138, 139, 141, 135, 93, 136, 97, 198, 153, 122, 101, 179, 138, 124, 149, 200, 153, 185, 158, 247, 154, 124, 155, 168, 141, 175, 158, 176, 137, 192, 142, 180, 158, 226, 158, 187, 137, 69, 154, 56, 156, 75, 156, 60, 0, 42, 0, 98, 79, 172, 82, 170, 101, 16, 103, 202, 128, 195, 131, 80, 136, 50, 158, 237, 109, 32, 110, 71, 125, 209, 127, 75, 132, 229, 142, 86, 133, 150, 116, 242, 120, 182, 135, 29, 157, 169, 116, 221, 135, 191, 119, 41, 159, 39, 134, 194, 95, 72, 112, 53, 98, 11, 107, 13, 108, 103, 124, 222, 129, 111, 118, 55, 136, 98, 136, 163, 142, 60, 136, 122, 148, 28, 158, 60, 0, 42, 0, 214, 83, 216, 122, 54, 90, 83, 138, 236, 122, 122, 135, 139, 159, 114, 159, 60, 0, 42, 0, 186, 83, 187, 83, 30, 82, 127, 84, 31, 85, 29, 128, 18, 150, 209, 137, 163, 141, 180, 149, 174, 158, 195, 149, 183, 137, 1, 159, 60, 0, 42, 0, 242, 143, 60, 0, 42, 0, 77, 89, 209, 92, 46, 95, 110, 96, 155, 96, 8, 87, 15, 87, 236, 104, 233, 153, 9, 148, 60, 0, 42, 0, 104, 81, 67, 103, 122, 79, 224, 139, 254, 89, 201, 108, 36, 109, 67, 131, 243, 98, 119, 114, 129, 143, 115, 85, 226, 87, 88, 90, 211, 96, 202, 117, 66, 120, 232, 148, 118, 110, 136, 114, 76, 123, 95, 125, 114, 132, 60, 100, 126, 105, 84, 116, 224, 137, 110, 138, 231, 141, 7, 143, 55, 135, 147, 146, 41, 106, 33, 142, 19, 126, 155, 145, 206, 149, 200, 156, 8, 155, 33, 154, 73, 91, 207, 93, 1, 156, 10, 107, 100, 159, 56, 136, 167, 152, 116, 152, 60, 0, 42, 0, 173, 114, 172, 114, 113, 108, 78, 117, 199, 112, 251, 126, 163, 125, 71, 134, 60, 0, 42, 0, 157, 82, 56, 82, 118, 114, 231, 82, 207, 151, 248, 82, 60, 0, 42, 0, 38, 105, 60, 0, 42, 0, 58, 127, 155, 132, 25, 150, 60, 0, 42, 0, 56, 118, 60, 0, 42, 0, 116, 83, 123, 83, 198, 87, 5, 93, 171, 96, 55, 116, 192, 150, 94, 120, 110, 120, 21, 150, 89, 88, 9, 100, 181, 118, 143, 120, 74, 158, 40, 97, 183, 105, 167, 88, 100, 97, 186, 120, 187, 120, 158, 141, 233, 113, 203, 149, 16, 121, 213, 149, 114, 157, 45, 121, 60, 0, 42, 0, 11, 89, 247, 86, 238, 92, 33, 144, 60, 0, 42, 0, 173, 91, 44, 94, 217, 136, 163, 127, 164, 127, 224, 136, 60, 0, 39, 0, 208, 253, 82, 0, 39, 0, 60, 0, 42, 0, 101, 84, 176, 128, 123, 136, 135, 136, 166, 134, 161, 136, 186, 134, 54, 113, 229, 154, 43, 86, 239, 154, 195, 113, 78, 126, 60, 0, 42, 0, 132, 81, 137, 81, 204, 89, 210, 130, 211, 103, 195, 115, 163, 90, 133, 132, 106, 106, 60, 0, 42, 0, 99, 122, 52, 81, 247, 82, 60, 112, 125, 115, 24, 134, 179, 121, 228, 116, 112, 122, 159, 142, 36, 155, 60, 0, 42, 0, 204, 88, 183, 86, 228, 88, 24, 101, 25, 114, 149, 126, 60, 0, 42, 0, 169, 139, 249, 97, 114, 139, 147, 139, 60, 0, 42, 0, 6, 90, 91, 131, 118, 153, 97, 104, 8, 91, 88, 133, 72, 106, 83, 137, 82, 153, 60, 0, 42, 0, 112, 98, 162, 150, 254, 100, 60, 0, 42, 0, 213, 126, 118, 144, 94, 126, 60, 0, 42, 0, 249, 96, 60, 0, 42, 0, 237, 112, 177, 113, 60, 0, 42, 0, 186, 78, 187, 78, 193, 78, 236, 88, 200, 95, 50, 103, 206, 95, 194, 121, 162, 130, 19, 146, 92, 155, 139, 146, 64, 157, 60, 0, 42, 0, 205, 95, 79, 131, 32, 104, 35, 104, 117, 131, 249, 121, 239, 104, 20, 122, 155, 125, 181, 142, 60, 0, 42, 0, 3, 82, 4, 82, 164, 139, 222, 78, 237, 78, 177, 139, 251, 78, 123, 92, 131, 92, 104, 98, 171, 126, 138, 89, 82, 103, 99, 114, 180, 126, 149, 128, 107, 143, 231, 151, 106, 153, 217, 89, 77, 121, 9, 125, 125, 136, 29, 125, 18, 138, 212, 142, 149, 104, 181, 136, 77, 125, 77, 129, 90, 132, 109, 151, 113, 151, 204, 151, 234, 152, 141, 138, 1, 153, 60, 0, 42, 0, 84, 98, 60, 0, 42, 0, 205, 78, 184, 143, 61, 121, 126, 150, 60, 0, 42, 0, 191, 130, 60, 0, 42, 0, 229, 101, 114, 154, 248, 86, 240, 145, 36, 146, 185, 153, 60, 0, 42, 0, 56, 131, 60, 0, 42, 0, 14, 98, 156, 128, 4, 104, 232, 114, 210, 126, 25, 131, 99, 131, 185, 91, 245, 92, 231, 107, 255, 112, 182, 90, 88, 93, 17, 100, 104, 125, 162, 127, 198, 90, 100, 93, 8, 100, 181, 105, 182, 110, 201, 132, 149, 105, 174, 105, 148, 113, 98, 116, 65, 122, 25, 126, 126, 135, 35, 137, 85, 149, 141, 135, 142, 135, 229, 153, 246, 154, 43, 91, 184, 93, 3, 114, 148, 147, 198, 93, 28, 112, 231, 102, 17, 136, 60, 0, 42, 0, 151, 81, 130, 91, 72, 87, 135, 80, 245, 142, 4, 108, 60, 0, 42, 0, 67, 122, 60, 0, 42, 0, 185, 83, 184, 121, 212, 103, 136, 124, 131, 90, 201, 99, 24, 110, 71, 132, 99, 113, 72, 116, 108, 129, 197, 124, 90, 135, 66, 142, 46, 143, 82, 147, 163, 151, 199, 116, 37, 154, 7, 156, 148, 157, 60, 0, 42, 0, 122, 105, 214, 151, 60, 0, 42, 0, 137, 128, 141, 91, 60, 0, 42, 0, 154, 144, 130, 89, 158, 79, 36, 94, 57, 131, 71, 104, 189, 136, 247, 148, 42, 110, 78, 123, 152, 132, 163, 146, 96, 133, 97, 135, 18, 81, 17, 157, 133, 86, 44, 91, 122, 91, 225, 111, 183, 133, 61, 157, 216, 102, 248, 113, 102, 137, 21, 136, 165, 152, 185, 145, 108, 152, 108, 156, 60, 0, 42, 0, 93, 108, 151, 128, 115, 78, 177, 143, 15, 145, 233, 100, 60, 0, 42, 0, 101, 81, 86, 98, 65, 103, 51, 109, 213, 85, 183, 90, 189, 110, 27, 127, 208, 132, 240, 156, 37, 137, 31, 126, 60, 0, 42, 0, 191, 93, 60, 0, 42, 0, 60, 99, 60, 0, 42, 0, 39, 88, 139, 100, 214, 88, 60, 0, 42, 0, 46, 150, 10, 103, 111, 143, 14, 128, 68, 80, 223, 142, 134, 90, 76, 116, 157, 120, 219, 125, 45, 143, 192, 116, 29, 121, 60, 0, 42, 0, 81, 90, 117, 104, 36, 117, 204, 125, 100, 133, 60, 0, 42, 0, 74, 133, 75, 133, 100, 106, 96, 126, 2, 134, 3, 134, 60, 0, 42, 0, 109, 108, 174, 130, 152, 103, 139, 134, 16, 149, 94, 116, 57, 135, 127, 119, 179, 146, 237, 146, 225, 83, 225, 88, 60, 0, 42, 0, 164, 119, 60, 0, 42, 0, 240, 149, 166, 109, 143, 149, 160, 149, 100, 111, 77, 106, 182, 129, 60, 0, 42, 0, 124, 99, 60, 0, 42, 0, 210, 83, 229, 130, 76, 80, 49, 95, 0, 145, 3, 110, 43, 113, 73, 105, 118, 93, 187, 132, 172, 123, 219, 123, 7, 114, 25, 156, 47, 156, 184, 157, 60, 0, 39, 0, 208, 253, 83, 0, 39, 0, 60, 0, 42, 0, 232, 78, 49, 99, 50, 99, 146, 100, 60, 0, 42, 0, 18, 109, 47, 138, 120, 151, 117, 111, 81, 112, 160, 142, 60, 0, 42, 0, 69, 83, 231, 108, 145, 148, 210, 152, 14, 129, 40, 132, 18, 146, 75, 100, 161, 150, 186, 153, 175, 152, 169, 133, 210, 106, 68, 134, 60, 0, 42, 0, 226, 107, 34, 97, 204, 99, 94, 88, 248, 107, 110, 129, 37, 86, 123, 86, 195, 156, 75, 152, 19, 156, 60, 0, 42, 0, 238, 85, 91, 141, 255, 80, 253, 140, 58, 124, 60, 0, 42, 0, 9, 78, 14, 95, 193, 83, 245, 107, 246, 107, 129, 83, 255, 107, 153, 114, 22, 155, 60, 0, 42, 0, 208, 78, 30, 79, 152, 80, 193, 124, 194, 124, 147, 153, 221, 124, 227, 124, 228, 124, 86, 126, 210, 147, 254, 147, 48, 151, 74, 153, 60, 0, 42, 0, 213, 79, 52, 94, 99, 101, 144, 149, 60, 0, 42, 0, 237, 88, 117, 106, 60, 0, 42, 0, 82, 104, 81, 104, 225, 105, 60, 0, 42, 0, 211, 85, 33, 100, 201, 120, 44, 137, 161, 152, 159, 147, 89, 152, 60, 0, 42, 0, 39, 78, 170, 85, 60, 0, 42, 0, 187, 99, 69, 97, 20, 100, 158, 110, 154, 154, 43, 127, 166, 135, 69, 126, 202, 129, 203, 156, 18, 154, 55, 154, 32, 156, 98, 156, 60, 0, 42, 0, 107, 98, 131, 99, 194, 90, 60, 0, 42, 0, 253, 87, 25, 118, 9, 108, 194, 119, 222, 154, 60, 0, 42, 0, 170, 149, 60, 0, 42, 0, 114, 130, 19, 109, 28, 104, 169, 109, 108, 85, 11, 110, 239, 148, 237, 150, 110, 107, 23, 116, 199, 85, 95, 116, 112, 107, 171, 146, 129, 111, 206, 97, 204, 100, 199, 111, 207, 111, 55, 118, 81, 122, 128, 111, 177, 116, 18, 112, 97, 122, 108, 126, 111, 122, 86, 143, 252, 147, 69, 139, 203, 152, 60, 0, 42, 0, 238, 104, 46, 105, 238, 105, 66, 137, 60, 0, 42, 0, 231, 80, 25, 155, 60, 0, 42, 0, 64, 103, 153, 108, 177, 126, 119, 78, 57, 82, 78, 82, 2, 120, 38, 85, 186, 107, 0, 115, 134, 124, 23, 125, 142, 131, 108, 104, 238, 107, 233, 148, 231, 117, 112, 120, 94, 113, 49, 133, 223, 136, 157, 105, 39, 106, 102, 155, 168, 156, 169, 147, 202, 155, 203, 155, 60, 0, 42, 0, 187, 80, 13, 81, 60, 0, 42, 0, 61, 80, 60, 85, 81, 85, 101, 85, 57, 94, 16, 132, 166, 83, 162, 85, 200, 94, 67, 107, 220, 127, 145, 123, 227, 127, 175, 149, 14, 151, 60, 0, 42, 0, 76, 126, 60, 0, 42, 0, 91, 123, 126, 145, 233, 123, 1, 124, 27, 124, 195, 145, 60, 0, 42, 0, 122, 126, 60, 0, 42, 0, 82, 102, 183, 149, 236, 102, 60, 0, 42, 0, 113, 92, 97, 95, 150, 144, 56, 87, 32, 82, 42, 82, 73, 103, 159, 130, 205, 89, 215, 89, 235, 130, 107, 136, 144, 148, 207, 87, 59, 99, 245, 103, 230, 114, 202, 115, 34, 130, 193, 117, 32, 129, 213, 142, 24, 123, 138, 149, 218, 141, 124, 82, 39, 100, 7, 86, 83, 94, 125, 113, 120, 111, 152, 111, 134, 106, 63, 126, 187, 129, 197, 155, 180, 127, 182, 127, 60, 0, 42, 0, 234, 149, 85, 150, 182, 112, 93, 150, 131, 149, 113, 102, 84, 113, 82, 119, 140, 113, 162, 137, 60, 0, 42, 0, 170, 139, 85, 108, 157, 117, 97, 82, 71, 98, 21, 138, 120, 141, 158, 99, 228, 145, 147, 80, 132, 85, 43, 105, 143, 146, 159, 154, 208, 80, 47, 145, 160, 88, 161, 88, 108, 111, 46, 127, 23, 91, 197, 100, 126, 101, 63, 106, 90, 107, 179, 129, 240, 120, 6, 139, 97, 141, 85, 126, 238, 135, 250, 135, 113, 139, 13, 141, 37, 148, 77, 153, 56, 154, 221, 156, 87, 112, 83, 156, 84, 156, 60, 0, 42, 0, 36, 79, 135, 107, 70, 85, 222, 137, 183, 80, 146, 88, 111, 97, 243, 110, 33, 111, 15, 133, 164, 107, 181, 113, 170, 135, 244, 137, 42, 139, 58, 155, 60, 0, 42, 0, 167, 87, 68, 98, 76, 102, 79, 141, 222, 140, 24, 141, 92, 148, 60, 0, 42, 0, 4, 78, 10, 78, 25, 92, 26, 92, 102, 96, 241, 126, 212, 125, 157, 151, 60, 0, 42, 0, 243, 136, 60, 0, 42, 0, 48, 95, 78, 99, 231, 112, 166, 131, 162, 104, 60, 113, 13, 122, 211, 101, 114, 123, 68, 130, 248, 134, 14, 143, 113, 133, 210, 113, 181, 152, 254, 154, 185, 155, 60, 0, 42, 0, 250, 82, 141, 130, 213, 130, 214, 103, 191, 115, 246, 151, 60, 0, 42, 0, 17, 92, 60, 0, 42, 0, 173, 82, 114, 83, 181, 144, 205, 126, 232, 84, 11, 90, 145, 136, 57, 125, 68, 119, 164, 125, 114, 111, 60, 0, 42, 0, 98, 89, 30, 115, 74, 141, 108, 117, 114, 117, 11, 143, 210, 140, 214, 140, 168, 106, 60, 0, 42, 0, 12, 130, 88, 79, 117, 134, 199, 134, 229, 134, 60, 0, 42, 0, 13, 130, 104, 99, 60, 0, 42, 0, 141, 83, 190, 139, 62, 121, 14, 130, 153, 83, 4, 92, 137, 109, 187, 109, 9, 110, 45, 138, 102, 141, 61, 95, 81, 97, 66, 100, 68, 100, 224, 110, 116, 97, 117, 100, 14, 133, 89, 107, 2, 136, 216, 151, 7, 154, 254, 97, 29, 101, 68, 112, 157, 158, 7, 107, 60, 0, 42, 0, 1, 140, 60, 0, 42, 0, 51, 117, 126, 92, 95, 98, 56, 79, 171, 142, 129, 79, 123, 84, 189, 89, 118, 124, 197, 126, 89, 127, 220, 139, 250, 89, 219, 103, 32, 108, 197, 115, 124, 122, 120, 124, 32, 90, 247, 92, 33, 117, 18, 119, 55, 120, 152, 131, 20, 88, 82, 101, 241, 109, 51, 125, 95, 81, 253, 104, 96, 132, 209, 136, 55, 138, 225, 132, 117, 138, 39, 117, 24, 133, 202, 113, 147, 133, 234, 153, 185, 156, 209, 102, 211, 155, 98, 157, 245, 155, 58, 156, 60, 0, 42, 0, 192, 78, 26, 117, 94, 121, 138, 105, 48, 156, 60, 0, 42, 0, 165, 144, 30, 95, 161, 91, 228, 119, 194, 84, 231, 119, 183, 91, 2, 140, 9, 140, 118, 90, 22, 110, 32, 138, 233, 91, 215, 138, 35, 152, 107, 155, 203, 102, 171, 119, 56, 91, 11, 112, 190, 137, 133, 139, 60, 0, 42, 0, 190, 128, 186, 79, 26, 102, 194, 128, 129, 109, 24, 119, 23, 110, 115, 121, 36, 129, 78, 129, 60, 97, 78, 97, 57, 105, 6, 118, 103, 127, 3, 135, 4, 135, 242, 110, 224, 146, 46, 118, 60, 0, 42, 0, 71, 83, 31, 117, 41, 150, 79, 84, 240, 88, 152, 101, 7, 102, 161, 103, 233, 108, 204, 114, 252, 130, 13, 104, 133, 107, 114, 114, 196, 115, 212, 122, 94, 150, 251, 102, 121, 150, 25, 123, 102, 110, 58, 113, 37, 117, 78, 146, 114, 128, 95, 147, 42, 159, 127, 157, 60, 0, 42, 0, 243, 126, 4, 126, 180, 97, 105, 126, 93, 139, 60, 0, 42, 0, 1, 119, 26, 119, 87, 80, 59, 110, 60, 0, 42, 0, 35, 87, 220, 128, 96, 102, 112, 82, 219, 118, 105, 82, 221, 82, 185, 140, 74, 93, 30, 116, 86, 128, 173, 88, 186, 105, 66, 133, 115, 106, 248, 140, 60, 0, 42, 0, 56, 92, 49, 89, 8, 94, 94, 84, 113, 134, 215, 139, 191, 144, 36, 158, 77, 92, 189, 101, 73, 109, 238, 114, 43, 94, 65, 125, 246, 145, 100, 110, 127, 110, 121, 132, 174, 110, 188, 110, 69, 115, 146, 132, 205, 132, 105, 138, 71, 146, 72, 146, 97, 116, 242, 156, 104, 135, 254, 156, 55, 137, 186, 156, 213, 111, 102, 147, 244, 155, 36, 156, 179, 157, 121, 137, 60, 0, 42, 0, 65, 83, 99, 153, 243, 119, 187, 143, 109, 78, 246, 101, 205, 122, 198, 139, 158, 91, 159, 91, 249, 101, 224, 152, 252, 89, 213, 92, 254, 98, 187, 112, 79, 121, 128, 134, 223, 152, 216, 87, 66, 102, 179, 131, 212, 91, 92, 110, 72, 144, 82, 88, 117, 93, 161, 110, 148, 132, 80, 146, 230, 91, 175, 105, 85, 135, 165, 156, 150, 155, 43, 159, 88, 139, 45, 159, 35, 156, 60, 0, 42, 0, 242, 83, 226, 119, 104, 78, 85, 140, 127, 79, 203, 89, 118, 154, 88, 81, 169, 91, 78, 92, 54, 123, 129, 105, 66, 146, 219, 153, 60, 0, 42, 0, 235, 88, 15, 108, 59, 121, 23, 78, 22, 78, 213, 78, 2, 94, 58, 121, 60, 79, 75, 83, 15, 95, 213, 95, 138, 78, 211, 83, 58, 98, 139, 78, 141, 79, 191, 82, 105, 84, 249, 103, 198, 137, 213, 139, 112, 153, 159, 81, 164, 91, 64, 96, 67, 96, 237, 98, 47, 102, 48, 102, 190, 103, 255, 103, 2, 119, 51, 141, 2, 144, 59, 104, 210, 112, 14, 119, 33, 119, 16, 130, 124, 143, 29, 144, 200, 148, 150, 137, 73, 140, 200, 145, 158, 90, 60, 93, 17, 95, 165, 95, 211, 99, 37, 140, 176, 140, 202, 145, 226, 82, 220, 85, 18, 95, 118, 113, 87, 119, 110, 123, 226, 137, 102, 138, 254, 142, 48, 146, 67, 146, 254, 152, 19, 130, 147, 138, 105, 144, 125, 146, 109, 89, 180, 146, 25, 153, 29, 153, 108, 86, 21, 91, 168, 111, 223, 138, 225, 138, 126, 144, 171, 135, 26, 139, 45, 124, 107, 137, 203, 145, 24, 156, 60, 0, 42, 0, 102, 79, 25, 83, 210, 123, 66, 124, 60, 0, 42, 0, 206, 83, 54, 101, 60, 0, 42, 0, 75, 98, 136, 91, 168, 87, 150, 153, 79, 130, 60, 0, 42, 0, 255, 91, 215, 83, 233, 114, 125, 81, 46, 85, 136, 99, 173, 109, 246, 126, 233, 117, 132, 129, 253, 88, 0, 89, 38, 118, 172, 125, 99, 115, 120, 115, 201, 147, 60, 0, 42, 0, 76, 98, 60, 0, 42, 0, 102, 78, 179, 107, 23, 92, 146, 98, 190, 126, 212, 83, 120, 103, 162, 103, 78, 150, 221, 89, 15, 80, 16, 80, 248, 102, 138, 107, 19, 125, 147, 99, 179, 104, 209, 109, 2, 113, 253, 131, 215, 142, 3, 145, 142, 117, 143, 117, 18, 130, 69, 100, 249, 107, 250, 107, 128, 125, 147, 143, 121, 116, 254, 141, 8, 142, 30, 106, 44, 133, 56, 143, 126, 106, 155, 155, 53, 81, 4, 101, 45, 112, 104, 157, 60, 0, 42, 0, 235, 121, 76, 90, 112, 91, 78, 141, 126, 88, 159, 113, 185, 116, 22, 141, 60, 0, 42, 0, 33, 159, 94, 92, 145, 102, 143, 102, 205, 158, 114, 127, 0, 135, 32, 159, 123, 111, 165, 133, 175, 133, 217, 102, 89, 118, 247, 133, 97, 137, 236, 124, 105, 137, 108, 92, 52, 136, 106, 156, 112, 156, 0, 158, 60, 0, 42, 0, 46, 103, 47, 103, 13, 98, 95, 103, 173, 108, 240, 143, 184, 79, 176, 84, 55, 96, 17, 104, 214, 122, 87, 131, 85, 96, 82, 99, 182, 94, 187, 94, 73, 125, 129, 132, 83, 136, 131, 150, 12, 92, 203, 136, 112, 101, 234, 122, 103, 129, 101, 146, 133, 88, 49, 111, 68, 111, 120, 101, 141, 111, 78, 140, 57, 106, 214, 111, 48, 147, 227, 147, 144, 157, 106, 134, 60, 0, 42, 0, 55, 82, 48, 85, 60, 0, 42, 0, 13, 128, 60, 0, 42, 0, 156, 138, 60, 0, 42, 0, 112, 136, 84, 100, 60, 0, 42, 0, 41, 117, 60, 0, 42, 0, 5, 94, 37, 94, 192, 135, 91, 83, 60, 0, 42, 0, 233, 149, 244, 98, 130, 149, 19, 104, 60, 0, 42, 0, 174, 109, 104, 129, 60, 0, 42, 0, 204, 83, 28, 151, 217, 150, 64, 91, 166, 154, 71, 91, 59, 154, 6, 107, 53, 121, 222, 157, 116, 158, 109, 130, 102, 154, 24, 158, 60, 0, 42, 0, 61, 114, 125, 88, 97, 97, 58, 111, 9, 106, 20, 126, 239, 147, 60, 0, 42, 0, 64, 112, 60, 0, 42, 0, 61, 129, 176, 138, 60, 0, 42, 0, 52, 108, 58, 108, 60, 0, 42, 0, 40, 94, 151, 109, 154, 109, 113, 121, 5, 122, 14, 122, 222, 136, 97, 119, 147, 119, 60, 0, 42, 0, 53, 108, 150, 149, 60, 0, 42, 0, 46, 84, 60, 0, 42, 0, 122, 152, 28, 130, 6, 152, 99, 133, 83, 106, 154, 119, 172, 119, 10, 155, 60, 0, 42, 0, 244, 139, 254, 84, 170, 138, 172, 138, 60, 0, 42, 0, 129, 89, 193, 112, 20, 103, 196, 148, 54, 107, 85, 120, 223, 119, 32, 100, 180, 132, 202, 105, 97, 115, 169, 120, 190, 123, 153, 147, 13, 114, 96, 148, 60, 0, 42, 0, 182, 83, 159, 126, 29, 78, 248, 83, 249, 124, 193, 121, 157, 84, 192, 108, 236, 79, 29, 96, 82, 134, 86, 96, 54, 158, 164, 90, 175, 101, 114, 125, 12, 127, 243, 134, 82, 105, 151, 121, 112, 146, 212, 152, 216, 81, 174, 83, 185, 105, 160, 121, 115, 127, 36, 135, 175, 146, 54, 149, 54, 86, 93, 86, 221, 94, 149, 100, 140, 111, 195, 120, 230, 125, 108, 133, 214, 146, 205, 113, 132, 135, 214, 135, 244, 135, 184, 152, 38, 154, 1, 148, 229, 157, 54, 159, 109, 124, 60, 0, 42, 0, 123, 107, 60, 0, 42, 0, 243, 93, 150, 78, 219, 86, 250, 91, 92, 108, 64, 79, 85, 81, 210, 89, 228, 108, 64, 121, 161, 79, 96, 91, 107, 103, 215, 108, 114, 153, 119, 154, 48, 90, 246, 103, 109, 114, 13, 109, 152, 109, 130, 128, 228, 152, 37, 123, 28, 128, 242, 145, 226, 122, 151, 137, 227, 85, 134, 128, 132, 140, 54, 146, 59, 146, 252, 152, 169, 121, 223, 153, 124, 133, 41, 81, 3, 112, 60, 0, 42, 0, 234, 95, 126, 103, 128, 103, 169, 103, 0, 90, 215, 103, 47, 80, 199, 81, 39, 93, 186, 94, 165, 104, 222, 109, 216, 131, 105, 93, 121, 120, 25, 135, 189, 97, 205, 111, 167, 106, 118, 147, 6, 155, 60, 0, 42, 0, 2, 96, 154, 96, 56, 128, 230, 122, 177, 80, 47, 97, 100, 105, 119, 93, 107, 97, 115, 128, 247, 153, 60, 0, 42, 0, 188, 139, 139, 91, 245, 139, 1, 144, 130, 152, 31, 138, 12, 152, 166, 138, 185, 147, 56, 153, 60, 0, 42, 0, 193, 81, 92, 99, 11, 145, 214, 85, 192, 94, 203, 94, 28, 100, 178, 110, 64, 115, 144, 132, 195, 132, 138, 153, 73, 100, 213, 152, 87, 100, 60, 149, 88, 130, 139, 135, 153, 145, 170, 147, 63, 153, 188, 152, 190, 152, 42, 154, 60, 0, 42, 0, 220, 83, 223, 83, 129, 80, 254, 85, 141, 119, 222, 100, 174, 133, 251, 100, 234, 133, 226, 106, 84, 124, 60, 0, 42, 0, 253, 85, 54, 118, 60, 0, 42, 0, 207, 130, 38, 117, 101, 145, 35, 122, 163, 122, 76, 122, 194, 155, 7, 134, 19, 134, 239, 106, 204, 86, 60, 0, 42, 0, 215, 79, 60, 0, 42, 0, 138, 115, 25, 89, 201, 139, 221, 108, 131, 128, 44, 109, 145, 109, 223, 115, 32, 125, 164, 131, 31, 144, 191, 91, 128, 104, 144, 107, 155, 124, 149, 154, 131, 80, 159, 124, 52, 138, 33, 140, 201, 85, 80, 88, 81, 88, 202, 90, 43, 97, 175, 110, 184, 110, 133, 128, 97, 144, 84, 158, 243, 80, 44, 97, 161, 105, 134, 129, 12, 133, 235, 137, 154, 141, 108, 144, 159, 97, 14, 106, 21, 106, 101, 111, 191, 120, 201, 146, 23, 153, 90, 111, 36, 126, 90, 106, 155, 116, 12, 124, 215, 133, 22, 139, 92, 142, 76, 154, 80, 156, 235, 157, 60, 0, 42, 0, 251, 114, 224, 117, 120, 145, 60, 0, 42, 0, 52, 83, 60, 0, 42, 0, 88, 121, 7, 123, 109, 123, 156, 132, 151, 123, 60, 0, 42, 0, 10, 89, 149, 130, 125, 134, 32, 80, 248, 84, 125, 109, 125, 131, 126, 131, 45, 119, 112, 132, 214, 110, 98, 119, 143, 125, 163, 113, 201, 111, 150, 151, 214, 150, 60, 0, 42, 0, 229, 126, 139, 150, 143, 150, 64, 144, 168, 150, 205, 116, 60, 0, 42, 0, 33, 112, 184, 129, 196, 154, 211, 154, 60, 0, 42, 0, 151, 78, 129, 92, 21, 120, 95, 121, 7, 140, 227, 87, 87, 93, 66, 144, 114, 107, 115, 107, 107, 113, 95, 119, 142, 120, 167, 150, 24, 91, 187, 111, 66, 122, 182, 138, 229, 140, 150, 106, 231, 113, 178, 116, 173, 121, 87, 122, 95, 122, 64, 126, 90, 137, 131, 144, 222, 101, 80, 126, 120, 126, 98, 139, 6, 148, 41, 148, 226, 151, 60, 0, 42, 0, 89, 91, 242, 114, 106, 131, 107, 91, 231, 152, 14, 100, 59, 115, 192, 132, 241, 152, 194, 105, 117, 133, 158, 133, 60, 0, 42, 0, 95, 99, 11, 123, 188, 150, 77, 123, 13, 100, 171, 105, 176, 123, 40, 124, 168, 147, 189, 157, 60, 0, 42, 0, 6, 85, 17, 90, 143, 131, 158, 80, 107, 104, 173, 104, 67, 119, 205, 85, 167, 127, 209, 132, 77, 100, 41, 127, 150, 141, 17, 124, 20, 124, 46, 126, 255, 154, 187, 155, 60, 0, 42, 0, 64, 98, 122, 78, 34, 85, 34, 125, 16, 116, 17, 116, 226, 96, 1, 149, 233, 85, 155, 102, 145, 110, 99, 116, 40, 137, 133, 116, 136, 147, 141, 147, 150, 147, 187, 147, 193, 147, 60, 0, 42, 0, 36, 144, 185, 110, 54, 135, 60, 0, 42, 0, 230, 85, 60, 0, 39, 0, 208, 253, 84, 0, 39, 0, 60, 0, 42, 0, 214, 78, 131, 91, 121, 89, 96, 114, 66, 121, 191, 141, 202, 148, 76, 88, 153, 105, 187, 110, 31, 137, 131, 86, 231, 149, 60, 0, 42, 0, 121, 142, 60, 0, 42, 0, 84, 88, 154, 110, 150, 88, 109, 115, 206, 156, 122, 115, 40, 156, 60, 0, 42, 0, 211, 98, 30, 99, 231, 114, 252, 149, 9, 93, 190, 109, 40, 100, 93, 144, 98, 144, 187, 105, 254, 107, 162, 121, 187, 100, 190, 111, 187, 138, 15, 142, 125, 106, 20, 147, 204, 111, 75, 142, 156, 151, 153, 155, 210, 149, 179, 151, 186, 86, 229, 149, 118, 139, 162, 142, 60, 0, 42, 0, 164, 79, 156, 84, 60, 0, 42, 0, 252, 86, 97, 91, 206, 128, 60, 0, 42, 0, 240, 83, 242, 101, 176, 144, 110, 87, 172, 98, 212, 130, 177, 103, 177, 112, 178, 112, 237, 131, 198, 141, 144, 156, 136, 123, 250, 129, 177, 152, 216, 153, 19, 81, 144, 155, 47, 91, 225, 100, 185, 133, 175, 106, 73, 124, 60, 0, 42, 0, 42, 89, 173, 81, 51, 89, 242, 95, 112, 108, 1, 96, 189, 128, 155, 148, 240, 108, 38, 130, 94, 145, 38, 146, 153, 110, 75, 97, 228, 113, 60, 0, 42, 0, 143, 124, 60, 0, 42, 0, 77, 87, 42, 141, 57, 96, 209, 117, 17, 130, 170, 140, 74, 100, 233, 110, 43, 118, 249, 100, 36, 101, 88, 112, 113, 118, 60, 0, 42, 0, 91, 87, 25, 102, 19, 80, 8, 140, 239, 144, 82, 90, 212, 96, 131, 137, 131, 105, 240, 117, 44, 149, 45, 140, 176, 88, 181, 88, 155, 97, 109, 111, 199, 138, 136, 145, 199, 88, 199, 102, 194, 113, 31, 147, 36, 153, 128, 106, 249, 120, 67, 152, 72, 127, 235, 133, 220, 88, 90, 139, 154, 140, 176, 145, 96, 139, 78, 127, 60, 0, 42, 0, 208, 95, 102, 87, 146, 136, 189, 148, 252, 131, 239, 107, 109, 146, 255, 85, 179, 97, 187, 97, 186, 102, 147, 145, 174, 116, 98, 137, 60, 0, 42, 0, 249, 83, 173, 112, 238, 87, 162, 99, 157, 80, 96, 110, 203, 80, 6, 86, 179, 120, 21, 130, 78, 107, 231, 140, 60, 0, 42, 0, 100, 108, 244, 148, 111, 110, 33, 86, 37, 128, 143, 82, 176, 127, 106, 135, 154, 133, 87, 149, 90, 142, 220, 147, 11, 148, 186, 151, 30, 159, 60, 0, 42, 0, 103, 153, 99, 87, 16, 85, 2, 88, 143, 80, 122, 85, 224, 104, 12, 145, 88, 88, 42, 100, 143, 110, 206, 132, 154, 150, 182, 105, 31, 111, 123, 113, 109, 116, 159, 121, 133, 129, 24, 106, 196, 120, 195, 124, 155, 129, 86, 106, 214, 123, 214, 124, 151, 135, 60, 142, 219, 124, 179, 135, 111, 141, 163, 145, 51, 153, 149, 147, 57, 153, 219, 149, 68, 153, 182, 157, 60, 0, 42, 0, 22, 79, 17, 94, 24, 80, 82, 80, 204, 109, 165, 80, 186, 142, 75, 149, 178, 147, 59, 81, 3, 98, 237, 102, 35, 114, 216, 119, 130, 148, 60, 0, 42, 0, 235, 112, 101, 100, 159, 141, 217, 113, 60, 0, 42, 0, 50, 89, 34, 95, 155, 109, 230, 126, 143, 99, 91, 125, 92, 138, 205, 90, 77, 94, 70, 97, 47, 100, 212, 110, 196, 105, 107, 116, 236, 151, 248, 152, 26, 126, 39, 126, 228, 111, 31, 139, 65, 143, 177, 151, 220, 151, 85, 153, 60, 0, 42, 0, 11, 83, 239, 143, 183, 84, 46, 109, 3, 144, 67, 104, 118, 150, 85, 85, 188, 104, 216, 109, 249, 126, 4, 132, 121, 121, 234, 136, 175, 125, 42, 135, 128, 151, 132, 145, 137, 151, 254, 146, 45, 147, 227, 153, 174, 106, 64, 153, 10, 154, 23, 159, 60, 0, 42, 0, 168, 139, 14, 138, 60, 0, 42, 0, 87, 89, 60, 0, 42, 0, 209, 95, 210, 95, 121, 114, 163, 140, 253, 148, 93, 97, 241, 146, 163, 135, 216, 135, 60, 0, 42, 0, 165, 113, 175, 129, 31, 159, 60, 0, 42, 0, 188, 117, 203, 117, 80, 94, 126, 129, 138, 138, 27, 111, 213, 110, 134, 144, 34, 126, 230, 153, 4, 139, 47, 81, 228, 133, 48, 154, 80, 124, 39, 156, 88, 124, 69, 134, 99, 154, 60, 0, 42, 0, 47, 151, 60, 0, 42, 0, 84, 82, 175, 104, 17, 149, 34, 142, 255, 100, 200, 157, 201, 157, 60, 0, 42, 0, 208, 130, 151, 83, 81, 131, 232, 126, 77, 80, 124, 85, 57, 93, 255, 96, 208, 99, 10, 122, 7, 127, 100, 127, 70, 144, 72, 158, 193, 85, 69, 116, 136, 125, 174, 120, 6, 137, 178, 95, 61, 111, 249, 125, 91, 133, 109, 135, 187, 146, 152, 152, 167, 141, 68, 142, 141, 145, 21, 139, 79, 142, 87, 147, 192, 156, 58, 157, 76, 152, 183, 155, 92, 157, 32, 154, 247, 155, 151, 157, 153, 157, 181, 121, 228, 157, 60, 0, 42, 0, 83, 79, 46, 99, 176, 142, 181, 154, 167, 155, 198, 142, 212, 154, 60, 0, 42, 0, 59, 98, 73, 92, 67, 82, 31, 109, 17, 103, 28, 80, 140, 96, 149, 109, 22, 144, 92, 92, 144, 96, 213, 96, 166, 99, 55, 144, 214, 96, 229, 99, 255, 102, 116, 105, 252, 136, 5, 137, 82, 107, 162, 107, 240, 154, 153, 133, 143, 86, 0, 155, 148, 86, 203, 116, 74, 124, 175, 141, 60, 0, 42, 0, 57, 123, 60, 0, 42, 0, 41, 89, 114, 81, 86, 90, 251, 109, 95, 145, 84, 151, 199, 158, 93, 151, 60, 0, 42, 0, 48, 117, 71, 92, 186, 108, 108, 96, 75, 117, 81, 117, 247, 118, 203, 128, 96, 117, 27, 117, 28, 117, 254, 131, 73, 110, 97, 88, 107, 88, 55, 100, 63, 146, 23, 150, 181, 120, 194, 125, 204, 120, 180, 122, 43, 157, 179, 116, 208, 149, 198, 157, 207, 157, 60, 0, 42, 0, 221, 95, 132, 107, 14, 80, 58, 85, 191, 96, 223, 109, 106, 102, 32, 116, 70, 129, 205, 137, 246, 117, 83, 119, 20, 130, 2, 153, 165, 137, 223, 140, 42, 147, 105, 147, 102, 151, 60, 0, 42, 0, 173, 99, 124, 119, 26, 130, 60, 0, 42, 0, 235, 101, 123, 79, 163, 94, 76, 96, 17, 99, 103, 121, 78, 128, 60, 0, 42, 0, 128, 130, 97, 103, 167, 92, 185, 92, 226, 143, 82, 121, 157, 104, 36, 123, 20, 132, 218, 132, 232, 132, 146, 141, 134, 159, 36, 106, 41, 135, 218, 146, 151, 151, 235, 154, 166, 156, 200, 155, 165, 147, 96, 159, 55, 156, 60, 0, 42, 0, 168, 91, 64, 102, 19, 103, 1, 129, 149, 122, 130, 138, 162, 101, 177, 122, 37, 91, 60, 0, 42, 0, 58, 119, 156, 124, 105, 125, 156, 137, 243, 141, 246, 124, 60, 0, 42, 0, 169, 135, 60, 0, 42, 0, 22, 94, 23, 96, 52, 141, 28, 132, 81, 128, 188, 140, 60, 0, 42, 0, 193, 148, 200, 134, 227, 80, 149, 146, 232, 146, 41, 157, 33, 148, 53, 148, 86, 154, 60, 0, 42, 0, 107, 84, 251, 152, 46, 153, 60, 0, 42, 0, 133, 83, 129, 94, 64, 108, 124, 130, 44, 84, 58, 117, 19, 128, 155, 83, 195, 112, 111, 104, 244, 112, 142, 125, 147, 151, 116, 128, 124, 128, 240, 94, 125, 128, 243, 94, 60, 0, 42, 0, 146, 144, 247, 94, 173, 78, 173, 94, 155, 131, 92, 80, 119, 90, 73, 93, 31, 110, 115, 123, 118, 132, 19, 135, 95, 105, 179, 105, 174, 149, 6, 151, 100, 128, 79, 135, 234, 138, 46, 159, 60, 0, 42, 0, 34, 87, 60, 117, 185, 79, 23, 90, 58, 99, 143, 109, 131, 104, 246, 112, 253, 115, 33, 129, 71, 130, 139, 152, 148, 138, 50, 152, 60, 0, 42, 0, 242, 86, 181, 112, 26, 144, 204, 117, 245, 85, 234, 132, 11, 106, 60, 0, 42, 0, 221, 78, 12, 84, 95, 79, 100, 95, 194, 92, 157, 94, 195, 84, 221, 92, 234, 114, 60, 131, 77, 102, 80, 104, 117, 109, 212, 112, 60, 120, 146, 134, 46, 119, 241, 121, 220, 148, 229, 122, 161, 124, 105, 123, 119, 138, 104, 141, 110, 145, 86, 146, 238, 80, 237, 82, 117, 146, 133, 146, 7, 153, 150, 156, 124, 111, 94, 115, 200, 102, 35, 103, 102, 106, 3, 108, 209, 113, 157, 114, 167, 129, 179, 119, 166, 155, 60, 0, 42, 0, 223, 126, 69, 99, 118, 104, 82, 123, 113, 125, 130, 125, 60, 0, 42, 0, 120, 96, 219, 117, 85, 136, 95, 97, 133, 97, 60, 0, 42, 0, 119, 80, 120, 80, 126, 90, 174, 90, 192, 146, 110, 147, 60, 0, 42, 0, 160, 78, 52, 89, 149, 98, 176, 154, 240, 125, 45, 152, 60, 0, 42, 0, 181, 89, 173, 148, 15, 125, 104, 101, 243, 152, 200, 158, 35, 134, 60, 0, 42, 0, 15, 144, 137, 125, 60, 0, 42, 0, 248, 81, 138, 91, 191, 121, 195, 121, 34, 96, 129, 122, 11, 85, 139, 109, 120, 99, 23, 88, 101, 110, 220, 117, 86, 132, 128, 93, 245, 146, 90, 157, 53, 159, 60, 0, 42, 0, 243, 86, 254, 86, 195, 81, 249, 92, 169, 94, 146, 95, 135, 96, 72, 99, 130, 109, 124, 131, 20, 144, 96, 92, 140, 104, 223, 131, 236, 99, 12, 122, 21, 87, 87, 88, 94, 93, 15, 118, 97, 123, 111, 129, 164, 132, 47, 146, 22, 87, 23, 87, 220, 94, 115, 111, 255, 141, 116, 145, 159, 153, 78, 147, 252, 153, 76, 157, 159, 157, 203, 157, 245, 157, 60, 0, 42, 0, 31, 87, 33, 87, 16, 84, 141, 148, 247, 145, 60, 0, 42, 0, 78, 81, 204, 143, 84, 81, 181, 131, 13, 88, 117, 157, 60, 0, 42, 0, 98, 108, 60, 0, 42, 0, 77, 110, 47, 115, 83, 113, 146, 140, 60, 0, 42, 0, 226, 86, 227, 86, 159, 98, 120, 82, 24, 87, 113, 97, 118, 100, 25, 111, 235, 105, 255, 123, 178, 106, 196, 147, 240, 124, 210, 157, 251, 157, 60, 0, 42, 0, 131, 117, 60, 0, 42, 0, 86, 95, 106, 110, 22, 137, 60, 0, 42, 0, 168, 99, 247, 132, 236, 133, 60, 0, 42, 0, 26, 95, 147, 152, 164, 150, 53, 92, 57, 152, 58, 152, 61, 152, 75, 155, 104, 122, 8, 134, 106, 142, 60, 0, 42, 0, 192, 79, 127, 129, 211, 80, 70, 142, 189, 154, 60, 0, 42, 0, 187, 79, 0, 144, 39, 90, 122, 113, 251, 134, 21, 135, 42, 137, 254, 153, 60, 0, 42, 0, 30, 84, 81, 84, 253, 101, 146, 109, 77, 85, 28, 103, 30, 113, 75, 86, 190, 102, 215, 158, 60, 0, 42, 0, 111, 92, 73, 87, 243, 95, 154, 130, 104, 153, 88, 140, 90, 140, 216, 142, 233, 152, 128, 156, 104, 155, 21, 151, 192, 129, 203, 129, 60, 0, 42, 0, 61, 108, 125, 117, 60, 0, 42, 0, 71, 78, 219, 78, 172, 139, 88, 98, 97, 98, 81, 108, 102, 153, 84, 103, 130, 79, 131, 84, 213, 98, 214, 98, 176, 108, 41, 99, 93, 99, 140, 131, 165, 136, 23, 138, 182, 109, 43, 129, 49, 129, 229, 152, 96, 155, 93, 154, 60, 0, 42, 0, 110, 154, 87, 79, 64, 150, 65, 150, 104, 87, 174, 92, 177, 108, 178, 108, 207, 114, 241, 143, 124, 154, 35, 120, 36, 120, 137, 136, 53, 158, 61, 125, 54, 88, 206, 141, 97, 145, 162, 120, 177, 153, 214, 105, 196, 153, 221, 153, 222, 153, 80, 106, 128, 155, 21, 157, 39, 159, 40, 154, 13, 159, 82, 154, 9, 159, 60, 0, 42, 0, 117, 95, 165, 89, 185, 94, 160, 90, 45, 105, 85, 105, 247, 90, 98, 106, 78, 157, 12, 155, 22, 156, 60, 0, 42, 0, 221, 103, 228, 107, 62, 85, 26, 132, 197, 141, 251, 107, 168, 123, 0, 134, 92, 124, 60, 0, 39, 0, 208, 253, 87, 0, 39, 0, 60, 0, 42, 0, 114, 92, 117, 122, 184, 82, 22, 99, 60, 109, 50, 90, 86, 117, 138, 122, 167, 90, 215, 85, 217, 134, 50, 100, 155, 110, 37, 111, 170, 122, 3, 159, 40, 101, 60, 0, 42, 0, 3, 90, 60, 0, 42, 0, 230, 116, 100, 79, 183, 144, 147, 84, 25, 120, 60, 0, 42, 0, 156, 136, 73, 128, 226, 85, 125, 129, 131, 129, 106, 137, 200, 151, 228, 151, 60, 0, 42, 0, 242, 116, 199, 84, 60, 0, 42, 0, 106, 107, 142, 85, 245, 122, 60, 0, 42, 0, 52, 93, 60, 0, 42, 0, 22, 89, 30, 89, 97, 152, 60, 0, 42, 0, 47, 95, 92, 82, 96, 90, 53, 94, 70, 88, 126, 110, 63, 135, 107, 111, 76, 140, 78, 95, 234, 88, 99, 112, 60, 0, 42, 0, 56, 78, 19, 82, 77, 108, 168, 126, 132, 130, 140, 91, 143, 92, 143, 98, 169, 115, 2, 123, 8, 125, 86, 99, 125, 152, 247, 112, 19, 116, 166, 140, 17, 152, 235, 127, 60, 0, 42, 0, 155, 91, 7, 80, 13, 85, 61, 99, 204, 118, 230, 87, 73, 90, 203, 96, 90, 102, 101, 102, 154, 104, 254, 126, 24, 129, 192, 131, 22, 132, 105, 102, 124, 102, 0, 105, 44, 116, 150, 118, 121, 117, 85, 119, 151, 120, 169, 125, 176, 125, 19, 143, 32, 142, 212, 146, 60, 0, 42, 0, 7, 78, 77, 83, 80, 83, 167, 89, 232, 95, 100, 103, 101, 99, 21, 129, 44, 132, 85, 129, 16, 143, 171, 111, 196, 146, 163, 119, 141, 133, 61, 147, 195, 135, 3, 141, 171, 147, 14, 141, 60, 0, 42, 0, 35, 92, 42, 92, 43, 92, 106, 108, 41, 92, 60, 0, 42, 0, 161, 78, 190, 78, 102, 81, 139, 115, 252, 78, 122, 95, 163, 131, 159, 134, 60, 0, 42, 0, 82, 127, 81, 127, 128, 95, 131, 95, 137, 103, 84, 127, 141, 95, 216, 96, 245, 131, 128, 102, 226, 104, 231, 134, 139, 143, 178, 125, 68, 135, 183, 138, 30, 143, 7, 112, 77, 155, 60, 0, 42, 0, 132, 89, 216, 95, 203, 143, 250, 101, 243, 118, 27, 103, 34, 103, 60, 0, 42, 0, 113, 83, 1, 90, 211, 112, 78, 80, 14, 132, 54, 144, 135, 150, 136, 150, 180, 85, 153, 90, 4, 97, 203, 99, 251, 99, 40, 110, 104, 132, 115, 132, 174, 95, 51, 105, 114, 105, 166, 110, 104, 113, 116, 138, 50, 135, 5, 126, 91, 135, 163, 137, 182, 93, 135, 133, 240, 113, 194, 156, 205, 93, 3, 156, 4, 156, 60, 0, 42, 0, 215, 86, 230, 151, 41, 87, 244, 86, 15, 94, 169, 108, 221, 143, 241, 149, 215, 92, 222, 92, 8, 109, 203, 151, 69, 104, 160, 109, 47, 85, 55, 94, 223, 96, 89, 120, 244, 126, 161, 85, 13, 87, 129, 90, 108, 93, 67, 94, 75, 110, 136, 110, 31, 116, 85, 144, 77, 111, 173, 125, 246, 132, 44, 145, 89, 111, 127, 111, 209, 120, 128, 145, 240, 111, 79, 147, 200, 149, 160, 155, 83, 118, 185, 137, 169, 114, 58, 151, 8, 107, 60, 0, 42, 0, 131, 83, 31, 79, 42, 79, 62, 92, 172, 126, 155, 130, 199, 130, 212, 89, 156, 112, 174, 115, 39, 109, 19, 90, 87, 92, 100, 99, 88, 109, 113, 131, 255, 139, 73, 80, 125, 80, 35, 93, 182, 104, 207, 117, 74, 120, 169, 154, 84, 93, 171, 95, 7, 97, 37, 115, 102, 132, 141, 132, 170, 154, 171, 154, 144, 102, 50, 105, 82, 113, 75, 116, 255, 117, 114, 129, 73, 130, 234, 151, 222, 80, 177, 100, 200, 120, 148, 156, 234, 91, 239, 125, 63, 133, 201, 138, 19, 142, 209, 151, 32, 152, 179, 133, 48, 81, 251, 111, 97, 147, 170, 155, 221, 88, 34, 112, 217, 151, 185, 152, 225, 151, 150, 101, 60, 0, 42, 0, 107, 83, 58, 78, 42, 103, 77, 79, 115, 84, 255, 130, 186, 112, 79, 117, 195, 128, 222, 83, 206, 142, 9, 92, 203, 131, 19, 140, 130, 85, 166, 90, 45, 110, 50, 114, 44, 115, 95, 113, 155, 88, 168, 120, 26, 133, 60, 135, 112, 97, 173, 113, 154, 114, 237, 125, 95, 135, 91, 136, 192, 97, 143, 116, 123, 127, 94, 136, 2, 139, 39, 153, 135, 155, 177, 135, 61, 137, 53, 153, 79, 155, 239, 133, 74, 143, 207, 147, 40, 151, 218, 156, 54, 134, 86, 153, 134, 139, 151, 142, 143, 139, 155, 142, 60, 0, 42, 0, 64, 113, 60, 0, 42, 0, 55, 102, 109, 88, 41, 110, 133, 105, 159, 107, 171, 110, 101, 116, 146, 143, 178, 105, 31, 118, 112, 133, 113, 140, 60, 143, 64, 143, 190, 147, 66, 153, 193, 156, 176, 151, 27, 156, 46, 156, 60, 0, 42, 0, 1, 83, 135, 101, 99, 95, 185, 126, 160, 130, 134, 112, 159, 115, 251, 149, 11, 125, 137, 134, 138, 134, 243, 115, 12, 150, 239, 150, 18, 118, 94, 128, 188, 153, 199, 153, 112, 155, 252, 156, 13, 157, 161, 135, 186, 149, 191, 149, 193, 135, 197, 149, 36, 159, 230, 149, 60, 0, 42, 0, 14, 82, 59, 84, 223, 95, 134, 98, 97, 84, 179, 128, 10, 125, 125, 104, 23, 129, 51, 122, 79, 122, 105, 122, 60, 0, 42, 0, 238, 149, 143, 89, 118, 108, 172, 131, 79, 85, 2, 110, 254, 99, 53, 100, 80, 152, 186, 116, 60, 0, 42, 0, 90, 84, 60, 0, 42, 0, 193, 127, 225, 85, 195, 110, 95, 158, 137, 135, 147, 147, 178, 157, 60, 0, 42, 0, 220, 82, 99, 89, 85, 88, 97, 93, 202, 132, 161, 102, 136, 119, 108, 128, 26, 101, 60, 0, 42, 0, 238, 116, 121, 133, 21, 117, 75, 127, 70, 159, 60, 0, 42, 0, 29, 99, 45, 80, 161, 109, 180, 131, 41, 85, 185, 109, 38, 110, 39, 115, 53, 132, 157, 122, 169, 122, 23, 135, 190, 100, 120, 135, 18, 142, 60, 0, 42, 0, 17, 98, 80, 90, 112, 99, 60, 0, 42, 0, 244, 78, 131, 108, 159, 128, 103, 83, 130, 103, 229, 129, 83, 80, 126, 99, 180, 109, 137, 90, 68, 94, 225, 99, 37, 110, 37, 113, 106, 120, 67, 105, 91, 129, 161, 101, 131, 119, 237, 100, 227, 111, 193, 116, 210, 129, 216, 150, 140, 159, 119, 159, 60, 0, 42, 0, 76, 78, 44, 87, 25, 95, 89, 108, 90, 108, 97, 108, 172, 144, 92, 84, 235, 93, 71, 103, 75, 92, 63, 109, 236, 139, 168, 148, 207, 112, 109, 82, 143, 122, 20, 145, 218, 85, 136, 138, 77, 107, 163, 138, 188, 123, 144, 135, 46, 157, 162, 147, 30, 156, 60, 0, 42, 0, 224, 101, 203, 107, 51, 84, 52, 84, 62, 84, 73, 84, 156, 130, 218, 144, 20, 85, 42, 90, 22, 109, 111, 109, 35, 131, 129, 131, 167, 104, 248, 115, 102, 121, 33, 113, 251, 148, 64, 158, 145, 121, 8, 135, 106, 133, 145, 116, 241, 135, 195, 155, 80, 157, 85, 139, 47, 159, 225, 157, 60, 0, 42, 0, 68, 78, 148, 78, 72, 83, 245, 78, 13, 79, 169, 89, 145, 94, 228, 95, 3, 96, 255, 101, 102, 107, 157, 115, 174, 79, 201, 79, 53, 80, 66, 99, 78, 85, 44, 90, 126, 114, 247, 115, 64, 100, 148, 120, 73, 158, 147, 113, 102, 116, 30, 130, 245, 90, 225, 94, 174, 97, 85, 111, 27, 81, 70, 106, 18, 117, 97, 157, 140, 142, 60, 0, 42, 0, 64, 81, 255, 82, 161, 82, 10, 98, 34, 150, 6, 79, 124, 92, 100, 98, 94, 87, 137, 92, 76, 103, 180, 130, 213, 143, 226, 95, 105, 114, 249, 119, 124, 83, 68, 101, 239, 139, 217, 82, 158, 96, 159, 96, 174, 96, 133, 124, 28, 144, 100, 102, 16, 113, 122, 90, 77, 93, 230, 117, 150, 150, 112, 151, 155, 154, 98, 88, 102, 89, 104, 93, 169, 110, 250, 150, 254, 150, 228, 91, 131, 113, 164, 138, 92, 158, 200, 146, 185, 122, 26, 151, 63, 159, 39, 151, 64, 159, 1, 134, 22, 154, 169, 157, 60, 0, 42, 0, 59, 147, 60, 0, 39, 0, 208, 253, 88, 0, 39, 0, 60, 0, 42, 0, 21, 89, 110, 81, 156, 144, 56, 84, 218, 95, 113, 98, 80, 108, 127, 137, 128, 137, 12, 94, 120, 98, 101, 83, 20, 102, 144, 103, 120, 122, 184, 128, 185, 128, 217, 79, 134, 95, 56, 96, 83, 96, 215, 144, 123, 153, 15, 85, 90, 89, 86, 92, 111, 96, 149, 96, 37, 108, 96, 109, 122, 114, 246, 114, 131, 131, 61, 85, 137, 96, 220, 96, 94, 102, 120, 104, 55, 107, 197, 109, 19, 110, 239, 112, 1, 113, 8, 113, 11, 116, 82, 120, 229, 131, 101, 141, 248, 145, 146, 80, 193, 96, 112, 102, 115, 102, 31, 113, 44, 113, 128, 114, 78, 119, 0, 122, 158, 124, 213, 127, 214, 127, 62, 130, 14, 145, 128, 83, 96, 93, 175, 95, 170, 110, 153, 118, 160, 132, 33, 149, 214, 80, 189, 105, 85, 113, 132, 113, 136, 113, 153, 113, 198, 125, 37, 135, 104, 140, 15, 153, 59, 86, 79, 86, 6, 91, 9, 91, 178, 93, 93, 111, 28, 118, 206, 120, 157, 129, 222, 81, 153, 97, 40, 106, 64, 106, 185, 113, 186, 113, 187, 113, 184, 122, 24, 126, 178, 127, 133, 135, 135, 135, 43, 147, 232, 113, 160, 114, 166, 119, 2, 121, 203, 135, 63, 140, 64, 140, 111, 140, 149, 140, 230, 124, 101, 126, 223, 150, 209, 155, 87, 157, 249, 137, 70, 139, 175, 145, 237, 147, 181, 150, 199, 93, 230, 102, 20, 114, 167, 114, 69, 145, 253, 137, 55, 159, 60, 0, 42, 0, 53, 136, 2, 158, 255, 137, 116, 148, 60, 0, 42, 0, 96, 78, 203, 144, 45, 94, 210, 127, 173, 136, 203, 137, 186, 85, 179, 90, 58, 105, 181, 132, 198, 132, 141, 93, 29, 111, 161, 137, 152, 141, 226, 105, 130, 133, 176, 150, 132, 106, 53, 139, 180, 147, 43, 151, 219, 156, 193, 152, 49, 154, 61, 154, 114, 137, 60, 156, 104, 154, 60, 0, 42, 0, 178, 103, 23, 109, 186, 115, 153, 95, 227, 148, 156, 85, 72, 132, 120, 132, 34, 146, 104, 146, 105, 146, 99, 92, 7, 111, 240, 132, 152, 97, 191, 102, 114, 106, 86, 107, 167, 121, 240, 138, 208, 88, 48, 126, 17, 139, 226, 135, 93, 142, 189, 116, 205, 86, 90, 156, 214, 119, 167, 142, 60, 0, 42, 0, 56, 83, 76, 83, 15, 98, 67, 92, 251, 124, 105, 153, 108, 84, 229, 95, 44, 96, 253, 119, 198, 126, 194, 79, 165, 84, 68, 96, 251, 118, 228, 144, 47, 107, 228, 126, 48, 125, 243, 145, 11, 150, 30, 105, 3, 130, 4, 130, 135, 141, 153, 150, 64, 97, 202, 110, 138, 121, 140, 125, 105, 141, 159, 150, 141, 88, 130, 113, 148, 114, 39, 122, 95, 111, 153, 111, 110, 133, 164, 137, 49, 98, 214, 158, 50, 98, 246, 120, 105, 134, 60, 153, 41, 155, 107, 126, 177, 86, 223, 149, 60, 151, 109, 92, 75, 136, 60, 0, 42, 0, 114, 134, 168, 117, 126, 134, 58, 140, 132, 80, 149, 149, 70, 113, 172, 152, 142, 119, 102, 135, 21, 156, 60, 0, 42, 0, 35, 83, 160, 79, 206, 114, 224, 79, 225, 92, 217, 103, 160, 112, 237, 114, 92, 150, 253, 92, 218, 112, 249, 114, 232, 115, 107, 121, 86, 120, 200, 127, 58, 130, 127, 150, 100, 120, 80, 144, 110, 101, 135, 102, 85, 116, 106, 123, 29, 130, 172, 120, 150, 143, 205, 120, 0, 126, 120, 133, 22, 126, 110, 141, 123, 155, 68, 143, 92, 147, 30, 151, 139, 147, 224, 158, 34, 154, 183, 157, 60, 0, 42, 0, 156, 149, 60, 0, 42, 0, 5, 78, 11, 78, 100, 78, 19, 84, 55, 87, 156, 117, 15, 89, 186, 104, 113, 119, 135, 86, 215, 97, 69, 127, 188, 147, 19, 89, 236, 147, 60, 0, 42, 0, 217, 78, 218, 78, 115, 92, 72, 81, 126, 89, 164, 126, 97, 79, 250, 95, 25, 108, 116, 103, 70, 121, 200, 121, 238, 130, 174, 103, 124, 124, 215, 115, 182, 131, 128, 99, 230, 148, 249, 141, 112, 145, 40, 149, 202, 80, 242, 80, 21, 86, 155, 146, 156, 156, 185, 102, 239, 151, 16, 91, 184, 97, 159, 133, 65, 147, 74, 126, 60, 137, 241, 151, 174, 155, 110, 142, 166, 153, 239, 94, 21, 101, 142, 126, 177, 157, 115, 137, 154, 142, 150, 126, 123, 156, 60, 0, 42, 0, 45, 79, 242, 149, 182, 89, 38, 95, 36, 141, 184, 84, 12, 85, 38, 99, 142, 109, 216, 128, 52, 90, 57, 90, 113, 90, 67, 125, 55, 130, 191, 134, 84, 136, 99, 85, 122, 110, 235, 117, 221, 134, 145, 149, 146, 149, 71, 158, 204, 90, 88, 136, 9, 117, 156, 146, 250, 90, 251, 90, 170, 97, 143, 100, 150, 111, 52, 122, 184, 138, 226, 140, 244, 138, 49, 143, 142, 145, 71, 118, 78, 118, 175, 119, 214, 133, 37, 121, 121, 158, 153, 158, 18, 141, 243, 157, 244, 157, 252, 157, 60, 0, 42, 0, 188, 81, 221, 114, 62, 102, 105, 150, 4, 93, 232, 107, 205, 112, 3, 115, 172, 134, 122, 150, 123, 141, 69, 123, 31, 92, 32, 92, 31, 100, 146, 121, 227, 141, 145, 146, 178, 123, 170, 150, 174, 93, 107, 115, 110, 115, 211, 133, 76, 147, 249, 113, 85, 152, 112, 94, 7, 101, 246, 106, 26, 134, 99, 139, 129, 115, 197, 151, 111, 152, 102, 112, 60, 0, 42, 0, 35, 79, 191, 83, 158, 84, 152, 92, 203, 130, 176, 115, 191, 126, 253, 129, 80, 150, 237, 89, 170, 91, 12, 119, 101, 150, 239, 84, 183, 87, 10, 90, 40, 90, 244, 92, 128, 109, 167, 131, 119, 150, 91, 102, 254, 115, 77, 120, 133, 153, 77, 119, 100, 125, 16, 127, 161, 127, 46, 115, 175, 124, 168, 127, 122, 129, 6, 135, 233, 80, 244, 80, 171, 125, 162, 138, 138, 100, 218, 125, 231, 146, 178, 97, 76, 106, 35, 126, 14, 147, 33, 153, 207, 88, 79, 140, 178, 158, 23, 112, 212, 129, 123, 115, 238, 124, 56, 159, 60, 0, 42, 0, 102, 148, 60, 0, 42, 0, 97, 78, 151, 130, 248, 118, 153, 153, 247, 144, 162, 83, 76, 85, 9, 145, 10, 145, 194, 94, 88, 110, 3, 127, 89, 132, 21, 145, 127, 105, 37, 122, 140, 133, 177, 123, 215, 125, 183, 129, 68, 137, 192, 95, 167, 154, 152, 158, 0, 107, 214, 116, 118, 149, 92, 156, 114, 148, 100, 154, 60, 0, 42, 0, 232, 116, 109, 79, 230, 139, 160, 94, 25, 104, 101, 121, 116, 125, 212, 127, 115, 138, 237, 141, 60, 0, 42, 0, 171, 78, 175, 78, 205, 84, 131, 134, 119, 153, 81, 102, 232, 152, 243, 96, 132, 146, 9, 153, 158, 156, 207, 102, 1, 136, 157, 155, 215, 155, 255, 151, 87, 153, 95, 153, 118, 156, 60, 0, 42, 0, 17, 84, 224, 89, 247, 93, 121, 152, 230, 115, 97, 140, 66, 88, 63, 127, 43, 132, 86, 136, 5, 152, 207, 80, 232, 82, 145, 93, 151, 146, 97, 106, 80, 137, 174, 86, 211, 135, 12, 148, 76, 156, 60, 0, 42, 0, 113, 112, 114, 112, 122, 84, 173, 103, 190, 79, 211, 84, 181, 103, 129, 154, 238, 84, 175, 91, 181, 91, 168, 94, 136, 109, 225, 126, 83, 134, 13, 144, 46, 158, 75, 90, 159, 104, 7, 113, 7, 115, 39, 132, 218, 117, 223, 117, 93, 120, 99, 120, 153, 122, 219, 127, 55, 132, 0, 149, 241, 99, 131, 125, 16, 86, 74, 107, 71, 111, 171, 123, 3, 142, 53, 86, 162, 97, 98, 115, 183, 146, 4, 151, 71, 95, 174, 129, 109, 133, 72, 155, 30, 157, 88, 122, 24, 124, 195, 133, 194, 135, 207, 135, 53, 157, 163, 86, 31, 112, 43, 124, 240, 135, 199, 154, 249, 106, 187, 86, 194, 86, 208, 154, 205, 157, 40, 136, 77, 154, 202, 107, 72, 134, 60, 0, 42, 0, 40, 109, 233, 144, 36, 93, 198, 109, 36, 138, 189, 107, 74, 123, 181, 138, 60, 0, 42, 0, 15, 92, 83, 102, 129, 102, 113, 123, 127, 123, 155, 118, 201, 102, 224, 123, 15, 139, 162, 118, 60, 0, 42, 0, 93, 91, 150, 128, 185, 82, 178, 84, 242, 79, 72, 101, 33, 104, 141, 109, 17, 123, 120, 85, 154, 80, 105, 101, 104, 138, 11, 86, 40, 86, 159, 138, 47, 86, 87, 107, 189, 113, 133, 101, 134, 101, 60, 0, 42, 0, 119, 96, 60, 0, 42, 0, 155, 78, 243, 99, 50, 115, 84, 105, 71, 107, 78, 135, 13, 136, 60, 0, 42, 0, 166, 82, 79, 83, 234, 101, 170, 144, 84, 83, 193, 128, 165, 87, 74, 89, 235, 92, 74, 96, 249, 98, 31, 99, 62, 99, 5, 129, 7, 129, 11, 129, 122, 136, 85, 80, 156, 101, 16, 140, 211, 127, 203, 85, 54, 97, 58, 100, 78, 116, 138, 125, 129, 113, 142, 129, 240, 82, 183, 100, 213, 100, 243, 125, 44, 127, 98, 135, 139, 151, 33, 152, 231, 138, 242, 113, 247, 100, 181, 151, 109, 137, 28, 101, 136, 126, 151, 139, 164, 159, 60, 0, 42, 0, 153, 81, 169, 81, 235, 91, 219, 133, 60, 0, 42, 0, 51, 79, 122, 112, 196, 108, 251, 108, 68, 121, 193, 126, 55, 127, 120, 83, 41, 109, 167, 112, 168, 112, 104, 83, 14, 90, 81, 92, 83, 92, 94, 80, 112, 80, 162, 95, 176, 104, 242, 112, 14, 113, 188, 121, 50, 125, 181, 78, 159, 90, 95, 92, 43, 110, 79, 125, 108, 125, 34, 140, 193, 80, 110, 88, 141, 105, 173, 105, 9, 137, 103, 86, 103, 92, 172, 102, 228, 125, 176, 93, 232, 94, 200, 97, 165, 111, 108, 115, 207, 124, 162, 133, 164, 133, 130, 144, 240, 151, 238, 113, 59, 137, 29, 139, 17, 89, 9, 112, 162, 151, 35, 112, 21, 114, 114, 126, 249, 135, 15, 136, 88, 159, 91, 159, 101, 159, 66, 159, 158, 142, 60, 0, 42, 0, 195, 95, 164, 144, 161, 89, 251, 95, 175, 130, 155, 143, 21, 102, 122, 103, 35, 107, 152, 112, 250, 118, 253, 79, 222, 96, 34, 138, 10, 146, 12, 149, 176, 101, 70, 107, 222, 94, 197, 146, 122, 86, 28, 91, 170, 133, 168, 153, 107, 148, 171, 153, 60, 0, 42, 0, 148, 103, 81, 137, 20, 148, 60, 0, 42, 0, 8, 79, 60, 0, 42, 0, 32, 150, 41, 79, 223, 86, 94, 91, 225, 79, 208, 142, 42, 129, 69, 136, 43, 138, 46, 113, 184, 153, 86, 152, 11, 130, 193, 145, 60, 0, 42, 0, 196, 95, 60, 0, 42, 0, 31, 102, 182, 87, 141, 154, 250, 96, 41, 115, 75, 113, 70, 116, 101, 129, 245, 134, 234, 137, 181, 123, 194, 123, 3, 139, 143, 155, 208, 102, 242, 137, 2, 154, 168, 118, 249, 155, 60, 0, 42, 0, 17, 82, 76, 136, 162, 144, 98, 95, 73, 150, 128, 79, 201, 144, 139, 87, 16, 109, 101, 131, 152, 148, 88, 150, 25, 90, 78, 120, 207, 148, 3, 146, 206, 110, 118, 146, 146, 146, 222, 146, 60, 0, 42, 0, 114, 119, 146, 145, 228, 100, 60, 0, 42, 0, 116, 81, 79, 103, 211, 89, 120, 94, 39, 96, 71, 131, 22, 80, 149, 131, 94, 90, 187, 96, 172, 109, 200, 125, 8, 130, 57, 91, 214, 129, 60, 0, 42, 0, 216, 84, 196, 136, 60, 0, 42, 0, 246, 81, 2, 83, 68, 81, 71, 81, 8, 83, 187, 139, 247, 95, 121, 108, 197, 84, 95, 96, 54, 109, 247, 128, 248, 128, 41, 138, 126, 138, 239, 140, 60, 0, 42, 0, 196, 150, 138, 113, 60, 0, 42, 0, 61, 113, 60, 0, 42, 0, 199, 139, 56, 113, 87, 138, 16, 89, 123, 101, 60, 0, 42, 0, 17, 79, 226, 79, 238, 79, 187, 84, 165, 94, 203, 112, 204, 112, 158, 127, 41, 129, 25, 129, 58, 158, 249, 129, 133, 140, 144, 153, 7, 106, 157, 146, 228, 154, 249, 154, 128, 147, 180, 155, 66, 157, 197, 147, 72, 153, 67, 156, 205, 152, 60, 0, 42, 0, 236, 130, 60, 0, 42, 0, 61, 103, 235, 110, 67, 111, 135, 125, 212, 124, 60, 0, 42, 0, 192, 121, 171, 92, 192, 92, 219, 115, 227, 126, 150, 136, 7, 116, 8, 149, 197, 85, 180, 110, 147, 116, 14, 137, 15, 137, 185, 146, 145, 135, 77, 126, 97, 126, 229, 147, 253, 147, 69, 159, 60, 0, 42, 0, 1, 84, 12, 98, 244, 101, 158, 117, 241, 118, 40, 107, 229, 128, 123, 152, 71, 102, 15, 138, 124, 152, 87, 134, 90, 134, 30, 140, 173, 90, 65, 94, 223, 99, 81, 110, 91, 134, 199, 136, 8, 152, 72, 105, 162, 122, 10, 152, 24, 86, 159, 88, 0, 151, 70, 155, 83, 86, 3, 91, 84, 107, 3, 126, 102, 133, 81, 135, 221, 138, 67, 139, 123, 126, 86, 155, 73, 154, 80, 148, 26, 155, 60, 0, 42, 0, 198, 79, 144, 95, 163, 132, 60, 0, 42, 0, 184, 139, 116, 84, 193, 89, 233, 139, 148, 81, 41, 104, 221, 115, 102, 80, 49, 138, 138, 102, 97, 138, 48, 122, 38, 145, 200, 124, 145, 145, 232, 118, 60, 0, 42, 0, 237, 101, 53, 79, 143, 94, 127, 108, 167, 130, 144, 79, 121, 83, 52, 96, 128, 108, 217, 83, 100, 96, 43, 102, 43, 109, 191, 87, 48, 107, 136, 107, 197, 112, 236, 115, 214, 82, 77, 101, 88, 101, 215, 82, 252, 112, 234, 126, 237, 126, 87, 145, 163, 85, 251, 88, 127, 90, 2, 103, 134, 110, 110, 125, 83, 128, 57, 138, 73, 97, 102, 113, 154, 125, 196, 132, 201, 140, 210, 105, 53, 111, 74, 111, 226, 118, 129, 119, 210, 125, 95, 128, 138, 146, 93, 115, 56, 122, 214, 125, 99, 155, 199, 133, 178, 119, 218, 133, 140, 126, 110, 156, 60, 0, 42, 0, 255, 132, 60, 0, 42, 0, 5, 84, 105, 143, 13, 102, 163, 91, 50, 95, 210, 142, 139, 104, 22, 140, 167, 85, 71, 88, 151, 90, 3, 97, 11, 97, 206, 99, 49, 132, 50, 132, 132, 102, 74, 113, 68, 116, 210, 132, 123, 119, 7, 81, 164, 121, 174, 123, 7, 126, 231, 127, 86, 135, 215, 146, 193, 97, 127, 133, 224, 138, 252, 138, 121, 147, 253, 153, 206, 119, 254, 127, 252, 133, 16, 134, 9, 136, 94, 139, 26, 156, 60, 0, 42, 0, 132, 115, 185, 115, 195, 117, 172, 96, 203, 101, 1, 116, 1, 135, 217, 90, 41, 111, 182, 102, 135, 116, 136, 106, 191, 116, 248, 97, 60, 0, 42, 0, 186, 84, 9, 144, 69, 102, 220, 112, 120, 144, 72, 152, 99, 118, 108, 118, 60, 0, 42, 0, 48, 96, 235, 108, 33, 102, 171, 112, 218, 126, 41, 119, 168, 136, 201, 148, 4, 116, 52, 119, 82, 136, 50, 110, 98, 125, 101, 105, 102, 105, 73, 146, 185, 120, 25, 133, 95, 149, 153, 151, 180, 152, 60, 126, 79, 126, 199, 147, 130, 139, 25, 141, 60, 0, 42, 0, 74, 82, 182, 117, 134, 132, 116, 151, 155, 133, 165, 143, 170, 143, 190, 151, 60, 0, 42, 0, 116, 122, 136, 101, 116, 78, 102, 91, 164, 92, 195, 92, 19, 131, 246, 108, 149, 136, 52, 158, 5, 142, 198, 88, 120, 91, 168, 93, 169, 111, 226, 113, 247, 137, 228, 150, 253, 157, 60, 0, 42, 0, 234, 150, 164, 129, 48, 106, 93, 130, 76, 143, 213, 156, 72, 156, 60, 0, 42, 0, 64, 136, 55, 84, 121, 87, 216, 114, 86, 104, 17, 140, 144, 141, 20, 139, 37, 112, 60, 0, 42, 0, 67, 87, 203, 82, 217, 87, 4, 113, 219, 82, 100, 88, 143, 113, 168, 122, 18, 133, 242, 82, 243, 82, 171, 133, 232, 153, 145, 86, 206, 88, 111, 115, 176, 133, 219, 102, 251, 113, 208, 129, 196, 119, 13, 134, 230, 88, 11, 114, 129, 126, 186, 145, 60, 0, 42, 0, 245, 94, 251, 91, 236, 101, 225, 93, 111, 154, 74, 103, 67, 117, 226, 139, 203, 92, 66, 96, 53, 109, 84, 109, 3, 125, 64, 131, 104, 131, 18, 104, 106, 104, 229, 107, 227, 115, 113, 80, 11, 92, 170, 95, 215, 99, 98, 138, 180, 153, 41, 145, 159, 156, 90, 86, 111, 111, 51, 101, 51, 106, 197, 113, 214, 113, 149, 116, 243, 135, 79, 156, 88, 156, 101, 112, 60, 0, 42, 0, 66, 83, 173, 139, 175, 139, 40, 79, 91, 108, 197, 143, 154, 79, 250, 93, 135, 95, 229, 114, 255, 143, 10, 144, 137, 107, 10, 138, 19, 138, 25, 138, 94, 89, 253, 93, 190, 107, 4, 122, 92, 144, 59, 97, 208, 140, 64, 86, 96, 111, 72, 133, 104, 152, 66, 148, 60, 0, 39, 0, 208, 253, 89, 0, 39, 0, 60, 0, 42, 0, 43, 78, 39, 87, 139, 83, 22, 84, 152, 94, 188, 98, 146, 103, 173, 87, 38, 158, 96, 104, 45, 158, 225, 87, 114, 91, 212, 148, 15, 105, 9, 157, 15, 147, 40, 157, 211, 88, 118, 157, 26, 148, 60, 0, 42, 0, 89, 114, 34, 79, 145, 83, 136, 92, 189, 130, 147, 83, 161, 115, 10, 116, 12, 123, 156, 134, 16, 88, 21, 93, 22, 93, 175, 109, 26, 115, 88, 116, 90, 119, 89, 136, 4, 111, 86, 159, 60, 0, 42, 0, 138, 83, 140, 94, 209, 84, 22, 85, 94, 85, 214, 117, 197, 150, 2, 118, 101, 133, 60, 0, 42, 0, 156, 82, 32, 87, 103, 143, 154, 78, 126, 137, 182, 139, 156, 78, 189, 114, 211, 143, 158, 78, 203, 142, 5, 90, 28, 99, 17, 120, 249, 79, 41, 108, 109, 90, 151, 99, 29, 138, 224, 99, 44, 108, 48, 115, 80, 128, 20, 87, 15, 122, 171, 122, 126, 159, 60, 0, 42, 0, 91, 78, 64, 84, 60, 0, 42, 0, 121, 96, 102, 82, 223, 112, 218, 115, 237, 128, 99, 80, 38, 93, 202, 109, 249, 109, 9, 113, 17, 113, 248, 131, 9, 150, 110, 110, 76, 129, 89, 113, 125, 120, 34, 145, 227, 90, 57, 111, 150, 93, 46, 106, 131, 145, 185, 149, 46, 91, 232, 97, 246, 123, 213, 97, 217, 129, 235, 158, 60, 0, 42, 0, 160, 139, 246, 94, 235, 149, 37, 78, 141, 89, 171, 130, 0, 138, 1, 138, 169, 92, 22, 102, 191, 108, 142, 112, 212, 144, 248, 89, 43, 90, 255, 114, 20, 120, 154, 131, 46, 90, 208, 118, 113, 85, 2, 116, 79, 120, 46, 138, 134, 149, 14, 150, 82, 93, 83, 93, 117, 123, 150, 125, 18, 135, 105, 88, 197, 99, 76, 105, 125, 138, 158, 120, 5, 133, 156, 152, 179, 83, 100, 134, 187, 149, 144, 106, 79, 152, 84, 152, 180, 86, 219, 88, 204, 93, 55, 124, 233, 106, 236, 158, 231, 88, 77, 91, 215, 93, 214, 93, 21, 107, 57, 121, 125, 158, 163, 158, 60, 0, 42, 0, 53, 89, 129, 98, 135, 108, 117, 78, 86, 81, 68, 89, 232, 79, 87, 81, 61, 83, 7, 95, 77, 136, 67, 80, 163, 83, 169, 99, 60, 119, 18, 132, 254, 144, 83, 145, 67, 93, 29, 97, 74, 98, 220, 99, 234, 104, 48, 110, 55, 110, 48, 116, 67, 144, 146, 150, 60, 105, 104, 127, 250, 136, 20, 111, 23, 137, 45, 98, 88, 135, 71, 155, 94, 86, 189, 142, 47, 126, 191, 106, 225, 158, 180, 83, 23, 117, 11, 156, 160, 157, 228, 158, 94, 159, 145, 159, 60, 81, 237, 158, 105, 152, 52, 159, 216, 93, 218, 93, 238, 102, 88, 155, 57, 159, 116, 159, 246, 158, 60, 0, 42, 0, 140, 83, 159, 89, 195, 137, 106, 114, 189, 84, 242, 89, 101, 95, 102, 95, 26, 120, 1, 85, 180, 91, 79, 102, 251, 112, 115, 130, 142, 137, 140, 154, 80, 80, 20, 113, 26, 140, 129, 150, 173, 85, 48, 88, 101, 101, 48, 113, 49, 113, 18, 115, 111, 120, 85, 132, 193, 150, 191, 80, 59, 105, 142, 110, 223, 110, 235, 156, 173, 83, 149, 88, 165, 102, 150, 113, 125, 145, 10, 91, 51, 140, 13, 153, 8, 157, 196, 113, 213, 113, 250, 138, 93, 141, 51, 155, 213, 102, 51, 157, 64, 145, 16, 154, 19, 154, 165, 86, 63, 91, 118, 130, 11, 141, 197, 142, 227, 102, 19, 114, 182, 145, 52, 154, 195, 157, 84, 112, 23, 141, 254, 137, 140, 139, 188, 145, 92, 153, 87, 154, 240, 157, 119, 130, 78, 112, 197, 145, 96, 154, 103, 112, 158, 139, 83, 140, 84, 140, 105, 112, 60, 0, 42, 0, 46, 89, 137, 84, 206, 89, 176, 98, 241, 108, 131, 107, 230, 128, 15, 119, 231, 121, 47, 158, 96, 146, 245, 150, 133, 151, 38, 157, 60, 0, 42, 0, 108, 98, 138, 127, 38, 150, 51, 150, 248, 101, 104, 103, 128, 112, 207, 152, 111, 79, 183, 82, 28, 108, 161, 117, 150, 148, 159, 87, 137, 95, 28, 102, 11, 109, 143, 127, 202, 112, 220, 115, 59, 119, 125, 150, 53, 93, 56, 93, 218, 99, 216, 134, 109, 101, 152, 102, 74, 105, 108, 113, 147, 121, 13, 118, 249, 138, 48, 143, 90, 147, 57, 157, 186, 152, 10, 148, 17, 156, 55, 151, 9, 158, 60, 0, 42, 0, 240, 78, 82, 79, 113, 87, 159, 92, 123, 81, 205, 103, 180, 112, 39, 108, 210, 117, 59, 125, 159, 80, 103, 105, 238, 142, 67, 97, 49, 108, 70, 135, 10, 153, 218, 153, 233, 97, 1, 101, 98, 118, 34, 121, 60, 0, 42, 0, 15, 96, 89, 96, 55, 104, 149, 127, 71, 138, 216, 105, 62, 111, 35, 106, 1, 112, 60, 0, 42, 0, 170, 127, 60, 0, 42, 0, 122, 94, 45, 89, 6, 84, 150, 89, 150, 103, 128, 107, 69, 121, 30, 138, 147, 85, 125, 132, 70, 105, 112, 129, 1, 157, 128, 144, 60, 0, 42, 0, 59, 114, 39, 92, 45, 92, 180, 128, 154, 87, 218, 89, 227, 92, 122, 143, 4, 80, 209, 112, 231, 115, 145, 122, 156, 80, 47, 88, 250, 99, 35, 140, 250, 142, 194, 85, 177, 90, 173, 95, 46, 97, 22, 100, 71, 100, 231, 110, 58, 115, 89, 144, 101, 144, 127, 100, 154, 102, 163, 105, 100, 116, 118, 116, 154, 146, 214, 152, 6, 153, 162, 93, 164, 93, 216, 120, 175, 122, 176, 122, 26, 153, 71, 126, 32, 139, 33, 139, 144, 147, 208, 156, 187, 152, 40, 134, 142, 144, 100, 152, 41, 156, 60, 0, 42, 0, 248, 78, 142, 91, 134, 92, 173, 98, 115, 103, 213, 114, 237, 130, 172, 84, 252, 103, 17, 119, 133, 122, 136, 122, 0, 130, 96, 80, 121, 90, 62, 93, 148, 110, 212, 132, 154, 105, 34, 157, 196, 149, 21, 154, 105, 159, 213, 157, 60, 0, 42, 0, 126, 122, 111, 131, 129, 137, 165, 148, 142, 136, 148, 122, 68, 123, 111, 132, 79, 138, 142, 113, 158, 137, 127, 151, 95, 115, 94, 158, 172, 133, 60, 159, 220, 102, 255, 113, 94, 130, 229, 133, 197, 119, 0, 128, 133, 126, 194, 157, 145, 139, 112, 148, 60, 0, 42, 0, 59, 80, 150, 99, 48, 105, 141, 102, 78, 86, 113, 111, 46, 136, 60, 0, 42, 0, 55, 114, 54, 128, 83, 99, 246, 99, 216, 148, 58, 114, 254, 145, 227, 146, 129, 147, 232, 100, 60, 0, 42, 0, 95, 78, 20, 84, 182, 81, 220, 87, 206, 145, 34, 86, 28, 111, 196, 88, 60, 0, 42, 0, 26, 78, 246, 83, 243, 102, 117, 152, 245, 102, 186, 144, 28, 89, 180, 98, 177, 78, 188, 103, 2, 109, 1, 152, 89, 99, 84, 102, 189, 103, 232, 112, 88, 85, 178, 109, 18, 140, 40, 88, 151, 107, 75, 129, 73, 132, 19, 145, 183, 88, 106, 105, 109, 105, 140, 153, 247, 80, 196, 102, 197, 102, 75, 107, 193, 113, 141, 116, 219, 100, 163, 118, 177, 119, 52, 145, 101, 151, 170, 93, 171, 93, 178, 111, 1, 139, 35, 153, 136, 86, 235, 100, 215, 102, 184, 119, 113, 147, 234, 100, 23, 114, 15, 121, 145, 147, 65, 153, 122, 157, 55, 148, 104, 151, 92, 154, 8, 158, 60, 0, 42, 0, 170, 78, 60, 0, 42, 0, 0, 78, 74, 78, 12, 95, 183, 143, 100, 136, 10, 79, 99, 136, 59, 83, 26, 84, 241, 88, 157, 79, 78, 121, 191, 84, 34, 109, 152, 96, 15, 110, 23, 115, 105, 117, 252, 144, 241, 148, 249, 88, 214, 99, 57, 107, 220, 134, 149, 121, 219, 90, 42, 111, 38, 122, 165, 146, 4, 91, 107, 86, 1, 89, 127, 116, 101, 158, 68, 126, 185, 106, 201, 107, 171, 145, 223, 158, 105, 139, 214, 157, 243, 158, 60, 0, 42, 0, 65, 78, 234, 78, 28, 83, 47, 87, 55, 89, 198, 143, 157, 81, 144, 91, 130, 108, 210, 139, 135, 79, 156, 91, 33, 96, 182, 108, 203, 114, 106, 136, 228, 143, 116, 153, 166, 84, 232, 89, 211, 92, 94, 96, 248, 98, 194, 103, 198, 115, 245, 116, 59, 141, 251, 143, 167, 91, 248, 93, 44, 95, 69, 98, 24, 104, 75, 104, 25, 119, 240, 128, 152, 136, 17, 138, 164, 140, 205, 117, 251, 121, 27, 128, 19, 132, 44, 105, 160, 127, 230, 134, 81, 138, 82, 138, 189, 140, 87, 144, 144, 90, 134, 102, 56, 105, 131, 138, 224, 141, 9, 152, 144, 152, 244, 152, 145, 117, 0, 81, 170, 113, 183, 123, 122, 144, 172, 93, 91, 95, 92, 95, 148, 135, 36, 152, 37, 152, 242, 91, 183, 93, 3, 124, 74, 152, 93, 95, 94, 95, 59, 139, 212, 147, 78, 124, 250, 137, 137, 139, 3, 158, 60, 0, 42, 0, 89, 78, 242, 93, 229, 78, 135, 148, 65, 79, 58, 101, 227, 119, 225, 130, 226, 130, 161, 94, 35, 130, 129, 134, 212, 145, 26, 80, 70, 98, 22, 123, 24, 144, 79, 145, 111, 80, 58, 93, 209, 101, 5, 105, 24, 146, 111, 146, 230, 156, 255, 136, 214, 101, 38, 142, 34, 143, 124, 101, 152, 135, 165, 106, 18, 121, 100, 130, 251, 135, 87, 152, 89, 143, 110, 159, 60, 0, 42, 0, 66, 78, 73, 78, 191, 78, 11, 95, 8, 82, 198, 95, 122, 130, 7, 83, 138, 128, 174, 139, 166, 78, 7, 79, 121, 92, 2, 95, 133, 130, 63, 79, 90, 79, 174, 82, 83, 84, 68, 87, 121, 95, 145, 98, 89, 103, 52, 128, 197, 130, 209, 139, 145, 144, 126, 79, 109, 84, 121, 84, 196, 92, 8, 96, 63, 96, 19, 102, 141, 103, 37, 107, 198, 108, 136, 112, 199, 121, 206, 126, 227, 139, 127, 154, 203, 79, 85, 89, 31, 94, 32, 94, 8, 95, 187, 103, 66, 109, 180, 115, 171, 117, 191, 127, 117, 136, 118, 143, 8, 85, 188, 87, 146, 96, 57, 99, 39, 104, 58, 104, 45, 107, 101, 109, 115, 109, 202, 118, 163, 136, 10, 140, 109, 150, 218, 82, 246, 87, 248, 87, 165, 96, 156, 99, 185, 107, 112, 117, 91, 120, 155, 127, 202, 127, 204, 127, 31, 132, 50, 138, 51, 138, 89, 140, 91, 140, 56, 144, 244, 145, 191, 150, 70, 94, 97, 101, 121, 102, 237, 104, 148, 107, 89, 110, 50, 113, 225, 134, 77, 138, 199, 141, 252, 142, 32, 146, 174, 154, 132, 78, 127, 81, 15, 97, 162, 110, 72, 115, 236, 117, 106, 119, 233, 122, 34, 127, 169, 127, 132, 128, 212, 136, 219, 136, 99, 138, 233, 82, 213, 90, 217, 94, 143, 105, 105, 111, 23, 118, 137, 129, 250, 132, 52, 135, 126, 151, 197, 153, 4, 81, 142, 100, 248, 105, 197, 107, 60, 0, 42, 0, 160, 113, 164, 113, 188, 113, 30, 118, 188, 138, 82, 149, 93, 158, 98, 158, 211, 158, 147, 82, 27, 87, 191, 88, 17, 91, 31, 91, 167, 93, 182, 97, 204, 97, 192, 102, 170, 107, 186, 111, 218, 113, 49, 118, 150, 119, 83, 122, 10, 126, 87, 130, 143, 133, 160, 135, 57, 137, 241, 91, 129, 101, 206, 102, 141, 106, 93, 107, 225, 113, 241, 113, 243, 127, 252, 127, 198, 129, 150, 140, 249, 140, 168, 155, 84, 118, 217, 133, 221, 133, 0, 141, 176, 147, 113, 149, 118, 126, 121, 126, 119, 140, 44, 151, 227, 155, 130, 157, 131, 157, 141, 157, 55, 112, 25, 134, 111, 139, 112, 139, 179, 145, 183, 145, 80, 153, 200, 86, 63, 148, 193, 157, 202, 157, 255, 97, 124, 137, 91, 154, 231, 157, 73, 134, 254, 157, 155, 139, 120, 159, 60, 0, 42, 0, 217, 86, 224, 86, 37, 150, 52, 150, 140, 79, 148, 87, 251, 89, 7, 109, 53, 131, 107, 131, 243, 151, 131, 154, 54, 104, 183, 107, 36, 108, 112, 150, 208, 81, 245, 121, 192, 136, 223, 148, 123, 150, 130, 150, 145, 85, 25, 88, 99, 90, 20, 97, 67, 123, 106, 125, 69, 107, 181, 110, 139, 121, 145, 132, 45, 133, 71, 97, 22, 118, 166, 146, 248, 125, 135, 151, 242, 138, 18, 151, 240, 153, 126, 86, 201, 149, 32, 151, 254, 151, 60, 0, 42, 0, 152, 81, 81, 78, 31, 84, 190, 114, 194, 130, 166, 101, 206, 112, 160, 87, 255, 108, 1, 87, 254, 92, 250, 114, 226, 115, 118, 131, 20, 138, 26, 138, 108, 90, 197, 91, 31, 93, 47, 93, 235, 109, 33, 138, 246, 148, 29, 146, 130, 159, 219, 110, 146, 120, 30, 145, 36, 89, 41, 133, 128, 146, 86, 86, 165, 107, 140, 116, 190, 138, 154, 86, 173, 106, 235, 135, 42, 151, 87, 159, 227, 157, 60, 0, 42, 0, 90, 78, 244, 94, 57, 92, 21, 95, 50, 84, 110, 153, 147, 134, 122, 141, 144, 150, 254, 109, 15, 146, 242, 152, 160, 150, 119, 151, 238, 152, 4, 103, 17, 143, 228, 120, 155, 141, 131, 106, 62, 118, 177, 150, 190, 93, 229, 111, 230, 111, 190, 135, 31, 134, 253, 106, 110, 118, 148, 139, 60, 0, 42, 0, 112, 83, 26, 131, 21, 109, 228, 128, 189, 87, 55, 88, 90, 110, 12, 115, 213, 94, 115, 145, 109, 97, 74, 118, 150, 97, 151, 97, 163, 155, 218, 97, 188, 106, 60, 0, 42, 0, 140, 124, 60, 0, 42, 0, 148, 94, 220, 95, 241, 130, 64, 80, 92, 104, 241, 115, 186, 131, 104, 85, 116, 90, 150, 90, 37, 97, 54, 110, 236, 126, 32, 103, 80, 113, 91, 116, 200, 90, 164, 120, 51, 149, 36, 86, 132, 100, 7, 117, 211, 125, 40, 127, 66, 127, 103, 135, 207, 140, 49, 106, 142, 116, 67, 127, 46, 137, 72, 147, 25, 151, 44, 157, 102, 158, 48, 91, 201, 97, 186, 129, 250, 151, 22, 117, 112, 158, 145, 157, 167, 157, 182, 86, 70, 91, 126, 91, 22, 101, 76, 127, 33, 134, 77, 139, 251, 106, 212, 116, 47, 121, 123, 139, 175, 157, 77, 148, 147, 126, 51, 136, 234, 157, 200, 142, 249, 157, 14, 158, 26, 158, 60, 0, 42, 0, 193, 118, 206, 143, 20, 131, 200, 118, 103, 131, 185, 131, 182, 85, 36, 132, 37, 132, 38, 132, 205, 134, 129, 110, 139, 110, 62, 132, 204, 80, 75, 88, 121, 105, 226, 110, 229, 132, 70, 111, 146, 113, 71, 135, 105, 116, 127, 135, 52, 91, 223, 113, 8, 126, 162, 135, 217, 111, 218, 111, 244, 111, 192, 133, 174, 137, 13, 139, 98, 141, 5, 112, 163, 147, 13, 101, 27, 112, 32, 112, 47, 112, 5, 136, 255, 106, 52, 112, 15, 141, 80, 112, 93, 124, 92, 112, 111, 124, 60, 0, 42, 0, 232, 119, 226, 144, 103, 109, 172, 104, 141, 152, 149, 152, 150, 152, 108, 100, 113, 95, 65, 111, 63, 118, 78, 122, 52, 152, 202, 93, 238, 94, 27, 148, 109, 118, 60, 0, 42, 0, 32, 102, 142, 102, 108, 120, 181, 90, 161, 129, 95, 86, 149, 151, 102, 156, 60, 0, 42, 0, 223, 84, 55, 85, 178, 85, 60, 0, 42, 0, 99, 79, 229, 98, 200, 117, 149, 144, 184, 94, 173, 80, 200, 85, 24, 145, 205, 150, 137, 88, 222, 90, 117, 97, 253, 110, 230, 105, 69, 114, 112, 86, 197, 88, 193, 100, 173, 111, 250, 144, 91, 149, 195, 129, 85, 118, 221, 150, 222, 147, 217, 156, 241, 94, 73, 112, 84, 153, 69, 156, 219, 157, 112, 118, 60, 0, 42, 0, 129, 85, 216, 99, 153, 152, 82, 152, 43, 156, 60, 0, 42, 0, 56, 108, 44, 117, 143, 84, 58, 96, 243, 108, 209, 79, 199, 82, 200, 82, 16, 104, 199, 87, 128, 96, 225, 103, 140, 109, 127, 96, 155, 80, 229, 96, 17, 97, 103, 110, 103, 120, 96, 138, 78, 88, 113, 93, 110, 95, 57, 97, 249, 134, 66, 97, 10, 142, 156, 121, 172, 156, 52, 142, 210, 155, 60, 0, 42, 0, 40, 117, 218, 130, 61, 120, 159, 145, 60, 0, 42, 0, 24, 79, 231, 95, 56, 101, 102, 84, 46, 96, 209, 108, 125, 94, 12, 144, 160, 96, 128, 158, 250, 110, 130, 97, 42, 81, 62, 145, 152, 86, 0, 112, 204, 106, 139, 126, 48, 128, 60, 0, 42, 0, 34, 92, 36, 92, 49, 117, 139, 108, 185, 114, 174, 144, 185, 108, 172, 128, 35, 96, 191, 101, 163, 117, 243, 92, 95, 109, 222, 121, 156, 131, 184, 131, 245, 144, 192, 148, 100, 80, 176, 134, 39, 138, 48, 144, 56, 110, 54, 115, 74, 144, 127, 156, 98, 105, 55, 115, 62, 146, 137, 156, 15, 143, 192, 153, 85, 133, 99, 135, 119, 155, 54, 143, 139, 155, 254, 106, 60, 0, 42, 0, 203, 83, 9, 103, 35, 78, 99, 83, 195, 130, 73, 145, 145, 127, 174, 94, 47, 104, 144, 127, 160, 131, 132, 104, 72, 128, 28, 129, 213, 148, 117, 110, 137, 121, 15, 135, 170, 146, 241, 105, 86, 114, 87, 114, 221, 158, 238, 97, 60, 0, 42, 0, 200, 83, 243, 83, 124, 94, 81, 79, 145, 79, 103, 91, 214, 114, 255, 124, 202, 84, 255, 86, 247, 89, 165, 91, 223, 92, 218, 103, 112, 114, 80, 121, 241, 139, 246, 143, 0, 85, 180, 134, 180, 78, 129, 140, 201, 145, 109, 145, 152, 138, 44, 159, 60, 0, 42, 0, 143, 132, 60, 0, 42, 0, 188, 143, 92, 98, 161, 126, 194, 143, 195, 143, 123, 122, 83, 150, 6, 125, 118, 134, 57, 85, 228, 109, 211, 118, 0, 118, 138, 123, 60, 0, 42, 0, 144, 78, 142, 78, 152, 144, 3, 79, 89, 79, 164, 89, 117, 98, 69, 103, 36, 107, 151, 115, 153, 115, 188, 101, 194, 118, 254, 129, 103, 136, 124, 156, 123, 78, 222, 79, 106, 81, 186, 121, 253, 122, 1, 130, 48, 131, 27, 90, 47, 90, 49, 90, 89, 104, 243, 114, 0, 140, 81, 145, 128, 153, 20, 110, 56, 132, 133, 150, 233, 150, 90, 155, 35, 88, 44, 88, 51, 93, 78, 93, 91, 93, 9, 97, 196, 99, 112, 105, 29, 110, 97, 110, 109, 117, 98, 120, 116, 129, 46, 132, 62, 144, 172, 154, 26, 97, 213, 101, 97, 105, 134, 105, 72, 107, 79, 114, 92, 116, 69, 130, 94, 134, 206, 137, 1, 111, 110, 119, 172, 122, 6, 130, 21, 137, 118, 107, 173, 127, 77, 133, 83, 135, 219, 138, 211, 150, 24, 153, 41, 91, 158, 111, 166, 137, 48, 142, 95, 107, 181, 116, 184, 135, 63, 143, 93, 147, 35, 139, 195, 154, 189, 155, 223, 101, 69, 124, 31, 154, 242, 155, 27, 134, 5, 156, 224, 157, 6, 158, 60, 0, 42, 0, 14, 78, 136, 78, 27, 79, 135, 91, 127, 92, 189, 127, 232, 150, 193, 79, 227, 79, 39, 99, 185, 121, 237, 139, 4, 87, 255, 92, 100, 121, 74, 80, 44, 83, 9, 87, 190, 94, 84, 101, 5, 145, 158, 101, 45, 132, 180, 80, 217, 91, 64, 105, 64, 116, 16, 118, 7, 130, 158, 138, 179, 122, 217, 146, 40, 152, 137, 159, 115, 86, 188, 93, 217, 97, 144, 140, 148, 101, 140, 158, 12, 134, 108, 159, 60, 0, 42, 0, 128, 128, 137, 115, 109, 154, 43, 87, 127, 128, 139, 130, 140, 130, 170, 89, 236, 95, 107, 153, 178, 128, 193, 144, 49, 102, 241, 114, 215, 121, 31, 131, 252, 79, 234, 92, 103, 95, 116, 109, 33, 120, 176, 148, 132, 152, 144, 85, 223, 87, 9, 88, 134, 96, 208, 96, 50, 107, 226, 109, 239, 109, 172, 136, 21, 140, 51, 144, 8, 150, 133, 85, 169, 85, 187, 85, 128, 90, 211, 91, 189, 94, 161, 95, 219, 104, 220, 104, 235, 104, 52, 113, 25, 116, 222, 119, 114, 120, 213, 136, 71, 144, 235, 152, 173, 153, 70, 158, 8, 97, 234, 110, 92, 113, 34, 122, 109, 127, 72, 130, 174, 132, 227, 132, 137, 138, 58, 146, 16, 152, 215, 90, 142, 93, 43, 98, 211, 107, 68, 115, 9, 118, 206, 125, 31, 135, 46, 135, 13, 143, 137, 146, 74, 86, 126, 97, 79, 111, 54, 122, 249, 132, 129, 133, 107, 140, 121, 144, 202, 146, 255, 156, 166, 111, 207, 113, 224, 113, 119, 133, 237, 138, 37, 147, 190, 149, 37, 157, 39, 157, 42, 157, 37, 81, 7, 121, 166, 121, 74, 155, 108, 158, 82, 118, 22, 121, 28, 121, 101, 122, 253, 123, 88, 126, 167, 145, 82, 157, 242, 106, 71, 153, 125, 139, 93, 143, 45, 148, 49, 151, 14, 107, 72, 154, 59, 155, 94, 124, 74, 156, 248, 157, 18, 158, 29, 107, 165, 159, 201, 142, 48, 155, 49, 155, 106, 112, 60, 0, 42, 0, 114, 124, 41, 114, 60, 0, 42, 0, 154, 111, 60, 0, 42, 0, 230, 86, 34, 158, 72, 82, 164, 81, 129, 96, 34, 119, 51, 158, 195, 91, 1, 110, 6, 110, 10, 110, 21, 110, 204, 96, 245, 109, 126, 132, 233, 104, 172, 132, 14, 135, 247, 136, 83, 158, 162, 123, 246, 156, 53, 135, 204, 153, 27, 157, 61, 91, 119, 157, 65, 112, 24, 159, 29, 159, 60, 0, 42, 0, 67, 81, 134, 81, 32, 141, 167, 144, 88, 84, 237, 86, 133, 108, 108, 103, 163, 87, 48, 114, 159, 140, 159, 83, 225, 84, 6, 87, 14, 123, 150, 134, 129, 136, 161, 83, 91, 145, 14, 87, 244, 99, 114, 110, 40, 115, 24, 127, 18, 131, 40, 146, 11, 159, 18, 87, 19, 87, 108, 88, 180, 90, 196, 90, 144, 110, 146, 110, 63, 115, 66, 115, 157, 132, 158, 105, 172, 105, 149, 143, 227, 125, 1, 126, 93, 135, 111, 135, 109, 155, 124, 106, 177, 127, 151, 133, 136, 135, 28, 139, 69, 143, 255, 158, 177, 147, 222, 106, 141, 144, 53, 154, 162, 157, 176, 157, 181, 83, 60, 0, 42, 0, 220, 143, 246, 118, 58, 144, 96, 144, 250, 146, 60, 0, 42, 0, 23, 89, 153, 128, 180, 89, 209, 130, 40, 96, 98, 150, 184, 87, 79, 136, 134, 80, 155, 90, 190, 99, 87, 116, 144, 121, 63, 97, 235, 136, 17, 137, 36, 137, 110, 86, 88, 152, 60, 0, 42, 0, 240, 102, 241, 102, 166, 126, 4, 125, 185, 123, 241, 119, 95, 95, 96, 95, 60, 0, 42, 0, 8, 103, 9, 98, 22, 82, 156, 89, 132, 92, 136, 98, 63, 121, 179, 92, 165, 115, 113, 96, 133, 96, 166, 96, 142, 134, 143, 134, 207, 142, 186, 148, 5, 150, 115, 99, 192, 141, 195, 141, 164, 124, 138, 141, 5, 146, 181, 124, 94, 146, 177, 149, 178, 149, 51, 91, 62, 106, 215, 123, 189, 93, 160, 159, 70, 124, 57, 112, 37, 134, 230, 158, 26, 114, 180, 121, 141, 142, 101, 124, 17, 158, 112, 124, 25, 158, 60, 0, 42, 0, 85, 102, 10, 127, 128, 132, 136, 102, 50, 108, 116, 113, 149, 132, 51, 108, 107, 89, 121, 135, 95, 141, 53, 152, 167, 153, 7, 141, 60, 0, 42, 0, 145, 78, 251, 82, 0, 83, 29, 79, 233, 86, 152, 89, 132, 108, 173, 126, 184, 130, 0, 102, 71, 117, 3, 119, 208, 121, 231, 144, 162, 109, 28, 125, 24, 128, 58, 128, 22, 145, 242, 150, 42, 97, 179, 110, 96, 123, 124, 123, 183, 132, 137, 113, 144, 111, 83, 133, 198, 146, 82, 106, 212, 123, 28, 126, 60, 0, 42, 0, 65, 81, 45, 150, 61, 89, 142, 98, 193, 114, 104, 150, 122, 131, 146, 107, 151, 85, 23, 146, 149, 150, 158, 107, 30, 137, 187, 153, 210, 120, 241, 140, 35, 151, 107, 159, 115, 159, 60, 0, 42, 0, 85, 91, 208, 143, 159, 103, 211, 144, 125, 96, 6, 145, 93, 145, 138, 80, 242, 96, 32, 97, 75, 144, 77, 97, 106, 129, 235, 151, 245, 151, 133, 113, 168, 113, 247, 125, 252, 125, 116, 133, 21, 126, 128, 133, 150, 145, 158, 145, 43, 153, 244, 133, 215, 151, 222, 151, 10, 134, 251, 151, 60, 0, 42, 0, 163, 98, 103, 126, 60, 0, 39, 0, 208, 253, 90, 0, 39, 0, 60, 0, 42, 0, 0, 94, 29, 83, 158, 108, 202, 143, 130, 84, 246, 98, 37, 125, 46, 125, 84, 146, 115, 155, 220, 129, 226, 129, 60, 0, 42, 0, 66, 103, 175, 108, 56, 120, 122, 80, 146, 85, 244, 151, 209, 150, 252, 120, 77, 137, 220, 150, 203, 86, 208, 86, 229, 150, 60, 0, 42, 0, 139, 84, 60, 0, 42, 0, 125, 112, 126, 112, 62, 117, 201, 84, 61, 104, 214, 112, 209, 131, 61, 110, 168, 110, 117, 119, 243, 140, 60, 0, 42, 0, 176, 91, 61, 93, 60, 0, 42, 0, 141, 81, 40, 87, 87, 98, 5, 109, 125, 143, 164, 80, 9, 143, 104, 145, 14, 81, 33, 126, 60, 0, 42, 0, 66, 81, 204, 124, 42, 124, 46, 124, 21, 148, 31, 148, 60, 0, 42, 0, 177, 84, 60, 0, 42, 0, 29, 102, 90, 104, 193, 91, 221, 99, 70, 86, 141, 100, 39, 81, 5, 101, 18, 101, 57, 81, 34, 101, 177, 141, 178, 141, 60, 0, 42, 0, 130, 102, 171, 102, 219, 140, 94, 141, 62, 147, 60, 145, 253, 111, 84, 142, 201, 116, 10, 141, 232, 147, 210, 116, 71, 145, 82, 112, 131, 139, 218, 116, 182, 121, 120, 137, 154, 139, 97, 153, 60, 0, 42, 0, 40, 83, 66, 114, 152, 127, 67, 141, 205, 140, 231, 129, 53, 133, 216, 140, 19, 141, 210, 154, 28, 141, 60, 0, 42, 0, 117, 154, 212, 153, 60, 0, 42, 0, 88, 89, 9, 95, 15, 129, 95, 88, 108, 132, 186, 146, 211, 129, 223, 129, 60, 0, 42, 0, 174, 80, 109, 144, 223, 124, 103, 142, 169, 145, 60, 0, 42, 0, 255, 81, 127, 148, 60, 0, 42, 0, 233, 101, 163, 103, 6, 104, 164, 134, 215, 104, 161, 111, 170, 116, 187, 133, 112, 126, 251, 133, 60, 0, 42, 0, 118, 112, 129, 118, 130, 118, 21, 85, 35, 85, 32, 144, 141, 104, 191, 85, 101, 97, 112, 113, 65, 130, 106, 86, 9, 124, 229, 113, 195, 122, 95, 139, 174, 141, 129, 142, 200, 122, 60, 0, 42, 0, 25, 82, 158, 98, 162, 108, 233, 98, 206, 108, 253, 108, 35, 141, 238, 143, 71, 82, 54, 85, 103, 85, 59, 94, 46, 123, 52, 130, 172, 140, 173, 110, 224, 119, 22, 86, 231, 90, 88, 94, 166, 123, 54, 133, 13, 106, 117, 107, 206, 138, 92, 141, 199, 100, 164, 111, 159, 118, 148, 119, 0, 124, 11, 121, 87, 137, 46, 139, 254, 140, 12, 136, 90, 159, 112, 159, 5, 158, 60, 0, 42, 0, 40, 89, 196, 78, 130, 94, 68, 108, 3, 102, 23, 102, 81, 99, 49, 93, 60, 0, 42, 0, 44, 79, 60, 0, 42, 0, 60, 141, 29, 98, 202, 140, 151, 156, 253, 155, 8, 136, 2, 156, 97, 156, 60, 0, 42, 0, 14, 96, 60, 0, 42, 0, 46, 140, 86, 139, 91, 139, 60, 0, 42, 0, 206, 86, 60, 0, 42, 0, 151, 88, 43, 145, 158, 88, 142, 97, 47, 127, 103, 106, 183, 113, 148, 116, 240, 119, 243, 120, 126, 127, 82, 126, 68, 139, 91, 156, 60, 0, 42, 0, 3, 149, 229, 146, 17, 117, 96, 141, 8, 141, 60, 0, 42, 0, 78, 98, 18, 84, 175, 98, 83, 89, 19, 99, 228, 103, 251, 103, 243, 84, 103, 80, 179, 85, 248, 99, 35, 110, 66, 105, 132, 82, 99, 100, 182, 118, 29, 106, 240, 137, 187, 118, 71, 139, 68, 159, 71, 159, 60, 0, 42, 0, 45, 103, 52, 117, 248, 149, 187, 134, 225, 148, 96, 113, 80, 114, 152, 149, 154, 123, 43, 128, 88, 147, 87, 139, 60, 0, 42, 0, 143, 83, 195, 98, 242, 130, 40, 119, 31, 120, 41, 100, 138, 156, 157, 156, 55, 142, 147, 155, 186, 155, 60, 0, 42, 0, 77, 78, 121, 112, 200, 139, 164, 84, 222, 103, 5, 104, 184, 112, 177, 91, 196, 117, 177, 134, 160, 110, 80, 138, 62, 100, 168, 105, 5, 151, 161, 145, 60, 0, 42, 0, 8, 89, 130, 124, 90, 99, 139, 101, 142, 101, 88, 100, 184, 105, 75, 159, 60, 0, 42, 0, 133, 91, 161, 106, 60, 0, 42, 0, 132, 122, 89, 146, 60, 0, 42, 0, 58, 80, 38, 120, 181, 80, 232, 91, 53, 118, 60, 0, 42, 0, 172, 103, 190, 108, 225, 107, 195, 101, 52, 104, 152, 124, 197, 134, 230, 152, 201, 96, 64, 138, 136, 141, 121, 138, 154, 149, 53, 140, 97, 86, 166, 93, 157, 133, 133, 144, 17, 151, 8, 108, 10, 108, 187, 119, 111, 158, 220, 101, 107, 139, 88, 153, 227, 156, 89, 154, 89, 155, 99, 156, 7, 158, 157, 139, 60, 0, 42, 0, 169, 101, 208, 152, 85, 92, 207, 118, 45, 93, 172, 101, 22, 116, 12, 100, 222, 118, 131, 93, 132, 93, 144, 105, 173, 152, 248, 90, 134, 145, 79, 106, 62, 143, 245, 158, 60, 0, 42, 0, 96, 83, 84, 79, 24, 98, 8, 104, 95, 104, 217, 122, 97, 80, 253, 126, 218, 131, 231, 104, 91, 110, 38, 98, 187, 125, 152, 93, 26, 143, 48, 98, 101, 134, 102, 134, 177, 137, 79, 143, 103, 139, 56, 134, 79, 154, 60, 0, 42, 0, 32, 95, 53, 95, 224, 122, 189, 80, 35, 145, 220, 90, 112, 95, 94, 97, 51, 111, 80, 115, 187, 124, 1, 133, 103, 144, 178, 102, 31, 106, 139, 116, 38, 153, 209, 135, 63, 154, 70, 156, 158, 158, 60, 0, 42, 0, 201, 78, 127, 149, 119, 149, 168, 109, 140, 99, 50, 111, 101, 94, 3, 121, 60, 0, 42, 0, 8, 78, 215, 78, 89, 98, 16, 94, 86, 103, 192, 128, 38, 141, 128, 124, 51, 94, 177, 109, 57, 129, 238, 117, 156, 150, 135, 88, 130, 93, 91, 94, 236, 140, 44, 118, 52, 118, 149, 119, 60, 0, 42, 0, 241, 147, 60, 0, 42, 0, 75, 79, 138, 148, 177, 89, 246, 93, 219, 98, 45, 102, 196, 118, 215, 145, 65, 85, 74, 146, 203, 153, 188, 122, 99, 147, 189, 118, 60, 0, 42, 0, 42, 114, 43, 114, 126, 98, 188, 108, 117, 116, 60, 0, 42, 0, 236, 83, 70, 81, 207, 139, 155, 103, 151, 87, 164, 112, 227, 114, 117, 141, 10, 123, 129, 128, 208, 101, 249, 104, 84, 138, 103, 113, 105, 127, 135, 128, 136, 128, 153, 141, 204, 102, 243, 113, 161, 155, 194, 106, 190, 119, 132, 127, 60, 0, 42, 0, 64, 127, 60, 0, 42, 0, 7, 135, 251, 85, 236, 90, 110, 144, 60, 0, 42, 0, 135, 83, 152, 98, 125, 107, 250, 119, 19, 120, 119, 124, 116, 134, 242, 84, 209, 87, 141, 124, 169, 136, 96, 85, 138, 96, 98, 102, 99, 102, 132, 143, 134, 85, 240, 134, 95, 138, 42, 140, 178, 153, 122, 100, 18, 143, 212, 120, 25, 143, 184, 146, 153, 143, 196, 135, 158, 86, 43, 139, 58, 139, 191, 155, 77, 143, 129, 139, 139, 139, 60, 0, 42, 0, 5, 128, 125, 78, 107, 85, 135, 121, 23, 149, 109, 141, 54, 137, 117, 137, 60, 0, 42, 0, 217, 143, 216, 103, 89, 109, 25, 144, 219, 109, 23, 133, 28, 106, 106, 111, 103, 158, 197, 135, 211, 157, 60, 0, 42, 0, 64, 119, 87, 132, 60, 0, 42, 0, 30, 141, 136, 148, 166, 79, 72, 109, 205, 115, 206, 115, 215, 128, 158, 140, 42, 94, 21, 104, 98, 104, 30, 119, 31, 119, 39, 120, 111, 121, 221, 145, 117, 80, 109, 104, 89, 145, 202, 91, 116, 132, 73, 144, 195, 90, 56, 100, 159, 101, 104, 105, 73, 115, 4, 117, 142, 121, 150, 132, 193, 132, 65, 146, 85, 151, 155, 105, 157, 107, 103, 116, 170, 120, 155, 121, 103, 111, 180, 123, 60, 106, 181, 111, 251, 129, 189, 133, 49, 147, 67, 143, 124, 147, 72, 124, 117, 156, 60, 0, 42, 0, 202, 139, 174, 98, 149, 103, 43, 95, 35, 102, 120, 143, 82, 92, 91, 117, 185, 117, 21, 119, 151, 136, 62, 125, 68, 128, 25, 132, 214, 136, 58, 138, 235, 142, 28, 127, 57, 122, 215, 153, 37, 126, 18, 155, 240, 158, 60, 0, 42, 0, 51, 87, 53, 150, 188, 126, 61, 117, 178, 79, 11, 99, 99, 150, 41, 158, 47, 99, 21, 103, 26, 104, 22, 125, 57, 119, 72, 141, 86, 145, 102, 88, 213, 99, 124, 125, 118, 101, 113, 116, 171, 138, 209, 140, 244, 146, 71, 149, 7, 151, 6, 157, 173, 147, 174, 147, 60, 0, 42, 0, 231, 81, 137, 78, 66, 79, 195, 89, 129, 95, 20, 96, 45, 114, 228, 139, 233, 87, 229, 92, 35, 99, 161, 112, 240, 114, 221, 112, 16, 119, 178, 148, 29, 93, 34, 93, 153, 99, 25, 115, 65, 119, 71, 128, 238, 148, 156, 90, 193, 99, 93, 123, 176, 95, 184, 132, 92, 119, 45, 142, 102, 146, 180, 95, 143, 123, 26, 147, 181, 95, 220, 123, 7, 155, 214, 155, 101, 118, 60, 0, 42, 0, 54, 108, 141, 98, 253, 124, 239, 98, 159, 99, 120, 102, 56, 97, 156, 100, 116, 101, 60, 0, 42, 0, 99, 107, 193, 139, 209, 144, 39, 94, 63, 101, 199, 117, 64, 94, 60, 138, 99, 88, 205, 138, 45, 145, 10, 157, 73, 139, 60, 0, 42, 0, 75, 78, 47, 101, 110, 83, 65, 108, 157, 130, 49, 84, 245, 93, 101, 108, 103, 87, 157, 103, 220, 108, 229, 119, 199, 126, 162, 128, 0, 104, 87, 121, 211, 121, 214, 121, 209, 128, 221, 128, 124, 136, 1, 80, 183, 117, 108, 121, 234, 121, 2, 129, 187, 150, 148, 104, 32, 98, 37, 105, 248, 129, 24, 100, 148, 121, 25, 122, 149, 125, 176, 105, 24, 135, 182, 153, 247, 156, 50, 157, 84, 126, 53, 134, 5, 159, 60, 0, 42, 0, 103, 98, 132, 79, 183, 89, 244, 118, 234, 89, 36, 80, 60, 80, 64, 128, 222, 145, 244, 87, 247, 87, 212, 109, 76, 128, 13, 105, 150, 107, 134, 114, 131, 121, 119, 125, 214, 141, 237, 85, 225, 116, 68, 146, 140, 88, 109, 100, 189, 153, 2, 91, 121, 97, 16, 111, 47, 142, 52, 106, 177, 129, 54, 126, 119, 128, 217, 135, 96, 142, 196, 142, 145, 142, 60, 0, 42, 0, 2, 89, 98, 107, 234, 83, 234, 81, 167, 82, 232, 101, 47, 150, 64, 87, 65, 87, 11, 94, 122, 98, 102, 108, 154, 108, 184, 126, 183, 130, 62, 96, 167, 98, 73, 121, 11, 131, 171, 84, 73, 96, 7, 99, 179, 103, 20, 109, 11, 120, 121, 136, 117, 143, 253, 109, 187, 117, 25, 125, 40, 138, 190, 141, 249, 142, 249, 158, 111, 145, 226, 133, 103, 137, 60, 0, 42, 0, 36, 150, 243, 129, 150, 130, 215, 95, 238, 95, 123, 98, 120, 140, 54, 82, 148, 83, 129, 87, 25, 94, 28, 94, 187, 108, 153, 112, 40, 141, 227, 143, 197, 144, 231, 79, 217, 92, 162, 94, 164, 94, 3, 99, 227, 103, 9, 104, 55, 109, 81, 121, 95, 150, 33, 90, 143, 95, 26, 99, 74, 102, 78, 104, 254, 114, 233, 121, 244, 129, 159, 136, 61, 141, 126, 143, 127, 78, 107, 80, 157, 95, 183, 99, 189, 104, 86, 105, 24, 115, 100, 117, 212, 117, 242, 121, 247, 121, 146, 122, 41, 125, 208, 127, 160, 136, 215, 137, 173, 140, 218, 148, 55, 158, 130, 80, 59, 93, 88, 95, 122, 102, 222, 110, 227, 117, 237, 134, 253, 142, 152, 154, 216, 91, 204, 94, 49, 100, 205, 110, 26, 122, 107, 123, 110, 127, 241, 141, 10, 143, 39, 149, 201, 150, 134, 88, 239, 110, 76, 111, 144, 117, 253, 136, 159, 137, 140, 138, 141, 146, 95, 94, 132, 97, 111, 100, 171, 113, 58, 122, 163, 129, 239, 137, 234, 140, 44, 142, 213, 146, 243, 100, 216, 101, 4, 112, 251, 125, 178, 150, 228, 153, 25, 157, 40, 81, 149, 82, 229, 97, 242, 100, 219, 106, 73, 122, 178, 135, 235, 97, 4, 141, 205, 106, 198, 116, 246, 137, 45, 154, 239, 155, 41, 121, 81, 140, 58, 154, 71, 154, 147, 142, 217, 157, 85, 148, 82, 140, 60, 0, 42, 0, 148, 95, 60, 0, 42, 0, 45, 78, 0, 79, 119, 108, 35, 82, 144, 89, 120, 95, 224, 95, 200, 108, 130, 112, 200, 126, 202, 103, 197, 118, 115, 136, 159, 148, 47, 130, 119, 136, 66, 125, 33, 146, 82, 94, 32, 133, 58, 149, 191, 146, 164, 135, 36, 157, 189, 135, 126, 147, 40, 159, 113, 142, 24, 148, 102, 124, 60, 0, 42, 0, 191, 128, 205, 121, 162, 81, 160, 85, 48, 92, 90, 88, 92, 88, 113, 107, 68, 113, 107, 129, 7, 118, 46, 122, 53, 142, 92, 122, 60, 0, 42, 0, 242, 78, 23, 79, 149, 89, 198, 114, 76, 121, 61, 131, 118, 136, 205, 145, 155, 134, 69, 80, 62, 119, 57, 88, 145, 90, 87, 123, 70, 136, 229, 138, 60, 0, 42, 0, 222, 93, 31, 130, 204, 139, 156, 79, 104, 84, 50, 109, 191, 112, 234, 139, 208, 112, 216, 115, 128, 143, 238, 144, 159, 95, 171, 99, 205, 109, 234, 119, 49, 144, 60, 158, 140, 85, 165, 124, 82, 141, 8, 143, 130, 146, 217, 140, 22, 143, 12, 151, 242, 153, 233, 118, 5, 139, 67, 157, 6, 154, 120, 139, 60, 0, 42, 0, 175, 89, 116, 143, 248, 142, 60, 0, 42, 0, 152, 128, 26, 94, 155, 117, 247, 131, 109, 102, 109, 119, 146, 123, 222, 155, 60, 0, 42, 0, 163, 126, 55, 79, 106, 84, 146, 84, 153, 91, 201, 126, 145, 81, 174, 84, 60, 102, 2, 125, 196, 128, 110, 131, 177, 118, 78, 145, 93, 102, 153, 124, 100, 132, 75, 138, 3, 117, 118, 138, 253, 80, 186, 118, 206, 153, 99, 86, 16, 126, 164, 154, 64, 124, 85, 124, 82, 124, 95, 154, 60, 0, 42, 0, 49, 103, 175, 82, 143, 79, 219, 139, 190, 144, 25, 109, 49, 131, 42, 104, 224, 115, 248, 139, 42, 115, 67, 120, 252, 121, 190, 136, 226, 148, 81, 125, 219, 134, 133, 138, 230, 141, 224, 105, 116, 111, 107, 135, 150, 146, 101, 106, 248, 138, 108, 140, 239, 153, 162, 155, 56, 157, 38, 112, 235, 106, 231, 106, 250, 155, 4, 159, 41, 136, 60, 0, 42, 0, 249, 122, 207, 108, 250, 122, 162, 112, 1, 123, 63, 131, 219, 112, 139, 122, 16, 144, 28, 123, 51, 130, 3, 118, 201, 123, 237, 113, 11, 136, 133, 142, 65, 156, 78, 91, 95, 112, 239, 102, 24, 107, 37, 114, 62, 136, 131, 148, 60, 0, 42, 0, 54, 78, 59, 78, 148, 91, 196, 98, 92, 127, 124, 150, 26, 110, 110, 113, 81, 113, 93, 138, 49, 86, 208, 111, 136, 158, 169, 119, 154, 82, 209, 86, 184, 101, 218, 119, 60, 0, 42, 0, 43, 79, 71, 79, 79, 79, 169, 82, 187, 126, 206, 130, 126, 87, 124, 103, 232, 108, 46, 141, 236, 143, 123, 154, 244, 88, 241, 103, 182, 107, 183, 112, 93, 121, 176, 117, 29, 119, 43, 120, 105, 121, 218, 122, 135, 131, 53, 125, 56, 125, 156, 127, 192, 134, 64, 93, 81, 123, 59, 138, 175, 140, 211, 141, 244, 142, 248, 148, 111, 123, 82, 146, 181, 153, 184, 88, 184, 123, 229, 127, 38, 106, 243, 146, 208, 153, 235, 123, 20, 151, 134, 158, 68, 148, 60, 0, 42, 0, 147, 98, 155, 106, 188, 129, 59, 124, 253, 154, 60, 0, 42, 0, 253, 98, 60, 0, 42, 0, 233, 141, 60, 0, 42, 0, 19, 78, 192, 83, 2, 92, 22, 120, 8, 92, 31, 145, 124, 88, 229, 90, 124, 116, 14, 117, 215, 120, 158, 129, 155, 152, 218, 120, 239, 138, 228, 135, 83, 152, 68, 156, 60, 0, 42, 0, 108, 143, 104, 91, 226, 142, 241, 122, 73, 143, 60, 0, 42, 0, 119, 112, 109, 85, 31, 88, 131, 132, 81, 116, 94, 129, 206, 80, 90, 141, 176, 100, 198, 123, 148, 153, 249, 123, 72, 137, 250, 140, 84, 139, 76, 153, 192, 86, 81, 124, 60, 0, 42, 0, 134, 89, 132, 94, 157, 89, 146, 94, 88, 131, 36, 90, 105, 104, 138, 131, 137, 104, 119, 110, 167, 124, 197, 136, 221, 136, 1, 106, 218, 124, 60, 0, 42, 0, 238, 88, 239, 88, 182, 114, 192, 114, 245, 88, 11, 113, 52, 111, 158, 100, 7, 98, 60, 0, 42, 0, 185, 150, 253, 143, 147, 154, 37, 149, 16, 147, 5, 154, 123, 157, 60, 0, 42, 0, 157, 108, 60, 0, 42, 0, 96, 87, 88, 104, 13, 123, 55, 90, 0, 127, 244, 96, 0, 117, 18, 127, 119, 117, 126, 120, 135, 129, 156, 88, 180, 125, 88, 141, 11, 126, 200, 138, 138, 145, 35, 147, 31, 153, 8, 121, 5, 141, 91, 143, 70, 148, 60, 0, 42, 0, 146, 91, 205, 143, 171, 128, 128, 122, 6, 140, 196, 138, 96, 136, 60, 0, 42, 0, 198, 81, 251, 87, 214, 81, 150, 110, 167, 125, 60, 0, 42, 0, 48, 138, 21, 122, 60, 0, 42, 0, 83, 83, 217, 98, 170, 112, 44, 80, 73, 99, 76, 104, 193, 104, 191, 109, 243, 104, 213, 105, 91, 122, 113, 122, 63, 136, 60, 0, 42, 0, 52, 87, 116, 95, 75, 108, 179, 114, 124, 112, 213, 83, 176, 89, 1, 131, 171, 101, 74, 109, 53, 78, 94, 109, 245, 112, 252, 139, 76, 145, 68, 85, 69, 85, 58, 90, 178, 104, 177, 101, 107, 102, 19, 105, 56, 116, 122, 120, 161, 122, 108, 127, 175, 100, 198, 100, 178, 101, 154, 121, 133, 82, 193, 138, 209, 138, 220, 146, 193, 111, 231, 123, 226, 100, 128, 101, 181, 101, 239, 111, 225, 106, 54, 139, 111, 149, 47, 148, 107, 157, 66, 112, 23, 136, 50, 148, 87, 124, 223, 157, 113, 124, 60, 0, 42, 0, 167, 122, 60, 0, 42, 0, 114, 78, 92, 91, 10, 131, 121, 81, 168, 84, 213, 89, 255, 89, 50, 131, 37, 104, 134, 115, 14, 125, 64, 141, 68, 141, 48, 93, 196, 109, 246, 121, 1, 127, 24, 140, 124, 141, 222, 85, 115, 91, 107, 93, 20, 105, 125, 110, 203, 110, 162, 124, 88, 132, 142, 143, 17, 145, 118, 91, 140, 121, 220, 137, 62, 138, 178, 140, 199, 140, 145, 141, 49, 149, 53, 122, 199, 125, 45, 146, 67, 149, 135, 159, 28, 143, 18, 159, 172, 111, 238, 138, 166, 141, 58, 143, 25, 147, 237, 154, 187, 156, 127, 147, 161, 147, 190, 116, 63, 152, 62, 152, 212, 155, 133, 157, 77, 159, 38, 156, 60, 0, 42, 0, 251, 132, 60, 0, 42, 0, 212, 78, 7, 84, 201, 89, 202, 89, 77, 103, 247, 119, 196, 121, 207, 128, 112, 84, 237, 121, 125, 124, 20, 128, 120, 134, 43, 123, 147, 104, 232, 145, 89, 85, 43, 125, 211, 110, 63, 138, 159, 105, 116, 106, 60, 0, 42, 0, 87, 91, 234, 129, 147, 130, 33, 131, 98, 131, 51, 80, 90, 82, 99, 96, 120, 114, 13, 110, 37, 119, 38, 119, 212, 128, 254, 128, 44, 111, 60, 0, 42, 0, 80, 91, 60, 0, 42, 0, 151, 91, 39, 80, 252, 126, 148, 154, 43, 88, 79, 93, 85, 93, 254, 96, 213, 104, 35, 115, 89, 129, 124, 132, 54, 105, 33, 103, 120, 93, 47, 122, 156, 125, 195, 125, 167, 113, 245, 125, 234, 127, 108, 135, 40, 142, 42, 142, 235, 120, 80, 147, 117, 140, 100, 142, 12, 154, 3, 155, 35, 154, 9, 155, 55, 155, 238, 155, 252, 155, 65, 148, 60, 0, 42, 0, 59, 96, 108, 80, 116, 99, 227, 96, 33, 97, 212, 99, 3, 100, 175, 80, 215, 132, 96, 100, 207, 125, 2, 126, 234, 113, 61, 126, 211, 147, 60, 0, 42, 0, 181, 126, 46, 102, 173, 117, 10, 80, 20, 115, 130, 120, 189, 124, 201, 124, 50, 118, 38, 126, 29, 147, 49, 126, 237, 124, 60, 0, 42, 0, 72, 111, 60, 0, 42, 0, 185, 144, 122, 154, 249, 139, 240, 144, 108, 150, 198, 131, 247, 104, 248, 104, 18, 145, 131, 123, 197, 125, 207, 138, 57, 145, 176, 156, 235, 155, 192, 158, 54, 154, 113, 159, 122, 159, 60, 0, 42, 0, 113, 141, 112, 141, 208, 155, 60, 0, 42, 0, 79, 89, 205, 99, 113, 105, 60, 0, 42, 0, 223, 121, 69, 132, 169, 132, 60, 0, 42, 0, 70, 83, 179, 141, 82, 83, 235, 84, 18, 93, 42, 93, 207, 101, 182, 80, 164, 123, 36, 142, 63, 142, 94, 149, 195, 147, 60, 0, 42, 0, 197, 139, 59, 150, 196, 126, 206, 79, 60, 114, 199, 115, 86, 121, 40, 85, 68, 125, 91, 138, 123, 151, 186, 147, 60, 0, 42, 0, 187, 148, 156, 142, 125, 148, 60, 0, 42, 0, 100, 126, 53, 127, 130, 126, 137, 126, 107, 124, 152, 126, 60, 0, 42, 0, 37, 101, 90, 148, 60, 0, 42, 0, 156, 83, 24, 103, 250, 85, 54, 106, 213, 135, 151, 126, 60, 0, 42, 0, 138, 93, 52, 86, 181, 93, 127, 86, 187, 116, 60, 0, 42, 0, 44, 104, 74, 125, 84, 145, 0, 103, 108, 102, 125, 121, 33, 122, 106, 127, 160, 143, 220, 105, 123, 145, 94, 133, 137, 145, 135, 106, 247, 146, 10, 147, 140, 106, 60, 0, 42, 0, 160, 103, 93, 122, 60, 0, 42, 0, 10, 92, 171, 88, 255, 88, 159, 93, 117, 144, 61, 106, 92, 126, 71, 127, 142, 157, 15, 148, 223, 156, 82, 156, 247, 157, 60, 0, 42, 0, 212, 80, 66, 86, 153, 100, 80, 139, 60, 0, 42, 0, 88, 99, 140, 146, 60, 0, 42, 0, 40, 102, 232, 121, 139, 131, 125, 99, 10, 105, 34, 116, 19, 122, 112, 123, 60, 146, 60, 0, 42, 0, 230, 93, 80, 79, 83, 126, 60, 0, 42, 0, 92, 79, 80, 87, 60, 150, 157, 92, 158, 92, 13, 96, 179, 79, 90, 121, 217, 128, 17, 85, 167, 94, 143, 136, 90, 80, 67, 132, 68, 132, 217, 132, 245, 152, 243, 124, 60, 0, 42, 0, 151, 84, 38, 0, 39, 0, 40, 0, 0, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 32, 50, 38, 0, 39, 0, 40, 0, 3, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 38, 50, 38, 0, 39, 0, 40, 0, 9, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 34, 50, 38, 0, 39, 0, 40, 0, 93, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 40, 50, 38, 0, 39, 0, 40, 0, 140, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 33, 50, 38, 0, 39, 0, 40, 0, 148, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 36, 50, 38, 0, 39, 0, 40, 0, 227, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 57, 50, 38, 0, 39, 0, 40, 0, 1, 79, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 61, 50, 38, 0, 39, 0, 40, 0, 17, 79, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 65, 50, 38, 0, 39, 0, 40, 0, 107, 81, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 39, 50, 38, 0, 39, 0, 40, 0, 109, 81, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 37, 50, 38, 0, 39, 0, 40, 0, 180, 82, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 56, 50, 38, 0, 39, 0, 40, 0, 65, 83, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 41, 50, 38, 0, 39, 0, 40, 0, 84, 83, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 63, 50, 38, 0, 39, 0, 40, 0, 13, 84, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 52, 50, 38, 0, 39, 0, 40, 0, 124, 84, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 58, 50, 38, 0, 39, 0, 40, 0, 219, 86, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 35, 50, 38, 0, 39, 0, 40, 0, 31, 87, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 47, 50, 38, 0, 39, 0, 40, 0, 102, 91, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 59, 50, 38, 0, 39, 0, 40, 0, 229, 101, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 48, 50, 38, 0, 39, 0, 40, 0, 8, 103, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 42, 50, 38, 0, 39, 0, 40, 0, 9, 103, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 50, 50, 38, 0, 39, 0, 40, 0, 40, 103, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 45, 50, 38, 0, 39, 0, 40, 0, 42, 104, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 49, 50, 38, 0, 39, 0, 40, 0, 52, 108, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 44, 50, 38, 0, 39, 0, 40, 0, 107, 112, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 43, 50, 38, 0, 39, 0, 40, 0, 121, 114, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 53, 50, 38, 0, 39, 0, 40, 0, 227, 118, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 50, 38, 0, 39, 0, 40, 0, 62, 121, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 51, 50, 38, 0, 39, 0, 40, 0, 93, 121, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 55, 50, 38, 0, 39, 0, 40, 0, 109, 121, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 64, 50, 38, 0, 39, 0, 40, 0, 234, 129, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 66, 50, 38, 0, 39, 0, 40, 0, 243, 129, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 67, 50, 38, 0, 39, 0, 40, 0, 161, 140, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 54, 50, 38, 0, 39, 0, 40, 0, 199, 140, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 62, 50, 38, 0, 39, 0, 40, 0, 209, 145, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 46, 50, 38, 0, 48, 0, 185, 112, 60, 0, 60, 0, 60, 0, 88, 51, 38, 0, 49, 0, 48, 0, 229, 101, 60, 0, 60, 0, 60, 0, 233, 51, 38, 0, 49, 0, 48, 0, 8, 103, 60, 0, 60, 0, 60, 0, 201, 50, 38, 0, 49, 0, 48, 0, 185, 112, 60, 0, 60, 0, 60, 0, 98, 51, 38, 0, 49, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 234, 51, 38, 0, 49, 0, 49, 0, 8, 103, 60, 0, 60, 0, 60, 0, 202, 50, 38, 0, 49, 0, 49, 0, 185, 112, 60, 0, 60, 0, 60, 0, 99, 51, 38, 0, 49, 0, 50, 0, 229, 101, 60, 0, 60, 0, 60, 0, 235, 51, 38, 0, 49, 0, 50, 0, 8, 103, 60, 0, 60, 0, 60, 0, 203, 50, 38, 0, 49, 0, 50, 0, 185, 112, 60, 0, 60, 0, 60, 0, 100, 51, 38, 0, 49, 0, 51, 0, 229, 101, 60, 0, 60, 0, 60, 0, 236, 51, 38, 0, 49, 0, 51, 0, 185, 112, 60, 0, 60, 0, 60, 0, 101, 51, 38, 0, 49, 0, 52, 0, 229, 101, 60, 0, 60, 0, 60, 0, 237, 51, 38, 0, 49, 0, 52, 0, 185, 112, 60, 0, 60, 0, 60, 0, 102, 51, 38, 0, 49, 0, 53, 0, 229, 101, 60, 0, 60, 0, 60, 0, 238, 51, 38, 0, 49, 0, 53, 0, 185, 112, 60, 0, 60, 0, 60, 0, 103, 51, 38, 0, 49, 0, 54, 0, 229, 101, 60, 0, 60, 0, 60, 0, 239, 51, 38, 0, 49, 0, 54, 0, 185, 112, 60, 0, 60, 0, 60, 0, 104, 51, 38, 0, 49, 0, 55, 0, 229, 101, 60, 0, 60, 0, 60, 0, 240, 51, 38, 0, 49, 0, 55, 0, 185, 112, 60, 0, 60, 0, 60, 0, 105, 51, 38, 0, 49, 0, 56, 0, 229, 101, 60, 0, 60, 0, 60, 0, 241, 51, 38, 0, 49, 0, 56, 0, 185, 112, 60, 0, 60, 0, 60, 0, 106, 51, 38, 0, 49, 0, 57, 0, 229, 101, 60, 0, 60, 0, 60, 0, 242, 51, 38, 0, 49, 0, 57, 0, 185, 112, 60, 0, 60, 0, 60, 0, 107, 51, 38, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 224, 51, 38, 0, 49, 0, 8, 103, 60, 0, 60, 0, 60, 0, 192, 50, 38, 0, 49, 0, 185, 112, 60, 0, 60, 0, 60, 0, 89, 51, 38, 0, 50, 0, 48, 0, 229, 101, 60, 0, 60, 0, 60, 0, 243, 51, 38, 0, 50, 0, 48, 0, 185, 112, 60, 0, 60, 0, 60, 0, 108, 51, 38, 0, 50, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 244, 51, 38, 0, 50, 0, 49, 0, 185, 112, 60, 0, 60, 0, 60, 0, 109, 51, 38, 0, 50, 0, 50, 0, 229, 101, 60, 0, 60, 0, 60, 0, 245, 51, 38, 0, 50, 0, 50, 0, 185, 112, 60, 0, 60, 0, 60, 0, 110, 51, 38, 0, 50, 0, 51, 0, 229, 101, 60, 0, 60, 0, 60, 0, 246, 51, 38, 0, 50, 0, 51, 0, 185, 112, 60, 0, 60, 0, 60, 0, 111, 51, 38, 0, 50, 0, 52, 0, 229, 101, 60, 0, 60, 0, 60, 0, 247, 51, 38, 0, 50, 0, 52, 0, 185, 112, 60, 0, 60, 0, 60, 0, 112, 51, 38, 0, 50, 0, 53, 0, 229, 101, 60, 0, 60, 0, 60, 0, 248, 51, 38, 0, 50, 0, 54, 0, 229, 101, 60, 0, 60, 0, 60, 0, 249, 51, 38, 0, 50, 0, 55, 0, 229, 101, 60, 0, 60, 0, 60, 0, 250, 51, 38, 0, 50, 0, 56, 0, 229, 101, 60, 0, 60, 0, 60, 0, 251, 51, 38, 0, 50, 0, 57, 0, 229, 101, 60, 0, 60, 0, 60, 0, 252, 51, 38, 0, 50, 0, 229, 101, 60, 0, 60, 0, 60, 0, 225, 51, 38, 0, 50, 0, 8, 103, 60, 0, 60, 0, 60, 0, 193, 50, 38, 0, 50, 0, 185, 112, 60, 0, 60, 0, 60, 0, 90, 51, 38, 0, 51, 0, 48, 0, 229, 101, 60, 0, 60, 0, 60, 0, 253, 51, 38, 0, 51, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 254, 51, 38, 0, 51, 0, 229, 101, 60, 0, 60, 0, 60, 0, 226, 51, 38, 0, 51, 0, 8, 103, 60, 0, 60, 0, 60, 0, 194, 50, 38, 0, 51, 0, 185, 112, 60, 0, 60, 0, 60, 0, 91, 51, 38, 0, 52, 0, 229, 101, 60, 0, 60, 0, 60, 0, 227, 51, 38, 0, 52, 0, 8, 103, 60, 0, 60, 0, 60, 0, 195, 50, 38, 0, 52, 0, 185, 112, 60, 0, 60, 0, 60, 0, 92, 51, 38, 0, 53, 0, 229, 101, 60, 0, 60, 0, 60, 0, 228, 51, 38, 0, 53, 0, 8, 103, 60, 0, 60, 0, 60, 0, 196, 50, 38, 0, 53, 0, 185, 112, 60, 0, 60, 0, 60, 0, 93, 51, 38, 0, 54, 0, 229, 101, 60, 0, 60, 0, 60, 0, 229, 51, 38, 0, 54, 0, 8, 103, 60, 0, 60, 0, 60, 0, 197, 50, 38, 0, 54, 0, 185, 112, 60, 0, 60, 0, 60, 0, 94, 51, 38, 0, 55, 0, 229, 101, 60, 0, 60, 0, 60, 0, 230, 51, 38, 0, 55, 0, 8, 103, 60, 0, 60, 0, 60, 0, 198, 50, 38, 0, 55, 0, 185, 112, 60, 0, 60, 0, 60, 0, 95, 51, 38, 0, 56, 0, 229, 101, 60, 0, 60, 0, 60, 0, 231, 51, 38, 0, 56, 0, 8, 103, 60, 0, 60, 0, 60, 0, 199, 50, 38, 0, 56, 0, 185, 112, 60, 0, 60, 0, 60, 0, 96, 51, 38, 0, 57, 0, 229, 101, 60, 0, 60, 0, 60, 0, 232, 51, 38, 0, 57, 0, 8, 103, 60, 0, 60, 0, 60, 0, 200, 50, 38, 0, 57, 0, 185, 112, 60, 0, 60, 0, 60, 0, 97, 51, 38, 0, 39, 0, 20, 48, 9, 78, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 65, 222, 38, 0, 39, 0, 20, 48, 140, 78, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 66, 222, 38, 0, 39, 0, 20, 48, 221, 82, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 71, 222, 38, 0, 39, 0, 20, 48, 137, 91, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 67, 222, 38, 0, 39, 0, 20, 48, 83, 98, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 69, 222, 38, 0, 39, 0, 20, 48, 87, 101, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 72, 222, 38, 0, 39, 0, 20, 48, 44, 103, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 64, 222, 38, 0, 39, 0, 20, 48, 185, 112, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 68, 222, 38, 0, 39, 0, 20, 48, 215, 118, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 70, 222, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 0, 47, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 146, 49, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 128, 50, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 60, 216, 41, 222, 38, 0, 1, 78, 60, 0, 60, 0, 60, 0, 156, 49, 38, 0, 3, 78, 60, 0, 60, 0, 60, 0, 134, 50, 38, 0, 9, 78, 60, 0, 60, 0, 60, 0, 148, 49, 38, 0, 9, 78, 60, 0, 60, 0, 60, 0, 130, 50, 38, 0, 9, 78, 60, 0, 60, 0, 60, 0, 60, 216, 42, 222, 38, 0, 10, 78, 60, 0, 60, 0, 60, 0, 150, 49, 38, 0, 10, 78, 60, 0, 60, 0, 60, 0, 164, 50, 38, 0, 11, 78, 60, 0, 60, 0, 60, 0, 152, 49, 38, 0, 11, 78, 60, 0, 60, 0, 60, 0, 166, 50, 38, 0, 25, 78, 60, 0, 60, 0, 60, 0, 155, 49, 38, 0, 40, 78, 60, 0, 60, 0, 60, 0, 1, 47, 38, 0, 45, 78, 60, 0, 60, 0, 60, 0, 151, 49, 38, 0, 45, 78, 60, 0, 60, 0, 60, 0, 165, 50, 38, 0, 45, 78, 60, 0, 60, 0, 60, 0, 60, 216, 45, 222, 38, 0, 54, 78, 60, 0, 60, 0, 60, 0, 2, 47, 38, 0, 63, 78, 60, 0, 60, 0, 60, 0, 3, 47, 38, 0, 89, 78, 60, 0, 60, 0, 60, 0, 4, 47, 38, 0, 89, 78, 60, 0, 60, 0, 60, 0, 154, 49, 38, 0, 93, 78, 60, 0, 60, 0, 60, 0, 136, 50, 38, 0, 133, 78, 60, 0, 60, 0, 60, 0, 5, 47, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 6, 47, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 147, 49, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 129, 50, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 60, 216, 20, 222, 38, 0, 148, 78, 60, 0, 60, 0, 60, 0, 132, 50, 38, 0, 160, 78, 60, 0, 60, 0, 60, 0, 7, 47, 38, 0, 164, 78, 60, 0, 60, 0, 60, 0, 60, 216, 24, 222, 38, 0, 186, 78, 60, 0, 60, 0, 60, 0, 8, 47, 38, 0, 186, 78, 60, 0, 60, 0, 60, 0, 159, 49, 38, 0, 228, 78, 140, 84, 60, 0, 60, 0, 60, 0, 255, 50, 38, 0, 1, 79, 60, 0, 60, 0, 60, 0, 173, 50, 38, 0, 17, 79, 60, 0, 60, 0, 60, 0, 161, 50, 38, 0, 42, 81, 60, 0, 60, 0, 60, 0, 157, 50, 38, 0, 63, 81, 60, 0, 60, 0, 60, 0, 9, 47, 38, 0, 101, 81, 60, 0, 60, 0, 60, 0, 10, 47, 38, 0, 107, 81, 60, 0, 60, 0, 60, 0, 11, 47, 38, 0, 107, 81, 60, 0, 60, 0, 60, 0, 135, 50, 38, 0, 109, 81, 60, 0, 60, 0, 60, 0, 133, 50, 38, 0, 130, 81, 60, 0, 60, 0, 60, 0, 12, 47, 38, 0, 141, 81, 60, 0, 60, 0, 60, 0, 60, 216, 30, 222, 38, 0, 150, 81, 60, 0, 60, 0, 60, 0, 13, 47, 38, 0, 153, 81, 60, 0, 60, 0, 60, 0, 162, 50, 38, 0, 171, 81, 60, 0, 60, 0, 60, 0, 14, 47, 38, 0, 224, 81, 60, 0, 60, 0, 60, 0, 15, 47, 38, 0, 245, 81, 60, 0, 60, 0, 60, 0, 16, 47, 38, 0, 0, 82, 60, 0, 60, 0, 60, 0, 17, 47, 38, 0, 29, 82, 60, 0, 60, 0, 60, 0, 60, 216, 32, 222, 38, 0, 77, 82, 60, 0, 60, 0, 60, 0, 60, 216, 28, 222, 38, 0, 114, 82, 60, 0, 60, 0, 60, 0, 60, 216, 57, 222, 38, 0, 155, 82, 60, 0, 60, 0, 60, 0, 18, 47, 38, 0, 180, 82, 60, 0, 60, 0, 60, 0, 152, 50, 38, 0, 249, 82, 60, 0, 60, 0, 60, 0, 19, 47, 38, 0, 21, 83, 60, 0, 60, 0, 60, 0, 20, 47, 38, 0, 26, 83, 60, 0, 60, 0, 60, 0, 21, 47, 38, 0, 56, 83, 60, 0, 60, 0, 60, 0, 22, 47, 38, 0, 59, 83, 60, 0, 60, 0, 60, 0, 169, 50, 38, 0, 65, 83, 60, 0, 60, 0, 60, 0, 23, 47, 38, 0, 65, 83, 60, 0, 60, 0, 60, 0, 56, 48, 38, 0, 65, 83, 60, 0, 60, 0, 60, 0, 137, 50, 38, 0, 68, 83, 60, 0, 60, 0, 60, 0, 57, 48, 38, 0, 69, 83, 60, 0, 60, 0, 60, 0, 58, 48, 38, 0, 84, 83, 60, 0, 60, 0, 60, 0, 175, 50, 38, 0, 92, 83, 60, 0, 60, 0, 60, 0, 24, 47, 38, 0, 105, 83, 60, 0, 60, 0, 60, 0, 25, 47, 38, 0, 112, 83, 60, 0, 60, 0, 60, 0, 158, 50, 38, 0, 130, 83, 60, 0, 60, 0, 60, 0, 26, 47, 38, 0, 182, 83, 60, 0, 60, 0, 60, 0, 27, 47, 38, 0, 200, 83, 60, 0, 60, 0, 60, 0, 28, 47, 38, 0, 204, 83, 60, 0, 60, 0, 60, 0, 60, 216, 18, 222, 38, 0, 227, 83, 60, 0, 60, 0, 60, 0, 29, 47, 38, 0, 239, 83, 60, 0, 60, 0, 60, 0, 60, 216, 81, 222, 38, 0, 243, 83, 60, 0, 60, 0, 60, 0, 168, 50, 38, 0, 243, 83, 60, 0, 60, 0, 60, 0, 60, 216, 46, 222, 38, 0, 8, 84, 60, 0, 60, 0, 60, 0, 60, 216, 52, 222, 38, 0, 13, 84, 60, 0, 60, 0, 60, 0, 148, 50, 38, 0, 57, 84, 60, 0, 60, 0, 60, 0, 60, 216, 37, 222, 38, 0, 79, 85, 60, 0, 60, 0, 60, 0, 68, 50, 38, 0, 182, 85, 60, 0, 60, 0, 60, 0, 60, 216, 58, 222, 38, 0, 215, 86, 60, 0, 60, 0, 60, 0, 30, 47, 38, 0, 219, 86, 60, 0, 60, 0, 60, 0, 149, 49, 38, 0, 219, 86, 60, 0, 60, 0, 60, 0, 131, 50, 38, 0, 31, 87, 60, 0, 60, 0, 60, 0, 31, 47, 38, 0, 31, 87, 60, 0, 60, 0, 60, 0, 143, 50, 38, 0, 48, 87, 60, 0, 60, 0, 60, 0, 158, 49, 38, 0, 235, 88, 60, 0, 60, 0, 60, 0, 32, 47, 38, 0, 240, 88, 60, 0, 60, 0, 60, 0, 60, 216, 36, 222, 38, 0, 2, 89, 60, 0, 60, 0, 60, 0, 33, 47, 38, 0, 10, 89, 60, 0, 60, 0, 60, 0, 34, 47, 38, 0, 21, 89, 60, 0, 60, 0, 60, 0, 35, 47, 38, 0, 26, 89, 60, 0, 60, 0, 60, 0, 60, 216, 21, 222, 38, 0, 28, 89, 60, 0, 60, 0, 60, 0, 176, 50, 38, 0, 39, 89, 60, 0, 60, 0, 60, 0, 36, 47, 38, 0, 39, 89, 99, 107, 60, 0, 60, 0, 60, 0, 125, 51, 38, 0, 41, 89, 60, 0, 60, 0, 60, 0, 157, 49, 38, 0, 41, 89, 60, 0, 60, 0, 60, 0, 60, 216, 23, 222, 38, 0, 115, 89, 60, 0, 60, 0, 60, 0, 37, 47, 38, 0, 115, 89, 60, 0, 60, 0, 60, 0, 155, 50, 38, 0, 80, 91, 60, 0, 60, 0, 60, 0, 38, 47, 38, 0, 87, 91, 60, 0, 60, 0, 60, 0, 60, 216, 17, 222, 38, 0, 102, 91, 60, 0, 60, 0, 60, 0, 171, 50, 38, 0, 128, 91, 60, 0, 60, 0, 60, 0, 39, 47, 38, 0, 151, 91, 60, 0, 60, 0, 60, 0, 170, 50, 38, 0, 248, 91, 60, 0, 60, 0, 60, 0, 40, 47, 38, 0, 15, 92, 60, 0, 60, 0, 60, 0, 41, 47, 38, 0, 34, 92, 60, 0, 60, 0, 60, 0, 42, 47, 38, 0, 56, 92, 60, 0, 60, 0, 60, 0, 43, 47, 38, 0, 110, 92, 60, 0, 60, 0, 60, 0, 44, 47, 38, 0, 113, 92, 60, 0, 60, 0, 60, 0, 45, 47, 38, 0, 219, 93, 60, 0, 60, 0, 60, 0, 46, 47, 38, 0, 229, 93, 60, 0, 60, 0, 60, 0, 47, 47, 38, 0, 230, 93, 60, 0, 60, 0, 60, 0, 167, 50, 38, 0, 230, 93, 60, 0, 60, 0, 60, 0, 60, 216, 44, 222, 38, 0, 241, 93, 60, 0, 60, 0, 60, 0, 48, 47, 38, 0, 254, 93, 60, 0, 60, 0, 60, 0, 49, 47, 38, 0, 114, 94, 60, 0, 60, 0, 60, 0, 50, 47, 38, 0, 115, 94, 16, 98, 60, 0, 60, 0, 60, 0, 123, 51, 38, 0, 122, 94, 60, 0, 60, 0, 60, 0, 51, 47, 38, 0, 124, 94, 60, 0, 60, 0, 60, 0, 69, 50, 38, 0, 127, 94, 60, 0, 60, 0, 60, 0, 52, 47, 38, 0, 244, 94, 60, 0, 60, 0, 60, 0, 53, 47, 38, 0, 254, 94, 60, 0, 60, 0, 60, 0, 54, 47, 38, 0, 11, 95, 60, 0, 60, 0, 60, 0, 55, 47, 38, 0, 19, 95, 60, 0, 60, 0, 60, 0, 56, 47, 38, 0, 80, 95, 60, 0, 60, 0, 60, 0, 57, 47, 38, 0, 97, 95, 60, 0, 60, 0, 60, 0, 58, 47, 38, 0, 115, 95, 60, 0, 60, 0, 60, 0, 59, 47, 38, 0, 140, 95, 60, 0, 60, 0, 60, 0, 60, 216, 29, 222, 38, 0, 151, 95, 60, 0, 60, 0, 60, 0, 60, 216, 80, 222, 38, 0, 195, 95, 60, 0, 60, 0, 60, 0, 60, 47, 38, 0, 8, 98, 60, 0, 60, 0, 60, 0, 61, 47, 38, 0, 54, 98, 60, 0, 60, 0, 60, 0, 62, 47, 38, 0, 75, 98, 60, 0, 60, 0, 60, 0, 63, 47, 38, 0, 75, 98, 60, 0, 60, 0, 60, 0, 60, 216, 16, 222, 38, 0, 83, 98, 60, 0, 60, 0, 60, 0, 60, 216, 49, 222, 38, 0, 149, 98, 60, 0, 60, 0, 60, 0, 60, 216, 39, 222, 38, 0, 7, 99, 60, 0, 60, 0, 60, 0, 60, 216, 47, 222, 38, 0, 85, 99, 60, 0, 60, 0, 60, 0, 60, 216, 40, 222, 38, 0, 47, 101, 60, 0, 60, 0, 60, 0, 64, 47, 38, 0, 52, 101, 60, 0, 60, 0, 60, 0, 65, 47, 38, 0, 135, 101, 60, 0, 60, 0, 60, 0, 66, 47, 38, 0, 135, 101, 60, 0, 60, 0, 60, 0, 70, 50, 38, 0, 151, 101, 60, 0, 60, 0, 60, 0, 67, 47, 38, 0, 153, 101, 60, 0, 60, 0, 60, 0, 60, 216, 27, 222, 38, 0, 164, 101, 60, 0, 60, 0, 60, 0, 68, 47, 38, 0, 176, 101, 60, 0, 60, 0, 60, 0, 60, 216, 31, 222, 38, 0, 185, 101, 60, 0, 60, 0, 60, 0, 69, 47, 38, 0, 224, 101, 60, 0, 60, 0, 60, 0, 70, 47, 38, 0, 229, 101, 60, 0, 60, 0, 60, 0, 71, 47, 38, 0, 229, 101, 60, 0, 60, 0, 60, 0, 144, 50, 38, 0, 14, 102, 187, 108, 60, 0, 60, 0, 60, 0, 126, 51, 38, 0, 32, 102, 60, 0, 60, 0, 60, 0, 60, 216, 25, 222, 38, 0, 45, 102, 140, 84, 60, 0, 60, 0, 60, 0, 124, 51, 38, 0, 240, 102, 60, 0, 60, 0, 60, 0, 72, 47, 38, 0, 8, 103, 60, 0, 60, 0, 60, 0, 73, 47, 38, 0, 8, 103, 60, 0, 60, 0, 60, 0, 138, 50, 38, 0, 8, 103, 60, 0, 60, 0, 60, 0, 60, 216, 55, 222, 38, 0, 9, 103, 60, 0, 60, 0, 60, 0, 146, 50, 38, 0, 9, 103, 60, 0, 60, 0, 60, 0, 60, 216, 54, 222, 38, 0, 40, 103, 60, 0, 60, 0, 60, 0, 74, 47, 38, 0, 40, 103, 60, 0, 60, 0, 60, 0, 141, 50, 38, 0, 42, 104, 60, 0, 60, 0, 60, 0, 145, 50, 38, 0, 42, 104, 15, 95, 26, 79, 62, 121, 60, 0, 60, 0, 60, 0, 127, 51, 38, 0, 32, 107, 60, 0, 60, 0, 60, 0, 75, 47, 38, 0, 98, 107, 60, 0, 60, 0, 60, 0, 76, 47, 38, 0, 99, 107, 60, 0, 60, 0, 60, 0, 163, 50, 38, 0, 121, 107, 60, 0, 60, 0, 60, 0, 77, 47, 38, 0, 179, 107, 60, 0, 60, 0, 60, 0, 78, 47, 38, 0, 203, 107, 60, 0, 60, 0, 60, 0, 79, 47, 38, 0, 205, 107, 60, 0, 60, 0, 60, 0, 159, 46, 38, 0, 212, 107, 60, 0, 60, 0, 60, 0, 80, 47, 38, 0, 219, 107, 60, 0, 60, 0, 60, 0, 81, 47, 38, 0, 15, 108, 60, 0, 60, 0, 60, 0, 82, 47, 38, 0, 20, 108, 60, 0, 60, 0, 60, 0, 83, 47, 38, 0, 52, 108, 60, 0, 60, 0, 60, 0, 84, 47, 38, 0, 52, 108, 60, 0, 60, 0, 60, 0, 140, 50, 38, 0, 232, 108, 60, 0, 60, 0, 60, 0, 159, 50, 38, 0, 128, 110, 60, 0, 60, 0, 60, 0, 60, 216, 53, 222, 38, 0, 20, 111, 60, 0, 60, 0, 60, 0, 60, 216, 38, 222, 38, 0, 107, 112, 60, 0, 60, 0, 60, 0, 85, 47, 38, 0, 107, 112, 60, 0, 60, 0, 60, 0, 139, 50, 38, 0, 33, 113, 60, 0, 60, 0, 60, 0, 60, 216, 26, 222, 38, 0, 42, 114, 60, 0, 60, 0, 60, 0, 86, 47, 38, 0, 54, 114, 60, 0, 60, 0, 60, 0, 87, 47, 38, 0, 59, 114, 60, 0, 60, 0, 60, 0, 88, 47, 38, 0, 63, 114, 60, 0, 60, 0, 60, 0, 89, 47, 38, 0, 71, 114, 60, 0, 60, 0, 60, 0, 90, 47, 38, 0, 89, 114, 60, 0, 60, 0, 60, 0, 91, 47, 38, 0, 91, 114, 60, 0, 60, 0, 60, 0, 92, 47, 38, 0, 121, 114, 60, 0, 60, 0, 60, 0, 149, 50, 38, 0, 172, 114, 60, 0, 60, 0, 60, 0, 93, 47, 38, 0, 132, 115, 60, 0, 60, 0, 60, 0, 94, 47, 38, 0, 137, 115, 60, 0, 60, 0, 60, 0, 95, 47, 38, 0, 220, 116, 60, 0, 60, 0, 60, 0, 96, 47, 38, 0, 230, 116, 60, 0, 60, 0, 60, 0, 97, 47, 38, 0, 24, 117, 60, 0, 60, 0, 60, 0, 98, 47, 38, 0, 31, 117, 60, 0, 60, 0, 60, 0, 99, 47, 38, 0, 31, 117, 60, 0, 60, 0, 60, 0, 60, 216, 34, 222, 38, 0, 40, 117, 60, 0, 60, 0, 60, 0, 100, 47, 38, 0, 48, 117, 60, 0, 60, 0, 60, 0, 101, 47, 38, 0, 50, 117, 60, 0, 60, 0, 60, 0, 153, 49, 38, 0, 51, 117, 60, 0, 60, 0, 60, 0, 60, 216, 56, 222, 38, 0, 55, 117, 60, 0, 60, 0, 60, 0, 154, 50, 38, 0, 139, 117, 60, 0, 60, 0, 60, 0, 102, 47, 38, 0, 146, 117, 60, 0, 60, 0, 60, 0, 103, 47, 38, 0, 118, 118, 60, 0, 60, 0, 60, 0, 104, 47, 38, 0, 125, 118, 60, 0, 60, 0, 60, 0, 105, 47, 38, 0, 174, 118, 60, 0, 60, 0, 60, 0, 106, 47, 38, 0, 191, 118, 60, 0, 60, 0, 60, 0, 107, 47, 38, 0, 227, 118, 60, 0, 60, 0, 60, 0, 172, 50, 38, 0, 238, 118, 60, 0, 60, 0, 60, 0, 108, 47, 38, 0, 219, 119, 60, 0, 60, 0, 60, 0, 109, 47, 38, 0, 226, 119, 60, 0, 60, 0, 60, 0, 110, 47, 38, 0, 243, 119, 60, 0, 60, 0, 60, 0, 111, 47, 38, 0, 58, 121, 60, 0, 60, 0, 60, 0, 112, 47, 38, 0, 62, 121, 60, 0, 60, 0, 60, 0, 147, 50, 38, 0, 93, 121, 60, 0, 60, 0, 60, 0, 151, 50, 38, 0, 129, 121, 60, 0, 60, 0, 60, 0, 60, 216, 50, 222, 38, 0, 184, 121, 60, 0, 60, 0, 60, 0, 113, 47, 38, 0, 190, 121, 60, 0, 60, 0, 60, 0, 114, 47, 38, 0, 216, 121, 60, 0, 60, 0, 60, 0, 153, 50, 38, 0, 116, 122, 60, 0, 60, 0, 60, 0, 115, 47, 38, 0, 122, 122, 60, 0, 60, 0, 60, 0, 60, 216, 51, 222, 38, 0, 203, 122, 60, 0, 60, 0, 60, 0, 116, 47, 38, 0, 249, 122, 60, 0, 60, 0, 60, 0, 117, 47, 38, 0, 143, 123, 60, 0, 60, 0, 60, 0, 71, 50, 38, 0, 115, 124, 60, 0, 60, 0, 60, 0, 118, 47, 38, 0, 248, 124, 60, 0, 60, 0, 60, 0, 119, 47, 38, 0, 66, 125, 60, 0, 60, 0, 60, 0, 60, 216, 33, 222, 38, 0, 54, 127, 60, 0, 60, 0, 60, 0, 120, 47, 38, 0, 81, 127, 60, 0, 60, 0, 60, 0, 121, 47, 38, 0, 138, 127, 60, 0, 60, 0, 60, 0, 122, 47, 38, 0, 189, 127, 60, 0, 60, 0, 60, 0, 123, 47, 38, 0, 1, 128, 60, 0, 60, 0, 60, 0, 124, 47, 38, 0, 12, 128, 60, 0, 60, 0, 60, 0, 125, 47, 38, 0, 18, 128, 60, 0, 60, 0, 60, 0, 126, 47, 38, 0, 51, 128, 60, 0, 60, 0, 60, 0, 127, 47, 38, 0, 127, 128, 60, 0, 60, 0, 60, 0, 128, 47, 38, 0, 137, 128, 60, 0, 60, 0, 60, 0, 129, 47, 38, 0, 227, 129, 60, 0, 60, 0, 60, 0, 130, 47, 38, 0, 234, 129, 60, 0, 60, 0, 60, 0, 131, 47, 38, 0, 243, 129, 60, 0, 60, 0, 60, 0, 132, 47, 38, 0, 252, 129, 60, 0, 60, 0, 60, 0, 133, 47, 38, 0, 12, 130, 60, 0, 60, 0, 60, 0, 134, 47, 38, 0, 27, 130, 60, 0, 60, 0, 60, 0, 135, 47, 38, 0, 31, 130, 60, 0, 60, 0, 60, 0, 136, 47, 38, 0, 110, 130, 60, 0, 60, 0, 60, 0, 137, 47, 38, 0, 114, 130, 60, 0, 60, 0, 60, 0, 138, 47, 38, 0, 120, 130, 60, 0, 60, 0, 60, 0, 139, 47, 38, 0, 77, 134, 60, 0, 60, 0, 60, 0, 140, 47, 38, 0, 107, 134, 60, 0, 60, 0, 60, 0, 141, 47, 38, 0, 64, 136, 60, 0, 60, 0, 60, 0, 142, 47, 38, 0, 76, 136, 60, 0, 60, 0, 60, 0, 143, 47, 38, 0, 99, 136, 60, 0, 60, 0, 60, 0, 144, 47, 38, 0, 126, 137, 60, 0, 60, 0, 60, 0, 145, 47, 38, 0, 139, 137, 60, 0, 60, 0, 60, 0, 146, 47, 38, 0, 210, 137, 60, 0, 60, 0, 60, 0, 147, 47, 38, 0, 227, 137, 60, 0, 60, 0, 60, 0, 60, 216, 22, 222, 38, 0, 0, 138, 60, 0, 60, 0, 60, 0, 148, 47, 38, 0, 55, 140, 60, 0, 60, 0, 60, 0, 149, 47, 38, 0, 70, 140, 60, 0, 60, 0, 60, 0, 150, 47, 38, 0, 85, 140, 60, 0, 60, 0, 60, 0, 151, 47, 38, 0, 120, 140, 60, 0, 60, 0, 60, 0, 152, 47, 38, 0, 157, 140, 60, 0, 60, 0, 60, 0, 153, 47, 38, 0, 161, 140, 60, 0, 60, 0, 60, 0, 150, 50, 38, 0, 169, 140, 60, 0, 60, 0, 60, 0, 60, 216, 35, 222, 38, 0, 199, 140, 60, 0, 60, 0, 60, 0, 174, 50, 38, 0, 100, 141, 60, 0, 60, 0, 60, 0, 154, 47, 38, 0, 112, 141, 60, 0, 60, 0, 60, 0, 155, 47, 38, 0, 112, 141, 60, 0, 60, 0, 60, 0, 60, 216, 48, 222, 38, 0, 179, 141, 60, 0, 60, 0, 60, 0, 156, 47, 38, 0, 171, 142, 60, 0, 60, 0, 60, 0, 157, 47, 38, 0, 202, 142, 60, 0, 60, 0, 60, 0, 158, 47, 38, 0, 155, 143, 60, 0, 60, 0, 60, 0, 159, 47, 38, 0, 176, 143, 60, 0, 60, 0, 60, 0, 160, 47, 38, 0, 181, 143, 60, 0, 60, 0, 60, 0, 161, 47, 38, 0, 74, 144, 60, 0, 60, 0, 60, 0, 60, 216, 43, 222, 38, 0, 105, 144, 60, 0, 60, 0, 60, 0, 156, 50, 38, 0, 145, 144, 60, 0, 60, 0, 60, 0, 162, 47, 38, 0, 73, 145, 60, 0, 60, 0, 60, 0, 163, 47, 38, 0, 77, 145, 60, 0, 60, 0, 60, 0, 60, 216, 59, 222, 38, 0, 198, 145, 60, 0, 60, 0, 60, 0, 164, 47, 38, 0, 204, 145, 60, 0, 60, 0, 60, 0, 165, 47, 38, 0, 209, 145, 60, 0, 60, 0, 60, 0, 166, 47, 38, 0, 209, 145, 60, 0, 60, 0, 60, 0, 142, 50, 38, 0, 119, 149, 60, 0, 60, 0, 60, 0, 167, 47, 38, 0, 128, 149, 60, 0, 60, 0, 60, 0, 168, 47, 38, 0, 28, 150, 60, 0, 60, 0, 60, 0, 169, 47, 38, 0, 182, 150, 60, 0, 60, 0, 60, 0, 170, 47, 38, 0, 185, 150, 60, 0, 60, 0, 60, 0, 171, 47, 38, 0, 232, 150, 60, 0, 60, 0, 60, 0, 172, 47, 38, 0, 81, 151, 60, 0, 60, 0, 60, 0, 173, 47, 38, 0, 94, 151, 60, 0, 60, 0, 60, 0, 174, 47, 38, 0, 98, 151, 60, 0, 60, 0, 60, 0, 175, 47, 38, 0, 105, 151, 60, 0, 60, 0, 60, 0, 176, 47, 38, 0, 203, 151, 60, 0, 60, 0, 60, 0, 177, 47, 38, 0, 237, 151, 60, 0, 60, 0, 60, 0, 178, 47, 38, 0, 243, 151, 60, 0, 60, 0, 60, 0, 179, 47, 38, 0, 1, 152, 60, 0, 60, 0, 60, 0, 180, 47, 38, 0, 5, 152, 60, 0, 60, 0, 60, 0, 160, 50, 38, 0, 168, 152, 60, 0, 60, 0, 60, 0, 181, 47, 38, 0, 219, 152, 60, 0, 60, 0, 60, 0, 182, 47, 38, 0, 223, 152, 60, 0, 60, 0, 60, 0, 183, 47, 38, 0, 150, 153, 60, 0, 60, 0, 60, 0, 184, 47, 38, 0, 153, 153, 60, 0, 60, 0, 60, 0, 185, 47, 38, 0, 172, 153, 60, 0, 60, 0, 60, 0, 186, 47, 38, 0, 168, 154, 60, 0, 60, 0, 60, 0, 187, 47, 38, 0, 216, 154, 60, 0, 60, 0, 60, 0, 188, 47, 38, 0, 223, 154, 60, 0, 60, 0, 60, 0, 189, 47, 38, 0, 37, 155, 60, 0, 60, 0, 60, 0, 190, 47, 38, 0, 47, 155, 60, 0, 60, 0, 60, 0, 191, 47, 38, 0, 50, 155, 60, 0, 60, 0, 60, 0, 192, 47, 38, 0, 60, 155, 60, 0, 60, 0, 60, 0, 193, 47, 38, 0, 90, 155, 60, 0, 60, 0, 60, 0, 194, 47, 38, 0, 229, 156, 60, 0, 60, 0, 60, 0, 195, 47, 38, 0, 117, 158, 60, 0, 60, 0, 60, 0, 196, 47, 38, 0, 127, 158, 60, 0, 60, 0, 60, 0, 197, 47, 38, 0, 165, 158, 60, 0, 60, 0, 60, 0, 198, 47, 38, 0, 187, 158, 60, 0, 60, 0, 60, 0, 199, 47, 38, 0, 195, 158, 60, 0, 60, 0, 60, 0, 200, 47, 38, 0, 205, 158, 60, 0, 60, 0, 60, 0, 201, 47, 38, 0, 209, 158, 60, 0, 60, 0, 60, 0, 202, 47, 38, 0, 249, 158, 60, 0, 60, 0, 60, 0, 203, 47, 38, 0, 253, 158, 60, 0, 60, 0, 60, 0, 204, 47, 38, 0, 14, 159, 60, 0, 60, 0, 60, 0, 205, 47, 38, 0, 19, 159, 60, 0, 60, 0, 60, 0, 206, 47, 38, 0, 32, 159, 60, 0, 60, 0, 60, 0, 207, 47, 38, 0, 59, 159, 60, 0, 60, 0, 60, 0, 208, 47, 38, 0, 74, 159, 60, 0, 60, 0, 60, 0, 209, 47, 38, 0, 82, 159, 60, 0, 60, 0, 60, 0, 210, 47, 38, 0, 141, 159, 60, 0, 60, 0, 60, 0, 211, 47, 38, 0, 156, 159, 60, 0, 60, 0, 60, 0, 212, 47, 38, 0, 159, 159, 60, 0, 60, 0, 60, 0, 243, 46, 38, 0, 160, 159, 60, 0, 60, 0, 60, 0, 213, 47, 38, 0, 107, 134, 60, 0, 205, 145, 134, 94, 47, 0, 134, 94, 38, 0, 30, 95, 60, 0, 136, 108, 51, 150, 47, 0, 51, 150, 38, 0, 186, 146, 60, 0, 207, 133, 135, 101, 47, 0, 135, 101, 0, 0, 239, 223, 208, 112, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 122, 0, 104, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 112, 0, 105, 0, 110, 0, 121, 0, 105, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 72, 0, 97, 0, 110, 0, 105, 0, 32, 0, 66, 0, 111, 0, 112, 0, 111, 0, 93, 0, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 112, 0, 114, 0, 105, 0, 109, 0, 97, 0, 114, 0, 121, 0, 32, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 97, 0, 98, 0, 108, 0, 101, 0, 93, 0, 60, 0, 60, 0, 201, 2, 60, 0, 60, 0, 202, 2, 60, 0, 60, 0, 199, 2, 60, 0, 60, 0, 203, 2, 60, 0, 60, 0, 217, 2, 60, 0, 60, 0, 234, 2, 60, 0, 60, 0, 235, 2, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 114, 0, 101, 0, 103, 0, 117, 0, 108, 0, 97, 0, 114, 0, 93, 0, 60, 0, 39, 0, 208, 253, 39, 0, 5, 49, 60, 0, 42, 0, 107, 81, 200, 78, 244, 93, 237, 83, 82, 98, 51, 103, 144, 115, 63, 89, 156, 92, 173, 130, 199, 92, 237, 103, 164, 117, 245, 84, 252, 93, 76, 99, 6, 123, 145, 124, 147, 127, 134, 134, 219, 145, 223, 145, 109, 59, 93, 140, 131, 156, 62, 76, 60, 0, 42, 0, 208, 83, 174, 114, 156, 98, 122, 87, 173, 89, 212, 98, 166, 112, 121, 118, 200, 128, 7, 131, 221, 131, 89, 138, 203, 141, 247, 142, 176, 152, 162, 88, 67, 155, 37, 159, 60, 0, 42, 0, 138, 98, 175, 148, 0, 146, 118, 151, 60, 0, 42, 0, 93, 87, 29, 95, 56, 114, 187, 87, 98, 127, 25, 128, 193, 141, 133, 156, 140, 156, 119, 127, 172, 67, 138, 155, 135, 137, 242, 119, 56, 151, 233, 88, 94, 112, 27, 107, 60, 0, 42, 0, 39, 84, 38, 125, 94, 155, 60, 0, 42, 0, 118, 118, 23, 94, 232, 98, 226, 108, 119, 118, 187, 115, 93, 82, 101, 82, 241, 84, 203, 118, 53, 120, 181, 148, 125, 153, 52, 125, 61, 127, 154, 136, 176, 136, 224, 131, 134, 120, 98, 146, 224, 80, 147, 93, 165, 100, 173, 100, 17, 153, 129, 155, 115, 142, 75, 154, 77, 156, 60, 0, 42, 0, 226, 78, 47, 79, 187, 114, 145, 128, 115, 154, 27, 94, 219, 114, 221, 116, 188, 79, 195, 82, 201, 128, 233, 130, 179, 78, 44, 99, 97, 109, 223, 116, 225, 121, 227, 144, 185, 148, 194, 148, 141, 58, 191, 61, 64, 62, 14, 63, 22, 129, 54, 130, 175, 136, 90, 83, 36, 110, 185, 136, 65, 158, 61, 97, 15, 100, 60, 115, 103, 132, 56, 146, 81, 146, 142, 153, 240, 80, 13, 59, 127, 113, 84, 114, 148, 123, 138, 129, 74, 130, 111, 75, 155, 153, 193, 153, 20, 133, 35, 142, 205, 146, 72, 149, 79, 70, 101, 216, 208, 220, 158, 153, 238, 153, 112, 140, 96, 216, 109, 223, 151, 86, 234, 97, 103, 58, 39, 59, 25, 124, 79, 137, 155, 147, 58, 153, 83, 157, 166, 114, 33, 121, 198, 154, 201, 154, 2, 107, 52, 121, 110, 137, 110, 148, 60, 0, 42, 0, 219, 141, 165, 123, 56, 124, 60, 0, 42, 0, 121, 91, 151, 106, 234, 124, 82, 139, 23, 134, 60, 0, 42, 0, 138, 46, 92, 83, 117, 85, 33, 132, 60, 0, 42, 0, 0, 99, 176, 99, 216, 100, 60, 0, 42, 0, 125, 118, 60, 0, 42, 0, 126, 118, 112, 79, 207, 103, 34, 104, 109, 99, 248, 116, 168, 124, 84, 125, 70, 100, 250, 100, 108, 137, 60, 0, 42, 0, 141, 94, 221, 98, 37, 141, 220, 98, 87, 101, 8, 115, 23, 122, 253, 134, 186, 124, 37, 53, 1, 141, 219, 151, 60, 0, 42, 0, 225, 122, 173, 133, 60, 0, 42, 0, 81, 83, 111, 103, 66, 150, 195, 118, 110, 104, 178, 96, 249, 99, 17, 105, 145, 120, 134, 121, 78, 158, 3, 147, 227, 133, 111, 157, 60, 0, 42, 0, 23, 83, 51, 57, 115, 146, 60, 0, 42, 0, 29, 141, 91, 91, 200, 114, 157, 140, 7, 89, 1, 102, 108, 114, 182, 144, 204, 128, 221, 130, 161, 148, 251, 79, 13, 80, 221, 54, 150, 96, 253, 114, 229, 144, 93, 80, 121, 80, 242, 58, 150, 104, 252, 115, 171, 136, 153, 80, 235, 96, 25, 113, 50, 116, 240, 142, 136, 143, 1, 145, 195, 80, 2, 97, 215, 62, 154, 120, 34, 62, 149, 114, 211, 132, 150, 138, 129, 151, 25, 137, 41, 143, 199, 146, 179, 154, 138, 97, 210, 124, 180, 151, 62, 148, 60, 0, 42, 0, 87, 84, 4, 85, 153, 121, 60, 0, 42, 0, 249, 82, 5, 83, 98, 91, 185, 103, 222, 128, 222, 130, 35, 123, 114, 113, 133, 159, 18, 137, 84, 133, 193, 149, 67, 137, 89, 159, 60, 0, 42, 0, 135, 122, 209, 90, 249, 150, 15, 66, 132, 133, 90, 216, 138, 220, 250, 63, 60, 0, 42, 0, 157, 91, 9, 96, 113, 153, 221, 79, 40, 158, 178, 91, 228, 115, 33, 88, 34, 88, 172, 90, 218, 91, 70, 132, 253, 152, 194, 153, 245, 156, 229, 125, 19, 137, 7, 157, 242, 140, 75, 75, 243, 91, 246, 91, 99, 216, 38, 220, 76, 151, 16, 77, 99, 216, 59, 220, 60, 0, 42, 0, 253, 82, 165, 98, 177, 98, 121, 140, 181, 141, 199, 148, 171, 134, 140, 136, 49, 88, 226, 131, 75, 146, 141, 156, 100, 151, 180, 102, 178, 154, 241, 154, 99, 134, 145, 155, 36, 81, 211, 102, 6, 114, 193, 95, 100, 148, 20, 158, 60, 0, 42, 0, 104, 79, 245, 133, 60, 0, 42, 0, 115, 98, 61, 101, 237, 115, 44, 130, 129, 152, 145, 101, 44, 100, 146, 101, 18, 152, 34, 118, 251, 156, 50, 66, 140, 135, 41, 137, 77, 118, 172, 143, 60, 0, 42, 0, 66, 87, 133, 92, 42, 150, 4, 102, 127, 103, 72, 114, 234, 116, 163, 148, 132, 124, 40, 130, 17, 146, 66, 135, 108, 155, 198, 149, 60, 0, 42, 0, 158, 82, 74, 83, 52, 79, 110, 98, 98, 87, 197, 89, 17, 96, 204, 98, 202, 126, 200, 103, 218, 121, 116, 110, 70, 125, 97, 146, 125, 151, 166, 143, 227, 116, 60, 0, 42, 0, 129, 135, 60, 0, 42, 0, 84, 89, 205, 108, 50, 141, 31, 104, 135, 114, 27, 149, 27, 147, 60, 0, 42, 0, 44, 103, 239, 130, 89, 89, 90, 117, 201, 127, 77, 105, 60, 0, 42, 0, 75, 87, 76, 87, 52, 80, 121, 99, 115, 104, 0, 110, 40, 123, 90, 216, 192, 223, 41, 144, 170, 100, 86, 115, 61, 143, 60, 0, 42, 0, 166, 144, 185, 87, 46, 94, 96, 99, 92, 109, 134, 104, 171, 144, 71, 94, 90, 94, 13, 126, 107, 94, 164, 151, 60, 0, 42, 0, 209, 126, 129, 125, 156, 105, 83, 114, 128, 129, 200, 154, 60, 0, 42, 0, 75, 59, 164, 115, 140, 134, 141, 80, 210, 104, 211, 104, 36, 140, 93, 88, 18, 100, 22, 122, 161, 132, 47, 135, 197, 120, 81, 149, 85, 130, 23, 139, 138, 147, 60, 0, 42, 0, 59, 79, 74, 121, 95, 89, 41, 93, 99, 125, 141, 149, 176, 80, 109, 93, 237, 117, 35, 86, 179, 125, 60, 0, 42, 0, 45, 117, 60, 0, 42, 0, 196, 87, 242, 87, 247, 126, 35, 116, 43, 116, 246, 131, 67, 126, 155, 151, 60, 0, 42, 0, 245, 108, 248, 143, 44, 144, 116, 88, 90, 149, 15, 117, 101, 216, 244, 221, 112, 75, 102, 142, 240, 147, 47, 136, 60, 0, 42, 0, 252, 99, 60, 0, 42, 0, 68, 92, 106, 80, 244, 107, 69, 105, 60, 144, 77, 140, 149, 135, 190, 156, 158, 147, 86, 157, 15, 156, 60, 0, 42, 0, 120, 131, 59, 159, 60, 0, 42, 0, 21, 83, 212, 107, 50, 56, 54, 89, 60, 103, 74, 79, 33, 84, 163, 89, 152, 108, 149, 117, 124, 95, 192, 103, 213, 121, 254, 79, 20, 123, 131, 124, 45, 130, 90, 85, 143, 58, 70, 123, 132, 123, 91, 128, 25, 145, 143, 140, 60, 0, 42, 0, 1, 94, 197, 95, 213, 107, 237, 149, 86, 79, 41, 54, 82, 87, 135, 94, 208, 139, 32, 54, 188, 89, 45, 96, 54, 96, 217, 57, 136, 103, 64, 117, 178, 144, 212, 84, 242, 103, 214, 107, 204, 115, 170, 117, 254, 130, 217, 107, 244, 114, 19, 123, 138, 124, 92, 131, 203, 148, 91, 150, 98, 90, 179, 94, 93, 101, 144, 104, 98, 117, 144, 136, 135, 149, 137, 149, 27, 88, 59, 95, 60, 95, 10, 97, 14, 97, 98, 110, 149, 118, 90, 123, 6, 132, 86, 138, 177, 140, 193, 140, 81, 141, 215, 110, 79, 113, 249, 117, 250, 117, 100, 119, 119, 129, 12, 135, 248, 141, 77, 146, 159, 149, 246, 152, 246, 85, 99, 94, 10, 95, 67, 95, 237, 110, 167, 120, 133, 123, 134, 123, 188, 125, 214, 132, 232, 136, 157, 153, 119, 111, 154, 113, 88, 115, 89, 115, 235, 62, 253, 132, 61, 133, 11, 72, 42, 145, 220, 153, 242, 154, 193, 88, 22, 91, 230, 94, 101, 61, 230, 123, 124, 127, 241, 137, 133, 155, 131, 101, 222, 111, 243, 123, 42, 126, 194, 129, 156, 133, 127, 144, 70, 153, 112, 89, 167, 116, 85, 142, 40, 145, 192, 154, 116, 126, 94, 137, 206, 147, 87, 69, 99, 137, 131, 142, 132, 142, 184, 151, 224, 151, 20, 141, 52, 148, 70, 154, 83, 155, 221, 157, 233, 157, 10, 159, 60, 0, 42, 0, 2, 83, 30, 132, 100, 94, 54, 91, 69, 137, 60, 0, 42, 0, 139, 97, 222, 135, 214, 156, 73, 156, 8, 159, 76, 134, 158, 159, 60, 0, 42, 0, 37, 82, 43, 82, 135, 84, 223, 67, 130, 131, 194, 134, 182, 95, 82, 137, 105, 142, 60, 0, 42, 0, 42, 118, 95, 118, 220, 63, 60, 0, 42, 0, 70, 95, 24, 67, 60, 0, 42, 0, 108, 112, 83, 103, 7, 104, 209, 152, 137, 154, 223, 154, 242, 109, 106, 95, 11, 115, 63, 129, 169, 152, 130, 88, 86, 94, 125, 100, 238, 110, 174, 152, 160, 154, 25, 106, 155, 113, 152, 129, 8, 133, 45, 118, 230, 120, 86, 149, 217, 152, 218, 152, 38, 81, 183, 152, 12, 112, 36, 139, 2, 114, 213, 129, 232, 133, 6, 141, 226, 147, 110, 122, 115, 149, 198, 152, 199, 152, 200, 152, 67, 154, 99, 148, 189, 75, 107, 154, 60, 0, 42, 0, 104, 136, 74, 90, 241, 136, 216, 138, 54, 147, 166, 106, 62, 137, 60, 0, 42, 0, 245, 79, 212, 156, 62, 156, 60, 0, 42, 0, 202, 152, 60, 0, 42, 0, 185, 143, 186, 143, 45, 120, 62, 123, 217, 99, 53, 115, 22, 127, 120, 113, 81, 114, 2, 117, 175, 123, 232, 125, 89, 135, 137, 144, 125, 147, 202, 156, 173, 151, 138, 144, 254, 155, 255, 155, 105, 124, 60, 0, 42, 0, 44, 141, 65, 98, 134, 122, 62, 83, 252, 96, 182, 140, 57, 132, 165, 120, 40, 122, 196, 124, 10, 137, 24, 157, 202, 133, 60, 0, 42, 0, 94, 83, 1, 95, 37, 83, 237, 95, 131, 98, 115, 108, 116, 108, 198, 145, 216, 83, 163, 115, 196, 130, 191, 79, 9, 89, 42, 102, 83, 59, 141, 137, 167, 95, 15, 127, 77, 144, 158, 149, 161, 143, 246, 125, 81, 130, 167, 143, 168, 143, 169, 143, 171, 143, 174, 143, 175, 143, 138, 139, 60, 0, 42, 0, 197, 92, 158, 112, 60, 0, 42, 0, 67, 108, 160, 144, 162, 115, 15, 120, 190, 91, 108, 95, 185, 104, 167, 80, 140, 101, 21, 105, 232, 110, 222, 62, 36, 127, 223, 105, 120, 116, 105, 140, 211, 140, 212, 140, 84, 149, 16, 81, 210, 111, 241, 111, 104, 134, 115, 140, 179, 106, 184, 116, 21, 112, 38, 151, 125, 126, 76, 148, 110, 152, 60, 0, 42, 0, 72, 100, 161, 107, 145, 129, 233, 154, 239, 100, 2, 155, 175, 107, 207, 129, 204, 154, 19, 155, 213, 154, 34, 155, 60, 0, 42, 0, 30, 108, 245, 111, 60, 0, 42, 0, 171, 81, 204, 78, 210, 78, 55, 108, 176, 81, 117, 81, 164, 99, 60, 0, 42, 0, 25, 78, 50, 96, 166, 98, 201, 121, 180, 144, 67, 150, 30, 102, 58, 102, 196, 103, 179, 112, 234, 130, 124, 153, 42, 119, 137, 122, 0, 58, 195, 134, 82, 100, 128, 121, 31, 122, 53, 146, 124, 146, 5, 153, 32, 153, 158, 151, 60, 0, 42, 0, 118, 94, 200, 52, 38, 78, 117, 79, 119, 94, 176, 94, 2, 80, 36, 104, 197, 117, 221, 122, 75, 80, 161, 80, 206, 91, 197, 104, 129, 138, 169, 155, 80, 151, 60, 0, 42, 0, 170, 87, 134, 151, 242, 146, 60, 0, 42, 0, 236, 92, 175, 94, 97, 102, 11, 144, 61, 146, 167, 138, 60, 0, 42, 0, 234, 156, 80, 143, 173, 145, 60, 0, 42, 0, 95, 83, 82, 216, 240, 220, 101, 136, 250, 84, 85, 99, 184, 85, 220, 136, 79, 157, 60, 0, 42, 0, 13, 78, 3, 94, 72, 79, 37, 84, 101, 107, 152, 84, 53, 54, 22, 96, 170, 98, 104, 107, 105, 107, 205, 60, 232, 103, 154, 148, 207, 82, 212, 87, 215, 87, 145, 96, 87, 99, 184, 148, 224, 87, 121, 131, 232, 144, 255, 116, 4, 142, 0, 133, 246, 144, 20, 153, 34, 153, 240, 123, 63, 124, 60, 0, 39, 0, 208, 253, 39, 0, 6, 49, 60, 0, 42, 0, 145, 89, 85, 216, 25, 220, 80, 216, 181, 221, 133, 118, 180, 141, 37, 130, 106, 85, 105, 132, 60, 0, 42, 0, 119, 103, 44, 114, 177, 99, 54, 116, 98, 123, 86, 111, 60, 0, 42, 0, 10, 94, 21, 94, 21, 96, 153, 136, 60, 0, 42, 0, 139, 148, 97, 87, 165, 92, 202, 108, 252, 108, 135, 152, 140, 110, 85, 146, 23, 152, 81, 111, 250, 147, 60, 0, 42, 0, 70, 90, 25, 86, 34, 133, 49, 145, 164, 118, 8, 139, 199, 106, 60, 0, 42, 0, 245, 83, 0, 92, 183, 148, 56, 123, 202, 153, 60, 0, 42, 0, 182, 92, 135, 112, 64, 101, 34, 102, 38, 109, 192, 115, 235, 143, 222, 112, 52, 120, 54, 120, 217, 145, 149, 124, 170, 132, 68, 155, 151, 145, 150, 74, 60, 0, 42, 0, 114, 104, 60, 0, 42, 0, 205, 98, 60, 0, 42, 0, 243, 79, 152, 95, 146, 99, 5, 115, 209, 104, 76, 114, 43, 143, 48, 124, 50, 124, 164, 114, 60, 0, 42, 0, 249, 94, 60, 0, 42, 0, 204, 84, 62, 109, 91, 59, 67, 110, 142, 132, 131, 147, 60, 0, 42, 0, 120, 84, 176, 54, 12, 96, 167, 128, 248, 103, 218, 128, 67, 136, 133, 145, 60, 0, 42, 0, 43, 150, 249, 87, 106, 150, 240, 107, 84, 141, 43, 149, 244, 136, 245, 136, 224, 140, 205, 153, 60, 0, 42, 0, 214, 79, 60, 0, 42, 0, 2, 79, 155, 108, 105, 79, 20, 94, 245, 89, 190, 101, 198, 101, 127, 109, 238, 115, 77, 145, 41, 123, 131, 66, 148, 143, 183, 153, 143, 93, 8, 151, 97, 143, 60, 0, 42, 0, 220, 132, 60, 0, 42, 0, 155, 98, 203, 98, 44, 129, 60, 0, 42, 0, 40, 82, 134, 84, 137, 87, 150, 94, 205, 114, 176, 112, 46, 114, 15, 83, 141, 136, 243, 142, 132, 151, 131, 158, 133, 158, 60, 0, 42, 0, 209, 141, 60, 0, 42, 0, 69, 89, 225, 108, 174, 112, 177, 117, 176, 118, 50, 120, 173, 158, 31, 121, 46, 121, 60, 0, 42, 0, 34, 132, 28, 137, 60, 0, 42, 0, 86, 82, 29, 90, 60, 0, 42, 0, 148, 98, 153, 98, 74, 99, 138, 99, 210, 136, 129, 123, 7, 147, 60, 0, 42, 0, 133, 84, 227, 84, 68, 90, 131, 114, 205, 94, 60, 0, 42, 0, 52, 52, 5, 119, 25, 120, 104, 117, 88, 111, 0, 101, 60, 0, 42, 0, 63, 114, 0, 109, 216, 118, 216, 141, 187, 90, 75, 94, 43, 100, 195, 105, 176, 132, 228, 118, 208, 120, 15, 126, 251, 120, 82, 142, 76, 216, 102, 220, 10, 112, 224, 135, 99, 142, 156, 147, 182, 151, 60, 0, 42, 0, 184, 81, 36, 82, 156, 108, 218, 98, 238, 108, 141, 112, 219, 83, 73, 114, 252, 118, 84, 117, 65, 128, 162, 136, 74, 138, 191, 110, 22, 152, 236, 146, 123, 137, 123, 148, 60, 0, 42, 0, 101, 157, 60, 0, 42, 0, 183, 85, 116, 86, 85, 107, 60, 0, 42, 0, 235, 116, 198, 118, 83, 110, 80, 132, 60, 0, 42, 0, 96, 84, 248, 127, 60, 0, 42, 0, 175, 85, 60, 0, 42, 0, 83, 78, 151, 108, 238, 128, 241, 150, 194, 110, 150, 129, 38, 74, 54, 151, 60, 0, 42, 0, 144, 83, 158, 94, 150, 83, 193, 101, 4, 144, 61, 130, 206, 90, 172, 95, 82, 216, 237, 221, 131, 135, 209, 156, 142, 159, 144, 159, 60, 0, 42, 0, 217, 85, 42, 128, 171, 137, 60, 0, 42, 0, 144, 112, 168, 128, 214, 128, 60, 0, 42, 0, 9, 83, 193, 54, 38, 96, 168, 98, 114, 96, 248, 57, 48, 120, 136, 104, 249, 112, 81, 120, 239, 142, 155, 149, 48, 111, 45, 86, 142, 111, 222, 120, 60, 0, 42, 0, 131, 130, 11, 103, 252, 122, 23, 80, 55, 99, 11, 88, 56, 95, 145, 131, 109, 95, 218, 104, 22, 105, 124, 120, 29, 122, 79, 158, 115, 88, 137, 97, 240, 105, 37, 106, 162, 113, 236, 132, 96, 216, 131, 220, 35, 143, 227, 123, 168, 129, 11, 147, 248, 151, 247, 123, 252, 154, 218, 135, 219, 135, 5, 155, 54, 77, 252, 151, 108, 157, 47, 154, 132, 126, 20, 155, 93, 148, 60, 0, 42, 0, 103, 99, 206, 109, 143, 118, 123, 82, 60, 0, 42, 0, 189, 99, 42, 105, 176, 120, 43, 142, 60, 0, 42, 0, 21, 78, 19, 79, 62, 79, 121, 98, 176, 126, 111, 87, 171, 98, 183, 98, 139, 112, 201, 114, 179, 144, 18, 120, 65, 216, 49, 223, 130, 96, 219, 121, 224, 121, 21, 125, 205, 148, 199, 101, 205, 127, 26, 128, 126, 140, 8, 146, 26, 146, 57, 146, 95, 146, 148, 146, 136, 82, 199, 120, 211, 153, 236, 154, 124, 86, 13, 147, 126, 155, 141, 155, 181, 97, 20, 121, 21, 121, 57, 151, 60, 0, 42, 0, 174, 118, 223, 52, 48, 150, 175, 92, 135, 103, 222, 107, 211, 114, 182, 128, 152, 130, 215, 107, 216, 107, 178, 117, 141, 134, 100, 85, 228, 87, 37, 93, 90, 216, 45, 222, 189, 134, 190, 134, 124, 140, 96, 216, 26, 222, 235, 144, 116, 150, 55, 113, 53, 116, 62, 129, 87, 129, 143, 156, 116, 127, 141, 129, 49, 135, 110, 155, 192, 88, 250, 123, 183, 135, 148, 140, 103, 157, 134, 127, 7, 103, 25, 159, 60, 0, 42, 0, 57, 83, 128, 94, 139, 117, 243, 78, 46, 87, 201, 130, 52, 129, 222, 117, 162, 146, 192, 138, 4, 157, 215, 100, 125, 86, 86, 118, 38, 76, 173, 86, 60, 0, 42, 0, 65, 92, 224, 109, 18, 110, 202, 99, 253, 145, 178, 90, 211, 90, 101, 119, 159, 143, 78, 111, 43, 122, 251, 80, 188, 111, 138, 86, 19, 117, 68, 68, 136, 117, 108, 139, 226, 149, 255, 157, 10, 158, 60, 0, 42, 0, 140, 105, 60, 0, 42, 0, 21, 108, 135, 100, 75, 216, 194, 220, 134, 100, 188, 102, 165, 119, 60, 0, 42, 0, 63, 78, 228, 130, 5, 148, 60, 0, 42, 0, 243, 90, 60, 0, 42, 0, 125, 82, 83, 97, 37, 127, 216, 152, 218, 101, 242, 127, 181, 135, 165, 114, 195, 152, 196, 152, 82, 155, 60, 0, 42, 0, 214, 90, 226, 116, 194, 122, 184, 133, 221, 149, 60, 0, 42, 0, 141, 107, 111, 95, 159, 119, 251, 123, 57, 126, 165, 145, 171, 118, 96, 152, 60, 0, 42, 0, 104, 121, 196, 80, 225, 82, 12, 86, 177, 95, 2, 111, 19, 59, 60, 0, 42, 0, 232, 86, 79, 80, 165, 90, 143, 114, 199, 123, 233, 127, 66, 147, 163, 157, 60, 0, 42, 0, 136, 154, 252, 128, 65, 129, 68, 105, 105, 105, 198, 140, 240, 141, 218, 138, 65, 142, 226, 153, 191, 154, 8, 154, 60, 0, 42, 0, 145, 137, 29, 140, 181, 140, 222, 138, 60, 0, 42, 0, 71, 114, 151, 154, 23, 154, 25, 154, 60, 0, 42, 0, 120, 155, 60, 0, 42, 0, 216, 89, 252, 98, 102, 56, 23, 121, 102, 122, 170, 153, 94, 154, 60, 0, 42, 0, 173, 115, 43, 141, 38, 90, 167, 140, 21, 116, 212, 90, 145, 152, 59, 152, 42, 91, 113, 115, 178, 133, 172, 86, 201, 119, 25, 136, 166, 152, 112, 152, 60, 0, 42, 0, 193, 84, 128, 105, 60, 0, 42, 0, 93, 114, 86, 108, 88, 128, 60, 0, 42, 0, 82, 78, 57, 117, 220, 79, 9, 90, 132, 109, 47, 120, 96, 128, 117, 130, 238, 122, 41, 152, 102, 216, 77, 222, 60, 0, 42, 0, 115, 94, 196, 139, 237, 81, 111, 84, 106, 87, 217, 108, 79, 92, 33, 94, 176, 103, 52, 109, 182, 115, 211, 128, 249, 130, 241, 144, 184, 62, 194, 62, 83, 131, 91, 92, 50, 94, 220, 109, 246, 116, 178, 134, 72, 94, 41, 113, 62, 127, 13, 132, 226, 134, 85, 138, 1, 117, 241, 132, 255, 142, 134, 156, 244, 81, 2, 66, 127, 97, 179, 123, 39, 143, 145, 97, 131, 155, 152, 106, 8, 124, 11, 134, 60, 0, 42, 0, 188, 92, 64, 88, 60, 0, 42, 0, 198, 78, 52, 101, 53, 101, 81, 98, 96, 150, 87, 86, 178, 100, 125, 111, 200, 100, 198, 155, 60, 0, 42, 0, 170, 62, 13, 83, 47, 129, 134, 131, 208, 131, 233, 131, 97, 132, 213, 80, 177, 132, 178, 132, 122, 145, 163, 88, 91, 115, 158, 116, 238, 111, 168, 119, 89, 122, 59, 66, 100, 149, 86, 216, 200, 220, 128, 126, 101, 137, 247, 147, 60, 0, 42, 0, 36, 87, 52, 103, 3, 87, 102, 109, 243, 112, 110, 102, 165, 110, 49, 140, 233, 138, 56, 106, 6, 108, 143, 106, 104, 149, 92, 139, 124, 142, 32, 148, 60, 0, 42, 0, 250, 148, 22, 130, 23, 130, 234, 146, 37, 59, 17, 112, 221, 102, 60, 0, 42, 0, 236, 93, 237, 93, 12, 141, 60, 0, 39, 0, 208, 253, 39, 0, 7, 49, 60, 0, 42, 0, 99, 84, 60, 0, 42, 0, 136, 89, 86, 91, 189, 90, 36, 91, 55, 91, 60, 0, 42, 0, 187, 158, 242, 117, 152, 114, 52, 133, 198, 135, 199, 135, 60, 0, 42, 0, 108, 154, 55, 52, 155, 115, 1, 120, 130, 134, 172, 153, 164, 110, 106, 116, 188, 120, 158, 135, 183, 147, 34, 156, 204, 157, 60, 0, 42, 0, 184, 114, 105, 103, 67, 121, 129, 149, 130, 154, 27, 85, 140, 80, 65, 115, 28, 86, 56, 58, 170, 105, 112, 119, 161, 121, 117, 127, 43, 55, 225, 153, 35, 121, 21, 155, 60, 0, 42, 0, 135, 78, 23, 84, 206, 85, 27, 86, 242, 90, 100, 144, 60, 0, 42, 0, 120, 100, 60, 0, 42, 0, 133, 94, 31, 140, 141, 153, 235, 90, 189, 158, 105, 100, 121, 100, 33, 106, 156, 129, 69, 106, 232, 120, 226, 124, 164, 86, 245, 100, 40, 139, 169, 86, 176, 86, 67, 153, 17, 134, 152, 82, 205, 154, 84, 155, 250, 64, 93, 153, 60, 0, 42, 0, 185, 98, 225, 97, 75, 74, 60, 0, 42, 0, 43, 103, 176, 82, 61, 87, 186, 89, 19, 94, 126, 107, 127, 107, 129, 107, 171, 108, 82, 216, 251, 220, 30, 94, 41, 102, 186, 103, 9, 131, 76, 150, 28, 85, 140, 118, 28, 119, 63, 119, 30, 120, 227, 121, 61, 119, 150, 124, 72, 125, 136, 131, 171, 131, 80, 110, 232, 134, 131, 140, 138, 140, 252, 85, 123, 88, 222, 91, 32, 111, 79, 115, 230, 132, 134, 146, 122, 151, 252, 90, 175, 102, 115, 60, 217, 158, 144, 119, 70, 149, 105, 155, 168, 88, 60, 118, 153, 119, 216, 158, 14, 112, 212, 135, 41, 139, 152, 140, 5, 114, 230, 133, 204, 147, 51, 121, 134, 126, 64, 154, 49, 128, 60, 0, 42, 0, 61, 96, 27, 92, 121, 155, 191, 158, 60, 0, 42, 0, 72, 78, 188, 158, 146, 86, 156, 86, 249, 111, 102, 118, 60, 0, 42, 0, 203, 87, 37, 55, 182, 133, 62, 151, 60, 0, 42, 0, 112, 78, 108, 131, 183, 140, 42, 86, 82, 133, 246, 157, 60, 0, 42, 0, 162, 82, 69, 79, 242, 88, 200, 143, 166, 158, 86, 83, 9, 129, 8, 129, 165, 158, 71, 136, 241, 82, 227, 140, 129, 144, 33, 151, 34, 151, 104, 216, 51, 221, 60, 0, 42, 0, 69, 84, 70, 87, 146, 108, 161, 108, 154, 103, 171, 115, 2, 104, 9, 119, 250, 130, 18, 90, 4, 129, 133, 104, 251, 115, 34, 129, 147, 131, 51, 88, 146, 90, 75, 93, 68, 110, 72, 110, 56, 115, 66, 119, 255, 144, 99, 105, 115, 105, 100, 113, 66, 116, 127, 132, 122, 88, 209, 105, 150, 121, 118, 145, 69, 149, 91, 158, 194, 146, 9, 151, 72, 122, 190, 95, 135, 147, 192, 119, 23, 101, 165, 157, 42, 134, 244, 158, 60, 0, 42, 0, 206, 107, 207, 107, 194, 81, 142, 127, 52, 99, 124, 109, 132, 90, 68, 93, 60, 110, 186, 90, 92, 129, 65, 149, 13, 91, 216, 113, 130, 147, 227, 158, 60, 0, 42, 0, 185, 89, 186, 98, 172, 108, 192, 101, 39, 102, 81, 59, 27, 119, 89, 121, 130, 136, 154, 90, 208, 91, 215, 117, 202, 141, 93, 113, 61, 155, 120, 119, 206, 151, 195, 123, 94, 135, 69, 155, 60, 0, 42, 0, 190, 142, 60, 0, 42, 0, 43, 115, 147, 140, 60, 0, 42, 0, 219, 107, 219, 119, 134, 103, 102, 114, 5, 131, 6, 131, 196, 101, 94, 127, 94, 81, 222, 142, 85, 145, 37, 88, 53, 110, 98, 216, 104, 221, 26, 149, 230, 154, 249, 90, 2, 108, 155, 114, 101, 135, 243, 154, 202, 135, 40, 147, 156, 157, 60, 0, 42, 0, 135, 81, 111, 83, 24, 89, 110, 78, 60, 98, 185, 54, 193, 92, 214, 108, 52, 102, 198, 148, 55, 123, 233, 132, 60, 0, 42, 0, 131, 81, 131, 118, 144, 81, 188, 130, 146, 81, 213, 103, 10, 119, 2, 131, 56, 141, 4, 128, 164, 136, 146, 137, 162, 90, 61, 94, 191, 140, 23, 97, 147, 102, 89, 105, 247, 107, 65, 116, 58, 132, 128, 119, 140, 140, 26, 145, 80, 135, 46, 145, 203, 97, 60, 0, 42, 0, 222, 84, 60, 0, 42, 0, 95, 114, 148, 79, 186, 82, 76, 59, 72, 96, 32, 109, 56, 119, 11, 140, 209, 134, 42, 127, 14, 142, 126, 146, 0, 139, 180, 119, 70, 126, 106, 147, 62, 157, 176, 158, 60, 0, 42, 0, 208, 103, 91, 216, 209, 220, 60, 0, 42, 0, 218, 90, 159, 152, 60, 0, 42, 0, 207, 89, 151, 96, 188, 52, 238, 134, 200, 80, 41, 140, 114, 97, 146, 153, 32, 106, 146, 119, 158, 119, 148, 151, 87, 216, 145, 220, 62, 139, 69, 153, 215, 156, 98, 152, 23, 155, 24, 155, 59, 156, 59, 136, 60, 0, 42, 0, 88, 92, 128, 110, 76, 119, 225, 110, 255, 110, 168, 135, 206, 135, 84, 137, 203, 147, 213, 119, 60, 0, 42, 0, 252, 102, 5, 59, 129, 88, 84, 94, 98, 97, 113, 100, 43, 111, 76, 115, 38, 127, 36, 145, 254, 105, 179, 113, 4, 133, 19, 133, 98, 56, 183, 111, 88, 149, 53, 126, 221, 147, 100, 216, 77, 221, 85, 69, 80, 216, 78, 221, 60, 0, 42, 0, 48, 134, 60, 0, 42, 0, 232, 149, 106, 98, 167, 115, 148, 148, 128, 149, 133, 149, 107, 99, 219, 131, 138, 116, 70, 147, 185, 78, 75, 134, 60, 0, 42, 0, 247, 149, 22, 113, 182, 96, 170, 102, 220, 113, 209, 97, 227, 97, 60, 0, 42, 0, 236, 78, 17, 80, 26, 105, 60, 0, 42, 0, 100, 114, 60, 0, 42, 0, 2, 84, 217, 95, 82, 108, 153, 144, 40, 92, 87, 103, 103, 103, 146, 130, 19, 108, 242, 118, 0, 123, 228, 84, 15, 90, 172, 94, 54, 57, 126, 96, 93, 109, 245, 114, 43, 131, 123, 114, 202, 62, 239, 145, 211, 148, 221, 117, 109, 120, 214, 134, 13, 66, 77, 216, 119, 222, 233, 146, 249, 153, 103, 216, 246, 221, 60, 0, 42, 0, 189, 131, 101, 120, 59, 131, 190, 131, 254, 88, 45, 111, 18, 59, 210, 135, 14, 136, 60, 0, 42, 0, 63, 117, 123, 134, 161, 81, 148, 131, 12, 132, 32, 132, 4, 64, 223, 118, 153, 132, 113, 135, 26, 81, 87, 106, 13, 117, 162, 119, 68, 133, 51, 145, 106, 94, 222, 97, 218, 102, 219, 111, 56, 145, 38, 103, 172, 106, 11, 108, 67, 68, 69, 68, 205, 155, 199, 119, 30, 121, 114, 158, 104, 130, 9, 134, 210, 119, 63, 151, 64, 151, 91, 153, 109, 152, 6, 159, 15, 158, 60, 0, 42, 0, 208, 82, 27, 115, 254, 116, 69, 64, 48, 149, 75, 130, 34, 135, 51, 147, 220, 97, 116, 115, 245, 97, 237, 155, 19, 136, 60, 0, 42, 0, 95, 91, 67, 216, 124, 221, 166, 104, 149, 110, 34, 89, 35, 89, 221, 68, 37, 151, 52, 55, 60, 0, 42, 0, 185, 99, 221, 100, 60, 0, 42, 0, 170, 84, 47, 119, 135, 119, 60, 0, 42, 0, 158, 81, 37, 95, 89, 127, 98, 121, 247, 143, 21, 115, 28, 140, 120, 138, 190, 132, 76, 95, 223, 100, 220, 124, 59, 126, 14, 139, 154, 145, 138, 158, 139, 158, 176, 121, 97, 151, 55, 55, 48, 112, 124, 115, 155, 158, 2, 98, 32, 101, 213, 116, 126, 149, 34, 114, 119, 69, 60, 134, 190, 145, 191, 145, 13, 158, 196, 145, 60, 0, 42, 0, 115, 124, 136, 130, 142, 79, 181, 108, 139, 127, 45, 95, 35, 109, 73, 101, 43, 119, 18, 129, 51, 110, 94, 132, 164, 146, 29, 133, 212, 111, 74, 91, 86, 112, 60, 0, 42, 0, 150, 81, 248, 124, 104, 108, 149, 108, 147, 91, 204, 108, 197, 137, 218, 92, 187, 91, 85, 121, 216, 121, 198, 91, 231, 109, 255, 109, 147, 137, 148, 137, 66, 94, 39, 140, 83, 88, 78, 94, 32, 58, 155, 137, 39, 86, 147, 105, 245, 110, 30, 111, 144, 113, 28, 135, 18, 106, 36, 133, 15, 159, 170, 81, 102, 94, 215, 111, 16, 139, 193, 106, 26, 124, 204, 133, 131, 127, 60, 0, 42, 0, 92, 78, 0, 84, 169, 84, 246, 84, 109, 91, 60, 0, 42, 0, 109, 112, 213, 112, 149, 137, 35, 100, 197, 110, 17, 133, 19, 157, 254, 123, 142, 133, 109, 94, 241, 97, 215, 106, 27, 136, 74, 136, 86, 148, 116, 156, 60, 0, 42, 0, 181, 85, 60, 0, 42, 0, 215, 130, 100, 52, 140, 90, 207, 99, 75, 158, 132, 119, 226, 125, 147, 157, 89, 156, 60, 0, 42, 0, 106, 103, 7, 119, 210, 121, 252, 109, 58, 110, 8, 127, 206, 123, 242, 125, 208, 133, 136, 144, 60, 0, 42, 0, 153, 89, 153, 94, 133, 115, 215, 122, 191, 94, 223, 94, 60, 0, 42, 0, 44, 140, 44, 139, 60, 0, 42, 0, 128, 91, 135, 130, 32, 119, 66, 90, 245, 126, 148, 90, 201, 104, 191, 125, 220, 125, 241, 129, 82, 135, 53, 91, 176, 106, 203, 106, 200, 119, 202, 119, 207, 119, 60, 0, 42, 0, 15, 78, 69, 108, 77, 81, 148, 108, 254, 158, 201, 82, 4, 119, 41, 90, 67, 55, 109, 80, 149, 81, 212, 82, 17, 110, 149, 85, 16, 97, 78, 110, 5, 127, 123, 125, 124, 129, 66, 132, 253, 158, 240, 62, 236, 125, 171, 158, 160, 111, 184, 155, 60, 0, 42, 0, 99, 151, 98, 151, 198, 124, 170, 158, 186, 158, 181, 158, 60, 0, 42, 0, 17, 108, 196, 89, 183, 92, 222, 95, 11, 96, 251, 101, 252, 101, 201, 115, 255, 118, 7, 120, 224, 130, 96, 127, 15, 93, 106, 99, 24, 116, 23, 127, 111, 101, 73, 116, 251, 117, 136, 120, 49, 146, 205, 125, 184, 74, 225, 125, 9, 147, 22, 157, 114, 147, 60, 0, 42, 0, 191, 118, 186, 81, 33, 82, 245, 149, 191, 98, 239, 108, 196, 82, 67, 101, 253, 149, 175, 96, 79, 101, 34, 123, 253, 96, 99, 110, 148, 149, 13, 97, 6, 59, 139, 102, 169, 149, 246, 80, 92, 97, 171, 97, 99, 111, 34, 124, 92, 216, 75, 220, 216, 156, 32, 136, 53, 156, 104, 216, 52, 221, 60, 0, 42, 0, 138, 87, 61, 123, 60, 0, 42, 0, 13, 84, 14, 102, 35, 158, 58, 109, 0, 119, 165, 81, 23, 131, 25, 103, 51, 119, 237, 148, 199, 90, 159, 110, 61, 115, 13, 145, 157, 102, 160, 105, 194, 132, 152, 146, 244, 156, 145, 119, 159, 135, 173, 137, 228, 74, 60, 0, 42, 0, 114, 79, 243, 89, 213, 81, 79, 97, 105, 145, 60, 0, 42, 0, 125, 84, 39, 105, 122, 138, 60, 0, 42, 0, 181, 99, 60, 0, 42, 0, 234, 107, 178, 88, 1, 108, 60, 0, 42, 0, 205, 107, 169, 78, 97, 114, 118, 87, 198, 89, 199, 98, 212, 92, 115, 114, 70, 117, 82, 117, 223, 128, 93, 117, 94, 117, 42, 120, 110, 117, 95, 216, 249, 223, 67, 216, 150, 222, 103, 146, 7, 142, 60, 0, 42, 0, 40, 103, 235, 78, 48, 103, 238, 118, 144, 108, 199, 114, 145, 112, 103, 114, 220, 130, 227, 107, 158, 134, 188, 148, 175, 131, 238, 150, 223, 82, 8, 55, 88, 105, 102, 119, 108, 146, 147, 88, 85, 94, 89, 94, 84, 97, 85, 97, 174, 102, 82, 130, 2, 151, 70, 122, 56, 126, 170, 151, 60, 0, 42, 0, 233, 81, 60, 0, 39, 0, 208, 253, 39, 0, 8, 49, 60, 0, 42, 0, 209, 83, 183, 108, 122, 118, 160, 80, 124, 118, 66, 95, 102, 145, 177, 145, 60, 0, 42, 0, 79, 78, 16, 79, 194, 89, 161, 87, 76, 109, 186, 117, 90, 127, 0, 150, 48, 104, 29, 120, 55, 131, 79, 123, 130, 119, 112, 127, 165, 149, 120, 127, 67, 106, 197, 133, 60, 0, 42, 0, 113, 79, 213, 108, 75, 112, 60, 0, 42, 0, 208, 115, 58, 116, 234, 154, 238, 154, 95, 133, 94, 216, 11, 223, 60, 0, 42, 0, 69, 147, 60, 0, 42, 0, 207, 78, 114, 87, 60, 0, 42, 0, 187, 104, 60, 0, 42, 0, 222, 152, 131, 89, 94, 151, 219, 152, 97, 85, 83, 90, 4, 110, 239, 126, 100, 216, 235, 221, 73, 98, 6, 115, 242, 131, 95, 151, 203, 125, 26, 135, 246, 136, 15, 151, 177, 156, 37, 153, 161, 153, 17, 154, 27, 154, 221, 152, 60, 0, 42, 0, 165, 128, 221, 109, 83, 129, 8, 66, 48, 135, 230, 135, 60, 0, 42, 0, 15, 103, 42, 83, 253, 139, 92, 89, 177, 96, 144, 101, 208, 104, 167, 105, 225, 127, 185, 138, 218, 123, 92, 133, 100, 216, 60, 221, 60, 0, 42, 0, 32, 84, 159, 94, 110, 103, 184, 108, 210, 114, 186, 128, 190, 130, 50, 102, 199, 128, 57, 141, 247, 79, 85, 82, 158, 83, 191, 117, 93, 92, 107, 150, 195, 94, 67, 62, 9, 132, 187, 140, 241, 117, 68, 149, 226, 94, 202, 102, 53, 73, 72, 118, 35, 159, 247, 111, 1, 76, 224, 106, 225, 155, 40, 148, 69, 151, 60, 0, 42, 0, 84, 90, 131, 102, 60, 0, 42, 0, 17, 125, 230, 136, 60, 0, 42, 0, 54, 127, 38, 84, 154, 89, 57, 127, 59, 127, 149, 107, 236, 150, 0, 157, 60, 0, 42, 0, 6, 94, 9, 138, 106, 117, 235, 82, 67, 86, 15, 91, 97, 94, 163, 97, 217, 101, 67, 133, 219, 101, 89, 126, 251, 127, 233, 133, 83, 143, 191, 152, 83, 124, 220, 152, 85, 156, 60, 0, 42, 0, 225, 81, 226, 81, 227, 81, 54, 56, 219, 95, 75, 103, 254, 119, 117, 124, 146, 148, 201, 103, 230, 112, 39, 130, 50, 123, 229, 104, 34, 110, 105, 113, 208, 125, 166, 88, 10, 106, 78, 106, 212, 113, 160, 116, 176, 129, 65, 126, 160, 133, 179, 127, 78, 137, 111, 142, 42, 112, 63, 112, 44, 121, 7, 148, 34, 148, 41, 134, 28, 136, 237, 157, 60, 0, 42, 0, 205, 83, 85, 98, 212, 143, 74, 72, 233, 145, 60, 0, 42, 0, 64, 216, 169, 221, 62, 108, 175, 114, 127, 89, 78, 108, 109, 153, 219, 108, 41, 141, 72, 117, 3, 131, 0, 64, 211, 142, 79, 90, 181, 104, 0, 61, 213, 118, 53, 123, 169, 140, 236, 142, 239, 152, 240, 152, 252, 110, 14, 91, 196, 123, 60, 0, 42, 0, 36, 130, 60, 0, 42, 0, 6, 82, 41, 84, 9, 94, 183, 126, 16, 102, 27, 108, 172, 130, 219, 84, 122, 81, 27, 125, 194, 127, 111, 136, 93, 81, 28, 138, 90, 145, 251, 104, 22, 146, 240, 150, 31, 59, 6, 103, 211, 113, 52, 153, 89, 153, 60, 0, 42, 0, 95, 87, 162, 89, 142, 92, 126, 108, 12, 103, 140, 103, 131, 112, 166, 128, 146, 127, 160, 134, 161, 134, 164, 104, 252, 104, 26, 113, 154, 153, 182, 132, 179, 88, 105, 94, 198, 111, 117, 155, 104, 106, 204, 113, 97, 133, 110, 140, 171, 150, 34, 159, 181, 127, 22, 159, 118, 140, 82, 143, 60, 148, 169, 153, 194, 158, 60, 0, 42, 0, 137, 124, 250, 158, 60, 0, 42, 0, 253, 78, 5, 95, 75, 89, 255, 95, 206, 121, 126, 80, 36, 97, 170, 124, 232, 80, 164, 97, 110, 89, 185, 129, 222, 124, 188, 156, 53, 112, 93, 156, 60, 0, 42, 0, 213, 122, 174, 142, 60, 0, 42, 0, 26, 83, 185, 101, 74, 87, 120, 108, 161, 144, 139, 103, 101, 114, 179, 130, 171, 148, 132, 134, 211, 109, 1, 146, 11, 157, 60, 0, 42, 0, 168, 89, 50, 150, 63, 98, 170, 128, 197, 87, 130, 156, 116, 155, 31, 156, 60, 0, 42, 0, 255, 78, 191, 139, 119, 95, 186, 126, 9, 102, 24, 102, 236, 116, 6, 119, 35, 80, 202, 101, 33, 125, 43, 130, 42, 138, 227, 154, 173, 157, 60, 0, 42, 0, 62, 101, 189, 141, 60, 0, 42, 0, 15, 88, 58, 147, 60, 0, 42, 0, 48, 78, 206, 152, 249, 78, 232, 81, 236, 81, 166, 89, 163, 108, 168, 108, 238, 81, 171, 103, 1, 92, 175, 117, 253, 118, 28, 120, 168, 152, 212, 54, 239, 92, 240, 92, 160, 68, 81, 80, 123, 104, 253, 112, 54, 93, 38, 115, 11, 149, 83, 105, 142, 114, 81, 132, 2, 135, 11, 118, 184, 120, 252, 80, 200, 123, 210, 146, 55, 145, 146, 106, 207, 149, 80, 140, 224, 147, 165, 52, 247, 91, 38, 56, 67, 112, 70, 145, 59, 151, 52, 134, 45, 136, 74, 151, 204, 152, 183, 158, 60, 0, 42, 0, 175, 81, 6, 89, 64, 99, 114, 109, 34, 144, 56, 88, 174, 153, 152, 125, 66, 130, 128, 57, 83, 100, 40, 111, 111, 61, 60, 0, 42, 0, 189, 139, 42, 85, 130, 137, 247, 138, 60, 0, 42, 0, 228, 81, 73, 89, 46, 117, 248, 79, 87, 110, 40, 113, 72, 113, 29, 127, 87, 141, 239, 156, 243, 156, 12, 157, 245, 140, 43, 126, 60, 0, 42, 0, 18, 116, 132, 110, 189, 147, 21, 134, 60, 0, 42, 0, 133, 137, 60, 0, 42, 0, 43, 89, 21, 79, 75, 84, 139, 89, 158, 144, 199, 89, 158, 115, 164, 128, 36, 96, 206, 103, 6, 120, 186, 87, 16, 90, 3, 92, 66, 131, 109, 136, 201, 101, 106, 59, 40, 125, 116, 131, 186, 141, 184, 158, 225, 117, 3, 122, 215, 141, 7, 146, 95, 123, 146, 125, 117, 91, 103, 140, 28, 145, 119, 101, 154, 129, 250, 156, 169, 158, 208, 124, 172, 158, 177, 158, 239, 97, 60, 0, 42, 0, 64, 78, 255, 93, 23, 95, 15, 79, 235, 81, 54, 117, 91, 79, 185, 81, 28, 82, 90, 91, 118, 98, 136, 84, 170, 92, 127, 95, 43, 96, 194, 98, 13, 103, 142, 103, 237, 108, 194, 126, 203, 126, 153, 130, 163, 130, 216, 79, 152, 87, 132, 58, 235, 103, 31, 108, 17, 109, 165, 112, 184, 115, 73, 117, 80, 117, 88, 127, 251, 130, 0, 131, 232, 151, 249, 84, 63, 104, 110, 109, 41, 120, 83, 121, 47, 131, 168, 134, 219, 144, 16, 83, 116, 104, 170, 109, 240, 112, 8, 116, 38, 123, 48, 123, 49, 125, 60, 125, 199, 127, 116, 130, 169, 131, 89, 134, 69, 94, 244, 104, 101, 125, 102, 127, 212, 131, 177, 136, 176, 124, 141, 125, 64, 130, 77, 132, 9, 135, 144, 143, 88, 146, 92, 146, 171, 152, 231, 156, 145, 105, 143, 121, 42, 122, 153, 123, 205, 151, 94, 94, 147, 111, 96, 135, 244, 154, 232, 138, 62, 142, 59, 143, 132, 155, 20, 157, 65, 118, 101, 216, 50, 223, 251, 158, 70, 137, 105, 157, 157, 157, 60, 0, 42, 0, 82, 84, 154, 98, 43, 117, 118, 78, 156, 94, 35, 95, 202, 98, 167, 101, 204, 79, 219, 79, 213, 128, 236, 156, 239, 79, 217, 144, 220, 145, 225, 145, 108, 99, 133, 143, 36, 113, 217, 118, 81, 129, 207, 110, 5, 135, 232, 66, 80, 129, 20, 143, 56, 86, 168, 100, 171, 100, 43, 152, 52, 155, 32, 124, 252, 158, 60, 0, 42, 0, 207, 46, 214, 46, 29, 150, 54, 114, 163, 139, 216, 78, 135, 89, 31, 141, 144, 84, 127, 87, 20, 57, 206, 122, 28, 150, 68, 150, 120, 154, 13, 89, 202, 92, 3, 138, 160, 140, 116, 141, 49, 57, 84, 121, 165, 134, 90, 150, 105, 80, 168, 81, 111, 82, 102, 90, 185, 134, 157, 136, 133, 80, 141, 90, 204, 91, 169, 95, 198, 61, 255, 121, 215, 134, 66, 138, 75, 141, 17, 87, 49, 105, 26, 127, 121, 129, 47, 132, 139, 156, 89, 141, 123, 75, 60, 59, 238, 125, 92, 135, 110, 135, 7, 137, 20, 137, 230, 140, 217, 153, 20, 91, 27, 126, 103, 133, 57, 143, 146, 155, 251, 140, 81, 147, 98, 147, 198, 156, 134, 137, 165, 153, 0, 70, 18, 156, 65, 216, 215, 220, 60, 0, 42, 0, 92, 145, 40, 105, 132, 137, 163, 121, 178, 155, 60, 0, 39, 0, 208, 253, 39, 0, 9, 49, 60, 0, 42, 0, 145, 84, 210, 84, 55, 128, 69, 131, 26, 123, 210, 85, 45, 100, 69, 216, 63, 220, 152, 100, 96, 86, 89, 216, 191, 221, 33, 137, 157, 147, 60, 0, 42, 0, 190, 143, 126, 84, 178, 89, 27, 96, 147, 108, 214, 143, 159, 112, 141, 127, 87, 117, 99, 117, 42, 123, 89, 131, 115, 82, 18, 83, 84, 123, 90, 138, 57, 144, 84, 144, 24, 150, 124, 151, 145, 151, 152, 133, 137, 147, 253, 135, 130, 142, 61, 148, 195, 151, 150, 159, 152, 159, 60, 0, 42, 0, 83, 98, 60, 0, 42, 0, 39, 89, 79, 108, 20, 119, 60, 0, 42, 0, 175, 87, 182, 88, 41, 118, 245, 113, 104, 126, 60, 0, 42, 0, 26, 86, 60, 0, 42, 0, 43, 57, 116, 96, 151, 95, 194, 109, 234, 96, 207, 104, 29, 149, 179, 95, 183, 95, 64, 147, 60, 0, 42, 0, 48, 87, 132, 118, 38, 129, 60, 0, 42, 0, 70, 84, 84, 84, 67, 115, 219, 97, 60, 0, 42, 0, 121, 107, 163, 80, 46, 144, 60, 0, 42, 0, 50, 52, 227, 78, 106, 143, 67, 216, 58, 220, 136, 87, 177, 92, 18, 94, 25, 117, 208, 126, 128, 154, 38, 94, 133, 95, 32, 96, 203, 103, 134, 107, 179, 115, 55, 141, 232, 143, 47, 94, 209, 142, 237, 87, 54, 94, 63, 125, 139, 136, 218, 142, 214, 62, 184, 140, 233, 142, 71, 116, 215, 94, 6, 66, 199, 83, 194, 59, 255, 125, 195, 102, 15, 157, 52, 98, 88, 216, 2, 221, 92, 130, 219, 158, 36, 124, 91, 142, 59, 112, 52, 151, 241, 158, 118, 137, 70, 151, 60, 0, 42, 0, 152, 155, 60, 0, 42, 0, 136, 46, 0, 82, 2, 82, 232, 83, 201, 95, 55, 103, 24, 108, 32, 130, 214, 145, 125, 156, 91, 155, 60, 0, 42, 0, 111, 99, 60, 0, 42, 0, 252, 91, 155, 92, 246, 92, 99, 99, 119, 121, 23, 100, 130, 121, 0, 56, 139, 93, 140, 93, 54, 58, 157, 150, 14, 92, 212, 88, 185, 93, 227, 100, 72, 142, 175, 150, 177, 121, 60, 0, 42, 0, 48, 82, 18, 80, 188, 96, 24, 113, 215, 118, 220, 118, 255, 131, 83, 144, 50, 122, 140, 123, 226, 127, 59, 122, 92, 136, 117, 86, 95, 136, 164, 106, 254, 113, 38, 67, 255, 127, 199, 142, 217, 116, 155, 126, 60, 0, 42, 0, 118, 92, 102, 150, 33, 105, 221, 105, 60, 0, 42, 0, 58, 84, 23, 85, 92, 81, 253, 144, 96, 81, 56, 133, 119, 106, 252, 123, 60, 0, 42, 0, 150, 98, 39, 150, 147, 103, 161, 103, 30, 85, 170, 134, 97, 150, 4, 146, 60, 0, 42, 0, 151, 101, 70, 140, 98, 109, 214, 144, 126, 153, 37, 155, 170, 104, 237, 107, 48, 129, 115, 131, 23, 144, 88, 145, 216, 117, 151, 149, 166, 122, 38, 75, 38, 155, 22, 153, 163, 101, 68, 216, 157, 220, 216, 149, 70, 69, 199, 122, 42, 155, 44, 155, 45, 155, 60, 0, 42, 0, 103, 78, 84, 130, 60, 0, 42, 0, 57, 78, 137, 89, 85, 83, 197, 98, 88, 83, 8, 119, 3, 120, 60, 128, 61, 128, 159, 68, 67, 128, 173, 142, 248, 144, 174, 85, 133, 90, 154, 107, 5, 118, 48, 83, 170, 123, 21, 152, 11, 81, 239, 82, 29, 137, 50, 145, 212, 100, 171, 107, 73, 118, 20, 117, 30, 124, 76, 137, 120, 128, 60, 0, 42, 0, 20, 79, 16, 82, 140, 98, 172, 115, 237, 116, 198, 128, 184, 117, 30, 125, 116, 136, 184, 99, 85, 141, 182, 78, 162, 100, 163, 100, 184, 111, 213, 158, 189, 129, 238, 158, 60, 0, 42, 0, 230, 101, 70, 79, 14, 94, 138, 108, 249, 108, 218, 114, 222, 139, 230, 103, 141, 117, 86, 85, 87, 85, 57, 95, 238, 96, 225, 109, 203, 134, 127, 85, 62, 95, 46, 108, 69, 129, 15, 132, 219, 137, 214, 53, 158, 122, 228, 80, 17, 135, 190, 153, 231, 154, 62, 86, 73, 86, 72, 95, 154, 97, 149, 138, 186, 97, 186, 102, 185, 111, 222, 132, 32, 157, 171, 121, 243, 153, 90, 118, 170, 86, 117, 126, 9, 141, 46, 151, 79, 153, 53, 74, 60, 0, 42, 0, 101, 98, 125, 98, 60, 0, 42, 0, 83, 95, 240, 115, 219, 148, 92, 123, 198, 136, 118, 117, 121, 86, 162, 111, 171, 116, 57, 124, 97, 130, 247, 135, 96, 137, 58, 148, 60, 0, 42, 0, 33, 99, 90, 81, 32, 140, 203, 100, 97, 139, 232, 158, 41, 101, 89, 112, 19, 107, 156, 139, 60, 0, 42, 0, 57, 108, 252, 81, 53, 87, 149, 91, 0, 120, 177, 87, 99, 104, 97, 131, 120, 90, 19, 97, 234, 131, 99, 93, 253, 116, 63, 144, 252, 150, 173, 120, 5, 81, 82, 111, 138, 119, 164, 141, 203, 88, 105, 133, 148, 106, 151, 116, 234, 118, 17, 121, 28, 124, 47, 134, 227, 149, 60, 0, 42, 0, 111, 112, 123, 118, 75, 140, 84, 86, 1, 91, 200, 113, 146, 116, 243, 122, 38, 124, 180, 137, 108, 142, 60, 0, 42, 0, 41, 103, 73, 123, 37, 98, 60, 0, 42, 0, 147, 144, 243, 81, 177, 88, 157, 93, 39, 145, 165, 150, 170, 119, 244, 120, 107, 149, 200, 106, 25, 148, 60, 0, 42, 0, 96, 130, 60, 0, 42, 0, 16, 108, 254, 78, 78, 79, 67, 89, 125, 95, 157, 127, 155, 136, 36, 88, 134, 141, 132, 150, 244, 110, 0, 106, 93, 149, 254, 120, 73, 147, 174, 151, 60, 0, 42, 0, 196, 114, 248, 94, 116, 124, 214, 130, 234, 143, 25, 85, 76, 101, 164, 109, 145, 104, 27, 123, 123, 131, 204, 137, 110, 151, 176, 153, 226, 154, 0, 86, 225, 90, 204, 110, 223, 127, 20, 152, 117, 101, 11, 133, 16, 133, 0, 74, 129, 86, 244, 123, 36, 74, 116, 140, 98, 142, 4, 155, 225, 133, 209, 147, 174, 66, 244, 124, 191, 137, 16, 158, 60, 0, 42, 0, 142, 83, 88, 87, 203, 139, 103, 84, 149, 94, 36, 95, 181, 98, 222, 98, 184, 144, 58, 150, 226, 103, 116, 114, 11, 131, 37, 120, 222, 87, 139, 99, 125, 59, 231, 131, 221, 137, 70, 138, 231, 142, 92, 128, 182, 154, 60, 0, 42, 0, 84, 87, 31, 95, 243, 101, 85, 103, 147, 115, 31, 96, 228, 79, 29, 94, 202, 87, 35, 90, 89, 80, 71, 85, 114, 85, 138, 104, 13, 113, 246, 115, 49, 119, 44, 123, 27, 140, 18, 144, 19, 144, 241, 145, 130, 90, 227, 104, 39, 110, 71, 119, 118, 121, 20, 127, 194, 131, 192, 80, 99, 129, 130, 132, 106, 146, 145, 88, 85, 100, 178, 120, 152, 121, 67, 135, 94, 144, 172, 88, 120, 97, 224, 125, 21, 133, 112, 144, 179, 93, 11, 117, 230, 138, 54, 142, 174, 135, 60, 0, 42, 0, 243, 155, 60, 0, 42, 0, 242, 85, 60, 0, 42, 0, 57, 114, 204, 141, 58, 137, 60, 0, 42, 0, 164, 87, 204, 92, 78, 96, 21, 99, 51, 102, 214, 126, 197, 128, 245, 130, 237, 143, 222, 116, 35, 119, 28, 98, 242, 67, 13, 140, 139, 85, 30, 88, 245, 96, 242, 99, 115, 117, 112, 125, 11, 128, 247, 129, 68, 138, 131, 141, 123, 149, 224, 83, 156, 107, 67, 114, 82, 114, 125, 93, 159, 120, 40, 135, 83, 130, 118, 135, 11, 137, 75, 216, 163, 221, 220, 138, 64, 142, 189, 156, 42, 59, 225, 102, 137, 117, 8, 156, 138, 117, 14, 108, 60, 0, 42, 0, 203, 84, 10, 128, 48, 119, 60, 0, 42, 0, 73, 94, 130, 117, 60, 0, 42, 0, 1, 82, 252, 83, 72, 108, 109, 134, 203, 81, 93, 89, 52, 95, 107, 95, 193, 134, 49, 116, 130, 140, 137, 120, 237, 156, 166, 107, 151, 119, 213, 150, 137, 155, 183, 156, 38, 159, 219, 155, 112, 157, 60, 0, 42, 0, 90, 98, 76, 92, 60, 0, 42, 0, 20, 95, 4, 79, 10, 84, 147, 148, 142, 122, 11, 138, 3, 140, 137, 99, 227, 145, 222, 148, 235, 148, 232, 122, 177, 146, 255, 150, 97, 155, 231, 132, 191, 138, 57, 118, 181, 122, 253, 146, 203, 133, 67, 148, 60, 0, 42, 0, 19, 124, 60, 0, 42, 0, 31, 78, 34, 78, 67, 216, 141, 220, 229, 148, 169, 146, 60, 0, 42, 0, 56, 117, 65, 101, 67, 216, 167, 221, 130, 99, 142, 80, 167, 83, 110, 93, 199, 110, 199, 105, 217, 105, 40, 118, 160, 152, 78, 142, 197, 93, 90, 152, 91, 152, 107, 118, 211, 93, 212, 93, 39, 101, 114, 118, 123, 159, 60, 0, 42, 0, 120, 81, 76, 89, 185, 112, 112, 90, 32, 115, 68, 67, 95, 101, 213, 141, 152, 120, 167, 132, 46, 142, 71, 133, 222, 158, 184, 86, 60, 0, 42, 0, 53, 117, 67, 79, 107, 87, 151, 94, 61, 150, 171, 87, 66, 98, 183, 115, 191, 148, 93, 90, 230, 96, 192, 109, 96, 89, 20, 116, 191, 107, 20, 135, 251, 150, 138, 88, 194, 88, 66, 106, 93, 106, 177, 111, 91, 151, 82, 216, 113, 222, 92, 118, 31, 124, 84, 154, 60, 0, 42, 0, 35, 105, 60, 0, 42, 0, 1, 78, 195, 78, 238, 83, 4, 94, 142, 115, 148, 117, 239, 118, 137, 148, 53, 128, 112, 134, 74, 145, 216, 145, 106, 151, 60, 0, 42, 0, 117, 89, 118, 152, 2, 152, 192, 58, 14, 159, 127, 93, 17, 159, 206, 111, 161, 133, 36, 148, 60, 0, 42, 0, 162, 139, 202, 95, 100, 153, 244, 119, 154, 91, 2, 138, 227, 152, 118, 85, 228, 148, 23, 105, 90, 129, 35, 132, 135, 120, 45, 149, 160, 120, 74, 135, 204, 146, 32, 147, 248, 120, 65, 152, 60, 0, 42, 0, 98, 128, 60, 0, 42, 0, 190, 83, 71, 216, 42, 220, 98, 82, 7, 150, 99, 119, 31, 86, 143, 145, 205, 149, 60, 0, 42, 0, 210, 107, 236, 114, 156, 109, 251, 139, 14, 110, 31, 105, 105, 60, 77, 114, 138, 114, 161, 120, 251, 136, 173, 138, 115, 135, 104, 115, 22, 147, 223, 81, 53, 83, 59, 91, 6, 112, 221, 106, 176, 107, 88, 114, 162, 114, 196, 116, 190, 118, 51, 154, 233, 158, 128, 139, 68, 140, 21, 141, 227, 151, 95, 148, 209, 154, 199, 151, 229, 151, 247, 158, 159, 139, 60, 0, 42, 0, 3, 123, 53, 88, 62, 94, 76, 141, 61, 116, 121, 119, 169, 137, 237, 140, 228, 123, 60, 0, 42, 0, 146, 89, 92, 103, 154, 128, 143, 130, 172, 89, 166, 94, 250, 121, 112, 131, 33, 110, 111, 151, 64, 149, 103, 216, 90, 221, 153, 135, 172, 107, 77, 147, 53, 124, 39, 136, 57, 136, 60, 0, 42, 0, 26, 89, 27, 89, 132, 84, 198, 84, 83, 117, 95, 82, 28, 93, 135, 99, 96, 101, 242, 107, 240, 136, 137, 86, 60, 0, 42, 0, 58, 89, 206, 148, 107, 82, 83, 101, 90, 101, 165, 85, 179, 96, 106, 101, 229, 117, 44, 146, 106, 89, 217, 81, 49, 142, 181, 155, 56, 148, 60, 0, 42, 0, 53, 103, 54, 103, 218, 84, 155, 87, 156, 87, 5, 99, 6, 99, 245, 87, 13, 127, 47, 105, 147, 141, 177, 142, 178, 142, 158, 125, 42, 73, 156, 97, 184, 78, 122, 147, 195, 142, 178, 86, 114, 89, 60, 0, 42, 0, 52, 82, 65, 82, 115, 153, 46, 92, 193, 103, 238, 103, 168, 112, 74, 150, 79, 150, 87, 104, 21, 88, 53, 130, 240, 96, 226, 141, 229, 141, 250, 141, 255, 152, 231, 62, 174, 88, 175, 88, 158, 93, 125, 157, 60, 0, 42, 0, 164, 103, 60, 0, 42, 0, 150, 87, 6, 88, 96, 88, 95, 93, 253, 117, 211, 120, 45, 157, 28, 148, 60, 0, 42, 0, 3, 58, 39, 152, 60, 0, 42, 0, 249, 91, 31, 150, 74, 81, 76, 81, 81, 81, 254, 91, 60, 96, 75, 121, 110, 150, 138, 150, 147, 120, 207, 65, 144, 125, 13, 92, 158, 97, 157, 97, 231, 111, 102, 149, 223, 97, 104, 137, 35, 59, 177, 133, 41, 112, 72, 139, 19, 148, 60, 0, 42, 0, 17, 128, 115, 80, 108, 82, 143, 90, 239, 122, 13, 137, 116, 147, 60, 0, 42, 0, 74, 216, 67, 223, 237, 119, 60, 0, 42, 0, 181, 107, 173, 101, 69, 88, 14, 127, 52, 105, 69, 113, 86, 116, 118, 129, 110, 132, 171, 120, 59, 149, 222, 125, 200, 107, 22, 124, 91, 147, 183, 101, 150, 142, 106, 124, 60, 0, 42, 0, 40, 84, 199, 96, 102, 101, 51, 135, 169, 88, 170, 88, 180, 100, 100, 115, 120, 86, 137, 100, 84, 106, 156, 114, 5, 121, 59, 69, 114, 142, 126, 142, 80, 154, 60, 0, 42, 0, 76, 216, 180, 223, 249, 118, 184, 141, 137, 142, 60, 0, 42, 0, 5, 79, 228, 86, 137, 94, 140, 108, 150, 112, 254, 118, 24, 120, 157, 148, 7, 144, 127, 152, 13, 146, 111, 105, 65, 144, 19, 152, 97, 111, 111, 144, 201, 113, 50, 142, 60, 0, 42, 0, 183, 120, 60, 0, 42, 0, 28, 78, 172, 81, 154, 84, 181, 54, 189, 92, 113, 103, 56, 102, 33, 108, 243, 130, 50, 80, 43, 158, 236, 87, 59, 90, 32, 93, 44, 93, 183, 109, 23, 123, 154, 95, 45, 108, 196, 131, 100, 216, 147, 220, 64, 135, 36, 157, 21, 159, 223, 155, 135, 157, 60, 0, 42, 0, 2, 58, 99, 132, 165, 88, 189, 123, 204, 138, 30, 91, 79, 216, 197, 223, 107, 133, 194, 97, 60, 0, 42, 0, 168, 82, 187, 81, 151, 79, 140, 87, 219, 89, 210, 92, 107, 96, 15, 99, 11, 104, 30, 109, 232, 128, 205, 81, 25, 98, 244, 128, 245, 143, 213, 82, 80, 120, 223, 104, 105, 110, 103, 125, 86, 129, 205, 80, 231, 153, 24, 151, 60, 0, 42, 0, 151, 155, 171, 157, 60, 0, 39, 0, 208, 253, 39, 0, 10, 49, 60, 0, 42, 0, 214, 78, 131, 91, 121, 89, 96, 114, 66, 121, 202, 148, 191, 141, 76, 88, 187, 110, 153, 105, 31, 137, 131, 86, 231, 149, 60, 0, 42, 0, 121, 142, 60, 0, 42, 0, 84, 88, 154, 110, 150, 88, 243, 53, 109, 115, 206, 156, 122, 115, 40, 156, 60, 0, 42, 0, 163, 78, 211, 98, 30, 99, 231, 114, 252, 149, 67, 216, 50, 221, 229, 54, 9, 93, 190, 109, 124, 59, 40, 100, 191, 64, 246, 141, 187, 105, 254, 107, 11, 66, 93, 144, 98, 144, 147, 52, 162, 121, 187, 138, 15, 142, 187, 100, 190, 111, 20, 147, 125, 106, 204, 111, 75, 142, 156, 151, 153, 155, 249, 63, 210, 149, 179, 151, 186, 86, 118, 139, 229, 149, 105, 216, 1, 222, 162, 142, 60, 0, 42, 0, 164, 79, 156, 84, 60, 0, 42, 0, 209, 95, 210, 95, 121, 114, 163, 140, 174, 134, 253, 148, 93, 97, 241, 146, 163, 135, 216, 135, 60, 0, 42, 0, 252, 86, 97, 91, 206, 128, 60, 0, 42, 0, 173, 81, 240, 83, 242, 101, 110, 87, 172, 98, 176, 144, 177, 103, 177, 112, 178, 112, 212, 130, 237, 131, 198, 141, 3, 59, 144, 156, 136, 123, 250, 129, 177, 152, 26, 66, 216, 153, 19, 81, 144, 155, 47, 91, 225, 100, 0, 62, 175, 106, 185, 133, 73, 124, 60, 0, 42, 0, 42, 89, 51, 89, 242, 95, 112, 108, 1, 96, 189, 128, 155, 148, 240, 108, 38, 130, 237, 72, 94, 145, 38, 146, 153, 110, 75, 97, 228, 113, 60, 0, 42, 0, 143, 124, 60, 0, 42, 0, 50, 89, 34, 95, 155, 109, 230, 126, 143, 99, 92, 138, 205, 90, 77, 94, 70, 97, 47, 100, 212, 110, 91, 125, 196, 105, 107, 116, 236, 151, 248, 152, 49, 66, 26, 126, 39, 126, 228, 111, 31, 139, 65, 143, 62, 73, 177, 151, 220, 151, 85, 153, 60, 0, 42, 0, 11, 83, 183, 84, 46, 109, 239, 143, 67, 104, 3, 144, 85, 85, 188, 104, 216, 109, 59, 65, 249, 126, 118, 150, 4, 132, 121, 121, 175, 125, 42, 135, 234, 136, 128, 151, 132, 145, 137, 151, 254, 146, 45, 147, 227, 153, 174, 106, 64, 153, 10, 154, 23, 159, 60, 0, 42, 0, 168, 139, 14, 138, 60, 0, 42, 0, 87, 89, 60, 0, 42, 0, 119, 80, 120, 80, 126, 90, 174, 90, 192, 146, 110, 147, 60, 0, 42, 0, 160, 78, 52, 89, 149, 98, 176, 154, 240, 125, 45, 152, 60, 0, 42, 0, 181, 89, 173, 148, 15, 125, 151, 58, 104, 101, 243, 152, 200, 158, 35, 134, 60, 0, 42, 0, 15, 144, 137, 125, 60, 0, 42, 0, 77, 87, 49, 54, 169, 98, 42, 141, 57, 96, 209, 117, 17, 130, 170, 140, 74, 100, 233, 110, 43, 118, 249, 100, 36, 101, 88, 112, 113, 118, 60, 0, 42, 0, 91, 87, 68, 216, 168, 222, 25, 102, 19, 80, 8, 140, 82, 90, 212, 96, 239, 144, 131, 137, 131, 105, 240, 117, 44, 149, 45, 140, 176, 88, 181, 88, 155, 97, 109, 111, 199, 138, 136, 145, 199, 88, 199, 102, 194, 113, 31, 147, 36, 153, 128, 106, 249, 120, 67, 152, 72, 127, 220, 88, 235, 133, 90, 139, 154, 140, 176, 145, 96, 139, 78, 127, 60, 0, 42, 0, 208, 95, 102, 87, 217, 67, 189, 148, 146, 136, 239, 107, 252, 131, 161, 71, 109, 146, 255, 85, 179, 97, 187, 97, 147, 145, 174, 116, 98, 137, 60, 0, 42, 0, 249, 83, 173, 112, 238, 87, 162, 99, 157, 80, 96, 110, 203, 80, 6, 86, 179, 120, 21, 130, 78, 107, 231, 140, 60, 0, 42, 0, 100, 108, 99, 87, 244, 148, 111, 110, 33, 86, 37, 128, 143, 82, 176, 127, 106, 135, 182, 71, 0, 72, 87, 149, 154, 133, 90, 142, 220, 147, 11, 148, 186, 151, 30, 159, 60, 0, 42, 0, 103, 153, 16, 85, 2, 88, 143, 80, 122, 85, 123, 59, 224, 104, 125, 52, 88, 88, 13, 55, 42, 100, 143, 110, 12, 145, 182, 105, 31, 111, 123, 113, 109, 116, 133, 129, 206, 132, 154, 150, 24, 106, 77, 216, 95, 222, 196, 120, 159, 121, 195, 124, 155, 129, 86, 106, 214, 123, 214, 124, 151, 135, 60, 142, 219, 124, 179, 135, 111, 141, 163, 145, 51, 153, 78, 66, 149, 147, 57, 153, 219, 149, 68, 153, 96, 216, 137, 221, 182, 157, 60, 0, 42, 0, 22, 79, 17, 94, 24, 80, 82, 80, 204, 109, 165, 80, 186, 142, 75, 149, 178, 147, 59, 81, 3, 98, 237, 102, 35, 114, 216, 119, 130, 148, 60, 0, 42, 0, 235, 112, 101, 100, 159, 141, 217, 113, 60, 0, 42, 0, 165, 113, 175, 129, 31, 159, 60, 0, 42, 0, 188, 117, 203, 117, 80, 94, 126, 129, 138, 138, 27, 111, 213, 110, 34, 126, 134, 144, 230, 153, 4, 139, 47, 81, 228, 133, 48, 154, 80, 124, 39, 156, 88, 124, 99, 154, 60, 0, 42, 0, 47, 151, 60, 0, 42, 0, 69, 134, 60, 0, 42, 0, 84, 82, 175, 104, 56, 53, 17, 149, 34, 142, 255, 100, 200, 157, 201, 157, 60, 0, 42, 0, 151, 83, 208, 130, 232, 126, 81, 131, 77, 80, 124, 85, 57, 93, 255, 96, 208, 99, 10, 122, 7, 127, 100, 127, 72, 158, 193, 85, 69, 116, 136, 125, 70, 144, 174, 120, 178, 95, 61, 111, 249, 125, 109, 135, 6, 137, 187, 146, 152, 152, 91, 133, 167, 141, 68, 142, 141, 145, 21, 139, 79, 142, 87, 147, 192, 156, 58, 157, 76, 152, 183, 155, 92, 157, 32, 154, 247, 155, 151, 157, 153, 157, 181, 121, 228, 157, 60, 0, 42, 0, 83, 79, 46, 99, 176, 142, 181, 154, 167, 155, 198, 142, 212, 154, 60, 0, 42, 0, 59, 98, 78, 216, 194, 221, 73, 92, 207, 143, 67, 82, 17, 103, 31, 109, 28, 80, 140, 96, 149, 109, 92, 92, 144, 96, 213, 96, 166, 99, 22, 144, 214, 96, 229, 99, 255, 102, 55, 144, 116, 105, 252, 136, 82, 107, 162, 107, 5, 137, 240, 154, 49, 59, 143, 86, 153, 133, 0, 155, 148, 86, 203, 116, 74, 124, 175, 141, 145, 77, 60, 0, 42, 0, 57, 123, 92, 93, 60, 0, 42, 0, 22, 94, 23, 96, 52, 141, 81, 128, 28, 132, 188, 140, 60, 0, 42, 0, 193, 148, 200, 134, 227, 80, 149, 146, 232, 146, 41, 157, 33, 148, 53, 148, 86, 154, 60, 0, 42, 0, 107, 84, 251, 152, 52, 77, 46, 153, 60, 0, 42, 0, 235, 101, 123, 79, 163, 94, 76, 96, 17, 99, 103, 121, 56, 59, 78, 128, 60, 0, 42, 0, 128, 130, 97, 103, 167, 92, 185, 92, 226, 143, 82, 121, 157, 104, 36, 123, 20, 132, 237, 71, 146, 141, 134, 159, 218, 132, 41, 135, 36, 106, 232, 132, 218, 146, 235, 154, 166, 156, 151, 151, 165, 147, 200, 155, 96, 159, 55, 156, 60, 0, 42, 0, 168, 91, 64, 102, 19, 103, 32, 62, 1, 129, 149, 122, 130, 138, 55, 72, 177, 122, 162, 101, 37, 91, 60, 0, 42, 0, 58, 119, 156, 124, 105, 125, 156, 137, 243, 141, 246, 124, 60, 0, 42, 0, 169, 135, 60, 0, 42, 0, 41, 89, 114, 81, 86, 90, 251, 109, 95, 145, 84, 151, 236, 66, 50, 59, 199, 158, 93, 151, 60, 0, 42, 0, 48, 117, 71, 92, 186, 108, 108, 96, 75, 117, 81, 117, 247, 118, 203, 128, 96, 117, 27, 117, 28, 117, 73, 110, 254, 131, 82, 72, 97, 88, 107, 88, 55, 100, 63, 146, 23, 150, 194, 125, 204, 120, 180, 122, 179, 116, 208, 149, 198, 157, 207, 157, 60, 0, 42, 0, 221, 95, 132, 107, 14, 80, 173, 53, 58, 85, 73, 54, 191, 96, 223, 109, 106, 102, 32, 116, 70, 129, 205, 137, 246, 117, 83, 119, 20, 130, 2, 153, 165, 137, 223, 140, 42, 147, 102, 151, 105, 147, 60, 0, 42, 0, 1, 52, 173, 99, 124, 119, 26, 130, 60, 0, 42, 0, 181, 120, 43, 157, 60, 0, 42, 0, 133, 83, 129, 94, 64, 108, 124, 130, 44, 84, 58, 117, 19, 128, 155, 83, 195, 112, 111, 104, 244, 112, 142, 125, 252, 66, 147, 151, 116, 128, 240, 94, 124, 128, 125, 128, 243, 94, 60, 0, 42, 0, 146, 144, 247, 94, 173, 78, 173, 94, 92, 80, 155, 131, 119, 90, 73, 93, 31, 110, 95, 105, 115, 123, 118, 132, 19, 135, 179, 105, 100, 128, 79, 135, 174, 149, 6, 151, 234, 138, 46, 159, 60, 0, 42, 0, 68, 216, 60, 222, 34, 87, 60, 117, 185, 79, 23, 90, 58, 99, 143, 109, 131, 104, 246, 112, 253, 115, 33, 129, 139, 152, 71, 130, 148, 138, 50, 152, 104, 216, 182, 222, 60, 0, 42, 0, 248, 81, 138, 91, 191, 121, 195, 121, 34, 96, 129, 122, 11, 85, 139, 109, 120, 99, 23, 88, 101, 110, 220, 117, 86, 132, 128, 93, 245, 146, 90, 157, 53, 159, 60, 0, 42, 0, 243, 86, 254, 86, 195, 81, 249, 92, 169, 94, 146, 95, 135, 96, 72, 99, 130, 109, 96, 92, 140, 104, 204, 62, 124, 131, 20, 144, 236, 99, 120, 59, 12, 122, 223, 131, 21, 87, 87, 88, 94, 93, 76, 216, 37, 222, 15, 118, 97, 123, 111, 129, 47, 146, 22, 87, 23, 87, 220, 94, 115, 111, 164, 132, 255, 141, 116, 145, 221, 72, 159, 153, 78, 147, 252, 153, 76, 157, 159, 157, 203, 157, 245, 157, 60, 0, 42, 0, 31, 87, 33, 87, 16, 84, 141, 148, 247, 145, 60, 0, 42, 0, 78, 81, 84, 81, 204, 143, 13, 88, 117, 157, 60, 0, 42, 0, 98, 108, 181, 131, 60, 0, 42, 0, 71, 78, 219, 78, 172, 139, 88, 98, 97, 98, 81, 108, 102, 153, 84, 103, 130, 79, 131, 84, 213, 98, 214, 98, 176, 108, 41, 99, 93, 99, 23, 138, 182, 109, 43, 129, 49, 129, 140, 131, 165, 136, 229, 152, 96, 155, 93, 154, 60, 0, 42, 0, 110, 154, 87, 79, 104, 87, 174, 92, 177, 108, 178, 108, 207, 114, 64, 150, 65, 150, 124, 154, 241, 143, 35, 120, 36, 120, 53, 158, 61, 125, 137, 136, 54, 88, 206, 141, 97, 145, 177, 153, 214, 105, 162, 120, 126, 75, 196, 153, 221, 153, 222, 153, 80, 106, 128, 155, 21, 157, 39, 159, 40, 154, 13, 159, 82, 154, 9, 159, 60, 0, 42, 0, 117, 95, 165, 89, 185, 94, 160, 90, 45, 105, 85, 105, 247, 90, 98, 106, 78, 157, 12, 155, 22, 156, 60, 0, 42, 0, 221, 103, 228, 107, 62, 85, 26, 132, 197, 141, 251, 107, 168, 123, 0, 134, 92, 124, 60, 0, 42, 0, 168, 99, 247, 132, 236, 133, 60, 0, 42, 0, 26, 95, 3, 64, 147, 152, 53, 92, 164, 150, 57, 152, 58, 152, 61, 152, 75, 155, 104, 122, 106, 142, 8, 134, 215, 63, 60, 0, 42, 0, 192, 79, 211, 80, 127, 129, 70, 142, 189, 154, 60, 0, 42, 0, 187, 79, 39, 90, 0, 144, 122, 113, 251, 134, 21, 135, 42, 137, 254, 153, 60, 0, 42, 0, 77, 110, 47, 115, 83, 113, 78, 71, 146, 140, 78, 77, 60, 0, 42, 0, 226, 86, 227, 86, 159, 98, 120, 82, 24, 87, 113, 97, 118, 100, 25, 111, 235, 105, 255, 123, 178, 106, 196, 147, 240, 124, 210, 157, 251, 157, 60, 0, 42, 0, 131, 117, 60, 0, 42, 0, 86, 95, 106, 110, 22, 137, 60, 0, 42, 0, 30, 84, 81, 84, 146, 109, 77, 85, 28, 103, 30, 113, 75, 86, 190, 102, 63, 59, 215, 158, 60, 0, 42, 0, 111, 92, 73, 87, 243, 95, 104, 153, 154, 130, 88, 140, 90, 140, 216, 142, 233, 152, 128, 156, 104, 155, 84, 58, 21, 151, 192, 129, 203, 129, 60, 0, 42, 0, 61, 108, 125, 117, 60, 0, 42, 0, 253, 101, 60, 0, 42, 0, 242, 86, 181, 112, 204, 117, 26, 144, 245, 85, 234, 132, 60, 0, 42, 0, 221, 78, 12, 84, 95, 79, 100, 95, 194, 92, 157, 94, 195, 84, 221, 92, 234, 114, 77, 102, 80, 104, 117, 109, 212, 112, 60, 120, 60, 131, 146, 134, 46, 119, 241, 121, 220, 148, 229, 122, 161, 124, 90, 216, 52, 222, 105, 123, 119, 138, 104, 141, 110, 145, 86, 146, 238, 80, 237, 82, 117, 146, 133, 146, 7, 153, 150, 156, 124, 111, 94, 115, 200, 102, 35, 103, 102, 106, 3, 108, 209, 113, 157, 114, 167, 129, 63, 63, 179, 119, 166, 155, 89, 216, 141, 220, 0, 77, 60, 0, 42, 0, 223, 126, 69, 99, 140, 58, 118, 104, 82, 123, 113, 125, 130, 125, 11, 106, 60, 0, 42, 0, 120, 96, 219, 117, 85, 136, 95, 97, 133, 97, 60, 0, 39, 0, 208, 253, 39, 0, 11, 49, 60, 0, 42, 0, 239, 85, 60, 0, 42, 0, 207, 98, 255, 98, 16, 99, 241, 85, 78, 149, 191, 147, 60, 0, 42, 0, 120, 78, 234, 84, 235, 150, 60, 0, 42, 0, 54, 84, 80, 84, 160, 89, 179, 126, 163, 144, 173, 128, 160, 148, 28, 90, 13, 125, 114, 136, 122, 99, 29, 123, 87, 216, 73, 222, 166, 136, 125, 140, 220, 142, 128, 140, 9, 146, 121, 151, 179, 132, 83, 216, 59, 222, 118, 155, 43, 66, 60, 0, 42, 0, 146, 117, 183, 139, 144, 98, 50, 119, 37, 138, 60, 0, 42, 0, 98, 84, 60, 0, 42, 0, 73, 129, 139, 113, 104, 100, 123, 91, 60, 0, 42, 0, 67, 78, 118, 89, 22, 108, 127, 130, 147, 117, 179, 89, 252, 94, 55, 80, 250, 143, 226, 145, 45, 91, 60, 0, 42, 0, 72, 89, 240, 103, 15, 128, 16, 128, 69, 216, 171, 223, 63, 110, 24, 132, 16, 159, 154, 135, 38, 137, 60, 147, 60, 0, 42, 0, 30, 90, 129, 153, 46, 129, 71, 129, 18, 153, 190, 155, 216, 155, 60, 0, 42, 0, 103, 81, 133, 81, 59, 52, 29, 108, 23, 147, 60, 0, 42, 0, 108, 91, 60, 0, 42, 0, 118, 84, 19, 96, 32, 99, 241, 92, 71, 120, 217, 148, 49, 115, 242, 134, 73, 138, 153, 120, 147, 100, 169, 93, 185, 97, 239, 135, 18, 89, 74, 139, 3, 148, 206, 93, 60, 0, 42, 0, 180, 87, 124, 96, 169, 96, 17, 129, 24, 83, 22, 88, 241, 96, 208, 90, 89, 116, 102, 129, 175, 120, 0, 71, 118, 115, 127, 115, 60, 0, 42, 0, 249, 149, 101, 90, 214, 109, 153, 149, 39, 155, 209, 129, 60, 0, 42, 0, 51, 129, 60, 0, 42, 0, 186, 127, 60, 0, 42, 0, 66, 85, 60, 0, 42, 0, 200, 105, 40, 128, 115, 115, 189, 106, 146, 147, 30, 148, 115, 139, 60, 0, 42, 0, 225, 86, 60, 0, 42, 0, 55, 117, 143, 103, 172, 103, 189, 79, 87, 83, 223, 103, 26, 90, 88, 117, 190, 150, 174, 131, 131, 85, 148, 102, 96, 105, 245, 138, 227, 150, 93, 216, 116, 221, 60, 0, 42, 0, 103, 141, 199, 99, 115, 110, 105, 129, 51, 132, 18, 66, 123, 135, 1, 98, 60, 0, 42, 0, 89, 216, 218, 222, 123, 90, 46, 59, 60, 0, 42, 0, 86, 144, 60, 0, 42, 0, 65, 96, 233, 90, 240, 90, 60, 0, 42, 0, 212, 86, 60, 0, 42, 0, 106, 78, 162, 86, 104, 139, 202, 86, 48, 136, 30, 155, 149, 153, 28, 107, 98, 153, 60, 0, 42, 0, 195, 100, 233, 102, 46, 101, 98, 112, 60, 0, 42, 0, 62, 81, 73, 159, 60, 0, 42, 0, 253, 128, 60, 0, 42, 0, 67, 216, 76, 222, 99, 216, 153, 221, 60, 0, 42, 0, 174, 89, 60, 0, 42, 0, 60, 92, 109, 87, 190, 55, 41, 96, 229, 108, 126, 124, 42, 80, 84, 92, 220, 121, 204, 148, 255, 87, 87, 90, 227, 109, 10, 115, 173, 134, 38, 70, 243, 144, 255, 104, 82, 216, 164, 221, 207, 70, 220, 141, 93, 129, 99, 128, 58, 135, 236, 137, 142, 140, 23, 143, 19, 151, 181, 156, 211, 155, 226, 155, 145, 158, 111, 159, 225, 129, 60, 0, 42, 0, 49, 79, 96, 79, 179, 98, 223, 98, 212, 114, 197, 103, 232, 130, 206, 101, 114, 102, 116, 91, 46, 146, 156, 153, 23, 81, 30, 81, 236, 100, 172, 150, 183, 106, 191, 133, 123, 128, 60, 0, 42, 0, 112, 92, 60, 108, 50, 79, 79, 216, 183, 220, 53, 102, 210, 128, 225, 143, 36, 119, 6, 144, 63, 83, 4, 88, 196, 96, 53, 97, 186, 110, 104, 119, 123, 129, 223, 90, 177, 102, 189, 138, 101, 64, 12, 126, 169, 129, 58, 91, 105, 216, 91, 222, 60, 0, 42, 0, 174, 136, 60, 0, 42, 0, 79, 99, 209, 99, 60, 0, 42, 0, 246, 130, 60, 0, 42, 0, 7, 94, 89, 216, 18, 222, 60, 87, 191, 103, 133, 109, 198, 117, 66, 128, 236, 129, 103, 150, 110, 85, 215, 96, 166, 85, 92, 101, 124, 110, 205, 131, 137, 150, 235, 85, 114, 93, 2, 142, 91, 86, 112, 100, 247, 105, 23, 142, 74, 149, 77, 149, 173, 93, 222, 123, 242, 129, 28, 147, 158, 152, 81, 142, 153, 86, 118, 128, 179, 147, 209, 149, 124, 91, 56, 55, 125, 91, 241, 106, 75, 124, 193, 86, 22, 134, 103, 159, 76, 216, 51, 220, 241, 124, 37, 136, 72, 148, 245, 124, 211, 86, 152, 139, 161, 142, 119, 148, 115, 152, 128, 148, 192, 75, 60, 0, 42, 0, 213, 93, 60, 0, 42, 0, 31, 158, 17, 131, 100, 59, 133, 136, 229, 156, 203, 90, 202, 136, 34, 106, 38, 133, 29, 91, 45, 137, 50, 91, 53, 55, 60, 0, 42, 0, 63, 92, 50, 129, 60, 0, 42, 0, 158, 89, 60, 0, 42, 0, 167, 46, 91, 114, 124, 108, 60, 0, 42, 0, 248, 95, 109, 98, 195, 114, 189, 126, 132, 112, 174, 148, 16, 125, 165, 131, 21, 146, 117, 151, 60, 0, 42, 0, 66, 136, 244, 66, 60, 0, 42, 0, 92, 114, 60, 0, 42, 0, 200, 98, 43, 133, 60, 0, 42, 0, 116, 94, 202, 121, 214, 84, 229, 121, 135, 156, 142, 155, 182, 156, 207, 158, 240, 155, 60, 0, 42, 0, 138, 109, 123, 99, 240, 109, 62, 113, 200, 141, 135, 143, 151, 143, 154, 100, 181, 100, 190, 120, 38, 143, 16, 124, 77, 142, 6, 101, 104, 142, 142, 142, 60, 0, 42, 0, 68, 83, 255, 94, 245, 95, 233, 89, 56, 85, 221, 87, 76, 130, 48, 159, 60, 0, 42, 0, 71, 157, 60, 0, 42, 0, 220, 86, 168, 96, 251, 66, 60, 0, 42, 0, 240, 98, 60, 0, 42, 0, 12, 129, 60, 0, 42, 0, 24, 90, 34, 91, 67, 91, 60, 0, 42, 0, 127, 145, 184, 145, 192, 145, 60, 0, 42, 0, 129, 91, 155, 84, 231, 98, 222, 114, 224, 103, 231, 130, 77, 128, 205, 91, 213, 91, 47, 117, 215, 91, 220, 91, 231, 91, 28, 81, 221, 81, 128, 86, 35, 91, 240, 100, 112, 115, 184, 106, 180, 133, 121, 128, 94, 216, 72, 223, 79, 148, 33, 155, 11, 158, 60, 0, 42, 0, 99, 106, 195, 119, 60, 0, 42, 0, 94, 79, 171, 79, 222, 108, 216, 111, 60, 0, 42, 0, 157, 111, 60, 0, 42, 0, 116, 89, 101, 91, 125, 154, 47, 123, 209, 153, 60, 0, 42, 0, 46, 79, 170, 82, 41, 95, 46, 120, 236, 128, 60, 0, 42, 0, 18, 96, 137, 80, 25, 100, 60, 0, 42, 0, 205, 144, 42, 99, 155, 104, 169, 80, 58, 81, 60, 0, 42, 0, 96, 106, 60, 0, 42, 0, 250, 139, 143, 85, 30, 97, 73, 216, 42, 223, 191, 99, 38, 100, 61, 144, 24, 149, 59, 100, 146, 105, 44, 122, 209, 124, 254, 138, 67, 142, 230, 97, 231, 97, 229, 124, 100, 122, 239, 124, 60, 0, 42, 0, 123, 89, 60, 0, 42, 0, 28, 110, 9, 59, 150, 102, 86, 113, 87, 113, 42, 153, 60, 0, 42, 0, 193, 158, 60, 0, 42, 0, 156, 81, 172, 79, 221, 84, 83, 109, 19, 129, 254, 121, 178, 143, 2, 81, 179, 143, 101, 86, 195, 111, 130, 106, 246, 113, 191, 129, 125, 133, 175, 121, 96, 122, 91, 137, 178, 145, 1, 107, 60, 0, 42, 0, 119, 126, 60, 0, 42, 0, 4, 95, 10, 99, 81, 118, 72, 159, 60, 0, 42, 0, 115, 89, 149, 148, 121, 124, 249, 145, 60, 0, 42, 0, 145, 108, 103, 96, 18, 103, 68, 136, 188, 70, 60, 0, 42, 0, 159, 117, 80, 134, 120, 120, 39, 118, 60, 0, 39, 0, 208, 253, 39, 0, 12, 49, 60, 0, 42, 0, 131, 87, 201, 98, 198, 103, 203, 127, 200, 131, 26, 100, 139, 144, 60, 0, 42, 0, 239, 101, 76, 82, 44, 120, 230, 99, 214, 120, 185, 86, 60, 0, 42, 0, 135, 85, 222, 133, 60, 0, 42, 0, 74, 129, 231, 99, 75, 105, 12, 118, 33, 135, 75, 135, 162, 143, 163, 143, 114, 135, 200, 129, 11, 101, 9, 114, 216, 129, 14, 155, 206, 116, 116, 149, 251, 155, 31, 136, 94, 148, 60, 0, 42, 0, 102, 85, 130, 110, 161, 151, 60, 0, 42, 0, 214, 86, 60, 0, 42, 0, 139, 128, 60, 0, 42, 0, 194, 78, 57, 53, 80, 78, 251, 83, 199, 95, 80, 98, 59, 108, 30, 150, 143, 115, 123, 130, 208, 108, 251, 122, 51, 120, 125, 105, 247, 151, 2, 106, 21, 124, 161, 57, 211, 156, 51, 156, 60, 0, 42, 0, 134, 78, 121, 153, 14, 153, 60, 0, 42, 0, 101, 103, 134, 79, 235, 79, 8, 80, 3, 93, 149, 95, 158, 109, 97, 90, 13, 93, 178, 94, 160, 95, 190, 104, 246, 109, 13, 115, 177, 131, 242, 144, 246, 104, 28, 116, 89, 123, 10, 132, 40, 144, 252, 148, 130, 123, 241, 66, 96, 216, 144, 220, 56, 147, 11, 154, 224, 155, 134, 157, 179, 158, 60, 0, 42, 0, 59, 85, 73, 141, 80, 119, 94, 119, 86, 141, 218, 140, 209, 111, 244, 140, 60, 152, 66, 152, 68, 216, 150, 220, 94, 118, 99, 157, 40, 112, 44, 112, 65, 124, 244, 106, 10, 62, 254, 133, 105, 118, 95, 124, 112, 137, 60, 0, 42, 0, 210, 82, 60, 0, 42, 0, 247, 150, 216, 90, 39, 127, 126, 117, 2, 133, 194, 100, 98, 61, 145, 106, 50, 126, 12, 121, 109, 149, 209, 106, 195, 116, 184, 127, 39, 121, 141, 126, 77, 127, 29, 136, 51, 148, 50, 134, 96, 143, 61, 81, 232, 88, 88, 148, 65, 151, 25, 107, 70, 134, 157, 126, 58, 159, 60, 0, 42, 0, 189, 83, 18, 128, 212, 139, 146, 87, 107, 125, 66, 129, 171, 80, 132, 138, 15, 106, 202, 120, 229, 120, 76, 133, 33, 81, 66, 66, 126, 133, 216, 88, 125, 61, 87, 118, 208, 106, 223, 133, 40, 121, 69, 112, 132, 139, 61, 134, 120, 148, 19, 158, 60, 0, 42, 0, 234, 108, 33, 109, 123, 124, 153, 109, 218, 109, 47, 125, 96, 216, 36, 220, 121, 145, 135, 146, 27, 152, 42, 152, 17, 147, 174, 72, 2, 101, 163, 152, 94, 152, 106, 66, 135, 126, 49, 134, 183, 121, 60, 0, 42, 0, 65, 88, 30, 86, 105, 156, 60, 0, 42, 0, 94, 99, 136, 100, 60, 0, 42, 0, 179, 82, 180, 82, 98, 114, 130, 122, 240, 84, 32, 85, 2, 93, 118, 109, 222, 82, 232, 117, 249, 148, 215, 80, 46, 86, 160, 55, 151, 93, 165, 97, 82, 216, 66, 222, 70, 118, 241, 120, 41, 124, 231, 135, 170, 145, 18, 148, 95, 152, 221, 154, 60, 0, 42, 0, 2, 128, 1, 128, 108, 79, 190, 84, 229, 89, 69, 96, 235, 114, 51, 104, 86, 131, 209, 148, 237, 65, 160, 146, 102, 111, 81, 106, 81, 143, 60, 0, 42, 0, 157, 109, 217, 112, 34, 128, 106, 145, 234, 90, 166, 97, 135, 111, 188, 142, 111, 106, 46, 128, 194, 142, 60, 0, 42, 0, 239, 115, 83, 120, 89, 216, 232, 222, 169, 124, 239, 134, 37, 103, 177, 155, 60, 0, 42, 0, 156, 119, 60, 0, 42, 0, 69, 82, 4, 90, 123, 80, 65, 90, 189, 85, 135, 110, 194, 80, 124, 105, 140, 132, 13, 86, 212, 94, 122, 97, 10, 111, 19, 106, 161, 113, 39, 128, 30, 133, 124, 135, 113, 144, 44, 128, 91, 130, 187, 135, 49, 139, 193, 142, 197, 154, 187, 151, 207, 154, 60, 0, 42, 0, 93, 93, 2, 100, 127, 88, 129, 93, 95, 100, 211, 123, 10, 117, 13, 124, 60, 0, 42, 0, 55, 53, 75, 150, 90, 92, 15, 111, 24, 118, 66, 149, 58, 118, 59, 118, 228, 147, 60, 0, 42, 0, 112, 81, 154, 92, 230, 98, 15, 104, 106, 90, 207, 96, 80, 93, 17, 150, 123, 132, 221, 132, 48, 140, 177, 83, 156, 111, 22, 81, 147, 101, 238, 123, 52, 137, 226, 97, 227, 113, 247, 113, 103, 149, 204, 149, 188, 116, 205, 133, 85, 137, 75, 139, 96, 216, 79, 221, 113, 94, 20, 101, 62, 112, 67, 124, 127, 126, 100, 137, 149, 101, 4, 107, 70, 112, 45, 134, 55, 121, 99, 124, 116, 137, 210, 86, 97, 112, 149, 139, 157, 142, 23, 107, 132, 148, 202, 151, 60, 0, 42, 0, 200, 137, 104, 109, 230, 54, 253, 99, 6, 127, 132, 105, 36, 111, 113, 127, 130, 145, 200, 88, 210, 97, 167, 137, 92, 58, 229, 100, 62, 91, 246, 97, 68, 91, 189, 137, 43, 67, 79, 91, 44, 101, 96, 112, 213, 86, 22, 107, 38, 114, 114, 152, 156, 126, 60, 0, 42, 0, 194, 112, 229, 110, 215, 113, 130, 86, 235, 111, 1, 114, 27, 114, 211, 116, 36, 114, 109, 148, 247, 124, 60, 0, 42, 0, 125, 137, 60, 0, 42, 0, 119, 85, 60, 0, 42, 0, 198, 82, 206, 144, 52, 107, 252, 114, 222, 144, 6, 150, 143, 101, 121, 104, 98, 60, 5, 116, 207, 90, 202, 94, 96, 120, 2, 122, 18, 149, 148, 105, 111, 116, 100, 123, 70, 130, 200, 132, 11, 135, 180, 142, 130, 135, 195, 146, 175, 147, 250, 153, 60, 0, 42, 0, 23, 103, 22, 103, 250, 112, 95, 54, 113, 88, 3, 106, 226, 132, 143, 138, 190, 59, 36, 103, 60, 0, 42, 0, 204, 87, 0, 93, 106, 109, 240, 58, 168, 131, 67, 216, 152, 222, 151, 132, 172, 149, 60, 0, 42, 0, 37, 85, 210, 144, 60, 0, 42, 0, 4, 54, 60, 0, 42, 0, 26, 93, 68, 88, 241, 104, 94, 105, 144, 120, 28, 122, 24, 143, 68, 66, 144, 133, 60, 0, 42, 0, 183, 81, 60, 0, 42, 0, 48, 80, 14, 88, 35, 97, 86, 119, 28, 142, 60, 0, 42, 0, 233, 84, 60, 0, 42, 0, 21, 82, 93, 103, 152, 83, 83, 82, 248, 114, 138, 154, 161, 96, 168, 104, 169, 104, 184, 104, 129, 114, 13, 116, 187, 121, 154, 124, 114, 131, 177, 85, 195, 104, 29, 61, 130, 114, 222, 131, 66, 158, 122, 82, 19, 111, 72, 62, 93, 119, 99, 123, 33, 127, 67, 130, 10, 135, 67, 216, 215, 222, 67, 216, 29, 223, 224, 90, 119, 91, 224, 118, 141, 140, 146, 57, 6, 106, 131, 116, 206, 124, 224, 132, 235, 146, 161, 156, 206, 158, 121, 127, 62, 133, 91, 216, 153, 222, 5, 147, 241, 123, 45, 126, 39, 69, 205, 135, 53, 137, 159, 86, 39, 139, 168, 145, 208, 145, 226, 150, 132, 101, 0, 60, 200, 116, 220, 133, 140, 144, 235, 147, 236, 155, 121, 157, 231, 158, 196, 86, 33, 136, 57, 154, 75, 91, 242, 94, 33, 101, 85, 112, 153, 82, 58, 134, 87, 148, 114, 122, 108, 124, 154, 126, 106, 154, 122, 156, 29, 158, 60, 0, 42, 0, 60, 121, 78, 103, 204, 145, 218, 79, 219, 92, 226, 92, 12, 90, 242, 92, 108, 109, 6, 116, 38, 144, 2, 149, 180, 124, 207, 136, 225, 136, 74, 140, 240, 146, 164, 156, 99, 81, 167, 111, 174, 121, 201, 155, 248, 135, 180, 145, 226, 156, 144, 144, 103, 156, 26, 107, 60, 0, 42, 0, 155, 82, 134, 83, 137, 83, 116, 92, 203, 122, 15, 84, 56, 103, 61, 78, 41, 82, 177, 82, 86, 84, 92, 87, 165, 108, 139, 79, 120, 53, 166, 92, 62, 98, 165, 103, 180, 108, 160, 117, 200, 130, 182, 150, 208, 79, 234, 79, 38, 57, 14, 104, 172, 117, 5, 120, 217, 130, 114, 141, 121, 143, 102, 52, 14, 85, 51, 90, 167, 96, 23, 104, 27, 104, 150, 109, 1, 115, 213, 115, 58, 120, 62, 120, 221, 121, 24, 131, 84, 131, 230, 144, 51, 85, 111, 90, 32, 123, 146, 124, 157, 124, 55, 129, 133, 131, 137, 131, 184, 134, 206, 134, 136, 80, 211, 81, 164, 83, 217, 104, 226, 117, 224, 134, 72, 138, 222, 141, 243, 150, 175, 83, 91, 88, 68, 97, 46, 100, 167, 110, 93, 146, 232, 156, 15, 59, 166, 102, 116, 107, 110, 116, 159, 125, 154, 132, 158, 132, 39, 135, 178, 83, 119, 135, 198, 102, 119, 107, 229, 123, 9, 69, 183, 150, 103, 216, 152, 221, 23, 157, 245, 82, 250, 62, 255, 120, 184, 150, 1, 155, 46, 81, 193, 93, 255, 111, 88, 118, 152, 147, 166, 86, 226, 88, 10, 101, 222, 102, 212, 106, 223, 106, 29, 112, 4, 114, 161, 114, 197, 116, 7, 136, 151, 158, 234, 106, 15, 114, 209, 116, 170, 118, 237, 118, 203, 119, 42, 121, 43, 121, 178, 121, 246, 133, 167, 52, 55, 81, 103, 118, 48, 121, 242, 124, 35, 136, 43, 136, 197, 157, 156, 158, 199, 86, 38, 101, 94, 216, 157, 220, 251, 137, 146, 142, 98, 143, 72, 145, 16, 107, 136, 139, 96, 216, 155, 222, 99, 143, 45, 101, 229, 116, 66, 151, 102, 216, 195, 221, 113, 156, 115, 156, 75, 151, 60, 0, 42, 0, 3, 104, 170, 106, 207, 155, 60, 0, 42, 0, 233, 79, 6, 80, 60, 0, 42, 0, 167, 84, 60, 0, 42, 0, 23, 82, 163, 82, 189, 81, 189, 82, 220, 57, 244, 89, 18, 99, 12, 109, 247, 84, 210, 87, 211, 87, 32, 57, 94, 59, 53, 104, 86, 109, 200, 112, 34, 131, 254, 143, 105, 99, 105, 59, 14, 115, 31, 129, 218, 134, 194, 136, 109, 113, 89, 119, 87, 128, 148, 141, 228, 93, 178, 152, 32, 81, 164, 155, 55, 157, 248, 100, 117, 115, 163, 114, 144, 142, 27, 155, 47, 59, 35, 155, 114, 156, 60, 0, 42, 0, 223, 107, 24, 99, 238, 112, 31, 115, 60, 0, 42, 0, 169, 100, 125, 142, 60, 0, 42, 0, 189, 143, 151, 117, 74, 128, 218, 80, 229, 91, 122, 93, 128, 97, 59, 111, 139, 129, 57, 86, 253, 90, 238, 91, 154, 93, 155, 93, 121, 101, 96, 115, 45, 127, 184, 102, 206, 113, 153, 116, 171, 129, 124, 144, 115, 61, 66, 118, 105, 158, 106, 92, 235, 94, 29, 124, 90, 126, 223, 135, 66, 140, 255, 140, 88, 142, 16, 148, 212, 133, 201, 152, 206, 154, 239, 157, 60, 0, 42, 0, 254, 83, 38, 65, 140, 148, 213, 145, 29, 145, 173, 97, 252, 132, 173, 119, 226, 102, 125, 149, 18, 114, 60, 0, 42, 0, 37, 92, 38, 92, 147, 112, 153, 101, 30, 92, 214, 94, 130, 100, 183, 122, 99, 149, 60, 0, 42, 0, 14, 114, 60, 0, 42, 0, 156, 110, 152, 113, 83, 142, 60, 0, 42, 0, 24, 82, 160, 108, 68, 117, 65, 109, 79, 109, 9, 116, 89, 117, 200, 101, 113, 117, 107, 120, 215, 136, 99, 216, 231, 222, 185, 90, 103, 93, 210, 101, 197, 132, 143, 153, 157, 154, 180, 105, 96, 116, 165, 132, 91, 144, 215, 152, 137, 82, 108, 116, 36, 118, 194, 120, 79, 149, 224, 153, 96, 158, 74, 106, 162, 116, 129, 117, 96, 149, 69, 118, 201, 135, 101, 216, 233, 221, 245, 153, 160, 86, 240, 97, 15, 112, 143, 147, 166, 147, 141, 158, 240, 133, 208, 147, 192, 152, 46, 154, 197, 152, 33, 156, 185, 157, 81, 154, 60, 0, 42, 0, 213, 57, 243, 103, 1, 104, 203, 115, 122, 104, 250, 126, 13, 149, 90, 146, 249, 152, 185, 125, 174, 113, 118, 127, 246, 146, 110, 106, 60, 91, 128, 127, 60, 0, 42, 0, 109, 81, 66, 117, 207, 127, 111, 88, 199, 94, 145, 111, 223, 120, 104, 158, 36, 151, 62, 153, 225, 150, 2, 148, 194, 152, 56, 155, 218, 157, 60, 0, 42, 0, 94, 104, 60, 0, 42, 0, 65, 89, 24, 94, 28, 96, 222, 143, 159, 109, 191, 104, 178, 131, 122, 70, 35, 144, 84, 128, 183, 78, 201, 94, 147, 110, 50, 83, 249, 85, 105, 89, 105, 97, 35, 111, 145, 113, 226, 136, 157, 137, 134, 82, 51, 83, 82, 86, 254, 90, 144, 97, 228, 105, 207, 120, 107, 128, 238, 132, 162, 156, 194, 111, 211, 111, 110, 128, 138, 135, 235, 113, 58, 126, 244, 127, 111, 128, 193, 129, 149, 133, 51, 137, 48, 139, 101, 142, 140, 147, 112, 149, 227, 106, 62, 124, 10, 136, 17, 155, 101, 73, 46, 148, 49, 156, 98, 124, 104, 124, 60, 0, 42, 0, 91, 101, 15, 116, 56, 129, 227, 136, 89, 100, 137, 116, 57, 133, 26, 91, 130, 101, 236, 59, 201, 129, 59, 145, 183, 127, 93, 137, 30, 134, 60, 0, 42, 0, 195, 126, 188, 112, 75, 96, 112, 109, 147, 107, 28, 88, 3, 55, 161, 90, 104, 60, 69, 110, 254, 148, 198, 80, 93, 105, 73, 113, 83, 116, 48, 132, 75, 111, 244, 125, 176, 111, 44, 147, 174, 107, 74, 147, 200, 147, 145, 61, 50, 112, 10, 156, 29, 134, 0, 98, 59, 55, 158, 126, 60, 0, 42, 0, 104, 128, 60, 0, 42, 0, 206, 98, 60, 0, 42, 0, 184, 83, 151, 103, 187, 144, 52, 78, 167, 81, 221, 119, 73, 85, 10, 93, 203, 109, 125, 102, 51, 116, 166, 124, 243, 117, 132, 120, 150, 123, 188, 124, 153, 93, 126, 111, 92, 115, 48, 145, 163, 150, 180, 101, 189, 102, 208, 113, 152, 116, 154, 143, 116, 144, 22, 151, 181, 119, 247, 120, 232, 129, 87, 126, 247, 127, 144, 158, 84, 143, 227, 88, 54, 112, 251, 147, 222, 156, 78, 154, 87, 156, 159, 158, 60, 0, 42, 0, 45, 52, 6, 58, 251, 131, 131, 78, 219, 81, 220, 81, 155, 100, 233, 94, 234, 94, 205, 97, 212, 97, 159, 111, 129, 106, 169, 106, 91, 118, 93, 118, 60, 0, 42, 0, 29, 84, 97, 96, 139, 96, 65, 141, 27, 113, 195, 140, 239, 80, 58, 133, 73, 106, 166, 129, 181, 149, 16, 117, 132, 117, 120, 142, 250, 133, 143, 142, 153, 142, 170, 142, 101, 143, 60, 0, 42, 0, 111, 130, 205, 79, 201, 81, 129, 104, 188, 109, 11, 105, 140, 143, 174, 124, 177, 124, 154, 88, 161, 125, 9, 142, 17, 106, 44, 143, 231, 124, 60, 0, 42, 0, 33, 78, 36, 78, 179, 52, 105, 81, 33, 85, 98, 85, 154, 99, 60, 129, 201, 125, 61, 135, 242, 136, 73, 155, 78, 155, 60, 0, 42, 0, 174, 78, 244, 84, 162, 96, 5, 140, 134, 143, 168, 85, 126, 102, 120, 110, 207, 145, 12, 143, 210, 138, 27, 143, 68, 147, 60, 0, 42, 0, 119, 113, 23, 124, 60, 0, 42, 0, 7, 48, 54, 79, 34, 82, 117, 112, 249, 86, 125, 87, 12, 89, 200, 89, 186, 92, 126, 95, 224, 108, 209, 114, 36, 102, 14, 103, 195, 103, 178, 115, 211, 130, 204, 81, 244, 116, 138, 118, 49, 120, 226, 121, 219, 122, 195, 148, 48, 158, 72, 90, 149, 99, 194, 104, 233, 109, 12, 116, 45, 123, 55, 125, 235, 126, 69, 67, 154, 127, 206, 127, 70, 128, 50, 130, 201, 134, 81, 136, 117, 150, 241, 131, 69, 138, 201, 141, 232, 142, 126, 121, 52, 146, 157, 149, 246, 150, 132, 159, 10, 66, 190, 125, 236, 136, 6, 133, 10, 151, 214, 153, 170, 111, 2, 147, 127, 155, 174, 156, 18, 157, 119, 158, 33, 59, 239, 113, 65, 66, 22, 69, 118, 133, 27, 151, 29, 151, 98, 159, 234, 155, 65, 91, 67, 145, 97, 159, 250, 106, 38, 134, 189, 145, 72, 151, 76, 216, 31, 223, 30, 107, 39, 114, 82, 216, 233, 222, 162, 158, 133, 69, 151, 159, 60, 0, 42, 0, 173, 92, 62, 150, 138, 136, 134, 152, 24, 152, 186, 93, 60, 0, 42, 0, 228, 78, 230, 83, 100, 84, 169, 112, 60, 0, 42, 0, 228, 132, 23, 151, 46, 112, 60, 0, 42, 0, 92, 86, 184, 100, 60, 0, 42, 0, 98, 83, 144, 94, 134, 87, 248, 108, 137, 112, 166, 130, 84, 59, 12, 104, 234, 128, 115, 143, 44, 158, 136, 115, 59, 130, 133, 152, 136, 156, 114, 155, 231, 118, 167, 86, 218, 88, 32, 56, 236, 94, 14, 101, 218, 106, 24, 112, 121, 115, 183, 116, 229, 102, 232, 106, 16, 114, 208, 116, 218, 129, 6, 134, 209, 119, 90, 124, 145, 126, 79, 127, 107, 130, 38, 136, 100, 143, 106, 148, 113, 152, 215, 154, 120, 156, 21, 158, 248, 158, 60, 0, 42, 0, 100, 83, 79, 134, 179, 99, 117, 158, 117, 120, 129, 156, 92, 134, 119, 88, 247, 110, 16, 106, 155, 111, 254, 132, 111, 155, 196, 100, 121, 106, 224, 120, 101, 149, 149, 86, 252, 100, 2, 112, 211, 106, 12, 108, 99, 130, 192, 147, 106, 130, 42, 148, 101, 148, 60, 0, 42, 0, 37, 87, 42, 117, 147, 79, 116, 87, 84, 95, 85, 95, 70, 150, 205, 92, 206, 82, 66, 141, 130, 143, 61, 90, 213, 109, 229, 109, 12, 110, 73, 120, 120, 150, 127, 158, 144, 58, 2, 105, 45, 116, 132, 121, 2, 68, 201, 131, 47, 144, 199, 80, 121, 82, 224, 82, 221, 118, 105, 119, 140, 120, 127, 121, 17, 122, 194, 140, 239, 141, 118, 88, 216, 94, 93, 100, 9, 111, 147, 123, 182, 124, 68, 216, 20, 220, 46, 98, 26, 106, 157, 113, 148, 129, 13, 133, 238, 137, 162, 141, 27, 142, 152, 143, 129, 145, 221, 59, 94, 111, 75, 122, 4, 147, 50, 147, 52, 147, 144, 116, 15, 124, 87, 133, 176, 135, 87, 142, 70, 143, 4, 154, 102, 216, 241, 220, 109, 158, 44, 124, 54, 124, 213, 147, 229, 155, 102, 157, 113, 157, 147, 158, 244, 147, 50, 151, 60, 154, 89, 124, 250, 157, 66, 134, 60, 0, 42, 0, 166, 103, 46, 130, 41, 146, 7, 108, 60, 0, 42, 0, 87, 127, 112, 85, 49, 152, 201, 86, 60, 0, 42, 0, 86, 127, 33, 115, 54, 129, 36, 105, 29, 132, 153, 137, 59, 144, 97, 129, 35, 149, 169, 123, 161, 154, 89, 149, 186, 135, 133, 127, 182, 137, 205, 147, 56, 81, 188, 137, 62, 154, 70, 216, 34, 221, 30, 101, 128, 115, 15, 107, 90, 216, 199, 220, 63, 134, 143, 144, 88, 154, 1, 158, 110, 124, 124, 148, 96, 153, 60, 0, 42, 0, 70, 82, 46, 80, 207, 132, 248, 136, 182, 142, 48, 118, 97, 58, 1, 60, 3, 136, 221, 129, 234, 102, 115, 118, 60, 0, 42, 0, 250, 108, 200, 92, 27, 109, 220, 126, 134, 154, 28, 109, 222, 115, 102, 131, 102, 120, 63, 123, 97, 125, 224, 85, 61, 132, 94, 100, 47, 111, 150, 114, 123, 146, 210, 150, 241, 153, 165, 155, 60, 157, 69, 157, 252, 111, 93, 216, 204, 223, 153, 126, 60, 0, 42, 0, 8, 90, 106, 91, 230, 92, 27, 99, 62, 104, 62, 158, 20, 129, 230, 110, 174, 146, 73, 157, 29, 87, 113, 89, 76, 91, 127, 91, 210, 93, 35, 101, 235, 102, 18, 107, 83, 112, 137, 127, 224, 129, 30, 87, 100, 112, 74, 134, 126, 148, 116, 118, 117, 118, 30, 158, 60, 0, 42, 0, 117, 83, 60, 0, 42, 0, 113, 78, 224, 145, 130, 78, 60, 0, 42, 0, 161, 98, 132, 99, 60, 0, 42, 0, 209, 78, 38, 79, 245, 86, 166, 108, 182, 126, 150, 79, 110, 143, 43, 80, 7, 87, 104, 90, 24, 93, 25, 93, 192, 96, 234, 109, 111, 150, 198, 104, 205, 61, 64, 129, 213, 131, 1, 66, 184, 125, 38, 135, 26, 142, 42, 143, 0, 147, 233, 155, 60, 0, 42, 0, 232, 87, 150, 120, 16, 122, 35, 128, 60, 0, 42, 0, 186, 139, 163, 110, 214, 138, 60, 0, 42, 0, 238, 120, 60, 0, 42, 0, 153, 159, 120, 92, 153, 84, 247, 108, 61, 102, 10, 104, 209, 115, 231, 128, 15, 131, 44, 119, 59, 120, 220, 122, 60, 123, 75, 128, 112, 110, 134, 150, 221, 110, 67, 216, 144, 223, 144, 93, 11, 111, 99, 56, 237, 123, 111, 133, 141, 159, 67, 118, 191, 122, 168, 86, 195, 93, 196, 93, 39, 112, 60, 124, 231, 147, 232, 102, 39, 103, 13, 60, 243, 106, 22, 114, 207, 116, 34, 134, 51, 151, 211, 119, 49, 121, 50, 121, 146, 159, 96, 124, 126, 128, 42, 136, 44, 136, 113, 137, 69, 140, 152, 142, 104, 148, 71, 151, 190, 75, 97, 154, 23, 158, 60, 0, 42, 0, 132, 87, 133, 87, 226, 98, 71, 150, 51, 61, 226, 123, 49, 81, 223, 88, 224, 88, 15, 101, 180, 150, 201, 122, 147, 159, 60, 0, 42, 0, 226, 84, 53, 99, 135, 104, 191, 95, 26, 141, 60, 0, 42, 0, 222, 105, 60, 0, 42, 0, 116, 154, 254, 149, 216, 144, 136, 105, 191, 153, 0, 108, 162, 129, 173, 149, 216, 133, 220, 157, 98, 154, 60, 0, 42, 0, 21, 84, 66, 84, 163, 79, 182, 79, 20, 99, 206, 54, 75, 99, 91, 99, 197, 101, 90, 59, 160, 104, 99, 121, 6, 122, 221, 148, 97, 92, 21, 127, 125, 125, 98, 92, 130, 129, 101, 92, 27, 137, 193, 146, 144, 129, 34, 81, 94, 122, 55, 126, 56, 137, 109, 122, 60, 0, 42, 0, 253, 91, 143, 87, 139, 95, 81, 134, 135, 115, 255, 126, 67, 216, 200, 221, 66, 93, 47, 108, 228, 110, 78, 132, 160, 125, 209, 125, 110, 97, 187, 123, 159, 129, 244, 82, 66, 126, 254, 111, 214, 106, 8, 114, 98, 148, 60, 0, 42, 0, 18, 113, 60, 0, 42, 0, 103, 117, 10, 149, 36, 122, 0, 53, 25, 87, 60, 58, 221, 146, 226, 146, 253, 100, 60, 0, 39, 0, 208, 253, 39, 0, 13, 49, 60, 0, 42, 0, 238, 101, 119, 84, 14, 86, 32, 86, 60, 0, 42, 0, 134, 148, 28, 92, 118, 86, 55, 147, 60, 0, 42, 0, 21, 92, 141, 115, 60, 0, 42, 0, 44, 92, 64, 155, 60, 0, 42, 0, 8, 98, 225, 78, 42, 87, 181, 114, 165, 126, 66, 216, 235, 223, 19, 98, 144, 128, 107, 114, 153, 117, 175, 84, 113, 114, 229, 84, 243, 128, 61, 158, 114, 82, 1, 100, 188, 136, 67, 216, 157, 222, 210, 110, 40, 98, 76, 107, 103, 216, 183, 223, 16, 157, 26, 157, 241, 100, 12, 139, 63, 157, 182, 147, 60, 0, 42, 0, 68, 84, 55, 56, 110, 79, 12, 83, 12, 99, 1, 150, 105, 151, 75, 101, 88, 59, 60, 104, 22, 131, 50, 155, 5, 97, 245, 129, 210, 134, 221, 85, 101, 88, 198, 110, 91, 132, 211, 136, 225, 137, 148, 150, 63, 100, 197, 105, 136, 129, 163, 149, 164, 149, 19, 66, 73, 149, 136, 151, 208, 151, 102, 115, 253, 138, 53, 143, 188, 154, 175, 155, 218, 151, 183, 151, 20, 154, 1, 70, 85, 143, 60, 0, 42, 0, 255, 84, 56, 130, 60, 0, 42, 0, 42, 78, 4, 84, 124, 134, 11, 80, 76, 120, 236, 148, 240, 85, 135, 123, 60, 0, 42, 0, 65, 95, 202, 106, 60, 0, 42, 0, 133, 79, 229, 139, 147, 87, 223, 89, 208, 92, 194, 144, 84, 150, 80, 102, 68, 131, 69, 141, 97, 117, 116, 121, 111, 125, 114, 138, 101, 140, 197, 140, 60, 0, 42, 0, 203, 95, 57, 101, 96, 125, 60, 0, 42, 0, 16, 78, 98, 78, 3, 83, 4, 83, 35, 150, 90, 103, 153, 148, 214, 118, 97, 100, 137, 110, 35, 146, 36, 98, 130, 105, 98, 132, 204, 140, 145, 150, 17, 111, 203, 132, 98, 53, 233, 105, 234, 105, 194, 116, 60, 0, 42, 0, 217, 126, 102, 125, 60, 0, 42, 0, 139, 118, 148, 127, 153, 127, 216, 154, 144, 118, 90, 216, 14, 221, 217, 154, 239, 129, 220, 110, 212, 105, 126, 119, 143, 129, 249, 105, 112, 106, 217, 123, 213, 124, 59, 153, 220, 106, 223, 151, 206, 157, 27, 159, 241, 157, 60, 0, 42, 0, 48, 89, 114, 103, 210, 131, 30, 100, 31, 127, 160, 102, 192, 105, 193, 105, 70, 61, 62, 122, 63, 122, 80, 149, 30, 126, 186, 106, 193, 133, 243, 133, 60, 0, 42, 0, 63, 84, 74, 84, 194, 82, 221, 83, 240, 139, 220, 144, 110, 121, 112, 121, 6, 149, 112, 113, 118, 123, 165, 138, 158, 121, 239, 146, 60, 0, 42, 0, 254, 82, 93, 79, 159, 108, 169, 148, 167, 136, 17, 127, 14, 146, 157, 110, 100, 146, 241, 125, 221, 123, 32, 137, 178, 151, 221, 151, 60, 0, 42, 0, 163, 92, 215, 114, 182, 130, 184, 103, 189, 115, 7, 128, 223, 130, 49, 123, 8, 128, 9, 128, 188, 134, 127, 140, 60, 0, 42, 0, 120, 87, 132, 103, 223, 139, 45, 141, 162, 87, 228, 89, 147, 81, 41, 131, 31, 89, 32, 89, 61, 138, 190, 90, 64, 95, 6, 100, 108, 138, 202, 150, 203, 105, 121, 113, 207, 137, 88, 144, 128, 100, 175, 137, 252, 140, 60, 0, 42, 0, 24, 117, 211, 95, 60, 101, 70, 103, 165, 61, 149, 115, 157, 128, 137, 130, 192, 143, 105, 87, 212, 108, 248, 119, 121, 78, 209, 103, 255, 122, 247, 130, 179, 117, 80, 145, 126, 78, 147, 124, 129, 78, 242, 81, 50, 92, 52, 92, 120, 123, 39, 111, 241, 156, 54, 92, 55, 92, 80, 155, 60, 0, 42, 0, 224, 78, 94, 98, 175, 118, 198, 121, 102, 136, 118, 141, 127, 104, 52, 123, 98, 101, 8, 122, 31, 97, 149, 141, 137, 111, 192, 100, 68, 106, 51, 124, 20, 156, 225, 156, 100, 156, 60, 0, 42, 0, 114, 94, 240, 101, 117, 108, 240, 118, 192, 126, 29, 80, 206, 81, 230, 109, 58, 125, 76, 138, 121, 94, 173, 154, 166, 105, 138, 106, 17, 141, 99, 141, 27, 141, 104, 112, 60, 0, 42, 0, 57, 104, 223, 141, 60, 0, 42, 0, 207, 84, 60, 0, 42, 0, 110, 130, 60, 0, 42, 0, 152, 78, 153, 78, 27, 131, 239, 99, 60, 0, 42, 0, 136, 81, 83, 127, 174, 81, 26, 82, 96, 103, 178, 126, 155, 128, 161, 92, 78, 59, 104, 114, 152, 117, 252, 119, 56, 127, 162, 148, 91, 82, 97, 127, 8, 88, 134, 99, 237, 145, 225, 104, 133, 114, 61, 88, 177, 125, 65, 127, 252, 146, 160, 147, 60, 0, 42, 0, 151, 92, 88, 63, 23, 93, 47, 110, 60, 0, 42, 0, 53, 113, 123, 123, 211, 105, 5, 98, 6, 98, 60, 0, 42, 0, 47, 82, 154, 94, 74, 117, 109, 109, 116, 62, 21, 128, 4, 100, 63, 113, 90, 125, 238, 131, 83, 141, 82, 158, 234, 125, 6, 126, 174, 127, 225, 140, 185, 127, 138, 157, 60, 0, 42, 0, 253, 84, 194, 87, 250, 92, 45, 99, 224, 126, 63, 128, 224, 144, 76, 216, 179, 221, 151, 104, 132, 131, 152, 63, 134, 125, 4, 67, 160, 156, 190, 154, 193, 155, 60, 0, 42, 0, 244, 102, 41, 88, 133, 102, 77, 76, 109, 76, 60, 0, 42, 0, 182, 99, 41, 105, 60, 0, 42, 0, 48, 79, 113, 84, 149, 84, 209, 89, 100, 91, 189, 108, 210, 108, 231, 103, 253, 130, 113, 143, 2, 85, 91, 127, 42, 158, 31, 123, 196, 134, 199, 131, 240, 131, 218, 137, 241, 142, 242, 142, 156, 143, 100, 145, 50, 146, 141, 123, 155, 123, 244, 90, 109, 106, 59, 68, 149, 155, 35, 157, 60, 0, 42, 0, 187, 157, 60, 0, 42, 0, 3, 89, 228, 83, 98, 98, 105, 108, 194, 139, 55, 140, 161, 128, 111, 114, 227, 46, 3, 85, 38, 64, 95, 127, 150, 127, 180, 148, 168, 154, 82, 85, 200, 109, 53, 129, 202, 134, 204, 134, 65, 138, 68, 158, 51, 92, 50, 97, 194, 107, 55, 146, 137, 153, 19, 159, 20, 159, 15, 86, 150, 105, 190, 105, 183, 118, 199, 132, 64, 122, 88, 158, 211, 124, 14, 126, 242, 111, 188, 118, 204, 129, 163, 133, 66, 143, 20, 112, 236, 118, 189, 119, 54, 153, 156, 77, 49, 136, 60, 0, 42, 0, 250, 86, 69, 101, 197, 81, 126, 152, 12, 88, 19, 93, 46, 93, 143, 104, 127, 114, 221, 104, 199, 150, 252, 117, 123, 121, 18, 122, 34, 149, 241, 80, 46, 147, 180, 156, 221, 155, 103, 152, 60, 0, 42, 0, 224, 92, 39, 144, 166, 80, 208, 123, 60, 0, 42, 0, 220, 116, 46, 82, 205, 128, 29, 104, 67, 216, 111, 221, 57, 158, 82, 128, 183, 68, 68, 107, 113, 113, 143, 141, 128, 82, 250, 125, 189, 146, 179, 152, 59, 142, 48, 157, 39, 154, 60, 0, 42, 0, 231, 83, 142, 81, 80, 82, 110, 82, 225, 91, 17, 66, 60, 0, 42, 0, 102, 83, 108, 87, 214, 139, 2, 99, 105, 85, 155, 99, 99, 127, 83, 125, 107, 127, 127, 138, 2, 137, 60, 0, 42, 0, 170, 152, 60, 0, 42, 0, 89, 84, 188, 84, 218, 87, 111, 85, 30, 93, 237, 144, 29, 88, 27, 146, 5, 149, 19, 86, 142, 88, 17, 118, 72, 135, 73, 95, 251, 53, 196, 111, 200, 135, 75, 147, 77, 95, 60, 0, 42, 0, 239, 86, 246, 86, 251, 86, 253, 86, 0, 87, 11, 87, 60, 94, 88, 129, 87, 94, 86, 97, 13, 111, 93, 128, 149, 129, 46, 133, 98, 134, 152, 153, 95, 216, 18, 220, 144, 65, 60, 0, 42, 0, 156, 103, 200, 96, 201, 109, 19, 115, 131, 153, 1, 105, 211, 131, 191, 124, 182, 125, 62, 135, 249, 136, 232, 105, 32, 143, 1, 147, 28, 153, 57, 148, 60, 0, 42, 0, 199, 143, 78, 144, 60, 0, 42, 0, 86, 78, 180, 99, 81, 100, 60, 0, 42, 0, 208, 98, 180, 103, 250, 103, 92, 216, 133, 222, 137, 123, 60, 0, 42, 0, 44, 89, 207, 83, 42, 96, 212, 57, 96, 96, 60, 0, 42, 0, 82, 95, 45, 87, 171, 89, 159, 159, 196, 137, 136, 118, 189, 144, 250, 149, 48, 94, 234, 115, 255, 128, 37, 131, 128, 78, 69, 120, 69, 65, 144, 122, 143, 137, 128, 80, 175, 90, 29, 105, 191, 136, 198, 94, 204, 144, 226, 90, 107, 100, 112, 116, 168, 149, 145, 156, 0, 91, 251, 105, 252, 105, 157, 116, 173, 129, 157, 135, 156, 159, 173, 155, 194, 93, 120, 107, 54, 155, 204, 116, 41, 154, 57, 155, 247, 106, 60, 0, 42, 0, 132, 91, 63, 108, 57, 103, 104, 143, 139, 94, 121, 79, 38, 83, 225, 139, 67, 216, 83, 220, 157, 87, 253, 89, 81, 96, 49, 101, 120, 118, 204, 142, 82, 150, 170, 94, 60, 155, 45, 83, 106, 121, 119, 102, 64, 110, 111, 67, 235, 134, 51, 53, 228, 137, 109, 138, 172, 83, 51, 58, 37, 73, 225, 135, 161, 119, 11, 124, 60, 0, 42, 0, 48, 101, 61, 82, 63, 82, 11, 102, 148, 112, 220, 103, 53, 141, 136, 58, 66, 104, 103, 104, 34, 105, 36, 115, 64, 123, 180, 140, 234, 141, 49, 83, 213, 132, 138, 82, 140, 82, 161, 93, 140, 100, 246, 105, 16, 66, 52, 68, 156, 106, 182, 119, 2, 124, 195, 106, 80, 118, 172, 121, 88, 137, 220, 156, 188, 151, 86, 156, 101, 156, 60, 0, 42, 0, 115, 81, 194, 137, 152, 91, 160, 81, 140, 137, 12, 80, 67, 216, 40, 221, 250, 104, 164, 122, 132, 132, 162, 149, 29, 118, 79, 118, 179, 137, 215, 149, 207, 156, 220, 149, 37, 156, 94, 156, 192, 137, 60, 0, 42, 0, 158, 131, 134, 153, 47, 116, 239, 117, 102, 123, 161, 123, 40, 143, 24, 130, 39, 147, 40, 153, 228, 156, 60, 0, 42, 0, 204, 107, 49, 78, 47, 141, 244, 108, 186, 96, 239, 96, 188, 99, 171, 109, 171, 140, 185, 96, 38, 61, 161, 59, 124, 121, 99, 97, 92, 100, 69, 111, 12, 106, 102, 144, 229, 118, 70, 127, 91, 70, 218, 150, 198, 147, 76, 112, 31, 114, 216, 116, 115, 158, 212, 119, 54, 121, 80, 127, 117, 148, 121, 156, 27, 158, 60, 0, 42, 0, 40, 78, 64, 216, 204, 220, 110, 136, 195, 96, 242, 126, 158, 136, 172, 136, 138, 143, 218, 110, 254, 110, 196, 125, 216, 132, 217, 120, 9, 133, 37, 143, 167, 156, 85, 216, 102, 221, 140, 155, 192, 155, 60, 0, 42, 0, 205, 104, 84, 119, 101, 54, 116, 119, 173, 116, 52, 139, 60, 0, 42, 0, 73, 81, 110, 112, 138, 79, 151, 112, 154, 112, 155, 112, 163, 84, 153, 87, 239, 89, 56, 109, 68, 104, 225, 112, 241, 128, 42, 131, 4, 143, 217, 80, 167, 146, 198, 158, 102, 216, 201, 221, 60, 0, 42, 0, 127, 94, 131, 94, 183, 114, 227, 94, 233, 129, 119, 115, 104, 216, 245, 221, 60, 0, 42, 0, 199, 79, 214, 115, 27, 144, 230, 129, 151, 100, 101, 216, 72, 220, 60, 0, 42, 0, 31, 107, 60, 0, 42, 0, 229, 93, 19, 95, 108, 81, 183, 83, 159, 82, 59, 101, 91, 103, 155, 79, 156, 115, 252, 124, 177, 128, 171, 91, 174, 91, 109, 96, 163, 134, 172, 142, 154, 159, 17, 83, 104, 88, 74, 94, 41, 97, 229, 137, 149, 113, 179, 142, 189, 120, 248, 154, 205, 76, 245, 137, 143, 159, 148, 159, 60, 0, 42, 0, 254, 94, 233, 93, 94, 108, 241, 98, 236, 57, 242, 98, 95, 59, 49, 104, 217, 115, 38, 63, 1, 143, 219, 146, 143, 151, 75, 76, 60, 0, 42, 0, 113, 81, 33, 141, 190, 127, 29, 85, 162, 140, 187, 131, 54, 53, 60, 0, 42, 0, 80, 97, 60, 0, 39, 0, 208, 253, 39, 0, 14, 49, 60, 0, 42, 0, 148, 84, 150, 84, 128, 85, 73, 136, 214, 100, 60, 0, 42, 0, 97, 83, 103, 79, 233, 128, 114, 146, 60, 0, 42, 0, 176, 87, 195, 136, 60, 0, 42, 0, 60, 83, 47, 62, 72, 216, 8, 222, 239, 103, 65, 114, 194, 115, 209, 121, 226, 128, 219, 130, 114, 143, 180, 117, 34, 120, 183, 141, 245, 104, 251, 142, 143, 152, 209, 85, 21, 100, 144, 114, 30, 122, 160, 122, 42, 132, 51, 146, 188, 105, 38, 68, 151, 152, 22, 106, 140, 119, 213, 120, 0, 67, 76, 135, 18, 147, 150, 133, 152, 145, 70, 152, 193, 154, 26, 121, 60, 0, 42, 0, 243, 88, 226, 99, 188, 107, 215, 127, 60, 0, 42, 0, 239, 83, 119, 87, 185, 55, 162, 92, 188, 58, 163, 112, 7, 110, 81, 93, 100, 101, 52, 110, 177, 93, 13, 121, 60, 0, 42, 0, 75, 81, 59, 82, 75, 82, 192, 82, 202, 82, 162, 91, 106, 96, 67, 216, 237, 220, 20, 90, 5, 92, 254, 139, 67, 216, 178, 221, 1, 88, 42, 108, 146, 154, 2, 127, 25, 97, 152, 110, 30, 149, 166, 120, 217, 125, 80, 130, 178, 138, 10, 121, 13, 154, 104, 216, 237, 220, 60, 0, 42, 0, 89, 93, 60, 0, 42, 0, 0, 95, 82, 89, 233, 99, 14, 149, 139, 149, 36, 73, 38, 148, 60, 0, 42, 0, 239, 81, 64, 82, 178, 87, 122, 96, 255, 149, 224, 148, 241, 81, 116, 82, 104, 97, 79, 88, 102, 93, 55, 97, 119, 105, 136, 132, 6, 143, 5, 86, 159, 102, 52, 149, 87, 64, 71, 147, 167, 147, 211, 149, 189, 152, 60, 0, 42, 0, 254, 95, 140, 112, 143, 112, 44, 107, 215, 112, 211, 82, 18, 97, 62, 97, 142, 147, 60, 0, 42, 0, 59, 92, 219, 154, 60, 0, 42, 0, 2, 78, 55, 101, 3, 128, 247, 98, 24, 109, 50, 104, 228, 112, 1, 122, 147, 156, 250, 113, 60, 0, 42, 0, 208, 148, 146, 114, 172, 146, 96, 151, 179, 155, 204, 155, 60, 0, 42, 0, 160, 98, 90, 216, 5, 223, 164, 130, 13, 119, 126, 82, 68, 95, 115, 100, 152, 119, 113, 64, 60, 0, 42, 0, 227, 83, 81, 216, 24, 222, 182, 82, 60, 0, 42, 0, 233, 83, 99, 98, 66, 101, 166, 81, 188, 91, 199, 91, 47, 60, 230, 145, 155, 122, 88, 123, 241, 110, 137, 119, 50, 133, 59, 133, 6, 124, 199, 157, 60, 0, 42, 0, 10, 82, 30, 104, 216, 82, 155, 159, 42, 88, 65, 93, 33, 98, 149, 159, 60, 0, 42, 0, 154, 81, 78, 87, 131, 79, 13, 120, 88, 80, 243, 87, 194, 96, 176, 131, 63, 107, 106, 88, 65, 107, 219, 105, 33, 143, 187, 106, 81, 152, 247, 122, 87, 143, 60, 0, 42, 0, 11, 119, 78, 136, 1, 93, 136, 88, 225, 120, 176, 119, 222, 149, 217, 119, 60, 0, 42, 0, 142, 128, 175, 128, 187, 128, 166, 87, 115, 96, 67, 85, 100, 140, 136, 159, 190, 88, 57, 147, 199, 97, 102, 159, 60, 0, 42, 0, 175, 99, 201, 136, 3, 137, 60, 0, 42, 0, 252, 95, 246, 149, 10, 120, 135, 124, 183, 94, 69, 216, 19, 220, 221, 90, 123, 93, 119, 97, 46, 111, 250, 105, 69, 122, 178, 65, 224, 124, 191, 142, 238, 147, 71, 156, 60, 0, 42, 0, 91, 98, 67, 100, 60, 0, 42, 0, 73, 72, 60, 0, 42, 0, 162, 78, 9, 79, 31, 83, 229, 86, 151, 98, 186, 114, 159, 144, 149, 112, 170, 148, 39, 146, 140, 149, 60, 0, 42, 0, 165, 82, 45, 84, 81, 87, 148, 89, 44, 150, 51, 99, 65, 120, 124, 114, 92, 120, 255, 148, 123, 120, 124, 100, 153, 138, 181, 146, 216, 64, 94, 147, 215, 147, 60, 0, 42, 0, 93, 98, 51, 82, 251, 119, 175, 103, 208, 128, 192, 144, 237, 84, 77, 104, 0, 88, 43, 93, 196, 53, 16, 87, 205, 141, 159, 122, 183, 154, 90, 216, 70, 220, 172, 155, 60, 0, 42, 0, 220, 114, 230, 130, 60, 0, 42, 0, 56, 56, 147, 94, 200, 79, 212, 126, 171, 94, 217, 121, 182, 141, 5, 113, 190, 85, 93, 125, 180, 136, 228, 136, 20, 118, 119, 145, 228, 94, 50, 137, 179, 86, 60, 0, 42, 0, 56, 89, 241, 89, 135, 138, 60, 0, 42, 0, 137, 79, 181, 84, 174, 87, 153, 146, 60, 0, 42, 0, 4, 52, 14, 99, 239, 128, 232, 141, 187, 154, 60, 0, 42, 0, 63, 130, 60, 0, 42, 0, 105, 98, 225, 98, 236, 98, 4, 99, 112, 104, 72, 123, 222, 134, 20, 150, 63, 132, 64, 132, 211, 94, 34, 152, 250, 154, 246, 111, 202, 149, 159, 151, 214, 97, 244, 100, 41, 151, 185, 151, 32, 155, 60, 0, 42, 0, 213, 151, 60, 0, 42, 0, 223, 57, 175, 132, 211, 100, 60, 0, 42, 0, 220, 93, 247, 81, 87, 87, 235, 95, 169, 79, 217, 84, 239, 114, 208, 144, 13, 129, 74, 88, 110, 60, 119, 123, 153, 156, 8, 81, 164, 88, 114, 86, 229, 94, 106, 115, 54, 145, 190, 129, 221, 101, 233, 124, 96, 156, 60, 0, 42, 0, 38, 87, 60, 0, 42, 0, 143, 78, 50, 82, 191, 92, 157, 96, 212, 118, 165, 122, 103, 128, 186, 122, 103, 134, 218, 149, 93, 152, 203, 93, 44, 134, 60, 0, 42, 0, 78, 89, 70, 102, 4, 152, 151, 153, 185, 85, 198, 99, 175, 69, 53, 144, 8, 145, 153, 154, 35, 98, 140, 102, 79, 105, 81, 105, 117, 132, 151, 150, 125, 119, 65, 155, 112, 135, 6, 73, 47, 152, 122, 54, 104, 147, 119, 147, 198, 106, 200, 133, 36, 154, 20, 89, 55, 134, 65, 134, 217, 93, 104, 216, 180, 221, 170, 114, 168, 142, 60, 0, 42, 0, 67, 113, 236, 141, 13, 152, 94, 142, 60, 0, 42, 0, 47, 92, 46, 83, 51, 107, 159, 85, 38, 97, 131, 110, 67, 129, 136, 153, 191, 90, 39, 97, 137, 132, 134, 119, 51, 86, 7, 91, 146, 97, 112, 111, 209, 123, 105, 128, 70, 70, 59, 106, 109, 128, 98, 133, 9, 139, 35, 124, 117, 128, 61, 153, 68, 124, 0, 148, 75, 153, 78, 148, 14, 76, 60, 0, 42, 0, 189, 91, 219, 91, 236, 91, 215, 129, 203, 154, 214, 154, 60, 0, 42, 0, 53, 107, 62, 107, 64, 107, 190, 122, 64, 69, 60, 0, 42, 0, 189, 122, 103, 148, 60, 0, 42, 0, 100, 87, 6, 102, 3, 88, 107, 90, 16, 93, 17, 93, 92, 102, 17, 115, 28, 113, 40, 116, 206, 131, 200, 136, 224, 154, 135, 140, 31, 149, 225, 154, 77, 158, 43, 135, 233, 136, 232, 154, 123, 116, 12, 137, 140, 145, 21, 147, 178, 156, 9, 154, 228, 155, 126, 157, 164, 157, 60, 0, 42, 0, 131, 96, 70, 99, 3, 150, 248, 88, 177, 104, 113, 120, 117, 121, 7, 122, 252, 88, 27, 122, 145, 125, 205, 136, 171, 149, 184, 149, 107, 159, 60, 0, 42, 0, 240, 86, 131, 109, 251, 58, 79, 119, 60, 0, 42, 0, 18, 88, 33, 92, 73, 111, 180, 113, 60, 0, 42, 0, 33, 83, 76, 52, 187, 82, 211, 139, 41, 83, 208, 84, 71, 96, 45, 109, 188, 144, 70, 104, 68, 120, 80, 123, 176, 68, 134, 138, 237, 142, 60, 0, 42, 0, 249, 95, 130, 98, 194, 114, 243, 139, 214, 142, 145, 138, 95, 157, 60, 0, 42, 0, 60, 89, 35, 81, 237, 97, 60, 0, 42, 0, 93, 83, 57, 87, 169, 126, 157, 144, 181, 81, 247, 101, 178, 92, 193, 108, 255, 119, 63, 102, 54, 141, 22, 119, 54, 119, 86, 125, 186, 140, 230, 142, 113, 146, 217, 88, 128, 142, 58, 145, 203, 158, 236, 97, 224, 102, 12, 114, 204, 119, 38, 121, 108, 122, 138, 126, 91, 148, 60, 0, 42, 0, 63, 120, 75, 125, 122, 123, 60, 0, 42, 0, 122, 122, 154, 54, 37, 80, 234, 87, 6, 93, 190, 96, 179, 109, 127, 120, 156, 123, 19, 147, 124, 157, 60, 0, 42, 0, 84, 91, 76, 216, 250, 223, 80, 96, 60, 0, 42, 0, 167, 99, 88, 216, 72, 220, 154, 151, 60, 0, 42, 0, 187, 142, 60, 0, 39, 0, 208, 253, 39, 0, 15, 49, 60, 0, 42, 0, 119, 86, 60, 0, 42, 0, 200, 84, 234, 148, 60, 0, 42, 0, 228, 134, 60, 0, 42, 0, 100, 89, 60, 0, 42, 0, 55, 78, 60, 0, 42, 0, 195, 139, 178, 98, 49, 107, 157, 85, 54, 138, 236, 85, 26, 136, 60, 0, 42, 0, 190, 121, 8, 84, 85, 79, 190, 82, 146, 83, 138, 84, 140, 84, 192, 89, 179, 108, 198, 92, 131, 58, 247, 102, 199, 103, 226, 114, 199, 118, 122, 124, 7, 125, 195, 144, 2, 150, 120, 153, 236, 84, 137, 58, 70, 101, 56, 104, 201, 118, 205, 118, 93, 85, 113, 59, 184, 109, 210, 118, 244, 121, 119, 131, 181, 134, 129, 159, 210, 96, 46, 110, 207, 131, 2, 132, 56, 138, 140, 152, 65, 105, 252, 107, 101, 138, 136, 140, 166, 71, 5, 143, 76, 146, 22, 150, 132, 156, 97, 53, 134, 113, 86, 158, 167, 158, 149, 111, 229, 63, 28, 152, 213, 123, 238, 127, 155, 135, 122, 155, 9, 121, 212, 149, 168, 151, 85, 159, 136, 137, 161, 157, 172, 118, 73, 148, 162, 159, 60, 0, 42, 0, 107, 79, 142, 87, 58, 141, 3, 113, 148, 136, 192, 140, 195, 85, 66, 113, 139, 120, 135, 113, 107, 141, 81, 61, 16, 137, 100, 158, 30, 59, 82, 122, 239, 127, 209, 88, 75, 118, 30, 139, 0, 114, 174, 157, 180, 157, 78, 151, 22, 158, 79, 151, 60, 0, 42, 0, 173, 124, 77, 151, 60, 0, 42, 0, 141, 84, 179, 84, 232, 85, 60, 0, 42, 0, 216, 143, 105, 91, 38, 152, 184, 154, 132, 144, 60, 0, 42, 0, 119, 109, 242, 128, 248, 112, 124, 145, 162, 145, 60, 0, 42, 0, 165, 78, 142, 89, 225, 57, 135, 154, 179, 91, 38, 108, 208, 85, 0, 153, 237, 153, 90, 216, 38, 221, 90, 153, 60, 0, 42, 0, 112, 88, 161, 86, 60, 0, 42, 0, 210, 158, 209, 158, 63, 86, 118, 111, 98, 216, 70, 223, 60, 0, 42, 0, 32, 131, 191, 132, 134, 86, 133, 133, 167, 133, 60, 0, 42, 0, 220, 107, 157, 134, 235, 107, 3, 105, 229, 85, 70, 115, 137, 140, 81, 86, 84, 115, 106, 140, 101, 216, 184, 221, 55, 86, 20, 59, 75, 115, 213, 138, 43, 81, 142, 86, 213, 88, 25, 56, 224, 111, 71, 124, 20, 136, 121, 139, 60, 0, 42, 0, 125, 89, 221, 144, 60, 0, 42, 0, 247, 83, 170, 54, 64, 55, 10, 102, 38, 102, 207, 121, 224, 84, 252, 92, 79, 96, 142, 96, 105, 109, 23, 128, 103, 102, 54, 59, 207, 109, 144, 80, 147, 118, 200, 110, 85, 128, 95, 134, 23, 145, 164, 102, 173, 102, 148, 111, 156, 118, 158, 118, 205, 102, 161, 118, 165, 118, 131, 133, 172, 147, 162, 152, 79, 112, 101, 152, 29, 156, 93, 112, 60, 0, 42, 0, 211, 122, 60, 0, 42, 0, 65, 159, 60, 0, 42, 0, 175, 79, 39, 57, 230, 119, 137, 85, 63, 94, 52, 115, 7, 145, 11, 59, 84, 132, 10, 118, 122, 119, 204, 123, 199, 124, 237, 127, 186, 154, 245, 127, 109, 147, 49, 153, 248, 155, 60, 0, 42, 0, 60, 84, 188, 114, 60, 0, 42, 0, 14, 84, 154, 83, 149, 87, 140, 95, 9, 109, 200, 144, 25, 80, 5, 144, 32, 88, 94, 140, 142, 156, 152, 156, 156, 155, 95, 156, 60, 0, 42, 0, 68, 79, 182, 112, 120, 152, 55, 57, 182, 134, 99, 145, 7, 152, 232, 90, 61, 140, 168, 97, 160, 153, 91, 107, 62, 159, 60, 0, 42, 0, 151, 144, 43, 84, 253, 81, 129, 84, 163, 128, 175, 144, 254, 81, 119, 134, 5, 85, 5, 87, 34, 90, 91, 109, 33, 93, 87, 102, 146, 104, 181, 109, 19, 113, 0, 116, 210, 91, 69, 93, 128, 59, 233, 151, 29, 117, 104, 123, 4, 66, 44, 135, 143, 111, 225, 146, 125, 155, 211, 151, 60, 0, 42, 0, 6, 78, 136, 83, 85, 127, 107, 109, 138, 85, 26, 150, 10, 133, 1, 62, 67, 140, 43, 155, 26, 54, 60, 0, 42, 0, 73, 108, 125, 92, 87, 108, 236, 149, 241, 101, 190, 92, 251, 84, 190, 87, 141, 96, 77, 99, 134, 109, 2, 115, 139, 58, 88, 102, 101, 102, 10, 113, 159, 131, 236, 145, 136, 149, 148, 118, 69, 119, 225, 131, 188, 80, 255, 134, 148, 152, 175, 153, 34, 111, 45, 135, 139, 140, 150, 100, 181, 102, 175, 113, 178, 146, 206, 146, 190, 97, 188, 100, 83, 216, 201, 220, 240, 127, 146, 135, 55, 152, 68, 152, 251, 153, 215, 150, 26, 112, 64, 139, 43, 134, 190, 157, 60, 0, 42, 0, 111, 81, 51, 114, 60, 0, 42, 0, 235, 98, 213, 117, 142, 151, 60, 0, 42, 0, 119, 79, 136, 95, 224, 114, 106, 138, 60, 0, 42, 0, 104, 96, 60, 0, 42, 0, 47, 89, 32, 60, 60, 0, 42, 0, 187, 101, 109, 103, 192, 130, 210, 143, 215, 126, 233, 115, 16, 123, 42, 130, 162, 134, 21, 70, 131, 152, 165, 140, 85, 123, 78, 125, 15, 152, 96, 216, 72, 220, 103, 155, 60, 0, 42, 0, 134, 108, 60, 0, 42, 0, 179, 87, 60, 0, 42, 0, 168, 78, 252, 84, 153, 96, 72, 85, 29, 129, 60, 0, 42, 0, 48, 53, 238, 89, 70, 96, 82, 96, 65, 104, 198, 112, 251, 128, 59, 158, 42, 106, 107, 106, 97, 136, 52, 157, 5, 134, 69, 148, 60, 0, 42, 0, 60, 88, 60, 0, 42, 0, 165, 109, 70, 157, 60, 0, 42, 0, 78, 78, 111, 78, 34, 83, 77, 134, 124, 84, 128, 87, 253, 95, 18, 102, 246, 102, 216, 108, 87, 96, 192, 112, 248, 130, 119, 143, 43, 83, 63, 85, 218, 96, 254, 57, 244, 109, 86, 134, 228, 142, 17, 86, 227, 91, 249, 110, 228, 68, 208, 150, 96, 94, 47, 98, 81, 107, 94, 216, 83, 222, 180, 129, 60, 139, 60, 0, 42, 0, 235, 86, 135, 98, 39, 95, 208, 114, 225, 128, 246, 88, 243, 116, 186, 150, 67, 216, 49, 221, 247, 88, 155, 101, 0, 113, 197, 53, 150, 85, 250, 88, 169, 90, 86, 110, 34, 115, 87, 125, 48, 100, 92, 105, 115, 113, 90, 116, 107, 132, 29, 86, 80, 216, 13, 223, 85, 158, 242, 105, 182, 123, 202, 124, 27, 133, 116, 135, 90, 136, 113, 155, 32, 126, 156, 135, 144, 145, 54, 152, 243, 137, 120, 147, 44, 153, 76, 75, 96, 157, 43, 112, 13, 155, 23, 156, 152, 157, 166, 157, 60, 0, 42, 0, 85, 78, 123, 108, 78, 134, 82, 109, 255, 79, 44, 85, 37, 116, 0, 132, 93, 134, 248, 110, 60, 0, 42, 0, 101, 78, 146, 78, 22, 95, 54, 98, 55, 98, 56, 98, 38, 62, 177, 81, 180, 81, 13, 94, 164, 98, 141, 108, 170, 108, 144, 130, 181, 92, 25, 96, 61, 98, 8, 102, 145, 103, 24, 96, 92, 121, 15, 123, 95, 90, 72, 98, 224, 116, 192, 85, 91, 105, 148, 125, 253, 150, 237, 90, 238, 90, 98, 100, 236, 110, 32, 145, 244, 105, 169, 113, 48, 133, 248, 156, 4, 124, 89, 147, 155, 86, 113, 158, 119, 139, 224, 156, 196, 151, 0, 152, 111, 156, 12, 158, 60, 0, 42, 0, 144, 124, 63, 147, 241, 155, 60, 0, 42, 0, 177, 130, 178, 130, 215, 84, 41, 86, 138, 132, 53, 147, 60, 0, 42, 0, 78, 83, 114, 53, 73, 59, 225, 89, 133, 154, 234, 145, 235, 145, 231, 148, 239, 131, 51, 100, 209, 110, 62, 115, 15, 56, 138, 57, 182, 100, 198, 120, 216, 146, 80, 133, 150, 135, 65, 139, 245, 147, 74, 154, 232, 157, 60, 0, 42, 0, 22, 83, 18, 82, 102, 53, 59, 89, 119, 53, 121, 103, 59, 117, 221, 139, 102, 104, 115, 90, 11, 93, 107, 117, 117, 117, 223, 137, 113, 138, 131, 82, 102, 100, 255, 90, 5, 91, 236, 105, 133, 111, 58, 106, 227, 138, 99, 126, 25, 130, 202, 158, 110, 139, 60, 0, 42, 0, 214, 87, 114, 90, 27, 105, 116, 120, 192, 124, 174, 138, 141, 116, 60, 0, 42, 0, 25, 84, 104, 82, 32, 128, 42, 149, 144, 82, 67, 147, 132, 86, 65, 140, 9, 101, 30, 154, 60, 0, 42, 0, 120, 79, 59, 109, 238, 121, 243, 121, 60, 0, 42, 0, 107, 112, 25, 79, 169, 144, 172, 148, 37, 146, 37, 89, 55, 111, 60, 0, 42, 0, 142, 108, 22, 98, 39, 141, 159, 84, 9, 120, 0, 65, 240, 79, 71, 99, 19, 119, 132, 149, 67, 216, 126, 221, 157, 99, 183, 131, 168, 140, 209, 96, 113, 110, 120, 121, 228, 101, 71, 105, 47, 58, 141, 121, 166, 132, 242, 53, 111, 89, 13, 151, 233, 111, 114, 115, 11, 139, 180, 106, 108, 149, 175, 86, 22, 112, 198, 119, 107, 122, 91, 76, 228, 102, 47, 128, 219, 129, 103, 130, 255, 133, 22, 136, 191, 86, 104, 118, 208, 119, 74, 148, 67, 151, 60, 0, 42, 0, 0, 96, 138, 95, 238, 109, 208, 105, 29, 142, 208, 97, 34, 137, 49, 137, 51, 55, 34, 56, 247, 97, 36, 112, 240, 106, 50, 128, 57, 134, 60, 0, 42, 0, 79, 87, 182, 84, 217, 138, 202, 88, 222, 88, 62, 134, 60, 0, 42, 0, 112, 112, 209, 57, 217, 139, 180, 84, 98, 96, 251, 98, 37, 99, 3, 109, 122, 134, 86, 102, 227, 112, 242, 115, 134, 136, 87, 140, 78, 90, 69, 86, 136, 90, 238, 99, 199, 61, 218, 127, 137, 143, 137, 102, 78, 105, 71, 113, 124, 138, 147, 150, 81, 94, 115, 119, 136, 121, 157, 100, 85, 86, 236, 127, 24, 137, 29, 143, 190, 158, 189, 95, 8, 112, 179, 150, 51, 134, 52, 156, 60, 0, 42, 0, 216, 86, 222, 86, 236, 86, 106, 79, 251, 94, 91, 96, 4, 109, 253, 94, 224, 112, 52, 131, 152, 134, 244, 143, 208, 117, 37, 144, 212, 134, 213, 134, 22, 135, 176, 155, 60, 0, 42, 0, 148, 96, 192, 107, 193, 107, 89, 216, 153, 223, 199, 107, 147, 106, 236, 113, 109, 139, 60, 0, 42, 0, 73, 83, 71, 108, 26, 79, 179, 139, 203, 108, 213, 84, 69, 216, 38, 223, 77, 109, 216, 126, 148, 130, 242, 139, 90, 96, 117, 96, 233, 112, 95, 131, 63, 141, 87, 95, 102, 102, 253, 121, 153, 85, 224, 96, 79, 110, 117, 125, 11, 127, 217, 127, 19, 150, 47, 83, 89, 95, 90, 95, 99, 57, 3, 103, 217, 110, 111, 138, 196, 140, 146, 152, 225, 80, 18, 86, 57, 58, 168, 138, 26, 87, 237, 91, 103, 97, 147, 97, 179, 102, 229, 105, 83, 111, 35, 118, 39, 133, 229, 73, 102, 86, 18, 91, 187, 95, 94, 106, 168, 107, 174, 111, 202, 111, 105, 115, 89, 133, 241, 138, 46, 152, 244, 113, 175, 116, 242, 123, 136, 133, 137, 133, 47, 153, 150, 86, 186, 119, 98, 122, 98, 126, 234, 135, 41, 59, 216, 106, 106, 126, 253, 127, 241, 133, 83, 139, 54, 81, 248, 147, 224, 149, 72, 91, 44, 148, 103, 151, 127, 139, 106, 152, 60, 0, 42, 0, 119, 92, 115, 112, 164, 116, 243, 97, 60, 0, 42, 0, 34, 107, 191, 114, 83, 107, 5, 157, 85, 62, 77, 157, 190, 86, 253, 97, 126, 115, 68, 145, 97, 107, 153, 139, 155, 140, 105, 154, 60, 0, 42, 0, 175, 115, 216, 92, 57, 109, 223, 114, 199, 144, 83, 104, 65, 131, 207, 91, 89, 125, 8, 132, 17, 132, 200, 150, 132, 125, 166, 127, 134, 140, 110, 146, 62, 149, 28, 87, 27, 91, 240, 91, 180, 111, 51, 127, 161, 73, 27, 150, 176, 116, 114, 140, 112, 147, 110, 149, 110, 158, 235, 124, 111, 126, 88, 143, 54, 148, 228, 149, 31, 155, 219, 116, 9, 77, 60, 0, 42, 0, 19, 127, 32, 66, 233, 125, 12, 101, 60, 0, 42, 0, 123, 94, 66, 89, 146, 128, 80, 89, 166, 91, 36, 85, 98, 99, 99, 109, 163, 109, 201, 112, 163, 96, 153, 104, 21, 113, 154, 85, 155, 85, 72, 93, 12, 97, 219, 99, 25, 110, 234, 117, 70, 119, 45, 144, 7, 59, 10, 59, 101, 113, 77, 116, 98, 140, 54, 111, 19, 118, 245, 105, 169, 156, 208, 100, 163, 111, 199, 155, 231, 133, 0, 156, 60, 0, 42, 0, 163, 119, 60, 0, 42, 0, 15, 102, 44, 102, 100, 131, 90, 90, 219, 96, 189, 109, 13, 150, 212, 104, 153, 107, 103, 119, 119, 132, 111, 119, 189, 149, 60, 0, 42, 0, 246, 95, 81, 109, 161, 104, 132, 153, 26, 88, 62, 110, 63, 116, 66, 155, 27, 153, 73, 126, 75, 143, 157, 75, 50, 159, 60, 0, 42, 0, 246, 155, 60, 0, 42, 0, 232, 139, 210, 79, 49, 80, 2, 87, 141, 99, 247, 109, 29, 113, 183, 110, 65, 97, 232, 137, 226, 138, 60, 0, 42, 0, 223, 93, 53, 56, 147, 128, 65, 136, 82, 131, 26, 103, 67, 88, 76, 97, 59, 59, 60, 0, 42, 0, 135, 118, 95, 80, 240, 81, 164, 85, 45, 88, 147, 90, 50, 93, 168, 95, 246, 96, 95, 110, 141, 150, 195, 158, 196, 158, 123, 105, 76, 113, 93, 116, 95, 132, 81, 144, 61, 149, 180, 88, 98, 111, 90, 115, 193, 123, 202, 123, 78, 130, 87, 135, 191, 113, 156, 116, 251, 138, 64, 118, 250, 120, 127, 65, 84, 122, 96, 147, 45, 153, 199, 156, 39, 124, 229, 135, 249, 151, 170, 141, 28, 154, 4, 148, 9, 156, 100, 81, 81, 156, 236, 157, 60, 0, 42, 0, 51, 96, 77, 96, 190, 112, 186, 91, 67, 102, 68, 102, 91, 89, 14, 140, 76, 94, 100, 138, 128, 113, 32, 68, 10, 139, 206, 106, 60, 0, 42, 0, 48, 97, 42, 58, 201, 110, 165, 105, 157, 118, 194, 102, 164, 147, 169, 118, 60, 0, 42, 0, 40, 126, 60, 0, 42, 0, 255, 83, 61, 84, 77, 84, 116, 112, 112, 143, 196, 84, 7, 138, 216, 112, 227, 142, 200, 99, 57, 110, 34, 113, 97, 120, 62, 140, 101, 216, 32, 220, 55, 143, 168, 133, 103, 147, 157, 86, 95, 143, 60, 0, 42, 0, 220, 78, 24, 95, 133, 89, 162, 126, 48, 84, 143, 91, 111, 108, 146, 115, 174, 126, 243, 149, 150, 91, 211, 108, 172, 87, 2, 90, 42, 109, 209, 122, 122, 66, 5, 125, 240, 130, 121, 134, 245, 92, 100, 109, 24, 125, 195, 127, 62, 128, 109, 131, 84, 120, 45, 125, 57, 140, 63, 158, 49, 110, 228, 122, 160, 124, 28, 146, 142, 149, 139, 125, 221, 127, 82, 132, 83, 132, 60, 140, 66, 111, 96, 216, 6, 223, 119, 146, 131, 151, 95, 155, 208, 146, 75, 95, 16, 151, 123, 133, 31, 151, 59, 157, 201, 158, 80, 216, 15, 221, 204, 158, 60, 0, 42, 0, 78, 102, 52, 59, 202, 85, 60, 0, 42, 0, 167, 139, 12, 138, 167, 149, 148, 100, 139, 111, 146, 111, 190, 146, 194, 149, 40, 155, 60, 0, 39, 0, 208, 253, 39, 0, 16, 49, 60, 0, 42, 0, 12, 78, 74, 216, 102, 222, 165, 139, 251, 81, 9, 82, 253, 83, 101, 153, 105, 78, 15, 82, 58, 103, 145, 115, 140, 128, 62, 87, 246, 119, 33, 158, 133, 103, 168, 130, 173, 84, 235, 89, 94, 82, 39, 85, 236, 89, 80, 92, 239, 121, 4, 123, 249, 143, 226, 152, 250, 87, 233, 126, 158, 85, 70, 93, 71, 93, 103, 101, 30, 103, 132, 114, 83, 123, 9, 127, 77, 141, 227, 82, 216, 85, 120, 117, 24, 122, 225, 141, 251, 141, 238, 156, 223, 80, 67, 216, 59, 223, 196, 107, 149, 123, 136, 146, 48, 86, 227, 105, 127, 117, 61, 122, 221, 125, 237, 137, 235, 140, 184, 142, 81, 159, 188, 88, 95, 106, 192, 111, 163, 116, 77, 122, 36, 147, 202, 100, 239, 120, 168, 121, 10, 124, 62, 126, 129, 127, 64, 137, 247, 140, 174, 150, 197, 106, 45, 128, 95, 142, 222, 150, 79, 139, 63, 145, 242, 151, 143, 157, 100, 139, 22, 148, 81, 153, 106, 118, 139, 142, 191, 151, 196, 157, 78, 159, 135, 127, 71, 148, 64, 134, 137, 137, 89, 148, 79, 159, 136, 127, 4, 158, 138, 137, 60, 0, 42, 0, 188, 78, 202, 83, 11, 79, 9, 84, 167, 126, 115, 83, 140, 92, 118, 95, 227, 95, 129, 103, 114, 108, 128, 118, 159, 78, 118, 79, 216, 139, 145, 148, 125, 83, 222, 89, 37, 96, 228, 114, 198, 144, 228, 99, 190, 117, 141, 118, 8, 123, 26, 125, 10, 129, 152, 71, 110, 80, 89, 83, 180, 94, 114, 59, 15, 113, 217, 137, 59, 140, 202, 53, 216, 104, 82, 110, 198, 150, 73, 88, 201, 90, 49, 97, 34, 98, 107, 105, 117, 105, 155, 107, 140, 141, 145, 143, 201, 105, 36, 128, 140, 129, 186, 132, 161, 146, 241, 53, 131, 100, 87, 111, 32, 118, 157, 121, 191, 123, 22, 142, 97, 158, 175, 93, 118, 106, 200, 111, 64, 133, 143, 135, 47, 143, 238, 100, 157, 106, 122, 133, 80, 142, 83, 147, 201, 133, 75, 137, 101, 130, 77, 124, 90, 143, 246, 147, 237, 76, 53, 151, 186, 157, 209, 157, 230, 150, 231, 150, 60, 0, 42, 0, 224, 81, 241, 93, 46, 78, 128, 89, 177, 114, 242, 108, 110, 134, 36, 99, 142, 99, 126, 156, 126, 94, 184, 57, 31, 98, 24, 146, 116, 93, 130, 158, 98, 155, 160, 100, 224, 100, 86, 122, 227, 135, 85, 155, 60, 0, 42, 0, 149, 46, 80, 95, 81, 95, 225, 101, 161, 139, 176, 139, 14, 79, 170, 126, 86, 87, 147, 89, 204, 95, 128, 98, 66, 82, 99, 91, 176, 130, 69, 150, 220, 84, 141, 87, 220, 92, 226, 101, 14, 109, 78, 109, 0, 125, 13, 131, 8, 138, 100, 82, 18, 125, 231, 126, 36, 131, 96, 131, 202, 137, 24, 138, 72, 80, 194, 91, 196, 91, 49, 56, 155, 95, 77, 57, 184, 96, 227, 101, 158, 104, 8, 110, 109, 121, 72, 88, 206, 96, 209, 62, 0, 68, 238, 129, 126, 81, 245, 117, 153, 125, 106, 132, 7, 133, 218, 136, 168, 102, 3, 111, 8, 111, 41, 122, 74, 122, 223, 132, 0, 137, 139, 138, 253, 141, 155, 150, 1, 151, 62, 155, 154, 156, 169, 102, 55, 122, 197, 138, 171, 156, 128, 81, 145, 82, 193, 102, 68, 122, 251, 154, 140, 86, 149, 106, 223, 111, 125, 127, 138, 133, 172, 137, 181, 106, 4, 62, 186, 133, 75, 157, 76, 159, 251, 97, 96, 118, 103, 122, 75, 126, 165, 154, 218, 155, 49, 112, 124, 126, 14, 134, 102, 216, 73, 221, 64, 156, 46, 134, 61, 151, 54, 156, 63, 156, 59, 134, 109, 156, 101, 154, 60, 0, 42, 0, 189, 78, 45, 106, 237, 94, 60, 0, 42, 0, 160, 82, 107, 78, 57, 89, 61, 79, 62, 89, 115, 79, 184, 98, 193, 98, 199, 108, 98, 62, 183, 103, 224, 107, 67, 109, 200, 115, 4, 131, 230, 143, 201, 87, 182, 91, 121, 109, 194, 117, 156, 104, 51, 123, 30, 128, 136, 136, 162, 80, 51, 115, 207, 141, 67, 216, 162, 222, 140, 114, 117, 129, 109, 132, 107, 146, 9, 86, 127, 146, 83, 149, 109, 140, 145, 140, 103, 216, 155, 221, 181, 147, 82, 69, 154, 158, 60, 0, 42, 0, 63, 87, 230, 95, 116, 98, 207, 144, 10, 85, 93, 96, 90, 131, 223, 144, 27, 98, 162, 131, 215, 148, 30, 98, 241, 134, 183, 136, 138, 152, 150, 59, 250, 134, 204, 136, 242, 141, 130, 151, 4, 153, 207, 146, 44, 152, 48, 152, 54, 157, 74, 157, 60, 0, 42, 0, 50, 117, 238, 78, 172, 92, 218, 83, 190, 115, 219, 128, 154, 101, 62, 141, 190, 148, 71, 80, 125, 90, 166, 95, 157, 101, 53, 105, 200, 140, 64, 146, 142, 105, 218, 105, 21, 118, 159, 106, 60, 0, 42, 0, 247, 78, 126, 154, 182, 103, 75, 216, 76, 220, 193, 90, 79, 94, 162, 105, 249, 80, 60, 122, 213, 153, 60, 0, 42, 0, 216, 124, 60, 0, 42, 0, 150, 117, 54, 150, 134, 118, 165, 99, 178, 99, 79, 216, 143, 221, 206, 117, 248, 121, 136, 85, 38, 88, 152, 90, 237, 99, 4, 105, 93, 110, 59, 129, 232, 131, 87, 136, 142, 150, 223, 85, 197, 90, 111, 113, 200, 64, 45, 122, 84, 135, 209, 100, 100, 118, 47, 139, 155, 157, 60, 0, 42, 0, 105, 83, 106, 83, 81, 91, 16, 92, 166, 139, 130, 130, 38, 82, 39, 82, 171, 82, 138, 92, 60, 82, 26, 53, 188, 82, 5, 102, 112, 103, 140, 117, 54, 54, 195, 54, 238, 98, 1, 109, 211, 126, 34, 80, 64, 104, 113, 136, 16, 138, 252, 143, 124, 80, 85, 90, 40, 93, 119, 99, 173, 131, 145, 80, 188, 85, 23, 58, 57, 64, 80, 125, 92, 125, 186, 136, 137, 152, 101, 93, 41, 58, 108, 105, 118, 105, 208, 110, 107, 119, 192, 123, 16, 135, 77, 135, 112, 138, 99, 146, 93, 155, 42, 98, 164, 105, 163, 120, 237, 122, 91, 216, 116, 221, 146, 156, 84, 111, 175, 127, 245, 132, 177, 138, 21, 142, 138, 151, 111, 94, 123, 147, 154, 155, 192, 93, 237, 106, 30, 136, 24, 136, 61, 136, 60, 0, 42, 0, 208, 89, 209, 107, 142, 90, 227, 137, 231, 137, 247, 152, 158, 106, 60, 0, 42, 0, 47, 78, 203, 78, 36, 84, 149, 92, 57, 56, 142, 94, 18, 98, 70, 92, 74, 92, 160, 115, 165, 130, 76, 117, 77, 117, 165, 117, 14, 120, 235, 139, 31, 80, 136, 96, 167, 134, 120, 136, 163, 95, 58, 88, 80, 105, 62, 116, 246, 134, 151, 114, 161, 138, 177, 154, 106, 155, 47, 137, 133, 147, 164, 142, 60, 0, 42, 0, 93, 104, 60, 0, 42, 0, 164, 78, 125, 130, 129, 130, 227, 89, 7, 90, 231, 92, 71, 109, 202, 144, 132, 154, 246, 128, 45, 131, 46, 131, 18, 105, 38, 113, 223, 134, 228, 141, 236, 80, 4, 86, 68, 216, 237, 223, 96, 134, 155, 156, 12, 91, 149, 93, 163, 93, 141, 97, 134, 111, 160, 129, 203, 113, 178, 129, 73, 133, 1, 121, 90, 122, 171, 155, 65, 157, 106, 158, 37, 124, 237, 135, 71, 143, 14, 148, 205, 157, 85, 154, 230, 157, 238, 157, 60, 0, 42, 0, 66, 59, 235, 129, 210, 137, 124, 79, 165, 79, 84, 96, 34, 99, 225, 114, 222, 126, 122, 153, 65, 99, 72, 102, 196, 112, 142, 118, 235, 119, 26, 129, 240, 148, 5, 100, 107, 110, 94, 125, 127, 82, 107, 101, 108, 110, 77, 113, 115, 129, 203, 140, 229, 80, 119, 100, 158, 102, 11, 142, 120, 146, 3, 153, 12, 81, 139, 82, 186, 95, 159, 100, 185, 100, 188, 95, 191, 97, 125, 101, 127, 101, 222, 113, 52, 127, 166, 150, 210, 102, 172, 116, 239, 119, 166, 118, 220, 135, 115, 126, 81, 139, 66, 91, 112, 58, 92, 216, 136, 220, 42, 101, 90, 112, 78, 156, 60, 0, 42, 0, 235, 83, 76, 84, 228, 92, 13, 99, 6, 138, 211, 115, 140, 122, 127, 143, 131, 143, 78, 101, 89, 101, 150, 122, 62, 60, 216, 110, 3, 143, 2, 86, 38, 86, 160, 101, 22, 111, 117, 145, 77, 86, 160, 93, 80, 111, 109, 86, 19, 91, 101, 115, 224, 133, 173, 141, 78, 143, 174, 145, 101, 139, 45, 59, 173, 118, 194, 145, 60, 0, 42, 0, 100, 157, 245, 106, 144, 126, 60, 0, 42, 0, 41, 78, 252, 82, 160, 126, 59, 103, 94, 114, 118, 122, 250, 124, 32, 158, 254, 124, 115, 141, 99, 216, 202, 220, 4, 150, 126, 85, 194, 99, 234, 99, 27, 132, 235, 99, 233, 156, 78, 100, 27, 106, 15, 155, 46, 155, 60, 0, 42, 0, 93, 78, 69, 78, 70, 78, 99, 78, 122, 89, 120, 112, 150, 115, 15, 130, 237, 151, 36, 125, 82, 145, 121, 149, 238, 151, 60, 0, 42, 0, 27, 83, 231, 101, 169, 62, 252, 129, 142, 84, 154, 117, 146, 65, 233, 103, 254, 103, 3, 80, 68, 99, 85, 104, 19, 83, 169, 83, 81, 101, 168, 90, 49, 92, 196, 94, 5, 130, 230, 80, 207, 94, 208, 94, 102, 97, 167, 107, 107, 158, 10, 130, 54, 83, 230, 155, 148, 158, 104, 159, 242, 157, 60, 0, 42, 0, 99, 108, 102, 103, 13, 107, 60, 0, 42, 0, 11, 98, 120, 89, 22, 92, 117, 94, 90, 87, 124, 107, 244, 149, 191, 81, 20, 98, 170, 115, 169, 128, 112, 130, 230, 89, 231, 89, 124, 81, 209, 118, 66, 80, 5, 88, 228, 96, 116, 59, 15, 115, 58, 123, 92, 140, 84, 110, 75, 114, 4, 127, 197, 131, 250, 131, 147, 149, 27, 100, 55, 105, 62, 105, 78, 113, 141, 114, 74, 116, 138, 120, 35, 127, 76, 132, 99, 140, 227, 118, 119, 119, 139, 123, 185, 132, 43, 106, 158, 113, 216, 125, 163, 156, 253, 156, 99, 158, 184, 113, 86, 216, 93, 220, 239, 123, 17, 126, 81, 133, 115, 133, 113, 130, 48, 153, 162, 153, 137, 158, 16, 112, 172, 151, 175, 151, 210, 156, 27, 121, 32, 67, 184, 137, 115, 157, 56, 112, 39, 148, 252, 106, 178, 107, 188, 157, 192, 151, 57, 156, 207, 86, 67, 134, 111, 148, 201, 151, 60, 0, 42, 0, 221, 86, 227, 98, 167, 103, 237, 79, 236, 103, 57, 80, 56, 99, 97, 99, 21, 123, 39, 131, 207, 81, 106, 82, 152, 104, 192, 104, 63, 88, 192, 99, 195, 99, 28, 105, 27, 110, 85, 110, 81, 119, 119, 120, 67, 138, 15, 149, 63, 95, 149, 102, 80, 116, 103, 123, 128, 123, 120, 125, 229, 136, 43, 140, 188, 141, 41, 98, 44, 98, 53, 58, 177, 120, 9, 81, 230, 127, 191, 100, 162, 106, 77, 216, 45, 223, 7, 139, 71, 142, 188, 119, 6, 121, 33, 124, 198, 133, 71, 137, 73, 137, 45, 139, 109, 126, 11, 155, 120, 158, 61, 112, 18, 136, 23, 148, 14, 156, 151, 82, 123, 158, 91, 124, 126, 139, 124, 158, 122, 137, 60, 0, 42, 0, 193, 137, 246, 78, 139, 137, 111, 153, 81, 82, 250, 94, 10, 109, 110, 114, 49, 141, 244, 79, 99, 82, 43, 104, 167, 109, 212, 115, 48, 130, 80, 131, 101, 80, 113, 82, 164, 95, 16, 110, 15, 140, 252, 145, 203, 91, 133, 110, 246, 129, 184, 136, 245, 141, 212, 101, 87, 105, 253, 107, 82, 216, 227, 221, 113, 129, 101, 132, 206, 140, 116, 146, 46, 149, 237, 80, 52, 58, 151, 105, 56, 111, 141, 82, 142, 82, 78, 61, 151, 111, 173, 123, 203, 124, 42, 133, 211, 138, 228, 140, 157, 141, 16, 142, 58, 142, 146, 82, 148, 82, 235, 138, 251, 146, 30, 153, 183, 119, 245, 120, 166, 133, 185, 135, 115, 147, 117, 147, 246, 100, 250, 111, 93, 126, 4, 60, 181, 137, 233, 147, 51, 112, 102, 130, 124, 139, 94, 143, 49, 148, 81, 148, 82, 148, 108, 148, 115, 148, 60, 0, 42, 0, 69, 95, 185, 88, 122, 106, 0, 121, 177, 107, 60, 0, 42, 0, 254, 93, 202, 78, 164, 101, 133, 148, 83, 81, 209, 145, 210, 145, 37, 109, 220, 119, 212, 137, 208, 87, 210, 115, 31, 125, 85, 131, 127, 136, 205, 96, 59, 88, 75, 123, 255, 145, 156, 93, 118, 158, 197, 158, 95, 137, 60, 0, 42, 0, 197, 78, 61, 92, 173, 79, 122, 83, 249, 93, 39, 125, 7, 88, 235, 131, 197, 80, 170, 83, 40, 140, 38, 149, 228, 90, 209, 94, 12, 111, 225, 118, 202, 125, 145, 153, 255, 105, 126, 116, 243, 132, 24, 81, 38, 147, 57, 139, 73, 153, 60, 0, 42, 0, 18, 79, 164, 82, 178, 82, 151, 89, 131, 103, 209, 143, 219, 143, 193, 82, 85, 109, 73, 102, 75, 102, 120, 109, 236, 112, 105, 131, 70, 141, 43, 85, 14, 116, 114, 121, 50, 144, 214, 91, 34, 100, 141, 110, 129, 121, 25, 127, 115, 151, 144, 88, 55, 58, 16, 59, 156, 102, 104, 116, 248, 80, 218, 81, 79, 107, 163, 107, 208, 137, 100, 86, 28, 59, 197, 111, 161, 116, 9, 126, 238, 140, 141, 86, 39, 91, 242, 59, 220, 111, 252, 113, 182, 116, 206, 133, 178, 137, 16, 141, 125, 159, 60, 0, 42, 0, 27, 120, 59, 116, 215, 88, 60, 0, 42, 0, 95, 108, 220, 89, 6, 92, 70, 109, 85, 117, 51, 131, 71, 140, 7, 92, 122, 117, 65, 132, 106, 100, 222, 127, 245, 80, 63, 111, 128, 135, 195, 88, 48, 127, 127, 106, 173, 107, 145, 133, 191, 135, 201, 156, 133, 117, 19, 121, 134, 117, 110, 126, 193, 151, 66, 156, 60, 0, 42, 0, 178, 139, 86, 89, 104, 104, 139, 80, 104, 89, 139, 132, 108, 89, 243, 105, 78, 115, 153, 129, 35, 133, 41, 128, 67, 66, 27, 139, 92, 152, 60, 0, 42, 0, 30, 83, 32, 83, 5, 89, 28, 95, 26, 109, 219, 126, 77, 150, 54, 95, 115, 125, 182, 136, 229, 82, 113, 145, 121, 93, 126, 100, 240, 110, 74, 95, 77, 216, 166, 222, 159, 114, 225, 124, 232, 124, 164, 145, 61, 139, 172, 145, 60, 0, 42, 0, 98, 103, 228, 106, 60, 0, 42, 0, 85, 87, 89, 87, 224, 93, 172, 78, 254, 108, 207, 126, 176, 78, 212, 121, 14, 131, 135, 109, 74, 131, 91, 90, 202, 96, 204, 101, 205, 101, 4, 115, 76, 125, 150, 131, 118, 102, 9, 122, 72, 129, 70, 131, 193, 131, 91, 119, 179, 124, 147, 125, 79, 132, 98, 81, 190, 124, 89, 128, 184, 156, 91, 157, 232, 155, 129, 157, 132, 157, 150, 158, 49, 159, 90, 154, 160, 158, 60, 0, 42, 0, 149, 78, 60, 78, 45, 82, 43, 54, 83, 87, 145, 91, 107, 108, 49, 150, 108, 108, 188, 128, 68, 82, 125, 122, 136, 152, 111, 102, 26, 152, 6, 81, 92, 94, 172, 97, 187, 102, 219, 113, 159, 116, 56, 152, 188, 97, 165, 116, 252, 135, 24, 71, 102, 139, 60, 0, 42, 0, 140, 89, 192, 81, 42, 95, 132, 95, 211, 79, 65, 216, 202, 221, 68, 109, 235, 128, 243, 143, 30, 80, 200, 81, 51, 95, 145, 95, 201, 117, 222, 122, 89, 90, 103, 90, 113, 104, 183, 104, 232, 109, 27, 129, 21, 144, 100, 216, 213, 221, 223, 122, 217, 117, 231, 122, 83, 151, 185, 80, 108, 101, 12, 59, 235, 122, 86, 151, 131, 88, 77, 115, 169, 138, 1, 142, 89, 151, 124, 71, 90, 151, 92, 149, 92, 151, 212, 102, 234, 111, 30, 112, 225, 147, 246, 122, 248, 122, 60, 0, 42, 0, 120, 106, 221, 113, 60, 0, 42, 0, 229, 81, 31, 82, 133, 98, 10, 83, 69, 92, 216, 98, 195, 108, 217, 114, 121, 154, 244, 130, 54, 99, 189, 117, 192, 117, 23, 119, 32, 120, 93, 127, 53, 90, 110, 90, 12, 93, 172, 99, 174, 104, 186, 109, 113, 150, 16, 105, 26, 116, 82, 129, 249, 131, 132, 141, 212, 141, 20, 149, 206, 150, 77, 130, 27, 135, 254, 136, 24, 142, 25, 142, 230, 146, 210, 153, 136, 155, 33, 157, 160, 151, 171, 151, 139, 157, 60, 0, 42, 0, 90, 216, 81, 221, 64, 92, 50, 54, 230, 108, 183, 79, 202, 114, 84, 104, 233, 107, 121, 85, 69, 90, 215, 109, 23, 113, 8, 105, 241, 107, 104, 110, 202, 131, 249, 144, 145, 114, 2, 143, 234, 80, 183, 124, 252, 141, 202, 138, 156, 141, 185, 142, 176, 149, 88, 106, 139, 106, 246, 153, 212, 76, 89, 157, 107, 142, 116, 157, 200, 93, 170, 157, 28, 134, 51, 159, 103, 154, 60, 0, 42, 0, 128, 84, 6, 95, 174, 108, 62, 78, 25, 99, 233, 119, 146, 131, 7, 105, 137, 105, 152, 105, 101, 123, 131, 159, 159, 132, 101, 128, 9, 130, 61, 142, 231, 100, 95, 159, 248, 106, 5, 107, 60, 0, 42, 0, 229, 83, 232, 93, 108, 78, 234, 93, 181, 139, 119, 81, 214, 89, 160, 92, 7, 96, 16, 96, 26, 96, 210, 98, 224, 98, 48, 109, 173, 144, 27, 102, 107, 107, 172, 112, 227, 130, 156, 148, 241, 79, 40, 80, 54, 80, 163, 81, 103, 82, 236, 121, 231, 87, 254, 87, 231, 96, 110, 99, 148, 124, 31, 128, 183, 134, 147, 136, 139, 114, 78, 138, 217, 141, 221, 141, 211, 152, 102, 140, 69, 146, 47, 149, 224, 91, 51, 97, 173, 122, 90, 128, 97, 134, 135, 82, 238, 82, 102, 92, 30, 142, 207, 153, 201, 88, 197, 97, 218, 100, 189, 111, 182, 122, 46, 66, 248, 146, 148, 155, 104, 92, 125, 144, 182, 152, 151, 140, 52, 124, 134, 142, 181, 145, 252, 97, 59, 148, 44, 59, 60, 0, 42, 0, 32, 114, 119, 137, 60, 0, 42, 0, 88, 86, 133, 100, 167, 100, 105, 92, 123, 142, 60, 0, 42, 0, 133, 78, 64, 216, 12, 221, 82, 91, 83, 91, 179, 81, 20, 82, 18, 108, 192, 139, 137, 98, 122, 108, 33, 95, 72, 59, 236, 108, 166, 115, 168, 115, 181, 130, 23, 99, 207, 115, 166, 117, 4, 120, 221, 126, 115, 134, 201, 137, 20, 80, 84, 99, 46, 107, 151, 134, 27, 93, 152, 99, 141, 101, 119, 104, 140, 107, 144, 137, 214, 137, 35, 138, 125, 141, 185, 141, 72, 72, 149, 80, 165, 83, 126, 59, 51, 113, 85, 125, 118, 125, 154, 137, 137, 141, 43, 144, 12, 146, 242, 55, 130, 82, 34, 53, 234, 82, 116, 116, 50, 140, 195, 153, 165, 93, 176, 97, 166, 113, 52, 114, 87, 115, 26, 118, 2, 157, 3, 157, 113, 86, 160, 97, 91, 106, 92, 106, 93, 133, 104, 133, 53, 114, 196, 129, 98, 149, 232, 135, 233, 135, 107, 92, 78, 139, 118, 142, 119, 142, 140, 157, 55, 83, 188, 86, 17, 114, 205, 119, 186, 137, 13, 148, 29, 148, 29, 114, 252, 137, 79, 95, 4, 98, 43, 101, 131, 115, 226, 157, 20, 107, 225, 119, 163, 159, 99, 216, 28, 220, 156, 140, 169, 142, 129, 148, 60, 0, 42, 0, 226, 89, 31, 90, 80, 99, 147, 109, 6, 113, 39, 129, 249, 116, 67, 158, 208, 136, 236, 82, 250, 72, 76, 149, 184, 147, 81, 157, 43, 148, 50, 136, 60, 0, 42, 0, 119, 83, 95, 84, 35, 94, 205, 87, 114, 99, 228, 131, 41, 149, 247, 61, 8, 147, 199, 129, 60, 0, 42, 0, 181, 82, 70, 89, 251, 93, 38, 80, 204, 82, 74, 104, 247, 114, 226, 126, 189, 150, 195, 109, 55, 119, 74, 119, 109, 125, 101, 127, 173, 69, 4, 145, 96, 119, 121, 125, 203, 150, 236, 152, 70, 216, 64, 220, 123, 97, 40, 133, 11, 153, 103, 115, 51, 126, 130, 127, 60, 0, 42, 0, 155, 81, 66, 216, 213, 220, 27, 84, 71, 87, 172, 54, 110, 108, 240, 89, 205, 142, 167, 148, 144, 134, 128, 136, 126, 104, 178, 118, 153, 131, 204, 131, 30, 146, 133, 120, 184, 118, 185, 118, 160, 137, 129, 146, 158, 146, 170, 156, 135, 158, 85, 147, 182, 155, 143, 158, 149, 158, 60, 0, 42, 0, 65, 84, 202, 79, 89, 63, 200, 87, 251, 92, 67, 99, 90, 109, 225, 144, 86, 150, 130, 153, 143, 154, 165, 53, 89, 102, 12, 113, 250, 115, 94, 216, 122, 221, 122, 52, 222, 104, 111, 117, 227, 122, 102, 55, 249, 65, 152, 123, 159, 123, 32, 135, 96, 216, 24, 223, 1, 81, 89, 53, 21, 153, 239, 91, 207, 97, 199, 113, 236, 111, 255, 153, 84, 157, 88, 157, 8, 101, 81, 69, 31, 101, 60, 0, 42, 0, 134, 46, 130, 81, 139, 81, 112, 87, 67, 98, 219, 87, 69, 125, 201, 153, 235, 153, 15, 134, 20, 134, 60, 0, 42, 0, 143, 81, 231, 86, 160, 56, 194, 108, 133, 112, 176, 79, 175, 112, 229, 143, 123, 109, 8, 144, 241, 112, 155, 72, 152, 122, 142, 152, 90, 113, 151, 125, 210, 80, 91, 113, 178, 113, 131, 111, 39, 137, 60, 0, 39, 0, 208, 253, 39, 0, 17, 49, 60, 0, 42, 0, 3, 78, 143, 108, 201, 143, 187, 89, 210, 103, 27, 80, 196, 81, 22, 104, 100, 104, 56, 90, 189, 96, 26, 98, 124, 104, 210, 109, 234, 144, 50, 101, 31, 103, 242, 104, 58, 107, 102, 60, 11, 132, 227, 134, 219, 80, 1, 86, 125, 97, 191, 105, 6, 111, 192, 125, 124, 97, 237, 105, 84, 216, 80, 222, 198, 138, 255, 138, 11, 151, 233, 69, 74, 142, 36, 59, 76, 155, 83, 73, 218, 147, 136, 157, 60, 0, 42, 0, 147, 78, 80, 159, 59, 87, 144, 92, 147, 92, 239, 95, 157, 78, 118, 81, 71, 89, 137, 101, 103, 107, 65, 117, 65, 121, 181, 128, 170, 130, 223, 79, 167, 117, 71, 121, 72, 121, 210, 122, 88, 82, 138, 101, 194, 101, 6, 128, 16, 129, 145, 134, 148, 134, 154, 134, 128, 152, 252, 87, 14, 93, 58, 94, 145, 99, 199, 109, 9, 115, 102, 117, 194, 141, 221, 142, 238, 145, 144, 154, 145, 154, 202, 104, 203, 104, 38, 116, 42, 116, 1, 132, 21, 132, 244, 134, 45, 97, 129, 120, 149, 120, 122, 121, 197, 68, 28, 149, 14, 152, 215, 101, 184, 124, 165, 125, 166, 125, 168, 125, 235, 68, 29, 135, 30, 135, 63, 155, 74, 159, 130, 116, 17, 142, 165, 121, 114, 133, 33, 147, 175, 156, 224, 97, 221, 111, 105, 216, 198, 221, 177, 106, 192, 106, 82, 216, 140, 222, 205, 129, 196, 133, 14, 154, 15, 154, 205, 156, 7, 76, 213, 155, 120, 157, 128, 157, 146, 158, 131, 126, 105, 130, 4, 134, 16, 136, 16, 155, 45, 156, 130, 115, 161, 158, 60, 0, 42, 0, 94, 78, 1, 79, 122, 92, 130, 92, 148, 144, 47, 84, 71, 84, 94, 103, 152, 115, 192, 118, 145, 130, 24, 85, 72, 140, 119, 141, 83, 85, 84, 85, 77, 90, 95, 85, 238, 126, 117, 102, 232, 104, 14, 65, 174, 125, 186, 125, 236, 138, 217, 149, 60, 0, 42, 0, 20, 108, 171, 139, 212, 95, 23, 108, 84, 108, 3, 95, 125, 108, 245, 119, 196, 143, 110, 84, 227, 108, 129, 112, 245, 118, 158, 130, 160, 84, 81, 89, 12, 120, 20, 104, 43, 107, 35, 108, 22, 138, 45, 85, 33, 89, 196, 104, 70, 110, 71, 110, 155, 120, 122, 132, 86, 100, 163, 102, 182, 120, 80, 86, 135, 97, 8, 117, 94, 216, 88, 220, 104, 86, 169, 97, 220, 120, 231, 120, 233, 120, 141, 70, 74, 127, 255, 135, 28, 159, 60, 0, 42, 0, 60, 127, 15, 110, 213, 125, 4, 137, 167, 135, 47, 124, 49, 124, 79, 124, 60, 0, 42, 0, 144, 99, 92, 132, 60, 0, 42, 0, 228, 98, 60, 0, 42, 0, 210, 141, 96, 145, 60, 0, 42, 0, 54, 87, 190, 81, 228, 52, 34, 94, 112, 96, 61, 109, 142, 107, 72, 120, 24, 97, 194, 154, 60, 0, 42, 0, 144, 151, 60, 0, 42, 0, 127, 118, 71, 216, 144, 221, 167, 73, 122, 128, 60, 0, 42, 0, 20, 78, 60, 0, 42, 0, 7, 82, 190, 89, 47, 96, 39, 83, 131, 122, 196, 144, 67, 216, 181, 220, 143, 96, 8, 99, 47, 109, 50, 57, 236, 96, 193, 109, 33, 123, 28, 97, 234, 134, 102, 57, 5, 103, 167, 123, 193, 125, 50, 149, 69, 216, 36, 220, 203, 123, 37, 142, 85, 122, 101, 147, 210, 133, 220, 155, 17, 148, 202, 122, 78, 216, 58, 220, 60, 0, 42, 0, 198, 130, 63, 80, 171, 90, 97, 124, 60, 0, 42, 0, 132, 96, 87, 120, 251, 144, 106, 93, 247, 141, 129, 82, 114, 101, 195, 107, 13, 142, 33, 145, 37, 145, 57, 149, 157, 88, 29, 152, 189, 88, 103, 94, 75, 216, 227, 223, 71, 106, 198, 113, 50, 127, 185, 154, 253, 120, 107, 147, 108, 147, 81, 126, 172, 141, 122, 142, 48, 148, 60, 0, 42, 0, 84, 78, 168, 79, 101, 104, 77, 131, 94, 131, 90, 120, 172, 85, 236, 131, 209, 80, 47, 140, 58, 86, 246, 90, 111, 55, 148, 97, 146, 151, 53, 106, 75, 106, 78, 133, 68, 118, 9, 64, 167, 119, 4, 121, 151, 67, 171, 141, 238, 133, 8, 148, 189, 151, 102, 152, 92, 216, 13, 221, 60, 0, 42, 0, 231, 93, 229, 145, 0, 97, 220, 154, 60, 0, 42, 0, 207, 79, 238, 139, 237, 92, 41, 94, 141, 122, 87, 150, 187, 107, 216, 127, 154, 138, 218, 154, 250, 80, 172, 100, 189, 100, 152, 151, 210, 151, 197, 122, 249, 127, 89, 139, 136, 142, 129, 54, 60, 0, 42, 0, 215, 105, 158, 114, 60, 0, 42, 0, 0, 52, 24, 78, 32, 78, 67, 216, 11, 220, 117, 87, 177, 54, 203, 60, 177, 144, 88, 96, 203, 121, 204, 121, 175, 134, 157, 90, 120, 105, 41, 132, 89, 158, 205, 123, 231, 125, 242, 132, 117, 135, 80, 122, 165, 141, 2, 74, 197, 156, 215, 135, 166, 151, 167, 151, 12, 156, 13, 156, 150, 157, 36, 136, 157, 159, 60, 0, 42, 0, 244, 83, 218, 86, 79, 98, 176, 114, 140, 115, 83, 108, 141, 128, 66, 108, 108, 134, 197, 108, 181, 66, 111, 134, 197, 79, 211, 137, 4, 138, 5, 138, 75, 145, 211, 145, 18, 85, 151, 55, 57, 57, 93, 59, 87, 109, 12, 125, 218, 145, 130, 104, 143, 107, 236, 107, 3, 116, 141, 131, 71, 141, 14, 144, 17, 144, 55, 93, 239, 93, 98, 57, 30, 110, 109, 110, 179, 118, 218, 118, 240, 93, 106, 113, 127, 125, 247, 134, 216, 136, 82, 144, 233, 137, 213, 140, 134, 116, 100, 135, 182, 146, 148, 145, 130, 155, 61, 159, 196, 155, 61, 156, 60, 0, 42, 0, 29, 100, 215, 124, 60, 0, 42, 0, 251, 145, 18, 134, 60, 0, 42, 0, 67, 83, 223, 78, 49, 87, 50, 87, 119, 89, 102, 98, 88, 108, 33, 150, 101, 79, 141, 92, 68, 103, 103, 108, 138, 130, 193, 143, 38, 107, 233, 116, 228, 129, 62, 131, 142, 148, 234, 98, 117, 114, 129, 124, 91, 81, 173, 96, 56, 140, 197, 148, 92, 90, 111, 91, 125, 114, 250, 145, 148, 99, 136, 134, 38, 140, 6, 146, 195, 150, 201, 80, 6, 97, 126, 123, 91, 146, 158, 154, 80, 158, 115, 97, 52, 100, 129, 100, 158, 123, 208, 138, 119, 144, 48, 137, 25, 139, 69, 152, 182, 106, 16, 101, 207, 106, 61, 124, 110, 157, 69, 91, 19, 101, 43, 154, 17, 101, 29, 155, 28, 155, 119, 58, 100, 124, 198, 151, 60, 0, 42, 0, 241, 78, 146, 92, 244, 95, 114, 98, 209, 98, 77, 82, 164, 148, 108, 107, 84, 134, 177, 148, 179, 148, 174, 99, 225, 142, 138, 90, 245, 99, 16, 146, 108, 151, 87, 146, 152, 88, 169, 105, 157, 123, 173, 146, 91, 111, 92, 111, 172, 127, 108, 106, 65, 133, 34, 147, 212, 158, 218, 158, 243, 111, 26, 154, 29, 154, 74, 112, 44, 156, 60, 0, 42, 0, 245, 81, 183, 128, 69, 109, 250, 109, 37, 129, 219, 85, 112, 93, 207, 105, 129, 129, 56, 135, 99, 144, 52, 140, 49, 127, 126, 126, 116, 139, 60, 0, 42, 0, 32, 107, 11, 82, 56, 52, 161, 130, 212, 79, 41, 80, 147, 96, 28, 131, 17, 88, 148, 80, 76, 93, 200, 104, 32, 105, 74, 97, 152, 118, 121, 88, 30, 55, 251, 55, 73, 107, 170, 125, 168, 132, 231, 105, 207, 123, 51, 133, 36, 143, 25, 81, 223, 123, 205, 88, 52, 126, 69, 73, 28, 156, 60, 0, 42, 0, 207, 122, 134, 147, 242, 147, 86, 124, 83, 148, 60, 0, 42, 0, 178, 78, 181, 79, 166, 148, 126, 136, 142, 154, 135, 90, 90, 93, 61, 107, 133, 125, 155, 138, 148, 93, 170, 137, 73, 152, 248, 153, 188, 155, 244, 91, 60, 0, 42, 0, 38, 54, 136, 94, 129, 58, 169, 130, 185, 130, 193, 87, 225, 115, 230, 121, 57, 128, 153, 134, 102, 99, 52, 116, 57, 116, 230, 131, 243, 131, 25, 146, 194, 150, 228, 82, 234, 85, 192, 90, 177, 110, 189, 121, 114, 151, 108, 97, 89, 86, 179, 101, 249, 156, 196, 97, 82, 58, 210, 100, 191, 111, 61, 118, 147, 135, 195, 97, 142, 106, 53, 155, 4, 136, 109, 157, 60, 0, 42, 0, 69, 87, 17, 102, 9, 123, 171, 104, 126, 141, 209, 91, 19, 149, 221, 91, 226, 91, 223, 146, 188, 135, 60, 0, 42, 0, 34, 84, 35, 84, 8, 57, 139, 98, 129, 108, 26, 85, 255, 99, 227, 131, 100, 216, 136, 220, 7, 100, 179, 100, 28, 66, 25, 112, 253, 133, 60, 0, 42, 0, 91, 84, 21, 98, 23, 98, 168, 101, 170, 103, 177, 115, 140, 127, 151, 127, 16, 115, 196, 141, 12, 105, 84, 129, 198, 85, 172, 110, 22, 149, 136, 93, 39, 98, 205, 105, 68, 114, 114, 116, 171, 127, 35, 135, 53, 149, 236, 123, 6, 147, 18, 139, 76, 142, 106, 149, 97, 142, 151, 147, 216, 147, 60, 0, 42, 0, 44, 78, 55, 95, 58, 95, 153, 88, 241, 90, 18, 111, 47, 106, 3, 133, 55, 133, 187, 88, 25, 91, 231, 94, 163, 106, 70, 114, 148, 133, 98, 130, 32, 134, 60, 0, 42, 0, 162, 98, 159, 127, 54, 100, 165, 127, 143, 88, 72, 126, 65, 137, 102, 126, 249, 147, 60, 0, 42, 0, 157, 112, 52, 85, 151, 113, 187, 127, 60, 0, 42, 0, 244, 85, 71, 115, 60, 0, 42, 0, 197, 114, 81, 151, 82, 151, 34, 108, 123, 143, 62, 80, 127, 83, 10, 87, 229, 87, 200, 91, 43, 108, 248, 109, 5, 110, 236, 144, 190, 80, 59, 135, 21, 143, 29, 74, 173, 156, 75, 148, 60, 0, 42, 0, 29, 89, 32, 117, 96, 82, 205, 82, 197, 96, 145, 107, 116, 102, 254, 104, 48, 108, 146, 102, 93, 132, 8, 106, 207, 100, 206, 100, 243, 59, 160, 106, 229, 158, 60, 0, 42, 0, 119, 152, 216, 130, 247, 139, 188, 94, 3, 152, 206, 94, 0, 111, 235, 61, 203, 138, 190, 106, 60, 0, 42, 0, 134, 94, 202, 81, 133, 99, 184, 107, 131, 120, 144, 123, 88, 151, 118, 97, 216, 120, 236, 120, 68, 127, 38, 139, 60, 0, 42, 0, 88, 120, 230, 106, 60, 0, 42, 0, 58, 83, 242, 102, 57, 79, 73, 79, 36, 83, 150, 92, 206, 139, 113, 154, 101, 87, 72, 92, 168, 92, 180, 92, 190, 98, 57, 150, 85, 59, 64, 109, 224, 128, 91, 121, 64, 83, 54, 125, 198, 134, 170, 136, 175, 142, 65, 123, 172, 124, 208, 134, 88, 138, 139, 141, 135, 93, 136, 97, 12, 66, 198, 153, 122, 101, 179, 138, 124, 149, 200, 153, 185, 158, 247, 154, 124, 155, 168, 141, 175, 158, 176, 137, 192, 142, 180, 158, 226, 158, 187, 137, 69, 154, 56, 156, 75, 156, 60, 0, 42, 0, 98, 79, 172, 82, 170, 101, 16, 103, 202, 128, 50, 158, 237, 109, 71, 125, 209, 127, 32, 110, 195, 131, 229, 142, 75, 132, 82, 216, 14, 222, 150, 116, 86, 133, 29, 157, 169, 116, 242, 120, 182, 135, 191, 119, 221, 135, 41, 159, 57, 55, 194, 95, 72, 112, 39, 134, 53, 98, 11, 107, 13, 108, 222, 129, 111, 118, 103, 124, 55, 136, 98, 136, 163, 142, 60, 136, 122, 148, 28, 158, 60, 0, 42, 0, 214, 83, 216, 122, 54, 90, 83, 138, 236, 122, 122, 135, 139, 159, 114, 159, 60, 0, 42, 0, 186, 83, 187, 83, 30, 82, 127, 84, 38, 60, 31, 85, 29, 128, 18, 150, 209, 137, 163, 141, 180, 149, 174, 158, 195, 149, 1, 159, 183, 137, 60, 0, 42, 0, 242, 143, 80, 136, 60, 0, 42, 0, 58, 127, 25, 150, 155, 132, 60, 0, 42, 0, 56, 118, 60, 0, 42, 0, 116, 83, 123, 83, 198, 87, 5, 93, 201, 91, 171, 96, 192, 150, 55, 116, 94, 120, 110, 120, 21, 150, 89, 88, 9, 100, 181, 118, 143, 120, 74, 158, 40, 97, 183, 105, 167, 88, 100, 97, 186, 120, 187, 120, 158, 141, 233, 113, 203, 149, 16, 121, 213, 149, 114, 157, 77, 112, 45, 121, 60, 0, 42, 0, 209, 92, 46, 95, 110, 96, 155, 96, 8, 87, 15, 87, 236, 104, 233, 153, 9, 148, 60, 0, 42, 0, 176, 52, 104, 81, 67, 103, 122, 79, 224, 139, 254, 89, 201, 108, 36, 109, 243, 98, 119, 114, 67, 131, 129, 143, 115, 85, 226, 87, 88, 90, 211, 96, 202, 117, 66, 120, 232, 148, 118, 110, 136, 114, 76, 123, 95, 125, 60, 100, 84, 116, 114, 132, 224, 137, 110, 138, 231, 141, 7, 143, 55, 135, 147, 146, 41, 106, 33, 142, 19, 126, 155, 145, 200, 156, 8, 155, 33, 154, 1, 156, 73, 91, 207, 93, 100, 159, 10, 107, 167, 152, 56, 136, 116, 152, 60, 0, 42, 0, 71, 216, 232, 223, 173, 114, 172, 114, 113, 108, 78, 117, 199, 112, 251, 126, 163, 125, 71, 134, 60, 0, 42, 0, 157, 82, 56, 82, 118, 114, 231, 82, 207, 151, 248, 82, 60, 0, 42, 0, 38, 105, 126, 105, 206, 149, 60, 0, 42, 0, 11, 89, 247, 86, 238, 92, 33, 144, 60, 0, 42, 0, 173, 91, 44, 94, 138, 58, 163, 127, 164, 127, 217, 136, 224, 136, 60, 0, 42, 0, 142, 130, 20, 83, 60, 0, 42, 0, 109, 83, 134, 91, 155, 144, 119, 122, 121, 122, 21, 131, 79, 104, 67, 65, 59, 123, 121, 141, 248, 96, 42, 113, 45, 113, 60, 116, 71, 123, 60, 130, 233, 134, 236, 134, 70, 216, 131, 221, 98, 113, 88, 119, 235, 141, 142, 146, 143, 119, 174, 122, 29, 81, 140, 97, 105, 106, 154, 116, 202, 116, 198, 122, 209, 133, 237, 133, 215, 116, 60, 0, 42, 0, 141, 113, 60, 0, 39, 0, 208, 253, 39, 0, 18, 49, 60, 0, 42, 0, 21, 89, 110, 81, 218, 95, 80, 108, 127, 137, 128, 137, 56, 84, 12, 94, 113, 98, 120, 98, 101, 83, 20, 102, 144, 103, 120, 122, 184, 128, 185, 128, 217, 79, 134, 95, 56, 96, 83, 96, 123, 153, 15, 85, 90, 89, 211, 54, 86, 92, 111, 96, 149, 96, 37, 108, 96, 109, 122, 114, 246, 114, 215, 144, 61, 85, 137, 96, 220, 96, 127, 99, 94, 102, 120, 104, 55, 107, 197, 109, 239, 112, 1, 113, 8, 113, 11, 116, 82, 120, 131, 131, 101, 141, 248, 145, 146, 80, 193, 96, 112, 102, 115, 102, 100, 60, 31, 113, 44, 113, 128, 114, 78, 119, 0, 122, 158, 124, 213, 127, 62, 130, 229, 131, 128, 83, 96, 93, 175, 95, 170, 110, 153, 118, 14, 145, 33, 149, 214, 80, 189, 105, 85, 113, 132, 113, 136, 113, 153, 113, 198, 125, 160, 132, 37, 135, 104, 140, 15, 153, 59, 86, 79, 86, 6, 91, 9, 91, 93, 111, 183, 63, 28, 118, 206, 120, 157, 129, 222, 81, 65, 216, 21, 222, 178, 93, 153, 97, 27, 59, 40, 106, 64, 106, 185, 113, 186, 113, 187, 113, 184, 122, 24, 126, 178, 127, 133, 135, 135, 135, 43, 147, 70, 60, 232, 113, 111, 64, 166, 119, 203, 135, 63, 140, 64, 140, 111, 140, 149, 140, 230, 124, 101, 126, 223, 150, 87, 157, 70, 139, 175, 145, 237, 147, 199, 93, 230, 102, 20, 114, 167, 114, 249, 137, 181, 150, 69, 145, 253, 137, 55, 159, 53, 136, 2, 158, 255, 137, 116, 148, 60, 0, 42, 0, 96, 78, 203, 144, 45, 94, 210, 127, 173, 136, 203, 137, 179, 90, 58, 105, 4, 56, 141, 93, 29, 111, 181, 132, 198, 132, 161, 137, 152, 141, 226, 105, 87, 58, 132, 106, 130, 133, 176, 150, 53, 139, 180, 147, 43, 151, 219, 156, 193, 152, 49, 154, 61, 154, 114, 137, 60, 156, 104, 154, 60, 0, 42, 0, 178, 103, 23, 109, 186, 115, 153, 95, 227, 148, 156, 85, 34, 146, 72, 132, 120, 132, 104, 146, 105, 146, 99, 92, 7, 111, 152, 97, 240, 132, 191, 102, 86, 107, 240, 138, 208, 88, 167, 121, 48, 126, 17, 139, 226, 135, 93, 142, 189, 116, 205, 86, 90, 156, 214, 119, 167, 142, 60, 0, 42, 0, 56, 83, 76, 83, 15, 98, 67, 92, 251, 124, 105, 153, 108, 84, 229, 95, 44, 96, 253, 119, 198, 126, 194, 79, 165, 84, 68, 96, 251, 118, 56, 57, 97, 59, 47, 107, 228, 126, 228, 144, 48, 125, 243, 145, 11, 150, 186, 85, 30, 105, 214, 127, 3, 130, 4, 130, 135, 141, 64, 97, 202, 110, 140, 125, 105, 141, 141, 88, 130, 113, 148, 114, 138, 121, 39, 122, 153, 150, 159, 150, 95, 111, 164, 137, 49, 98, 153, 111, 42, 66, 110, 133, 214, 158, 50, 98, 246, 120, 142, 75, 105, 134, 60, 153, 41, 155, 107, 126, 177, 86, 223, 149, 13, 62, 60, 151, 109, 92, 75, 136, 60, 0, 42, 0, 19, 110, 114, 106, 160, 114, 2, 121, 209, 155, 60, 0, 42, 0, 114, 134, 168, 117, 126, 134, 32, 53, 58, 140, 132, 80, 149, 149, 70, 113, 117, 113, 172, 152, 92, 216, 205, 221, 142, 119, 102, 135, 21, 156, 60, 0, 42, 0, 35, 83, 160, 79, 206, 114, 224, 79, 225, 92, 217, 103, 160, 112, 237, 114, 253, 92, 218, 112, 249, 114, 232, 115, 92, 150, 86, 120, 107, 121, 200, 127, 58, 130, 199, 53, 100, 120, 127, 150, 110, 101, 135, 102, 85, 116, 106, 123, 29, 130, 80, 144, 172, 120, 150, 143, 205, 120, 0, 126, 22, 126, 110, 141, 123, 155, 120, 133, 68, 143, 92, 147, 30, 151, 139, 147, 224, 158, 34, 154, 183, 157, 60, 0, 42, 0, 156, 149, 60, 0, 42, 0, 5, 78, 11, 78, 100, 78, 19, 84, 156, 117, 15, 89, 72, 54, 113, 119, 135, 86, 215, 97, 69, 127, 19, 89, 188, 147, 236, 147, 60, 0, 42, 0, 55, 87, 186, 104, 138, 110, 60, 0, 42, 0, 155, 78, 84, 60, 243, 99, 50, 115, 84, 105, 71, 107, 78, 135, 13, 136, 60, 0, 42, 0, 166, 82, 79, 83, 234, 101, 170, 144, 84, 83, 193, 128, 165, 87, 74, 96, 249, 98, 31, 99, 74, 89, 235, 92, 62, 99, 5, 129, 7, 129, 122, 136, 85, 80, 156, 101, 104, 59, 16, 140, 191, 53, 211, 127, 203, 85, 54, 97, 58, 100, 78, 116, 138, 125, 91, 216, 195, 220, 102, 54, 129, 113, 142, 129, 240, 82, 183, 100, 243, 125, 44, 127, 98, 135, 139, 151, 33, 152, 213, 100, 231, 138, 242, 113, 102, 58, 247, 100, 181, 151, 28, 101, 136, 126, 109, 137, 151, 139, 164, 159, 60, 0, 42, 0, 153, 81, 169, 81, 235, 91, 219, 133, 60, 0, 42, 0, 51, 79, 122, 112, 120, 83, 196, 108, 251, 108, 193, 126, 55, 127, 41, 109, 167, 112, 68, 121, 14, 90, 81, 92, 83, 92, 74, 216, 63, 223, 94, 80, 112, 80, 104, 83, 162, 95, 176, 104, 242, 112, 14, 113, 50, 125, 181, 78, 159, 90, 95, 92, 43, 110, 81, 216, 200, 221, 188, 121, 79, 125, 108, 125, 34, 140, 193, 80, 110, 88, 141, 105, 173, 105, 61, 61, 103, 92, 172, 102, 228, 125, 9, 137, 103, 86, 176, 93, 232, 94, 200, 97, 165, 111, 108, 115, 207, 124, 93, 216, 53, 223, 240, 151, 238, 113, 162, 133, 164, 133, 59, 137, 29, 139, 130, 144, 244, 153, 9, 112, 162, 151, 35, 112, 21, 114, 79, 66, 114, 126, 249, 135, 15, 136, 88, 159, 91, 159, 101, 159, 66, 159, 158, 142, 60, 0, 42, 0, 11, 129, 17, 89, 60, 0, 42, 0, 113, 112, 114, 112, 122, 84, 173, 103, 190, 79, 211, 84, 181, 103, 129, 154, 238, 84, 175, 91, 181, 91, 168, 94, 136, 109, 225, 126, 83, 134, 46, 158, 75, 90, 159, 104, 7, 113, 7, 115, 13, 144, 218, 117, 223, 117, 93, 120, 99, 120, 153, 122, 39, 132, 0, 149, 241, 99, 131, 125, 219, 127, 55, 132, 11, 86, 16, 86, 74, 107, 71, 111, 171, 123, 3, 142, 53, 86, 71, 95, 162, 97, 75, 58, 98, 115, 183, 146, 4, 151, 174, 129, 109, 133, 30, 157, 88, 122, 24, 124, 194, 135, 207, 135, 72, 155, 53, 157, 163, 86, 43, 124, 195, 133, 240, 135, 31, 112, 249, 106, 81, 216, 115, 220, 199, 154, 187, 86, 194, 86, 40, 136, 77, 154, 208, 154, 202, 107, 72, 134, 60, 0, 42, 0, 40, 109, 53, 59, 5, 123, 36, 93, 198, 109, 36, 138, 233, 144, 189, 107, 74, 123, 181, 138, 60, 0, 42, 0, 140, 46, 141, 46, 15, 92, 83, 102, 129, 102, 113, 123, 127, 123, 155, 118, 201, 102, 224, 123, 162, 118, 15, 139, 60, 0, 42, 0, 93, 91, 150, 128, 185, 82, 178, 84, 242, 79, 72, 101, 33, 104, 141, 109, 17, 123, 120, 85, 154, 80, 105, 101, 219, 53, 104, 138, 40, 86, 159, 138, 47, 86, 87, 107, 189, 113, 169, 151, 133, 101, 134, 101, 103, 69, 60, 0, 42, 0, 119, 96, 231, 110, 60, 0, 42, 0, 17, 79, 226, 79, 187, 84, 165, 94, 220, 60, 238, 79, 203, 112, 204, 112, 158, 127, 25, 129, 41, 129, 58, 158, 249, 129, 133, 140, 144, 153, 157, 146, 228, 154, 7, 106, 249, 154, 128, 147, 66, 157, 197, 147, 72, 153, 67, 156, 205, 152, 60, 0, 42, 0, 236, 130, 60, 0, 42, 0, 61, 103, 135, 125, 235, 110, 212, 124, 60, 0, 42, 0, 192, 121, 171, 92, 192, 92, 219, 115, 227, 126, 7, 116, 150, 136, 8, 149, 197, 85, 180, 110, 147, 116, 14, 137, 15, 137, 185, 146, 145, 135, 77, 126, 97, 126, 229, 147, 253, 147, 69, 159, 60, 0, 42, 0, 180, 155, 60, 0, 42, 0, 62, 53, 217, 78, 218, 78, 115, 92, 72, 81, 126, 89, 164, 126, 97, 79, 250, 95, 25, 108, 116, 103, 200, 121, 174, 103, 70, 121, 124, 124, 238, 130, 215, 115, 128, 99, 182, 131, 46, 138, 230, 148, 77, 216, 25, 221, 202, 80, 249, 141, 112, 145, 40, 149, 21, 86, 155, 146, 156, 156, 239, 151, 16, 91, 184, 97, 185, 102, 65, 147, 159, 133, 60, 137, 241, 151, 174, 155, 110, 142, 166, 153, 76, 77, 239, 94, 21, 101, 142, 126, 177, 157, 154, 142, 150, 126, 115, 137, 123, 156, 60, 0, 42, 0, 45, 79, 242, 149, 182, 89, 38, 95, 36, 141, 184, 84, 38, 99, 216, 128, 52, 90, 12, 85, 99, 85, 57, 90, 113, 90, 142, 109, 67, 125, 55, 130, 191, 134, 84, 136, 121, 59, 235, 117, 221, 134, 145, 149, 146, 149, 71, 158, 204, 90, 88, 136, 156, 146, 250, 90, 251, 90, 170, 97, 143, 100, 150, 111, 9, 117, 52, 122, 184, 138, 226, 140, 215, 59, 106, 61, 197, 113, 244, 138, 49, 143, 142, 145, 71, 118, 78, 118, 175, 119, 214, 133, 10, 60, 37, 121, 84, 69, 121, 158, 153, 158, 18, 141, 243, 157, 244, 157, 252, 157, 60, 0, 42, 0, 188, 81, 221, 114, 229, 57, 62, 102, 4, 93, 96, 59, 232, 107, 205, 112, 3, 115, 172, 134, 105, 150, 123, 141, 122, 150, 69, 123, 31, 92, 32, 92, 31, 100, 227, 141, 14, 59, 146, 121, 145, 146, 178, 123, 174, 93, 107, 115, 170, 150, 110, 115, 76, 147, 249, 113, 211, 133, 85, 152, 112, 94, 7, 101, 246, 106, 99, 139, 26, 134, 6, 70, 106, 73, 129, 115, 197, 151, 111, 152, 102, 112, 60, 0, 42, 0, 35, 79, 191, 83, 152, 92, 65, 216, 14, 223, 176, 115, 191, 126, 253, 129, 203, 130, 158, 84, 237, 89, 170, 91, 12, 119, 80, 150, 239, 84, 183, 87, 10, 90, 40, 90, 244, 92, 135, 58, 128, 109, 101, 150, 91, 102, 254, 115, 77, 120, 167, 131, 119, 150, 133, 153, 77, 119, 100, 125, 16, 127, 161, 127, 46, 115, 175, 124, 168, 127, 122, 129, 6, 135, 223, 73, 233, 80, 244, 80, 69, 216, 22, 220, 171, 125, 162, 138, 138, 100, 218, 125, 231, 146, 178, 97, 23, 59, 76, 106, 35, 126, 14, 147, 33, 153, 207, 88, 79, 140, 60, 73, 178, 158, 23, 112, 212, 129, 123, 115, 238, 124, 56, 159, 60, 0, 42, 0, 242, 80, 74, 126, 102, 148, 60, 0, 42, 0, 195, 95, 161, 89, 251, 95, 155, 143, 164, 144, 21, 102, 122, 103, 35, 107, 152, 112, 175, 130, 250, 118, 253, 79, 98, 59, 222, 96, 34, 138, 10, 146, 12, 149, 176, 101, 70, 107, 222, 94, 197, 146, 28, 91, 170, 133, 168, 153, 107, 148, 171, 153, 60, 0, 42, 0, 148, 103, 38, 55, 81, 137, 20, 148, 60, 0, 42, 0, 8, 79, 60, 0, 42, 0, 48, 52, 41, 79, 223, 86, 32, 150, 94, 91, 68, 59, 82, 216, 242, 220, 225, 79, 208, 142, 42, 129, 69, 136, 43, 138, 46, 113, 97, 113, 184, 153, 11, 130, 86, 152, 193, 145, 60, 0, 42, 0, 196, 95, 122, 86, 60, 0, 42, 0, 97, 78, 151, 130, 248, 118, 153, 153, 162, 83, 76, 85, 247, 144, 194, 94, 88, 110, 3, 127, 9, 145, 10, 145, 37, 122, 89, 132, 21, 145, 177, 123, 215, 125, 183, 129, 140, 133, 68, 137, 192, 95, 167, 154, 152, 158, 0, 107, 214, 116, 118, 149, 114, 148, 100, 154, 60, 0, 42, 0, 109, 79, 232, 116, 230, 139, 160, 94, 25, 104, 101, 121, 116, 125, 212, 127, 115, 138, 237, 141, 60, 0, 42, 0, 171, 78, 175, 78, 205, 84, 119, 153, 81, 102, 232, 152, 243, 96, 132, 146, 9, 153, 158, 156, 207, 102, 157, 155, 1, 136, 215, 155, 255, 151, 87, 153, 95, 153, 118, 156, 60, 0, 42, 0, 17, 84, 224, 89, 247, 93, 131, 134, 121, 152, 230, 115, 66, 88, 63, 127, 86, 136, 97, 140, 5, 152, 43, 132, 207, 80, 232, 82, 151, 146, 145, 93, 97, 106, 192, 149, 174, 86, 211, 135, 80, 137, 12, 148, 76, 156, 60, 0, 42, 0, 127, 105, 92, 156, 60, 0, 42, 0, 31, 102, 182, 87, 141, 154, 250, 96, 41, 115, 75, 113, 70, 116, 101, 129, 245, 134, 234, 137, 181, 123, 194, 123, 143, 155, 208, 102, 242, 137, 95, 147, 2, 154, 168, 118, 249, 155, 60, 0, 42, 0, 17, 82, 76, 136, 98, 95, 162, 144, 128, 79, 73, 150, 139, 87, 16, 109, 201, 144, 152, 148, 25, 90, 101, 131, 88, 150, 78, 120, 207, 148, 3, 146, 206, 110, 118, 146, 146, 146, 222, 146, 60, 0, 42, 0, 24, 58, 114, 119, 146, 145, 228, 100, 60, 0, 42, 0, 116, 81, 79, 103, 211, 89, 120, 94, 39, 96, 22, 80, 71, 131, 94, 90, 187, 96, 172, 109, 149, 131, 200, 125, 8, 130, 57, 91, 214, 129, 60, 0, 42, 0, 216, 84, 196, 136, 3, 139, 60, 0, 42, 0, 1, 84, 12, 98, 244, 101, 158, 117, 241, 118, 40, 107, 229, 128, 123, 152, 71, 102, 15, 138, 124, 152, 87, 134, 90, 134, 30, 140, 173, 90, 65, 94, 223, 99, 81, 110, 81, 216, 224, 223, 91, 134, 199, 136, 8, 152, 72, 105, 162, 122, 10, 152, 24, 86, 0, 151, 83, 86, 159, 88, 3, 91, 3, 126, 81, 135, 70, 155, 84, 107, 102, 133, 221, 138, 67, 139, 123, 126, 80, 148, 73, 154, 26, 155, 86, 155, 60, 0, 42, 0, 198, 79, 144, 95, 163, 132, 60, 0, 42, 0, 184, 139, 116, 84, 193, 89, 233, 139, 148, 81, 41, 104, 221, 115, 102, 80, 49, 138, 138, 102, 97, 138, 48, 122, 200, 124, 38, 145, 145, 145, 232, 118, 60, 0, 42, 0, 237, 101, 53, 79, 143, 94, 127, 108, 144, 79, 121, 83, 52, 96, 128, 108, 167, 130, 217, 83, 100, 96, 43, 102, 43, 109, 15, 70, 191, 87, 48, 107, 136, 107, 197, 112, 236, 115, 214, 82, 77, 101, 88, 101, 215, 82, 252, 112, 234, 126, 237, 126, 87, 145, 163, 85, 251, 88, 127, 90, 2, 103, 134, 110, 110, 125, 57, 138, 73, 97, 102, 113, 201, 140, 210, 105, 53, 111, 74, 111, 226, 118, 129, 119, 210, 125, 95, 128, 196, 132, 138, 146, 93, 115, 56, 122, 214, 125, 99, 155, 178, 119, 199, 133, 218, 133, 140, 126, 110, 156, 60, 0, 42, 0, 83, 128, 154, 125, 255, 132, 60, 0, 42, 0, 74, 82, 182, 117, 225, 62, 134, 132, 116, 151, 254, 53, 165, 143, 170, 143, 155, 133, 190, 151, 60, 0, 42, 0, 116, 122, 136, 101, 116, 78, 102, 91, 164, 92, 195, 92, 246, 108, 19, 131, 52, 158, 149, 136, 5, 142, 198, 88, 120, 91, 168, 93, 169, 111, 226, 113, 133, 61, 247, 137, 228, 150, 253, 157, 60, 0, 42, 0, 234, 150, 213, 156, 72, 156, 60, 0, 42, 0, 64, 136, 112, 53, 55, 84, 121, 87, 216, 114, 86, 104, 17, 140, 144, 141, 20, 139, 37, 112, 60, 0, 42, 0, 164, 129, 48, 106, 93, 130, 76, 143, 60, 0, 42, 0, 5, 84, 105, 143, 13, 102, 163, 91, 50, 95, 210, 142, 139, 104, 22, 140, 167, 85, 71, 88, 151, 90, 3, 97, 11, 97, 206, 99, 132, 102, 74, 113, 68, 116, 49, 132, 50, 132, 123, 119, 210, 132, 7, 81, 174, 123, 7, 126, 231, 127, 86, 135, 215, 146, 193, 97, 164, 121, 224, 138, 252, 138, 127, 133, 121, 147, 253, 153, 254, 127, 9, 136, 206, 119, 252, 133, 16, 134, 94, 139, 60, 0, 42, 0, 132, 115, 185, 115, 35, 54, 195, 117, 172, 96, 203, 101, 1, 116, 1, 135, 217, 90, 41, 111, 182, 102, 135, 116, 98, 64, 136, 106, 191, 116, 248, 97, 60, 0, 42, 0, 186, 84, 69, 102, 220, 112, 9, 144, 53, 53, 120, 144, 72, 152, 99, 118, 108, 118, 60, 0, 42, 0, 48, 96, 235, 108, 230, 57, 33, 102, 171, 112, 218, 126, 41, 119, 201, 148, 4, 116, 52, 119, 82, 136, 168, 136, 50, 110, 98, 125, 101, 105, 102, 105, 73, 146, 185, 120, 25, 133, 95, 149, 153, 151, 180, 152, 60, 126, 79, 126, 199, 147, 130, 139, 25, 141, 60, 0, 42, 0, 26, 156, 60, 0, 42, 0, 67, 87, 203, 82, 217, 87, 4, 113, 219, 82, 100, 88, 143, 113, 168, 122, 242, 82, 18, 133, 243, 82, 232, 153, 206, 88, 111, 115, 171, 133, 219, 102, 251, 113, 208, 129, 176, 133, 196, 119, 129, 126, 13, 134, 230, 88, 186, 145, 60, 0, 42, 0, 251, 91, 245, 94, 236, 101, 111, 154, 74, 103, 225, 93, 67, 117, 226, 139, 203, 92, 66, 96, 53, 109, 84, 109, 3, 125, 18, 104, 106, 104, 229, 107, 227, 115, 64, 131, 104, 131, 113, 80, 4, 55, 11, 92, 170, 95, 215, 99, 101, 216, 176, 221, 98, 138, 180, 153, 198, 105, 67, 111, 159, 156, 90, 86, 76, 61, 111, 111, 41, 145, 51, 101, 51, 106, 214, 113, 149, 116, 243, 135, 79, 156, 88, 156, 101, 112, 60, 0, 42, 0, 66, 83, 173, 139, 175, 139, 40, 79, 91, 108, 197, 143, 154, 79, 250, 93, 135, 95, 229, 114, 137, 107, 10, 138, 19, 138, 25, 138, 255, 143, 10, 144, 94, 89, 253, 93, 190, 107, 4, 122, 59, 97, 208, 140, 92, 144, 64, 86, 96, 111, 72, 133, 85, 157, 11, 114, 104, 152, 66, 148, 60, 0, 42, 0, 145, 86, 60, 0, 42, 0, 246, 81, 68, 81, 43, 52, 71, 81, 8, 83, 187, 139, 247, 95, 121, 108, 197, 84, 95, 96, 54, 109, 247, 128, 248, 128, 41, 138, 126, 138, 239, 140, 60, 0, 42, 0, 196, 150, 138, 113, 60, 0, 42, 0, 61, 113, 60, 0, 42, 0, 199, 139, 56, 113, 87, 138, 16, 89, 123, 101, 75, 216, 204, 223, 60, 0, 39, 0, 208, 253, 39, 0, 19, 49, 60, 0, 42, 0, 75, 78, 47, 101, 110, 83, 65, 108, 49, 84, 245, 93, 101, 108, 103, 87, 157, 103, 220, 108, 229, 119, 199, 126, 162, 128, 157, 130, 0, 104, 211, 121, 214, 121, 209, 128, 221, 128, 1, 80, 183, 117, 87, 121, 234, 121, 2, 129, 124, 136, 187, 150, 148, 104, 108, 121, 37, 105, 248, 129, 32, 98, 24, 100, 25, 122, 149, 125, 176, 105, 148, 121, 24, 135, 182, 153, 196, 59, 247, 156, 50, 157, 68, 157, 84, 126, 5, 159, 53, 134, 60, 0, 42, 0, 103, 98, 132, 79, 183, 89, 244, 118, 234, 89, 36, 80, 60, 80, 64, 128, 222, 145, 244, 87, 247, 87, 71, 216, 28, 222, 212, 109, 76, 128, 173, 140, 13, 105, 150, 107, 134, 114, 119, 125, 1, 137, 214, 141, 237, 85, 225, 116, 131, 121, 68, 146, 140, 88, 109, 100, 189, 153, 2, 91, 121, 97, 16, 111, 47, 142, 52, 106, 177, 129, 40, 81, 54, 126, 119, 128, 217, 135, 96, 142, 196, 142, 145, 142, 60, 0, 42, 0, 2, 89, 98, 107, 234, 83, 167, 82, 232, 101, 64, 87, 65, 87, 11, 94, 122, 98, 102, 108, 154, 108, 184, 126, 47, 150, 62, 96, 167, 98, 79, 216, 181, 220, 183, 130, 171, 84, 73, 96, 7, 99, 179, 103, 20, 109, 11, 120, 73, 121, 117, 143, 187, 117, 219, 65, 25, 125, 121, 136, 253, 109, 40, 138, 190, 141, 249, 142, 32, 73, 249, 158, 111, 145, 226, 133, 103, 137, 60, 0, 42, 0, 243, 129, 36, 150, 215, 95, 238, 95, 123, 98, 150, 130, 120, 140, 97, 216, 104, 220, 54, 82, 148, 83, 129, 87, 25, 94, 28, 94, 187, 108, 153, 112, 40, 141, 217, 92, 162, 94, 164, 94, 3, 99, 227, 103, 9, 104, 55, 109, 244, 129, 227, 143, 197, 144, 33, 90, 143, 95, 73, 216, 122, 222, 26, 99, 74, 102, 78, 104, 254, 114, 81, 121, 233, 121, 61, 141, 126, 143, 95, 150, 127, 78, 107, 80, 64, 216, 183, 222, 157, 95, 1, 58, 183, 99, 189, 104, 86, 105, 24, 115, 100, 117, 212, 117, 242, 121, 247, 121, 146, 122, 41, 125, 208, 127, 159, 136, 160, 136, 215, 137, 218, 148, 55, 158, 130, 80, 68, 216, 118, 223, 59, 93, 88, 95, 122, 102, 222, 110, 227, 117, 237, 134, 253, 142, 152, 154, 216, 91, 204, 94, 49, 100, 205, 110, 26, 122, 107, 123, 110, 127, 241, 141, 10, 143, 39, 149, 201, 150, 134, 88, 70, 216, 57, 220, 239, 110, 76, 111, 144, 117, 186, 67, 253, 136, 159, 137, 140, 138, 141, 146, 95, 94, 132, 97, 111, 100, 171, 113, 58, 122, 251, 125, 45, 68, 163, 129, 239, 137, 234, 140, 44, 142, 213, 146, 243, 100, 216, 101, 4, 112, 228, 153, 25, 157, 149, 82, 229, 97, 219, 106, 73, 122, 69, 66, 178, 135, 9, 54, 49, 55, 235, 97, 242, 100, 4, 141, 205, 106, 198, 116, 246, 137, 239, 155, 41, 121, 81, 140, 45, 154, 58, 154, 71, 154, 147, 142, 217, 157, 85, 148, 82, 140, 60, 0, 42, 0, 234, 81, 231, 79, 148, 95, 34, 139, 60, 0, 42, 0, 78, 98, 18, 84, 175, 98, 83, 89, 19, 99, 228, 103, 251, 103, 243, 84, 103, 80, 179, 85, 248, 99, 35, 110, 66, 105, 132, 82, 99, 100, 182, 118, 29, 106, 240, 137, 187, 118, 71, 139, 68, 159, 71, 159, 60, 0, 42, 0, 45, 103, 52, 117, 248, 149, 187, 134, 225, 148, 96, 113, 80, 114, 152, 149, 154, 123, 43, 128, 88, 147, 87, 139, 60, 0, 42, 0, 143, 83, 195, 98, 242, 130, 40, 119, 31, 120, 41, 100, 138, 156, 157, 156, 55, 142, 147, 155, 186, 155, 98, 69, 60, 0, 42, 0, 77, 78, 121, 112, 200, 139, 164, 84, 222, 103, 5, 104, 184, 112, 177, 91, 196, 117, 177, 134, 80, 138, 62, 100, 160, 110, 168, 105, 94, 216, 10, 222, 5, 151, 161, 145, 60, 0, 42, 0, 7, 135, 251, 85, 236, 90, 110, 144, 60, 0, 42, 0, 135, 83, 152, 98, 125, 107, 250, 119, 19, 120, 119, 124, 116, 134, 242, 84, 209, 87, 89, 59, 141, 124, 96, 85, 138, 96, 98, 102, 99, 102, 111, 59, 169, 136, 132, 143, 134, 85, 240, 134, 95, 138, 253, 65, 42, 140, 178, 153, 122, 100, 18, 143, 212, 120, 25, 143, 184, 146, 153, 143, 196, 135, 158, 86, 43, 139, 58, 139, 191, 155, 77, 143, 129, 139, 139, 139, 60, 0, 42, 0, 5, 128, 125, 78, 107, 85, 23, 149, 135, 121, 109, 141, 54, 137, 117, 137, 60, 0, 42, 0, 217, 143, 216, 103, 89, 109, 219, 109, 25, 144, 28, 106, 106, 111, 23, 133, 103, 158, 197, 135, 38, 72, 211, 157, 60, 0, 42, 0, 64, 119, 87, 132, 60, 0, 42, 0, 90, 99, 139, 101, 142, 101, 88, 100, 184, 105, 29, 69, 75, 159, 103, 216, 115, 220, 60, 0, 42, 0, 133, 91, 161, 106, 60, 0, 42, 0, 132, 122, 89, 146, 60, 0, 42, 0, 58, 80, 38, 120, 181, 80, 232, 91, 53, 118, 60, 0, 42, 0, 8, 89, 130, 124, 60, 0, 42, 0, 75, 79, 138, 148, 177, 89, 246, 93, 219, 98, 45, 102, 196, 118, 215, 145, 65, 85, 74, 146, 203, 153, 188, 122, 99, 147, 189, 118, 60, 0, 42, 0, 42, 114, 126, 98, 188, 108, 117, 116, 3, 66, 60, 0, 42, 0, 236, 83, 70, 81, 207, 139, 155, 103, 151, 87, 164, 112, 227, 114, 117, 141, 10, 123, 129, 128, 208, 101, 249, 104, 110, 67, 84, 138, 103, 113, 105, 127, 7, 66, 135, 128, 136, 128, 153, 141, 204, 102, 243, 113, 161, 155, 194, 106, 190, 119, 132, 127, 60, 0, 42, 0, 165, 46, 43, 114, 64, 127, 60, 0, 42, 0, 222, 93, 31, 130, 204, 139, 156, 79, 104, 84, 50, 109, 234, 139, 208, 112, 216, 115, 128, 143, 159, 95, 4, 58, 171, 99, 205, 109, 234, 119, 238, 144, 60, 158, 140, 85, 165, 124, 82, 141, 49, 144, 8, 143, 223, 68, 130, 146, 217, 140, 22, 143, 12, 151, 233, 118, 5, 139, 67, 157, 6, 154, 120, 139, 60, 0, 42, 0, 175, 89, 116, 143, 248, 142, 60, 0, 42, 0, 152, 128, 26, 94, 155, 117, 109, 102, 247, 131, 109, 119, 146, 123, 222, 155, 60, 0, 42, 0, 163, 126, 55, 79, 106, 84, 146, 84, 153, 91, 201, 126, 145, 81, 174, 84, 60, 102, 2, 125, 196, 128, 177, 118, 110, 131, 78, 145, 93, 102, 153, 124, 115, 52, 75, 138, 100, 132, 118, 138, 3, 117, 253, 80, 186, 118, 206, 153, 99, 86, 80, 216, 242, 223, 16, 126, 164, 154, 64, 124, 85, 124, 82, 124, 95, 154, 60, 0, 42, 0, 191, 112, 242, 153, 60, 0, 42, 0, 190, 108, 225, 107, 195, 101, 52, 104, 152, 124, 197, 134, 230, 152, 201, 96, 64, 138, 136, 141, 121, 138, 154, 149, 53, 140, 97, 86, 166, 93, 17, 151, 8, 108, 10, 108, 157, 133, 133, 144, 187, 119, 111, 158, 220, 101, 107, 139, 88, 153, 227, 156, 89, 154, 89, 155, 99, 156, 7, 158, 60, 0, 42, 0, 157, 139, 60, 0, 42, 0, 169, 101, 208, 152, 85, 92, 207, 118, 45, 93, 172, 101, 45, 56, 43, 105, 22, 116, 10, 55, 12, 100, 222, 118, 131, 93, 132, 93, 144, 105, 173, 152, 248, 90, 134, 145, 79, 106, 106, 64, 62, 143, 98, 216, 192, 222, 94, 216, 6, 223, 72, 216, 122, 221, 245, 158, 60, 0, 42, 0, 96, 83, 84, 79, 24, 98, 8, 104, 95, 104, 217, 122, 97, 80, 253, 126, 231, 104, 91, 110, 218, 131, 38, 98, 187, 125, 152, 93, 26, 143, 48, 98, 101, 134, 102, 134, 177, 137, 79, 143, 103, 139, 102, 216, 71, 221, 79, 154, 56, 134, 60, 0, 42, 0, 30, 141, 48, 54, 136, 148, 166, 79, 72, 109, 205, 115, 206, 115, 215, 128, 158, 140, 152, 53, 42, 94, 21, 104, 98, 104, 30, 119, 31, 119, 39, 120, 221, 145, 117, 80, 109, 104, 111, 121, 89, 145, 202, 91, 195, 90, 71, 216, 200, 222, 56, 100, 159, 101, 104, 105, 73, 115, 116, 132, 73, 144, 65, 146, 85, 151, 155, 105, 157, 107, 103, 116, 4, 117, 170, 120, 142, 121, 150, 132, 193, 132, 103, 111, 155, 121, 180, 123, 60, 106, 181, 111, 251, 129, 49, 147, 67, 143, 124, 147, 189, 133, 72, 124, 117, 156, 60, 0, 42, 0, 49, 52, 202, 139, 174, 98, 149, 103, 43, 95, 35, 102, 120, 143, 82, 92, 91, 117, 185, 117, 21, 119, 62, 125, 68, 128, 151, 136, 58, 138, 235, 142, 124, 125, 28, 127, 214, 136, 180, 74, 57, 122, 215, 153, 37, 126, 18, 155, 240, 158, 60, 0, 42, 0, 51, 87, 188, 126, 53, 150, 61, 117, 178, 79, 11, 99, 182, 66, 41, 158, 47, 99, 21, 103, 26, 104, 22, 125, 99, 150, 57, 119, 72, 141, 86, 145, 213, 99, 102, 88, 113, 116, 171, 138, 209, 140, 244, 71, 118, 101, 71, 149, 7, 151, 6, 157, 173, 147, 174, 147, 60, 0, 42, 0, 25, 132, 244, 146, 60, 0, 42, 0, 32, 95, 53, 95, 224, 122, 189, 80, 135, 88, 220, 90, 112, 95, 73, 216, 84, 221, 94, 97, 51, 111, 80, 115, 187, 124, 35, 145, 178, 102, 31, 106, 80, 216, 100, 223, 139, 116, 1, 133, 103, 144, 38, 153, 209, 135, 63, 154, 70, 156, 158, 158, 60, 0, 42, 0, 201, 78, 127, 149, 119, 149, 168, 109, 140, 99, 50, 111, 3, 121, 60, 0, 42, 0, 8, 78, 215, 78, 89, 98, 16, 94, 86, 103, 192, 128, 38, 141, 51, 94, 177, 109, 57, 129, 238, 117, 130, 93, 91, 94, 156, 150, 236, 140, 44, 118, 52, 118, 149, 119, 60, 0, 42, 0, 128, 124, 101, 94, 241, 147, 35, 148, 60, 0, 42, 0, 137, 78, 66, 79, 195, 89, 129, 95, 20, 96, 45, 114, 228, 139, 67, 216, 156, 220, 229, 92, 35, 99, 161, 112, 240, 114, 221, 112, 16, 119, 178, 148, 233, 87, 29, 93, 34, 93, 153, 99, 25, 115, 65, 119, 71, 128, 97, 216, 19, 223, 238, 148, 156, 90, 193, 99, 93, 123, 57, 59, 176, 95, 92, 119, 102, 146, 180, 95, 184, 132, 143, 123, 181, 95, 45, 142, 220, 123, 26, 147, 7, 155, 214, 155, 101, 118, 60, 0, 42, 0, 54, 108, 141, 98, 253, 124, 239, 98, 159, 99, 120, 102, 56, 97, 156, 100, 116, 101, 60, 0, 42, 0, 99, 107, 193, 139, 39, 94, 63, 101, 209, 144, 199, 117, 64, 94, 60, 138, 99, 88, 187, 64, 95, 216, 132, 221, 35, 66, 205, 138, 45, 145, 10, 157, 73, 139, 109, 73, 60, 0, 42, 0, 231, 81, 60, 0, 42, 0, 49, 103, 175, 82, 143, 79, 219, 139, 25, 109, 190, 144, 42, 104, 224, 115, 49, 131, 248, 139, 42, 115, 67, 120, 252, 121, 226, 148, 81, 125, 219, 134, 190, 136, 133, 138, 230, 141, 224, 105, 107, 135, 150, 146, 101, 106, 116, 111, 248, 138, 108, 140, 239, 153, 162, 155, 56, 157, 38, 112, 235, 106, 231, 106, 250, 155, 4, 159, 41, 136, 60, 0, 42, 0, 249, 122, 207, 108, 250, 122, 162, 112, 1, 123, 219, 112, 139, 122, 63, 131, 28, 123, 51, 130, 16, 144, 3, 118, 201, 123, 237, 113, 11, 136, 133, 142, 65, 156, 78, 91, 95, 112, 239, 102, 24, 107, 37, 114, 62, 136, 60, 0, 42, 0, 54, 78, 59, 78, 67, 216, 13, 220, 148, 91, 196, 98, 92, 127, 26, 110, 124, 150, 110, 113, 93, 138, 81, 113, 49, 86, 208, 111, 136, 158, 169, 119, 154, 82, 209, 86, 184, 101, 218, 119, 60, 0, 42, 0, 43, 79, 71, 79, 79, 79, 169, 82, 187, 126, 126, 87, 124, 103, 232, 108, 206, 130, 46, 141, 236, 143, 123, 154, 244, 88, 241, 103, 182, 107, 183, 112, 176, 117, 29, 119, 43, 120, 93, 121, 218, 122, 105, 121, 53, 125, 56, 125, 156, 127, 135, 131, 192, 134, 64, 93, 81, 123, 59, 138, 175, 140, 211, 141, 244, 142, 248, 148, 111, 123, 82, 146, 181, 153, 184, 123, 229, 127, 38, 106, 243, 146, 208, 153, 235, 123, 20, 151, 134, 158, 68, 148, 60, 0, 42, 0, 184, 88, 60, 0, 42, 0, 147, 98, 155, 106, 188, 129, 253, 154, 59, 124, 60, 0, 42, 0, 83, 83, 217, 98, 170, 112, 44, 80, 73, 99, 76, 104, 193, 104, 191, 109, 243, 104, 91, 122, 87, 216, 26, 223, 147, 75, 113, 122, 63, 136, 60, 0, 42, 0, 52, 87, 116, 95, 75, 108, 179, 114, 124, 112, 213, 83, 176, 89, 171, 101, 74, 109, 1, 131, 53, 78, 94, 109, 252, 139, 76, 145, 68, 85, 69, 85, 58, 90, 251, 57, 108, 59, 178, 104, 245, 112, 177, 101, 107, 102, 19, 105, 56, 116, 1, 68, 84, 216, 205, 221, 122, 120, 161, 122, 108, 127, 178, 101, 213, 105, 175, 100, 198, 100, 154, 121, 117, 65, 193, 138, 209, 138, 220, 146, 193, 111, 231, 123, 98, 216, 231, 221, 226, 100, 128, 101, 181, 101, 239, 111, 80, 216, 248, 223, 225, 106, 54, 139, 111, 149, 107, 157, 66, 112, 23, 136, 47, 148, 50, 148, 87, 124, 223, 157, 113, 124, 60, 0, 42, 0, 133, 82, 60, 0, 42, 0, 167, 122, 60, 0, 42, 0, 253, 98, 60, 0, 42, 0, 233, 141, 60, 0, 42, 0, 185, 150, 253, 143, 147, 154, 37, 149, 16, 147, 5, 154, 123, 157, 60, 0, 42, 0, 157, 108, 60, 0, 42, 0, 96, 87, 88, 104, 13, 123, 55, 90, 0, 127, 244, 96, 18, 127, 119, 117, 0, 117, 126, 120, 180, 125, 135, 129, 88, 141, 156, 88, 200, 138, 138, 145, 11, 126, 35, 147, 31, 153, 8, 121, 5, 141, 117, 139, 91, 143, 70, 148, 60, 0, 42, 0, 19, 78, 192, 83, 2, 92, 22, 120, 8, 92, 124, 88, 229, 90, 31, 145, 124, 116, 215, 120, 158, 129, 155, 152, 14, 117, 218, 120, 239, 138, 228, 135, 83, 152, 68, 156, 60, 0, 42, 0, 108, 143, 104, 91, 226, 142, 241, 122, 113, 72, 73, 143, 60, 0, 42, 0, 119, 112, 109, 85, 31, 88, 81, 116, 94, 129, 131, 132, 206, 80, 90, 141, 176, 100, 198, 123, 148, 153, 249, 123, 250, 140, 72, 137, 84, 139, 76, 153, 192, 86, 81, 124, 60, 0, 42, 0, 146, 91, 171, 128, 205, 143, 128, 122, 6, 140, 196, 138, 96, 136, 60, 0, 42, 0, 198, 81, 251, 87, 150, 110, 167, 125, 60, 0, 42, 0, 48, 138, 21, 122, 60, 0, 42, 0, 214, 81, 60, 0, 42, 0, 134, 89, 132, 94, 157, 89, 36, 90, 105, 104, 88, 131, 137, 104, 138, 131, 119, 110, 167, 124, 197, 136, 221, 136, 1, 106, 218, 124, 60, 0, 42, 0, 238, 88, 239, 88, 182, 114, 192, 114, 245, 88, 11, 113, 52, 111, 158, 100, 7, 98, 60, 0, 42, 0, 146, 94, 60, 0, 42, 0, 45, 78, 0, 79, 119, 108, 35, 82, 144, 89, 120, 95, 224, 95, 200, 108, 130, 112, 200, 126, 90, 216, 21, 223, 202, 103, 197, 118, 159, 148, 215, 65, 47, 130, 115, 136, 119, 136, 66, 125, 33, 146, 82, 94, 58, 149, 32, 133, 191, 146, 164, 135, 189, 135, 126, 147, 40, 159, 113, 142, 24, 148, 100, 216, 94, 221, 102, 124, 60, 0, 42, 0, 191, 128, 205, 121, 162, 81, 160, 85, 48, 92, 90, 88, 92, 88, 113, 107, 68, 113, 107, 129, 7, 118, 46, 122, 53, 142, 92, 122, 60, 0, 42, 0, 242, 78, 23, 79, 149, 89, 198, 114, 76, 121, 205, 145, 61, 131, 155, 134, 118, 136, 69, 80, 62, 119, 57, 88, 145, 90, 87, 123, 70, 136, 229, 138, 60, 0, 42, 0, 218, 143, 60, 0, 39, 0, 208, 253, 39, 0, 20, 49, 60, 0, 42, 0, 3, 84, 153, 79, 231, 84, 104, 95, 245, 128, 169, 134, 49, 158, 53, 119, 30, 123, 171, 85, 251, 116, 53, 138, 228, 85, 184, 90, 244, 117, 122, 125, 91, 100, 68, 86, 186, 138, 157, 119, 31, 157, 173, 135, 97, 118, 93, 159, 81, 155, 114, 95, 208, 158, 60, 0, 42, 0, 27, 95, 96, 108, 112, 154, 123, 87, 187, 92, 223, 143, 1, 99, 254, 122, 12, 131, 109, 107, 78, 131, 179, 134, 127, 141, 66, 123, 190, 140, 141, 141, 69, 144, 179, 153, 142, 123, 95, 144, 128, 88, 38, 111, 31, 142, 234, 123, 114, 144, 24, 139, 98, 216, 50, 222, 60, 0, 42, 0, 58, 92, 250, 83, 78, 84, 136, 79, 118, 83, 127, 159, 145, 87, 227, 128, 101, 96, 180, 61, 142, 124, 188, 66, 59, 128, 135, 134, 60, 107, 111, 107, 178, 136, 179, 136, 237, 136, 121, 146, 82, 159, 43, 137, 60, 0, 42, 0, 115, 95, 241, 83, 165, 101, 88, 103, 123, 112, 100, 141, 108, 153, 182, 98, 197, 82, 92, 96, 189, 112, 209, 82, 196, 127, 197, 127, 85, 101, 254, 112, 211, 117, 123, 85, 97, 57, 65, 110, 115, 120, 237, 152, 186, 80, 248, 117, 95, 129, 238, 141, 83, 146, 244, 150, 8, 118, 228, 127, 144, 146, 87, 97, 143, 97, 27, 118, 232, 127, 107, 144, 190, 113, 216, 97, 169, 141, 32, 72, 78, 153, 146, 157, 216, 157, 60, 0, 42, 0, 155, 89, 182, 158, 60, 0, 42, 0, 201, 83, 96, 98, 72, 103, 81, 216, 47, 223, 158, 128, 255, 129, 63, 99, 91, 80, 210, 99, 247, 99, 135, 153, 207, 85, 159, 146, 56, 149, 128, 117, 86, 130, 100, 147, 55, 153, 60, 0, 42, 0, 197, 121, 158, 87, 229, 103, 44, 131, 54, 131, 86, 93, 57, 115, 107, 151, 61, 100, 103, 138, 223, 91, 206, 105, 180, 120, 87, 216, 215, 222, 171, 106, 101, 216, 212, 223, 60, 0, 42, 0, 105, 136, 69, 142, 114, 149, 84, 148, 60, 0, 42, 0, 124, 89, 74, 108, 148, 92, 152, 79, 231, 139, 249, 89, 1, 125, 238, 93, 213, 71, 107, 138, 60, 0, 42, 0, 102, 143, 33, 79, 202, 142, 229, 79, 23, 120, 19, 85, 151, 131, 104, 120, 252, 134, 60, 0, 42, 0, 111, 98, 86, 80, 163, 71, 166, 100, 60, 0, 42, 0, 110, 92, 123, 95, 124, 87, 224, 143, 52, 57, 226, 112, 237, 63, 45, 53, 69, 128, 163, 99, 105, 120, 25, 152, 185, 95, 164, 100, 136, 111, 26, 59, 246, 82, 174, 119, 32, 71, 33, 114, 60, 0, 42, 0, 134, 130, 198, 98, 151, 148, 245, 145, 60, 0, 42, 0, 170, 79, 244, 103, 122, 140, 97, 121, 141, 85, 21, 81, 92, 159, 60, 0, 42, 0, 29, 131, 60, 0, 42, 0, 127, 134, 131, 136, 13, 138, 37, 118, 6, 136, 198, 86, 60, 0, 42, 0, 132, 98, 72, 216, 126, 220, 40, 95, 10, 96, 41, 107, 158, 148, 44, 138, 47, 113, 81, 216, 116, 222, 133, 141, 20, 146, 230, 82, 255, 74, 60, 0, 42, 0, 74, 114, 65, 102, 226, 93, 227, 93, 29, 103, 12, 159, 5, 111, 27, 145, 50, 86, 20, 106, 110, 111, 178, 122, 122, 127, 72, 143, 2, 159, 63, 139, 60, 0, 42, 0, 53, 84, 146, 112, 39, 119, 35, 113, 124, 113, 168, 158, 208, 93, 60, 0, 42, 0, 230, 78, 239, 78, 22, 128, 216, 137, 60, 0, 42, 0, 189, 98, 100, 90, 10, 100, 51, 118, 216, 123, 168, 114, 171, 114, 60, 0, 42, 0, 199, 78, 30, 96, 230, 79, 49, 94, 38, 104, 198, 96, 44, 125, 248, 126, 6, 105, 116, 117, 82, 125, 215, 131, 1, 97, 151, 118, 32, 122, 121, 123, 103, 145, 108, 145, 162, 125, 239, 136, 12, 142, 20, 81, 212, 150, 139, 86, 38, 91, 108, 94, 228, 97, 253, 113, 181, 133, 224, 150, 135, 117, 76, 124, 138, 142, 187, 145, 142, 139, 144, 139, 60, 0, 42, 0, 17, 78, 18, 78, 28, 84, 123, 103, 125, 103, 180, 79, 98, 80, 133, 119, 156, 145, 193, 119, 87, 155, 60, 0, 42, 0, 237, 129, 240, 129, 160, 107, 90, 144, 60, 0, 42, 0, 191, 143, 199, 137, 0, 100, 180, 104, 152, 137, 70, 146, 231, 136, 211, 146, 104, 94, 92, 137, 25, 101, 60, 0, 42, 0, 117, 90, 23, 140, 206, 104, 121, 110, 139, 153, 88, 113, 133, 121, 32, 127, 221, 80, 81, 115, 73, 135, 151, 138, 3, 81, 11, 91, 219, 94, 121, 111, 122, 111, 254, 125, 182, 111, 219, 120, 203, 146, 218, 107, 170, 121, 97, 149, 13, 112, 236, 135, 61, 145, 51, 81, 150, 82, 254, 135, 181, 86, 201, 93, 58, 112, 65, 145, 3, 107, 143, 126, 146, 126, 148, 142, 117, 149, 108, 130, 146, 139, 113, 148, 94, 153, 60, 0, 42, 0, 167, 78, 44, 82, 245, 101, 51, 78, 186, 101, 80, 109, 87, 82, 4, 140, 116, 85, 34, 117, 35, 117, 242, 148, 16, 150, 119, 82, 135, 132, 124, 93, 76, 100, 251, 110, 61, 86, 93, 94, 194, 138, 179, 149, 163, 154, 192, 113, 70, 133, 5, 124, 129, 81, 95, 126, 66, 139, 180, 143, 223, 147, 225, 149, 197, 86, 91, 112, 135, 139, 60, 0, 42, 0, 207, 95, 95, 120, 4, 59, 114, 100, 244, 97, 164, 152, 250, 97, 188, 127, 194, 151, 107, 152, 60, 0, 42, 0, 229, 88, 60, 0, 42, 0, 187, 98, 117, 99, 244, 144, 27, 116, 212, 85, 157, 125, 139, 119, 195, 138, 221, 140, 29, 126, 19, 139, 60, 0, 42, 0, 24, 92, 227, 129, 241, 95, 136, 108, 137, 108, 176, 143, 72, 150, 231, 143, 184, 91, 30, 131, 80, 101, 104, 102, 128, 131, 144, 131, 38, 138, 12, 140, 217, 142, 115, 150, 22, 97, 232, 99, 2, 146, 65, 113, 117, 88, 14, 118, 4, 106, 47, 133, 3, 151, 246, 138, 180, 135, 188, 133, 142, 158, 223, 102, 208, 157, 60, 0, 42, 0, 187, 141, 118, 120, 156, 120, 139, 88, 38, 89, 227, 120, 56, 142, 86, 147, 2, 141, 166, 145, 60, 0, 42, 0, 162, 117, 108, 136, 128, 159, 129, 141, 130, 141, 135, 105, 83, 159, 44, 81, 84, 159, 45, 81, 171, 86, 54, 140, 236, 106, 111, 137, 150, 139, 60, 0, 42, 0, 229, 112, 60, 0, 42, 0, 37, 79, 12, 102, 0, 80, 60, 90, 208, 109, 22, 115, 10, 150, 127, 102, 41, 116, 214, 131, 32, 149, 238, 136, 41, 147, 182, 149, 179, 156, 150, 75, 231, 155, 104, 216, 135, 220, 26, 159, 60, 0, 42, 0, 231, 78, 79, 81, 160, 128, 204, 130, 120, 149, 29, 92, 127, 80, 56, 94, 156, 95, 250, 116, 7, 132, 30, 117, 120, 129, 23, 86, 114, 88, 230, 90, 122, 116, 147, 129, 255, 146, 31, 81, 144, 86, 191, 156, 219, 147, 104, 156, 60, 0, 42, 0, 130, 83, 58, 87, 54, 102, 221, 96, 52, 88, 94, 101, 216, 80, 176, 83, 224, 94, 5, 108, 249, 146, 60, 0, 42, 0, 5, 96, 154, 115, 69, 117, 33, 80, 47, 155, 49, 85, 181, 96, 59, 113, 82, 116, 162, 102, 124, 117, 175, 138, 212, 151, 60, 0, 42, 0, 25, 105, 208, 135, 60, 0, 42, 0, 223, 108, 55, 150, 253, 103, 47, 114, 126, 109, 240, 121, 65, 80, 207, 134, 245, 54, 230, 104, 94, 110, 90, 114, 36, 116, 106, 141, 220, 80, 134, 97, 90, 100, 49, 122, 87, 151, 144, 100, 145, 100, 253, 125, 85, 106, 160, 119, 108, 141, 51, 152, 137, 106, 192, 122, 106, 122, 246, 135, 243, 147, 255, 147, 83, 153, 60, 0, 42, 0, 30, 78, 16, 98, 62, 103, 72, 84, 127, 98, 168, 103, 218, 139, 87, 78, 206, 87, 13, 90, 248, 92, 6, 109, 213, 144, 88, 78, 213, 87, 172, 91, 48, 99, 95, 102, 249, 115, 0, 129, 129, 99, 245, 115, 154, 122, 45, 129, 127, 131, 214, 148, 24, 88, 233, 96, 214, 104, 9, 105, 11, 122, 108, 123, 126, 125, 77, 88, 86, 88, 151, 110, 128, 120, 248, 65, 206, 136, 160, 138, 123, 117, 114, 145, 238, 146, 149, 97, 130, 111, 132, 111, 89, 106, 153, 106, 19, 112, 242, 97, 96, 216, 93, 221, 44, 154, 60, 0, 42, 0, 177, 79, 142, 95, 156, 96, 139, 154, 177, 94, 30, 144, 72, 119, 1, 154, 60, 0, 42, 0, 228, 121, 60, 0, 42, 0, 206, 155, 60, 0, 42, 0, 250, 81, 128, 92, 29, 82, 116, 100, 23, 106, 153, 140, 99, 159, 60, 0, 42, 0, 13, 82, 187, 130, 100, 150, 168, 83, 96, 140, 4, 149, 176, 90, 193, 110, 33, 128, 13, 135, 142, 141, 79, 146, 207, 150, 81, 53, 147, 114, 162, 132, 173, 132, 97, 56, 218, 94, 79, 133, 228, 146, 113, 106, 232, 123, 110, 94, 201, 106, 219, 150, 229, 106, 248, 133, 112, 142, 135, 142, 181, 157, 149, 142, 60, 0, 42, 0, 230, 81, 117, 103, 64, 120, 168, 80, 24, 105, 110, 105, 90, 105, 26, 137, 203, 111, 50, 81, 154, 106, 14, 121, 109, 159, 0, 158, 124, 159, 60, 0, 42, 0, 141, 78, 4, 89, 204, 122, 53, 96, 192, 98, 204, 126, 86, 140, 247, 103, 42, 107, 208, 122, 246, 79, 74, 101, 92, 117, 71, 54, 241, 87, 255, 115, 64, 125, 85, 134, 151, 80, 33, 116, 16, 100, 192, 110, 230, 137, 16, 145, 0, 142, 166, 149, 10, 81, 60, 86, 235, 132, 212, 138, 183, 97, 182, 101, 92, 107, 197, 129, 220, 158, 248, 137, 215, 119, 60, 0, 42, 0, 139, 105, 123, 106, 180, 116, 245, 135, 60, 0, 42, 0, 59, 107, 88, 107, 60, 0, 42, 0, 52, 144, 20, 142, 51, 98, 60, 0, 42, 0, 204, 46, 205, 46, 182, 143, 181, 143, 139, 60, 21, 90, 22, 90, 217, 96, 176, 109, 240, 126, 124, 90, 79, 129, 141, 143, 206, 68, 107, 145, 189, 125, 160, 141, 31, 143, 138, 159, 201, 100, 237, 120, 91, 126, 96, 107, 189, 86, 106, 159, 97, 148, 60, 0, 42, 0, 227, 99, 11, 100, 60, 0, 42, 0, 151, 129, 60, 0, 42, 0, 92, 85, 44, 86, 170, 129, 57, 142, 60, 0, 42, 0, 57, 84, 138, 112, 60, 0, 42, 0, 130, 87, 21, 80, 192, 87, 118, 99, 240, 104, 14, 105, 68, 129, 114, 150, 37, 100, 217, 131, 36, 149, 204, 105, 160, 123, 24, 147, 64, 152, 154, 147, 60, 0, 42, 0, 161, 159, 60, 0, 42, 0, 219, 93, 221, 93, 26, 108, 127, 122, 118, 82, 45, 115, 79, 116, 60, 0, 42, 0, 29, 79, 32, 79, 33, 130, 41, 130, 57, 130, 12, 87, 179, 80, 61, 105, 68, 144, 183, 102, 197, 123, 50, 143, 60, 0, 42, 0, 27, 130, 72, 131, 152, 85, 66, 107, 226, 80, 51, 142, 60, 0, 42, 0, 76, 108, 50, 78, 148, 115, 143, 148, 231, 145, 215, 140, 168, 157, 60, 0, 42, 0, 254, 101, 118, 103, 37, 102, 62, 88, 139, 90, 5, 132, 153, 102, 63, 105, 67, 116, 186, 123, 125, 135, 65, 106, 52, 143, 165, 129, 196, 106, 6, 156, 158, 157, 104, 216, 185, 220, 60, 0, 42, 0, 175, 126, 7, 85, 113, 109, 20, 125, 89, 150, 243, 109, 35, 129, 188, 131, 123, 110, 137, 114, 227, 110, 81, 158, 24, 111, 147, 132, 244, 132, 135, 145, 149, 145, 30, 147, 217, 155, 137, 157, 60, 0, 42, 0, 70, 80, 247, 96, 54, 132, 118, 119, 240, 140, 34, 136, 60, 0, 42, 0, 5, 82, 174, 117, 147, 122, 151, 122, 78, 114, 80, 100, 85, 114, 33, 118, 187, 122, 60, 0, 42, 0, 138, 94, 64, 114, 71, 86, 98, 94, 60, 0, 42, 0, 239, 149, 184, 80, 100, 100, 226, 120, 214, 149, 60, 0, 42, 0, 27, 82, 6, 96, 49, 82, 79, 82, 89, 82, 212, 81, 117, 82, 52, 97, 60, 0, 42, 0, 69, 81, 178, 81, 225, 95, 150, 108, 58, 131, 122, 109, 235, 115, 192, 127, 2, 130, 3, 86, 79, 100, 184, 95, 131, 97, 167, 97, 93, 136, 185, 65, 127, 127, 95, 130, 86, 142, 60, 0, 42, 0, 107, 134, 7, 93, 8, 93, 128, 150, 223, 125, 105, 135, 8, 137, 242, 135, 30, 114, 60, 0, 42, 0, 160, 91, 235, 87, 65, 216, 165, 221, 245, 91, 60, 0, 42, 0, 243, 148, 240, 99, 131, 146, 60, 0, 39, 0, 208, 253, 39, 0, 21, 49, 60, 0, 42, 0, 56, 92, 49, 89, 8, 94, 94, 84, 113, 134, 215, 139, 36, 158, 77, 92, 189, 101, 73, 109, 238, 114, 191, 144, 43, 94, 65, 125, 246, 145, 100, 110, 127, 110, 31, 146, 174, 110, 188, 110, 69, 115, 121, 132, 105, 138, 71, 146, 72, 146, 97, 116, 146, 132, 205, 132, 242, 156, 104, 135, 254, 156, 94, 216, 121, 222, 186, 156, 213, 111, 89, 216, 136, 220, 55, 137, 102, 147, 104, 216, 20, 220, 244, 155, 36, 156, 179, 157, 121, 137, 60, 0, 42, 0, 65, 83, 99, 153, 243, 119, 109, 78, 187, 143, 246, 101, 198, 139, 67, 216, 52, 220, 158, 91, 159, 91, 249, 101, 252, 89, 213, 92, 254, 98, 187, 112, 128, 134, 223, 152, 224, 152, 216, 87, 66, 102, 79, 121, 179, 131, 212, 91, 92, 110, 82, 88, 161, 110, 72, 144, 80, 146, 230, 91, 175, 105, 148, 132, 85, 135, 21, 66, 165, 156, 43, 159, 88, 139, 45, 159, 35, 156, 60, 0, 42, 0, 242, 83, 226, 119, 104, 78, 85, 140, 127, 79, 203, 89, 118, 154, 88, 81, 169, 91, 78, 92, 54, 123, 66, 146, 219, 153, 60, 0, 42, 0, 235, 88, 15, 108, 59, 121, 23, 78, 22, 78, 213, 78, 2, 94, 172, 46, 58, 121, 75, 83, 15, 95, 213, 95, 174, 46, 138, 78, 60, 79, 211, 83, 58, 98, 139, 78, 141, 79, 191, 82, 105, 84, 249, 103, 213, 139, 112, 153, 159, 81, 164, 91, 64, 96, 67, 96, 237, 98, 47, 102, 48, 102, 190, 103, 255, 103, 2, 119, 198, 137, 51, 141, 59, 104, 210, 112, 14, 119, 33, 119, 16, 130, 124, 143, 2, 144, 200, 148, 150, 137, 73, 140, 29, 144, 200, 145, 158, 90, 60, 93, 17, 95, 165, 95, 211, 99, 37, 140, 176, 140, 202, 145, 226, 82, 220, 85, 18, 95, 87, 119, 110, 123, 226, 137, 102, 138, 254, 142, 48, 146, 67, 146, 254, 152, 19, 130, 147, 138, 125, 146, 109, 89, 105, 144, 45, 73, 180, 146, 25, 153, 108, 86, 21, 91, 168, 111, 223, 138, 225, 138, 29, 153, 171, 135, 26, 139, 126, 144, 45, 124, 203, 145, 107, 137, 60, 0, 42, 0, 102, 79, 205, 122, 25, 83, 117, 93, 129, 105, 118, 113, 210, 123, 150, 155, 66, 124, 24, 156, 60, 0, 42, 0, 64, 103, 153, 108, 177, 126, 119, 78, 57, 82, 78, 82, 2, 120, 38, 85, 0, 115, 134, 124, 23, 125, 108, 104, 186, 107, 238, 107, 58, 62, 142, 131, 233, 148, 231, 117, 112, 120, 94, 113, 223, 136, 157, 105, 39, 106, 49, 133, 102, 155, 168, 156, 202, 155, 203, 155, 169, 147, 60, 0, 42, 0, 187, 80, 13, 81, 60, 0, 42, 0, 61, 80, 60, 85, 81, 85, 101, 85, 57, 94, 166, 83, 162, 85, 16, 132, 200, 94, 60, 60, 67, 107, 220, 127, 9, 66, 145, 123, 227, 127, 175, 149, 14, 151, 60, 0, 42, 0, 76, 126, 60, 0, 42, 0, 98, 89, 30, 115, 74, 141, 108, 117, 114, 117, 11, 143, 210, 140, 214, 140, 168, 106, 60, 0, 42, 0, 12, 130, 88, 79, 117, 134, 199, 134, 229, 134, 95, 216, 115, 221, 60, 0, 42, 0, 13, 130, 104, 99, 55, 75, 60, 0, 42, 0, 141, 83, 190, 139, 62, 121, 153, 83, 4, 92, 137, 109, 187, 109, 9, 110, 45, 138, 102, 141, 61, 95, 81, 97, 66, 100, 68, 100, 224, 110, 116, 97, 117, 100, 14, 133, 89, 107, 216, 151, 7, 154, 2, 136, 19, 71, 254, 97, 29, 101, 68, 112, 157, 158, 7, 107, 60, 0, 42, 0, 14, 130, 60, 0, 42, 0, 19, 61, 91, 123, 126, 145, 233, 123, 1, 124, 27, 124, 195, 145, 60, 0, 42, 0, 122, 126, 60, 0, 42, 0, 82, 102, 32, 59, 183, 149, 236, 102, 60, 0, 42, 0, 48, 95, 78, 99, 231, 112, 162, 104, 166, 131, 60, 113, 13, 122, 211, 101, 114, 123, 68, 130, 248, 134, 14, 143, 210, 113, 48, 66, 181, 152, 254, 154, 185, 155, 60, 0, 42, 0, 250, 82, 141, 130, 214, 103, 191, 115, 213, 130, 240, 122, 246, 151, 60, 0, 42, 0, 17, 92, 96, 216, 24, 222, 32, 69, 60, 0, 42, 0, 173, 82, 114, 83, 205, 126, 181, 144, 232, 84, 11, 90, 57, 125, 145, 136, 68, 119, 164, 125, 114, 111, 60, 0, 42, 0, 113, 133, 60, 0, 42, 0, 206, 83, 54, 101, 60, 0, 42, 0, 75, 98, 136, 91, 168, 87, 150, 153, 79, 130, 60, 0, 42, 0, 255, 91, 215, 83, 233, 114, 125, 81, 46, 85, 136, 99, 173, 109, 246, 126, 233, 117, 253, 88, 172, 125, 0, 89, 38, 118, 245, 70, 120, 115, 201, 147, 60, 0, 42, 0, 76, 98, 99, 115, 60, 0, 42, 0, 113, 92, 97, 95, 150, 144, 32, 82, 42, 82, 73, 103, 205, 89, 215, 89, 159, 130, 144, 148, 245, 103, 230, 114, 202, 115, 34, 130, 235, 130, 107, 136, 207, 87, 193, 117, 213, 142, 59, 99, 24, 123, 32, 129, 218, 141, 124, 82, 39, 100, 7, 86, 83, 94, 125, 113, 120, 111, 152, 111, 134, 106, 63, 126, 187, 129, 180, 127, 182, 127, 197, 155, 17, 60, 60, 0, 42, 0, 94, 216, 62, 222, 60, 0, 42, 0, 234, 149, 85, 150, 131, 149, 93, 150, 142, 58, 113, 102, 84, 113, 82, 119, 140, 113, 162, 137, 60, 0, 42, 0, 170, 139, 85, 108, 204, 56, 157, 117, 97, 82, 71, 98, 21, 138, 120, 141, 158, 99, 228, 145, 147, 80, 132, 85, 143, 146, 159, 154, 208, 80, 160, 88, 161, 88, 108, 111, 85, 216, 49, 221, 46, 127, 47, 145, 23, 91, 197, 100, 63, 106, 90, 107, 179, 129, 73, 76, 240, 120, 6, 139, 97, 141, 32, 65, 85, 126, 238, 135, 71, 73, 250, 135, 113, 139, 13, 141, 37, 148, 77, 153, 56, 154, 221, 156, 87, 112, 83, 156, 84, 156, 60, 0, 42, 0, 56, 87, 99, 103, 138, 149, 126, 101, 60, 0, 42, 0, 51, 117, 126, 92, 95, 98, 56, 79, 171, 142, 129, 79, 123, 84, 189, 89, 118, 124, 197, 126, 220, 139, 250, 89, 219, 103, 32, 108, 197, 115, 124, 122, 120, 124, 32, 90, 247, 92, 33, 117, 18, 119, 55, 120, 82, 101, 241, 109, 51, 125, 152, 131, 95, 81, 253, 104, 55, 138, 96, 132, 209, 136, 117, 138, 67, 216, 221, 222, 39, 117, 225, 132, 24, 133, 202, 113, 234, 153, 185, 156, 209, 102, 147, 133, 98, 157, 245, 155, 58, 156, 60, 0, 42, 0, 192, 78, 26, 117, 94, 121, 32, 76, 60, 0, 42, 0, 30, 95, 165, 144, 161, 91, 228, 119, 194, 84, 231, 119, 183, 91, 2, 140, 9, 140, 118, 90, 22, 110, 32, 138, 233, 91, 215, 138, 35, 152, 107, 155, 203, 102, 37, 152, 171, 119, 56, 91, 11, 112, 2, 60, 190, 137, 133, 139, 60, 0, 42, 0, 190, 128, 186, 79, 26, 102, 194, 128, 129, 109, 24, 119, 23, 110, 36, 129, 115, 121, 78, 129, 60, 97, 78, 97, 57, 105, 6, 118, 103, 127, 3, 135, 4, 135, 242, 110, 224, 146, 46, 118, 60, 0, 42, 0, 20, 88, 138, 105, 48, 156, 60, 0, 42, 0, 36, 79, 135, 107, 70, 85, 222, 137, 183, 80, 146, 88, 111, 97, 243, 110, 33, 111, 164, 107, 181, 113, 15, 133, 170, 135, 244, 137, 42, 139, 58, 155, 60, 0, 42, 0, 167, 87, 68, 98, 76, 102, 79, 141, 222, 140, 24, 141, 92, 148, 60, 0, 42, 0, 4, 78, 10, 78, 25, 92, 26, 92, 102, 96, 241, 126, 212, 125, 157, 151, 60, 0, 42, 0, 233, 78, 243, 136, 60, 0, 42, 0, 71, 83, 31, 117, 79, 84, 240, 88, 152, 101, 41, 150, 7, 102, 233, 108, 204, 114, 13, 104, 133, 107, 114, 114, 196, 115, 252, 130, 97, 60, 94, 150, 25, 123, 121, 150, 102, 110, 58, 113, 37, 117, 78, 146, 114, 128, 42, 159, 127, 157, 60, 0, 42, 0, 243, 126, 180, 97, 105, 126, 93, 139, 60, 0, 42, 0, 1, 119, 26, 119, 87, 80, 59, 110, 60, 0, 42, 0, 35, 87, 220, 128, 96, 102, 112, 82, 219, 118, 105, 82, 221, 82, 185, 140, 74, 93, 30, 116, 86, 128, 173, 88, 186, 105, 66, 133, 248, 140, 60, 0, 42, 0, 212, 122, 251, 102, 115, 106, 60, 0, 42, 0, 102, 78, 179, 107, 23, 92, 146, 98, 190, 126, 80, 52, 212, 83, 120, 103, 162, 103, 221, 89, 78, 150, 15, 80, 16, 80, 248, 102, 138, 107, 19, 125, 147, 99, 179, 104, 209, 109, 2, 113, 215, 142, 142, 117, 143, 117, 18, 130, 253, 131, 3, 145, 69, 100, 249, 107, 128, 125, 147, 143, 254, 141, 8, 142, 30, 106, 121, 116, 44, 133, 56, 143, 126, 106, 155, 155, 4, 101, 53, 81, 104, 157, 60, 0, 42, 0, 235, 121, 76, 90, 112, 91, 78, 141, 126, 88, 159, 113, 185, 116, 22, 141, 60, 0, 42, 0, 33, 159, 94, 92, 145, 102, 205, 158, 143, 102, 114, 127, 0, 135, 32, 159, 123, 111, 217, 102, 255, 62, 165, 133, 175, 133, 89, 118, 247, 133, 97, 137, 108, 92, 105, 137, 131, 148, 60, 0, 42, 0, 46, 103, 47, 103, 13, 98, 95, 103, 173, 108, 184, 79, 193, 81, 176, 84, 55, 96, 17, 104, 214, 122, 240, 143, 85, 96, 82, 99, 87, 131, 182, 94, 187, 94, 73, 125, 83, 136, 12, 92, 131, 150, 112, 101, 234, 122, 103, 129, 129, 132, 203, 136, 101, 146, 133, 88, 49, 111, 68, 111, 120, 101, 141, 111, 78, 140, 57, 106, 214, 111, 48, 147, 64, 66, 227, 147, 144, 157, 106, 134, 60, 0, 42, 0, 45, 112, 236, 124, 52, 136, 106, 156, 112, 156, 60, 0, 42, 0, 55, 82, 48, 85, 60, 0, 42, 0, 13, 128, 60, 0, 42, 0, 156, 138, 60, 0, 42, 0, 244, 139, 254, 84, 170, 138, 172, 138, 60, 0, 42, 0, 129, 89, 193, 112, 20, 103, 196, 148, 54, 107, 85, 120, 223, 119, 32, 100, 202, 105, 169, 120, 180, 132, 97, 115, 190, 123, 88, 216, 151, 220, 153, 147, 13, 114, 96, 148, 60, 0, 42, 0, 112, 136, 84, 100, 96, 216, 46, 221, 60, 0, 42, 0, 41, 117, 60, 0, 42, 0, 5, 94, 37, 94, 77, 216, 26, 222, 192, 135, 91, 83, 60, 0, 42, 0, 1, 140, 61, 129, 176, 138, 60, 0, 42, 0, 52, 108, 60, 0, 42, 0, 40, 94, 151, 109, 154, 109, 113, 121, 5, 122, 14, 122, 97, 119, 222, 136, 147, 119, 60, 0, 42, 0, 53, 108, 58, 108, 150, 149, 60, 0, 42, 0, 233, 149, 244, 98, 130, 149, 19, 104, 60, 0, 42, 0, 174, 109, 104, 129, 60, 0, 42, 0, 46, 84, 60, 0, 42, 0, 122, 152, 28, 130, 6, 152, 83, 106, 154, 119, 99, 133, 172, 119, 10, 155, 60, 0, 42, 0, 204, 83, 28, 151, 217, 150, 64, 91, 166, 154, 71, 91, 59, 154, 6, 107, 53, 121, 222, 157, 116, 158, 118, 66, 109, 130, 102, 154, 24, 158, 60, 0, 42, 0, 61, 114, 125, 88, 97, 97, 58, 111, 9, 106, 20, 126, 238, 69, 60, 0, 42, 0, 64, 112, 60, 0, 42, 0, 239, 147, 60, 0, 39, 0, 208, 253, 39, 0, 22, 49, 60, 0, 42, 0, 156, 46, 229, 101, 114, 154, 248, 86, 240, 145, 36, 146, 185, 153, 60, 0, 42, 0, 249, 96, 60, 0, 42, 0, 237, 112, 177, 113, 60, 0, 42, 0, 6, 90, 118, 153, 97, 104, 91, 131, 8, 91, 72, 106, 88, 133, 83, 137, 82, 153, 60, 0, 42, 0, 112, 98, 162, 150, 254, 100, 60, 0, 42, 0, 213, 126, 118, 144, 94, 126, 60, 0, 42, 0, 185, 83, 184, 121, 212, 103, 131, 90, 201, 99, 24, 110, 99, 113, 72, 116, 19, 68, 71, 132, 197, 124, 90, 135, 66, 142, 46, 143, 82, 147, 163, 151, 199, 116, 37, 154, 7, 156, 148, 157, 60, 0, 42, 0, 136, 124, 122, 105, 214, 151, 60, 0, 42, 0, 188, 46, 137, 128, 141, 91, 108, 129, 60, 0, 42, 0, 101, 84, 176, 128, 166, 134, 123, 136, 135, 136, 186, 134, 161, 136, 54, 113, 229, 154, 43, 86, 35, 55, 239, 154, 195, 113, 78, 126, 51, 69, 60, 0, 42, 0, 132, 81, 137, 81, 204, 89, 211, 103, 195, 115, 210, 130, 163, 90, 106, 106, 60, 0, 42, 0, 133, 132, 60, 0, 42, 0, 186, 78, 187, 78, 193, 78, 236, 88, 200, 95, 50, 103, 206, 95, 194, 121, 162, 130, 19, 146, 92, 155, 139, 146, 62, 67, 64, 157, 60, 0, 42, 0, 205, 95, 32, 104, 35, 104, 79, 131, 249, 121, 117, 131, 239, 104, 20, 122, 60, 0, 42, 0, 3, 82, 4, 82, 164, 139, 222, 78, 237, 78, 177, 139, 251, 78, 123, 92, 131, 92, 52, 56, 104, 98, 171, 126, 182, 46, 88, 216, 76, 223, 138, 89, 82, 103, 99, 114, 180, 126, 149, 128, 107, 143, 231, 151, 106, 153, 217, 89, 77, 121, 9, 125, 29, 125, 125, 136, 18, 138, 212, 142, 149, 104, 62, 62, 224, 142, 238, 65, 77, 125, 77, 129, 181, 136, 109, 151, 113, 151, 204, 151, 234, 152, 90, 132, 141, 138, 1, 153, 60, 0, 42, 0, 155, 125, 181, 142, 60, 0, 42, 0, 99, 122, 52, 81, 247, 82, 60, 112, 125, 115, 24, 134, 228, 116, 179, 121, 112, 122, 116, 66, 159, 142, 36, 155, 60, 0, 42, 0, 204, 88, 183, 86, 228, 88, 24, 101, 25, 114, 149, 126, 60, 0, 42, 0, 169, 139, 249, 97, 114, 139, 147, 139, 60, 0, 42, 0, 84, 98, 60, 0, 42, 0, 205, 78, 66, 216, 168, 223, 65, 59, 184, 143, 61, 121, 177, 62, 126, 150, 60, 0, 42, 0, 191, 130, 60, 0, 42, 0, 130, 89, 154, 144, 158, 79, 36, 94, 71, 104, 57, 131, 247, 148, 42, 110, 78, 123, 189, 136, 152, 132, 163, 146, 97, 135, 18, 81, 96, 133, 17, 157, 133, 86, 44, 91, 122, 91, 225, 111, 61, 157, 216, 102, 248, 113, 183, 133, 21, 136, 102, 137, 165, 152, 185, 145, 108, 152, 108, 156, 60, 0, 42, 0, 93, 108, 151, 128, 115, 78, 177, 143, 15, 145, 233, 100, 60, 0, 42, 0, 101, 81, 51, 109, 213, 85, 183, 90, 189, 110, 27, 127, 208, 132, 31, 126, 37, 137, 60, 0, 42, 0, 86, 98, 65, 103, 188, 143, 240, 156, 191, 93, 60, 0, 42, 0, 60, 99, 60, 0, 42, 0, 124, 99, 60, 0, 42, 0, 210, 83, 229, 130, 49, 95, 76, 80, 3, 110, 43, 113, 0, 145, 73, 105, 187, 132, 172, 123, 219, 123, 7, 114, 25, 156, 47, 156, 184, 157, 60, 0, 42, 0, 118, 93, 60, 0, 42, 0, 81, 90, 117, 104, 36, 117, 204, 125, 100, 133, 16, 75, 60, 0, 42, 0, 100, 106, 74, 133, 75, 133, 96, 126, 2, 134, 3, 134, 60, 0, 42, 0, 109, 108, 152, 103, 174, 130, 50, 65, 139, 134, 16, 149, 94, 116, 127, 119, 57, 135, 179, 146, 237, 146, 225, 83, 98, 216, 232, 221, 171, 58, 225, 88, 60, 0, 42, 0, 39, 88, 139, 100, 214, 88, 60, 0, 42, 0, 46, 150, 10, 103, 111, 143, 14, 128, 68, 80, 223, 142, 134, 90, 76, 116, 157, 120, 219, 125, 244, 68, 45, 143, 192, 116, 29, 121, 60, 0, 42, 0, 164, 119, 60, 0, 42, 0, 240, 149, 166, 109, 143, 149, 160, 149, 100, 111, 77, 106, 182, 129, 60, 0, 42, 0, 56, 131, 60, 0, 42, 0, 14, 98, 156, 128, 4, 104, 232, 114, 210, 126, 185, 91, 92, 59, 231, 107, 25, 131, 99, 131, 255, 112, 70, 72, 88, 93, 239, 65, 104, 125, 162, 127, 182, 90, 198, 90, 100, 93, 8, 100, 17, 100, 182, 110, 209, 56, 149, 105, 174, 105, 181, 105, 148, 113, 98, 116, 201, 132, 65, 122, 126, 135, 85, 149, 25, 126, 141, 135, 142, 135, 35, 137, 229, 153, 246, 154, 43, 91, 184, 93, 3, 114, 148, 147, 28, 112, 198, 93, 231, 102, 17, 136, 104, 216, 254, 220, 60, 0, 42, 0, 151, 81, 130, 91, 72, 87, 74, 216, 232, 222, 135, 80, 245, 142, 4, 108, 60, 0, 42, 0, 39, 157, 60, 0, 42, 0, 67, 122, 60, 0, 39, 0, 208, 253, 39, 0, 23, 49, 60, 0, 42, 0, 114, 78, 92, 91, 121, 81, 168, 84, 213, 89, 255, 89, 10, 131, 37, 104, 134, 115, 14, 125, 50, 131, 64, 141, 68, 141, 196, 109, 246, 121, 1, 127, 24, 140, 222, 85, 115, 91, 107, 93, 20, 105, 125, 110, 203, 110, 162, 124, 142, 143, 118, 91, 88, 132, 220, 137, 62, 138, 178, 140, 199, 140, 145, 141, 17, 145, 49, 149, 140, 121, 53, 122, 199, 125, 45, 146, 67, 149, 135, 159, 28, 143, 172, 111, 238, 138, 166, 141, 58, 143, 25, 147, 237, 154, 187, 156, 18, 159, 127, 147, 161, 147, 190, 116, 63, 152, 62, 152, 212, 155, 133, 157, 77, 159, 105, 216, 203, 221, 38, 156, 60, 0, 42, 0, 251, 132, 60, 0, 42, 0, 64, 216, 148, 220, 212, 78, 7, 84, 77, 103, 173, 62, 201, 89, 202, 89, 247, 119, 196, 121, 112, 84, 125, 124, 20, 128, 207, 128, 120, 134, 237, 121, 147, 104, 43, 123, 232, 145, 89, 85, 43, 125, 211, 110, 63, 138, 159, 105, 60, 0, 42, 0, 87, 91, 234, 129, 147, 130, 216, 57, 51, 80, 90, 82, 99, 96, 120, 114, 33, 131, 13, 110, 37, 119, 38, 119, 212, 128, 254, 128, 44, 111, 116, 60, 60, 0, 42, 0, 80, 91, 48, 93, 116, 106, 60, 0, 42, 0, 0, 94, 29, 83, 158, 108, 130, 84, 202, 143, 246, 98, 37, 125, 46, 125, 84, 146, 115, 155, 220, 129, 226, 129, 60, 0, 42, 0, 66, 103, 56, 120, 122, 80, 146, 85, 244, 151, 209, 150, 75, 216, 141, 221, 187, 93, 252, 120, 77, 137, 220, 150, 203, 86, 208, 86, 229, 150, 60, 0, 42, 0, 139, 84, 60, 0, 42, 0, 25, 82, 158, 98, 162, 108, 233, 98, 206, 108, 253, 108, 35, 141, 71, 82, 238, 143, 93, 131, 54, 85, 103, 85, 59, 94, 46, 123, 52, 130, 172, 140, 173, 110, 224, 119, 22, 86, 231, 90, 88, 94, 166, 123, 13, 106, 206, 138, 92, 141, 199, 100, 164, 111, 159, 118, 148, 119, 0, 124, 11, 121, 46, 139, 254, 140, 12, 136, 87, 137, 90, 159, 106, 69, 112, 159, 5, 158, 60, 0, 42, 0, 40, 89, 196, 78, 130, 94, 68, 108, 3, 102, 23, 102, 81, 99, 49, 93, 60, 0, 42, 0, 44, 79, 54, 133, 60, 0, 42, 0, 125, 112, 126, 112, 62, 117, 201, 84, 61, 104, 214, 112, 61, 110, 209, 131, 117, 119, 243, 140, 60, 0, 42, 0, 176, 91, 61, 93, 193, 67, 60, 0, 42, 0, 141, 81, 40, 87, 87, 98, 162, 79, 5, 109, 125, 143, 164, 80, 9, 143, 104, 145, 14, 81, 33, 126, 60, 0, 42, 0, 60, 141, 29, 98, 202, 140, 151, 156, 253, 155, 8, 136, 2, 156, 97, 156, 60, 0, 42, 0, 174, 80, 109, 144, 223, 124, 103, 142, 169, 145, 60, 0, 42, 0, 255, 81, 127, 148, 60, 0, 42, 0, 233, 101, 163, 103, 164, 134, 215, 104, 161, 111, 170, 116, 187, 133, 112, 126, 251, 133, 60, 0, 42, 0, 118, 112, 129, 118, 130, 118, 21, 85, 35, 85, 141, 104, 32, 144, 191, 85, 65, 130, 101, 97, 106, 86, 229, 113, 9, 124, 95, 139, 174, 141, 129, 142, 195, 122, 200, 122, 60, 0, 42, 0, 6, 104, 60, 0, 42, 0, 185, 144, 122, 154, 249, 139, 240, 144, 108, 150, 247, 104, 248, 104, 198, 131, 18, 145, 131, 123, 197, 125, 207, 138, 176, 156, 57, 145, 235, 155, 192, 158, 54, 154, 113, 159, 122, 159, 60, 0, 42, 0, 113, 141, 112, 141, 60, 0, 42, 0, 79, 89, 205, 99, 113, 105, 117, 61, 60, 0, 42, 0, 208, 155, 60, 0, 42, 0, 66, 81, 204, 124, 42, 124, 46, 124, 21, 148, 31, 148, 60, 0, 42, 0, 177, 84, 60, 0, 42, 0, 29, 102, 175, 108, 90, 104, 193, 91, 235, 60, 221, 99, 70, 86, 141, 100, 39, 81, 5, 101, 18, 101, 57, 81, 34, 101, 177, 141, 56, 121, 178, 141, 60, 0, 42, 0, 130, 102, 6, 53, 171, 102, 219, 140, 94, 141, 62, 147, 253, 111, 84, 142, 60, 145, 201, 116, 10, 141, 232, 147, 210, 116, 58, 55, 82, 112, 131, 139, 71, 145, 218, 116, 182, 121, 120, 137, 154, 139, 97, 153, 60, 0, 42, 0, 14, 96, 60, 0, 42, 0, 46, 140, 86, 139, 91, 139, 60, 0, 42, 0, 206, 86, 60, 0, 42, 0, 40, 83, 66, 114, 152, 127, 67, 141, 205, 140, 231, 129, 216, 140, 53, 133, 19, 141, 210, 154, 28, 141, 60, 0, 42, 0, 117, 154, 212, 153, 60, 0, 42, 0, 56, 54, 88, 89, 9, 95, 15, 129, 95, 88, 108, 132, 186, 146, 211, 129, 223, 129, 60, 0, 42, 0, 151, 88, 158, 88, 142, 97, 47, 127, 43, 145, 103, 106, 183, 113, 148, 116, 240, 119, 243, 120, 126, 127, 82, 126, 68, 139, 60, 0, 42, 0, 3, 149, 229, 146, 96, 141, 17, 117, 8, 141, 60, 0, 42, 0, 91, 156, 60, 0, 42, 0, 223, 121, 69, 132, 169, 132, 60, 0, 42, 0, 70, 83, 179, 141, 82, 83, 235, 84, 18, 93, 42, 93, 207, 101, 182, 80, 164, 123, 36, 142, 63, 142, 94, 149, 195, 147, 60, 0, 42, 0, 197, 139, 196, 126, 59, 150, 206, 79, 60, 114, 199, 115, 86, 121, 68, 125, 91, 138, 123, 151, 3, 69, 60, 0, 42, 0, 186, 147, 60, 0, 42, 0, 40, 102, 232, 121, 125, 99, 139, 131, 10, 105, 34, 116, 19, 122, 112, 123, 60, 146, 60, 0, 42, 0, 64, 216, 135, 220, 230, 93, 80, 79, 40, 85, 83, 126, 60, 0, 42, 0, 92, 79, 80, 87, 52, 54, 157, 92, 158, 92, 13, 96, 60, 150, 179, 79, 217, 128, 17, 85, 167, 94, 90, 121, 90, 80, 48, 57, 110, 59, 143, 136, 67, 132, 68, 132, 245, 152, 0, 54, 243, 124, 60, 0, 42, 0, 151, 84, 217, 132, 60, 0, 42, 0, 156, 83, 24, 103, 112, 59, 250, 85, 54, 106, 213, 135, 151, 126, 60, 0, 42, 0, 138, 93, 52, 86, 181, 93, 127, 86, 187, 116, 60, 0, 42, 0, 44, 104, 74, 125, 84, 145, 0, 103, 108, 102, 125, 121, 33, 122, 106, 127, 160, 143, 220, 105, 123, 145, 137, 145, 94, 133, 247, 146, 10, 147, 135, 106, 140, 106, 60, 0, 42, 0, 160, 103, 93, 122, 60, 0, 42, 0, 187, 148, 96, 216, 86, 222, 156, 142, 125, 148, 60, 0, 42, 0, 100, 126, 53, 127, 130, 126, 137, 126, 107, 124, 152, 126, 60, 0, 42, 0, 37, 101, 90, 148, 60, 0, 42, 0, 10, 92, 171, 88, 255, 88, 159, 93, 61, 106, 117, 144, 92, 126, 71, 127, 15, 148, 223, 156, 82, 156, 247, 157, 60, 0, 42, 0, 212, 80, 66, 86, 153, 100, 63, 69, 80, 139, 60, 0, 42, 0, 88, 99, 140, 146, 60, 0, 42, 0, 142, 157, 60, 0, 42, 0, 151, 91, 39, 80, 252, 126, 148, 154, 43, 88, 79, 93, 85, 93, 254, 96, 213, 104, 35, 115, 89, 129, 33, 103, 54, 105, 124, 132, 120, 93, 83, 64, 47, 122, 156, 125, 195, 125, 167, 113, 38, 66, 245, 125, 234, 127, 108, 135, 40, 142, 42, 142, 235, 120, 80, 147, 117, 140, 100, 142, 12, 154, 3, 155, 35, 154, 9, 155, 55, 155, 238, 155, 252, 155, 65, 148, 60, 0, 42, 0, 59, 96, 108, 80, 116, 99, 227, 96, 33, 97, 212, 99, 3, 100, 175, 80, 211, 61, 96, 100, 207, 125, 215, 132, 2, 126, 61, 126, 211, 147, 60, 0, 42, 0, 181, 126, 46, 102, 173, 117, 10, 80, 20, 115, 130, 120, 189, 124, 201, 124, 50, 118, 38, 126, 29, 147, 49, 126, 237, 124, 60, 0, 42, 0, 72, 111, 60, 0, 39, 0, 208, 253, 39, 0, 24, 49, 60, 0, 42, 0, 114, 84, 181, 117, 124, 141, 104, 80, 128, 141, 208, 141, 102, 216, 40, 222, 180, 154, 18, 126, 0, 136, 202, 154, 121, 159, 60, 0, 42, 0, 205, 139, 57, 54, 144, 87, 204, 103, 193, 115, 50, 88, 96, 121, 8, 131, 40, 131, 53, 57, 247, 116, 208, 70, 94, 138, 157, 143, 158, 143, 72, 97, 6, 117, 193, 120, 204, 150, 15, 74, 90, 158, 205, 124, 164, 143, 250, 152, 8, 153, 2, 54, 40, 91, 232, 111, 139, 133, 28, 157, 32, 121, 173, 143, 191, 157, 192, 157, 60, 0, 42, 0, 100, 107, 76, 79, 218, 108, 188, 115, 137, 118, 42, 125, 134, 155, 60, 0, 42, 0, 63, 103, 33, 107, 58, 79, 125, 79, 58, 82, 62, 82, 155, 94, 40, 104, 38, 131, 88, 125, 191, 131, 211, 134, 80, 141, 220, 140, 134, 135, 94, 58, 60, 0, 42, 0, 147, 86, 230, 100, 3, 101, 60, 0, 42, 0, 36, 121, 60, 0, 42, 0, 106, 144, 195, 86, 60, 0, 42, 0, 138, 81, 140, 81, 167, 79, 149, 83, 123, 96, 250, 98, 75, 109, 71, 101, 95, 117, 116, 80, 160, 83, 39, 123, 163, 124, 193, 94, 251, 96, 44, 110, 86, 123, 23, 132, 94, 123, 116, 123, 52, 132, 132, 88, 50, 58, 163, 123, 219, 132, 161, 97, 87, 216, 33, 220, 14, 124, 60, 0, 42, 0, 114, 80, 71, 90, 28, 115, 60, 0, 42, 0, 77, 98, 178, 114, 80, 103, 34, 141, 161, 140, 193, 136, 168, 110, 148, 126, 60, 0, 42, 0, 221, 107, 199, 145, 56, 80, 75, 85, 192, 91, 105, 95, 161, 99, 108, 119, 244, 141, 181, 125, 41, 142, 60, 0, 42, 0, 240, 87, 204, 104, 220, 131, 33, 133, 41, 126, 60, 0, 42, 0, 161, 100, 205, 100, 217, 124, 60, 0, 42, 0, 250, 102, 249, 102, 8, 86, 134, 93, 21, 111, 253, 105, 248, 132, 85, 216, 98, 221, 90, 130, 172, 135, 63, 137, 234, 147, 60, 0, 42, 0, 120, 130, 73, 131, 58, 97, 198, 97, 50, 154, 60, 0, 42, 0, 143, 128, 53, 145, 89, 137, 60, 0, 42, 0, 190, 46, 121, 130, 60, 0, 42, 0, 209, 81, 74, 110, 96, 129, 143, 143, 51, 143, 60, 0, 42, 0, 194, 83, 195, 83, 196, 83, 225, 152, 150, 154, 197, 83, 176, 85, 76, 110, 170, 80, 32, 91, 16, 153, 66, 154, 60, 0, 42, 0, 139, 107, 149, 134, 237, 96, 152, 107, 90, 97, 69, 135, 89, 97, 221, 69, 49, 91, 54, 136, 58, 136, 60, 0, 42, 0, 232, 96, 1, 103, 88, 97, 175, 97, 71, 122, 248, 123, 234, 158, 242, 158, 60, 0, 42, 0, 127, 112, 209, 54, 186, 99, 113, 91, 178, 124, 123, 100, 175, 111, 230, 113, 168, 116, 146, 133, 50, 139, 60, 0, 42, 0, 15, 81, 24, 114, 60, 0, 42, 0, 126, 93, 60, 0, 42, 0, 145, 92, 148, 109, 18, 123, 163, 104, 60, 0, 42, 0, 211, 78, 250, 78, 39, 79, 167, 108, 205, 130, 39, 158, 9, 80, 49, 130, 150, 80, 98, 93, 196, 110, 74, 115, 188, 132, 89, 130, 165, 135, 172, 157, 60, 0, 42, 0, 207, 133, 118, 148, 60, 0, 42, 0, 110, 65, 246, 140, 60, 0, 42, 0, 248, 111, 73, 127, 12, 107, 60, 0, 42, 0, 253, 102, 76, 86, 60, 0, 42, 0, 66, 92, 254, 102, 100, 92, 146, 93, 29, 59, 242, 122, 85, 66, 83, 154, 60, 0, 42, 0, 109, 142, 60, 0, 42, 0, 151, 124, 213, 137, 129, 158, 132, 158, 164, 158, 60, 0, 42, 0, 130, 95, 130, 107, 60, 0, 42, 0, 195, 79, 29, 115, 40, 129, 98, 145, 4, 118, 142, 138, 151, 141, 72, 86, 177, 97, 31, 133, 39, 142, 139, 145, 47, 118, 7, 124, 44, 126, 89, 142, 68, 73, 0, 159, 116, 142, 117, 142, 99, 152, 60, 0, 42, 0, 19, 100, 115, 116, 174, 100, 203, 120, 115, 144, 73, 142, 157, 145, 60, 0, 42, 0, 107, 59, 88, 134, 228, 117, 73, 119, 236, 119, 111, 93, 115, 93, 171, 132, 22, 133, 126, 158, 66, 145, 122, 158, 166, 142, 60, 0, 42, 0, 30, 129, 60, 0, 42, 0, 73, 82, 82, 82, 157, 83, 14, 89, 43, 99, 170, 99, 157, 131, 161, 131, 174, 101, 228, 104, 42, 144, 9, 149, 25, 149, 204, 132, 117, 107, 188, 146, 47, 147, 60, 0, 42, 0, 20, 93, 172, 80, 215, 81, 30, 127, 148, 88, 32, 55, 137, 93, 91, 97, 103, 100, 177, 105, 85, 115, 239, 105, 234, 120, 23, 126, 217, 147, 60, 0, 42, 0, 60, 111, 128, 116, 161, 141, 160, 118, 60, 0, 42, 0, 28, 79, 240, 95, 169, 117, 5, 80, 139, 124, 35, 125, 198, 127, 3, 129, 6, 129, 80, 85, 91, 85, 180, 96, 236, 109, 243, 107, 32, 113, 58, 129, 3, 132, 1, 118, 185, 124, 183, 125, 224, 127, 172, 129, 181, 129, 226, 111, 193, 122, 71, 152, 206, 129, 74, 137, 60, 0, 42, 0, 124, 78, 60, 0, 42, 0, 70, 108, 186, 100, 209, 146, 105, 149, 127, 142, 27, 101, 165, 142, 121, 148, 60, 0, 42, 0, 213, 106, 209, 93, 17, 107, 115, 122, 60, 0, 42, 0, 156, 122, 169, 107, 182, 113, 225, 123, 18, 124, 196, 122, 40, 114, 60, 0, 42, 0, 81, 103, 168, 144, 180, 118, 6, 142, 138, 111, 244, 122, 60, 0, 42, 0, 88, 91, 159, 79, 245, 98, 60, 0, 42, 0, 12, 82, 214, 95, 60, 0, 42, 0, 248, 91, 11, 84, 127, 124, 60, 0, 42, 0, 6, 83, 234, 86, 241, 86, 233, 95, 158, 103, 193, 130, 49, 96, 164, 96, 69, 216, 223, 223, 199, 104, 39, 113, 82, 216, 218, 221, 113, 132, 23, 111, 97, 128, 162, 154, 176, 102, 5, 106, 44, 106, 156, 113, 125, 116, 129, 116, 235, 125, 102, 128, 106, 128, 239, 132, 37, 133, 155, 119, 234, 113, 245, 123, 112, 128, 204, 135, 111, 147, 113, 126, 230, 147, 24, 154, 68, 154, 60, 0, 42, 0, 206, 78, 27, 78, 147, 95, 67, 90, 110, 91, 150, 95, 158, 95, 176, 96, 217, 109, 46, 116, 82, 97, 14, 111, 64, 111, 104, 111, 180, 138, 232, 140, 233, 140, 55, 106, 226, 83, 194, 133, 71, 112, 9, 107, 28, 114, 60, 0, 42, 0, 129, 97, 37, 139, 60, 0, 42, 0, 16, 131, 60, 0, 39, 0, 208, 253, 39, 0, 25, 49, 60, 0, 42, 0, 182, 83, 159, 126, 29, 78, 248, 83, 168, 62, 249, 124, 193, 121, 157, 84, 192, 108, 236, 79, 29, 96, 82, 134, 54, 158, 164, 90, 175, 101, 114, 125, 12, 127, 243, 134, 82, 105, 50, 61, 112, 146, 212, 152, 216, 81, 174, 83, 185, 105, 151, 121, 115, 127, 36, 135, 54, 149, 54, 86, 93, 86, 70, 216, 82, 220, 221, 94, 149, 100, 140, 111, 195, 120, 160, 121, 86, 216, 232, 223, 230, 125, 214, 146, 205, 113, 108, 133, 132, 135, 46, 69, 214, 135, 244, 135, 184, 152, 38, 154, 1, 148, 100, 216, 112, 221, 229, 157, 54, 159, 109, 124, 60, 0, 42, 0, 123, 107, 60, 0, 42, 0, 243, 93, 150, 78, 219, 86, 250, 91, 92, 108, 64, 79, 161, 79, 85, 81, 210, 89, 96, 91, 107, 103, 215, 108, 228, 108, 64, 121, 114, 153, 119, 154, 82, 59, 246, 103, 109, 114, 48, 90, 13, 109, 152, 109, 130, 128, 228, 152, 37, 123, 28, 128, 242, 145, 226, 122, 151, 137, 227, 85, 134, 128, 132, 140, 54, 146, 59, 146, 252, 152, 223, 153, 169, 121, 41, 81, 124, 133, 3, 112, 60, 0, 42, 0, 86, 96, 175, 146, 60, 0, 42, 0, 232, 78, 49, 99, 50, 99, 146, 100, 60, 0, 42, 0, 18, 109, 47, 138, 120, 151, 117, 111, 81, 112, 160, 142, 60, 0, 42, 0, 69, 83, 231, 108, 210, 152, 14, 129, 18, 146, 40, 132, 75, 100, 175, 152, 186, 153, 169, 133, 210, 106, 68, 134, 60, 0, 42, 0, 66, 216, 191, 223, 161, 150, 60, 0, 42, 0, 170, 149, 60, 0, 42, 0, 114, 130, 19, 109, 28, 104, 169, 109, 108, 85, 239, 148, 110, 107, 23, 116, 237, 150, 199, 85, 95, 116, 112, 107, 171, 146, 129, 111, 206, 97, 204, 100, 199, 111, 207, 111, 55, 118, 81, 122, 128, 111, 177, 116, 18, 112, 97, 122, 108, 126, 86, 143, 252, 147, 69, 139, 203, 152, 60, 0, 42, 0, 11, 110, 111, 122, 60, 0, 42, 0, 226, 107, 34, 97, 204, 99, 94, 88, 248, 107, 110, 129, 123, 86, 195, 156, 75, 152, 4, 76, 19, 156, 60, 0, 42, 0, 34, 66, 60, 0, 42, 0, 238, 85, 91, 141, 255, 80, 253, 140, 58, 124, 60, 0, 42, 0, 37, 86, 60, 0, 42, 0, 187, 99, 154, 154, 69, 97, 20, 100, 158, 110, 43, 127, 69, 126, 202, 129, 203, 156, 18, 154, 55, 154, 32, 156, 98, 156, 60, 0, 42, 0, 107, 98, 131, 99, 194, 90, 69, 69, 60, 0, 42, 0, 253, 87, 25, 118, 9, 108, 194, 119, 222, 154, 60, 0, 42, 0, 166, 135, 60, 0, 42, 0, 92, 99, 192, 94, 138, 153, 214, 85, 203, 94, 28, 100, 73, 100, 178, 110, 64, 115, 11, 145, 87, 100, 75, 216, 183, 220, 144, 132, 195, 132, 60, 149, 213, 152, 45, 66, 88, 130, 139, 135, 153, 145, 170, 147, 63, 153, 188, 152, 190, 152, 42, 154, 60, 0, 42, 0, 220, 83, 223, 83, 129, 80, 254, 85, 141, 119, 222, 100, 174, 133, 251, 100, 226, 106, 234, 133, 84, 124, 60, 0, 42, 0, 253, 85, 132, 129, 54, 118, 60, 0, 42, 0, 9, 78, 14, 95, 193, 83, 245, 107, 0, 66, 255, 107, 153, 114, 22, 155, 60, 0, 42, 0, 208, 78, 30, 79, 152, 80, 193, 124, 194, 124, 147, 153, 221, 124, 227, 124, 228, 124, 86, 126, 210, 147, 254, 147, 48, 151, 74, 153, 60, 0, 42, 0, 213, 79, 52, 94, 183, 96, 99, 101, 144, 149, 60, 0, 42, 0, 237, 88, 246, 107, 129, 83, 117, 106, 60, 0, 42, 0, 238, 104, 46, 105, 238, 105, 66, 137, 60, 0, 42, 0, 82, 104, 81, 104, 60, 0, 42, 0, 211, 85, 33, 100, 201, 120, 44, 137, 161, 152, 159, 147, 89, 152, 60, 0, 42, 0, 39, 78, 170, 85, 60, 0, 42, 0, 225, 105, 60, 0, 42, 0, 231, 80, 143, 68, 25, 155, 60, 0, 42, 0, 207, 130, 38, 117, 101, 145, 35, 122, 163, 122, 76, 122, 239, 106, 7, 134, 19, 134, 204, 86, 60, 0, 42, 0, 215, 79, 60, 0, 42, 0, 138, 115, 25, 89, 201, 139, 221, 108, 131, 128, 44, 109, 145, 109, 223, 115, 32, 125, 191, 91, 128, 104, 144, 107, 155, 124, 164, 131, 31, 144, 149, 154, 131, 80, 159, 124, 52, 138, 33, 140, 201, 85, 80, 88, 81, 88, 202, 90, 43, 97, 175, 110, 184, 110, 133, 128, 84, 158, 243, 80, 4, 53, 44, 97, 161, 105, 134, 129, 235, 137, 154, 141, 97, 144, 159, 97, 14, 106, 21, 106, 90, 111, 101, 111, 82, 216, 130, 220, 191, 120, 12, 133, 108, 144, 201, 146, 23, 153, 24, 59, 90, 106, 155, 116, 36, 126, 12, 124, 22, 139, 92, 142, 215, 133, 76, 154, 80, 156, 235, 157, 60, 0, 42, 0, 194, 155, 60, 0, 42, 0, 6, 85, 214, 54, 17, 90, 107, 104, 173, 104, 143, 131, 158, 80, 67, 119, 205, 85, 167, 127, 77, 100, 41, 127, 209, 132, 150, 141, 17, 124, 11, 69, 20, 124, 46, 126, 255, 154, 187, 155, 60, 0, 42, 0, 64, 98, 122, 78, 189, 58, 34, 85, 34, 125, 16, 116, 226, 96, 1, 149, 233, 85, 75, 216, 98, 220, 145, 110, 155, 102, 99, 116, 133, 116, 40, 137, 155, 69, 136, 147, 141, 147, 150, 147, 187, 147, 193, 147, 60, 0, 42, 0, 36, 144, 185, 110, 54, 135, 60, 0, 42, 0, 17, 116, 230, 85, 60, 0, 42, 0, 10, 89, 149, 130, 125, 134, 32, 80, 248, 84, 125, 109, 45, 119, 125, 131, 126, 131, 214, 110, 98, 119, 143, 125, 112, 132, 163, 113, 201, 111, 150, 151, 214, 150, 60, 0, 42, 0, 229, 126, 139, 150, 64, 144, 143, 150, 168, 150, 205, 116, 60, 0, 42, 0, 184, 129, 73, 75, 33, 112, 196, 154, 211, 154, 60, 0, 42, 0, 151, 78, 129, 92, 21, 120, 95, 121, 7, 140, 227, 87, 87, 93, 71, 216, 213, 222, 114, 107, 115, 107, 107, 113, 95, 119, 142, 120, 66, 144, 66, 122, 182, 138, 229, 140, 24, 91, 187, 111, 167, 150, 133, 106, 150, 106, 231, 113, 178, 116, 87, 122, 64, 126, 131, 144, 173, 121, 95, 122, 80, 126, 222, 101, 120, 126, 90, 137, 98, 139, 6, 148, 41, 148, 226, 151, 60, 0, 42, 0, 251, 114, 224, 117, 120, 145, 60, 0, 42, 0, 52, 83, 60, 0, 42, 0, 88, 121, 7, 123, 109, 123, 151, 123, 156, 132, 60, 0, 42, 0, 89, 91, 242, 114, 107, 91, 106, 131, 231, 152, 14, 100, 59, 115, 241, 152, 194, 105, 192, 132, 117, 133, 158, 133, 60, 0, 42, 0, 95, 99, 11, 123, 188, 150, 77, 123, 13, 100, 171, 105, 176, 123, 40, 124, 168, 147, 189, 157, 60, 0, 42, 0, 234, 95, 126, 103, 128, 103, 0, 90, 215, 103, 47, 80, 199, 81, 39, 93, 186, 94, 165, 104, 222, 109, 216, 131, 105, 93, 121, 120, 25, 135, 189, 97, 205, 111, 100, 216, 237, 220, 167, 106, 118, 147, 6, 155, 60, 0, 42, 0, 2, 96, 154, 96, 56, 128, 230, 122, 177, 80, 47, 97, 38, 58, 100, 105, 119, 93, 107, 97, 115, 128, 247, 153, 60, 0, 42, 0, 188, 139, 139, 91, 245, 139, 1, 144, 130, 152, 31, 138, 12, 152, 166, 138, 56, 153, 60, 0, 42, 0, 169, 103, 185, 147, 60, 0, 39, 0, 208, 253, 39, 0, 26, 49, 60, 0, 42, 0, 117, 84, 63, 150, 21, 149, 60, 0, 42, 0, 196, 85, 60, 0, 42, 0, 74, 85, 60, 0, 39, 0, 208, 253, 39, 0, 27, 49, 60, 0, 42, 0, 148, 85, 98, 86, 60, 0, 42, 0, 230, 84, 60, 0, 42, 0, 125, 123, 60, 0, 39, 0, 208, 253, 39, 0, 28, 49, 60, 0, 42, 0, 184, 89, 191, 89, 182, 148, 63, 90, 64, 90, 89, 92, 254, 117, 60, 0, 42, 0, 185, 139, 42, 84, 238, 86, 215, 143, 196, 79, 37, 90, 232, 92, 233, 92, 144, 109, 244, 115, 170, 131, 27, 138, 146, 118, 75, 119, 9, 65, 11, 146, 7, 149, 69, 158, 254, 134, 144, 138, 192, 120, 31, 152, 157, 152, 100, 155, 178, 150, 77, 152, 93, 157, 94, 157, 76, 139, 68, 216, 69, 221, 42, 156, 60, 0, 42, 0, 153, 103, 8, 120, 11, 152, 65, 86, 0, 154, 60, 0, 42, 0, 132, 83, 117, 92, 57, 98, 122, 107, 67, 84, 139, 92, 124, 98, 40, 150, 93, 84, 202, 130, 109, 143, 56, 150, 162, 84, 185, 84, 67, 216, 101, 220, 169, 87, 246, 89, 201, 92, 16, 120, 14, 83, 118, 96, 40, 120, 101, 68, 133, 134, 127, 153, 84, 80, 126, 83, 10, 88, 170, 96, 160, 99, 101, 117, 101, 117, 70, 120, 20, 140, 219, 142, 15, 150, 193, 53, 46, 88, 63, 93, 225, 96, 21, 97, 66, 110, 72, 71, 95, 140, 246, 142, 2, 145, 42, 146, 197, 94, 36, 100, 57, 100, 153, 59, 39, 116, 109, 129, 60, 132, 123, 138, 76, 144, 79, 144, 235, 80, 65, 135, 55, 149, 87, 158, 30, 152, 154, 152, 19, 153, 101, 155, 105, 86, 90, 133, 168, 137, 228, 138, 188, 149, 41, 153, 150, 140, 84, 147, 196, 156, 94, 107, 78, 152, 24, 121, 238, 106, 16, 156, 154, 157, 141, 139, 67, 159, 105, 148, 118, 159, 119, 156, 60, 0, 42, 0, 220, 100, 72, 157, 60, 0, 39, 0, 208, 253, 39, 0, 30, 49, 60, 0, 42, 0, 192, 84, 206, 84, 9, 85, 195, 87, 45, 90, 40, 99, 56, 107, 243, 85, 190, 110, 176, 146, 63, 149, 111, 86, 132, 147, 60, 0, 42, 0, 64, 85, 113, 99, 145, 118, 176, 110, 10, 86, 113, 101, 115, 101, 154, 118, 76, 118, 3, 154, 60, 0, 42, 0, 208, 107, 57, 102, 62, 90, 238, 119, 60, 133, 183, 142, 237, 111, 45, 151, 249, 133, 68, 151, 60, 0, 42, 0, 12, 79, 126, 130, 55, 54, 49, 114, 57, 120, 75, 120, 210, 53, 204, 85, 103, 88, 210, 90, 27, 97, 141, 120, 152, 150, 198, 83, 167, 102, 119, 116, 161, 149, 254, 80, 108, 65, 33, 91, 211, 97, 210, 88, 221, 97, 214, 102, 166, 116, 134, 133, 50, 153, 49, 157, 167, 118, 185, 119, 61, 69, 164, 153, 25, 121, 106, 139, 122, 139, 64, 148, 73, 151, 60, 0, 42, 0, 107, 156, 60, 0, 39, 0, 208, 253, 39, 0, 31, 49, 60, 0, 42, 0, 246, 139, 146, 138, 60, 0, 39, 0, 208, 253, 39, 0, 32, 49, 60, 0, 42, 0, 249, 81, 109, 53, 234, 103, 142, 104, 234, 142, 10, 114, 60, 0, 42, 0, 86, 101, 171, 83, 247, 85, 248, 85, 133, 93, 210, 94, 246, 110, 83, 115, 82, 115, 158, 150, 110, 100, 76, 216, 129, 222, 172, 113, 136, 116, 28, 133, 104, 144, 221, 120, 241, 127, 113, 128, 175, 135, 250, 127, 55, 139, 56, 139, 204, 156, 214, 147, 50, 156, 212, 157, 7, 159, 60, 0, 42, 0, 157, 98, 215, 98, 186, 130, 132, 136, 122, 149, 170, 90, 188, 90, 86, 137, 60, 0, 42, 0, 45, 54, 153, 92, 119, 98, 115, 87, 135, 87, 176, 92, 97, 89, 101, 89, 178, 80, 103, 89, 156, 154, 28, 55, 239, 90, 96, 97, 75, 216, 198, 220, 186, 88, 180, 93, 202, 97, 217, 100, 179, 111, 169, 150, 202, 147, 65, 154, 60, 0, 42, 0, 246, 127, 60, 0, 39, 0, 208, 253, 39, 0, 33, 49, 60, 0, 42, 0, 180, 139, 164, 108, 39, 107, 180, 107, 239, 116, 37, 158, 120, 88, 26, 111, 80, 107, 198, 107, 176, 113, 14, 157, 12, 117, 68, 216, 191, 220, 51, 139, 217, 106, 194, 147, 215, 157, 114, 73, 60, 0, 42, 0, 146, 129, 117, 159, 60, 0, 42, 0, 65, 216, 118, 222, 24, 84, 85, 84, 118, 80, 98, 129, 20, 86, 150, 52, 38, 128, 69, 133, 213, 133, 60, 0, 42, 0, 4, 96, 106, 97, 60, 0, 42, 0, 242, 133, 60, 0, 39, 0, 208, 253, 39, 0, 34, 49, 60, 0, 42, 0, 137, 91, 146, 79, 214, 92, 73, 104, 40, 108, 181, 94, 2, 64, 25, 140, 149, 90, 244, 131, 247, 117, 100, 129, 59, 132, 74, 132, 76, 158, 157, 138, 237, 132, 140, 151, 141, 151, 230, 118, 243, 138, 163, 153, 235, 118, 106, 157, 253, 151, 149, 157, 60, 0, 42, 0, 181, 115, 125, 85, 248, 150, 17, 81, 60, 0, 42, 0, 181, 87, 250, 79, 53, 85, 239, 87, 245, 148, 222, 99, 140, 150, 111, 127, 168, 146, 60, 0, 42, 0, 180, 114, 184, 92, 9, 99, 29, 109, 72, 104, 250, 128, 76, 131, 123, 140, 19, 88, 105, 90, 65, 65, 123, 102, 151, 102, 12, 147, 199, 149, 159, 155, 151, 75, 239, 158, 60, 0, 39, 0, 208, 253, 39, 0, 35, 49, 60, 0, 42, 0, 64, 89, 105, 96, 69, 216, 239, 223, 126, 113, 189, 132, 60, 0, 42, 0, 206, 92, 60, 0, 42, 0, 65, 100, 83, 75, 60, 0, 39, 0, 208, 253, 39, 0, 36, 49, 60, 0, 42, 0, 174, 128, 175, 154, 60, 0, 42, 0, 108, 83, 135, 92, 2, 102, 59, 102, 162, 68, 127, 59, 60, 0, 42, 0, 138, 103, 206, 118, 160, 145, 60, 0, 39, 0, 208, 253, 39, 0, 37, 49, 60, 0, 42, 0, 165, 151, 60, 0, 39, 0, 208, 253, 39, 0, 38, 49, 60, 0, 42, 0, 63, 81, 12, 128, 80, 81, 149, 79, 82, 81, 207, 92, 15, 109, 81, 150, 45, 104, 249, 128, 75, 131, 50, 85, 56, 158, 171, 124, 79, 128, 187, 136, 0, 143, 233, 66, 149, 156, 245, 154, 173, 150, 158, 155, 47, 157, 92, 143, 60, 0, 42, 0, 188, 83, 18, 92, 19, 92, 20, 92, 51, 128, 49, 109, 233, 143, 117, 153, 46, 104, 230, 107, 229, 115, 210, 148, 62, 114, 12, 153, 236, 153, 190, 133, 135, 144, 176, 141, 60, 0, 42, 0, 140, 78, 13, 95, 16, 95, 116, 79, 53, 82, 161, 84, 197, 54, 48, 141, 174, 140, 72, 136, 179, 140, 128, 138, 122, 146, 50, 106, 60, 0, 39, 0, 208, 253, 39, 0, 39, 49, 60, 0, 42, 0, 0, 78, 74, 78, 12, 95, 100, 136, 10, 79, 99, 136, 59, 83, 26, 84, 241, 88, 157, 79, 191, 84, 67, 216, 150, 220, 196, 54, 214, 60, 34, 109, 78, 121, 77, 216, 63, 220, 152, 96, 23, 115, 241, 148, 249, 88, 214, 99, 57, 107, 220, 134, 252, 144, 219, 90, 42, 111, 149, 121, 38, 122, 165, 146, 4, 91, 107, 86, 1, 89, 127, 116, 101, 158, 68, 126, 209, 74, 185, 106, 201, 107, 171, 145, 223, 158, 105, 139, 214, 157, 243, 158, 60, 0, 42, 0, 65, 78, 234, 78, 28, 83, 47, 87, 55, 89, 157, 81, 144, 91, 130, 108, 210, 139, 198, 143, 135, 79, 156, 91, 33, 96, 182, 108, 203, 114, 116, 153, 166, 84, 232, 89, 211, 92, 248, 93, 44, 95, 94, 96, 248, 98, 194, 103, 198, 115, 106, 136, 59, 141, 228, 143, 167, 91, 69, 98, 24, 104, 75, 104, 191, 62, 245, 116, 25, 119, 240, 128, 17, 138, 164, 140, 251, 143, 205, 117, 251, 121, 27, 128, 152, 136, 210, 81, 160, 127, 19, 132, 230, 134, 81, 138, 82, 138, 189, 140, 144, 90, 134, 102, 56, 105, 131, 138, 224, 141, 87, 144, 9, 152, 144, 152, 244, 152, 145, 117, 0, 81, 170, 113, 183, 123, 172, 93, 91, 95, 92, 95, 148, 135, 122, 144, 36, 152, 242, 91, 183, 93, 3, 124, 74, 152, 101, 216, 231, 220, 76, 76, 93, 95, 94, 95, 59, 139, 212, 147, 250, 137, 24, 60, 137, 139, 3, 158, 60, 0, 42, 0, 89, 78, 242, 93, 229, 78, 135, 148, 65, 79, 66, 216, 203, 223, 58, 101, 227, 119, 148, 128, 161, 94, 35, 130, 225, 130, 226, 130, 129, 134, 212, 145, 26, 80, 70, 98, 79, 145, 111, 80, 22, 123, 24, 144, 58, 93, 209, 101, 5, 105, 230, 156, 111, 146, 214, 101, 255, 136, 38, 142, 34, 143, 124, 101, 152, 135, 231, 73, 165, 106, 114, 75, 18, 121, 100, 130, 251, 135, 87, 152, 89, 143, 110, 159, 60, 0, 42, 0, 66, 78, 64, 216, 134, 220, 73, 78, 191, 78, 11, 95, 8, 82, 198, 95, 138, 128, 122, 130, 174, 139, 166, 78, 57, 52, 7, 79, 121, 92, 2, 95, 63, 79, 90, 79, 174, 82, 83, 84, 68, 87, 121, 95, 145, 98, 89, 103, 52, 128, 133, 130, 209, 139, 145, 144, 126, 79, 109, 84, 121, 84, 196, 92, 8, 96, 63, 96, 19, 102, 141, 103, 37, 107, 209, 60, 198, 108, 136, 112, 199, 121, 206, 126, 197, 130, 227, 139, 127, 154, 203, 79, 85, 89, 31, 94, 32, 94, 8, 95, 72, 216, 215, 223, 187, 103, 2, 109, 66, 109, 180, 115, 171, 117, 191, 127, 88, 216, 248, 223, 118, 143, 101, 52, 8, 85, 188, 87, 146, 96, 57, 99, 89, 99, 39, 104, 58, 104, 45, 107, 101, 109, 115, 109, 202, 118, 117, 136, 10, 140, 218, 82, 246, 87, 248, 87, 165, 96, 156, 99, 185, 107, 112, 117, 91, 120, 155, 127, 202, 127, 204, 127, 163, 136, 50, 138, 51, 138, 89, 140, 91, 140, 244, 145, 109, 150, 191, 150, 70, 94, 97, 101, 121, 102, 237, 104, 148, 107, 89, 110, 50, 113, 82, 216, 127, 221, 59, 67, 225, 134, 77, 138, 199, 141, 252, 142, 56, 144, 32, 146, 132, 78, 127, 81, 52, 53, 15, 97, 162, 110, 72, 115, 236, 117, 106, 119, 233, 122, 252, 65, 34, 127, 169, 127, 132, 128, 212, 136, 219, 136, 99, 138, 174, 154, 233, 82, 213, 90, 217, 94, 143, 105, 23, 118, 137, 129, 52, 135, 126, 151, 197, 153, 4, 81, 137, 57, 142, 100, 248, 105, 197, 107, 105, 111, 160, 113, 164, 113, 30, 118, 250, 132, 188, 138, 82, 149, 93, 158, 98, 158, 211, 158, 147, 82, 1, 54, 27, 87, 191, 88, 17, 91, 31, 91, 167, 93, 182, 97, 204, 97, 192, 102, 170, 107, 105, 61, 186, 111, 188, 113, 218, 113, 49, 118, 150, 119, 83, 122, 10, 126, 87, 130, 160, 135, 241, 91, 129, 101, 206, 102, 141, 106, 93, 107, 225, 113, 241, 113, 243, 127, 252, 127, 198, 129, 143, 133, 57, 137, 249, 140, 168, 155, 84, 118, 0, 141, 176, 147, 113, 149, 118, 126, 121, 126, 217, 133, 221, 133, 119, 140, 44, 151, 227, 155, 130, 157, 131, 157, 164, 57, 55, 112, 111, 139, 112, 139, 179, 145, 183, 145, 80, 153, 68, 216, 59, 221, 85, 216, 37, 223, 25, 134, 63, 148, 193, 157, 202, 157, 200, 86, 255, 97, 91, 154, 231, 157, 254, 157, 73, 134, 124, 137, 120, 159, 155, 139, 60, 0, 42, 0, 7, 83, 183, 143, 105, 117, 44, 105, 31, 132, 141, 157, 78, 124, 60, 0, 42, 0, 43, 78, 39, 87, 139, 83, 22, 84, 152, 94, 188, 98, 146, 103, 173, 87, 38, 158, 96, 104, 45, 158, 225, 87, 114, 91, 15, 105, 9, 157, 15, 147, 40, 157, 211, 88, 118, 157, 26, 148, 60, 0, 42, 0, 89, 114, 34, 79, 145, 83, 136, 92, 147, 83, 161, 115, 189, 130, 12, 123, 156, 134, 16, 88, 21, 93, 22, 93, 175, 109, 26, 115, 10, 116, 88, 116, 90, 119, 89, 136, 4, 111, 86, 159, 105, 216, 50, 222, 60, 0, 42, 0, 83, 216, 19, 221, 138, 83, 140, 94, 209, 84, 22, 85, 94, 85, 214, 117, 197, 150, 2, 118, 94, 216, 221, 221, 101, 133, 60, 0, 42, 0, 156, 82, 32, 87, 103, 143, 154, 78, 126, 137, 182, 139, 156, 78, 189, 114, 158, 78, 203, 142, 211, 143, 5, 90, 28, 99, 17, 120, 249, 79, 41, 108, 109, 90, 151, 99, 29, 138, 212, 148, 224, 99, 44, 108, 48, 115, 80, 128, 89, 216, 22, 223, 20, 87, 15, 122, 171, 122, 126, 159, 60, 0, 42, 0, 132, 46, 91, 78, 64, 84, 60, 0, 42, 0, 223, 84, 55, 85, 178, 85, 60, 0, 42, 0, 59, 80, 150, 99, 141, 102, 48, 105, 78, 86, 113, 111, 46, 136, 60, 0, 42, 0, 55, 114, 54, 128, 83, 99, 216, 148, 246, 99, 254, 145, 58, 114, 227, 146, 232, 100, 129, 147, 60, 0, 42, 0, 95, 78, 20, 84, 182, 81, 220, 87, 206, 145, 34, 86, 28, 111, 196, 88, 60, 0, 42, 0, 26, 78, 246, 83, 243, 102, 117, 152, 245, 102, 28, 89, 180, 98, 186, 144, 177, 78, 188, 103, 1, 152, 84, 102, 189, 103, 232, 112, 161, 53, 88, 85, 178, 109, 18, 140, 40, 88, 151, 107, 75, 129, 106, 105, 109, 105, 73, 132, 19, 145, 140, 153, 247, 80, 75, 107, 183, 88, 14, 66, 101, 151, 170, 93, 171, 93, 219, 100, 196, 102, 197, 102, 178, 111, 193, 113, 90, 216, 28, 220, 1, 139, 52, 145, 35, 153, 136, 86, 235, 100, 215, 102, 163, 118, 177, 119, 113, 147, 234, 100, 184, 119, 15, 121, 166, 66, 145, 147, 65, 153, 122, 157, 23, 114, 55, 148, 104, 151, 92, 154, 8, 158, 60, 0, 42, 0, 170, 78, 60, 0, 42, 0, 122, 94, 45, 89, 6, 84, 150, 89, 150, 103, 128, 107, 69, 121, 30, 138, 147, 85, 70, 105, 112, 129, 125, 132, 1, 67, 69, 70, 1, 157, 128, 144, 60, 0, 42, 0, 59, 114, 39, 92, 45, 92, 180, 128, 154, 87, 218, 89, 227, 92, 154, 64, 122, 143, 4, 80, 209, 112, 231, 115, 145, 122, 156, 80, 47, 88, 250, 99, 35, 140, 250, 142, 126, 52, 194, 85, 177, 90, 173, 95, 46, 97, 22, 100, 71, 100, 58, 115, 49, 58, 154, 102, 163, 105, 100, 116, 118, 116, 89, 144, 101, 144, 154, 146, 214, 152, 6, 153, 162, 93, 164, 93, 175, 122, 176, 122, 4, 69, 26, 153, 71, 126, 32, 139, 33, 139, 91, 216, 190, 223, 144, 147, 208, 156, 25, 75, 187, 152, 40, 134, 142, 144, 100, 152, 41, 156, 60, 0, 42, 0, 248, 78, 142, 91, 134, 92, 173, 98, 115, 103, 213, 114, 172, 84, 252, 103, 237, 130, 17, 119, 133, 122, 136, 122, 0, 130, 96, 80, 121, 90, 62, 93, 148, 110, 154, 105, 212, 132, 34, 157, 60, 159, 196, 149, 21, 154, 105, 159, 213, 157, 60, 0, 42, 0, 126, 122, 129, 137, 165, 148, 84, 216, 82, 220, 111, 131, 148, 122, 142, 136, 68, 123, 79, 138, 111, 132, 142, 113, 158, 137, 127, 151, 95, 115, 94, 158, 86, 216, 175, 222, 172, 133, 220, 102, 255, 113, 94, 130, 197, 119, 229, 133, 0, 128, 133, 126, 194, 157, 145, 139, 112, 148, 60, 0, 42, 0, 24, 79, 231, 95, 56, 101, 102, 84, 46, 96, 209, 108, 125, 94, 160, 96, 12, 144, 128, 158, 250, 110, 130, 97, 42, 81, 152, 86, 0, 112, 62, 145, 204, 106, 139, 126, 48, 128, 60, 0, 42, 0, 34, 92, 36, 92, 49, 117, 139, 108, 185, 114, 85, 63, 76, 216, 222, 223, 185, 108, 172, 128, 174, 144, 35, 96, 191, 101, 163, 117, 243, 92, 95, 109, 222, 121, 67, 67, 192, 148, 100, 80, 156, 131, 184, 131, 176, 134, 39, 138, 56, 110, 54, 115, 48, 144, 245, 144, 127, 156, 98, 105, 55, 115, 74, 144, 62, 146, 137, 156, 15, 143, 192, 153, 99, 135, 119, 155, 85, 133, 54, 143, 139, 155, 254, 106, 60, 0, 42, 0, 203, 83, 9, 103, 35, 78, 99, 83, 73, 145, 195, 130, 173, 61, 145, 127, 174, 94, 47, 104, 144, 127, 132, 104, 72, 128, 28, 129, 160, 131, 213, 148, 117, 110, 15, 135, 137, 121, 170, 146, 0, 75, 241, 105, 86, 114, 221, 158, 238, 97, 60, 0, 42, 0, 128, 46, 200, 83, 243, 83, 124, 94, 81, 79, 145, 79, 214, 114, 255, 124, 202, 84, 255, 86, 247, 89, 165, 91, 223, 92, 218, 103, 112, 114, 241, 139, 0, 85, 80, 121, 246, 143, 1, 64, 180, 134, 180, 78, 129, 140, 201, 145, 109, 145, 152, 138, 44, 159, 60, 0, 42, 0, 103, 91, 143, 132, 87, 114, 60, 0, 42, 0, 121, 96, 102, 82, 223, 112, 218, 115, 237, 128, 99, 80, 113, 85, 38, 93, 63, 57, 202, 109, 249, 109, 9, 113, 17, 113, 155, 71, 9, 150, 110, 110, 18, 115, 76, 129, 248, 131, 89, 113, 125, 120, 227, 90, 57, 111, 103, 65, 34, 145, 131, 145, 185, 149, 46, 91, 232, 97, 246, 123, 213, 97, 217, 129, 235, 158, 60, 0, 42, 0, 160, 139, 37, 78, 0, 138, 1, 138, 169, 92, 246, 94, 22, 102, 191, 108, 142, 112, 171, 130, 141, 89, 248, 89, 20, 120, 46, 90, 208, 118, 43, 90, 255, 114, 2, 116, 79, 120, 212, 144, 134, 149, 14, 150, 82, 93, 83, 93, 122, 110, 154, 131, 105, 88, 197, 99, 76, 105, 254, 65, 125, 138, 158, 120, 117, 123, 150, 125, 18, 135, 5, 133, 156, 152, 100, 134, 187, 149, 179, 83, 144, 106, 79, 152, 84, 152, 219, 88, 55, 124, 180, 86, 204, 93, 233, 106, 236, 158, 215, 93, 231, 88, 77, 91, 214, 93, 125, 158, 57, 121, 163, 158, 60, 0, 42, 0, 53, 89, 129, 98, 135, 108, 117, 78, 86, 81, 68, 89, 232, 79, 87, 81, 61, 83, 7, 95, 77, 136, 67, 80, 163, 83, 169, 99, 60, 119, 83, 145, 67, 93, 29, 97, 74, 98, 220, 99, 122, 59, 234, 104, 48, 110, 55, 110, 48, 116, 18, 132, 254, 144, 60, 105, 104, 127, 67, 144, 146, 150, 20, 111, 250, 136, 150, 93, 45, 98, 88, 135, 23, 137, 94, 86, 189, 142, 71, 155, 47, 126, 191, 106, 19, 154, 225, 158, 180, 83, 84, 216, 44, 220, 11, 156, 160, 157, 228, 158, 94, 159, 145, 159, 23, 117, 237, 158, 60, 81, 105, 152, 52, 159, 216, 93, 218, 93, 57, 159, 238, 102, 88, 155, 116, 159, 246, 158, 60, 0, 42, 0, 140, 83, 235, 149, 159, 89, 195, 137, 106, 114, 189, 84, 242, 89, 101, 95, 102, 95, 26, 120, 1, 85, 180, 91, 79, 102, 115, 130, 142, 137, 140, 154, 80, 80, 20, 113, 26, 140, 173, 85, 48, 88, 101, 101, 251, 112, 48, 113, 49, 113, 111, 120, 129, 150, 193, 150, 191, 80, 59, 105, 142, 110, 223, 110, 85, 132, 235, 156, 173, 83, 149, 88, 165, 102, 125, 145, 10, 91, 51, 140, 100, 216, 192, 220, 13, 153, 8, 157, 243, 61, 252, 61, 196, 113, 213, 113, 250, 138, 93, 141, 173, 72, 51, 155, 213, 102, 51, 157, 16, 154, 165, 86, 63, 91, 118, 130, 11, 141, 64, 145, 4, 74, 225, 76, 43, 59, 227, 102, 19, 114, 182, 145, 52, 154, 195, 157, 84, 112, 23, 141, 254, 137, 140, 139, 188, 145, 92, 153, 87, 154, 240, 157, 119, 130, 105, 216, 74, 222, 78, 112, 96, 154, 103, 112, 158, 139, 83, 140, 197, 145, 84, 140, 105, 112, 60, 0, 42, 0, 46, 106, 150, 113, 197, 142, 21, 107, 60, 0, 42, 0, 217, 86, 224, 86, 37, 150, 52, 150, 140, 79, 148, 87, 251, 89, 226, 57, 7, 109, 243, 151, 131, 154, 54, 104, 183, 107, 36, 108, 53, 131, 208, 81, 99, 90, 4, 65, 245, 121, 107, 131, 223, 148, 112, 150, 145, 85, 25, 88, 20, 97, 67, 123, 106, 125, 192, 136, 123, 150, 130, 150, 69, 107, 181, 110, 71, 97, 127, 100, 22, 118, 139, 121, 166, 146, 248, 125, 45, 133, 135, 151, 242, 138, 18, 151, 100, 216, 236, 220, 240, 153, 126, 86, 201, 149, 32, 151, 143, 61, 254, 151, 60, 0, 42, 0, 152, 81, 81, 78, 66, 53, 31, 84, 190, 114, 166, 101, 194, 130, 160, 87, 255, 108, 151, 53, 1, 87, 254, 92, 206, 112, 250, 114, 226, 115, 20, 138, 26, 138, 108, 90, 197, 91, 31, 93, 47, 93, 235, 109, 118, 131, 33, 138, 246, 148, 29, 146, 130, 159, 219, 110, 146, 120, 36, 89, 30, 145, 128, 146, 86, 86, 165, 107, 140, 116, 41, 133, 190, 138, 154, 86, 173, 106, 235, 135, 38, 77, 42, 151, 87, 159, 227, 157, 60, 0, 42, 0, 90, 78, 64, 216, 202, 220, 244, 94, 57, 92, 21, 95, 50, 84, 110, 153, 147, 134, 122, 141, 254, 109, 15, 146, 144, 150, 242, 152, 119, 151, 238, 152, 4, 103, 17, 143, 160, 150, 228, 120, 155, 141, 62, 118, 190, 93, 131, 106, 229, 111, 230, 111, 190, 135, 177, 150, 253, 106, 31, 134, 110, 118, 148, 139, 60, 0, 42, 0, 112, 83, 21, 109, 228, 128, 189, 87, 26, 131, 55, 88, 90, 110, 12, 115, 213, 94, 145, 132, 115, 145, 109, 97, 150, 97, 151, 97, 74, 118, 163, 155, 101, 56, 218, 97, 188, 106, 67, 69, 60, 0, 42, 0, 140, 124, 60, 0, 42, 0, 46, 89, 137, 84, 206, 89, 176, 98, 241, 108, 131, 107, 230, 128, 15, 119, 231, 121, 47, 158, 96, 146, 245, 150, 133, 151, 38, 157, 60, 0, 42, 0, 108, 98, 138, 127, 248, 101, 104, 103, 128, 112, 38, 150, 51, 150, 207, 152, 111, 79, 183, 82, 28, 108, 161, 117, 150, 148, 159, 87, 137, 95, 28, 102, 11, 109, 143, 127, 202, 112, 220, 115, 90, 216, 117, 223, 59, 119, 53, 93, 56, 93, 218, 99, 216, 134, 125, 150, 109, 101, 152, 102, 74, 105, 108, 113, 13, 118, 147, 121, 21, 59, 249, 138, 48, 143, 90, 147, 57, 157, 186, 152, 10, 148, 17, 156, 55, 151, 9, 158, 60, 0, 42, 0, 240, 78, 82, 79, 113, 87, 159, 92, 123, 81, 205, 103, 180, 112, 39, 108, 210, 117, 59, 125, 159, 80, 238, 142, 67, 97, 103, 105, 49, 108, 70, 135, 10, 153, 218, 153, 38, 53, 233, 97, 1, 101, 70, 68, 98, 118, 60, 0, 42, 0, 15, 96, 89, 96, 55, 104, 149, 127, 71, 138, 216, 105, 62, 111, 35, 106, 1, 112, 60, 0, 42, 0, 77, 89, 170, 127, 34, 121, 60, 0, 42, 0, 148, 94, 220, 95, 241, 130, 92, 104, 64, 80, 104, 85, 116, 90, 236, 126, 186, 131, 67, 216, 4, 222, 68, 216, 152, 223, 150, 90, 54, 110, 208, 72, 200, 90, 32, 103, 80, 113, 91, 116, 36, 86, 132, 100, 164, 120, 40, 127, 66, 127, 207, 140, 51, 149, 49, 106, 142, 116, 7, 117, 211, 125, 103, 135, 67, 127, 46, 137, 72, 147, 44, 157, 102, 158, 48, 91, 201, 97, 186, 129, 25, 151, 250, 151, 112, 158, 22, 117, 145, 157, 182, 86, 70, 91, 126, 91, 22, 101, 76, 127, 77, 139, 167, 157, 251, 106, 212, 116, 47, 121, 33, 134, 123, 139, 175, 157, 77, 148, 147, 126, 51, 136, 234, 157, 249, 157, 14, 158, 26, 158, 60, 0, 42, 0, 193, 118, 206, 143, 200, 118, 20, 131, 103, 131, 182, 85, 185, 131, 205, 134, 97, 216, 15, 223, 129, 110, 139, 110, 36, 132, 38, 132, 204, 80, 75, 88, 121, 105, 226, 110, 37, 132, 62, 132, 70, 111, 146, 113, 229, 132, 71, 135, 105, 116, 148, 66, 127, 135, 52, 91, 8, 126, 162, 135, 217, 111, 218, 111, 244, 111, 223, 113, 174, 137, 13, 139, 98, 141, 5, 112, 192, 133, 163, 147, 13, 101, 27, 112, 32, 112, 5, 136, 74, 216, 79, 221, 47, 112, 52, 112, 15, 141, 255, 106, 93, 124, 111, 124, 60, 0, 42, 0, 232, 119, 103, 109, 226, 144, 172, 104, 141, 152, 149, 152, 150, 152, 108, 100, 44, 75, 113, 95, 65, 111, 132, 116, 63, 118, 78, 122, 52, 152, 202, 93, 238, 94, 109, 118, 60, 0, 42, 0, 32, 102, 108, 120, 181, 90, 142, 102, 161, 129, 95, 86, 149, 151, 27, 148, 102, 156, 60, 0, 42, 0, 241, 115, 37, 97, 4, 126, 26, 101, 80, 112, 92, 112, 200, 142, 60, 0, 39, 0, 208, 253, 39, 0, 40, 49, 60, 0, 42, 0, 76, 78, 44, 87, 25, 95, 89, 108, 90, 108, 97, 108, 92, 84, 235, 93, 71, 103, 172, 144, 75, 92, 63, 109, 236, 139, 168, 148, 207, 112, 109, 82, 143, 122, 218, 85, 20, 145, 77, 107, 163, 138, 188, 123, 144, 135, 46, 157, 162, 147, 30, 156, 60, 0, 42, 0, 224, 101, 203, 107, 51, 84, 52, 84, 62, 84, 73, 84, 156, 130, 20, 85, 42, 90, 22, 109, 111, 109, 218, 144, 167, 104, 248, 115, 35, 131, 129, 131, 33, 113, 102, 121, 251, 148, 64, 158, 8, 135, 136, 138, 145, 121, 145, 116, 106, 133, 241, 135, 195, 155, 80, 157, 85, 139, 47, 159, 225, 157, 60, 0, 42, 0, 148, 78, 72, 83, 245, 78, 13, 79, 169, 89, 145, 94, 228, 95, 3, 96, 255, 101, 51, 59, 102, 107, 157, 115, 174, 79, 201, 79, 53, 80, 66, 99, 78, 85, 44, 90, 126, 114, 247, 115, 64, 100, 148, 120, 73, 158, 147, 113, 102, 116, 30, 130, 245, 90, 225, 94, 174, 97, 85, 111, 27, 81, 70, 106, 18, 117, 97, 157, 140, 142, 60, 0, 42, 0, 64, 81, 255, 82, 51, 52, 161, 82, 10, 98, 6, 79, 124, 92, 100, 98, 34, 150, 94, 87, 137, 92, 76, 103, 226, 95, 105, 114, 249, 119, 180, 130, 213, 143, 124, 83, 68, 101, 239, 139, 158, 96, 159, 96, 174, 96, 133, 124, 217, 82, 100, 102, 16, 113, 28, 144, 122, 90, 77, 93, 230, 117, 112, 151, 155, 154, 98, 88, 102, 89, 104, 93, 169, 110, 150, 150, 250, 150, 254, 150, 228, 91, 131, 113, 164, 138, 92, 158, 200, 146, 185, 122, 123, 144, 1, 74, 26, 151, 63, 159, 39, 151, 22, 154, 64, 159, 1, 134, 169, 157, 60, 0, 42, 0, 68, 78, 59, 147, 60, 0, 42, 0, 117, 122, 184, 82, 22, 99, 60, 109, 50, 90, 86, 117, 138, 122, 167, 90, 217, 134, 215, 85, 50, 100, 155, 110, 37, 111, 170, 122, 3, 159, 40, 101, 60, 0, 42, 0, 3, 90, 60, 0, 42, 0, 230, 116, 100, 79, 147, 84, 218, 57, 183, 144, 60, 0, 42, 0, 73, 128, 156, 136, 226, 85, 125, 129, 131, 129, 106, 137, 200, 151, 228, 151, 60, 0, 42, 0, 114, 92, 199, 84, 242, 116, 60, 0, 42, 0, 29, 99, 45, 80, 161, 109, 41, 85, 185, 109, 180, 131, 38, 110, 39, 115, 157, 122, 53, 132, 23, 135, 169, 122, 120, 135, 18, 142, 190, 100, 60, 0, 42, 0, 17, 98, 80, 90, 112, 99, 0, 76, 60, 0, 42, 0, 244, 78, 131, 108, 159, 128, 103, 83, 130, 103, 229, 129, 83, 80, 126, 99, 180, 109, 137, 90, 68, 94, 225, 99, 37, 110, 37, 113, 106, 120, 67, 105, 91, 129, 161, 101, 131, 119, 67, 216, 223, 223, 237, 100, 227, 111, 140, 159, 193, 116, 210, 129, 216, 150, 119, 159, 60, 0, 42, 0, 106, 107, 142, 85, 245, 122, 60, 0, 42, 0, 52, 93, 60, 0, 42, 0, 22, 89, 67, 216, 59, 220, 30, 89, 97, 152, 60, 0, 42, 0, 113, 83, 1, 90, 67, 216, 255, 220, 211, 112, 78, 80, 180, 85, 78, 54, 153, 90, 4, 97, 203, 99, 251, 99, 40, 110, 14, 132, 54, 144, 135, 150, 136, 150, 174, 95, 51, 105, 114, 105, 166, 110, 104, 113, 104, 132, 115, 132, 116, 138, 50, 135, 91, 135, 163, 137, 240, 113, 135, 133, 194, 156, 3, 156, 4, 156, 205, 93, 60, 0, 42, 0, 215, 86, 230, 151, 41, 87, 244, 86, 15, 94, 169, 108, 241, 149, 221, 143, 215, 92, 222, 92, 8, 109, 203, 151, 69, 104, 160, 109, 47, 85, 55, 94, 223, 96, 89, 120, 244, 126, 161, 85, 13, 87, 129, 90, 67, 94, 75, 110, 136, 110, 31, 116, 44, 145, 108, 93, 85, 144, 77, 111, 173, 125, 89, 111, 127, 111, 209, 120, 246, 132, 128, 145, 240, 111, 79, 147, 200, 149, 160, 155, 83, 118, 96, 66, 185, 137, 58, 151, 8, 107, 169, 114, 60, 0, 42, 0, 131, 83, 31, 79, 42, 79, 62, 92, 172, 126, 212, 89, 79, 59, 156, 112, 174, 115, 155, 130, 199, 130, 39, 109, 19, 90, 87, 92, 88, 109, 255, 139, 73, 80, 125, 80, 35, 93, 182, 104, 207, 117, 74, 120, 113, 131, 234, 91, 84, 93, 171, 95, 7, 97, 37, 115, 169, 154, 144, 102, 50, 105, 82, 113, 75, 116, 255, 117, 251, 65, 114, 129, 73, 130, 102, 132, 141, 132, 234, 151, 170, 154, 171, 154, 222, 80, 148, 156, 48, 81, 177, 100, 200, 120, 86, 216, 228, 223, 239, 125, 201, 138, 19, 142, 209, 151, 32, 152, 25, 59, 63, 133, 97, 216, 232, 221, 97, 147, 170, 155, 251, 111, 179, 133, 217, 151, 185, 152, 221, 88, 34, 112, 36, 134, 225, 151, 150, 101, 60, 0, 42, 0, 107, 83, 58, 78, 42, 103, 77, 79, 115, 84, 186, 112, 79, 117, 195, 128, 255, 130, 222, 83, 206, 142, 9, 92, 19, 140, 130, 85, 166, 90, 45, 110, 201, 61, 50, 114, 44, 115, 203, 131, 95, 113, 168, 120, 60, 135, 112, 97, 173, 113, 154, 114, 237, 125, 26, 133, 95, 135, 91, 136, 192, 97, 143, 116, 123, 127, 94, 136, 2, 139, 39, 153, 135, 155, 177, 135, 61, 137, 53, 153, 74, 143, 79, 155, 239, 133, 207, 147, 40, 151, 218, 156, 86, 153, 54, 134, 134, 139, 151, 142, 129, 76, 143, 139, 155, 142, 60, 0, 42, 0, 100, 99, 64, 113, 155, 88, 5, 126, 182, 93, 60, 0, 42, 0, 47, 95, 92, 82, 96, 90, 53, 94, 70, 88, 126, 110, 63, 135, 107, 111, 76, 140, 78, 95, 234, 88, 99, 112, 60, 0, 42, 0, 56, 78, 19, 82, 77, 108, 168, 126, 140, 91, 143, 92, 143, 98, 132, 130, 169, 115, 8, 125, 86, 99, 125, 152, 247, 112, 19, 116, 17, 152, 200, 75, 235, 127, 60, 0, 42, 0, 155, 91, 7, 80, 13, 85, 61, 99, 204, 118, 230, 87, 73, 90, 203, 96, 90, 102, 105, 102, 154, 104, 254, 126, 24, 129, 124, 102, 0, 105, 44, 116, 150, 118, 192, 131, 22, 132, 121, 117, 85, 119, 151, 120, 169, 125, 176, 125, 19, 143, 32, 142, 196, 146, 212, 146, 74, 74, 54, 55, 60, 0, 42, 0, 7, 78, 77, 83, 80, 83, 167, 89, 232, 95, 101, 99, 21, 129, 166, 140, 85, 129, 44, 132, 16, 143, 171, 111, 61, 147, 141, 133, 195, 135, 3, 141, 171, 147, 81, 73, 14, 141, 60, 0, 42, 0, 156, 144, 100, 103, 2, 123, 60, 0, 42, 0, 55, 102, 41, 110, 109, 88, 8, 59, 133, 105, 171, 110, 146, 143, 159, 107, 101, 116, 31, 118, 112, 133, 60, 143, 113, 140, 64, 143, 193, 156, 176, 151, 102, 216, 72, 221, 27, 156, 46, 156, 60, 0, 42, 0, 1, 83, 135, 101, 99, 95, 185, 126, 134, 112, 159, 115, 160, 130, 251, 149, 11, 125, 137, 134, 138, 134, 243, 115, 12, 150, 29, 116, 239, 150, 18, 118, 94, 128, 188, 153, 199, 153, 112, 155, 252, 156, 13, 157, 161, 135, 186, 149, 191, 149, 193, 135, 197, 149, 36, 159, 230, 149, 60, 0, 42, 0, 14, 82, 59, 84, 223, 95, 134, 98, 97, 84, 179, 128, 10, 125, 125, 104, 23, 129, 51, 122, 79, 122, 105, 122, 60, 0, 42, 0, 238, 149, 143, 89, 118, 108, 79, 85, 2, 110, 172, 131, 254, 99, 53, 100, 80, 152, 186, 116, 60, 0, 42, 0, 90, 84, 43, 146, 190, 147, 60, 0, 42, 0, 70, 216, 193, 223, 35, 92, 41, 92, 42, 92, 43, 92, 106, 108, 60, 0, 42, 0, 161, 78, 190, 78, 102, 81, 139, 115, 252, 78, 122, 95, 159, 134, 163, 131, 60, 0, 42, 0, 193, 52, 82, 127, 81, 127, 128, 95, 131, 95, 137, 103, 84, 127, 141, 95, 216, 96, 128, 102, 226, 104, 245, 131, 231, 134, 139, 143, 178, 125, 68, 135, 183, 138, 30, 143, 7, 112, 77, 155, 60, 0, 42, 0, 132, 89, 216, 95, 250, 101, 243, 118, 203, 143, 27, 103, 34, 103, 17, 73, 60, 0, 42, 0, 57, 113, 60, 0, 42, 0, 193, 127, 225, 85, 195, 110, 95, 158, 137, 135, 147, 147, 178, 157, 60, 0, 42, 0, 34, 54, 220, 82, 99, 89, 85, 88, 97, 93, 161, 102, 37, 68, 202, 132, 136, 119, 108, 128, 50, 55, 60, 0, 42, 0, 238, 116, 121, 133, 21, 117, 75, 127, 70, 159, 60, 0, 39, 0, 208, 253, 39, 0, 41, 49, 60, 0, 42, 0, 92, 98, 161, 126, 194, 143, 195, 143, 123, 122, 6, 125, 118, 134, 83, 150, 57, 85, 228, 109, 211, 118, 250, 107, 0, 118, 138, 123, 60, 0, 42, 0, 144, 78, 142, 78, 3, 79, 165, 54, 48, 59, 152, 144, 89, 79, 164, 89, 117, 98, 69, 103, 36, 107, 151, 115, 153, 115, 188, 101, 194, 118, 254, 129, 124, 156, 123, 78, 222, 79, 106, 81, 186, 121, 253, 122, 66, 67, 103, 136, 27, 90, 47, 90, 49, 90, 89, 104, 243, 114, 1, 130, 48, 131, 0, 140, 81, 145, 128, 153, 20, 110, 233, 150, 90, 155, 35, 88, 44, 88, 51, 93, 78, 93, 91, 93, 101, 57, 9, 97, 196, 99, 112, 105, 29, 110, 97, 110, 109, 117, 98, 120, 116, 129, 56, 132, 133, 150, 26, 97, 213, 101, 2, 59, 97, 105, 134, 105, 72, 107, 79, 114, 92, 116, 69, 130, 46, 132, 94, 134, 206, 137, 62, 144, 172, 154, 1, 111, 110, 119, 172, 122, 118, 107, 173, 127, 51, 68, 83, 135, 21, 137, 219, 138, 211, 150, 24, 153, 59, 76, 158, 111, 6, 130, 77, 133, 166, 137, 48, 142, 41, 91, 184, 135, 63, 143, 93, 147, 98, 216, 54, 222, 95, 107, 181, 116, 35, 139, 189, 155, 31, 154, 195, 154, 223, 101, 69, 124, 27, 134, 5, 156, 163, 76, 224, 157, 6, 158, 60, 0, 42, 0, 14, 78, 136, 78, 27, 79, 135, 91, 127, 92, 189, 127, 215, 46, 232, 150, 193, 79, 227, 79, 185, 121, 237, 139, 4, 87, 255, 92, 74, 80, 49, 53, 44, 83, 9, 87, 190, 94, 84, 101, 100, 121, 98, 55, 158, 101, 5, 145, 180, 80, 217, 91, 64, 105, 64, 116, 16, 118, 45, 132, 195, 56, 7, 130, 158, 138, 179, 122, 217, 146, 40, 152, 137, 159, 115, 86, 144, 140, 188, 93, 217, 97, 155, 57, 148, 101, 140, 158, 12, 134, 108, 159, 60, 0, 42, 0, 128, 128, 137, 115, 109, 154, 43, 87, 127, 128, 170, 89, 236, 95, 139, 130, 140, 130, 107, 153, 178, 128, 34, 57, 49, 102, 241, 114, 215, 121, 193, 144, 252, 79, 234, 92, 103, 95, 116, 109, 33, 120, 31, 131, 176, 148, 132, 152, 223, 87, 9, 88, 161, 95, 134, 96, 208, 96, 50, 107, 226, 109, 239, 109, 21, 140, 8, 150, 133, 85, 144, 85, 169, 85, 187, 85, 128, 90, 211, 91, 189, 94, 219, 104, 220, 104, 235, 104, 52, 113, 25, 116, 222, 119, 114, 120, 91, 216, 64, 220, 51, 144, 235, 152, 173, 153, 70, 158, 8, 97, 211, 107, 234, 110, 92, 113, 99, 63, 34, 122, 109, 127, 72, 130, 213, 136, 137, 138, 71, 144, 58, 146, 16, 152, 215, 90, 142, 93, 43, 98, 68, 115, 9, 118, 5, 66, 237, 66, 206, 125, 174, 132, 227, 132, 31, 135, 46, 135, 59, 70, 13, 143, 137, 146, 74, 86, 126, 97, 79, 111, 54, 122, 249, 132, 202, 146, 255, 156, 73, 216, 176, 220, 166, 111, 207, 113, 166, 121, 237, 138, 107, 140, 121, 144, 37, 147, 190, 149, 37, 157, 42, 157, 37, 81, 69, 216, 138, 220, 224, 113, 253, 123, 119, 133, 129, 133, 108, 158, 82, 118, 7, 121, 22, 121, 88, 126, 93, 216, 80, 220, 167, 145, 74, 155, 82, 157, 28, 121, 101, 122, 71, 153, 242, 106, 49, 151, 125, 139, 93, 143, 45, 148, 14, 107, 72, 154, 59, 155, 94, 124, 74, 156, 248, 157, 29, 107, 18, 158, 165, 159, 201, 142, 48, 155, 49, 155, 106, 112, 114, 124, 41, 114, 60, 0, 42, 0, 39, 99, 98, 131, 154, 111, 242, 155, 60, 0, 42, 0, 240, 102, 241, 102, 166, 126, 4, 125, 185, 123, 95, 95, 241, 119, 96, 95, 60, 0, 42, 0, 157, 46, 8, 103, 9, 98, 22, 82, 156, 89, 132, 92, 136, 98, 69, 216, 223, 222, 179, 92, 165, 115, 63, 121, 113, 96, 149, 65, 133, 96, 166, 96, 142, 134, 143, 134, 207, 142, 186, 148, 5, 150, 115, 99, 192, 141, 195, 141, 71, 72, 164, 124, 138, 141, 5, 146, 181, 124, 94, 146, 38, 73, 177, 149, 178, 149, 62, 106, 215, 123, 51, 91, 189, 93, 101, 216, 229, 220, 160, 159, 48, 55, 38, 59, 57, 112, 70, 124, 230, 158, 26, 114, 37, 134, 141, 142, 180, 121, 101, 124, 17, 158, 112, 124, 25, 158, 60, 0, 42, 0, 230, 86, 34, 158, 72, 82, 164, 81, 129, 96, 34, 119, 51, 158, 195, 91, 245, 109, 1, 110, 6, 110, 10, 110, 21, 110, 204, 96, 233, 104, 147, 63, 126, 132, 14, 135, 83, 158, 162, 123, 172, 132, 53, 135, 247, 136, 246, 156, 93, 72, 204, 153, 27, 157, 61, 91, 73, 73, 119, 157, 65, 112, 24, 159, 29, 159, 60, 0, 42, 0, 67, 81, 134, 81, 32, 141, 88, 84, 237, 86, 133, 108, 167, 144, 108, 103, 163, 87, 48, 114, 159, 140, 159, 83, 225, 84, 6, 87, 14, 123, 150, 134, 129, 136, 161, 83, 14, 87, 244, 99, 114, 110, 40, 115, 24, 127, 18, 131, 11, 159, 18, 87, 19, 87, 108, 88, 180, 90, 196, 90, 144, 110, 146, 110, 63, 115, 66, 115, 158, 105, 172, 105, 157, 132, 149, 143, 227, 125, 1, 126, 93, 135, 111, 135, 109, 155, 174, 76, 124, 106, 177, 127, 136, 135, 151, 133, 28, 139, 69, 143, 255, 158, 177, 147, 222, 106, 141, 144, 53, 154, 162, 157, 176, 157, 181, 83, 60, 0, 42, 0, 220, 143, 246, 118, 58, 144, 96, 144, 250, 146, 60, 0, 42, 0, 23, 89, 153, 128, 180, 89, 71, 59, 40, 96, 209, 130, 184, 87, 79, 136, 98, 150, 134, 80, 155, 90, 190, 99, 87, 116, 63, 97, 144, 121, 235, 136, 17, 137, 110, 86, 36, 137, 88, 152, 60, 0, 42, 0, 91, 145, 40, 146, 60, 0, 42, 0, 85, 102, 10, 127, 136, 102, 116, 113, 128, 132, 107, 89, 50, 108, 51, 108, 149, 132, 21, 126, 121, 135, 95, 141, 53, 152, 7, 141, 167, 153, 60, 0, 42, 0, 145, 78, 251, 82, 0, 83, 233, 86, 152, 89, 132, 108, 173, 126, 0, 102, 184, 130, 71, 117, 3, 119, 208, 121, 162, 109, 28, 125, 24, 128, 58, 128, 231, 144, 242, 150, 42, 97, 179, 110, 96, 123, 124, 123, 22, 145, 178, 105, 137, 113, 183, 132, 144, 111, 198, 146, 82, 106, 212, 123, 28, 126, 83, 133, 66, 153, 60, 0, 42, 0, 65, 81, 61, 89, 142, 98, 193, 114, 45, 150, 104, 150, 146, 107, 122, 131, 151, 85, 30, 73, 23, 146, 149, 150, 158, 107, 187, 153, 210, 120, 30, 137, 241, 140, 35, 151, 115, 159, 60, 0, 42, 0, 85, 91, 159, 103, 208, 143, 125, 96, 211, 144, 93, 145, 138, 80, 242, 96, 32, 97, 6, 145, 77, 97, 106, 129, 75, 144, 235, 151, 245, 151, 133, 113, 168, 113, 247, 125, 252, 125, 176, 76, 116, 133, 150, 145, 128, 133, 158, 145, 43, 153, 215, 151, 244, 133, 222, 151, 251, 151, 10, 134, 60, 0, 42, 0, 163, 98, 103, 126, 60, 0, 42, 0, 99, 79, 229, 98, 200, 117, 149, 144, 184, 94, 173, 80, 200, 85, 205, 150, 137, 88, 222, 90, 117, 97, 253, 110, 24, 145, 230, 105, 112, 86, 197, 88, 193, 100, 173, 111, 250, 144, 91, 149, 195, 129, 85, 118, 221, 150, 222, 147, 217, 156, 241, 94, 73, 112, 84, 153, 69, 156, 219, 157, 112, 118, 60, 0, 42, 0, 129, 85, 216, 99, 69, 114, 153, 152, 82, 152, 43, 156, 60, 0, 42, 0, 56, 108, 44, 117, 143, 84, 243, 108, 209, 79, 199, 82, 200, 82, 16, 104, 199, 87, 128, 96, 225, 103, 140, 109, 127, 96, 4, 61, 155, 80, 229, 96, 17, 97, 103, 110, 103, 120, 96, 138, 78, 88, 113, 93, 110, 95, 57, 97, 249, 134, 66, 97, 10, 142, 156, 121, 172, 156, 52, 142, 210, 155, 60, 0, 42, 0, 40, 117, 218, 130, 159, 145, 60, 0, 42, 0, 58, 96, 61, 120, 38, 0, 39, 0, 40, 0, 39, 0, 0, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 32, 50, 38, 0, 39, 0, 40, 0, 39, 0, 3, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 38, 50, 38, 0, 39, 0, 40, 0, 39, 0, 9, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 34, 50, 38, 0, 39, 0, 40, 0, 39, 0, 93, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 40, 50, 38, 0, 39, 0, 40, 0, 39, 0, 140, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 33, 50, 38, 0, 39, 0, 40, 0, 39, 0, 148, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 36, 50, 38, 0, 39, 0, 40, 0, 39, 0, 227, 78, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 57, 50, 38, 0, 39, 0, 40, 0, 39, 0, 1, 79, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 61, 50, 38, 0, 39, 0, 40, 0, 39, 0, 17, 79, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 65, 50, 38, 0, 39, 0, 40, 0, 39, 0, 107, 81, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 39, 50, 38, 0, 39, 0, 40, 0, 39, 0, 109, 81, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 37, 50, 38, 0, 39, 0, 40, 0, 39, 0, 180, 82, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 56, 50, 38, 0, 39, 0, 40, 0, 39, 0, 65, 83, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 41, 50, 38, 0, 39, 0, 40, 0, 39, 0, 84, 83, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 63, 50, 38, 0, 39, 0, 40, 0, 39, 0, 13, 84, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 52, 50, 38, 0, 39, 0, 40, 0, 39, 0, 124, 84, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 58, 50, 38, 0, 39, 0, 40, 0, 39, 0, 219, 86, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 35, 50, 38, 0, 39, 0, 40, 0, 39, 0, 31, 87, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 47, 50, 38, 0, 39, 0, 40, 0, 39, 0, 102, 91, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 59, 50, 38, 0, 39, 0, 40, 0, 39, 0, 229, 101, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 48, 50, 38, 0, 39, 0, 40, 0, 39, 0, 8, 103, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 42, 50, 38, 0, 39, 0, 40, 0, 39, 0, 9, 103, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 50, 50, 38, 0, 39, 0, 40, 0, 39, 0, 40, 103, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 45, 50, 38, 0, 39, 0, 40, 0, 39, 0, 42, 104, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 49, 50, 38, 0, 39, 0, 40, 0, 39, 0, 52, 108, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 44, 50, 38, 0, 39, 0, 40, 0, 39, 0, 107, 112, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 43, 50, 38, 0, 39, 0, 40, 0, 39, 0, 121, 114, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 53, 50, 38, 0, 39, 0, 40, 0, 39, 0, 227, 118, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 50, 38, 0, 39, 0, 40, 0, 39, 0, 62, 121, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 51, 50, 38, 0, 39, 0, 40, 0, 39, 0, 93, 121, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 55, 50, 38, 0, 39, 0, 40, 0, 39, 0, 109, 121, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 64, 50, 38, 0, 39, 0, 40, 0, 39, 0, 234, 129, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 66, 50, 38, 0, 39, 0, 40, 0, 39, 0, 243, 129, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 67, 50, 38, 0, 39, 0, 40, 0, 39, 0, 161, 140, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 54, 50, 38, 0, 39, 0, 40, 0, 39, 0, 199, 140, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 62, 50, 38, 0, 39, 0, 40, 0, 39, 0, 209, 145, 39, 0, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 46, 50, 38, 0, 48, 0, 185, 112, 60, 0, 60, 0, 60, 0, 88, 51, 38, 0, 49, 0, 48, 0, 229, 101, 60, 0, 60, 0, 60, 0, 233, 51, 38, 0, 49, 0, 48, 0, 8, 103, 60, 0, 60, 0, 60, 0, 201, 50, 38, 0, 49, 0, 48, 0, 185, 112, 60, 0, 60, 0, 60, 0, 98, 51, 38, 0, 49, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 234, 51, 38, 0, 49, 0, 49, 0, 8, 103, 60, 0, 60, 0, 60, 0, 202, 50, 38, 0, 49, 0, 49, 0, 185, 112, 60, 0, 60, 0, 60, 0, 99, 51, 38, 0, 49, 0, 50, 0, 229, 101, 60, 0, 60, 0, 60, 0, 235, 51, 38, 0, 49, 0, 50, 0, 8, 103, 60, 0, 60, 0, 60, 0, 203, 50, 38, 0, 49, 0, 50, 0, 185, 112, 60, 0, 60, 0, 60, 0, 100, 51, 38, 0, 49, 0, 51, 0, 229, 101, 60, 0, 60, 0, 60, 0, 236, 51, 38, 0, 49, 0, 51, 0, 185, 112, 60, 0, 60, 0, 60, 0, 101, 51, 38, 0, 49, 0, 52, 0, 229, 101, 60, 0, 60, 0, 60, 0, 237, 51, 38, 0, 49, 0, 52, 0, 185, 112, 60, 0, 60, 0, 60, 0, 102, 51, 38, 0, 49, 0, 53, 0, 229, 101, 60, 0, 60, 0, 60, 0, 238, 51, 38, 0, 49, 0, 53, 0, 185, 112, 60, 0, 60, 0, 60, 0, 103, 51, 38, 0, 49, 0, 54, 0, 229, 101, 60, 0, 60, 0, 60, 0, 239, 51, 38, 0, 49, 0, 54, 0, 185, 112, 60, 0, 60, 0, 60, 0, 104, 51, 38, 0, 49, 0, 55, 0, 229, 101, 60, 0, 60, 0, 60, 0, 240, 51, 38, 0, 49, 0, 55, 0, 185, 112, 60, 0, 60, 0, 60, 0, 105, 51, 38, 0, 49, 0, 56, 0, 229, 101, 60, 0, 60, 0, 60, 0, 241, 51, 38, 0, 49, 0, 56, 0, 185, 112, 60, 0, 60, 0, 60, 0, 106, 51, 38, 0, 49, 0, 57, 0, 229, 101, 60, 0, 60, 0, 60, 0, 242, 51, 38, 0, 49, 0, 57, 0, 185, 112, 60, 0, 60, 0, 60, 0, 107, 51, 38, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 224, 51, 38, 0, 49, 0, 8, 103, 60, 0, 60, 0, 60, 0, 192, 50, 38, 0, 49, 0, 185, 112, 60, 0, 60, 0, 60, 0, 89, 51, 38, 0, 50, 0, 48, 0, 229, 101, 60, 0, 60, 0, 60, 0, 243, 51, 38, 0, 50, 0, 48, 0, 185, 112, 60, 0, 60, 0, 60, 0, 108, 51, 38, 0, 50, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 244, 51, 38, 0, 50, 0, 49, 0, 185, 112, 60, 0, 60, 0, 60, 0, 109, 51, 38, 0, 50, 0, 50, 0, 229, 101, 60, 0, 60, 0, 60, 0, 245, 51, 38, 0, 50, 0, 50, 0, 185, 112, 60, 0, 60, 0, 60, 0, 110, 51, 38, 0, 50, 0, 51, 0, 229, 101, 60, 0, 60, 0, 60, 0, 246, 51, 38, 0, 50, 0, 51, 0, 185, 112, 60, 0, 60, 0, 60, 0, 111, 51, 38, 0, 50, 0, 52, 0, 229, 101, 60, 0, 60, 0, 60, 0, 247, 51, 38, 0, 50, 0, 52, 0, 185, 112, 60, 0, 60, 0, 60, 0, 112, 51, 38, 0, 50, 0, 53, 0, 229, 101, 60, 0, 60, 0, 60, 0, 248, 51, 38, 0, 50, 0, 54, 0, 229, 101, 60, 0, 60, 0, 60, 0, 249, 51, 38, 0, 50, 0, 55, 0, 229, 101, 60, 0, 60, 0, 60, 0, 250, 51, 38, 0, 50, 0, 56, 0, 229, 101, 60, 0, 60, 0, 60, 0, 251, 51, 38, 0, 50, 0, 57, 0, 229, 101, 60, 0, 60, 0, 60, 0, 252, 51, 38, 0, 50, 0, 229, 101, 60, 0, 60, 0, 60, 0, 225, 51, 38, 0, 50, 0, 8, 103, 60, 0, 60, 0, 60, 0, 193, 50, 38, 0, 50, 0, 185, 112, 60, 0, 60, 0, 60, 0, 90, 51, 38, 0, 51, 0, 48, 0, 229, 101, 60, 0, 60, 0, 60, 0, 253, 51, 38, 0, 51, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 254, 51, 38, 0, 51, 0, 229, 101, 60, 0, 60, 0, 60, 0, 226, 51, 38, 0, 51, 0, 8, 103, 60, 0, 60, 0, 60, 0, 194, 50, 38, 0, 51, 0, 185, 112, 60, 0, 60, 0, 60, 0, 91, 51, 38, 0, 52, 0, 229, 101, 60, 0, 60, 0, 60, 0, 227, 51, 38, 0, 52, 0, 8, 103, 60, 0, 60, 0, 60, 0, 195, 50, 38, 0, 52, 0, 185, 112, 60, 0, 60, 0, 60, 0, 92, 51, 38, 0, 53, 0, 229, 101, 60, 0, 60, 0, 60, 0, 228, 51, 38, 0, 53, 0, 8, 103, 60, 0, 60, 0, 60, 0, 196, 50, 38, 0, 53, 0, 185, 112, 60, 0, 60, 0, 60, 0, 93, 51, 38, 0, 54, 0, 229, 101, 60, 0, 60, 0, 60, 0, 229, 51, 38, 0, 54, 0, 8, 103, 60, 0, 60, 0, 60, 0, 197, 50, 38, 0, 54, 0, 185, 112, 60, 0, 60, 0, 60, 0, 94, 51, 38, 0, 55, 0, 229, 101, 60, 0, 60, 0, 60, 0, 230, 51, 38, 0, 55, 0, 8, 103, 60, 0, 60, 0, 60, 0, 198, 50, 38, 0, 55, 0, 185, 112, 60, 0, 60, 0, 60, 0, 95, 51, 38, 0, 56, 0, 229, 101, 60, 0, 60, 0, 60, 0, 231, 51, 38, 0, 56, 0, 8, 103, 60, 0, 60, 0, 60, 0, 199, 50, 38, 0, 56, 0, 185, 112, 60, 0, 60, 0, 60, 0, 96, 51, 38, 0, 57, 0, 229, 101, 60, 0, 60, 0, 60, 0, 232, 51, 38, 0, 57, 0, 8, 103, 60, 0, 60, 0, 60, 0, 200, 50, 38, 0, 57, 0, 185, 112, 60, 0, 60, 0, 60, 0, 97, 51, 38, 0, 20, 48, 9, 78, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 65, 222, 38, 0, 20, 48, 140, 78, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 66, 222, 38, 0, 20, 48, 221, 82, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 71, 222, 38, 0, 20, 48, 137, 91, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 67, 222, 38, 0, 20, 48, 83, 98, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 69, 222, 38, 0, 20, 48, 87, 101, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 72, 222, 38, 0, 20, 48, 44, 103, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 64, 222, 38, 0, 20, 48, 185, 112, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 68, 222, 38, 0, 20, 48, 215, 118, 21, 48, 60, 0, 60, 0, 60, 0, 60, 216, 70, 222, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 0, 47, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 146, 49, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 128, 50, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 60, 216, 41, 222, 38, 0, 1, 78, 60, 0, 60, 0, 60, 0, 156, 49, 38, 0, 3, 78, 60, 0, 60, 0, 60, 0, 134, 50, 38, 0, 9, 78, 60, 0, 60, 0, 60, 0, 148, 49, 38, 0, 9, 78, 60, 0, 60, 0, 60, 0, 130, 50, 38, 0, 9, 78, 60, 0, 60, 0, 60, 0, 60, 216, 42, 222, 38, 0, 10, 78, 60, 0, 60, 0, 60, 0, 150, 49, 38, 0, 10, 78, 60, 0, 60, 0, 60, 0, 164, 50, 38, 0, 11, 78, 60, 0, 60, 0, 60, 0, 152, 49, 38, 0, 11, 78, 60, 0, 60, 0, 60, 0, 166, 50, 38, 0, 25, 78, 60, 0, 60, 0, 60, 0, 155, 49, 38, 0, 40, 78, 60, 0, 60, 0, 60, 0, 1, 47, 38, 0, 45, 78, 60, 0, 60, 0, 60, 0, 151, 49, 38, 0, 45, 78, 60, 0, 60, 0, 60, 0, 165, 50, 38, 0, 45, 78, 60, 0, 60, 0, 60, 0, 60, 216, 45, 222, 38, 0, 54, 78, 60, 0, 60, 0, 60, 0, 2, 47, 38, 0, 63, 78, 60, 0, 60, 0, 60, 0, 3, 47, 38, 0, 89, 78, 60, 0, 60, 0, 60, 0, 4, 47, 38, 0, 89, 78, 60, 0, 60, 0, 60, 0, 154, 49, 38, 0, 93, 78, 60, 0, 60, 0, 60, 0, 136, 50, 38, 0, 133, 78, 60, 0, 60, 0, 60, 0, 5, 47, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 6, 47, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 147, 49, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 129, 50, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 60, 216, 20, 222, 38, 0, 148, 78, 60, 0, 60, 0, 60, 0, 132, 50, 38, 0, 160, 78, 60, 0, 60, 0, 60, 0, 7, 47, 38, 0, 164, 78, 60, 0, 60, 0, 60, 0, 60, 216, 24, 222, 38, 0, 186, 78, 60, 0, 60, 0, 60, 0, 8, 47, 38, 0, 186, 78, 60, 0, 60, 0, 60, 0, 159, 49, 38, 0, 1, 79, 60, 0, 60, 0, 60, 0, 173, 50, 38, 0, 17, 79, 60, 0, 60, 0, 60, 0, 161, 50, 38, 0, 42, 81, 60, 0, 60, 0, 60, 0, 157, 50, 38, 0, 63, 81, 60, 0, 60, 0, 60, 0, 9, 47, 38, 0, 101, 81, 60, 0, 60, 0, 60, 0, 10, 47, 38, 0, 107, 81, 60, 0, 60, 0, 60, 0, 11, 47, 38, 0, 107, 81, 60, 0, 60, 0, 60, 0, 135, 50, 38, 0, 109, 81, 60, 0, 60, 0, 60, 0, 133, 50, 38, 0, 130, 81, 60, 0, 60, 0, 60, 0, 12, 47, 38, 0, 141, 81, 60, 0, 60, 0, 60, 0, 60, 216, 30, 222, 38, 0, 150, 81, 60, 0, 60, 0, 60, 0, 13, 47, 38, 0, 153, 81, 60, 0, 60, 0, 60, 0, 162, 50, 38, 0, 171, 81, 60, 0, 60, 0, 60, 0, 14, 47, 38, 0, 224, 81, 60, 0, 60, 0, 60, 0, 15, 47, 38, 0, 245, 81, 60, 0, 60, 0, 60, 0, 16, 47, 38, 0, 0, 82, 60, 0, 60, 0, 60, 0, 17, 47, 38, 0, 29, 82, 60, 0, 60, 0, 60, 0, 60, 216, 32, 222, 38, 0, 77, 82, 60, 0, 60, 0, 60, 0, 60, 216, 28, 222, 38, 0, 114, 82, 60, 0, 60, 0, 60, 0, 60, 216, 57, 222, 38, 0, 155, 82, 60, 0, 60, 0, 60, 0, 18, 47, 38, 0, 180, 82, 60, 0, 60, 0, 60, 0, 152, 50, 38, 0, 249, 82, 60, 0, 60, 0, 60, 0, 19, 47, 38, 0, 21, 83, 60, 0, 60, 0, 60, 0, 20, 47, 38, 0, 26, 83, 60, 0, 60, 0, 60, 0, 21, 47, 38, 0, 56, 83, 60, 0, 60, 0, 60, 0, 22, 47, 38, 0, 59, 83, 60, 0, 60, 0, 60, 0, 169, 50, 38, 0, 65, 83, 60, 0, 60, 0, 60, 0, 23, 47, 38, 0, 65, 83, 60, 0, 60, 0, 60, 0, 56, 48, 38, 0, 65, 83, 60, 0, 60, 0, 60, 0, 137, 50, 38, 0, 68, 83, 60, 0, 60, 0, 60, 0, 57, 48, 38, 0, 69, 83, 60, 0, 60, 0, 60, 0, 58, 48, 38, 0, 84, 83, 60, 0, 60, 0, 60, 0, 175, 50, 38, 0, 92, 83, 60, 0, 60, 0, 60, 0, 24, 47, 38, 0, 105, 83, 60, 0, 60, 0, 60, 0, 25, 47, 38, 0, 112, 83, 60, 0, 60, 0, 60, 0, 158, 50, 38, 0, 130, 83, 60, 0, 60, 0, 60, 0, 26, 47, 38, 0, 182, 83, 60, 0, 60, 0, 60, 0, 27, 47, 38, 0, 200, 83, 60, 0, 60, 0, 60, 0, 28, 47, 38, 0, 204, 83, 60, 0, 60, 0, 60, 0, 60, 216, 18, 222, 38, 0, 227, 83, 60, 0, 60, 0, 60, 0, 29, 47, 38, 0, 239, 83, 60, 0, 60, 0, 60, 0, 60, 216, 81, 222, 38, 0, 243, 83, 60, 0, 60, 0, 60, 0, 168, 50, 38, 0, 243, 83, 60, 0, 60, 0, 60, 0, 60, 216, 46, 222, 38, 0, 8, 84, 60, 0, 60, 0, 60, 0, 60, 216, 52, 222, 38, 0, 13, 84, 60, 0, 60, 0, 60, 0, 148, 50, 38, 0, 57, 84, 60, 0, 60, 0, 60, 0, 60, 216, 37, 222, 38, 0, 79, 85, 60, 0, 60, 0, 60, 0, 68, 50, 38, 0, 182, 85, 60, 0, 60, 0, 60, 0, 60, 216, 58, 222, 38, 0, 215, 86, 60, 0, 60, 0, 60, 0, 30, 47, 38, 0, 219, 86, 60, 0, 60, 0, 60, 0, 149, 49, 38, 0, 219, 86, 60, 0, 60, 0, 60, 0, 131, 50, 38, 0, 31, 87, 60, 0, 60, 0, 60, 0, 31, 47, 38, 0, 31, 87, 60, 0, 60, 0, 60, 0, 143, 50, 38, 0, 48, 87, 60, 0, 60, 0, 60, 0, 158, 49, 38, 0, 235, 88, 60, 0, 60, 0, 60, 0, 32, 47, 38, 0, 240, 88, 60, 0, 60, 0, 60, 0, 60, 216, 36, 222, 38, 0, 2, 89, 60, 0, 60, 0, 60, 0, 33, 47, 38, 0, 10, 89, 60, 0, 60, 0, 60, 0, 34, 47, 38, 0, 21, 89, 60, 0, 60, 0, 60, 0, 35, 47, 38, 0, 26, 89, 60, 0, 60, 0, 60, 0, 60, 216, 21, 222, 38, 0, 28, 89, 60, 0, 60, 0, 60, 0, 176, 50, 38, 0, 39, 89, 60, 0, 60, 0, 60, 0, 36, 47, 38, 0, 39, 89, 99, 107, 60, 0, 60, 0, 60, 0, 125, 51, 38, 0, 41, 89, 60, 0, 60, 0, 60, 0, 157, 49, 38, 0, 41, 89, 60, 0, 60, 0, 60, 0, 60, 216, 23, 222, 38, 0, 115, 89, 60, 0, 60, 0, 60, 0, 37, 47, 38, 0, 115, 89, 60, 0, 60, 0, 60, 0, 155, 50, 38, 0, 80, 91, 60, 0, 60, 0, 60, 0, 38, 47, 38, 0, 87, 91, 60, 0, 60, 0, 60, 0, 60, 216, 17, 222, 38, 0, 102, 91, 60, 0, 60, 0, 60, 0, 171, 50, 38, 0, 128, 91, 60, 0, 60, 0, 60, 0, 39, 47, 38, 0, 151, 91, 60, 0, 60, 0, 60, 0, 170, 50, 38, 0, 248, 91, 60, 0, 60, 0, 60, 0, 40, 47, 38, 0, 15, 92, 60, 0, 60, 0, 60, 0, 41, 47, 38, 0, 34, 92, 60, 0, 60, 0, 60, 0, 42, 47, 38, 0, 56, 92, 60, 0, 60, 0, 60, 0, 43, 47, 38, 0, 110, 92, 60, 0, 60, 0, 60, 0, 44, 47, 38, 0, 113, 92, 60, 0, 60, 0, 60, 0, 45, 47, 38, 0, 219, 93, 60, 0, 60, 0, 60, 0, 46, 47, 38, 0, 229, 93, 60, 0, 60, 0, 60, 0, 47, 47, 38, 0, 230, 93, 60, 0, 60, 0, 60, 0, 167, 50, 38, 0, 230, 93, 60, 0, 60, 0, 60, 0, 60, 216, 44, 222, 38, 0, 241, 93, 60, 0, 60, 0, 60, 0, 48, 47, 38, 0, 254, 93, 60, 0, 60, 0, 60, 0, 49, 47, 38, 0, 114, 94, 60, 0, 60, 0, 60, 0, 50, 47, 38, 0, 115, 94, 16, 98, 60, 0, 60, 0, 60, 0, 123, 51, 38, 0, 122, 94, 60, 0, 60, 0, 60, 0, 51, 47, 38, 0, 124, 94, 60, 0, 60, 0, 60, 0, 69, 50, 38, 0, 127, 94, 60, 0, 60, 0, 60, 0, 52, 47, 38, 0, 244, 94, 60, 0, 60, 0, 60, 0, 53, 47, 38, 0, 254, 94, 60, 0, 60, 0, 60, 0, 54, 47, 38, 0, 11, 95, 60, 0, 60, 0, 60, 0, 55, 47, 38, 0, 19, 95, 60, 0, 60, 0, 60, 0, 56, 47, 38, 0, 80, 95, 60, 0, 60, 0, 60, 0, 57, 47, 38, 0, 97, 95, 60, 0, 60, 0, 60, 0, 58, 47, 38, 0, 115, 95, 60, 0, 60, 0, 60, 0, 59, 47, 38, 0, 140, 95, 60, 0, 60, 0, 60, 0, 60, 216, 29, 222, 38, 0, 151, 95, 60, 0, 60, 0, 60, 0, 60, 216, 80, 222, 38, 0, 195, 95, 60, 0, 60, 0, 60, 0, 60, 47, 38, 0, 8, 98, 60, 0, 60, 0, 60, 0, 61, 47, 38, 0, 54, 98, 60, 0, 60, 0, 60, 0, 62, 47, 38, 0, 75, 98, 60, 0, 60, 0, 60, 0, 63, 47, 38, 0, 75, 98, 60, 0, 60, 0, 60, 0, 60, 216, 16, 222, 38, 0, 83, 98, 60, 0, 60, 0, 60, 0, 60, 216, 49, 222, 38, 0, 149, 98, 60, 0, 60, 0, 60, 0, 60, 216, 39, 222, 38, 0, 7, 99, 60, 0, 60, 0, 60, 0, 60, 216, 47, 222, 38, 0, 85, 99, 60, 0, 60, 0, 60, 0, 60, 216, 40, 222, 38, 0, 47, 101, 60, 0, 60, 0, 60, 0, 64, 47, 38, 0, 52, 101, 60, 0, 60, 0, 60, 0, 65, 47, 38, 0, 135, 101, 60, 0, 60, 0, 60, 0, 66, 47, 38, 0, 135, 101, 60, 0, 60, 0, 60, 0, 70, 50, 38, 0, 151, 101, 60, 0, 60, 0, 60, 0, 67, 47, 38, 0, 153, 101, 60, 0, 60, 0, 60, 0, 60, 216, 27, 222, 38, 0, 164, 101, 60, 0, 60, 0, 60, 0, 68, 47, 38, 0, 176, 101, 60, 0, 60, 0, 60, 0, 60, 216, 31, 222, 38, 0, 185, 101, 60, 0, 60, 0, 60, 0, 69, 47, 38, 0, 224, 101, 60, 0, 60, 0, 60, 0, 70, 47, 38, 0, 229, 101, 60, 0, 60, 0, 60, 0, 71, 47, 38, 0, 229, 101, 60, 0, 60, 0, 60, 0, 144, 50, 38, 0, 14, 102, 187, 108, 60, 0, 60, 0, 60, 0, 126, 51, 38, 0, 32, 102, 60, 0, 60, 0, 60, 0, 60, 216, 25, 222, 38, 0, 45, 102, 140, 84, 60, 0, 60, 0, 60, 0, 124, 51, 38, 0, 240, 102, 60, 0, 60, 0, 60, 0, 72, 47, 38, 0, 8, 103, 60, 0, 60, 0, 60, 0, 73, 47, 38, 0, 8, 103, 60, 0, 60, 0, 60, 0, 138, 50, 38, 0, 8, 103, 60, 0, 60, 0, 60, 0, 60, 216, 55, 222, 38, 0, 9, 103, 60, 0, 60, 0, 60, 0, 146, 50, 38, 0, 9, 103, 60, 0, 60, 0, 60, 0, 60, 216, 54, 222, 38, 0, 40, 103, 60, 0, 60, 0, 60, 0, 74, 47, 38, 0, 40, 103, 60, 0, 60, 0, 60, 0, 141, 50, 38, 0, 42, 104, 60, 0, 60, 0, 60, 0, 145, 50, 38, 0, 42, 104, 15, 95, 26, 79, 62, 121, 60, 0, 60, 0, 60, 0, 127, 51, 38, 0, 32, 107, 60, 0, 60, 0, 60, 0, 75, 47, 38, 0, 98, 107, 60, 0, 60, 0, 60, 0, 76, 47, 38, 0, 99, 107, 60, 0, 60, 0, 60, 0, 163, 50, 38, 0, 121, 107, 60, 0, 60, 0, 60, 0, 77, 47, 38, 0, 179, 107, 60, 0, 60, 0, 60, 0, 78, 47, 38, 0, 203, 107, 60, 0, 60, 0, 60, 0, 79, 47, 38, 0, 205, 107, 60, 0, 60, 0, 60, 0, 159, 46, 38, 0, 212, 107, 60, 0, 60, 0, 60, 0, 80, 47, 38, 0, 219, 107, 60, 0, 60, 0, 60, 0, 81, 47, 38, 0, 15, 108, 60, 0, 60, 0, 60, 0, 82, 47, 38, 0, 20, 108, 60, 0, 60, 0, 60, 0, 83, 47, 38, 0, 52, 108, 60, 0, 60, 0, 60, 0, 84, 47, 38, 0, 52, 108, 60, 0, 60, 0, 60, 0, 140, 50, 38, 0, 232, 108, 60, 0, 60, 0, 60, 0, 159, 50, 38, 0, 128, 110, 60, 0, 60, 0, 60, 0, 60, 216, 53, 222, 38, 0, 20, 111, 60, 0, 60, 0, 60, 0, 60, 216, 38, 222, 38, 0, 107, 112, 60, 0, 60, 0, 60, 0, 85, 47, 38, 0, 107, 112, 60, 0, 60, 0, 60, 0, 139, 50, 38, 0, 33, 113, 60, 0, 60, 0, 60, 0, 60, 216, 26, 222, 38, 0, 42, 114, 60, 0, 60, 0, 60, 0, 86, 47, 38, 0, 54, 114, 60, 0, 60, 0, 60, 0, 87, 47, 38, 0, 59, 114, 60, 0, 60, 0, 60, 0, 88, 47, 38, 0, 63, 114, 60, 0, 60, 0, 60, 0, 89, 47, 38, 0, 71, 114, 60, 0, 60, 0, 60, 0, 90, 47, 38, 0, 89, 114, 60, 0, 60, 0, 60, 0, 91, 47, 38, 0, 91, 114, 60, 0, 60, 0, 60, 0, 92, 47, 38, 0, 121, 114, 60, 0, 60, 0, 60, 0, 149, 50, 38, 0, 172, 114, 60, 0, 60, 0, 60, 0, 93, 47, 38, 0, 132, 115, 60, 0, 60, 0, 60, 0, 94, 47, 38, 0, 137, 115, 60, 0, 60, 0, 60, 0, 95, 47, 38, 0, 220, 116, 60, 0, 60, 0, 60, 0, 96, 47, 38, 0, 230, 116, 60, 0, 60, 0, 60, 0, 97, 47, 38, 0, 24, 117, 60, 0, 60, 0, 60, 0, 98, 47, 38, 0, 31, 117, 60, 0, 60, 0, 60, 0, 99, 47, 38, 0, 31, 117, 60, 0, 60, 0, 60, 0, 60, 216, 34, 222, 38, 0, 40, 117, 60, 0, 60, 0, 60, 0, 100, 47, 38, 0, 48, 117, 60, 0, 60, 0, 60, 0, 101, 47, 38, 0, 50, 117, 60, 0, 60, 0, 60, 0, 153, 49, 38, 0, 51, 117, 60, 0, 60, 0, 60, 0, 60, 216, 56, 222, 38, 0, 55, 117, 60, 0, 60, 0, 60, 0, 154, 50, 38, 0, 139, 117, 60, 0, 60, 0, 60, 0, 102, 47, 38, 0, 146, 117, 60, 0, 60, 0, 60, 0, 103, 47, 38, 0, 118, 118, 60, 0, 60, 0, 60, 0, 104, 47, 38, 0, 125, 118, 60, 0, 60, 0, 60, 0, 105, 47, 38, 0, 174, 118, 60, 0, 60, 0, 60, 0, 106, 47, 38, 0, 191, 118, 60, 0, 60, 0, 60, 0, 107, 47, 38, 0, 227, 118, 60, 0, 60, 0, 60, 0, 172, 50, 38, 0, 238, 118, 60, 0, 60, 0, 60, 0, 108, 47, 38, 0, 219, 119, 60, 0, 60, 0, 60, 0, 109, 47, 38, 0, 226, 119, 60, 0, 60, 0, 60, 0, 110, 47, 38, 0, 243, 119, 60, 0, 60, 0, 60, 0, 111, 47, 38, 0, 58, 121, 60, 0, 60, 0, 60, 0, 112, 47, 38, 0, 62, 121, 60, 0, 60, 0, 60, 0, 147, 50, 38, 0, 93, 121, 60, 0, 60, 0, 60, 0, 151, 50, 38, 0, 129, 121, 60, 0, 60, 0, 60, 0, 60, 216, 50, 222, 38, 0, 184, 121, 60, 0, 60, 0, 60, 0, 113, 47, 38, 0, 190, 121, 60, 0, 60, 0, 60, 0, 114, 47, 38, 0, 216, 121, 60, 0, 60, 0, 60, 0, 153, 50, 38, 0, 116, 122, 60, 0, 60, 0, 60, 0, 115, 47, 38, 0, 122, 122, 60, 0, 60, 0, 60, 0, 60, 216, 51, 222, 38, 0, 203, 122, 60, 0, 60, 0, 60, 0, 116, 47, 38, 0, 249, 122, 60, 0, 60, 0, 60, 0, 117, 47, 38, 0, 143, 123, 60, 0, 60, 0, 60, 0, 71, 50, 38, 0, 115, 124, 60, 0, 60, 0, 60, 0, 118, 47, 38, 0, 248, 124, 60, 0, 60, 0, 60, 0, 119, 47, 38, 0, 66, 125, 60, 0, 60, 0, 60, 0, 60, 216, 33, 222, 38, 0, 54, 127, 60, 0, 60, 0, 60, 0, 120, 47, 38, 0, 81, 127, 60, 0, 60, 0, 60, 0, 121, 47, 38, 0, 138, 127, 60, 0, 60, 0, 60, 0, 122, 47, 38, 0, 189, 127, 60, 0, 60, 0, 60, 0, 123, 47, 38, 0, 1, 128, 60, 0, 60, 0, 60, 0, 124, 47, 38, 0, 12, 128, 60, 0, 60, 0, 60, 0, 125, 47, 38, 0, 18, 128, 60, 0, 60, 0, 60, 0, 126, 47, 38, 0, 51, 128, 60, 0, 60, 0, 60, 0, 127, 47, 38, 0, 127, 128, 60, 0, 60, 0, 60, 0, 128, 47, 38, 0, 137, 128, 60, 0, 60, 0, 60, 0, 129, 47, 38, 0, 227, 129, 60, 0, 60, 0, 60, 0, 130, 47, 38, 0, 234, 129, 60, 0, 60, 0, 60, 0, 131, 47, 38, 0, 243, 129, 60, 0, 60, 0, 60, 0, 132, 47, 38, 0, 252, 129, 60, 0, 60, 0, 60, 0, 133, 47, 38, 0, 12, 130, 60, 0, 60, 0, 60, 0, 134, 47, 38, 0, 27, 130, 60, 0, 60, 0, 60, 0, 135, 47, 38, 0, 31, 130, 60, 0, 60, 0, 60, 0, 136, 47, 38, 0, 110, 130, 60, 0, 60, 0, 60, 0, 137, 47, 38, 0, 114, 130, 60, 0, 60, 0, 60, 0, 138, 47, 38, 0, 120, 130, 60, 0, 60, 0, 60, 0, 139, 47, 38, 0, 77, 134, 60, 0, 60, 0, 60, 0, 140, 47, 38, 0, 107, 134, 60, 0, 60, 0, 60, 0, 141, 47, 38, 0, 64, 136, 60, 0, 60, 0, 60, 0, 142, 47, 38, 0, 76, 136, 60, 0, 60, 0, 60, 0, 143, 47, 38, 0, 99, 136, 60, 0, 60, 0, 60, 0, 144, 47, 38, 0, 126, 137, 60, 0, 60, 0, 60, 0, 145, 47, 38, 0, 139, 137, 60, 0, 60, 0, 60, 0, 146, 47, 38, 0, 210, 137, 60, 0, 60, 0, 60, 0, 147, 47, 38, 0, 227, 137, 60, 0, 60, 0, 60, 0, 60, 216, 22, 222, 38, 0, 0, 138, 60, 0, 60, 0, 60, 0, 148, 47, 38, 0, 55, 140, 60, 0, 60, 0, 60, 0, 149, 47, 38, 0, 70, 140, 60, 0, 60, 0, 60, 0, 150, 47, 38, 0, 85, 140, 60, 0, 60, 0, 60, 0, 151, 47, 38, 0, 120, 140, 60, 0, 60, 0, 60, 0, 152, 47, 38, 0, 157, 140, 60, 0, 60, 0, 60, 0, 153, 47, 38, 0, 161, 140, 60, 0, 60, 0, 60, 0, 150, 50, 38, 0, 169, 140, 60, 0, 60, 0, 60, 0, 60, 216, 35, 222, 38, 0, 199, 140, 60, 0, 60, 0, 60, 0, 174, 50, 38, 0, 100, 141, 60, 0, 60, 0, 60, 0, 154, 47, 38, 0, 112, 141, 60, 0, 60, 0, 60, 0, 155, 47, 38, 0, 112, 141, 60, 0, 60, 0, 60, 0, 60, 216, 48, 222, 38, 0, 179, 141, 60, 0, 60, 0, 60, 0, 156, 47, 38, 0, 171, 142, 60, 0, 60, 0, 60, 0, 157, 47, 38, 0, 202, 142, 60, 0, 60, 0, 60, 0, 158, 47, 38, 0, 155, 143, 60, 0, 60, 0, 60, 0, 159, 47, 38, 0, 176, 143, 60, 0, 60, 0, 60, 0, 160, 47, 38, 0, 181, 143, 60, 0, 60, 0, 60, 0, 161, 47, 38, 0, 74, 144, 60, 0, 60, 0, 60, 0, 60, 216, 43, 222, 38, 0, 105, 144, 60, 0, 60, 0, 60, 0, 156, 50, 38, 0, 145, 144, 60, 0, 60, 0, 60, 0, 162, 47, 38, 0, 73, 145, 60, 0, 60, 0, 60, 0, 163, 47, 38, 0, 198, 145, 60, 0, 60, 0, 60, 0, 164, 47, 38, 0, 204, 145, 60, 0, 60, 0, 60, 0, 165, 47, 38, 0, 209, 145, 60, 0, 60, 0, 60, 0, 166, 47, 38, 0, 209, 145, 60, 0, 60, 0, 60, 0, 142, 50, 38, 0, 119, 149, 60, 0, 60, 0, 60, 0, 167, 47, 38, 0, 128, 149, 60, 0, 60, 0, 60, 0, 168, 47, 38, 0, 28, 150, 60, 0, 60, 0, 60, 0, 169, 47, 38, 0, 182, 150, 60, 0, 60, 0, 60, 0, 170, 47, 38, 0, 185, 150, 60, 0, 60, 0, 60, 0, 171, 47, 38, 0, 232, 150, 60, 0, 60, 0, 60, 0, 172, 47, 38, 0, 81, 151, 60, 0, 60, 0, 60, 0, 173, 47, 38, 0, 94, 151, 60, 0, 60, 0, 60, 0, 174, 47, 38, 0, 98, 151, 60, 0, 60, 0, 60, 0, 175, 47, 38, 0, 105, 151, 60, 0, 60, 0, 60, 0, 176, 47, 38, 0, 203, 151, 60, 0, 60, 0, 60, 0, 177, 47, 38, 0, 237, 151, 60, 0, 60, 0, 60, 0, 178, 47, 38, 0, 243, 151, 60, 0, 60, 0, 60, 0, 179, 47, 38, 0, 1, 152, 60, 0, 60, 0, 60, 0, 180, 47, 38, 0, 5, 152, 60, 0, 60, 0, 60, 0, 160, 50, 38, 0, 168, 152, 60, 0, 60, 0, 60, 0, 181, 47, 38, 0, 219, 152, 60, 0, 60, 0, 60, 0, 182, 47, 38, 0, 223, 152, 60, 0, 60, 0, 60, 0, 183, 47, 38, 0, 150, 153, 60, 0, 60, 0, 60, 0, 184, 47, 38, 0, 153, 153, 60, 0, 60, 0, 60, 0, 185, 47, 38, 0, 172, 153, 60, 0, 60, 0, 60, 0, 186, 47, 38, 0, 168, 154, 60, 0, 60, 0, 60, 0, 187, 47, 38, 0, 216, 154, 60, 0, 60, 0, 60, 0, 188, 47, 38, 0, 223, 154, 60, 0, 60, 0, 60, 0, 189, 47, 38, 0, 37, 155, 60, 0, 60, 0, 60, 0, 190, 47, 38, 0, 47, 155, 60, 0, 60, 0, 60, 0, 191, 47, 38, 0, 50, 155, 60, 0, 60, 0, 60, 0, 192, 47, 38, 0, 60, 155, 60, 0, 60, 0, 60, 0, 193, 47, 38, 0, 90, 155, 60, 0, 60, 0, 60, 0, 194, 47, 38, 0, 229, 156, 60, 0, 60, 0, 60, 0, 195, 47, 38, 0, 117, 158, 60, 0, 60, 0, 60, 0, 196, 47, 38, 0, 127, 158, 60, 0, 60, 0, 60, 0, 197, 47, 38, 0, 165, 158, 60, 0, 60, 0, 60, 0, 198, 47, 38, 0, 187, 158, 60, 0, 60, 0, 60, 0, 199, 47, 38, 0, 195, 158, 60, 0, 60, 0, 60, 0, 200, 47, 38, 0, 205, 158, 60, 0, 60, 0, 60, 0, 201, 47, 38, 0, 209, 158, 60, 0, 60, 0, 60, 0, 202, 47, 38, 0, 249, 158, 60, 0, 60, 0, 60, 0, 203, 47, 38, 0, 253, 158, 60, 0, 60, 0, 60, 0, 204, 47, 38, 0, 14, 159, 60, 0, 60, 0, 60, 0, 205, 47, 38, 0, 19, 159, 60, 0, 60, 0, 60, 0, 206, 47, 38, 0, 32, 159, 60, 0, 60, 0, 60, 0, 207, 47, 38, 0, 59, 159, 60, 0, 60, 0, 60, 0, 208, 47, 38, 0, 74, 159, 60, 0, 60, 0, 60, 0, 209, 47, 38, 0, 82, 159, 60, 0, 60, 0, 60, 0, 210, 47, 38, 0, 141, 159, 60, 0, 60, 0, 60, 0, 211, 47, 38, 0, 156, 159, 60, 0, 60, 0, 60, 0, 212, 47, 38, 0, 159, 159, 60, 0, 60, 0, 60, 0, 243, 46, 38, 0, 160, 159, 60, 0, 60, 0, 60, 0, 213, 47, 0, 0, 239, 223, 24, 114, 91, 0, 105, 0, 109, 0, 112, 0, 111, 0, 114, 0, 116, 0, 32, 0, 122, 0, 104, 0, 45, 0, 117, 0, 45, 0, 99, 0, 111, 0, 45, 0, 112, 0, 114, 0, 105, 0, 118, 0, 97, 0, 116, 0, 101, 0, 45, 0, 112, 0, 105, 0, 110, 0, 121, 0, 105, 0, 110, 0, 93, 0, 91, 0, 114, 0, 101, 0, 111, 0, 114, 0, 100, 0, 101, 0, 114, 0, 32, 0, 72, 0, 97, 0, 110, 0, 105, 0, 32, 0, 66, 0, 111, 0, 112, 0, 111, 0, 93, 0, 38, 0, 91, 0, 108, 0, 97, 0, 115, 0, 116, 0, 32, 0, 114, 0, 101, 0, 103, 0, 117, 0, 108, 0, 97, 0, 114, 0, 93, 0, 60, 0, 39, 0, 208, 253, 1, 40, 39, 0, 60, 0, 42, 0, 0, 78, 40, 78, 54, 78, 63, 78, 64, 78, 65, 78, 89, 78, 90, 78, 132, 46, 91, 78, 64, 216, 202, 220, 64, 216, 203, 220, 64, 216, 204, 220, 64, 216, 205, 220, 64, 216, 209, 220, 133, 78, 64, 216, 12, 221, 6, 48, 7, 48, 33, 48, 37, 48, 59, 48, 60, 0, 39, 0, 208, 253, 2, 40, 39, 0, 60, 0, 42, 0, 1, 78, 2, 78, 3, 78, 4, 78, 5, 78, 6, 78, 41, 78, 55, 78, 66, 78, 67, 78, 68, 78, 64, 216, 134, 220, 64, 216, 135, 220, 64, 216, 138, 220, 92, 78, 93, 78, 134, 78, 64, 216, 14, 221, 140, 78, 160, 78, 186, 78, 187, 78, 63, 81, 101, 81, 107, 81, 134, 46, 130, 81, 150, 81, 171, 81, 224, 81, 135, 46, 245, 81, 0, 82, 1, 82, 136, 46, 2, 82, 155, 82, 249, 82, 21, 83, 26, 83, 56, 83, 65, 83, 138, 46, 92, 83, 105, 83, 130, 83, 182, 83, 128, 46, 200, 83, 220, 93, 160, 139, 207, 46, 214, 46, 99, 216, 15, 222, 34, 48, 36, 48, 38, 48, 60, 0, 39, 0, 208, 253, 3, 40, 39, 0, 60, 0, 42, 0, 7, 78, 8, 78, 9, 78, 10, 78, 11, 78, 12, 78, 144, 78, 68, 83, 42, 78, 43, 78, 4, 52, 56, 78, 73, 78, 69, 78, 70, 78, 71, 78, 72, 78, 74, 78, 94, 78, 95, 78, 96, 78, 135, 78, 141, 78, 142, 78, 143, 78, 161, 78, 188, 78, 189, 78, 190, 78, 191, 78, 64, 216, 164, 221, 64, 81, 102, 81, 225, 81, 226, 81, 227, 81, 3, 82, 4, 82, 156, 82, 250, 82, 66, 83, 67, 83, 106, 83, 107, 83, 62, 53, 201, 83, 227, 83, 215, 86, 31, 87, 235, 88, 2, 89, 10, 89, 21, 89, 39, 89, 40, 89, 115, 89, 80, 91, 81, 91, 82, 91, 83, 91, 128, 91, 248, 91, 140, 46, 141, 46, 15, 92, 34, 92, 70, 216, 193, 223, 70, 216, 194, 223, 56, 92, 110, 92, 113, 92, 219, 93, 221, 93, 71, 216, 232, 223, 229, 93, 241, 93, 242, 93, 243, 93, 254, 93, 114, 94, 97, 78, 122, 94, 127, 94, 244, 94, 254, 94, 11, 95, 19, 95, 80, 95, 149, 46, 81, 95, 97, 95, 115, 95, 196, 95, 76, 98, 77, 98, 53, 108, 173, 114, 159, 126, 190, 46, 121, 130, 204, 46, 232, 149, 29, 150, 222, 152, 99, 153, 108, 154, 5, 48, 35, 48, 39, 48, 60, 0, 39, 0, 208, 253, 4, 40, 39, 0, 60, 0, 42, 0, 13, 78, 14, 78, 15, 78, 16, 78, 17, 78, 18, 78, 19, 78, 45, 78, 46, 78, 47, 78, 48, 78, 57, 78, 58, 78, 75, 78, 76, 78, 57, 92, 99, 78, 100, 78, 101, 78, 102, 78, 136, 78, 145, 78, 146, 78, 147, 78, 148, 78, 149, 78, 150, 78, 162, 78, 163, 78, 192, 78, 193, 78, 194, 78, 195, 78, 196, 78, 197, 78, 198, 78, 199, 78, 200, 78, 201, 78, 202, 78, 203, 78, 204, 78, 205, 78, 206, 78, 207, 78, 208, 78, 209, 78, 210, 78, 211, 78, 65, 81, 66, 81, 67, 81, 103, 81, 108, 81, 109, 81, 110, 81, 111, 81, 131, 81, 132, 81, 133, 81, 134, 81, 135, 81, 136, 81, 151, 81, 152, 81, 193, 52, 228, 81, 65, 216, 48, 222, 246, 81, 65, 216, 118, 222, 5, 82, 6, 82, 7, 82, 8, 82, 157, 82, 158, 82, 251, 82, 252, 82, 253, 82, 254, 82, 255, 82, 0, 83, 1, 83, 2, 83, 22, 83, 43, 53, 57, 83, 58, 83, 69, 83, 70, 83, 71, 83, 72, 83, 57, 53, 93, 83, 94, 83, 108, 83, 131, 83, 132, 83, 133, 83, 134, 83, 183, 83, 184, 83, 185, 83, 202, 83, 203, 83, 204, 83, 205, 83, 206, 83, 32, 87, 33, 87, 68, 216, 60, 222, 236, 88, 3, 89, 41, 89, 42, 89, 43, 89, 44, 89, 45, 89, 84, 91, 16, 92, 17, 92, 35, 92, 36, 92, 58, 92, 111, 92, 98, 78, 114, 92, 244, 93, 60, 0, 42, 0, 255, 93, 0, 94, 1, 94, 123, 94, 255, 94, 0, 95, 12, 95, 20, 95, 21, 95, 22, 95, 73, 216, 188, 220, 195, 95, 198, 95, 8, 98, 54, 98, 55, 98, 56, 98, 75, 98, 78, 98, 74, 216, 102, 222, 47, 101, 52, 101, 53, 101, 135, 101, 151, 101, 164, 101, 185, 101, 224, 101, 225, 101, 156, 46, 229, 101, 240, 102, 157, 46, 8, 103, 40, 103, 41, 103, 76, 216, 180, 223, 32, 107, 98, 107, 176, 159, 121, 107, 179, 107, 203, 107, 204, 107, 212, 107, 219, 107, 15, 108, 20, 108, 52, 108, 107, 112, 108, 112, 42, 114, 165, 46, 43, 114, 54, 114, 59, 114, 44, 78, 63, 114, 71, 114, 89, 114, 38, 62, 167, 46, 91, 114, 92, 114, 172, 114, 139, 115, 82, 216, 233, 220, 59, 121, 85, 216, 5, 222, 83, 127, 2, 128, 128, 128, 188, 46, 193, 137, 161, 139, 162, 139, 163, 139, 164, 139, 165, 139, 29, 141, 102, 143, 205, 46, 182, 143, 233, 149, 230, 151, 206, 152, 40, 48, 41, 48, 60, 0, 39, 0, 208, 253, 5, 40, 39, 0, 60, 0, 42, 0, 23, 78, 20, 78, 21, 78, 22, 78, 24, 78, 25, 78, 26, 78, 27, 78, 28, 78, 29, 78, 0, 52, 49, 78, 59, 78, 60, 78, 77, 78, 78, 78, 79, 78, 80, 78, 64, 216, 148, 220, 103, 78, 151, 78, 212, 78, 213, 78, 214, 78, 215, 78, 216, 78, 217, 78, 218, 78, 219, 78, 220, 78, 221, 78, 222, 78, 223, 78, 224, 78, 225, 78, 226, 78, 227, 78, 228, 78, 229, 78, 230, 78, 231, 78, 232, 78, 233, 78, 234, 78, 235, 78, 236, 78, 237, 78, 48, 52, 49, 52, 50, 52, 51, 52, 52, 52, 53, 52, 54, 52, 55, 52, 64, 216, 169, 221, 64, 216, 171, 221, 68, 81, 69, 81, 176, 52, 112, 81, 137, 81, 138, 81, 139, 81, 140, 81, 65, 216, 71, 221, 153, 81, 154, 81, 172, 81, 173, 81, 174, 81, 175, 81, 229, 81, 230, 81, 231, 81, 247, 81, 248, 81, 249, 81, 250, 81, 251, 81, 9, 82, 10, 82, 11, 82, 12, 82, 13, 82, 159, 82, 160, 82, 161, 82, 162, 82, 3, 83, 4, 83, 5, 83, 6, 83, 7, 83, 23, 83, 27, 83, 28, 83, 29, 83, 30, 83, 73, 83, 74, 83, 76, 83, 95, 83, 96, 83, 97, 83, 98, 83, 109, 83, 110, 83, 111, 83, 66, 216, 17, 222, 135, 83, 136, 83, 137, 83, 186, 83, 187, 83, 188, 83, 207, 83, 208, 83, 209, 83, 228, 83, 229, 83, 230, 83, 231, 83, 232, 83, 233, 83, 234, 83, 235, 83, 236, 83, 60, 0, 42, 0, 237, 83, 238, 83, 239, 83, 240, 83, 241, 83, 242, 83, 243, 83, 244, 83, 245, 83, 246, 83, 247, 83, 248, 83, 249, 83, 250, 83, 251, 83, 252, 83, 253, 83, 254, 83, 66, 216, 159, 223, 66, 216, 168, 223, 66, 216, 169, 223, 216, 86, 217, 86, 218, 86, 219, 86, 220, 86, 177, 159, 34, 87, 35, 87, 36, 87, 37, 87, 38, 87, 39, 87, 38, 54, 237, 88, 4, 89, 22, 89, 23, 89, 24, 89, 46, 89, 47, 89, 48, 89, 49, 89, 50, 89, 51, 89, 52, 89, 116, 89, 117, 89, 118, 89, 85, 91, 129, 91, 130, 91, 131, 91, 132, 91, 249, 91, 18, 92, 19, 92, 20, 92, 21, 92, 59, 92, 60, 92, 115, 92, 116, 92, 117, 92, 118, 92, 119, 92, 230, 93, 231, 93, 232, 93, 2, 94, 3, 94, 4, 94, 5, 94, 50, 56, 51, 56, 115, 94, 124, 94, 128, 94, 129, 94, 130, 94, 131, 94, 1, 95, 13, 95, 23, 95, 24, 95, 82, 95, 212, 56, 197, 95, 250, 56, 199, 95, 201, 95, 202, 95, 73, 216, 175, 221, 9, 98, 10, 98, 11, 98, 57, 98, 80, 98, 81, 98, 82, 98, 83, 98, 84, 98, 85, 98, 86, 98, 79, 98, 165, 101, 230, 101, 231, 101, 76, 216, 61, 221, 241, 102, 42, 103, 43, 103, 44, 103, 45, 103, 46, 103, 47, 103, 48, 103, 99, 107, 122, 107, 205, 107, 16, 108, 17, 108, 21, 108, 58, 108, 54, 108, 55, 108, 56, 108, 57, 108, 60, 0, 42, 0, 59, 108, 62, 108, 63, 108, 64, 108, 65, 108, 67, 108, 68, 108, 69, 108, 71, 108, 72, 108, 73, 108, 109, 112, 174, 114, 175, 114, 176, 114, 132, 115, 137, 115, 138, 115, 140, 115, 141, 115, 220, 116, 230, 116, 24, 117, 31, 117, 40, 117, 41, 117, 48, 117, 49, 117, 50, 117, 51, 117, 52, 117, 53, 117, 139, 117, 83, 216, 19, 221, 170, 46, 146, 117, 118, 118, 125, 118, 174, 118, 191, 118, 238, 118, 219, 119, 226, 119, 243, 119, 58, 121, 172, 46, 184, 121, 190, 121, 116, 122, 203, 122, 160, 126, 82, 127, 88, 216, 106, 222, 88, 216, 107, 222, 138, 128, 122, 130, 100, 136, 93, 216, 7, 222, 166, 139, 167, 139, 168, 139, 169, 139, 170, 139, 171, 139, 172, 139, 173, 139, 174, 139, 175, 139, 176, 139, 177, 139, 103, 143, 183, 143, 146, 144, 147, 144, 133, 148, 127, 149, 234, 149, 30, 150, 31, 150, 100, 153, 101, 153, 109, 154, 31, 158, 153, 159, 60, 0, 39, 0, 208, 253, 6, 40, 39, 0, 60, 0, 42, 0, 30, 78, 31, 78, 32, 78, 33, 78, 34, 78, 1, 52, 81, 78, 82, 78, 83, 78, 84, 78, 104, 78, 105, 78, 106, 78, 107, 78, 108, 78, 109, 78, 110, 78, 111, 78, 112, 78, 137, 78, 152, 78, 153, 78, 154, 78, 164, 78, 165, 78, 166, 78, 167, 78, 43, 52, 238, 78, 239, 78, 240, 78, 241, 78, 242, 78, 243, 78, 244, 78, 245, 78, 246, 78, 247, 78, 248, 78, 249, 78, 250, 78, 251, 78, 252, 78, 253, 78, 254, 78, 255, 78, 0, 79, 1, 79, 2, 79, 3, 79, 4, 79, 5, 79, 6, 79, 7, 79, 8, 79, 9, 79, 10, 79, 11, 79, 12, 79, 13, 79, 14, 79, 15, 79, 16, 79, 17, 79, 18, 79, 19, 79, 20, 79, 21, 79, 22, 79, 23, 79, 24, 79, 25, 79, 26, 79, 27, 79, 28, 79, 29, 79, 30, 79, 31, 79, 32, 79, 33, 79, 34, 79, 35, 79, 36, 79, 37, 79, 38, 79, 39, 79, 40, 79, 41, 79, 42, 79, 43, 79, 44, 79, 100, 79, 56, 52, 57, 52, 59, 52, 63, 52, 64, 52, 64, 216, 193, 221, 64, 216, 212, 221, 70, 81, 71, 81, 72, 81, 73, 81, 74, 81, 104, 81, 61, 108, 113, 81, 114, 81, 115, 81, 116, 81, 141, 81, 142, 81, 155, 81, 156, 81, 176, 81, 177, 81, 178, 81, 179, 81, 180, 81, 65, 216, 179, 221, 232, 81, 233, 81, 234, 81, 235, 81, 252, 81, 14, 82, 15, 82, 16, 82, 60, 0, 42, 0, 17, 82, 18, 82, 19, 82, 20, 82, 21, 82, 22, 82, 23, 82, 24, 82, 25, 82, 26, 82, 27, 82, 163, 82, 164, 82, 165, 82, 166, 82, 167, 82, 168, 82, 8, 83, 31, 83, 32, 83, 33, 83, 34, 83, 75, 83, 77, 83, 78, 83, 79, 83, 80, 83, 59, 53, 112, 83, 113, 83, 138, 83, 139, 83, 140, 83, 141, 83, 66, 53, 189, 83, 190, 83, 210, 83, 255, 83, 0, 84, 1, 84, 2, 84, 3, 84, 4, 84, 5, 84, 6, 84, 7, 84, 8, 84, 9, 84, 10, 84, 11, 84, 12, 84, 13, 84, 14, 84, 15, 84, 16, 84, 17, 84, 18, 84, 19, 84, 20, 84, 21, 84, 22, 84, 23, 84, 102, 53, 66, 216, 191, 223, 66, 216, 198, 223, 221, 86, 222, 86, 223, 86, 224, 86, 225, 86, 226, 86, 227, 86, 40, 87, 41, 87, 42, 87, 43, 87, 44, 87, 45, 87, 46, 87, 47, 87, 48, 87, 49, 87, 50, 87, 51, 87, 52, 87, 53, 87, 54, 87, 55, 87, 56, 87, 57, 87, 58, 87, 68, 216, 79, 222, 238, 88, 5, 89, 25, 89, 26, 89, 27, 89, 53, 89, 54, 89, 55, 89, 56, 89, 57, 89, 58, 89, 59, 89, 60, 89, 119, 89, 120, 89, 121, 89, 122, 89, 123, 89, 124, 89, 125, 89, 126, 89, 127, 89, 128, 89, 129, 89, 130, 89, 131, 89, 132, 89, 133, 89, 134, 89, 135, 89, 136, 89, 165, 54, 69, 216, 184, 222, 69, 216, 186, 222, 86, 91, 60, 0, 42, 0, 87, 91, 88, 91, 89, 91, 70, 216, 68, 221, 133, 91, 134, 91, 135, 91, 136, 91, 137, 91, 250, 91, 251, 91, 252, 91, 22, 92, 23, 92, 24, 92, 37, 92, 38, 92, 39, 92, 61, 92, 71, 216, 42, 220, 112, 92, 120, 92, 121, 92, 122, 92, 123, 92, 124, 92, 125, 92, 126, 92, 127, 92, 128, 92, 129, 92, 130, 92, 131, 92, 71, 216, 70, 221, 222, 93, 223, 93, 233, 93, 234, 93, 6, 94, 7, 94, 8, 94, 52, 56, 53, 56, 54, 56, 116, 94, 117, 94, 118, 94, 132, 94, 133, 94, 134, 94, 245, 94, 2, 95, 14, 95, 15, 95, 16, 95, 25, 95, 26, 95, 27, 95, 28, 95, 83, 95, 116, 95, 117, 95, 73, 216, 193, 220, 200, 95, 203, 95, 207, 95, 211, 95, 212, 95, 213, 95, 214, 95, 217, 95, 218, 95, 219, 95, 73, 216, 190, 221, 12, 98, 13, 98, 14, 98, 15, 98, 16, 98, 87, 98, 88, 98, 89, 98, 90, 98, 91, 98, 92, 98, 93, 98, 94, 98, 96, 98, 97, 98, 98, 98, 99, 98, 100, 98, 101, 98, 102, 98, 103, 98, 104, 98, 105, 98, 106, 98, 107, 98, 108, 98, 95, 98, 48, 101, 54, 101, 55, 101, 232, 101, 233, 101, 234, 101, 235, 101, 236, 101, 237, 101, 238, 101, 239, 101, 242, 102, 243, 102, 48, 59, 9, 103, 49, 103, 50, 103, 51, 103, 52, 103, 53, 103, 54, 103, 55, 103, 56, 103, 57, 103, 58, 103, 59, 103, 60, 103, 60, 0, 42, 0, 61, 103, 62, 103, 63, 103, 64, 103, 65, 103, 66, 103, 67, 103, 65, 59, 33, 107, 34, 107, 100, 107, 123, 107, 206, 107, 213, 107, 18, 108, 22, 108, 23, 108, 24, 108, 60, 108, 70, 108, 74, 108, 75, 108, 76, 108, 77, 108, 78, 108, 79, 108, 80, 108, 81, 108, 82, 108, 83, 108, 84, 108, 85, 108, 87, 108, 88, 108, 89, 108, 90, 108, 91, 108, 92, 108, 93, 108, 95, 108, 96, 108, 97, 108, 98, 108, 99, 108, 100, 108, 119, 108, 110, 112, 111, 112, 112, 112, 113, 112, 114, 112, 115, 112, 55, 114, 93, 114, 94, 114, 95, 114, 177, 114, 178, 114, 179, 114, 180, 114, 181, 114, 183, 114, 184, 114, 81, 216, 37, 223, 142, 115, 143, 115, 144, 115, 145, 115, 168, 62, 169, 62, 170, 62, 42, 117, 54, 117, 126, 118, 127, 118, 174, 46, 60, 121, 117, 122, 249, 122, 115, 124, 248, 124, 249, 124, 161, 126, 162, 126, 163, 126, 164, 126, 165, 126, 166, 126, 167, 126, 168, 126, 169, 126, 170, 126, 171, 126, 54, 127, 81, 127, 138, 127, 88, 216, 75, 223, 182, 46, 88, 216, 76, 223, 189, 127, 1, 128, 3, 128, 12, 128, 18, 128, 51, 128, 127, 128, 89, 216, 18, 222, 137, 128, 139, 128, 140, 128, 141, 128, 142, 128, 227, 129, 234, 129, 243, 129, 252, 129, 90, 216, 81, 221, 12, 130, 27, 130, 31, 130, 110, 130, 114, 130, 120, 130, 145, 68, 123, 130, 124, 130, 125, 130, 126, 130, 60, 0, 42, 0, 127, 130, 128, 130, 129, 130, 130, 130, 77, 134, 107, 134, 64, 136, 76, 136, 99, 136, 126, 137, 127, 137, 128, 137, 194, 137, 178, 139, 179, 139, 180, 139, 181, 139, 182, 139, 183, 139, 184, 139, 185, 139, 186, 139, 187, 139, 188, 139, 189, 139, 190, 139, 191, 139, 192, 139, 30, 141, 31, 141, 32, 141, 113, 141, 104, 143, 184, 143, 185, 143, 186, 143, 187, 143, 188, 143, 189, 143, 36, 250, 148, 144, 150, 144, 151, 144, 152, 144, 153, 144, 154, 144, 155, 144, 156, 144, 157, 144, 134, 148, 135, 148, 235, 149, 236, 149, 237, 149, 238, 149, 239, 149, 32, 150, 33, 150, 34, 150, 35, 150, 36, 150, 117, 152, 102, 153, 103, 153, 110, 154, 111, 154, 112, 154, 80, 159, 60, 0, 39, 0, 208, 253, 7, 40, 39, 0, 60, 0, 42, 0, 35, 78, 36, 78, 37, 78, 61, 78, 64, 216, 33, 220, 50, 78, 64, 216, 104, 220, 85, 78, 113, 78, 114, 78, 138, 78, 64, 216, 24, 221, 156, 78, 168, 78, 169, 78, 170, 78, 45, 79, 46, 79, 47, 79, 48, 79, 49, 79, 50, 79, 51, 79, 52, 79, 53, 79, 54, 79, 55, 79, 56, 79, 57, 79, 58, 79, 59, 79, 60, 79, 61, 79, 62, 79, 63, 79, 64, 79, 65, 79, 66, 79, 67, 79, 68, 79, 69, 79, 70, 79, 71, 79, 72, 79, 73, 79, 74, 79, 75, 79, 77, 79, 78, 79, 79, 79, 80, 79, 81, 79, 82, 79, 83, 79, 84, 79, 85, 79, 86, 79, 87, 79, 88, 79, 89, 79, 90, 79, 91, 79, 92, 79, 93, 79, 94, 79, 95, 79, 96, 79, 97, 79, 98, 79, 99, 79, 101, 79, 102, 79, 103, 79, 104, 79, 70, 52, 64, 216, 242, 221, 75, 81, 76, 81, 77, 81, 78, 81, 79, 81, 80, 81, 81, 81, 179, 52, 117, 81, 143, 81, 157, 81, 181, 81, 182, 81, 183, 81, 184, 81, 185, 81, 186, 81, 187, 81, 200, 52, 236, 81, 28, 82, 29, 82, 30, 82, 31, 82, 32, 82, 33, 82, 34, 82, 35, 82, 36, 82, 37, 82, 38, 82, 39, 82, 40, 82, 41, 82, 42, 82, 43, 82, 44, 82, 45, 82, 223, 52, 169, 82, 170, 82, 171, 82, 172, 82, 173, 82, 174, 82, 175, 82, 176, 82, 177, 82, 178, 82, 179, 82, 180, 82, 9, 83, 60, 0, 42, 0, 66, 216, 213, 220, 35, 83, 36, 83, 37, 83, 48, 53, 59, 83, 55, 53, 99, 83, 100, 83, 114, 83, 115, 83, 116, 83, 117, 83, 142, 83, 143, 83, 144, 83, 145, 83, 191, 83, 211, 83, 24, 84, 25, 84, 26, 84, 27, 84, 28, 84, 29, 84, 30, 84, 31, 84, 32, 84, 33, 84, 34, 84, 35, 84, 36, 84, 37, 84, 38, 84, 39, 84, 40, 84, 41, 84, 42, 84, 43, 84, 44, 84, 45, 84, 46, 84, 47, 84, 48, 84, 49, 84, 50, 84, 51, 84, 52, 84, 53, 84, 54, 84, 55, 84, 56, 84, 57, 84, 58, 84, 59, 84, 60, 84, 61, 84, 62, 84, 63, 84, 64, 84, 65, 84, 66, 84, 67, 84, 68, 84, 69, 84, 70, 84, 71, 84, 72, 84, 73, 84, 74, 84, 75, 84, 76, 84, 77, 84, 78, 84, 79, 84, 80, 84, 81, 84, 82, 84, 83, 84, 84, 84, 85, 84, 86, 84, 87, 84, 88, 84, 89, 84, 90, 84, 91, 84, 92, 84, 109, 53, 112, 53, 114, 53, 66, 216, 203, 223, 66, 216, 226, 223, 66, 216, 235, 223, 66, 216, 251, 223, 66, 216, 255, 223, 228, 86, 229, 86, 230, 86, 231, 86, 232, 86, 233, 86, 234, 86, 235, 86, 236, 86, 237, 86, 238, 86, 239, 86, 240, 86, 241, 86, 242, 86, 243, 86, 244, 86, 245, 86, 59, 87, 60, 87, 61, 87, 62, 87, 63, 87, 64, 87, 65, 87, 66, 87, 67, 87, 68, 87, 69, 87, 70, 87, 71, 87, 72, 87, 60, 0, 42, 0, 73, 87, 74, 87, 75, 87, 76, 87, 77, 87, 78, 87, 79, 87, 80, 87, 81, 87, 82, 87, 83, 87, 84, 87, 85, 87, 86, 87, 87, 87, 88, 87, 89, 87, 90, 87, 91, 87, 92, 87, 93, 87, 94, 87, 95, 87, 96, 87, 41, 54, 43, 54, 45, 54, 46, 54, 48, 54, 68, 216, 124, 222, 239, 88, 240, 88, 241, 88, 242, 88, 243, 88, 6, 89, 11, 89, 61, 89, 62, 89, 63, 89, 64, 89, 65, 89, 66, 89, 137, 89, 138, 89, 139, 89, 140, 89, 142, 89, 143, 89, 144, 89, 145, 89, 146, 89, 147, 89, 148, 89, 149, 89, 150, 89, 151, 89, 152, 89, 153, 89, 154, 89, 155, 89, 156, 89, 157, 89, 158, 89, 159, 89, 160, 89, 161, 89, 162, 89, 163, 89, 164, 89, 165, 89, 166, 89, 167, 89, 168, 89, 169, 89, 170, 89, 171, 89, 170, 54, 172, 54, 69, 216, 192, 222, 69, 216, 193, 222, 69, 216, 194, 222, 69, 216, 211, 222, 69, 216, 213, 222, 90, 91, 91, 91, 92, 91, 93, 91, 94, 91, 138, 91, 139, 91, 140, 91, 141, 91, 142, 91, 143, 91, 144, 91, 145, 91, 146, 91, 253, 91, 254, 91, 255, 91, 40, 92, 41, 92, 42, 92, 43, 92, 44, 92, 62, 92, 63, 92, 64, 92, 65, 92, 66, 92, 67, 92, 132, 92, 133, 92, 134, 92, 135, 92, 136, 92, 137, 92, 138, 92, 139, 92, 140, 92, 142, 92, 143, 92, 144, 92, 145, 92, 146, 92, 147, 92, 148, 92, 60, 0, 42, 0, 149, 92, 150, 92, 151, 92, 152, 92, 153, 92, 154, 92, 155, 92, 156, 92, 71, 216, 83, 221, 71, 216, 94, 221, 141, 92, 224, 93, 235, 93, 245, 93, 9, 94, 10, 94, 11, 94, 12, 94, 13, 94, 14, 94, 15, 94, 16, 94, 55, 56, 56, 56, 57, 56, 59, 56, 135, 94, 136, 94, 137, 94, 138, 94, 139, 94, 140, 94, 141, 94, 142, 94, 143, 94, 144, 94, 145, 94, 146, 94, 147, 94, 148, 94, 247, 94, 3, 95, 4, 95, 5, 95, 29, 95, 30, 95, 31, 95, 32, 95, 72, 216, 189, 223, 98, 95, 99, 95, 100, 95, 118, 95, 119, 95, 120, 95, 121, 95, 122, 95, 123, 95, 200, 159, 201, 159, 73, 216, 201, 220, 73, 216, 204, 220, 204, 95, 205, 95, 206, 95, 208, 95, 209, 95, 210, 95, 215, 95, 216, 95, 220, 95, 0, 57, 223, 95, 225, 95, 227, 95, 228, 95, 230, 95, 231, 95, 232, 95, 234, 95, 235, 95, 236, 95, 237, 95, 238, 95, 239, 95, 240, 95, 241, 95, 242, 95, 243, 95, 244, 95, 246, 95, 247, 95, 248, 95, 249, 95, 250, 95, 251, 95, 252, 95, 254, 95, 0, 96, 3, 96, 4, 96, 5, 96, 6, 96, 8, 57, 17, 98, 18, 98, 19, 98, 74, 216, 147, 221, 58, 98, 59, 98, 60, 98, 109, 98, 110, 98, 111, 98, 112, 98, 113, 98, 114, 98, 115, 98, 116, 98, 117, 98, 118, 98, 119, 98, 121, 98, 122, 98, 123, 98, 124, 98, 125, 98, 126, 98, 60, 0, 42, 0, 128, 98, 129, 98, 130, 98, 131, 98, 132, 98, 133, 98, 134, 98, 135, 98, 136, 98, 137, 98, 138, 98, 139, 98, 140, 98, 141, 98, 142, 98, 143, 98, 144, 98, 145, 98, 146, 98, 147, 98, 148, 98, 149, 98, 150, 98, 151, 98, 152, 98, 153, 98, 154, 98, 155, 98, 156, 98, 157, 98, 158, 98, 159, 98, 160, 98, 161, 98, 162, 98, 163, 98, 164, 98, 165, 98, 209, 57, 120, 98, 56, 101, 57, 101, 58, 101, 59, 101, 60, 101, 75, 216, 239, 222, 136, 101, 152, 101, 240, 101, 241, 101, 242, 101, 243, 101, 244, 101, 245, 101, 246, 101, 247, 101, 248, 101, 244, 102, 245, 102, 68, 103, 69, 103, 70, 103, 71, 103, 72, 103, 73, 103, 74, 103, 75, 103, 76, 103, 77, 103, 78, 103, 79, 103, 80, 103, 81, 103, 82, 103, 83, 103, 84, 103, 85, 103, 86, 103, 87, 103, 88, 103, 89, 103, 90, 103, 91, 103, 92, 103, 93, 103, 94, 103, 95, 103, 96, 103, 97, 103, 98, 103, 99, 103, 100, 103, 101, 103, 102, 103, 103, 103, 104, 103, 105, 103, 129, 103, 66, 59, 67, 59, 68, 59, 69, 59, 70, 59, 76, 216, 204, 223, 36, 107, 101, 107, 80, 60, 124, 107, 78, 216, 194, 221, 207, 107, 208, 107, 78, 216, 238, 222, 220, 107, 221, 107, 25, 108, 26, 108, 66, 108, 86, 108, 94, 108, 101, 108, 102, 108, 103, 108, 104, 108, 105, 108, 106, 108, 107, 108, 109, 108, 110, 108, 60, 0, 42, 0, 111, 108, 112, 108, 113, 108, 114, 108, 115, 108, 116, 108, 117, 108, 118, 108, 120, 108, 121, 108, 122, 108, 123, 108, 124, 108, 125, 108, 126, 108, 127, 108, 129, 108, 130, 108, 131, 108, 132, 108, 133, 108, 134, 108, 135, 108, 136, 108, 137, 108, 139, 108, 140, 108, 141, 108, 142, 108, 143, 108, 144, 108, 145, 108, 146, 108, 148, 108, 149, 108, 150, 108, 152, 108, 153, 108, 154, 108, 155, 108, 156, 108, 158, 108, 159, 108, 160, 108, 161, 108, 162, 108, 163, 108, 164, 108, 165, 108, 166, 108, 167, 108, 168, 108, 169, 108, 170, 108, 79, 216, 153, 220, 79, 216, 154, 220, 79, 216, 155, 220, 116, 112, 117, 112, 118, 112, 119, 112, 120, 112, 121, 112, 122, 112, 123, 112, 124, 112, 125, 112, 126, 112, 127, 112, 128, 112, 165, 61, 96, 114, 97, 114, 98, 114, 99, 114, 100, 114, 81, 216, 24, 222, 182, 114, 185, 114, 186, 114, 187, 114, 188, 114, 189, 114, 190, 114, 191, 114, 193, 114, 194, 114, 195, 114, 196, 114, 197, 114, 198, 114, 199, 114, 200, 114, 81, 216, 47, 223, 146, 115, 147, 115, 148, 115, 149, 115, 150, 115, 151, 115, 152, 115, 153, 115, 154, 115, 155, 115, 173, 62, 82, 216, 240, 220, 82, 216, 241, 220, 82, 216, 242, 220, 82, 216, 243, 220, 231, 116, 43, 117, 44, 117, 55, 117, 56, 117, 57, 117, 58, 117, 60, 117, 85, 63, 147, 117, 148, 117, 149, 117, 150, 117, 151, 117, 128, 118, 129, 118, 130, 118, 131, 118, 192, 118, 60, 0, 42, 0, 193, 118, 239, 118, 227, 119, 244, 119, 245, 119, 246, 119, 85, 216, 25, 220, 61, 121, 191, 121, 192, 121, 193, 121, 194, 121, 195, 121, 38, 65, 118, 122, 119, 122, 204, 122, 205, 122, 250, 124, 251, 124, 182, 126, 172, 126, 173, 126, 174, 126, 175, 126, 176, 126, 177, 126, 178, 126, 179, 126, 180, 126, 181, 126, 183, 126, 184, 126, 185, 126, 186, 126, 187, 126, 188, 126, 189, 126, 190, 126, 85, 127, 52, 128, 144, 128, 145, 128, 146, 128, 147, 128, 148, 128, 149, 128, 150, 128, 151, 128, 152, 128, 153, 128, 154, 128, 155, 128, 156, 128, 157, 128, 158, 128, 159, 128, 160, 128, 235, 129, 111, 130, 131, 130, 132, 130, 133, 130, 134, 130, 135, 130, 136, 130, 137, 130, 138, 130, 139, 130, 140, 130, 141, 130, 142, 130, 143, 130, 144, 130, 145, 130, 146, 130, 147, 130, 149, 130, 150, 130, 151, 130, 90, 216, 5, 223, 90, 216, 10, 223, 108, 134, 93, 216, 12, 222, 139, 137, 195, 137, 210, 137, 198, 46, 94, 216, 178, 220, 0, 138, 1, 138, 193, 139, 194, 139, 195, 139, 196, 139, 197, 139, 198, 139, 199, 139, 200, 139, 201, 139, 202, 139, 203, 139, 204, 139, 205, 139, 206, 139, 207, 139, 208, 139, 209, 139, 210, 139, 55, 140, 70, 140, 85, 140, 120, 140, 157, 140, 33, 141, 34, 141, 100, 141, 112, 141, 179, 141, 202, 46, 171, 142, 202, 142, 105, 143, 106, 143, 107, 143, 155, 143, 176, 143, 181, 143, 60, 0, 42, 0, 225, 93, 190, 143, 191, 143, 192, 143, 193, 143, 194, 143, 195, 143, 196, 143, 197, 143, 198, 143, 199, 143, 200, 143, 201, 143, 97, 216, 104, 220, 97, 216, 108, 220, 97, 216, 115, 220, 145, 144, 158, 144, 159, 144, 160, 144, 161, 144, 162, 144, 163, 144, 164, 144, 165, 144, 166, 144, 167, 144, 168, 144, 169, 144, 170, 144, 172, 144, 97, 216, 170, 222, 97, 216, 171, 222, 73, 145, 198, 145, 204, 145, 136, 148, 137, 148, 138, 148, 139, 148, 140, 148, 240, 149, 241, 149, 242, 149, 243, 149, 244, 149, 245, 149, 246, 149, 247, 149, 37, 150, 38, 150, 39, 150, 40, 150, 41, 150, 42, 150, 43, 150, 44, 150, 45, 150, 46, 150, 47, 150, 48, 150, 49, 150, 50, 150, 51, 150, 52, 150, 53, 150, 54, 150, 99, 216, 54, 222, 99, 216, 57, 222, 231, 151, 207, 152, 104, 153, 105, 153, 106, 153, 107, 153, 108, 153, 109, 153, 110, 153, 113, 154, 114, 154, 115, 154, 116, 154, 32, 158, 33, 158, 166, 158, 159, 159, 60, 0, 39, 0, 208, 253, 8, 40, 39, 0, 60, 0, 42, 0, 38, 78, 39, 78, 51, 78, 86, 78, 115, 78, 116, 78, 117, 78, 118, 78, 119, 78, 120, 78, 64, 216, 238, 220, 139, 78, 155, 78, 157, 78, 158, 78, 159, 78, 171, 78, 172, 78, 45, 52, 76, 79, 105, 79, 106, 79, 107, 79, 108, 79, 109, 79, 110, 79, 111, 79, 112, 79, 113, 79, 114, 79, 115, 79, 116, 79, 117, 79, 118, 79, 119, 79, 120, 79, 121, 79, 122, 79, 123, 79, 124, 79, 125, 79, 126, 79, 127, 79, 128, 79, 129, 79, 130, 79, 131, 79, 132, 79, 133, 79, 134, 79, 135, 79, 136, 79, 137, 79, 138, 79, 139, 79, 140, 79, 141, 79, 142, 79, 143, 79, 144, 79, 145, 79, 146, 79, 147, 79, 148, 79, 149, 79, 150, 79, 151, 79, 152, 79, 153, 79, 154, 79, 155, 79, 156, 79, 157, 79, 158, 79, 159, 79, 160, 79, 161, 79, 162, 79, 163, 79, 164, 79, 165, 79, 166, 79, 167, 79, 168, 79, 169, 79, 170, 79, 171, 79, 172, 79, 173, 79, 199, 159, 74, 52, 76, 52, 80, 52, 64, 216, 4, 222, 64, 216, 12, 222, 64, 216, 20, 222, 82, 81, 83, 81, 84, 81, 85, 81, 86, 81, 65, 216, 135, 220, 105, 81, 118, 81, 119, 81, 120, 81, 144, 81, 158, 81, 188, 81, 189, 81, 190, 81, 191, 81, 192, 81, 65, 216, 195, 221, 237, 81, 238, 81, 239, 81, 253, 81, 46, 82, 47, 82, 48, 82, 49, 82, 50, 82, 51, 82, 52, 82, 53, 82, 54, 82, 55, 82, 60, 0, 42, 0, 56, 82, 57, 82, 58, 82, 59, 82, 60, 82, 61, 82, 62, 82, 63, 82, 64, 82, 65, 82, 66, 82, 228, 52, 230, 52, 65, 216, 14, 223, 70, 82, 181, 82, 182, 82, 183, 82, 184, 82, 185, 82, 186, 82, 187, 82, 188, 82, 189, 82, 190, 82, 191, 82, 26, 53, 198, 82, 10, 83, 11, 83, 12, 83, 66, 216, 22, 221, 38, 83, 60, 83, 81, 83, 82, 83, 83, 83, 84, 83, 85, 83, 86, 83, 101, 83, 102, 83, 103, 83, 66, 216, 231, 221, 118, 83, 119, 83, 120, 83, 121, 83, 122, 83, 146, 83, 147, 83, 148, 83, 149, 83, 66, 216, 80, 222, 192, 83, 193, 83, 194, 83, 212, 83, 213, 83, 214, 83, 215, 83, 216, 83, 93, 84, 94, 84, 95, 84, 96, 84, 97, 84, 98, 84, 99, 84, 100, 84, 101, 84, 102, 84, 103, 84, 104, 84, 105, 84, 106, 84, 107, 84, 108, 84, 109, 84, 110, 84, 111, 84, 113, 84, 115, 84, 116, 84, 117, 84, 118, 84, 119, 84, 120, 84, 121, 84, 122, 84, 123, 84, 124, 84, 125, 84, 126, 84, 127, 84, 128, 84, 129, 84, 130, 84, 131, 84, 132, 84, 133, 84, 134, 84, 135, 84, 136, 84, 137, 84, 138, 84, 139, 84, 140, 84, 141, 84, 142, 84, 143, 84, 144, 84, 145, 84, 146, 84, 147, 84, 148, 84, 149, 84, 150, 84, 151, 84, 152, 84, 153, 84, 154, 84, 155, 84, 156, 84, 157, 84, 119, 53, 120, 53, 67, 216, 11, 220, 60, 0, 42, 0, 67, 216, 13, 220, 67, 216, 32, 220, 67, 216, 52, 220, 67, 216, 58, 220, 67, 216, 59, 220, 67, 216, 65, 220, 67, 216, 66, 220, 67, 216, 67, 220, 246, 86, 247, 86, 248, 86, 249, 86, 250, 86, 251, 86, 252, 86, 253, 86, 254, 86, 32, 54, 97, 87, 98, 87, 99, 87, 100, 87, 101, 87, 102, 87, 103, 87, 104, 87, 105, 87, 106, 87, 107, 87, 108, 87, 109, 87, 110, 87, 111, 87, 112, 87, 113, 87, 114, 87, 115, 87, 116, 87, 117, 87, 118, 87, 119, 87, 120, 87, 121, 87, 122, 87, 123, 87, 124, 87, 125, 87, 126, 87, 127, 87, 128, 87, 129, 87, 130, 87, 131, 87, 132, 87, 133, 87, 134, 87, 135, 87, 136, 87, 137, 87, 138, 87, 49, 54, 50, 54, 51, 54, 52, 54, 53, 54, 68, 216, 168, 222, 68, 216, 169, 222, 68, 216, 176, 222, 7, 89, 12, 89, 28, 89, 29, 89, 67, 89, 68, 89, 69, 89, 70, 89, 71, 89, 72, 89, 73, 89, 75, 89, 76, 89, 77, 89, 69, 216, 10, 222, 84, 89, 172, 89, 173, 89, 174, 89, 175, 89, 176, 89, 177, 89, 178, 89, 179, 89, 180, 89, 181, 89, 182, 89, 183, 89, 184, 89, 185, 89, 186, 89, 187, 89, 188, 89, 189, 89, 190, 89, 191, 89, 192, 89, 193, 89, 194, 89, 195, 89, 196, 89, 197, 89, 198, 89, 199, 89, 200, 89, 201, 89, 202, 89, 203, 89, 204, 89, 205, 89, 206, 89, 207, 89, 208, 89, 209, 89, 210, 89, 211, 89, 212, 89, 214, 89, 215, 89, 176, 54, 60, 0, 42, 0, 177, 54, 181, 54, 185, 54, 188, 54, 193, 54, 69, 216, 223, 222, 69, 216, 230, 222, 69, 216, 231, 222, 69, 216, 232, 222, 69, 216, 250, 222, 69, 216, 251, 222, 69, 216, 252, 222, 69, 216, 254, 222, 95, 91, 96, 91, 97, 91, 98, 91, 99, 91, 100, 91, 101, 91, 102, 91, 103, 91, 64, 55, 70, 216, 88, 221, 147, 91, 148, 91, 149, 91, 150, 91, 151, 91, 152, 91, 153, 91, 154, 91, 155, 91, 156, 91, 157, 91, 158, 91, 159, 91, 160, 91, 161, 91, 70, 216, 219, 221, 0, 92, 25, 92, 26, 92, 45, 92, 68, 92, 69, 92, 70, 92, 71, 92, 72, 92, 73, 92, 74, 92, 157, 92, 158, 92, 159, 92, 160, 92, 161, 92, 162, 92, 163, 92, 164, 92, 165, 92, 166, 92, 167, 92, 168, 92, 169, 92, 170, 92, 171, 92, 172, 92, 173, 92, 174, 92, 175, 92, 176, 92, 177, 92, 178, 92, 179, 92, 180, 92, 181, 92, 182, 92, 183, 92, 184, 92, 185, 92, 186, 92, 187, 92, 188, 92, 189, 92, 190, 92, 191, 92, 192, 92, 193, 92, 194, 92, 195, 92, 196, 92, 197, 92, 185, 55, 190, 55, 71, 216, 144, 221, 246, 93, 48, 56, 17, 94, 18, 94, 19, 94, 20, 94, 21, 94, 22, 94, 23, 94, 24, 94, 25, 94, 26, 94, 27, 94, 28, 94, 72, 216, 126, 220, 119, 94, 120, 94, 152, 94, 149, 94, 150, 94, 151, 94, 153, 94, 154, 94, 156, 94, 157, 94, 158, 94, 159, 94, 117, 56, 246, 94, 248, 94, 249, 94, 160, 56, 6, 95, 60, 0, 42, 0, 33, 95, 34, 95, 35, 95, 36, 95, 37, 95, 38, 95, 39, 95, 40, 95, 41, 95, 42, 95, 72, 216, 208, 223, 84, 95, 85, 95, 204, 56, 73, 216, 139, 220, 124, 95, 125, 95, 126, 95, 127, 95, 128, 95, 129, 95, 130, 95, 131, 95, 132, 95, 221, 95, 222, 95, 224, 95, 226, 95, 229, 95, 233, 95, 245, 95, 253, 95, 255, 95, 1, 96, 2, 96, 7, 96, 8, 96, 9, 96, 10, 96, 11, 96, 12, 96, 13, 96, 15, 96, 16, 96, 17, 96, 19, 96, 20, 96, 21, 96, 22, 96, 23, 96, 25, 96, 26, 96, 27, 96, 28, 96, 30, 96, 31, 96, 33, 96, 34, 96, 38, 96, 39, 96, 41, 96, 42, 96, 43, 96, 44, 96, 45, 96, 46, 96, 47, 96, 48, 96, 50, 96, 51, 96, 52, 96, 53, 96, 54, 96, 58, 96, 61, 96, 62, 96, 63, 96, 20, 57, 73, 216, 27, 222, 73, 216, 28, 222, 73, 216, 43, 222, 20, 98, 21, 98, 22, 98, 23, 98, 61, 98, 62, 98, 63, 98, 64, 98, 127, 98, 166, 98, 167, 98, 168, 98, 169, 98, 170, 98, 171, 98, 172, 98, 173, 98, 174, 98, 175, 98, 176, 98, 177, 98, 178, 98, 179, 98, 180, 98, 181, 98, 182, 98, 183, 98, 184, 98, 185, 98, 186, 98, 187, 98, 188, 98, 189, 98, 190, 98, 191, 98, 192, 98, 193, 98, 194, 98, 195, 98, 196, 98, 197, 98, 198, 98, 199, 98, 200, 98, 201, 98, 202, 98, 203, 98, 60, 0, 42, 0, 204, 98, 205, 98, 206, 98, 208, 98, 209, 98, 210, 98, 211, 98, 212, 98, 213, 98, 214, 98, 215, 98, 216, 98, 217, 98, 218, 98, 219, 98, 221, 98, 222, 98, 223, 98, 224, 98, 225, 98, 226, 98, 227, 98, 228, 98, 229, 98, 230, 98, 231, 98, 232, 98, 233, 98, 212, 57, 213, 57, 217, 57, 218, 57, 220, 57, 221, 57, 222, 57, 223, 57, 74, 216, 207, 222, 74, 216, 213, 222, 61, 101, 62, 101, 129, 58, 137, 101, 76, 216, 68, 220, 166, 101, 167, 101, 168, 101, 169, 101, 186, 101, 187, 101, 188, 101, 76, 216, 218, 220, 249, 101, 250, 101, 251, 101, 252, 101, 253, 101, 254, 101, 255, 101, 0, 102, 1, 102, 2, 102, 3, 102, 4, 102, 5, 102, 6, 102, 7, 102, 8, 102, 9, 102, 10, 102, 11, 102, 12, 102, 13, 102, 14, 102, 15, 102, 16, 102, 17, 102, 18, 102, 19, 102, 20, 102, 21, 102, 22, 102, 23, 102, 24, 102, 25, 102, 246, 102, 10, 103, 11, 103, 12, 103, 13, 103, 51, 59, 106, 103, 107, 103, 108, 103, 109, 103, 110, 103, 111, 103, 112, 103, 113, 103, 114, 103, 115, 103, 116, 103, 117, 103, 118, 103, 119, 103, 120, 103, 121, 103, 122, 103, 123, 103, 124, 103, 125, 103, 126, 103, 127, 103, 128, 103, 130, 103, 131, 103, 132, 103, 133, 103, 134, 103, 135, 103, 136, 103, 137, 103, 138, 103, 139, 103, 140, 103, 141, 103, 142, 103, 143, 103, 144, 103, 60, 0, 42, 0, 145, 103, 146, 103, 147, 103, 148, 103, 149, 103, 150, 103, 151, 103, 152, 103, 153, 103, 154, 103, 155, 103, 156, 103, 157, 103, 158, 103, 159, 103, 160, 103, 161, 103, 162, 103, 163, 103, 164, 103, 165, 103, 166, 103, 167, 103, 168, 103, 169, 103, 170, 103, 171, 103, 172, 103, 173, 103, 71, 59, 72, 59, 73, 59, 74, 59, 75, 59, 76, 59, 77, 59, 78, 59, 79, 59, 80, 59, 76, 216, 222, 223, 76, 216, 230, 223, 76, 216, 244, 223, 76, 216, 245, 223, 76, 216, 249, 223, 76, 216, 250, 223, 76, 216, 254, 223, 77, 216, 0, 220, 249, 103, 35, 107, 37, 107, 38, 107, 39, 107, 32, 60, 102, 107, 103, 107, 104, 107, 105, 107, 125, 107, 126, 107, 127, 107, 128, 107, 129, 107, 180, 107, 209, 107, 222, 107, 223, 107, 19, 108, 27, 108, 28, 108, 29, 108, 108, 108, 128, 108, 138, 108, 147, 108, 157, 108, 171, 108, 172, 108, 173, 108, 174, 108, 176, 108, 177, 108, 178, 108, 179, 108, 180, 108, 181, 108, 182, 108, 183, 108, 184, 108, 185, 108, 186, 108, 187, 108, 188, 108, 189, 108, 190, 108, 191, 108, 192, 108, 193, 108, 194, 108, 195, 108, 196, 108, 197, 108, 198, 108, 199, 108, 200, 108, 202, 108, 203, 108, 204, 108, 205, 108, 206, 108, 207, 108, 208, 108, 209, 108, 210, 108, 211, 108, 212, 108, 213, 108, 214, 108, 215, 108, 216, 108, 217, 108, 219, 108, 220, 108, 221, 108, 222, 108, 223, 108, 224, 108, 225, 108, 226, 108, 60, 0, 42, 0, 227, 108, 228, 108, 229, 108, 230, 108, 231, 108, 232, 108, 233, 108, 234, 108, 235, 108, 236, 108, 237, 108, 238, 108, 239, 108, 241, 108, 242, 108, 243, 108, 247, 108, 248, 108, 249, 108, 250, 108, 251, 108, 252, 108, 253, 108, 254, 108, 203, 60, 205, 60, 209, 60, 79, 216, 181, 220, 79, 216, 183, 220, 79, 216, 199, 220, 79, 216, 200, 220, 79, 216, 201, 220, 48, 109, 135, 112, 129, 112, 130, 112, 131, 112, 132, 112, 133, 112, 134, 112, 136, 112, 137, 112, 138, 112, 139, 112, 140, 112, 141, 112, 142, 112, 143, 112, 144, 112, 145, 112, 146, 112, 147, 112, 148, 112, 149, 112, 150, 112, 151, 112, 152, 112, 153, 112, 154, 112, 155, 112, 156, 112, 157, 112, 158, 112, 80, 216, 163, 221, 80, 216, 164, 221, 80, 216, 165, 221, 80, 216, 172, 221, 80, 216, 181, 221, 44, 114, 45, 114, 56, 114, 64, 114, 72, 114, 101, 114, 102, 114, 103, 114, 104, 114, 105, 114, 106, 114, 107, 114, 108, 114, 47, 62, 81, 216, 42, 222, 192, 114, 201, 114, 203, 114, 204, 114, 205, 114, 206, 114, 207, 114, 208, 114, 209, 114, 210, 114, 211, 114, 212, 114, 213, 114, 214, 114, 215, 114, 216, 114, 217, 114, 218, 114, 219, 114, 220, 114, 221, 114, 222, 114, 98, 62, 105, 62, 156, 115, 157, 115, 158, 115, 159, 115, 160, 115, 161, 115, 162, 115, 163, 115, 164, 115, 165, 115, 166, 115, 167, 115, 168, 115, 169, 115, 170, 115, 171, 115, 172, 115, 173, 115, 174, 115, 175, 115, 176, 115, 60, 0, 42, 0, 177, 115, 82, 216, 251, 220, 82, 216, 255, 220, 82, 216, 0, 221, 82, 216, 1, 221, 82, 216, 12, 221, 221, 116, 232, 116, 233, 116, 25, 117, 59, 117, 61, 117, 62, 117, 63, 117, 64, 117, 65, 117, 66, 117, 67, 117, 68, 117, 69, 117, 140, 117, 152, 117, 153, 117, 154, 117, 155, 117, 156, 117, 157, 117, 158, 117, 159, 117, 160, 117, 161, 117, 119, 118, 132, 118, 175, 118, 194, 118, 240, 118, 241, 118, 242, 118, 243, 118, 244, 118, 245, 118, 228, 119, 229, 119, 247, 119, 248, 119, 249, 119, 250, 119, 251, 119, 252, 119, 253, 119, 254, 119, 255, 119, 0, 120, 1, 120, 85, 216, 37, 220, 62, 121, 63, 121, 64, 121, 65, 121, 66, 121, 67, 121, 196, 121, 197, 121, 198, 121, 199, 121, 200, 121, 201, 121, 202, 121, 120, 122, 121, 122, 122, 122, 123, 122, 146, 65, 206, 122, 207, 122, 196, 65, 250, 122, 251, 122, 116, 124, 117, 124, 118, 124, 252, 124, 253, 124, 254, 124, 255, 124, 181, 66, 191, 126, 192, 126, 193, 126, 194, 126, 195, 126, 196, 126, 197, 126, 198, 126, 199, 126, 200, 126, 201, 126, 202, 126, 203, 126, 204, 126, 205, 126, 206, 126, 207, 126, 208, 126, 55, 127, 84, 127, 86, 127, 87, 127, 89, 127, 139, 127, 140, 127, 5, 128, 19, 128, 53, 128, 131, 128, 143, 128, 161, 128, 162, 128, 163, 128, 164, 128, 165, 128, 166, 128, 167, 128, 168, 128, 169, 128, 170, 128, 171, 128, 172, 128, 60, 0, 42, 0, 173, 128, 174, 128, 175, 128, 176, 128, 177, 128, 178, 128, 179, 128, 180, 128, 181, 128, 182, 128, 183, 128, 184, 128, 185, 128, 186, 128, 187, 128, 188, 128, 189, 128, 190, 128, 191, 128, 192, 128, 193, 128, 217, 67, 228, 129, 229, 129, 253, 129, 254, 129, 13, 130, 14, 130, 15, 130, 32, 130, 112, 130, 152, 130, 153, 130, 154, 130, 155, 130, 156, 130, 157, 130, 158, 130, 159, 130, 160, 130, 161, 130, 162, 130, 163, 130, 164, 130, 165, 130, 166, 130, 167, 130, 168, 130, 169, 130, 170, 130, 171, 130, 172, 130, 173, 130, 174, 130, 175, 130, 176, 130, 177, 130, 178, 130, 179, 130, 180, 130, 181, 130, 182, 130, 183, 130, 184, 130, 185, 130, 186, 130, 188, 130, 189, 130, 190, 130, 192, 130, 193, 130, 194, 130, 195, 130, 196, 130, 197, 130, 198, 130, 199, 130, 200, 130, 201, 130, 202, 130, 203, 130, 204, 130, 205, 130, 206, 130, 207, 130, 62, 131, 90, 216, 19, 223, 90, 216, 21, 223, 90, 216, 35, 223, 90, 216, 40, 223, 191, 130, 78, 134, 79, 134, 109, 134, 110, 134, 111, 134, 112, 134, 113, 134, 114, 134, 93, 216, 224, 221, 101, 136, 104, 136, 196, 137, 197, 137, 211, 139, 212, 139, 213, 139, 214, 139, 215, 139, 216, 139, 217, 139, 218, 139, 219, 139, 220, 139, 221, 139, 222, 139, 223, 139, 224, 139, 225, 139, 226, 139, 227, 139, 228, 139, 229, 139, 230, 139, 231, 139, 232, 139, 233, 139, 86, 140, 60, 0, 42, 0, 35, 141, 36, 141, 37, 141, 38, 141, 39, 141, 40, 141, 41, 141, 42, 141, 43, 141, 44, 141, 45, 141, 46, 141, 47, 141, 203, 142, 108, 143, 109, 143, 110, 143, 111, 143, 112, 143, 202, 143, 203, 143, 204, 143, 205, 143, 206, 143, 207, 143, 208, 143, 209, 143, 210, 143, 211, 143, 212, 143, 213, 143, 214, 143, 215, 143, 216, 143, 217, 143, 218, 143, 219, 143, 220, 143, 221, 143, 222, 143, 223, 143, 236, 143, 97, 216, 130, 220, 173, 144, 174, 144, 175, 144, 176, 144, 177, 144, 178, 144, 179, 144, 180, 144, 181, 144, 182, 144, 183, 144, 184, 144, 185, 144, 186, 144, 187, 144, 97, 216, 178, 222, 97, 216, 188, 222, 199, 145, 209, 145, 210, 145, 141, 148, 142, 148, 143, 148, 144, 148, 145, 148, 146, 148, 147, 148, 148, 148, 149, 148, 150, 148, 151, 148, 119, 149, 120, 149, 128, 149, 248, 149, 249, 149, 28, 150, 55, 150, 56, 150, 57, 150, 58, 150, 59, 150, 60, 150, 61, 150, 62, 150, 63, 150, 64, 150, 65, 150, 66, 150, 67, 150, 68, 150, 69, 150, 70, 150, 71, 150, 72, 150, 73, 150, 182, 150, 185, 150, 215, 46, 232, 150, 81, 151, 82, 151, 94, 151, 99, 151, 118, 152, 119, 152, 111, 153, 112, 153, 113, 153, 114, 153, 115, 153, 116, 153, 117, 154, 118, 154, 119, 154, 120, 154, 121, 154, 122, 154, 123, 154, 124, 154, 125, 154, 126, 154, 127, 154, 128, 154, 124, 156, 60, 0, 42, 0, 34, 158, 35, 158, 36, 158, 254, 158, 33, 159, 127, 159, 60, 0, 39, 0, 208, 253, 9, 40, 39, 0, 60, 0, 42, 0, 52, 78, 62, 78, 87, 78, 121, 78, 122, 78, 123, 78, 124, 78, 32, 52, 173, 78, 174, 78, 175, 78, 176, 78, 177, 78, 178, 78, 174, 79, 175, 79, 176, 79, 177, 79, 178, 79, 179, 79, 180, 79, 181, 79, 182, 79, 183, 79, 184, 79, 185, 79, 186, 79, 187, 79, 188, 79, 189, 79, 190, 79, 191, 79, 192, 79, 193, 79, 194, 79, 195, 79, 196, 79, 197, 79, 198, 79, 199, 79, 200, 79, 201, 79, 202, 79, 203, 79, 204, 79, 205, 79, 206, 79, 207, 79, 208, 79, 209, 79, 210, 79, 211, 79, 212, 79, 213, 79, 214, 79, 215, 79, 216, 79, 217, 79, 218, 79, 219, 79, 220, 79, 221, 79, 222, 79, 223, 79, 224, 79, 225, 79, 226, 79, 227, 79, 228, 79, 229, 79, 230, 79, 231, 79, 232, 79, 233, 79, 234, 79, 235, 79, 236, 79, 237, 79, 64, 216, 57, 222, 64, 216, 91, 222, 87, 81, 88, 81, 89, 81, 65, 216, 142, 220, 106, 81, 121, 81, 123, 81, 145, 81, 146, 81, 159, 81, 160, 81, 193, 81, 194, 81, 195, 81, 65, 216, 202, 221, 65, 216, 208, 221, 65, 216, 213, 221, 254, 81, 67, 82, 68, 82, 69, 82, 71, 82, 72, 82, 73, 82, 74, 82, 75, 82, 76, 82, 77, 82, 78, 82, 79, 82, 80, 82, 81, 82, 192, 82, 193, 82, 194, 82, 195, 82, 196, 82, 197, 82, 199, 82, 200, 82, 201, 82, 202, 82, 203, 82, 13, 83, 39, 83, 40, 83, 41, 83, 61, 83, 60, 0, 42, 0, 87, 83, 88, 83, 123, 83, 124, 83, 125, 83, 150, 83, 151, 83, 152, 83, 153, 83, 154, 83, 155, 83, 217, 83, 218, 83, 219, 83, 220, 83, 221, 83, 112, 84, 114, 84, 158, 84, 159, 84, 160, 84, 161, 84, 162, 84, 163, 84, 164, 84, 165, 84, 166, 84, 167, 84, 168, 84, 169, 84, 170, 84, 171, 84, 172, 84, 173, 84, 174, 84, 175, 84, 176, 84, 177, 84, 178, 84, 179, 84, 180, 84, 181, 84, 182, 84, 183, 84, 184, 84, 185, 84, 186, 84, 187, 84, 188, 84, 189, 84, 190, 84, 191, 84, 192, 84, 193, 84, 194, 84, 195, 84, 196, 84, 197, 84, 198, 84, 199, 84, 200, 84, 201, 84, 202, 84, 203, 84, 204, 84, 205, 84, 206, 84, 207, 84, 208, 84, 209, 84, 210, 84, 211, 84, 212, 84, 213, 84, 214, 84, 215, 84, 216, 84, 217, 84, 218, 84, 219, 84, 220, 84, 221, 84, 222, 84, 223, 84, 132, 53, 67, 216, 83, 220, 67, 216, 101, 220, 67, 216, 119, 220, 67, 216, 120, 220, 67, 216, 124, 220, 67, 216, 141, 220, 67, 216, 150, 220, 67, 216, 156, 220, 255, 86, 0, 87, 34, 54, 68, 216, 217, 221, 139, 87, 140, 87, 141, 87, 142, 87, 143, 87, 144, 87, 145, 87, 146, 87, 147, 87, 148, 87, 149, 87, 150, 87, 151, 87, 152, 87, 153, 87, 154, 87, 155, 87, 156, 87, 157, 87, 158, 87, 159, 87, 160, 87, 161, 87, 162, 87, 163, 87, 164, 87, 165, 87, 166, 87, 167, 87, 168, 87, 169, 87, 60, 0, 42, 0, 170, 87, 171, 87, 172, 87, 173, 87, 174, 87, 175, 87, 176, 87, 177, 87, 178, 87, 179, 87, 180, 87, 181, 87, 206, 87, 54, 54, 55, 54, 56, 54, 57, 54, 59, 54, 62, 54, 68, 216, 227, 222, 244, 88, 245, 88, 8, 89, 9, 89, 13, 89, 78, 89, 79, 89, 80, 89, 81, 89, 82, 89, 83, 89, 85, 89, 86, 89, 154, 54, 69, 216, 19, 222, 69, 216, 25, 222, 213, 89, 141, 89, 216, 89, 217, 89, 218, 89, 219, 89, 220, 89, 221, 89, 222, 89, 223, 89, 224, 89, 225, 89, 226, 89, 227, 89, 228, 89, 229, 89, 230, 89, 231, 89, 232, 89, 233, 89, 234, 89, 235, 89, 237, 89, 238, 89, 239, 89, 240, 89, 241, 89, 242, 89, 243, 89, 244, 89, 245, 89, 246, 89, 247, 89, 248, 89, 249, 89, 250, 89, 251, 89, 252, 89, 253, 89, 254, 89, 255, 89, 0, 90, 1, 90, 2, 90, 3, 90, 5, 90, 6, 90, 7, 90, 8, 90, 195, 54, 196, 54, 197, 54, 199, 54, 200, 54, 69, 216, 13, 223, 69, 216, 16, 223, 69, 216, 38, 223, 13, 90, 104, 91, 105, 91, 106, 91, 162, 91, 163, 91, 164, 91, 165, 91, 166, 91, 168, 91, 169, 91, 170, 91, 171, 91, 70, 216, 243, 221, 1, 92, 2, 92, 6, 92, 27, 92, 28, 92, 29, 92, 46, 92, 47, 92, 75, 92, 76, 92, 77, 92, 78, 92, 79, 92, 198, 92, 199, 92, 200, 92, 201, 92, 202, 92, 203, 92, 204, 92, 205, 92, 60, 0, 42, 0, 206, 92, 207, 92, 208, 92, 209, 92, 210, 92, 211, 92, 212, 92, 213, 92, 214, 92, 215, 92, 216, 92, 217, 92, 218, 92, 219, 92, 220, 92, 221, 92, 222, 92, 223, 92, 224, 92, 225, 92, 226, 92, 227, 92, 228, 92, 229, 92, 230, 92, 231, 92, 71, 216, 182, 221, 71, 216, 186, 221, 71, 216, 202, 221, 71, 216, 209, 221, 248, 92, 236, 93, 237, 93, 247, 93, 248, 93, 249, 93, 250, 93, 251, 93, 29, 94, 30, 94, 31, 94, 32, 94, 33, 94, 34, 94, 35, 94, 36, 94, 37, 94, 38, 94, 39, 94, 72, 216, 154, 220, 125, 94, 155, 94, 160, 94, 161, 94, 162, 94, 163, 94, 164, 94, 165, 94, 166, 94, 72, 216, 8, 222, 176, 94, 250, 94, 251, 94, 252, 94, 72, 216, 33, 223, 72, 216, 37, 223, 7, 95, 8, 95, 43, 95, 44, 95, 45, 95, 46, 95, 47, 95, 72, 216, 215, 223, 86, 95, 101, 95, 102, 95, 133, 95, 134, 95, 135, 95, 136, 95, 137, 95, 138, 95, 139, 95, 140, 95, 141, 95, 148, 95, 14, 96, 18, 96, 24, 96, 29, 96, 32, 96, 35, 96, 36, 96, 37, 96, 40, 96, 49, 96, 55, 96, 56, 96, 57, 96, 59, 96, 60, 96, 64, 96, 66, 96, 67, 96, 68, 96, 69, 96, 70, 96, 71, 96, 72, 96, 73, 96, 74, 96, 76, 96, 77, 96, 78, 96, 81, 96, 82, 96, 83, 96, 84, 96, 87, 96, 88, 96, 91, 96, 92, 96, 94, 96, 95, 96, 96, 96, 97, 96, 98, 96, 60, 0, 42, 0, 100, 96, 102, 96, 104, 96, 106, 96, 107, 96, 108, 96, 110, 96, 112, 96, 113, 96, 114, 96, 120, 96, 121, 96, 122, 96, 123, 96, 124, 96, 125, 96, 34, 57, 38, 57, 39, 57, 45, 57, 73, 216, 104, 222, 24, 98, 65, 98, 66, 98, 67, 98, 207, 98, 220, 98, 216, 57, 234, 98, 235, 98, 236, 98, 237, 98, 238, 98, 239, 98, 240, 98, 241, 98, 244, 98, 245, 98, 246, 98, 247, 98, 248, 98, 249, 98, 250, 98, 251, 98, 252, 98, 253, 98, 254, 98, 0, 99, 1, 99, 2, 99, 3, 99, 4, 99, 5, 99, 6, 99, 7, 99, 9, 99, 10, 99, 11, 99, 12, 99, 13, 99, 14, 99, 15, 99, 17, 99, 18, 99, 19, 99, 20, 99, 21, 99, 22, 99, 23, 99, 24, 99, 28, 99, 29, 99, 30, 99, 31, 99, 32, 99, 33, 99, 34, 99, 35, 99, 36, 99, 37, 99, 38, 99, 39, 99, 225, 57, 226, 57, 229, 57, 230, 57, 74, 216, 230, 222, 74, 216, 232, 222, 74, 216, 14, 223, 74, 216, 34, 223, 49, 101, 63, 101, 64, 101, 65, 101, 66, 101, 67, 101, 68, 101, 69, 101, 131, 58, 132, 58, 133, 58, 170, 101, 171, 101, 188, 58, 189, 58, 189, 101, 190, 101, 191, 101, 192, 101, 198, 58, 226, 101, 26, 102, 27, 102, 28, 102, 29, 102, 30, 102, 31, 102, 32, 102, 33, 102, 34, 102, 35, 102, 36, 102, 37, 102, 38, 102, 39, 102, 40, 102, 41, 102, 42, 102, 60, 0, 42, 0, 43, 102, 44, 102, 45, 102, 46, 102, 47, 102, 48, 102, 49, 102, 50, 102, 51, 102, 52, 102, 53, 102, 54, 102, 55, 102, 56, 102, 57, 102, 58, 102, 59, 102, 60, 102, 61, 102, 62, 102, 63, 102, 222, 58, 224, 58, 76, 216, 125, 221, 76, 216, 130, 221, 247, 102, 14, 103, 15, 103, 16, 103, 174, 103, 175, 103, 176, 103, 177, 103, 178, 103, 179, 103, 180, 103, 181, 103, 182, 103, 183, 103, 184, 103, 185, 103, 186, 103, 187, 103, 188, 103, 190, 103, 191, 103, 192, 103, 193, 103, 194, 103, 195, 103, 196, 103, 197, 103, 198, 103, 199, 103, 200, 103, 201, 103, 202, 103, 203, 103, 204, 103, 205, 103, 206, 103, 207, 103, 208, 103, 209, 103, 210, 103, 211, 103, 212, 103, 213, 103, 214, 103, 215, 103, 216, 103, 217, 103, 218, 103, 219, 103, 220, 103, 221, 103, 222, 103, 223, 103, 224, 103, 226, 103, 227, 103, 228, 103, 229, 103, 230, 103, 231, 103, 232, 103, 233, 103, 234, 103, 235, 103, 236, 103, 237, 103, 238, 103, 239, 103, 240, 103, 241, 103, 242, 103, 243, 103, 245, 103, 246, 103, 247, 103, 248, 103, 250, 103, 251, 103, 252, 103, 253, 103, 254, 103, 255, 103, 0, 104, 1, 104, 2, 104, 3, 104, 4, 104, 5, 104, 6, 104, 7, 104, 8, 104, 9, 104, 10, 104, 11, 104, 12, 104, 13, 104, 14, 104, 15, 104, 16, 104, 17, 104, 81, 59, 82, 59, 83, 59, 60, 0, 42, 0, 84, 59, 85, 59, 86, 59, 87, 59, 82, 104, 40, 107, 41, 107, 42, 107, 38, 60, 106, 107, 107, 107, 84, 60, 130, 107, 131, 107, 132, 107, 133, 107, 134, 107, 135, 107, 181, 107, 182, 107, 210, 107, 214, 107, 215, 107, 216, 107, 139, 60, 224, 107, 225, 107, 78, 216, 90, 223, 30, 108, 31, 108, 32, 108, 33, 108, 34, 108, 151, 108, 175, 108, 201, 108, 244, 108, 246, 108, 218, 108, 255, 108, 0, 109, 1, 109, 2, 109, 3, 109, 4, 109, 5, 109, 6, 109, 7, 109, 8, 109, 9, 109, 10, 109, 11, 109, 12, 109, 14, 109, 15, 109, 16, 109, 17, 109, 18, 109, 19, 109, 20, 109, 21, 109, 23, 109, 24, 109, 25, 109, 26, 109, 27, 109, 29, 109, 30, 109, 31, 109, 32, 109, 33, 109, 34, 109, 35, 109, 36, 109, 37, 109, 38, 109, 39, 109, 40, 109, 41, 109, 42, 109, 43, 109, 44, 109, 45, 109, 46, 109, 49, 109, 50, 109, 51, 109, 52, 109, 53, 109, 54, 109, 55, 109, 56, 109, 57, 109, 58, 109, 59, 109, 60, 109, 61, 109, 62, 109, 63, 109, 64, 109, 65, 109, 66, 109, 67, 109, 68, 109, 69, 109, 71, 109, 72, 109, 73, 109, 74, 109, 75, 109, 76, 109, 77, 109, 78, 109, 79, 109, 80, 109, 81, 109, 82, 109, 83, 109, 84, 109, 85, 109, 214, 60, 220, 60, 79, 216, 252, 220, 79, 216, 253, 220, 79, 216, 254, 220, 79, 216, 255, 220, 159, 112, 160, 112, 60, 0, 42, 0, 161, 112, 162, 112, 163, 112, 164, 112, 165, 112, 166, 112, 167, 112, 168, 112, 169, 112, 170, 112, 171, 112, 172, 112, 173, 112, 174, 112, 175, 112, 176, 112, 177, 112, 178, 112, 179, 112, 180, 112, 181, 112, 182, 112, 183, 112, 184, 112, 185, 112, 186, 112, 187, 112, 188, 112, 189, 112, 190, 112, 191, 112, 192, 112, 193, 112, 194, 112, 195, 112, 173, 61, 80, 216, 205, 221, 80, 216, 226, 221, 46, 114, 47, 114, 48, 114, 81, 216, 5, 221, 60, 114, 65, 114, 73, 114, 74, 114, 109, 114, 110, 114, 111, 114, 112, 114, 113, 114, 114, 114, 115, 114, 116, 114, 117, 114, 202, 114, 223, 114, 224, 114, 225, 114, 226, 114, 227, 114, 228, 114, 229, 114, 230, 114, 231, 114, 232, 114, 233, 114, 234, 114, 235, 114, 236, 114, 237, 114, 238, 114, 239, 114, 240, 114, 241, 114, 242, 114, 81, 216, 143, 223, 133, 115, 177, 62, 178, 115, 179, 115, 180, 115, 181, 115, 182, 115, 183, 115, 184, 115, 185, 115, 187, 115, 189, 115, 190, 115, 191, 115, 192, 115, 193, 115, 194, 115, 195, 115, 196, 115, 197, 115, 198, 115, 199, 115, 200, 115, 201, 115, 202, 115, 203, 115, 204, 115, 205, 115, 206, 115, 207, 115, 208, 115, 209, 115, 82, 216, 22, 221, 82, 216, 23, 221, 82, 216, 25, 221, 82, 216, 47, 221, 82, 216, 51, 221, 82, 216, 52, 221, 234, 116, 235, 116, 236, 116, 237, 116, 238, 116, 239, 116, 240, 116, 241, 116, 242, 116, 26, 117, 32, 117, 45, 117, 46, 117, 60, 0, 42, 0, 70, 117, 71, 117, 72, 117, 73, 117, 74, 117, 75, 117, 76, 117, 77, 117, 78, 117, 79, 117, 80, 117, 81, 117, 82, 117, 83, 117, 88, 63, 89, 63, 162, 117, 163, 117, 164, 117, 165, 117, 166, 117, 167, 117, 168, 117, 169, 117, 170, 117, 171, 117, 172, 117, 173, 117, 174, 117, 175, 117, 186, 117, 124, 63, 120, 118, 121, 118, 122, 118, 133, 118, 134, 118, 135, 118, 136, 118, 195, 118, 196, 118, 197, 118, 198, 118, 199, 118, 200, 118, 246, 118, 247, 118, 248, 118, 249, 118, 250, 118, 251, 118, 252, 118, 253, 118, 254, 118, 255, 118, 0, 119, 1, 119, 2, 119, 3, 119, 4, 119, 5, 119, 6, 119, 7, 119, 8, 119, 9, 119, 10, 119, 11, 119, 12, 119, 13, 119, 29, 64, 84, 216, 43, 221, 220, 119, 230, 119, 231, 119, 232, 119, 245, 108, 2, 120, 3, 120, 4, 120, 5, 120, 6, 120, 7, 120, 8, 120, 9, 120, 10, 120, 11, 120, 12, 120, 13, 120, 14, 120, 15, 120, 16, 120, 17, 120, 18, 120, 19, 120, 20, 120, 21, 120, 22, 120, 23, 120, 24, 120, 25, 120, 26, 120, 27, 120, 28, 120, 154, 64, 85, 216, 47, 220, 85, 216, 48, 220, 68, 121, 69, 121, 70, 121, 71, 121, 72, 121, 73, 121, 74, 121, 75, 121, 76, 121, 77, 121, 78, 121, 198, 137, 0, 65, 85, 216, 53, 222, 185, 121, 186, 121, 203, 121, 204, 121, 205, 121, 206, 121, 207, 121, 208, 121, 60, 0, 42, 0, 209, 121, 210, 121, 211, 121, 212, 121, 213, 121, 214, 121, 215, 121, 49, 65, 50, 65, 85, 216, 114, 223, 124, 122, 125, 122, 126, 122, 127, 122, 128, 122, 129, 122, 130, 122, 131, 122, 149, 65, 86, 216, 70, 221, 208, 122, 209, 122, 210, 122, 211, 122, 212, 122, 213, 122, 214, 122, 215, 122, 252, 122, 253, 122, 254, 122, 255, 122, 0, 123, 1, 123, 2, 123, 3, 123, 86, 216, 233, 222, 4, 90, 119, 124, 120, 124, 121, 124, 122, 124, 123, 124, 124, 124, 125, 124, 126, 124, 127, 124, 128, 124, 129, 124, 130, 124, 122, 66, 0, 125, 1, 125, 2, 125, 3, 125, 4, 125, 5, 125, 6, 125, 7, 125, 8, 125, 9, 125, 182, 66, 183, 66, 185, 66, 209, 126, 210, 126, 211, 126, 212, 126, 213, 126, 214, 126, 215, 126, 216, 126, 217, 126, 218, 126, 219, 126, 220, 126, 221, 126, 222, 126, 223, 126, 56, 127, 66, 67, 88, 127, 90, 127, 141, 127, 142, 127, 143, 127, 145, 127, 88, 216, 81, 223, 190, 127, 191, 127, 88, 216, 245, 223, 88, 216, 248, 223, 7, 128, 13, 128, 14, 128, 15, 128, 16, 128, 17, 128, 20, 128, 54, 128, 55, 128, 17, 103, 194, 128, 195, 128, 196, 128, 197, 128, 198, 128, 199, 128, 200, 128, 201, 128, 202, 128, 203, 128, 204, 128, 205, 128, 206, 128, 207, 128, 208, 128, 209, 128, 210, 128, 211, 128, 213, 128, 214, 128, 215, 128, 216, 128, 217, 128, 218, 128, 219, 128, 220, 128, 60, 0, 42, 0, 221, 128, 222, 128, 223, 128, 224, 128, 225, 128, 226, 128, 227, 128, 228, 128, 229, 128, 230, 128, 231, 128, 232, 128, 233, 128, 234, 128, 235, 128, 9, 129, 223, 67, 244, 129, 255, 129, 33, 130, 34, 130, 35, 130, 36, 130, 148, 130, 208, 130, 209, 130, 210, 130, 211, 130, 212, 130, 213, 130, 214, 130, 215, 130, 216, 130, 217, 130, 218, 130, 219, 130, 220, 130, 221, 130, 222, 130, 223, 130, 224, 130, 225, 130, 226, 130, 227, 130, 228, 130, 229, 130, 230, 130, 231, 130, 232, 130, 233, 130, 234, 130, 235, 130, 236, 130, 237, 130, 238, 130, 239, 130, 240, 130, 241, 130, 242, 130, 243, 130, 244, 130, 245, 130, 246, 130, 247, 130, 248, 130, 249, 130, 250, 130, 251, 130, 252, 130, 253, 130, 254, 130, 255, 130, 0, 131, 1, 131, 2, 131, 3, 131, 4, 131, 5, 131, 6, 131, 7, 131, 9, 131, 10, 131, 11, 131, 12, 131, 13, 131, 14, 131, 15, 131, 16, 131, 17, 131, 19, 131, 20, 131, 21, 131, 90, 216, 80, 223, 90, 216, 81, 223, 90, 216, 82, 223, 90, 216, 83, 223, 90, 216, 91, 223, 58, 131, 80, 134, 115, 134, 116, 134, 117, 134, 118, 134, 119, 134, 120, 134, 121, 134, 122, 134, 123, 134, 124, 134, 125, 134, 126, 134, 127, 134, 128, 134, 129, 134, 130, 134, 131, 134, 65, 136, 66, 136, 15, 70, 77, 136, 78, 136, 93, 216, 228, 221, 102, 136, 103, 136, 105, 136, 106, 136, 107, 136, 108, 136, 60, 0, 42, 0, 93, 216, 50, 222, 93, 216, 57, 222, 129, 137, 140, 137, 199, 137, 200, 137, 201, 137, 211, 137, 212, 137, 2, 138, 3, 138, 4, 138, 5, 138, 6, 138, 7, 138, 8, 138, 174, 70, 234, 139, 235, 139, 236, 139, 237, 139, 238, 139, 239, 139, 240, 139, 241, 139, 242, 139, 243, 139, 244, 139, 245, 139, 246, 139, 158, 140, 159, 140, 160, 140, 48, 141, 49, 141, 50, 141, 51, 141, 52, 141, 53, 141, 54, 141, 55, 141, 56, 141, 57, 141, 58, 141, 59, 141, 114, 141, 115, 141, 116, 141, 117, 141, 170, 159, 180, 141, 204, 142, 205, 142, 113, 143, 114, 143, 115, 143, 116, 143, 117, 143, 118, 143, 119, 143, 120, 143, 121, 143, 122, 143, 123, 143, 224, 143, 225, 143, 226, 143, 227, 143, 228, 143, 229, 143, 230, 143, 231, 143, 232, 143, 233, 143, 234, 143, 235, 143, 237, 143, 238, 143, 239, 143, 240, 143, 241, 143, 242, 143, 243, 143, 188, 144, 189, 144, 190, 144, 191, 144, 192, 144, 193, 144, 194, 144, 195, 144, 196, 144, 197, 144, 198, 144, 199, 144, 200, 144, 201, 144, 202, 144, 203, 144, 205, 144, 206, 144, 207, 144, 208, 144, 209, 144, 211, 144, 213, 144, 97, 216, 216, 222, 97, 216, 230, 222, 241, 144, 74, 145, 75, 145, 205, 145, 211, 145, 212, 145, 152, 148, 153, 148, 154, 148, 155, 148, 156, 148, 157, 148, 158, 148, 159, 148, 160, 148, 161, 148, 162, 148, 163, 148, 164, 148, 165, 148, 60, 0, 42, 0, 166, 148, 167, 148, 168, 148, 169, 148, 170, 148, 171, 148, 172, 148, 173, 148, 174, 148, 175, 148, 129, 149, 130, 149, 250, 149, 251, 149, 252, 149, 253, 149, 254, 149, 255, 149, 0, 150, 1, 150, 2, 150, 74, 150, 75, 150, 76, 150, 77, 150, 78, 150, 79, 150, 80, 150, 81, 150, 82, 150, 83, 150, 84, 150, 85, 150, 99, 216, 101, 222, 99, 216, 102, 222, 98, 151, 105, 151, 203, 151, 232, 151, 237, 151, 243, 151, 1, 152, 120, 152, 121, 152, 122, 152, 123, 152, 168, 152, 208, 152, 209, 152, 210, 152, 219, 152, 223, 152, 224, 152, 222, 46, 117, 153, 118, 153, 119, 153, 120, 153, 121, 153, 122, 153, 123, 153, 124, 153, 150, 153, 102, 216, 16, 220, 153, 153, 129, 154, 130, 154, 131, 154, 132, 154, 133, 154, 134, 154, 135, 154, 136, 154, 137, 154, 227, 46, 236, 156, 37, 158, 38, 158, 39, 158, 40, 158, 41, 158, 60, 0, 39, 0, 208, 253, 10, 40, 39, 0, 60, 0, 42, 0, 64, 216, 62, 220, 53, 78, 88, 78, 125, 78, 179, 78, 238, 79, 239, 79, 240, 79, 241, 79, 242, 79, 243, 79, 244, 79, 245, 79, 246, 79, 247, 79, 248, 79, 249, 79, 250, 79, 251, 79, 252, 79, 253, 79, 254, 79, 255, 79, 0, 80, 1, 80, 2, 80, 3, 80, 4, 80, 5, 80, 6, 80, 7, 80, 8, 80, 9, 80, 10, 80, 11, 80, 12, 80, 13, 80, 14, 80, 15, 80, 16, 80, 17, 80, 18, 80, 19, 80, 20, 80, 21, 80, 22, 80, 23, 80, 24, 80, 25, 80, 26, 80, 27, 80, 28, 80, 29, 80, 30, 80, 31, 80, 32, 80, 33, 80, 34, 80, 35, 80, 36, 80, 37, 80, 38, 80, 39, 80, 40, 80, 41, 80, 42, 80, 43, 80, 44, 80, 45, 80, 46, 80, 47, 80, 48, 80, 49, 80, 50, 80, 51, 80, 52, 80, 53, 80, 54, 80, 55, 80, 56, 80, 57, 80, 58, 80, 59, 80, 60, 80, 61, 80, 62, 80, 63, 80, 101, 52, 102, 52, 64, 216, 116, 222, 64, 216, 117, 222, 64, 216, 153, 222, 64, 216, 158, 222, 64, 216, 160, 222, 86, 80, 90, 81, 91, 81, 65, 216, 145, 220, 65, 216, 146, 220, 122, 81, 124, 81, 147, 81, 148, 81, 161, 81, 162, 81, 163, 81, 164, 81, 165, 81, 166, 81, 167, 81, 65, 216, 142, 221, 196, 81, 197, 81, 198, 81, 199, 81, 200, 81, 201, 81, 202, 81, 203, 81, 204, 81, 205, 81, 206, 81, 65, 216, 223, 221, 65, 216, 224, 221, 65, 216, 86, 222, 82, 82, 83, 82, 84, 82, 60, 0, 42, 0, 85, 82, 86, 82, 87, 82, 88, 82, 89, 82, 90, 82, 91, 82, 92, 82, 93, 82, 94, 82, 95, 82, 96, 82, 97, 82, 98, 82, 99, 82, 100, 82, 101, 82, 102, 82, 103, 82, 65, 216, 49, 223, 204, 82, 205, 82, 206, 82, 207, 82, 208, 82, 209, 82, 66, 216, 115, 220, 14, 83, 42, 83, 43, 83, 168, 159, 127, 83, 156, 83, 157, 83, 158, 83, 159, 83, 82, 134, 66, 216, 111, 222, 222, 83, 223, 83, 224, 84, 225, 84, 226, 84, 227, 84, 228, 84, 229, 84, 230, 84, 231, 84, 232, 84, 233, 84, 234, 84, 235, 84, 236, 84, 237, 84, 238, 84, 239, 84, 240, 84, 241, 84, 242, 84, 243, 84, 244, 84, 245, 84, 246, 84, 247, 84, 248, 84, 249, 84, 250, 84, 251, 84, 252, 84, 253, 84, 254, 84, 255, 84, 0, 85, 1, 85, 2, 85, 3, 85, 4, 85, 5, 85, 6, 85, 7, 85, 8, 85, 9, 85, 10, 85, 11, 85, 13, 85, 14, 85, 15, 85, 16, 85, 17, 85, 18, 85, 19, 85, 20, 85, 21, 85, 22, 85, 23, 85, 24, 85, 25, 85, 26, 85, 27, 85, 28, 85, 29, 85, 30, 85, 31, 85, 32, 85, 33, 85, 34, 85, 35, 85, 36, 85, 37, 85, 38, 85, 39, 85, 151, 53, 152, 53, 67, 216, 181, 220, 67, 216, 184, 220, 67, 216, 207, 220, 67, 216, 211, 220, 67, 216, 212, 220, 67, 216, 213, 220, 67, 216, 214, 220, 67, 216, 221, 220, 67, 216, 237, 220, 67, 216, 255, 220, 67, 216, 21, 221, 1, 87, 2, 87, 3, 87, 4, 87, 60, 0, 42, 0, 5, 87, 6, 87, 35, 54, 182, 87, 183, 87, 184, 87, 185, 87, 186, 87, 187, 87, 188, 87, 189, 87, 190, 87, 191, 87, 192, 87, 193, 87, 194, 87, 195, 87, 196, 87, 197, 87, 198, 87, 199, 87, 200, 87, 201, 87, 202, 87, 203, 87, 204, 87, 205, 87, 207, 87, 208, 87, 209, 87, 210, 87, 211, 87, 212, 87, 213, 87, 214, 87, 215, 87, 216, 87, 217, 87, 218, 87, 219, 87, 15, 250, 68, 216, 254, 222, 68, 216, 2, 223, 68, 216, 3, 223, 68, 216, 4, 223, 68, 216, 5, 223, 50, 88, 246, 88, 14, 89, 15, 89, 30, 89, 69, 216, 150, 221, 74, 89, 87, 89, 88, 89, 89, 89, 90, 89, 236, 89, 9, 90, 10, 90, 11, 90, 12, 90, 14, 90, 15, 90, 16, 90, 17, 90, 18, 90, 19, 90, 20, 90, 21, 90, 22, 90, 23, 90, 24, 90, 25, 90, 26, 90, 27, 90, 28, 90, 29, 90, 30, 90, 31, 90, 32, 90, 33, 90, 34, 90, 35, 90, 36, 90, 37, 90, 38, 90, 39, 90, 40, 90, 41, 90, 42, 90, 45, 90, 46, 90, 47, 90, 48, 90, 49, 90, 50, 90, 51, 90, 52, 90, 206, 54, 209, 54, 211, 54, 212, 54, 214, 54, 221, 54, 225, 54, 226, 54, 69, 216, 58, 223, 69, 216, 59, 223, 69, 216, 60, 223, 69, 216, 87, 223, 107, 91, 108, 91, 109, 91, 67, 55, 70, 216, 106, 221, 167, 91, 172, 91, 173, 91, 174, 91, 175, 91, 176, 91, 177, 91, 178, 91, 179, 91, 180, 91, 181, 91, 182, 91, 60, 0, 42, 0, 183, 91, 184, 91, 185, 91, 186, 91, 187, 91, 188, 91, 189, 91, 190, 91, 3, 92, 4, 92, 5, 92, 80, 92, 81, 92, 82, 92, 83, 92, 84, 92, 85, 92, 86, 92, 87, 92, 88, 92, 151, 55, 71, 216, 112, 220, 232, 92, 233, 92, 234, 92, 235, 92, 236, 92, 237, 92, 238, 92, 239, 92, 240, 92, 241, 92, 242, 92, 243, 92, 244, 92, 245, 92, 246, 92, 247, 92, 249, 92, 250, 92, 251, 92, 252, 92, 253, 92, 254, 92, 255, 92, 0, 93, 1, 93, 2, 93, 3, 93, 4, 93, 5, 93, 214, 55, 71, 216, 235, 221, 71, 216, 249, 221, 238, 93, 43, 56, 252, 93, 40, 94, 41, 94, 42, 94, 43, 94, 44, 94, 45, 94, 46, 94, 47, 94, 48, 94, 49, 94, 72, 216, 199, 220, 72, 216, 193, 221, 72, 216, 195, 221, 167, 94, 168, 94, 169, 94, 170, 94, 171, 94, 172, 94, 173, 94, 174, 94, 175, 94, 253, 94, 9, 95, 48, 95, 49, 95, 50, 95, 51, 95, 103, 95, 104, 95, 73, 216, 145, 220, 142, 95, 143, 95, 144, 95, 145, 95, 146, 95, 147, 95, 149, 95, 73, 216, 237, 220, 65, 96, 75, 96, 79, 96, 80, 96, 85, 96, 86, 96, 89, 96, 90, 96, 93, 96, 99, 96, 101, 96, 103, 96, 105, 96, 109, 96, 111, 96, 115, 96, 116, 96, 117, 96, 118, 96, 119, 96, 32, 57, 43, 57, 126, 96, 128, 96, 129, 96, 130, 96, 131, 96, 132, 96, 133, 96, 135, 96, 136, 96, 139, 96, 60, 0, 42, 0, 140, 96, 141, 96, 142, 96, 143, 96, 145, 96, 146, 96, 147, 96, 148, 96, 149, 96, 150, 96, 151, 96, 153, 96, 154, 96, 155, 96, 156, 96, 157, 96, 158, 96, 159, 96, 162, 96, 166, 96, 167, 96, 169, 96, 173, 96, 174, 96, 175, 96, 49, 57, 51, 57, 52, 57, 54, 57, 55, 57, 56, 57, 57, 57, 73, 216, 122, 222, 73, 216, 150, 222, 73, 216, 152, 222, 25, 98, 68, 98, 69, 98, 70, 98, 71, 98, 242, 98, 243, 98, 255, 98, 8, 99, 16, 99, 25, 99, 26, 99, 27, 99, 236, 57, 40, 99, 41, 99, 42, 99, 43, 99, 44, 99, 45, 99, 46, 99, 47, 99, 48, 99, 49, 99, 51, 99, 52, 99, 53, 99, 54, 99, 55, 99, 56, 99, 57, 99, 58, 99, 60, 99, 61, 99, 62, 99, 63, 99, 64, 99, 65, 99, 66, 99, 67, 99, 68, 99, 69, 99, 70, 99, 71, 99, 72, 99, 73, 99, 74, 99, 75, 99, 76, 99, 77, 99, 78, 99, 79, 99, 80, 99, 81, 99, 82, 99, 83, 99, 84, 99, 85, 99, 86, 99, 87, 99, 88, 99, 89, 99, 90, 99, 91, 99, 92, 99, 93, 99, 94, 99, 95, 99, 96, 99, 97, 99, 98, 99, 99, 99, 100, 99, 248, 57, 74, 216, 63, 223, 74, 216, 67, 223, 74, 216, 106, 223, 228, 99, 70, 101, 71, 101, 72, 101, 73, 101, 74, 101, 75, 101, 76, 101, 135, 58, 136, 58, 137, 58, 76, 216, 75, 220, 138, 101, 139, 101, 153, 101, 154, 101, 60, 0, 42, 0, 193, 101, 194, 101, 195, 101, 196, 101, 197, 101, 198, 101, 202, 101, 201, 58, 64, 102, 65, 102, 66, 102, 67, 102, 68, 102, 69, 102, 70, 102, 71, 102, 72, 102, 73, 102, 74, 102, 75, 102, 76, 102, 77, 102, 78, 102, 79, 102, 80, 102, 81, 102, 82, 102, 83, 102, 84, 102, 85, 102, 86, 102, 76, 216, 164, 221, 76, 216, 165, 221, 95, 102, 96, 102, 248, 102, 250, 102, 251, 102, 18, 103, 19, 103, 20, 103, 21, 103, 52, 59, 53, 59, 23, 103, 189, 103, 225, 103, 244, 103, 18, 104, 19, 104, 20, 104, 21, 104, 22, 104, 23, 104, 24, 104, 25, 104, 26, 104, 27, 104, 28, 104, 29, 104, 30, 104, 31, 104, 32, 104, 33, 104, 34, 104, 35, 104, 36, 104, 37, 104, 38, 104, 39, 104, 40, 104, 41, 104, 42, 104, 43, 104, 44, 104, 45, 104, 46, 104, 47, 104, 48, 104, 49, 104, 50, 104, 51, 104, 52, 104, 53, 104, 54, 104, 55, 104, 56, 104, 57, 104, 58, 104, 59, 104, 60, 104, 61, 104, 62, 104, 63, 104, 64, 104, 65, 104, 66, 104, 67, 104, 68, 104, 69, 104, 70, 104, 71, 104, 72, 104, 73, 104, 74, 104, 75, 104, 76, 104, 77, 104, 78, 104, 79, 104, 80, 104, 81, 104, 83, 104, 84, 104, 85, 104, 86, 104, 87, 104, 88, 104, 89, 104, 90, 104, 91, 104, 92, 104, 93, 104, 94, 104, 95, 104, 96, 104, 97, 104, 98, 104, 60, 0, 42, 0, 99, 104, 100, 104, 101, 104, 102, 104, 103, 104, 104, 104, 105, 104, 106, 104, 88, 59, 89, 59, 90, 59, 91, 59, 92, 59, 93, 59, 94, 59, 95, 59, 96, 59, 97, 59, 98, 59, 99, 59, 100, 59, 101, 59, 102, 59, 103, 59, 77, 216, 63, 220, 77, 216, 80, 220, 77, 216, 111, 220, 77, 216, 114, 220, 43, 107, 44, 107, 45, 107, 46, 107, 47, 107, 48, 107, 49, 107, 52, 107, 108, 107, 109, 107, 136, 107, 137, 107, 138, 107, 139, 107, 97, 60, 183, 107, 217, 107, 226, 107, 227, 107, 228, 107, 229, 107, 230, 107, 231, 107, 232, 107, 233, 107, 234, 107, 35, 108, 36, 108, 37, 108, 38, 108, 39, 108, 40, 108, 41, 108, 240, 108, 28, 109, 47, 109, 70, 109, 13, 109, 22, 109, 86, 109, 87, 109, 88, 109, 89, 109, 90, 109, 91, 109, 92, 109, 93, 109, 94, 109, 95, 109, 96, 109, 97, 109, 98, 109, 99, 109, 100, 109, 101, 109, 102, 109, 103, 109, 104, 109, 105, 109, 106, 109, 107, 109, 108, 109, 109, 109, 110, 109, 111, 109, 112, 109, 113, 109, 114, 109, 115, 109, 116, 109, 117, 109, 118, 109, 119, 109, 120, 109, 121, 109, 122, 109, 123, 109, 124, 109, 125, 109, 126, 109, 127, 109, 128, 109, 129, 109, 130, 109, 131, 109, 132, 109, 133, 109, 134, 109, 135, 109, 136, 109, 137, 109, 138, 109, 139, 109, 140, 109, 141, 109, 143, 109, 144, 109, 145, 109, 146, 109, 147, 109, 60, 0, 42, 0, 148, 109, 149, 109, 150, 109, 151, 109, 152, 109, 154, 109, 155, 109, 156, 109, 157, 109, 158, 109, 159, 109, 160, 109, 161, 109, 162, 109, 163, 109, 164, 109, 165, 109, 166, 109, 167, 109, 168, 109, 169, 109, 239, 60, 79, 216, 64, 221, 79, 216, 91, 221, 196, 112, 197, 112, 198, 112, 199, 112, 200, 112, 201, 112, 202, 112, 203, 112, 204, 112, 205, 112, 206, 112, 207, 112, 208, 112, 209, 112, 210, 112, 211, 112, 212, 112, 213, 112, 214, 112, 215, 112, 216, 112, 217, 112, 218, 112, 219, 112, 220, 112, 221, 112, 222, 112, 223, 112, 224, 112, 225, 112, 226, 112, 227, 112, 228, 112, 229, 112, 230, 112, 231, 112, 232, 112, 233, 112, 234, 112, 235, 112, 236, 112, 237, 112, 238, 112, 180, 61, 80, 216, 252, 221, 80, 216, 27, 222, 49, 114, 57, 114, 66, 114, 81, 216, 120, 221, 32, 62, 118, 114, 119, 114, 120, 114, 121, 114, 122, 114, 243, 114, 244, 114, 245, 114, 246, 114, 247, 114, 248, 114, 249, 114, 250, 114, 251, 114, 252, 114, 253, 114, 254, 114, 0, 115, 1, 115, 2, 115, 3, 115, 116, 62, 134, 115, 186, 115, 188, 115, 184, 62, 210, 115, 211, 115, 212, 115, 213, 115, 214, 115, 215, 115, 216, 115, 217, 115, 218, 115, 219, 115, 220, 115, 221, 115, 222, 115, 223, 115, 224, 115, 225, 115, 226, 115, 227, 115, 228, 115, 229, 115, 230, 115, 231, 115, 232, 115, 233, 115, 234, 115, 235, 115, 236, 115, 60, 0, 42, 0, 237, 115, 238, 115, 239, 115, 240, 115, 241, 115, 242, 115, 9, 116, 191, 62, 194, 62, 199, 62, 82, 216, 62, 221, 82, 216, 63, 221, 82, 216, 64, 221, 82, 216, 65, 221, 82, 216, 66, 221, 82, 216, 67, 221, 249, 115, 222, 116, 223, 116, 243, 116, 244, 116, 245, 116, 33, 117, 84, 117, 85, 117, 86, 117, 87, 117, 88, 117, 89, 117, 90, 117, 91, 117, 92, 117, 93, 117, 94, 117, 95, 117, 141, 117, 176, 117, 177, 117, 178, 117, 179, 117, 180, 117, 182, 117, 183, 117, 184, 117, 185, 117, 187, 117, 188, 117, 189, 117, 190, 117, 191, 117, 192, 117, 193, 117, 194, 117, 195, 117, 196, 117, 197, 117, 198, 117, 199, 117, 200, 117, 201, 117, 96, 117, 138, 118, 139, 118, 140, 118, 141, 118, 83, 216, 92, 223, 176, 118, 177, 118, 237, 63, 201, 118, 202, 118, 203, 118, 204, 118, 205, 118, 206, 118, 207, 118, 208, 118, 209, 118, 0, 64, 84, 216, 82, 220, 14, 119, 15, 119, 16, 119, 17, 119, 18, 119, 19, 119, 20, 119, 21, 119, 22, 119, 23, 119, 24, 119, 25, 119, 26, 119, 27, 119, 28, 119, 29, 119, 30, 119, 31, 119, 32, 119, 33, 119, 34, 119, 35, 119, 36, 119, 39, 119, 40, 119, 41, 119, 42, 119, 43, 119, 44, 119, 63, 119, 38, 64, 84, 216, 72, 221, 221, 119, 233, 119, 29, 120, 30, 120, 31, 120, 32, 120, 33, 120, 34, 120, 35, 120, 36, 120, 37, 120, 39, 120, 40, 120, 41, 120, 42, 120, 43, 120, 60, 0, 42, 0, 44, 120, 45, 120, 46, 120, 47, 120, 48, 120, 49, 120, 50, 120, 51, 120, 52, 120, 53, 120, 54, 120, 55, 120, 56, 120, 57, 120, 58, 120, 59, 120, 60, 120, 61, 120, 62, 120, 63, 120, 64, 120, 65, 120, 168, 64, 85, 216, 70, 220, 85, 216, 108, 220, 85, 216, 110, 220, 79, 121, 80, 121, 81, 121, 82, 121, 83, 121, 84, 121, 85, 121, 86, 121, 87, 121, 88, 121, 89, 121, 90, 121, 91, 121, 92, 121, 93, 121, 94, 121, 95, 121, 96, 121, 98, 121, 2, 65, 3, 65, 85, 216, 81, 222, 85, 216, 61, 223, 216, 121, 217, 121, 218, 121, 219, 121, 220, 121, 221, 121, 222, 121, 223, 121, 224, 121, 225, 121, 226, 121, 227, 121, 228, 121, 229, 121, 230, 121, 231, 121, 232, 121, 233, 121, 234, 121, 235, 121, 236, 121, 237, 121, 238, 121, 239, 121, 240, 121, 132, 122, 133, 122, 134, 122, 135, 122, 136, 122, 137, 122, 138, 122, 139, 122, 140, 122, 141, 122, 142, 122, 86, 216, 86, 221, 216, 122, 217, 122, 218, 122, 219, 122, 220, 122, 221, 122, 222, 122, 202, 65, 5, 123, 6, 123, 7, 123, 8, 123, 9, 123, 10, 123, 11, 123, 12, 123, 13, 123, 14, 123, 15, 123, 16, 123, 17, 123, 18, 123, 19, 123, 20, 123, 21, 123, 215, 65, 219, 65, 4, 123, 131, 124, 132, 124, 133, 124, 134, 124, 135, 124, 136, 124, 137, 124, 138, 124, 139, 124, 140, 124, 141, 124, 142, 124, 143, 124, 144, 124, 60, 0, 42, 0, 145, 124, 10, 125, 11, 125, 12, 125, 13, 125, 14, 125, 15, 125, 16, 125, 17, 125, 18, 125, 19, 125, 20, 125, 21, 125, 22, 125, 23, 125, 24, 125, 25, 125, 26, 125, 27, 125, 28, 125, 29, 125, 30, 125, 31, 125, 32, 125, 33, 125, 34, 125, 35, 125, 36, 125, 37, 125, 38, 125, 39, 125, 188, 66, 224, 126, 225, 126, 226, 126, 227, 126, 228, 126, 229, 126, 230, 126, 231, 126, 232, 126, 57, 127, 58, 127, 60, 127, 67, 67, 91, 127, 92, 127, 93, 127, 94, 127, 95, 127, 96, 127, 97, 127, 98, 127, 144, 127, 146, 127, 147, 127, 148, 127, 150, 127, 151, 127, 152, 127, 153, 127, 192, 127, 193, 127, 194, 127, 195, 127, 196, 127, 197, 127, 198, 127, 89, 216, 2, 220, 89, 216, 16, 220, 89, 216, 17, 220, 89, 216, 18, 220, 4, 128, 6, 128, 10, 128, 21, 128, 22, 128, 23, 128, 24, 128, 25, 128, 56, 128, 57, 128, 58, 128, 59, 128, 60, 128, 61, 128, 62, 128, 63, 128, 64, 128, 65, 128, 66, 128, 129, 128, 130, 128, 237, 128, 238, 128, 239, 128, 240, 128, 241, 128, 242, 128, 243, 128, 244, 128, 245, 128, 246, 128, 247, 128, 248, 128, 249, 128, 250, 128, 251, 128, 252, 128, 253, 128, 255, 128, 0, 129, 1, 129, 2, 129, 3, 129, 4, 129, 5, 129, 6, 129, 7, 129, 8, 129, 10, 129, 11, 129, 12, 129, 13, 129, 14, 129, 15, 129, 16, 129, 17, 129, 60, 0, 42, 0, 18, 129, 19, 129, 237, 67, 89, 216, 175, 222, 89, 216, 177, 222, 89, 216, 181, 222, 236, 129, 237, 129, 90, 216, 57, 221, 0, 130, 1, 130, 16, 130, 37, 130, 38, 130, 39, 130, 40, 130, 41, 130, 42, 130, 43, 130, 44, 130, 45, 130, 46, 130, 47, 130, 48, 130, 49, 130, 101, 68, 115, 130, 187, 130, 159, 68, 160, 68, 162, 68, 8, 131, 22, 131, 23, 131, 24, 131, 25, 131, 26, 131, 27, 131, 28, 131, 30, 131, 31, 131, 32, 131, 33, 131, 34, 131, 36, 131, 37, 131, 38, 131, 39, 131, 40, 131, 41, 131, 42, 131, 43, 131, 44, 131, 45, 131, 46, 131, 47, 131, 48, 131, 49, 131, 50, 131, 51, 131, 52, 131, 53, 131, 54, 131, 55, 131, 56, 131, 57, 131, 60, 131, 61, 131, 63, 131, 64, 131, 65, 131, 66, 131, 67, 131, 68, 131, 69, 131, 71, 131, 72, 131, 73, 131, 74, 131, 75, 131, 76, 131, 77, 131, 78, 131, 79, 131, 80, 131, 81, 131, 82, 131, 84, 131, 85, 131, 86, 131, 87, 131, 88, 131, 90, 131, 91, 131, 92, 131, 93, 131, 94, 131, 95, 131, 96, 131, 97, 131, 98, 131, 99, 131, 100, 131, 101, 131, 102, 131, 103, 131, 104, 131, 105, 131, 106, 131, 108, 131, 109, 131, 110, 131, 111, 131, 90, 216, 117, 223, 90, 216, 130, 223, 90, 216, 150, 223, 90, 216, 151, 223, 90, 216, 157, 223, 90, 216, 179, 223, 83, 131, 81, 134, 83, 134, 84, 134, 132, 134, 133, 134, 134, 134, 135, 134, 137, 134, 60, 0, 42, 0, 138, 134, 139, 134, 140, 134, 141, 134, 142, 134, 143, 134, 144, 134, 145, 134, 146, 134, 147, 134, 148, 134, 149, 134, 150, 134, 151, 134, 152, 134, 153, 134, 154, 134, 155, 134, 156, 134, 157, 134, 158, 134, 159, 134, 160, 134, 161, 134, 162, 134, 163, 134, 164, 134, 165, 134, 166, 134, 167, 134, 168, 134, 169, 134, 170, 134, 172, 134, 67, 136, 68, 136, 79, 136, 21, 70, 109, 136, 110, 136, 111, 136, 112, 136, 113, 136, 114, 136, 115, 136, 116, 136, 117, 136, 118, 136, 119, 136, 120, 136, 121, 136, 122, 136, 123, 136, 124, 136, 125, 136, 126, 136, 127, 136, 128, 136, 129, 136, 130, 136, 131, 136, 132, 136, 133, 136, 134, 136, 135, 136, 93, 216, 85, 222, 93, 216, 86, 222, 93, 216, 87, 222, 141, 137, 142, 137, 202, 137, 9, 138, 10, 138, 11, 138, 12, 138, 13, 138, 14, 138, 15, 138, 16, 138, 17, 138, 18, 138, 19, 138, 20, 138, 21, 138, 22, 138, 23, 138, 24, 138, 25, 138, 26, 138, 94, 216, 103, 221, 247, 139, 248, 139, 249, 139, 250, 139, 251, 139, 252, 139, 253, 139, 254, 139, 255, 139, 0, 140, 1, 140, 2, 140, 3, 140, 4, 140, 5, 140, 6, 140, 7, 140, 8, 140, 9, 140, 10, 140, 56, 140, 71, 140, 72, 140, 87, 140, 121, 140, 122, 140, 123, 140, 161, 140, 162, 140, 163, 140, 164, 140, 95, 216, 47, 221, 60, 141, 61, 141, 62, 141, 63, 141, 64, 141, 65, 141, 60, 0, 42, 0, 66, 141, 67, 141, 68, 141, 69, 141, 70, 141, 118, 141, 119, 141, 120, 141, 152, 71, 181, 141, 182, 141, 183, 141, 184, 141, 213, 71, 172, 142, 206, 142, 207, 142, 208, 142, 209, 142, 210, 142, 211, 142, 212, 142, 213, 142, 124, 143, 125, 143, 126, 143, 127, 143, 128, 143, 129, 143, 130, 143, 131, 143, 177, 143, 244, 143, 245, 143, 246, 143, 247, 143, 248, 143, 249, 143, 250, 143, 251, 143, 252, 143, 253, 143, 254, 143, 255, 143, 0, 144, 1, 144, 2, 144, 3, 144, 4, 144, 5, 144, 6, 144, 7, 144, 8, 144, 9, 144, 10, 144, 149, 144, 214, 144, 215, 144, 216, 144, 217, 144, 218, 144, 219, 144, 220, 144, 221, 144, 222, 144, 223, 144, 224, 144, 225, 144, 226, 144, 227, 144, 228, 144, 229, 144, 230, 144, 231, 144, 76, 145, 77, 145, 78, 145, 79, 145, 80, 145, 81, 145, 82, 145, 213, 145, 214, 145, 215, 145, 216, 145, 217, 145, 218, 145, 219, 145, 220, 145, 221, 145, 222, 145, 223, 145, 224, 145, 225, 145, 226, 145, 176, 148, 177, 148, 178, 148, 179, 148, 180, 148, 181, 148, 182, 148, 183, 148, 184, 148, 185, 148, 186, 148, 187, 148, 188, 148, 189, 148, 190, 148, 191, 148, 192, 148, 193, 148, 194, 148, 195, 148, 196, 148, 197, 148, 198, 148, 199, 148, 200, 148, 201, 148, 202, 148, 203, 148, 204, 148, 205, 148, 206, 148, 131, 149, 132, 149, 60, 0, 42, 0, 133, 149, 99, 216, 202, 220, 99, 216, 205, 220, 3, 150, 4, 150, 5, 150, 6, 150, 86, 150, 87, 150, 88, 150, 89, 150, 91, 150, 92, 150, 93, 150, 94, 150, 95, 150, 96, 150, 97, 150, 98, 150, 99, 150, 100, 150, 101, 150, 102, 150, 103, 150, 104, 150, 105, 150, 90, 150, 99, 216, 151, 222, 186, 150, 187, 150, 188, 150, 189, 150, 190, 150, 124, 152, 125, 152, 126, 152, 127, 152, 128, 152, 129, 152, 130, 152, 131, 152, 132, 152, 226, 152, 227, 152, 228, 152, 125, 153, 126, 153, 127, 153, 128, 153, 129, 153, 130, 153, 172, 153, 138, 154, 139, 154, 140, 154, 141, 154, 142, 154, 143, 154, 168, 154, 216, 154, 223, 154, 37, 155, 47, 155, 50, 155, 60, 155, 125, 156, 42, 158, 43, 158, 44, 158, 45, 158, 46, 158, 47, 158, 48, 158, 49, 158, 50, 158, 51, 158, 52, 158, 53, 158, 54, 158, 128, 159, 60, 0, 39, 0, 208, 253, 11, 40, 39, 0, 60, 0, 42, 0, 126, 78, 127, 78, 128, 78, 34, 52, 64, 80, 65, 80, 66, 80, 67, 80, 68, 80, 69, 80, 70, 80, 71, 80, 72, 80, 73, 80, 74, 80, 75, 80, 76, 80, 77, 80, 78, 80, 79, 80, 80, 80, 81, 80, 82, 80, 83, 80, 84, 80, 85, 80, 87, 80, 88, 80, 89, 80, 90, 80, 91, 80, 92, 80, 93, 80, 94, 80, 95, 80, 96, 80, 97, 80, 98, 80, 99, 80, 100, 80, 101, 80, 102, 80, 103, 80, 105, 80, 106, 80, 107, 80, 108, 80, 109, 80, 110, 80, 111, 80, 112, 80, 113, 80, 114, 80, 115, 80, 116, 80, 117, 80, 118, 80, 119, 80, 120, 80, 121, 80, 122, 80, 123, 80, 124, 80, 125, 80, 126, 80, 127, 80, 100, 52, 64, 216, 183, 222, 64, 216, 191, 222, 64, 216, 192, 222, 64, 216, 229, 222, 92, 81, 93, 81, 94, 81, 125, 81, 149, 81, 188, 52, 168, 81, 207, 81, 208, 81, 209, 81, 65, 216, 235, 221, 240, 81, 104, 82, 106, 82, 107, 82, 108, 82, 109, 82, 110, 82, 111, 82, 112, 82, 113, 82, 118, 82, 65, 216, 121, 223, 210, 82, 211, 82, 212, 82, 213, 82, 214, 82, 216, 82, 217, 82, 218, 82, 32, 53, 15, 83, 16, 83, 19, 83, 24, 83, 25, 83, 45, 53, 44, 83, 45, 83, 46, 83, 49, 53, 62, 83, 63, 83, 64, 83, 89, 83, 104, 83, 126, 83, 160, 83, 161, 83, 162, 83, 163, 83, 169, 83, 195, 83, 196, 83, 12, 85, 40, 85, 60, 0, 42, 0, 41, 85, 42, 85, 43, 85, 44, 85, 45, 85, 46, 85, 47, 85, 48, 85, 49, 85, 50, 85, 51, 85, 52, 85, 53, 85, 54, 85, 55, 85, 56, 85, 57, 85, 58, 85, 59, 85, 60, 85, 61, 85, 62, 85, 63, 85, 64, 85, 65, 85, 66, 85, 67, 85, 68, 85, 69, 85, 70, 85, 71, 85, 72, 85, 73, 85, 74, 85, 75, 85, 76, 85, 77, 85, 78, 85, 79, 85, 80, 85, 81, 85, 82, 85, 83, 85, 84, 85, 85, 85, 86, 85, 87, 85, 88, 85, 90, 85, 91, 85, 92, 85, 93, 85, 94, 85, 96, 85, 97, 85, 98, 85, 99, 85, 100, 85, 101, 85, 102, 85, 103, 85, 104, 85, 105, 85, 106, 85, 108, 85, 109, 85, 110, 85, 111, 85, 112, 85, 113, 85, 114, 85, 115, 85, 116, 85, 117, 85, 118, 85, 119, 85, 120, 85, 121, 85, 161, 53, 165, 53, 173, 53, 67, 216, 40, 221, 67, 216, 49, 221, 67, 216, 50, 221, 67, 216, 70, 221, 67, 216, 71, 221, 67, 216, 72, 221, 67, 216, 73, 221, 67, 216, 76, 221, 67, 216, 77, 221, 67, 216, 78, 221, 67, 216, 111, 221, 67, 216, 113, 221, 67, 216, 116, 221, 67, 216, 124, 221, 67, 216, 126, 221, 67, 216, 127, 221, 67, 216, 150, 221, 67, 216, 156, 221, 67, 216, 167, 221, 67, 216, 178, 221, 107, 85, 182, 85, 7, 87, 8, 87, 9, 87, 10, 87, 11, 87, 15, 87, 220, 87, 221, 87, 222, 87, 223, 87, 224, 87, 225, 87, 226, 87, 227, 87, 228, 87, 229, 87, 230, 87, 231, 87, 232, 87, 233, 87, 235, 87, 236, 87, 237, 87, 238, 87, 239, 87, 60, 0, 42, 0, 240, 87, 241, 87, 242, 87, 243, 87, 244, 87, 245, 87, 246, 87, 247, 87, 248, 87, 249, 87, 250, 87, 251, 87, 252, 87, 253, 87, 254, 87, 255, 87, 0, 88, 1, 88, 2, 88, 3, 88, 4, 88, 5, 88, 6, 88, 7, 88, 8, 88, 9, 88, 10, 88, 11, 88, 12, 88, 13, 88, 14, 88, 15, 88, 16, 88, 17, 88, 18, 88, 19, 88, 20, 88, 21, 88, 71, 54, 72, 54, 73, 54, 68, 216, 54, 223, 68, 216, 58, 223, 234, 87, 53, 88, 247, 88, 248, 88, 31, 89, 32, 89, 91, 89, 92, 89, 93, 89, 94, 89, 69, 216, 62, 222, 95, 89, 98, 89, 43, 90, 61, 90, 44, 90, 53, 90, 54, 90, 55, 90, 56, 90, 57, 90, 58, 90, 59, 90, 60, 90, 62, 90, 63, 90, 64, 90, 65, 90, 66, 90, 67, 90, 68, 90, 69, 90, 70, 90, 71, 90, 72, 90, 73, 90, 74, 90, 75, 90, 76, 90, 77, 90, 78, 90, 79, 90, 80, 90, 81, 90, 82, 90, 83, 90, 84, 90, 85, 90, 86, 90, 87, 90, 88, 90, 89, 90, 90, 90, 91, 90, 92, 90, 93, 90, 94, 90, 95, 90, 96, 90, 97, 90, 98, 90, 99, 90, 100, 90, 101, 90, 102, 90, 103, 90, 104, 90, 105, 90, 106, 90, 107, 90, 108, 90, 109, 90, 110, 90, 111, 90, 112, 90, 113, 90, 114, 90, 115, 90, 116, 90, 117, 90, 118, 90, 229, 54, 230, 54, 69, 216, 108, 223, 69, 216, 109, 223, 60, 0, 42, 0, 69, 216, 110, 223, 69, 216, 111, 223, 69, 216, 112, 223, 69, 216, 113, 223, 69, 216, 115, 223, 69, 216, 116, 223, 142, 90, 110, 91, 111, 91, 112, 91, 114, 91, 191, 91, 192, 91, 193, 91, 194, 91, 195, 91, 196, 91, 197, 91, 198, 91, 199, 91, 200, 91, 201, 91, 70, 216, 45, 222, 70, 216, 52, 222, 7, 92, 8, 92, 9, 92, 89, 92, 90, 92, 91, 92, 92, 92, 93, 92, 96, 92, 6, 93, 7, 93, 8, 93, 9, 93, 10, 93, 11, 93, 12, 93, 13, 93, 14, 93, 15, 93, 16, 93, 17, 93, 18, 93, 19, 93, 20, 93, 21, 93, 22, 93, 23, 93, 24, 93, 25, 93, 26, 93, 27, 93, 28, 93, 29, 93, 30, 93, 31, 93, 32, 93, 33, 93, 34, 93, 35, 93, 36, 93, 37, 93, 38, 93, 39, 93, 40, 93, 41, 93, 42, 93, 43, 93, 44, 93, 45, 93, 46, 93, 47, 93, 48, 93, 71, 216, 28, 222, 71, 216, 35, 222, 71, 216, 55, 222, 71, 216, 61, 222, 226, 93, 227, 93, 49, 56, 50, 94, 51, 94, 52, 94, 53, 94, 54, 94, 55, 94, 56, 94, 57, 94, 58, 94, 59, 94, 60, 94, 62, 94, 177, 94, 178, 94, 179, 94, 180, 94, 181, 94, 182, 94, 183, 94, 184, 94, 185, 94, 186, 94, 187, 94, 188, 94, 190, 94, 52, 95, 53, 95, 54, 95, 55, 95, 56, 95, 57, 95, 72, 216, 250, 223, 87, 95, 105, 95, 107, 95, 108, 95, 150, 95, 151, 95, 152, 95, 153, 95, 155, 95, 156, 95, 157, 95, 158, 95, 159, 95, 60, 0, 42, 0, 160, 95, 161, 95, 162, 95, 163, 95, 164, 95, 73, 216, 19, 221, 73, 216, 27, 221, 127, 96, 134, 96, 137, 96, 138, 96, 144, 96, 152, 96, 160, 96, 161, 96, 163, 96, 164, 96, 165, 96, 168, 96, 170, 96, 171, 96, 172, 96, 48, 57, 50, 57, 53, 57, 59, 57, 176, 96, 177, 96, 180, 96, 181, 96, 183, 96, 184, 96, 186, 96, 187, 96, 188, 96, 189, 96, 190, 96, 191, 96, 192, 96, 194, 96, 195, 96, 197, 96, 198, 96, 199, 96, 200, 96, 202, 96, 203, 96, 205, 96, 207, 96, 208, 96, 211, 96, 212, 96, 213, 96, 215, 96, 216, 96, 217, 96, 218, 96, 219, 96, 220, 96, 221, 96, 222, 96, 223, 96, 228, 96, 230, 96, 231, 96, 232, 96, 236, 96, 237, 96, 238, 96, 239, 96, 63, 57, 77, 57, 73, 216, 244, 222, 73, 216, 245, 222, 73, 216, 246, 222, 73, 216, 18, 223, 73, 216, 20, 223, 73, 216, 27, 223, 73, 216, 31, 223, 26, 98, 27, 98, 28, 98, 29, 98, 72, 98, 50, 99, 59, 99, 101, 99, 102, 99, 103, 99, 104, 99, 105, 99, 106, 99, 107, 99, 108, 99, 109, 99, 110, 99, 111, 99, 112, 99, 113, 99, 114, 99, 115, 99, 116, 99, 117, 99, 118, 99, 119, 99, 120, 99, 121, 99, 122, 99, 123, 99, 124, 99, 125, 99, 126, 99, 127, 99, 128, 99, 129, 99, 130, 99, 131, 99, 132, 99, 133, 99, 134, 99, 135, 99, 136, 99, 137, 99, 138, 99, 139, 99, 141, 99, 142, 99, 143, 99, 60, 0, 42, 0, 144, 99, 145, 99, 146, 99, 147, 99, 149, 99, 150, 99, 151, 99, 152, 99, 153, 99, 154, 99, 155, 99, 156, 99, 157, 99, 158, 99, 159, 99, 160, 99, 161, 99, 162, 99, 164, 99, 165, 99, 166, 99, 167, 99, 168, 99, 169, 99, 170, 99, 171, 99, 172, 99, 173, 99, 174, 99, 175, 99, 179, 99, 180, 99, 181, 99, 182, 99, 183, 99, 184, 99, 185, 99, 186, 99, 187, 99, 188, 99, 189, 99, 251, 57, 254, 57, 0, 58, 1, 58, 2, 58, 3, 58, 4, 58, 6, 58, 74, 216, 202, 223, 74, 216, 206, 223, 178, 99, 95, 85, 77, 101, 78, 101, 79, 101, 80, 101, 81, 101, 82, 101, 83, 101, 84, 101, 85, 101, 86, 101, 87, 101, 88, 101, 89, 101, 90, 101, 91, 101, 138, 58, 139, 58, 140, 58, 141, 58, 142, 58, 93, 101, 141, 101, 142, 101, 143, 101, 155, 101, 156, 101, 172, 101, 173, 101, 199, 101, 200, 101, 201, 101, 203, 101, 204, 101, 205, 101, 206, 101, 207, 101, 203, 58, 227, 101, 215, 82, 87, 102, 88, 102, 89, 102, 90, 102, 91, 102, 92, 102, 93, 102, 94, 102, 97, 102, 98, 102, 99, 102, 100, 102, 101, 102, 102, 102, 103, 102, 104, 102, 105, 102, 253, 102, 240, 58, 242, 58, 76, 216, 179, 221, 76, 216, 200, 221, 76, 216, 201, 221, 249, 102, 252, 102, 22, 103, 24, 103, 25, 103, 26, 103, 27, 103, 54, 59, 55, 59, 107, 104, 108, 104, 109, 104, 110, 104, 60, 0, 42, 0, 111, 104, 112, 104, 113, 104, 114, 104, 115, 104, 116, 104, 117, 104, 118, 104, 119, 104, 120, 104, 121, 104, 122, 104, 123, 104, 124, 104, 125, 104, 126, 104, 127, 104, 128, 104, 129, 104, 130, 104, 131, 104, 132, 104, 133, 104, 134, 104, 135, 104, 136, 104, 137, 104, 138, 104, 139, 104, 140, 104, 141, 104, 142, 104, 143, 104, 144, 104, 145, 104, 146, 104, 147, 104, 148, 104, 149, 104, 150, 104, 151, 104, 152, 104, 153, 104, 154, 104, 155, 104, 156, 104, 157, 104, 158, 104, 159, 104, 160, 104, 161, 104, 162, 104, 163, 104, 164, 104, 165, 104, 166, 104, 167, 104, 168, 104, 169, 104, 170, 104, 171, 104, 172, 104, 173, 104, 174, 104, 175, 104, 176, 104, 177, 104, 178, 104, 179, 104, 181, 104, 182, 104, 183, 104, 184, 104, 185, 104, 186, 104, 187, 104, 188, 104, 189, 104, 190, 104, 191, 104, 192, 104, 193, 104, 194, 104, 104, 59, 105, 59, 106, 59, 107, 59, 108, 59, 109, 59, 110, 59, 111, 59, 112, 59, 113, 59, 114, 59, 115, 59, 116, 59, 117, 59, 119, 59, 86, 105, 50, 107, 51, 107, 53, 107, 54, 107, 55, 107, 56, 107, 47, 60, 140, 107, 141, 107, 142, 107, 143, 107, 144, 107, 145, 107, 146, 107, 147, 107, 98, 60, 184, 107, 185, 107, 186, 107, 187, 107, 235, 107, 236, 107, 237, 107, 238, 107, 42, 108, 43, 108, 79, 216, 99, 220, 142, 109, 235, 60, 60, 0, 42, 0, 153, 109, 170, 109, 171, 109, 172, 109, 173, 109, 174, 109, 175, 109, 176, 109, 177, 109, 178, 109, 179, 109, 180, 109, 181, 109, 182, 109, 183, 109, 184, 109, 185, 109, 186, 109, 187, 109, 188, 109, 189, 109, 190, 109, 191, 109, 192, 109, 193, 109, 194, 109, 195, 109, 196, 109, 197, 109, 198, 109, 199, 109, 200, 109, 201, 109, 202, 109, 203, 109, 204, 109, 205, 109, 206, 109, 207, 109, 208, 109, 209, 109, 210, 109, 211, 109, 212, 109, 213, 109, 214, 109, 215, 109, 216, 109, 217, 109, 218, 109, 219, 109, 220, 109, 221, 109, 222, 109, 223, 109, 224, 109, 225, 109, 226, 109, 227, 109, 228, 109, 229, 109, 230, 109, 231, 109, 232, 109, 233, 109, 234, 109, 235, 109, 236, 109, 237, 109, 238, 109, 239, 109, 240, 109, 241, 109, 242, 109, 243, 109, 244, 109, 245, 109, 246, 109, 247, 109, 248, 109, 249, 109, 250, 109, 251, 109, 253, 109, 255, 109, 0, 110, 1, 110, 2, 110, 4, 110, 5, 110, 6, 110, 7, 110, 8, 110, 9, 110, 10, 110, 11, 110, 12, 110, 13, 110, 14, 110, 15, 110, 16, 110, 17, 110, 18, 110, 19, 110, 20, 110, 21, 110, 22, 110, 23, 110, 26, 110, 116, 110, 0, 61, 4, 61, 79, 216, 126, 221, 79, 216, 143, 221, 79, 216, 182, 221, 79, 216, 183, 221, 79, 216, 184, 221, 79, 216, 185, 221, 79, 216, 186, 221, 79, 216, 187, 221, 79, 216, 188, 221, 79, 216, 189, 221, 79, 216, 227, 221, 79, 216, 248, 221, 239, 112, 240, 112, 241, 112, 242, 112, 60, 0, 42, 0, 243, 112, 244, 112, 245, 112, 246, 112, 247, 112, 248, 112, 249, 112, 250, 112, 252, 112, 253, 112, 254, 112, 255, 112, 0, 113, 1, 113, 2, 113, 3, 113, 4, 113, 5, 113, 6, 113, 7, 113, 8, 113, 9, 113, 10, 113, 11, 113, 12, 113, 13, 113, 14, 113, 15, 113, 16, 113, 17, 113, 18, 113, 19, 113, 21, 113, 22, 113, 23, 113, 24, 113, 191, 61, 80, 216, 75, 222, 80, 216, 86, 222, 80, 216, 89, 222, 20, 113, 61, 114, 123, 114, 124, 114, 125, 114, 126, 114, 127, 114, 129, 114, 58, 62, 60, 62, 62, 62, 64, 62, 81, 216, 101, 222, 255, 114, 4, 115, 5, 115, 7, 115, 8, 115, 9, 115, 10, 115, 13, 115, 14, 115, 15, 115, 16, 115, 17, 115, 19, 115, 20, 115, 21, 115, 22, 115, 23, 115, 24, 115, 25, 115, 26, 115, 27, 115, 28, 115, 29, 115, 30, 115, 31, 115, 32, 115, 33, 115, 42, 115, 135, 115, 136, 115, 243, 115, 244, 115, 245, 115, 246, 115, 248, 115, 250, 115, 251, 115, 252, 115, 253, 115, 254, 115, 255, 115, 0, 116, 1, 116, 2, 116, 3, 116, 4, 116, 5, 116, 6, 116, 7, 116, 8, 116, 10, 116, 11, 116, 12, 116, 13, 116, 14, 116, 15, 116, 16, 116, 17, 116, 18, 116, 19, 116, 202, 62, 204, 62, 208, 62, 82, 216, 98, 221, 82, 216, 99, 221, 82, 216, 116, 221, 82, 216, 117, 221, 82, 216, 118, 221, 224, 116, 14, 63, 246, 116, 247, 116, 248, 116, 38, 63, 27, 117, 60, 0, 42, 0, 28, 117, 34, 117, 35, 117, 97, 117, 98, 117, 99, 117, 100, 117, 101, 117, 102, 117, 103, 117, 105, 117, 112, 117, 181, 117, 202, 117, 203, 117, 204, 117, 205, 117, 206, 117, 207, 117, 208, 117, 209, 117, 210, 117, 211, 117, 212, 117, 213, 117, 214, 117, 137, 118, 142, 118, 143, 118, 144, 118, 145, 118, 178, 118, 210, 118, 211, 118, 212, 118, 213, 118, 214, 118, 215, 118, 216, 118, 219, 118, 1, 64, 2, 64, 37, 119, 38, 119, 45, 119, 46, 119, 47, 119, 48, 119, 49, 119, 50, 119, 51, 119, 52, 119, 53, 119, 54, 119, 55, 119, 56, 119, 57, 119, 58, 119, 59, 119, 60, 119, 61, 119, 62, 119, 65, 119, 84, 216, 125, 221, 84, 216, 126, 221, 64, 119, 234, 119, 235, 119, 38, 120, 66, 120, 67, 120, 68, 120, 69, 120, 70, 120, 71, 120, 72, 120, 73, 120, 74, 120, 75, 120, 76, 120, 77, 120, 78, 120, 79, 120, 80, 120, 81, 120, 82, 120, 83, 120, 84, 120, 85, 120, 86, 120, 87, 120, 88, 120, 89, 120, 90, 120, 91, 120, 85, 216, 154, 220, 97, 121, 99, 121, 100, 121, 101, 121, 103, 121, 104, 121, 105, 121, 106, 121, 107, 121, 108, 121, 109, 121, 110, 121, 111, 121, 4, 65, 150, 137, 85, 216, 131, 222, 187, 121, 241, 121, 242, 121, 243, 121, 244, 121, 245, 121, 246, 121, 247, 121, 248, 121, 249, 121, 250, 121, 251, 121, 252, 121, 253, 121, 254, 121, 6, 122, 60, 0, 42, 0, 59, 65, 65, 65, 67, 65, 69, 65, 143, 122, 144, 122, 145, 122, 146, 122, 147, 122, 148, 122, 149, 122, 154, 122, 225, 122, 22, 123, 23, 123, 24, 123, 25, 123, 26, 123, 27, 123, 28, 123, 29, 123, 30, 123, 31, 123, 32, 123, 33, 123, 34, 123, 35, 123, 36, 123, 37, 123, 38, 123, 39, 123, 40, 123, 41, 123, 42, 123, 43, 123, 44, 123, 45, 123, 46, 123, 47, 123, 48, 123, 49, 123, 50, 123, 51, 123, 52, 123, 53, 123, 54, 123, 55, 123, 56, 123, 57, 123, 58, 123, 59, 123, 60, 123, 61, 123, 62, 123, 104, 117, 146, 124, 147, 124, 148, 124, 149, 124, 150, 124, 151, 124, 152, 124, 153, 124, 154, 124, 155, 124, 156, 124, 157, 124, 163, 124, 87, 216, 73, 222, 40, 125, 41, 125, 44, 125, 45, 125, 46, 125, 47, 125, 48, 125, 49, 125, 50, 125, 51, 125, 52, 125, 53, 125, 54, 125, 55, 125, 56, 125, 57, 125, 58, 125, 59, 125, 60, 125, 61, 125, 62, 125, 63, 125, 64, 125, 65, 125, 66, 125, 67, 125, 68, 125, 69, 125, 70, 125, 71, 125, 72, 125, 73, 125, 74, 125, 75, 125, 76, 125, 87, 216, 225, 223, 87, 216, 226, 223, 233, 126, 234, 126, 235, 126, 236, 126, 237, 126, 238, 126, 239, 126, 240, 126, 241, 126, 242, 126, 243, 126, 244, 126, 245, 126, 246, 126, 247, 126, 248, 126, 249, 126, 250, 126, 251, 126, 252, 126, 253, 126, 254, 126, 60, 0, 42, 0, 255, 126, 0, 127, 1, 127, 59, 127, 61, 127, 68, 67, 69, 67, 99, 127, 149, 127, 154, 127, 155, 127, 156, 127, 157, 127, 158, 127, 159, 127, 199, 127, 200, 127, 201, 127, 202, 127, 203, 127, 204, 127, 205, 127, 206, 127, 207, 127, 208, 127, 209, 127, 210, 127, 8, 128, 9, 128, 26, 128, 27, 128, 28, 128, 29, 128, 30, 128, 31, 128, 67, 128, 68, 128, 69, 128, 70, 128, 71, 128, 72, 128, 73, 128, 74, 128, 75, 128, 76, 128, 77, 128, 236, 128, 21, 129, 22, 129, 23, 129, 24, 129, 25, 129, 26, 129, 27, 129, 28, 129, 29, 129, 30, 129, 31, 129, 33, 129, 34, 129, 35, 129, 36, 129, 37, 129, 38, 129, 39, 129, 40, 129, 41, 129, 42, 129, 43, 129, 44, 129, 45, 129, 46, 129, 47, 129, 48, 129, 49, 129, 50, 129, 51, 129, 52, 129, 53, 129, 54, 129, 55, 129, 56, 129, 240, 67, 242, 67, 89, 216, 218, 222, 89, 216, 232, 222, 89, 216, 252, 222, 90, 216, 14, 221, 2, 130, 17, 130, 50, 130, 51, 130, 52, 130, 53, 130, 54, 130, 55, 130, 56, 130, 57, 130, 58, 130, 59, 130, 90, 216, 45, 222, 90, 216, 46, 222, 116, 130, 89, 131, 107, 131, 29, 131, 35, 131, 112, 131, 113, 131, 114, 131, 115, 131, 116, 131, 117, 131, 118, 131, 119, 131, 120, 131, 121, 131, 122, 131, 123, 131, 124, 131, 125, 131, 126, 131, 127, 131, 128, 131, 129, 131, 130, 131, 131, 131, 132, 131, 60, 0, 42, 0, 133, 131, 134, 131, 135, 131, 136, 131, 137, 131, 138, 131, 139, 131, 140, 131, 141, 131, 142, 131, 143, 131, 144, 131, 145, 131, 146, 131, 147, 131, 148, 131, 149, 131, 150, 131, 151, 131, 152, 131, 153, 131, 155, 131, 156, 131, 157, 131, 158, 131, 159, 131, 160, 131, 161, 131, 162, 131, 163, 131, 164, 131, 165, 131, 166, 131, 167, 131, 168, 131, 169, 131, 170, 131, 171, 131, 172, 131, 174, 131, 175, 131, 176, 131, 177, 131, 178, 131, 179, 131, 180, 131, 181, 131, 182, 131, 183, 131, 184, 131, 185, 131, 186, 131, 188, 131, 189, 131, 202, 159, 90, 216, 192, 223, 90, 216, 247, 223, 91, 216, 33, 220, 173, 131, 106, 95, 85, 134, 86, 134, 87, 134, 88, 134, 89, 134, 90, 134, 171, 134, 173, 134, 174, 134, 175, 134, 176, 134, 177, 134, 178, 134, 179, 134, 180, 134, 181, 134, 182, 134, 183, 134, 184, 134, 185, 134, 186, 134, 187, 134, 188, 134, 189, 134, 190, 134, 191, 134, 192, 134, 193, 134, 194, 134, 195, 134, 196, 134, 197, 134, 198, 134, 199, 134, 200, 134, 201, 134, 202, 134, 203, 134, 204, 134, 205, 134, 206, 134, 207, 134, 33, 250, 166, 69, 92, 216, 103, 222, 92, 216, 128, 222, 92, 216, 133, 222, 92, 216, 139, 222, 69, 136, 80, 136, 81, 136, 82, 136, 83, 136, 84, 136, 136, 136, 137, 136, 138, 136, 139, 136, 140, 136, 141, 136, 142, 136, 143, 136, 144, 136, 145, 136, 146, 136, 147, 136, 148, 136, 149, 136, 60, 0, 42, 0, 150, 136, 151, 136, 152, 136, 153, 136, 154, 136, 155, 136, 156, 136, 157, 136, 158, 136, 159, 136, 160, 136, 161, 136, 162, 136, 163, 136, 164, 136, 165, 136, 166, 136, 167, 136, 168, 136, 169, 136, 170, 136, 171, 136, 172, 136, 173, 136, 174, 136, 176, 136, 38, 70, 175, 136, 130, 137, 143, 137, 144, 137, 145, 137, 146, 137, 147, 137, 148, 137, 122, 70, 203, 137, 213, 137, 214, 137, 215, 137, 216, 137, 217, 137, 94, 216, 200, 220, 27, 138, 28, 138, 29, 138, 30, 138, 31, 138, 32, 138, 33, 138, 34, 138, 35, 138, 36, 138, 37, 138, 38, 138, 39, 138, 40, 138, 41, 138, 42, 138, 43, 138, 44, 138, 45, 138, 46, 138, 47, 138, 48, 138, 49, 138, 50, 138, 51, 138, 187, 70, 188, 70, 94, 216, 122, 221, 11, 140, 12, 140, 13, 140, 14, 140, 15, 140, 16, 140, 17, 140, 18, 140, 19, 140, 20, 140, 21, 140, 22, 140, 23, 140, 24, 140, 25, 140, 26, 140, 27, 140, 28, 140, 29, 140, 30, 140, 57, 140, 58, 140, 59, 140, 73, 140, 88, 140, 89, 140, 90, 140, 91, 140, 92, 140, 93, 140, 70, 71, 71, 71, 124, 140, 125, 140, 165, 140, 166, 140, 167, 140, 168, 140, 169, 140, 170, 140, 171, 140, 172, 140, 173, 140, 174, 140, 71, 141, 72, 141, 73, 141, 74, 141, 101, 141, 102, 141, 103, 141, 121, 141, 122, 141, 123, 141, 124, 141, 125, 141, 126, 141, 60, 0, 42, 0, 127, 141, 35, 250, 155, 71, 185, 141, 186, 141, 187, 141, 189, 141, 190, 141, 191, 141, 192, 141, 193, 141, 194, 141, 195, 141, 196, 141, 173, 142, 174, 142, 175, 142, 96, 216, 24, 222, 96, 216, 26, 222, 214, 142, 215, 142, 216, 142, 217, 142, 218, 142, 219, 142, 220, 142, 221, 142, 222, 142, 223, 142, 224, 142, 225, 142, 226, 142, 227, 142, 70, 72, 71, 72, 72, 72, 73, 72, 74, 72, 96, 216, 205, 222, 132, 143, 133, 143, 134, 143, 11, 144, 12, 144, 13, 144, 14, 144, 15, 144, 16, 144, 17, 144, 18, 144, 19, 144, 20, 144, 21, 144, 22, 144, 23, 144, 24, 144, 25, 144, 26, 144, 27, 144, 28, 144, 29, 144, 30, 144, 31, 144, 32, 144, 33, 144, 34, 144, 35, 144, 36, 144, 37, 144, 38, 144, 39, 144, 155, 72, 97, 216, 1, 221, 171, 144, 212, 144, 232, 144, 233, 144, 234, 144, 235, 144, 236, 144, 237, 144, 238, 144, 239, 144, 240, 144, 242, 144, 243, 144, 244, 144, 247, 144, 248, 144, 253, 144, 97, 216, 15, 223, 97, 216, 19, 223, 83, 145, 84, 145, 85, 145, 86, 145, 87, 145, 88, 145, 89, 145, 90, 145, 91, 145, 92, 145, 93, 145, 94, 145, 237, 72, 98, 216, 4, 220, 200, 145, 206, 145, 227, 145, 228, 145, 229, 145, 230, 145, 231, 145, 232, 145, 233, 145, 234, 145, 235, 145, 236, 145, 237, 145, 238, 145, 239, 145, 240, 145, 241, 145, 242, 145, 243, 145, 244, 145, 245, 145, 246, 145, 60, 0, 42, 0, 247, 145, 248, 145, 249, 145, 250, 145, 251, 145, 252, 145, 98, 216, 72, 221, 98, 216, 73, 221, 207, 148, 208, 148, 209, 148, 210, 148, 211, 148, 212, 148, 213, 148, 214, 148, 215, 148, 216, 148, 217, 148, 218, 148, 219, 148, 220, 148, 221, 148, 222, 148, 223, 148, 224, 148, 225, 148, 226, 148, 227, 148, 228, 148, 229, 148, 230, 148, 231, 148, 232, 148, 233, 148, 234, 148, 235, 148, 236, 148, 237, 148, 238, 148, 239, 148, 240, 148, 241, 148, 242, 148, 243, 148, 244, 148, 245, 148, 246, 148, 247, 148, 121, 149, 122, 149, 134, 149, 135, 149, 136, 149, 137, 149, 138, 149, 99, 216, 210, 220, 7, 150, 8, 150, 9, 150, 10, 150, 11, 150, 12, 150, 13, 150, 14, 150, 15, 150, 16, 150, 106, 150, 107, 150, 108, 150, 109, 150, 110, 150, 111, 150, 112, 150, 113, 150, 115, 150, 116, 150, 117, 150, 118, 150, 119, 150, 120, 150, 121, 150, 122, 150, 124, 150, 99, 216, 172, 222, 99, 216, 178, 222, 99, 216, 179, 222, 191, 150, 192, 150, 233, 150, 234, 150, 235, 150, 100, 216, 121, 220, 100, 216, 213, 221, 100, 216, 235, 221, 106, 151, 223, 122, 224, 122, 2, 152, 3, 152, 4, 152, 133, 152, 134, 152, 135, 152, 136, 152, 225, 152, 229, 152, 230, 152, 131, 153, 132, 153, 133, 153, 134, 153, 151, 153, 144, 154, 145, 154, 146, 154, 147, 154, 148, 154, 149, 154, 150, 154, 217, 154, 90, 155, 126, 156, 229, 156, 55, 158, 56, 158, 57, 158, 58, 158, 60, 0, 42, 0, 59, 158, 60, 158, 61, 158, 62, 158, 63, 158, 117, 158, 127, 158, 165, 158, 184, 158, 187, 158, 210, 158, 129, 159, 154, 159, 155, 159, 60, 0, 39, 0, 208, 253, 12, 40, 39, 0, 60, 0, 42, 0, 64, 216, 70, 220, 129, 78, 180, 78, 181, 78, 104, 80, 128, 80, 129, 80, 130, 80, 131, 80, 132, 80, 133, 80, 134, 80, 135, 80, 136, 80, 137, 80, 138, 80, 139, 80, 140, 80, 141, 80, 142, 80, 143, 80, 144, 80, 145, 80, 146, 80, 147, 80, 148, 80, 149, 80, 150, 80, 151, 80, 152, 80, 153, 80, 154, 80, 155, 80, 156, 80, 157, 80, 158, 80, 159, 80, 160, 80, 161, 80, 162, 80, 163, 80, 164, 80, 165, 80, 166, 80, 167, 80, 168, 80, 169, 80, 115, 52, 122, 52, 64, 216, 10, 223, 64, 216, 37, 223, 95, 81, 96, 81, 65, 216, 163, 220, 0, 103, 210, 81, 211, 81, 212, 81, 213, 81, 214, 81, 241, 81, 242, 81, 255, 81, 105, 82, 114, 82, 115, 82, 116, 82, 117, 82, 251, 52, 219, 82, 220, 82, 221, 82, 222, 82, 17, 83, 18, 83, 66, 216, 35, 221, 66, 216, 84, 221, 56, 53, 90, 83, 164, 83, 165, 83, 166, 83, 167, 83, 168, 83, 66, 216, 138, 222, 197, 83, 89, 85, 122, 85, 123, 85, 124, 85, 125, 85, 126, 85, 127, 85, 128, 85, 129, 85, 130, 85, 131, 85, 132, 85, 133, 85, 134, 85, 135, 85, 136, 85, 137, 85, 138, 85, 139, 85, 140, 85, 142, 85, 143, 85, 144, 85, 145, 85, 146, 85, 147, 85, 148, 85, 149, 85, 150, 85, 151, 85, 152, 85, 153, 85, 154, 85, 155, 85, 156, 85, 157, 85, 158, 85, 159, 85, 160, 85, 161, 85, 162, 85, 163, 85, 60, 0, 42, 0, 164, 85, 165, 85, 166, 85, 167, 85, 168, 85, 169, 85, 170, 85, 171, 85, 172, 85, 173, 85, 174, 85, 175, 85, 176, 85, 177, 85, 178, 85, 179, 85, 180, 85, 181, 85, 183, 85, 184, 85, 185, 85, 186, 85, 187, 85, 188, 85, 189, 85, 190, 85, 69, 86, 191, 53, 193, 53, 196, 53, 197, 53, 199, 53, 202, 53, 206, 53, 67, 216, 200, 221, 67, 216, 4, 222, 67, 216, 9, 222, 67, 216, 10, 222, 67, 216, 13, 222, 67, 216, 14, 222, 67, 216, 15, 222, 67, 216, 16, 222, 67, 216, 17, 222, 67, 216, 22, 222, 67, 216, 29, 222, 222, 85, 12, 87, 13, 87, 14, 87, 16, 87, 22, 88, 23, 88, 24, 88, 25, 88, 26, 88, 27, 88, 28, 88, 29, 88, 30, 88, 31, 88, 32, 88, 33, 88, 34, 88, 35, 88, 36, 88, 37, 88, 38, 88, 39, 88, 40, 88, 41, 88, 42, 88, 43, 88, 44, 88, 45, 88, 46, 88, 47, 88, 48, 88, 49, 88, 51, 88, 52, 88, 54, 88, 55, 88, 56, 88, 57, 88, 58, 88, 59, 88, 60, 88, 62, 88, 63, 88, 64, 88, 65, 88, 66, 88, 68, 88, 69, 88, 70, 88, 71, 88, 72, 88, 78, 54, 68, 216, 117, 223, 68, 216, 118, 223, 68, 216, 142, 223, 68, 216, 152, 223, 68, 216, 156, 223, 249, 88, 250, 88, 251, 88, 33, 89, 96, 89, 97, 89, 99, 89, 100, 89, 101, 89, 119, 90, 120, 90, 121, 90, 122, 90, 123, 90, 124, 90, 125, 90, 126, 90, 127, 90, 128, 90, 129, 90, 130, 90, 131, 90, 132, 90, 133, 90, 134, 90, 60, 0, 42, 0, 135, 90, 136, 90, 137, 90, 138, 90, 139, 90, 140, 90, 141, 90, 143, 90, 145, 90, 146, 90, 147, 90, 148, 90, 149, 90, 150, 90, 151, 90, 152, 90, 153, 90, 154, 90, 155, 90, 156, 90, 157, 90, 158, 90, 159, 90, 160, 90, 161, 90, 162, 90, 163, 90, 164, 90, 165, 90, 166, 90, 167, 90, 168, 90, 169, 90, 170, 90, 171, 90, 172, 90, 173, 90, 174, 90, 175, 90, 207, 90, 245, 54, 0, 55, 1, 55, 3, 55, 4, 55, 69, 216, 171, 223, 69, 216, 176, 223, 69, 216, 177, 223, 69, 216, 178, 223, 69, 216, 179, 223, 69, 216, 180, 223, 69, 216, 181, 223, 69, 216, 195, 223, 69, 216, 199, 223, 69, 216, 217, 223, 69, 216, 218, 223, 69, 216, 219, 223, 69, 216, 220, 223, 69, 216, 223, 223, 113, 91, 70, 216, 124, 221, 70, 216, 128, 221, 115, 91, 202, 91, 203, 91, 204, 91, 205, 91, 206, 91, 207, 91, 208, 91, 209, 91, 210, 91, 211, 91, 212, 91, 213, 91, 234, 91, 98, 55, 70, 216, 69, 222, 70, 216, 75, 222, 10, 92, 11, 92, 12, 92, 30, 92, 48, 92, 49, 92, 94, 92, 95, 92, 97, 92, 71, 216, 162, 220, 71, 216, 165, 220, 49, 93, 50, 93, 51, 93, 52, 93, 53, 93, 54, 93, 55, 93, 56, 93, 57, 93, 58, 93, 59, 93, 60, 93, 61, 93, 62, 93, 63, 93, 64, 93, 65, 93, 66, 93, 67, 93, 68, 93, 69, 93, 70, 93, 71, 93, 72, 93, 73, 93, 75, 93, 76, 93, 77, 93, 78, 93, 79, 93, 80, 93, 81, 93, 82, 93, 83, 93, 84, 93, 85, 93, 86, 93, 87, 93, 60, 0, 42, 0, 88, 93, 89, 93, 90, 93, 91, 93, 92, 93, 93, 93, 17, 250, 71, 216, 137, 222, 71, 216, 164, 222, 71, 216, 168, 222, 107, 93, 239, 93, 45, 56, 253, 93, 72, 216, 69, 220, 72, 216, 73, 220, 61, 94, 63, 94, 64, 94, 65, 94, 66, 94, 67, 94, 68, 94, 69, 94, 70, 94, 71, 94, 73, 94, 72, 216, 252, 220, 72, 94, 126, 94, 189, 94, 191, 94, 192, 94, 193, 94, 194, 94, 195, 94, 202, 94, 196, 94, 17, 95, 58, 95, 59, 95, 60, 95, 61, 95, 62, 95, 88, 95, 109, 95, 154, 95, 165, 95, 166, 95, 167, 95, 168, 95, 169, 95, 170, 95, 171, 95, 73, 216, 48, 221, 178, 96, 179, 96, 182, 96, 185, 96, 193, 96, 196, 96, 201, 96, 204, 96, 206, 96, 209, 96, 210, 96, 214, 96, 224, 96, 225, 96, 226, 96, 227, 96, 229, 96, 233, 96, 234, 96, 235, 96, 240, 96, 241, 96, 242, 96, 244, 96, 245, 96, 246, 96, 248, 96, 250, 96, 251, 96, 252, 96, 253, 96, 254, 96, 255, 96, 0, 97, 3, 97, 4, 97, 5, 97, 7, 97, 9, 97, 10, 97, 11, 97, 12, 97, 14, 97, 16, 97, 17, 97, 18, 97, 19, 97, 20, 97, 21, 97, 22, 97, 24, 97, 28, 97, 29, 97, 30, 97, 32, 97, 33, 97, 34, 97, 35, 97, 36, 97, 37, 97, 38, 97, 104, 97, 97, 57, 98, 57, 101, 57, 73, 216, 42, 223, 73, 216, 117, 223, 73, 216, 129, 223, 30, 98, 31, 98, 184, 57, 73, 98, 74, 98, 60, 0, 42, 0, 140, 99, 148, 99, 163, 99, 176, 99, 177, 99, 190, 99, 191, 99, 192, 99, 193, 99, 194, 99, 195, 99, 196, 99, 198, 99, 199, 99, 200, 99, 201, 99, 202, 99, 203, 99, 204, 99, 205, 99, 206, 99, 207, 99, 208, 99, 209, 99, 210, 99, 211, 99, 212, 99, 213, 99, 214, 99, 215, 99, 216, 99, 217, 99, 218, 99, 219, 99, 220, 99, 221, 99, 222, 99, 223, 99, 224, 99, 225, 99, 226, 99, 227, 99, 229, 99, 230, 99, 232, 99, 233, 99, 234, 99, 236, 99, 237, 99, 238, 99, 239, 99, 240, 99, 242, 99, 243, 99, 244, 99, 245, 99, 246, 99, 247, 99, 248, 99, 249, 99, 250, 99, 251, 99, 252, 99, 253, 99, 254, 99, 255, 99, 0, 100, 1, 100, 2, 100, 3, 100, 4, 100, 5, 100, 82, 100, 23, 58, 24, 58, 75, 216, 38, 220, 75, 216, 39, 220, 75, 216, 56, 220, 75, 216, 76, 220, 75, 216, 81, 220, 75, 216, 85, 220, 97, 100, 50, 101, 92, 101, 94, 101, 95, 101, 96, 101, 97, 101, 98, 101, 99, 101, 100, 101, 101, 101, 102, 101, 103, 101, 104, 101, 105, 101, 106, 101, 143, 58, 144, 58, 151, 58, 75, 216, 116, 223, 140, 101, 144, 101, 145, 101, 157, 101, 184, 58, 76, 216, 125, 220, 76, 216, 126, 220, 158, 101, 174, 101, 175, 101, 192, 58, 177, 101, 208, 101, 209, 101, 76, 216, 3, 221, 106, 102, 107, 102, 108, 102, 109, 102, 110, 102, 111, 102, 112, 102, 113, 102, 114, 102, 115, 102, 116, 102, 117, 102, 60, 0, 42, 0, 118, 102, 119, 102, 121, 102, 122, 102, 123, 102, 124, 102, 125, 102, 126, 102, 127, 102, 128, 102, 129, 102, 130, 102, 131, 102, 145, 102, 245, 58, 251, 58, 0, 59, 76, 216, 234, 221, 76, 216, 247, 221, 76, 216, 248, 221, 76, 216, 249, 221, 76, 216, 15, 222, 254, 102, 255, 102, 1, 103, 2, 103, 28, 103, 29, 103, 30, 103, 31, 103, 56, 59, 180, 104, 195, 104, 196, 104, 197, 104, 198, 104, 199, 104, 200, 104, 201, 104, 202, 104, 203, 104, 204, 104, 205, 104, 206, 104, 207, 104, 208, 104, 209, 104, 210, 104, 211, 104, 212, 104, 213, 104, 214, 104, 215, 104, 216, 104, 217, 104, 218, 104, 219, 104, 220, 104, 221, 104, 222, 104, 223, 104, 224, 104, 225, 104, 226, 104, 227, 104, 228, 104, 229, 104, 230, 104, 231, 104, 232, 104, 233, 104, 234, 104, 235, 104, 236, 104, 237, 104, 238, 104, 239, 104, 240, 104, 241, 104, 242, 104, 243, 104, 244, 104, 245, 104, 246, 104, 247, 104, 248, 104, 249, 104, 250, 104, 251, 104, 252, 104, 253, 104, 254, 104, 255, 104, 0, 105, 1, 105, 2, 105, 3, 105, 4, 105, 5, 105, 6, 105, 7, 105, 8, 105, 9, 105, 10, 105, 11, 105, 12, 105, 13, 105, 14, 105, 15, 105, 16, 105, 17, 105, 18, 105, 19, 105, 20, 105, 21, 105, 22, 105, 23, 105, 24, 105, 25, 105, 26, 105, 27, 105, 28, 105, 29, 105, 30, 105, 31, 105, 32, 105, 33, 105, 34, 105, 60, 0, 42, 0, 35, 105, 36, 105, 37, 105, 38, 105, 39, 105, 40, 105, 41, 105, 42, 105, 43, 105, 44, 105, 45, 105, 46, 105, 118, 59, 120, 59, 121, 59, 122, 59, 123, 59, 124, 59, 125, 59, 126, 59, 127, 59, 128, 59, 77, 216, 229, 220, 77, 216, 25, 221, 77, 216, 48, 221, 110, 105, 112, 105, 19, 250, 57, 107, 58, 107, 59, 107, 60, 107, 61, 107, 62, 107, 63, 107, 110, 107, 111, 107, 148, 107, 149, 107, 150, 107, 151, 107, 152, 107, 153, 107, 154, 107, 100, 60, 102, 60, 104, 60, 105, 60, 188, 107, 189, 107, 190, 107, 239, 107, 240, 107, 241, 107, 242, 107, 243, 107, 244, 107, 245, 107, 246, 107, 44, 108, 45, 108, 46, 108, 47, 108, 48, 108, 252, 109, 254, 109, 3, 110, 24, 110, 25, 110, 27, 110, 28, 110, 29, 110, 30, 110, 31, 110, 32, 110, 33, 110, 34, 110, 35, 110, 36, 110, 37, 110, 38, 110, 39, 110, 40, 110, 41, 110, 42, 110, 43, 110, 44, 110, 45, 110, 46, 110, 47, 110, 48, 110, 49, 110, 50, 110, 51, 110, 52, 110, 53, 110, 54, 110, 55, 110, 56, 110, 57, 110, 58, 110, 59, 110, 60, 110, 61, 110, 62, 110, 63, 110, 64, 110, 65, 110, 66, 110, 67, 110, 68, 110, 69, 110, 70, 110, 71, 110, 72, 110, 73, 110, 74, 110, 75, 110, 76, 110, 77, 110, 78, 110, 79, 110, 80, 110, 81, 110, 82, 110, 83, 110, 84, 110, 85, 110, 60, 0, 42, 0, 86, 110, 87, 110, 88, 110, 89, 110, 90, 110, 91, 110, 92, 110, 93, 110, 94, 110, 95, 110, 96, 110, 97, 110, 98, 110, 99, 110, 100, 110, 101, 110, 102, 110, 103, 110, 104, 110, 105, 110, 106, 110, 107, 110, 109, 110, 110, 110, 111, 110, 112, 110, 113, 110, 114, 110, 115, 110, 117, 110, 118, 110, 119, 110, 120, 110, 121, 110, 122, 110, 123, 110, 124, 110, 125, 110, 126, 110, 127, 110, 128, 110, 129, 110, 130, 110, 131, 110, 132, 110, 133, 110, 134, 110, 135, 110, 136, 110, 137, 110, 138, 110, 139, 110, 140, 110, 203, 110, 222, 110, 18, 61, 19, 61, 29, 61, 32, 61, 34, 61, 79, 216, 6, 222, 79, 216, 17, 222, 79, 216, 44, 222, 79, 216, 45, 222, 79, 216, 46, 222, 79, 216, 47, 222, 79, 216, 48, 222, 79, 216, 49, 222, 79, 216, 57, 222, 251, 112, 25, 113, 26, 113, 27, 113, 28, 113, 29, 113, 30, 113, 31, 113, 32, 113, 33, 113, 34, 113, 35, 113, 36, 113, 37, 113, 38, 113, 39, 113, 40, 113, 41, 113, 42, 113, 43, 113, 44, 113, 45, 113, 46, 113, 47, 113, 48, 113, 49, 113, 50, 113, 51, 113, 52, 113, 53, 113, 54, 113, 55, 113, 56, 113, 57, 113, 58, 113, 59, 113, 60, 113, 61, 113, 62, 113, 63, 113, 64, 113, 110, 113, 198, 61, 199, 61, 201, 61, 204, 61, 205, 61, 80, 216, 118, 222, 80, 216, 119, 222, 80, 216, 120, 222, 80, 216, 132, 222, 80, 216, 147, 222, 80, 216, 149, 222, 80, 216, 165, 222, 50, 114, 75, 114, 76, 114, 77, 114, 81, 216, 200, 221, 60, 0, 42, 0, 90, 114, 128, 114, 130, 114, 131, 114, 132, 114, 133, 114, 134, 114, 135, 114, 136, 114, 137, 114, 138, 114, 139, 114, 67, 62, 81, 216, 116, 222, 6, 115, 11, 115, 12, 115, 18, 115, 43, 115, 34, 115, 35, 115, 36, 115, 37, 115, 38, 115, 39, 115, 40, 115, 41, 115, 44, 115, 45, 115, 47, 115, 48, 115, 49, 115, 50, 115, 51, 115, 52, 115, 53, 115, 54, 115, 56, 115, 57, 115, 81, 216, 224, 223, 247, 115, 20, 116, 21, 116, 22, 116, 23, 116, 24, 116, 25, 116, 26, 116, 27, 116, 28, 116, 29, 116, 31, 116, 32, 116, 33, 116, 34, 116, 35, 116, 36, 116, 37, 116, 38, 116, 40, 116, 41, 116, 42, 116, 43, 116, 44, 116, 45, 116, 46, 116, 47, 116, 48, 116, 49, 116, 50, 116, 51, 116, 52, 116, 53, 116, 54, 116, 55, 116, 56, 116, 57, 116, 58, 116, 59, 116, 60, 116, 209, 62, 214, 62, 218, 62, 219, 62, 82, 216, 123, 221, 82, 216, 127, 221, 82, 216, 130, 221, 82, 216, 136, 221, 82, 216, 137, 221, 82, 216, 138, 221, 82, 216, 139, 221, 82, 216, 140, 221, 82, 216, 141, 221, 82, 216, 142, 221, 82, 216, 143, 221, 82, 216, 148, 221, 82, 216, 164, 221, 82, 216, 167, 221, 82, 216, 169, 221, 82, 216, 171, 221, 82, 216, 172, 221, 82, 216, 173, 221, 249, 116, 250, 116, 251, 116, 252, 116, 36, 117, 37, 117, 38, 117, 47, 117, 83, 216, 9, 220, 106, 117, 107, 117, 108, 117, 109, 117, 110, 117, 111, 117, 114, 117, 115, 117, 116, 117, 113, 117, 142, 117, 143, 117, 215, 117, 216, 117, 217, 117, 218, 117, 219, 117, 60, 0, 42, 0, 220, 117, 221, 117, 222, 117, 223, 117, 224, 117, 225, 117, 226, 117, 227, 117, 228, 117, 229, 117, 230, 117, 231, 117, 232, 117, 233, 117, 234, 117, 235, 117, 147, 63, 152, 63, 83, 216, 184, 221, 123, 118, 124, 118, 146, 118, 147, 118, 148, 118, 149, 118, 150, 118, 179, 118, 180, 118, 217, 118, 218, 118, 220, 118, 3, 64, 66, 119, 67, 119, 68, 119, 69, 119, 70, 119, 71, 119, 72, 119, 73, 119, 74, 119, 75, 119, 76, 119, 77, 119, 78, 119, 79, 119, 80, 119, 81, 119, 57, 64, 64, 64, 222, 119, 223, 119, 236, 119, 237, 119, 92, 120, 93, 120, 94, 120, 95, 120, 96, 120, 97, 120, 98, 120, 99, 120, 100, 120, 101, 120, 102, 120, 103, 120, 104, 120, 105, 120, 106, 120, 107, 120, 108, 120, 109, 120, 110, 120, 111, 120, 112, 120, 113, 120, 114, 120, 115, 120, 116, 120, 117, 120, 118, 120, 119, 120, 180, 64, 102, 121, 112, 121, 113, 121, 114, 121, 115, 121, 116, 121, 117, 121, 118, 121, 119, 121, 120, 121, 132, 121, 9, 65, 85, 216, 149, 222, 188, 121, 255, 121, 0, 122, 1, 122, 2, 122, 3, 122, 4, 122, 5, 122, 7, 122, 8, 122, 9, 122, 10, 122, 11, 122, 12, 122, 13, 122, 14, 122, 85, 216, 199, 223, 150, 122, 151, 122, 152, 122, 153, 122, 155, 122, 156, 122, 157, 122, 226, 122, 227, 122, 228, 122, 229, 122, 230, 122, 231, 122, 63, 123, 64, 123, 60, 0, 42, 0, 65, 123, 66, 123, 67, 123, 68, 123, 69, 123, 70, 123, 71, 123, 72, 123, 73, 123, 74, 123, 75, 123, 76, 123, 77, 123, 78, 123, 79, 123, 80, 123, 81, 123, 82, 123, 83, 123, 84, 123, 85, 123, 86, 123, 87, 123, 88, 123, 89, 123, 90, 123, 91, 123, 92, 123, 93, 123, 237, 65, 238, 65, 239, 65, 108, 123, 158, 124, 159, 124, 160, 124, 161, 124, 162, 124, 164, 124, 165, 124, 166, 124, 167, 124, 168, 124, 169, 124, 170, 124, 171, 124, 172, 124, 173, 124, 131, 66, 42, 125, 43, 125, 77, 125, 78, 125, 79, 125, 80, 125, 81, 125, 82, 125, 83, 125, 84, 125, 85, 125, 86, 125, 87, 125, 88, 125, 89, 125, 90, 125, 92, 125, 93, 125, 94, 125, 95, 125, 96, 125, 97, 125, 98, 125, 99, 125, 100, 125, 101, 125, 102, 125, 103, 125, 104, 125, 105, 125, 106, 125, 107, 125, 108, 125, 109, 125, 110, 125, 111, 125, 112, 125, 113, 125, 114, 125, 115, 125, 116, 125, 117, 125, 118, 125, 119, 125, 126, 125, 59, 67, 2, 127, 3, 127, 4, 127, 5, 127, 6, 127, 7, 127, 8, 127, 9, 127, 10, 127, 11, 127, 12, 127, 13, 127, 14, 127, 15, 127, 16, 127, 17, 127, 18, 127, 19, 127, 20, 127, 21, 127, 22, 127, 23, 127, 24, 127, 62, 127, 63, 127, 64, 127, 100, 127, 101, 127, 102, 127, 160, 127, 161, 127, 162, 127, 110, 67, 60, 0, 42, 0, 111, 67, 211, 127, 212, 127, 213, 127, 214, 127, 215, 127, 216, 127, 217, 127, 218, 127, 11, 128, 32, 128, 78, 128, 79, 128, 80, 128, 81, 128, 82, 128, 83, 128, 84, 128, 96, 128, 89, 216, 114, 221, 212, 128, 254, 128, 20, 129, 32, 129, 57, 129, 58, 129, 59, 129, 60, 129, 61, 129, 62, 129, 63, 129, 64, 129, 65, 129, 66, 129, 67, 129, 68, 129, 69, 129, 70, 129, 71, 129, 72, 129, 73, 129, 74, 129, 75, 129, 76, 129, 77, 129, 78, 129, 79, 129, 81, 129, 82, 129, 83, 129, 84, 129, 85, 129, 86, 129, 87, 129, 88, 129, 89, 129, 90, 129, 57, 59, 0, 68, 1, 68, 2, 68, 89, 216, 22, 223, 89, 216, 65, 223, 116, 129, 230, 129, 238, 129, 239, 129, 240, 129, 90, 216, 17, 221, 245, 129, 246, 129, 247, 129, 248, 129, 249, 129, 3, 130, 4, 130, 18, 130, 28, 130, 60, 130, 61, 130, 62, 130, 63, 130, 90, 216, 52, 222, 117, 130, 18, 131, 59, 131, 70, 131, 176, 68, 183, 68, 189, 68, 192, 68, 154, 131, 190, 131, 191, 131, 192, 131, 193, 131, 194, 131, 195, 131, 196, 131, 197, 131, 198, 131, 199, 131, 200, 131, 201, 131, 202, 131, 203, 131, 204, 131, 205, 131, 206, 131, 207, 131, 208, 131, 209, 131, 210, 131, 211, 131, 212, 131, 213, 131, 214, 131, 215, 131, 216, 131, 218, 131, 219, 131, 220, 131, 221, 131, 222, 131, 223, 131, 224, 131, 225, 131, 226, 131, 60, 0, 42, 0, 227, 131, 228, 131, 229, 131, 230, 131, 231, 131, 232, 131, 233, 131, 234, 131, 235, 131, 236, 131, 237, 131, 238, 131, 239, 131, 240, 131, 241, 131, 242, 131, 243, 131, 244, 131, 245, 131, 246, 131, 247, 131, 248, 131, 249, 131, 250, 131, 251, 131, 252, 131, 253, 131, 254, 131, 255, 131, 0, 132, 1, 132, 2, 132, 3, 132, 4, 132, 5, 132, 6, 132, 7, 132, 8, 132, 9, 132, 10, 132, 11, 132, 12, 132, 13, 132, 14, 132, 15, 132, 16, 132, 17, 132, 18, 132, 19, 132, 20, 132, 21, 132, 22, 132, 23, 132, 24, 132, 25, 132, 26, 132, 27, 132, 28, 132, 29, 132, 30, 132, 31, 132, 32, 132, 33, 132, 34, 132, 35, 132, 36, 132, 38, 132, 39, 132, 87, 132, 91, 216, 64, 220, 91, 216, 65, 220, 91, 216, 70, 220, 91, 216, 126, 220, 91, 216, 127, 220, 91, 216, 128, 220, 91, 216, 129, 220, 91, 216, 130, 220, 91, 216, 164, 220, 91, 216, 183, 220, 91, 216, 184, 220, 91, 216, 189, 220, 91, 216, 192, 220, 56, 132, 91, 134, 93, 134, 136, 134, 208, 134, 209, 134, 210, 134, 211, 134, 212, 134, 213, 134, 215, 134, 216, 134, 217, 134, 218, 134, 219, 134, 220, 134, 221, 134, 222, 134, 223, 134, 224, 134, 225, 134, 226, 134, 227, 134, 228, 134, 229, 134, 230, 134, 231, 134, 232, 134, 233, 134, 234, 134, 235, 134, 236, 134, 237, 134, 238, 134, 239, 134, 240, 134, 241, 134, 242, 134, 243, 134, 244, 134, 173, 69, 175, 69, 179, 69, 92, 216, 178, 222, 92, 216, 182, 222, 70, 136, 60, 0, 42, 0, 71, 136, 72, 136, 73, 136, 85, 136, 86, 136, 87, 136, 177, 136, 178, 136, 179, 136, 180, 136, 181, 136, 182, 136, 183, 136, 184, 136, 185, 136, 186, 136, 187, 136, 188, 136, 189, 136, 190, 136, 191, 136, 192, 136, 193, 136, 194, 136, 195, 136, 196, 136, 197, 136, 198, 136, 199, 136, 200, 136, 201, 136, 215, 136, 93, 216, 148, 222, 1, 137, 131, 137, 132, 137, 153, 137, 149, 137, 151, 137, 152, 137, 154, 137, 204, 137, 205, 137, 218, 137, 219, 137, 221, 137, 222, 137, 52, 138, 53, 138, 54, 138, 55, 138, 56, 138, 57, 138, 58, 138, 59, 138, 60, 138, 61, 138, 64, 138, 65, 138, 66, 138, 67, 138, 68, 138, 69, 138, 70, 138, 71, 138, 72, 138, 73, 138, 74, 138, 75, 138, 76, 138, 77, 138, 78, 138, 79, 138, 80, 138, 81, 138, 82, 138, 83, 138, 84, 138, 85, 138, 86, 138, 87, 138, 88, 138, 89, 138, 90, 138, 91, 138, 92, 138, 93, 138, 94, 138, 95, 138, 96, 138, 207, 70, 208, 70, 94, 216, 160, 221, 31, 140, 32, 140, 33, 140, 34, 140, 35, 140, 36, 140, 37, 140, 38, 140, 39, 140, 94, 140, 95, 140, 96, 140, 97, 140, 72, 71, 126, 140, 127, 140, 128, 140, 129, 140, 130, 140, 131, 140, 95, 216, 177, 220, 175, 140, 176, 140, 177, 140, 179, 140, 180, 140, 181, 140, 182, 140, 183, 140, 184, 140, 185, 140, 186, 140, 187, 140, 188, 140, 189, 140, 60, 0, 42, 0, 190, 140, 191, 140, 192, 140, 193, 140, 95, 216, 83, 221, 95, 216, 84, 221, 75, 141, 76, 141, 77, 141, 78, 141, 79, 141, 80, 141, 81, 141, 82, 141, 83, 141, 84, 141, 85, 141, 128, 141, 129, 141, 130, 141, 131, 141, 132, 141, 133, 141, 134, 141, 135, 141, 136, 141, 137, 141, 138, 141, 139, 141, 161, 71, 163, 71, 197, 141, 198, 141, 199, 141, 200, 141, 201, 141, 202, 141, 203, 141, 204, 141, 205, 141, 206, 141, 207, 141, 209, 141, 210, 141, 211, 141, 212, 141, 213, 141, 214, 141, 215, 141, 216, 141, 217, 141, 218, 141, 219, 141, 220, 141, 221, 141, 222, 141, 245, 141, 237, 71, 95, 216, 249, 223, 96, 216, 2, 220, 96, 216, 9, 220, 176, 142, 228, 142, 229, 142, 230, 142, 231, 142, 232, 142, 233, 142, 234, 142, 235, 142, 236, 142, 238, 142, 239, 142, 240, 142, 241, 142, 242, 142, 243, 142, 244, 142, 245, 142, 246, 142, 247, 142, 248, 142, 249, 142, 250, 142, 251, 142, 252, 142, 253, 142, 171, 159, 82, 72, 96, 216, 226, 222, 135, 143, 136, 143, 137, 143, 138, 143, 139, 143, 140, 143, 141, 143, 142, 143, 156, 143, 157, 143, 97, 216, 18, 220, 40, 144, 41, 144, 42, 144, 43, 144, 44, 144, 45, 144, 46, 144, 47, 144, 48, 144, 49, 144, 50, 144, 51, 144, 52, 144, 53, 144, 54, 144, 55, 144, 56, 144, 57, 144, 58, 144, 59, 144, 97, 216, 60, 221, 97, 216, 61, 221, 245, 144, 249, 144, 251, 144, 252, 144, 254, 144, 60, 0, 42, 0, 255, 144, 0, 145, 1, 145, 2, 145, 3, 145, 4, 145, 5, 145, 6, 145, 7, 145, 8, 145, 9, 145, 10, 145, 208, 72, 44, 145, 95, 145, 96, 145, 97, 145, 98, 145, 99, 145, 100, 145, 101, 145, 243, 72, 201, 145, 202, 145, 207, 145, 253, 145, 254, 145, 255, 145, 0, 146, 1, 146, 2, 146, 3, 146, 4, 146, 5, 146, 6, 146, 7, 146, 8, 146, 9, 146, 10, 146, 11, 146, 12, 146, 13, 146, 14, 146, 15, 146, 16, 146, 17, 146, 18, 146, 19, 146, 20, 146, 21, 146, 22, 146, 23, 146, 24, 146, 25, 146, 26, 146, 27, 146, 28, 146, 29, 146, 30, 146, 31, 146, 32, 146, 33, 146, 34, 146, 35, 146, 36, 146, 37, 146, 38, 146, 39, 146, 40, 146, 41, 146, 42, 146, 43, 146, 44, 146, 30, 73, 32, 73, 98, 216, 86, 221, 98, 216, 100, 221, 248, 148, 249, 148, 250, 148, 251, 148, 252, 148, 253, 148, 254, 148, 255, 148, 0, 149, 1, 149, 2, 149, 3, 149, 4, 149, 5, 149, 6, 149, 7, 149, 8, 149, 9, 149, 10, 149, 11, 149, 12, 149, 13, 149, 14, 149, 15, 149, 16, 149, 17, 149, 18, 149, 19, 149, 20, 149, 21, 149, 123, 149, 139, 149, 140, 149, 141, 149, 142, 149, 143, 149, 144, 149, 145, 149, 146, 149, 147, 149, 148, 149, 149, 149, 150, 149, 151, 149, 17, 150, 18, 150, 19, 150, 20, 150, 21, 150, 114, 150, 123, 150, 60, 0, 42, 0, 125, 150, 126, 150, 127, 150, 128, 150, 129, 150, 130, 150, 131, 150, 132, 150, 133, 150, 134, 150, 135, 150, 136, 150, 137, 150, 138, 150, 139, 150, 140, 150, 141, 150, 142, 150, 144, 150, 99, 216, 217, 222, 99, 216, 231, 222, 193, 150, 194, 150, 195, 150, 196, 150, 197, 150, 198, 150, 199, 150, 200, 150, 236, 150, 237, 150, 238, 150, 239, 150, 240, 150, 241, 150, 242, 150, 243, 150, 100, 216, 136, 220, 100, 216, 139, 220, 83, 151, 84, 151, 95, 151, 107, 151, 108, 151, 109, 151, 110, 151, 111, 151, 112, 151, 113, 151, 204, 151, 233, 151, 5, 152, 6, 152, 7, 152, 8, 152, 137, 152, 138, 152, 139, 152, 140, 152, 141, 152, 142, 152, 143, 152, 169, 152, 170, 152, 101, 216, 158, 221, 211, 152, 231, 152, 232, 152, 233, 152, 234, 152, 235, 152, 237, 152, 239, 152, 240, 152, 242, 152, 135, 153, 136, 153, 138, 153, 139, 153, 173, 153, 174, 153, 151, 154, 152, 154, 153, 154, 154, 154, 155, 154, 169, 154, 224, 154, 50, 76, 127, 156, 128, 156, 129, 156, 130, 156, 131, 156, 230, 156, 64, 158, 65, 158, 66, 158, 67, 158, 68, 158, 69, 158, 70, 158, 71, 158, 72, 158, 195, 158, 196, 158, 205, 158, 209, 158, 249, 158, 11, 159, 130, 159, 60, 0, 39, 0, 208, 253, 13, 40, 39, 0, 60, 0, 42, 0, 130, 78, 131, 78, 132, 78, 182, 78, 183, 78, 170, 80, 171, 80, 172, 80, 173, 80, 174, 80, 175, 80, 176, 80, 177, 80, 178, 80, 179, 80, 180, 80, 181, 80, 182, 80, 183, 80, 184, 80, 185, 80, 186, 80, 187, 80, 188, 80, 189, 80, 190, 80, 191, 80, 192, 80, 193, 80, 194, 80, 195, 80, 196, 80, 197, 80, 198, 80, 199, 80, 200, 80, 201, 80, 202, 80, 203, 80, 204, 80, 205, 80, 125, 52, 126, 52, 64, 216, 65, 223, 64, 216, 69, 223, 64, 216, 70, 223, 64, 216, 71, 223, 64, 216, 181, 223, 97, 81, 126, 81, 127, 81, 215, 81, 119, 82, 120, 82, 121, 82, 122, 82, 123, 82, 124, 82, 125, 82, 126, 82, 127, 82, 223, 82, 224, 82, 225, 82, 226, 82, 227, 82, 228, 82, 229, 82, 230, 82, 231, 82, 47, 83, 50, 53, 51, 53, 52, 53, 128, 83, 129, 83, 170, 83, 171, 83, 175, 83, 224, 83, 141, 85, 191, 85, 192, 85, 193, 85, 194, 85, 195, 85, 196, 85, 197, 85, 198, 85, 199, 85, 200, 85, 201, 85, 202, 85, 203, 85, 204, 85, 205, 85, 206, 85, 207, 85, 208, 85, 209, 85, 210, 85, 211, 85, 212, 85, 213, 85, 214, 85, 215, 85, 216, 85, 217, 85, 218, 85, 219, 85, 220, 85, 221, 85, 223, 85, 224, 85, 225, 85, 226, 85, 227, 85, 228, 85, 229, 85, 230, 85, 231, 85, 232, 85, 233, 85, 234, 85, 235, 85, 236, 85, 237, 85, 238, 85, 60, 0, 42, 0, 239, 85, 240, 85, 241, 85, 242, 85, 243, 85, 244, 85, 245, 85, 210, 53, 214, 53, 219, 53, 221, 53, 224, 53, 67, 216, 76, 222, 67, 216, 109, 222, 67, 216, 115, 222, 67, 216, 117, 222, 67, 216, 118, 222, 67, 216, 119, 222, 67, 216, 120, 222, 67, 216, 121, 222, 67, 216, 122, 222, 67, 216, 123, 222, 67, 216, 140, 222, 67, 216, 150, 222, 67, 216, 152, 222, 67, 216, 157, 222, 67, 216, 162, 222, 67, 216, 170, 222, 67, 216, 171, 222, 67, 216, 172, 222, 67, 216, 182, 222, 17, 87, 18, 87, 19, 87, 20, 87, 21, 87, 61, 88, 67, 88, 73, 88, 74, 88, 75, 88, 76, 88, 77, 88, 78, 88, 79, 88, 80, 88, 81, 88, 82, 88, 83, 88, 84, 88, 85, 88, 86, 88, 87, 88, 88, 88, 89, 88, 90, 88, 91, 88, 92, 88, 93, 88, 94, 88, 95, 88, 96, 88, 97, 88, 98, 88, 99, 88, 100, 88, 101, 88, 102, 88, 103, 88, 104, 88, 105, 88, 106, 88, 107, 88, 108, 88, 109, 88, 110, 88, 111, 88, 112, 88, 113, 88, 95, 54, 68, 216, 197, 223, 68, 216, 198, 223, 252, 88, 102, 89, 103, 89, 104, 89, 69, 216, 97, 222, 144, 90, 176, 90, 177, 90, 178, 90, 179, 90, 180, 90, 181, 90, 182, 90, 183, 90, 184, 90, 185, 90, 186, 90, 187, 90, 188, 90, 189, 90, 190, 90, 191, 90, 192, 90, 193, 90, 194, 90, 195, 90, 196, 90, 197, 90, 198, 90, 199, 90, 200, 90, 201, 90, 202, 90, 203, 90, 204, 90, 205, 90, 206, 90, 208, 90, 209, 90, 210, 90, 211, 90, 212, 90, 8, 55, 10, 55, 13, 55, 69, 216, 239, 223, 60, 0, 42, 0, 69, 216, 245, 223, 69, 216, 246, 223, 69, 216, 248, 223, 69, 216, 249, 223, 69, 216, 250, 223, 69, 216, 251, 223, 69, 216, 252, 223, 116, 91, 70, 216, 131, 221, 70, 216, 136, 221, 118, 91, 214, 91, 215, 91, 216, 91, 217, 91, 218, 91, 219, 91, 220, 91, 221, 91, 102, 55, 70, 216, 99, 222, 31, 92, 32, 92, 50, 92, 51, 92, 52, 92, 71, 216, 172, 220, 74, 93, 94, 93, 95, 93, 96, 93, 97, 93, 98, 93, 99, 93, 100, 93, 101, 93, 102, 93, 103, 93, 104, 93, 105, 93, 106, 93, 108, 93, 109, 93, 110, 93, 111, 93, 112, 93, 113, 93, 114, 93, 115, 93, 116, 93, 117, 93, 118, 93, 242, 55, 248, 55, 71, 216, 200, 222, 71, 216, 213, 222, 240, 93, 74, 94, 75, 94, 76, 94, 77, 94, 78, 94, 79, 94, 121, 94, 197, 94, 198, 94, 199, 94, 200, 94, 201, 94, 203, 94, 204, 94, 72, 216, 124, 222, 18, 95, 63, 95, 64, 95, 65, 95, 66, 95, 89, 95, 90, 95, 73, 216, 101, 220, 110, 95, 172, 95, 173, 95, 174, 95, 175, 95, 176, 95, 243, 96, 247, 96, 249, 96, 1, 97, 2, 97, 6, 97, 8, 97, 13, 97, 15, 97, 23, 97, 25, 97, 26, 97, 27, 97, 31, 97, 99, 57, 100, 57, 102, 57, 39, 97, 41, 97, 42, 97, 43, 97, 45, 97, 46, 97, 47, 97, 48, 97, 49, 97, 50, 97, 52, 97, 53, 97, 54, 97, 55, 97, 57, 97, 58, 97, 60, 97, 61, 97, 62, 97, 64, 97, 67, 97, 68, 97, 69, 97, 70, 97, 73, 97, 60, 0, 42, 0, 74, 97, 76, 97, 77, 97, 78, 97, 79, 97, 81, 97, 73, 216, 150, 223, 73, 216, 180, 223, 73, 216, 181, 223, 73, 216, 205, 223, 38, 98, 32, 98, 33, 98, 34, 98, 35, 98, 36, 98, 37, 98, 231, 99, 235, 99, 241, 99, 6, 100, 7, 100, 8, 100, 9, 100, 10, 100, 11, 100, 12, 100, 13, 100, 14, 100, 15, 100, 16, 100, 17, 100, 18, 100, 19, 100, 20, 100, 21, 100, 22, 100, 23, 100, 24, 100, 25, 100, 26, 100, 27, 100, 28, 100, 29, 100, 30, 100, 31, 100, 32, 100, 33, 100, 34, 100, 35, 100, 36, 100, 37, 100, 38, 100, 39, 100, 40, 100, 41, 100, 42, 100, 44, 100, 45, 100, 46, 100, 47, 100, 48, 100, 50, 100, 51, 100, 53, 100, 54, 100, 55, 100, 56, 100, 57, 100, 58, 100, 60, 100, 61, 100, 62, 100, 64, 100, 65, 100, 66, 100, 67, 100, 68, 100, 69, 100, 70, 100, 71, 100, 72, 100, 73, 100, 74, 100, 32, 58, 35, 58, 38, 58, 41, 58, 42, 58, 75, 216, 98, 220, 75, 216, 136, 220, 75, 216, 155, 220, 75, 216, 161, 220, 75, 216, 169, 220, 75, 216, 178, 220, 197, 99, 49, 100, 109, 101, 107, 101, 108, 101, 110, 101, 111, 101, 112, 101, 146, 101, 101, 216, 176, 221, 159, 101, 176, 101, 210, 101, 211, 101, 212, 101, 213, 101, 228, 101, 120, 102, 132, 102, 133, 102, 134, 102, 135, 102, 136, 102, 137, 102, 138, 102, 139, 102, 140, 102, 141, 102, 142, 102, 143, 102, 144, 102, 146, 102, 147, 102, 60, 0, 42, 0, 148, 102, 149, 102, 150, 102, 151, 102, 152, 102, 153, 102, 1, 59, 2, 59, 3, 59, 4, 59, 5, 59, 6, 59, 7, 59, 8, 59, 9, 59, 10, 59, 11, 59, 12, 59, 76, 216, 37, 222, 76, 216, 47, 222, 76, 216, 49, 222, 76, 216, 50, 222, 76, 216, 51, 222, 76, 216, 52, 222, 3, 103, 32, 103, 47, 105, 48, 105, 49, 105, 50, 105, 51, 105, 52, 105, 53, 105, 54, 105, 55, 105, 56, 105, 57, 105, 58, 105, 59, 105, 60, 105, 61, 105, 62, 105, 63, 105, 64, 105, 65, 105, 66, 105, 67, 105, 68, 105, 69, 105, 70, 105, 71, 105, 72, 105, 73, 105, 74, 105, 75, 105, 76, 105, 77, 105, 78, 105, 79, 105, 80, 105, 81, 105, 82, 105, 83, 105, 84, 105, 85, 105, 87, 105, 88, 105, 89, 105, 90, 105, 91, 105, 92, 105, 93, 105, 94, 105, 95, 105, 96, 105, 97, 105, 98, 105, 99, 105, 100, 105, 101, 105, 102, 105, 103, 105, 104, 105, 105, 105, 106, 105, 107, 105, 108, 105, 109, 105, 111, 105, 113, 105, 114, 105, 115, 105, 116, 105, 117, 105, 118, 105, 119, 105, 120, 105, 121, 105, 122, 105, 123, 105, 124, 105, 125, 105, 126, 105, 127, 105, 128, 105, 129, 105, 130, 105, 131, 105, 132, 105, 133, 105, 134, 105, 135, 105, 136, 105, 137, 105, 139, 105, 140, 105, 148, 105, 152, 105, 149, 59, 150, 59, 153, 59, 77, 216, 81, 221, 77, 216, 90, 221, 77, 216, 103, 221, 77, 216, 149, 221, 77, 216, 153, 221, 77, 216, 156, 221, 64, 107, 60, 0, 42, 0, 65, 107, 66, 107, 67, 107, 68, 107, 69, 107, 70, 107, 71, 107, 72, 107, 60, 60, 62, 60, 113, 107, 114, 107, 115, 107, 155, 107, 156, 107, 110, 60, 191, 107, 192, 107, 193, 107, 194, 107, 78, 216, 167, 222, 211, 107, 247, 107, 248, 107, 249, 107, 250, 107, 251, 107, 252, 107, 253, 107, 49, 108, 108, 110, 141, 110, 142, 110, 143, 110, 144, 110, 145, 110, 146, 110, 147, 110, 148, 110, 149, 110, 150, 110, 151, 110, 152, 110, 153, 110, 154, 110, 155, 110, 156, 110, 157, 110, 158, 110, 159, 110, 160, 110, 161, 110, 162, 110, 163, 110, 164, 110, 165, 110, 166, 110, 167, 110, 168, 110, 169, 110, 170, 110, 171, 110, 172, 110, 173, 110, 174, 110, 175, 110, 176, 110, 177, 110, 178, 110, 179, 110, 180, 110, 181, 110, 182, 110, 183, 110, 184, 110, 185, 110, 186, 110, 187, 110, 188, 110, 189, 110, 190, 110, 191, 110, 192, 110, 193, 110, 194, 110, 195, 110, 196, 110, 197, 110, 198, 110, 199, 110, 200, 110, 201, 110, 202, 110, 205, 110, 207, 110, 208, 110, 209, 110, 210, 110, 211, 110, 212, 110, 214, 110, 215, 110, 216, 110, 217, 110, 219, 110, 220, 110, 221, 110, 223, 110, 224, 110, 225, 110, 226, 110, 227, 110, 228, 110, 229, 110, 230, 110, 231, 110, 232, 110, 233, 110, 234, 110, 19, 111, 38, 61, 50, 61, 51, 61, 59, 61, 79, 216, 136, 222, 79, 216, 137, 222, 79, 216, 138, 222, 79, 216, 139, 222, 60, 0, 42, 0, 79, 216, 185, 222, 79, 216, 191, 222, 218, 110, 65, 113, 66, 113, 67, 113, 68, 113, 69, 113, 70, 113, 71, 113, 72, 113, 73, 113, 74, 113, 75, 113, 76, 113, 77, 113, 78, 113, 79, 113, 80, 113, 81, 113, 82, 113, 83, 113, 84, 113, 86, 113, 87, 113, 88, 113, 89, 113, 90, 113, 92, 113, 93, 113, 94, 113, 95, 113, 96, 113, 97, 113, 98, 113, 99, 113, 100, 113, 101, 113, 102, 113, 103, 113, 104, 113, 105, 113, 106, 113, 107, 113, 108, 113, 109, 113, 111, 113, 112, 113, 113, 113, 114, 113, 115, 113, 116, 113, 117, 113, 118, 113, 119, 113, 120, 113, 122, 113, 161, 59, 211, 61, 219, 61, 80, 216, 191, 222, 80, 216, 193, 222, 80, 216, 201, 222, 80, 216, 202, 222, 80, 216, 238, 222, 80, 216, 250, 222, 81, 216, 33, 221, 58, 114, 67, 114, 78, 114, 79, 114, 80, 114, 81, 114, 82, 114, 140, 114, 141, 114, 142, 114, 143, 114, 144, 114, 145, 114, 72, 62, 81, 216, 151, 222, 46, 115, 55, 115, 65, 115, 58, 115, 59, 115, 60, 115, 61, 115, 62, 115, 63, 115, 64, 115, 66, 115, 69, 115, 70, 115, 71, 115, 72, 115, 73, 115, 74, 115, 82, 216, 18, 220, 82, 216, 35, 220, 39, 116, 215, 62, 30, 116, 61, 116, 62, 116, 63, 116, 64, 116, 65, 116, 66, 116, 67, 116, 68, 116, 69, 116, 70, 116, 71, 116, 72, 116, 73, 116, 74, 116, 75, 116, 76, 116, 77, 116, 78, 116, 79, 116, 80, 116, 81, 116, 82, 116, 83, 116, 84, 116, 60, 0, 42, 0, 85, 116, 86, 116, 87, 116, 88, 116, 89, 116, 90, 116, 91, 116, 92, 116, 93, 116, 94, 116, 95, 116, 222, 62, 225, 62, 226, 62, 82, 216, 183, 221, 82, 216, 184, 221, 82, 216, 185, 221, 82, 216, 186, 221, 82, 216, 187, 221, 82, 216, 197, 221, 82, 216, 208, 221, 82, 216, 218, 221, 82, 216, 222, 221, 82, 216, 223, 221, 82, 216, 227, 221, 82, 216, 229, 221, 111, 116, 225, 116, 253, 116, 254, 116, 255, 116, 1, 117, 29, 117, 30, 117, 117, 117, 118, 117, 119, 117, 120, 117, 121, 117, 122, 117, 99, 63, 83, 216, 158, 220, 83, 216, 159, 220, 236, 117, 237, 117, 238, 117, 239, 117, 240, 117, 241, 117, 242, 117, 243, 117, 244, 117, 245, 117, 246, 117, 247, 117, 248, 117, 249, 117, 250, 117, 251, 117, 252, 117, 253, 117, 254, 117, 255, 117, 0, 118, 1, 118, 2, 118, 3, 118, 4, 118, 5, 118, 6, 118, 83, 216, 234, 221, 83, 216, 235, 221, 15, 118, 16, 118, 151, 118, 152, 118, 153, 118, 83, 216, 130, 223, 83, 216, 134, 223, 181, 118, 221, 118, 222, 118, 223, 118, 4, 64, 82, 119, 83, 119, 84, 119, 85, 119, 86, 119, 87, 119, 88, 119, 89, 119, 90, 119, 91, 119, 92, 119, 93, 119, 94, 119, 95, 119, 96, 119, 98, 119, 99, 119, 100, 119, 101, 119, 102, 119, 103, 119, 104, 119, 105, 119, 106, 119, 107, 119, 108, 119, 109, 119, 69, 64, 84, 216, 205, 221, 84, 216, 227, 221, 84, 216, 230, 221, 84, 216, 231, 221, 97, 119, 121, 119, 224, 119, 238, 119, 120, 120, 121, 120, 122, 120, 123, 120, 124, 120, 125, 120, 127, 120, 128, 120, 60, 0, 42, 0, 129, 120, 130, 120, 131, 120, 132, 120, 133, 120, 134, 120, 135, 120, 136, 120, 137, 120, 138, 120, 139, 120, 140, 120, 141, 120, 142, 120, 143, 120, 144, 120, 145, 120, 146, 120, 147, 120, 148, 120, 149, 120, 150, 120, 151, 120, 152, 120, 153, 120, 154, 120, 155, 120, 156, 120, 176, 120, 187, 64, 191, 64, 121, 121, 122, 121, 123, 121, 124, 121, 125, 121, 126, 121, 127, 121, 128, 121, 129, 121, 130, 121, 131, 121, 133, 121, 134, 121, 14, 65, 189, 121, 44, 132, 15, 122, 16, 122, 17, 122, 18, 122, 19, 122, 20, 122, 21, 122, 22, 122, 23, 122, 24, 122, 25, 122, 26, 122, 27, 122, 28, 122, 29, 122, 30, 122, 31, 122, 32, 122, 33, 122, 34, 122, 35, 122, 36, 122, 37, 122, 85, 216, 223, 223, 85, 216, 224, 223, 85, 216, 225, 223, 158, 122, 159, 122, 160, 122, 161, 122, 162, 122, 163, 122, 164, 122, 165, 122, 166, 122, 167, 122, 86, 216, 172, 221, 232, 122, 233, 122, 234, 122, 235, 122, 207, 65, 86, 216, 149, 222, 94, 123, 95, 123, 96, 123, 97, 123, 98, 123, 99, 123, 100, 123, 101, 123, 102, 123, 103, 123, 104, 123, 105, 123, 106, 123, 107, 123, 109, 123, 110, 123, 111, 123, 112, 123, 113, 123, 114, 123, 115, 123, 116, 123, 118, 123, 119, 123, 120, 123, 121, 123, 122, 123, 123, 123, 124, 123, 125, 123, 126, 123, 127, 123, 128, 123, 192, 123, 248, 65, 249, 65, 251, 65, 252, 65, 60, 0, 42, 0, 253, 65, 254, 65, 255, 65, 0, 66, 86, 216, 116, 223, 86, 216, 137, 223, 174, 124, 175, 124, 176, 124, 177, 124, 178, 124, 179, 124, 180, 124, 181, 124, 192, 124, 140, 66, 87, 216, 129, 222, 87, 216, 130, 222, 87, 216, 131, 222, 91, 125, 120, 125, 121, 125, 122, 125, 123, 125, 124, 125, 125, 125, 127, 125, 128, 125, 129, 125, 130, 125, 131, 125, 132, 125, 133, 125, 134, 125, 135, 125, 136, 125, 137, 125, 138, 125, 139, 125, 140, 125, 141, 125, 142, 125, 143, 125, 144, 125, 145, 125, 146, 125, 147, 125, 148, 125, 149, 125, 151, 125, 152, 125, 153, 125, 154, 125, 155, 125, 88, 216, 33, 220, 88, 216, 41, 220, 25, 127, 26, 127, 27, 127, 28, 127, 29, 127, 30, 127, 31, 127, 32, 127, 33, 127, 34, 127, 35, 127, 36, 127, 103, 127, 104, 127, 105, 127, 106, 127, 107, 127, 108, 127, 109, 127, 110, 127, 114, 127, 88, 216, 208, 222, 163, 127, 164, 127, 165, 127, 166, 127, 167, 127, 168, 127, 169, 127, 170, 127, 219, 127, 220, 127, 221, 127, 33, 128, 34, 128, 85, 128, 86, 128, 87, 128, 88, 128, 132, 128, 133, 128, 134, 128, 80, 94, 33, 103, 91, 129, 92, 129, 93, 129, 94, 129, 95, 129, 96, 129, 97, 129, 98, 129, 99, 129, 100, 129, 101, 129, 102, 129, 103, 129, 104, 129, 105, 129, 106, 129, 107, 129, 108, 129, 109, 129, 110, 129, 111, 129, 112, 129, 113, 129, 114, 129, 115, 129, 117, 129, 118, 129, 119, 129, 60, 0, 42, 0, 120, 129, 121, 129, 122, 129, 123, 129, 124, 129, 125, 129, 126, 129, 19, 68, 5, 130, 29, 130, 64, 130, 65, 130, 66, 130, 67, 130, 68, 130, 69, 130, 70, 130, 71, 130, 72, 130, 73, 130, 90, 216, 66, 222, 187, 131, 195, 68, 197, 68, 206, 68, 217, 131, 37, 132, 40, 132, 41, 132, 42, 132, 43, 132, 45, 132, 46, 132, 47, 132, 48, 132, 49, 132, 50, 132, 51, 132, 52, 132, 53, 132, 54, 132, 55, 132, 57, 132, 58, 132, 59, 132, 60, 132, 61, 132, 62, 132, 63, 132, 64, 132, 65, 132, 66, 132, 67, 132, 68, 132, 69, 132, 70, 132, 71, 132, 72, 132, 73, 132, 74, 132, 75, 132, 76, 132, 77, 132, 78, 132, 79, 132, 80, 132, 81, 132, 82, 132, 83, 132, 84, 132, 85, 132, 86, 132, 88, 132, 89, 132, 90, 132, 91, 132, 92, 132, 93, 132, 94, 132, 95, 132, 96, 132, 97, 132, 98, 132, 99, 132, 100, 132, 101, 132, 102, 132, 103, 132, 104, 132, 105, 132, 106, 132, 107, 132, 108, 132, 109, 132, 110, 132, 111, 132, 112, 132, 113, 132, 114, 132, 115, 132, 116, 132, 117, 132, 118, 132, 119, 132, 120, 132, 121, 132, 122, 132, 123, 132, 124, 132, 125, 132, 126, 132, 127, 132, 128, 132, 129, 132, 130, 132, 131, 132, 132, 132, 133, 132, 134, 132, 135, 132, 136, 132, 137, 132, 139, 132, 140, 132, 141, 132, 142, 132, 143, 132, 91, 216, 195, 220, 60, 0, 42, 0, 91, 216, 209, 220, 91, 216, 34, 221, 91, 216, 35, 221, 91, 216, 36, 221, 91, 216, 37, 221, 91, 216, 38, 221, 91, 216, 39, 221, 91, 216, 40, 221, 91, 216, 41, 221, 91, 216, 42, 221, 91, 216, 81, 221, 197, 132, 200, 132, 241, 132, 7, 133, 92, 134, 94, 134, 95, 134, 214, 134, 245, 134, 246, 134, 247, 134, 248, 134, 249, 134, 250, 134, 251, 134, 252, 134, 253, 134, 254, 134, 255, 134, 0, 135, 1, 135, 2, 135, 3, 135, 4, 135, 5, 135, 6, 135, 7, 135, 8, 135, 9, 135, 10, 135, 11, 135, 12, 135, 13, 135, 14, 135, 15, 135, 16, 135, 19, 135, 20, 135, 21, 135, 22, 135, 23, 135, 70, 135, 92, 216, 230, 222, 77, 135, 88, 136, 89, 136, 202, 136, 203, 136, 204, 136, 205, 136, 206, 136, 207, 136, 208, 136, 209, 136, 210, 136, 211, 136, 212, 136, 213, 136, 214, 136, 216, 136, 217, 136, 218, 136, 219, 136, 220, 136, 221, 136, 222, 136, 223, 136, 224, 136, 225, 136, 227, 136, 228, 136, 229, 136, 133, 137, 155, 137, 156, 137, 128, 70, 206, 137, 220, 137, 223, 137, 224, 137, 225, 137, 226, 137, 227, 137, 228, 137, 229, 137, 230, 137, 231, 137, 62, 138, 63, 138, 97, 138, 98, 138, 99, 138, 100, 138, 101, 138, 102, 138, 103, 138, 104, 138, 105, 138, 106, 138, 107, 138, 108, 138, 109, 138, 110, 138, 111, 138, 112, 138, 113, 138, 114, 138, 115, 138, 116, 138, 117, 138, 118, 138, 119, 138, 120, 138, 121, 138, 122, 138, 123, 138, 124, 138, 60, 0, 42, 0, 125, 138, 126, 138, 127, 138, 128, 138, 129, 138, 130, 138, 131, 138, 132, 138, 133, 138, 134, 138, 135, 138, 136, 138, 137, 138, 138, 138, 160, 138, 94, 216, 221, 221, 40, 140, 41, 140, 42, 140, 43, 140, 44, 140, 60, 140, 74, 140, 75, 140, 54, 71, 98, 140, 99, 140, 100, 140, 101, 140, 102, 140, 132, 140, 133, 140, 134, 140, 135, 140, 136, 140, 137, 140, 138, 140, 95, 216, 197, 220, 178, 140, 194, 140, 195, 140, 196, 140, 197, 140, 198, 140, 199, 140, 200, 140, 201, 140, 202, 140, 203, 140, 204, 140, 205, 140, 206, 140, 95, 216, 102, 221, 95, 216, 115, 221, 95, 216, 132, 221, 86, 141, 87, 141, 104, 141, 105, 141, 106, 141, 140, 141, 141, 141, 142, 141, 143, 141, 144, 141, 145, 141, 146, 141, 147, 141, 148, 141, 166, 71, 208, 141, 188, 141, 223, 141, 224, 141, 225, 141, 226, 141, 227, 141, 228, 141, 229, 141, 230, 141, 231, 141, 232, 141, 233, 141, 234, 141, 235, 141, 236, 141, 237, 141, 238, 141, 239, 141, 240, 141, 241, 141, 242, 141, 243, 141, 244, 141, 246, 141, 247, 141, 248, 141, 249, 141, 250, 141, 251, 141, 96, 216, 30, 220, 96, 216, 35, 220, 96, 216, 36, 220, 177, 142, 178, 142, 55, 72, 237, 142, 254, 142, 255, 142, 0, 143, 1, 143, 2, 143, 3, 143, 4, 143, 5, 143, 6, 143, 7, 143, 8, 143, 9, 143, 10, 143, 11, 143, 12, 143, 143, 143, 144, 143, 145, 143, 146, 143, 147, 143, 148, 143, 60, 0, 42, 0, 158, 143, 159, 143, 160, 143, 178, 143, 60, 144, 61, 144, 62, 144, 63, 144, 64, 144, 65, 144, 66, 144, 67, 144, 68, 144, 69, 144, 70, 144, 71, 144, 72, 144, 73, 144, 74, 144, 75, 144, 76, 144, 77, 144, 78, 144, 79, 144, 80, 144, 81, 144, 82, 144, 83, 144, 84, 144, 85, 144, 86, 144, 87, 144, 97, 216, 108, 221, 204, 144, 11, 145, 12, 145, 13, 145, 14, 145, 15, 145, 16, 145, 17, 145, 18, 145, 19, 145, 20, 145, 21, 145, 22, 145, 23, 145, 102, 145, 103, 145, 104, 145, 105, 145, 106, 145, 107, 145, 108, 145, 109, 145, 110, 145, 111, 145, 112, 145, 113, 145, 98, 216, 43, 220, 46, 146, 47, 146, 48, 146, 49, 146, 50, 146, 51, 146, 52, 146, 53, 146, 54, 146, 55, 146, 56, 146, 57, 146, 58, 146, 59, 146, 60, 146, 61, 146, 62, 146, 63, 146, 64, 146, 65, 146, 66, 146, 67, 146, 68, 146, 69, 146, 70, 146, 71, 146, 72, 146, 73, 146, 74, 146, 75, 146, 76, 146, 77, 146, 78, 146, 79, 146, 80, 146, 81, 146, 82, 146, 83, 146, 84, 146, 85, 146, 86, 146, 87, 146, 88, 146, 89, 146, 90, 146, 91, 146, 92, 146, 93, 146, 94, 146, 95, 146, 96, 146, 97, 146, 98, 146, 99, 146, 100, 146, 101, 146, 102, 146, 103, 146, 104, 146, 105, 146, 106, 146, 107, 146, 108, 146, 109, 146, 110, 146, 111, 146, 112, 146, 113, 146, 60, 0, 42, 0, 114, 146, 115, 146, 116, 146, 143, 146, 175, 159, 98, 216, 104, 221, 98, 216, 108, 221, 98, 216, 109, 221, 98, 216, 126, 221, 22, 149, 23, 149, 24, 149, 25, 149, 26, 149, 27, 149, 28, 149, 29, 149, 30, 149, 31, 149, 32, 149, 33, 149, 34, 149, 35, 149, 36, 149, 37, 149, 38, 149, 39, 149, 40, 149, 41, 149, 42, 149, 43, 149, 44, 149, 45, 149, 46, 149, 47, 149, 48, 149, 49, 149, 137, 73, 152, 149, 153, 149, 154, 149, 155, 149, 156, 149, 157, 149, 158, 149, 159, 149, 160, 149, 22, 150, 23, 150, 24, 150, 25, 150, 143, 150, 145, 150, 146, 150, 147, 150, 148, 150, 149, 150, 150, 150, 151, 150, 152, 150, 41, 250, 223, 73, 201, 150, 202, 150, 203, 150, 205, 150, 206, 150, 207, 150, 244, 150, 245, 150, 246, 150, 247, 150, 248, 150, 249, 150, 250, 150, 251, 150, 252, 150, 253, 150, 254, 150, 100, 216, 147, 220, 85, 151, 86, 151, 114, 151, 115, 151, 116, 151, 117, 151, 118, 151, 119, 151, 120, 151, 121, 151, 234, 151, 235, 151, 238, 151, 244, 151, 245, 151, 9, 152, 10, 152, 11, 152, 12, 152, 13, 152, 14, 152, 15, 152, 16, 152, 17, 152, 18, 152, 19, 152, 180, 74, 144, 152, 145, 152, 146, 152, 147, 152, 148, 152, 149, 152, 150, 152, 171, 152, 172, 152, 212, 152, 236, 152, 238, 152, 241, 152, 38, 75, 243, 152, 244, 152, 245, 152, 246, 152, 247, 152, 249, 152, 251, 152, 60, 0, 42, 0, 252, 152, 253, 152, 254, 152, 255, 152, 137, 153, 140, 153, 141, 153, 142, 153, 143, 153, 144, 153, 154, 153, 175, 153, 176, 153, 177, 153, 178, 153, 179, 153, 180, 153, 181, 153, 156, 154, 157, 154, 158, 154, 159, 154, 170, 154, 171, 154, 172, 154, 173, 154, 174, 154, 102, 216, 40, 222, 225, 154, 226, 154, 61, 155, 91, 155, 92, 155, 93, 155, 94, 155, 132, 156, 133, 156, 134, 156, 135, 156, 136, 156, 137, 156, 138, 156, 139, 156, 140, 156, 141, 156, 142, 156, 143, 156, 144, 156, 231, 156, 232, 156, 233, 156, 234, 156, 235, 156, 237, 156, 238, 156, 239, 156, 240, 156, 73, 158, 74, 158, 75, 158, 76, 158, 77, 158, 78, 158, 79, 158, 80, 158, 81, 158, 82, 158, 83, 158, 84, 158, 128, 158, 129, 158, 130, 158, 104, 216, 255, 222, 253, 158, 12, 159, 14, 159, 19, 159, 20, 159, 32, 159, 131, 159, 132, 159, 133, 159, 134, 159, 60, 0, 39, 0, 208, 253, 14, 40, 39, 0, 60, 0, 42, 0, 64, 216, 78, 220, 72, 216, 161, 221, 206, 80, 207, 80, 208, 80, 209, 80, 210, 80, 211, 80, 212, 80, 213, 80, 214, 80, 215, 80, 216, 80, 217, 80, 218, 80, 219, 80, 220, 80, 221, 80, 222, 80, 223, 80, 224, 80, 225, 80, 226, 80, 227, 80, 228, 80, 229, 80, 230, 80, 231, 80, 232, 80, 233, 80, 234, 80, 235, 80, 236, 80, 237, 80, 238, 80, 239, 80, 240, 80, 241, 80, 243, 80, 244, 80, 247, 80, 64, 216, 126, 223, 64, 216, 127, 223, 64, 216, 128, 223, 64, 216, 160, 223, 64, 216, 167, 223, 242, 80, 98, 81, 169, 81, 216, 81, 243, 81, 244, 81, 128, 82, 129, 82, 130, 82, 131, 82, 132, 82, 0, 53, 4, 53, 6, 53, 232, 82, 233, 82, 234, 82, 235, 82, 236, 82, 237, 82, 34, 53, 48, 83, 49, 83, 50, 83, 53, 53, 172, 83, 173, 83, 174, 83, 176, 83, 81, 53, 198, 83, 66, 216, 13, 223, 14, 250, 97, 53, 246, 85, 247, 85, 248, 85, 249, 85, 250, 85, 251, 85, 252, 85, 253, 85, 254, 85, 255, 85, 0, 86, 1, 86, 2, 86, 3, 86, 4, 86, 5, 86, 6, 86, 7, 86, 8, 86, 9, 86, 10, 86, 11, 86, 12, 86, 13, 86, 14, 86, 15, 86, 16, 86, 17, 86, 18, 86, 19, 86, 20, 86, 21, 86, 22, 86, 23, 86, 24, 86, 25, 86, 26, 86, 27, 86, 28, 86, 29, 86, 30, 86, 33, 86, 34, 86, 35, 86, 36, 86, 37, 86, 38, 86, 39, 86, 60, 0, 42, 0, 81, 86, 67, 216, 215, 222, 67, 216, 216, 222, 67, 216, 221, 222, 67, 216, 248, 222, 67, 216, 249, 222, 67, 216, 250, 222, 67, 216, 251, 222, 67, 216, 29, 223, 67, 216, 38, 223, 67, 216, 45, 223, 67, 216, 46, 223, 67, 216, 48, 223, 67, 216, 49, 223, 67, 216, 59, 223, 67, 216, 76, 223, 31, 86, 40, 86, 22, 87, 23, 87, 24, 87, 25, 87, 114, 88, 115, 88, 116, 88, 117, 88, 118, 88, 119, 88, 120, 88, 121, 88, 122, 88, 123, 88, 124, 88, 125, 88, 126, 88, 127, 88, 129, 88, 130, 88, 131, 88, 132, 88, 133, 88, 134, 88, 135, 88, 136, 88, 137, 88, 138, 88, 139, 88, 140, 88, 141, 88, 142, 88, 143, 88, 144, 88, 145, 88, 146, 88, 147, 88, 148, 88, 149, 88, 150, 88, 151, 88, 152, 88, 153, 88, 154, 88, 155, 88, 97, 54, 101, 54, 102, 54, 68, 216, 237, 223, 68, 216, 254, 223, 69, 216, 19, 220, 69, 216, 22, 220, 173, 88, 253, 88, 254, 88, 16, 89, 34, 89, 35, 89, 36, 89, 37, 89, 105, 89, 106, 89, 107, 89, 108, 89, 213, 90, 214, 90, 215, 90, 216, 90, 217, 90, 218, 90, 219, 90, 220, 90, 221, 90, 222, 90, 223, 90, 224, 90, 225, 90, 226, 90, 227, 90, 228, 90, 229, 90, 230, 90, 231, 90, 232, 90, 233, 90, 234, 90, 235, 90, 236, 90, 237, 90, 238, 90, 239, 90, 240, 90, 241, 90, 242, 90, 28, 55, 30, 55, 32, 55, 34, 55, 70, 216, 32, 220, 70, 216, 40, 220, 70, 216, 41, 220, 70, 216, 42, 220, 70, 216, 45, 220, 70, 216, 57, 220, 70, 216, 58, 220, 70, 216, 59, 220, 70, 216, 64, 220, 70, 216, 69, 220, 243, 90, 117, 91, 60, 0, 42, 0, 119, 91, 70, 216, 150, 221, 222, 91, 223, 91, 224, 91, 225, 91, 226, 91, 227, 91, 228, 91, 229, 91, 230, 91, 231, 91, 232, 91, 13, 92, 33, 92, 98, 92, 99, 92, 119, 93, 120, 93, 121, 93, 122, 93, 123, 93, 124, 93, 125, 93, 126, 93, 127, 93, 128, 93, 129, 93, 130, 93, 131, 93, 132, 93, 133, 93, 134, 93, 135, 93, 136, 93, 137, 93, 138, 93, 139, 93, 140, 93, 141, 93, 142, 93, 251, 55, 0, 56, 1, 56, 4, 56, 81, 94, 82, 94, 83, 94, 84, 94, 85, 94, 86, 94, 87, 94, 88, 94, 89, 94, 91, 94, 72, 216, 42, 221, 99, 94, 205, 94, 206, 94, 207, 94, 208, 94, 209, 94, 210, 94, 211, 94, 212, 94, 213, 94, 214, 94, 215, 94, 216, 94, 217, 94, 220, 94, 10, 95, 67, 95, 68, 95, 69, 95, 195, 56, 70, 95, 111, 95, 112, 95, 209, 56, 177, 95, 179, 95, 180, 95, 73, 216, 84, 221, 182, 95, 40, 97, 44, 97, 51, 97, 56, 97, 59, 97, 63, 97, 65, 97, 66, 97, 71, 97, 72, 97, 75, 97, 80, 97, 120, 57, 82, 97, 83, 97, 84, 97, 86, 97, 88, 97, 90, 97, 91, 97, 94, 97, 95, 97, 96, 97, 97, 97, 98, 97, 99, 97, 101, 97, 105, 97, 106, 97, 108, 97, 111, 97, 113, 97, 114, 97, 115, 97, 116, 97, 117, 97, 119, 97, 122, 97, 123, 97, 125, 97, 128, 97, 129, 97, 134, 97, 60, 0, 42, 0, 136, 97, 128, 57, 74, 216, 3, 220, 39, 98, 40, 98, 41, 98, 42, 98, 43, 98, 44, 98, 43, 100, 52, 100, 59, 100, 63, 100, 75, 100, 76, 100, 77, 100, 78, 100, 79, 100, 80, 100, 81, 100, 83, 100, 84, 100, 85, 100, 87, 100, 88, 100, 89, 100, 90, 100, 91, 100, 92, 100, 93, 100, 94, 100, 95, 100, 96, 100, 98, 100, 99, 100, 100, 100, 101, 100, 102, 100, 103, 100, 106, 100, 107, 100, 108, 100, 109, 100, 113, 100, 114, 100, 115, 100, 116, 100, 117, 100, 118, 100, 119, 100, 120, 100, 122, 100, 123, 100, 124, 100, 125, 100, 126, 100, 127, 100, 129, 100, 130, 100, 132, 100, 135, 100, 47, 58, 49, 58, 50, 58, 51, 58, 52, 58, 53, 58, 54, 58, 55, 58, 56, 58, 57, 58, 75, 216, 183, 220, 75, 216, 194, 220, 75, 216, 198, 220, 75, 216, 201, 220, 75, 216, 7, 221, 75, 216, 8, 221, 75, 216, 18, 221, 86, 100, 166, 100, 113, 101, 114, 101, 115, 101, 160, 101, 161, 101, 178, 101, 76, 216, 183, 220, 214, 101, 215, 101, 154, 102, 155, 102, 156, 102, 157, 102, 158, 102, 159, 102, 160, 102, 161, 102, 162, 102, 163, 102, 164, 102, 165, 102, 166, 102, 167, 102, 168, 102, 13, 59, 14, 59, 15, 59, 16, 59, 76, 216, 86, 222, 76, 216, 94, 222, 76, 216, 98, 222, 4, 103, 5, 103, 34, 103, 58, 59, 59, 59, 138, 105, 141, 105, 142, 105, 143, 105, 144, 105, 145, 105, 146, 105, 147, 105, 149, 105, 150, 105, 151, 105, 153, 105, 60, 0, 42, 0, 154, 105, 155, 105, 156, 105, 157, 105, 158, 105, 159, 105, 160, 105, 161, 105, 162, 105, 163, 105, 164, 105, 165, 105, 166, 105, 167, 105, 168, 105, 169, 105, 170, 105, 171, 105, 172, 105, 173, 105, 174, 105, 175, 105, 176, 105, 177, 105, 178, 105, 179, 105, 180, 105, 181, 105, 182, 105, 183, 105, 184, 105, 185, 105, 186, 105, 187, 105, 188, 105, 189, 105, 190, 105, 191, 105, 192, 105, 193, 105, 194, 105, 195, 105, 196, 105, 197, 105, 198, 105, 199, 105, 200, 105, 201, 105, 202, 105, 203, 105, 204, 105, 205, 105, 206, 105, 207, 105, 208, 105, 209, 105, 210, 105, 211, 105, 212, 105, 213, 105, 214, 105, 215, 105, 216, 105, 217, 105, 218, 105, 219, 105, 220, 105, 221, 105, 222, 105, 223, 105, 224, 105, 225, 105, 46, 106, 20, 250, 188, 59, 77, 216, 187, 221, 77, 216, 205, 221, 77, 216, 206, 221, 77, 216, 207, 221, 77, 216, 243, 221, 77, 216, 0, 222, 3, 106, 73, 107, 74, 107, 75, 107, 76, 107, 77, 107, 112, 107, 116, 107, 157, 107, 158, 107, 159, 107, 160, 107, 161, 107, 195, 107, 196, 107, 78, 216, 250, 222, 254, 107, 50, 108, 51, 108, 206, 110, 204, 110, 235, 110, 236, 110, 237, 110, 238, 110, 239, 110, 240, 110, 241, 110, 242, 110, 243, 110, 244, 110, 245, 110, 246, 110, 247, 110, 248, 110, 249, 110, 250, 110, 251, 110, 252, 110, 253, 110, 254, 110, 255, 110, 1, 111, 2, 111, 3, 111, 4, 111, 5, 111, 60, 0, 42, 0, 6, 111, 7, 111, 8, 111, 9, 111, 10, 111, 12, 111, 13, 111, 14, 111, 15, 111, 17, 111, 18, 111, 20, 111, 21, 111, 22, 111, 23, 111, 24, 111, 25, 111, 26, 111, 27, 111, 28, 111, 29, 111, 30, 111, 31, 111, 32, 111, 33, 111, 34, 111, 35, 111, 36, 111, 37, 111, 39, 111, 40, 111, 41, 111, 42, 111, 43, 111, 44, 111, 45, 111, 46, 111, 47, 111, 48, 111, 49, 111, 50, 111, 51, 111, 52, 111, 53, 111, 54, 111, 55, 111, 56, 111, 57, 111, 58, 111, 59, 111, 60, 111, 62, 111, 64, 111, 66, 111, 67, 111, 68, 111, 69, 111, 70, 111, 71, 111, 72, 111, 73, 111, 74, 111, 75, 111, 76, 111, 77, 111, 61, 61, 70, 61, 79, 216, 215, 222, 79, 216, 247, 222, 79, 216, 248, 222, 79, 216, 249, 222, 79, 216, 250, 222, 79, 216, 251, 222, 79, 216, 252, 222, 79, 216, 53, 223, 79, 216, 65, 223, 78, 111, 115, 111, 85, 113, 91, 113, 121, 113, 123, 113, 124, 113, 125, 113, 126, 113, 127, 113, 128, 113, 129, 113, 130, 113, 131, 113, 132, 113, 133, 113, 134, 113, 135, 113, 136, 113, 137, 113, 138, 113, 139, 113, 140, 113, 141, 113, 142, 113, 143, 113, 144, 113, 145, 113, 146, 113, 147, 113, 148, 113, 149, 113, 150, 113, 151, 113, 152, 113, 153, 113, 184, 132, 231, 61, 232, 61, 80, 216, 13, 223, 80, 216, 26, 223, 80, 216, 52, 223, 80, 216, 72, 223, 51, 114, 62, 114, 68, 114, 83, 114, 84, 114, 34, 62, 146, 114, 147, 114, 148, 114, 60, 0, 42, 0, 149, 114, 150, 114, 151, 114, 67, 115, 68, 115, 76, 115, 77, 115, 79, 115, 80, 115, 81, 115, 82, 115, 83, 115, 84, 115, 85, 115, 96, 116, 97, 116, 98, 116, 99, 116, 100, 116, 101, 116, 102, 116, 103, 116, 104, 116, 106, 116, 107, 116, 109, 116, 110, 116, 112, 116, 113, 116, 114, 116, 115, 116, 116, 116, 117, 116, 118, 116, 119, 116, 120, 116, 231, 62, 233, 62, 82, 216, 236, 221, 82, 216, 237, 221, 82, 216, 246, 221, 82, 216, 247, 221, 82, 216, 248, 221, 82, 216, 249, 221, 82, 216, 251, 221, 82, 216, 14, 222, 82, 216, 18, 222, 82, 216, 19, 222, 0, 117, 2, 117, 3, 117, 4, 117, 5, 117, 6, 117, 82, 216, 110, 223, 39, 117, 123, 117, 124, 117, 125, 117, 144, 117, 145, 117, 7, 118, 8, 118, 9, 118, 10, 118, 11, 118, 12, 118, 13, 118, 14, 118, 17, 118, 18, 118, 19, 118, 20, 118, 21, 118, 22, 118, 23, 118, 24, 118, 39, 118, 182, 118, 183, 118, 184, 118, 185, 118, 224, 118, 225, 118, 226, 118, 227, 118, 110, 119, 111, 119, 112, 119, 113, 119, 114, 119, 115, 119, 116, 119, 117, 119, 118, 119, 119, 119, 120, 119, 122, 119, 123, 119, 124, 119, 125, 119, 126, 119, 127, 119, 128, 119, 129, 119, 130, 119, 131, 119, 132, 119, 133, 119, 134, 119, 83, 64, 84, 216, 32, 222, 84, 216, 33, 222, 126, 120, 157, 120, 158, 120, 159, 120, 160, 120, 161, 120, 162, 120, 163, 120, 164, 120, 165, 120, 166, 120, 167, 120, 168, 120, 169, 120, 170, 120, 60, 0, 42, 0, 171, 120, 172, 120, 173, 120, 174, 120, 175, 120, 177, 120, 178, 120, 179, 120, 180, 120, 181, 120, 182, 120, 183, 120, 184, 120, 185, 120, 193, 120, 200, 64, 135, 121, 136, 121, 137, 121, 138, 121, 139, 121, 140, 121, 141, 121, 142, 121, 143, 121, 144, 121, 145, 121, 146, 121, 147, 121, 148, 121, 149, 121, 150, 121, 151, 121, 152, 121, 153, 121, 38, 122, 39, 122, 40, 122, 41, 122, 42, 122, 43, 122, 44, 122, 45, 122, 46, 122, 47, 122, 48, 122, 49, 122, 50, 122, 51, 122, 74, 122, 103, 65, 53, 122, 168, 122, 169, 122, 170, 122, 171, 122, 172, 122, 173, 122, 86, 216, 204, 221, 236, 122, 237, 122, 239, 122, 240, 122, 86, 216, 156, 222, 238, 122, 117, 123, 129, 123, 130, 123, 131, 123, 132, 123, 133, 123, 134, 123, 135, 123, 136, 123, 137, 123, 138, 123, 139, 123, 140, 123, 141, 123, 142, 123, 143, 123, 144, 123, 145, 123, 146, 123, 147, 123, 148, 123, 149, 123, 150, 123, 151, 123, 152, 123, 153, 123, 154, 123, 155, 123, 156, 123, 157, 123, 158, 123, 159, 123, 160, 123, 161, 123, 162, 123, 163, 123, 164, 123, 165, 123, 166, 123, 167, 123, 168, 123, 169, 123, 170, 123, 171, 123, 1, 66, 2, 66, 3, 66, 4, 66, 5, 66, 6, 66, 7, 66, 8, 66, 9, 66, 10, 66, 11, 66, 12, 66, 13, 66, 86, 216, 179, 223, 86, 216, 180, 223, 86, 216, 198, 223, 184, 123, 182, 124, 183, 124, 60, 0, 42, 0, 184, 124, 185, 124, 186, 124, 187, 124, 188, 124, 189, 124, 190, 124, 191, 124, 193, 124, 87, 216, 166, 222, 87, 216, 188, 222, 150, 125, 208, 125, 156, 125, 157, 125, 158, 125, 159, 125, 160, 125, 161, 125, 162, 125, 163, 125, 164, 125, 165, 125, 166, 125, 167, 125, 168, 125, 169, 125, 170, 125, 171, 125, 172, 125, 173, 125, 174, 125, 175, 125, 176, 125, 177, 125, 178, 125, 179, 125, 180, 125, 181, 125, 182, 125, 183, 125, 184, 125, 185, 125, 186, 125, 187, 125, 188, 125, 189, 125, 190, 125, 191, 125, 192, 125, 193, 125, 194, 125, 195, 125, 196, 125, 197, 125, 198, 125, 199, 125, 200, 125, 201, 125, 202, 125, 203, 125, 204, 125, 205, 125, 206, 125, 207, 125, 209, 125, 210, 125, 212, 125, 213, 125, 232, 66, 233, 66, 236, 66, 237, 66, 241, 66, 88, 216, 72, 220, 88, 216, 100, 220, 37, 127, 38, 127, 39, 127, 40, 127, 41, 127, 42, 127, 43, 127, 65, 127, 66, 127, 111, 127, 112, 127, 113, 127, 115, 127, 116, 127, 171, 127, 222, 127, 223, 127, 224, 127, 225, 127, 226, 127, 227, 127, 228, 127, 89, 216, 74, 220, 229, 127, 35, 128, 36, 128, 37, 128, 89, 128, 90, 128, 91, 128, 92, 128, 93, 128, 94, 128, 95, 128, 97, 128, 98, 128, 99, 128, 186, 67, 89, 216, 160, 221, 135, 128, 136, 128, 80, 129, 127, 129, 128, 129, 129, 129, 130, 129, 131, 129, 132, 129, 133, 129, 134, 129, 135, 129, 136, 129, 60, 0, 42, 0, 137, 129, 138, 129, 139, 129, 140, 129, 141, 129, 142, 129, 143, 129, 145, 129, 32, 68, 37, 68, 38, 68, 89, 216, 153, 223, 89, 216, 179, 223, 89, 216, 180, 223, 89, 216, 204, 223, 231, 129, 250, 129, 7, 130, 19, 130, 20, 130, 21, 130, 30, 130, 74, 130, 75, 130, 76, 130, 77, 130, 90, 216, 81, 222, 90, 216, 82, 222, 221, 68, 222, 68, 223, 68, 225, 68, 228, 68, 233, 68, 234, 68, 235, 68, 236, 68, 144, 132, 145, 132, 146, 132, 147, 132, 148, 132, 149, 132, 150, 132, 151, 132, 152, 132, 153, 132, 154, 132, 155, 132, 156, 132, 157, 132, 158, 132, 159, 132, 160, 132, 161, 132, 162, 132, 163, 132, 164, 132, 165, 132, 166, 132, 167, 132, 168, 132, 169, 132, 170, 132, 171, 132, 172, 132, 173, 132, 174, 132, 175, 132, 176, 132, 177, 132, 178, 132, 179, 132, 180, 132, 181, 132, 182, 132, 183, 132, 185, 132, 186, 132, 187, 132, 188, 132, 189, 132, 190, 132, 191, 132, 192, 132, 193, 132, 194, 132, 195, 132, 196, 132, 198, 132, 199, 132, 201, 132, 202, 132, 203, 132, 204, 132, 205, 132, 206, 132, 207, 132, 208, 132, 209, 132, 210, 132, 211, 132, 212, 132, 213, 132, 214, 132, 215, 132, 216, 132, 217, 132, 218, 132, 219, 132, 220, 132, 221, 132, 223, 132, 225, 132, 226, 132, 227, 132, 228, 132, 230, 132, 91, 216, 116, 221, 91, 216, 160, 221, 91, 216, 161, 221, 91, 216, 162, 221, 91, 216, 163, 221, 91, 216, 164, 221, 91, 216, 165, 221, 91, 216, 166, 221, 91, 216, 167, 221, 91, 216, 174, 221, 60, 0, 42, 0, 91, 216, 220, 221, 91, 216, 234, 221, 91, 216, 235, 221, 91, 216, 240, 221, 229, 132, 96, 134, 97, 134, 92, 216, 205, 221, 17, 135, 18, 135, 43, 135, 24, 135, 25, 135, 26, 135, 27, 135, 28, 135, 29, 135, 30, 135, 31, 135, 32, 135, 33, 135, 34, 135, 35, 135, 36, 135, 37, 135, 38, 135, 39, 135, 40, 135, 41, 135, 42, 135, 44, 135, 45, 135, 46, 135, 47, 135, 48, 135, 49, 135, 50, 135, 51, 135, 52, 135, 53, 135, 54, 135, 55, 135, 56, 135, 57, 135, 58, 135, 59, 135, 60, 135, 61, 135, 62, 135, 63, 135, 64, 135, 65, 135, 66, 135, 67, 135, 68, 135, 69, 135, 71, 135, 72, 135, 73, 135, 74, 135, 75, 135, 85, 135, 107, 135, 226, 136, 231, 136, 232, 136, 233, 136, 234, 136, 235, 136, 236, 136, 237, 136, 238, 136, 239, 136, 240, 136, 241, 136, 242, 136, 243, 136, 244, 136, 245, 136, 246, 136, 247, 136, 248, 136, 249, 136, 250, 136, 251, 136, 252, 136, 253, 136, 254, 136, 255, 136, 0, 137, 2, 137, 3, 137, 4, 137, 59, 70, 26, 137, 157, 137, 158, 137, 159, 137, 160, 137, 161, 137, 207, 137, 232, 137, 233, 137, 234, 137, 235, 137, 139, 138, 140, 138, 141, 138, 142, 138, 143, 138, 144, 138, 145, 138, 146, 138, 147, 138, 148, 138, 150, 138, 151, 138, 152, 138, 153, 138, 154, 138, 155, 138, 156, 138, 157, 138, 158, 138, 159, 138, 161, 138, 162, 138, 163, 138, 60, 0, 42, 0, 164, 138, 165, 138, 166, 138, 167, 138, 168, 138, 169, 138, 170, 138, 171, 138, 172, 138, 173, 138, 174, 138, 94, 216, 253, 221, 94, 216, 10, 222, 94, 216, 14, 222, 45, 140, 46, 140, 47, 140, 48, 140, 49, 140, 50, 140, 61, 140, 103, 140, 104, 140, 105, 140, 106, 140, 139, 140, 140, 140, 141, 140, 207, 140, 208, 140, 209, 140, 210, 140, 211, 140, 212, 140, 213, 140, 214, 140, 215, 140, 216, 140, 98, 53, 95, 216, 143, 221, 95, 216, 152, 221, 88, 141, 89, 141, 90, 141, 91, 141, 107, 141, 149, 141, 150, 141, 151, 141, 152, 141, 153, 141, 154, 141, 252, 141, 253, 141, 254, 141, 255, 141, 0, 142, 1, 142, 2, 142, 3, 142, 4, 142, 5, 142, 6, 142, 7, 142, 8, 142, 9, 142, 10, 142, 11, 142, 12, 142, 13, 142, 14, 142, 244, 71, 96, 216, 72, 220, 179, 142, 180, 142, 181, 142, 13, 143, 14, 143, 15, 143, 16, 143, 17, 143, 18, 143, 19, 143, 20, 143, 21, 143, 96, 216, 6, 223, 96, 216, 24, 223, 149, 143, 150, 143, 151, 143, 161, 143, 162, 143, 163, 143, 88, 144, 89, 144, 90, 144, 91, 144, 92, 144, 93, 144, 94, 144, 95, 144, 96, 144, 97, 144, 98, 144, 99, 144, 100, 144, 101, 144, 210, 144, 24, 145, 25, 145, 26, 145, 27, 145, 28, 145, 29, 145, 30, 145, 31, 145, 32, 145, 33, 145, 34, 145, 35, 145, 36, 145, 37, 145, 114, 145, 115, 145, 116, 145, 117, 145, 118, 145, 119, 145, 60, 0, 42, 0, 120, 145, 121, 145, 122, 145, 123, 145, 124, 145, 125, 145, 126, 145, 127, 145, 250, 72, 45, 146, 117, 146, 118, 146, 119, 146, 120, 146, 121, 146, 122, 146, 123, 146, 125, 146, 126, 146, 127, 146, 128, 146, 129, 146, 130, 146, 131, 146, 132, 146, 133, 146, 134, 146, 135, 146, 136, 146, 137, 146, 138, 146, 139, 146, 140, 146, 141, 146, 142, 146, 144, 146, 145, 146, 146, 146, 147, 146, 148, 146, 149, 146, 150, 146, 151, 146, 152, 146, 153, 146, 154, 146, 155, 146, 156, 146, 157, 146, 158, 146, 159, 146, 160, 146, 161, 146, 162, 146, 163, 146, 164, 146, 165, 146, 166, 146, 167, 146, 168, 146, 169, 146, 170, 146, 171, 146, 172, 146, 173, 146, 174, 146, 175, 146, 176, 146, 177, 146, 36, 73, 37, 73, 38, 73, 42, 73, 98, 216, 137, 221, 98, 216, 168, 221, 98, 216, 170, 221, 98, 216, 171, 221, 238, 146, 124, 146, 50, 149, 51, 149, 52, 149, 53, 149, 54, 149, 55, 149, 56, 149, 57, 149, 58, 149, 59, 149, 60, 149, 61, 149, 62, 149, 63, 149, 64, 149, 65, 149, 66, 149, 67, 149, 68, 149, 69, 149, 161, 149, 162, 149, 163, 149, 164, 149, 165, 149, 166, 149, 167, 149, 168, 149, 169, 149, 170, 149, 26, 150, 153, 150, 154, 150, 155, 150, 156, 150, 157, 150, 158, 150, 159, 150, 160, 150, 161, 150, 204, 150, 208, 150, 209, 150, 210, 150, 99, 216, 197, 223, 255, 150, 0, 151, 1, 151, 15, 74, 60, 0, 42, 0, 100, 216, 175, 220, 100, 216, 176, 220, 100, 216, 177, 220, 87, 151, 88, 151, 89, 151, 100, 151, 122, 151, 123, 151, 124, 151, 125, 151, 126, 151, 127, 151, 128, 151, 129, 151, 130, 151, 131, 151, 132, 151, 133, 151, 134, 151, 205, 151, 206, 151, 236, 151, 246, 151, 247, 151, 101, 216, 29, 220, 25, 152, 20, 152, 21, 152, 22, 152, 23, 152, 24, 152, 26, 152, 184, 74, 151, 152, 173, 152, 174, 152, 175, 152, 176, 152, 177, 152, 255, 74, 0, 75, 101, 216, 184, 221, 214, 152, 213, 152, 215, 152, 248, 152, 44, 75, 0, 153, 1, 153, 2, 153, 3, 153, 4, 153, 5, 153, 6, 153, 7, 153, 9, 153, 12, 153, 14, 153, 15, 153, 145, 153, 146, 153, 155, 153, 156, 153, 157, 153, 111, 75, 182, 153, 183, 153, 184, 153, 185, 153, 186, 153, 187, 153, 188, 153, 189, 153, 190, 153, 191, 153, 192, 153, 193, 153, 194, 153, 195, 153, 196, 153, 197, 153, 198, 153, 199, 153, 123, 75, 126, 75, 160, 154, 161, 154, 162, 154, 175, 154, 176, 154, 177, 154, 200, 75, 218, 154, 227, 154, 228, 154, 229, 154, 230, 154, 231, 154, 232, 154, 233, 154, 234, 154, 102, 216, 5, 223, 102, 216, 14, 223, 38, 155, 62, 155, 63, 155, 64, 155, 65, 155, 66, 155, 103, 216, 173, 220, 95, 155, 96, 155, 97, 155, 98, 155, 103, 216, 90, 221, 145, 156, 146, 156, 147, 156, 148, 156, 149, 156, 150, 156, 151, 156, 152, 156, 153, 156, 154, 156, 155, 156, 156, 156, 60, 0, 42, 0, 157, 156, 158, 156, 159, 156, 241, 156, 242, 156, 243, 156, 244, 156, 245, 156, 246, 156, 85, 158, 86, 158, 87, 158, 89, 158, 90, 158, 91, 158, 92, 158, 167, 158, 188, 158, 189, 158, 59, 159, 74, 159, 135, 159, 136, 159, 60, 0, 39, 0, 208, 253, 15, 40, 39, 0, 60, 0, 42, 0, 245, 80, 246, 80, 248, 80, 249, 80, 250, 80, 251, 80, 252, 80, 253, 80, 254, 80, 255, 80, 0, 81, 1, 81, 2, 81, 3, 81, 4, 81, 5, 81, 6, 81, 7, 81, 8, 81, 9, 81, 10, 81, 11, 81, 12, 81, 13, 81, 14, 81, 15, 81, 147, 52, 150, 52, 152, 52, 64, 216, 201, 223, 64, 216, 203, 223, 48, 81, 217, 81, 218, 81, 219, 81, 220, 81, 65, 216, 17, 222, 133, 82, 134, 82, 135, 82, 136, 82, 137, 82, 138, 82, 139, 82, 140, 82, 141, 82, 142, 82, 143, 82, 238, 82, 239, 82, 240, 82, 241, 82, 242, 82, 20, 83, 51, 83, 177, 83, 178, 83, 66, 216, 180, 222, 199, 83, 89, 53, 83, 86, 32, 86, 41, 86, 42, 86, 43, 86, 44, 86, 45, 86, 46, 86, 47, 86, 48, 86, 49, 86, 50, 86, 51, 86, 52, 86, 53, 86, 54, 86, 55, 86, 56, 86, 57, 86, 58, 86, 59, 86, 60, 86, 61, 86, 62, 86, 63, 86, 64, 86, 65, 86, 66, 86, 67, 86, 68, 86, 70, 86, 71, 86, 72, 86, 73, 86, 74, 86, 75, 86, 76, 86, 77, 86, 78, 86, 79, 86, 80, 86, 82, 86, 84, 86, 86, 86, 87, 86, 88, 86, 89, 86, 90, 86, 91, 86, 92, 86, 93, 86, 116, 86, 241, 53, 242, 53, 243, 53, 67, 216, 100, 223, 67, 216, 141, 223, 67, 216, 144, 223, 67, 216, 173, 223, 67, 216, 180, 223, 67, 216, 181, 223, 67, 216, 182, 223, 67, 216, 188, 223, 68, 216, 20, 220, 26, 87, 128, 88, 156, 88, 157, 88, 60, 0, 42, 0, 158, 88, 159, 88, 160, 88, 161, 88, 162, 88, 163, 88, 164, 88, 165, 88, 166, 88, 167, 88, 169, 88, 170, 88, 171, 88, 172, 88, 174, 88, 175, 88, 176, 88, 177, 88, 178, 88, 179, 88, 180, 88, 181, 88, 182, 88, 183, 88, 184, 88, 185, 88, 69, 216, 36, 220, 69, 216, 63, 220, 69, 216, 82, 220, 69, 216, 84, 220, 69, 216, 85, 220, 255, 88, 0, 89, 69, 216, 119, 221, 38, 89, 109, 89, 244, 90, 245, 90, 246, 90, 247, 90, 248, 90, 249, 90, 250, 90, 251, 90, 252, 90, 253, 90, 254, 90, 255, 90, 0, 91, 1, 91, 2, 91, 3, 91, 4, 91, 5, 91, 6, 91, 7, 91, 8, 91, 9, 91, 10, 91, 11, 91, 12, 91, 13, 91, 14, 91, 15, 91, 35, 55, 37, 55, 38, 55, 70, 216, 82, 220, 70, 216, 94, 220, 70, 216, 97, 220, 70, 216, 98, 220, 70, 216, 99, 220, 70, 216, 100, 220, 70, 216, 119, 220, 70, 216, 123, 220, 70, 216, 131, 220, 70, 216, 132, 220, 70, 216, 133, 220, 233, 91, 235, 91, 236, 91, 237, 91, 238, 91, 111, 55, 14, 92, 70, 216, 68, 223, 53, 92, 100, 92, 101, 92, 102, 92, 103, 92, 160, 55, 143, 93, 144, 93, 145, 93, 146, 93, 147, 93, 148, 93, 149, 93, 150, 93, 151, 93, 152, 93, 153, 93, 154, 93, 155, 93, 156, 93, 157, 93, 158, 93, 159, 93, 160, 93, 161, 93, 162, 93, 163, 93, 164, 93, 165, 93, 15, 56, 71, 216, 15, 223, 71, 216, 21, 223, 228, 93, 90, 94, 92, 94, 93, 94, 94, 94, 95, 94, 96, 94, 97, 94, 98, 94, 100, 94, 60, 0, 42, 0, 101, 94, 97, 56, 105, 94, 218, 94, 219, 94, 221, 94, 222, 94, 223, 94, 224, 94, 225, 94, 226, 94, 227, 94, 228, 94, 71, 95, 72, 95, 73, 95, 113, 95, 178, 95, 181, 95, 183, 95, 184, 95, 185, 95, 186, 95, 85, 97, 87, 97, 89, 97, 92, 97, 93, 97, 100, 97, 102, 97, 103, 97, 107, 97, 110, 97, 112, 97, 118, 97, 120, 97, 121, 97, 124, 97, 126, 97, 127, 97, 130, 97, 131, 97, 132, 97, 133, 97, 135, 97, 109, 97, 137, 97, 139, 97, 141, 97, 142, 97, 143, 97, 144, 97, 146, 97, 147, 97, 148, 97, 149, 97, 152, 97, 154, 97, 155, 97, 156, 97, 158, 97, 159, 97, 161, 97, 162, 97, 163, 97, 164, 97, 166, 97, 167, 97, 170, 97, 171, 97, 172, 97, 173, 97, 174, 97, 175, 97, 176, 97, 177, 97, 179, 97, 137, 57, 138, 57, 146, 57, 74, 216, 95, 220, 74, 216, 96, 220, 74, 216, 113, 220, 45, 98, 46, 98, 47, 98, 104, 100, 105, 100, 110, 100, 111, 100, 112, 100, 121, 100, 128, 100, 131, 100, 60, 58, 133, 100, 134, 100, 136, 100, 138, 100, 139, 100, 140, 100, 141, 100, 142, 100, 143, 100, 144, 100, 145, 100, 146, 100, 147, 100, 148, 100, 149, 100, 150, 100, 151, 100, 152, 100, 153, 100, 154, 100, 155, 100, 156, 100, 157, 100, 158, 100, 159, 100, 160, 100, 161, 100, 162, 100, 163, 100, 164, 100, 165, 100, 167, 100, 168, 100, 60, 0, 42, 0, 169, 100, 170, 100, 171, 100, 172, 100, 173, 100, 174, 100, 175, 100, 176, 100, 177, 100, 178, 100, 179, 100, 180, 100, 181, 100, 182, 100, 183, 100, 184, 100, 185, 100, 186, 100, 198, 100, 75, 58, 75, 216, 68, 221, 75, 216, 76, 221, 75, 216, 103, 221, 117, 101, 118, 101, 119, 101, 120, 101, 121, 101, 122, 101, 123, 101, 75, 216, 204, 223, 179, 101, 76, 216, 188, 220, 169, 102, 170, 102, 171, 102, 172, 102, 173, 102, 174, 102, 175, 102, 176, 102, 177, 102, 178, 102, 179, 102, 180, 102, 181, 102, 182, 102, 183, 102, 17, 59, 18, 59, 19, 59, 20, 59, 21, 59, 22, 59, 76, 216, 129, 222, 76, 216, 138, 222, 188, 102, 164, 129, 60, 59, 61, 59, 62, 59, 233, 105, 226, 105, 227, 105, 228, 105, 229, 105, 230, 105, 231, 105, 232, 105, 234, 105, 235, 105, 236, 105, 237, 105, 238, 105, 239, 105, 240, 105, 241, 105, 242, 105, 243, 105, 244, 105, 245, 105, 246, 105, 247, 105, 248, 105, 249, 105, 250, 105, 251, 105, 252, 105, 253, 105, 254, 105, 255, 105, 0, 106, 1, 106, 2, 106, 4, 106, 5, 106, 6, 106, 7, 106, 8, 106, 9, 106, 10, 106, 11, 106, 12, 106, 13, 106, 14, 106, 15, 106, 16, 106, 17, 106, 18, 106, 19, 106, 20, 106, 21, 106, 22, 106, 23, 106, 24, 106, 25, 106, 26, 106, 27, 106, 28, 106, 29, 106, 30, 106, 31, 106, 32, 106, 33, 106, 34, 106, 35, 106, 36, 106, 37, 106, 60, 0, 42, 0, 38, 106, 39, 106, 41, 106, 42, 106, 43, 106, 44, 106, 45, 106, 47, 106, 48, 106, 49, 106, 101, 106, 190, 59, 194, 59, 196, 59, 77, 216, 23, 222, 77, 216, 26, 222, 77, 216, 60, 222, 77, 216, 64, 222, 77, 216, 89, 222, 77, 216, 95, 222, 77, 216, 119, 222, 78, 107, 79, 107, 80, 107, 81, 107, 82, 107, 83, 107, 117, 107, 118, 107, 162, 107, 163, 107, 164, 107, 165, 107, 166, 107, 115, 60, 116, 60, 197, 107, 198, 107, 255, 107, 0, 108, 1, 108, 2, 108, 213, 110, 0, 111, 16, 111, 38, 111, 63, 111, 65, 111, 11, 111, 61, 111, 79, 111, 80, 111, 81, 111, 82, 111, 83, 111, 84, 111, 85, 111, 86, 111, 87, 111, 88, 111, 89, 111, 90, 111, 91, 111, 92, 111, 93, 111, 95, 111, 96, 111, 97, 111, 98, 111, 99, 111, 100, 111, 101, 111, 102, 111, 103, 111, 104, 111, 105, 111, 106, 111, 107, 111, 108, 111, 109, 111, 110, 111, 111, 111, 112, 111, 113, 111, 114, 111, 116, 111, 117, 111, 118, 111, 119, 111, 120, 111, 121, 111, 122, 111, 123, 111, 124, 111, 125, 111, 126, 111, 127, 111, 129, 111, 130, 111, 132, 111, 133, 111, 134, 111, 135, 111, 136, 111, 137, 111, 138, 111, 139, 111, 140, 111, 141, 111, 142, 111, 143, 111, 144, 111, 145, 111, 146, 111, 147, 111, 148, 111, 149, 111, 150, 111, 151, 111, 152, 111, 154, 111, 155, 111, 156, 111, 157, 111, 208, 111, 76, 61, 78, 61, 81, 61, 60, 0, 42, 0, 79, 216, 74, 223, 79, 216, 97, 223, 79, 216, 127, 223, 79, 216, 128, 223, 79, 216, 129, 223, 79, 216, 130, 223, 79, 216, 143, 223, 79, 216, 180, 223, 79, 216, 183, 223, 198, 111, 166, 113, 154, 113, 155, 113, 156, 113, 157, 113, 158, 113, 159, 113, 160, 113, 161, 113, 162, 113, 163, 113, 164, 113, 165, 113, 167, 113, 168, 113, 169, 113, 170, 113, 171, 113, 172, 113, 173, 113, 174, 113, 175, 113, 176, 113, 177, 113, 178, 113, 179, 113, 180, 113, 181, 113, 217, 158, 166, 159, 235, 61, 80, 216, 98, 223, 80, 216, 99, 223, 80, 216, 100, 223, 80, 216, 101, 223, 80, 216, 140, 223, 80, 216, 150, 223, 80, 216, 156, 223, 85, 86, 52, 114, 69, 114, 85, 114, 86, 114, 87, 114, 152, 114, 153, 114, 154, 114, 155, 114, 81, 216, 212, 222, 78, 115, 75, 115, 86, 115, 87, 115, 88, 115, 89, 115, 90, 115, 91, 115, 92, 115, 93, 115, 94, 115, 95, 115, 96, 115, 97, 115, 98, 115, 100, 115, 82, 216, 130, 220, 105, 116, 108, 116, 121, 116, 122, 116, 123, 116, 124, 116, 125, 116, 126, 116, 128, 116, 129, 116, 130, 116, 131, 116, 132, 116, 133, 116, 134, 116, 135, 116, 136, 116, 137, 116, 138, 116, 139, 116, 140, 116, 142, 116, 147, 116, 235, 62, 236, 62, 240, 62, 243, 62, 244, 62, 82, 216, 21, 222, 82, 216, 33, 222, 82, 216, 34, 222, 82, 216, 35, 222, 82, 216, 36, 222, 82, 216, 37, 222, 82, 216, 38, 222, 82, 216, 39, 222, 82, 216, 40, 222, 82, 216, 41, 222, 82, 216, 42, 222, 82, 216, 62, 222, 7, 117, 8, 117, 9, 117, 83, 63, 82, 216, 245, 223, 126, 117, 127, 117, 31, 118, 25, 118, 26, 118, 27, 118, 28, 118, 60, 0, 42, 0, 29, 118, 30, 118, 32, 118, 33, 118, 34, 118, 35, 118, 36, 118, 37, 118, 38, 118, 40, 118, 41, 118, 42, 118, 43, 118, 183, 63, 192, 63, 83, 216, 59, 222, 83, 216, 80, 222, 154, 118, 155, 118, 156, 118, 157, 118, 158, 118, 229, 63, 83, 216, 151, 223, 83, 216, 154, 223, 186, 118, 228, 118, 135, 119, 136, 119, 137, 119, 138, 119, 139, 119, 140, 119, 141, 119, 142, 119, 143, 119, 144, 119, 145, 119, 146, 119, 147, 119, 87, 64, 84, 216, 80, 222, 186, 120, 187, 120, 188, 120, 189, 120, 190, 120, 191, 120, 192, 120, 194, 120, 195, 120, 196, 120, 197, 120, 198, 120, 199, 120, 200, 120, 201, 120, 202, 120, 203, 120, 204, 120, 205, 120, 206, 120, 207, 120, 208, 120, 209, 120, 210, 120, 211, 120, 212, 120, 213, 120, 215, 120, 216, 120, 217, 120, 228, 120, 85, 216, 49, 221, 85, 216, 53, 221, 85, 216, 63, 221, 154, 121, 155, 121, 156, 121, 157, 121, 158, 121, 159, 121, 160, 121, 161, 121, 162, 121, 163, 121, 85, 216, 227, 222, 52, 122, 54, 122, 55, 122, 56, 122, 57, 122, 58, 122, 59, 122, 60, 122, 61, 122, 62, 122, 63, 122, 64, 122, 65, 122, 66, 122, 67, 122, 108, 65, 110, 65, 117, 65, 86, 216, 87, 220, 174, 122, 175, 122, 176, 122, 177, 122, 178, 122, 179, 122, 180, 122, 172, 123, 173, 123, 174, 123, 175, 123, 176, 123, 177, 123, 178, 123, 179, 123, 180, 123, 181, 123, 182, 123, 183, 123, 185, 123, 186, 123, 187, 123, 60, 0, 42, 0, 188, 123, 189, 123, 190, 123, 191, 123, 193, 123, 194, 123, 195, 123, 196, 123, 197, 123, 198, 123, 199, 123, 200, 123, 202, 123, 203, 123, 204, 123, 205, 123, 206, 123, 207, 123, 208, 123, 209, 123, 210, 123, 211, 123, 14, 66, 15, 66, 16, 66, 17, 66, 18, 66, 19, 66, 20, 66, 21, 66, 26, 66, 28, 66, 32, 66, 34, 66, 35, 66, 38, 66, 41, 66, 86, 216, 228, 223, 86, 216, 232, 223, 87, 216, 1, 220, 87, 216, 6, 220, 194, 124, 195, 124, 196, 124, 197, 124, 198, 124, 199, 124, 200, 124, 201, 124, 202, 124, 203, 124, 204, 124, 205, 124, 206, 124, 148, 66, 87, 216, 215, 222, 87, 216, 216, 222, 211, 125, 214, 125, 215, 125, 216, 125, 217, 125, 218, 125, 219, 125, 220, 125, 221, 125, 222, 125, 223, 125, 224, 125, 225, 125, 226, 125, 227, 125, 228, 125, 229, 125, 230, 125, 231, 125, 232, 125, 233, 125, 234, 125, 235, 125, 236, 125, 237, 125, 238, 125, 239, 125, 240, 125, 241, 125, 242, 125, 243, 125, 244, 125, 245, 125, 246, 125, 247, 125, 248, 125, 249, 125, 250, 125, 251, 125, 252, 125, 253, 125, 254, 125, 255, 125, 0, 126, 1, 126, 2, 126, 3, 126, 4, 126, 5, 126, 6, 126, 7, 126, 244, 66, 251, 66, 252, 66, 0, 67, 1, 67, 4, 67, 88, 216, 131, 220, 88, 216, 151, 220, 88, 216, 164, 220, 88, 216, 165, 220, 62, 67, 44, 127, 45, 127, 46, 127, 47, 127, 117, 127, 118, 127, 119, 127, 120, 127, 172, 127, 60, 0, 42, 0, 173, 127, 174, 127, 175, 127, 176, 127, 230, 127, 231, 127, 232, 127, 233, 127, 234, 127, 235, 127, 236, 127, 237, 127, 89, 216, 105, 220, 38, 128, 39, 128, 172, 67, 100, 128, 101, 128, 102, 128, 103, 128, 104, 128, 105, 128, 106, 128, 107, 128, 89, 216, 173, 221, 146, 129, 147, 129, 148, 129, 149, 129, 150, 129, 151, 129, 152, 129, 153, 129, 154, 129, 155, 129, 156, 129, 157, 129, 158, 129, 159, 129, 160, 129, 161, 129, 162, 129, 163, 129, 45, 68, 51, 68, 52, 68, 241, 129, 22, 130, 23, 130, 78, 130, 79, 130, 80, 130, 81, 130, 82, 130, 83, 130, 84, 130, 122, 68, 138, 132, 244, 68, 224, 132, 231, 132, 232, 132, 233, 132, 234, 132, 235, 132, 236, 132, 237, 132, 238, 132, 239, 132, 240, 132, 242, 132, 243, 132, 244, 132, 245, 132, 246, 132, 247, 132, 248, 132, 249, 132, 250, 132, 251, 132, 252, 132, 253, 132, 254, 132, 255, 132, 0, 133, 1, 133, 2, 133, 3, 133, 4, 133, 5, 133, 6, 133, 8, 133, 9, 133, 10, 133, 11, 133, 12, 133, 13, 133, 14, 133, 15, 133, 16, 133, 17, 133, 18, 133, 19, 133, 20, 133, 21, 133, 22, 133, 23, 133, 24, 133, 25, 133, 26, 133, 27, 133, 28, 133, 29, 133, 30, 133, 31, 133, 32, 133, 33, 133, 34, 133, 35, 133, 36, 133, 37, 133, 38, 133, 39, 133, 40, 133, 41, 133, 42, 133, 43, 133, 44, 133, 60, 0, 42, 0, 45, 133, 46, 133, 47, 133, 48, 133, 49, 133, 50, 133, 51, 133, 52, 133, 53, 133, 54, 133, 55, 133, 56, 133, 57, 133, 58, 133, 59, 133, 60, 133, 91, 216, 0, 222, 91, 216, 5, 222, 91, 216, 7, 222, 91, 216, 18, 222, 91, 216, 66, 222, 91, 216, 67, 222, 91, 216, 68, 222, 91, 216, 69, 222, 91, 216, 110, 222, 91, 216, 114, 222, 91, 216, 119, 222, 91, 216, 132, 222, 61, 133, 79, 133, 98, 134, 76, 135, 78, 135, 79, 135, 80, 135, 81, 135, 82, 135, 83, 135, 84, 135, 86, 135, 87, 135, 88, 135, 89, 135, 90, 135, 91, 135, 92, 135, 93, 135, 94, 135, 95, 135, 96, 135, 97, 135, 98, 135, 99, 135, 100, 135, 101, 135, 102, 135, 103, 135, 104, 135, 105, 135, 106, 135, 108, 135, 109, 135, 110, 135, 111, 135, 112, 135, 113, 135, 114, 135, 115, 135, 116, 135, 117, 135, 118, 135, 119, 135, 120, 135, 122, 135, 123, 135, 124, 135, 125, 135, 126, 135, 127, 135, 128, 135, 225, 135, 92, 216, 82, 223, 92, 216, 154, 223, 130, 135, 90, 136, 91, 136, 92, 136, 93, 136, 93, 216, 253, 221, 230, 136, 5, 137, 6, 137, 7, 137, 8, 137, 9, 137, 10, 137, 11, 137, 12, 137, 13, 137, 14, 137, 15, 137, 16, 137, 17, 137, 18, 137, 19, 137, 20, 137, 21, 137, 22, 137, 23, 137, 24, 137, 25, 137, 27, 137, 28, 137, 29, 137, 69, 70, 70, 70, 93, 216, 15, 223, 162, 137, 163, 137, 164, 137, 165, 137, 94, 216, 88, 220, 169, 137, 208, 137, 209, 137, 236, 137, 237, 137, 238, 137, 60, 0, 42, 0, 239, 137, 240, 137, 149, 138, 175, 138, 176, 138, 177, 138, 178, 138, 179, 138, 180, 138, 181, 138, 182, 138, 183, 138, 184, 138, 185, 138, 186, 138, 187, 138, 188, 138, 189, 138, 190, 138, 191, 138, 192, 138, 193, 138, 194, 138, 195, 138, 196, 138, 197, 138, 198, 138, 199, 138, 200, 138, 201, 138, 202, 138, 203, 138, 204, 138, 205, 138, 206, 138, 207, 138, 208, 138, 209, 138, 210, 138, 211, 138, 212, 138, 213, 138, 214, 138, 215, 138, 216, 138, 217, 138, 218, 138, 233, 138, 245, 70, 247, 70, 94, 216, 62, 222, 94, 216, 83, 222, 94, 216, 89, 222, 219, 138, 248, 138, 51, 140, 52, 140, 53, 140, 62, 140, 76, 140, 77, 140, 78, 140, 94, 216, 239, 223, 94, 216, 244, 223, 108, 140, 142, 140, 143, 140, 217, 140, 218, 140, 219, 140, 220, 140, 221, 140, 222, 140, 223, 140, 224, 140, 225, 140, 226, 140, 227, 140, 228, 140, 229, 140, 230, 140, 231, 140, 232, 140, 233, 140, 234, 140, 235, 140, 236, 140, 124, 71, 237, 140, 92, 141, 109, 141, 155, 141, 156, 141, 157, 141, 158, 141, 159, 141, 160, 141, 161, 141, 162, 141, 163, 141, 164, 141, 15, 142, 16, 142, 17, 142, 18, 142, 19, 142, 20, 142, 21, 142, 22, 142, 23, 142, 24, 142, 25, 142, 26, 142, 27, 142, 28, 142, 29, 142, 30, 142, 31, 142, 32, 142, 33, 142, 34, 142, 35, 142, 36, 142, 37, 142, 38, 142, 39, 142, 40, 142, 41, 142, 60, 0, 42, 0, 42, 142, 44, 142, 45, 142, 46, 142, 47, 142, 58, 142, 0, 72, 1, 72, 11, 72, 96, 216, 131, 220, 96, 216, 144, 220, 43, 142, 55, 142, 182, 142, 183, 142, 184, 142, 185, 142, 186, 142, 187, 142, 188, 142, 96, 216, 86, 222, 22, 143, 23, 143, 24, 143, 25, 143, 26, 143, 27, 143, 28, 143, 29, 143, 30, 143, 31, 143, 32, 143, 33, 143, 34, 143, 35, 143, 36, 143, 37, 143, 38, 143, 39, 143, 40, 143, 41, 143, 42, 143, 43, 143, 44, 143, 93, 72, 96, 216, 47, 223, 96, 216, 58, 223, 152, 143, 164, 143, 179, 143, 102, 144, 103, 144, 104, 144, 105, 144, 106, 144, 107, 144, 108, 144, 109, 144, 110, 144, 111, 144, 112, 144, 113, 144, 115, 144, 119, 144, 246, 144, 38, 145, 39, 145, 41, 145, 42, 145, 43, 145, 45, 145, 46, 145, 47, 145, 48, 145, 49, 145, 50, 145, 221, 72, 128, 145, 129, 145, 130, 145, 131, 145, 132, 145, 133, 145, 134, 145, 135, 145, 136, 145, 137, 145, 138, 145, 139, 145, 140, 145, 178, 146, 179, 146, 180, 146, 181, 146, 182, 146, 183, 146, 184, 146, 185, 146, 186, 146, 187, 146, 188, 146, 189, 146, 190, 146, 191, 146, 192, 146, 193, 146, 194, 146, 195, 146, 196, 146, 197, 146, 198, 146, 199, 146, 200, 146, 201, 146, 202, 146, 204, 146, 205, 146, 206, 146, 207, 146, 208, 146, 209, 146, 210, 146, 211, 146, 212, 146, 213, 146, 214, 146, 215, 146, 216, 146, 60, 0, 42, 0, 217, 146, 218, 146, 219, 146, 220, 146, 221, 146, 222, 146, 223, 146, 224, 146, 225, 146, 226, 146, 227, 146, 228, 146, 229, 146, 230, 146, 231, 146, 232, 146, 233, 146, 234, 146, 235, 146, 236, 146, 237, 146, 239, 146, 240, 146, 241, 146, 242, 146, 243, 146, 244, 146, 245, 146, 246, 146, 39, 250, 45, 73, 47, 73, 48, 73, 98, 216, 184, 221, 98, 216, 188, 221, 98, 216, 192, 221, 98, 216, 220, 221, 98, 216, 222, 221, 98, 216, 225, 221, 98, 216, 227, 221, 98, 216, 228, 221, 70, 149, 71, 149, 72, 149, 73, 149, 74, 149, 75, 149, 76, 149, 77, 149, 78, 149, 79, 149, 80, 149, 81, 149, 82, 149, 83, 149, 84, 149, 85, 149, 124, 149, 171, 149, 172, 149, 173, 149, 174, 149, 175, 149, 176, 149, 177, 149, 178, 149, 179, 149, 180, 149, 99, 216, 52, 221, 162, 150, 163, 150, 164, 150, 165, 150, 229, 73, 211, 150, 2, 151, 3, 151, 4, 151, 5, 151, 6, 151, 7, 151, 8, 151, 9, 151, 10, 151, 100, 216, 192, 220, 90, 151, 96, 151, 101, 151, 135, 151, 136, 151, 137, 151, 138, 151, 139, 151, 140, 151, 141, 151, 142, 151, 143, 151, 144, 151, 145, 151, 146, 151, 207, 151, 208, 151, 209, 151, 239, 151, 101, 216, 32, 220, 27, 152, 28, 152, 29, 152, 30, 152, 31, 152, 32, 152, 33, 152, 34, 152, 35, 152, 38, 152, 39, 152, 40, 152, 41, 152, 42, 152, 43, 152, 44, 152, 152, 152, 153, 152, 154, 152, 155, 152, 156, 152, 157, 152, 178, 152, 60, 0, 42, 0, 179, 152, 216, 152, 250, 152, 8, 153, 10, 153, 11, 153, 13, 153, 55, 75, 17, 153, 18, 153, 19, 153, 20, 153, 21, 153, 22, 153, 23, 153, 24, 153, 25, 153, 147, 153, 148, 153, 200, 153, 201, 153, 202, 153, 203, 153, 204, 153, 205, 153, 206, 153, 207, 153, 208, 153, 209, 153, 210, 153, 211, 153, 212, 153, 213, 153, 214, 153, 215, 153, 216, 153, 217, 153, 218, 153, 219, 153, 220, 153, 221, 153, 222, 153, 223, 153, 224, 153, 102, 216, 164, 220, 163, 154, 178, 154, 179, 154, 180, 154, 181, 154, 182, 154, 183, 154, 219, 154, 235, 154, 236, 154, 238, 154, 239, 154, 240, 154, 241, 154, 242, 154, 243, 154, 244, 154, 39, 155, 67, 155, 68, 155, 69, 155, 70, 155, 32, 76, 99, 155, 100, 155, 101, 155, 102, 155, 103, 155, 104, 155, 105, 155, 106, 155, 107, 155, 108, 155, 109, 155, 110, 155, 111, 155, 112, 155, 113, 155, 114, 155, 115, 155, 116, 155, 117, 155, 118, 155, 119, 155, 120, 155, 121, 155, 59, 76, 62, 76, 103, 216, 124, 221, 160, 156, 161, 156, 162, 156, 163, 156, 164, 156, 165, 156, 166, 156, 167, 156, 168, 156, 169, 156, 170, 156, 171, 156, 172, 156, 247, 156, 248, 156, 249, 156, 250, 156, 251, 156, 252, 156, 253, 156, 254, 156, 255, 156, 0, 157, 1, 157, 2, 157, 3, 157, 4, 157, 5, 157, 6, 157, 7, 157, 8, 157, 9, 157, 11, 157, 12, 157, 60, 0, 42, 0, 13, 157, 14, 157, 174, 76, 176, 76, 183, 76, 103, 216, 183, 223, 88, 158, 93, 158, 94, 158, 95, 158, 96, 158, 97, 158, 98, 158, 99, 158, 100, 158, 118, 158, 131, 158, 132, 158, 104, 216, 147, 222, 168, 158, 169, 158, 170, 158, 171, 158, 185, 158, 190, 158, 206, 158, 168, 88, 211, 158, 15, 159, 16, 159, 17, 159, 81, 159, 82, 159, 137, 159, 138, 159, 60, 0, 39, 0, 208, 253, 16, 40, 39, 0, 60, 0, 42, 0, 184, 78, 16, 81, 17, 81, 18, 81, 19, 81, 20, 81, 21, 81, 22, 81, 23, 81, 24, 81, 25, 81, 26, 81, 27, 81, 28, 81, 29, 81, 30, 81, 43, 81, 64, 216, 245, 223, 64, 216, 252, 223, 99, 81, 65, 216, 252, 220, 128, 81, 170, 81, 221, 81, 222, 81, 65, 216, 21, 222, 144, 82, 145, 82, 146, 82, 147, 82, 148, 82, 243, 82, 52, 83, 225, 83, 94, 86, 95, 86, 96, 86, 97, 86, 98, 86, 99, 86, 100, 86, 101, 86, 102, 86, 103, 86, 104, 86, 105, 86, 106, 86, 107, 86, 108, 86, 109, 86, 110, 86, 111, 86, 112, 86, 113, 86, 114, 86, 115, 86, 117, 86, 118, 86, 119, 86, 120, 86, 121, 86, 122, 86, 123, 86, 124, 86, 251, 53, 254, 53, 0, 54, 1, 54, 2, 54, 3, 54, 4, 54, 67, 216, 223, 223, 67, 216, 234, 223, 67, 216, 235, 223, 67, 216, 236, 223, 67, 216, 237, 223, 68, 216, 29, 220, 68, 216, 30, 220, 27, 87, 28, 87, 186, 88, 187, 88, 188, 88, 189, 88, 190, 88, 191, 88, 192, 88, 193, 88, 194, 88, 195, 88, 196, 88, 197, 88, 198, 88, 199, 88, 200, 88, 201, 88, 202, 88, 203, 88, 204, 88, 179, 159, 1, 89, 110, 89, 111, 89, 16, 91, 17, 91, 18, 91, 19, 91, 20, 91, 21, 91, 22, 91, 23, 91, 24, 91, 25, 91, 26, 91, 27, 91, 28, 91, 29, 91, 30, 91, 31, 91, 32, 91, 33, 91, 34, 91, 52, 91, 43, 55, 44, 55, 45, 55, 70, 216, 158, 220, 70, 216, 159, 220, 60, 0, 42, 0, 70, 216, 160, 220, 70, 216, 161, 220, 70, 216, 162, 220, 70, 216, 190, 220, 70, 216, 191, 220, 40, 91, 120, 91, 121, 91, 239, 91, 240, 91, 166, 93, 167, 93, 168, 93, 169, 93, 170, 93, 171, 93, 172, 93, 173, 93, 174, 93, 175, 93, 176, 93, 177, 93, 178, 93, 179, 93, 180, 93, 181, 93, 182, 93, 102, 94, 103, 94, 104, 94, 98, 56, 99, 56, 100, 56, 72, 216, 91, 221, 111, 94, 229, 94, 230, 94, 231, 94, 232, 94, 233, 94, 234, 94, 74, 95, 75, 95, 91, 95, 92, 95, 73, 216, 113, 220, 73, 216, 176, 220, 187, 95, 188, 95, 138, 97, 140, 97, 145, 97, 150, 97, 151, 97, 153, 97, 157, 97, 160, 97, 165, 97, 168, 97, 169, 97, 178, 97, 180, 97, 182, 97, 183, 97, 184, 97, 185, 97, 186, 97, 189, 97, 190, 97, 191, 97, 192, 97, 193, 97, 196, 97, 197, 97, 198, 97, 200, 97, 202, 97, 204, 97, 205, 97, 206, 97, 207, 97, 208, 97, 210, 97, 211, 97, 212, 97, 153, 57, 74, 216, 173, 220, 74, 216, 193, 220, 187, 97, 49, 98, 48, 98, 137, 100, 187, 100, 188, 100, 189, 100, 190, 100, 191, 100, 192, 100, 193, 100, 194, 100, 195, 100, 196, 100, 197, 100, 199, 100, 200, 100, 201, 100, 203, 100, 204, 100, 205, 100, 207, 100, 208, 100, 209, 100, 210, 100, 211, 100, 212, 100, 213, 100, 214, 100, 215, 100, 217, 100, 218, 100, 219, 100, 220, 100, 221, 100, 222, 100, 82, 58, 84, 58, 87, 58, 75, 216, 141, 221, 60, 0, 42, 0, 75, 216, 149, 221, 75, 216, 160, 221, 75, 216, 163, 221, 75, 216, 164, 221, 75, 216, 183, 221, 243, 100, 51, 101, 116, 101, 124, 101, 125, 101, 126, 101, 127, 101, 75, 216, 227, 223, 147, 101, 162, 101, 180, 101, 216, 101, 217, 101, 184, 102, 185, 102, 186, 102, 187, 102, 189, 102, 190, 102, 191, 102, 192, 102, 193, 102, 194, 102, 195, 102, 196, 102, 197, 102, 198, 102, 199, 102, 200, 102, 201, 102, 202, 102, 203, 102, 204, 102, 205, 102, 167, 159, 23, 59, 24, 59, 25, 59, 26, 59, 27, 59, 28, 59, 29, 59, 30, 59, 31, 59, 76, 216, 137, 222, 76, 216, 171, 222, 76, 216, 172, 222, 76, 216, 173, 222, 207, 102, 6, 103, 49, 59, 35, 103, 36, 103, 63, 59, 40, 106, 116, 106, 50, 106, 51, 106, 52, 106, 53, 106, 54, 106, 55, 106, 56, 106, 57, 106, 58, 106, 59, 106, 60, 106, 61, 106, 62, 106, 63, 106, 64, 106, 65, 106, 66, 106, 67, 106, 68, 106, 69, 106, 70, 106, 71, 106, 72, 106, 73, 106, 74, 106, 75, 106, 76, 106, 77, 106, 78, 106, 79, 106, 80, 106, 81, 106, 82, 106, 83, 106, 84, 106, 85, 106, 86, 106, 87, 106, 88, 106, 89, 106, 90, 106, 91, 106, 92, 106, 93, 106, 94, 106, 95, 106, 96, 106, 97, 106, 98, 106, 99, 106, 100, 106, 102, 106, 103, 106, 104, 106, 105, 106, 106, 106, 107, 106, 108, 106, 109, 106, 110, 106, 111, 106, 112, 106, 113, 106, 114, 106, 115, 106, 117, 106, 118, 106, 60, 0, 42, 0, 119, 106, 120, 106, 121, 106, 122, 106, 123, 106, 124, 106, 215, 59, 221, 59, 77, 216, 142, 222, 77, 216, 158, 222, 77, 216, 166, 222, 77, 216, 173, 222, 77, 216, 186, 222, 77, 216, 223, 222, 77, 216, 238, 222, 84, 107, 85, 107, 86, 107, 87, 107, 88, 107, 89, 107, 90, 107, 119, 107, 167, 107, 168, 107, 169, 107, 170, 107, 171, 107, 199, 107, 200, 107, 3, 108, 4, 108, 5, 108, 6, 108, 7, 108, 94, 111, 131, 111, 153, 111, 158, 111, 159, 111, 160, 111, 161, 111, 162, 111, 163, 111, 164, 111, 165, 111, 166, 111, 167, 111, 168, 111, 170, 111, 171, 111, 172, 111, 173, 111, 174, 111, 175, 111, 176, 111, 177, 111, 178, 111, 179, 111, 180, 111, 181, 111, 182, 111, 183, 111, 184, 111, 185, 111, 186, 111, 187, 111, 188, 111, 189, 111, 190, 111, 191, 111, 192, 111, 193, 111, 194, 111, 195, 111, 196, 111, 197, 111, 199, 111, 200, 111, 201, 111, 202, 111, 203, 111, 205, 111, 206, 111, 207, 111, 209, 111, 210, 111, 211, 111, 214, 111, 95, 61, 98, 61, 101, 61, 105, 61, 106, 61, 79, 216, 197, 223, 79, 216, 235, 223, 79, 216, 236, 223, 79, 216, 237, 223, 79, 216, 238, 223, 79, 216, 239, 223, 79, 216, 240, 223, 80, 216, 17, 220, 4, 112, 182, 113, 183, 113, 184, 113, 185, 113, 186, 113, 187, 113, 188, 113, 189, 113, 190, 113, 191, 113, 192, 113, 193, 113, 194, 113, 195, 113, 196, 113, 197, 113, 198, 113, 199, 113, 200, 113, 201, 113, 202, 113, 203, 113, 204, 113, 205, 113, 206, 113, 60, 0, 42, 0, 207, 113, 208, 113, 209, 113, 210, 113, 211, 113, 212, 113, 213, 113, 214, 113, 215, 113, 216, 113, 217, 113, 218, 113, 219, 113, 220, 113, 221, 113, 222, 113, 243, 61, 244, 61, 247, 61, 252, 61, 253, 61, 80, 216, 189, 223, 80, 216, 193, 223, 80, 216, 233, 223, 80, 216, 234, 223, 80, 216, 242, 223, 156, 114, 157, 114, 158, 114, 159, 114, 99, 115, 101, 115, 102, 115, 103, 115, 104, 115, 105, 115, 106, 115, 107, 115, 108, 115, 109, 115, 127, 116, 141, 116, 143, 116, 145, 116, 146, 116, 148, 116, 149, 116, 150, 116, 152, 116, 153, 116, 154, 116, 155, 116, 156, 116, 157, 116, 158, 116, 159, 116, 160, 116, 161, 116, 163, 116, 164, 116, 252, 62, 82, 216, 66, 222, 82, 216, 69, 222, 82, 216, 74, 222, 82, 216, 78, 222, 82, 216, 79, 222, 82, 216, 80, 222, 82, 216, 81, 222, 82, 216, 93, 222, 82, 216, 101, 222, 82, 216, 102, 222, 82, 216, 103, 222, 162, 116, 226, 116, 82, 216, 15, 223, 10, 117, 11, 117, 12, 117, 13, 117, 14, 117, 128, 117, 129, 117, 130, 117, 83, 216, 201, 220, 44, 118, 45, 118, 46, 118, 47, 118, 48, 118, 49, 118, 50, 118, 51, 118, 52, 118, 53, 118, 54, 118, 55, 118, 56, 118, 57, 118, 58, 118, 59, 118, 60, 118, 61, 118, 62, 118, 63, 118, 74, 118, 200, 63, 159, 118, 160, 118, 161, 118, 187, 118, 229, 118, 230, 118, 231, 118, 84, 216, 157, 220, 148, 119, 149, 119, 150, 119, 151, 119, 152, 119, 153, 119, 154, 119, 155, 119, 156, 119, 157, 119, 158, 119, 159, 119, 160, 119, 161, 119, 60, 0, 42, 0, 162, 119, 163, 119, 98, 64, 101, 64, 106, 64, 84, 216, 153, 222, 165, 119, 214, 120, 220, 120, 218, 120, 219, 120, 221, 120, 222, 120, 223, 120, 224, 120, 225, 120, 226, 120, 227, 120, 229, 120, 230, 120, 231, 120, 232, 120, 233, 120, 234, 120, 235, 120, 236, 120, 237, 120, 238, 120, 216, 64, 85, 216, 91, 221, 85, 216, 92, 221, 85, 216, 93, 221, 85, 216, 94, 221, 85, 216, 98, 221, 85, 216, 101, 221, 85, 216, 102, 221, 164, 121, 165, 121, 166, 121, 169, 121, 85, 216, 246, 222, 68, 122, 69, 122, 70, 122, 71, 122, 72, 122, 75, 122, 76, 122, 77, 122, 78, 122, 79, 122, 80, 122, 81, 122, 82, 122, 124, 65, 86, 216, 93, 220, 86, 216, 114, 220, 83, 122, 181, 122, 182, 122, 183, 122, 184, 122, 185, 122, 186, 122, 187, 122, 188, 122, 189, 122, 178, 65, 241, 122, 86, 216, 174, 222, 201, 123, 212, 123, 213, 123, 214, 123, 215, 123, 216, 123, 217, 123, 218, 123, 219, 123, 220, 123, 221, 123, 222, 123, 223, 123, 224, 123, 225, 123, 226, 123, 227, 123, 228, 123, 229, 123, 230, 123, 231, 123, 232, 123, 233, 123, 234, 123, 235, 123, 236, 123, 237, 123, 238, 123, 239, 123, 17, 124, 42, 66, 43, 66, 45, 66, 46, 66, 48, 66, 49, 66, 50, 66, 87, 216, 33, 220, 87, 216, 74, 220, 87, 216, 101, 220, 249, 123, 207, 124, 208, 124, 209, 124, 210, 124, 211, 124, 212, 124, 213, 124, 214, 124, 215, 124, 216, 124, 0, 74, 8, 126, 9, 126, 10, 126, 11, 126, 12, 126, 13, 126, 60, 0, 42, 0, 14, 126, 15, 126, 16, 126, 17, 126, 18, 126, 19, 126, 20, 126, 21, 126, 22, 126, 23, 126, 24, 126, 25, 126, 26, 126, 27, 126, 28, 126, 29, 126, 30, 126, 31, 126, 32, 126, 33, 126, 34, 126, 35, 126, 36, 126, 37, 126, 38, 126, 39, 126, 40, 126, 10, 67, 48, 127, 49, 127, 50, 127, 51, 127, 52, 127, 67, 127, 121, 127, 122, 127, 123, 127, 124, 127, 177, 127, 178, 127, 238, 127, 239, 127, 240, 127, 241, 127, 40, 128, 41, 128, 42, 128, 89, 216, 18, 221, 108, 128, 109, 128, 193, 67, 89, 216, 191, 221, 110, 128, 144, 129, 37, 103, 166, 129, 167, 129, 168, 129, 169, 129, 170, 129, 171, 129, 172, 129, 173, 129, 174, 129, 175, 129, 176, 129, 177, 129, 178, 129, 179, 129, 180, 129, 181, 129, 182, 129, 59, 68, 90, 216, 28, 220, 90, 216, 110, 220, 183, 129, 185, 129, 242, 129, 251, 129, 8, 130, 6, 130, 9, 130, 24, 130, 85, 130, 86, 130, 87, 130, 88, 130, 89, 130, 222, 132, 0, 69, 3, 69, 4, 69, 9, 69, 11, 69, 62, 133, 63, 133, 64, 133, 65, 133, 66, 133, 67, 133, 68, 133, 69, 133, 70, 133, 71, 133, 72, 133, 73, 133, 74, 133, 75, 133, 76, 133, 77, 133, 78, 133, 80, 133, 81, 133, 82, 133, 83, 133, 84, 133, 85, 133, 86, 133, 88, 133, 89, 133, 90, 133, 91, 133, 92, 133, 93, 133, 94, 133, 95, 133, 96, 133, 97, 133, 60, 0, 42, 0, 98, 133, 99, 133, 100, 133, 101, 133, 102, 133, 103, 133, 104, 133, 105, 133, 106, 133, 107, 133, 108, 133, 109, 133, 110, 133, 111, 133, 112, 133, 113, 133, 114, 133, 115, 133, 116, 133, 117, 133, 91, 216, 136, 222, 91, 216, 139, 222, 91, 216, 153, 222, 91, 216, 208, 222, 91, 216, 209, 222, 91, 216, 210, 222, 91, 216, 211, 222, 91, 216, 212, 222, 91, 216, 213, 222, 91, 216, 214, 222, 91, 216, 215, 222, 91, 216, 38, 223, 140, 133, 31, 250, 99, 134, 100, 134, 101, 134, 102, 134, 121, 135, 129, 135, 131, 135, 132, 135, 133, 135, 134, 135, 135, 135, 136, 135, 137, 135, 138, 135, 139, 135, 140, 135, 141, 135, 142, 135, 143, 135, 144, 135, 145, 135, 146, 135, 147, 135, 148, 135, 149, 135, 150, 135, 151, 135, 152, 135, 153, 135, 154, 135, 155, 135, 156, 135, 157, 135, 158, 135, 159, 135, 160, 135, 161, 135, 162, 135, 163, 135, 164, 135, 165, 135, 166, 135, 167, 135, 168, 135, 169, 135, 218, 69, 221, 69, 94, 136, 95, 136, 96, 136, 97, 136, 23, 70, 93, 216, 254, 221, 30, 137, 31, 137, 32, 137, 33, 137, 34, 137, 35, 137, 36, 137, 37, 137, 38, 137, 39, 137, 40, 137, 41, 137, 42, 137, 43, 137, 44, 137, 45, 137, 46, 137, 47, 137, 48, 137, 49, 137, 50, 137, 52, 137, 79, 70, 93, 216, 53, 223, 93, 216, 54, 223, 93, 216, 65, 223, 166, 137, 167, 137, 168, 137, 170, 137, 94, 216, 112, 220, 241, 137, 161, 70, 220, 138, 221, 138, 222, 138, 223, 138, 224, 138, 225, 138, 226, 138, 227, 138, 60, 0, 42, 0, 228, 138, 229, 138, 230, 138, 231, 138, 232, 138, 234, 138, 235, 138, 236, 138, 237, 138, 238, 138, 239, 138, 240, 138, 241, 138, 242, 138, 243, 138, 244, 138, 245, 138, 246, 138, 247, 138, 249, 138, 250, 138, 251, 138, 252, 138, 253, 138, 254, 138, 255, 138, 0, 139, 1, 139, 2, 139, 3, 139, 94, 216, 121, 222, 94, 216, 132, 222, 20, 139, 107, 140, 109, 140, 110, 140, 78, 71, 95, 216, 108, 220, 144, 140, 145, 140, 146, 140, 147, 140, 238, 140, 239, 140, 240, 140, 241, 140, 242, 140, 243, 140, 244, 140, 245, 140, 95, 216, 189, 221, 93, 141, 94, 141, 95, 141, 96, 141, 108, 141, 110, 141, 165, 141, 166, 141, 167, 141, 95, 216, 46, 223, 48, 142, 49, 142, 50, 142, 51, 142, 52, 142, 53, 142, 54, 142, 56, 142, 57, 142, 59, 142, 60, 142, 61, 142, 62, 142, 63, 142, 64, 142, 65, 142, 66, 142, 67, 142, 68, 142, 69, 142, 96, 216, 189, 220, 96, 216, 190, 220, 189, 142, 190, 142, 45, 143, 46, 143, 47, 143, 48, 143, 49, 143, 50, 143, 51, 143, 52, 143, 53, 143, 54, 143, 55, 143, 56, 143, 57, 143, 58, 143, 59, 143, 60, 143, 96, 216, 101, 223, 153, 143, 154, 143, 165, 143, 166, 143, 167, 143, 168, 143, 169, 143, 170, 143, 114, 144, 116, 144, 117, 144, 118, 144, 120, 144, 121, 144, 122, 144, 123, 144, 124, 144, 134, 144, 173, 72, 97, 216, 232, 221, 97, 216, 244, 221, 97, 216, 0, 222, 250, 144, 51, 145, 52, 145, 53, 145, 60, 0, 42, 0, 54, 145, 55, 145, 141, 145, 142, 145, 143, 145, 144, 145, 145, 145, 146, 145, 147, 145, 148, 145, 149, 145, 150, 145, 151, 145, 0, 73, 6, 73, 203, 146, 247, 146, 248, 146, 249, 146, 250, 146, 251, 146, 252, 146, 253, 146, 254, 146, 255, 146, 0, 147, 1, 147, 2, 147, 3, 147, 4, 147, 5, 147, 6, 147, 7, 147, 8, 147, 9, 147, 10, 147, 11, 147, 12, 147, 13, 147, 14, 147, 15, 147, 16, 147, 17, 147, 18, 147, 19, 147, 20, 147, 21, 147, 22, 147, 23, 147, 24, 147, 25, 147, 26, 147, 27, 147, 28, 147, 29, 147, 30, 147, 31, 147, 32, 147, 33, 147, 34, 147, 35, 147, 36, 147, 37, 147, 38, 147, 39, 147, 41, 147, 42, 147, 43, 147, 44, 147, 45, 147, 46, 147, 47, 147, 48, 147, 49, 147, 50, 147, 51, 147, 52, 147, 53, 147, 54, 147, 55, 147, 56, 147, 57, 147, 58, 147, 59, 147, 60, 147, 61, 147, 62, 147, 63, 147, 64, 147, 65, 147, 66, 147, 67, 147, 68, 147, 69, 147, 70, 147, 72, 147, 172, 159, 40, 250, 53, 73, 98, 216, 231, 221, 98, 216, 232, 221, 98, 216, 249, 221, 98, 216, 250, 221, 98, 216, 251, 221, 98, 216, 252, 221, 98, 216, 15, 222, 98, 216, 22, 222, 98, 216, 37, 222, 98, 216, 41, 222, 122, 147, 86, 149, 87, 149, 88, 149, 89, 149, 90, 149, 91, 149, 92, 149, 93, 149, 94, 149, 95, 149, 96, 149, 181, 149, 182, 149, 184, 149, 185, 149, 186, 149, 187, 149, 188, 149, 60, 0, 42, 0, 189, 149, 190, 149, 191, 149, 193, 149, 194, 149, 161, 73, 167, 73, 205, 149, 27, 150, 166, 150, 167, 150, 168, 150, 169, 150, 170, 150, 171, 150, 231, 73, 183, 150, 212, 150, 213, 150, 11, 151, 12, 151, 13, 151, 14, 151, 15, 151, 16, 151, 17, 151, 18, 151, 19, 151, 20, 151, 21, 151, 22, 151, 23, 151, 29, 74, 100, 216, 228, 220, 100, 216, 229, 220, 100, 216, 236, 220, 100, 216, 237, 220, 91, 151, 92, 151, 102, 151, 147, 151, 148, 151, 149, 151, 150, 151, 151, 151, 152, 151, 153, 151, 210, 151, 240, 151, 248, 151, 36, 152, 37, 152, 45, 152, 46, 152, 47, 152, 48, 152, 49, 152, 50, 152, 51, 152, 52, 152, 53, 152, 54, 152, 55, 152, 56, 152, 57, 152, 58, 152, 59, 152, 60, 152, 61, 152, 101, 216, 208, 220, 101, 216, 217, 220, 101, 216, 218, 220, 158, 152, 159, 152, 160, 152, 161, 152, 180, 152, 181, 152, 101, 216, 215, 221, 217, 152, 218, 152, 16, 153, 29, 153, 26, 153, 27, 153, 28, 153, 30, 153, 31, 153, 32, 153, 33, 153, 34, 153, 35, 153, 36, 153, 38, 153, 39, 153, 40, 153, 41, 153, 101, 216, 32, 223, 52, 153, 158, 153, 159, 153, 160, 153, 225, 153, 226, 153, 227, 153, 228, 153, 229, 153, 230, 153, 231, 153, 232, 153, 233, 153, 234, 153, 235, 153, 236, 153, 237, 153, 238, 153, 239, 153, 240, 153, 241, 153, 242, 153, 102, 216, 209, 220, 184, 154, 185, 154, 186, 154, 187, 154, 188, 154, 102, 216, 77, 222, 191, 154, 60, 0, 42, 0, 237, 154, 245, 154, 246, 154, 247, 154, 248, 154, 249, 154, 250, 154, 251, 154, 7, 155, 40, 155, 51, 155, 71, 155, 122, 155, 123, 155, 124, 155, 125, 155, 126, 155, 127, 155, 128, 155, 129, 155, 130, 155, 131, 155, 132, 155, 133, 155, 135, 155, 136, 155, 137, 155, 138, 155, 139, 155, 140, 155, 141, 155, 142, 155, 143, 155, 144, 155, 145, 155, 146, 155, 147, 155, 148, 155, 149, 155, 150, 155, 151, 155, 152, 155, 163, 155, 64, 76, 71, 76, 73, 76, 103, 216, 152, 221, 103, 216, 155, 221, 173, 156, 174, 156, 175, 156, 176, 156, 177, 156, 178, 156, 179, 156, 180, 156, 181, 156, 182, 156, 183, 156, 184, 156, 185, 156, 186, 156, 187, 156, 10, 157, 15, 157, 16, 157, 17, 157, 18, 157, 19, 157, 20, 157, 21, 157, 22, 157, 23, 157, 24, 157, 25, 157, 26, 157, 27, 157, 29, 157, 30, 157, 31, 157, 32, 157, 33, 157, 34, 157, 35, 157, 36, 157, 37, 157, 38, 157, 39, 157, 40, 157, 41, 157, 42, 157, 43, 157, 44, 157, 103, 216, 222, 223, 101, 158, 102, 158, 103, 158, 104, 158, 119, 158, 126, 158, 133, 158, 134, 158, 135, 158, 136, 158, 104, 216, 159, 222, 172, 158, 173, 158, 174, 158, 52, 77, 186, 158, 197, 158, 198, 158, 212, 158, 213, 158, 214, 158, 215, 158, 216, 158, 250, 158, 18, 159, 60, 159, 61, 159, 83, 159, 141, 159, 156, 159, 60, 0, 39, 0, 208, 253, 17, 40, 39, 0, 60, 0, 42, 0, 31, 81, 32, 81, 33, 81, 34, 81, 35, 81, 36, 81, 37, 81, 38, 81, 39, 81, 40, 81, 41, 81, 42, 81, 44, 81, 65, 216, 19, 220, 65, 216, 20, 220, 65, 216, 31, 220, 50, 81, 223, 81, 65, 216, 25, 222, 65, 216, 26, 222, 149, 82, 244, 82, 245, 82, 246, 82, 37, 53, 38, 53, 53, 83, 179, 83, 83, 53, 66, 216, 143, 223, 125, 86, 126, 86, 127, 86, 128, 86, 129, 86, 130, 86, 131, 86, 132, 86, 133, 86, 134, 86, 135, 86, 136, 86, 137, 86, 138, 86, 139, 86, 140, 86, 141, 86, 142, 86, 143, 86, 144, 86, 145, 86, 146, 86, 147, 86, 68, 216, 79, 220, 68, 216, 92, 220, 68, 216, 111, 220, 68, 216, 117, 220, 68, 216, 118, 220, 68, 216, 119, 220, 68, 216, 120, 220, 68, 216, 123, 220, 68, 216, 136, 220, 205, 88, 206, 88, 207, 88, 208, 88, 209, 88, 210, 88, 211, 88, 212, 88, 213, 88, 214, 88, 215, 88, 122, 54, 69, 216, 138, 220, 69, 216, 151, 220, 69, 216, 146, 222, 35, 91, 36, 91, 37, 91, 38, 91, 39, 91, 41, 91, 42, 91, 43, 91, 44, 91, 45, 91, 46, 91, 47, 91, 48, 91, 49, 91, 50, 91, 51, 91, 53, 91, 54, 91, 55, 91, 70, 216, 209, 220, 70, 216, 214, 220, 70, 216, 215, 220, 70, 216, 216, 220, 70, 216, 217, 220, 122, 91, 123, 91, 241, 91, 242, 91, 54, 92, 55, 92, 104, 92, 183, 93, 184, 93, 185, 93, 186, 93, 188, 93, 189, 93, 190, 93, 191, 93, 25, 56, 71, 216, 106, 223, 187, 93, 106, 94, 107, 94, 108, 94, 101, 56, 102, 56, 76, 95, 189, 95, 190, 95, 73, 216, 141, 221, 60, 0, 42, 0, 181, 97, 188, 97, 194, 97, 195, 97, 199, 97, 201, 97, 203, 97, 209, 97, 215, 97, 217, 97, 218, 97, 219, 97, 220, 97, 221, 97, 222, 97, 224, 97, 225, 97, 226, 97, 228, 97, 229, 97, 230, 97, 231, 97, 232, 97, 74, 216, 247, 220, 50, 98, 52, 98, 202, 100, 206, 100, 216, 100, 223, 100, 224, 100, 225, 100, 226, 100, 227, 100, 228, 100, 230, 100, 232, 100, 233, 100, 235, 100, 236, 100, 237, 100, 238, 100, 239, 100, 240, 100, 241, 100, 92, 58, 94, 58, 75, 216, 238, 221, 128, 101, 129, 101, 130, 101, 131, 101, 163, 101, 181, 101, 182, 101, 218, 101, 206, 102, 208, 102, 209, 102, 210, 102, 211, 102, 212, 102, 213, 102, 214, 102, 215, 102, 218, 102, 32, 59, 33, 59, 34, 59, 76, 216, 210, 222, 217, 102, 50, 59, 64, 59, 125, 106, 126, 106, 127, 106, 128, 106, 129, 106, 130, 106, 131, 106, 132, 106, 133, 106, 134, 106, 135, 106, 136, 106, 137, 106, 138, 106, 139, 106, 140, 106, 141, 106, 142, 106, 143, 106, 144, 106, 145, 106, 146, 106, 147, 106, 148, 106, 149, 106, 150, 106, 151, 106, 152, 106, 153, 106, 154, 106, 155, 106, 156, 106, 157, 106, 158, 106, 159, 106, 160, 106, 161, 106, 162, 106, 163, 106, 164, 106, 165, 106, 166, 106, 167, 106, 168, 106, 169, 106, 170, 106, 236, 59, 242, 59, 243, 59, 244, 59, 77, 216, 3, 223, 77, 216, 22, 223, 77, 216, 32, 223, 77, 216, 45, 223, 77, 216, 47, 223, 60, 0, 42, 0, 77, 216, 63, 223, 219, 106, 91, 107, 92, 107, 93, 107, 70, 60, 172, 107, 173, 107, 174, 107, 218, 107, 8, 108, 9, 108, 10, 108, 169, 111, 204, 111, 128, 111, 212, 111, 213, 111, 215, 111, 216, 111, 217, 111, 218, 111, 219, 111, 220, 111, 221, 111, 222, 111, 223, 111, 224, 111, 225, 111, 226, 111, 227, 111, 228, 111, 229, 111, 230, 111, 231, 111, 232, 111, 233, 111, 234, 111, 235, 111, 236, 111, 237, 111, 238, 111, 239, 111, 240, 111, 241, 111, 242, 111, 244, 111, 245, 111, 246, 111, 248, 111, 111, 61, 115, 61, 117, 61, 79, 216, 192, 223, 80, 216, 57, 220, 80, 216, 58, 220, 80, 216, 59, 220, 80, 216, 60, 220, 80, 216, 61, 220, 80, 216, 87, 220, 223, 113, 224, 113, 225, 113, 226, 113, 227, 113, 228, 113, 229, 113, 230, 113, 231, 113, 232, 113, 233, 113, 234, 113, 235, 113, 236, 113, 237, 113, 238, 113, 239, 113, 240, 113, 241, 113, 242, 113, 243, 113, 244, 113, 245, 113, 246, 113, 247, 113, 0, 62, 1, 62, 2, 62, 80, 216, 248, 223, 81, 216, 4, 220, 53, 114, 70, 114, 160, 114, 85, 62, 110, 115, 111, 115, 112, 115, 113, 115, 114, 115, 115, 115, 116, 115, 144, 116, 151, 116, 250, 62, 165, 116, 166, 116, 168, 116, 169, 116, 170, 116, 171, 116, 172, 116, 173, 116, 174, 116, 175, 116, 176, 116, 177, 116, 178, 116, 179, 116, 180, 116, 255, 62, 0, 63, 1, 63, 82, 216, 113, 222, 82, 216, 119, 222, 82, 216, 120, 222, 82, 216, 121, 222, 82, 216, 122, 222, 15, 117, 60, 0, 42, 0, 16, 117, 17, 117, 18, 117, 63, 63, 131, 117, 132, 117, 83, 216, 217, 220, 64, 118, 65, 118, 66, 118, 67, 118, 68, 118, 69, 118, 70, 118, 71, 118, 72, 118, 73, 118, 75, 118, 76, 118, 77, 118, 78, 118, 83, 216, 165, 222, 83, 216, 167, 222, 162, 118, 163, 118, 164, 118, 165, 118, 188, 118, 232, 118, 233, 118, 234, 118, 9, 64, 164, 119, 166, 119, 167, 119, 168, 119, 169, 119, 170, 119, 171, 119, 172, 119, 173, 119, 174, 119, 175, 119, 176, 119, 177, 119, 178, 119, 179, 119, 180, 119, 181, 119, 182, 119, 183, 119, 111, 64, 113, 64, 84, 216, 199, 222, 239, 119, 240, 119, 239, 120, 240, 120, 241, 120, 242, 120, 243, 120, 244, 120, 245, 120, 246, 120, 247, 120, 248, 120, 249, 120, 250, 120, 251, 120, 252, 120, 253, 120, 254, 120, 255, 120, 0, 121, 1, 121, 2, 121, 3, 121, 4, 121, 5, 121, 223, 64, 85, 216, 129, 221, 85, 216, 132, 221, 85, 216, 143, 221, 167, 121, 168, 121, 170, 121, 171, 121, 85, 216, 6, 223, 73, 122, 84, 122, 85, 122, 86, 122, 87, 122, 88, 122, 89, 122, 90, 122, 91, 122, 92, 122, 93, 122, 94, 122, 127, 65, 190, 122, 191, 122, 192, 122, 193, 122, 194, 122, 185, 65, 242, 122, 243, 122, 244, 122, 86, 216, 175, 222, 21, 124, 240, 123, 241, 123, 242, 123, 243, 123, 244, 123, 245, 123, 246, 123, 247, 123, 248, 123, 250, 123, 251, 123, 252, 123, 253, 123, 254, 123, 255, 123, 0, 124, 60, 0, 42, 0, 1, 124, 2, 124, 3, 124, 4, 124, 5, 124, 6, 124, 7, 124, 8, 124, 9, 124, 10, 124, 11, 124, 12, 124, 13, 124, 14, 124, 15, 124, 16, 124, 18, 124, 19, 124, 20, 124, 22, 124, 23, 124, 59, 66, 64, 66, 65, 66, 66, 66, 67, 66, 68, 66, 69, 66, 87, 216, 145, 220, 87, 216, 164, 220, 87, 216, 192, 220, 24, 124, 217, 124, 218, 124, 219, 124, 220, 124, 221, 124, 222, 124, 223, 124, 224, 124, 225, 124, 226, 124, 232, 124, 162, 66, 41, 126, 42, 126, 43, 126, 44, 126, 45, 126, 46, 126, 47, 126, 48, 126, 49, 126, 50, 126, 51, 126, 52, 126, 53, 126, 54, 126, 55, 126, 56, 126, 57, 126, 58, 126, 59, 126, 60, 126, 61, 126, 62, 126, 63, 126, 64, 126, 65, 126, 66, 126, 67, 126, 68, 126, 69, 126, 70, 126, 71, 126, 73, 126, 74, 126, 76, 126, 77, 126, 88, 216, 2, 221, 88, 216, 33, 221, 72, 126, 68, 127, 69, 127, 70, 127, 125, 127, 126, 127, 127, 127, 129, 127, 88, 216, 190, 223, 242, 127, 243, 127, 244, 127, 245, 127, 246, 127, 252, 127, 89, 216, 132, 220, 89, 216, 136, 220, 89, 216, 137, 220, 43, 128, 44, 128, 111, 128, 112, 128, 113, 128, 114, 128, 115, 128, 116, 128, 89, 216, 38, 222, 165, 129, 184, 129, 186, 129, 187, 129, 188, 129, 189, 129, 190, 129, 191, 129, 192, 129, 193, 129, 194, 129, 195, 129, 196, 129, 197, 129, 198, 129, 199, 129, 200, 129, 201, 129, 202, 129, 204, 129, 60, 0, 42, 0, 90, 216, 70, 220, 90, 216, 94, 220, 232, 129, 233, 129, 90, 216, 242, 221, 90, 130, 91, 130, 92, 130, 93, 130, 113, 130, 22, 69, 27, 69, 29, 69, 32, 69, 38, 69, 39, 69, 87, 133, 118, 133, 119, 133, 120, 133, 121, 133, 122, 133, 123, 133, 124, 133, 125, 133, 126, 133, 127, 133, 128, 133, 129, 133, 130, 133, 131, 133, 132, 133, 133, 133, 134, 133, 135, 133, 136, 133, 137, 133, 138, 133, 139, 133, 141, 133, 142, 133, 143, 133, 144, 133, 145, 133, 146, 133, 147, 133, 148, 133, 149, 133, 150, 133, 151, 133, 152, 133, 153, 133, 154, 133, 155, 133, 156, 133, 157, 133, 158, 133, 159, 133, 160, 133, 161, 133, 162, 133, 163, 133, 164, 133, 165, 133, 166, 133, 167, 133, 168, 133, 170, 133, 171, 133, 172, 133, 174, 133, 169, 159, 91, 216, 115, 223, 91, 216, 116, 223, 91, 216, 159, 223, 91, 216, 161, 223, 173, 133, 175, 133, 103, 134, 104, 134, 170, 135, 171, 135, 172, 135, 173, 135, 174, 135, 175, 135, 176, 135, 177, 135, 178, 135, 179, 135, 180, 135, 181, 135, 182, 135, 183, 135, 184, 135, 185, 135, 186, 135, 187, 135, 188, 135, 189, 135, 190, 135, 191, 135, 192, 135, 193, 135, 194, 135, 195, 135, 196, 135, 197, 135, 198, 135, 199, 135, 200, 135, 201, 135, 202, 135, 203, 135, 204, 135, 205, 135, 206, 135, 207, 135, 208, 135, 209, 135, 210, 135, 233, 69, 234, 69, 238, 69, 92, 216, 255, 223, 93, 216, 34, 220, 222, 135, 51, 137, 60, 0, 42, 0, 53, 137, 54, 137, 55, 137, 56, 137, 57, 137, 58, 137, 59, 137, 60, 137, 61, 137, 62, 137, 63, 137, 64, 137, 65, 137, 66, 137, 67, 137, 68, 137, 69, 137, 84, 137, 91, 70, 82, 137, 203, 159, 93, 216, 94, 223, 171, 137, 172, 137, 173, 137, 174, 137, 175, 137, 242, 137, 243, 137, 94, 216, 36, 221, 4, 139, 5, 139, 6, 139, 7, 139, 8, 139, 9, 139, 10, 139, 11, 139, 12, 139, 13, 139, 14, 139, 15, 139, 16, 139, 17, 139, 18, 139, 19, 139, 21, 139, 22, 139, 23, 139, 24, 139, 25, 139, 26, 139, 27, 139, 28, 139, 29, 139, 30, 139, 31, 139, 32, 139, 33, 139, 34, 139, 0, 71, 94, 216, 189, 222, 94, 216, 190, 222, 38, 71, 63, 140, 64, 140, 65, 140, 79, 140, 111, 140, 112, 140, 113, 140, 114, 140, 115, 140, 148, 140, 149, 140, 150, 140, 246, 140, 247, 140, 248, 140, 249, 140, 250, 140, 251, 140, 252, 140, 253, 140, 95, 216, 220, 221, 97, 141, 98, 141, 111, 141, 168, 141, 70, 142, 71, 142, 72, 142, 73, 142, 74, 142, 75, 142, 76, 142, 77, 142, 78, 142, 79, 142, 80, 142, 81, 142, 82, 142, 83, 142, 96, 216, 232, 220, 96, 216, 233, 220, 96, 216, 244, 220, 61, 143, 62, 143, 63, 143, 64, 143, 65, 143, 66, 143, 67, 143, 68, 143, 69, 143, 96, 216, 109, 223, 96, 216, 125, 223, 171, 143, 125, 144, 126, 144, 127, 144, 128, 144, 129, 144, 130, 144, 131, 144, 132, 144, 133, 144, 137, 144, 60, 0, 42, 0, 174, 72, 97, 216, 11, 222, 56, 145, 57, 145, 152, 145, 153, 145, 154, 145, 155, 145, 156, 145, 157, 145, 158, 145, 159, 145, 160, 145, 161, 145, 162, 145, 163, 145, 164, 145, 98, 216, 51, 221, 40, 147, 71, 147, 73, 147, 74, 147, 75, 147, 76, 147, 77, 147, 78, 147, 79, 147, 80, 147, 81, 147, 82, 147, 83, 147, 84, 147, 85, 147, 86, 147, 87, 147, 88, 147, 89, 147, 90, 147, 91, 147, 92, 147, 93, 147, 94, 147, 95, 147, 96, 147, 97, 147, 98, 147, 99, 147, 100, 147, 101, 147, 102, 147, 103, 147, 104, 147, 105, 147, 106, 147, 107, 147, 108, 147, 109, 147, 110, 147, 111, 147, 112, 147, 113, 147, 114, 147, 115, 147, 116, 147, 117, 147, 118, 147, 119, 147, 120, 147, 121, 147, 123, 147, 124, 147, 125, 147, 126, 147, 127, 147, 128, 147, 129, 147, 130, 147, 131, 147, 132, 147, 133, 147, 134, 147, 135, 147, 60, 73, 98, 216, 50, 222, 98, 216, 54, 222, 98, 216, 68, 222, 98, 216, 69, 222, 98, 216, 70, 222, 98, 216, 71, 222, 98, 216, 72, 222, 98, 216, 73, 222, 98, 216, 74, 222, 98, 216, 75, 222, 98, 216, 89, 222, 98, 216, 90, 222, 98, 216, 129, 222, 98, 216, 130, 222, 98, 216, 131, 222, 161, 147, 175, 147, 97, 149, 98, 149, 99, 149, 100, 149, 101, 149, 102, 149, 103, 149, 104, 149, 105, 149, 106, 149, 107, 149, 192, 149, 183, 149, 195, 149, 196, 149, 197, 149, 198, 149, 199, 149, 200, 149, 201, 149, 202, 149, 203, 149, 204, 149, 206, 149, 207, 149, 172, 150, 173, 150, 174, 150, 60, 0, 42, 0, 175, 150, 176, 150, 177, 150, 178, 150, 184, 150, 214, 150, 1, 74, 2, 74, 24, 151, 25, 151, 26, 151, 27, 151, 28, 151, 29, 151, 30, 151, 31, 151, 32, 151, 36, 74, 100, 216, 13, 221, 100, 216, 16, 221, 33, 151, 74, 74, 154, 151, 155, 151, 156, 151, 157, 151, 158, 151, 159, 151, 160, 151, 161, 151, 211, 151, 212, 151, 213, 151, 241, 151, 64, 152, 65, 152, 66, 152, 67, 152, 68, 152, 69, 152, 70, 152, 71, 152, 72, 152, 73, 152, 74, 152, 209, 74, 101, 216, 229, 220, 101, 216, 231, 220, 182, 152, 183, 152, 16, 75, 101, 216, 233, 221, 101, 216, 244, 221, 32, 75, 37, 153, 42, 153, 43, 153, 44, 153, 45, 153, 47, 153, 48, 153, 49, 153, 50, 153, 51, 153, 53, 153, 55, 153, 101, 216, 50, 223, 66, 153, 70, 153, 152, 153, 161, 153, 162, 153, 163, 153, 112, 75, 114, 75, 244, 153, 245, 153, 246, 153, 247, 153, 248, 153, 249, 153, 250, 153, 251, 153, 252, 153, 253, 153, 254, 153, 255, 153, 0, 154, 1, 154, 2, 154, 3, 154, 142, 75, 144, 75, 102, 216, 234, 220, 243, 153, 164, 154, 189, 154, 190, 154, 252, 154, 253, 154, 254, 154, 255, 154, 0, 155, 1, 155, 2, 155, 52, 155, 72, 155, 73, 155, 134, 155, 153, 155, 154, 155, 155, 155, 156, 155, 157, 155, 158, 155, 159, 155, 160, 155, 161, 155, 162, 155, 164, 155, 165, 155, 166, 155, 167, 155, 168, 155, 169, 155, 170, 155, 171, 155, 172, 155, 60, 0, 42, 0, 173, 155, 174, 155, 175, 155, 176, 155, 177, 155, 178, 155, 179, 155, 180, 155, 186, 155, 75, 76, 76, 76, 77, 76, 206, 155, 188, 156, 189, 156, 191, 156, 192, 156, 193, 156, 194, 156, 195, 156, 196, 156, 197, 156, 198, 156, 199, 156, 200, 156, 201, 156, 202, 156, 203, 156, 190, 156, 28, 157, 45, 157, 46, 157, 47, 157, 48, 157, 49, 157, 50, 157, 51, 157, 52, 157, 53, 157, 54, 157, 55, 157, 56, 157, 57, 157, 58, 157, 59, 157, 60, 157, 61, 157, 62, 157, 63, 157, 64, 157, 65, 157, 66, 157, 67, 157, 68, 157, 69, 157, 70, 157, 71, 157, 72, 157, 73, 157, 205, 76, 212, 76, 104, 216, 20, 220, 103, 157, 105, 158, 106, 158, 107, 158, 108, 158, 137, 158, 138, 158, 139, 158, 104, 216, 178, 222, 175, 158, 176, 158, 199, 158, 200, 158, 201, 158, 207, 158, 218, 158, 219, 158, 220, 158, 221, 158, 222, 158, 105, 216, 52, 220, 251, 158, 255, 158, 34, 159, 35, 159, 36, 159, 62, 159, 63, 159, 75, 159, 105, 216, 198, 221, 84, 159, 98, 159, 139, 159, 140, 159, 160, 159, 60, 0, 39, 0, 208, 253, 18, 40, 39, 0, 60, 0, 42, 0, 45, 81, 46, 81, 47, 81, 49, 81, 175, 52, 65, 216, 254, 220, 129, 81, 65, 216, 165, 221, 66, 216, 194, 222, 226, 83, 148, 86, 149, 86, 150, 86, 151, 86, 152, 86, 153, 86, 154, 86, 155, 86, 156, 86, 157, 86, 158, 86, 159, 86, 160, 86, 161, 86, 162, 86, 163, 86, 164, 86, 9, 54, 68, 216, 150, 220, 68, 216, 157, 220, 68, 216, 180, 220, 68, 216, 191, 220, 68, 216, 192, 220, 68, 216, 193, 220, 68, 216, 199, 220, 68, 216, 200, 220, 68, 216, 201, 220, 68, 216, 207, 220, 68, 216, 211, 220, 174, 86, 216, 88, 217, 88, 69, 216, 182, 220, 17, 89, 19, 89, 112, 89, 56, 91, 58, 91, 59, 91, 60, 91, 48, 55, 49, 55, 70, 216, 250, 220, 105, 92, 106, 92, 192, 93, 193, 93, 194, 93, 109, 94, 110, 94, 235, 94, 77, 95, 93, 95, 94, 95, 213, 97, 214, 97, 216, 97, 223, 97, 227, 97, 155, 57, 233, 97, 234, 97, 235, 97, 237, 97, 238, 97, 240, 97, 241, 97, 243, 97, 244, 97, 161, 57, 51, 98, 231, 100, 234, 100, 229, 100, 242, 100, 244, 100, 245, 100, 246, 100, 247, 100, 248, 100, 249, 100, 250, 100, 251, 100, 252, 100, 253, 100, 254, 100, 255, 100, 1, 101, 2, 101, 3, 101, 4, 101, 5, 101, 6, 101, 97, 58, 102, 58, 103, 58, 75, 216, 13, 222, 1, 141, 171, 58, 76, 216, 102, 220, 148, 101, 76, 216, 142, 220, 183, 101, 219, 101, 216, 102, 219, 102, 220, 102, 35, 59, 36, 59, 37, 59, 38, 59, 39, 59, 40, 59, 76, 216, 224, 222, 76, 216, 225, 222, 38, 103, 171, 106, 172, 106, 60, 0, 42, 0, 173, 106, 174, 106, 175, 106, 176, 106, 177, 106, 178, 106, 179, 106, 180, 106, 181, 106, 182, 106, 183, 106, 184, 106, 185, 106, 186, 106, 187, 106, 188, 106, 189, 106, 190, 106, 191, 106, 192, 106, 193, 106, 194, 106, 195, 106, 196, 106, 197, 106, 198, 106, 199, 106, 200, 106, 201, 106, 202, 106, 77, 216, 102, 223, 77, 216, 129, 223, 225, 106, 237, 106, 94, 107, 95, 107, 120, 107, 175, 107, 201, 107, 11, 108, 247, 111, 249, 111, 250, 111, 251, 111, 252, 111, 253, 111, 254, 111, 255, 111, 0, 112, 1, 112, 2, 112, 3, 112, 5, 112, 6, 112, 7, 112, 8, 112, 9, 112, 10, 112, 11, 112, 12, 112, 13, 112, 14, 112, 15, 112, 16, 112, 17, 112, 18, 112, 19, 112, 20, 112, 125, 61, 80, 216, 133, 220, 80, 216, 139, 220, 80, 216, 140, 220, 80, 216, 141, 220, 80, 216, 145, 220, 38, 112, 248, 113, 249, 113, 250, 113, 251, 113, 252, 113, 253, 113, 254, 113, 255, 113, 0, 114, 1, 114, 3, 114, 4, 62, 81, 216, 53, 220, 81, 216, 54, 220, 90, 216, 168, 221, 117, 115, 118, 115, 119, 115, 167, 116, 181, 116, 182, 116, 184, 116, 185, 116, 187, 116, 188, 116, 190, 116, 191, 116, 192, 116, 193, 116, 194, 116, 82, 216, 140, 222, 82, 216, 147, 222, 82, 216, 148, 222, 82, 216, 149, 222, 82, 216, 150, 222, 82, 216, 164, 222, 82, 216, 165, 222, 82, 216, 166, 222, 82, 216, 167, 222, 19, 117, 20, 117, 21, 117, 133, 117, 79, 118, 80, 118, 81, 118, 82, 118, 83, 118, 84, 118, 85, 118, 86, 118, 87, 118, 88, 118, 60, 0, 42, 0, 89, 118, 90, 118, 91, 118, 92, 118, 93, 118, 94, 118, 100, 118, 166, 118, 167, 118, 168, 118, 83, 216, 169, 223, 189, 118, 249, 63, 235, 118, 236, 118, 184, 119, 185, 119, 186, 119, 187, 119, 188, 119, 189, 119, 190, 119, 191, 119, 192, 119, 193, 119, 194, 119, 84, 216, 216, 222, 6, 121, 7, 121, 8, 121, 9, 121, 10, 121, 11, 121, 12, 121, 13, 121, 14, 121, 15, 121, 16, 121, 17, 121, 18, 121, 19, 121, 20, 121, 21, 121, 22, 121, 172, 121, 173, 121, 174, 121, 175, 121, 32, 65, 95, 122, 96, 122, 97, 122, 98, 122, 99, 122, 129, 65, 86, 216, 200, 220, 196, 122, 197, 122, 245, 122, 25, 124, 26, 124, 27, 124, 28, 124, 29, 124, 30, 124, 31, 124, 32, 124, 33, 124, 34, 124, 35, 124, 36, 124, 37, 124, 38, 124, 39, 124, 40, 124, 41, 124, 42, 124, 43, 124, 45, 124, 46, 124, 47, 124, 48, 124, 49, 124, 50, 124, 78, 66, 85, 66, 87, 216, 193, 220, 87, 216, 254, 220, 87, 216, 32, 221, 227, 124, 228, 124, 229, 124, 230, 124, 231, 124, 166, 66, 87, 216, 26, 223, 78, 126, 79, 126, 80, 126, 81, 126, 82, 126, 83, 126, 84, 126, 85, 126, 86, 126, 87, 126, 88, 126, 89, 126, 90, 126, 91, 126, 92, 126, 93, 126, 94, 126, 95, 126, 96, 126, 97, 126, 98, 126, 99, 126, 100, 126, 101, 126, 103, 126, 113, 126, 24, 67, 88, 216, 89, 221, 88, 216, 90, 221, 88, 216, 91, 221, 88, 216, 92, 221, 71, 127, 60, 0, 42, 0, 72, 127, 73, 127, 88, 216, 88, 222, 128, 127, 130, 127, 179, 127, 180, 127, 181, 127, 247, 127, 248, 127, 249, 127, 250, 127, 251, 127, 151, 67, 89, 216, 141, 220, 89, 216, 152, 220, 45, 128, 46, 128, 117, 128, 118, 128, 119, 128, 205, 129, 206, 129, 207, 129, 208, 129, 209, 129, 210, 129, 211, 129, 67, 68, 68, 68, 69, 68, 90, 216, 136, 220, 10, 130, 25, 130, 94, 130, 95, 130, 96, 130, 143, 68, 46, 69, 51, 69, 54, 69, 59, 69, 61, 69, 63, 69, 64, 69, 67, 69, 68, 69, 169, 133, 176, 133, 177, 133, 178, 133, 179, 133, 180, 133, 181, 133, 182, 133, 183, 133, 184, 133, 185, 133, 186, 133, 187, 133, 188, 133, 189, 133, 190, 133, 191, 133, 192, 133, 193, 133, 194, 133, 195, 133, 196, 133, 197, 133, 198, 133, 199, 133, 200, 133, 201, 133, 202, 133, 203, 133, 204, 133, 205, 133, 206, 133, 207, 133, 208, 133, 210, 133, 211, 133, 91, 216, 190, 223, 91, 216, 222, 223, 91, 216, 223, 223, 92, 216, 14, 220, 105, 134, 155, 69, 215, 135, 211, 135, 212, 135, 214, 135, 216, 135, 217, 135, 218, 135, 219, 135, 220, 135, 221, 135, 223, 135, 224, 135, 226, 135, 227, 135, 228, 135, 229, 135, 230, 135, 231, 135, 232, 135, 233, 135, 234, 135, 235, 135, 236, 135, 237, 135, 238, 135, 239, 135, 240, 135, 241, 135, 242, 135, 243, 135, 244, 135, 245, 135, 14, 136, 93, 216, 80, 220, 93, 216, 132, 220, 93, 216, 134, 220, 70, 137, 71, 137, 72, 137, 60, 0, 42, 0, 73, 137, 74, 137, 75, 137, 76, 137, 77, 137, 78, 137, 79, 137, 80, 137, 81, 137, 83, 137, 85, 137, 93, 216, 132, 223, 93, 216, 133, 223, 134, 137, 176, 137, 177, 137, 178, 137, 179, 137, 141, 70, 244, 137, 100, 157, 35, 139, 36, 139, 37, 139, 38, 139, 39, 139, 40, 139, 41, 139, 42, 139, 43, 139, 44, 139, 45, 139, 46, 139, 47, 139, 48, 139, 49, 139, 50, 139, 51, 139, 52, 139, 53, 139, 54, 139, 55, 139, 56, 139, 57, 139, 58, 139, 59, 139, 60, 139, 61, 139, 62, 139, 94, 216, 244, 222, 71, 139, 66, 140, 80, 140, 116, 140, 117, 140, 79, 71, 151, 140, 152, 140, 153, 140, 254, 140, 255, 140, 0, 141, 2, 141, 3, 141, 4, 141, 5, 141, 169, 141, 84, 142, 85, 142, 86, 142, 87, 142, 88, 142, 89, 142, 90, 142, 91, 142, 92, 142, 93, 142, 94, 142, 95, 142, 96, 142, 97, 142, 98, 142, 99, 142, 100, 142, 101, 142, 102, 142, 103, 142, 110, 142, 128, 142, 32, 72, 96, 216, 46, 221, 105, 142, 191, 142, 192, 142, 193, 142, 96, 216, 124, 222, 70, 143, 71, 143, 72, 143, 73, 143, 74, 143, 75, 143, 76, 143, 113, 72, 96, 216, 138, 223, 172, 143, 135, 144, 136, 144, 97, 216, 37, 222, 40, 145, 58, 145, 59, 145, 60, 145, 61, 145, 62, 145, 165, 145, 166, 145, 167, 145, 168, 145, 169, 145, 170, 145, 171, 145, 172, 145, 208, 145, 136, 147, 137, 147, 138, 147, 139, 147, 60, 0, 42, 0, 140, 147, 141, 147, 142, 147, 143, 147, 144, 147, 145, 147, 146, 147, 147, 147, 148, 147, 149, 147, 150, 147, 151, 147, 152, 147, 153, 147, 154, 147, 155, 147, 156, 147, 157, 147, 158, 147, 159, 147, 160, 147, 162, 147, 163, 147, 164, 147, 165, 147, 166, 147, 167, 147, 168, 147, 170, 147, 171, 147, 172, 147, 173, 147, 174, 147, 176, 147, 177, 147, 178, 147, 179, 147, 180, 147, 181, 147, 182, 147, 183, 147, 184, 147, 185, 147, 186, 147, 187, 147, 188, 147, 189, 147, 190, 147, 191, 147, 178, 159, 62, 73, 68, 73, 69, 73, 71, 73, 98, 216, 154, 222, 98, 216, 155, 222, 98, 216, 156, 222, 98, 216, 192, 222, 98, 216, 198, 222, 98, 216, 203, 222, 98, 216, 204, 222, 98, 216, 206, 222, 108, 149, 109, 149, 110, 149, 111, 149, 112, 149, 113, 149, 208, 149, 209, 149, 210, 149, 211, 149, 212, 149, 213, 149, 214, 149, 215, 149, 216, 149, 99, 216, 153, 221, 179, 150, 215, 150, 216, 150, 217, 150, 218, 150, 219, 150, 220, 150, 221, 150, 222, 150, 223, 150, 224, 150, 3, 74, 226, 150, 34, 151, 35, 151, 36, 151, 37, 151, 38, 74, 93, 151, 162, 151, 163, 151, 164, 151, 165, 151, 166, 151, 167, 151, 168, 151, 169, 151, 170, 151, 171, 151, 172, 151, 173, 151, 174, 151, 175, 151, 176, 151, 214, 151, 215, 151, 216, 151, 217, 151, 218, 151, 150, 74, 249, 151, 250, 151, 101, 216, 51, 220, 63, 152, 62, 152, 75, 152, 76, 152, 77, 152, 78, 152, 79, 152, 60, 0, 42, 0, 80, 152, 81, 152, 82, 152, 83, 152, 84, 152, 85, 152, 162, 152, 163, 152, 184, 152, 185, 152, 186, 152, 46, 153, 73, 75, 75, 75, 76, 75, 54, 153, 56, 153, 57, 153, 58, 153, 59, 153, 60, 153, 61, 153, 62, 153, 63, 153, 64, 153, 65, 153, 164, 153, 165, 153, 4, 154, 5, 154, 6, 154, 7, 154, 8, 154, 9, 154, 10, 154, 11, 154, 12, 154, 13, 154, 14, 154, 15, 154, 16, 154, 17, 154, 18, 154, 19, 154, 147, 75, 150, 75, 151, 75, 102, 216, 241, 220, 102, 216, 250, 220, 102, 216, 3, 221, 102, 216, 5, 221, 192, 154, 193, 154, 220, 154, 3, 155, 4, 155, 5, 155, 6, 155, 8, 155, 0, 76, 1, 76, 41, 155, 53, 155, 54, 155, 74, 155, 75, 155, 76, 155, 77, 155, 78, 155, 79, 155, 38, 76, 181, 155, 182, 155, 183, 155, 184, 155, 185, 155, 187, 155, 188, 155, 189, 155, 190, 155, 191, 155, 192, 155, 193, 155, 194, 155, 195, 155, 196, 155, 198, 155, 199, 155, 200, 155, 201, 155, 202, 155, 203, 155, 204, 155, 205, 155, 207, 155, 208, 155, 209, 155, 210, 155, 211, 155, 253, 155, 87, 76, 103, 216, 246, 221, 204, 156, 205, 156, 206, 156, 207, 156, 208, 156, 209, 156, 210, 156, 74, 157, 75, 157, 76, 157, 77, 157, 78, 157, 79, 157, 80, 157, 81, 157, 82, 157, 83, 157, 84, 157, 85, 157, 86, 157, 87, 157, 88, 157, 89, 157, 90, 157, 91, 157, 92, 157, 60, 0, 42, 0, 93, 157, 94, 157, 95, 157, 96, 157, 98, 157, 99, 157, 101, 157, 109, 158, 110, 158, 111, 158, 112, 158, 140, 158, 141, 158, 142, 158, 143, 158, 144, 158, 38, 77, 104, 216, 180, 222, 104, 216, 182, 222, 104, 216, 186, 222, 104, 216, 189, 222, 177, 158, 178, 158, 54, 77, 191, 158, 202, 158, 203, 158, 223, 158, 224, 158, 225, 158, 0, 159, 1, 159, 2, 159, 118, 77, 21, 159, 22, 159, 37, 159, 38, 159, 39, 159, 40, 159, 41, 159, 42, 159, 43, 159, 44, 159, 76, 159, 85, 159, 142, 159, 60, 0, 39, 0, 208, 253, 19, 40, 39, 0, 60, 0, 42, 0, 38, 52, 51, 81, 52, 81, 53, 81, 163, 52, 150, 82, 247, 82, 248, 82, 54, 83, 180, 83, 225, 88, 165, 86, 166, 86, 167, 86, 168, 86, 169, 86, 170, 86, 171, 86, 172, 86, 173, 86, 175, 86, 176, 86, 68, 216, 228, 220, 68, 216, 244, 220, 68, 216, 245, 220, 68, 216, 246, 220, 218, 88, 219, 88, 220, 88, 221, 88, 222, 88, 223, 88, 224, 88, 226, 88, 18, 89, 61, 91, 57, 91, 62, 91, 63, 91, 50, 55, 51, 55, 52, 55, 53, 55, 70, 216, 3, 221, 70, 216, 4, 221, 70, 216, 5, 221, 124, 91, 243, 91, 244, 91, 245, 91, 107, 92, 195, 93, 196, 93, 197, 93, 32, 56, 34, 56, 71, 216, 158, 223, 71, 216, 161, 223, 112, 94, 72, 216, 115, 221, 236, 94, 237, 94, 144, 159, 95, 95, 191, 95, 236, 97, 239, 97, 242, 97, 245, 97, 246, 97, 247, 97, 74, 216, 38, 221, 74, 216, 57, 221, 251, 97, 0, 101, 7, 101, 8, 101, 9, 101, 10, 101, 11, 101, 12, 101, 13, 101, 14, 101, 15, 101, 16, 101, 18, 101, 75, 216, 54, 222, 75, 216, 66, 222, 132, 101, 220, 101, 221, 101, 222, 101, 221, 102, 222, 102, 223, 102, 224, 102, 225, 102, 226, 102, 41, 59, 42, 59, 203, 106, 204, 106, 205, 106, 206, 106, 207, 106, 208, 106, 209, 106, 210, 106, 211, 106, 212, 106, 213, 106, 214, 106, 215, 106, 216, 106, 217, 106, 218, 106, 220, 106, 221, 106, 222, 106, 223, 106, 224, 106, 226, 106, 227, 106, 228, 106, 229, 106, 230, 106, 235, 106, 0, 60, 60, 0, 42, 0, 1, 60, 2, 60, 4, 60, 77, 216, 162, 223, 77, 216, 188, 223, 77, 216, 194, 223, 231, 106, 96, 107, 176, 107, 177, 107, 78, 216, 219, 222, 12, 108, 243, 111, 21, 112, 22, 112, 23, 112, 24, 112, 25, 112, 26, 112, 27, 112, 28, 112, 29, 112, 30, 112, 31, 112, 32, 112, 33, 112, 34, 112, 35, 112, 36, 112, 37, 112, 39, 112, 40, 112, 41, 112, 43, 112, 44, 112, 45, 112, 46, 112, 133, 61, 136, 61, 138, 61, 80, 216, 201, 220, 80, 216, 225, 220, 107, 216, 255, 221, 2, 114, 4, 114, 5, 114, 6, 114, 7, 114, 8, 114, 9, 114, 10, 114, 12, 114, 13, 114, 14, 114, 21, 114, 6, 62, 81, 216, 90, 220, 81, 216, 91, 220, 88, 114, 161, 114, 162, 114, 163, 114, 164, 114, 165, 114, 166, 114, 120, 115, 121, 115, 122, 115, 183, 116, 189, 116, 195, 116, 196, 116, 197, 116, 198, 116, 199, 116, 200, 116, 201, 116, 202, 116, 203, 116, 4, 63, 6, 63, 82, 216, 177, 222, 82, 216, 178, 222, 82, 216, 179, 222, 82, 216, 186, 222, 82, 216, 187, 222, 82, 216, 188, 222, 227, 116, 22, 117, 134, 117, 135, 117, 95, 118, 96, 118, 97, 118, 99, 118, 169, 118, 84, 216, 44, 220, 195, 119, 196, 119, 197, 119, 198, 119, 199, 119, 200, 119, 201, 119, 202, 119, 84, 216, 14, 223, 84, 216, 17, 223, 84, 216, 19, 223, 241, 119, 23, 121, 24, 121, 25, 121, 26, 121, 27, 121, 28, 121, 29, 121, 30, 121, 31, 121, 32, 121, 33, 121, 85, 216, 185, 221, 176, 121, 177, 121, 85, 216, 29, 223, 100, 122, 101, 122, 102, 122, 103, 122, 60, 0, 42, 0, 104, 122, 105, 122, 106, 122, 107, 122, 86, 216, 222, 220, 86, 216, 225, 220, 198, 122, 44, 124, 51, 124, 52, 124, 53, 124, 54, 124, 55, 124, 56, 124, 57, 124, 58, 124, 59, 124, 60, 124, 61, 124, 62, 124, 63, 124, 64, 124, 65, 124, 66, 124, 79, 66, 96, 66, 87, 216, 48, 221, 87, 216, 67, 221, 233, 124, 234, 124, 235, 124, 236, 124, 237, 124, 87, 216, 75, 223, 75, 126, 102, 126, 104, 126, 105, 126, 106, 126, 107, 126, 108, 126, 109, 126, 110, 126, 111, 126, 112, 126, 114, 126, 115, 126, 116, 126, 117, 126, 118, 126, 119, 126, 120, 126, 121, 126, 122, 126, 32, 67, 53, 127, 74, 127, 75, 127, 131, 127, 132, 127, 133, 127, 134, 127, 182, 127, 183, 127, 184, 127, 185, 127, 253, 127, 254, 127, 120, 128, 203, 129, 212, 129, 213, 129, 215, 129, 216, 129, 70, 68, 90, 216, 138, 220, 90, 216, 147, 220, 90, 216, 38, 221, 11, 130, 26, 130, 97, 130, 98, 130, 99, 130, 100, 130, 101, 130, 118, 130, 69, 69, 70, 69, 81, 69, 209, 133, 213, 133, 214, 133, 215, 133, 216, 133, 217, 133, 218, 133, 219, 133, 220, 133, 221, 133, 222, 133, 223, 133, 224, 133, 225, 133, 226, 133, 227, 133, 228, 133, 229, 133, 230, 133, 231, 133, 232, 133, 233, 133, 234, 133, 235, 133, 236, 133, 237, 133, 239, 133, 240, 133, 241, 133, 242, 133, 243, 133, 244, 133, 245, 133, 92, 216, 75, 220, 92, 216, 82, 220, 92, 216, 83, 220, 247, 133, 248, 133, 1, 136, 60, 0, 42, 0, 213, 135, 246, 135, 247, 135, 248, 135, 249, 135, 250, 135, 251, 135, 252, 135, 253, 135, 254, 135, 255, 135, 0, 136, 2, 136, 3, 136, 4, 136, 5, 136, 6, 136, 7, 136, 8, 136, 9, 136, 10, 136, 11, 136, 12, 136, 13, 136, 15, 136, 30, 136, 86, 137, 87, 137, 88, 137, 89, 137, 90, 137, 91, 137, 92, 137, 93, 137, 94, 137, 95, 137, 96, 137, 97, 137, 98, 137, 135, 137, 136, 137, 180, 137, 181, 137, 182, 137, 183, 137, 184, 137, 245, 137, 246, 137, 63, 139, 64, 139, 65, 139, 66, 139, 67, 139, 68, 139, 70, 139, 72, 139, 73, 139, 74, 139, 75, 139, 76, 139, 78, 139, 79, 139, 80, 139, 81, 139, 82, 139, 83, 139, 84, 139, 85, 139, 86, 139, 87, 139, 88, 139, 89, 139, 90, 139, 91, 139, 92, 139, 94, 216, 6, 223, 94, 216, 11, 223, 94, 216, 24, 223, 54, 140, 67, 140, 119, 140, 118, 140, 154, 140, 6, 141, 7, 141, 8, 141, 9, 141, 10, 141, 11, 141, 12, 141, 170, 141, 171, 141, 172, 141, 173, 141, 104, 142, 106, 142, 107, 142, 108, 142, 109, 142, 111, 142, 112, 142, 113, 142, 114, 142, 115, 142, 116, 142, 117, 142, 118, 142, 119, 142, 120, 142, 121, 142, 122, 142, 123, 142, 124, 142, 125, 142, 126, 142, 127, 142, 38, 72, 96, 216, 79, 221, 96, 216, 93, 221, 96, 216, 111, 221, 135, 142, 194, 142, 195, 142, 196, 142, 197, 142, 77, 143, 78, 143, 79, 143, 60, 0, 42, 0, 80, 143, 81, 143, 82, 143, 83, 143, 84, 143, 173, 143, 180, 143, 138, 144, 139, 144, 140, 144, 97, 216, 59, 222, 63, 145, 64, 145, 66, 145, 173, 145, 174, 145, 175, 145, 176, 145, 177, 145, 17, 73, 169, 147, 192, 147, 193, 147, 194, 147, 195, 147, 196, 147, 197, 147, 198, 147, 199, 147, 200, 147, 201, 147, 202, 147, 203, 147, 204, 147, 205, 147, 206, 147, 207, 147, 208, 147, 209, 147, 210, 147, 211, 147, 212, 147, 213, 147, 214, 147, 215, 147, 216, 147, 217, 147, 218, 147, 219, 147, 220, 147, 221, 147, 222, 147, 223, 147, 224, 147, 225, 147, 226, 147, 227, 147, 228, 147, 229, 147, 230, 147, 231, 147, 232, 147, 233, 147, 234, 147, 235, 147, 236, 147, 237, 147, 238, 147, 239, 147, 240, 147, 241, 147, 242, 147, 249, 147, 73, 73, 81, 73, 83, 73, 98, 216, 222, 222, 98, 216, 223, 222, 98, 216, 224, 222, 98, 216, 225, 222, 98, 216, 226, 222, 98, 216, 227, 222, 98, 216, 229, 222, 98, 216, 234, 222, 98, 216, 252, 222, 98, 216, 12, 223, 114, 149, 125, 149, 217, 149, 218, 149, 219, 149, 220, 149, 221, 149, 180, 150, 225, 150, 227, 150, 4, 74, 38, 151, 39, 151, 40, 151, 41, 151, 42, 151, 43, 151, 44, 151, 45, 151, 100, 216, 60, 221, 100, 216, 77, 221, 100, 216, 91, 221, 97, 151, 177, 151, 178, 151, 179, 151, 180, 151, 181, 151, 182, 151, 183, 151, 220, 151, 221, 151, 222, 151, 223, 151, 242, 151, 251, 151, 252, 151, 86, 152, 87, 152, 88, 152, 89, 152, 90, 152, 60, 0, 42, 0, 91, 152, 92, 152, 93, 152, 94, 152, 228, 74, 164, 152, 187, 152, 188, 152, 189, 152, 190, 152, 191, 152, 192, 152, 25, 75, 67, 153, 83, 75, 68, 153, 69, 153, 71, 153, 72, 153, 73, 153, 166, 153, 167, 153, 102, 216, 87, 220, 20, 154, 21, 154, 22, 154, 23, 154, 24, 154, 25, 154, 26, 154, 27, 154, 28, 154, 29, 154, 30, 154, 31, 154, 32, 154, 33, 154, 34, 154, 35, 154, 36, 154, 37, 154, 38, 154, 39, 154, 40, 154, 157, 75, 102, 216, 47, 221, 165, 154, 194, 154, 195, 154, 197, 154, 9, 155, 10, 155, 11, 155, 12, 155, 13, 155, 14, 155, 15, 155, 4, 76, 7, 76, 55, 155, 197, 155, 212, 155, 213, 155, 214, 155, 215, 155, 216, 155, 217, 155, 218, 155, 219, 155, 220, 155, 221, 155, 222, 155, 223, 155, 224, 155, 225, 155, 226, 155, 227, 155, 228, 155, 229, 155, 230, 155, 231, 155, 232, 155, 233, 155, 234, 155, 235, 155, 236, 155, 237, 155, 238, 155, 239, 155, 240, 155, 241, 155, 242, 155, 243, 155, 244, 155, 245, 155, 91, 76, 103, 216, 6, 222, 103, 216, 45, 222, 250, 155, 211, 156, 212, 156, 213, 156, 214, 156, 215, 156, 216, 156, 217, 156, 218, 156, 219, 156, 97, 157, 102, 157, 104, 157, 105, 157, 106, 157, 107, 157, 108, 157, 109, 157, 110, 157, 111, 157, 112, 157, 113, 157, 114, 157, 115, 157, 116, 157, 117, 157, 118, 157, 119, 157, 120, 157, 121, 157, 60, 0, 42, 0, 122, 157, 123, 157, 124, 157, 125, 157, 126, 157, 127, 157, 128, 157, 129, 157, 130, 157, 131, 157, 132, 157, 133, 157, 134, 157, 135, 157, 136, 157, 137, 157, 138, 157, 139, 157, 140, 157, 141, 157, 142, 157, 143, 157, 145, 157, 225, 76, 226, 76, 104, 216, 135, 220, 113, 158, 114, 158, 120, 158, 145, 158, 146, 158, 147, 158, 148, 158, 149, 158, 150, 158, 151, 158, 179, 158, 180, 158, 192, 158, 76, 77, 226, 158, 227, 158, 252, 158, 3, 159, 4, 159, 119, 77, 23, 159, 45, 159, 64, 159, 65, 159, 77, 159, 86, 159, 87, 159, 88, 159, 143, 159, 105, 216, 169, 222, 60, 0, 39, 0, 208, 253, 20, 40, 39, 0, 60, 0, 42, 0, 54, 81, 165, 52, 55, 83, 177, 86, 178, 86, 179, 86, 180, 86, 181, 86, 182, 86, 183, 86, 184, 86, 185, 86, 68, 216, 47, 221, 188, 86, 37, 54, 227, 88, 228, 88, 229, 88, 69, 216, 232, 220, 64, 91, 65, 91, 66, 91, 67, 91, 68, 91, 69, 91, 70, 91, 54, 55, 55, 55, 56, 55, 70, 216, 16, 221, 70, 216, 17, 221, 70, 216, 18, 221, 70, 216, 21, 221, 125, 91, 126, 91, 246, 91, 198, 93, 199, 93, 200, 93, 201, 93, 202, 93, 204, 93, 113, 94, 72, 216, 122, 221, 238, 94, 239, 94, 240, 94, 192, 95, 193, 95, 248, 97, 164, 57, 249, 97, 250, 97, 74, 216, 79, 221, 19, 101, 20, 101, 21, 101, 22, 101, 23, 101, 24, 101, 25, 101, 26, 101, 112, 58, 133, 101, 134, 101, 223, 101, 227, 102, 228, 102, 229, 102, 230, 102, 231, 102, 232, 102, 43, 59, 76, 216, 0, 223, 39, 103, 232, 106, 233, 106, 234, 106, 236, 106, 238, 106, 239, 106, 240, 106, 241, 106, 242, 106, 243, 106, 244, 106, 245, 106, 246, 106, 9, 60, 10, 60, 13, 60, 17, 60, 77, 216, 213, 223, 77, 216, 214, 223, 77, 216, 215, 223, 249, 106, 42, 112, 47, 112, 48, 112, 49, 112, 50, 112, 51, 112, 52, 112, 53, 112, 54, 112, 55, 112, 56, 112, 57, 112, 58, 112, 59, 112, 60, 112, 61, 112, 62, 112, 63, 112, 64, 112, 65, 112, 143, 61, 145, 61, 80, 216, 236, 220, 80, 216, 4, 221, 66, 112, 11, 114, 15, 114, 16, 114, 17, 114, 18, 114, 19, 114, 20, 114, 60, 0, 42, 0, 22, 114, 23, 114, 24, 114, 10, 62, 81, 216, 115, 220, 81, 216, 135, 220, 81, 216, 136, 220, 167, 114, 168, 114, 81, 216, 6, 223, 123, 115, 124, 115, 125, 115, 186, 116, 204, 116, 205, 116, 206, 116, 207, 116, 208, 116, 209, 116, 210, 116, 7, 63, 82, 216, 192, 222, 82, 216, 199, 222, 136, 117, 137, 117, 98, 118, 101, 118, 102, 118, 170, 118, 171, 118, 190, 118, 250, 63, 237, 118, 203, 119, 204, 119, 205, 119, 206, 119, 207, 119, 242, 119, 34, 121, 35, 121, 36, 121, 37, 121, 38, 121, 39, 121, 40, 121, 41, 121, 42, 121, 43, 121, 44, 121, 178, 121, 108, 122, 109, 122, 110, 122, 111, 122, 199, 122, 246, 122, 247, 122, 67, 124, 68, 124, 69, 124, 70, 124, 71, 124, 72, 124, 73, 124, 74, 124, 75, 124, 76, 124, 77, 124, 78, 124, 79, 124, 85, 124, 238, 124, 239, 124, 240, 124, 174, 66, 123, 126, 124, 126, 125, 126, 126, 126, 127, 126, 128, 126, 129, 126, 130, 126, 131, 126, 38, 67, 88, 216, 173, 221, 88, 216, 174, 221, 88, 216, 178, 221, 76, 127, 88, 216, 53, 223, 186, 127, 255, 127, 0, 128, 47, 128, 121, 128, 122, 128, 123, 128, 124, 128, 214, 129, 217, 129, 218, 129, 219, 129, 220, 129, 90, 216, 181, 221, 102, 130, 103, 130, 104, 130, 105, 130, 36, 134, 82, 69, 84, 69, 85, 69, 87, 69, 88, 69, 92, 69, 238, 133, 246, 133, 249, 133, 250, 133, 251, 133, 252, 133, 253, 133, 254, 133, 255, 133, 0, 134, 1, 134, 60, 0, 42, 0, 2, 134, 3, 134, 4, 134, 5, 134, 6, 134, 7, 134, 8, 134, 9, 134, 10, 134, 11, 134, 12, 134, 13, 134, 14, 134, 15, 134, 16, 134, 17, 134, 19, 134, 20, 134, 34, 134, 92, 216, 136, 220, 92, 216, 173, 220, 92, 216, 174, 220, 92, 216, 175, 220, 18, 134, 27, 134, 48, 134, 16, 136, 17, 136, 18, 136, 19, 136, 20, 136, 21, 136, 22, 136, 23, 136, 24, 136, 25, 136, 0, 70, 1, 70, 99, 137, 100, 137, 101, 137, 102, 137, 103, 137, 104, 137, 185, 137, 186, 137, 187, 137, 247, 137, 248, 137, 249, 137, 77, 139, 93, 139, 94, 139, 95, 139, 96, 139, 97, 139, 98, 139, 99, 139, 100, 139, 101, 139, 102, 139, 103, 139, 104, 139, 105, 139, 106, 139, 107, 139, 108, 139, 109, 139, 110, 139, 111, 139, 112, 139, 113, 139, 114, 139, 19, 71, 24, 71, 94, 216, 56, 223, 94, 216, 57, 223, 94, 216, 58, 223, 81, 140, 95, 216, 18, 220, 13, 141, 14, 141, 15, 141, 174, 141, 129, 142, 130, 142, 131, 142, 132, 142, 133, 142, 134, 142, 136, 142, 137, 142, 96, 216, 137, 221, 198, 142, 85, 143, 86, 143, 87, 143, 88, 143, 89, 143, 90, 143, 174, 143, 141, 144, 65, 145, 67, 145, 178, 145, 179, 145, 180, 145, 181, 145, 182, 145, 183, 145, 184, 145, 203, 145, 243, 147, 245, 147, 246, 147, 247, 147, 248, 147, 250, 147, 251, 147, 252, 147, 253, 147, 254, 147, 255, 147, 0, 148, 1, 148, 2, 148, 3, 148, 4, 148, 60, 0, 42, 0, 5, 148, 6, 148, 7, 148, 8, 148, 9, 148, 10, 148, 11, 148, 12, 148, 13, 148, 14, 148, 15, 148, 16, 148, 17, 148, 18, 148, 19, 148, 20, 148, 21, 148, 22, 148, 23, 148, 24, 148, 25, 148, 26, 148, 27, 148, 28, 148, 29, 148, 30, 148, 31, 148, 32, 148, 33, 148, 34, 148, 35, 148, 36, 148, 37, 148, 38, 148, 39, 148, 40, 148, 98, 216, 19, 223, 98, 216, 33, 223, 98, 216, 34, 223, 98, 216, 43, 223, 98, 216, 44, 223, 98, 216, 45, 223, 98, 216, 47, 223, 98, 216, 70, 223, 98, 216, 76, 223, 98, 216, 78, 223, 98, 216, 80, 223, 47, 148, 60, 148, 115, 149, 116, 149, 222, 149, 223, 149, 224, 149, 225, 149, 99, 216, 185, 221, 181, 150, 46, 151, 47, 151, 48, 151, 49, 151, 51, 151, 52, 151, 100, 216, 94, 221, 100, 216, 112, 221, 75, 74, 184, 151, 185, 151, 186, 151, 187, 151, 219, 151, 224, 151, 253, 151, 254, 151, 255, 151, 95, 152, 96, 152, 97, 152, 98, 152, 99, 152, 165, 152, 193, 152, 194, 152, 195, 152, 196, 152, 74, 153, 75, 153, 76, 153, 77, 153, 78, 153, 80, 153, 81, 153, 82, 153, 83, 153, 89, 153, 168, 153, 41, 154, 42, 154, 43, 154, 44, 154, 45, 154, 46, 154, 47, 154, 48, 154, 49, 154, 50, 154, 51, 154, 52, 154, 53, 154, 54, 154, 55, 154, 56, 154, 102, 216, 69, 221, 102, 216, 71, 221, 102, 216, 72, 221, 102, 216, 73, 221, 102, 216, 93, 221, 102, 216, 106, 221, 166, 154, 167, 154, 196, 154, 198, 154, 199, 154, 200, 154, 201, 154, 202, 154, 203, 154, 204, 154, 60, 0, 42, 0, 16, 155, 17, 155, 18, 155, 19, 155, 42, 155, 56, 155, 80, 155, 251, 155, 246, 155, 247, 155, 248, 155, 249, 155, 252, 155, 254, 155, 255, 155, 0, 156, 1, 156, 2, 156, 3, 156, 4, 156, 5, 156, 6, 156, 7, 156, 8, 156, 9, 156, 10, 156, 11, 156, 12, 156, 13, 156, 14, 156, 15, 156, 16, 156, 17, 156, 18, 156, 19, 156, 20, 156, 21, 156, 22, 156, 23, 156, 24, 156, 25, 156, 26, 156, 27, 156, 32, 156, 109, 76, 103, 216, 104, 222, 64, 156, 220, 156, 221, 156, 222, 156, 223, 156, 144, 157, 146, 157, 147, 157, 148, 157, 149, 157, 150, 157, 151, 157, 152, 157, 153, 157, 154, 157, 155, 157, 156, 157, 157, 157, 158, 157, 159, 157, 160, 157, 161, 157, 162, 157, 163, 157, 164, 157, 165, 157, 166, 157, 167, 157, 168, 157, 169, 157, 170, 157, 171, 157, 237, 76, 104, 216, 185, 220, 104, 216, 225, 220, 104, 216, 237, 220, 104, 216, 243, 220, 104, 216, 248, 220, 191, 157, 121, 158, 152, 158, 153, 158, 154, 158, 155, 158, 181, 158, 104, 216, 81, 223, 193, 158, 104, 216, 169, 223, 77, 77, 228, 158, 229, 158, 230, 158, 231, 158, 232, 158, 233, 158, 234, 158, 105, 216, 91, 220, 13, 159, 46, 159, 47, 159, 48, 159, 105, 216, 203, 221, 89, 159, 90, 159, 91, 159, 93, 159, 94, 159, 95, 159, 96, 159, 97, 159, 99, 159, 145, 159, 60, 0, 39, 0, 208, 253, 21, 40, 39, 0, 60, 0, 42, 0, 55, 81, 56, 81, 57, 81, 58, 81, 167, 52, 100, 81, 151, 82, 152, 82, 66, 216, 44, 220, 91, 83, 186, 86, 187, 86, 189, 86, 190, 86, 191, 86, 192, 86, 193, 86, 194, 86, 195, 86, 196, 86, 205, 86, 68, 216, 59, 221, 68, 216, 61, 221, 68, 216, 69, 221, 68, 216, 72, 221, 68, 216, 79, 221, 230, 88, 69, 216, 253, 220, 20, 89, 71, 91, 72, 91, 73, 91, 57, 55, 70, 216, 28, 221, 247, 91, 108, 92, 203, 93, 205, 93, 207, 93, 208, 93, 38, 56, 241, 94, 194, 95, 252, 97, 253, 97, 254, 97, 17, 101, 27, 101, 28, 101, 29, 101, 75, 216, 120, 222, 149, 101, 233, 102, 76, 216, 10, 223, 7, 103, 248, 106, 250, 106, 251, 106, 252, 106, 253, 106, 254, 106, 255, 106, 0, 107, 1, 107, 2, 107, 3, 107, 4, 107, 5, 107, 12, 107, 21, 60, 178, 107, 67, 112, 68, 112, 69, 112, 70, 112, 71, 112, 72, 112, 73, 112, 74, 112, 75, 112, 76, 112, 77, 112, 79, 112, 80, 112, 80, 216, 15, 221, 80, 216, 25, 221, 25, 114, 26, 114, 27, 114, 13, 62, 81, 216, 185, 220, 29, 114, 126, 115, 211, 116, 212, 116, 214, 116, 82, 216, 202, 222, 82, 216, 209, 222, 23, 117, 103, 118, 104, 118, 105, 118, 106, 118, 107, 118, 215, 63, 172, 118, 83, 216, 184, 223, 208, 119, 209, 119, 210, 119, 211, 119, 45, 121, 46, 121, 47, 121, 48, 121, 49, 121, 50, 121, 51, 121, 52, 121, 85, 216, 213, 221, 85, 216, 219, 221, 85, 216, 37, 223, 86, 216, 3, 221, 195, 122, 200, 122, 201, 122, 86, 124, 80, 124, 60, 0, 42, 0, 81, 124, 82, 124, 83, 124, 84, 124, 106, 66, 87, 216, 153, 221, 242, 124, 132, 126, 133, 126, 134, 126, 135, 126, 136, 126, 137, 126, 138, 126, 139, 126, 140, 126, 141, 126, 142, 126, 143, 126, 144, 126, 77, 127, 187, 127, 188, 127, 154, 67, 48, 128, 221, 129, 106, 130, 212, 133, 97, 69, 98, 69, 21, 134, 22, 134, 23, 134, 24, 134, 25, 134, 26, 134, 28, 134, 29, 134, 30, 134, 31, 134, 32, 134, 33, 134, 35, 134, 37, 134, 38, 134, 39, 134, 40, 134, 41, 134, 42, 134, 43, 134, 45, 134, 46, 134, 47, 134, 92, 216, 205, 220, 92, 216, 210, 220, 92, 216, 248, 220, 92, 216, 9, 221, 92, 216, 12, 221, 26, 136, 27, 136, 28, 136, 29, 136, 31, 136, 32, 136, 33, 136, 34, 136, 35, 136, 36, 136, 2, 70, 3, 70, 41, 136, 43, 136, 74, 136, 105, 137, 106, 137, 107, 137, 108, 137, 109, 137, 110, 137, 188, 137, 189, 137, 250, 137, 69, 139, 115, 139, 116, 139, 117, 139, 118, 139, 119, 139, 120, 139, 121, 139, 122, 139, 123, 139, 124, 139, 125, 139, 94, 216, 72, 223, 16, 141, 17, 141, 18, 141, 19, 141, 20, 141, 99, 141, 175, 141, 176, 141, 138, 142, 139, 142, 140, 142, 141, 142, 142, 142, 143, 142, 96, 216, 175, 221, 96, 216, 188, 221, 199, 142, 91, 143, 92, 143, 93, 143, 94, 143, 95, 143, 175, 143, 142, 144, 68, 145, 69, 145, 70, 145, 185, 145, 186, 145, 187, 145, 244, 147, 41, 148, 42, 148, 60, 0, 42, 0, 43, 148, 44, 148, 45, 148, 46, 148, 48, 148, 49, 148, 50, 148, 51, 148, 52, 148, 53, 148, 54, 148, 55, 148, 56, 148, 57, 148, 58, 148, 59, 148, 61, 148, 62, 148, 63, 148, 64, 148, 65, 148, 101, 73, 98, 216, 99, 223, 98, 216, 100, 223, 98, 216, 101, 223, 98, 216, 102, 223, 98, 216, 108, 223, 98, 216, 143, 223, 226, 149, 227, 149, 228, 149, 229, 149, 230, 149, 228, 150, 50, 151, 53, 151, 54, 151, 55, 151, 56, 151, 57, 151, 58, 151, 59, 151, 103, 151, 188, 151, 189, 151, 190, 151, 191, 151, 225, 151, 226, 151, 164, 74, 101, 216, 63, 220, 100, 152, 101, 152, 102, 152, 103, 152, 104, 152, 166, 152, 197, 152, 198, 152, 199, 152, 200, 152, 201, 152, 202, 152, 220, 152, 79, 153, 86, 153, 87, 153, 88, 153, 169, 153, 57, 154, 58, 154, 59, 154, 60, 154, 61, 154, 62, 154, 63, 154, 64, 154, 65, 154, 66, 154, 67, 154, 68, 154, 69, 154, 70, 154, 71, 154, 173, 159, 205, 154, 206, 154, 207, 154, 20, 155, 21, 155, 22, 155, 23, 155, 24, 155, 57, 155, 58, 155, 81, 155, 82, 155, 83, 155, 84, 155, 28, 156, 29, 156, 30, 156, 31, 156, 33, 156, 34, 156, 35, 156, 36, 156, 37, 156, 38, 156, 39, 156, 40, 156, 41, 156, 42, 156, 43, 156, 44, 156, 45, 156, 46, 156, 47, 156, 48, 156, 119, 76, 123, 76, 125, 76, 103, 216, 172, 222, 163, 76, 164, 76, 224, 156, 225, 156, 226, 156, 60, 0, 42, 0, 227, 156, 172, 157, 173, 157, 174, 157, 175, 157, 176, 157, 177, 157, 178, 157, 179, 157, 180, 157, 181, 157, 182, 157, 183, 157, 184, 157, 185, 157, 186, 157, 187, 157, 188, 157, 189, 157, 190, 157, 192, 157, 193, 157, 194, 157, 195, 157, 196, 157, 197, 157, 198, 157, 199, 157, 200, 157, 201, 157, 202, 157, 204, 157, 205, 157, 206, 157, 207, 157, 104, 216, 254, 220, 104, 216, 7, 221, 104, 216, 35, 221, 122, 158, 123, 158, 156, 158, 157, 158, 78, 77, 235, 158, 236, 158, 237, 158, 238, 158, 239, 158, 5, 159, 24, 159, 25, 159, 26, 159, 27, 159, 49, 159, 78, 159, 92, 159, 100, 159, 101, 159, 102, 159, 103, 159, 104, 159, 105, 159, 105, 216, 1, 222, 146, 159, 157, 159, 161, 159, 60, 0, 39, 0, 208, 253, 22, 40, 39, 0, 60, 0, 42, 0, 185, 78, 59, 81, 60, 81, 65, 216, 101, 220, 66, 216, 121, 221, 197, 86, 198, 86, 199, 86, 200, 86, 201, 86, 202, 86, 203, 86, 206, 86, 24, 54, 29, 87, 113, 89, 74, 91, 75, 91, 76, 91, 58, 55, 70, 216, 34, 221, 70, 216, 39, 221, 127, 91, 206, 93, 209, 93, 210, 93, 211, 93, 212, 93, 213, 93, 215, 93, 242, 94, 78, 95, 114, 95, 255, 97, 2, 98, 74, 216, 103, 221, 74, 216, 107, 221, 53, 98, 30, 101, 31, 101, 32, 101, 33, 101, 34, 101, 36, 101, 38, 101, 39, 101, 75, 216, 139, 222, 76, 216, 51, 220, 44, 59, 45, 59, 247, 106, 6, 107, 7, 107, 8, 107, 9, 107, 10, 107, 11, 107, 13, 107, 14, 107, 24, 60, 97, 107, 13, 108, 81, 112, 82, 112, 84, 112, 85, 112, 86, 112, 87, 112, 88, 112, 80, 216, 63, 221, 80, 216, 64, 221, 80, 216, 68, 221, 28, 114, 30, 114, 31, 114, 32, 114, 169, 114, 127, 115, 128, 115, 213, 116, 215, 116, 216, 116, 217, 116, 228, 116, 138, 117, 108, 118, 109, 118, 110, 118, 83, 216, 14, 223, 173, 118, 53, 121, 248, 64, 85, 216, 224, 221, 179, 121, 180, 121, 112, 122, 113, 122, 202, 122, 248, 122, 87, 124, 88, 124, 89, 124, 90, 124, 91, 124, 92, 124, 93, 124, 95, 124, 96, 124, 97, 124, 87, 216, 185, 221, 241, 124, 244, 124, 145, 126, 146, 126, 43, 67, 88, 216, 221, 221, 78, 127, 79, 127, 88, 216, 97, 222, 135, 127, 49, 128, 50, 128, 125, 128, 126, 128, 222, 129, 223, 129, 90, 216, 250, 221, 107, 130, 60, 0, 42, 0, 103, 69, 106, 69, 109, 69, 44, 134, 50, 134, 51, 134, 52, 134, 53, 134, 54, 134, 55, 134, 92, 216, 13, 221, 92, 216, 38, 221, 92, 216, 39, 221, 37, 136, 38, 136, 39, 136, 40, 136, 42, 136, 44, 136, 6, 70, 111, 137, 112, 137, 113, 137, 114, 137, 190, 137, 191, 137, 94, 216, 157, 220, 251, 137, 252, 137, 126, 139, 127, 139, 128, 139, 129, 139, 130, 139, 131, 139, 132, 139, 133, 139, 32, 71, 94, 216, 101, 223, 134, 139, 68, 140, 21, 141, 22, 141, 23, 141, 24, 141, 144, 142, 145, 142, 146, 142, 147, 142, 148, 142, 149, 142, 150, 142, 151, 142, 154, 142, 96, 143, 97, 143, 98, 143, 71, 145, 72, 145, 66, 148, 67, 148, 68, 148, 69, 148, 70, 148, 71, 148, 72, 148, 73, 148, 74, 148, 75, 148, 76, 148, 77, 148, 78, 148, 79, 148, 80, 148, 81, 148, 82, 148, 83, 148, 84, 148, 103, 148, 106, 73, 109, 73, 98, 216, 153, 223, 98, 216, 156, 223, 98, 216, 157, 223, 117, 149, 118, 149, 126, 149, 231, 149, 60, 151, 61, 151, 62, 151, 63, 151, 64, 151, 100, 216, 156, 221, 192, 151, 193, 151, 194, 151, 195, 151, 227, 151, 105, 152, 106, 152, 107, 152, 203, 152, 84, 153, 85, 153, 90, 153, 91, 153, 101, 216, 212, 223, 72, 154, 73, 154, 74, 154, 75, 154, 76, 154, 77, 154, 78, 154, 79, 154, 80, 154, 81, 154, 82, 154, 83, 154, 84, 154, 85, 154, 102, 216, 157, 221, 208, 154, 210, 154, 221, 154, 29, 155, 25, 155, 60, 0, 42, 0, 26, 155, 27, 155, 28, 155, 14, 76, 102, 216, 213, 223, 43, 155, 59, 155, 85, 155, 86, 155, 49, 156, 50, 156, 51, 156, 52, 156, 53, 156, 54, 156, 55, 156, 56, 156, 57, 156, 58, 156, 59, 156, 60, 156, 61, 156, 62, 156, 63, 156, 65, 156, 66, 156, 67, 156, 68, 156, 69, 156, 70, 156, 71, 156, 72, 156, 224, 157, 129, 76, 133, 76, 103, 216, 176, 222, 103, 216, 195, 222, 73, 156, 228, 156, 203, 157, 208, 157, 209, 157, 210, 157, 211, 157, 212, 157, 213, 157, 214, 157, 215, 157, 216, 157, 217, 157, 218, 157, 219, 157, 220, 157, 221, 157, 222, 157, 223, 157, 104, 216, 51, 221, 104, 216, 52, 221, 104, 216, 80, 221, 233, 157, 245, 157, 115, 158, 116, 158, 158, 158, 104, 216, 223, 222, 182, 158, 208, 158, 240, 158, 241, 158, 50, 159, 51, 159, 52, 159, 53, 159, 66, 159, 106, 159, 107, 159, 108, 159, 156, 77, 147, 159, 148, 159, 149, 159, 162, 159, 60, 0, 39, 0, 208, 253, 23, 40, 39, 0, 60, 0, 42, 0, 61, 81, 153, 82, 154, 82, 66, 216, 205, 222, 204, 86, 207, 86, 208, 86, 26, 54, 231, 88, 232, 88, 69, 216, 130, 221, 114, 89, 77, 91, 214, 93, 216, 93, 218, 93, 79, 95, 0, 98, 1, 98, 3, 98, 4, 98, 35, 101, 37, 101, 40, 101, 41, 101, 42, 101, 43, 101, 119, 58, 150, 101, 234, 102, 235, 102, 236, 102, 46, 59, 15, 107, 16, 107, 17, 107, 18, 107, 78, 216, 58, 220, 202, 107, 83, 112, 89, 112, 90, 112, 91, 112, 92, 112, 80, 216, 78, 221, 80, 216, 85, 221, 80, 216, 86, 221, 80, 216, 87, 221, 80, 216, 92, 221, 80, 216, 95, 221, 33, 114, 34, 114, 81, 216, 188, 220, 81, 216, 206, 220, 81, 216, 211, 220, 81, 216, 214, 220, 129, 115, 130, 115, 131, 115, 218, 116, 82, 216, 223, 222, 111, 118, 112, 118, 212, 119, 54, 121, 55, 121, 181, 121, 94, 124, 98, 124, 99, 124, 100, 124, 101, 124, 102, 124, 103, 124, 104, 124, 116, 66, 245, 124, 147, 126, 148, 126, 149, 126, 150, 126, 226, 129, 90, 216, 199, 220, 108, 130, 119, 69, 49, 134, 56, 134, 57, 134, 58, 134, 59, 134, 60, 134, 61, 134, 62, 134, 63, 134, 64, 134, 65, 134, 92, 216, 27, 222, 52, 136, 45, 136, 46, 136, 47, 136, 48, 136, 49, 136, 50, 136, 51, 136, 115, 137, 116, 137, 118, 137, 137, 137, 253, 137, 254, 137, 135, 139, 136, 139, 137, 139, 138, 139, 139, 139, 140, 139, 141, 139, 142, 139, 143, 139, 144, 139, 69, 140, 25, 141, 26, 141, 177, 141, 152, 142, 153, 142, 155, 142, 60, 0, 42, 0, 156, 142, 96, 216, 155, 222, 99, 143, 100, 143, 143, 144, 144, 144, 188, 145, 85, 148, 86, 148, 87, 148, 88, 148, 89, 148, 90, 148, 91, 148, 92, 148, 93, 148, 94, 148, 95, 148, 96, 148, 97, 148, 98, 148, 99, 148, 100, 148, 101, 148, 102, 148, 114, 73, 98, 216, 185, 223, 98, 216, 194, 223, 98, 216, 197, 223, 98, 216, 212, 223, 98, 216, 215, 223, 98, 216, 217, 223, 98, 216, 218, 223, 65, 151, 53, 74, 100, 216, 168, 221, 104, 151, 196, 151, 197, 151, 0, 152, 108, 152, 109, 152, 110, 152, 111, 152, 167, 152, 92, 153, 170, 153, 86, 154, 87, 154, 88, 154, 89, 154, 90, 154, 91, 154, 92, 154, 209, 154, 211, 154, 212, 154, 222, 154, 30, 155, 31, 155, 32, 155, 74, 156, 75, 156, 76, 156, 77, 156, 78, 156, 79, 156, 80, 156, 81, 156, 82, 156, 83, 156, 84, 156, 85, 156, 86, 156, 87, 156, 88, 156, 89, 156, 90, 156, 91, 156, 103, 216, 248, 222, 106, 156, 225, 157, 226, 157, 227, 157, 228, 157, 229, 157, 230, 157, 231, 157, 232, 157, 234, 157, 235, 157, 236, 157, 237, 157, 238, 157, 239, 157, 240, 157, 241, 157, 242, 157, 243, 157, 244, 157, 246, 157, 247, 157, 248, 157, 251, 157, 252, 157, 0, 77, 104, 216, 146, 221, 104, 216, 147, 221, 104, 216, 171, 221, 159, 158, 194, 158, 242, 158, 243, 158, 244, 158, 6, 159, 7, 159, 28, 159, 54, 159, 55, 159, 56, 159, 57, 159, 137, 77, 67, 159, 68, 159, 79, 159, 109, 159, 110, 159, 111, 159, 60, 0, 42, 0, 112, 159, 113, 159, 105, 216, 50, 222, 60, 0, 39, 0, 208, 253, 24, 40, 39, 0, 60, 0, 42, 0, 62, 81, 65, 216, 215, 220, 209, 86, 210, 86, 211, 86, 68, 216, 128, 221, 233, 88, 129, 54, 78, 91, 79, 91, 109, 92, 217, 93, 74, 216, 128, 221, 44, 101, 45, 101, 237, 102, 238, 102, 19, 107, 20, 107, 21, 107, 93, 112, 94, 112, 95, 112, 96, 112, 97, 112, 80, 216, 97, 221, 35, 114, 219, 116, 229, 116, 113, 118, 114, 118, 83, 216, 194, 223, 213, 119, 215, 119, 214, 119, 56, 121, 250, 64, 182, 121, 183, 121, 115, 122, 114, 122, 106, 124, 118, 66, 151, 126, 80, 127, 136, 127, 137, 127, 109, 130, 119, 130, 120, 69, 67, 134, 69, 134, 92, 216, 100, 221, 92, 216, 101, 221, 53, 136, 54, 136, 55, 136, 56, 136, 57, 136, 58, 136, 75, 136, 98, 136, 117, 137, 119, 137, 93, 216, 204, 223, 145, 139, 146, 139, 147, 139, 148, 139, 149, 139, 150, 139, 27, 141, 157, 142, 158, 142, 159, 142, 160, 142, 200, 142, 189, 145, 190, 145, 191, 145, 192, 145, 194, 145, 104, 148, 105, 148, 106, 148, 107, 148, 108, 148, 98, 216, 231, 223, 98, 216, 232, 223, 98, 216, 233, 223, 98, 216, 234, 223, 98, 216, 235, 223, 98, 216, 236, 223, 98, 216, 245, 223, 229, 150, 230, 150, 66, 151, 67, 151, 68, 151, 69, 151, 70, 151, 71, 151, 72, 151, 198, 151, 199, 151, 200, 151, 228, 151, 229, 151, 101, 216, 72, 220, 112, 152, 93, 153, 93, 154, 94, 154, 95, 154, 213, 154, 33, 155, 34, 155, 44, 155, 45, 155, 87, 155, 88, 155, 89, 155, 103, 216, 62, 221, 92, 156, 93, 156, 94, 156, 95, 156, 96, 156, 60, 0, 42, 0, 97, 156, 98, 156, 99, 156, 100, 156, 101, 156, 102, 156, 103, 156, 105, 156, 107, 156, 103, 216, 35, 223, 112, 156, 250, 157, 249, 157, 253, 157, 254, 157, 255, 157, 0, 158, 1, 158, 2, 158, 3, 158, 4, 158, 5, 158, 6, 158, 7, 158, 8, 158, 9, 158, 10, 158, 7, 77, 9, 77, 104, 216, 180, 221, 104, 216, 181, 221, 124, 158, 125, 158, 160, 158, 30, 159, 69, 159, 70, 159, 114, 159, 115, 159, 116, 159, 117, 159, 118, 159, 119, 159, 60, 0, 39, 0, 208, 253, 25, 40, 39, 0, 60, 0, 42, 0, 212, 86, 213, 86, 68, 216, 135, 221, 234, 88, 243, 94, 5, 98, 6, 98, 46, 101, 184, 101, 239, 102, 47, 59, 22, 107, 23, 107, 24, 107, 25, 107, 26, 107, 27, 107, 29, 107, 98, 112, 99, 112, 36, 114, 37, 114, 38, 114, 170, 114, 83, 216, 6, 221, 216, 119, 217, 119, 225, 119, 57, 121, 105, 124, 107, 124, 108, 124, 109, 124, 110, 124, 246, 124, 152, 126, 153, 126, 154, 126, 155, 126, 224, 129, 225, 129, 66, 134, 70, 134, 71, 134, 72, 134, 73, 134, 59, 136, 93, 216, 116, 221, 120, 137, 121, 137, 122, 137, 123, 137, 124, 137, 138, 137, 192, 137, 255, 137, 151, 139, 152, 139, 153, 139, 82, 140, 68, 71, 155, 140, 28, 141, 95, 216, 77, 222, 161, 142, 162, 142, 163, 142, 164, 142, 165, 142, 193, 145, 109, 148, 110, 148, 111, 148, 112, 148, 113, 148, 114, 148, 115, 148, 98, 216, 255, 223, 99, 216, 3, 220, 73, 151, 113, 152, 114, 152, 94, 153, 95, 153, 149, 153, 189, 75, 102, 216, 195, 221, 102, 216, 201, 221, 214, 154, 35, 155, 104, 156, 108, 156, 109, 156, 110, 156, 111, 156, 103, 216, 48, 223, 11, 158, 12, 158, 13, 158, 14, 158, 15, 158, 16, 158, 17, 158, 18, 158, 104, 216, 223, 221, 161, 158, 204, 158, 245, 158, 8, 159, 9, 159, 29, 159, 31, 159, 71, 159, 120, 159, 121, 159, 122, 159, 123, 159, 105, 216, 74, 222, 163, 159, 60, 0, 39, 0, 208, 253, 26, 40, 39, 0, 60, 0, 42, 0, 54, 53, 30, 87, 59, 55, 96, 95, 28, 107, 14, 108, 78, 112, 100, 112, 102, 112, 82, 216, 226, 222, 115, 118, 218, 119, 111, 124, 112, 124, 87, 216, 14, 222, 243, 124, 68, 134, 106, 134, 60, 136, 154, 139, 155, 139, 95, 216, 79, 222, 178, 141, 166, 142, 167, 142, 195, 145, 196, 145, 116, 148, 117, 148, 118, 148, 119, 148, 120, 148, 121, 148, 122, 148, 99, 216, 9, 220, 99, 216, 28, 220, 99, 216, 29, 220, 74, 151, 201, 151, 96, 154, 97, 154, 98, 154, 99, 154, 190, 75, 101, 154, 215, 154, 113, 156, 114, 156, 115, 156, 116, 156, 117, 156, 118, 156, 19, 158, 20, 158, 104, 216, 245, 221, 246, 158, 10, 159, 105, 216, 91, 222, 164, 159, 165, 159, 60, 0, 39, 0, 208, 253, 27, 40, 39, 0, 60, 0, 42, 0, 78, 216, 26, 223, 101, 112, 103, 112, 104, 112, 80, 216, 119, 221, 80, 216, 122, 221, 171, 114, 247, 124, 156, 126, 157, 126, 74, 134, 61, 136, 62, 136, 63, 136, 125, 137, 156, 139, 157, 139, 158, 139, 83, 140, 156, 140, 169, 142, 170, 142, 201, 142, 101, 143, 197, 145, 123, 148, 124, 148, 125, 148, 126, 148, 99, 216, 35, 220, 99, 216, 38, 220, 75, 151, 76, 151, 77, 151, 78, 151, 115, 152, 116, 152, 204, 152, 205, 152, 221, 152, 96, 153, 97, 153, 171, 153, 100, 154, 102, 154, 103, 154, 174, 159, 36, 155, 46, 155, 48, 155, 119, 156, 120, 156, 21, 158, 22, 158, 23, 158, 247, 158, 72, 159, 60, 0, 39, 0, 208, 253, 28, 40, 39, 0, 60, 0, 42, 0, 214, 86, 7, 98, 75, 216, 179, 222, 76, 216, 31, 223, 30, 107, 31, 107, 39, 114, 82, 216, 233, 222, 116, 118, 220, 63, 92, 216, 117, 221, 76, 134, 84, 140, 168, 142, 96, 216, 7, 222, 127, 148, 128, 148, 129, 148, 130, 148, 99, 216, 43, 220, 99, 216, 48, 220, 231, 150, 104, 154, 105, 154, 192, 75, 24, 158, 25, 158, 26, 158, 104, 216, 32, 222, 162, 158, 248, 158, 58, 159, 124, 159, 125, 159, 158, 159, 60, 0, 39, 0, 208, 253, 29, 40, 39, 0, 60, 0, 42, 0, 40, 114, 86, 216, 84, 222, 158, 126, 75, 134, 159, 139, 131, 148, 132, 148, 121, 73, 79, 151, 106, 154, 49, 155, 121, 156, 27, 158, 28, 158, 183, 158, 60, 0, 39, 0, 208, 253, 30, 40, 39, 0, 60, 0, 42, 0, 181, 83, 117, 118, 144, 65, 113, 124, 133, 69, 99, 216, 57, 220, 202, 151, 98, 153, 107, 154, 103, 216, 115, 220, 122, 156, 29, 158, 30, 158, 104, 216, 51, 222, 145, 77, 60, 0, 39, 0, 208, 253, 31, 40, 39, 0, 60, 0, 42, 0, 105, 112, 93, 216, 163, 221, 98, 216, 13, 221, 16, 77, 163, 158, 60, 0, 39, 0, 208, 253, 32, 40, 39, 0, 60, 0, 42, 0, 106, 112, 114, 124, 99, 216, 59, 220, 150, 159, 60, 0, 39, 0, 208, 253, 33, 40, 39, 0, 60, 0, 42, 0, 70, 216, 59, 221, 41, 114, 123, 156, 164, 158, 151, 159, 60, 0, 39, 0, 208, 253, 35, 40, 39, 0, 60, 0, 42, 0, 126, 159, 60, 0, 39, 0, 208, 253, 36, 40, 39, 0, 60, 0, 42, 0, 73, 159, 60, 0, 39, 0, 208, 253, 39, 40, 39, 0, 60, 0, 42, 0, 80, 151, 60, 0, 39, 0, 208, 253, 48, 40, 39, 0, 60, 0, 42, 0, 152, 159, 38, 0, 39, 0, 40, 0, 0, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 32, 50, 38, 0, 39, 0, 40, 0, 3, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 38, 50, 38, 0, 39, 0, 40, 0, 9, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 34, 50, 38, 0, 39, 0, 40, 0, 93, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 40, 50, 38, 0, 39, 0, 40, 0, 140, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 33, 50, 38, 0, 39, 0, 40, 0, 148, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 36, 50, 38, 0, 39, 0, 40, 0, 227, 78, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 57, 50, 38, 0, 39, 0, 40, 0, 1, 79, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 61, 50, 38, 0, 39, 0, 40, 0, 17, 79, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 65, 50, 38, 0, 39, 0, 40, 0, 107, 81, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 39, 50, 38, 0, 39, 0, 40, 0, 109, 81, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 37, 50, 38, 0, 39, 0, 40, 0, 180, 82, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 56, 50, 38, 0, 39, 0, 40, 0, 65, 83, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 41, 50, 38, 0, 39, 0, 40, 0, 84, 83, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 63, 50, 38, 0, 39, 0, 40, 0, 13, 84, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 52, 50, 38, 0, 39, 0, 40, 0, 124, 84, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 58, 50, 38, 0, 39, 0, 40, 0, 219, 86, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 35, 50, 38, 0, 39, 0, 40, 0, 31, 87, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 47, 50, 38, 0, 39, 0, 40, 0, 102, 91, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 59, 50, 38, 0, 39, 0, 40, 0, 229, 101, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 48, 50, 38, 0, 39, 0, 40, 0, 8, 103, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 42, 50, 38, 0, 39, 0, 40, 0, 9, 103, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 50, 50, 38, 0, 39, 0, 40, 0, 40, 103, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 45, 50, 38, 0, 39, 0, 40, 0, 42, 104, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 49, 50, 38, 0, 39, 0, 40, 0, 52, 108, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 44, 50, 38, 0, 39, 0, 40, 0, 107, 112, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 43, 50, 38, 0, 39, 0, 40, 0, 121, 114, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 53, 50, 38, 0, 39, 0, 40, 0, 227, 118, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 60, 50, 38, 0, 39, 0, 40, 0, 62, 121, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 51, 50, 38, 0, 39, 0, 40, 0, 93, 121, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 55, 50, 38, 0, 39, 0, 40, 0, 109, 121, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 64, 50, 38, 0, 39, 0, 40, 0, 234, 129, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 66, 50, 38, 0, 39, 0, 40, 0, 243, 129, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 67, 50, 38, 0, 39, 0, 40, 0, 161, 140, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 54, 50, 38, 0, 39, 0, 40, 0, 199, 140, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 62, 50, 38, 0, 39, 0, 40, 0, 209, 145, 41, 0, 39, 0, 60, 0, 60, 0, 60, 0, 46, 50, 38, 0, 48, 0, 185, 112, 60, 0, 60, 0, 60, 0, 88, 51, 38, 0, 49, 0, 48, 0, 229, 101, 60, 0, 60, 0, 60, 0, 233, 51, 38, 0, 49, 0, 48, 0, 8, 103, 60, 0, 60, 0, 60, 0, 201, 50, 38, 0, 49, 0, 48, 0, 185, 112, 60, 0, 60, 0, 60, 0, 98, 51, 38, 0, 49, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 234, 51, 38, 0, 49, 0, 49, 0, 8, 103, 60, 0, 60, 0, 60, 0, 202, 50, 38, 0, 49, 0, 49, 0, 185, 112, 60, 0, 60, 0, 60, 0, 99, 51, 38, 0, 49, 0, 50, 0, 229, 101, 60, 0, 60, 0, 60, 0, 235, 51, 38, 0, 49, 0, 50, 0, 8, 103, 60, 0, 60, 0, 60, 0, 203, 50, 38, 0, 49, 0, 50, 0, 185, 112, 60, 0, 60, 0, 60, 0, 100, 51, 38, 0, 49, 0, 51, 0, 229, 101, 60, 0, 60, 0, 60, 0, 236, 51, 38, 0, 49, 0, 51, 0, 185, 112, 60, 0, 60, 0, 60, 0, 101, 51, 38, 0, 49, 0, 52, 0, 229, 101, 60, 0, 60, 0, 60, 0, 237, 51, 38, 0, 49, 0, 52, 0, 185, 112, 60, 0, 60, 0, 60, 0, 102, 51, 38, 0, 49, 0, 53, 0, 229, 101, 60, 0, 60, 0, 60, 0, 238, 51, 38, 0, 49, 0, 53, 0, 185, 112, 60, 0, 60, 0, 60, 0, 103, 51, 38, 0, 49, 0, 54, 0, 229, 101, 60, 0, 60, 0, 60, 0, 239, 51, 38, 0, 49, 0, 54, 0, 185, 112, 60, 0, 60, 0, 60, 0, 104, 51, 38, 0, 49, 0, 55, 0, 229, 101, 60, 0, 60, 0, 60, 0, 240, 51, 38, 0, 49, 0, 55, 0, 185, 112, 60, 0, 60, 0, 60, 0, 105, 51, 38, 0, 49, 0, 56, 0, 229, 101, 60, 0, 60, 0, 60, 0, 241, 51, 38, 0, 49, 0, 56, 0, 185, 112, 60, 0, 60, 0, 60, 0, 106, 51, 38, 0, 49, 0, 57, 0, 229, 101, 60, 0, 60, 0, 60, 0, 242, 51, 38, 0, 49, 0, 57, 0, 185, 112, 60, 0, 60, 0, 60, 0, 107, 51, 38, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 224, 51, 38, 0, 49, 0, 8, 103, 60, 0, 60, 0, 60, 0, 192, 50, 38, 0, 49, 0, 185, 112, 60, 0, 60, 0, 60, 0, 89, 51, 38, 0, 50, 0, 48, 0, 229, 101, 60, 0, 60, 0, 60, 0, 243, 51, 38, 0, 50, 0, 48, 0, 185, 112, 60, 0, 60, 0, 60, 0, 108, 51, 38, 0, 50, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 244, 51, 38, 0, 50, 0, 49, 0, 185, 112, 60, 0, 60, 0, 60, 0, 109, 51, 38, 0, 50, 0, 50, 0, 229, 101, 60, 0, 60, 0, 60, 0, 245, 51, 38, 0, 50, 0, 50, 0, 185, 112, 60, 0, 60, 0, 60, 0, 110, 51, 38, 0, 50, 0, 51, 0, 229, 101, 60, 0, 60, 0, 60, 0, 246, 51, 38, 0, 50, 0, 51, 0, 185, 112, 60, 0, 60, 0, 60, 0, 111, 51, 38, 0, 50, 0, 52, 0, 229, 101, 60, 0, 60, 0, 60, 0, 247, 51, 38, 0, 50, 0, 52, 0, 185, 112, 60, 0, 60, 0, 60, 0, 112, 51, 38, 0, 50, 0, 53, 0, 229, 101, 60, 0, 60, 0, 60, 0, 248, 51, 38, 0, 50, 0, 54, 0, 229, 101, 60, 0, 60, 0, 60, 0, 249, 51, 38, 0, 50, 0, 55, 0, 229, 101, 60, 0, 60, 0, 60, 0, 250, 51, 38, 0, 50, 0, 56, 0, 229, 101, 60, 0, 60, 0, 60, 0, 251, 51, 38, 0, 50, 0, 57, 0, 229, 101, 60, 0, 60, 0, 60, 0, 252, 51, 38, 0, 50, 0, 229, 101, 60, 0, 60, 0, 60, 0, 225, 51, 38, 0, 50, 0, 8, 103, 60, 0, 60, 0, 60, 0, 193, 50, 38, 0, 50, 0, 185, 112, 60, 0, 60, 0, 60, 0, 90, 51, 38, 0, 51, 0, 48, 0, 229, 101, 60, 0, 60, 0, 60, 0, 253, 51, 38, 0, 51, 0, 49, 0, 229, 101, 60, 0, 60, 0, 60, 0, 254, 51, 38, 0, 51, 0, 229, 101, 60, 0, 60, 0, 60, 0, 226, 51, 38, 0, 51, 0, 8, 103, 60, 0, 60, 0, 60, 0, 194, 50, 38, 0, 51, 0, 185, 112, 60, 0, 60, 0, 60, 0, 91, 51, 38, 0, 52, 0, 229, 101, 60, 0, 60, 0, 60, 0, 227, 51, 38, 0, 52, 0, 8, 103, 60, 0, 60, 0, 60, 0, 195, 50, 38, 0, 52, 0, 185, 112, 60, 0, 60, 0, 60, 0, 92, 51, 38, 0, 53, 0, 229, 101, 60, 0, 60, 0, 60, 0, 228, 51, 38, 0, 53, 0, 8, 103, 60, 0, 60, 0, 60, 0, 196, 50, 38, 0, 53, 0, 185, 112, 60, 0, 60, 0, 60, 0, 93, 51, 38, 0, 54, 0, 229, 101, 60, 0, 60, 0, 60, 0, 229, 51, 38, 0, 54, 0, 8, 103, 60, 0, 60, 0, 60, 0, 197, 50, 38, 0, 54, 0, 185, 112, 60, 0, 60, 0, 60, 0, 94, 51, 38, 0, 55, 0, 229, 101, 60, 0, 60, 0, 60, 0, 230, 51, 38, 0, 55, 0, 8, 103, 60, 0, 60, 0, 60, 0, 198, 50, 38, 0, 55, 0, 185, 112, 60, 0, 60, 0, 60, 0, 95, 51, 38, 0, 56, 0, 229, 101, 60, 0, 60, 0, 60, 0, 231, 51, 38, 0, 56, 0, 8, 103, 60, 0, 60, 0, 60, 0, 199, 50, 38, 0, 56, 0, 185, 112, 60, 0, 60, 0, 60, 0, 96, 51, 38, 0, 57, 0, 229, 101, 60, 0, 60, 0, 60, 0, 232, 51, 38, 0, 57, 0, 8, 103, 60, 0, 60, 0, 60, 0, 200, 50, 38, 0, 57, 0, 185, 112, 60, 0, 60, 0, 60, 0, 97, 51, 38, 0, 39, 0, 20, 48, 9, 78, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 65, 222, 38, 0, 39, 0, 20, 48, 140, 78, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 66, 222, 38, 0, 39, 0, 20, 48, 221, 82, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 71, 222, 38, 0, 39, 0, 20, 48, 137, 91, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 67, 222, 38, 0, 39, 0, 20, 48, 83, 98, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 69, 222, 38, 0, 39, 0, 20, 48, 87, 101, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 72, 222, 38, 0, 39, 0, 20, 48, 44, 103, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 64, 222, 38, 0, 39, 0, 20, 48, 185, 112, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 68, 222, 38, 0, 39, 0, 20, 48, 215, 118, 21, 48, 39, 0, 60, 0, 60, 0, 60, 0, 60, 216, 70, 222, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 0, 47, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 146, 49, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 128, 50, 38, 0, 0, 78, 60, 0, 60, 0, 60, 0, 60, 216, 41, 222, 38, 0, 1, 78, 60, 0, 60, 0, 60, 0, 156, 49, 38, 0, 3, 78, 60, 0, 60, 0, 60, 0, 134, 50, 38, 0, 9, 78, 60, 0, 60, 0, 60, 0, 148, 49, 38, 0, 9, 78, 60, 0, 60, 0, 60, 0, 130, 50, 38, 0, 9, 78, 60, 0, 60, 0, 60, 0, 60, 216, 42, 222, 38, 0, 10, 78, 60, 0, 60, 0, 60, 0, 150, 49, 38, 0, 10, 78, 60, 0, 60, 0, 60, 0, 164, 50, 38, 0, 11, 78, 60, 0, 60, 0, 60, 0, 152, 49, 38, 0, 11, 78, 60, 0, 60, 0, 60, 0, 166, 50, 38, 0, 25, 78, 60, 0, 60, 0, 60, 0, 155, 49, 38, 0, 40, 78, 60, 0, 60, 0, 60, 0, 1, 47, 38, 0, 45, 78, 60, 0, 60, 0, 60, 0, 151, 49, 38, 0, 45, 78, 60, 0, 60, 0, 60, 0, 165, 50, 38, 0, 45, 78, 60, 0, 60, 0, 60, 0, 60, 216, 45, 222, 38, 0, 54, 78, 60, 0, 60, 0, 60, 0, 2, 47, 38, 0, 63, 78, 60, 0, 60, 0, 60, 0, 3, 47, 38, 0, 89, 78, 60, 0, 60, 0, 60, 0, 4, 47, 38, 0, 89, 78, 60, 0, 60, 0, 60, 0, 154, 49, 38, 0, 93, 78, 60, 0, 60, 0, 60, 0, 136, 50, 38, 0, 133, 78, 60, 0, 60, 0, 60, 0, 5, 47, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 6, 47, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 147, 49, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 129, 50, 38, 0, 140, 78, 60, 0, 60, 0, 60, 0, 60, 216, 20, 222, 38, 0, 148, 78, 60, 0, 60, 0, 60, 0, 132, 50, 38, 0, 160, 78, 60, 0, 60, 0, 60, 0, 7, 47, 38, 0, 164, 78, 60, 0, 60, 0, 60, 0, 60, 216, 24, 222, 38, 0, 186, 78, 60, 0, 60, 0, 60, 0, 8, 47, 38, 0, 186, 78, 60, 0, 60, 0, 60, 0, 159, 49, 38, 0, 228, 78, 140, 84, 60, 0, 60, 0, 60, 0, 255, 50, 38, 0, 1, 79, 60, 0, 60, 0, 60, 0, 173, 50, 38, 0, 17, 79, 60, 0, 60, 0, 60, 0, 161, 50, 38, 0, 42, 81, 60, 0, 60, 0, 60, 0, 157, 50, 38, 0, 63, 81, 60, 0, 60, 0, 60, 0, 9, 47, 38, 0, 101, 81, 60, 0, 60, 0, 60, 0, 10, 47, 38, 0, 107, 81, 60, 0, 60, 0, 60, 0, 11, 47, 38, 0, 107, 81, 60, 0, 60, 0, 60, 0, 135, 50, 38, 0, 109, 81, 60, 0, 60, 0, 60, 0, 133, 50, 38, 0, 130, 81, 60, 0, 60, 0, 60, 0, 12, 47, 38, 0, 141, 81, 60, 0, 60, 0, 60, 0, 60, 216, 30, 222, 38, 0, 150, 81, 60, 0, 60, 0, 60, 0, 13, 47, 38, 0, 153, 81, 60, 0, 60, 0, 60, 0, 162, 50, 38, 0, 171, 81, 60, 0, 60, 0, 60, 0, 14, 47, 38, 0, 224, 81, 60, 0, 60, 0, 60, 0, 15, 47, 38, 0, 245, 81, 60, 0, 60, 0, 60, 0, 16, 47, 38, 0, 0, 82, 60, 0, 60, 0, 60, 0, 17, 47, 38, 0, 29, 82, 60, 0, 60, 0, 60, 0, 60, 216, 32, 222, 38, 0, 77, 82, 60, 0, 60, 0, 60, 0, 60, 216, 28, 222, 38, 0, 114, 82, 60, 0, 60, 0, 60, 0, 60, 216, 57, 222, 38, 0, 155, 82, 60, 0, 60, 0, 60, 0, 18, 47, 38, 0, 180, 82, 60, 0, 60, 0, 60, 0, 152, 50, 38, 0, 249, 82, 60, 0, 60, 0, 60, 0, 19, 47, 38, 0, 21, 83, 60, 0, 60, 0, 60, 0, 20, 47, 38, 0, 26, 83, 60, 0, 60, 0, 60, 0, 21, 47, 38, 0, 56, 83, 60, 0, 60, 0, 60, 0, 22, 47, 38, 0, 59, 83, 60, 0, 60, 0, 60, 0, 169, 50, 38, 0, 65, 83, 60, 0, 60, 0, 60, 0, 23, 47, 38, 0, 65, 83, 60, 0, 60, 0, 60, 0, 56, 48, 38, 0, 65, 83, 60, 0, 60, 0, 60, 0, 137, 50, 38, 0, 68, 83, 60, 0, 60, 0, 60, 0, 57, 48, 38, 0, 69, 83, 60, 0, 60, 0, 60, 0, 58, 48, 38, 0, 84, 83, 60, 0, 60, 0, 60, 0, 175, 50, 38, 0, 92, 83, 60, 0, 60, 0, 60, 0, 24, 47, 38, 0, 105, 83, 60, 0, 60, 0, 60, 0, 25, 47, 38, 0, 112, 83, 60, 0, 60, 0, 60, 0, 158, 50, 38, 0, 130, 83, 60, 0, 60, 0, 60, 0, 26, 47, 38, 0, 182, 83, 60, 0, 60, 0, 60, 0, 27, 47, 38, 0, 200, 83, 60, 0, 60, 0, 60, 0, 28, 47, 38, 0, 204, 83, 60, 0, 60, 0, 60, 0, 60, 216, 18, 222, 38, 0, 227, 83, 60, 0, 60, 0, 60, 0, 29, 47, 38, 0, 239, 83, 60, 0, 60, 0, 60, 0, 60, 216, 81, 222, 38, 0, 243, 83, 60, 0, 60, 0, 60, 0, 168, 50, 38, 0, 243, 83, 60, 0, 60, 0, 60, 0, 60, 216, 46, 222, 38, 0, 8, 84, 60, 0, 60, 0, 60, 0, 60, 216, 52, 222, 38, 0, 13, 84, 60, 0, 60, 0, 60, 0, 148, 50, 38, 0, 57, 84, 60, 0, 60, 0, 60, 0, 60, 216, 37, 222, 38, 0, 79, 85, 60, 0, 60, 0, 60, 0, 68, 50, 38, 0, 182, 85, 60, 0, 60, 0, 60, 0, 60, 216, 58, 222, 38, 0, 215, 86, 60, 0, 60, 0, 60, 0, 30, 47, 38, 0, 219, 86, 60, 0, 60, 0, 60, 0, 149, 49, 38, 0, 219, 86, 60, 0, 60, 0, 60, 0, 131, 50, 38, 0, 31, 87, 60, 0, 60, 0, 60, 0, 31, 47, 38, 0, 31, 87, 60, 0, 60, 0, 60, 0, 143, 50, 38, 0, 48, 87, 60, 0, 60, 0, 60, 0, 158, 49, 38, 0, 235, 88, 60, 0, 60, 0, 60, 0, 32, 47, 38, 0, 240, 88, 60, 0, 60, 0, 60, 0, 60, 216, 36, 222, 38, 0, 2, 89, 60, 0, 60, 0, 60, 0, 33, 47, 38, 0, 10, 89, 60, 0, 60, 0, 60, 0, 34, 47, 38, 0, 21, 89, 60, 0, 60, 0, 60, 0, 35, 47, 38, 0, 26, 89, 60, 0, 60, 0, 60, 0, 60, 216, 21, 222, 38, 0, 28, 89, 60, 0, 60, 0, 60, 0, 176, 50, 38, 0, 39, 89, 60, 0, 60, 0, 60, 0, 36, 47, 38, 0, 39, 89, 99, 107, 60, 0, 60, 0, 60, 0, 125, 51, 38, 0, 41, 89, 60, 0, 60, 0, 60, 0, 157, 49, 38, 0, 41, 89, 60, 0, 60, 0, 60, 0, 60, 216, 23, 222, 38, 0, 115, 89, 60, 0, 60, 0, 60, 0, 37, 47, 38, 0, 115, 89, 60, 0, 60, 0, 60, 0, 155, 50, 38, 0, 80, 91, 60, 0, 60, 0, 60, 0, 38, 47, 38, 0, 87, 91, 60, 0, 60, 0, 60, 0, 60, 216, 17, 222, 38, 0, 102, 91, 60, 0, 60, 0, 60, 0, 171, 50, 38, 0, 128, 91, 60, 0, 60, 0, 60, 0, 39, 47, 38, 0, 151, 91, 60, 0, 60, 0, 60, 0, 170, 50, 38, 0, 248, 91, 60, 0, 60, 0, 60, 0, 40, 47, 38, 0, 15, 92, 60, 0, 60, 0, 60, 0, 41, 47, 38, 0, 34, 92, 60, 0, 60, 0, 60, 0, 42, 47, 38, 0, 56, 92, 60, 0, 60, 0, 60, 0, 43, 47, 38, 0, 110, 92, 60, 0, 60, 0, 60, 0, 44, 47, 38, 0, 113, 92, 60, 0, 60, 0, 60, 0, 45, 47, 38, 0, 219, 93, 60, 0, 60, 0, 60, 0, 46, 47, 38, 0, 229, 93, 60, 0, 60, 0, 60, 0, 47, 47, 38, 0, 230, 93, 60, 0, 60, 0, 60, 0, 167, 50, 38, 0, 230, 93, 60, 0, 60, 0, 60, 0, 60, 216, 44, 222, 38, 0, 241, 93, 60, 0, 60, 0, 60, 0, 48, 47, 38, 0, 254, 93, 60, 0, 60, 0, 60, 0, 49, 47, 38, 0, 114, 94, 60, 0, 60, 0, 60, 0, 50, 47, 38, 0, 115, 94, 16, 98, 60, 0, 60, 0, 60, 0, 123, 51, 38, 0, 122, 94, 60, 0, 60, 0, 60, 0, 51, 47, 38, 0, 124, 94, 60, 0, 60, 0, 60, 0, 69, 50, 38, 0, 127, 94, 60, 0, 60, 0, 60, 0, 52, 47, 38, 0, 244, 94, 60, 0, 60, 0, 60, 0, 53, 47, 38, 0, 254, 94, 60, 0, 60, 0, 60, 0, 54, 47, 38, 0, 11, 95, 60, 0, 60, 0, 60, 0, 55, 47, 38, 0, 19, 95, 60, 0, 60, 0, 60, 0, 56, 47, 38, 0, 80, 95, 60, 0, 60, 0, 60, 0, 57, 47, 38, 0, 97, 95, 60, 0, 60, 0, 60, 0, 58, 47, 38, 0, 115, 95, 60, 0, 60, 0, 60, 0, 59, 47, 38, 0, 140, 95, 60, 0, 60, 0, 60, 0, 60, 216, 29, 222, 38, 0, 151, 95, 60, 0, 60, 0, 60, 0, 60, 216, 80, 222, 38, 0, 195, 95, 60, 0, 60, 0, 60, 0, 60, 47, 38, 0, 8, 98, 60, 0, 60, 0, 60, 0, 61, 47, 38, 0, 54, 98, 60, 0, 60, 0, 60, 0, 62, 47, 38, 0, 75, 98, 60, 0, 60, 0, 60, 0, 63, 47, 38, 0, 75, 98, 60, 0, 60, 0, 60, 0, 60, 216, 16, 222, 38, 0, 83, 98, 60, 0, 60, 0, 60, 0, 60, 216, 49, 222, 38, 0, 149, 98, 60, 0, 60, 0, 60, 0, 60, 216, 39, 222, 38, 0, 7, 99, 60, 0, 60, 0, 60, 0, 60, 216, 47, 222, 38, 0, 85, 99, 60, 0, 60, 0, 60, 0, 60, 216, 40, 222, 38, 0, 47, 101, 60, 0, 60, 0, 60, 0, 64, 47, 38, 0, 52, 101, 60, 0, 60, 0, 60, 0, 65, 47, 38, 0, 135, 101, 60, 0, 60, 0, 60, 0, 66, 47, 38, 0, 135, 101, 60, 0, 60, 0, 60, 0, 70, 50, 38, 0, 151, 101, 60, 0, 60, 0, 60, 0, 67, 47, 38, 0, 153, 101, 60, 0, 60, 0, 60, 0, 60, 216, 27, 222, 38, 0, 164, 101, 60, 0, 60, 0, 60, 0, 68, 47, 38, 0, 176, 101, 60, 0, 60, 0, 60, 0, 60, 216, 31, 222, 38, 0, 185, 101, 60, 0, 60, 0, 60, 0, 69, 47, 38, 0, 224, 101, 60, 0, 60, 0, 60, 0, 70, 47, 38, 0, 229, 101, 60, 0, 60, 0, 60, 0, 71, 47, 38, 0, 229, 101, 60, 0, 60, 0, 60, 0, 144, 50, 38, 0, 14, 102, 187, 108, 60, 0, 60, 0, 60, 0, 126, 51, 38, 0, 32, 102, 60, 0, 60, 0, 60, 0, 60, 216, 25, 222, 38, 0, 45, 102, 140, 84, 60, 0, 60, 0, 60, 0, 124, 51, 38, 0, 240, 102, 60, 0, 60, 0, 60, 0, 72, 47, 38, 0, 8, 103, 60, 0, 60, 0, 60, 0, 73, 47, 38, 0, 8, 103, 60, 0, 60, 0, 60, 0, 138, 50, 38, 0, 8, 103, 60, 0, 60, 0, 60, 0, 60, 216, 55, 222, 38, 0, 9, 103, 60, 0, 60, 0, 60, 0, 146, 50, 38, 0, 9, 103, 60, 0, 60, 0, 60, 0, 60, 216, 54, 222, 38, 0, 40, 103, 60, 0, 60, 0, 60, 0, 74, 47, 38, 0, 40, 103, 60, 0, 60, 0, 60, 0, 141, 50, 38, 0, 42, 104, 60, 0, 60, 0, 60, 0, 145, 50, 38, 0, 42, 104, 15, 95, 26, 79, 62, 121, 60, 0, 60, 0, 60, 0, 127, 51, 38, 0, 32, 107, 60, 0, 60, 0, 60, 0, 75, 47, 38, 0, 98, 107, 60, 0, 60, 0, 60, 0, 76, 47, 38, 0, 99, 107, 60, 0, 60, 0, 60, 0, 163, 50, 38, 0, 121, 107, 60, 0, 60, 0, 60, 0, 77, 47, 38, 0, 179, 107, 60, 0, 60, 0, 60, 0, 78, 47, 38, 0, 203, 107, 60, 0, 60, 0, 60, 0, 79, 47, 38, 0, 205, 107, 60, 0, 60, 0, 60, 0, 159, 46, 38, 0, 212, 107, 60, 0, 60, 0, 60, 0, 80, 47, 38, 0, 219, 107, 60, 0, 60, 0, 60, 0, 81, 47, 38, 0, 15, 108, 60, 0, 60, 0, 60, 0, 82, 47, 38, 0, 20, 108, 60, 0, 60, 0, 60, 0, 83, 47, 38, 0, 52, 108, 60, 0, 60, 0, 60, 0, 84, 47, 38, 0, 52, 108, 60, 0, 60, 0, 60, 0, 140, 50, 38, 0, 232, 108, 60, 0, 60, 0, 60, 0, 159, 50, 38, 0, 128, 110, 60, 0, 60, 0, 60, 0, 60, 216, 53, 222, 38, 0, 20, 111, 60, 0, 60, 0, 60, 0, 60, 216, 38, 222, 38, 0, 107, 112, 60, 0, 60, 0, 60, 0, 85, 47, 38, 0, 107, 112, 60, 0, 60, 0, 60, 0, 139, 50, 38, 0, 33, 113, 60, 0, 60, 0, 60, 0, 60, 216, 26, 222, 38, 0, 42, 114, 60, 0, 60, 0, 60, 0, 86, 47, 38, 0, 54, 114, 60, 0, 60, 0, 60, 0, 87, 47, 38, 0, 59, 114, 60, 0, 60, 0, 60, 0, 88, 47, 38, 0, 63, 114, 60, 0, 60, 0, 60, 0, 89, 47, 38, 0, 71, 114, 60, 0, 60, 0, 60, 0, 90, 47, 38, 0, 89, 114, 60, 0, 60, 0, 60, 0, 91, 47, 38, 0, 91, 114, 60, 0, 60, 0, 60, 0, 92, 47, 38, 0, 121, 114, 60, 0, 60, 0, 60, 0, 149, 50, 38, 0, 172, 114, 60, 0, 60, 0, 60, 0, 93, 47, 38, 0, 132, 115, 60, 0, 60, 0, 60, 0, 94, 47, 38, 0, 137, 115, 60, 0, 60, 0, 60, 0, 95, 47, 38, 0, 220, 116, 60, 0, 60, 0, 60, 0, 96, 47, 38, 0, 230, 116, 60, 0, 60, 0, 60, 0, 97, 47, 38, 0, 24, 117, 60, 0, 60, 0, 60, 0, 98, 47, 38, 0, 31, 117, 60, 0, 60, 0, 60, 0, 99, 47, 38, 0, 31, 117, 60, 0, 60, 0, 60, 0, 60, 216, 34, 222, 38, 0, 40, 117, 60, 0, 60, 0, 60, 0, 100, 47, 38, 0, 48, 117, 60, 0, 60, 0, 60, 0, 101, 47, 38, 0, 50, 117, 60, 0, 60, 0, 60, 0, 153, 49, 38, 0, 51, 117, 60, 0, 60, 0, 60, 0, 60, 216, 56, 222, 38, 0, 55, 117, 60, 0, 60, 0, 60, 0, 154, 50, 38, 0, 139, 117, 60, 0, 60, 0, 60, 0, 102, 47, 38, 0, 146, 117, 60, 0, 60, 0, 60, 0, 103, 47, 38, 0, 118, 118, 60, 0, 60, 0, 60, 0, 104, 47, 38, 0, 125, 118, 60, 0, 60, 0, 60, 0, 105, 47, 38, 0, 174, 118, 60, 0, 60, 0, 60, 0, 106, 47, 38, 0, 191, 118, 60, 0, 60, 0, 60, 0, 107, 47, 38, 0, 227, 118, 60, 0, 60, 0, 60, 0, 172, 50, 38, 0, 238, 118, 60, 0, 60, 0, 60, 0, 108, 47, 38, 0, 219, 119, 60, 0, 60, 0, 60, 0, 109, 47, 38, 0, 226, 119, 60, 0, 60, 0, 60, 0, 110, 47, 38, 0, 243, 119, 60, 0, 60, 0, 60, 0, 111, 47, 38, 0, 58, 121, 60, 0, 60, 0, 60, 0, 112, 47, 38, 0, 62, 121, 60, 0, 60, 0, 60, 0, 147, 50, 38, 0, 93, 121, 60, 0, 60, 0, 60, 0, 151, 50, 38, 0, 129, 121, 60, 0, 60, 0, 60, 0, 60, 216, 50, 222, 38, 0, 184, 121, 60, 0, 60, 0, 60, 0, 113, 47, 38, 0, 190, 121, 60, 0, 60, 0, 60, 0, 114, 47, 38, 0, 216, 121, 60, 0, 60, 0, 60, 0, 153, 50, 38, 0, 116, 122, 60, 0, 60, 0, 60, 0, 115, 47, 38, 0, 122, 122, 60, 0, 60, 0, 60, 0, 60, 216, 51, 222, 38, 0, 203, 122, 60, 0, 60, 0, 60, 0, 116, 47, 38, 0, 249, 122, 60, 0, 60, 0, 60, 0, 117, 47, 38, 0, 143, 123, 60, 0, 60, 0, 60, 0, 71, 50, 38, 0, 115, 124, 60, 0, 60, 0, 60, 0, 118, 47, 38, 0, 248, 124, 60, 0, 60, 0, 60, 0, 119, 47, 38, 0, 66, 125, 60, 0, 60, 0, 60, 0, 60, 216, 33, 222, 38, 0, 54, 127, 60, 0, 60, 0, 60, 0, 120, 47, 38, 0, 81, 127, 60, 0, 60, 0, 60, 0, 121, 47, 38, 0, 138, 127, 60, 0, 60, 0, 60, 0, 122, 47, 38, 0, 189, 127, 60, 0, 60, 0, 60, 0, 123, 47, 38, 0, 1, 128, 60, 0, 60, 0, 60, 0, 124, 47, 38, 0, 12, 128, 60, 0, 60, 0, 60, 0, 125, 47, 38, 0, 18, 128, 60, 0, 60, 0, 60, 0, 126, 47, 38, 0, 51, 128, 60, 0, 60, 0, 60, 0, 127, 47, 38, 0, 127, 128, 60, 0, 60, 0, 60, 0, 128, 47, 38, 0, 137, 128, 60, 0, 60, 0, 60, 0, 129, 47, 38, 0, 227, 129, 60, 0, 60, 0, 60, 0, 130, 47, 38, 0, 234, 129, 60, 0, 60, 0, 60, 0, 131, 47, 38, 0, 243, 129, 60, 0, 60, 0, 60, 0, 132, 47, 38, 0, 252, 129, 60, 0, 60, 0, 60, 0, 133, 47, 38, 0, 12, 130, 60, 0, 60, 0, 60, 0, 134, 47, 38, 0, 27, 130, 60, 0, 60, 0, 60, 0, 135, 47, 38, 0, 31, 130, 60, 0, 60, 0, 60, 0, 136, 47, 38, 0, 110, 130, 60, 0, 60, 0, 60, 0, 137, 47, 38, 0, 114, 130, 60, 0, 60, 0, 60, 0, 138, 47, 38, 0, 120, 130, 60, 0, 60, 0, 60, 0, 139, 47, 38, 0, 77, 134, 60, 0, 60, 0, 60, 0, 140, 47, 38, 0, 107, 134, 60, 0, 60, 0, 60, 0, 141, 47, 38, 0, 64, 136, 60, 0, 60, 0, 60, 0, 142, 47, 38, 0, 76, 136, 60, 0, 60, 0, 60, 0, 143, 47, 38, 0, 99, 136, 60, 0, 60, 0, 60, 0, 144, 47, 38, 0, 126, 137, 60, 0, 60, 0, 60, 0, 145, 47, 38, 0, 139, 137, 60, 0, 60, 0, 60, 0, 146, 47, 38, 0, 210, 137, 60, 0, 60, 0, 60, 0, 147, 47, 38, 0, 227, 137, 60, 0, 60, 0, 60, 0, 60, 216, 22, 222, 38, 0, 0, 138, 60, 0, 60, 0, 60, 0, 148, 47, 38, 0, 55, 140, 60, 0, 60, 0, 60, 0, 149, 47, 38, 0, 70, 140, 60, 0, 60, 0, 60, 0, 150, 47, 38, 0, 85, 140, 60, 0, 60, 0, 60, 0, 151, 47, 38, 0, 120, 140, 60, 0, 60, 0, 60, 0, 152, 47, 38, 0, 157, 140, 60, 0, 60, 0, 60, 0, 153, 47, 38, 0, 161, 140, 60, 0, 60, 0, 60, 0, 150, 50, 38, 0, 169, 140, 60, 0, 60, 0, 60, 0, 60, 216, 35, 222, 38, 0, 199, 140, 60, 0, 60, 0, 60, 0, 174, 50, 38, 0, 100, 141, 60, 0, 60, 0, 60, 0, 154, 47, 38, 0, 112, 141, 60, 0, 60, 0, 60, 0, 155, 47, 38, 0, 112, 141, 60, 0, 60, 0, 60, 0, 60, 216, 48, 222, 38, 0, 179, 141, 60, 0, 60, 0, 60, 0, 156, 47, 38, 0, 171, 142, 60, 0, 60, 0, 60, 0, 157, 47, 38, 0, 202, 142, 60, 0, 60, 0, 60, 0, 158, 47, 38, 0, 155, 143, 60, 0, 60, 0, 60, 0, 159, 47, 38, 0, 176, 143, 60, 0, 60, 0, 60, 0, 160, 47, 38, 0, 181, 143, 60, 0, 60, 0, 60, 0, 161, 47, 38, 0, 74, 144, 60, 0, 60, 0, 60, 0, 60, 216, 43, 222, 38, 0, 105, 144, 60, 0, 60, 0, 60, 0, 156, 50, 38, 0, 145, 144, 60, 0, 60, 0, 60, 0, 162, 47, 38, 0, 73, 145, 60, 0, 60, 0, 60, 0, 163, 47, 38, 0, 77, 145, 60, 0, 60, 0, 60, 0, 60, 216, 59, 222, 38, 0, 198, 145, 60, 0, 60, 0, 60, 0, 164, 47, 38, 0, 204, 145, 60, 0, 60, 0, 60, 0, 165, 47, 38, 0, 209, 145, 60, 0, 60, 0, 60, 0, 166, 47, 38, 0, 209, 145, 60, 0, 60, 0, 60, 0, 142, 50, 38, 0, 119, 149, 60, 0, 60, 0, 60, 0, 167, 47, 38, 0, 128, 149, 60, 0, 60, 0, 60, 0, 168, 47, 38, 0, 28, 150, 60, 0, 60, 0, 60, 0, 169, 47, 38, 0, 182, 150, 60, 0, 60, 0, 60, 0, 170, 47, 38, 0, 185, 150, 60, 0, 60, 0, 60, 0, 171, 47, 38, 0, 232, 150, 60, 0, 60, 0, 60, 0, 172, 47, 38, 0, 81, 151, 60, 0, 60, 0, 60, 0, 173, 47, 38, 0, 94, 151, 60, 0, 60, 0, 60, 0, 174, 47, 38, 0, 98, 151, 60, 0, 60, 0, 60, 0, 175, 47, 38, 0, 105, 151, 60, 0, 60, 0, 60, 0, 176, 47, 38, 0, 203, 151, 60, 0, 60, 0, 60, 0, 177, 47, 38, 0, 237, 151, 60, 0, 60, 0, 60, 0, 178, 47, 38, 0, 243, 151, 60, 0, 60, 0, 60, 0, 179, 47, 38, 0, 1, 152, 60, 0, 60, 0, 60, 0, 180, 47, 38, 0, 5, 152, 60, 0, 60, 0, 60, 0, 160, 50, 38, 0, 168, 152, 60, 0, 60, 0, 60, 0, 181, 47, 38, 0, 219, 152, 60, 0, 60, 0, 60, 0, 182, 47, 38, 0, 223, 152, 60, 0, 60, 0, 60, 0, 183, 47, 38, 0, 150, 153, 60, 0, 60, 0, 60, 0, 184, 47, 38, 0, 153, 153, 60, 0, 60, 0, 60, 0, 185, 47, 38, 0, 172, 153, 60, 0, 60, 0, 60, 0, 186, 47, 38, 0, 168, 154, 60, 0, 60, 0, 60, 0, 187, 47, 38, 0, 216, 154, 60, 0, 60, 0, 60, 0, 188, 47, 38, 0, 223, 154, 60, 0, 60, 0, 60, 0, 189, 47, 38, 0, 37, 155, 60, 0, 60, 0, 60, 0, 190, 47, 38, 0, 47, 155, 60, 0, 60, 0, 60, 0, 191, 47, 38, 0, 50, 155, 60, 0, 60, 0, 60, 0, 192, 47, 38, 0, 60, 155, 60, 0, 60, 0, 60, 0, 193, 47, 38, 0, 90, 155, 60, 0, 60, 0, 60, 0, 194, 47, 38, 0, 229, 156, 60, 0, 60, 0, 60, 0, 195, 47, 38, 0, 117, 158, 60, 0, 60, 0, 60, 0, 196, 47, 38, 0, 127, 158, 60, 0, 60, 0, 60, 0, 197, 47, 38, 0, 165, 158, 60, 0, 60, 0, 60, 0, 198, 47, 38, 0, 187, 158, 60, 0, 60, 0, 60, 0, 199, 47, 38, 0, 195, 158, 60, 0, 60, 0, 60, 0, 200, 47, 38, 0, 205, 158, 60, 0, 60, 0, 60, 0, 201, 47, 38, 0, 209, 158, 60, 0, 60, 0, 60, 0, 202, 47, 38, 0, 249, 158, 60, 0, 60, 0, 60, 0, 203, 47, 38, 0, 253, 158, 60, 0, 60, 0, 60, 0, 204, 47, 38, 0, 14, 159, 60, 0, 60, 0, 60, 0, 205, 47, 38, 0, 19, 159, 60, 0, 60, 0, 60, 0, 206, 47, 38, 0, 32, 159, 60, 0, 60, 0, 60, 0, 207, 47, 38, 0, 59, 159, 60, 0, 60, 0, 60, 0, 208, 47, 38, 0, 74, 159, 60, 0, 60, 0, 60, 0, 209, 47, 38, 0, 82, 159, 60, 0, 60, 0, 60, 0, 210, 47, 38, 0, 141, 159, 60, 0, 60, 0, 60, 0, 211, 47, 38, 0, 156, 159, 60, 0, 60, 0, 60, 0, 212, 47, 38, 0, 159, 159, 60, 0, 60, 0, 60, 0, 243, 46, 38, 0, 160, 159, 60, 0, 60, 0, 60, 0, 213, 47, 0, 0, 2, 0, 59, 0, 32, 0, 97, 0, 1, 0, 170, 170, 170, 170, 222, 137, 1, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 100, 0, 0, 0, 100, 1, 0, 0, 148, 106, 1, 0, 148, 106, 1, 0, 116, 114, 1, 0, 116, 114, 1, 0, 48, 119, 1, 0, 48, 119, 1, 0, 146, 133, 1, 0, 146, 133, 1, 0, 194, 137, 1, 0, 25, 0, 0, 0, 17, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 96, 127, 0, 0, 123, 99, 0, 118, 123, 127, 0, 0, 126, 224, 255, 0, 254, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 0, 251, 252, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 254, 255, 50, 105, 114, 84, 1, 0, 192, 8, 122, 21, 64, 0, 192, 0, 96, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 109, 0, 117, 0, 122, 0, 130, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 169, 0, 176, 0, 184, 0, 189, 0, 48, 0, 197, 0, 205, 0, 213, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 221, 0, 229, 0, 237, 0, 245, 0, 253, 0, 5, 1, 13, 1, 21, 1, 29, 1, 37, 1, 45, 1, 53, 1, 61, 1, 69, 1, 77, 1, 85, 1, 93, 1, 101, 1, 109, 1, 117, 1, 125, 1, 133, 1, 141, 1, 149, 1, 157, 1, 165, 1, 173, 1, 181, 1, 189, 1, 197, 1, 205, 1, 213, 1, 221, 1, 229, 1, 237, 1, 245, 1, 253, 1, 5, 2, 13, 2, 21, 2, 29, 2, 37, 2, 45, 2, 53, 2, 61, 2, 69, 2, 77, 2, 85, 2, 93, 2, 101, 2, 109, 2, 117, 2, 125, 2, 133, 2, 141, 2, 149, 2, 157, 2, 164, 2, 172, 2, 180, 2, 188, 2, 196, 2, 204, 2, 212, 2, 220, 2, 228, 2, 236, 2, 244, 2, 252, 2, 4, 3, 12, 3, 20, 3, 28, 3, 36, 3, 44, 3, 52, 3, 60, 3, 68, 3, 76, 3, 84, 3, 92, 3, 100, 3, 108, 3, 116, 3, 124, 3, 132, 3, 140, 3, 148, 3, 156, 3, 164, 3, 172, 3, 180, 3, 188, 3, 196, 3, 204, 3, 212, 3, 220, 3, 228, 3, 236, 3, 244, 3, 252, 3, 4, 4, 12, 4, 20, 4, 28, 4, 36, 4, 44, 4, 52, 4, 60, 4, 68, 4, 76, 4, 84, 4, 92, 4, 100, 4, 108, 4, 116, 4, 124, 4, 132, 4, 139, 4, 147, 4, 155, 4, 163, 4, 171, 4, 179, 4, 187, 4, 195, 4, 203, 4, 211, 4, 219, 4, 227, 4, 235, 4, 243, 4, 251, 4, 3, 5, 11, 5, 19, 5, 27, 5, 35, 5, 43, 5, 51, 5, 59, 5, 67, 5, 75, 5, 83, 5, 91, 5, 99, 5, 107, 5, 115, 5, 123, 5, 131, 5, 139, 5, 147, 5, 155, 5, 163, 5, 171, 5, 179, 5, 187, 5, 195, 5, 203, 5, 211, 5, 219, 5, 227, 5, 235, 5, 243, 5, 251, 5, 2, 6, 10, 6, 18, 6, 26, 6, 32, 6, 40, 6, 47, 6, 55, 6, 63, 6, 71, 6, 79, 6, 86, 6, 94, 6, 102, 6, 110, 6, 118, 6, 126, 6, 134, 6, 142, 6, 150, 6, 158, 6, 166, 6, 174, 6, 182, 6, 190, 6, 198, 6, 206, 6, 214, 6, 222, 6, 230, 6, 238, 6, 246, 6, 254, 6, 6, 7, 14, 7, 22, 7, 30, 7, 38, 7, 46, 7, 54, 7, 62, 7, 70, 7, 78, 7, 86, 7, 94, 7, 102, 7, 108, 7, 116, 7, 124, 7, 131, 7, 139, 7, 147, 7, 154, 7, 162, 7, 170, 7, 178, 7, 186, 7, 194, 7, 202, 7, 210, 7, 218, 7, 226, 7, 234, 7, 242, 7, 250, 7, 2, 8, 10, 8, 18, 8, 26, 8, 34, 8, 42, 8, 50, 8, 58, 8, 66, 8, 74, 8, 82, 8, 90, 8, 98, 8, 106, 8, 114, 8, 122, 8, 130, 8, 138, 8, 146, 8, 154, 8, 162, 8, 170, 8, 178, 8, 185, 8, 193, 8, 201, 8, 209, 8, 217, 8, 225, 8, 233, 8, 240, 8, 248, 8, 0, 9, 8, 9, 16, 9, 24, 9, 32, 9, 40, 9, 48, 9, 56, 9, 64, 9, 72, 9, 80, 9, 88, 9, 96, 9, 104, 9, 112, 9, 120, 9, 128, 9, 136, 9, 144, 9, 152, 9, 160, 9, 168, 9, 176, 9, 184, 9, 192, 9, 200, 9, 208, 9, 216, 9, 224, 9, 232, 9, 240, 9, 248, 9, 0, 10, 8, 10, 16, 10, 24, 10, 32, 10, 40, 10, 48, 10, 56, 10, 64, 10, 72, 10, 80, 10, 88, 10, 96, 10, 104, 10, 112, 10, 120, 10, 128, 10, 136, 10, 144, 10, 152, 10, 160, 10, 168, 10, 176, 10, 184, 10, 192, 10, 200, 10, 208, 10, 216, 10, 224, 10, 232, 10, 240, 10, 248, 10, 0, 11, 8, 11, 16, 11, 24, 11, 32, 11, 40, 11, 48, 11, 56, 11, 64, 11, 72, 11, 80, 11, 88, 11, 96, 11, 104, 11, 112, 11, 120, 11, 128, 11, 136, 11, 144, 11, 152, 11, 160, 11, 168, 11, 176, 11, 184, 11, 192, 11, 200, 11, 208, 11, 216, 11, 224, 11, 232, 11, 240, 11, 248, 11, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 40, 12, 48, 12, 56, 12, 64, 12, 72, 12, 80, 12, 88, 12, 96, 12, 104, 12, 112, 12, 120, 12, 128, 12, 136, 12, 144, 12, 152, 12, 160, 12, 168, 12, 176, 12, 184, 12, 192, 12, 200, 12, 208, 12, 216, 12, 224, 12, 232, 12, 240, 12, 48, 0, 48, 0, 48, 0, 48, 0, 248, 12, 0, 13, 8, 13, 16, 13, 24, 13, 32, 13, 40, 13, 48, 13, 56, 13, 64, 13, 72, 13, 80, 13, 88, 13, 96, 13, 104, 13, 112, 13, 120, 13, 128, 13, 136, 13, 144, 13, 152, 13, 160, 13, 168, 13, 176, 13, 184, 13, 192, 13, 200, 13, 208, 13, 216, 13, 224, 13, 232, 13, 240, 13, 248, 13, 0, 14, 6, 14, 14, 14, 22, 14, 30, 14, 38, 14, 46, 14, 54, 14, 62, 14, 70, 14, 78, 14, 86, 14, 94, 14, 101, 14, 109, 14, 117, 14, 125, 14, 133, 14, 141, 14, 149, 14, 157, 14, 165, 14, 173, 14, 181, 14, 189, 14, 197, 14, 205, 14, 213, 14, 221, 14, 229, 14, 237, 14, 245, 14, 253, 14, 5, 15, 13, 15, 21, 15, 29, 15, 37, 15, 45, 15, 53, 15, 61, 15, 69, 15, 77, 15, 85, 15, 93, 15, 101, 15, 109, 15, 117, 15, 125, 15, 133, 15, 141, 15, 149, 15, 157, 15, 165, 15, 173, 15, 181, 15, 189, 15, 197, 15, 205, 15, 213, 15, 221, 15, 229, 15, 237, 15, 245, 15, 253, 15, 5, 16, 13, 16, 48, 0, 48, 0, 48, 0, 48, 0, 21, 16, 29, 16, 37, 16, 45, 16, 53, 16, 61, 16, 69, 16, 77, 16, 48, 0, 48, 0, 84, 16, 92, 16, 100, 16, 108, 16, 116, 16, 124, 16, 132, 16, 140, 16, 148, 16, 156, 16, 164, 16, 172, 16, 180, 16, 188, 16, 196, 16, 204, 16, 212, 16, 214, 16, 222, 16, 230, 16, 238, 16, 246, 16, 254, 16, 6, 17, 14, 17, 22, 17, 30, 17, 38, 17, 46, 17, 54, 17, 62, 17, 70, 17, 78, 17, 86, 17, 94, 17, 102, 17, 110, 17, 118, 17, 126, 17, 134, 17, 142, 17, 150, 17, 158, 17, 166, 17, 174, 17, 182, 17, 190, 17, 198, 17, 206, 17, 214, 17, 222, 17, 230, 17, 238, 17, 246, 17, 254, 17, 6, 18, 14, 18, 22, 18, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 25, 18, 33, 18, 41, 18, 49, 18, 57, 18, 59, 18, 67, 18, 75, 18, 83, 18, 91, 18, 99, 18, 107, 18, 115, 18, 123, 18, 131, 18, 139, 18, 147, 18, 155, 18, 163, 18, 171, 18, 179, 18, 187, 18, 195, 18, 203, 18, 211, 18, 48, 0, 217, 18, 225, 18, 233, 18, 241, 18, 249, 18, 1, 19, 9, 19, 12, 19, 20, 19, 28, 19, 36, 19, 44, 19, 52, 19, 60, 19, 68, 19, 48, 0, 74, 19, 82, 19, 90, 19, 98, 19, 106, 19, 114, 19, 122, 19, 130, 19, 138, 19, 146, 19, 154, 19, 162, 19, 170, 19, 178, 19, 186, 19, 48, 0, 48, 0, 48, 0, 193, 19, 201, 19, 209, 19, 217, 19, 225, 19, 233, 19, 241, 19, 249, 19, 1, 20, 9, 20, 48, 0, 48, 0, 17, 20, 25, 20, 33, 20, 41, 20, 49, 20, 57, 20, 65, 20, 73, 20, 81, 20, 89, 20, 97, 20, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 105, 21, 161, 0, 48, 0, 48, 0, 153, 0, 153, 0, 153, 0, 113, 21, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 105, 20, 113, 20, 121, 20, 129, 20, 137, 20, 145, 20, 153, 20, 161, 20, 169, 20, 177, 20, 185, 20, 193, 20, 201, 20, 209, 20, 217, 20, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 180, 1, 232, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 128, 8, 225, 20, 233, 20, 241, 20, 249, 20, 1, 21, 9, 21, 17, 21, 25, 21, 33, 21, 41, 21, 49, 21, 57, 21, 65, 21, 73, 21, 81, 21, 89, 21, 97, 21, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 36, 0, 202, 193, 36, 0, 202, 226, 36, 0, 202, 3, 37, 0, 202, 36, 37, 0, 202, 69, 37, 0, 202, 102, 37, 0, 202, 135, 37, 0, 202, 168, 37, 0, 202, 201, 37, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 134, 2, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 38, 8, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 201, 6, 11, 0, 201, 70, 14, 0, 201, 134, 17, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 198, 20, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 70, 28, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 38, 31, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 198, 36, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 201, 166, 39, 0, 201, 230, 42, 0, 201, 38, 46, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 102, 49, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 134, 3, 0, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 35, 0, 196, 160, 156, 46, 198, 65, 0, 0, 134, 3, 0, 50, 201, 230, 56, 0, 196, 150, 156, 50, 198, 129, 0, 0, 134, 3, 0, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 198, 65, 1, 0, 134, 3, 0, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 198, 129, 1, 0, 134, 3, 0, 82, 196, 142, 156, 82, 201, 198, 59, 0, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 36, 0, 198, 193, 1, 0, 5, 3, 0, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 36, 0, 196, 160, 5, 46, 198, 1, 2, 0, 5, 3, 0, 50, 201, 102, 63, 0, 196, 150, 5, 50, 198, 65, 2, 0, 5, 3, 0, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 198, 1, 3, 0, 5, 3, 0, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 198, 65, 3, 0, 5, 3, 0, 82, 196, 142, 5, 82, 201, 70, 66, 0, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 134, 2, 0, 42, 5, 2, 0, 42, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 69, 0, 201, 134, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 50, 5, 2, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 75, 0, 201, 70, 78, 0, 198, 97, 0, 0, 198, 33, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 58, 5, 2, 0, 58, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 81, 0, 201, 6, 84, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 68, 5, 3, 0, 68, 201, 230, 86, 0, 201, 198, 89, 0, 198, 33, 1, 0, 198, 225, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 70, 5, 2, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 82, 5, 2, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 92, 0, 201, 230, 95, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 99, 0, 201, 6, 102, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 104, 0, 201, 38, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 198, 225, 1, 0, 198, 161, 0, 0, 198, 97, 2, 0, 198, 97, 1, 0, 198, 33, 3, 0, 198, 161, 1, 0, 198, 97, 3, 0, 134, 70, 0, 82, 5, 70, 0, 82, 134, 71, 0, 82, 5, 71, 0, 82, 134, 72, 0, 82, 5, 72, 0, 82, 134, 73, 0, 82, 5, 73, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 111, 0, 201, 70, 114, 0, 197, 98, 8, 0, 197, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 1, 0, 198, 193, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 117, 0, 201, 6, 120, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 122, 0, 201, 198, 125, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 201, 166, 128, 0, 201, 134, 131, 0, 201, 102, 134, 0, 201, 70, 137, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 140, 0, 201, 6, 143, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 66, 5, 3, 0, 66, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 145, 0, 201, 198, 148, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 151, 0, 201, 134, 154, 0, 201, 102, 157, 0, 201, 70, 160, 0, 201, 38, 163, 0, 201, 6, 166, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 197, 226, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 168, 0, 201, 38, 172, 0, 201, 102, 175, 0, 201, 166, 178, 0, 201, 230, 181, 0, 201, 38, 185, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 24, 0, 197, 226, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 188, 0, 201, 70, 191, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 194, 0, 201, 6, 197, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 50, 5, 71, 0, 50, 134, 73, 0, 50, 5, 73, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 199, 0, 201, 198, 202, 0, 201, 166, 205, 0, 201, 134, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 5, 0, 197, 34, 6, 0, 198, 130, 8, 0, 198, 2, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 211, 0, 201, 70, 214, 0, 201, 38, 217, 0, 201, 102, 220, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 6, 0, 197, 98, 7, 0, 198, 130, 9, 0, 198, 2, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 223, 0, 201, 230, 226, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 136, 5, 5, 0, 138, 192, 0, 0, 0, 5, 5, 0, 139, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 220, 5, 5, 0, 158, 5, 5, 0, 156, 5, 5, 0, 159, 5, 5, 0, 157, 5, 5, 12, 220, 5, 5, 4, 199, 5, 5, 20, 220, 5, 5, 18, 220, 5, 5, 3, 199, 5, 5, 2, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 199, 5, 5, 97, 199, 5, 5, 96, 199, 192, 0, 0, 0, 5, 5, 99, 199, 5, 5, 95, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 199, 5, 5, 223, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 160, 192, 0, 0, 0, 5, 5, 5, 199, 5, 5, 21, 220, 192, 0, 0, 0, 5, 5, 6, 199, 5, 5, 40, 220, 5, 5, 135, 200, 5, 5, 144, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 161, 5, 5, 7, 199, 192, 0, 0, 0, 5, 5, 100, 199, 5, 5, 41, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 220, 5, 5, 0, 140, 192, 0, 0, 0, 5, 5, 0, 163, 192, 0, 0, 0, 5, 5, 13, 220, 5, 5, 0, 164, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 199, 192, 0, 0, 0, 5, 5, 101, 199, 5, 5, 103, 199, 5, 5, 102, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 199, 5, 5, 225, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 201, 192, 0, 0, 0, 5, 5, 172, 204, 5, 5, 0, 137, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 220, 5, 5, 0, 141, 5, 5, 0, 166, 5, 5, 0, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 206, 5, 5, 219, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 210, 5, 5, 19, 232, 5, 5, 20, 232, 192, 0, 0, 0, 5, 5, 0, 142, 192, 0, 0, 0, 5, 5, 10, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 201, 5, 5, 0, 143, 5, 5, 14, 220, 5, 5, 0, 167, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 199, 5, 5, 13, 199, 5, 5, 22, 220, 5, 5, 14, 199, 5, 5, 12, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 199, 192, 0, 0, 0, 5, 5, 160, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 201, 5, 5, 27, 203, 192, 0, 0, 0, 5, 5, 0, 168, 5, 5, 15, 199, 192, 0, 0, 0, 5, 5, 228, 199, 5, 5, 230, 199, 5, 5, 229, 199, 192, 0, 0, 0, 5, 5, 136, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 201, 5, 5, 163, 201, 5, 5, 28, 203, 5, 5, 29, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 248, 5, 5, 0, 144, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 199, 5, 5, 16, 199, 5, 5, 23, 220, 5, 5, 18, 199, 5, 5, 24, 199, 192, 0, 0, 0, 5, 5, 19, 199, 5, 5, 20, 199, 5, 5, 25, 220, 5, 5, 24, 220, 5, 5, 22, 199, 5, 5, 23, 199, 192, 0, 0, 0, 5, 5, 21, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 199, 5, 5, 107, 199, 5, 5, 108, 199, 5, 5, 109, 199, 5, 5, 105, 199, 5, 5, 112, 199, 5, 5, 47, 220, 192, 0, 0, 0, 5, 5, 43, 220, 5, 5, 46, 220, 5, 5, 113, 199, 5, 5, 129, 199, 192, 0, 0, 0, 5, 5, 45, 220, 192, 0, 0, 0, 5, 5, 110, 199, 5, 5, 111, 199, 5, 5, 104, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 220, 5, 5, 44, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 199, 5, 5, 83, 220, 5, 5, 240, 199, 5, 5, 244, 199, 5, 5, 93, 220, 5, 5, 81, 220, 5, 5, 241, 199, 5, 5, 85, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 199, 192, 0, 0, 0, 5, 5, 245, 199, 192, 0, 0, 0, 5, 5, 231, 199, 5, 5, 84, 220, 5, 5, 246, 199, 5, 5, 88, 220, 192, 0, 0, 0, 5, 5, 92, 220, 5, 5, 89, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 220, 5, 5, 232, 199, 5, 5, 234, 199, 5, 5, 247, 199, 192, 0, 0, 0, 5, 5, 236, 199, 5, 5, 78, 220, 5, 5, 239, 199, 5, 5, 237, 199, 5, 5, 238, 199, 5, 5, 94, 220, 5, 5, 91, 220, 5, 5, 82, 220, 5, 5, 235, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 220, 5, 5, 233, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 220, 5, 5, 200, 220, 192, 0, 0, 0, 5, 5, 158, 200, 5, 5, 145, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 220, 5, 5, 142, 200, 192, 0, 0, 0, 5, 5, 160, 200, 192, 0, 0, 0, 5, 5, 150, 200, 192, 0, 0, 0, 5, 5, 149, 200, 5, 5, 189, 220, 5, 5, 153, 200, 5, 5, 148, 200, 5, 5, 194, 220, 5, 5, 202, 220, 192, 0, 0, 0, 5, 5, 198, 220, 192, 0, 0, 0, 5, 5, 151, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 200, 5, 5, 139, 200, 5, 5, 163, 200, 5, 5, 191, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 221, 5, 5, 137, 200, 5, 5, 159, 200, 5, 5, 138, 200, 5, 5, 146, 200, 5, 5, 147, 200, 5, 5, 196, 220, 5, 5, 192, 220, 5, 5, 152, 200, 5, 5, 144, 200, 5, 5, 188, 220, 5, 5, 140, 200, 5, 5, 199, 220, 5, 5, 161, 200, 5, 5, 164, 200, 5, 5, 143, 200, 5, 5, 156, 200, 5, 5, 162, 200, 5, 5, 141, 200, 5, 5, 197, 220, 5, 5, 157, 200, 5, 5, 203, 220, 5, 5, 190, 220, 5, 5, 155, 200, 5, 5, 193, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 220, 192, 0, 0, 0, 5, 5, 177, 201, 5, 5, 158, 221, 5, 5, 170, 221, 5, 5, 169, 201, 192, 0, 0, 0, 5, 5, 171, 221, 5, 5, 164, 201, 5, 5, 174, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 201, 5, 5, 152, 221, 5, 5, 175, 201, 5, 5, 151, 221, 5, 5, 155, 221, 5, 5, 162, 221, 5, 5, 160, 221, 5, 5, 183, 201, 5, 5, 178, 201, 5, 5, 146, 221, 5, 5, 148, 221, 5, 5, 180, 201, 5, 5, 168, 201, 5, 5, 149, 221, 5, 5, 161, 221, 5, 5, 168, 221, 5, 5, 173, 201, 5, 5, 154, 221, 5, 5, 147, 221, 5, 5, 172, 201, 5, 5, 150, 221, 5, 5, 176, 201, 5, 5, 153, 221, 192, 0, 0, 0, 5, 5, 171, 201, 192, 0, 0, 0, 5, 5, 166, 201, 192, 0, 0, 0, 5, 5, 181, 201, 5, 5, 163, 221, 5, 5, 182, 201, 5, 5, 167, 221, 192, 0, 0, 0, 5, 5, 165, 221, 5, 5, 169, 221, 5, 5, 179, 201, 5, 5, 157, 221, 5, 5, 145, 221, 192, 0, 0, 0, 5, 5, 159, 221, 5, 5, 170, 201, 5, 5, 164, 221, 5, 5, 165, 201, 5, 5, 166, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 203, 5, 5, 32, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 222, 5, 5, 5, 223, 192, 0, 0, 0, 5, 5, 31, 203, 5, 5, 39, 203, 5, 5, 51, 203, 192, 0, 0, 0, 5, 5, 11, 223, 5, 5, 9, 223, 5, 5, 4, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 203, 5, 5, 10, 223, 5, 5, 254, 222, 5, 5, 47, 203, 5, 5, 38, 203, 5, 5, 46, 203, 5, 5, 249, 222, 192, 0, 0, 0, 5, 5, 7, 223, 192, 0, 0, 0, 5, 5, 252, 222, 5, 5, 42, 203, 5, 5, 253, 222, 192, 0, 0, 0, 5, 5, 248, 222, 5, 5, 49, 203, 5, 5, 36, 203, 5, 5, 45, 203, 5, 5, 35, 203, 192, 0, 0, 0, 5, 5, 250, 222, 5, 5, 255, 222, 192, 0, 0, 0, 5, 5, 8, 223, 5, 5, 43, 203, 5, 5, 40, 203, 5, 5, 3, 223, 5, 5, 48, 203, 5, 5, 6, 223, 5, 5, 2, 223, 5, 5, 37, 203, 5, 5, 50, 203, 5, 5, 41, 203, 5, 5, 34, 203, 5, 5, 30, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 223, 192, 0, 0, 0, 5, 5, 199, 204, 5, 5, 177, 204, 192, 0, 0, 0, 5, 5, 193, 204, 192, 0, 0, 0, 5, 5, 198, 204, 5, 5, 198, 224, 5, 5, 197, 224, 5, 5, 202, 224, 5, 5, 203, 224, 5, 5, 180, 204, 192, 0, 0, 0, 5, 5, 189, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 204, 192, 0, 0, 0, 5, 5, 190, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 224, 5, 5, 183, 204, 5, 5, 192, 224, 192, 0, 0, 0, 5, 5, 204, 204, 192, 0, 0, 0, 5, 5, 195, 204, 5, 5, 174, 204, 5, 5, 175, 204, 5, 5, 211, 224, 5, 5, 138, 206, 192, 0, 0, 0, 5, 5, 188, 204, 5, 5, 187, 204, 5, 5, 193, 224, 5, 5, 191, 204, 5, 5, 235, 226, 5, 5, 182, 204, 5, 5, 204, 224, 5, 5, 197, 204, 5, 5, 196, 204, 5, 5, 186, 204, 5, 5, 196, 224, 5, 5, 205, 224, 192, 0, 0, 0, 5, 5, 190, 224, 5, 5, 185, 204, 5, 5, 206, 224, 5, 5, 194, 204, 5, 5, 194, 224, 5, 5, 176, 204, 192, 0, 0, 0, 5, 5, 179, 204, 5, 5, 178, 204, 5, 5, 207, 224, 5, 5, 192, 204, 5, 5, 181, 204, 5, 5, 201, 204, 5, 5, 203, 204, 5, 5, 201, 224, 5, 5, 200, 204, 192, 0, 0, 0, 5, 5, 209, 224, 5, 5, 195, 224, 5, 5, 210, 224, 192, 0, 0, 0, 5, 5, 199, 224, 192, 0, 0, 0, 5, 5, 208, 224, 192, 0, 0, 0, 5, 5, 200, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 226, 5, 5, 231, 226, 192, 0, 0, 0, 5, 5, 126, 206, 192, 0, 0, 0, 5, 5, 236, 226, 5, 5, 242, 226, 5, 5, 125, 206, 5, 5, 229, 226, 5, 5, 129, 206, 5, 5, 233, 226, 5, 5, 226, 226, 5, 5, 127, 206, 5, 5, 230, 226, 5, 5, 132, 206, 5, 5, 137, 206, 192, 0, 0, 0, 5, 5, 247, 226, 192, 0, 0, 0, 5, 5, 225, 226, 192, 0, 0, 0, 5, 5, 133, 206, 192, 0, 0, 0, 5, 5, 246, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 206, 5, 5, 232, 226, 5, 5, 124, 206, 5, 5, 227, 226, 5, 5, 223, 226, 5, 5, 237, 226, 5, 5, 224, 226, 5, 5, 222, 226, 5, 5, 234, 226, 5, 5, 240, 226, 5, 5, 241, 226, 5, 5, 130, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 229, 5, 5, 238, 226, 5, 5, 221, 226, 5, 5, 239, 226, 192, 0, 0, 0, 5, 5, 140, 206, 5, 5, 244, 226, 5, 5, 139, 206, 5, 5, 220, 226, 192, 0, 0, 0, 5, 5, 228, 226, 5, 5, 245, 226, 5, 5, 135, 206, 5, 5, 134, 206, 5, 5, 131, 206, 5, 5, 136, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 208, 192, 0, 0, 0, 5, 5, 104, 229, 5, 5, 97, 229, 192, 0, 0, 0, 5, 5, 107, 208, 192, 0, 0, 0, 5, 5, 105, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 229, 5, 5, 98, 229, 5, 5, 106, 208, 5, 5, 99, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 208, 5, 5, 103, 229, 192, 0, 0, 0, 5, 5, 93, 229, 5, 5, 92, 229, 5, 5, 111, 208, 192, 0, 0, 0, 5, 5, 112, 208, 5, 5, 108, 208, 5, 5, 113, 208, 5, 5, 91, 229, 5, 5, 102, 229, 5, 5, 100, 229, 5, 5, 94, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 208, 5, 5, 96, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 210, 5, 5, 69, 210, 5, 5, 25, 232, 5, 5, 78, 210, 5, 5, 31, 232, 5, 5, 34, 232, 5, 5, 71, 210, 5, 5, 72, 210, 5, 5, 28, 232, 5, 5, 70, 210, 5, 5, 37, 232, 5, 5, 76, 210, 5, 5, 38, 232, 192, 0, 0, 0, 5, 5, 33, 232, 5, 5, 77, 210, 192, 0, 0, 0, 5, 5, 22, 232, 5, 5, 74, 210, 5, 5, 23, 232, 192, 0, 0, 0, 5, 5, 32, 232, 5, 5, 30, 232, 192, 0, 0, 0, 5, 5, 26, 232, 5, 5, 73, 210, 5, 5, 24, 232, 5, 5, 79, 210, 5, 5, 29, 232, 5, 5, 36, 232, 5, 5, 27, 232, 5, 5, 35, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 212, 5, 5, 15, 212, 192, 0, 0, 0, 5, 5, 16, 212, 192, 0, 0, 0, 5, 5, 210, 234, 5, 5, 202, 234, 5, 5, 14, 212, 5, 5, 11, 212, 5, 5, 203, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 212, 5, 5, 206, 234, 192, 0, 0, 0, 5, 5, 208, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 234, 5, 5, 209, 234, 5, 5, 10, 212, 5, 5, 201, 234, 5, 5, 8, 212, 5, 5, 204, 234, 5, 5, 19, 212, 5, 5, 207, 234, 192, 0, 0, 0, 5, 5, 211, 234, 5, 5, 12, 212, 5, 5, 9, 212, 5, 5, 213, 234, 5, 5, 212, 234, 5, 5, 17, 212, 192, 0, 0, 0, 5, 5, 205, 234, 192, 0, 0, 0, 5, 5, 144, 213, 5, 5, 63, 237, 192, 0, 0, 0, 5, 5, 60, 237, 5, 5, 145, 213, 192, 0, 0, 0, 5, 5, 143, 213, 192, 0, 0, 0, 5, 5, 67, 237, 5, 5, 64, 237, 5, 5, 57, 237, 5, 5, 142, 213, 192, 0, 0, 0, 5, 5, 146, 213, 5, 5, 58, 237, 5, 5, 141, 213, 5, 5, 149, 213, 5, 5, 61, 237, 5, 5, 62, 237, 5, 5, 147, 213, 5, 5, 148, 213, 5, 5, 68, 237, 5, 5, 65, 237, 5, 5, 66, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 215, 5, 5, 195, 239, 5, 5, 26, 215, 5, 5, 192, 239, 5, 5, 28, 215, 5, 5, 30, 215, 192, 0, 0, 0, 5, 5, 193, 239, 5, 5, 27, 215, 192, 0, 0, 0, 5, 5, 194, 239, 192, 0, 0, 0, 5, 5, 191, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 216, 5, 5, 2, 242, 5, 5, 72, 216, 5, 5, 254, 241, 192, 0, 0, 0, 5, 5, 255, 241, 5, 5, 253, 241, 5, 5, 252, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 242, 5, 5, 70, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 243, 5, 5, 215, 243, 192, 0, 0, 0, 5, 5, 59, 237, 5, 5, 213, 243, 5, 5, 73, 216, 5, 5, 11, 218, 5, 5, 71, 245, 5, 5, 72, 245, 192, 0, 0, 0, 5, 5, 254, 218, 5, 5, 255, 218, 5, 5, 149, 247, 5, 5, 148, 247, 5, 5, 84, 219, 5, 5, 83, 219, 5, 5, 224, 248, 192, 0, 0, 0, 5, 5, 0, 145, 5, 5, 0, 169, 5, 5, 26, 199, 192, 0, 0, 0, 5, 5, 25, 199, 5, 5, 115, 199, 5, 5, 114, 199, 5, 5, 250, 199, 5, 5, 249, 199, 5, 5, 251, 199, 5, 5, 248, 199, 192, 0, 0, 0, 5, 5, 166, 200, 5, 5, 165, 200, 5, 5, 167, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 201, 192, 0, 0, 0, 5, 5, 184, 201, 5, 5, 186, 201, 192, 0, 0, 0, 5, 5, 52, 203, 192, 0, 0, 0, 5, 5, 0, 127, 5, 5, 212, 224, 5, 5, 0, 128, 5, 5, 141, 206, 5, 5, 0, 130, 5, 5, 0, 129, 5, 5, 106, 229, 192, 0, 0, 0, 5, 5, 0, 131, 5, 5, 20, 212, 5, 5, 0, 132, 192, 0, 0, 0, 5, 5, 0, 146, 192, 0, 0, 0, 5, 5, 27, 199, 5, 5, 252, 199, 5, 5, 187, 201, 192, 0, 0, 0, 5, 5, 0, 147, 5, 5, 30, 199, 5, 5, 28, 199, 5, 5, 29, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 200, 5, 5, 189, 201, 5, 5, 188, 201, 5, 5, 190, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 220, 192, 0, 0, 0, 5, 5, 116, 199, 5, 5, 117, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 199, 192, 0, 0, 0, 5, 5, 204, 220, 192, 0, 0, 0, 5, 5, 54, 203, 5, 5, 53, 203, 5, 5, 214, 224, 5, 5, 213, 224, 5, 5, 142, 206, 192, 0, 0, 0, 5, 5, 31, 199, 5, 5, 26, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 221, 192, 0, 0, 0, 5, 5, 55, 203, 192, 0, 0, 0, 5, 5, 208, 204, 192, 0, 0, 0, 5, 5, 206, 204, 5, 5, 207, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 215, 192, 0, 0, 0, 5, 5, 118, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 199, 5, 5, 95, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 200, 5, 5, 170, 200, 192, 0, 0, 0, 5, 5, 205, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 221, 5, 5, 191, 201, 5, 5, 174, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 224, 5, 5, 217, 224, 5, 5, 211, 204, 192, 0, 0, 0, 5, 5, 218, 224, 192, 0, 0, 0, 5, 5, 215, 224, 5, 5, 212, 204, 5, 5, 210, 204, 5, 5, 209, 204, 5, 5, 219, 224, 192, 0, 0, 0, 5, 5, 248, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 232, 5, 5, 216, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 213, 5, 5, 33, 215, 5, 5, 196, 239, 192, 0, 0, 0, 5, 5, 0, 148, 5, 5, 0, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 206, 5, 5, 115, 208, 192, 0, 0, 0, 5, 5, 21, 212, 192, 0, 0, 0, 5, 5, 8, 220, 5, 5, 32, 199, 192, 0, 0, 0, 5, 5, 121, 199, 5, 5, 119, 199, 5, 5, 120, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 149, 5, 5, 0, 150, 192, 0, 0, 0, 5, 5, 0, 170, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 199, 5, 5, 34, 199, 5, 5, 35, 199, 5, 5, 97, 220, 5, 5, 122, 199, 192, 0, 0, 0, 5, 5, 48, 220, 192, 0, 0, 0, 5, 5, 5, 200, 192, 0, 0, 0, 5, 5, 98, 220, 5, 5, 3, 200, 5, 5, 4, 200, 5, 5, 96, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 200, 5, 5, 2, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 220, 5, 5, 3, 203, 5, 5, 207, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 200, 5, 5, 171, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 200, 5, 5, 173, 200, 5, 5, 174, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 201, 192, 0, 0, 0, 5, 5, 197, 201, 5, 5, 179, 221, 5, 5, 176, 221, 5, 5, 177, 221, 192, 0, 0, 0, 5, 5, 175, 221, 5, 5, 199, 201, 5, 5, 195, 201, 5, 5, 194, 201, 192, 0, 0, 0, 5, 5, 196, 201, 5, 5, 193, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 201, 192, 0, 0, 0, 5, 5, 57, 203, 5, 5, 13, 223, 192, 0, 0, 0, 5, 5, 178, 221, 5, 5, 62, 203, 192, 0, 0, 0, 5, 5, 14, 223, 5, 5, 58, 203, 5, 5, 61, 203, 5, 5, 60, 203, 5, 5, 59, 203, 5, 5, 56, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 224, 192, 0, 0, 0, 5, 5, 215, 204, 5, 5, 225, 224, 5, 5, 213, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 224, 5, 5, 216, 204, 5, 5, 214, 204, 5, 5, 217, 204, 5, 5, 223, 224, 5, 5, 224, 224, 192, 0, 0, 0, 5, 5, 220, 224, 5, 5, 226, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 208, 5, 5, 144, 206, 5, 5, 249, 226, 5, 5, 251, 226, 5, 5, 250, 226, 5, 5, 252, 226, 5, 5, 145, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 208, 192, 0, 0, 0, 5, 5, 117, 208, 5, 5, 118, 208, 192, 0, 0, 0, 5, 5, 81, 210, 5, 5, 41, 232, 192, 0, 0, 0, 5, 5, 40, 232, 5, 5, 42, 232, 5, 5, 43, 232, 5, 5, 82, 210, 192, 0, 0, 0, 5, 5, 80, 210, 5, 5, 217, 234, 5, 5, 218, 234, 5, 5, 23, 212, 5, 5, 22, 212, 5, 5, 48, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 213, 5, 5, 152, 213, 5, 5, 153, 213, 5, 5, 155, 213, 5, 5, 69, 237, 5, 5, 70, 237, 5, 5, 154, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 215, 192, 0, 0, 0, 5, 5, 35, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 245, 5, 5, 151, 247, 5, 5, 150, 247, 5, 5, 225, 248, 192, 0, 0, 0, 5, 5, 0, 151, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 199, 5, 5, 123, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 200, 5, 5, 178, 200, 5, 5, 176, 200, 5, 5, 179, 200, 5, 5, 209, 220, 5, 5, 210, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 201, 5, 5, 180, 221, 192, 0, 0, 0, 5, 5, 201, 201, 192, 0, 0, 0, 5, 5, 15, 223, 5, 5, 66, 203, 5, 5, 16, 223, 5, 5, 65, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 203, 192, 0, 0, 0, 5, 5, 64, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 206, 5, 5, 254, 226, 192, 0, 0, 0, 5, 5, 149, 206, 5, 5, 253, 226, 5, 5, 79, 207, 5, 5, 148, 206, 5, 5, 147, 206, 192, 0, 0, 0, 5, 5, 122, 208, 192, 0, 0, 0, 5, 5, 121, 208, 5, 5, 120, 208, 5, 5, 83, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 210, 5, 5, 87, 210, 5, 5, 85, 210, 192, 0, 0, 0, 5, 5, 84, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 234, 192, 0, 0, 0, 5, 5, 220, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 237, 5, 5, 156, 213, 5, 5, 71, 237, 192, 0, 0, 0, 5, 5, 36, 215, 5, 5, 4, 242, 5, 5, 74, 216, 192, 0, 0, 0, 5, 5, 74, 245, 5, 5, 158, 218, 192, 0, 0, 0, 5, 5, 0, 171, 5, 5, 36, 199, 5, 5, 27, 220, 192, 0, 0, 0, 5, 5, 37, 199, 5, 5, 38, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 199, 5, 5, 126, 199, 192, 0, 0, 0, 5, 5, 8, 200, 5, 5, 211, 220, 5, 5, 181, 221, 5, 5, 182, 221, 192, 0, 0, 0, 5, 5, 67, 203, 5, 5, 228, 224, 5, 5, 151, 206, 5, 5, 150, 206, 5, 5, 109, 229, 5, 5, 108, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 152, 5, 5, 39, 199, 5, 5, 127, 199, 192, 0, 0, 0, 5, 5, 152, 206, 5, 5, 9, 220, 192, 0, 0, 0, 5, 5, 49, 220, 5, 5, 128, 199, 192, 0, 0, 0, 5, 5, 101, 220, 5, 5, 10, 200, 5, 5, 9, 200, 5, 5, 100, 220, 5, 5, 180, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 226, 192, 0, 0, 0, 5, 5, 88, 210, 5, 5, 221, 234, 5, 5, 24, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 246, 192, 0, 0, 0, 5, 5, 40, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 221, 5, 5, 17, 223, 5, 5, 155, 206, 5, 5, 153, 206, 5, 5, 154, 206, 5, 5, 0, 153, 192, 0, 0, 0, 5, 5, 0, 172, 192, 0, 0, 0, 5, 5, 43, 199, 192, 0, 0, 0, 5, 5, 42, 199, 5, 5, 41, 199, 5, 5, 131, 199, 5, 5, 130, 199, 192, 0, 0, 0, 5, 5, 50, 220, 5, 5, 102, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 201, 5, 5, 203, 201, 5, 5, 205, 201, 5, 5, 204, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 208, 192, 0, 0, 0, 5, 5, 0, 154, 192, 0, 0, 0, 5, 5, 44, 199, 192, 0, 0, 0, 5, 5, 133, 199, 5, 5, 132, 199, 192, 0, 0, 0, 5, 5, 212, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 220, 192, 0, 0, 0, 5, 5, 135, 199, 5, 5, 134, 199, 5, 5, 11, 200, 5, 5, 12, 200, 5, 5, 213, 220, 5, 5, 181, 200, 192, 0, 0, 0, 5, 5, 182, 200, 192, 0, 0, 0, 5, 5, 208, 201, 5, 5, 209, 201, 5, 5, 210, 201, 192, 0, 0, 0, 5, 5, 69, 203, 5, 5, 18, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 220, 192, 0, 0, 0, 5, 5, 45, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 220, 5, 5, 215, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 221, 192, 0, 0, 0, 5, 5, 185, 221, 192, 0, 0, 0, 5, 5, 20, 223, 5, 5, 19, 223, 5, 5, 22, 223, 5, 5, 21, 223, 5, 5, 70, 203, 192, 0, 0, 0, 5, 5, 2, 227, 5, 5, 221, 204, 5, 5, 229, 224, 5, 5, 220, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 229, 5, 5, 124, 208, 192, 0, 0, 0, 5, 5, 111, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 234, 5, 5, 25, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 213, 192, 0, 0, 0, 5, 5, 75, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 220, 192, 0, 0, 0, 5, 5, 136, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 155, 5, 5, 0, 173, 5, 5, 47, 199, 5, 5, 46, 199, 192, 0, 0, 0, 5, 5, 48, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 201, 192, 0, 0, 0, 5, 5, 211, 201, 5, 5, 213, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 204, 192, 0, 0, 0, 5, 5, 198, 239, 5, 5, 92, 217, 5, 5, 0, 174, 5, 5, 138, 199, 5, 5, 153, 199, 5, 5, 148, 199, 192, 0, 0, 0, 5, 5, 143, 199, 5, 5, 142, 199, 5, 5, 149, 199, 5, 5, 147, 199, 5, 5, 140, 199, 5, 5, 154, 199, 5, 5, 141, 199, 5, 5, 137, 199, 5, 5, 152, 199, 5, 5, 151, 199, 5, 5, 150, 199, 5, 5, 139, 199, 192, 0, 0, 0, 5, 5, 146, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 199, 5, 5, 144, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 200, 192, 0, 0, 0, 5, 5, 24, 200, 5, 5, 20, 200, 192, 0, 0, 0, 5, 5, 26, 200, 5, 5, 104, 220, 5, 5, 23, 200, 5, 5, 13, 200, 5, 5, 16, 200, 5, 5, 19, 200, 5, 5, 15, 200, 5, 5, 22, 200, 5, 5, 25, 200, 5, 5, 14, 200, 5, 5, 17, 200, 5, 5, 21, 200, 5, 5, 27, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 220, 5, 5, 221, 220, 192, 0, 0, 0, 5, 5, 195, 200, 5, 5, 222, 220, 5, 5, 183, 200, 5, 5, 185, 200, 5, 5, 209, 200, 5, 5, 204, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 220, 5, 5, 223, 220, 5, 5, 187, 200, 5, 5, 189, 200, 5, 5, 228, 220, 5, 5, 196, 200, 5, 5, 218, 220, 5, 5, 208, 200, 5, 5, 210, 200, 5, 5, 184, 200, 5, 5, 201, 200, 192, 0, 0, 0, 5, 5, 216, 220, 5, 5, 207, 200, 192, 0, 0, 0, 5, 5, 192, 200, 192, 0, 0, 0, 5, 5, 202, 200, 5, 5, 203, 200, 5, 5, 217, 220, 5, 5, 200, 200, 5, 5, 198, 200, 192, 0, 0, 0, 5, 5, 199, 200, 5, 5, 205, 200, 5, 5, 225, 220, 5, 5, 186, 200, 192, 0, 0, 0, 5, 5, 206, 200, 5, 5, 227, 220, 5, 5, 194, 200, 5, 5, 191, 200, 192, 0, 0, 0, 5, 5, 220, 220, 5, 5, 190, 200, 5, 5, 230, 220, 5, 5, 193, 200, 192, 0, 0, 0, 5, 5, 197, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 200, 5, 5, 226, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 221, 5, 5, 201, 221, 5, 5, 231, 201, 5, 5, 204, 221, 5, 5, 206, 221, 5, 5, 195, 221, 5, 5, 198, 221, 5, 5, 205, 221, 5, 5, 232, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 221, 5, 5, 196, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 221, 5, 5, 37, 223, 5, 5, 227, 201, 5, 5, 41, 223, 5, 5, 214, 201, 5, 5, 197, 221, 5, 5, 215, 201, 5, 5, 228, 201, 5, 5, 221, 201, 5, 5, 217, 201, 192, 0, 0, 0, 5, 5, 193, 221, 5, 5, 220, 201, 5, 5, 225, 201, 5, 5, 234, 201, 5, 5, 194, 221, 5, 5, 187, 221, 5, 5, 219, 201, 5, 5, 188, 221, 5, 5, 190, 221, 192, 0, 0, 0, 5, 5, 222, 201, 192, 0, 0, 0, 5, 5, 224, 201, 5, 5, 186, 221, 5, 5, 191, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 201, 5, 5, 229, 201, 5, 5, 199, 221, 5, 5, 235, 201, 192, 0, 0, 0, 5, 5, 226, 201, 5, 5, 189, 221, 5, 5, 223, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 201, 5, 5, 216, 201, 192, 0, 0, 0, 5, 5, 203, 221, 192, 0, 0, 0, 5, 5, 230, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 223, 5, 5, 24, 223, 5, 5, 39, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 223, 5, 5, 78, 203, 5, 5, 92, 203, 5, 5, 74, 203, 5, 5, 91, 203, 5, 5, 83, 203, 5, 5, 88, 203, 5, 5, 72, 203, 5, 5, 25, 223, 5, 5, 31, 223, 5, 5, 87, 203, 5, 5, 43, 223, 5, 5, 89, 203, 192, 0, 0, 0, 5, 5, 79, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 223, 5, 5, 30, 223, 5, 5, 77, 203, 192, 0, 0, 0, 5, 5, 23, 223, 5, 5, 90, 203, 5, 5, 38, 223, 5, 5, 82, 203, 5, 5, 40, 223, 5, 5, 93, 203, 5, 5, 73, 203, 5, 5, 84, 203, 5, 5, 81, 203, 5, 5, 28, 223, 5, 5, 85, 203, 5, 5, 34, 223, 5, 5, 35, 223, 5, 5, 80, 203, 5, 5, 86, 203, 5, 5, 76, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 203, 5, 5, 27, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 223, 192, 0, 0, 0, 5, 5, 249, 224, 5, 5, 235, 204, 5, 5, 231, 224, 192, 0, 0, 0, 5, 5, 236, 224, 5, 5, 228, 204, 5, 5, 239, 204, 5, 5, 234, 224, 5, 5, 223, 204, 5, 5, 233, 204, 5, 5, 238, 204, 5, 5, 241, 224, 192, 0, 0, 0, 5, 5, 234, 204, 5, 5, 237, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 224, 5, 5, 229, 204, 5, 5, 235, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 224, 5, 5, 248, 224, 192, 0, 0, 0, 5, 5, 231, 204, 5, 5, 246, 224, 5, 5, 227, 204, 5, 5, 242, 204, 192, 0, 0, 0, 5, 5, 238, 224, 192, 0, 0, 0, 5, 5, 225, 204, 192, 0, 0, 0, 5, 5, 251, 224, 5, 5, 239, 224, 5, 5, 243, 224, 5, 5, 230, 204, 5, 5, 241, 204, 5, 5, 240, 224, 5, 5, 236, 204, 5, 5, 245, 224, 5, 5, 252, 224, 5, 5, 17, 227, 192, 0, 0, 0, 5, 5, 250, 224, 5, 5, 243, 204, 5, 5, 224, 204, 5, 5, 242, 224, 5, 5, 233, 224, 192, 0, 0, 0, 5, 5, 232, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 224, 5, 5, 240, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 227, 192, 0, 0, 0, 5, 5, 175, 206, 5, 5, 23, 227, 5, 5, 173, 206, 5, 5, 170, 206, 5, 5, 14, 227, 5, 5, 166, 206, 5, 5, 18, 227, 5, 5, 177, 206, 5, 5, 8, 227, 5, 5, 13, 227, 5, 5, 12, 227, 5, 5, 226, 204, 5, 5, 172, 206, 5, 5, 21, 227, 192, 0, 0, 0, 5, 5, 24, 227, 5, 5, 5, 227, 192, 0, 0, 0, 5, 5, 142, 208, 192, 0, 0, 0, 5, 5, 25, 227, 5, 5, 178, 206, 192, 0, 0, 0, 5, 5, 164, 206, 5, 5, 161, 206, 5, 5, 16, 227, 5, 5, 158, 206, 192, 0, 0, 0, 5, 5, 22, 227, 192, 0, 0, 0, 5, 5, 165, 206, 5, 5, 26, 227, 192, 0, 0, 0, 5, 5, 6, 227, 5, 5, 20, 227, 5, 5, 168, 206, 5, 5, 7, 227, 5, 5, 10, 227, 5, 5, 15, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 206, 5, 5, 167, 206, 5, 5, 179, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 206, 192, 0, 0, 0, 5, 5, 162, 206, 5, 5, 60, 207, 192, 0, 0, 0, 5, 5, 163, 206, 5, 5, 11, 227, 5, 5, 176, 206, 5, 5, 171, 206, 5, 5, 19, 227, 5, 5, 160, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 227, 5, 5, 4, 227, 5, 5, 116, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 208, 5, 5, 128, 208, 5, 5, 127, 229, 5, 5, 148, 208, 5, 5, 130, 229, 5, 5, 126, 208, 5, 5, 123, 229, 5, 5, 132, 208, 5, 5, 138, 208, 5, 5, 159, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 208, 5, 5, 120, 229, 5, 5, 149, 208, 5, 5, 129, 208, 5, 5, 137, 208, 5, 5, 128, 229, 5, 5, 57, 232, 5, 5, 133, 229, 5, 5, 121, 229, 192, 0, 0, 0, 5, 5, 112, 229, 5, 5, 125, 229, 5, 5, 119, 229, 5, 5, 135, 208, 5, 5, 131, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 208, 5, 5, 151, 208, 5, 5, 144, 208, 192, 0, 0, 0, 5, 5, 133, 208, 5, 5, 130, 208, 192, 0, 0, 0, 5, 5, 141, 208, 192, 0, 0, 0, 5, 5, 132, 229, 5, 5, 118, 229, 5, 5, 124, 229, 5, 5, 126, 229, 5, 5, 114, 229, 5, 5, 129, 229, 5, 5, 127, 208, 5, 5, 113, 229, 192, 0, 0, 0, 5, 5, 134, 208, 5, 5, 150, 208, 5, 5, 146, 208, 5, 5, 115, 229, 5, 5, 140, 208, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 208, 5, 5, 143, 208, 5, 5, 139, 208, 192, 0, 0, 0, 5, 5, 122, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 232, 5, 5, 51, 232, 192, 0, 0, 0, 5, 5, 66, 232, 5, 5, 44, 232, 5, 5, 53, 232, 5, 5, 102, 210, 5, 5, 103, 210, 5, 5, 95, 210, 5, 5, 62, 232, 5, 5, 105, 210, 5, 5, 49, 232, 5, 5, 48, 232, 5, 5, 46, 232, 5, 5, 64, 232, 5, 5, 93, 210, 5, 5, 58, 232, 5, 5, 47, 232, 5, 5, 96, 210, 5, 5, 56, 232, 5, 5, 91, 210, 5, 5, 52, 232, 5, 5, 59, 232, 5, 5, 61, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 232, 5, 5, 100, 210, 5, 5, 45, 232, 5, 5, 94, 210, 5, 5, 50, 232, 192, 0, 0, 0, 5, 5, 89, 210, 192, 0, 0, 0, 5, 5, 101, 210, 5, 5, 60, 232, 5, 5, 97, 210, 5, 5, 98, 210, 5, 5, 104, 210, 5, 5, 92, 210, 5, 5, 0, 133, 5, 5, 90, 210, 5, 5, 54, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 212, 5, 5, 36, 212, 192, 0, 0, 0, 5, 5, 237, 234, 5, 5, 233, 234, 192, 0, 0, 0, 5, 5, 227, 234, 5, 5, 30, 212, 5, 5, 26, 212, 5, 5, 236, 234, 5, 5, 27, 212, 5, 5, 230, 234, 5, 5, 232, 234, 192, 0, 0, 0, 5, 5, 235, 234, 192, 0, 0, 0, 5, 5, 32, 212, 192, 0, 0, 0, 5, 5, 39, 212, 5, 5, 33, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 234, 5, 5, 34, 212, 5, 5, 35, 212, 5, 5, 228, 234, 5, 5, 40, 212, 192, 0, 0, 0, 5, 5, 226, 234, 5, 5, 231, 234, 5, 5, 31, 212, 5, 5, 224, 234, 5, 5, 37, 212, 5, 5, 29, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 212, 5, 5, 229, 234, 5, 5, 234, 234, 192, 0, 0, 0, 5, 5, 38, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 234, 192, 0, 0, 0, 5, 5, 164, 213, 5, 5, 89, 237, 192, 0, 0, 0, 5, 5, 86, 237, 192, 0, 0, 0, 5, 5, 158, 213, 5, 5, 170, 213, 5, 5, 171, 213, 192, 0, 0, 0, 5, 5, 161, 213, 5, 5, 84, 237, 5, 5, 163, 213, 5, 5, 76, 237, 5, 5, 169, 213, 192, 0, 0, 0, 5, 5, 88, 237, 5, 5, 160, 213, 5, 5, 90, 237, 5, 5, 159, 213, 192, 0, 0, 0, 5, 5, 85, 237, 5, 5, 87, 237, 5, 5, 162, 213, 5, 5, 83, 237, 5, 5, 77, 237, 5, 5, 74, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 229, 5, 5, 80, 237, 192, 0, 0, 0, 5, 5, 73, 237, 5, 5, 79, 237, 5, 5, 78, 237, 192, 0, 0, 0, 5, 5, 75, 237, 192, 0, 0, 0, 5, 5, 166, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 213, 5, 5, 81, 237, 5, 5, 37, 215, 5, 5, 82, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 239, 192, 0, 0, 0, 5, 5, 200, 239, 192, 0, 0, 0, 5, 5, 49, 215, 5, 5, 204, 239, 5, 5, 41, 215, 5, 5, 45, 215, 5, 5, 203, 239, 192, 0, 0, 0, 5, 5, 44, 215, 5, 5, 40, 215, 5, 5, 43, 215, 5, 5, 38, 215, 5, 5, 48, 215, 5, 5, 205, 239, 5, 5, 201, 239, 5, 5, 47, 215, 5, 5, 199, 239, 5, 5, 46, 215, 5, 5, 206, 239, 5, 5, 202, 239, 5, 5, 168, 213, 192, 0, 0, 0, 5, 5, 50, 215, 5, 5, 208, 239, 5, 5, 42, 215, 5, 5, 39, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 242, 5, 5, 13, 242, 5, 5, 76, 216, 5, 5, 14, 242, 5, 5, 12, 242, 5, 5, 10, 242, 5, 5, 9, 242, 5, 5, 78, 216, 5, 5, 8, 242, 5, 5, 79, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 242, 5, 5, 7, 242, 5, 5, 75, 216, 5, 5, 80, 216, 5, 5, 77, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 242, 192, 0, 0, 0, 5, 5, 93, 217, 192, 0, 0, 0, 5, 5, 218, 243, 5, 5, 216, 243, 5, 5, 221, 243, 5, 5, 219, 243, 192, 0, 0, 0, 5, 5, 217, 243, 5, 5, 220, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 218, 5, 5, 78, 245, 5, 5, 79, 245, 5, 5, 159, 218, 192, 0, 0, 0, 5, 5, 80, 245, 5, 5, 76, 245, 5, 5, 81, 245, 5, 5, 77, 245, 5, 5, 94, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 246, 5, 5, 152, 246, 5, 5, 162, 218, 5, 5, 151, 246, 5, 5, 161, 218, 5, 5, 160, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 218, 5, 5, 153, 247, 5, 5, 154, 247, 192, 0, 0, 0, 5, 5, 3, 219, 5, 5, 2, 219, 5, 5, 4, 219, 5, 5, 152, 247, 192, 0, 0, 0, 5, 5, 75, 248, 5, 5, 74, 248, 192, 0, 0, 0, 5, 5, 85, 219, 5, 5, 87, 219, 5, 5, 86, 219, 5, 5, 76, 248, 5, 5, 142, 219, 5, 5, 85, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 219, 192, 0, 0, 0, 5, 5, 84, 249, 5, 5, 157, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 199, 5, 5, 156, 199, 192, 0, 0, 0, 5, 5, 30, 200, 5, 5, 29, 200, 5, 5, 106, 220, 5, 5, 28, 200, 5, 5, 105, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 200, 5, 5, 233, 220, 192, 0, 0, 0, 5, 5, 232, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 200, 5, 5, 214, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 220, 192, 0, 0, 0, 5, 5, 212, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 221, 192, 0, 0, 0, 5, 5, 208, 221, 5, 5, 236, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 203, 192, 0, 0, 0, 5, 5, 253, 224, 5, 5, 254, 224, 5, 5, 244, 204, 5, 5, 245, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 227, 5, 5, 180, 206, 5, 5, 182, 206, 5, 5, 27, 227, 5, 5, 181, 206, 5, 5, 134, 229, 5, 5, 152, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 210, 5, 5, 107, 210, 5, 5, 67, 232, 192, 0, 0, 0, 5, 5, 43, 212, 192, 0, 0, 0, 5, 5, 42, 212, 192, 0, 0, 0, 5, 5, 91, 237, 5, 5, 210, 239, 5, 5, 209, 239, 192, 0, 0, 0, 5, 5, 217, 249, 5, 5, 0, 175, 5, 5, 30, 220, 192, 0, 0, 0, 5, 5, 51, 220, 5, 5, 52, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 200, 5, 5, 37, 200, 5, 5, 108, 220, 192, 0, 0, 0, 5, 5, 35, 200, 5, 5, 34, 200, 5, 5, 107, 220, 5, 5, 36, 200, 5, 5, 32, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 200, 5, 5, 109, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 200, 192, 0, 0, 0, 5, 5, 217, 200, 5, 5, 234, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 220, 192, 0, 0, 0, 5, 5, 219, 200, 192, 0, 0, 0, 5, 5, 237, 220, 5, 5, 215, 200, 5, 5, 238, 220, 5, 5, 236, 220, 5, 5, 218, 200, 5, 5, 220, 200, 5, 5, 223, 200, 5, 5, 222, 200, 5, 5, 216, 200, 5, 5, 239, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 201, 5, 5, 221, 221, 192, 0, 0, 0, 5, 5, 243, 201, 192, 0, 0, 0, 5, 5, 242, 201, 192, 0, 0, 0, 5, 5, 222, 221, 5, 5, 240, 201, 5, 5, 239, 201, 5, 5, 212, 221, 192, 0, 0, 0, 5, 5, 211, 221, 192, 0, 0, 0, 5, 5, 209, 221, 5, 5, 214, 221, 5, 5, 213, 221, 5, 5, 210, 221, 5, 5, 219, 221, 5, 5, 220, 221, 5, 5, 217, 221, 5, 5, 215, 221, 5, 5, 238, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 221, 5, 5, 244, 201, 5, 5, 223, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 221, 192, 0, 0, 0, 5, 5, 95, 203, 5, 5, 237, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 203, 5, 5, 48, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 203, 5, 5, 52, 223, 5, 5, 58, 223, 192, 0, 0, 0, 5, 5, 49, 223, 5, 5, 53, 223, 5, 5, 55, 223, 5, 5, 57, 223, 5, 5, 51, 223, 192, 0, 0, 0, 5, 5, 50, 223, 5, 5, 45, 223, 5, 5, 46, 223, 5, 5, 97, 203, 192, 0, 0, 0, 5, 5, 99, 203, 5, 5, 98, 203, 5, 5, 47, 223, 5, 5, 56, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 223, 5, 5, 10, 225, 192, 0, 0, 0, 5, 5, 9, 225, 5, 5, 14, 225, 5, 5, 5, 225, 192, 0, 0, 0, 5, 5, 8, 225, 5, 5, 7, 225, 192, 0, 0, 0, 5, 5, 11, 225, 192, 0, 0, 0, 5, 5, 15, 225, 5, 5, 246, 204, 5, 5, 249, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 225, 5, 5, 12, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 204, 5, 5, 255, 224, 192, 0, 0, 0, 5, 5, 100, 203, 5, 5, 42, 227, 5, 5, 13, 225, 192, 0, 0, 0, 5, 5, 4, 225, 192, 0, 0, 0, 5, 5, 247, 204, 5, 5, 3, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 206, 5, 5, 187, 206, 5, 5, 49, 227, 5, 5, 31, 227, 5, 5, 45, 227, 5, 5, 188, 206, 5, 5, 47, 227, 192, 0, 0, 0, 5, 5, 53, 227, 192, 0, 0, 0, 5, 5, 57, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 227, 5, 5, 36, 227, 5, 5, 44, 227, 192, 0, 0, 0, 5, 5, 58, 227, 5, 5, 56, 227, 5, 5, 46, 227, 5, 5, 41, 227, 5, 5, 34, 227, 5, 5, 142, 229, 5, 5, 32, 227, 5, 5, 192, 206, 5, 5, 39, 227, 5, 5, 193, 206, 5, 5, 189, 206, 5, 5, 29, 227, 5, 5, 51, 227, 5, 5, 37, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 227, 5, 5, 54, 227, 5, 5, 190, 206, 192, 0, 0, 0, 5, 5, 60, 227, 5, 5, 184, 206, 5, 5, 186, 206, 5, 5, 43, 227, 5, 5, 38, 227, 5, 5, 250, 204, 5, 5, 185, 206, 5, 5, 40, 227, 5, 5, 55, 227, 5, 5, 59, 227, 5, 5, 50, 227, 192, 0, 0, 0, 5, 5, 52, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 229, 192, 0, 0, 0, 5, 5, 146, 229, 5, 5, 145, 229, 5, 5, 160, 208, 5, 5, 138, 229, 192, 0, 0, 0, 5, 5, 161, 208, 5, 5, 159, 208, 192, 0, 0, 0, 5, 5, 140, 229, 5, 5, 156, 208, 5, 5, 144, 229, 192, 0, 0, 0, 5, 5, 139, 229, 5, 5, 141, 229, 5, 5, 135, 229, 5, 5, 154, 208, 192, 0, 0, 0, 5, 5, 154, 229, 5, 5, 153, 229, 5, 5, 150, 229, 5, 5, 153, 208, 5, 5, 157, 208, 5, 5, 158, 208, 5, 5, 2, 225, 5, 5, 147, 229, 5, 5, 155, 208, 5, 5, 191, 206, 5, 5, 149, 229, 5, 5, 136, 229, 5, 5, 152, 229, 5, 5, 151, 229, 192, 0, 0, 0, 5, 5, 155, 229, 192, 0, 0, 0, 5, 5, 81, 232, 192, 0, 0, 0, 5, 5, 148, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 229, 5, 5, 73, 232, 5, 5, 117, 210, 5, 5, 120, 210, 5, 5, 115, 210, 5, 5, 72, 232, 5, 5, 76, 232, 5, 5, 71, 232, 192, 0, 0, 0, 5, 5, 109, 210, 5, 5, 119, 210, 5, 5, 68, 232, 5, 5, 113, 210, 5, 5, 75, 232, 192, 0, 0, 0, 5, 5, 111, 210, 5, 5, 110, 210, 5, 5, 78, 232, 5, 5, 112, 210, 5, 5, 80, 232, 192, 0, 0, 0, 5, 5, 77, 232, 5, 5, 108, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 210, 5, 5, 82, 232, 5, 5, 70, 232, 5, 5, 79, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 210, 192, 0, 0, 0, 5, 5, 116, 210, 192, 0, 0, 0, 5, 5, 74, 232, 192, 0, 0, 0, 5, 5, 83, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 234, 5, 5, 44, 212, 5, 5, 251, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 212, 5, 5, 247, 234, 5, 5, 254, 234, 5, 5, 239, 234, 5, 5, 51, 212, 5, 5, 45, 212, 5, 5, 244, 234, 5, 5, 172, 213, 5, 5, 243, 234, 5, 5, 252, 234, 5, 5, 46, 212, 192, 0, 0, 0, 5, 5, 50, 212, 5, 5, 242, 234, 5, 5, 248, 234, 5, 5, 253, 234, 5, 5, 238, 234, 5, 5, 48, 212, 5, 5, 246, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 234, 5, 5, 2, 235, 5, 5, 240, 234, 5, 5, 249, 234, 192, 0, 0, 0, 5, 5, 47, 212, 5, 5, 255, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 213, 5, 5, 93, 237, 5, 5, 174, 213, 5, 5, 173, 213, 5, 5, 95, 237, 5, 5, 100, 237, 192, 0, 0, 0, 5, 5, 96, 237, 192, 0, 0, 0, 5, 5, 99, 237, 5, 5, 179, 213, 192, 0, 0, 0, 5, 5, 24, 215, 5, 5, 178, 213, 192, 0, 0, 0, 5, 5, 92, 237, 5, 5, 98, 237, 192, 0, 0, 0, 5, 5, 177, 213, 5, 5, 97, 237, 192, 0, 0, 0, 5, 5, 94, 237, 192, 0, 0, 0, 5, 5, 175, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 239, 192, 0, 0, 0, 5, 5, 217, 239, 5, 5, 212, 239, 5, 5, 52, 215, 5, 5, 214, 239, 192, 0, 0, 0, 5, 5, 51, 215, 5, 5, 216, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 215, 5, 5, 218, 239, 5, 5, 53, 215, 5, 5, 211, 239, 5, 5, 213, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 216, 5, 5, 17, 242, 192, 0, 0, 0, 5, 5, 83, 216, 5, 5, 18, 242, 5, 5, 82, 216, 5, 5, 16, 242, 5, 5, 81, 216, 5, 5, 15, 242, 192, 0, 0, 0, 5, 5, 96, 217, 5, 5, 95, 217, 5, 5, 82, 245, 5, 5, 84, 245, 192, 0, 0, 0, 5, 5, 83, 245, 5, 5, 13, 218, 5, 5, 14, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 218, 5, 5, 153, 246, 5, 5, 164, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 248, 5, 5, 226, 248, 5, 5, 184, 219, 192, 0, 0, 0, 5, 5, 0, 176, 5, 5, 49, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 208, 5, 5, 163, 208, 192, 0, 0, 0, 5, 5, 84, 232, 5, 5, 52, 212, 5, 5, 3, 235, 5, 5, 101, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 221, 5, 5, 60, 223, 5, 5, 16, 225, 5, 5, 251, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 245, 192, 0, 0, 0, 5, 5, 5, 219, 5, 5, 0, 177, 5, 5, 158, 199, 5, 5, 53, 220, 192, 0, 0, 0, 5, 5, 38, 200, 5, 5, 39, 200, 192, 0, 0, 0, 5, 5, 245, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 206, 192, 0, 0, 0, 5, 5, 54, 212, 192, 0, 0, 0, 5, 5, 55, 212, 5, 5, 53, 212, 192, 0, 0, 0, 5, 5, 0, 178, 192, 0, 0, 0, 5, 5, 50, 199, 5, 5, 52, 199, 5, 5, 51, 199, 5, 5, 32, 220, 5, 5, 53, 199, 5, 5, 159, 199, 5, 5, 54, 220, 192, 0, 0, 0, 5, 5, 160, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 200, 5, 5, 41, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 220, 192, 0, 0, 0, 5, 5, 226, 200, 192, 0, 0, 0, 5, 5, 241, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 201, 5, 5, 225, 221, 192, 0, 0, 0, 5, 5, 247, 201, 5, 5, 248, 201, 5, 5, 246, 201, 5, 5, 17, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 203, 5, 5, 105, 203, 5, 5, 107, 203, 5, 5, 104, 203, 192, 0, 0, 0, 5, 5, 61, 223, 5, 5, 250, 201, 5, 5, 103, 203, 192, 0, 0, 0, 5, 5, 252, 204, 5, 5, 253, 204, 192, 0, 0, 0, 5, 5, 254, 204, 192, 0, 0, 0, 5, 5, 61, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 208, 5, 5, 156, 229, 5, 5, 195, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 210, 192, 0, 0, 0, 5, 5, 57, 212, 5, 5, 56, 212, 5, 5, 4, 235, 192, 0, 0, 0, 5, 5, 180, 213, 5, 5, 55, 215, 192, 0, 0, 0, 5, 5, 222, 243, 5, 5, 77, 248, 5, 5, 228, 248, 5, 5, 0, 179, 5, 5, 161, 199, 192, 0, 0, 0, 5, 5, 162, 199, 5, 5, 116, 220, 5, 5, 43, 200, 5, 5, 46, 200, 192, 0, 0, 0, 5, 5, 114, 220, 5, 5, 112, 220, 5, 5, 45, 200, 5, 5, 115, 220, 5, 5, 117, 220, 5, 5, 111, 220, 5, 5, 48, 200, 5, 5, 47, 200, 5, 5, 44, 200, 5, 5, 42, 200, 5, 5, 113, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 200, 5, 5, 246, 220, 5, 5, 249, 220, 5, 5, 248, 220, 192, 0, 0, 0, 5, 5, 228, 200, 5, 5, 236, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 200, 5, 5, 245, 220, 5, 5, 243, 220, 5, 5, 232, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 200, 5, 5, 230, 200, 192, 0, 0, 0, 5, 5, 244, 220, 5, 5, 251, 220, 5, 5, 247, 220, 5, 5, 231, 200, 5, 5, 235, 200, 5, 5, 238, 200, 5, 5, 242, 220, 5, 5, 250, 220, 5, 5, 229, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 201, 5, 5, 9, 202, 192, 0, 0, 0, 5, 5, 237, 221, 5, 5, 230, 221, 5, 5, 10, 202, 5, 5, 241, 221, 5, 5, 226, 221, 5, 5, 233, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 201, 5, 5, 227, 221, 5, 5, 252, 201, 5, 5, 234, 221, 5, 5, 238, 221, 5, 5, 251, 201, 192, 0, 0, 0, 5, 5, 239, 221, 5, 5, 232, 221, 192, 0, 0, 0, 5, 5, 235, 221, 192, 0, 0, 0, 5, 5, 12, 202, 5, 5, 3, 202, 5, 5, 242, 221, 5, 5, 240, 221, 192, 0, 0, 0, 5, 5, 8, 202, 5, 5, 6, 202, 5, 5, 231, 221, 5, 5, 5, 202, 5, 5, 229, 221, 5, 5, 228, 221, 5, 5, 4, 202, 5, 5, 2, 202, 5, 5, 11, 202, 5, 5, 7, 202, 5, 5, 253, 201, 192, 0, 0, 0, 5, 5, 236, 221, 192, 0, 0, 0, 5, 5, 109, 203, 192, 0, 0, 0, 5, 5, 116, 203, 5, 5, 75, 223, 5, 5, 108, 203, 5, 5, 67, 223, 5, 5, 63, 223, 192, 0, 0, 0, 5, 5, 79, 223, 5, 5, 62, 223, 192, 0, 0, 0, 5, 5, 111, 203, 5, 5, 72, 223, 5, 5, 114, 203, 5, 5, 117, 203, 192, 0, 0, 0, 5, 5, 112, 203, 5, 5, 76, 223, 5, 5, 115, 203, 192, 0, 0, 0, 5, 5, 7, 205, 5, 5, 82, 223, 5, 5, 64, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 223, 5, 5, 73, 223, 5, 5, 77, 223, 5, 5, 81, 223, 5, 5, 78, 223, 5, 5, 71, 223, 5, 5, 74, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 223, 5, 5, 119, 203, 5, 5, 70, 223, 5, 5, 69, 223, 5, 5, 80, 223, 5, 5, 110, 203, 5, 5, 65, 223, 5, 5, 118, 203, 192, 0, 0, 0, 5, 5, 113, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 205, 5, 5, 25, 225, 192, 0, 0, 0, 5, 5, 12, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 225, 192, 0, 0, 0, 5, 5, 255, 204, 192, 0, 0, 0, 5, 5, 6, 205, 192, 0, 0, 0, 5, 5, 22, 225, 5, 5, 19, 225, 5, 5, 24, 225, 5, 5, 2, 205, 5, 5, 18, 225, 192, 0, 0, 0, 5, 5, 5, 205, 5, 5, 3, 205, 192, 0, 0, 0, 5, 5, 26, 225, 5, 5, 4, 205, 5, 5, 8, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 205, 192, 0, 0, 0, 5, 5, 11, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 225, 5, 5, 21, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 225, 192, 0, 0, 0, 5, 5, 68, 227, 5, 5, 196, 206, 5, 5, 170, 229, 5, 5, 67, 227, 5, 5, 86, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 206, 192, 0, 0, 0, 5, 5, 84, 227, 192, 0, 0, 0, 5, 5, 201, 206, 5, 5, 197, 206, 5, 5, 93, 227, 5, 5, 77, 227, 5, 5, 80, 227, 192, 0, 0, 0, 5, 5, 205, 206, 5, 5, 90, 227, 5, 5, 82, 227, 5, 5, 198, 206, 5, 5, 206, 206, 192, 0, 0, 0, 5, 5, 87, 227, 5, 5, 85, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 227, 5, 5, 91, 227, 5, 5, 79, 227, 5, 5, 74, 227, 192, 0, 0, 0, 5, 5, 64, 227, 5, 5, 92, 227, 5, 5, 76, 227, 5, 5, 63, 227, 192, 0, 0, 0, 5, 5, 204, 206, 5, 5, 81, 227, 5, 5, 94, 227, 5, 5, 78, 227, 5, 5, 66, 227, 5, 5, 71, 227, 5, 5, 62, 227, 192, 0, 0, 0, 5, 5, 203, 206, 192, 0, 0, 0, 5, 5, 75, 227, 5, 5, 72, 227, 5, 5, 199, 206, 5, 5, 65, 227, 192, 0, 0, 0, 5, 5, 89, 227, 5, 5, 200, 206, 192, 0, 0, 0, 5, 5, 73, 227, 5, 5, 69, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 208, 5, 5, 163, 229, 192, 0, 0, 0, 5, 5, 160, 229, 5, 5, 177, 229, 5, 5, 165, 229, 5, 5, 178, 229, 192, 0, 0, 0, 5, 5, 167, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 229, 5, 5, 171, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 229, 5, 5, 175, 229, 5, 5, 179, 229, 192, 0, 0, 0, 5, 5, 83, 227, 5, 5, 181, 229, 5, 5, 104, 232, 192, 0, 0, 0, 5, 5, 168, 208, 5, 5, 182, 229, 5, 5, 158, 229, 5, 5, 168, 229, 192, 0, 0, 0, 5, 5, 173, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 208, 5, 5, 169, 208, 5, 5, 180, 229, 5, 5, 183, 229, 5, 5, 162, 229, 5, 5, 159, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 229, 5, 5, 164, 229, 5, 5, 170, 208, 192, 0, 0, 0, 5, 5, 176, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 229, 192, 0, 0, 0, 5, 5, 169, 229, 5, 5, 157, 229, 5, 5, 92, 232, 5, 5, 90, 232, 5, 5, 130, 210, 5, 5, 128, 210, 5, 5, 100, 232, 5, 5, 91, 232, 5, 5, 101, 232, 5, 5, 97, 232, 5, 5, 89, 232, 5, 5, 103, 232, 5, 5, 88, 232, 5, 5, 95, 232, 5, 5, 127, 210, 5, 5, 126, 210, 5, 5, 125, 210, 5, 5, 93, 232, 5, 5, 98, 232, 5, 5, 122, 210, 5, 5, 129, 210, 192, 0, 0, 0, 5, 5, 86, 232, 192, 0, 0, 0, 5, 5, 96, 232, 5, 5, 85, 232, 5, 5, 94, 232, 5, 5, 123, 210, 5, 5, 99, 232, 5, 5, 87, 232, 5, 5, 124, 210, 5, 5, 102, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 235, 5, 5, 62, 212, 5, 5, 61, 212, 5, 5, 63, 212, 5, 5, 20, 235, 5, 5, 10, 235, 5, 5, 16, 235, 5, 5, 5, 235, 5, 5, 19, 235, 5, 5, 18, 235, 5, 5, 22, 235, 5, 5, 15, 235, 5, 5, 58, 212, 5, 5, 14, 235, 5, 5, 64, 212, 192, 0, 0, 0, 5, 5, 7, 235, 5, 5, 59, 212, 192, 0, 0, 0, 5, 5, 21, 235, 5, 5, 60, 212, 5, 5, 9, 235, 5, 5, 12, 235, 5, 5, 17, 235, 5, 5, 11, 235, 5, 5, 6, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 235, 5, 5, 103, 237, 5, 5, 184, 213, 5, 5, 106, 237, 5, 5, 105, 237, 5, 5, 108, 237, 5, 5, 110, 237, 192, 0, 0, 0, 5, 5, 182, 213, 192, 0, 0, 0, 5, 5, 104, 237, 192, 0, 0, 0, 5, 5, 102, 237, 192, 0, 0, 0, 5, 5, 111, 237, 5, 5, 109, 237, 5, 5, 107, 237, 192, 0, 0, 0, 5, 5, 113, 237, 192, 0, 0, 0, 5, 5, 112, 237, 5, 5, 186, 213, 5, 5, 181, 213, 192, 0, 0, 0, 5, 5, 183, 213, 5, 5, 185, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 237, 5, 5, 225, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 239, 5, 5, 223, 239, 192, 0, 0, 0, 5, 5, 226, 239, 5, 5, 219, 239, 192, 0, 0, 0, 5, 5, 220, 239, 5, 5, 228, 239, 5, 5, 221, 239, 192, 0, 0, 0, 5, 5, 56, 215, 5, 5, 230, 239, 192, 0, 0, 0, 5, 5, 229, 239, 5, 5, 222, 239, 192, 0, 0, 0, 5, 5, 22, 242, 5, 5, 87, 216, 5, 5, 20, 242, 5, 5, 25, 242, 5, 5, 24, 242, 5, 5, 227, 239, 192, 0, 0, 0, 5, 5, 86, 216, 192, 0, 0, 0, 5, 5, 23, 242, 5, 5, 19, 242, 5, 5, 27, 242, 5, 5, 26, 242, 5, 5, 85, 216, 192, 0, 0, 0, 5, 5, 21, 242, 192, 0, 0, 0, 5, 5, 57, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 243, 5, 5, 86, 245, 5, 5, 87, 245, 5, 5, 88, 245, 5, 5, 165, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 218, 192, 0, 0, 0, 5, 5, 154, 246, 192, 0, 0, 0, 5, 5, 156, 247, 5, 5, 155, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 248, 5, 5, 79, 248, 5, 5, 229, 248, 5, 5, 86, 249, 192, 0, 0, 0, 5, 5, 0, 180, 5, 5, 0, 181, 192, 0, 0, 0, 5, 5, 0, 182, 5, 5, 54, 199, 5, 5, 163, 199, 5, 5, 118, 220, 5, 5, 49, 200, 5, 5, 50, 200, 192, 0, 0, 0, 5, 5, 241, 200, 5, 5, 242, 200, 5, 5, 240, 200, 5, 5, 239, 200, 192, 0, 0, 0, 5, 5, 13, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 221, 5, 5, 15, 202, 5, 5, 14, 202, 5, 5, 244, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 203, 192, 0, 0, 0, 5, 5, 14, 205, 5, 5, 28, 225, 192, 0, 0, 0, 5, 5, 96, 227, 192, 0, 0, 0, 5, 5, 207, 206, 5, 5, 172, 208, 5, 5, 95, 227, 5, 5, 171, 208, 192, 0, 0, 0, 5, 5, 65, 212, 192, 0, 0, 0, 5, 5, 23, 235, 5, 5, 58, 215, 192, 0, 0, 0, 5, 5, 88, 216, 5, 5, 28, 242, 192, 0, 0, 0, 5, 5, 167, 218, 192, 0, 0, 0, 5, 5, 88, 219, 192, 0, 0, 0, 5, 5, 55, 220, 192, 0, 0, 0, 5, 5, 164, 199, 5, 5, 56, 220, 5, 5, 53, 200, 192, 0, 0, 0, 5, 5, 51, 200, 5, 5, 52, 200, 5, 5, 54, 200, 192, 0, 0, 0, 5, 5, 244, 200, 5, 5, 243, 200, 192, 0, 0, 0, 5, 5, 252, 220, 5, 5, 245, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 220, 5, 5, 245, 221, 192, 0, 0, 0, 5, 5, 246, 221, 192, 0, 0, 0, 5, 5, 16, 202, 5, 5, 18, 202, 5, 5, 20, 202, 5, 5, 17, 202, 5, 5, 21, 202, 5, 5, 19, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 203, 5, 5, 121, 203, 5, 5, 123, 203, 5, 5, 125, 203, 5, 5, 122, 203, 5, 5, 29, 225, 5, 5, 83, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 225, 5, 5, 30, 225, 5, 5, 20, 205, 192, 0, 0, 0, 5, 5, 16, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 205, 5, 5, 19, 205, 5, 5, 21, 205, 5, 5, 18, 205, 192, 0, 0, 0, 5, 5, 23, 205, 5, 5, 22, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 206, 5, 5, 98, 227, 5, 5, 97, 227, 5, 5, 211, 206, 192, 0, 0, 0, 5, 5, 210, 206, 5, 5, 209, 206, 5, 5, 213, 206, 5, 5, 208, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 229, 5, 5, 186, 229, 5, 5, 174, 208, 5, 5, 185, 229, 5, 5, 190, 229, 192, 0, 0, 0, 5, 5, 176, 208, 5, 5, 188, 229, 5, 5, 173, 208, 5, 5, 175, 208, 5, 5, 187, 229, 192, 0, 0, 0, 5, 5, 105, 232, 192, 0, 0, 0, 5, 5, 106, 232, 5, 5, 107, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 212, 5, 5, 74, 212, 5, 5, 24, 235, 5, 5, 68, 212, 5, 5, 72, 212, 5, 5, 25, 235, 5, 5, 73, 212, 5, 5, 69, 212, 5, 5, 70, 212, 5, 5, 67, 212, 5, 5, 71, 212, 5, 5, 189, 213, 5, 5, 184, 229, 5, 5, 190, 213, 5, 5, 188, 213, 192, 0, 0, 0, 5, 5, 187, 213, 5, 5, 231, 239, 5, 5, 59, 215, 5, 5, 29, 242, 5, 5, 30, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 218, 5, 5, 168, 218, 192, 0, 0, 0, 5, 5, 0, 183, 192, 0, 0, 0, 5, 5, 55, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 203, 192, 0, 0, 0, 5, 5, 32, 225, 5, 5, 24, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 206, 5, 5, 215, 206, 5, 5, 214, 206, 5, 5, 177, 208, 5, 5, 178, 208, 5, 5, 191, 229, 5, 5, 75, 212, 5, 5, 60, 215, 5, 5, 0, 184, 5, 5, 33, 220, 5, 5, 55, 199, 5, 5, 57, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 220, 5, 5, 56, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 185, 192, 0, 0, 0, 5, 5, 56, 199, 5, 5, 120, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 220, 192, 0, 0, 0, 5, 5, 255, 220, 192, 0, 0, 0, 5, 5, 246, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 229, 5, 5, 179, 208, 192, 0, 0, 0, 5, 5, 109, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 216, 5, 5, 0, 186, 5, 5, 9, 199, 5, 5, 57, 199, 5, 5, 58, 220, 5, 5, 165, 199, 192, 0, 0, 0, 5, 5, 250, 200, 5, 5, 249, 200, 5, 5, 247, 200, 5, 5, 248, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 221, 5, 5, 24, 202, 5, 5, 25, 202, 5, 5, 248, 221, 5, 5, 23, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 203, 5, 5, 84, 223, 5, 5, 129, 203, 5, 5, 127, 203, 5, 5, 128, 203, 5, 5, 27, 205, 5, 5, 25, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 225, 5, 5, 26, 205, 5, 5, 33, 225, 192, 0, 0, 0, 5, 5, 15, 205, 5, 5, 99, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 206, 5, 5, 219, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 206, 192, 0, 0, 0, 5, 5, 76, 212, 5, 5, 26, 235, 5, 5, 191, 213, 5, 5, 192, 213, 192, 0, 0, 0, 5, 5, 115, 237, 5, 5, 90, 216, 5, 5, 224, 243, 5, 5, 225, 243, 192, 0, 0, 0, 5, 5, 6, 219, 5, 5, 87, 249, 5, 5, 16, 220, 5, 5, 58, 199, 192, 0, 0, 0, 5, 5, 0, 187, 192, 0, 0, 0, 5, 5, 60, 220, 5, 5, 59, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 200, 5, 5, 122, 220, 5, 5, 123, 220, 5, 5, 121, 220, 192, 0, 0, 0, 5, 5, 124, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 221, 192, 0, 0, 0, 5, 5, 4, 221, 5, 5, 6, 221, 5, 5, 8, 221, 5, 5, 5, 221, 5, 5, 254, 200, 5, 5, 2, 221, 192, 0, 0, 0, 5, 5, 3, 221, 5, 5, 251, 200, 5, 5, 252, 200, 5, 5, 7, 221, 5, 5, 10, 221, 5, 5, 253, 200, 5, 5, 11, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 222, 192, 0, 0, 0, 5, 5, 2, 222, 5, 5, 251, 221, 5, 5, 27, 202, 5, 5, 5, 222, 5, 5, 3, 222, 5, 5, 250, 221, 5, 5, 9, 222, 5, 5, 12, 222, 5, 5, 7, 222, 5, 5, 254, 221, 5, 5, 29, 202, 5, 5, 6, 222, 5, 5, 30, 202, 5, 5, 255, 221, 5, 5, 4, 222, 5, 5, 249, 221, 5, 5, 253, 221, 5, 5, 11, 222, 5, 5, 31, 202, 192, 0, 0, 0, 5, 5, 32, 202, 192, 0, 0, 0, 5, 5, 252, 221, 5, 5, 10, 222, 5, 5, 26, 202, 5, 5, 28, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 223, 5, 5, 94, 223, 5, 5, 100, 223, 5, 5, 93, 223, 5, 5, 95, 223, 5, 5, 89, 223, 5, 5, 87, 223, 192, 0, 0, 0, 5, 5, 102, 223, 5, 5, 99, 223, 5, 5, 85, 223, 192, 0, 0, 0, 5, 5, 132, 203, 5, 5, 97, 223, 5, 5, 98, 223, 192, 0, 0, 0, 5, 5, 96, 223, 5, 5, 88, 223, 5, 5, 86, 223, 5, 5, 131, 203, 5, 5, 92, 223, 5, 5, 90, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 223, 5, 5, 103, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 205, 192, 0, 0, 0, 5, 5, 31, 205, 192, 0, 0, 0, 5, 5, 35, 225, 5, 5, 28, 205, 5, 5, 37, 225, 192, 0, 0, 0, 5, 5, 33, 205, 5, 5, 38, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 205, 192, 0, 0, 0, 5, 5, 34, 205, 5, 5, 39, 225, 5, 5, 104, 223, 5, 5, 41, 225, 192, 0, 0, 0, 5, 5, 30, 205, 192, 0, 0, 0, 5, 5, 29, 205, 192, 0, 0, 0, 5, 5, 36, 225, 5, 5, 40, 225, 5, 5, 35, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 206, 5, 5, 220, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 227, 5, 5, 105, 227, 5, 5, 107, 227, 5, 5, 222, 206, 5, 5, 110, 227, 192, 0, 0, 0, 5, 5, 226, 206, 5, 5, 112, 227, 192, 0, 0, 0, 5, 5, 228, 206, 192, 0, 0, 0, 5, 5, 224, 206, 5, 5, 232, 206, 192, 0, 0, 0, 5, 5, 229, 206, 5, 5, 103, 227, 5, 5, 223, 206, 192, 0, 0, 0, 5, 5, 102, 227, 5, 5, 100, 227, 5, 5, 114, 227, 5, 5, 104, 227, 192, 0, 0, 0, 5, 5, 225, 206, 5, 5, 113, 227, 5, 5, 230, 206, 5, 5, 109, 227, 5, 5, 108, 227, 5, 5, 231, 206, 5, 5, 106, 227, 5, 5, 227, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 227, 192, 0, 0, 0, 5, 5, 111, 227, 5, 5, 207, 229, 5, 5, 214, 229, 5, 5, 203, 229, 5, 5, 182, 208, 5, 5, 199, 229, 5, 5, 215, 229, 5, 5, 193, 229, 5, 5, 212, 229, 5, 5, 210, 229, 5, 5, 204, 229, 192, 0, 0, 0, 5, 5, 213, 229, 5, 5, 206, 229, 192, 0, 0, 0, 5, 5, 198, 229, 5, 5, 216, 229, 5, 5, 196, 229, 5, 5, 209, 229, 5, 5, 194, 229, 192, 0, 0, 0, 5, 5, 217, 229, 192, 0, 0, 0, 5, 5, 183, 208, 192, 0, 0, 0, 5, 5, 211, 229, 5, 5, 112, 232, 5, 5, 197, 229, 5, 5, 180, 208, 192, 0, 0, 0, 5, 5, 201, 229, 192, 0, 0, 0, 5, 5, 181, 208, 5, 5, 200, 229, 5, 5, 205, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 232, 5, 5, 111, 232, 192, 0, 0, 0, 5, 5, 113, 232, 192, 0, 0, 0, 5, 5, 118, 232, 5, 5, 117, 232, 5, 5, 131, 210, 192, 0, 0, 0, 5, 5, 195, 229, 5, 5, 115, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 210, 192, 0, 0, 0, 5, 5, 110, 232, 5, 5, 114, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 235, 192, 0, 0, 0, 5, 5, 40, 235, 5, 5, 31, 235, 192, 0, 0, 0, 5, 5, 38, 235, 5, 5, 29, 235, 5, 5, 37, 235, 5, 5, 41, 235, 5, 5, 28, 235, 5, 5, 32, 235, 5, 5, 27, 235, 192, 0, 0, 0, 5, 5, 77, 212, 192, 0, 0, 0, 5, 5, 30, 235, 5, 5, 78, 212, 5, 5, 36, 235, 5, 5, 35, 235, 5, 5, 34, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 237, 5, 5, 121, 237, 5, 5, 194, 213, 5, 5, 122, 237, 192, 0, 0, 0, 5, 5, 117, 237, 192, 0, 0, 0, 5, 5, 116, 237, 5, 5, 126, 237, 192, 0, 0, 0, 5, 5, 124, 237, 5, 5, 193, 213, 5, 5, 127, 237, 5, 5, 118, 237, 5, 5, 123, 237, 5, 5, 125, 237, 5, 5, 120, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 239, 5, 5, 240, 239, 5, 5, 234, 239, 5, 5, 239, 239, 192, 0, 0, 0, 5, 5, 232, 239, 5, 5, 242, 239, 5, 5, 238, 239, 5, 5, 243, 239, 5, 5, 237, 239, 5, 5, 233, 239, 5, 5, 241, 239, 192, 0, 0, 0, 5, 5, 244, 239, 5, 5, 236, 239, 192, 0, 0, 0, 5, 5, 31, 242, 5, 5, 94, 216, 192, 0, 0, 0, 5, 5, 92, 216, 192, 0, 0, 0, 5, 5, 91, 216, 5, 5, 93, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 243, 192, 0, 0, 0, 5, 5, 238, 244, 5, 5, 89, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 246, 5, 5, 156, 246, 192, 0, 0, 0, 5, 5, 169, 218, 192, 0, 0, 0, 5, 5, 157, 247, 192, 0, 0, 0, 5, 5, 7, 219, 192, 0, 0, 0, 5, 5, 158, 247, 192, 0, 0, 0, 5, 5, 81, 248, 5, 5, 90, 219, 192, 0, 0, 0, 5, 5, 89, 219, 5, 5, 80, 248, 5, 5, 143, 219, 192, 0, 0, 0, 5, 5, 230, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 188, 5, 5, 58, 200, 5, 5, 125, 220, 5, 5, 12, 221, 5, 5, 142, 201, 5, 5, 233, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 189, 5, 5, 168, 199, 5, 5, 167, 199, 5, 5, 166, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 205, 192, 0, 0, 0, 5, 5, 120, 232, 5, 5, 0, 190, 5, 5, 0, 191, 5, 5, 0, 192, 5, 5, 59, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 203, 192, 0, 0, 0, 5, 5, 105, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 208, 5, 5, 0, 193, 5, 5, 34, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 199, 5, 5, 170, 199, 5, 5, 61, 220, 192, 0, 0, 0, 5, 5, 59, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 221, 192, 0, 0, 0, 5, 5, 2, 201, 192, 0, 0, 0, 5, 5, 14, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 222, 5, 5, 36, 202, 5, 5, 35, 202, 5, 5, 13, 222, 5, 5, 33, 202, 5, 5, 15, 222, 5, 5, 34, 202, 5, 5, 37, 202, 192, 0, 0, 0, 5, 5, 134, 203, 192, 0, 0, 0, 5, 5, 136, 203, 5, 5, 109, 223, 5, 5, 106, 223, 5, 5, 107, 223, 5, 5, 108, 223, 5, 5, 110, 223, 5, 5, 135, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 225, 5, 5, 42, 225, 192, 0, 0, 0, 5, 5, 39, 205, 192, 0, 0, 0, 5, 5, 38, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 206, 5, 5, 117, 227, 192, 0, 0, 0, 5, 5, 235, 206, 5, 5, 237, 206, 5, 5, 234, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 208, 5, 5, 116, 227, 192, 0, 0, 0, 5, 5, 187, 208, 5, 5, 219, 229, 192, 0, 0, 0, 5, 5, 188, 208, 5, 5, 218, 229, 5, 5, 185, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 232, 5, 5, 125, 232, 5, 5, 133, 210, 5, 5, 124, 232, 5, 5, 122, 232, 5, 5, 121, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 235, 5, 5, 83, 212, 5, 5, 81, 212, 192, 0, 0, 0, 5, 5, 82, 212, 5, 5, 42, 235, 5, 5, 43, 235, 192, 0, 0, 0, 5, 5, 79, 212, 5, 5, 131, 237, 5, 5, 129, 237, 192, 0, 0, 0, 5, 5, 196, 213, 5, 5, 130, 237, 5, 5, 197, 213, 5, 5, 195, 213, 5, 5, 80, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 239, 5, 5, 245, 239, 5, 5, 246, 239, 5, 5, 128, 237, 5, 5, 33, 242, 5, 5, 95, 216, 5, 5, 32, 242, 5, 5, 227, 243, 5, 5, 228, 243, 5, 5, 248, 239, 5, 5, 90, 245, 192, 0, 0, 0, 5, 5, 0, 194, 5, 5, 171, 199, 5, 5, 61, 200, 5, 5, 126, 220, 5, 5, 60, 200, 192, 0, 0, 0, 5, 5, 39, 202, 5, 5, 134, 210, 192, 0, 0, 0, 5, 5, 60, 199, 5, 5, 172, 199, 5, 5, 137, 203, 5, 5, 189, 208, 192, 0, 0, 0, 5, 5, 62, 220, 192, 0, 0, 0, 5, 5, 63, 220, 192, 0, 0, 0, 5, 5, 127, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 201, 5, 5, 18, 221, 5, 5, 16, 221, 5, 5, 5, 201, 5, 5, 15, 221, 5, 5, 17, 221, 5, 5, 19, 221, 192, 0, 0, 0, 5, 5, 3, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 202, 5, 5, 44, 202, 5, 5, 41, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 202, 5, 5, 114, 223, 5, 5, 42, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 203, 192, 0, 0, 0, 5, 5, 113, 223, 5, 5, 115, 223, 5, 5, 112, 223, 5, 5, 116, 223, 5, 5, 139, 203, 5, 5, 42, 205, 5, 5, 44, 225, 192, 0, 0, 0, 5, 5, 46, 225, 5, 5, 40, 205, 5, 5, 47, 225, 5, 5, 41, 205, 5, 5, 45, 225, 192, 0, 0, 0, 5, 5, 111, 223, 5, 5, 118, 227, 5, 5, 121, 227, 5, 5, 122, 227, 5, 5, 119, 227, 5, 5, 241, 206, 5, 5, 240, 206, 5, 5, 238, 206, 5, 5, 239, 206, 5, 5, 120, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 208, 5, 5, 192, 208, 192, 0, 0, 0, 5, 5, 193, 208, 5, 5, 126, 232, 5, 5, 128, 232, 5, 5, 130, 232, 5, 5, 136, 210, 5, 5, 135, 210, 5, 5, 190, 208, 5, 5, 129, 232, 5, 5, 127, 232, 192, 0, 0, 0, 5, 5, 48, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 235, 5, 5, 52, 235, 5, 5, 84, 212, 5, 5, 53, 235, 5, 5, 50, 235, 5, 5, 85, 212, 5, 5, 47, 235, 5, 5, 45, 235, 5, 5, 51, 235, 5, 5, 199, 213, 5, 5, 133, 237, 5, 5, 49, 235, 5, 5, 201, 213, 5, 5, 134, 237, 5, 5, 200, 213, 5, 5, 203, 213, 5, 5, 135, 237, 5, 5, 198, 213, 5, 5, 202, 213, 192, 0, 0, 0, 5, 5, 253, 239, 5, 5, 251, 239, 5, 5, 250, 239, 5, 5, 252, 239, 5, 5, 249, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 218, 192, 0, 0, 0, 5, 5, 157, 246, 5, 5, 158, 246, 192, 0, 0, 0, 5, 5, 159, 247, 5, 5, 82, 248, 5, 5, 218, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 202, 5, 5, 6, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 195, 5, 5, 61, 199, 192, 0, 0, 0, 5, 5, 173, 199, 5, 5, 128, 220, 192, 0, 0, 0, 5, 5, 7, 201, 5, 5, 20, 221, 192, 0, 0, 0, 5, 5, 117, 223, 5, 5, 141, 203, 192, 0, 0, 0, 5, 5, 86, 212, 5, 5, 0, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 210, 5, 5, 0, 197, 5, 5, 62, 199, 5, 5, 63, 199, 192, 0, 0, 0, 5, 5, 175, 199, 5, 5, 174, 199, 192, 0, 0, 0, 5, 5, 129, 220, 5, 5, 63, 200, 192, 0, 0, 0, 5, 5, 21, 221, 192, 0, 0, 0, 5, 5, 8, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 222, 5, 5, 18, 222, 5, 5, 19, 222, 192, 0, 0, 0, 5, 5, 46, 202, 5, 5, 47, 202, 5, 5, 16, 222, 5, 5, 48, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 203, 5, 5, 118, 223, 192, 0, 0, 0, 5, 5, 49, 225, 5, 5, 43, 205, 192, 0, 0, 0, 5, 5, 48, 225, 192, 0, 0, 0, 5, 5, 243, 206, 5, 5, 123, 227, 5, 5, 244, 206, 5, 5, 124, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 235, 5, 5, 54, 235, 192, 0, 0, 0, 5, 5, 87, 212, 192, 0, 0, 0, 5, 5, 204, 213, 5, 5, 136, 237, 5, 5, 61, 215, 5, 5, 254, 239, 5, 5, 96, 216, 192, 0, 0, 0, 5, 5, 91, 219, 5, 5, 232, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 222, 192, 0, 0, 0, 5, 5, 119, 223, 5, 5, 245, 206, 5, 5, 220, 229, 5, 5, 138, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 201, 192, 0, 0, 0, 5, 5, 9, 201, 5, 5, 143, 203, 192, 0, 0, 0, 5, 5, 50, 225, 192, 0, 0, 0, 5, 5, 247, 206, 5, 5, 11, 208, 5, 5, 248, 206, 5, 5, 246, 206, 5, 5, 195, 208, 192, 0, 0, 0, 5, 5, 56, 235, 5, 5, 88, 212, 5, 5, 205, 213, 192, 0, 0, 0, 5, 5, 17, 220, 5, 5, 130, 220, 192, 0, 0, 0, 5, 5, 23, 221, 5, 5, 11, 201, 5, 5, 22, 221, 5, 5, 12, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 202, 5, 5, 23, 222, 5, 5, 22, 222, 5, 5, 51, 202, 5, 5, 49, 202, 5, 5, 50, 202, 5, 5, 21, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 203, 5, 5, 120, 223, 5, 5, 148, 203, 5, 5, 144, 203, 5, 5, 150, 203, 5, 5, 146, 203, 5, 5, 147, 203, 5, 5, 149, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 205, 5, 5, 45, 205, 5, 5, 44, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 227, 5, 5, 249, 206, 5, 5, 252, 206, 5, 5, 250, 206, 192, 0, 0, 0, 5, 5, 125, 227, 5, 5, 255, 206, 192, 0, 0, 0, 5, 5, 251, 206, 5, 5, 127, 227, 5, 5, 254, 206, 5, 5, 253, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 229, 5, 5, 221, 229, 192, 0, 0, 0, 5, 5, 198, 208, 5, 5, 196, 208, 5, 5, 197, 208, 5, 5, 223, 229, 5, 5, 139, 210, 5, 5, 133, 232, 5, 5, 140, 210, 5, 5, 132, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 213, 5, 5, 57, 235, 5, 5, 206, 213, 192, 0, 0, 0, 5, 5, 89, 212, 192, 0, 0, 0, 5, 5, 35, 242, 5, 5, 255, 239, 5, 5, 97, 216, 5, 5, 34, 242, 5, 5, 91, 245, 5, 5, 159, 246, 5, 5, 160, 246, 192, 0, 0, 0, 5, 5, 64, 199, 192, 0, 0, 0, 5, 5, 176, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 201, 5, 5, 16, 201, 192, 0, 0, 0, 5, 5, 133, 220, 5, 5, 26, 221, 5, 5, 25, 221, 5, 5, 24, 221, 192, 0, 0, 0, 5, 5, 132, 220, 5, 5, 131, 220, 5, 5, 65, 200, 5, 5, 15, 201, 5, 5, 13, 201, 5, 5, 64, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 202, 5, 5, 24, 222, 192, 0, 0, 0, 5, 5, 54, 202, 5, 5, 31, 221, 192, 0, 0, 0, 5, 5, 33, 221, 5, 5, 32, 221, 5, 5, 25, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 221, 192, 0, 0, 0, 5, 5, 20, 201, 5, 5, 18, 201, 192, 0, 0, 0, 5, 5, 27, 221, 5, 5, 29, 221, 5, 5, 35, 221, 192, 0, 0, 0, 5, 5, 17, 201, 192, 0, 0, 0, 5, 5, 30, 221, 5, 5, 39, 221, 5, 5, 56, 202, 192, 0, 0, 0, 5, 5, 36, 221, 5, 5, 19, 201, 192, 0, 0, 0, 5, 5, 34, 221, 5, 5, 37, 221, 192, 0, 0, 0, 5, 5, 55, 202, 192, 0, 0, 0, 5, 5, 57, 202, 5, 5, 38, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 222, 5, 5, 32, 222, 5, 5, 30, 222, 5, 5, 44, 222, 5, 5, 39, 222, 5, 5, 155, 203, 5, 5, 58, 202, 5, 5, 40, 222, 5, 5, 43, 222, 5, 5, 151, 203, 5, 5, 42, 222, 5, 5, 59, 202, 5, 5, 64, 202, 5, 5, 62, 202, 5, 5, 33, 222, 192, 0, 0, 0, 5, 5, 28, 222, 5, 5, 35, 222, 5, 5, 69, 202, 5, 5, 46, 222, 5, 5, 152, 203, 5, 5, 36, 222, 192, 0, 0, 0, 5, 5, 153, 203, 5, 5, 65, 202, 5, 5, 38, 222, 192, 0, 0, 0, 5, 5, 135, 223, 5, 5, 154, 203, 5, 5, 27, 222, 5, 5, 66, 202, 5, 5, 156, 203, 5, 5, 67, 202, 5, 5, 63, 202, 5, 5, 68, 202, 5, 5, 37, 222, 5, 5, 26, 222, 5, 5, 41, 222, 5, 5, 60, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 222, 5, 5, 34, 222, 5, 5, 31, 222, 5, 5, 61, 202, 192, 0, 0, 0, 5, 5, 121, 223, 192, 0, 0, 0, 5, 5, 122, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 223, 5, 5, 54, 225, 5, 5, 133, 223, 5, 5, 162, 203, 5, 5, 136, 223, 5, 5, 126, 223, 5, 5, 161, 203, 5, 5, 128, 223, 192, 0, 0, 0, 5, 5, 129, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 223, 5, 5, 157, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 223, 5, 5, 123, 223, 5, 5, 51, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 223, 5, 5, 47, 205, 5, 5, 52, 225, 5, 5, 130, 223, 192, 0, 0, 0, 5, 5, 51, 225, 5, 5, 125, 223, 5, 5, 134, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 203, 5, 5, 48, 205, 5, 5, 166, 203, 5, 5, 49, 205, 5, 5, 138, 223, 5, 5, 53, 225, 5, 5, 159, 203, 5, 5, 53, 205, 5, 5, 165, 203, 5, 5, 164, 203, 5, 5, 163, 203, 5, 5, 52, 205, 5, 5, 139, 223, 5, 5, 54, 205, 5, 5, 158, 203, 192, 0, 0, 0, 5, 5, 124, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 207, 5, 5, 57, 225, 5, 5, 59, 225, 192, 0, 0, 0, 5, 5, 61, 225, 5, 5, 55, 205, 5, 5, 61, 205, 5, 5, 130, 227, 5, 5, 65, 225, 5, 5, 56, 225, 5, 5, 4, 207, 5, 5, 128, 227, 192, 0, 0, 0, 5, 5, 60, 205, 5, 5, 58, 205, 5, 5, 67, 225, 192, 0, 0, 0, 5, 5, 129, 227, 192, 0, 0, 0, 5, 5, 58, 225, 192, 0, 0, 0, 5, 5, 59, 205, 5, 5, 62, 225, 5, 5, 62, 205, 5, 5, 64, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 205, 5, 5, 63, 225, 5, 5, 66, 225, 5, 5, 60, 225, 192, 0, 0, 0, 5, 5, 56, 205, 5, 5, 5, 207, 192, 0, 0, 0, 5, 5, 55, 225, 5, 5, 3, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 227, 5, 5, 141, 227, 5, 5, 201, 208, 192, 0, 0, 0, 5, 5, 8, 207, 5, 5, 13, 207, 5, 5, 202, 208, 5, 5, 143, 227, 5, 5, 20, 207, 5, 5, 225, 229, 5, 5, 133, 227, 5, 5, 12, 207, 5, 5, 15, 207, 5, 5, 10, 207, 5, 5, 131, 227, 5, 5, 145, 227, 5, 5, 148, 227, 5, 5, 240, 229, 192, 0, 0, 0, 5, 5, 146, 227, 5, 5, 229, 229, 5, 5, 11, 207, 5, 5, 18, 207, 5, 5, 22, 207, 5, 5, 140, 227, 5, 5, 224, 229, 5, 5, 144, 227, 5, 5, 7, 207, 5, 5, 226, 229, 5, 5, 147, 227, 5, 5, 228, 229, 5, 5, 136, 227, 192, 0, 0, 0, 5, 5, 199, 208, 192, 0, 0, 0, 5, 5, 134, 227, 5, 5, 135, 227, 5, 5, 17, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 207, 5, 5, 138, 227, 5, 5, 21, 207, 5, 5, 142, 227, 5, 5, 14, 207, 5, 5, 139, 227, 192, 0, 0, 0, 5, 5, 19, 207, 5, 5, 203, 208, 5, 5, 200, 208, 5, 5, 227, 229, 192, 0, 0, 0, 5, 5, 137, 227, 192, 0, 0, 0, 5, 5, 9, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 208, 5, 5, 212, 208, 5, 5, 231, 229, 5, 5, 145, 210, 5, 5, 210, 208, 5, 5, 235, 229, 5, 5, 214, 208, 5, 5, 134, 232, 5, 5, 237, 229, 5, 5, 147, 210, 5, 5, 206, 208, 5, 5, 209, 208, 5, 5, 238, 229, 192, 0, 0, 0, 5, 5, 239, 229, 5, 5, 245, 229, 5, 5, 216, 208, 5, 5, 148, 210, 192, 0, 0, 0, 5, 5, 241, 229, 5, 5, 246, 229, 5, 5, 234, 229, 5, 5, 158, 210, 192, 0, 0, 0, 5, 5, 149, 210, 5, 5, 215, 208, 5, 5, 232, 229, 5, 5, 247, 229, 192, 0, 0, 0, 5, 5, 157, 210, 5, 5, 213, 208, 5, 5, 142, 210, 5, 5, 244, 229, 192, 0, 0, 0, 5, 5, 217, 208, 5, 5, 236, 229, 5, 5, 230, 229, 5, 5, 207, 208, 5, 5, 233, 229, 192, 0, 0, 0, 5, 5, 242, 229, 192, 0, 0, 0, 5, 5, 141, 210, 5, 5, 146, 210, 5, 5, 204, 208, 5, 5, 243, 229, 192, 0, 0, 0, 5, 5, 144, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 210, 5, 5, 59, 235, 5, 5, 145, 232, 192, 0, 0, 0, 5, 5, 137, 232, 5, 5, 58, 235, 192, 0, 0, 0, 5, 5, 141, 232, 5, 5, 143, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 232, 192, 0, 0, 0, 5, 5, 155, 210, 192, 0, 0, 0, 5, 5, 139, 232, 5, 5, 159, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 210, 5, 5, 91, 212, 5, 5, 146, 232, 5, 5, 60, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 210, 5, 5, 138, 232, 5, 5, 142, 232, 5, 5, 90, 212, 5, 5, 143, 210, 5, 5, 135, 232, 5, 5, 136, 232, 5, 5, 92, 212, 5, 5, 151, 210, 5, 5, 153, 210, 5, 5, 150, 210, 5, 5, 144, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 235, 5, 5, 65, 235, 5, 5, 70, 235, 5, 5, 212, 213, 5, 5, 77, 235, 192, 0, 0, 0, 5, 5, 98, 212, 192, 0, 0, 0, 5, 5, 97, 212, 5, 5, 72, 235, 192, 0, 0, 0, 5, 5, 211, 213, 5, 5, 61, 235, 5, 5, 96, 212, 192, 0, 0, 0, 5, 5, 76, 235, 5, 5, 94, 212, 5, 5, 95, 212, 192, 0, 0, 0, 5, 5, 73, 235, 5, 5, 154, 237, 5, 5, 209, 213, 5, 5, 211, 208, 192, 0, 0, 0, 5, 5, 75, 235, 5, 5, 216, 213, 5, 5, 67, 235, 192, 0, 0, 0, 5, 5, 210, 213, 192, 0, 0, 0, 5, 5, 215, 213, 5, 5, 62, 235, 5, 5, 66, 235, 5, 5, 63, 235, 5, 5, 69, 235, 5, 5, 99, 212, 5, 5, 208, 213, 5, 5, 93, 212, 192, 0, 0, 0, 5, 5, 140, 237, 5, 5, 71, 235, 192, 0, 0, 0, 5, 5, 214, 213, 192, 0, 0, 0, 5, 5, 217, 213, 192, 0, 0, 0, 5, 5, 68, 235, 192, 0, 0, 0, 5, 5, 213, 213, 5, 5, 139, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 237, 5, 5, 65, 215, 5, 5, 138, 237, 5, 5, 13, 240, 5, 5, 153, 237, 5, 5, 221, 213, 192, 0, 0, 0, 5, 5, 219, 213, 5, 5, 63, 215, 5, 5, 150, 237, 5, 5, 146, 237, 5, 5, 225, 213, 192, 0, 0, 0, 5, 5, 4, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 213, 5, 5, 145, 237, 192, 0, 0, 0, 5, 5, 2, 240, 192, 0, 0, 0, 5, 5, 149, 237, 192, 0, 0, 0, 5, 5, 152, 237, 5, 5, 143, 237, 192, 0, 0, 0, 5, 5, 224, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 213, 5, 5, 3, 240, 5, 5, 64, 215, 5, 5, 151, 237, 5, 5, 220, 213, 5, 5, 222, 213, 5, 5, 148, 237, 5, 5, 226, 213, 5, 5, 147, 237, 5, 5, 142, 237, 5, 5, 141, 237, 5, 5, 62, 215, 5, 5, 155, 237, 5, 5, 6, 240, 5, 5, 37, 242, 5, 5, 67, 215, 192, 0, 0, 0, 5, 5, 12, 240, 192, 0, 0, 0, 5, 5, 10, 240, 192, 0, 0, 0, 5, 5, 38, 242, 192, 0, 0, 0, 5, 5, 68, 215, 5, 5, 11, 240, 192, 0, 0, 0, 5, 5, 8, 240, 5, 5, 99, 216, 5, 5, 36, 242, 192, 0, 0, 0, 5, 5, 5, 240, 5, 5, 7, 240, 5, 5, 100, 216, 5, 5, 70, 215, 5, 5, 98, 216, 5, 5, 69, 215, 5, 5, 102, 216, 5, 5, 9, 240, 5, 5, 66, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 243, 192, 0, 0, 0, 5, 5, 229, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 242, 5, 5, 230, 243, 5, 5, 40, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 217, 5, 5, 42, 242, 5, 5, 41, 242, 5, 5, 101, 216, 5, 5, 39, 242, 5, 5, 43, 242, 5, 5, 238, 243, 5, 5, 234, 243, 5, 5, 236, 243, 192, 0, 0, 0, 5, 5, 231, 243, 5, 5, 232, 243, 192, 0, 0, 0, 5, 5, 235, 243, 5, 5, 233, 243, 5, 5, 19, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 218, 5, 5, 21, 218, 5, 5, 20, 218, 5, 5, 170, 218, 5, 5, 161, 246, 5, 5, 171, 218, 5, 5, 92, 245, 5, 5, 8, 219, 5, 5, 160, 247, 5, 5, 9, 219, 5, 5, 92, 219, 5, 5, 144, 219, 5, 5, 233, 248, 192, 0, 0, 0, 5, 5, 234, 248, 5, 5, 235, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 250, 5, 5, 65, 199, 5, 5, 65, 220, 5, 5, 177, 199, 192, 0, 0, 0, 5, 5, 67, 200, 5, 5, 68, 200, 5, 5, 66, 200, 192, 0, 0, 0, 5, 5, 69, 200, 5, 5, 22, 201, 5, 5, 21, 201, 192, 0, 0, 0, 5, 5, 47, 222, 5, 5, 71, 202, 5, 5, 70, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 225, 5, 5, 23, 207, 5, 5, 24, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 208, 5, 5, 147, 232, 5, 5, 160, 210, 5, 5, 161, 210, 5, 5, 149, 232, 5, 5, 151, 232, 5, 5, 150, 232, 192, 0, 0, 0, 5, 5, 79, 235, 192, 0, 0, 0, 5, 5, 78, 235, 5, 5, 100, 212, 5, 5, 80, 235, 192, 0, 0, 0, 5, 5, 156, 237, 5, 5, 227, 213, 192, 0, 0, 0, 5, 5, 71, 215, 192, 0, 0, 0, 5, 5, 103, 216, 5, 5, 100, 217, 5, 5, 104, 216, 192, 0, 0, 0, 5, 5, 66, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 222, 5, 5, 73, 202, 5, 5, 72, 202, 5, 5, 74, 202, 5, 5, 167, 203, 5, 5, 140, 223, 5, 5, 141, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 225, 5, 5, 63, 205, 5, 5, 25, 207, 5, 5, 219, 208, 5, 5, 248, 229, 5, 5, 67, 199, 192, 0, 0, 0, 5, 5, 0, 198, 5, 5, 68, 199, 192, 0, 0, 0, 5, 5, 66, 220, 5, 5, 181, 199, 5, 5, 180, 199, 5, 5, 178, 199, 5, 5, 179, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 200, 5, 5, 140, 220, 5, 5, 142, 220, 5, 5, 71, 200, 5, 5, 134, 220, 192, 0, 0, 0, 5, 5, 135, 220, 192, 0, 0, 0, 5, 5, 141, 220, 5, 5, 137, 220, 5, 5, 139, 220, 5, 5, 70, 200, 5, 5, 136, 220, 5, 5, 143, 220, 5, 5, 138, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 201, 5, 5, 38, 201, 5, 5, 36, 201, 5, 5, 50, 221, 5, 5, 47, 221, 5, 5, 55, 221, 5, 5, 34, 201, 5, 5, 56, 221, 192, 0, 0, 0, 5, 5, 27, 201, 5, 5, 53, 221, 192, 0, 0, 0, 5, 5, 33, 201, 5, 5, 49, 221, 5, 5, 48, 221, 5, 5, 31, 201, 5, 5, 54, 221, 5, 5, 32, 201, 5, 5, 75, 202, 5, 5, 26, 201, 5, 5, 51, 221, 192, 0, 0, 0, 5, 5, 41, 221, 5, 5, 23, 201, 192, 0, 0, 0, 5, 5, 42, 201, 5, 5, 46, 221, 5, 5, 52, 221, 5, 5, 28, 201, 5, 5, 30, 201, 192, 0, 0, 0, 5, 5, 42, 221, 192, 0, 0, 0, 5, 5, 43, 221, 5, 5, 44, 221, 192, 0, 0, 0, 5, 5, 41, 201, 5, 5, 35, 201, 5, 5, 40, 201, 5, 5, 45, 221, 5, 5, 39, 201, 5, 5, 25, 201, 5, 5, 24, 201, 5, 5, 37, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 202, 5, 5, 60, 222, 5, 5, 53, 222, 5, 5, 84, 202, 5, 5, 103, 202, 5, 5, 49, 222, 5, 5, 56, 222, 5, 5, 58, 222, 5, 5, 61, 222, 5, 5, 98, 202, 192, 0, 0, 0, 5, 5, 57, 222, 5, 5, 50, 222, 5, 5, 96, 202, 5, 5, 54, 222, 192, 0, 0, 0, 5, 5, 62, 222, 5, 5, 81, 202, 192, 0, 0, 0, 5, 5, 59, 222, 5, 5, 91, 202, 5, 5, 90, 202, 5, 5, 52, 222, 5, 5, 79, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 202, 192, 0, 0, 0, 5, 5, 78, 202, 192, 0, 0, 0, 5, 5, 102, 202, 5, 5, 94, 202, 5, 5, 88, 202, 5, 5, 76, 202, 5, 5, 55, 222, 5, 5, 87, 202, 5, 5, 77, 202, 5, 5, 95, 202, 5, 5, 104, 202, 5, 5, 142, 223, 5, 5, 92, 202, 5, 5, 51, 222, 5, 5, 82, 202, 5, 5, 85, 202, 5, 5, 86, 202, 192, 0, 0, 0, 5, 5, 100, 202, 5, 5, 101, 202, 5, 5, 99, 202, 5, 5, 93, 202, 5, 5, 97, 202, 5, 5, 83, 202, 5, 5, 168, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 223, 5, 5, 180, 203, 5, 5, 172, 203, 5, 5, 174, 203, 5, 5, 179, 203, 5, 5, 160, 223, 5, 5, 177, 203, 5, 5, 70, 225, 5, 5, 64, 205, 5, 5, 182, 203, 5, 5, 145, 223, 5, 5, 153, 223, 5, 5, 178, 203, 5, 5, 152, 223, 5, 5, 149, 223, 5, 5, 157, 223, 5, 5, 159, 223, 5, 5, 171, 203, 5, 5, 175, 203, 5, 5, 181, 203, 5, 5, 66, 205, 5, 5, 154, 223, 5, 5, 173, 203, 5, 5, 184, 203, 5, 5, 147, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 203, 5, 5, 65, 205, 5, 5, 170, 203, 192, 0, 0, 0, 5, 5, 144, 223, 5, 5, 151, 223, 5, 5, 143, 223, 5, 5, 146, 223, 5, 5, 150, 223, 5, 5, 71, 225, 5, 5, 183, 203, 192, 0, 0, 0, 5, 5, 155, 223, 5, 5, 156, 223, 5, 5, 158, 223, 5, 5, 169, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 205, 5, 5, 83, 225, 5, 5, 78, 205, 5, 5, 79, 205, 5, 5, 73, 225, 5, 5, 89, 225, 192, 0, 0, 0, 5, 5, 69, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 227, 5, 5, 91, 225, 5, 5, 85, 225, 192, 0, 0, 0, 5, 5, 76, 225, 192, 0, 0, 0, 5, 5, 94, 225, 5, 5, 79, 225, 5, 5, 75, 205, 5, 5, 172, 227, 5, 5, 82, 225, 5, 5, 77, 205, 5, 5, 68, 205, 192, 0, 0, 0, 5, 5, 96, 225, 5, 5, 84, 225, 5, 5, 71, 205, 5, 5, 77, 225, 5, 5, 74, 225, 5, 5, 75, 225, 5, 5, 72, 205, 5, 5, 90, 225, 5, 5, 97, 225, 5, 5, 74, 205, 5, 5, 81, 225, 5, 5, 80, 225, 5, 5, 82, 205, 5, 5, 81, 205, 5, 5, 67, 205, 5, 5, 73, 205, 5, 5, 76, 205, 5, 5, 93, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 225, 5, 5, 70, 205, 5, 5, 72, 225, 5, 5, 95, 225, 5, 5, 86, 225, 5, 5, 88, 225, 5, 5, 92, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 227, 192, 0, 0, 0, 5, 5, 33, 207, 5, 5, 53, 207, 5, 5, 52, 207, 192, 0, 0, 0, 5, 5, 41, 207, 192, 0, 0, 0, 5, 5, 168, 227, 5, 5, 164, 227, 5, 5, 161, 227, 5, 5, 179, 227, 5, 5, 36, 207, 5, 5, 28, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 227, 5, 5, 235, 208, 5, 5, 32, 207, 5, 5, 174, 227, 192, 0, 0, 0, 5, 5, 54, 207, 5, 5, 51, 207, 5, 5, 170, 227, 5, 5, 153, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 207, 5, 5, 176, 227, 5, 5, 152, 227, 5, 5, 39, 207, 5, 5, 43, 207, 5, 5, 175, 227, 192, 0, 0, 0, 5, 5, 162, 227, 5, 5, 44, 207, 5, 5, 38, 207, 5, 5, 151, 227, 192, 0, 0, 0, 5, 5, 221, 208, 5, 5, 178, 227, 5, 5, 160, 227, 5, 5, 49, 207, 5, 5, 163, 227, 5, 5, 177, 227, 5, 5, 48, 207, 192, 0, 0, 0, 5, 5, 249, 229, 192, 0, 0, 0, 5, 5, 29, 207, 5, 5, 158, 227, 5, 5, 34, 207, 5, 5, 45, 207, 192, 0, 0, 0, 5, 5, 40, 207, 5, 5, 167, 227, 5, 5, 157, 227, 5, 5, 155, 227, 5, 5, 173, 227, 5, 5, 26, 207, 5, 5, 46, 207, 5, 5, 30, 207, 5, 5, 220, 208, 5, 5, 171, 227, 5, 5, 31, 207, 192, 0, 0, 0, 5, 5, 27, 207, 5, 5, 42, 207, 5, 5, 37, 207, 5, 5, 35, 207, 5, 5, 159, 227, 5, 5, 47, 207, 5, 5, 156, 227, 5, 5, 169, 227, 5, 5, 165, 227, 5, 5, 251, 229, 5, 5, 250, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 227, 5, 5, 13, 230, 192, 0, 0, 0, 5, 5, 223, 208, 192, 0, 0, 0, 5, 5, 19, 230, 5, 5, 2, 230, 5, 5, 16, 230, 5, 5, 152, 232, 5, 5, 226, 208, 5, 5, 20, 230, 5, 5, 23, 230, 5, 5, 225, 208, 5, 5, 5, 230, 5, 5, 22, 230, 5, 5, 21, 230, 5, 5, 227, 208, 5, 5, 252, 229, 5, 5, 222, 208, 5, 5, 230, 208, 192, 0, 0, 0, 5, 5, 228, 208, 5, 5, 18, 230, 192, 0, 0, 0, 5, 5, 8, 230, 5, 5, 232, 208, 5, 5, 25, 230, 5, 5, 17, 230, 5, 5, 26, 230, 5, 5, 240, 208, 5, 5, 238, 208, 5, 5, 15, 230, 5, 5, 14, 230, 192, 0, 0, 0, 5, 5, 12, 230, 5, 5, 6, 230, 5, 5, 231, 208, 192, 0, 0, 0, 5, 5, 229, 208, 5, 5, 78, 225, 5, 5, 253, 229, 192, 0, 0, 0, 5, 5, 182, 232, 5, 5, 254, 229, 5, 5, 224, 208, 5, 5, 237, 208, 5, 5, 154, 232, 192, 0, 0, 0, 5, 5, 233, 208, 5, 5, 234, 208, 5, 5, 255, 229, 5, 5, 24, 230, 5, 5, 153, 232, 5, 5, 9, 230, 5, 5, 4, 230, 5, 5, 236, 208, 5, 5, 10, 230, 5, 5, 7, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 232, 5, 5, 176, 232, 5, 5, 181, 232, 5, 5, 169, 232, 5, 5, 172, 210, 5, 5, 186, 232, 5, 5, 169, 210, 5, 5, 155, 232, 192, 0, 0, 0, 5, 5, 156, 232, 5, 5, 162, 210, 5, 5, 171, 210, 5, 5, 163, 232, 5, 5, 174, 210, 5, 5, 175, 210, 5, 5, 164, 232, 192, 0, 0, 0, 5, 5, 177, 232, 5, 5, 183, 232, 5, 5, 170, 210, 192, 0, 0, 0, 5, 5, 164, 210, 5, 5, 162, 232, 5, 5, 158, 232, 5, 5, 185, 232, 5, 5, 167, 232, 5, 5, 168, 232, 5, 5, 159, 232, 5, 5, 179, 232, 5, 5, 170, 232, 5, 5, 180, 232, 5, 5, 172, 232, 192, 0, 0, 0, 5, 5, 165, 210, 5, 5, 81, 235, 5, 5, 168, 210, 5, 5, 166, 210, 5, 5, 184, 232, 5, 5, 175, 232, 5, 5, 171, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 232, 5, 5, 110, 212, 5, 5, 174, 232, 5, 5, 173, 210, 5, 5, 166, 232, 192, 0, 0, 0, 5, 5, 165, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 210, 5, 5, 163, 210, 5, 5, 102, 235, 5, 5, 178, 232, 5, 5, 173, 232, 192, 0, 0, 0, 5, 5, 161, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 235, 192, 0, 0, 0, 5, 5, 82, 235, 5, 5, 93, 235, 192, 0, 0, 0, 5, 5, 100, 235, 5, 5, 108, 212, 5, 5, 239, 208, 5, 5, 98, 235, 5, 5, 103, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 212, 5, 5, 105, 235, 192, 0, 0, 0, 5, 5, 83, 235, 5, 5, 96, 235, 5, 5, 84, 235, 5, 5, 95, 235, 5, 5, 106, 212, 5, 5, 99, 235, 5, 5, 11, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 235, 5, 5, 91, 235, 5, 5, 109, 212, 5, 5, 170, 237, 5, 5, 228, 213, 192, 0, 0, 0, 5, 5, 104, 235, 5, 5, 103, 235, 5, 5, 111, 212, 5, 5, 157, 237, 5, 5, 229, 213, 5, 5, 158, 237, 192, 0, 0, 0, 5, 5, 87, 235, 5, 5, 88, 235, 5, 5, 85, 235, 5, 5, 90, 235, 5, 5, 86, 235, 5, 5, 107, 235, 5, 5, 105, 212, 5, 5, 230, 213, 5, 5, 107, 212, 5, 5, 112, 212, 192, 0, 0, 0, 5, 5, 89, 235, 192, 0, 0, 0, 5, 5, 101, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 237, 192, 0, 0, 0, 5, 5, 101, 212, 5, 5, 233, 213, 5, 5, 20, 240, 5, 5, 166, 237, 5, 5, 165, 237, 5, 5, 167, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 237, 5, 5, 234, 213, 192, 0, 0, 0, 5, 5, 240, 213, 5, 5, 237, 213, 192, 0, 0, 0, 5, 5, 238, 213, 5, 5, 159, 237, 5, 5, 162, 237, 5, 5, 172, 237, 5, 5, 246, 213, 5, 5, 244, 213, 192, 0, 0, 0, 5, 5, 163, 237, 5, 5, 3, 230, 5, 5, 231, 213, 5, 5, 169, 237, 5, 5, 160, 237, 192, 0, 0, 0, 5, 5, 247, 213, 5, 5, 168, 237, 5, 5, 104, 212, 5, 5, 236, 213, 5, 5, 92, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 213, 192, 0, 0, 0, 5, 5, 243, 213, 5, 5, 245, 213, 5, 5, 242, 213, 5, 5, 241, 213, 192, 0, 0, 0, 5, 5, 235, 213, 5, 5, 171, 237, 5, 5, 232, 213, 5, 5, 248, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 215, 5, 5, 76, 215, 5, 5, 19, 240, 5, 5, 85, 215, 5, 5, 82, 215, 192, 0, 0, 0, 5, 5, 73, 215, 5, 5, 80, 215, 5, 5, 21, 240, 5, 5, 78, 215, 5, 5, 72, 215, 192, 0, 0, 0, 5, 5, 79, 215, 192, 0, 0, 0, 5, 5, 18, 240, 5, 5, 106, 216, 5, 5, 74, 215, 192, 0, 0, 0, 5, 5, 81, 215, 5, 5, 105, 216, 5, 5, 17, 240, 5, 5, 16, 240, 192, 0, 0, 0, 5, 5, 83, 215, 192, 0, 0, 0, 5, 5, 84, 215, 192, 0, 0, 0, 5, 5, 15, 240, 5, 5, 14, 240, 5, 5, 107, 216, 5, 5, 24, 240, 5, 5, 77, 215, 5, 5, 22, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 216, 192, 0, 0, 0, 5, 5, 113, 216, 5, 5, 47, 242, 5, 5, 49, 242, 192, 0, 0, 0, 5, 5, 110, 216, 192, 0, 0, 0, 5, 5, 50, 242, 5, 5, 46, 242, 192, 0, 0, 0, 5, 5, 48, 242, 5, 5, 111, 216, 5, 5, 114, 216, 192, 0, 0, 0, 5, 5, 45, 242, 5, 5, 109, 216, 5, 5, 112, 216, 5, 5, 102, 217, 5, 5, 23, 240, 5, 5, 101, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 217, 5, 5, 242, 243, 192, 0, 0, 0, 5, 5, 105, 217, 5, 5, 106, 217, 5, 5, 245, 243, 5, 5, 241, 243, 5, 5, 103, 217, 5, 5, 239, 243, 5, 5, 23, 218, 5, 5, 243, 243, 192, 0, 0, 0, 5, 5, 244, 243, 5, 5, 240, 243, 192, 0, 0, 0, 5, 5, 104, 217, 5, 5, 93, 245, 192, 0, 0, 0, 5, 5, 96, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 245, 5, 5, 95, 245, 5, 5, 98, 245, 5, 5, 24, 218, 5, 5, 94, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 246, 5, 5, 173, 218, 5, 5, 164, 246, 5, 5, 163, 246, 5, 5, 162, 246, 5, 5, 172, 218, 5, 5, 174, 218, 192, 0, 0, 0, 5, 5, 161, 247, 5, 5, 11, 219, 5, 5, 10, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 248, 5, 5, 83, 248, 5, 5, 86, 248, 5, 5, 145, 219, 5, 5, 93, 219, 5, 5, 237, 248, 5, 5, 85, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 248, 5, 5, 147, 219, 5, 5, 146, 219, 5, 5, 185, 219, 5, 5, 88, 249, 5, 5, 158, 249, 5, 5, 69, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 230, 5, 5, 25, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 200, 5, 5, 57, 221, 5, 5, 45, 201, 5, 5, 43, 201, 192, 0, 0, 0, 5, 5, 44, 201, 192, 0, 0, 0, 5, 5, 63, 222, 5, 5, 105, 202, 5, 5, 185, 203, 192, 0, 0, 0, 5, 5, 161, 223, 192, 0, 0, 0, 5, 5, 162, 223, 192, 0, 0, 0, 5, 5, 186, 203, 5, 5, 99, 225, 192, 0, 0, 0, 5, 5, 83, 205, 5, 5, 84, 205, 5, 5, 98, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 207, 192, 0, 0, 0, 5, 5, 57, 207, 192, 0, 0, 0, 5, 5, 180, 227, 5, 5, 64, 207, 5, 5, 63, 207, 5, 5, 56, 207, 5, 5, 59, 207, 5, 5, 62, 207, 5, 5, 58, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 230, 5, 5, 55, 207, 5, 5, 242, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 208, 5, 5, 245, 208, 5, 5, 30, 230, 5, 5, 33, 230, 5, 5, 243, 208, 5, 5, 28, 230, 5, 5, 32, 230, 192, 0, 0, 0, 5, 5, 29, 230, 192, 0, 0, 0, 5, 5, 177, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 212, 5, 5, 108, 235, 5, 5, 86, 215, 5, 5, 249, 213, 5, 5, 173, 237, 5, 5, 250, 213, 5, 5, 251, 213, 5, 5, 175, 237, 5, 5, 174, 237, 5, 5, 176, 237, 5, 5, 27, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 240, 5, 5, 52, 242, 5, 5, 51, 242, 5, 5, 115, 216, 5, 5, 116, 216, 5, 5, 99, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 208, 5, 5, 246, 208, 5, 5, 188, 232, 192, 0, 0, 0, 5, 5, 246, 243, 5, 5, 12, 219, 5, 5, 238, 248, 5, 5, 71, 199, 192, 0, 0, 0, 5, 5, 85, 205, 192, 0, 0, 0, 5, 5, 66, 207, 5, 5, 65, 207, 5, 5, 35, 230, 5, 5, 36, 230, 5, 5, 178, 210, 5, 5, 109, 235, 5, 5, 114, 212, 5, 5, 28, 240, 192, 0, 0, 0, 5, 5, 72, 199, 5, 5, 182, 199, 192, 0, 0, 0, 5, 5, 106, 202, 5, 5, 64, 222, 192, 0, 0, 0, 5, 5, 163, 223, 5, 5, 187, 203, 5, 5, 67, 207, 192, 0, 0, 0, 5, 5, 37, 230, 5, 5, 248, 208, 5, 5, 179, 210, 192, 0, 0, 0, 5, 5, 177, 237, 5, 5, 178, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 242, 5, 5, 108, 217, 5, 5, 159, 249, 5, 5, 73, 199, 192, 0, 0, 0, 5, 5, 65, 222, 5, 5, 107, 202, 5, 5, 188, 203, 192, 0, 0, 0, 5, 5, 164, 223, 192, 0, 0, 0, 5, 5, 86, 205, 5, 5, 103, 225, 5, 5, 101, 225, 5, 5, 102, 225, 5, 5, 87, 205, 5, 5, 100, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 207, 5, 5, 70, 207, 5, 5, 181, 227, 5, 5, 71, 207, 5, 5, 68, 207, 5, 5, 38, 230, 192, 0, 0, 0, 5, 5, 39, 230, 5, 5, 189, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 212, 5, 5, 115, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 242, 5, 5, 247, 243, 192, 0, 0, 0, 5, 5, 101, 245, 5, 5, 100, 245, 5, 5, 166, 246, 192, 0, 0, 0, 5, 5, 35, 220, 5, 5, 189, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 199, 5, 5, 183, 199, 192, 0, 0, 0, 5, 5, 75, 200, 5, 5, 74, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 200, 5, 5, 77, 200, 5, 5, 145, 220, 5, 5, 144, 220, 5, 5, 58, 221, 5, 5, 46, 201, 5, 5, 61, 221, 5, 5, 60, 221, 5, 5, 59, 221, 5, 5, 62, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 202, 5, 5, 71, 222, 5, 5, 67, 222, 5, 5, 76, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 202, 192, 0, 0, 0, 5, 5, 113, 202, 5, 5, 72, 222, 5, 5, 68, 222, 5, 5, 75, 222, 5, 5, 112, 202, 5, 5, 119, 202, 5, 5, 70, 222, 5, 5, 66, 222, 5, 5, 118, 202, 5, 5, 73, 222, 5, 5, 111, 202, 5, 5, 74, 222, 5, 5, 114, 202, 5, 5, 116, 202, 5, 5, 78, 222, 5, 5, 77, 222, 5, 5, 69, 222, 5, 5, 110, 202, 5, 5, 109, 202, 5, 5, 117, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 223, 5, 5, 175, 223, 192, 0, 0, 0, 5, 5, 195, 203, 5, 5, 192, 203, 5, 5, 166, 223, 5, 5, 171, 223, 192, 0, 0, 0, 5, 5, 198, 203, 5, 5, 190, 203, 5, 5, 170, 223, 5, 5, 193, 203, 5, 5, 196, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 223, 192, 0, 0, 0, 5, 5, 191, 203, 5, 5, 178, 223, 5, 5, 194, 203, 192, 0, 0, 0, 5, 5, 197, 203, 5, 5, 167, 223, 5, 5, 172, 223, 5, 5, 176, 223, 5, 5, 168, 223, 5, 5, 165, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 223, 5, 5, 174, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 205, 5, 5, 88, 205, 5, 5, 91, 205, 192, 0, 0, 0, 5, 5, 94, 205, 192, 0, 0, 0, 5, 5, 106, 225, 192, 0, 0, 0, 5, 5, 89, 205, 5, 5, 104, 225, 192, 0, 0, 0, 5, 5, 93, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 205, 192, 0, 0, 0, 5, 5, 107, 225, 5, 5, 92, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 227, 5, 5, 73, 207, 5, 5, 184, 227, 5, 5, 186, 227, 5, 5, 72, 207, 5, 5, 77, 207, 5, 5, 105, 225, 192, 0, 0, 0, 5, 5, 183, 227, 5, 5, 187, 227, 192, 0, 0, 0, 5, 5, 74, 207, 5, 5, 182, 227, 5, 5, 76, 207, 192, 0, 0, 0, 5, 5, 75, 207, 192, 0, 0, 0, 5, 5, 46, 230, 192, 0, 0, 0, 5, 5, 41, 230, 192, 0, 0, 0, 5, 5, 249, 208, 5, 5, 253, 208, 5, 5, 250, 208, 5, 5, 44, 230, 5, 5, 47, 230, 192, 0, 0, 0, 5, 5, 251, 208, 192, 0, 0, 0, 5, 5, 252, 208, 5, 5, 3, 209, 5, 5, 198, 232, 5, 5, 45, 230, 5, 5, 255, 208, 5, 5, 42, 230, 5, 5, 40, 230, 192, 0, 0, 0, 5, 5, 2, 209, 192, 0, 0, 0, 5, 5, 43, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 210, 192, 0, 0, 0, 5, 5, 190, 232, 5, 5, 182, 210, 5, 5, 183, 210, 5, 5, 181, 210, 5, 5, 195, 232, 5, 5, 194, 232, 5, 5, 191, 232, 5, 5, 187, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 232, 5, 5, 254, 208, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 232, 5, 5, 192, 232, 5, 5, 184, 210, 5, 5, 180, 210, 5, 5, 186, 210, 5, 5, 196, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 212, 192, 0, 0, 0, 5, 5, 112, 235, 5, 5, 111, 235, 5, 5, 110, 235, 5, 5, 117, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 212, 5, 5, 181, 237, 5, 5, 184, 237, 5, 5, 253, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 213, 5, 5, 185, 237, 5, 5, 180, 237, 5, 5, 255, 213, 5, 5, 182, 237, 192, 0, 0, 0, 5, 5, 254, 213, 5, 5, 179, 237, 192, 0, 0, 0, 5, 5, 183, 237, 5, 5, 92, 215, 5, 5, 89, 215, 5, 5, 37, 240, 5, 5, 36, 240, 192, 0, 0, 0, 5, 5, 35, 240, 5, 5, 30, 240, 192, 0, 0, 0, 5, 5, 31, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 215, 192, 0, 0, 0, 5, 5, 87, 215, 5, 5, 91, 215, 5, 5, 29, 240, 5, 5, 88, 215, 5, 5, 32, 240, 5, 5, 33, 240, 5, 5, 38, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 216, 192, 0, 0, 0, 5, 5, 250, 243, 5, 5, 117, 216, 5, 5, 248, 243, 5, 5, 249, 243, 5, 5, 109, 217, 5, 5, 26, 218, 5, 5, 102, 245, 192, 0, 0, 0, 5, 5, 25, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 246, 5, 5, 169, 246, 192, 0, 0, 0, 5, 5, 175, 218, 192, 0, 0, 0, 5, 5, 167, 246, 5, 5, 13, 219, 192, 0, 0, 0, 5, 5, 239, 248, 5, 5, 148, 219, 5, 5, 89, 249, 5, 5, 90, 249, 192, 0, 0, 0, 5, 5, 75, 199, 192, 0, 0, 0, 5, 5, 78, 200, 5, 5, 79, 200, 5, 5, 47, 201, 192, 0, 0, 0, 5, 5, 79, 222, 5, 5, 199, 203, 5, 5, 96, 205, 5, 5, 78, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 206, 192, 0, 0, 0, 5, 5, 4, 209, 5, 5, 5, 209, 5, 5, 114, 208, 5, 5, 48, 230, 192, 0, 0, 0, 5, 5, 188, 210, 5, 5, 114, 235, 5, 5, 113, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 199, 5, 5, 80, 200, 5, 5, 80, 222, 5, 5, 121, 202, 192, 0, 0, 0, 5, 5, 120, 202, 192, 0, 0, 0, 5, 5, 179, 223, 5, 5, 180, 223, 192, 0, 0, 0, 5, 5, 108, 225, 5, 5, 109, 225, 5, 5, 97, 205, 5, 5, 98, 205, 192, 0, 0, 0, 5, 5, 99, 205, 5, 5, 188, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 207, 192, 0, 0, 0, 5, 5, 7, 209, 192, 0, 0, 0, 5, 5, 6, 209, 5, 5, 199, 232, 5, 5, 214, 233, 5, 5, 115, 235, 5, 5, 39, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 217, 5, 5, 176, 218, 5, 5, 77, 199, 192, 0, 0, 0, 5, 5, 186, 199, 5, 5, 187, 199, 5, 5, 185, 199, 5, 5, 188, 199, 5, 5, 184, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 200, 192, 0, 0, 0, 5, 5, 153, 220, 5, 5, 82, 200, 5, 5, 84, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 220, 5, 5, 147, 220, 5, 5, 150, 220, 5, 5, 149, 220, 5, 5, 152, 220, 5, 5, 81, 200, 5, 5, 146, 220, 5, 5, 151, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 221, 5, 5, 57, 201, 5, 5, 64, 221, 5, 5, 68, 221, 5, 5, 56, 201, 192, 0, 0, 0, 5, 5, 72, 221, 5, 5, 67, 221, 5, 5, 70, 221, 5, 5, 49, 201, 5, 5, 50, 201, 5, 5, 51, 201, 5, 5, 52, 201, 192, 0, 0, 0, 5, 5, 59, 201, 192, 0, 0, 0, 5, 5, 66, 221, 5, 5, 54, 201, 5, 5, 60, 201, 192, 0, 0, 0, 5, 5, 65, 221, 5, 5, 71, 221, 192, 0, 0, 0, 5, 5, 53, 201, 5, 5, 69, 221, 5, 5, 55, 201, 5, 5, 48, 201, 5, 5, 58, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 202, 192, 0, 0, 0, 5, 5, 82, 222, 5, 5, 122, 202, 192, 0, 0, 0, 5, 5, 132, 202, 5, 5, 133, 202, 5, 5, 125, 202, 5, 5, 143, 202, 5, 5, 127, 202, 5, 5, 90, 222, 5, 5, 138, 202, 5, 5, 85, 222, 5, 5, 128, 202, 5, 5, 100, 222, 5, 5, 101, 222, 5, 5, 93, 222, 5, 5, 86, 222, 5, 5, 141, 202, 5, 5, 98, 222, 5, 5, 136, 202, 5, 5, 134, 202, 192, 0, 0, 0, 5, 5, 99, 222, 192, 0, 0, 0, 5, 5, 97, 222, 5, 5, 89, 222, 5, 5, 81, 222, 5, 5, 88, 222, 5, 5, 129, 202, 192, 0, 0, 0, 5, 5, 135, 202, 192, 0, 0, 0, 5, 5, 123, 202, 5, 5, 92, 222, 5, 5, 91, 222, 5, 5, 83, 222, 192, 0, 0, 0, 5, 5, 137, 202, 5, 5, 95, 222, 5, 5, 84, 222, 5, 5, 140, 202, 5, 5, 102, 222, 5, 5, 124, 202, 192, 0, 0, 0, 5, 5, 131, 202, 5, 5, 87, 222, 5, 5, 96, 222, 5, 5, 139, 202, 192, 0, 0, 0, 5, 5, 126, 202, 5, 5, 130, 202, 192, 0, 0, 0, 5, 5, 94, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 223, 5, 5, 207, 203, 5, 5, 220, 203, 192, 0, 0, 0, 5, 5, 207, 223, 5, 5, 197, 223, 5, 5, 213, 203, 5, 5, 195, 223, 5, 5, 206, 203, 5, 5, 190, 223, 5, 5, 216, 203, 5, 5, 203, 223, 5, 5, 184, 223, 5, 5, 186, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 223, 5, 5, 181, 223, 5, 5, 202, 223, 5, 5, 218, 223, 5, 5, 211, 203, 5, 5, 191, 223, 5, 5, 209, 223, 192, 0, 0, 0, 5, 5, 183, 223, 5, 5, 216, 223, 5, 5, 217, 223, 5, 5, 220, 223, 5, 5, 211, 223, 5, 5, 196, 223, 5, 5, 204, 223, 5, 5, 217, 203, 5, 5, 204, 203, 5, 5, 212, 203, 5, 5, 224, 203, 5, 5, 201, 203, 5, 5, 203, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 223, 5, 5, 221, 203, 5, 5, 214, 203, 5, 5, 214, 223, 5, 5, 185, 223, 5, 5, 223, 203, 5, 5, 218, 203, 5, 5, 194, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 203, 5, 5, 201, 223, 5, 5, 193, 223, 5, 5, 215, 203, 5, 5, 213, 223, 5, 5, 205, 223, 192, 0, 0, 0, 5, 5, 209, 203, 5, 5, 219, 223, 5, 5, 192, 223, 5, 5, 205, 203, 5, 5, 210, 223, 5, 5, 200, 223, 5, 5, 210, 203, 5, 5, 206, 223, 5, 5, 202, 203, 5, 5, 182, 223, 5, 5, 219, 203, 5, 5, 114, 205, 5, 5, 208, 203, 5, 5, 199, 223, 5, 5, 198, 223, 5, 5, 187, 223, 192, 0, 0, 0, 5, 5, 215, 223, 192, 0, 0, 0, 5, 5, 208, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 225, 5, 5, 121, 205, 5, 5, 130, 225, 192, 0, 0, 0, 5, 5, 118, 225, 5, 5, 110, 205, 5, 5, 122, 205, 192, 0, 0, 0, 5, 5, 111, 225, 192, 0, 0, 0, 5, 5, 120, 225, 5, 5, 128, 225, 192, 0, 0, 0, 5, 5, 110, 225, 5, 5, 137, 225, 5, 5, 100, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 225, 5, 5, 131, 225, 192, 0, 0, 0, 5, 5, 132, 225, 5, 5, 108, 205, 5, 5, 119, 205, 5, 5, 122, 225, 192, 0, 0, 0, 5, 5, 123, 225, 5, 5, 133, 225, 5, 5, 124, 225, 192, 0, 0, 0, 5, 5, 119, 225, 5, 5, 113, 225, 5, 5, 114, 225, 5, 5, 127, 225, 5, 5, 121, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 205, 5, 5, 105, 205, 5, 5, 135, 225, 5, 5, 115, 225, 5, 5, 117, 205, 5, 5, 113, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 205, 5, 5, 123, 205, 5, 5, 106, 205, 5, 5, 118, 205, 5, 5, 126, 225, 5, 5, 120, 205, 5, 5, 103, 205, 192, 0, 0, 0, 5, 5, 102, 205, 5, 5, 112, 225, 192, 0, 0, 0, 5, 5, 116, 225, 5, 5, 111, 205, 5, 5, 134, 225, 5, 5, 125, 225, 5, 5, 117, 225, 5, 5, 115, 205, 5, 5, 112, 205, 192, 0, 0, 0, 5, 5, 104, 205, 5, 5, 107, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 227, 192, 0, 0, 0, 5, 5, 193, 227, 5, 5, 194, 227, 5, 5, 198, 227, 192, 0, 0, 0, 5, 5, 213, 227, 5, 5, 210, 227, 192, 0, 0, 0, 5, 5, 203, 227, 5, 5, 202, 227, 5, 5, 87, 207, 5, 5, 206, 227, 5, 5, 221, 227, 5, 5, 189, 227, 192, 0, 0, 0, 5, 5, 222, 227, 5, 5, 208, 227, 5, 5, 226, 227, 5, 5, 214, 227, 5, 5, 86, 207, 5, 5, 212, 227, 5, 5, 81, 207, 5, 5, 102, 207, 5, 5, 92, 207, 192, 0, 0, 0, 5, 5, 96, 207, 5, 5, 95, 207, 5, 5, 190, 227, 192, 0, 0, 0, 5, 5, 219, 227, 5, 5, 225, 227, 5, 5, 217, 227, 5, 5, 224, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 227, 5, 5, 191, 227, 5, 5, 223, 227, 5, 5, 207, 227, 5, 5, 84, 207, 5, 5, 97, 207, 192, 0, 0, 0, 5, 5, 216, 227, 5, 5, 90, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 227, 5, 5, 192, 227, 5, 5, 98, 207, 192, 0, 0, 0, 5, 5, 100, 207, 5, 5, 218, 227, 5, 5, 101, 207, 5, 5, 83, 207, 5, 5, 199, 227, 5, 5, 220, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 207, 5, 5, 99, 207, 5, 5, 201, 227, 5, 5, 211, 227, 5, 5, 196, 227, 5, 5, 200, 227, 5, 5, 94, 207, 5, 5, 195, 227, 5, 5, 82, 207, 5, 5, 91, 207, 5, 5, 88, 207, 5, 5, 204, 227, 5, 5, 109, 205, 5, 5, 71, 230, 5, 5, 85, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 207, 192, 0, 0, 0, 5, 5, 74, 230, 5, 5, 96, 230, 5, 5, 85, 230, 5, 5, 27, 209, 192, 0, 0, 0, 5, 5, 22, 209, 5, 5, 68, 230, 5, 5, 23, 209, 5, 5, 84, 230, 192, 0, 0, 0, 5, 5, 77, 230, 5, 5, 90, 230, 5, 5, 19, 209, 5, 5, 50, 230, 5, 5, 92, 230, 5, 5, 9, 209, 5, 5, 58, 230, 5, 5, 12, 209, 5, 5, 11, 209, 192, 0, 0, 0, 5, 5, 28, 209, 192, 0, 0, 0, 5, 5, 52, 230, 5, 5, 86, 230, 5, 5, 87, 230, 5, 5, 14, 209, 5, 5, 10, 209, 5, 5, 66, 230, 192, 0, 0, 0, 5, 5, 21, 209, 5, 5, 61, 230, 192, 0, 0, 0, 5, 5, 88, 230, 5, 5, 17, 209, 5, 5, 80, 230, 5, 5, 93, 230, 5, 5, 55, 230, 5, 5, 60, 230, 5, 5, 54, 230, 192, 0, 0, 0, 5, 5, 16, 209, 5, 5, 73, 230, 5, 5, 221, 232, 5, 5, 56, 230, 5, 5, 20, 209, 5, 5, 65, 230, 5, 5, 89, 230, 5, 5, 15, 209, 5, 5, 62, 230, 5, 5, 59, 230, 5, 5, 76, 230, 5, 5, 18, 209, 5, 5, 8, 209, 5, 5, 30, 209, 5, 5, 79, 230, 5, 5, 78, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 230, 5, 5, 13, 209, 5, 5, 91, 230, 5, 5, 67, 230, 5, 5, 69, 230, 192, 0, 0, 0, 5, 5, 82, 230, 5, 5, 81, 230, 5, 5, 49, 230, 5, 5, 24, 209, 5, 5, 26, 209, 5, 5, 57, 230, 5, 5, 64, 230, 5, 5, 72, 230, 5, 5, 25, 209, 5, 5, 63, 230, 5, 5, 75, 230, 5, 5, 94, 230, 192, 0, 0, 0, 5, 5, 83, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 232, 5, 5, 196, 210, 192, 0, 0, 0, 5, 5, 239, 232, 5, 5, 218, 232, 5, 5, 223, 232, 5, 5, 216, 232, 192, 0, 0, 0, 5, 5, 235, 232, 5, 5, 202, 232, 5, 5, 209, 232, 192, 0, 0, 0, 5, 5, 233, 232, 5, 5, 243, 232, 5, 5, 219, 232, 192, 0, 0, 0, 5, 5, 206, 232, 5, 5, 225, 232, 5, 5, 230, 232, 5, 5, 210, 232, 192, 0, 0, 0, 5, 5, 227, 232, 5, 5, 207, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 232, 5, 5, 215, 232, 5, 5, 198, 210, 5, 5, 234, 232, 5, 5, 232, 232, 192, 0, 0, 0, 5, 5, 203, 232, 5, 5, 237, 232, 192, 0, 0, 0, 5, 5, 238, 232, 5, 5, 240, 232, 5, 5, 202, 210, 5, 5, 194, 210, 192, 0, 0, 0, 5, 5, 197, 227, 5, 5, 211, 232, 5, 5, 229, 232, 5, 5, 212, 232, 5, 5, 191, 210, 5, 5, 207, 210, 5, 5, 236, 232, 5, 5, 205, 210, 5, 5, 201, 210, 5, 5, 201, 232, 5, 5, 193, 210, 192, 0, 0, 0, 5, 5, 204, 232, 5, 5, 206, 210, 192, 0, 0, 0, 5, 5, 220, 232, 5, 5, 200, 232, 192, 0, 0, 0, 5, 5, 199, 210, 5, 5, 224, 232, 5, 5, 208, 232, 5, 5, 200, 210, 5, 5, 217, 232, 5, 5, 190, 210, 5, 5, 29, 209, 5, 5, 226, 232, 5, 5, 70, 230, 5, 5, 205, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 232, 5, 5, 195, 210, 5, 5, 228, 232, 5, 5, 192, 210, 5, 5, 222, 232, 5, 5, 203, 210, 5, 5, 213, 232, 5, 5, 242, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 210, 5, 5, 99, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 235, 5, 5, 127, 235, 192, 0, 0, 0, 5, 5, 149, 235, 5, 5, 125, 235, 192, 0, 0, 0, 5, 5, 143, 235, 5, 5, 189, 210, 5, 5, 122, 212, 5, 5, 121, 235, 5, 5, 148, 235, 192, 0, 0, 0, 5, 5, 126, 235, 5, 5, 140, 235, 5, 5, 127, 212, 5, 5, 120, 212, 192, 0, 0, 0, 5, 5, 146, 235, 192, 0, 0, 0, 5, 5, 119, 235, 5, 5, 145, 235, 192, 0, 0, 0, 5, 5, 138, 212, 5, 5, 136, 235, 5, 5, 152, 235, 5, 5, 136, 212, 5, 5, 128, 235, 5, 5, 121, 212, 5, 5, 130, 235, 5, 5, 144, 235, 5, 5, 130, 212, 5, 5, 123, 235, 5, 5, 134, 212, 5, 5, 124, 212, 5, 5, 132, 235, 5, 5, 122, 235, 5, 5, 116, 235, 192, 0, 0, 0, 5, 5, 142, 235, 5, 5, 131, 212, 5, 5, 151, 235, 5, 5, 117, 235, 5, 5, 128, 212, 192, 0, 0, 0, 5, 5, 138, 235, 192, 0, 0, 0, 5, 5, 129, 212, 5, 5, 124, 235, 5, 5, 135, 235, 5, 5, 131, 235, 5, 5, 133, 235, 192, 0, 0, 0, 5, 5, 123, 212, 5, 5, 150, 235, 5, 5, 137, 212, 5, 5, 134, 235, 192, 0, 0, 0, 5, 5, 153, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 235, 5, 5, 139, 235, 5, 5, 126, 212, 5, 5, 135, 212, 5, 5, 133, 212, 5, 5, 120, 235, 5, 5, 141, 235, 5, 5, 132, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 212, 5, 5, 137, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 237, 192, 0, 0, 0, 5, 5, 193, 237, 5, 5, 189, 237, 5, 5, 209, 237, 5, 5, 206, 237, 5, 5, 4, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 237, 5, 5, 196, 237, 5, 5, 15, 214, 5, 5, 203, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 237, 5, 5, 202, 237, 5, 5, 12, 214, 5, 5, 222, 237, 192, 0, 0, 0, 5, 5, 219, 237, 5, 5, 205, 237, 5, 5, 190, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 237, 5, 5, 212, 237, 5, 5, 8, 214, 5, 5, 200, 237, 5, 5, 195, 237, 5, 5, 186, 237, 5, 5, 5, 214, 5, 5, 13, 214, 192, 0, 0, 0, 5, 5, 215, 237, 5, 5, 14, 214, 5, 5, 187, 237, 5, 5, 221, 237, 5, 5, 208, 237, 5, 5, 214, 237, 5, 5, 11, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 237, 192, 0, 0, 0, 5, 5, 218, 237, 192, 0, 0, 0, 5, 5, 16, 214, 192, 0, 0, 0, 5, 5, 10, 214, 5, 5, 204, 237, 5, 5, 191, 237, 5, 5, 223, 237, 5, 5, 188, 237, 5, 5, 216, 237, 5, 5, 7, 214, 192, 0, 0, 0, 5, 5, 198, 237, 192, 0, 0, 0, 5, 5, 199, 237, 5, 5, 6, 214, 5, 5, 3, 214, 5, 5, 194, 237, 5, 5, 9, 214, 192, 0, 0, 0, 5, 5, 2, 214, 192, 0, 0, 0, 5, 5, 217, 237, 5, 5, 220, 237, 5, 5, 201, 237, 5, 5, 52, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 240, 192, 0, 0, 0, 5, 5, 40, 240, 5, 5, 105, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 215, 5, 5, 99, 215, 5, 5, 95, 215, 5, 5, 54, 240, 5, 5, 65, 240, 5, 5, 93, 215, 5, 5, 46, 240, 5, 5, 55, 240, 5, 5, 207, 237, 5, 5, 56, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 215, 192, 0, 0, 0, 5, 5, 71, 240, 5, 5, 104, 215, 5, 5, 107, 215, 5, 5, 42, 240, 192, 0, 0, 0, 5, 5, 103, 215, 192, 0, 0, 0, 5, 5, 69, 240, 5, 5, 70, 240, 5, 5, 60, 240, 5, 5, 59, 240, 5, 5, 51, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 240, 5, 5, 68, 240, 5, 5, 67, 240, 192, 0, 0, 0, 5, 5, 98, 215, 5, 5, 96, 215, 5, 5, 53, 240, 5, 5, 50, 240, 192, 0, 0, 0, 5, 5, 47, 240, 5, 5, 66, 240, 5, 5, 106, 215, 5, 5, 64, 240, 5, 5, 102, 215, 5, 5, 101, 215, 192, 0, 0, 0, 5, 5, 58, 240, 192, 0, 0, 0, 5, 5, 41, 240, 5, 5, 43, 240, 5, 5, 45, 240, 5, 5, 63, 240, 5, 5, 57, 240, 5, 5, 97, 215, 192, 0, 0, 0, 5, 5, 48, 240, 192, 0, 0, 0, 5, 5, 62, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 216, 5, 5, 73, 242, 5, 5, 119, 216, 5, 5, 58, 242, 192, 0, 0, 0, 5, 5, 69, 242, 5, 5, 121, 216, 5, 5, 76, 242, 192, 0, 0, 0, 5, 5, 65, 242, 192, 0, 0, 0, 5, 5, 60, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 242, 5, 5, 56, 242, 5, 5, 67, 242, 192, 0, 0, 0, 5, 5, 128, 216, 5, 5, 72, 242, 5, 5, 78, 242, 5, 5, 66, 242, 5, 5, 120, 216, 5, 5, 68, 242, 5, 5, 57, 242, 5, 5, 127, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 242, 5, 5, 62, 242, 5, 5, 123, 216, 192, 0, 0, 0, 5, 5, 64, 242, 5, 5, 61, 242, 5, 5, 129, 216, 5, 5, 63, 242, 5, 5, 122, 216, 5, 5, 125, 216, 5, 5, 71, 242, 5, 5, 59, 242, 5, 5, 74, 242, 192, 0, 0, 0, 5, 5, 70, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 217, 5, 5, 7, 244, 5, 5, 117, 217, 5, 5, 118, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 217, 5, 5, 6, 244, 192, 0, 0, 0, 5, 5, 3, 244, 5, 5, 4, 244, 5, 5, 115, 217, 5, 5, 252, 243, 5, 5, 2, 244, 5, 5, 114, 217, 192, 0, 0, 0, 5, 5, 253, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 217, 5, 5, 113, 217, 192, 0, 0, 0, 5, 5, 251, 243, 5, 5, 255, 243, 5, 5, 5, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 245, 5, 5, 105, 245, 5, 5, 114, 245, 192, 0, 0, 0, 5, 5, 113, 245, 5, 5, 111, 245, 5, 5, 106, 245, 192, 0, 0, 0, 5, 5, 30, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 245, 5, 5, 29, 218, 5, 5, 124, 216, 5, 5, 110, 245, 5, 5, 28, 218, 5, 5, 115, 245, 5, 5, 109, 245, 5, 5, 104, 245, 5, 5, 254, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 218, 192, 0, 0, 0, 5, 5, 103, 245, 5, 5, 173, 246, 192, 0, 0, 0, 5, 5, 172, 246, 5, 5, 112, 245, 5, 5, 177, 218, 192, 0, 0, 0, 5, 5, 176, 246, 5, 5, 177, 246, 5, 5, 171, 246, 5, 5, 175, 246, 192, 0, 0, 0, 5, 5, 170, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 247, 5, 5, 174, 246, 5, 5, 16, 219, 5, 5, 14, 219, 5, 5, 163, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 247, 192, 0, 0, 0, 5, 5, 162, 247, 5, 5, 164, 247, 5, 5, 15, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 248, 5, 5, 89, 248, 5, 5, 94, 219, 5, 5, 87, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 248, 5, 5, 149, 219, 5, 5, 240, 248, 5, 5, 241, 248, 5, 5, 91, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 219, 5, 5, 162, 249, 5, 5, 160, 249, 5, 5, 161, 249, 5, 5, 163, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 250, 192, 0, 0, 0, 5, 5, 78, 199, 5, 5, 85, 200, 192, 0, 0, 0, 5, 5, 144, 202, 192, 0, 0, 0, 5, 5, 103, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 225, 5, 5, 140, 225, 192, 0, 0, 0, 5, 5, 139, 225, 192, 0, 0, 0, 5, 5, 141, 225, 5, 5, 103, 207, 5, 5, 228, 227, 5, 5, 142, 225, 192, 0, 0, 0, 5, 5, 227, 227, 5, 5, 229, 227, 5, 5, 230, 227, 5, 5, 97, 230, 5, 5, 32, 209, 5, 5, 98, 230, 5, 5, 100, 230, 5, 5, 33, 209, 5, 5, 31, 209, 5, 5, 99, 230, 192, 0, 0, 0, 5, 5, 249, 232, 5, 5, 247, 232, 5, 5, 246, 232, 192, 0, 0, 0, 5, 5, 245, 232, 5, 5, 244, 232, 5, 5, 208, 210, 5, 5, 248, 232, 5, 5, 139, 212, 5, 5, 154, 235, 5, 5, 156, 235, 5, 5, 140, 212, 5, 5, 155, 235, 5, 5, 18, 214, 192, 0, 0, 0, 5, 5, 17, 214, 5, 5, 224, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 240, 5, 5, 72, 240, 5, 5, 74, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 215, 192, 0, 0, 0, 5, 5, 79, 242, 5, 5, 130, 216, 192, 0, 0, 0, 5, 5, 8, 244, 5, 5, 119, 217, 5, 5, 116, 245, 5, 5, 95, 219, 5, 5, 79, 199, 5, 5, 189, 199, 5, 5, 86, 200, 5, 5, 61, 201, 5, 5, 145, 202, 5, 5, 146, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 238, 5, 5, 109, 215, 5, 5, 120, 217, 5, 5, 80, 199, 192, 0, 0, 0, 5, 5, 87, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 222, 5, 5, 147, 202, 5, 5, 104, 222, 192, 0, 0, 0, 5, 5, 222, 223, 5, 5, 226, 203, 5, 5, 223, 223, 192, 0, 0, 0, 5, 5, 227, 203, 192, 0, 0, 0, 5, 5, 145, 225, 5, 5, 125, 205, 5, 5, 124, 205, 192, 0, 0, 0, 5, 5, 235, 227, 5, 5, 233, 227, 5, 5, 234, 227, 5, 5, 232, 227, 192, 0, 0, 0, 5, 5, 231, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 230, 5, 5, 104, 230, 5, 5, 35, 209, 5, 5, 102, 230, 5, 5, 34, 209, 5, 5, 103, 230, 192, 0, 0, 0, 5, 5, 250, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 235, 5, 5, 158, 235, 5, 5, 159, 235, 192, 0, 0, 0, 5, 5, 227, 237, 5, 5, 226, 237, 5, 5, 19, 214, 5, 5, 225, 237, 5, 5, 228, 237, 5, 5, 75, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 240, 5, 5, 77, 240, 192, 0, 0, 0, 5, 5, 80, 242, 5, 5, 131, 216, 5, 5, 121, 217, 5, 5, 117, 245, 192, 0, 0, 0, 5, 5, 17, 219, 5, 5, 36, 220, 192, 0, 0, 0, 5, 5, 228, 203, 5, 5, 224, 223, 5, 5, 126, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 207, 192, 0, 0, 0, 5, 5, 36, 209, 5, 5, 105, 230, 192, 0, 0, 0, 5, 5, 211, 210, 5, 5, 210, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 235, 5, 5, 161, 235, 5, 5, 20, 214, 5, 5, 21, 214, 5, 5, 79, 240, 5, 5, 78, 240, 5, 5, 9, 244, 5, 5, 243, 248, 5, 5, 81, 199, 5, 5, 37, 220, 5, 5, 190, 199, 192, 0, 0, 0, 5, 5, 62, 201, 5, 5, 73, 221, 192, 0, 0, 0, 5, 5, 229, 203, 5, 5, 212, 210, 5, 5, 82, 199, 192, 0, 0, 0, 5, 5, 225, 223, 5, 5, 230, 203, 5, 5, 226, 223, 192, 0, 0, 0, 5, 5, 132, 216, 5, 5, 83, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 222, 192, 0, 0, 0, 5, 5, 227, 223, 192, 0, 0, 0, 5, 5, 150, 225, 5, 5, 149, 225, 5, 5, 147, 225, 192, 0, 0, 0, 5, 5, 146, 225, 5, 5, 151, 225, 5, 5, 148, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 207, 5, 5, 106, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 209, 5, 5, 106, 230, 192, 0, 0, 0, 5, 5, 107, 230, 5, 5, 108, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 232, 5, 5, 255, 232, 5, 5, 253, 232, 192, 0, 0, 0, 5, 5, 251, 232, 5, 5, 252, 232, 5, 5, 213, 210, 5, 5, 162, 235, 5, 5, 231, 237, 5, 5, 230, 237, 5, 5, 229, 237, 5, 5, 232, 237, 5, 5, 81, 240, 5, 5, 80, 240, 5, 5, 110, 215, 5, 5, 82, 240, 192, 0, 0, 0, 5, 5, 133, 216, 5, 5, 81, 242, 192, 0, 0, 0, 5, 5, 10, 244, 5, 5, 118, 245, 5, 5, 90, 248, 192, 0, 0, 0, 5, 5, 84, 199, 5, 5, 192, 199, 5, 5, 191, 199, 192, 0, 0, 0, 5, 5, 148, 202, 5, 5, 38, 220, 5, 5, 67, 220, 5, 5, 88, 200, 192, 0, 0, 0, 5, 5, 154, 220, 5, 5, 74, 221, 5, 5, 75, 221, 5, 5, 149, 202, 192, 0, 0, 0, 5, 5, 107, 222, 192, 0, 0, 0, 5, 5, 231, 203, 5, 5, 228, 223, 5, 5, 229, 223, 192, 0, 0, 0, 5, 5, 127, 205, 5, 5, 131, 205, 5, 5, 152, 225, 5, 5, 130, 205, 5, 5, 128, 205, 5, 5, 129, 205, 192, 0, 0, 0, 5, 5, 236, 227, 5, 5, 107, 207, 5, 5, 40, 209, 192, 0, 0, 0, 5, 5, 38, 209, 5, 5, 39, 209, 5, 5, 109, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 212, 5, 5, 85, 199, 192, 0, 0, 0, 5, 5, 68, 220, 192, 0, 0, 0, 5, 5, 193, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 199, 5, 5, 70, 220, 5, 5, 195, 199, 5, 5, 194, 199, 5, 5, 63, 201, 5, 5, 69, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 220, 5, 5, 161, 220, 5, 5, 162, 220, 5, 5, 98, 200, 5, 5, 99, 200, 5, 5, 158, 220, 5, 5, 94, 200, 192, 0, 0, 0, 5, 5, 156, 220, 192, 0, 0, 0, 5, 5, 160, 220, 5, 5, 95, 200, 192, 0, 0, 0, 5, 5, 90, 200, 192, 0, 0, 0, 5, 5, 91, 200, 192, 0, 0, 0, 5, 5, 97, 200, 5, 5, 157, 220, 5, 5, 89, 200, 5, 5, 64, 201, 5, 5, 92, 200, 5, 5, 93, 200, 5, 5, 96, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 221, 5, 5, 90, 221, 5, 5, 77, 221, 5, 5, 76, 201, 5, 5, 84, 221, 5, 5, 71, 201, 5, 5, 78, 221, 192, 0, 0, 0, 5, 5, 86, 221, 192, 0, 0, 0, 5, 5, 83, 221, 5, 5, 74, 201, 5, 5, 82, 221, 5, 5, 81, 201, 5, 5, 91, 221, 5, 5, 83, 201, 192, 0, 0, 0, 5, 5, 85, 201, 192, 0, 0, 0, 5, 5, 76, 221, 192, 0, 0, 0, 5, 5, 72, 201, 5, 5, 93, 221, 192, 0, 0, 0, 5, 5, 79, 201, 5, 5, 82, 201, 192, 0, 0, 0, 5, 5, 130, 222, 5, 5, 66, 201, 5, 5, 89, 201, 5, 5, 80, 201, 5, 5, 79, 221, 5, 5, 69, 201, 5, 5, 84, 201, 5, 5, 87, 221, 5, 5, 67, 201, 5, 5, 68, 201, 5, 5, 128, 222, 5, 5, 80, 221, 5, 5, 75, 201, 5, 5, 86, 201, 5, 5, 94, 221, 5, 5, 81, 221, 5, 5, 73, 201, 192, 0, 0, 0, 5, 5, 78, 201, 5, 5, 108, 222, 5, 5, 87, 201, 5, 5, 88, 221, 5, 5, 77, 201, 192, 0, 0, 0, 5, 5, 88, 201, 5, 5, 65, 201, 5, 5, 85, 221, 5, 5, 70, 201, 5, 5, 89, 221, 5, 5, 129, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 202, 5, 5, 177, 202, 5, 5, 115, 222, 5, 5, 168, 202, 192, 0, 0, 0, 5, 5, 136, 222, 5, 5, 153, 202, 192, 0, 0, 0, 5, 5, 156, 202, 5, 5, 127, 222, 192, 0, 0, 0, 5, 5, 113, 222, 5, 5, 117, 222, 5, 5, 164, 202, 5, 5, 166, 202, 5, 5, 120, 222, 5, 5, 173, 202, 5, 5, 159, 202, 5, 5, 157, 202, 5, 5, 158, 202, 5, 5, 172, 202, 5, 5, 132, 222, 5, 5, 167, 202, 5, 5, 119, 222, 5, 5, 121, 222, 5, 5, 165, 202, 5, 5, 170, 202, 5, 5, 122, 222, 5, 5, 135, 222, 192, 0, 0, 0, 5, 5, 232, 203, 5, 5, 176, 202, 192, 0, 0, 0, 5, 5, 154, 202, 5, 5, 134, 222, 192, 0, 0, 0, 5, 5, 138, 222, 5, 5, 118, 222, 5, 5, 140, 222, 5, 5, 125, 222, 5, 5, 163, 202, 5, 5, 114, 222, 5, 5, 162, 202, 5, 5, 180, 202, 5, 5, 169, 202, 192, 0, 0, 0, 5, 5, 112, 222, 5, 5, 238, 223, 5, 5, 175, 202, 5, 5, 179, 202, 5, 5, 126, 222, 5, 5, 131, 222, 192, 0, 0, 0, 5, 5, 181, 202, 5, 5, 174, 202, 5, 5, 160, 202, 5, 5, 150, 202, 192, 0, 0, 0, 5, 5, 155, 202, 192, 0, 0, 0, 5, 5, 116, 222, 5, 5, 151, 202, 5, 5, 139, 222, 192, 0, 0, 0, 5, 5, 110, 222, 5, 5, 109, 222, 5, 5, 123, 222, 5, 5, 111, 222, 5, 5, 178, 202, 5, 5, 132, 205, 5, 5, 171, 202, 5, 5, 124, 222, 5, 5, 152, 202, 192, 0, 0, 0, 5, 5, 247, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 223, 5, 5, 248, 223, 192, 0, 0, 0, 5, 5, 251, 223, 5, 5, 240, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 223, 5, 5, 3, 224, 5, 5, 5, 224, 5, 5, 237, 223, 5, 5, 233, 203, 5, 5, 238, 203, 5, 5, 157, 225, 5, 5, 254, 203, 5, 5, 252, 223, 5, 5, 6, 224, 5, 5, 244, 223, 5, 5, 236, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 225, 5, 5, 241, 203, 5, 5, 249, 223, 5, 5, 241, 223, 5, 5, 243, 223, 5, 5, 246, 203, 192, 0, 0, 0, 5, 5, 246, 223, 5, 5, 240, 203, 5, 5, 233, 223, 5, 5, 255, 223, 192, 0, 0, 0, 5, 5, 4, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 203, 192, 0, 0, 0, 5, 5, 249, 203, 5, 5, 230, 223, 5, 5, 251, 203, 5, 5, 235, 203, 5, 5, 255, 203, 5, 5, 2, 224, 5, 5, 232, 223, 5, 5, 252, 203, 5, 5, 178, 225, 5, 5, 133, 222, 5, 5, 239, 203, 5, 5, 234, 203, 5, 5, 239, 223, 5, 5, 231, 223, 5, 5, 253, 203, 5, 5, 245, 203, 5, 5, 250, 223, 5, 5, 250, 203, 5, 5, 248, 203, 5, 5, 242, 223, 5, 5, 242, 203, 5, 5, 234, 223, 5, 5, 243, 203, 5, 5, 244, 203, 5, 5, 235, 223, 5, 5, 253, 223, 5, 5, 236, 203, 5, 5, 247, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 225, 5, 5, 140, 205, 5, 5, 145, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 225, 5, 5, 175, 225, 5, 5, 171, 225, 5, 5, 158, 225, 5, 5, 161, 225, 5, 5, 154, 225, 5, 5, 155, 225, 5, 5, 152, 205, 5, 5, 137, 205, 5, 5, 170, 225, 5, 5, 179, 225, 5, 5, 147, 205, 5, 5, 133, 205, 192, 0, 0, 0, 5, 5, 142, 205, 5, 5, 162, 225, 5, 5, 144, 205, 5, 5, 163, 225, 5, 5, 173, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 205, 5, 5, 188, 225, 5, 5, 156, 225, 5, 5, 139, 205, 5, 5, 138, 205, 5, 5, 150, 205, 5, 5, 153, 225, 5, 5, 186, 225, 5, 5, 174, 225, 5, 5, 187, 225, 5, 5, 181, 225, 5, 5, 167, 225, 5, 5, 182, 225, 192, 0, 0, 0, 5, 5, 176, 225, 5, 5, 185, 225, 5, 5, 183, 225, 5, 5, 151, 205, 5, 5, 168, 225, 5, 5, 136, 205, 5, 5, 135, 205, 5, 5, 143, 205, 5, 5, 149, 205, 5, 5, 180, 225, 5, 5, 148, 205, 5, 5, 165, 225, 5, 5, 108, 207, 192, 0, 0, 0, 5, 5, 189, 225, 5, 5, 164, 225, 5, 5, 159, 225, 5, 5, 141, 205, 5, 5, 153, 205, 5, 5, 134, 205, 192, 0, 0, 0, 5, 5, 172, 225, 5, 5, 177, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 207, 5, 5, 238, 227, 5, 5, 242, 227, 192, 0, 0, 0, 5, 5, 123, 207, 5, 5, 121, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 207, 5, 5, 240, 227, 5, 5, 239, 227, 5, 5, 132, 207, 192, 0, 0, 0, 5, 5, 245, 227, 5, 5, 126, 207, 192, 0, 0, 0, 5, 5, 9, 228, 5, 5, 21, 228, 5, 5, 109, 207, 5, 5, 6, 228, 5, 5, 253, 227, 5, 5, 144, 207, 5, 5, 237, 227, 192, 0, 0, 0, 5, 5, 11, 228, 192, 0, 0, 0, 5, 5, 141, 207, 5, 5, 129, 207, 5, 5, 140, 207, 5, 5, 119, 207, 5, 5, 249, 227, 5, 5, 13, 228, 5, 5, 5, 228, 5, 5, 120, 207, 5, 5, 114, 207, 5, 5, 19, 228, 192, 0, 0, 0, 5, 5, 12, 228, 5, 5, 14, 228, 5, 5, 122, 207, 5, 5, 130, 207, 5, 5, 16, 228, 5, 5, 247, 227, 5, 5, 10, 228, 5, 5, 252, 227, 5, 5, 18, 228, 5, 5, 135, 207, 5, 5, 111, 207, 5, 5, 133, 207, 5, 5, 3, 228, 5, 5, 255, 227, 5, 5, 2, 228, 5, 5, 124, 207, 5, 5, 251, 227, 5, 5, 250, 227, 5, 5, 113, 207, 5, 5, 244, 227, 5, 5, 20, 228, 5, 5, 115, 207, 5, 5, 254, 227, 5, 5, 145, 207, 192, 0, 0, 0, 5, 5, 139, 207, 5, 5, 243, 227, 5, 5, 136, 207, 5, 5, 134, 207, 5, 5, 143, 207, 5, 5, 7, 228, 5, 5, 138, 207, 5, 5, 166, 225, 5, 5, 8, 228, 5, 5, 137, 207, 5, 5, 15, 228, 5, 5, 110, 207, 5, 5, 4, 228, 5, 5, 128, 207, 5, 5, 246, 227, 5, 5, 127, 207, 192, 0, 0, 0, 5, 5, 125, 207, 5, 5, 117, 207, 5, 5, 116, 207, 5, 5, 110, 230, 5, 5, 17, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 230, 192, 0, 0, 0, 5, 5, 118, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 209, 5, 5, 131, 207, 5, 5, 51, 209, 5, 5, 125, 230, 5, 5, 65, 209, 192, 0, 0, 0, 5, 5, 113, 230, 5, 5, 48, 209, 5, 5, 44, 209, 5, 5, 143, 230, 5, 5, 50, 209, 5, 5, 54, 209, 5, 5, 49, 209, 5, 5, 58, 209, 5, 5, 147, 230, 5, 5, 138, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 230, 5, 5, 63, 209, 5, 5, 57, 209, 5, 5, 132, 230, 5, 5, 41, 209, 5, 5, 144, 230, 5, 5, 137, 230, 5, 5, 45, 209, 5, 5, 126, 230, 5, 5, 60, 209, 5, 5, 155, 230, 5, 5, 156, 230, 192, 0, 0, 0, 5, 5, 42, 209, 5, 5, 142, 230, 5, 5, 62, 209, 5, 5, 130, 230, 5, 5, 116, 230, 5, 5, 117, 230, 5, 5, 66, 209, 5, 5, 121, 230, 5, 5, 128, 230, 5, 5, 122, 230, 192, 0, 0, 0, 5, 5, 64, 209, 5, 5, 72, 209, 5, 5, 118, 230, 5, 5, 111, 230, 5, 5, 112, 230, 192, 0, 0, 0, 5, 5, 114, 230, 5, 5, 47, 209, 5, 5, 127, 230, 192, 0, 0, 0, 5, 5, 61, 209, 5, 5, 70, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 230, 5, 5, 153, 230, 5, 5, 145, 230, 5, 5, 43, 209, 5, 5, 151, 230, 5, 5, 55, 209, 192, 0, 0, 0, 5, 5, 53, 209, 192, 0, 0, 0, 5, 5, 157, 230, 5, 5, 52, 209, 5, 5, 135, 230, 5, 5, 123, 230, 5, 5, 133, 230, 5, 5, 75, 209, 5, 5, 139, 230, 5, 5, 136, 230, 5, 5, 119, 230, 5, 5, 71, 209, 5, 5, 149, 230, 5, 5, 146, 230, 5, 5, 154, 230, 5, 5, 46, 209, 5, 5, 134, 230, 5, 5, 74, 209, 192, 0, 0, 0, 5, 5, 141, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 209, 5, 5, 59, 209, 192, 0, 0, 0, 5, 5, 140, 230, 5, 5, 73, 209, 5, 5, 124, 230, 5, 5, 241, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 230, 5, 5, 148, 230, 5, 5, 152, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 230, 5, 5, 68, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 233, 5, 5, 31, 233, 5, 5, 5, 233, 5, 5, 219, 210, 192, 0, 0, 0, 5, 5, 30, 233, 5, 5, 8, 233, 5, 5, 9, 233, 192, 0, 0, 0, 5, 5, 229, 210, 5, 5, 39, 233, 5, 5, 224, 210, 5, 5, 29, 233, 192, 0, 0, 0, 5, 5, 2, 233, 5, 5, 230, 210, 5, 5, 220, 210, 5, 5, 17, 233, 5, 5, 7, 233, 5, 5, 10, 233, 5, 5, 34, 233, 5, 5, 214, 210, 5, 5, 41, 233, 5, 5, 33, 233, 5, 5, 223, 210, 5, 5, 22, 233, 5, 5, 234, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 210, 5, 5, 227, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 233, 5, 5, 215, 210, 5, 5, 20, 233, 5, 5, 11, 233, 5, 5, 24, 233, 5, 5, 36, 233, 5, 5, 235, 210, 192, 0, 0, 0, 5, 5, 217, 210, 5, 5, 19, 233, 192, 0, 0, 0, 5, 5, 12, 233, 5, 5, 226, 210, 192, 0, 0, 0, 5, 5, 225, 210, 5, 5, 15, 233, 5, 5, 25, 233, 5, 5, 35, 233, 5, 5, 6, 233, 5, 5, 16, 233, 5, 5, 218, 210, 5, 5, 26, 233, 5, 5, 231, 210, 5, 5, 222, 210, 5, 5, 13, 233, 5, 5, 221, 210, 5, 5, 4, 233, 5, 5, 18, 233, 5, 5, 38, 233, 5, 5, 67, 209, 5, 5, 169, 212, 5, 5, 21, 233, 5, 5, 163, 235, 5, 5, 23, 233, 5, 5, 37, 233, 5, 5, 228, 210, 5, 5, 14, 233, 5, 5, 216, 210, 5, 5, 232, 210, 5, 5, 39, 214, 5, 5, 3, 233, 192, 0, 0, 0, 5, 5, 28, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 235, 5, 5, 166, 212, 5, 5, 180, 235, 5, 5, 185, 235, 5, 5, 155, 212, 192, 0, 0, 0, 5, 5, 165, 235, 5, 5, 168, 212, 192, 0, 0, 0, 5, 5, 146, 212, 5, 5, 164, 235, 5, 5, 197, 235, 5, 5, 170, 212, 5, 5, 168, 235, 5, 5, 184, 235, 192, 0, 0, 0, 5, 5, 170, 235, 5, 5, 200, 235, 5, 5, 196, 235, 5, 5, 144, 212, 5, 5, 154, 212, 5, 5, 4, 238, 5, 5, 167, 212, 5, 5, 152, 212, 5, 5, 166, 235, 192, 0, 0, 0, 5, 5, 195, 235, 5, 5, 156, 212, 5, 5, 192, 235, 5, 5, 205, 235, 5, 5, 172, 235, 5, 5, 181, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 235, 5, 5, 193, 235, 5, 5, 151, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 235, 5, 5, 145, 212, 5, 5, 143, 212, 5, 5, 161, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 235, 5, 5, 174, 235, 5, 5, 175, 235, 192, 0, 0, 0, 5, 5, 199, 235, 192, 0, 0, 0, 5, 5, 204, 235, 5, 5, 202, 235, 5, 5, 149, 212, 5, 5, 206, 235, 5, 5, 153, 212, 5, 5, 160, 212, 192, 0, 0, 0, 5, 5, 167, 235, 5, 5, 234, 237, 5, 5, 176, 235, 192, 0, 0, 0, 5, 5, 147, 212, 5, 5, 165, 212, 5, 5, 162, 212, 5, 5, 150, 212, 5, 5, 186, 235, 5, 5, 171, 235, 5, 5, 163, 212, 5, 5, 188, 235, 5, 5, 157, 212, 5, 5, 159, 212, 5, 5, 142, 212, 192, 0, 0, 0, 5, 5, 190, 235, 5, 5, 182, 235, 5, 5, 169, 235, 5, 5, 158, 212, 5, 5, 198, 235, 5, 5, 201, 235, 5, 5, 178, 235, 5, 5, 189, 235, 192, 0, 0, 0, 5, 5, 148, 212, 5, 5, 22, 214, 5, 5, 187, 235, 5, 5, 233, 237, 192, 0, 0, 0, 5, 5, 194, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 235, 5, 5, 243, 237, 5, 5, 253, 237, 5, 5, 25, 214, 5, 5, 252, 237, 5, 5, 10, 238, 5, 5, 27, 214, 5, 5, 250, 237, 192, 0, 0, 0, 5, 5, 254, 237, 5, 5, 38, 214, 192, 0, 0, 0, 5, 5, 245, 237, 5, 5, 30, 214, 192, 0, 0, 0, 5, 5, 3, 238, 5, 5, 90, 240, 5, 5, 42, 214, 5, 5, 41, 214, 5, 5, 5, 238, 5, 5, 242, 237, 5, 5, 15, 238, 5, 5, 36, 214, 192, 0, 0, 0, 5, 5, 26, 214, 5, 5, 8, 238, 192, 0, 0, 0, 5, 5, 12, 238, 5, 5, 17, 238, 5, 5, 6, 238, 5, 5, 248, 237, 5, 5, 29, 214, 5, 5, 32, 214, 5, 5, 40, 214, 5, 5, 35, 214, 192, 0, 0, 0, 5, 5, 251, 237, 5, 5, 183, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 237, 5, 5, 16, 238, 5, 5, 31, 214, 192, 0, 0, 0, 5, 5, 34, 214, 5, 5, 18, 238, 5, 5, 23, 214, 5, 5, 7, 238, 5, 5, 235, 237, 5, 5, 13, 238, 5, 5, 142, 216, 192, 0, 0, 0, 5, 5, 249, 237, 192, 0, 0, 0, 5, 5, 24, 214, 5, 5, 244, 237, 5, 5, 28, 214, 5, 5, 236, 237, 5, 5, 164, 212, 5, 5, 240, 237, 192, 0, 0, 0, 5, 5, 11, 238, 5, 5, 241, 237, 5, 5, 239, 237, 5, 5, 33, 214, 192, 0, 0, 0, 5, 5, 9, 238, 192, 0, 0, 0, 5, 5, 238, 237, 5, 5, 2, 238, 5, 5, 255, 237, 5, 5, 14, 238, 5, 5, 246, 237, 5, 5, 37, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 240, 192, 0, 0, 0, 5, 5, 123, 215, 5, 5, 113, 215, 5, 5, 105, 240, 5, 5, 85, 240, 5, 5, 115, 215, 5, 5, 97, 240, 5, 5, 122, 215, 5, 5, 117, 215, 5, 5, 95, 240, 5, 5, 83, 242, 5, 5, 101, 240, 5, 5, 107, 240, 5, 5, 100, 240, 5, 5, 83, 240, 5, 5, 98, 240, 5, 5, 109, 240, 5, 5, 111, 240, 5, 5, 112, 215, 5, 5, 110, 240, 5, 5, 118, 215, 5, 5, 124, 215, 192, 0, 0, 0, 5, 5, 121, 215, 192, 0, 0, 0, 5, 5, 104, 240, 5, 5, 120, 215, 5, 5, 99, 240, 192, 0, 0, 0, 5, 5, 87, 240, 5, 5, 92, 240, 192, 0, 0, 0, 5, 5, 103, 240, 5, 5, 119, 215, 5, 5, 116, 215, 5, 5, 111, 215, 5, 5, 114, 215, 5, 5, 91, 240, 192, 0, 0, 0, 5, 5, 237, 237, 5, 5, 86, 240, 5, 5, 89, 240, 5, 5, 106, 240, 5, 5, 94, 240, 5, 5, 84, 240, 5, 5, 82, 242, 5, 5, 108, 240, 5, 5, 88, 240, 5, 5, 102, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 242, 5, 5, 146, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 216, 5, 5, 87, 242, 5, 5, 93, 242, 5, 5, 91, 242, 5, 5, 136, 216, 5, 5, 137, 216, 5, 5, 144, 216, 5, 5, 94, 242, 5, 5, 86, 242, 5, 5, 139, 216, 192, 0, 0, 0, 5, 5, 90, 242, 5, 5, 89, 242, 5, 5, 95, 242, 5, 5, 145, 216, 192, 0, 0, 0, 5, 5, 140, 216, 5, 5, 143, 216, 5, 5, 88, 242, 5, 5, 147, 216, 5, 5, 141, 216, 5, 5, 148, 216, 5, 5, 135, 216, 5, 5, 92, 242, 192, 0, 0, 0, 5, 5, 84, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 217, 5, 5, 20, 244, 5, 5, 22, 244, 192, 0, 0, 0, 5, 5, 124, 217, 5, 5, 18, 244, 5, 5, 19, 244, 5, 5, 14, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 240, 5, 5, 15, 244, 5, 5, 125, 217, 5, 5, 11, 244, 192, 0, 0, 0, 5, 5, 122, 217, 5, 5, 24, 244, 5, 5, 123, 217, 5, 5, 12, 244, 5, 5, 13, 244, 5, 5, 17, 244, 5, 5, 128, 217, 192, 0, 0, 0, 5, 5, 127, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 244, 5, 5, 36, 218, 5, 5, 122, 245, 5, 5, 128, 245, 5, 5, 37, 218, 5, 5, 119, 245, 5, 5, 34, 218, 5, 5, 31, 218, 5, 5, 130, 245, 5, 5, 35, 218, 192, 0, 0, 0, 5, 5, 32, 218, 5, 5, 121, 245, 5, 5, 124, 245, 5, 5, 125, 245, 5, 5, 126, 245, 5, 5, 129, 245, 192, 0, 0, 0, 5, 5, 21, 244, 5, 5, 120, 245, 5, 5, 33, 218, 5, 5, 127, 245, 5, 5, 131, 245, 5, 5, 123, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 246, 5, 5, 179, 218, 5, 5, 183, 246, 5, 5, 180, 218, 5, 5, 191, 246, 5, 5, 182, 246, 5, 5, 179, 246, 192, 0, 0, 0, 5, 5, 181, 246, 5, 5, 185, 246, 5, 5, 188, 246, 5, 5, 187, 246, 5, 5, 190, 246, 5, 5, 178, 246, 192, 0, 0, 0, 5, 5, 178, 218, 5, 5, 186, 246, 5, 5, 189, 246, 5, 5, 192, 246, 5, 5, 184, 246, 5, 5, 167, 247, 5, 5, 168, 247, 5, 5, 172, 247, 5, 5, 173, 247, 192, 0, 0, 0, 5, 5, 170, 247, 5, 5, 171, 247, 5, 5, 169, 247, 192, 0, 0, 0, 5, 5, 18, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 219, 5, 5, 94, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 248, 5, 5, 92, 248, 5, 5, 93, 248, 5, 5, 97, 219, 192, 0, 0, 0, 5, 5, 245, 248, 5, 5, 244, 248, 192, 0, 0, 0, 5, 5, 94, 249, 5, 5, 186, 219, 5, 5, 92, 249, 5, 5, 95, 249, 5, 5, 93, 249, 5, 5, 164, 249, 5, 5, 220, 219, 5, 5, 234, 219, 5, 5, 239, 249, 5, 5, 218, 249, 192, 0, 0, 0, 5, 5, 238, 249, 5, 5, 30, 250, 5, 5, 31, 250, 5, 5, 86, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 200, 5, 5, 163, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 221, 192, 0, 0, 0, 5, 5, 90, 201, 192, 0, 0, 0, 5, 5, 93, 201, 192, 0, 0, 0, 5, 5, 96, 221, 192, 0, 0, 0, 5, 5, 91, 201, 5, 5, 92, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 222, 5, 5, 151, 222, 5, 5, 146, 222, 5, 5, 143, 222, 5, 5, 145, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 222, 5, 5, 184, 202, 5, 5, 144, 222, 5, 5, 141, 222, 5, 5, 182, 202, 5, 5, 148, 222, 192, 0, 0, 0, 5, 5, 142, 222, 5, 5, 186, 202, 5, 5, 150, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 224, 192, 0, 0, 0, 5, 5, 12, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 224, 192, 0, 0, 0, 5, 5, 2, 204, 5, 5, 5, 204, 5, 5, 7, 204, 5, 5, 9, 204, 5, 5, 6, 204, 5, 5, 11, 224, 5, 5, 10, 224, 192, 0, 0, 0, 5, 5, 4, 204, 5, 5, 13, 224, 5, 5, 14, 224, 192, 0, 0, 0, 5, 5, 7, 224, 5, 5, 8, 204, 192, 0, 0, 0, 5, 5, 3, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 225, 5, 5, 204, 225, 5, 5, 205, 225, 5, 5, 158, 205, 192, 0, 0, 0, 5, 5, 154, 205, 5, 5, 194, 225, 192, 0, 0, 0, 5, 5, 202, 225, 5, 5, 207, 225, 5, 5, 159, 205, 192, 0, 0, 0, 5, 5, 192, 225, 5, 5, 198, 225, 5, 5, 191, 225, 5, 5, 201, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 225, 5, 5, 155, 205, 5, 5, 157, 205, 5, 5, 206, 225, 192, 0, 0, 0, 5, 5, 190, 225, 5, 5, 193, 225, 5, 5, 199, 225, 192, 0, 0, 0, 5, 5, 200, 225, 5, 5, 208, 225, 5, 5, 196, 225, 192, 0, 0, 0, 5, 5, 156, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 207, 5, 5, 28, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 228, 5, 5, 26, 228, 192, 0, 0, 0, 5, 5, 38, 228, 5, 5, 24, 228, 5, 5, 37, 228, 5, 5, 146, 207, 5, 5, 22, 228, 5, 5, 161, 230, 5, 5, 32, 228, 5, 5, 149, 207, 192, 0, 0, 0, 5, 5, 33, 228, 5, 5, 36, 228, 192, 0, 0, 0, 5, 5, 40, 228, 192, 0, 0, 0, 5, 5, 29, 228, 192, 0, 0, 0, 5, 5, 34, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 207, 5, 5, 148, 207, 5, 5, 39, 228, 5, 5, 27, 228, 5, 5, 23, 228, 5, 5, 41, 228, 192, 0, 0, 0, 5, 5, 31, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 228, 192, 0, 0, 0, 5, 5, 76, 209, 5, 5, 77, 209, 5, 5, 171, 230, 5, 5, 83, 209, 192, 0, 0, 0, 5, 5, 159, 230, 5, 5, 168, 230, 5, 5, 158, 230, 5, 5, 80, 209, 5, 5, 166, 230, 5, 5, 164, 230, 192, 0, 0, 0, 5, 5, 165, 230, 5, 5, 78, 209, 192, 0, 0, 0, 5, 5, 169, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 230, 5, 5, 160, 230, 5, 5, 79, 209, 5, 5, 163, 230, 5, 5, 167, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 233, 5, 5, 54, 233, 5, 5, 56, 233, 5, 5, 62, 233, 192, 0, 0, 0, 5, 5, 248, 210, 5, 5, 42, 233, 192, 0, 0, 0, 5, 5, 240, 210, 192, 0, 0, 0, 5, 5, 57, 233, 5, 5, 245, 210, 5, 5, 63, 233, 5, 5, 236, 210, 192, 0, 0, 0, 5, 5, 60, 233, 192, 0, 0, 0, 5, 5, 44, 233, 5, 5, 61, 233, 5, 5, 43, 233, 192, 0, 0, 0, 5, 5, 250, 210, 192, 0, 0, 0, 5, 5, 55, 233, 5, 5, 237, 210, 5, 5, 64, 233, 192, 0, 0, 0, 5, 5, 242, 210, 5, 5, 48, 233, 5, 5, 247, 210, 5, 5, 59, 233, 5, 5, 46, 233, 5, 5, 53, 233, 5, 5, 49, 233, 5, 5, 45, 233, 5, 5, 239, 210, 5, 5, 246, 210, 5, 5, 244, 210, 5, 5, 241, 210, 5, 5, 249, 210, 5, 5, 238, 210, 5, 5, 52, 233, 192, 0, 0, 0, 5, 5, 243, 210, 192, 0, 0, 0, 5, 5, 82, 209, 192, 0, 0, 0, 5, 5, 58, 233, 192, 0, 0, 0, 5, 5, 50, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 235, 192, 0, 0, 0, 5, 5, 173, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 235, 5, 5, 216, 235, 5, 5, 212, 235, 192, 0, 0, 0, 5, 5, 175, 212, 5, 5, 211, 235, 5, 5, 215, 235, 5, 5, 207, 235, 192, 0, 0, 0, 5, 5, 209, 235, 5, 5, 174, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 235, 5, 5, 208, 235, 192, 0, 0, 0, 5, 5, 176, 212, 192, 0, 0, 0, 5, 5, 171, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 235, 192, 0, 0, 0, 5, 5, 172, 212, 5, 5, 24, 238, 5, 5, 21, 238, 5, 5, 33, 238, 5, 5, 27, 238, 5, 5, 29, 238, 5, 5, 43, 214, 5, 5, 23, 238, 5, 5, 31, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 238, 5, 5, 28, 238, 192, 0, 0, 0, 5, 5, 34, 238, 5, 5, 46, 214, 5, 5, 25, 238, 5, 5, 32, 238, 192, 0, 0, 0, 5, 5, 44, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 238, 5, 5, 22, 238, 5, 5, 45, 214, 5, 5, 19, 238, 5, 5, 35, 238, 192, 0, 0, 0, 5, 5, 26, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 240, 5, 5, 131, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 240, 5, 5, 124, 240, 5, 5, 125, 215, 5, 5, 114, 240, 5, 5, 117, 240, 5, 5, 118, 240, 5, 5, 113, 240, 5, 5, 135, 215, 5, 5, 136, 215, 5, 5, 112, 240, 5, 5, 127, 240, 5, 5, 122, 240, 5, 5, 129, 215, 5, 5, 126, 215, 5, 5, 121, 240, 5, 5, 119, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 215, 5, 5, 123, 240, 5, 5, 127, 215, 192, 0, 0, 0, 5, 5, 128, 215, 192, 0, 0, 0, 5, 5, 120, 240, 5, 5, 130, 215, 5, 5, 116, 240, 192, 0, 0, 0, 5, 5, 125, 240, 5, 5, 133, 215, 5, 5, 128, 240, 5, 5, 129, 240, 5, 5, 134, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 216, 5, 5, 157, 216, 5, 5, 96, 242, 5, 5, 102, 242, 192, 0, 0, 0, 5, 5, 100, 242, 5, 5, 153, 216, 5, 5, 152, 216, 5, 5, 149, 216, 5, 5, 98, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 216, 5, 5, 154, 216, 5, 5, 151, 216, 192, 0, 0, 0, 5, 5, 101, 242, 5, 5, 97, 242, 5, 5, 99, 242, 192, 0, 0, 0, 5, 5, 156, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 217, 5, 5, 27, 244, 192, 0, 0, 0, 5, 5, 129, 217, 5, 5, 130, 217, 5, 5, 29, 244, 5, 5, 131, 217, 5, 5, 26, 244, 192, 0, 0, 0, 5, 5, 25, 244, 5, 5, 135, 245, 5, 5, 28, 244, 192, 0, 0, 0, 5, 5, 136, 245, 5, 5, 38, 218, 5, 5, 134, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 245, 192, 0, 0, 0, 5, 5, 132, 245, 5, 5, 39, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 246, 5, 5, 194, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 247, 5, 5, 175, 247, 5, 5, 19, 219, 192, 0, 0, 0, 5, 5, 174, 247, 5, 5, 95, 248, 5, 5, 96, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 248, 5, 5, 96, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 249, 5, 5, 10, 250, 5, 5, 254, 219, 5, 5, 20, 250, 5, 5, 87, 199, 192, 0, 0, 0, 5, 5, 187, 202, 5, 5, 188, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 216, 5, 5, 88, 199, 192, 0, 0, 0, 5, 5, 189, 202, 5, 5, 160, 205, 5, 5, 251, 210, 5, 5, 89, 199, 192, 0, 0, 0, 5, 5, 151, 207, 5, 5, 177, 212, 5, 5, 39, 220, 192, 0, 0, 0, 5, 5, 16, 224, 5, 5, 209, 225, 192, 0, 0, 0, 5, 5, 218, 235, 192, 0, 0, 0, 5, 5, 159, 216, 5, 5, 90, 199, 5, 5, 190, 202, 5, 5, 17, 224, 5, 5, 18, 224, 5, 5, 172, 230, 5, 5, 84, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 210, 5, 5, 219, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 214, 192, 0, 0, 0, 5, 5, 40, 218, 5, 5, 91, 199, 5, 5, 173, 230, 5, 5, 92, 199, 192, 0, 0, 0, 5, 5, 102, 200, 5, 5, 164, 220, 5, 5, 101, 200, 5, 5, 96, 201, 5, 5, 95, 201, 5, 5, 94, 201, 5, 5, 97, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 202, 192, 0, 0, 0, 5, 5, 192, 202, 5, 5, 152, 222, 192, 0, 0, 0, 5, 5, 19, 224, 192, 0, 0, 0, 5, 5, 22, 224, 5, 5, 13, 204, 5, 5, 20, 224, 192, 0, 0, 0, 5, 5, 12, 204, 5, 5, 21, 224, 5, 5, 14, 204, 192, 0, 0, 0, 5, 5, 212, 225, 5, 5, 211, 225, 5, 5, 210, 225, 5, 5, 161, 205, 192, 0, 0, 0, 5, 5, 43, 228, 5, 5, 44, 228, 5, 5, 152, 207, 5, 5, 42, 228, 5, 5, 45, 228, 5, 5, 86, 209, 5, 5, 153, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 209, 5, 5, 177, 230, 5, 5, 176, 230, 192, 0, 0, 0, 5, 5, 174, 230, 5, 5, 175, 230, 192, 0, 0, 0, 5, 5, 178, 230, 5, 5, 67, 233, 5, 5, 66, 233, 5, 5, 70, 233, 192, 0, 0, 0, 5, 5, 69, 233, 5, 5, 68, 233, 5, 5, 178, 212, 5, 5, 222, 235, 192, 0, 0, 0, 5, 5, 221, 235, 5, 5, 179, 212, 5, 5, 220, 235, 5, 5, 36, 238, 192, 0, 0, 0, 5, 5, 37, 238, 5, 5, 48, 214, 192, 0, 0, 0, 5, 5, 130, 240, 5, 5, 131, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 245, 5, 5, 41, 218, 5, 5, 140, 245, 5, 5, 139, 245, 5, 5, 137, 245, 5, 5, 138, 245, 5, 5, 20, 219, 5, 5, 195, 246, 5, 5, 97, 248, 5, 5, 166, 249, 192, 0, 0, 0, 5, 5, 93, 199, 192, 0, 0, 0, 5, 5, 72, 220, 5, 5, 197, 199, 5, 5, 73, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 220, 5, 5, 166, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 221, 192, 0, 0, 0, 5, 5, 98, 221, 5, 5, 193, 202, 5, 5, 102, 221, 5, 5, 98, 201, 5, 5, 100, 221, 5, 5, 97, 201, 5, 5, 104, 221, 5, 5, 101, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 222, 5, 5, 23, 224, 5, 5, 154, 222, 5, 5, 161, 222, 192, 0, 0, 0, 5, 5, 194, 202, 192, 0, 0, 0, 5, 5, 197, 202, 5, 5, 162, 222, 5, 5, 158, 222, 192, 0, 0, 0, 5, 5, 159, 222, 192, 0, 0, 0, 5, 5, 153, 222, 5, 5, 196, 202, 5, 5, 155, 222, 5, 5, 195, 202, 5, 5, 160, 222, 192, 0, 0, 0, 5, 5, 157, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 224, 5, 5, 16, 204, 5, 5, 17, 204, 192, 0, 0, 0, 5, 5, 30, 224, 5, 5, 24, 224, 192, 0, 0, 0, 5, 5, 29, 224, 192, 0, 0, 0, 5, 5, 25, 224, 5, 5, 15, 204, 5, 5, 28, 224, 5, 5, 26, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 225, 5, 5, 215, 225, 192, 0, 0, 0, 5, 5, 217, 225, 5, 5, 166, 205, 5, 5, 165, 205, 5, 5, 163, 205, 5, 5, 214, 225, 5, 5, 219, 225, 5, 5, 162, 205, 5, 5, 164, 205, 5, 5, 216, 225, 5, 5, 53, 228, 5, 5, 213, 225, 5, 5, 220, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 228, 5, 5, 52, 228, 192, 0, 0, 0, 5, 5, 51, 228, 5, 5, 180, 230, 5, 5, 191, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 228, 192, 0, 0, 0, 5, 5, 49, 228, 5, 5, 179, 230, 5, 5, 157, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 207, 5, 5, 47, 228, 5, 5, 50, 228, 5, 5, 158, 207, 192, 0, 0, 0, 5, 5, 155, 207, 5, 5, 154, 207, 5, 5, 46, 228, 5, 5, 55, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 230, 5, 5, 189, 230, 192, 0, 0, 0, 5, 5, 88, 209, 5, 5, 188, 230, 5, 5, 185, 230, 192, 0, 0, 0, 5, 5, 90, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 230, 5, 5, 183, 230, 5, 5, 186, 230, 5, 5, 184, 230, 5, 5, 89, 209, 5, 5, 190, 230, 5, 5, 87, 209, 5, 5, 253, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 233, 5, 5, 73, 233, 5, 5, 71, 233, 192, 0, 0, 0, 5, 5, 2, 211, 5, 5, 255, 210, 5, 5, 75, 233, 192, 0, 0, 0, 5, 5, 72, 233, 5, 5, 223, 235, 5, 5, 180, 212, 5, 5, 254, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 233, 5, 5, 76, 233, 192, 0, 0, 0, 5, 5, 226, 235, 5, 5, 224, 235, 5, 5, 49, 214, 192, 0, 0, 0, 5, 5, 181, 212, 5, 5, 225, 235, 5, 5, 39, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 214, 5, 5, 38, 238, 5, 5, 47, 238, 5, 5, 46, 238, 5, 5, 44, 238, 192, 0, 0, 0, 5, 5, 43, 238, 5, 5, 40, 238, 5, 5, 41, 238, 5, 5, 42, 238, 5, 5, 45, 238, 5, 5, 48, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 240, 5, 5, 133, 240, 5, 5, 134, 240, 5, 5, 137, 215, 5, 5, 132, 240, 5, 5, 138, 240, 5, 5, 137, 240, 5, 5, 135, 240, 192, 0, 0, 0, 5, 5, 104, 242, 5, 5, 105, 242, 5, 5, 160, 216, 192, 0, 0, 0, 5, 5, 161, 216, 5, 5, 103, 242, 192, 0, 0, 0, 5, 5, 134, 217, 5, 5, 30, 244, 5, 5, 133, 217, 5, 5, 42, 218, 192, 0, 0, 0, 5, 5, 43, 218, 5, 5, 182, 218, 5, 5, 197, 246, 5, 5, 196, 246, 5, 5, 177, 247, 5, 5, 98, 248, 5, 5, 98, 219, 5, 5, 248, 248, 5, 5, 247, 248, 5, 5, 249, 248, 5, 5, 198, 199, 5, 5, 31, 224, 5, 5, 167, 205, 5, 5, 159, 207, 5, 5, 56, 228, 5, 5, 199, 199, 5, 5, 74, 220, 5, 5, 94, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 221, 5, 5, 107, 221, 5, 5, 108, 221, 5, 5, 105, 221, 5, 5, 99, 201, 5, 5, 106, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 222, 192, 0, 0, 0, 5, 5, 200, 202, 5, 5, 168, 222, 5, 5, 164, 222, 5, 5, 167, 222, 192, 0, 0, 0, 5, 5, 163, 222, 5, 5, 202, 202, 5, 5, 166, 222, 192, 0, 0, 0, 5, 5, 199, 202, 5, 5, 198, 202, 192, 0, 0, 0, 5, 5, 201, 202, 5, 5, 169, 222, 5, 5, 165, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 204, 5, 5, 24, 204, 5, 5, 39, 224, 5, 5, 38, 224, 5, 5, 37, 224, 5, 5, 18, 204, 5, 5, 46, 224, 5, 5, 36, 224, 192, 0, 0, 0, 5, 5, 20, 204, 5, 5, 225, 225, 192, 0, 0, 0, 5, 5, 43, 224, 5, 5, 41, 224, 5, 5, 23, 204, 192, 0, 0, 0, 5, 5, 33, 224, 5, 5, 44, 224, 192, 0, 0, 0, 5, 5, 35, 224, 5, 5, 45, 224, 5, 5, 42, 224, 5, 5, 34, 224, 192, 0, 0, 0, 5, 5, 19, 204, 5, 5, 47, 224, 5, 5, 32, 224, 5, 5, 22, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 225, 5, 5, 221, 225, 5, 5, 232, 225, 192, 0, 0, 0, 5, 5, 224, 225, 5, 5, 235, 225, 5, 5, 236, 225, 5, 5, 222, 225, 5, 5, 234, 225, 5, 5, 231, 225, 5, 5, 229, 225, 5, 5, 233, 225, 5, 5, 173, 205, 192, 0, 0, 0, 5, 5, 171, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 225, 192, 0, 0, 0, 5, 5, 223, 225, 192, 0, 0, 0, 5, 5, 226, 225, 5, 5, 237, 225, 5, 5, 228, 225, 5, 5, 172, 205, 5, 5, 40, 224, 192, 0, 0, 0, 5, 5, 168, 205, 5, 5, 170, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 228, 5, 5, 59, 228, 5, 5, 57, 228, 192, 0, 0, 0, 5, 5, 58, 228, 192, 0, 0, 0, 5, 5, 65, 228, 192, 0, 0, 0, 5, 5, 66, 228, 5, 5, 62, 228, 5, 5, 164, 207, 5, 5, 67, 228, 5, 5, 64, 228, 5, 5, 61, 228, 192, 0, 0, 0, 5, 5, 162, 207, 5, 5, 60, 228, 5, 5, 160, 207, 5, 5, 163, 207, 5, 5, 63, 228, 5, 5, 71, 228, 5, 5, 169, 205, 5, 5, 161, 207, 5, 5, 69, 228, 5, 5, 68, 228, 5, 5, 165, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 230, 5, 5, 100, 209, 192, 0, 0, 0, 5, 5, 203, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 230, 5, 5, 198, 230, 5, 5, 94, 209, 5, 5, 202, 230, 5, 5, 201, 230, 5, 5, 95, 209, 5, 5, 101, 209, 192, 0, 0, 0, 5, 5, 102, 209, 5, 5, 204, 230, 5, 5, 92, 209, 5, 5, 195, 230, 5, 5, 193, 230, 5, 5, 199, 230, 5, 5, 192, 230, 5, 5, 99, 209, 5, 5, 194, 230, 5, 5, 200, 230, 5, 5, 206, 230, 5, 5, 93, 209, 5, 5, 98, 209, 5, 5, 96, 209, 5, 5, 97, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 211, 5, 5, 84, 233, 5, 5, 8, 211, 5, 5, 88, 233, 192, 0, 0, 0, 5, 5, 78, 233, 192, 0, 0, 0, 5, 5, 89, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 233, 5, 5, 80, 233, 192, 0, 0, 0, 5, 5, 90, 233, 5, 5, 87, 233, 5, 5, 85, 233, 5, 5, 86, 233, 5, 5, 82, 233, 5, 5, 81, 233, 192, 0, 0, 0, 5, 5, 91, 233, 5, 5, 5, 211, 192, 0, 0, 0, 5, 5, 83, 233, 192, 0, 0, 0, 5, 5, 10, 211, 5, 5, 4, 211, 5, 5, 11, 211, 5, 5, 12, 211, 192, 0, 0, 0, 5, 5, 7, 211, 5, 5, 6, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 235, 5, 5, 183, 212, 5, 5, 182, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 235, 192, 0, 0, 0, 5, 5, 51, 214, 5, 5, 184, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 212, 5, 5, 233, 235, 5, 5, 3, 211, 5, 5, 185, 212, 5, 5, 229, 235, 5, 5, 231, 235, 5, 5, 228, 235, 192, 0, 0, 0, 5, 5, 230, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 238, 5, 5, 54, 238, 5, 5, 54, 214, 5, 5, 139, 240, 5, 5, 55, 214, 5, 5, 53, 238, 192, 0, 0, 0, 5, 5, 53, 214, 192, 0, 0, 0, 5, 5, 55, 238, 5, 5, 52, 238, 5, 5, 49, 238, 5, 5, 56, 238, 5, 5, 50, 238, 5, 5, 51, 238, 5, 5, 52, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 242, 192, 0, 0, 0, 5, 5, 143, 240, 192, 0, 0, 0, 5, 5, 142, 240, 5, 5, 144, 240, 192, 0, 0, 0, 5, 5, 106, 242, 5, 5, 140, 215, 192, 0, 0, 0, 5, 5, 140, 240, 192, 0, 0, 0, 5, 5, 138, 215, 192, 0, 0, 0, 5, 5, 142, 215, 5, 5, 141, 215, 5, 5, 141, 240, 5, 5, 145, 240, 192, 0, 0, 0, 5, 5, 139, 215, 192, 0, 0, 0, 5, 5, 113, 242, 5, 5, 164, 216, 5, 5, 135, 217, 5, 5, 165, 216, 5, 5, 162, 216, 5, 5, 110, 242, 5, 5, 108, 242, 192, 0, 0, 0, 5, 5, 111, 242, 192, 0, 0, 0, 5, 5, 114, 242, 5, 5, 163, 216, 5, 5, 112, 242, 5, 5, 107, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 244, 5, 5, 36, 244, 5, 5, 144, 245, 5, 5, 31, 244, 192, 0, 0, 0, 5, 5, 198, 246, 5, 5, 37, 244, 192, 0, 0, 0, 5, 5, 44, 218, 5, 5, 35, 244, 5, 5, 136, 217, 5, 5, 32, 244, 5, 5, 34, 244, 5, 5, 38, 244, 5, 5, 145, 245, 192, 0, 0, 0, 5, 5, 143, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 218, 5, 5, 142, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 219, 5, 5, 100, 248, 5, 5, 21, 219, 5, 5, 102, 248, 5, 5, 99, 248, 5, 5, 101, 248, 5, 5, 150, 219, 5, 5, 97, 249, 5, 5, 200, 199, 5, 5, 171, 222, 5, 5, 238, 225, 5, 5, 239, 225, 5, 5, 166, 207, 5, 5, 92, 233, 5, 5, 143, 215, 5, 5, 46, 218, 5, 5, 99, 219, 5, 5, 98, 249, 5, 5, 201, 199, 192, 0, 0, 0, 5, 5, 172, 222, 5, 5, 0, 134, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 224, 192, 0, 0, 0, 5, 5, 49, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 225, 5, 5, 241, 225, 5, 5, 167, 207, 5, 5, 168, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 230, 192, 0, 0, 0, 5, 5, 95, 233, 5, 5, 94, 233, 5, 5, 93, 233, 5, 5, 234, 235, 192, 0, 0, 0, 5, 5, 235, 235, 5, 5, 236, 235, 5, 5, 187, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 238, 5, 5, 59, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 240, 5, 5, 144, 215, 5, 5, 145, 215, 192, 0, 0, 0, 5, 5, 118, 242, 5, 5, 115, 242, 5, 5, 116, 242, 5, 5, 117, 242, 5, 5, 40, 244, 5, 5, 39, 244, 5, 5, 137, 217, 5, 5, 146, 245, 5, 5, 178, 247, 5, 5, 202, 199, 192, 0, 0, 0, 5, 5, 25, 204, 192, 0, 0, 0, 5, 5, 169, 207, 5, 5, 96, 233, 192, 0, 0, 0, 5, 5, 203, 199, 192, 0, 0, 0, 5, 5, 242, 225, 5, 5, 170, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 209, 5, 5, 104, 209, 192, 0, 0, 0, 5, 5, 204, 199, 5, 5, 205, 199, 5, 5, 168, 220, 5, 5, 101, 201, 5, 5, 100, 201, 5, 5, 26, 204, 5, 5, 50, 224, 5, 5, 208, 230, 5, 5, 206, 199, 5, 5, 207, 199, 5, 5, 208, 199, 5, 5, 209, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 201, 5, 5, 103, 201, 5, 5, 111, 221, 5, 5, 110, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 202, 5, 5, 175, 222, 5, 5, 173, 222, 5, 5, 174, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 224, 5, 5, 52, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 204, 5, 5, 28, 204, 192, 0, 0, 0, 5, 5, 29, 204, 5, 5, 27, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 205, 5, 5, 177, 205, 5, 5, 243, 225, 5, 5, 176, 205, 5, 5, 175, 205, 192, 0, 0, 0, 5, 5, 244, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 207, 5, 5, 73, 228, 5, 5, 72, 228, 5, 5, 171, 207, 5, 5, 172, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 209, 5, 5, 105, 209, 5, 5, 210, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 230, 5, 5, 174, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 211, 5, 5, 98, 233, 5, 5, 14, 211, 5, 5, 97, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 235, 5, 5, 61, 238, 5, 5, 56, 214, 5, 5, 147, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 242, 192, 0, 0, 0, 5, 5, 48, 218, 5, 5, 47, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 219, 5, 5, 210, 199, 5, 5, 176, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 207, 5, 5, 238, 235, 5, 5, 188, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 221, 5, 5, 113, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 222, 5, 5, 205, 202, 5, 5, 206, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 204, 5, 5, 35, 204, 5, 5, 32, 204, 5, 5, 33, 204, 192, 0, 0, 0, 5, 5, 53, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 224, 5, 5, 31, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 225, 192, 0, 0, 0, 5, 5, 182, 205, 5, 5, 183, 205, 192, 0, 0, 0, 5, 5, 178, 207, 5, 5, 251, 225, 192, 0, 0, 0, 5, 5, 188, 205, 5, 5, 186, 205, 5, 5, 252, 225, 5, 5, 247, 225, 5, 5, 185, 205, 5, 5, 184, 205, 5, 5, 179, 205, 5, 5, 250, 225, 5, 5, 249, 225, 5, 5, 246, 225, 5, 5, 187, 205, 192, 0, 0, 0, 5, 5, 248, 225, 5, 5, 180, 205, 192, 0, 0, 0, 5, 5, 181, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 207, 5, 5, 77, 228, 5, 5, 78, 228, 5, 5, 180, 207, 5, 5, 74, 228, 5, 5, 76, 228, 5, 5, 80, 228, 5, 5, 79, 228, 5, 5, 75, 228, 192, 0, 0, 0, 5, 5, 176, 207, 5, 5, 177, 207, 192, 0, 0, 0, 5, 5, 218, 230, 5, 5, 111, 209, 5, 5, 110, 209, 5, 5, 212, 230, 5, 5, 108, 209, 192, 0, 0, 0, 5, 5, 215, 230, 5, 5, 112, 209, 5, 5, 216, 230, 5, 5, 113, 209, 5, 5, 213, 230, 5, 5, 107, 209, 5, 5, 109, 209, 5, 5, 217, 230, 192, 0, 0, 0, 5, 5, 214, 230, 5, 5, 211, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 233, 192, 0, 0, 0, 5, 5, 100, 233, 5, 5, 16, 211, 5, 5, 19, 211, 5, 5, 18, 211, 5, 5, 23, 211, 5, 5, 22, 211, 5, 5, 112, 233, 5, 5, 110, 233, 5, 5, 103, 233, 5, 5, 107, 233, 5, 5, 106, 233, 5, 5, 20, 211, 5, 5, 109, 233, 5, 5, 105, 233, 5, 5, 113, 233, 5, 5, 104, 233, 5, 5, 21, 211, 5, 5, 15, 211, 5, 5, 17, 211, 192, 0, 0, 0, 5, 5, 102, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 235, 5, 5, 192, 212, 5, 5, 244, 235, 5, 5, 191, 212, 5, 5, 241, 235, 5, 5, 190, 212, 192, 0, 0, 0, 5, 5, 101, 233, 5, 5, 108, 233, 5, 5, 243, 235, 192, 0, 0, 0, 5, 5, 193, 212, 5, 5, 245, 235, 5, 5, 242, 235, 5, 5, 239, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 238, 5, 5, 68, 238, 5, 5, 70, 238, 5, 5, 66, 238, 5, 5, 65, 238, 5, 5, 63, 238, 5, 5, 59, 214, 5, 5, 57, 214, 5, 5, 62, 214, 5, 5, 63, 214, 5, 5, 67, 238, 5, 5, 60, 214, 5, 5, 62, 238, 5, 5, 61, 214, 5, 5, 189, 212, 5, 5, 69, 238, 5, 5, 58, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 240, 192, 0, 0, 0, 5, 5, 148, 240, 5, 5, 156, 240, 5, 5, 150, 240, 5, 5, 155, 240, 5, 5, 152, 240, 5, 5, 146, 215, 5, 5, 154, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 215, 192, 0, 0, 0, 5, 5, 148, 215, 192, 0, 0, 0, 5, 5, 153, 240, 5, 5, 151, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 216, 5, 5, 120, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 216, 5, 5, 123, 242, 5, 5, 121, 242, 5, 5, 122, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 244, 192, 0, 0, 0, 5, 5, 140, 217, 5, 5, 45, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 217, 5, 5, 46, 244, 5, 5, 139, 217, 5, 5, 43, 244, 5, 5, 47, 244, 192, 0, 0, 0, 5, 5, 41, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 218, 5, 5, 147, 245, 5, 5, 50, 218, 5, 5, 184, 218, 192, 0, 0, 0, 5, 5, 42, 244, 5, 5, 185, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 219, 5, 5, 179, 247, 192, 0, 0, 0, 5, 5, 102, 219, 5, 5, 103, 248, 5, 5, 101, 219, 192, 0, 0, 0, 5, 5, 250, 248, 5, 5, 187, 219, 5, 5, 188, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 204, 5, 5, 55, 224, 192, 0, 0, 0, 5, 5, 114, 209, 5, 5, 115, 209, 5, 5, 211, 199, 5, 5, 103, 200, 5, 5, 169, 220, 192, 0, 0, 0, 5, 5, 114, 221, 5, 5, 104, 201, 192, 0, 0, 0, 5, 5, 207, 202, 192, 0, 0, 0, 5, 5, 37, 204, 5, 5, 38, 204, 5, 5, 39, 204, 5, 5, 82, 228, 5, 5, 253, 225, 5, 5, 189, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 207, 5, 5, 81, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 230, 5, 5, 117, 209, 192, 0, 0, 0, 5, 5, 219, 230, 5, 5, 116, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 233, 5, 5, 64, 214, 5, 5, 74, 238, 5, 5, 71, 238, 5, 5, 72, 238, 5, 5, 73, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 242, 192, 0, 0, 0, 5, 5, 48, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 246, 5, 5, 199, 246, 192, 0, 0, 0, 5, 5, 104, 248, 5, 5, 212, 199, 5, 5, 178, 222, 5, 5, 190, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 209, 5, 5, 115, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 235, 192, 0, 0, 0, 5, 5, 65, 214, 5, 5, 157, 240, 192, 0, 0, 0, 5, 5, 49, 244, 5, 5, 201, 246, 5, 5, 213, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 202, 5, 5, 42, 204, 5, 5, 56, 224, 5, 5, 43, 204, 5, 5, 41, 204, 192, 0, 0, 0, 5, 5, 40, 204, 5, 5, 254, 225, 5, 5, 191, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 205, 5, 5, 193, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 207, 5, 5, 83, 228, 5, 5, 182, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 230, 5, 5, 184, 207, 5, 5, 119, 209, 5, 5, 116, 233, 5, 5, 24, 211, 5, 5, 25, 211, 192, 0, 0, 0, 5, 5, 194, 212, 192, 0, 0, 0, 5, 5, 195, 212, 5, 5, 66, 214, 5, 5, 150, 215, 5, 5, 158, 240, 5, 5, 149, 215, 192, 0, 0, 0, 5, 5, 125, 242, 5, 5, 169, 216, 192, 0, 0, 0, 5, 5, 50, 244, 5, 5, 202, 246, 5, 5, 214, 199, 5, 5, 105, 201, 5, 5, 181, 222, 5, 5, 180, 222, 5, 5, 209, 202, 5, 5, 179, 222, 5, 5, 210, 202, 5, 5, 182, 222, 192, 0, 0, 0, 5, 5, 62, 224, 5, 5, 46, 204, 5, 5, 45, 204, 5, 5, 64, 224, 5, 5, 63, 224, 5, 5, 50, 204, 192, 0, 0, 0, 5, 5, 49, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 204, 192, 0, 0, 0, 5, 5, 58, 224, 5, 5, 59, 224, 5, 5, 60, 224, 192, 0, 0, 0, 5, 5, 51, 204, 5, 5, 57, 224, 5, 5, 47, 204, 5, 5, 61, 224, 5, 5, 48, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 226, 5, 5, 7, 226, 5, 5, 5, 226, 5, 5, 4, 226, 192, 0, 0, 0, 5, 5, 8, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 226, 5, 5, 10, 226, 5, 5, 2, 226, 192, 0, 0, 0, 5, 5, 255, 225, 192, 0, 0, 0, 5, 5, 195, 205, 5, 5, 196, 205, 192, 0, 0, 0, 5, 5, 11, 226, 5, 5, 6, 226, 192, 0, 0, 0, 5, 5, 92, 228, 192, 0, 0, 0, 5, 5, 12, 226, 5, 5, 197, 205, 5, 5, 194, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 228, 192, 0, 0, 0, 5, 5, 85, 228, 192, 0, 0, 0, 5, 5, 87, 228, 5, 5, 88, 228, 5, 5, 90, 228, 5, 5, 89, 228, 5, 5, 94, 228, 5, 5, 188, 207, 5, 5, 185, 207, 5, 5, 189, 207, 5, 5, 84, 228, 5, 5, 190, 207, 5, 5, 93, 228, 5, 5, 187, 207, 5, 5, 91, 228, 5, 5, 186, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 230, 5, 5, 226, 230, 5, 5, 222, 230, 5, 5, 223, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 230, 5, 5, 229, 230, 5, 5, 230, 230, 5, 5, 225, 230, 5, 5, 228, 230, 5, 5, 120, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 233, 192, 0, 0, 0, 5, 5, 125, 233, 5, 5, 117, 233, 5, 5, 121, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 233, 5, 5, 122, 233, 5, 5, 26, 211, 5, 5, 34, 211, 192, 0, 0, 0, 5, 5, 29, 211, 5, 5, 118, 233, 5, 5, 119, 233, 5, 5, 199, 212, 5, 5, 37, 211, 5, 5, 30, 211, 192, 0, 0, 0, 5, 5, 35, 211, 5, 5, 28, 211, 5, 5, 124, 233, 5, 5, 36, 211, 5, 5, 123, 233, 5, 5, 32, 211, 5, 5, 27, 211, 5, 5, 33, 211, 5, 5, 127, 233, 5, 5, 251, 235, 5, 5, 253, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 235, 5, 5, 197, 212, 5, 5, 254, 235, 5, 5, 198, 212, 5, 5, 252, 235, 5, 5, 247, 235, 5, 5, 250, 235, 5, 5, 255, 235, 5, 5, 196, 212, 5, 5, 249, 235, 192, 0, 0, 0, 5, 5, 68, 214, 5, 5, 78, 238, 5, 5, 77, 238, 192, 0, 0, 0, 5, 5, 71, 214, 5, 5, 69, 214, 5, 5, 75, 238, 5, 5, 67, 214, 5, 5, 76, 238, 192, 0, 0, 0, 5, 5, 70, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 240, 192, 0, 0, 0, 5, 5, 168, 240, 192, 0, 0, 0, 5, 5, 167, 240, 5, 5, 159, 240, 5, 5, 163, 240, 5, 5, 162, 240, 5, 5, 160, 240, 5, 5, 152, 215, 5, 5, 153, 215, 5, 5, 151, 215, 5, 5, 161, 240, 5, 5, 164, 240, 5, 5, 165, 240, 192, 0, 0, 0, 5, 5, 154, 215, 192, 0, 0, 0, 5, 5, 174, 216, 5, 5, 133, 242, 192, 0, 0, 0, 5, 5, 171, 216, 5, 5, 127, 242, 5, 5, 173, 216, 5, 5, 175, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 216, 5, 5, 132, 242, 5, 5, 128, 242, 5, 5, 170, 216, 5, 5, 131, 242, 5, 5, 126, 242, 5, 5, 130, 242, 5, 5, 129, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 244, 5, 5, 143, 217, 5, 5, 144, 217, 5, 5, 141, 217, 192, 0, 0, 0, 5, 5, 142, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 244, 192, 0, 0, 0, 5, 5, 150, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 218, 192, 0, 0, 0, 5, 5, 148, 245, 5, 5, 149, 245, 192, 0, 0, 0, 5, 5, 203, 246, 5, 5, 206, 246, 5, 5, 204, 246, 5, 5, 205, 246, 5, 5, 180, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 219, 5, 5, 251, 248, 5, 5, 99, 249, 192, 0, 0, 0, 5, 5, 189, 219, 5, 5, 167, 249, 5, 5, 168, 249, 5, 5, 235, 219, 5, 5, 215, 199, 5, 5, 52, 204, 192, 0, 0, 0, 5, 5, 231, 230, 192, 0, 0, 0, 5, 5, 128, 233, 192, 0, 0, 0, 5, 5, 216, 199, 5, 5, 106, 201, 192, 0, 0, 0, 5, 5, 211, 202, 192, 0, 0, 0, 5, 5, 65, 224, 5, 5, 66, 224, 5, 5, 198, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 230, 5, 5, 121, 209, 5, 5, 38, 211, 5, 5, 176, 216, 5, 5, 134, 242, 5, 5, 151, 245, 5, 5, 207, 246, 5, 5, 217, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 222, 5, 5, 183, 222, 5, 5, 185, 222, 5, 5, 187, 222, 5, 5, 186, 222, 5, 5, 184, 222, 5, 5, 212, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 204, 5, 5, 75, 224, 192, 0, 0, 0, 5, 5, 70, 224, 5, 5, 67, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 204, 5, 5, 56, 204, 5, 5, 73, 224, 5, 5, 72, 224, 5, 5, 71, 224, 5, 5, 68, 224, 5, 5, 69, 224, 5, 5, 76, 224, 5, 5, 54, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 205, 192, 0, 0, 0, 5, 5, 208, 205, 5, 5, 207, 205, 5, 5, 21, 226, 5, 5, 15, 226, 5, 5, 13, 226, 192, 0, 0, 0, 5, 5, 205, 205, 5, 5, 101, 228, 5, 5, 200, 205, 5, 5, 18, 226, 5, 5, 22, 226, 5, 5, 24, 226, 5, 5, 20, 226, 5, 5, 14, 226, 5, 5, 206, 205, 5, 5, 19, 226, 5, 5, 17, 226, 5, 5, 199, 205, 5, 5, 25, 226, 5, 5, 209, 205, 5, 5, 23, 226, 5, 5, 203, 205, 5, 5, 16, 226, 192, 0, 0, 0, 5, 5, 204, 205, 5, 5, 201, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 207, 192, 0, 0, 0, 5, 5, 102, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 228, 5, 5, 97, 228, 5, 5, 99, 228, 192, 0, 0, 0, 5, 5, 100, 228, 5, 5, 98, 228, 5, 5, 193, 207, 192, 0, 0, 0, 5, 5, 103, 228, 192, 0, 0, 0, 5, 5, 96, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 230, 5, 5, 122, 209, 5, 5, 244, 230, 192, 0, 0, 0, 5, 5, 233, 230, 192, 0, 0, 0, 5, 5, 245, 230, 192, 0, 0, 0, 5, 5, 234, 230, 5, 5, 235, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 230, 5, 5, 242, 230, 5, 5, 239, 230, 5, 5, 191, 207, 5, 5, 123, 209, 5, 5, 237, 230, 5, 5, 240, 230, 5, 5, 124, 209, 5, 5, 241, 230, 5, 5, 238, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 233, 192, 0, 0, 0, 5, 5, 143, 233, 5, 5, 45, 211, 192, 0, 0, 0, 5, 5, 7, 236, 5, 5, 48, 211, 5, 5, 141, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 233, 5, 5, 133, 233, 5, 5, 135, 233, 5, 5, 136, 233, 5, 5, 129, 233, 192, 0, 0, 0, 5, 5, 44, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 211, 192, 0, 0, 0, 5, 5, 39, 211, 5, 5, 132, 233, 192, 0, 0, 0, 5, 5, 46, 211, 192, 0, 0, 0, 5, 5, 47, 211, 5, 5, 131, 233, 5, 5, 134, 233, 5, 5, 142, 233, 5, 5, 41, 211, 5, 5, 42, 211, 5, 5, 140, 233, 5, 5, 130, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 236, 5, 5, 201, 212, 5, 5, 11, 236, 5, 5, 137, 233, 5, 5, 13, 236, 5, 5, 205, 212, 5, 5, 14, 236, 5, 5, 10, 236, 192, 0, 0, 0, 5, 5, 202, 212, 5, 5, 6, 236, 5, 5, 204, 212, 5, 5, 3, 236, 5, 5, 8, 236, 5, 5, 12, 236, 5, 5, 5, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 211, 192, 0, 0, 0, 5, 5, 2, 236, 5, 5, 203, 212, 5, 5, 4, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 214, 5, 5, 80, 238, 5, 5, 78, 214, 192, 0, 0, 0, 5, 5, 76, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 212, 192, 0, 0, 0, 5, 5, 87, 238, 5, 5, 88, 238, 5, 5, 73, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 238, 5, 5, 89, 238, 5, 5, 75, 214, 5, 5, 72, 214, 5, 5, 82, 238, 5, 5, 79, 238, 5, 5, 84, 238, 5, 5, 81, 238, 5, 5, 79, 214, 5, 5, 83, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 238, 5, 5, 77, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 215, 5, 5, 175, 240, 192, 0, 0, 0, 5, 5, 169, 240, 5, 5, 173, 240, 5, 5, 179, 240, 5, 5, 180, 240, 5, 5, 176, 240, 5, 5, 177, 240, 5, 5, 174, 240, 192, 0, 0, 0, 5, 5, 171, 240, 192, 0, 0, 0, 5, 5, 158, 215, 5, 5, 155, 215, 5, 5, 170, 240, 5, 5, 172, 240, 192, 0, 0, 0, 5, 5, 157, 215, 5, 5, 178, 240, 192, 0, 0, 0, 5, 5, 180, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 242, 5, 5, 135, 242, 5, 5, 179, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 216, 192, 0, 0, 0, 5, 5, 142, 242, 5, 5, 178, 216, 5, 5, 138, 242, 5, 5, 139, 242, 5, 5, 136, 242, 5, 5, 143, 242, 5, 5, 53, 244, 192, 0, 0, 0, 5, 5, 181, 216, 5, 5, 137, 242, 192, 0, 0, 0, 5, 5, 144, 242, 5, 5, 141, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 244, 192, 0, 0, 0, 5, 5, 145, 217, 192, 0, 0, 0, 5, 5, 58, 244, 5, 5, 60, 244, 5, 5, 59, 244, 5, 5, 55, 244, 5, 5, 56, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 245, 192, 0, 0, 0, 5, 5, 52, 218, 192, 0, 0, 0, 5, 5, 153, 245, 5, 5, 155, 245, 5, 5, 152, 245, 5, 5, 157, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 245, 192, 0, 0, 0, 5, 5, 209, 246, 5, 5, 212, 246, 5, 5, 208, 246, 5, 5, 186, 218, 5, 5, 210, 246, 5, 5, 211, 246, 5, 5, 213, 246, 5, 5, 187, 218, 5, 5, 189, 218, 5, 5, 188, 218, 5, 5, 181, 247, 192, 0, 0, 0, 5, 5, 183, 247, 192, 0, 0, 0, 5, 5, 182, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 249, 5, 5, 169, 249, 5, 5, 218, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 221, 5, 5, 213, 202, 5, 5, 190, 222, 5, 5, 214, 202, 5, 5, 215, 202, 5, 5, 189, 222, 192, 0, 0, 0, 5, 5, 81, 224, 5, 5, 80, 224, 5, 5, 57, 204, 5, 5, 60, 204, 5, 5, 59, 204, 5, 5, 58, 204, 5, 5, 77, 224, 5, 5, 79, 224, 5, 5, 78, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 226, 5, 5, 211, 205, 5, 5, 32, 226, 5, 5, 31, 226, 5, 5, 30, 226, 5, 5, 26, 226, 5, 5, 210, 205, 5, 5, 214, 205, 5, 5, 217, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 205, 5, 5, 27, 226, 5, 5, 29, 226, 5, 5, 216, 205, 5, 5, 215, 205, 5, 5, 213, 205, 5, 5, 212, 205, 5, 5, 110, 228, 192, 0, 0, 0, 5, 5, 108, 228, 5, 5, 104, 228, 5, 5, 194, 207, 192, 0, 0, 0, 5, 5, 105, 228, 5, 5, 195, 207, 5, 5, 106, 228, 5, 5, 107, 228, 5, 5, 109, 228, 192, 0, 0, 0, 5, 5, 196, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 230, 192, 0, 0, 0, 5, 5, 248, 230, 5, 5, 247, 230, 5, 5, 246, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 233, 5, 5, 49, 211, 192, 0, 0, 0, 5, 5, 144, 233, 5, 5, 146, 233, 192, 0, 0, 0, 5, 5, 50, 211, 192, 0, 0, 0, 5, 5, 51, 211, 5, 5, 145, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 236, 192, 0, 0, 0, 5, 5, 16, 236, 5, 5, 17, 236, 192, 0, 0, 0, 5, 5, 208, 212, 5, 5, 206, 212, 5, 5, 207, 212, 5, 5, 25, 236, 192, 0, 0, 0, 5, 5, 24, 236, 5, 5, 21, 236, 5, 5, 20, 236, 5, 5, 19, 236, 5, 5, 18, 236, 5, 5, 22, 236, 5, 5, 15, 236, 192, 0, 0, 0, 5, 5, 90, 238, 5, 5, 95, 238, 5, 5, 93, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 238, 5, 5, 91, 238, 5, 5, 94, 238, 192, 0, 0, 0, 5, 5, 181, 240, 192, 0, 0, 0, 5, 5, 159, 215, 5, 5, 182, 216, 5, 5, 146, 242, 192, 0, 0, 0, 5, 5, 183, 216, 5, 5, 145, 242, 5, 5, 62, 244, 5, 5, 61, 244, 5, 5, 146, 217, 192, 0, 0, 0, 5, 5, 158, 245, 5, 5, 53, 218, 5, 5, 214, 246, 5, 5, 103, 219, 5, 5, 106, 248, 192, 0, 0, 0, 5, 5, 102, 249, 5, 5, 101, 249, 5, 5, 75, 220, 5, 5, 61, 204, 5, 5, 62, 204, 5, 5, 111, 228, 192, 0, 0, 0, 5, 5, 53, 211, 5, 5, 219, 199, 5, 5, 109, 201, 5, 5, 108, 201, 5, 5, 107, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 202, 5, 5, 216, 202, 192, 0, 0, 0, 5, 5, 65, 204, 192, 0, 0, 0, 5, 5, 83, 224, 5, 5, 86, 224, 5, 5, 84, 224, 192, 0, 0, 0, 5, 5, 63, 204, 5, 5, 64, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 224, 5, 5, 85, 224, 192, 0, 0, 0, 5, 5, 225, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 226, 5, 5, 41, 226, 5, 5, 40, 226, 5, 5, 222, 205, 5, 5, 35, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 205, 5, 5, 219, 205, 192, 0, 0, 0, 5, 5, 223, 205, 5, 5, 221, 205, 192, 0, 0, 0, 5, 5, 224, 205, 5, 5, 38, 226, 5, 5, 33, 226, 5, 5, 34, 226, 5, 5, 37, 226, 5, 5, 36, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 228, 5, 5, 115, 228, 5, 5, 113, 228, 192, 0, 0, 0, 5, 5, 112, 228, 5, 5, 197, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 209, 192, 0, 0, 0, 5, 5, 250, 230, 5, 5, 252, 230, 5, 5, 254, 230, 5, 5, 128, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 209, 192, 0, 0, 0, 5, 5, 251, 230, 5, 5, 127, 209, 5, 5, 253, 230, 5, 5, 125, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 233, 5, 5, 148, 233, 5, 5, 151, 233, 5, 5, 155, 233, 5, 5, 57, 211, 5, 5, 153, 233, 192, 0, 0, 0, 5, 5, 152, 233, 5, 5, 149, 233, 5, 5, 150, 233, 5, 5, 55, 211, 5, 5, 156, 233, 5, 5, 54, 211, 192, 0, 0, 0, 5, 5, 59, 211, 5, 5, 58, 211, 5, 5, 56, 211, 192, 0, 0, 0, 5, 5, 154, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 236, 192, 0, 0, 0, 5, 5, 30, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 212, 5, 5, 29, 236, 5, 5, 28, 236, 5, 5, 210, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 214, 192, 0, 0, 0, 5, 5, 97, 238, 192, 0, 0, 0, 5, 5, 85, 214, 5, 5, 81, 214, 5, 5, 83, 214, 192, 0, 0, 0, 5, 5, 80, 214, 5, 5, 82, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 240, 192, 0, 0, 0, 5, 5, 162, 215, 5, 5, 184, 240, 5, 5, 183, 240, 192, 0, 0, 0, 5, 5, 27, 236, 5, 5, 164, 215, 5, 5, 163, 215, 5, 5, 160, 215, 5, 5, 161, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 242, 192, 0, 0, 0, 5, 5, 149, 242, 5, 5, 184, 216, 5, 5, 150, 242, 192, 0, 0, 0, 5, 5, 152, 242, 5, 5, 148, 242, 5, 5, 147, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 244, 5, 5, 149, 217, 5, 5, 147, 217, 5, 5, 148, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 245, 5, 5, 160, 245, 5, 5, 55, 218, 192, 0, 0, 0, 5, 5, 54, 218, 5, 5, 216, 246, 5, 5, 217, 246, 5, 5, 215, 246, 192, 0, 0, 0, 5, 5, 107, 248, 5, 5, 108, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 199, 5, 5, 170, 220, 5, 5, 110, 201, 192, 0, 0, 0, 5, 5, 192, 222, 5, 5, 219, 202, 5, 5, 218, 202, 5, 5, 193, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 224, 5, 5, 66, 204, 5, 5, 87, 224, 5, 5, 67, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 205, 5, 5, 44, 226, 5, 5, 42, 226, 5, 5, 48, 226, 5, 5, 227, 205, 5, 5, 43, 226, 5, 5, 47, 226, 5, 5, 45, 226, 5, 5, 46, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 228, 5, 5, 118, 228, 192, 0, 0, 0, 5, 5, 198, 207, 192, 0, 0, 0, 5, 5, 117, 228, 5, 5, 199, 207, 5, 5, 132, 209, 5, 5, 131, 209, 5, 5, 130, 209, 5, 5, 255, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 233, 5, 5, 60, 211, 5, 5, 61, 211, 192, 0, 0, 0, 5, 5, 159, 233, 5, 5, 158, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 236, 5, 5, 212, 212, 5, 5, 211, 212, 5, 5, 33, 236, 5, 5, 34, 236, 192, 0, 0, 0, 5, 5, 87, 214, 5, 5, 86, 214, 192, 0, 0, 0, 5, 5, 188, 240, 5, 5, 98, 238, 5, 5, 100, 238, 5, 5, 99, 238, 5, 5, 187, 240, 5, 5, 185, 240, 5, 5, 189, 240, 5, 5, 186, 240, 192, 0, 0, 0, 5, 5, 165, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 242, 5, 5, 185, 216, 5, 5, 154, 242, 5, 5, 155, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 217, 5, 5, 151, 217, 192, 0, 0, 0, 5, 5, 190, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 219, 5, 5, 221, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 226, 5, 5, 228, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 208, 5, 5, 95, 208, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 209, 5, 5, 3, 231, 5, 5, 133, 209, 5, 5, 2, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 233, 192, 0, 0, 0, 5, 5, 213, 212, 5, 5, 35, 236, 5, 5, 214, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 218, 5, 5, 218, 246, 192, 0, 0, 0, 5, 5, 104, 200, 5, 5, 220, 202, 5, 5, 194, 222, 192, 0, 0, 0, 5, 5, 69, 204, 192, 0, 0, 0, 5, 5, 68, 204, 5, 5, 90, 224, 5, 5, 91, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 226, 5, 5, 53, 226, 5, 5, 229, 205, 192, 0, 0, 0, 5, 5, 55, 226, 5, 5, 58, 226, 5, 5, 56, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 226, 5, 5, 54, 226, 5, 5, 50, 226, 5, 5, 230, 205, 5, 5, 59, 226, 5, 5, 52, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 228, 5, 5, 205, 207, 5, 5, 136, 228, 5, 5, 202, 207, 192, 0, 0, 0, 5, 5, 129, 228, 5, 5, 206, 207, 192, 0, 0, 0, 5, 5, 200, 207, 192, 0, 0, 0, 5, 5, 124, 228, 5, 5, 137, 228, 5, 5, 125, 228, 5, 5, 122, 228, 5, 5, 204, 207, 192, 0, 0, 0, 5, 5, 201, 207, 192, 0, 0, 0, 5, 5, 128, 228, 5, 5, 131, 228, 5, 5, 203, 207, 5, 5, 132, 228, 5, 5, 207, 207, 5, 5, 133, 228, 5, 5, 123, 228, 5, 5, 130, 228, 5, 5, 134, 228, 5, 5, 126, 228, 5, 5, 121, 228, 5, 5, 119, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 231, 5, 5, 12, 231, 5, 5, 137, 209, 5, 5, 120, 228, 5, 5, 7, 231, 5, 5, 135, 209, 5, 5, 4, 231, 5, 5, 142, 209, 5, 5, 8, 231, 5, 5, 141, 209, 5, 5, 9, 231, 5, 5, 143, 209, 5, 5, 138, 209, 5, 5, 144, 209, 5, 5, 139, 209, 192, 0, 0, 0, 5, 5, 140, 209, 192, 0, 0, 0, 5, 5, 136, 209, 192, 0, 0, 0, 5, 5, 11, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 211, 5, 5, 172, 233, 192, 0, 0, 0, 5, 5, 175, 233, 5, 5, 163, 233, 5, 5, 168, 233, 5, 5, 162, 233, 5, 5, 66, 211, 192, 0, 0, 0, 5, 5, 166, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 233, 5, 5, 65, 211, 192, 0, 0, 0, 5, 5, 171, 233, 5, 5, 170, 233, 5, 5, 167, 233, 5, 5, 169, 233, 5, 5, 165, 233, 5, 5, 218, 212, 5, 5, 174, 233, 5, 5, 62, 211, 5, 5, 173, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 236, 192, 0, 0, 0, 5, 5, 225, 212, 5, 5, 46, 236, 192, 0, 0, 0, 5, 5, 224, 212, 5, 5, 36, 236, 192, 0, 0, 0, 5, 5, 38, 236, 5, 5, 217, 212, 5, 5, 43, 236, 5, 5, 42, 236, 5, 5, 45, 236, 5, 5, 222, 212, 5, 5, 40, 236, 5, 5, 39, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 212, 5, 5, 216, 212, 5, 5, 41, 236, 5, 5, 219, 212, 5, 5, 47, 236, 5, 5, 49, 236, 192, 0, 0, 0, 5, 5, 44, 236, 5, 5, 37, 236, 5, 5, 220, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 214, 5, 5, 215, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 238, 5, 5, 88, 214, 192, 0, 0, 0, 5, 5, 106, 238, 192, 0, 0, 0, 5, 5, 89, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 214, 5, 5, 109, 238, 192, 0, 0, 0, 5, 5, 101, 238, 5, 5, 223, 212, 5, 5, 107, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 238, 192, 0, 0, 0, 5, 5, 63, 211, 5, 5, 94, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 214, 192, 0, 0, 0, 5, 5, 92, 214, 5, 5, 93, 214, 192, 0, 0, 0, 5, 5, 168, 215, 5, 5, 108, 238, 5, 5, 102, 238, 5, 5, 96, 214, 192, 0, 0, 0, 5, 5, 105, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 240, 5, 5, 194, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 240, 5, 5, 166, 215, 5, 5, 196, 240, 5, 5, 170, 215, 5, 5, 202, 240, 5, 5, 193, 240, 5, 5, 190, 240, 5, 5, 205, 240, 5, 5, 191, 216, 5, 5, 171, 215, 5, 5, 201, 240, 5, 5, 191, 240, 5, 5, 169, 215, 5, 5, 195, 240, 5, 5, 173, 215, 5, 5, 192, 240, 5, 5, 197, 240, 5, 5, 172, 215, 5, 5, 200, 240, 5, 5, 203, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 242, 5, 5, 174, 242, 5, 5, 158, 242, 5, 5, 165, 242, 5, 5, 163, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 216, 5, 5, 170, 242, 5, 5, 198, 240, 192, 0, 0, 0, 5, 5, 161, 242, 192, 0, 0, 0, 5, 5, 171, 242, 5, 5, 188, 216, 5, 5, 160, 242, 5, 5, 159, 242, 5, 5, 169, 242, 5, 5, 166, 242, 5, 5, 168, 242, 192, 0, 0, 0, 5, 5, 156, 242, 5, 5, 172, 242, 5, 5, 186, 216, 192, 0, 0, 0, 5, 5, 167, 242, 5, 5, 176, 242, 5, 5, 164, 242, 5, 5, 190, 216, 5, 5, 187, 216, 5, 5, 162, 242, 5, 5, 157, 242, 5, 5, 175, 242, 5, 5, 167, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 244, 5, 5, 70, 244, 5, 5, 155, 217, 5, 5, 68, 244, 5, 5, 67, 244, 5, 5, 157, 217, 5, 5, 73, 244, 5, 5, 156, 217, 192, 0, 0, 0, 5, 5, 74, 244, 5, 5, 71, 244, 5, 5, 153, 217, 5, 5, 72, 244, 5, 5, 65, 244, 5, 5, 154, 217, 5, 5, 152, 217, 5, 5, 164, 245, 5, 5, 69, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 218, 5, 5, 58, 218, 5, 5, 163, 245, 192, 0, 0, 0, 5, 5, 165, 245, 5, 5, 162, 245, 5, 5, 59, 218, 5, 5, 56, 218, 5, 5, 57, 218, 5, 5, 61, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 218, 192, 0, 0, 0, 5, 5, 223, 246, 192, 0, 0, 0, 5, 5, 222, 246, 5, 5, 220, 246, 5, 5, 219, 246, 5, 5, 221, 246, 192, 0, 0, 0, 5, 5, 192, 218, 5, 5, 194, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 247, 5, 5, 184, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 248, 192, 0, 0, 0, 5, 5, 111, 248, 5, 5, 113, 248, 5, 5, 112, 248, 5, 5, 110, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 219, 5, 5, 104, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 219, 5, 5, 152, 219, 5, 5, 154, 219, 5, 5, 253, 248, 5, 5, 252, 248, 192, 0, 0, 0, 5, 5, 170, 249, 5, 5, 103, 249, 5, 5, 171, 249, 5, 5, 221, 219, 192, 0, 0, 0, 5, 5, 222, 219, 5, 5, 219, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 220, 5, 5, 105, 200, 192, 0, 0, 0, 5, 5, 195, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 224, 5, 5, 94, 224, 5, 5, 92, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 204, 192, 0, 0, 0, 5, 5, 95, 224, 5, 5, 96, 224, 5, 5, 97, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 226, 5, 5, 66, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 226, 5, 5, 231, 205, 5, 5, 62, 226, 192, 0, 0, 0, 5, 5, 63, 226, 5, 5, 65, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 226, 5, 5, 208, 207, 192, 0, 0, 0, 5, 5, 138, 228, 5, 5, 210, 207, 5, 5, 140, 228, 5, 5, 209, 207, 5, 5, 139, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 231, 5, 5, 145, 209, 192, 0, 0, 0, 5, 5, 16, 231, 5, 5, 13, 231, 5, 5, 141, 228, 192, 0, 0, 0, 5, 5, 146, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 233, 192, 0, 0, 0, 5, 5, 67, 211, 5, 5, 176, 233, 5, 5, 68, 211, 5, 5, 177, 233, 5, 5, 69, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 212, 5, 5, 55, 236, 5, 5, 52, 236, 5, 5, 54, 236, 5, 5, 227, 212, 5, 5, 228, 212, 5, 5, 53, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 238, 192, 0, 0, 0, 5, 5, 97, 214, 5, 5, 113, 238, 5, 5, 112, 238, 192, 0, 0, 0, 5, 5, 0, 135, 192, 0, 0, 0, 5, 5, 209, 240, 5, 5, 210, 240, 5, 5, 206, 240, 192, 0, 0, 0, 5, 5, 207, 240, 5, 5, 174, 215, 5, 5, 175, 215, 5, 5, 208, 240, 192, 0, 0, 0, 5, 5, 197, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 216, 5, 5, 198, 216, 5, 5, 194, 216, 5, 5, 196, 216, 5, 5, 192, 216, 192, 0, 0, 0, 5, 5, 195, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 217, 5, 5, 177, 242, 192, 0, 0, 0, 5, 5, 167, 245, 192, 0, 0, 0, 5, 5, 166, 245, 192, 0, 0, 0, 5, 5, 224, 246, 5, 5, 195, 218, 5, 5, 196, 218, 5, 5, 115, 248, 5, 5, 186, 247, 192, 0, 0, 0, 5, 5, 114, 248, 192, 0, 0, 0, 5, 5, 172, 249, 5, 5, 240, 249, 5, 5, 106, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 201, 192, 0, 0, 0, 5, 5, 196, 222, 5, 5, 221, 202, 192, 0, 0, 0, 5, 5, 73, 204, 5, 5, 100, 224, 5, 5, 71, 204, 5, 5, 98, 224, 5, 5, 76, 204, 5, 5, 72, 204, 5, 5, 77, 204, 5, 5, 75, 204, 5, 5, 99, 224, 5, 5, 74, 204, 5, 5, 235, 205, 5, 5, 234, 205, 5, 5, 77, 226, 5, 5, 243, 205, 5, 5, 70, 226, 5, 5, 76, 226, 5, 5, 239, 205, 5, 5, 69, 226, 5, 5, 75, 226, 5, 5, 73, 226, 5, 5, 238, 205, 5, 5, 240, 205, 5, 5, 72, 226, 5, 5, 233, 205, 5, 5, 71, 226, 5, 5, 244, 205, 5, 5, 241, 205, 5, 5, 245, 205, 5, 5, 242, 205, 5, 5, 68, 226, 5, 5, 67, 226, 5, 5, 74, 226, 5, 5, 236, 205, 5, 5, 232, 205, 5, 5, 237, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 228, 5, 5, 149, 228, 192, 0, 0, 0, 5, 5, 151, 209, 5, 5, 148, 228, 192, 0, 0, 0, 5, 5, 214, 207, 5, 5, 221, 207, 5, 5, 218, 207, 5, 5, 224, 207, 5, 5, 223, 207, 5, 5, 219, 207, 192, 0, 0, 0, 5, 5, 142, 228, 5, 5, 145, 228, 192, 0, 0, 0, 5, 5, 144, 228, 5, 5, 215, 207, 5, 5, 146, 228, 5, 5, 155, 228, 5, 5, 216, 207, 5, 5, 143, 228, 5, 5, 152, 228, 5, 5, 153, 228, 5, 5, 217, 207, 5, 5, 150, 228, 5, 5, 222, 207, 5, 5, 212, 207, 5, 5, 220, 207, 5, 5, 147, 228, 5, 5, 211, 207, 5, 5, 151, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 231, 5, 5, 24, 231, 5, 5, 148, 209, 5, 5, 31, 231, 5, 5, 28, 231, 5, 5, 20, 231, 5, 5, 29, 231, 5, 5, 150, 209, 5, 5, 21, 231, 192, 0, 0, 0, 5, 5, 17, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 211, 5, 5, 26, 231, 192, 0, 0, 0, 5, 5, 147, 209, 5, 5, 32, 231, 192, 0, 0, 0, 5, 5, 154, 209, 5, 5, 156, 209, 5, 5, 19, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 209, 5, 5, 22, 231, 5, 5, 149, 209, 5, 5, 30, 231, 5, 5, 23, 231, 5, 5, 27, 231, 192, 0, 0, 0, 5, 5, 25, 231, 5, 5, 152, 209, 5, 5, 18, 231, 5, 5, 157, 209, 5, 5, 213, 207, 5, 5, 153, 209, 5, 5, 158, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 211, 5, 5, 185, 233, 5, 5, 187, 233, 5, 5, 189, 233, 5, 5, 193, 233, 192, 0, 0, 0, 5, 5, 183, 233, 5, 5, 181, 233, 5, 5, 73, 211, 192, 0, 0, 0, 5, 5, 188, 233, 5, 5, 192, 233, 5, 5, 184, 233, 5, 5, 180, 233, 192, 0, 0, 0, 5, 5, 179, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 233, 5, 5, 182, 233, 5, 5, 186, 233, 5, 5, 74, 211, 192, 0, 0, 0, 5, 5, 72, 211, 5, 5, 194, 233, 5, 5, 70, 211, 5, 5, 191, 233, 192, 0, 0, 0, 5, 5, 72, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 212, 5, 5, 64, 236, 5, 5, 110, 214, 5, 5, 74, 236, 5, 5, 234, 212, 5, 5, 78, 236, 5, 5, 240, 212, 5, 5, 59, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 236, 5, 5, 56, 236, 192, 0, 0, 0, 5, 5, 77, 236, 5, 5, 60, 236, 192, 0, 0, 0, 5, 5, 247, 212, 5, 5, 244, 212, 5, 5, 76, 236, 5, 5, 70, 236, 5, 5, 230, 212, 5, 5, 238, 212, 5, 5, 237, 212, 192, 0, 0, 0, 5, 5, 236, 212, 5, 5, 242, 212, 192, 0, 0, 0, 5, 5, 57, 236, 5, 5, 243, 212, 5, 5, 71, 236, 5, 5, 239, 212, 5, 5, 229, 212, 5, 5, 73, 236, 5, 5, 232, 212, 5, 5, 233, 212, 5, 5, 241, 212, 5, 5, 62, 236, 5, 5, 61, 236, 5, 5, 58, 236, 192, 0, 0, 0, 5, 5, 66, 236, 5, 5, 63, 236, 5, 5, 67, 236, 5, 5, 246, 212, 192, 0, 0, 0, 5, 5, 79, 236, 5, 5, 235, 212, 5, 5, 68, 236, 5, 5, 69, 236, 192, 0, 0, 0, 5, 5, 65, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 238, 5, 5, 102, 214, 5, 5, 111, 214, 5, 5, 107, 214, 5, 5, 115, 238, 192, 0, 0, 0, 5, 5, 104, 214, 5, 5, 108, 214, 5, 5, 127, 238, 5, 5, 98, 214, 5, 5, 119, 238, 192, 0, 0, 0, 5, 5, 106, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 238, 5, 5, 117, 238, 5, 5, 105, 214, 5, 5, 109, 214, 5, 5, 116, 238, 192, 0, 0, 0, 5, 5, 103, 214, 192, 0, 0, 0, 5, 5, 126, 238, 5, 5, 100, 214, 5, 5, 125, 238, 5, 5, 124, 238, 5, 5, 112, 214, 5, 5, 132, 237, 5, 5, 99, 214, 192, 0, 0, 0, 5, 5, 123, 238, 5, 5, 114, 238, 192, 0, 0, 0, 5, 5, 113, 214, 5, 5, 121, 238, 5, 5, 101, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 215, 5, 5, 183, 215, 5, 5, 176, 215, 5, 5, 223, 240, 5, 5, 214, 240, 5, 5, 226, 240, 5, 5, 218, 240, 5, 5, 224, 240, 5, 5, 184, 215, 5, 5, 177, 215, 5, 5, 211, 240, 5, 5, 217, 240, 5, 5, 227, 240, 5, 5, 220, 240, 5, 5, 225, 240, 5, 5, 213, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 240, 5, 5, 179, 215, 5, 5, 219, 240, 5, 5, 182, 215, 5, 5, 181, 215, 5, 5, 215, 240, 5, 5, 216, 240, 5, 5, 212, 240, 5, 5, 222, 240, 5, 5, 180, 215, 5, 5, 229, 240, 5, 5, 228, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 242, 5, 5, 184, 242, 5, 5, 205, 216, 192, 0, 0, 0, 5, 5, 178, 242, 5, 5, 199, 216, 5, 5, 215, 216, 5, 5, 190, 242, 5, 5, 207, 216, 5, 5, 203, 216, 5, 5, 181, 242, 5, 5, 210, 216, 5, 5, 213, 216, 5, 5, 192, 242, 5, 5, 202, 216, 5, 5, 183, 242, 5, 5, 211, 216, 5, 5, 194, 242, 5, 5, 191, 242, 5, 5, 179, 242, 5, 5, 206, 216, 5, 5, 200, 216, 5, 5, 214, 216, 5, 5, 186, 242, 5, 5, 209, 216, 5, 5, 180, 242, 5, 5, 204, 216, 5, 5, 193, 242, 5, 5, 208, 216, 5, 5, 201, 216, 5, 5, 187, 242, 5, 5, 212, 216, 5, 5, 185, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 244, 5, 5, 83, 244, 5, 5, 164, 217, 5, 5, 86, 244, 5, 5, 159, 217, 5, 5, 160, 217, 5, 5, 78, 244, 5, 5, 85, 244, 5, 5, 80, 244, 5, 5, 165, 217, 5, 5, 162, 217, 192, 0, 0, 0, 5, 5, 76, 244, 192, 0, 0, 0, 5, 5, 161, 217, 5, 5, 82, 244, 5, 5, 84, 244, 5, 5, 163, 217, 5, 5, 81, 244, 5, 5, 79, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 245, 5, 5, 65, 218, 5, 5, 66, 218, 5, 5, 62, 218, 192, 0, 0, 0, 5, 5, 63, 218, 192, 0, 0, 0, 5, 5, 173, 245, 5, 5, 171, 245, 192, 0, 0, 0, 5, 5, 175, 245, 5, 5, 68, 218, 5, 5, 176, 245, 5, 5, 169, 245, 5, 5, 168, 245, 5, 5, 172, 245, 5, 5, 170, 245, 5, 5, 64, 218, 5, 5, 174, 245, 5, 5, 225, 246, 5, 5, 199, 218, 5, 5, 198, 218, 5, 5, 226, 246, 192, 0, 0, 0, 5, 5, 228, 246, 5, 5, 227, 246, 5, 5, 200, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 247, 5, 5, 188, 247, 5, 5, 189, 247, 192, 0, 0, 0, 5, 5, 187, 247, 5, 5, 190, 247, 5, 5, 27, 219, 5, 5, 192, 247, 192, 0, 0, 0, 5, 5, 26, 219, 192, 0, 0, 0, 5, 5, 116, 248, 192, 0, 0, 0, 5, 5, 155, 219, 5, 5, 157, 219, 5, 5, 254, 248, 5, 5, 156, 219, 5, 5, 104, 249, 5, 5, 174, 249, 5, 5, 176, 249, 5, 5, 173, 249, 5, 5, 175, 249, 5, 5, 241, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 200, 192, 0, 0, 0, 5, 5, 78, 204, 5, 5, 195, 225, 5, 5, 246, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 207, 5, 5, 34, 231, 5, 5, 35, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 240, 5, 5, 216, 216, 5, 5, 195, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 217, 192, 0, 0, 0, 5, 5, 179, 245, 5, 5, 178, 245, 5, 5, 201, 218, 5, 5, 193, 247, 192, 0, 0, 0, 5, 5, 117, 248, 5, 5, 190, 219, 5, 5, 171, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 202, 5, 5, 112, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 226, 5, 5, 78, 226, 5, 5, 82, 226, 5, 5, 80, 226, 5, 5, 247, 205, 5, 5, 81, 226, 5, 5, 79, 226, 192, 0, 0, 0, 5, 5, 157, 228, 192, 0, 0, 0, 5, 5, 36, 231, 5, 5, 37, 231, 5, 5, 197, 233, 5, 5, 198, 233, 5, 5, 77, 211, 5, 5, 78, 211, 5, 5, 196, 233, 5, 5, 199, 233, 5, 5, 195, 233, 5, 5, 76, 211, 192, 0, 0, 0, 5, 5, 248, 212, 192, 0, 0, 0, 5, 5, 79, 211, 5, 5, 80, 236, 192, 0, 0, 0, 5, 5, 114, 214, 5, 5, 128, 238, 5, 5, 115, 214, 192, 0, 0, 0, 5, 5, 185, 215, 5, 5, 233, 240, 5, 5, 231, 240, 5, 5, 232, 240, 5, 5, 198, 242, 5, 5, 197, 242, 5, 5, 196, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 245, 192, 0, 0, 0, 5, 5, 67, 218, 5, 5, 181, 245, 5, 5, 118, 248, 5, 5, 191, 219, 5, 5, 105, 249, 5, 5, 108, 200, 5, 5, 224, 202, 5, 5, 223, 202, 5, 5, 103, 224, 5, 5, 79, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 224, 5, 5, 85, 226, 192, 0, 0, 0, 5, 5, 248, 205, 5, 5, 158, 228, 5, 5, 84, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 207, 5, 5, 161, 228, 5, 5, 159, 228, 5, 5, 160, 228, 5, 5, 226, 207, 192, 0, 0, 0, 5, 5, 39, 231, 5, 5, 40, 231, 5, 5, 38, 231, 192, 0, 0, 0, 5, 5, 82, 211, 5, 5, 201, 233, 5, 5, 200, 233, 5, 5, 202, 233, 5, 5, 81, 211, 5, 5, 80, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 238, 5, 5, 131, 238, 192, 0, 0, 0, 5, 5, 116, 214, 5, 5, 130, 238, 5, 5, 234, 240, 5, 5, 186, 215, 5, 5, 88, 244, 192, 0, 0, 0, 5, 5, 87, 244, 5, 5, 69, 218, 5, 5, 182, 245, 5, 5, 71, 218, 5, 5, 70, 218, 5, 5, 229, 246, 5, 5, 194, 247, 5, 5, 28, 219, 5, 5, 109, 200, 5, 5, 104, 224, 5, 5, 80, 204, 5, 5, 88, 226, 5, 5, 250, 205, 5, 5, 87, 226, 5, 5, 86, 226, 192, 0, 0, 0, 5, 5, 249, 205, 192, 0, 0, 0, 5, 5, 167, 228, 192, 0, 0, 0, 5, 5, 169, 228, 5, 5, 162, 228, 5, 5, 163, 228, 5, 5, 228, 207, 5, 5, 164, 228, 5, 5, 229, 207, 5, 5, 168, 228, 5, 5, 165, 228, 5, 5, 166, 228, 5, 5, 230, 207, 192, 0, 0, 0, 5, 5, 160, 209, 5, 5, 161, 209, 192, 0, 0, 0, 5, 5, 41, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 233, 5, 5, 204, 233, 192, 0, 0, 0, 5, 5, 84, 236, 5, 5, 251, 212, 5, 5, 249, 212, 5, 5, 250, 212, 5, 5, 81, 236, 5, 5, 82, 236, 192, 0, 0, 0, 5, 5, 83, 236, 5, 5, 136, 238, 192, 0, 0, 0, 5, 5, 137, 238, 5, 5, 117, 214, 5, 5, 134, 238, 5, 5, 133, 238, 5, 5, 135, 238, 5, 5, 132, 238, 5, 5, 189, 215, 5, 5, 235, 240, 5, 5, 187, 215, 5, 5, 188, 215, 5, 5, 200, 242, 5, 5, 217, 216, 5, 5, 199, 242, 5, 5, 198, 243, 192, 0, 0, 0, 5, 5, 89, 244, 5, 5, 90, 244, 5, 5, 167, 217, 192, 0, 0, 0, 5, 5, 168, 217, 5, 5, 218, 216, 5, 5, 183, 245, 5, 5, 184, 245, 5, 5, 230, 246, 5, 5, 202, 218, 5, 5, 110, 200, 192, 0, 0, 0, 5, 5, 111, 200, 5, 5, 81, 204, 5, 5, 225, 202, 5, 5, 251, 205, 5, 5, 105, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 209, 5, 5, 112, 200, 5, 5, 83, 204, 5, 5, 106, 224, 5, 5, 107, 224, 5, 5, 82, 204, 5, 5, 84, 204, 5, 5, 113, 200, 192, 0, 0, 0, 5, 5, 108, 224, 5, 5, 253, 205, 5, 5, 89, 226, 5, 5, 255, 205, 5, 5, 252, 205, 5, 5, 254, 205, 192, 0, 0, 0, 5, 5, 172, 228, 5, 5, 231, 207, 192, 0, 0, 0, 5, 5, 171, 228, 5, 5, 170, 228, 192, 0, 0, 0, 5, 5, 205, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 236, 192, 0, 0, 0, 5, 5, 118, 214, 192, 0, 0, 0, 5, 5, 190, 215, 5, 5, 237, 240, 5, 5, 236, 240, 192, 0, 0, 0, 5, 5, 201, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 200, 5, 5, 116, 221, 5, 5, 197, 222, 5, 5, 85, 204, 5, 5, 109, 224, 192, 0, 0, 0, 5, 5, 91, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 206, 5, 5, 90, 226, 5, 5, 3, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 207, 5, 5, 173, 228, 5, 5, 175, 228, 192, 0, 0, 0, 5, 5, 232, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 231, 5, 5, 44, 231, 5, 5, 42, 231, 5, 5, 163, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 211, 192, 0, 0, 0, 5, 5, 84, 211, 192, 0, 0, 0, 5, 5, 253, 212, 192, 0, 0, 0, 5, 5, 87, 236, 5, 5, 86, 236, 5, 5, 252, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 216, 5, 5, 221, 216, 5, 5, 219, 216, 5, 5, 220, 216, 5, 5, 223, 216, 192, 0, 0, 0, 5, 5, 91, 244, 5, 5, 170, 217, 5, 5, 169, 217, 5, 5, 185, 245, 5, 5, 231, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 219, 5, 5, 106, 219, 5, 5, 115, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 225, 192, 0, 0, 0, 5, 5, 86, 211, 5, 5, 164, 209, 5, 5, 85, 211, 5, 5, 254, 212, 192, 0, 0, 0, 5, 5, 116, 200, 5, 5, 76, 220, 5, 5, 117, 200, 5, 5, 118, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 222, 5, 5, 119, 221, 192, 0, 0, 0, 5, 5, 120, 221, 5, 5, 114, 201, 192, 0, 0, 0, 5, 5, 117, 221, 5, 5, 113, 201, 192, 0, 0, 0, 5, 5, 116, 201, 5, 5, 118, 221, 5, 5, 118, 201, 5, 5, 117, 201, 5, 5, 121, 221, 5, 5, 115, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 202, 5, 5, 228, 202, 5, 5, 200, 222, 192, 0, 0, 0, 5, 5, 227, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 202, 5, 5, 234, 202, 5, 5, 231, 202, 192, 0, 0, 0, 5, 5, 203, 222, 5, 5, 199, 222, 5, 5, 235, 202, 192, 0, 0, 0, 5, 5, 229, 202, 5, 5, 119, 201, 192, 0, 0, 0, 5, 5, 233, 202, 5, 5, 202, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 222, 192, 0, 0, 0, 5, 5, 226, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 224, 5, 5, 89, 204, 5, 5, 90, 204, 5, 5, 117, 224, 192, 0, 0, 0, 5, 5, 111, 224, 5, 5, 114, 224, 5, 5, 123, 224, 5, 5, 121, 224, 192, 0, 0, 0, 5, 5, 91, 204, 5, 5, 127, 224, 5, 5, 94, 204, 5, 5, 124, 224, 5, 5, 116, 224, 5, 5, 113, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 231, 5, 5, 122, 224, 5, 5, 86, 204, 5, 5, 125, 224, 5, 5, 110, 224, 5, 5, 119, 224, 5, 5, 88, 204, 5, 5, 93, 204, 5, 5, 120, 224, 5, 5, 97, 204, 5, 5, 95, 204, 192, 0, 0, 0, 5, 5, 112, 224, 5, 5, 92, 204, 192, 0, 0, 0, 5, 5, 118, 224, 5, 5, 96, 204, 5, 5, 87, 204, 5, 5, 126, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 206, 192, 0, 0, 0, 5, 5, 17, 206, 5, 5, 6, 206, 5, 5, 4, 206, 5, 5, 93, 226, 5, 5, 12, 206, 5, 5, 9, 206, 5, 5, 95, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 206, 5, 5, 94, 226, 5, 5, 92, 226, 5, 5, 97, 226, 5, 5, 16, 206, 5, 5, 14, 206, 5, 5, 45, 231, 192, 0, 0, 0, 5, 5, 98, 226, 5, 5, 96, 226, 5, 5, 5, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 206, 5, 5, 10, 206, 192, 0, 0, 0, 5, 5, 13, 206, 192, 0, 0, 0, 5, 5, 15, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 228, 5, 5, 235, 207, 192, 0, 0, 0, 5, 5, 176, 228, 5, 5, 178, 228, 192, 0, 0, 0, 5, 5, 179, 228, 192, 0, 0, 0, 5, 5, 187, 228, 5, 5, 183, 228, 5, 5, 181, 228, 192, 0, 0, 0, 5, 5, 184, 228, 5, 5, 188, 228, 5, 5, 236, 207, 5, 5, 240, 207, 5, 5, 177, 228, 192, 0, 0, 0, 5, 5, 186, 228, 192, 0, 0, 0, 5, 5, 238, 207, 192, 0, 0, 0, 5, 5, 237, 207, 5, 5, 182, 228, 5, 5, 180, 228, 192, 0, 0, 0, 5, 5, 234, 207, 5, 5, 239, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 209, 5, 5, 54, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 231, 5, 5, 172, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 231, 5, 5, 219, 233, 192, 0, 0, 0, 5, 5, 171, 209, 5, 5, 51, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 231, 5, 5, 167, 209, 5, 5, 173, 209, 5, 5, 53, 231, 5, 5, 169, 209, 5, 5, 50, 231, 5, 5, 255, 212, 5, 5, 168, 209, 5, 5, 49, 231, 5, 5, 174, 209, 5, 5, 166, 209, 5, 5, 165, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 233, 5, 5, 209, 233, 192, 0, 0, 0, 5, 5, 215, 233, 192, 0, 0, 0, 5, 5, 207, 233, 5, 5, 220, 233, 5, 5, 212, 233, 192, 0, 0, 0, 5, 5, 206, 233, 5, 5, 90, 211, 5, 5, 96, 211, 5, 5, 217, 233, 192, 0, 0, 0, 5, 5, 210, 233, 192, 0, 0, 0, 5, 5, 93, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 211, 5, 5, 218, 233, 5, 5, 88, 211, 5, 5, 87, 211, 5, 5, 213, 233, 5, 5, 92, 211, 5, 5, 175, 209, 192, 0, 0, 0, 5, 5, 216, 233, 5, 5, 208, 233, 5, 5, 89, 211, 5, 5, 94, 211, 5, 5, 95, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 213, 5, 5, 2, 213, 192, 0, 0, 0, 5, 5, 7, 213, 5, 5, 90, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 236, 5, 5, 91, 236, 5, 5, 4, 213, 5, 5, 88, 236, 5, 5, 5, 213, 5, 5, 94, 236, 5, 5, 93, 236, 5, 5, 92, 236, 192, 0, 0, 0, 5, 5, 3, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 238, 192, 0, 0, 0, 5, 5, 146, 238, 5, 5, 124, 214, 5, 5, 145, 238, 5, 5, 123, 214, 5, 5, 119, 214, 5, 5, 120, 214, 5, 5, 121, 214, 5, 5, 142, 238, 5, 5, 141, 238, 5, 5, 122, 214, 192, 0, 0, 0, 5, 5, 144, 238, 5, 5, 140, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 240, 5, 5, 250, 240, 5, 5, 193, 215, 5, 5, 192, 215, 192, 0, 0, 0, 5, 5, 244, 240, 5, 5, 246, 240, 192, 0, 0, 0, 5, 5, 241, 240, 192, 0, 0, 0, 5, 5, 245, 240, 5, 5, 239, 240, 5, 5, 248, 240, 5, 5, 191, 215, 5, 5, 247, 240, 5, 5, 243, 240, 192, 0, 0, 0, 5, 5, 249, 240, 192, 0, 0, 0, 5, 5, 242, 240, 5, 5, 226, 216, 5, 5, 202, 242, 5, 5, 208, 242, 5, 5, 230, 216, 5, 5, 232, 216, 5, 5, 229, 216, 5, 5, 228, 216, 192, 0, 0, 0, 5, 5, 227, 216, 5, 5, 225, 216, 5, 5, 203, 242, 5, 5, 206, 242, 5, 5, 224, 216, 5, 5, 207, 242, 192, 0, 0, 0, 5, 5, 231, 216, 5, 5, 205, 242, 192, 0, 0, 0, 5, 5, 204, 242, 5, 5, 171, 217, 192, 0, 0, 0, 5, 5, 172, 217, 5, 5, 94, 244, 5, 5, 92, 244, 5, 5, 93, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 245, 192, 0, 0, 0, 5, 5, 186, 245, 5, 5, 72, 218, 5, 5, 233, 246, 5, 5, 203, 218, 5, 5, 232, 246, 192, 0, 0, 0, 5, 5, 196, 247, 5, 5, 119, 248, 5, 5, 108, 219, 5, 5, 177, 249, 5, 5, 178, 249, 5, 5, 158, 219, 5, 5, 119, 200, 192, 0, 0, 0, 5, 5, 236, 202, 5, 5, 55, 231, 5, 5, 8, 213, 5, 5, 233, 216, 5, 5, 209, 242, 5, 5, 120, 200, 192, 0, 0, 0, 5, 5, 19, 206, 5, 5, 18, 206, 5, 5, 56, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 240, 5, 5, 121, 200, 5, 5, 98, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 231, 5, 5, 58, 231, 5, 5, 59, 231, 5, 5, 9, 213, 5, 5, 194, 215, 5, 5, 122, 200, 192, 0, 0, 0, 5, 5, 237, 202, 5, 5, 128, 224, 5, 5, 20, 206, 5, 5, 99, 226, 5, 5, 241, 207, 192, 0, 0, 0, 5, 5, 60, 231, 5, 5, 97, 211, 192, 0, 0, 0, 5, 5, 10, 213, 5, 5, 195, 215, 5, 5, 234, 216, 5, 5, 173, 217, 5, 5, 234, 246, 5, 5, 123, 200, 5, 5, 238, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 206, 5, 5, 189, 228, 5, 5, 176, 209, 192, 0, 0, 0, 5, 5, 11, 213, 5, 5, 95, 236, 5, 5, 147, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 200, 5, 5, 177, 209, 5, 5, 221, 233, 5, 5, 12, 213, 5, 5, 125, 200, 5, 5, 204, 222, 5, 5, 129, 224, 5, 5, 99, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 206, 192, 0, 0, 0, 5, 5, 22, 206, 5, 5, 23, 206, 5, 5, 25, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 228, 5, 5, 191, 228, 5, 5, 193, 228, 5, 5, 242, 207, 5, 5, 244, 207, 5, 5, 243, 207, 5, 5, 190, 228, 5, 5, 245, 207, 5, 5, 192, 228, 192, 0, 0, 0, 5, 5, 61, 231, 5, 5, 62, 231, 192, 0, 0, 0, 5, 5, 63, 231, 5, 5, 224, 233, 192, 0, 0, 0, 5, 5, 225, 233, 192, 0, 0, 0, 5, 5, 223, 233, 5, 5, 226, 233, 192, 0, 0, 0, 5, 5, 98, 211, 192, 0, 0, 0, 5, 5, 222, 233, 192, 0, 0, 0, 5, 5, 13, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 238, 5, 5, 148, 238, 5, 5, 151, 238, 5, 5, 153, 238, 5, 5, 150, 238, 5, 5, 149, 238, 192, 0, 0, 0, 5, 5, 252, 240, 5, 5, 253, 240, 5, 5, 254, 240, 5, 5, 196, 215, 5, 5, 197, 215, 5, 5, 211, 242, 5, 5, 210, 242, 5, 5, 212, 242, 192, 0, 0, 0, 5, 5, 96, 244, 5, 5, 95, 244, 192, 0, 0, 0, 5, 5, 189, 245, 192, 0, 0, 0, 5, 5, 190, 245, 5, 5, 188, 245, 192, 0, 0, 0, 5, 5, 204, 218, 192, 0, 0, 0, 5, 5, 235, 246, 5, 5, 236, 246, 192, 0, 0, 0, 5, 5, 120, 248, 5, 5, 255, 248, 5, 5, 106, 249, 5, 5, 126, 200, 5, 5, 120, 201, 192, 0, 0, 0, 5, 5, 235, 216, 5, 5, 127, 200, 192, 0, 0, 0, 5, 5, 195, 228, 5, 5, 64, 231, 192, 0, 0, 0, 5, 5, 198, 219, 5, 5, 172, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 220, 5, 5, 175, 220, 5, 5, 128, 200, 5, 5, 176, 220, 5, 5, 174, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 221, 5, 5, 130, 221, 5, 5, 124, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 221, 5, 5, 122, 201, 192, 0, 0, 0, 5, 5, 123, 201, 5, 5, 125, 221, 5, 5, 123, 221, 5, 5, 122, 221, 5, 5, 126, 221, 5, 5, 121, 201, 5, 5, 127, 221, 5, 5, 130, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 222, 5, 5, 241, 202, 5, 5, 208, 222, 5, 5, 210, 222, 192, 0, 0, 0, 5, 5, 240, 202, 5, 5, 215, 222, 5, 5, 244, 202, 5, 5, 205, 222, 5, 5, 219, 222, 5, 5, 225, 222, 5, 5, 251, 202, 5, 5, 222, 222, 5, 5, 248, 202, 192, 0, 0, 0, 5, 5, 212, 222, 5, 5, 218, 222, 5, 5, 220, 222, 192, 0, 0, 0, 5, 5, 207, 222, 5, 5, 247, 202, 5, 5, 242, 202, 5, 5, 213, 222, 5, 5, 249, 202, 5, 5, 252, 202, 5, 5, 246, 202, 192, 0, 0, 0, 5, 5, 239, 202, 5, 5, 217, 222, 5, 5, 211, 222, 5, 5, 224, 222, 5, 5, 254, 202, 5, 5, 250, 202, 5, 5, 245, 202, 5, 5, 216, 222, 5, 5, 26, 206, 5, 5, 214, 222, 5, 5, 243, 202, 5, 5, 253, 202, 192, 0, 0, 0, 5, 5, 206, 222, 192, 0, 0, 0, 5, 5, 221, 222, 5, 5, 223, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 204, 5, 5, 110, 204, 5, 5, 118, 204, 5, 5, 115, 204, 5, 5, 137, 224, 5, 5, 140, 224, 5, 5, 111, 204, 192, 0, 0, 0, 5, 5, 131, 224, 192, 0, 0, 0, 5, 5, 104, 204, 5, 5, 114, 204, 192, 0, 0, 0, 5, 5, 117, 204, 5, 5, 119, 204, 5, 5, 152, 224, 5, 5, 143, 224, 192, 0, 0, 0, 5, 5, 103, 204, 5, 5, 151, 224, 5, 5, 107, 204, 5, 5, 105, 204, 5, 5, 100, 204, 5, 5, 135, 224, 192, 0, 0, 0, 5, 5, 150, 224, 5, 5, 139, 224, 5, 5, 142, 224, 5, 5, 155, 224, 192, 0, 0, 0, 5, 5, 120, 204, 5, 5, 149, 224, 5, 5, 112, 204, 5, 5, 144, 224, 5, 5, 154, 224, 5, 5, 141, 224, 5, 5, 145, 224, 5, 5, 148, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 224, 5, 5, 153, 224, 5, 5, 147, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 224, 192, 0, 0, 0, 5, 5, 136, 224, 5, 5, 113, 204, 5, 5, 108, 204, 5, 5, 101, 204, 5, 5, 106, 204, 5, 5, 102, 204, 5, 5, 121, 204, 5, 5, 134, 224, 5, 5, 118, 226, 5, 5, 109, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 224, 5, 5, 29, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 226, 5, 5, 40, 206, 192, 0, 0, 0, 5, 5, 105, 226, 192, 0, 0, 0, 5, 5, 116, 226, 5, 5, 112, 226, 192, 0, 0, 0, 5, 5, 131, 226, 192, 0, 0, 0, 5, 5, 123, 226, 192, 0, 0, 0, 5, 5, 113, 226, 192, 0, 0, 0, 5, 5, 122, 226, 5, 5, 107, 226, 5, 5, 111, 226, 5, 5, 134, 226, 5, 5, 43, 206, 5, 5, 126, 226, 5, 5, 117, 226, 5, 5, 27, 206, 5, 5, 132, 226, 5, 5, 103, 226, 192, 0, 0, 0, 5, 5, 125, 226, 192, 0, 0, 0, 5, 5, 42, 206, 5, 5, 37, 206, 5, 5, 102, 226, 5, 5, 35, 206, 5, 5, 34, 206, 5, 5, 39, 206, 5, 5, 124, 226, 5, 5, 31, 206, 5, 5, 38, 206, 5, 5, 138, 224, 5, 5, 65, 231, 5, 5, 119, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 226, 5, 5, 41, 206, 5, 5, 110, 226, 5, 5, 114, 226, 5, 5, 44, 206, 5, 5, 104, 226, 5, 5, 128, 226, 192, 0, 0, 0, 5, 5, 127, 226, 5, 5, 135, 226, 5, 5, 33, 206, 5, 5, 30, 206, 5, 5, 133, 226, 5, 5, 129, 226, 5, 5, 120, 226, 5, 5, 115, 226, 5, 5, 36, 206, 5, 5, 32, 206, 5, 5, 106, 226, 5, 5, 28, 206, 5, 5, 130, 226, 5, 5, 29, 206, 192, 0, 0, 0, 5, 5, 108, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 228, 5, 5, 203, 228, 5, 5, 208, 228, 5, 5, 228, 228, 5, 5, 5, 208, 5, 5, 249, 207, 192, 0, 0, 0, 5, 5, 200, 228, 5, 5, 6, 208, 5, 5, 7, 208, 5, 5, 211, 228, 5, 5, 218, 228, 5, 5, 224, 228, 192, 0, 0, 0, 5, 5, 205, 228, 192, 0, 0, 0, 5, 5, 212, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 208, 5, 5, 226, 228, 5, 5, 221, 228, 5, 5, 3, 208, 5, 5, 255, 207, 5, 5, 217, 228, 5, 5, 213, 228, 5, 5, 199, 228, 5, 5, 246, 207, 5, 5, 204, 228, 5, 5, 196, 228, 192, 0, 0, 0, 5, 5, 254, 207, 5, 5, 2, 208, 5, 5, 209, 228, 5, 5, 206, 228, 5, 5, 251, 207, 5, 5, 222, 228, 5, 5, 248, 207, 5, 5, 207, 228, 5, 5, 229, 228, 5, 5, 215, 228, 192, 0, 0, 0, 5, 5, 214, 228, 5, 5, 247, 207, 192, 0, 0, 0, 5, 5, 4, 208, 192, 0, 0, 0, 5, 5, 250, 207, 5, 5, 197, 228, 5, 5, 202, 228, 5, 5, 219, 228, 5, 5, 225, 228, 5, 5, 9, 208, 5, 5, 198, 228, 5, 5, 210, 228, 5, 5, 216, 228, 5, 5, 253, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 228, 5, 5, 220, 228, 5, 5, 223, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 207, 192, 0, 0, 0, 5, 5, 81, 231, 5, 5, 69, 231, 5, 5, 185, 209, 5, 5, 98, 231, 5, 5, 106, 231, 5, 5, 109, 231, 5, 5, 183, 209, 5, 5, 77, 231, 5, 5, 102, 231, 5, 5, 78, 231, 5, 5, 92, 231, 5, 5, 196, 209, 5, 5, 88, 231, 5, 5, 193, 209, 192, 0, 0, 0, 5, 5, 89, 231, 5, 5, 66, 231, 192, 0, 0, 0, 5, 5, 103, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 209, 5, 5, 100, 231, 5, 5, 90, 231, 5, 5, 111, 231, 5, 5, 85, 231, 5, 5, 4, 234, 192, 0, 0, 0, 5, 5, 114, 231, 5, 5, 200, 209, 5, 5, 83, 231, 5, 5, 95, 231, 5, 5, 203, 209, 5, 5, 182, 209, 5, 5, 87, 231, 5, 5, 112, 231, 5, 5, 80, 231, 5, 5, 73, 231, 5, 5, 84, 231, 192, 0, 0, 0, 5, 5, 72, 231, 5, 5, 70, 231, 5, 5, 178, 209, 5, 5, 104, 231, 5, 5, 79, 231, 5, 5, 107, 231, 192, 0, 0, 0, 5, 5, 108, 231, 5, 5, 186, 209, 5, 5, 191, 209, 5, 5, 187, 209, 5, 5, 195, 209, 5, 5, 99, 231, 5, 5, 188, 209, 5, 5, 91, 231, 5, 5, 75, 231, 192, 0, 0, 0, 5, 5, 180, 209, 5, 5, 67, 231, 5, 5, 101, 231, 5, 5, 110, 231, 5, 5, 74, 231, 5, 5, 194, 209, 5, 5, 115, 231, 5, 5, 86, 231, 192, 0, 0, 0, 5, 5, 82, 231, 192, 0, 0, 0, 5, 5, 179, 209, 5, 5, 199, 209, 192, 0, 0, 0, 5, 5, 97, 231, 5, 5, 201, 209, 192, 0, 0, 0, 5, 5, 93, 231, 5, 5, 190, 209, 5, 5, 184, 209, 5, 5, 192, 209, 5, 5, 181, 209, 5, 5, 198, 209, 5, 5, 94, 231, 5, 5, 76, 231, 5, 5, 96, 231, 5, 5, 71, 231, 5, 5, 105, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 234, 192, 0, 0, 0, 5, 5, 25, 234, 5, 5, 52, 211, 5, 5, 8, 234, 192, 0, 0, 0, 5, 5, 6, 234, 5, 5, 239, 233, 5, 5, 102, 211, 5, 5, 255, 233, 5, 5, 21, 234, 5, 5, 251, 233, 5, 5, 110, 211, 5, 5, 20, 234, 5, 5, 249, 233, 5, 5, 197, 209, 5, 5, 11, 234, 5, 5, 250, 233, 5, 5, 18, 234, 5, 5, 109, 211, 5, 5, 101, 211, 192, 0, 0, 0, 5, 5, 228, 233, 5, 5, 236, 233, 192, 0, 0, 0, 5, 5, 7, 234, 5, 5, 253, 233, 5, 5, 24, 234, 5, 5, 2, 234, 5, 5, 115, 211, 5, 5, 19, 234, 192, 0, 0, 0, 5, 5, 106, 211, 192, 0, 0, 0, 5, 5, 5, 234, 5, 5, 13, 234, 5, 5, 240, 233, 5, 5, 12, 234, 192, 0, 0, 0, 5, 5, 29, 234, 5, 5, 235, 233, 5, 5, 14, 234, 192, 0, 0, 0, 5, 5, 27, 234, 192, 0, 0, 0, 5, 5, 229, 233, 5, 5, 189, 209, 192, 0, 0, 0, 5, 5, 243, 233, 5, 5, 242, 233, 5, 5, 108, 211, 192, 0, 0, 0, 5, 5, 246, 233, 5, 5, 248, 233, 5, 5, 9, 234, 5, 5, 26, 234, 5, 5, 111, 211, 192, 0, 0, 0, 5, 5, 112, 211, 192, 0, 0, 0, 5, 5, 234, 233, 5, 5, 104, 211, 5, 5, 238, 233, 5, 5, 22, 234, 5, 5, 113, 211, 192, 0, 0, 0, 5, 5, 105, 211, 5, 5, 107, 211, 5, 5, 114, 211, 5, 5, 28, 234, 5, 5, 15, 234, 5, 5, 10, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 233, 5, 5, 244, 233, 5, 5, 103, 211, 5, 5, 230, 233, 5, 5, 100, 211, 5, 5, 254, 233, 5, 5, 231, 233, 5, 5, 252, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 233, 5, 5, 23, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 233, 5, 5, 17, 234, 5, 5, 232, 233, 5, 5, 26, 213, 5, 5, 135, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 236, 5, 5, 126, 236, 5, 5, 18, 213, 5, 5, 111, 236, 5, 5, 121, 236, 5, 5, 21, 213, 5, 5, 114, 236, 5, 5, 19, 213, 5, 5, 99, 236, 5, 5, 129, 236, 5, 5, 98, 236, 5, 5, 117, 236, 192, 0, 0, 0, 5, 5, 97, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 236, 5, 5, 124, 236, 5, 5, 122, 236, 5, 5, 110, 236, 5, 5, 105, 236, 5, 5, 103, 236, 192, 0, 0, 0, 5, 5, 104, 236, 5, 5, 123, 236, 5, 5, 134, 236, 5, 5, 112, 236, 5, 5, 20, 213, 192, 0, 0, 0, 5, 5, 107, 236, 192, 0, 0, 0, 5, 5, 127, 236, 192, 0, 0, 0, 5, 5, 23, 213, 5, 5, 106, 236, 5, 5, 100, 236, 5, 5, 116, 236, 5, 5, 27, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 213, 5, 5, 24, 213, 5, 5, 108, 236, 5, 5, 102, 236, 192, 0, 0, 0, 5, 5, 17, 213, 5, 5, 16, 234, 5, 5, 16, 213, 5, 5, 119, 236, 192, 0, 0, 0, 5, 5, 14, 213, 5, 5, 29, 213, 5, 5, 22, 213, 5, 5, 120, 236, 5, 5, 109, 236, 5, 5, 101, 236, 5, 5, 128, 236, 5, 5, 113, 236, 5, 5, 28, 213, 5, 5, 132, 236, 5, 5, 25, 213, 5, 5, 131, 236, 192, 0, 0, 0, 5, 5, 125, 236, 5, 5, 130, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 238, 5, 5, 194, 238, 5, 5, 181, 238, 5, 5, 180, 238, 5, 5, 176, 238, 5, 5, 136, 214, 192, 0, 0, 0, 5, 5, 128, 214, 5, 5, 196, 238, 5, 5, 195, 238, 5, 5, 227, 233, 5, 5, 173, 238, 5, 5, 177, 238, 5, 5, 171, 238, 192, 0, 0, 0, 5, 5, 190, 238, 5, 5, 175, 238, 192, 0, 0, 0, 5, 5, 197, 238, 5, 5, 168, 238, 5, 5, 166, 238, 5, 5, 178, 238, 5, 5, 188, 238, 5, 5, 183, 238, 5, 5, 138, 214, 5, 5, 157, 238, 192, 0, 0, 0, 5, 5, 187, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 214, 5, 5, 247, 233, 5, 5, 169, 238, 5, 5, 160, 238, 5, 5, 163, 238, 5, 5, 201, 238, 5, 5, 170, 238, 5, 5, 161, 238, 5, 5, 159, 238, 5, 5, 156, 238, 192, 0, 0, 0, 5, 5, 132, 214, 5, 5, 179, 238, 5, 5, 131, 214, 5, 5, 135, 214, 5, 5, 174, 238, 5, 5, 182, 238, 5, 5, 125, 214, 5, 5, 198, 238, 5, 5, 202, 238, 5, 5, 127, 214, 192, 0, 0, 0, 5, 5, 165, 238, 5, 5, 185, 238, 5, 5, 189, 238, 5, 5, 162, 238, 5, 5, 199, 238, 5, 5, 134, 214, 192, 0, 0, 0, 5, 5, 133, 214, 5, 5, 154, 238, 5, 5, 137, 214, 5, 5, 192, 238, 5, 5, 164, 238, 5, 5, 184, 238, 5, 5, 158, 238, 5, 5, 172, 238, 5, 5, 167, 238, 5, 5, 129, 214, 5, 5, 130, 214, 5, 5, 186, 238, 5, 5, 203, 238, 5, 5, 200, 238, 5, 5, 191, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 238, 192, 0, 0, 0, 5, 5, 126, 214, 5, 5, 17, 241, 192, 0, 0, 0, 5, 5, 8, 241, 5, 5, 11, 241, 192, 0, 0, 0, 5, 5, 203, 215, 5, 5, 13, 241, 5, 5, 2, 241, 5, 5, 9, 241, 5, 5, 15, 241, 5, 5, 200, 215, 5, 5, 204, 215, 5, 5, 198, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 241, 5, 5, 20, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 241, 192, 0, 0, 0, 5, 5, 5, 241, 5, 5, 29, 241, 5, 5, 23, 241, 5, 5, 255, 240, 5, 5, 231, 242, 5, 5, 7, 241, 5, 5, 199, 215, 192, 0, 0, 0, 5, 5, 18, 241, 192, 0, 0, 0, 5, 5, 28, 241, 5, 5, 207, 215, 192, 0, 0, 0, 5, 5, 25, 241, 5, 5, 6, 241, 5, 5, 12, 241, 5, 5, 16, 241, 5, 5, 10, 241, 5, 5, 30, 241, 5, 5, 27, 241, 5, 5, 24, 241, 5, 5, 201, 215, 5, 5, 202, 215, 5, 5, 206, 215, 5, 5, 3, 241, 5, 5, 31, 241, 5, 5, 205, 215, 5, 5, 21, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 241, 5, 5, 244, 242, 5, 5, 225, 242, 5, 5, 230, 242, 5, 5, 243, 242, 5, 5, 229, 242, 5, 5, 221, 242, 5, 5, 226, 242, 192, 0, 0, 0, 5, 5, 238, 216, 192, 0, 0, 0, 5, 5, 214, 242, 5, 5, 238, 242, 5, 5, 240, 242, 5, 5, 213, 242, 5, 5, 237, 216, 5, 5, 242, 242, 5, 5, 234, 242, 5, 5, 244, 216, 5, 5, 241, 242, 5, 5, 227, 242, 5, 5, 246, 216, 5, 5, 219, 242, 5, 5, 26, 241, 5, 5, 235, 242, 5, 5, 232, 242, 5, 5, 215, 242, 5, 5, 246, 242, 5, 5, 240, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 216, 5, 5, 217, 242, 5, 5, 233, 242, 192, 0, 0, 0, 5, 5, 245, 242, 5, 5, 236, 242, 5, 5, 223, 242, 5, 5, 243, 216, 5, 5, 239, 216, 5, 5, 237, 242, 5, 5, 224, 242, 5, 5, 247, 242, 5, 5, 218, 242, 5, 5, 228, 242, 5, 5, 239, 242, 5, 5, 220, 242, 5, 5, 222, 242, 192, 0, 0, 0, 5, 5, 182, 217, 5, 5, 216, 242, 5, 5, 245, 216, 5, 5, 175, 217, 5, 5, 236, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 216, 5, 5, 179, 217, 5, 5, 112, 244, 192, 0, 0, 0, 5, 5, 102, 244, 5, 5, 97, 244, 5, 5, 103, 244, 5, 5, 113, 244, 5, 5, 117, 244, 5, 5, 116, 244, 5, 5, 181, 217, 5, 5, 180, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 244, 5, 5, 118, 244, 5, 5, 107, 244, 5, 5, 99, 244, 192, 0, 0, 0, 5, 5, 101, 244, 5, 5, 100, 244, 5, 5, 106, 244, 5, 5, 111, 244, 5, 5, 98, 244, 5, 5, 105, 244, 5, 5, 110, 244, 5, 5, 178, 217, 192, 0, 0, 0, 5, 5, 108, 244, 192, 0, 0, 0, 5, 5, 176, 217, 5, 5, 109, 244, 5, 5, 174, 217, 5, 5, 177, 217, 5, 5, 206, 245, 5, 5, 114, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 218, 192, 0, 0, 0, 5, 5, 198, 245, 5, 5, 202, 245, 5, 5, 194, 245, 5, 5, 197, 245, 192, 0, 0, 0, 5, 5, 205, 245, 5, 5, 74, 218, 5, 5, 210, 245, 5, 5, 203, 245, 192, 0, 0, 0, 5, 5, 195, 245, 5, 5, 211, 245, 5, 5, 204, 245, 5, 5, 77, 218, 5, 5, 78, 218, 5, 5, 208, 245, 192, 0, 0, 0, 5, 5, 196, 245, 5, 5, 73, 218, 5, 5, 75, 218, 5, 5, 191, 245, 5, 5, 199, 245, 5, 5, 193, 245, 192, 0, 0, 0, 5, 5, 209, 245, 5, 5, 207, 245, 5, 5, 192, 245, 5, 5, 200, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 246, 5, 5, 79, 218, 5, 5, 201, 245, 5, 5, 206, 218, 5, 5, 208, 218, 5, 5, 205, 218, 192, 0, 0, 0, 5, 5, 248, 246, 5, 5, 240, 246, 5, 5, 238, 246, 5, 5, 242, 246, 5, 5, 239, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 246, 5, 5, 246, 246, 5, 5, 209, 218, 5, 5, 211, 218, 192, 0, 0, 0, 5, 5, 245, 246, 5, 5, 212, 218, 5, 5, 210, 218, 5, 5, 247, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 219, 5, 5, 197, 247, 5, 5, 203, 247, 5, 5, 31, 219, 5, 5, 241, 246, 5, 5, 202, 247, 192, 0, 0, 0, 5, 5, 209, 247, 5, 5, 200, 247, 5, 5, 207, 247, 5, 5, 206, 247, 5, 5, 237, 246, 5, 5, 201, 247, 5, 5, 115, 244, 5, 5, 210, 247, 5, 5, 199, 247, 5, 5, 204, 247, 192, 0, 0, 0, 5, 5, 208, 247, 5, 5, 198, 247, 192, 0, 0, 0, 5, 5, 124, 248, 5, 5, 30, 219, 5, 5, 205, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 249, 5, 5, 125, 248, 5, 5, 123, 248, 5, 5, 121, 248, 5, 5, 122, 248, 5, 5, 126, 248, 192, 0, 0, 0, 5, 5, 159, 219, 5, 5, 4, 249, 5, 5, 2, 249, 5, 5, 7, 249, 5, 5, 5, 249, 192, 0, 0, 0, 5, 5, 8, 249, 5, 5, 160, 219, 5, 5, 3, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 249, 5, 5, 180, 249, 5, 5, 181, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 250, 5, 5, 11, 250, 5, 5, 177, 220, 5, 5, 255, 202, 192, 0, 0, 0, 5, 5, 122, 204, 192, 0, 0, 0, 5, 5, 137, 226, 5, 5, 136, 226, 5, 5, 45, 206, 5, 5, 10, 208, 5, 5, 231, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 228, 192, 0, 0, 0, 5, 5, 204, 209, 5, 5, 117, 211, 192, 0, 0, 0, 5, 5, 116, 211, 5, 5, 118, 211, 192, 0, 0, 0, 5, 5, 136, 236, 5, 5, 204, 238, 5, 5, 32, 241, 5, 5, 34, 241, 5, 5, 33, 241, 192, 0, 0, 0, 5, 5, 247, 216, 5, 5, 248, 242, 5, 5, 119, 244, 5, 5, 241, 249, 5, 5, 129, 200, 192, 0, 0, 0, 5, 5, 228, 222, 5, 5, 229, 222, 5, 5, 227, 222, 5, 5, 226, 222, 5, 5, 2, 203, 192, 0, 0, 0, 5, 5, 159, 224, 5, 5, 157, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 224, 192, 0, 0, 0, 5, 5, 123, 204, 5, 5, 125, 204, 5, 5, 124, 204, 5, 5, 158, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 226, 5, 5, 146, 226, 5, 5, 144, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 206, 5, 5, 147, 226, 5, 5, 51, 206, 5, 5, 141, 226, 5, 5, 156, 226, 192, 0, 0, 0, 5, 5, 158, 226, 5, 5, 142, 226, 192, 0, 0, 0, 5, 5, 48, 206, 5, 5, 159, 226, 5, 5, 154, 226, 5, 5, 140, 226, 5, 5, 145, 226, 5, 5, 155, 226, 5, 5, 151, 226, 5, 5, 148, 226, 192, 0, 0, 0, 5, 5, 53, 206, 5, 5, 157, 226, 5, 5, 143, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 226, 5, 5, 138, 226, 5, 5, 52, 206, 5, 5, 49, 206, 5, 5, 150, 226, 192, 0, 0, 0, 5, 5, 153, 226, 5, 5, 139, 226, 5, 5, 50, 206, 5, 5, 47, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 208, 5, 5, 238, 228, 5, 5, 19, 208, 192, 0, 0, 0, 5, 5, 241, 228, 5, 5, 244, 228, 5, 5, 16, 208, 5, 5, 14, 208, 5, 5, 233, 228, 5, 5, 242, 228, 5, 5, 240, 228, 5, 5, 237, 228, 5, 5, 245, 228, 5, 5, 246, 228, 5, 5, 248, 228, 5, 5, 249, 228, 5, 5, 232, 228, 5, 5, 13, 208, 5, 5, 235, 228, 5, 5, 234, 228, 5, 5, 247, 228, 5, 5, 15, 208, 5, 5, 236, 228, 5, 5, 17, 208, 5, 5, 12, 208, 5, 5, 239, 228, 5, 5, 21, 208, 192, 0, 0, 0, 5, 5, 18, 208, 5, 5, 243, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 209, 5, 5, 130, 231, 192, 0, 0, 0, 5, 5, 119, 231, 5, 5, 208, 209, 192, 0, 0, 0, 5, 5, 35, 234, 5, 5, 128, 231, 5, 5, 116, 231, 5, 5, 206, 209, 5, 5, 121, 231, 5, 5, 209, 209, 5, 5, 125, 231, 5, 5, 123, 231, 5, 5, 212, 209, 5, 5, 205, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 231, 5, 5, 120, 231, 5, 5, 210, 209, 192, 0, 0, 0, 5, 5, 118, 231, 192, 0, 0, 0, 5, 5, 129, 231, 5, 5, 127, 231, 5, 5, 122, 231, 5, 5, 124, 231, 5, 5, 126, 231, 5, 5, 207, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 234, 5, 5, 42, 234, 5, 5, 32, 234, 5, 5, 38, 234, 5, 5, 119, 211, 5, 5, 34, 234, 5, 5, 125, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 211, 192, 0, 0, 0, 5, 5, 123, 211, 5, 5, 41, 234, 5, 5, 126, 211, 5, 5, 127, 211, 5, 5, 31, 234, 5, 5, 44, 234, 5, 5, 128, 211, 5, 5, 122, 211, 5, 5, 121, 211, 5, 5, 40, 234, 5, 5, 129, 211, 5, 5, 30, 234, 5, 5, 33, 234, 5, 5, 43, 234, 5, 5, 39, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 236, 5, 5, 157, 236, 5, 5, 120, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 213, 5, 5, 145, 236, 5, 5, 166, 236, 5, 5, 146, 236, 5, 5, 31, 213, 192, 0, 0, 0, 5, 5, 143, 236, 192, 0, 0, 0, 5, 5, 152, 236, 5, 5, 144, 236, 5, 5, 33, 213, 5, 5, 138, 236, 5, 5, 165, 236, 5, 5, 34, 213, 5, 5, 162, 236, 5, 5, 163, 236, 5, 5, 139, 236, 5, 5, 39, 213, 5, 5, 154, 236, 192, 0, 0, 0, 5, 5, 148, 236, 5, 5, 155, 236, 5, 5, 142, 236, 192, 0, 0, 0, 5, 5, 167, 236, 5, 5, 159, 236, 5, 5, 153, 236, 5, 5, 137, 236, 5, 5, 35, 213, 5, 5, 160, 236, 192, 0, 0, 0, 5, 5, 38, 213, 5, 5, 164, 236, 192, 0, 0, 0, 5, 5, 158, 236, 5, 5, 32, 213, 5, 5, 156, 236, 192, 0, 0, 0, 5, 5, 150, 236, 5, 5, 30, 213, 5, 5, 141, 236, 5, 5, 149, 236, 5, 5, 161, 236, 5, 5, 147, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 214, 5, 5, 37, 234, 5, 5, 222, 238, 5, 5, 231, 238, 5, 5, 221, 238, 5, 5, 217, 238, 5, 5, 214, 238, 5, 5, 150, 214, 5, 5, 212, 238, 5, 5, 37, 213, 5, 5, 205, 238, 5, 5, 148, 214, 5, 5, 211, 238, 5, 5, 147, 214, 5, 5, 216, 238, 5, 5, 213, 238, 5, 5, 229, 238, 5, 5, 224, 238, 5, 5, 218, 238, 5, 5, 223, 238, 5, 5, 143, 214, 5, 5, 215, 238, 5, 5, 234, 238, 5, 5, 206, 238, 5, 5, 207, 238, 5, 5, 230, 238, 5, 5, 144, 214, 5, 5, 235, 238, 5, 5, 146, 214, 5, 5, 236, 238, 5, 5, 220, 238, 5, 5, 140, 236, 5, 5, 226, 238, 5, 5, 219, 238, 5, 5, 228, 238, 5, 5, 225, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 238, 5, 5, 141, 214, 5, 5, 233, 238, 5, 5, 142, 214, 5, 5, 208, 238, 5, 5, 145, 214, 5, 5, 44, 241, 5, 5, 227, 238, 5, 5, 232, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 241, 5, 5, 140, 214, 5, 5, 208, 215, 5, 5, 51, 241, 5, 5, 47, 241, 192, 0, 0, 0, 5, 5, 45, 241, 5, 5, 40, 241, 5, 5, 55, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 215, 192, 0, 0, 0, 5, 5, 36, 241, 5, 5, 48, 241, 5, 5, 49, 241, 5, 5, 39, 241, 5, 5, 38, 241, 5, 5, 52, 241, 192, 0, 0, 0, 5, 5, 42, 241, 5, 5, 37, 241, 5, 5, 43, 241, 192, 0, 0, 0, 5, 5, 54, 241, 5, 5, 35, 241, 5, 5, 53, 241, 5, 5, 50, 241, 5, 5, 210, 215, 5, 5, 209, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 215, 5, 5, 46, 241, 5, 5, 136, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 242, 5, 5, 253, 216, 5, 5, 254, 242, 5, 5, 251, 242, 5, 5, 4, 243, 5, 5, 10, 243, 5, 5, 253, 242, 192, 0, 0, 0, 5, 5, 19, 243, 5, 5, 250, 216, 5, 5, 13, 243, 5, 5, 2, 243, 5, 5, 14, 243, 5, 5, 9, 243, 5, 5, 16, 243, 5, 5, 255, 242, 5, 5, 255, 216, 5, 5, 254, 216, 5, 5, 3, 243, 5, 5, 17, 243, 5, 5, 249, 242, 5, 5, 15, 243, 5, 5, 248, 216, 192, 0, 0, 0, 5, 5, 7, 243, 5, 5, 6, 243, 5, 5, 11, 243, 5, 5, 252, 242, 5, 5, 252, 216, 192, 0, 0, 0, 5, 5, 2, 217, 5, 5, 5, 243, 5, 5, 12, 243, 5, 5, 3, 217, 5, 5, 8, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 216, 5, 5, 251, 216, 5, 5, 132, 244, 5, 5, 130, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 244, 5, 5, 134, 244, 5, 5, 138, 244, 192, 0, 0, 0, 5, 5, 123, 244, 5, 5, 131, 244, 5, 5, 142, 244, 5, 5, 18, 243, 5, 5, 127, 244, 5, 5, 186, 217, 5, 5, 209, 238, 5, 5, 122, 244, 5, 5, 135, 244, 5, 5, 129, 244, 5, 5, 126, 244, 5, 5, 121, 244, 5, 5, 120, 244, 5, 5, 141, 244, 192, 0, 0, 0, 5, 5, 125, 244, 5, 5, 124, 244, 5, 5, 184, 217, 5, 5, 133, 244, 192, 0, 0, 0, 5, 5, 183, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 217, 5, 5, 128, 244, 5, 5, 140, 244, 192, 0, 0, 0, 5, 5, 215, 245, 5, 5, 216, 245, 192, 0, 0, 0, 5, 5, 83, 218, 5, 5, 213, 245, 5, 5, 80, 218, 5, 5, 221, 245, 192, 0, 0, 0, 5, 5, 84, 218, 5, 5, 223, 245, 5, 5, 212, 245, 5, 5, 139, 244, 5, 5, 225, 245, 5, 5, 214, 245, 192, 0, 0, 0, 5, 5, 81, 218, 5, 5, 220, 245, 192, 0, 0, 0, 5, 5, 222, 245, 5, 5, 217, 245, 5, 5, 224, 245, 5, 5, 219, 245, 5, 5, 218, 245, 5, 5, 82, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 246, 5, 5, 251, 246, 192, 0, 0, 0, 5, 5, 253, 246, 5, 5, 213, 218, 5, 5, 214, 218, 5, 5, 254, 246, 5, 5, 252, 246, 192, 0, 0, 0, 5, 5, 249, 246, 192, 0, 0, 0, 5, 5, 213, 247, 5, 5, 216, 247, 5, 5, 212, 247, 192, 0, 0, 0, 5, 5, 35, 219, 5, 5, 214, 247, 5, 5, 34, 219, 5, 5, 33, 219, 5, 5, 32, 219, 5, 5, 215, 247, 5, 5, 131, 248, 5, 5, 129, 248, 192, 0, 0, 0, 5, 5, 128, 248, 5, 5, 211, 247, 5, 5, 130, 248, 5, 5, 217, 247, 5, 5, 127, 248, 192, 0, 0, 0, 5, 5, 11, 249, 5, 5, 231, 248, 5, 5, 9, 249, 5, 5, 161, 219, 5, 5, 10, 249, 5, 5, 12, 249, 192, 0, 0, 0, 5, 5, 110, 249, 5, 5, 192, 219, 5, 5, 109, 249, 5, 5, 108, 249, 5, 5, 193, 219, 192, 0, 0, 0, 5, 5, 223, 219, 5, 5, 220, 249, 5, 5, 243, 249, 5, 5, 242, 249, 5, 5, 244, 249, 5, 5, 130, 200, 5, 5, 160, 224, 192, 0, 0, 0, 5, 5, 160, 226, 5, 5, 161, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 231, 192, 0, 0, 0, 5, 5, 218, 247, 5, 5, 111, 249, 5, 5, 131, 200, 5, 5, 126, 204, 5, 5, 161, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 228, 5, 5, 22, 208, 192, 0, 0, 0, 5, 5, 133, 231, 5, 5, 132, 231, 5, 5, 213, 209, 192, 0, 0, 0, 5, 5, 130, 211, 5, 5, 237, 238, 5, 5, 151, 214, 192, 0, 0, 0, 5, 5, 152, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 215, 5, 5, 194, 219, 5, 5, 132, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 224, 5, 5, 4, 203, 5, 5, 164, 224, 5, 5, 163, 224, 5, 5, 127, 204, 192, 0, 0, 0, 5, 5, 162, 226, 192, 0, 0, 0, 5, 5, 169, 226, 5, 5, 54, 206, 5, 5, 167, 226, 5, 5, 165, 226, 192, 0, 0, 0, 5, 5, 172, 226, 5, 5, 163, 226, 5, 5, 164, 226, 5, 5, 55, 206, 192, 0, 0, 0, 5, 5, 59, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 226, 5, 5, 58, 206, 5, 5, 171, 226, 5, 5, 168, 226, 5, 5, 166, 226, 5, 5, 56, 206, 5, 5, 57, 206, 5, 5, 170, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 208, 5, 5, 251, 228, 192, 0, 0, 0, 5, 5, 29, 208, 5, 5, 13, 229, 5, 5, 28, 208, 5, 5, 15, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 229, 5, 5, 26, 208, 5, 5, 14, 229, 192, 0, 0, 0, 5, 5, 252, 228, 5, 5, 27, 208, 5, 5, 10, 229, 5, 5, 6, 229, 5, 5, 8, 229, 5, 5, 2, 229, 5, 5, 9, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 208, 5, 5, 5, 229, 192, 0, 0, 0, 5, 5, 4, 229, 5, 5, 254, 228, 192, 0, 0, 0, 5, 5, 11, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 229, 5, 5, 253, 228, 192, 0, 0, 0, 5, 5, 255, 228, 5, 5, 25, 208, 5, 5, 12, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 209, 5, 5, 144, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 231, 5, 5, 142, 231, 5, 5, 137, 231, 5, 5, 136, 231, 5, 5, 134, 231, 192, 0, 0, 0, 5, 5, 141, 231, 5, 5, 143, 231, 5, 5, 139, 231, 192, 0, 0, 0, 5, 5, 138, 231, 5, 5, 214, 209, 5, 5, 215, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 231, 5, 5, 138, 211, 5, 5, 46, 234, 5, 5, 52, 234, 5, 5, 47, 234, 5, 5, 48, 234, 192, 0, 0, 0, 5, 5, 53, 234, 192, 0, 0, 0, 5, 5, 140, 211, 192, 0, 0, 0, 5, 5, 132, 211, 5, 5, 139, 211, 5, 5, 45, 234, 5, 5, 135, 231, 5, 5, 135, 211, 5, 5, 133, 211, 5, 5, 51, 234, 5, 5, 50, 234, 5, 5, 134, 211, 5, 5, 136, 211, 5, 5, 49, 234, 5, 5, 131, 211, 192, 0, 0, 0, 5, 5, 137, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 236, 5, 5, 46, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 236, 5, 5, 180, 236, 192, 0, 0, 0, 5, 5, 175, 236, 5, 5, 48, 213, 5, 5, 179, 236, 5, 5, 171, 236, 5, 5, 172, 236, 5, 5, 40, 213, 5, 5, 42, 213, 192, 0, 0, 0, 5, 5, 177, 236, 5, 5, 169, 236, 5, 5, 44, 213, 5, 5, 43, 213, 5, 5, 173, 236, 5, 5, 178, 236, 5, 5, 176, 236, 5, 5, 45, 213, 5, 5, 174, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 231, 5, 5, 41, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 238, 5, 5, 245, 238, 5, 5, 154, 214, 192, 0, 0, 0, 5, 5, 249, 238, 5, 5, 158, 214, 5, 5, 241, 238, 5, 5, 239, 238, 192, 0, 0, 0, 5, 5, 248, 238, 192, 0, 0, 0, 5, 5, 153, 214, 5, 5, 247, 238, 5, 5, 155, 214, 5, 5, 156, 214, 5, 5, 240, 238, 5, 5, 157, 214, 5, 5, 246, 238, 5, 5, 242, 238, 5, 5, 243, 238, 5, 5, 244, 238, 5, 5, 47, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 241, 5, 5, 68, 241, 192, 0, 0, 0, 5, 5, 218, 215, 5, 5, 63, 241, 5, 5, 65, 241, 192, 0, 0, 0, 5, 5, 216, 215, 5, 5, 57, 241, 5, 5, 61, 241, 192, 0, 0, 0, 5, 5, 64, 241, 5, 5, 214, 215, 5, 5, 217, 215, 5, 5, 67, 241, 5, 5, 59, 241, 5, 5, 60, 241, 5, 5, 66, 241, 5, 5, 58, 241, 5, 5, 62, 241, 5, 5, 215, 215, 5, 5, 21, 243, 192, 0, 0, 0, 5, 5, 20, 243, 5, 5, 5, 217, 5, 5, 26, 243, 5, 5, 7, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 217, 5, 5, 22, 243, 5, 5, 8, 217, 5, 5, 23, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 243, 5, 5, 27, 243, 192, 0, 0, 0, 5, 5, 6, 217, 192, 0, 0, 0, 5, 5, 147, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 244, 192, 0, 0, 0, 5, 5, 144, 244, 5, 5, 146, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 244, 5, 5, 148, 244, 5, 5, 149, 244, 5, 5, 25, 243, 5, 5, 143, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 218, 5, 5, 229, 245, 5, 5, 232, 245, 5, 5, 234, 245, 5, 5, 227, 245, 5, 5, 228, 245, 5, 5, 231, 245, 5, 5, 233, 245, 5, 5, 88, 218, 5, 5, 86, 218, 5, 5, 85, 218, 5, 5, 230, 245, 5, 5, 226, 245, 5, 5, 255, 246, 5, 5, 215, 218, 192, 0, 0, 0, 5, 5, 2, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 247, 5, 5, 36, 219, 5, 5, 222, 247, 5, 5, 37, 219, 5, 5, 219, 247, 5, 5, 221, 247, 5, 5, 110, 219, 192, 0, 0, 0, 5, 5, 132, 248, 5, 5, 109, 219, 5, 5, 15, 249, 5, 5, 14, 249, 192, 0, 0, 0, 5, 5, 13, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 249, 5, 5, 183, 249, 5, 5, 185, 249, 5, 5, 184, 249, 192, 0, 0, 0, 5, 5, 178, 220, 5, 5, 133, 200, 192, 0, 0, 0, 5, 5, 128, 204, 5, 5, 16, 229, 5, 5, 217, 209, 192, 0, 0, 0, 5, 5, 54, 234, 5, 5, 187, 217, 192, 0, 0, 0, 5, 5, 235, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 208, 192, 0, 0, 0, 5, 5, 147, 231, 5, 5, 218, 209, 5, 5, 149, 231, 5, 5, 148, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 234, 5, 5, 141, 211, 5, 5, 182, 236, 5, 5, 185, 236, 5, 5, 184, 236, 192, 0, 0, 0, 5, 5, 183, 236, 5, 5, 250, 238, 5, 5, 252, 238, 5, 5, 251, 238, 192, 0, 0, 0, 5, 5, 220, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 215, 192, 0, 0, 0, 5, 5, 9, 217, 5, 5, 28, 243, 5, 5, 30, 243, 5, 5, 29, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 245, 5, 5, 236, 245, 192, 0, 0, 0, 5, 5, 3, 247, 5, 5, 216, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 219, 5, 5, 134, 248, 5, 5, 133, 248, 5, 5, 224, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 201, 5, 5, 165, 224, 5, 5, 129, 204, 5, 5, 19, 229, 5, 5, 17, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 229, 5, 5, 151, 231, 5, 5, 152, 231, 5, 5, 62, 234, 5, 5, 150, 231, 192, 0, 0, 0, 5, 5, 56, 234, 5, 5, 60, 234, 5, 5, 59, 234, 5, 5, 61, 234, 5, 5, 142, 211, 5, 5, 58, 234, 5, 5, 57, 234, 5, 5, 63, 234, 192, 0, 0, 0, 5, 5, 188, 236, 5, 5, 186, 236, 192, 0, 0, 0, 5, 5, 187, 236, 5, 5, 255, 238, 5, 5, 253, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 238, 5, 5, 69, 241, 5, 5, 31, 243, 5, 5, 32, 243, 5, 5, 189, 217, 192, 0, 0, 0, 5, 5, 238, 245, 5, 5, 4, 247, 5, 5, 217, 218, 192, 0, 0, 0, 5, 5, 223, 247, 5, 5, 135, 248, 5, 5, 111, 219, 192, 0, 0, 0, 5, 5, 16, 249, 5, 5, 186, 249, 5, 5, 126, 201, 192, 0, 0, 0, 5, 5, 131, 204, 5, 5, 132, 204, 5, 5, 166, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 224, 5, 5, 130, 204, 192, 0, 0, 0, 5, 5, 65, 206, 192, 0, 0, 0, 5, 5, 63, 206, 192, 0, 0, 0, 5, 5, 62, 206, 5, 5, 69, 206, 5, 5, 61, 206, 5, 5, 70, 206, 5, 5, 174, 226, 5, 5, 67, 206, 192, 0, 0, 0, 5, 5, 64, 206, 5, 5, 68, 206, 5, 5, 66, 206, 5, 5, 60, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 208, 192, 0, 0, 0, 5, 5, 33, 208, 5, 5, 23, 229, 5, 5, 38, 208, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 208, 5, 5, 34, 208, 192, 0, 0, 0, 5, 5, 35, 208, 192, 0, 0, 0, 5, 5, 21, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 208, 192, 0, 0, 0, 5, 5, 22, 229, 5, 5, 37, 208, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 229, 5, 5, 36, 208, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 209, 192, 0, 0, 0, 5, 5, 231, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 231, 5, 5, 230, 209, 5, 5, 219, 209, 5, 5, 223, 209, 192, 0, 0, 0, 5, 5, 161, 211, 5, 5, 68, 234, 5, 5, 157, 231, 5, 5, 224, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 231, 5, 5, 161, 231, 5, 5, 228, 209, 192, 0, 0, 0, 5, 5, 163, 231, 192, 0, 0, 0, 5, 5, 165, 231, 192, 0, 0, 0, 5, 5, 166, 231, 5, 5, 154, 231, 5, 5, 153, 231, 5, 5, 167, 231, 5, 5, 227, 209, 5, 5, 164, 231, 5, 5, 162, 231, 192, 0, 0, 0, 5, 5, 225, 209, 5, 5, 221, 209, 5, 5, 232, 209, 5, 5, 158, 231, 5, 5, 159, 231, 5, 5, 156, 231, 192, 0, 0, 0, 5, 5, 226, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 209, 192, 0, 0, 0, 5, 5, 220, 209, 5, 5, 67, 234, 5, 5, 156, 211, 5, 5, 151, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 211, 192, 0, 0, 0, 5, 5, 162, 211, 5, 5, 147, 211, 192, 0, 0, 0, 5, 5, 143, 211, 5, 5, 158, 211, 5, 5, 155, 211, 5, 5, 157, 211, 192, 0, 0, 0, 5, 5, 148, 211, 5, 5, 153, 211, 5, 5, 144, 211, 5, 5, 145, 211, 5, 5, 75, 234, 5, 5, 72, 234, 5, 5, 64, 234, 5, 5, 69, 234, 192, 0, 0, 0, 5, 5, 159, 211, 5, 5, 76, 234, 5, 5, 160, 211, 5, 5, 150, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 234, 192, 0, 0, 0, 5, 5, 74, 234, 5, 5, 70, 234, 5, 5, 73, 234, 5, 5, 71, 234, 5, 5, 154, 211, 5, 5, 65, 234, 5, 5, 149, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 236, 5, 5, 50, 213, 5, 5, 53, 213, 192, 0, 0, 0, 5, 5, 193, 236, 192, 0, 0, 0, 5, 5, 61, 213, 5, 5, 192, 236, 5, 5, 55, 213, 192, 0, 0, 0, 5, 5, 163, 214, 5, 5, 194, 236, 192, 0, 0, 0, 5, 5, 60, 213, 5, 5, 190, 236, 5, 5, 62, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 213, 192, 0, 0, 0, 5, 5, 152, 211, 5, 5, 54, 213, 192, 0, 0, 0, 5, 5, 52, 213, 5, 5, 56, 213, 5, 5, 58, 213, 5, 5, 49, 213, 5, 5, 63, 213, 5, 5, 59, 213, 192, 0, 0, 0, 5, 5, 57, 213, 5, 5, 189, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 214, 192, 0, 0, 0, 5, 5, 166, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 214, 192, 0, 0, 0, 5, 5, 4, 239, 5, 5, 174, 214, 5, 5, 16, 239, 5, 5, 9, 239, 5, 5, 159, 214, 5, 5, 17, 239, 5, 5, 11, 239, 5, 5, 169, 214, 5, 5, 12, 239, 192, 0, 0, 0, 5, 5, 168, 214, 5, 5, 15, 239, 5, 5, 162, 214, 5, 5, 13, 239, 5, 5, 3, 239, 5, 5, 161, 214, 5, 5, 80, 241, 5, 5, 167, 214, 192, 0, 0, 0, 5, 5, 164, 214, 192, 0, 0, 0, 5, 5, 172, 214, 192, 0, 0, 0, 5, 5, 2, 239, 192, 0, 0, 0, 5, 5, 6, 239, 5, 5, 160, 214, 5, 5, 5, 239, 5, 5, 7, 239, 5, 5, 8, 239, 5, 5, 171, 214, 5, 5, 10, 239, 5, 5, 14, 239, 5, 5, 18, 239, 192, 0, 0, 0, 5, 5, 175, 214, 5, 5, 226, 215, 5, 5, 75, 241, 5, 5, 81, 241, 5, 5, 78, 241, 5, 5, 70, 241, 5, 5, 82, 241, 5, 5, 71, 241, 192, 0, 0, 0, 5, 5, 77, 241, 192, 0, 0, 0, 5, 5, 221, 215, 5, 5, 227, 215, 5, 5, 83, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 215, 192, 0, 0, 0, 5, 5, 233, 215, 5, 5, 228, 215, 5, 5, 85, 241, 5, 5, 79, 241, 5, 5, 224, 215, 5, 5, 72, 241, 5, 5, 234, 215, 5, 5, 73, 241, 5, 5, 74, 241, 5, 5, 235, 215, 5, 5, 232, 215, 5, 5, 165, 214, 192, 0, 0, 0, 5, 5, 222, 215, 5, 5, 86, 241, 5, 5, 236, 215, 192, 0, 0, 0, 5, 5, 229, 215, 5, 5, 84, 241, 5, 5, 225, 215, 5, 5, 230, 215, 5, 5, 231, 215, 192, 0, 0, 0, 5, 5, 17, 217, 5, 5, 37, 243, 5, 5, 46, 243, 5, 5, 43, 243, 5, 5, 45, 243, 192, 0, 0, 0, 5, 5, 14, 217, 5, 5, 38, 243, 192, 0, 0, 0, 5, 5, 44, 243, 5, 5, 10, 217, 5, 5, 40, 243, 5, 5, 18, 217, 5, 5, 36, 243, 5, 5, 41, 243, 5, 5, 48, 243, 5, 5, 76, 241, 5, 5, 42, 243, 5, 5, 35, 243, 5, 5, 11, 217, 5, 5, 34, 243, 5, 5, 12, 217, 5, 5, 49, 243, 5, 5, 13, 217, 5, 5, 47, 243, 5, 5, 16, 217, 5, 5, 33, 243, 192, 0, 0, 0, 5, 5, 15, 217, 192, 0, 0, 0, 5, 5, 39, 243, 5, 5, 153, 244, 5, 5, 167, 244, 5, 5, 162, 244, 5, 5, 164, 244, 5, 5, 152, 244, 5, 5, 190, 217, 192, 0, 0, 0, 5, 5, 151, 244, 5, 5, 193, 217, 5, 5, 192, 217, 192, 0, 0, 0, 5, 5, 166, 244, 5, 5, 158, 244, 5, 5, 155, 244, 5, 5, 161, 244, 192, 0, 0, 0, 5, 5, 154, 244, 192, 0, 0, 0, 5, 5, 156, 244, 5, 5, 165, 244, 5, 5, 163, 244, 192, 0, 0, 0, 5, 5, 191, 217, 5, 5, 170, 244, 5, 5, 168, 244, 5, 5, 159, 244, 5, 5, 169, 244, 5, 5, 160, 244, 192, 0, 0, 0, 5, 5, 242, 245, 5, 5, 89, 218, 5, 5, 249, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 247, 5, 5, 96, 218, 5, 5, 157, 244, 5, 5, 240, 245, 5, 5, 92, 218, 5, 5, 241, 245, 5, 5, 246, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 218, 5, 5, 95, 218, 5, 5, 239, 245, 5, 5, 248, 245, 5, 5, 250, 245, 5, 5, 243, 245, 5, 5, 245, 245, 5, 5, 247, 245, 5, 5, 244, 245, 5, 5, 251, 245, 5, 5, 91, 218, 5, 5, 97, 218, 5, 5, 93, 218, 192, 0, 0, 0, 5, 5, 90, 218, 5, 5, 7, 247, 192, 0, 0, 0, 5, 5, 222, 218, 5, 5, 5, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 247, 192, 0, 0, 0, 5, 5, 10, 247, 5, 5, 220, 218, 5, 5, 11, 247, 5, 5, 8, 247, 192, 0, 0, 0, 5, 5, 6, 247, 5, 5, 223, 218, 5, 5, 219, 218, 5, 5, 12, 247, 192, 0, 0, 0, 5, 5, 221, 218, 5, 5, 218, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 219, 5, 5, 225, 247, 5, 5, 224, 247, 5, 5, 227, 247, 5, 5, 228, 247, 192, 0, 0, 0, 5, 5, 41, 219, 5, 5, 136, 248, 5, 5, 141, 248, 5, 5, 112, 219, 192, 0, 0, 0, 5, 5, 138, 248, 192, 0, 0, 0, 5, 5, 137, 248, 5, 5, 140, 248, 5, 5, 139, 248, 192, 0, 0, 0, 5, 5, 20, 249, 192, 0, 0, 0, 5, 5, 162, 219, 5, 5, 19, 249, 5, 5, 17, 249, 192, 0, 0, 0, 5, 5, 18, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 219, 5, 5, 195, 219, 5, 5, 112, 249, 5, 5, 113, 249, 5, 5, 197, 219, 192, 0, 0, 0, 5, 5, 187, 249, 5, 5, 188, 249, 5, 5, 236, 219, 192, 0, 0, 0, 5, 5, 242, 219, 192, 0, 0, 0, 5, 5, 245, 249, 5, 5, 22, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 201, 192, 0, 0, 0, 5, 5, 24, 229, 192, 0, 0, 0, 5, 5, 25, 229, 5, 5, 77, 234, 5, 5, 195, 236, 5, 5, 19, 239, 5, 5, 20, 217, 192, 0, 0, 0, 5, 5, 19, 217, 5, 5, 171, 244, 5, 5, 252, 245, 192, 0, 0, 0, 5, 5, 21, 249, 5, 5, 128, 201, 5, 5, 175, 226, 5, 5, 71, 206, 5, 5, 41, 208, 5, 5, 79, 234, 5, 5, 78, 234, 5, 5, 176, 214, 5, 5, 20, 239, 5, 5, 177, 214, 5, 5, 50, 243, 5, 5, 194, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 219, 5, 5, 129, 201, 5, 5, 230, 222, 5, 5, 176, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 208, 192, 0, 0, 0, 5, 5, 26, 229, 5, 5, 27, 229, 192, 0, 0, 0, 5, 5, 168, 231, 192, 0, 0, 0, 5, 5, 233, 209, 5, 5, 163, 211, 192, 0, 0, 0, 5, 5, 81, 234, 5, 5, 80, 234, 5, 5, 82, 234, 192, 0, 0, 0, 5, 5, 196, 236, 5, 5, 197, 236, 5, 5, 64, 213, 5, 5, 237, 215, 5, 5, 178, 214, 5, 5, 238, 215, 192, 0, 0, 0, 5, 5, 54, 243, 5, 5, 51, 243, 5, 5, 53, 243, 5, 5, 52, 243, 5, 5, 21, 217, 192, 0, 0, 0, 5, 5, 172, 244, 5, 5, 254, 245, 5, 5, 253, 245, 5, 5, 131, 221, 5, 5, 73, 206, 5, 5, 72, 206, 5, 5, 177, 226, 192, 0, 0, 0, 5, 5, 28, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 231, 5, 5, 169, 231, 5, 5, 234, 209, 192, 0, 0, 0, 5, 5, 84, 234, 5, 5, 85, 234, 5, 5, 83, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 211, 5, 5, 164, 211, 192, 0, 0, 0, 5, 5, 66, 213, 5, 5, 65, 213, 192, 0, 0, 0, 5, 5, 21, 239, 5, 5, 89, 241, 5, 5, 87, 241, 5, 5, 88, 241, 5, 5, 239, 215, 5, 5, 56, 243, 5, 5, 55, 243, 192, 0, 0, 0, 5, 5, 175, 244, 5, 5, 174, 244, 5, 5, 173, 244, 5, 5, 255, 245, 192, 0, 0, 0, 5, 5, 246, 249, 5, 5, 130, 201, 5, 5, 133, 204, 192, 0, 0, 0, 5, 5, 134, 204, 5, 5, 74, 206, 5, 5, 75, 206, 5, 5, 179, 226, 5, 5, 178, 226, 5, 5, 29, 229, 192, 0, 0, 0, 5, 5, 48, 208, 5, 5, 46, 208, 5, 5, 43, 208, 5, 5, 47, 208, 5, 5, 45, 208, 5, 5, 44, 208, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 209, 5, 5, 173, 231, 192, 0, 0, 0, 5, 5, 170, 211, 5, 5, 237, 209, 5, 5, 242, 209, 5, 5, 175, 231, 5, 5, 244, 209, 5, 5, 243, 209, 5, 5, 246, 209, 5, 5, 174, 231, 5, 5, 171, 231, 5, 5, 240, 209, 5, 5, 236, 209, 5, 5, 238, 209, 5, 5, 172, 231, 5, 5, 245, 209, 5, 5, 241, 209, 5, 5, 239, 209, 5, 5, 172, 211, 5, 5, 171, 211, 5, 5, 169, 211, 5, 5, 173, 211, 192, 0, 0, 0, 5, 5, 167, 211, 5, 5, 168, 211, 192, 0, 0, 0, 5, 5, 166, 211, 192, 0, 0, 0, 5, 5, 86, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 236, 192, 0, 0, 0, 5, 5, 68, 213, 5, 5, 69, 213, 5, 5, 67, 213, 192, 0, 0, 0, 5, 5, 198, 236, 192, 0, 0, 0, 5, 5, 200, 236, 192, 0, 0, 0, 5, 5, 24, 239, 5, 5, 26, 239, 192, 0, 0, 0, 5, 5, 187, 214, 5, 5, 27, 239, 5, 5, 180, 214, 5, 5, 23, 239, 5, 5, 179, 214, 5, 5, 189, 214, 5, 5, 185, 214, 5, 5, 186, 214, 5, 5, 182, 214, 5, 5, 22, 239, 5, 5, 181, 214, 5, 5, 28, 239, 5, 5, 25, 239, 192, 0, 0, 0, 5, 5, 188, 214, 192, 0, 0, 0, 5, 5, 183, 214, 5, 5, 184, 214, 5, 5, 91, 241, 192, 0, 0, 0, 5, 5, 93, 241, 5, 5, 92, 241, 192, 0, 0, 0, 5, 5, 94, 241, 5, 5, 240, 215, 5, 5, 90, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 217, 5, 5, 57, 243, 5, 5, 22, 217, 5, 5, 26, 217, 5, 5, 24, 217, 5, 5, 23, 217, 5, 5, 176, 244, 192, 0, 0, 0, 5, 5, 179, 244, 192, 0, 0, 0, 5, 5, 178, 244, 192, 0, 0, 0, 5, 5, 177, 244, 5, 5, 195, 217, 5, 5, 2, 246, 5, 5, 3, 246, 5, 5, 98, 218, 5, 5, 4, 246, 5, 5, 99, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 218, 192, 0, 0, 0, 5, 5, 224, 218, 5, 5, 229, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 219, 5, 5, 230, 247, 5, 5, 142, 248, 5, 5, 113, 219, 5, 5, 114, 219, 192, 0, 0, 0, 5, 5, 22, 249, 192, 0, 0, 0, 5, 5, 199, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 201, 192, 0, 0, 0, 5, 5, 50, 208, 5, 5, 49, 208, 5, 5, 87, 234, 5, 5, 88, 234, 192, 0, 0, 0, 5, 5, 70, 213, 5, 5, 95, 241, 5, 5, 190, 214, 5, 5, 96, 241, 5, 5, 58, 243, 5, 5, 132, 201, 192, 0, 0, 0, 5, 5, 168, 224, 5, 5, 136, 204, 5, 5, 135, 204, 192, 0, 0, 0, 5, 5, 180, 226, 5, 5, 76, 206, 5, 5, 181, 226, 5, 5, 32, 229, 192, 0, 0, 0, 5, 5, 31, 229, 192, 0, 0, 0, 5, 5, 30, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 231, 5, 5, 249, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 231, 5, 5, 248, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 231, 5, 5, 247, 209, 192, 0, 0, 0, 5, 5, 90, 234, 5, 5, 93, 234, 5, 5, 91, 234, 5, 5, 92, 234, 5, 5, 96, 234, 5, 5, 89, 234, 5, 5, 97, 234, 5, 5, 94, 234, 5, 5, 95, 234, 5, 5, 72, 213, 5, 5, 201, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 213, 192, 0, 0, 0, 5, 5, 32, 239, 5, 5, 30, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 214, 5, 5, 29, 239, 5, 5, 31, 239, 192, 0, 0, 0, 5, 5, 192, 214, 192, 0, 0, 0, 5, 5, 97, 241, 192, 0, 0, 0, 5, 5, 98, 241, 5, 5, 27, 217, 192, 0, 0, 0, 5, 5, 6, 246, 5, 5, 8, 246, 5, 5, 5, 246, 5, 5, 7, 246, 5, 5, 13, 247, 5, 5, 231, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 249, 5, 5, 133, 201, 5, 5, 137, 204, 5, 5, 182, 226, 5, 5, 184, 226, 5, 5, 183, 226, 192, 0, 0, 0, 5, 5, 35, 229, 5, 5, 52, 208, 192, 0, 0, 0, 5, 5, 33, 229, 192, 0, 0, 0, 5, 5, 51, 208, 5, 5, 36, 229, 192, 0, 0, 0, 5, 5, 37, 229, 5, 5, 34, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 231, 5, 5, 3, 210, 5, 5, 182, 231, 5, 5, 188, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 209, 5, 5, 255, 209, 5, 5, 181, 231, 5, 5, 250, 209, 5, 5, 185, 231, 5, 5, 103, 234, 5, 5, 254, 209, 192, 0, 0, 0, 5, 5, 180, 231, 192, 0, 0, 0, 5, 5, 186, 231, 5, 5, 183, 231, 5, 5, 189, 231, 5, 5, 179, 231, 5, 5, 187, 231, 5, 5, 253, 209, 5, 5, 2, 210, 5, 5, 184, 231, 5, 5, 251, 209, 192, 0, 0, 0, 5, 5, 175, 211, 5, 5, 99, 234, 5, 5, 174, 211, 5, 5, 106, 234, 5, 5, 105, 234, 5, 5, 181, 211, 192, 0, 0, 0, 5, 5, 182, 211, 5, 5, 107, 234, 5, 5, 176, 211, 5, 5, 104, 234, 5, 5, 180, 211, 5, 5, 109, 234, 5, 5, 100, 234, 192, 0, 0, 0, 5, 5, 102, 234, 5, 5, 177, 211, 5, 5, 98, 234, 5, 5, 101, 234, 5, 5, 108, 234, 5, 5, 178, 211, 5, 5, 110, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 211, 192, 0, 0, 0, 5, 5, 73, 213, 5, 5, 206, 236, 5, 5, 212, 236, 5, 5, 204, 236, 5, 5, 213, 236, 192, 0, 0, 0, 5, 5, 203, 236, 5, 5, 208, 236, 5, 5, 214, 236, 5, 5, 211, 236, 5, 5, 210, 236, 5, 5, 209, 236, 192, 0, 0, 0, 5, 5, 202, 236, 5, 5, 207, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 236, 192, 0, 0, 0, 5, 5, 197, 214, 5, 5, 194, 214, 5, 5, 41, 239, 5, 5, 46, 239, 5, 5, 48, 239, 5, 5, 45, 239, 5, 5, 38, 239, 5, 5, 40, 239, 5, 5, 50, 239, 5, 5, 47, 239, 5, 5, 42, 239, 5, 5, 51, 239, 5, 5, 39, 239, 5, 5, 49, 239, 5, 5, 195, 214, 5, 5, 201, 214, 5, 5, 199, 214, 5, 5, 33, 239, 5, 5, 200, 214, 5, 5, 196, 214, 5, 5, 34, 239, 5, 5, 36, 239, 5, 5, 35, 239, 5, 5, 43, 239, 5, 5, 44, 239, 192, 0, 0, 0, 5, 5, 198, 214, 192, 0, 0, 0, 5, 5, 193, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 239, 192, 0, 0, 0, 5, 5, 108, 241, 5, 5, 242, 215, 192, 0, 0, 0, 5, 5, 99, 241, 5, 5, 243, 215, 5, 5, 246, 215, 5, 5, 104, 241, 192, 0, 0, 0, 5, 5, 101, 241, 5, 5, 245, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 241, 5, 5, 106, 241, 5, 5, 100, 241, 5, 5, 109, 241, 5, 5, 102, 241, 5, 5, 107, 241, 5, 5, 244, 215, 192, 0, 0, 0, 5, 5, 241, 215, 5, 5, 103, 241, 192, 0, 0, 0, 5, 5, 64, 243, 5, 5, 30, 217, 5, 5, 28, 217, 5, 5, 31, 217, 5, 5, 29, 217, 5, 5, 63, 243, 5, 5, 60, 243, 5, 5, 59, 243, 192, 0, 0, 0, 5, 5, 62, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 243, 5, 5, 193, 244, 5, 5, 201, 217, 5, 5, 184, 244, 5, 5, 183, 244, 192, 0, 0, 0, 5, 5, 196, 217, 5, 5, 189, 244, 5, 5, 188, 244, 5, 5, 180, 244, 5, 5, 191, 244, 5, 5, 185, 244, 5, 5, 200, 217, 5, 5, 182, 244, 5, 5, 190, 244, 5, 5, 181, 244, 5, 5, 197, 217, 5, 5, 199, 217, 5, 5, 186, 244, 5, 5, 198, 217, 5, 5, 187, 244, 192, 0, 0, 0, 5, 5, 192, 244, 5, 5, 12, 246, 192, 0, 0, 0, 5, 5, 104, 218, 5, 5, 9, 246, 192, 0, 0, 0, 5, 5, 13, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 218, 5, 5, 11, 246, 5, 5, 106, 218, 192, 0, 0, 0, 5, 5, 103, 218, 192, 0, 0, 0, 5, 5, 10, 246, 192, 0, 0, 0, 5, 5, 105, 218, 5, 5, 14, 246, 5, 5, 100, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 218, 5, 5, 229, 218, 192, 0, 0, 0, 5, 5, 16, 247, 5, 5, 228, 218, 5, 5, 14, 247, 5, 5, 102, 218, 5, 5, 15, 247, 5, 5, 226, 218, 5, 5, 43, 219, 5, 5, 45, 219, 5, 5, 233, 247, 5, 5, 44, 219, 5, 5, 232, 247, 192, 0, 0, 0, 5, 5, 147, 248, 5, 5, 115, 219, 5, 5, 146, 248, 5, 5, 116, 219, 5, 5, 144, 248, 5, 5, 143, 248, 5, 5, 148, 248, 5, 5, 149, 248, 5, 5, 23, 249, 192, 0, 0, 0, 5, 5, 145, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 249, 5, 5, 114, 249, 5, 5, 115, 249, 5, 5, 116, 249, 5, 5, 225, 219, 192, 0, 0, 0, 5, 5, 191, 249, 5, 5, 190, 249, 5, 5, 189, 249, 5, 5, 222, 249, 192, 0, 0, 0, 5, 5, 12, 250, 5, 5, 247, 249, 5, 5, 243, 219, 5, 5, 134, 201, 5, 5, 77, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 217, 192, 0, 0, 0, 5, 5, 15, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 249, 5, 5, 135, 201, 5, 5, 5, 203, 5, 5, 139, 204, 5, 5, 138, 204, 192, 0, 0, 0, 5, 5, 80, 206, 192, 0, 0, 0, 5, 5, 185, 226, 5, 5, 78, 206, 5, 5, 186, 226, 5, 5, 79, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 229, 5, 5, 38, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 208, 5, 5, 41, 229, 5, 5, 40, 229, 5, 5, 39, 229, 5, 5, 54, 208, 5, 5, 43, 229, 5, 5, 44, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 231, 5, 5, 195, 231, 5, 5, 199, 231, 5, 5, 200, 231, 5, 5, 206, 231, 192, 0, 0, 0, 5, 5, 202, 231, 5, 5, 204, 231, 192, 0, 0, 0, 5, 5, 196, 231, 5, 5, 191, 231, 192, 0, 0, 0, 5, 5, 203, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 231, 5, 5, 198, 231, 5, 5, 201, 231, 5, 5, 192, 231, 5, 5, 5, 210, 5, 5, 194, 231, 5, 5, 193, 231, 5, 5, 4, 210, 5, 5, 6, 210, 192, 0, 0, 0, 5, 5, 186, 211, 5, 5, 112, 234, 5, 5, 114, 234, 5, 5, 113, 234, 5, 5, 118, 234, 5, 5, 184, 211, 192, 0, 0, 0, 5, 5, 115, 234, 5, 5, 111, 234, 5, 5, 116, 234, 5, 5, 117, 234, 5, 5, 185, 211, 5, 5, 187, 211, 5, 5, 119, 234, 192, 0, 0, 0, 5, 5, 218, 236, 5, 5, 217, 236, 192, 0, 0, 0, 5, 5, 215, 236, 5, 5, 216, 236, 5, 5, 75, 213, 5, 5, 77, 213, 5, 5, 74, 213, 5, 5, 76, 213, 5, 5, 58, 239, 5, 5, 59, 239, 5, 5, 54, 239, 192, 0, 0, 0, 5, 5, 55, 239, 5, 5, 204, 214, 5, 5, 209, 214, 5, 5, 203, 214, 5, 5, 210, 214, 5, 5, 205, 214, 5, 5, 56, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 239, 5, 5, 53, 239, 5, 5, 211, 214, 5, 5, 207, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 214, 5, 5, 208, 214, 192, 0, 0, 0, 5, 5, 52, 239, 192, 0, 0, 0, 5, 5, 112, 241, 5, 5, 248, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 241, 5, 5, 250, 215, 5, 5, 117, 241, 5, 5, 113, 241, 5, 5, 111, 241, 5, 5, 116, 241, 5, 5, 249, 215, 5, 5, 115, 241, 192, 0, 0, 0, 5, 5, 247, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 217, 5, 5, 36, 217, 5, 5, 66, 243, 192, 0, 0, 0, 5, 5, 34, 217, 5, 5, 65, 243, 5, 5, 32, 217, 5, 5, 35, 217, 5, 5, 194, 244, 5, 5, 195, 244, 5, 5, 196, 244, 5, 5, 203, 217, 192, 0, 0, 0, 5, 5, 197, 244, 192, 0, 0, 0, 5, 5, 204, 217, 5, 5, 108, 218, 5, 5, 18, 246, 5, 5, 19, 246, 5, 5, 17, 246, 5, 5, 16, 246, 5, 5, 20, 246, 5, 5, 107, 218, 5, 5, 20, 247, 5, 5, 18, 247, 5, 5, 19, 247, 5, 5, 21, 247, 5, 5, 17, 247, 5, 5, 22, 247, 5, 5, 235, 247, 192, 0, 0, 0, 5, 5, 236, 247, 5, 5, 234, 247, 5, 5, 46, 219, 5, 5, 150, 248, 5, 5, 117, 219, 5, 5, 151, 248, 5, 5, 25, 249, 5, 5, 24, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 201, 5, 5, 7, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 215, 192, 0, 0, 0, 5, 5, 251, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 218, 5, 5, 197, 218, 5, 5, 47, 219, 5, 5, 137, 201, 5, 5, 81, 206, 5, 5, 189, 211, 192, 0, 0, 0, 5, 5, 21, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 201, 192, 0, 0, 0, 5, 5, 141, 201, 5, 5, 140, 201, 5, 5, 139, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 221, 192, 0, 0, 0, 5, 5, 232, 222, 192, 0, 0, 0, 5, 5, 234, 222, 5, 5, 6, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 203, 5, 5, 231, 222, 5, 5, 233, 222, 5, 5, 7, 203, 5, 5, 236, 222, 5, 5, 235, 222, 5, 5, 237, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 224, 5, 5, 170, 224, 5, 5, 142, 204, 5, 5, 169, 224, 5, 5, 147, 204, 5, 5, 144, 204, 5, 5, 141, 204, 192, 0, 0, 0, 5, 5, 148, 204, 192, 0, 0, 0, 5, 5, 143, 204, 5, 5, 146, 204, 192, 0, 0, 0, 5, 5, 145, 204, 5, 5, 171, 224, 192, 0, 0, 0, 5, 5, 140, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 206, 5, 5, 188, 226, 5, 5, 194, 226, 5, 5, 84, 206, 5, 5, 91, 206, 192, 0, 0, 0, 5, 5, 86, 206, 5, 5, 191, 226, 5, 5, 193, 226, 5, 5, 89, 206, 5, 5, 187, 226, 5, 5, 190, 226, 5, 5, 85, 206, 5, 5, 82, 206, 5, 5, 189, 226, 5, 5, 88, 206, 5, 5, 192, 226, 5, 5, 90, 206, 5, 5, 83, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 229, 5, 5, 49, 229, 5, 5, 56, 208, 192, 0, 0, 0, 5, 5, 66, 208, 5, 5, 62, 208, 5, 5, 47, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 208, 5, 5, 63, 208, 5, 5, 68, 208, 5, 5, 58, 208, 192, 0, 0, 0, 5, 5, 55, 208, 5, 5, 57, 208, 5, 5, 69, 208, 5, 5, 48, 229, 5, 5, 61, 208, 5, 5, 64, 208, 5, 5, 60, 208, 5, 5, 65, 208, 5, 5, 50, 229, 5, 5, 67, 208, 5, 5, 59, 208, 5, 5, 45, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 231, 5, 5, 8, 210, 5, 5, 209, 231, 192, 0, 0, 0, 5, 5, 10, 210, 5, 5, 12, 210, 192, 0, 0, 0, 5, 5, 208, 231, 5, 5, 9, 210, 5, 5, 13, 210, 192, 0, 0, 0, 5, 5, 11, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 211, 5, 5, 124, 234, 5, 5, 203, 211, 5, 5, 121, 234, 192, 0, 0, 0, 5, 5, 204, 211, 5, 5, 193, 211, 192, 0, 0, 0, 5, 5, 122, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 211, 192, 0, 0, 0, 5, 5, 123, 234, 5, 5, 191, 211, 5, 5, 190, 211, 192, 0, 0, 0, 5, 5, 201, 211, 5, 5, 200, 211, 5, 5, 199, 211, 5, 5, 197, 211, 5, 5, 202, 211, 5, 5, 120, 234, 5, 5, 192, 211, 5, 5, 194, 211, 5, 5, 196, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 213, 5, 5, 83, 213, 192, 0, 0, 0, 5, 5, 87, 213, 5, 5, 81, 213, 5, 5, 86, 213, 5, 5, 84, 213, 192, 0, 0, 0, 5, 5, 79, 213, 192, 0, 0, 0, 5, 5, 85, 213, 5, 5, 82, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 239, 5, 5, 214, 214, 5, 5, 212, 214, 192, 0, 0, 0, 5, 5, 64, 239, 192, 0, 0, 0, 5, 5, 215, 214, 5, 5, 213, 214, 5, 5, 62, 239, 5, 5, 61, 239, 192, 0, 0, 0, 5, 5, 2, 216, 5, 5, 60, 239, 5, 5, 254, 215, 5, 5, 253, 215, 5, 5, 118, 241, 5, 5, 216, 214, 5, 5, 255, 215, 5, 5, 119, 241, 5, 5, 4, 216, 5, 5, 120, 241, 5, 5, 3, 216, 5, 5, 38, 217, 5, 5, 68, 243, 5, 5, 37, 217, 5, 5, 42, 217, 5, 5, 40, 217, 5, 5, 41, 217, 5, 5, 206, 217, 5, 5, 39, 217, 5, 5, 67, 243, 5, 5, 121, 241, 5, 5, 205, 217, 5, 5, 207, 217, 192, 0, 0, 0, 5, 5, 110, 218, 5, 5, 111, 218, 192, 0, 0, 0, 5, 5, 23, 247, 192, 0, 0, 0, 5, 5, 164, 219, 5, 5, 163, 219, 5, 5, 143, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 220, 5, 5, 92, 206, 192, 0, 0, 0, 5, 5, 180, 220, 5, 5, 181, 220, 5, 5, 179, 220, 192, 0, 0, 0, 5, 5, 182, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 221, 5, 5, 134, 221, 5, 5, 139, 221, 5, 5, 135, 221, 5, 5, 144, 201, 5, 5, 147, 201, 192, 0, 0, 0, 5, 5, 136, 221, 5, 5, 146, 201, 5, 5, 138, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 222, 5, 5, 242, 222, 5, 5, 11, 203, 5, 5, 238, 222, 5, 5, 241, 222, 5, 5, 239, 222, 5, 5, 9, 203, 5, 5, 12, 203, 192, 0, 0, 0, 5, 5, 10, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 224, 5, 5, 178, 224, 5, 5, 175, 224, 192, 0, 0, 0, 5, 5, 151, 204, 192, 0, 0, 0, 5, 5, 152, 204, 192, 0, 0, 0, 5, 5, 177, 224, 192, 0, 0, 0, 5, 5, 179, 224, 5, 5, 181, 224, 192, 0, 0, 0, 5, 5, 149, 204, 5, 5, 180, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 229, 5, 5, 176, 224, 5, 5, 195, 226, 5, 5, 204, 226, 5, 5, 202, 226, 5, 5, 197, 226, 5, 5, 198, 226, 5, 5, 203, 226, 5, 5, 205, 226, 5, 5, 94, 206, 192, 0, 0, 0, 5, 5, 200, 226, 5, 5, 196, 226, 5, 5, 93, 206, 5, 5, 95, 206, 5, 5, 199, 226, 5, 5, 206, 226, 5, 5, 201, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 208, 5, 5, 60, 229, 5, 5, 52, 229, 5, 5, 58, 229, 5, 5, 59, 229, 5, 5, 72, 208, 192, 0, 0, 0, 5, 5, 51, 229, 5, 5, 53, 229, 5, 5, 173, 224, 5, 5, 55, 229, 5, 5, 56, 229, 5, 5, 54, 229, 5, 5, 15, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 231, 5, 5, 122, 241, 5, 5, 217, 231, 5, 5, 214, 231, 5, 5, 73, 208, 5, 5, 17, 210, 5, 5, 213, 231, 5, 5, 219, 231, 5, 5, 218, 231, 5, 5, 14, 210, 5, 5, 222, 231, 5, 5, 212, 231, 5, 5, 221, 231, 5, 5, 210, 231, 5, 5, 220, 231, 5, 5, 215, 231, 5, 5, 16, 210, 192, 0, 0, 0, 5, 5, 131, 234, 192, 0, 0, 0, 5, 5, 126, 234, 5, 5, 132, 234, 5, 5, 127, 234, 5, 5, 125, 234, 5, 5, 128, 234, 5, 5, 205, 211, 192, 0, 0, 0, 5, 5, 130, 234, 192, 0, 0, 0, 5, 5, 129, 234, 5, 5, 206, 211, 5, 5, 89, 213, 5, 5, 88, 213, 5, 5, 225, 236, 5, 5, 228, 236, 5, 5, 220, 236, 5, 5, 224, 236, 5, 5, 90, 213, 5, 5, 223, 236, 5, 5, 221, 236, 5, 5, 227, 236, 5, 5, 222, 236, 5, 5, 219, 236, 5, 5, 226, 236, 192, 0, 0, 0, 5, 5, 70, 239, 5, 5, 219, 214, 5, 5, 198, 244, 5, 5, 67, 239, 5, 5, 68, 239, 5, 5, 66, 239, 5, 5, 211, 231, 5, 5, 218, 214, 5, 5, 71, 239, 5, 5, 65, 239, 5, 5, 217, 214, 5, 5, 220, 214, 5, 5, 69, 239, 5, 5, 123, 241, 5, 5, 5, 216, 5, 5, 124, 241, 5, 5, 125, 241, 192, 0, 0, 0, 5, 5, 69, 243, 5, 5, 43, 217, 5, 5, 199, 244, 5, 5, 200, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 244, 5, 5, 23, 246, 5, 5, 22, 246, 5, 5, 25, 247, 192, 0, 0, 0, 5, 5, 24, 247, 5, 5, 238, 247, 5, 5, 239, 247, 5, 5, 237, 247, 5, 5, 152, 248, 5, 5, 118, 219, 5, 5, 148, 201, 5, 5, 154, 204, 5, 5, 153, 204, 5, 5, 98, 206, 5, 5, 97, 206, 5, 5, 208, 226, 5, 5, 209, 226, 5, 5, 207, 226, 192, 0, 0, 0, 5, 5, 96, 206, 5, 5, 64, 229, 192, 0, 0, 0, 5, 5, 65, 229, 5, 5, 61, 229, 5, 5, 74, 208, 5, 5, 62, 229, 192, 0, 0, 0, 5, 5, 63, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 231, 5, 5, 227, 231, 5, 5, 223, 231, 5, 5, 226, 231, 5, 5, 18, 210, 5, 5, 224, 231, 5, 5, 19, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 232, 5, 5, 209, 211, 5, 5, 208, 211, 192, 0, 0, 0, 5, 5, 207, 211, 192, 0, 0, 0, 5, 5, 133, 234, 5, 5, 134, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 236, 5, 5, 232, 236, 5, 5, 94, 213, 5, 5, 91, 213, 192, 0, 0, 0, 5, 5, 93, 213, 5, 5, 92, 213, 5, 5, 231, 236, 5, 5, 229, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 239, 5, 5, 75, 239, 5, 5, 76, 239, 5, 5, 224, 214, 5, 5, 77, 239, 5, 5, 72, 239, 5, 5, 73, 239, 5, 5, 221, 214, 192, 0, 0, 0, 5, 5, 222, 214, 5, 5, 74, 239, 5, 5, 223, 214, 192, 0, 0, 0, 5, 5, 129, 241, 192, 0, 0, 0, 5, 5, 130, 241, 5, 5, 127, 241, 5, 5, 128, 241, 5, 5, 6, 216, 5, 5, 126, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 243, 5, 5, 70, 243, 5, 5, 72, 243, 5, 5, 46, 217, 5, 5, 76, 243, 5, 5, 45, 217, 5, 5, 74, 243, 5, 5, 77, 243, 5, 5, 75, 243, 5, 5, 71, 243, 5, 5, 44, 217, 192, 0, 0, 0, 5, 5, 203, 244, 192, 0, 0, 0, 5, 5, 204, 244, 5, 5, 202, 244, 192, 0, 0, 0, 5, 5, 206, 244, 5, 5, 208, 217, 5, 5, 209, 217, 5, 5, 25, 246, 5, 5, 113, 218, 5, 5, 205, 244, 5, 5, 24, 246, 5, 5, 112, 218, 5, 5, 28, 247, 5, 5, 29, 247, 5, 5, 230, 218, 5, 5, 27, 247, 192, 0, 0, 0, 5, 5, 26, 247, 192, 0, 0, 0, 5, 5, 240, 247, 5, 5, 48, 219, 192, 0, 0, 0, 5, 5, 26, 249, 5, 5, 119, 249, 5, 5, 118, 249, 192, 0, 0, 0, 5, 5, 200, 219, 5, 5, 226, 219, 5, 5, 120, 249, 5, 5, 223, 249, 192, 0, 0, 0, 5, 5, 244, 219, 5, 5, 149, 201, 5, 5, 13, 203, 192, 0, 0, 0, 5, 5, 210, 211, 192, 0, 0, 0, 5, 5, 231, 218, 5, 5, 150, 201, 5, 5, 155, 204, 5, 5, 75, 208, 5, 5, 20, 210, 5, 5, 210, 217, 5, 5, 14, 203, 192, 0, 0, 0, 5, 5, 183, 224, 5, 5, 182, 224, 5, 5, 210, 226, 192, 0, 0, 0, 5, 5, 101, 206, 5, 5, 99, 206, 5, 5, 103, 206, 5, 5, 212, 226, 192, 0, 0, 0, 5, 5, 102, 206, 5, 5, 100, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 226, 5, 5, 78, 208, 5, 5, 71, 229, 192, 0, 0, 0, 5, 5, 77, 208, 5, 5, 79, 208, 5, 5, 76, 229, 5, 5, 81, 208, 5, 5, 73, 229, 5, 5, 74, 229, 5, 5, 66, 229, 5, 5, 80, 208, 5, 5, 77, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 229, 192, 0, 0, 0, 5, 5, 69, 229, 5, 5, 67, 229, 5, 5, 76, 208, 192, 0, 0, 0, 5, 5, 75, 229, 5, 5, 70, 229, 5, 5, 72, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 231, 192, 0, 0, 0, 5, 5, 238, 231, 5, 5, 236, 231, 5, 5, 228, 231, 5, 5, 243, 231, 5, 5, 231, 231, 5, 5, 241, 231, 5, 5, 248, 231, 5, 5, 240, 231, 5, 5, 28, 210, 192, 0, 0, 0, 5, 5, 24, 210, 5, 5, 229, 231, 192, 0, 0, 0, 5, 5, 235, 231, 5, 5, 26, 210, 192, 0, 0, 0, 5, 5, 234, 231, 5, 5, 27, 210, 5, 5, 29, 210, 5, 5, 237, 231, 192, 0, 0, 0, 5, 5, 21, 210, 5, 5, 22, 210, 5, 5, 249, 231, 5, 5, 247, 231, 192, 0, 0, 0, 5, 5, 246, 231, 5, 5, 232, 231, 192, 0, 0, 0, 5, 5, 244, 231, 192, 0, 0, 0, 5, 5, 25, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 210, 5, 5, 245, 231, 5, 5, 230, 231, 5, 5, 233, 231, 5, 5, 242, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 237, 5, 5, 144, 234, 192, 0, 0, 0, 5, 5, 137, 234, 5, 5, 156, 234, 5, 5, 165, 234, 5, 5, 140, 234, 5, 5, 221, 211, 192, 0, 0, 0, 5, 5, 153, 234, 5, 5, 211, 211, 5, 5, 213, 211, 5, 5, 225, 211, 5, 5, 138, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 211, 5, 5, 216, 211, 5, 5, 226, 211, 5, 5, 215, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 211, 5, 5, 146, 234, 192, 0, 0, 0, 5, 5, 135, 234, 5, 5, 222, 211, 5, 5, 145, 234, 5, 5, 218, 211, 5, 5, 163, 234, 5, 5, 223, 211, 5, 5, 161, 234, 5, 5, 149, 234, 5, 5, 159, 234, 5, 5, 220, 211, 5, 5, 136, 234, 5, 5, 162, 234, 5, 5, 157, 234, 192, 0, 0, 0, 5, 5, 164, 234, 5, 5, 212, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 211, 5, 5, 217, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 234, 192, 0, 0, 0, 5, 5, 150, 234, 5, 5, 154, 234, 192, 0, 0, 0, 5, 5, 158, 234, 5, 5, 219, 211, 5, 5, 141, 234, 5, 5, 139, 234, 5, 5, 151, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 234, 5, 5, 147, 234, 192, 0, 0, 0, 5, 5, 152, 234, 5, 5, 155, 234, 192, 0, 0, 0, 5, 5, 160, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 236, 192, 0, 0, 0, 5, 5, 95, 213, 5, 5, 245, 236, 5, 5, 237, 236, 5, 5, 100, 213, 5, 5, 104, 213, 5, 5, 254, 236, 5, 5, 8, 237, 5, 5, 247, 236, 5, 5, 96, 213, 192, 0, 0, 0, 5, 5, 251, 236, 5, 5, 143, 234, 192, 0, 0, 0, 5, 5, 97, 213, 5, 5, 4, 237, 5, 5, 9, 237, 5, 5, 255, 236, 192, 0, 0, 0, 5, 5, 3, 237, 5, 5, 12, 237, 5, 5, 5, 237, 5, 5, 241, 236, 5, 5, 250, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 213, 192, 0, 0, 0, 5, 5, 101, 213, 5, 5, 239, 236, 5, 5, 252, 236, 5, 5, 99, 213, 5, 5, 246, 236, 5, 5, 98, 213, 5, 5, 6, 237, 5, 5, 243, 236, 5, 5, 236, 236, 5, 5, 102, 213, 5, 5, 11, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 236, 5, 5, 2, 237, 5, 5, 253, 236, 5, 5, 248, 236, 5, 5, 234, 236, 5, 5, 233, 236, 5, 5, 242, 236, 5, 5, 7, 237, 5, 5, 103, 213, 5, 5, 10, 237, 5, 5, 240, 236, 5, 5, 244, 236, 5, 5, 229, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 214, 5, 5, 232, 214, 5, 5, 111, 239, 5, 5, 108, 239, 5, 5, 84, 239, 5, 5, 227, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 214, 5, 5, 233, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 239, 5, 5, 231, 214, 5, 5, 94, 239, 5, 5, 80, 239, 5, 5, 81, 239, 5, 5, 225, 214, 5, 5, 110, 239, 5, 5, 235, 214, 5, 5, 97, 239, 5, 5, 102, 239, 5, 5, 96, 239, 5, 5, 148, 241, 5, 5, 92, 239, 5, 5, 100, 239, 5, 5, 98, 239, 5, 5, 85, 239, 5, 5, 79, 239, 5, 5, 106, 239, 5, 5, 234, 214, 5, 5, 107, 239, 192, 0, 0, 0, 5, 5, 101, 239, 192, 0, 0, 0, 5, 5, 90, 239, 5, 5, 88, 239, 5, 5, 83, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 239, 5, 5, 104, 239, 5, 5, 87, 239, 5, 5, 103, 239, 5, 5, 109, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 214, 192, 0, 0, 0, 5, 5, 99, 239, 5, 5, 105, 239, 5, 5, 95, 239, 5, 5, 89, 239, 5, 5, 228, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 236, 5, 5, 93, 239, 5, 5, 236, 214, 5, 5, 86, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 241, 5, 5, 9, 216, 5, 5, 153, 241, 5, 5, 150, 241, 5, 5, 169, 241, 5, 5, 13, 216, 192, 0, 0, 0, 5, 5, 166, 241, 5, 5, 158, 241, 5, 5, 168, 241, 5, 5, 143, 241, 5, 5, 156, 241, 192, 0, 0, 0, 5, 5, 15, 216, 192, 0, 0, 0, 5, 5, 135, 241, 192, 0, 0, 0, 5, 5, 133, 241, 5, 5, 167, 241, 192, 0, 0, 0, 5, 5, 165, 241, 5, 5, 164, 241, 5, 5, 147, 241, 5, 5, 146, 241, 5, 5, 136, 241, 5, 5, 17, 216, 192, 0, 0, 0, 5, 5, 142, 241, 5, 5, 152, 241, 5, 5, 163, 241, 5, 5, 20, 216, 5, 5, 170, 241, 192, 0, 0, 0, 5, 5, 54, 217, 5, 5, 22, 216, 5, 5, 16, 216, 5, 5, 140, 241, 192, 0, 0, 0, 5, 5, 149, 241, 5, 5, 132, 241, 5, 5, 134, 241, 5, 5, 7, 216, 5, 5, 19, 216, 5, 5, 12, 216, 5, 5, 141, 241, 5, 5, 157, 241, 5, 5, 151, 241, 5, 5, 18, 216, 5, 5, 131, 241, 5, 5, 49, 217, 5, 5, 159, 241, 5, 5, 162, 241, 5, 5, 14, 216, 192, 0, 0, 0, 5, 5, 145, 241, 5, 5, 21, 216, 5, 5, 11, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 216, 5, 5, 155, 241, 5, 5, 161, 241, 5, 5, 8, 216, 192, 0, 0, 0, 5, 5, 138, 241, 5, 5, 160, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 241, 5, 5, 83, 243, 192, 0, 0, 0, 5, 5, 88, 243, 5, 5, 51, 217, 5, 5, 53, 217, 5, 5, 94, 243, 5, 5, 47, 217, 5, 5, 109, 243, 5, 5, 100, 243, 5, 5, 89, 243, 5, 5, 90, 243, 5, 5, 99, 243, 192, 0, 0, 0, 5, 5, 60, 217, 5, 5, 98, 243, 5, 5, 82, 243, 5, 5, 97, 243, 5, 5, 85, 243, 5, 5, 110, 243, 5, 5, 59, 217, 5, 5, 57, 217, 5, 5, 86, 243, 192, 0, 0, 0, 5, 5, 105, 243, 192, 0, 0, 0, 5, 5, 91, 243, 5, 5, 104, 243, 192, 0, 0, 0, 5, 5, 106, 243, 5, 5, 81, 243, 5, 5, 52, 217, 192, 0, 0, 0, 5, 5, 107, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 243, 192, 0, 0, 0, 5, 5, 56, 217, 5, 5, 92, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 217, 5, 5, 101, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 217, 5, 5, 87, 243, 5, 5, 102, 243, 192, 0, 0, 0, 5, 5, 96, 243, 5, 5, 137, 241, 5, 5, 103, 243, 5, 5, 84, 243, 192, 0, 0, 0, 5, 5, 55, 217, 192, 0, 0, 0, 5, 5, 108, 243, 192, 0, 0, 0, 5, 5, 48, 217, 5, 5, 79, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 244, 5, 5, 213, 244, 5, 5, 212, 217, 192, 0, 0, 0, 5, 5, 208, 244, 5, 5, 223, 244, 5, 5, 215, 244, 5, 5, 93, 243, 192, 0, 0, 0, 5, 5, 225, 244, 5, 5, 209, 244, 192, 0, 0, 0, 5, 5, 211, 217, 5, 5, 219, 244, 5, 5, 213, 217, 5, 5, 221, 217, 5, 5, 219, 217, 5, 5, 221, 244, 5, 5, 220, 217, 5, 5, 211, 244, 192, 0, 0, 0, 5, 5, 212, 244, 5, 5, 217, 244, 5, 5, 222, 244, 192, 0, 0, 0, 5, 5, 78, 243, 5, 5, 214, 217, 5, 5, 230, 244, 5, 5, 227, 244, 5, 5, 231, 244, 5, 5, 218, 244, 5, 5, 214, 244, 5, 5, 228, 244, 5, 5, 36, 246, 5, 5, 216, 244, 192, 0, 0, 0, 5, 5, 217, 217, 192, 0, 0, 0, 5, 5, 216, 217, 5, 5, 80, 243, 5, 5, 218, 217, 5, 5, 224, 244, 5, 5, 226, 244, 5, 5, 215, 217, 5, 5, 229, 244, 5, 5, 207, 244, 192, 0, 0, 0, 5, 5, 210, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 246, 192, 0, 0, 0, 5, 5, 29, 246, 5, 5, 117, 218, 5, 5, 43, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 246, 5, 5, 118, 218, 192, 0, 0, 0, 5, 5, 38, 246, 192, 0, 0, 0, 5, 5, 34, 246, 5, 5, 123, 218, 5, 5, 44, 246, 5, 5, 28, 246, 5, 5, 31, 246, 5, 5, 115, 218, 5, 5, 46, 246, 192, 0, 0, 0, 5, 5, 39, 246, 5, 5, 42, 246, 5, 5, 121, 218, 5, 5, 126, 218, 5, 5, 124, 218, 5, 5, 35, 246, 5, 5, 30, 246, 192, 0, 0, 0, 5, 5, 119, 218, 5, 5, 120, 218, 5, 5, 26, 246, 5, 5, 116, 218, 192, 0, 0, 0, 5, 5, 114, 218, 5, 5, 122, 218, 5, 5, 41, 246, 5, 5, 125, 218, 192, 0, 0, 0, 5, 5, 37, 246, 5, 5, 47, 246, 5, 5, 127, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 246, 192, 0, 0, 0, 5, 5, 40, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 247, 5, 5, 56, 247, 5, 5, 44, 247, 5, 5, 51, 247, 5, 5, 32, 246, 5, 5, 49, 247, 5, 5, 32, 247, 5, 5, 54, 247, 5, 5, 234, 218, 5, 5, 36, 247, 5, 5, 53, 247, 5, 5, 43, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 247, 5, 5, 45, 247, 192, 0, 0, 0, 5, 5, 50, 247, 5, 5, 52, 247, 5, 5, 30, 247, 5, 5, 55, 247, 5, 5, 41, 247, 5, 5, 46, 247, 5, 5, 34, 247, 5, 5, 38, 247, 5, 5, 57, 247, 5, 5, 48, 247, 5, 5, 31, 247, 5, 5, 35, 247, 5, 5, 37, 247, 5, 5, 47, 247, 192, 0, 0, 0, 5, 5, 232, 218, 5, 5, 40, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 247, 5, 5, 244, 247, 5, 5, 249, 247, 5, 5, 55, 219, 5, 5, 251, 247, 192, 0, 0, 0, 5, 5, 49, 219, 192, 0, 0, 0, 5, 5, 247, 247, 5, 5, 253, 247, 5, 5, 54, 219, 5, 5, 50, 219, 192, 0, 0, 0, 5, 5, 51, 219, 5, 5, 243, 247, 5, 5, 250, 247, 5, 5, 53, 219, 5, 5, 248, 247, 5, 5, 52, 219, 5, 5, 242, 247, 5, 5, 246, 247, 5, 5, 245, 247, 192, 0, 0, 0, 5, 5, 241, 247, 5, 5, 252, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 219, 5, 5, 159, 248, 5, 5, 162, 248, 5, 5, 158, 248, 5, 5, 160, 248, 5, 5, 161, 248, 5, 5, 155, 248, 5, 5, 156, 248, 5, 5, 153, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 248, 5, 5, 154, 248, 5, 5, 120, 219, 5, 5, 121, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 249, 192, 0, 0, 0, 5, 5, 30, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 249, 5, 5, 31, 249, 192, 0, 0, 0, 5, 5, 166, 219, 192, 0, 0, 0, 5, 5, 27, 249, 5, 5, 165, 219, 5, 5, 167, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 249, 5, 5, 123, 249, 5, 5, 201, 219, 5, 5, 121, 249, 192, 0, 0, 0, 5, 5, 193, 249, 5, 5, 192, 249, 5, 5, 194, 249, 5, 5, 228, 219, 5, 5, 195, 249, 5, 5, 227, 219, 5, 5, 196, 249, 5, 5, 224, 249, 5, 5, 227, 249, 5, 5, 226, 249, 5, 5, 237, 219, 5, 5, 225, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 219, 5, 5, 245, 219, 5, 5, 246, 219, 5, 5, 252, 219, 5, 5, 14, 250, 5, 5, 15, 250, 5, 5, 13, 250, 5, 5, 23, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 229, 5, 5, 250, 231, 5, 5, 112, 239, 5, 5, 48, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 203, 192, 0, 0, 0, 5, 5, 156, 204, 5, 5, 104, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 229, 192, 0, 0, 0, 5, 5, 80, 229, 5, 5, 82, 208, 192, 0, 0, 0, 5, 5, 32, 210, 5, 5, 252, 231, 5, 5, 251, 231, 5, 5, 36, 210, 5, 5, 31, 210, 5, 5, 253, 231, 5, 5, 33, 210, 5, 5, 35, 210, 5, 5, 34, 210, 5, 5, 30, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 234, 5, 5, 167, 234, 192, 0, 0, 0, 5, 5, 168, 234, 5, 5, 166, 234, 192, 0, 0, 0, 5, 5, 106, 213, 192, 0, 0, 0, 5, 5, 109, 213, 5, 5, 111, 213, 5, 5, 110, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 213, 5, 5, 108, 213, 192, 0, 0, 0, 5, 5, 114, 239, 5, 5, 113, 239, 5, 5, 238, 214, 5, 5, 115, 239, 192, 0, 0, 0, 5, 5, 116, 239, 5, 5, 239, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 241, 5, 5, 176, 241, 5, 5, 116, 243, 192, 0, 0, 0, 5, 5, 174, 241, 5, 5, 175, 241, 5, 5, 23, 216, 5, 5, 171, 241, 5, 5, 179, 241, 5, 5, 173, 241, 5, 5, 177, 241, 5, 5, 112, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 243, 192, 0, 0, 0, 5, 5, 115, 243, 5, 5, 65, 217, 5, 5, 111, 243, 5, 5, 64, 217, 5, 5, 113, 243, 5, 5, 61, 217, 5, 5, 62, 217, 5, 5, 63, 217, 5, 5, 172, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 217, 5, 5, 234, 244, 5, 5, 232, 244, 5, 5, 233, 244, 5, 5, 222, 217, 5, 5, 225, 217, 5, 5, 223, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 246, 5, 5, 50, 246, 5, 5, 128, 218, 192, 0, 0, 0, 5, 5, 59, 247, 5, 5, 61, 247, 5, 5, 60, 247, 5, 5, 235, 218, 5, 5, 56, 219, 5, 5, 2, 248, 5, 5, 255, 247, 5, 5, 254, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 203, 192, 0, 0, 0, 5, 5, 77, 220, 192, 0, 0, 0, 5, 5, 186, 220, 5, 5, 134, 200, 5, 5, 184, 220, 5, 5, 187, 220, 5, 5, 185, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 221, 192, 0, 0, 0, 5, 5, 154, 201, 192, 0, 0, 0, 5, 5, 155, 201, 5, 5, 143, 221, 5, 5, 152, 201, 5, 5, 142, 221, 5, 5, 140, 221, 5, 5, 153, 201, 5, 5, 151, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 222, 5, 5, 246, 222, 5, 5, 20, 203, 5, 5, 245, 222, 5, 5, 244, 222, 192, 0, 0, 0, 5, 5, 19, 203, 5, 5, 18, 203, 192, 0, 0, 0, 5, 5, 22, 203, 5, 5, 247, 222, 5, 5, 21, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 224, 5, 5, 158, 204, 5, 5, 159, 204, 5, 5, 160, 204, 5, 5, 189, 224, 5, 5, 185, 224, 5, 5, 157, 204, 5, 5, 186, 224, 192, 0, 0, 0, 5, 5, 187, 224, 5, 5, 184, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 206, 5, 5, 213, 226, 5, 5, 109, 206, 5, 5, 112, 206, 5, 5, 214, 226, 192, 0, 0, 0, 5, 5, 107, 206, 5, 5, 105, 206, 5, 5, 106, 206, 5, 5, 110, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 208, 5, 5, 83, 229, 5, 5, 91, 208, 5, 5, 82, 229, 192, 0, 0, 0, 5, 5, 85, 229, 5, 5, 87, 208, 5, 5, 84, 229, 5, 5, 44, 210, 5, 5, 85, 208, 5, 5, 88, 208, 5, 5, 84, 208, 5, 5, 89, 208, 5, 5, 90, 208, 5, 5, 86, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 229, 5, 5, 40, 210, 5, 5, 255, 231, 192, 0, 0, 0, 5, 5, 5, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 232, 5, 5, 45, 210, 5, 5, 41, 210, 5, 5, 42, 210, 5, 5, 254, 231, 5, 5, 2, 232, 5, 5, 3, 232, 5, 5, 37, 210, 5, 5, 39, 210, 192, 0, 0, 0, 5, 5, 43, 210, 5, 5, 38, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 234, 5, 5, 170, 234, 5, 5, 171, 234, 5, 5, 230, 211, 5, 5, 231, 211, 192, 0, 0, 0, 5, 5, 173, 234, 5, 5, 229, 211, 5, 5, 112, 213, 192, 0, 0, 0, 5, 5, 114, 213, 5, 5, 113, 213, 192, 0, 0, 0, 5, 5, 14, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 237, 5, 5, 118, 239, 192, 0, 0, 0, 5, 5, 117, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 216, 5, 5, 25, 216, 5, 5, 180, 241, 5, 5, 26, 216, 192, 0, 0, 0, 5, 5, 119, 243, 192, 0, 0, 0, 5, 5, 117, 243, 192, 0, 0, 0, 5, 5, 118, 243, 5, 5, 66, 217, 192, 0, 0, 0, 5, 5, 235, 244, 5, 5, 129, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 217, 5, 5, 23, 203, 192, 0, 0, 0, 5, 5, 113, 206, 5, 5, 215, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 229, 5, 5, 92, 208, 5, 5, 46, 210, 5, 5, 6, 232, 5, 5, 8, 232, 5, 5, 48, 210, 5, 5, 47, 210, 5, 5, 49, 210, 5, 5, 50, 210, 5, 5, 7, 232, 5, 5, 234, 211, 5, 5, 235, 211, 5, 5, 233, 211, 5, 5, 115, 213, 5, 5, 232, 211, 5, 5, 174, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 213, 5, 5, 119, 239, 5, 5, 181, 241, 5, 5, 27, 216, 5, 5, 68, 217, 5, 5, 236, 244, 5, 5, 240, 244, 5, 5, 228, 217, 5, 5, 237, 244, 5, 5, 229, 217, 5, 5, 227, 217, 5, 5, 241, 244, 5, 5, 230, 217, 5, 5, 239, 244, 192, 0, 0, 0, 5, 5, 51, 246, 5, 5, 226, 217, 5, 5, 130, 218, 192, 0, 0, 0, 5, 5, 124, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 203, 5, 5, 94, 208, 5, 5, 93, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 210, 5, 5, 10, 232, 5, 5, 9, 232, 5, 5, 52, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 234, 5, 5, 239, 211, 5, 5, 236, 211, 5, 5, 177, 234, 5, 5, 238, 211, 5, 5, 175, 234, 5, 5, 237, 211, 192, 0, 0, 0, 5, 5, 176, 234, 192, 0, 0, 0, 5, 5, 16, 237, 5, 5, 117, 213, 192, 0, 0, 0, 5, 5, 122, 239, 192, 0, 0, 0, 5, 5, 240, 214, 5, 5, 120, 239, 5, 5, 241, 214, 5, 5, 242, 214, 5, 5, 121, 239, 5, 5, 243, 214, 192, 0, 0, 0, 5, 5, 182, 241, 192, 0, 0, 0, 5, 5, 31, 216, 5, 5, 28, 216, 5, 5, 33, 216, 5, 5, 184, 241, 5, 5, 29, 216, 5, 5, 183, 241, 5, 5, 32, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 216, 192, 0, 0, 0, 5, 5, 122, 243, 5, 5, 124, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 217, 5, 5, 123, 243, 5, 5, 70, 217, 5, 5, 121, 243, 5, 5, 120, 243, 192, 0, 0, 0, 5, 5, 243, 244, 5, 5, 242, 244, 5, 5, 231, 217, 5, 5, 244, 244, 5, 5, 56, 246, 5, 5, 132, 218, 5, 5, 55, 246, 5, 5, 52, 246, 5, 5, 131, 218, 5, 5, 53, 246, 5, 5, 54, 246, 192, 0, 0, 0, 5, 5, 62, 247, 5, 5, 63, 247, 5, 5, 236, 218, 192, 0, 0, 0, 5, 5, 59, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 219, 5, 5, 58, 219, 5, 5, 4, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 219, 5, 5, 123, 219, 5, 5, 163, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 219, 5, 5, 126, 249, 5, 5, 204, 219, 192, 0, 0, 0, 5, 5, 125, 249, 5, 5, 127, 249, 5, 5, 203, 219, 5, 5, 197, 249, 192, 0, 0, 0, 5, 5, 249, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 211, 192, 0, 0, 0, 5, 5, 17, 237, 192, 0, 0, 0, 5, 5, 123, 239, 5, 5, 34, 216, 5, 5, 35, 216, 192, 0, 0, 0, 5, 5, 26, 203, 192, 0, 0, 0, 5, 5, 244, 214, 5, 5, 133, 218, 5, 5, 161, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 216, 192, 0, 0, 0, 5, 5, 168, 219, 5, 5, 162, 204, 5, 5, 87, 229, 192, 0, 0, 0, 5, 5, 11, 232, 192, 0, 0, 0, 5, 5, 13, 232, 192, 0, 0, 0, 5, 5, 12, 232, 192, 0, 0, 0, 5, 5, 182, 234, 5, 5, 179, 234, 5, 5, 241, 211, 192, 0, 0, 0, 5, 5, 242, 211, 5, 5, 180, 234, 5, 5, 181, 234, 192, 0, 0, 0, 5, 5, 19, 237, 5, 5, 24, 237, 5, 5, 118, 213, 5, 5, 18, 237, 5, 5, 20, 237, 5, 5, 27, 237, 5, 5, 22, 237, 5, 5, 26, 237, 5, 5, 23, 237, 5, 5, 21, 237, 5, 5, 25, 237, 5, 5, 119, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 239, 192, 0, 0, 0, 5, 5, 124, 239, 5, 5, 246, 214, 192, 0, 0, 0, 5, 5, 245, 214, 5, 5, 125, 239, 5, 5, 247, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 241, 5, 5, 37, 216, 5, 5, 185, 241, 5, 5, 125, 243, 192, 0, 0, 0, 5, 5, 127, 243, 5, 5, 129, 243, 5, 5, 128, 243, 192, 0, 0, 0, 5, 5, 71, 217, 5, 5, 126, 243, 5, 5, 251, 244, 5, 5, 232, 217, 5, 5, 249, 244, 5, 5, 252, 244, 5, 5, 233, 217, 192, 0, 0, 0, 5, 5, 247, 244, 192, 0, 0, 0, 5, 5, 250, 244, 5, 5, 248, 244, 5, 5, 245, 244, 5, 5, 234, 217, 5, 5, 246, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 246, 5, 5, 58, 246, 192, 0, 0, 0, 5, 5, 64, 247, 192, 0, 0, 0, 5, 5, 65, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 248, 192, 0, 0, 0, 5, 5, 125, 219, 192, 0, 0, 0, 5, 5, 124, 219, 5, 5, 32, 249, 5, 5, 33, 249, 5, 5, 205, 219, 5, 5, 128, 249, 192, 0, 0, 0, 5, 5, 238, 219, 192, 0, 0, 0, 5, 5, 163, 204, 5, 5, 53, 210, 5, 5, 29, 237, 5, 5, 28, 237, 5, 5, 128, 239, 5, 5, 127, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 217, 5, 5, 131, 243, 5, 5, 130, 243, 5, 5, 255, 244, 5, 5, 253, 244, 5, 5, 2, 245, 5, 5, 254, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 218, 5, 5, 60, 246, 5, 5, 61, 246, 5, 5, 62, 246, 192, 0, 0, 0, 5, 5, 6, 248, 192, 0, 0, 0, 5, 5, 164, 248, 192, 0, 0, 0, 5, 5, 129, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 241, 5, 5, 132, 243, 192, 0, 0, 0, 5, 5, 165, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 213, 192, 0, 0, 0, 5, 5, 189, 241, 5, 5, 235, 217, 5, 5, 3, 245, 5, 5, 135, 218, 192, 0, 0, 0, 5, 5, 66, 247, 5, 5, 67, 247, 5, 5, 60, 219, 5, 5, 34, 249, 5, 5, 166, 204, 5, 5, 97, 208, 5, 5, 98, 208, 5, 5, 88, 229, 5, 5, 54, 210, 5, 5, 55, 210, 5, 5, 14, 232, 5, 5, 56, 210, 192, 0, 0, 0, 5, 5, 246, 211, 192, 0, 0, 0, 5, 5, 248, 211, 5, 5, 184, 234, 5, 5, 185, 234, 5, 5, 183, 234, 5, 5, 243, 211, 5, 5, 244, 211, 5, 5, 247, 211, 5, 5, 245, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 237, 5, 5, 121, 213, 5, 5, 122, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 239, 5, 5, 250, 214, 5, 5, 130, 239, 5, 5, 129, 239, 192, 0, 0, 0, 5, 5, 134, 239, 5, 5, 248, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 216, 192, 0, 0, 0, 5, 5, 131, 239, 5, 5, 136, 239, 5, 5, 133, 239, 5, 5, 132, 239, 192, 0, 0, 0, 5, 5, 249, 214, 192, 0, 0, 0, 5, 5, 42, 216, 192, 0, 0, 0, 5, 5, 191, 241, 5, 5, 38, 216, 192, 0, 0, 0, 5, 5, 192, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 241, 192, 0, 0, 0, 5, 5, 41, 216, 5, 5, 39, 216, 5, 5, 43, 216, 192, 0, 0, 0, 5, 5, 40, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 243, 192, 0, 0, 0, 5, 5, 138, 243, 5, 5, 134, 243, 5, 5, 137, 243, 5, 5, 73, 217, 192, 0, 0, 0, 5, 5, 182, 242, 5, 5, 136, 243, 5, 5, 135, 243, 192, 0, 0, 0, 5, 5, 238, 217, 5, 5, 236, 217, 5, 5, 239, 217, 5, 5, 237, 217, 5, 5, 4, 245, 5, 5, 5, 245, 5, 5, 6, 245, 5, 5, 240, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 246, 5, 5, 137, 218, 5, 5, 64, 246, 192, 0, 0, 0, 5, 5, 138, 218, 5, 5, 63, 246, 5, 5, 65, 246, 5, 5, 136, 218, 5, 5, 71, 247, 5, 5, 68, 247, 192, 0, 0, 0, 5, 5, 69, 247, 5, 5, 70, 247, 5, 5, 7, 248, 5, 5, 62, 219, 192, 0, 0, 0, 5, 5, 61, 219, 192, 0, 0, 0, 5, 5, 166, 248, 5, 5, 165, 248, 5, 5, 126, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 219, 5, 5, 206, 219, 5, 5, 229, 219, 5, 5, 198, 249, 5, 5, 250, 249, 5, 5, 251, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 204, 5, 5, 15, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 234, 5, 5, 31, 237, 5, 5, 32, 237, 5, 5, 123, 213, 192, 0, 0, 0, 5, 5, 124, 213, 5, 5, 137, 239, 5, 5, 251, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 217, 192, 0, 0, 0, 5, 5, 7, 245, 192, 0, 0, 0, 5, 5, 241, 217, 5, 5, 69, 246, 5, 5, 139, 218, 5, 5, 68, 246, 5, 5, 70, 246, 5, 5, 67, 246, 5, 5, 10, 248, 5, 5, 72, 247, 5, 5, 73, 247, 192, 0, 0, 0, 5, 5, 237, 218, 192, 0, 0, 0, 5, 5, 9, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 248, 192, 0, 0, 0, 5, 5, 167, 248, 5, 5, 252, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 206, 5, 5, 216, 226, 192, 0, 0, 0, 5, 5, 89, 229, 192, 0, 0, 0, 5, 5, 57, 210, 192, 0, 0, 0, 5, 5, 60, 210, 5, 5, 58, 210, 5, 5, 16, 232, 192, 0, 0, 0, 5, 5, 62, 210, 192, 0, 0, 0, 5, 5, 59, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 210, 192, 0, 0, 0, 5, 5, 250, 211, 192, 0, 0, 0, 5, 5, 187, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 234, 5, 5, 139, 239, 192, 0, 0, 0, 5, 5, 249, 211, 5, 5, 251, 211, 5, 5, 252, 211, 192, 0, 0, 0, 5, 5, 34, 237, 192, 0, 0, 0, 5, 5, 33, 237, 5, 5, 125, 213, 192, 0, 0, 0, 5, 5, 126, 213, 192, 0, 0, 0, 5, 5, 35, 237, 5, 5, 138, 239, 5, 5, 128, 213, 5, 5, 252, 214, 192, 0, 0, 0, 5, 5, 127, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 216, 5, 5, 140, 239, 5, 5, 254, 214, 5, 5, 253, 214, 5, 5, 141, 239, 5, 5, 144, 239, 5, 5, 142, 239, 5, 5, 143, 239, 5, 5, 255, 214, 192, 0, 0, 0, 5, 5, 50, 216, 5, 5, 48, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 216, 5, 5, 194, 241, 192, 0, 0, 0, 5, 5, 49, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 241, 5, 5, 139, 243, 192, 0, 0, 0, 5, 5, 195, 241, 5, 5, 46, 216, 5, 5, 196, 241, 5, 5, 142, 243, 5, 5, 140, 243, 5, 5, 141, 243, 5, 5, 146, 243, 5, 5, 245, 217, 5, 5, 145, 243, 5, 5, 148, 243, 5, 5, 147, 243, 5, 5, 144, 243, 5, 5, 143, 243, 192, 0, 0, 0, 5, 5, 75, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 245, 192, 0, 0, 0, 5, 5, 9, 245, 5, 5, 244, 217, 5, 5, 242, 217, 5, 5, 243, 217, 192, 0, 0, 0, 5, 5, 8, 245, 192, 0, 0, 0, 5, 5, 73, 246, 192, 0, 0, 0, 5, 5, 140, 218, 192, 0, 0, 0, 5, 5, 72, 246, 5, 5, 71, 246, 5, 5, 141, 218, 192, 0, 0, 0, 5, 5, 78, 247, 5, 5, 77, 247, 192, 0, 0, 0, 5, 5, 75, 247, 192, 0, 0, 0, 5, 5, 74, 247, 5, 5, 239, 218, 5, 5, 238, 218, 5, 5, 79, 247, 5, 5, 168, 248, 5, 5, 127, 219, 5, 5, 12, 248, 5, 5, 63, 219, 5, 5, 11, 248, 5, 5, 76, 247, 192, 0, 0, 0, 5, 5, 169, 248, 5, 5, 170, 219, 192, 0, 0, 0, 5, 5, 230, 219, 5, 5, 199, 249, 192, 0, 0, 0, 5, 5, 253, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 204, 5, 5, 90, 229, 5, 5, 149, 243, 5, 5, 171, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 237, 5, 5, 36, 237, 5, 5, 197, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 243, 192, 0, 0, 0, 5, 5, 150, 243, 192, 0, 0, 0, 5, 5, 246, 217, 5, 5, 74, 246, 5, 5, 75, 246, 5, 5, 240, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 249, 5, 5, 115, 206, 5, 5, 64, 210, 5, 5, 63, 210, 5, 5, 189, 234, 5, 5, 191, 234, 5, 5, 254, 211, 5, 5, 190, 234, 5, 5, 253, 211, 5, 5, 255, 211, 5, 5, 192, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 237, 5, 5, 41, 237, 5, 5, 40, 237, 192, 0, 0, 0, 5, 5, 43, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 213, 5, 5, 42, 237, 5, 5, 38, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 237, 192, 0, 0, 0, 5, 5, 151, 239, 192, 0, 0, 0, 5, 5, 154, 239, 5, 5, 156, 239, 5, 5, 146, 239, 5, 5, 150, 239, 5, 5, 147, 239, 5, 5, 3, 215, 5, 5, 6, 215, 5, 5, 8, 215, 5, 5, 148, 239, 5, 5, 149, 239, 5, 5, 7, 215, 5, 5, 152, 239, 5, 5, 155, 239, 5, 5, 153, 239, 5, 5, 9, 215, 192, 0, 0, 0, 5, 5, 5, 215, 5, 5, 145, 239, 5, 5, 2, 215, 192, 0, 0, 0, 5, 5, 4, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 216, 5, 5, 203, 241, 5, 5, 201, 241, 5, 5, 200, 241, 192, 0, 0, 0, 5, 5, 206, 241, 192, 0, 0, 0, 5, 5, 205, 241, 5, 5, 204, 241, 192, 0, 0, 0, 5, 5, 199, 241, 5, 5, 51, 216, 5, 5, 198, 241, 192, 0, 0, 0, 5, 5, 202, 241, 5, 5, 53, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 243, 192, 0, 0, 0, 5, 5, 158, 243, 5, 5, 155, 243, 5, 5, 157, 243, 5, 5, 156, 243, 5, 5, 153, 243, 5, 5, 159, 243, 5, 5, 162, 243, 5, 5, 160, 243, 5, 5, 161, 243, 5, 5, 77, 217, 192, 0, 0, 0, 5, 5, 76, 217, 5, 5, 152, 243, 5, 5, 163, 243, 5, 5, 15, 245, 5, 5, 18, 245, 5, 5, 20, 245, 5, 5, 19, 245, 192, 0, 0, 0, 5, 5, 13, 245, 5, 5, 17, 245, 5, 5, 12, 245, 192, 0, 0, 0, 5, 5, 14, 245, 5, 5, 247, 217, 5, 5, 11, 245, 192, 0, 0, 0, 5, 5, 16, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 246, 5, 5, 77, 246, 5, 5, 142, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 218, 5, 5, 76, 246, 5, 5, 81, 246, 5, 5, 87, 246, 5, 5, 79, 246, 5, 5, 86, 246, 192, 0, 0, 0, 5, 5, 83, 246, 192, 0, 0, 0, 5, 5, 82, 246, 5, 5, 85, 246, 5, 5, 80, 246, 5, 5, 78, 246, 192, 0, 0, 0, 5, 5, 84, 246, 192, 0, 0, 0, 5, 5, 86, 247, 5, 5, 84, 247, 5, 5, 241, 218, 5, 5, 83, 247, 5, 5, 89, 247, 5, 5, 87, 247, 192, 0, 0, 0, 5, 5, 242, 218, 5, 5, 82, 247, 5, 5, 80, 247, 192, 0, 0, 0, 5, 5, 81, 247, 5, 5, 244, 218, 5, 5, 85, 247, 5, 5, 243, 218, 5, 5, 88, 247, 5, 5, 13, 248, 5, 5, 19, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 248, 5, 5, 67, 219, 5, 5, 20, 248, 5, 5, 66, 219, 5, 5, 18, 248, 5, 5, 17, 248, 5, 5, 65, 219, 5, 5, 16, 248, 5, 5, 64, 219, 5, 5, 15, 248, 192, 0, 0, 0, 5, 5, 175, 248, 5, 5, 177, 248, 5, 5, 176, 248, 192, 0, 0, 0, 5, 5, 173, 248, 5, 5, 129, 219, 5, 5, 170, 248, 5, 5, 174, 248, 5, 5, 179, 248, 192, 0, 0, 0, 5, 5, 178, 248, 5, 5, 171, 248, 5, 5, 172, 248, 5, 5, 128, 219, 5, 5, 35, 249, 5, 5, 173, 219, 192, 0, 0, 0, 5, 5, 36, 249, 5, 5, 171, 219, 5, 5, 172, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 249, 5, 5, 207, 219, 5, 5, 228, 249, 192, 0, 0, 0, 5, 5, 239, 219, 192, 0, 0, 0, 5, 5, 255, 249, 5, 5, 240, 219, 5, 5, 2, 250, 5, 5, 3, 250, 5, 5, 17, 250, 5, 5, 16, 250, 5, 5, 255, 219, 5, 5, 27, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 234, 192, 0, 0, 0, 5, 5, 193, 234, 192, 0, 0, 0, 5, 5, 130, 213, 5, 5, 131, 213, 5, 5, 45, 237, 192, 0, 0, 0, 5, 5, 157, 239, 5, 5, 209, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 215, 5, 5, 54, 216, 5, 5, 207, 241, 192, 0, 0, 0, 5, 5, 210, 241, 5, 5, 55, 216, 192, 0, 0, 0, 5, 5, 164, 243, 5, 5, 208, 241, 5, 5, 21, 245, 5, 5, 248, 217, 5, 5, 89, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 247, 5, 5, 90, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 248, 192, 0, 0, 0, 5, 5, 68, 219, 5, 5, 180, 248, 5, 5, 176, 219, 5, 5, 130, 219, 5, 5, 174, 219, 5, 5, 175, 219, 5, 5, 131, 249, 5, 5, 231, 219, 192, 0, 0, 0, 5, 5, 117, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 226, 192, 0, 0, 0, 5, 5, 2, 212, 192, 0, 0, 0, 5, 5, 46, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 213, 5, 5, 47, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 239, 5, 5, 158, 239, 5, 5, 57, 216, 5, 5, 11, 215, 5, 5, 12, 215, 192, 0, 0, 0, 5, 5, 162, 239, 5, 5, 161, 239, 5, 5, 160, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 241, 5, 5, 214, 241, 192, 0, 0, 0, 5, 5, 213, 241, 5, 5, 212, 241, 5, 5, 56, 216, 5, 5, 168, 243, 5, 5, 166, 243, 5, 5, 165, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 243, 192, 0, 0, 0, 5, 5, 249, 217, 5, 5, 24, 245, 5, 5, 25, 245, 5, 5, 250, 217, 192, 0, 0, 0, 5, 5, 23, 245, 192, 0, 0, 0, 5, 5, 91, 246, 5, 5, 90, 246, 5, 5, 93, 246, 5, 5, 144, 218, 5, 5, 92, 246, 192, 0, 0, 0, 5, 5, 93, 247, 5, 5, 95, 247, 5, 5, 94, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 248, 5, 5, 25, 248, 5, 5, 23, 248, 5, 5, 24, 248, 5, 5, 181, 248, 5, 5, 131, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 249, 5, 5, 38, 249, 5, 5, 39, 249, 192, 0, 0, 0, 5, 5, 208, 219, 5, 5, 232, 219, 5, 5, 4, 250, 5, 5, 118, 206, 192, 0, 0, 0, 5, 5, 13, 215, 5, 5, 58, 216, 5, 5, 26, 245, 192, 0, 0, 0, 5, 5, 182, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 250, 5, 5, 218, 226, 192, 0, 0, 0, 5, 5, 2, 220, 5, 5, 119, 206, 5, 5, 215, 241, 192, 0, 0, 0, 5, 5, 27, 245, 192, 0, 0, 0, 5, 5, 94, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 248, 5, 5, 183, 248, 5, 5, 120, 206, 192, 0, 0, 0, 5, 5, 48, 237, 5, 5, 49, 237, 192, 0, 0, 0, 5, 5, 133, 213, 5, 5, 134, 213, 5, 5, 164, 239, 5, 5, 15, 215, 5, 5, 14, 215, 5, 5, 163, 239, 192, 0, 0, 0, 5, 5, 169, 243, 192, 0, 0, 0, 5, 5, 28, 245, 5, 5, 30, 245, 5, 5, 29, 245, 5, 5, 253, 217, 5, 5, 252, 217, 5, 5, 251, 217, 192, 0, 0, 0, 5, 5, 70, 219, 5, 5, 27, 248, 192, 0, 0, 0, 5, 5, 69, 219, 5, 5, 185, 248, 5, 5, 184, 248, 192, 0, 0, 0, 5, 5, 209, 219, 5, 5, 132, 249, 5, 5, 99, 208, 5, 5, 195, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 237, 5, 5, 50, 237, 5, 5, 51, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 239, 192, 0, 0, 0, 5, 5, 168, 239, 5, 5, 165, 239, 5, 5, 172, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 215, 5, 5, 171, 239, 5, 5, 167, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 239, 5, 5, 170, 239, 5, 5, 169, 239, 5, 5, 16, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 241, 5, 5, 221, 241, 5, 5, 219, 241, 5, 5, 228, 241, 5, 5, 220, 241, 192, 0, 0, 0, 5, 5, 216, 241, 192, 0, 0, 0, 5, 5, 222, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 241, 5, 5, 178, 243, 5, 5, 218, 241, 5, 5, 229, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 241, 5, 5, 59, 216, 5, 5, 224, 241, 5, 5, 223, 241, 192, 0, 0, 0, 5, 5, 227, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 243, 5, 5, 173, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 243, 192, 0, 0, 0, 5, 5, 180, 243, 5, 5, 175, 243, 5, 5, 179, 243, 192, 0, 0, 0, 5, 5, 177, 243, 5, 5, 176, 243, 5, 5, 174, 243, 192, 0, 0, 0, 5, 5, 171, 243, 192, 0, 0, 0, 5, 5, 80, 217, 5, 5, 79, 217, 192, 0, 0, 0, 5, 5, 81, 217, 5, 5, 78, 217, 5, 5, 181, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 245, 5, 5, 39, 245, 192, 0, 0, 0, 5, 5, 37, 245, 5, 5, 41, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 245, 192, 0, 0, 0, 5, 5, 34, 245, 5, 5, 4, 218, 5, 5, 35, 245, 192, 0, 0, 0, 5, 5, 33, 245, 5, 5, 40, 245, 192, 0, 0, 0, 5, 5, 32, 245, 5, 5, 31, 245, 5, 5, 3, 218, 5, 5, 255, 217, 5, 5, 254, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 245, 5, 5, 103, 246, 5, 5, 109, 246, 5, 5, 147, 218, 5, 5, 104, 246, 192, 0, 0, 0, 5, 5, 107, 246, 5, 5, 111, 246, 5, 5, 148, 218, 5, 5, 106, 246, 192, 0, 0, 0, 5, 5, 98, 246, 192, 0, 0, 0, 5, 5, 97, 246, 5, 5, 110, 246, 5, 5, 101, 246, 192, 0, 0, 0, 5, 5, 99, 246, 5, 5, 108, 246, 5, 5, 100, 246, 5, 5, 146, 218, 5, 5, 145, 218, 192, 0, 0, 0, 5, 5, 95, 246, 5, 5, 96, 246, 5, 5, 105, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 247, 5, 5, 101, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 247, 5, 5, 105, 247, 5, 5, 103, 247, 5, 5, 97, 247, 5, 5, 108, 247, 192, 0, 0, 0, 5, 5, 96, 247, 192, 0, 0, 0, 5, 5, 246, 218, 5, 5, 104, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 247, 5, 5, 245, 218, 5, 5, 107, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 248, 5, 5, 29, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 248, 192, 0, 0, 0, 5, 5, 32, 248, 5, 5, 35, 248, 5, 5, 72, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 248, 5, 5, 34, 248, 192, 0, 0, 0, 5, 5, 28, 248, 5, 5, 31, 248, 5, 5, 71, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 219, 5, 5, 196, 248, 5, 5, 191, 248, 5, 5, 195, 248, 192, 0, 0, 0, 5, 5, 198, 248, 5, 5, 194, 248, 192, 0, 0, 0, 5, 5, 190, 248, 192, 0, 0, 0, 5, 5, 135, 219, 5, 5, 193, 248, 5, 5, 197, 248, 5, 5, 134, 219, 5, 5, 188, 248, 5, 5, 102, 247, 5, 5, 192, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 248, 192, 0, 0, 0, 5, 5, 186, 248, 192, 0, 0, 0, 5, 5, 187, 248, 5, 5, 132, 219, 5, 5, 43, 249, 5, 5, 45, 249, 5, 5, 48, 249, 5, 5, 44, 249, 5, 5, 49, 249, 192, 0, 0, 0, 5, 5, 42, 249, 192, 0, 0, 0, 5, 5, 40, 249, 192, 0, 0, 0, 5, 5, 177, 219, 5, 5, 46, 249, 5, 5, 179, 219, 5, 5, 178, 219, 5, 5, 41, 249, 5, 5, 47, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 249, 5, 5, 210, 219, 5, 5, 138, 249, 192, 0, 0, 0, 5, 5, 136, 249, 5, 5, 133, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 249, 5, 5, 134, 249, 5, 5, 200, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 249, 5, 5, 201, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 249, 192, 0, 0, 0, 5, 5, 230, 249, 5, 5, 229, 249, 5, 5, 232, 249, 192, 0, 0, 0, 5, 5, 248, 219, 5, 5, 249, 219, 5, 5, 24, 250, 5, 5, 28, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 208, 5, 5, 17, 232, 5, 5, 198, 234, 192, 0, 0, 0, 5, 5, 3, 212, 5, 5, 196, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 237, 5, 5, 54, 237, 5, 5, 137, 213, 5, 5, 135, 213, 5, 5, 55, 237, 5, 5, 136, 213, 5, 5, 179, 239, 192, 0, 0, 0, 5, 5, 182, 239, 5, 5, 176, 239, 5, 5, 183, 239, 5, 5, 175, 239, 5, 5, 177, 239, 192, 0, 0, 0, 5, 5, 178, 239, 5, 5, 181, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 215, 5, 5, 186, 239, 5, 5, 185, 239, 5, 5, 18, 215, 5, 5, 180, 239, 5, 5, 184, 239, 5, 5, 19, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 241, 192, 0, 0, 0, 5, 5, 64, 216, 192, 0, 0, 0, 5, 5, 234, 241, 5, 5, 60, 216, 192, 0, 0, 0, 5, 5, 231, 241, 5, 5, 237, 241, 5, 5, 240, 241, 192, 0, 0, 0, 5, 5, 65, 216, 192, 0, 0, 0, 5, 5, 236, 241, 5, 5, 233, 241, 5, 5, 241, 241, 5, 5, 232, 241, 192, 0, 0, 0, 5, 5, 238, 241, 5, 5, 61, 216, 192, 0, 0, 0, 5, 5, 230, 241, 5, 5, 62, 216, 192, 0, 0, 0, 5, 5, 63, 216, 5, 5, 235, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 243, 5, 5, 186, 243, 5, 5, 187, 243, 5, 5, 190, 243, 5, 5, 188, 243, 192, 0, 0, 0, 5, 5, 182, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 243, 5, 5, 195, 243, 5, 5, 189, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 217, 192, 0, 0, 0, 5, 5, 197, 243, 5, 5, 194, 243, 5, 5, 83, 217, 5, 5, 196, 243, 5, 5, 183, 243, 5, 5, 192, 243, 5, 5, 193, 243, 192, 0, 0, 0, 5, 5, 191, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 245, 5, 5, 48, 245, 5, 5, 51, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 245, 192, 0, 0, 0, 5, 5, 5, 218, 5, 5, 53, 245, 5, 5, 44, 245, 5, 5, 54, 245, 192, 0, 0, 0, 5, 5, 50, 245, 5, 5, 52, 245, 5, 5, 56, 245, 5, 5, 49, 245, 5, 5, 57, 245, 5, 5, 47, 245, 5, 5, 43, 245, 5, 5, 6, 218, 192, 0, 0, 0, 5, 5, 55, 245, 5, 5, 7, 218, 5, 5, 150, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 243, 5, 5, 137, 246, 5, 5, 128, 246, 5, 5, 152, 218, 5, 5, 124, 246, 5, 5, 153, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 246, 5, 5, 127, 246, 5, 5, 117, 246, 5, 5, 151, 218, 5, 5, 130, 246, 5, 5, 125, 246, 5, 5, 126, 246, 192, 0, 0, 0, 5, 5, 112, 246, 5, 5, 119, 246, 5, 5, 134, 246, 192, 0, 0, 0, 5, 5, 131, 246, 192, 0, 0, 0, 5, 5, 123, 246, 192, 0, 0, 0, 5, 5, 135, 246, 5, 5, 118, 246, 5, 5, 113, 246, 5, 5, 132, 246, 192, 0, 0, 0, 5, 5, 115, 246, 5, 5, 129, 246, 5, 5, 120, 246, 5, 5, 136, 246, 5, 5, 116, 246, 5, 5, 149, 218, 5, 5, 114, 246, 5, 5, 121, 246, 5, 5, 122, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 247, 192, 0, 0, 0, 5, 5, 113, 247, 192, 0, 0, 0, 5, 5, 118, 247, 192, 0, 0, 0, 5, 5, 130, 247, 5, 5, 121, 247, 5, 5, 114, 247, 5, 5, 110, 247, 5, 5, 123, 247, 5, 5, 116, 247, 5, 5, 119, 247, 5, 5, 112, 247, 5, 5, 126, 247, 5, 5, 109, 247, 5, 5, 117, 247, 5, 5, 122, 247, 5, 5, 124, 247, 5, 5, 127, 247, 5, 5, 111, 247, 192, 0, 0, 0, 5, 5, 131, 247, 5, 5, 132, 247, 5, 5, 125, 247, 5, 5, 129, 247, 5, 5, 120, 247, 192, 0, 0, 0, 5, 5, 52, 248, 5, 5, 55, 248, 192, 0, 0, 0, 5, 5, 73, 219, 192, 0, 0, 0, 5, 5, 54, 248, 5, 5, 59, 248, 5, 5, 57, 248, 5, 5, 74, 219, 5, 5, 48, 248, 5, 5, 38, 248, 5, 5, 37, 248, 5, 5, 76, 219, 5, 5, 50, 248, 5, 5, 51, 248, 5, 5, 47, 248, 5, 5, 39, 248, 192, 0, 0, 0, 5, 5, 44, 248, 5, 5, 128, 247, 192, 0, 0, 0, 5, 5, 40, 248, 5, 5, 75, 219, 5, 5, 46, 248, 192, 0, 0, 0, 5, 5, 45, 248, 192, 0, 0, 0, 5, 5, 41, 248, 5, 5, 53, 248, 192, 0, 0, 0, 5, 5, 42, 248, 5, 5, 203, 248, 5, 5, 56, 248, 5, 5, 58, 248, 5, 5, 49, 248, 5, 5, 43, 248, 5, 5, 204, 248, 5, 5, 206, 248, 5, 5, 200, 248, 5, 5, 136, 219, 192, 0, 0, 0, 5, 5, 213, 248, 5, 5, 211, 248, 5, 5, 137, 219, 5, 5, 210, 248, 5, 5, 209, 248, 5, 5, 202, 248, 5, 5, 199, 248, 5, 5, 205, 248, 5, 5, 214, 248, 5, 5, 201, 248, 5, 5, 207, 248, 192, 0, 0, 0, 5, 5, 58, 249, 5, 5, 63, 249, 5, 5, 53, 249, 5, 5, 56, 249, 5, 5, 180, 219, 5, 5, 60, 249, 192, 0, 0, 0, 5, 5, 67, 249, 5, 5, 208, 248, 192, 0, 0, 0, 5, 5, 54, 249, 5, 5, 64, 249, 5, 5, 68, 249, 5, 5, 59, 249, 5, 5, 52, 249, 5, 5, 62, 249, 192, 0, 0, 0, 5, 5, 61, 249, 5, 5, 66, 249, 5, 5, 65, 249, 5, 5, 212, 248, 5, 5, 57, 249, 5, 5, 51, 249, 5, 5, 55, 249, 5, 5, 211, 219, 5, 5, 212, 219, 5, 5, 50, 249, 192, 0, 0, 0, 5, 5, 149, 249, 5, 5, 140, 249, 5, 5, 148, 249, 5, 5, 145, 249, 5, 5, 146, 249, 5, 5, 139, 249, 5, 5, 142, 249, 5, 5, 150, 249, 5, 5, 144, 249, 5, 5, 143, 249, 5, 5, 141, 249, 192, 0, 0, 0, 5, 5, 147, 249, 192, 0, 0, 0, 5, 5, 203, 249, 192, 0, 0, 0, 5, 5, 204, 249, 192, 0, 0, 0, 5, 5, 206, 249, 5, 5, 205, 249, 5, 5, 208, 249, 5, 5, 207, 249, 5, 5, 234, 249, 5, 5, 233, 249, 5, 5, 5, 250, 192, 0, 0, 0, 5, 5, 6, 250, 192, 0, 0, 0, 5, 5, 19, 250, 5, 5, 253, 219, 5, 5, 3, 220, 192, 0, 0, 0, 5, 5, 29, 250, 5, 5, 4, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 218, 5, 5, 60, 248, 192, 0, 0, 0, 5, 5, 213, 219, 5, 5, 214, 219, 192, 0, 0, 0, 5, 5, 102, 208, 5, 5, 199, 234, 192, 0, 0, 0, 5, 5, 4, 212, 5, 5, 187, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 241, 5, 5, 244, 241, 5, 5, 242, 241, 5, 5, 201, 243, 5, 5, 200, 243, 5, 5, 84, 217, 5, 5, 59, 245, 5, 5, 202, 243, 5, 5, 58, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 246, 5, 5, 154, 218, 5, 5, 156, 218, 5, 5, 138, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 218, 192, 0, 0, 0, 5, 5, 133, 247, 5, 5, 135, 247, 5, 5, 134, 247, 5, 5, 61, 248, 5, 5, 77, 219, 192, 0, 0, 0, 5, 5, 181, 219, 5, 5, 151, 249, 5, 5, 209, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 250, 5, 5, 103, 208, 192, 0, 0, 0, 5, 5, 56, 237, 192, 0, 0, 0, 5, 5, 21, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 241, 5, 5, 245, 241, 192, 0, 0, 0, 5, 5, 203, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 218, 5, 5, 248, 218, 5, 5, 215, 248, 5, 5, 25, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 208, 5, 5, 138, 213, 192, 0, 0, 0, 5, 5, 22, 215, 192, 0, 0, 0, 5, 5, 140, 246, 192, 0, 0, 0, 5, 5, 69, 249, 5, 5, 65, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 219, 5, 5, 66, 210, 5, 5, 23, 215, 5, 5, 85, 217, 5, 5, 70, 249, 5, 5, 67, 210, 192, 0, 0, 0, 5, 5, 188, 239, 5, 5, 67, 216, 5, 5, 247, 241, 5, 5, 248, 241, 192, 0, 0, 0, 5, 5, 66, 216, 192, 0, 0, 0, 5, 5, 205, 243, 5, 5, 89, 217, 5, 5, 87, 217, 5, 5, 88, 217, 5, 5, 86, 217, 5, 5, 60, 245, 5, 5, 8, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 247, 5, 5, 136, 247, 5, 5, 139, 247, 5, 5, 138, 247, 5, 5, 249, 218, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 248, 192, 0, 0, 0, 5, 5, 64, 248, 5, 5, 63, 248, 5, 5, 78, 219, 5, 5, 216, 248, 192, 0, 0, 0, 5, 5, 71, 249, 5, 5, 72, 249, 5, 5, 182, 219, 5, 5, 210, 249, 5, 5, 235, 249, 5, 5, 250, 219, 192, 0, 0, 0, 5, 5, 18, 232, 5, 5, 249, 241, 5, 5, 206, 243, 5, 5, 141, 246, 5, 5, 200, 234, 192, 0, 0, 0, 5, 5, 207, 243, 5, 5, 62, 245, 5, 5, 61, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 249, 5, 5, 215, 219, 192, 0, 0, 0, 5, 5, 211, 249, 5, 5, 236, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 212, 5, 5, 189, 239, 5, 5, 190, 239, 192, 0, 0, 0, 5, 5, 250, 241, 5, 5, 6, 212, 192, 0, 0, 0, 5, 5, 9, 218, 5, 5, 63, 245, 192, 0, 0, 0, 5, 5, 66, 248, 5, 5, 79, 219, 5, 5, 67, 248, 5, 5, 65, 248, 5, 5, 74, 249, 192, 0, 0, 0, 5, 5, 152, 249, 192, 0, 0, 0, 5, 5, 7, 212, 192, 0, 0, 0, 5, 5, 210, 243, 5, 5, 209, 243, 5, 5, 208, 243, 5, 5, 64, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 245, 5, 5, 67, 245, 5, 5, 66, 245, 5, 5, 65, 245, 5, 5, 10, 218, 5, 5, 142, 246, 5, 5, 141, 247, 5, 5, 250, 218, 5, 5, 140, 247, 5, 5, 68, 248, 5, 5, 219, 248, 5, 5, 218, 248, 5, 5, 138, 219, 5, 5, 217, 248, 5, 5, 77, 249, 5, 5, 76, 249, 5, 5, 75, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 213, 192, 0, 0, 0, 5, 5, 251, 241, 5, 5, 90, 217, 192, 0, 0, 0, 5, 5, 143, 246, 5, 5, 144, 246, 5, 5, 220, 248, 5, 5, 78, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 249, 5, 5, 212, 249, 5, 5, 7, 250, 5, 5, 34, 250, 5, 5, 140, 213, 5, 5, 91, 217, 5, 5, 69, 245, 5, 5, 145, 246, 5, 5, 69, 248, 5, 5, 79, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 215, 192, 0, 0, 0, 5, 5, 211, 243, 5, 5, 70, 245, 5, 5, 146, 246, 5, 5, 147, 246, 5, 5, 148, 246, 5, 5, 146, 247, 192, 0, 0, 0, 5, 5, 142, 247, 5, 5, 80, 219, 5, 5, 145, 247, 5, 5, 144, 247, 5, 5, 251, 218, 5, 5, 143, 247, 5, 5, 253, 218, 192, 0, 0, 0, 5, 5, 252, 218, 5, 5, 71, 248, 5, 5, 70, 248, 5, 5, 81, 219, 5, 5, 82, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 219, 5, 5, 221, 248, 5, 5, 139, 219, 192, 0, 0, 0, 5, 5, 82, 249, 5, 5, 83, 249, 5, 5, 81, 249, 5, 5, 80, 249, 5, 5, 217, 219, 192, 0, 0, 0, 5, 5, 154, 249, 5, 5, 155, 249, 5, 5, 156, 249, 5, 5, 216, 219, 5, 5, 213, 249, 5, 5, 216, 249, 5, 5, 215, 249, 5, 5, 214, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 218, 5, 5, 147, 247, 5, 5, 72, 248, 192, 0, 0, 0, 5, 5, 141, 219, 5, 5, 222, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 243, 192, 0, 0, 0, 5, 5, 223, 248, 192, 0, 0, 0, 5, 5, 237, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 206, 5, 5, 47, 201, 5, 5, 135, 201, 5, 5, 168, 211, 5, 5, 228, 210, 5, 5, 135, 200, 5, 5, 153, 199, 5, 5, 69, 216, 5, 5, 69, 216, 5, 5, 104, 203, 5, 5, 14, 203, 5, 5, 136, 208, 5, 5, 248, 201, 5, 5, 21, 218, 5, 5, 23, 219, 5, 5, 67, 218, 5, 5, 160, 219, 5, 5, 255, 216, 5, 5, 44, 213, 5, 5, 164, 219, 5, 5, 13, 214, 5, 5, 246, 203, 5, 5, 157, 205, 5, 5, 173, 205, 5, 5, 101, 211, 5, 5, 208, 211, 5, 5, 53, 216, 5, 5, 68, 210, 5, 5, 182, 200, 5, 5, 15, 219, 5, 5, 19, 219, 5, 5, 30, 219, 5, 5, 4, 220, 5, 5, 181, 208, 5, 5, 140, 216, 5, 5, 176, 217, 5, 5, 215, 218, 5, 5, 76, 202, 5, 5, 72, 218, 5, 5, 35, 219, 5, 5, 190, 208, 5, 5, 99, 205, 5, 5, 133, 205, 5, 5, 162, 205, 5, 5, 150, 204, 5, 5, 172, 201, 5, 5, 170, 200, 5, 5, 120, 208, 5, 5, 78, 215, 5, 5, 30, 218, 5, 5, 181, 218, 5, 5, 149, 215, 5, 5, 110, 200, 5, 5, 209, 218, 5, 5, 117, 211, 5, 5, 177, 211, 5, 5, 59, 219, 5, 5, 17, 215, 5, 5, 212, 219, 5, 5, 43, 211, 5, 5, 50, 211, 5, 5, 234, 212, 5, 5, 92, 231, 5, 5, 15, 216, 5, 5, 102, 208, 5, 5, 171, 214, 5, 5, 14, 218, 5, 5, 7, 201, 5, 5, 104, 219, 5, 5, 106, 219, 5, 5, 94, 201, 5, 5, 75, 214, 5, 5, 172, 211, 5, 5, 236, 211, 5, 5, 96, 217, 5, 5, 76, 212, 5, 5, 10, 214, 5, 5, 133, 207, 5, 5, 151, 212, 5, 5, 221, 207, 5, 5, 202, 216, 5, 5, 158, 204, 5, 5, 146, 206, 5, 5, 117, 200, 5, 5, 150, 213, 5, 5, 210, 204, 5, 5, 54, 211, 5, 5, 233, 212, 5, 5, 187, 209, 5, 5, 84, 208, 5, 5, 112, 219, 5, 5, 142, 223, 5, 5, 13, 214, 5, 5, 229, 215, 5, 5, 7, 199, 5, 5, 67, 212, 5, 5, 151, 203, 5, 5, 159, 207, 5, 5, 174, 207, 5, 5, 127, 199, 5, 5, 138, 242, 5, 5, 33, 203, 5, 5, 196, 208, 5, 5, 4, 199, 5, 5, 154, 202, 5, 5, 251, 213, 5, 5, 237, 205, 5, 5, 156, 206, 5, 5, 108, 210, 5, 5, 44, 204, 5, 5, 106, 211, 5, 5, 57, 213, 5, 5, 104, 207, 5, 5, 137, 201, 5, 5, 67, 201, 5, 5, 181, 203, 5, 5, 107, 204, 5, 5, 26, 207, 5, 5, 171, 207, 5, 5, 29, 203, 5, 5, 187, 201, 192, 0, 0, 0, 5, 5, 81, 207, 5, 5, 158, 217, 5, 5, 120, 201, 5, 5, 160, 214, 5, 5, 20, 210, 5, 5, 74, 216, 5, 5, 194, 200, 5, 5, 0, 179, 5, 5, 18, 218, 5, 5, 87, 205, 5, 5, 124, 217, 5, 5, 187, 218, 5, 5, 238, 214, 5, 5, 255, 219, 5, 5, 155, 218, 5, 5, 23, 215, 5, 5, 0, 151, 5, 5, 87, 215, 5, 5, 109, 215, 5, 5, 151, 248, 5, 5, 61, 200, 5, 5, 219, 213, 5, 5, 144, 219, 5, 5, 244, 213, 5, 5, 160, 212, 5, 5, 240, 210, 5, 5, 50, 238, 192, 0, 0, 0, 5, 5, 99, 214, 5, 5, 222, 216, 5, 5, 207, 214, 5, 5, 128, 214, 5, 5, 59, 208, 5, 5, 51, 217, 5, 5, 2, 200, 5, 5, 7, 200, 5, 5, 82, 203, 5, 5, 158, 205, 5, 5, 215, 209, 5, 5, 57, 213, 5, 5, 135, 210, 5, 5, 56, 202, 5, 5, 51, 207, 5, 5, 131, 216, 5, 5, 56, 218, 5, 5, 134, 217, 5, 5, 111, 199, 5, 5, 208, 221, 5, 5, 67, 212, 5, 5, 92, 216, 5, 5, 46, 222, 5, 5, 21, 204, 5, 5, 51, 214, 5, 5, 227, 207, 5, 5, 233, 207, 5, 5, 221, 211, 5, 5, 239, 211, 5, 5, 203, 219, 5, 5, 122, 213, 5, 5, 171, 201, 5, 5, 146, 217, 5, 5, 230, 218, 5, 5, 67, 217, 5, 5, 200, 208, 5, 5, 0, 142, 5, 5, 13, 212, 5, 5, 187, 213, 5, 5, 249, 200, 5, 5, 85, 205, 5, 5, 13, 214, 5, 5, 132, 215, 5, 5, 167, 216, 5, 5, 178, 238, 5, 5, 3, 216, 5, 5, 68, 216, 5, 5, 183, 210, 5, 5, 152, 201, 5, 5, 153, 213, 5, 5, 86, 222, 5, 5, 219, 203, 5, 5, 236, 203, 5, 5, 230, 210, 5, 5, 169, 205, 5, 5, 178, 205, 5, 5, 191, 207, 5, 5, 239, 205, 5, 5, 136, 218, 5, 5, 28, 199, 5, 5, 227, 213, 5, 5, 86, 208, 5, 5, 203, 204, 5, 5, 229, 206, 5, 5, 136, 207, 5, 5, 208, 214, 5, 5, 147, 203, 5, 5, 152, 210, 5, 5, 110, 205, 5, 5, 159, 207, 5, 5, 42, 210, 5, 5, 173, 200, 5, 5, 14, 200, 5, 5, 192, 213, 5, 5, 110, 202, 5, 5, 49, 201, 5, 5, 99, 207, 5, 5, 155, 202, 5, 5, 163, 207, 5, 5, 107, 209, 5, 5, 185, 215, 192, 0, 0, 0, 5, 5, 137, 211, 5, 5, 150, 201, 5, 5, 226, 217, 5, 5, 153, 206, 5, 5, 226, 210, 5, 5, 183, 200, 5, 5, 127, 215, 5, 5, 140, 215, 5, 5, 208, 218, 192, 0, 0, 0, 5, 5, 178, 219, 5, 5, 181, 219, 5, 5, 131, 202, 5, 5, 120, 207, 5, 5, 233, 216, 5, 5, 221, 199, 5, 5, 200, 207, 5, 5, 208, 207, 5, 5, 193, 202, 5, 5, 186, 202, 5, 5, 91, 218, 5, 5, 17, 199, 5, 5, 39, 206, 5, 5, 196, 201, 5, 5, 34, 199, 5, 5, 139, 203, 5, 5, 85, 202, 5, 5, 175, 215, 5, 5, 53, 200, 5, 5, 240, 203, 5, 5, 254, 213, 5, 5, 247, 215, 5, 5, 131, 200, 5, 5, 160, 204, 5, 5, 124, 201, 5, 5, 84, 212, 5, 5, 0, 169, 5, 5, 51, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 210, 192, 0, 0, 0, 5, 5, 251, 208, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 239, 192, 0, 0, 0, 5, 5, 191, 205, 192, 0, 0, 0, 5, 5, 215, 205, 5, 5, 194, 207, 5, 5, 207, 212, 5, 5, 240, 211, 5, 5, 228, 212, 5, 5, 109, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 210, 5, 5, 73, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 210, 5, 5, 249, 211, 5, 5, 46, 216, 5, 5, 74, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 203, 5, 5, 8, 212, 5, 5, 167, 200, 5, 5, 64, 203, 5, 5, 85, 210, 5, 5, 206, 201, 5, 5, 130, 208, 5, 5, 32, 212, 5, 5, 44, 215, 192, 0, 0, 0, 5, 5, 24, 215, 5, 5, 191, 213, 5, 5, 16, 220, 5, 5, 59, 205, 5, 5, 211, 208, 5, 5, 221, 213, 5, 5, 19, 218, 5, 5, 61, 207, 5, 5, 189, 203, 5, 5, 254, 208, 5, 5, 96, 207, 5, 5, 139, 205, 5, 5, 131, 207, 5, 5, 153, 212, 5, 5, 82, 209, 192, 0, 0, 0, 5, 5, 94, 209, 5, 5, 46, 211, 5, 5, 213, 202, 5, 5, 58, 204, 5, 5, 59, 204, 5, 5, 211, 205, 5, 5, 214, 205, 5, 5, 216, 205, 5, 5, 208, 212, 5, 5, 206, 212, 5, 5, 82, 214, 5, 5, 67, 204, 5, 5, 63, 211, 5, 5, 99, 214, 5, 5, 183, 215, 5, 5, 209, 216, 5, 5, 79, 211, 5, 5, 225, 202, 5, 5, 18, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 209, 5, 5, 153, 214, 5, 5, 218, 209, 5, 5, 230, 215, 5, 5, 191, 217, 5, 5, 67, 213, 5, 5, 98, 218, 192, 0, 0, 0, 5, 5, 11, 210, 5, 5, 130, 218, 5, 5, 60, 219, 5, 5, 40, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 201, 192, 0, 0, 0, 5, 5, 252, 199, 5, 5, 149, 221, 5, 5, 114, 199, 5, 5, 31, 215, 5, 5, 63, 203, 5, 5, 0, 171, 5, 5, 130, 208, 5, 5, 169, 206, 5, 5, 151, 208, 5, 5, 60, 232, 5, 5, 112, 210, 5, 5, 175, 213, 5, 5, 249, 201, 5, 5, 250, 201, 5, 5, 203, 206, 5, 5, 227, 239, 5, 5, 52, 235, 5, 5, 51, 235, 5, 5, 247, 206, 5, 5, 133, 232, 5, 5, 16, 207, 5, 5, 150, 210, 5, 5, 149, 210, 5, 5, 221, 213, 192, 0, 0, 0, 5, 5, 19, 218, 5, 5, 104, 216, 5, 5, 16, 230, 5, 5, 170, 210, 5, 5, 239, 208, 5, 5, 56, 207, 5, 5, 251, 208, 5, 5, 99, 205, 5, 5, 80, 207, 5, 5, 54, 201, 5, 5, 80, 199, 5, 5, 104, 207, 5, 5, 236, 203, 192, 0, 0, 0, 5, 5, 67, 209, 5, 5, 153, 212, 192, 0, 0, 0, 5, 5, 82, 209, 5, 5, 174, 216, 5, 5, 158, 216, 5, 5, 197, 199, 192, 0, 0, 0, 5, 5, 229, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 238, 5, 5, 59, 214, 5, 5, 191, 205, 5, 5, 184, 207, 5, 5, 210, 202, 5, 5, 227, 230, 192, 0, 0, 0, 5, 5, 82, 238, 5, 5, 188, 240, 5, 5, 63, 211, 192, 0, 0, 0, 5, 5, 75, 211, 5, 5, 99, 214, 5, 5, 34, 231, 5, 5, 225, 202, 5, 5, 28, 206, 5, 5, 186, 209, 5, 5, 44, 241, 5, 5, 24, 243, 5, 5, 187, 217, 5, 5, 218, 209, 5, 5, 169, 214, 5, 5, 165, 214, 5, 5, 164, 214, 5, 5, 230, 215, 5, 5, 229, 215, 5, 5, 233, 215, 5, 5, 191, 217, 5, 5, 162, 219, 5, 5, 98, 218, 5, 5, 249, 215, 5, 5, 2, 216, 5, 5, 73, 243, 5, 5, 235, 236, 5, 5, 81, 229, 5, 5, 130, 218, 5, 5, 240, 211, 192, 0, 0, 0, 5, 5, 60, 219, 192, 0, 0, 0, 5, 5, 40, 216, 5, 5, 94, 247, 5, 5, 69, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 161, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 200, 5, 5, 44, 203, 5, 5, 4, 223, 192, 0, 0, 0, 5, 5, 122, 206, 5, 5, 108, 208, 5, 5, 8, 212, 5, 5, 15, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 200, 5, 5, 184, 201, 192, 0, 0, 0, 5, 5, 188, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 199, 5, 5, 254, 199, 192, 0, 0, 0, 5, 5, 31, 199, 5, 5, 206, 204, 192, 0, 0, 0, 5, 5, 118, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 220, 5, 5, 0, 170, 192, 0, 0, 0, 5, 5, 193, 201, 5, 5, 178, 221, 5, 5, 116, 208, 5, 5, 81, 210, 192, 0, 0, 0, 5, 5, 63, 203, 5, 5, 64, 203, 5, 5, 85, 210, 5, 5, 0, 171, 5, 5, 125, 199, 5, 5, 126, 199, 5, 5, 127, 199, 5, 5, 131, 199, 5, 5, 206, 201, 5, 5, 123, 208, 5, 5, 181, 200, 192, 0, 0, 0, 5, 5, 219, 204, 5, 5, 219, 204, 5, 5, 219, 204, 192, 0, 0, 0, 5, 5, 100, 200, 5, 5, 47, 199, 5, 5, 222, 204, 192, 0, 0, 0, 5, 5, 147, 199, 5, 5, 151, 199, 5, 5, 26, 200, 192, 0, 0, 0, 5, 5, 200, 200, 5, 5, 193, 200, 5, 5, 232, 201, 5, 5, 39, 223, 192, 0, 0, 0, 5, 5, 224, 204, 192, 0, 0, 0, 5, 5, 176, 206, 5, 5, 159, 209, 5, 5, 159, 209, 5, 5, 151, 208, 5, 5, 150, 208, 5, 5, 139, 208, 5, 5, 66, 232, 5, 5, 43, 212, 5, 5, 32, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 213, 5, 5, 34, 199, 192, 0, 0, 0, 5, 5, 100, 203, 5, 5, 34, 227, 5, 5, 59, 227, 5, 5, 96, 203, 5, 5, 2, 225, 5, 5, 158, 208, 5, 5, 98, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 220, 5, 5, 39, 200, 5, 5, 54, 212, 5, 5, 195, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 205, 5, 5, 5, 205, 192, 0, 0, 0, 5, 5, 109, 203, 5, 5, 199, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 213, 5, 5, 87, 245, 5, 5, 87, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 232, 5, 5, 67, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 185, 192, 0, 0, 0, 5, 5, 217, 206, 5, 5, 16, 220, 192, 0, 0, 0, 5, 5, 2, 221, 192, 0, 0, 0, 5, 5, 194, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 229, 5, 5, 38, 235, 5, 5, 142, 201, 5, 5, 233, 206, 192, 0, 0, 0, 5, 5, 184, 208, 5, 5, 43, 225, 5, 5, 186, 208, 5, 5, 128, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 223, 5, 5, 122, 227, 5, 5, 240, 206, 5, 5, 190, 208, 192, 0, 0, 0, 5, 5, 0, 195, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 226, 5, 5, 17, 222, 5, 5, 17, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 201, 5, 5, 248, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 201, 5, 5, 15, 201, 192, 0, 0, 0, 5, 5, 59, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 205, 192, 0, 0, 0, 5, 5, 22, 207, 5, 5, 143, 210, 5, 5, 151, 210, 5, 5, 150, 210, 5, 5, 151, 210, 5, 5, 71, 235, 5, 5, 221, 213, 5, 5, 62, 215, 5, 5, 224, 213, 5, 5, 147, 237, 5, 5, 44, 242, 5, 5, 19, 218, 5, 5, 21, 218, 5, 5, 69, 200, 5, 5, 24, 207, 192, 0, 0, 0, 5, 5, 98, 202, 5, 5, 86, 202, 5, 5, 76, 205, 192, 0, 0, 0, 5, 5, 77, 205, 5, 5, 171, 203, 5, 5, 53, 207, 5, 5, 39, 207, 5, 5, 78, 225, 192, 0, 0, 0, 5, 5, 167, 232, 5, 5, 152, 232, 5, 5, 37, 207, 192, 0, 0, 0, 5, 5, 228, 213, 192, 0, 0, 0, 5, 5, 3, 230, 5, 5, 107, 235, 192, 0, 0, 0, 5, 5, 61, 207, 5, 5, 177, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 205, 5, 5, 89, 205, 192, 0, 0, 0, 5, 5, 254, 208, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 203, 5, 5, 142, 206, 5, 5, 114, 208, 192, 0, 0, 0, 5, 5, 203, 222, 192, 0, 0, 0, 5, 5, 99, 205, 5, 5, 80, 207, 5, 5, 214, 233, 5, 5, 55, 201, 5, 5, 59, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 223, 5, 5, 81, 222, 192, 0, 0, 0, 5, 5, 96, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 225, 5, 5, 75, 230, 192, 0, 0, 0, 5, 5, 210, 232, 5, 5, 138, 212, 192, 0, 0, 0, 5, 5, 70, 242, 192, 0, 0, 0, 5, 5, 124, 216, 192, 0, 0, 0, 5, 5, 85, 200, 192, 0, 0, 0, 5, 5, 73, 240, 192, 0, 0, 0, 5, 5, 209, 210, 5, 5, 158, 235, 5, 5, 104, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 200, 192, 0, 0, 0, 5, 5, 172, 202, 5, 5, 134, 222, 5, 5, 77, 221, 5, 5, 184, 225, 5, 5, 244, 203, 5, 5, 139, 205, 5, 5, 236, 203, 5, 5, 147, 205, 5, 5, 138, 205, 5, 5, 151, 205, 192, 0, 0, 0, 5, 5, 231, 223, 5, 5, 41, 209, 5, 5, 56, 209, 192, 0, 0, 0, 5, 5, 67, 209, 5, 5, 221, 210, 192, 0, 0, 0, 5, 5, 125, 207, 5, 5, 32, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 237, 5, 5, 188, 246, 192, 0, 0, 0, 5, 5, 31, 218, 192, 0, 0, 0, 5, 5, 169, 247, 5, 5, 92, 201, 192, 0, 0, 0, 5, 5, 7, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 238, 192, 0, 0, 0, 5, 5, 254, 219, 5, 5, 158, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 209, 5, 5, 221, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 218, 5, 5, 94, 199, 192, 0, 0, 0, 5, 5, 202, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 211, 5, 5, 229, 235, 5, 5, 55, 238, 5, 5, 45, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 222, 192, 0, 0, 0, 5, 5, 174, 207, 192, 0, 0, 0, 5, 5, 108, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 205, 5, 5, 195, 205, 5, 5, 227, 230, 192, 0, 0, 0, 5, 5, 71, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 207, 5, 5, 43, 211, 5, 5, 82, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 212, 5, 5, 33, 226, 192, 0, 0, 0, 5, 5, 82, 214, 5, 5, 27, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 214, 5, 5, 168, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 240, 192, 0, 0, 0, 5, 5, 177, 242, 192, 0, 0, 0, 5, 5, 73, 204, 192, 0, 0, 0, 5, 5, 19, 231, 192, 0, 0, 0, 5, 5, 246, 212, 192, 0, 0, 0, 5, 5, 208, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 240, 192, 0, 0, 0, 5, 5, 158, 228, 192, 0, 0, 0, 5, 5, 225, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 209, 5, 5, 91, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 226, 5, 5, 60, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 221, 5, 5, 122, 201, 5, 5, 240, 202, 192, 0, 0, 0, 5, 5, 246, 202, 5, 5, 239, 202, 5, 5, 243, 202, 5, 5, 105, 204, 192, 0, 0, 0, 5, 5, 107, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 207, 5, 5, 72, 231, 5, 5, 189, 209, 5, 5, 130, 226, 5, 5, 196, 209, 5, 5, 193, 209, 5, 5, 200, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 233, 5, 5, 177, 238, 5, 5, 182, 238, 192, 0, 0, 0, 5, 5, 10, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 204, 5, 5, 117, 211, 5, 5, 247, 216, 5, 5, 119, 244, 5, 5, 50, 206, 192, 0, 0, 0, 5, 5, 39, 234, 5, 5, 117, 231, 5, 5, 44, 241, 5, 5, 139, 236, 5, 5, 140, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 238, 5, 5, 139, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 200, 192, 0, 0, 0, 5, 5, 135, 231, 5, 5, 49, 234, 192, 0, 0, 0, 5, 5, 173, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 211, 5, 5, 233, 215, 5, 5, 162, 219, 5, 5, 129, 201, 192, 0, 0, 0, 5, 5, 45, 208, 5, 5, 239, 209, 5, 5, 199, 219, 5, 5, 76, 206, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 209, 5, 5, 33, 229, 5, 5, 98, 234, 192, 0, 0, 0, 5, 5, 79, 206, 5, 5, 249, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 220, 5, 5, 173, 224, 5, 5, 128, 234, 192, 0, 0, 0, 5, 5, 228, 236, 5, 5, 213, 211, 5, 5, 90, 239, 5, 5, 88, 239, 5, 5, 104, 213, 5, 5, 32, 246, 5, 5, 37, 247, 192, 0, 0, 0, 5, 5, 32, 210, 192, 0, 0, 0, 5, 5, 116, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 232, 5, 5, 241, 239, 5, 5, 242, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 239, 192, 0, 0, 0, 5, 5, 114, 206, 192, 0, 0, 0, 5, 5, 196, 241, 5, 5, 75, 246, 5, 5, 42, 237, 5, 5, 161, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 247, 5, 5, 102, 247, 5, 5, 177, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 208, 192, 0, 0, 0, 5, 5, 18, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 239, 5, 5, 63, 245, 5, 5, 139, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 134, 2, 4, 0, 0, 0, 42, 0, 134, 3, 4, 0, 0, 0, 50, 0, 134, 2, 4, 0, 0, 0, 50, 0, 134, 3, 4, 0, 0, 0, 58, 0, 134, 2, 4, 0, 0, 0, 58, 0, 134, 3, 4, 0, 0, 0, 66, 0, 134, 2, 4, 0, 0, 0, 66, 0, 134, 3, 4, 0, 0, 0, 68, 0, 134, 2, 4, 0, 0, 0, 68, 0, 134, 3, 4, 0, 0, 0, 70, 0, 134, 2, 4, 0, 0, 0, 70, 0, 134, 3, 4, 0, 0, 0, 82, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 2, 4, 0, 0, 0, 82, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 3, 0, 42, 5, 162, 0, 0, 134, 2, 0, 42, 5, 162, 0, 0, 5, 3, 0, 42, 5, 162, 0, 0, 5, 2, 0, 42, 5, 162, 0, 0, 134, 3, 0, 50, 5, 162, 0, 0, 134, 2, 0, 50, 5, 162, 0, 0, 5, 3, 0, 50, 5, 162, 0, 0, 5, 2, 0, 50, 5, 162, 0, 0, 134, 3, 0, 58, 5, 162, 0, 0, 134, 2, 0, 58, 5, 162, 0, 0, 5, 3, 0, 58, 5, 162, 0, 0, 5, 2, 0, 58, 5, 162, 0, 0, 134, 3, 0, 68, 5, 160, 0, 0, 134, 2, 0, 68, 5, 160, 0, 0, 5, 3, 0, 68, 5, 160, 0, 0, 5, 2, 0, 68, 5, 160, 0, 0, 134, 3, 0, 82, 5, 162, 0, 0, 134, 2, 0, 82, 5, 162, 0, 0, 134, 74, 0, 82, 5, 162, 0, 0, 5, 3, 0, 82, 5, 162, 0, 0, 5, 2, 0, 82, 5, 162, 0, 0, 5, 74, 0, 82, 5, 162, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 134, 3, 0, 70, 5, 162, 0, 0, 134, 2, 0, 70, 5, 162, 0, 0, 5, 3, 0, 70, 5, 162, 0, 0, 5, 2, 0, 70, 5, 162, 0, 0, 134, 3, 0, 50, 5, 160, 0, 0, 134, 2, 0, 50, 5, 160, 0, 0, 5, 3, 0, 50, 5, 160, 0, 0, 5, 2, 0, 50, 5, 160, 0, 0, 134, 3, 0, 42, 5, 200, 0, 0, 134, 2, 0, 42, 5, 200, 0, 0, 5, 3, 0, 42, 5, 200, 0, 0, 5, 2, 0, 42, 5, 200, 0, 0, 134, 2, 0, 50, 5, 138, 0, 0, 5, 2, 0, 50, 5, 138, 0, 0, 134, 2, 0, 50, 5, 136, 0, 0, 5, 2, 0, 50, 5, 136, 0, 0, 134, 3, 0, 50, 5, 204, 0, 0, 134, 2, 0, 50, 5, 204, 0, 0, 5, 3, 0, 50, 5, 204, 0, 0, 5, 2, 0, 50, 5, 204, 0, 0, 134, 3, 0, 50, 5, 208, 0, 0, 134, 2, 0, 50, 5, 208, 0, 0, 5, 3, 0, 50, 5, 208, 0, 0, 5, 2, 0, 50, 5, 208, 0, 0, 134, 3, 0, 58, 5, 208, 0, 0, 134, 2, 0, 58, 5, 208, 0, 0, 5, 3, 0, 58, 5, 208, 0, 0, 5, 2, 0, 58, 5, 208, 0, 0, 134, 3, 0, 66, 5, 196, 0, 0, 134, 2, 0, 66, 5, 196, 0, 0, 5, 3, 0, 66, 5, 196, 0, 0, 5, 2, 0, 66, 5, 196, 0, 0, 134, 3, 0, 68, 5, 196, 0, 0, 134, 2, 0, 68, 5, 196, 0, 0, 5, 3, 0, 68, 5, 196, 0, 0, 5, 2, 0, 68, 5, 196, 0, 0, 134, 3, 0, 68, 5, 210, 0, 0, 134, 2, 0, 68, 5, 210, 0, 0, 5, 3, 0, 68, 5, 210, 0, 0, 5, 2, 0, 68, 5, 210, 0, 0, 134, 3, 0, 68, 5, 204, 0, 0, 134, 2, 0, 68, 5, 204, 0, 0, 5, 3, 0, 68, 5, 204, 0, 0, 5, 2, 0, 68, 5, 204, 0, 0, 134, 2, 0, 70, 5, 138, 0, 0, 5, 2, 0, 70, 5, 138, 0, 0, 134, 2, 0, 70, 5, 136, 0, 0, 5, 2, 0, 70, 5, 136, 0, 0, 134, 3, 0, 82, 5, 198, 0, 0, 134, 2, 0, 82, 5, 198, 0, 0, 134, 74, 0, 82, 5, 198, 0, 0, 5, 3, 0, 82, 5, 198, 0, 0, 5, 2, 0, 82, 5, 198, 0, 0, 5, 74, 0, 82, 5, 198, 0, 0, 134, 3, 0, 82, 5, 208, 0, 0, 134, 2, 0, 82, 5, 208, 0, 0, 134, 74, 0, 82, 5, 208, 0, 0, 5, 3, 0, 82, 5, 208, 0, 0, 5, 2, 0, 82, 5, 208, 0, 0, 5, 74, 0, 82, 5, 208, 0, 0, 134, 3, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 204, 0, 0, 134, 74, 0, 82, 5, 204, 0, 0, 5, 3, 0, 82, 5, 204, 0, 0, 5, 2, 0, 82, 5, 204, 0, 0, 5, 74, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 150, 0, 0, 5, 2, 0, 82, 5, 150, 0, 0, 134, 3, 0, 42, 5, 196, 0, 0, 134, 2, 0, 42, 5, 196, 0, 0, 5, 3, 0, 42, 5, 196, 0, 0, 5, 2, 0, 42, 5, 196, 0, 0, 134, 3, 0, 50, 5, 196, 0, 0, 134, 2, 0, 50, 5, 196, 0, 0, 5, 3, 0, 50, 5, 196, 0, 0, 5, 2, 0, 50, 5, 196, 0, 0, 134, 3, 0, 58, 5, 196, 0, 0, 134, 2, 0, 58, 5, 196, 0, 0, 5, 3, 0, 58, 5, 196, 0, 0, 5, 2, 0, 58, 5, 196, 0, 0, 134, 3, 0, 70, 5, 196, 0, 0, 134, 2, 0, 70, 5, 196, 0, 0, 5, 3, 0, 70, 5, 196, 0, 0, 5, 2, 0, 70, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 196, 0, 0, 134, 2, 0, 82, 5, 196, 0, 0, 134, 74, 0, 82, 5, 196, 0, 0, 5, 3, 0, 82, 5, 196, 0, 0, 5, 2, 0, 82, 5, 196, 0, 0, 5, 74, 0, 82, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 5, 0, 12, 3, 7, 0, 12, 3, 198, 161, 64, 3, 198, 129, 65, 3, 0, 234, 134, 3, 0, 3, 198, 129, 1, 3, 0, 234, 134, 3, 4, 3, 0, 234, 134, 2, 0, 50, 156, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 134, 2, 12, 3, 0, 192, 198, 97, 64, 3, 0, 192, 198, 65, 65, 3, 0, 242, 134, 3, 0, 3, 0, 192, 198, 65, 1, 3, 0, 242, 134, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 58, 156, 5, 5, 0, 12, 3, 9, 0, 12, 3, 0, 192, 198, 161, 64, 3, 0, 192, 198, 129, 65, 3, 0, 250, 134, 3, 0, 3, 0, 192, 198, 129, 1, 3, 0, 250, 134, 3, 4, 3, 0, 250, 134, 2, 0, 66, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 0, 192, 198, 225, 64, 3, 0, 192, 198, 193, 65, 3, 255, 255, 0, 66, 134, 3, 0, 3, 0, 192, 198, 193, 1, 3, 255, 255, 0, 66, 134, 3, 4, 3, 255, 255, 0, 66, 134, 2, 0, 68, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 33, 64, 3, 1, 192, 198, 1, 65, 3, 255, 255, 0, 68, 134, 3, 0, 3, 1, 192, 198, 1, 1, 3, 255, 255, 0, 68, 134, 3, 4, 3, 255, 255, 0, 68, 134, 2, 0, 70, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 97, 64, 3, 1, 192, 198, 65, 65, 3, 255, 255, 0, 70, 134, 3, 0, 3, 1, 192, 198, 65, 1, 3, 255, 255, 0, 70, 134, 3, 4, 3, 255, 255, 0, 70, 134, 2, 0, 82, 156, 5, 7, 0, 12, 3, 14, 0, 12, 3, 1, 192, 198, 161, 64, 3, 1, 192, 198, 129, 65, 3, 255, 255, 0, 82, 134, 3, 68, 3, 255, 255, 0, 82, 134, 71, 0, 3, 1, 192, 198, 129, 1, 3, 255, 255, 0, 82, 134, 3, 4, 3, 255, 255, 0, 82, 134, 2, 8, 3, 197, 127, 0, 82, 134, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 0, 42, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 1, 192, 198, 225, 64, 3, 1, 192, 198, 193, 65, 3, 0, 234, 5, 3, 0, 3, 1, 192, 198, 193, 1, 3, 0, 234, 5, 3, 4, 3, 0, 234, 5, 2, 0, 50, 5, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 5, 2, 12, 3, 2, 192, 198, 33, 64, 3, 2, 192, 198, 1, 65, 3, 0, 242, 5, 3, 0, 3, 2, 192, 198, 1, 1, 3, 0, 242, 5, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 58, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 2, 192, 198, 97, 64, 3, 2, 192, 198, 65, 65, 3, 0, 250, 5, 3, 0, 3, 2, 192, 198, 65, 1, 3, 0, 250, 5, 3, 4, 3, 0, 250, 5, 2, 0, 66, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 161, 64, 3, 2, 192, 198, 129, 65, 3, 255, 255, 0, 66, 5, 3, 0, 3, 2, 192, 198, 129, 1, 3, 255, 255, 0, 66, 5, 3, 4, 3, 255, 255, 0, 66, 5, 2, 0, 68, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 225, 64, 3, 2, 192, 198, 193, 65, 3, 255, 255, 0, 68, 5, 3, 0, 3, 2, 192, 198, 193, 1, 3, 255, 255, 0, 68, 5, 3, 4, 3, 255, 255, 0, 68, 5, 2, 0, 70, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 3, 192, 198, 33, 64, 3, 3, 192, 198, 1, 65, 3, 255, 255, 0, 70, 5, 3, 0, 3, 3, 192, 198, 1, 1, 3, 255, 255, 0, 70, 5, 3, 4, 3, 255, 255, 0, 70, 5, 2, 0, 82, 5, 5, 7, 0, 12, 3, 14, 0, 12, 3, 3, 192, 198, 97, 64, 3, 3, 192, 198, 65, 65, 3, 255, 255, 0, 82, 5, 3, 68, 3, 255, 255, 0, 82, 5, 71, 0, 3, 3, 192, 198, 65, 1, 3, 255, 255, 0, 82, 5, 3, 4, 3, 255, 255, 0, 82, 5, 2, 8, 3, 197, 127, 0, 82, 5, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 82, 134, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 5, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 82, 5, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 42, 196, 162, 5, 0, 12, 3, 8, 0, 12, 3, 3, 192, 198, 194, 64, 3, 3, 192, 198, 130, 65, 3, 197, 162, 0, 3, 3, 192, 198, 130, 1, 3, 197, 162, 4, 3, 0, 192, 197, 98, 5, 42, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 66, 64, 3, 4, 192, 198, 2, 65, 3, 0, 192, 197, 162, 0, 3, 4, 192, 198, 2, 1, 3, 0, 192, 197, 162, 4, 3, 0, 192, 197, 226, 156, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 194, 64, 3, 4, 192, 198, 130, 65, 3, 1, 192, 197, 34, 0, 3, 4, 192, 198, 130, 1, 3, 1, 192, 197, 34, 4, 3, 1, 192, 197, 98, 5, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 66, 64, 3, 5, 192, 198, 2, 65, 3, 1, 192, 197, 162, 0, 3, 5, 192, 198, 2, 1, 3, 1, 192, 197, 162, 4, 3, 1, 192, 197, 226, 156, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 194, 64, 3, 5, 192, 198, 130, 65, 3, 2, 192, 197, 34, 0, 3, 5, 192, 198, 130, 1, 3, 2, 192, 197, 34, 4, 3, 2, 192, 197, 98, 5, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 66, 64, 3, 6, 192, 198, 2, 65, 3, 2, 192, 197, 162, 0, 3, 6, 192, 198, 2, 1, 3, 2, 192, 197, 162, 4, 3, 2, 192, 197, 226, 156, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 194, 64, 3, 6, 192, 198, 130, 65, 3, 3, 192, 197, 34, 0, 3, 6, 192, 198, 130, 1, 3, 3, 192, 197, 34, 4, 3, 3, 192, 197, 98, 5, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 7, 192, 198, 66, 64, 3, 7, 192, 198, 2, 65, 3, 3, 192, 197, 162, 0, 3, 7, 192, 198, 2, 1, 3, 3, 192, 197, 162, 4, 3, 3, 192, 197, 226, 156, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 4, 192, 197, 162, 12, 3, 7, 192, 198, 194, 64, 3, 7, 192, 198, 130, 65, 3, 4, 192, 197, 34, 0, 3, 7, 192, 198, 130, 1, 3, 4, 192, 197, 34, 4, 3, 4, 192, 197, 98, 5, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 5, 192, 197, 98, 12, 3, 8, 192, 198, 66, 64, 3, 8, 192, 198, 2, 65, 3, 4, 192, 197, 226, 0, 3, 8, 192, 198, 2, 1, 3, 4, 192, 197, 226, 4, 3, 5, 192, 197, 34, 156, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 8, 192, 198, 194, 64, 3, 8, 192, 198, 130, 65, 3, 5, 192, 197, 162, 0, 3, 8, 192, 198, 130, 1, 3, 5, 192, 197, 162, 4, 3, 5, 192, 197, 226, 5, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 9, 192, 198, 66, 64, 3, 9, 192, 198, 2, 65, 3, 6, 192, 197, 34, 0, 3, 9, 192, 198, 2, 1, 3, 6, 192, 197, 34, 4, 3, 6, 192, 197, 98, 156, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 34, 12, 3, 9, 192, 198, 194, 64, 3, 9, 192, 198, 130, 65, 3, 6, 192, 197, 162, 0, 3, 9, 192, 198, 130, 1, 3, 6, 192, 197, 162, 4, 3, 6, 192, 197, 226, 5, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 226, 12, 3, 10, 192, 198, 66, 64, 3, 10, 192, 198, 2, 65, 3, 7, 192, 197, 98, 0, 3, 10, 192, 198, 2, 1, 3, 7, 192, 197, 98, 4, 3, 7, 192, 197, 162, 156, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 10, 192, 198, 194, 64, 3, 10, 192, 198, 130, 65, 3, 8, 192, 197, 34, 0, 3, 10, 192, 198, 130, 1, 3, 8, 192, 197, 34, 4, 3, 8, 192, 197, 98, 5, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 66, 64, 3, 11, 192, 198, 2, 65, 3, 8, 192, 197, 162, 0, 3, 11, 192, 198, 2, 1, 3, 8, 192, 197, 162, 4, 3, 8, 192, 197, 226, 156, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 194, 64, 3, 11, 192, 198, 130, 65, 3, 9, 192, 197, 34, 0, 3, 11, 192, 198, 130, 1, 3, 9, 192, 197, 34, 4, 3, 9, 192, 197, 98, 5, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 66, 64, 3, 12, 192, 198, 2, 65, 3, 9, 192, 197, 162, 0, 3, 12, 192, 198, 2, 1, 3, 9, 192, 197, 162, 4, 3, 9, 192, 197, 226, 156, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 194, 64, 3, 12, 192, 198, 130, 65, 3, 10, 192, 197, 34, 0, 3, 12, 192, 198, 130, 1, 3, 10, 192, 197, 34, 4, 3, 10, 192, 197, 98, 5, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 66, 64, 3, 13, 192, 198, 2, 65, 3, 10, 192, 197, 162, 0, 3, 13, 192, 198, 2, 1, 3, 10, 192, 197, 162, 4, 3, 10, 192, 197, 226, 156, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 194, 64, 3, 13, 192, 198, 130, 65, 3, 12, 192, 197, 34, 0, 3, 13, 192, 198, 130, 1, 3, 12, 192, 197, 34, 4, 3, 12, 192, 197, 98, 5, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 66, 64, 3, 14, 192, 198, 2, 65, 3, 12, 192, 197, 162, 0, 3, 14, 192, 198, 2, 1, 3, 12, 192, 197, 162, 4, 3, 12, 192, 197, 226, 156, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 194, 64, 3, 14, 192, 198, 130, 65, 3, 13, 192, 197, 34, 0, 3, 14, 192, 198, 130, 1, 3, 13, 192, 197, 34, 4, 3, 13, 192, 197, 98, 5, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 66, 64, 3, 15, 192, 198, 2, 65, 3, 13, 192, 197, 162, 0, 3, 15, 192, 198, 2, 1, 3, 13, 192, 197, 162, 4, 3, 13, 192, 197, 226, 156, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 194, 64, 3, 15, 192, 198, 130, 65, 3, 14, 192, 197, 34, 0, 3, 15, 192, 198, 130, 1, 3, 14, 192, 197, 34, 4, 3, 14, 192, 197, 98, 5, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 66, 64, 3, 16, 192, 198, 2, 65, 3, 14, 192, 197, 162, 0, 3, 16, 192, 198, 2, 1, 3, 14, 192, 197, 162, 4, 3, 14, 192, 197, 226, 156, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 194, 64, 3, 16, 192, 198, 130, 65, 3, 15, 192, 197, 34, 0, 3, 16, 192, 198, 130, 1, 3, 15, 192, 197, 34, 4, 3, 15, 192, 197, 98, 5, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 66, 64, 3, 17, 192, 198, 2, 65, 3, 15, 192, 197, 162, 0, 3, 17, 192, 198, 2, 1, 3, 15, 192, 197, 162, 4, 3, 15, 192, 197, 226, 156, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 194, 64, 3, 17, 192, 198, 130, 65, 3, 16, 192, 197, 34, 0, 3, 17, 192, 198, 130, 1, 3, 16, 192, 197, 34, 4, 3, 16, 192, 197, 98, 5, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 66, 64, 3, 18, 192, 198, 2, 65, 3, 16, 192, 197, 162, 0, 3, 18, 192, 198, 2, 1, 3, 16, 192, 197, 162, 4, 3, 16, 192, 197, 226, 156, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 194, 64, 3, 18, 192, 198, 130, 65, 3, 17, 192, 197, 34, 0, 3, 18, 192, 198, 130, 1, 3, 17, 192, 197, 34, 4, 3, 17, 192, 197, 98, 5, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 66, 64, 3, 19, 192, 198, 2, 65, 3, 17, 192, 197, 162, 0, 3, 19, 192, 198, 2, 1, 3, 17, 192, 197, 162, 4, 3, 17, 192, 197, 226, 156, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 194, 64, 3, 19, 192, 198, 130, 65, 3, 18, 192, 197, 34, 0, 3, 19, 192, 198, 130, 1, 3, 18, 192, 197, 34, 4, 3, 18, 192, 197, 98, 5, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 20, 192, 198, 66, 64, 3, 20, 192, 198, 2, 65, 3, 18, 192, 197, 162, 0, 3, 20, 192, 198, 2, 1, 3, 18, 192, 197, 162, 4, 3, 18, 192, 197, 226, 156, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 20, 192, 197, 162, 12, 3, 20, 192, 198, 194, 64, 3, 20, 192, 198, 130, 65, 3, 20, 192, 197, 34, 0, 3, 20, 192, 198, 130, 1, 3, 20, 192, 197, 34, 4, 3, 20, 192, 197, 98, 5, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 21, 192, 197, 98, 12, 3, 21, 192, 198, 66, 64, 3, 21, 192, 198, 2, 65, 3, 20, 192, 197, 226, 0, 3, 21, 192, 198, 2, 1, 3, 20, 192, 197, 226, 4, 3, 21, 192, 197, 34, 156, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 34, 12, 3, 21, 192, 198, 194, 64, 3, 21, 192, 198, 130, 65, 3, 21, 192, 197, 162, 0, 3, 21, 192, 198, 130, 1, 3, 21, 192, 197, 162, 4, 3, 21, 192, 197, 226, 5, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 226, 12, 3, 22, 192, 198, 66, 64, 3, 22, 192, 198, 2, 65, 3, 22, 192, 197, 98, 0, 3, 22, 192, 198, 2, 1, 3, 22, 192, 197, 98, 4, 3, 22, 192, 197, 162, 156, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 23, 192, 197, 162, 12, 3, 22, 192, 198, 194, 64, 3, 22, 192, 198, 130, 65, 3, 23, 192, 197, 34, 0, 3, 22, 192, 198, 130, 1, 3, 23, 192, 197, 34, 4, 3, 23, 192, 197, 98, 5, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 24, 192, 197, 98, 12, 3, 23, 192, 198, 66, 64, 3, 23, 192, 198, 2, 65, 3, 23, 192, 197, 226, 0, 3, 23, 192, 198, 2, 1, 3, 23, 192, 197, 226, 4, 3, 24, 192, 197, 34, 156, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 23, 192, 198, 194, 64, 3, 23, 192, 198, 130, 65, 3, 25, 192, 197, 34, 0, 3, 23, 192, 198, 130, 1, 3, 25, 192, 197, 34, 4, 3, 25, 192, 197, 98, 5, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 66, 64, 3, 24, 192, 198, 2, 65, 3, 25, 192, 197, 162, 0, 3, 24, 192, 198, 2, 1, 3, 25, 192, 197, 162, 4, 3, 25, 192, 197, 226, 156, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 194, 64, 3, 24, 192, 198, 130, 65, 3, 26, 192, 197, 34, 0, 3, 24, 192, 198, 130, 1, 3, 26, 192, 197, 34, 4, 3, 26, 192, 197, 98, 5, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 66, 64, 3, 25, 192, 198, 2, 65, 3, 26, 192, 197, 162, 0, 3, 25, 192, 198, 2, 1, 3, 26, 192, 197, 162, 4, 3, 26, 192, 197, 226, 156, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 194, 64, 3, 25, 192, 198, 130, 65, 3, 27, 192, 197, 34, 0, 3, 25, 192, 198, 130, 1, 3, 27, 192, 197, 34, 4, 3, 27, 192, 197, 98, 5, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 66, 64, 3, 26, 192, 198, 2, 65, 3, 27, 192, 197, 162, 0, 3, 26, 192, 198, 2, 1, 3, 27, 192, 197, 162, 4, 3, 27, 192, 197, 226, 156, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 194, 64, 3, 26, 192, 198, 130, 65, 3, 28, 192, 197, 34, 0, 3, 26, 192, 198, 130, 1, 3, 28, 192, 197, 34, 4, 3, 28, 192, 197, 98, 5, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 66, 64, 3, 27, 192, 198, 2, 65, 3, 28, 192, 197, 162, 0, 3, 27, 192, 198, 2, 1, 3, 28, 192, 197, 162, 4, 3, 28, 192, 197, 226, 29, 0, 197, 35, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 227, 64, 3, 27, 192, 198, 131, 65, 3, 29, 192, 197, 131, 0, 3, 27, 192, 198, 131, 1, 3, 29, 192, 197, 131, 4, 3, 29, 192, 197, 227, 30, 0, 197, 67, 5, 0, 12, 3, 9, 0, 12, 3, 28, 192, 198, 163, 64, 3, 28, 192, 198, 67, 65, 3, 30, 192, 197, 163, 0, 3, 28, 192, 198, 67, 1, 3, 30, 192, 197, 163, 4, 3, 31, 192, 197, 3, 156, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 31, 192, 197, 226, 12, 3, 29, 192, 198, 66, 64, 3, 29, 192, 198, 2, 65, 3, 31, 192, 197, 98, 0, 3, 29, 192, 198, 2, 1, 3, 31, 192, 197, 98, 4, 3, 31, 192, 197, 162, 5, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 32, 192, 197, 162, 12, 3, 29, 192, 198, 194, 64, 3, 29, 192, 198, 130, 65, 3, 32, 192, 197, 34, 0, 3, 29, 192, 198, 130, 1, 3, 32, 192, 197, 34, 4, 3, 32, 192, 197, 98, 32, 0, 197, 227, 6, 0, 8, 3, 12, 0, 8, 3, 34, 192, 197, 3, 12, 3, 30, 192, 198, 99, 64, 3, 30, 192, 198, 3, 65, 3, 33, 192, 197, 67, 0, 3, 30, 192, 198, 3, 1, 3, 33, 192, 197, 67, 4, 3, 33, 192, 197, 163, 34, 0, 197, 99, 6, 0, 8, 3, 12, 0, 8, 3, 35, 192, 197, 131, 12, 3, 31, 192, 198, 35, 64, 3, 30, 192, 198, 195, 65, 3, 34, 192, 197, 195, 0, 3, 30, 192, 198, 195, 1, 3, 34, 192, 197, 195, 4, 3, 35, 192, 197, 35, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 28, 4, 185, 76, 185, 80, 185, 84, 30, 4, 186, 96, 185, 100, 185, 104, 32, 4, 34, 4, 36, 4, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 38, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 40, 4, 168, 60, 168, 64, 168, 68, 42, 4, 168, 76, 168, 80, 168, 84, 44, 4, 168, 96, 168, 100, 168, 104, 46, 4, 48, 4, 50, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 52, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 89, 56, 57, 56, 217, 57, 121, 58, 57, 58, 25, 58, 0, 8, 217, 66, 121, 72, 57, 72, 54, 4, 58, 74, 89, 88, 57, 88, 218, 89, 58, 90, 217, 68, 122, 114, 89, 120, 57, 120, 217, 121, 121, 122, 57, 122, 136, 14, 185, 122, 89, 152, 57, 152, 217, 153, 56, 4, 153, 169, 185, 176, 2, 8, 72, 56, 40, 56, 200, 57, 104, 58, 40, 58, 8, 58, 4, 8, 200, 66, 104, 72, 40, 72, 58, 4, 40, 74, 72, 88, 40, 88, 200, 89, 40, 90, 200, 68, 104, 114, 72, 120, 40, 120, 200, 121, 104, 122, 40, 122, 128, 14, 168, 122, 72, 152, 40, 152, 200, 153, 60, 4, 136, 169, 168, 176, 40, 170, 25, 56, 8, 56, 185, 57, 168, 57, 62, 4, 64, 4, 153, 65, 136, 65, 217, 65, 200, 65, 153, 66, 136, 66, 249, 65, 232, 65, 249, 69, 232, 69, 25, 71, 8, 71, 25, 72, 8, 72, 186, 73, 168, 73, 154, 74, 136, 74, 66, 4, 68, 4, 89, 72, 72, 72, 217, 81, 200, 81, 185, 81, 168, 81, 153, 82, 136, 82, 217, 82, 200, 82, 217, 85, 200, 85, 25, 87, 8, 87, 122, 90, 104, 90, 25, 88, 8, 88, 186, 89, 168, 89, 70, 4, 72, 4, 154, 90, 168, 92, 6, 8, 8, 8, 218, 97, 200, 97, 217, 102, 200, 102, 168, 132, 153, 105, 136, 105, 217, 106, 200, 106, 249, 105, 232, 105, 57, 107, 40, 107, 25, 107, 8, 107, 25, 112, 8, 112, 74, 4, 76, 4, 57, 112, 40, 112, 10, 8, 185, 116, 168, 116, 25, 120, 8, 120, 185, 121, 168, 121, 89, 122, 72, 122, 12, 8, 14, 8, 153, 137, 136, 137, 217, 138, 200, 138, 249, 137, 232, 137, 154, 141, 136, 141, 218, 141, 200, 141, 218, 142, 200, 142, 250, 141, 232, 141, 217, 146, 200, 146, 249, 145, 232, 145, 185, 148, 168, 148, 121, 154, 104, 154, 25, 152, 8, 152, 185, 153, 168, 153, 25, 154, 8, 154, 89, 154, 72, 154, 78, 4, 80, 4, 217, 161, 200, 161, 217, 169, 200, 169, 57, 170, 153, 173, 136, 173, 153, 174, 136, 174, 249, 173, 232, 173, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 72, 255, 5, 186, 88, 255, 5, 185, 108, 255, 5, 186, 112, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 255, 5, 168, 72, 255, 5, 168, 88, 255, 5, 168, 108, 255, 5, 168, 112, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 218, 73, 255, 5, 217, 152, 255, 5, 200, 73, 255, 5, 200, 152, 255, 5, 249, 58, 255, 5, 232, 58, 255, 5, 250, 74, 255, 5, 232, 74, 255, 5, 250, 90, 255, 5, 232, 90, 255, 5, 218, 114, 255, 5, 200, 114, 255, 5, 249, 154, 255, 5, 232, 154, 255, 1, 170, 170, 3, 0, 68, 0, 59, 0, 32, 0, 27, 206, 0, 16, 140, 28, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 118, 113, 1, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 76, 0, 0, 0, 76, 1, 0, 0, 44, 82, 1, 0, 44, 82, 1, 0, 12, 90, 1, 0, 12, 90, 1, 0, 200, 94, 1, 0, 200, 94, 1, 0, 42, 109, 1, 0, 42, 109, 1, 0, 90, 113, 1, 0, 25, 0, 0, 0, 17, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 254, 255, 50, 105, 114, 84, 1, 0, 192, 8, 245, 19, 64, 0, 192, 0, 96, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 109, 0, 117, 0, 122, 0, 130, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 169, 0, 176, 0, 184, 0, 189, 0, 48, 0, 197, 0, 205, 0, 213, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 221, 0, 229, 0, 237, 0, 245, 0, 253, 0, 5, 1, 13, 1, 21, 1, 29, 1, 37, 1, 45, 1, 53, 1, 61, 1, 69, 1, 77, 1, 85, 1, 93, 1, 101, 1, 109, 1, 117, 1, 125, 1, 133, 1, 140, 1, 147, 1, 154, 1, 162, 1, 170, 1, 178, 1, 186, 1, 194, 1, 202, 1, 210, 1, 218, 1, 226, 1, 234, 1, 242, 1, 250, 1, 2, 2, 10, 2, 18, 2, 26, 2, 34, 2, 42, 2, 50, 2, 58, 2, 66, 2, 74, 2, 82, 2, 90, 2, 98, 2, 106, 2, 114, 2, 122, 2, 130, 2, 138, 2, 146, 2, 154, 2, 162, 2, 170, 2, 178, 2, 186, 2, 194, 2, 202, 2, 210, 2, 218, 2, 226, 2, 234, 2, 242, 2, 250, 2, 2, 3, 10, 3, 18, 3, 26, 3, 34, 3, 42, 3, 50, 3, 58, 3, 66, 3, 74, 3, 82, 3, 90, 3, 98, 3, 105, 3, 113, 3, 121, 3, 129, 3, 137, 3, 144, 3, 152, 3, 160, 3, 168, 3, 176, 3, 184, 3, 192, 3, 200, 3, 208, 3, 216, 3, 224, 3, 232, 3, 240, 3, 248, 3, 255, 3, 7, 4, 15, 4, 21, 4, 27, 4, 35, 4, 43, 4, 51, 4, 59, 4, 67, 4, 75, 4, 83, 4, 91, 4, 99, 4, 107, 4, 115, 4, 123, 4, 131, 4, 139, 4, 147, 4, 155, 4, 163, 4, 171, 4, 179, 4, 187, 4, 194, 4, 202, 4, 210, 4, 218, 4, 226, 4, 234, 4, 242, 4, 250, 4, 2, 5, 9, 5, 17, 5, 25, 5, 33, 5, 41, 5, 49, 5, 57, 5, 65, 5, 73, 5, 81, 5, 89, 5, 97, 5, 105, 5, 113, 5, 121, 5, 129, 5, 137, 5, 145, 5, 153, 5, 161, 5, 169, 5, 175, 5, 183, 5, 191, 5, 198, 5, 206, 5, 214, 5, 222, 5, 230, 5, 238, 5, 246, 5, 254, 5, 6, 6, 14, 6, 22, 6, 30, 6, 38, 6, 46, 6, 54, 6, 62, 6, 70, 6, 78, 6, 86, 6, 94, 6, 102, 6, 110, 6, 118, 6, 126, 6, 134, 6, 142, 6, 149, 6, 157, 6, 165, 6, 172, 6, 180, 6, 188, 6, 196, 6, 204, 6, 212, 6, 220, 6, 228, 6, 236, 6, 244, 6, 252, 6, 4, 7, 12, 7, 20, 7, 28, 7, 36, 7, 44, 7, 52, 7, 60, 7, 68, 7, 76, 7, 84, 7, 92, 7, 100, 7, 108, 7, 116, 7, 124, 7, 132, 7, 140, 7, 148, 7, 156, 7, 164, 7, 172, 7, 180, 7, 188, 7, 196, 7, 204, 7, 212, 7, 220, 7, 228, 7, 236, 7, 244, 7, 48, 0, 48, 0, 48, 0, 252, 7, 4, 8, 12, 8, 20, 8, 28, 8, 36, 8, 44, 8, 52, 8, 60, 8, 68, 8, 76, 8, 84, 8, 92, 8, 100, 8, 108, 8, 116, 8, 124, 8, 132, 8, 140, 8, 148, 8, 156, 8, 164, 8, 172, 8, 180, 8, 188, 8, 196, 8, 204, 8, 212, 8, 220, 8, 228, 8, 236, 8, 244, 8, 252, 8, 4, 9, 12, 9, 20, 9, 28, 9, 36, 9, 44, 9, 48, 9, 56, 9, 63, 9, 71, 9, 79, 9, 87, 9, 95, 9, 103, 9, 111, 9, 118, 9, 125, 9, 133, 9, 141, 9, 149, 9, 157, 9, 165, 9, 172, 9, 180, 9, 188, 9, 196, 9, 204, 9, 212, 9, 220, 9, 228, 9, 236, 9, 244, 9, 252, 9, 4, 10, 12, 10, 20, 10, 28, 10, 36, 10, 44, 10, 52, 10, 60, 10, 68, 10, 76, 10, 84, 10, 91, 10, 99, 10, 107, 10, 115, 10, 123, 10, 131, 10, 139, 10, 147, 10, 155, 10, 163, 10, 171, 10, 179, 10, 187, 10, 195, 10, 203, 10, 210, 10, 218, 10, 226, 10, 234, 10, 242, 10, 250, 10, 2, 11, 10, 11, 18, 11, 26, 11, 34, 11, 42, 11, 50, 11, 58, 11, 66, 11, 74, 11, 82, 11, 90, 11, 97, 11, 105, 11, 113, 11, 121, 11, 129, 11, 137, 11, 144, 11, 152, 11, 160, 11, 168, 11, 176, 11, 184, 11, 190, 11, 198, 11, 206, 11, 214, 11, 222, 11, 230, 11, 238, 11, 246, 11, 254, 11, 6, 12, 14, 12, 22, 12, 30, 12, 38, 12, 46, 12, 54, 12, 58, 12, 66, 12, 74, 12, 82, 12, 90, 12, 97, 12, 105, 12, 48, 0, 110, 12, 48, 0, 117, 12, 48, 0, 48, 0, 48, 0, 121, 12, 129, 12, 48, 0, 137, 12, 145, 12, 153, 12, 161, 12, 169, 12, 177, 12, 185, 12, 193, 12, 201, 12, 209, 12, 217, 12, 225, 12, 233, 12, 241, 12, 249, 12, 0, 13, 8, 13, 16, 13, 24, 13, 32, 13, 40, 13, 48, 13, 56, 13, 64, 13, 72, 13, 80, 13, 88, 13, 96, 13, 104, 13, 112, 13, 120, 13, 128, 13, 136, 13, 144, 13, 152, 13, 160, 13, 168, 13, 176, 13, 184, 13, 192, 13, 200, 13, 208, 13, 216, 13, 224, 13, 232, 13, 240, 13, 248, 13, 0, 14, 8, 14, 16, 14, 24, 14, 32, 14, 40, 14, 48, 14, 56, 14, 64, 14, 72, 14, 80, 14, 88, 14, 96, 14, 104, 14, 111, 14, 119, 14, 127, 14, 135, 14, 143, 14, 151, 14, 159, 14, 167, 14, 175, 14, 183, 14, 191, 14, 199, 14, 207, 14, 215, 14, 223, 14, 231, 14, 239, 14, 247, 14, 255, 14, 7, 15, 15, 15, 23, 15, 31, 15, 39, 15, 47, 15, 55, 15, 63, 15, 71, 15, 48, 0, 79, 15, 87, 15, 95, 15, 97, 15, 105, 15, 108, 15, 115, 15, 48, 0, 48, 0, 48, 0, 122, 15, 129, 15, 48, 0, 136, 15, 48, 0, 144, 15, 152, 15, 160, 15, 168, 15, 176, 15, 184, 15, 192, 15, 200, 15, 48, 0, 48, 0, 48, 0, 207, 15, 215, 15, 223, 15, 231, 15, 239, 15, 247, 15, 255, 15, 7, 16, 15, 16, 23, 16, 31, 16, 38, 16, 46, 16, 54, 16, 61, 16, 48, 0, 48, 0, 48, 0, 48, 0, 68, 16, 76, 16, 84, 16, 92, 16, 100, 16, 108, 16, 116, 16, 124, 16, 132, 16, 140, 16, 148, 16, 156, 16, 164, 16, 172, 16, 180, 16, 188, 16, 196, 16, 204, 16, 212, 16, 219, 16, 48, 0, 48, 0, 48, 0, 48, 0, 227, 16, 233, 16, 238, 16, 244, 16, 48, 0, 48, 0, 247, 16, 48, 0, 255, 16, 4, 17, 48, 0, 10, 17, 48, 0, 48, 0, 16, 17, 48, 0, 24, 17, 31, 17, 39, 17, 47, 17, 55, 17, 63, 17, 71, 17, 79, 17, 87, 17, 48, 0, 48, 0, 48, 0, 95, 17, 103, 17, 111, 17, 119, 17, 127, 17, 134, 17, 142, 17, 150, 17, 158, 17, 166, 17, 174, 17, 182, 17, 190, 17, 198, 17, 206, 17, 48, 0, 213, 17, 48, 0, 48, 0, 48, 0, 219, 17, 227, 17, 235, 17, 240, 17, 248, 17, 253, 17, 2, 18, 6, 18, 14, 18, 22, 18, 29, 18, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 34, 18, 42, 18, 50, 18, 58, 18, 66, 18, 74, 18, 82, 18, 90, 18, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 96, 18, 104, 18, 112, 18, 120, 18, 128, 18, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 129, 18, 137, 18, 145, 18, 153, 18, 161, 18, 169, 18, 177, 18, 185, 18, 193, 18, 201, 18, 209, 18, 214, 18, 222, 18, 230, 18, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 228, 19, 161, 0, 48, 0, 48, 0, 153, 0, 153, 0, 153, 0, 236, 19, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 232, 18, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 240, 18, 248, 18, 0, 19, 8, 19, 16, 19, 24, 19, 32, 19, 40, 19, 48, 19, 56, 19, 64, 19, 72, 19, 80, 19, 88, 19, 94, 19, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 180, 1, 232, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 128, 8, 102, 19, 110, 19, 118, 19, 126, 19, 134, 19, 142, 19, 150, 19, 158, 19, 166, 19, 174, 19, 182, 19, 190, 19, 198, 19, 206, 19, 213, 19, 48, 0, 220, 19, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 36, 0, 202, 193, 36, 0, 202, 226, 36, 0, 202, 3, 37, 0, 202, 36, 37, 0, 202, 69, 37, 0, 202, 102, 37, 0, 202, 135, 37, 0, 202, 168, 37, 0, 202, 201, 37, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 134, 2, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 38, 8, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 201, 6, 11, 0, 201, 70, 14, 0, 201, 134, 17, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 198, 20, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 70, 28, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 38, 31, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 198, 36, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 201, 166, 39, 0, 201, 230, 42, 0, 201, 38, 46, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 102, 49, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 134, 3, 0, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 35, 0, 196, 160, 156, 46, 198, 65, 0, 0, 134, 3, 0, 50, 201, 230, 56, 0, 196, 150, 156, 50, 198, 129, 0, 0, 134, 3, 0, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 198, 65, 1, 0, 134, 3, 0, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 198, 129, 1, 0, 134, 3, 0, 82, 196, 142, 156, 82, 201, 198, 59, 0, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 36, 0, 198, 193, 1, 0, 5, 3, 0, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 36, 0, 196, 160, 5, 46, 198, 1, 2, 0, 5, 3, 0, 50, 201, 102, 63, 0, 196, 150, 5, 50, 198, 65, 2, 0, 5, 3, 0, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 198, 1, 3, 0, 5, 3, 0, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 198, 65, 3, 0, 5, 3, 0, 82, 196, 142, 5, 82, 201, 70, 66, 0, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 134, 2, 0, 42, 5, 2, 0, 42, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 69, 0, 201, 134, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 50, 5, 2, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 75, 0, 201, 70, 78, 0, 198, 97, 0, 0, 198, 33, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 58, 5, 2, 0, 58, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 81, 0, 201, 6, 84, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 68, 5, 3, 0, 68, 201, 230, 86, 0, 201, 198, 89, 0, 198, 33, 1, 0, 198, 225, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 70, 5, 2, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 82, 5, 2, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 92, 0, 201, 230, 95, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 99, 0, 201, 6, 102, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 104, 0, 201, 38, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 198, 225, 1, 0, 198, 161, 0, 0, 198, 97, 2, 0, 198, 97, 1, 0, 198, 33, 3, 0, 198, 161, 1, 0, 198, 97, 3, 0, 134, 70, 0, 82, 5, 70, 0, 82, 134, 71, 0, 82, 5, 71, 0, 82, 134, 72, 0, 82, 5, 72, 0, 82, 134, 73, 0, 82, 5, 73, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 111, 0, 201, 70, 114, 0, 197, 98, 8, 0, 197, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 1, 0, 198, 193, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 117, 0, 201, 6, 120, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 122, 0, 201, 198, 125, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 201, 166, 128, 0, 201, 134, 131, 0, 201, 102, 134, 0, 201, 70, 137, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 140, 0, 201, 6, 143, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 66, 5, 3, 0, 66, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 145, 0, 201, 198, 148, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 151, 0, 201, 134, 154, 0, 201, 102, 157, 0, 201, 70, 160, 0, 201, 38, 163, 0, 201, 6, 166, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 197, 226, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 168, 0, 201, 38, 172, 0, 201, 102, 175, 0, 201, 166, 178, 0, 201, 230, 181, 0, 201, 38, 185, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 24, 0, 197, 226, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 188, 0, 201, 70, 191, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 194, 0, 201, 6, 197, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 50, 5, 71, 0, 50, 134, 73, 0, 50, 5, 73, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 199, 0, 201, 198, 202, 0, 201, 166, 205, 0, 201, 134, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 5, 0, 197, 34, 6, 0, 198, 130, 8, 0, 198, 2, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 211, 0, 201, 70, 214, 0, 201, 38, 217, 0, 201, 102, 220, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 6, 0, 197, 98, 7, 0, 198, 130, 9, 0, 198, 2, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 223, 0, 201, 230, 226, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 236, 5, 5, 215, 225, 192, 0, 0, 0, 5, 5, 3, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 234, 5, 5, 121, 237, 5, 5, 221, 232, 5, 5, 15, 233, 5, 5, 58, 235, 5, 5, 110, 238, 5, 5, 119, 224, 5, 5, 221, 236, 192, 0, 0, 0, 5, 5, 112, 238, 5, 5, 15, 225, 192, 0, 0, 0, 5, 5, 22, 238, 5, 5, 84, 232, 5, 5, 115, 238, 5, 5, 94, 233, 192, 0, 0, 0, 5, 5, 114, 232, 5, 5, 89, 224, 5, 5, 73, 236, 5, 5, 78, 225, 5, 5, 225, 225, 5, 5, 187, 233, 5, 5, 117, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 225, 192, 0, 0, 0, 5, 5, 7, 230, 5, 5, 5, 236, 192, 0, 0, 0, 5, 5, 227, 232, 5, 5, 121, 238, 192, 0, 0, 0, 5, 5, 234, 226, 5, 5, 242, 235, 5, 5, 186, 242, 5, 5, 222, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 226, 192, 0, 0, 0, 5, 5, 40, 225, 192, 0, 0, 0, 5, 5, 35, 230, 192, 0, 0, 0, 5, 5, 136, 238, 192, 0, 0, 0, 5, 5, 162, 234, 5, 5, 124, 225, 5, 5, 194, 234, 5, 5, 7, 238, 192, 0, 0, 0, 5, 5, 224, 229, 5, 5, 7, 229, 5, 5, 123, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 231, 192, 0, 0, 0, 5, 5, 239, 228, 192, 0, 0, 0, 5, 5, 125, 238, 5, 5, 186, 230, 5, 5, 121, 236, 192, 0, 0, 0, 5, 5, 188, 237, 5, 5, 242, 234, 5, 5, 87, 237, 5, 5, 167, 227, 5, 5, 60, 226, 5, 5, 192, 229, 192, 0, 0, 0, 5, 5, 219, 231, 5, 5, 150, 231, 5, 5, 73, 232, 192, 0, 0, 0, 5, 5, 40, 227, 192, 0, 0, 0, 5, 5, 229, 224, 5, 5, 102, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 238, 5, 5, 241, 228, 5, 5, 22, 232, 5, 5, 70, 236, 5, 5, 40, 235, 5, 5, 95, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 230, 5, 5, 122, 230, 192, 0, 0, 0, 5, 5, 201, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 230, 192, 0, 0, 0, 5, 5, 218, 236, 5, 5, 169, 237, 192, 0, 0, 0, 5, 5, 96, 233, 5, 5, 54, 226, 5, 5, 109, 238, 5, 5, 204, 236, 5, 5, 131, 229, 192, 0, 0, 0, 5, 5, 24, 237, 5, 5, 181, 227, 5, 5, 141, 238, 5, 5, 253, 234, 5, 5, 218, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 238, 192, 0, 0, 0, 5, 5, 251, 235, 5, 5, 127, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 238, 5, 5, 27, 239, 5, 5, 176, 234, 5, 5, 70, 229, 192, 0, 0, 0, 5, 5, 137, 228, 5, 5, 87, 227, 5, 5, 116, 236, 5, 5, 182, 224, 5, 5, 146, 227, 5, 5, 28, 231, 192, 0, 0, 0, 5, 5, 101, 235, 5, 5, 213, 228, 5, 5, 94, 234, 5, 5, 11, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 232, 5, 5, 29, 239, 192, 0, 0, 0, 5, 5, 32, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 232, 5, 5, 181, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 236, 5, 5, 80, 233, 5, 5, 170, 232, 5, 5, 184, 238, 5, 5, 182, 238, 5, 5, 146, 238, 5, 5, 196, 228, 5, 5, 239, 231, 5, 5, 12, 225, 192, 0, 0, 0, 5, 5, 183, 238, 5, 5, 191, 228, 5, 5, 183, 228, 192, 0, 0, 0, 5, 5, 180, 232, 5, 5, 77, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 230, 192, 0, 0, 0, 5, 5, 146, 224, 5, 5, 62, 238, 5, 5, 105, 233, 5, 5, 247, 233, 5, 5, 124, 237, 5, 5, 170, 226, 5, 5, 65, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 239, 5, 5, 188, 238, 5, 5, 44, 232, 192, 0, 0, 0, 5, 5, 186, 238, 192, 0, 0, 0, 5, 5, 116, 225, 5, 5, 56, 230, 5, 5, 104, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 238, 192, 0, 0, 0, 5, 5, 91, 236, 5, 5, 187, 238, 5, 5, 205, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 236, 192, 0, 0, 0, 5, 5, 231, 237, 5, 5, 190, 238, 192, 0, 0, 0, 5, 5, 193, 238, 5, 5, 110, 228, 5, 5, 75, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 232, 192, 0, 0, 0, 5, 5, 115, 226, 192, 0, 0, 0, 5, 5, 88, 226, 192, 0, 0, 0, 5, 5, 23, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 238, 5, 5, 85, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 235, 5, 5, 47, 228, 5, 5, 146, 226, 5, 5, 59, 226, 5, 5, 179, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 237, 5, 5, 183, 236, 5, 5, 255, 227, 5, 5, 241, 227, 5, 5, 189, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 232, 5, 5, 200, 234, 5, 5, 37, 225, 192, 0, 0, 0, 5, 5, 191, 238, 192, 0, 0, 0, 5, 5, 11, 233, 5, 5, 194, 238, 5, 5, 127, 230, 5, 5, 195, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 234, 5, 5, 197, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 224, 5, 5, 20, 227, 192, 0, 0, 0, 5, 5, 205, 238, 192, 0, 0, 0, 5, 5, 0, 199, 192, 0, 0, 0, 5, 5, 48, 230, 192, 0, 0, 0, 5, 5, 44, 233, 192, 0, 0, 0, 5, 5, 193, 233, 192, 0, 0, 0, 5, 5, 194, 233, 5, 5, 206, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 234, 5, 5, 165, 225, 5, 5, 15, 238, 5, 5, 98, 238, 5, 5, 197, 236, 192, 0, 0, 0, 5, 5, 65, 234, 192, 0, 0, 0, 5, 5, 128, 227, 192, 0, 0, 0, 5, 5, 204, 238, 5, 5, 8, 239, 5, 5, 210, 236, 5, 5, 201, 238, 5, 5, 134, 226, 5, 5, 101, 238, 5, 5, 202, 238, 5, 5, 198, 238, 5, 5, 203, 238, 5, 5, 73, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 236, 5, 5, 192, 238, 5, 5, 9, 239, 192, 0, 0, 0, 5, 5, 199, 238, 192, 0, 0, 0, 5, 5, 169, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 236, 5, 5, 0, 185, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 228, 5, 5, 208, 238, 192, 0, 0, 0, 5, 5, 207, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 238, 5, 5, 216, 238, 192, 0, 0, 0, 5, 5, 213, 238, 5, 5, 87, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 238, 5, 5, 196, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 224, 5, 5, 210, 238, 192, 0, 0, 0, 5, 5, 231, 229, 192, 0, 0, 0, 5, 5, 106, 233, 192, 0, 0, 0, 5, 5, 212, 238, 192, 0, 0, 0, 5, 5, 209, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 226, 192, 0, 0, 0, 5, 5, 84, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 230, 192, 0, 0, 0, 5, 5, 146, 228, 5, 5, 156, 237, 5, 5, 156, 224, 5, 5, 74, 232, 5, 5, 119, 229, 5, 5, 215, 238, 192, 0, 0, 0, 5, 5, 217, 238, 192, 0, 0, 0, 5, 5, 4, 235, 5, 5, 160, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 225, 5, 5, 37, 226, 5, 5, 222, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 239, 5, 5, 80, 232, 5, 5, 232, 229, 5, 5, 226, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 233, 5, 5, 147, 226, 192, 0, 0, 0, 5, 5, 223, 238, 192, 0, 0, 0, 5, 5, 225, 238, 5, 5, 0, 223, 5, 5, 211, 236, 5, 5, 227, 238, 192, 0, 0, 0, 5, 5, 155, 235, 192, 0, 0, 0, 5, 5, 224, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 238, 192, 0, 0, 0, 5, 5, 220, 238, 5, 5, 241, 229, 5, 5, 221, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 228, 5, 5, 180, 235, 5, 5, 157, 226, 192, 0, 0, 0, 5, 5, 17, 229, 192, 0, 0, 0, 5, 5, 231, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 238, 192, 0, 0, 0, 5, 5, 0, 145, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 238, 5, 5, 20, 224, 192, 0, 0, 0, 5, 5, 233, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 225, 192, 0, 0, 0, 5, 5, 33, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 234, 5, 5, 164, 227, 5, 5, 100, 236, 192, 0, 0, 0, 5, 5, 237, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 228, 192, 0, 0, 0, 5, 5, 197, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 238, 5, 5, 25, 229, 192, 0, 0, 0, 5, 5, 240, 238, 5, 5, 229, 238, 5, 5, 69, 231, 192, 0, 0, 0, 5, 5, 232, 238, 5, 5, 235, 238, 5, 5, 234, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 237, 192, 0, 0, 0, 5, 5, 195, 237, 192, 0, 0, 0, 5, 5, 101, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 228, 5, 5, 244, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 238, 192, 0, 0, 0, 5, 5, 245, 238, 5, 5, 205, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 238, 192, 0, 0, 0, 5, 5, 93, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 231, 5, 5, 111, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 238, 5, 5, 191, 224, 5, 5, 138, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 225, 192, 0, 0, 0, 5, 5, 248, 238, 192, 0, 0, 0, 5, 5, 249, 238, 5, 5, 28, 225, 5, 5, 250, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 238, 5, 5, 243, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 239, 5, 5, 236, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 238, 5, 5, 254, 238, 5, 5, 3, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 233, 192, 0, 0, 0, 5, 5, 119, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 238, 5, 5, 4, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 226, 5, 5, 111, 238, 5, 5, 28, 237, 192, 0, 0, 0, 5, 5, 250, 236, 5, 5, 172, 235, 5, 5, 254, 224, 5, 5, 135, 237, 192, 0, 0, 0, 5, 5, 64, 235, 5, 5, 54, 227, 192, 0, 0, 0, 5, 5, 87, 229, 192, 0, 0, 0, 5, 5, 232, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 234, 5, 5, 26, 239, 5, 5, 28, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 225, 192, 0, 0, 0, 5, 5, 235, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 169, 5, 5, 253, 226, 5, 5, 67, 230, 5, 5, 14, 239, 192, 0, 0, 0, 5, 5, 166, 229, 5, 5, 6, 227, 192, 0, 0, 0, 5, 5, 44, 227, 5, 5, 161, 235, 5, 5, 86, 224, 5, 5, 8, 232, 5, 5, 13, 229, 5, 5, 188, 225, 5, 5, 54, 238, 192, 0, 0, 0, 5, 5, 46, 236, 5, 5, 86, 228, 5, 5, 124, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 228, 5, 5, 18, 239, 5, 5, 179, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 226, 5, 5, 157, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 224, 5, 5, 43, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 230, 5, 5, 44, 239, 5, 5, 191, 232, 192, 0, 0, 0, 5, 5, 138, 235, 192, 0, 0, 0, 5, 5, 42, 229, 5, 5, 109, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 227, 192, 0, 0, 0, 5, 5, 45, 239, 192, 0, 0, 0, 5, 5, 249, 236, 5, 5, 46, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 239, 5, 5, 226, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 226, 5, 5, 87, 224, 5, 5, 40, 239, 5, 5, 255, 224, 5, 5, 36, 229, 192, 0, 0, 0, 5, 5, 130, 229, 5, 5, 69, 236, 5, 5, 206, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 225, 5, 5, 42, 239, 5, 5, 41, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 232, 192, 0, 0, 0, 5, 5, 42, 238, 5, 5, 43, 239, 192, 0, 0, 0, 5, 5, 3, 230, 192, 0, 0, 0, 5, 5, 202, 225, 5, 5, 50, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 228, 192, 0, 0, 0, 5, 5, 79, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 230, 192, 0, 0, 0, 5, 5, 99, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 228, 5, 5, 72, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 239, 192, 0, 0, 0, 5, 5, 224, 231, 192, 0, 0, 0, 5, 5, 57, 229, 5, 5, 219, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 225, 192, 0, 0, 0, 5, 5, 177, 239, 5, 5, 173, 235, 192, 0, 0, 0, 5, 5, 115, 234, 5, 5, 0, 154, 5, 5, 18, 225, 5, 5, 9, 228, 5, 5, 178, 239, 5, 5, 97, 227, 192, 0, 0, 0, 5, 5, 54, 237, 5, 5, 142, 225, 5, 5, 203, 225, 5, 5, 162, 238, 5, 5, 176, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 226, 5, 5, 82, 232, 5, 5, 163, 238, 192, 0, 0, 0, 5, 5, 59, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 239, 5, 5, 164, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 235, 5, 5, 190, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 245, 5, 5, 26, 230, 5, 5, 63, 230, 5, 5, 68, 237, 5, 5, 201, 226, 5, 5, 46, 225, 192, 0, 0, 0, 5, 5, 17, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 231, 5, 5, 229, 229, 192, 0, 0, 0, 5, 5, 78, 224, 192, 0, 0, 0, 5, 5, 165, 238, 5, 5, 34, 227, 192, 0, 0, 0, 5, 5, 149, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 237, 5, 5, 158, 233, 5, 5, 143, 232, 5, 5, 240, 232, 5, 5, 70, 225, 5, 5, 88, 229, 192, 0, 0, 0, 5, 5, 72, 227, 192, 0, 0, 0, 5, 5, 167, 238, 5, 5, 168, 238, 5, 5, 31, 226, 5, 5, 49, 228, 5, 5, 70, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 235, 192, 0, 0, 0, 5, 5, 169, 238, 5, 5, 50, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 227, 5, 5, 113, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 234, 192, 0, 0, 0, 5, 5, 236, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 238, 192, 0, 0, 0, 5, 5, 170, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 220, 192, 0, 0, 0, 5, 5, 21, 229, 192, 0, 0, 0, 5, 5, 65, 233, 5, 5, 100, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 238, 192, 0, 0, 0, 5, 5, 153, 228, 192, 0, 0, 0, 5, 5, 176, 238, 5, 5, 175, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 229, 192, 0, 0, 0, 5, 5, 144, 232, 5, 5, 0, 202, 5, 5, 248, 226, 5, 5, 67, 228, 5, 5, 11, 235, 5, 5, 165, 239, 5, 5, 29, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 225, 5, 5, 10, 238, 5, 5, 112, 231, 5, 5, 167, 228, 5, 5, 166, 239, 5, 5, 167, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 229, 5, 5, 206, 228, 5, 5, 183, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 239, 5, 5, 100, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 236, 192, 0, 0, 0, 5, 5, 233, 230, 192, 0, 0, 0, 5, 5, 223, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 239, 192, 0, 0, 0, 5, 5, 191, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 239, 192, 0, 0, 0, 5, 5, 61, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 239, 5, 5, 23, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 227, 5, 5, 10, 235, 5, 5, 26, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 218, 5, 5, 76, 225, 192, 0, 0, 0, 5, 5, 175, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 239, 192, 0, 0, 0, 5, 5, 207, 240, 5, 5, 23, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 238, 5, 5, 191, 227, 5, 5, 15, 224, 192, 0, 0, 0, 5, 5, 241, 224, 5, 5, 154, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 228, 5, 5, 123, 229, 192, 0, 0, 0, 5, 5, 51, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 231, 5, 5, 122, 232, 5, 5, 81, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 238, 5, 5, 74, 231, 192, 0, 0, 0, 5, 5, 76, 233, 192, 0, 0, 0, 5, 5, 41, 232, 192, 0, 0, 0, 5, 5, 114, 238, 192, 0, 0, 0, 5, 5, 61, 233, 5, 5, 255, 234, 5, 5, 236, 227, 5, 5, 0, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 227, 5, 5, 131, 235, 192, 0, 0, 0, 5, 5, 14, 224, 5, 5, 82, 238, 5, 5, 45, 238, 192, 0, 0, 0, 5, 5, 125, 225, 5, 5, 156, 230, 5, 5, 53, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 224, 192, 0, 0, 0, 5, 5, 115, 224, 192, 0, 0, 0, 5, 5, 67, 224, 5, 5, 22, 241, 5, 5, 108, 237, 5, 5, 52, 229, 5, 5, 84, 230, 5, 5, 161, 238, 5, 5, 89, 230, 192, 0, 0, 0, 5, 5, 160, 238, 5, 5, 236, 234, 192, 0, 0, 0, 5, 5, 106, 239, 192, 0, 0, 0, 5, 5, 216, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 238, 5, 5, 180, 230, 5, 5, 147, 236, 5, 5, 187, 234, 192, 0, 0, 0, 5, 5, 36, 228, 5, 5, 148, 232, 5, 5, 121, 230, 192, 0, 0, 0, 5, 5, 27, 229, 5, 5, 140, 235, 192, 0, 0, 0, 5, 5, 107, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 224, 192, 0, 0, 0, 5, 5, 42, 226, 5, 5, 88, 234, 5, 5, 228, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 229, 192, 0, 0, 0, 5, 5, 237, 235, 5, 5, 23, 236, 5, 5, 147, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 229, 192, 0, 0, 0, 5, 5, 163, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 238, 192, 0, 0, 0, 5, 5, 253, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 235, 5, 5, 149, 238, 192, 0, 0, 0, 5, 5, 150, 238, 5, 5, 59, 235, 192, 0, 0, 0, 5, 5, 20, 225, 5, 5, 243, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 235, 192, 0, 0, 0, 5, 5, 222, 232, 5, 5, 138, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 236, 5, 5, 162, 224, 5, 5, 32, 228, 5, 5, 195, 236, 5, 5, 167, 233, 5, 5, 74, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 233, 192, 0, 0, 0, 5, 5, 129, 232, 5, 5, 122, 233, 5, 5, 66, 224, 5, 5, 198, 235, 192, 0, 0, 0, 5, 5, 149, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 239, 5, 5, 214, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 229, 5, 5, 25, 227, 5, 5, 18, 229, 5, 5, 55, 230, 192, 0, 0, 0, 5, 5, 26, 241, 5, 5, 25, 241, 5, 5, 201, 237, 5, 5, 158, 228, 5, 5, 137, 237, 5, 5, 0, 166, 5, 5, 217, 225, 5, 5, 85, 229, 5, 5, 4, 234, 5, 5, 23, 241, 5, 5, 85, 233, 5, 5, 196, 236, 192, 0, 0, 0, 5, 5, 155, 238, 5, 5, 74, 236, 5, 5, 119, 227, 5, 5, 186, 233, 5, 5, 26, 234, 192, 0, 0, 0, 5, 5, 27, 241, 5, 5, 200, 225, 5, 5, 24, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 236, 192, 0, 0, 0, 5, 5, 238, 224, 5, 5, 235, 226, 192, 0, 0, 0, 5, 5, 30, 241, 192, 0, 0, 0, 5, 5, 129, 227, 5, 5, 26, 228, 5, 5, 205, 225, 192, 0, 0, 0, 5, 5, 102, 234, 5, 5, 3, 231, 5, 5, 165, 227, 5, 5, 222, 229, 5, 5, 124, 234, 5, 5, 106, 235, 5, 5, 28, 241, 5, 5, 61, 235, 192, 0, 0, 0, 5, 5, 104, 230, 5, 5, 29, 241, 5, 5, 152, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 229, 192, 0, 0, 0, 5, 5, 41, 230, 5, 5, 134, 234, 5, 5, 139, 236, 5, 5, 99, 226, 5, 5, 37, 241, 192, 0, 0, 0, 5, 5, 40, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 226, 5, 5, 0, 167, 5, 5, 14, 226, 5, 5, 106, 226, 192, 0, 0, 0, 5, 5, 94, 227, 5, 5, 89, 234, 5, 5, 96, 229, 5, 5, 173, 233, 5, 5, 24, 232, 192, 0, 0, 0, 5, 5, 182, 237, 5, 5, 41, 241, 192, 0, 0, 0, 5, 5, 250, 234, 5, 5, 205, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 235, 5, 5, 47, 225, 192, 0, 0, 0, 5, 5, 223, 234, 5, 5, 162, 227, 192, 0, 0, 0, 5, 5, 249, 234, 192, 0, 0, 0, 5, 5, 241, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 225, 192, 0, 0, 0, 5, 5, 228, 224, 192, 0, 0, 0, 5, 5, 218, 226, 5, 5, 31, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 231, 192, 0, 0, 0, 5, 5, 32, 241, 5, 5, 33, 241, 5, 5, 34, 241, 5, 5, 127, 231, 5, 5, 35, 241, 5, 5, 38, 241, 5, 5, 3, 237, 5, 5, 39, 241, 192, 0, 0, 0, 5, 5, 60, 232, 5, 5, 240, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 231, 192, 0, 0, 0, 5, 5, 46, 241, 192, 0, 0, 0, 5, 5, 51, 241, 192, 0, 0, 0, 5, 5, 234, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 241, 5, 5, 62, 241, 5, 5, 206, 234, 192, 0, 0, 0, 5, 5, 121, 227, 5, 5, 50, 241, 5, 5, 44, 241, 5, 5, 162, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 233, 5, 5, 166, 227, 5, 5, 4, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 229, 192, 0, 0, 0, 5, 5, 42, 241, 192, 0, 0, 0, 5, 5, 49, 241, 192, 0, 0, 0, 5, 5, 156, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 237, 5, 5, 127, 227, 192, 0, 0, 0, 5, 5, 248, 228, 5, 5, 175, 236, 5, 5, 180, 226, 192, 0, 0, 0, 5, 5, 242, 237, 192, 0, 0, 0, 5, 5, 43, 241, 5, 5, 18, 227, 5, 5, 51, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 230, 5, 5, 47, 241, 5, 5, 48, 241, 192, 0, 0, 0, 5, 5, 52, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 241, 5, 5, 73, 241, 192, 0, 0, 0, 5, 5, 59, 241, 5, 5, 58, 241, 5, 5, 55, 238, 5, 5, 71, 241, 5, 5, 72, 241, 5, 5, 135, 243, 5, 5, 59, 236, 5, 5, 54, 241, 192, 0, 0, 0, 5, 5, 53, 229, 192, 0, 0, 0, 5, 5, 45, 237, 192, 0, 0, 0, 5, 5, 84, 229, 5, 5, 56, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 241, 192, 0, 0, 0, 5, 5, 254, 235, 192, 0, 0, 0, 5, 5, 66, 241, 5, 5, 0, 133, 5, 5, 217, 231, 5, 5, 55, 241, 192, 0, 0, 0, 5, 5, 151, 227, 192, 0, 0, 0, 5, 5, 23, 226, 5, 5, 149, 234, 5, 5, 82, 227, 5, 5, 39, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 241, 5, 5, 100, 235, 5, 5, 0, 131, 5, 5, 75, 241, 5, 5, 53, 241, 5, 5, 250, 235, 5, 5, 57, 241, 5, 5, 60, 241, 5, 5, 61, 241, 5, 5, 64, 241, 192, 0, 0, 0, 5, 5, 185, 227, 192, 0, 0, 0, 5, 5, 68, 241, 5, 5, 69, 241, 192, 0, 0, 0, 5, 5, 70, 241, 5, 5, 74, 241, 5, 5, 76, 241, 5, 5, 166, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 226, 5, 5, 122, 231, 5, 5, 78, 241, 5, 5, 26, 233, 5, 5, 240, 229, 5, 5, 42, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 229, 5, 5, 112, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 237, 5, 5, 82, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 224, 192, 0, 0, 0, 5, 5, 145, 227, 5, 5, 80, 241, 192, 0, 0, 0, 5, 5, 169, 239, 192, 0, 0, 0, 5, 5, 26, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 233, 5, 5, 55, 225, 192, 0, 0, 0, 5, 5, 0, 132, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 241, 5, 5, 34, 234, 5, 5, 86, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 241, 192, 0, 0, 0, 5, 5, 83, 241, 5, 5, 84, 241, 5, 5, 205, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 241, 192, 0, 0, 0, 5, 5, 179, 227, 192, 0, 0, 0, 5, 5, 121, 233, 5, 5, 192, 234, 5, 5, 105, 241, 5, 5, 196, 224, 192, 0, 0, 0, 5, 5, 104, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 241, 192, 0, 0, 0, 5, 5, 147, 234, 5, 5, 96, 241, 192, 0, 0, 0, 5, 5, 94, 241, 192, 0, 0, 0, 5, 5, 92, 229, 5, 5, 50, 238, 192, 0, 0, 0, 5, 5, 12, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 241, 5, 5, 0, 127, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 241, 5, 5, 100, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 231, 5, 5, 244, 232, 5, 5, 156, 229, 5, 5, 89, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 231, 192, 0, 0, 0, 5, 5, 144, 238, 5, 5, 93, 241, 5, 5, 92, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 241, 5, 5, 102, 241, 5, 5, 103, 241, 5, 5, 123, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 241, 5, 5, 64, 234, 192, 0, 0, 0, 5, 5, 115, 241, 192, 0, 0, 0, 5, 5, 50, 229, 5, 5, 113, 241, 5, 5, 209, 234, 5, 5, 109, 241, 5, 5, 6, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 229, 5, 5, 112, 241, 5, 5, 159, 227, 5, 5, 98, 227, 5, 5, 107, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 241, 192, 0, 0, 0, 5, 5, 117, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 225, 5, 5, 123, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 235, 5, 5, 122, 227, 192, 0, 0, 0, 5, 5, 114, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 241, 192, 0, 0, 0, 5, 5, 77, 237, 192, 0, 0, 0, 5, 5, 91, 241, 192, 0, 0, 0, 5, 5, 171, 231, 192, 0, 0, 0, 5, 5, 111, 241, 192, 0, 0, 0, 5, 5, 233, 236, 192, 0, 0, 0, 5, 5, 120, 241, 5, 5, 121, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 241, 5, 5, 183, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 241, 5, 5, 140, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 241, 5, 5, 108, 241, 5, 5, 226, 232, 5, 5, 131, 241, 192, 0, 0, 0, 5, 5, 116, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 233, 5, 5, 133, 241, 192, 0, 0, 0, 5, 5, 119, 241, 192, 0, 0, 0, 5, 5, 227, 234, 192, 0, 0, 0, 5, 5, 191, 233, 5, 5, 143, 241, 5, 5, 136, 241, 5, 5, 132, 241, 192, 0, 0, 0, 5, 5, 141, 241, 192, 0, 0, 0, 5, 5, 124, 241, 5, 5, 129, 241, 5, 5, 130, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 241, 5, 5, 138, 241, 192, 0, 0, 0, 5, 5, 142, 241, 192, 0, 0, 0, 5, 5, 125, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 233, 5, 5, 151, 241, 192, 0, 0, 0, 5, 5, 152, 241, 5, 5, 148, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 241, 5, 5, 62, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 241, 192, 0, 0, 0, 5, 5, 182, 226, 5, 5, 145, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 241, 5, 5, 127, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 241, 5, 5, 149, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 241, 5, 5, 154, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 238, 5, 5, 202, 224, 192, 0, 0, 0, 5, 5, 90, 238, 192, 0, 0, 0, 5, 5, 183, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 241, 192, 0, 0, 0, 5, 5, 21, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 241, 5, 5, 159, 241, 5, 5, 65, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 241, 5, 5, 155, 241, 5, 5, 161, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 241, 192, 0, 0, 0, 5, 5, 166, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 232, 5, 5, 120, 238, 5, 5, 61, 237, 5, 5, 168, 241, 5, 5, 103, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 241, 5, 5, 170, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 241, 5, 5, 165, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 227, 5, 5, 67, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 232, 5, 5, 192, 233, 192, 0, 0, 0, 5, 5, 176, 241, 5, 5, 232, 227, 5, 5, 130, 238, 5, 5, 134, 236, 5, 5, 177, 241, 5, 5, 127, 234, 192, 0, 0, 0, 5, 5, 18, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 241, 192, 0, 0, 0, 5, 5, 2, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 229, 5, 5, 75, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 234, 5, 5, 178, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 241, 5, 5, 32, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 227, 5, 5, 118, 234, 5, 5, 181, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 231, 5, 5, 182, 241, 192, 0, 0, 0, 5, 5, 4, 237, 192, 0, 0, 0, 5, 5, 134, 232, 5, 5, 184, 241, 5, 5, 183, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 237, 5, 5, 191, 239, 5, 5, 193, 239, 192, 0, 0, 0, 5, 5, 192, 239, 5, 5, 59, 227, 5, 5, 196, 239, 5, 5, 197, 239, 5, 5, 176, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 239, 5, 5, 187, 224, 5, 5, 199, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 228, 192, 0, 0, 0, 5, 5, 197, 237, 192, 0, 0, 0, 5, 5, 200, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 226, 192, 0, 0, 0, 5, 5, 184, 239, 5, 5, 10, 234, 5, 5, 62, 229, 5, 5, 197, 227, 5, 5, 102, 238, 5, 5, 95, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 228, 5, 5, 15, 234, 5, 5, 198, 239, 5, 5, 0, 177, 5, 5, 5, 235, 5, 5, 110, 226, 5, 5, 39, 238, 5, 5, 228, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 229, 192, 0, 0, 0, 5, 5, 21, 234, 192, 0, 0, 0, 5, 5, 207, 239, 5, 5, 201, 239, 5, 5, 220, 231, 5, 5, 203, 239, 192, 0, 0, 0, 5, 5, 208, 239, 192, 0, 0, 0, 5, 5, 187, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 239, 5, 5, 76, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 239, 5, 5, 205, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 225, 5, 5, 150, 229, 5, 5, 74, 230, 5, 5, 202, 239, 5, 5, 204, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 235, 5, 5, 213, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 229, 5, 5, 218, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 239, 5, 5, 186, 239, 5, 5, 12, 227, 5, 5, 251, 236, 5, 5, 212, 239, 192, 0, 0, 0, 5, 5, 93, 229, 5, 5, 216, 239, 192, 0, 0, 0, 5, 5, 185, 239, 192, 0, 0, 0, 5, 5, 190, 225, 192, 0, 0, 0, 5, 5, 211, 239, 5, 5, 113, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 239, 192, 0, 0, 0, 5, 5, 217, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 226, 5, 5, 0, 129, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 224, 5, 5, 215, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 239, 192, 0, 0, 0, 5, 5, 244, 231, 5, 5, 220, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 239, 5, 5, 223, 239, 5, 5, 222, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 239, 192, 0, 0, 0, 5, 5, 228, 236, 5, 5, 118, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 239, 192, 0, 0, 0, 5, 5, 227, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 239, 5, 5, 164, 231, 5, 5, 11, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 239, 192, 0, 0, 0, 5, 5, 31, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 226, 5, 5, 102, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 239, 192, 0, 0, 0, 5, 5, 232, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 239, 192, 0, 0, 0, 5, 5, 239, 239, 5, 5, 2, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 233, 5, 5, 199, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 234, 5, 5, 241, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 239, 192, 0, 0, 0, 5, 5, 245, 239, 5, 5, 242, 239, 192, 0, 0, 0, 5, 5, 223, 228, 192, 0, 0, 0, 5, 5, 153, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 233, 5, 5, 31, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 232, 5, 5, 244, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 237, 5, 5, 188, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 231, 5, 5, 13, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 233, 5, 5, 169, 232, 192, 0, 0, 0, 5, 5, 33, 238, 192, 0, 0, 0, 5, 5, 57, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 242, 192, 0, 0, 0, 5, 5, 32, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 226, 192, 0, 0, 0, 5, 5, 60, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 239, 5, 5, 30, 235, 5, 5, 156, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 239, 5, 5, 24, 226, 192, 0, 0, 0, 5, 5, 77, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 242, 5, 5, 55, 242, 192, 0, 0, 0, 5, 5, 109, 225, 192, 0, 0, 0, 5, 5, 72, 234, 5, 5, 7, 234, 5, 5, 136, 226, 192, 0, 0, 0, 5, 5, 126, 238, 5, 5, 33, 236, 5, 5, 109, 227, 192, 0, 0, 0, 5, 5, 69, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 236, 5, 5, 112, 229, 5, 5, 64, 228, 5, 5, 25, 226, 192, 0, 0, 0, 5, 5, 201, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 240, 5, 5, 164, 239, 192, 0, 0, 0, 5, 5, 15, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 232, 5, 5, 52, 231, 5, 5, 132, 226, 192, 0, 0, 0, 5, 5, 114, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 229, 5, 5, 78, 238, 192, 0, 0, 0, 5, 5, 25, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 224, 5, 5, 204, 240, 5, 5, 127, 228, 5, 5, 51, 234, 5, 5, 206, 240, 192, 0, 0, 0, 5, 5, 205, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 225, 192, 0, 0, 0, 5, 5, 29, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 231, 5, 5, 111, 231, 192, 0, 0, 0, 5, 5, 50, 231, 192, 0, 0, 0, 5, 5, 89, 228, 5, 5, 249, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 243, 5, 5, 199, 232, 5, 5, 150, 243, 5, 5, 183, 234, 192, 0, 0, 0, 5, 5, 31, 238, 5, 5, 178, 226, 5, 5, 144, 230, 192, 0, 0, 0, 5, 5, 177, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 226, 5, 5, 59, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 236, 5, 5, 155, 243, 192, 0, 0, 0, 5, 5, 244, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 243, 5, 5, 159, 243, 5, 5, 145, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 226, 5, 5, 152, 243, 5, 5, 153, 243, 5, 5, 157, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 231, 5, 5, 162, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 230, 192, 0, 0, 0, 5, 5, 2, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 243, 5, 5, 90, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 228, 5, 5, 23, 227, 5, 5, 160, 243, 5, 5, 171, 235, 5, 5, 199, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 243, 5, 5, 170, 243, 192, 0, 0, 0, 5, 5, 58, 236, 192, 0, 0, 0, 5, 5, 120, 228, 5, 5, 167, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 243, 192, 0, 0, 0, 5, 5, 187, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 243, 192, 0, 0, 0, 5, 5, 138, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 238, 192, 0, 0, 0, 5, 5, 184, 234, 192, 0, 0, 0, 5, 5, 152, 234, 5, 5, 78, 230, 5, 5, 165, 243, 5, 5, 166, 243, 5, 5, 141, 228, 5, 5, 168, 243, 5, 5, 173, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 243, 192, 0, 0, 0, 5, 5, 178, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 229, 5, 5, 47, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 243, 192, 0, 0, 0, 5, 5, 40, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 236, 5, 5, 174, 243, 192, 0, 0, 0, 5, 5, 175, 243, 192, 0, 0, 0, 5, 5, 130, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 234, 5, 5, 181, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 236, 5, 5, 185, 243, 5, 5, 52, 233, 5, 5, 189, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 230, 5, 5, 188, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 243, 5, 5, 41, 235, 192, 0, 0, 0, 5, 5, 200, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 228, 5, 5, 231, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 243, 192, 0, 0, 0, 5, 5, 195, 243, 192, 0, 0, 0, 5, 5, 200, 243, 192, 0, 0, 0, 5, 5, 202, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 243, 5, 5, 173, 225, 192, 0, 0, 0, 5, 5, 198, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 231, 192, 0, 0, 0, 5, 5, 192, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 243, 5, 5, 205, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 243, 192, 0, 0, 0, 5, 5, 36, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 238, 5, 5, 215, 243, 192, 0, 0, 0, 5, 5, 216, 243, 5, 5, 99, 229, 5, 5, 34, 237, 192, 0, 0, 0, 5, 5, 68, 238, 5, 5, 96, 225, 5, 5, 235, 233, 5, 5, 212, 243, 5, 5, 143, 238, 5, 5, 60, 238, 5, 5, 120, 235, 192, 0, 0, 0, 5, 5, 213, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 243, 5, 5, 46, 228, 5, 5, 22, 227, 5, 5, 213, 243, 5, 5, 219, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 227, 5, 5, 119, 230, 192, 0, 0, 0, 5, 5, 119, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 233, 5, 5, 138, 243, 192, 0, 0, 0, 5, 5, 214, 243, 192, 0, 0, 0, 5, 5, 139, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 243, 5, 5, 100, 231, 192, 0, 0, 0, 5, 5, 248, 233, 5, 5, 79, 243, 5, 5, 92, 237, 192, 0, 0, 0, 5, 5, 224, 236, 5, 5, 118, 233, 5, 5, 0, 144, 192, 0, 0, 0, 5, 5, 202, 233, 5, 5, 164, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 243, 192, 0, 0, 0, 5, 5, 80, 243, 192, 0, 0, 0, 5, 5, 72, 238, 5, 5, 45, 227, 5, 5, 244, 237, 5, 5, 222, 225, 5, 5, 170, 234, 5, 5, 95, 236, 5, 5, 4, 224, 5, 5, 83, 233, 192, 0, 0, 0, 5, 5, 4, 225, 5, 5, 51, 233, 5, 5, 89, 229, 5, 5, 209, 235, 5, 5, 112, 233, 5, 5, 82, 243, 5, 5, 210, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 235, 5, 5, 254, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 227, 5, 5, 29, 236, 5, 5, 116, 235, 5, 5, 66, 228, 192, 0, 0, 0, 5, 5, 83, 243, 5, 5, 189, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 229, 5, 5, 84, 224, 5, 5, 218, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 228, 192, 0, 0, 0, 5, 5, 52, 228, 5, 5, 142, 236, 5, 5, 226, 230, 5, 5, 104, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 230, 192, 0, 0, 0, 5, 5, 96, 227, 5, 5, 242, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 232, 5, 5, 21, 231, 5, 5, 168, 224, 192, 0, 0, 0, 5, 5, 37, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 243, 5, 5, 18, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 243, 192, 0, 0, 0, 5, 5, 90, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 225, 5, 5, 12, 226, 5, 5, 190, 233, 5, 5, 228, 235, 5, 5, 148, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 233, 192, 0, 0, 0, 5, 5, 120, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 233, 192, 0, 0, 0, 5, 5, 121, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 234, 5, 5, 94, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 235, 192, 0, 0, 0, 5, 5, 25, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 226, 5, 5, 210, 243, 5, 5, 82, 228, 192, 0, 0, 0, 5, 5, 216, 224, 192, 0, 0, 0, 5, 5, 16, 233, 192, 0, 0, 0, 5, 5, 211, 243, 5, 5, 188, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 240, 192, 0, 0, 0, 5, 5, 186, 236, 5, 5, 209, 240, 192, 0, 0, 0, 5, 5, 54, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 233, 5, 5, 144, 236, 5, 5, 249, 224, 5, 5, 134, 243, 5, 5, 71, 231, 5, 5, 205, 228, 5, 5, 202, 234, 5, 5, 88, 231, 5, 5, 4, 229, 5, 5, 202, 231, 5, 5, 159, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 232, 5, 5, 71, 234, 5, 5, 186, 228, 5, 5, 245, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 233, 5, 5, 227, 231, 5, 5, 136, 243, 5, 5, 146, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 233, 192, 0, 0, 0, 5, 5, 122, 234, 5, 5, 109, 230, 192, 0, 0, 0, 5, 5, 139, 243, 192, 0, 0, 0, 5, 5, 108, 230, 5, 5, 140, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 243, 5, 5, 135, 234, 192, 0, 0, 0, 5, 5, 251, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 236, 5, 5, 200, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 236, 192, 0, 0, 0, 5, 5, 230, 233, 5, 5, 21, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 241, 5, 5, 201, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 241, 5, 5, 207, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 224, 192, 0, 0, 0, 5, 5, 203, 241, 5, 5, 206, 226, 5, 5, 205, 241, 5, 5, 206, 241, 5, 5, 208, 241, 5, 5, 146, 225, 5, 5, 209, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 241, 5, 5, 216, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 236, 192, 0, 0, 0, 5, 5, 214, 241, 5, 5, 213, 241, 5, 5, 53, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 241, 192, 0, 0, 0, 5, 5, 18, 237, 192, 0, 0, 0, 5, 5, 210, 241, 192, 0, 0, 0, 5, 5, 218, 241, 5, 5, 0, 148, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 241, 192, 0, 0, 0, 5, 5, 133, 229, 192, 0, 0, 0, 5, 5, 217, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 241, 5, 5, 223, 241, 5, 5, 117, 230, 192, 0, 0, 0, 5, 5, 35, 226, 192, 0, 0, 0, 5, 5, 234, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 241, 5, 5, 225, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 241, 5, 5, 3, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 225, 192, 0, 0, 0, 5, 5, 246, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 241, 192, 0, 0, 0, 5, 5, 227, 241, 5, 5, 226, 241, 192, 0, 0, 0, 5, 5, 30, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 237, 5, 5, 228, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 241, 5, 5, 234, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 241, 5, 5, 242, 241, 5, 5, 56, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 241, 192, 0, 0, 0, 5, 5, 240, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 241, 192, 0, 0, 0, 5, 5, 241, 241, 5, 5, 237, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 244, 192, 0, 0, 0, 5, 5, 34, 225, 5, 5, 235, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 235, 5, 5, 204, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 226, 5, 5, 97, 238, 5, 5, 75, 232, 5, 5, 12, 229, 5, 5, 2, 227, 192, 0, 0, 0, 5, 5, 239, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 224, 5, 5, 183, 239, 192, 0, 0, 0, 5, 5, 42, 228, 5, 5, 101, 236, 5, 5, 196, 233, 5, 5, 0, 171, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 239, 5, 5, 187, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 224, 5, 5, 110, 233, 5, 5, 120, 224, 192, 0, 0, 0, 5, 5, 163, 233, 5, 5, 65, 226, 192, 0, 0, 0, 5, 5, 68, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 241, 5, 5, 122, 237, 5, 5, 189, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 241, 5, 5, 133, 231, 5, 5, 87, 234, 192, 0, 0, 0, 5, 5, 249, 229, 5, 5, 187, 241, 5, 5, 241, 237, 5, 5, 107, 224, 5, 5, 210, 237, 5, 5, 179, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 225, 5, 5, 175, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 235, 5, 5, 0, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 241, 5, 5, 189, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 241, 5, 5, 192, 241, 5, 5, 183, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 230, 192, 0, 0, 0, 5, 5, 194, 241, 5, 5, 143, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 241, 5, 5, 33, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 241, 5, 5, 43, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 226, 5, 5, 223, 231, 5, 5, 80, 231, 192, 0, 0, 0, 5, 5, 94, 224, 192, 0, 0, 0, 5, 5, 168, 235, 192, 0, 0, 0, 5, 5, 53, 244, 5, 5, 211, 227, 5, 5, 201, 236, 5, 5, 182, 236, 192, 0, 0, 0, 5, 5, 55, 227, 5, 5, 79, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 238, 192, 0, 0, 0, 5, 5, 110, 232, 5, 5, 49, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 225, 5, 5, 81, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 235, 5, 5, 86, 230, 5, 5, 80, 242, 192, 0, 0, 0, 5, 5, 110, 229, 5, 5, 152, 236, 5, 5, 175, 225, 5, 5, 82, 242, 5, 5, 194, 225, 192, 0, 0, 0, 5, 5, 243, 230, 5, 5, 241, 226, 192, 0, 0, 0, 5, 5, 162, 226, 192, 0, 0, 0, 5, 5, 151, 231, 5, 5, 101, 226, 5, 5, 84, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 242, 5, 5, 254, 225, 5, 5, 103, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 242, 192, 0, 0, 0, 5, 5, 86, 242, 5, 5, 154, 233, 5, 5, 65, 229, 5, 5, 171, 236, 5, 5, 85, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 229, 5, 5, 178, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 242, 5, 5, 90, 242, 5, 5, 148, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 242, 192, 0, 0, 0, 5, 5, 94, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 239, 192, 0, 0, 0, 5, 5, 9, 236, 5, 5, 92, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 240, 5, 5, 113, 238, 5, 5, 54, 229, 5, 5, 181, 239, 5, 5, 128, 236, 5, 5, 30, 232, 5, 5, 110, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 240, 192, 0, 0, 0, 5, 5, 53, 224, 5, 5, 18, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 233, 192, 0, 0, 0, 5, 5, 21, 241, 192, 0, 0, 0, 5, 5, 255, 226, 192, 0, 0, 0, 5, 5, 145, 236, 192, 0, 0, 0, 5, 5, 153, 226, 5, 5, 157, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 225, 5, 5, 117, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 230, 5, 5, 74, 235, 5, 5, 177, 227, 192, 0, 0, 0, 5, 5, 143, 243, 5, 5, 142, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 243, 192, 0, 0, 0, 5, 5, 158, 234, 192, 0, 0, 0, 5, 5, 213, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 225, 5, 5, 65, 232, 192, 0, 0, 0, 5, 5, 146, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 243, 192, 0, 0, 0, 5, 5, 61, 227, 5, 5, 137, 225, 192, 0, 0, 0, 5, 5, 100, 230, 5, 5, 132, 243, 5, 5, 131, 243, 5, 5, 133, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 242, 5, 5, 164, 235, 192, 0, 0, 0, 5, 5, 104, 234, 192, 0, 0, 0, 5, 5, 27, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 224, 5, 5, 73, 224, 192, 0, 0, 0, 5, 5, 80, 224, 5, 5, 177, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 237, 5, 5, 162, 236, 192, 0, 0, 0, 5, 5, 252, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 241, 192, 0, 0, 0, 5, 5, 111, 236, 192, 0, 0, 0, 5, 5, 211, 224, 5, 5, 41, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 234, 5, 5, 167, 237, 5, 5, 254, 241, 192, 0, 0, 0, 5, 5, 226, 228, 5, 5, 119, 225, 192, 0, 0, 0, 5, 5, 255, 241, 5, 5, 142, 227, 5, 5, 2, 242, 5, 5, 194, 227, 5, 5, 113, 230, 5, 5, 3, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 235, 192, 0, 0, 0, 5, 5, 119, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 242, 192, 0, 0, 0, 5, 5, 155, 225, 5, 5, 139, 231, 5, 5, 5, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 242, 192, 0, 0, 0, 5, 5, 225, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 242, 5, 5, 186, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 242, 192, 0, 0, 0, 5, 5, 154, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 242, 5, 5, 229, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 235, 5, 5, 96, 242, 5, 5, 54, 224, 5, 5, 120, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 228, 5, 5, 172, 232, 192, 0, 0, 0, 5, 5, 99, 242, 5, 5, 50, 246, 5, 5, 49, 246, 5, 5, 19, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 242, 5, 5, 204, 237, 5, 5, 182, 234, 5, 5, 172, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 237, 5, 5, 103, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 242, 5, 5, 120, 229, 192, 0, 0, 0, 5, 5, 109, 242, 5, 5, 101, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 235, 192, 0, 0, 0, 5, 5, 168, 227, 5, 5, 105, 242, 5, 5, 116, 226, 5, 5, 195, 227, 5, 5, 8, 234, 5, 5, 199, 233, 5, 5, 100, 242, 5, 5, 102, 242, 5, 5, 106, 242, 5, 5, 107, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 242, 5, 5, 71, 237, 5, 5, 115, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 231, 192, 0, 0, 0, 5, 5, 170, 237, 5, 5, 134, 231, 5, 5, 124, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 242, 192, 0, 0, 0, 5, 5, 114, 242, 5, 5, 247, 229, 5, 5, 184, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 225, 5, 5, 121, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 228, 5, 5, 113, 242, 5, 5, 170, 235, 5, 5, 11, 237, 5, 5, 117, 242, 5, 5, 42, 227, 5, 5, 118, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 238, 5, 5, 51, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 242, 192, 0, 0, 0, 5, 5, 55, 246, 5, 5, 126, 242, 5, 5, 111, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 229, 192, 0, 0, 0, 5, 5, 224, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 229, 192, 0, 0, 0, 5, 5, 149, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 246, 5, 5, 53, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 227, 5, 5, 57, 246, 5, 5, 202, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 246, 5, 5, 144, 227, 5, 5, 47, 226, 5, 5, 127, 242, 5, 5, 232, 225, 5, 5, 77, 234, 5, 5, 249, 226, 192, 0, 0, 0, 5, 5, 26, 235, 5, 5, 35, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 226, 192, 0, 0, 0, 5, 5, 122, 242, 5, 5, 123, 242, 5, 5, 125, 242, 5, 5, 124, 242, 5, 5, 59, 231, 5, 5, 128, 242, 192, 0, 0, 0, 5, 5, 179, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 242, 5, 5, 70, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 242, 5, 5, 105, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 242, 192, 0, 0, 0, 5, 5, 234, 227, 192, 0, 0, 0, 5, 5, 129, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 242, 5, 5, 136, 242, 192, 0, 0, 0, 5, 5, 132, 242, 192, 0, 0, 0, 5, 5, 12, 235, 5, 5, 184, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 237, 192, 0, 0, 0, 5, 5, 96, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 246, 5, 5, 210, 235, 5, 5, 131, 242, 192, 0, 0, 0, 5, 5, 251, 230, 192, 0, 0, 0, 5, 5, 139, 242, 5, 5, 13, 224, 192, 0, 0, 0, 5, 5, 144, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 242, 5, 5, 151, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 232, 5, 5, 142, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 228, 5, 5, 169, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 242, 192, 0, 0, 0, 5, 5, 143, 242, 192, 0, 0, 0, 5, 5, 31, 235, 5, 5, 140, 242, 192, 0, 0, 0, 5, 5, 193, 234, 5, 5, 237, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 225, 5, 5, 19, 229, 5, 5, 142, 224, 5, 5, 231, 224, 192, 0, 0, 0, 5, 5, 23, 224, 5, 5, 137, 242, 5, 5, 141, 224, 5, 5, 132, 225, 5, 5, 51, 227, 5, 5, 32, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 235, 5, 5, 149, 242, 192, 0, 0, 0, 5, 5, 220, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 232, 5, 5, 160, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 242, 5, 5, 10, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 246, 192, 0, 0, 0, 5, 5, 236, 236, 5, 5, 214, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 246, 5, 5, 151, 242, 5, 5, 118, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 226, 5, 5, 145, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 242, 5, 5, 117, 226, 192, 0, 0, 0, 5, 5, 146, 242, 5, 5, 140, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 225, 192, 0, 0, 0, 5, 5, 153, 242, 192, 0, 0, 0, 5, 5, 213, 227, 192, 0, 0, 0, 5, 5, 55, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 227, 5, 5, 58, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 242, 192, 0, 0, 0, 5, 5, 66, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 242, 5, 5, 91, 227, 5, 5, 62, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 235, 192, 0, 0, 0, 5, 5, 0, 161, 5, 5, 64, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 246, 5, 5, 171, 229, 192, 0, 0, 0, 5, 5, 159, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 246, 192, 0, 0, 0, 5, 5, 222, 226, 192, 0, 0, 0, 5, 5, 6, 235, 5, 5, 29, 245, 5, 5, 189, 235, 5, 5, 151, 233, 5, 5, 182, 232, 5, 5, 47, 235, 5, 5, 227, 224, 5, 5, 234, 234, 5, 5, 178, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 242, 5, 5, 4, 228, 5, 5, 30, 245, 5, 5, 109, 237, 192, 0, 0, 0, 5, 5, 255, 231, 5, 5, 31, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 245, 192, 0, 0, 0, 5, 5, 34, 245, 5, 5, 33, 245, 192, 0, 0, 0, 5, 5, 36, 245, 5, 5, 35, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 228, 192, 0, 0, 0, 5, 5, 37, 245, 192, 0, 0, 0, 5, 5, 102, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 225, 5, 5, 113, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 246, 5, 5, 25, 246, 5, 5, 85, 226, 5, 5, 245, 233, 5, 5, 64, 224, 192, 0, 0, 0, 5, 5, 27, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 233, 5, 5, 28, 246, 5, 5, 29, 246, 192, 0, 0, 0, 5, 5, 116, 233, 5, 5, 212, 240, 5, 5, 129, 224, 5, 5, 76, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 231, 5, 5, 0, 165, 5, 5, 108, 225, 5, 5, 179, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 232, 5, 5, 194, 237, 192, 0, 0, 0, 5, 5, 147, 229, 5, 5, 213, 240, 5, 5, 230, 232, 5, 5, 37, 236, 5, 5, 104, 231, 5, 5, 0, 197, 5, 5, 208, 224, 5, 5, 165, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 192, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 226, 192, 0, 0, 0, 5, 5, 130, 237, 5, 5, 234, 224, 5, 5, 44, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 229, 5, 5, 0, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 236, 5, 5, 129, 233, 5, 5, 19, 238, 192, 0, 0, 0, 5, 5, 112, 234, 5, 5, 236, 225, 5, 5, 69, 229, 5, 5, 139, 237, 192, 0, 0, 0, 5, 5, 155, 226, 5, 5, 155, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 240, 5, 5, 101, 229, 5, 5, 125, 230, 5, 5, 66, 232, 192, 0, 0, 0, 5, 5, 180, 227, 5, 5, 6, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 231, 5, 5, 3, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 231, 192, 0, 0, 0, 5, 5, 215, 240, 5, 5, 238, 235, 5, 5, 5, 225, 192, 0, 0, 0, 5, 5, 248, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 226, 192, 0, 0, 0, 5, 5, 4, 238, 5, 5, 123, 225, 5, 5, 172, 224, 5, 5, 26, 231, 5, 5, 79, 231, 5, 5, 151, 229, 5, 5, 216, 240, 192, 0, 0, 0, 5, 5, 0, 198, 5, 5, 136, 231, 5, 5, 42, 230, 192, 0, 0, 0, 5, 5, 41, 227, 192, 0, 0, 0, 5, 5, 10, 229, 5, 5, 146, 234, 5, 5, 0, 172, 192, 0, 0, 0, 5, 5, 137, 234, 5, 5, 218, 240, 5, 5, 252, 228, 5, 5, 44, 238, 5, 5, 217, 240, 5, 5, 128, 237, 5, 5, 0, 187, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 230, 5, 5, 96, 228, 192, 0, 0, 0, 5, 5, 167, 236, 5, 5, 163, 229, 5, 5, 101, 231, 5, 5, 98, 224, 5, 5, 67, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 229, 5, 5, 97, 233, 5, 5, 219, 240, 5, 5, 172, 237, 192, 0, 0, 0, 5, 5, 4, 227, 192, 0, 0, 0, 5, 5, 141, 232, 5, 5, 165, 233, 192, 0, 0, 0, 5, 5, 221, 240, 5, 5, 73, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 231, 5, 5, 21, 238, 5, 5, 78, 233, 5, 5, 41, 231, 192, 0, 0, 0, 5, 5, 240, 224, 5, 5, 38, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 237, 5, 5, 114, 245, 5, 5, 0, 146, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 237, 5, 5, 118, 230, 192, 0, 0, 0, 5, 5, 230, 234, 5, 5, 252, 233, 5, 5, 132, 235, 5, 5, 57, 231, 5, 5, 138, 225, 5, 5, 220, 240, 5, 5, 165, 237, 5, 5, 39, 228, 5, 5, 227, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 130, 192, 0, 0, 0, 5, 5, 118, 231, 5, 5, 102, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 240, 5, 5, 96, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 234, 5, 5, 224, 240, 192, 0, 0, 0, 5, 5, 107, 234, 5, 5, 144, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 238, 192, 0, 0, 0, 5, 5, 223, 240, 5, 5, 0, 164, 5, 5, 102, 227, 5, 5, 19, 233, 5, 5, 89, 231, 5, 5, 22, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 229, 5, 5, 236, 233, 192, 0, 0, 0, 5, 5, 97, 228, 5, 5, 203, 227, 5, 5, 143, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 231, 192, 0, 0, 0, 5, 5, 234, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 240, 5, 5, 11, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 225, 5, 5, 171, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 240, 5, 5, 83, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 235, 192, 0, 0, 0, 5, 5, 184, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 226, 5, 5, 120, 233, 5, 5, 204, 225, 5, 5, 233, 240, 192, 0, 0, 0, 5, 5, 118, 237, 192, 0, 0, 0, 5, 5, 229, 240, 5, 5, 41, 234, 5, 5, 34, 232, 192, 0, 0, 0, 5, 5, 137, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 236, 192, 0, 0, 0, 5, 5, 32, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 230, 192, 0, 0, 0, 5, 5, 25, 234, 5, 5, 210, 224, 192, 0, 0, 0, 5, 5, 161, 228, 192, 0, 0, 0, 5, 5, 100, 229, 5, 5, 128, 234, 5, 5, 16, 236, 5, 5, 101, 225, 192, 0, 0, 0, 5, 5, 232, 240, 5, 5, 225, 240, 5, 5, 235, 240, 192, 0, 0, 0, 5, 5, 116, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 230, 5, 5, 230, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 237, 5, 5, 127, 225, 192, 0, 0, 0, 5, 5, 176, 224, 192, 0, 0, 0, 5, 5, 236, 240, 192, 0, 0, 0, 5, 5, 246, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 240, 192, 0, 0, 0, 5, 5, 245, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 238, 5, 5, 243, 240, 5, 5, 238, 230, 5, 5, 61, 234, 192, 0, 0, 0, 5, 5, 161, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 240, 192, 0, 0, 0, 5, 5, 239, 240, 5, 5, 237, 234, 192, 0, 0, 0, 5, 5, 33, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 229, 5, 5, 234, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 240, 192, 0, 0, 0, 5, 5, 254, 236, 192, 0, 0, 0, 5, 5, 226, 240, 192, 0, 0, 0, 5, 5, 238, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 229, 192, 0, 0, 0, 5, 5, 240, 240, 5, 5, 175, 224, 5, 5, 221, 226, 5, 5, 79, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 240, 5, 5, 252, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 224, 5, 5, 30, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 225, 5, 5, 228, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 240, 5, 5, 205, 233, 192, 0, 0, 0, 5, 5, 215, 226, 192, 0, 0, 0, 5, 5, 251, 240, 5, 5, 254, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 234, 192, 0, 0, 0, 5, 5, 0, 191, 5, 5, 104, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 224, 192, 0, 0, 0, 5, 5, 117, 245, 192, 0, 0, 0, 5, 5, 248, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 233, 5, 5, 247, 240, 5, 5, 0, 184, 5, 5, 53, 236, 5, 5, 85, 224, 192, 0, 0, 0, 5, 5, 11, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 240, 192, 0, 0, 0, 5, 5, 160, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 225, 192, 0, 0, 0, 5, 5, 12, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 231, 5, 5, 7, 231, 5, 5, 5, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 230, 192, 0, 0, 0, 5, 5, 2, 241, 5, 5, 29, 229, 192, 0, 0, 0, 5, 5, 212, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 224, 5, 5, 214, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 233, 5, 5, 4, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 232, 5, 5, 97, 224, 5, 5, 99, 225, 192, 0, 0, 0, 5, 5, 25, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 231, 192, 0, 0, 0, 5, 5, 6, 241, 5, 5, 7, 241, 192, 0, 0, 0, 5, 5, 9, 241, 192, 0, 0, 0, 5, 5, 101, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 241, 192, 0, 0, 0, 5, 5, 200, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 224, 5, 5, 104, 232, 192, 0, 0, 0, 5, 5, 11, 241, 192, 0, 0, 0, 5, 5, 94, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 241, 5, 5, 118, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 241, 192, 0, 0, 0, 5, 5, 13, 241, 192, 0, 0, 0, 5, 5, 125, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 241, 5, 5, 181, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 245, 5, 5, 138, 245, 5, 5, 115, 233, 192, 0, 0, 0, 5, 5, 200, 238, 5, 5, 184, 226, 192, 0, 0, 0, 5, 5, 247, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 226, 5, 5, 174, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 235, 5, 5, 245, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 230, 192, 0, 0, 0, 5, 5, 244, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 245, 5, 5, 0, 155, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 228, 192, 0, 0, 0, 5, 5, 250, 229, 192, 0, 0, 0, 5, 5, 52, 224, 5, 5, 194, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 226, 5, 5, 224, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 245, 5, 5, 224, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 233, 192, 0, 0, 0, 5, 5, 69, 232, 192, 0, 0, 0, 5, 5, 171, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 237, 5, 5, 81, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 245, 5, 5, 0, 189, 192, 0, 0, 0, 5, 5, 231, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 225, 192, 0, 0, 0, 5, 5, 25, 230, 192, 0, 0, 0, 5, 5, 130, 249, 5, 5, 135, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 237, 192, 0, 0, 0, 5, 5, 235, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 228, 5, 5, 252, 224, 192, 0, 0, 0, 5, 5, 159, 226, 192, 0, 0, 0, 5, 5, 102, 237, 192, 0, 0, 0, 5, 5, 79, 246, 192, 0, 0, 0, 5, 5, 7, 226, 192, 0, 0, 0, 5, 5, 181, 233, 5, 5, 152, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 245, 5, 5, 71, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 231, 192, 0, 0, 0, 5, 5, 235, 245, 5, 5, 234, 245, 5, 5, 107, 230, 5, 5, 233, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 235, 5, 5, 236, 245, 192, 0, 0, 0, 5, 5, 237, 245, 5, 5, 83, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 245, 5, 5, 16, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 234, 192, 0, 0, 0, 5, 5, 56, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 232, 5, 5, 129, 225, 5, 5, 245, 228, 5, 5, 202, 237, 5, 5, 57, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 235, 5, 5, 199, 235, 5, 5, 46, 245, 5, 5, 47, 245, 5, 5, 48, 245, 5, 5, 103, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 233, 5, 5, 129, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 245, 192, 0, 0, 0, 5, 5, 0, 152, 5, 5, 52, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 245, 192, 0, 0, 0, 5, 5, 185, 224, 192, 0, 0, 0, 5, 5, 253, 230, 5, 5, 248, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 236, 5, 5, 14, 235, 5, 5, 53, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 245, 192, 0, 0, 0, 5, 5, 157, 235, 5, 5, 165, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 225, 192, 0, 0, 0, 5, 5, 199, 230, 5, 5, 96, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 237, 192, 0, 0, 0, 5, 5, 101, 233, 192, 0, 0, 0, 5, 5, 59, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 245, 5, 5, 61, 245, 5, 5, 60, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 237, 192, 0, 0, 0, 5, 5, 76, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 245, 192, 0, 0, 0, 5, 5, 222, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 228, 5, 5, 14, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 232, 5, 5, 118, 235, 5, 5, 64, 245, 5, 5, 32, 237, 5, 5, 67, 245, 5, 5, 69, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 245, 192, 0, 0, 0, 5, 5, 68, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 235, 192, 0, 0, 0, 5, 5, 238, 227, 192, 0, 0, 0, 5, 5, 217, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 231, 5, 5, 220, 228, 5, 5, 20, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 232, 192, 0, 0, 0, 5, 5, 211, 228, 5, 5, 70, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 237, 192, 0, 0, 0, 5, 5, 71, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 231, 5, 5, 0, 147, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 245, 5, 5, 9, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 233, 192, 0, 0, 0, 5, 5, 76, 245, 5, 5, 77, 245, 5, 5, 250, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 237, 192, 0, 0, 0, 5, 5, 124, 232, 5, 5, 75, 236, 5, 5, 240, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 245, 192, 0, 0, 0, 5, 5, 152, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 230, 192, 0, 0, 0, 5, 5, 74, 237, 5, 5, 66, 234, 5, 5, 92, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 237, 5, 5, 194, 236, 5, 5, 151, 245, 5, 5, 183, 231, 192, 0, 0, 0, 5, 5, 148, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 233, 5, 5, 178, 245, 192, 0, 0, 0, 5, 5, 180, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 234, 192, 0, 0, 0, 5, 5, 201, 224, 192, 0, 0, 0, 5, 5, 252, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 245, 192, 0, 0, 0, 5, 5, 36, 231, 192, 0, 0, 0, 5, 5, 204, 234, 5, 5, 15, 231, 5, 5, 28, 224, 5, 5, 79, 237, 192, 0, 0, 0, 5, 5, 147, 233, 192, 0, 0, 0, 5, 5, 250, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 231, 5, 5, 28, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 232, 192, 0, 0, 0, 5, 5, 37, 237, 5, 5, 136, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 226, 192, 0, 0, 0, 5, 5, 118, 244, 5, 5, 250, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 244, 192, 0, 0, 0, 5, 5, 216, 229, 5, 5, 169, 235, 5, 5, 128, 224, 5, 5, 95, 225, 192, 0, 0, 0, 5, 5, 116, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 225, 192, 0, 0, 0, 5, 5, 117, 244, 5, 5, 150, 233, 5, 5, 208, 226, 5, 5, 81, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 236, 5, 5, 119, 244, 5, 5, 122, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 227, 192, 0, 0, 0, 5, 5, 11, 224, 5, 5, 170, 228, 192, 0, 0, 0, 5, 5, 51, 245, 5, 5, 123, 244, 192, 0, 0, 0, 5, 5, 126, 244, 192, 0, 0, 0, 5, 5, 131, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 244, 192, 0, 0, 0, 5, 5, 197, 233, 5, 5, 0, 195, 192, 0, 0, 0, 5, 5, 27, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 244, 192, 0, 0, 0, 5, 5, 177, 234, 192, 0, 0, 0, 5, 5, 130, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 237, 192, 0, 0, 0, 5, 5, 32, 230, 5, 5, 124, 244, 192, 0, 0, 0, 5, 5, 188, 230, 192, 0, 0, 0, 5, 5, 79, 227, 5, 5, 180, 237, 5, 5, 128, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 233, 5, 5, 56, 237, 192, 0, 0, 0, 5, 5, 120, 244, 192, 0, 0, 0, 5, 5, 125, 244, 5, 5, 127, 244, 192, 0, 0, 0, 5, 5, 59, 232, 5, 5, 121, 226, 192, 0, 0, 0, 5, 5, 129, 244, 192, 0, 0, 0, 5, 5, 105, 229, 5, 5, 138, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 244, 192, 0, 0, 0, 5, 5, 141, 244, 5, 5, 76, 228, 5, 5, 63, 228, 5, 5, 147, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 244, 192, 0, 0, 0, 5, 5, 146, 244, 5, 5, 88, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 182, 5, 5, 25, 231, 5, 5, 192, 226, 5, 5, 6, 232, 5, 5, 158, 232, 5, 5, 193, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 244, 5, 5, 140, 244, 5, 5, 142, 244, 192, 0, 0, 0, 5, 5, 69, 227, 5, 5, 144, 244, 5, 5, 99, 238, 192, 0, 0, 0, 5, 5, 97, 231, 192, 0, 0, 0, 5, 5, 148, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 229, 5, 5, 133, 244, 5, 5, 9, 238, 192, 0, 0, 0, 5, 5, 66, 230, 5, 5, 173, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 244, 192, 0, 0, 0, 5, 5, 95, 233, 5, 5, 145, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 237, 192, 0, 0, 0, 5, 5, 72, 224, 5, 5, 107, 237, 5, 5, 134, 244, 5, 5, 136, 244, 5, 5, 230, 225, 5, 5, 139, 244, 192, 0, 0, 0, 5, 5, 149, 244, 5, 5, 162, 229, 192, 0, 0, 0, 5, 5, 149, 233, 192, 0, 0, 0, 5, 5, 164, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 231, 5, 5, 223, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 244, 5, 5, 248, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 244, 5, 5, 153, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 236, 5, 5, 125, 227, 5, 5, 237, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 226, 5, 5, 38, 237, 5, 5, 167, 244, 192, 0, 0, 0, 5, 5, 166, 244, 5, 5, 164, 244, 5, 5, 68, 227, 5, 5, 46, 234, 5, 5, 158, 244, 5, 5, 190, 234, 5, 5, 126, 229, 192, 0, 0, 0, 5, 5, 0, 150, 5, 5, 169, 244, 5, 5, 168, 244, 192, 0, 0, 0, 5, 5, 46, 238, 192, 0, 0, 0, 5, 5, 156, 244, 192, 0, 0, 0, 5, 5, 99, 234, 5, 5, 225, 232, 192, 0, 0, 0, 5, 5, 200, 227, 5, 5, 169, 228, 5, 5, 162, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 244, 5, 5, 155, 244, 5, 5, 157, 244, 5, 5, 141, 225, 5, 5, 159, 244, 5, 5, 71, 232, 5, 5, 163, 244, 5, 5, 165, 244, 5, 5, 126, 228, 5, 5, 28, 238, 192, 0, 0, 0, 5, 5, 176, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 244, 192, 0, 0, 0, 5, 5, 106, 234, 5, 5, 174, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 230, 192, 0, 0, 0, 5, 5, 160, 244, 192, 0, 0, 0, 5, 5, 189, 230, 5, 5, 0, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 230, 5, 5, 248, 234, 5, 5, 208, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 233, 192, 0, 0, 0, 5, 5, 57, 234, 5, 5, 139, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 233, 192, 0, 0, 0, 5, 5, 171, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 228, 192, 0, 0, 0, 5, 5, 177, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 230, 192, 0, 0, 0, 5, 5, 9, 232, 192, 0, 0, 0, 5, 5, 75, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 228, 192, 0, 0, 0, 5, 5, 179, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 232, 192, 0, 0, 0, 5, 5, 184, 244, 5, 5, 204, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 244, 5, 5, 43, 227, 192, 0, 0, 0, 5, 5, 179, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 244, 192, 0, 0, 0, 5, 5, 192, 237, 5, 5, 35, 238, 192, 0, 0, 0, 5, 5, 189, 244, 192, 0, 0, 0, 5, 5, 133, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 244, 5, 5, 181, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 225, 192, 0, 0, 0, 5, 5, 53, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 244, 5, 5, 205, 229, 192, 0, 0, 0, 5, 5, 192, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 229, 5, 5, 199, 244, 5, 5, 208, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 244, 192, 0, 0, 0, 5, 5, 185, 226, 192, 0, 0, 0, 5, 5, 195, 244, 192, 0, 0, 0, 5, 5, 206, 236, 5, 5, 202, 244, 5, 5, 203, 244, 5, 5, 205, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 229, 5, 5, 219, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 244, 5, 5, 0, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 244, 5, 5, 85, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 244, 192, 0, 0, 0, 5, 5, 213, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 244, 192, 0, 0, 0, 5, 5, 201, 244, 192, 0, 0, 0, 5, 5, 204, 244, 192, 0, 0, 0, 5, 5, 193, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 244, 5, 5, 220, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 224, 192, 0, 0, 0, 5, 5, 222, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 244, 5, 5, 226, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 237, 192, 0, 0, 0, 5, 5, 9, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 244, 192, 0, 0, 0, 5, 5, 147, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 244, 192, 0, 0, 0, 5, 5, 149, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 244, 5, 5, 230, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 244, 5, 5, 226, 224, 192, 0, 0, 0, 5, 5, 233, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 244, 192, 0, 0, 0, 5, 5, 242, 244, 5, 5, 241, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 244, 192, 0, 0, 0, 5, 5, 245, 244, 5, 5, 208, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 232, 5, 5, 72, 225, 5, 5, 198, 227, 5, 5, 150, 235, 5, 5, 213, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 245, 192, 0, 0, 0, 5, 5, 215, 245, 5, 5, 253, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 245, 5, 5, 128, 235, 192, 0, 0, 0, 5, 5, 58, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 237, 5, 5, 173, 237, 5, 5, 69, 225, 5, 5, 121, 224, 5, 5, 252, 234, 5, 5, 11, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 225, 192, 0, 0, 0, 5, 5, 188, 233, 5, 5, 79, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 244, 5, 5, 249, 244, 5, 5, 32, 236, 5, 5, 251, 244, 192, 0, 0, 0, 5, 5, 115, 225, 5, 5, 250, 244, 192, 0, 0, 0, 5, 5, 231, 235, 5, 5, 128, 233, 5, 5, 140, 224, 192, 0, 0, 0, 5, 5, 254, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 244, 5, 5, 253, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 244, 5, 5, 2, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 245, 5, 5, 125, 231, 5, 5, 6, 226, 192, 0, 0, 0, 5, 5, 135, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 225, 192, 0, 0, 0, 5, 5, 233, 227, 5, 5, 227, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 234, 192, 0, 0, 0, 5, 5, 30, 231, 192, 0, 0, 0, 5, 5, 197, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 225, 5, 5, 133, 238, 5, 5, 38, 224, 5, 5, 45, 224, 5, 5, 47, 224, 5, 5, 194, 231, 192, 0, 0, 0, 5, 5, 46, 224, 192, 0, 0, 0, 5, 5, 177, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 245, 5, 5, 115, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 245, 192, 0, 0, 0, 5, 5, 123, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 245, 5, 5, 127, 245, 5, 5, 126, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 245, 192, 0, 0, 0, 5, 5, 109, 233, 5, 5, 129, 238, 5, 5, 247, 230, 192, 0, 0, 0, 5, 5, 171, 230, 5, 5, 28, 232, 5, 5, 129, 245, 5, 5, 48, 231, 192, 0, 0, 0, 5, 5, 130, 245, 5, 5, 131, 245, 5, 5, 132, 245, 5, 5, 107, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 226, 192, 0, 0, 0, 5, 5, 133, 245, 5, 5, 100, 232, 192, 0, 0, 0, 5, 5, 135, 245, 192, 0, 0, 0, 5, 5, 86, 227, 5, 5, 43, 236, 5, 5, 0, 143, 5, 5, 134, 245, 5, 5, 136, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 225, 5, 5, 112, 230, 5, 5, 106, 232, 192, 0, 0, 0, 5, 5, 137, 245, 192, 0, 0, 0, 5, 5, 170, 233, 5, 5, 189, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 234, 5, 5, 187, 237, 5, 5, 117, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 239, 5, 5, 243, 227, 192, 0, 0, 0, 5, 5, 108, 227, 5, 5, 192, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 242, 5, 5, 7, 233, 192, 0, 0, 0, 5, 5, 107, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 235, 5, 5, 191, 242, 5, 5, 202, 232, 5, 5, 3, 227, 5, 5, 123, 228, 5, 5, 242, 224, 5, 5, 243, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 242, 5, 5, 199, 242, 5, 5, 174, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 234, 192, 0, 0, 0, 5, 5, 35, 228, 192, 0, 0, 0, 5, 5, 200, 242, 192, 0, 0, 0, 5, 5, 201, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 232, 5, 5, 112, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 232, 5, 5, 94, 236, 5, 5, 238, 234, 192, 0, 0, 0, 5, 5, 194, 242, 5, 5, 202, 242, 192, 0, 0, 0, 5, 5, 50, 233, 5, 5, 219, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 232, 5, 5, 195, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 246, 5, 5, 196, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 232, 192, 0, 0, 0, 5, 5, 170, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 227, 192, 0, 0, 0, 5, 5, 193, 230, 192, 0, 0, 0, 5, 5, 193, 242, 5, 5, 129, 231, 5, 5, 236, 229, 5, 5, 129, 230, 5, 5, 147, 224, 192, 0, 0, 0, 5, 5, 203, 242, 5, 5, 182, 227, 5, 5, 19, 231, 192, 0, 0, 0, 5, 5, 206, 242, 5, 5, 8, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 242, 5, 5, 211, 242, 5, 5, 133, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 226, 5, 5, 191, 236, 192, 0, 0, 0, 5, 5, 220, 237, 5, 5, 131, 237, 5, 5, 21, 227, 5, 5, 100, 237, 5, 5, 14, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 235, 5, 5, 120, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 232, 5, 5, 112, 224, 192, 0, 0, 0, 5, 5, 224, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 242, 5, 5, 205, 242, 5, 5, 62, 226, 5, 5, 213, 242, 5, 5, 210, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 231, 192, 0, 0, 0, 5, 5, 212, 242, 5, 5, 161, 231, 5, 5, 100, 224, 5, 5, 32, 232, 192, 0, 0, 0, 5, 5, 70, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 238, 192, 0, 0, 0, 5, 5, 203, 229, 5, 5, 215, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 242, 5, 5, 219, 242, 5, 5, 5, 234, 5, 5, 209, 242, 192, 0, 0, 0, 5, 5, 176, 236, 192, 0, 0, 0, 5, 5, 33, 224, 5, 5, 70, 246, 5, 5, 207, 242, 5, 5, 208, 242, 192, 0, 0, 0, 5, 5, 214, 242, 5, 5, 144, 235, 5, 5, 229, 231, 5, 5, 69, 237, 5, 5, 220, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 236, 5, 5, 223, 242, 192, 0, 0, 0, 5, 5, 229, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 235, 192, 0, 0, 0, 5, 5, 228, 242, 5, 5, 234, 242, 5, 5, 141, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 228, 192, 0, 0, 0, 5, 5, 222, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 227, 5, 5, 230, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 226, 5, 5, 236, 237, 5, 5, 238, 242, 192, 0, 0, 0, 5, 5, 233, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 242, 192, 0, 0, 0, 5, 5, 254, 227, 5, 5, 151, 234, 5, 5, 36, 232, 5, 5, 141, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 230, 192, 0, 0, 0, 5, 5, 224, 242, 192, 0, 0, 0, 5, 5, 53, 232, 5, 5, 122, 228, 5, 5, 139, 228, 5, 5, 225, 242, 192, 0, 0, 0, 5, 5, 53, 238, 5, 5, 158, 224, 192, 0, 0, 0, 5, 5, 231, 242, 5, 5, 51, 228, 5, 5, 235, 242, 192, 0, 0, 0, 5, 5, 251, 227, 5, 5, 236, 242, 5, 5, 108, 231, 5, 5, 237, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 237, 5, 5, 47, 229, 192, 0, 0, 0, 5, 5, 246, 242, 192, 0, 0, 0, 5, 5, 243, 242, 192, 0, 0, 0, 5, 5, 247, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 227, 5, 5, 181, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 226, 5, 5, 240, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 227, 5, 5, 204, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 235, 5, 5, 38, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 236, 192, 0, 0, 0, 5, 5, 73, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 242, 192, 0, 0, 0, 5, 5, 244, 242, 5, 5, 245, 242, 5, 5, 69, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 234, 192, 0, 0, 0, 5, 5, 190, 229, 5, 5, 241, 242, 5, 5, 248, 229, 5, 5, 242, 242, 5, 5, 232, 234, 192, 0, 0, 0, 5, 5, 209, 227, 5, 5, 171, 225, 192, 0, 0, 0, 5, 5, 211, 232, 5, 5, 117, 228, 5, 5, 119, 237, 5, 5, 234, 232, 5, 5, 150, 226, 5, 5, 8, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 243, 5, 5, 248, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 242, 5, 5, 197, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 238, 5, 5, 252, 242, 5, 5, 117, 235, 5, 5, 251, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 230, 5, 5, 38, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 231, 192, 0, 0, 0, 5, 5, 48, 234, 5, 5, 6, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 243, 5, 5, 253, 242, 192, 0, 0, 0, 5, 5, 2, 243, 5, 5, 133, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 236, 192, 0, 0, 0, 5, 5, 4, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 236, 5, 5, 93, 225, 192, 0, 0, 0, 5, 5, 196, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 233, 192, 0, 0, 0, 5, 5, 56, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 227, 192, 0, 0, 0, 5, 5, 3, 236, 192, 0, 0, 0, 5, 5, 73, 234, 5, 5, 71, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 236, 192, 0, 0, 0, 5, 5, 9, 243, 5, 5, 67, 238, 5, 5, 254, 242, 192, 0, 0, 0, 5, 5, 115, 228, 5, 5, 3, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 236, 192, 0, 0, 0, 5, 5, 7, 243, 5, 5, 56, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 232, 5, 5, 255, 225, 192, 0, 0, 0, 5, 5, 78, 237, 5, 5, 111, 224, 5, 5, 21, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 234, 192, 0, 0, 0, 5, 5, 11, 243, 192, 0, 0, 0, 5, 5, 212, 234, 192, 0, 0, 0, 5, 5, 207, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 243, 192, 0, 0, 0, 5, 5, 86, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 236, 192, 0, 0, 0, 5, 5, 242, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 231, 5, 5, 22, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 234, 5, 5, 13, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 243, 5, 5, 19, 243, 192, 0, 0, 0, 5, 5, 176, 227, 192, 0, 0, 0, 5, 5, 94, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 234, 5, 5, 72, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 229, 192, 0, 0, 0, 5, 5, 116, 228, 5, 5, 17, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 243, 5, 5, 6, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 243, 192, 0, 0, 0, 5, 5, 30, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 233, 192, 0, 0, 0, 5, 5, 24, 243, 5, 5, 15, 243, 192, 0, 0, 0, 5, 5, 35, 243, 192, 0, 0, 0, 5, 5, 188, 232, 5, 5, 33, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 231, 5, 5, 32, 243, 192, 0, 0, 0, 5, 5, 31, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 225, 5, 5, 38, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 243, 192, 0, 0, 0, 5, 5, 188, 227, 192, 0, 0, 0, 5, 5, 64, 238, 5, 5, 43, 234, 5, 5, 204, 245, 192, 0, 0, 0, 5, 5, 34, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 237, 5, 5, 23, 243, 5, 5, 26, 243, 5, 5, 162, 230, 5, 5, 28, 243, 192, 0, 0, 0, 5, 5, 115, 230, 5, 5, 174, 229, 5, 5, 120, 230, 192, 0, 0, 0, 5, 5, 83, 224, 5, 5, 14, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 229, 5, 5, 19, 236, 5, 5, 44, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 243, 5, 5, 50, 243, 5, 5, 166, 230, 192, 0, 0, 0, 5, 5, 27, 243, 192, 0, 0, 0, 5, 5, 46, 243, 192, 0, 0, 0, 5, 5, 156, 233, 192, 0, 0, 0, 5, 5, 116, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 243, 5, 5, 42, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 243, 192, 0, 0, 0, 5, 5, 195, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 232, 192, 0, 0, 0, 5, 5, 98, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 234, 5, 5, 203, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 243, 192, 0, 0, 0, 5, 5, 49, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 243, 192, 0, 0, 0, 5, 5, 59, 243, 192, 0, 0, 0, 5, 5, 58, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 224, 5, 5, 53, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 243, 5, 5, 54, 243, 5, 5, 176, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 162, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 228, 192, 0, 0, 0, 5, 5, 65, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 243, 5, 5, 82, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 243, 192, 0, 0, 0, 5, 5, 69, 243, 5, 5, 66, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 243, 5, 5, 70, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 243, 5, 5, 73, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 228, 5, 5, 21, 246, 5, 5, 246, 230, 192, 0, 0, 0, 5, 5, 158, 225, 5, 5, 226, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 230, 5, 5, 64, 237, 192, 0, 0, 0, 5, 5, 240, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 238, 192, 0, 0, 0, 5, 5, 40, 237, 5, 5, 143, 224, 5, 5, 240, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 230, 5, 5, 48, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 224, 192, 0, 0, 0, 5, 5, 146, 232, 5, 5, 71, 229, 5, 5, 242, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 245, 5, 5, 243, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 245, 5, 5, 20, 229, 5, 5, 27, 234, 5, 5, 158, 231, 5, 5, 232, 228, 192, 0, 0, 0, 5, 5, 248, 245, 192, 0, 0, 0, 5, 5, 92, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 245, 5, 5, 88, 237, 5, 5, 187, 225, 192, 0, 0, 0, 5, 5, 244, 245, 5, 5, 254, 229, 5, 5, 253, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 245, 5, 5, 180, 233, 5, 5, 173, 229, 5, 5, 90, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 230, 192, 0, 0, 0, 5, 5, 250, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 227, 5, 5, 189, 229, 192, 0, 0, 0, 5, 5, 2, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 229, 192, 0, 0, 0, 5, 5, 73, 226, 5, 5, 21, 233, 5, 5, 249, 245, 5, 5, 242, 227, 192, 0, 0, 0, 5, 5, 40, 234, 5, 5, 203, 228, 5, 5, 168, 232, 192, 0, 0, 0, 5, 5, 33, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 234, 192, 0, 0, 0, 5, 5, 175, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 235, 5, 5, 106, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 245, 192, 0, 0, 0, 5, 5, 208, 227, 5, 5, 253, 245, 192, 0, 0, 0, 5, 5, 22, 246, 5, 5, 24, 224, 5, 5, 111, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 245, 5, 5, 28, 236, 5, 5, 255, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 246, 192, 0, 0, 0, 5, 5, 221, 227, 192, 0, 0, 0, 5, 5, 85, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 246, 192, 0, 0, 0, 5, 5, 245, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 230, 192, 0, 0, 0, 5, 5, 23, 246, 5, 5, 133, 237, 5, 5, 4, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 246, 5, 5, 2, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 246, 192, 0, 0, 0, 5, 5, 9, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 246, 5, 5, 15, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 233, 5, 5, 14, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 156, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 246, 192, 0, 0, 0, 5, 5, 12, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 246, 5, 5, 22, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 237, 192, 0, 0, 0, 5, 5, 17, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 246, 192, 0, 0, 0, 5, 5, 20, 238, 192, 0, 0, 0, 5, 5, 132, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 245, 5, 5, 0, 140, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 229, 5, 5, 172, 226, 5, 5, 67, 236, 5, 5, 0, 180, 5, 5, 209, 225, 192, 0, 0, 0, 5, 5, 127, 238, 192, 0, 0, 0, 5, 5, 168, 233, 192, 0, 0, 0, 5, 5, 187, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 231, 5, 5, 0, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 231, 5, 5, 141, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 235, 192, 0, 0, 0, 5, 5, 103, 231, 192, 0, 0, 0, 5, 5, 104, 245, 192, 0, 0, 0, 5, 5, 24, 231, 192, 0, 0, 0, 5, 5, 27, 231, 5, 5, 184, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 245, 5, 5, 39, 231, 192, 0, 0, 0, 5, 5, 9, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 245, 5, 5, 106, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 234, 5, 5, 24, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 245, 5, 5, 108, 245, 5, 5, 36, 235, 5, 5, 209, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 225, 5, 5, 110, 245, 192, 0, 0, 0, 5, 5, 111, 245, 192, 0, 0, 0, 5, 5, 112, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 232, 5, 5, 13, 242, 192, 0, 0, 0, 5, 5, 78, 226, 5, 5, 14, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 242, 192, 0, 0, 0, 5, 5, 34, 238, 5, 5, 16, 242, 5, 5, 17, 242, 5, 5, 190, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 242, 5, 5, 125, 229, 5, 5, 18, 242, 5, 5, 170, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 242, 5, 5, 20, 242, 192, 0, 0, 0, 5, 5, 174, 227, 192, 0, 0, 0, 5, 5, 22, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 227, 192, 0, 0, 0, 5, 5, 253, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 231, 192, 0, 0, 0, 5, 5, 143, 227, 5, 5, 148, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 242, 5, 5, 25, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 225, 5, 5, 57, 235, 5, 5, 70, 233, 5, 5, 24, 242, 5, 5, 168, 237, 5, 5, 238, 236, 5, 5, 26, 242, 5, 5, 30, 242, 5, 5, 27, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 242, 5, 5, 212, 229, 192, 0, 0, 0, 5, 5, 32, 242, 5, 5, 33, 242, 5, 5, 177, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 242, 192, 0, 0, 0, 5, 5, 31, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 242, 192, 0, 0, 0, 5, 5, 40, 242, 5, 5, 186, 224, 5, 5, 34, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 230, 5, 5, 126, 224, 5, 5, 39, 242, 5, 5, 38, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 242, 5, 5, 41, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 235, 5, 5, 251, 237, 5, 5, 174, 230, 5, 5, 44, 242, 192, 0, 0, 0, 5, 5, 79, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 244, 5, 5, 45, 242, 5, 5, 42, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 227, 5, 5, 5, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 242, 192, 0, 0, 0, 5, 5, 247, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 242, 5, 5, 251, 233, 192, 0, 0, 0, 5, 5, 52, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 230, 192, 0, 0, 0, 5, 5, 227, 236, 192, 0, 0, 0, 5, 5, 175, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 234, 192, 0, 0, 0, 5, 5, 187, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 244, 5, 5, 199, 227, 5, 5, 78, 235, 192, 0, 0, 0, 5, 5, 43, 230, 5, 5, 67, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 244, 5, 5, 95, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 244, 192, 0, 0, 0, 5, 5, 64, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 244, 5, 5, 69, 244, 5, 5, 248, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 237, 192, 0, 0, 0, 5, 5, 63, 244, 5, 5, 63, 226, 5, 5, 65, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 244, 192, 0, 0, 0, 5, 5, 247, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 244, 192, 0, 0, 0, 5, 5, 76, 244, 192, 0, 0, 0, 5, 5, 75, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 232, 192, 0, 0, 0, 5, 5, 175, 229, 5, 5, 215, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 230, 5, 5, 74, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 244, 5, 5, 242, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 244, 5, 5, 89, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 244, 5, 5, 83, 244, 192, 0, 0, 0, 5, 5, 85, 244, 192, 0, 0, 0, 5, 5, 81, 244, 192, 0, 0, 0, 5, 5, 88, 244, 192, 0, 0, 0, 5, 5, 87, 244, 192, 0, 0, 0, 5, 5, 86, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 230, 5, 5, 91, 232, 5, 5, 193, 231, 5, 5, 135, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 244, 192, 0, 0, 0, 5, 5, 93, 244, 192, 0, 0, 0, 5, 5, 95, 244, 5, 5, 169, 227, 5, 5, 82, 244, 5, 5, 92, 244, 192, 0, 0, 0, 5, 5, 208, 232, 5, 5, 232, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 236, 5, 5, 96, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 244, 192, 0, 0, 0, 5, 5, 101, 244, 5, 5, 102, 244, 192, 0, 0, 0, 5, 5, 239, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 244, 192, 0, 0, 0, 5, 5, 108, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 244, 192, 0, 0, 0, 5, 5, 105, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 244, 5, 5, 106, 244, 5, 5, 107, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 227, 192, 0, 0, 0, 5, 5, 90, 247, 192, 0, 0, 0, 5, 5, 91, 247, 192, 0, 0, 0, 5, 5, 210, 231, 5, 5, 0, 200, 5, 5, 159, 232, 192, 0, 0, 0, 5, 5, 153, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 234, 5, 5, 39, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 245, 192, 0, 0, 0, 5, 5, 225, 231, 5, 5, 67, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 240, 192, 0, 0, 0, 5, 5, 42, 245, 192, 0, 0, 0, 5, 5, 43, 245, 192, 0, 0, 0, 5, 5, 44, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 226, 5, 5, 20, 241, 5, 5, 53, 233, 192, 0, 0, 0, 5, 5, 76, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 236, 5, 5, 162, 233, 192, 0, 0, 0, 5, 5, 154, 226, 5, 5, 92, 247, 5, 5, 32, 224, 192, 0, 0, 0, 5, 5, 84, 243, 5, 5, 75, 234, 5, 5, 187, 236, 5, 5, 71, 228, 5, 5, 42, 233, 192, 0, 0, 0, 5, 5, 191, 225, 192, 0, 0, 0, 5, 5, 54, 231, 5, 5, 193, 225, 192, 0, 0, 0, 5, 5, 158, 246, 5, 5, 189, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 244, 192, 0, 0, 0, 5, 5, 159, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 246, 5, 5, 181, 228, 192, 0, 0, 0, 5, 5, 160, 246, 5, 5, 207, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 230, 5, 5, 182, 239, 5, 5, 163, 246, 5, 5, 201, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 230, 5, 5, 12, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 246, 192, 0, 0, 0, 5, 5, 7, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 228, 5, 5, 165, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 233, 192, 0, 0, 0, 5, 5, 93, 236, 5, 5, 136, 247, 192, 0, 0, 0, 5, 5, 137, 247, 192, 0, 0, 0, 5, 5, 138, 247, 5, 5, 16, 230, 192, 0, 0, 0, 5, 5, 225, 226, 5, 5, 250, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 247, 192, 0, 0, 0, 5, 5, 117, 231, 5, 5, 139, 247, 5, 5, 39, 236, 192, 0, 0, 0, 5, 5, 142, 247, 5, 5, 0, 170, 5, 5, 184, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 236, 5, 5, 141, 247, 192, 0, 0, 0, 5, 5, 41, 225, 5, 5, 126, 226, 5, 5, 148, 247, 5, 5, 147, 247, 5, 5, 197, 231, 5, 5, 143, 247, 5, 5, 144, 247, 5, 5, 61, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 247, 5, 5, 158, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 234, 5, 5, 254, 228, 5, 5, 34, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 247, 5, 5, 149, 247, 5, 5, 146, 247, 5, 5, 93, 224, 192, 0, 0, 0, 5, 5, 176, 237, 5, 5, 170, 236, 5, 5, 227, 228, 5, 5, 140, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 236, 192, 0, 0, 0, 5, 5, 218, 237, 5, 5, 141, 227, 5, 5, 151, 247, 192, 0, 0, 0, 5, 5, 241, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 229, 5, 5, 153, 247, 5, 5, 156, 247, 192, 0, 0, 0, 5, 5, 155, 247, 5, 5, 158, 247, 5, 5, 154, 247, 192, 0, 0, 0, 5, 5, 206, 227, 5, 5, 157, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 234, 5, 5, 160, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 247, 5, 5, 164, 247, 5, 5, 91, 225, 192, 0, 0, 0, 5, 5, 159, 247, 192, 0, 0, 0, 5, 5, 165, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 247, 192, 0, 0, 0, 5, 5, 166, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 247, 192, 0, 0, 0, 5, 5, 167, 247, 5, 5, 170, 247, 5, 5, 172, 247, 192, 0, 0, 0, 5, 5, 173, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 234, 5, 5, 176, 247, 192, 0, 0, 0, 5, 5, 175, 247, 192, 0, 0, 0, 5, 5, 64, 230, 5, 5, 169, 247, 5, 5, 123, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 225, 5, 5, 79, 224, 5, 5, 13, 234, 192, 0, 0, 0, 5, 5, 178, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 247, 5, 5, 126, 237, 5, 5, 181, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 247, 192, 0, 0, 0, 5, 5, 183, 247, 5, 5, 180, 247, 5, 5, 177, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 135, 5, 5, 185, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 247, 192, 0, 0, 0, 5, 5, 189, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 247, 192, 0, 0, 0, 5, 5, 186, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 225, 192, 0, 0, 0, 5, 5, 0, 181, 5, 5, 0, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 237, 192, 0, 0, 0, 5, 5, 156, 225, 192, 0, 0, 0, 5, 5, 162, 228, 5, 5, 218, 227, 5, 5, 85, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 134, 192, 0, 0, 0, 5, 5, 87, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 237, 5, 5, 240, 247, 192, 0, 0, 0, 5, 5, 241, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 237, 5, 5, 172, 231, 192, 0, 0, 0, 5, 5, 161, 236, 192, 0, 0, 0, 5, 5, 122, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 246, 5, 5, 0, 153, 5, 5, 101, 237, 5, 5, 4, 236, 5, 5, 80, 228, 5, 5, 130, 227, 192, 0, 0, 0, 5, 5, 132, 229, 192, 0, 0, 0, 5, 5, 187, 226, 5, 5, 153, 225, 5, 5, 144, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 231, 5, 5, 216, 225, 192, 0, 0, 0, 5, 5, 126, 246, 5, 5, 170, 230, 192, 0, 0, 0, 5, 5, 191, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 235, 5, 5, 129, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 231, 192, 0, 0, 0, 5, 5, 20, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 246, 5, 5, 131, 246, 5, 5, 194, 230, 192, 0, 0, 0, 5, 5, 64, 229, 192, 0, 0, 0, 5, 5, 128, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 246, 5, 5, 132, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 237, 5, 5, 229, 230, 192, 0, 0, 0, 5, 5, 133, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 246, 192, 0, 0, 0, 5, 5, 83, 237, 5, 5, 215, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 246, 192, 0, 0, 0, 5, 5, 214, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 246, 5, 5, 128, 229, 5, 5, 26, 229, 5, 5, 138, 246, 192, 0, 0, 0, 5, 5, 83, 234, 192, 0, 0, 0, 5, 5, 17, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 238, 5, 5, 166, 237, 192, 0, 0, 0, 5, 5, 142, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 246, 5, 5, 140, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 246, 5, 5, 210, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 233, 5, 5, 145, 246, 5, 5, 243, 225, 192, 0, 0, 0, 5, 5, 146, 246, 5, 5, 38, 231, 192, 0, 0, 0, 5, 5, 144, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 228, 5, 5, 131, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 246, 5, 5, 134, 238, 5, 5, 147, 246, 5, 5, 150, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 230, 5, 5, 14, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 246, 5, 5, 148, 246, 5, 5, 49, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 246, 5, 5, 159, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 246, 5, 5, 154, 246, 192, 0, 0, 0, 5, 5, 175, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 231, 192, 0, 0, 0, 5, 5, 72, 232, 192, 0, 0, 0, 5, 5, 5, 238, 5, 5, 161, 225, 192, 0, 0, 0, 5, 5, 174, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 234, 192, 0, 0, 0, 5, 5, 156, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 237, 192, 0, 0, 0, 5, 5, 157, 246, 192, 0, 0, 0, 5, 5, 170, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 230, 5, 5, 242, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 233, 5, 5, 103, 236, 192, 0, 0, 0, 5, 5, 184, 237, 192, 0, 0, 0, 5, 5, 66, 247, 192, 0, 0, 0, 5, 5, 6, 229, 192, 0, 0, 0, 5, 5, 145, 228, 5, 5, 67, 247, 5, 5, 4, 226, 5, 5, 0, 137, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 246, 192, 0, 0, 0, 5, 5, 73, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 235, 5, 5, 69, 226, 5, 5, 127, 229, 5, 5, 74, 246, 5, 5, 147, 230, 5, 5, 238, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 232, 5, 5, 63, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 246, 5, 5, 188, 231, 192, 0, 0, 0, 5, 5, 6, 236, 192, 0, 0, 0, 5, 5, 23, 238, 5, 5, 76, 246, 5, 5, 77, 246, 192, 0, 0, 0, 5, 5, 24, 236, 192, 0, 0, 0, 5, 5, 81, 246, 5, 5, 82, 246, 192, 0, 0, 0, 5, 5, 86, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 246, 192, 0, 0, 0, 5, 5, 88, 246, 5, 5, 96, 246, 5, 5, 152, 237, 192, 0, 0, 0, 5, 5, 91, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 246, 5, 5, 80, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 233, 5, 5, 36, 237, 5, 5, 83, 246, 5, 5, 84, 246, 5, 5, 85, 246, 5, 5, 87, 246, 192, 0, 0, 0, 5, 5, 227, 229, 192, 0, 0, 0, 5, 5, 27, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 227, 192, 0, 0, 0, 5, 5, 98, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 246, 192, 0, 0, 0, 5, 5, 92, 246, 192, 0, 0, 0, 5, 5, 97, 246, 192, 0, 0, 0, 5, 5, 18, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 233, 5, 5, 94, 246, 5, 5, 95, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 246, 5, 5, 60, 230, 5, 5, 164, 236, 5, 5, 93, 246, 5, 5, 151, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 246, 192, 0, 0, 0, 5, 5, 199, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 230, 5, 5, 0, 139, 5, 5, 229, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 224, 192, 0, 0, 0, 5, 5, 102, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 234, 5, 5, 186, 225, 192, 0, 0, 0, 5, 5, 103, 246, 5, 5, 101, 246, 5, 5, 105, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 225, 192, 0, 0, 0, 5, 5, 106, 246, 192, 0, 0, 0, 5, 5, 107, 246, 192, 0, 0, 0, 5, 5, 110, 246, 192, 0, 0, 0, 5, 5, 42, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 231, 5, 5, 94, 228, 5, 5, 108, 246, 5, 5, 24, 234, 5, 5, 166, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 246, 5, 5, 196, 229, 5, 5, 98, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 246, 5, 5, 80, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 246, 192, 0, 0, 0, 5, 5, 117, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 233, 5, 5, 30, 246, 5, 5, 219, 229, 192, 0, 0, 0, 5, 5, 39, 233, 192, 0, 0, 0, 5, 5, 31, 246, 5, 5, 18, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 246, 192, 0, 0, 0, 5, 5, 17, 232, 5, 5, 33, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 238, 5, 5, 39, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 246, 5, 5, 34, 246, 5, 5, 35, 246, 5, 5, 17, 238, 5, 5, 49, 233, 5, 5, 234, 233, 5, 5, 40, 246, 192, 0, 0, 0, 5, 5, 38, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 235, 192, 0, 0, 0, 5, 5, 42, 246, 5, 5, 211, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 228, 192, 0, 0, 0, 5, 5, 41, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 225, 5, 5, 8, 228, 192, 0, 0, 0, 5, 5, 43, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 239, 5, 5, 201, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 230, 5, 5, 44, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 236, 5, 5, 122, 238, 5, 5, 213, 229, 192, 0, 0, 0, 5, 5, 95, 232, 5, 5, 126, 227, 192, 0, 0, 0, 5, 5, 185, 235, 5, 5, 185, 233, 192, 0, 0, 0, 5, 5, 116, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 224, 192, 0, 0, 0, 5, 5, 113, 232, 192, 0, 0, 0, 5, 5, 228, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 229, 5, 5, 241, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 247, 5, 5, 237, 224, 192, 0, 0, 0, 5, 5, 90, 232, 5, 5, 35, 236, 192, 0, 0, 0, 5, 5, 214, 237, 192, 0, 0, 0, 5, 5, 72, 247, 192, 0, 0, 0, 5, 5, 70, 247, 192, 0, 0, 0, 5, 5, 15, 228, 5, 5, 224, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 236, 192, 0, 0, 0, 5, 5, 240, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 235, 192, 0, 0, 0, 5, 5, 76, 247, 5, 5, 75, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 224, 192, 0, 0, 0, 5, 5, 20, 233, 5, 5, 172, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 188, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 247, 192, 0, 0, 0, 5, 5, 9, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 247, 192, 0, 0, 0, 5, 5, 79, 247, 192, 0, 0, 0, 5, 5, 150, 225, 5, 5, 74, 228, 5, 5, 14, 228, 192, 0, 0, 0, 5, 5, 217, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 235, 192, 0, 0, 0, 5, 5, 235, 228, 5, 5, 112, 232, 5, 5, 194, 247, 5, 5, 195, 247, 5, 5, 97, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 225, 5, 5, 196, 247, 5, 5, 117, 234, 192, 0, 0, 0, 5, 5, 86, 232, 5, 5, 93, 237, 192, 0, 0, 0, 5, 5, 197, 247, 192, 0, 0, 0, 5, 5, 198, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 236, 5, 5, 221, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 247, 5, 5, 159, 228, 5, 5, 42, 225, 5, 5, 233, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 225, 5, 5, 233, 234, 192, 0, 0, 0, 5, 5, 107, 229, 5, 5, 201, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 229, 5, 5, 200, 247, 192, 0, 0, 0, 5, 5, 203, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 247, 5, 5, 204, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 228, 5, 5, 229, 228, 5, 5, 114, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 229, 192, 0, 0, 0, 5, 5, 105, 234, 5, 5, 193, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 228, 192, 0, 0, 0, 5, 5, 3, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 237, 5, 5, 141, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 248, 192, 0, 0, 0, 5, 5, 193, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 248, 5, 5, 145, 248, 192, 0, 0, 0, 5, 5, 0, 168, 192, 0, 0, 0, 5, 5, 143, 248, 192, 0, 0, 0, 5, 5, 147, 248, 5, 5, 237, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 248, 192, 0, 0, 0, 5, 5, 124, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 224, 5, 5, 146, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 248, 192, 0, 0, 0, 5, 5, 169, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 248, 192, 0, 0, 0, 5, 5, 156, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 248, 5, 5, 157, 248, 5, 5, 145, 226, 192, 0, 0, 0, 5, 5, 29, 224, 192, 0, 0, 0, 5, 5, 152, 248, 5, 5, 148, 248, 5, 5, 178, 225, 192, 0, 0, 0, 5, 5, 154, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 248, 192, 0, 0, 0, 5, 5, 159, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 248, 192, 0, 0, 0, 5, 5, 83, 228, 192, 0, 0, 0, 5, 5, 71, 230, 192, 0, 0, 0, 5, 5, 160, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 248, 192, 0, 0, 0, 5, 5, 150, 248, 192, 0, 0, 0, 5, 5, 160, 225, 192, 0, 0, 0, 5, 5, 188, 228, 5, 5, 166, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 226, 5, 5, 124, 229, 5, 5, 14, 238, 5, 5, 108, 234, 192, 0, 0, 0, 5, 5, 106, 225, 192, 0, 0, 0, 5, 5, 155, 224, 192, 0, 0, 0, 5, 5, 162, 248, 192, 0, 0, 0, 5, 5, 163, 248, 5, 5, 246, 232, 192, 0, 0, 0, 5, 5, 167, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 248, 192, 0, 0, 0, 5, 5, 171, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 248, 5, 5, 172, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 248, 192, 0, 0, 0, 5, 5, 118, 229, 192, 0, 0, 0, 5, 5, 11, 225, 192, 0, 0, 0, 5, 5, 170, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 232, 192, 0, 0, 0, 5, 5, 98, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 224, 5, 5, 16, 228, 192, 0, 0, 0, 5, 5, 223, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 248, 5, 5, 179, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 227, 5, 5, 184, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 248, 5, 5, 176, 248, 5, 5, 180, 248, 5, 5, 137, 230, 5, 5, 182, 248, 5, 5, 185, 248, 5, 5, 178, 248, 5, 5, 109, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 238, 5, 5, 204, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 248, 192, 0, 0, 0, 5, 5, 80, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 226, 5, 5, 191, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 248, 5, 5, 193, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 248, 192, 0, 0, 0, 5, 5, 196, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 248, 5, 5, 195, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 248, 5, 5, 217, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 224, 5, 5, 206, 248, 5, 5, 205, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 230, 5, 5, 13, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 229, 5, 5, 244, 248, 5, 5, 63, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 248, 5, 5, 235, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 231, 192, 0, 0, 0, 5, 5, 80, 225, 5, 5, 99, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 248, 5, 5, 247, 248, 5, 5, 249, 248, 5, 5, 214, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 248, 192, 0, 0, 0, 5, 5, 19, 237, 5, 5, 238, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 230, 5, 5, 251, 248, 5, 5, 216, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 248, 5, 5, 253, 248, 5, 5, 215, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 249, 192, 0, 0, 0, 5, 5, 255, 248, 5, 5, 4, 249, 192, 0, 0, 0, 5, 5, 175, 227, 192, 0, 0, 0, 5, 5, 2, 249, 5, 5, 3, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 226, 5, 5, 35, 234, 5, 5, 6, 249, 192, 0, 0, 0, 5, 5, 150, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 237, 5, 5, 67, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 233, 192, 0, 0, 0, 5, 5, 243, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 243, 5, 5, 236, 228, 5, 5, 244, 243, 5, 5, 158, 227, 5, 5, 245, 243, 5, 5, 67, 235, 5, 5, 246, 243, 5, 5, 14, 237, 5, 5, 38, 228, 5, 5, 247, 243, 5, 5, 248, 243, 5, 5, 61, 228, 5, 5, 178, 232, 5, 5, 203, 234, 5, 5, 249, 243, 192, 0, 0, 0, 5, 5, 57, 225, 5, 5, 250, 243, 5, 5, 242, 232, 5, 5, 205, 226, 5, 5, 47, 231, 192, 0, 0, 0, 5, 5, 75, 238, 5, 5, 130, 230, 5, 5, 109, 226, 5, 5, 203, 237, 5, 5, 222, 234, 5, 5, 90, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 231, 5, 5, 251, 243, 5, 5, 87, 235, 5, 5, 252, 243, 5, 5, 253, 243, 5, 5, 254, 243, 5, 5, 255, 229, 5, 5, 87, 238, 5, 5, 48, 233, 5, 5, 48, 235, 5, 5, 189, 237, 5, 5, 227, 237, 5, 5, 255, 243, 5, 5, 0, 203, 5, 5, 2, 244, 5, 5, 3, 244, 5, 5, 28, 233, 5, 5, 131, 236, 5, 5, 217, 228, 5, 5, 4, 244, 5, 5, 0, 209, 5, 5, 190, 232, 5, 5, 175, 228, 5, 5, 5, 244, 5, 5, 166, 232, 192, 0, 0, 0, 5, 5, 6, 244, 5, 5, 246, 227, 5, 5, 236, 226, 5, 5, 216, 235, 5, 5, 7, 244, 5, 5, 143, 230, 5, 5, 38, 229, 5, 5, 152, 228, 5, 5, 109, 234, 5, 5, 8, 244, 5, 5, 9, 244, 5, 5, 28, 229, 5, 5, 187, 235, 192, 0, 0, 0, 5, 5, 227, 233, 5, 5, 44, 234, 5, 5, 60, 228, 5, 5, 10, 244, 5, 5, 24, 228, 5, 5, 205, 235, 5, 5, 11, 244, 192, 0, 0, 0, 5, 5, 206, 235, 5, 5, 12, 244, 5, 5, 13, 244, 5, 5, 60, 225, 5, 5, 14, 244, 5, 5, 15, 244, 5, 5, 62, 233, 5, 5, 196, 234, 5, 5, 230, 230, 5, 5, 17, 244, 5, 5, 31, 224, 5, 5, 13, 225, 192, 0, 0, 0, 5, 5, 18, 244, 5, 5, 19, 244, 5, 5, 73, 238, 5, 5, 112, 237, 5, 5, 20, 244, 5, 5, 116, 230, 5, 5, 40, 238, 5, 5, 21, 244, 5, 5, 22, 244, 5, 5, 23, 244, 5, 5, 90, 228, 5, 5, 235, 230, 5, 5, 172, 229, 5, 5, 24, 244, 5, 5, 25, 244, 5, 5, 25, 228, 192, 0, 0, 0, 5, 5, 26, 244, 5, 5, 27, 244, 5, 5, 16, 244, 5, 5, 8, 226, 5, 5, 28, 244, 192, 0, 0, 0, 5, 5, 29, 244, 5, 5, 30, 244, 5, 5, 202, 227, 5, 5, 182, 225, 5, 5, 110, 230, 5, 5, 62, 224, 5, 5, 31, 244, 5, 5, 7, 237, 5, 5, 32, 244, 5, 5, 179, 226, 5, 5, 34, 244, 5, 5, 33, 244, 5, 5, 130, 226, 192, 0, 0, 0, 5, 5, 35, 244, 5, 5, 180, 224, 5, 5, 36, 244, 5, 5, 37, 244, 5, 5, 38, 244, 5, 5, 39, 244, 5, 5, 40, 244, 5, 5, 41, 244, 5, 5, 42, 244, 5, 5, 153, 236, 5, 5, 241, 233, 5, 5, 43, 244, 5, 5, 44, 244, 5, 5, 45, 244, 5, 5, 46, 244, 5, 5, 9, 233, 5, 5, 47, 244, 5, 5, 48, 244, 5, 5, 49, 244, 5, 5, 50, 244, 5, 5, 51, 244, 5, 5, 151, 228, 5, 5, 52, 244, 5, 5, 136, 248, 192, 0, 0, 0, 5, 5, 203, 226, 192, 0, 0, 0, 5, 5, 145, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 248, 192, 0, 0, 0, 5, 5, 138, 248, 5, 5, 139, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 227, 5, 5, 178, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 238, 5, 5, 99, 227, 192, 0, 0, 0, 5, 5, 134, 230, 5, 5, 167, 246, 192, 0, 0, 0, 5, 5, 57, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 246, 192, 0, 0, 0, 5, 5, 168, 246, 5, 5, 0, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 246, 5, 5, 134, 237, 5, 5, 93, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 246, 5, 5, 142, 233, 192, 0, 0, 0, 5, 5, 172, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 236, 192, 0, 0, 0, 5, 5, 62, 232, 192, 0, 0, 0, 5, 5, 198, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 248, 5, 5, 181, 235, 5, 5, 239, 248, 192, 0, 0, 0, 5, 5, 83, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 248, 5, 5, 242, 248, 192, 0, 0, 0, 5, 5, 243, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 239, 5, 5, 242, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 243, 5, 5, 226, 236, 192, 0, 0, 0, 5, 5, 10, 249, 192, 0, 0, 0, 5, 5, 228, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 247, 192, 0, 0, 0, 5, 5, 130, 236, 192, 0, 0, 0, 5, 5, 11, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 235, 5, 5, 12, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 225, 5, 5, 94, 225, 5, 5, 14, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 249, 5, 5, 15, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 249, 192, 0, 0, 0, 5, 5, 100, 227, 5, 5, 0, 157, 192, 0, 0, 0, 5, 5, 18, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 226, 5, 5, 129, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 236, 5, 5, 185, 229, 192, 0, 0, 0, 5, 5, 72, 229, 5, 5, 119, 245, 5, 5, 143, 237, 5, 5, 62, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 247, 5, 5, 48, 226, 5, 5, 159, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 231, 192, 0, 0, 0, 5, 5, 243, 247, 192, 0, 0, 0, 5, 5, 244, 247, 5, 5, 239, 226, 5, 5, 245, 247, 5, 5, 118, 227, 5, 5, 23, 237, 5, 5, 0, 176, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 247, 192, 0, 0, 0, 5, 5, 248, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 247, 5, 5, 250, 247, 5, 5, 251, 247, 5, 5, 253, 247, 5, 5, 252, 247, 5, 5, 153, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 247, 192, 0, 0, 0, 5, 5, 50, 226, 192, 0, 0, 0, 5, 5, 2, 248, 5, 5, 66, 236, 5, 5, 203, 240, 5, 5, 198, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 224, 192, 0, 0, 0, 5, 5, 122, 225, 192, 0, 0, 0, 5, 5, 244, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 231, 5, 5, 3, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 230, 5, 5, 69, 230, 5, 5, 190, 237, 5, 5, 5, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 248, 192, 0, 0, 0, 5, 5, 242, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 231, 192, 0, 0, 0, 5, 5, 9, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 248, 5, 5, 73, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 246, 5, 5, 67, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 233, 5, 5, 114, 236, 192, 0, 0, 0, 5, 5, 189, 233, 5, 5, 136, 237, 192, 0, 0, 0, 5, 5, 194, 232, 192, 0, 0, 0, 5, 5, 201, 229, 5, 5, 17, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 235, 192, 0, 0, 0, 5, 5, 240, 237, 192, 0, 0, 0, 5, 5, 253, 225, 5, 5, 204, 226, 5, 5, 148, 245, 5, 5, 194, 226, 192, 0, 0, 0, 5, 5, 147, 245, 5, 5, 192, 224, 5, 5, 29, 227, 5, 5, 185, 237, 192, 0, 0, 0, 5, 5, 138, 226, 5, 5, 96, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 228, 5, 5, 84, 226, 5, 5, 154, 245, 192, 0, 0, 0, 5, 5, 155, 245, 5, 5, 0, 151, 5, 5, 91, 229, 192, 0, 0, 0, 5, 5, 153, 245, 5, 5, 239, 236, 192, 0, 0, 0, 5, 5, 156, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 226, 192, 0, 0, 0, 5, 5, 150, 245, 5, 5, 152, 245, 5, 5, 54, 233, 5, 5, 229, 237, 5, 5, 125, 237, 5, 5, 136, 235, 5, 5, 163, 245, 5, 5, 208, 234, 5, 5, 164, 245, 192, 0, 0, 0, 5, 5, 128, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 224, 5, 5, 166, 245, 5, 5, 255, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 231, 5, 5, 167, 245, 192, 0, 0, 0, 5, 5, 165, 245, 5, 5, 163, 231, 5, 5, 162, 245, 5, 5, 66, 233, 5, 5, 169, 245, 5, 5, 0, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 238, 5, 5, 239, 247, 192, 0, 0, 0, 5, 5, 158, 245, 5, 5, 159, 245, 5, 5, 160, 245, 5, 5, 161, 245, 5, 5, 170, 245, 5, 5, 186, 243, 5, 5, 173, 245, 192, 0, 0, 0, 5, 5, 116, 229, 5, 5, 92, 236, 5, 5, 171, 245, 5, 5, 176, 245, 5, 5, 224, 226, 5, 5, 172, 245, 192, 0, 0, 0, 5, 5, 136, 228, 192, 0, 0, 0, 5, 5, 174, 235, 192, 0, 0, 0, 5, 5, 0, 149, 192, 0, 0, 0, 5, 5, 177, 245, 5, 5, 66, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 230, 5, 5, 41, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 245, 5, 5, 175, 245, 5, 5, 50, 237, 5, 5, 14, 232, 5, 5, 58, 231, 5, 5, 179, 245, 5, 5, 107, 231, 5, 5, 33, 239, 192, 0, 0, 0, 5, 5, 109, 224, 192, 0, 0, 0, 5, 5, 184, 245, 192, 0, 0, 0, 5, 5, 147, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 226, 192, 0, 0, 0, 5, 5, 139, 234, 5, 5, 185, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 245, 192, 0, 0, 0, 5, 5, 251, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 234, 192, 0, 0, 0, 5, 5, 186, 245, 192, 0, 0, 0, 5, 5, 154, 229, 5, 5, 76, 236, 5, 5, 187, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 226, 5, 5, 161, 226, 192, 0, 0, 0, 5, 5, 188, 245, 5, 5, 61, 232, 5, 5, 173, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 245, 5, 5, 191, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 235, 192, 0, 0, 0, 5, 5, 198, 245, 192, 0, 0, 0, 5, 5, 194, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 245, 5, 5, 217, 232, 192, 0, 0, 0, 5, 5, 50, 236, 5, 5, 192, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 226, 5, 5, 81, 235, 5, 5, 75, 231, 5, 5, 195, 245, 5, 5, 196, 245, 5, 5, 54, 234, 5, 5, 130, 234, 5, 5, 0, 208, 192, 0, 0, 0, 5, 5, 202, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 245, 192, 0, 0, 0, 5, 5, 110, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 226, 192, 0, 0, 0, 5, 5, 203, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 234, 5, 5, 10, 231, 5, 5, 29, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 245, 192, 0, 0, 0, 5, 5, 182, 231, 192, 0, 0, 0, 5, 5, 206, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 242, 5, 5, 210, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 234, 5, 5, 211, 245, 5, 5, 57, 224, 5, 5, 169, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 245, 192, 0, 0, 0, 5, 5, 207, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 238, 192, 0, 0, 0, 5, 5, 211, 248, 5, 5, 16, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 237, 5, 5, 208, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 237, 5, 5, 246, 228, 192, 0, 0, 0, 5, 5, 207, 248, 192, 0, 0, 0, 5, 5, 60, 236, 5, 5, 208, 248, 5, 5, 209, 248, 192, 0, 0, 0, 5, 5, 210, 248, 5, 5, 247, 228, 5, 5, 209, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 233, 5, 5, 33, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 248, 192, 0, 0, 0, 5, 5, 132, 233, 192, 0, 0, 0, 5, 5, 78, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 242, 5, 5, 176, 233, 192, 0, 0, 0, 5, 5, 2, 235, 5, 5, 233, 237, 192, 0, 0, 0, 5, 5, 213, 248, 5, 5, 214, 248, 5, 5, 215, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 248, 192, 0, 0, 0, 5, 5, 111, 227, 5, 5, 219, 248, 5, 5, 0, 193, 5, 5, 216, 248, 192, 0, 0, 0, 5, 5, 217, 248, 5, 5, 112, 228, 5, 5, 145, 224, 192, 0, 0, 0, 5, 5, 222, 248, 5, 5, 223, 248, 5, 5, 30, 226, 5, 5, 108, 224, 5, 5, 71, 235, 5, 5, 220, 248, 5, 5, 38, 225, 192, 0, 0, 0, 5, 5, 221, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 234, 192, 0, 0, 0, 5, 5, 226, 248, 192, 0, 0, 0, 5, 5, 227, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 233, 192, 0, 0, 0, 5, 5, 229, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 249, 5, 5, 6, 230, 5, 5, 88, 228, 192, 0, 0, 0, 5, 5, 233, 232, 5, 5, 20, 236, 5, 5, 145, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 239, 5, 5, 105, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 239, 5, 5, 0, 138, 5, 5, 251, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 228, 192, 0, 0, 0, 5, 5, 255, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 238, 192, 0, 0, 0, 5, 5, 253, 239, 5, 5, 229, 236, 192, 0, 0, 0, 5, 5, 22, 233, 5, 5, 2, 240, 5, 5, 252, 239, 192, 0, 0, 0, 5, 5, 3, 240, 5, 5, 168, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 240, 5, 5, 13, 240, 5, 5, 5, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 234, 5, 5, 179, 237, 192, 0, 0, 0, 5, 5, 23, 240, 192, 0, 0, 0, 5, 5, 21, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 240, 5, 5, 180, 228, 5, 5, 83, 230, 192, 0, 0, 0, 5, 5, 254, 239, 5, 5, 19, 240, 5, 5, 22, 240, 5, 5, 6, 240, 5, 5, 104, 226, 5, 5, 0, 163, 5, 5, 15, 240, 5, 5, 148, 235, 5, 5, 9, 240, 5, 5, 184, 227, 192, 0, 0, 0, 5, 5, 82, 226, 5, 5, 20, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 240, 5, 5, 7, 240, 5, 5, 93, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 235, 5, 5, 8, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 234, 5, 5, 10, 240, 192, 0, 0, 0, 5, 5, 11, 240, 5, 5, 16, 240, 5, 5, 17, 240, 5, 5, 144, 224, 5, 5, 25, 240, 5, 5, 209, 233, 192, 0, 0, 0, 5, 5, 9, 237, 5, 5, 35, 240, 5, 5, 39, 240, 5, 5, 2, 234, 5, 5, 46, 240, 192, 0, 0, 0, 5, 5, 237, 230, 5, 5, 36, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 229, 5, 5, 33, 240, 192, 0, 0, 0, 5, 5, 0, 216, 5, 5, 10, 227, 5, 5, 45, 240, 5, 5, 27, 240, 192, 0, 0, 0, 5, 5, 12, 240, 5, 5, 30, 240, 5, 5, 212, 232, 5, 5, 108, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 224, 192, 0, 0, 0, 5, 5, 148, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 240, 192, 0, 0, 0, 5, 5, 221, 231, 192, 0, 0, 0, 5, 5, 38, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 238, 5, 5, 181, 230, 5, 5, 76, 226, 5, 5, 83, 232, 5, 5, 175, 230, 5, 5, 42, 240, 5, 5, 32, 240, 5, 5, 51, 240, 5, 5, 28, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 240, 192, 0, 0, 0, 5, 5, 209, 228, 5, 5, 31, 240, 192, 0, 0, 0, 5, 5, 40, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 240, 5, 5, 44, 240, 192, 0, 0, 0, 5, 5, 64, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 240, 5, 5, 72, 240, 5, 5, 47, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 228, 5, 5, 62, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 230, 5, 5, 163, 224, 5, 5, 66, 240, 192, 0, 0, 0, 5, 5, 58, 240, 192, 0, 0, 0, 5, 5, 55, 240, 192, 0, 0, 0, 5, 5, 68, 240, 5, 5, 54, 240, 5, 5, 132, 236, 5, 5, 164, 224, 192, 0, 0, 0, 5, 5, 183, 232, 5, 5, 195, 232, 5, 5, 67, 240, 192, 0, 0, 0, 5, 5, 53, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 228, 5, 5, 60, 240, 192, 0, 0, 0, 5, 5, 153, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 240, 5, 5, 102, 228, 5, 5, 48, 240, 5, 5, 212, 227, 192, 0, 0, 0, 5, 5, 221, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 228, 5, 5, 49, 240, 5, 5, 50, 240, 192, 0, 0, 0, 5, 5, 57, 240, 5, 5, 62, 240, 5, 5, 65, 240, 5, 5, 140, 225, 192, 0, 0, 0, 5, 5, 185, 232, 5, 5, 247, 227, 5, 5, 70, 240, 5, 5, 69, 240, 5, 5, 157, 236, 5, 5, 71, 240, 5, 5, 73, 240, 5, 5, 75, 240, 5, 5, 133, 236, 5, 5, 74, 240, 5, 5, 76, 240, 5, 5, 77, 240, 5, 5, 61, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 227, 5, 5, 79, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 240, 5, 5, 87, 240, 5, 5, 90, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 240, 5, 5, 240, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 240, 5, 5, 84, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 240, 5, 5, 85, 240, 192, 0, 0, 0, 5, 5, 94, 240, 192, 0, 0, 0, 5, 5, 82, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 240, 5, 5, 89, 240, 5, 5, 83, 240, 5, 5, 16, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 240, 5, 5, 157, 229, 5, 5, 243, 229, 5, 5, 80, 240, 5, 5, 81, 240, 192, 0, 0, 0, 5, 5, 88, 240, 5, 5, 3, 228, 5, 5, 91, 240, 5, 5, 154, 236, 5, 5, 96, 240, 192, 0, 0, 0, 5, 5, 97, 240, 5, 5, 173, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 240, 5, 5, 98, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 240, 192, 0, 0, 0, 5, 5, 17, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 229, 192, 0, 0, 0, 5, 5, 40, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 240, 192, 0, 0, 0, 5, 5, 107, 240, 192, 0, 0, 0, 5, 5, 101, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 224, 5, 5, 105, 240, 192, 0, 0, 0, 5, 5, 113, 240, 5, 5, 96, 224, 5, 5, 123, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 231, 5, 5, 118, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 240, 5, 5, 44, 230, 5, 5, 92, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 240, 5, 5, 117, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 240, 192, 0, 0, 0, 5, 5, 115, 240, 5, 5, 45, 234, 192, 0, 0, 0, 5, 5, 111, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 240, 5, 5, 206, 230, 5, 5, 222, 231, 5, 5, 198, 234, 5, 5, 114, 240, 192, 0, 0, 0, 5, 5, 110, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 240, 5, 5, 132, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 236, 5, 5, 156, 236, 5, 5, 121, 240, 5, 5, 108, 235, 5, 5, 216, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 240, 5, 5, 140, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 238, 192, 0, 0, 0, 5, 5, 127, 240, 5, 5, 126, 240, 5, 5, 228, 226, 5, 5, 124, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 231, 192, 0, 0, 0, 5, 5, 227, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 240, 192, 0, 0, 0, 5, 5, 171, 227, 5, 5, 51, 237, 5, 5, 141, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 225, 192, 0, 0, 0, 5, 5, 128, 240, 192, 0, 0, 0, 5, 5, 135, 229, 5, 5, 137, 240, 192, 0, 0, 0, 5, 5, 133, 240, 192, 0, 0, 0, 5, 5, 131, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 240, 5, 5, 130, 240, 5, 5, 132, 240, 192, 0, 0, 0, 5, 5, 125, 228, 5, 5, 138, 240, 192, 0, 0, 0, 5, 5, 139, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 240, 192, 0, 0, 0, 5, 5, 207, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 231, 192, 0, 0, 0, 5, 5, 154, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 237, 5, 5, 153, 240, 5, 5, 150, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 240, 192, 0, 0, 0, 5, 5, 149, 240, 192, 0, 0, 0, 5, 5, 142, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 232, 5, 5, 148, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 240, 5, 5, 238, 233, 192, 0, 0, 0, 5, 5, 147, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 229, 192, 0, 0, 0, 5, 5, 43, 228, 5, 5, 151, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 240, 192, 0, 0, 0, 5, 5, 156, 240, 5, 5, 145, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 230, 192, 0, 0, 0, 5, 5, 163, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 240, 192, 0, 0, 0, 5, 5, 136, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 231, 5, 5, 125, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 232, 5, 5, 160, 240, 5, 5, 162, 240, 5, 5, 164, 240, 5, 5, 166, 240, 5, 5, 0, 136, 5, 5, 44, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 240, 5, 5, 132, 228, 5, 5, 207, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 240, 192, 0, 0, 0, 5, 5, 30, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 240, 5, 5, 174, 240, 5, 5, 178, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 221, 5, 5, 186, 240, 192, 0, 0, 0, 5, 5, 181, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 235, 5, 5, 185, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 240, 192, 0, 0, 0, 5, 5, 147, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 240, 5, 5, 139, 233, 5, 5, 189, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 240, 192, 0, 0, 0, 5, 5, 187, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 224, 5, 5, 240, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 240, 192, 0, 0, 0, 5, 5, 126, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 240, 192, 0, 0, 0, 5, 5, 196, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 248, 5, 5, 178, 227, 5, 5, 90, 230, 5, 5, 116, 231, 5, 5, 111, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 225, 5, 5, 28, 248, 192, 0, 0, 0, 5, 5, 29, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 227, 5, 5, 31, 248, 5, 5, 33, 248, 5, 5, 32, 248, 5, 5, 224, 233, 5, 5, 50, 235, 5, 5, 30, 248, 5, 5, 82, 233, 5, 5, 99, 236, 5, 5, 148, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 234, 5, 5, 36, 248, 5, 5, 0, 212, 5, 5, 35, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 248, 192, 0, 0, 0, 5, 5, 139, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 235, 5, 5, 38, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 248, 5, 5, 59, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 248, 5, 5, 34, 248, 5, 5, 43, 248, 5, 5, 41, 248, 192, 0, 0, 0, 5, 5, 37, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 248, 5, 5, 48, 248, 5, 5, 50, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 248, 5, 5, 46, 248, 5, 5, 44, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 248, 192, 0, 0, 0, 5, 5, 123, 232, 5, 5, 31, 233, 192, 0, 0, 0, 5, 5, 52, 248, 5, 5, 26, 227, 5, 5, 136, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 248, 5, 5, 53, 248, 5, 5, 60, 248, 5, 5, 66, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 248, 5, 5, 150, 234, 192, 0, 0, 0, 5, 5, 249, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 248, 5, 5, 64, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 248, 5, 5, 161, 230, 192, 0, 0, 0, 5, 5, 141, 237, 5, 5, 56, 248, 5, 5, 57, 248, 5, 5, 59, 248, 5, 5, 63, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 248, 5, 5, 174, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 226, 192, 0, 0, 0, 5, 5, 143, 233, 192, 0, 0, 0, 5, 5, 122, 226, 5, 5, 67, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 248, 5, 5, 70, 248, 5, 5, 73, 248, 5, 5, 71, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 236, 5, 5, 61, 248, 192, 0, 0, 0, 5, 5, 131, 234, 192, 0, 0, 0, 5, 5, 231, 234, 5, 5, 183, 237, 192, 0, 0, 0, 5, 5, 79, 248, 192, 0, 0, 0, 5, 5, 225, 230, 192, 0, 0, 0, 5, 5, 76, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 229, 5, 5, 88, 248, 5, 5, 74, 248, 192, 0, 0, 0, 5, 5, 77, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 248, 5, 5, 86, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 236, 5, 5, 81, 248, 5, 5, 177, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 248, 192, 0, 0, 0, 5, 5, 129, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 227, 192, 0, 0, 0, 5, 5, 101, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 248, 5, 5, 100, 248, 5, 5, 102, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 248, 192, 0, 0, 0, 5, 5, 93, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 227, 192, 0, 0, 0, 5, 5, 211, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 248, 5, 5, 99, 248, 5, 5, 89, 248, 5, 5, 90, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 248, 5, 5, 112, 248, 192, 0, 0, 0, 5, 5, 109, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 248, 192, 0, 0, 0, 5, 5, 186, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 248, 5, 5, 115, 248, 5, 5, 110, 248, 192, 0, 0, 0, 5, 5, 104, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 248, 192, 0, 0, 0, 5, 5, 116, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 248, 5, 5, 118, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 248, 5, 5, 106, 248, 5, 5, 119, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 248, 192, 0, 0, 0, 5, 5, 196, 232, 5, 5, 128, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 248, 5, 5, 58, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 248, 5, 5, 156, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 235, 5, 5, 18, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 228, 192, 0, 0, 0, 5, 5, 247, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 227, 5, 5, 11, 242, 5, 5, 86, 236, 5, 5, 206, 247, 5, 5, 117, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 224, 5, 5, 207, 247, 192, 0, 0, 0, 5, 5, 254, 232, 5, 5, 222, 224, 192, 0, 0, 0, 5, 5, 30, 239, 192, 0, 0, 0, 5, 5, 161, 233, 192, 0, 0, 0, 5, 5, 208, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 247, 5, 5, 232, 248, 5, 5, 210, 247, 192, 0, 0, 0, 5, 5, 252, 236, 5, 5, 211, 247, 192, 0, 0, 0, 5, 5, 0, 158, 5, 5, 233, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 225, 192, 0, 0, 0, 5, 5, 159, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 247, 192, 0, 0, 0, 5, 5, 31, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 224, 192, 0, 0, 0, 5, 5, 38, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 224, 5, 5, 27, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 238, 5, 5, 213, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 239, 192, 0, 0, 0, 5, 5, 117, 236, 5, 5, 243, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 248, 5, 5, 153, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 247, 5, 5, 219, 247, 5, 5, 111, 229, 5, 5, 220, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 247, 5, 5, 221, 247, 192, 0, 0, 0, 5, 5, 17, 233, 5, 5, 165, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 230, 5, 5, 80, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 247, 192, 0, 0, 0, 5, 5, 225, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 227, 192, 0, 0, 0, 5, 5, 0, 219, 5, 5, 229, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 247, 5, 5, 222, 247, 5, 5, 230, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 234, 5, 5, 233, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 247, 192, 0, 0, 0, 5, 5, 234, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 248, 5, 5, 193, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 235, 192, 0, 0, 0, 5, 5, 62, 236, 192, 0, 0, 0, 5, 5, 9, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 228, 5, 5, 47, 227, 192, 0, 0, 0, 5, 5, 58, 227, 5, 5, 223, 230, 5, 5, 113, 233, 5, 5, 94, 245, 5, 5, 170, 229, 5, 5, 35, 229, 5, 5, 95, 245, 5, 5, 96, 245, 5, 5, 97, 245, 192, 0, 0, 0, 5, 5, 98, 245, 5, 5, 99, 245, 5, 5, 100, 245, 5, 5, 101, 245, 5, 5, 149, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 249, 192, 0, 0, 0, 5, 5, 134, 249, 192, 0, 0, 0, 5, 5, 132, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 228, 192, 0, 0, 0, 5, 5, 135, 249, 5, 5, 31, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 236, 5, 5, 56, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 239, 5, 5, 54, 228, 5, 5, 223, 225, 5, 5, 176, 226, 5, 5, 175, 232, 5, 5, 21, 228, 5, 5, 48, 239, 5, 5, 49, 239, 5, 5, 50, 234, 5, 5, 163, 232, 5, 5, 50, 239, 5, 5, 33, 232, 192, 0, 0, 0, 5, 5, 233, 235, 5, 5, 125, 236, 5, 5, 234, 235, 5, 5, 55, 228, 192, 0, 0, 0, 5, 5, 128, 228, 5, 5, 244, 227, 5, 5, 51, 239, 5, 5, 52, 239, 5, 5, 252, 235, 5, 5, 53, 239, 5, 5, 195, 235, 5, 5, 39, 226, 5, 5, 131, 230, 192, 0, 0, 0, 5, 5, 203, 233, 5, 5, 131, 226, 5, 5, 40, 233, 5, 5, 89, 226, 5, 5, 37, 229, 5, 5, 178, 237, 5, 5, 54, 239, 5, 5, 55, 239, 5, 5, 226, 231, 5, 5, 85, 238, 5, 5, 84, 233, 192, 0, 0, 0, 5, 5, 89, 237, 5, 5, 219, 233, 5, 5, 159, 237, 5, 5, 56, 239, 5, 5, 237, 237, 5, 5, 68, 225, 5, 5, 58, 239, 5, 5, 57, 239, 192, 0, 0, 0, 5, 5, 127, 236, 5, 5, 59, 239, 5, 5, 60, 239, 5, 5, 61, 239, 5, 5, 114, 233, 5, 5, 62, 239, 5, 5, 73, 233, 5, 5, 63, 239, 5, 5, 64, 239, 5, 5, 233, 224, 5, 5, 253, 237, 5, 5, 65, 239, 5, 5, 192, 227, 5, 5, 134, 225, 5, 5, 66, 239, 5, 5, 67, 239, 5, 5, 66, 227, 5, 5, 227, 235, 5, 5, 124, 236, 5, 5, 68, 239, 5, 5, 183, 226, 5, 5, 98, 235, 5, 5, 171, 224, 5, 5, 69, 239, 5, 5, 70, 239, 192, 0, 0, 0, 5, 5, 185, 228, 5, 5, 244, 234, 5, 5, 225, 236, 5, 5, 71, 239, 5, 5, 13, 235, 5, 5, 72, 239, 5, 5, 199, 236, 5, 5, 245, 227, 5, 5, 73, 239, 5, 5, 177, 233, 5, 5, 204, 233, 5, 5, 74, 239, 5, 5, 109, 232, 5, 5, 252, 237, 5, 5, 75, 239, 5, 5, 121, 231, 5, 5, 247, 225, 5, 5, 76, 239, 5, 5, 98, 226, 5, 5, 90, 229, 5, 5, 77, 239, 5, 5, 78, 239, 5, 5, 169, 233, 5, 5, 79, 239, 5, 5, 207, 225, 5, 5, 80, 239, 5, 5, 12, 230, 5, 5, 41, 238, 5, 5, 81, 239, 5, 5, 20, 234, 192, 0, 0, 0, 5, 5, 126, 236, 5, 5, 23, 231, 5, 5, 82, 239, 5, 5, 213, 225, 5, 5, 225, 227, 5, 5, 83, 239, 5, 5, 137, 235, 5, 5, 84, 239, 5, 5, 85, 239, 5, 5, 213, 234, 5, 5, 86, 239, 5, 5, 87, 239, 5, 5, 88, 239, 5, 5, 179, 224, 5, 5, 91, 239, 5, 5, 89, 239, 5, 5, 30, 236, 5, 5, 90, 239, 5, 5, 219, 230, 5, 5, 92, 239, 192, 0, 0, 0, 5, 5, 93, 239, 5, 5, 94, 239, 5, 5, 95, 239, 5, 5, 145, 235, 5, 5, 57, 236, 5, 5, 0, 215, 5, 5, 96, 239, 5, 5, 45, 232, 5, 5, 97, 239, 5, 5, 197, 228, 5, 5, 167, 230, 5, 5, 98, 239, 5, 5, 99, 239, 5, 5, 5, 231, 5, 5, 19, 234, 5, 5, 100, 239, 5, 5, 101, 239, 5, 5, 168, 229, 5, 5, 249, 231, 5, 5, 102, 239, 5, 5, 103, 239, 5, 5, 54, 232, 5, 5, 104, 239, 5, 5, 105, 239, 5, 5, 28, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 225, 5, 5, 34, 249, 192, 0, 0, 0, 5, 5, 35, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 235, 5, 5, 207, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 227, 5, 5, 245, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 249, 5, 5, 8, 224, 5, 5, 174, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 227, 5, 5, 126, 249, 192, 0, 0, 0, 5, 5, 184, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 224, 5, 5, 154, 237, 5, 5, 174, 226, 192, 0, 0, 0, 5, 5, 5, 227, 5, 5, 130, 224, 5, 5, 66, 237, 5, 5, 69, 235, 5, 5, 0, 186, 5, 5, 123, 237, 5, 5, 7, 228, 5, 5, 216, 237, 5, 5, 77, 226, 5, 5, 12, 234, 5, 5, 216, 231, 5, 5, 63, 224, 5, 5, 14, 227, 5, 5, 12, 238, 5, 5, 53, 227, 5, 5, 55, 226, 5, 5, 106, 228, 5, 5, 80, 245, 5, 5, 81, 245, 5, 5, 85, 234, 5, 5, 71, 227, 5, 5, 82, 245, 5, 5, 117, 225, 5, 5, 185, 230, 5, 5, 103, 226, 5, 5, 138, 227, 5, 5, 83, 245, 5, 5, 70, 237, 5, 5, 84, 245, 5, 5, 70, 228, 5, 5, 239, 227, 5, 5, 85, 245, 5, 5, 40, 230, 5, 5, 96, 230, 5, 5, 49, 237, 5, 5, 56, 238, 5, 5, 86, 245, 5, 5, 87, 245, 5, 5, 90, 245, 5, 5, 88, 245, 5, 5, 89, 245, 5, 5, 30, 233, 5, 5, 167, 226, 5, 5, 250, 225, 5, 5, 91, 245, 5, 5, 136, 233, 5, 5, 13, 233, 5, 5, 71, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 242, 5, 5, 166, 231, 5, 5, 92, 245, 5, 5, 159, 229, 192, 0, 0, 0, 5, 5, 38, 238, 5, 5, 93, 245, 5, 5, 26, 238, 5, 5, 220, 232, 5, 5, 159, 238, 5, 5, 153, 238, 5, 5, 48, 237, 192, 0, 0, 0, 5, 5, 75, 237, 5, 5, 4, 233, 5, 5, 160, 236, 5, 5, 199, 226, 5, 5, 250, 224, 192, 0, 0, 0, 5, 5, 34, 233, 5, 5, 32, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 227, 192, 0, 0, 0, 5, 5, 33, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 249, 5, 5, 164, 226, 5, 5, 132, 237, 5, 5, 195, 226, 5, 5, 20, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 249, 5, 5, 198, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 237, 5, 5, 121, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 249, 192, 0, 0, 0, 5, 5, 81, 238, 5, 5, 131, 231, 5, 5, 72, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 249, 192, 0, 0, 0, 5, 5, 75, 249, 192, 0, 0, 0, 5, 5, 74, 249, 192, 0, 0, 0, 5, 5, 200, 237, 5, 5, 73, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 237, 5, 5, 76, 249, 192, 0, 0, 0, 5, 5, 84, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 173, 5, 5, 208, 225, 192, 0, 0, 0, 5, 5, 81, 249, 5, 5, 82, 249, 192, 0, 0, 0, 5, 5, 160, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 249, 5, 5, 78, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 249, 5, 5, 83, 249, 192, 0, 0, 0, 5, 5, 14, 229, 5, 5, 80, 249, 5, 5, 238, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 249, 5, 5, 91, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 229, 192, 0, 0, 0, 5, 5, 70, 227, 5, 5, 68, 249, 5, 5, 85, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 234, 192, 0, 0, 0, 5, 5, 105, 228, 192, 0, 0, 0, 5, 5, 86, 249, 5, 5, 87, 249, 5, 5, 89, 249, 5, 5, 29, 226, 5, 5, 90, 249, 192, 0, 0, 0, 5, 5, 93, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 249, 5, 5, 173, 236, 192, 0, 0, 0, 5, 5, 8, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 249, 5, 5, 15, 229, 5, 5, 96, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 234, 5, 5, 99, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 224, 5, 5, 71, 238, 192, 0, 0, 0, 5, 5, 97, 249, 192, 0, 0, 0, 5, 5, 98, 249, 5, 5, 100, 249, 192, 0, 0, 0, 5, 5, 106, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 249, 5, 5, 101, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 249, 5, 5, 108, 249, 5, 5, 109, 249, 192, 0, 0, 0, 5, 5, 63, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 243, 5, 5, 144, 225, 5, 5, 107, 249, 5, 5, 112, 249, 5, 5, 253, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 249, 5, 5, 111, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 225, 5, 5, 160, 224, 192, 0, 0, 0, 5, 5, 116, 249, 5, 5, 113, 249, 192, 0, 0, 0, 5, 5, 15, 226, 192, 0, 0, 0, 5, 5, 117, 249, 192, 0, 0, 0, 5, 5, 114, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 225, 192, 0, 0, 0, 5, 5, 60, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 249, 192, 0, 0, 0, 5, 5, 21, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 249, 5, 5, 121, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 249, 192, 0, 0, 0, 5, 5, 123, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 233, 5, 5, 252, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 224, 5, 5, 80, 237, 5, 5, 64, 227, 5, 5, 207, 235, 192, 0, 0, 0, 5, 5, 5, 245, 5, 5, 24, 238, 5, 5, 6, 245, 5, 5, 126, 230, 5, 5, 205, 232, 5, 5, 150, 227, 5, 5, 7, 245, 5, 5, 8, 245, 5, 5, 9, 245, 5, 5, 239, 237, 5, 5, 10, 245, 5, 5, 11, 245, 5, 5, 13, 245, 5, 5, 12, 245, 5, 5, 14, 245, 5, 5, 15, 245, 5, 5, 99, 232, 5, 5, 16, 245, 5, 5, 42, 237, 5, 5, 17, 245, 5, 5, 156, 228, 192, 0, 0, 0, 5, 5, 18, 245, 5, 5, 19, 245, 5, 5, 157, 228, 5, 5, 20, 245, 5, 5, 156, 226, 5, 5, 8, 230, 5, 5, 21, 245, 5, 5, 16, 224, 5, 5, 228, 227, 5, 5, 73, 227, 5, 5, 22, 245, 192, 0, 0, 0, 5, 5, 23, 245, 5, 5, 24, 245, 5, 5, 25, 245, 5, 5, 142, 226, 5, 5, 29, 228, 192, 0, 0, 0, 5, 5, 130, 233, 5, 5, 144, 241, 5, 5, 255, 236, 5, 5, 53, 235, 5, 5, 103, 237, 5, 5, 26, 245, 5, 5, 142, 237, 5, 5, 27, 245, 5, 5, 151, 235, 5, 5, 16, 227, 192, 0, 0, 0, 5, 5, 65, 225, 5, 5, 55, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 224, 5, 5, 69, 224, 192, 0, 0, 0, 5, 5, 70, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 224, 5, 5, 200, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 230, 5, 5, 105, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 232, 5, 5, 202, 236, 192, 0, 0, 0, 5, 5, 29, 232, 5, 5, 236, 235, 192, 0, 0, 0, 5, 5, 81, 227, 5, 5, 157, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 236, 192, 0, 0, 0, 5, 5, 29, 237, 5, 5, 202, 228, 192, 0, 0, 0, 5, 5, 94, 243, 5, 5, 75, 226, 5, 5, 95, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 227, 5, 5, 146, 237, 192, 0, 0, 0, 5, 5, 198, 228, 5, 5, 8, 237, 5, 5, 189, 234, 5, 5, 244, 229, 5, 5, 243, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 234, 192, 0, 0, 0, 5, 5, 98, 243, 5, 5, 96, 243, 5, 5, 100, 243, 192, 0, 0, 0, 5, 5, 102, 243, 5, 5, 99, 243, 5, 5, 167, 225, 5, 5, 234, 231, 192, 0, 0, 0, 5, 5, 212, 225, 5, 5, 97, 243, 192, 0, 0, 0, 5, 5, 148, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 230, 5, 5, 35, 224, 5, 5, 19, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 234, 5, 5, 201, 233, 5, 5, 104, 233, 5, 5, 47, 234, 5, 5, 104, 243, 5, 5, 103, 243, 5, 5, 76, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 235, 5, 5, 235, 235, 5, 5, 105, 243, 192, 0, 0, 0, 5, 5, 108, 243, 192, 0, 0, 0, 5, 5, 114, 234, 5, 5, 254, 237, 5, 5, 107, 243, 5, 5, 181, 225, 192, 0, 0, 0, 5, 5, 120, 234, 192, 0, 0, 0, 5, 5, 109, 243, 5, 5, 240, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 234, 5, 5, 56, 227, 192, 0, 0, 0, 5, 5, 99, 233, 5, 5, 235, 224, 5, 5, 213, 233, 5, 5, 62, 237, 5, 5, 110, 243, 5, 5, 128, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 243, 5, 5, 120, 225, 5, 5, 114, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 243, 5, 5, 112, 243, 192, 0, 0, 0, 5, 5, 113, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 230, 5, 5, 36, 224, 192, 0, 0, 0, 5, 5, 212, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 226, 5, 5, 232, 233, 192, 0, 0, 0, 5, 5, 115, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 224, 192, 0, 0, 0, 5, 5, 44, 226, 5, 5, 118, 243, 5, 5, 116, 243, 5, 5, 117, 243, 5, 5, 152, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 236, 5, 5, 120, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 243, 5, 5, 52, 232, 192, 0, 0, 0, 5, 5, 55, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 237, 5, 5, 138, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 243, 5, 5, 95, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 243, 192, 0, 0, 0, 5, 5, 58, 224, 5, 5, 49, 236, 192, 0, 0, 0, 5, 5, 126, 243, 5, 5, 128, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 236, 192, 0, 0, 0, 5, 5, 163, 225, 192, 0, 0, 0, 5, 5, 57, 244, 192, 0, 0, 0, 5, 5, 127, 239, 192, 0, 0, 0, 5, 5, 130, 239, 192, 0, 0, 0, 5, 5, 128, 239, 192, 0, 0, 0, 5, 5, 129, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 239, 5, 5, 165, 235, 5, 5, 45, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 235, 192, 0, 0, 0, 5, 5, 131, 239, 192, 0, 0, 0, 5, 5, 188, 236, 5, 5, 92, 227, 5, 5, 138, 239, 5, 5, 115, 232, 192, 0, 0, 0, 5, 5, 134, 239, 5, 5, 133, 239, 5, 5, 27, 233, 5, 5, 135, 239, 192, 0, 0, 0, 5, 5, 137, 239, 5, 5, 76, 238, 5, 5, 136, 239, 5, 5, 36, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 239, 5, 5, 140, 239, 192, 0, 0, 0, 5, 5, 144, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 229, 5, 5, 139, 239, 5, 5, 142, 239, 5, 5, 177, 237, 192, 0, 0, 0, 5, 5, 145, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 239, 5, 5, 148, 239, 5, 5, 116, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 229, 5, 5, 147, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 239, 5, 5, 25, 237, 5, 5, 123, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 239, 192, 0, 0, 0, 5, 5, 77, 227, 192, 0, 0, 0, 5, 5, 152, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 225, 5, 5, 153, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 226, 192, 0, 0, 0, 5, 5, 154, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 239, 5, 5, 157, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 239, 192, 0, 0, 0, 5, 5, 158, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 236, 5, 5, 36, 249, 5, 5, 119, 232, 5, 5, 49, 238, 5, 5, 168, 231, 5, 5, 38, 249, 5, 5, 39, 249, 5, 5, 37, 249, 192, 0, 0, 0, 5, 5, 242, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 237, 5, 5, 6, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 249, 5, 5, 41, 249, 5, 5, 90, 227, 5, 5, 40, 249, 5, 5, 210, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 249, 5, 5, 188, 229, 192, 0, 0, 0, 5, 5, 6, 225, 192, 0, 0, 0, 5, 5, 100, 234, 5, 5, 45, 249, 5, 5, 43, 249, 5, 5, 130, 228, 5, 5, 48, 249, 192, 0, 0, 0, 5, 5, 49, 249, 5, 5, 155, 228, 5, 5, 190, 230, 5, 5, 109, 229, 5, 5, 221, 233, 5, 5, 50, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 249, 5, 5, 47, 249, 5, 5, 86, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 249, 192, 0, 0, 0, 5, 5, 54, 225, 192, 0, 0, 0, 5, 5, 91, 238, 192, 0, 0, 0, 5, 5, 81, 225, 5, 5, 51, 249, 5, 5, 54, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 249, 5, 5, 55, 249, 5, 5, 167, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 230, 5, 5, 137, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 249, 5, 5, 57, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 249, 5, 5, 60, 249, 5, 5, 61, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 249, 5, 5, 62, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 224, 192, 0, 0, 0, 5, 5, 198, 236, 5, 5, 107, 233, 192, 0, 0, 0, 5, 5, 217, 229, 5, 5, 230, 237, 5, 5, 68, 236, 5, 5, 9, 230, 192, 0, 0, 0, 5, 5, 190, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 246, 5, 5, 181, 246, 5, 5, 182, 246, 5, 5, 155, 237, 5, 5, 218, 225, 5, 5, 184, 246, 5, 5, 183, 246, 5, 5, 185, 246, 5, 5, 186, 246, 5, 5, 39, 232, 5, 5, 187, 246, 5, 5, 188, 246, 192, 0, 0, 0, 5, 5, 70, 226, 5, 5, 206, 225, 5, 5, 189, 246, 5, 5, 191, 246, 192, 0, 0, 0, 5, 5, 190, 246, 192, 0, 0, 0, 5, 5, 186, 226, 5, 5, 192, 246, 5, 5, 193, 246, 5, 5, 194, 246, 5, 5, 19, 226, 5, 5, 200, 224, 5, 5, 225, 237, 5, 5, 44, 231, 5, 5, 19, 224, 5, 5, 202, 226, 5, 5, 195, 246, 5, 5, 196, 246, 5, 5, 17, 237, 5, 5, 87, 232, 5, 5, 41, 229, 5, 5, 241, 234, 5, 5, 7, 227, 5, 5, 198, 246, 5, 5, 197, 246, 5, 5, 200, 246, 5, 5, 199, 246, 5, 5, 105, 231, 5, 5, 201, 246, 5, 5, 202, 246, 5, 5, 48, 232, 5, 5, 203, 246, 5, 5, 49, 232, 5, 5, 204, 246, 5, 5, 99, 224, 5, 5, 205, 246, 5, 5, 206, 246, 5, 5, 207, 246, 5, 5, 208, 246, 5, 5, 209, 246, 5, 5, 88, 238, 5, 5, 210, 246, 5, 5, 211, 246, 5, 5, 72, 228, 5, 5, 212, 246, 5, 5, 189, 236, 5, 5, 86, 234, 5, 5, 104, 224, 5, 5, 47, 230, 5, 5, 213, 246, 5, 5, 40, 232, 5, 5, 179, 230, 192, 0, 0, 0, 5, 5, 214, 246, 5, 5, 215, 246, 5, 5, 216, 246, 5, 5, 217, 246, 5, 5, 218, 246, 5, 5, 219, 246, 5, 5, 220, 246, 192, 0, 0, 0, 5, 5, 221, 246, 5, 5, 222, 246, 5, 5, 223, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 246, 5, 5, 225, 246, 5, 5, 226, 246, 5, 5, 228, 246, 5, 5, 227, 246, 192, 0, 0, 0, 5, 5, 229, 246, 5, 5, 103, 234, 5, 5, 105, 230, 5, 5, 230, 246, 5, 5, 231, 246, 5, 5, 232, 246, 5, 5, 81, 237, 5, 5, 233, 246, 5, 5, 43, 235, 5, 5, 234, 246, 5, 5, 235, 246, 192, 0, 0, 0, 5, 5, 236, 246, 5, 5, 237, 246, 5, 5, 239, 246, 5, 5, 238, 246, 5, 5, 240, 246, 5, 5, 233, 226, 5, 5, 2, 231, 5, 5, 241, 246, 5, 5, 242, 246, 5, 5, 144, 228, 5, 5, 83, 236, 5, 5, 181, 224, 5, 5, 243, 246, 5, 5, 244, 246, 5, 5, 245, 246, 5, 5, 140, 236, 5, 5, 246, 246, 5, 5, 13, 238, 5, 5, 247, 246, 5, 5, 238, 231, 192, 0, 0, 0, 5, 5, 248, 246, 5, 5, 249, 246, 5, 5, 252, 229, 5, 5, 250, 246, 5, 5, 114, 235, 5, 5, 244, 233, 5, 5, 252, 246, 5, 5, 251, 246, 5, 5, 22, 225, 5, 5, 76, 227, 5, 5, 253, 246, 5, 5, 254, 246, 5, 5, 184, 235, 5, 5, 255, 246, 5, 5, 2, 247, 5, 5, 124, 226, 5, 5, 149, 235, 5, 5, 3, 247, 5, 5, 4, 247, 5, 5, 5, 247, 5, 5, 209, 232, 5, 5, 60, 234, 5, 5, 6, 247, 5, 5, 7, 247, 5, 5, 8, 247, 5, 5, 9, 247, 5, 5, 10, 247, 5, 5, 144, 237, 5, 5, 11, 247, 5, 5, 103, 225, 5, 5, 176, 230, 5, 5, 12, 247, 192, 0, 0, 0, 5, 5, 13, 247, 5, 5, 14, 247, 5, 5, 15, 247, 192, 0, 0, 0, 5, 5, 23, 235, 5, 5, 16, 247, 5, 5, 136, 230, 5, 5, 50, 225, 5, 5, 36, 238, 5, 5, 195, 228, 192, 0, 0, 0, 5, 5, 63, 235, 5, 5, 19, 247, 5, 5, 17, 247, 5, 5, 18, 247, 5, 5, 20, 247, 5, 5, 221, 225, 5, 5, 108, 228, 5, 5, 16, 229, 5, 5, 210, 230, 5, 5, 21, 247, 5, 5, 22, 247, 192, 0, 0, 0, 5, 5, 23, 247, 5, 5, 31, 247, 5, 5, 24, 247, 5, 5, 25, 247, 5, 5, 26, 247, 5, 5, 68, 232, 5, 5, 65, 247, 5, 5, 5, 226, 5, 5, 27, 247, 192, 0, 0, 0, 5, 5, 28, 247, 5, 5, 29, 247, 5, 5, 252, 225, 5, 5, 196, 230, 5, 5, 30, 247, 192, 0, 0, 0, 5, 5, 32, 247, 5, 5, 33, 247, 5, 5, 34, 247, 5, 5, 162, 237, 192, 0, 0, 0, 5, 5, 35, 247, 5, 5, 93, 231, 192, 0, 0, 0, 5, 5, 36, 247, 5, 5, 94, 231, 5, 5, 37, 247, 5, 5, 38, 247, 5, 5, 216, 226, 5, 5, 0, 213, 5, 5, 39, 247, 5, 5, 40, 247, 5, 5, 41, 247, 192, 0, 0, 0, 5, 5, 42, 247, 5, 5, 43, 247, 5, 5, 44, 247, 5, 5, 45, 247, 192, 0, 0, 0, 5, 5, 46, 247, 5, 5, 225, 228, 5, 5, 49, 247, 5, 5, 47, 247, 5, 5, 48, 247, 192, 0, 0, 0, 5, 5, 50, 247, 5, 5, 51, 247, 5, 5, 23, 230, 5, 5, 52, 247, 5, 5, 53, 247, 5, 5, 54, 247, 5, 5, 55, 247, 5, 5, 56, 247, 5, 5, 57, 247, 5, 5, 58, 247, 5, 5, 59, 247, 5, 5, 60, 247, 5, 5, 194, 229, 192, 0, 0, 0, 5, 5, 61, 247, 5, 5, 245, 229, 5, 5, 62, 247, 5, 5, 63, 247, 5, 5, 64, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 230, 5, 5, 163, 242, 5, 5, 255, 232, 5, 5, 164, 242, 192, 0, 0, 0, 5, 5, 51, 224, 5, 5, 226, 234, 5, 5, 45, 225, 5, 5, 210, 232, 5, 5, 165, 242, 5, 5, 72, 235, 5, 5, 166, 242, 5, 5, 84, 228, 5, 5, 167, 242, 5, 5, 168, 242, 5, 5, 204, 230, 5, 5, 82, 237, 5, 5, 60, 231, 5, 5, 63, 227, 5, 5, 221, 234, 5, 5, 169, 242, 5, 5, 252, 230, 5, 5, 170, 242, 192, 0, 0, 0, 5, 5, 61, 226, 5, 5, 231, 226, 5, 5, 132, 227, 5, 5, 171, 242, 5, 5, 172, 242, 5, 5, 22, 237, 5, 5, 173, 242, 192, 0, 0, 0, 5, 5, 174, 242, 5, 5, 255, 235, 5, 5, 175, 242, 5, 5, 176, 242, 5, 5, 177, 242, 5, 5, 178, 242, 5, 5, 12, 236, 5, 5, 179, 242, 5, 5, 183, 224, 5, 5, 165, 229, 5, 5, 180, 242, 192, 0, 0, 0, 5, 5, 149, 229, 5, 5, 181, 242, 5, 5, 182, 242, 5, 5, 183, 242, 192, 0, 0, 0, 5, 5, 184, 242, 5, 5, 185, 242, 192, 0, 0, 0, 5, 5, 171, 226, 5, 5, 108, 239, 192, 0, 0, 0, 5, 5, 11, 226, 192, 0, 0, 0, 5, 5, 110, 239, 5, 5, 109, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 239, 5, 5, 86, 226, 5, 5, 42, 236, 5, 5, 137, 236, 5, 5, 163, 237, 5, 5, 165, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 238, 5, 5, 114, 239, 5, 5, 113, 239, 192, 0, 0, 0, 5, 5, 0, 128, 5, 5, 141, 234, 192, 0, 0, 0, 5, 5, 115, 239, 192, 0, 0, 0, 5, 5, 177, 226, 5, 5, 58, 228, 5, 5, 101, 230, 5, 5, 76, 230, 5, 5, 220, 224, 5, 5, 116, 239, 192, 0, 0, 0, 5, 5, 82, 230, 5, 5, 22, 231, 5, 5, 131, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 239, 5, 5, 2, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 239, 192, 0, 0, 0, 5, 5, 238, 225, 5, 5, 12, 237, 192, 0, 0, 0, 5, 5, 25, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 239, 5, 5, 27, 237, 5, 5, 77, 235, 5, 5, 167, 231, 192, 0, 0, 0, 5, 5, 120, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 239, 192, 0, 0, 0, 5, 5, 122, 239, 5, 5, 52, 230, 5, 5, 49, 234, 5, 5, 85, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 236, 5, 5, 73, 230, 192, 0, 0, 0, 5, 5, 123, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 233, 192, 0, 0, 0, 5, 5, 124, 239, 192, 0, 0, 0, 5, 5, 226, 233, 5, 5, 146, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 239, 5, 5, 0, 141, 5, 5, 46, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 249, 5, 5, 167, 249, 5, 5, 55, 231, 192, 0, 0, 0, 5, 5, 152, 232, 5, 5, 25, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 235, 5, 5, 249, 235, 5, 5, 31, 228, 5, 5, 33, 227, 192, 0, 0, 0, 5, 5, 68, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 225, 5, 5, 172, 236, 5, 5, 168, 249, 5, 5, 23, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 236, 5, 5, 141, 249, 5, 5, 221, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 230, 5, 5, 193, 229, 192, 0, 0, 0, 5, 5, 0, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 235, 192, 0, 0, 0, 5, 5, 190, 235, 5, 5, 145, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 235, 192, 0, 0, 0, 5, 5, 144, 249, 5, 5, 160, 237, 5, 5, 146, 249, 5, 5, 191, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 228, 5, 5, 148, 249, 5, 5, 147, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 233, 192, 0, 0, 0, 5, 5, 52, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 249, 192, 0, 0, 0, 5, 5, 94, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 178, 5, 5, 189, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 232, 5, 5, 140, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 228, 192, 0, 0, 0, 5, 5, 189, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 226, 192, 0, 0, 0, 5, 5, 75, 229, 5, 5, 216, 230, 5, 5, 236, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 228, 5, 5, 217, 235, 192, 0, 0, 0, 5, 5, 0, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 235, 192, 0, 0, 0, 5, 5, 0, 142, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 249, 5, 5, 247, 249, 192, 0, 0, 0, 5, 5, 248, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 249, 192, 0, 0, 0, 5, 5, 60, 224, 192, 0, 0, 0, 5, 5, 249, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 249, 192, 0, 0, 0, 5, 5, 253, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 234, 5, 5, 173, 232, 192, 0, 0, 0, 5, 5, 93, 227, 5, 5, 112, 244, 5, 5, 113, 244, 5, 5, 114, 244, 5, 5, 238, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 236, 192, 0, 0, 0, 5, 5, 33, 237, 5, 5, 24, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 236, 5, 5, 219, 225, 5, 5, 108, 232, 5, 5, 10, 248, 5, 5, 102, 235, 5, 5, 175, 233, 5, 5, 193, 235, 5, 5, 73, 244, 5, 5, 161, 234, 5, 5, 31, 227, 5, 5, 17, 226, 5, 5, 11, 248, 5, 5, 0, 194, 5, 5, 200, 233, 5, 5, 12, 248, 5, 5, 244, 236, 5, 5, 85, 230, 5, 5, 54, 230, 5, 5, 230, 231, 5, 5, 221, 228, 5, 5, 13, 248, 5, 5, 69, 228, 192, 0, 0, 0, 5, 5, 14, 248, 5, 5, 15, 248, 192, 0, 0, 0, 5, 5, 16, 248, 5, 5, 87, 236, 5, 5, 215, 231, 192, 0, 0, 0, 5, 5, 129, 234, 5, 5, 17, 248, 192, 0, 0, 0, 5, 5, 163, 236, 5, 5, 81, 229, 5, 5, 62, 234, 192, 0, 0, 0, 5, 5, 18, 248, 5, 5, 19, 248, 5, 5, 11, 236, 5, 5, 38, 226, 5, 5, 20, 248, 5, 5, 21, 248, 5, 5, 183, 225, 5, 5, 22, 248, 5, 5, 23, 248, 192, 0, 0, 0, 5, 5, 184, 224, 5, 5, 24, 248, 5, 5, 25, 248, 5, 5, 135, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 245, 5, 5, 220, 245, 5, 5, 221, 245, 192, 0, 0, 0, 5, 5, 222, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 231, 5, 5, 223, 245, 5, 5, 224, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 226, 5, 5, 81, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 242, 5, 5, 21, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 250, 5, 5, 24, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 242, 192, 0, 0, 0, 5, 5, 18, 228, 192, 0, 0, 0, 5, 5, 60, 242, 5, 5, 61, 242, 5, 5, 62, 242, 5, 5, 63, 242, 5, 5, 64, 242, 5, 5, 65, 242, 5, 5, 79, 226, 5, 5, 143, 236, 5, 5, 114, 228, 5, 5, 108, 233, 5, 5, 3, 224, 5, 5, 195, 233, 192, 0, 0, 0, 5, 5, 66, 242, 5, 5, 52, 226, 5, 5, 164, 232, 5, 5, 67, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 228, 192, 0, 0, 0, 5, 5, 91, 224, 5, 5, 68, 242, 192, 0, 0, 0, 5, 5, 46, 226, 5, 5, 69, 242, 5, 5, 63, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 242, 5, 5, 82, 235, 5, 5, 49, 227, 5, 5, 71, 242, 5, 5, 139, 229, 192, 0, 0, 0, 5, 5, 72, 242, 5, 5, 178, 224, 192, 0, 0, 0, 5, 5, 73, 242, 192, 0, 0, 0, 5, 5, 61, 230, 5, 5, 74, 242, 5, 5, 75, 242, 5, 5, 160, 230, 5, 5, 76, 242, 5, 5, 77, 242, 5, 5, 78, 242, 5, 5, 117, 233, 5, 5, 132, 238, 5, 5, 17, 239, 5, 5, 91, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 230, 5, 5, 246, 236, 5, 5, 142, 234, 5, 5, 229, 235, 5, 5, 245, 224, 5, 5, 127, 232, 192, 0, 0, 0, 5, 5, 113, 224, 5, 5, 103, 230, 5, 5, 218, 243, 5, 5, 89, 233, 5, 5, 219, 243, 5, 5, 220, 243, 5, 5, 2, 229, 5, 5, 221, 243, 5, 5, 18, 238, 5, 5, 143, 234, 5, 5, 223, 243, 5, 5, 77, 228, 5, 5, 222, 243, 5, 5, 224, 243, 5, 5, 225, 243, 5, 5, 150, 230, 192, 0, 0, 0, 5, 5, 140, 228, 5, 5, 226, 243, 5, 5, 142, 230, 5, 5, 89, 227, 5, 5, 227, 243, 192, 0, 0, 0, 5, 5, 228, 243, 5, 5, 236, 224, 5, 5, 31, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 229, 5, 5, 229, 243, 5, 5, 19, 232, 5, 5, 230, 243, 5, 5, 231, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 243, 5, 5, 207, 231, 5, 5, 233, 243, 192, 0, 0, 0, 5, 5, 229, 232, 5, 5, 234, 243, 5, 5, 235, 243, 5, 5, 236, 243, 5, 5, 85, 243, 5, 5, 237, 243, 5, 5, 238, 243, 5, 5, 138, 230, 5, 5, 239, 243, 5, 5, 240, 243, 5, 5, 246, 237, 5, 5, 241, 243, 192, 0, 0, 0, 5, 5, 242, 243, 5, 5, 27, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 249, 5, 5, 254, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 250, 5, 5, 2, 250, 5, 5, 83, 227, 192, 0, 0, 0, 5, 5, 5, 250, 192, 0, 0, 0, 5, 5, 6, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 250, 5, 5, 7, 250, 5, 5, 10, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 250, 5, 5, 12, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 250, 192, 0, 0, 0, 5, 5, 228, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 250, 192, 0, 0, 0, 5, 5, 27, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 250, 192, 0, 0, 0, 5, 5, 32, 250, 192, 0, 0, 0, 5, 5, 29, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 250, 192, 0, 0, 0, 5, 5, 31, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 243, 5, 5, 65, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 229, 5, 5, 250, 227, 5, 5, 15, 250, 5, 5, 233, 231, 5, 5, 14, 250, 192, 0, 0, 0, 5, 5, 16, 250, 5, 5, 18, 250, 5, 5, 17, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 250, 192, 0, 0, 0, 5, 5, 210, 234, 192, 0, 0, 0, 5, 5, 20, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 249, 192, 0, 0, 0, 5, 5, 91, 230, 5, 5, 182, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 249, 5, 5, 184, 249, 5, 5, 185, 249, 5, 5, 186, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 249, 192, 0, 0, 0, 5, 5, 9, 224, 5, 5, 189, 249, 192, 0, 0, 0, 5, 5, 190, 249, 5, 5, 191, 249, 5, 5, 192, 249, 192, 0, 0, 0, 5, 5, 193, 249, 5, 5, 194, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 249, 5, 5, 196, 249, 5, 5, 66, 235, 192, 0, 0, 0, 5, 5, 197, 249, 5, 5, 198, 249, 5, 5, 199, 249, 5, 5, 200, 249, 5, 5, 201, 249, 5, 5, 202, 249, 5, 5, 218, 229, 5, 5, 203, 249, 5, 5, 204, 249, 5, 5, 205, 249, 5, 5, 206, 249, 5, 5, 207, 249, 192, 0, 0, 0, 5, 5, 208, 249, 192, 0, 0, 0, 5, 5, 209, 249, 5, 5, 210, 249, 192, 0, 0, 0, 5, 5, 211, 249, 5, 5, 212, 249, 5, 5, 213, 249, 5, 5, 214, 249, 5, 5, 215, 249, 5, 5, 216, 249, 5, 5, 217, 249, 5, 5, 218, 249, 5, 5, 212, 228, 192, 0, 0, 0, 5, 5, 219, 249, 5, 5, 220, 249, 5, 5, 221, 249, 5, 5, 222, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 232, 5, 5, 223, 249, 5, 5, 224, 249, 5, 5, 225, 249, 5, 5, 226, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 249, 5, 5, 228, 249, 5, 5, 229, 249, 5, 5, 230, 249, 5, 5, 231, 249, 5, 5, 232, 249, 5, 5, 233, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 249, 5, 5, 235, 249, 5, 5, 236, 249, 5, 5, 76, 224, 5, 5, 237, 249, 5, 5, 238, 249, 5, 5, 239, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 249, 5, 5, 241, 249, 5, 5, 37, 230, 5, 5, 242, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 231, 5, 5, 93, 247, 5, 5, 22, 228, 5, 5, 94, 247, 5, 5, 255, 230, 192, 0, 0, 0, 5, 5, 124, 231, 5, 5, 239, 235, 192, 0, 0, 0, 5, 5, 95, 247, 5, 5, 96, 247, 5, 5, 97, 247, 5, 5, 98, 247, 5, 5, 99, 247, 5, 5, 240, 235, 192, 0, 0, 0, 5, 5, 34, 236, 192, 0, 0, 0, 5, 5, 101, 247, 5, 5, 100, 247, 5, 5, 247, 236, 192, 0, 0, 0, 5, 5, 140, 234, 5, 5, 102, 247, 5, 5, 104, 247, 5, 5, 103, 247, 5, 5, 105, 247, 5, 5, 106, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 226, 5, 5, 107, 247, 5, 5, 154, 227, 192, 0, 0, 0, 5, 5, 108, 247, 5, 5, 109, 247, 5, 5, 23, 229, 5, 5, 110, 247, 5, 5, 36, 226, 5, 5, 111, 247, 5, 5, 112, 247, 5, 5, 113, 247, 5, 5, 114, 247, 5, 5, 149, 232, 5, 5, 115, 247, 5, 5, 116, 247, 192, 0, 0, 0, 5, 5, 117, 247, 5, 5, 184, 231, 192, 0, 0, 0, 5, 5, 118, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 247, 192, 0, 0, 0, 5, 5, 120, 247, 5, 5, 3, 250, 192, 0, 0, 0, 5, 5, 121, 247, 5, 5, 122, 247, 5, 5, 123, 247, 192, 0, 0, 0, 5, 5, 124, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 247, 5, 5, 137, 227, 192, 0, 0, 0, 5, 5, 126, 247, 5, 5, 127, 247, 5, 5, 128, 247, 5, 5, 129, 247, 5, 5, 130, 247, 5, 5, 131, 247, 5, 5, 132, 247, 5, 5, 134, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 236, 5, 5, 133, 247, 192, 0, 0, 0, 5, 5, 135, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 249, 5, 5, 97, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 250, 5, 5, 44, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 250, 5, 5, 92, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 250, 192, 0, 0, 0, 5, 5, 49, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 230, 192, 0, 0, 0, 5, 5, 39, 250, 5, 5, 40, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 233, 5, 5, 210, 229, 5, 5, 82, 247, 192, 0, 0, 0, 5, 5, 140, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 250, 5, 5, 51, 250, 5, 5, 52, 250, 192, 0, 0, 0, 5, 5, 54, 250, 5, 5, 53, 250, 192, 0, 0, 0, 5, 5, 55, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 250, 192, 0, 0, 0, 5, 5, 57, 250, 192, 0, 0, 0, 5, 5, 56, 250, 5, 5, 59, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 246, 192, 0, 0, 0, 5, 5, 124, 246, 5, 5, 125, 246, 192, 0, 0, 0, 5, 5, 162, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 249, 192, 0, 0, 0, 5, 5, 164, 249, 5, 5, 220, 225, 192, 0, 0, 0, 5, 5, 138, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 238, 192, 0, 0, 0, 5, 5, 247, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 233, 192, 0, 0, 0, 5, 5, 61, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 250, 192, 0, 0, 0, 5, 5, 64, 250, 192, 0, 0, 0, 5, 5, 37, 224, 192, 0, 0, 0, 5, 5, 66, 250, 5, 5, 67, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 232, 5, 5, 230, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 224, 5, 5, 153, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 249, 5, 5, 46, 230, 5, 5, 155, 249, 5, 5, 156, 249, 5, 5, 157, 249, 5, 5, 158, 249, 5, 5, 159, 249, 5, 5, 160, 249, 5, 5, 131, 232, 5, 5, 161, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 230, 5, 5, 250, 226, 5, 5, 122, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 227, 5, 5, 11, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 238, 5, 5, 105, 238, 5, 5, 106, 238, 5, 5, 107, 238, 5, 5, 108, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 227, 5, 5, 40, 225, 5, 5, 18, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 232, 5, 5, 190, 228, 5, 5, 152, 229, 5, 5, 52, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 230, 5, 5, 139, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 230, 5, 5, 189, 229, 5, 5, 77, 244, 5, 5, 140, 230, 5, 5, 188, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 229, 5, 5, 180, 229, 5, 5, 181, 229, 5, 5, 177, 229, 5, 5, 179, 229, 192, 0, 0, 0, 5, 5, 206, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 230, 5, 5, 94, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 229, 5, 5, 196, 229, 192, 0, 0, 0, 5, 5, 193, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 230, 5, 5, 197, 229, 192, 0, 0, 0, 5, 5, 82, 230, 5, 5, 191, 229, 5, 5, 201, 229, 192, 0, 0, 0, 5, 5, 50, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 230, 5, 5, 52, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 225, 192, 0, 0, 0, 5, 5, 113, 231, 5, 5, 114, 230, 192, 0, 0, 0, 5, 5, 15, 224, 192, 0, 0, 0, 5, 5, 65, 224, 192, 0, 0, 0, 5, 5, 119, 224, 5, 5, 224, 230, 192, 0, 0, 0, 5, 5, 243, 233, 192, 0, 0, 0, 5, 5, 219, 232, 5, 5, 63, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 224, 5, 5, 50, 233, 5, 5, 78, 233, 5, 5, 212, 232, 5, 5, 123, 230, 5, 5, 124, 230, 5, 5, 11, 230, 192, 0, 0, 0, 5, 5, 3, 230, 5, 5, 4, 230, 192, 0, 0, 0, 5, 5, 6, 230, 192, 0, 0, 0, 5, 5, 9, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 231, 192, 0, 0, 0, 5, 5, 107, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 229, 5, 5, 238, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 230, 5, 5, 29, 230, 5, 5, 254, 235, 5, 5, 28, 230, 5, 5, 27, 230, 192, 0, 0, 0, 5, 5, 246, 229, 5, 5, 84, 231, 5, 5, 83, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 230, 5, 5, 180, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 229, 5, 5, 43, 230, 192, 0, 0, 0, 5, 5, 49, 230, 5, 5, 4, 248, 192, 0, 0, 0, 5, 5, 45, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 229, 192, 0, 0, 0, 5, 5, 63, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 230, 5, 5, 15, 230, 5, 5, 88, 243, 5, 5, 88, 231, 5, 5, 25, 230, 192, 0, 0, 0, 5, 5, 17, 230, 192, 0, 0, 0, 5, 5, 168, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 230, 5, 5, 65, 230, 5, 5, 57, 230, 5, 5, 58, 230, 5, 5, 62, 230, 5, 5, 60, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 230, 5, 5, 102, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 230, 192, 0, 0, 0, 5, 5, 223, 229, 5, 5, 114, 230, 5, 5, 73, 230, 5, 5, 229, 229, 5, 5, 222, 229, 5, 5, 108, 230, 5, 5, 107, 236, 5, 5, 216, 229, 5, 5, 208, 229, 5, 5, 70, 231, 5, 5, 215, 229, 5, 5, 233, 229, 5, 5, 174, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 229, 192, 0, 0, 0, 5, 5, 74, 231, 5, 5, 77, 231, 5, 5, 41, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 250, 5, 5, 32, 230, 5, 5, 38, 230, 192, 0, 0, 0, 5, 5, 234, 229, 5, 5, 156, 248, 5, 5, 235, 229, 192, 0, 0, 0, 5, 5, 217, 237, 192, 0, 0, 0, 5, 5, 80, 233, 5, 5, 164, 224, 5, 5, 70, 225, 5, 5, 82, 232, 5, 5, 254, 225, 5, 5, 146, 234, 5, 5, 35, 234, 5, 5, 92, 237, 5, 5, 234, 225, 5, 5, 7, 224, 192, 0, 0, 0, 5, 5, 166, 235, 5, 5, 131, 228, 192, 0, 0, 0, 5, 5, 148, 229, 5, 5, 111, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 237, 5, 5, 122, 236, 5, 5, 219, 229, 5, 5, 49, 233, 5, 5, 97, 235, 5, 5, 151, 226, 5, 5, 228, 228, 5, 5, 215, 228, 5, 5, 226, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 236, 5, 5, 242, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 235, 5, 5, 236, 232, 5, 5, 232, 230, 5, 5, 233, 230, 5, 5, 92, 232, 5, 5, 14, 224, 5, 5, 122, 227, 192, 0, 0, 0, 5, 5, 27, 232, 192, 0, 0, 0, 5, 5, 17, 231, 192, 0, 0, 0, 5, 5, 148, 243, 5, 5, 234, 227, 5, 5, 58, 229, 5, 5, 73, 237, 192, 0, 0, 0, 5, 5, 250, 230, 5, 5, 56, 228, 5, 5, 140, 233, 5, 5, 189, 230, 5, 5, 85, 227, 5, 5, 250, 242, 192, 0, 0, 0, 5, 5, 3, 238, 192, 0, 0, 0, 5, 5, 47, 238, 5, 5, 12, 224, 5, 5, 39, 233, 5, 5, 33, 246, 5, 5, 17, 232, 192, 0, 0, 0, 5, 5, 84, 238, 5, 5, 17, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 226, 5, 5, 142, 233, 5, 5, 143, 237, 5, 5, 16, 225, 5, 5, 249, 239, 5, 5, 249, 239, 5, 5, 8, 238, 5, 5, 136, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 243, 5, 5, 113, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 229, 5, 5, 139, 232, 192, 0, 0, 0, 5, 5, 254, 224, 5, 5, 45, 228, 5, 5, 180, 236, 5, 5, 23, 233, 5, 5, 122, 227, 5, 5, 95, 241, 5, 5, 123, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 236, 5, 5, 26, 224, 5, 5, 184, 243, 192, 0, 0, 0, 5, 5, 90, 242, 192, 0, 0, 0, 5, 5, 134, 224, 5, 5, 8, 242, 5, 5, 141, 242, 5, 5, 55, 233, 5, 5, 236, 236, 5, 5, 73, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 225, 5, 5, 241, 240, 5, 5, 205, 233, 5, 5, 244, 240, 5, 5, 0, 155, 5, 5, 105, 232, 5, 5, 180, 229, 5, 5, 181, 234, 5, 5, 120, 237, 5, 5, 111, 225, 192, 0, 0, 0, 5, 5, 65, 230, 192, 0, 0, 0, 5, 5, 58, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 238, 5, 5, 72, 232, 5, 5, 34, 229, 5, 5, 78, 226, 5, 5, 251, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 227, 192, 0, 0, 0, 5, 5, 217, 234, 5, 5, 122, 236, 5, 5, 64, 233, 5, 5, 191, 237, 192, 0, 0, 0, 5, 5, 51, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 237, 5, 5, 212, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 247, 5, 5, 166, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 229, 5, 5, 162, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 231, 5, 5, 4, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 232, 5, 5, 105, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 230, 5, 5, 125, 234, 192, 0, 0, 0, 5, 5, 13, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 237, 192, 0, 0, 0, 5, 5, 191, 232, 5, 5, 249, 236, 192, 0, 0, 0, 5, 5, 226, 225, 5, 5, 130, 229, 192, 0, 0, 0, 5, 5, 177, 239, 5, 5, 176, 232, 192, 0, 0, 0, 5, 5, 88, 229, 192, 0, 0, 0, 5, 5, 226, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 236, 5, 5, 233, 230, 5, 5, 92, 232, 5, 5, 23, 233, 5, 5, 0, 218, 5, 5, 76, 225, 5, 5, 15, 224, 5, 5, 236, 227, 5, 5, 14, 224, 5, 5, 101, 224, 5, 5, 36, 228, 192, 0, 0, 0, 5, 5, 102, 232, 5, 5, 102, 232, 5, 5, 102, 232, 192, 0, 0, 0, 5, 5, 226, 227, 5, 5, 32, 228, 5, 5, 173, 239, 192, 0, 0, 0, 5, 5, 158, 228, 5, 5, 23, 241, 5, 5, 30, 241, 192, 0, 0, 0, 5, 5, 22, 235, 5, 5, 228, 224, 5, 5, 234, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 233, 5, 5, 6, 233, 5, 5, 123, 241, 192, 0, 0, 0, 5, 5, 77, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 232, 5, 5, 33, 238, 5, 5, 225, 224, 5, 5, 226, 239, 5, 5, 232, 239, 5, 5, 163, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 226, 192, 0, 0, 0, 5, 5, 29, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 233, 5, 5, 121, 228, 5, 5, 137, 225, 5, 5, 208, 240, 192, 0, 0, 0, 5, 5, 122, 234, 5, 5, 148, 243, 192, 0, 0, 0, 5, 5, 201, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 241, 192, 0, 0, 0, 5, 5, 230, 235, 5, 5, 204, 224, 192, 0, 0, 0, 5, 5, 15, 239, 192, 0, 0, 0, 5, 5, 183, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 242, 5, 5, 154, 233, 5, 5, 178, 229, 192, 0, 0, 0, 5, 5, 199, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 232, 5, 5, 204, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 225, 5, 5, 213, 227, 5, 5, 55, 233, 5, 5, 213, 227, 192, 0, 0, 0, 5, 5, 73, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 224, 5, 5, 31, 245, 192, 0, 0, 0, 5, 5, 5, 224, 5, 5, 0, 172, 5, 5, 22, 229, 192, 0, 0, 0, 5, 5, 166, 234, 5, 5, 214, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 236, 192, 0, 0, 0, 5, 5, 12, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 230, 5, 5, 224, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 230, 5, 5, 231, 230, 5, 5, 92, 238, 192, 0, 0, 0, 5, 5, 155, 245, 192, 0, 0, 0, 5, 5, 180, 229, 5, 5, 181, 234, 192, 0, 0, 0, 5, 5, 117, 244, 5, 5, 116, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 231, 5, 5, 85, 227, 5, 5, 65, 230, 5, 5, 120, 227, 5, 5, 204, 228, 5, 5, 95, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 226, 5, 5, 12, 243, 192, 0, 0, 0, 5, 5, 58, 238, 5, 5, 185, 225, 192, 0, 0, 0, 5, 5, 3, 236, 5, 5, 203, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 243, 192, 0, 0, 0, 5, 5, 73, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 234, 192, 0, 0, 0, 5, 5, 5, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 246, 5, 5, 34, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 237, 5, 5, 150, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 246, 5, 5, 93, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 226, 5, 5, 72, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 231, 5, 5, 200, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 248, 5, 5, 210, 248, 5, 5, 65, 225, 192, 0, 0, 0, 5, 5, 3, 240, 5, 5, 229, 236, 5, 5, 179, 237, 5, 5, 183, 229, 5, 5, 184, 227, 5, 5, 82, 226, 5, 5, 243, 235, 5, 5, 108, 229, 192, 0, 0, 0, 5, 5, 212, 232, 192, 0, 0, 0, 5, 5, 185, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 230, 192, 0, 0, 0, 5, 5, 8, 238, 192, 0, 0, 0, 5, 5, 3, 229, 5, 5, 40, 229, 5, 5, 135, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 173, 5, 5, 74, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 230, 192, 0, 0, 0, 5, 5, 123, 246, 5, 5, 162, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 134, 2, 4, 0, 0, 0, 42, 0, 134, 3, 4, 0, 0, 0, 50, 0, 134, 2, 4, 0, 0, 0, 50, 0, 134, 3, 4, 0, 0, 0, 58, 0, 134, 2, 4, 0, 0, 0, 58, 0, 134, 3, 4, 0, 0, 0, 66, 0, 134, 2, 4, 0, 0, 0, 66, 0, 134, 3, 4, 0, 0, 0, 68, 0, 134, 2, 4, 0, 0, 0, 68, 0, 134, 3, 4, 0, 0, 0, 70, 0, 134, 2, 4, 0, 0, 0, 70, 0, 134, 3, 4, 0, 0, 0, 82, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 2, 4, 0, 0, 0, 82, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 0, 0, 0, 134, 3, 0, 42, 5, 162, 0, 0, 134, 2, 0, 42, 5, 162, 0, 0, 5, 3, 0, 42, 5, 162, 0, 0, 5, 2, 0, 42, 5, 162, 0, 0, 134, 3, 0, 50, 5, 162, 0, 0, 134, 2, 0, 50, 5, 162, 0, 0, 5, 3, 0, 50, 5, 162, 0, 0, 5, 2, 0, 50, 5, 162, 0, 0, 134, 3, 0, 58, 5, 162, 0, 0, 134, 2, 0, 58, 5, 162, 0, 0, 5, 3, 0, 58, 5, 162, 0, 0, 5, 2, 0, 58, 5, 162, 0, 0, 134, 3, 0, 68, 5, 160, 0, 0, 134, 2, 0, 68, 5, 160, 0, 0, 5, 3, 0, 68, 5, 160, 0, 0, 5, 2, 0, 68, 5, 160, 0, 0, 134, 3, 0, 82, 5, 162, 0, 0, 134, 2, 0, 82, 5, 162, 0, 0, 134, 74, 0, 82, 5, 162, 0, 0, 5, 3, 0, 82, 5, 162, 0, 0, 5, 2, 0, 82, 5, 162, 0, 0, 5, 74, 0, 82, 5, 162, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 134, 3, 0, 70, 5, 162, 0, 0, 134, 2, 0, 70, 5, 162, 0, 0, 5, 3, 0, 70, 5, 162, 0, 0, 5, 2, 0, 70, 5, 162, 0, 0, 134, 3, 0, 50, 5, 160, 0, 0, 134, 2, 0, 50, 5, 160, 0, 0, 5, 3, 0, 50, 5, 160, 0, 0, 5, 2, 0, 50, 5, 160, 0, 0, 134, 3, 0, 42, 5, 200, 0, 0, 134, 2, 0, 42, 5, 200, 0, 0, 5, 3, 0, 42, 5, 200, 0, 0, 5, 2, 0, 42, 5, 200, 0, 0, 134, 2, 0, 50, 5, 138, 0, 0, 5, 2, 0, 50, 5, 138, 0, 0, 134, 2, 0, 50, 5, 136, 0, 0, 5, 2, 0, 50, 5, 136, 0, 0, 134, 3, 0, 50, 5, 204, 0, 0, 134, 2, 0, 50, 5, 204, 0, 0, 5, 3, 0, 50, 5, 204, 0, 0, 5, 2, 0, 50, 5, 204, 0, 0, 134, 3, 0, 50, 5, 208, 0, 0, 134, 2, 0, 50, 5, 208, 0, 0, 5, 3, 0, 50, 5, 208, 0, 0, 5, 2, 0, 50, 5, 208, 0, 0, 134, 3, 0, 58, 5, 208, 0, 0, 134, 2, 0, 58, 5, 208, 0, 0, 5, 3, 0, 58, 5, 208, 0, 0, 5, 2, 0, 58, 5, 208, 0, 0, 134, 3, 0, 66, 5, 196, 0, 0, 134, 2, 0, 66, 5, 196, 0, 0, 5, 3, 0, 66, 5, 196, 0, 0, 5, 2, 0, 66, 5, 196, 0, 0, 134, 3, 0, 68, 5, 196, 0, 0, 134, 2, 0, 68, 5, 196, 0, 0, 5, 3, 0, 68, 5, 196, 0, 0, 5, 2, 0, 68, 5, 196, 0, 0, 134, 3, 0, 68, 5, 210, 0, 0, 134, 2, 0, 68, 5, 210, 0, 0, 5, 3, 0, 68, 5, 210, 0, 0, 5, 2, 0, 68, 5, 210, 0, 0, 134, 3, 0, 68, 5, 204, 0, 0, 134, 2, 0, 68, 5, 204, 0, 0, 5, 3, 0, 68, 5, 204, 0, 0, 5, 2, 0, 68, 5, 204, 0, 0, 134, 2, 0, 70, 5, 138, 0, 0, 5, 2, 0, 70, 5, 138, 0, 0, 134, 2, 0, 70, 5, 136, 0, 0, 5, 2, 0, 70, 5, 136, 0, 0, 134, 3, 0, 82, 5, 198, 0, 0, 134, 2, 0, 82, 5, 198, 0, 0, 134, 74, 0, 82, 5, 198, 0, 0, 5, 3, 0, 82, 5, 198, 0, 0, 5, 2, 0, 82, 5, 198, 0, 0, 5, 74, 0, 82, 5, 198, 0, 0, 134, 3, 0, 82, 5, 208, 0, 0, 134, 2, 0, 82, 5, 208, 0, 0, 134, 74, 0, 82, 5, 208, 0, 0, 5, 3, 0, 82, 5, 208, 0, 0, 5, 2, 0, 82, 5, 208, 0, 0, 5, 74, 0, 82, 5, 208, 0, 0, 134, 3, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 204, 0, 0, 134, 74, 0, 82, 5, 204, 0, 0, 5, 3, 0, 82, 5, 204, 0, 0, 5, 2, 0, 82, 5, 204, 0, 0, 5, 74, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 150, 0, 0, 5, 2, 0, 82, 5, 150, 0, 0, 134, 3, 0, 42, 5, 196, 0, 0, 134, 2, 0, 42, 5, 196, 0, 0, 5, 3, 0, 42, 5, 196, 0, 0, 5, 2, 0, 42, 5, 196, 0, 0, 134, 3, 0, 50, 5, 196, 0, 0, 134, 2, 0, 50, 5, 196, 0, 0, 5, 3, 0, 50, 5, 196, 0, 0, 5, 2, 0, 50, 5, 196, 0, 0, 134, 3, 0, 58, 5, 196, 0, 0, 134, 2, 0, 58, 5, 196, 0, 0, 5, 3, 0, 58, 5, 196, 0, 0, 5, 2, 0, 58, 5, 196, 0, 0, 134, 3, 0, 70, 5, 196, 0, 0, 134, 2, 0, 70, 5, 196, 0, 0, 5, 3, 0, 70, 5, 196, 0, 0, 5, 2, 0, 70, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 196, 0, 0, 134, 2, 0, 82, 5, 196, 0, 0, 134, 74, 0, 82, 5, 196, 0, 0, 5, 3, 0, 82, 5, 196, 0, 0, 5, 2, 0, 82, 5, 196, 0, 0, 5, 74, 0, 82, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 5, 0, 12, 3, 7, 0, 12, 3, 198, 161, 64, 3, 198, 129, 65, 3, 0, 234, 134, 3, 0, 3, 198, 129, 1, 3, 0, 234, 134, 3, 4, 3, 0, 234, 134, 2, 0, 50, 156, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 134, 2, 12, 3, 0, 192, 198, 97, 64, 3, 0, 192, 198, 65, 65, 3, 0, 242, 134, 3, 0, 3, 0, 192, 198, 65, 1, 3, 0, 242, 134, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 58, 156, 5, 5, 0, 12, 3, 9, 0, 12, 3, 0, 192, 198, 161, 64, 3, 0, 192, 198, 129, 65, 3, 0, 250, 134, 3, 0, 3, 0, 192, 198, 129, 1, 3, 0, 250, 134, 3, 4, 3, 0, 250, 134, 2, 0, 66, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 0, 192, 198, 225, 64, 3, 0, 192, 198, 193, 65, 3, 255, 255, 0, 66, 134, 3, 0, 3, 0, 192, 198, 193, 1, 3, 255, 255, 0, 66, 134, 3, 4, 3, 255, 255, 0, 66, 134, 2, 0, 68, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 33, 64, 3, 1, 192, 198, 1, 65, 3, 255, 255, 0, 68, 134, 3, 0, 3, 1, 192, 198, 1, 1, 3, 255, 255, 0, 68, 134, 3, 4, 3, 255, 255, 0, 68, 134, 2, 0, 70, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 97, 64, 3, 1, 192, 198, 65, 65, 3, 255, 255, 0, 70, 134, 3, 0, 3, 1, 192, 198, 65, 1, 3, 255, 255, 0, 70, 134, 3, 4, 3, 255, 255, 0, 70, 134, 2, 0, 82, 156, 5, 7, 0, 12, 3, 14, 0, 12, 3, 1, 192, 198, 161, 64, 3, 1, 192, 198, 129, 65, 3, 255, 255, 0, 82, 134, 3, 68, 3, 255, 255, 0, 82, 134, 71, 0, 3, 1, 192, 198, 129, 1, 3, 255, 255, 0, 82, 134, 3, 4, 3, 255, 255, 0, 82, 134, 2, 8, 3, 197, 127, 0, 82, 134, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 0, 42, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 1, 192, 198, 225, 64, 3, 1, 192, 198, 193, 65, 3, 0, 234, 5, 3, 0, 3, 1, 192, 198, 193, 1, 3, 0, 234, 5, 3, 4, 3, 0, 234, 5, 2, 0, 50, 5, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 5, 2, 12, 3, 2, 192, 198, 33, 64, 3, 2, 192, 198, 1, 65, 3, 0, 242, 5, 3, 0, 3, 2, 192, 198, 1, 1, 3, 0, 242, 5, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 58, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 2, 192, 198, 97, 64, 3, 2, 192, 198, 65, 65, 3, 0, 250, 5, 3, 0, 3, 2, 192, 198, 65, 1, 3, 0, 250, 5, 3, 4, 3, 0, 250, 5, 2, 0, 66, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 161, 64, 3, 2, 192, 198, 129, 65, 3, 255, 255, 0, 66, 5, 3, 0, 3, 2, 192, 198, 129, 1, 3, 255, 255, 0, 66, 5, 3, 4, 3, 255, 255, 0, 66, 5, 2, 0, 68, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 225, 64, 3, 2, 192, 198, 193, 65, 3, 255, 255, 0, 68, 5, 3, 0, 3, 2, 192, 198, 193, 1, 3, 255, 255, 0, 68, 5, 3, 4, 3, 255, 255, 0, 68, 5, 2, 0, 70, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 3, 192, 198, 33, 64, 3, 3, 192, 198, 1, 65, 3, 255, 255, 0, 70, 5, 3, 0, 3, 3, 192, 198, 1, 1, 3, 255, 255, 0, 70, 5, 3, 4, 3, 255, 255, 0, 70, 5, 2, 0, 82, 5, 5, 7, 0, 12, 3, 14, 0, 12, 3, 3, 192, 198, 97, 64, 3, 3, 192, 198, 65, 65, 3, 255, 255, 0, 82, 5, 3, 68, 3, 255, 255, 0, 82, 5, 71, 0, 3, 3, 192, 198, 65, 1, 3, 255, 255, 0, 82, 5, 3, 4, 3, 255, 255, 0, 82, 5, 2, 8, 3, 197, 127, 0, 82, 5, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 82, 134, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 5, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 82, 5, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 42, 196, 162, 5, 0, 12, 3, 8, 0, 12, 3, 3, 192, 198, 194, 64, 3, 3, 192, 198, 130, 65, 3, 197, 162, 0, 3, 3, 192, 198, 130, 1, 3, 197, 162, 4, 3, 0, 192, 197, 98, 5, 42, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 66, 64, 3, 4, 192, 198, 2, 65, 3, 0, 192, 197, 162, 0, 3, 4, 192, 198, 2, 1, 3, 0, 192, 197, 162, 4, 3, 0, 192, 197, 226, 156, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 194, 64, 3, 4, 192, 198, 130, 65, 3, 1, 192, 197, 34, 0, 3, 4, 192, 198, 130, 1, 3, 1, 192, 197, 34, 4, 3, 1, 192, 197, 98, 5, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 66, 64, 3, 5, 192, 198, 2, 65, 3, 1, 192, 197, 162, 0, 3, 5, 192, 198, 2, 1, 3, 1, 192, 197, 162, 4, 3, 1, 192, 197, 226, 156, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 194, 64, 3, 5, 192, 198, 130, 65, 3, 2, 192, 197, 34, 0, 3, 5, 192, 198, 130, 1, 3, 2, 192, 197, 34, 4, 3, 2, 192, 197, 98, 5, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 66, 64, 3, 6, 192, 198, 2, 65, 3, 2, 192, 197, 162, 0, 3, 6, 192, 198, 2, 1, 3, 2, 192, 197, 162, 4, 3, 2, 192, 197, 226, 156, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 194, 64, 3, 6, 192, 198, 130, 65, 3, 3, 192, 197, 34, 0, 3, 6, 192, 198, 130, 1, 3, 3, 192, 197, 34, 4, 3, 3, 192, 197, 98, 5, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 7, 192, 198, 66, 64, 3, 7, 192, 198, 2, 65, 3, 3, 192, 197, 162, 0, 3, 7, 192, 198, 2, 1, 3, 3, 192, 197, 162, 4, 3, 3, 192, 197, 226, 156, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 4, 192, 197, 162, 12, 3, 7, 192, 198, 194, 64, 3, 7, 192, 198, 130, 65, 3, 4, 192, 197, 34, 0, 3, 7, 192, 198, 130, 1, 3, 4, 192, 197, 34, 4, 3, 4, 192, 197, 98, 5, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 5, 192, 197, 98, 12, 3, 8, 192, 198, 66, 64, 3, 8, 192, 198, 2, 65, 3, 4, 192, 197, 226, 0, 3, 8, 192, 198, 2, 1, 3, 4, 192, 197, 226, 4, 3, 5, 192, 197, 34, 156, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 8, 192, 198, 194, 64, 3, 8, 192, 198, 130, 65, 3, 5, 192, 197, 162, 0, 3, 8, 192, 198, 130, 1, 3, 5, 192, 197, 162, 4, 3, 5, 192, 197, 226, 5, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 9, 192, 198, 66, 64, 3, 9, 192, 198, 2, 65, 3, 6, 192, 197, 34, 0, 3, 9, 192, 198, 2, 1, 3, 6, 192, 197, 34, 4, 3, 6, 192, 197, 98, 156, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 34, 12, 3, 9, 192, 198, 194, 64, 3, 9, 192, 198, 130, 65, 3, 6, 192, 197, 162, 0, 3, 9, 192, 198, 130, 1, 3, 6, 192, 197, 162, 4, 3, 6, 192, 197, 226, 5, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 226, 12, 3, 10, 192, 198, 66, 64, 3, 10, 192, 198, 2, 65, 3, 7, 192, 197, 98, 0, 3, 10, 192, 198, 2, 1, 3, 7, 192, 197, 98, 4, 3, 7, 192, 197, 162, 156, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 10, 192, 198, 194, 64, 3, 10, 192, 198, 130, 65, 3, 8, 192, 197, 34, 0, 3, 10, 192, 198, 130, 1, 3, 8, 192, 197, 34, 4, 3, 8, 192, 197, 98, 5, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 66, 64, 3, 11, 192, 198, 2, 65, 3, 8, 192, 197, 162, 0, 3, 11, 192, 198, 2, 1, 3, 8, 192, 197, 162, 4, 3, 8, 192, 197, 226, 156, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 194, 64, 3, 11, 192, 198, 130, 65, 3, 9, 192, 197, 34, 0, 3, 11, 192, 198, 130, 1, 3, 9, 192, 197, 34, 4, 3, 9, 192, 197, 98, 5, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 66, 64, 3, 12, 192, 198, 2, 65, 3, 9, 192, 197, 162, 0, 3, 12, 192, 198, 2, 1, 3, 9, 192, 197, 162, 4, 3, 9, 192, 197, 226, 156, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 194, 64, 3, 12, 192, 198, 130, 65, 3, 10, 192, 197, 34, 0, 3, 12, 192, 198, 130, 1, 3, 10, 192, 197, 34, 4, 3, 10, 192, 197, 98, 5, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 66, 64, 3, 13, 192, 198, 2, 65, 3, 10, 192, 197, 162, 0, 3, 13, 192, 198, 2, 1, 3, 10, 192, 197, 162, 4, 3, 10, 192, 197, 226, 156, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 194, 64, 3, 13, 192, 198, 130, 65, 3, 12, 192, 197, 34, 0, 3, 13, 192, 198, 130, 1, 3, 12, 192, 197, 34, 4, 3, 12, 192, 197, 98, 5, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 66, 64, 3, 14, 192, 198, 2, 65, 3, 12, 192, 197, 162, 0, 3, 14, 192, 198, 2, 1, 3, 12, 192, 197, 162, 4, 3, 12, 192, 197, 226, 156, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 194, 64, 3, 14, 192, 198, 130, 65, 3, 13, 192, 197, 34, 0, 3, 14, 192, 198, 130, 1, 3, 13, 192, 197, 34, 4, 3, 13, 192, 197, 98, 5, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 66, 64, 3, 15, 192, 198, 2, 65, 3, 13, 192, 197, 162, 0, 3, 15, 192, 198, 2, 1, 3, 13, 192, 197, 162, 4, 3, 13, 192, 197, 226, 156, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 194, 64, 3, 15, 192, 198, 130, 65, 3, 14, 192, 197, 34, 0, 3, 15, 192, 198, 130, 1, 3, 14, 192, 197, 34, 4, 3, 14, 192, 197, 98, 5, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 66, 64, 3, 16, 192, 198, 2, 65, 3, 14, 192, 197, 162, 0, 3, 16, 192, 198, 2, 1, 3, 14, 192, 197, 162, 4, 3, 14, 192, 197, 226, 156, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 194, 64, 3, 16, 192, 198, 130, 65, 3, 15, 192, 197, 34, 0, 3, 16, 192, 198, 130, 1, 3, 15, 192, 197, 34, 4, 3, 15, 192, 197, 98, 5, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 66, 64, 3, 17, 192, 198, 2, 65, 3, 15, 192, 197, 162, 0, 3, 17, 192, 198, 2, 1, 3, 15, 192, 197, 162, 4, 3, 15, 192, 197, 226, 156, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 194, 64, 3, 17, 192, 198, 130, 65, 3, 16, 192, 197, 34, 0, 3, 17, 192, 198, 130, 1, 3, 16, 192, 197, 34, 4, 3, 16, 192, 197, 98, 5, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 66, 64, 3, 18, 192, 198, 2, 65, 3, 16, 192, 197, 162, 0, 3, 18, 192, 198, 2, 1, 3, 16, 192, 197, 162, 4, 3, 16, 192, 197, 226, 156, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 194, 64, 3, 18, 192, 198, 130, 65, 3, 17, 192, 197, 34, 0, 3, 18, 192, 198, 130, 1, 3, 17, 192, 197, 34, 4, 3, 17, 192, 197, 98, 5, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 66, 64, 3, 19, 192, 198, 2, 65, 3, 17, 192, 197, 162, 0, 3, 19, 192, 198, 2, 1, 3, 17, 192, 197, 162, 4, 3, 17, 192, 197, 226, 156, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 194, 64, 3, 19, 192, 198, 130, 65, 3, 18, 192, 197, 34, 0, 3, 19, 192, 198, 130, 1, 3, 18, 192, 197, 34, 4, 3, 18, 192, 197, 98, 5, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 20, 192, 198, 66, 64, 3, 20, 192, 198, 2, 65, 3, 18, 192, 197, 162, 0, 3, 20, 192, 198, 2, 1, 3, 18, 192, 197, 162, 4, 3, 18, 192, 197, 226, 156, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 20, 192, 197, 162, 12, 3, 20, 192, 198, 194, 64, 3, 20, 192, 198, 130, 65, 3, 20, 192, 197, 34, 0, 3, 20, 192, 198, 130, 1, 3, 20, 192, 197, 34, 4, 3, 20, 192, 197, 98, 5, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 21, 192, 197, 98, 12, 3, 21, 192, 198, 66, 64, 3, 21, 192, 198, 2, 65, 3, 20, 192, 197, 226, 0, 3, 21, 192, 198, 2, 1, 3, 20, 192, 197, 226, 4, 3, 21, 192, 197, 34, 156, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 34, 12, 3, 21, 192, 198, 194, 64, 3, 21, 192, 198, 130, 65, 3, 21, 192, 197, 162, 0, 3, 21, 192, 198, 130, 1, 3, 21, 192, 197, 162, 4, 3, 21, 192, 197, 226, 5, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 226, 12, 3, 22, 192, 198, 66, 64, 3, 22, 192, 198, 2, 65, 3, 22, 192, 197, 98, 0, 3, 22, 192, 198, 2, 1, 3, 22, 192, 197, 98, 4, 3, 22, 192, 197, 162, 156, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 23, 192, 197, 162, 12, 3, 22, 192, 198, 194, 64, 3, 22, 192, 198, 130, 65, 3, 23, 192, 197, 34, 0, 3, 22, 192, 198, 130, 1, 3, 23, 192, 197, 34, 4, 3, 23, 192, 197, 98, 5, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 24, 192, 197, 98, 12, 3, 23, 192, 198, 66, 64, 3, 23, 192, 198, 2, 65, 3, 23, 192, 197, 226, 0, 3, 23, 192, 198, 2, 1, 3, 23, 192, 197, 226, 4, 3, 24, 192, 197, 34, 156, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 23, 192, 198, 194, 64, 3, 23, 192, 198, 130, 65, 3, 25, 192, 197, 34, 0, 3, 23, 192, 198, 130, 1, 3, 25, 192, 197, 34, 4, 3, 25, 192, 197, 98, 5, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 66, 64, 3, 24, 192, 198, 2, 65, 3, 25, 192, 197, 162, 0, 3, 24, 192, 198, 2, 1, 3, 25, 192, 197, 162, 4, 3, 25, 192, 197, 226, 156, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 194, 64, 3, 24, 192, 198, 130, 65, 3, 26, 192, 197, 34, 0, 3, 24, 192, 198, 130, 1, 3, 26, 192, 197, 34, 4, 3, 26, 192, 197, 98, 5, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 66, 64, 3, 25, 192, 198, 2, 65, 3, 26, 192, 197, 162, 0, 3, 25, 192, 198, 2, 1, 3, 26, 192, 197, 162, 4, 3, 26, 192, 197, 226, 156, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 194, 64, 3, 25, 192, 198, 130, 65, 3, 27, 192, 197, 34, 0, 3, 25, 192, 198, 130, 1, 3, 27, 192, 197, 34, 4, 3, 27, 192, 197, 98, 5, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 66, 64, 3, 26, 192, 198, 2, 65, 3, 27, 192, 197, 162, 0, 3, 26, 192, 198, 2, 1, 3, 27, 192, 197, 162, 4, 3, 27, 192, 197, 226, 156, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 194, 64, 3, 26, 192, 198, 130, 65, 3, 28, 192, 197, 34, 0, 3, 26, 192, 198, 130, 1, 3, 28, 192, 197, 34, 4, 3, 28, 192, 197, 98, 5, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 66, 64, 3, 27, 192, 198, 2, 65, 3, 28, 192, 197, 162, 0, 3, 27, 192, 198, 2, 1, 3, 28, 192, 197, 162, 4, 3, 28, 192, 197, 226, 29, 0, 197, 35, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 227, 64, 3, 27, 192, 198, 131, 65, 3, 29, 192, 197, 131, 0, 3, 27, 192, 198, 131, 1, 3, 29, 192, 197, 131, 4, 3, 29, 192, 197, 227, 30, 0, 197, 67, 5, 0, 12, 3, 9, 0, 12, 3, 28, 192, 198, 163, 64, 3, 28, 192, 198, 67, 65, 3, 30, 192, 197, 163, 0, 3, 28, 192, 198, 67, 1, 3, 30, 192, 197, 163, 4, 3, 31, 192, 197, 3, 156, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 31, 192, 197, 226, 12, 3, 29, 192, 198, 66, 64, 3, 29, 192, 198, 2, 65, 3, 31, 192, 197, 98, 0, 3, 29, 192, 198, 2, 1, 3, 31, 192, 197, 98, 4, 3, 31, 192, 197, 162, 5, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 32, 192, 197, 162, 12, 3, 29, 192, 198, 194, 64, 3, 29, 192, 198, 130, 65, 3, 32, 192, 197, 34, 0, 3, 29, 192, 198, 130, 1, 3, 32, 192, 197, 34, 4, 3, 32, 192, 197, 98, 32, 0, 197, 227, 6, 0, 8, 3, 12, 0, 8, 3, 34, 192, 197, 3, 12, 3, 30, 192, 198, 99, 64, 3, 30, 192, 198, 3, 65, 3, 33, 192, 197, 67, 0, 3, 30, 192, 198, 3, 1, 3, 33, 192, 197, 67, 4, 3, 33, 192, 197, 163, 34, 0, 197, 99, 6, 0, 8, 3, 12, 0, 8, 3, 35, 192, 197, 131, 12, 3, 31, 192, 198, 35, 64, 3, 30, 192, 198, 195, 65, 3, 34, 192, 197, 195, 0, 3, 30, 192, 198, 195, 1, 3, 34, 192, 197, 195, 4, 3, 35, 192, 197, 35, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 28, 4, 185, 76, 185, 80, 185, 84, 30, 4, 186, 96, 185, 100, 185, 104, 32, 4, 34, 4, 36, 4, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 38, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 40, 4, 168, 60, 168, 64, 168, 68, 42, 4, 168, 76, 168, 80, 168, 84, 44, 4, 168, 96, 168, 100, 168, 104, 46, 4, 48, 4, 50, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 52, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 89, 56, 57, 56, 217, 57, 121, 58, 57, 58, 25, 58, 0, 8, 217, 66, 121, 72, 57, 72, 54, 4, 58, 74, 89, 88, 57, 88, 218, 89, 58, 90, 217, 68, 122, 114, 89, 120, 57, 120, 217, 121, 121, 122, 57, 122, 136, 14, 185, 122, 89, 152, 57, 152, 217, 153, 56, 4, 153, 169, 185, 176, 2, 8, 72, 56, 40, 56, 200, 57, 104, 58, 40, 58, 8, 58, 4, 8, 200, 66, 104, 72, 40, 72, 58, 4, 40, 74, 72, 88, 40, 88, 200, 89, 40, 90, 200, 68, 104, 114, 72, 120, 40, 120, 200, 121, 104, 122, 40, 122, 128, 14, 168, 122, 72, 152, 40, 152, 200, 153, 60, 4, 136, 169, 168, 176, 40, 170, 25, 56, 8, 56, 185, 57, 168, 57, 62, 4, 64, 4, 153, 65, 136, 65, 217, 65, 200, 65, 153, 66, 136, 66, 249, 65, 232, 65, 249, 69, 232, 69, 25, 71, 8, 71, 25, 72, 8, 72, 186, 73, 168, 73, 154, 74, 136, 74, 66, 4, 68, 4, 89, 72, 72, 72, 217, 81, 200, 81, 185, 81, 168, 81, 153, 82, 136, 82, 217, 82, 200, 82, 217, 85, 200, 85, 25, 87, 8, 87, 122, 90, 104, 90, 25, 88, 8, 88, 186, 89, 168, 89, 70, 4, 72, 4, 154, 90, 168, 92, 6, 8, 8, 8, 218, 97, 200, 97, 217, 102, 200, 102, 168, 132, 153, 105, 136, 105, 217, 106, 200, 106, 249, 105, 232, 105, 57, 107, 40, 107, 25, 107, 8, 107, 25, 112, 8, 112, 74, 4, 76, 4, 57, 112, 40, 112, 10, 8, 185, 116, 168, 116, 25, 120, 8, 120, 185, 121, 168, 121, 89, 122, 72, 122, 12, 8, 14, 8, 153, 137, 136, 137, 217, 138, 200, 138, 249, 137, 232, 137, 154, 141, 136, 141, 218, 141, 200, 141, 218, 142, 200, 142, 250, 141, 232, 141, 217, 146, 200, 146, 249, 145, 232, 145, 185, 148, 168, 148, 121, 154, 104, 154, 25, 152, 8, 152, 185, 153, 168, 153, 25, 154, 8, 154, 89, 154, 72, 154, 78, 4, 80, 4, 217, 161, 200, 161, 217, 169, 200, 169, 57, 170, 153, 173, 136, 173, 153, 174, 136, 174, 249, 173, 232, 173, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 72, 255, 5, 186, 88, 255, 5, 185, 108, 255, 5, 186, 112, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 255, 5, 168, 72, 255, 5, 168, 88, 255, 5, 168, 108, 255, 5, 168, 112, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 218, 73, 255, 5, 217, 152, 255, 5, 200, 73, 255, 5, 200, 152, 255, 5, 249, 58, 255, 5, 232, 58, 255, 5, 250, 74, 255, 5, 232, 74, 255, 5, 250, 90, 255, 5, 232, 90, 255, 5, 218, 114, 255, 5, 200, 114, 255, 5, 249, 154, 255, 5, 232, 154, 255, 1, 170, 170, 3, 0, 68, 0, 59, 0, 32, 0, 155, 48, 1, 16, 214, 1, 0, 96, 1, 0, 0, 96, 98, 186, 1, 0, 24, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 72, 0, 0, 0, 72, 1, 0, 0, 248, 132, 1, 0, 248, 132, 1, 0, 104, 143, 1, 0, 104, 143, 1, 0, 40, 158, 1, 0, 40, 158, 1, 0, 44, 179, 1, 0, 26, 182, 1, 0, 74, 186, 1, 0, 17, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 254, 255, 50, 105, 114, 84, 1, 0, 0, 9, 26, 23, 64, 0, 192, 0, 96, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 109, 0, 117, 0, 122, 0, 130, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 169, 0, 176, 0, 184, 0, 189, 0, 48, 0, 197, 0, 205, 0, 213, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 218, 0, 48, 0, 48, 0, 226, 0, 234, 0, 242, 0, 250, 0, 2, 1, 10, 1, 18, 1, 26, 1, 48, 0, 33, 1, 35, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 42, 1, 48, 0, 48, 0, 48, 0, 48, 0, 50, 1, 58, 1, 48, 0, 66, 1, 74, 1, 82, 1, 85, 1, 48, 0, 48, 0, 93, 1, 101, 1, 48, 0, 48, 0, 48, 0, 109, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 117, 1, 125, 1, 133, 1, 141, 1, 149, 1, 157, 1, 165, 1, 173, 1, 181, 1, 189, 1, 197, 1, 205, 1, 213, 1, 221, 1, 229, 1, 237, 1, 245, 1, 253, 1, 5, 2, 13, 2, 21, 2, 29, 2, 37, 2, 45, 2, 53, 2, 61, 2, 69, 2, 77, 2, 85, 2, 93, 2, 101, 2, 109, 2, 117, 2, 125, 2, 133, 2, 141, 2, 149, 2, 157, 2, 165, 2, 173, 2, 181, 2, 189, 2, 197, 2, 205, 2, 213, 2, 221, 2, 229, 2, 237, 2, 245, 2, 253, 2, 5, 3, 13, 3, 21, 3, 29, 3, 37, 3, 45, 3, 53, 3, 61, 3, 69, 3, 77, 3, 85, 3, 93, 3, 101, 3, 109, 3, 117, 3, 125, 3, 133, 3, 141, 3, 149, 3, 157, 3, 165, 3, 173, 3, 181, 3, 189, 3, 197, 3, 205, 3, 213, 3, 221, 3, 229, 3, 237, 3, 245, 3, 253, 3, 5, 4, 13, 4, 21, 4, 29, 4, 37, 4, 45, 4, 53, 4, 61, 4, 69, 4, 77, 4, 85, 4, 93, 4, 101, 4, 109, 4, 117, 4, 125, 4, 133, 4, 141, 4, 149, 4, 157, 4, 165, 4, 173, 4, 181, 4, 189, 4, 197, 4, 205, 4, 213, 4, 221, 4, 229, 4, 237, 4, 245, 4, 253, 4, 5, 5, 13, 5, 21, 5, 29, 5, 37, 5, 45, 5, 53, 5, 61, 5, 69, 5, 77, 5, 85, 5, 93, 5, 101, 5, 109, 5, 117, 5, 125, 5, 133, 5, 141, 5, 149, 5, 157, 5, 165, 5, 173, 5, 181, 5, 189, 5, 197, 5, 205, 5, 213, 5, 221, 5, 229, 5, 237, 5, 245, 5, 253, 5, 5, 6, 13, 6, 21, 6, 29, 6, 37, 6, 45, 6, 53, 6, 61, 6, 69, 6, 77, 6, 85, 6, 93, 6, 101, 6, 109, 6, 117, 6, 125, 6, 133, 6, 141, 6, 149, 6, 157, 6, 165, 6, 173, 6, 181, 6, 189, 6, 197, 6, 205, 6, 213, 6, 221, 6, 229, 6, 237, 6, 245, 6, 253, 6, 5, 7, 13, 7, 21, 7, 29, 7, 37, 7, 45, 7, 53, 7, 61, 7, 69, 7, 77, 7, 85, 7, 93, 7, 101, 7, 109, 7, 117, 7, 125, 7, 133, 7, 141, 7, 149, 7, 157, 7, 165, 7, 173, 7, 181, 7, 189, 7, 197, 7, 205, 7, 213, 7, 221, 7, 229, 7, 237, 7, 245, 7, 253, 7, 5, 8, 13, 8, 21, 8, 29, 8, 37, 8, 45, 8, 53, 8, 61, 8, 69, 8, 77, 8, 85, 8, 93, 8, 101, 8, 109, 8, 117, 8, 125, 8, 133, 8, 141, 8, 149, 8, 157, 8, 165, 8, 173, 8, 181, 8, 189, 8, 197, 8, 205, 8, 213, 8, 221, 8, 229, 8, 237, 8, 245, 8, 253, 8, 5, 9, 13, 9, 21, 9, 29, 9, 37, 9, 45, 9, 53, 9, 61, 9, 69, 9, 77, 9, 85, 9, 93, 9, 101, 9, 109, 9, 117, 9, 125, 9, 133, 9, 141, 9, 149, 9, 157, 9, 165, 9, 173, 9, 181, 9, 189, 9, 197, 9, 205, 9, 213, 9, 221, 9, 229, 9, 237, 9, 245, 9, 253, 9, 5, 10, 13, 10, 21, 10, 29, 10, 37, 10, 45, 10, 53, 10, 61, 10, 69, 10, 77, 10, 85, 10, 93, 10, 101, 10, 109, 10, 117, 10, 125, 10, 133, 10, 141, 10, 149, 10, 157, 10, 165, 10, 173, 10, 181, 10, 189, 10, 197, 10, 205, 10, 213, 10, 221, 10, 229, 10, 237, 10, 245, 10, 253, 10, 5, 11, 13, 11, 21, 11, 29, 11, 37, 11, 45, 11, 53, 11, 61, 11, 69, 11, 77, 11, 85, 11, 93, 11, 101, 11, 109, 11, 117, 11, 125, 11, 133, 11, 141, 11, 149, 11, 157, 11, 165, 11, 173, 11, 181, 11, 189, 11, 197, 11, 205, 11, 213, 11, 221, 11, 229, 11, 237, 11, 245, 11, 253, 11, 5, 12, 13, 12, 21, 12, 29, 12, 37, 12, 45, 12, 53, 12, 61, 12, 69, 12, 77, 12, 85, 12, 93, 12, 101, 12, 109, 12, 117, 12, 125, 12, 133, 12, 141, 12, 149, 12, 157, 12, 165, 12, 173, 12, 181, 12, 189, 12, 197, 12, 205, 12, 213, 12, 221, 12, 229, 12, 237, 12, 245, 12, 253, 12, 5, 13, 13, 13, 21, 13, 29, 13, 37, 13, 45, 13, 53, 13, 61, 13, 69, 13, 77, 13, 85, 13, 93, 13, 101, 13, 109, 13, 117, 13, 125, 13, 133, 13, 141, 13, 149, 13, 157, 13, 165, 13, 173, 13, 181, 13, 189, 13, 197, 13, 205, 13, 213, 13, 221, 13, 229, 13, 237, 13, 245, 13, 253, 13, 5, 14, 13, 14, 21, 14, 29, 14, 37, 14, 45, 14, 53, 14, 61, 14, 69, 14, 77, 14, 85, 14, 93, 14, 101, 14, 109, 14, 117, 14, 125, 14, 133, 14, 141, 14, 149, 14, 157, 14, 165, 14, 173, 14, 181, 14, 189, 14, 197, 14, 205, 14, 213, 14, 221, 14, 229, 14, 237, 14, 245, 14, 253, 14, 5, 15, 13, 15, 21, 15, 29, 15, 37, 15, 45, 15, 53, 15, 61, 15, 69, 15, 77, 15, 85, 15, 93, 15, 101, 15, 109, 15, 117, 15, 125, 15, 133, 15, 141, 15, 149, 15, 157, 15, 165, 15, 173, 15, 181, 15, 189, 15, 197, 15, 205, 15, 213, 15, 221, 15, 229, 15, 237, 15, 245, 15, 253, 15, 5, 16, 13, 16, 21, 16, 29, 16, 37, 16, 45, 16, 53, 16, 61, 16, 69, 16, 77, 16, 85, 16, 93, 16, 101, 16, 109, 16, 117, 16, 125, 16, 133, 16, 141, 16, 149, 16, 157, 16, 165, 16, 173, 16, 181, 16, 189, 16, 197, 16, 205, 16, 213, 16, 221, 16, 229, 16, 237, 16, 245, 16, 253, 16, 5, 17, 13, 17, 21, 17, 29, 17, 37, 17, 45, 17, 53, 17, 61, 17, 69, 17, 77, 17, 85, 17, 93, 17, 101, 17, 109, 17, 117, 17, 125, 17, 133, 17, 141, 17, 149, 17, 157, 17, 165, 17, 173, 17, 181, 17, 189, 17, 197, 17, 205, 17, 213, 17, 221, 17, 229, 17, 237, 17, 245, 17, 253, 17, 5, 18, 13, 18, 21, 18, 29, 18, 37, 18, 45, 18, 53, 18, 61, 18, 69, 18, 77, 18, 85, 18, 93, 18, 101, 18, 109, 18, 117, 18, 125, 18, 133, 18, 141, 18, 149, 18, 157, 18, 165, 18, 173, 18, 181, 18, 189, 18, 197, 18, 205, 18, 213, 18, 221, 18, 229, 18, 237, 18, 245, 18, 253, 18, 5, 19, 13, 19, 21, 19, 29, 19, 37, 19, 45, 19, 53, 19, 61, 19, 69, 19, 77, 19, 85, 19, 93, 19, 101, 19, 109, 19, 117, 19, 125, 19, 133, 19, 141, 19, 149, 19, 157, 19, 165, 19, 173, 19, 181, 19, 189, 19, 197, 19, 205, 19, 213, 19, 221, 19, 229, 19, 237, 19, 245, 19, 253, 19, 5, 20, 13, 20, 21, 20, 29, 20, 37, 20, 45, 20, 53, 20, 61, 20, 69, 20, 77, 20, 85, 20, 93, 20, 101, 20, 109, 20, 117, 20, 125, 20, 133, 20, 141, 20, 149, 20, 157, 20, 165, 20, 173, 20, 181, 20, 189, 20, 197, 20, 205, 20, 213, 20, 221, 20, 229, 20, 237, 20, 245, 20, 253, 20, 5, 21, 13, 21, 21, 21, 29, 21, 37, 21, 45, 21, 53, 21, 61, 21, 69, 21, 77, 21, 85, 21, 93, 21, 101, 21, 109, 21, 117, 21, 125, 21, 133, 21, 141, 21, 149, 21, 157, 21, 165, 21, 173, 21, 181, 21, 189, 21, 197, 21, 205, 21, 213, 21, 221, 21, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 1, 23, 161, 0, 48, 0, 48, 0, 153, 0, 9, 23, 153, 0, 17, 23, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 229, 21, 237, 21, 245, 21, 253, 21, 5, 22, 13, 22, 21, 22, 29, 22, 37, 22, 45, 22, 53, 22, 61, 22, 69, 22, 77, 22, 85, 22, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 92, 22, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 180, 1, 232, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 128, 8, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 192, 8, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 97, 22, 105, 22, 113, 22, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 121, 22, 129, 22, 137, 22, 145, 22, 153, 22, 161, 22, 169, 22, 177, 22, 185, 22, 193, 22, 201, 22, 209, 22, 217, 22, 225, 22, 233, 22, 241, 22, 249, 22, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 41, 0, 202, 65, 39, 0, 202, 130, 39, 0, 202, 195, 39, 0, 202, 4, 40, 0, 202, 69, 40, 0, 202, 134, 40, 0, 202, 199, 40, 0, 202, 8, 41, 0, 202, 73, 41, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 134, 2, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 38, 8, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 201, 6, 11, 0, 201, 70, 14, 0, 201, 134, 17, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 198, 20, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 70, 28, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 38, 31, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 198, 36, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 201, 166, 39, 0, 201, 230, 42, 0, 201, 38, 46, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 102, 49, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 134, 3, 0, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 66, 117, 0, 196, 160, 156, 46, 198, 65, 0, 0, 134, 3, 0, 50, 201, 230, 56, 0, 196, 150, 156, 50, 198, 129, 0, 0, 134, 3, 0, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 198, 65, 1, 0, 134, 3, 0, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 198, 129, 1, 0, 134, 3, 0, 82, 196, 142, 156, 82, 201, 198, 59, 0, 196, 136, 156, 90, 160, 5, 38, 93, 197, 130, 117, 0, 198, 193, 1, 0, 5, 3, 0, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 194, 117, 0, 196, 160, 5, 46, 198, 1, 2, 0, 5, 3, 0, 50, 201, 102, 63, 0, 196, 150, 5, 50, 198, 65, 2, 0, 5, 3, 0, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 198, 1, 3, 0, 5, 3, 0, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 198, 65, 3, 0, 5, 3, 0, 82, 196, 142, 5, 82, 201, 70, 66, 0, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 134, 2, 0, 42, 5, 2, 0, 42, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 69, 0, 201, 134, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 50, 5, 2, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 75, 0, 201, 70, 78, 0, 198, 97, 0, 0, 198, 33, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 58, 5, 2, 0, 58, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 81, 0, 201, 6, 84, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 68, 5, 3, 0, 68, 201, 230, 86, 0, 201, 198, 89, 0, 198, 33, 1, 0, 198, 225, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 70, 5, 2, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 82, 5, 2, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 92, 0, 201, 230, 95, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 99, 0, 201, 6, 102, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 104, 0, 201, 38, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 198, 225, 1, 0, 198, 161, 0, 0, 198, 97, 2, 0, 198, 97, 1, 0, 198, 33, 3, 0, 198, 161, 1, 0, 198, 97, 3, 0, 134, 70, 0, 82, 5, 70, 0, 82, 134, 71, 0, 82, 5, 71, 0, 82, 134, 72, 0, 82, 5, 72, 0, 82, 134, 73, 0, 82, 5, 73, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 111, 0, 201, 70, 114, 0, 197, 98, 8, 0, 197, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 1, 0, 198, 193, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 117, 0, 201, 6, 120, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 122, 0, 201, 198, 125, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 201, 166, 128, 0, 201, 134, 131, 0, 201, 102, 134, 0, 201, 70, 137, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 140, 0, 201, 6, 143, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 66, 5, 3, 0, 66, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 145, 0, 201, 198, 148, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 151, 0, 201, 134, 154, 0, 201, 102, 157, 0, 201, 70, 160, 0, 201, 38, 163, 0, 201, 6, 166, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 197, 226, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 168, 0, 201, 38, 172, 0, 201, 102, 175, 0, 201, 166, 178, 0, 201, 230, 181, 0, 201, 38, 185, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 24, 0, 197, 226, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 188, 0, 201, 70, 191, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 194, 0, 201, 6, 197, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 50, 5, 71, 0, 50, 134, 73, 0, 50, 5, 73, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 199, 0, 201, 198, 202, 0, 201, 166, 205, 0, 201, 134, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 5, 0, 197, 34, 6, 0, 198, 130, 8, 0, 198, 2, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 211, 0, 201, 70, 214, 0, 201, 38, 217, 0, 201, 102, 220, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 6, 0, 197, 98, 7, 0, 198, 130, 9, 0, 198, 2, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 223, 0, 201, 230, 226, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 172, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 27, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 57, 238, 6, 5, 126, 187, 6, 5, 5, 248, 6, 5, 126, 213, 7, 5, 180, 238, 6, 5, 181, 197, 9, 5, 6, 182, 6, 5, 63, 227, 7, 5, 120, 218, 6, 5, 219, 181, 6, 5, 50, 219, 7, 5, 161, 168, 6, 5, 118, 196, 6, 5, 96, 208, 6, 5, 91, 171, 6, 5, 172, 192, 6, 5, 129, 215, 6, 5, 104, 178, 6, 5, 95, 202, 6, 5, 81, 169, 6, 5, 22, 170, 6, 5, 137, 182, 6, 5, 105, 231, 8, 5, 204, 221, 6, 5, 244, 171, 6, 5, 59, 195, 6, 5, 149, 173, 6, 5, 57, 223, 6, 5, 86, 241, 6, 5, 74, 199, 6, 5, 234, 228, 7, 5, 132, 227, 6, 5, 3, 222, 6, 5, 159, 246, 6, 5, 48, 224, 6, 5, 168, 230, 6, 5, 155, 177, 7, 5, 148, 211, 6, 5, 96, 249, 6, 5, 133, 208, 6, 5, 70, 177, 6, 5, 60, 233, 6, 5, 24, 241, 6, 5, 166, 221, 6, 5, 223, 173, 6, 5, 72, 220, 6, 5, 203, 175, 6, 5, 203, 185, 6, 5, 173, 192, 6, 5, 165, 195, 6, 5, 236, 184, 6, 5, 111, 237, 6, 5, 13, 187, 6, 5, 254, 239, 6, 5, 232, 185, 6, 5, 221, 238, 6, 5, 204, 185, 6, 5, 192, 192, 6, 5, 73, 220, 6, 5, 223, 174, 6, 5, 213, 233, 6, 5, 84, 185, 6, 5, 229, 189, 7, 5, 95, 222, 6, 5, 80, 246, 6, 5, 6, 214, 7, 5, 161, 229, 6, 5, 91, 180, 6, 5, 167, 195, 6, 5, 138, 182, 6, 5, 49, 230, 7, 5, 186, 218, 6, 5, 253, 242, 8, 5, 7, 243, 7, 5, 189, 209, 6, 5, 145, 215, 6, 5, 160, 246, 6, 5, 167, 177, 6, 5, 119, 222, 6, 5, 50, 230, 6, 5, 23, 170, 6, 5, 74, 207, 6, 5, 4, 222, 6, 5, 222, 214, 7, 5, 15, 223, 7, 5, 200, 191, 6, 5, 82, 245, 6, 5, 74, 184, 6, 5, 125, 237, 6, 5, 13, 212, 6, 5, 87, 213, 6, 5, 247, 235, 6, 5, 87, 211, 6, 5, 2, 218, 6, 5, 42, 235, 6, 5, 10, 242, 6, 5, 143, 186, 6, 5, 62, 228, 6, 5, 188, 184, 7, 5, 114, 221, 6, 5, 255, 240, 6, 5, 85, 226, 6, 5, 31, 213, 6, 5, 102, 210, 6, 5, 140, 171, 6, 5, 230, 168, 6, 5, 245, 212, 6, 5, 254, 208, 6, 5, 192, 209, 6, 5, 75, 207, 6, 5, 244, 221, 6, 5, 206, 221, 6, 5, 10, 222, 6, 5, 244, 218, 6, 5, 155, 188, 6, 5, 98, 235, 6, 5, 99, 202, 6, 5, 235, 247, 6, 5, 78, 208, 6, 5, 97, 208, 6, 5, 154, 183, 6, 5, 176, 228, 6, 5, 39, 237, 6, 5, 226, 241, 6, 5, 107, 201, 6, 5, 220, 181, 6, 5, 188, 201, 6, 5, 246, 181, 6, 5, 13, 242, 6, 5, 12, 219, 6, 5, 250, 173, 6, 5, 82, 249, 6, 5, 197, 246, 6, 5, 186, 196, 6, 5, 237, 220, 6, 5, 221, 175, 6, 5, 128, 247, 6, 5, 164, 185, 6, 5, 233, 219, 6, 5, 148, 172, 6, 5, 144, 189, 6, 5, 29, 175, 6, 5, 124, 235, 6, 5, 24, 234, 6, 5, 63, 238, 6, 5, 26, 236, 6, 5, 13, 194, 6, 5, 196, 194, 6, 5, 87, 236, 6, 5, 80, 186, 6, 5, 92, 180, 6, 5, 246, 221, 6, 5, 202, 246, 6, 5, 164, 169, 6, 5, 228, 174, 7, 5, 182, 249, 6, 5, 191, 249, 6, 5, 23, 221, 6, 5, 213, 173, 6, 5, 218, 233, 6, 5, 254, 173, 6, 5, 72, 176, 6, 5, 244, 238, 6, 5, 67, 241, 6, 5, 206, 170, 6, 5, 69, 202, 7, 5, 170, 195, 6, 5, 41, 245, 6, 5, 225, 207, 6, 5, 83, 184, 6, 5, 116, 202, 6, 5, 142, 248, 6, 5, 227, 241, 6, 5, 169, 216, 6, 5, 179, 182, 6, 5, 175, 208, 6, 5, 117, 185, 6, 5, 245, 228, 6, 5, 179, 196, 6, 5, 177, 239, 6, 5, 8, 238, 6, 5, 84, 183, 6, 5, 180, 182, 6, 5, 222, 221, 6, 5, 98, 222, 6, 5, 162, 232, 6, 5, 187, 206, 6, 5, 83, 186, 6, 5, 37, 185, 6, 5, 234, 170, 6, 5, 98, 180, 6, 5, 164, 173, 6, 5, 120, 185, 6, 5, 82, 187, 6, 5, 162, 241, 6, 5, 237, 210, 6, 5, 122, 205, 6, 5, 165, 205, 6, 5, 234, 206, 6, 5, 177, 206, 6, 5, 203, 190, 6, 5, 172, 222, 6, 5, 246, 188, 6, 5, 192, 246, 6, 5, 170, 208, 6, 5, 242, 179, 6, 5, 104, 186, 6, 5, 175, 222, 6, 5, 20, 170, 6, 5, 165, 214, 6, 5, 222, 174, 6, 5, 237, 204, 6, 5, 59, 187, 6, 5, 38, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 5, 204, 221, 6, 5, 221, 210, 6, 5, 139, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 8, 5, 57, 238, 8, 5, 6, 182, 8, 5, 169, 219, 7, 5, 109, 223, 7, 5, 185, 220, 8, 5, 67, 247, 7, 5, 211, 231, 6, 5, 99, 193, 6, 5, 180, 238, 6, 5, 99, 171, 6, 5, 226, 179, 7, 5, 77, 226, 6, 5, 168, 178, 6, 5, 120, 218, 198, 131, 31, 0, 198, 227, 31, 0, 198, 67, 32, 0, 198, 163, 32, 0, 198, 3, 33, 0, 198, 99, 33, 0, 198, 195, 33, 0, 198, 35, 34, 0, 198, 131, 34, 0, 198, 227, 34, 0, 198, 67, 35, 0, 198, 163, 35, 0, 198, 3, 36, 0, 198, 99, 36, 0, 197, 227, 35, 0, 198, 195, 36, 0, 198, 35, 37, 0, 198, 131, 37, 0, 198, 227, 37, 0, 197, 67, 36, 0, 198, 67, 38, 0, 198, 163, 38, 0, 197, 163, 36, 0, 197, 3, 37, 0, 198, 3, 39, 0, 198, 99, 39, 0, 198, 195, 39, 0, 198, 35, 40, 0, 197, 99, 37, 0, 198, 131, 40, 0, 197, 195, 37, 0, 198, 227, 40, 0, 197, 35, 38, 0, 198, 67, 41, 0, 197, 131, 38, 0, 197, 227, 38, 0, 6, 5, 204, 229, 6, 5, 88, 241, 6, 5, 161, 229, 6, 5, 50, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 5, 57, 238, 7, 5, 6, 182, 7, 5, 169, 219, 6, 5, 109, 223, 6, 5, 80, 230, 6, 5, 198, 204, 6, 5, 63, 214, 6, 5, 161, 168, 6, 5, 169, 196, 6, 5, 204, 221, 7, 5, 7, 243, 6, 5, 200, 191, 6, 5, 15, 223, 6, 5, 189, 209, 6, 5, 170, 195, 6, 5, 132, 227, 6, 5, 186, 218, 6, 5, 207, 247, 7, 5, 63, 241, 6, 5, 246, 220, 6, 5, 29, 209, 6, 5, 213, 225, 6, 5, 52, 172, 6, 5, 39, 248, 6, 5, 83, 201, 6, 5, 106, 208, 6, 5, 22, 210, 6, 5, 148, 211, 6, 5, 71, 222, 6, 5, 17, 241, 6, 5, 29, 240, 6, 5, 31, 248, 6, 5, 228, 232, 6, 5, 87, 234, 6, 5, 144, 233, 6, 5, 66, 246, 6, 5, 185, 220, 7, 5, 67, 247, 6, 5, 211, 231, 7, 5, 30, 250, 7, 5, 87, 241, 6, 5, 64, 238, 6, 5, 97, 249, 6, 5, 101, 235, 6, 5, 166, 193, 6, 5, 199, 214, 6, 5, 32, 249, 6, 5, 103, 233, 6, 5, 3, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 39, 0, 197, 130, 39, 0, 197, 194, 39, 0, 197, 2, 40, 0, 197, 66, 40, 0, 197, 130, 40, 0, 197, 194, 40, 0, 197, 2, 41, 0, 197, 66, 41, 0, 197, 131, 41, 0, 197, 227, 41, 0, 197, 67, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 42, 0, 197, 34, 43, 0, 197, 98, 43, 0, 197, 162, 43, 0, 197, 226, 43, 0, 197, 34, 44, 0, 197, 98, 44, 0, 197, 162, 44, 0, 197, 226, 44, 0, 197, 34, 45, 0, 197, 99, 45, 0, 197, 195, 45, 0, 197, 35, 46, 0, 197, 131, 46, 0, 197, 227, 46, 0, 197, 67, 47, 0, 197, 163, 47, 0, 197, 3, 48, 0, 197, 99, 48, 0, 197, 195, 48, 0, 197, 35, 49, 0, 197, 131, 49, 0, 197, 227, 49, 0, 197, 67, 50, 0, 197, 163, 50, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 51, 0, 197, 66, 51, 0, 197, 130, 51, 0, 197, 194, 51, 0, 197, 4, 52, 0, 197, 130, 52, 0, 197, 194, 52, 0, 197, 2, 53, 0, 197, 66, 53, 0, 197, 130, 53, 0, 197, 194, 53, 0, 197, 2, 54, 0, 197, 66, 54, 0, 197, 130, 54, 0, 197, 195, 54, 0, 197, 35, 55, 0, 197, 131, 55, 0, 197, 227, 55, 0, 197, 67, 56, 0, 197, 163, 56, 0, 197, 3, 57, 0, 197, 99, 57, 0, 197, 195, 57, 0, 197, 35, 58, 0, 197, 131, 58, 0, 197, 227, 58, 0, 197, 67, 59, 0, 197, 163, 59, 0, 197, 3, 60, 0, 197, 99, 60, 0, 197, 195, 60, 0, 197, 35, 61, 0, 197, 131, 61, 0, 197, 227, 61, 0, 197, 67, 62, 0, 197, 163, 62, 0, 192, 0, 0, 0, 5, 5, 57, 238, 5, 5, 226, 179, 5, 5, 190, 198, 5, 5, 63, 214, 5, 5, 184, 220, 5, 5, 210, 231, 5, 5, 16, 188, 5, 5, 142, 228, 5, 5, 47, 245, 5, 5, 169, 219, 5, 5, 185, 220, 5, 5, 211, 231, 5, 5, 249, 191, 5, 5, 10, 172, 5, 5, 221, 241, 5, 5, 151, 208, 5, 5, 168, 184, 5, 5, 88, 175, 5, 5, 89, 175, 5, 5, 77, 248, 5, 5, 66, 216, 5, 5, 206, 212, 5, 5, 7, 222, 5, 5, 6, 222, 5, 5, 7, 217, 5, 5, 99, 171, 5, 5, 251, 237, 5, 5, 189, 176, 5, 5, 19, 180, 5, 5, 59, 223, 5, 5, 86, 174, 5, 5, 15, 180, 5, 5, 8, 217, 5, 5, 90, 203, 5, 5, 16, 180, 5, 5, 64, 241, 5, 5, 91, 203, 5, 5, 84, 236, 5, 5, 125, 171, 5, 5, 208, 219, 5, 5, 126, 187, 5, 5, 148, 196, 5, 5, 152, 185, 5, 5, 226, 235, 5, 5, 209, 215, 5, 5, 67, 247, 5, 5, 174, 192, 5, 5, 135, 195, 5, 5, 70, 183, 5, 5, 218, 186, 5, 5, 228, 175, 5, 5, 65, 173, 5, 5, 222, 203, 5, 5, 210, 248, 5, 5, 5, 248, 5, 5, 160, 168, 5, 5, 96, 228, 5, 5, 214, 177, 5, 5, 88, 229, 5, 5, 6, 248, 5, 5, 51, 196, 5, 5, 102, 202, 5, 5, 39, 197, 5, 5, 126, 213, 5, 5, 200, 183, 5, 5, 99, 238, 5, 5, 218, 238, 5, 5, 254, 209, 5, 5, 79, 230, 5, 5, 170, 196, 5, 5, 171, 196, 5, 5, 220, 227, 5, 5, 135, 207, 5, 5, 219, 238, 5, 5, 58, 238, 5, 5, 79, 246, 5, 5, 22, 230, 5, 5, 166, 244, 5, 5, 141, 189, 5, 5, 26, 182, 5, 5, 139, 201, 5, 5, 213, 239, 5, 5, 156, 213, 5, 5, 50, 212, 5, 5, 12, 216, 5, 5, 214, 189, 5, 5, 176, 186, 5, 5, 94, 174, 5, 5, 101, 174, 5, 5, 180, 238, 5, 5, 253, 239, 5, 5, 49, 236, 5, 5, 207, 208, 5, 5, 169, 196, 5, 5, 197, 214, 5, 5, 243, 237, 5, 5, 46, 231, 5, 5, 159, 232, 5, 5, 169, 184, 5, 5, 172, 196, 5, 5, 212, 231, 5, 5, 226, 189, 5, 5, 118, 222, 5, 5, 83, 180, 5, 5, 245, 221, 5, 5, 255, 191, 5, 5, 48, 210, 5, 5, 38, 193, 5, 5, 59, 197, 5, 5, 208, 221, 5, 5, 99, 207, 5, 5, 142, 189, 5, 5, 220, 206, 5, 5, 58, 206, 5, 5, 252, 248, 5, 5, 46, 219, 5, 5, 100, 235, 5, 5, 142, 236, 5, 5, 43, 184, 5, 5, 19, 220, 5, 5, 227, 209, 5, 5, 200, 184, 5, 5, 151, 224, 5, 5, 143, 241, 5, 5, 57, 177, 5, 5, 152, 245, 5, 5, 205, 184, 5, 5, 234, 246, 5, 5, 36, 187, 5, 5, 207, 184, 5, 5, 60, 206, 5, 5, 7, 204, 5, 5, 63, 239, 5, 5, 181, 197, 5, 5, 155, 201, 5, 5, 212, 206, 5, 5, 222, 241, 5, 5, 19, 246, 5, 5, 15, 222, 5, 5, 18, 222, 5, 5, 6, 182, 5, 5, 156, 175, 5, 5, 128, 241, 5, 5, 229, 199, 5, 5, 127, 241, 5, 5, 64, 243, 5, 5, 227, 189, 5, 5, 102, 214, 5, 5, 80, 230, 5, 5, 50, 196, 5, 5, 108, 223, 5, 5, 75, 224, 5, 5, 165, 185, 5, 5, 166, 185, 5, 5, 25, 236, 5, 5, 92, 233, 5, 5, 28, 236, 5, 5, 110, 214, 5, 5, 31, 236, 5, 5, 107, 192, 5, 5, 63, 227, 5, 5, 167, 228, 5, 5, 177, 198, 5, 5, 156, 177, 5, 5, 156, 194, 5, 5, 219, 187, 5, 5, 228, 238, 5, 5, 62, 173, 5, 5, 3, 189, 5, 5, 173, 209, 5, 5, 56, 238, 5, 5, 199, 232, 5, 5, 13, 196, 5, 5, 218, 226, 5, 5, 102, 203, 5, 5, 200, 232, 5, 5, 17, 196, 5, 5, 5, 238, 5, 5, 95, 216, 5, 5, 184, 171, 5, 5, 106, 241, 5, 5, 171, 233, 5, 5, 255, 177, 5, 5, 245, 202, 5, 5, 65, 181, 5, 5, 231, 207, 5, 5, 120, 218, 5, 5, 121, 218, 5, 5, 94, 192, 5, 5, 95, 192, 5, 5, 168, 228, 5, 5, 220, 238, 5, 5, 65, 221, 5, 5, 122, 218, 5, 5, 137, 201, 5, 5, 227, 179, 5, 5, 81, 244, 5, 5, 191, 195, 5, 5, 5, 214, 5, 5, 52, 175, 5, 5, 162, 168, 5, 5, 39, 245, 5, 5, 166, 195, 5, 5, 136, 195, 5, 5, 92, 171, 5, 5, 181, 218, 5, 5, 188, 176, 5, 5, 149, 183, 5, 5, 178, 219, 5, 5, 72, 206, 5, 5, 93, 171, 5, 5, 113, 172, 5, 5, 59, 249, 5, 5, 8, 222, 5, 5, 174, 224, 5, 5, 48, 245, 5, 5, 76, 184, 5, 5, 225, 231, 5, 5, 226, 231, 5, 5, 221, 227, 5, 5, 44, 189, 5, 5, 3, 227, 5, 5, 146, 218, 5, 5, 30, 215, 5, 5, 217, 184, 5, 5, 85, 185, 5, 5, 171, 171, 5, 5, 170, 177, 5, 5, 124, 204, 5, 5, 182, 238, 5, 5, 207, 173, 5, 5, 124, 173, 5, 5, 129, 219, 5, 5, 125, 173, 5, 5, 100, 238, 5, 5, 190, 209, 5, 5, 240, 207, 5, 5, 147, 218, 5, 5, 112, 182, 5, 5, 208, 173, 5, 5, 79, 237, 5, 5, 93, 215, 5, 5, 111, 247, 5, 5, 32, 213, 5, 5, 250, 229, 5, 5, 82, 230, 5, 5, 12, 194, 5, 5, 118, 193, 5, 5, 176, 237, 5, 5, 72, 183, 5, 5, 114, 172, 5, 5, 149, 218, 5, 5, 171, 228, 5, 5, 54, 183, 5, 5, 200, 178, 5, 5, 159, 182, 5, 5, 68, 247, 5, 5, 199, 214, 5, 5, 121, 212, 5, 5, 130, 241, 5, 5, 153, 179, 5, 5, 6, 181, 5, 5, 119, 230, 5, 5, 229, 238, 5, 5, 242, 233, 5, 5, 178, 198, 5, 5, 62, 238, 5, 5, 97, 192, 5, 5, 2, 168, 5, 5, 83, 230, 5, 5, 197, 192, 5, 5, 203, 183, 5, 5, 27, 182, 5, 5, 87, 234, 5, 5, 216, 195, 5, 5, 207, 212, 5, 5, 244, 177, 5, 5, 163, 183, 5, 5, 139, 225, 5, 5, 112, 247, 5, 5, 5, 241, 5, 5, 201, 191, 5, 5, 69, 191, 5, 5, 223, 241, 5, 5, 29, 177, 5, 5, 67, 243, 5, 5, 179, 219, 5, 5, 26, 229, 5, 5, 210, 175, 5, 5, 212, 173, 5, 5, 248, 235, 5, 5, 104, 232, 5, 5, 161, 220, 5, 5, 106, 173, 5, 5, 73, 206, 5, 5, 115, 172, 5, 5, 200, 235, 5, 5, 244, 233, 5, 5, 27, 229, 5, 5, 23, 248, 5, 5, 88, 244, 5, 5, 18, 232, 5, 5, 140, 211, 5, 5, 172, 171, 5, 5, 44, 186, 5, 5, 153, 210, 5, 5, 174, 210, 5, 5, 148, 233, 5, 5, 31, 169, 5, 5, 199, 234, 5, 5, 40, 204, 5, 5, 171, 247, 5, 5, 22, 221, 5, 5, 91, 217, 5, 5, 140, 176, 5, 5, 232, 169, 5, 5, 11, 222, 5, 5, 40, 193, 5, 5, 208, 212, 5, 5, 233, 238, 5, 5, 112, 223, 5, 5, 184, 238, 5, 5, 20, 246, 5, 5, 102, 179, 5, 5, 230, 187, 5, 5, 228, 206, 5, 5, 9, 178, 5, 5, 24, 248, 5, 5, 12, 172, 5, 5, 92, 217, 5, 5, 26, 170, 5, 5, 68, 245, 5, 5, 133, 176, 5, 5, 90, 229, 5, 5, 201, 178, 5, 5, 25, 248, 5, 5, 31, 250, 5, 5, 89, 241, 5, 5, 80, 237, 5, 5, 39, 226, 5, 5, 250, 244, 5, 5, 157, 188, 5, 5, 48, 170, 5, 5, 245, 227, 5, 5, 238, 220, 5, 5, 131, 241, 5, 5, 234, 238, 5, 5, 206, 183, 5, 5, 33, 250, 5, 5, 249, 185, 5, 5, 80, 211, 5, 5, 5, 227, 5, 5, 154, 210, 5, 5, 231, 231, 5, 5, 136, 217, 5, 5, 187, 240, 5, 5, 63, 228, 5, 5, 39, 215, 5, 5, 89, 222, 5, 5, 85, 198, 5, 5, 83, 169, 5, 5, 123, 212, 5, 5, 43, 191, 5, 5, 218, 188, 5, 5, 108, 201, 5, 5, 190, 232, 5, 5, 112, 185, 5, 5, 46, 237, 5, 5, 232, 168, 5, 5, 46, 182, 5, 5, 51, 209, 5, 5, 43, 193, 5, 5, 9, 182, 5, 5, 126, 171, 5, 5, 108, 192, 5, 5, 252, 188, 5, 5, 196, 191, 5, 5, 72, 187, 5, 5, 209, 217, 5, 5, 135, 226, 5, 5, 197, 194, 5, 5, 141, 176, 5, 5, 245, 238, 5, 5, 247, 221, 5, 5, 28, 234, 5, 5, 24, 221, 5, 5, 228, 227, 5, 5, 137, 198, 5, 5, 123, 246, 5, 5, 148, 184, 5, 5, 162, 200, 5, 5, 109, 238, 5, 5, 207, 174, 5, 5, 127, 199, 5, 5, 250, 186, 5, 5, 109, 202, 5, 5, 171, 239, 5, 5, 19, 222, 5, 5, 80, 208, 5, 5, 202, 247, 5, 5, 203, 234, 5, 5, 90, 241, 5, 5, 35, 168, 5, 5, 149, 205, 5, 5, 152, 209, 5, 5, 222, 181, 5, 5, 77, 206, 5, 5, 53, 180, 5, 5, 236, 172, 5, 5, 148, 174, 5, 5, 203, 235, 5, 5, 210, 185, 5, 5, 130, 247, 5, 5, 67, 238, 5, 5, 16, 219, 5, 5, 66, 177, 5, 5, 170, 231, 5, 5, 117, 223, 5, 5, 172, 177, 5, 5, 236, 205, 5, 5, 223, 224, 5, 5, 198, 194, 5, 5, 174, 245, 5, 5, 159, 172, 5, 5, 13, 216, 5, 5, 142, 199, 5, 5, 247, 172, 5, 5, 81, 211, 5, 5, 101, 211, 5, 5, 193, 195, 5, 5, 91, 230, 5, 5, 108, 189, 5, 5, 133, 196, 5, 5, 137, 174, 5, 5, 248, 245, 5, 5, 39, 250, 5, 5, 93, 175, 5, 5, 96, 216, 5, 5, 238, 205, 5, 5, 3, 197, 5, 5, 196, 222, 5, 5, 238, 226, 5, 5, 94, 221, 5, 5, 186, 227, 5, 5, 180, 171, 5, 5, 25, 210, 5, 5, 243, 232, 5, 5, 210, 170, 5, 5, 181, 227, 5, 5, 228, 241, 5, 5, 116, 231, 5, 5, 225, 176, 5, 5, 102, 181, 5, 5, 46, 217, 5, 5, 179, 234, 5, 5, 19, 187, 5, 5, 108, 199, 5, 5, 92, 230, 5, 5, 48, 198, 5, 5, 6, 239, 5, 5, 48, 184, 5, 5, 76, 203, 5, 5, 206, 249, 5, 5, 43, 216, 5, 5, 117, 202, 5, 5, 229, 240, 5, 5, 176, 191, 5, 5, 80, 196, 5, 5, 148, 215, 5, 5, 192, 219, 5, 5, 120, 212, 5, 5, 236, 223, 5, 5, 225, 183, 5, 5, 185, 230, 5, 5, 70, 202, 5, 5, 49, 184, 5, 5, 158, 213, 5, 5, 103, 169, 5, 5, 144, 241, 5, 5, 118, 214, 5, 5, 172, 231, 5, 5, 247, 233, 5, 5, 88, 234, 5, 5, 229, 241, 5, 5, 24, 179, 5, 5, 214, 173, 5, 5, 54, 175, 5, 5, 213, 246, 5, 5, 145, 236, 5, 5, 233, 202, 5, 5, 118, 202, 5, 5, 163, 200, 5, 5, 65, 223, 5, 5, 205, 193, 5, 5, 89, 234, 5, 5, 53, 184, 5, 5, 212, 191, 5, 5, 78, 197, 5, 5, 74, 233, 5, 5, 245, 211, 5, 5, 21, 194, 5, 5, 26, 171, 5, 5, 166, 175, 5, 5, 229, 182, 5, 5, 137, 183, 5, 5, 36, 236, 5, 5, 64, 168, 5, 5, 173, 169, 5, 5, 25, 242, 5, 5, 224, 233, 5, 5, 34, 170, 5, 5, 218, 189, 5, 5, 108, 173, 5, 5, 100, 246, 5, 5, 130, 171, 5, 5, 191, 196, 5, 5, 133, 237, 5, 5, 32, 177, 5, 5, 234, 202, 5, 5, 115, 228, 5, 5, 164, 200, 5, 5, 119, 172, 5, 5, 151, 249, 5, 5, 155, 185, 5, 5, 192, 186, 5, 5, 174, 169, 5, 5, 114, 226, 5, 5, 128, 222, 5, 5, 129, 222, 5, 5, 241, 207, 5, 5, 135, 178, 5, 5, 28, 225, 5, 5, 202, 197, 5, 5, 10, 176, 5, 5, 52, 234, 5, 5, 165, 212, 5, 5, 141, 225, 5, 5, 134, 189, 5, 5, 193, 238, 5, 5, 68, 214, 5, 5, 51, 226, 5, 5, 241, 184, 5, 5, 84, 196, 5, 5, 151, 195, 5, 5, 51, 224, 5, 5, 163, 173, 5, 5, 79, 195, 5, 5, 167, 182, 5, 5, 127, 246, 5, 5, 53, 199, 5, 5, 154, 197, 5, 5, 98, 249, 5, 5, 79, 197, 5, 5, 150, 215, 5, 5, 129, 210, 5, 5, 79, 206, 5, 5, 190, 248, 5, 5, 233, 229, 5, 5, 138, 206, 5, 5, 152, 223, 5, 5, 241, 201, 5, 5, 177, 191, 5, 5, 27, 180, 5, 5, 86, 249, 5, 5, 223, 169, 5, 5, 93, 230, 5, 5, 80, 197, 5, 5, 8, 210, 5, 5, 57, 172, 5, 5, 208, 193, 5, 5, 194, 244, 5, 5, 226, 237, 5, 5, 128, 246, 5, 5, 46, 220, 5, 5, 173, 216, 5, 5, 83, 211, 5, 5, 50, 240, 5, 5, 60, 174, 5, 5, 104, 215, 5, 5, 150, 236, 5, 5, 71, 219, 5, 5, 120, 247, 5, 5, 62, 176, 5, 5, 107, 193, 5, 5, 223, 192, 5, 5, 42, 229, 5, 5, 236, 241, 5, 5, 134, 171, 5, 5, 114, 219, 5, 5, 254, 225, 5, 5, 204, 228, 5, 5, 64, 213, 5, 5, 214, 236, 5, 5, 87, 183, 5, 5, 142, 225, 5, 5, 2, 230, 5, 5, 155, 181, 5, 5, 116, 233, 5, 5, 221, 173, 5, 5, 148, 221, 5, 5, 140, 198, 5, 5, 30, 179, 5, 5, 45, 250, 5, 5, 204, 172, 5, 5, 221, 226, 5, 5, 178, 169, 5, 5, 163, 233, 5, 5, 190, 190, 5, 5, 189, 237, 5, 5, 255, 244, 5, 5, 94, 175, 5, 5, 56, 236, 5, 5, 42, 241, 5, 5, 22, 194, 5, 5, 189, 234, 5, 5, 129, 244, 5, 5, 96, 176, 5, 5, 94, 184, 5, 5, 8, 170, 5, 5, 235, 246, 5, 5, 134, 249, 5, 5, 159, 208, 5, 5, 123, 192, 5, 5, 198, 238, 5, 5, 164, 233, 5, 5, 178, 235, 5, 5, 47, 172, 5, 5, 183, 180, 5, 5, 167, 172, 5, 5, 188, 245, 5, 5, 219, 211, 5, 5, 57, 227, 5, 5, 58, 227, 5, 5, 179, 169, 5, 5, 163, 243, 5, 5, 30, 205, 5, 5, 84, 195, 5, 5, 43, 229, 5, 5, 59, 183, 5, 5, 131, 173, 5, 5, 37, 187, 5, 5, 214, 223, 5, 5, 253, 246, 5, 5, 253, 223, 5, 5, 162, 231, 5, 5, 99, 184, 5, 5, 242, 242, 5, 5, 239, 218, 5, 5, 150, 202, 5, 5, 146, 211, 5, 5, 122, 243, 5, 5, 119, 194, 5, 5, 202, 206, 5, 5, 67, 169, 5, 5, 69, 179, 5, 5, 105, 225, 5, 5, 126, 188, 5, 5, 90, 195, 5, 5, 217, 230, 5, 5, 129, 220, 5, 5, 153, 215, 5, 5, 217, 197, 5, 5, 121, 172, 5, 5, 173, 175, 5, 5, 180, 219, 5, 5, 183, 169, 5, 5, 80, 233, 5, 5, 238, 240, 5, 5, 137, 237, 5, 5, 78, 225, 5, 5, 135, 224, 5, 5, 89, 237, 5, 5, 38, 182, 5, 5, 135, 171, 5, 5, 60, 193, 5, 5, 169, 177, 5, 5, 192, 243, 5, 5, 144, 225, 5, 5, 89, 186, 5, 5, 54, 171, 5, 5, 145, 175, 5, 5, 171, 211, 5, 5, 80, 172, 5, 5, 194, 201, 5, 5, 12, 177, 5, 5, 192, 240, 5, 5, 5, 244, 5, 5, 140, 249, 5, 5, 239, 169, 5, 5, 171, 223, 5, 5, 144, 168, 5, 5, 216, 175, 5, 5, 244, 241, 5, 5, 196, 244, 5, 5, 197, 249, 5, 5, 165, 220, 5, 5, 248, 175, 5, 5, 102, 196, 5, 5, 244, 174, 5, 5, 44, 220, 5, 5, 44, 188, 5, 5, 21, 245, 5, 5, 182, 216, 5, 5, 227, 236, 5, 5, 47, 179, 5, 5, 178, 233, 5, 5, 35, 205, 5, 5, 184, 169, 5, 5, 116, 213, 5, 5, 199, 195, 5, 5, 55, 203, 5, 5, 170, 205, 5, 5, 246, 206, 5, 5, 65, 215, 5, 5, 247, 231, 5, 5, 80, 225, 5, 5, 124, 240, 5, 5, 72, 180, 5, 5, 106, 248, 5, 5, 229, 232, 5, 5, 134, 220, 5, 5, 20, 216, 5, 5, 142, 196, 5, 5, 183, 227, 5, 5, 15, 250, 5, 5, 25, 214, 5, 5, 238, 230, 5, 5, 93, 201, 5, 5, 155, 173, 5, 5, 9, 187, 5, 5, 122, 203, 5, 5, 87, 214, 5, 5, 65, 174, 5, 5, 26, 173, 5, 5, 65, 229, 5, 5, 39, 192, 5, 5, 162, 171, 5, 5, 100, 191, 5, 5, 225, 175, 5, 5, 188, 226, 5, 5, 32, 178, 5, 5, 221, 194, 5, 5, 201, 196, 5, 5, 14, 220, 5, 5, 62, 183, 5, 5, 138, 232, 5, 5, 18, 197, 5, 5, 181, 181, 5, 5, 170, 194, 5, 5, 46, 194, 5, 5, 29, 227, 5, 5, 26, 204, 5, 5, 206, 171, 5, 5, 138, 186, 5, 5, 248, 231, 5, 5, 13, 224, 5, 5, 139, 232, 5, 5, 100, 194, 5, 5, 20, 209, 5, 5, 25, 238, 5, 5, 244, 195, 5, 5, 124, 193, 5, 5, 53, 216, 5, 5, 54, 213, 5, 5, 100, 183, 5, 5, 190, 247, 5, 5, 16, 168, 5, 5, 166, 219, 5, 5, 158, 238, 5, 5, 63, 198, 5, 5, 107, 211, 5, 5, 31, 173, 5, 5, 88, 239, 5, 5, 92, 178, 5, 5, 63, 196, 5, 5, 22, 235, 5, 5, 150, 199, 5, 5, 239, 193, 5, 5, 182, 175, 5, 5, 235, 177, 5, 5, 227, 194, 5, 5, 45, 220, 5, 5, 195, 243, 5, 5, 107, 172, 5, 5, 65, 171, 5, 5, 62, 168, 5, 5, 28, 219, 5, 5, 244, 224, 5, 5, 73, 175, 5, 5, 252, 172, 5, 5, 223, 200, 5, 5, 165, 210, 5, 5, 212, 195, 5, 5, 164, 215, 5, 5, 252, 207, 5, 5, 109, 230, 5, 5, 66, 211, 5, 5, 252, 216, 5, 5, 166, 210, 5, 5, 144, 173, 5, 5, 208, 203, 5, 5, 201, 201, 5, 5, 4, 206, 5, 5, 196, 199, 5, 5, 142, 169, 5, 5, 105, 242, 5, 5, 255, 170, 5, 5, 210, 243, 5, 5, 44, 247, 5, 5, 144, 223, 5, 5, 17, 241, 5, 5, 106, 188, 5, 5, 221, 216, 5, 5, 46, 174, 5, 5, 188, 202, 5, 5, 234, 225, 5, 5, 77, 229, 5, 5, 16, 205, 5, 5, 150, 175, 5, 5, 45, 173, 5, 5, 82, 218, 5, 5, 156, 222, 5, 5, 141, 191, 5, 5, 213, 202, 5, 5, 124, 206, 5, 5, 213, 243, 5, 5, 172, 211, 5, 5, 148, 225, 5, 5, 187, 236, 5, 5, 178, 201, 5, 5, 61, 210, 5, 5, 219, 181, 5, 5, 114, 230, 5, 5, 96, 243, 5, 5, 197, 243, 5, 5, 170, 242, 5, 5, 65, 234, 5, 5, 11, 175, 5, 5, 88, 245, 5, 5, 66, 234, 5, 5, 228, 231, 5, 5, 248, 186, 5, 5, 219, 180, 5, 5, 252, 198, 5, 5, 220, 180, 5, 5, 153, 208, 5, 5, 137, 227, 5, 5, 126, 173, 5, 5, 221, 181, 5, 5, 221, 180, 5, 5, 223, 181, 5, 5, 169, 195, 5, 5, 139, 227, 5, 5, 113, 223, 5, 5, 143, 236, 5, 5, 146, 236, 5, 5, 250, 221, 192, 0, 0, 0, 5, 5, 66, 178, 5, 5, 52, 215, 5, 5, 78, 180, 5, 5, 8, 183, 5, 5, 82, 207, 5, 5, 47, 221, 5, 5, 79, 180, 192, 0, 0, 0, 5, 5, 37, 196, 5, 5, 85, 202, 5, 5, 247, 190, 5, 5, 50, 219, 5, 5, 169, 228, 5, 5, 115, 210, 5, 5, 207, 217, 5, 5, 92, 203, 5, 5, 145, 241, 5, 5, 161, 168, 5, 5, 205, 185, 5, 5, 198, 204, 5, 5, 169, 230, 5, 5, 67, 188, 5, 5, 209, 200, 5, 5, 241, 185, 5, 5, 78, 226, 5, 5, 187, 186, 5, 5, 46, 234, 5, 5, 96, 171, 5, 5, 111, 214, 5, 5, 69, 197, 5, 5, 87, 179, 5, 5, 255, 248, 5, 5, 5, 183, 5, 5, 83, 237, 5, 5, 141, 193, 5, 5, 103, 222, 5, 5, 237, 192, 5, 5, 64, 239, 5, 5, 4, 193, 5, 5, 88, 173, 5, 5, 118, 196, 5, 5, 107, 207, 5, 5, 72, 218, 5, 5, 116, 210, 5, 5, 171, 242, 5, 5, 96, 207, 5, 5, 254, 184, 5, 5, 73, 218, 5, 5, 157, 172, 5, 5, 119, 196, 5, 5, 158, 172, 5, 5, 187, 243, 5, 5, 160, 186, 5, 5, 128, 196, 5, 5, 110, 207, 5, 5, 176, 247, 5, 5, 112, 207, 5, 5, 28, 186, 5, 5, 186, 234, 5, 5, 160, 208, 5, 5, 96, 208, 5, 5, 236, 218, 5, 5, 212, 239, 5, 5, 144, 233, 5, 5, 135, 198, 5, 5, 21, 198, 5, 5, 100, 211, 5, 5, 105, 238, 5, 5, 46, 208, 5, 5, 26, 222, 5, 5, 190, 186, 5, 5, 245, 207, 5, 5, 99, 247, 5, 5, 81, 197, 5, 5, 143, 242, 5, 5, 35, 209, 5, 5, 79, 199, 5, 5, 223, 203, 5, 5, 95, 184, 5, 5, 145, 233, 5, 5, 124, 208, 5, 5, 91, 171, 5, 5, 20, 180, 5, 5, 252, 224, 5, 5, 2, 185, 5, 5, 118, 183, 5, 5, 95, 171, 5, 5, 232, 189, 5, 5, 12, 175, 5, 5, 184, 197, 5, 5, 233, 189, 5, 5, 202, 199, 5, 5, 245, 237, 5, 5, 240, 201, 5, 5, 31, 212, 5, 5, 207, 183, 5, 5, 255, 208, 5, 5, 52, 180, 5, 5, 68, 232, 5, 5, 182, 203, 5, 5, 21, 215, 5, 5, 134, 193, 5, 5, 76, 196, 5, 5, 188, 223, 5, 5, 187, 207, 5, 5, 97, 227, 5, 5, 69, 214, 5, 5, 125, 186, 5, 5, 180, 248, 5, 5, 153, 223, 5, 5, 85, 196, 5, 5, 77, 203, 5, 5, 213, 216, 5, 5, 132, 179, 5, 5, 57, 204, 5, 5, 63, 180, 5, 5, 242, 184, 5, 5, 212, 193, 5, 5, 183, 239, 5, 5, 213, 176, 5, 5, 0, 148, 5, 5, 151, 202, 5, 5, 3, 176, 5, 5, 53, 209, 5, 5, 182, 248, 5, 5, 13, 177, 5, 5, 79, 223, 5, 5, 47, 181, 5, 5, 245, 195, 5, 5, 8, 204, 5, 5, 9, 204, 5, 5, 67, 211, 5, 5, 3, 231, 5, 5, 135, 180, 5, 5, 172, 192, 5, 5, 74, 182, 5, 5, 75, 182, 5, 5, 76, 182, 5, 5, 134, 183, 5, 5, 213, 196, 5, 5, 141, 175, 5, 5, 18, 246, 5, 5, 73, 183, 5, 5, 193, 209, 5, 5, 162, 246, 5, 5, 204, 183, 5, 5, 74, 183, 5, 5, 168, 213, 5, 5, 78, 183, 5, 5, 96, 198, 5, 5, 191, 190, 5, 5, 102, 198, 5, 5, 208, 184, 5, 5, 189, 178, 5, 5, 198, 213, 5, 5, 129, 215, 5, 5, 63, 234, 5, 5, 138, 199, 5, 5, 78, 227, 5, 5, 98, 168, 5, 5, 103, 175, 5, 5, 251, 191, 5, 5, 76, 178, 5, 5, 245, 187, 5, 5, 248, 187, 5, 5, 10, 244, 5, 5, 104, 178, 5, 5, 127, 179, 5, 5, 105, 178, 5, 5, 143, 218, 5, 5, 144, 218, 5, 5, 234, 175, 5, 5, 249, 182, 5, 5, 67, 216, 5, 5, 222, 238, 5, 5, 252, 191, 5, 5, 127, 198, 5, 5, 146, 215, 5, 5, 68, 177, 5, 5, 110, 175, 5, 5, 188, 229, 5, 5, 2, 192, 5, 5, 245, 177, 5, 5, 23, 234, 5, 5, 53, 190, 5, 5, 97, 228, 5, 5, 185, 197, 5, 5, 247, 201, 5, 5, 9, 243, 5, 5, 180, 203, 5, 5, 132, 204, 5, 5, 41, 244, 5, 5, 3, 185, 5, 5, 252, 175, 5, 5, 208, 183, 5, 5, 105, 175, 5, 5, 184, 217, 5, 5, 130, 179, 5, 5, 76, 220, 5, 5, 2, 209, 5, 5, 41, 204, 5, 5, 70, 247, 5, 5, 32, 212, 5, 5, 36, 171, 5, 5, 65, 195, 5, 5, 66, 195, 5, 5, 79, 212, 5, 5, 103, 202, 5, 5, 77, 220, 5, 5, 37, 171, 5, 5, 63, 173, 5, 5, 53, 196, 5, 5, 144, 186, 5, 5, 170, 185, 5, 5, 134, 178, 5, 5, 254, 175, 5, 5, 230, 199, 5, 5, 92, 199, 5, 5, 70, 181, 5, 5, 10, 182, 5, 5, 203, 246, 5, 5, 228, 222, 5, 5, 10, 218, 5, 5, 20, 220, 5, 5, 142, 176, 5, 5, 253, 198, 5, 5, 70, 195, 5, 5, 95, 187, 5, 5, 143, 176, 5, 5, 96, 187, 5, 5, 97, 198, 5, 5, 71, 181, 5, 5, 205, 192, 5, 5, 48, 226, 5, 5, 59, 196, 5, 5, 28, 205, 5, 5, 137, 206, 5, 5, 49, 244, 5, 5, 142, 242, 5, 5, 93, 177, 5, 5, 90, 235, 5, 5, 254, 198, 5, 5, 132, 200, 5, 5, 98, 215, 5, 5, 21, 220, 5, 5, 255, 175, 5, 5, 162, 186, 5, 5, 16, 194, 5, 5, 94, 177, 5, 5, 250, 201, 5, 5, 243, 225, 5, 5, 230, 182, 5, 5, 246, 213, 5, 5, 68, 173, 5, 5, 121, 214, 5, 5, 2, 176, 5, 5, 87, 249, 5, 5, 13, 185, 5, 5, 85, 228, 5, 5, 146, 171, 5, 5, 14, 192, 5, 5, 30, 181, 5, 5, 189, 216, 5, 5, 123, 220, 5, 5, 115, 180, 5, 5, 23, 194, 5, 5, 218, 192, 5, 5, 147, 171, 5, 5, 52, 236, 5, 5, 82, 197, 5, 5, 187, 191, 5, 5, 155, 221, 5, 5, 213, 193, 5, 5, 40, 181, 5, 5, 184, 180, 5, 5, 37, 230, 5, 5, 163, 186, 5, 5, 96, 184, 5, 5, 153, 221, 5, 5, 28, 194, 5, 5, 99, 185, 5, 5, 133, 177, 5, 5, 103, 198, 5, 5, 4, 176, 5, 5, 207, 175, 5, 5, 76, 173, 5, 5, 151, 227, 5, 5, 171, 205, 5, 5, 12, 202, 5, 5, 201, 212, 5, 5, 97, 220, 5, 5, 92, 213, 5, 5, 70, 199, 5, 5, 215, 194, 5, 5, 153, 186, 5, 5, 246, 215, 5, 5, 225, 197, 5, 5, 66, 190, 5, 5, 134, 244, 5, 5, 230, 248, 5, 5, 3, 203, 5, 5, 106, 197, 5, 5, 232, 212, 5, 5, 154, 204, 5, 5, 110, 187, 5, 5, 228, 194, 5, 5, 111, 187, 5, 5, 50, 194, 5, 5, 51, 194, 5, 5, 91, 225, 5, 5, 190, 191, 5, 5, 5, 193, 5, 5, 61, 194, 5, 5, 101, 239, 5, 5, 62, 194, 5, 5, 45, 247, 5, 5, 46, 173, 5, 5, 5, 194, 5, 5, 87, 209, 5, 5, 59, 202, 5, 5, 19, 248, 5, 5, 95, 202, 5, 5, 22, 236, 5, 5, 9, 218, 5, 5, 29, 169, 5, 5, 207, 185, 5, 5, 37, 193, 5, 5, 116, 230, 5, 5, 226, 206, 5, 5, 181, 203, 5, 5, 217, 195, 5, 5, 36, 199, 5, 5, 99, 233, 5, 5, 163, 246, 5, 5, 51, 180, 5, 5, 26, 248, 5, 5, 141, 211, 5, 5, 67, 195, 5, 5, 137, 217, 5, 5, 217, 220, 5, 5, 235, 238, 5, 5, 201, 247, 5, 5, 92, 209, 5, 5, 104, 202, 5, 5, 218, 195, 5, 5, 82, 201, 5, 5, 83, 201, 5, 5, 152, 197, 5, 5, 75, 199, 5, 5, 47, 237, 5, 5, 46, 228, 5, 5, 72, 233, 5, 5, 153, 209, 5, 5, 173, 199, 5, 5, 71, 195, 5, 5, 183, 203, 5, 5, 158, 188, 5, 5, 20, 222, 5, 5, 255, 198, 5, 5, 223, 195, 5, 5, 73, 185, 5, 5, 181, 171, 5, 5, 6, 209, 5, 5, 231, 174, 5, 5, 38, 201, 5, 5, 230, 240, 5, 5, 231, 240, 5, 5, 156, 208, 5, 5, 2, 199, 5, 5, 134, 235, 5, 5, 155, 197, 5, 5, 190, 216, 5, 5, 154, 205, 5, 5, 22, 172, 5, 5, 25, 208, 5, 5, 234, 174, 5, 5, 158, 201, 5, 5, 123, 198, 5, 5, 161, 208, 5, 5, 66, 180, 5, 5, 216, 234, 5, 5, 219, 234, 5, 5, 129, 198, 5, 5, 133, 230, 5, 5, 31, 239, 5, 5, 137, 235, 5, 5, 218, 229, 5, 5, 156, 221, 5, 5, 90, 201, 5, 5, 203, 209, 5, 5, 172, 205, 5, 5, 117, 213, 5, 5, 57, 222, 5, 5, 32, 192, 5, 5, 128, 216, 5, 5, 141, 194, 5, 5, 183, 173, 5, 5, 12, 218, 5, 5, 230, 232, 5, 5, 77, 239, 5, 5, 226, 197, 5, 5, 58, 182, 5, 5, 136, 197, 5, 5, 30, 227, 5, 5, 107, 197, 5, 5, 236, 177, 5, 5, 127, 233, 5, 5, 236, 206, 5, 5, 142, 235, 5, 5, 143, 235, 5, 5, 23, 206, 5, 5, 175, 202, 5, 5, 235, 173, 5, 5, 83, 218, 5, 5, 14, 218, 5, 5, 81, 169, 5, 5, 210, 220, 5, 5, 65, 243, 5, 5, 149, 196, 5, 5, 124, 169, 5, 5, 248, 185, 5, 5, 115, 230, 5, 5, 66, 243, 5, 5, 160, 229, 5, 5, 64, 234, 5, 5, 170, 184, 5, 5, 171, 184, 5, 5, 82, 169, 5, 5, 153, 176, 5, 5, 225, 238, 5, 5, 67, 234, 5, 5, 147, 212, 5, 5, 215, 196, 5, 5, 182, 225, 5, 5, 113, 185, 5, 5, 17, 214, 5, 5, 150, 181, 5, 5, 87, 212, 5, 5, 245, 183, 5, 5, 219, 185, 5, 5, 134, 177, 5, 5, 194, 196, 5, 5, 226, 185, 5, 5, 22, 170, 5, 5, 52, 190, 5, 5, 159, 169, 5, 5, 86, 210, 5, 5, 90, 222, 5, 5, 137, 182, 5, 5, 184, 196, 5, 5, 101, 238, 5, 5, 149, 243, 5, 5, 130, 194, 5, 5, 179, 198, 5, 5, 131, 194, 5, 5, 172, 199, 5, 5, 143, 189, 5, 5, 169, 231, 5, 5, 93, 217, 5, 5, 77, 182, 5, 5, 73, 187, 5, 5, 72, 216, 5, 5, 237, 243, 5, 5, 176, 199, 5, 5, 208, 182, 5, 5, 155, 189, 5, 5, 237, 241, 5, 5, 85, 187, 5, 5, 23, 200, 5, 5, 92, 191, 5, 5, 230, 177, 5, 5, 109, 187, 5, 5, 252, 202, 5, 5, 4, 203, 5, 5, 42, 224, 5, 5, 136, 180, 5, 5, 207, 196, 5, 5, 255, 197, 5, 5, 105, 231, 5, 5, 29, 213, 5, 5, 89, 217, 5, 5, 64, 238, 5, 5, 204, 198, 5, 5, 147, 236, 5, 5, 189, 170, 5, 5, 179, 210, 5, 5, 107, 217, 5, 5, 204, 221, 5, 5, 197, 235, 5, 5, 29, 215, 5, 5, 221, 210, 5, 5, 139, 219, 5, 5, 190, 249, 5, 5, 113, 221, 5, 5, 81, 230, 5, 5, 65, 191, 5, 5, 30, 169, 5, 5, 12, 222, 5, 5, 106, 231, 5, 5, 143, 228, 5, 5, 33, 190, 5, 5, 100, 233, 5, 5, 144, 228, 5, 5, 148, 169, 5, 5, 192, 249, 5, 5, 187, 248, 5, 5, 103, 233, 5, 5, 216, 177, 5, 5, 231, 206, 5, 5, 26, 210, 5, 5, 218, 177, 5, 5, 124, 192, 5, 5, 196, 171, 5, 5, 238, 222, 5, 5, 244, 171, 5, 5, 198, 199, 5, 5, 199, 170, 5, 5, 2, 172, 5, 5, 249, 244, 5, 5, 84, 198, 5, 5, 74, 205, 5, 5, 65, 241, 5, 5, 119, 205, 5, 5, 179, 230, 5, 5, 165, 186, 5, 5, 253, 229, 5, 5, 159, 233, 5, 5, 59, 195, 5, 5, 60, 195, 5, 5, 87, 229, 5, 5, 91, 168, 5, 5, 228, 216, 5, 5, 81, 246, 5, 5, 97, 207, 5, 5, 29, 240, 5, 5, 201, 228, 5, 5, 218, 220, 5, 5, 104, 192, 5, 5, 20, 218, 5, 5, 57, 206, 5, 5, 208, 174, 5, 5, 142, 197, 5, 5, 155, 233, 5, 5, 204, 234, 5, 5, 194, 195, 5, 5, 21, 218, 5, 5, 130, 230, 5, 5, 111, 192, 5, 5, 156, 181, 5, 5, 174, 216, 5, 5, 231, 230, 5, 5, 174, 219, 5, 5, 149, 173, 5, 5, 25, 229, 5, 5, 132, 181, 5, 5, 198, 226, 5, 5, 96, 202, 5, 5, 116, 245, 5, 5, 17, 188, 5, 5, 97, 202, 5, 5, 13, 236, 5, 5, 228, 235, 5, 5, 198, 236, 5, 5, 244, 220, 5, 5, 250, 178, 5, 5, 155, 244, 5, 5, 57, 212, 5, 5, 249, 235, 5, 5, 68, 216, 5, 5, 252, 235, 5, 5, 204, 246, 5, 5, 160, 172, 5, 5, 59, 212, 5, 5, 251, 225, 5, 5, 249, 201, 5, 5, 248, 220, 5, 5, 129, 189, 5, 5, 205, 226, 5, 5, 226, 249, 5, 5, 95, 177, 5, 5, 231, 182, 5, 5, 181, 242, 5, 5, 168, 172, 5, 5, 187, 242, 5, 5, 164, 232, 5, 5, 151, 236, 5, 5, 152, 202, 5, 5, 218, 197, 5, 5, 52, 220, 5, 5, 70, 179, 5, 5, 113, 175, 5, 5, 195, 196, 5, 5, 200, 195, 5, 5, 104, 168, 5, 5, 91, 187, 5, 5, 232, 236, 5, 5, 80, 223, 5, 5, 159, 202, 5, 5, 156, 173, 5, 5, 220, 200, 5, 5, 168, 202, 5, 5, 119, 236, 5, 5, 180, 236, 5, 5, 228, 242, 5, 5, 57, 223, 5, 5, 206, 185, 5, 5, 219, 203, 5, 5, 243, 218, 5, 5, 182, 217, 5, 5, 183, 217, 5, 5, 242, 181, 5, 5, 187, 201, 5, 5, 114, 180, 5, 5, 105, 232, 5, 5, 78, 248, 5, 5, 171, 219, 5, 5, 72, 172, 5, 5, 73, 172, 5, 5, 74, 172, 5, 5, 77, 172, 5, 5, 12, 168, 5, 5, 199, 177, 5, 5, 86, 241, 5, 5, 198, 172, 5, 5, 96, 192, 5, 5, 62, 241, 5, 5, 245, 222, 5, 5, 110, 182, 5, 5, 93, 222, 5, 5, 184, 186, 5, 5, 186, 168, 5, 5, 20, 182, 5, 5, 112, 219, 5, 5, 16, 222, 5, 5, 123, 222, 5, 5, 205, 248, 5, 5, 174, 217, 5, 5, 101, 222, 5, 5, 207, 170, 5, 5, 207, 234, 5, 5, 101, 193, 5, 5, 37, 212, 5, 5, 212, 223, 5, 5, 112, 192, 5, 5, 96, 229, 5, 5, 213, 223, 5, 5, 205, 179, 5, 5, 101, 219, 5, 5, 207, 176, 5, 5, 74, 199, 5, 5, 76, 186, 5, 5, 58, 197, 5, 5, 125, 204, 5, 5, 161, 186, 5, 5, 106, 178, 5, 5, 76, 199, 5, 5, 161, 246, 5, 5, 251, 194, 5, 5, 87, 245, 5, 5, 164, 168, 5, 5, 228, 179, 5, 5, 243, 198, 5, 5, 228, 224, 5, 5, 224, 174, 5, 5, 243, 221, 5, 5, 87, 241, 5, 5, 35, 217, 5, 5, 224, 213, 5, 5, 252, 237, 5, 5, 115, 188, 5, 5, 60, 223, 5, 5, 74, 225, 5, 5, 204, 174, 5, 5, 140, 201, 5, 5, 128, 179, 5, 5, 253, 191, 5, 5, 158, 203, 5, 5, 25, 189, 5, 5, 208, 208, 5, 5, 138, 234, 5, 5, 39, 207, 5, 5, 146, 174, 5, 5, 153, 185, 5, 5, 255, 234, 5, 5, 113, 237, 5, 5, 60, 249, 5, 5, 156, 188, 5, 5, 98, 192, 5, 5, 154, 179, 5, 5, 71, 177, 5, 5, 4, 227, 5, 5, 29, 209, 5, 5, 127, 189, 5, 5, 100, 202, 5, 5, 134, 227, 5, 5, 218, 232, 5, 5, 122, 244, 5, 5, 213, 231, 5, 5, 244, 237, 5, 5, 234, 205, 5, 5, 229, 235, 5, 5, 213, 206, 5, 5, 217, 211, 5, 5, 186, 191, 5, 5, 65, 238, 5, 5, 22, 198, 5, 5, 90, 175, 5, 5, 20, 204, 5, 5, 194, 227, 5, 5, 214, 239, 5, 5, 219, 182, 5, 5, 27, 170, 5, 5, 158, 216, 5, 5, 159, 216, 5, 5, 137, 195, 5, 5, 13, 172, 5, 5, 155, 183, 5, 5, 225, 168, 5, 5, 241, 180, 5, 5, 250, 182, 5, 5, 99, 181, 5, 5, 243, 187, 5, 5, 202, 226, 5, 5, 38, 199, 5, 5, 28, 223, 5, 5, 203, 214, 5, 5, 48, 189, 5, 5, 84, 246, 5, 5, 3, 240, 5, 5, 51, 230, 5, 5, 52, 230, 5, 5, 200, 173, 5, 5, 231, 209, 5, 5, 125, 235, 5, 5, 171, 230, 5, 5, 6, 176, 5, 5, 75, 180, 5, 5, 189, 229, 5, 5, 125, 189, 5, 5, 26, 189, 5, 5, 53, 230, 5, 5, 71, 185, 5, 5, 50, 236, 5, 5, 47, 198, 5, 5, 235, 205, 5, 5, 138, 181, 5, 5, 111, 185, 5, 5, 140, 207, 5, 5, 163, 177, 5, 5, 204, 214, 5, 5, 89, 174, 5, 5, 54, 230, 5, 5, 72, 185, 5, 5, 165, 183, 5, 5, 252, 194, 5, 5, 27, 189, 5, 5, 205, 174, 5, 5, 116, 221, 5, 5, 232, 209, 5, 5, 195, 227, 5, 5, 40, 184, 5, 5, 236, 238, 5, 5, 164, 177, 5, 5, 218, 211, 5, 5, 105, 202, 5, 5, 206, 169, 5, 5, 174, 242, 5, 5, 147, 187, 5, 5, 210, 229, 5, 5, 176, 215, 5, 5, 29, 230, 5, 5, 142, 181, 5, 5, 169, 221, 5, 5, 143, 197, 5, 5, 144, 212, 5, 5, 192, 229, 5, 5, 107, 210, 5, 5, 171, 206, 5, 5, 126, 204, 5, 5, 59, 218, 5, 5, 8, 241, 5, 5, 255, 178, 5, 5, 131, 247, 5, 5, 21, 222, 5, 5, 172, 247, 5, 5, 195, 226, 5, 5, 111, 231, 5, 5, 246, 238, 5, 5, 232, 214, 5, 5, 169, 213, 5, 5, 67, 249, 5, 5, 45, 186, 5, 5, 93, 176, 5, 5, 91, 229, 5, 5, 183, 234, 5, 5, 148, 188, 5, 5, 64, 210, 5, 5, 137, 184, 5, 5, 103, 212, 5, 5, 247, 238, 5, 5, 241, 232, 5, 5, 25, 221, 5, 5, 145, 189, 5, 5, 56, 209, 5, 5, 124, 177, 5, 5, 185, 217, 5, 5, 36, 197, 5, 5, 246, 187, 5, 5, 152, 243, 5, 5, 229, 227, 5, 5, 27, 181, 5, 5, 255, 213, 5, 5, 80, 212, 5, 5, 38, 171, 5, 5, 213, 183, 5, 5, 25, 237, 5, 5, 159, 188, 5, 5, 173, 243, 5, 5, 160, 188, 5, 5, 205, 187, 5, 5, 187, 196, 5, 5, 226, 240, 5, 5, 80, 184, 5, 5, 110, 177, 5, 5, 173, 247, 5, 5, 65, 228, 5, 5, 79, 198, 5, 5, 46, 186, 5, 5, 80, 198, 5, 5, 51, 250, 5, 5, 15, 172, 5, 5, 215, 204, 5, 5, 21, 180, 5, 5, 54, 211, 5, 5, 224, 224, 5, 5, 63, 223, 5, 5, 106, 232, 5, 5, 210, 191, 5, 5, 236, 214, 5, 5, 11, 182, 5, 5, 145, 181, 5, 5, 254, 186, 5, 5, 169, 244, 5, 5, 117, 231, 5, 5, 117, 238, 5, 5, 178, 203, 5, 5, 2, 249, 5, 5, 209, 208, 5, 5, 43, 208, 5, 5, 178, 246, 5, 5, 183, 237, 5, 5, 11, 192, 5, 5, 177, 247, 5, 5, 93, 185, 5, 5, 197, 222, 5, 5, 203, 243, 5, 5, 73, 233, 5, 5, 206, 187, 5, 5, 4, 191, 5, 5, 128, 199, 5, 5, 92, 190, 5, 5, 184, 225, 5, 5, 23, 232, 5, 5, 146, 181, 5, 5, 56, 235, 5, 5, 90, 234, 5, 5, 148, 187, 5, 5, 202, 236, 5, 5, 109, 201, 5, 5, 69, 238, 5, 5, 0, 133, 5, 5, 151, 213, 5, 5, 75, 221, 5, 5, 9, 227, 5, 5, 30, 189, 5, 5, 71, 234, 5, 5, 28, 181, 5, 5, 76, 228, 5, 5, 201, 187, 5, 5, 177, 243, 5, 5, 94, 241, 5, 5, 224, 179, 5, 5, 2, 212, 5, 5, 201, 232, 5, 5, 0, 132, 5, 5, 163, 185, 5, 5, 177, 199, 5, 5, 15, 236, 5, 5, 128, 177, 5, 5, 244, 232, 5, 5, 60, 170, 5, 5, 72, 191, 5, 5, 197, 210, 5, 5, 22, 190, 5, 5, 61, 234, 5, 5, 144, 199, 5, 5, 53, 181, 5, 5, 2, 183, 5, 5, 207, 192, 5, 5, 102, 211, 5, 5, 150, 209, 5, 5, 184, 240, 5, 5, 119, 188, 5, 5, 182, 242, 5, 5, 23, 205, 5, 5, 2, 214, 5, 5, 51, 207, 5, 5, 95, 185, 5, 5, 195, 211, 5, 5, 149, 174, 5, 5, 221, 220, 5, 5, 246, 201, 5, 5, 228, 209, 5, 5, 193, 249, 5, 5, 173, 188, 5, 5, 96, 199, 5, 5, 247, 232, 5, 5, 117, 232, 5, 5, 86, 201, 5, 5, 148, 171, 5, 5, 123, 245, 5, 5, 128, 244, 5, 5, 103, 203, 5, 5, 174, 168, 5, 5, 210, 208, 5, 5, 189, 203, 5, 5, 52, 224, 5, 5, 239, 183, 5, 5, 4, 172, 5, 5, 232, 187, 5, 5, 4, 189, 5, 5, 188, 185, 5, 5, 38, 223, 5, 5, 149, 185, 5, 5, 104, 241, 5, 5, 207, 236, 5, 5, 54, 186, 5, 5, 84, 186, 5, 5, 207, 169, 5, 5, 250, 187, 5, 5, 132, 224, 5, 5, 44, 176, 5, 5, 18, 239, 5, 5, 0, 134, 5, 5, 81, 193, 5, 5, 84, 227, 5, 5, 24, 232, 5, 5, 116, 228, 5, 5, 128, 202, 5, 5, 191, 230, 5, 5, 103, 225, 5, 5, 42, 250, 5, 5, 52, 217, 5, 5, 216, 173, 5, 5, 57, 230, 5, 5, 25, 244, 5, 5, 16, 236, 5, 5, 76, 180, 5, 5, 208, 214, 5, 5, 221, 178, 5, 5, 162, 216, 5, 5, 201, 206, 5, 5, 105, 209, 5, 5, 244, 185, 5, 5, 88, 180, 5, 5, 186, 217, 5, 5, 87, 201, 5, 5, 93, 203, 5, 5, 152, 224, 5, 5, 26, 244, 5, 5, 152, 190, 5, 5, 78, 201, 5, 5, 23, 220, 5, 5, 15, 192, 5, 5, 210, 249, 5, 5, 236, 229, 5, 5, 132, 183, 5, 5, 231, 195, 5, 5, 219, 189, 5, 5, 242, 214, 5, 5, 104, 222, 5, 5, 248, 228, 5, 5, 229, 222, 5, 5, 165, 173, 5, 5, 230, 181, 5, 5, 141, 202, 5, 5, 236, 215, 5, 5, 65, 168, 5, 5, 50, 244, 5, 5, 185, 240, 5, 5, 225, 210, 5, 5, 122, 241, 5, 5, 115, 226, 5, 5, 189, 200, 5, 5, 47, 220, 5, 5, 201, 230, 5, 5, 36, 228, 5, 5, 156, 189, 5, 5, 0, 145, 5, 5, 76, 245, 5, 5, 127, 211, 5, 5, 26, 199, 5, 5, 215, 248, 5, 5, 216, 248, 5, 5, 163, 220, 5, 5, 31, 179, 5, 5, 7, 189, 5, 5, 224, 203, 5, 5, 0, 131, 5, 5, 58, 172, 5, 5, 165, 232, 5, 5, 198, 227, 5, 5, 95, 230, 5, 5, 204, 229, 5, 5, 38, 177, 5, 5, 48, 220, 5, 5, 90, 186, 5, 5, 211, 214, 5, 5, 212, 214, 5, 5, 189, 225, 5, 5, 17, 178, 5, 5, 18, 178, 5, 5, 13, 238, 5, 5, 75, 249, 5, 5, 38, 170, 5, 5, 39, 177, 5, 5, 199, 175, 5, 5, 179, 188, 5, 5, 17, 236, 5, 5, 214, 214, 5, 5, 127, 245, 5, 5, 181, 182, 5, 5, 94, 203, 5, 5, 35, 232, 5, 5, 5, 213, 5, 5, 49, 220, 5, 5, 159, 200, 5, 5, 51, 244, 5, 5, 54, 240, 5, 5, 169, 186, 5, 5, 232, 211, 5, 5, 153, 245, 5, 5, 237, 219, 5, 5, 101, 248, 5, 5, 4, 211, 5, 5, 151, 187, 5, 5, 105, 206, 5, 5, 101, 236, 5, 5, 205, 178, 5, 5, 218, 217, 5, 5, 70, 173, 5, 5, 154, 171, 5, 5, 255, 179, 5, 5, 37, 201, 5, 5, 79, 233, 5, 5, 7, 197, 5, 5, 106, 225, 5, 5, 240, 174, 5, 5, 255, 225, 5, 5, 60, 168, 5, 5, 160, 196, 5, 5, 24, 178, 5, 5, 81, 198, 5, 5, 219, 240, 5, 5, 101, 229, 5, 5, 32, 210, 5, 5, 130, 220, 5, 5, 44, 242, 5, 5, 132, 245, 5, 5, 137, 200, 5, 5, 40, 195, 5, 5, 111, 189, 5, 5, 20, 188, 5, 5, 193, 179, 5, 5, 145, 247, 5, 5, 251, 172, 5, 5, 78, 228, 5, 5, 175, 211, 5, 5, 32, 242, 5, 5, 189, 239, 5, 5, 164, 243, 5, 5, 119, 237, 5, 5, 193, 211, 5, 5, 163, 208, 5, 5, 181, 189, 5, 5, 104, 243, 5, 5, 223, 175, 5, 5, 85, 191, 5, 5, 161, 190, 5, 5, 162, 190, 5, 5, 80, 231, 5, 5, 151, 188, 5, 5, 23, 192, 5, 5, 25, 200, 5, 5, 100, 247, 5, 5, 253, 228, 5, 5, 53, 220, 5, 5, 224, 234, 5, 5, 194, 190, 5, 5, 43, 181, 5, 5, 8, 211, 5, 5, 9, 235, 5, 5, 107, 203, 5, 5, 45, 242, 5, 5, 209, 219, 5, 5, 158, 174, 5, 5, 19, 216, 5, 5, 218, 236, 5, 5, 226, 177, 5, 5, 146, 212, 5, 5, 78, 172, 5, 5, 8, 202, 5, 5, 186, 240, 5, 5, 130, 244, 5, 5, 209, 228, 5, 5, 181, 208, 5, 5, 116, 240, 5, 5, 137, 212, 5, 5, 6, 172, 5, 5, 249, 199, 5, 5, 52, 231, 5, 5, 46, 242, 5, 5, 41, 195, 5, 5, 32, 205, 5, 5, 115, 199, 5, 5, 29, 244, 5, 5, 252, 189, 5, 5, 10, 226, 5, 5, 142, 237, 5, 5, 225, 188, 5, 5, 0, 130, 5, 5, 126, 234, 5, 5, 191, 215, 5, 5, 243, 219, 5, 5, 193, 240, 5, 5, 3, 224, 5, 5, 96, 189, 5, 5, 120, 233, 5, 5, 7, 168, 5, 5, 139, 224, 5, 5, 214, 206, 5, 5, 205, 172, 5, 5, 224, 187, 5, 5, 221, 198, 5, 5, 114, 177, 5, 5, 201, 219, 5, 5, 242, 173, 5, 5, 54, 219, 5, 5, 191, 223, 5, 5, 53, 228, 5, 5, 33, 192, 5, 5, 69, 212, 5, 5, 40, 230, 5, 5, 134, 215, 5, 5, 58, 222, 5, 5, 127, 185, 5, 5, 17, 249, 5, 5, 42, 195, 5, 5, 126, 201, 5, 5, 212, 229, 5, 5, 69, 228, 5, 5, 135, 223, 5, 5, 160, 174, 5, 5, 97, 188, 5, 5, 172, 224, 192, 0, 0, 0, 5, 5, 207, 187, 5, 5, 158, 224, 5, 5, 129, 216, 5, 5, 11, 211, 5, 5, 153, 188, 5, 5, 141, 246, 5, 5, 164, 219, 5, 5, 220, 209, 5, 5, 151, 185, 5, 5, 224, 209, 5, 5, 177, 179, 5, 5, 0, 140, 5, 5, 186, 215, 5, 5, 254, 226, 5, 5, 95, 170, 5, 5, 106, 168, 5, 5, 107, 168, 5, 5, 246, 202, 5, 5, 228, 249, 5, 5, 113, 245, 5, 5, 119, 209, 5, 5, 223, 223, 5, 5, 215, 223, 5, 5, 67, 225, 5, 5, 232, 178, 5, 5, 88, 214, 5, 5, 11, 195, 5, 5, 20, 175, 5, 5, 171, 194, 5, 5, 104, 198, 5, 5, 81, 225, 5, 5, 99, 220, 5, 5, 139, 172, 5, 5, 67, 193, 5, 5, 0, 150, 5, 5, 83, 233, 5, 5, 118, 213, 5, 5, 37, 205, 5, 5, 138, 184, 5, 5, 106, 186, 5, 5, 17, 233, 5, 5, 161, 189, 5, 5, 101, 191, 5, 5, 158, 187, 5, 5, 221, 211, 5, 5, 249, 231, 5, 5, 58, 244, 5, 5, 138, 173, 5, 5, 162, 234, 5, 5, 218, 213, 5, 5, 157, 178, 5, 5, 215, 206, 5, 5, 205, 206, 5, 5, 192, 189, 5, 5, 231, 201, 5, 5, 117, 180, 5, 5, 139, 184, 5, 5, 89, 225, 5, 5, 248, 237, 5, 5, 242, 169, 5, 5, 66, 240, 5, 5, 159, 219, 5, 5, 12, 195, 5, 5, 116, 208, 5, 5, 84, 233, 5, 5, 27, 190, 5, 5, 223, 206, 5, 5, 68, 218, 5, 5, 200, 175, 5, 5, 158, 212, 5, 5, 94, 201, 5, 5, 86, 233, 5, 5, 43, 192, 5, 5, 15, 248, 5, 5, 192, 173, 5, 5, 34, 200, 5, 5, 233, 249, 5, 5, 22, 233, 5, 5, 87, 223, 5, 5, 103, 188, 5, 5, 66, 184, 5, 5, 128, 203, 5, 5, 23, 216, 5, 5, 248, 230, 5, 5, 183, 175, 5, 5, 80, 173, 5, 5, 36, 178, 5, 5, 247, 188, 5, 5, 220, 235, 5, 5, 129, 181, 5, 5, 16, 250, 5, 5, 59, 182, 5, 5, 165, 174, 5, 5, 29, 191, 5, 5, 208, 243, 5, 5, 245, 175, 5, 5, 234, 176, 5, 5, 33, 178, 5, 5, 85, 242, 5, 5, 201, 227, 5, 5, 190, 172, 5, 5, 16, 195, 5, 5, 230, 237, 5, 5, 249, 230, 5, 5, 254, 214, 5, 5, 100, 188, 5, 5, 5, 203, 5, 5, 166, 234, 5, 5, 176, 178, 5, 5, 31, 191, 5, 5, 243, 239, 5, 5, 11, 214, 5, 5, 176, 197, 5, 5, 134, 216, 5, 5, 186, 235, 5, 5, 14, 211, 5, 5, 71, 205, 5, 5, 88, 223, 5, 5, 175, 236, 5, 5, 181, 240, 5, 5, 117, 177, 5, 5, 213, 244, 5, 5, 194, 211, 5, 5, 193, 247, 5, 5, 250, 195, 5, 5, 109, 211, 5, 5, 113, 191, 5, 5, 183, 233, 5, 5, 2, 215, 5, 5, 191, 181, 5, 5, 33, 244, 5, 5, 87, 238, 5, 5, 77, 222, 5, 5, 19, 195, 5, 5, 251, 242, 5, 5, 0, 143, 5, 5, 202, 240, 5, 5, 240, 197, 5, 5, 153, 199, 5, 5, 255, 241, 5, 5, 138, 212, 5, 5, 146, 178, 5, 5, 142, 184, 5, 5, 24, 189, 5, 5, 249, 180, 5, 5, 57, 178, 5, 5, 233, 233, 5, 5, 160, 219, 5, 5, 236, 212, 5, 5, 41, 213, 5, 5, 208, 239, 5, 5, 235, 249, 5, 5, 68, 211, 5, 5, 241, 178, 5, 5, 29, 201, 5, 5, 185, 224, 5, 5, 191, 191, 5, 5, 30, 219, 5, 5, 89, 188, 5, 5, 219, 231, 5, 5, 41, 238, 5, 5, 37, 181, 5, 5, 56, 213, 5, 5, 75, 175, 5, 5, 10, 193, 5, 5, 254, 195, 5, 5, 107, 188, 5, 5, 70, 226, 5, 5, 145, 173, 5, 5, 146, 235, 5, 5, 134, 207, 5, 5, 40, 172, 5, 5, 72, 226, 5, 5, 136, 205, 5, 5, 132, 191, 5, 5, 222, 171, 5, 5, 19, 241, 5, 5, 27, 211, 5, 5, 247, 239, 5, 5, 12, 190, 5, 5, 138, 207, 5, 5, 41, 189, 5, 5, 144, 245, 5, 5, 40, 202, 5, 5, 168, 204, 5, 5, 229, 187, 5, 5, 49, 210, 5, 5, 38, 233, 5, 5, 79, 209, 5, 5, 251, 236, 5, 5, 196, 202, 5, 5, 93, 205, 5, 5, 238, 204, 5, 5, 82, 209, 5, 5, 46, 178, 5, 5, 47, 174, 5, 5, 146, 213, 5, 5, 43, 213, 5, 5, 235, 232, 5, 5, 235, 191, 5, 5, 83, 209, 5, 5, 153, 231, 5, 5, 68, 181, 5, 5, 122, 199, 5, 5, 125, 236, 5, 5, 49, 173, 5, 5, 90, 240, 5, 5, 93, 218, 5, 5, 100, 179, 5, 5, 136, 200, 5, 5, 219, 224, 5, 5, 44, 233, 5, 5, 2, 198, 5, 5, 90, 176, 5, 5, 102, 190, 5, 5, 241, 191, 5, 5, 116, 248, 5, 5, 36, 211, 5, 5, 45, 233, 5, 5, 46, 172, 5, 5, 52, 202, 5, 5, 94, 173, 5, 5, 6, 173, 5, 5, 219, 202, 5, 5, 155, 239, 5, 5, 107, 206, 5, 5, 51, 210, 5, 5, 170, 243, 5, 5, 235, 223, 5, 5, 101, 231, 5, 5, 101, 244, 5, 5, 198, 193, 5, 5, 171, 243, 5, 5, 20, 248, 5, 5, 247, 200, 5, 5, 43, 211, 5, 5, 47, 210, 5, 5, 21, 201, 5, 5, 212, 204, 5, 5, 234, 228, 5, 5, 40, 191, 5, 5, 167, 239, 5, 5, 36, 217, 5, 5, 109, 223, 5, 5, 49, 211, 5, 5, 202, 193, 5, 5, 41, 191, 5, 5, 245, 233, 5, 5, 168, 239, 5, 5, 21, 210, 5, 5, 148, 227, 5, 5, 149, 227, 5, 5, 7, 181, 5, 5, 181, 198, 5, 5, 140, 242, 5, 5, 129, 196, 5, 5, 63, 213, 5, 5, 68, 243, 5, 5, 154, 176, 5, 5, 170, 189, 5, 5, 42, 191, 5, 5, 175, 242, 5, 5, 100, 181, 5, 5, 165, 187, 5, 5, 97, 200, 5, 5, 155, 176, 5, 5, 250, 226, 5, 5, 95, 227, 5, 5, 237, 228, 5, 5, 74, 206, 5, 5, 166, 187, 5, 5, 49, 218, 5, 5, 188, 218, 5, 5, 43, 204, 5, 5, 123, 186, 5, 5, 167, 187, 5, 5, 225, 224, 5, 5, 168, 187, 5, 5, 96, 227, 5, 5, 95, 241, 5, 5, 169, 187, 5, 5, 221, 239, 5, 5, 178, 191, 5, 5, 38, 214, 5, 5, 233, 241, 5, 5, 251, 187, 5, 5, 183, 242, 5, 5, 81, 206, 5, 5, 202, 217, 5, 5, 238, 241, 5, 5, 176, 216, 5, 5, 170, 187, 5, 5, 214, 175, 5, 5, 254, 228, 5, 5, 189, 242, 5, 5, 203, 217, 5, 5, 100, 199, 5, 5, 42, 214, 5, 5, 197, 242, 5, 5, 198, 242, 5, 5, 45, 236, 5, 5, 111, 227, 5, 5, 119, 227, 5, 5, 120, 227, 5, 5, 152, 227, 5, 5, 66, 206, 5, 5, 105, 191, 5, 5, 102, 239, 5, 5, 126, 190, 5, 5, 39, 206, 5, 5, 50, 206, 5, 5, 132, 227, 5, 5, 23, 236, 5, 5, 133, 227, 5, 5, 236, 226, 5, 5, 150, 221, 5, 5, 16, 214, 5, 5, 146, 205, 5, 5, 139, 199, 5, 5, 227, 235, 5, 5, 188, 243, 5, 5, 236, 228, 5, 5, 86, 185, 5, 5, 12, 242, 5, 5, 23, 230, 5, 5, 25, 187, 5, 5, 33, 213, 5, 5, 102, 238, 5, 5, 168, 178, 5, 5, 32, 215, 5, 5, 33, 215, 5, 5, 244, 245, 5, 5, 200, 248, 5, 5, 77, 178, 5, 5, 20, 215, 5, 5, 214, 231, 5, 5, 75, 220, 5, 5, 200, 199, 5, 5, 150, 173, 5, 5, 105, 214, 5, 5, 251, 210, 5, 5, 93, 209, 5, 5, 3, 192, 5, 5, 75, 193, 5, 5, 166, 246, 5, 5, 167, 246, 5, 5, 16, 169, 5, 5, 133, 235, 5, 5, 237, 238, 5, 5, 147, 216, 5, 5, 141, 207, 5, 5, 23, 198, 5, 5, 238, 218, 5, 5, 205, 227, 5, 5, 140, 182, 5, 5, 221, 169, 5, 5, 222, 169, 5, 5, 13, 225, 5, 5, 136, 198, 5, 5, 91, 190, 5, 5, 34, 250, 5, 5, 39, 199, 5, 5, 49, 170, 5, 5, 54, 196, 5, 5, 17, 179, 5, 5, 10, 196, 5, 5, 199, 192, 5, 5, 140, 199, 5, 5, 251, 178, 5, 5, 11, 196, 5, 5, 131, 193, 5, 5, 26, 225, 5, 5, 106, 202, 5, 5, 208, 168, 5, 5, 122, 230, 5, 5, 18, 183, 5, 5, 150, 248, 5, 5, 208, 213, 5, 5, 33, 169, 5, 5, 102, 225, 5, 5, 51, 200, 5, 5, 98, 217, 5, 5, 61, 225, 5, 5, 87, 246, 5, 5, 248, 227, 5, 5, 196, 184, 5, 5, 170, 213, 5, 5, 104, 179, 5, 5, 166, 186, 5, 5, 125, 210, 5, 5, 231, 224, 5, 5, 212, 212, 5, 5, 120, 196, 5, 5, 81, 237, 5, 5, 150, 183, 5, 5, 141, 168, 5, 5, 150, 205, 5, 5, 10, 217, 5, 5, 175, 209, 5, 5, 244, 198, 5, 5, 21, 186, 5, 5, 126, 235, 5, 5, 189, 168, 5, 5, 181, 174, 5, 5, 225, 173, 5, 5, 44, 204, 5, 5, 29, 248, 5, 5, 81, 184, 5, 5, 146, 189, 5, 5, 205, 246, 5, 5, 9, 176, 5, 5, 124, 200, 5, 5, 12, 205, 5, 5, 13, 205, 5, 5, 77, 205, 5, 5, 142, 168, 5, 5, 173, 177, 5, 5, 81, 212, 5, 5, 28, 209, 5, 5, 30, 234, 5, 5, 54, 180, 5, 5, 208, 192, 5, 5, 219, 188, 5, 5, 9, 206, 5, 5, 105, 176, 5, 5, 210, 174, 5, 5, 190, 201, 5, 5, 152, 184, 5, 5, 172, 239, 5, 5, 130, 189, 5, 5, 208, 180, 5, 5, 91, 245, 5, 5, 226, 183, 5, 5, 255, 186, 5, 5, 129, 237, 5, 5, 54, 181, 5, 5, 55, 181, 5, 5, 76, 187, 5, 5, 216, 172, 5, 5, 51, 237, 5, 5, 219, 239, 5, 5, 29, 182, 5, 5, 25, 186, 5, 5, 178, 242, 5, 5, 181, 179, 5, 5, 105, 233, 5, 5, 24, 199, 5, 5, 177, 220, 5, 5, 97, 222, 5, 5, 147, 181, 5, 5, 128, 171, 5, 5, 106, 179, 5, 5, 58, 189, 5, 5, 233, 235, 5, 5, 129, 199, 5, 5, 159, 177, 5, 5, 86, 198, 5, 5, 80, 178, 5, 5, 98, 198, 5, 5, 74, 188, 5, 5, 82, 210, 5, 5, 63, 168, 5, 5, 5, 234, 5, 5, 118, 232, 5, 5, 240, 242, 5, 5, 48, 169, 5, 5, 175, 183, 5, 5, 211, 168, 5, 5, 19, 239, 5, 5, 33, 240, 5, 5, 31, 188, 5, 5, 211, 234, 5, 5, 11, 176, 5, 5, 114, 216, 5, 5, 189, 185, 5, 5, 0, 135, 5, 5, 245, 169, 5, 5, 155, 182, 5, 5, 22, 218, 5, 5, 233, 240, 5, 5, 51, 198, 5, 5, 52, 193, 5, 5, 26, 179, 5, 5, 217, 206, 5, 5, 72, 201, 5, 5, 145, 197, 5, 5, 95, 174, 5, 5, 85, 220, 5, 5, 178, 195, 5, 5, 124, 245, 5, 5, 190, 203, 5, 5, 191, 203, 5, 5, 23, 172, 5, 5, 102, 174, 5, 5, 23, 190, 5, 5, 24, 172, 5, 5, 223, 221, 5, 5, 135, 235, 5, 5, 149, 187, 5, 5, 122, 196, 5, 5, 246, 237, 5, 5, 226, 210, 5, 5, 206, 178, 5, 5, 33, 242, 5, 5, 29, 172, 5, 5, 237, 235, 5, 5, 219, 217, 5, 5, 80, 224, 5, 5, 6, 213, 5, 5, 177, 216, 5, 5, 119, 228, 5, 5, 88, 197, 5, 5, 97, 206, 5, 5, 26, 246, 5, 5, 54, 199, 5, 5, 40, 175, 5, 5, 29, 180, 5, 5, 189, 177, 5, 5, 76, 225, 5, 5, 66, 168, 5, 5, 67, 172, 5, 5, 169, 175, 5, 5, 246, 169, 5, 5, 141, 198, 5, 5, 131, 246, 5, 5, 58, 181, 5, 5, 32, 239, 5, 5, 132, 246, 5, 5, 33, 239, 5, 5, 112, 212, 5, 5, 21, 192, 5, 5, 181, 248, 5, 5, 130, 214, 5, 5, 227, 219, 5, 5, 89, 197, 5, 5, 134, 210, 5, 5, 98, 199, 5, 5, 8, 199, 5, 5, 104, 225, 5, 5, 53, 200, 5, 5, 181, 210, 5, 5, 143, 193, 5, 5, 209, 180, 5, 5, 197, 195, 5, 5, 15, 185, 5, 5, 34, 242, 5, 5, 157, 181, 5, 5, 167, 212, 5, 5, 127, 186, 5, 5, 141, 227, 5, 5, 242, 201, 5, 5, 176, 182, 5, 5, 3, 236, 5, 5, 152, 215, 5, 5, 54, 200, 5, 5, 80, 168, 5, 5, 43, 221, 5, 5, 79, 181, 5, 5, 88, 210, 5, 5, 87, 227, 5, 5, 113, 174, 5, 5, 190, 239, 5, 5, 167, 191, 5, 5, 82, 170, 5, 5, 56, 203, 5, 5, 152, 187, 5, 5, 194, 179, 5, 5, 102, 248, 5, 5, 135, 189, 5, 5, 107, 169, 5, 5, 108, 169, 5, 5, 163, 241, 5, 5, 209, 178, 5, 5, 86, 207, 5, 5, 43, 195, 5, 5, 64, 219, 5, 5, 16, 238, 5, 5, 201, 185, 5, 5, 131, 198, 5, 5, 101, 249, 5, 5, 164, 241, 5, 5, 195, 190, 5, 5, 164, 181, 5, 5, 138, 237, 5, 5, 219, 236, 5, 5, 133, 169, 5, 5, 110, 176, 5, 5, 156, 207, 5, 5, 153, 173, 5, 5, 158, 180, 5, 5, 6, 228, 5, 5, 34, 240, 5, 5, 123, 183, 5, 5, 122, 247, 5, 5, 155, 195, 5, 5, 183, 195, 5, 5, 23, 189, 5, 5, 20, 185, 5, 5, 29, 176, 5, 5, 218, 193, 5, 5, 188, 213, 5, 5, 191, 201, 5, 5, 224, 232, 5, 5, 187, 190, 5, 5, 233, 201, 5, 5, 192, 180, 5, 5, 88, 228, 5, 5, 10, 235, 5, 5, 232, 192, 5, 5, 134, 192, 5, 5, 147, 199, 5, 5, 125, 240, 5, 5, 181, 224, 5, 5, 121, 174, 5, 5, 244, 240, 5, 5, 107, 198, 5, 5, 4, 224, 5, 5, 5, 224, 5, 5, 229, 221, 5, 5, 113, 208, 5, 5, 188, 224, 5, 5, 220, 229, 5, 5, 122, 174, 5, 5, 112, 227, 5, 5, 109, 225, 5, 5, 30, 218, 5, 5, 102, 247, 5, 5, 160, 202, 5, 5, 103, 247, 5, 5, 71, 169, 5, 5, 156, 219, 5, 5, 253, 243, 5, 5, 210, 180, 5, 5, 98, 226, 5, 5, 147, 230, 5, 5, 74, 246, 5, 5, 138, 235, 5, 5, 128, 185, 5, 5, 5, 246, 5, 5, 8, 168, 5, 5, 220, 185, 5, 5, 112, 236, 5, 5, 144, 198, 5, 5, 99, 226, 5, 5, 199, 242, 5, 5, 141, 229, 5, 5, 179, 233, 5, 5, 201, 204, 5, 5, 216, 187, 5, 5, 67, 201, 5, 5, 139, 173, 5, 5, 172, 212, 5, 5, 255, 169, 5, 5, 18, 174, 5, 5, 179, 205, 5, 5, 126, 205, 5, 5, 203, 211, 5, 5, 160, 215, 5, 5, 170, 207, 5, 5, 120, 209, 5, 5, 83, 248, 5, 5, 6, 223, 5, 5, 164, 222, 5, 5, 57, 205, 5, 5, 197, 174, 5, 5, 21, 207, 5, 5, 240, 170, 5, 5, 104, 196, 5, 5, 180, 172, 5, 5, 216, 222, 5, 5, 19, 247, 5, 5, 60, 245, 5, 5, 155, 198, 5, 5, 196, 240, 5, 5, 118, 179, 5, 5, 31, 174, 5, 5, 144, 246, 5, 5, 137, 231, 5, 5, 156, 187, 5, 5, 230, 215, 5, 5, 242, 195, 5, 5, 52, 179, 5, 5, 166, 220, 5, 5, 205, 209, 5, 5, 15, 177, 5, 5, 233, 236, 5, 5, 190, 224, 5, 5, 102, 244, 5, 5, 112, 215, 5, 5, 212, 215, 5, 5, 84, 203, 5, 5, 107, 229, 5, 5, 161, 248, 5, 5, 251, 215, 5, 5, 104, 244, 5, 5, 163, 234, 5, 5, 136, 220, 5, 5, 137, 220, 5, 5, 202, 168, 5, 5, 27, 214, 5, 5, 151, 199, 5, 5, 45, 180, 5, 5, 91, 182, 5, 5, 37, 218, 5, 5, 136, 209, 5, 5, 245, 180, 5, 5, 246, 180, 5, 5, 3, 250, 5, 5, 53, 179, 5, 5, 161, 221, 5, 5, 86, 181, 5, 5, 89, 181, 5, 5, 38, 225, 5, 5, 192, 178, 5, 5, 171, 209, 5, 5, 35, 183, 5, 5, 207, 190, 5, 5, 39, 225, 5, 5, 161, 177, 5, 5, 21, 238, 5, 5, 60, 248, 5, 5, 52, 194, 5, 5, 152, 168, 5, 5, 218, 215, 5, 5, 54, 192, 5, 5, 254, 215, 5, 5, 29, 199, 5, 5, 103, 239, 5, 5, 20, 213, 5, 5, 130, 170, 5, 5, 119, 179, 5, 5, 103, 194, 5, 5, 250, 237, 5, 5, 203, 240, 5, 5, 109, 235, 5, 5, 44, 225, 5, 5, 10, 201, 5, 5, 111, 197, 5, 5, 93, 190, 5, 5, 96, 178, 5, 5, 92, 218, 5, 5, 169, 215, 5, 5, 147, 235, 5, 5, 150, 232, 5, 5, 95, 231, 5, 5, 234, 188, 5, 5, 17, 168, 5, 5, 244, 235, 5, 5, 129, 178, 5, 5, 108, 188, 5, 5, 66, 219, 5, 5, 255, 195, 5, 5, 202, 201, 5, 5, 219, 199, 5, 5, 94, 205, 5, 5, 126, 236, 5, 5, 54, 225, 5, 5, 81, 229, 5, 5, 94, 190, 5, 5, 18, 205, 5, 5, 19, 205, 5, 5, 102, 219, 5, 5, 197, 202, 5, 5, 253, 203, 5, 5, 94, 218, 5, 5, 50, 173, 5, 5, 155, 235, 5, 5, 131, 236, 5, 5, 179, 201, 5, 5, 222, 168, 5, 5, 94, 228, 5, 5, 3, 222, 5, 5, 123, 218, 5, 5, 196, 219, 5, 5, 133, 248, 5, 5, 134, 248, 5, 5, 117, 221, 5, 5, 66, 238, 5, 5, 229, 206, 5, 5, 239, 198, 5, 5, 35, 248, 5, 5, 137, 248, 5, 5, 176, 189, 5, 5, 178, 189, 5, 5, 86, 200, 5, 5, 77, 238, 5, 5, 182, 189, 5, 5, 225, 234, 5, 5, 92, 200, 5, 5, 110, 222, 5, 5, 68, 207, 5, 5, 4, 250, 5, 5, 111, 222, 5, 5, 88, 238, 5, 5, 159, 246, 5, 5, 75, 186, 5, 5, 157, 175, 5, 5, 132, 194, 5, 5, 119, 183, 5, 5, 166, 169, 5, 5, 182, 244, 5, 5, 211, 170, 5, 5, 48, 224, 5, 5, 48, 218, 5, 5, 45, 204, 5, 5, 85, 184, 5, 5, 96, 177, 5, 5, 216, 231, 5, 5, 85, 234, 5, 5, 200, 233, 5, 5, 78, 210, 5, 5, 223, 231, 5, 5, 12, 200, 5, 5, 168, 230, 5, 5, 81, 228, 5, 5, 234, 242, 5, 5, 98, 207, 5, 5, 238, 223, 5, 5, 25, 181, 5, 5, 26, 181, 5, 5, 3, 238, 5, 5, 187, 216, 5, 5, 82, 228, 5, 5, 29, 186, 5, 5, 30, 186, 5, 5, 244, 214, 5, 5, 39, 208, 5, 5, 40, 208, 5, 5, 240, 239, 5, 5, 206, 191, 5, 5, 32, 174, 5, 5, 155, 177, 5, 5, 80, 244, 5, 5, 77, 226, 5, 5, 251, 224, 5, 5, 162, 183, 5, 5, 183, 186, 5, 5, 112, 237, 5, 5, 24, 237, 5, 5, 69, 188, 5, 5, 55, 185, 5, 5, 167, 221, 5, 5, 158, 225, 5, 5, 253, 224, 5, 5, 64, 227, 5, 5, 139, 236, 5, 5, 24, 170, 5, 5, 103, 238, 5, 5, 123, 199, 5, 5, 39, 193, 5, 5, 38, 181, 5, 5, 54, 190, 5, 5, 195, 199, 5, 5, 98, 243, 5, 5, 41, 193, 5, 5, 168, 168, 5, 5, 212, 181, 5, 5, 235, 202, 5, 5, 148, 190, 5, 5, 202, 178, 5, 5, 144, 236, 5, 5, 94, 212, 5, 5, 151, 197, 5, 5, 112, 214, 5, 5, 11, 210, 5, 5, 135, 183, 5, 5, 106, 233, 5, 5, 56, 183, 5, 5, 88, 179, 5, 5, 197, 217, 5, 5, 242, 199, 5, 5, 184, 249, 5, 5, 150, 190, 5, 5, 237, 214, 5, 5, 91, 198, 5, 5, 124, 244, 5, 5, 208, 169, 5, 5, 7, 239, 5, 5, 117, 194, 5, 5, 210, 225, 5, 5, 250, 243, 5, 5, 217, 169, 5, 5, 192, 230, 5, 5, 239, 190, 5, 5, 210, 182, 5, 5, 133, 179, 5, 5, 213, 235, 5, 5, 234, 169, 5, 5, 113, 179, 5, 5, 145, 168, 5, 5, 228, 220, 5, 5, 219, 229, 5, 5, 204, 187, 5, 5, 146, 168, 5, 5, 148, 230, 5, 5, 147, 168, 5, 5, 121, 194, 5, 5, 253, 202, 5, 5, 46, 181, 5, 5, 58, 243, 5, 5, 122, 194, 5, 5, 73, 222, 5, 5, 64, 183, 5, 5, 226, 191, 5, 5, 145, 170, 5, 5, 40, 206, 5, 5, 69, 181, 5, 5, 148, 211, 5, 5, 135, 211, 5, 5, 239, 179, 5, 5, 255, 209, 5, 5, 34, 215, 5, 5, 128, 193, 5, 5, 176, 224, 5, 5, 173, 196, 5, 5, 157, 211, 5, 5, 233, 172, 5, 5, 113, 188, 5, 5, 229, 231, 5, 5, 117, 182, 5, 5, 175, 192, 5, 5, 41, 223, 5, 5, 15, 219, 5, 5, 178, 182, 5, 5, 194, 228, 5, 5, 46, 189, 5, 5, 118, 248, 5, 5, 77, 184, 5, 5, 172, 206, 5, 5, 215, 177, 5, 5, 154, 218, 5, 5, 166, 183, 5, 5, 75, 196, 5, 5, 85, 236, 5, 5, 220, 187, 5, 5, 201, 229, 5, 5, 71, 247, 5, 5, 228, 211, 5, 5, 167, 180, 5, 5, 200, 192, 5, 5, 40, 199, 5, 5, 113, 247, 5, 5, 114, 237, 5, 5, 219, 195, 5, 5, 69, 243, 5, 5, 201, 208, 5, 5, 156, 183, 5, 5, 147, 174, 5, 5, 10, 243, 5, 5, 120, 248, 5, 5, 86, 211, 5, 5, 199, 236, 5, 5, 233, 209, 5, 5, 215, 233, 5, 5, 19, 183, 5, 5, 28, 170, 5, 5, 132, 241, 5, 5, 24, 228, 5, 5, 75, 183, 5, 5, 145, 228, 5, 5, 152, 182, 5, 5, 84, 230, 5, 5, 16, 242, 5, 5, 26, 187, 5, 5, 170, 180, 5, 5, 190, 168, 5, 5, 123, 210, 5, 5, 159, 247, 5, 5, 206, 248, 5, 5, 70, 245, 5, 5, 125, 177, 5, 5, 5, 210, 5, 5, 236, 242, 5, 5, 69, 227, 5, 5, 20, 232, 5, 5, 124, 246, 5, 5, 92, 181, 5, 5, 201, 207, 5, 5, 94, 209, 5, 5, 66, 214, 5, 5, 53, 170, 5, 5, 26, 221, 5, 5, 69, 216, 5, 5, 93, 181, 5, 5, 161, 188, 5, 5, 184, 234, 5, 5, 28, 182, 5, 5, 21, 246, 5, 5, 229, 208, 5, 5, 34, 169, 5, 5, 176, 209, 5, 5, 167, 183, 5, 5, 46, 204, 5, 5, 61, 249, 5, 5, 62, 249, 5, 5, 248, 221, 5, 5, 74, 218, 5, 5, 80, 220, 5, 5, 26, 237, 5, 5, 243, 206, 5, 5, 128, 195, 5, 5, 47, 186, 5, 5, 114, 223, 5, 5, 48, 234, 5, 5, 32, 229, 5, 5, 3, 249, 5, 5, 62, 197, 5, 5, 81, 220, 5, 5, 130, 213, 5, 5, 162, 218, 5, 5, 130, 237, 5, 5, 55, 180, 5, 5, 89, 194, 5, 5, 127, 222, 5, 5, 113, 192, 5, 5, 153, 184, 5, 5, 219, 232, 5, 5, 34, 190, 5, 5, 127, 197, 5, 5, 158, 194, 5, 5, 26, 186, 5, 5, 110, 201, 5, 5, 139, 193, 5, 5, 140, 193, 5, 5, 118, 238, 5, 5, 224, 210, 5, 5, 126, 246, 5, 5, 12, 192, 5, 5, 16, 192, 5, 5, 113, 232, 5, 5, 9, 189, 5, 5, 2, 187, 5, 5, 25, 198, 5, 5, 124, 199, 5, 5, 203, 236, 5, 5, 52, 209, 5, 5, 184, 203, 5, 5, 125, 212, 5, 5, 148, 181, 5, 5, 96, 241, 5, 5, 94, 236, 5, 5, 238, 172, 5, 5, 31, 221, 5, 5, 173, 239, 5, 5, 216, 221, 5, 5, 77, 187, 5, 5, 211, 217, 5, 5, 4, 249, 5, 5, 150, 223, 5, 5, 202, 228, 5, 5, 59, 189, 5, 5, 61, 228, 5, 5, 29, 205, 5, 5, 33, 236, 5, 5, 102, 218, 5, 5, 159, 194, 5, 5, 29, 206, 5, 5, 159, 213, 5, 5, 119, 232, 5, 5, 222, 220, 5, 5, 73, 202, 5, 5, 96, 174, 5, 5, 160, 233, 5, 5, 52, 207, 5, 5, 176, 183, 5, 5, 133, 224, 5, 5, 149, 207, 5, 5, 36, 229, 5, 5, 5, 199, 5, 5, 73, 176, 5, 5, 74, 176, 5, 5, 239, 226, 5, 5, 228, 210, 5, 5, 35, 234, 5, 5, 28, 210, 5, 5, 150, 241, 5, 5, 236, 209, 5, 5, 247, 213, 5, 5, 108, 210, 5, 5, 128, 197, 5, 5, 37, 221, 5, 5, 223, 246, 5, 5, 252, 187, 5, 5, 27, 179, 5, 5, 123, 248, 5, 5, 103, 181, 5, 5, 138, 213, 5, 5, 188, 227, 5, 5, 120, 232, 5, 5, 158, 208, 5, 5, 58, 230, 5, 5, 95, 236, 5, 5, 96, 230, 5, 5, 0, 136, 5, 5, 99, 236, 5, 5, 151, 241, 5, 5, 123, 223, 5, 5, 152, 241, 5, 5, 49, 228, 5, 5, 129, 202, 5, 5, 28, 232, 5, 5, 229, 196, 5, 5, 176, 217, 5, 5, 153, 248, 5, 5, 73, 214, 5, 5, 29, 232, 5, 5, 217, 248, 5, 5, 30, 180, 5, 5, 109, 173, 5, 5, 158, 205, 5, 5, 0, 158, 5, 5, 94, 181, 5, 5, 95, 181, 5, 5, 31, 205, 5, 5, 137, 208, 5, 5, 191, 176, 5, 5, 3, 214, 5, 5, 8, 197, 5, 5, 217, 213, 5, 5, 59, 172, 5, 5, 65, 204, 5, 5, 120, 228, 5, 5, 20, 171, 5, 5, 2, 233, 5, 5, 161, 222, 5, 5, 213, 214, 5, 5, 15, 191, 5, 5, 126, 182, 5, 5, 248, 229, 5, 5, 80, 219, 5, 5, 31, 225, 5, 5, 182, 182, 5, 5, 183, 182, 5, 5, 85, 195, 5, 5, 79, 226, 5, 5, 135, 210, 5, 5, 220, 217, 5, 5, 81, 196, 5, 5, 153, 191, 5, 5, 23, 196, 5, 5, 57, 215, 5, 5, 110, 179, 5, 5, 54, 234, 5, 5, 248, 189, 5, 5, 86, 228, 5, 5, 170, 200, 5, 5, 74, 170, 5, 5, 191, 239, 5, 5, 46, 175, 5, 5, 97, 210, 5, 5, 97, 184, 5, 5, 91, 196, 5, 5, 82, 206, 5, 5, 81, 168, 5, 5, 213, 200, 5, 5, 55, 200, 5, 5, 228, 239, 5, 5, 38, 236, 5, 5, 230, 196, 5, 5, 142, 202, 5, 5, 90, 179, 5, 5, 30, 232, 5, 5, 24, 190, 5, 5, 60, 190, 5, 5, 55, 240, 5, 5, 18, 173, 5, 5, 80, 221, 5, 5, 222, 226, 5, 5, 84, 178, 5, 5, 190, 237, 5, 5, 141, 230, 5, 5, 46, 210, 5, 5, 75, 176, 5, 5, 108, 193, 5, 5, 59, 227, 5, 5, 226, 234, 5, 5, 47, 242, 5, 5, 255, 228, 5, 5, 41, 179, 5, 5, 247, 218, 5, 5, 191, 207, 5, 5, 227, 177, 5, 5, 73, 219, 5, 5, 100, 216, 5, 5, 16, 191, 5, 5, 5, 230, 5, 5, 108, 215, 5, 5, 30, 176, 5, 5, 183, 208, 5, 5, 100, 184, 5, 5, 130, 195, 5, 5, 185, 180, 5, 5, 149, 238, 5, 5, 123, 247, 5, 5, 157, 207, 5, 5, 196, 190, 5, 5, 139, 208, 5, 5, 42, 168, 5, 5, 56, 240, 5, 5, 11, 235, 5, 5, 44, 195, 5, 5, 210, 228, 5, 5, 209, 207, 5, 5, 243, 242, 5, 5, 41, 246, 5, 5, 15, 217, 5, 5, 49, 222, 5, 5, 172, 233, 5, 5, 26, 228, 5, 5, 57, 203, 5, 5, 119, 207, 5, 5, 78, 218, 5, 5, 70, 223, 5, 5, 65, 213, 5, 5, 102, 229, 5, 5, 52, 228, 5, 5, 228, 176, 5, 5, 183, 189, 5, 5, 136, 168, 5, 5, 91, 195, 5, 5, 109, 169, 5, 5, 152, 234, 5, 5, 60, 227, 5, 5, 42, 187, 5, 5, 118, 175, 5, 5, 143, 237, 5, 5, 48, 213, 5, 5, 53, 231, 5, 5, 200, 242, 5, 5, 179, 240, 5, 5, 204, 218, 5, 5, 55, 219, 5, 5, 161, 174, 5, 5, 143, 204, 5, 5, 194, 207, 5, 5, 17, 212, 5, 5, 137, 168, 5, 5, 174, 206, 5, 5, 32, 186, 5, 5, 26, 200, 5, 5, 130, 216, 5, 5, 121, 193, 5, 5, 226, 219, 5, 5, 194, 245, 5, 5, 201, 242, 5, 5, 45, 195, 5, 5, 209, 218, 5, 5, 40, 209, 5, 5, 63, 240, 5, 5, 135, 192, 5, 5, 6, 224, 5, 5, 238, 210, 5, 5, 42, 232, 5, 5, 165, 225, 5, 5, 63, 212, 5, 5, 44, 201, 5, 5, 90, 210, 5, 5, 97, 169, 5, 5, 9, 168, 5, 5, 49, 213, 5, 5, 141, 213, 5, 5, 78, 239, 5, 5, 104, 213, 5, 5, 74, 242, 5, 5, 160, 201, 5, 5, 49, 235, 5, 5, 241, 206, 5, 5, 82, 238, 5, 5, 23, 245, 5, 5, 165, 198, 5, 5, 197, 240, 5, 5, 183, 210, 5, 5, 20, 202, 5, 5, 233, 178, 5, 5, 47, 187, 5, 5, 69, 236, 5, 5, 203, 195, 5, 5, 84, 248, 5, 5, 140, 173, 5, 5, 59, 244, 5, 5, 236, 187, 5, 5, 120, 210, 5, 5, 129, 201, 5, 5, 66, 209, 5, 5, 114, 245, 5, 5, 2, 190, 5, 5, 3, 190, 5, 5, 148, 168, 5, 5, 121, 210, 5, 5, 213, 215, 5, 5, 216, 206, 5, 5, 129, 213, 5, 5, 69, 186, 5, 5, 105, 230, 5, 5, 24, 216, 5, 5, 29, 228, 5, 5, 244, 244, 5, 5, 187, 208, 5, 5, 46, 232, 5, 5, 47, 232, 5, 5, 129, 209, 5, 5, 129, 203, 5, 5, 6, 203, 5, 5, 69, 190, 5, 5, 51, 187, 5, 5, 177, 178, 5, 5, 147, 246, 5, 5, 167, 234, 5, 5, 86, 238, 5, 5, 62, 190, 5, 5, 250, 230, 5, 5, 35, 200, 5, 5, 106, 218, 5, 5, 251, 230, 5, 5, 237, 236, 5, 5, 32, 173, 5, 5, 174, 194, 5, 5, 196, 207, 5, 5, 135, 182, 5, 5, 60, 182, 5, 5, 254, 231, 5, 5, 104, 239, 5, 5, 114, 191, 5, 5, 20, 195, 5, 5, 123, 184, 5, 5, 78, 222, 5, 5, 131, 170, 5, 5, 140, 220, 5, 5, 89, 224, 5, 5, 219, 215, 5, 5, 42, 203, 5, 5, 127, 190, 5, 5, 234, 233, 5, 5, 242, 210, 5, 5, 46, 180, 5, 5, 105, 239, 5, 5, 81, 172, 5, 5, 18, 168, 5, 5, 229, 210, 5, 5, 69, 211, 5, 5, 175, 206, 5, 5, 172, 226, 5, 5, 76, 175, 5, 5, 2, 196, 5, 5, 124, 176, 5, 5, 201, 241, 5, 5, 144, 213, 5, 5, 228, 218, 5, 5, 31, 219, 5, 5, 10, 210, 5, 5, 75, 236, 5, 5, 246, 224, 5, 5, 82, 240, 5, 5, 170, 215, 5, 5, 244, 210, 5, 5, 34, 243, 5, 5, 134, 240, 5, 5, 145, 208, 5, 5, 21, 170, 5, 5, 176, 206, 5, 5, 89, 221, 5, 5, 59, 234, 5, 5, 192, 210, 5, 5, 137, 180, 5, 5, 196, 204, 5, 5, 165, 242, 5, 5, 14, 201, 5, 5, 252, 236, 5, 5, 248, 222, 5, 5, 108, 204, 5, 5, 246, 194, 5, 5, 230, 210, 5, 5, 16, 201, 5, 5, 86, 215, 5, 5, 91, 240, 5, 5, 250, 222, 5, 5, 144, 191, 5, 5, 247, 217, 5, 5, 94, 208, 5, 5, 57, 202, 5, 5, 41, 206, 5, 5, 132, 236, 5, 5, 252, 247, 5, 5, 18, 201, 5, 5, 96, 249, 5, 5, 61, 195, 5, 5, 182, 197, 5, 5, 183, 197, 5, 5, 63, 199, 5, 5, 115, 243, 5, 5, 173, 206, 5, 5, 81, 249, 5, 5, 65, 177, 5, 5, 110, 224, 5, 5, 209, 183, 5, 5, 162, 169, 5, 5, 253, 248, 5, 5, 70, 233, 5, 5, 246, 233, 5, 5, 37, 208, 5, 5, 118, 223, 5, 5, 226, 224, 5, 5, 84, 169, 5, 5, 206, 192, 5, 5, 48, 186, 5, 5, 136, 211, 5, 5, 101, 235, 5, 5, 91, 241, 5, 5, 96, 248, 5, 5, 209, 187, 5, 5, 30, 206, 5, 5, 113, 224, 5, 5, 63, 210, 5, 5, 211, 208, 5, 5, 192, 176, 5, 5, 58, 215, 5, 5, 162, 222, 5, 5, 104, 172, 5, 5, 238, 235, 5, 5, 18, 249, 5, 5, 162, 210, 5, 5, 190, 183, 5, 5, 27, 249, 5, 5, 21, 202, 5, 5, 110, 235, 5, 5, 239, 171, 5, 5, 32, 219, 5, 5, 253, 209, 5, 5, 31, 211, 5, 5, 32, 211, 5, 5, 92, 240, 5, 5, 42, 206, 5, 5, 133, 208, 5, 5, 53, 211, 5, 5, 237, 218, 5, 5, 175, 224, 5, 5, 67, 187, 5, 5, 190, 244, 5, 5, 230, 216, 5, 5, 224, 241, 5, 5, 96, 222, 5, 5, 34, 168, 5, 5, 79, 227, 5, 5, 179, 223, 5, 5, 101, 228, 5, 5, 13, 219, 5, 5, 177, 237, 5, 5, 49, 189, 5, 5, 106, 238, 5, 5, 55, 196, 5, 5, 173, 248, 5, 5, 100, 208, 5, 5, 7, 248, 5, 5, 78, 178, 5, 5, 54, 189, 5, 5, 97, 249, 5, 5, 189, 186, 5, 5, 174, 247, 5, 5, 252, 179, 5, 5, 114, 228, 5, 5, 110, 238, 5, 5, 100, 169, 5, 5, 212, 221, 5, 5, 213, 221, 5, 5, 39, 175, 5, 5, 73, 221, 5, 5, 3, 199, 5, 5, 4, 235, 5, 5, 27, 222, 5, 5, 97, 241, 5, 5, 151, 190, 5, 5, 127, 238, 5, 5, 164, 226, 5, 5, 251, 221, 5, 5, 114, 232, 5, 5, 213, 185, 5, 5, 97, 174, 5, 5, 52, 218, 5, 5, 214, 185, 5, 5, 248, 232, 5, 5, 185, 243, 5, 5, 173, 244, 5, 5, 105, 169, 5, 5, 222, 187, 5, 5, 208, 236, 5, 5, 249, 232, 5, 5, 53, 193, 5, 5, 77, 221, 5, 5, 4, 174, 5, 5, 200, 218, 5, 5, 236, 190, 5, 5, 105, 208, 5, 5, 80, 199, 5, 5, 160, 199, 5, 5, 51, 171, 5, 5, 248, 223, 5, 5, 60, 172, 5, 5, 206, 243, 5, 5, 224, 192, 5, 5, 147, 242, 5, 5, 225, 192, 5, 5, 229, 239, 5, 5, 107, 208, 5, 5, 81, 199, 5, 5, 178, 216, 5, 5, 222, 188, 5, 5, 191, 245, 5, 5, 34, 194, 5, 5, 101, 184, 5, 5, 60, 211, 5, 5, 136, 171, 5, 5, 118, 190, 5, 5, 210, 207, 5, 5, 152, 216, 5, 5, 6, 188, 5, 5, 48, 242, 5, 5, 226, 221, 5, 5, 61, 211, 5, 5, 236, 195, 5, 5, 63, 211, 5, 5, 8, 247, 5, 5, 245, 241, 5, 5, 111, 169, 5, 5, 161, 199, 5, 5, 64, 211, 5, 5, 154, 216, 5, 5, 121, 209, 5, 5, 226, 172, 5, 5, 101, 197, 5, 5, 164, 186, 5, 5, 155, 216, 5, 5, 162, 189, 5, 5, 153, 230, 5, 5, 123, 203, 5, 5, 234, 221, 5, 5, 65, 211, 5, 5, 197, 244, 5, 5, 83, 221, 5, 5, 69, 229, 5, 5, 146, 233, 5, 5, 162, 199, 5, 5, 106, 191, 5, 5, 130, 203, 5, 5, 67, 198, 5, 5, 128, 190, 5, 5, 121, 239, 5, 5, 168, 238, 5, 5, 121, 169, 5, 5, 110, 216, 5, 5, 41, 175, 5, 5, 122, 169, 5, 5, 110, 183, 5, 5, 70, 177, 5, 5, 218, 180, 5, 5, 110, 223, 5, 5, 160, 235, 5, 5, 115, 178, 5, 5, 8, 206, 5, 5, 222, 180, 5, 5, 100, 222, 5, 5, 225, 213, 5, 5, 80, 183, 5, 5, 79, 248, 5, 5, 177, 183, 5, 5, 249, 220, 5, 5, 6, 199, 5, 5, 90, 194, 5, 5, 95, 194, 5, 5, 81, 248, 5, 5, 98, 229, 5, 5, 2, 250, 5, 5, 179, 235, 5, 5, 210, 222, 5, 5, 229, 180, 5, 5, 127, 178, 5, 5, 60, 233, 5, 5, 62, 195, 5, 5, 216, 220, 5, 5, 243, 181, 5, 5, 244, 181, 5, 5, 245, 181, 5, 5, 144, 184, 5, 5, 129, 193, 5, 5, 120, 222, 5, 5, 249, 173, 5, 5, 186, 220, 5, 5, 187, 220, 5, 5, 64, 209, 5, 5, 141, 184, 5, 5, 130, 173, 5, 5, 172, 203, 5, 5, 80, 232, 5, 5, 81, 232, 5, 5, 100, 200, 5, 5, 24, 241, 5, 5, 162, 228, 5, 5, 25, 241, 5, 5, 168, 203, 5, 5, 169, 203, 5, 5, 126, 237, 5, 5, 43, 207, 5, 5, 166, 228, 5, 5, 163, 228, 5, 5, 164, 228, 5, 5, 146, 184, 5, 5, 127, 237, 5, 5, 75, 181, 5, 5, 22, 200, 5, 5, 101, 247, 5, 5, 197, 196, 5, 5, 209, 184, 5, 5, 95, 186, 5, 5, 210, 184, 5, 5, 173, 227, 5, 5, 214, 184, 5, 5, 215, 184, 5, 5, 166, 221, 5, 5, 255, 239, 5, 5, 203, 174, 5, 5, 188, 198, 5, 5, 124, 210, 5, 5, 192, 195, 5, 5, 28, 229, 5, 5, 245, 210, 5, 5, 255, 196, 5, 5, 44, 213, 5, 5, 191, 172, 5, 5, 108, 231, 5, 5, 7, 170, 5, 5, 216, 196, 5, 5, 142, 195, 5, 5, 86, 226, 5, 5, 99, 217, 5, 5, 47, 226, 5, 5, 143, 195, 5, 5, 32, 230, 5, 5, 151, 179, 5, 5, 174, 221, 5, 5, 252, 221, 5, 5, 175, 213, 5, 5, 17, 192, 5, 5, 161, 233, 5, 5, 227, 245, 5, 5, 162, 233, 5, 5, 130, 210, 5, 5, 233, 244, 5, 5, 193, 230, 5, 5, 37, 229, 5, 5, 9, 207, 5, 5, 96, 181, 5, 5, 64, 205, 5, 5, 183, 213, 5, 5, 55, 226, 5, 5, 234, 182, 5, 5, 169, 222, 5, 5, 173, 233, 5, 5, 106, 227, 5, 5, 248, 205, 5, 5, 251, 205, 5, 5, 86, 231, 5, 5, 193, 172, 5, 5, 255, 205, 5, 5, 108, 197, 5, 5, 184, 233, 5, 5, 118, 197, 5, 5, 179, 197, 5, 5, 144, 203, 5, 5, 249, 197, 5, 5, 185, 222, 5, 5, 156, 231, 5, 5, 223, 173, 5, 5, 204, 227, 5, 5, 172, 210, 5, 5, 72, 220, 5, 5, 44, 228, 5, 5, 227, 231, 5, 5, 98, 202, 5, 5, 133, 181, 5, 5, 66, 191, 5, 5, 67, 191, 5, 5, 214, 204, 5, 5, 230, 238, 5, 5, 200, 214, 5, 5, 150, 218, 5, 5, 120, 230, 5, 5, 26, 188, 5, 5, 20, 221, 5, 5, 225, 241, 5, 5, 104, 175, 5, 5, 76, 224, 5, 5, 201, 214, 5, 5, 151, 218, 5, 5, 11, 243, 5, 5, 17, 169, 5, 5, 178, 237, 5, 5, 92, 168, 5, 5, 250, 235, 5, 5, 123, 230, 5, 5, 68, 195, 5, 5, 136, 181, 5, 5, 99, 192, 5, 5, 40, 215, 5, 5, 20, 183, 5, 5, 102, 228, 5, 5, 106, 214, 5, 5, 185, 172, 5, 5, 94, 215, 5, 5, 107, 214, 5, 5, 235, 172, 5, 5, 138, 195, 5, 5, 94, 217, 5, 5, 25, 185, 5, 5, 107, 232, 5, 5, 139, 168, 5, 5, 210, 200, 5, 5, 116, 178, 5, 5, 169, 168, 5, 5, 36, 250, 5, 5, 37, 250, 5, 5, 82, 237, 5, 5, 63, 197, 5, 5, 7, 185, 5, 5, 245, 198, 5, 5, 10, 186, 5, 5, 102, 235, 5, 5, 209, 213, 5, 5, 110, 202, 5, 5, 143, 226, 5, 5, 100, 217, 5, 5, 89, 236, 5, 5, 214, 183, 5, 5, 119, 234, 5, 5, 100, 193, 5, 5, 119, 204, 5, 5, 249, 227, 5, 5, 248, 212, 5, 5, 143, 168, 5, 5, 174, 177, 5, 5, 204, 199, 5, 5, 14, 243, 5, 5, 101, 217, 5, 5, 239, 189, 5, 5, 230, 213, 5, 5, 230, 208, 5, 5, 73, 168, 5, 5, 144, 226, 5, 5, 47, 204, 5, 5, 182, 174, 5, 5, 171, 213, 5, 5, 22, 180, 5, 5, 30, 188, 5, 5, 231, 199, 5, 5, 120, 234, 5, 5, 101, 207, 5, 5, 7, 227, 5, 5, 103, 235, 5, 5, 248, 238, 5, 5, 208, 170, 5, 5, 164, 188, 5, 5, 171, 168, 5, 5, 148, 206, 5, 5, 149, 181, 5, 5, 86, 184, 5, 5, 167, 235, 5, 5, 182, 179, 5, 5, 153, 205, 5, 5, 216, 181, 5, 5, 225, 181, 5, 5, 154, 184, 5, 5, 198, 217, 5, 5, 56, 180, 5, 5, 119, 238, 5, 5, 178, 209, 5, 5, 217, 221, 5, 5, 36, 168, 5, 5, 242, 228, 5, 5, 111, 190, 5, 5, 214, 246, 5, 5, 103, 208, 5, 5, 71, 202, 5, 5, 209, 192, 5, 5, 10, 227, 5, 5, 243, 228, 5, 5, 98, 241, 5, 5, 16, 215, 5, 5, 173, 231, 5, 5, 72, 202, 5, 5, 131, 237, 5, 5, 253, 194, 5, 5, 27, 246, 5, 5, 31, 206, 5, 5, 160, 194, 5, 5, 104, 181, 5, 5, 105, 181, 5, 5, 26, 242, 5, 5, 107, 233, 5, 5, 246, 171, 5, 5, 46, 216, 5, 5, 50, 218, 5, 5, 85, 183, 5, 5, 86, 183, 5, 5, 67, 210, 5, 5, 74, 202, 5, 5, 35, 241, 5, 5, 121, 232, 5, 5, 201, 218, 5, 5, 118, 178, 5, 5, 38, 221, 5, 5, 98, 174, 5, 5, 98, 227, 5, 5, 190, 185, 5, 5, 52, 198, 5, 5, 76, 185, 5, 5, 178, 231, 5, 5, 222, 239, 5, 5, 234, 241, 5, 5, 73, 201, 5, 5, 154, 198, 5, 5, 88, 201, 5, 5, 165, 200, 5, 5, 72, 232, 5, 5, 23, 218, 5, 5, 55, 199, 5, 5, 31, 175, 5, 5, 32, 175, 5, 5, 199, 224, 5, 5, 225, 203, 5, 5, 58, 190, 5, 5, 231, 196, 5, 5, 171, 200, 5, 5, 131, 214, 5, 5, 240, 208, 5, 5, 56, 200, 5, 5, 57, 200, 5, 5, 194, 249, 5, 5, 128, 186, 5, 5, 11, 177, 5, 5, 4, 236, 5, 5, 5, 236, 5, 5, 26, 185, 5, 5, 83, 206, 5, 5, 84, 206, 5, 5, 232, 201, 5, 5, 206, 197, 5, 5, 31, 181, 5, 5, 34, 246, 5, 5, 153, 187, 5, 5, 230, 239, 5, 5, 31, 180, 5, 5, 254, 187, 5, 5, 35, 246, 5, 5, 44, 229, 5, 5, 54, 233, 5, 5, 7, 213, 5, 5, 57, 236, 5, 5, 154, 223, 5, 5, 86, 195, 5, 5, 235, 169, 5, 5, 195, 249, 5, 5, 99, 199, 5, 5, 32, 180, 5, 5, 235, 244, 5, 5, 129, 186, 5, 5, 231, 239, 5, 5, 11, 249, 5, 5, 87, 244, 5, 5, 197, 190, 5, 5, 165, 241, 5, 5, 80, 228, 5, 5, 60, 237, 5, 5, 91, 183, 5, 5, 65, 217, 5, 5, 61, 237, 5, 5, 2, 226, 5, 5, 199, 238, 5, 5, 254, 246, 5, 5, 50, 222, 5, 5, 186, 243, 5, 5, 191, 237, 5, 5, 165, 181, 5, 5, 50, 248, 5, 5, 132, 198, 5, 5, 14, 206, 5, 5, 157, 236, 5, 5, 192, 207, 5, 5, 7, 188, 5, 5, 24, 192, 5, 5, 25, 192, 5, 5, 163, 190, 5, 5, 223, 226, 5, 5, 158, 221, 5, 5, 158, 207, 5, 5, 154, 215, 5, 5, 142, 230, 5, 5, 166, 241, 5, 5, 102, 249, 5, 5, 215, 200, 5, 5, 248, 198, 5, 5, 107, 236, 5, 5, 108, 236, 5, 5, 49, 229, 5, 5, 103, 249, 5, 5, 220, 172, 5, 5, 81, 224, 5, 5, 205, 218, 5, 5, 216, 198, 5, 5, 101, 216, 5, 5, 167, 241, 5, 5, 144, 214, 5, 5, 55, 205, 5, 5, 113, 227, 5, 5, 211, 180, 5, 5, 232, 230, 5, 5, 221, 229, 5, 5, 122, 172, 5, 5, 88, 178, 5, 5, 210, 218, 5, 5, 95, 195, 5, 5, 108, 198, 5, 5, 144, 204, 5, 5, 149, 230, 5, 5, 159, 223, 5, 5, 242, 215, 5, 5, 19, 249, 5, 5, 2, 229, 5, 5, 240, 169, 5, 5, 71, 179, 5, 5, 81, 177, 5, 5, 135, 215, 5, 5, 245, 240, 5, 5, 12, 211, 5, 5, 82, 177, 5, 5, 184, 192, 5, 5, 230, 221, 5, 5, 120, 219, 5, 5, 174, 223, 5, 5, 111, 249, 5, 5, 143, 194, 5, 5, 124, 203, 5, 5, 166, 198, 5, 5, 77, 173, 5, 5, 209, 179, 5, 5, 184, 172, 5, 5, 243, 179, 5, 5, 91, 227, 5, 5, 58, 205, 5, 5, 61, 245, 5, 5, 240, 244, 5, 5, 241, 244, 5, 5, 108, 168, 5, 5, 140, 172, 5, 5, 116, 217, 5, 5, 194, 215, 5, 5, 16, 177, 5, 5, 232, 249, 5, 5, 124, 178, 5, 5, 125, 178, 5, 5, 57, 231, 5, 5, 75, 242, 5, 5, 134, 212, 5, 5, 230, 204, 5, 5, 231, 232, 5, 5, 194, 172, 5, 5, 163, 171, 5, 5, 105, 216, 5, 5, 175, 194, 5, 5, 71, 236, 5, 5, 95, 201, 5, 5, 8, 245, 5, 5, 236, 203, 5, 5, 131, 203, 5, 5, 132, 203, 5, 5, 187, 195, 5, 5, 193, 178, 5, 5, 87, 181, 5, 5, 5, 250, 5, 5, 17, 195, 5, 5, 112, 187, 5, 5, 160, 237, 5, 5, 176, 194, 5, 5, 161, 237, 5, 5, 230, 197, 5, 5, 214, 244, 5, 5, 106, 239, 5, 5, 111, 235, 5, 5, 75, 210, 5, 5, 36, 238, 5, 5, 37, 238, 5, 5, 162, 238, 5, 5, 21, 211, 5, 5, 88, 232, 5, 5, 145, 192, 5, 5, 187, 233, 5, 5, 251, 198, 5, 5, 252, 230, 5, 5, 62, 179, 5, 5, 153, 168, 5, 5, 234, 249, 5, 5, 227, 228, 5, 5, 169, 238, 5, 5, 229, 218, 5, 5, 130, 178, 5, 5, 123, 204, 5, 5, 118, 195, 5, 5, 2, 242, 5, 5, 37, 243, 5, 5, 21, 240, 5, 5, 62, 219, 5, 5, 123, 195, 5, 5, 181, 202, 5, 5, 60, 187, 5, 5, 239, 204, 5, 5, 240, 204, 5, 5, 78, 179, 5, 5, 232, 218, 5, 5, 35, 231, 5, 5, 30, 197, 5, 5, 51, 173, 5, 5, 172, 240, 5, 5, 240, 199, 5, 5, 127, 236, 5, 5, 231, 228, 5, 5, 81, 210, 5, 5, 248, 217, 5, 5, 206, 173, 5, 5, 255, 176, 5, 5, 43, 206, 5, 5, 82, 179, 5, 5, 83, 179, 5, 5, 38, 211, 5, 5, 134, 236, 5, 5, 133, 236, 5, 5, 191, 236, 5, 5, 14, 200, 5, 5, 192, 236, 5, 5, 203, 175, 5, 5, 137, 199, 5, 5, 204, 175, 5, 5, 127, 247, 5, 5, 182, 190, 5, 5, 12, 196, 5, 5, 162, 235, 5, 5, 186, 173, 5, 5, 187, 173, 5, 5, 206, 203, 5, 5, 203, 185, 5, 5, 30, 250, 5, 5, 39, 216, 5, 5, 57, 197, 5, 5, 233, 185, 5, 5, 60, 197, 5, 5, 30, 230, 5, 5, 60, 214, 5, 5, 61, 214, 5, 5, 239, 172, 5, 5, 66, 217, 5, 5, 76, 217, 5, 5, 173, 192, 5, 5, 181, 238, 5, 5, 107, 223, 5, 5, 163, 168, 5, 5, 85, 246, 5, 5, 71, 245, 5, 5, 220, 232, 5, 5, 128, 238, 5, 5, 195, 195, 5, 5, 204, 235, 5, 5, 153, 197, 5, 5, 175, 168, 5, 5, 214, 235, 5, 5, 165, 195, 5, 5, 201, 183, 5, 5, 148, 243, 5, 5, 44, 170, 5, 5, 9, 222, 5, 5, 11, 172, 5, 5, 229, 179, 5, 5, 235, 222, 5, 5, 55, 182, 5, 5, 250, 210, 5, 5, 168, 221, 5, 5, 251, 182, 5, 5, 240, 211, 5, 5, 168, 246, 5, 5, 177, 230, 5, 5, 235, 189, 5, 5, 10, 178, 5, 5, 238, 228, 5, 5, 50, 245, 5, 5, 140, 225, 5, 5, 175, 177, 5, 5, 95, 209, 5, 5, 124, 212, 5, 5, 241, 211, 5, 5, 180, 226, 5, 5, 141, 171, 5, 5, 237, 202, 5, 5, 206, 246, 5, 5, 163, 247, 5, 5, 176, 171, 5, 5, 207, 246, 5, 5, 25, 179, 5, 5, 102, 209, 5, 5, 8, 239, 5, 5, 9, 239, 5, 5, 176, 213, 5, 5, 22, 215, 5, 5, 144, 197, 5, 5, 17, 219, 5, 5, 236, 222, 5, 5, 180, 177, 5, 5, 69, 246, 5, 5, 17, 223, 5, 5, 47, 216, 5, 5, 180, 245, 5, 5, 178, 221, 5, 5, 53, 218, 5, 5, 48, 231, 5, 5, 49, 169, 5, 5, 187, 177, 5, 5, 33, 187, 5, 5, 55, 175, 5, 5, 184, 213, 5, 5, 55, 245, 5, 5, 193, 219, 5, 5, 87, 228, 5, 5, 190, 177, 5, 5, 249, 228, 5, 5, 132, 173, 5, 5, 50, 220, 5, 5, 132, 214, 5, 5, 52, 244, 5, 5, 171, 187, 5, 5, 120, 207, 5, 5, 159, 180, 5, 5, 112, 189, 5, 5, 72, 246, 5, 5, 153, 234, 5, 5, 111, 208, 5, 5, 3, 229, 5, 5, 6, 230, 5, 5, 5, 184, 5, 5, 50, 239, 5, 5, 54, 169, 5, 5, 189, 213, 5, 5, 195, 179, 5, 5, 221, 185, 5, 5, 18, 212, 5, 5, 241, 190, 5, 5, 166, 225, 5, 5, 114, 208, 5, 5, 122, 193, 5, 5, 225, 225, 5, 5, 26, 191, 5, 5, 85, 247, 5, 5, 100, 220, 5, 5, 22, 207, 5, 5, 206, 209, 5, 5, 241, 170, 5, 5, 173, 187, 5, 5, 60, 244, 5, 5, 207, 209, 5, 5, 55, 169, 5, 5, 62, 245, 5, 5, 65, 196, 5, 5, 81, 173, 5, 5, 24, 184, 5, 5, 27, 247, 5, 5, 165, 189, 5, 5, 61, 182, 5, 5, 246, 175, 5, 5, 112, 170, 5, 5, 123, 170, 5, 5, 45, 245, 5, 5, 126, 208, 5, 5, 255, 215, 5, 5, 15, 173, 5, 5, 36, 183, 5, 5, 5, 208, 5, 5, 57, 169, 5, 5, 77, 175, 5, 5, 220, 208, 5, 5, 131, 175, 5, 5, 117, 195, 5, 5, 95, 232, 5, 5, 236, 200, 5, 5, 236, 184, 5, 5, 166, 213, 5, 5, 195, 210, 5, 5, 130, 193, 5, 5, 124, 171, 5, 5, 127, 171, 5, 5, 49, 234, 5, 5, 247, 184, 5, 5, 111, 237, 5, 5, 147, 190, 5, 5, 88, 241, 5, 5, 11, 241, 5, 5, 182, 192, 5, 5, 13, 187, 5, 5, 30, 213, 5, 5, 199, 226, 5, 5, 82, 244, 5, 5, 14, 187, 5, 5, 119, 248, 5, 5, 63, 209, 5, 5, 212, 216, 5, 5, 50, 170, 5, 5, 111, 216, 5, 5, 8, 181, 5, 5, 243, 175, 5, 5, 71, 187, 5, 5, 14, 236, 5, 5, 243, 168, 5, 5, 139, 195, 5, 5, 200, 234, 5, 5, 75, 205, 5, 5, 85, 230, 5, 5, 121, 248, 5, 5, 107, 199, 5, 5, 47, 240, 5, 5, 2, 179, 5, 5, 82, 212, 5, 5, 105, 179, 5, 5, 230, 235, 5, 5, 202, 208, 5, 5, 171, 185, 5, 5, 144, 176, 5, 5, 44, 184, 5, 5, 8, 227, 5, 5, 58, 212, 5, 5, 221, 182, 5, 5, 192, 232, 5, 5, 189, 238, 5, 5, 215, 246, 5, 5, 136, 226, 5, 5, 216, 246, 5, 5, 91, 234, 5, 5, 171, 180, 5, 5, 43, 250, 5, 5, 250, 232, 5, 5, 99, 227, 5, 5, 83, 187, 5, 5, 110, 199, 5, 5, 53, 207, 5, 5, 219, 226, 5, 5, 69, 241, 5, 5, 247, 171, 5, 5, 129, 171, 5, 5, 142, 174, 5, 5, 172, 200, 5, 5, 75, 170, 5, 5, 125, 192, 5, 5, 39, 168, 5, 5, 204, 222, 5, 5, 164, 198, 5, 5, 191, 240, 5, 5, 25, 228, 5, 5, 155, 223, 5, 5, 205, 222, 5, 5, 206, 216, 5, 5, 49, 242, 5, 5, 239, 241, 5, 5, 205, 208, 5, 5, 192, 223, 5, 5, 172, 172, 5, 5, 168, 232, 5, 5, 236, 182, 5, 5, 198, 196, 5, 5, 175, 181, 5, 5, 43, 187, 5, 5, 202, 204, 5, 5, 54, 220, 5, 5, 247, 202, 5, 5, 45, 201, 5, 5, 193, 223, 5, 5, 9, 247, 5, 5, 31, 172, 5, 5, 208, 216, 5, 5, 202, 196, 5, 5, 199, 196, 5, 5, 204, 195, 5, 5, 109, 168, 5, 5, 111, 200, 5, 5, 38, 205, 5, 5, 37, 240, 5, 5, 173, 203, 5, 5, 198, 177, 5, 5, 180, 205, 5, 5, 79, 239, 5, 5, 127, 175, 5, 5, 33, 173, 5, 5, 121, 227, 5, 5, 89, 223, 5, 5, 231, 233, 5, 5, 206, 208, 5, 5, 157, 173, 5, 5, 106, 230, 5, 5, 240, 182, 5, 5, 16, 187, 5, 5, 120, 199, 5, 5, 154, 199, 5, 5, 132, 170, 5, 5, 220, 215, 5, 5, 188, 233, 5, 5, 11, 204, 5, 5, 12, 204, 5, 5, 145, 203, 5, 5, 95, 205, 5, 5, 20, 193, 5, 5, 173, 240, 5, 5, 10, 232, 5, 5, 213, 226, 5, 5, 213, 240, 5, 5, 58, 202, 5, 5, 215, 226, 5, 5, 254, 239, 5, 5, 159, 235, 5, 5, 82, 236, 5, 5, 217, 226, 5, 5, 216, 178, 5, 5, 255, 211, 5, 5, 14, 194, 5, 5, 44, 191, 5, 5, 6, 210, 5, 5, 45, 191, 5, 5, 232, 185, 5, 5, 222, 210, 5, 5, 90, 198, 5, 5, 200, 170, 5, 5, 231, 238, 5, 5, 228, 214, 5, 5, 121, 211, 5, 5, 55, 183, 5, 5, 37, 197, 5, 5, 148, 236, 5, 5, 10, 239, 5, 5, 251, 243, 5, 5, 113, 170, 5, 5, 221, 238, 5, 5, 59, 238, 5, 5, 7, 182, 5, 5, 170, 219, 5, 5, 13, 222, 5, 5, 8, 182, 5, 5, 51, 222, 5, 5, 59, 222, 5, 5, 204, 185, 5, 5, 152, 179, 5, 5, 2, 240, 5, 5, 228, 189, 5, 5, 202, 183, 5, 5, 45, 189, 5, 5, 24, 230, 5, 5, 170, 227, 5, 5, 177, 174, 5, 5, 133, 194, 5, 5, 209, 168, 5, 5, 71, 221, 5, 5, 18, 179, 5, 5, 18, 245, 5, 5, 188, 197, 5, 5, 159, 225, 5, 5, 45, 184, 5, 5, 3, 179, 5, 5, 47, 208, 5, 5, 21, 232, 5, 5, 172, 189, 5, 5, 175, 173, 5, 5, 142, 211, 5, 5, 77, 196, 5, 5, 224, 245, 5, 5, 129, 238, 5, 5, 83, 208, 5, 5, 199, 217, 5, 5, 84, 228, 5, 5, 193, 220, 5, 5, 115, 219, 5, 5, 5, 235, 5, 5, 86, 196, 5, 5, 134, 179, 5, 5, 19, 245, 5, 5, 137, 194, 5, 5, 210, 215, 5, 5, 168, 212, 5, 5, 19, 178, 5, 5, 211, 215, 5, 5, 83, 170, 5, 5, 84, 170, 5, 5, 255, 220, 5, 5, 25, 178, 5, 5, 228, 193, 5, 5, 33, 186, 5, 5, 101, 185, 5, 5, 24, 182, 5, 5, 96, 170, 5, 5, 71, 199, 5, 5, 235, 193, 5, 5, 46, 171, 5, 5, 27, 233, 5, 5, 37, 178, 5, 5, 159, 187, 5, 5, 145, 194, 5, 5, 88, 189, 5, 5, 57, 208, 5, 5, 163, 187, 5, 5, 93, 228, 5, 5, 10, 198, 5, 5, 192, 192, 5, 5, 193, 192, 5, 5, 24, 187, 5, 5, 51, 178, 5, 5, 151, 205, 5, 5, 152, 205, 5, 5, 164, 227, 5, 5, 82, 191, 5, 5, 255, 246, 5, 5, 93, 191, 5, 5, 94, 191, 5, 5, 163, 238, 5, 5, 164, 238, 5, 5, 172, 238, 5, 5, 173, 238, 5, 5, 5, 243, 5, 5, 6, 243, 5, 5, 73, 220, 5, 5, 26, 234, 5, 5, 162, 229, 5, 5, 6, 227, 5, 5, 204, 236, 5, 5, 205, 236, 5, 5, 27, 242, 5, 5, 150, 174, 5, 5, 61, 172, 5, 5, 236, 170, 5, 5, 135, 179, 5, 5, 52, 171, 5, 5, 169, 212, 5, 5, 246, 240, 5, 5, 93, 213, 5, 5, 24, 245, 5, 5, 167, 240, 5, 5, 175, 174, 5, 5, 223, 174, 5, 5, 201, 248, 5, 5, 23, 228, 5, 5, 100, 192, 5, 5, 161, 182, 5, 5, 72, 247, 5, 5, 238, 238, 5, 5, 172, 228, 5, 5, 224, 173, 5, 5, 31, 170, 5, 5, 203, 178, 5, 5, 48, 204, 5, 5, 215, 183, 5, 5, 177, 228, 5, 5, 22, 246, 5, 5, 223, 176, 5, 5, 178, 228, 5, 5, 78, 196, 5, 5, 181, 177, 5, 5, 186, 230, 5, 5, 205, 235, 5, 5, 253, 188, 5, 5, 52, 237, 5, 5, 79, 190, 5, 5, 10, 206, 5, 5, 131, 189, 5, 5, 181, 228, 5, 5, 138, 174, 5, 5, 224, 246, 5, 5, 180, 234, 5, 5, 87, 196, 5, 5, 100, 227, 5, 5, 190, 176, 5, 5, 66, 247, 5, 5, 166, 200, 5, 5, 193, 176, 5, 5, 159, 178, 5, 5, 246, 211, 5, 5, 78, 231, 5, 5, 36, 180, 5, 5, 226, 192, 5, 5, 133, 173, 5, 5, 236, 246, 5, 5, 194, 176, 5, 5, 139, 247, 5, 5, 173, 200, 5, 5, 50, 242, 5, 5, 165, 233, 5, 5, 154, 195, 5, 5, 29, 194, 5, 5, 52, 222, 5, 5, 109, 193, 5, 5, 214, 170, 5, 5, 198, 190, 5, 5, 102, 184, 5, 5, 180, 235, 5, 5, 50, 229, 5, 5, 64, 212, 5, 5, 144, 237, 5, 5, 217, 228, 5, 5, 233, 230, 5, 5, 44, 246, 5, 5, 119, 213, 5, 5, 15, 226, 5, 5, 165, 178, 5, 5, 49, 246, 5, 5, 52, 246, 5, 5, 41, 171, 5, 5, 166, 178, 5, 5, 22, 175, 5, 5, 232, 173, 5, 5, 229, 194, 5, 5, 115, 191, 5, 5, 233, 194, 5, 5, 35, 191, 5, 5, 178, 207, 5, 5, 25, 205, 5, 5, 180, 232, 5, 5, 145, 169, 5, 5, 160, 217, 5, 5, 213, 233, 5, 5, 3, 234, 5, 5, 45, 170, 5, 5, 223, 238, 5, 5, 141, 201, 5, 5, 124, 218, 5, 5, 107, 178, 5, 5, 249, 179, 5, 5, 165, 184, 5, 5, 201, 192, 5, 5, 133, 218, 5, 5, 126, 218, 5, 5, 97, 173, 5, 5, 60, 225, 5, 5, 211, 225, 5, 5, 212, 225, 5, 5, 189, 184, 5, 5, 224, 214, 5, 5, 14, 222, 5, 5, 69, 177, 5, 5, 199, 246, 5, 5, 195, 228, 5, 5, 40, 207, 5, 5, 172, 230, 5, 5, 56, 182, 5, 5, 48, 240, 5, 5, 112, 226, 5, 5, 231, 208, 5, 5, 190, 229, 5, 5, 73, 247, 5, 5, 13, 175, 5, 5, 127, 230, 5, 5, 101, 192, 5, 5, 86, 230, 5, 5, 112, 231, 5, 5, 76, 193, 5, 5, 6, 241, 5, 5, 146, 228, 5, 5, 157, 176, 5, 5, 146, 223, 5, 5, 141, 199, 5, 5, 17, 242, 5, 5, 201, 170, 5, 5, 200, 246, 5, 5, 108, 214, 5, 5, 30, 177, 5, 5, 251, 173, 5, 5, 254, 224, 5, 5, 206, 227, 5, 5, 95, 215, 5, 5, 223, 210, 5, 5, 163, 191, 5, 5, 69, 234, 5, 5, 90, 211, 5, 5, 186, 199, 5, 5, 232, 231, 5, 5, 216, 233, 5, 5, 160, 198, 5, 5, 147, 189, 5, 5, 118, 198, 5, 5, 57, 183, 5, 5, 78, 190, 5, 5, 2, 225, 5, 5, 167, 223, 5, 5, 87, 230, 5, 5, 225, 211, 5, 5, 160, 173, 5, 5, 253, 175, 5, 5, 64, 197, 5, 5, 249, 238, 5, 5, 101, 169, 5, 5, 176, 173, 5, 5, 232, 208, 5, 5, 104, 212, 5, 5, 38, 250, 5, 5, 97, 244, 5, 5, 101, 237, 5, 5, 70, 197, 5, 5, 35, 169, 5, 5, 145, 211, 5, 5, 65, 210, 5, 5, 23, 246, 5, 5, 242, 211, 5, 5, 16, 172, 5, 5, 181, 226, 5, 5, 244, 189, 5, 5, 240, 189, 5, 5, 71, 197, 5, 5, 126, 177, 5, 5, 238, 202, 5, 5, 66, 223, 5, 5, 53, 175, 5, 5, 22, 179, 5, 5, 182, 177, 5, 5, 111, 238, 5, 5, 82, 227, 5, 5, 32, 241, 5, 5, 170, 183, 5, 5, 114, 192, 5, 5, 148, 212, 5, 5, 50, 234, 5, 5, 238, 242, 5, 5, 126, 210, 5, 5, 185, 186, 5, 5, 216, 183, 5, 5, 113, 231, 5, 5, 54, 170, 5, 5, 9, 241, 5, 5, 70, 216, 5, 5, 64, 235, 5, 5, 160, 176, 5, 5, 102, 171, 5, 5, 233, 190, 5, 5, 205, 234, 5, 5, 159, 175, 5, 5, 55, 170, 5, 5, 198, 222, 5, 5, 187, 230, 5, 5, 15, 225, 5, 5, 227, 240, 5, 5, 133, 249, 5, 5, 224, 180, 5, 5, 148, 209, 5, 5, 174, 246, 5, 5, 250, 238, 5, 5, 28, 222, 5, 5, 119, 210, 5, 5, 168, 235, 5, 5, 29, 222, 5, 5, 118, 231, 5, 5, 111, 201, 5, 5, 10, 189, 5, 5, 178, 199, 5, 5, 154, 209, 5, 5, 179, 246, 5, 5, 108, 233, 5, 5, 52, 203, 5, 5, 137, 226, 5, 5, 234, 190, 5, 5, 183, 179, 5, 5, 120, 188, 5, 5, 64, 199, 5, 5, 78, 187, 5, 5, 11, 189, 5, 5, 188, 230, 5, 5, 199, 194, 5, 5, 202, 222, 5, 5, 68, 223, 5, 5, 152, 189, 5, 5, 11, 217, 5, 5, 102, 237, 5, 5, 78, 191, 5, 5, 46, 191, 5, 5, 232, 174, 5, 5, 82, 193, 5, 5, 120, 238, 5, 5, 72, 234, 5, 5, 186, 186, 5, 5, 21, 204, 5, 5, 5, 191, 5, 5, 88, 249, 5, 5, 208, 234, 5, 5, 212, 174, 5, 5, 188, 220, 5, 5, 154, 211, 5, 5, 2, 189, 5, 5, 213, 181, 5, 5, 4, 199, 5, 5, 57, 180, 5, 5, 88, 226, 5, 5, 215, 185, 5, 5, 200, 217, 5, 5, 194, 230, 5, 5, 23, 215, 5, 5, 16, 243, 5, 5, 150, 212, 5, 5, 25, 199, 5, 5, 158, 178, 5, 5, 79, 191, 5, 5, 151, 181, 5, 5, 91, 233, 5, 5, 52, 227, 5, 5, 51, 236, 5, 5, 99, 198, 5, 5, 161, 172, 5, 5, 83, 210, 5, 5, 119, 243, 5, 5, 48, 207, 5, 5, 237, 240, 5, 5, 234, 240, 5, 5, 144, 242, 5, 5, 218, 212, 5, 5, 83, 200, 5, 5, 239, 215, 5, 5, 17, 243, 5, 5, 35, 242, 5, 5, 101, 227, 5, 5, 152, 195, 5, 5, 202, 230, 5, 5, 128, 245, 5, 5, 22, 204, 5, 5, 52, 226, 5, 5, 32, 188, 5, 5, 121, 188, 5, 5, 74, 216, 5, 5, 56, 226, 5, 5, 25, 172, 5, 5, 20, 239, 5, 5, 151, 215, 5, 5, 58, 191, 5, 5, 195, 230, 5, 5, 175, 169, 5, 5, 244, 206, 5, 5, 71, 238, 5, 5, 5, 189, 5, 5, 168, 223, 5, 5, 201, 217, 5, 5, 139, 174, 5, 5, 232, 199, 5, 5, 134, 230, 5, 5, 135, 230, 5, 5, 13, 241, 5, 5, 255, 201, 5, 5, 104, 203, 5, 5, 157, 190, 5, 5, 161, 176, 5, 5, 34, 239, 5, 5, 18, 243, 5, 5, 130, 202, 5, 5, 50, 211, 5, 5, 84, 210, 5, 5, 158, 181, 5, 5, 24, 218, 5, 5, 45, 235, 5, 5, 53, 215, 5, 5, 136, 230, 5, 5, 9, 209, 5, 5, 195, 176, 5, 5, 211, 182, 5, 5, 152, 169, 5, 5, 161, 178, 5, 5, 40, 177, 5, 5, 166, 173, 5, 5, 235, 207, 5, 5, 143, 202, 5, 5, 227, 192, 5, 5, 226, 186, 5, 5, 221, 186, 5, 5, 55, 234, 5, 5, 136, 178, 5, 5, 74, 214, 5, 5, 56, 199, 5, 5, 116, 226, 5, 5, 85, 206, 5, 5, 218, 230, 5, 5, 142, 198, 5, 5, 128, 187, 5, 5, 182, 210, 5, 5, 191, 216, 5, 5, 57, 175, 5, 5, 242, 180, 5, 5, 182, 187, 5, 5, 207, 244, 5, 5, 24, 196, 5, 5, 121, 228, 5, 5, 152, 242, 5, 5, 182, 195, 5, 5, 233, 192, 5, 5, 214, 200, 5, 5, 36, 242, 5, 5, 220, 191, 5, 5, 188, 188, 5, 5, 221, 217, 5, 5, 32, 225, 5, 5, 57, 226, 5, 5, 60, 226, 5, 5, 5, 211, 5, 5, 182, 228, 5, 5, 76, 176, 5, 5, 157, 189, 5, 5, 154, 191, 5, 5, 203, 230, 5, 5, 152, 173, 5, 5, 225, 233, 5, 5, 250, 228, 5, 5, 86, 191, 5, 5, 166, 181, 5, 5, 156, 224, 5, 5, 136, 249, 5, 5, 144, 193, 5, 5, 239, 240, 5, 5, 111, 179, 5, 5, 90, 197, 5, 5, 94, 172, 5, 5, 114, 174, 5, 5, 162, 178, 5, 5, 185, 169, 5, 5, 77, 216, 5, 5, 87, 172, 5, 5, 20, 178, 5, 5, 222, 186, 5, 5, 81, 181, 5, 5, 89, 210, 5, 5, 123, 243, 5, 5, 206, 232, 5, 5, 155, 248, 5, 5, 196, 179, 5, 5, 199, 190, 5, 5, 64, 176, 5, 5, 238, 216, 5, 5, 117, 218, 5, 5, 7, 234, 5, 5, 173, 172, 5, 5, 191, 170, 5, 5, 13, 209, 5, 5, 104, 249, 5, 5, 3, 226, 5, 5, 41, 216, 5, 5, 64, 175, 5, 5, 190, 169, 5, 5, 12, 235, 5, 5, 211, 228, 5, 5, 121, 185, 5, 5, 66, 215, 5, 5, 51, 229, 5, 5, 63, 242, 5, 5, 168, 241, 5, 5, 85, 170, 5, 5, 13, 235, 5, 5, 164, 190, 5, 5, 16, 209, 5, 5, 86, 170, 5, 5, 65, 239, 5, 5, 164, 208, 5, 5, 240, 240, 5, 5, 124, 198, 5, 5, 85, 178, 5, 5, 192, 239, 5, 5, 167, 181, 5, 5, 13, 174, 5, 5, 124, 207, 5, 5, 27, 215, 5, 5, 12, 199, 5, 5, 179, 241, 5, 5, 10, 168, 5, 5, 80, 216, 5, 5, 158, 236, 5, 5, 178, 211, 5, 5, 227, 184, 5, 5, 124, 243, 5, 5, 137, 249, 5, 5, 154, 219, 5, 5, 243, 201, 5, 5, 60, 183, 5, 5, 57, 240, 5, 5, 27, 200, 5, 5, 28, 200, 5, 5, 35, 218, 5, 5, 222, 185, 5, 5, 84, 243, 5, 5, 7, 224, 5, 5, 14, 224, 5, 5, 151, 214, 5, 5, 145, 237, 5, 5, 172, 223, 5, 5, 248, 190, 5, 5, 136, 192, 5, 5, 96, 186, 5, 5, 102, 197, 5, 5, 5, 176, 5, 5, 184, 210, 5, 5, 121, 233, 5, 5, 109, 198, 5, 5, 63, 246, 5, 5, 247, 240, 5, 5, 150, 172, 5, 5, 218, 235, 5, 5, 103, 221, 5, 5, 200, 171, 5, 5, 125, 198, 5, 5, 247, 242, 5, 5, 131, 231, 5, 5, 183, 191, 5, 5, 249, 240, 5, 5, 92, 237, 5, 5, 161, 202, 5, 5, 211, 219, 5, 5, 167, 225, 5, 5, 200, 239, 5, 5, 114, 176, 5, 5, 232, 234, 5, 5, 157, 215, 5, 5, 10, 225, 5, 5, 188, 190, 5, 5, 126, 243, 5, 5, 104, 221, 5, 5, 54, 209, 5, 5, 247, 185, 5, 5, 2, 221, 5, 5, 198, 176, 5, 5, 94, 213, 5, 5, 208, 209, 5, 5, 212, 209, 5, 5, 174, 187, 5, 5, 255, 174, 5, 5, 96, 172, 5, 5, 91, 172, 5, 5, 89, 172, 5, 5, 17, 177, 5, 5, 21, 209, 5, 5, 216, 225, 5, 5, 25, 245, 5, 5, 55, 227, 5, 5, 149, 168, 5, 5, 7, 223, 5, 5, 23, 207, 5, 5, 228, 186, 5, 5, 38, 218, 5, 5, 30, 244, 5, 5, 203, 196, 5, 5, 107, 191, 5, 5, 105, 198, 5, 5, 248, 202, 5, 5, 226, 211, 5, 5, 175, 223, 5, 5, 133, 216, 5, 5, 39, 240, 5, 5, 20, 206, 5, 5, 167, 220, 5, 5, 111, 229, 5, 5, 153, 227, 5, 5, 248, 206, 5, 5, 71, 215, 5, 5, 6, 221, 5, 5, 198, 240, 5, 5, 219, 216, 5, 5, 167, 198, 5, 5, 59, 179, 5, 5, 148, 246, 5, 5, 39, 205, 5, 5, 170, 197, 5, 5, 93, 214, 5, 5, 89, 214, 5, 5, 86, 242, 5, 5, 199, 213, 5, 5, 125, 203, 5, 5, 211, 176, 5, 5, 16, 241, 5, 5, 23, 175, 5, 5, 28, 247, 5, 5, 56, 227, 5, 5, 66, 174, 5, 5, 255, 214, 5, 5, 117, 217, 5, 5, 180, 212, 5, 5, 202, 169, 5, 5, 29, 171, 5, 5, 253, 216, 5, 5, 177, 194, 5, 5, 105, 244, 5, 5, 250, 174, 5, 5, 7, 203, 5, 5, 202, 213, 5, 5, 36, 200, 5, 5, 108, 191, 5, 5, 25, 216, 5, 5, 128, 174, 5, 5, 41, 240, 5, 5, 42, 240, 5, 5, 89, 231, 5, 5, 4, 231, 5, 5, 38, 178, 5, 5, 40, 225, 5, 5, 88, 181, 5, 5, 231, 180, 5, 5, 230, 180, 5, 5, 21, 224, 5, 5, 241, 197, 5, 5, 182, 172, 5, 5, 23, 233, 5, 5, 62, 182, 5, 5, 63, 183, 5, 5, 96, 201, 5, 5, 130, 201, 5, 5, 24, 175, 5, 5, 238, 187, 5, 5, 3, 215, 5, 5, 48, 232, 5, 5, 22, 209, 5, 5, 66, 196, 5, 5, 163, 203, 5, 5, 107, 230, 5, 5, 97, 172, 5, 5, 231, 197, 5, 5, 235, 176, 5, 5, 145, 232, 5, 5, 68, 225, 5, 5, 143, 221, 5, 5, 241, 212, 5, 5, 107, 239, 5, 5, 185, 175, 5, 5, 255, 231, 5, 5, 76, 210, 5, 5, 39, 178, 5, 5, 69, 225, 5, 5, 68, 196, 5, 5, 162, 223, 5, 5, 55, 188, 5, 5, 234, 194, 5, 5, 117, 229, 5, 5, 29, 235, 5, 5, 47, 180, 5, 5, 143, 216, 5, 5, 138, 216, 5, 5, 112, 197, 5, 5, 151, 172, 5, 5, 31, 199, 5, 5, 189, 233, 5, 5, 83, 240, 5, 5, 154, 168, 5, 5, 133, 207, 5, 5, 108, 239, 5, 5, 13, 204, 5, 5, 248, 219, 5, 5, 68, 198, 5, 5, 84, 190, 5, 5, 238, 207, 5, 5, 11, 201, 5, 5, 19, 168, 5, 5, 14, 204, 5, 5, 78, 236, 5, 5, 119, 200, 5, 5, 220, 231, 5, 5, 5, 175, 5, 5, 3, 242, 5, 5, 44, 240, 5, 5, 166, 177, 5, 5, 31, 208, 5, 5, 22, 168, 5, 5, 6, 208, 5, 5, 235, 180, 5, 5, 173, 214, 5, 5, 90, 209, 5, 5, 226, 200, 5, 5, 239, 207, 5, 5, 78, 175, 5, 5, 46, 247, 5, 5, 187, 211, 5, 5, 188, 211, 5, 5, 76, 236, 5, 5, 97, 237, 5, 5, 223, 171, 5, 5, 51, 247, 5, 5, 221, 199, 5, 5, 197, 199, 5, 5, 85, 241, 5, 5, 199, 183, 5, 5, 169, 204, 5, 5, 221, 208, 5, 5, 135, 174, 5, 5, 149, 191, 5, 5, 100, 173, 5, 5, 34, 208, 5, 5, 15, 201, 5, 5, 86, 190, 5, 5, 55, 235, 5, 5, 99, 218, 5, 5, 102, 173, 5, 5, 21, 193, 5, 5, 124, 197, 5, 5, 103, 190, 5, 5, 13, 221, 5, 5, 159, 239, 5, 5, 73, 203, 5, 5, 45, 210, 5, 5, 69, 208, 5, 5, 149, 225, 5, 5, 11, 198, 5, 5, 32, 185, 5, 5, 33, 185, 5, 5, 141, 248, 5, 5, 84, 185, 5, 5, 8, 243, 5, 5, 117, 230, 5, 5, 127, 193, 5, 5, 139, 234, 5, 5, 192, 222, 5, 5, 193, 218, 5, 5, 107, 231, 5, 5, 87, 174, 5, 5, 247, 229, 5, 5, 140, 195, 5, 5, 89, 185, 5, 5, 135, 193, 5, 5, 178, 215, 5, 5, 208, 191, 5, 5, 179, 215, 5, 5, 251, 244, 5, 5, 64, 180, 5, 5, 75, 214, 5, 5, 84, 193, 5, 5, 191, 179, 5, 5, 90, 244, 5, 5, 87, 193, 5, 5, 183, 192, 5, 5, 107, 246, 5, 5, 133, 198, 5, 5, 128, 192, 5, 5, 253, 199, 5, 5, 181, 184, 5, 5, 187, 178, 5, 5, 6, 245, 5, 5, 195, 215, 5, 5, 103, 185, 5, 5, 236, 193, 5, 5, 106, 195, 5, 5, 76, 242, 5, 5, 237, 193, 5, 5, 172, 236, 5, 5, 186, 205, 5, 5, 166, 189, 5, 5, 10, 245, 5, 5, 145, 231, 5, 5, 147, 231, 5, 5, 70, 176, 5, 5, 203, 177, 5, 5, 162, 217, 5, 5, 229, 189, 5, 5, 230, 189, 5, 5, 231, 189, 5, 5, 134, 181, 5, 5, 17, 222, 5, 5, 46, 226, 5, 5, 100, 207, 5, 5, 241, 189, 5, 5, 111, 202, 5, 5, 153, 182, 5, 5, 150, 224, 5, 5, 187, 170, 5, 5, 107, 179, 5, 5, 121, 196, 5, 5, 178, 220, 5, 5, 130, 238, 5, 5, 194, 238, 5, 5, 124, 220, 5, 5, 249, 189, 5, 5, 187, 182, 5, 5, 159, 236, 5, 5, 95, 222, 5, 5, 117, 222, 5, 5, 49, 172, 5, 5, 121, 244, 5, 5, 37, 217, 5, 5, 142, 201, 5, 5, 8, 214, 5, 5, 165, 168, 5, 5, 154, 177, 5, 5, 180, 218, 5, 5, 111, 182, 5, 5, 51, 219, 5, 5, 189, 243, 5, 5, 223, 227, 5, 5, 49, 245, 5, 5, 150, 179, 5, 5, 175, 198, 5, 5, 114, 241, 5, 5, 91, 199, 5, 5, 218, 184, 5, 5, 21, 221, 5, 5, 199, 172, 5, 5, 224, 227, 5, 5, 77, 186, 5, 5, 77, 199, 5, 5, 121, 230, 5, 5, 171, 178, 5, 5, 35, 215, 5, 5, 122, 246, 5, 5, 152, 218, 5, 5, 101, 200, 5, 5, 222, 207, 5, 5, 224, 219, 5, 5, 38, 237, 5, 5, 91, 211, 5, 5, 32, 169, 5, 5, 220, 173, 5, 5, 111, 218, 5, 5, 173, 230, 5, 5, 96, 215, 5, 5, 254, 168, 5, 5, 77, 193, 5, 5, 133, 241, 5, 5, 210, 183, 5, 5, 140, 168, 5, 5, 178, 230, 5, 5, 209, 212, 5, 5, 169, 246, 5, 5, 201, 246, 5, 5, 139, 181, 5, 5, 172, 178, 5, 5, 84, 245, 5, 5, 90, 174, 5, 5, 202, 192, 5, 5, 140, 236, 5, 5, 187, 199, 5, 5, 202, 170, 5, 5, 174, 173, 5, 5, 214, 196, 5, 5, 191, 229, 5, 5, 171, 189, 5, 5, 12, 243, 5, 5, 189, 197, 5, 5, 204, 168, 5, 5, 160, 216, 5, 5, 246, 177, 5, 5, 58, 246, 5, 5, 99, 243, 5, 5, 103, 228, 5, 5, 104, 210, 5, 5, 239, 238, 5, 5, 121, 222, 5, 5, 70, 248, 5, 5, 248, 213, 5, 5, 65, 227, 5, 5, 85, 180, 5, 5, 182, 198, 5, 5, 117, 245, 5, 5, 249, 213, 5, 5, 41, 184, 5, 5, 75, 212, 5, 5, 188, 168, 5, 5, 131, 168, 5, 5, 42, 244, 5, 5, 150, 227, 5, 5, 67, 199, 5, 5, 70, 206, 5, 5, 226, 215, 5, 5, 145, 243, 5, 5, 236, 189, 5, 5, 125, 169, 5, 5, 103, 171, 5, 5, 175, 246, 5, 5, 149, 212, 5, 5, 23, 210, 5, 5, 17, 172, 5, 5, 213, 212, 5, 5, 232, 224, 5, 5, 179, 237, 5, 5, 222, 245, 5, 5, 123, 244, 5, 5, 27, 237, 5, 5, 126, 169, 5, 5, 149, 188, 5, 5, 156, 210, 5, 5, 4, 238, 5, 5, 4, 179, 5, 5, 230, 174, 5, 5, 214, 212, 5, 5, 42, 193, 5, 5, 89, 209, 5, 5, 202, 207, 5, 5, 238, 173, 5, 5, 231, 235, 5, 5, 45, 175, 5, 5, 102, 217, 5, 5, 4, 209, 5, 5, 160, 175, 5, 5, 44, 193, 5, 5, 217, 183, 5, 5, 156, 244, 5, 5, 8, 248, 5, 5, 217, 177, 5, 5, 243, 172, 5, 5, 177, 209, 5, 5, 193, 210, 5, 5, 125, 200, 5, 5, 46, 184, 5, 5, 76, 212, 5, 5, 36, 169, 5, 5, 244, 211, 5, 5, 218, 203, 5, 5, 221, 209, 5, 5, 179, 186, 5, 5, 97, 215, 5, 5, 65, 197, 5, 5, 195, 224, 5, 5, 191, 168, 5, 5, 230, 227, 5, 5, 231, 227, 5, 5, 133, 168, 5, 5, 217, 196, 5, 5, 188, 248, 5, 5, 34, 212, 5, 5, 72, 245, 5, 5, 246, 168, 5, 5, 244, 168, 5, 5, 5, 179, 5, 5, 155, 210, 5, 5, 72, 197, 5, 5, 102, 200, 5, 5, 14, 205, 5, 5, 203, 193, 5, 5, 15, 215, 5, 5, 188, 240, 5, 5, 211, 200, 5, 5, 55, 211, 5, 5, 142, 171, 5, 5, 44, 244, 5, 5, 49, 215, 5, 5, 249, 188, 5, 5, 103, 200, 5, 5, 30, 222, 5, 5, 75, 195, 5, 5, 60, 246, 5, 5, 51, 211, 5, 5, 235, 185, 5, 5, 236, 185, 5, 5, 215, 217, 5, 5, 240, 222, 5, 5, 67, 177, 5, 5, 153, 243, 5, 5, 193, 198, 5, 5, 121, 238, 5, 5, 109, 233, 5, 5, 162, 172, 5, 5, 6, 191, 5, 5, 131, 213, 5, 5, 75, 248, 5, 5, 218, 221, 5, 5, 222, 209, 5, 5, 227, 168, 5, 5, 184, 174, 5, 5, 168, 186, 5, 5, 217, 246, 5, 5, 104, 200, 5, 5, 56, 181, 5, 5, 57, 181, 5, 5, 180, 246, 5, 5, 75, 216, 5, 5, 74, 168, 5, 5, 122, 211, 5, 5, 249, 245, 5, 5, 114, 185, 5, 5, 254, 194, 5, 5, 131, 199, 5, 5, 58, 180, 5, 5, 223, 209, 5, 5, 138, 226, 5, 5, 185, 203, 5, 5, 125, 244, 5, 5, 239, 205, 5, 5, 184, 179, 5, 5, 47, 228, 5, 5, 195, 197, 5, 5, 179, 203, 5, 5, 42, 197, 5, 5, 225, 246, 5, 5, 32, 206, 5, 5, 34, 236, 5, 5, 232, 229, 5, 5, 196, 224, 5, 5, 110, 233, 5, 5, 66, 210, 5, 5, 65, 178, 5, 5, 200, 194, 5, 5, 28, 246, 5, 5, 179, 192, 5, 5, 7, 191, 5, 5, 25, 232, 5, 5, 230, 241, 5, 5, 0, 137, 5, 5, 233, 227, 5, 5, 169, 211, 5, 5, 97, 177, 5, 5, 185, 171, 5, 5, 191, 185, 5, 5, 40, 226, 5, 5, 252, 245, 5, 5, 103, 174, 5, 5, 130, 219, 5, 5, 131, 219, 5, 5, 41, 199, 5, 5, 189, 207, 5, 5, 123, 211, 5, 5, 222, 196, 5, 5, 163, 212, 5, 5, 209, 193, 5, 5, 21, 239, 5, 5, 240, 226, 5, 5, 86, 220, 5, 5, 63, 219, 5, 5, 117, 228, 5, 5, 111, 233, 5, 5, 203, 172, 5, 5, 120, 183, 5, 5, 204, 194, 5, 5, 94, 230, 5, 5, 53, 198, 5, 5, 192, 196, 5, 5, 47, 227, 5, 5, 84, 200, 5, 5, 213, 191, 5, 5, 102, 227, 5, 5, 191, 248, 5, 5, 250, 213, 5, 5, 241, 205, 5, 5, 176, 168, 5, 5, 33, 188, 5, 5, 194, 220, 5, 5, 247, 210, 5, 5, 129, 197, 5, 5, 86, 244, 5, 5, 203, 222, 5, 5, 235, 237, 5, 5, 203, 197, 5, 5, 5, 172, 5, 5, 107, 228, 5, 5, 26, 172, 5, 5, 19, 250, 5, 5, 9, 238, 5, 5, 184, 244, 5, 5, 240, 205, 5, 5, 189, 223, 5, 5, 234, 227, 5, 5, 79, 201, 5, 5, 122, 224, 5, 5, 50, 169, 5, 5, 210, 193, 5, 5, 153, 190, 5, 5, 119, 178, 5, 5, 38, 229, 5, 5, 148, 228, 5, 5, 120, 216, 5, 5, 198, 212, 5, 5, 243, 220, 5, 5, 196, 203, 5, 5, 241, 208, 5, 5, 227, 207, 5, 5, 56, 184, 5, 5, 235, 168, 5, 5, 91, 197, 5, 5, 114, 178, 5, 5, 249, 229, 5, 5, 0, 146, 5, 5, 146, 197, 5, 5, 24, 243, 5, 5, 135, 249, 5, 5, 240, 173, 5, 5, 13, 176, 5, 5, 87, 195, 5, 5, 86, 227, 5, 5, 224, 169, 5, 5, 241, 209, 5, 5, 206, 210, 5, 5, 111, 219, 5, 5, 24, 250, 5, 5, 3, 230, 5, 5, 76, 214, 5, 5, 242, 231, 5, 5, 107, 174, 5, 5, 68, 179, 5, 5, 225, 219, 5, 5, 71, 206, 5, 5, 214, 216, 5, 5, 16, 185, 5, 5, 32, 181, 5, 5, 105, 222, 5, 5, 159, 179, 5, 5, 251, 213, 5, 5, 9, 179, 5, 5, 43, 245, 5, 5, 179, 191, 5, 5, 180, 192, 5, 5, 162, 225, 5, 5, 13, 215, 5, 5, 133, 214, 5, 5, 247, 211, 5, 5, 133, 222, 5, 5, 61, 215, 5, 5, 66, 204, 5, 5, 227, 237, 5, 5, 39, 236, 5, 5, 207, 197, 5, 5, 36, 246, 5, 5, 95, 203, 5, 5, 170, 186, 5, 5, 35, 239, 5, 5, 217, 191, 5, 5, 127, 220, 5, 5, 61, 246, 5, 5, 61, 206, 5, 5, 62, 172, 5, 5, 77, 225, 5, 5, 229, 173, 5, 5, 97, 171, 5, 5, 34, 195, 5, 5, 58, 226, 5, 5, 65, 199, 5, 5, 167, 227, 5, 5, 152, 236, 5, 5, 100, 177, 5, 5, 140, 247, 5, 5, 232, 196, 5, 5, 131, 226, 5, 5, 105, 215, 5, 5, 33, 199, 5, 5, 228, 168, 5, 5, 236, 211, 5, 5, 35, 195, 5, 5, 121, 205, 5, 5, 177, 186, 5, 5, 59, 209, 5, 5, 88, 195, 5, 5, 237, 246, 5, 5, 253, 177, 5, 5, 42, 208, 5, 5, 103, 172, 5, 5, 210, 219, 5, 5, 223, 186, 5, 5, 202, 212, 5, 5, 244, 242, 5, 5, 176, 211, 5, 5, 219, 193, 5, 5, 42, 246, 5, 5, 161, 196, 5, 5, 220, 193, 5, 5, 169, 241, 5, 5, 113, 236, 5, 5, 250, 199, 5, 5, 39, 210, 5, 5, 34, 189, 5, 5, 248, 218, 5, 5, 46, 213, 5, 5, 212, 228, 5, 5, 155, 219, 5, 5, 185, 249, 5, 5, 14, 235, 5, 5, 184, 208, 5, 5, 4, 226, 5, 5, 248, 210, 5, 5, 206, 172, 5, 5, 53, 222, 5, 5, 138, 249, 5, 5, 6, 246, 5, 5, 78, 238, 5, 5, 181, 235, 5, 5, 220, 240, 5, 5, 168, 170, 5, 5, 62, 237, 5, 5, 165, 190, 5, 5, 160, 236, 5, 5, 207, 243, 5, 5, 69, 168, 5, 5, 154, 234, 5, 5, 41, 236, 5, 5, 7, 230, 5, 5, 240, 198, 5, 5, 196, 175, 5, 5, 119, 192, 5, 5, 61, 226, 5, 5, 134, 200, 5, 5, 141, 200, 5, 5, 14, 174, 5, 5, 92, 198, 5, 5, 162, 196, 5, 5, 163, 196, 5, 5, 109, 227, 5, 5, 46, 195, 5, 5, 17, 191, 5, 5, 168, 185, 5, 5, 43, 175, 5, 5, 15, 233, 5, 5, 197, 179, 5, 5, 93, 233, 5, 5, 190, 242, 5, 5, 106, 215, 5, 5, 236, 237, 5, 5, 207, 172, 5, 5, 131, 244, 5, 5, 153, 169, 5, 5, 139, 237, 5, 5, 213, 228, 5, 5, 6, 170, 5, 5, 4, 201, 5, 5, 206, 229, 5, 5, 164, 216, 5, 5, 10, 173, 5, 5, 100, 185, 5, 5, 56, 205, 5, 5, 139, 249, 5, 5, 169, 185, 5, 5, 212, 194, 5, 5, 34, 186, 5, 5, 165, 216, 5, 5, 211, 218, 5, 5, 31, 218, 5, 5, 47, 175, 5, 5, 197, 175, 5, 5, 238, 244, 5, 5, 126, 224, 5, 5, 115, 224, 5, 5, 201, 171, 5, 5, 176, 175, 5, 5, 206, 218, 5, 5, 72, 169, 5, 5, 73, 177, 5, 5, 212, 219, 5, 5, 222, 198, 5, 5, 146, 237, 5, 5, 122, 178, 5, 5, 110, 246, 5, 5, 147, 211, 5, 5, 129, 200, 5, 5, 156, 193, 5, 5, 194, 223, 5, 5, 87, 217, 5, 5, 59, 185, 5, 5, 82, 232, 5, 5, 48, 223, 5, 5, 202, 219, 5, 5, 237, 195, 5, 5, 215, 208, 5, 5, 176, 181, 5, 5, 15, 176, 5, 5, 179, 211, 5, 5, 98, 220, 5, 5, 201, 224, 5, 5, 160, 244, 5, 5, 110, 225, 5, 5, 20, 212, 5, 5, 6, 169, 5, 5, 115, 177, 5, 5, 162, 202, 5, 5, 168, 225, 5, 5, 184, 189, 5, 5, 10, 247, 5, 5, 55, 228, 5, 5, 42, 190, 5, 5, 72, 215, 5, 5, 207, 229, 5, 5, 228, 215, 5, 5, 100, 226, 5, 5, 195, 245, 5, 5, 177, 181, 5, 5, 122, 233, 5, 5, 181, 211, 5, 5, 230, 217, 5, 5, 221, 172, 5, 5, 178, 244, 5, 5, 131, 185, 5, 5, 99, 230, 5, 5, 217, 181, 5, 5, 3, 221, 5, 5, 176, 198, 5, 5, 4, 221, 5, 5, 143, 222, 5, 5, 240, 168, 5, 5, 147, 237, 5, 5, 78, 171, 5, 5, 200, 223, 5, 5, 19, 225, 5, 5, 146, 219, 5, 5, 78, 173, 5, 5, 142, 224, 5, 5, 165, 196, 5, 5, 21, 175, 5, 5, 239, 175, 5, 5, 178, 186, 5, 5, 114, 171, 5, 5, 126, 183, 5, 5, 233, 222, 5, 5, 54, 179, 5, 5, 250, 214, 5, 5, 202, 223, 5, 5, 187, 244, 5, 5, 39, 203, 5, 5, 67, 174, 5, 5, 162, 174, 5, 5, 229, 186, 5, 5, 181, 205, 5, 5, 159, 206, 5, 5, 59, 205, 5, 5, 142, 249, 5, 5, 176, 184, 5, 5, 4, 190, 5, 5, 135, 244, 5, 5, 249, 175, 5, 5, 155, 225, 5, 5, 67, 190, 5, 5, 18, 177, 5, 5, 252, 209, 5, 5, 72, 209, 5, 5, 98, 194, 5, 5, 48, 187, 5, 5, 166, 240, 5, 5, 145, 246, 5, 5, 114, 168, 5, 5, 29, 247, 5, 5, 15, 211, 5, 5, 24, 207, 5, 5, 99, 173, 5, 5, 72, 199, 5, 5, 106, 175, 5, 5, 7, 221, 5, 5, 154, 227, 5, 5, 222, 194, 5, 5, 62, 209, 5, 5, 68, 209, 5, 5, 137, 245, 5, 5, 106, 172, 5, 5, 47, 199, 5, 5, 242, 170, 5, 5, 144, 194, 5, 5, 152, 237, 5, 5, 41, 186, 5, 5, 73, 215, 5, 5, 174, 203, 5, 5, 44, 192, 5, 5, 67, 240, 5, 5, 177, 197, 5, 5, 123, 213, 5, 5, 122, 213, 5, 5, 81, 201, 5, 5, 250, 180, 5, 5, 142, 232, 5, 5, 65, 219, 5, 5, 113, 187, 5, 5, 209, 243, 5, 5, 89, 239, 5, 5, 49, 232, 5, 5, 70, 174, 5, 5, 71, 174, 5, 5, 132, 219, 5, 5, 74, 210, 5, 5, 100, 189, 5, 5, 90, 223, 5, 5, 48, 188, 5, 5, 22, 187, 5, 5, 118, 177, 5, 5, 17, 250, 5, 5, 212, 210, 5, 5, 10, 204, 5, 5, 64, 246, 5, 5, 30, 191, 5, 5, 140, 248, 5, 5, 230, 194, 5, 5, 187, 192, 5, 5, 134, 172, 5, 5, 2, 178, 5, 5, 3, 178, 5, 5, 233, 173, 5, 5, 164, 171, 5, 5, 222, 173, 5, 5, 178, 197, 5, 5, 67, 184, 5, 5, 117, 203, 5, 5, 229, 169, 5, 5, 68, 184, 5, 5, 54, 216, 5, 5, 165, 171, 5, 5, 77, 177, 5, 5, 226, 248, 5, 5, 108, 248, 5, 5, 66, 229, 5, 5, 12, 214, 5, 5, 166, 216, 5, 5, 247, 180, 5, 5, 213, 210, 5, 5, 43, 190, 5, 5, 128, 233, 5, 5, 72, 205, 5, 5, 231, 194, 5, 5, 247, 176, 5, 5, 207, 224, 5, 5, 56, 188, 5, 5, 55, 216, 5, 5, 244, 229, 5, 5, 241, 193, 5, 5, 231, 184, 5, 5, 204, 240, 5, 5, 164, 201, 5, 5, 57, 210, 5, 5, 132, 205, 5, 5, 141, 220, 5, 5, 227, 248, 5, 5, 67, 244, 5, 5, 14, 214, 5, 5, 86, 176, 5, 5, 63, 192, 5, 5, 68, 178, 5, 5, 249, 219, 5, 5, 135, 172, 5, 5, 200, 216, 5, 5, 198, 216, 5, 5, 176, 190, 5, 5, 54, 195, 5, 5, 135, 216, 5, 5, 136, 199, 5, 5, 237, 177, 5, 5, 129, 233, 5, 5, 83, 198, 5, 5, 40, 213, 5, 5, 229, 168, 5, 5, 156, 168, 5, 5, 113, 197, 5, 5, 31, 238, 5, 5, 192, 181, 5, 5, 242, 207, 5, 5, 217, 223, 5, 5, 58, 208, 5, 5, 188, 192, 5, 5, 247, 224, 5, 5, 236, 248, 5, 5, 131, 178, 5, 5, 45, 234, 5, 5, 13, 201, 5, 5, 41, 172, 5, 5, 52, 197, 5, 5, 242, 237, 5, 5, 49, 219, 5, 5, 46, 238, 5, 5, 42, 238, 5, 5, 168, 210, 5, 5, 15, 230, 5, 5, 119, 195, 5, 5, 82, 171, 5, 5, 70, 211, 5, 5, 107, 185, 5, 5, 47, 247, 5, 5, 37, 247, 5, 5, 114, 200, 5, 5, 80, 209, 5, 5, 73, 194, 5, 5, 137, 233, 5, 5, 211, 203, 5, 5, 22, 225, 5, 5, 241, 168, 5, 5, 219, 223, 5, 5, 137, 205, 5, 5, 69, 206, 5, 5, 113, 218, 5, 5, 247, 225, 5, 5, 12, 212, 5, 5, 98, 237, 5, 5, 225, 201, 5, 5, 42, 172, 5, 5, 157, 222, 5, 5, 211, 243, 5, 5, 218, 210, 5, 5, 96, 232, 5, 5, 76, 198, 5, 5, 194, 191, 5, 5, 198, 202, 5, 5, 150, 200, 5, 5, 146, 190, 5, 5, 147, 240, 5, 5, 96, 205, 5, 5, 20, 205, 5, 5, 81, 215, 5, 5, 82, 215, 5, 5, 212, 243, 5, 5, 87, 215, 5, 5, 237, 200, 5, 5, 11, 232, 5, 5, 93, 240, 5, 5, 181, 207, 5, 5, 95, 218, 5, 5, 17, 173, 5, 5, 226, 229, 5, 5, 251, 176, 5, 5, 140, 233, 5, 5, 14, 221, 5, 5, 127, 206, 5, 5, 77, 198, 5, 5, 70, 208, 5, 5, 174, 174, 5, 5, 214, 243, 5, 5, 44, 206, 5, 5, 23, 225, 5, 5, 224, 249, 5, 5, 220, 202, 5, 5, 84, 179, 5, 5, 60, 228, 5, 5, 71, 178, 5, 5, 248, 194, 5, 5, 12, 198, 5, 5, 19, 201, 5, 5, 227, 202, 5, 5, 59, 210, 5, 5, 80, 246, 5, 5, 94, 187, 5, 5, 79, 187, 5, 5, 80, 214, 5, 5, 188, 235, 5, 5, 6, 214, 5, 5, 7, 214, 5, 5, 94, 222, 5, 5, 191, 198, 5, 5, 7, 241, 5, 5, 166, 184, 5, 5, 185, 238, 5, 5, 208, 185, 5, 5, 192, 184, 5, 5, 255, 168, 5, 5, 174, 182, 5, 5, 70, 246, 5, 5, 232, 213, 5, 5, 67, 179, 5, 5, 78, 199, 5, 5, 7, 209, 5, 5, 131, 230, 5, 5, 124, 186, 5, 5, 174, 188, 5, 5, 165, 172, 5, 5, 75, 233, 5, 5, 85, 208, 5, 5, 167, 175, 5, 5, 118, 185, 5, 5, 222, 178, 5, 5, 217, 234, 5, 5, 6, 195, 5, 5, 10, 209, 5, 5, 8, 174, 5, 5, 196, 196, 5, 5, 44, 221, 5, 5, 41, 181, 5, 5, 240, 241, 5, 5, 237, 174, 5, 5, 103, 168, 5, 5, 247, 168, 5, 5, 218, 234, 5, 5, 7, 195, 5, 5, 42, 181, 5, 5, 35, 203, 5, 5, 9, 211, 5, 5, 76, 170, 5, 5, 154, 173, 5, 5, 92, 179, 5, 5, 33, 181, 5, 5, 51, 239, 5, 5, 223, 184, 5, 5, 194, 219, 5, 5, 249, 198, 5, 5, 220, 236, 5, 5, 243, 180, 5, 5, 26, 192, 5, 5, 72, 227, 5, 5, 81, 233, 5, 5, 34, 181, 5, 5, 216, 194, 5, 5, 98, 196, 5, 5, 64, 237, 5, 5, 189, 231, 5, 5, 17, 209, 5, 5, 212, 222, 5, 5, 0, 151, 5, 5, 247, 215, 5, 5, 0, 152, 5, 5, 65, 246, 5, 5, 238, 178, 5, 5, 8, 246, 5, 5, 192, 183, 5, 5, 219, 222, 5, 5, 133, 203, 5, 5, 119, 217, 5, 5, 86, 234, 5, 5, 209, 238, 5, 5, 235, 194, 5, 5, 142, 220, 5, 5, 236, 194, 5, 5, 237, 248, 5, 5, 122, 239, 5, 5, 43, 203, 5, 5, 140, 170, 5, 5, 46, 202, 5, 5, 89, 233, 5, 5, 90, 233, 5, 5, 161, 229, 5, 5, 99, 235, 5, 5, 113, 214, 5, 5, 122, 214, 5, 5, 185, 244, 5, 5, 55, 171, 5, 5, 208, 197, 5, 5, 186, 244, 5, 5, 46, 201, 5, 5, 212, 182, 5, 5, 5, 169, 5, 5, 7, 169, 5, 5, 224, 200, 5, 5, 5, 242, 5, 5, 243, 200, 5, 5, 86, 229, 5, 5, 91, 180, 5, 5, 118, 221, 5, 5, 171, 203, 5, 5, 104, 193, 5, 5, 179, 189, 5, 5, 117, 233, 5, 5, 110, 193, 5, 5, 242, 241, 5, 5, 196, 245, 5, 5, 13, 195, 5, 5, 13, 230, 5, 5, 170, 226, 5, 5, 108, 180, 5, 5, 167, 195, 5, 5, 225, 174, 5, 5, 217, 239, 5, 5, 47, 184, 5, 5, 180, 215, 5, 5, 231, 244, 5, 5, 138, 217, 5, 5, 208, 248, 5, 5, 236, 244, 5, 5, 191, 180, 5, 5, 102, 177, 5, 5, 71, 223, 5, 5, 229, 233, 5, 5, 219, 248, 5, 5, 228, 248, 5, 5, 136, 216, 5, 5, 240, 203, 5, 5, 238, 248, 5, 5, 186, 175, 5, 5, 205, 180, 5, 5, 21, 248, 5, 5, 138, 182, 5, 5, 66, 173, 5, 5, 72, 188, 5, 5, 138, 241, 5, 5, 175, 221, 5, 5, 126, 212, 5, 5, 33, 241, 5, 5, 204, 207, 5, 5, 61, 212, 5, 5, 123, 214, 5, 5, 202, 244, 5, 5, 80, 207, 5, 5, 242, 205, 5, 5, 127, 212, 5, 5, 223, 212, 5, 5, 138, 204, 5, 5, 179, 183, 5, 5, 168, 182, 5, 5, 46, 235, 5, 5, 25, 196, 5, 5, 26, 196, 5, 5, 160, 210, 5, 5, 196, 249, 5, 5, 97, 245, 5, 5, 200, 238, 5, 5, 145, 204, 5, 5, 200, 220, 5, 5, 35, 194, 5, 5, 180, 241, 5, 5, 206, 238, 5, 5, 158, 214, 5, 5, 38, 247, 5, 5, 64, 182, 5, 5, 97, 213, 5, 5, 65, 182, 5, 5, 222, 244, 5, 5, 157, 199, 5, 5, 103, 224, 5, 5, 213, 241, 5, 5, 49, 230, 5, 5, 194, 192, 5, 5, 210, 192, 5, 5, 228, 192, 5, 5, 221, 191, 5, 5, 186, 218, 5, 5, 8, 178, 5, 5, 185, 196, 5, 5, 164, 246, 5, 5, 12, 244, 5, 5, 101, 233, 5, 5, 134, 226, 5, 5, 161, 235, 5, 5, 198, 234, 5, 5, 136, 184, 5, 5, 131, 200, 5, 5, 237, 184, 5, 5, 29, 188, 5, 5, 229, 224, 5, 5, 19, 179, 5, 5, 140, 234, 5, 5, 64, 173, 5, 5, 209, 221, 5, 5, 203, 199, 5, 5, 42, 237, 5, 5, 214, 221, 5, 5, 197, 228, 5, 5, 233, 208, 5, 5, 234, 208, 5, 5, 196, 227, 5, 5, 25, 176, 5, 5, 88, 230, 5, 5, 73, 243, 5, 5, 167, 169, 5, 5, 93, 168, 5, 5, 84, 244, 5, 5, 18, 169, 5, 5, 69, 195, 5, 5, 52, 200, 5, 5, 119, 221, 5, 5, 242, 189, 5, 5, 163, 182, 5, 5, 116, 188, 5, 5, 97, 187, 5, 5, 107, 173, 5, 5, 3, 235, 5, 5, 30, 209, 5, 5, 150, 191, 5, 5, 254, 182, 5, 5, 148, 216, 5, 5, 148, 189, 5, 5, 251, 238, 5, 5, 180, 230, 5, 5, 219, 233, 5, 5, 90, 236, 5, 5, 85, 244, 5, 5, 164, 182, 5, 5, 27, 225, 5, 5, 95, 221, 5, 5, 73, 197, 5, 5, 53, 237, 5, 5, 204, 243, 5, 5, 106, 171, 5, 5, 4, 234, 5, 5, 176, 240, 5, 5, 66, 235, 5, 5, 233, 213, 5, 5, 225, 245, 5, 5, 52, 204, 5, 5, 27, 176, 5, 5, 117, 188, 5, 5, 205, 207, 5, 5, 21, 250, 5, 5, 103, 209, 5, 5, 212, 170, 5, 5, 209, 234, 5, 5, 151, 191, 5, 5, 73, 245, 5, 5, 149, 249, 5, 5, 31, 222, 5, 5, 32, 222, 5, 5, 21, 242, 5, 5, 226, 182, 5, 5, 185, 179, 5, 5, 103, 207, 5, 5, 176, 210, 5, 5, 151, 173, 5, 5, 140, 229, 5, 5, 25, 180, 5, 5, 0, 157, 5, 5, 107, 171, 5, 5, 94, 168, 5, 5, 178, 247, 5, 5, 218, 204, 5, 5, 70, 232, 5, 5, 207, 199, 5, 5, 165, 226, 5, 5, 185, 173, 5, 5, 224, 221, 5, 5, 237, 190, 5, 5, 238, 190, 5, 5, 58, 235, 5, 5, 243, 199, 5, 5, 146, 234, 5, 5, 205, 194, 5, 5, 226, 195, 5, 5, 226, 246, 5, 5, 227, 195, 5, 5, 179, 220, 5, 5, 13, 227, 5, 5, 95, 189, 5, 5, 209, 236, 5, 5, 156, 184, 5, 5, 204, 232, 5, 5, 69, 220, 5, 5, 61, 233, 5, 5, 10, 238, 5, 5, 50, 243, 5, 5, 11, 191, 5, 5, 255, 187, 5, 5, 37, 188, 5, 5, 57, 198, 5, 5, 122, 228, 5, 5, 125, 232, 5, 5, 58, 200, 5, 5, 184, 247, 5, 5, 204, 230, 5, 5, 104, 174, 5, 5, 152, 221, 5, 5, 250, 171, 5, 5, 129, 245, 5, 5, 130, 245, 5, 5, 139, 230, 5, 5, 123, 228, 5, 5, 83, 191, 5, 5, 124, 188, 5, 5, 9, 174, 5, 5, 129, 228, 5, 5, 118, 226, 5, 5, 220, 248, 5, 5, 241, 249, 5, 5, 166, 247, 5, 5, 41, 214, 5, 5, 62, 196, 5, 5, 219, 230, 5, 5, 115, 220, 5, 5, 161, 210, 5, 5, 220, 230, 5, 5, 193, 216, 5, 5, 216, 214, 5, 5, 30, 196, 5, 5, 86, 187, 5, 5, 62, 246, 5, 5, 52, 239, 5, 5, 2, 247, 5, 5, 82, 168, 5, 5, 130, 228, 5, 5, 227, 203, 5, 5, 108, 203, 5, 5, 114, 173, 5, 5, 184, 228, 5, 5, 62, 233, 5, 5, 217, 243, 5, 5, 188, 182, 5, 5, 17, 235, 5, 5, 202, 185, 5, 5, 150, 238, 5, 5, 190, 231, 5, 5, 53, 243, 5, 5, 21, 191, 5, 5, 191, 234, 5, 5, 18, 209, 5, 5, 254, 199, 5, 5, 229, 237, 5, 5, 177, 240, 5, 5, 171, 222, 5, 5, 57, 229, 5, 5, 170, 222, 5, 5, 197, 216, 5, 5, 125, 207, 5, 5, 34, 210, 5, 5, 229, 193, 5, 5, 159, 211, 5, 5, 83, 168, 5, 5, 65, 237, 5, 5, 31, 176, 5, 5, 153, 237, 5, 5, 159, 224, 5, 5, 44, 214, 5, 5, 44, 209, 5, 5, 223, 194, 5, 5, 112, 198, 5, 5, 61, 185, 5, 5, 223, 229, 5, 5, 169, 173, 5, 5, 251, 214, 5, 5, 132, 188, 5, 5, 234, 236, 5, 5, 169, 202, 5, 5, 13, 168, 5, 5, 244, 192, 5, 5, 254, 192, 5, 5, 236, 207, 5, 5, 218, 243, 5, 5, 185, 233, 5, 5, 133, 188, 5, 5, 213, 209, 5, 5, 126, 209, 5, 5, 170, 176, 5, 5, 188, 210, 5, 5, 31, 245, 5, 5, 109, 191, 5, 5, 138, 169, 5, 5, 51, 188, 5, 5, 51, 235, 5, 5, 218, 175, 5, 5, 137, 203, 5, 5, 252, 231, 5, 5, 70, 225, 5, 5, 69, 196, 5, 5, 124, 213, 5, 5, 241, 203, 5, 5, 202, 227, 5, 5, 90, 231, 5, 5, 109, 239, 5, 5, 6, 193, 5, 5, 251, 190, 5, 5, 200, 177, 5, 5, 26, 238, 5, 5, 27, 238, 5, 5, 176, 202, 5, 5, 45, 225, 5, 5, 37, 227, 5, 5, 66, 233, 5, 5, 241, 182, 5, 5, 87, 221, 5, 5, 105, 245, 5, 5, 137, 188, 5, 5, 123, 239, 5, 5, 210, 232, 5, 5, 18, 234, 5, 5, 60, 221, 5, 5, 239, 194, 5, 5, 143, 169, 5, 5, 110, 196, 5, 5, 246, 236, 5, 5, 23, 168, 5, 5, 43, 238, 5, 5, 36, 219, 5, 5, 179, 222, 5, 5, 8, 208, 5, 5, 150, 235, 5, 5, 216, 237, 5, 5, 59, 214, 5, 5, 189, 202, 5, 5, 26, 174, 5, 5, 222, 199, 5, 5, 219, 179, 5, 5, 165, 203, 5, 5, 2, 237, 5, 5, 242, 191, 5, 5, 101, 205, 5, 5, 36, 231, 5, 5, 234, 218, 5, 5, 245, 204, 5, 5, 58, 210, 5, 5, 145, 206, 5, 5, 45, 206, 5, 5, 71, 220, 5, 5, 150, 225, 5, 5, 193, 236, 5, 5, 254, 247, 5, 5, 253, 242, 5, 5, 254, 242, 5, 5, 90, 217, 5, 5, 253, 237, 5, 5, 200, 185, 5, 5, 255, 237, 5, 5, 149, 189, 5, 5, 165, 188, 5, 5, 130, 222, 5, 5, 138, 172, 5, 5, 137, 172, 5, 5, 130, 221, 5, 5, 19, 207, 5, 5, 189, 172, 5, 5, 192, 172, 5, 5, 62, 226, 5, 5, 240, 249, 5, 5, 95, 172, 5, 5, 227, 234, 5, 5, 95, 191, 5, 5, 14, 240, 5, 5, 82, 216, 5, 5, 15, 183, 5, 5, 26, 213, 5, 5, 7, 243, 5, 5, 63, 241, 5, 5, 68, 219, 5, 5, 162, 212, 5, 5, 22, 183, 5, 5, 218, 183, 5, 5, 53, 204, 5, 5, 206, 182, 5, 5, 139, 217, 5, 5, 50, 226, 5, 5, 155, 211, 5, 5, 166, 226, 5, 5, 43, 223, 5, 5, 253, 245, 5, 5, 65, 201, 5, 5, 64, 201, 5, 5, 227, 249, 5, 5, 36, 209, 5, 5, 186, 190, 5, 5, 199, 228, 5, 5, 199, 227, 5, 5, 188, 173, 5, 5, 27, 192, 5, 5, 81, 214, 5, 5, 60, 240, 5, 5, 110, 249, 5, 5, 200, 228, 5, 5, 38, 227, 5, 5, 71, 201, 5, 5, 97, 201, 5, 5, 9, 208, 5, 5, 246, 204, 5, 5, 189, 209, 5, 5, 185, 178, 5, 5, 89, 229, 5, 5, 91, 209, 5, 5, 215, 169, 5, 5, 143, 244, 5, 5, 190, 222, 5, 5, 191, 222, 5, 5, 191, 209, 5, 5, 200, 247, 5, 5, 125, 218, 5, 5, 166, 168, 5, 5, 37, 214, 5, 5, 51, 181, 5, 5, 52, 181, 5, 5, 108, 178, 5, 5, 101, 202, 5, 5, 69, 187, 5, 5, 4, 192, 5, 5, 151, 196, 5, 5, 25, 170, 5, 5, 113, 234, 5, 5, 88, 174, 5, 5, 138, 176, 5, 5, 16, 220, 5, 5, 52, 219, 5, 5, 160, 243, 5, 5, 208, 217, 5, 5, 41, 215, 5, 5, 134, 241, 5, 5, 193, 184, 5, 5, 31, 230, 5, 5, 200, 172, 5, 5, 78, 220, 5, 5, 164, 235, 5, 5, 78, 182, 5, 5, 124, 230, 5, 5, 63, 249, 5, 5, 68, 202, 5, 5, 47, 234, 5, 5, 50, 172, 5, 5, 59, 177, 5, 5, 155, 218, 5, 5, 230, 170, 5, 5, 227, 227, 5, 5, 20, 179, 5, 5, 51, 245, 5, 5, 44, 207, 5, 5, 226, 174, 5, 5, 240, 238, 5, 5, 173, 184, 5, 5, 209, 185, 5, 5, 168, 180, 5, 5, 248, 201, 5, 5, 205, 214, 5, 5, 193, 222, 5, 5, 4, 185, 5, 5, 142, 226, 5, 5, 153, 194, 5, 5, 135, 208, 5, 5, 147, 228, 5, 5, 161, 200, 5, 5, 175, 196, 5, 5, 45, 207, 5, 5, 43, 237, 5, 5, 197, 206, 5, 5, 191, 208, 5, 5, 119, 223, 5, 5, 177, 242, 5, 5, 73, 188, 5, 5, 222, 182, 5, 5, 149, 169, 5, 5, 72, 195, 5, 5, 23, 180, 5, 5, 56, 185, 5, 5, 180, 237, 5, 5, 234, 231, 5, 5, 142, 175, 5, 5, 26, 176, 5, 5, 234, 211, 5, 5, 124, 222, 5, 5, 55, 190, 5, 5, 220, 233, 5, 5, 91, 175, 5, 5, 30, 248, 5, 5, 92, 175, 5, 5, 147, 223, 5, 5, 19, 169, 5, 5, 148, 223, 5, 5, 105, 192, 5, 5, 254, 229, 5, 5, 222, 195, 5, 5, 22, 186, 5, 5, 10, 192, 5, 5, 76, 207, 5, 5, 249, 212, 5, 5, 56, 170, 5, 5, 179, 228, 5, 5, 95, 168, 5, 5, 143, 182, 5, 5, 23, 183, 5, 5, 252, 238, 5, 5, 219, 183, 5, 5, 24, 210, 5, 5, 181, 230, 5, 5, 243, 189, 5, 5, 232, 235, 5, 5, 86, 180, 5, 5, 239, 233, 5, 5, 223, 245, 5, 5, 115, 237, 5, 5, 221, 203, 5, 5, 93, 219, 5, 5, 126, 181, 5, 5, 144, 207, 5, 5, 90, 245, 5, 5, 181, 187, 5, 5, 88, 246, 5, 5, 158, 176, 5, 5, 117, 243, 5, 5, 254, 249, 5, 5, 120, 221, 5, 5, 125, 222, 5, 5, 13, 244, 5, 5, 23, 179, 5, 5, 112, 202, 5, 5, 78, 205, 5, 5, 204, 193, 5, 5, 91, 174, 5, 5, 149, 223, 5, 5, 181, 215, 5, 5, 79, 183, 5, 5, 199, 244, 5, 5, 242, 232, 5, 5, 238, 231, 5, 5, 95, 199, 5, 5, 177, 213, 5, 5, 234, 224, 5, 5, 75, 231, 5, 5, 181, 246, 5, 5, 180, 186, 5, 5, 245, 232, 5, 5, 120, 193, 5, 5, 48, 193, 5, 5, 13, 186, 5, 5, 86, 169, 5, 5, 104, 209, 5, 5, 11, 239, 5, 5, 6, 238, 5, 5, 11, 238, 5, 5, 33, 222, 5, 5, 252, 210, 5, 5, 32, 170, 5, 5, 76, 181, 5, 5, 122, 238, 5, 5, 54, 204, 5, 5, 108, 171, 5, 5, 159, 210, 5, 5, 126, 200, 5, 5, 166, 188, 5, 5, 38, 169, 5, 5, 79, 182, 5, 5, 77, 247, 5, 5, 183, 177, 5, 5, 106, 176, 5, 5, 84, 237, 5, 5, 171, 183, 5, 5, 233, 168, 5, 5, 168, 209, 5, 5, 201, 184, 5, 5, 67, 214, 5, 5, 76, 218, 5, 5, 245, 218, 5, 5, 113, 207, 5, 5, 213, 220, 5, 5, 151, 223, 5, 5, 160, 245, 5, 5, 174, 231, 5, 5, 99, 241, 5, 5, 32, 221, 5, 5, 98, 187, 5, 5, 34, 228, 5, 5, 170, 244, 5, 5, 27, 210, 5, 5, 58, 211, 5, 5, 235, 240, 5, 5, 6, 179, 5, 5, 218, 246, 5, 5, 126, 244, 5, 5, 217, 172, 5, 5, 15, 178, 5, 5, 52, 186, 5, 5, 20, 172, 5, 5, 189, 196, 5, 5, 99, 168, 5, 5, 227, 183, 5, 5, 206, 193, 5, 5, 172, 168, 5, 5, 77, 181, 5, 5, 206, 198, 5, 5, 12, 210, 5, 5, 36, 248, 5, 5, 61, 170, 5, 5, 183, 204, 5, 5, 248, 172, 5, 5, 87, 220, 5, 5, 124, 223, 5, 5, 163, 175, 5, 5, 106, 212, 5, 5, 22, 222, 5, 5, 181, 186, 5, 5, 127, 244, 5, 5, 184, 237, 5, 5, 54, 174, 5, 5, 190, 196, 5, 5, 34, 222, 5, 5, 93, 246, 5, 5, 184, 204, 5, 5, 147, 207, 5, 5, 119, 202, 5, 5, 195, 218, 5, 5, 171, 244, 5, 5, 14, 244, 5, 5, 231, 170, 5, 5, 252, 244, 5, 5, 219, 246, 5, 5, 219, 204, 5, 5, 59, 180, 5, 5, 81, 205, 5, 5, 124, 221, 5, 5, 120, 202, 5, 5, 212, 200, 5, 5, 232, 240, 5, 5, 199, 222, 5, 5, 174, 235, 5, 5, 242, 222, 5, 5, 239, 214, 5, 5, 181, 245, 5, 5, 70, 214, 5, 5, 131, 202, 5, 5, 131, 238, 5, 5, 193, 232, 5, 5, 254, 245, 5, 5, 132, 202, 5, 5, 235, 219, 5, 5, 146, 186, 5, 5, 128, 198, 5, 5, 211, 169, 5, 5, 135, 218, 5, 5, 76, 233, 5, 5, 234, 168, 5, 5, 136, 218, 5, 5, 131, 171, 5, 5, 6, 249, 5, 5, 56, 175, 5, 5, 22, 239, 5, 5, 146, 176, 5, 5, 187, 234, 5, 5, 207, 247, 5, 5, 24, 194, 5, 5, 237, 249, 5, 5, 228, 181, 5, 5, 250, 181, 5, 5, 70, 241, 5, 5, 35, 182, 5, 5, 237, 185, 5, 5, 195, 198, 5, 5, 114, 201, 5, 5, 203, 244, 5, 5, 192, 203, 5, 5, 179, 239, 5, 5, 103, 237, 5, 5, 175, 188, 5, 5, 161, 185, 5, 5, 23, 239, 5, 5, 38, 222, 5, 5, 119, 185, 5, 5, 178, 243, 5, 5, 33, 206, 5, 5, 240, 183, 5, 5, 80, 195, 5, 5, 12, 189, 5, 5, 101, 187, 5, 5, 187, 225, 5, 5, 5, 187, 5, 5, 246, 228, 5, 5, 180, 199, 5, 5, 19, 219, 5, 5, 77, 168, 5, 5, 37, 168, 5, 5, 156, 197, 5, 5, 132, 238, 5, 5, 192, 248, 5, 5, 97, 199, 5, 5, 227, 246, 5, 5, 234, 216, 5, 5, 14, 227, 5, 5, 199, 219, 5, 5, 198, 219, 5, 5, 115, 190, 5, 5, 5, 197, 5, 5, 193, 196, 5, 5, 128, 235, 5, 5, 78, 181, 5, 5, 151, 248, 5, 5, 153, 241, 5, 5, 205, 243, 192, 0, 0, 0, 5, 5, 51, 240, 5, 5, 81, 195, 5, 5, 185, 204, 5, 5, 253, 244, 5, 5, 235, 235, 5, 5, 105, 218, 5, 5, 182, 245, 5, 5, 82, 178, 5, 5, 71, 214, 5, 5, 16, 216, 5, 5, 59, 190, 5, 5, 102, 187, 5, 5, 118, 194, 5, 5, 124, 248, 5, 5, 175, 235, 5, 5, 136, 224, 5, 5, 29, 220, 5, 5, 189, 245, 5, 5, 151, 169, 5, 5, 207, 226, 5, 5, 105, 200, 5, 5, 92, 196, 5, 5, 245, 213, 5, 5, 225, 169, 5, 5, 246, 183, 5, 5, 81, 219, 5, 5, 48, 227, 5, 5, 209, 197, 5, 5, 205, 230, 5, 5, 47, 201, 5, 5, 187, 204, 5, 5, 88, 183, 5, 5, 77, 214, 5, 5, 195, 229, 5, 5, 31, 198, 5, 5, 224, 184, 5, 5, 249, 223, 5, 5, 78, 203, 5, 5, 59, 217, 5, 5, 242, 226, 5, 5, 73, 241, 5, 5, 152, 207, 5, 5, 51, 169, 5, 5, 24, 205, 5, 5, 152, 212, 5, 5, 126, 248, 5, 5, 32, 179, 5, 5, 7, 235, 5, 5, 107, 227, 5, 5, 28, 244, 5, 5, 100, 168, 5, 5, 130, 186, 5, 5, 77, 170, 5, 5, 225, 178, 5, 5, 2, 188, 5, 5, 73, 249, 5, 5, 106, 246, 5, 5, 169, 218, 5, 5, 180, 169, 5, 5, 195, 185, 5, 5, 214, 193, 5, 5, 158, 190, 5, 5, 124, 228, 5, 5, 170, 211, 5, 5, 56, 193, 5, 5, 147, 226, 5, 5, 229, 192, 5, 5, 3, 233, 5, 5, 243, 205, 5, 5, 165, 191, 5, 5, 197, 220, 5, 5, 188, 172, 5, 5, 29, 183, 5, 5, 156, 223, 5, 5, 38, 208, 5, 5, 63, 230, 5, 5, 2, 202, 5, 5, 3, 202, 5, 5, 99, 180, 5, 5, 150, 216, 5, 5, 162, 240, 5, 5, 137, 224, 5, 5, 233, 196, 5, 5, 244, 225, 5, 5, 166, 233, 5, 5, 87, 200, 5, 5, 218, 248, 5, 5, 134, 222, 5, 5, 9, 173, 5, 5, 127, 182, 5, 5, 45, 229, 5, 5, 93, 196, 5, 5, 4, 202, 5, 5, 53, 171, 5, 5, 217, 231, 5, 5, 151, 183, 5, 5, 190, 225, 5, 5, 238, 246, 5, 5, 174, 200, 5, 5, 242, 202, 5, 5, 215, 193, 5, 5, 193, 248, 5, 5, 67, 204, 5, 5, 9, 202, 5, 5, 245, 214, 5, 5, 115, 171, 5, 5, 88, 206, 5, 5, 162, 176, 5, 5, 155, 215, 5, 5, 140, 208, 5, 5, 145, 214, 5, 5, 146, 214, 5, 5, 69, 172, 5, 5, 142, 187, 5, 5, 20, 173, 5, 5, 163, 178, 5, 5, 213, 182, 5, 5, 249, 211, 5, 5, 68, 169, 5, 5, 69, 169, 5, 5, 157, 191, 5, 5, 105, 249, 5, 5, 115, 174, 5, 5, 16, 244, 5, 5, 129, 192, 5, 5, 153, 202, 5, 5, 170, 212, 5, 5, 51, 242, 5, 5, 52, 242, 5, 5, 132, 186, 5, 5, 60, 198, 5, 5, 68, 180, 5, 5, 107, 225, 5, 5, 18, 185, 5, 5, 185, 228, 5, 5, 42, 179, 5, 5, 103, 177, 5, 5, 87, 182, 5, 5, 56, 174, 5, 5, 4, 245, 5, 5, 217, 214, 5, 5, 155, 242, 5, 5, 161, 236, 5, 5, 53, 242, 5, 5, 204, 217, 5, 5, 53, 239, 5, 5, 10, 220, 5, 5, 139, 218, 5, 5, 16, 176, 5, 5, 234, 201, 5, 5, 82, 214, 5, 5, 195, 248, 5, 5, 6, 184, 5, 5, 217, 198, 5, 5, 179, 200, 5, 5, 168, 249, 5, 5, 169, 249, 5, 5, 98, 245, 5, 5, 193, 186, 5, 5, 9, 183, 5, 5, 30, 183, 5, 5, 48, 221, 5, 5, 194, 216, 5, 5, 139, 210, 5, 5, 131, 228, 5, 5, 187, 187, 5, 5, 166, 205, 5, 5, 96, 188, 5, 5, 47, 195, 5, 5, 201, 238, 5, 5, 61, 175, 5, 5, 43, 197, 5, 5, 13, 197, 5, 5, 116, 174, 5, 5, 25, 250, 5, 5, 80, 203, 5, 5, 187, 215, 5, 5, 135, 246, 5, 5, 17, 176, 5, 5, 240, 235, 5, 5, 236, 196, 5, 5, 154, 169, 5, 5, 166, 194, 5, 5, 221, 248, 5, 5, 20, 249, 5, 5, 56, 171, 5, 5, 171, 212, 5, 5, 4, 180, 5, 5, 144, 175, 5, 5, 115, 173, 5, 5, 220, 207, 5, 5, 25, 190, 5, 5, 221, 193, 5, 5, 44, 187, 5, 5, 126, 231, 5, 5, 126, 180, 5, 5, 156, 215, 5, 5, 140, 178, 5, 5, 103, 187, 5, 5, 93, 179, 5, 5, 113, 206, 5, 5, 108, 246, 5, 5, 15, 218, 5, 5, 57, 209, 5, 5, 58, 184, 5, 5, 176, 185, 5, 5, 203, 212, 5, 5, 131, 220, 5, 5, 142, 238, 5, 5, 27, 228, 5, 5, 11, 220, 5, 5, 60, 181, 5, 5, 228, 237, 5, 5, 109, 184, 5, 5, 58, 229, 5, 5, 218, 228, 5, 5, 195, 180, 5, 5, 111, 193, 5, 5, 109, 249, 5, 5, 157, 193, 5, 5, 151, 238, 5, 5, 105, 221, 5, 5, 54, 231, 5, 5, 228, 236, 5, 5, 167, 236, 5, 5, 217, 175, 5, 5, 158, 193, 5, 5, 32, 176, 5, 5, 246, 241, 5, 5, 192, 188, 5, 5, 133, 244, 5, 5, 11, 230, 5, 5, 75, 213, 5, 5, 11, 170, 5, 5, 121, 237, 5, 5, 222, 191, 5, 5, 159, 234, 5, 5, 119, 219, 5, 5, 66, 237, 5, 5, 142, 200, 5, 5, 114, 236, 5, 5, 220, 169, 5, 5, 22, 191, 5, 5, 255, 199, 5, 5, 156, 195, 5, 5, 2, 200, 5, 5, 75, 223, 5, 5, 95, 183, 5, 5, 95, 233, 5, 5, 28, 228, 5, 5, 239, 246, 5, 5, 36, 194, 5, 5, 209, 209, 5, 5, 126, 207, 5, 5, 147, 175, 5, 5, 251, 189, 5, 5, 189, 189, 5, 5, 61, 203, 5, 5, 235, 201, 5, 5, 228, 226, 5, 5, 35, 210, 5, 5, 181, 241, 5, 5, 50, 241, 5, 5, 164, 207, 5, 5, 173, 223, 5, 5, 77, 235, 5, 5, 78, 235, 5, 5, 90, 237, 5, 5, 197, 245, 5, 5, 76, 213, 5, 5, 22, 238, 5, 5, 137, 192, 5, 5, 96, 195, 5, 5, 23, 238, 5, 5, 146, 175, 5, 5, 18, 181, 5, 5, 170, 241, 5, 5, 186, 249, 5, 5, 219, 228, 5, 5, 165, 207, 5, 5, 63, 226, 5, 5, 130, 192, 5, 5, 97, 195, 5, 5, 110, 198, 5, 5, 17, 217, 5, 5, 126, 240, 5, 5, 10, 219, 5, 5, 204, 190, 5, 5, 36, 205, 5, 5, 149, 201, 5, 5, 231, 217, 5, 5, 173, 232, 5, 5, 152, 213, 5, 5, 254, 221, 5, 5, 182, 184, 5, 5, 34, 225, 5, 5, 6, 201, 5, 5, 143, 229, 5, 5, 182, 241, 5, 5, 43, 174, 5, 5, 225, 205, 5, 5, 45, 197, 5, 5, 68, 221, 5, 5, 192, 175, 5, 5, 12, 170, 5, 5, 180, 233, 5, 5, 114, 193, 5, 5, 80, 239, 5, 5, 242, 244, 5, 5, 20, 184, 5, 5, 182, 211, 5, 5, 117, 208, 5, 5, 50, 201, 5, 5, 215, 218, 5, 5, 107, 186, 5, 5, 47, 194, 5, 5, 46, 197, 5, 5, 182, 224, 5, 5, 194, 237, 5, 5, 205, 245, 5, 5, 61, 169, 5, 5, 37, 220, 5, 5, 207, 242, 5, 5, 79, 249, 5, 5, 45, 209, 5, 5, 15, 224, 5, 5, 123, 193, 5, 5, 154, 237, 5, 5, 107, 195, 5, 5, 250, 190, 5, 5, 248, 184, 5, 5, 214, 182, 5, 5, 179, 244, 5, 5, 113, 215, 5, 5, 206, 206, 5, 5, 127, 224, 5, 5, 208, 242, 5, 5, 181, 233, 5, 5, 216, 218, 5, 5, 235, 221, 5, 5, 114, 246, 5, 5, 19, 177, 5, 5, 148, 229, 5, 5, 229, 226, 5, 5, 151, 204, 5, 5, 212, 218, 5, 5, 114, 225, 5, 5, 36, 218, 5, 5, 188, 244, 5, 5, 81, 223, 5, 5, 162, 221, 5, 5, 204, 224, 5, 5, 227, 198, 5, 5, 239, 230, 5, 5, 101, 186, 5, 5, 90, 214, 5, 5, 62, 185, 5, 5, 63, 185, 5, 5, 119, 224, 5, 5, 21, 212, 5, 5, 170, 225, 5, 5, 132, 185, 5, 5, 33, 176, 5, 5, 73, 179, 5, 5, 128, 211, 5, 5, 141, 192, 5, 5, 50, 223, 5, 5, 38, 186, 5, 5, 19, 176, 5, 5, 196, 215, 5, 5, 224, 172, 5, 5, 137, 215, 5, 5, 81, 190, 5, 5, 171, 207, 5, 5, 237, 234, 5, 5, 31, 185, 5, 5, 42, 185, 5, 5, 196, 248, 5, 5, 11, 228, 5, 5, 21, 216, 5, 5, 106, 237, 5, 5, 74, 179, 5, 5, 115, 193, 5, 5, 146, 198, 5, 5, 246, 249, 5, 5, 126, 178, 5, 5, 231, 204, 5, 5, 59, 171, 5, 5, 219, 247, 5, 5, 200, 219, 5, 5, 61, 231, 5, 5, 45, 192, 5, 5, 254, 202, 5, 5, 110, 191, 5, 5, 200, 240, 5, 5, 165, 215, 5, 5, 189, 187, 5, 5, 183, 184, 5, 5, 186, 184, 5, 5, 156, 227, 5, 5, 70, 190, 5, 5, 94, 214, 5, 5, 12, 220, 5, 5, 21, 177, 5, 5, 177, 212, 5, 5, 81, 241, 5, 5, 195, 189, 5, 5, 124, 194, 5, 5, 7, 190, 5, 5, 174, 190, 5, 5, 114, 187, 5, 5, 16, 211, 5, 5, 90, 239, 5, 5, 45, 185, 5, 5, 169, 198, 5, 5, 52, 187, 5, 5, 53, 187, 5, 5, 143, 172, 5, 5, 30, 207, 5, 5, 210, 195, 5, 5, 239, 178, 5, 5, 131, 248, 5, 5, 151, 201, 5, 5, 69, 201, 5, 5, 19, 174, 5, 5, 171, 176, 5, 5, 22, 202, 5, 5, 101, 234, 5, 5, 199, 216, 5, 5, 9, 223, 5, 5, 92, 182, 5, 5, 2, 227, 5, 5, 232, 186, 5, 5, 63, 244, 5, 5, 22, 224, 5, 5, 196, 201, 5, 5, 129, 205, 5, 5, 87, 203, 5, 5, 125, 208, 5, 5, 43, 205, 5, 5, 193, 173, 5, 5, 23, 224, 5, 5, 230, 198, 5, 5, 107, 175, 5, 5, 120, 225, 5, 5, 247, 170, 5, 5, 187, 205, 5, 5, 166, 196, 5, 5, 165, 245, 5, 5, 137, 244, 5, 5, 151, 222, 5, 5, 32, 245, 5, 5, 250, 206, 5, 5, 69, 209, 5, 5, 241, 210, 5, 5, 108, 237, 5, 5, 154, 226, 5, 5, 178, 212, 5, 5, 63, 248, 5, 5, 38, 220, 5, 5, 5, 231, 5, 5, 239, 217, 5, 5, 16, 189, 5, 5, 168, 193, 5, 5, 172, 176, 5, 5, 46, 192, 5, 5, 72, 236, 5, 5, 215, 215, 5, 5, 119, 235, 5, 5, 74, 240, 5, 5, 18, 182, 5, 5, 189, 235, 5, 5, 151, 246, 5, 5, 28, 216, 5, 5, 229, 249, 5, 5, 205, 176, 5, 5, 47, 214, 5, 5, 222, 222, 5, 5, 68, 190, 5, 5, 42, 200, 5, 5, 212, 245, 5, 5, 7, 250, 5, 5, 35, 243, 5, 5, 143, 220, 5, 5, 6, 231, 5, 5, 37, 176, 5, 5, 120, 179, 5, 5, 44, 182, 5, 5, 230, 184, 5, 5, 74, 209, 5, 5, 110, 230, 5, 5, 2, 216, 5, 5, 108, 218, 5, 5, 28, 204, 5, 5, 161, 204, 5, 5, 29, 216, 5, 5, 146, 232, 5, 5, 109, 219, 5, 5, 93, 182, 5, 5, 246, 244, 5, 5, 15, 228, 5, 5, 121, 201, 5, 5, 93, 243, 5, 5, 33, 223, 5, 5, 251, 180, 5, 5, 73, 174, 5, 5, 124, 225, 5, 5, 253, 207, 5, 5, 24, 197, 5, 5, 131, 174, 5, 5, 30, 224, 5, 5, 242, 197, 5, 5, 243, 197, 5, 5, 121, 179, 5, 5, 116, 191, 5, 5, 55, 192, 5, 5, 173, 211, 5, 5, 233, 232, 5, 5, 30, 228, 5, 5, 78, 211, 5, 5, 87, 219, 5, 5, 223, 247, 5, 5, 39, 227, 5, 5, 107, 244, 5, 5, 40, 183, 5, 5, 254, 216, 5, 5, 80, 218, 5, 5, 17, 189, 5, 5, 120, 215, 5, 5, 71, 186, 5, 5, 195, 204, 5, 5, 133, 201, 5, 5, 46, 185, 5, 5, 130, 175, 5, 5, 91, 231, 5, 5, 164, 221, 5, 5, 80, 249, 5, 5, 197, 219, 5, 5, 154, 192, 5, 5, 81, 180, 5, 5, 40, 196, 5, 5, 133, 205, 5, 5, 87, 194, 5, 5, 193, 175, 5, 5, 215, 242, 5, 5, 211, 224, 5, 5, 154, 222, 5, 5, 106, 194, 5, 5, 49, 225, 5, 5, 16, 204, 5, 5, 112, 211, 5, 5, 18, 240, 5, 5, 64, 231, 5, 5, 126, 191, 5, 5, 104, 220, 5, 5, 250, 249, 5, 5, 53, 235, 5, 5, 77, 174, 5, 5, 249, 184, 5, 5, 25, 197, 5, 5, 253, 249, 5, 5, 124, 239, 5, 5, 139, 216, 5, 5, 49, 214, 5, 5, 122, 236, 5, 5, 165, 201, 5, 5, 104, 183, 5, 5, 62, 191, 5, 5, 97, 178, 5, 5, 11, 193, 5, 5, 94, 224, 5, 5, 240, 171, 5, 5, 204, 213, 5, 5, 132, 174, 5, 5, 151, 175, 5, 5, 71, 248, 5, 5, 115, 187, 5, 5, 155, 192, 5, 5, 134, 195, 5, 5, 117, 193, 5, 5, 201, 216, 5, 5, 191, 244, 5, 5, 242, 193, 5, 5, 222, 215, 5, 5, 150, 178, 5, 5, 211, 238, 5, 5, 25, 171, 5, 5, 164, 223, 5, 5, 236, 220, 5, 5, 17, 204, 5, 5, 182, 202, 5, 5, 228, 172, 5, 5, 10, 208, 5, 5, 248, 239, 5, 5, 204, 225, 5, 5, 249, 224, 5, 5, 146, 208, 5, 5, 176, 214, 5, 5, 158, 227, 5, 5, 71, 171, 5, 5, 230, 191, 5, 5, 17, 193, 5, 5, 80, 215, 5, 5, 170, 210, 5, 5, 73, 211, 5, 5, 92, 238, 5, 5, 69, 185, 5, 5, 148, 198, 5, 5, 45, 240, 5, 5, 131, 211, 5, 5, 211, 216, 5, 5, 178, 236, 5, 5, 177, 214, 5, 5, 130, 208, 5, 5, 108, 245, 5, 5, 119, 187, 5, 5, 39, 176, 5, 5, 70, 192, 5, 5, 7, 200, 5, 5, 223, 213, 5, 5, 197, 178, 5, 5, 132, 175, 5, 5, 144, 185, 5, 5, 147, 208, 5, 5, 21, 241, 5, 5, 53, 247, 5, 5, 246, 190, 5, 5, 83, 215, 5, 5, 205, 201, 5, 5, 169, 201, 5, 5, 151, 219, 5, 5, 139, 205, 5, 5, 190, 202, 5, 5, 254, 176, 5, 5, 26, 206, 5, 5, 223, 208, 5, 5, 137, 191, 5, 5, 211, 211, 5, 5, 230, 205, 5, 5, 48, 247, 5, 5, 50, 185, 5, 5, 139, 180, 5, 5, 223, 242, 5, 5, 199, 202, 5, 5, 245, 182, 5, 5, 240, 248, 5, 5, 221, 223, 5, 5, 20, 203, 5, 5, 151, 194, 5, 5, 137, 175, 5, 5, 225, 216, 5, 5, 231, 247, 5, 5, 102, 205, 5, 5, 129, 236, 5, 5, 205, 202, 5, 5, 230, 247, 5, 5, 49, 174, 5, 5, 124, 195, 5, 5, 202, 181, 5, 5, 234, 223, 5, 5, 88, 190, 5, 5, 33, 211, 5, 5, 118, 242, 5, 5, 247, 204, 5, 5, 205, 200, 5, 5, 30, 195, 5, 5, 97, 232, 5, 5, 66, 187, 5, 5, 53, 197, 5, 5, 43, 233, 5, 5, 111, 204, 5, 5, 97, 240, 5, 5, 193, 193, 5, 5, 26, 240, 5, 5, 61, 241, 5, 5, 152, 240, 5, 5, 183, 232, 5, 5, 119, 211, 5, 5, 232, 171, 5, 5, 53, 173, 5, 5, 244, 200, 5, 5, 55, 197, 5, 5, 252, 222, 5, 5, 17, 221, 5, 5, 24, 229, 5, 5, 209, 176, 5, 5, 250, 217, 5, 5, 163, 217, 5, 5, 131, 172, 5, 5, 210, 196, 5, 5, 124, 242, 5, 5, 131, 206, 5, 5, 224, 202, 5, 5, 2, 177, 5, 5, 46, 206, 5, 5, 73, 178, 5, 5, 15, 198, 5, 5, 135, 236, 5, 5, 22, 201, 5, 5, 250, 200, 5, 5, 255, 247, 5, 5, 184, 201, 5, 5, 94, 202, 5, 5, 224, 168, 5, 5, 55, 210, 5, 5, 131, 242, 5, 5, 114, 204, 5, 5, 23, 187, 5, 5, 145, 215, 5, 5, 139, 176, 5, 5, 96, 190, 5, 5, 221, 233, 5, 5, 135, 241, 5, 5, 253, 238, 5, 5, 45, 215, 5, 5, 199, 211, 5, 5, 143, 234, 5, 5, 177, 173, 5, 5, 164, 175, 5, 5, 243, 214, 5, 5, 121, 198, 5, 5, 24, 239, 5, 5, 204, 197, 5, 5, 121, 231, 5, 5, 212, 234, 5, 5, 150, 188, 5, 5, 37, 242, 5, 5, 24, 200, 5, 5, 41, 201, 5, 5, 167, 199, 5, 5, 45, 223, 5, 5, 206, 230, 5, 5, 0, 138, 5, 5, 79, 238, 5, 5, 83, 214, 5, 5, 194, 175, 5, 5, 216, 174, 5, 5, 102, 216, 5, 5, 168, 199, 5, 5, 143, 198, 5, 5, 169, 199, 5, 5, 145, 198, 5, 5, 224, 175, 5, 5, 55, 220, 5, 5, 148, 186, 5, 5, 195, 239, 5, 5, 230, 233, 5, 5, 96, 233, 5, 5, 183, 241, 5, 5, 161, 215, 5, 5, 18, 233, 5, 5, 28, 238, 5, 5, 104, 185, 5, 5, 42, 230, 5, 5, 84, 225, 5, 5, 246, 195, 5, 5, 205, 211, 5, 5, 167, 189, 5, 5, 66, 226, 5, 5, 98, 190, 5, 5, 168, 234, 5, 5, 139, 212, 5, 5, 92, 231, 5, 5, 87, 233, 5, 5, 195, 175, 5, 5, 9, 221, 5, 5, 144, 220, 5, 5, 240, 187, 5, 5, 187, 175, 5, 5, 125, 239, 5, 5, 199, 181, 5, 5, 205, 241, 5, 5, 89, 176, 5, 5, 105, 190, 5, 5, 160, 246, 5, 5, 66, 246, 5, 5, 132, 176, 5, 5, 14, 172, 5, 5, 89, 230, 5, 5, 114, 214, 5, 5, 18, 172, 5, 5, 19, 172, 5, 5, 77, 228, 5, 5, 74, 197, 5, 5, 100, 215, 5, 5, 187, 174, 5, 5, 241, 219, 5, 5, 217, 174, 5, 5, 245, 219, 5, 5, 104, 247, 5, 5, 83, 224, 5, 5, 84, 224, 5, 5, 170, 202, 5, 5, 64, 244, 5, 5, 194, 241, 5, 5, 177, 202, 5, 5, 61, 187, 5, 5, 167, 177, 5, 5, 135, 181, 5, 5, 106, 223, 5, 5, 132, 193, 5, 5, 118, 245, 5, 5, 96, 209, 5, 5, 97, 209, 5, 5, 116, 237, 5, 5, 98, 209, 5, 5, 224, 176, 5, 5, 29, 237, 5, 5, 113, 226, 5, 5, 125, 221, 5, 5, 184, 177, 5, 5, 162, 220, 5, 5, 213, 234, 5, 5, 209, 235, 5, 5, 131, 222, 5, 5, 85, 172, 5, 5, 210, 197, 5, 5, 108, 213, 5, 5, 25, 215, 5, 5, 60, 217, 5, 5, 250, 223, 5, 5, 192, 216, 5, 5, 103, 243, 5, 5, 54, 203, 5, 5, 54, 239, 5, 5, 159, 183, 5, 5, 136, 246, 5, 5, 17, 238, 5, 5, 88, 172, 5, 5, 158, 191, 5, 5, 228, 177, 5, 5, 131, 192, 5, 5, 206, 179, 5, 5, 206, 245, 5, 5, 107, 243, 5, 5, 144, 229, 5, 5, 102, 175, 5, 5, 79, 171, 5, 5, 67, 226, 5, 5, 247, 195, 5, 5, 171, 220, 5, 5, 244, 239, 5, 5, 141, 179, 5, 5, 204, 196, 5, 5, 117, 191, 5, 5, 5, 177, 5, 5, 110, 239, 5, 5, 238, 177, 5, 5, 178, 180, 5, 5, 107, 194, 5, 5, 68, 203, 5, 5, 84, 171, 5, 5, 140, 180, 5, 5, 187, 193, 5, 5, 194, 193, 5, 5, 119, 222, 5, 5, 200, 211, 5, 5, 190, 180, 5, 5, 37, 248, 5, 5, 180, 239, 5, 5, 215, 216, 5, 5, 36, 239, 5, 5, 24, 220, 5, 5, 49, 216, 5, 5, 241, 198, 5, 5, 57, 233, 5, 5, 215, 235, 5, 5, 115, 179, 5, 5, 59, 191, 5, 5, 60, 191, 5, 5, 102, 186, 5, 5, 248, 215, 5, 5, 40, 192, 5, 5, 91, 239, 5, 5, 206, 211, 5, 5, 61, 191, 5, 5, 202, 180, 5, 5, 93, 238, 5, 5, 50, 233, 5, 5, 50, 230, 5, 5, 217, 186, 5, 5, 172, 209, 5, 5, 185, 207, 5, 5, 186, 207, 5, 5, 0, 156, 5, 5, 129, 195, 5, 5, 121, 180, 5, 5, 77, 242, 5, 5, 23, 170, 5, 5, 46, 170, 5, 5, 62, 170, 5, 5, 253, 212, 5, 5, 254, 212, 5, 5, 67, 170, 5, 5, 40, 173, 5, 5, 74, 207, 5, 5, 92, 188, 5, 5, 55, 172, 5, 5, 250, 212, 5, 5, 177, 203, 5, 5, 49, 193, 5, 5, 201, 244, 5, 5, 153, 219, 5, 5, 199, 209, 5, 5, 35, 228, 5, 5, 176, 235, 5, 5, 251, 181, 5, 5, 202, 218, 5, 5, 73, 232, 5, 5, 6, 197, 5, 5, 169, 209, 5, 5, 95, 188, 5, 5, 61, 217, 5, 5, 100, 180, 5, 5, 30, 220, 5, 5, 65, 225, 5, 5, 113, 212, 5, 5, 14, 197, 5, 5, 35, 181, 5, 5, 43, 177, 5, 5, 9, 170, 5, 5, 172, 219, 5, 5, 173, 219, 5, 5, 127, 207, 5, 5, 157, 219, 5, 5, 144, 222, 5, 5, 145, 222, 5, 5, 39, 228, 5, 5, 193, 188, 5, 5, 37, 194, 5, 5, 205, 224, 5, 5, 175, 219, 5, 5, 228, 205, 5, 5, 170, 209, 5, 5, 89, 207, 5, 5, 40, 227, 5, 5, 241, 218, 5, 5, 158, 173, 5, 5, 48, 214, 5, 5, 222, 205, 5, 5, 218, 244, 5, 5, 229, 219, 5, 5, 219, 244, 5, 5, 11, 208, 5, 5, 140, 205, 5, 5, 164, 217, 5, 5, 223, 179, 5, 5, 4, 222, 5, 5, 199, 178, 5, 5, 228, 208, 5, 5, 186, 197, 5, 5, 46, 207, 5, 5, 222, 214, 5, 5, 121, 213, 5, 5, 3, 210, 5, 5, 225, 214, 5, 5, 109, 178, 5, 5, 233, 231, 5, 5, 205, 175, 5, 5, 255, 182, 5, 5, 48, 237, 5, 5, 117, 210, 5, 5, 90, 171, 5, 5, 228, 183, 5, 5, 33, 221, 5, 5, 26, 180, 5, 5, 171, 216, 5, 5, 240, 214, 5, 5, 181, 239, 5, 5, 196, 230, 5, 5, 223, 187, 5, 5, 86, 237, 5, 5, 38, 168, 5, 5, 37, 236, 5, 5, 9, 199, 5, 5, 179, 216, 5, 5, 42, 236, 5, 5, 37, 180, 5, 5, 26, 178, 5, 5, 15, 206, 5, 5, 195, 216, 5, 5, 93, 237, 5, 5, 54, 243, 5, 5, 57, 243, 5, 5, 15, 223, 5, 5, 26, 223, 5, 5, 57, 246, 5, 5, 94, 171, 5, 5, 224, 240, 5, 5, 75, 178, 5, 5, 16, 223, 5, 5, 143, 201, 5, 5, 173, 210, 5, 5, 218, 227, 5, 5, 115, 182, 5, 5, 68, 187, 5, 5, 200, 226, 5, 5, 82, 246, 5, 5, 42, 217, 5, 5, 47, 171, 5, 5, 83, 244, 5, 5, 152, 208, 5, 5, 246, 176, 5, 5, 68, 191, 5, 5, 129, 179, 5, 5, 25, 188, 5, 5, 234, 172, 5, 5, 202, 248, 5, 5, 227, 175, 5, 5, 98, 228, 5, 5, 118, 182, 5, 5, 157, 177, 5, 5, 174, 230, 5, 5, 225, 227, 5, 5, 41, 207, 5, 5, 40, 217, 5, 5, 226, 214, 5, 5, 121, 220, 5, 5, 154, 213, 5, 5, 27, 188, 5, 5, 36, 215, 5, 5, 25, 230, 5, 5, 26, 230, 5, 5, 201, 235, 5, 5, 111, 223, 5, 5, 44, 219, 5, 5, 234, 185, 5, 5, 88, 194, 5, 5, 178, 174, 5, 5, 27, 230, 5, 5, 143, 227, 5, 5, 174, 196, 5, 5, 86, 225, 5, 5, 86, 246, 5, 5, 170, 246, 5, 5, 42, 215, 5, 5, 98, 208, 5, 5, 78, 186, 5, 5, 165, 228, 5, 5, 56, 196, 5, 5, 57, 196, 5, 5, 91, 219, 5, 5, 24, 198, 5, 5, 50, 189, 5, 5, 255, 224, 5, 5, 3, 218, 5, 5, 102, 192, 5, 5, 203, 170, 5, 5, 204, 170, 5, 5, 238, 184, 5, 5, 202, 229, 5, 5, 69, 247, 5, 5, 141, 182, 5, 5, 70, 234, 5, 5, 190, 197, 5, 5, 215, 189, 5, 5, 89, 211, 5, 5, 229, 214, 5, 5, 21, 183, 5, 5, 201, 234, 5, 5, 206, 234, 5, 5, 161, 216, 5, 5, 107, 238, 5, 5, 251, 229, 5, 5, 70, 243, 5, 5, 176, 242, 5, 5, 86, 188, 5, 5, 141, 236, 201, 32, 230, 0, 5, 5, 253, 173, 5, 5, 11, 178, 5, 5, 27, 241, 5, 5, 9, 181, 5, 5, 237, 189, 5, 5, 207, 191, 5, 5, 65, 214, 5, 5, 194, 209, 5, 5, 152, 211, 5, 5, 142, 207, 5, 5, 127, 177, 5, 5, 154, 208, 5, 5, 99, 208, 5, 5, 14, 175, 5, 5, 51, 212, 5, 5, 29, 170, 5, 5, 17, 220, 5, 5, 171, 246, 5, 5, 122, 212, 5, 5, 33, 212, 5, 5, 149, 248, 5, 5, 150, 243, 5, 5, 250, 185, 5, 5, 133, 204, 5, 5, 143, 207, 5, 5, 43, 244, 5, 5, 76, 183, 5, 5, 198, 211, 5, 5, 107, 202, 5, 5, 75, 206, 5, 5, 116, 172, 5, 5, 77, 183, 5, 5, 239, 228, 5, 5, 238, 189, 5, 5, 99, 209, 5, 5, 203, 207, 5, 5, 194, 222, 5, 5, 38, 197, 5, 5, 161, 243, 5, 5, 232, 227, 5, 5, 250, 227, 5, 5, 251, 227, 5, 5, 162, 188, 5, 5, 113, 202, 5, 5, 81, 208, 5, 5, 112, 238, 5, 5, 21, 182, 5, 5, 223, 182, 5, 5, 30, 241, 5, 5, 87, 226, 5, 5, 208, 246, 5, 5, 85, 245, 5, 5, 49, 186, 5, 5, 200, 244, 5, 5, 91, 236, 5, 5, 64, 223, 5, 5, 205, 199, 5, 5, 130, 196, 5, 5, 218, 196, 5, 5, 150, 233, 5, 5, 44, 217, 5, 5, 254, 238, 5, 5, 45, 193, 5, 5, 74, 247, 5, 5, 212, 217, 5, 5, 210, 213, 5, 5, 71, 191, 5, 5, 101, 208, 5, 5, 209, 169, 5, 5, 45, 244, 5, 5, 236, 247, 5, 5, 146, 201, 5, 5, 10, 241, 5, 5, 50, 186, 5, 5, 55, 189, 5, 5, 197, 184, 5, 5, 47, 182, 5, 5, 102, 207, 5, 5, 120, 223, 5, 5, 150, 189, 5, 5, 172, 213, 5, 5, 134, 176, 5, 5, 119, 182, 5, 5, 89, 246, 5, 5, 240, 223, 5, 5, 82, 211, 5, 5, 53, 174, 5, 5, 49, 204, 5, 5, 95, 212, 5, 5, 143, 171, 5, 5, 233, 214, 5, 5, 115, 223, 5, 5, 127, 210, 5, 5, 2, 197, 5, 5, 140, 219, 5, 5, 31, 248, 5, 5, 121, 221, 5, 5, 214, 201, 5, 5, 65, 235, 5, 5, 192, 197, 5, 5, 220, 183, 5, 5, 35, 212, 5, 5, 5, 209, 5, 5, 5, 225, 5, 5, 28, 237, 5, 5, 177, 192, 5, 5, 228, 240, 5, 5, 220, 186, 5, 5, 252, 169, 5, 5, 105, 235, 5, 5, 216, 204, 5, 5, 79, 205, 5, 5, 12, 178, 5, 5, 147, 206, 5, 5, 151, 233, 5, 5, 211, 213, 5, 5, 46, 244, 5, 5, 14, 196, 5, 5, 220, 239, 5, 5, 14, 212, 5, 5, 76, 195, 5, 5, 7, 238, 5, 5, 8, 191, 5, 5, 47, 191, 5, 5, 190, 243, 5, 5, 99, 174, 5, 5, 174, 239, 5, 5, 244, 228, 5, 5, 132, 189, 5, 5, 17, 194, 5, 5, 54, 237, 5, 5, 186, 203, 5, 5, 126, 223, 5, 5, 211, 192, 5, 5, 226, 181, 5, 5, 31, 234, 5, 5, 229, 183, 5, 5, 133, 219, 5, 5, 234, 219, 5, 5, 182, 246, 5, 5, 31, 240, 5, 5, 59, 230, 5, 5, 76, 231, 5, 5, 194, 198, 5, 5, 205, 247, 5, 5, 135, 194, 5, 5, 149, 206, 5, 5, 153, 206, 5, 5, 75, 168, 5, 5, 60, 180, 5, 5, 49, 226, 5, 5, 155, 209, 5, 5, 215, 201, 5, 5, 70, 238, 5, 5, 84, 208, 5, 5, 213, 217, 5, 5, 172, 195, 5, 5, 234, 213, 5, 5, 35, 229, 5, 5, 52, 233, 5, 5, 156, 233, 5, 5, 60, 189, 5, 5, 210, 234, 5, 5, 242, 223, 5, 5, 179, 199, 5, 5, 185, 225, 5, 5, 76, 216, 5, 5, 66, 197, 5, 5, 248, 181, 5, 5, 132, 247, 5, 5, 53, 219, 5, 5, 178, 213, 5, 5, 169, 235, 5, 5, 73, 234, 5, 5, 220, 246, 5, 5, 3, 187, 5, 5, 112, 190, 5, 5, 32, 209, 5, 5, 197, 191, 5, 5, 48, 228, 5, 5, 24, 215, 5, 5, 3, 212, 5, 5, 33, 230, 5, 5, 103, 217, 5, 5, 136, 204, 5, 5, 12, 239, 5, 5, 50, 193, 5, 5, 82, 196, 5, 5, 130, 215, 5, 5, 92, 194, 5, 5, 161, 194, 5, 5, 175, 245, 5, 5, 176, 221, 5, 5, 209, 248, 5, 5, 163, 172, 5, 5, 30, 182, 5, 5, 73, 191, 5, 5, 212, 192, 5, 5, 135, 204, 5, 5, 67, 173, 5, 5, 164, 191, 5, 5, 217, 189, 5, 5, 103, 211, 5, 5, 170, 235, 5, 5, 224, 195, 5, 5, 193, 203, 5, 5, 53, 217, 5, 5, 39, 229, 5, 5, 161, 245, 5, 5, 54, 198, 5, 5, 253, 187, 5, 5, 52, 169, 5, 5, 54, 207, 5, 5, 211, 248, 5, 5, 36, 241, 5, 5, 197, 230, 5, 5, 186, 171, 5, 5, 94, 180, 5, 5, 154, 190, 5, 5, 65, 189, 5, 5, 25, 239, 5, 5, 39, 214, 5, 5, 161, 240, 5, 5, 3, 201, 5, 5, 122, 188, 5, 5, 74, 201, 5, 5, 19, 188, 5, 5, 75, 202, 5, 5, 173, 185, 5, 5, 241, 183, 5, 5, 60, 230, 5, 5, 53, 203, 5, 5, 45, 176, 5, 5, 121, 183, 5, 5, 26, 239, 5, 5, 28, 242, 5, 5, 15, 227, 5, 5, 89, 201, 5, 5, 213, 187, 5, 5, 228, 195, 5, 5, 54, 193, 5, 5, 16, 175, 5, 5, 136, 196, 5, 5, 190, 207, 5, 5, 53, 224, 5, 5, 57, 174, 5, 5, 128, 212, 5, 5, 122, 232, 5, 5, 97, 221, 5, 5, 103, 227, 5, 5, 98, 200, 5, 5, 160, 213, 5, 5, 254, 210, 5, 5, 34, 188, 5, 5, 21, 196, 5, 5, 251, 232, 5, 5, 250, 220, 5, 5, 205, 210, 5, 5, 85, 227, 5, 5, 236, 240, 5, 5, 77, 233, 5, 5, 26, 232, 5, 5, 241, 226, 5, 5, 106, 181, 5, 5, 243, 223, 5, 5, 197, 227, 5, 5, 130, 197, 5, 5, 186, 172, 5, 5, 53, 226, 5, 5, 133, 202, 5, 5, 18, 223, 5, 5, 127, 223, 5, 5, 217, 201, 5, 5, 19, 223, 5, 5, 160, 225, 5, 5, 123, 180, 5, 5, 124, 201, 5, 5, 167, 200, 5, 5, 239, 202, 5, 5, 247, 228, 5, 5, 234, 229, 5, 5, 78, 243, 5, 5, 155, 190, 5, 5, 223, 178, 5, 5, 6, 189, 5, 5, 105, 219, 5, 5, 25, 194, 5, 5, 42, 245, 5, 5, 236, 219, 5, 5, 247, 183, 5, 5, 224, 186, 5, 5, 56, 234, 5, 5, 106, 222, 5, 5, 243, 222, 5, 5, 6, 236, 5, 5, 77, 176, 5, 5, 56, 245, 5, 5, 14, 238, 5, 5, 57, 199, 5, 5, 4, 230, 5, 5, 3, 188, 5, 5, 238, 227, 5, 5, 33, 180, 5, 5, 180, 188, 5, 5, 237, 229, 5, 5, 234, 196, 5, 5, 251, 220, 5, 5, 79, 203, 5, 5, 155, 191, 5, 5, 200, 224, 5, 5, 194, 248, 5, 5, 112, 179, 5, 5, 78, 216, 5, 5, 160, 178, 5, 5, 160, 197, 5, 5, 12, 249, 5, 5, 207, 230, 5, 5, 55, 233, 5, 5, 134, 214, 5, 5, 91, 186, 5, 5, 183, 187, 5, 5, 58, 236, 5, 5, 226, 203, 5, 5, 143, 225, 5, 5, 141, 247, 5, 5, 199, 212, 5, 5, 125, 188, 5, 5, 110, 173, 5, 5, 135, 222, 5, 5, 78, 214, 5, 5, 146, 182, 5, 5, 133, 246, 5, 5, 159, 205, 5, 5, 98, 210, 5, 5, 9, 197, 5, 5, 191, 225, 5, 5, 196, 176, 5, 5, 218, 201, 5, 5, 163, 245, 5, 5, 185, 213, 5, 5, 202, 182, 5, 5, 157, 223, 5, 5, 117, 226, 5, 5, 45, 213, 5, 5, 21, 178, 5, 5, 38, 242, 5, 5, 136, 210, 5, 5, 123, 241, 5, 5, 160, 205, 5, 5, 243, 184, 5, 5, 108, 208, 5, 5, 94, 196, 5, 5, 68, 204, 5, 5, 86, 206, 5, 5, 232, 239, 5, 5, 41, 177, 5, 5, 144, 217, 5, 5, 80, 190, 5, 5, 39, 242, 5, 5, 207, 210, 5, 5, 45, 221, 5, 5, 235, 170, 5, 5, 48, 176, 5, 5, 158, 189, 5, 5, 153, 242, 5, 5, 175, 200, 5, 5, 180, 191, 5, 5, 180, 216, 5, 5, 59, 236, 5, 5, 132, 215, 5, 5, 80, 226, 5, 5, 194, 208, 5, 5, 186, 246, 5, 5, 8, 240, 5, 5, 192, 171, 5, 5, 226, 169, 5, 5, 148, 242, 5, 5, 205, 229, 5, 5, 117, 219, 5, 5, 184, 182, 5, 5, 181, 216, 5, 5, 149, 242, 5, 5, 247, 198, 5, 5, 230, 192, 5, 5, 252, 220, 5, 5, 150, 242, 5, 5, 238, 219, 5, 5, 161, 205, 5, 5, 90, 249, 5, 5, 125, 180, 5, 5, 72, 238, 5, 5, 30, 194, 5, 5, 162, 208, 5, 5, 4, 212, 5, 5, 208, 230, 5, 5, 158, 241, 5, 5, 151, 242, 5, 5, 81, 221, 5, 5, 99, 221, 5, 5, 249, 218, 5, 5, 166, 190, 5, 5, 11, 248, 5, 5, 222, 193, 5, 5, 158, 211, 5, 5, 171, 241, 5, 5, 67, 217, 5, 5, 224, 226, 5, 5, 145, 217, 5, 5, 174, 180, 5, 5, 88, 182, 5, 5, 132, 244, 5, 5, 197, 171, 5, 5, 8, 230, 5, 5, 238, 229, 5, 5, 43, 179, 5, 5, 214, 228, 5, 5, 142, 229, 5, 5, 22, 219, 5, 5, 174, 233, 5, 5, 174, 172, 5, 5, 103, 229, 5, 5, 181, 188, 5, 5, 27, 185, 5, 5, 162, 236, 5, 5, 73, 189, 5, 5, 75, 235, 5, 5, 89, 208, 5, 5, 250, 198, 5, 5, 83, 207, 5, 5, 58, 240, 5, 5, 163, 236, 5, 5, 46, 241, 5, 5, 35, 189, 5, 5, 195, 208, 5, 5, 149, 221, 5, 5, 193, 207, 5, 5, 181, 243, 5, 5, 168, 191, 5, 5, 16, 210, 5, 5, 87, 187, 5, 5, 241, 174, 5, 5, 168, 181, 5, 5, 5, 212, 5, 5, 159, 207, 5, 5, 58, 203, 5, 5, 246, 214, 5, 5, 247, 214, 5, 5, 160, 207, 5, 5, 97, 226, 5, 5, 214, 176, 5, 5, 4, 229, 5, 5, 79, 172, 5, 5, 144, 227, 5, 5, 165, 208, 5, 5, 87, 191, 5, 5, 116, 209, 5, 5, 155, 234, 5, 5, 132, 192, 5, 5, 142, 212, 5, 5, 150, 193, 5, 5, 216, 193, 5, 5, 185, 189, 5, 5, 138, 183, 5, 5, 169, 232, 5, 5, 55, 239, 5, 5, 35, 240, 5, 5, 5, 245, 5, 5, 227, 221, 5, 5, 48, 195, 5, 5, 62, 174, 5, 5, 200, 190, 5, 5, 79, 225, 5, 5, 172, 241, 5, 5, 87, 170, 5, 5, 14, 209, 5, 5, 181, 221, 5, 5, 88, 227, 5, 5, 133, 221, 5, 5, 241, 240, 5, 5, 15, 197, 5, 5, 69, 180, 5, 5, 165, 227, 5, 5, 213, 194, 5, 5, 217, 194, 5, 5, 68, 217, 5, 5, 64, 236, 5, 5, 88, 225, 5, 5, 228, 204, 5, 5, 223, 191, 5, 5, 191, 242, 5, 5, 40, 210, 5, 5, 40, 169, 5, 5, 77, 241, 5, 5, 225, 217, 5, 5, 127, 248, 5, 5, 109, 203, 5, 5, 21, 173, 5, 5, 36, 232, 5, 5, 50, 176, 5, 5, 10, 211, 5, 5, 21, 249, 5, 5, 89, 228, 5, 5, 182, 221, 5, 5, 10, 207, 5, 5, 121, 240, 5, 5, 139, 200, 5, 5, 29, 200, 5, 5, 124, 183, 5, 5, 38, 194, 5, 5, 228, 234, 5, 5, 33, 205, 5, 5, 5, 229, 5, 5, 177, 184, 5, 5, 159, 171, 5, 5, 122, 240, 5, 5, 212, 213, 5, 5, 238, 195, 5, 5, 229, 236, 5, 5, 111, 225, 5, 5, 202, 242, 5, 5, 160, 224, 5, 5, 203, 242, 5, 5, 249, 202, 5, 5, 192, 237, 5, 5, 248, 207, 5, 5, 183, 248, 5, 5, 123, 174, 5, 5, 13, 199, 5, 5, 9, 225, 5, 5, 189, 224, 5, 5, 56, 228, 5, 5, 129, 204, 5, 5, 255, 185, 5, 5, 213, 219, 5, 5, 41, 209, 5, 5, 176, 244, 5, 5, 231, 221, 5, 5, 66, 239, 5, 5, 102, 206, 5, 5, 188, 206, 5, 5, 43, 214, 5, 5, 220, 228, 5, 5, 163, 202, 5, 5, 182, 243, 5, 5, 150, 230, 5, 5, 234, 230, 5, 5, 145, 229, 5, 5, 188, 215, 5, 5, 56, 244, 5, 5, 184, 221, 5, 5, 8, 224, 5, 5, 0, 149, 5, 5, 131, 216, 5, 5, 195, 223, 5, 5, 85, 243, 5, 5, 127, 234, 5, 5, 196, 239, 5, 5, 213, 218, 5, 5, 182, 191, 5, 5, 9, 224, 5, 5, 170, 224, 5, 5, 185, 210, 5, 5, 183, 224, 5, 5, 185, 221, 5, 5, 56, 219, 5, 5, 0, 139, 5, 5, 44, 212, 5, 5, 177, 175, 5, 5, 114, 175, 5, 5, 54, 212, 5, 5, 213, 229, 5, 5, 123, 172, 5, 5, 216, 208, 5, 5, 129, 185, 5, 5, 72, 179, 5, 5, 129, 188, 5, 5, 249, 190, 5, 5, 132, 231, 5, 5, 22, 249, 5, 5, 229, 178, 5, 5, 11, 247, 5, 5, 39, 234, 5, 5, 62, 184, 5, 5, 98, 195, 5, 5, 40, 190, 5, 5, 102, 185, 5, 5, 77, 249, 5, 5, 169, 225, 5, 5, 229, 225, 5, 5, 58, 224, 5, 5, 97, 170, 5, 5, 9, 195, 5, 5, 96, 191, 5, 5, 132, 187, 5, 5, 237, 239, 5, 5, 41, 185, 5, 5, 229, 204, 5, 5, 3, 247, 5, 5, 230, 236, 5, 5, 5, 221, 5, 5, 12, 207, 5, 5, 127, 240, 5, 5, 52, 176, 5, 5, 17, 206, 5, 5, 27, 201, 5, 5, 36, 206, 5, 5, 148, 237, 5, 5, 57, 171, 5, 5, 20, 225, 5, 5, 64, 242, 5, 5, 114, 234, 5, 5, 5, 190, 5, 5, 98, 170, 5, 5, 243, 170, 5, 5, 20, 247, 5, 5, 142, 194, 5, 5, 85, 199, 5, 5, 110, 221, 5, 5, 168, 220, 5, 5, 211, 178, 5, 5, 118, 208, 5, 5, 110, 168, 5, 5, 127, 205, 5, 5, 223, 189, 5, 5, 163, 189, 5, 5, 15, 241, 5, 5, 79, 173, 5, 5, 134, 182, 5, 5, 199, 240, 5, 5, 134, 187, 5, 5, 13, 207, 5, 5, 209, 216, 5, 5, 189, 241, 5, 5, 120, 213, 5, 5, 245, 192, 5, 5, 11, 236, 5, 5, 191, 173, 5, 5, 91, 214, 5, 5, 87, 231, 5, 5, 246, 192, 5, 5, 182, 205, 5, 5, 40, 205, 5, 5, 232, 204, 5, 5, 205, 195, 5, 5, 175, 187, 5, 5, 199, 176, 5, 5, 65, 205, 5, 5, 149, 246, 5, 5, 185, 184, 5, 5, 216, 215, 5, 5, 13, 202, 5, 5, 170, 236, 5, 5, 141, 172, 5, 5, 14, 195, 5, 5, 165, 176, 5, 5, 55, 176, 5, 5, 155, 227, 5, 5, 204, 211, 5, 5, 228, 225, 5, 5, 249, 237, 5, 5, 58, 231, 5, 5, 119, 208, 5, 5, 115, 225, 5, 5, 122, 209, 5, 5, 169, 220, 5, 5, 49, 188, 5, 5, 250, 202, 5, 5, 7, 201, 5, 5, 57, 228, 5, 5, 198, 174, 5, 5, 212, 184, 5, 5, 127, 183, 5, 5, 50, 235, 5, 5, 83, 238, 5, 5, 25, 207, 5, 5, 95, 249, 5, 5, 69, 207, 5, 5, 168, 198, 5, 5, 160, 206, 5, 5, 157, 212, 5, 5, 217, 222, 5, 5, 44, 245, 5, 5, 26, 245, 5, 5, 139, 248, 5, 5, 238, 234, 5, 5, 172, 190, 5, 5, 205, 191, 5, 5, 48, 194, 5, 5, 70, 236, 5, 5, 8, 223, 5, 5, 126, 203, 5, 5, 25, 177, 5, 5, 16, 226, 5, 5, 107, 237, 5, 5, 101, 194, 5, 5, 200, 176, 5, 5, 168, 240, 5, 5, 83, 189, 5, 5, 115, 234, 5, 5, 218, 222, 5, 5, 230, 186, 5, 5, 129, 240, 5, 5, 19, 233, 5, 5, 161, 249, 5, 5, 70, 200, 5, 5, 239, 234, 5, 5, 64, 203, 5, 5, 21, 247, 5, 5, 8, 229, 5, 5, 52, 213, 5, 5, 87, 242, 5, 5, 18, 195, 5, 5, 215, 213, 5, 5, 90, 178, 5, 5, 111, 191, 5, 5, 112, 195, 5, 5, 108, 230, 5, 5, 239, 211, 5, 5, 146, 192, 5, 5, 11, 212, 5, 5, 12, 229, 5, 5, 28, 224, 5, 5, 116, 215, 5, 5, 117, 215, 5, 5, 253, 230, 5, 5, 195, 205, 5, 5, 141, 231, 5, 5, 221, 235, 5, 5, 22, 181, 5, 5, 208, 190, 5, 5, 23, 209, 5, 5, 108, 219, 5, 5, 24, 224, 5, 5, 120, 201, 5, 5, 210, 245, 5, 5, 201, 176, 5, 5, 81, 239, 5, 5, 166, 245, 5, 5, 91, 228, 5, 5, 138, 220, 5, 5, 41, 225, 5, 5, 194, 173, 5, 5, 187, 235, 5, 5, 37, 200, 5, 5, 231, 237, 5, 5, 227, 220, 5, 5, 122, 227, 5, 5, 220, 247, 5, 5, 136, 219, 5, 5, 248, 188, 5, 5, 124, 170, 5, 5, 102, 220, 5, 5, 34, 173, 5, 5, 35, 173, 5, 5, 176, 222, 5, 5, 35, 227, 5, 5, 13, 214, 5, 5, 237, 203, 5, 5, 13, 229, 5, 5, 254, 219, 5, 5, 247, 219, 5, 5, 129, 174, 5, 5, 145, 196, 5, 5, 130, 174, 5, 5, 71, 190, 5, 5, 178, 194, 5, 5, 131, 201, 5, 5, 234, 173, 5, 5, 228, 184, 5, 5, 62, 177, 5, 5, 101, 189, 5, 5, 91, 223, 5, 5, 220, 222, 5, 5, 159, 212, 5, 5, 12, 188, 5, 5, 90, 243, 5, 5, 203, 204, 5, 5, 102, 189, 5, 5, 25, 184, 5, 5, 134, 188, 5, 5, 194, 188, 5, 5, 50, 232, 5, 5, 53, 194, 5, 5, 103, 220, 5, 5, 142, 231, 5, 5, 139, 242, 5, 5, 130, 205, 5, 5, 221, 200, 5, 5, 84, 211, 5, 5, 202, 241, 5, 5, 15, 204, 5, 5, 172, 208, 5, 5, 17, 244, 5, 5, 58, 178, 5, 5, 177, 190, 5, 5, 68, 244, 5, 5, 190, 233, 5, 5, 95, 242, 5, 5, 86, 202, 5, 5, 79, 222, 5, 5, 112, 235, 5, 5, 94, 204, 5, 5, 153, 228, 5, 5, 42, 249, 5, 5, 205, 240, 5, 5, 118, 191, 5, 5, 108, 172, 5, 5, 66, 203, 5, 5, 122, 179, 5, 5, 38, 238, 5, 5, 155, 168, 5, 5, 129, 190, 5, 5, 213, 245, 5, 5, 37, 173, 5, 5, 32, 207, 5, 5, 4, 178, 5, 5, 40, 178, 5, 5, 111, 239, 5, 5, 90, 224, 5, 5, 55, 213, 5, 5, 114, 197, 5, 5, 208, 224, 5, 5, 140, 216, 5, 5, 56, 192, 5, 5, 234, 248, 5, 5, 13, 203, 5, 5, 110, 211, 5, 5, 160, 187, 5, 5, 251, 195, 5, 5, 37, 183, 5, 5, 250, 219, 5, 5, 147, 192, 5, 5, 62, 224, 5, 5, 119, 191, 5, 5, 149, 175, 5, 5, 213, 224, 5, 5, 163, 223, 5, 5, 245, 179, 5, 5, 251, 219, 5, 5, 16, 248, 5, 5, 195, 200, 5, 5, 66, 171, 5, 5, 14, 203, 5, 5, 93, 208, 5, 5, 198, 221, 5, 5, 223, 222, 5, 5, 127, 208, 5, 5, 85, 211, 5, 5, 138, 240, 5, 5, 139, 240, 5, 5, 7, 208, 5, 5, 3, 196, 5, 5, 174, 214, 5, 5, 141, 170, 5, 5, 12, 193, 5, 5, 109, 188, 5, 5, 33, 219, 5, 5, 52, 177, 5, 5, 16, 230, 5, 5, 176, 225, 5, 5, 22, 240, 5, 5, 23, 240, 5, 5, 232, 180, 5, 5, 125, 176, 5, 5, 227, 191, 5, 5, 222, 216, 5, 5, 30, 201, 5, 5, 71, 198, 5, 5, 0, 162, 5, 5, 30, 214, 5, 5, 239, 248, 5, 5, 16, 229, 5, 5, 67, 171, 5, 5, 114, 186, 5, 5, 125, 215, 5, 5, 140, 240, 5, 5, 68, 171, 5, 5, 115, 200, 5, 5, 244, 182, 5, 5, 126, 172, 5, 5, 137, 207, 5, 5, 74, 194, 5, 5, 78, 229, 5, 5, 168, 206, 5, 5, 223, 243, 5, 5, 25, 206, 5, 5, 183, 202, 5, 5, 20, 241, 5, 5, 109, 237, 5, 5, 138, 205, 5, 5, 145, 223, 5, 5, 39, 247, 5, 5, 145, 240, 5, 5, 138, 180, 5, 5, 192, 228, 5, 5, 37, 191, 5, 5, 201, 233, 5, 5, 26, 212, 5, 5, 90, 221, 5, 5, 3, 171, 5, 5, 43, 173, 5, 5, 138, 209, 5, 5, 170, 204, 5, 5, 184, 193, 5, 5, 58, 214, 5, 5, 2, 220, 5, 5, 134, 174, 5, 5, 119, 186, 5, 5, 73, 171, 5, 5, 236, 191, 5, 5, 153, 232, 5, 5, 97, 205, 5, 5, 167, 216, 5, 5, 64, 188, 5, 5, 148, 240, 5, 5, 233, 218, 5, 5, 200, 202, 5, 5, 113, 196, 5, 5, 39, 233, 5, 5, 149, 240, 5, 5, 71, 224, 5, 5, 82, 229, 5, 5, 203, 233, 5, 5, 87, 190, 5, 5, 132, 235, 5, 5, 229, 247, 5, 5, 241, 204, 5, 5, 201, 200, 5, 5, 236, 180, 5, 5, 102, 182, 5, 5, 209, 189, 5, 5, 202, 200, 5, 5, 158, 222, 5, 5, 128, 204, 5, 5, 150, 240, 5, 5, 65, 208, 5, 5, 27, 193, 5, 5, 71, 203, 5, 5, 76, 194, 5, 5, 153, 240, 5, 5, 68, 183, 5, 5, 254, 203, 5, 5, 148, 239, 5, 5, 191, 193, 5, 5, 40, 243, 5, 5, 52, 173, 5, 5, 208, 177, 5, 5, 84, 218, 5, 5, 2, 194, 5, 5, 238, 200, 5, 5, 103, 182, 5, 5, 12, 223, 5, 5, 167, 242, 5, 5, 245, 248, 5, 5, 111, 183, 5, 5, 15, 221, 5, 5, 208, 201, 5, 5, 239, 200, 5, 5, 208, 176, 5, 5, 161, 217, 5, 5, 214, 240, 5, 5, 127, 215, 5, 5, 49, 182, 5, 5, 238, 186, 5, 5, 7, 198, 5, 5, 16, 237, 5, 5, 143, 188, 5, 5, 155, 240, 5, 5, 137, 219, 5, 5, 230, 243, 5, 5, 47, 206, 5, 5, 7, 237, 5, 5, 53, 202, 5, 5, 95, 208, 5, 5, 158, 220, 5, 5, 24, 225, 5, 5, 72, 178, 5, 5, 249, 194, 5, 5, 95, 173, 5, 5, 157, 240, 5, 5, 146, 188, 5, 5, 223, 168, 5, 5, 253, 247, 5, 5, 20, 201, 5, 5, 248, 200, 5, 5, 60, 210, 5, 5, 95, 228, 5, 5, 51, 206, 5, 5, 196, 235, 5, 5, 103, 232, 5, 5, 19, 237, 5, 5, 253, 184, 5, 5, 23, 237, 5, 5, 136, 242, 5, 5, 200, 191, 5, 5, 229, 170, 5, 5, 212, 208, 5, 5, 249, 186, 5, 5, 173, 178, 5, 5, 255, 190, 5, 5, 239, 232, 5, 5, 240, 232, 5, 5, 2, 191, 5, 5, 28, 189, 5, 5, 42, 204, 5, 5, 22, 244, 5, 5, 100, 248, 5, 5, 176, 196, 5, 5, 167, 244, 5, 5, 149, 233, 5, 5, 227, 174, 5, 5, 204, 248, 5, 5, 174, 243, 5, 5, 175, 243, 5, 5, 102, 172, 5, 5, 44, 237, 5, 5, 234, 214, 5, 5, 75, 247, 5, 5, 24, 183, 5, 5, 94, 211, 5, 5, 131, 196, 5, 5, 165, 229, 5, 5, 9, 214, 5, 5, 255, 238, 5, 5, 80, 205, 5, 5, 7, 176, 5, 5, 215, 212, 5, 5, 119, 198, 5, 5, 36, 212, 5, 5, 92, 236, 5, 5, 120, 198, 5, 5, 72, 212, 5, 5, 196, 209, 5, 5, 201, 173, 5, 5, 170, 203, 5, 5, 99, 187, 5, 5, 184, 198, 5, 5, 10, 181, 5, 5, 251, 186, 5, 5, 222, 233, 5, 5, 209, 246, 5, 5, 252, 186, 5, 5, 253, 186, 5, 5, 33, 229, 5, 5, 235, 215, 5, 5, 228, 170, 5, 5, 129, 177, 5, 5, 175, 231, 5, 5, 29, 246, 5, 5, 238, 247, 5, 5, 246, 198, 5, 5, 92, 245, 5, 5, 230, 183, 5, 5, 193, 168, 5, 5, 157, 233, 5, 5, 158, 233, 5, 5, 127, 204, 5, 5, 189, 248, 5, 5, 67, 235, 5, 5, 75, 197, 5, 5, 75, 225, 5, 5, 96, 212, 5, 5, 134, 196, 5, 5, 84, 212, 5, 5, 235, 224, 5, 5, 236, 224, 5, 5, 109, 171, 5, 5, 85, 237, 5, 5, 251, 226, 5, 5, 112, 220, 5, 5, 38, 248, 5, 5, 172, 244, 5, 5, 89, 179, 5, 5, 93, 229, 5, 5, 219, 221, 5, 5, 51, 203, 5, 5, 233, 174, 5, 5, 235, 190, 5, 5, 133, 247, 5, 5, 153, 189, 5, 5, 42, 223, 5, 5, 26, 201, 5, 5, 206, 226, 5, 5, 206, 194, 5, 5, 214, 234, 5, 5, 13, 189, 5, 5, 5, 218, 5, 5, 194, 203, 5, 5, 156, 190, 5, 5, 56, 237, 5, 5, 92, 234, 5, 5, 93, 234, 5, 5, 74, 232, 5, 5, 218, 239, 5, 5, 36, 230, 5, 5, 135, 247, 5, 5, 134, 237, 5, 5, 39, 222, 5, 5, 203, 228, 5, 5, 16, 227, 5, 5, 213, 208, 5, 5, 179, 243, 5, 5, 122, 198, 5, 5, 32, 189, 5, 5, 125, 201, 5, 5, 179, 231, 5, 5, 241, 247, 5, 5, 59, 235, 5, 5, 31, 246, 5, 5, 236, 213, 5, 5, 53, 236, 5, 5, 50, 191, 5, 5, 6, 187, 5, 5, 227, 173, 5, 5, 12, 191, 5, 5, 196, 198, 5, 5, 83, 197, 5, 5, 84, 182, 5, 5, 195, 220, 5, 5, 12, 238, 5, 5, 80, 191, 192, 0, 0, 0, 5, 5, 154, 225, 5, 5, 229, 195, 5, 5, 118, 218, 5, 5, 195, 203, 5, 5, 209, 230, 5, 5, 248, 183, 5, 5, 137, 196, 5, 5, 167, 233, 5, 5, 40, 214, 5, 5, 208, 226, 5, 5, 212, 248, 5, 5, 243, 226, 5, 5, 109, 228, 5, 5, 215, 187, 5, 5, 153, 212, 5, 5, 66, 201, 5, 5, 210, 236, 5, 5, 220, 234, 5, 5, 89, 183, 5, 5, 238, 174, 5, 5, 203, 218, 5, 5, 180, 189, 5, 5, 210, 230, 5, 5, 136, 222, 5, 5, 223, 188, 5, 5, 136, 235, 5, 5, 113, 199, 5, 5, 131, 197, 5, 5, 4, 233, 5, 5, 211, 230, 5, 5, 60, 236, 5, 5, 38, 188, 5, 5, 138, 248, 5, 5, 58, 198, 5, 5, 33, 179, 5, 5, 168, 233, 5, 5, 126, 192, 5, 5, 140, 230, 5, 5, 61, 236, 5, 5, 244, 205, 5, 5, 4, 188, 5, 5, 215, 236, 5, 5, 159, 190, 5, 5, 234, 207, 5, 5, 10, 197, 5, 5, 137, 178, 5, 5, 186, 169, 5, 5, 31, 183, 5, 5, 24, 204, 5, 5, 62, 200, 5, 5, 181, 191, 5, 5, 200, 227, 5, 5, 221, 230, 5, 5, 44, 177, 5, 5, 66, 230, 5, 5, 36, 189, 5, 5, 203, 173, 5, 5, 59, 184, 5, 5, 9, 230, 5, 5, 167, 194, 5, 5, 163, 176, 5, 5, 139, 183, 5, 5, 190, 213, 5, 5, 239, 216, 5, 5, 118, 219, 5, 5, 222, 230, 5, 5, 240, 216, 5, 5, 252, 233, 5, 5, 180, 173, 5, 5, 221, 236, 5, 5, 222, 236, 5, 5, 56, 239, 5, 5, 219, 197, 5, 5, 54, 242, 5, 5, 28, 185, 5, 5, 66, 218, 5, 5, 11, 213, 5, 5, 83, 234, 5, 5, 29, 185, 5, 5, 134, 221, 5, 5, 167, 173, 5, 5, 198, 220, 5, 5, 81, 234, 5, 5, 208, 210, 5, 5, 177, 185, 5, 5, 139, 229, 5, 5, 16, 174, 5, 5, 226, 188, 5, 5, 18, 200, 5, 5, 105, 247, 5, 5, 196, 180, 5, 5, 164, 231, 5, 5, 23, 191, 5, 5, 140, 183, 5, 5, 62, 203, 5, 5, 18, 235, 5, 5, 9, 234, 5, 5, 205, 190, 5, 5, 218, 194, 5, 5, 159, 193, 5, 5, 99, 170, 5, 5, 61, 240, 5, 5, 13, 248, 5, 5, 59, 229, 5, 5, 146, 227, 5, 5, 116, 220, 5, 5, 240, 230, 5, 5, 160, 211, 5, 5, 161, 211, 5, 5, 24, 173, 5, 5, 66, 236, 5, 5, 138, 196, 5, 5, 143, 196, 5, 5, 65, 242, 5, 5, 214, 207, 5, 5, 34, 220, 5, 5, 106, 229, 5, 5, 148, 244, 5, 5, 235, 195, 5, 5, 245, 216, 5, 5, 251, 218, 5, 5, 166, 207, 5, 5, 169, 190, 5, 5, 233, 234, 5, 5, 100, 245, 5, 5, 221, 228, 5, 5, 89, 182, 5, 5, 72, 217, 5, 5, 85, 224, 5, 5, 67, 237, 5, 5, 202, 203, 5, 5, 12, 248, 5, 5, 51, 195, 5, 5, 31, 244, 5, 5, 149, 186, 5, 5, 89, 169, 5, 5, 190, 189, 5, 5, 55, 243, 5, 5, 42, 210, 5, 5, 60, 222, 5, 5, 115, 203, 5, 5, 171, 170, 5, 5, 39, 186, 5, 5, 189, 227, 5, 5, 116, 225, 5, 5, 204, 173, 5, 5, 101, 220, 5, 5, 215, 181, 5, 5, 207, 171, 5, 5, 243, 190, 5, 5, 125, 233, 5, 5, 138, 231, 5, 5, 154, 230, 5, 5, 241, 230, 5, 5, 130, 243, 5, 5, 201, 188, 5, 5, 228, 188, 5, 5, 242, 230, 5, 5, 89, 243, 5, 5, 80, 234, 5, 5, 251, 209, 5, 5, 118, 220, 5, 5, 246, 216, 5, 5, 209, 237, 5, 5, 139, 235, 5, 5, 120, 208, 5, 5, 255, 202, 5, 5, 130, 240, 5, 5, 102, 230, 5, 5, 217, 218, 5, 5, 225, 185, 5, 5, 235, 236, 5, 5, 237, 215, 5, 5, 130, 204, 5, 5, 243, 230, 5, 5, 114, 170, 5, 5, 248, 170, 5, 5, 173, 176, 5, 5, 188, 205, 5, 5, 169, 193, 5, 5, 165, 222, 5, 5, 92, 239, 5, 5, 44, 205, 5, 5, 179, 212, 5, 5, 61, 224, 5, 5, 93, 239, 5, 5, 220, 225, 5, 5, 232, 197, 5, 5, 115, 249, 5, 5, 131, 243, 5, 5, 8, 190, 5, 5, 159, 238, 5, 5, 30, 247, 5, 5, 115, 168, 5, 5, 112, 229, 5, 5, 192, 204, 5, 5, 52, 188, 5, 5, 207, 211, 5, 5, 119, 218, 5, 5, 144, 196, 5, 5, 31, 207, 5, 5, 74, 200, 5, 5, 172, 220, 5, 5, 6, 177, 5, 5, 108, 244, 5, 5, 176, 193, 5, 5, 7, 231, 5, 5, 8, 231, 5, 5, 9, 231, 5, 5, 94, 239, 5, 5, 88, 233, 5, 5, 4, 175, 5, 5, 211, 190, 5, 5, 86, 173, 5, 5, 29, 238, 5, 5, 46, 225, 5, 5, 70, 218, 5, 5, 241, 236, 5, 5, 190, 235, 5, 5, 3, 216, 5, 5, 70, 198, 5, 5, 178, 178, 5, 5, 23, 181, 5, 5, 56, 221, 5, 5, 181, 194, 5, 5, 41, 183, 5, 5, 96, 223, 5, 5, 138, 203, 5, 5, 96, 242, 5, 5, 242, 203, 5, 5, 41, 227, 5, 5, 206, 220, 5, 5, 42, 183, 5, 5, 94, 182, 5, 5, 242, 236, 5, 5, 191, 235, 5, 5, 28, 201, 5, 5, 197, 207, 5, 5, 157, 225, 5, 5, 112, 239, 5, 5, 146, 196, 5, 5, 237, 207, 5, 5, 117, 196, 5, 5, 237, 194, 5, 5, 135, 240, 5, 5, 97, 242, 5, 5, 126, 239, 5, 5, 113, 235, 5, 5, 227, 200, 5, 5, 11, 225, 5, 5, 35, 244, 5, 5, 110, 172, 5, 5, 95, 224, 5, 5, 16, 231, 5, 5, 42, 218, 5, 5, 145, 249, 5, 5, 21, 203, 5, 5, 63, 191, 5, 5, 248, 247, 5, 5, 197, 233, 5, 5, 102, 204, 5, 5, 229, 228, 5, 5, 127, 239, 5, 5, 136, 233, 5, 5, 106, 245, 5, 5, 127, 191, 5, 5, 146, 177, 5, 5, 113, 211, 5, 5, 228, 200, 5, 5, 37, 219, 5, 5, 93, 232, 5, 5, 237, 188, 5, 5, 151, 235, 5, 5, 5, 196, 5, 5, 80, 175, 5, 5, 152, 178, 5, 5, 217, 237, 5, 5, 238, 188, 5, 5, 31, 201, 5, 5, 6, 171, 5, 5, 230, 218, 5, 5, 191, 202, 5, 5, 143, 209, 5, 5, 144, 169, 5, 5, 124, 219, 5, 5, 27, 206, 5, 5, 151, 200, 5, 5, 102, 168, 5, 5, 156, 235, 5, 5, 223, 199, 5, 5, 55, 223, 5, 5, 152, 203, 5, 5, 206, 202, 5, 5, 103, 205, 5, 5, 250, 197, 5, 5, 167, 203, 5, 5, 3, 237, 5, 5, 37, 231, 5, 5, 204, 233, 5, 5, 248, 204, 5, 5, 47, 238, 5, 5, 83, 172, 5, 5, 96, 218, 5, 5, 43, 243, 5, 5, 32, 201, 5, 5, 210, 176, 5, 5, 8, 198, 5, 5, 38, 175, 5, 5, 239, 186, 5, 5, 126, 197, 5, 5, 237, 173, 5, 5, 73, 208, 5, 5, 151, 225, 5, 5, 34, 201, 5, 5, 2, 248, 5, 5, 23, 201, 5, 5, 115, 204, 5, 5, 10, 177, 5, 5, 138, 242, 5, 5, 82, 245, 5, 5, 83, 245, 5, 5, 235, 211, 5, 5, 24, 246, 5, 5, 85, 212, 5, 5, 55, 174, 5, 5, 179, 242, 5, 5, 3, 168, 5, 5, 104, 229, 5, 5, 68, 188, 5, 5, 233, 197, 5, 5, 244, 197, 5, 5, 74, 184, 5, 5, 233, 237, 5, 5, 210, 168, 5, 5, 175, 179, 5, 5, 238, 237, 5, 5, 125, 237, 5, 5, 207, 249, 5, 5, 5, 223, 5, 5, 254, 181, 5, 5, 13, 212, 5, 5, 244, 175, 5, 5, 207, 198, 5, 5, 238, 243, 5, 5, 207, 179, 5, 5, 197, 215, 5, 5, 201, 240, 5, 5, 223, 215, 5, 5, 87, 213, 5, 5, 20, 169, 5, 5, 38, 212, 5, 5, 184, 173, 5, 5, 151, 193, 5, 5, 250, 211, 5, 5, 127, 180, 5, 5, 238, 175, 5, 5, 184, 241, 5, 5, 149, 244, 5, 5, 172, 170, 5, 5, 208, 179, 5, 5, 62, 169, 5, 5, 208, 171, 5, 5, 240, 175, 5, 5, 82, 241, 5, 5, 83, 241, 5, 5, 141, 180, 5, 5, 247, 235, 5, 5, 63, 174, 5, 5, 87, 211, 5, 5, 88, 211, 5, 5, 153, 213, 5, 5, 152, 196, 5, 5, 151, 209, 5, 5, 177, 224, 5, 5, 174, 209, 5, 5, 84, 201, 5, 5, 156, 218, 5, 5, 37, 207, 5, 5, 144, 182, 5, 5, 77, 207, 5, 5, 197, 209, 5, 5, 8, 185, 5, 5, 128, 230, 5, 5, 201, 236, 5, 5, 91, 185, 5, 5, 168, 169, 5, 5, 125, 223, 5, 5, 18, 194, 5, 5, 82, 186, 5, 5, 100, 241, 5, 5, 94, 185, 5, 5, 126, 221, 5, 5, 179, 209, 5, 5, 7, 179, 5, 5, 50, 215, 5, 5, 11, 218, 5, 5, 216, 217, 5, 5, 89, 249, 5, 5, 213, 225, 5, 5, 198, 230, 5, 5, 56, 207, 5, 5, 43, 199, 5, 5, 59, 215, 5, 5, 97, 230, 5, 5, 131, 186, 5, 5, 223, 230, 5, 5, 5, 202, 5, 5, 10, 202, 5, 5, 4, 214, 5, 5, 28, 192, 5, 5, 19, 185, 5, 5, 137, 246, 5, 5, 212, 169, 5, 5, 226, 217, 5, 5, 51, 176, 5, 5, 128, 180, 5, 5, 94, 197, 5, 5, 64, 193, 5, 5, 152, 193, 5, 5, 96, 183, 5, 5, 66, 213, 5, 5, 223, 198, 5, 5, 16, 197, 5, 5, 198, 198, 5, 5, 125, 175, 5, 5, 139, 231, 5, 5, 195, 169, 5, 5, 161, 206, 5, 5, 160, 195, 5, 5, 180, 206, 5, 5, 176, 219, 5, 5, 113, 229, 5, 5, 90, 207, 5, 5, 252, 180, 5, 5, 42, 227, 5, 5, 30, 216, 5, 5, 146, 194, 5, 5, 17, 231, 5, 5, 192, 202, 5, 5, 142, 180, 5, 5, 213, 203, 5, 5, 254, 211, 5, 5, 100, 213, 5, 5, 229, 171, 5, 5, 38, 231, 5, 5, 50, 175, 5, 5, 22, 229, 5, 5, 16, 200, 5, 5, 51, 175, 5, 5, 2, 218, 5, 5, 255, 217, 5, 5, 187, 168, 5, 5, 116, 182, 5, 5, 38, 217, 5, 5, 176, 192, 5, 5, 246, 172, 5, 5, 203, 248, 5, 5, 72, 168, 5, 5, 87, 185, 5, 5, 135, 248, 5, 5, 15, 187, 5, 5, 196, 206, 5, 5, 28, 241, 5, 5, 183, 198, 5, 5, 173, 171, 5, 5, 126, 189, 5, 5, 29, 236, 5, 5, 215, 239, 5, 5, 97, 190, 5, 5, 136, 248, 5, 5, 100, 243, 5, 5, 188, 199, 5, 5, 92, 211, 5, 5, 217, 178, 5, 5, 189, 199, 5, 5, 114, 247, 5, 5, 195, 209, 5, 5, 163, 169, 5, 5, 216, 212, 5, 5, 4, 197, 5, 5, 113, 238, 5, 5, 122, 221, 5, 5, 83, 212, 5, 5, 171, 231, 5, 5, 252, 227, 5, 5, 173, 189, 5, 5, 50, 204, 5, 5, 224, 182, 5, 5, 251, 212, 5, 5, 157, 210, 5, 5, 181, 237, 5, 5, 92, 241, 5, 5, 11, 186, 5, 5, 127, 235, 5, 5, 219, 196, 5, 5, 13, 178, 5, 5, 177, 171, 5, 5, 105, 199, 5, 5, 69, 232, 5, 5, 56, 211, 5, 5, 113, 190, 5, 5, 254, 188, 5, 5, 201, 194, 5, 5, 167, 188, 5, 5, 93, 245, 5, 5, 115, 192, 5, 5, 206, 235, 5, 5, 88, 220, 5, 5, 197, 224, 5, 5, 196, 218, 5, 5, 102, 222, 5, 5, 11, 227, 5, 5, 112, 201, 5, 5, 122, 180, 5, 5, 176, 231, 5, 5, 177, 221, 5, 5, 145, 199, 5, 5, 30, 246, 5, 5, 22, 242, 5, 5, 111, 224, 5, 5, 154, 241, 5, 5, 68, 170, 5, 5, 55, 207, 5, 5, 199, 230, 5, 5, 157, 197, 5, 5, 251, 201, 5, 5, 180, 231, 5, 5, 223, 239, 5, 5, 39, 224, 5, 5, 42, 201, 5, 5, 176, 169, 5, 5, 228, 246, 5, 5, 96, 236, 5, 5, 25, 220, 5, 5, 134, 202, 5, 5, 35, 188, 5, 5, 75, 232, 5, 5, 27, 196, 5, 5, 248, 211, 5, 5, 189, 182, 5, 5, 5, 233, 5, 5, 248, 168, 5, 5, 135, 214, 5, 5, 137, 210, 5, 5, 237, 170, 5, 5, 36, 240, 5, 5, 176, 200, 5, 5, 197, 203, 5, 5, 145, 193, 5, 5, 184, 215, 5, 5, 59, 200, 5, 5, 223, 236, 5, 5, 184, 187, 5, 5, 152, 249, 5, 5, 51, 208, 5, 5, 111, 173, 5, 5, 73, 238, 5, 5, 240, 246, 5, 5, 37, 246, 5, 5, 7, 236, 5, 5, 26, 208, 5, 5, 48, 172, 5, 5, 226, 176, 5, 5, 229, 220, 5, 5, 198, 203, 5, 5, 91, 179, 5, 5, 109, 206, 5, 5, 186, 189, 5, 5, 106, 249, 5, 5, 104, 187, 5, 5, 52, 229, 5, 5, 92, 183, 5, 5, 239, 229, 5, 5, 192, 242, 5, 5, 8, 234, 5, 5, 210, 247, 5, 5, 72, 207, 5, 5, 105, 229, 5, 5, 208, 175, 5, 5, 133, 232, 5, 5, 145, 227, 5, 5, 43, 236, 5, 5, 70, 210, 5, 5, 94, 233, 5, 5, 61, 193, 5, 5, 113, 189, 5, 5, 169, 170, 5, 5, 47, 241, 5, 5, 51, 241, 5, 5, 161, 207, 5, 5, 222, 172, 5, 5, 149, 237, 5, 5, 116, 224, 5, 5, 202, 171, 5, 5, 42, 209, 5, 5, 41, 190, 5, 5, 204, 242, 5, 5, 196, 223, 5, 5, 203, 206, 5, 5, 205, 242, 5, 5, 165, 177, 5, 5, 78, 242, 5, 5, 186, 221, 5, 5, 98, 188, 5, 5, 189, 215, 5, 5, 67, 239, 5, 5, 198, 245, 5, 5, 124, 172, 5, 5, 104, 188, 5, 5, 26, 207, 5, 5, 105, 196, 5, 5, 125, 194, 5, 5, 123, 209, 5, 5, 27, 245, 5, 5, 27, 173, 5, 5, 116, 168, 5, 5, 111, 168, 5, 5, 101, 188, 5, 5, 20, 177, 5, 5, 230, 169, 5, 5, 234, 197, 5, 5, 125, 170, 5, 5, 115, 170, 5, 5, 209, 190, 5, 5, 28, 214, 5, 5, 238, 203, 5, 5, 245, 234, 5, 5, 36, 227, 5, 5, 212, 237, 5, 5, 134, 203, 5, 5, 51, 223, 5, 5, 24, 233, 5, 5, 114, 222, 5, 5, 248, 180, 5, 5, 21, 195, 5, 5, 136, 185, 5, 5, 173, 197, 5, 5, 133, 180, 5, 5, 120, 191, 5, 5, 155, 199, 5, 5, 89, 232, 5, 5, 191, 233, 5, 5, 191, 224, 5, 5, 90, 232, 5, 5, 148, 235, 5, 5, 71, 211, 5, 5, 177, 170, 5, 5, 228, 191, 5, 5, 130, 211, 5, 5, 32, 208, 5, 5, 212, 203, 5, 5, 94, 210, 5, 5, 17, 187, 5, 5, 115, 222, 5, 5, 98, 205, 5, 5, 193, 224, 5, 5, 155, 232, 5, 5, 66, 208, 5, 5, 85, 218, 5, 5, 104, 190, 5, 5, 95, 210, 5, 5, 128, 206, 5, 5, 100, 232, 5, 5, 195, 214, 5, 5, 13, 198, 5, 5, 42, 235, 5, 5, 203, 208, 5, 5, 7, 249, 5, 5, 12, 206, 5, 5, 86, 205, 5, 5, 10, 242, 5, 5, 237, 223, 5, 5, 170, 228, 5, 5, 39, 217, 5, 5, 145, 184, 5, 5, 230, 179, 5, 5, 145, 201, 5, 5, 167, 168, 5, 5, 5, 192, 5, 5, 51, 189, 5, 5, 21, 179, 5, 5, 229, 175, 5, 5, 194, 184, 5, 5, 177, 196, 5, 5, 136, 241, 5, 5, 206, 214, 5, 5, 137, 241, 5, 5, 161, 173, 5, 5, 184, 206, 5, 5, 56, 189, 5, 5, 90, 230, 5, 5, 168, 183, 5, 5, 166, 229, 5, 5, 144, 195, 5, 5, 253, 235, 5, 5, 49, 171, 5, 5, 209, 170, 5, 5, 65, 169, 5, 5, 15, 243, 5, 5, 193, 197, 5, 5, 223, 207, 5, 5, 194, 197, 5, 5, 104, 228, 5, 5, 136, 193, 5, 5, 145, 207, 5, 5, 247, 177, 5, 5, 136, 213, 5, 5, 34, 229, 5, 5, 109, 190, 5, 5, 109, 232, 5, 5, 182, 215, 5, 5, 55, 204, 5, 5, 185, 177, 5, 5, 13, 239, 5, 5, 59, 168, 5, 5, 179, 213, 5, 5, 108, 179, 5, 5, 231, 183, 5, 5, 43, 235, 5, 5, 77, 231, 5, 5, 145, 171, 5, 5, 135, 176, 5, 5, 14, 186, 5, 5, 102, 193, 5, 5, 214, 220, 5, 5, 235, 213, 5, 5, 104, 176, 5, 5, 208, 198, 5, 5, 77, 218, 5, 5, 127, 221, 5, 5, 34, 221, 5, 5, 123, 238, 5, 5, 208, 249, 5, 5, 51, 193, 5, 5, 236, 208, 5, 5, 89, 220, 5, 5, 186, 204, 5, 5, 63, 170, 5, 5, 176, 245, 5, 5, 177, 245, 5, 5, 196, 197, 5, 5, 50, 182, 5, 5, 220, 204, 5, 5, 179, 195, 5, 5, 2, 195, 5, 5, 26, 194, 5, 5, 135, 202, 5, 5, 7, 187, 5, 5, 240, 231, 5, 5, 136, 247, 5, 5, 238, 185, 5, 5, 54, 236, 5, 5, 121, 234, 5, 5, 57, 237, 5, 5, 188, 234, 5, 5, 154, 206, 5, 5, 244, 223, 5, 5, 208, 247, 5, 5, 115, 216, 5, 5, 224, 239, 5, 5, 177, 235, 5, 5, 106, 169, 5, 5, 252, 181, 5, 5, 222, 232, 5, 5, 135, 237, 5, 5, 181, 231, 5, 5, 76, 188, 5, 5, 34, 187, 5, 5, 17, 175, 5, 5, 215, 234, 5, 5, 2, 169, 5, 5, 129, 212, 5, 5, 115, 201, 5, 5, 52, 178, 5, 5, 52, 240, 5, 5, 13, 191, 5, 5, 171, 229, 5, 5, 108, 181, 5, 5, 108, 174, 5, 5, 34, 179, 5, 5, 98, 230, 5, 5, 64, 230, 5, 5, 105, 174, 5, 5, 59, 198, 5, 5, 153, 207, 5, 5, 181, 169, 5, 5, 244, 226, 5, 5, 126, 232, 5, 5, 170, 175, 5, 5, 5, 188, 5, 5, 47, 235, 5, 5, 102, 236, 5, 5, 62, 217, 5, 5, 72, 235, 5, 5, 48, 201, 5, 5, 77, 202, 5, 5, 124, 234, 5, 5, 249, 183, 5, 5, 139, 204, 5, 5, 254, 235, 5, 5, 212, 230, 5, 5, 69, 204, 5, 5, 6, 202, 5, 5, 232, 195, 5, 5, 36, 203, 5, 5, 154, 224, 5, 5, 155, 224, 5, 5, 90, 183, 5, 5, 110, 228, 5, 5, 114, 179, 5, 5, 140, 213, 5, 5, 237, 244, 5, 5, 242, 219, 5, 5, 242, 208, 5, 5, 55, 242, 5, 5, 237, 196, 5, 5, 241, 173, 5, 5, 180, 200, 5, 5, 243, 208, 5, 5, 159, 221, 5, 5, 6, 229, 5, 5, 119, 226, 5, 5, 174, 175, 5, 5, 26, 250, 5, 5, 248, 169, 5, 5, 58, 174, 5, 5, 221, 189, 5, 5, 147, 214, 5, 5, 178, 181, 5, 5, 63, 200, 5, 5, 116, 173, 5, 5, 148, 214, 5, 5, 249, 169, 5, 5, 132, 228, 5, 5, 167, 205, 5, 5, 197, 176, 5, 5, 208, 186, 5, 5, 164, 236, 5, 5, 137, 179, 5, 5, 187, 169, 5, 5, 228, 203, 5, 5, 122, 216, 5, 5, 12, 213, 5, 5, 237, 211, 5, 5, 25, 218, 5, 5, 222, 248, 5, 5, 123, 216, 5, 5, 51, 182, 5, 5, 184, 195, 5, 5, 241, 216, 5, 5, 160, 180, 5, 5, 157, 195, 5, 5, 169, 191, 5, 5, 247, 241, 5, 5, 128, 207, 5, 5, 167, 207, 5, 5, 34, 176, 5, 5, 19, 235, 5, 5, 11, 226, 5, 5, 10, 234, 5, 5, 197, 177, 5, 5, 252, 218, 5, 5, 245, 208, 5, 5, 160, 193, 5, 5, 60, 229, 5, 5, 74, 219, 5, 5, 170, 190, 5, 5, 123, 233, 5, 5, 209, 175, 5, 5, 230, 193, 5, 5, 104, 248, 5, 5, 168, 173, 5, 5, 63, 203, 5, 5, 232, 217, 5, 5, 191, 231, 5, 5, 197, 180, 5, 5, 245, 242, 5, 5, 8, 236, 5, 5, 91, 210, 5, 5, 191, 189, 5, 5, 62, 240, 5, 5, 185, 241, 5, 5, 206, 190, 5, 5, 96, 219, 5, 5, 244, 219, 5, 5, 152, 204, 5, 5, 192, 221, 5, 5, 218, 218, 5, 5, 161, 224, 5, 5, 155, 237, 5, 5, 146, 229, 5, 5, 103, 230, 5, 5, 207, 245, 5, 5, 243, 195, 5, 5, 132, 240, 5, 5, 190, 206, 5, 5, 171, 225, 5, 5, 155, 204, 5, 5, 117, 225, 5, 5, 171, 202, 5, 5, 51, 201, 5, 5, 45, 187, 5, 5, 9, 246, 5, 5, 198, 215, 5, 5, 74, 177, 5, 5, 227, 197, 5, 5, 86, 245, 5, 5, 156, 237, 5, 5, 14, 168, 5, 5, 60, 171, 5, 5, 148, 222, 5, 5, 141, 173, 5, 5, 75, 200, 5, 5, 85, 248, 5, 5, 174, 176, 5, 5, 211, 195, 5, 5, 89, 238, 5, 5, 26, 177, 5, 5, 175, 176, 5, 5, 166, 214, 5, 5, 23, 202, 5, 5, 67, 196, 5, 5, 163, 224, 5, 5, 79, 217, 5, 5, 52, 235, 5, 5, 117, 168, 5, 5, 40, 203, 5, 5, 229, 207, 5, 5, 33, 245, 5, 5, 245, 239, 5, 5, 30, 238, 5, 5, 75, 240, 5, 5, 118, 229, 5, 5, 201, 205, 5, 5, 72, 230, 5, 5, 179, 178, 5, 5, 128, 234, 5, 5, 109, 244, 5, 5, 192, 235, 5, 5, 151, 217, 5, 5, 98, 178, 5, 5, 243, 203, 5, 5, 139, 203, 5, 5, 255, 216, 5, 5, 31, 224, 5, 5, 212, 190, 5, 5, 55, 187, 5, 5, 29, 214, 5, 5, 70, 196, 5, 5, 95, 182, 5, 5, 248, 195, 5, 5, 162, 204, 5, 5, 57, 192, 5, 5, 121, 191, 5, 5, 71, 196, 5, 5, 24, 168, 5, 5, 146, 170, 5, 5, 111, 172, 5, 5, 155, 217, 5, 5, 18, 244, 5, 5, 59, 178, 5, 5, 240, 194, 5, 5, 145, 187, 5, 5, 72, 225, 5, 5, 128, 191, 5, 5, 132, 190, 5, 5, 255, 219, 5, 5, 96, 224, 5, 5, 108, 226, 5, 5, 152, 175, 5, 5, 206, 241, 5, 5, 6, 196, 5, 5, 99, 205, 5, 5, 72, 171, 5, 5, 166, 222, 5, 5, 209, 229, 5, 5, 236, 249, 5, 5, 233, 200, 5, 5, 100, 231, 5, 5, 51, 249, 5, 5, 54, 235, 5, 5, 78, 219, 5, 5, 17, 230, 5, 5, 187, 184, 5, 5, 170, 201, 5, 5, 143, 180, 5, 5, 201, 202, 5, 5, 54, 247, 5, 5, 6, 219, 5, 5, 47, 202, 5, 5, 225, 243, 5, 5, 3, 217, 5, 5, 73, 226, 5, 5, 64, 187, 5, 5, 70, 224, 5, 5, 154, 200, 5, 5, 249, 204, 5, 5, 104, 205, 5, 5, 207, 202, 5, 5, 228, 243, 5, 5, 33, 201, 5, 5, 98, 240, 5, 5, 71, 208, 5, 5, 184, 232, 5, 5, 6, 217, 5, 5, 240, 186, 5, 5, 155, 178, 5, 5, 232, 243, 5, 5, 143, 190, 5, 5, 143, 186, 5, 5, 178, 171, 5, 5, 188, 179, 5, 5, 187, 171, 5, 5, 250, 189, 5, 5, 142, 246, 5, 5, 105, 213, 5, 5, 45, 169, 5, 5, 88, 218, 5, 5, 228, 202, 5, 5, 62, 228, 192, 0, 0, 0, 5, 5, 189, 232, 5, 5, 43, 215, 5, 5, 21, 169, 5, 5, 140, 212, 5, 5, 165, 182, 5, 5, 248, 177, 5, 5, 227, 229, 5, 5, 201, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 228, 5, 5, 174, 189, 5, 5, 56, 204, 5, 5, 124, 238, 5, 5, 182, 213, 5, 5, 111, 176, 5, 5, 236, 168, 5, 5, 132, 197, 5, 5, 134, 173, 5, 5, 154, 174, 192, 0, 0, 0, 5, 5, 86, 178, 5, 5, 27, 208, 5, 5, 30, 172, 5, 5, 156, 248, 5, 5, 191, 213, 5, 5, 173, 170, 5, 5, 188, 247, 5, 5, 199, 245, 192, 0, 0, 0, 5, 5, 115, 176, 5, 5, 68, 240, 5, 5, 252, 214, 5, 5, 44, 232, 5, 5, 60, 205, 5, 5, 60, 179, 5, 5, 208, 211, 5, 5, 251, 207, 5, 5, 86, 248, 5, 5, 2, 170, 5, 5, 29, 204, 5, 5, 118, 244, 5, 5, 111, 230, 5, 5, 57, 213, 5, 5, 239, 177, 5, 5, 229, 229, 5, 5, 86, 240, 5, 5, 181, 236, 5, 5, 188, 184, 5, 5, 176, 177, 5, 5, 66, 221, 5, 5, 94, 226, 5, 5, 95, 226, 5, 5, 9, 188, 5, 5, 136, 173, 5, 5, 114, 221, 5, 5, 188, 216, 5, 5, 39, 221, 5, 5, 71, 173, 5, 5, 72, 173, 5, 5, 82, 219, 5, 5, 135, 221, 5, 5, 226, 223, 5, 5, 54, 221, 5, 5, 255, 240, 5, 5, 234, 222, 5, 5, 147, 205, 5, 5, 42, 184, 5, 5, 225, 240, 5, 5, 244, 169, 5, 5, 136, 183, 5, 5, 62, 211, 5, 5, 85, 226, 5, 5, 26, 241, 5, 5, 99, 193, 5, 5, 19, 221, 5, 5, 144, 244, 5, 5, 101, 179, 5, 5, 205, 183, 5, 5, 22, 210, 5, 5, 66, 179, 5, 5, 157, 213, 5, 5, 203, 226, 5, 5, 56, 190, 5, 5, 237, 226, 5, 5, 247, 245, 5, 5, 176, 243, 5, 5, 243, 207, 5, 5, 57, 170, 5, 5, 58, 170, 5, 5, 199, 204, 5, 5, 165, 235, 5, 5, 134, 204, 5, 5, 162, 173, 5, 5, 180, 209, 5, 5, 74, 243, 5, 5, 123, 182, 5, 5, 232, 183, 5, 5, 172, 185, 5, 5, 89, 226, 5, 5, 145, 195, 5, 5, 146, 195, 5, 5, 4, 218, 5, 5, 94, 229, 5, 5, 233, 183, 5, 5, 90, 226, 5, 5, 181, 209, 5, 5, 29, 181, 5, 5, 40, 212, 5, 5, 93, 194, 5, 5, 50, 228, 5, 5, 132, 177, 5, 5, 29, 210, 5, 5, 137, 204, 5, 5, 218, 169, 5, 5, 228, 245, 5, 5, 168, 175, 5, 5, 183, 209, 5, 5, 184, 209, 5, 5, 166, 172, 5, 5, 93, 226, 5, 5, 158, 184, 5, 5, 69, 170, 5, 5, 135, 177, 5, 5, 241, 246, 5, 5, 62, 206, 5, 5, 136, 214, 5, 5, 63, 206, 5, 5, 9, 212, 5, 5, 74, 238, 5, 5, 57, 182, 5, 5, 61, 190, 5, 5, 231, 220, 5, 5, 173, 241, 5, 5, 186, 209, 5, 5, 61, 198, 5, 5, 37, 239, 5, 5, 140, 204, 5, 5, 232, 220, 5, 5, 198, 179, 5, 5, 62, 175, 5, 5, 63, 190, 5, 5, 56, 178, 5, 5, 158, 248, 5, 5, 34, 192, 5, 5, 134, 228, 5, 5, 97, 194, 5, 5, 125, 174, 5, 5, 170, 173, 5, 5, 219, 227, 5, 5, 163, 201, 5, 5, 47, 192, 5, 5, 212, 172, 5, 5, 163, 204, 5, 5, 215, 179, 5, 5, 163, 227, 5, 5, 32, 204, 5, 5, 110, 194, 5, 5, 112, 194, 5, 5, 82, 175, 5, 5, 58, 213, 5, 5, 220, 179, 5, 5, 222, 179, 5, 5, 31, 213, 5, 5, 73, 195, 5, 5, 16, 178, 5, 5, 140, 222, 5, 5, 141, 222, 5, 5, 22, 247, 5, 5, 157, 238, 5, 5, 102, 210, 5, 5, 4, 210, 5, 5, 231, 179, 5, 5, 30, 170, 5, 5, 32, 195, 5, 5, 120, 203, 5, 5, 9, 185, 5, 5, 92, 185, 5, 5, 188, 196, 5, 5, 164, 247, 5, 5, 215, 231, 5, 5, 122, 220, 5, 5, 141, 234, 5, 5, 163, 211, 5, 5, 114, 202, 5, 5, 49, 237, 5, 5, 39, 174, 5, 5, 34, 241, 5, 5, 173, 168, 5, 5, 147, 195, 5, 5, 197, 197, 5, 5, 119, 214, 5, 5, 159, 231, 5, 5, 31, 177, 5, 5, 64, 170, 5, 5, 14, 239, 5, 5, 121, 202, 5, 5, 150, 249, 5, 5, 235, 175, 5, 5, 81, 183, 5, 5, 40, 248, 5, 5, 97, 212, 5, 5, 255, 212, 5, 5, 203, 184, 5, 5, 212, 198, 5, 5, 94, 176, 5, 5, 91, 235, 5, 5, 101, 246, 5, 5, 251, 177, 5, 5, 229, 245, 5, 5, 31, 182, 5, 5, 187, 246, 5, 5, 223, 225, 5, 5, 223, 196, 5, 5, 120, 192, 5, 5, 232, 182, 5, 5, 224, 196, 5, 5, 91, 220, 5, 5, 55, 193, 5, 5, 44, 235, 5, 5, 174, 244, 5, 5, 132, 171, 5, 5, 255, 210, 5, 5, 71, 246, 5, 5, 189, 240, 5, 5, 88, 196, 5, 5, 222, 217, 5, 5, 224, 225, 5, 5, 253, 226, 5, 5, 138, 238, 5, 5, 36, 195, 5, 5, 46, 229, 5, 5, 53, 191, 5, 5, 16, 225, 5, 5, 87, 237, 5, 5, 239, 174, 5, 5, 242, 246, 5, 5, 250, 188, 5, 5, 18, 236, 5, 5, 211, 207, 5, 5, 103, 180, 5, 5, 99, 196, 5, 5, 7, 233, 5, 5, 53, 227, 5, 5, 89, 227, 5, 5, 60, 207, 5, 5, 36, 213, 5, 5, 8, 233, 5, 5, 40, 224, 5, 5, 183, 183, 5, 5, 154, 202, 5, 5, 4, 247, 5, 5, 83, 177, 5, 5, 44, 181, 5, 5, 143, 230, 5, 5, 32, 220, 5, 5, 91, 201, 5, 5, 108, 222, 5, 5, 167, 190, 5, 5, 37, 232, 5, 5, 68, 239, 5, 5, 241, 169, 5, 5, 58, 245, 5, 5, 209, 186, 5, 5, 35, 225, 5, 5, 238, 182, 5, 5, 178, 206, 5, 5, 230, 203, 5, 5, 163, 174, 5, 5, 238, 192, 5, 5, 122, 226, 5, 5, 45, 168, 5, 5, 245, 174, 5, 5, 100, 170, 5, 5, 101, 170, 5, 5, 246, 208, 5, 5, 135, 186, 5, 5, 212, 180, 5, 5, 97, 181, 5, 5, 61, 229, 5, 5, 125, 241, 5, 5, 46, 177, 5, 5, 20, 236, 5, 5, 246, 247, 5, 5, 230, 176, 5, 5, 229, 177, 5, 5, 106, 221, 5, 5, 107, 247, 5, 5, 251, 174, 5, 5, 79, 242, 5, 5, 115, 189, 5, 5, 98, 183, 5, 5, 143, 200, 5, 5, 69, 237, 5, 5, 20, 174, 5, 5, 114, 227, 5, 5, 248, 241, 5, 5, 157, 187, 5, 5, 174, 229, 5, 5, 173, 190, 5, 5, 118, 199, 5, 5, 116, 193, 5, 5, 201, 239, 5, 5, 82, 239, 5, 5, 66, 205, 5, 5, 228, 219, 5, 5, 235, 197, 5, 5, 2, 175, 5, 5, 254, 230, 5, 5, 197, 186, 5, 5, 95, 239, 5, 5, 149, 229, 5, 5, 148, 192, 5, 5, 241, 175, 5, 5, 9, 169, 5, 5, 102, 191, 5, 5, 156, 204, 5, 5, 4, 173, 5, 5, 112, 222, 5, 5, 166, 211, 5, 5, 75, 179, 5, 5, 160, 177, 5, 5, 44, 171, 5, 5, 21, 225, 5, 5, 64, 245, 5, 5, 250, 170, 5, 5, 112, 221, 5, 5, 238, 176, 5, 5, 142, 206, 5, 5, 113, 239, 5, 5, 156, 249, 5, 5, 48, 175, 5, 5, 65, 245, 5, 5, 198, 244, 5, 5, 224, 223, 5, 5, 252, 219, 5, 5, 19, 218, 5, 5, 170, 179, 5, 5, 68, 205, 5, 5, 69, 205, 5, 5, 131, 209, 5, 5, 141, 216, 5, 5, 19, 240, 5, 5, 169, 240, 5, 5, 217, 190, 5, 5, 33, 184, 5, 5, 141, 203, 5, 5, 234, 204, 5, 5, 31, 216, 5, 5, 165, 204, 5, 5, 98, 201, 5, 5, 57, 232, 5, 5, 242, 182, 5, 5, 240, 177, 5, 5, 40, 240, 5, 5, 235, 188, 5, 5, 0, 154, 5, 5, 13, 169, 5, 5, 58, 232, 5, 5, 198, 186, 5, 5, 120, 187, 5, 5, 124, 211, 5, 5, 110, 242, 5, 5, 20, 229, 5, 5, 134, 239, 5, 5, 209, 240, 5, 5, 42, 213, 5, 5, 203, 201, 5, 5, 184, 202, 5, 5, 180, 222, 5, 5, 45, 178, 5, 5, 18, 204, 5, 5, 124, 179, 5, 5, 19, 204, 5, 5, 199, 200, 5, 5, 45, 171, 5, 5, 22, 193, 5, 5, 171, 174, 5, 5, 99, 237, 5, 5, 62, 235, 5, 5, 56, 195, 5, 5, 56, 246, 5, 5, 139, 207, 5, 5, 214, 202, 5, 5, 244, 191, 5, 5, 206, 200, 5, 5, 80, 192, 5, 5, 81, 179, 5, 5, 63, 235, 5, 5, 175, 240, 5, 5, 27, 240, 5, 5, 167, 217, 5, 5, 218, 240, 5, 5, 25, 225, 5, 5, 85, 179, 5, 5, 146, 206, 5, 5, 54, 206, 5, 5, 55, 206, 5, 5, 140, 171, 5, 5, 144, 171, 5, 5, 80, 187, 5, 5, 194, 168, 5, 5, 22, 182, 5, 5, 174, 178, 5, 5, 23, 182, 5, 5, 230, 168, 5, 5, 231, 168, 5, 5, 64, 216, 5, 5, 106, 192, 5, 5, 23, 244, 5, 5, 24, 244, 5, 5, 108, 207, 5, 5, 169, 178, 5, 5, 229, 211, 5, 5, 33, 195, 5, 5, 189, 190, 5, 5, 30, 187, 5, 5, 136, 176, 5, 5, 58, 204, 5, 5, 34, 185, 5, 5, 106, 209, 5, 5, 116, 192, 5, 5, 208, 194, 5, 5, 200, 212, 5, 5, 38, 185, 5, 5, 0, 147, 5, 5, 110, 181, 5, 5, 127, 188, 5, 5, 42, 188, 5, 5, 88, 170, 5, 5, 133, 228, 5, 5, 65, 175, 5, 5, 158, 215, 5, 5, 235, 230, 5, 5, 0, 153, 5, 5, 65, 233, 5, 5, 135, 188, 5, 5, 252, 190, 5, 5, 136, 188, 5, 5, 69, 244, 5, 5, 28, 177, 5, 5, 138, 188, 5, 5, 69, 233, 5, 5, 32, 238, 5, 5, 221, 213, 5, 5, 140, 188, 5, 5, 242, 194, 5, 5, 26, 168, 5, 5, 21, 234, 5, 5, 254, 190, 5, 5, 208, 202, 5, 5, 113, 213, 5, 5, 215, 188, 5, 5, 27, 195, 5, 5, 245, 212, 5, 5, 219, 184, 5, 5, 98, 212, 5, 5, 182, 247, 5, 5, 32, 198, 5, 5, 69, 217, 5, 5, 60, 177, 5, 5, 32, 218, 5, 5, 136, 244, 5, 5, 108, 186, 5, 5, 36, 198, 5, 5, 37, 198, 5, 5, 191, 247, 5, 5, 139, 244, 5, 5, 115, 186, 5, 5, 81, 245, 5, 5, 144, 180, 5, 5, 254, 208, 5, 5, 207, 214, 5, 5, 110, 240, 5, 5, 139, 241, 5, 5, 150, 169, 5, 5, 74, 245, 5, 5, 78, 247, 5, 5, 141, 212, 5, 5, 168, 188, 5, 5, 113, 240, 5, 5, 176, 188, 5, 5, 27, 239, 5, 5, 149, 171, 5, 5, 118, 228, 5, 5, 177, 188, 5, 5, 96, 168, 5, 5, 234, 244, 5, 5, 100, 236, 5, 5, 142, 193, 5, 5, 182, 188, 5, 5, 124, 241, 5, 5, 233, 199, 5, 5, 128, 182, 5, 5, 175, 184, 5, 5, 138, 178, 5, 5, 15, 212, 5, 5, 60, 184, 5, 5, 70, 217, 5, 5, 154, 221, 5, 5, 141, 178, 5, 5, 173, 205, 5, 5, 239, 244, 5, 5, 249, 207, 5, 5, 24, 202, 5, 5, 206, 195, 5, 5, 240, 234, 5, 5, 166, 193, 5, 5, 22, 212, 5, 5, 233, 186, 5, 5, 52, 168, 5, 5, 92, 205, 5, 5, 197, 234, 5, 5, 154, 247, 5, 5, 99, 178, 5, 5, 55, 168, 5, 5, 120, 186, 5, 5, 209, 202, 5, 5, 192, 209, 5, 5, 232, 179, 5, 5, 239, 184, 5, 5, 142, 234, 5, 5, 47, 207, 5, 5, 198, 228, 5, 5, 125, 246, 5, 5, 235, 214, 5, 5, 230, 242, 5, 5, 91, 226, 5, 5, 161, 232, 5, 5, 3, 181, 5, 5, 223, 233, 5, 5, 119, 231, 5, 5, 39, 212, 5, 5, 82, 183, 5, 5, 11, 181, 5, 5, 237, 208, 5, 5, 33, 209, 5, 5, 146, 221, 5, 5, 35, 222, 5, 5, 75, 243, 5, 5, 157, 208, 5, 5, 8, 212, 5, 5, 166, 182, 5, 5, 192, 208, 5, 5, 219, 177, 5, 5, 148, 207, 5, 5, 114, 207, 5, 5, 152, 198, 5, 5, 115, 232, 5, 5, 69, 199, 5, 5, 40, 222, 5, 5, 31, 237, 5, 5, 32, 246, 5, 5, 185, 237, 5, 5, 40, 221, 5, 5, 214, 191, 5, 5, 158, 177, 5, 5, 230, 245, 5, 5, 209, 199, 5, 5, 225, 196, 5, 5, 98, 221, 5, 5, 133, 238, 5, 5, 147, 221, 5, 5, 208, 207, 5, 5, 107, 209, 5, 5, 41, 248, 5, 5, 183, 245, 5, 5, 184, 245, 5, 5, 136, 208, 5, 5, 41, 222, 5, 5, 145, 242, 5, 5, 189, 179, 5, 5, 180, 210, 5, 5, 91, 249, 5, 5, 92, 249, 5, 5, 202, 173, 5, 5, 158, 244, 5, 5, 69, 235, 5, 5, 111, 171, 5, 5, 86, 208, 5, 5, 222, 204, 5, 5, 56, 224, 5, 5, 19, 227, 5, 5, 44, 208, 5, 5, 225, 179, 5, 5, 35, 179, 5, 5, 105, 210, 5, 5, 37, 209, 5, 5, 73, 235, 5, 5, 155, 174, 5, 5, 211, 199, 5, 5, 161, 197, 5, 5, 156, 209, 5, 5, 2, 246, 5, 5, 173, 226, 5, 5, 58, 237, 5, 5, 153, 236, 5, 5, 113, 209, 5, 5, 121, 247, 5, 5, 108, 209, 5, 5, 170, 245, 5, 5, 38, 246, 5, 5, 162, 207, 5, 5, 138, 224, 5, 5, 225, 220, 5, 5, 43, 188, 5, 5, 168, 190, 5, 5, 44, 179, 5, 5, 143, 174, 5, 5, 84, 177, 5, 5, 163, 197, 5, 5, 111, 181, 5, 5, 11, 207, 5, 5, 129, 232, 5, 5, 224, 230, 5, 5, 99, 200, 5, 5, 191, 200, 5, 5, 223, 193, 5, 5, 117, 220, 5, 5, 123, 226, 5, 5, 143, 187, 5, 5, 135, 228, 5, 5, 244, 201, 5, 5, 61, 222, 5, 5, 247, 216, 5, 5, 203, 203, 5, 5, 9, 236, 5, 5, 34, 196, 5, 5, 46, 246, 5, 5, 14, 202, 5, 5, 192, 200, 5, 5, 86, 224, 5, 5, 167, 197, 5, 5, 24, 223, 5, 5, 59, 224, 5, 5, 118, 180, 5, 5, 102, 170, 5, 5, 50, 213, 5, 5, 210, 209, 5, 5, 160, 191, 5, 5, 186, 210, 5, 5, 174, 205, 5, 5, 69, 239, 5, 5, 99, 195, 5, 5, 63, 172, 5, 5, 167, 247, 5, 5, 190, 241, 5, 5, 161, 191, 5, 5, 204, 206, 5, 5, 218, 231, 5, 5, 43, 234, 5, 5, 27, 191, 5, 5, 144, 187, 5, 5, 183, 243, 5, 5, 65, 176, 5, 5, 161, 193, 5, 5, 215, 207, 5, 5, 162, 180, 5, 5, 116, 189, 5, 5, 21, 235, 5, 5, 132, 226, 5, 5, 4, 200, 5, 5, 43, 185, 5, 5, 98, 219, 5, 5, 129, 207, 5, 5, 241, 234, 5, 5, 40, 182, 5, 5, 14, 230, 5, 5, 185, 208, 5, 5, 95, 175, 5, 5, 33, 200, 5, 5, 45, 208, 5, 5, 224, 229, 5, 5, 88, 199, 5, 5, 93, 178, 5, 5, 244, 173, 5, 5, 231, 198, 5, 5, 216, 223, 5, 5, 166, 231, 5, 5, 250, 216, 5, 5, 132, 209, 5, 5, 48, 209, 5, 5, 251, 206, 5, 5, 25, 223, 5, 5, 70, 244, 5, 5, 66, 245, 5, 5, 114, 239, 5, 5, 142, 179, 5, 5, 73, 199, 5, 5, 133, 209, 5, 5, 34, 223, 5, 5, 179, 176, 5, 5, 27, 205, 5, 5, 166, 174, 5, 5, 252, 206, 5, 5, 109, 213, 5, 5, 74, 174, 5, 5, 56, 187, 5, 5, 254, 207, 5, 5, 155, 228, 5, 5, 103, 219, 5, 5, 125, 213, 5, 5, 18, 231, 5, 5, 32, 216, 5, 5, 31, 214, 5, 5, 18, 248, 5, 5, 194, 178, 5, 5, 88, 221, 5, 5, 35, 223, 5, 5, 164, 203, 5, 5, 236, 173, 5, 5, 59, 232, 5, 5, 156, 198, 5, 5, 33, 238, 5, 5, 250, 234, 5, 5, 44, 227, 5, 5, 163, 209, 5, 5, 246, 203, 5, 5, 116, 187, 5, 5, 68, 194, 5, 5, 44, 238, 5, 5, 27, 168, 5, 5, 133, 191, 5, 5, 220, 244, 5, 5, 248, 193, 5, 5, 121, 186, 5, 5, 109, 245, 5, 5, 157, 217, 5, 5, 180, 207, 5, 5, 97, 175, 5, 5, 230, 219, 5, 5, 79, 211, 5, 5, 153, 235, 5, 5, 220, 237, 5, 5, 232, 191, 5, 5, 12, 208, 5, 5, 148, 208, 5, 5, 147, 213, 5, 5, 149, 208, 5, 5, 206, 201, 5, 5, 224, 199, 5, 5, 3, 198, 5, 5, 35, 235, 5, 5, 150, 208, 5, 5, 245, 191, 5, 5, 106, 205, 5, 5, 18, 208, 5, 5, 250, 204, 5, 5, 241, 186, 5, 5, 18, 207, 5, 5, 103, 231, 5, 5, 191, 175, 5, 5, 152, 225, 5, 5, 159, 198, 5, 5, 22, 248, 5, 5, 75, 207, 5, 5, 173, 195, 5, 5, 180, 195, 5, 5, 56, 242, 5, 5, 47, 223, 5, 5, 57, 244, 5, 5, 16, 198, 5, 5, 244, 221, 5, 5, 186, 238, 5, 5, 74, 221, 5, 5, 90, 246, 5, 5, 109, 189, 5, 5, 76, 221, 5, 5, 159, 240, 5, 5, 40, 197, 5, 5, 142, 247, 5, 5, 209, 194, 5, 5, 85, 177, 5, 5, 189, 180, 5, 5, 0, 159, 5, 5, 241, 194, 5, 5, 110, 244, 5, 5, 4, 243, 5, 5, 220, 168, 5, 5, 206, 221, 5, 5, 251, 179, 5, 5, 230, 214, 5, 5, 8, 192, 5, 5, 64, 249, 5, 5, 198, 184, 5, 5, 129, 230, 5, 5, 119, 245, 5, 5, 93, 199, 5, 5, 10, 185, 5, 5, 114, 231, 5, 5, 80, 182, 5, 5, 206, 199, 5, 5, 79, 178, 5, 5, 185, 206, 5, 5, 22, 220, 5, 5, 220, 177, 5, 5, 198, 197, 5, 5, 122, 202, 5, 5, 172, 183, 5, 5, 238, 208, 5, 5, 127, 181, 5, 5, 211, 191, 5, 5, 162, 198, 5, 5, 183, 246, 5, 5, 238, 214, 5, 5, 138, 198, 5, 5, 148, 195, 5, 5, 50, 171, 5, 5, 143, 181, 5, 5, 35, 236, 5, 5, 217, 212, 5, 5, 120, 245, 5, 5, 97, 236, 5, 5, 77, 224, 5, 5, 80, 248, 5, 5, 215, 173, 5, 5, 12, 181, 5, 5, 66, 228, 5, 5, 206, 236, 5, 5, 174, 195, 5, 5, 83, 183, 5, 5, 36, 182, 5, 5, 109, 209, 5, 5, 159, 244, 5, 5, 226, 196, 5, 5, 29, 242, 5, 5, 213, 198, 5, 5, 255, 227, 5, 5, 2, 228, 5, 5, 8, 179, 5, 5, 195, 244, 5, 5, 185, 245, 5, 5, 152, 181, 5, 5, 242, 183, 5, 5, 185, 209, 5, 5, 42, 248, 5, 5, 133, 200, 5, 5, 166, 170, 5, 5, 143, 211, 5, 5, 161, 213, 5, 5, 151, 212, 5, 5, 59, 204, 5, 5, 99, 212, 5, 5, 148, 201, 5, 5, 237, 213, 5, 5, 150, 171, 5, 5, 238, 213, 5, 5, 41, 221, 5, 5, 162, 243, 5, 5, 4, 168, 5, 5, 136, 202, 5, 5, 223, 204, 5, 5, 17, 227, 5, 5, 3, 241, 5, 5, 137, 202, 5, 5, 210, 199, 5, 5, 143, 175, 5, 5, 42, 199, 5, 5, 223, 217, 5, 5, 211, 247, 5, 5, 181, 199, 5, 5, 38, 187, 5, 5, 159, 181, 5, 5, 68, 210, 5, 5, 26, 215, 5, 5, 162, 205, 5, 5, 47, 229, 5, 5, 5, 168, 5, 5, 156, 185, 5, 5, 127, 232, 5, 5, 36, 234, 5, 5, 103, 236, 5, 5, 67, 180, 5, 5, 154, 212, 5, 5, 213, 230, 5, 5, 116, 201, 5, 5, 69, 189, 5, 5, 46, 223, 5, 5, 183, 231, 5, 5, 240, 215, 5, 5, 226, 216, 5, 5, 251, 228, 5, 5, 17, 216, 5, 5, 38, 239, 5, 5, 44, 199, 5, 5, 9, 233, 5, 5, 27, 218, 5, 5, 98, 173, 5, 5, 52, 201, 5, 5, 37, 189, 5, 5, 174, 241, 5, 5, 10, 233, 5, 5, 187, 231, 5, 5, 66, 207, 5, 5, 155, 206, 5, 5, 242, 240, 5, 5, 218, 173, 5, 5, 230, 173, 5, 5, 10, 230, 5, 5, 141, 204, 5, 5, 178, 240, 5, 5, 57, 207, 5, 5, 28, 218, 5, 5, 224, 236, 5, 5, 33, 220, 5, 5, 89, 200, 5, 5, 57, 242, 5, 5, 242, 174, 5, 5, 26, 190, 5, 5, 123, 205, 5, 5, 29, 174, 5, 5, 224, 193, 5, 5, 165, 211, 5, 5, 160, 223, 5, 5, 223, 248, 5, 5, 46, 199, 5, 5, 173, 212, 5, 5, 67, 236, 5, 5, 159, 248, 5, 5, 58, 199, 5, 5, 109, 174, 5, 5, 152, 214, 5, 5, 153, 249, 5, 5, 216, 216, 5, 5, 231, 203, 5, 5, 35, 198, 5, 5, 160, 171, 5, 5, 6, 180, 5, 5, 247, 208, 5, 5, 139, 179, 5, 5, 162, 193, 5, 5, 227, 188, 5, 5, 175, 205, 5, 5, 11, 168, 5, 5, 87, 224, 5, 5, 33, 218, 5, 5, 236, 201, 5, 5, 155, 169, 5, 5, 238, 239, 5, 5, 227, 180, 5, 5, 100, 230, 5, 5, 153, 214, 5, 5, 98, 206, 5, 5, 136, 228, 5, 5, 95, 179, 5, 5, 73, 210, 5, 5, 191, 169, 5, 5, 249, 214, 5, 5, 30, 174, 5, 5, 75, 219, 5, 5, 116, 236, 5, 5, 210, 179, 5, 5, 8, 180, 5, 5, 129, 180, 5, 5, 9, 228, 5, 5, 108, 195, 5, 5, 64, 240, 5, 5, 193, 170, 5, 5, 15, 199, 5, 5, 116, 170, 5, 5, 108, 229, 5, 5, 52, 223, 5, 5, 208, 245, 5, 5, 199, 180, 5, 5, 194, 231, 5, 5, 91, 178, 5, 5, 13, 226, 5, 5, 93, 210, 5, 5, 204, 212, 5, 5, 238, 193, 5, 5, 55, 179, 5, 5, 82, 225, 5, 5, 227, 172, 5, 5, 103, 226, 5, 5, 253, 214, 5, 5, 20, 181, 5, 5, 99, 183, 5, 5, 80, 235, 5, 5, 39, 218, 5, 5, 40, 218, 5, 5, 191, 206, 5, 5, 227, 185, 5, 5, 214, 210, 5, 5, 25, 224, 5, 5, 116, 181, 5, 5, 117, 176, 5, 5, 157, 204, 5, 5, 92, 223, 5, 5, 121, 225, 5, 5, 76, 169, 5, 5, 44, 190, 5, 5, 233, 212, 5, 5, 67, 229, 5, 5, 203, 219, 5, 5, 197, 201, 5, 5, 76, 177, 5, 5, 105, 226, 5, 5, 196, 231, 5, 5, 255, 230, 5, 5, 8, 203, 5, 5, 23, 212, 5, 5, 14, 229, 5, 5, 110, 243, 5, 5, 213, 180, 5, 5, 139, 245, 5, 5, 232, 198, 5, 5, 135, 200, 5, 5, 87, 248, 5, 5, 162, 237, 5, 5, 137, 187, 5, 5, 90, 248, 5, 5, 38, 173, 5, 5, 4, 215, 5, 5, 118, 168, 5, 5, 160, 212, 5, 5, 204, 204, 5, 5, 134, 205, 5, 5, 157, 198, 5, 5, 250, 175, 5, 5, 33, 174, 5, 5, 16, 240, 5, 5, 199, 201, 5, 5, 251, 170, 5, 5, 5, 215, 5, 5, 75, 209, 5, 5, 6, 215, 5, 5, 22, 177, 5, 5, 121, 249, 5, 5, 220, 216, 5, 5, 87, 176, 5, 5, 94, 206, 5, 5, 64, 192, 5, 5, 146, 220, 5, 5, 99, 201, 5, 5, 152, 217, 5, 5, 111, 244, 5, 5, 195, 178, 5, 5, 69, 194, 5, 5, 148, 231, 5, 5, 247, 203, 5, 5, 13, 180, 5, 5, 50, 225, 5, 5, 218, 190, 5, 5, 25, 212, 5, 5, 167, 243, 5, 5, 5, 216, 5, 5, 213, 178, 5, 5, 185, 202, 5, 5, 70, 194, 5, 5, 183, 194, 5, 5, 19, 231, 5, 5, 46, 245, 5, 5, 33, 216, 5, 5, 253, 180, 5, 5, 249, 193, 5, 5, 106, 242, 5, 5, 169, 248, 5, 5, 209, 188, 5, 5, 19, 199, 5, 5, 72, 244, 5, 5, 167, 201, 5, 5, 120, 195, 5, 5, 153, 175, 5, 5, 48, 238, 5, 5, 44, 218, 5, 5, 100, 178, 5, 5, 212, 238, 5, 5, 111, 194, 5, 5, 242, 212, 5, 5, 243, 212, 5, 5, 111, 242, 5, 5, 132, 213, 5, 5, 201, 181, 5, 5, 29, 168, 5, 5, 238, 198, 5, 5, 188, 193, 5, 5, 112, 242, 5, 5, 79, 219, 5, 5, 13, 208, 5, 5, 101, 212, 5, 5, 128, 176, 5, 5, 224, 171, 5, 5, 100, 237, 5, 5, 210, 206, 5, 5, 43, 172, 5, 5, 61, 232, 5, 5, 225, 199, 5, 5, 172, 201, 5, 5, 207, 201, 5, 5, 58, 247, 5, 5, 202, 202, 5, 5, 210, 202, 5, 5, 104, 182, 5, 5, 47, 218, 5, 5, 102, 212, 5, 5, 99, 240, 5, 5, 215, 202, 5, 5, 251, 204, 5, 5, 252, 204, 5, 5, 145, 209, 5, 5, 234, 171, 5, 5, 253, 222, 5, 5, 242, 186, 5, 5, 245, 200, 5, 5, 44, 172, 5, 5, 136, 236, 5, 5, 10, 222, 5, 5, 5, 222, 5, 5, 67, 202, 5, 5, 183, 218, 5, 5, 246, 220, 5, 5, 13, 243, 5, 5, 116, 223, 5, 5, 103, 214, 5, 5, 178, 224, 5, 5, 198, 206, 5, 5, 152, 233, 5, 5, 117, 237, 5, 5, 235, 231, 5, 5, 115, 214, 5, 5, 116, 214, 5, 5, 176, 246, 5, 5, 233, 169, 5, 5, 223, 180, 5, 5, 115, 247, 5, 5, 163, 218, 5, 5, 68, 238, 5, 5, 220, 221, 5, 5, 101, 241, 5, 5, 221, 246, 5, 5, 146, 226, 5, 5, 234, 183, 5, 5, 87, 184, 5, 5, 104, 208, 5, 5, 209, 249, 5, 5, 94, 246, 5, 5, 43, 224, 5, 5, 206, 207, 5, 5, 40, 250, 5, 5, 104, 217, 5, 5, 245, 189, 5, 5, 39, 248, 5, 5, 67, 221, 5, 5, 78, 224, 5, 5, 107, 176, 5, 5, 250, 172, 5, 5, 48, 208, 5, 5, 245, 205, 5, 5, 235, 241, 5, 5, 194, 232, 5, 5, 65, 230, 5, 5, 139, 226, 5, 5, 115, 213, 5, 5, 43, 248, 5, 5, 84, 187, 5, 5, 182, 231, 5, 5, 102, 246, 5, 5, 231, 192, 5, 5, 77, 185, 5, 5, 186, 245, 5, 5, 78, 185, 5, 5, 20, 223, 5, 5, 233, 195, 5, 5, 100, 221, 5, 5, 159, 184, 5, 5, 88, 200, 5, 5, 36, 179, 5, 5, 120, 178, 5, 5, 218, 191, 5, 5, 194, 225, 5, 5, 149, 214, 5, 5, 133, 186, 5, 5, 227, 186, 5, 5, 242, 249, 5, 5, 80, 204, 5, 5, 169, 205, 5, 5, 116, 171, 5, 5, 239, 195, 5, 5, 121, 178, 5, 5, 138, 246, 5, 5, 168, 205, 5, 5, 22, 173, 5, 5, 89, 170, 5, 5, 154, 245, 5, 5, 24, 191, 5, 5, 78, 241, 5, 5, 133, 231, 5, 5, 197, 239, 5, 5, 28, 249, 5, 5, 224, 191, 5, 5, 200, 245, 5, 5, 10, 184, 5, 5, 246, 242, 5, 5, 69, 230, 5, 5, 83, 232, 5, 5, 68, 237, 5, 5, 111, 246, 5, 5, 81, 238, 5, 5, 168, 207, 5, 5, 76, 223, 5, 5, 48, 179, 5, 5, 192, 169, 5, 5, 229, 248, 5, 5, 209, 245, 5, 5, 251, 240, 5, 5, 247, 192, 5, 5, 81, 185, 5, 5, 118, 225, 5, 5, 82, 223, 5, 5, 207, 206, 5, 5, 206, 224, 5, 5, 113, 184, 5, 5, 23, 235, 5, 5, 167, 214, 5, 5, 107, 242, 5, 5, 93, 231, 5, 5, 58, 192, 5, 5, 141, 223, 5, 5, 39, 173, 5, 5, 41, 178, 5, 5, 117, 187, 5, 5, 97, 224, 5, 5, 87, 202, 5, 5, 114, 211, 5, 5, 63, 208, 5, 5, 133, 178, 5, 5, 193, 202, 5, 5, 87, 218, 5, 5, 44, 243, 5, 5, 37, 226, 5, 5, 233, 243, 5, 5, 230, 201, 5, 5, 244, 218, 5, 5, 231, 241, 5, 5, 146, 241, 5, 5, 252, 201, 5, 5, 169, 233, 5, 5, 124, 216, 5, 5, 155, 188, 5, 5, 80, 227, 5, 5, 118, 234, 5, 5, 62, 223, 5, 5, 127, 218, 5, 5, 81, 227, 5, 5, 65, 249, 5, 5, 215, 172, 5, 5, 220, 184, 5, 5, 2, 239, 5, 5, 236, 231, 5, 5, 104, 171, 5, 5, 196, 210, 5, 5, 12, 217, 5, 5, 13, 217, 5, 5, 98, 247, 5, 5, 58, 183, 5, 5, 118, 188, 5, 5, 76, 243, 5, 5, 209, 198, 5, 5, 193, 208, 5, 5, 95, 246, 5, 5, 18, 196, 5, 5, 33, 170, 5, 5, 96, 246, 5, 5, 23, 242, 5, 5, 106, 208, 5, 5, 111, 199, 5, 5, 39, 169, 5, 5, 219, 212, 5, 5, 131, 210, 5, 5, 138, 202, 5, 5, 37, 241, 5, 5, 187, 249, 5, 5, 220, 212, 5, 5, 188, 171, 5, 5, 60, 204, 5, 5, 110, 209, 5, 5, 145, 174, 5, 5, 198, 210, 5, 5, 116, 216, 5, 5, 32, 237, 5, 5, 22, 250, 5, 5, 229, 246, 5, 5, 103, 246, 5, 5, 160, 222, 5, 5, 76, 197, 5, 5, 68, 249, 5, 5, 198, 191, 5, 5, 18, 192, 5, 5, 59, 174, 5, 5, 20, 227, 5, 5, 243, 246, 5, 5, 199, 191, 5, 5, 183, 188, 5, 5, 184, 239, 5, 5, 13, 249, 5, 5, 244, 246, 5, 5, 37, 195, 5, 5, 138, 218, 5, 5, 173, 180, 5, 5, 139, 238, 5, 5, 212, 247, 5, 5, 84, 191, 5, 5, 105, 211, 5, 5, 103, 184, 5, 5, 225, 230, 5, 5, 58, 185, 5, 5, 53, 201, 5, 5, 184, 183, 5, 5, 216, 235, 5, 5, 21, 223, 5, 5, 246, 205, 5, 5, 90, 200, 5, 5, 226, 184, 5, 5, 31, 196, 5, 5, 5, 226, 5, 5, 117, 174, 5, 5, 110, 227, 5, 5, 199, 220, 5, 5, 22, 223, 5, 5, 46, 236, 5, 5, 99, 206, 5, 5, 176, 205, 5, 5, 136, 186, 5, 5, 27, 250, 5, 5, 140, 218, 5, 5, 186, 248, 5, 5, 73, 169, 5, 5, 249, 168, 5, 5, 35, 192, 5, 5, 112, 246, 5, 5, 12, 247, 5, 5, 93, 200, 5, 5, 237, 201, 5, 5, 174, 212, 5, 5, 224, 198, 5, 5, 117, 171, 5, 5, 66, 175, 5, 5, 243, 249, 5, 5, 66, 242, 5, 5, 228, 223, 5, 5, 65, 206, 5, 5, 174, 232, 5, 5, 84, 238, 5, 5, 140, 231, 5, 5, 194, 170, 5, 5, 248, 192, 5, 5, 21, 184, 5, 5, 53, 213, 5, 5, 183, 211, 5, 5, 52, 195, 5, 5, 108, 247, 5, 5, 112, 249, 5, 5, 193, 234, 5, 5, 68, 174, 5, 5, 143, 178, 5, 5, 197, 229, 5, 5, 51, 232, 5, 5, 35, 249, 5, 5, 88, 242, 5, 5, 255, 192, 5, 5, 246, 234, 5, 5, 239, 245, 5, 5, 31, 247, 5, 5, 147, 178, 5, 5, 125, 193, 5, 5, 48, 192, 5, 5, 64, 185, 5, 5, 65, 185, 5, 5, 110, 186, 5, 5, 219, 218, 5, 5, 91, 224, 5, 5, 242, 218, 5, 5, 7, 193, 5, 5, 170, 198, 5, 5, 216, 209, 5, 5, 98, 172, 5, 5, 177, 207, 5, 5, 49, 247, 5, 5, 249, 192, 5, 5, 196, 205, 5, 5, 230, 223, 5, 5, 59, 192, 5, 5, 170, 240, 5, 5, 198, 229, 5, 5, 23, 217, 5, 5, 253, 219, 5, 5, 232, 188, 5, 5, 115, 239, 5, 5, 219, 190, 5, 5, 88, 216, 5, 5, 189, 192, 5, 5, 98, 224, 5, 5, 32, 233, 5, 5, 32, 214, 5, 5, 184, 194, 5, 5, 197, 248, 5, 5, 110, 247, 5, 5, 255, 249, 5, 5, 5, 206, 5, 5, 99, 224, 5, 5, 116, 211, 5, 5, 3, 220, 5, 5, 134, 191, 5, 5, 81, 218, 5, 5, 190, 211, 5, 5, 113, 242, 5, 5, 133, 213, 5, 5, 23, 193, 5, 5, 178, 227, 5, 5, 199, 229, 5, 5, 80, 174, 5, 5, 233, 191, 5, 5, 226, 199, 5, 5, 7, 206, 5, 5, 10, 171, 5, 5, 5, 220, 5, 5, 89, 218, 5, 5, 198, 248, 5, 5, 61, 202, 5, 5, 3, 177, 5, 5, 98, 235, 5, 5, 45, 228, 5, 5, 153, 196, 5, 5, 231, 216, 5, 5, 182, 230, 5, 5, 232, 216, 5, 5, 52, 199, 5, 5, 118, 241, 5, 5, 35, 221, 5, 5, 60, 196, 5, 5, 200, 237, 5, 5, 206, 175, 5, 5, 176, 248, 5, 5, 83, 227, 5, 5, 85, 201, 5, 5, 73, 216, 5, 5, 192, 244, 5, 5, 186, 237, 5, 5, 188, 170, 5, 5, 96, 169, 5, 5, 187, 237, 5, 5, 112, 171, 5, 5, 51, 228, 5, 5, 242, 247, 5, 5, 3, 195, 5, 5, 48, 216, 5, 5, 156, 179, 5, 5, 38, 230, 5, 5, 39, 187, 5, 5, 136, 237, 5, 5, 245, 246, 5, 5, 236, 175, 5, 5, 205, 237, 5, 5, 168, 226, 5, 5, 8, 195, 5, 5, 237, 175, 5, 5, 139, 196, 5, 5, 11, 233, 5, 5, 110, 174, 5, 5, 83, 199, 5, 5, 4, 177, 5, 5, 241, 229, 5, 5, 30, 178, 5, 5, 102, 199, 5, 5, 225, 198, 5, 5, 224, 248, 5, 5, 160, 234, 5, 5, 229, 223, 5, 5, 195, 186, 5, 5, 234, 199, 5, 5, 105, 180, 5, 5, 251, 248, 5, 5, 140, 235, 5, 5, 242, 229, 5, 5, 58, 228, 5, 5, 47, 236, 5, 5, 191, 241, 5, 5, 103, 197, 5, 5, 251, 216, 5, 5, 163, 237, 5, 5, 164, 237, 5, 5, 171, 226, 5, 5, 195, 173, 5, 5, 251, 241, 5, 5, 106, 226, 5, 5, 171, 179, 5, 5, 115, 197, 5, 5, 175, 203, 5, 5, 10, 231, 5, 5, 158, 230, 5, 5, 236, 199, 5, 5, 242, 175, 5, 5, 79, 245, 5, 5, 170, 199, 5, 5, 171, 199, 5, 5, 235, 204, 5, 5, 78, 174, 5, 5, 53, 177, 5, 5, 142, 203, 5, 5, 36, 244, 5, 5, 9, 177, 5, 5, 58, 216, 5, 5, 4, 217, 5, 5, 110, 180, 5, 5, 40, 244, 5, 5, 21, 205, 5, 5, 93, 216, 5, 5, 99, 202, 5, 5, 158, 175, 5, 5, 210, 221, 5, 5, 82, 184, 5, 5, 46, 215, 5, 5, 165, 175, 5, 5, 61, 189, 5, 5, 120, 214, 5, 5, 93, 188, 5, 5, 128, 221, 5, 5, 3, 183, 5, 5, 200, 222, 5, 5, 204, 208, 5, 5, 175, 217, 5, 5, 254, 244, 5, 5, 44, 248, 5, 5, 61, 204, 5, 5, 224, 204, 5, 5, 133, 171, 5, 5, 89, 196, 5, 5, 97, 196, 5, 5, 20, 245, 5, 5, 237, 168, 5, 5, 133, 223, 5, 5, 62, 198, 5, 5, 74, 189, 5, 5, 22, 227, 5, 5, 170, 223, 5, 5, 100, 196, 5, 5, 164, 179, 5, 5, 70, 239, 5, 5, 213, 222, 5, 5, 95, 196, 5, 5, 178, 217, 5, 5, 109, 195, 5, 5, 164, 213, 5, 5, 186, 180, 5, 5, 25, 202, 5, 5, 98, 248, 5, 5, 195, 172, 5, 5, 180, 178, 5, 5, 63, 177, 5, 5, 79, 228, 5, 5, 115, 196, 5, 5, 150, 198, 5, 5, 116, 196, 5, 5, 235, 247, 5, 5, 237, 247, 5, 5, 147, 201, 5, 5, 164, 212, 5, 5, 147, 241, 5, 5, 185, 174, 5, 5, 202, 184, 5, 5, 49, 207, 5, 5, 239, 247, 5, 5, 105, 228, 5, 5, 157, 180, 5, 5, 19, 192, 5, 5, 207, 194, 5, 5, 177, 168, 5, 5, 44, 224, 5, 5, 117, 192, 5, 5, 149, 216, 5, 5, 95, 245, 5, 5, 123, 224, 5, 5, 255, 235, 5, 5, 152, 248, 5, 5, 184, 242, 5, 5, 246, 189, 5, 5, 77, 188, 5, 5, 78, 233, 5, 5, 187, 172, 5, 5, 70, 170, 5, 5, 35, 170, 5, 5, 211, 193, 5, 5, 195, 238, 5, 5, 34, 180, 5, 5, 94, 220, 5, 5, 132, 221, 5, 5, 113, 177, 5, 5, 226, 178, 5, 5, 244, 247, 5, 5, 242, 209, 5, 5, 156, 174, 5, 5, 57, 186, 5, 5, 144, 202, 5, 5, 79, 216, 5, 5, 11, 209, 5, 5, 88, 169, 5, 5, 148, 226, 5, 5, 130, 223, 5, 5, 250, 183, 5, 5, 169, 172, 5, 5, 227, 169, 5, 5, 37, 182, 5, 5, 136, 177, 5, 5, 72, 249, 5, 5, 37, 179, 5, 5, 70, 204, 5, 5, 53, 244, 5, 5, 138, 211, 5, 5, 251, 183, 5, 5, 17, 186, 5, 5, 86, 182, 5, 5, 57, 193, 5, 5, 225, 184, 5, 5, 129, 182, 5, 5, 253, 221, 5, 5, 105, 207, 5, 5, 227, 213, 5, 5, 76, 226, 5, 5, 146, 193, 5, 5, 235, 216, 5, 5, 225, 204, 5, 5, 12, 209, 5, 5, 167, 170, 5, 5, 156, 206, 5, 5, 105, 187, 5, 5, 111, 217, 5, 5, 190, 174, 5, 5, 193, 239, 5, 5, 206, 237, 5, 5, 79, 232, 5, 5, 39, 170, 5, 5, 236, 216, 5, 5, 106, 200, 5, 5, 186, 178, 5, 5, 58, 233, 5, 5, 185, 195, 5, 5, 227, 217, 5, 5, 125, 224, 5, 5, 23, 219, 5, 5, 39, 182, 5, 5, 182, 199, 5, 5, 51, 248, 5, 5, 49, 227, 5, 5, 29, 192, 5, 5, 138, 177, 5, 5, 82, 188, 5, 5, 175, 172, 5, 5, 124, 247, 5, 5, 84, 199, 5, 5, 181, 200, 5, 5, 90, 170, 5, 5, 62, 220, 5, 5, 55, 178, 5, 5, 43, 246, 5, 5, 177, 172, 5, 5, 187, 183, 5, 5, 86, 243, 5, 5, 115, 227, 5, 5, 238, 211, 5, 5, 15, 202, 5, 5, 55, 201, 5, 5, 44, 197, 5, 5, 210, 186, 5, 5, 231, 193, 5, 5, 10, 188, 5, 5, 24, 227, 5, 5, 192, 231, 5, 5, 13, 247, 5, 5, 118, 174, 5, 5, 45, 224, 5, 5, 62, 222, 5, 5, 57, 248, 5, 5, 28, 250, 5, 5, 63, 233, 5, 5, 201, 220, 5, 5, 225, 226, 5, 5, 178, 172, 5, 5, 109, 236, 5, 5, 80, 185, 5, 5, 148, 199, 5, 5, 211, 184, 5, 5, 67, 175, 5, 5, 183, 199, 5, 5, 30, 185, 5, 5, 87, 243, 5, 5, 196, 211, 5, 5, 67, 215, 5, 5, 64, 233, 5, 5, 232, 193, 5, 5, 253, 213, 5, 5, 183, 200, 5, 5, 171, 249, 5, 5, 41, 170, 5, 5, 117, 170, 5, 5, 118, 170, 5, 5, 158, 185, 5, 5, 240, 224, 5, 5, 182, 186, 5, 5, 126, 241, 5, 5, 167, 193, 5, 5, 144, 178, 5, 5, 67, 186, 5, 5, 196, 174, 5, 5, 50, 246, 5, 5, 217, 216, 5, 5, 57, 220, 5, 5, 168, 247, 5, 5, 183, 205, 5, 5, 209, 171, 5, 5, 41, 192, 5, 5, 232, 203, 5, 5, 47, 224, 5, 5, 64, 198, 5, 5, 22, 184, 5, 5, 151, 244, 5, 5, 68, 186, 5, 5, 59, 199, 5, 5, 114, 215, 5, 5, 74, 215, 5, 5, 65, 198, 5, 5, 21, 176, 5, 5, 211, 186, 5, 5, 160, 242, 5, 5, 181, 172, 5, 5, 198, 249, 5, 5, 237, 171, 5, 5, 61, 244, 5, 5, 83, 216, 5, 5, 40, 228, 5, 5, 117, 206, 5, 5, 231, 177, 5, 5, 20, 233, 5, 5, 122, 219, 5, 5, 54, 194, 5, 5, 24, 235, 5, 5, 174, 170, 5, 5, 128, 224, 5, 5, 176, 232, 5, 5, 86, 232, 5, 5, 200, 213, 5, 5, 211, 245, 5, 5, 14, 234, 5, 5, 196, 189, 5, 5, 160, 238, 5, 5, 61, 248, 5, 5, 3, 243, 5, 5, 35, 176, 5, 5, 21, 206, 5, 5, 44, 230, 5, 5, 48, 180, 5, 5, 53, 223, 5, 5, 149, 192, 5, 5, 100, 195, 5, 5, 213, 190, 5, 5, 15, 234, 5, 5, 218, 207, 5, 5, 136, 182, 5, 5, 219, 175, 5, 5, 109, 248, 5, 5, 67, 213, 5, 5, 101, 183, 5, 5, 247, 247, 5, 5, 214, 190, 5, 5, 86, 216, 5, 5, 117, 189, 5, 5, 20, 217, 5, 5, 197, 208, 5, 5, 166, 215, 5, 5, 70, 186, 5, 5, 38, 200, 5, 5, 91, 222, 5, 5, 61, 205, 5, 5, 94, 243, 5, 5, 205, 188, 5, 5, 125, 225, 5, 5, 36, 243, 5, 5, 49, 175, 5, 5, 47, 185, 5, 5, 218, 182, 5, 5, 123, 219, 5, 5, 53, 246, 5, 5, 5, 186, 5, 5, 22, 211, 5, 5, 168, 215, 5, 5, 67, 233, 5, 5, 7, 177, 5, 5, 15, 205, 5, 5, 182, 212, 5, 5, 165, 180, 5, 5, 178, 202, 5, 5, 133, 170, 5, 5, 235, 248, 5, 5, 128, 175, 5, 5, 64, 220, 5, 5, 201, 174, 5, 5, 66, 248, 5, 5, 201, 215, 5, 5, 236, 204, 5, 5, 225, 200, 5, 5, 177, 193, 5, 5, 36, 172, 5, 5, 33, 202, 5, 5, 129, 191, 5, 5, 134, 170, 5, 5, 240, 178, 5, 5, 180, 176, 5, 5, 77, 236, 5, 5, 183, 212, 5, 5, 99, 172, 5, 5, 111, 248, 5, 5, 21, 213, 5, 5, 110, 213, 5, 5, 82, 180, 5, 5, 114, 242, 5, 5, 222, 208, 5, 5, 157, 227, 5, 5, 71, 244, 5, 5, 65, 220, 5, 5, 118, 187, 5, 5, 170, 238, 5, 5, 10, 190, 5, 5, 87, 173, 5, 5, 89, 199, 5, 5, 239, 176, 5, 5, 205, 213, 5, 5, 34, 244, 5, 5, 65, 192, 5, 5, 92, 187, 5, 5, 32, 224, 5, 5, 62, 205, 5, 5, 183, 172, 5, 5, 202, 205, 5, 5, 216, 210, 5, 5, 145, 224, 5, 5, 8, 177, 5, 5, 146, 179, 5, 5, 146, 224, 5, 5, 152, 201, 5, 5, 203, 180, 5, 5, 116, 203, 5, 5, 33, 233, 5, 5, 225, 171, 5, 5, 131, 208, 5, 5, 66, 220, 5, 5, 101, 178, 5, 5, 146, 203, 5, 5, 242, 177, 5, 5, 125, 179, 5, 5, 71, 184, 5, 5, 250, 193, 5, 5, 24, 209, 5, 5, 45, 200, 5, 5, 206, 177, 5, 5, 188, 194, 5, 5, 181, 178, 5, 5, 221, 190, 5, 5, 129, 224, 5, 5, 100, 201, 5, 5, 199, 243, 5, 5, 40, 233, 5, 5, 209, 205, 5, 5, 85, 222, 5, 5, 200, 243, 5, 5, 12, 215, 5, 5, 252, 211, 5, 5, 178, 214, 5, 5, 253, 211, 5, 5, 232, 184, 5, 5, 121, 197, 5, 5, 210, 205, 5, 5, 204, 205, 5, 5, 128, 236, 5, 5, 238, 171, 5, 5, 61, 178, 5, 5, 168, 219, 5, 5, 73, 248, 5, 5, 6, 186, 5, 5, 84, 215, 5, 5, 28, 203, 5, 5, 38, 172, 5, 5, 196, 247, 5, 5, 203, 200, 5, 5, 92, 222, 5, 5, 240, 200, 5, 5, 47, 200, 5, 5, 214, 241, 5, 5, 39, 243, 5, 5, 110, 188, 5, 5, 219, 245, 5, 5, 250, 224, 5, 5, 74, 226, 5, 5, 34, 211, 5, 5, 83, 175, 5, 5, 163, 192, 5, 5, 176, 238, 5, 5, 188, 214, 5, 5, 237, 225, 5, 5, 117, 248, 5, 5, 198, 247, 5, 5, 71, 182, 5, 5, 222, 223, 5, 5, 197, 247, 5, 5, 175, 215, 5, 5, 248, 248, 5, 5, 239, 225, 5, 5, 219, 205, 5, 5, 107, 205, 5, 5, 7, 194, 5, 5, 42, 228, 5, 5, 156, 240, 5, 5, 127, 242, 5, 5, 208, 200, 5, 5, 255, 204, 5, 5, 94, 216, 5, 5, 33, 203, 5, 5, 249, 200, 5, 5, 91, 215, 5, 5, 46, 243, 5, 5, 96, 247, 5, 5, 165, 217, 5, 5, 34, 203, 5, 5, 185, 170, 5, 5, 207, 180, 5, 5, 222, 249, 5, 5, 62, 202, 5, 5, 105, 223, 5, 5, 134, 206, 5, 5, 158, 240, 5, 5, 48, 243, 5, 5, 250, 248, 5, 5, 137, 242, 5, 5, 78, 208, 5, 5, 218, 178, 5, 5, 81, 182, 5, 5, 27, 221, 5, 5, 121, 245, 5, 5, 36, 221, 5, 5, 150, 211, 5, 5, 169, 188, 5, 5, 216, 201, 5, 5, 239, 231, 5, 5, 69, 249, 5, 5, 128, 210, 5, 5, 72, 177, 5, 5, 54, 245, 5, 5, 51, 215, 5, 5, 183, 244, 5, 5, 36, 170, 5, 5, 23, 169, 5, 5, 137, 230, 5, 5, 26, 220, 5, 5, 163, 198, 5, 5, 246, 218, 5, 5, 52, 183, 5, 5, 71, 170, 5, 5, 33, 177, 5, 5, 46, 240, 5, 5, 125, 245, 5, 5, 87, 208, 5, 5, 12, 225, 5, 5, 247, 189, 5, 5, 178, 168, 5, 5, 145, 202, 5, 5, 206, 184, 5, 5, 84, 197, 5, 5, 240, 213, 5, 5, 114, 209, 5, 5, 218, 176, 5, 5, 204, 244, 5, 5, 185, 247, 5, 5, 157, 174, 5, 5, 251, 223, 5, 5, 174, 226, 5, 5, 146, 202, 5, 5, 226, 230, 5, 5, 254, 223, 5, 5, 75, 189, 5, 5, 23, 227, 5, 5, 23, 249, 5, 5, 170, 172, 5, 5, 27, 243, 5, 5, 146, 247, 5, 5, 229, 203, 5, 5, 128, 248, 5, 5, 238, 168, 5, 5, 80, 201, 5, 5, 61, 183, 5, 5, 233, 181, 5, 5, 112, 217, 5, 5, 189, 188, 5, 5, 82, 203, 5, 5, 134, 232, 5, 5, 11, 184, 5, 5, 83, 203, 5, 5, 105, 172, 5, 5, 35, 196, 5, 5, 80, 202, 5, 5, 30, 243, 5, 5, 184, 205, 5, 5, 19, 197, 5, 5, 159, 214, 5, 5, 47, 177, 5, 5, 250, 168, 5, 5, 28, 245, 5, 5, 233, 203, 5, 5, 154, 249, 5, 5, 38, 196, 5, 5, 190, 187, 5, 5, 28, 190, 5, 5, 181, 219, 5, 5, 182, 219, 5, 5, 122, 225, 5, 5, 196, 170, 5, 5, 254, 218, 5, 5, 176, 208, 5, 5, 118, 189, 5, 5, 195, 234, 5, 5, 155, 249, 5, 5, 197, 189, 5, 5, 55, 194, 5, 5, 198, 243, 5, 5, 120, 176, 5, 5, 27, 202, 5, 5, 192, 233, 5, 5, 196, 183, 5, 5, 186, 211, 5, 5, 203, 169, 5, 5, 112, 186, 5, 5, 117, 234, 5, 5, 48, 185, 5, 5, 126, 225, 5, 5, 88, 217, 5, 5, 70, 193, 5, 5, 136, 172, 5, 5, 132, 248, 5, 5, 129, 225, 5, 5, 59, 208, 5, 5, 184, 219, 5, 5, 66, 183, 5, 5, 7, 244, 5, 5, 171, 198, 5, 5, 147, 194, 5, 5, 76, 209, 5, 5, 185, 219, 5, 5, 186, 219, 5, 5, 189, 211, 5, 5, 25, 231, 5, 5, 89, 203, 5, 5, 149, 194, 5, 5, 158, 199, 5, 5, 241, 171, 5, 5, 137, 190, 5, 5, 183, 222, 5, 5, 160, 249, 5, 5, 156, 232, 5, 5, 191, 211, 5, 5, 160, 227, 5, 5, 39, 211, 5, 5, 211, 202, 5, 5, 50, 250, 5, 5, 247, 178, 5, 5, 40, 211, 5, 5, 178, 226, 5, 5, 36, 201, 5, 5, 97, 208, 5, 5, 61, 223, 5, 5, 154, 196, 5, 5, 109, 231, 5, 5, 211, 185, 5, 5, 59, 246, 5, 5, 156, 196, 5, 5, 93, 241, 5, 5, 213, 192, 5, 5, 240, 172, 5, 5, 179, 247, 5, 5, 171, 235, 5, 5, 2, 243, 5, 5, 62, 189, 5, 5, 120, 241, 5, 5, 170, 188, 5, 5, 106, 228, 5, 5, 164, 218, 5, 5, 194, 229, 5, 5, 168, 229, 5, 5, 54, 217, 5, 5, 239, 209, 5, 5, 8, 249, 5, 5, 71, 227, 5, 5, 96, 211, 5, 5, 152, 183, 5, 5, 219, 192, 5, 5, 132, 222, 5, 5, 46, 176, 5, 5, 221, 212, 5, 5, 255, 245, 5, 5, 27, 220, 5, 5, 66, 189, 5, 5, 188, 246, 5, 5, 118, 192, 5, 5, 6, 183, 5, 5, 79, 243, 5, 5, 166, 218, 5, 5, 252, 177, 5, 5, 181, 195, 5, 5, 245, 223, 5, 5, 169, 182, 5, 5, 153, 224, 5, 5, 34, 177, 5, 5, 180, 196, 5, 5, 154, 243, 5, 5, 226, 168, 5, 5, 196, 195, 5, 5, 180, 183, 5, 5, 246, 246, 5, 5, 84, 214, 5, 5, 76, 249, 5, 5, 58, 175, 5, 5, 70, 189, 5, 5, 155, 243, 5, 5, 219, 201, 5, 5, 123, 231, 5, 5, 252, 183, 5, 5, 170, 233, 5, 5, 46, 221, 5, 5, 155, 171, 5, 5, 46, 248, 5, 5, 108, 217, 5, 5, 71, 204, 5, 5, 47, 248, 5, 5, 224, 220, 5, 5, 244, 184, 5, 5, 88, 237, 5, 5, 253, 183, 5, 5, 5, 228, 5, 5, 232, 245, 5, 5, 191, 177, 5, 5, 171, 175, 5, 5, 179, 221, 5, 5, 83, 247, 5, 5, 31, 232, 5, 5, 211, 249, 5, 5, 123, 196, 5, 5, 41, 169, 5, 5, 146, 217, 5, 5, 115, 209, 5, 5, 206, 222, 5, 5, 238, 249, 5, 5, 212, 199, 5, 5, 28, 196, 5, 5, 171, 218, 5, 5, 83, 188, 5, 5, 175, 233, 5, 5, 92, 195, 5, 5, 215, 247, 5, 5, 63, 175, 5, 5, 172, 186, 5, 5, 239, 168, 5, 5, 220, 197, 5, 5, 213, 199, 5, 5, 187, 189, 5, 5, 148, 176, 5, 5, 119, 190, 5, 5, 178, 185, 5, 5, 163, 225, 5, 5, 93, 195, 5, 5, 116, 199, 5, 5, 214, 194, 5, 5, 228, 217, 5, 5, 167, 184, 5, 5, 157, 206, 5, 5, 76, 235, 5, 5, 237, 169, 5, 5, 130, 232, 5, 5, 7, 184, 5, 5, 160, 185, 5, 5, 70, 180, 5, 5, 207, 218, 5, 5, 175, 226, 5, 5, 194, 239, 5, 5, 192, 201, 5, 5, 176, 233, 5, 5, 164, 197, 5, 5, 229, 234, 5, 5, 160, 184, 5, 5, 199, 179, 5, 5, 50, 227, 5, 5, 72, 223, 5, 5, 139, 194, 5, 5, 195, 232, 5, 5, 88, 191, 5, 5, 221, 197, 5, 5, 139, 246, 5, 5, 233, 193, 5, 5, 168, 197, 5, 5, 164, 174, 5, 5, 168, 208, 5, 5, 7, 246, 5, 5, 249, 205, 5, 5, 119, 174, 5, 5, 73, 217, 5, 5, 146, 222, 5, 5, 60, 169, 5, 5, 51, 227, 5, 5, 16, 233, 5, 5, 121, 190, 5, 5, 103, 216, 5, 5, 196, 185, 5, 5, 116, 234, 5, 5, 12, 226, 5, 5, 77, 227, 5, 5, 124, 233, 5, 5, 80, 189, 5, 5, 134, 231, 5, 5, 12, 184, 5, 5, 209, 226, 5, 5, 60, 224, 5, 5, 228, 180, 5, 5, 94, 200, 5, 5, 188, 183, 5, 5, 36, 196, 5, 5, 253, 189, 5, 5, 113, 246, 5, 5, 110, 236, 5, 5, 141, 196, 5, 5, 128, 183, 5, 5, 239, 192, 5, 5, 234, 234, 5, 5, 141, 218, 5, 5, 113, 249, 5, 5, 243, 173, 5, 5, 64, 181, 5, 5, 172, 202, 5, 5, 18, 206, 5, 5, 85, 203, 5, 5, 71, 175, 5, 5, 7, 218, 5, 5, 226, 220, 5, 5, 160, 214, 5, 5, 161, 214, 5, 5, 184, 248, 5, 5, 162, 214, 5, 5, 137, 228, 5, 5, 162, 215, 5, 5, 140, 232, 5, 5, 113, 222, 5, 5, 9, 229, 5, 5, 218, 214, 5, 5, 196, 225, 5, 5, 138, 228, 5, 5, 21, 185, 5, 5, 188, 228, 5, 5, 243, 169, 5, 5, 162, 248, 5, 5, 65, 172, 5, 5, 191, 187, 5, 5, 48, 177, 5, 5, 90, 206, 5, 5, 191, 204, 5, 5, 219, 214, 5, 5, 7, 245, 5, 5, 119, 170, 5, 5, 82, 176, 5, 5, 90, 204, 5, 5, 141, 208, 5, 5, 92, 214, 5, 5, 84, 216, 5, 5, 104, 226, 5, 5, 114, 249, 5, 5, 135, 187, 5, 5, 172, 249, 5, 5, 244, 230, 5, 5, 36, 249, 5, 5, 57, 234, 5, 5, 98, 203, 5, 5, 207, 195, 5, 5, 192, 182, 5, 5, 83, 219, 5, 5, 249, 208, 5, 5, 80, 242, 5, 5, 143, 249, 5, 5, 90, 182, 5, 5, 19, 206, 5, 5, 242, 234, 5, 5, 69, 240, 5, 5, 190, 220, 5, 5, 11, 215, 5, 5, 247, 234, 5, 5, 177, 232, 5, 5, 170, 193, 5, 5, 16, 199, 5, 5, 143, 232, 5, 5, 76, 219, 5, 5, 142, 208, 5, 5, 49, 192, 5, 5, 201, 180, 5, 5, 35, 175, 5, 5, 61, 179, 5, 5, 250, 208, 5, 5, 188, 208, 5, 5, 210, 242, 5, 5, 186, 233, 5, 5, 117, 169, 5, 5, 93, 223, 5, 5, 21, 217, 5, 5, 175, 170, 5, 5, 145, 190, 5, 5, 181, 185, 5, 5, 176, 176, 5, 5, 171, 208, 5, 5, 114, 229, 5, 5, 117, 184, 5, 5, 70, 229, 5, 5, 67, 227, 5, 5, 3, 186, 5, 5, 198, 208, 5, 5, 130, 233, 5, 5, 65, 203, 5, 5, 116, 249, 5, 5, 219, 170, 5, 5, 132, 243, 5, 5, 203, 239, 5, 5, 17, 226, 5, 5, 154, 186, 5, 5, 40, 247, 5, 5, 133, 243, 5, 5, 72, 174, 5, 5, 36, 173, 5, 5, 201, 177, 5, 5, 197, 231, 5, 5, 211, 242, 5, 5, 144, 249, 5, 5, 169, 234, 5, 5, 142, 221, 5, 5, 224, 228, 5, 5, 182, 185, 5, 5, 25, 235, 5, 5, 136, 240, 5, 5, 252, 195, 5, 5, 116, 239, 5, 5, 164, 248, 5, 5, 189, 210, 5, 5, 56, 169, 5, 5, 111, 186, 5, 5, 24, 212, 5, 5, 194, 247, 5, 5, 178, 193, 5, 5, 98, 176, 5, 5, 241, 217, 5, 5, 10, 223, 5, 5, 135, 243, 5, 5, 199, 231, 5, 5, 23, 177, 5, 5, 11, 231, 5, 5, 220, 218, 5, 5, 174, 225, 5, 5, 124, 184, 5, 5, 95, 243, 5, 5, 247, 173, 5, 5, 67, 185, 5, 5, 61, 219, 5, 5, 201, 189, 5, 5, 196, 243, 5, 5, 231, 225, 5, 5, 147, 232, 5, 5, 29, 224, 5, 5, 241, 245, 5, 5, 157, 249, 5, 5, 175, 225, 5, 5, 244, 190, 5, 5, 71, 172, 5, 5, 135, 170, 5, 5, 146, 183, 5, 5, 240, 176, 5, 5, 34, 202, 5, 5, 147, 224, 5, 5, 177, 236, 5, 5, 96, 231, 5, 5, 159, 249, 5, 5, 166, 201, 5, 5, 174, 197, 5, 5, 171, 215, 5, 5, 34, 207, 5, 5, 153, 246, 5, 5, 6, 206, 5, 5, 217, 209, 5, 5, 111, 213, 5, 5, 15, 203, 5, 5, 60, 208, 5, 5, 85, 235, 5, 5, 146, 249, 5, 5, 66, 192, 5, 5, 105, 220, 5, 5, 100, 224, 5, 5, 98, 182, 5, 5, 24, 206, 5, 5, 250, 169, 5, 5, 91, 238, 5, 5, 217, 219, 5, 5, 164, 209, 5, 5, 166, 237, 5, 5, 231, 215, 5, 5, 172, 191, 5, 5, 4, 232, 5, 5, 13, 193, 5, 5, 61, 220, 5, 5, 133, 234, 5, 5, 71, 218, 5, 5, 86, 235, 5, 5, 104, 224, 5, 5, 8, 216, 5, 5, 113, 244, 5, 5, 32, 250, 5, 5, 119, 246, 5, 5, 149, 220, 5, 5, 187, 219, 5, 5, 249, 203, 5, 5, 115, 242, 5, 5, 66, 182, 5, 5, 147, 203, 5, 5, 88, 176, 5, 5, 8, 250, 5, 5, 75, 194, 5, 5, 116, 218, 5, 5, 89, 173, 5, 5, 88, 219, 5, 5, 134, 234, 5, 5, 135, 191, 5, 5, 75, 190, 5, 5, 218, 249, 5, 5, 26, 231, 5, 5, 233, 215, 5, 5, 146, 243, 5, 5, 162, 177, 5, 5, 144, 221, 5, 5, 138, 191, 5, 5, 152, 231, 5, 5, 4, 220, 5, 5, 251, 193, 5, 5, 113, 194, 5, 5, 138, 190, 5, 5, 20, 244, 5, 5, 184, 176, 5, 5, 205, 233, 5, 5, 244, 194, 5, 5, 150, 170, 5, 5, 47, 178, 5, 5, 140, 239, 5, 5, 120, 211, 5, 5, 105, 224, 5, 5, 141, 239, 5, 5, 68, 220, 5, 5, 174, 234, 5, 5, 28, 193, 5, 5, 75, 171, 5, 5, 142, 215, 5, 5, 241, 200, 5, 5, 35, 214, 5, 5, 157, 235, 5, 5, 220, 249, 5, 5, 189, 214, 5, 5, 191, 212, 5, 5, 222, 237, 5, 5, 146, 209, 5, 5, 228, 201, 5, 5, 141, 233, 5, 5, 221, 249, 5, 5, 227, 199, 5, 5, 22, 241, 5, 5, 252, 234, 5, 5, 173, 201, 5, 5, 12, 232, 5, 5, 54, 173, 5, 5, 247, 194, 5, 5, 108, 205, 5, 5, 55, 173, 5, 5, 103, 240, 5, 5, 54, 172, 5, 5, 97, 218, 5, 5, 16, 232, 5, 5, 231, 249, 5, 5, 223, 249, 5, 5, 169, 206, 5, 5, 63, 202, 5, 5, 156, 178, 5, 5, 25, 201, 5, 5, 185, 201, 5, 5, 74, 203, 5, 5, 58, 223, 5, 5, 150, 196, 5, 5, 115, 241, 5, 5, 47, 189, 5, 5, 170, 247, 5, 5, 230, 231, 5, 5, 88, 185, 5, 5, 255, 242, 5, 5, 103, 192, 5, 5, 99, 228, 5, 5, 201, 199, 5, 5, 198, 192, 5, 5, 153, 218, 5, 5, 29, 229, 5, 5, 71, 243, 5, 5, 52, 189, 5, 5, 42, 176, 5, 5, 210, 212, 5, 5, 18, 220, 5, 5, 5, 185, 5, 5, 234, 209, 5, 5, 157, 218, 5, 5, 148, 249, 5, 5, 76, 206, 5, 5, 252, 182, 5, 5, 172, 246, 5, 5, 163, 229, 5, 5, 162, 182, 5, 5, 27, 248, 5, 5, 246, 245, 5, 5, 93, 211, 5, 5, 122, 222, 5, 5, 110, 232, 5, 5, 240, 184, 5, 5, 153, 233, 5, 5, 221, 183, 5, 5, 50, 203, 5, 5, 205, 249, 5, 5, 28, 221, 5, 5, 115, 231, 5, 5, 91, 246, 5, 5, 76, 247, 5, 5, 175, 247, 5, 5, 37, 169, 5, 5, 222, 183, 5, 5, 161, 175, 5, 5, 220, 220, 5, 5, 3, 239, 5, 5, 15, 196, 5, 5, 177, 177, 5, 5, 59, 169, 5, 5, 197, 218, 5, 5, 77, 195, 5, 5, 109, 199, 5, 5, 114, 218, 5, 5, 186, 179, 5, 5, 75, 188, 5, 5, 74, 191, 5, 5, 159, 185, 5, 5, 68, 235, 5, 5, 136, 194, 5, 5, 150, 206, 5, 5, 199, 197, 5, 5, 202, 194, 5, 5, 46, 227, 5, 5, 192, 185, 5, 5, 252, 232, 5, 5, 158, 197, 5, 5, 122, 234, 5, 5, 122, 231, 5, 5, 65, 224, 5, 5, 161, 225, 5, 5, 220, 192, 5, 5, 253, 225, 5, 5, 22, 192, 5, 5, 221, 234, 5, 5, 72, 204, 5, 5, 59, 240, 5, 5, 222, 234, 5, 5, 215, 214, 5, 5, 185, 182, 5, 5, 78, 176, 5, 5, 189, 220, 5, 5, 129, 187, 5, 5, 141, 221, 5, 5, 252, 228, 5, 5, 138, 208, 5, 5, 107, 222, 5, 5, 236, 169, 5, 5, 59, 175, 5, 5, 192, 225, 5, 5, 188, 204, 5, 5, 6, 218, 5, 5, 99, 249, 5, 5, 2, 245, 5, 5, 125, 228, 5, 5, 13, 206, 5, 5, 154, 248, 5, 5, 14, 249, 5, 5, 11, 199, 5, 5, 170, 232, 5, 5, 153, 193, 5, 5, 166, 208, 5, 5, 5, 201, 5, 5, 6, 226, 5, 5, 196, 208, 5, 5, 30, 192, 5, 5, 51, 243, 5, 5, 89, 191, 5, 5, 73, 223, 5, 5, 59, 181, 5, 5, 193, 180, 5, 5, 215, 170, 5, 5, 131, 232, 5, 5, 253, 185, 5, 5, 157, 248, 5, 5, 144, 190, 5, 5, 45, 179, 5, 5, 250, 205, 5, 5, 170, 170, 5, 5, 244, 208, 5, 5, 193, 242, 5, 5, 240, 195, 5, 5, 110, 184, 5, 5, 57, 219, 5, 5, 238, 245, 5, 5, 141, 183, 5, 5, 14, 177, 5, 5, 60, 185, 5, 5, 25, 173, 5, 5, 16, 202, 5, 5, 71, 239, 5, 5, 163, 193, 5, 5, 58, 171, 5, 5, 95, 213, 5, 5, 27, 207, 5, 5, 161, 201, 5, 5, 70, 240, 5, 5, 143, 224, 5, 5, 159, 209, 5, 5, 215, 219, 5, 5, 131, 233, 5, 5, 135, 203, 5, 5, 139, 220, 5, 5, 106, 244, 5, 5, 104, 194, 5, 5, 141, 215, 5, 5, 4, 216, 5, 5, 130, 190, 5, 5, 238, 194, 5, 5, 219, 249, 5, 5, 154, 183, 5, 5, 154, 233, 5, 5, 11, 185, 5, 5, 157, 183, 5, 5, 16, 218, 5, 5, 158, 183, 5, 5, 10, 215, 5, 5, 156, 171, 5, 5, 192, 213, 5, 5, 225, 232, 5, 5, 111, 245, 5, 5, 22, 185, 5, 5, 71, 240, 5, 5, 76, 240, 5, 5, 223, 216, 5, 5, 221, 231, 5, 5, 234, 186, 5, 5, 9, 250, 5, 5, 52, 225, 5, 5, 79, 174, 5, 5, 7, 215, 5, 5, 230, 229, 5, 5, 94, 240, 5, 5, 174, 201, 5, 5, 59, 225, 5, 5, 109, 205, 5, 5, 244, 186, 5, 5, 176, 228, 5, 5, 175, 228, 5, 5, 255, 184, 5, 5, 180, 228, 5, 5, 18, 188, 5, 5, 108, 206, 5, 5, 104, 206, 5, 5, 235, 183, 5, 5, 29, 221, 5, 5, 32, 182, 5, 5, 55, 186, 5, 5, 9, 248, 5, 5, 227, 196, 5, 5, 81, 207, 5, 5, 85, 186, 5, 5, 239, 208, 5, 5, 14, 185, 5, 5, 212, 168, 5, 5, 171, 186, 5, 5, 7, 226, 5, 5, 165, 197, 5, 5, 8, 184, 5, 5, 107, 221, 5, 5, 168, 236, 5, 5, 101, 245, 5, 5, 244, 249, 5, 5, 173, 186, 5, 5, 225, 248, 5, 5, 67, 242, 5, 5, 14, 247, 5, 5, 70, 168, 5, 5, 41, 182, 5, 5, 8, 201, 5, 5, 173, 222, 5, 5, 83, 223, 5, 5, 16, 213, 5, 5, 208, 206, 5, 5, 193, 204, 5, 5, 217, 168, 5, 5, 43, 182, 5, 5, 35, 202, 5, 5, 196, 173, 5, 5, 119, 229, 5, 5, 126, 170, 5, 5, 14, 193, 5, 5, 112, 244, 5, 5, 26, 175, 5, 5, 197, 204, 5, 5, 67, 192, 5, 5, 175, 197, 5, 5, 132, 208, 5, 5, 110, 245, 5, 5, 121, 206, 5, 5, 25, 213, 5, 5, 85, 192, 5, 5, 91, 192, 5, 5, 48, 206, 5, 5, 39, 237, 5, 5, 82, 208, 5, 5, 177, 215, 5, 5, 130, 177, 5, 5, 188, 207, 5, 5, 55, 237, 5, 5, 71, 241, 5, 5, 68, 241, 5, 5, 26, 183, 5, 5, 179, 168, 5, 5, 35, 185, 5, 5, 104, 237, 5, 5, 86, 186, 5, 5, 183, 215, 5, 5, 239, 243, 5, 5, 36, 185, 5, 5, 73, 204, 5, 5, 39, 239, 5, 5, 48, 248, 5, 5, 207, 178, 5, 5, 94, 234, 5, 5, 227, 215, 5, 5, 143, 238, 5, 5, 132, 232, 5, 5, 208, 218, 5, 5, 55, 218, 5, 5, 56, 218, 5, 5, 229, 215, 5, 5, 122, 190, 5, 5, 140, 224, 5, 5, 135, 232, 5, 5, 72, 239, 5, 5, 110, 237, 5, 5, 199, 215, 5, 5, 118, 215, 5, 5, 195, 241, 5, 5, 183, 185, 5, 5, 113, 195, 5, 5, 92, 225, 5, 5, 216, 242, 5, 5, 12, 231, 5, 5, 100, 182, 5, 5, 108, 220, 5, 5, 45, 183, 5, 5, 109, 220, 5, 5, 47, 203, 5, 5, 171, 201, 5, 5, 185, 185, 5, 5, 126, 211, 5, 5, 238, 215, 5, 5, 103, 173, 5, 5, 226, 241, 5, 5, 243, 185, 5, 5, 15, 239, 5, 5, 18, 175, 5, 5, 211, 229, 5, 5, 7, 183, 5, 5, 67, 189, 5, 5, 235, 174, 5, 5, 236, 174, 5, 5, 35, 177, 5, 5, 254, 183, 5, 5, 184, 231, 5, 5, 219, 169, 5, 5, 40, 239, 5, 5, 127, 200, 5, 5, 41, 239, 5, 5, 224, 212, 5, 5, 74, 204, 5, 5, 200, 204, 5, 5, 247, 246, 5, 5, 147, 217, 5, 5, 49, 231, 5, 5, 119, 233, 5, 5, 196, 232, 5, 5, 227, 230, 5, 5, 228, 230, 5, 5, 242, 198, 5, 5, 50, 216, 5, 5, 90, 191, 5, 5, 18, 191, 5, 5, 12, 233, 5, 5, 56, 220, 5, 5, 81, 189, 5, 5, 99, 194, 5, 5, 234, 178, 5, 5, 49, 177, 5, 5, 215, 182, 5, 5, 145, 178, 5, 5, 58, 220, 5, 5, 252, 174, 5, 5, 62, 248, 5, 5, 240, 193, 5, 5, 26, 235, 5, 5, 3, 175, 5, 5, 68, 213, 5, 5, 117, 249, 5, 5, 113, 228, 5, 5, 32, 191, 5, 5, 119, 189, 5, 5, 206, 188, 5, 5, 233, 188, 5, 5, 57, 188, 5, 5, 119, 168, 5, 5, 98, 213, 5, 5, 128, 239, 5, 5, 16, 203, 5, 5, 121, 189, 5, 5, 122, 168, 5, 5, 250, 203, 5, 5, 145, 212, 5, 5, 59, 216, 5, 5, 125, 168, 5, 5, 67, 182, 5, 5, 129, 239, 5, 5, 139, 191, 5, 5, 36, 235, 5, 5, 153, 178, 5, 5, 221, 237, 5, 5, 107, 201, 5, 5, 106, 201, 5, 5, 192, 198, 5, 5, 116, 207, 5, 5, 151, 245, 5, 5, 124, 214, 5, 5, 15, 186, 5, 5, 18, 186, 5, 5, 16, 186, 5, 5, 190, 179, 5, 5, 200, 179, 5, 5, 220, 181, 5, 5, 230, 222, 5, 5, 70, 219, 5, 5, 13, 210, 5, 5, 14, 210, 5, 5, 182, 180, 5, 5, 188, 201, 5, 5, 204, 226, 5, 5, 70, 249, 5, 5, 174, 185, 5, 5, 209, 173, 5, 5, 123, 188, 5, 5, 80, 243, 5, 5, 213, 168, 5, 5, 225, 212, 5, 5, 140, 238, 5, 5, 131, 223, 5, 5, 187, 217, 5, 5, 58, 193, 5, 5, 85, 197, 5, 5, 188, 191, 5, 5, 119, 175, 5, 5, 127, 201, 5, 5, 100, 206, 5, 5, 142, 192, 5, 5, 90, 225, 5, 5, 222, 211, 5, 5, 45, 205, 5, 5, 129, 211, 5, 5, 126, 194, 5, 5, 70, 212, 5, 5, 152, 244, 5, 5, 47, 205, 5, 5, 71, 192, 5, 5, 134, 201, 5, 5, 237, 191, 5, 5, 23, 241, 5, 5, 147, 209, 5, 5, 89, 190, 5, 5, 246, 181, 5, 5, 241, 238, 5, 5, 234, 179, 5, 5, 234, 237, 5, 5, 111, 177, 5, 5, 169, 223, 5, 5, 117, 216, 5, 5, 81, 243, 5, 5, 213, 174, 5, 5, 221, 177, 5, 5, 222, 177, 5, 5, 68, 189, 5, 5, 193, 185, 5, 5, 129, 246, 5, 5, 41, 212, 5, 5, 2, 211, 5, 5, 224, 177, 5, 5, 233, 245, 5, 5, 228, 173, 5, 5, 75, 204, 5, 5, 39, 246, 5, 5, 74, 241, 5, 5, 68, 228, 5, 5, 121, 203, 5, 5, 226, 204, 5, 5, 134, 246, 5, 5, 59, 211, 5, 5, 140, 226, 5, 5, 234, 181, 5, 5, 44, 236, 5, 5, 184, 226, 5, 5, 150, 186, 5, 5, 230, 234, 5, 5, 243, 202, 5, 5, 130, 188, 5, 5, 160, 221, 5, 5, 204, 203, 5, 5, 155, 213, 5, 5, 39, 196, 5, 5, 104, 197, 5, 5, 42, 170, 5, 5, 14, 179, 5, 5, 176, 187, 5, 5, 175, 229, 5, 5, 243, 234, 5, 5, 162, 213, 5, 5, 166, 176, 5, 5, 9, 180, 5, 5, 142, 210, 5, 5, 232, 226, 5, 5, 49, 197, 5, 5, 177, 176, 5, 5, 235, 199, 5, 5, 9, 203, 5, 5, 39, 200, 5, 5, 178, 176, 5, 5, 10, 203, 5, 5, 225, 229, 5, 5, 40, 200, 5, 5, 17, 203, 5, 5, 22, 203, 5, 5, 181, 176, 5, 5, 120, 168, 5, 5, 137, 221, 5, 5, 176, 223, 5, 5, 211, 226, 5, 5, 46, 200, 5, 5, 28, 211, 5, 5, 154, 246, 5, 5, 243, 177, 5, 5, 74, 211, 5, 5, 65, 216, 5, 5, 171, 210, 5, 5, 212, 226, 5, 5, 214, 226, 5, 5, 2, 205, 5, 5, 13, 242, 5, 5, 9, 242, 5, 5, 96, 245, 5, 5, 128, 223, 5, 5, 241, 223, 5, 5, 73, 239, 5, 5, 10, 224, 5, 5, 136, 223, 5, 5, 102, 245, 5, 5, 103, 245, 5, 5, 12, 219, 5, 5, 226, 238, 5, 5, 136, 201, 5, 5, 6, 192, 5, 5, 41, 217, 5, 5, 21, 199, 5, 5, 153, 172, 5, 5, 90, 185, 5, 5, 174, 171, 5, 5, 149, 190, 5, 5, 183, 190, 5, 5, 206, 174, 5, 5, 158, 218, 5, 5, 71, 233, 5, 5, 45, 219, 5, 5, 162, 247, 5, 5, 235, 242, 5, 5, 169, 180, 5, 5, 6, 185, 5, 5, 194, 218, 5, 5, 195, 184, 5, 5, 201, 172, 5, 5, 252, 229, 5, 5, 127, 173, 5, 5, 81, 186, 5, 5, 92, 246, 5, 5, 247, 187, 5, 5, 169, 183, 5, 5, 201, 182, 5, 5, 25, 183, 5, 5, 105, 212, 5, 5, 73, 212, 5, 5, 137, 193, 5, 5, 154, 182, 5, 5, 175, 248, 5, 5, 31, 241, 5, 5, 235, 209, 5, 5, 89, 168, 5, 5, 22, 199, 5, 5, 60, 218, 5, 5, 212, 185, 5, 5, 19, 242, 5, 5, 193, 229, 5, 5, 128, 237, 5, 5, 117, 214, 5, 5, 252, 212, 5, 5, 131, 215, 5, 5, 183, 230, 5, 5, 184, 230, 5, 5, 225, 182, 5, 5, 23, 199, 5, 5, 58, 196, 5, 5, 3, 225, 5, 5, 93, 221, 5, 5, 97, 247, 5, 5, 52, 245, 5, 5, 104, 233, 5, 5, 96, 221, 5, 5, 95, 229, 5, 5, 180, 247, 5, 5, 187, 179, 5, 5, 249, 177, 5, 5, 227, 182, 5, 5, 195, 168, 5, 5, 182, 171, 5, 5, 140, 217, 5, 5, 92, 226, 5, 5, 170, 169, 5, 5, 145, 186, 5, 5, 227, 224, 5, 5, 66, 249, 5, 5, 207, 182, 5, 5, 97, 246, 5, 5, 177, 210, 5, 5, 180, 213, 5, 5, 93, 249, 5, 5, 50, 184, 5, 5, 74, 212, 5, 5, 178, 245, 5, 5, 27, 232, 5, 5, 41, 250, 5, 5, 107, 212, 5, 5, 103, 193, 5, 5, 151, 221, 5, 5, 98, 246, 5, 5, 87, 169, 5, 5, 182, 209, 5, 5, 105, 217, 5, 5, 175, 189, 5, 5, 210, 198, 5, 5, 211, 174, 5, 5, 32, 240, 5, 5, 144, 234, 5, 5, 30, 237, 5, 5, 221, 204, 5, 5, 61, 180, 5, 5, 87, 198, 5, 5, 82, 205, 5, 5, 83, 196, 5, 5, 144, 211, 5, 5, 55, 236, 5, 5, 52, 212, 5, 5, 132, 199, 5, 5, 134, 238, 5, 5, 8, 187, 5, 5, 214, 187, 5, 5, 96, 185, 5, 5, 65, 180, 5, 5, 151, 174, 5, 5, 165, 194, 5, 5, 74, 234, 5, 5, 75, 234, 5, 5, 229, 181, 5, 5, 78, 168, 5, 5, 14, 189, 5, 5, 72, 213, 5, 5, 122, 210, 5, 5, 94, 249, 5, 5, 35, 187, 5, 5, 106, 174, 5, 5, 167, 226, 5, 5, 104, 246, 5, 5, 36, 177, 5, 5, 150, 207, 5, 5, 112, 233, 5, 5, 37, 177, 5, 5, 113, 233, 5, 5, 233, 206, 5, 5, 232, 206, 5, 5, 121, 192, 5, 5, 114, 233, 5, 5, 52, 211, 5, 5, 146, 199, 5, 5, 143, 219, 5, 5, 252, 243, 5, 5, 125, 214, 5, 5, 85, 210, 5, 5, 88, 208, 5, 5, 104, 211, 5, 5, 35, 206, 5, 5, 149, 228, 5, 5, 193, 171, 5, 5, 196, 229, 5, 5, 126, 228, 5, 5, 96, 234, 5, 5, 210, 194, 5, 5, 96, 196, 5, 5, 75, 241, 5, 5, 8, 189, 5, 5, 92, 177, 5, 5, 199, 203, 5, 5, 92, 220, 5, 5, 245, 226, 5, 5, 154, 207, 5, 5, 49, 176, 5, 5, 101, 221, 5, 5, 133, 215, 5, 5, 170, 178, 5, 5, 133, 197, 5, 5, 227, 176, 5, 5, 95, 234, 5, 5, 249, 233, 5, 5, 239, 227, 5, 5, 77, 212, 5, 5, 111, 174, 5, 5, 110, 210, 5, 5, 19, 214, 5, 5, 101, 180, 5, 5, 240, 227, 5, 5, 246, 210, 5, 5, 87, 210, 5, 5, 35, 213, 5, 5, 92, 186, 5, 5, 110, 206, 5, 5, 147, 202, 5, 5, 37, 203, 5, 5, 57, 245, 5, 5, 45, 177, 5, 5, 49, 195, 5, 5, 96, 203, 5, 5, 13, 223, 5, 5, 13, 213, 5, 5, 238, 170, 5, 5, 89, 206, 5, 5, 74, 213, 5, 5, 193, 201, 5, 5, 30, 200, 5, 5, 18, 176, 5, 5, 27, 178, 5, 5, 120, 226, 5, 5, 111, 210, 5, 5, 32, 196, 5, 5, 250, 209, 5, 5, 140, 200, 5, 5, 18, 238, 5, 5, 65, 236, 5, 5, 172, 218, 5, 5, 102, 221, 5, 5, 79, 176, 5, 5, 64, 184, 5, 5, 61, 184, 5, 5, 238, 196, 5, 5, 203, 182, 5, 5, 190, 215, 5, 5, 151, 228, 5, 5, 71, 180, 5, 5, 14, 213, 5, 5, 172, 187, 5, 5, 107, 249, 5, 5, 5, 180, 5, 5, 12, 230, 5, 5, 169, 207, 5, 5, 141, 210, 5, 5, 105, 248, 5, 5, 246, 174, 5, 5, 215, 176, 5, 5, 114, 206, 5, 5, 220, 211, 5, 5, 49, 179, 5, 5, 46, 168, 5, 5, 11, 234, 5, 5, 92, 210, 5, 5, 214, 222, 5, 5, 244, 222, 5, 5, 43, 210, 5, 5, 127, 243, 5, 5, 106, 247, 5, 5, 253, 218, 5, 5, 179, 181, 5, 5, 158, 219, 5, 5, 116, 227, 5, 5, 122, 237, 5, 5, 39, 194, 5, 5, 62, 229, 5, 5, 219, 194, 5, 5, 175, 241, 5, 5, 65, 193, 5, 5, 198, 180, 5, 5, 103, 170, 5, 5, 137, 173, 5, 5, 111, 184, 5, 5, 136, 232, 5, 5, 187, 210, 5, 5, 169, 208, 5, 5, 70, 228, 5, 5, 226, 225, 5, 5, 182, 227, 5, 5, 63, 169, 5, 5, 138, 215, 5, 5, 252, 205, 5, 5, 71, 228, 5, 5, 109, 222, 5, 5, 119, 225, 5, 5, 16, 224, 5, 5, 160, 248, 5, 5, 133, 185, 5, 5, 83, 239, 5, 5, 210, 171, 5, 5, 127, 203, 5, 5, 143, 192, 5, 5, 17, 213, 5, 5, 126, 233, 5, 5, 44, 185, 5, 5, 2, 206, 5, 5, 80, 171, 5, 5, 209, 211, 5, 5, 142, 173, 5, 5, 189, 205, 5, 5, 178, 187, 5, 5, 55, 212, 5, 5, 198, 175, 5, 5, 249, 170, 5, 5, 127, 194, 5, 5, 193, 183, 5, 5, 123, 225, 5, 5, 70, 209, 5, 5, 2, 231, 5, 5, 88, 248, 5, 5, 22, 206, 5, 5, 179, 194, 5, 5, 180, 240, 5, 5, 229, 205, 5, 5, 32, 247, 5, 5, 118, 235, 5, 5, 64, 177, 5, 5, 30, 204, 5, 5, 43, 227, 5, 5, 184, 212, 5, 5, 191, 210, 5, 5, 201, 175, 5, 5, 140, 203, 5, 5, 51, 177, 5, 5, 57, 187, 5, 5, 28, 233, 5, 5, 221, 225, 5, 5, 96, 182, 5, 5, 152, 246, 5, 5, 182, 194, 5, 5, 145, 220, 5, 5, 168, 189, 5, 5, 50, 177, 5, 5, 110, 219, 5, 5, 178, 232, 5, 5, 72, 224, 5, 5, 65, 183, 5, 5, 84, 240, 5, 5, 106, 220, 5, 5, 72, 248, 5, 5, 6, 178, 5, 5, 156, 199, 5, 5, 115, 211, 5, 5, 216, 227, 5, 5, 23, 203, 5, 5, 142, 170, 5, 5, 208, 240, 5, 5, 245, 197, 5, 5, 188, 175, 5, 5, 130, 239, 5, 5, 149, 197, 5, 5, 149, 200, 5, 5, 44, 203, 5, 5, 218, 219, 5, 5, 217, 227, 5, 5, 116, 186, 5, 5, 179, 214, 5, 5, 56, 177, 5, 5, 85, 171, 5, 5, 152, 235, 5, 5, 101, 210, 5, 5, 18, 230, 5, 5, 3, 244, 5, 5, 154, 232, 5, 5, 7, 171, 5, 5, 60, 234, 5, 5, 163, 199, 5, 5, 152, 200, 5, 5, 79, 236, 5, 5, 105, 205, 5, 5, 243, 191, 5, 5, 158, 243, 5, 5, 144, 206, 5, 5, 166, 217, 5, 5, 4, 244, 5, 5, 49, 206, 5, 5, 152, 210, 5, 5, 159, 243, 5, 5, 250, 173, 5, 5, 47, 215, 5, 5, 255, 229, 5, 5, 21, 187, 5, 5, 242, 243, 5, 5, 248, 203, 5, 5, 18, 187, 5, 5, 82, 249, 5, 5, 195, 194, 5, 5, 3, 211, 5, 5, 99, 175, 5, 5, 234, 192, 5, 5, 40, 185, 5, 5, 100, 175, 5, 5, 143, 208, 5, 5, 23, 211, 5, 5, 197, 246, 5, 5, 230, 246, 5, 5, 122, 185, 5, 5, 40, 194, 5, 5, 201, 179, 5, 5, 109, 246, 5, 5, 98, 234, 5, 5, 241, 224, 5, 5, 214, 245, 5, 5, 186, 196, 5, 5, 111, 232, 5, 5, 140, 241, 5, 5, 202, 172, 5, 5, 188, 237, 5, 5, 148, 241, 5, 5, 19, 175, 5, 5, 127, 231, 5, 5, 128, 231, 5, 5, 200, 196, 5, 5, 192, 241, 5, 5, 249, 241, 5, 5, 58, 234, 5, 5, 50, 197, 5, 5, 205, 196, 5, 5, 255, 233, 5, 5, 237, 220, 5, 5, 242, 220, 5, 5, 247, 220, 5, 5, 178, 196, 5, 5, 42, 222, 5, 5, 17, 225, 5, 5, 142, 222, 5, 5, 69, 222, 5, 5, 124, 226, 5, 5, 83, 225, 5, 5, 55, 214, 5, 5, 56, 214, 5, 5, 213, 186, 5, 5, 76, 190, 5, 5, 133, 226, 5, 5, 221, 175, 5, 5, 30, 223, 5, 5, 193, 231, 5, 5, 104, 230, 5, 5, 128, 247, 5, 5, 110, 178, 5, 5, 211, 175, 5, 5, 90, 220, 5, 5, 190, 238, 5, 5, 83, 182, 5, 5, 231, 211, 5, 5, 6, 225, 5, 5, 85, 182, 5, 5, 24, 169, 5, 5, 212, 175, 5, 5, 78, 188, 5, 5, 170, 182, 5, 5, 3, 169, 5, 5, 37, 170, 5, 5, 84, 205, 5, 5, 81, 247, 5, 5, 27, 194, 5, 5, 120, 172, 5, 5, 76, 204, 5, 5, 245, 247, 5, 5, 54, 244, 5, 5, 80, 181, 5, 5, 194, 171, 5, 5, 32, 232, 5, 5, 150, 185, 5, 5, 213, 175, 5, 5, 185, 231, 5, 5, 87, 205, 5, 5, 242, 216, 5, 5, 62, 212, 5, 5, 229, 230, 5, 5, 125, 199, 5, 5, 13, 184, 5, 5, 32, 244, 5, 5, 129, 183, 5, 5, 17, 202, 5, 5, 202, 220, 5, 5, 186, 241, 5, 5, 56, 201, 5, 5, 246, 226, 5, 5, 68, 242, 5, 5, 63, 229, 5, 5, 211, 171, 5, 5, 29, 208, 5, 5, 227, 210, 5, 5, 244, 196, 5, 5, 215, 190, 5, 5, 99, 222, 5, 5, 17, 199, 5, 5, 220, 170, 5, 5, 214, 209, 5, 5, 213, 179, 5, 5, 148, 178, 5, 5, 78, 169, 5, 5, 211, 172, 5, 5, 117, 239, 5, 5, 204, 223, 5, 5, 127, 172, 5, 5, 145, 172, 5, 5, 48, 205, 5, 5, 204, 177, 5, 5, 120, 235, 5, 5, 218, 237, 5, 5, 27, 175, 5, 5, 182, 178, 5, 5, 62, 178, 5, 5, 224, 215, 5, 5, 141, 205, 5, 5, 213, 238, 5, 5, 162, 192, 5, 5, 79, 194, 5, 5, 238, 191, 5, 5, 16, 208, 5, 5, 190, 214, 5, 5, 143, 205, 5, 5, 110, 205, 5, 5, 58, 173, 5, 5, 2, 223, 5, 5, 164, 185, 5, 5, 75, 203, 5, 5, 138, 193, 5, 5, 179, 193, 5, 5, 233, 219, 5, 5, 211, 236, 5, 5, 255, 183, 5, 5, 163, 213, 5, 5, 253, 236, 5, 5, 15, 237, 5, 5, 148, 172, 5, 5, 156, 172, 5, 5, 224, 238, 5, 5, 144, 201, 5, 5, 201, 226, 5, 5, 154, 194, 5, 5, 0, 167, 5, 5, 2, 210, 5, 5, 141, 226, 5, 5, 155, 194, 5, 5, 63, 195, 5, 5, 161, 212, 5, 5, 100, 228, 5, 5, 232, 238, 5, 5, 242, 172, 5, 5, 134, 208, 5, 5, 79, 208, 5, 5, 190, 184, 5, 5, 37, 215, 5, 5, 14, 242, 5, 5, 15, 242, 5, 5, 211, 220, 5, 5, 227, 216, 5, 5, 166, 180, 5, 5, 234, 189, 5, 5, 202, 214, 5, 5, 42, 207, 5, 5, 83, 249, 5, 5, 75, 191, 5, 5, 49, 224, 5, 5, 198, 246, 5, 5, 160, 232, 5, 5, 247, 212, 5, 5, 211, 183, 5, 5, 207, 227, 5, 5, 30, 229, 5, 5, 55, 230, 5, 5, 83, 246, 5, 5, 231, 214, 5, 5, 79, 220, 5, 5, 164, 229, 5, 5, 147, 215, 5, 5, 128, 218, 5, 5, 212, 183, 5, 5, 68, 199, 5, 5, 141, 195, 5, 5, 76, 205, 5, 5, 202, 234, 5, 5, 7, 192, 5, 5, 112, 216, 5, 5, 109, 214, 5, 5, 86, 236, 5, 5, 253, 182, 5, 5, 170, 168, 5, 5, 92, 219, 5, 5, 217, 233, 5, 5, 203, 192, 5, 5, 20, 190, 5, 5, 21, 190, 5, 5, 142, 182, 5, 5, 125, 230, 5, 5, 191, 197, 5, 5, 9, 186, 5, 5, 173, 246, 5, 5, 72, 243, 5, 5, 113, 216, 5, 5, 132, 168, 5, 5, 112, 175, 5, 5, 109, 207, 5, 5, 251, 235, 5, 5, 220, 182, 5, 5, 185, 218, 5, 5, 70, 188, 5, 5, 156, 176, 5, 5, 216, 239, 5, 5, 66, 241, 5, 5, 205, 170, 5, 5, 242, 238, 5, 5, 63, 216, 5, 5, 31, 229, 5, 5, 108, 202, 5, 5, 34, 213, 5, 5, 140, 181, 5, 5, 108, 232, 5, 5, 128, 173, 5, 5, 117, 172, 5, 5, 28, 248, 5, 5, 225, 223, 5, 5, 250, 225, 5, 5, 237, 242, 5, 5, 75, 218, 5, 5, 51, 204, 5, 5, 233, 224, 5, 5, 212, 220, 5, 5, 219, 178, 5, 5, 182, 208, 5, 5, 203, 216, 5, 5, 115, 202, 5, 5, 2, 241, 5, 5, 205, 198, 5, 5, 198, 209, 5, 5, 169, 169, 5, 5, 85, 169, 5, 5, 12, 186, 5, 5, 235, 208, 5, 5, 187, 238, 5, 5, 188, 238, 5, 5, 67, 197, 5, 5, 127, 213, 5, 5, 113, 219, 5, 5, 106, 199, 5, 5, 57, 211, 5, 5, 158, 210, 5, 5, 179, 171, 5, 5, 105, 171, 5, 5, 82, 220, 5, 5, 112, 234, 5, 5, 182, 237, 5, 5, 237, 231, 5, 5, 216, 169, 5, 5, 57, 189, 5, 5, 49, 240, 5, 5, 157, 244, 5, 5, 24, 180, 5, 5, 220, 196, 5, 5, 179, 179, 5, 5, 249, 210, 5, 5, 199, 184, 5, 5, 151, 189, 5, 5, 173, 213, 5, 5, 146, 207, 5, 5, 223, 183, 5, 5, 123, 221, 5, 5, 51, 186, 5, 5, 59, 170, 5, 5, 92, 229, 5, 5, 224, 183, 5, 5, 207, 248, 5, 5, 111, 207, 5, 5, 121, 182, 5, 5, 46, 193, 5, 5, 78, 207, 5, 5, 79, 207, 5, 5, 192, 168, 5, 5, 108, 176, 5, 5, 100, 209, 5, 5, 254, 248, 5, 5, 177, 246, 5, 5, 183, 174, 5, 5, 214, 192, 5, 5, 16, 196, 5, 5, 217, 204, 5, 5, 159, 176, 5, 5, 235, 210, 5, 5, 194, 242, 5, 5, 104, 235, 5, 5, 112, 240, 5, 5, 233, 216, 5, 5, 115, 185, 5, 5, 34, 209, 5, 5, 123, 202, 5, 5, 198, 218, 5, 5, 30, 240, 5, 5, 167, 185, 5, 5, 149, 215, 5, 5, 254, 172, 5, 5, 3, 174, 5, 5, 24, 242, 5, 5, 87, 188, 5, 5, 84, 249, 5, 5, 187, 203, 5, 5, 61, 230, 5, 5, 215, 192, 5, 5, 31, 187, 5, 5, 145, 176, 5, 5, 207, 193, 5, 5, 109, 176, 5, 5, 27, 186, 5, 5, 4, 187, 5, 5, 50, 207, 5, 5, 218, 172, 5, 5, 162, 194, 5, 5, 163, 194, 5, 5, 236, 183, 5, 5, 149, 241, 5, 5, 206, 247, 5, 5, 5, 249, 5, 5, 91, 194, 5, 5, 48, 191, 5, 5, 175, 239, 5, 5, 219, 172, 5, 5, 33, 182, 5, 5, 192, 218, 5, 5, 18, 219, 5, 5, 15, 175, 5, 5, 67, 207, 5, 5, 12, 227, 5, 5, 116, 247, 5, 5, 44, 215, 5, 5, 240, 247, 5, 5, 172, 235, 5, 5, 114, 190, 5, 5, 173, 183, 5, 5, 214, 217, 5, 5, 155, 184, 5, 5, 112, 177, 5, 5, 19, 196, 5, 5, 51, 234, 5, 5, 222, 175, 5, 5, 149, 172, 5, 5, 20, 196, 5, 5, 227, 181, 5, 5, 76, 168, 5, 5, 14, 216, 5, 5, 186, 174, 5, 5, 134, 218, 5, 5, 19, 194, 5, 5, 252, 225, 5, 5, 184, 190, 5, 5, 181, 213, 5, 5, 124, 202, 5, 5, 175, 195, 5, 5, 113, 201, 5, 5, 201, 222, 5, 5, 122, 248, 5, 5, 131, 177, 5, 5, 79, 193, 5, 5, 103, 218, 5, 5, 65, 170, 5, 5, 164, 172, 5, 5, 15, 216, 5, 5, 76, 191, 5, 5, 216, 192, 5, 5, 81, 178, 5, 5, 85, 249, 5, 5, 199, 218, 5, 5, 152, 191, 5, 5, 32, 234, 5, 5, 151, 206, 5, 5, 114, 240, 5, 5, 173, 235, 5, 5, 225, 195, 5, 5, 112, 224, 5, 5, 176, 239, 5, 5, 221, 206, 5, 5, 63, 189, 5, 5, 181, 247, 5, 5, 201, 237, 5, 5, 172, 180, 5, 5, 40, 229, 5, 5, 253, 201, 5, 5, 95, 180, 5, 5, 178, 183, 5, 5, 137, 218, 5, 5, 225, 239, 5, 5, 178, 188, 5, 5, 19, 170, 5, 5, 27, 172, 5, 5, 102, 243, 5, 5, 224, 178, 5, 5, 104, 227, 5, 5, 54, 224, 5, 5, 55, 224, 5, 5, 100, 174, 5, 5, 5, 174, 5, 5, 62, 230, 5, 5, 39, 171, 5, 5, 200, 230, 5, 5, 194, 185, 5, 5, 139, 202, 5, 5, 18, 214, 5, 5, 45, 248, 5, 5, 111, 209, 5, 5, 140, 202, 5, 5, 125, 248, 5, 5, 23, 250, 5, 5, 235, 227, 5, 5, 55, 217, 5, 5, 28, 220, 5, 5, 134, 224, 5, 5, 6, 174, 5, 5, 166, 212, 5, 5, 41, 197, 5, 5, 151, 207, 5, 5, 246, 207, 5, 5, 53, 234, 5, 5, 22, 196, 5, 5, 217, 173, 5, 5, 42, 221, 5, 5, 29, 198, 5, 5, 98, 236, 5, 5, 220, 226, 5, 5, 38, 241, 5, 5, 98, 177, 5, 5, 206, 186, 5, 5, 36, 188, 5, 5, 72, 241, 5, 5, 99, 177, 5, 5, 83, 193, 5, 5, 173, 228, 5, 5, 246, 223, 5, 5, 97, 211, 5, 5, 196, 220, 5, 5, 123, 232, 5, 5, 75, 201, 5, 5, 243, 183, 5, 5, 107, 181, 5, 5, 112, 209, 5, 5, 203, 229, 5, 5, 82, 195, 5, 5, 30, 210, 5, 5, 200, 209, 5, 5, 139, 198, 5, 5, 168, 200, 5, 5, 240, 202, 5, 5, 225, 221, 5, 5, 235, 229, 5, 5, 140, 227, 5, 5, 241, 231, 5, 5, 215, 191, 5, 5, 39, 241, 5, 5, 115, 240, 5, 5, 53, 240, 5, 5, 246, 185, 5, 5, 47, 176, 5, 5, 64, 207, 5, 5, 65, 207, 5, 5, 147, 176, 5, 5, 127, 228, 5, 5, 29, 196, 5, 5, 38, 179, 5, 5, 141, 217, 5, 5, 35, 180, 5, 5, 147, 193, 5, 5, 167, 249, 5, 5, 58, 186, 5, 5, 128, 200, 5, 5, 163, 205, 5, 5, 11, 197, 5, 5, 99, 229, 5, 5, 33, 198, 5, 5, 6, 211, 5, 5, 60, 200, 5, 5, 184, 188, 5, 5, 21, 214, 5, 5, 180, 243, 5, 5, 57, 185, 5, 5, 185, 187, 5, 5, 2, 184, 5, 5, 87, 206, 5, 5, 112, 173, 5, 5, 60, 175, 5, 5, 158, 223, 5, 5, 14, 176, 5, 5, 3, 245, 5, 5, 228, 207, 5, 5, 68, 172, 5, 5, 196, 168, 5, 5, 7, 202, 5, 5, 108, 227, 5, 5, 157, 171, 5, 5, 39, 188, 5, 5, 130, 169, 5, 5, 163, 216, 5, 5, 147, 197, 5, 5, 214, 230, 5, 5, 118, 216, 5, 5, 10, 179, 5, 5, 38, 195, 5, 5, 20, 214, 5, 5, 83, 178, 5, 5, 198, 195, 5, 5, 18, 216, 5, 5, 237, 224, 5, 5, 175, 185, 5, 5, 36, 190, 5, 5, 59, 186, 5, 5, 77, 204, 5, 5, 186, 182, 5, 5, 121, 216, 5, 5, 40, 168, 5, 5, 183, 228, 5, 5, 247, 169, 5, 5, 165, 247, 5, 5, 62, 236, 5, 5, 235, 196, 5, 5, 148, 193, 5, 5, 6, 204, 5, 5, 64, 225, 5, 5, 137, 222, 5, 5, 96, 226, 5, 5, 139, 178, 5, 5, 220, 189, 5, 5, 137, 214, 5, 5, 185, 188, 5, 5, 42, 177, 5, 5, 193, 225, 5, 5, 28, 176, 5, 5, 78, 170, 5, 5, 135, 173, 5, 5, 116, 190, 5, 5, 235, 182, 5, 5, 177, 200, 5, 5, 79, 214, 5, 5, 247, 207, 5, 5, 186, 213, 5, 5, 205, 228, 5, 5, 22, 178, 5, 5, 51, 220, 5, 5, 117, 190, 5, 5, 154, 236, 5, 5, 141, 238, 5, 5, 149, 226, 5, 5, 138, 214, 5, 5, 128, 228, 5, 5, 171, 172, 5, 5, 15, 210, 5, 5, 234, 245, 5, 5, 37, 228, 5, 5, 159, 196, 5, 5, 183, 226, 5, 5, 111, 206, 5, 5, 79, 170, 5, 5, 42, 239, 5, 5, 10, 212, 5, 5, 245, 171, 5, 5, 78, 212, 5, 5, 2, 180, 5, 5, 117, 240, 5, 5, 118, 240, 5, 5, 119, 240, 5, 5, 6, 233, 5, 5, 144, 219, 5, 5, 16, 217, 5, 5, 218, 198, 5, 5, 226, 232, 5, 5, 150, 228, 5, 5, 243, 241, 5, 5, 176, 241, 5, 5, 104, 184, 5, 5, 59, 203, 5, 5, 15, 235, 5, 5, 16, 235, 5, 5, 41, 210, 5, 5, 176, 172, 5, 5, 240, 229, 5, 5, 63, 176, 5, 5, 13, 233, 5, 5, 159, 241, 5, 5, 192, 170, 5, 5, 121, 207, 5, 5, 43, 168, 5, 5, 169, 181, 5, 5, 158, 206, 5, 5, 123, 240, 5, 5, 107, 200, 5, 5, 108, 200, 5, 5, 96, 194, 5, 5, 167, 208, 5, 5, 46, 250, 5, 5, 47, 250, 5, 5, 188, 249, 5, 5, 110, 169, 5, 5, 250, 218, 5, 5, 81, 231, 5, 5, 19, 238, 5, 5, 44, 168, 5, 5, 148, 217, 5, 5, 154, 193, 5, 5, 9, 184, 5, 5, 16, 206, 5, 5, 33, 196, 5, 5, 143, 212, 5, 5, 93, 183, 5, 5, 76, 189, 5, 5, 77, 189, 5, 5, 114, 189, 5, 5, 225, 236, 5, 5, 90, 227, 5, 5, 171, 245, 5, 5, 24, 249, 5, 5, 171, 232, 5, 5, 173, 218, 5, 5, 123, 185, 5, 5, 14, 215, 5, 5, 196, 216, 5, 5, 90, 208, 5, 5, 201, 190, 5, 5, 49, 221, 5, 5, 22, 214, 5, 5, 178, 184, 5, 5, 44, 180, 5, 5, 186, 247, 5, 5, 41, 194, 5, 5, 53, 229, 5, 5, 198, 171, 5, 5, 215, 228, 5, 5, 233, 211, 5, 5, 235, 192, 5, 5, 188, 189, 5, 5, 254, 243, 5, 5, 62, 193, 5, 5, 194, 180, 5, 5, 207, 237, 5, 5, 57, 224, 5, 5, 164, 176, 5, 5, 229, 217, 5, 5, 216, 228, 5, 5, 192, 245, 5, 5, 251, 199, 5, 5, 226, 226, 5, 5, 159, 191, 5, 5, 82, 231, 5, 5, 183, 221, 5, 5, 248, 214, 5, 5, 216, 200, 5, 5, 108, 249, 5, 5, 120, 237, 5, 5, 154, 242, 5, 5, 163, 207, 5, 5, 52, 243, 5, 5, 207, 222, 5, 5, 46, 179, 5, 5, 103, 248, 5, 5, 194, 186, 5, 5, 79, 218, 5, 5, 92, 235, 5, 5, 75, 173, 5, 5, 106, 198, 5, 5, 31, 200, 5, 5, 30, 190, 5, 5, 120, 194, 5, 5, 34, 205, 5, 5, 54, 229, 5, 5, 6, 212, 5, 5, 112, 241, 5, 5, 197, 223, 5, 5, 198, 239, 5, 5, 187, 221, 5, 5, 53, 176, 5, 5, 232, 221, 5, 5, 56, 243, 5, 5, 201, 245, 5, 5, 76, 201, 5, 5, 24, 219, 5, 5, 250, 207, 5, 5, 164, 202, 5, 5, 17, 218, 5, 5, 46, 224, 5, 5, 206, 242, 5, 5, 165, 202, 5, 5, 47, 197, 5, 5, 236, 230, 5, 5, 74, 169, 5, 5, 115, 175, 5, 5, 181, 234, 5, 5, 117, 227, 5, 5, 146, 204, 5, 5, 225, 191, 5, 5, 96, 179, 5, 5, 159, 215, 5, 5, 189, 249, 5, 5, 242, 213, 5, 5, 86, 177, 5, 5, 156, 242, 5, 5, 120, 175, 5, 5, 69, 242, 5, 5, 135, 199, 5, 5, 19, 212, 5, 5, 23, 214, 5, 5, 24, 214, 5, 5, 247, 209, 5, 5, 49, 223, 5, 5, 55, 231, 5, 5, 32, 183, 5, 5, 88, 243, 5, 5, 45, 246, 5, 5, 164, 193, 5, 5, 138, 192, 5, 5, 121, 219, 5, 5, 125, 172, 5, 5, 214, 181, 5, 5, 53, 208, 5, 5, 88, 188, 5, 5, 117, 224, 5, 5, 202, 245, 5, 5, 43, 209, 5, 5, 198, 223, 5, 5, 235, 234, 5, 5, 147, 204, 5, 5, 56, 231, 5, 5, 97, 186, 5, 5, 49, 201, 5, 5, 214, 218, 5, 5, 222, 229, 5, 5, 184, 184, 5, 5, 105, 177, 5, 5, 188, 221, 5, 5, 112, 225, 5, 5, 139, 206, 5, 5, 58, 219, 5, 5, 141, 224, 5, 5, 20, 235, 5, 5, 193, 169, 5, 5, 193, 237, 5, 5, 107, 187, 5, 5, 104, 170, 5, 5, 141, 249, 5, 5, 133, 187, 5, 5, 48, 250, 5, 5, 150, 226, 5, 5, 179, 172, 5, 5, 131, 212, 5, 5, 218, 200, 5, 5, 42, 178, 5, 5, 240, 192, 5, 5, 18, 202, 5, 5, 52, 221, 5, 5, 68, 201, 5, 5, 70, 242, 5, 5, 84, 204, 5, 5, 128, 240, 5, 5, 124, 209, 5, 5, 165, 179, 5, 5, 151, 226, 5, 5, 106, 207, 5, 5, 255, 226, 5, 5, 178, 175, 5, 5, 175, 212, 5, 5, 48, 168, 5, 5, 251, 202, 5, 5, 167, 176, 5, 5, 88, 231, 5, 5, 193, 213, 5, 5, 18, 217, 5, 5, 208, 195, 5, 5, 56, 176, 5, 5, 110, 195, 5, 5, 10, 229, 5, 5, 168, 227, 5, 5, 142, 172, 5, 5, 89, 242, 5, 5, 84, 239, 5, 5, 58, 249, 5, 5, 162, 203, 5, 5, 105, 170, 5, 5, 128, 205, 5, 5, 254, 234, 5, 5, 32, 172, 5, 5, 29, 245, 5, 5, 162, 201, 5, 5, 217, 215, 5, 5, 28, 207, 5, 5, 117, 236, 5, 5, 91, 204, 5, 5, 236, 192, 5, 5, 244, 170, 5, 5, 136, 187, 5, 5, 21, 188, 5, 5, 235, 178, 5, 5, 17, 224, 5, 5, 185, 205, 5, 5, 8, 221, 5, 5, 170, 220, 5, 5, 236, 178, 5, 5, 217, 208, 5, 5, 141, 235, 5, 5, 29, 207, 5, 5, 212, 171, 5, 5, 56, 179, 5, 5, 87, 177, 5, 5, 164, 245, 5, 5, 55, 221, 5, 5, 81, 235, 5, 5, 109, 229, 5, 5, 193, 189, 5, 5, 112, 168, 5, 5, 91, 208, 5, 5, 41, 205, 5, 5, 231, 176, 5, 5, 86, 247, 5, 5, 70, 172, 5, 5, 219, 213, 5, 5, 123, 194, 5, 5, 121, 208, 5, 5, 168, 176, 5, 5, 240, 210, 5, 5, 103, 191, 5, 5, 171, 197, 5, 5, 241, 239, 5, 5, 49, 194, 5, 5, 194, 210, 5, 5, 152, 222, 5, 5, 199, 239, 5, 5, 177, 187, 5, 5, 17, 174, 5, 5, 6, 190, 5, 5, 35, 220, 5, 5, 86, 199, 5, 5, 163, 215, 5, 5, 179, 206, 5, 5, 243, 243, 5, 5, 62, 244, 5, 5, 214, 215, 5, 5, 80, 180, 5, 5, 41, 203, 5, 5, 27, 204, 5, 5, 87, 199, 5, 5, 0, 160, 5, 5, 120, 170, 5, 5, 28, 202, 5, 5, 71, 229, 5, 5, 150, 192, 5, 5, 119, 215, 5, 5, 163, 221, 5, 5, 63, 182, 5, 5, 255, 207, 5, 5, 223, 211, 5, 5, 82, 173, 5, 5, 97, 179, 5, 5, 222, 235, 5, 5, 180, 194, 5, 5, 85, 219, 5, 5, 86, 219, 5, 5, 198, 201, 5, 5, 196, 241, 5, 5, 26, 216, 5, 5, 126, 175, 5, 5, 45, 190, 5, 5, 171, 193, 5, 5, 224, 206, 5, 5, 91, 243, 5, 5, 91, 169, 5, 5, 56, 241, 5, 5, 150, 217, 5, 5, 197, 205, 5, 5, 107, 218, 5, 5, 112, 191, 5, 5, 186, 181, 5, 5, 18, 226, 5, 5, 216, 182, 5, 5, 236, 197, 5, 5, 248, 249, 5, 5, 53, 182, 5, 5, 26, 219, 5, 5, 38, 183, 5, 5, 41, 200, 5, 5, 32, 223, 5, 5, 84, 219, 5, 5, 21, 236, 5, 5, 170, 234, 5, 5, 118, 184, 5, 5, 237, 197, 5, 5, 94, 178, 5, 5, 71, 230, 5, 5, 49, 180, 5, 5, 94, 223, 5, 5, 29, 233, 5, 5, 143, 231, 5, 5, 233, 204, 5, 5, 150, 229, 5, 5, 205, 220, 5, 5, 168, 214, 5, 5, 172, 193, 5, 5, 134, 243, 5, 5, 120, 224, 5, 5, 95, 204, 5, 5, 98, 242, 5, 5, 198, 231, 5, 5, 228, 229, 5, 5, 151, 192, 5, 5, 89, 189, 5, 5, 143, 223, 5, 5, 111, 211, 5, 5, 200, 201, 5, 5, 30, 235, 5, 5, 136, 243, 5, 5, 90, 242, 5, 5, 62, 231, 5, 5, 139, 188, 5, 5, 99, 169, 5, 5, 90, 188, 5, 5, 20, 168, 5, 5, 228, 228, 5, 5, 122, 191, 5, 5, 123, 191, 5, 5, 8, 193, 5, 5, 126, 176, 5, 5, 175, 232, 5, 5, 156, 228, 5, 5, 218, 208, 5, 5, 118, 239, 5, 5, 239, 201, 5, 5, 105, 194, 5, 5, 109, 172, 5, 5, 57, 221, 5, 5, 221, 215, 5, 5, 18, 203, 5, 5, 228, 198, 5, 5, 217, 242, 5, 5, 144, 177, 5, 5, 69, 226, 5, 5, 94, 225, 5, 5, 94, 235, 5, 5, 136, 170, 5, 5, 215, 244, 5, 5, 121, 224, 5, 5, 2, 232, 5, 5, 97, 182, 5, 5, 246, 179, 5, 5, 193, 233, 5, 5, 113, 186, 5, 5, 194, 233, 5, 5, 177, 222, 5, 5, 63, 194, 5, 5, 91, 188, 5, 5, 39, 189, 5, 5, 150, 219, 5, 5, 235, 233, 5, 5, 144, 235, 5, 5, 214, 237, 5, 5, 251, 168, 5, 5, 218, 223, 5, 5, 178, 222, 5, 5, 149, 235, 5, 5, 233, 180, 5, 5, 145, 213, 5, 5, 76, 229, 5, 5, 72, 211, 5, 5, 79, 175, 5, 5, 218, 206, 5, 5, 34, 219, 5, 5, 106, 213, 5, 5, 248, 224, 5, 5, 15, 193, 5, 5, 19, 244, 5, 5, 24, 174, 5, 5, 215, 245, 5, 5, 3, 182, 5, 5, 169, 210, 5, 5, 141, 240, 5, 5, 68, 185, 5, 5, 206, 176, 5, 5, 34, 233, 5, 5, 175, 214, 5, 5, 45, 182, 5, 5, 243, 193, 5, 5, 249, 234, 5, 5, 8, 200, 5, 5, 158, 192, 5, 5, 198, 170, 5, 5, 173, 179, 5, 5, 128, 208, 5, 5, 229, 200, 5, 5, 4, 196, 201, 224, 230, 0, 5, 5, 199, 208, 5, 5, 2, 217, 5, 5, 89, 216, 5, 5, 91, 232, 5, 5, 155, 203, 5, 5, 224, 211, 5, 5, 60, 232, 5, 5, 33, 224, 5, 5, 231, 205, 5, 5, 135, 239, 5, 5, 251, 234, 5, 5, 147, 233, 5, 5, 41, 202, 5, 5, 136, 239, 5, 5, 138, 200, 5, 5, 204, 201, 5, 5, 22, 195, 5, 5, 242, 178, 5, 5, 194, 246, 5, 5, 157, 169, 5, 5, 235, 225, 5, 5, 219, 237, 5, 5, 140, 209, 5, 5, 178, 190, 5, 5, 4, 171, 5, 5, 68, 182, 5, 5, 220, 223, 5, 5, 53, 225, 5, 5, 169, 227, 5, 5, 5, 217, 5, 5, 34, 216, 5, 5, 128, 229, 5, 5, 171, 204, 5, 5, 130, 191, 5, 5, 212, 211, 5, 5, 70, 185, 5, 5, 140, 243, 5, 5, 120, 169, 5, 5, 203, 202, 5, 5, 181, 222, 5, 5, 133, 175, 5, 5, 0, 163, 5, 5, 33, 204, 5, 5, 21, 244, 5, 5, 37, 235, 5, 5, 168, 216, 5, 5, 204, 200, 5, 5, 239, 191, 5, 5, 41, 228, 5, 5, 163, 230, 5, 5, 89, 219, 5, 5, 90, 219, 5, 5, 183, 214, 5, 5, 21, 189, 5, 5, 100, 205, 5, 5, 232, 223, 5, 5, 179, 227, 5, 5, 17, 208, 5, 5, 143, 243, 5, 5, 206, 213, 5, 5, 7, 242, 5, 5, 154, 235, 5, 5, 24, 193, 5, 5, 126, 196, 5, 5, 38, 235, 5, 5, 84, 209, 5, 5, 29, 217, 5, 5, 233, 223, 5, 5, 127, 196, 5, 5, 192, 212, 5, 5, 35, 211, 5, 5, 243, 171, 5, 5, 86, 218, 5, 5, 149, 239, 5, 5, 98, 232, 5, 5, 217, 241, 5, 5, 31, 197, 5, 5, 48, 203, 5, 5, 49, 203, 5, 5, 25, 240, 5, 5, 225, 215, 5, 5, 95, 240, 5, 5, 242, 204, 5, 5, 75, 227, 5, 5, 32, 190, 5, 5, 41, 243, 5, 5, 109, 204, 5, 5, 159, 217, 5, 5, 172, 237, 5, 5, 105, 182, 5, 5, 182, 207, 5, 5, 65, 188, 5, 5, 241, 199, 5, 5, 242, 200, 5, 5, 29, 193, 5, 5, 102, 178, 5, 5, 36, 207, 5, 5, 229, 201, 5, 5, 175, 201, 5, 5, 38, 191, 5, 5, 112, 183, 5, 5, 120, 246, 5, 5, 133, 229, 5, 5, 13, 200, 5, 5, 16, 245, 5, 5, 90, 190, 5, 5, 54, 202, 5, 5, 31, 193, 5, 5, 72, 208, 5, 5, 210, 201, 5, 5, 95, 190, 5, 5, 129, 206, 5, 5, 86, 192, 5, 5, 15, 200, 5, 5, 220, 205, 5, 5, 199, 193, 5, 5, 152, 219, 5, 5, 240, 225, 5, 5, 182, 201, 5, 5, 8, 218, 5, 5, 51, 233, 5, 5, 163, 239, 5, 5, 52, 206, 5, 5, 232, 207, 5, 5, 34, 171, 5, 5, 144, 189, 5, 5, 216, 189, 5, 5, 120, 205, 5, 5, 164, 211, 5, 5, 11, 206, 5, 5, 67, 223, 5, 5, 253, 232, 5, 5, 101, 215, 5, 5, 172, 175, 5, 5, 159, 189, 5, 5, 149, 234, 5, 5, 80, 177, 5, 5, 3, 184, 5, 5, 150, 234, 5, 5, 156, 234, 5, 5, 125, 205, 5, 5, 222, 189, 5, 5, 187, 241, 5, 5, 131, 188, 5, 5, 172, 194, 5, 5, 98, 197, 5, 5, 179, 187, 5, 5, 140, 169, 5, 5, 120, 236, 5, 5, 11, 245, 5, 5, 12, 245, 5, 5, 237, 199, 5, 5, 69, 171, 5, 5, 149, 231, 5, 5, 227, 222, 5, 5, 29, 175, 5, 5, 43, 217, 5, 5, 131, 179, 5, 5, 178, 192, 5, 5, 45, 217, 5, 5, 235, 179, 5, 5, 170, 221, 5, 5, 158, 231, 5, 5, 200, 197, 5, 5, 122, 245, 5, 5, 239, 220, 5, 5, 121, 241, 5, 5, 249, 187, 5, 5, 71, 249, 5, 5, 64, 189, 5, 5, 9, 191, 5, 5, 244, 207, 5, 5, 154, 185, 5, 5, 50, 224, 5, 5, 160, 231, 5, 5, 255, 172, 5, 5, 221, 221, 5, 5, 191, 238, 5, 5, 186, 206, 5, 5, 202, 232, 5, 5, 147, 182, 5, 5, 153, 181, 5, 5, 180, 168, 5, 5, 214, 174, 5, 5, 54, 215, 5, 5, 169, 229, 5, 5, 170, 229, 5, 5, 94, 219, 5, 5, 66, 169, 5, 5, 2, 213, 5, 5, 19, 243, 5, 5, 20, 243, 5, 5, 30, 198, 5, 5, 126, 214, 5, 5, 18, 227, 5, 5, 5, 240, 5, 5, 127, 214, 5, 5, 86, 172, 5, 5, 185, 242, 5, 5, 205, 197, 5, 5, 51, 191, 5, 5, 119, 216, 5, 5, 128, 214, 5, 5, 119, 247, 5, 5, 2, 236, 5, 5, 94, 188, 5, 5, 201, 209, 5, 5, 174, 228, 5, 5, 27, 183, 5, 5, 28, 183, 5, 5, 79, 188, 5, 5, 216, 185, 5, 5, 15, 244, 5, 5, 91, 184, 5, 5, 63, 218, 5, 5, 153, 195, 5, 5, 244, 183, 5, 5, 152, 174, 5, 5, 89, 180, 5, 5, 131, 169, 5, 5, 76, 232, 5, 5, 138, 210, 5, 5, 192, 177, 5, 5, 14, 217, 5, 5, 43, 241, 5, 5, 175, 244, 5, 5, 187, 213, 5, 5, 188, 174, 5, 5, 105, 241, 5, 5, 186, 188, 5, 5, 233, 187, 5, 5, 86, 197, 5, 5, 148, 202, 5, 5, 98, 184, 5, 5, 65, 218, 5, 5, 146, 244, 5, 5, 19, 186, 5, 5, 8, 213, 5, 5, 9, 213, 5, 5, 33, 232, 5, 5, 49, 248, 5, 5, 136, 179, 5, 5, 40, 171, 5, 5, 113, 171, 5, 5, 60, 186, 5, 5, 205, 244, 5, 5, 109, 217, 5, 5, 240, 220, 5, 5, 187, 226, 5, 5, 78, 204, 5, 5, 93, 186, 5, 5, 23, 178, 5, 5, 94, 186, 5, 5, 120, 240, 5, 5, 149, 202, 5, 5, 61, 174, 5, 5, 113, 217, 5, 5, 158, 209, 5, 5, 124, 185, 5, 5, 149, 176, 5, 5, 54, 191, 5, 5, 55, 191, 5, 5, 61, 207, 5, 5, 105, 184, 5, 5, 63, 237, 5, 5, 54, 228, 5, 5, 200, 203, 5, 5, 216, 247, 5, 5, 80, 238, 5, 5, 38, 232, 5, 5, 109, 200, 5, 5, 168, 194, 5, 5, 155, 202, 5, 5, 57, 239, 5, 5, 194, 213, 5, 5, 85, 214, 5, 5, 203, 187, 5, 5, 241, 220, 5, 5, 144, 238, 5, 5, 186, 228, 5, 5, 117, 209, 5, 5, 243, 216, 5, 5, 81, 216, 5, 5, 88, 187, 5, 5, 244, 216, 5, 5, 5, 247, 5, 5, 245, 206, 5, 5, 118, 201, 5, 5, 133, 245, 5, 5, 88, 193, 5, 5, 71, 210, 5, 5, 74, 223, 5, 5, 150, 214, 5, 5, 12, 234, 5, 5, 158, 195, 5, 5, 74, 217, 5, 5, 203, 220, 5, 5, 248, 240, 5, 5, 91, 193, 5, 5, 190, 227, 5, 5, 219, 173, 5, 5, 194, 169, 5, 5, 115, 181, 5, 5, 45, 188, 5, 5, 174, 222, 5, 5, 48, 235, 5, 5, 97, 183, 5, 5, 108, 221, 5, 5, 109, 221, 5, 5, 63, 184, 5, 5, 137, 232, 5, 5, 112, 245, 5, 5, 70, 230, 5, 5, 14, 184, 5, 5, 19, 202, 5, 5, 57, 201, 5, 5, 106, 170, 5, 5, 121, 175, 5, 5, 157, 242, 5, 5, 79, 241, 5, 5, 101, 195, 5, 5, 28, 178, 5, 5, 111, 236, 5, 5, 227, 226, 5, 5, 116, 179, 5, 5, 191, 227, 5, 5, 56, 191, 5, 5, 243, 229, 5, 5, 115, 246, 5, 5, 161, 223, 5, 5, 193, 182, 5, 5, 245, 196, 5, 5, 122, 208, 5, 5, 163, 214, 5, 5, 164, 214, 5, 5, 81, 242, 5, 5, 66, 198, 5, 5, 144, 200, 5, 5, 30, 208, 5, 5, 192, 215, 5, 5, 84, 223, 5, 5, 245, 230, 5, 5, 91, 206, 5, 5, 173, 202, 5, 5, 211, 179, 5, 5, 155, 226, 5, 5, 197, 225, 5, 5, 71, 200, 5, 5, 11, 188, 5, 5, 50, 188, 5, 5, 82, 242, 5, 5, 75, 169, 5, 5, 204, 182, 5, 5, 18, 213, 5, 5, 223, 228, 5, 5, 244, 180, 5, 5, 85, 239, 5, 5, 162, 242, 5, 5, 171, 224, 5, 5, 237, 217, 5, 5, 139, 215, 5, 5, 97, 219, 5, 5, 143, 210, 5, 5, 183, 216, 5, 5, 110, 229, 5, 5, 99, 203, 5, 5, 192, 187, 5, 5, 90, 228, 5, 5, 38, 180, 5, 5, 182, 181, 5, 5, 26, 169, 5, 5, 57, 179, 5, 5, 189, 228, 5, 5, 90, 172, 5, 5, 94, 237, 5, 5, 131, 240, 5, 5, 161, 187, 5, 5, 28, 173, 5, 5, 10, 180, 5, 5, 145, 200, 5, 5, 233, 198, 5, 5, 102, 195, 5, 5, 97, 233, 5, 5, 233, 226, 5, 5, 132, 207, 5, 5, 171, 234, 5, 5, 144, 208, 5, 5, 197, 241, 5, 5, 55, 195, 5, 5, 236, 221, 5, 5, 27, 235, 5, 5, 216, 190, 5, 5, 173, 236, 5, 5, 176, 170, 5, 5, 255, 218, 5, 5, 225, 228, 5, 5, 119, 184, 5, 5, 212, 242, 5, 5, 219, 207, 5, 5, 115, 229, 5, 5, 26, 184, 5, 5, 27, 219, 5, 5, 132, 233, 5, 5, 57, 241, 5, 5, 80, 217, 5, 5, 91, 207, 5, 5, 167, 231, 5, 5, 72, 240, 5, 5, 194, 221, 5, 5, 36, 175, 5, 5, 93, 225, 5, 5, 221, 247, 5, 5, 118, 249, 5, 5, 19, 226, 5, 5, 120, 184, 5, 5, 213, 242, 5, 5, 5, 200, 5, 5, 2, 208, 5, 5, 148, 200, 5, 5, 132, 180, 5, 5, 198, 189, 5, 5, 22, 217, 5, 5, 214, 179, 5, 5, 174, 202, 5, 5, 245, 229, 5, 5, 59, 243, 5, 5, 179, 217, 5, 5, 44, 210, 5, 5, 46, 205, 5, 5, 36, 176, 5, 5, 221, 218, 5, 5, 133, 240, 5, 5, 102, 194, 5, 5, 46, 169, 5, 5, 59, 201, 5, 5, 65, 212, 5, 5, 97, 223, 5, 5, 13, 231, 5, 5, 151, 176, 5, 5, 14, 231, 5, 5, 218, 242, 5, 5, 215, 229, 5, 5, 19, 203, 5, 5, 205, 223, 5, 5, 11, 169, 5, 5, 224, 218, 5, 5, 225, 218, 5, 5, 156, 192, 5, 5, 45, 230, 5, 5, 132, 234, 5, 5, 58, 188, 5, 5, 142, 216, 5, 5, 165, 238, 5, 5, 14, 170, 5, 5, 46, 190, 5, 5, 127, 225, 5, 5, 210, 238, 5, 5, 177, 180, 5, 5, 19, 210, 5, 5, 207, 188, 5, 5, 202, 189, 5, 5, 54, 187, 5, 5, 192, 206, 5, 5, 49, 209, 5, 5, 119, 239, 5, 5, 181, 229, 5, 5, 137, 240, 5, 5, 218, 225, 5, 5, 89, 247, 5, 5, 128, 172, 5, 5, 216, 219, 5, 5, 170, 214, 5, 5, 14, 207, 5, 5, 179, 226, 5, 5, 173, 220, 5, 5, 83, 222, 5, 5, 146, 172, 5, 5, 168, 174, 5, 5, 65, 179, 5, 5, 121, 168, 5, 5, 203, 205, 5, 5, 124, 229, 5, 5, 50, 247, 5, 5, 130, 225, 5, 5, 22, 174, 5, 5, 99, 213, 5, 5, 153, 217, 5, 5, 22, 213, 5, 5, 207, 241, 5, 5, 71, 194, 5, 5, 120, 206, 5, 5, 49, 205, 5, 5, 157, 216, 5, 5, 91, 247, 5, 5, 24, 240, 5, 5, 108, 194, 5, 5, 237, 222, 5, 5, 184, 229, 5, 5, 35, 233, 5, 5, 158, 228, 5, 5, 143, 245, 5, 5, 168, 245, 5, 5, 181, 206, 5, 5, 182, 206, 5, 5, 164, 187, 5, 5, 166, 204, 5, 5, 94, 207, 5, 5, 20, 231, 5, 5, 182, 176, 5, 5, 37, 202, 5, 5, 16, 207, 5, 5, 36, 233, 5, 5, 173, 173, 5, 5, 35, 245, 5, 5, 70, 207, 5, 5, 236, 232, 5, 5, 141, 209, 5, 5, 230, 249, 5, 5, 98, 223, 5, 5, 26, 217, 5, 5, 219, 225, 5, 5, 155, 246, 5, 5, 188, 212, 5, 5, 189, 212, 5, 5, 243, 194, 5, 5, 156, 217, 5, 5, 30, 171, 5, 5, 148, 203, 5, 5, 27, 212, 5, 5, 93, 187, 5, 5, 98, 231, 5, 5, 190, 192, 5, 5, 92, 248, 5, 5, 222, 190, 5, 5, 205, 182, 5, 5, 101, 201, 5, 5, 247, 197, 5, 5, 248, 197, 5, 5, 136, 191, 5, 5, 249, 239, 5, 5, 44, 173, 5, 5, 189, 194, 5, 5, 150, 220, 5, 5, 77, 210, 5, 5, 41, 233, 5, 5, 73, 230, 5, 5, 37, 175, 5, 5, 193, 235, 5, 5, 99, 223, 5, 5, 134, 175, 5, 5, 81, 174, 5, 5, 63, 178, 5, 5, 89, 202, 5, 5, 206, 233, 5, 5, 151, 220, 5, 5, 214, 238, 5, 5, 73, 196, 5, 5, 147, 177, 5, 5, 47, 173, 5, 5, 8, 215, 5, 5, 101, 176, 5, 5, 211, 232, 5, 5, 10, 221, 5, 5, 143, 206, 5, 5, 144, 216, 5, 5, 151, 240, 5, 5, 5, 173, 5, 5, 194, 202, 5, 5, 94, 244, 5, 5, 39, 235, 5, 5, 29, 203, 5, 5, 249, 247, 5, 5, 76, 244, 5, 5, 98, 233, 5, 5, 71, 207, 5, 5, 207, 233, 5, 5, 191, 214, 5, 5, 235, 218, 5, 5, 3, 194, 5, 5, 35, 208, 5, 5, 111, 188, 5, 5, 39, 219, 5, 5, 240, 191, 5, 5, 246, 248, 5, 5, 126, 195, 5, 5, 148, 213, 5, 5, 154, 188, 5, 5, 224, 208, 5, 5, 108, 182, 5, 5, 209, 201, 5, 5, 125, 195, 5, 5, 157, 200, 5, 5, 26, 209, 5, 5, 55, 202, 5, 5, 67, 176, 5, 5, 212, 202, 5, 5, 33, 217, 5, 5, 41, 211, 5, 5, 111, 205, 5, 5, 180, 180, 5, 5, 48, 233, 5, 5, 234, 247, 5, 5, 3, 205, 5, 5, 56, 202, 5, 5, 4, 205, 5, 5, 114, 183, 5, 5, 232, 237, 5, 5, 28, 213, 5, 5, 52, 210, 5, 5, 122, 186, 5, 5, 141, 197, 5, 5, 104, 240, 5, 5, 186, 222, 5, 5, 42, 231, 5, 5, 92, 172, 5, 5, 168, 217, 5, 5, 252, 217, 5, 5, 181, 180, 5, 5, 93, 172, 5, 5, 8, 207, 5, 5, 171, 217, 5, 5, 127, 195, 5, 5, 3, 248, 5, 5, 199, 248, 5, 5, 124, 235, 5, 5, 185, 190, 5, 5, 153, 211, 5, 5, 108, 212, 5, 5, 156, 211, 5, 5, 250, 233, 5, 5, 125, 247, 5, 5, 235, 206, 5, 5, 14, 182, 5, 5, 82, 198, 5, 5, 225, 208, 5, 5, 157, 231, 5, 5, 24, 234, 5, 5, 149, 236, 5, 5, 153, 198, 5, 5, 241, 242, 5, 5, 142, 217, 5, 5, 79, 204, 5, 5, 74, 235, 5, 5, 208, 222, 5, 5, 34, 232, 5, 5, 54, 227, 5, 5, 227, 232, 5, 5, 50, 195, 5, 5, 43, 232, 5, 5, 10, 236, 5, 5, 199, 189, 5, 5, 116, 229, 5, 5, 149, 178, 5, 5, 25, 175, 5, 5, 120, 229, 5, 5, 151, 178, 5, 5, 179, 248, 5, 5, 18, 189, 5, 5, 169, 217, 5, 5, 63, 238, 5, 5, 61, 238, 5, 5, 3, 172, 5, 5, 221, 184, 5, 5, 141, 241, 5, 5, 19, 171, 5, 5, 229, 172, 5, 5, 114, 238, 5, 5, 83, 220, 5, 5, 38, 174, 5, 5, 174, 183, 5, 5, 127, 187, 5, 5, 4, 183, 5, 5, 232, 222, 5, 5, 74, 195, 5, 5, 237, 209, 5, 5, 79, 247, 5, 5, 250, 177, 5, 5, 16, 239, 5, 5, 117, 247, 5, 5, 82, 247, 5, 5, 149, 195, 5, 5, 184, 246, 5, 5, 115, 233, 5, 5, 61, 218, 5, 5, 99, 246, 5, 5, 165, 218, 5, 5, 98, 216, 5, 5, 176, 195, 5, 5, 26, 198, 5, 5, 186, 242, 5, 5, 207, 207, 5, 5, 2, 173, 5, 5, 134, 168, 5, 5, 236, 210, 5, 5, 10, 191, 5, 5, 62, 218, 5, 5, 59, 193, 5, 5, 3, 228, 5, 5, 120, 204, 5, 5, 193, 177, 5, 5, 132, 169, 5, 5, 86, 212, 5, 5, 204, 237, 5, 5, 44, 250, 5, 5, 72, 170, 5, 5, 223, 220, 5, 5, 62, 225, 5, 5, 87, 197, 5, 5, 221, 188, 5, 5, 106, 235, 5, 5, 123, 234, 5, 5, 231, 245, 5, 5, 135, 238, 5, 5, 243, 211, 5, 5, 151, 171, 5, 5, 204, 178, 5, 5, 67, 228, 5, 5, 92, 184, 5, 5, 130, 187, 5, 5, 231, 246, 5, 5, 248, 246, 5, 5, 64, 218, 5, 5, 42, 212, 5, 5, 28, 239, 5, 5, 117, 207, 5, 5, 236, 227, 5, 5, 240, 209, 5, 5, 252, 185, 5, 5, 70, 235, 5, 5, 126, 245, 5, 5, 106, 217, 5, 5, 177, 169, 5, 5, 40, 242, 5, 5, 50, 231, 5, 5, 50, 208, 5, 5, 189, 171, 5, 5, 158, 171, 5, 5, 4, 184, 5, 5, 218, 174, 5, 5, 215, 174, 5, 5, 114, 199, 5, 5, 170, 218, 5, 5, 138, 194, 5, 5, 85, 193, 5, 5, 31, 194, 5, 5, 195, 171, 5, 5, 89, 195, 5, 5, 231, 181, 5, 5, 97, 185, 5, 5, 20, 219, 5, 5, 213, 247, 5, 5, 40, 187, 5, 5, 185, 239, 5, 5, 53, 172, 5, 5, 201, 203, 5, 5, 88, 198, 5, 5, 62, 234, 5, 5, 129, 248, 5, 5, 53, 178, 5, 5, 157, 234, 5, 5, 61, 200, 5, 5, 34, 199, 5, 5, 239, 210, 5, 5, 211, 222, 5, 5, 89, 193, 5, 5, 91, 200, 5, 5, 120, 174, 5, 5, 81, 202, 5, 5, 134, 197, 5, 5, 50, 221, 5, 5, 252, 213, 5, 5, 125, 185, 5, 5, 74, 239, 5, 5, 58, 242, 5, 5, 235, 245, 5, 5, 142, 204, 5, 5, 75, 217, 5, 5, 54, 218, 5, 5, 241, 192, 5, 5, 75, 239, 5, 5, 7, 172, 5, 5, 130, 248, 5, 5, 23, 223, 5, 5, 36, 220, 5, 5, 57, 218, 5, 5, 78, 202, 5, 5, 244, 202, 5, 5, 38, 203, 5, 5, 117, 199, 5, 5, 225, 193, 5, 5, 153, 183, 5, 5, 12, 173, 5, 5, 107, 170, 5, 5, 65, 200, 5, 5, 195, 225, 5, 5, 248, 242, 5, 5, 85, 204, 5, 5, 219, 174, 5, 5, 117, 173, 5, 5, 68, 175, 5, 5, 36, 181, 5, 5, 21, 171, 5, 5, 97, 203, 5, 5, 192, 220, 5, 5, 116, 212, 5, 5, 117, 212, 5, 5, 191, 182, 5, 5, 158, 242, 5, 5, 140, 206, 5, 5, 193, 187, 5, 5, 169, 236, 5, 5, 130, 180, 5, 5, 64, 226, 5, 5, 23, 247, 5, 5, 242, 196, 5, 5, 205, 238, 5, 5, 154, 214, 5, 5, 188, 187, 5, 5, 174, 186, 5, 5, 35, 199, 5, 5, 86, 214, 5, 5, 65, 226, 5, 5, 14, 226, 5, 5, 114, 184, 5, 5, 34, 175, 5, 5, 182, 233, 5, 5, 195, 170, 5, 5, 212, 179, 5, 5, 72, 200, 5, 5, 187, 180, 5, 5, 129, 234, 5, 5, 130, 234, 5, 5, 229, 188, 5, 5, 249, 242, 5, 5, 92, 169, 5, 5, 114, 169, 5, 5, 115, 184, 5, 5, 193, 241, 5, 5, 166, 227, 5, 5, 171, 236, 5, 5, 28, 191, 5, 5, 196, 169, 5, 5, 148, 175, 5, 5, 253, 205, 5, 5, 90, 212, 5, 5, 232, 177, 5, 5, 108, 243, 5, 5, 184, 224, 5, 5, 4, 186, 5, 5, 116, 177, 5, 5, 82, 190, 5, 5, 222, 218, 5, 5, 250, 242, 5, 5, 60, 219, 5, 5, 18, 210, 5, 5, 147, 196, 5, 5, 162, 224, 5, 5, 12, 169, 5, 5, 192, 227, 5, 5, 221, 174, 5, 5, 204, 219, 5, 5, 243, 210, 5, 5, 77, 240, 5, 5, 162, 195, 5, 5, 77, 215, 5, 5, 85, 190, 5, 5, 121, 199, 5, 5, 11, 203, 5, 5, 222, 200, 5, 5, 29, 202, 5, 5, 157, 245, 5, 5, 196, 221, 5, 5, 3, 206, 5, 5, 120, 239, 5, 5, 178, 179, 5, 5, 198, 233, 5, 5, 5, 232, 5, 5, 125, 229, 5, 5, 23, 171, 5, 5, 144, 172, 5, 5, 60, 192, 5, 5, 232, 215, 5, 5, 13, 220, 5, 5, 93, 169, 5, 5, 179, 232, 5, 5, 127, 170, 5, 5, 34, 184, 5, 5, 244, 193, 5, 5, 112, 248, 5, 5, 245, 193, 5, 5, 54, 177, 5, 5, 159, 192, 5, 5, 241, 177, 5, 5, 168, 243, 5, 5, 99, 182, 5, 5, 220, 171, 5, 5, 234, 232, 5, 5, 240, 233, 5, 5, 42, 171, 5, 5, 109, 218, 5, 5, 15, 207, 5, 5, 230, 200, 5, 5, 138, 168, 5, 5, 73, 244, 5, 5, 121, 187, 5, 5, 155, 172, 5, 5, 101, 224, 5, 5, 117, 211, 5, 5, 16, 173, 5, 5, 46, 203, 5, 5, 151, 170, 5, 5, 190, 195, 5, 5, 60, 178, 5, 5, 182, 222, 5, 5, 73, 225, 5, 5, 152, 170, 5, 5, 234, 200, 5, 5, 37, 184, 5, 5, 38, 219, 5, 5, 195, 246, 5, 5, 238, 180, 5, 5, 184, 222, 5, 5, 72, 228, 5, 5, 86, 222, 5, 5, 242, 168, 5, 5, 139, 233, 5, 5, 233, 171, 5, 5, 50, 174, 5, 5, 207, 200, 5, 5, 253, 204, 5, 5, 72, 231, 5, 5, 14, 232, 5, 5, 246, 200, 5, 5, 158, 245, 5, 5, 210, 177, 5, 5, 56, 197, 5, 5, 234, 243, 5, 5, 203, 221, 5, 5, 9, 194, 5, 5, 48, 212, 5, 5, 160, 239, 5, 5, 253, 200, 5, 5, 26, 236, 5, 5, 175, 230, 5, 5, 176, 230, 5, 5, 202, 237, 5, 5, 131, 183, 5, 5, 33, 225, 5, 5, 106, 184, 5, 5, 148, 183, 5, 5, 131, 184, 5, 5, 219, 168, 5, 5, 213, 188, 5, 5, 88, 192, 5, 5, 93, 192, 5, 5, 13, 194, 5, 5, 191, 186, 5, 5, 213, 170, 5, 5, 212, 236, 5, 5, 41, 187, 5, 5, 211, 197, 5, 5, 84, 213, 5, 5, 118, 207, 5, 5, 109, 208, 5, 5, 110, 208, 5, 5, 214, 208, 5, 5, 46, 222, 5, 5, 134, 223, 5, 5, 11, 173, 5, 5, 115, 206, 5, 5, 222, 197, 5, 5, 115, 208, 5, 5, 176, 226, 5, 5, 2, 203, 5, 5, 210, 237, 5, 5, 24, 247, 5, 5, 38, 198, 5, 5, 59, 231, 5, 5, 119, 220, 5, 5, 226, 228, 5, 5, 144, 231, 5, 5, 126, 226, 5, 5, 203, 241, 5, 5, 12, 201, 5, 5, 193, 181, 5, 5, 163, 180, 5, 5, 106, 216, 5, 5, 71, 212, 5, 5, 16, 193, 5, 5, 50, 209, 5, 5, 142, 240, 5, 5, 42, 186, 5, 5, 128, 217, 5, 5, 13, 245, 5, 5, 7, 196, 5, 5, 199, 186, 5, 5, 183, 178, 5, 5, 77, 194, 5, 5, 122, 206, 5, 5, 194, 217, 5, 5, 189, 193, 5, 5, 21, 229, 5, 5, 4, 198, 5, 5, 132, 217, 5, 5, 125, 206, 5, 5, 17, 201, 5, 5, 91, 221, 5, 5, 248, 178, 5, 5, 204, 186, 5, 5, 11, 194, 5, 5, 188, 186, 5, 5, 200, 236, 5, 5, 28, 187, 5, 5, 102, 208, 5, 5, 23, 222, 5, 5, 8, 173, 5, 5, 2, 201, 5, 5, 201, 197, 5, 5, 221, 192, 5, 5, 51, 231, 5, 5, 227, 178, 5, 5, 121, 226, 5, 5, 188, 241, 5, 5, 40, 186, 5, 5, 249, 195, 5, 5, 189, 217, 5, 5, 196, 194, 5, 5, 47, 217, 5, 5, 177, 195, 5, 5, 219, 176, 5, 5, 212, 197, 5, 5, 249, 246, 5, 5, 210, 173, 5, 5, 122, 192, 5, 5, 61, 186, 5, 5, 29, 178, 5, 5, 29, 249, 5, 5, 11, 179, 5, 5, 164, 220, 5, 5, 64, 190, 5, 5, 233, 217, 5, 5, 130, 185, 5, 5, 63, 222, 5, 5, 131, 195, 5, 5, 89, 187, 5, 5, 223, 185, 5, 5, 179, 175, 5, 5, 132, 195, 5, 5, 184, 191, 5, 5, 77, 217, 5, 5, 13, 234, 5, 5, 18, 224, 5, 5, 144, 210, 5, 5, 50, 192, 5, 5, 190, 205, 5, 5, 33, 247, 5, 5, 138, 244, 5, 5, 137, 170, 5, 5, 19, 234, 5, 5, 205, 189, 5, 5, 174, 220, 5, 5, 229, 185, 5, 5, 55, 247, 5, 5, 114, 235, 5, 5, 190, 175, 5, 5, 30, 231, 5, 5, 177, 238, 5, 5, 221, 202, 5, 5, 9, 198, 5, 5, 40, 231, 5, 5, 9, 237, 5, 5, 44, 231, 5, 5, 87, 236, 5, 5, 88, 236, 5, 5, 253, 179, 5, 5, 88, 184, 5, 5, 48, 217, 5, 5, 49, 217, 5, 5, 255, 194, 5, 5, 31, 189, 5, 5, 217, 192, 5, 5, 124, 182, 5, 5, 210, 235, 5, 5, 157, 179, 5, 5, 98, 189, 5, 5, 3, 173, 5, 5, 209, 225, 5, 5, 147, 234, 5, 5, 83, 195, 5, 5, 136, 238, 5, 5, 167, 218, 5, 5, 211, 235, 5, 5, 226, 239, 5, 5, 125, 220, 5, 5, 241, 214, 5, 5, 237, 227, 5, 5, 222, 192, 5, 5, 212, 235, 5, 5, 227, 239, 5, 5, 109, 181, 5, 5, 10, 183, 5, 5, 40, 236, 5, 5, 118, 237, 5, 5, 183, 223, 5, 5, 82, 221, 5, 5, 233, 239, 5, 5, 226, 233, 5, 5, 213, 197, 5, 5, 56, 233, 5, 5, 106, 210, 5, 5, 10, 174, 5, 5, 44, 241, 5, 5, 189, 246, 5, 5, 76, 234, 5, 5, 171, 182, 5, 5, 251, 233, 5, 5, 179, 173, 5, 5, 253, 220, 5, 5, 104, 236, 5, 5, 134, 219, 5, 5, 185, 248, 5, 5, 190, 234, 5, 5, 43, 239, 5, 5, 44, 239, 5, 5, 255, 223, 5, 5, 159, 174, 5, 5, 152, 188, 5, 5, 51, 221, 5, 5, 190, 188, 5, 5, 231, 234, 5, 5, 236, 245, 5, 5, 52, 248, 5, 5, 73, 246, 5, 5, 31, 186, 5, 5, 30, 249, 5, 5, 78, 249, 5, 5, 208, 244, 5, 5, 98, 186, 5, 5, 107, 184, 5, 5, 226, 193, 5, 5, 202, 179, 5, 5, 81, 204, 5, 5, 12, 179, 5, 5, 105, 237, 5, 5, 156, 202, 5, 5, 72, 210, 5, 5, 43, 212, 5, 5, 187, 247, 5, 5, 245, 184, 5, 5, 58, 239, 5, 5, 92, 197, 5, 5, 208, 237, 5, 5, 177, 244, 5, 5, 145, 238, 5, 5, 146, 238, 5, 5, 177, 217, 5, 5, 99, 245, 5, 5, 195, 213, 5, 5, 91, 170, 5, 5, 84, 234, 5, 5, 114, 217, 5, 5, 197, 168, 5, 5, 139, 177, 5, 5, 212, 249, 5, 5, 164, 225, 5, 5, 14, 248, 5, 5, 112, 176, 5, 5, 134, 245, 5, 5, 243, 240, 5, 5, 192, 234, 5, 5, 182, 235, 5, 5, 76, 239, 5, 5, 242, 190, 5, 5, 195, 188, 5, 5, 64, 222, 5, 5, 225, 172, 5, 5, 82, 233, 5, 5, 189, 221, 5, 5, 255, 188, 5, 5, 241, 172, 5, 5, 35, 186, 5, 5, 90, 187, 5, 5, 234, 217, 5, 5, 97, 191, 5, 5, 103, 195, 5, 5, 65, 190, 5, 5, 161, 184, 5, 5, 197, 232, 5, 5, 222, 228, 5, 5, 53, 221, 5, 5, 189, 247, 5, 5, 25, 227, 5, 5, 54, 208, 5, 5, 210, 244, 5, 5, 58, 209, 5, 5, 180, 181, 5, 5, 25, 191, 5, 5, 115, 236, 5, 5, 77, 234, 5, 5, 175, 186, 5, 5, 16, 182, 5, 5, 137, 171, 5, 5, 169, 226, 5, 5, 152, 238, 5, 5, 195, 201, 5, 5, 217, 247, 5, 5, 184, 199, 5, 5, 126, 199, 5, 5, 41, 230, 5, 5, 71, 242, 5, 5, 227, 225, 5, 5, 250, 192, 5, 5, 25, 247, 5, 5, 178, 218, 5, 5, 232, 176, 5, 5, 70, 201, 5, 5, 117, 181, 5, 5, 193, 199, 5, 5, 211, 181, 5, 5, 70, 222, 5, 5, 248, 226, 5, 5, 31, 178, 5, 5, 197, 169, 5, 5, 29, 173, 5, 5, 110, 241, 5, 5, 48, 199, 5, 5, 51, 216, 5, 5, 104, 216, 5, 5, 231, 222, 5, 5, 49, 168, 5, 5, 250, 241, 5, 5, 85, 233, 5, 5, 124, 174, 5, 5, 161, 195, 5, 5, 141, 232, 5, 5, 43, 230, 5, 5, 155, 230, 5, 5, 82, 185, 5, 5, 185, 223, 5, 5, 252, 171, 5, 5, 104, 191, 5, 5, 106, 196, 5, 5, 39, 223, 5, 5, 10, 246, 5, 5, 40, 223, 5, 5, 131, 180, 5, 5, 29, 190, 5, 5, 171, 173, 5, 5, 14, 223, 5, 5, 114, 195, 5, 5, 18, 199, 5, 5, 120, 217, 5, 5, 202, 176, 5, 5, 59, 233, 5, 5, 92, 224, 5, 5, 190, 228, 5, 5, 52, 232, 5, 5, 217, 182, 5, 5, 167, 174, 5, 5, 209, 224, 5, 5, 96, 239, 5, 5, 190, 210, 5, 5, 246, 239, 5, 5, 169, 179, 5, 5, 38, 213, 5, 5, 231, 248, 5, 5, 83, 173, 5, 5, 245, 173, 5, 5, 178, 248, 5, 5, 2, 193, 5, 5, 95, 214, 5, 5, 42, 225, 5, 5, 165, 248, 5, 5, 72, 229, 5, 5, 246, 196, 5, 5, 210, 216, 5, 5, 50, 180, 5, 5, 75, 246, 5, 5, 65, 244, 5, 5, 173, 249, 5, 5, 75, 215, 5, 5, 232, 248, 5, 5, 112, 203, 5, 5, 56, 194, 5, 5, 184, 175, 5, 5, 105, 188, 5, 5, 103, 206, 5, 5, 84, 221, 5, 5, 22, 171, 5, 5, 72, 190, 5, 5, 79, 213, 5, 5, 198, 241, 5, 5, 216, 179, 5, 5, 172, 234, 5, 5, 86, 213, 5, 5, 80, 222, 5, 5, 31, 235, 5, 5, 81, 222, 5, 5, 185, 191, 5, 5, 73, 190, 5, 5, 194, 181, 5, 5, 126, 247, 5, 5, 63, 179, 5, 5, 135, 233, 5, 5, 29, 184, 5, 5, 46, 214, 5, 5, 234, 226, 5, 5, 64, 194, 5, 5, 220, 214, 5, 5, 99, 242, 5, 5, 43, 249, 5, 5, 91, 248, 5, 5, 94, 231, 5, 5, 124, 191, 5, 5, 205, 239, 5, 5, 53, 168, 5, 5, 54, 232, 5, 5, 36, 210, 5, 5, 23, 174, 5, 5, 133, 183, 5, 5, 224, 247, 5, 5, 70, 237, 5, 5, 243, 236, 5, 5, 220, 190, 5, 5, 32, 235, 5, 5, 141, 185, 5, 5, 184, 211, 5, 5, 96, 214, 5, 5, 162, 209, 5, 5, 39, 238, 5, 5, 121, 229, 5, 5, 16, 234, 5, 5, 233, 225, 5, 5, 155, 247, 5, 5, 147, 220, 5, 5, 246, 193, 5, 5, 222, 213, 5, 5, 43, 200, 5, 5, 245, 190, 5, 5, 231, 191, 5, 5, 108, 185, 5, 5, 143, 240, 5, 5, 55, 208, 5, 5, 68, 233, 5, 5, 129, 208, 5, 5, 97, 231, 5, 5, 203, 215, 5, 5, 248, 173, 5, 5, 131, 235, 5, 5, 25, 226, 5, 5, 34, 224, 5, 5, 79, 169, 5, 5, 202, 174, 5, 5, 78, 215, 5, 5, 84, 222, 5, 5, 128, 194, 5, 5, 219, 242, 5, 5, 199, 233, 5, 5, 236, 188, 5, 5, 177, 225, 5, 5, 167, 237, 5, 5, 168, 237, 5, 5, 73, 205, 5, 5, 210, 241, 5, 5, 5, 171, 5, 5, 212, 176, 5, 5, 224, 216, 5, 5, 38, 202, 5, 5, 77, 209, 5, 5, 78, 209, 5, 5, 175, 220, 5, 5, 145, 245, 5, 5, 61, 209, 5, 5, 252, 193, 5, 5, 74, 244, 5, 5, 57, 195, 5, 5, 24, 203, 5, 5, 50, 205, 5, 5, 112, 172, 5, 5, 213, 211, 5, 5, 146, 187, 5, 5, 65, 231, 5, 5, 241, 248, 5, 5, 123, 168, 5, 5, 124, 168, 5, 5, 214, 195, 5, 5, 146, 245, 5, 5, 174, 238, 5, 5, 169, 189, 5, 5, 152, 194, 5, 5, 255, 206, 5, 5, 199, 173, 5, 5, 62, 188, 5, 5, 48, 190, 5, 5, 90, 173, 5, 5, 173, 234, 5, 5, 114, 244, 5, 5, 8, 220, 5, 5, 31, 231, 5, 5, 140, 244, 5, 5, 237, 180, 5, 5, 78, 246, 5, 5, 79, 210, 5, 5, 235, 200, 5, 5, 124, 181, 5, 5, 96, 240, 5, 5, 251, 197, 5, 5, 74, 192, 5, 5, 18, 250, 5, 5, 245, 194, 5, 5, 242, 171, 5, 5, 140, 191, 5, 5, 114, 248, 5, 5, 76, 230, 5, 5, 99, 244, 5, 5, 154, 244, 5, 5, 240, 221, 5, 5, 61, 216, 5, 5, 55, 225, 5, 5, 100, 244, 5, 5, 51, 214, 5, 5, 145, 221, 5, 5, 40, 235, 5, 5, 37, 244, 5, 5, 58, 225, 5, 5, 69, 178, 5, 5, 106, 224, 5, 5, 99, 232, 5, 5, 77, 192, 5, 5, 26, 195, 5, 5, 74, 196, 5, 5, 15, 245, 5, 5, 50, 210, 5, 5, 96, 238, 5, 5, 0, 164, 5, 5, 223, 244, 5, 5, 59, 213, 5, 5, 64, 191, 5, 5, 77, 190, 5, 5, 150, 239, 5, 5, 151, 239, 5, 5, 152, 220, 5, 5, 100, 218, 5, 5, 134, 211, 5, 5, 143, 215, 5, 5, 240, 180, 5, 5, 221, 224, 5, 5, 14, 190, 5, 5, 158, 247, 5, 5, 112, 188, 5, 5, 30, 168, 5, 5, 100, 240, 5, 5, 80, 194, 5, 5, 120, 242, 5, 5, 8, 194, 5, 5, 147, 191, 5, 5, 147, 180, 5, 5, 149, 245, 5, 5, 88, 235, 5, 5, 231, 243, 5, 5, 211, 201, 5, 5, 92, 221, 5, 5, 135, 229, 5, 5, 96, 173, 5, 5, 225, 202, 5, 5, 178, 238, 5, 5, 227, 170, 5, 5, 150, 245, 5, 5, 10, 237, 5, 5, 205, 181, 5, 5, 86, 175, 5, 5, 137, 229, 5, 5, 87, 175, 5, 5, 224, 237, 5, 5, 59, 173, 5, 5, 101, 218, 5, 5, 28, 240, 5, 5, 251, 200, 5, 5, 51, 174, 5, 5, 142, 233, 5, 5, 44, 211, 5, 5, 106, 190, 5, 5, 235, 243, 5, 5, 165, 239, 5, 5, 74, 178, 5, 5, 230, 244, 5, 5, 20, 237, 5, 5, 156, 180, 5, 5, 81, 236, 5, 5, 195, 192, 5, 5, 248, 179, 5, 5, 75, 184, 5, 5, 145, 218, 5, 5, 250, 191, 5, 5, 64, 195, 5, 5, 97, 189, 5, 5, 208, 225, 5, 5, 98, 218, 5, 5, 120, 220, 5, 5, 223, 214, 5, 5, 222, 227, 5, 5, 198, 235, 5, 5, 227, 238, 5, 5, 199, 235, 5, 5, 196, 192, 5, 5, 148, 218, 5, 5, 116, 194, 5, 5, 70, 191, 5, 5, 197, 211, 5, 5, 61, 197, 5, 5, 27, 236, 5, 5, 103, 210, 5, 5, 182, 234, 5, 5, 98, 181, 5, 5, 101, 206, 5, 5, 68, 234, 5, 5, 178, 223, 5, 5, 130, 183, 5, 5, 245, 220, 5, 5, 160, 182, 5, 5, 187, 197, 5, 5, 67, 246, 5, 5, 79, 186, 5, 5, 147, 188, 5, 5, 167, 213, 5, 5, 203, 249, 5, 5, 211, 221, 5, 5, 82, 234, 5, 5, 168, 244, 5, 5, 239, 223, 5, 5, 221, 245, 5, 5, 252, 178, 5, 5, 129, 247, 5, 5, 103, 176, 5, 5, 95, 217, 5, 5, 89, 245, 5, 5, 51, 170, 5, 5, 243, 238, 5, 5, 108, 238, 5, 5, 174, 199, 5, 5, 189, 201, 5, 5, 24, 222, 5, 5, 167, 186, 5, 5, 171, 221, 5, 5, 109, 192, 5, 5, 3, 191, 5, 5, 92, 174, 5, 5, 203, 247, 5, 5, 30, 221, 5, 5, 57, 190, 5, 5, 14, 178, 5, 5, 23, 186, 5, 5, 210, 217, 5, 5, 74, 187, 5, 5, 166, 235, 5, 5, 4, 239, 5, 5, 25, 246, 5, 5, 149, 184, 5, 5, 191, 232, 5, 5, 237, 172, 5, 5, 175, 191, 5, 5, 185, 234, 5, 5, 134, 247, 5, 5, 150, 195, 5, 5, 34, 230, 5, 5, 232, 241, 5, 5, 44, 216, 5, 5, 132, 230, 5, 5, 74, 185, 5, 5, 102, 241, 5, 5, 77, 191, 5, 5, 190, 199, 5, 5, 37, 223, 5, 5, 180, 223, 5, 5, 210, 181, 5, 5, 205, 216, 5, 5, 209, 247, 5, 5, 164, 249, 5, 5, 174, 211, 5, 5, 124, 180, 5, 5, 213, 248, 5, 5, 209, 182, 5, 5, 7, 199, 5, 5, 41, 229, 5, 5, 155, 241, 5, 5, 18, 221, 5, 5, 78, 221, 5, 5, 158, 179, 5, 5, 69, 173, 5, 5, 105, 203, 5, 5, 177, 248, 5, 5, 79, 224, 5, 5, 29, 225, 5, 5, 79, 221, 5, 5, 29, 239, 5, 5, 157, 209, 5, 5, 11, 174, 5, 5, 192, 179, 5, 5, 240, 190, 5, 5, 32, 194, 5, 5, 118, 233, 5, 5, 129, 235, 5, 5, 15, 238, 5, 5, 100, 229, 5, 5, 160, 181, 5, 5, 41, 242, 5, 5, 8, 235, 5, 5, 19, 173, 5, 5, 15, 249, 5, 5, 41, 168, 5, 5, 216, 236, 5, 5, 39, 179, 5, 5, 52, 208, 5, 5, 73, 213, 5, 5, 151, 234, 5, 5, 65, 209, 5, 5, 67, 178, 5, 5, 2, 224, 5, 5, 177, 233, 5, 5, 140, 237, 5, 5, 70, 169, 5, 5, 54, 222, 5, 5, 62, 215, 5, 5, 112, 208, 5, 5, 201, 195, 5, 5, 247, 206, 5, 5, 135, 245, 5, 5, 234, 193, 5, 5, 60, 209, 5, 5, 37, 225, 5, 5, 98, 244, 5, 5, 22, 216, 5, 5, 219, 200, 5, 5, 45, 214, 5, 5, 228, 197, 5, 5, 238, 236, 5, 5, 140, 215, 5, 5, 212, 244, 5, 5, 48, 174, 5, 5, 80, 186, 5, 5, 55, 215, 5, 5, 71, 189, 5, 5, 161, 231, 5, 5, 127, 192, 5, 5, 82, 189, 5, 5, 237, 187, 5, 5, 38, 189, 5, 5, 21, 231, 5, 5, 22, 231, 5, 5, 193, 191, 5, 5, 149, 203, 5, 5, 23, 188, 5, 5, 148, 180, 5, 5, 5, 205, 5, 5, 92, 180, 5, 5, 94, 194, 5, 5, 209, 214, 5, 5, 47, 222, 5, 5, 82, 202, 5, 5, 175, 178, 5, 5, 92, 228, 5, 5, 13, 170, 5, 5, 221, 222, 5, 5, 151, 232, 5, 5, 106, 183, 5, 5, 59, 247, 5, 5, 65, 247, 5, 5, 21, 237, 5, 5, 22, 237, 5, 5, 246, 221, 5, 5, 162, 175, 5, 5, 14, 191, 5, 5, 209, 227, 5, 5, 45, 239, 5, 5, 210, 227, 5, 5, 46, 239, 5, 5, 149, 193, 5, 5, 183, 168, 5, 5, 136, 189, 5, 5, 170, 181, 5, 5, 116, 175, 5, 5, 223, 232, 5, 5, 171, 190, 5, 5, 165, 193, 5, 5, 27, 199, 5, 5, 162, 184, 5, 5, 99, 197, 5, 5, 191, 183, 5, 5, 246, 230, 5, 5, 61, 171, 5, 5, 102, 188, 5, 5, 91, 242, 5, 5, 225, 247, 5, 5, 71, 193, 5, 5, 43, 183, 5, 5, 23, 231, 5, 5, 221, 171, 5, 5, 151, 229, 5, 5, 133, 190, 5, 5, 70, 171, 5, 5, 243, 178, 5, 5, 123, 249, 5, 5, 47, 183, 5, 5, 142, 239, 5, 5, 202, 246, 5, 5, 127, 169, 5, 5, 249, 172, 5, 5, 79, 168, 5, 5, 10, 213, 5, 5, 243, 209, 5, 5, 226, 212, 5, 5, 20, 186, 5, 5, 245, 209, 5, 5, 107, 241, 5, 5, 138, 179, 5, 5, 118, 209, 5, 5, 137, 223, 5, 5, 99, 234, 5, 5, 123, 190, 5, 5, 66, 200, 5, 5, 196, 188, 5, 5, 99, 188, 5, 5, 125, 209, 5, 5, 84, 168, 5, 5, 130, 207, 5, 5, 26, 202, 5, 5, 145, 210, 5, 5, 43, 170, 5, 5, 4, 242, 5, 5, 72, 193, 5, 5, 147, 227, 5, 5, 73, 207, 5, 5, 23, 213, 5, 5, 24, 231, 5, 5, 131, 239, 5, 5, 120, 197, 5, 5, 139, 209, 5, 5, 108, 175, 5, 5, 56, 225, 5, 5, 107, 190, 5, 5, 18, 198, 5, 5, 164, 169, 5, 5, 179, 245, 5, 5, 180, 242, 5, 5, 89, 184, 5, 5, 52, 172, 5, 5, 245, 185, 5, 5, 214, 225, 5, 5, 137, 238, 5, 5, 81, 188, 5, 5, 111, 228, 5, 5, 139, 213, 5, 5, 219, 191, 5, 5, 130, 182, 5, 5, 18, 225, 5, 5, 225, 186, 5, 5, 55, 244, 5, 5, 250, 246, 5, 5, 13, 182, 5, 5, 53, 248, 5, 5, 55, 222, 5, 5, 92, 170, 5, 5, 31, 249, 5, 5, 15, 182, 5, 5, 106, 187, 5, 5, 85, 213, 5, 5, 190, 170, 5, 5, 222, 206, 5, 5, 195, 177, 5, 5, 157, 221, 5, 5, 214, 199, 5, 5, 237, 182, 5, 5, 185, 226, 5, 5, 147, 238, 5, 5, 191, 174, 5, 5, 122, 207, 5, 5, 224, 188, 5, 5, 93, 170, 5, 5, 177, 205, 5, 5, 25, 204, 5, 5, 98, 191, 5, 5, 163, 184, 5, 5, 77, 213, 5, 5, 32, 249, 5, 5, 112, 193, 5, 5, 236, 234, 5, 5, 91, 244, 5, 5, 220, 194, 5, 5, 164, 184, 5, 5, 241, 243, 5, 5, 43, 194, 5, 5, 73, 240, 5, 5, 219, 235, 5, 5, 11, 246, 5, 5, 234, 220, 5, 5, 62, 171, 5, 5, 63, 171, 5, 5, 78, 217, 5, 5, 235, 220, 5, 5, 232, 175, 5, 5, 244, 243, 5, 5, 150, 247, 5, 5, 194, 200, 5, 5, 219, 243, 5, 5, 152, 176, 5, 5, 246, 173, 5, 5, 181, 220, 5, 5, 127, 226, 5, 5, 118, 212, 5, 5, 184, 185, 5, 5, 53, 232, 5, 5, 237, 206, 5, 5, 57, 194, 5, 5, 93, 224, 5, 5, 121, 184, 5, 5, 85, 225, 5, 5, 203, 176, 5, 5, 204, 176, 5, 5, 34, 247, 5, 5, 51, 192, 5, 5, 63, 245, 5, 5, 164, 180, 5, 5, 253, 195, 5, 5, 78, 234, 5, 5, 66, 176, 5, 5, 111, 243, 5, 5, 119, 169, 5, 5, 184, 243, 5, 5, 196, 200, 5, 5, 147, 183, 5, 5, 133, 172, 5, 5, 68, 192, 5, 5, 165, 221, 5, 5, 132, 239, 5, 5, 113, 248, 5, 5, 127, 184, 5, 5, 43, 186, 5, 5, 167, 219, 5, 5, 75, 244, 5, 5, 150, 203, 5, 5, 137, 239, 5, 5, 252, 168, 5, 5, 36, 174, 5, 5, 159, 228, 5, 5, 52, 247, 5, 5, 170, 248, 5, 5, 8, 171, 5, 5, 63, 243, 5, 5, 120, 244, 5, 5, 48, 178, 5, 5, 226, 243, 5, 5, 254, 236, 5, 5, 34, 214, 5, 5, 153, 220, 5, 5, 161, 228, 5, 5, 154, 240, 5, 5, 8, 196, 5, 5, 250, 184, 5, 5, 64, 232, 5, 5, 245, 243, 5, 5, 159, 170, 5, 5, 149, 180, 5, 5, 167, 222, 5, 5, 8, 237, 5, 5, 182, 220, 5, 5, 89, 235, 5, 5, 26, 205, 5, 5, 252, 184, 5, 5, 247, 243, 5, 5, 161, 169, 5, 5, 172, 245, 5, 5, 78, 184, 5, 5, 172, 242, 5, 5, 242, 185, 5, 5, 51, 172, 5, 5, 47, 244, 5, 5, 22, 232, 5, 5, 245, 168, 5, 5, 53, 245, 5, 5, 209, 191, 5, 5, 210, 246, 5, 5, 122, 182, 5, 5, 14, 225, 5, 5, 137, 213, 5, 5, 186, 170, 5, 5, 24, 186, 5, 5, 32, 248, 5, 5, 219, 186, 5, 5, 12, 182, 5, 5, 20, 194, 5, 5, 210, 169, 5, 5, 36, 222, 5, 5, 182, 226, 5, 5, 100, 187, 5, 5, 208, 199, 5, 5, 186, 177, 5, 5, 115, 207, 5, 5, 228, 182, 5, 5, 220, 188, 5, 5, 125, 238, 5, 5, 89, 244, 5, 5, 232, 246, 5, 5, 105, 193, 5, 5, 81, 191, 5, 5, 9, 249, 5, 5, 23, 204, 5, 5, 155, 205, 5, 5, 240, 243, 5, 5, 10, 249, 5, 5, 157, 184, 5, 5, 230, 195, 5, 5, 63, 217, 5, 5, 3, 246, 5, 5, 190, 200, 5, 5, 230, 220, 5, 5, 108, 184, 5, 5, 161, 180, 5, 5, 31, 192, 5, 5, 163, 222, 5, 5, 180, 220, 5, 5, 150, 176, 5, 5, 94, 170, 5, 5, 147, 247, 5, 5, 179, 185, 5, 5, 114, 212, 5, 5, 254, 177, 5, 5, 193, 200, 5, 5, 142, 183, 5, 5, 163, 248, 5, 5, 116, 184, 5, 5, 107, 248, 5, 5, 165, 219, 5, 5, 66, 244, 5, 5, 244, 236, 5, 5, 220, 243, 5, 5, 60, 243, 5, 5, 119, 244, 5, 5, 148, 220, 5, 5, 144, 240, 5, 5, 251, 184, 5, 5, 228, 174, 5, 5, 215, 230, 5, 5, 254, 220, 5, 5, 38, 210, 5, 5, 26, 227, 5, 5, 135, 231, 5, 5, 64, 174, 5, 5, 230, 188, 5, 5, 75, 174, 5, 5, 156, 245, 5, 5, 200, 231, 5, 5, 131, 225, 5, 5, 182, 249, 5, 5, 181, 249, 5, 5, 125, 202, 5, 5, 157, 196, 5, 5, 90, 184, 5, 5, 94, 245, 5, 5, 222, 184, 5, 5, 210, 214, 5, 5, 126, 220, 5, 5, 237, 216, 5, 5, 6, 240, 5, 5, 78, 232, 5, 5, 16, 249, 5, 5, 214, 197, 5, 5, 151, 216, 5, 5, 189, 174, 5, 5, 95, 176, 5, 5, 41, 174, 5, 5, 42, 174, 5, 5, 203, 179, 5, 5, 239, 196, 5, 5, 181, 173, 5, 5, 210, 178, 5, 5, 129, 231, 5, 5, 209, 244, 5, 5, 223, 197, 5, 5, 28, 243, 5, 5, 115, 217, 5, 5, 139, 192, 5, 5, 193, 174, 5, 5, 122, 175, 5, 5, 152, 186, 5, 5, 130, 235, 5, 5, 33, 249, 5, 5, 152, 226, 5, 5, 61, 181, 5, 5, 205, 203, 5, 5, 229, 184, 5, 5, 144, 224, 5, 5, 233, 176, 5, 5, 60, 231, 5, 5, 104, 245, 5, 5, 19, 224, 5, 5, 17, 240, 5, 5, 22, 197, 5, 5, 58, 194, 5, 5, 41, 218, 5, 5, 156, 225, 5, 5, 83, 176, 5, 5, 27, 177, 5, 5, 191, 205, 5, 5, 190, 217, 5, 5, 95, 178, 5, 5, 24, 217, 5, 5, 44, 249, 5, 5, 22, 226, 5, 5, 126, 217, 5, 5, 6, 175, 5, 5, 226, 190, 5, 5, 35, 216, 5, 5, 9, 216, 5, 5, 23, 195, 5, 5, 38, 244, 5, 5, 75, 226, 5, 5, 5, 182, 5, 5, 215, 243, 5, 5, 216, 243, 5, 5, 191, 249, 5, 5, 230, 211, 5, 5, 128, 169, 5, 5, 112, 199, 5, 5, 214, 198, 5, 5, 4, 181, 5, 5, 215, 197, 5, 5, 181, 183, 5, 5, 28, 174, 5, 5, 217, 193, 5, 5, 175, 182, 5, 5, 190, 246, 5, 5, 179, 224, 5, 5, 25, 243, 5, 5, 214, 168, 5, 5, 139, 214, 5, 5, 26, 243, 5, 5, 185, 215, 5, 5, 38, 228, 5, 5, 238, 224, 5, 5, 59, 239, 5, 5, 209, 210, 5, 5, 82, 204, 5, 5, 212, 207, 5, 5, 198, 168, 5, 5, 176, 179, 5, 5, 101, 199, 5, 5, 7, 228, 5, 5, 63, 193, 5, 5, 97, 176, 5, 5, 93, 212, 5, 5, 17, 215, 5, 5, 54, 248, 5, 5, 240, 196, 5, 5, 94, 179, 5, 5, 140, 246, 5, 5, 185, 183, 5, 5, 16, 212, 5, 5, 95, 197, 5, 5, 96, 220, 5, 5, 236, 171, 5, 5, 140, 210, 5, 5, 93, 197, 5, 5, 157, 202, 5, 5, 162, 185, 5, 5, 153, 238, 5, 5, 36, 192, 5, 5, 82, 181, 5, 5, 84, 232, 5, 5, 169, 194, 5, 5, 83, 181, 5, 5, 218, 247, 5, 5, 235, 217, 5, 5, 133, 199, 5, 5, 76, 248, 5, 5, 108, 187, 5, 5, 248, 216, 5, 5, 19, 200, 5, 5, 198, 232, 5, 5, 247, 174, 5, 5, 178, 205, 5, 5, 78, 213, 5, 5, 15, 247, 5, 5, 92, 193, 5, 5, 177, 226, 5, 5, 64, 172, 5, 5, 42, 194, 5, 5, 141, 177, 5, 5, 243, 215, 5, 5, 108, 170, 5, 5, 244, 231, 5, 5, 84, 181, 5, 5, 37, 192, 5, 5, 20, 197, 5, 5, 251, 192, 5, 5, 149, 222, 5, 5, 123, 227, 5, 5, 180, 175, 5, 5, 107, 196, 5, 5, 13, 211, 5, 5, 21, 233, 5, 5, 33, 172, 5, 5, 108, 235, 5, 5, 61, 177, 5, 5, 188, 209, 5, 5, 150, 222, 5, 5, 86, 203, 5, 5, 250, 240, 5, 5, 224, 194, 5, 5, 72, 175, 5, 5, 249, 215, 5, 5, 160, 209, 5, 5, 210, 224, 5, 5, 59, 194, 5, 5, 169, 214, 5, 5, 246, 229, 5, 5, 73, 229, 5, 5, 69, 176, 5, 5, 115, 195, 5, 5, 152, 192, 5, 5, 17, 211, 5, 5, 247, 196, 5, 5, 18, 211, 5, 5, 92, 206, 5, 5, 192, 205, 5, 5, 245, 201, 5, 5, 83, 190, 5, 5, 109, 197, 5, 5, 199, 174, 5, 5, 140, 228, 5, 5, 240, 217, 5, 5, 246, 225, 5, 5, 215, 171, 5, 5, 199, 249, 5, 5, 87, 216, 5, 5, 207, 238, 5, 5, 236, 176, 5, 5, 119, 249, 5, 5, 66, 172, 5, 5, 120, 249, 5, 5, 205, 212, 5, 5, 35, 247, 5, 5, 47, 246, 5, 5, 98, 179, 5, 5, 150, 246, 5, 5, 204, 241, 5, 5, 48, 181, 5, 5, 24, 181, 5, 5, 226, 175, 5, 5, 253, 240, 5, 5, 109, 247, 5, 5, 64, 179, 5, 5, 163, 244, 5, 5, 34, 174, 5, 5, 202, 175, 5, 5, 60, 194, 5, 5, 155, 186, 5, 5, 128, 225, 5, 5, 51, 197, 5, 5, 30, 184, 5, 5, 200, 249, 5, 5, 217, 179, 5, 5, 81, 213, 5, 5, 2, 219, 5, 5, 185, 211, 5, 5, 23, 226, 5, 5, 230, 172, 5, 5, 184, 227, 5, 5, 247, 193, 5, 5, 132, 178, 5, 5, 78, 177, 5, 5, 98, 214, 5, 5, 214, 224, 5, 5, 204, 215, 5, 5, 217, 210, 5, 5, 77, 179, 5, 5, 26, 226, 5, 5, 160, 192, 5, 5, 26, 211, 5, 5, 254, 206, 5, 5, 131, 204, 5, 5, 224, 243, 5, 5, 143, 170, 5, 5, 28, 175, 5, 5, 205, 205, 5, 5, 151, 203, 5, 5, 241, 176, 5, 5, 96, 225, 5, 5, 207, 177, 5, 5, 35, 224, 5, 5, 99, 231, 5, 5, 21, 200, 5, 5, 72, 192, 5, 5, 156, 246, 5, 5, 206, 215, 5, 5, 244, 178, 5, 5, 28, 212, 5, 5, 124, 249, 5, 5, 25, 203, 5, 5, 4, 170, 5, 5, 8, 244, 5, 5, 219, 210, 5, 5, 43, 171, 5, 5, 180, 227, 5, 5, 28, 197, 5, 5, 184, 178, 5, 5, 197, 172, 5, 5, 8, 232, 5, 5, 101, 182, 5, 5, 138, 175, 5, 5, 94, 247, 5, 5, 254, 180, 5, 5, 168, 171, 5, 5, 243, 176, 5, 5, 244, 176, 5, 5, 252, 197, 5, 5, 253, 197, 5, 5, 34, 204, 5, 5, 187, 224, 5, 5, 10, 216, 5, 5, 180, 197, 5, 5, 52, 214, 5, 5, 118, 203, 5, 5, 255, 180, 5, 5, 250, 176, 5, 5, 236, 186, 5, 5, 39, 244, 5, 5, 149, 177, 5, 5, 156, 170, 5, 5, 157, 170, 5, 5, 250, 247, 5, 5, 122, 197, 5, 5, 135, 175, 5, 5, 62, 216, 5, 5, 5, 181, 5, 5, 84, 175, 5, 5, 81, 192, 5, 5, 113, 230, 5, 5, 45, 243, 5, 5, 220, 210, 5, 5, 35, 204, 5, 5, 214, 203, 5, 5, 158, 246, 5, 5, 222, 202, 5, 5, 62, 247, 5, 5, 56, 173, 5, 5, 140, 175, 5, 5, 206, 180, 5, 5, 136, 229, 5, 5, 6, 205, 5, 5, 36, 204, 5, 5, 15, 232, 5, 5, 138, 229, 5, 5, 216, 249, 5, 5, 252, 200, 5, 5, 212, 233, 5, 5, 90, 218, 5, 5, 138, 219, 5, 5, 45, 211, 5, 5, 222, 224, 5, 5, 170, 217, 5, 5, 169, 192, 5, 5, 252, 176, 5, 5, 91, 177, 5, 5, 104, 231, 5, 5, 17, 200, 5, 5, 19, 198, 5, 5, 37, 204, 5, 5, 23, 221, 5, 5, 217, 185, 5, 5, 225, 177, 5, 5, 11, 183, 5, 5, 110, 217, 5, 5, 41, 226, 5, 5, 62, 181, 5, 5, 63, 181, 5, 5, 224, 185, 5, 5, 60, 201, 5, 5, 142, 218, 5, 5, 141, 206, 5, 5, 0, 161, 5, 5, 52, 192, 5, 5, 23, 197, 5, 5, 145, 225, 5, 5, 60, 199, 5, 5, 132, 201, 5, 5, 176, 236, 5, 5, 198, 207, 5, 5, 172, 198, 5, 5, 129, 217, 5, 5, 51, 205, 5, 5, 135, 201, 5, 5, 67, 181, 5, 5, 157, 246, 5, 5, 255, 236, 5, 5, 44, 226, 5, 5, 154, 178, 5, 5, 106, 240, 5, 5, 133, 242, 5, 5, 213, 173, 5, 5, 32, 236, 5, 5, 81, 187, 5, 5, 27, 198, 5, 5, 97, 229, 5, 5, 21, 243, 5, 5, 248, 233, 5, 5, 188, 177, 5, 5, 6, 235, 5, 5, 125, 182, 5, 5, 168, 218, 5, 5, 93, 220, 5, 5, 191, 199, 5, 5, 138, 222, 5, 5, 211, 227, 5, 5, 12, 174, 5, 5, 194, 177, 5, 5, 161, 181, 5, 5, 244, 209, 5, 5, 140, 214, 5, 5, 84, 207, 5, 5, 72, 219, 5, 5, 192, 199, 5, 5, 107, 215, 5, 5, 97, 248, 5, 5, 33, 189, 5, 5, 160, 189, 5, 5, 149, 217, 5, 5, 215, 199, 5, 5, 13, 179, 5, 5, 83, 204, 5, 5, 196, 177, 5, 5, 101, 168, 5, 5, 237, 245, 5, 5, 131, 182, 5, 5, 185, 199, 5, 5, 91, 237, 5, 5, 155, 212, 5, 5, 188, 169, 5, 5, 62, 186, 5, 5, 63, 186, 5, 5, 88, 212, 5, 5, 55, 248, 5, 5, 240, 218, 5, 5, 171, 181, 5, 5, 199, 168, 5, 5, 161, 247, 5, 5, 191, 246, 5, 5, 141, 237, 5, 5, 219, 198, 5, 5, 60, 239, 5, 5, 6, 247, 5, 5, 65, 222, 5, 5, 196, 213, 5, 5, 235, 181, 5, 5, 239, 185, 5, 5, 17, 197, 5, 5, 10, 195, 5, 5, 10, 187, 5, 5, 197, 188, 5, 5, 111, 198, 5, 5, 236, 217, 5, 5, 148, 247, 5, 5, 193, 243, 5, 5, 16, 247, 5, 5, 233, 220, 5, 5, 111, 203, 5, 5, 83, 242, 5, 5, 204, 220, 5, 5, 54, 241, 5, 5, 152, 228, 5, 5, 15, 240, 5, 5, 138, 245, 5, 5, 139, 228, 5, 5, 65, 184, 5, 5, 184, 216, 5, 5, 151, 247, 5, 5, 146, 210, 5, 5, 238, 201, 5, 5, 140, 245, 5, 5, 9, 245, 5, 5, 113, 203, 5, 5, 40, 249, 5, 5, 33, 191, 5, 5, 191, 228, 5, 5, 84, 176, 5, 5, 194, 187, 5, 5, 147, 198, 5, 5, 208, 238, 5, 5, 181, 212, 5, 5, 167, 215, 5, 5, 138, 187, 5, 5, 215, 210, 5, 5, 201, 213, 5, 5, 212, 186, 5, 5, 200, 169, 5, 5, 93, 206, 5, 5, 251, 211, 5, 5, 88, 203, 5, 5, 77, 219, 5, 5, 3, 219, 5, 5, 157, 192, 5, 5, 71, 237, 5, 5, 55, 232, 5, 5, 220, 175, 5, 5, 217, 176, 5, 5, 38, 176, 5, 5, 142, 185, 5, 5, 59, 241, 5, 5, 40, 189, 5, 5, 153, 222, 5, 5, 125, 184, 5, 5, 45, 249, 5, 5, 31, 184, 5, 5, 152, 229, 5, 5, 231, 169, 5, 5, 247, 244, 5, 5, 208, 241, 5, 5, 153, 229, 5, 5, 178, 225, 5, 5, 117, 186, 5, 5, 217, 245, 5, 5, 202, 231, 5, 5, 220, 242, 5, 5, 206, 205, 5, 5, 190, 194, 5, 5, 197, 173, 5, 5, 99, 248, 5, 5, 129, 229, 5, 5, 173, 191, 5, 5, 121, 235, 5, 5, 148, 245, 5, 5, 24, 195, 5, 5, 14, 245, 5, 5, 254, 171, 5, 5, 123, 201, 5, 5, 48, 183, 5, 5, 69, 182, 5, 5, 252, 203, 5, 5, 146, 185, 5, 5, 6, 220, 5, 5, 151, 198, 5, 5, 139, 190, 5, 5, 216, 238, 5, 5, 164, 192, 5, 5, 171, 248, 5, 5, 241, 181, 5, 5, 121, 242, 5, 5, 81, 194, 5, 5, 43, 189, 5, 5, 177, 201, 5, 5, 136, 212, 5, 5, 223, 202, 5, 5, 226, 202, 5, 5, 112, 205, 5, 5, 38, 204, 5, 5, 211, 173, 5, 5, 24, 236, 5, 5, 70, 187, 5, 5, 2, 235, 5, 5, 171, 177, 5, 5, 159, 218, 5, 5, 95, 248, 5, 5, 144, 181, 5, 5, 78, 206, 5, 5, 69, 219, 5, 5, 29, 189, 5, 5, 53, 186, 5, 5, 211, 198, 5, 5, 83, 205, 5, 5, 160, 247, 5, 5, 185, 246, 5, 5, 17, 239, 5, 5, 154, 189, 5, 5, 226, 245, 5, 5, 126, 202, 5, 5, 132, 237, 5, 5, 172, 216, 5, 5, 43, 222, 5, 5, 191, 243, 5, 5, 233, 246, 5, 5, 4, 195, 5, 5, 137, 247, 5, 5, 217, 217, 5, 5, 156, 205, 5, 5, 5, 195, 5, 5, 131, 245, 5, 5, 57, 184, 5, 5, 106, 203, 5, 5, 210, 210, 5, 5, 189, 169, 5, 5, 19, 191, 5, 5, 131, 187, 5, 5, 187, 228, 5, 5, 81, 203, 5, 5, 80, 176, 5, 5, 25, 249, 5, 5, 216, 176, 5, 5, 15, 184, 5, 5, 140, 192, 5, 5, 147, 229, 5, 5, 147, 222, 5, 5, 132, 212, 5, 5, 209, 242, 5, 5, 195, 231, 5, 5, 211, 210, 5, 5, 193, 205, 5, 5, 142, 245, 5, 5, 244, 203, 5, 5, 218, 233, 5, 5, 64, 186, 5, 5, 113, 176, 5, 5, 116, 176, 5, 5, 51, 213, 5, 5, 245, 249, 5, 5, 218, 170, 5, 5, 146, 200, 5, 5, 147, 200, 5, 5, 121, 176, 5, 5, 95, 235, 5, 5, 44, 169, 5, 5, 221, 170, 5, 5, 222, 170, 5, 5, 223, 170, 5, 5, 96, 235, 5, 5, 224, 170, 5, 5, 14, 169, 5, 5, 129, 176, 5, 5, 225, 170, 5, 5, 226, 170, 5, 5, 254, 173, 5, 5, 47, 219, 5, 5, 106, 211, 5, 5, 108, 211, 5, 5, 91, 173, 5, 5, 72, 176, 5, 5, 71, 176, 5, 5, 60, 238, 5, 5, 182, 218, 5, 5, 164, 170, 5, 5, 165, 170, 5, 5, 207, 221, 5, 5, 113, 241, 5, 5, 119, 203, 5, 5, 120, 177, 5, 5, 7, 173, 5, 5, 191, 184, 5, 5, 38, 215, 5, 5, 116, 241, 5, 5, 117, 241, 5, 5, 227, 214, 5, 5, 202, 235, 5, 5, 104, 238, 5, 5, 198, 187, 5, 5, 227, 206, 5, 5, 64, 214, 5, 5, 151, 243, 5, 5, 196, 228, 5, 5, 138, 227, 5, 5, 174, 248, 5, 5, 111, 240, 5, 5, 121, 177, 5, 5, 116, 243, 5, 5, 220, 195, 5, 5, 71, 188, 5, 5, 30, 236, 5, 5, 113, 182, 5, 5, 126, 230, 5, 5, 122, 177, 5, 5, 101, 181, 5, 5, 208, 187, 5, 5, 159, 245, 5, 5, 123, 177, 5, 5, 221, 195, 5, 5, 229, 242, 5, 5, 240, 228, 5, 5, 236, 202, 5, 5, 180, 174, 5, 5, 226, 173, 5, 5, 175, 210, 5, 5, 145, 226, 5, 5, 211, 246, 5, 5, 115, 238, 5, 5, 132, 196, 5, 5, 47, 193, 5, 5, 2, 174, 5, 5, 178, 177, 5, 5, 247, 181, 5, 5, 220, 178, 5, 5, 231, 213, 5, 5, 33, 248, 5, 5, 180, 179, 5, 5, 48, 244, 5, 5, 183, 225, 5, 5, 195, 222, 5, 5, 253, 227, 5, 5, 196, 217, 5, 5, 79, 196, 5, 5, 49, 191, 5, 5, 62, 180, 5, 5, 103, 241, 5, 5, 49, 208, 5, 5, 253, 169, 5, 5, 13, 192, 5, 5, 7, 210, 5, 5, 126, 238, 5, 5, 78, 195, 5, 5, 143, 248, 5, 5, 188, 203, 5, 5, 207, 235, 5, 5, 187, 227, 5, 5, 181, 223, 5, 5, 37, 222, 5, 5, 186, 225, 5, 5, 60, 212, 5, 5, 133, 189, 5, 5, 178, 210, 5, 5, 13, 181, 5, 5, 135, 196, 5, 5, 57, 235, 5, 5, 208, 235, 5, 5, 248, 171, 5, 5, 12, 241, 5, 5, 254, 232, 5, 5, 56, 217, 5, 5, 76, 227, 5, 5, 243, 247, 5, 5, 57, 217, 5, 5, 28, 179, 5, 5, 29, 179, 5, 5, 105, 227, 5, 5, 90, 196, 5, 5, 54, 226, 5, 5, 96, 180, 5, 5, 196, 238, 5, 5, 162, 245, 5, 5, 252, 226, 5, 5, 20, 187, 5, 5, 138, 230, 5, 5, 44, 222, 5, 5, 140, 174, 5, 5, 247, 223, 5, 5, 27, 244, 5, 5, 51, 218, 5, 5, 122, 183, 5, 5, 241, 202, 5, 5, 169, 224, 5, 5, 52, 191, 5, 5, 76, 202, 5, 5, 87, 186, 5, 5, 178, 200, 5, 5, 228, 169, 5, 5, 101, 177, 5, 5, 216, 197, 5, 5, 254, 169, 5, 5, 160, 190, 5, 5, 168, 177, 5, 5, 164, 205, 5, 5, 45, 241, 5, 5, 143, 247, 5, 5, 234, 195, 5, 5, 42, 242, 5, 5, 68, 176, 5, 5, 244, 199, 5, 5, 206, 228, 5, 5, 59, 226, 5, 5, 47, 239, 5, 5, 137, 177, 5, 5, 231, 242, 5, 5, 112, 206, 5, 5, 10, 170, 5, 5, 177, 211, 5, 5, 177, 241, 5, 5, 87, 178, 5, 5, 68, 224, 5, 5, 16, 181, 5, 5, 82, 224, 5, 5, 165, 236, 5, 5, 215, 175, 5, 5, 192, 174, 5, 5, 8, 226, 5, 5, 59, 242, 5, 5, 228, 221, 5, 5, 193, 245, 5, 5, 48, 241, 5, 5, 125, 243, 5, 5, 172, 181, 5, 5, 216, 170, 5, 5, 199, 187, 5, 5, 173, 181, 5, 5, 188, 231, 5, 5, 202, 190, 5, 5, 71, 217, 5, 5, 142, 178, 5, 5, 140, 177, 5, 5, 7, 229, 5, 5, 33, 210, 5, 5, 148, 238, 5, 5, 36, 186, 5, 5, 150, 237, 5, 5, 101, 175, 5, 5, 148, 204, 5, 5, 217, 235, 5, 5, 202, 224, 5, 5, 50, 179, 5, 5, 194, 174, 5, 5, 232, 242, 5, 5, 11, 224, 5, 5, 203, 224, 5, 5, 136, 215, 5, 5, 189, 206, 5, 5, 151, 237, 5, 5, 231, 186, 5, 5, 30, 245, 5, 5, 113, 168, 5, 5, 71, 222, 5, 5, 45, 172, 5, 5, 253, 174, 5, 5, 20, 224, 5, 5, 6, 244, 5, 5, 115, 245, 5, 5, 21, 181, 5, 5, 58, 179, 5, 5, 42, 205, 5, 5, 200, 174, 5, 5, 75, 177, 5, 5, 239, 203, 5, 5, 6, 250, 5, 5, 115, 218, 5, 5, 76, 215, 5, 5, 60, 235, 5, 5, 92, 242, 5, 5, 161, 238, 5, 5, 187, 181, 5, 5, 136, 203, 5, 5, 116, 197, 5, 5, 82, 222, 5, 5, 138, 170, 5, 5, 124, 237, 5, 5, 238, 206, 5, 5, 195, 233, 5, 5, 102, 224, 5, 5, 212, 187, 5, 5, 216, 244, 5, 5, 230, 225, 5, 5, 4, 182, 5, 5, 200, 208, 5, 5, 178, 170, 5, 5, 181, 170, 5, 5, 130, 200, 5, 5, 42, 202, 5, 5, 224, 242, 5, 5, 173, 237, 5, 5, 130, 206, 5, 5, 92, 202, 5, 5, 244, 238, 5, 5, 216, 226, 5, 5, 188, 178, 5, 5, 198, 214, 5, 5, 190, 240, 5, 5, 74, 220, 5, 5, 242, 187, 5, 5, 129, 241, 5, 5, 38, 207, 5, 5, 14, 219, 5, 5, 229, 216, 5, 5, 170, 230, 5, 5, 199, 199, 5, 5, 164, 183, 5, 5, 180, 198, 5, 5, 48, 171, 5, 5, 139, 182, 5, 5, 25, 234, 5, 5, 230, 209, 5, 5, 214, 233, 5, 5, 70, 221, 5, 5, 47, 169, 5, 5, 173, 242, 5, 5, 58, 177, 5, 5, 202, 191, 5, 5, 102, 233, 5, 5, 53, 169, 5, 5, 28, 230, 5, 5, 68, 197, 5, 5, 29, 241, 5, 5, 244, 187, 5, 5, 230, 224, 5, 5, 9, 217, 5, 5, 52, 170, 5, 5, 211, 212, 5, 5, 100, 171, 5, 5, 219, 220, 5, 5, 165, 169, 5, 5, 64, 228, 5, 5, 253, 178, 5, 5, 162, 249, 5, 5, 2, 238, 5, 5, 220, 203, 5, 5, 175, 199, 5, 5, 29, 187, 5, 5, 204, 247, 5, 5, 172, 221, 5, 5, 94, 199, 5, 5, 20, 242, 5, 5, 150, 184, 5, 5, 163, 188, 5, 5, 71, 216, 5, 5, 212, 246, 5, 5, 110, 192, 5, 5, 110, 190, 5, 5, 128, 189, 5, 5, 29, 234, 5, 5, 157, 194, 5, 5, 47, 231, 5, 5, 46, 187, 5, 5, 168, 211, 5, 5, 39, 201, 5, 5, 78, 193, 5, 5, 143, 199, 5, 5, 68, 246, 5, 5, 58, 201, 5, 5, 118, 243, 5, 5, 93, 236, 5, 5, 93, 174, 5, 5, 93, 180, 5, 5, 189, 230, 5, 5, 237, 205, 5, 5, 51, 184, 5, 5, 56, 230, 5, 5, 237, 183, 5, 5, 75, 185, 5, 5, 114, 188, 5, 5, 40, 201, 5, 5, 80, 193, 5, 5, 187, 185, 5, 5, 49, 198, 5, 5, 160, 240, 5, 5, 183, 171, 5, 5, 120, 231, 5, 5, 171, 169, 5, 5, 127, 202, 5, 5, 77, 243, 5, 5, 28, 172, 5, 5, 53, 233, 5, 5, 72, 214, 5, 5, 3, 213, 5, 5, 175, 216, 5, 5, 150, 187, 5, 5, 138, 247, 5, 5, 30, 225, 5, 5, 165, 249, 5, 5, 174, 213, 5, 5, 169, 200, 5, 5, 132, 210, 5, 5, 239, 173, 5, 5, 40, 241, 5, 5, 34, 172, 5, 5, 163, 232, 5, 5, 223, 177, 5, 5, 12, 197, 5, 5, 206, 240, 5, 5, 241, 215, 5, 5, 75, 238, 5, 5, 77, 180, 5, 5, 155, 236, 5, 5, 155, 207, 5, 5, 116, 219, 5, 5, 182, 169, 5, 5, 162, 181, 5, 5, 139, 222, 5, 5, 162, 197, 5, 5, 241, 241, 5, 5, 120, 243, 5, 5, 110, 189, 5, 5, 245, 199, 5, 5, 166, 232, 5, 5, 167, 232, 5, 5, 190, 223, 5, 5, 108, 225, 5, 5, 39, 209, 5, 5, 230, 230, 5, 5, 48, 219, 5, 5, 175, 175, 5, 5, 26, 249, 5, 5, 170, 249, 5, 5, 20, 238, 5, 5, 39, 230, 5, 5, 172, 232, 5, 5, 82, 243, 5, 5, 128, 188, 5, 5, 194, 240, 5, 5, 40, 170, 5, 5, 123, 207, 5, 5, 189, 173, 5, 5, 189, 183, 5, 5, 161, 203, 5, 5, 239, 239, 5, 5, 82, 248, 5, 5, 254, 189, 5, 5, 244, 215, 5, 5, 68, 236, 5, 5, 22, 245, 5, 5, 20, 207, 5, 5, 245, 215, 5, 5, 194, 234, 5, 5, 190, 178, 5, 5, 147, 170, 5, 5, 184, 235, 5, 5, 121, 170, 5, 5, 103, 244, 5, 5, 11, 229, 5, 5, 76, 246, 5, 5, 131, 207, 5, 5, 135, 220, 5, 5, 234, 203, 5, 5, 220, 213, 5, 5, 233, 177, 5, 5, 3, 208, 5, 5, 34, 238, 5, 5, 154, 172, 5, 5, 152, 199, 5, 5, 102, 183, 5, 5, 4, 208, 5, 5, 174, 249, 5, 5, 218, 199, 5, 5, 45, 203, 5, 5, 222, 243, 5, 5, 41, 173, 5, 5, 18, 241, 5, 5, 69, 192, 5, 5, 248, 236, 5, 5, 48, 173, 5, 5, 89, 177, 5, 5, 106, 204, 5, 5, 101, 190, 5, 5, 39, 231, 5, 5, 109, 183, 5, 5, 229, 243, 5, 5, 216, 202, 5, 5, 67, 241, 5, 5, 236, 179, 5, 5, 50, 217, 5, 5, 214, 248, 5, 5, 130, 212, 5, 5, 183, 247, 5, 5, 197, 238, 5, 5, 204, 184, 5, 5, 156, 241, 5, 5, 181, 196, 5, 5, 156, 236, 5, 5, 239, 249, 5, 5, 85, 207, 5, 5, 4, 246, 5, 5, 223, 234, 5, 5, 102, 180, 5, 5, 190, 245, 5, 5, 12, 183, 5, 5, 188, 242, 5, 5, 134, 184, 5, 5, 121, 243, 5, 5, 7, 225, 5, 5, 81, 226, 5, 5, 18, 215, 5, 5, 8, 228, 5, 5, 229, 176, 5, 5, 234, 187, 5, 5, 65, 186, 5, 5, 227, 223, 5, 5, 241, 213, 5, 5, 69, 175, 5, 5, 194, 243, 5, 5, 55, 209, 5, 5, 128, 201, 5, 5, 81, 176, 5, 5, 70, 175, 5, 5, 109, 241, 5, 5, 27, 227, 5, 5, 193, 246, 5, 5, 245, 231, 5, 5, 140, 194, 5, 5, 126, 174, 5, 5, 38, 240, 5, 5, 124, 227, 5, 5, 15, 195, 5, 5, 172, 207, 5, 5, 119, 199, 5, 5, 41, 224, 5, 5, 220, 201, 5, 5, 26, 214, 5, 5, 247, 249, 5, 5, 217, 187, 5, 5, 236, 236, 5, 5, 63, 220, 5, 5, 231, 210, 5, 5, 15, 229, 5, 5, 194, 205, 5, 5, 9, 201, 5, 5, 73, 236, 5, 5, 199, 225, 5, 5, 109, 212, 5, 5, 245, 244, 5, 5, 57, 176, 5, 5, 43, 225, 5, 5, 249, 249, 5, 5, 166, 248, 5, 5, 237, 176, 5, 5, 76, 200, 5, 5, 24, 226, 5, 5, 56, 232, 5, 5, 119, 180, 5, 5, 203, 189, 5, 5, 196, 234, 5, 5, 44, 234, 5, 5, 71, 225, 5, 5, 82, 217, 5, 5, 58, 176, 5, 5, 137, 243, 5, 5, 244, 213, 5, 5, 235, 198, 5, 5, 206, 223, 5, 5, 56, 208, 5, 5, 242, 217, 5, 5, 96, 175, 5, 5, 79, 177, 5, 5, 138, 243, 5, 5, 4, 241, 5, 5, 97, 168, 5, 5, 181, 244, 5, 5, 218, 187, 5, 5, 132, 225, 5, 5, 148, 194, 5, 5, 112, 213, 5, 5, 37, 174, 5, 5, 116, 242, 5, 5, 39, 202, 5, 5, 9, 244, 5, 5, 102, 201, 5, 5, 94, 238, 5, 5, 150, 194, 5, 5, 255, 171, 5, 5, 25, 195, 5, 5, 32, 231, 5, 5, 57, 225, 5, 5, 25, 182, 5, 5, 118, 211, 5, 5, 152, 239, 5, 5, 90, 202, 5, 5, 123, 197, 5, 5, 4, 237, 5, 5, 153, 239, 5, 5, 232, 210, 5, 5, 41, 219, 5, 5, 158, 235, 5, 5, 85, 175, 5, 5, 11, 237, 5, 5, 112, 204, 5, 5, 74, 208, 5, 5, 75, 208, 5, 5, 233, 210, 5, 5, 2, 234, 5, 5, 28, 195, 5, 5, 67, 220, 5, 5, 77, 208, 5, 5, 17, 237, 5, 5, 206, 170, 5, 5, 56, 172, 5, 5, 48, 222, 5, 5, 108, 241, 5, 5, 56, 222, 5, 5, 87, 222, 5, 5, 69, 202, 201, 160, 231, 0, 5, 5, 247, 237, 5, 5, 110, 203, 5, 5, 43, 202, 5, 5, 170, 195, 5, 5, 171, 195, 5, 5, 51, 217, 5, 5, 192, 238, 5, 5, 160, 203, 5, 5, 111, 178, 5, 5, 75, 245, 5, 5, 237, 179, 5, 5, 239, 213, 5, 5, 58, 217, 5, 5, 181, 168, 5, 5, 54, 184, 5, 5, 187, 245, 5, 5, 130, 246, 5, 5, 182, 168, 5, 5, 59, 206, 5, 5, 55, 184, 5, 5, 9, 210, 5, 5, 160, 179, 5, 5, 128, 220, 5, 5, 40, 216, 5, 5, 82, 199, 5, 5, 231, 175, 5, 5, 74, 249, 5, 5, 114, 182, 5, 5, 37, 190, 5, 5, 38, 190, 5, 5, 40, 188, 5, 5, 17, 185, 5, 5, 141, 214, 5, 5, 58, 207, 5, 5, 189, 218, 5, 5, 40, 179, 5, 5, 132, 223, 5, 5, 124, 231, 5, 5, 48, 239, 5, 5, 245, 172, 5, 5, 180, 221, 5, 5, 136, 227, 5, 5, 216, 230, 5, 5, 151, 211, 5, 5, 60, 215, 5, 5, 64, 217, 5, 5, 33, 194, 5, 5, 47, 213, 5, 5, 239, 237, 5, 5, 186, 195, 5, 5, 206, 168, 5, 5, 148, 182, 5, 5, 15, 174, 5, 5, 38, 234, 5, 5, 90, 180, 5, 5, 29, 243, 5, 5, 63, 215, 5, 5, 186, 183, 5, 5, 227, 212, 5, 5, 246, 209, 5, 5, 227, 233, 5, 5, 112, 181, 5, 5, 224, 197, 5, 5, 17, 181, 5, 5, 254, 185, 5, 5, 9, 240, 5, 5, 109, 215, 5, 5, 25, 169, 5, 5, 145, 219, 5, 5, 129, 218, 5, 5, 182, 173, 5, 5, 98, 211, 5, 5, 13, 183, 5, 5, 105, 243, 5, 5, 202, 238, 5, 5, 125, 216, 5, 5, 228, 212, 5, 5, 154, 187, 5, 5, 78, 189, 5, 5, 235, 239, 5, 5, 34, 198, 5, 5, 163, 179, 5, 5, 61, 239, 5, 5, 84, 247, 5, 5, 83, 231, 5, 5, 179, 184, 5, 5, 190, 218, 5, 5, 204, 191, 5, 5, 8, 225, 5, 5, 186, 198, 5, 5, 195, 242, 5, 5, 89, 205, 5, 5, 174, 181, 5, 5, 126, 216, 5, 5, 45, 181, 5, 5, 37, 249, 5, 5, 163, 210, 5, 5, 118, 227, 5, 5, 66, 222, 5, 5, 248, 208, 5, 5, 66, 186, 5, 5, 226, 198, 5, 5, 86, 204, 5, 5, 118, 171, 5, 5, 138, 223, 5, 5, 103, 186, 5, 5, 203, 171, 5, 5, 229, 212, 5, 5, 72, 242, 5, 5, 139, 223, 5, 5, 29, 250, 5, 5, 251, 171, 5, 5, 52, 241, 5, 5, 101, 226, 5, 5, 113, 193, 5, 5, 203, 245, 5, 5, 255, 221, 5, 5, 67, 222, 5, 5, 143, 246, 5, 5, 96, 197, 5, 5, 13, 173, 5, 5, 190, 221, 5, 5, 191, 221, 5, 5, 79, 235, 5, 5, 77, 245, 5, 5, 134, 169, 5, 5, 198, 188, 5, 5, 109, 170, 5, 5, 136, 221, 5, 5, 123, 175, 5, 5, 233, 221, 5, 5, 204, 171, 5, 5, 58, 248, 5, 5, 248, 174, 5, 5, 156, 243, 5, 5, 213, 213, 5, 5, 28, 227, 5, 5, 111, 215, 5, 5, 16, 184, 5, 5, 193, 244, 5, 5, 190, 204, 5, 5, 68, 215, 5, 5, 17, 184, 5, 5, 166, 202, 5, 5, 31, 243, 5, 5, 230, 212, 5, 5, 34, 237, 5, 5, 42, 169, 5, 5, 161, 171, 5, 5, 104, 195, 5, 5, 2, 186, 5, 5, 215, 222, 5, 5, 48, 246, 5, 5, 187, 209, 5, 5, 84, 231, 5, 5, 85, 231, 5, 5, 51, 179, 5, 5, 66, 193, 5, 5, 211, 209, 5, 5, 66, 225, 5, 5, 124, 190, 5, 5, 203, 238, 5, 5, 77, 223, 5, 5, 216, 199, 5, 5, 89, 198, 5, 5, 160, 169, 5, 5, 44, 194, 5, 5, 31, 227, 5, 5, 40, 234, 5, 5, 84, 189, 5, 5, 225, 194, 5, 5, 220, 174, 5, 5, 17, 182, 5, 5, 162, 206, 5, 5, 119, 171, 5, 5, 72, 222, 5, 5, 161, 209, 5, 5, 68, 193, 5, 5, 242, 239, 5, 5, 39, 198, 5, 5, 149, 247, 5, 5, 44, 175, 5, 5, 207, 232, 5, 5, 32, 227, 5, 5, 127, 209, 5, 5, 221, 201, 5, 5, 42, 192, 5, 5, 84, 242, 5, 5, 244, 234, 5, 5, 131, 218, 5, 5, 20, 250, 5, 5, 26, 247, 5, 5, 249, 216, 5, 5, 132, 220, 5, 5, 254, 174, 5, 5, 85, 232, 5, 5, 41, 234, 5, 5, 238, 217, 5, 5, 231, 212, 5, 5, 189, 226, 5, 5, 222, 247, 5, 5, 232, 232, 5, 5, 46, 209, 5, 5, 130, 199, 5, 5, 157, 237, 5, 5, 250, 231, 5, 5, 45, 232, 5, 5, 102, 234, 5, 5, 40, 198, 5, 5, 209, 172, 5, 5, 119, 201, 5, 5, 144, 192, 5, 5, 37, 213, 5, 5, 25, 219, 5, 5, 92, 208, 5, 5, 85, 238, 5, 5, 202, 239, 5, 5, 11, 187, 5, 5, 71, 168, 5, 5, 18, 180, 5, 5, 80, 241, 5, 5, 246, 219, 5, 5, 199, 198, 5, 5, 115, 215, 5, 5, 37, 206, 5, 5, 85, 223, 5, 5, 0, 141, 5, 5, 166, 179, 5, 5, 53, 188, 5, 5, 99, 219, 5, 5, 74, 222, 5, 5, 49, 199, 5, 5, 81, 217, 5, 5, 25, 233, 5, 5, 141, 245, 5, 5, 131, 234, 5, 5, 255, 243, 5, 5, 20, 226, 5, 5, 107, 177, 5, 5, 156, 186, 5, 5, 103, 189, 5, 5, 88, 247, 5, 5, 61, 227, 5, 5, 254, 205, 5, 5, 175, 207, 5, 5, 61, 201, 5, 5, 154, 228, 5, 5, 232, 233, 5, 5, 92, 243, 5, 5, 201, 169, 5, 5, 157, 230, 5, 5, 26, 224, 5, 5, 93, 242, 5, 5, 30, 173, 5, 5, 11, 180, 5, 5, 216, 171, 5, 5, 54, 188, 5, 5, 94, 193, 5, 5, 87, 189, 5, 5, 248, 176, 5, 5, 103, 183, 5, 5, 14, 173, 5, 5, 141, 228, 5, 5, 36, 247, 5, 5, 95, 223, 5, 5, 28, 235, 5, 5, 47, 190, 5, 5, 252, 241, 5, 5, 156, 226, 5, 5, 217, 199, 5, 5, 233, 248, 5, 5, 67, 206, 5, 5, 42, 234, 5, 5, 156, 216, 5, 5, 111, 221, 5, 5, 13, 188, 5, 5, 68, 206, 5, 5, 240, 237, 5, 5, 129, 175, 5, 5, 117, 244, 5, 5, 248, 196, 5, 5, 144, 232, 5, 5, 190, 226, 5, 5, 62, 207, 5, 5, 57, 214, 5, 5, 30, 202, 5, 5, 46, 212, 5, 5, 100, 219, 5, 5, 127, 174, 5, 5, 83, 185, 5, 5, 83, 202, 5, 5, 217, 225, 5, 5, 98, 171, 5, 5, 64, 248, 5, 5, 12, 246, 5, 5, 92, 227, 5, 5, 194, 204, 5, 5, 251, 249, 5, 5, 117, 197, 5, 5, 159, 173, 5, 5, 233, 242, 5, 5, 65, 194, 5, 5, 23, 185, 5, 5, 172, 179, 5, 5, 201, 225, 5, 5, 143, 173, 5, 5, 95, 206, 5, 5, 195, 187, 5, 5, 96, 204, 5, 5, 237, 212, 5, 5, 198, 205, 5, 5, 36, 202, 5, 5, 202, 215, 5, 5, 254, 213, 5, 5, 150, 197, 5, 5, 251, 208, 5, 5, 252, 249, 5, 5, 185, 212, 5, 5, 85, 168, 5, 5, 238, 212, 5, 5, 148, 232, 5, 5, 242, 235, 5, 5, 146, 248, 5, 5, 224, 201, 5, 5, 234, 198, 5, 5, 61, 199, 5, 5, 212, 224, 5, 5, 77, 200, 5, 5, 134, 180, 5, 5, 118, 210, 5, 5, 22, 176, 5, 5, 46, 249, 5, 5, 51, 246, 5, 5, 214, 169, 5, 5, 24, 211, 5, 5, 158, 249, 5, 5, 59, 176, 5, 5, 47, 225, 5, 5, 12, 180, 5, 5, 171, 214, 5, 5, 121, 215, 5, 5, 167, 248, 5, 5, 61, 192, 5, 5, 100, 242, 5, 5, 213, 195, 5, 5, 214, 186, 5, 5, 93, 207, 5, 5, 119, 173, 5, 5, 128, 226, 5, 5, 15, 231, 5, 5, 67, 203, 5, 5, 202, 225, 5, 5, 139, 186, 5, 5, 108, 177, 5, 5, 224, 222, 5, 5, 216, 245, 5, 5, 199, 205, 5, 5, 33, 208, 5, 5, 200, 205, 5, 5, 31, 190, 5, 5, 24, 171, 5, 5, 143, 184, 5, 5, 184, 200, 5, 5, 30, 199, 5, 5, 149, 182, 5, 5, 166, 230, 5, 5, 20, 210, 5, 5, 157, 228, 5, 5, 221, 243, 5, 5, 224, 189, 5, 5, 167, 178, 5, 5, 3, 232, 5, 5, 69, 213, 5, 5, 192, 191, 5, 5, 114, 203, 5, 5, 48, 182, 5, 5, 230, 207, 5, 5, 114, 198, 5, 5, 78, 240, 5, 5, 214, 178, 5, 5, 69, 203, 5, 5, 162, 187, 5, 5, 92, 232, 5, 5, 179, 180, 5, 5, 126, 227, 5, 5, 17, 229, 5, 5, 122, 249, 5, 5, 128, 184, 5, 5, 4, 219, 5, 5, 161, 192, 5, 5, 197, 181, 5, 5, 43, 198, 5, 5, 35, 174, 5, 5, 27, 226, 5, 5, 153, 244, 5, 5, 11, 190, 5, 5, 72, 237, 5, 5, 204, 180, 5, 5, 203, 231, 5, 5, 209, 241, 5, 5, 50, 199, 5, 5, 138, 221, 5, 5, 223, 190, 5, 5, 79, 229, 5, 5, 129, 184, 5, 5, 80, 245, 5, 5, 213, 172, 5, 5, 90, 216, 5, 5, 199, 221, 5, 5, 42, 189, 5, 5, 9, 200, 5, 5, 129, 226, 5, 5, 165, 209, 5, 5, 6, 216, 5, 5, 7, 216, 5, 5, 122, 189, 5, 5, 62, 227, 5, 5, 183, 176, 5, 5, 134, 190, 5, 5, 45, 238, 5, 5, 253, 208, 5, 5, 72, 194, 5, 5, 188, 180, 5, 5, 66, 194, 5, 5, 165, 223, 5, 5, 10, 200, 5, 5, 206, 189, 5, 5, 33, 235, 5, 5, 66, 181, 5, 5, 121, 195, 5, 5, 218, 245, 5, 5, 179, 170, 5, 5, 92, 247, 5, 5, 49, 249, 5, 5, 105, 234, 5, 5, 241, 237, 5, 5, 199, 207, 5, 5, 7, 212, 5, 5, 0, 144, 5, 5, 164, 195, 5, 5, 173, 215, 5, 5, 179, 207, 5, 5, 164, 224, 5, 5, 148, 177, 5, 5, 80, 169, 5, 5, 205, 231, 5, 5, 26, 203, 5, 5, 165, 224, 5, 5, 126, 198, 5, 5, 172, 204, 5, 5, 169, 237, 5, 5, 49, 238, 5, 5, 132, 211, 5, 5, 216, 229, 5, 5, 231, 218, 5, 5, 134, 225, 5, 5, 166, 224, 5, 5, 207, 215, 5, 5, 195, 202, 5, 5, 54, 223, 5, 5, 23, 176, 5, 5, 226, 171, 5, 5, 29, 212, 5, 5, 119, 177, 5, 5, 17, 170, 5, 5, 206, 219, 5, 5, 24, 185, 5, 5, 50, 249, 5, 5, 47, 230, 5, 5, 146, 240, 5, 5, 253, 190, 5, 5, 161, 226, 5, 5, 173, 204, 5, 5, 115, 198, 5, 5, 130, 224, 5, 5, 41, 220, 5, 5, 207, 223, 5, 5, 160, 228, 5, 5, 141, 188, 5, 5, 16, 246, 5, 5, 17, 246, 5, 5, 62, 201, 5, 5, 138, 239, 5, 5, 222, 242, 5, 5, 147, 225, 5, 5, 29, 211, 5, 5, 66, 231, 5, 5, 73, 193, 5, 5, 110, 185, 5, 5, 193, 206, 5, 5, 138, 197, 5, 5, 186, 223, 5, 5, 214, 249, 5, 5, 167, 224, 5, 5, 222, 231, 5, 5, 107, 183, 5, 5, 154, 229, 5, 5, 226, 209, 5, 5, 142, 205, 5, 5, 168, 224, 5, 5, 214, 211, 5, 5, 202, 249, 5, 5, 159, 227, 5, 5, 108, 234, 5, 5, 237, 186, 5, 5, 87, 235, 5, 5, 70, 203, 5, 5, 116, 222, 5, 5, 157, 168, 5, 5, 17, 207, 5, 5, 142, 209, 5, 5, 123, 206, 5, 5, 148, 170, 5, 5, 130, 229, 5, 5, 175, 204, 5, 5, 246, 178, 5, 5, 188, 219, 5, 5, 147, 249, 5, 5, 175, 238, 5, 5, 211, 205, 5, 5, 127, 168, 5, 5, 51, 199, 5, 5, 208, 215, 5, 5, 24, 177, 5, 5, 100, 214, 5, 5, 147, 173, 5, 5, 97, 225, 5, 5, 35, 207, 5, 5, 211, 240, 5, 5, 92, 173, 5, 5, 108, 183, 5, 5, 114, 196, 5, 5, 9, 171, 5, 5, 225, 222, 5, 5, 70, 205, 5, 5, 135, 234, 5, 5, 185, 176, 5, 5, 243, 204, 5, 5, 227, 243, 5, 5, 78, 194, 5, 5, 147, 172, 5, 5, 48, 202, 5, 5, 224, 231, 5, 5, 33, 231, 5, 5, 173, 198, 5, 5, 11, 223, 5, 5, 5, 170, 5, 5, 67, 245, 5, 5, 174, 215, 5, 5, 82, 174, 5, 5, 216, 205, 5, 5, 49, 190, 5, 5, 165, 192, 5, 5, 36, 214, 5, 5, 142, 191, 5, 5, 234, 215, 5, 5, 216, 213, 5, 5, 255, 203, 5, 5, 7, 220, 5, 5, 136, 234, 5, 5, 189, 219, 5, 5, 83, 174, 5, 5, 48, 200, 5, 5, 102, 223, 5, 5, 177, 204, 5, 5, 80, 210, 5, 5, 228, 190, 5, 5, 128, 213, 5, 5, 107, 224, 5, 5, 106, 182, 5, 5, 36, 216, 5, 5, 206, 217, 5, 5, 75, 237, 5, 5, 98, 225, 5, 5, 237, 232, 5, 5, 5, 198, 5, 5, 191, 194, 5, 5, 11, 250, 5, 5, 153, 203, 5, 5, 92, 216, 5, 5, 103, 201, 5, 5, 239, 180, 5, 5, 241, 233, 5, 5, 201, 243, 5, 5, 78, 192, 5, 5, 4, 194, 5, 5, 95, 247, 5, 5, 198, 178, 5, 5, 246, 235, 5, 5, 174, 240, 5, 5, 215, 180, 5, 5, 6, 198, 5, 5, 133, 211, 5, 5, 202, 243, 5, 5, 53, 214, 5, 5, 192, 226, 5, 5, 107, 182, 5, 5, 84, 174, 5, 5, 247, 179, 5, 5, 154, 220, 5, 5, 95, 198, 5, 5, 192, 193, 5, 5, 247, 182, 5, 5, 108, 224, 5, 5, 144, 205, 5, 5, 140, 197, 5, 5, 145, 191, 5, 5, 122, 242, 5, 5, 31, 203, 5, 5, 243, 248, 5, 5, 11, 216, 5, 5, 82, 194, 5, 5, 247, 248, 5, 5, 176, 201, 5, 5, 160, 170, 5, 5, 193, 226, 5, 5, 140, 190, 5, 5, 52, 238, 5, 5, 50, 181, 5, 5, 197, 187, 5, 5, 64, 178, 5, 5, 125, 197, 5, 5, 49, 183, 5, 5, 150, 177, 5, 5, 205, 169, 5, 5, 156, 239, 5, 5, 31, 168, 5, 5, 132, 249, 5, 5, 224, 235, 5, 5, 174, 179, 5, 5, 69, 248, 5, 5, 22, 189, 5, 5, 172, 248, 5, 5, 87, 192, 5, 5, 42, 211, 5, 5, 216, 188, 5, 5, 246, 191, 5, 5, 186, 216, 5, 5, 76, 171, 5, 5, 102, 240, 5, 5, 50, 200, 5, 5, 75, 211, 5, 5, 177, 234, 5, 5, 83, 194, 5, 5, 84, 194, 5, 5, 144, 215, 5, 5, 232, 172, 5, 5, 64, 247, 5, 5, 226, 208, 5, 5, 60, 202, 5, 5, 180, 201, 5, 5, 89, 192, 5, 5, 225, 249, 5, 5, 228, 199, 5, 5, 183, 220, 5, 5, 197, 212, 5, 5, 158, 200, 5, 5, 152, 180, 5, 5, 56, 223, 5, 5, 92, 176, 5, 5, 28, 206, 5, 5, 17, 171, 5, 5, 146, 169, 5, 5, 145, 205, 5, 5, 158, 232, 5, 5, 165, 199, 5, 5, 7, 205, 5, 5, 207, 181, 5, 5, 113, 205, 5, 5, 237, 233, 5, 5, 85, 194, 5, 5, 35, 201, 5, 5, 235, 171, 5, 5, 200, 193, 5, 5, 225, 237, 5, 5, 60, 173, 5, 5, 187, 232, 5, 5, 86, 194, 5, 5, 45, 231, 5, 5, 245, 186, 5, 5, 132, 172, 5, 5, 46, 211, 5, 5, 212, 201, 5, 5, 253, 176, 5, 5, 172, 217, 5, 5, 49, 212, 5, 5, 135, 206, 5, 5, 217, 249, 5, 5, 53, 206, 5, 5, 11, 244, 5, 5, 48, 211, 5, 5, 20, 198, 5, 5, 153, 225, 5, 5, 4, 248, 5, 5, 254, 200, 5, 5, 168, 195, 5, 5, 140, 184, 5, 5, 183, 238, 5, 5, 173, 245, 5, 5, 233, 179, 5, 5, 69, 245, 5, 5, 207, 213, 5, 5, 159, 203, 5, 5, 135, 227, 5, 5, 48, 215, 5, 5, 230, 175, 5, 5, 84, 220, 5, 5, 141, 219, 5, 5, 82, 182, 5, 5, 155, 179, 5, 5, 224, 207, 5, 5, 149, 211, 5, 5, 50, 237, 5, 5, 244, 172, 5, 5, 33, 234, 5, 5, 174, 184, 5, 5, 21, 172, 5, 5, 4, 225, 5, 5, 77, 197, 5, 5, 14, 181, 5, 5, 178, 173, 5, 5, 80, 247, 5, 5, 238, 209, 5, 5, 172, 169, 5, 5, 12, 185, 5, 5, 22, 169, 5, 5, 99, 215, 5, 5, 203, 237, 5, 5, 97, 216, 5, 5, 28, 198, 5, 5, 35, 230, 5, 5, 251, 185, 5, 5, 185, 198, 5, 5, 145, 182, 5, 5, 203, 191, 5, 5, 70, 227, 5, 5, 95, 211, 5, 5, 205, 168, 5, 5, 30, 242, 5, 5, 102, 215, 5, 5, 33, 246, 5, 5, 103, 215, 5, 5, 88, 186, 5, 5, 152, 171, 5, 5, 215, 198, 5, 5, 226, 213, 5, 5, 249, 171, 5, 5, 190, 171, 5, 5, 22, 243, 5, 5, 215, 249, 5, 5, 202, 209, 5, 5, 63, 225, 5, 5, 106, 193, 5, 5, 109, 179, 5, 5, 41, 241, 5, 5, 186, 226, 5, 5, 191, 171, 5, 5, 62, 204, 5, 5, 44, 223, 5, 5, 56, 215, 5, 5, 104, 207, 5, 5, 129, 169, 5, 5, 45, 222, 5, 5, 71, 235, 5, 5, 180, 224, 5, 5, 73, 170, 5, 5, 133, 210, 5, 5, 222, 212, 5, 5, 39, 181, 5, 5, 37, 234, 5, 5, 197, 198, 5, 5, 117, 201, 5, 5, 253, 181, 5, 5, 59, 207, 5, 5, 239, 235, 5, 5, 76, 241, 5, 5, 112, 174, 5, 5, 86, 193, 5, 5, 237, 237, 5, 5, 69, 210, 5, 5, 251, 246, 5, 5, 54, 178, 5, 5, 21, 227, 5, 5, 247, 205, 5, 5, 161, 179, 5, 5, 186, 239, 5, 5, 101, 198, 5, 5, 147, 244, 5, 5, 214, 247, 5, 5, 79, 231, 5, 5, 3, 180, 5, 5, 17, 180, 5, 5, 243, 231, 5, 5, 39, 190, 5, 5, 224, 217, 5, 5, 31, 220, 5, 5, 202, 187, 5, 5, 162, 179, 5, 5, 157, 185, 5, 5, 38, 209, 5, 5, 40, 246, 5, 5, 239, 219, 5, 5, 211, 194, 5, 5, 76, 238, 5, 5, 73, 173, 5, 5, 42, 175, 5, 5, 87, 225, 5, 5, 67, 168, 5, 5, 234, 239, 5, 5, 21, 219, 5, 5, 56, 248, 5, 5, 92, 201, 5, 5, 54, 214, 5, 5, 67, 230, 5, 5, 182, 200, 5, 5, 215, 225, 5, 5, 60, 203, 5, 5, 45, 199, 5, 5, 14, 233, 5, 5, 157, 224, 5, 5, 79, 202, 5, 5, 116, 244, 5, 5, 117, 175, 5, 5, 155, 187, 5, 5, 79, 185, 5, 5, 113, 181, 5, 5, 125, 234, 5, 5, 104, 177, 5, 5, 64, 206, 5, 5, 94, 183, 5, 5, 228, 233, 5, 5, 189, 204, 5, 5, 93, 198, 5, 5, 227, 193, 5, 5, 95, 219, 5, 5, 245, 225, 5, 5, 54, 201, 5, 5, 153, 216, 5, 5, 241, 196, 5, 5, 0, 129, 5, 5, 193, 215, 5, 5, 155, 245, 5, 5, 180, 211, 5, 5, 106, 177, 5, 5, 87, 207, 5, 5, 213, 169, 5, 5, 155, 214, 5, 5, 164, 178, 5, 5, 14, 199, 5, 5, 67, 200, 5, 5, 118, 173, 5, 5, 237, 230, 5, 5, 137, 186, 5, 5, 116, 206, 5, 5, 20, 176, 5, 5, 145, 248, 5, 5, 202, 195, 5, 5, 17, 247, 5, 5, 246, 231, 5, 5, 148, 197, 5, 5, 189, 191, 5, 5, 115, 212, 5, 5, 36, 225, 5, 5, 7, 180, 5, 5, 45, 194, 5, 5, 100, 197, 5, 5, 28, 208, 5, 5, 34, 249, 5, 5, 85, 216, 5, 5, 65, 240, 5, 5, 113, 198, 5, 5, 200, 215, 5, 5, 86, 223, 5, 5, 183, 181, 5, 5, 210, 172, 5, 5, 250, 215, 5, 5, 87, 247, 5, 5, 200, 180, 5, 5, 203, 223, 5, 5, 210, 190, 5, 5, 125, 190, 5, 5, 0, 142, 5, 5, 176, 180, 5, 5, 195, 207, 5, 5, 67, 205, 5, 5, 38, 249, 5, 5, 239, 182, 5, 5, 173, 207, 5, 5, 134, 209, 5, 5, 13, 246, 5, 5, 217, 171, 5, 5, 137, 185, 5, 5, 19, 211, 5, 5, 146, 225, 5, 5, 137, 197, 5, 5, 20, 211, 5, 5, 225, 209, 5, 5, 158, 204, 5, 5, 66, 185, 5, 5, 77, 169, 5, 5, 97, 239, 5, 5, 69, 193, 5, 5, 64, 171, 5, 5, 223, 218, 5, 5, 252, 170, 5, 5, 95, 225, 5, 5, 33, 207, 5, 5, 119, 206, 5, 5, 3, 170, 5, 5, 207, 240, 5, 5, 111, 196, 5, 5, 212, 178, 5, 5, 201, 249, 5, 5, 82, 235, 5, 5, 164, 204, 5, 5, 42, 173, 5, 5, 246, 197, 5, 5, 176, 203, 5, 5, 33, 214, 5, 5, 135, 205, 5, 5, 234, 180, 5, 5, 231, 200, 5, 5, 50, 214, 5, 5, 249, 176, 5, 5, 205, 215, 5, 5, 196, 178, 5, 5, 234, 191, 5, 5, 168, 201, 5, 5, 135, 190, 5, 5, 242, 248, 5, 5, 27, 203, 5, 5, 139, 239, 5, 5, 231, 172, 5, 5, 11, 171, 5, 5, 155, 200, 5, 5, 57, 173, 5, 5, 185, 232, 5, 5, 41, 245, 5, 5, 129, 173, 5, 5, 182, 196, 5, 5, 135, 168, 5, 5, 204, 179, 5, 5, 121, 217, 5, 5, 166, 203, 5, 5, 68, 208, 5, 5, 40, 245, 5, 5, 225, 207, 5, 5, 199, 206, 5, 5, 241, 222, 5, 5, 114, 220, 5, 5, 216, 191, 5, 5, 226, 207, 5, 5, 105, 236, 5, 5, 80, 170, 5, 5, 41, 188, 5, 5, 81, 170, 5, 5, 95, 220, 5, 5, 94, 198, 5, 5, 187, 198, 5, 5, 238, 169, 5, 5, 79, 189, 5, 5, 106, 219, 5, 5, 195, 219, 5, 5, 39, 232, 5, 5, 40, 232, 5, 5, 155, 193, 5, 5, 15, 209, 5, 5, 163, 231, 5, 5, 27, 223, 5, 5, 104, 180, 5, 5, 150, 244, 5, 5, 99, 210, 5, 5, 211, 244, 5, 5, 156, 212, 5, 5, 209, 231, 5, 5, 87, 204, 5, 5, 217, 170, 5, 5, 110, 170, 5, 5, 107, 219, 5, 5, 15, 168, 5, 5, 196, 186, 5, 5, 134, 185, 5, 5, 135, 185, 5, 5, 42, 182, 5, 5, 181, 175, 5, 5, 99, 189, 5, 5, 49, 187, 5, 5, 19, 209, 5, 5, 232, 219, 5, 5, 95, 200, 5, 5, 77, 201, 5, 5, 226, 205, 5, 5, 230, 226, 5, 5, 59, 220, 5, 5, 21, 197, 5, 5, 32, 243, 5, 5, 33, 243, 5, 5, 84, 173, 5, 5, 191, 217, 5, 5, 31, 204, 5, 5, 120, 173, 5, 5, 70, 220, 5, 5, 96, 200, 5, 5, 74, 236, 5, 5, 182, 229, 5, 5, 121, 236, 5, 5, 195, 181, 5, 5, 162, 191, 5, 5, 101, 242, 5, 5, 183, 229, 5, 5, 106, 189, 5, 5, 94, 169, 5, 5, 104, 189, 5, 5, 193, 217, 5, 5, 196, 237, 5, 5, 185, 229, 5, 5, 28, 169, 5, 5, 86, 168, 5, 5, 18, 229, 5, 5, 209, 239, 5, 5, 116, 200, 5, 5, 43, 218, 5, 5, 232, 200, 5, 5, 120, 180, 5, 5, 243, 217, 5, 5, 105, 183, 5, 5, 109, 226, 5, 5, 30, 211, 5, 5, 217, 224, 5, 5, 116, 198, 5, 5, 210, 188, 5, 5, 45, 218, 5, 5, 251, 175, 5, 5, 200, 186, 5, 5, 109, 180, 5, 5, 221, 214, 5, 5, 239, 199, 5, 5, 136, 225, 5, 5, 202, 186, 5, 5, 107, 213, 5, 5, 158, 198, 5, 5, 154, 231, 5, 5, 143, 191, 5, 5, 93, 173, 5, 5, 60, 213, 5, 5, 103, 178, 5, 5, 141, 190, 5, 5, 220, 224, 5, 5, 187, 229, 5, 5, 186, 224, 5, 5, 221, 207, 5, 5, 239, 222, 5, 5, 110, 220, 5, 5, 83, 236, 5, 5, 28, 188, 5, 5, 47, 170, 5, 5, 200, 229, 5, 5, 247, 175, 5, 5, 104, 219, 5, 5, 241, 228, 5, 5, 19, 232, 5, 5, 53, 189, 5, 5, 133, 193, 5, 5, 3, 209, 5, 5, 161, 198, 5, 5, 233, 207, 5, 5, 145, 244, 5, 5, 96, 210, 5, 5, 32, 187, 5, 5, 167, 229, 5, 5, 198, 224, 5, 5, 8, 209, 5, 5, 224, 205, 5, 5, 100, 198, 5, 5, 34, 182, 5, 5, 116, 185, 5, 5, 171, 188, 5, 5, 85, 200, 5, 5, 158, 196, 5, 5, 23, 243, 5, 5, 43, 201, 5, 5, 116, 180, 5, 5, 43, 242, 5, 5, 63, 236, 5, 5, 113, 173, 5, 5, 125, 231, 5, 5, 172, 229, 5, 5, 156, 191, 5, 5, 106, 236, 5, 5, 163, 181, 5, 5, 74, 173, 5, 5, 217, 200, 5, 5, 188, 217, 5, 5, 91, 191, 5, 5, 110, 200, 5, 5, 29, 218, 5, 5, 199, 188, 5, 5, 102, 226, 5, 5, 142, 177, 5, 5, 18, 218, 5, 5, 22, 188, 5, 5, 131, 190, 5, 5, 83, 184, 5, 5, 73, 184, 5, 5, 138, 201, 5, 5, 217, 180, 5, 5, 243, 233, 5, 5, 31, 215, 5, 5, 118, 230, 5, 5, 172, 184, 5, 5, 196, 246, 5, 5, 169, 239, 5, 5, 40, 237, 5, 5, 84, 180, 5, 5, 137, 181, 5, 5, 115, 221, 5, 5, 15, 169, 5, 5, 110, 212, 5, 5, 37, 199, 5, 5, 97, 243, 5, 5, 67, 219, 5, 5, 165, 246, 5, 5, 246, 212, 5, 5, 52, 196, 5, 5, 151, 182, 5, 5, 41, 237, 5, 5, 170, 239, 5, 5, 245, 245, 5, 5, 31, 195, 5, 5, 52, 174, 5, 5, 141, 181, 5, 5, 96, 217, 5, 5, 254, 178, 5, 5, 204, 249, 5, 5, 35, 250, 5, 5, 103, 179, 5, 5, 118, 204, 5, 5, 0, 128, 5, 5, 246, 227, 5, 5, 247, 227, 5, 5, 147, 169, 5, 5, 101, 171, 5, 5, 79, 184, 5, 5, 204, 192, 5, 5, 148, 205, 5, 5, 11, 205, 5, 5, 255, 173, 5, 5, 27, 234, 5, 5, 72, 181, 5, 5, 63, 205, 5, 5, 101, 209, 5, 5, 134, 194, 5, 5, 126, 222, 5, 5, 73, 181, 5, 5, 112, 232, 5, 5, 224, 181, 5, 5, 75, 187, 5, 5, 119, 241, 5, 5, 151, 184, 5, 5, 111, 220, 5, 5, 50, 198, 5, 5, 45, 216, 5, 5, 34, 234, 5, 5, 43, 176, 5, 5, 93, 184, 5, 5, 66, 170, 5, 5, 177, 231, 5, 5, 113, 220, 5, 5, 129, 221, 5, 5, 222, 246, 5, 5, 10, 214, 5, 5, 87, 180, 5, 5, 239, 242, 5, 5, 250, 245, 5, 5, 111, 175, 5, 5, 116, 232, 5, 5, 117, 178, 5, 5, 253, 210, 5, 5, 101, 243, 5, 5, 71, 232, 5, 5, 111, 212, 5, 5, 233, 182, 5, 5, 166, 249, 5, 5, 30, 239, 5, 5, 225, 180, 5, 5, 80, 206, 5, 5, 182, 239, 5, 5, 228, 196, 5, 5, 12, 176, 5, 5, 7, 174, 5, 5, 4, 213, 5, 5, 63, 204, 5, 5, 188, 225, 5, 5, 124, 232, 5, 5, 157, 205, 5, 5, 131, 221, 5, 5, 77, 232, 5, 5, 187, 239, 5, 5, 10, 248, 5, 5, 59, 237, 5, 5, 184, 218, 5, 5, 186, 231, 5, 5, 33, 175, 5, 5, 217, 236, 5, 5, 188, 239, 5, 5, 209, 222, 5, 5, 208, 178, 5, 5, 160, 241, 5, 5, 227, 204, 5, 5, 207, 228, 5, 5, 208, 228, 5, 5, 7, 211, 5, 5, 226, 180, 5, 5, 66, 224, 5, 5, 68, 168, 5, 5, 192, 190, 5, 5, 39, 195, 5, 5, 67, 224, 5, 5, 7, 240, 5, 5, 180, 184, 5, 5, 166, 236, 5, 5, 20, 191, 5, 5, 126, 185, 5, 5, 106, 243, 5, 5, 144, 230, 5, 5, 246, 199, 5, 5, 6, 168, 5, 5, 130, 231, 5, 5, 113, 225, 5, 5, 242, 192, 5, 5, 59, 245, 5, 5, 123, 178, 5, 5, 105, 168, 5, 5, 136, 231, 5, 5, 11, 240, 5, 5, 147, 219, 5, 5, 112, 218, 5, 5, 235, 203, 5, 5, 172, 227, 5, 5, 191, 178, 5, 5, 232, 194, 5, 5, 88, 224, 5, 5, 69, 224, 5, 5, 151, 168, 5, 5, 87, 232, 5, 5, 34, 183, 5, 5, 167, 210, 5, 5, 237, 181, 5, 5, 62, 192, 5, 5, 128, 178, 5, 5, 63, 231, 5, 5, 20, 240, 5, 5, 41, 247, 5, 5, 36, 191, 5, 5, 17, 205, 5, 5, 34, 231, 5, 5, 116, 202, 5, 5, 179, 202, 5, 5, 186, 202, 5, 5, 142, 248, 5, 5, 177, 189, 5, 5, 105, 246, 5, 5, 124, 224, 5, 5, 159, 197, 5, 5, 31, 210, 5, 5, 49, 239, 5, 5, 26, 218, 5, 5, 226, 236, 5, 5, 127, 216, 5, 5, 64, 215, 5, 5, 79, 234, 5, 5, 19, 236, 5, 5, 133, 192, 5, 5, 134, 186, 5, 5, 120, 190, 5, 5, 18, 247, 5, 5, 37, 186, 5, 5, 166, 197, 5, 5, 118, 176, 5, 5, 195, 240, 5, 5, 243, 196, 5, 5, 124, 175, 5, 5, 164, 189, 5, 5, 166, 243, 5, 5, 163, 206, 5, 5, 199, 241, 5, 5, 74, 175, 5, 5, 143, 179, 5, 5, 64, 224, 5, 5, 63, 188, 5, 5, 19, 230, 5, 5, 247, 222, 5, 5, 235, 186, 5, 5, 136, 175, 5, 5, 169, 243, 5, 5, 210, 240, 5, 5, 73, 192, 5, 5, 27, 231, 5, 5, 81, 175, 5, 5, 208, 204, 5, 5, 44, 202, 5, 5, 37, 210, 5, 5, 115, 235, 5, 5, 172, 243, 5, 5, 170, 192, 5, 5, 171, 192, 5, 5, 227, 241, 5, 5, 161, 241, 5, 5, 117, 235, 5, 5, 229, 209, 5, 5, 160, 183, 5, 5, 240, 219, 5, 5, 204, 209, 5, 5, 173, 229, 5, 5, 14, 183, 5, 5, 53, 212, 5, 5, 83, 243, 5, 5, 158, 202, 5, 5, 249, 174, 5, 5, 35, 237, 5, 5, 88, 204, 5, 5, 159, 201, 5, 5, 61, 168, 5, 5, 98, 169, 5, 5, 151, 230, 5, 5, 117, 179, 5, 5, 89, 178, 5, 5, 255, 189, 5, 5, 152, 230, 5, 5, 167, 179, 5, 5, 164, 234, 5, 5, 252, 192, 5, 5, 215, 209, 5, 5, 21, 174, 5, 5, 26, 233, 5, 5, 180, 244, 5, 5, 231, 226, 5, 5, 14, 246, 5, 5, 135, 212, 5, 5, 176, 207, 5, 5, 92, 204, 5, 5, 97, 214, 5, 5, 152, 247, 5, 5, 229, 191, 5, 5, 60, 220, 5, 5, 194, 182, 5, 5, 90, 189, 5, 5, 217, 244, 5, 5, 206, 239, 5, 5, 147, 210, 5, 5, 67, 248, 5, 5, 215, 227, 5, 5, 245, 203, 5, 5, 97, 204, 5, 5, 74, 180, 5, 5, 79, 240, 5, 5, 159, 230, 5, 5, 103, 204, 5, 5, 246, 222, 5, 5, 104, 204, 5, 5, 204, 231, 5, 5, 92, 189, 5, 5, 210, 239, 5, 5, 239, 206, 5, 5, 240, 206, 5, 5, 112, 243, 5, 5, 206, 204, 5, 5, 41, 208, 5, 5, 74, 171, 5, 5, 161, 230, 5, 5, 131, 229, 5, 5, 120, 200, 5, 5, 250, 239, 5, 5, 67, 231, 5, 5, 143, 239, 5, 5, 0, 165, 5, 5, 49, 178, 5, 5, 242, 225, 5, 5, 190, 219, 5, 5, 123, 242, 5, 5, 217, 205, 5, 5, 244, 204, 5, 5, 209, 177, 5, 5, 166, 192, 5, 5, 56, 212, 5, 5, 77, 237, 5, 5, 221, 168, 5, 5, 244, 212, 5, 5, 23, 229, 5, 5, 113, 183, 5, 5, 155, 231, 5, 5, 32, 193, 5, 5, 219, 206, 5, 5, 19, 208, 5, 5, 20, 208, 5, 5, 181, 201, 5, 5, 229, 202, 5, 5, 247, 191, 5, 5, 0, 166, 5, 5, 248, 182, 5, 5, 212, 177, 5, 5, 8, 205, 5, 5, 113, 204, 5, 5, 33, 168, 5, 5, 115, 183, 5, 5, 232, 202, 5, 5, 123, 169, 5, 5, 241, 188, 5, 5, 242, 188, 5, 5, 244, 188, 5, 5, 139, 171, 5, 5, 169, 216, 5, 5, 170, 216, 5, 5, 101, 196, 5, 5, 82, 226, 5, 5, 204, 245, 5, 5, 103, 196, 5, 5, 69, 174, 5, 5, 218, 216, 5, 5, 108, 196, 5, 5, 109, 196, 5, 5, 123, 179, 5, 5, 112, 196, 5, 5, 84, 226, 5, 5, 179, 182, 5, 5, 190, 182, 5, 5, 201, 198, 5, 5, 64, 208, 5, 5, 175, 208, 5, 5, 174, 208, 5, 5, 136, 169, 5, 5, 35, 238, 5, 5, 130, 226, 5, 5, 146, 191, 5, 5, 53, 238, 5, 5, 117, 185, 5, 5, 238, 179, 5, 5, 223, 172, 5, 5, 110, 215, 5, 5, 174, 218, 5, 5, 228, 178, 5, 5, 175, 180, 5, 5, 145, 230, 5, 5, 175, 218, 5, 5, 132, 216, 5, 5, 241, 195, 5, 5, 93, 235, 5, 5, 99, 211, 5, 5, 207, 168, 5, 5, 12, 240, 5, 5, 135, 219, 5, 5, 248, 209, 5, 5, 128, 209, 5, 5, 213, 249, 5, 5, 143, 177, 5, 5, 43, 169, 5, 5, 86, 239, 5, 5, 211, 237, 5, 5, 198, 225, 5, 5, 198, 169, 5, 5, 53, 195, 5, 5, 85, 189, 5, 5, 89, 212, 5, 5, 36, 237, 5, 5, 120, 171, 5, 5, 204, 239, 5, 5, 138, 185, 5, 5, 200, 225, 5, 5, 116, 195, 5, 5, 133, 233, 5, 5, 50, 168, 5, 5, 51, 168, 5, 5, 251, 188, 5, 5, 240, 185, 5, 5, 19, 215, 5, 5, 145, 177, 5, 5, 27, 216, 5, 5, 210, 226, 5, 5, 253, 206, 5, 5, 182, 240, 5, 5, 63, 224, 5, 5, 157, 226, 5, 5, 56, 216, 5, 5, 83, 235, 5, 5, 66, 199, 5, 5, 251, 169, 5, 5, 215, 224, 5, 5, 191, 220, 5, 5, 123, 171, 5, 5, 117, 200, 5, 5, 252, 196, 5, 5, 160, 200, 5, 5, 202, 233, 5, 5, 5, 219, 5, 5, 58, 169, 5, 5, 218, 181, 5, 5, 27, 217, 5, 5, 28, 217, 5, 5, 211, 188, 5, 5, 60, 216, 5, 5, 218, 209, 5, 5, 253, 196, 5, 5, 180, 193, 5, 5, 182, 170, 5, 5, 215, 178, 5, 5, 185, 193, 5, 5, 157, 229, 5, 5, 179, 225, 5, 5, 7, 186, 5, 5, 218, 224, 5, 5, 204, 169, 5, 5, 138, 233, 5, 5, 30, 212, 5, 5, 147, 185, 5, 5, 153, 170, 5, 5, 121, 200, 5, 5, 99, 225, 5, 5, 53, 205, 5, 5, 123, 187, 5, 5, 37, 216, 5, 5, 97, 235, 5, 5, 82, 192, 5, 5, 196, 193, 5, 5, 114, 194, 5, 5, 104, 173, 5, 5, 151, 177, 5, 5, 16, 190, 5, 5, 101, 232, 5, 5, 92, 215, 5, 5, 153, 180, 5, 5, 73, 228, 5, 5, 201, 193, 5, 5, 255, 200, 5, 5, 245, 228, 5, 5, 176, 218, 5, 5, 23, 184, 5, 5, 216, 207, 5, 5, 13, 218, 5, 5, 139, 185, 5, 5, 74, 229, 5, 5, 57, 216, 5, 5, 15, 188, 5, 5, 172, 173, 5, 5, 118, 200, 5, 5, 11, 219, 5, 5, 141, 243, 5, 5, 11, 221, 5, 5, 83, 229, 5, 5, 145, 185, 5, 5, 253, 168, 5, 5, 180, 225, 5, 5, 8, 186, 5, 5, 142, 243, 5, 5, 51, 185, 5, 5, 154, 170, 5, 5, 85, 229, 5, 5, 109, 224, 5, 5, 150, 180, 5, 5, 74, 228, 5, 5, 154, 180, 5, 5, 235, 228, 5, 5, 160, 218, 5, 5, 238, 183, 5, 5, 8, 188, 5, 5, 64, 229, 5, 5, 128, 243, 5, 5, 172, 225, 5, 5, 179, 196, 5, 5, 183, 196, 5, 5, 253, 231, 5, 5, 196, 233, 5, 5, 6, 232, 5, 5, 75, 192, 5, 5, 177, 239, 5, 5, 165, 243, 5, 5, 129, 243, 5, 5, 215, 220, 5, 5, 150, 201, 5, 5, 186, 212, 5, 5, 227, 190, 5, 5, 85, 240, 5, 5, 144, 243, 5, 5, 193, 212, 5, 5, 57, 168, 5, 5, 211, 239, 5, 5, 214, 232, 5, 5, 17, 190, 5, 5, 8, 238, 5, 5, 241, 179, 5, 5, 207, 216, 5, 5, 247, 199, 5, 5, 228, 232, 5, 5, 31, 223, 5, 5, 235, 187, 5, 5, 158, 234, 5, 5, 154, 238, 5, 5, 161, 234, 5, 5, 128, 181, 5, 5, 184, 223, 5, 5, 20, 200, 5, 5, 156, 214, 5, 5, 84, 188, 5, 5, 73, 242, 5, 5, 112, 228, 5, 5, 8, 169, 5, 5, 19, 181, 5, 5, 237, 178, 5, 5, 234, 177, 5, 5, 45, 212, 5, 5, 214, 213, 5, 5, 122, 204, 5, 5, 231, 173, 5, 5, 64, 196, 5, 5, 222, 201, 5, 5, 204, 188, 5, 5, 252, 215, 5, 5, 188, 181, 5, 5, 118, 181, 5, 5, 75, 229, 5, 5, 134, 233, 5, 5, 112, 200, 5, 5, 85, 221, 5, 5, 166, 238, 5, 5, 167, 238, 5, 5, 210, 187, 5, 5, 216, 180, 5, 5, 253, 241, 5, 5, 165, 213, 5, 5, 223, 201, 5, 5, 69, 184, 5, 5, 95, 193, 5, 5, 68, 227, 5, 5, 125, 191, 5, 5, 6, 200, 5, 5, 96, 193, 5, 5, 106, 206, 5, 5, 249, 226, 5, 5, 76, 174, 5, 5, 171, 240, 5, 5, 61, 243, 5, 5, 204, 189, 5, 5, 59, 188, 5, 5, 72, 196, 5, 5, 174, 227, 5, 5, 175, 227, 5, 5, 143, 213, 5, 5, 197, 200, 5, 5, 176, 227, 5, 5, 53, 249, 5, 5, 52, 249, 5, 5, 24, 176, 5, 5, 14, 180, 5, 5, 198, 200, 5, 5, 51, 225, 5, 5, 60, 188, 5, 5, 79, 215, 5, 5, 236, 198, 5, 5, 55, 177, 5, 5, 61, 235, 5, 5, 107, 216, 5, 5, 171, 238, 5, 5, 162, 219, 5, 5, 30, 226, 5, 5, 121, 181, 5, 5, 200, 181, 5, 5, 123, 236, 5, 5, 208, 229, 5, 5, 149, 198, 5, 5, 222, 240, 5, 5, 93, 248, 5, 5, 124, 236, 5, 5, 94, 232, 5, 5, 254, 233, 5, 5, 215, 238, 5, 5, 252, 242, 5, 5, 207, 219, 5, 5, 79, 179, 5, 5, 80, 179, 5, 5, 129, 194, 5, 5, 238, 199, 5, 5, 227, 201, 5, 5, 104, 201, 5, 5, 114, 213, 5, 5, 83, 228, 5, 5, 4, 207, 5, 5, 245, 176, 5, 5, 174, 237, 5, 5, 144, 188, 5, 5, 38, 216, 5, 5, 142, 186, 5, 5, 223, 235, 5, 5, 189, 236, 5, 5, 148, 191, 5, 5, 105, 173, 5, 5, 42, 219, 5, 5, 22, 208, 5, 5, 77, 171, 5, 5, 102, 232, 5, 5, 150, 213, 5, 5, 114, 205, 5, 5, 24, 201, 5, 5, 47, 211, 5, 5, 254, 217, 5, 5, 254, 237, 5, 5, 240, 179, 5, 5, 204, 216, 5, 5, 231, 187, 5, 5, 221, 232, 5, 5, 29, 223, 5, 5, 145, 234, 5, 5, 148, 234, 5, 5, 108, 228, 5, 5, 126, 186, 5, 5, 15, 181, 5, 5, 129, 214, 5, 5, 4, 169, 5, 5, 182, 223, 5, 5, 80, 188, 5, 5, 31, 242, 5, 5, 88, 205, 5, 5, 121, 204, 5, 5, 228, 213, 5, 5, 61, 196, 5, 5, 94, 195, 5, 5, 90, 193, 5, 5, 247, 226, 5, 5, 191, 188, 5, 5, 163, 240, 5, 5, 140, 196, 5, 5, 220, 198, 5, 5, 155, 238, 5, 5, 142, 213, 5, 5, 99, 191, 5, 5, 171, 227, 5, 5, 46, 188, 5, 5, 164, 240, 5, 5, 165, 240, 5, 5, 229, 198, 5, 5, 21, 226, 5, 5, 221, 240, 5, 5, 189, 181, 5, 5, 89, 248, 5, 5, 118, 236, 5, 5, 119, 181, 5, 5, 25, 211, 5, 5, 242, 206, 5, 5, 76, 179, 5, 5, 205, 219, 5, 5, 142, 188, 5, 5, 226, 201, 5, 5, 101, 173, 5, 5, 40, 219, 5, 5, 149, 213, 5, 5, 253, 217, 5, 5, 84, 183, 5, 5, 239, 170, 5, 5, 151, 186, 5, 5, 18, 184, 5, 5, 165, 231, 5, 5, 243, 244, 5, 5, 245, 170, 5, 5, 149, 219, 5, 5, 200, 168, 5, 5, 242, 224, 5, 5, 207, 203, 5, 5, 157, 186, 5, 5, 84, 235, 5, 5, 207, 220, 5, 5, 119, 197, 5, 5, 2, 171, 5, 5, 100, 223, 5, 5, 84, 229, 5, 5, 74, 237, 5, 5, 171, 237, 5, 5, 209, 223, 5, 5, 117, 198, 5, 5, 210, 223, 5, 5, 70, 182, 5, 5, 176, 204, 5, 5, 69, 231, 5, 5, 209, 204, 5, 5, 101, 213, 5, 5, 102, 213, 5, 5, 179, 204, 5, 5, 12, 171, 5, 5, 13, 171, 5, 5, 14, 171, 5, 5, 156, 203, 5, 5, 15, 171, 5, 5, 9, 220, 5, 5, 116, 183, 5, 5, 111, 234, 5, 5, 71, 183, 5, 5, 45, 237, 5, 5, 232, 244, 5, 5, 232, 170, 5, 5, 142, 219, 5, 5, 97, 197, 5, 5, 78, 223, 5, 5, 201, 223, 5, 5, 158, 237, 5, 5, 153, 204, 5, 5, 96, 213, 5, 5, 253, 170, 5, 5, 254, 170, 5, 5, 180, 182, 5, 5, 72, 182, 5, 5, 200, 182, 5, 5, 177, 182, 5, 5, 222, 221, 5, 5, 215, 221, 5, 5, 75, 172, 5, 5, 20, 192, 5, 5, 254, 179, 5, 5, 129, 223, 5, 5, 241, 227, 5, 5, 206, 244, 5, 5, 114, 224, 5, 5, 205, 232, 5, 5, 212, 227, 5, 5, 177, 218, 5, 5, 60, 242, 5, 5, 169, 197, 5, 5, 243, 174, 5, 5, 13, 240, 5, 5, 132, 182, 5, 5, 133, 182, 5, 5, 118, 224, 5, 5, 10, 240, 5, 5, 73, 227, 5, 5, 156, 238, 5, 5, 49, 250, 5, 5, 111, 170, 5, 5, 133, 195, 5, 5, 173, 225, 5, 5, 112, 169, 5, 5, 122, 176, 5, 5, 196, 226, 5, 5, 140, 223, 5, 5, 113, 169, 5, 5, 68, 222, 5, 5, 85, 181, 5, 5, 225, 187, 5, 5, 179, 218, 5, 5, 125, 226, 5, 5, 226, 194, 5, 5, 93, 193, 5, 5, 121, 171, 5, 5, 159, 237, 5, 5, 33, 227, 5, 5, 123, 176, 5, 5, 208, 232, 5, 5, 95, 237, 5, 5, 172, 197, 5, 5, 255, 181, 5, 5, 239, 236, 5, 5, 157, 201, 5, 5, 247, 230, 5, 5, 82, 172, 5, 5, 166, 171, 5, 5, 112, 210, 5, 5, 190, 181, 5, 5, 35, 172, 5, 5, 69, 198, 5, 5, 107, 180, 5, 5, 27, 224, 5, 5, 200, 241, 5, 5, 75, 222, 5, 5, 165, 237, 5, 5, 171, 191, 5, 5, 196, 187, 5, 5, 76, 222, 5, 5, 67, 194, 5, 5, 168, 248, 5, 5, 122, 171, 5, 5, 149, 232, 5, 5, 37, 172, 5, 5, 40, 238, 5, 5, 48, 225, 5, 5, 196, 182, 5, 5, 34, 245, 5, 5, 122, 229, 5, 5, 215, 186, 5, 5, 196, 181, 5, 5, 162, 211, 5, 5, 139, 243, 5, 5, 207, 189, 5, 5, 224, 190, 5, 5, 197, 226, 5, 5, 131, 191, 5, 5, 181, 193, 5, 5, 123, 189, 5, 5, 25, 168, 5, 5, 133, 225, 5, 5, 16, 183, 5, 5, 126, 229, 5, 5, 118, 186, 5, 5, 214, 172, 5, 5, 187, 223, 5, 5, 135, 225, 5, 5, 227, 171, 5, 5, 49, 185, 5, 5, 150, 231, 5, 5, 44, 200, 5, 5, 207, 204, 5, 5, 208, 223, 5, 5, 205, 225, 5, 5, 50, 238, 5, 5, 155, 229, 5, 5, 81, 209, 5, 5, 137, 225, 5, 5, 2, 207, 5, 5, 149, 170, 5, 5, 119, 242, 5, 5, 109, 234, 5, 5, 215, 195, 5, 5, 191, 219, 5, 5, 49, 200, 5, 5, 115, 248, 5, 5, 155, 220, 5, 5, 7, 175, 5, 5, 50, 178, 5, 5, 154, 239, 5, 5, 79, 192, 5, 5, 110, 218, 5, 5, 85, 174, 5, 5, 215, 240, 5, 5, 181, 225, 5, 5, 134, 229, 5, 5, 215, 232, 5, 5, 224, 244, 5, 5, 17, 183, 5, 5, 228, 187, 5, 5, 21, 208, 5, 5, 12, 237, 5, 5, 88, 209, 5, 5, 61, 173, 5, 5, 216, 232, 5, 5, 136, 206, 5, 5, 236, 243, 5, 5, 56, 210, 5, 5, 205, 221, 5, 5, 250, 179, 5, 5, 254, 191, 5, 5, 226, 227, 5, 5, 101, 225, 5, 5, 208, 227, 5, 5, 110, 231, 5, 5, 161, 218, 5, 5, 18, 242, 5, 5, 229, 174, 5, 5, 120, 182, 5, 5, 4, 240, 5, 5, 15, 194, 5, 5, 25, 222, 5, 5, 102, 169, 5, 5, 121, 223, 5, 5, 74, 181, 5, 5, 116, 238, 5, 5, 249, 181, 5, 5, 104, 218, 5, 5, 203, 232, 5, 5, 172, 188, 5, 5, 156, 201, 5, 5, 203, 194, 5, 5, 190, 230, 5, 5, 110, 171, 5, 5, 153, 171, 5, 5, 97, 180, 5, 5, 154, 181, 5, 5, 157, 241, 5, 5, 109, 210, 5, 5, 55, 198, 5, 5, 186, 187, 5, 5, 166, 191, 5, 5, 128, 232, 5, 5, 207, 186, 5, 5, 208, 172, 5, 5, 32, 200, 5, 5, 99, 186, 5, 5, 199, 223, 5, 5, 23, 173, 5, 5, 24, 238, 5, 5, 67, 209, 5, 5, 205, 171, 5, 5, 149, 204, 5, 5, 100, 234, 5, 5, 209, 195, 5, 5, 249, 206, 5, 5, 183, 219, 5, 5, 110, 248, 5, 5, 54, 210, 5, 5, 98, 222, 5, 5, 248, 199, 5, 5, 180, 187, 5, 5, 162, 232, 5, 5, 33, 183, 5, 5, 213, 171, 5, 5, 164, 210, 5, 5, 122, 170, 5, 5, 218, 171, 5, 5, 125, 227, 5, 5, 239, 187, 5, 5, 197, 182, 5, 5, 182, 193, 5, 5, 54, 168, 5, 5, 28, 168, 5, 5, 132, 184, 5, 5, 9, 232, 5, 5, 62, 243, 5, 5, 236, 233, 5, 5, 50, 183, 5, 5, 134, 213, 5, 5, 238, 233, 5, 5, 187, 206, 5, 5, 61, 242, 5, 5, 125, 183, 5, 5, 47, 188, 5, 5, 230, 178, 5, 5, 10, 228, 5, 5, 136, 245, 5, 5, 195, 174, 5, 5, 183, 235, 5, 5, 59, 248, 5, 5, 116, 246, 5, 5, 133, 212, 5, 5, 253, 233, 5, 5, 191, 218, 5, 5, 148, 219, 5, 5, 109, 243, 5, 5, 176, 229, 5, 5, 146, 246, 5, 5, 34, 178, 5, 5, 227, 205, 5, 5, 55, 241, 5, 5, 214, 171, 5, 5, 115, 169, 5, 5, 229, 197, 5, 5, 12, 228, 5, 5, 87, 239, 5, 5, 118, 217, 5, 5, 177, 229, 5, 5, 122, 217, 5, 5, 124, 196, 5, 5, 229, 213, 5, 5, 78, 245, 5, 5, 163, 242, 5, 5, 119, 212, 5, 5, 192, 247, 5, 5, 105, 197, 5, 5, 65, 248, 5, 5, 139, 211, 5, 5, 249, 196, 5, 5, 234, 212, 5, 5, 249, 243, 5, 5, 126, 193, 5, 5, 93, 204, 5, 5, 240, 245, 5, 5, 243, 224, 5, 5, 122, 184, 5, 5, 96, 237, 5, 5, 2, 222, 5, 5, 128, 170, 5, 5, 13, 228, 5, 5, 14, 228, 5, 5, 142, 223, 5, 5, 159, 204, 5, 5, 209, 206, 5, 5, 82, 213, 5, 5, 203, 225, 5, 5, 42, 247, 5, 5, 226, 218, 5, 5, 232, 225, 5, 5, 73, 180, 5, 5, 145, 235, 5, 5, 205, 217, 5, 5, 58, 221, 5, 5, 125, 196, 5, 5, 2, 182, 5, 5, 226, 187, 5, 5, 219, 171, 5, 5, 226, 247, 5, 5, 207, 239, 5, 5, 164, 206, 5, 5, 153, 247, 5, 5, 43, 178, 5, 5, 227, 187, 5, 5, 167, 204, 5, 5, 26, 197, 5, 5, 177, 223, 5, 5, 108, 216, 5, 5, 63, 207, 5, 5, 63, 201, 5, 5, 61, 188, 5, 5, 127, 227, 5, 5, 34, 235, 5, 5, 193, 227, 5, 5, 72, 198, 5, 5, 130, 181, 5, 5, 144, 174, 5, 5, 20, 234, 5, 5, 0, 155, 5, 5, 207, 205, 5, 5, 147, 248, 5, 5, 157, 247, 5, 5, 12, 221, 5, 5, 83, 213, 5, 5, 79, 200, 5, 5, 206, 225, 5, 5, 185, 200, 5, 5, 125, 249, 5, 5, 20, 199, 5, 5, 180, 214, 5, 5, 181, 214, 5, 5, 249, 236, 5, 5, 198, 182, 5, 5, 220, 219, 5, 5, 250, 236, 5, 5, 148, 185, 5, 5, 197, 237, 5, 5, 164, 230, 5, 5, 90, 213, 5, 5, 186, 176, 5, 5, 91, 213, 5, 5, 126, 215, 5, 5, 199, 182, 5, 5, 229, 190, 5, 5, 124, 215, 5, 5, 195, 191, 5, 5, 215, 241, 5, 5, 33, 226, 5, 5, 246, 217, 5, 5, 207, 231, 5, 5, 127, 249, 5, 5, 11, 200, 5, 5, 7, 219, 5, 5, 101, 223, 5, 5, 159, 186, 5, 5, 19, 228, 5, 5, 63, 187, 5, 5, 211, 223, 5, 5, 88, 215, 5, 5, 136, 174, 5, 5, 56, 247, 5, 5, 178, 204, 5, 5, 195, 212, 5, 5, 236, 225, 5, 5, 70, 231, 5, 5, 152, 172, 5, 5, 145, 180, 5, 5, 5, 237, 5, 5, 225, 242, 5, 5, 178, 249, 5, 5, 221, 219, 5, 5, 156, 220, 5, 5, 193, 214, 5, 5, 60, 247, 5, 5, 251, 222, 5, 5, 218, 205, 5, 5, 71, 231, 5, 5, 126, 206, 5, 5, 36, 245, 5, 5, 144, 209, 5, 5, 158, 168, 5, 5, 84, 172, 5, 5, 16, 171, 5, 5, 187, 176, 5, 5, 133, 217, 5, 5, 158, 170, 5, 5, 61, 247, 5, 5, 125, 242, 5, 5, 176, 234, 5, 5, 50, 190, 5, 5, 169, 171, 5, 5, 36, 224, 5, 5, 49, 233, 5, 5, 3, 204, 5, 5, 17, 245, 5, 5, 2, 181, 5, 5, 182, 204, 5, 5, 21, 228, 5, 5, 196, 172, 5, 5, 126, 179, 5, 5, 192, 194, 5, 5, 194, 226, 5, 5, 13, 237, 5, 5, 132, 206, 5, 5, 226, 244, 5, 5, 48, 196, 5, 5, 161, 239, 5, 5, 54, 238, 5, 5, 243, 227, 5, 5, 135, 213, 5, 5, 199, 247, 5, 5, 18, 237, 5, 5, 9, 205, 5, 5, 233, 205, 5, 5, 241, 225, 5, 5, 188, 232, 5, 5, 36, 193, 5, 5, 3, 223, 5, 5, 35, 197, 5, 5, 74, 231, 5, 5, 108, 190, 5, 5, 64, 202, 5, 5, 18, 171, 5, 5, 183, 206, 5, 5, 11, 242, 5, 5, 244, 227, 5, 5, 163, 235, 5, 5, 179, 174, 5, 5, 97, 217, 5, 5, 187, 218, 5, 5, 175, 171, 5, 5, 223, 205, 5, 5, 248, 243, 5, 5, 249, 221, 5, 5, 122, 223, 5, 5, 84, 184, 5, 5, 221, 196, 5, 5, 163, 249, 5, 5, 34, 248, 5, 5, 254, 227, 5, 5, 137, 211, 5, 5, 119, 193, 5, 5, 5, 239, 5, 5, 179, 177, 5, 5, 246, 232, 5, 5, 200, 206, 5, 5, 178, 239, 5, 5, 164, 194, 5, 5, 35, 190, 5, 5, 152, 206, 5, 5, 221, 187, 5, 5, 71, 213, 5, 5, 233, 170, 5, 5, 254, 201, 5, 5, 141, 174, 5, 5, 213, 236, 5, 5, 6, 234, 5, 5, 99, 216, 5, 5, 56, 198, 5, 5, 142, 214, 5, 5, 143, 214, 5, 5, 10, 199, 5, 5, 144, 248, 5, 5, 100, 249, 5, 5, 252, 223, 5, 5, 76, 172, 5, 5, 88, 213, 5, 5, 7, 247, 5, 5, 252, 199, 5, 5, 214, 219, 5, 5, 146, 230, 5, 5, 150, 168, 5, 5, 150, 204, 5, 5, 69, 215, 5, 5, 133, 220, 5, 5, 246, 170, 5, 5, 118, 206, 5, 5, 169, 176, 5, 5, 85, 173, 5, 5, 195, 247, 5, 5, 25, 193, 5, 5, 249, 222, 5, 5, 181, 232, 5, 5, 83, 186, 5, 5, 48, 229, 5, 5, 55, 229, 5, 5, 56, 229, 5, 5, 178, 241, 5, 5, 246, 184, 5, 5, 62, 239, 5, 5, 90, 168, 5, 5, 66, 227, 5, 5, 159, 195, 5, 5, 137, 169, 5, 5, 199, 169, 5, 5, 99, 176, 5, 5, 42, 226, 5, 5, 15, 179, 5, 5, 103, 199, 5, 5, 211, 187, 5, 5, 253, 215, 5, 5, 120, 189, 5, 5, 134, 199, 5, 5, 140, 185, 5, 5, 185, 227, 5, 5, 198, 185, 5, 5, 80, 213, 5, 5, 144, 170, 5, 5, 237, 198, 5, 5, 28, 215, 5, 5, 211, 241, 5, 5, 73, 224, 5, 5, 52, 205, 5, 5, 230, 171, 5, 5, 42, 233, 5, 5, 64, 169, 5, 5, 231, 171, 5, 5, 100, 176, 5, 5, 164, 199, 5, 5, 86, 171, 5, 5, 85, 209, 5, 5, 154, 203, 5, 5, 54, 205, 5, 5, 46, 233, 5, 5, 151, 180, 5, 5, 246, 243, 5, 5, 74, 224, 5, 5, 45, 226, 5, 5, 88, 171, 5, 5, 166, 199, 5, 5, 115, 205, 5, 5, 37, 185, 5, 5, 39, 185, 5, 5, 52, 216, 5, 5, 189, 198, 5, 5, 42, 216, 5, 5, 105, 201, 5, 5, 231, 219, 5, 5, 234, 170, 5, 5, 64, 200, 5, 5, 68, 200, 5, 5, 231, 178, 5, 5, 172, 182, 5, 5, 103, 234, 5, 5, 67, 218, 5, 5, 88, 207, 5, 5, 35, 178, 5, 5, 73, 200, 5, 5, 81, 171, 5, 5, 52, 182, 5, 5, 158, 226, 5, 5, 235, 212, 5, 5, 47, 249, 5, 5, 54, 182, 5, 5, 69, 218, 5, 5, 68, 226, 5, 5, 139, 169, 5, 5, 129, 170, 5, 5, 92, 207, 5, 5, 27, 184, 5, 5, 238, 181, 5, 5, 227, 218, 5, 5, 124, 217, 5, 5, 228, 185, 5, 5, 104, 234, 5, 5, 113, 200, 5, 5, 9, 193, 5, 5, 187, 212, 5, 5, 74, 248, 5, 5, 208, 220, 5, 5, 148, 224, 5, 5, 71, 226, 5, 5, 187, 202, 5, 5, 83, 171, 5, 5, 126, 249, 5, 5, 245, 178, 5, 5, 190, 212, 5, 5, 166, 223, 5, 5, 54, 246, 5, 5, 245, 217, 5, 5, 128, 249, 5, 5, 36, 223, 5, 5, 253, 193, 5, 5, 32, 228, 5, 5, 210, 189, 5, 5, 153, 200, 5, 5, 167, 196, 5, 5, 192, 214, 5, 5, 30, 203, 5, 5, 242, 245, 5, 5, 87, 171, 5, 5, 196, 212, 5, 5, 211, 206, 5, 5, 177, 219, 5, 5, 5, 207, 5, 5, 6, 207, 5, 5, 15, 220, 5, 5, 178, 234, 5, 5, 215, 203, 5, 5, 90, 215, 5, 5, 89, 215, 5, 5, 53, 210, 5, 5, 142, 190, 5, 5, 122, 200, 5, 5, 76, 211, 5, 5, 89, 171, 5, 5, 216, 203, 5, 5, 91, 218, 5, 5, 98, 180, 5, 5, 106, 180, 5, 5, 100, 210, 5, 5, 105, 189, 5, 5, 151, 231, 5, 5, 111, 180, 5, 5, 24, 188, 5, 5, 113, 180, 5, 5, 112, 180, 5, 5, 168, 196, 5, 5, 164, 173, 5, 5, 134, 242, 5, 5, 135, 242, 5, 5, 120, 185, 5, 5, 245, 236, 5, 5, 70, 184, 5, 5, 145, 216, 5, 5, 62, 187, 5, 5, 129, 249, 5, 5, 210, 204, 5, 5, 65, 187, 5, 5, 176, 220, 5, 5, 126, 242, 5, 5, 82, 187, 5, 5, 213, 207, 5, 5, 243, 192, 5, 5, 157, 214, 5, 5, 147, 184, 5, 5, 3, 200, 5, 5, 170, 191, 5, 5, 201, 168, 5, 5, 243, 213, 5, 5, 217, 207, 5, 5, 165, 234, 5, 5, 174, 236, 5, 5, 30, 233, 5, 5, 100, 203, 5, 5, 108, 242, 5, 5, 177, 227, 5, 5, 99, 214, 5, 5, 193, 228, 5, 5, 101, 203, 5, 5, 127, 229, 5, 5, 216, 184, 5, 5, 172, 174, 5, 5, 103, 213, 5, 5, 155, 170, 5, 5, 86, 209, 5, 5, 191, 192, 5, 5, 175, 234, 5, 5, 98, 175, 5, 5, 194, 236, 5, 5, 227, 244, 5, 5, 162, 241, 5, 5, 113, 178, 5, 5, 130, 218, 5, 5, 105, 195, 5, 5, 185, 168, 5, 5, 86, 189, 5, 5, 242, 227, 5, 5, 168, 179, 5, 5, 186, 192, 5, 5, 248, 234, 5, 5, 120, 181, 5, 5, 184, 181, 5, 5, 39, 220, 5, 5, 85, 188, 5, 5, 214, 227, 5, 5, 135, 209, 5, 5, 163, 195, 5, 5, 86, 221, 5, 5, 27, 169, 5, 5, 214, 242, 5, 5, 19, 213, 5, 5, 131, 205, 5, 5, 178, 229, 5, 5, 200, 189, 5, 5, 91, 205, 5, 5, 157, 243, 5, 5, 157, 182, 5, 5, 39, 183, 5, 5, 249, 209, 5, 5, 58, 241, 5, 5, 89, 213, 5, 5, 73, 209, 5, 5, 208, 188, 5, 5, 201, 231, 5, 5, 125, 217, 5, 5, 14, 188, 5, 5, 239, 212, 5, 5, 98, 204, 5, 5, 16, 228, 5, 5, 167, 171, 5, 5, 83, 217, 5, 5, 203, 213, 5, 5, 32, 184, 5, 5, 139, 170, 5, 5, 137, 176, 5, 5, 123, 229, 5, 5, 250, 196, 5, 5, 144, 179, 5, 5, 218, 168, 5, 5, 60, 241, 5, 5, 140, 187, 5, 5, 240, 212, 5, 5, 200, 210, 5, 5, 17, 234, 5, 5, 245, 224, 5, 5, 141, 169, 5, 5, 126, 184, 5, 5, 164, 244, 5, 5, 110, 197, 5, 5, 72, 186, 5, 5, 238, 221, 5, 5, 39, 180, 5, 5, 213, 177, 5, 5, 216, 224, 5, 5, 122, 195, 5, 5, 155, 222, 5, 5, 139, 189, 5, 5, 212, 232, 5, 5, 239, 181, 5, 5, 87, 168, 5, 5, 19, 229, 5, 5, 107, 245, 5, 5, 227, 247, 5, 5, 43, 240, 5, 5, 209, 203, 5, 5, 165, 206, 5, 5, 45, 227, 5, 5, 43, 226, 5, 5, 133, 239, 5, 5, 138, 171, 5, 5, 80, 229, 5, 5, 185, 194, 5, 5, 104, 199, 5, 5, 58, 187, 5, 5, 7, 232, 5, 5, 143, 185, 5, 5, 57, 191, 5, 5, 122, 201, 5, 5, 35, 184, 5, 5, 202, 198, 5, 5, 106, 234, 5, 5, 49, 181, 5, 5, 44, 198, 5, 5, 31, 226, 5, 5, 173, 208, 5, 5, 209, 220, 5, 5, 165, 244, 5, 5, 149, 224, 5, 5, 109, 216, 5, 5, 212, 241, 5, 5, 113, 210, 5, 5, 147, 245, 5, 5, 141, 187, 5, 5, 199, 185, 5, 5, 231, 223, 5, 5, 74, 230, 5, 5, 85, 217, 5, 5, 107, 220, 5, 5, 15, 214, 5, 5, 179, 190, 5, 5, 160, 226, 5, 5, 88, 202, 5, 5, 42, 220, 5, 5, 43, 220, 5, 5, 203, 198, 5, 5, 14, 208, 5, 5, 133, 174, 5, 5, 45, 202, 5, 5, 183, 249, 5, 5, 28, 231, 5, 5, 254, 240, 5, 5, 61, 221, 5, 5, 54, 249, 5, 5, 184, 214, 5, 5, 55, 246, 5, 5, 213, 232, 5, 5, 114, 210, 5, 5, 60, 176, 5, 5, 26, 193, 5, 5, 148, 179, 5, 5, 91, 216, 5, 5, 141, 186, 5, 5, 169, 247, 5, 5, 41, 180, 5, 5, 186, 200, 5, 5, 246, 182, 5, 5, 149, 210, 5, 5, 144, 239, 5, 5, 80, 200, 5, 5, 212, 205, 5, 5, 208, 196, 5, 5, 122, 173, 5, 5, 43, 196, 5, 5, 96, 206, 5, 5, 107, 204, 5, 5, 176, 249, 5, 5, 49, 202, 5, 5, 36, 208, 5, 5, 130, 249, 5, 5, 57, 247, 5, 5, 204, 210, 5, 5, 225, 189, 5, 5, 216, 241, 5, 5, 16, 179, 5, 5, 200, 221, 5, 5, 63, 221, 5, 5, 180, 190, 5, 5, 34, 226, 5, 5, 124, 189, 5, 5, 22, 234, 5, 5, 232, 247, 5, 5, 156, 200, 5, 5, 131, 249, 5, 5, 93, 244, 5, 5, 183, 170, 5, 5, 184, 170, 5, 5, 181, 190, 5, 5, 249, 217, 5, 5, 95, 244, 5, 5, 232, 228, 5, 5, 233, 228, 5, 5, 218, 241, 5, 5, 40, 176, 5, 5, 8, 219, 5, 5, 221, 179, 5, 5, 230, 190, 5, 5, 72, 203, 5, 5, 182, 236, 5, 5, 30, 217, 5, 5, 31, 217, 5, 5, 254, 193, 5, 5, 18, 170, 5, 5, 203, 181, 5, 5, 76, 237, 5, 5, 133, 184, 5, 5, 163, 219, 5, 5, 233, 184, 5, 5, 168, 231, 5, 5, 33, 228, 5, 5, 211, 189, 5, 5, 88, 222, 5, 5, 125, 219, 5, 5, 41, 235, 5, 5, 158, 229, 5, 5, 172, 215, 5, 5, 145, 188, 5, 5, 48, 230, 5, 5, 158, 182, 5, 5, 222, 219, 5, 5, 180, 204, 5, 5, 194, 206, 5, 5, 242, 221, 5, 5, 201, 221, 5, 5, 203, 186, 5, 5, 57, 249, 5, 5, 238, 225, 5, 5, 194, 224, 5, 5, 175, 237, 5, 5, 125, 181, 5, 5, 223, 240, 5, 5, 128, 215, 5, 5, 194, 214, 5, 5, 159, 229, 5, 5, 126, 219, 5, 5, 69, 221, 5, 5, 32, 203, 5, 5, 128, 168, 5, 5, 154, 201, 5, 5, 39, 191, 5, 5, 27, 209, 5, 5, 33, 193, 5, 5, 163, 226, 5, 5, 134, 217, 5, 5, 197, 193, 5, 5, 64, 221, 5, 5, 7, 207, 5, 5, 73, 231, 5, 5, 86, 217, 5, 5, 28, 171, 5, 5, 34, 193, 5, 5, 30, 193, 5, 5, 251, 247, 5, 5, 115, 194, 5, 5, 110, 234, 5, 5, 94, 248, 5, 5, 216, 240, 5, 5, 37, 245, 5, 5, 174, 198, 5, 5, 123, 235, 5, 5, 32, 171, 5, 5, 128, 242, 5, 5, 135, 217, 5, 5, 238, 232, 5, 5, 170, 171, 5, 5, 250, 194, 5, 5, 194, 235, 5, 5, 37, 224, 5, 5, 231, 190, 5, 5, 13, 250, 5, 5, 159, 220, 5, 5, 160, 220, 5, 5, 73, 182, 5, 5, 124, 187, 5, 5, 4, 204, 5, 5, 195, 235, 5, 5, 190, 208, 5, 5, 102, 231, 5, 5, 115, 244, 5, 5, 186, 232, 5, 5, 69, 183, 5, 5, 205, 186, 5, 5, 140, 189, 5, 5, 159, 199, 5, 5, 96, 244, 5, 5, 223, 219, 5, 5, 228, 244, 5, 5, 235, 184, 5, 5, 125, 187, 5, 5, 183, 240, 5, 5, 93, 202, 5, 5, 148, 173, 5, 5, 183, 201, 5, 5, 187, 222, 5, 5, 32, 168, 5, 5, 43, 219, 5, 5, 35, 193, 5, 5, 253, 234, 5, 5, 18, 190, 5, 5, 188, 222, 5, 5, 230, 202, 5, 5, 217, 203, 5, 5, 231, 202, 5, 5, 227, 208, 5, 5, 220, 245, 5, 5, 217, 232, 5, 5, 209, 181, 5, 5, 116, 205, 5, 5, 246, 186, 5, 5, 65, 202, 5, 5, 17, 232, 5, 5, 142, 241, 5, 5, 112, 178, 5, 5, 181, 192, 5, 5, 49, 241, 5, 5, 213, 227, 5, 5, 124, 205, 5, 5, 156, 182, 5, 5, 184, 168, 5, 5, 200, 188, 5, 5, 215, 168, 5, 5, 197, 213, 5, 5, 199, 210, 5, 5, 90, 205, 5, 5, 53, 241, 5, 5, 161, 244, 5, 5, 112, 184, 5, 5, 216, 168, 5, 5, 135, 169, 5, 5, 137, 189, 5, 5, 15, 213, 5, 5, 239, 224, 5, 5, 50, 187, 5, 5, 111, 195, 5, 5, 200, 198, 5, 5, 68, 229, 5, 5, 236, 181, 5, 5, 34, 227, 5, 5, 92, 244, 5, 5, 138, 189, 5, 5, 149, 199, 5, 5, 253, 192, 5, 5, 173, 194, 5, 5, 251, 231, 5, 5, 162, 244, 5, 5, 209, 232, 5, 5, 185, 235, 5, 5, 197, 185, 5, 5, 31, 202, 5, 5, 12, 203, 5, 5, 173, 193, 5, 5, 84, 202, 5, 5, 237, 221, 5, 5, 159, 226, 5, 5, 139, 187, 5, 5, 40, 220, 5, 5, 175, 190, 5, 5, 41, 198, 5, 5, 3, 193, 5, 5, 252, 240, 5, 5, 185, 216, 5, 5, 99, 204, 5, 5, 172, 214, 5, 5, 175, 249, 5, 5, 195, 182, 5, 5, 78, 200, 5, 5, 121, 173, 5, 5, 140, 186, 5, 5, 148, 210, 5, 5, 201, 210, 5, 5, 145, 179, 5, 5, 41, 196, 5, 5, 59, 221, 5, 5, 197, 221, 5, 5, 48, 249, 5, 5, 67, 183, 5, 5, 218, 179, 5, 5, 16, 170, 5, 5, 146, 173, 5, 5, 28, 226, 5, 5, 156, 229, 5, 5, 230, 228, 5, 5, 161, 219, 5, 5, 198, 181, 5, 5, 25, 217, 5, 5, 130, 184, 5, 5, 225, 190, 5, 5, 244, 217, 5, 5, 109, 194, 5, 5, 180, 170, 5, 5, 219, 219, 5, 5, 126, 168, 5, 5, 182, 214, 5, 5, 192, 224, 5, 5, 201, 186, 5, 5, 170, 237, 5, 5, 66, 212, 5, 5, 186, 193, 5, 5, 153, 201, 5, 5, 27, 171, 5, 5, 122, 235, 5, 5, 31, 171, 5, 5, 3, 207, 5, 5, 25, 209, 5, 5, 212, 240, 5, 5, 132, 229, 5, 5, 68, 231, 5, 5, 122, 187, 5, 5, 157, 220, 5, 5, 2, 204, 5, 5, 12, 250, 5, 5, 15, 190, 5, 5, 234, 184, 5, 5, 91, 202, 5, 5, 225, 244, 5, 5, 216, 186, 5, 5, 237, 210, 5, 5, 204, 238, 5, 5, 19, 184, 5, 5, 167, 202, 5, 5, 164, 196, 5, 5, 253, 171, 5, 5, 231, 236, 5, 5, 52, 184, 5, 5, 140, 179, 5, 5, 38, 192, 5, 5, 143, 183, 5, 5, 59, 219, 5, 5, 213, 184, 5, 5, 193, 221, 5, 5, 144, 183, 5, 5, 47, 209, 5, 5, 116, 169, 5, 5, 161, 242, 5, 5, 117, 246, 5, 5, 9, 190, 5, 5, 137, 216, 5, 5, 194, 183, 5, 5, 10, 169, 5, 5, 179, 229, 5, 5, 174, 193, 5, 5, 195, 221, 5, 5, 94, 242, 5, 5, 161, 183, 5, 5, 123, 237, 5, 5, 238, 197, 5, 5, 239, 197, 5, 5, 39, 213, 5, 5, 99, 190, 5, 5, 15, 246, 5, 5, 118, 169, 5, 5, 240, 236, 5, 5, 241, 235, 5, 5, 77, 246, 5, 5, 150, 182, 5, 5, 145, 183, 5, 5, 180, 229, 5, 5, 210, 211, 5, 5, 202, 177, 5, 5, 105, 185, 5, 5, 29, 219, 5, 5, 100, 204, 5, 5, 219, 208, 5, 5, 28, 184, 5, 5, 17, 228, 5, 5, 252, 208, 5, 5, 180, 202, 5, 5, 197, 170, 5, 5, 43, 247, 5, 5, 106, 185, 5, 5, 164, 242, 5, 5, 127, 176, 5, 5, 154, 217, 5, 5, 31, 233, 5, 5, 169, 174, 5, 5, 44, 178, 5, 5, 251, 196, 5, 5, 195, 237, 5, 5, 73, 186, 5, 5, 90, 247, 5, 5, 102, 242, 5, 5, 37, 237, 5, 5, 103, 242, 5, 5, 243, 235, 5, 5, 191, 226, 5, 5, 104, 242, 5, 5, 107, 226, 5, 5, 80, 240, 5, 5, 214, 180, 5, 5, 46, 230, 5, 5, 240, 181, 5, 5, 158, 186, 5, 5, 21, 168, 5, 5, 118, 246, 5, 5, 247, 236, 5, 5, 19, 189, 5, 5, 37, 233, 5, 5, 97, 193, 5, 5, 210, 203, 5, 5, 228, 247, 5, 5, 73, 237, 5, 5, 29, 226, 5, 5, 93, 189, 5, 5, 166, 206, 5, 5, 35, 219, 5, 5, 166, 209, 5, 5, 109, 185, 5, 5, 132, 218, 5, 5, 186, 194, 5, 5, 107, 234, 5, 5, 156, 247, 5, 5, 170, 174, 5, 5, 167, 206, 5, 5, 20, 189, 5, 5, 202, 210, 5, 5, 131, 181, 5, 5, 38, 206, 5, 5, 98, 193, 5, 5, 18, 193, 5, 5, 128, 227, 5, 5, 100, 190, 5, 5, 31, 228, 5, 5, 8, 172, 5, 5, 75, 230, 5, 5, 139, 197, 5, 5, 117, 242, 5, 5, 228, 171, 5, 5, 73, 198, 5, 5, 74, 198, 5, 5, 15, 170, 5, 5, 29, 231, 5, 5, 75, 198, 5, 5, 27, 197, 5, 5, 93, 227, 5, 5, 42, 196, 5, 5, 32, 226, 5, 5, 122, 181, 5, 5, 123, 181, 5, 5, 194, 199, 5, 5, 208, 189, 5, 5, 112, 230, 5, 5, 62, 221, 5, 5, 200, 200, 5, 5, 29, 195, 5, 5, 47, 212, 5, 5, 213, 205, 5, 5, 24, 213, 5, 5, 159, 222, 5, 5, 38, 184, 5, 5, 56, 168, 5, 5, 244, 248, 5, 5, 194, 212, 5, 5, 146, 216, 5, 5, 85, 215, 5, 5, 158, 169, 5, 5, 149, 179, 5, 5, 214, 205, 5, 5, 46, 218, 5, 5, 190, 193, 5, 5, 29, 197, 5, 5, 142, 227, 5, 5, 245, 235, 5, 5, 166, 242, 5, 5, 185, 214, 5, 5, 50, 202, 5, 5, 51, 238, 5, 5, 148, 248, 5, 5, 62, 199, 5, 5, 90, 181, 5, 5, 81, 200, 5, 5, 140, 221, 5, 5, 186, 214, 5, 5, 44, 196, 5, 5, 145, 239, 5, 5, 146, 239, 5, 5, 45, 196, 5, 5, 55, 249, 5, 5, 187, 200, 5, 5, 42, 180, 5, 5, 101, 214, 5, 5, 61, 176, 5, 5, 186, 185, 5, 5, 254, 196, 5, 5, 254, 197, 5, 5, 147, 239, 5, 5, 10, 250, 5, 5, 76, 192, 5, 5, 226, 222, 5, 5, 88, 240, 5, 5, 8, 175, 5, 5, 189, 208, 5, 5, 9, 219, 5, 5, 58, 168, 5, 5, 32, 217, 5, 5, 35, 226, 5, 5, 212, 189, 5, 5, 36, 226, 5, 5, 204, 181, 5, 5, 58, 195, 5, 5, 95, 207, 5, 5, 39, 184, 5, 5, 41, 176, 5, 5, 129, 227, 5, 5, 183, 236, 5, 5, 214, 188, 5, 5, 226, 242, 5, 5, 70, 213, 5, 5, 82, 200, 5, 5, 183, 207, 5, 5, 213, 189, 5, 5, 89, 240, 5, 5, 233, 175, 5, 5, 165, 230, 5, 5, 32, 197, 5, 5, 43, 180, 5, 5, 129, 172, 5, 5, 173, 182, 5, 5, 239, 188, 5, 5, 101, 240, 5, 5, 227, 242, 5, 5, 13, 232, 5, 5, 217, 229, 5, 5, 202, 221, 5, 5, 240, 188, 5, 5, 139, 175, 5, 5, 138, 225, 5, 5, 208, 231, 5, 5, 127, 219, 5, 5, 181, 204, 5, 5, 167, 192, 5, 5, 74, 186, 5, 5, 195, 193, 5, 5, 131, 224, 5, 5, 66, 188, 5, 5, 130, 176, 5, 5, 131, 176, 5, 5, 157, 239, 5, 5, 223, 237, 5, 5, 6, 237, 5, 5, 83, 192, 5, 5, 217, 202, 5, 5, 110, 226, 5, 5, 90, 199, 5, 5, 248, 225, 5, 5, 249, 225, 5, 5, 158, 239, 5, 5, 130, 227, 5, 5, 195, 206, 5, 5, 47, 233, 5, 5, 52, 185, 5, 5, 111, 226, 5, 5, 27, 174, 5, 5, 168, 192, 5, 5, 161, 227, 5, 5, 169, 245, 5, 5, 129, 168, 5, 5, 199, 237, 5, 5, 97, 238, 5, 5, 215, 211, 5, 5, 9, 175, 5, 5, 63, 247, 5, 5, 211, 204, 5, 5, 217, 240, 5, 5, 232, 205, 5, 5, 161, 170, 5, 5, 254, 222, 5, 5, 249, 248, 5, 5, 219, 241, 5, 5, 78, 230, 5, 5, 14, 198, 5, 5, 252, 239, 5, 5, 38, 226, 5, 5, 103, 223, 5, 5, 193, 194, 5, 5, 162, 239, 5, 5, 51, 190, 5, 5, 162, 170, 5, 5, 105, 240, 5, 5, 38, 224, 5, 5, 232, 190, 5, 5, 109, 182, 5, 5, 194, 194, 5, 5, 157, 203, 5, 5, 14, 237, 5, 5, 54, 185, 5, 5, 212, 196, 5, 5, 65, 232, 5, 5, 66, 232, 5, 5, 131, 227, 5, 5, 225, 206, 5, 5, 14, 250, 5, 5, 129, 242, 5, 5, 107, 240, 5, 5, 221, 205, 5, 5, 162, 227, 5, 5, 67, 232, 5, 5, 116, 235, 5, 5, 164, 239, 5, 5, 61, 213, 5, 5, 189, 222, 5, 5, 133, 206, 5, 5, 43, 231, 5, 5, 179, 238, 5, 5, 92, 192, 5, 5, 79, 244, 5, 5, 220, 241, 5, 5, 229, 244, 5, 5, 55, 238, 5, 5, 78, 237, 5, 5, 62, 213, 5, 5, 77, 211, 5, 5, 19, 190, 5, 5, 76, 208, 5, 5, 108, 240, 5, 5, 24, 208, 5, 5, 249, 178, 5, 5, 47, 243, 5, 5, 130, 242, 5, 5, 213, 201, 5, 5, 9, 172, 5, 5, 117, 205, 5, 5, 243, 188, 5, 5, 10, 205, 5, 5, 4, 223, 5, 5, 49, 243, 5, 5, 109, 240, 5, 5, 247, 186, 5, 5, 173, 217, 5, 5, 66, 202, 5, 5, 56, 206, 5, 5, 234, 210, 5, 5, 155, 196, 5, 5, 9, 192, 5, 5, 141, 242, 5, 5, 31, 209, 5, 5, 173, 221, 5, 5, 202, 211, 5, 5, 234, 235, 5, 5, 118, 172, 5, 5, 104, 169, 5, 5, 251, 245, 5, 5, 56, 186, 5, 5, 28, 180, 5, 5, 85, 205, 5, 5, 236, 235, 5, 5, 255, 232, 5, 5, 33, 237, 5, 5, 64, 204, 5, 5, 153, 174, 5, 5, 143, 217, 5, 5, 146, 242, 5, 5, 107, 235, 5, 5, 4, 228, 5, 5, 69, 223, 5, 5, 252, 246, 5, 5, 232, 181, 5, 5, 147, 186, 5, 5, 97, 234, 5, 5, 15, 189, 5, 5, 144, 247, 5, 5, 98, 185, 5, 5, 34, 206, 5, 5, 72, 189, 5, 5, 68, 230, 5, 5, 199, 171, 5, 5, 11, 202, 5, 5, 135, 197, 5, 5, 100, 186, 5, 5, 114, 181, 5, 5, 62, 242, 5, 5, 41, 232, 5, 5, 9, 226, 5, 5, 101, 230, 5, 5, 34, 218, 5, 5, 186, 208, 5, 5, 47, 168, 5, 5, 69, 200, 5, 5, 156, 169, 5, 5, 176, 212, 5, 5, 70, 215, 5, 5, 54, 176, 5, 5, 180, 185, 5, 5, 159, 242, 5, 5, 12, 224, 5, 5, 194, 189, 5, 5, 202, 188, 5, 5, 185, 181, 5, 5, 109, 186, 5, 5, 19, 217, 5, 5, 119, 176, 5, 5, 174, 207, 5, 5, 156, 230, 5, 5, 98, 239, 5, 5, 213, 237, 5, 5, 214, 229, 5, 5, 160, 204, 5, 5, 153, 192, 5, 5, 99, 239, 5, 5, 175, 193, 5, 5, 231, 188, 5, 5, 90, 238, 5, 5, 81, 240, 5, 5, 167, 245, 5, 5, 205, 204, 5, 5, 143, 203, 5, 5, 187, 194, 5, 5, 206, 196, 5, 5, 109, 242, 5, 5, 208, 205, 5, 5, 136, 190, 5, 5, 221, 244, 5, 5, 87, 240, 5, 5, 13, 190, 5, 5, 15, 208, 5, 5, 243, 186, 5, 5, 255, 222, 5, 5, 122, 205, 5, 5, 188, 195, 5, 5, 101, 204, 5, 5, 255, 193, 5, 5, 62, 232, 5, 5, 90, 177, 5, 5, 6, 194, 5, 5, 10, 194, 5, 5, 137, 236, 5, 5, 88, 177, 5, 5, 165, 205, 5, 5, 14, 241, 5, 5, 220, 176, 5, 5, 185, 192, 5, 5, 91, 212, 5, 5, 221, 176, 5, 5, 92, 212, 5, 5, 68, 248, 5, 5, 42, 198, 5, 5, 17, 248, 5, 5, 183, 193, 5, 5, 61, 208, 5, 5, 62, 208, 5, 5, 6, 242, 5, 5, 174, 204, 5, 5, 25, 174, 5, 5, 45, 198, 5, 5, 251, 203, 5, 5, 150, 210, 5, 5, 187, 214, 5, 5, 215, 205, 5, 5, 209, 196, 5, 5, 46, 198, 5, 5, 46, 196, 5, 5, 204, 202, 5, 5, 182, 232, 5, 5, 63, 232, 5, 5, 74, 193, 5, 5, 67, 208, 5, 5, 218, 202, 5, 5, 16, 221, 5, 5, 38, 245, 5, 5, 5, 204, 5, 5, 49, 196, 5, 5, 196, 214, 5, 5, 116, 204, 5, 5, 138, 236, 5, 5, 222, 176, 5, 5, 234, 206, 5, 5, 230, 206, 5, 5, 203, 188, 5, 5, 205, 173, 5, 5, 195, 183, 5, 5, 177, 208, 5, 5, 178, 208, 5, 5, 197, 183, 5, 5, 100, 212, 5, 5, 192, 217, 5, 5, 127, 217, 5, 5, 167, 209, 5, 5, 198, 183, 5, 5, 152, 232, 5, 5, 188, 200, 5, 5, 130, 217, 5, 5, 180, 208, 5, 5, 10, 175, 5, 5, 117, 183, 5, 5, 182, 183, 5, 5, 123, 217, 5, 5, 179, 208, 5, 5, 177, 206, 5, 5, 136, 207, 5, 5, 71, 209, 5, 5, 34, 191, 5, 5, 149, 209, 5, 5, 177, 249, 5, 5, 167, 211, 5, 5, 51, 183, 5, 5, 203, 190, 5, 5, 193, 190, 5, 5, 189, 195, 5, 5, 12, 187, 5, 5, 83, 226, 5, 5, 74, 227, 5, 5, 91, 189, 5, 5, 74, 190, 5, 5, 220, 199, 5, 5, 94, 189, 5, 5, 172, 222, 5, 5, 32, 202, 5, 5, 203, 210, 5, 5, 176, 174, 5, 5, 246, 188, 5, 5, 245, 188, 5, 5, 100, 239, 5, 5, 122, 215, 5, 5, 5, 178, 5, 5, 146, 231, 5, 5, 203, 227, 5, 5, 137, 209, 5, 5, 130, 209, 5, 5, 123, 215, 5, 5, 205, 177, 5, 5, 189, 175, 5, 5, 84, 241, 5, 5, 99, 179, 5, 5, 95, 238, 5, 5, 206, 231, 5, 5, 179, 236, 5, 5, 131, 217, 5, 5, 200, 207, 5, 5, 184, 236, 5, 5, 202, 216, 5, 5, 42, 243, 5, 5, 51, 202, 5, 5, 70, 178, 5, 5, 146, 180, 5, 5, 100, 172, 5, 5, 80, 236, 5, 5, 130, 236, 5, 5, 188, 236, 5, 5, 7, 178, 5, 5, 88, 168, 5, 5, 243, 245, 5, 5, 211, 177, 5, 5, 101, 172, 5, 5, 98, 238, 5, 5, 184, 207, 5, 5, 248, 244, 5, 5, 197, 236, 5, 5, 155, 180, 5, 5, 118, 205, 5, 5, 192, 246, 5, 5, 53, 183, 5, 5, 36, 184, 5, 5, 72, 184, 5, 5, 170, 208, 5, 5, 155, 208, 5, 5, 221, 242, 5, 5, 242, 176, 5, 5, 195, 217, 5, 5, 198, 173, 5, 5, 59, 228, 5, 5, 233, 247, 5, 5, 121, 246, 5, 5, 23, 208, 5, 5, 130, 168, 5, 5, 33, 171, 5, 5, 22, 228, 5, 5, 163, 170, 5, 5, 196, 242, 5, 5, 190, 173, 5, 5, 20, 228, 5, 5, 242, 179, 5, 5, 123, 208, 5, 5, 17, 210, 5, 5, 244, 179, 5, 5, 41, 249, 5, 5, 104, 186, 5, 5, 105, 186, 5, 5, 40, 180, 5, 5, 46, 183, 5, 5, 207, 225, 5, 5, 168, 242, 5, 5, 27, 213, 5, 5, 123, 173, 5, 5, 53, 185, 5, 5, 9, 215, 5, 5, 169, 242, 5, 5, 100, 225, 5, 5, 222, 225, 5, 5, 175, 222, 5, 5, 168, 222, 5, 5, 44, 183, 5, 5, 243, 182, 5, 5, 186, 229, 5, 5, 203, 168, 5, 5, 147, 179, 5, 5, 18, 228, 5, 5, 93, 247, 5, 5, 158, 217, 5, 5, 139, 221, 5, 5, 239, 221, 5, 5, 111, 241, 5, 5, 241, 221, 5, 5, 235, 226, 5, 5, 77, 230, 5, 5, 33, 197, 5, 5, 47, 196, 5, 5, 174, 191, 5, 5, 34, 197, 5, 5, 190, 236, 5, 5, 94, 227, 5, 5, 104, 223, 5, 5, 41, 231, 5, 5, 157, 232, 5, 5, 195, 236, 5, 5, 186, 201, 5, 5, 20, 170, 5, 5, 215, 237, 5, 5, 84, 217, 5, 5, 241, 187, 5, 5, 160, 230, 5, 5, 162, 230, 5, 5, 107, 189, 5, 5, 211, 233, 5, 5, 206, 181, 5, 5, 141, 244, 5, 5, 137, 234, 5, 5, 231, 229, 5, 5, 142, 244, 5, 5, 125, 211, 5, 5, 62, 210, 5, 5, 165, 214, 5, 5, 189, 244, 5, 5, 19, 193, 5, 5, 56, 249, 5, 5, 84, 192, 5, 5, 90, 192, 5, 5, 104, 214, 5, 5, 53, 192, 5, 5, 222, 174, 5, 5, 44, 174, 5, 5, 45, 174, 5, 5, 212, 188, 5, 5, 12, 236, 5, 5, 251, 239, 5, 5, 208, 233, 5, 5, 95, 169, 5, 5, 77, 244, 5, 5, 209, 233, 5, 5, 253, 172, 5, 5, 173, 174, 5, 5, 185, 236, 5, 5, 54, 197, 5, 5, 162, 226, 5, 5, 110, 204, 5, 5, 105, 204, 5, 5, 109, 175, 5, 5, 251, 217, 5, 5, 210, 233, 5, 5, 32, 199, 5, 5, 37, 211, 5, 5, 211, 196, 5, 5, 198, 237, 5, 5, 91, 176, 5, 5, 113, 243, 5, 5, 8, 242, 5, 5, 154, 175, 5, 5, 217, 238, 5, 5, 151, 210, 5, 5, 78, 244, 5, 5, 179, 249, 5, 5, 181, 217, 5, 5, 114, 243, 5, 5, 196, 236, 5, 5, 216, 211, 5, 5, 208, 181, 5, 5, 21, 230, 5, 5, 166, 239, 5, 5, 102, 176, 5, 5, 180, 249, 5, 5, 86, 179, 5, 5, 155, 175, 5, 5, 9, 196, 5, 5, 48, 236, 5, 5, 209, 174, 5, 5, 40, 174, 5, 5, 187, 188, 5, 5, 236, 239, 5, 5, 48, 197, 5, 5, 89, 204, 5, 5, 90, 169, 5, 5, 153, 226, 5, 5, 39, 249, 5, 5, 28, 199, 5, 5, 254, 241, 5, 5, 85, 176, 5, 5, 180, 217, 5, 5, 20, 230, 5, 5, 237, 204, 5, 5, 67, 212, 5, 5, 230, 185, 5, 5, 68, 212, 5, 5, 186, 236, 5, 5, 254, 204, 5, 5, 22, 205, 5, 5, 231, 185, 5, 5, 134, 198, 5, 5, 152, 177, 5, 5, 117, 204, 5, 5, 153, 177, 5, 5, 213, 204, 5, 5, 218, 185, 5, 5, 130, 198, 5, 5, 59, 187, 5, 5, 34, 217, 5, 5, 35, 171, 5, 5, 27, 187, 5, 5, 38, 243, 5, 5, 8, 176, 5, 5, 217, 188, 5, 5, 17, 198, 5, 5, 143, 233, 5, 5, 132, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 214, 5, 5, 200, 185, 5, 5, 213, 173, 5, 5, 112, 193, 5, 5, 40, 190, 5, 5, 228, 175, 5, 5, 58, 197, 5, 5, 59, 187, 5, 5, 59, 187, 5, 5, 237, 214, 5, 5, 170, 195, 5, 5, 137, 200, 5, 5, 11, 210, 5, 5, 15, 201, 5, 5, 206, 200, 5, 5, 121, 206, 5, 5, 129, 206, 5, 5, 120, 206, 5, 5, 140, 206, 5, 5, 130, 206, 5, 5, 151, 201, 5, 5, 149, 206, 5, 5, 125, 201, 5, 5, 154, 206, 5, 5, 158, 206, 5, 5, 128, 201, 5, 5, 164, 206, 5, 5, 60, 206, 5, 5, 57, 206, 5, 5, 244, 200, 5, 5, 32, 201, 5, 5, 242, 200, 5, 5, 56, 206, 5, 5, 215, 200, 5, 5, 30, 201, 5, 5, 229, 200, 5, 5, 234, 200, 5, 5, 125, 200, 5, 5, 152, 200, 5, 5, 157, 200, 5, 5, 45, 201, 5, 5, 64, 201, 5, 5, 74, 201, 5, 5, 42, 201, 5, 5, 39, 201, 5, 5, 162, 200, 5, 5, 240, 201, 5, 5, 90, 201, 5, 5, 132, 205, 5, 5, 139, 205, 5, 5, 103, 205, 5, 5, 92, 205, 5, 5, 107, 201, 5, 5, 100, 205, 5, 5, 125, 205, 5, 5, 178, 205, 5, 5, 217, 205, 5, 5, 131, 205, 5, 5, 221, 205, 5, 5, 175, 205, 5, 5, 169, 205, 5, 5, 18, 206, 5, 5, 163, 205, 5, 5, 198, 205, 5, 5, 165, 205, 5, 5, 103, 206, 5, 5, 18, 205, 5, 5, 121, 211, 5, 5, 255, 204, 5, 5, 2, 205, 5, 5, 84, 201, 5, 5, 197, 201, 5, 5, 177, 205, 5, 5, 159, 201, 5, 5, 202, 201, 5, 5, 251, 205, 5, 5, 43, 205, 5, 5, 218, 201, 5, 5, 65, 205, 5, 5, 219, 201, 5, 5, 6, 206, 5, 5, 63, 205, 5, 5, 158, 201, 5, 5, 136, 201, 5, 5, 9, 204, 5, 5, 57, 204, 5, 5, 237, 201, 5, 5, 90, 204, 5, 5, 77, 204, 5, 5, 63, 204, 5, 5, 147, 180, 5, 5, 221, 209, 5, 5, 151, 201, 5, 5, 184, 211, 5, 5, 214, 177, 5, 5, 65, 211, 5, 5, 145, 211, 5, 5, 12, 206, 5, 5, 37, 239, 5, 5, 159, 169, 5, 5, 25, 212, 5, 5, 210, 170, 5, 5, 102, 184, 5, 5, 10, 172, 5, 5, 101, 208, 5, 5, 219, 222, 5, 5, 153, 224, 5, 5, 73, 172, 5, 5, 156, 219, 5, 5, 146, 221, 5, 5, 19, 238, 5, 5, 39, 223, 5, 5, 24, 220, 5, 5, 254, 173, 201, 32, 230, 0, 5, 5, 218, 221, 5, 5, 113, 219, 5, 5, 61, 206, 5, 5, 62, 206, 5, 5, 102, 203, 5, 5, 92, 203, 5, 5, 77, 203, 5, 5, 78, 203, 5, 5, 89, 203, 5, 5, 75, 203, 5, 5, 112, 203, 5, 5, 110, 203, 5, 5, 175, 202, 5, 5, 235, 205, 5, 5, 148, 211, 5, 5, 95, 205, 5, 5, 242, 205, 5, 5, 25, 206, 5, 5, 202, 202, 5, 5, 226, 205, 5, 5, 64, 202, 5, 5, 204, 202, 5, 5, 32, 202, 5, 5, 95, 202, 5, 5, 176, 202, 5, 5, 177, 202, 5, 5, 223, 202, 5, 5, 195, 210, 5, 5, 7, 203, 5, 5, 73, 203, 5, 5, 212, 210, 5, 5, 250, 202, 5, 5, 62, 203, 5, 5, 40, 203, 5, 5, 196, 210, 5, 5, 65, 203, 5, 5, 22, 203, 5, 5, 215, 210, 5, 5, 251, 202, 5, 5, 241, 202, 5, 5, 69, 203, 5, 5, 180, 203, 5, 5, 181, 203, 5, 5, 202, 236, 5, 5, 194, 203, 5, 5, 201, 203, 5, 5, 39, 223, 5, 5, 247, 202, 5, 5, 223, 210, 5, 5, 206, 210, 5, 5, 68, 203, 5, 5, 28, 203, 5, 5, 212, 203, 5, 5, 124, 204, 5, 5, 43, 204, 5, 5, 65, 211, 5, 5, 123, 204, 5, 5, 238, 202, 5, 5, 55, 204, 5, 5, 132, 240, 5, 5, 73, 204, 5, 5, 75, 204, 5, 5, 86, 204, 5, 5, 88, 204, 5, 5, 113, 204, 5, 5, 122, 204, 5, 5, 109, 202, 5, 5, 87, 202, 5, 5, 90, 202, 5, 5, 186, 202, 5, 5, 166, 181, 5, 5, 155, 201, 5, 5, 122, 203, 5, 5, 130, 203, 5, 5, 245, 210, 5, 5, 171, 203, 5, 5, 151, 201, 5, 5, 138, 203, 5, 5, 141, 203, 5, 5, 162, 203, 5, 5, 136, 203, 5, 5, 237, 204, 5, 5, 53, 243, 5, 5, 67, 219, 5, 5, 154, 204, 5, 5, 91, 175, 5, 5, 183, 204, 5, 5, 136, 204, 5, 5, 129, 204, 5, 5, 139, 204, 5, 5, 137, 204, 5, 5, 141, 204, 5, 5, 96, 211, 5, 5, 227, 201, 5, 5, 198, 204, 5, 5, 186, 205, 5, 5, 157, 205, 5, 5, 79, 206, 5, 5, 84, 206, 5, 5, 86, 206, 5, 5, 93, 206, 5, 5, 10, 206, 5, 5, 161, 202, 5, 5, 131, 202, 5, 5, 12, 206, 5, 5, 227, 204, 5, 5, 103, 202, 5, 5, 100, 202, 5, 5, 255, 205, 5, 5, 251, 238, 5, 5, 68, 202, 5, 5, 2, 202, 5, 5, 127, 210, 5, 5, 77, 202, 5, 5, 154, 202, 5, 5, 35, 202, 5, 5, 81, 202, 5, 5, 78, 202, 5, 5, 69, 202, 5, 5, 44, 202, 5, 5, 179, 210, 5, 5, 185, 210, 5, 5, 20, 204, 5, 5, 242, 203, 5, 5, 243, 203, 5, 5, 33, 204, 5, 5, 235, 203, 5, 5, 4, 204, 5, 5, 5, 204, 5, 5, 221, 203, 5, 5, 226, 203, 5, 5, 248, 203, 5, 5, 99, 202, 5, 5, 144, 202, 5, 5, 145, 202, 5, 5, 136, 248, 5, 5, 209, 246, 5, 5, 240, 221, 5, 5, 65, 221, 5, 5, 219, 172, 5, 5, 142, 176, 5, 5, 67, 216, 5, 5, 171, 180, 5, 5, 195, 224, 5, 5, 126, 225, 5, 5, 190, 244, 5, 5, 60, 180, 5, 5, 138, 169, 5, 5, 31, 184, 5, 5, 24, 234, 5, 5, 134, 194, 5, 5, 13, 194, 5, 5, 111, 200, 5, 5, 114, 230, 5, 5, 252, 189, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 247, 192, 0, 0, 0, 5, 5, 193, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 231, 5, 5, 210, 247, 5, 5, 27, 239, 5, 5, 67, 202, 5, 5, 67, 221, 5, 5, 194, 232, 5, 5, 10, 184, 5, 5, 103, 196, 5, 5, 38, 196, 5, 5, 226, 241, 192, 0, 0, 0, 5, 5, 29, 217, 192, 0, 0, 0, 5, 5, 224, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 239, 5, 5, 77, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 182, 5, 5, 140, 223, 5, 5, 215, 186, 5, 5, 240, 188, 5, 5, 40, 201, 5, 5, 179, 202, 5, 5, 91, 230, 5, 5, 14, 220, 5, 5, 153, 208, 5, 5, 156, 208, 5, 5, 128, 216, 5, 5, 148, 169, 5, 5, 151, 188, 5, 5, 81, 225, 5, 5, 2, 215, 5, 5, 188, 213, 5, 5, 136, 209, 5, 5, 193, 172, 5, 5, 223, 173, 5, 5, 58, 191, 5, 5, 105, 198, 5, 5, 105, 244, 5, 5, 135, 174, 5, 5, 10, 209, 5, 5, 210, 192, 5, 5, 170, 222, 5, 5, 152, 207, 5, 5, 213, 187, 5, 5, 11, 248, 5, 5, 49, 188, 5, 5, 12, 248, 5, 5, 83, 245, 5, 5, 26, 250, 5, 5, 155, 169, 5, 5, 246, 220, 5, 5, 176, 246, 5, 5, 116, 214, 5, 5, 101, 241, 5, 5, 209, 249, 5, 5, 39, 248, 5, 5, 224, 191, 5, 5, 200, 245, 5, 5, 110, 186, 5, 5, 83, 227, 5, 5, 100, 195, 5, 5, 65, 203, 5, 5, 252, 195, 5, 5, 98, 182, 5, 5, 173, 222, 5, 5, 151, 245, 5, 5, 99, 175, 5, 5, 156, 172, 5, 5, 156, 172, 5, 5, 171, 245, 5, 5, 229, 188, 5, 5, 46, 222, 5, 5, 39, 238, 5, 5, 214, 195, 5, 5, 62, 171, 5, 5, 120, 244, 5, 5, 71, 176, 5, 5, 47, 239, 5, 5, 37, 210, 5, 5, 214, 232, 5, 5, 143, 213, 5, 5, 79, 191, 192, 0, 0, 0, 5, 5, 213, 186, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 171, 5, 5, 202, 199, 5, 5, 207, 217, 5, 5, 28, 234, 5, 5, 11, 175, 5, 5, 4, 193, 5, 5, 230, 240, 5, 5, 210, 220, 5, 5, 151, 188, 5, 5, 189, 225, 5, 5, 85, 191, 5, 5, 69, 228, 5, 5, 102, 247, 5, 5, 35, 183, 5, 5, 144, 236, 5, 5, 208, 169, 5, 5, 74, 170, 5, 5, 124, 176, 5, 5, 109, 168, 5, 5, 79, 239, 5, 5, 61, 172, 5, 5, 144, 237, 5, 5, 182, 228, 5, 5, 104, 221, 5, 5, 63, 242, 5, 5, 105, 244, 5, 5, 149, 168, 5, 5, 135, 174, 5, 5, 203, 177, 5, 5, 169, 241, 5, 5, 194, 223, 5, 5, 114, 171, 5, 5, 103, 168, 5, 5, 193, 216, 5, 5, 64, 201, 5, 5, 199, 228, 5, 5, 51, 245, 5, 5, 167, 177, 5, 5, 24, 220, 5, 5, 136, 204, 5, 5, 237, 239, 5, 5, 22, 249, 5, 5, 49, 188, 5, 5, 113, 196, 5, 5, 12, 248, 5, 5, 32, 216, 5, 5, 244, 197, 5, 5, 116, 182, 5, 5, 210, 247, 5, 5, 9, 246, 5, 5, 115, 176, 5, 5, 56, 190, 5, 5, 197, 186, 5, 5, 149, 229, 5, 5, 27, 239, 5, 5, 154, 221, 5, 5, 125, 246, 5, 5, 163, 197, 5, 5, 170, 245, 5, 5, 105, 226, 5, 5, 171, 226, 5, 5, 100, 195, 5, 5, 216, 201, 5, 5, 163, 225, 5, 5, 65, 203, 5, 5, 192, 213, 5, 5, 151, 245, 5, 5, 184, 190, 5, 5, 36, 190, 5, 5, 59, 243, 5, 5, 232, 215, 5, 5, 131, 184, 5, 5, 46, 222, 5, 5, 169, 179, 5, 5, 224, 247, 5, 5, 210, 216, 5, 5, 39, 238, 5, 5, 184, 211, 5, 5, 99, 242, 5, 5, 214, 195, 5, 5, 227, 170, 5, 5, 120, 244, 5, 5, 153, 222, 5, 5, 200, 174, 5, 5, 206, 223, 5, 5, 40, 234, 5, 5, 10, 248, 5, 5, 37, 210, 5, 5, 103, 196, 5, 5, 253, 168, 5, 5, 214, 232, 5, 5, 128, 181, 5, 5, 143, 213, 5, 5, 242, 245, 5, 5, 59, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 181, 5, 5, 67, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 100, 232, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 95, 222, 6, 5, 81, 249, 6, 5, 245, 222, 192, 0, 0, 0, 6, 5, 6, 182, 6, 5, 25, 181, 6, 5, 131, 195, 6, 5, 77, 226, 6, 5, 156, 194, 6, 5, 176, 240, 6, 5, 66, 230, 6, 5, 171, 203, 6, 5, 98, 215, 6, 5, 131, 189, 6, 5, 187, 243, 6, 5, 229, 233, 6, 5, 105, 175, 6, 5, 83, 247, 6, 5, 114, 221, 6, 5, 130, 182, 6, 5, 117, 221, 6, 5, 6, 176, 6, 5, 170, 236, 6, 5, 65, 227, 6, 5, 5, 172, 6, 5, 57, 238, 6, 5, 169, 219, 6, 5, 48, 241, 6, 5, 30, 250, 6, 5, 67, 247, 6, 5, 87, 241, 6, 5, 180, 246, 6, 5, 182, 249, 6, 5, 154, 177, 6, 5, 239, 195, 6, 5, 52, 199, 6, 5, 156, 188, 6, 5, 10, 207, 6, 5, 63, 241, 6, 5, 7, 243, 6, 5, 19, 221, 6, 5, 99, 185, 6, 5, 116, 240, 6, 5, 130, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 227, 113, 0, 197, 67, 114, 0, 197, 163, 114, 0, 197, 3, 115, 0, 197, 99, 115, 0, 197, 195, 115, 0, 197, 35, 116, 0, 197, 131, 116, 0, 197, 227, 116, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 159, 178, 6, 5, 243, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 202, 5, 5, 96, 228, 5, 5, 99, 238, 192, 0, 0, 0, 5, 5, 154, 210, 5, 5, 91, 230, 5, 5, 186, 227, 5, 5, 130, 171, 5, 5, 163, 243, 5, 5, 183, 169, 5, 5, 14, 220, 5, 5, 229, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 208, 5, 5, 139, 227, 5, 5, 247, 190, 5, 5, 69, 197, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 210, 5, 5, 187, 243, 192, 0, 0, 0, 5, 5, 236, 218, 5, 5, 143, 242, 5, 5, 92, 171, 5, 5, 20, 180, 5, 5, 202, 199, 192, 0, 0, 0, 5, 5, 129, 215, 5, 5, 143, 218, 192, 0, 0, 0, 5, 5, 253, 198, 5, 5, 137, 206, 5, 5, 99, 185, 5, 5, 76, 173, 192, 0, 0, 0, 5, 5, 230, 240, 5, 5, 156, 208, 5, 5, 128, 216, 5, 5, 210, 220, 5, 5, 82, 169, 5, 5, 153, 176, 5, 5, 159, 169, 5, 5, 65, 191, 5, 5, 148, 169, 5, 5, 196, 171, 5, 5, 104, 192, 5, 5, 111, 192, 5, 5, 174, 216, 5, 5, 174, 216, 5, 5, 174, 216, 192, 0, 0, 0, 5, 5, 255, 190, 5, 5, 96, 192, 5, 5, 213, 223, 192, 0, 0, 0, 5, 5, 251, 194, 5, 5, 224, 174, 5, 5, 113, 237, 5, 5, 106, 232, 5, 5, 171, 230, 5, 5, 89, 174, 5, 5, 131, 247, 5, 5, 145, 181, 5, 5, 210, 208, 5, 5, 103, 225, 5, 5, 211, 214, 5, 5, 35, 232, 5, 5, 130, 220, 5, 5, 130, 220, 5, 5, 85, 191, 5, 5, 158, 174, 5, 5, 130, 244, 5, 5, 142, 237, 5, 5, 119, 227, 5, 5, 81, 225, 5, 5, 120, 227, 5, 5, 100, 188, 5, 5, 138, 212, 5, 5, 67, 216, 5, 5, 133, 248, 5, 5, 95, 174, 5, 5, 131, 246, 5, 5, 141, 227, 5, 5, 30, 234, 5, 5, 110, 176, 5, 5, 133, 169, 5, 5, 53, 179, 192, 0, 0, 0, 5, 5, 229, 206, 5, 5, 178, 189, 5, 5, 119, 183, 5, 5, 25, 181, 5, 5, 39, 208, 5, 5, 228, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 192, 5, 5, 150, 241, 5, 5, 188, 227, 5, 5, 130, 213, 5, 5, 97, 184, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 218, 5, 5, 14, 201, 5, 5, 14, 201, 192, 0, 0, 0, 5, 5, 147, 242, 5, 5, 8, 247, 5, 5, 65, 211, 5, 5, 121, 169, 192, 0, 0, 0, 5, 5, 100, 222, 5, 5, 90, 194, 5, 5, 51, 178, 5, 5, 24, 241, 192, 0, 0, 0, 5, 5, 106, 227, 5, 5, 223, 173, 5, 5, 120, 234, 5, 5, 40, 215, 192, 0, 0, 0, 5, 5, 157, 236, 192, 0, 0, 0, 5, 5, 71, 179, 5, 5, 19, 249, 5, 5, 77, 173, 5, 5, 162, 235, 5, 5, 186, 173, 192, 0, 0, 0, 5, 5, 214, 235, 5, 5, 17, 223, 5, 5, 120, 207, 5, 5, 36, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 171, 5, 5, 75, 170, 5, 5, 204, 222, 5, 5, 45, 201, 192, 0, 0, 0, 5, 5, 232, 185, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 241, 5, 5, 159, 225, 5, 5, 159, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 234, 5, 5, 135, 179, 192, 0, 0, 0, 5, 5, 36, 180, 5, 5, 133, 218, 5, 5, 199, 246, 5, 5, 186, 199, 5, 5, 144, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 191, 192, 0, 0, 0, 5, 5, 242, 180, 5, 5, 114, 176, 5, 5, 188, 190, 5, 5, 104, 221, 5, 5, 188, 190, 5, 5, 39, 205, 5, 5, 105, 244, 5, 5, 145, 232, 5, 5, 63, 183, 5, 5, 97, 172, 5, 5, 6, 208, 5, 5, 135, 174, 5, 5, 15, 201, 5, 5, 87, 174, 5, 5, 84, 193, 5, 5, 91, 199, 5, 5, 126, 169, 5, 5, 191, 168, 5, 5, 129, 197, 192, 0, 0, 0, 5, 5, 117, 228, 5, 5, 131, 213, 5, 5, 243, 220, 5, 5, 225, 219, 5, 5, 119, 192, 192, 0, 0, 0, 5, 5, 237, 195, 5, 5, 113, 236, 5, 5, 152, 236, 192, 0, 0, 0, 5, 5, 72, 209, 5, 5, 144, 194, 5, 5, 30, 191, 5, 5, 222, 194, 192, 0, 0, 0, 5, 5, 10, 209, 5, 5, 98, 196, 192, 0, 0, 0, 5, 5, 228, 192, 5, 5, 130, 222, 5, 5, 226, 195, 192, 0, 0, 0, 5, 5, 170, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 207, 5, 5, 160, 208, 5, 5, 240, 249, 5, 5, 44, 214, 5, 5, 235, 209, 192, 0, 0, 0, 5, 5, 64, 201, 5, 5, 199, 228, 5, 5, 110, 249, 5, 5, 205, 214, 5, 5, 230, 170, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 186, 5, 5, 10, 192, 5, 5, 198, 219, 5, 5, 152, 207, 192, 0, 0, 0, 5, 5, 100, 168, 5, 5, 211, 169, 5, 5, 20, 249, 192, 0, 0, 0, 5, 5, 133, 244, 5, 5, 154, 237, 5, 5, 186, 184, 5, 5, 236, 220, 192, 0, 0, 0, 5, 5, 48, 247, 192, 0, 0, 0, 5, 5, 139, 176, 192, 0, 0, 0, 5, 5, 168, 234, 192, 0, 0, 0, 5, 5, 83, 224, 5, 5, 144, 229, 5, 5, 24, 220, 5, 5, 49, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 182, 192, 0, 0, 0, 5, 5, 91, 236, 5, 5, 209, 169, 5, 5, 42, 215, 5, 5, 59, 230, 5, 5, 3, 212, 5, 5, 213, 187, 5, 5, 136, 204, 5, 5, 122, 188, 5, 5, 228, 195, 5, 5, 254, 210, 192, 0, 0, 0, 5, 5, 178, 213, 5, 5, 27, 185, 5, 5, 64, 236, 192, 0, 0, 0, 5, 5, 22, 249, 5, 5, 72, 179, 192, 0, 0, 0, 5, 5, 59, 236, 5, 5, 194, 173, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 183, 5, 5, 40, 243, 5, 5, 113, 196, 5, 5, 148, 240, 192, 0, 0, 0, 5, 5, 127, 215, 5, 5, 174, 243, 5, 5, 100, 248, 5, 5, 75, 225, 192, 0, 0, 0, 5, 5, 196, 180, 192, 0, 0, 0, 5, 5, 173, 176, 192, 0, 0, 0, 5, 5, 10, 177, 5, 5, 244, 197, 5, 5, 149, 244, 192, 0, 0, 0, 5, 5, 223, 230, 5, 5, 195, 169, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 224, 5, 5, 170, 228, 192, 0, 0, 0, 5, 5, 15, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 177, 5, 5, 185, 241, 5, 5, 9, 246, 5, 5, 163, 224, 5, 5, 3, 217, 192, 0, 0, 0, 5, 5, 82, 219, 192, 0, 0, 0, 5, 5, 176, 243, 192, 0, 0, 0, 5, 5, 37, 239, 192, 0, 0, 0, 5, 5, 248, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 245, 5, 5, 184, 245, 5, 5, 184, 245, 5, 5, 163, 197, 192, 0, 0, 0, 5, 5, 244, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 234, 5, 5, 175, 205, 5, 5, 105, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 249, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 184, 5, 5, 160, 222, 192, 0, 0, 0, 5, 5, 110, 186, 5, 5, 249, 192, 5, 5, 198, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 248, 5, 5, 247, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 169, 192, 0, 0, 0, 5, 5, 149, 194, 5, 5, 185, 219, 5, 5, 213, 192, 192, 0, 0, 0, 5, 5, 237, 169, 192, 0, 0, 0, 5, 5, 36, 249, 5, 5, 144, 249, 5, 5, 217, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 173, 192, 0, 0, 0, 5, 5, 104, 237, 5, 5, 125, 168, 5, 5, 151, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 213, 192, 0, 0, 0, 5, 5, 181, 176, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 242, 5, 5, 36, 177, 192, 0, 0, 0, 5, 5, 13, 213, 5, 5, 179, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 241, 5, 5, 128, 231, 5, 5, 116, 176, 192, 0, 0, 0, 5, 5, 202, 214, 5, 5, 14, 242, 5, 5, 83, 246, 5, 5, 82, 201, 5, 5, 20, 190, 5, 5, 142, 182, 5, 5, 251, 235, 5, 5, 106, 199, 192, 0, 0, 0, 5, 5, 113, 219, 5, 5, 254, 172, 5, 5, 199, 218, 5, 5, 82, 195, 5, 5, 61, 230, 5, 5, 64, 207, 5, 5, 10, 179, 5, 5, 171, 245, 5, 5, 181, 213, 5, 5, 11, 197, 5, 5, 33, 198, 5, 5, 68, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 213, 5, 5, 208, 195, 5, 5, 87, 177, 192, 0, 0, 0, 5, 5, 84, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 211, 5, 5, 125, 205, 5, 5, 237, 199, 5, 5, 149, 231, 5, 5, 152, 174, 5, 5, 54, 215, 5, 5, 157, 242, 5, 5, 194, 213, 5, 5, 59, 243, 5, 5, 211, 179, 5, 5, 221, 247, 5, 5, 151, 176, 192, 0, 0, 0, 5, 5, 93, 187, 5, 5, 211, 232, 192, 0, 0, 0, 5, 5, 179, 248, 5, 5, 63, 238, 192, 0, 0, 0, 5, 5, 142, 204, 5, 5, 23, 223, 192, 0, 0, 0, 5, 5, 169, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 174, 5, 5, 99, 242, 5, 5, 227, 170, 5, 5, 246, 221, 192, 0, 0, 0, 5, 5, 225, 186, 5, 5, 93, 170, 5, 5, 252, 184, 5, 5, 210, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 168, 5, 5, 217, 193, 5, 5, 78, 213, 192, 0, 0, 0, 5, 5, 168, 218, 5, 5, 153, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 214, 5, 5, 174, 213, 5, 5, 26, 249, 192, 0, 0, 0, 5, 5, 189, 173, 5, 5, 203, 171, 5, 5, 28, 235, 5, 5, 47, 190, 5, 5, 119, 171, 5, 5, 234, 215, 5, 5, 201, 243, 192, 0, 0, 0, 5, 5, 94, 198, 192, 0, 0, 0, 5, 5, 70, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 215, 5, 5, 252, 230, 5, 5, 112, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 181, 5, 5, 128, 181, 5, 5, 165, 213, 192, 0, 0, 0, 5, 5, 20, 192, 192, 0, 0, 0, 5, 5, 196, 181, 5, 5, 62, 243, 5, 5, 115, 169, 5, 5, 193, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 245, 5, 5, 30, 193, 5, 5, 174, 193, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 206, 192, 0, 0, 0, 5, 5, 192, 246, 5, 5, 155, 208, 5, 5, 121, 246, 5, 5, 123, 208, 5, 5, 46, 183, 5, 5, 20, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 134, 2, 4, 0, 0, 0, 42, 0, 134, 3, 4, 0, 0, 0, 50, 0, 134, 2, 4, 0, 0, 0, 50, 0, 134, 3, 4, 0, 0, 0, 58, 0, 134, 2, 4, 0, 0, 0, 58, 0, 134, 3, 4, 0, 0, 0, 66, 0, 134, 2, 4, 0, 0, 0, 66, 0, 134, 3, 4, 0, 0, 0, 68, 0, 134, 2, 4, 0, 0, 0, 68, 0, 134, 3, 4, 0, 0, 0, 70, 0, 134, 2, 4, 0, 0, 0, 70, 0, 134, 3, 4, 0, 0, 0, 82, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 2, 4, 0, 0, 0, 82, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 57, 238, 28, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 6, 182, 24, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 169, 219, 26, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 109, 223, 12, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 80, 230, 23, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 198, 204, 18, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 63, 214, 27, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 161, 168, 19, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 169, 196, 25, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 204, 221, 16, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 7, 243, 8, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 200, 191, 3, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 15, 223, 4, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 189, 209, 6, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 132, 227, 11, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 186, 218, 9, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 207, 247, 5, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 63, 241, 7, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 29, 209, 14, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 213, 225, 2, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 83, 201, 17, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 170, 177, 22, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 145, 189, 13, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 101, 235, 10, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 199, 214, 21, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 103, 233, 15, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 87, 234, 20, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 212, 55, 18, 254, 0, 0, 0, 0, 134, 3, 0, 42, 5, 162, 0, 0, 134, 2, 0, 42, 5, 162, 0, 0, 5, 3, 0, 42, 5, 162, 0, 0, 5, 2, 0, 42, 5, 162, 0, 0, 134, 3, 0, 50, 5, 162, 0, 0, 134, 2, 0, 50, 5, 162, 0, 0, 5, 3, 0, 50, 5, 162, 0, 0, 5, 2, 0, 50, 5, 162, 0, 0, 134, 3, 0, 58, 5, 162, 0, 0, 134, 2, 0, 58, 5, 162, 0, 0, 5, 3, 0, 58, 5, 162, 0, 0, 5, 2, 0, 58, 5, 162, 0, 0, 134, 3, 0, 68, 5, 160, 0, 0, 134, 2, 0, 68, 5, 160, 0, 0, 5, 3, 0, 68, 5, 160, 0, 0, 5, 2, 0, 68, 5, 160, 0, 0, 134, 3, 0, 82, 5, 162, 0, 0, 134, 2, 0, 82, 5, 162, 0, 0, 134, 74, 0, 82, 5, 162, 0, 0, 5, 3, 0, 82, 5, 162, 0, 0, 5, 2, 0, 82, 5, 162, 0, 0, 5, 74, 0, 82, 5, 162, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 134, 3, 0, 70, 5, 162, 0, 0, 134, 2, 0, 70, 5, 162, 0, 0, 5, 3, 0, 70, 5, 162, 0, 0, 5, 2, 0, 70, 5, 162, 0, 0, 134, 3, 0, 50, 5, 160, 0, 0, 134, 2, 0, 50, 5, 160, 0, 0, 5, 3, 0, 50, 5, 160, 0, 0, 5, 2, 0, 50, 5, 160, 0, 0, 134, 3, 0, 42, 5, 200, 0, 0, 134, 2, 0, 42, 5, 200, 0, 0, 5, 3, 0, 42, 5, 200, 0, 0, 5, 2, 0, 42, 5, 200, 0, 0, 134, 2, 0, 50, 5, 138, 0, 0, 5, 2, 0, 50, 5, 138, 0, 0, 134, 2, 0, 50, 5, 136, 0, 0, 5, 2, 0, 50, 5, 136, 0, 0, 134, 3, 0, 50, 5, 204, 0, 0, 134, 2, 0, 50, 5, 204, 0, 0, 5, 3, 0, 50, 5, 204, 0, 0, 5, 2, 0, 50, 5, 204, 0, 0, 134, 3, 0, 50, 5, 208, 0, 0, 134, 2, 0, 50, 5, 208, 0, 0, 5, 3, 0, 50, 5, 208, 0, 0, 5, 2, 0, 50, 5, 208, 0, 0, 134, 3, 0, 58, 5, 208, 0, 0, 134, 2, 0, 58, 5, 208, 0, 0, 5, 3, 0, 58, 5, 208, 0, 0, 5, 2, 0, 58, 5, 208, 0, 0, 134, 3, 0, 66, 5, 196, 0, 0, 134, 2, 0, 66, 5, 196, 0, 0, 5, 3, 0, 66, 5, 196, 0, 0, 5, 2, 0, 66, 5, 196, 0, 0, 134, 3, 0, 68, 5, 196, 0, 0, 134, 2, 0, 68, 5, 196, 0, 0, 5, 3, 0, 68, 5, 196, 0, 0, 5, 2, 0, 68, 5, 196, 0, 0, 134, 3, 0, 68, 5, 210, 0, 0, 134, 2, 0, 68, 5, 210, 0, 0, 5, 3, 0, 68, 5, 210, 0, 0, 5, 2, 0, 68, 5, 210, 0, 0, 134, 3, 0, 68, 5, 204, 0, 0, 134, 2, 0, 68, 5, 204, 0, 0, 5, 3, 0, 68, 5, 204, 0, 0, 5, 2, 0, 68, 5, 204, 0, 0, 134, 2, 0, 70, 5, 138, 0, 0, 5, 2, 0, 70, 5, 138, 0, 0, 134, 2, 0, 70, 5, 136, 0, 0, 5, 2, 0, 70, 5, 136, 0, 0, 134, 3, 0, 82, 5, 198, 0, 0, 134, 2, 0, 82, 5, 198, 0, 0, 134, 74, 0, 82, 5, 198, 0, 0, 5, 3, 0, 82, 5, 198, 0, 0, 5, 2, 0, 82, 5, 198, 0, 0, 5, 74, 0, 82, 5, 198, 0, 0, 134, 3, 0, 82, 5, 208, 0, 0, 134, 2, 0, 82, 5, 208, 0, 0, 134, 74, 0, 82, 5, 208, 0, 0, 5, 3, 0, 82, 5, 208, 0, 0, 5, 2, 0, 82, 5, 208, 0, 0, 5, 74, 0, 82, 5, 208, 0, 0, 134, 3, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 204, 0, 0, 134, 74, 0, 82, 5, 204, 0, 0, 5, 3, 0, 82, 5, 204, 0, 0, 5, 2, 0, 82, 5, 204, 0, 0, 5, 74, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 150, 0, 0, 5, 2, 0, 82, 5, 150, 0, 0, 134, 3, 0, 42, 5, 196, 0, 0, 134, 2, 0, 42, 5, 196, 0, 0, 5, 3, 0, 42, 5, 196, 0, 0, 5, 2, 0, 42, 5, 196, 0, 0, 134, 3, 0, 50, 5, 196, 0, 0, 134, 2, 0, 50, 5, 196, 0, 0, 5, 3, 0, 50, 5, 196, 0, 0, 5, 2, 0, 50, 5, 196, 0, 0, 134, 3, 0, 58, 5, 196, 0, 0, 134, 2, 0, 58, 5, 196, 0, 0, 5, 3, 0, 58, 5, 196, 0, 0, 5, 2, 0, 58, 5, 196, 0, 0, 134, 3, 0, 70, 5, 196, 0, 0, 134, 2, 0, 70, 5, 196, 0, 0, 5, 3, 0, 70, 5, 196, 0, 0, 5, 2, 0, 70, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 196, 0, 0, 134, 2, 0, 82, 5, 196, 0, 0, 134, 74, 0, 82, 5, 196, 0, 0, 5, 3, 0, 82, 5, 196, 0, 0, 5, 2, 0, 82, 5, 196, 0, 0, 5, 74, 0, 82, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 142, 9, 5, 5, 170, 195, 6, 5, 144, 9, 5, 5, 142, 9, 5, 5, 246, 220, 13, 5, 144, 9, 5, 5, 142, 9, 5, 5, 52, 172, 8, 5, 144, 9, 5, 5, 142, 9, 5, 5, 39, 248, 12, 5, 144, 9, 5, 5, 142, 9, 5, 5, 166, 193, 14, 5, 144, 9, 5, 5, 142, 9, 5, 5, 32, 249, 7, 5, 144, 9, 5, 5, 142, 9, 5, 5, 231, 192, 11, 5, 144, 9, 5, 5, 142, 9, 5, 5, 82, 249, 10, 5, 144, 9, 5, 5, 142, 9, 5, 5, 197, 246, 9, 5, 144, 9, 5, 5, 0, 21, 17, 5, 7, 243, 5, 5, 0, 23, 16, 5, 7, 243, 5, 5, 0, 25, 15, 5, 7, 243, 5, 5, 0, 27, 14, 5, 7, 243, 5, 5, 0, 29, 13, 5, 7, 243, 5, 5, 0, 31, 12, 5, 7, 243, 5, 5, 0, 33, 11, 5, 7, 243, 5, 5, 0, 35, 10, 5, 7, 243, 5, 5, 0, 37, 9, 5, 7, 243, 5, 5, 0, 21, 5, 5, 0, 19, 20, 5, 7, 243, 5, 5, 0, 21, 5, 5, 0, 21, 19, 5, 7, 243, 5, 5, 0, 21, 5, 5, 0, 23, 18, 5, 7, 243, 5, 5, 124, 204, 7, 5, 160, 188, 5, 5, 0, 19, 30, 5, 89, 179, 5, 5, 0, 21, 19, 5, 89, 179, 5, 5, 0, 23, 13, 5, 89, 179, 5, 5, 0, 25, 12, 5, 89, 179, 5, 5, 0, 27, 11, 5, 89, 179, 5, 5, 0, 29, 10, 5, 89, 179, 5, 5, 0, 31, 9, 5, 89, 179, 5, 5, 0, 33, 8, 5, 89, 179, 5, 5, 0, 35, 7, 5, 89, 179, 5, 5, 0, 37, 6, 5, 89, 179, 5, 5, 0, 21, 5, 5, 0, 19, 29, 5, 89, 179, 5, 5, 0, 21, 5, 5, 0, 21, 28, 5, 89, 179, 5, 5, 0, 21, 5, 5, 0, 23, 27, 5, 89, 179, 5, 5, 0, 21, 5, 5, 0, 25, 26, 5, 89, 179, 5, 5, 0, 21, 5, 5, 0, 27, 25, 5, 89, 179, 5, 5, 0, 21, 5, 5, 0, 29, 24, 5, 89, 179, 5, 5, 0, 21, 5, 5, 0, 31, 23, 5, 89, 179, 5, 5, 0, 21, 5, 5, 0, 33, 22, 5, 89, 179, 5, 5, 0, 21, 5, 5, 0, 35, 21, 5, 89, 179, 5, 5, 0, 21, 5, 5, 0, 37, 20, 5, 89, 179, 5, 5, 0, 23, 5, 5, 0, 19, 18, 5, 89, 179, 5, 5, 0, 23, 5, 5, 0, 21, 17, 5, 89, 179, 5, 5, 0, 23, 5, 5, 0, 23, 16, 5, 89, 179, 5, 5, 0, 23, 5, 5, 0, 25, 15, 5, 89, 179, 5, 5, 0, 23, 5, 5, 0, 27, 14, 5, 89, 179, 5, 5, 166, 213, 6, 5, 87, 174, 5, 5, 73, 245, 6, 5, 160, 188, 5, 5, 155, 177, 7, 5, 66, 246, 5, 5, 30, 209, 6, 5, 208, 246, 5, 5, 207, 247, 5, 5, 13, 222, 5, 5, 69, 191, 7, 5, 246, 220, 5, 5, 0, 21, 28, 5, 186, 218, 5, 5, 0, 23, 17, 5, 186, 218, 5, 5, 0, 25, 14, 5, 186, 218, 5, 5, 0, 27, 13, 5, 186, 218, 5, 5, 0, 29, 12, 5, 186, 218, 5, 5, 0, 31, 11, 5, 186, 218, 5, 5, 0, 33, 10, 5, 186, 218, 5, 5, 0, 35, 9, 5, 186, 218, 5, 5, 0, 37, 8, 5, 186, 218, 5, 5, 0, 21, 5, 5, 0, 19, 38, 5, 186, 218, 5, 5, 0, 21, 5, 5, 0, 21, 37, 5, 186, 218, 5, 5, 0, 21, 5, 5, 0, 23, 36, 5, 186, 218, 5, 5, 0, 21, 5, 5, 0, 25, 35, 5, 186, 218, 5, 5, 0, 21, 5, 5, 0, 27, 34, 5, 186, 218, 5, 5, 0, 21, 5, 5, 0, 29, 33, 5, 186, 218, 5, 5, 0, 21, 5, 5, 0, 31, 32, 5, 186, 218, 5, 5, 0, 21, 5, 5, 0, 33, 31, 5, 186, 218, 5, 5, 0, 21, 5, 5, 0, 35, 30, 5, 186, 218, 5, 5, 0, 21, 5, 5, 0, 37, 29, 5, 186, 218, 5, 5, 0, 23, 5, 5, 0, 19, 27, 5, 186, 218, 5, 5, 0, 23, 5, 5, 0, 21, 26, 5, 186, 218, 5, 5, 0, 23, 5, 5, 0, 23, 25, 5, 186, 218, 5, 5, 0, 23, 5, 5, 0, 25, 24, 5, 186, 218, 5, 5, 0, 23, 5, 5, 0, 27, 23, 5, 186, 218, 5, 5, 0, 23, 5, 5, 0, 29, 22, 5, 186, 218, 5, 5, 0, 23, 5, 5, 0, 31, 21, 5, 186, 218, 5, 5, 0, 23, 5, 5, 0, 33, 20, 5, 186, 218, 5, 5, 0, 23, 5, 5, 0, 35, 19, 5, 186, 218, 5, 5, 0, 23, 5, 5, 0, 37, 18, 5, 186, 218, 5, 5, 0, 25, 5, 5, 0, 19, 16, 5, 186, 218, 5, 5, 0, 25, 5, 5, 0, 21, 15, 5, 186, 218, 5, 5, 72, 221, 5, 5, 41, 237, 5, 5, 2, 244, 5, 5, 161, 229, 5, 5, 30, 175, 5, 5, 212, 216, 69, 5, 39, 172, 5, 160, 0, 0, 69, 5, 91, 181, 5, 138, 0, 0, 69, 5, 91, 181, 5, 136, 0, 0, 69, 5, 91, 181, 5, 142, 0, 0, 69, 5, 91, 181, 5, 150, 0, 0, 69, 5, 219, 209, 5, 154, 0, 0, 69, 5, 192, 211, 5, 138, 0, 0, 69, 5, 192, 211, 5, 136, 0, 0, 69, 5, 192, 211, 5, 142, 0, 0, 69, 5, 192, 211, 5, 154, 0, 0, 69, 5, 192, 211, 5, 150, 0, 0, 69, 5, 225, 235, 5, 136, 0, 0, 69, 5, 39, 172, 5, 136, 0, 0, 69, 5, 39, 172, 5, 142, 0, 0, 69, 5, 39, 172, 5, 156, 0, 0, 69, 5, 39, 172, 5, 144, 0, 0, 69, 5, 109, 177, 5, 144, 0, 0, 69, 5, 91, 181, 5, 164, 0, 0, 69, 5, 91, 181, 5, 140, 0, 0, 69, 5, 91, 181, 5, 156, 0, 0, 69, 5, 91, 181, 5, 162, 0, 0, 69, 5, 91, 181, 5, 144, 0, 0, 69, 5, 135, 184, 5, 142, 0, 0, 69, 5, 135, 184, 5, 140, 0, 0, 69, 5, 135, 184, 5, 156, 0, 0, 69, 5, 135, 184, 5, 160, 0, 0, 69, 5, 200, 187, 5, 142, 0, 0, 69, 5, 248, 191, 5, 142, 0, 0, 69, 5, 78, 198, 5, 160, 0, 0, 69, 5, 123, 200, 5, 136, 0, 0, 69, 5, 123, 200, 5, 160, 0, 0, 69, 5, 123, 200, 5, 144, 0, 0, 69, 5, 219, 209, 5, 136, 0, 0, 69, 5, 219, 209, 5, 160, 0, 0, 69, 5, 219, 209, 5, 144, 0, 0, 69, 5, 192, 211, 5, 164, 0, 0, 69, 5, 192, 211, 5, 140, 0, 0, 69, 5, 192, 211, 5, 152, 0, 0, 69, 5, 58, 218, 5, 136, 0, 0, 69, 5, 58, 218, 5, 160, 0, 0, 69, 5, 58, 218, 5, 144, 0, 0, 69, 5, 128, 219, 5, 136, 0, 0, 69, 5, 128, 219, 5, 142, 0, 0, 69, 5, 128, 219, 5, 160, 0, 0, 69, 5, 128, 219, 5, 144, 0, 0, 69, 5, 173, 224, 5, 160, 0, 0, 69, 5, 173, 224, 5, 144, 0, 0, 69, 5, 43, 228, 5, 142, 0, 0, 69, 5, 225, 235, 5, 142, 0, 0, 69, 5, 225, 235, 5, 150, 0, 0, 69, 5, 147, 243, 5, 136, 0, 0, 69, 5, 147, 243, 5, 156, 0, 0, 69, 5, 147, 243, 5, 144, 0, 0, 69, 5, 192, 211, 5, 190, 0, 0, 69, 5, 192, 211, 5, 144, 0, 0, 69, 5, 0, 127, 5, 150, 0, 0, 5, 164, 0, 0, 69, 5, 0, 127, 5, 156, 0, 0, 5, 164, 0, 0, 69, 5, 135, 184, 5, 144, 0, 0, 69, 5, 78, 198, 5, 144, 0, 0, 69, 5, 192, 211, 5, 162, 0, 0, 69, 5, 192, 211, 5, 162, 0, 0, 5, 164, 0, 0, 69, 5, 135, 184, 5, 136, 0, 0, 69, 5, 219, 209, 5, 138, 0, 0, 69, 5, 0, 127, 5, 146, 0, 0, 5, 136, 0, 0, 69, 5, 91, 181, 5, 184, 0, 0, 69, 5, 91, 181, 5, 188, 0, 0, 69, 5, 192, 211, 5, 184, 0, 0, 69, 5, 192, 211, 5, 188, 0, 0, 69, 5, 58, 218, 5, 184, 0, 0, 69, 5, 58, 218, 5, 188, 0, 0, 69, 5, 128, 219, 5, 202, 0, 0, 69, 5, 173, 224, 5, 202, 0, 0, 69, 5, 200, 187, 5, 144, 0, 0, 69, 5, 91, 181, 5, 160, 0, 0, 69, 5, 192, 211, 5, 150, 0, 0, 5, 164, 0, 0, 69, 5, 192, 211, 5, 154, 0, 0, 5, 164, 0, 0, 69, 5, 192, 211, 5, 156, 0, 0, 69, 5, 192, 211, 5, 156, 0, 0, 5, 164, 0, 0, 69, 5, 225, 235, 5, 164, 0, 0, 69, 5, 159, 168, 5, 156, 0, 0, 69, 5, 159, 168, 5, 196, 0, 0, 69, 5, 159, 168, 5, 210, 0, 0, 69, 5, 39, 172, 5, 160, 0, 0, 5, 136, 0, 0, 69, 5, 109, 177, 5, 156, 0, 0, 69, 5, 109, 177, 5, 196, 0, 0, 69, 5, 109, 177, 5, 210, 0, 0, 69, 5, 109, 177, 5, 160, 0, 0, 69, 5, 109, 177, 5, 204, 0, 0, 69, 5, 91, 181, 5, 164, 0, 0, 5, 138, 0, 0, 69, 5, 91, 181, 5, 164, 0, 0, 5, 136, 0, 0, 69, 5, 91, 181, 5, 204, 0, 0, 69, 5, 91, 181, 5, 208, 0, 0, 69, 5, 91, 181, 5, 160, 0, 0, 5, 140, 0, 0, 69, 5, 19, 182, 5, 156, 0, 0, 69, 5, 135, 184, 5, 164, 0, 0, 69, 5, 200, 187, 5, 156, 0, 0, 69, 5, 200, 187, 5, 196, 0, 0, 69, 5, 200, 187, 5, 150, 0, 0, 69, 5, 200, 187, 5, 160, 0, 0, 69, 5, 200, 187, 5, 206, 0, 0, 69, 5, 78, 198, 5, 136, 0, 0, 69, 5, 78, 198, 5, 196, 0, 0, 69, 5, 78, 198, 5, 210, 0, 0, 69, 5, 123, 200, 5, 196, 0, 0, 69, 5, 123, 200, 5, 196, 0, 0, 5, 164, 0, 0, 69, 5, 123, 200, 5, 210, 0, 0, 69, 5, 123, 200, 5, 204, 0, 0, 69, 5, 170, 206, 5, 136, 0, 0, 69, 5, 170, 206, 5, 156, 0, 0, 69, 5, 170, 206, 5, 196, 0, 0, 69, 5, 219, 209, 5, 156, 0, 0, 69, 5, 219, 209, 5, 196, 0, 0, 69, 5, 219, 209, 5, 210, 0, 0, 69, 5, 219, 209, 5, 204, 0, 0, 69, 5, 192, 211, 5, 154, 0, 0, 5, 136, 0, 0, 69, 5, 192, 211, 5, 154, 0, 0, 5, 150, 0, 0, 69, 5, 192, 211, 5, 164, 0, 0, 5, 138, 0, 0, 69, 5, 192, 211, 5, 164, 0, 0, 5, 136, 0, 0, 69, 5, 227, 211, 5, 136, 0, 0, 69, 5, 227, 211, 5, 156, 0, 0, 69, 5, 58, 218, 5, 156, 0, 0, 69, 5, 58, 218, 5, 196, 0, 0, 69, 5, 58, 218, 5, 196, 0, 0, 5, 164, 0, 0, 69, 5, 58, 218, 5, 210, 0, 0, 69, 5, 128, 219, 5, 156, 0, 0, 69, 5, 128, 219, 5, 196, 0, 0, 69, 5, 128, 219, 5, 136, 0, 0, 5, 156, 0, 0, 69, 5, 128, 219, 5, 144, 0, 0, 5, 156, 0, 0, 69, 5, 128, 219, 5, 196, 0, 0, 5, 156, 0, 0, 69, 5, 173, 224, 5, 156, 0, 0, 69, 5, 173, 224, 5, 196, 0, 0, 69, 5, 173, 224, 5, 210, 0, 0, 69, 5, 173, 224, 5, 204, 0, 0, 69, 5, 43, 228, 5, 138, 0, 0, 69, 5, 43, 228, 5, 136, 0, 0, 69, 5, 43, 228, 5, 150, 0, 0, 69, 5, 43, 228, 5, 156, 0, 0, 69, 5, 43, 228, 5, 196, 0, 0, 69, 5, 167, 230, 5, 156, 0, 0, 69, 5, 167, 230, 5, 150, 0, 0, 69, 5, 225, 235, 5, 156, 0, 0, 69, 5, 147, 243, 5, 142, 0, 0, 69, 5, 147, 243, 5, 196, 0, 0, 69, 5, 147, 243, 5, 210, 0, 0, 69, 5, 0, 127, 5, 142, 0, 0, 5, 136, 0, 0, 69, 5, 0, 127, 5, 142, 0, 0, 5, 138, 0, 0, 69, 5, 0, 127, 5, 142, 0, 0, 5, 182, 0, 0, 69, 5, 0, 127, 5, 142, 0, 0, 5, 154, 0, 0, 69, 5, 0, 127, 5, 196, 0, 0, 5, 142, 0, 0, 69, 5, 0, 127, 5, 140, 0, 0, 5, 136, 0, 0, 69, 5, 0, 127, 5, 140, 0, 0, 5, 138, 0, 0, 69, 5, 0, 127, 5, 140, 0, 0, 5, 182, 0, 0, 69, 5, 0, 127, 5, 140, 0, 0, 5, 154, 0, 0, 69, 5, 0, 127, 5, 196, 0, 0, 5, 140, 0, 0, 69, 5, 91, 181, 5, 196, 0, 0, 69, 5, 91, 181, 5, 182, 0, 0, 69, 5, 91, 181, 5, 154, 0, 0, 69, 5, 91, 181, 5, 142, 0, 0, 5, 136, 0, 0, 69, 5, 91, 181, 5, 142, 0, 0, 5, 138, 0, 0, 69, 5, 91, 181, 5, 142, 0, 0, 5, 182, 0, 0, 69, 5, 91, 181, 5, 142, 0, 0, 5, 154, 0, 0, 69, 5, 91, 181, 5, 196, 0, 0, 5, 142, 0, 0, 69, 5, 192, 211, 5, 196, 0, 0, 69, 5, 192, 211, 5, 182, 0, 0, 69, 5, 192, 211, 5, 142, 0, 0, 5, 136, 0, 0, 69, 5, 192, 211, 5, 142, 0, 0, 5, 138, 0, 0, 69, 5, 192, 211, 5, 142, 0, 0, 5, 182, 0, 0, 69, 5, 192, 211, 5, 142, 0, 0, 5, 154, 0, 0, 69, 5, 192, 211, 5, 196, 0, 0, 5, 142, 0, 0, 69, 5, 192, 211, 5, 190, 0, 0, 5, 136, 0, 0, 69, 5, 192, 211, 5, 190, 0, 0, 5, 138, 0, 0, 69, 5, 192, 211, 5, 190, 0, 0, 5, 182, 0, 0, 69, 5, 192, 211, 5, 190, 0, 0, 5, 154, 0, 0, 69, 5, 192, 211, 5, 190, 0, 0, 5, 196, 0, 0, 69, 5, 225, 235, 5, 138, 0, 0, 69, 5, 225, 235, 5, 196, 0, 0, 69, 5, 225, 235, 5, 182, 0, 0, 69, 5, 225, 235, 5, 154, 0, 0, 5, 5, 82, 10, 5, 5, 215, 169, 8, 5, 84, 10, 5, 5, 82, 10, 5, 5, 169, 219, 14, 5, 84, 10, 5, 5, 82, 10, 5, 5, 6, 182, 13, 5, 84, 10, 5, 5, 82, 10, 5, 5, 34, 168, 11, 5, 84, 10, 5, 5, 82, 10, 5, 5, 89, 179, 7, 5, 84, 10, 5, 5, 82, 10, 5, 5, 154, 177, 10, 5, 84, 10, 5, 5, 82, 10, 5, 5, 138, 178, 6, 5, 84, 10, 5, 5, 82, 10, 5, 5, 156, 221, 12, 5, 84, 10, 5, 5, 82, 10, 5, 5, 247, 168, 9, 5, 84, 10, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 0, 42, 156, 5, 5, 0, 12, 3, 7, 0, 12, 3, 198, 161, 64, 3, 198, 129, 65, 3, 0, 234, 134, 3, 0, 3, 198, 129, 1, 3, 0, 234, 134, 3, 4, 3, 0, 234, 134, 2, 0, 50, 156, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 134, 2, 12, 3, 0, 192, 198, 97, 64, 3, 0, 192, 198, 65, 65, 3, 0, 242, 134, 3, 0, 3, 0, 192, 198, 65, 1, 3, 0, 242, 134, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 58, 156, 5, 5, 0, 12, 3, 9, 0, 12, 3, 0, 192, 198, 161, 64, 3, 0, 192, 198, 129, 65, 3, 0, 250, 134, 3, 0, 3, 0, 192, 198, 129, 1, 3, 0, 250, 134, 3, 4, 3, 0, 250, 134, 2, 0, 66, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 0, 192, 198, 225, 64, 3, 0, 192, 198, 193, 65, 3, 255, 255, 0, 66, 134, 3, 0, 3, 0, 192, 198, 193, 1, 3, 255, 255, 0, 66, 134, 3, 4, 3, 255, 255, 0, 66, 134, 2, 0, 68, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 33, 64, 3, 1, 192, 198, 1, 65, 3, 255, 255, 0, 68, 134, 3, 0, 3, 1, 192, 198, 1, 1, 3, 255, 255, 0, 68, 134, 3, 4, 3, 255, 255, 0, 68, 134, 2, 0, 70, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 97, 64, 3, 1, 192, 198, 65, 65, 3, 255, 255, 0, 70, 134, 3, 0, 3, 1, 192, 198, 65, 1, 3, 255, 255, 0, 70, 134, 3, 4, 3, 255, 255, 0, 70, 134, 2, 0, 82, 156, 5, 7, 0, 12, 3, 14, 0, 12, 3, 1, 192, 198, 161, 64, 3, 1, 192, 198, 129, 65, 3, 255, 255, 0, 82, 134, 3, 68, 3, 255, 255, 0, 82, 134, 71, 0, 3, 1, 192, 198, 129, 1, 3, 255, 255, 0, 82, 134, 3, 4, 3, 255, 255, 0, 82, 134, 2, 8, 3, 197, 127, 0, 82, 134, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 0, 42, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 1, 192, 198, 225, 64, 3, 1, 192, 198, 193, 65, 3, 0, 234, 5, 3, 0, 3, 1, 192, 198, 193, 1, 3, 0, 234, 5, 3, 4, 3, 0, 234, 5, 2, 0, 50, 5, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 5, 2, 12, 3, 2, 192, 198, 33, 64, 3, 2, 192, 198, 1, 65, 3, 0, 242, 5, 3, 0, 3, 2, 192, 198, 1, 1, 3, 0, 242, 5, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 58, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 2, 192, 198, 97, 64, 3, 2, 192, 198, 65, 65, 3, 0, 250, 5, 3, 0, 3, 2, 192, 198, 65, 1, 3, 0, 250, 5, 3, 4, 3, 0, 250, 5, 2, 0, 66, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 161, 64, 3, 2, 192, 198, 129, 65, 3, 255, 255, 0, 66, 5, 3, 0, 3, 2, 192, 198, 129, 1, 3, 255, 255, 0, 66, 5, 3, 4, 3, 255, 255, 0, 66, 5, 2, 0, 68, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 225, 64, 3, 2, 192, 198, 193, 65, 3, 255, 255, 0, 68, 5, 3, 0, 3, 2, 192, 198, 193, 1, 3, 255, 255, 0, 68, 5, 3, 4, 3, 255, 255, 0, 68, 5, 2, 0, 70, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 3, 192, 198, 33, 64, 3, 3, 192, 198, 1, 65, 3, 255, 255, 0, 70, 5, 3, 0, 3, 3, 192, 198, 1, 1, 3, 255, 255, 0, 70, 5, 3, 4, 3, 255, 255, 0, 70, 5, 2, 0, 82, 5, 5, 7, 0, 12, 3, 14, 0, 12, 3, 3, 192, 198, 97, 64, 3, 3, 192, 198, 65, 65, 3, 255, 255, 0, 82, 5, 3, 68, 3, 255, 255, 0, 82, 5, 71, 0, 3, 3, 192, 198, 65, 1, 3, 255, 255, 0, 82, 5, 3, 4, 3, 255, 255, 0, 82, 5, 2, 8, 3, 197, 127, 0, 82, 5, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 82, 134, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 5, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 82, 5, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 42, 196, 162, 5, 0, 12, 3, 8, 0, 12, 3, 3, 192, 198, 194, 64, 3, 3, 192, 198, 130, 65, 3, 197, 162, 0, 3, 3, 192, 198, 130, 1, 3, 197, 162, 4, 3, 0, 192, 197, 98, 5, 42, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 66, 64, 3, 4, 192, 198, 2, 65, 3, 0, 192, 197, 162, 0, 3, 4, 192, 198, 2, 1, 3, 0, 192, 197, 162, 4, 3, 0, 192, 197, 226, 156, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 194, 64, 3, 4, 192, 198, 130, 65, 3, 1, 192, 197, 34, 0, 3, 4, 192, 198, 130, 1, 3, 1, 192, 197, 34, 4, 3, 1, 192, 197, 98, 5, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 66, 64, 3, 5, 192, 198, 2, 65, 3, 1, 192, 197, 162, 0, 3, 5, 192, 198, 2, 1, 3, 1, 192, 197, 162, 4, 3, 1, 192, 197, 226, 156, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 194, 64, 3, 5, 192, 198, 130, 65, 3, 2, 192, 197, 34, 0, 3, 5, 192, 198, 130, 1, 3, 2, 192, 197, 34, 4, 3, 2, 192, 197, 98, 5, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 66, 64, 3, 6, 192, 198, 2, 65, 3, 2, 192, 197, 162, 0, 3, 6, 192, 198, 2, 1, 3, 2, 192, 197, 162, 4, 3, 2, 192, 197, 226, 156, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 194, 64, 3, 6, 192, 198, 130, 65, 3, 3, 192, 197, 34, 0, 3, 6, 192, 198, 130, 1, 3, 3, 192, 197, 34, 4, 3, 3, 192, 197, 98, 5, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 7, 192, 198, 66, 64, 3, 7, 192, 198, 2, 65, 3, 3, 192, 197, 162, 0, 3, 7, 192, 198, 2, 1, 3, 3, 192, 197, 162, 4, 3, 3, 192, 197, 226, 156, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 4, 192, 197, 162, 12, 3, 7, 192, 198, 194, 64, 3, 7, 192, 198, 130, 65, 3, 4, 192, 197, 34, 0, 3, 7, 192, 198, 130, 1, 3, 4, 192, 197, 34, 4, 3, 4, 192, 197, 98, 5, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 5, 192, 197, 98, 12, 3, 8, 192, 198, 66, 64, 3, 8, 192, 198, 2, 65, 3, 4, 192, 197, 226, 0, 3, 8, 192, 198, 2, 1, 3, 4, 192, 197, 226, 4, 3, 5, 192, 197, 34, 156, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 8, 192, 198, 194, 64, 3, 8, 192, 198, 130, 65, 3, 5, 192, 197, 162, 0, 3, 8, 192, 198, 130, 1, 3, 5, 192, 197, 162, 4, 3, 5, 192, 197, 226, 5, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 9, 192, 198, 66, 64, 3, 9, 192, 198, 2, 65, 3, 6, 192, 197, 34, 0, 3, 9, 192, 198, 2, 1, 3, 6, 192, 197, 34, 4, 3, 6, 192, 197, 98, 156, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 34, 12, 3, 9, 192, 198, 194, 64, 3, 9, 192, 198, 130, 65, 3, 6, 192, 197, 162, 0, 3, 9, 192, 198, 130, 1, 3, 6, 192, 197, 162, 4, 3, 6, 192, 197, 226, 5, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 226, 12, 3, 10, 192, 198, 66, 64, 3, 10, 192, 198, 2, 65, 3, 7, 192, 197, 98, 0, 3, 10, 192, 198, 2, 1, 3, 7, 192, 197, 98, 4, 3, 7, 192, 197, 162, 156, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 10, 192, 198, 194, 64, 3, 10, 192, 198, 130, 65, 3, 8, 192, 197, 34, 0, 3, 10, 192, 198, 130, 1, 3, 8, 192, 197, 34, 4, 3, 8, 192, 197, 98, 5, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 66, 64, 3, 11, 192, 198, 2, 65, 3, 8, 192, 197, 162, 0, 3, 11, 192, 198, 2, 1, 3, 8, 192, 197, 162, 4, 3, 8, 192, 197, 226, 156, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 194, 64, 3, 11, 192, 198, 130, 65, 3, 9, 192, 197, 34, 0, 3, 11, 192, 198, 130, 1, 3, 9, 192, 197, 34, 4, 3, 9, 192, 197, 98, 5, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 66, 64, 3, 12, 192, 198, 2, 65, 3, 9, 192, 197, 162, 0, 3, 12, 192, 198, 2, 1, 3, 9, 192, 197, 162, 4, 3, 9, 192, 197, 226, 156, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 194, 64, 3, 12, 192, 198, 130, 65, 3, 10, 192, 197, 34, 0, 3, 12, 192, 198, 130, 1, 3, 10, 192, 197, 34, 4, 3, 10, 192, 197, 98, 5, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 66, 64, 3, 13, 192, 198, 2, 65, 3, 10, 192, 197, 162, 0, 3, 13, 192, 198, 2, 1, 3, 10, 192, 197, 162, 4, 3, 10, 192, 197, 226, 156, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 194, 64, 3, 13, 192, 198, 130, 65, 3, 12, 192, 197, 34, 0, 3, 13, 192, 198, 130, 1, 3, 12, 192, 197, 34, 4, 3, 12, 192, 197, 98, 5, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 66, 64, 3, 14, 192, 198, 2, 65, 3, 12, 192, 197, 162, 0, 3, 14, 192, 198, 2, 1, 3, 12, 192, 197, 162, 4, 3, 12, 192, 197, 226, 156, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 194, 64, 3, 14, 192, 198, 130, 65, 3, 13, 192, 197, 34, 0, 3, 14, 192, 198, 130, 1, 3, 13, 192, 197, 34, 4, 3, 13, 192, 197, 98, 5, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 66, 64, 3, 15, 192, 198, 2, 65, 3, 13, 192, 197, 162, 0, 3, 15, 192, 198, 2, 1, 3, 13, 192, 197, 162, 4, 3, 13, 192, 197, 226, 156, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 194, 64, 3, 15, 192, 198, 130, 65, 3, 14, 192, 197, 34, 0, 3, 15, 192, 198, 130, 1, 3, 14, 192, 197, 34, 4, 3, 14, 192, 197, 98, 5, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 66, 64, 3, 16, 192, 198, 2, 65, 3, 14, 192, 197, 162, 0, 3, 16, 192, 198, 2, 1, 3, 14, 192, 197, 162, 4, 3, 14, 192, 197, 226, 156, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 194, 64, 3, 16, 192, 198, 130, 65, 3, 15, 192, 197, 34, 0, 3, 16, 192, 198, 130, 1, 3, 15, 192, 197, 34, 4, 3, 15, 192, 197, 98, 5, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 66, 64, 3, 17, 192, 198, 2, 65, 3, 15, 192, 197, 162, 0, 3, 17, 192, 198, 2, 1, 3, 15, 192, 197, 162, 4, 3, 15, 192, 197, 226, 156, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 194, 64, 3, 17, 192, 198, 130, 65, 3, 16, 192, 197, 34, 0, 3, 17, 192, 198, 130, 1, 3, 16, 192, 197, 34, 4, 3, 16, 192, 197, 98, 5, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 66, 64, 3, 18, 192, 198, 2, 65, 3, 16, 192, 197, 162, 0, 3, 18, 192, 198, 2, 1, 3, 16, 192, 197, 162, 4, 3, 16, 192, 197, 226, 156, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 194, 64, 3, 18, 192, 198, 130, 65, 3, 17, 192, 197, 34, 0, 3, 18, 192, 198, 130, 1, 3, 17, 192, 197, 34, 4, 3, 17, 192, 197, 98, 5, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 66, 64, 3, 19, 192, 198, 2, 65, 3, 17, 192, 197, 162, 0, 3, 19, 192, 198, 2, 1, 3, 17, 192, 197, 162, 4, 3, 17, 192, 197, 226, 156, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 194, 64, 3, 19, 192, 198, 130, 65, 3, 18, 192, 197, 34, 0, 3, 19, 192, 198, 130, 1, 3, 18, 192, 197, 34, 4, 3, 18, 192, 197, 98, 5, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 20, 192, 198, 66, 64, 3, 20, 192, 198, 2, 65, 3, 18, 192, 197, 162, 0, 3, 20, 192, 198, 2, 1, 3, 18, 192, 197, 162, 4, 3, 18, 192, 197, 226, 156, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 20, 192, 197, 162, 12, 3, 20, 192, 198, 194, 64, 3, 20, 192, 198, 130, 65, 3, 20, 192, 197, 34, 0, 3, 20, 192, 198, 130, 1, 3, 20, 192, 197, 34, 4, 3, 20, 192, 197, 98, 5, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 21, 192, 197, 98, 12, 3, 21, 192, 198, 66, 64, 3, 21, 192, 198, 2, 65, 3, 20, 192, 197, 226, 0, 3, 21, 192, 198, 2, 1, 3, 20, 192, 197, 226, 4, 3, 21, 192, 197, 34, 156, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 34, 12, 3, 21, 192, 198, 194, 64, 3, 21, 192, 198, 130, 65, 3, 21, 192, 197, 162, 0, 3, 21, 192, 198, 130, 1, 3, 21, 192, 197, 162, 4, 3, 21, 192, 197, 226, 5, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 226, 12, 3, 22, 192, 198, 66, 64, 3, 22, 192, 198, 2, 65, 3, 22, 192, 197, 98, 0, 3, 22, 192, 198, 2, 1, 3, 22, 192, 197, 98, 4, 3, 22, 192, 197, 162, 156, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 23, 192, 197, 162, 12, 3, 22, 192, 198, 194, 64, 3, 22, 192, 198, 130, 65, 3, 23, 192, 197, 34, 0, 3, 22, 192, 198, 130, 1, 3, 23, 192, 197, 34, 4, 3, 23, 192, 197, 98, 5, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 24, 192, 197, 98, 12, 3, 23, 192, 198, 66, 64, 3, 23, 192, 198, 2, 65, 3, 23, 192, 197, 226, 0, 3, 23, 192, 198, 2, 1, 3, 23, 192, 197, 226, 4, 3, 24, 192, 197, 34, 156, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 23, 192, 198, 194, 64, 3, 23, 192, 198, 130, 65, 3, 25, 192, 197, 34, 0, 3, 23, 192, 198, 130, 1, 3, 25, 192, 197, 34, 4, 3, 25, 192, 197, 98, 5, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 66, 64, 3, 24, 192, 198, 2, 65, 3, 25, 192, 197, 162, 0, 3, 24, 192, 198, 2, 1, 3, 25, 192, 197, 162, 4, 3, 25, 192, 197, 226, 156, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 194, 64, 3, 24, 192, 198, 130, 65, 3, 26, 192, 197, 34, 0, 3, 24, 192, 198, 130, 1, 3, 26, 192, 197, 34, 4, 3, 26, 192, 197, 98, 5, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 66, 64, 3, 25, 192, 198, 2, 65, 3, 26, 192, 197, 162, 0, 3, 25, 192, 198, 2, 1, 3, 26, 192, 197, 162, 4, 3, 26, 192, 197, 226, 156, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 194, 64, 3, 25, 192, 198, 130, 65, 3, 27, 192, 197, 34, 0, 3, 25, 192, 198, 130, 1, 3, 27, 192, 197, 34, 4, 3, 27, 192, 197, 98, 5, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 66, 64, 3, 26, 192, 198, 2, 65, 3, 27, 192, 197, 162, 0, 3, 26, 192, 198, 2, 1, 3, 27, 192, 197, 162, 4, 3, 27, 192, 197, 226, 156, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 194, 64, 3, 26, 192, 198, 130, 65, 3, 28, 192, 197, 34, 0, 3, 26, 192, 198, 130, 1, 3, 28, 192, 197, 34, 4, 3, 28, 192, 197, 98, 5, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 66, 64, 3, 27, 192, 198, 2, 65, 3, 28, 192, 197, 162, 0, 3, 27, 192, 198, 2, 1, 3, 28, 192, 197, 162, 4, 3, 28, 192, 197, 226, 29, 0, 197, 35, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 227, 64, 3, 27, 192, 198, 131, 65, 3, 29, 192, 197, 131, 0, 3, 27, 192, 198, 131, 1, 3, 29, 192, 197, 131, 4, 3, 29, 192, 197, 227, 30, 0, 197, 67, 5, 0, 12, 3, 9, 0, 12, 3, 28, 192, 198, 163, 64, 3, 28, 192, 198, 67, 65, 3, 30, 192, 197, 163, 0, 3, 28, 192, 198, 67, 1, 3, 30, 192, 197, 163, 4, 3, 31, 192, 197, 3, 156, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 31, 192, 197, 226, 12, 3, 29, 192, 198, 66, 64, 3, 29, 192, 198, 2, 65, 3, 31, 192, 197, 98, 0, 3, 29, 192, 198, 2, 1, 3, 31, 192, 197, 98, 4, 3, 31, 192, 197, 162, 5, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 32, 192, 197, 162, 12, 3, 29, 192, 198, 194, 64, 3, 29, 192, 198, 130, 65, 3, 32, 192, 197, 34, 0, 3, 29, 192, 198, 130, 1, 3, 32, 192, 197, 34, 4, 3, 32, 192, 197, 98, 32, 0, 197, 227, 6, 0, 8, 3, 12, 0, 8, 3, 34, 192, 197, 3, 12, 3, 30, 192, 198, 99, 64, 3, 30, 192, 198, 3, 65, 3, 33, 192, 197, 67, 0, 3, 30, 192, 198, 3, 1, 3, 33, 192, 197, 67, 4, 3, 33, 192, 197, 163, 34, 0, 197, 99, 6, 0, 8, 3, 12, 0, 8, 3, 35, 192, 197, 131, 12, 3, 31, 192, 198, 35, 64, 3, 30, 192, 198, 195, 65, 3, 34, 192, 197, 195, 0, 3, 30, 192, 198, 195, 1, 3, 34, 192, 197, 195, 4, 3, 35, 192, 197, 35, 252, 173, 5, 5, 48, 0, 51, 150, 63, 192, 197, 2, 130, 172, 5, 5, 48, 0, 135, 101, 63, 192, 197, 66, 118, 247, 5, 5, 48, 0, 134, 94, 63, 192, 197, 130, 41, 0, 198, 161, 0, 0, 218, 0, 40, 2, 141, 1, 102, 30, 199, 0, 180, 30, 100, 0, 212, 30, 50, 0, 226, 30, 25, 0, 246, 30, 14, 0, 246, 30, 113, 192, 197, 98, 248, 30, 113, 192, 197, 162, 42, 33, 255, 255, 78, 198, 69, 5, 43, 33, 255, 255, 69, 127, 196, 146, 226, 30, 112, 192, 197, 131, 242, 30, 112, 192, 197, 226, 244, 30, 113, 192, 197, 34, 218, 30, 12, 0, 218, 30, 111, 192, 197, 3, 220, 30, 111, 192, 197, 99, 222, 30, 111, 192, 197, 195, 224, 30, 112, 192, 197, 35, 212, 30, 109, 192, 197, 227, 214, 30, 110, 192, 197, 67, 216, 30, 110, 192, 197, 163, 194, 30, 23, 0, 204, 30, 12, 0, 204, 30, 108, 192, 197, 162, 206, 30, 108, 192, 197, 226, 208, 30, 109, 192, 197, 35, 210, 30, 109, 192, 197, 131, 194, 30, 107, 192, 197, 131, 196, 30, 107, 192, 197, 227, 198, 30, 108, 192, 197, 67, 186, 30, 12, 0, 186, 30, 106, 192, 197, 66, 188, 30, 106, 192, 197, 130, 190, 30, 106, 192, 197, 195, 192, 30, 107, 192, 197, 35, 180, 30, 105, 192, 197, 67, 182, 30, 105, 192, 197, 163, 184, 30, 106, 192, 197, 2, 142, 30, 50, 0, 166, 30, 23, 0, 172, 30, 12, 0, 172, 30, 103, 192, 197, 195, 174, 30, 104, 192, 197, 35, 176, 30, 104, 192, 197, 131, 178, 30, 104, 192, 197, 227, 166, 30, 102, 192, 197, 163, 168, 30, 103, 192, 197, 3, 170, 30, 103, 192, 197, 99, 148, 30, 14, 0, 148, 30, 102, 192, 197, 2, 160, 30, 255, 255, 69, 127, 196, 196, 162, 30, 255, 255, 69, 127, 196, 182, 164, 30, 102, 192, 197, 67, 142, 30, 101, 192, 197, 66, 144, 30, 101, 192, 197, 130, 146, 30, 101, 192, 197, 194, 128, 30, 23, 0, 134, 30, 12, 0, 134, 30, 100, 192, 197, 66, 136, 30, 100, 192, 197, 130, 138, 30, 100, 192, 197, 194, 140, 30, 101, 192, 197, 2, 128, 30, 99, 192, 197, 130, 130, 30, 99, 192, 197, 194, 132, 30, 100, 192, 197, 2, 108, 30, 9, 0, 108, 30, 98, 192, 197, 194, 110, 30, 99, 192, 197, 2, 112, 30, 99, 192, 197, 66, 102, 30, 97, 192, 197, 195, 104, 30, 98, 192, 197, 35, 106, 30, 98, 192, 197, 130, 42, 30, 98, 0, 74, 30, 48, 0, 88, 30, 23, 0, 94, 30, 12, 0, 94, 30, 96, 192, 197, 162, 96, 30, 96, 192, 197, 226, 98, 30, 97, 192, 197, 34, 100, 30, 97, 192, 197, 99, 88, 30, 95, 192, 197, 194, 90, 30, 96, 192, 197, 2, 92, 30, 96, 192, 197, 67, 80, 30, 12, 0, 80, 30, 94, 192, 197, 131, 82, 30, 94, 192, 197, 227, 84, 30, 95, 192, 197, 66, 86, 30, 95, 192, 197, 130, 74, 30, 93, 192, 197, 130, 76, 30, 93, 192, 197, 195, 78, 30, 94, 192, 197, 35, 60, 30, 23, 0, 66, 30, 12, 0, 66, 30, 92, 192, 197, 130, 68, 30, 92, 192, 197, 194, 70, 30, 93, 192, 197, 2, 72, 30, 93, 192, 197, 66, 60, 30, 91, 192, 197, 194, 62, 30, 92, 192, 197, 2, 64, 30, 92, 192, 197, 66, 52, 30, 12, 0, 52, 30, 90, 192, 197, 162, 54, 30, 90, 192, 197, 226, 56, 30, 91, 192, 197, 35, 58, 30, 91, 192, 197, 130, 42, 30, 89, 192, 197, 226, 48, 30, 90, 192, 197, 34, 50, 30, 90, 192, 197, 98, 14, 30, 48, 0, 28, 30, 23, 0, 34, 30, 12, 0, 34, 30, 88, 192, 197, 226, 36, 30, 89, 192, 197, 34, 38, 30, 89, 192, 197, 98, 40, 30, 89, 192, 197, 162, 28, 30, 88, 192, 197, 3, 30, 30, 88, 192, 197, 98, 32, 30, 88, 192, 197, 162, 20, 30, 12, 0, 20, 30, 86, 192, 197, 195, 22, 30, 87, 192, 197, 35, 24, 30, 87, 192, 197, 130, 26, 30, 87, 192, 197, 194, 14, 30, 86, 192, 197, 2, 16, 30, 86, 192, 197, 66, 18, 30, 86, 192, 197, 130, 0, 30, 24, 0, 6, 30, 12, 0, 6, 30, 84, 192, 197, 226, 8, 30, 85, 192, 197, 35, 10, 30, 85, 192, 197, 130, 12, 30, 85, 192, 197, 194, 0, 30, 255, 255, 69, 127, 196, 200, 2, 30, 84, 192, 197, 98, 4, 30, 84, 192, 197, 162, 46, 2, 9, 0, 46, 2, 83, 192, 197, 130, 48, 2, 83, 192, 197, 195, 50, 2, 84, 192, 197, 34, 40, 2, 82, 192, 197, 130, 42, 2, 82, 192, 197, 195, 44, 2, 83, 192, 197, 35, 28, 1, 202, 0, 120, 1, 105, 0, 248, 1, 54, 0, 14, 2, 27, 0, 24, 2, 16, 0, 24, 2, 81, 192, 197, 194, 26, 2, 82, 192, 197, 2, 30, 2, 82, 192, 197, 66, 38, 2, 240, 127, 69, 127, 196, 156, 4, 3, 77, 192, 197, 227, 14, 2, 81, 192, 197, 2, 16, 2, 81, 192, 197, 66, 18, 2, 81, 192, 197, 130, 2, 2, 13, 0, 2, 2, 255, 255, 69, 127, 196, 188, 4, 2, 80, 192, 197, 66, 6, 2, 80, 192, 197, 130, 12, 2, 80, 192, 197, 194, 248, 1, 79, 192, 197, 162, 250, 1, 79, 192, 197, 227, 0, 2, 255, 255, 69, 127, 196, 184, 222, 1, 23, 0, 232, 1, 12, 0, 232, 1, 78, 192, 197, 130, 234, 1, 78, 192, 197, 194, 236, 1, 79, 192, 197, 3, 244, 1, 79, 192, 197, 98, 222, 1, 77, 192, 197, 131, 224, 1, 77, 192, 197, 227, 230, 1, 78, 192, 197, 66, 125, 1, 13, 0, 125, 1, 76, 192, 197, 194, 160, 1, 77, 192, 197, 2, 205, 1, 255, 255, 69, 127, 196, 144, 209, 1, 77, 192, 197, 66, 120, 1, 76, 192, 197, 2, 121, 1, 76, 192, 197, 66, 123, 1, 76, 192, 197, 130, 76, 1, 48, 0, 92, 1, 23, 0, 98, 1, 12, 0, 98, 1, 75, 192, 197, 2, 100, 1, 75, 192, 197, 66, 116, 1, 75, 192, 197, 130, 118, 1, 75, 192, 197, 194, 92, 1, 74, 192, 197, 66, 94, 1, 74, 192, 197, 130, 96, 1, 74, 192, 197, 194, 84, 1, 12, 0, 84, 1, 73, 192, 197, 66, 86, 1, 73, 192, 197, 130, 88, 1, 73, 192, 197, 194, 90, 1, 74, 192, 197, 2, 76, 1, 72, 192, 197, 130, 78, 1, 72, 192, 197, 194, 80, 1, 73, 192, 197, 2, 54, 1, 23, 0, 61, 1, 12, 0, 61, 1, 71, 192, 197, 130, 67, 1, 71, 192, 197, 194, 69, 1, 72, 192, 197, 2, 71, 1, 72, 192, 197, 66, 54, 1, 70, 192, 197, 194, 57, 1, 71, 192, 197, 2, 59, 1, 71, 192, 197, 66, 34, 1, 9, 0, 34, 1, 70, 192, 197, 2, 36, 1, 70, 192, 197, 66, 52, 1, 70, 192, 197, 130, 28, 1, 69, 192, 197, 66, 30, 1, 69, 192, 197, 130, 32, 1, 69, 192, 197, 194, 196, 0, 100, 0, 221, 0, 51, 0, 12, 1, 23, 0, 20, 1, 12, 0, 20, 1, 68, 192, 197, 66, 22, 1, 68, 192, 197, 130, 24, 1, 68, 192, 197, 194, 26, 1, 69, 192, 197, 2, 12, 1, 67, 192, 197, 130, 14, 1, 67, 192, 197, 194, 18, 1, 68, 192, 197, 2, 4, 1, 13, 0, 4, 1, 255, 255, 69, 127, 196, 162, 6, 1, 66, 192, 197, 194, 8, 1, 67, 192, 197, 2, 10, 1, 67, 192, 197, 66, 221, 0, 66, 192, 197, 130, 0, 1, 255, 255, 69, 127, 196, 164, 2, 1, 255, 255, 69, 127, 196, 140, 203, 0, 23, 0, 211, 0, 12, 0, 211, 0, 65, 192, 197, 130, 212, 0, 65, 192, 197, 194, 213, 0, 66, 192, 197, 2, 214, 0, 66, 192, 197, 66, 203, 0, 64, 192, 197, 194, 209, 0, 65, 192, 197, 2, 210, 0, 65, 192, 197, 66, 200, 0, 9, 0, 200, 0, 64, 192, 197, 2, 201, 0, 64, 192, 197, 66, 202, 0, 64, 192, 197, 130, 196, 0, 255, 255, 69, 127, 196, 150, 197, 0, 255, 255, 69, 127, 196, 146, 199, 0, 63, 192, 197, 194, 79, 0, 62, 0, 88, 0, 30, 0, 192, 0, 16, 0, 192, 0, 255, 255, 69, 127, 196, 138, 193, 0, 255, 255, 69, 127, 196, 136, 194, 0, 255, 255, 69, 127, 196, 142, 195, 0, 255, 255, 69, 127, 196, 154, 88, 0, 255, 255, 167, 230, 69, 5, 89, 0, 255, 255, 225, 235, 69, 5, 90, 0, 255, 255, 147, 243, 69, 5, 82, 0, 16, 0, 82, 0, 255, 255, 58, 218, 69, 5, 83, 0, 255, 255, 128, 219, 69, 5, 84, 0, 255, 255, 173, 224, 69, 5, 87, 0, 255, 255, 43, 228, 69, 5, 79, 0, 255, 255, 192, 211, 69, 5, 80, 0, 255, 255, 227, 211, 69, 5, 81, 0, 255, 255, 62, 214, 69, 5, 71, 0, 30, 0, 75, 0, 16, 0, 75, 0, 255, 255, 78, 198, 69, 5, 76, 0, 255, 255, 123, 200, 69, 5, 77, 0, 255, 255, 170, 206, 69, 5, 78, 0, 255, 255, 219, 209, 69, 5, 71, 0, 255, 255, 135, 184, 69, 5, 72, 0, 255, 255, 200, 187, 69, 5, 74, 0, 255, 255, 248, 191, 69, 5, 68, 0, 12, 0, 68, 0, 255, 255, 109, 177, 69, 5, 69, 0, 255, 255, 91, 181, 69, 5, 70, 0, 255, 255, 19, 182, 69, 5, 65, 0, 255, 255, 0, 127, 69, 5, 66, 0, 255, 255, 159, 168, 69, 5, 67, 0, 255, 255, 39, 172, 69, 5, 118, 1, 65, 0, 73, 0, 74, 0, 76, 0, 77, 0, 85, 0, 87, 0, 91, 0, 192, 0, 198, 0, 199, 0, 204, 0, 209, 0, 215, 0, 221, 0, 222, 0, 0, 1, 1, 1, 2, 1, 3, 1, 4, 1, 5, 1, 6, 1, 7, 1, 8, 1, 9, 1, 10, 1, 11, 1, 12, 1, 13, 1, 14, 1, 15, 1, 18, 1, 19, 1, 20, 1, 21, 1, 22, 1, 23, 1, 24, 1, 25, 1, 26, 1, 27, 1, 28, 1, 29, 1, 30, 1, 31, 1, 32, 1, 33, 1, 34, 1, 35, 1, 36, 1, 37, 1, 52, 1, 53, 1, 54, 1, 55, 1, 57, 1, 58, 1, 59, 1, 60, 1, 61, 1, 62, 1, 67, 1, 68, 1, 69, 1, 70, 1, 71, 1, 72, 1, 76, 1, 77, 1, 78, 1, 79, 1, 80, 1, 81, 1, 84, 1, 85, 1, 86, 1, 87, 1, 88, 1, 89, 1, 90, 1, 91, 1, 92, 1, 93, 1, 94, 1, 95, 1, 96, 1, 97, 1, 98, 1, 99, 1, 100, 1, 101, 1, 116, 1, 117, 1, 118, 1, 119, 1, 120, 1, 122, 1, 123, 1, 124, 1, 125, 1, 126, 1, 160, 1, 161, 1, 205, 1, 206, 1, 209, 1, 210, 1, 222, 1, 223, 1, 224, 1, 225, 1, 230, 1, 231, 1, 232, 1, 233, 1, 234, 1, 235, 1, 236, 1, 237, 1, 244, 1, 245, 1, 248, 1, 249, 1, 250, 1, 251, 1, 0, 2, 1, 2, 2, 2, 3, 2, 4, 2, 5, 2, 6, 2, 7, 2, 12, 2, 13, 2, 14, 2, 15, 2, 16, 2, 17, 2, 18, 2, 19, 2, 24, 2, 25, 2, 26, 2, 27, 2, 30, 2, 31, 2, 38, 2, 39, 2, 40, 2, 41, 2, 42, 2, 43, 2, 44, 2, 45, 2, 46, 2, 47, 2, 48, 2, 49, 2, 50, 2, 51, 2, 0, 30, 1, 30, 2, 30, 3, 30, 4, 30, 5, 30, 6, 30, 7, 30, 8, 30, 9, 30, 10, 30, 11, 30, 12, 30, 13, 30, 14, 30, 15, 30, 16, 30, 17, 30, 18, 30, 19, 30, 20, 30, 21, 30, 22, 30, 23, 30, 24, 30, 25, 30, 26, 30, 27, 30, 28, 30, 29, 30, 30, 30, 31, 30, 32, 30, 33, 30, 34, 30, 35, 30, 36, 30, 37, 30, 38, 30, 39, 30, 40, 30, 41, 30, 42, 30, 43, 30, 48, 30, 49, 30, 50, 30, 51, 30, 52, 30, 53, 30, 54, 30, 55, 30, 56, 30, 57, 30, 58, 30, 59, 30, 60, 30, 61, 30, 62, 30, 63, 30, 64, 30, 65, 30, 66, 30, 67, 30, 68, 30, 69, 30, 70, 30, 71, 30, 72, 30, 73, 30, 74, 30, 75, 30, 76, 30, 77, 30, 78, 30, 79, 30, 80, 30, 81, 30, 82, 30, 83, 30, 84, 30, 85, 30, 86, 30, 87, 30, 88, 30, 89, 30, 90, 30, 91, 30, 92, 30, 93, 30, 94, 30, 95, 30, 96, 30, 97, 30, 98, 30, 99, 30, 100, 30, 101, 30, 102, 30, 103, 30, 104, 30, 105, 30, 106, 30, 107, 30, 108, 30, 109, 30, 110, 30, 111, 30, 112, 30, 113, 30, 128, 30, 129, 30, 130, 30, 131, 30, 132, 30, 133, 30, 134, 30, 135, 30, 136, 30, 137, 30, 138, 30, 139, 30, 140, 30, 141, 30, 142, 30, 143, 30, 144, 30, 145, 30, 146, 30, 147, 30, 148, 30, 149, 30, 160, 30, 161, 30, 162, 30, 163, 30, 164, 30, 165, 30, 166, 30, 167, 30, 168, 30, 169, 30, 170, 30, 171, 30, 172, 30, 173, 30, 174, 30, 175, 30, 176, 30, 177, 30, 178, 30, 179, 30, 180, 30, 181, 30, 182, 30, 183, 30, 184, 30, 185, 30, 186, 30, 187, 30, 188, 30, 189, 30, 190, 30, 191, 30, 192, 30, 193, 30, 194, 30, 195, 30, 196, 30, 197, 30, 198, 30, 199, 30, 204, 30, 205, 30, 206, 30, 207, 30, 208, 30, 209, 30, 210, 30, 211, 30, 212, 30, 213, 30, 214, 30, 215, 30, 216, 30, 217, 30, 218, 30, 219, 30, 220, 30, 221, 30, 222, 30, 223, 30, 224, 30, 225, 30, 226, 30, 227, 30, 242, 30, 243, 30, 244, 30, 245, 30, 246, 30, 247, 30, 248, 30, 249, 30, 42, 33, 44, 33, 134, 94, 135, 94, 135, 101, 136, 101, 51, 150, 52, 150, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 28, 4, 185, 76, 185, 80, 185, 84, 30, 4, 186, 96, 185, 100, 185, 104, 32, 4, 34, 4, 36, 4, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 38, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 40, 4, 168, 60, 168, 64, 168, 68, 42, 4, 168, 76, 168, 80, 168, 84, 44, 4, 168, 96, 168, 100, 168, 104, 46, 4, 48, 4, 50, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 52, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 89, 56, 57, 56, 217, 57, 121, 58, 57, 58, 25, 58, 0, 8, 217, 66, 121, 72, 57, 72, 54, 4, 58, 74, 89, 88, 57, 88, 218, 89, 58, 90, 217, 68, 122, 114, 89, 120, 57, 120, 217, 121, 121, 122, 57, 122, 136, 14, 185, 122, 89, 152, 57, 152, 217, 153, 56, 4, 153, 169, 185, 176, 2, 8, 72, 56, 40, 56, 200, 57, 104, 58, 40, 58, 8, 58, 4, 8, 200, 66, 104, 72, 40, 72, 58, 4, 40, 74, 72, 88, 40, 88, 200, 89, 40, 90, 200, 68, 104, 114, 72, 120, 40, 120, 200, 121, 104, 122, 40, 122, 128, 14, 168, 122, 72, 152, 40, 152, 200, 153, 60, 4, 136, 169, 168, 176, 40, 170, 25, 56, 8, 56, 185, 57, 168, 57, 62, 4, 64, 4, 153, 65, 136, 65, 217, 65, 200, 65, 153, 66, 136, 66, 249, 65, 232, 65, 249, 69, 232, 69, 25, 71, 8, 71, 25, 72, 8, 72, 186, 73, 168, 73, 154, 74, 136, 74, 66, 4, 68, 4, 89, 72, 72, 72, 217, 81, 200, 81, 185, 81, 168, 81, 153, 82, 136, 82, 217, 82, 200, 82, 217, 85, 200, 85, 25, 87, 8, 87, 122, 90, 104, 90, 25, 88, 8, 88, 186, 89, 168, 89, 70, 4, 72, 4, 154, 90, 168, 92, 6, 8, 8, 8, 218, 97, 200, 97, 217, 102, 200, 102, 168, 132, 153, 105, 136, 105, 217, 106, 200, 106, 249, 105, 232, 105, 57, 107, 40, 107, 25, 107, 8, 107, 25, 112, 8, 112, 74, 4, 76, 4, 57, 112, 40, 112, 10, 8, 185, 116, 168, 116, 25, 120, 8, 120, 185, 121, 168, 121, 89, 122, 72, 122, 12, 8, 14, 8, 153, 137, 136, 137, 217, 138, 200, 138, 249, 137, 232, 137, 154, 141, 136, 141, 218, 141, 200, 141, 218, 142, 200, 142, 250, 141, 232, 141, 217, 146, 200, 146, 249, 145, 232, 145, 185, 148, 168, 148, 121, 154, 104, 154, 25, 152, 8, 152, 185, 153, 168, 153, 25, 154, 8, 154, 89, 154, 72, 154, 78, 4, 80, 4, 217, 161, 200, 161, 217, 169, 200, 169, 57, 170, 153, 173, 136, 173, 153, 174, 136, 174, 249, 173, 232, 173, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 72, 255, 5, 186, 88, 255, 5, 185, 108, 255, 5, 186, 112, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 255, 5, 168, 72, 255, 5, 168, 88, 255, 5, 168, 108, 255, 5, 168, 112, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 218, 73, 255, 5, 217, 152, 255, 5, 200, 73, 255, 5, 200, 152, 255, 5, 249, 58, 255, 5, 232, 58, 255, 5, 250, 74, 255, 5, 232, 74, 255, 5, 250, 90, 255, 5, 232, 90, 255, 5, 218, 114, 255, 5, 200, 114, 255, 5, 249, 154, 255, 5, 232, 154, 255, 1, 170, 170, 3, 0, 68, 0, 59, 0, 32, 0, 255, 140, 1, 16, 219, 79, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 34, 84, 3, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 100, 0, 0, 0, 100, 1, 0, 0, 68, 45, 3, 0, 68, 45, 3, 0, 180, 55, 3, 0, 180, 55, 3, 0, 24, 64, 3, 0, 24, 64, 3, 0, 196, 79, 3, 0, 214, 79, 3, 0, 6, 84, 3, 0, 17, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 41, 127, 0, 0, 96, 125, 0, 0, 123, 44, 0, 118, 123, 125, 0, 0, 126, 169, 255, 0, 254, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 0, 249, 250, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 254, 255, 50, 105, 114, 84, 1, 0, 112, 14, 239, 48, 64, 0, 192, 0, 96, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 109, 0, 117, 0, 122, 0, 130, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 169, 0, 176, 0, 184, 0, 189, 0, 48, 0, 197, 0, 205, 0, 213, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 221, 0, 229, 0, 237, 0, 245, 0, 253, 0, 5, 1, 13, 1, 21, 1, 29, 1, 37, 1, 45, 1, 48, 0, 52, 1, 60, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 67, 1, 48, 0, 48, 0, 48, 0, 48, 0, 75, 1, 83, 1, 48, 0, 91, 1, 99, 1, 107, 1, 110, 1, 48, 0, 48, 0, 118, 1, 126, 1, 48, 0, 48, 0, 48, 0, 134, 1, 142, 1, 150, 1, 158, 1, 165, 1, 173, 1, 181, 1, 189, 1, 197, 1, 205, 1, 213, 1, 221, 1, 229, 1, 236, 1, 244, 1, 252, 1, 4, 2, 12, 2, 20, 2, 28, 2, 36, 2, 44, 2, 51, 2, 59, 2, 67, 2, 75, 2, 83, 2, 91, 2, 99, 2, 103, 2, 111, 2, 119, 2, 125, 2, 133, 2, 141, 2, 48, 0, 149, 2, 48, 0, 157, 2, 165, 2, 171, 2, 179, 2, 187, 2, 195, 2, 203, 2, 211, 2, 219, 2, 226, 2, 234, 2, 242, 2, 250, 2, 2, 3, 10, 3, 18, 3, 24, 3, 32, 3, 40, 3, 48, 3, 56, 3, 64, 3, 72, 3, 80, 3, 88, 3, 96, 3, 104, 3, 112, 3, 120, 3, 128, 3, 136, 3, 142, 3, 48, 0, 148, 3, 156, 3, 164, 3, 172, 3, 180, 3, 188, 3, 196, 3, 203, 3, 211, 3, 218, 3, 226, 3, 234, 3, 242, 3, 250, 3, 48, 0, 0, 4, 8, 4, 16, 4, 24, 4, 31, 4, 39, 4, 47, 4, 55, 4, 62, 4, 70, 4, 78, 4, 86, 4, 94, 4, 102, 4, 110, 4, 115, 4, 122, 4, 130, 4, 138, 4, 146, 4, 154, 4, 162, 4, 169, 4, 177, 4, 183, 4, 190, 4, 197, 4, 205, 4, 213, 4, 221, 4, 229, 4, 237, 4, 245, 4, 48, 0, 253, 4, 5, 5, 13, 5, 21, 5, 26, 5, 30, 5, 37, 5, 45, 5, 53, 5, 61, 5, 69, 5, 77, 5, 84, 5, 91, 5, 99, 5, 107, 5, 115, 5, 123, 5, 131, 5, 139, 5, 147, 5, 154, 5, 161, 5, 166, 5, 174, 5, 182, 5, 189, 5, 196, 5, 203, 5, 211, 5, 219, 5, 227, 5, 232, 5, 240, 5, 248, 5, 255, 5, 3, 6, 11, 6, 19, 6, 22, 6, 28, 6, 36, 6, 44, 6, 51, 6, 59, 6, 64, 6, 71, 6, 75, 6, 83, 6, 91, 6, 99, 6, 107, 6, 114, 6, 121, 6, 129, 6, 131, 6, 139, 6, 147, 6, 155, 6, 161, 6, 48, 0, 164, 6, 171, 6, 178, 6, 185, 6, 193, 6, 201, 6, 207, 6, 212, 6, 220, 6, 228, 6, 236, 6, 48, 0, 244, 6, 252, 6, 4, 7, 11, 7, 19, 7, 27, 7, 33, 7, 41, 7, 49, 7, 56, 7, 62, 7, 66, 7, 72, 7, 48, 0, 48, 0, 48, 0, 80, 7, 88, 7, 96, 7, 104, 7, 112, 7, 120, 7, 128, 7, 136, 7, 144, 7, 152, 7, 160, 7, 168, 7, 176, 7, 184, 7, 192, 7, 200, 7, 208, 7, 216, 7, 224, 7, 232, 7, 240, 7, 248, 7, 0, 8, 8, 8, 16, 8, 24, 8, 32, 8, 40, 8, 48, 8, 56, 8, 64, 8, 72, 8, 80, 8, 88, 8, 96, 8, 104, 8, 112, 8, 120, 8, 128, 8, 136, 8, 144, 8, 152, 8, 160, 8, 168, 8, 176, 8, 184, 8, 192, 8, 200, 8, 208, 8, 216, 8, 224, 8, 232, 8, 240, 8, 248, 8, 0, 9, 8, 9, 16, 9, 24, 9, 32, 9, 40, 9, 48, 9, 56, 9, 64, 9, 72, 9, 80, 9, 88, 9, 96, 9, 104, 9, 112, 9, 120, 9, 128, 9, 136, 9, 144, 9, 152, 9, 160, 9, 168, 9, 176, 9, 184, 9, 192, 9, 200, 9, 208, 9, 216, 9, 224, 9, 232, 9, 240, 9, 248, 9, 0, 10, 8, 10, 16, 10, 24, 10, 32, 10, 40, 10, 48, 10, 56, 10, 64, 10, 72, 10, 80, 10, 88, 10, 96, 10, 104, 10, 112, 10, 120, 10, 128, 10, 136, 10, 144, 10, 152, 10, 160, 10, 168, 10, 176, 10, 184, 10, 192, 10, 200, 10, 208, 10, 216, 10, 224, 10, 232, 10, 240, 10, 248, 10, 0, 11, 8, 11, 16, 11, 24, 11, 32, 11, 40, 11, 48, 11, 56, 11, 64, 11, 72, 11, 80, 11, 88, 11, 96, 11, 104, 11, 112, 11, 120, 11, 128, 11, 136, 11, 144, 11, 152, 11, 160, 11, 168, 11, 176, 11, 184, 11, 192, 11, 200, 11, 208, 11, 216, 11, 224, 11, 232, 11, 240, 11, 248, 11, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 40, 12, 48, 12, 56, 12, 64, 12, 72, 12, 80, 12, 88, 12, 96, 12, 104, 12, 112, 12, 120, 12, 128, 12, 136, 12, 144, 12, 152, 12, 160, 12, 168, 12, 176, 12, 184, 12, 192, 12, 200, 12, 208, 12, 216, 12, 224, 12, 232, 12, 240, 12, 248, 12, 0, 13, 8, 13, 16, 13, 24, 13, 32, 13, 40, 13, 48, 13, 56, 13, 64, 13, 72, 13, 80, 13, 88, 13, 96, 13, 104, 13, 112, 13, 120, 13, 128, 13, 136, 13, 144, 13, 152, 13, 160, 13, 168, 13, 176, 13, 184, 13, 192, 13, 200, 13, 208, 13, 216, 13, 224, 13, 232, 13, 240, 13, 248, 13, 0, 14, 8, 14, 16, 14, 24, 14, 32, 14, 40, 14, 48, 14, 56, 14, 64, 14, 72, 14, 80, 14, 88, 14, 96, 14, 104, 14, 112, 14, 120, 14, 128, 14, 136, 14, 144, 14, 152, 14, 160, 14, 168, 14, 176, 14, 184, 14, 192, 14, 200, 14, 208, 14, 216, 14, 224, 14, 232, 14, 240, 14, 248, 14, 0, 15, 8, 15, 16, 15, 24, 15, 32, 15, 40, 15, 48, 15, 56, 15, 64, 15, 72, 15, 80, 15, 88, 15, 96, 15, 104, 15, 112, 15, 120, 15, 128, 15, 136, 15, 144, 15, 152, 15, 160, 15, 168, 15, 176, 15, 184, 15, 192, 15, 200, 15, 208, 15, 216, 15, 224, 15, 232, 15, 240, 15, 248, 15, 0, 16, 8, 16, 16, 16, 24, 16, 32, 16, 40, 16, 48, 16, 56, 16, 64, 16, 72, 16, 80, 16, 88, 16, 96, 16, 104, 16, 112, 16, 120, 16, 128, 16, 136, 16, 144, 16, 152, 16, 160, 16, 168, 16, 176, 16, 184, 16, 192, 16, 200, 16, 208, 16, 216, 16, 224, 16, 232, 16, 240, 16, 248, 16, 0, 17, 8, 17, 16, 17, 24, 17, 32, 17, 40, 17, 48, 17, 56, 17, 64, 17, 72, 17, 80, 17, 88, 17, 96, 17, 104, 17, 112, 17, 120, 17, 128, 17, 136, 17, 144, 17, 152, 17, 160, 17, 168, 17, 176, 17, 184, 17, 192, 17, 200, 17, 208, 17, 216, 17, 224, 17, 232, 17, 240, 17, 248, 17, 0, 18, 8, 18, 16, 18, 24, 18, 32, 18, 40, 18, 48, 18, 56, 18, 64, 18, 72, 18, 80, 18, 88, 18, 96, 18, 104, 18, 112, 18, 120, 18, 128, 18, 136, 18, 144, 18, 152, 18, 160, 18, 168, 18, 176, 18, 184, 18, 192, 18, 200, 18, 208, 18, 216, 18, 224, 18, 232, 18, 240, 18, 248, 18, 0, 19, 8, 19, 16, 19, 24, 19, 32, 19, 40, 19, 48, 19, 56, 19, 64, 19, 72, 19, 80, 19, 88, 19, 96, 19, 104, 19, 112, 19, 120, 19, 128, 19, 136, 19, 144, 19, 152, 19, 160, 19, 168, 19, 176, 19, 184, 19, 192, 19, 200, 19, 208, 19, 216, 19, 224, 19, 232, 19, 240, 19, 248, 19, 0, 20, 8, 20, 16, 20, 24, 20, 32, 20, 40, 20, 48, 20, 56, 20, 64, 20, 72, 20, 80, 20, 88, 20, 96, 20, 104, 20, 112, 20, 120, 20, 128, 20, 136, 20, 144, 20, 152, 20, 160, 20, 168, 20, 176, 20, 184, 20, 192, 20, 200, 20, 208, 20, 216, 20, 224, 20, 232, 20, 240, 20, 248, 20, 0, 21, 8, 21, 16, 21, 24, 21, 32, 21, 40, 21, 48, 21, 56, 21, 64, 21, 72, 21, 80, 21, 88, 21, 96, 21, 104, 21, 112, 21, 120, 21, 128, 21, 136, 21, 144, 21, 152, 21, 160, 21, 168, 21, 176, 21, 184, 21, 192, 21, 200, 21, 208, 21, 216, 21, 224, 21, 232, 21, 240, 21, 248, 21, 0, 22, 8, 22, 16, 22, 24, 22, 32, 22, 40, 22, 48, 22, 56, 22, 64, 22, 72, 22, 80, 22, 88, 22, 96, 22, 104, 22, 112, 22, 120, 22, 128, 22, 136, 22, 144, 22, 152, 22, 160, 22, 168, 22, 176, 22, 184, 22, 192, 22, 200, 22, 208, 22, 216, 22, 224, 22, 232, 22, 240, 22, 248, 22, 0, 23, 8, 23, 16, 23, 24, 23, 32, 23, 40, 23, 48, 23, 56, 23, 64, 23, 72, 23, 80, 23, 88, 23, 96, 23, 104, 23, 112, 23, 120, 23, 128, 23, 136, 23, 144, 23, 152, 23, 160, 23, 168, 23, 176, 23, 184, 23, 192, 23, 200, 23, 208, 23, 216, 23, 224, 23, 232, 23, 240, 23, 248, 23, 0, 24, 8, 24, 16, 24, 24, 24, 32, 24, 40, 24, 48, 24, 56, 24, 64, 24, 72, 24, 80, 24, 88, 24, 96, 24, 104, 24, 112, 24, 120, 24, 128, 24, 136, 24, 144, 24, 152, 24, 160, 24, 168, 24, 176, 24, 184, 24, 192, 24, 200, 24, 208, 24, 216, 24, 224, 24, 232, 24, 240, 24, 248, 24, 0, 25, 8, 25, 16, 25, 24, 25, 32, 25, 40, 25, 48, 25, 56, 25, 64, 25, 72, 25, 80, 25, 88, 25, 96, 25, 104, 25, 112, 25, 120, 25, 128, 25, 136, 25, 144, 25, 152, 25, 160, 25, 168, 25, 176, 25, 184, 25, 192, 25, 200, 25, 208, 25, 216, 25, 224, 25, 232, 25, 240, 25, 248, 25, 0, 26, 8, 26, 16, 26, 24, 26, 32, 26, 40, 26, 48, 26, 56, 26, 64, 26, 72, 26, 80, 26, 88, 26, 96, 26, 104, 26, 112, 26, 120, 26, 128, 26, 136, 26, 144, 26, 152, 26, 160, 26, 168, 26, 176, 26, 184, 26, 192, 26, 200, 26, 208, 26, 216, 26, 224, 26, 232, 26, 240, 26, 248, 26, 0, 27, 8, 27, 16, 27, 24, 27, 32, 27, 40, 27, 48, 27, 56, 27, 64, 27, 72, 27, 80, 27, 88, 27, 96, 27, 104, 27, 112, 27, 120, 27, 128, 27, 136, 27, 144, 27, 152, 27, 160, 27, 168, 27, 176, 27, 184, 27, 191, 27, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 208, 48, 161, 0, 48, 0, 48, 0, 153, 0, 216, 48, 224, 48, 230, 48, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 199, 27, 207, 27, 215, 27, 223, 27, 231, 27, 239, 27, 247, 27, 255, 27, 7, 28, 15, 28, 23, 28, 31, 28, 39, 28, 47, 28, 55, 28, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 62, 28, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 180, 1, 232, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 128, 8, 64, 0, 191, 8, 254, 8, 62, 9, 126, 9, 190, 9, 254, 9, 62, 10, 125, 10, 189, 10, 253, 10, 60, 11, 124, 11, 188, 11, 252, 11, 60, 12, 123, 12, 187, 12, 251, 12, 58, 13, 122, 13, 186, 13, 240, 13, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 48, 14, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 67, 28, 75, 28, 83, 28, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 91, 28, 99, 28, 105, 28, 112, 28, 48, 0, 118, 28, 123, 28, 128, 28, 48, 0, 48, 0, 48, 0, 48, 0, 135, 28, 143, 28, 149, 28, 156, 28, 162, 28, 169, 28, 176, 28, 182, 28, 190, 28, 198, 28, 205, 28, 211, 28, 218, 28, 226, 28, 228, 28, 236, 28, 244, 28, 250, 28, 253, 28, 5, 29, 48, 0, 48, 0, 13, 29, 20, 29, 28, 29, 31, 29, 37, 29, 48, 0, 48, 0, 45, 29, 48, 0, 50, 29, 57, 29, 65, 29, 73, 29, 77, 29, 84, 29, 89, 29, 95, 29, 48, 0, 48, 0, 48, 0, 48, 0, 101, 29, 105, 29, 48, 0, 110, 29, 48, 0, 48, 0, 48, 0, 48, 0, 117, 29, 48, 0, 121, 29, 48, 0, 48, 0, 126, 29, 48, 0, 132, 29, 140, 29, 143, 29, 149, 29, 48, 0, 48, 0, 48, 0, 156, 29, 160, 29, 48, 0, 165, 29, 170, 29, 176, 29, 179, 29, 187, 29, 48, 0, 192, 29, 48, 0, 48, 0, 48, 0, 197, 29, 205, 29, 213, 29, 221, 29, 229, 29, 237, 29, 245, 29, 252, 29, 4, 30, 12, 30, 19, 30, 27, 30, 35, 30, 41, 30, 48, 30, 53, 30, 61, 30, 69, 30, 75, 30, 48, 0, 82, 30, 48, 0, 90, 30, 95, 30, 102, 30, 110, 30, 116, 30, 124, 30, 131, 30, 139, 30, 146, 30, 153, 30, 158, 30, 163, 30, 171, 30, 179, 30, 183, 30, 48, 0, 191, 30, 199, 30, 206, 30, 214, 30, 222, 30, 229, 30, 48, 0, 235, 30, 243, 30, 48, 0, 251, 30, 48, 0, 253, 30, 48, 0, 48, 0, 4, 31, 12, 31, 16, 31, 48, 0, 24, 31, 48, 0, 32, 31, 40, 31, 43, 31, 48, 0, 50, 31, 56, 31, 48, 0, 64, 31, 69, 31, 77, 31, 84, 31, 92, 31, 48, 0, 98, 31, 104, 31, 48, 0, 110, 31, 48, 0, 48, 0, 48, 0, 118, 31, 126, 31, 48, 0, 48, 0, 48, 0, 132, 31, 139, 31, 48, 0, 147, 31, 151, 31, 156, 31, 164, 31, 172, 31, 180, 31, 187, 31, 195, 31, 201, 31, 48, 0, 207, 31, 215, 31, 223, 31, 48, 0, 231, 31, 239, 31, 247, 31, 255, 31, 7, 32, 15, 32, 22, 32, 30, 32, 38, 32, 45, 32, 52, 32, 60, 32, 48, 0, 66, 32, 73, 32, 48, 0, 78, 32, 85, 32, 93, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 100, 32, 48, 0, 48, 0, 48, 0, 108, 32, 48, 0, 48, 0, 114, 32, 48, 0, 118, 32, 48, 0, 126, 32, 48, 0, 48, 0, 48, 0, 48, 0, 133, 32, 48, 0, 141, 32, 146, 32, 153, 32, 159, 32, 166, 32, 174, 32, 48, 0, 48, 0, 182, 32, 189, 32, 195, 32, 48, 0, 201, 32, 48, 0, 48, 0, 207, 32, 210, 32, 218, 32, 48, 0, 224, 32, 48, 0, 48, 0, 231, 32, 234, 32, 242, 32, 48, 0, 249, 32, 251, 32, 48, 0, 3, 33, 6, 33, 13, 33, 48, 0, 21, 33, 29, 33, 48, 0, 35, 33, 48, 0, 48, 0, 38, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 33, 48, 0, 48, 0, 48, 0, 48, 33, 56, 33, 62, 33, 48, 0, 48, 0, 48, 0, 69, 33, 75, 33, 80, 33, 88, 33, 93, 33, 97, 33, 104, 33, 109, 33, 48, 0, 115, 33, 120, 33, 48, 0, 48, 0, 128, 33, 136, 33, 48, 0, 142, 33, 149, 33, 48, 0, 48, 0, 156, 33, 162, 33, 170, 33, 48, 0, 173, 33, 181, 33, 187, 33, 193, 33, 48, 0, 201, 33, 48, 0, 202, 33, 210, 33, 48, 0, 215, 33, 223, 33, 226, 33, 48, 0, 233, 33, 240, 33, 247, 33, 255, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 6, 34, 48, 0, 48, 0, 11, 34, 18, 34, 23, 34, 31, 34, 39, 34, 45, 34, 48, 0, 48, 0, 51, 34, 48, 0, 48, 0, 58, 34, 65, 34, 73, 34, 79, 34, 87, 34, 95, 34, 48, 0, 102, 34, 48, 0, 109, 34, 116, 34, 121, 34, 129, 34, 48, 0, 135, 34, 140, 34, 144, 34, 152, 34, 154, 34, 161, 34, 165, 34, 48, 0, 170, 34, 48, 0, 48, 0, 48, 0, 174, 34, 48, 0, 48, 0, 180, 34, 188, 34, 48, 0, 192, 34, 199, 34, 206, 34, 214, 34, 218, 34, 226, 34, 48, 0, 234, 34, 235, 34, 48, 0, 243, 34, 251, 34, 2, 35, 8, 35, 14, 35, 21, 35, 28, 35, 34, 35, 42, 35, 50, 35, 56, 35, 60, 35, 68, 35, 76, 35, 48, 0, 48, 0, 48, 0, 48, 0, 84, 35, 90, 35, 98, 35, 106, 35, 107, 35, 115, 35, 120, 35, 48, 0, 48, 0, 48, 0, 127, 35, 129, 35, 136, 35, 141, 35, 148, 35, 151, 35, 159, 35, 166, 35, 170, 35, 178, 35, 185, 35, 193, 35, 201, 35, 207, 35, 215, 35, 222, 35, 230, 35, 238, 35, 246, 35, 48, 0, 254, 35, 6, 36, 14, 36, 22, 36, 48, 0, 28, 36, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 36, 36, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 43, 36, 45, 36, 52, 36, 59, 36, 48, 0, 66, 36, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 74, 36, 76, 36, 83, 36, 90, 36, 93, 36, 48, 0, 48, 0, 101, 36, 108, 36, 116, 36, 120, 36, 48, 0, 128, 36, 135, 36, 140, 36, 48, 0, 48, 0, 147, 36, 150, 36, 158, 36, 164, 36, 48, 0, 172, 36, 180, 36, 188, 36, 196, 36, 200, 36, 208, 36, 214, 36, 219, 36, 224, 36, 232, 36, 48, 0, 239, 36, 48, 0, 245, 36, 253, 36, 4, 37, 11, 37, 19, 37, 27, 37, 48, 0, 35, 37, 41, 37, 49, 37, 57, 37, 48, 0, 64, 37, 71, 37, 78, 37, 85, 37, 93, 37, 98, 37, 105, 37, 112, 37, 118, 37, 126, 37, 131, 37, 139, 37, 147, 37, 153, 37, 160, 37, 163, 37, 169, 37, 176, 37, 183, 37, 186, 37, 194, 37, 48, 0, 201, 37, 209, 37, 48, 0, 211, 37, 48, 0, 48, 0, 218, 37, 48, 0, 221, 37, 228, 37, 48, 0, 235, 37, 241, 37, 48, 0, 247, 37, 48, 0, 254, 37, 5, 38, 48, 0, 48, 0, 10, 38, 48, 0, 48, 0, 18, 38, 22, 38, 30, 38, 48, 0, 48, 0, 38, 38, 48, 0, 48, 0, 44, 38, 52, 38, 59, 38, 67, 38, 75, 38, 83, 38, 90, 38, 97, 38, 105, 38, 112, 38, 120, 38, 128, 38, 136, 38, 143, 38, 150, 38, 158, 38, 166, 38, 171, 38, 48, 0, 48, 0, 176, 38, 48, 0, 48, 0, 48, 0, 180, 38, 186, 38, 48, 0, 48, 0, 48, 0, 189, 38, 48, 0, 197, 38, 48, 0, 204, 38, 48, 0, 48, 0, 48, 0, 48, 0, 209, 38, 48, 0, 216, 38, 48, 0, 219, 38, 226, 38, 48, 0, 48, 0, 233, 38, 48, 0, 48, 0, 238, 38, 48, 0, 242, 38, 48, 0, 250, 38, 1, 39, 9, 39, 48, 0, 14, 39, 18, 39, 48, 0, 23, 39, 48, 0, 48, 0, 48, 0, 48, 0, 31, 39, 37, 39, 40, 39, 48, 0, 48, 0, 48, 39, 56, 39, 48, 0, 64, 39, 68, 39, 48, 0, 73, 39, 48, 0, 80, 39, 48, 0, 87, 39, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 92, 39, 99, 39, 106, 39, 111, 39, 115, 39, 48, 0, 48, 0, 48, 0, 48, 0, 122, 39, 130, 39, 138, 39, 146, 39, 150, 39, 157, 39, 165, 39, 172, 39, 175, 39, 181, 39, 48, 0, 189, 39, 48, 0, 48, 0, 197, 39, 204, 39, 212, 39, 48, 0, 220, 39, 48, 0, 48, 0, 227, 39, 235, 39, 48, 0, 48, 0, 238, 39, 246, 39, 48, 0, 48, 0, 252, 39, 4, 40, 12, 40, 48, 0, 19, 40, 48, 0, 48, 0, 25, 40, 30, 40, 48, 0, 48, 0, 48, 0, 34, 40, 48, 0, 40, 40, 48, 40, 48, 0, 54, 40, 48, 0, 48, 0, 48, 0, 57, 40, 63, 40, 67, 40, 74, 40, 81, 40, 89, 40, 97, 40, 103, 40, 110, 40, 114, 40, 121, 40, 129, 40, 130, 40, 48, 0, 138, 40, 146, 40, 48, 0, 148, 40, 155, 40, 48, 0, 48, 0, 162, 40, 48, 0, 168, 40, 48, 0, 176, 40, 183, 40, 191, 40, 48, 0, 198, 40, 48, 0, 205, 40, 48, 0, 48, 0, 48, 0, 48, 0, 213, 40, 48, 0, 221, 40, 227, 40, 234, 40, 242, 40, 249, 40, 48, 0, 48, 0, 1, 41, 9, 41, 11, 41, 48, 0, 48, 0, 19, 41, 24, 41, 48, 0, 48, 0, 48, 0, 32, 41, 40, 41, 48, 0, 48, 0, 44, 41, 48, 0, 48, 0, 49, 41, 55, 41, 48, 0, 48, 0, 60, 41, 48, 0, 68, 41, 76, 41, 48, 0, 82, 41, 88, 41, 95, 41, 48, 0, 48, 0, 48, 0, 102, 41, 48, 0, 48, 0, 107, 41, 48, 0, 115, 41, 48, 0, 48, 0, 123, 41, 130, 41, 48, 0, 48, 0, 48, 0, 135, 41, 141, 41, 148, 41, 156, 41, 48, 0, 164, 41, 48, 0, 166, 41, 173, 41, 179, 41, 48, 0, 183, 41, 48, 0, 191, 41, 199, 41, 205, 41, 48, 0, 212, 41, 48, 0, 217, 41, 224, 41, 231, 41, 48, 0, 48, 0, 237, 41, 48, 0, 243, 41, 48, 0, 250, 41, 2, 42, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 9, 42, 17, 42, 21, 42, 28, 42, 36, 42, 44, 42, 52, 42, 55, 42, 48, 0, 63, 42, 71, 42, 73, 42, 81, 42, 88, 42, 96, 42, 48, 0, 48, 0, 104, 42, 108, 42, 113, 42, 48, 0, 121, 42, 125, 42, 133, 42, 141, 42, 48, 0, 149, 42, 154, 42, 161, 42, 48, 0, 168, 42, 48, 0, 48, 0, 175, 42, 48, 0, 179, 42, 185, 42, 193, 42, 201, 42, 48, 0, 209, 42, 48, 0, 215, 42, 48, 0, 221, 42, 226, 42, 231, 42, 236, 42, 243, 42, 250, 42, 48, 0, 1, 43, 48, 0, 48, 0, 7, 43, 48, 0, 11, 43, 48, 0, 48, 0, 18, 43, 26, 43, 30, 43, 48, 0, 37, 43, 48, 0, 48, 0, 41, 43, 48, 0, 46, 43, 48, 0, 48, 0, 53, 43, 48, 0, 61, 43, 65, 43, 48, 0, 72, 43, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 75, 43, 48, 0, 83, 43, 48, 0, 91, 43, 99, 43, 103, 43, 110, 43, 48, 0, 116, 43, 48, 0, 48, 0, 48, 0, 122, 43, 126, 43, 134, 43, 48, 0, 142, 43, 48, 0, 147, 43, 48, 0, 48, 0, 151, 43, 158, 43, 163, 43, 171, 43, 177, 43, 48, 0, 48, 0, 184, 43, 48, 0, 191, 43, 48, 0, 199, 43, 200, 43, 208, 43, 216, 43, 220, 43, 228, 43, 235, 43, 242, 43, 244, 43, 48, 0, 252, 43, 3, 44, 10, 44, 17, 44, 24, 44, 48, 0, 48, 0, 48, 0, 29, 44, 35, 44, 48, 0, 48, 0, 40, 44, 48, 0, 44, 44, 51, 44, 48, 0, 48, 0, 56, 44, 60, 44, 67, 44, 74, 44, 81, 44, 89, 44, 48, 0, 48, 0, 48, 0, 97, 44, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 102, 44, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 106, 44, 114, 44, 122, 44, 128, 44, 48, 0, 136, 44, 141, 44, 48, 0, 149, 44, 48, 0, 155, 44, 160, 44, 168, 44, 174, 44, 179, 44, 48, 0, 48, 0, 187, 44, 48, 0, 194, 44, 200, 44, 208, 44, 48, 0, 215, 44, 223, 44, 230, 44, 237, 44, 48, 0, 244, 44, 252, 44, 48, 0, 48, 0, 48, 0, 0, 45, 48, 0, 48, 0, 6, 45, 14, 45, 48, 0, 48, 0, 48, 0, 22, 45, 23, 45, 48, 0, 31, 45, 48, 0, 48, 0, 48, 0, 37, 45, 45, 45, 52, 45, 48, 0, 48, 0, 48, 0, 59, 45, 67, 45, 74, 45, 79, 45, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 86, 45, 92, 45, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 97, 45, 101, 45, 107, 45, 114, 45, 122, 45, 128, 45, 136, 45, 144, 45, 152, 45, 159, 45, 166, 45, 48, 0, 174, 45, 48, 0, 182, 45, 190, 45, 198, 45, 206, 45, 213, 45, 221, 45, 226, 45, 234, 45, 242, 45, 249, 45, 1, 46, 9, 46, 48, 0, 48, 0, 48, 0, 48, 0, 16, 46, 48, 0, 48, 0, 21, 46, 48, 0, 48, 0, 27, 46, 34, 46, 48, 0, 48, 0, 41, 46, 45, 46, 48, 0, 52, 46, 55, 46, 61, 46, 66, 46, 73, 46, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 80, 46, 48, 0, 48, 0, 48, 0, 82, 46, 89, 46, 94, 46, 102, 46, 109, 46, 114, 46, 119, 46, 127, 46, 135, 46, 140, 46, 148, 46, 151, 46, 157, 46, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 160, 46, 168, 46, 176, 46, 48, 0, 48, 0, 48, 0, 180, 46, 187, 46, 48, 0, 48, 0, 48, 0, 48, 0, 189, 46, 197, 46, 204, 46, 210, 46, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 218, 46, 48, 0, 48, 0, 48, 0, 48, 0, 223, 46, 48, 0, 229, 46, 48, 0, 234, 46, 48, 0, 48, 0, 241, 46, 245, 46, 251, 46, 3, 47, 8, 47, 15, 47, 23, 47, 26, 47, 48, 0, 34, 47, 48, 0, 48, 0, 40, 47, 45, 47, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 52, 47, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 56, 47, 48, 0, 48, 0, 48, 0, 48, 0, 62, 47, 48, 0, 67, 47, 48, 0, 48, 0, 48, 0, 71, 47, 79, 47, 86, 47, 94, 47, 48, 0, 48, 0, 101, 47, 108, 47, 113, 47, 48, 0, 119, 47, 48, 0, 124, 47, 132, 47, 134, 47, 48, 0, 142, 47, 48, 0, 48, 0, 48, 0, 147, 47, 153, 47, 48, 0, 161, 47, 48, 0, 48, 0, 48, 0, 168, 47, 170, 47, 48, 0, 178, 47, 186, 47, 194, 47, 200, 47, 48, 0, 205, 47, 211, 47, 217, 47, 225, 47, 48, 0, 233, 47, 48, 0, 48, 0, 238, 47, 246, 47, 254, 47, 6, 48, 48, 0, 48, 0, 14, 48, 48, 0, 48, 0, 20, 48, 48, 0, 48, 0, 48, 0, 23, 48, 29, 48, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 36, 48, 48, 0, 44, 48, 48, 48, 54, 48, 48, 0, 48, 0, 61, 48, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 64, 48, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 72, 48, 80, 48, 88, 48, 96, 48, 104, 48, 112, 48, 120, 48, 128, 48, 136, 48, 144, 48, 152, 48, 160, 48, 168, 48, 176, 48, 184, 48, 192, 48, 200, 48, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 41, 0, 202, 65, 39, 0, 202, 130, 39, 0, 202, 195, 39, 0, 202, 4, 40, 0, 202, 69, 40, 0, 202, 134, 40, 0, 202, 199, 40, 0, 202, 8, 41, 0, 202, 73, 41, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 134, 2, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 38, 8, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 201, 6, 11, 0, 201, 70, 14, 0, 201, 134, 17, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 198, 20, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 70, 28, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 38, 31, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 198, 36, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 201, 166, 39, 0, 201, 230, 42, 0, 201, 38, 46, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 102, 49, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 134, 3, 0, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 98, 66, 0, 196, 160, 156, 46, 198, 65, 0, 0, 134, 3, 0, 50, 201, 230, 56, 0, 196, 150, 156, 50, 198, 129, 0, 0, 134, 3, 0, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 198, 65, 1, 0, 134, 3, 0, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 198, 129, 1, 0, 134, 3, 0, 82, 196, 142, 156, 82, 201, 198, 59, 0, 196, 136, 156, 90, 160, 5, 38, 93, 197, 162, 66, 0, 198, 193, 1, 0, 5, 3, 0, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 226, 66, 0, 196, 160, 5, 46, 198, 1, 2, 0, 5, 3, 0, 50, 201, 102, 63, 0, 196, 150, 5, 50, 198, 65, 2, 0, 5, 3, 0, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 198, 1, 3, 0, 5, 3, 0, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 198, 65, 3, 0, 5, 3, 0, 82, 196, 142, 5, 82, 201, 70, 66, 0, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 134, 2, 0, 42, 5, 2, 0, 42, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 69, 0, 201, 134, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 50, 5, 2, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 75, 0, 201, 70, 78, 0, 198, 97, 0, 0, 198, 33, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 58, 5, 2, 0, 58, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 81, 0, 201, 6, 84, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 68, 5, 3, 0, 68, 201, 230, 86, 0, 201, 198, 89, 0, 198, 33, 1, 0, 198, 225, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 70, 5, 2, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 82, 5, 2, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 92, 0, 201, 230, 95, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 99, 0, 201, 6, 102, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 104, 0, 201, 38, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 198, 225, 1, 0, 198, 161, 0, 0, 198, 97, 2, 0, 198, 97, 1, 0, 198, 33, 3, 0, 198, 161, 1, 0, 198, 97, 3, 0, 134, 70, 0, 82, 5, 70, 0, 82, 134, 71, 0, 82, 5, 71, 0, 82, 134, 72, 0, 82, 5, 72, 0, 82, 134, 73, 0, 82, 5, 73, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 111, 0, 201, 70, 114, 0, 197, 98, 8, 0, 197, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 1, 0, 198, 193, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 117, 0, 201, 6, 120, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 122, 0, 201, 198, 125, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 201, 166, 128, 0, 201, 134, 131, 0, 201, 102, 134, 0, 201, 70, 137, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 140, 0, 201, 6, 143, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 66, 5, 3, 0, 66, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 145, 0, 201, 198, 148, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 151, 0, 201, 134, 154, 0, 201, 102, 157, 0, 201, 70, 160, 0, 201, 38, 163, 0, 201, 6, 166, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 197, 226, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 168, 0, 201, 38, 172, 0, 201, 102, 175, 0, 201, 166, 178, 0, 201, 230, 181, 0, 201, 38, 185, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 24, 0, 197, 226, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 188, 0, 201, 70, 191, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 194, 0, 201, 6, 197, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 50, 5, 71, 0, 50, 134, 73, 0, 50, 5, 73, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 199, 0, 201, 198, 202, 0, 201, 166, 205, 0, 201, 134, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 5, 0, 197, 34, 6, 0, 198, 130, 8, 0, 198, 2, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 211, 0, 201, 70, 214, 0, 201, 38, 217, 0, 201, 102, 220, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 6, 0, 197, 98, 7, 0, 198, 130, 9, 0, 198, 2, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 223, 0, 201, 230, 226, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 136, 192, 0, 0, 0, 5, 5, 21, 157, 5, 5, 26, 157, 5, 5, 30, 157, 192, 0, 0, 0, 5, 5, 38, 157, 192, 0, 0, 0, 5, 5, 116, 157, 5, 5, 117, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 158, 5, 5, 70, 158, 192, 0, 0, 0, 6, 5, 120, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 158, 192, 0, 0, 0, 5, 5, 99, 158, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 159, 192, 0, 0, 0, 5, 5, 174, 159, 192, 0, 0, 0, 5, 5, 165, 161, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 161, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 158, 192, 0, 0, 0, 5, 5, 152, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 165, 192, 0, 0, 0, 5, 5, 154, 157, 5, 5, 119, 158, 192, 0, 0, 0, 5, 5, 47, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 157, 5, 5, 111, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 252, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 0, 128, 6, 5, 0, 129, 6, 5, 0, 130, 6, 5, 0, 131, 7, 5, 0, 134, 6, 5, 0, 143, 9, 5, 14, 157, 6, 5, 15, 157, 7, 5, 16, 157, 6, 5, 18, 157, 6, 5, 19, 157, 7, 5, 20, 157, 6, 5, 22, 157, 6, 5, 23, 157, 6, 5, 24, 157, 6, 5, 25, 157, 6, 5, 27, 157, 6, 5, 28, 157, 6, 5, 32, 157, 6, 5, 33, 157, 6, 5, 34, 157, 6, 5, 35, 157, 6, 5, 36, 157, 8, 5, 37, 157, 6, 5, 39, 157, 6, 5, 40, 157, 6, 5, 41, 157, 6, 5, 42, 157, 6, 5, 44, 157, 6, 5, 100, 157, 6, 5, 101, 157, 7, 5, 102, 157, 6, 5, 103, 157, 6, 5, 104, 157, 6, 5, 105, 157, 6, 5, 106, 157, 6, 5, 107, 157, 7, 5, 109, 157, 6, 5, 110, 157, 6, 5, 114, 157, 6, 5, 115, 157, 6, 5, 118, 157, 6, 5, 119, 157, 6, 5, 122, 157, 6, 5, 123, 157, 6, 5, 124, 157, 6, 5, 125, 157, 6, 5, 128, 157, 6, 5, 129, 157, 6, 5, 132, 157, 6, 5, 133, 157, 6, 5, 135, 157, 6, 5, 136, 157, 6, 5, 137, 157, 6, 5, 138, 157, 6, 5, 139, 157, 6, 5, 140, 157, 6, 5, 141, 157, 6, 5, 144, 157, 6, 5, 145, 157, 6, 5, 49, 158, 6, 5, 51, 158, 6, 5, 52, 158, 7, 5, 55, 158, 6, 5, 58, 158, 6, 5, 59, 158, 7, 5, 61, 158, 6, 5, 62, 158, 6, 5, 63, 158, 6, 5, 64, 158, 6, 5, 65, 158, 7, 5, 68, 158, 6, 5, 69, 158, 8, 5, 71, 158, 7, 5, 72, 158, 6, 5, 75, 158, 6, 5, 76, 158, 6, 5, 78, 158, 6, 5, 79, 158, 6, 5, 80, 158, 6, 5, 82, 158, 6, 5, 83, 158, 6, 5, 84, 158, 6, 5, 85, 158, 7, 5, 86, 158, 7, 5, 87, 158, 6, 5, 89, 158, 6, 5, 92, 158, 6, 5, 93, 158, 6, 5, 95, 158, 6, 5, 96, 158, 6, 5, 97, 158, 6, 5, 100, 158, 6, 5, 102, 158, 6, 5, 144, 159, 6, 5, 145, 159, 6, 5, 149, 159, 6, 5, 150, 159, 6, 5, 151, 159, 7, 5, 152, 159, 6, 5, 153, 159, 6, 5, 155, 159, 6, 5, 161, 159, 6, 5, 164, 159, 6, 5, 165, 159, 6, 5, 166, 159, 6, 5, 167, 159, 6, 5, 168, 159, 6, 5, 169, 159, 6, 5, 170, 159, 6, 5, 171, 159, 6, 5, 172, 159, 6, 5, 173, 159, 6, 5, 175, 159, 6, 5, 176, 159, 6, 5, 177, 159, 6, 5, 178, 159, 6, 5, 168, 161, 6, 5, 169, 161, 6, 5, 170, 161, 6, 5, 183, 161, 6, 5, 184, 161, 6, 5, 185, 161, 6, 5, 189, 161, 6, 5, 190, 161, 6, 5, 192, 161, 6, 5, 193, 161, 6, 5, 194, 161, 6, 5, 195, 161, 6, 5, 197, 161, 6, 5, 202, 161, 6, 5, 203, 161, 6, 5, 204, 161, 6, 5, 205, 161, 6, 5, 207, 161, 6, 5, 208, 161, 6, 5, 209, 161, 6, 5, 210, 161, 6, 5, 211, 161, 6, 5, 212, 161, 6, 5, 222, 161, 6, 5, 223, 161, 6, 5, 224, 161, 6, 5, 225, 161, 6, 5, 226, 161, 6, 5, 227, 161, 6, 5, 123, 165, 6, 5, 125, 165, 6, 5, 128, 165, 6, 5, 148, 165, 6, 5, 149, 165, 6, 5, 150, 165, 6, 5, 151, 165, 6, 5, 152, 165, 6, 5, 155, 165, 7, 5, 156, 165, 6, 5, 157, 165, 6, 5, 159, 165, 6, 5, 160, 165, 6, 5, 164, 165, 6, 5, 165, 165, 6, 5, 166, 165, 6, 5, 183, 165, 6, 5, 200, 165, 6, 5, 201, 165, 6, 5, 202, 165, 7, 5, 71, 171, 6, 5, 84, 171, 6, 5, 86, 171, 6, 5, 89, 171, 6, 5, 109, 171, 6, 5, 110, 171, 6, 5, 112, 171, 6, 5, 113, 171, 6, 5, 115, 171, 6, 5, 192, 177, 6, 5, 193, 177, 6, 5, 194, 177, 6, 5, 196, 177, 6, 5, 197, 177, 6, 5, 198, 177, 6, 5, 203, 177, 6, 5, 207, 177, 6, 5, 208, 177, 6, 5, 219, 177, 6, 5, 221, 177, 6, 5, 50, 185, 6, 5, 57, 185, 6, 5, 58, 185, 6, 5, 59, 185, 6, 5, 60, 185, 6, 5, 61, 185, 6, 5, 62, 185, 6, 5, 63, 185, 6, 5, 89, 193, 6, 5, 91, 193, 6, 5, 101, 193, 6, 5, 102, 193, 6, 5, 103, 193, 6, 5, 105, 193, 6, 5, 231, 201, 6, 5, 233, 201, 6, 5, 234, 201, 6, 5, 235, 201, 6, 5, 200, 209, 6, 5, 202, 209, 6, 5, 203, 209, 6, 5, 205, 209, 6, 5, 244, 216, 6, 5, 245, 216, 6, 5, 41, 224, 6, 5, 51, 230, 6, 5, 52, 230, 6, 5, 32, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 157, 5, 5, 0, 145, 5, 5, 0, 146, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 147, 5, 5, 50, 157, 5, 5, 161, 157, 5, 5, 51, 157, 5, 5, 0, 148, 5, 5, 52, 157, 5, 5, 162, 157, 5, 5, 124, 158, 5, 5, 125, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 5, 37, 157, 6, 5, 61, 157, 6, 5, 252, 157, 5, 5, 0, 149, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 8, 5, 0, 128, 8, 5, 14, 157, 8, 5, 56, 157, 7, 5, 25, 159, 7, 5, 57, 157, 8, 5, 171, 157, 7, 5, 58, 157, 6, 5, 157, 159, 6, 5, 0, 134, 6, 5, 132, 158, 6, 5, 0, 151, 7, 5, 23, 158, 6, 5, 169, 160, 6, 5, 16, 157, 198, 131, 31, 0, 198, 227, 31, 0, 198, 67, 32, 0, 198, 163, 32, 0, 198, 3, 33, 0, 198, 99, 33, 0, 198, 195, 33, 0, 198, 35, 34, 0, 198, 131, 34, 0, 198, 227, 34, 0, 198, 67, 35, 0, 198, 163, 35, 0, 198, 3, 36, 0, 198, 99, 36, 0, 197, 227, 35, 0, 198, 195, 36, 0, 198, 35, 37, 0, 198, 131, 37, 0, 198, 227, 37, 0, 197, 67, 36, 0, 198, 67, 38, 0, 198, 163, 38, 0, 197, 163, 36, 0, 197, 3, 37, 0, 198, 3, 39, 0, 198, 99, 39, 0, 198, 195, 39, 0, 198, 35, 40, 0, 197, 99, 37, 0, 198, 131, 40, 0, 197, 195, 37, 0, 198, 227, 40, 0, 197, 35, 38, 0, 198, 67, 41, 0, 197, 131, 38, 0, 197, 227, 38, 0, 6, 5, 246, 185, 6, 5, 77, 159, 6, 5, 61, 158, 6, 5, 171, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 5, 0, 128, 7, 5, 14, 157, 7, 5, 56, 157, 6, 5, 25, 159, 6, 5, 188, 157, 6, 5, 218, 157, 6, 5, 0, 153, 6, 5, 20, 157, 6, 5, 11, 157, 6, 5, 37, 157, 7, 5, 71, 158, 6, 5, 87, 158, 6, 5, 86, 158, 6, 5, 72, 158, 6, 5, 71, 171, 6, 5, 102, 157, 6, 5, 68, 158, 6, 5, 189, 180, 7, 5, 77, 161, 6, 5, 63, 170, 6, 5, 140, 160, 6, 5, 197, 181, 6, 5, 114, 184, 6, 5, 161, 182, 6, 5, 159, 162, 6, 5, 170, 182, 6, 5, 23, 165, 6, 5, 109, 157, 6, 5, 186, 222, 6, 5, 65, 230, 6, 5, 116, 160, 6, 5, 139, 169, 6, 5, 178, 201, 6, 5, 23, 160, 6, 5, 193, 158, 6, 5, 118, 159, 6, 5, 57, 157, 7, 5, 171, 157, 6, 5, 58, 157, 7, 5, 67, 159, 7, 5, 7, 159, 6, 5, 166, 162, 6, 5, 158, 167, 6, 5, 150, 167, 6, 5, 48, 213, 6, 5, 7, 160, 6, 5, 41, 208, 6, 5, 165, 166, 6, 5, 72, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 39, 0, 197, 130, 39, 0, 197, 194, 39, 0, 197, 2, 40, 0, 197, 66, 40, 0, 197, 130, 40, 0, 197, 194, 40, 0, 197, 2, 41, 0, 197, 66, 41, 0, 197, 131, 41, 0, 197, 227, 41, 0, 197, 67, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 226, 42, 0, 197, 34, 43, 0, 197, 98, 43, 0, 197, 162, 43, 0, 197, 226, 43, 0, 197, 34, 44, 0, 197, 98, 44, 0, 197, 162, 44, 0, 197, 226, 44, 0, 197, 34, 45, 0, 197, 99, 45, 0, 197, 195, 45, 0, 197, 35, 46, 0, 197, 131, 46, 0, 197, 227, 46, 0, 197, 67, 47, 0, 197, 163, 47, 0, 197, 3, 48, 0, 197, 99, 48, 0, 197, 195, 48, 0, 197, 35, 49, 0, 197, 131, 49, 0, 197, 227, 49, 0, 197, 67, 50, 0, 197, 163, 50, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 2, 51, 0, 197, 66, 51, 0, 197, 130, 51, 0, 197, 194, 51, 0, 197, 4, 52, 0, 197, 130, 52, 0, 197, 194, 52, 0, 197, 2, 53, 0, 197, 66, 53, 0, 197, 130, 53, 0, 197, 194, 53, 0, 197, 2, 54, 0, 197, 66, 54, 0, 197, 130, 54, 0, 197, 195, 54, 0, 197, 35, 55, 0, 197, 131, 55, 0, 197, 227, 55, 0, 197, 67, 56, 0, 197, 163, 56, 0, 197, 3, 57, 0, 197, 99, 57, 0, 197, 195, 57, 0, 197, 35, 58, 0, 197, 131, 58, 0, 197, 227, 58, 0, 197, 67, 59, 0, 197, 163, 59, 0, 197, 3, 60, 0, 197, 99, 60, 0, 197, 195, 60, 0, 197, 35, 61, 0, 197, 131, 61, 0, 197, 227, 61, 0, 197, 67, 62, 0, 197, 163, 62, 0, 192, 0, 0, 0, 5, 5, 137, 158, 5, 5, 219, 159, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 171, 192, 0, 0, 0, 5, 5, 83, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 159, 192, 0, 0, 0, 5, 5, 18, 166, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 158, 5, 5, 175, 158, 5, 5, 176, 158, 5, 5, 177, 158, 5, 5, 178, 158, 5, 5, 179, 158, 5, 5, 180, 158, 5, 5, 181, 158, 5, 5, 52, 160, 5, 5, 53, 160, 192, 0, 0, 0, 5, 5, 54, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 160, 5, 5, 56, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 166, 192, 0, 0, 0, 5, 5, 91, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 185, 5, 5, 72, 178, 5, 5, 73, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 193, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 202, 5, 5, 27, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 217, 192, 0, 0, 0, 5, 5, 23, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 238, 192, 0, 0, 0, 5, 5, 26, 242, 192, 0, 0, 0, 5, 5, 152, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 235, 5, 5, 186, 158, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 166, 192, 0, 0, 0, 5, 5, 140, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 210, 192, 0, 0, 0, 5, 5, 16, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 185, 192, 0, 0, 0, 5, 5, 23, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 230, 5, 5, 79, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 157, 192, 0, 0, 0, 5, 5, 188, 185, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 162, 5, 5, 192, 185, 5, 5, 56, 202, 5, 5, 57, 202, 5, 5, 58, 202, 5, 5, 27, 210, 5, 5, 220, 249, 5, 5, 167, 162, 5, 5, 188, 193, 5, 5, 2, 158, 192, 0, 0, 0, 5, 5, 115, 160, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 210, 192, 0, 0, 0, 5, 5, 82, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 210, 5, 5, 133, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 162, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 162, 192, 0, 0, 0, 5, 5, 251, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 166, 5, 5, 254, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 178, 5, 5, 225, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 194, 192, 0, 0, 0, 5, 5, 13, 194, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 194, 5, 5, 15, 194, 192, 0, 0, 0, 5, 5, 16, 194, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 202, 192, 0, 0, 0, 5, 5, 123, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 217, 5, 5, 108, 217, 5, 5, 109, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 224, 192, 0, 0, 0, 5, 5, 111, 224, 5, 5, 112, 224, 5, 5, 113, 224, 5, 5, 114, 224, 5, 5, 115, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 246, 192, 0, 0, 0, 5, 5, 183, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 167, 192, 0, 0, 0, 5, 5, 114, 172, 5, 5, 243, 178, 192, 0, 0, 0, 5, 5, 39, 242, 5, 5, 34, 159, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 163, 192, 0, 0, 0, 5, 5, 60, 163, 192, 0, 0, 0, 5, 5, 61, 163, 5, 5, 62, 163, 192, 0, 0, 0, 5, 5, 63, 163, 5, 5, 62, 167, 5, 5, 63, 167, 5, 5, 64, 167, 5, 5, 65, 167, 5, 5, 66, 167, 5, 5, 160, 172, 5, 5, 161, 172, 5, 5, 162, 172, 5, 5, 163, 172, 192, 0, 0, 0, 5, 5, 164, 172, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 186, 5, 5, 121, 186, 5, 5, 122, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 202, 192, 0, 0, 0, 5, 5, 148, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 210, 5, 5, 150, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 163, 192, 0, 0, 0, 5, 5, 113, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 167, 5, 5, 130, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 167, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 167, 192, 0, 0, 0, 5, 5, 232, 172, 5, 5, 233, 172, 5, 5, 234, 172, 192, 0, 0, 0, 5, 5, 235, 172, 5, 5, 236, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 179, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 179, 192, 0, 0, 0, 5, 5, 88, 179, 5, 5, 89, 179, 192, 0, 0, 0, 5, 5, 90, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 179, 5, 5, 93, 179, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 186, 5, 5, 207, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 194, 5, 5, 155, 194, 192, 0, 0, 0, 5, 5, 156, 194, 5, 5, 157, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 202, 192, 0, 0, 0, 5, 5, 238, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 210, 192, 0, 0, 0, 5, 5, 198, 210, 192, 0, 0, 0, 5, 5, 199, 210, 192, 0, 0, 0, 5, 5, 200, 210, 5, 5, 187, 217, 192, 0, 0, 0, 5, 5, 188, 217, 5, 5, 189, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 224, 5, 5, 169, 224, 5, 5, 170, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 235, 5, 5, 85, 235, 5, 5, 255, 238, 5, 5, 2, 239, 5, 5, 3, 239, 5, 5, 4, 239, 5, 5, 51, 242, 5, 5, 52, 242, 5, 5, 53, 242, 5, 5, 180, 244, 5, 5, 110, 246, 5, 5, 222, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 167, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 210, 5, 5, 2, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 239, 192, 0, 0, 0, 5, 5, 17, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 179, 192, 0, 0, 0, 5, 5, 253, 194, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 167, 5, 5, 38, 187, 5, 5, 74, 159, 5, 5, 75, 159, 5, 5, 3, 161, 5, 5, 4, 161, 5, 5, 5, 161, 5, 5, 185, 163, 5, 5, 186, 163, 5, 5, 187, 163, 192, 0, 0, 0, 5, 5, 188, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 217, 5, 5, 203, 224, 5, 5, 204, 224, 5, 5, 205, 224, 5, 5, 176, 230, 5, 5, 177, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 179, 192, 0, 0, 0, 5, 5, 165, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 173, 5, 5, 167, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 179, 192, 0, 0, 0, 5, 5, 168, 173, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 187, 5, 5, 14, 180, 5, 5, 107, 187, 5, 5, 15, 180, 5, 5, 16, 180, 5, 5, 108, 187, 5, 5, 17, 180, 5, 5, 18, 180, 5, 5, 19, 180, 5, 5, 20, 180, 192, 0, 0, 0, 5, 5, 109, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 195, 5, 5, 105, 195, 5, 5, 87, 203, 5, 5, 88, 203, 5, 5, 106, 195, 5, 5, 89, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 218, 5, 5, 75, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 225, 192, 0, 0, 0, 5, 5, 103, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 164, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 168, 5, 5, 161, 168, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 173, 5, 5, 162, 168, 5, 5, 163, 168, 192, 0, 0, 0, 5, 5, 164, 168, 5, 5, 165, 168, 5, 5, 166, 168, 5, 5, 167, 168, 192, 0, 0, 0, 5, 5, 231, 173, 5, 5, 232, 173, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 173, 5, 5, 234, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 180, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 187, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 187, 192, 0, 0, 0, 5, 5, 255, 187, 5, 5, 2, 188, 5, 5, 3, 188, 5, 5, 4, 188, 5, 5, 5, 188, 192, 0, 0, 0, 5, 5, 6, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 195, 5, 5, 189, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 203, 5, 5, 203, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 211, 192, 0, 0, 0, 5, 5, 149, 211, 5, 5, 150, 211, 5, 5, 151, 211, 5, 5, 152, 211, 5, 5, 153, 211, 5, 5, 154, 211, 5, 5, 155, 211, 5, 5, 156, 211, 5, 5, 157, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 225, 192, 0, 0, 0, 5, 5, 44, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 230, 192, 0, 0, 0, 5, 5, 228, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 235, 5, 5, 139, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 168, 192, 0, 0, 0, 5, 5, 247, 173, 5, 5, 248, 173, 5, 5, 249, 173, 192, 0, 0, 0, 5, 5, 109, 180, 5, 5, 110, 180, 5, 5, 111, 180, 5, 5, 26, 188, 5, 5, 27, 188, 5, 5, 28, 188, 5, 5, 29, 188, 5, 5, 30, 188, 5, 5, 212, 195, 5, 5, 213, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 173, 5, 5, 253, 173, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 174, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 180, 192, 0, 0, 0, 5, 5, 47, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 174, 192, 0, 0, 0, 5, 5, 45, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 188, 192, 0, 0, 0, 5, 5, 69, 188, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 196, 5, 5, 249, 203, 5, 5, 250, 203, 5, 5, 251, 203, 5, 5, 252, 203, 5, 5, 253, 203, 5, 5, 254, 203, 5, 5, 255, 203, 5, 5, 2, 204, 5, 5, 3, 204, 5, 5, 4, 204, 5, 5, 5, 204, 5, 5, 6, 204, 5, 5, 191, 211, 5, 5, 192, 211, 5, 5, 193, 211, 5, 5, 194, 211, 5, 5, 173, 218, 5, 5, 174, 218, 5, 5, 175, 218, 5, 5, 176, 218, 5, 5, 177, 218, 5, 5, 178, 218, 5, 5, 87, 225, 5, 5, 88, 225, 5, 5, 89, 225, 5, 5, 90, 225, 5, 5, 91, 225, 5, 5, 92, 225, 5, 5, 93, 225, 5, 5, 94, 225, 5, 5, 95, 225, 5, 5, 248, 230, 5, 5, 249, 230, 5, 5, 250, 230, 5, 5, 151, 235, 5, 5, 152, 235, 5, 5, 153, 235, 5, 5, 154, 235, 5, 5, 155, 235, 5, 5, 156, 235, 5, 5, 60, 239, 5, 5, 61, 239, 5, 5, 97, 242, 5, 5, 139, 246, 5, 5, 140, 246, 5, 5, 208, 247, 5, 5, 110, 249, 5, 5, 76, 161, 5, 5, 102, 225, 5, 5, 253, 230, 5, 5, 221, 168, 5, 5, 159, 180, 5, 5, 160, 180, 5, 5, 80, 188, 5, 5, 81, 188, 5, 5, 19, 196, 5, 5, 52, 199, 5, 5, 201, 211, 5, 5, 202, 211, 5, 5, 183, 218, 5, 5, 184, 218, 5, 5, 185, 218, 5, 5, 105, 225, 5, 5, 254, 230, 5, 5, 97, 161, 5, 5, 135, 164, 5, 5, 136, 164, 5, 5, 137, 164, 5, 5, 138, 164, 5, 5, 139, 164, 5, 5, 35, 169, 5, 5, 36, 169, 5, 5, 37, 169, 5, 5, 38, 169, 5, 5, 39, 169, 5, 5, 40, 169, 5, 5, 41, 169, 5, 5, 42, 169, 5, 5, 43, 169, 5, 5, 44, 169, 5, 5, 148, 174, 5, 5, 149, 174, 5, 5, 150, 174, 5, 5, 151, 174, 5, 5, 152, 174, 5, 5, 153, 174, 5, 5, 154, 174, 5, 5, 253, 180, 5, 5, 254, 180, 5, 5, 255, 180, 5, 5, 2, 181, 5, 5, 3, 181, 5, 5, 4, 181, 5, 5, 5, 181, 5, 5, 6, 181, 5, 5, 7, 181, 5, 5, 8, 181, 5, 5, 9, 181, 5, 5, 10, 181, 5, 5, 11, 181, 5, 5, 12, 181, 5, 5, 13, 181, 5, 5, 14, 181, 5, 5, 169, 188, 5, 5, 170, 188, 5, 5, 171, 188, 5, 5, 172, 188, 5, 5, 173, 188, 5, 5, 174, 188, 5, 5, 175, 188, 5, 5, 176, 188, 5, 5, 177, 188, 5, 5, 178, 188, 5, 5, 179, 188, 5, 5, 180, 188, 5, 5, 181, 188, 5, 5, 182, 188, 5, 5, 129, 196, 5, 5, 183, 188, 5, 5, 130, 196, 5, 5, 131, 196, 5, 5, 132, 196, 5, 5, 133, 196, 5, 5, 134, 196, 5, 5, 135, 196, 5, 5, 136, 196, 5, 5, 137, 196, 5, 5, 138, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 204, 5, 5, 108, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 212, 192, 0, 0, 0, 5, 5, 10, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 219, 192, 0, 0, 0, 5, 5, 12, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 231, 5, 5, 49, 231, 5, 5, 50, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 239, 5, 5, 90, 239, 5, 5, 91, 239, 192, 0, 0, 0, 5, 5, 92, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 242, 5, 5, 114, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 204, 192, 0, 0, 0, 5, 5, 126, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 181, 5, 5, 200, 188, 192, 0, 0, 0, 5, 5, 161, 196, 192, 0, 0, 0, 5, 5, 162, 196, 192, 0, 0, 0, 5, 5, 163, 196, 5, 5, 164, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 219, 5, 5, 34, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 169, 192, 0, 0, 0, 5, 5, 159, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 197, 5, 5, 47, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 197, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 197, 192, 0, 0, 0, 5, 5, 50, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 204, 5, 5, 239, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 204, 192, 0, 0, 0, 5, 5, 154, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 219, 192, 0, 0, 0, 5, 5, 125, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 226, 5, 5, 14, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 231, 192, 0, 0, 0, 5, 5, 108, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 239, 192, 0, 0, 0, 5, 5, 129, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 242, 192, 0, 0, 0, 5, 5, 142, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 197, 5, 5, 103, 197, 192, 0, 0, 0, 5, 5, 104, 197, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 197, 5, 5, 106, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 212, 5, 5, 203, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 226, 5, 5, 66, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 226, 5, 5, 69, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 231, 5, 5, 142, 231, 5, 5, 143, 231, 192, 0, 0, 0, 5, 5, 248, 235, 192, 0, 0, 0, 5, 5, 145, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 181, 192, 0, 0, 0, 5, 5, 213, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 189, 192, 0, 0, 0, 5, 5, 137, 189, 192, 0, 0, 0, 5, 5, 138, 189, 192, 0, 0, 0, 5, 5, 139, 189, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 161, 5, 5, 159, 161, 5, 5, 160, 161, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 181, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 182, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 189, 192, 0, 0, 0, 5, 5, 202, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 189, 5, 5, 199, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 197, 5, 5, 93, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 197, 5, 5, 202, 197, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 205, 5, 5, 132, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 212, 192, 0, 0, 0, 5, 5, 254, 212, 192, 0, 0, 0, 5, 5, 226, 219, 5, 5, 227, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 219, 5, 5, 230, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 231, 192, 0, 0, 0, 5, 5, 109, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 231, 5, 5, 176, 231, 5, 5, 177, 231, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 239, 192, 0, 0, 0, 5, 5, 170, 239, 5, 5, 174, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 219, 192, 0, 0, 0, 5, 5, 28, 165, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 175, 5, 5, 168, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 236, 5, 5, 185, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 182, 5, 5, 2, 190, 5, 5, 3, 190, 5, 5, 24, 198, 5, 5, 202, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 231, 192, 0, 0, 0, 5, 5, 236, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 246, 192, 0, 0, 0, 5, 5, 218, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 176, 192, 0, 0, 0, 5, 5, 166, 182, 5, 5, 167, 182, 5, 5, 71, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 176, 5, 5, 35, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 190, 192, 0, 0, 0, 5, 5, 92, 190, 192, 0, 0, 0, 5, 5, 93, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 220, 192, 0, 0, 0, 5, 5, 104, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 232, 192, 0, 0, 0, 5, 5, 90, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 250, 192, 0, 0, 0, 5, 5, 80, 170, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 198, 5, 5, 151, 198, 5, 5, 152, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 206, 5, 5, 118, 206, 192, 0, 0, 0, 5, 5, 119, 206, 5, 5, 120, 206, 5, 5, 121, 206, 5, 5, 122, 206, 5, 5, 123, 206, 5, 5, 124, 206, 5, 5, 200, 213, 5, 5, 201, 213, 5, 5, 202, 213, 5, 5, 203, 213, 5, 5, 204, 213, 5, 5, 205, 213, 5, 5, 206, 213, 5, 5, 207, 213, 5, 5, 208, 213, 5, 5, 209, 213, 5, 5, 210, 213, 5, 5, 211, 213, 5, 5, 212, 213, 5, 5, 151, 220, 5, 5, 152, 220, 5, 5, 153, 220, 5, 5, 154, 220, 5, 5, 155, 220, 5, 5, 156, 220, 5, 5, 157, 220, 5, 5, 158, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 220, 192, 0, 0, 0, 5, 5, 160, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 220, 192, 0, 0, 0, 5, 5, 162, 220, 5, 5, 163, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 220, 5, 5, 23, 227, 5, 5, 24, 227, 192, 0, 0, 0, 5, 5, 25, 227, 5, 5, 26, 227, 192, 0, 0, 0, 5, 5, 27, 227, 5, 5, 28, 227, 5, 5, 29, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 232, 5, 5, 81, 232, 5, 5, 82, 232, 5, 5, 83, 232, 5, 5, 84, 232, 5, 5, 85, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 236, 5, 5, 242, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 247, 192, 0, 0, 0, 5, 5, 224, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 170, 5, 5, 88, 176, 5, 5, 89, 176, 192, 0, 0, 0, 5, 5, 90, 176, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 214, 5, 5, 35, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 214, 5, 5, 37, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 220, 5, 5, 239, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 220, 5, 5, 241, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 198, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 176, 5, 5, 39, 183, 5, 5, 223, 190, 5, 5, 224, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 198, 5, 5, 249, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 183, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 191, 192, 0, 0, 0, 5, 5, 47, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 199, 5, 5, 54, 199, 5, 5, 55, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 214, 5, 5, 103, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 221, 5, 5, 48, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 236, 5, 5, 193, 236, 5, 5, 194, 236, 5, 5, 38, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 236, 192, 0, 0, 0, 5, 5, 213, 161, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 183, 5, 5, 152, 183, 192, 0, 0, 0, 5, 5, 153, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 207, 192, 0, 0, 0, 5, 5, 18, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 214, 5, 5, 122, 214, 5, 5, 123, 214, 192, 0, 0, 0, 5, 5, 124, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 214, 5, 5, 127, 214, 5, 5, 128, 214, 5, 5, 129, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 227, 5, 5, 142, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 227, 192, 0, 0, 0, 5, 5, 144, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 232, 192, 0, 0, 0, 5, 5, 198, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 232, 5, 5, 201, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 236, 192, 0, 0, 0, 5, 5, 206, 236, 192, 0, 0, 0, 5, 5, 207, 236, 5, 5, 208, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 236, 5, 5, 210, 236, 5, 5, 50, 240, 5, 5, 51, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 240, 5, 5, 10, 243, 192, 0, 0, 0, 5, 5, 11, 243, 5, 5, 12, 243, 192, 0, 0, 0, 5, 5, 13, 243, 5, 5, 14, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 245, 5, 5, 51, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 246, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 248, 5, 5, 231, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 199, 192, 0, 0, 0, 5, 5, 246, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 233, 5, 5, 54, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 243, 5, 5, 64, 243, 5, 5, 90, 245, 5, 5, 91, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 184, 192, 0, 0, 0, 5, 5, 10, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 221, 5, 5, 246, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 192, 5, 5, 35, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 200, 5, 5, 88, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 222, 192, 0, 0, 0, 5, 5, 54, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 249, 192, 0, 0, 0, 5, 5, 67, 192, 5, 5, 68, 192, 5, 5, 103, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 228, 5, 5, 93, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 184, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 200, 192, 0, 0, 0, 5, 5, 155, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 222, 5, 5, 140, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 192, 5, 5, 128, 192, 5, 5, 129, 192, 5, 5, 130, 192, 5, 5, 131, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 228, 5, 5, 188, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 201, 192, 0, 0, 0, 5, 5, 73, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 216, 5, 5, 39, 216, 5, 5, 40, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 223, 192, 0, 0, 0, 5, 5, 38, 223, 5, 5, 39, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 234, 192, 0, 0, 0, 5, 5, 216, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 237, 5, 5, 218, 237, 192, 0, 0, 0, 5, 5, 219, 237, 192, 0, 0, 0, 5, 5, 39, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 241, 192, 0, 0, 0, 5, 5, 41, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 223, 192, 0, 0, 0, 5, 5, 70, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 227, 5, 5, 68, 234, 5, 5, 69, 234, 5, 5, 255, 237, 5, 5, 62, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 234, 192, 0, 0, 0, 5, 5, 7, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 234, 5, 5, 220, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 216, 5, 5, 138, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 238, 192, 0, 0, 0, 5, 5, 53, 238, 5, 5, 54, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 216, 5, 5, 135, 234, 192, 0, 0, 0, 5, 5, 136, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 234, 192, 0, 0, 0, 5, 5, 154, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 238, 5, 5, 86, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 249, 5, 5, 9, 250, 192, 0, 0, 0, 5, 5, 109, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 238, 5, 5, 100, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 223, 192, 0, 0, 0, 5, 5, 226, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 229, 192, 0, 0, 0, 5, 5, 228, 229, 192, 0, 0, 0, 5, 5, 199, 234, 5, 5, 200, 234, 5, 5, 201, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 246, 192, 0, 0, 0, 5, 5, 17, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 246, 5, 5, 20, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 224, 192, 0, 0, 0, 5, 5, 12, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 241, 5, 5, 246, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 249, 192, 0, 0, 0, 5, 5, 84, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 230, 192, 0, 0, 0, 5, 5, 191, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 242, 5, 5, 123, 244, 5, 5, 66, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 238, 5, 5, 13, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 128, 5, 5, 0, 151, 5, 5, 0, 152, 5, 5, 0, 153, 5, 5, 0, 154, 5, 5, 0, 155, 5, 5, 0, 156, 5, 5, 54, 157, 5, 5, 55, 157, 5, 5, 56, 157, 5, 5, 57, 157, 5, 5, 58, 157, 5, 5, 59, 157, 5, 5, 164, 157, 5, 5, 165, 157, 5, 5, 166, 157, 5, 5, 167, 157, 5, 5, 168, 157, 5, 5, 169, 157, 5, 5, 170, 157, 5, 5, 128, 158, 5, 5, 129, 158, 5, 5, 130, 158, 5, 5, 127, 158, 5, 5, 131, 158, 5, 5, 132, 158, 5, 5, 133, 158, 5, 5, 134, 158, 5, 5, 135, 158, 5, 5, 136, 158, 5, 5, 214, 159, 5, 5, 215, 159, 5, 5, 216, 159, 5, 5, 217, 159, 5, 5, 218, 159, 5, 5, 33, 162, 5, 5, 34, 162, 5, 5, 35, 162, 5, 5, 254, 165, 5, 5, 255, 165, 5, 5, 0, 129, 5, 5, 2, 157, 5, 5, 62, 157, 5, 5, 63, 157, 5, 5, 94, 158, 5, 5, 171, 157, 5, 5, 172, 157, 5, 5, 173, 157, 5, 5, 174, 157, 5, 5, 138, 158, 5, 5, 38, 162, 5, 5, 2, 166, 5, 5, 145, 171, 5, 5, 240, 177, 5, 5, 0, 130, 5, 5, 3, 157, 5, 5, 65, 157, 5, 5, 175, 157, 5, 5, 176, 157, 5, 5, 139, 158, 5, 5, 140, 158, 5, 5, 36, 162, 5, 5, 146, 171, 5, 5, 0, 131, 5, 5, 0, 132, 5, 5, 0, 133, 5, 5, 4, 157, 5, 5, 5, 157, 5, 5, 6, 157, 5, 5, 67, 157, 5, 5, 68, 157, 5, 5, 69, 157, 5, 5, 70, 157, 5, 5, 66, 157, 5, 5, 71, 157, 5, 5, 177, 157, 5, 5, 178, 157, 5, 5, 141, 158, 5, 5, 142, 158, 5, 5, 143, 158, 5, 5, 144, 158, 5, 5, 220, 159, 5, 5, 221, 159, 5, 5, 222, 159, 5, 5, 223, 159, 5, 5, 40, 162, 5, 5, 3, 166, 5, 5, 147, 171, 5, 5, 241, 177, 5, 5, 0, 134, 5, 5, 0, 135, 5, 5, 0, 137, 5, 5, 10, 157, 5, 5, 11, 157, 5, 5, 72, 157, 5, 5, 73, 157, 5, 5, 74, 157, 5, 5, 134, 157, 5, 5, 35, 158, 5, 5, 180, 157, 5, 5, 181, 157, 5, 5, 182, 157, 5, 5, 183, 157, 5, 5, 146, 158, 5, 5, 224, 159, 5, 5, 225, 159, 5, 5, 226, 159, 5, 5, 227, 159, 5, 5, 228, 159, 5, 5, 229, 159, 5, 5, 230, 159, 5, 5, 231, 159, 5, 5, 232, 159, 5, 5, 41, 162, 5, 5, 42, 162, 5, 5, 4, 166, 5, 5, 5, 166, 5, 5, 6, 166, 5, 5, 7, 166, 5, 5, 8, 166, 5, 5, 9, 166, 5, 5, 148, 171, 5, 5, 149, 171, 5, 5, 150, 171, 5, 5, 151, 171, 5, 5, 242, 177, 5, 5, 80, 185, 5, 5, 81, 185, 5, 5, 82, 185, 5, 5, 112, 193, 5, 5, 239, 201, 5, 5, 240, 201, 5, 5, 241, 201, 5, 5, 0, 143, 5, 5, 12, 157, 5, 5, 75, 157, 5, 5, 184, 157, 5, 5, 233, 159, 5, 5, 43, 162, 5, 5, 11, 166, 5, 5, 14, 157, 5, 5, 76, 157, 5, 5, 77, 157, 5, 5, 78, 157, 5, 5, 60, 157, 5, 5, 185, 157, 5, 5, 186, 157, 5, 5, 187, 157, 5, 5, 188, 157, 5, 5, 189, 157, 5, 5, 190, 157, 5, 5, 147, 158, 5, 5, 234, 159, 5, 5, 235, 159, 5, 5, 236, 159, 5, 5, 12, 166, 5, 5, 45, 162, 5, 5, 13, 166, 5, 5, 14, 166, 5, 5, 15, 166, 5, 5, 15, 157, 5, 5, 79, 157, 5, 5, 191, 157, 5, 5, 192, 157, 5, 5, 237, 159, 5, 5, 238, 159, 5, 5, 239, 159, 5, 5, 240, 159, 5, 5, 46, 162, 5, 5, 47, 162, 5, 5, 48, 162, 5, 5, 16, 166, 5, 5, 17, 166, 5, 5, 153, 171, 5, 5, 154, 171, 5, 5, 155, 171, 5, 5, 156, 171, 5, 5, 157, 171, 5, 5, 158, 171, 5, 5, 243, 177, 5, 5, 113, 193, 5, 5, 114, 193, 5, 5, 242, 201, 5, 5, 243, 201, 5, 5, 45, 224, 5, 5, 91, 246, 5, 5, 16, 157, 5, 5, 17, 157, 5, 5, 80, 157, 5, 5, 81, 157, 5, 5, 82, 157, 5, 5, 83, 157, 5, 5, 193, 157, 5, 5, 194, 157, 5, 5, 195, 157, 5, 5, 196, 157, 5, 5, 197, 157, 5, 5, 198, 157, 5, 5, 199, 157, 5, 5, 200, 157, 5, 5, 201, 157, 5, 5, 202, 157, 5, 5, 203, 157, 5, 5, 204, 157, 5, 5, 205, 157, 5, 5, 206, 157, 5, 5, 207, 157, 5, 5, 208, 157, 5, 5, 209, 157, 5, 5, 210, 157, 5, 5, 211, 157, 5, 5, 212, 157, 5, 5, 148, 158, 5, 5, 149, 158, 5, 5, 150, 158, 5, 5, 151, 158, 5, 5, 152, 158, 5, 5, 153, 158, 5, 5, 154, 158, 5, 5, 155, 158, 5, 5, 156, 158, 5, 5, 157, 158, 5, 5, 158, 158, 5, 5, 159, 158, 5, 5, 160, 158, 5, 5, 161, 158, 5, 5, 162, 158, 5, 5, 163, 158, 5, 5, 164, 158, 5, 5, 165, 158, 5, 5, 166, 158, 5, 5, 167, 158, 5, 5, 168, 158, 5, 5, 169, 158, 5, 5, 170, 158, 5, 5, 171, 158, 5, 5, 172, 158, 5, 5, 173, 158, 5, 5, 242, 159, 5, 5, 243, 159, 5, 5, 244, 159, 5, 5, 245, 159, 5, 5, 246, 159, 5, 5, 247, 159, 5, 5, 248, 159, 5, 5, 249, 159, 5, 5, 250, 159, 5, 5, 251, 159, 5, 5, 252, 159, 5, 5, 253, 159, 5, 5, 254, 159, 5, 5, 255, 159, 5, 5, 2, 160, 5, 5, 3, 160, 5, 5, 4, 160, 5, 5, 5, 160, 5, 5, 6, 160, 5, 5, 7, 160, 5, 5, 8, 160, 5, 5, 9, 160, 5, 5, 10, 160, 5, 5, 11, 160, 5, 5, 12, 160, 5, 5, 13, 160, 5, 5, 14, 160, 5, 5, 15, 160, 5, 5, 16, 160, 5, 5, 17, 160, 5, 5, 18, 160, 5, 5, 19, 160, 5, 5, 20, 160, 5, 5, 21, 160, 5, 5, 22, 160, 5, 5, 23, 160, 5, 5, 24, 160, 5, 5, 25, 160, 5, 5, 26, 160, 5, 5, 27, 160, 5, 5, 28, 160, 5, 5, 29, 160, 5, 5, 30, 160, 5, 5, 31, 160, 5, 5, 32, 160, 5, 5, 33, 160, 5, 5, 34, 160, 5, 5, 35, 160, 5, 5, 36, 160, 5, 5, 37, 160, 5, 5, 38, 160, 5, 5, 39, 160, 5, 5, 40, 160, 5, 5, 41, 160, 5, 5, 42, 160, 5, 5, 43, 160, 5, 5, 44, 160, 5, 5, 45, 160, 5, 5, 46, 160, 5, 5, 47, 160, 5, 5, 48, 160, 5, 5, 49, 160, 5, 5, 50, 160, 5, 5, 49, 162, 5, 5, 50, 162, 5, 5, 51, 162, 5, 5, 52, 162, 5, 5, 53, 162, 5, 5, 54, 162, 5, 5, 55, 162, 5, 5, 56, 162, 5, 5, 57, 162, 5, 5, 58, 162, 5, 5, 59, 162, 5, 5, 60, 162, 5, 5, 61, 162, 5, 5, 62, 162, 5, 5, 63, 162, 5, 5, 64, 162, 5, 5, 65, 162, 5, 5, 66, 162, 5, 5, 67, 162, 5, 5, 68, 162, 5, 5, 69, 162, 5, 5, 70, 162, 5, 5, 71, 162, 5, 5, 72, 162, 5, 5, 73, 162, 5, 5, 74, 162, 5, 5, 75, 162, 5, 5, 76, 162, 5, 5, 77, 162, 5, 5, 78, 162, 5, 5, 79, 162, 5, 5, 19, 166, 5, 5, 80, 162, 5, 5, 81, 162, 5, 5, 82, 162, 5, 5, 83, 162, 5, 5, 84, 162, 5, 5, 85, 162, 5, 5, 86, 162, 5, 5, 87, 162, 5, 5, 88, 162, 5, 5, 89, 162, 5, 5, 90, 162, 5, 5, 91, 162, 5, 5, 92, 162, 5, 5, 93, 162, 5, 5, 94, 162, 5, 5, 95, 162, 5, 5, 96, 162, 5, 5, 97, 162, 5, 5, 98, 162, 5, 5, 99, 162, 5, 5, 100, 162, 5, 5, 101, 162, 5, 5, 102, 162, 5, 5, 51, 160, 5, 5, 103, 162, 5, 5, 104, 162, 5, 5, 105, 162, 5, 5, 106, 162, 5, 5, 20, 166, 5, 5, 21, 166, 5, 5, 22, 166, 5, 5, 23, 166, 5, 5, 24, 166, 5, 5, 25, 166, 5, 5, 26, 166, 5, 5, 27, 166, 5, 5, 28, 166, 5, 5, 29, 166, 5, 5, 30, 166, 5, 5, 31, 166, 5, 5, 32, 166, 5, 5, 33, 166, 5, 5, 34, 166, 5, 5, 35, 166, 5, 5, 36, 166, 5, 5, 37, 166, 5, 5, 38, 166, 5, 5, 39, 166, 5, 5, 40, 166, 5, 5, 41, 166, 5, 5, 42, 166, 5, 5, 43, 166, 5, 5, 44, 166, 5, 5, 45, 166, 5, 5, 46, 166, 5, 5, 47, 166, 5, 5, 48, 166, 5, 5, 49, 166, 5, 5, 50, 166, 5, 5, 51, 166, 5, 5, 52, 166, 5, 5, 53, 166, 5, 5, 54, 166, 5, 5, 55, 166, 5, 5, 56, 166, 5, 5, 57, 166, 5, 5, 58, 166, 5, 5, 59, 166, 5, 5, 60, 166, 5, 5, 61, 166, 5, 5, 62, 166, 5, 5, 63, 166, 5, 5, 64, 166, 5, 5, 65, 166, 5, 5, 66, 166, 5, 5, 67, 166, 5, 5, 68, 166, 5, 5, 69, 166, 5, 5, 70, 166, 5, 5, 71, 166, 5, 5, 72, 166, 5, 5, 73, 166, 5, 5, 74, 166, 5, 5, 75, 166, 5, 5, 76, 166, 5, 5, 77, 166, 5, 5, 78, 166, 5, 5, 79, 166, 5, 5, 80, 166, 5, 5, 81, 166, 5, 5, 82, 166, 5, 5, 83, 166, 5, 5, 84, 166, 5, 5, 85, 166, 5, 5, 86, 166, 5, 5, 87, 166, 5, 5, 88, 166, 5, 5, 159, 171, 5, 5, 160, 171, 5, 5, 161, 171, 5, 5, 162, 171, 5, 5, 163, 171, 5, 5, 164, 171, 5, 5, 165, 171, 5, 5, 166, 171, 5, 5, 167, 171, 5, 5, 168, 171, 5, 5, 169, 171, 5, 5, 170, 171, 5, 5, 171, 171, 5, 5, 172, 171, 5, 5, 173, 171, 5, 5, 174, 171, 5, 5, 175, 171, 5, 5, 176, 171, 5, 5, 177, 171, 5, 5, 178, 171, 5, 5, 179, 171, 5, 5, 180, 171, 5, 5, 181, 171, 5, 5, 182, 171, 5, 5, 183, 171, 5, 5, 184, 171, 5, 5, 185, 171, 5, 5, 186, 171, 5, 5, 187, 171, 5, 5, 188, 171, 5, 5, 189, 171, 5, 5, 190, 171, 5, 5, 191, 171, 5, 5, 192, 171, 5, 5, 193, 171, 5, 5, 194, 171, 5, 5, 195, 171, 5, 5, 196, 171, 5, 5, 197, 171, 5, 5, 198, 171, 5, 5, 199, 171, 5, 5, 200, 171, 5, 5, 201, 171, 5, 5, 202, 171, 5, 5, 203, 171, 5, 5, 204, 171, 5, 5, 205, 171, 5, 5, 206, 171, 5, 5, 207, 171, 5, 5, 208, 171, 5, 5, 209, 171, 5, 5, 210, 171, 5, 5, 211, 171, 5, 5, 212, 171, 5, 5, 213, 171, 5, 5, 214, 171, 5, 5, 215, 171, 5, 5, 216, 171, 5, 5, 217, 171, 5, 5, 218, 171, 5, 5, 219, 171, 5, 5, 220, 171, 5, 5, 221, 171, 5, 5, 222, 171, 5, 5, 244, 177, 5, 5, 245, 177, 5, 5, 246, 177, 5, 5, 247, 177, 5, 5, 248, 177, 5, 5, 249, 177, 5, 5, 250, 177, 5, 5, 251, 177, 5, 5, 252, 177, 5, 5, 253, 177, 5, 5, 254, 177, 5, 5, 255, 177, 5, 5, 2, 178, 5, 5, 3, 178, 5, 5, 4, 178, 5, 5, 5, 178, 5, 5, 6, 178, 5, 5, 7, 178, 5, 5, 8, 178, 5, 5, 9, 178, 5, 5, 10, 178, 5, 5, 11, 178, 5, 5, 12, 178, 5, 5, 13, 178, 5, 5, 14, 178, 5, 5, 15, 178, 5, 5, 16, 178, 5, 5, 17, 178, 5, 5, 18, 178, 5, 5, 19, 178, 5, 5, 20, 178, 5, 5, 21, 178, 5, 5, 22, 178, 5, 5, 23, 178, 5, 5, 24, 178, 5, 5, 25, 178, 5, 5, 26, 178, 5, 5, 27, 178, 5, 5, 28, 178, 5, 5, 29, 178, 5, 5, 30, 178, 5, 5, 31, 178, 5, 5, 32, 178, 5, 5, 33, 178, 5, 5, 34, 178, 5, 5, 35, 178, 5, 5, 36, 178, 5, 5, 37, 178, 5, 5, 38, 178, 5, 5, 39, 178, 5, 5, 40, 178, 5, 5, 41, 178, 5, 5, 42, 178, 5, 5, 43, 178, 5, 5, 44, 178, 5, 5, 45, 178, 5, 5, 46, 178, 5, 5, 47, 178, 5, 5, 48, 178, 5, 5, 49, 178, 5, 5, 50, 178, 5, 5, 51, 178, 5, 5, 52, 178, 5, 5, 53, 178, 5, 5, 54, 178, 5, 5, 55, 178, 5, 5, 56, 178, 5, 5, 57, 178, 5, 5, 58, 178, 5, 5, 59, 178, 5, 5, 60, 178, 5, 5, 61, 178, 5, 5, 62, 178, 5, 5, 63, 178, 5, 5, 64, 178, 5, 5, 65, 178, 5, 5, 66, 178, 5, 5, 67, 178, 5, 5, 68, 178, 5, 5, 69, 178, 5, 5, 70, 178, 5, 5, 71, 178, 5, 5, 84, 185, 5, 5, 85, 185, 5, 5, 86, 185, 5, 5, 87, 185, 5, 5, 88, 185, 5, 5, 89, 185, 5, 5, 90, 185, 5, 5, 91, 185, 5, 5, 92, 185, 5, 5, 93, 185, 5, 5, 94, 185, 5, 5, 95, 185, 5, 5, 96, 185, 5, 5, 97, 185, 5, 5, 98, 185, 5, 5, 99, 185, 5, 5, 100, 185, 5, 5, 101, 185, 5, 5, 102, 185, 5, 5, 103, 185, 5, 5, 104, 185, 5, 5, 105, 185, 5, 5, 79, 178, 5, 5, 106, 185, 5, 5, 107, 185, 5, 5, 108, 185, 5, 5, 109, 185, 5, 5, 110, 185, 5, 5, 111, 185, 5, 5, 112, 185, 5, 5, 113, 185, 5, 5, 114, 185, 5, 5, 115, 185, 5, 5, 116, 185, 5, 5, 117, 185, 5, 5, 118, 185, 5, 5, 119, 185, 5, 5, 120, 185, 5, 5, 121, 185, 5, 5, 122, 185, 5, 5, 115, 193, 5, 5, 123, 185, 5, 5, 124, 185, 5, 5, 125, 185, 5, 5, 126, 185, 5, 5, 127, 185, 5, 5, 128, 185, 5, 5, 129, 185, 5, 5, 130, 185, 5, 5, 131, 185, 5, 5, 132, 185, 5, 5, 133, 185, 5, 5, 134, 185, 5, 5, 135, 185, 5, 5, 136, 185, 5, 5, 137, 185, 5, 5, 138, 185, 5, 5, 139, 185, 5, 5, 140, 185, 5, 5, 141, 185, 5, 5, 142, 185, 5, 5, 143, 185, 5, 5, 144, 185, 5, 5, 145, 185, 5, 5, 116, 193, 5, 5, 117, 193, 5, 5, 118, 193, 5, 5, 119, 193, 5, 5, 120, 193, 5, 5, 121, 193, 5, 5, 122, 193, 5, 5, 123, 193, 5, 5, 124, 193, 5, 5, 125, 193, 5, 5, 126, 193, 5, 5, 127, 193, 5, 5, 128, 193, 5, 5, 129, 193, 5, 5, 130, 193, 5, 5, 131, 193, 5, 5, 132, 193, 5, 5, 133, 193, 5, 5, 134, 193, 5, 5, 135, 193, 5, 5, 136, 193, 5, 5, 137, 193, 5, 5, 138, 193, 5, 5, 139, 193, 5, 5, 140, 193, 5, 5, 141, 193, 5, 5, 142, 193, 5, 5, 143, 193, 5, 5, 144, 193, 5, 5, 145, 193, 5, 5, 146, 193, 5, 5, 147, 193, 5, 5, 148, 193, 5, 5, 149, 193, 5, 5, 150, 193, 5, 5, 151, 193, 5, 5, 152, 193, 5, 5, 153, 193, 5, 5, 154, 193, 5, 5, 155, 193, 5, 5, 156, 193, 5, 5, 157, 193, 5, 5, 244, 201, 5, 5, 245, 201, 5, 5, 246, 201, 5, 5, 247, 201, 5, 5, 248, 201, 5, 5, 249, 201, 5, 5, 250, 201, 5, 5, 251, 201, 5, 5, 252, 201, 5, 5, 253, 201, 5, 5, 254, 201, 5, 5, 255, 201, 5, 5, 2, 202, 5, 5, 3, 202, 5, 5, 4, 202, 5, 5, 5, 202, 5, 5, 6, 202, 5, 5, 7, 202, 5, 5, 8, 202, 5, 5, 9, 202, 5, 5, 10, 202, 5, 5, 11, 202, 5, 5, 12, 202, 5, 5, 13, 202, 5, 5, 14, 202, 5, 5, 15, 202, 5, 5, 16, 202, 5, 5, 17, 202, 5, 5, 18, 202, 5, 5, 19, 202, 5, 5, 20, 202, 5, 5, 21, 202, 5, 5, 22, 202, 5, 5, 23, 202, 5, 5, 24, 202, 5, 5, 25, 202, 5, 5, 213, 209, 5, 5, 214, 209, 5, 5, 215, 209, 5, 5, 216, 209, 5, 5, 217, 209, 5, 5, 218, 209, 5, 5, 219, 209, 5, 5, 220, 209, 5, 5, 221, 209, 5, 5, 222, 209, 5, 5, 223, 209, 5, 5, 224, 209, 5, 5, 225, 209, 5, 5, 226, 209, 5, 5, 227, 209, 5, 5, 228, 209, 5, 5, 229, 209, 5, 5, 230, 209, 5, 5, 231, 209, 5, 5, 232, 209, 5, 5, 233, 209, 5, 5, 234, 209, 5, 5, 235, 209, 5, 5, 236, 209, 5, 5, 237, 209, 5, 5, 238, 209, 5, 5, 239, 209, 5, 5, 240, 209, 5, 5, 241, 209, 5, 5, 242, 209, 5, 5, 243, 209, 5, 5, 244, 209, 5, 5, 245, 209, 5, 5, 246, 209, 5, 5, 247, 209, 5, 5, 248, 209, 5, 5, 3, 210, 5, 5, 249, 209, 5, 5, 250, 209, 5, 5, 249, 216, 5, 5, 250, 216, 5, 5, 251, 209, 5, 5, 251, 216, 5, 5, 252, 216, 5, 5, 253, 216, 5, 5, 254, 216, 5, 5, 255, 216, 5, 5, 2, 217, 5, 5, 3, 217, 5, 5, 4, 217, 5, 5, 5, 217, 5, 5, 6, 217, 5, 5, 7, 217, 5, 5, 8, 217, 5, 5, 9, 217, 5, 5, 10, 217, 5, 5, 11, 217, 5, 5, 12, 217, 5, 5, 13, 217, 5, 5, 14, 217, 5, 5, 15, 217, 5, 5, 16, 217, 5, 5, 17, 217, 5, 5, 18, 217, 5, 5, 19, 217, 5, 5, 20, 217, 5, 5, 46, 224, 5, 5, 47, 224, 5, 5, 48, 224, 5, 5, 49, 224, 5, 5, 50, 224, 5, 5, 51, 224, 5, 5, 52, 224, 5, 5, 53, 224, 5, 5, 54, 224, 5, 5, 55, 224, 5, 5, 56, 224, 5, 5, 57, 224, 5, 5, 58, 224, 5, 5, 59, 224, 5, 5, 60, 224, 5, 5, 54, 230, 5, 5, 55, 230, 5, 5, 56, 230, 5, 5, 57, 230, 5, 5, 58, 230, 5, 5, 59, 230, 5, 5, 60, 230, 5, 5, 61, 230, 5, 5, 62, 230, 5, 5, 63, 230, 5, 5, 64, 230, 5, 5, 65, 230, 5, 5, 61, 224, 5, 5, 66, 230, 5, 5, 34, 235, 5, 5, 35, 235, 5, 5, 36, 235, 5, 5, 26, 217, 5, 5, 37, 235, 5, 5, 70, 230, 5, 5, 217, 238, 5, 5, 218, 238, 5, 5, 219, 238, 5, 5, 25, 242, 5, 5, 148, 244, 5, 5, 149, 244, 5, 5, 150, 244, 5, 5, 151, 244, 5, 5, 92, 246, 5, 5, 93, 246, 5, 5, 176, 247, 5, 5, 182, 248, 5, 5, 18, 157, 5, 5, 85, 157, 5, 5, 213, 157, 5, 5, 214, 157, 5, 5, 215, 157, 5, 5, 184, 158, 5, 5, 185, 158, 5, 5, 59, 160, 5, 5, 60, 160, 5, 5, 61, 160, 5, 5, 62, 160, 5, 5, 63, 160, 5, 5, 109, 162, 5, 5, 110, 162, 5, 5, 111, 162, 5, 5, 112, 162, 5, 5, 113, 162, 5, 5, 114, 162, 5, 5, 115, 162, 5, 5, 96, 166, 5, 5, 97, 166, 5, 5, 98, 166, 5, 5, 99, 166, 5, 5, 100, 166, 5, 5, 225, 171, 5, 5, 226, 171, 5, 5, 227, 171, 5, 5, 80, 178, 5, 5, 81, 178, 5, 5, 151, 185, 5, 5, 152, 185, 5, 5, 153, 185, 5, 5, 162, 193, 5, 5, 163, 193, 5, 5, 33, 202, 5, 5, 4, 210, 5, 5, 64, 224, 5, 5, 153, 244, 5, 5, 19, 157, 5, 5, 86, 157, 5, 5, 216, 157, 5, 5, 64, 160, 5, 5, 102, 166, 5, 5, 229, 171, 5, 5, 20, 157, 5, 5, 217, 157, 5, 5, 218, 157, 5, 5, 219, 157, 5, 5, 220, 157, 5, 5, 187, 158, 5, 5, 66, 160, 5, 5, 67, 160, 5, 5, 68, 160, 5, 5, 69, 160, 5, 5, 117, 162, 5, 5, 103, 166, 5, 5, 104, 166, 5, 5, 105, 166, 5, 5, 230, 171, 5, 5, 84, 178, 5, 5, 231, 171, 5, 5, 85, 178, 5, 5, 154, 185, 5, 5, 34, 202, 5, 5, 35, 202, 5, 5, 66, 224, 5, 5, 40, 235, 5, 5, 22, 157, 5, 5, 221, 157, 5, 5, 222, 157, 5, 5, 223, 157, 5, 5, 224, 157, 5, 5, 225, 157, 5, 5, 226, 157, 5, 5, 188, 158, 5, 5, 189, 158, 5, 5, 190, 158, 5, 5, 191, 158, 5, 5, 70, 160, 5, 5, 71, 160, 5, 5, 118, 162, 5, 5, 106, 166, 5, 5, 232, 171, 5, 5, 233, 171, 5, 5, 86, 178, 5, 5, 87, 178, 5, 5, 155, 185, 5, 5, 23, 157, 5, 5, 227, 157, 5, 5, 228, 157, 5, 5, 193, 158, 5, 5, 194, 158, 5, 5, 72, 160, 5, 5, 73, 160, 5, 5, 119, 162, 5, 5, 107, 166, 5, 5, 234, 171, 5, 5, 235, 171, 5, 5, 88, 178, 5, 5, 89, 178, 5, 5, 90, 178, 5, 5, 91, 178, 5, 5, 92, 178, 5, 5, 93, 178, 5, 5, 94, 178, 5, 5, 157, 185, 5, 5, 5, 210, 5, 5, 67, 224, 5, 5, 24, 157, 5, 5, 195, 158, 5, 5, 196, 158, 5, 5, 197, 158, 5, 5, 198, 158, 5, 5, 74, 160, 5, 5, 75, 160, 5, 5, 76, 160, 5, 5, 77, 160, 5, 5, 78, 160, 5, 5, 120, 162, 5, 5, 121, 162, 5, 5, 122, 162, 5, 5, 123, 162, 5, 5, 124, 162, 5, 5, 125, 162, 5, 5, 126, 162, 5, 5, 108, 166, 5, 5, 109, 166, 5, 5, 110, 166, 5, 5, 111, 166, 5, 5, 112, 166, 5, 5, 236, 171, 5, 5, 237, 171, 5, 5, 238, 171, 5, 5, 96, 178, 5, 5, 97, 178, 5, 5, 98, 178, 5, 5, 99, 178, 5, 5, 100, 178, 5, 5, 101, 178, 5, 5, 102, 178, 5, 5, 103, 178, 5, 5, 104, 178, 5, 5, 105, 178, 5, 5, 106, 178, 5, 5, 158, 185, 5, 5, 159, 185, 5, 5, 160, 185, 5, 5, 166, 193, 5, 5, 167, 193, 5, 5, 168, 193, 5, 5, 169, 193, 5, 5, 170, 193, 5, 5, 36, 202, 5, 5, 6, 210, 5, 5, 27, 217, 5, 5, 28, 217, 5, 5, 29, 217, 5, 5, 30, 217, 5, 5, 68, 224, 5, 5, 69, 224, 5, 5, 71, 230, 5, 5, 25, 157, 5, 5, 87, 157, 5, 5, 88, 157, 5, 5, 89, 157, 5, 5, 230, 157, 5, 5, 199, 158, 5, 5, 200, 158, 5, 5, 201, 158, 5, 5, 80, 160, 5, 5, 81, 160, 5, 5, 82, 160, 5, 5, 83, 160, 5, 5, 128, 162, 5, 5, 114, 166, 5, 5, 115, 166, 5, 5, 116, 166, 5, 5, 162, 185, 5, 5, 171, 193, 5, 5, 172, 193, 5, 5, 7, 210, 5, 5, 8, 210, 5, 5, 27, 157, 5, 5, 232, 157, 5, 5, 202, 158, 5, 5, 203, 158, 5, 5, 204, 158, 5, 5, 205, 158, 5, 5, 206, 158, 5, 5, 84, 160, 5, 5, 117, 166, 5, 5, 242, 171, 5, 5, 173, 193, 5, 5, 28, 157, 5, 5, 29, 157, 5, 5, 31, 157, 5, 5, 90, 157, 5, 5, 91, 157, 5, 5, 234, 157, 5, 5, 235, 157, 5, 5, 236, 157, 5, 5, 237, 157, 5, 5, 207, 158, 5, 5, 208, 158, 5, 5, 209, 158, 5, 5, 210, 158, 5, 5, 211, 158, 5, 5, 85, 160, 5, 5, 86, 160, 5, 5, 87, 160, 5, 5, 88, 160, 5, 5, 89, 160, 5, 5, 90, 160, 5, 5, 91, 160, 5, 5, 92, 160, 5, 5, 93, 160, 5, 5, 94, 160, 5, 5, 95, 160, 5, 5, 96, 160, 5, 5, 97, 160, 5, 5, 98, 160, 5, 5, 129, 162, 5, 5, 130, 162, 5, 5, 131, 162, 5, 5, 132, 162, 5, 5, 133, 162, 5, 5, 134, 162, 5, 5, 135, 162, 5, 5, 136, 162, 5, 5, 137, 162, 5, 5, 138, 162, 5, 5, 139, 162, 5, 5, 140, 162, 5, 5, 141, 162, 5, 5, 142, 162, 5, 5, 143, 162, 5, 5, 144, 162, 5, 5, 145, 162, 5, 5, 146, 162, 5, 5, 118, 166, 5, 5, 119, 166, 5, 5, 120, 166, 5, 5, 121, 166, 5, 5, 122, 166, 5, 5, 123, 166, 5, 5, 124, 166, 5, 5, 125, 166, 5, 5, 126, 166, 5, 5, 127, 166, 5, 5, 128, 166, 5, 5, 129, 166, 5, 5, 130, 166, 5, 5, 131, 166, 5, 5, 132, 166, 5, 5, 133, 166, 5, 5, 134, 166, 5, 5, 135, 166, 5, 5, 136, 166, 5, 5, 137, 166, 5, 5, 138, 166, 5, 5, 243, 171, 5, 5, 244, 171, 5, 5, 245, 171, 5, 5, 142, 166, 5, 5, 246, 171, 5, 5, 247, 171, 5, 5, 248, 171, 5, 5, 249, 171, 5, 5, 250, 171, 5, 5, 251, 171, 5, 5, 252, 171, 5, 5, 253, 171, 5, 5, 254, 171, 5, 5, 255, 171, 5, 5, 2, 172, 5, 5, 110, 178, 5, 5, 111, 178, 5, 5, 112, 178, 5, 5, 113, 178, 5, 5, 114, 178, 5, 5, 115, 178, 5, 5, 116, 178, 5, 5, 117, 178, 5, 5, 118, 178, 5, 5, 119, 178, 5, 5, 120, 178, 5, 5, 121, 178, 5, 5, 122, 178, 5, 5, 123, 178, 5, 5, 124, 178, 5, 5, 125, 178, 5, 5, 126, 178, 5, 5, 127, 178, 5, 5, 128, 178, 5, 5, 129, 178, 5, 5, 130, 178, 5, 5, 131, 178, 5, 5, 163, 185, 5, 5, 174, 193, 5, 5, 164, 185, 5, 5, 165, 185, 5, 5, 166, 185, 5, 5, 167, 185, 5, 5, 168, 185, 5, 5, 169, 185, 5, 5, 170, 185, 5, 5, 171, 185, 5, 5, 175, 193, 5, 5, 176, 193, 5, 5, 177, 193, 5, 5, 178, 193, 5, 5, 172, 185, 5, 5, 37, 202, 5, 5, 38, 202, 5, 5, 39, 202, 5, 5, 40, 202, 5, 5, 41, 202, 5, 5, 42, 202, 5, 5, 43, 202, 5, 5, 44, 202, 5, 5, 45, 202, 5, 5, 9, 210, 5, 5, 10, 210, 5, 5, 11, 210, 5, 5, 12, 210, 5, 5, 13, 210, 5, 5, 32, 217, 5, 5, 33, 217, 5, 5, 34, 217, 5, 5, 35, 217, 5, 5, 36, 217, 5, 5, 37, 217, 5, 5, 38, 217, 5, 5, 39, 217, 5, 5, 40, 217, 5, 5, 41, 217, 5, 5, 42, 217, 5, 5, 71, 224, 5, 5, 72, 224, 5, 5, 73, 224, 5, 5, 74, 224, 5, 5, 75, 224, 5, 5, 74, 230, 5, 5, 221, 238, 5, 5, 154, 244, 5, 5, 155, 244, 5, 5, 177, 247, 5, 5, 178, 247, 5, 5, 32, 157, 5, 5, 92, 157, 5, 5, 238, 157, 5, 5, 239, 157, 5, 5, 212, 158, 5, 5, 213, 158, 5, 5, 214, 158, 5, 5, 215, 158, 5, 5, 99, 160, 5, 5, 100, 160, 5, 5, 101, 160, 5, 5, 102, 160, 5, 5, 103, 160, 5, 5, 104, 160, 5, 5, 148, 162, 5, 5, 149, 162, 5, 5, 150, 162, 5, 5, 151, 162, 5, 5, 152, 162, 5, 5, 153, 162, 5, 5, 154, 162, 5, 5, 155, 162, 5, 5, 156, 162, 5, 5, 157, 162, 5, 5, 158, 162, 5, 5, 159, 162, 5, 5, 143, 166, 5, 5, 144, 166, 5, 5, 145, 166, 5, 5, 146, 166, 5, 5, 147, 166, 5, 5, 148, 166, 5, 5, 149, 166, 5, 5, 150, 166, 5, 5, 151, 166, 5, 5, 152, 166, 5, 5, 153, 166, 5, 5, 3, 172, 5, 5, 4, 172, 5, 5, 5, 172, 5, 5, 6, 172, 5, 5, 7, 172, 5, 5, 8, 172, 5, 5, 155, 166, 5, 5, 9, 172, 5, 5, 10, 172, 5, 5, 11, 172, 5, 5, 12, 172, 5, 5, 13, 172, 5, 5, 133, 178, 5, 5, 134, 178, 5, 5, 135, 178, 5, 5, 136, 178, 5, 5, 137, 178, 5, 5, 138, 178, 5, 5, 174, 185, 5, 5, 175, 185, 5, 5, 176, 185, 5, 5, 177, 185, 5, 5, 178, 185, 5, 5, 49, 188, 5, 5, 179, 185, 5, 5, 180, 185, 5, 5, 181, 185, 5, 5, 180, 193, 5, 5, 181, 193, 5, 5, 182, 193, 5, 5, 183, 193, 5, 5, 46, 202, 5, 5, 47, 202, 5, 5, 48, 202, 5, 5, 49, 202, 5, 5, 50, 202, 5, 5, 51, 202, 5, 5, 52, 202, 5, 5, 53, 202, 5, 5, 54, 202, 5, 5, 17, 210, 5, 5, 18, 210, 5, 5, 19, 210, 5, 5, 20, 210, 5, 5, 21, 210, 5, 5, 22, 210, 5, 5, 43, 217, 5, 5, 44, 217, 5, 5, 45, 217, 5, 5, 46, 217, 5, 5, 47, 217, 5, 5, 76, 224, 5, 5, 75, 230, 5, 5, 76, 230, 5, 5, 77, 230, 5, 5, 222, 238, 5, 5, 223, 238, 5, 5, 33, 157, 5, 5, 93, 157, 5, 5, 240, 157, 5, 5, 241, 157, 5, 5, 242, 157, 5, 5, 243, 157, 5, 5, 244, 157, 5, 5, 245, 157, 5, 5, 246, 157, 5, 5, 247, 157, 5, 5, 216, 158, 5, 5, 217, 158, 5, 5, 218, 158, 5, 5, 219, 158, 5, 5, 220, 158, 5, 5, 105, 160, 5, 5, 160, 162, 5, 5, 156, 166, 5, 5, 157, 166, 5, 5, 158, 166, 5, 5, 14, 172, 5, 5, 140, 178, 5, 5, 183, 185, 5, 5, 184, 185, 5, 5, 184, 193, 5, 5, 185, 193, 5, 5, 185, 185, 5, 5, 48, 217, 5, 5, 34, 157, 5, 5, 248, 157, 5, 5, 221, 158, 5, 5, 186, 185, 5, 5, 187, 185, 5, 5, 35, 157, 5, 5, 222, 158, 5, 5, 223, 158, 5, 5, 224, 158, 5, 5, 225, 158, 5, 5, 106, 160, 5, 5, 107, 160, 5, 5, 108, 160, 5, 5, 109, 160, 5, 5, 162, 162, 5, 5, 163, 162, 5, 5, 164, 162, 5, 5, 160, 166, 5, 5, 15, 172, 5, 5, 16, 172, 5, 5, 17, 172, 5, 5, 141, 178, 5, 5, 142, 178, 5, 5, 189, 185, 5, 5, 190, 185, 5, 5, 191, 185, 5, 5, 55, 202, 5, 5, 24, 210, 5, 5, 25, 210, 5, 5, 26, 210, 5, 5, 49, 217, 5, 5, 77, 224, 5, 5, 80, 230, 5, 5, 224, 238, 5, 5, 27, 242, 5, 5, 36, 157, 5, 5, 250, 157, 5, 5, 251, 157, 5, 5, 166, 162, 5, 5, 161, 166, 5, 5, 18, 172, 5, 5, 193, 185, 5, 5, 194, 185, 5, 5, 195, 185, 5, 5, 37, 157, 5, 5, 94, 157, 5, 5, 95, 157, 5, 5, 61, 157, 5, 5, 252, 157, 5, 5, 253, 157, 5, 5, 254, 157, 5, 5, 255, 157, 5, 5, 226, 158, 5, 5, 227, 158, 5, 5, 110, 160, 5, 5, 228, 158, 5, 5, 111, 160, 5, 5, 112, 160, 5, 5, 113, 160, 5, 5, 114, 160, 5, 5, 162, 166, 5, 5, 163, 166, 5, 5, 164, 166, 5, 5, 165, 166, 5, 5, 166, 166, 5, 5, 167, 166, 5, 5, 19, 172, 5, 5, 20, 172, 5, 5, 196, 185, 5, 5, 189, 193, 5, 5, 157, 244, 5, 5, 39, 157, 5, 5, 3, 158, 5, 5, 4, 158, 5, 5, 229, 158, 5, 5, 230, 158, 5, 5, 231, 158, 5, 5, 232, 158, 5, 5, 168, 162, 5, 5, 169, 162, 5, 5, 168, 166, 5, 5, 169, 166, 5, 5, 170, 166, 5, 5, 197, 185, 5, 5, 40, 157, 5, 5, 96, 157, 5, 5, 97, 157, 5, 5, 5, 158, 5, 5, 233, 158, 5, 5, 234, 158, 5, 5, 235, 158, 5, 5, 116, 160, 5, 5, 117, 160, 5, 5, 170, 162, 5, 5, 171, 162, 5, 5, 172, 162, 5, 5, 173, 162, 5, 5, 172, 166, 5, 5, 173, 166, 5, 5, 174, 166, 5, 5, 175, 166, 5, 5, 176, 166, 5, 5, 21, 172, 5, 5, 22, 172, 5, 5, 23, 172, 5, 5, 198, 185, 5, 5, 144, 178, 5, 5, 59, 202, 5, 5, 60, 202, 5, 5, 41, 157, 5, 5, 6, 158, 5, 5, 7, 158, 5, 5, 8, 158, 5, 5, 9, 158, 5, 5, 237, 158, 5, 5, 238, 158, 5, 5, 239, 158, 5, 5, 118, 160, 5, 5, 119, 160, 5, 5, 120, 160, 5, 5, 121, 160, 5, 5, 174, 162, 5, 5, 175, 162, 5, 5, 176, 162, 5, 5, 177, 162, 5, 5, 177, 166, 5, 5, 178, 166, 5, 5, 179, 166, 5, 5, 180, 166, 5, 5, 24, 172, 5, 5, 25, 172, 5, 5, 26, 172, 5, 5, 27, 172, 5, 5, 28, 172, 5, 5, 29, 172, 5, 5, 145, 178, 5, 5, 146, 178, 5, 5, 147, 178, 5, 5, 148, 178, 5, 5, 199, 185, 5, 5, 200, 185, 5, 5, 201, 185, 5, 5, 202, 185, 5, 5, 190, 193, 5, 5, 191, 193, 5, 5, 192, 193, 5, 5, 193, 193, 5, 5, 194, 193, 5, 5, 203, 185, 5, 5, 61, 202, 5, 5, 62, 202, 5, 5, 28, 210, 5, 5, 29, 210, 5, 5, 30, 210, 5, 5, 63, 202, 5, 5, 31, 210, 5, 5, 50, 217, 5, 5, 51, 217, 5, 5, 81, 230, 5, 5, 225, 238, 5, 5, 137, 250, 5, 5, 42, 157, 5, 5, 10, 158, 5, 5, 11, 158, 5, 5, 12, 158, 5, 5, 240, 158, 5, 5, 241, 158, 5, 5, 242, 158, 5, 5, 123, 160, 5, 5, 124, 160, 5, 5, 178, 162, 5, 5, 182, 166, 5, 5, 183, 166, 5, 5, 184, 166, 5, 5, 204, 185, 5, 5, 205, 185, 5, 5, 196, 193, 5, 5, 33, 210, 5, 5, 53, 217, 5, 5, 44, 157, 5, 5, 99, 157, 5, 5, 13, 158, 5, 5, 14, 158, 5, 5, 15, 158, 5, 5, 16, 158, 5, 5, 17, 158, 5, 5, 243, 158, 5, 5, 244, 158, 5, 5, 245, 158, 5, 5, 125, 160, 5, 5, 179, 162, 5, 5, 185, 166, 5, 5, 186, 166, 5, 5, 187, 166, 5, 5, 188, 166, 5, 5, 189, 166, 5, 5, 30, 172, 5, 5, 31, 172, 5, 5, 32, 172, 5, 5, 33, 172, 5, 5, 34, 172, 5, 5, 151, 178, 5, 5, 152, 178, 5, 5, 64, 202, 5, 5, 78, 224, 5, 5, 43, 235, 5, 5, 100, 157, 5, 5, 246, 158, 5, 5, 247, 158, 5, 5, 248, 158, 5, 5, 249, 158, 5, 5, 250, 158, 5, 5, 251, 158, 5, 5, 252, 158, 5, 5, 253, 158, 5, 5, 254, 158, 5, 5, 255, 158, 5, 5, 2, 159, 5, 5, 3, 159, 5, 5, 4, 159, 5, 5, 5, 159, 5, 5, 6, 159, 5, 5, 7, 159, 5, 5, 8, 159, 5, 5, 9, 159, 5, 5, 10, 159, 5, 5, 11, 159, 5, 5, 12, 159, 5, 5, 13, 159, 5, 5, 14, 159, 5, 5, 15, 159, 5, 5, 16, 159, 5, 5, 17, 159, 5, 5, 18, 159, 5, 5, 126, 160, 5, 5, 127, 160, 5, 5, 128, 160, 5, 5, 129, 160, 5, 5, 130, 160, 5, 5, 131, 160, 5, 5, 132, 160, 5, 5, 133, 160, 5, 5, 134, 160, 5, 5, 135, 160, 5, 5, 136, 160, 5, 5, 137, 160, 5, 5, 138, 160, 5, 5, 139, 160, 5, 5, 140, 160, 5, 5, 141, 160, 5, 5, 142, 160, 5, 5, 143, 160, 5, 5, 144, 160, 5, 5, 145, 160, 5, 5, 146, 160, 5, 5, 147, 160, 5, 5, 148, 160, 5, 5, 149, 160, 5, 5, 150, 160, 5, 5, 180, 162, 5, 5, 181, 162, 5, 5, 182, 162, 5, 5, 183, 162, 5, 5, 184, 162, 5, 5, 185, 162, 5, 5, 186, 162, 5, 5, 187, 162, 5, 5, 188, 162, 5, 5, 189, 162, 5, 5, 190, 162, 5, 5, 191, 162, 5, 5, 192, 162, 5, 5, 193, 162, 5, 5, 194, 162, 5, 5, 195, 162, 5, 5, 196, 162, 5, 5, 197, 162, 5, 5, 198, 162, 5, 5, 199, 162, 5, 5, 200, 162, 5, 5, 201, 162, 5, 5, 202, 162, 5, 5, 203, 162, 5, 5, 204, 162, 5, 5, 205, 162, 5, 5, 206, 162, 5, 5, 207, 162, 5, 5, 208, 162, 5, 5, 209, 162, 5, 5, 210, 162, 5, 5, 211, 162, 5, 5, 212, 162, 5, 5, 213, 162, 5, 5, 214, 162, 5, 5, 215, 162, 5, 5, 216, 162, 5, 5, 217, 162, 5, 5, 218, 162, 5, 5, 219, 162, 5, 5, 220, 162, 5, 5, 221, 162, 5, 5, 222, 162, 5, 5, 223, 162, 5, 5, 224, 162, 5, 5, 225, 162, 5, 5, 226, 162, 5, 5, 227, 162, 5, 5, 228, 162, 5, 5, 229, 162, 5, 5, 230, 162, 5, 5, 231, 162, 5, 5, 232, 162, 5, 5, 233, 162, 5, 5, 234, 162, 5, 5, 235, 162, 5, 5, 236, 162, 5, 5, 237, 162, 5, 5, 238, 162, 5, 5, 239, 162, 5, 5, 240, 162, 5, 5, 241, 162, 5, 5, 242, 162, 5, 5, 243, 162, 5, 5, 244, 162, 5, 5, 245, 162, 5, 5, 246, 162, 5, 5, 247, 162, 5, 5, 248, 162, 5, 5, 190, 166, 5, 5, 191, 166, 5, 5, 192, 166, 5, 5, 193, 166, 5, 5, 194, 166, 5, 5, 195, 166, 5, 5, 196, 166, 5, 5, 197, 166, 5, 5, 198, 166, 5, 5, 199, 166, 5, 5, 200, 166, 5, 5, 201, 166, 5, 5, 202, 166, 5, 5, 203, 166, 5, 5, 204, 166, 5, 5, 205, 166, 5, 5, 206, 166, 5, 5, 207, 166, 5, 5, 208, 166, 5, 5, 35, 172, 5, 5, 209, 166, 5, 5, 36, 172, 5, 5, 210, 166, 5, 5, 211, 166, 5, 5, 212, 166, 5, 5, 213, 166, 5, 5, 214, 166, 5, 5, 215, 166, 5, 5, 216, 166, 5, 5, 217, 166, 5, 5, 218, 166, 5, 5, 219, 166, 5, 5, 220, 166, 5, 5, 221, 166, 5, 5, 222, 166, 5, 5, 223, 166, 5, 5, 224, 166, 5, 5, 225, 166, 5, 5, 226, 166, 5, 5, 227, 166, 5, 5, 228, 166, 5, 5, 229, 166, 5, 5, 230, 166, 5, 5, 231, 166, 5, 5, 232, 166, 5, 5, 233, 166, 5, 5, 234, 166, 5, 5, 235, 166, 5, 5, 236, 166, 5, 5, 237, 166, 5, 5, 238, 166, 5, 5, 239, 166, 5, 5, 240, 166, 5, 5, 241, 166, 5, 5, 242, 166, 5, 5, 243, 166, 5, 5, 244, 166, 5, 5, 245, 166, 5, 5, 246, 166, 5, 5, 247, 166, 5, 5, 248, 166, 5, 5, 249, 166, 5, 5, 250, 166, 5, 5, 251, 166, 5, 5, 252, 166, 5, 5, 37, 172, 5, 5, 38, 172, 5, 5, 39, 172, 5, 5, 40, 172, 5, 5, 41, 172, 5, 5, 42, 172, 5, 5, 43, 172, 5, 5, 44, 172, 5, 5, 45, 172, 5, 5, 46, 172, 5, 5, 47, 172, 5, 5, 48, 172, 5, 5, 49, 172, 5, 5, 50, 172, 5, 5, 51, 172, 5, 5, 52, 172, 5, 5, 53, 172, 5, 5, 54, 172, 5, 5, 55, 172, 5, 5, 56, 172, 5, 5, 57, 172, 5, 5, 58, 172, 5, 5, 59, 172, 5, 5, 60, 172, 5, 5, 61, 172, 5, 5, 62, 172, 5, 5, 63, 172, 5, 5, 64, 172, 5, 5, 65, 172, 5, 5, 66, 172, 5, 5, 67, 172, 5, 5, 68, 172, 5, 5, 69, 172, 5, 5, 70, 172, 5, 5, 71, 172, 5, 5, 72, 172, 5, 5, 73, 172, 5, 5, 74, 172, 5, 5, 75, 172, 5, 5, 76, 172, 5, 5, 77, 172, 5, 5, 78, 172, 5, 5, 79, 172, 5, 5, 80, 172, 5, 5, 81, 172, 5, 5, 82, 172, 5, 5, 83, 172, 5, 5, 84, 172, 5, 5, 85, 172, 5, 5, 86, 172, 5, 5, 87, 172, 5, 5, 88, 172, 5, 5, 89, 172, 5, 5, 90, 172, 5, 5, 91, 172, 5, 5, 92, 172, 5, 5, 93, 172, 5, 5, 94, 172, 5, 5, 95, 172, 5, 5, 96, 172, 5, 5, 97, 172, 5, 5, 98, 172, 5, 5, 99, 172, 5, 5, 100, 172, 5, 5, 101, 172, 5, 5, 102, 172, 5, 5, 153, 178, 5, 5, 154, 178, 5, 5, 155, 178, 5, 5, 156, 178, 5, 5, 157, 178, 5, 5, 158, 178, 5, 5, 159, 178, 5, 5, 160, 178, 5, 5, 161, 178, 5, 5, 162, 178, 5, 5, 163, 178, 5, 5, 164, 178, 5, 5, 165, 178, 5, 5, 166, 178, 5, 5, 167, 178, 5, 5, 168, 178, 5, 5, 169, 178, 5, 5, 170, 178, 5, 5, 171, 178, 5, 5, 172, 178, 5, 5, 173, 178, 5, 5, 174, 178, 5, 5, 175, 178, 5, 5, 176, 178, 5, 5, 177, 178, 5, 5, 178, 178, 5, 5, 179, 178, 5, 5, 180, 178, 5, 5, 181, 178, 5, 5, 182, 178, 5, 5, 183, 178, 5, 5, 184, 178, 5, 5, 185, 178, 5, 5, 186, 178, 5, 5, 187, 178, 5, 5, 188, 178, 5, 5, 189, 178, 5, 5, 190, 178, 5, 5, 191, 178, 5, 5, 192, 178, 5, 5, 193, 178, 5, 5, 194, 178, 5, 5, 195, 178, 5, 5, 196, 178, 5, 5, 206, 185, 5, 5, 197, 178, 5, 5, 198, 178, 5, 5, 199, 178, 5, 5, 200, 178, 5, 5, 201, 178, 5, 5, 202, 178, 5, 5, 203, 178, 5, 5, 204, 178, 5, 5, 205, 178, 5, 5, 206, 178, 5, 5, 207, 178, 5, 5, 208, 178, 5, 5, 209, 178, 5, 5, 210, 178, 5, 5, 211, 178, 5, 5, 212, 178, 5, 5, 213, 178, 5, 5, 214, 178, 5, 5, 215, 178, 5, 5, 216, 178, 5, 5, 217, 178, 5, 5, 218, 178, 5, 5, 219, 178, 5, 5, 220, 178, 5, 5, 221, 178, 5, 5, 222, 178, 5, 5, 223, 178, 5, 5, 207, 185, 5, 5, 208, 185, 5, 5, 209, 185, 5, 5, 210, 185, 5, 5, 211, 185, 5, 5, 212, 185, 5, 5, 213, 185, 5, 5, 214, 185, 5, 5, 215, 185, 5, 5, 216, 185, 5, 5, 217, 185, 5, 5, 218, 185, 5, 5, 219, 185, 5, 5, 220, 185, 5, 5, 221, 185, 5, 5, 222, 185, 5, 5, 223, 185, 5, 5, 224, 185, 5, 5, 225, 185, 5, 5, 226, 185, 5, 5, 227, 185, 5, 5, 228, 185, 5, 5, 229, 185, 5, 5, 230, 185, 5, 5, 231, 185, 5, 5, 232, 185, 5, 5, 233, 185, 5, 5, 234, 185, 5, 5, 235, 185, 5, 5, 236, 185, 5, 5, 237, 185, 5, 5, 238, 185, 5, 5, 239, 185, 5, 5, 240, 185, 5, 5, 241, 185, 5, 5, 242, 185, 5, 5, 243, 185, 5, 5, 244, 185, 5, 5, 245, 185, 5, 5, 246, 185, 5, 5, 247, 185, 5, 5, 248, 185, 5, 5, 249, 185, 5, 5, 250, 185, 5, 5, 251, 185, 5, 5, 252, 185, 5, 5, 253, 185, 5, 5, 254, 185, 5, 5, 255, 185, 5, 5, 197, 193, 5, 5, 2, 186, 5, 5, 3, 186, 5, 5, 4, 186, 5, 5, 5, 186, 5, 5, 6, 186, 5, 5, 10, 188, 5, 5, 7, 186, 5, 5, 8, 186, 5, 5, 9, 186, 5, 5, 10, 186, 5, 5, 11, 186, 5, 5, 12, 186, 5, 5, 13, 186, 5, 5, 14, 186, 5, 5, 15, 186, 5, 5, 16, 186, 5, 5, 17, 186, 5, 5, 55, 186, 5, 5, 18, 186, 5, 5, 19, 186, 5, 5, 20, 186, 5, 5, 21, 186, 5, 5, 22, 186, 5, 5, 23, 186, 5, 5, 24, 186, 5, 5, 25, 186, 5, 5, 26, 186, 5, 5, 27, 186, 5, 5, 28, 186, 5, 5, 29, 186, 5, 5, 30, 186, 5, 5, 31, 186, 5, 5, 198, 193, 5, 5, 199, 193, 5, 5, 200, 193, 5, 5, 201, 193, 5, 5, 202, 193, 5, 5, 203, 193, 5, 5, 204, 193, 5, 5, 205, 193, 5, 5, 206, 193, 5, 5, 207, 193, 5, 5, 208, 193, 5, 5, 209, 193, 5, 5, 210, 193, 5, 5, 211, 193, 5, 5, 212, 193, 5, 5, 213, 193, 5, 5, 214, 193, 5, 5, 215, 193, 5, 5, 216, 193, 5, 5, 65, 202, 5, 5, 217, 193, 5, 5, 218, 193, 5, 5, 219, 193, 5, 5, 220, 193, 5, 5, 221, 193, 5, 5, 222, 193, 5, 5, 223, 193, 5, 5, 224, 193, 5, 5, 225, 193, 5, 5, 226, 193, 5, 5, 227, 193, 5, 5, 228, 193, 5, 5, 229, 193, 5, 5, 230, 193, 5, 5, 231, 193, 5, 5, 232, 193, 5, 5, 233, 193, 5, 5, 234, 193, 5, 5, 235, 193, 5, 5, 236, 193, 5, 5, 237, 193, 5, 5, 238, 193, 5, 5, 239, 193, 5, 5, 240, 193, 5, 5, 241, 193, 5, 5, 242, 193, 5, 5, 243, 193, 5, 5, 244, 193, 5, 5, 245, 193, 5, 5, 246, 193, 5, 5, 247, 193, 5, 5, 248, 193, 5, 5, 249, 193, 5, 5, 250, 193, 5, 5, 251, 193, 5, 5, 252, 193, 5, 5, 253, 193, 5, 5, 254, 193, 5, 5, 255, 193, 5, 5, 2, 194, 5, 5, 56, 186, 5, 5, 3, 194, 5, 5, 4, 194, 5, 5, 5, 194, 5, 5, 6, 194, 5, 5, 7, 194, 5, 5, 8, 194, 5, 5, 9, 194, 5, 5, 10, 194, 5, 5, 66, 202, 5, 5, 67, 202, 5, 5, 68, 202, 5, 5, 69, 202, 5, 5, 70, 202, 5, 5, 71, 202, 5, 5, 72, 202, 5, 5, 73, 202, 5, 5, 74, 202, 5, 5, 75, 202, 5, 5, 76, 202, 5, 5, 77, 202, 5, 5, 78, 202, 5, 5, 79, 202, 5, 5, 80, 202, 5, 5, 81, 202, 5, 5, 82, 202, 5, 5, 83, 202, 5, 5, 84, 202, 5, 5, 85, 202, 5, 5, 86, 202, 5, 5, 87, 202, 5, 5, 88, 202, 5, 5, 89, 202, 5, 5, 90, 202, 5, 5, 91, 202, 5, 5, 92, 202, 5, 5, 93, 202, 5, 5, 94, 202, 5, 5, 95, 202, 5, 5, 96, 202, 5, 5, 30, 194, 5, 5, 97, 202, 5, 5, 98, 202, 5, 5, 99, 202, 5, 5, 100, 202, 5, 5, 101, 202, 5, 5, 102, 202, 5, 5, 103, 202, 5, 5, 104, 202, 5, 5, 105, 202, 5, 5, 106, 202, 5, 5, 107, 202, 5, 5, 108, 202, 5, 5, 109, 202, 5, 5, 110, 202, 5, 5, 111, 202, 5, 5, 112, 202, 5, 5, 113, 202, 5, 5, 114, 202, 5, 5, 115, 202, 5, 5, 116, 202, 5, 5, 117, 202, 5, 5, 118, 202, 5, 5, 119, 202, 5, 5, 37, 210, 5, 5, 38, 210, 5, 5, 39, 210, 5, 5, 40, 210, 5, 5, 41, 210, 5, 5, 42, 210, 5, 5, 43, 210, 5, 5, 44, 210, 5, 5, 45, 210, 5, 5, 46, 210, 5, 5, 47, 210, 5, 5, 48, 210, 5, 5, 49, 210, 5, 5, 50, 210, 5, 5, 51, 210, 5, 5, 52, 210, 5, 5, 53, 210, 5, 5, 54, 210, 5, 5, 55, 210, 5, 5, 56, 210, 5, 5, 57, 210, 5, 5, 58, 210, 5, 5, 59, 210, 5, 5, 60, 210, 5, 5, 61, 210, 5, 5, 62, 210, 5, 5, 63, 210, 5, 5, 64, 210, 5, 5, 65, 210, 5, 5, 66, 210, 5, 5, 67, 210, 5, 5, 68, 210, 5, 5, 69, 210, 5, 5, 70, 210, 5, 5, 71, 210, 5, 5, 72, 210, 5, 5, 73, 210, 5, 5, 74, 210, 5, 5, 75, 210, 5, 5, 76, 210, 5, 5, 77, 210, 5, 5, 101, 210, 5, 5, 56, 217, 5, 5, 78, 210, 5, 5, 79, 210, 5, 5, 80, 210, 5, 5, 81, 210, 5, 5, 82, 210, 5, 5, 83, 210, 5, 5, 84, 210, 5, 5, 102, 210, 5, 5, 57, 217, 5, 5, 58, 217, 5, 5, 59, 217, 5, 5, 60, 217, 5, 5, 61, 217, 5, 5, 62, 217, 5, 5, 63, 217, 5, 5, 64, 217, 5, 5, 65, 217, 5, 5, 66, 217, 5, 5, 67, 217, 5, 5, 68, 217, 5, 5, 69, 217, 5, 5, 70, 217, 5, 5, 71, 217, 5, 5, 72, 217, 5, 5, 73, 217, 5, 5, 74, 217, 5, 5, 75, 217, 5, 5, 76, 217, 5, 5, 77, 217, 5, 5, 78, 217, 5, 5, 79, 217, 5, 5, 80, 217, 5, 5, 81, 217, 5, 5, 82, 217, 5, 5, 83, 217, 5, 5, 84, 217, 5, 5, 11, 194, 5, 5, 85, 217, 5, 5, 86, 217, 5, 5, 87, 217, 5, 5, 88, 217, 5, 5, 89, 217, 5, 5, 90, 217, 5, 5, 91, 217, 5, 5, 92, 217, 5, 5, 93, 217, 5, 5, 94, 217, 5, 5, 95, 217, 5, 5, 85, 210, 5, 5, 96, 217, 5, 5, 55, 217, 5, 5, 97, 217, 5, 5, 175, 219, 5, 5, 98, 217, 5, 5, 99, 217, 5, 5, 100, 217, 5, 5, 101, 217, 5, 5, 102, 217, 5, 5, 103, 217, 5, 5, 104, 217, 5, 5, 105, 217, 5, 5, 79, 224, 5, 5, 80, 224, 5, 5, 81, 224, 5, 5, 82, 224, 5, 5, 83, 224, 5, 5, 84, 224, 5, 5, 85, 224, 5, 5, 86, 224, 5, 5, 87, 224, 5, 5, 88, 224, 5, 5, 89, 224, 5, 5, 90, 224, 5, 5, 91, 224, 5, 5, 92, 224, 5, 5, 93, 224, 5, 5, 94, 224, 5, 5, 95, 224, 5, 5, 96, 224, 5, 5, 97, 224, 5, 5, 98, 224, 5, 5, 99, 224, 5, 5, 100, 224, 5, 5, 106, 217, 5, 5, 101, 224, 5, 5, 102, 224, 5, 5, 103, 224, 5, 5, 104, 224, 5, 5, 105, 224, 5, 5, 106, 224, 5, 5, 107, 224, 5, 5, 108, 224, 5, 5, 84, 230, 5, 5, 85, 230, 5, 5, 86, 230, 5, 5, 87, 230, 5, 5, 88, 230, 5, 5, 89, 230, 5, 5, 90, 230, 5, 5, 91, 230, 5, 5, 92, 230, 5, 5, 93, 230, 5, 5, 94, 230, 5, 5, 95, 230, 5, 5, 96, 230, 5, 5, 97, 230, 5, 5, 98, 230, 5, 5, 99, 230, 5, 5, 100, 230, 5, 5, 101, 230, 5, 5, 102, 230, 5, 5, 103, 230, 5, 5, 104, 230, 5, 5, 105, 230, 5, 5, 106, 230, 5, 5, 44, 235, 5, 5, 45, 235, 5, 5, 46, 235, 5, 5, 47, 235, 5, 5, 48, 235, 5, 5, 49, 235, 5, 5, 50, 235, 5, 5, 51, 235, 5, 5, 52, 235, 5, 5, 53, 235, 5, 5, 54, 235, 5, 5, 55, 235, 5, 5, 56, 235, 5, 5, 57, 235, 5, 5, 58, 235, 5, 5, 59, 235, 5, 5, 60, 235, 5, 5, 227, 238, 5, 5, 228, 238, 5, 5, 229, 238, 5, 5, 230, 238, 5, 5, 231, 238, 5, 5, 232, 238, 5, 5, 233, 238, 5, 5, 234, 238, 5, 5, 235, 238, 5, 5, 73, 235, 5, 5, 236, 238, 5, 5, 237, 238, 5, 5, 28, 242, 5, 5, 29, 242, 5, 5, 30, 242, 5, 5, 31, 242, 5, 5, 32, 242, 5, 5, 33, 242, 5, 5, 34, 242, 5, 5, 35, 242, 5, 5, 36, 242, 5, 5, 158, 244, 5, 5, 159, 244, 5, 5, 38, 242, 5, 5, 160, 244, 5, 5, 161, 244, 5, 5, 162, 244, 5, 5, 163, 244, 5, 5, 164, 244, 5, 5, 165, 244, 5, 5, 166, 244, 5, 5, 167, 244, 5, 5, 96, 246, 5, 5, 97, 246, 5, 5, 98, 246, 5, 5, 99, 246, 5, 5, 100, 246, 5, 5, 101, 246, 5, 5, 102, 246, 5, 5, 180, 247, 5, 5, 168, 244, 5, 5, 103, 246, 5, 5, 181, 247, 5, 5, 182, 247, 5, 5, 184, 248, 5, 5, 185, 248, 5, 5, 186, 248, 5, 5, 100, 249, 5, 5, 101, 249, 5, 5, 85, 250, 5, 5, 101, 157, 5, 5, 22, 159, 5, 5, 23, 159, 5, 5, 24, 159, 5, 5, 25, 159, 5, 5, 26, 159, 5, 5, 154, 160, 5, 5, 155, 160, 5, 5, 156, 160, 5, 5, 157, 160, 5, 5, 158, 160, 5, 5, 159, 160, 5, 5, 160, 160, 5, 5, 3, 163, 5, 5, 4, 163, 5, 5, 5, 163, 5, 5, 6, 163, 5, 5, 7, 163, 5, 5, 8, 163, 5, 5, 9, 163, 5, 5, 10, 163, 5, 5, 11, 163, 5, 5, 12, 163, 5, 5, 13, 163, 5, 5, 14, 163, 5, 5, 15, 163, 5, 5, 16, 163, 5, 5, 17, 163, 5, 5, 18, 163, 5, 5, 19, 163, 5, 5, 20, 163, 5, 5, 10, 167, 5, 5, 11, 167, 5, 5, 12, 167, 5, 5, 13, 167, 5, 5, 14, 167, 5, 5, 15, 167, 5, 5, 16, 167, 5, 5, 17, 167, 5, 5, 18, 167, 5, 5, 112, 172, 5, 5, 113, 172, 5, 5, 237, 178, 5, 5, 238, 178, 5, 5, 239, 178, 5, 5, 240, 178, 5, 5, 241, 178, 5, 5, 242, 178, 5, 5, 57, 186, 5, 5, 58, 186, 5, 5, 59, 186, 5, 5, 60, 186, 5, 5, 61, 186, 5, 5, 31, 194, 5, 5, 32, 194, 5, 5, 33, 194, 5, 5, 62, 186, 5, 5, 34, 194, 5, 5, 144, 202, 5, 5, 145, 202, 5, 5, 146, 202, 5, 5, 147, 202, 5, 5, 148, 202, 5, 5, 103, 210, 5, 5, 104, 210, 5, 5, 105, 210, 5, 5, 106, 210, 5, 5, 119, 217, 5, 5, 123, 224, 5, 5, 124, 224, 5, 5, 105, 246, 5, 5, 221, 249, 5, 5, 102, 157, 5, 5, 18, 158, 5, 5, 19, 158, 5, 5, 28, 159, 5, 5, 29, 159, 5, 5, 30, 159, 5, 5, 31, 159, 5, 5, 32, 159, 5, 5, 33, 159, 5, 5, 161, 160, 5, 5, 162, 160, 5, 5, 163, 160, 5, 5, 164, 160, 5, 5, 165, 160, 5, 5, 166, 160, 5, 5, 167, 160, 5, 5, 168, 160, 5, 5, 169, 160, 5, 5, 170, 160, 5, 5, 171, 160, 5, 5, 172, 160, 5, 5, 173, 160, 5, 5, 174, 160, 5, 5, 175, 160, 5, 5, 176, 160, 5, 5, 177, 160, 5, 5, 178, 160, 5, 5, 179, 160, 5, 5, 21, 163, 5, 5, 22, 163, 5, 5, 23, 163, 5, 5, 24, 163, 5, 5, 25, 163, 5, 5, 26, 163, 5, 5, 27, 163, 5, 5, 28, 163, 5, 5, 29, 163, 5, 5, 30, 163, 5, 5, 31, 163, 5, 5, 32, 163, 5, 5, 33, 163, 5, 5, 34, 163, 5, 5, 35, 163, 5, 5, 36, 163, 5, 5, 37, 163, 5, 5, 38, 163, 5, 5, 39, 163, 5, 5, 40, 163, 5, 5, 41, 163, 5, 5, 42, 163, 5, 5, 43, 163, 5, 5, 44, 163, 5, 5, 45, 163, 5, 5, 46, 163, 5, 5, 47, 163, 5, 5, 48, 163, 5, 5, 49, 163, 5, 5, 50, 163, 5, 5, 51, 163, 5, 5, 52, 163, 5, 5, 53, 163, 5, 5, 54, 163, 5, 5, 55, 163, 5, 5, 56, 163, 5, 5, 57, 163, 5, 5, 58, 163, 5, 5, 20, 167, 5, 5, 21, 167, 5, 5, 22, 167, 5, 5, 23, 167, 5, 5, 24, 167, 5, 5, 25, 167, 5, 5, 26, 167, 5, 5, 27, 167, 5, 5, 28, 167, 5, 5, 29, 167, 5, 5, 30, 167, 5, 5, 31, 167, 5, 5, 32, 167, 5, 5, 33, 167, 5, 5, 34, 167, 5, 5, 35, 167, 5, 5, 36, 167, 5, 5, 37, 167, 5, 5, 38, 167, 5, 5, 39, 167, 5, 5, 40, 167, 5, 5, 41, 167, 5, 5, 42, 167, 5, 5, 43, 167, 5, 5, 44, 167, 5, 5, 45, 167, 5, 5, 46, 167, 5, 5, 47, 167, 5, 5, 48, 167, 5, 5, 49, 167, 5, 5, 50, 167, 5, 5, 51, 167, 5, 5, 52, 167, 5, 5, 53, 167, 5, 5, 54, 167, 5, 5, 55, 167, 5, 5, 56, 167, 5, 5, 57, 167, 5, 5, 58, 167, 5, 5, 59, 167, 5, 5, 60, 167, 5, 5, 61, 167, 5, 5, 116, 172, 5, 5, 117, 172, 5, 5, 118, 172, 5, 5, 119, 172, 5, 5, 120, 172, 5, 5, 121, 172, 5, 5, 122, 172, 5, 5, 123, 172, 5, 5, 124, 172, 5, 5, 125, 172, 5, 5, 126, 172, 5, 5, 127, 172, 5, 5, 128, 172, 5, 5, 129, 172, 5, 5, 130, 172, 5, 5, 131, 172, 5, 5, 132, 172, 5, 5, 133, 172, 5, 5, 134, 172, 5, 5, 135, 172, 5, 5, 136, 172, 5, 5, 137, 172, 5, 5, 138, 172, 5, 5, 139, 172, 5, 5, 140, 172, 5, 5, 141, 172, 5, 5, 142, 172, 5, 5, 143, 172, 5, 5, 144, 172, 5, 5, 145, 172, 5, 5, 146, 172, 5, 5, 147, 172, 5, 5, 148, 172, 5, 5, 149, 172, 5, 5, 150, 172, 5, 5, 151, 172, 5, 5, 152, 172, 5, 5, 153, 172, 5, 5, 154, 172, 5, 5, 155, 172, 5, 5, 156, 172, 5, 5, 157, 172, 5, 5, 158, 172, 5, 5, 244, 178, 5, 5, 245, 178, 5, 5, 246, 178, 5, 5, 247, 178, 5, 5, 248, 178, 5, 5, 249, 178, 5, 5, 250, 178, 5, 5, 251, 178, 5, 5, 252, 178, 5, 5, 253, 178, 5, 5, 254, 178, 5, 5, 255, 178, 5, 5, 2, 179, 5, 5, 3, 179, 5, 5, 4, 179, 5, 5, 5, 179, 5, 5, 6, 179, 5, 5, 7, 179, 5, 5, 8, 179, 5, 5, 9, 179, 5, 5, 10, 179, 5, 5, 11, 179, 5, 5, 12, 179, 5, 5, 13, 179, 5, 5, 159, 172, 5, 5, 14, 179, 5, 5, 15, 179, 5, 5, 16, 179, 5, 5, 17, 179, 5, 5, 18, 179, 5, 5, 19, 179, 5, 5, 20, 179, 5, 5, 21, 179, 5, 5, 22, 179, 5, 5, 23, 179, 5, 5, 24, 179, 5, 5, 25, 179, 5, 5, 26, 179, 5, 5, 63, 186, 5, 5, 64, 186, 5, 5, 65, 186, 5, 5, 66, 186, 5, 5, 67, 186, 5, 5, 68, 186, 5, 5, 69, 186, 5, 5, 70, 186, 5, 5, 71, 186, 5, 5, 72, 186, 5, 5, 73, 186, 5, 5, 74, 186, 5, 5, 75, 186, 5, 5, 76, 186, 5, 5, 125, 186, 5, 5, 77, 186, 5, 5, 78, 186, 5, 5, 79, 186, 5, 5, 80, 186, 5, 5, 81, 186, 5, 5, 82, 186, 5, 5, 83, 186, 5, 5, 84, 186, 5, 5, 85, 186, 5, 5, 86, 186, 5, 5, 87, 186, 5, 5, 88, 186, 5, 5, 89, 186, 5, 5, 90, 186, 5, 5, 91, 186, 5, 5, 92, 186, 5, 5, 93, 186, 5, 5, 94, 186, 5, 5, 95, 186, 5, 5, 96, 186, 5, 5, 97, 186, 5, 5, 98, 186, 5, 5, 99, 186, 5, 5, 100, 186, 5, 5, 101, 186, 5, 5, 102, 186, 5, 5, 103, 186, 5, 5, 104, 186, 5, 5, 105, 186, 5, 5, 106, 186, 5, 5, 107, 186, 5, 5, 108, 186, 5, 5, 109, 186, 5, 5, 110, 186, 5, 5, 111, 186, 5, 5, 112, 186, 5, 5, 113, 186, 5, 5, 114, 186, 5, 5, 115, 186, 5, 5, 116, 186, 5, 5, 117, 186, 5, 5, 118, 186, 5, 5, 119, 186, 5, 5, 35, 194, 5, 5, 36, 194, 5, 5, 37, 194, 5, 5, 38, 194, 5, 5, 39, 194, 5, 5, 40, 194, 5, 5, 41, 194, 5, 5, 42, 194, 5, 5, 43, 194, 5, 5, 44, 194, 5, 5, 45, 194, 5, 5, 46, 194, 5, 5, 47, 194, 5, 5, 48, 194, 5, 5, 49, 194, 5, 5, 50, 194, 5, 5, 51, 194, 5, 5, 52, 194, 5, 5, 53, 194, 5, 5, 54, 194, 5, 5, 55, 194, 5, 5, 56, 194, 5, 5, 57, 194, 5, 5, 58, 194, 5, 5, 59, 194, 5, 5, 60, 194, 5, 5, 61, 194, 5, 5, 62, 194, 5, 5, 33, 179, 5, 5, 63, 194, 5, 5, 64, 194, 5, 5, 126, 186, 5, 5, 65, 194, 5, 5, 66, 194, 5, 5, 67, 194, 5, 5, 68, 194, 5, 5, 69, 194, 5, 5, 70, 194, 5, 5, 71, 194, 5, 5, 149, 202, 5, 5, 72, 194, 5, 5, 73, 194, 5, 5, 74, 194, 5, 5, 75, 194, 5, 5, 76, 194, 5, 5, 150, 202, 5, 5, 77, 194, 5, 5, 78, 194, 5, 5, 79, 194, 5, 5, 80, 194, 5, 5, 81, 194, 5, 5, 151, 202, 5, 5, 152, 202, 5, 5, 153, 202, 5, 5, 154, 202, 5, 5, 155, 202, 5, 5, 156, 202, 5, 5, 157, 202, 5, 5, 158, 202, 5, 5, 159, 202, 5, 5, 160, 202, 5, 5, 161, 202, 5, 5, 162, 202, 5, 5, 163, 202, 5, 5, 164, 202, 5, 5, 165, 202, 5, 5, 166, 202, 5, 5, 167, 202, 5, 5, 168, 202, 5, 5, 169, 202, 5, 5, 170, 202, 5, 5, 171, 202, 5, 5, 172, 202, 5, 5, 173, 202, 5, 5, 174, 202, 5, 5, 175, 202, 5, 5, 176, 202, 5, 5, 177, 202, 5, 5, 178, 202, 5, 5, 179, 202, 5, 5, 180, 202, 5, 5, 181, 202, 5, 5, 182, 202, 5, 5, 183, 202, 5, 5, 184, 202, 5, 5, 185, 202, 5, 5, 186, 202, 5, 5, 187, 202, 5, 5, 188, 202, 5, 5, 189, 202, 5, 5, 190, 202, 5, 5, 191, 202, 5, 5, 107, 210, 5, 5, 108, 210, 5, 5, 109, 210, 5, 5, 110, 210, 5, 5, 111, 210, 5, 5, 112, 210, 5, 5, 113, 210, 5, 5, 114, 210, 5, 5, 115, 210, 5, 5, 116, 210, 5, 5, 117, 210, 5, 5, 118, 210, 5, 5, 119, 210, 5, 5, 120, 210, 5, 5, 120, 217, 5, 5, 121, 210, 5, 5, 122, 210, 5, 5, 123, 210, 5, 5, 124, 210, 5, 5, 125, 210, 5, 5, 126, 210, 5, 5, 127, 210, 5, 5, 128, 210, 5, 5, 129, 210, 5, 5, 130, 210, 5, 5, 131, 210, 5, 5, 132, 210, 5, 5, 133, 210, 5, 5, 134, 210, 5, 5, 135, 210, 5, 5, 136, 210, 5, 5, 137, 210, 5, 5, 138, 210, 5, 5, 139, 210, 5, 5, 140, 210, 5, 5, 141, 210, 5, 5, 142, 210, 5, 5, 143, 210, 5, 5, 144, 210, 5, 5, 145, 210, 5, 5, 146, 210, 5, 5, 147, 210, 5, 5, 121, 217, 5, 5, 122, 217, 5, 5, 123, 217, 5, 5, 124, 217, 5, 5, 125, 217, 5, 5, 126, 217, 5, 5, 127, 217, 5, 5, 128, 217, 5, 5, 129, 217, 5, 5, 130, 217, 5, 5, 131, 217, 5, 5, 132, 217, 5, 5, 35, 224, 5, 5, 133, 217, 5, 5, 134, 217, 5, 5, 135, 217, 5, 5, 136, 217, 5, 5, 155, 210, 5, 5, 137, 217, 5, 5, 138, 217, 5, 5, 139, 217, 5, 5, 140, 217, 5, 5, 141, 217, 5, 5, 142, 217, 5, 5, 143, 217, 5, 5, 144, 217, 5, 5, 145, 217, 5, 5, 146, 217, 5, 5, 147, 217, 5, 5, 148, 217, 5, 5, 125, 224, 5, 5, 126, 224, 5, 5, 127, 224, 5, 5, 128, 224, 5, 5, 129, 224, 5, 5, 130, 224, 5, 5, 131, 224, 5, 5, 132, 224, 5, 5, 133, 224, 5, 5, 134, 224, 5, 5, 135, 224, 5, 5, 136, 224, 5, 5, 137, 224, 5, 5, 138, 224, 5, 5, 139, 224, 5, 5, 140, 224, 5, 5, 141, 224, 5, 5, 142, 224, 5, 5, 143, 224, 5, 5, 116, 230, 5, 5, 117, 230, 5, 5, 118, 230, 5, 5, 119, 230, 5, 5, 120, 230, 5, 5, 121, 230, 5, 5, 122, 230, 5, 5, 123, 230, 5, 5, 124, 230, 5, 5, 125, 230, 5, 5, 126, 230, 5, 5, 74, 235, 5, 5, 75, 235, 5, 5, 242, 238, 5, 5, 243, 238, 5, 5, 244, 238, 5, 5, 245, 238, 5, 5, 246, 238, 5, 5, 247, 238, 5, 5, 248, 238, 5, 5, 226, 238, 5, 5, 249, 238, 5, 5, 40, 242, 5, 5, 41, 242, 5, 5, 42, 242, 5, 5, 174, 244, 5, 5, 184, 247, 5, 5, 185, 247, 5, 5, 188, 248, 5, 5, 103, 249, 5, 5, 103, 157, 5, 5, 21, 158, 5, 5, 35, 159, 5, 5, 181, 160, 5, 5, 65, 163, 5, 5, 66, 163, 5, 5, 67, 163, 5, 5, 68, 163, 5, 5, 69, 163, 5, 5, 167, 172, 5, 5, 168, 172, 5, 5, 34, 179, 5, 5, 127, 186, 5, 5, 128, 186, 5, 5, 88, 194, 5, 5, 89, 194, 5, 5, 90, 194, 5, 5, 195, 202, 5, 5, 156, 210, 5, 5, 157, 210, 5, 5, 154, 217, 5, 5, 155, 217, 5, 5, 145, 224, 5, 5, 104, 157, 5, 5, 22, 158, 5, 5, 36, 159, 5, 5, 182, 160, 5, 5, 70, 163, 5, 5, 70, 167, 5, 5, 169, 172, 5, 5, 170, 172, 5, 5, 105, 157, 5, 5, 71, 163, 5, 5, 71, 167, 5, 5, 171, 172, 5, 5, 35, 179, 5, 5, 36, 179, 5, 5, 158, 210, 5, 5, 77, 235, 5, 5, 250, 238, 5, 5, 78, 235, 5, 5, 176, 244, 5, 5, 106, 157, 5, 5, 37, 159, 5, 5, 38, 159, 5, 5, 39, 159, 5, 5, 183, 160, 5, 5, 184, 160, 5, 5, 185, 160, 5, 5, 72, 167, 5, 5, 73, 167, 5, 5, 37, 179, 5, 5, 129, 186, 5, 5, 130, 186, 5, 5, 91, 194, 5, 5, 159, 210, 5, 5, 160, 210, 5, 5, 161, 210, 5, 5, 162, 210, 5, 5, 157, 217, 5, 5, 107, 157, 5, 5, 108, 157, 5, 5, 23, 158, 5, 5, 24, 158, 5, 5, 25, 158, 5, 5, 26, 158, 5, 5, 27, 158, 5, 5, 40, 159, 5, 5, 41, 159, 5, 5, 42, 159, 5, 5, 43, 159, 5, 5, 44, 159, 5, 5, 45, 159, 5, 5, 46, 159, 5, 5, 186, 160, 5, 5, 187, 160, 5, 5, 188, 160, 5, 5, 189, 160, 5, 5, 190, 160, 5, 5, 191, 160, 5, 5, 192, 160, 5, 5, 193, 160, 5, 5, 72, 163, 5, 5, 73, 163, 5, 5, 74, 163, 5, 5, 75, 163, 5, 5, 76, 163, 5, 5, 77, 163, 5, 5, 74, 167, 5, 5, 75, 167, 5, 5, 76, 167, 5, 5, 77, 167, 5, 5, 78, 167, 5, 5, 79, 167, 5, 5, 80, 167, 5, 5, 39, 179, 5, 5, 81, 167, 5, 5, 82, 167, 5, 5, 83, 167, 5, 5, 172, 172, 5, 5, 173, 172, 5, 5, 174, 172, 5, 5, 175, 172, 5, 5, 176, 172, 5, 5, 177, 172, 5, 5, 85, 167, 5, 5, 178, 172, 5, 5, 179, 172, 5, 5, 40, 179, 5, 5, 41, 179, 5, 5, 42, 179, 5, 5, 43, 179, 5, 5, 131, 186, 5, 5, 132, 186, 5, 5, 133, 186, 5, 5, 134, 186, 5, 5, 136, 186, 5, 5, 92, 194, 5, 5, 93, 194, 5, 5, 137, 186, 5, 5, 94, 194, 5, 5, 95, 194, 5, 5, 96, 194, 5, 5, 196, 202, 5, 5, 197, 202, 5, 5, 198, 202, 5, 5, 163, 210, 5, 5, 164, 210, 5, 5, 165, 210, 5, 5, 166, 210, 5, 5, 158, 217, 5, 5, 146, 224, 5, 5, 147, 224, 5, 5, 79, 235, 5, 5, 106, 246, 5, 5, 187, 247, 5, 5, 109, 157, 5, 5, 47, 159, 5, 5, 48, 159, 5, 5, 49, 159, 5, 5, 194, 160, 5, 5, 195, 160, 5, 5, 196, 160, 5, 5, 197, 160, 5, 5, 198, 160, 5, 5, 199, 160, 5, 5, 200, 160, 5, 5, 201, 160, 5, 5, 202, 160, 5, 5, 203, 160, 5, 5, 204, 160, 5, 5, 205, 160, 5, 5, 206, 160, 5, 5, 207, 160, 5, 5, 208, 160, 5, 5, 209, 160, 5, 5, 210, 160, 5, 5, 211, 160, 5, 5, 78, 163, 5, 5, 79, 163, 5, 5, 80, 163, 5, 5, 81, 163, 5, 5, 184, 172, 5, 5, 82, 163, 5, 5, 83, 163, 5, 5, 84, 163, 5, 5, 85, 163, 5, 5, 86, 163, 5, 5, 87, 163, 5, 5, 88, 163, 5, 5, 89, 163, 5, 5, 90, 163, 5, 5, 91, 163, 5, 5, 92, 163, 5, 5, 93, 163, 5, 5, 94, 163, 5, 5, 95, 163, 5, 5, 96, 163, 5, 5, 97, 163, 5, 5, 98, 163, 5, 5, 99, 163, 5, 5, 100, 163, 5, 5, 101, 163, 5, 5, 102, 163, 5, 5, 103, 163, 5, 5, 104, 163, 5, 5, 105, 163, 5, 5, 106, 163, 5, 5, 107, 163, 5, 5, 108, 163, 5, 5, 109, 163, 5, 5, 110, 163, 5, 5, 111, 163, 5, 5, 86, 167, 5, 5, 87, 167, 5, 5, 88, 167, 5, 5, 89, 167, 5, 5, 90, 167, 5, 5, 91, 167, 5, 5, 92, 167, 5, 5, 93, 167, 5, 5, 94, 167, 5, 5, 95, 167, 5, 5, 96, 167, 5, 5, 97, 167, 5, 5, 98, 167, 5, 5, 99, 167, 5, 5, 100, 167, 5, 5, 101, 167, 5, 5, 102, 167, 5, 5, 103, 167, 5, 5, 104, 167, 5, 5, 105, 167, 5, 5, 106, 167, 5, 5, 107, 167, 5, 5, 108, 167, 5, 5, 109, 167, 5, 5, 110, 167, 5, 5, 111, 167, 5, 5, 112, 167, 5, 5, 113, 167, 5, 5, 114, 167, 5, 5, 115, 167, 5, 5, 116, 167, 5, 5, 117, 167, 5, 5, 118, 167, 5, 5, 119, 167, 5, 5, 120, 167, 5, 5, 121, 167, 5, 5, 122, 167, 5, 5, 123, 167, 5, 5, 124, 167, 5, 5, 125, 167, 5, 5, 126, 167, 5, 5, 183, 172, 5, 5, 127, 167, 5, 5, 128, 167, 5, 5, 185, 172, 5, 5, 186, 172, 5, 5, 187, 172, 5, 5, 188, 172, 5, 5, 189, 172, 5, 5, 190, 172, 5, 5, 191, 172, 5, 5, 192, 172, 5, 5, 193, 172, 5, 5, 194, 172, 5, 5, 195, 172, 5, 5, 196, 172, 5, 5, 197, 172, 5, 5, 198, 172, 5, 5, 199, 172, 5, 5, 200, 172, 5, 5, 201, 172, 5, 5, 202, 172, 5, 5, 203, 172, 5, 5, 204, 172, 5, 5, 44, 179, 5, 5, 205, 172, 5, 5, 206, 172, 5, 5, 207, 172, 5, 5, 208, 172, 5, 5, 209, 172, 5, 5, 210, 172, 5, 5, 211, 172, 5, 5, 212, 172, 5, 5, 213, 172, 5, 5, 214, 172, 5, 5, 215, 172, 5, 5, 216, 172, 5, 5, 217, 172, 5, 5, 218, 172, 5, 5, 219, 172, 5, 5, 220, 172, 5, 5, 221, 172, 5, 5, 222, 172, 5, 5, 223, 172, 5, 5, 224, 172, 5, 5, 225, 172, 5, 5, 226, 172, 5, 5, 227, 172, 5, 5, 64, 176, 5, 5, 228, 172, 5, 5, 229, 172, 5, 5, 230, 172, 5, 5, 231, 172, 5, 5, 45, 179, 5, 5, 46, 179, 5, 5, 47, 179, 5, 5, 48, 179, 5, 5, 240, 172, 5, 5, 49, 179, 5, 5, 50, 179, 5, 5, 51, 179, 5, 5, 52, 179, 5, 5, 53, 179, 5, 5, 54, 179, 5, 5, 55, 179, 5, 5, 56, 179, 5, 5, 57, 179, 5, 5, 58, 179, 5, 5, 59, 179, 5, 5, 60, 179, 5, 5, 61, 179, 5, 5, 62, 179, 5, 5, 63, 179, 5, 5, 64, 179, 5, 5, 65, 179, 5, 5, 66, 179, 5, 5, 67, 179, 5, 5, 68, 179, 5, 5, 69, 179, 5, 5, 70, 179, 5, 5, 71, 179, 5, 5, 72, 179, 5, 5, 73, 179, 5, 5, 74, 179, 5, 5, 75, 179, 5, 5, 76, 179, 5, 5, 77, 179, 5, 5, 138, 186, 5, 5, 140, 186, 5, 5, 78, 179, 5, 5, 79, 179, 5, 5, 80, 179, 5, 5, 81, 179, 5, 5, 82, 179, 5, 5, 83, 179, 5, 5, 84, 179, 5, 5, 85, 179, 5, 5, 141, 186, 5, 5, 142, 186, 5, 5, 143, 186, 5, 5, 144, 186, 5, 5, 145, 186, 5, 5, 146, 186, 5, 5, 147, 186, 5, 5, 148, 186, 5, 5, 139, 186, 5, 5, 149, 186, 5, 5, 150, 186, 5, 5, 151, 186, 5, 5, 152, 186, 5, 5, 153, 186, 5, 5, 154, 186, 5, 5, 155, 186, 5, 5, 156, 186, 5, 5, 157, 186, 5, 5, 158, 186, 5, 5, 159, 186, 5, 5, 160, 186, 5, 5, 161, 186, 5, 5, 162, 186, 5, 5, 163, 186, 5, 5, 164, 186, 5, 5, 165, 186, 5, 5, 166, 186, 5, 5, 167, 186, 5, 5, 168, 186, 5, 5, 169, 186, 5, 5, 170, 186, 5, 5, 171, 186, 5, 5, 172, 186, 5, 5, 173, 186, 5, 5, 174, 186, 5, 5, 175, 186, 5, 5, 176, 186, 5, 5, 177, 186, 5, 5, 178, 186, 5, 5, 179, 186, 5, 5, 180, 186, 5, 5, 181, 186, 5, 5, 182, 186, 5, 5, 183, 186, 5, 5, 184, 186, 5, 5, 185, 186, 5, 5, 186, 186, 5, 5, 187, 186, 5, 5, 188, 186, 5, 5, 189, 186, 5, 5, 190, 186, 5, 5, 191, 186, 5, 5, 192, 186, 5, 5, 193, 186, 5, 5, 194, 186, 5, 5, 195, 186, 5, 5, 196, 186, 5, 5, 197, 186, 5, 5, 198, 186, 5, 5, 199, 186, 5, 5, 200, 186, 5, 5, 201, 186, 5, 5, 202, 186, 5, 5, 203, 186, 5, 5, 204, 186, 5, 5, 205, 186, 5, 5, 97, 194, 5, 5, 98, 194, 5, 5, 99, 194, 5, 5, 100, 194, 5, 5, 101, 194, 5, 5, 102, 194, 5, 5, 103, 194, 5, 5, 104, 194, 5, 5, 105, 194, 5, 5, 106, 194, 5, 5, 107, 194, 5, 5, 108, 194, 5, 5, 109, 194, 5, 5, 110, 194, 5, 5, 111, 194, 5, 5, 112, 194, 5, 5, 113, 194, 5, 5, 114, 194, 5, 5, 115, 194, 5, 5, 116, 194, 5, 5, 117, 194, 5, 5, 118, 194, 5, 5, 119, 194, 5, 5, 216, 186, 5, 5, 120, 194, 5, 5, 200, 202, 5, 5, 121, 194, 5, 5, 122, 194, 5, 5, 123, 194, 5, 5, 124, 194, 5, 5, 125, 194, 5, 5, 126, 194, 5, 5, 127, 194, 5, 5, 128, 194, 5, 5, 129, 194, 5, 5, 130, 194, 5, 5, 131, 194, 5, 5, 132, 194, 5, 5, 133, 194, 5, 5, 134, 194, 5, 5, 135, 194, 5, 5, 136, 194, 5, 5, 137, 194, 5, 5, 138, 194, 5, 5, 139, 194, 5, 5, 140, 194, 5, 5, 141, 194, 5, 5, 142, 194, 5, 5, 143, 194, 5, 5, 144, 194, 5, 5, 145, 194, 5, 5, 146, 194, 5, 5, 147, 194, 5, 5, 148, 194, 5, 5, 149, 194, 5, 5, 150, 194, 5, 5, 151, 194, 5, 5, 201, 202, 5, 5, 202, 202, 5, 5, 203, 202, 5, 5, 204, 202, 5, 5, 205, 202, 5, 5, 206, 202, 5, 5, 207, 202, 5, 5, 208, 202, 5, 5, 209, 202, 5, 5, 210, 202, 5, 5, 211, 202, 5, 5, 212, 202, 5, 5, 213, 202, 5, 5, 214, 202, 5, 5, 215, 202, 5, 5, 216, 202, 5, 5, 217, 202, 5, 5, 218, 202, 5, 5, 219, 202, 5, 5, 220, 202, 5, 5, 221, 202, 5, 5, 222, 202, 5, 5, 223, 202, 5, 5, 224, 202, 5, 5, 225, 202, 5, 5, 226, 202, 5, 5, 227, 202, 5, 5, 228, 202, 5, 5, 229, 202, 5, 5, 230, 202, 5, 5, 231, 202, 5, 5, 152, 194, 5, 5, 232, 202, 5, 5, 233, 202, 5, 5, 234, 202, 5, 5, 235, 202, 5, 5, 236, 202, 5, 5, 167, 210, 5, 5, 168, 210, 5, 5, 169, 210, 5, 5, 170, 210, 5, 5, 171, 210, 5, 5, 172, 210, 5, 5, 173, 210, 5, 5, 174, 210, 5, 5, 175, 210, 5, 5, 176, 210, 5, 5, 177, 210, 5, 5, 178, 210, 5, 5, 179, 210, 5, 5, 180, 210, 5, 5, 181, 210, 5, 5, 182, 210, 5, 5, 183, 210, 5, 5, 184, 210, 5, 5, 185, 210, 5, 5, 186, 210, 5, 5, 187, 210, 5, 5, 188, 210, 5, 5, 189, 210, 5, 5, 190, 210, 5, 5, 191, 210, 5, 5, 192, 210, 5, 5, 193, 210, 5, 5, 194, 210, 5, 5, 195, 210, 5, 5, 196, 210, 5, 5, 211, 210, 5, 5, 159, 217, 5, 5, 160, 217, 5, 5, 161, 217, 5, 5, 162, 217, 5, 5, 163, 217, 5, 5, 164, 217, 5, 5, 165, 217, 5, 5, 166, 217, 5, 5, 167, 217, 5, 5, 168, 217, 5, 5, 169, 217, 5, 5, 170, 217, 5, 5, 171, 217, 5, 5, 172, 217, 5, 5, 173, 217, 5, 5, 174, 217, 5, 5, 175, 217, 5, 5, 176, 217, 5, 5, 177, 217, 5, 5, 178, 217, 5, 5, 179, 217, 5, 5, 180, 217, 5, 5, 181, 217, 5, 5, 182, 217, 5, 5, 183, 217, 5, 5, 184, 217, 5, 5, 185, 217, 5, 5, 186, 217, 5, 5, 148, 224, 5, 5, 149, 224, 5, 5, 150, 224, 5, 5, 151, 224, 5, 5, 152, 224, 5, 5, 153, 224, 5, 5, 154, 224, 5, 5, 155, 224, 5, 5, 156, 224, 5, 5, 157, 224, 5, 5, 158, 224, 5, 5, 159, 224, 5, 5, 160, 224, 5, 5, 161, 224, 5, 5, 162, 224, 5, 5, 163, 224, 5, 5, 164, 224, 5, 5, 165, 224, 5, 5, 166, 224, 5, 5, 131, 230, 5, 5, 132, 230, 5, 5, 133, 230, 5, 5, 134, 230, 5, 5, 135, 230, 5, 5, 178, 224, 5, 5, 136, 230, 5, 5, 137, 230, 5, 5, 138, 230, 5, 5, 139, 230, 5, 5, 140, 230, 5, 5, 141, 230, 5, 5, 142, 230, 5, 5, 143, 230, 5, 5, 144, 230, 5, 5, 145, 230, 5, 5, 146, 230, 5, 5, 167, 224, 5, 5, 147, 230, 5, 5, 148, 230, 5, 5, 149, 230, 5, 5, 80, 235, 5, 5, 252, 238, 5, 5, 81, 235, 5, 5, 82, 235, 5, 5, 83, 235, 5, 5, 251, 238, 5, 5, 253, 238, 5, 5, 254, 238, 5, 5, 44, 242, 5, 5, 45, 242, 5, 5, 46, 242, 5, 5, 47, 242, 5, 5, 48, 242, 5, 5, 49, 242, 5, 5, 50, 242, 5, 5, 177, 244, 5, 5, 178, 244, 5, 5, 179, 244, 5, 5, 107, 246, 5, 5, 108, 246, 5, 5, 109, 246, 5, 5, 188, 247, 5, 5, 190, 248, 5, 5, 191, 248, 5, 5, 110, 157, 5, 5, 111, 157, 5, 5, 112, 157, 5, 5, 113, 157, 5, 5, 28, 158, 5, 5, 50, 159, 5, 5, 215, 160, 5, 5, 216, 160, 5, 5, 217, 160, 5, 5, 218, 160, 5, 5, 119, 163, 5, 5, 120, 163, 5, 5, 121, 163, 5, 5, 122, 163, 5, 5, 123, 163, 5, 5, 143, 167, 5, 5, 144, 167, 5, 5, 145, 167, 5, 5, 146, 167, 5, 5, 147, 167, 5, 5, 148, 167, 5, 5, 149, 167, 5, 5, 150, 167, 5, 5, 151, 167, 5, 5, 241, 172, 5, 5, 242, 172, 5, 5, 243, 172, 5, 5, 98, 179, 5, 5, 99, 179, 5, 5, 100, 179, 5, 5, 217, 186, 5, 5, 218, 186, 5, 5, 219, 186, 5, 5, 172, 194, 5, 5, 220, 186, 5, 5, 175, 194, 5, 5, 248, 202, 5, 5, 212, 210, 5, 5, 251, 202, 5, 5, 213, 210, 5, 5, 179, 224, 5, 5, 180, 224, 5, 5, 155, 230, 5, 5, 156, 230, 5, 5, 8, 239, 5, 5, 58, 242, 5, 5, 59, 242, 5, 5, 113, 246, 5, 5, 114, 157, 5, 5, 51, 159, 5, 5, 52, 159, 5, 5, 53, 159, 5, 5, 54, 159, 5, 5, 220, 160, 5, 5, 221, 160, 5, 5, 222, 160, 5, 5, 223, 160, 5, 5, 224, 160, 5, 5, 124, 163, 5, 5, 125, 163, 5, 5, 126, 163, 5, 5, 127, 163, 5, 5, 128, 163, 5, 5, 129, 163, 5, 5, 130, 163, 5, 5, 131, 163, 5, 5, 132, 163, 5, 5, 154, 167, 5, 5, 155, 167, 5, 5, 156, 167, 5, 5, 157, 167, 5, 5, 158, 167, 5, 5, 159, 167, 5, 5, 160, 167, 5, 5, 161, 167, 5, 5, 162, 167, 5, 5, 163, 167, 5, 5, 164, 167, 5, 5, 165, 167, 5, 5, 166, 167, 5, 5, 167, 167, 5, 5, 168, 167, 5, 5, 244, 172, 5, 5, 245, 172, 5, 5, 246, 172, 5, 5, 247, 172, 5, 5, 248, 172, 5, 5, 103, 179, 5, 5, 249, 172, 5, 5, 250, 172, 5, 5, 251, 172, 5, 5, 252, 172, 5, 5, 104, 179, 5, 5, 105, 179, 5, 5, 106, 179, 5, 5, 107, 179, 5, 5, 108, 179, 5, 5, 109, 179, 5, 5, 110, 179, 5, 5, 111, 179, 5, 5, 112, 179, 5, 5, 113, 179, 5, 5, 114, 179, 5, 5, 115, 179, 5, 5, 116, 179, 5, 5, 117, 179, 5, 5, 118, 179, 5, 5, 119, 179, 5, 5, 120, 179, 5, 5, 121, 179, 5, 5, 122, 179, 5, 5, 221, 186, 5, 5, 222, 186, 5, 5, 223, 186, 5, 5, 224, 186, 5, 5, 225, 186, 5, 5, 226, 186, 5, 5, 227, 186, 5, 5, 228, 186, 5, 5, 229, 186, 5, 5, 230, 186, 5, 5, 231, 186, 5, 5, 176, 194, 5, 5, 177, 194, 5, 5, 178, 194, 5, 5, 179, 194, 5, 5, 180, 194, 5, 5, 181, 194, 5, 5, 182, 194, 5, 5, 183, 194, 5, 5, 184, 194, 5, 5, 185, 194, 5, 5, 186, 194, 5, 5, 187, 194, 5, 5, 252, 202, 5, 5, 253, 202, 5, 5, 254, 202, 5, 5, 255, 202, 5, 5, 2, 203, 5, 5, 3, 203, 5, 5, 4, 203, 5, 5, 5, 203, 5, 5, 215, 210, 5, 5, 216, 210, 5, 5, 217, 210, 5, 5, 218, 210, 5, 5, 219, 210, 5, 5, 220, 210, 5, 5, 221, 210, 5, 5, 222, 210, 5, 5, 223, 210, 5, 5, 224, 210, 5, 5, 225, 210, 5, 5, 201, 217, 5, 5, 188, 194, 5, 5, 202, 217, 5, 5, 203, 217, 5, 5, 204, 217, 5, 5, 205, 217, 5, 5, 181, 224, 5, 5, 182, 224, 5, 5, 157, 230, 5, 5, 158, 230, 5, 5, 9, 239, 5, 5, 10, 239, 5, 5, 11, 239, 5, 5, 60, 242, 5, 5, 182, 244, 5, 5, 115, 157, 5, 5, 55, 159, 5, 5, 225, 160, 5, 5, 226, 160, 5, 5, 227, 160, 5, 5, 133, 163, 5, 5, 134, 163, 5, 5, 135, 163, 5, 5, 170, 167, 5, 5, 254, 172, 5, 5, 255, 172, 5, 5, 123, 179, 5, 5, 124, 179, 5, 5, 125, 179, 5, 5, 2, 173, 5, 5, 234, 186, 5, 5, 235, 186, 5, 5, 236, 186, 5, 5, 192, 194, 5, 5, 193, 194, 5, 5, 194, 194, 5, 5, 226, 210, 5, 5, 207, 217, 5, 5, 118, 157, 5, 5, 29, 158, 5, 5, 30, 158, 5, 5, 56, 159, 5, 5, 57, 159, 5, 5, 58, 159, 5, 5, 59, 159, 5, 5, 228, 160, 5, 5, 229, 160, 5, 5, 230, 160, 5, 5, 171, 167, 5, 5, 172, 167, 5, 5, 3, 173, 5, 5, 4, 173, 5, 5, 5, 173, 5, 5, 195, 194, 5, 5, 8, 203, 5, 5, 9, 203, 5, 5, 227, 210, 5, 5, 119, 157, 5, 5, 31, 158, 5, 5, 32, 158, 5, 5, 231, 160, 5, 5, 232, 160, 5, 5, 233, 160, 5, 5, 136, 163, 5, 5, 137, 163, 5, 5, 138, 163, 5, 5, 139, 163, 5, 5, 140, 163, 5, 5, 173, 167, 5, 5, 6, 173, 5, 5, 7, 173, 5, 5, 196, 194, 5, 5, 197, 194, 5, 5, 10, 203, 5, 5, 11, 203, 5, 5, 12, 203, 5, 5, 209, 217, 5, 5, 159, 230, 5, 5, 160, 230, 5, 5, 122, 157, 5, 5, 179, 157, 5, 5, 33, 158, 5, 5, 60, 159, 5, 5, 61, 159, 5, 5, 234, 160, 5, 5, 141, 163, 5, 5, 142, 163, 5, 5, 143, 163, 5, 5, 144, 163, 5, 5, 145, 163, 5, 5, 146, 163, 5, 5, 174, 167, 5, 5, 175, 167, 5, 5, 176, 167, 5, 5, 177, 167, 5, 5, 178, 167, 5, 5, 179, 167, 5, 5, 180, 167, 5, 5, 8, 173, 5, 5, 9, 173, 5, 5, 10, 173, 5, 5, 11, 173, 5, 5, 12, 173, 5, 5, 126, 179, 5, 5, 127, 179, 5, 5, 128, 179, 5, 5, 129, 179, 5, 5, 130, 179, 5, 5, 131, 179, 5, 5, 132, 179, 5, 5, 133, 179, 5, 5, 134, 179, 5, 5, 237, 186, 5, 5, 238, 186, 5, 5, 239, 186, 5, 5, 240, 186, 5, 5, 241, 186, 5, 5, 198, 194, 5, 5, 199, 194, 5, 5, 242, 186, 5, 5, 200, 194, 5, 5, 228, 210, 5, 5, 229, 210, 5, 5, 210, 217, 5, 5, 211, 217, 5, 5, 212, 217, 5, 5, 213, 217, 5, 5, 161, 230, 5, 5, 87, 235, 5, 5, 88, 235, 5, 5, 12, 239, 5, 5, 183, 244, 5, 5, 192, 248, 5, 5, 123, 157, 5, 5, 34, 158, 5, 5, 236, 160, 5, 5, 124, 157, 5, 5, 36, 158, 5, 5, 62, 159, 5, 5, 63, 159, 5, 5, 64, 159, 5, 5, 65, 159, 5, 5, 66, 159, 5, 5, 237, 160, 5, 5, 238, 160, 5, 5, 239, 160, 5, 5, 240, 160, 5, 5, 241, 160, 5, 5, 242, 160, 5, 5, 243, 160, 5, 5, 244, 160, 5, 5, 245, 160, 5, 5, 246, 160, 5, 5, 247, 160, 5, 5, 248, 160, 5, 5, 147, 163, 5, 5, 148, 163, 5, 5, 149, 163, 5, 5, 150, 163, 5, 5, 151, 163, 5, 5, 152, 163, 5, 5, 153, 163, 5, 5, 154, 163, 5, 5, 155, 163, 5, 5, 173, 163, 5, 5, 156, 163, 5, 5, 157, 163, 5, 5, 158, 163, 5, 5, 159, 163, 5, 5, 160, 163, 5, 5, 161, 163, 5, 5, 162, 163, 5, 5, 163, 163, 5, 5, 164, 163, 5, 5, 165, 163, 5, 5, 166, 163, 5, 5, 167, 163, 5, 5, 168, 163, 5, 5, 169, 163, 5, 5, 170, 163, 5, 5, 181, 167, 5, 5, 182, 167, 5, 5, 183, 167, 5, 5, 184, 167, 5, 5, 185, 167, 5, 5, 186, 167, 5, 5, 187, 167, 5, 5, 188, 167, 5, 5, 189, 167, 5, 5, 190, 167, 5, 5, 191, 167, 5, 5, 192, 167, 5, 5, 193, 167, 5, 5, 194, 167, 5, 5, 195, 167, 5, 5, 196, 167, 5, 5, 197, 167, 5, 5, 198, 167, 5, 5, 199, 167, 5, 5, 200, 167, 5, 5, 201, 167, 5, 5, 202, 167, 5, 5, 203, 167, 5, 5, 204, 167, 5, 5, 205, 167, 5, 5, 206, 167, 5, 5, 207, 167, 5, 5, 208, 167, 5, 5, 209, 167, 5, 5, 210, 167, 5, 5, 211, 167, 5, 5, 212, 167, 5, 5, 213, 167, 5, 5, 214, 167, 5, 5, 215, 167, 5, 5, 216, 167, 5, 5, 217, 167, 5, 5, 218, 167, 5, 5, 219, 167, 5, 5, 220, 167, 5, 5, 221, 167, 5, 5, 13, 173, 5, 5, 14, 173, 5, 5, 15, 173, 5, 5, 16, 173, 5, 5, 17, 173, 5, 5, 18, 173, 5, 5, 19, 173, 5, 5, 20, 173, 5, 5, 21, 173, 5, 5, 22, 173, 5, 5, 23, 173, 5, 5, 24, 173, 5, 5, 25, 173, 5, 5, 26, 173, 5, 5, 27, 173, 5, 5, 28, 173, 5, 5, 29, 173, 5, 5, 30, 173, 5, 5, 31, 173, 5, 5, 32, 173, 5, 5, 33, 173, 5, 5, 34, 173, 5, 5, 35, 173, 5, 5, 36, 173, 5, 5, 37, 173, 5, 5, 38, 173, 5, 5, 39, 173, 5, 5, 40, 173, 5, 5, 41, 173, 5, 5, 42, 173, 5, 5, 43, 173, 5, 5, 44, 173, 5, 5, 45, 173, 5, 5, 46, 173, 5, 5, 137, 179, 5, 5, 138, 179, 5, 5, 139, 179, 5, 5, 140, 179, 5, 5, 141, 179, 5, 5, 142, 179, 5, 5, 143, 179, 5, 5, 144, 179, 5, 5, 145, 179, 5, 5, 146, 179, 5, 5, 147, 179, 5, 5, 148, 179, 5, 5, 149, 179, 5, 5, 150, 179, 5, 5, 151, 179, 5, 5, 152, 179, 5, 5, 51, 173, 5, 5, 153, 179, 5, 5, 154, 179, 5, 5, 155, 179, 5, 5, 156, 179, 5, 5, 157, 179, 5, 5, 158, 179, 5, 5, 159, 179, 5, 5, 160, 179, 5, 5, 161, 179, 5, 5, 162, 179, 5, 5, 163, 179, 5, 5, 164, 179, 5, 5, 165, 179, 5, 5, 243, 186, 5, 5, 244, 186, 5, 5, 245, 186, 5, 5, 246, 186, 5, 5, 247, 186, 5, 5, 248, 186, 5, 5, 249, 186, 5, 5, 250, 186, 5, 5, 251, 186, 5, 5, 252, 186, 5, 5, 253, 186, 5, 5, 254, 186, 5, 5, 255, 186, 5, 5, 2, 187, 5, 5, 3, 187, 5, 5, 4, 187, 5, 5, 5, 187, 5, 5, 6, 187, 5, 5, 7, 187, 5, 5, 8, 187, 5, 5, 9, 187, 5, 5, 10, 187, 5, 5, 11, 187, 5, 5, 12, 187, 5, 5, 13, 187, 5, 5, 14, 187, 5, 5, 15, 187, 5, 5, 16, 187, 5, 5, 17, 187, 5, 5, 18, 187, 5, 5, 19, 187, 5, 5, 20, 187, 5, 5, 21, 187, 5, 5, 22, 187, 5, 5, 23, 187, 5, 5, 24, 187, 5, 5, 25, 187, 5, 5, 26, 187, 5, 5, 27, 187, 5, 5, 28, 187, 5, 5, 29, 187, 5, 5, 30, 187, 5, 5, 31, 187, 5, 5, 203, 194, 5, 5, 204, 194, 5, 5, 205, 194, 5, 5, 206, 194, 5, 5, 207, 194, 5, 5, 208, 194, 5, 5, 209, 194, 5, 5, 210, 194, 5, 5, 211, 194, 5, 5, 212, 194, 5, 5, 213, 194, 5, 5, 214, 194, 5, 5, 215, 194, 5, 5, 216, 194, 5, 5, 217, 194, 5, 5, 218, 194, 5, 5, 219, 194, 5, 5, 220, 194, 5, 5, 221, 194, 5, 5, 222, 194, 5, 5, 223, 194, 5, 5, 224, 194, 5, 5, 225, 194, 5, 5, 226, 194, 5, 5, 227, 194, 5, 5, 14, 203, 5, 5, 228, 194, 5, 5, 229, 194, 5, 5, 230, 194, 5, 5, 231, 194, 5, 5, 232, 194, 5, 5, 233, 194, 5, 5, 234, 194, 5, 5, 235, 194, 5, 5, 236, 194, 5, 5, 237, 194, 5, 5, 238, 194, 5, 5, 239, 194, 5, 5, 240, 194, 5, 5, 241, 194, 5, 5, 242, 194, 5, 5, 243, 194, 5, 5, 244, 194, 5, 5, 245, 194, 5, 5, 246, 194, 5, 5, 15, 203, 5, 5, 16, 203, 5, 5, 17, 203, 5, 5, 18, 203, 5, 5, 19, 203, 5, 5, 20, 203, 5, 5, 21, 203, 5, 5, 22, 203, 5, 5, 23, 203, 5, 5, 24, 203, 5, 5, 25, 203, 5, 5, 26, 203, 5, 5, 27, 203, 5, 5, 251, 194, 5, 5, 28, 203, 5, 5, 29, 203, 5, 5, 30, 203, 5, 5, 31, 203, 5, 5, 32, 203, 5, 5, 33, 203, 5, 5, 34, 203, 5, 5, 35, 203, 5, 5, 36, 203, 5, 5, 37, 203, 5, 5, 38, 203, 5, 5, 230, 210, 5, 5, 231, 210, 5, 5, 232, 210, 5, 5, 233, 210, 5, 5, 234, 210, 5, 5, 235, 210, 5, 5, 236, 210, 5, 5, 237, 210, 5, 5, 238, 210, 5, 5, 239, 210, 5, 5, 240, 210, 5, 5, 241, 210, 5, 5, 242, 210, 5, 5, 243, 210, 5, 5, 244, 210, 5, 5, 245, 210, 5, 5, 246, 210, 5, 5, 247, 210, 5, 5, 248, 210, 5, 5, 249, 210, 5, 5, 250, 210, 5, 5, 251, 210, 5, 5, 252, 210, 5, 5, 253, 210, 5, 5, 215, 217, 5, 5, 216, 217, 5, 5, 217, 217, 5, 5, 218, 217, 5, 5, 219, 217, 5, 5, 220, 217, 5, 5, 221, 217, 5, 5, 222, 217, 5, 5, 223, 217, 5, 5, 224, 217, 5, 5, 225, 217, 5, 5, 226, 217, 5, 5, 227, 217, 5, 5, 228, 217, 5, 5, 229, 217, 5, 5, 230, 217, 5, 5, 231, 217, 5, 5, 232, 217, 5, 5, 233, 217, 5, 5, 234, 217, 5, 5, 235, 217, 5, 5, 236, 217, 5, 5, 237, 217, 5, 5, 183, 224, 5, 5, 184, 224, 5, 5, 185, 224, 5, 5, 186, 224, 5, 5, 187, 224, 5, 5, 188, 224, 5, 5, 189, 224, 5, 5, 190, 224, 5, 5, 191, 224, 5, 5, 192, 224, 5, 5, 193, 224, 5, 5, 194, 224, 5, 5, 195, 224, 5, 5, 196, 224, 5, 5, 197, 224, 5, 5, 198, 224, 5, 5, 199, 224, 5, 5, 162, 230, 5, 5, 163, 230, 5, 5, 164, 230, 5, 5, 165, 230, 5, 5, 172, 230, 5, 5, 166, 230, 5, 5, 167, 230, 5, 5, 168, 230, 5, 5, 169, 230, 5, 5, 89, 235, 5, 5, 90, 235, 5, 5, 91, 235, 5, 5, 13, 239, 5, 5, 14, 239, 5, 5, 15, 239, 5, 5, 61, 242, 5, 5, 62, 242, 5, 5, 63, 242, 5, 5, 64, 242, 5, 5, 65, 242, 5, 5, 184, 244, 5, 5, 66, 242, 5, 5, 185, 244, 5, 5, 114, 246, 5, 5, 186, 244, 5, 5, 187, 244, 5, 5, 115, 246, 5, 5, 116, 246, 5, 5, 117, 246, 5, 5, 118, 246, 5, 5, 119, 246, 5, 5, 189, 247, 5, 5, 120, 246, 5, 5, 190, 247, 5, 5, 193, 248, 5, 5, 191, 247, 5, 5, 125, 157, 5, 5, 45, 157, 5, 5, 126, 157, 5, 5, 250, 160, 5, 5, 251, 160, 5, 5, 174, 163, 5, 5, 167, 165, 5, 5, 36, 187, 5, 5, 37, 187, 5, 5, 241, 217, 5, 5, 128, 157, 5, 5, 67, 159, 5, 5, 68, 159, 5, 5, 69, 159, 5, 5, 252, 160, 5, 5, 253, 160, 5, 5, 175, 163, 5, 5, 52, 173, 5, 5, 53, 173, 5, 5, 169, 179, 5, 5, 252, 194, 5, 5, 43, 203, 5, 5, 129, 157, 5, 5, 130, 157, 5, 5, 131, 157, 5, 5, 37, 158, 5, 5, 176, 163, 5, 5, 225, 167, 5, 5, 54, 173, 5, 5, 55, 173, 5, 5, 56, 173, 5, 5, 57, 173, 5, 5, 58, 173, 5, 5, 171, 179, 5, 5, 254, 194, 5, 5, 132, 157, 5, 5, 38, 158, 5, 5, 39, 158, 5, 5, 40, 158, 5, 5, 70, 159, 5, 5, 71, 159, 5, 5, 72, 159, 5, 5, 73, 159, 5, 5, 254, 160, 5, 5, 255, 160, 5, 5, 2, 161, 5, 5, 177, 163, 5, 5, 178, 163, 5, 5, 179, 163, 5, 5, 180, 163, 5, 5, 181, 163, 5, 5, 182, 163, 5, 5, 183, 163, 5, 5, 184, 163, 5, 5, 227, 167, 5, 5, 228, 167, 5, 5, 229, 167, 5, 5, 230, 167, 5, 5, 231, 167, 5, 5, 232, 167, 5, 5, 233, 167, 5, 5, 234, 167, 5, 5, 235, 167, 5, 5, 236, 167, 5, 5, 237, 167, 5, 5, 238, 167, 5, 5, 59, 173, 5, 5, 60, 173, 5, 5, 61, 173, 5, 5, 62, 173, 5, 5, 63, 173, 5, 5, 64, 173, 5, 5, 65, 173, 5, 5, 66, 173, 5, 5, 67, 173, 5, 5, 68, 173, 5, 5, 69, 173, 5, 5, 172, 179, 5, 5, 173, 179, 5, 5, 174, 179, 5, 5, 175, 179, 5, 5, 176, 179, 5, 5, 177, 179, 5, 5, 178, 179, 5, 5, 179, 179, 5, 5, 180, 179, 5, 5, 181, 179, 5, 5, 39, 187, 5, 5, 40, 187, 5, 5, 41, 187, 5, 5, 42, 187, 5, 5, 43, 187, 5, 5, 44, 187, 5, 5, 45, 187, 5, 5, 46, 187, 5, 5, 47, 187, 5, 5, 48, 187, 5, 5, 49, 187, 5, 5, 3, 195, 5, 5, 50, 187, 5, 5, 4, 195, 5, 5, 5, 195, 5, 5, 6, 195, 5, 5, 7, 195, 5, 5, 8, 195, 5, 5, 9, 195, 5, 5, 10, 195, 5, 5, 11, 195, 5, 5, 12, 195, 5, 5, 15, 195, 5, 5, 13, 195, 5, 5, 44, 203, 5, 5, 45, 203, 5, 5, 46, 203, 5, 5, 47, 203, 5, 5, 48, 203, 5, 5, 49, 203, 5, 5, 219, 206, 5, 5, 4, 211, 5, 5, 5, 211, 5, 5, 6, 211, 5, 5, 7, 211, 5, 5, 8, 211, 5, 5, 9, 211, 5, 5, 10, 211, 5, 5, 11, 211, 5, 5, 12, 211, 5, 5, 242, 217, 5, 5, 13, 211, 5, 5, 243, 217, 5, 5, 244, 217, 5, 5, 245, 217, 5, 5, 246, 217, 5, 5, 247, 217, 5, 5, 248, 217, 5, 5, 249, 217, 5, 5, 15, 211, 5, 5, 250, 217, 5, 5, 251, 217, 5, 5, 200, 224, 5, 5, 201, 224, 5, 5, 202, 224, 5, 5, 253, 217, 5, 5, 173, 230, 5, 5, 174, 230, 5, 5, 175, 230, 5, 5, 92, 235, 5, 5, 93, 235, 5, 5, 207, 224, 5, 5, 20, 239, 5, 5, 67, 242, 5, 5, 133, 157, 5, 5, 76, 159, 5, 5, 6, 161, 5, 5, 7, 161, 5, 5, 8, 161, 5, 5, 240, 167, 5, 5, 241, 167, 5, 5, 50, 203, 5, 5, 135, 157, 5, 5, 41, 158, 5, 5, 77, 159, 5, 5, 71, 173, 5, 5, 16, 195, 5, 5, 136, 157, 5, 5, 78, 159, 5, 5, 79, 159, 5, 5, 80, 159, 5, 5, 81, 159, 5, 5, 9, 161, 5, 5, 10, 161, 5, 5, 11, 161, 5, 5, 189, 163, 5, 5, 190, 163, 5, 5, 191, 163, 5, 5, 192, 163, 5, 5, 193, 163, 5, 5, 194, 163, 5, 5, 195, 163, 5, 5, 196, 163, 5, 5, 197, 163, 5, 5, 198, 163, 5, 5, 199, 163, 5, 5, 200, 163, 5, 5, 201, 163, 5, 5, 202, 163, 5, 5, 243, 167, 5, 5, 244, 167, 5, 5, 245, 167, 5, 5, 242, 167, 5, 5, 246, 167, 5, 5, 247, 167, 5, 5, 72, 173, 5, 5, 248, 167, 5, 5, 249, 167, 5, 5, 250, 167, 5, 5, 251, 167, 5, 5, 73, 173, 5, 5, 74, 173, 5, 5, 75, 173, 5, 5, 76, 173, 5, 5, 77, 173, 5, 5, 78, 173, 5, 5, 79, 173, 5, 5, 185, 179, 5, 5, 186, 179, 5, 5, 187, 179, 5, 5, 188, 179, 5, 5, 189, 179, 5, 5, 190, 179, 5, 5, 191, 179, 5, 5, 192, 179, 5, 5, 193, 179, 5, 5, 81, 173, 5, 5, 51, 187, 5, 5, 52, 187, 5, 5, 53, 187, 5, 5, 54, 187, 5, 5, 55, 187, 5, 5, 56, 187, 5, 5, 57, 187, 5, 5, 58, 187, 5, 5, 59, 187, 5, 5, 60, 187, 5, 5, 61, 187, 5, 5, 62, 187, 5, 5, 17, 195, 5, 5, 63, 187, 5, 5, 18, 195, 5, 5, 19, 195, 5, 5, 20, 195, 5, 5, 21, 195, 5, 5, 22, 195, 5, 5, 24, 195, 5, 5, 51, 203, 5, 5, 52, 203, 5, 5, 53, 203, 5, 5, 54, 203, 5, 5, 55, 203, 5, 5, 23, 195, 5, 5, 56, 203, 5, 5, 57, 203, 5, 5, 16, 211, 5, 5, 17, 211, 5, 5, 18, 211, 5, 5, 19, 211, 5, 5, 20, 211, 5, 5, 21, 211, 5, 5, 22, 211, 5, 5, 23, 211, 5, 5, 24, 211, 5, 5, 25, 211, 5, 5, 26, 211, 5, 5, 27, 211, 5, 5, 28, 211, 5, 5, 254, 217, 5, 5, 255, 217, 5, 5, 29, 211, 5, 5, 2, 218, 5, 5, 3, 218, 5, 5, 4, 218, 5, 5, 5, 218, 5, 5, 6, 218, 5, 5, 7, 218, 5, 5, 8, 218, 5, 5, 9, 218, 5, 5, 208, 224, 5, 5, 209, 224, 5, 5, 210, 224, 5, 5, 211, 224, 5, 5, 212, 224, 5, 5, 213, 224, 5, 5, 94, 235, 5, 5, 22, 239, 5, 5, 23, 239, 5, 5, 69, 242, 5, 5, 70, 242, 5, 5, 71, 242, 5, 5, 189, 244, 5, 5, 121, 246, 5, 5, 104, 249, 5, 5, 137, 157, 5, 5, 12, 161, 5, 5, 253, 167, 5, 5, 203, 163, 5, 5, 254, 167, 5, 5, 255, 167, 5, 5, 82, 173, 5, 5, 83, 173, 5, 5, 84, 173, 5, 5, 194, 179, 5, 5, 138, 157, 5, 5, 42, 158, 5, 5, 43, 158, 5, 5, 82, 159, 5, 5, 13, 161, 5, 5, 204, 163, 5, 5, 205, 163, 5, 5, 206, 163, 5, 5, 3, 168, 5, 5, 87, 173, 5, 5, 88, 173, 5, 5, 195, 179, 5, 5, 30, 211, 5, 5, 139, 157, 5, 5, 44, 158, 5, 5, 83, 159, 5, 5, 14, 161, 5, 5, 15, 161, 5, 5, 16, 161, 5, 5, 25, 195, 5, 5, 59, 203, 5, 5, 140, 157, 5, 5, 45, 158, 5, 5, 46, 158, 5, 5, 47, 158, 5, 5, 84, 159, 5, 5, 85, 159, 5, 5, 17, 161, 5, 5, 18, 161, 5, 5, 19, 161, 5, 5, 20, 161, 5, 5, 207, 163, 5, 5, 208, 163, 5, 5, 209, 163, 5, 5, 210, 163, 5, 5, 4, 168, 5, 5, 5, 168, 5, 5, 6, 168, 5, 5, 7, 168, 5, 5, 8, 168, 5, 5, 9, 168, 5, 5, 10, 168, 5, 5, 11, 168, 5, 5, 12, 168, 5, 5, 13, 168, 5, 5, 89, 173, 5, 5, 90, 173, 5, 5, 91, 173, 5, 5, 92, 173, 5, 5, 93, 173, 5, 5, 196, 179, 5, 5, 197, 179, 5, 5, 198, 179, 5, 5, 199, 179, 5, 5, 64, 187, 5, 5, 65, 187, 5, 5, 66, 187, 5, 5, 67, 187, 5, 5, 68, 187, 5, 5, 69, 187, 5, 5, 26, 195, 5, 5, 27, 195, 5, 5, 28, 195, 5, 5, 29, 195, 5, 5, 30, 195, 5, 5, 60, 203, 5, 5, 61, 203, 5, 5, 62, 203, 5, 5, 63, 203, 5, 5, 31, 211, 5, 5, 32, 211, 5, 5, 33, 211, 5, 5, 35, 211, 5, 5, 10, 218, 5, 5, 11, 218, 5, 5, 12, 218, 5, 5, 214, 224, 5, 5, 215, 224, 5, 5, 178, 230, 5, 5, 95, 235, 5, 5, 122, 246, 5, 5, 192, 247, 5, 5, 141, 157, 5, 5, 143, 157, 5, 5, 86, 159, 5, 5, 21, 161, 5, 5, 15, 168, 5, 5, 16, 168, 5, 5, 95, 173, 5, 5, 71, 187, 5, 5, 31, 195, 5, 5, 64, 203, 5, 5, 65, 203, 5, 5, 216, 224, 5, 5, 217, 224, 5, 5, 96, 235, 5, 5, 97, 235, 5, 5, 25, 239, 5, 5, 223, 249, 5, 5, 144, 157, 5, 5, 212, 163, 5, 5, 213, 163, 5, 5, 214, 163, 5, 5, 96, 173, 5, 5, 97, 173, 5, 5, 200, 179, 5, 5, 201, 179, 5, 5, 72, 187, 5, 5, 151, 191, 5, 5, 73, 187, 5, 5, 74, 187, 5, 5, 32, 195, 5, 5, 67, 203, 5, 5, 36, 211, 5, 5, 37, 211, 5, 5, 13, 218, 5, 5, 123, 246, 5, 5, 145, 157, 5, 5, 22, 161, 5, 5, 23, 161, 5, 5, 215, 163, 5, 5, 216, 163, 5, 5, 217, 163, 5, 5, 218, 163, 5, 5, 219, 163, 5, 5, 220, 163, 5, 5, 19, 168, 5, 5, 20, 168, 5, 5, 21, 168, 5, 5, 22, 168, 5, 5, 23, 168, 5, 5, 24, 168, 5, 5, 25, 168, 5, 5, 26, 168, 5, 5, 27, 168, 5, 5, 98, 173, 5, 5, 99, 173, 5, 5, 100, 173, 5, 5, 101, 173, 5, 5, 102, 173, 5, 5, 103, 173, 5, 5, 104, 173, 5, 5, 105, 173, 5, 5, 106, 173, 5, 5, 203, 179, 5, 5, 204, 179, 5, 5, 205, 179, 5, 5, 206, 179, 5, 5, 207, 179, 5, 5, 208, 179, 5, 5, 107, 173, 5, 5, 209, 179, 5, 5, 75, 187, 5, 5, 76, 187, 5, 5, 77, 187, 5, 5, 78, 187, 5, 5, 33, 195, 5, 5, 79, 187, 5, 5, 80, 187, 5, 5, 81, 187, 5, 5, 82, 187, 5, 5, 83, 187, 5, 5, 84, 187, 5, 5, 85, 187, 5, 5, 86, 187, 5, 5, 87, 187, 5, 5, 88, 187, 5, 5, 34, 195, 5, 5, 35, 195, 5, 5, 36, 195, 5, 5, 37, 195, 5, 5, 38, 195, 5, 5, 39, 195, 5, 5, 40, 195, 5, 5, 68, 203, 5, 5, 69, 203, 5, 5, 70, 203, 5, 5, 71, 203, 5, 5, 72, 203, 5, 5, 39, 211, 5, 5, 14, 218, 5, 5, 40, 211, 5, 5, 41, 211, 5, 5, 15, 218, 5, 5, 43, 211, 5, 5, 16, 218, 5, 5, 17, 218, 5, 5, 18, 218, 5, 5, 19, 218, 5, 5, 220, 224, 5, 5, 221, 224, 5, 5, 179, 230, 5, 5, 180, 230, 5, 5, 26, 239, 5, 5, 72, 242, 5, 5, 73, 242, 5, 5, 190, 244, 5, 5, 49, 158, 5, 5, 146, 157, 5, 5, 88, 159, 5, 5, 50, 158, 5, 5, 90, 159, 5, 5, 25, 161, 5, 5, 91, 159, 5, 5, 92, 159, 5, 5, 26, 161, 5, 5, 225, 163, 5, 5, 226, 163, 5, 5, 227, 163, 5, 5, 27, 161, 5, 5, 228, 163, 5, 5, 229, 163, 5, 5, 230, 163, 5, 5, 28, 161, 5, 5, 29, 161, 5, 5, 30, 161, 5, 5, 31, 161, 5, 5, 231, 163, 5, 5, 232, 163, 5, 5, 32, 161, 5, 5, 33, 161, 5, 5, 34, 161, 5, 5, 233, 163, 5, 5, 28, 168, 5, 5, 29, 168, 5, 5, 235, 163, 5, 5, 30, 168, 5, 5, 236, 163, 5, 5, 31, 168, 5, 5, 237, 163, 5, 5, 238, 163, 5, 5, 32, 168, 5, 5, 239, 163, 5, 5, 240, 163, 5, 5, 241, 163, 5, 5, 33, 168, 5, 5, 242, 163, 5, 5, 243, 163, 5, 5, 244, 163, 5, 5, 245, 163, 5, 5, 246, 163, 5, 5, 247, 163, 5, 5, 248, 163, 5, 5, 249, 163, 5, 5, 250, 163, 5, 5, 251, 163, 5, 5, 252, 163, 5, 5, 34, 168, 5, 5, 253, 163, 5, 5, 254, 163, 5, 5, 255, 163, 5, 5, 2, 164, 5, 5, 3, 164, 5, 5, 4, 164, 5, 5, 5, 164, 5, 5, 35, 168, 5, 5, 6, 164, 5, 5, 36, 168, 5, 5, 7, 164, 5, 5, 37, 168, 5, 5, 38, 168, 5, 5, 8, 164, 5, 5, 9, 164, 5, 5, 10, 164, 5, 5, 11, 164, 5, 5, 39, 168, 5, 5, 40, 168, 5, 5, 41, 168, 5, 5, 42, 168, 5, 5, 43, 168, 5, 5, 44, 168, 5, 5, 45, 168, 5, 5, 108, 173, 5, 5, 46, 168, 5, 5, 47, 168, 5, 5, 48, 168, 5, 5, 109, 173, 5, 5, 49, 168, 5, 5, 50, 168, 5, 5, 51, 168, 5, 5, 52, 168, 5, 5, 53, 168, 5, 5, 110, 173, 5, 5, 54, 168, 5, 5, 55, 168, 5, 5, 56, 168, 5, 5, 57, 168, 5, 5, 111, 173, 5, 5, 58, 168, 5, 5, 59, 168, 5, 5, 112, 173, 5, 5, 60, 168, 5, 5, 61, 168, 5, 5, 113, 173, 5, 5, 114, 173, 5, 5, 115, 173, 5, 5, 62, 168, 5, 5, 63, 168, 5, 5, 116, 173, 5, 5, 64, 168, 5, 5, 65, 168, 5, 5, 66, 168, 5, 5, 67, 168, 5, 5, 68, 168, 5, 5, 69, 168, 5, 5, 70, 168, 5, 5, 71, 168, 5, 5, 117, 173, 5, 5, 72, 168, 5, 5, 73, 168, 5, 5, 74, 168, 5, 5, 75, 168, 5, 5, 76, 168, 5, 5, 118, 173, 5, 5, 119, 173, 5, 5, 120, 173, 5, 5, 77, 168, 5, 5, 121, 173, 5, 5, 122, 173, 5, 5, 78, 168, 5, 5, 79, 168, 5, 5, 80, 168, 5, 5, 123, 173, 5, 5, 211, 179, 5, 5, 124, 173, 5, 5, 125, 173, 5, 5, 126, 173, 5, 5, 127, 173, 5, 5, 128, 173, 5, 5, 129, 173, 5, 5, 130, 173, 5, 5, 131, 173, 5, 5, 132, 173, 5, 5, 212, 179, 5, 5, 133, 173, 5, 5, 134, 173, 5, 5, 135, 173, 5, 5, 213, 179, 5, 5, 214, 179, 5, 5, 136, 173, 5, 5, 137, 173, 5, 5, 138, 173, 5, 5, 139, 173, 5, 5, 215, 179, 5, 5, 216, 179, 5, 5, 140, 173, 5, 5, 141, 173, 5, 5, 217, 179, 5, 5, 218, 179, 5, 5, 142, 173, 5, 5, 143, 173, 5, 5, 219, 179, 5, 5, 144, 173, 5, 5, 145, 173, 5, 5, 146, 173, 5, 5, 147, 173, 5, 5, 148, 173, 5, 5, 220, 179, 5, 5, 149, 173, 5, 5, 221, 179, 5, 5, 150, 173, 5, 5, 222, 179, 5, 5, 151, 173, 5, 5, 223, 179, 5, 5, 152, 173, 5, 5, 153, 173, 5, 5, 154, 173, 5, 5, 224, 179, 5, 5, 155, 173, 5, 5, 225, 179, 5, 5, 156, 173, 5, 5, 157, 173, 5, 5, 158, 173, 5, 5, 226, 179, 5, 5, 227, 179, 5, 5, 228, 179, 5, 5, 229, 179, 5, 5, 230, 179, 5, 5, 159, 173, 5, 5, 160, 173, 5, 5, 161, 173, 5, 5, 162, 173, 5, 5, 163, 173, 5, 5, 164, 173, 5, 5, 233, 179, 5, 5, 91, 187, 5, 5, 234, 179, 5, 5, 235, 179, 5, 5, 236, 179, 5, 5, 237, 179, 5, 5, 238, 179, 5, 5, 239, 179, 5, 5, 92, 187, 5, 5, 240, 179, 5, 5, 241, 179, 5, 5, 93, 187, 5, 5, 94, 187, 5, 5, 242, 179, 5, 5, 243, 179, 5, 5, 244, 179, 5, 5, 245, 179, 5, 5, 246, 179, 5, 5, 95, 187, 5, 5, 247, 179, 5, 5, 248, 179, 5, 5, 249, 179, 5, 5, 250, 179, 5, 5, 251, 179, 5, 5, 252, 179, 5, 5, 253, 179, 5, 5, 96, 187, 5, 5, 254, 179, 5, 5, 255, 179, 5, 5, 2, 180, 5, 5, 3, 180, 5, 5, 4, 180, 5, 5, 5, 180, 5, 5, 6, 180, 5, 5, 97, 187, 5, 5, 98, 187, 5, 5, 7, 180, 5, 5, 99, 187, 5, 5, 100, 187, 5, 5, 101, 187, 5, 5, 8, 180, 5, 5, 9, 180, 5, 5, 102, 187, 5, 5, 10, 180, 5, 5, 103, 187, 5, 5, 104, 187, 5, 5, 105, 187, 5, 5, 11, 180, 5, 5, 12, 180, 5, 5, 13, 180, 5, 5, 110, 187, 5, 5, 111, 187, 5, 5, 42, 195, 5, 5, 43, 195, 5, 5, 112, 187, 5, 5, 113, 187, 5, 5, 44, 195, 5, 5, 114, 187, 5, 5, 115, 187, 5, 5, 45, 195, 5, 5, 116, 187, 5, 5, 117, 187, 5, 5, 118, 187, 5, 5, 119, 187, 5, 5, 120, 187, 5, 5, 121, 187, 5, 5, 122, 187, 5, 5, 46, 195, 5, 5, 123, 187, 5, 5, 124, 187, 5, 5, 47, 195, 5, 5, 125, 187, 5, 5, 126, 187, 5, 5, 127, 187, 5, 5, 128, 187, 5, 5, 48, 195, 5, 5, 129, 187, 5, 5, 130, 187, 5, 5, 49, 195, 5, 5, 131, 187, 5, 5, 50, 195, 5, 5, 132, 187, 5, 5, 133, 187, 5, 5, 51, 195, 5, 5, 52, 195, 5, 5, 134, 187, 5, 5, 135, 187, 5, 5, 136, 187, 5, 5, 53, 195, 5, 5, 137, 187, 5, 5, 138, 187, 5, 5, 139, 187, 5, 5, 140, 187, 5, 5, 141, 187, 5, 5, 142, 187, 5, 5, 143, 187, 5, 5, 144, 187, 5, 5, 145, 187, 5, 5, 54, 195, 5, 5, 55, 195, 5, 5, 56, 195, 5, 5, 57, 195, 5, 5, 146, 187, 5, 5, 58, 195, 5, 5, 147, 187, 5, 5, 148, 187, 5, 5, 149, 187, 5, 5, 59, 195, 5, 5, 60, 195, 5, 5, 61, 195, 5, 5, 150, 187, 5, 5, 151, 187, 5, 5, 152, 187, 5, 5, 153, 187, 5, 5, 62, 195, 5, 5, 63, 195, 5, 5, 64, 195, 5, 5, 73, 203, 5, 5, 65, 195, 5, 5, 66, 195, 5, 5, 67, 195, 5, 5, 74, 203, 5, 5, 68, 195, 5, 5, 75, 203, 5, 5, 69, 195, 5, 5, 70, 195, 5, 5, 71, 195, 5, 5, 72, 195, 5, 5, 73, 195, 5, 5, 74, 195, 5, 5, 75, 195, 5, 5, 76, 203, 5, 5, 77, 203, 5, 5, 76, 195, 5, 5, 77, 195, 5, 5, 78, 195, 5, 5, 78, 203, 5, 5, 79, 195, 5, 5, 79, 203, 5, 5, 80, 195, 5, 5, 81, 195, 5, 5, 82, 195, 5, 5, 83, 195, 5, 5, 80, 203, 5, 5, 84, 195, 5, 5, 81, 203, 5, 5, 85, 195, 5, 5, 86, 195, 5, 5, 87, 195, 5, 5, 88, 195, 5, 5, 89, 195, 5, 5, 90, 195, 5, 5, 91, 195, 5, 5, 82, 203, 5, 5, 92, 195, 5, 5, 83, 203, 5, 5, 84, 203, 5, 5, 85, 203, 5, 5, 93, 195, 5, 5, 94, 195, 5, 5, 95, 195, 5, 5, 86, 203, 5, 5, 96, 195, 5, 5, 97, 195, 5, 5, 98, 195, 5, 5, 99, 195, 5, 5, 100, 195, 5, 5, 101, 195, 5, 5, 102, 195, 5, 5, 90, 203, 5, 5, 44, 211, 5, 5, 91, 203, 5, 5, 92, 203, 5, 5, 93, 203, 5, 5, 45, 211, 5, 5, 94, 203, 5, 5, 95, 203, 5, 5, 96, 203, 5, 5, 97, 203, 5, 5, 98, 203, 5, 5, 99, 203, 5, 5, 46, 211, 5, 5, 100, 203, 5, 5, 101, 203, 5, 5, 102, 203, 5, 5, 103, 203, 5, 5, 47, 211, 5, 5, 104, 203, 5, 5, 105, 203, 5, 5, 48, 211, 5, 5, 106, 203, 5, 5, 107, 203, 5, 5, 108, 203, 5, 5, 49, 211, 5, 5, 109, 203, 5, 5, 50, 211, 5, 5, 51, 211, 5, 5, 110, 203, 5, 5, 111, 203, 5, 5, 112, 203, 5, 5, 113, 203, 5, 5, 52, 211, 5, 5, 53, 211, 5, 5, 114, 203, 5, 5, 115, 203, 5, 5, 54, 211, 5, 5, 116, 203, 5, 5, 117, 203, 5, 5, 118, 203, 5, 5, 119, 203, 5, 5, 55, 211, 5, 5, 120, 203, 5, 5, 57, 211, 5, 5, 58, 211, 5, 5, 59, 211, 5, 5, 20, 218, 5, 5, 60, 211, 5, 5, 21, 218, 5, 5, 61, 211, 5, 5, 22, 218, 5, 5, 62, 211, 5, 5, 63, 211, 5, 5, 23, 218, 5, 5, 24, 218, 5, 5, 64, 211, 5, 5, 65, 211, 5, 5, 66, 211, 5, 5, 67, 211, 5, 5, 68, 211, 5, 5, 69, 211, 5, 5, 25, 218, 5, 5, 70, 211, 5, 5, 26, 218, 5, 5, 27, 218, 5, 5, 103, 195, 5, 5, 71, 211, 5, 5, 72, 211, 5, 5, 28, 218, 5, 5, 73, 211, 5, 5, 42, 218, 5, 5, 29, 218, 5, 5, 74, 211, 5, 5, 30, 218, 5, 5, 75, 211, 5, 5, 76, 211, 5, 5, 77, 211, 5, 5, 78, 211, 5, 5, 79, 211, 5, 5, 31, 218, 5, 5, 80, 211, 5, 5, 32, 218, 5, 5, 33, 218, 5, 5, 81, 211, 5, 5, 82, 211, 5, 5, 34, 218, 5, 5, 83, 211, 5, 5, 35, 218, 5, 5, 36, 218, 5, 5, 84, 211, 5, 5, 85, 211, 5, 5, 37, 218, 5, 5, 38, 218, 5, 5, 39, 218, 5, 5, 40, 218, 5, 5, 86, 211, 5, 5, 41, 218, 5, 5, 87, 211, 5, 5, 43, 218, 5, 5, 222, 224, 5, 5, 44, 218, 5, 5, 223, 224, 5, 5, 45, 218, 5, 5, 46, 218, 5, 5, 47, 218, 5, 5, 48, 218, 5, 5, 224, 224, 5, 5, 49, 218, 5, 5, 50, 218, 5, 5, 51, 218, 5, 5, 52, 218, 5, 5, 225, 224, 5, 5, 226, 224, 5, 5, 53, 218, 5, 5, 227, 224, 5, 5, 54, 218, 5, 5, 55, 218, 5, 5, 56, 218, 5, 5, 228, 224, 5, 5, 57, 218, 5, 5, 58, 218, 5, 5, 229, 224, 5, 5, 59, 218, 5, 5, 60, 218, 5, 5, 61, 218, 5, 5, 62, 218, 5, 5, 230, 224, 5, 5, 63, 218, 5, 5, 64, 218, 5, 5, 231, 224, 5, 5, 232, 224, 5, 5, 65, 218, 5, 5, 66, 218, 5, 5, 67, 218, 5, 5, 68, 218, 5, 5, 69, 218, 5, 5, 70, 218, 5, 5, 71, 218, 5, 5, 72, 218, 5, 5, 233, 224, 5, 5, 73, 218, 5, 5, 234, 224, 5, 5, 182, 230, 5, 5, 235, 224, 5, 5, 236, 224, 5, 5, 237, 224, 5, 5, 238, 224, 5, 5, 239, 224, 5, 5, 7, 225, 5, 5, 183, 230, 5, 5, 240, 224, 5, 5, 241, 224, 5, 5, 242, 224, 5, 5, 243, 224, 5, 5, 244, 224, 5, 5, 184, 230, 5, 5, 185, 230, 5, 5, 245, 224, 5, 5, 246, 224, 5, 5, 247, 224, 5, 5, 186, 230, 5, 5, 248, 224, 5, 5, 187, 230, 5, 5, 249, 224, 5, 5, 188, 230, 5, 5, 250, 224, 5, 5, 251, 224, 5, 5, 252, 224, 5, 5, 253, 224, 5, 5, 254, 224, 5, 5, 189, 230, 5, 5, 255, 224, 5, 5, 2, 225, 5, 5, 3, 225, 5, 5, 98, 235, 5, 5, 99, 235, 5, 5, 190, 230, 5, 5, 100, 235, 5, 5, 191, 230, 5, 5, 192, 230, 5, 5, 193, 230, 5, 5, 194, 230, 5, 5, 195, 230, 5, 5, 196, 230, 5, 5, 101, 235, 5, 5, 197, 230, 5, 5, 198, 230, 5, 5, 199, 230, 5, 5, 102, 235, 5, 5, 200, 230, 5, 5, 201, 230, 5, 5, 202, 230, 5, 5, 203, 230, 5, 5, 204, 230, 5, 5, 104, 235, 5, 5, 105, 235, 5, 5, 106, 235, 5, 5, 27, 239, 5, 5, 107, 235, 5, 5, 108, 235, 5, 5, 28, 239, 5, 5, 109, 235, 5, 5, 110, 235, 5, 5, 29, 239, 5, 5, 111, 235, 5, 5, 112, 235, 5, 5, 30, 239, 5, 5, 31, 239, 5, 5, 32, 239, 5, 5, 74, 242, 5, 5, 76, 242, 5, 5, 77, 242, 5, 5, 35, 239, 5, 5, 191, 244, 5, 5, 192, 244, 5, 5, 193, 244, 5, 5, 124, 246, 5, 5, 193, 247, 5, 5, 194, 247, 5, 5, 125, 246, 5, 5, 195, 247, 5, 5, 196, 247, 5, 5, 105, 249, 5, 5, 106, 249, 5, 5, 86, 250, 5, 5, 51, 158, 5, 5, 94, 159, 5, 5, 95, 159, 5, 5, 96, 159, 5, 5, 36, 161, 5, 5, 37, 161, 5, 5, 38, 161, 5, 5, 39, 161, 5, 5, 40, 161, 5, 5, 13, 164, 5, 5, 14, 164, 5, 5, 15, 164, 5, 5, 85, 168, 5, 5, 86, 168, 5, 5, 87, 168, 5, 5, 88, 168, 5, 5, 170, 173, 5, 5, 24, 180, 5, 5, 163, 187, 5, 5, 164, 187, 5, 5, 165, 187, 5, 5, 166, 187, 5, 5, 110, 195, 5, 5, 111, 195, 5, 5, 126, 203, 5, 5, 127, 203, 5, 5, 128, 203, 5, 5, 129, 203, 5, 5, 130, 203, 5, 5, 131, 203, 5, 5, 125, 203, 5, 5, 90, 211, 5, 5, 91, 211, 5, 5, 92, 211, 5, 5, 93, 211, 5, 5, 94, 211, 5, 5, 95, 211, 5, 5, 80, 218, 5, 5, 81, 218, 5, 5, 82, 218, 5, 5, 9, 225, 5, 5, 8, 225, 5, 5, 206, 230, 5, 5, 114, 235, 5, 5, 207, 230, 5, 5, 128, 246, 5, 5, 52, 158, 5, 5, 53, 158, 5, 5, 54, 158, 5, 5, 97, 159, 5, 5, 17, 164, 5, 5, 18, 164, 5, 5, 19, 164, 5, 5, 89, 168, 5, 5, 90, 168, 5, 5, 91, 168, 5, 5, 92, 168, 5, 5, 171, 173, 5, 5, 172, 173, 5, 5, 173, 173, 5, 5, 25, 180, 5, 5, 26, 180, 5, 5, 27, 180, 5, 5, 28, 180, 5, 5, 167, 187, 5, 5, 113, 195, 5, 5, 114, 195, 5, 5, 55, 158, 5, 5, 147, 157, 5, 5, 148, 157, 5, 5, 56, 158, 5, 5, 105, 159, 5, 5, 98, 159, 5, 5, 99, 159, 5, 5, 100, 159, 5, 5, 101, 159, 5, 5, 102, 159, 5, 5, 103, 159, 5, 5, 104, 159, 5, 5, 41, 161, 5, 5, 42, 161, 5, 5, 43, 161, 5, 5, 44, 161, 5, 5, 45, 161, 5, 5, 46, 161, 5, 5, 47, 161, 5, 5, 48, 161, 5, 5, 62, 161, 5, 5, 49, 161, 5, 5, 50, 161, 5, 5, 51, 161, 5, 5, 52, 161, 5, 5, 53, 161, 5, 5, 54, 161, 5, 5, 55, 161, 5, 5, 56, 161, 5, 5, 57, 161, 5, 5, 58, 161, 5, 5, 59, 161, 5, 5, 60, 161, 5, 5, 61, 161, 5, 5, 20, 164, 5, 5, 21, 164, 5, 5, 22, 164, 5, 5, 23, 164, 5, 5, 24, 164, 5, 5, 25, 164, 5, 5, 26, 164, 5, 5, 27, 164, 5, 5, 28, 164, 5, 5, 29, 164, 5, 5, 30, 164, 5, 5, 76, 164, 5, 5, 31, 164, 5, 5, 32, 164, 5, 5, 33, 164, 5, 5, 34, 164, 5, 5, 35, 164, 5, 5, 36, 164, 5, 5, 93, 168, 5, 5, 37, 164, 5, 5, 38, 164, 5, 5, 39, 164, 5, 5, 40, 164, 5, 5, 41, 164, 5, 5, 42, 164, 5, 5, 43, 164, 5, 5, 44, 164, 5, 5, 45, 164, 5, 5, 46, 164, 5, 5, 47, 164, 5, 5, 48, 164, 5, 5, 49, 164, 5, 5, 50, 164, 5, 5, 51, 164, 5, 5, 52, 164, 5, 5, 53, 164, 5, 5, 54, 164, 5, 5, 55, 164, 5, 5, 56, 164, 5, 5, 57, 164, 5, 5, 58, 164, 5, 5, 59, 164, 5, 5, 60, 164, 5, 5, 61, 164, 5, 5, 62, 164, 5, 5, 63, 164, 5, 5, 64, 164, 5, 5, 65, 164, 5, 5, 66, 164, 5, 5, 67, 164, 5, 5, 68, 164, 5, 5, 69, 164, 5, 5, 70, 164, 5, 5, 71, 164, 5, 5, 72, 164, 5, 5, 73, 164, 5, 5, 74, 164, 5, 5, 94, 168, 5, 5, 95, 168, 5, 5, 96, 168, 5, 5, 97, 168, 5, 5, 98, 168, 5, 5, 99, 168, 5, 5, 100, 168, 5, 5, 101, 168, 5, 5, 102, 168, 5, 5, 103, 168, 5, 5, 104, 168, 5, 5, 105, 168, 5, 5, 106, 168, 5, 5, 107, 168, 5, 5, 108, 168, 5, 5, 109, 168, 5, 5, 110, 168, 5, 5, 111, 168, 5, 5, 112, 168, 5, 5, 113, 168, 5, 5, 114, 168, 5, 5, 115, 168, 5, 5, 116, 168, 5, 5, 117, 168, 5, 5, 118, 168, 5, 5, 119, 168, 5, 5, 120, 168, 5, 5, 121, 168, 5, 5, 122, 168, 5, 5, 123, 168, 5, 5, 124, 168, 5, 5, 125, 168, 5, 5, 126, 168, 5, 5, 127, 168, 5, 5, 128, 168, 5, 5, 129, 168, 5, 5, 130, 168, 5, 5, 131, 168, 5, 5, 132, 168, 5, 5, 133, 168, 5, 5, 134, 168, 5, 5, 174, 173, 5, 5, 135, 168, 5, 5, 136, 168, 5, 5, 137, 168, 5, 5, 138, 168, 5, 5, 139, 168, 5, 5, 140, 168, 5, 5, 141, 168, 5, 5, 142, 168, 5, 5, 143, 168, 5, 5, 144, 168, 5, 5, 145, 168, 5, 5, 146, 168, 5, 5, 175, 173, 5, 5, 147, 168, 5, 5, 148, 168, 5, 5, 149, 168, 5, 5, 150, 168, 5, 5, 151, 168, 5, 5, 152, 168, 5, 5, 153, 168, 5, 5, 154, 168, 5, 5, 155, 168, 5, 5, 156, 168, 5, 5, 157, 168, 5, 5, 158, 168, 5, 5, 159, 168, 5, 5, 177, 173, 5, 5, 178, 173, 5, 5, 179, 173, 5, 5, 180, 173, 5, 5, 181, 173, 5, 5, 182, 173, 5, 5, 183, 173, 5, 5, 184, 173, 5, 5, 29, 180, 5, 5, 30, 180, 5, 5, 185, 173, 5, 5, 186, 173, 5, 5, 187, 173, 5, 5, 188, 173, 5, 5, 189, 173, 5, 5, 190, 173, 5, 5, 191, 173, 5, 5, 192, 173, 5, 5, 193, 173, 5, 5, 194, 173, 5, 5, 195, 173, 5, 5, 31, 180, 5, 5, 196, 173, 5, 5, 197, 173, 5, 5, 198, 173, 5, 5, 199, 173, 5, 5, 200, 173, 5, 5, 201, 173, 5, 5, 202, 173, 5, 5, 203, 173, 5, 5, 32, 180, 5, 5, 204, 173, 5, 5, 205, 173, 5, 5, 206, 173, 5, 5, 207, 173, 5, 5, 208, 173, 5, 5, 209, 173, 5, 5, 210, 173, 5, 5, 33, 180, 5, 5, 211, 173, 5, 5, 212, 173, 5, 5, 213, 173, 5, 5, 214, 173, 5, 5, 215, 173, 5, 5, 216, 173, 5, 5, 217, 173, 5, 5, 218, 173, 5, 5, 34, 180, 5, 5, 35, 180, 5, 5, 36, 180, 5, 5, 219, 173, 5, 5, 220, 173, 5, 5, 221, 173, 5, 5, 222, 173, 5, 5, 223, 173, 5, 5, 224, 173, 5, 5, 225, 173, 5, 5, 226, 173, 5, 5, 227, 173, 5, 5, 228, 173, 5, 5, 229, 173, 5, 5, 230, 173, 5, 5, 38, 180, 5, 5, 39, 180, 5, 5, 40, 180, 5, 5, 41, 180, 5, 5, 42, 180, 5, 5, 43, 180, 5, 5, 44, 180, 5, 5, 45, 180, 5, 5, 46, 180, 5, 5, 47, 180, 5, 5, 168, 187, 5, 5, 48, 180, 5, 5, 49, 180, 5, 5, 50, 180, 5, 5, 51, 180, 5, 5, 52, 180, 5, 5, 53, 180, 5, 5, 54, 180, 5, 5, 55, 180, 5, 5, 169, 187, 5, 5, 56, 180, 5, 5, 57, 180, 5, 5, 58, 180, 5, 5, 59, 180, 5, 5, 60, 180, 5, 5, 61, 180, 5, 5, 62, 180, 5, 5, 63, 180, 5, 5, 64, 180, 5, 5, 65, 180, 5, 5, 66, 180, 5, 5, 67, 180, 5, 5, 68, 180, 5, 5, 69, 180, 5, 5, 70, 180, 5, 5, 71, 180, 5, 5, 72, 180, 5, 5, 73, 180, 5, 5, 74, 180, 5, 5, 75, 180, 5, 5, 76, 180, 5, 5, 77, 180, 5, 5, 78, 180, 5, 5, 79, 180, 5, 5, 80, 180, 5, 5, 81, 180, 5, 5, 82, 180, 5, 5, 83, 180, 5, 5, 84, 180, 5, 5, 85, 180, 5, 5, 86, 180, 5, 5, 87, 180, 5, 5, 88, 180, 5, 5, 89, 180, 5, 5, 90, 180, 5, 5, 91, 180, 5, 5, 92, 180, 5, 5, 93, 180, 5, 5, 94, 180, 5, 5, 95, 180, 5, 5, 96, 180, 5, 5, 170, 187, 5, 5, 171, 187, 5, 5, 172, 187, 5, 5, 173, 187, 5, 5, 174, 187, 5, 5, 175, 187, 5, 5, 176, 187, 5, 5, 177, 187, 5, 5, 178, 187, 5, 5, 179, 187, 5, 5, 180, 187, 5, 5, 181, 187, 5, 5, 182, 187, 5, 5, 183, 187, 5, 5, 184, 187, 5, 5, 185, 187, 5, 5, 186, 187, 5, 5, 187, 187, 5, 5, 188, 187, 5, 5, 189, 187, 5, 5, 190, 187, 5, 5, 191, 187, 5, 5, 192, 187, 5, 5, 193, 187, 5, 5, 194, 187, 5, 5, 195, 187, 5, 5, 196, 187, 5, 5, 197, 187, 5, 5, 198, 187, 5, 5, 199, 187, 5, 5, 200, 187, 5, 5, 201, 187, 5, 5, 202, 187, 5, 5, 203, 187, 5, 5, 204, 187, 5, 5, 205, 187, 5, 5, 206, 187, 5, 5, 207, 187, 5, 5, 208, 187, 5, 5, 115, 195, 5, 5, 209, 187, 5, 5, 210, 187, 5, 5, 211, 187, 5, 5, 212, 187, 5, 5, 213, 187, 5, 5, 214, 187, 5, 5, 215, 187, 5, 5, 116, 195, 5, 5, 216, 187, 5, 5, 217, 187, 5, 5, 218, 187, 5, 5, 219, 187, 5, 5, 220, 187, 5, 5, 221, 187, 5, 5, 222, 187, 5, 5, 223, 187, 5, 5, 224, 187, 5, 5, 225, 187, 5, 5, 226, 187, 5, 5, 227, 187, 5, 5, 228, 187, 5, 5, 229, 187, 5, 5, 117, 195, 5, 5, 230, 187, 5, 5, 231, 187, 5, 5, 232, 187, 5, 5, 233, 187, 5, 5, 234, 187, 5, 5, 235, 187, 5, 5, 236, 187, 5, 5, 237, 187, 5, 5, 238, 187, 5, 5, 239, 187, 5, 5, 240, 187, 5, 5, 241, 187, 5, 5, 118, 195, 5, 5, 119, 195, 5, 5, 9, 188, 5, 5, 242, 187, 5, 5, 243, 187, 5, 5, 244, 187, 5, 5, 245, 187, 5, 5, 246, 187, 5, 5, 247, 187, 5, 5, 248, 187, 5, 5, 249, 187, 5, 5, 250, 187, 5, 5, 251, 187, 5, 5, 252, 187, 5, 5, 120, 195, 5, 5, 121, 195, 5, 5, 122, 195, 5, 5, 123, 195, 5, 5, 124, 195, 5, 5, 125, 195, 5, 5, 126, 195, 5, 5, 210, 203, 5, 5, 127, 195, 5, 5, 128, 195, 5, 5, 129, 195, 5, 5, 130, 195, 5, 5, 131, 195, 5, 5, 132, 195, 5, 5, 133, 195, 5, 5, 134, 195, 5, 5, 135, 195, 5, 5, 136, 195, 5, 5, 137, 195, 5, 5, 138, 195, 5, 5, 139, 195, 5, 5, 140, 195, 5, 5, 141, 195, 5, 5, 142, 195, 5, 5, 143, 195, 5, 5, 144, 195, 5, 5, 145, 195, 5, 5, 146, 195, 5, 5, 147, 195, 5, 5, 148, 195, 5, 5, 149, 195, 5, 5, 150, 195, 5, 5, 151, 195, 5, 5, 152, 195, 5, 5, 153, 195, 5, 5, 154, 195, 5, 5, 155, 195, 5, 5, 156, 195, 5, 5, 101, 180, 5, 5, 157, 195, 5, 5, 158, 195, 5, 5, 132, 203, 5, 5, 159, 195, 5, 5, 160, 195, 5, 5, 161, 195, 5, 5, 133, 203, 5, 5, 162, 195, 5, 5, 163, 195, 5, 5, 164, 195, 5, 5, 165, 195, 5, 5, 166, 195, 5, 5, 134, 203, 5, 5, 167, 195, 5, 5, 168, 195, 5, 5, 169, 195, 5, 5, 170, 195, 5, 5, 171, 195, 5, 5, 172, 195, 5, 5, 173, 195, 5, 5, 174, 195, 5, 5, 175, 195, 5, 5, 176, 195, 5, 5, 177, 195, 5, 5, 178, 195, 5, 5, 179, 195, 5, 5, 180, 195, 5, 5, 181, 195, 5, 5, 182, 195, 5, 5, 183, 195, 5, 5, 184, 195, 5, 5, 185, 195, 5, 5, 186, 195, 5, 5, 135, 203, 5, 5, 136, 203, 5, 5, 137, 203, 5, 5, 138, 203, 5, 5, 139, 203, 5, 5, 140, 203, 5, 5, 141, 203, 5, 5, 142, 203, 5, 5, 143, 203, 5, 5, 144, 203, 5, 5, 145, 203, 5, 5, 146, 203, 5, 5, 147, 203, 5, 5, 148, 203, 5, 5, 149, 203, 5, 5, 150, 203, 5, 5, 151, 203, 5, 5, 152, 203, 5, 5, 153, 203, 5, 5, 154, 203, 5, 5, 155, 203, 5, 5, 156, 203, 5, 5, 157, 203, 5, 5, 158, 203, 5, 5, 159, 203, 5, 5, 160, 203, 5, 5, 161, 203, 5, 5, 162, 203, 5, 5, 163, 203, 5, 5, 164, 203, 5, 5, 165, 203, 5, 5, 166, 203, 5, 5, 167, 203, 5, 5, 168, 203, 5, 5, 169, 203, 5, 5, 170, 203, 5, 5, 171, 203, 5, 5, 96, 211, 5, 5, 172, 203, 5, 5, 173, 203, 5, 5, 174, 203, 5, 5, 175, 203, 5, 5, 176, 203, 5, 5, 211, 203, 5, 5, 177, 203, 5, 5, 178, 203, 5, 5, 97, 211, 5, 5, 179, 203, 5, 5, 180, 203, 5, 5, 181, 203, 5, 5, 182, 203, 5, 5, 183, 203, 5, 5, 184, 203, 5, 5, 98, 211, 5, 5, 185, 203, 5, 5, 186, 203, 5, 5, 187, 203, 5, 5, 99, 211, 5, 5, 188, 203, 5, 5, 189, 203, 5, 5, 190, 203, 5, 5, 191, 203, 5, 5, 192, 203, 5, 5, 193, 203, 5, 5, 194, 203, 5, 5, 195, 203, 5, 5, 196, 203, 5, 5, 197, 203, 5, 5, 198, 203, 5, 5, 100, 211, 5, 5, 101, 211, 5, 5, 102, 211, 5, 5, 103, 211, 5, 5, 104, 211, 5, 5, 105, 211, 5, 5, 106, 211, 5, 5, 187, 195, 5, 5, 107, 211, 5, 5, 108, 211, 5, 5, 109, 211, 5, 5, 165, 211, 5, 5, 110, 211, 5, 5, 111, 211, 5, 5, 112, 211, 5, 5, 113, 211, 5, 5, 114, 211, 5, 5, 115, 211, 5, 5, 116, 211, 5, 5, 117, 211, 5, 5, 118, 211, 5, 5, 119, 211, 5, 5, 196, 195, 5, 5, 120, 211, 5, 5, 121, 211, 5, 5, 122, 211, 5, 5, 123, 211, 5, 5, 124, 211, 5, 5, 125, 211, 5, 5, 83, 218, 5, 5, 84, 218, 5, 5, 126, 211, 5, 5, 127, 211, 5, 5, 128, 211, 5, 5, 129, 211, 5, 5, 85, 218, 5, 5, 86, 218, 5, 5, 87, 218, 5, 5, 130, 211, 5, 5, 131, 211, 5, 5, 132, 211, 5, 5, 133, 211, 5, 5, 134, 211, 5, 5, 135, 211, 5, 5, 136, 211, 5, 5, 137, 211, 5, 5, 88, 218, 5, 5, 138, 211, 5, 5, 139, 211, 5, 5, 140, 211, 5, 5, 141, 211, 5, 5, 142, 211, 5, 5, 143, 211, 5, 5, 89, 218, 5, 5, 144, 211, 5, 5, 145, 211, 5, 5, 90, 218, 5, 5, 146, 211, 5, 5, 92, 218, 5, 5, 93, 218, 5, 5, 147, 211, 5, 5, 94, 218, 5, 5, 10, 225, 5, 5, 95, 218, 5, 5, 96, 218, 5, 5, 97, 218, 5, 5, 98, 218, 5, 5, 99, 218, 5, 5, 100, 218, 5, 5, 101, 218, 5, 5, 102, 218, 5, 5, 103, 218, 5, 5, 104, 218, 5, 5, 105, 218, 5, 5, 106, 218, 5, 5, 107, 218, 5, 5, 108, 218, 5, 5, 109, 218, 5, 5, 110, 218, 5, 5, 111, 218, 5, 5, 112, 218, 5, 5, 113, 218, 5, 5, 114, 218, 5, 5, 115, 218, 5, 5, 116, 218, 5, 5, 117, 218, 5, 5, 118, 218, 5, 5, 119, 218, 5, 5, 120, 218, 5, 5, 121, 218, 5, 5, 122, 218, 5, 5, 166, 211, 5, 5, 123, 218, 5, 5, 124, 218, 5, 5, 125, 218, 5, 5, 126, 218, 5, 5, 127, 218, 5, 5, 128, 218, 5, 5, 129, 218, 5, 5, 130, 218, 5, 5, 131, 218, 5, 5, 132, 218, 5, 5, 133, 218, 5, 5, 134, 218, 5, 5, 135, 218, 5, 5, 136, 218, 5, 5, 137, 218, 5, 5, 138, 218, 5, 5, 139, 218, 5, 5, 140, 218, 5, 5, 141, 218, 5, 5, 142, 218, 5, 5, 11, 225, 5, 5, 12, 225, 5, 5, 13, 225, 5, 5, 14, 225, 5, 5, 15, 225, 5, 5, 16, 225, 5, 5, 17, 225, 5, 5, 18, 225, 5, 5, 19, 225, 5, 5, 20, 225, 5, 5, 21, 225, 5, 5, 143, 218, 5, 5, 22, 225, 5, 5, 23, 225, 5, 5, 24, 225, 5, 5, 208, 230, 5, 5, 25, 225, 5, 5, 26, 225, 5, 5, 27, 225, 5, 5, 209, 230, 5, 5, 28, 225, 5, 5, 29, 225, 5, 5, 30, 225, 5, 5, 31, 225, 5, 5, 32, 225, 5, 5, 33, 225, 5, 5, 34, 225, 5, 5, 35, 225, 5, 5, 36, 225, 5, 5, 210, 230, 5, 5, 37, 225, 5, 5, 38, 225, 5, 5, 39, 225, 5, 5, 40, 225, 5, 5, 41, 225, 5, 5, 42, 225, 5, 5, 211, 230, 5, 5, 212, 230, 5, 5, 213, 230, 5, 5, 214, 230, 5, 5, 215, 230, 5, 5, 216, 230, 5, 5, 117, 235, 5, 5, 217, 230, 5, 5, 115, 235, 5, 5, 218, 230, 5, 5, 219, 230, 5, 5, 116, 235, 5, 5, 220, 230, 5, 5, 221, 230, 5, 5, 222, 230, 5, 5, 223, 230, 5, 5, 224, 230, 5, 5, 225, 230, 5, 5, 226, 230, 5, 5, 118, 235, 5, 5, 52, 225, 5, 5, 119, 235, 5, 5, 120, 235, 5, 5, 121, 235, 5, 5, 122, 235, 5, 5, 123, 235, 5, 5, 124, 235, 5, 5, 125, 235, 5, 5, 126, 235, 5, 5, 127, 235, 5, 5, 128, 235, 5, 5, 129, 235, 5, 5, 130, 235, 5, 5, 36, 239, 5, 5, 131, 235, 5, 5, 132, 235, 5, 5, 133, 235, 5, 5, 134, 235, 5, 5, 135, 235, 5, 5, 136, 235, 5, 5, 37, 239, 5, 5, 38, 239, 5, 5, 39, 239, 5, 5, 40, 239, 5, 5, 41, 239, 5, 5, 42, 239, 5, 5, 43, 239, 5, 5, 44, 239, 5, 5, 45, 239, 5, 5, 46, 239, 5, 5, 194, 244, 5, 5, 47, 239, 5, 5, 79, 242, 5, 5, 80, 242, 5, 5, 81, 242, 5, 5, 82, 242, 5, 5, 83, 242, 5, 5, 84, 242, 5, 5, 85, 242, 5, 5, 86, 242, 5, 5, 195, 244, 5, 5, 196, 244, 5, 5, 197, 244, 5, 5, 129, 246, 5, 5, 130, 246, 5, 5, 131, 246, 5, 5, 132, 246, 5, 5, 133, 246, 5, 5, 197, 247, 5, 5, 134, 246, 5, 5, 198, 247, 5, 5, 135, 246, 5, 5, 136, 246, 5, 5, 199, 247, 5, 5, 200, 247, 5, 5, 201, 247, 5, 5, 202, 247, 5, 5, 195, 248, 5, 5, 196, 248, 5, 5, 107, 249, 5, 5, 58, 158, 5, 5, 63, 161, 5, 5, 239, 173, 5, 5, 197, 195, 5, 5, 53, 225, 5, 5, 59, 158, 5, 5, 60, 158, 5, 5, 64, 161, 5, 5, 65, 161, 5, 5, 77, 164, 5, 5, 78, 164, 5, 5, 79, 164, 5, 5, 80, 164, 5, 5, 81, 164, 5, 5, 170, 168, 5, 5, 171, 168, 5, 5, 240, 173, 5, 5, 241, 173, 5, 5, 242, 173, 5, 5, 243, 173, 5, 5, 244, 173, 5, 5, 245, 173, 5, 5, 246, 173, 5, 5, 102, 180, 5, 5, 103, 180, 5, 5, 104, 180, 5, 5, 105, 180, 5, 5, 106, 180, 5, 5, 107, 180, 5, 5, 108, 180, 5, 5, 11, 188, 5, 5, 12, 188, 5, 5, 13, 188, 5, 5, 14, 188, 5, 5, 15, 188, 5, 5, 16, 188, 5, 5, 17, 188, 5, 5, 18, 188, 5, 5, 19, 188, 5, 5, 20, 188, 5, 5, 21, 188, 5, 5, 22, 188, 5, 5, 23, 188, 5, 5, 24, 188, 5, 5, 25, 188, 5, 5, 198, 195, 5, 5, 31, 188, 5, 5, 199, 195, 5, 5, 200, 195, 5, 5, 201, 195, 5, 5, 202, 195, 5, 5, 203, 195, 5, 5, 204, 195, 5, 5, 205, 195, 5, 5, 206, 195, 5, 5, 207, 195, 5, 5, 208, 195, 5, 5, 209, 195, 5, 5, 210, 195, 5, 5, 211, 195, 5, 5, 213, 203, 5, 5, 214, 203, 5, 5, 212, 203, 5, 5, 215, 203, 5, 5, 216, 203, 5, 5, 217, 203, 5, 5, 167, 211, 5, 5, 168, 211, 5, 5, 169, 211, 5, 5, 54, 225, 5, 5, 148, 218, 5, 5, 149, 218, 5, 5, 150, 218, 5, 5, 151, 218, 5, 5, 152, 218, 5, 5, 153, 218, 5, 5, 154, 218, 5, 5, 55, 225, 5, 5, 56, 225, 5, 5, 57, 225, 5, 5, 58, 225, 5, 5, 230, 230, 5, 5, 231, 230, 5, 5, 232, 230, 5, 5, 233, 230, 5, 5, 50, 239, 5, 5, 88, 242, 5, 5, 89, 242, 5, 5, 61, 158, 5, 5, 83, 164, 5, 5, 173, 168, 5, 5, 113, 180, 5, 5, 114, 180, 5, 5, 216, 195, 5, 5, 32, 188, 5, 5, 33, 188, 5, 5, 34, 188, 5, 5, 217, 195, 5, 5, 218, 195, 5, 5, 218, 203, 5, 5, 60, 225, 5, 5, 144, 235, 5, 5, 199, 244, 5, 5, 204, 247, 5, 5, 62, 158, 5, 5, 84, 164, 5, 5, 115, 180, 5, 5, 116, 180, 5, 5, 35, 188, 5, 5, 36, 188, 5, 5, 219, 195, 5, 5, 223, 195, 5, 5, 220, 203, 5, 5, 170, 211, 5, 5, 171, 211, 5, 5, 61, 225, 5, 5, 234, 230, 5, 5, 63, 158, 5, 5, 106, 159, 5, 5, 175, 168, 5, 5, 176, 168, 5, 5, 177, 168, 5, 5, 178, 168, 5, 5, 250, 173, 5, 5, 251, 173, 5, 5, 37, 188, 5, 5, 38, 188, 5, 5, 224, 195, 5, 5, 225, 195, 5, 5, 221, 203, 5, 5, 227, 195, 5, 5, 172, 211, 5, 5, 156, 218, 5, 5, 62, 225, 5, 5, 235, 230, 5, 5, 236, 230, 5, 5, 146, 235, 5, 5, 108, 249, 5, 5, 64, 158, 5, 5, 179, 168, 5, 5, 180, 168, 5, 5, 181, 168, 5, 5, 254, 173, 5, 5, 255, 173, 5, 5, 2, 174, 5, 5, 3, 174, 5, 5, 117, 180, 5, 5, 118, 180, 5, 5, 119, 180, 5, 5, 120, 180, 5, 5, 121, 180, 5, 5, 122, 180, 5, 5, 39, 188, 5, 5, 40, 188, 5, 5, 41, 188, 5, 5, 123, 180, 5, 5, 42, 188, 5, 5, 43, 188, 5, 5, 44, 188, 5, 5, 45, 188, 5, 5, 46, 188, 5, 5, 228, 195, 5, 5, 229, 195, 5, 5, 222, 203, 5, 5, 223, 203, 5, 5, 224, 203, 5, 5, 225, 203, 5, 5, 174, 211, 5, 5, 175, 211, 5, 5, 63, 225, 5, 5, 64, 225, 5, 5, 237, 230, 5, 5, 147, 235, 5, 5, 51, 239, 5, 5, 52, 239, 5, 5, 53, 239, 5, 5, 90, 242, 5, 5, 65, 158, 5, 5, 66, 158, 5, 5, 5, 174, 5, 5, 48, 188, 5, 5, 226, 203, 5, 5, 68, 158, 5, 5, 107, 159, 5, 5, 108, 159, 5, 5, 66, 161, 5, 5, 67, 161, 5, 5, 68, 161, 5, 5, 69, 161, 5, 5, 70, 161, 5, 5, 71, 161, 5, 5, 72, 161, 5, 5, 73, 161, 5, 5, 85, 164, 5, 5, 86, 164, 5, 5, 87, 164, 5, 5, 88, 164, 5, 5, 89, 164, 5, 5, 90, 164, 5, 5, 91, 164, 5, 5, 92, 164, 5, 5, 93, 164, 5, 5, 183, 168, 5, 5, 184, 168, 5, 5, 185, 168, 5, 5, 186, 168, 5, 5, 187, 168, 5, 5, 188, 168, 5, 5, 189, 168, 5, 5, 190, 168, 5, 5, 191, 168, 5, 5, 192, 168, 5, 5, 193, 168, 5, 5, 194, 168, 5, 5, 195, 168, 5, 5, 196, 168, 5, 5, 197, 168, 5, 5, 198, 168, 5, 5, 199, 168, 5, 5, 200, 168, 5, 5, 201, 168, 5, 5, 202, 168, 5, 5, 203, 168, 5, 5, 204, 168, 5, 5, 205, 168, 5, 5, 206, 168, 5, 5, 207, 168, 5, 5, 208, 168, 5, 5, 209, 168, 5, 5, 210, 168, 5, 5, 211, 168, 5, 5, 212, 168, 5, 5, 213, 168, 5, 5, 214, 168, 5, 5, 215, 168, 5, 5, 6, 174, 5, 5, 7, 174, 5, 5, 8, 174, 5, 5, 9, 174, 5, 5, 10, 174, 5, 5, 11, 174, 5, 5, 12, 174, 5, 5, 13, 174, 5, 5, 14, 174, 5, 5, 15, 174, 5, 5, 16, 174, 5, 5, 17, 174, 5, 5, 18, 174, 5, 5, 19, 174, 5, 5, 20, 174, 5, 5, 21, 174, 5, 5, 22, 174, 5, 5, 23, 174, 5, 5, 24, 174, 5, 5, 25, 174, 5, 5, 26, 174, 5, 5, 27, 174, 5, 5, 28, 174, 5, 5, 29, 174, 5, 5, 30, 174, 5, 5, 31, 174, 5, 5, 32, 174, 5, 5, 33, 174, 5, 5, 34, 174, 5, 5, 35, 174, 5, 5, 36, 174, 5, 5, 37, 174, 5, 5, 38, 174, 5, 5, 39, 174, 5, 5, 40, 174, 5, 5, 41, 174, 5, 5, 42, 174, 5, 5, 43, 174, 5, 5, 125, 180, 5, 5, 126, 180, 5, 5, 127, 180, 5, 5, 128, 180, 5, 5, 129, 180, 5, 5, 130, 180, 5, 5, 131, 180, 5, 5, 132, 180, 5, 5, 133, 180, 5, 5, 134, 180, 5, 5, 135, 180, 5, 5, 136, 180, 5, 5, 137, 180, 5, 5, 138, 180, 5, 5, 139, 180, 5, 5, 140, 180, 5, 5, 141, 180, 5, 5, 142, 180, 5, 5, 143, 180, 5, 5, 144, 180, 5, 5, 145, 180, 5, 5, 146, 180, 5, 5, 147, 180, 5, 5, 50, 188, 5, 5, 51, 188, 5, 5, 52, 188, 5, 5, 53, 188, 5, 5, 54, 188, 5, 5, 55, 188, 5, 5, 56, 188, 5, 5, 57, 188, 5, 5, 150, 180, 5, 5, 151, 180, 5, 5, 58, 188, 5, 5, 59, 188, 5, 5, 60, 188, 5, 5, 61, 188, 5, 5, 62, 188, 5, 5, 63, 188, 5, 5, 64, 188, 5, 5, 65, 188, 5, 5, 66, 188, 5, 5, 231, 195, 5, 5, 232, 195, 5, 5, 233, 195, 5, 5, 234, 195, 5, 5, 235, 195, 5, 5, 236, 195, 5, 5, 237, 195, 5, 5, 238, 195, 5, 5, 239, 195, 5, 5, 240, 195, 5, 5, 241, 195, 5, 5, 242, 195, 5, 5, 243, 195, 5, 5, 244, 195, 5, 5, 227, 203, 5, 5, 245, 195, 5, 5, 246, 195, 5, 5, 247, 195, 5, 5, 248, 195, 5, 5, 249, 195, 5, 5, 250, 195, 5, 5, 251, 195, 5, 5, 252, 195, 5, 5, 253, 195, 5, 5, 254, 195, 5, 5, 255, 195, 5, 5, 228, 203, 5, 5, 229, 203, 5, 5, 230, 203, 5, 5, 231, 203, 5, 5, 232, 203, 5, 5, 233, 203, 5, 5, 234, 203, 5, 5, 235, 203, 5, 5, 236, 203, 5, 5, 237, 203, 5, 5, 238, 203, 5, 5, 239, 203, 5, 5, 240, 203, 5, 5, 2, 196, 5, 5, 241, 203, 5, 5, 242, 203, 5, 5, 243, 203, 5, 5, 244, 203, 5, 5, 245, 203, 5, 5, 246, 203, 5, 5, 247, 203, 5, 5, 248, 203, 5, 5, 176, 211, 5, 5, 177, 211, 5, 5, 178, 211, 5, 5, 179, 211, 5, 5, 180, 211, 5, 5, 181, 211, 5, 5, 182, 211, 5, 5, 183, 211, 5, 5, 184, 211, 5, 5, 185, 211, 5, 5, 186, 211, 5, 5, 187, 211, 5, 5, 188, 211, 5, 5, 189, 211, 5, 5, 190, 211, 5, 5, 158, 218, 5, 5, 159, 218, 5, 5, 160, 218, 5, 5, 161, 218, 5, 5, 162, 218, 5, 5, 163, 218, 5, 5, 164, 218, 5, 5, 165, 218, 5, 5, 166, 218, 5, 5, 167, 218, 5, 5, 168, 218, 5, 5, 169, 218, 5, 5, 170, 218, 5, 5, 171, 218, 5, 5, 172, 218, 5, 5, 65, 225, 5, 5, 66, 225, 5, 5, 67, 225, 5, 5, 68, 225, 5, 5, 181, 218, 5, 5, 69, 225, 5, 5, 70, 225, 5, 5, 71, 225, 5, 5, 72, 225, 5, 5, 73, 225, 5, 5, 74, 225, 5, 5, 75, 225, 5, 5, 76, 225, 5, 5, 77, 225, 5, 5, 78, 225, 5, 5, 79, 225, 5, 5, 80, 225, 5, 5, 81, 225, 5, 5, 82, 225, 5, 5, 83, 225, 5, 5, 84, 225, 5, 5, 85, 225, 5, 5, 238, 230, 5, 5, 100, 225, 5, 5, 239, 230, 5, 5, 240, 230, 5, 5, 241, 230, 5, 5, 242, 230, 5, 5, 243, 230, 5, 5, 244, 230, 5, 5, 245, 230, 5, 5, 246, 230, 5, 5, 148, 235, 5, 5, 252, 230, 5, 5, 247, 230, 5, 5, 149, 235, 5, 5, 150, 235, 5, 5, 54, 239, 5, 5, 55, 239, 5, 5, 56, 239, 5, 5, 57, 239, 5, 5, 58, 239, 5, 5, 59, 239, 5, 5, 91, 242, 5, 5, 92, 242, 5, 5, 93, 242, 5, 5, 94, 242, 5, 5, 95, 242, 5, 5, 96, 242, 5, 5, 200, 244, 5, 5, 205, 247, 5, 5, 206, 247, 5, 5, 207, 247, 5, 5, 197, 248, 5, 5, 198, 248, 5, 5, 109, 249, 5, 5, 69, 158, 5, 5, 110, 159, 5, 5, 74, 161, 5, 5, 75, 161, 5, 5, 94, 164, 5, 5, 95, 164, 5, 5, 216, 168, 5, 5, 48, 174, 5, 5, 152, 180, 5, 5, 73, 188, 5, 5, 153, 180, 5, 5, 154, 180, 5, 5, 74, 188, 5, 5, 67, 188, 5, 5, 11, 196, 5, 5, 12, 196, 5, 5, 165, 193, 5, 5, 13, 196, 5, 5, 14, 196, 5, 5, 13, 204, 5, 5, 198, 211, 5, 5, 199, 211, 5, 5, 101, 225, 5, 5, 202, 244, 5, 5, 71, 158, 5, 5, 77, 161, 5, 5, 217, 168, 5, 5, 218, 168, 5, 5, 219, 168, 5, 5, 220, 168, 5, 5, 49, 174, 5, 5, 50, 174, 5, 5, 51, 174, 5, 5, 128, 176, 5, 5, 155, 180, 5, 5, 156, 180, 5, 5, 157, 180, 5, 5, 158, 180, 5, 5, 75, 188, 5, 5, 161, 180, 5, 5, 76, 188, 5, 5, 77, 188, 5, 5, 78, 188, 5, 5, 79, 188, 5, 5, 15, 196, 5, 5, 16, 196, 5, 5, 17, 196, 5, 5, 18, 196, 5, 5, 14, 204, 5, 5, 220, 206, 5, 5, 200, 211, 5, 5, 103, 225, 5, 5, 104, 225, 5, 5, 105, 227, 5, 5, 159, 235, 5, 5, 99, 242, 5, 5, 72, 158, 5, 5, 73, 158, 5, 5, 111, 159, 5, 5, 112, 159, 5, 5, 113, 159, 5, 5, 114, 159, 5, 5, 115, 159, 5, 5, 116, 159, 5, 5, 117, 159, 5, 5, 78, 161, 5, 5, 79, 161, 5, 5, 80, 161, 5, 5, 81, 161, 5, 5, 82, 161, 5, 5, 83, 161, 5, 5, 84, 161, 5, 5, 85, 161, 5, 5, 86, 161, 5, 5, 87, 161, 5, 5, 88, 161, 5, 5, 89, 161, 5, 5, 90, 161, 5, 5, 91, 161, 5, 5, 92, 161, 5, 5, 93, 161, 5, 5, 94, 161, 5, 5, 95, 161, 5, 5, 96, 161, 5, 5, 96, 164, 5, 5, 97, 164, 5, 5, 98, 164, 5, 5, 99, 164, 5, 5, 100, 164, 5, 5, 101, 164, 5, 5, 102, 164, 5, 5, 103, 164, 5, 5, 104, 164, 5, 5, 105, 164, 5, 5, 106, 164, 5, 5, 107, 164, 5, 5, 108, 164, 5, 5, 109, 164, 5, 5, 110, 164, 5, 5, 111, 164, 5, 5, 112, 164, 5, 5, 113, 164, 5, 5, 114, 164, 5, 5, 115, 164, 5, 5, 116, 164, 5, 5, 117, 164, 5, 5, 118, 164, 5, 5, 119, 164, 5, 5, 120, 164, 5, 5, 121, 164, 5, 5, 122, 164, 5, 5, 123, 164, 5, 5, 124, 164, 5, 5, 125, 164, 5, 5, 126, 164, 5, 5, 127, 164, 5, 5, 128, 164, 5, 5, 129, 164, 5, 5, 130, 164, 5, 5, 131, 164, 5, 5, 132, 164, 5, 5, 133, 164, 5, 5, 222, 168, 5, 5, 223, 168, 5, 5, 224, 168, 5, 5, 225, 168, 5, 5, 226, 168, 5, 5, 227, 168, 5, 5, 228, 168, 5, 5, 229, 168, 5, 5, 230, 168, 5, 5, 231, 168, 5, 5, 232, 168, 5, 5, 233, 168, 5, 5, 234, 168, 5, 5, 235, 168, 5, 5, 236, 168, 5, 5, 237, 168, 5, 5, 238, 168, 5, 5, 239, 168, 5, 5, 240, 168, 5, 5, 241, 168, 5, 5, 242, 168, 5, 5, 243, 168, 5, 5, 244, 168, 5, 5, 134, 164, 5, 5, 245, 168, 5, 5, 246, 168, 5, 5, 247, 168, 5, 5, 248, 168, 5, 5, 249, 168, 5, 5, 250, 168, 5, 5, 251, 168, 5, 5, 252, 168, 5, 5, 253, 168, 5, 5, 254, 168, 5, 5, 255, 168, 5, 5, 2, 169, 5, 5, 3, 169, 5, 5, 4, 169, 5, 5, 5, 169, 5, 5, 6, 169, 5, 5, 7, 169, 5, 5, 8, 169, 5, 5, 9, 169, 5, 5, 10, 169, 5, 5, 11, 169, 5, 5, 12, 169, 5, 5, 13, 169, 5, 5, 14, 169, 5, 5, 15, 169, 5, 5, 16, 169, 5, 5, 17, 169, 5, 5, 18, 169, 5, 5, 19, 169, 5, 5, 20, 169, 5, 5, 21, 169, 5, 5, 22, 169, 5, 5, 23, 169, 5, 5, 24, 169, 5, 5, 25, 169, 5, 5, 26, 169, 5, 5, 27, 169, 5, 5, 28, 169, 5, 5, 29, 169, 5, 5, 30, 169, 5, 5, 31, 169, 5, 5, 32, 169, 5, 5, 33, 169, 5, 5, 34, 169, 5, 5, 52, 174, 5, 5, 53, 174, 5, 5, 54, 174, 5, 5, 55, 174, 5, 5, 56, 174, 5, 5, 57, 174, 5, 5, 58, 174, 5, 5, 59, 174, 5, 5, 60, 174, 5, 5, 61, 174, 5, 5, 62, 174, 5, 5, 63, 174, 5, 5, 64, 174, 5, 5, 65, 174, 5, 5, 66, 174, 5, 5, 162, 180, 5, 5, 67, 174, 5, 5, 68, 174, 5, 5, 69, 174, 5, 5, 70, 174, 5, 5, 71, 174, 5, 5, 72, 174, 5, 5, 73, 174, 5, 5, 74, 174, 5, 5, 75, 174, 5, 5, 76, 174, 5, 5, 77, 174, 5, 5, 78, 174, 5, 5, 79, 174, 5, 5, 80, 174, 5, 5, 81, 174, 5, 5, 82, 174, 5, 5, 83, 174, 5, 5, 84, 174, 5, 5, 85, 174, 5, 5, 86, 174, 5, 5, 87, 174, 5, 5, 88, 174, 5, 5, 89, 174, 5, 5, 90, 174, 5, 5, 91, 174, 5, 5, 92, 174, 5, 5, 93, 174, 5, 5, 94, 174, 5, 5, 95, 174, 5, 5, 96, 174, 5, 5, 97, 174, 5, 5, 98, 174, 5, 5, 99, 174, 5, 5, 100, 174, 5, 5, 101, 174, 5, 5, 163, 180, 5, 5, 102, 174, 5, 5, 103, 174, 5, 5, 104, 174, 5, 5, 105, 174, 5, 5, 106, 174, 5, 5, 107, 174, 5, 5, 108, 174, 5, 5, 109, 174, 5, 5, 110, 174, 5, 5, 111, 174, 5, 5, 112, 174, 5, 5, 113, 174, 5, 5, 114, 174, 5, 5, 115, 174, 5, 5, 116, 174, 5, 5, 117, 174, 5, 5, 118, 174, 5, 5, 119, 174, 5, 5, 164, 180, 5, 5, 120, 174, 5, 5, 121, 174, 5, 5, 122, 174, 5, 5, 123, 174, 5, 5, 53, 169, 5, 5, 124, 174, 5, 5, 125, 174, 5, 5, 126, 174, 5, 5, 127, 174, 5, 5, 128, 174, 5, 5, 129, 174, 5, 5, 130, 174, 5, 5, 131, 174, 5, 5, 132, 174, 5, 5, 133, 174, 5, 5, 134, 174, 5, 5, 135, 174, 5, 5, 136, 174, 5, 5, 137, 174, 5, 5, 138, 174, 5, 5, 139, 174, 5, 5, 140, 174, 5, 5, 141, 174, 5, 5, 142, 174, 5, 5, 143, 174, 5, 5, 144, 174, 5, 5, 145, 174, 5, 5, 146, 174, 5, 5, 147, 174, 5, 5, 165, 180, 5, 5, 166, 180, 5, 5, 167, 180, 5, 5, 168, 180, 5, 5, 169, 180, 5, 5, 170, 180, 5, 5, 171, 180, 5, 5, 172, 180, 5, 5, 173, 180, 5, 5, 174, 180, 5, 5, 175, 180, 5, 5, 176, 180, 5, 5, 177, 180, 5, 5, 178, 180, 5, 5, 179, 180, 5, 5, 180, 180, 5, 5, 181, 180, 5, 5, 182, 180, 5, 5, 183, 180, 5, 5, 184, 180, 5, 5, 185, 180, 5, 5, 186, 180, 5, 5, 187, 180, 5, 5, 188, 180, 5, 5, 189, 180, 5, 5, 190, 180, 5, 5, 191, 180, 5, 5, 192, 180, 5, 5, 193, 180, 5, 5, 194, 180, 5, 5, 195, 180, 5, 5, 196, 180, 5, 5, 197, 180, 5, 5, 198, 180, 5, 5, 199, 180, 5, 5, 200, 180, 5, 5, 201, 180, 5, 5, 202, 180, 5, 5, 203, 180, 5, 5, 204, 180, 5, 5, 205, 180, 5, 5, 206, 180, 5, 5, 207, 180, 5, 5, 208, 180, 5, 5, 209, 180, 5, 5, 210, 180, 5, 5, 211, 180, 5, 5, 212, 180, 5, 5, 213, 180, 5, 5, 214, 180, 5, 5, 215, 180, 5, 5, 216, 180, 5, 5, 217, 180, 5, 5, 218, 180, 5, 5, 219, 180, 5, 5, 220, 180, 5, 5, 221, 180, 5, 5, 222, 180, 5, 5, 223, 180, 5, 5, 224, 180, 5, 5, 225, 180, 5, 5, 226, 180, 5, 5, 227, 180, 5, 5, 228, 180, 5, 5, 155, 174, 5, 5, 229, 180, 5, 5, 230, 180, 5, 5, 231, 180, 5, 5, 232, 180, 5, 5, 233, 180, 5, 5, 234, 180, 5, 5, 235, 180, 5, 5, 236, 180, 5, 5, 237, 180, 5, 5, 238, 180, 5, 5, 239, 180, 5, 5, 240, 180, 5, 5, 241, 180, 5, 5, 242, 180, 5, 5, 243, 180, 5, 5, 244, 180, 5, 5, 245, 180, 5, 5, 246, 180, 5, 5, 247, 180, 5, 5, 248, 180, 5, 5, 249, 180, 5, 5, 250, 180, 5, 5, 251, 180, 5, 5, 252, 180, 5, 5, 82, 188, 5, 5, 83, 188, 5, 5, 84, 188, 5, 5, 85, 188, 5, 5, 86, 188, 5, 5, 87, 188, 5, 5, 88, 188, 5, 5, 89, 188, 5, 5, 90, 188, 5, 5, 91, 188, 5, 5, 92, 188, 5, 5, 93, 188, 5, 5, 94, 188, 5, 5, 95, 188, 5, 5, 96, 188, 5, 5, 97, 188, 5, 5, 98, 188, 5, 5, 99, 188, 5, 5, 100, 188, 5, 5, 101, 188, 5, 5, 102, 188, 5, 5, 103, 188, 5, 5, 104, 188, 5, 5, 105, 188, 5, 5, 106, 188, 5, 5, 107, 188, 5, 5, 108, 188, 5, 5, 109, 188, 5, 5, 110, 188, 5, 5, 111, 188, 5, 5, 112, 188, 5, 5, 113, 188, 5, 5, 114, 188, 5, 5, 115, 188, 5, 5, 116, 188, 5, 5, 117, 188, 5, 5, 118, 188, 5, 5, 119, 188, 5, 5, 120, 188, 5, 5, 121, 188, 5, 5, 122, 188, 5, 5, 123, 188, 5, 5, 124, 188, 5, 5, 125, 188, 5, 5, 126, 188, 5, 5, 127, 188, 5, 5, 128, 188, 5, 5, 129, 188, 5, 5, 130, 188, 5, 5, 131, 188, 5, 5, 132, 188, 5, 5, 133, 188, 5, 5, 134, 188, 5, 5, 135, 188, 5, 5, 136, 188, 5, 5, 137, 188, 5, 5, 138, 188, 5, 5, 139, 188, 5, 5, 140, 188, 5, 5, 141, 188, 5, 5, 142, 188, 5, 5, 143, 188, 5, 5, 144, 188, 5, 5, 145, 188, 5, 5, 146, 188, 5, 5, 147, 188, 5, 5, 148, 188, 5, 5, 149, 188, 5, 5, 150, 188, 5, 5, 151, 188, 5, 5, 152, 188, 5, 5, 153, 188, 5, 5, 154, 188, 5, 5, 20, 196, 5, 5, 155, 188, 5, 5, 156, 188, 5, 5, 157, 188, 5, 5, 158, 188, 5, 5, 159, 188, 5, 5, 160, 188, 5, 5, 161, 188, 5, 5, 162, 188, 5, 5, 163, 188, 5, 5, 164, 188, 5, 5, 165, 188, 5, 5, 166, 188, 5, 5, 167, 188, 5, 5, 168, 188, 5, 5, 21, 196, 5, 5, 22, 196, 5, 5, 23, 196, 5, 5, 24, 196, 5, 5, 25, 196, 5, 5, 26, 196, 5, 5, 27, 196, 5, 5, 28, 196, 5, 5, 29, 196, 5, 5, 30, 196, 5, 5, 31, 196, 5, 5, 32, 196, 5, 5, 33, 196, 5, 5, 34, 196, 5, 5, 35, 196, 5, 5, 36, 196, 5, 5, 37, 196, 5, 5, 38, 196, 5, 5, 39, 196, 5, 5, 40, 196, 5, 5, 41, 196, 5, 5, 42, 196, 5, 5, 43, 196, 5, 5, 44, 196, 5, 5, 45, 196, 5, 5, 46, 196, 5, 5, 47, 196, 5, 5, 48, 196, 5, 5, 49, 196, 5, 5, 50, 196, 5, 5, 51, 196, 5, 5, 52, 196, 5, 5, 53, 196, 5, 5, 54, 196, 5, 5, 55, 196, 5, 5, 56, 196, 5, 5, 57, 196, 5, 5, 58, 196, 5, 5, 59, 196, 5, 5, 60, 196, 5, 5, 61, 196, 5, 5, 62, 196, 5, 5, 63, 196, 5, 5, 64, 196, 5, 5, 65, 196, 5, 5, 66, 196, 5, 5, 67, 196, 5, 5, 68, 196, 5, 5, 69, 196, 5, 5, 70, 196, 5, 5, 71, 196, 5, 5, 72, 196, 5, 5, 73, 196, 5, 5, 74, 196, 5, 5, 75, 196, 5, 5, 76, 196, 5, 5, 77, 196, 5, 5, 78, 196, 5, 5, 79, 196, 5, 5, 80, 196, 5, 5, 81, 196, 5, 5, 82, 196, 5, 5, 83, 196, 5, 5, 84, 196, 5, 5, 85, 196, 5, 5, 86, 196, 5, 5, 87, 196, 5, 5, 88, 196, 5, 5, 89, 196, 5, 5, 90, 196, 5, 5, 91, 196, 5, 5, 92, 196, 5, 5, 93, 196, 5, 5, 94, 196, 5, 5, 95, 196, 5, 5, 96, 196, 5, 5, 97, 196, 5, 5, 98, 196, 5, 5, 99, 196, 5, 5, 100, 196, 5, 5, 101, 196, 5, 5, 102, 196, 5, 5, 103, 196, 5, 5, 104, 196, 5, 5, 105, 196, 5, 5, 106, 196, 5, 5, 107, 196, 5, 5, 108, 196, 5, 5, 109, 196, 5, 5, 110, 196, 5, 5, 111, 196, 5, 5, 112, 196, 5, 5, 113, 196, 5, 5, 114, 196, 5, 5, 115, 196, 5, 5, 116, 196, 5, 5, 117, 196, 5, 5, 118, 196, 5, 5, 119, 196, 5, 5, 120, 196, 5, 5, 121, 196, 5, 5, 122, 196, 5, 5, 123, 196, 5, 5, 124, 196, 5, 5, 125, 196, 5, 5, 126, 196, 5, 5, 127, 196, 5, 5, 128, 196, 5, 5, 15, 204, 5, 5, 16, 204, 5, 5, 17, 204, 5, 5, 18, 204, 5, 5, 19, 204, 5, 5, 20, 204, 5, 5, 21, 204, 5, 5, 22, 204, 5, 5, 23, 204, 5, 5, 24, 204, 5, 5, 25, 204, 5, 5, 26, 204, 5, 5, 27, 204, 5, 5, 28, 204, 5, 5, 29, 204, 5, 5, 30, 204, 5, 5, 31, 204, 5, 5, 32, 204, 5, 5, 33, 204, 5, 5, 34, 204, 5, 5, 35, 204, 5, 5, 36, 204, 5, 5, 37, 204, 5, 5, 38, 204, 5, 5, 39, 204, 5, 5, 40, 204, 5, 5, 41, 204, 5, 5, 42, 204, 5, 5, 43, 204, 5, 5, 44, 204, 5, 5, 45, 204, 5, 5, 46, 204, 5, 5, 47, 204, 5, 5, 48, 204, 5, 5, 49, 204, 5, 5, 50, 204, 5, 5, 51, 204, 5, 5, 52, 204, 5, 5, 53, 204, 5, 5, 184, 188, 5, 5, 54, 204, 5, 5, 55, 204, 5, 5, 56, 204, 5, 5, 57, 204, 5, 5, 58, 204, 5, 5, 59, 204, 5, 5, 60, 204, 5, 5, 61, 204, 5, 5, 62, 204, 5, 5, 63, 204, 5, 5, 64, 204, 5, 5, 65, 204, 5, 5, 66, 204, 5, 5, 67, 204, 5, 5, 68, 204, 5, 5, 69, 204, 5, 5, 70, 204, 5, 5, 71, 204, 5, 5, 72, 204, 5, 5, 73, 204, 5, 5, 74, 204, 5, 5, 75, 204, 5, 5, 76, 204, 5, 5, 142, 196, 5, 5, 77, 204, 5, 5, 143, 196, 5, 5, 78, 204, 5, 5, 79, 204, 5, 5, 80, 204, 5, 5, 81, 204, 5, 5, 82, 204, 5, 5, 83, 204, 5, 5, 84, 204, 5, 5, 85, 204, 5, 5, 86, 204, 5, 5, 87, 204, 5, 5, 88, 204, 5, 5, 89, 204, 5, 5, 90, 204, 5, 5, 91, 204, 5, 5, 92, 204, 5, 5, 93, 204, 5, 5, 94, 204, 5, 5, 95, 204, 5, 5, 96, 204, 5, 5, 97, 204, 5, 5, 98, 204, 5, 5, 99, 204, 5, 5, 100, 204, 5, 5, 101, 204, 5, 5, 102, 204, 5, 5, 203, 211, 5, 5, 103, 204, 5, 5, 104, 204, 5, 5, 204, 211, 5, 5, 205, 211, 5, 5, 206, 211, 5, 5, 207, 211, 5, 5, 208, 211, 5, 5, 209, 211, 5, 5, 210, 211, 5, 5, 105, 204, 5, 5, 211, 211, 5, 5, 212, 211, 5, 5, 213, 211, 5, 5, 106, 204, 5, 5, 214, 211, 5, 5, 215, 211, 5, 5, 216, 211, 5, 5, 217, 211, 5, 5, 218, 211, 5, 5, 219, 211, 5, 5, 220, 211, 5, 5, 221, 211, 5, 5, 222, 211, 5, 5, 223, 211, 5, 5, 224, 211, 5, 5, 225, 211, 5, 5, 226, 211, 5, 5, 227, 211, 5, 5, 228, 211, 5, 5, 229, 211, 5, 5, 230, 211, 5, 5, 231, 211, 5, 5, 232, 211, 5, 5, 233, 211, 5, 5, 234, 211, 5, 5, 235, 211, 5, 5, 236, 211, 5, 5, 237, 211, 5, 5, 238, 211, 5, 5, 239, 211, 5, 5, 240, 211, 5, 5, 241, 211, 5, 5, 242, 211, 5, 5, 243, 211, 5, 5, 244, 211, 5, 5, 245, 211, 5, 5, 246, 211, 5, 5, 247, 211, 5, 5, 248, 211, 5, 5, 249, 211, 5, 5, 250, 211, 5, 5, 251, 211, 5, 5, 252, 211, 5, 5, 253, 211, 5, 5, 254, 211, 5, 5, 255, 211, 5, 5, 2, 212, 5, 5, 3, 212, 5, 5, 4, 212, 5, 5, 5, 212, 5, 5, 6, 212, 5, 5, 7, 212, 5, 5, 8, 212, 5, 5, 9, 212, 5, 5, 10, 212, 5, 5, 11, 212, 5, 5, 12, 212, 5, 5, 13, 212, 5, 5, 14, 212, 5, 5, 15, 212, 5, 5, 16, 212, 5, 5, 17, 212, 5, 5, 18, 212, 5, 5, 19, 212, 5, 5, 20, 212, 5, 5, 21, 212, 5, 5, 22, 212, 5, 5, 23, 212, 5, 5, 24, 212, 5, 5, 25, 212, 5, 5, 26, 212, 5, 5, 27, 212, 5, 5, 28, 212, 5, 5, 29, 212, 5, 5, 30, 212, 5, 5, 31, 212, 5, 5, 32, 212, 5, 5, 187, 218, 5, 5, 188, 218, 5, 5, 189, 218, 5, 5, 190, 218, 5, 5, 191, 218, 5, 5, 192, 218, 5, 5, 193, 218, 5, 5, 186, 218, 5, 5, 194, 218, 5, 5, 195, 218, 5, 5, 196, 218, 5, 5, 197, 218, 5, 5, 198, 218, 5, 5, 199, 218, 5, 5, 200, 218, 5, 5, 201, 218, 5, 5, 202, 218, 5, 5, 203, 218, 5, 5, 204, 218, 5, 5, 205, 218, 5, 5, 206, 218, 5, 5, 207, 218, 5, 5, 208, 218, 5, 5, 209, 218, 5, 5, 210, 218, 5, 5, 211, 218, 5, 5, 212, 218, 5, 5, 213, 218, 5, 5, 214, 218, 5, 5, 215, 218, 5, 5, 216, 218, 5, 5, 217, 218, 5, 5, 218, 218, 5, 5, 42, 212, 5, 5, 219, 218, 5, 5, 220, 218, 5, 5, 221, 218, 5, 5, 222, 218, 5, 5, 223, 218, 5, 5, 224, 218, 5, 5, 225, 218, 5, 5, 226, 218, 5, 5, 227, 218, 5, 5, 228, 218, 5, 5, 229, 218, 5, 5, 230, 218, 5, 5, 231, 218, 5, 5, 232, 218, 5, 5, 233, 218, 5, 5, 234, 218, 5, 5, 235, 218, 5, 5, 236, 218, 5, 5, 237, 218, 5, 5, 238, 218, 5, 5, 239, 218, 5, 5, 240, 218, 5, 5, 241, 218, 5, 5, 242, 218, 5, 5, 243, 218, 5, 5, 244, 218, 5, 5, 245, 218, 5, 5, 246, 218, 5, 5, 247, 218, 5, 5, 248, 218, 5, 5, 249, 218, 5, 5, 250, 218, 5, 5, 251, 218, 5, 5, 252, 218, 5, 5, 253, 218, 5, 5, 254, 218, 5, 5, 106, 225, 5, 5, 255, 218, 5, 5, 2, 219, 5, 5, 3, 219, 5, 5, 4, 219, 5, 5, 5, 219, 5, 5, 33, 212, 5, 5, 6, 219, 5, 5, 7, 219, 5, 5, 8, 219, 5, 5, 108, 225, 5, 5, 109, 225, 5, 5, 110, 225, 5, 5, 111, 225, 5, 5, 112, 225, 5, 5, 113, 225, 5, 5, 114, 225, 5, 5, 115, 225, 5, 5, 116, 225, 5, 5, 117, 225, 5, 5, 118, 225, 5, 5, 119, 225, 5, 5, 120, 225, 5, 5, 121, 225, 5, 5, 122, 225, 5, 5, 123, 225, 5, 5, 124, 225, 5, 5, 125, 225, 5, 5, 126, 225, 5, 5, 127, 225, 5, 5, 128, 225, 5, 5, 129, 225, 5, 5, 130, 225, 5, 5, 131, 225, 5, 5, 132, 225, 5, 5, 133, 225, 5, 5, 134, 225, 5, 5, 135, 225, 5, 5, 136, 225, 5, 5, 137, 225, 5, 5, 138, 225, 5, 5, 139, 225, 5, 5, 140, 225, 5, 5, 141, 225, 5, 5, 142, 225, 5, 5, 143, 225, 5, 5, 144, 225, 5, 5, 145, 225, 5, 5, 146, 225, 5, 5, 147, 225, 5, 5, 148, 225, 5, 5, 149, 225, 5, 5, 150, 225, 5, 5, 151, 225, 5, 5, 152, 225, 5, 5, 153, 225, 5, 5, 154, 225, 5, 5, 155, 225, 5, 5, 156, 225, 5, 5, 157, 225, 5, 5, 158, 225, 5, 5, 9, 219, 5, 5, 159, 225, 5, 5, 160, 225, 5, 5, 161, 225, 5, 5, 162, 225, 5, 5, 163, 225, 5, 5, 164, 225, 5, 5, 165, 225, 5, 5, 166, 225, 5, 5, 167, 225, 5, 5, 168, 225, 5, 5, 169, 225, 5, 5, 170, 225, 5, 5, 171, 225, 5, 5, 172, 225, 5, 5, 107, 225, 5, 5, 173, 225, 5, 5, 174, 225, 5, 5, 175, 225, 5, 5, 176, 225, 5, 5, 177, 225, 5, 5, 178, 225, 5, 5, 179, 225, 5, 5, 180, 225, 5, 5, 255, 230, 5, 5, 2, 231, 5, 5, 3, 231, 5, 5, 4, 231, 5, 5, 5, 231, 5, 5, 6, 231, 5, 5, 7, 231, 5, 5, 8, 231, 5, 5, 9, 231, 5, 5, 10, 231, 5, 5, 11, 231, 5, 5, 12, 231, 5, 5, 13, 231, 5, 5, 14, 231, 5, 5, 15, 231, 5, 5, 16, 231, 5, 5, 17, 231, 5, 5, 18, 231, 5, 5, 19, 231, 5, 5, 20, 231, 5, 5, 21, 231, 5, 5, 22, 231, 5, 5, 23, 231, 5, 5, 24, 231, 5, 5, 25, 231, 5, 5, 26, 231, 5, 5, 27, 231, 5, 5, 28, 231, 5, 5, 29, 231, 5, 5, 30, 231, 5, 5, 31, 231, 5, 5, 32, 231, 5, 5, 33, 231, 5, 5, 34, 231, 5, 5, 35, 231, 5, 5, 36, 231, 5, 5, 37, 231, 5, 5, 38, 231, 5, 5, 39, 231, 5, 5, 40, 231, 5, 5, 41, 231, 5, 5, 42, 231, 5, 5, 43, 231, 5, 5, 44, 231, 5, 5, 45, 231, 5, 5, 46, 231, 5, 5, 160, 235, 5, 5, 161, 235, 5, 5, 162, 235, 5, 5, 163, 235, 5, 5, 164, 235, 5, 5, 165, 235, 5, 5, 166, 235, 5, 5, 167, 235, 5, 5, 168, 235, 5, 5, 169, 235, 5, 5, 170, 235, 5, 5, 171, 235, 5, 5, 172, 235, 5, 5, 173, 235, 5, 5, 174, 235, 5, 5, 175, 235, 5, 5, 176, 235, 5, 5, 177, 235, 5, 5, 178, 235, 5, 5, 179, 235, 5, 5, 180, 235, 5, 5, 181, 235, 5, 5, 182, 235, 5, 5, 183, 235, 5, 5, 184, 235, 5, 5, 185, 235, 5, 5, 186, 235, 5, 5, 187, 235, 5, 5, 188, 235, 5, 5, 189, 235, 5, 5, 190, 235, 5, 5, 191, 235, 5, 5, 62, 239, 5, 5, 63, 239, 5, 5, 64, 239, 5, 5, 65, 239, 5, 5, 66, 239, 5, 5, 67, 239, 5, 5, 68, 239, 5, 5, 69, 239, 5, 5, 70, 239, 5, 5, 71, 239, 5, 5, 72, 239, 5, 5, 73, 239, 5, 5, 74, 239, 5, 5, 75, 239, 5, 5, 76, 239, 5, 5, 77, 239, 5, 5, 57, 231, 5, 5, 78, 239, 5, 5, 79, 239, 5, 5, 80, 239, 5, 5, 81, 239, 5, 5, 82, 239, 5, 5, 194, 235, 5, 5, 83, 239, 5, 5, 84, 239, 5, 5, 85, 239, 5, 5, 86, 239, 5, 5, 87, 239, 5, 5, 96, 239, 5, 5, 100, 242, 5, 5, 101, 242, 5, 5, 102, 242, 5, 5, 88, 239, 5, 5, 103, 242, 5, 5, 195, 235, 5, 5, 104, 242, 5, 5, 105, 242, 5, 5, 106, 242, 5, 5, 107, 242, 5, 5, 108, 242, 5, 5, 109, 242, 5, 5, 110, 242, 5, 5, 111, 242, 5, 5, 112, 242, 5, 5, 141, 246, 5, 5, 203, 244, 5, 5, 120, 242, 5, 5, 204, 244, 5, 5, 205, 244, 5, 5, 206, 244, 5, 5, 207, 244, 5, 5, 208, 244, 5, 5, 209, 244, 5, 5, 210, 244, 5, 5, 211, 244, 5, 5, 212, 244, 5, 5, 213, 244, 5, 5, 214, 244, 5, 5, 215, 244, 5, 5, 142, 246, 5, 5, 143, 246, 5, 5, 144, 246, 5, 5, 145, 246, 5, 5, 146, 246, 5, 5, 147, 246, 5, 5, 216, 244, 5, 5, 148, 246, 5, 5, 149, 246, 5, 5, 209, 247, 5, 5, 210, 247, 5, 5, 211, 247, 5, 5, 212, 247, 5, 5, 199, 248, 5, 5, 200, 248, 5, 5, 201, 248, 5, 5, 111, 249, 5, 5, 112, 249, 5, 5, 113, 249, 5, 5, 114, 249, 5, 5, 115, 249, 5, 5, 116, 249, 5, 5, 224, 249, 5, 5, 117, 249, 5, 5, 89, 250, 5, 5, 90, 250, 5, 5, 75, 158, 5, 5, 98, 161, 5, 5, 99, 161, 5, 5, 54, 169, 5, 5, 141, 164, 5, 5, 55, 169, 5, 5, 56, 169, 5, 5, 57, 169, 5, 5, 156, 174, 5, 5, 157, 174, 5, 5, 158, 174, 5, 5, 19, 181, 5, 5, 20, 181, 5, 5, 21, 181, 5, 5, 22, 181, 5, 5, 23, 181, 5, 5, 24, 181, 5, 5, 25, 181, 5, 5, 185, 188, 5, 5, 186, 188, 5, 5, 26, 181, 5, 5, 187, 188, 5, 5, 188, 188, 5, 5, 189, 188, 5, 5, 190, 188, 5, 5, 145, 196, 5, 5, 146, 196, 5, 5, 147, 196, 5, 5, 148, 196, 5, 5, 149, 196, 5, 5, 150, 196, 5, 5, 151, 196, 5, 5, 116, 204, 5, 5, 117, 204, 5, 5, 118, 204, 5, 5, 119, 204, 5, 5, 120, 204, 5, 5, 121, 204, 5, 5, 122, 204, 5, 5, 123, 204, 5, 5, 124, 204, 5, 5, 43, 212, 5, 5, 44, 212, 5, 5, 45, 212, 5, 5, 46, 212, 5, 5, 47, 212, 5, 5, 20, 219, 5, 5, 21, 219, 5, 5, 22, 219, 5, 5, 23, 219, 5, 5, 24, 219, 5, 5, 25, 219, 5, 5, 190, 225, 5, 5, 191, 225, 5, 5, 192, 225, 5, 5, 193, 225, 5, 5, 194, 225, 5, 5, 195, 225, 5, 5, 196, 225, 5, 5, 58, 231, 5, 5, 59, 231, 5, 5, 60, 231, 5, 5, 196, 235, 5, 5, 197, 235, 5, 5, 97, 239, 5, 5, 151, 246, 5, 5, 76, 158, 5, 5, 118, 159, 5, 5, 100, 161, 5, 5, 142, 164, 5, 5, 59, 169, 5, 5, 60, 169, 5, 5, 61, 169, 5, 5, 62, 169, 5, 5, 160, 174, 5, 5, 161, 174, 5, 5, 27, 181, 5, 5, 28, 181, 5, 5, 152, 196, 5, 5, 153, 196, 5, 5, 48, 212, 5, 5, 127, 204, 5, 5, 128, 204, 5, 5, 129, 204, 5, 5, 49, 212, 5, 5, 26, 219, 5, 5, 27, 219, 5, 5, 197, 225, 5, 5, 198, 235, 5, 5, 78, 158, 5, 5, 119, 159, 5, 5, 101, 161, 5, 5, 144, 164, 5, 5, 63, 169, 5, 5, 64, 169, 5, 5, 65, 169, 5, 5, 66, 169, 5, 5, 67, 169, 5, 5, 163, 174, 5, 5, 164, 174, 5, 5, 165, 174, 5, 5, 166, 174, 5, 5, 167, 174, 5, 5, 168, 174, 5, 5, 29, 181, 5, 5, 30, 181, 5, 5, 31, 181, 5, 5, 32, 181, 5, 5, 192, 188, 5, 5, 193, 188, 5, 5, 194, 188, 5, 5, 195, 188, 5, 5, 196, 188, 5, 5, 197, 188, 5, 5, 198, 188, 5, 5, 199, 188, 5, 5, 154, 196, 5, 5, 155, 196, 5, 5, 156, 196, 5, 5, 157, 196, 5, 5, 158, 196, 5, 5, 159, 196, 5, 5, 160, 196, 5, 5, 130, 204, 5, 5, 131, 204, 5, 5, 50, 212, 5, 5, 51, 212, 5, 5, 52, 212, 5, 5, 53, 212, 5, 5, 54, 212, 5, 5, 28, 219, 5, 5, 29, 219, 5, 5, 30, 219, 5, 5, 31, 219, 5, 5, 32, 219, 5, 5, 198, 225, 5, 5, 199, 225, 5, 5, 200, 225, 5, 5, 201, 225, 5, 5, 202, 225, 5, 5, 62, 231, 5, 5, 63, 231, 5, 5, 64, 231, 5, 5, 199, 235, 5, 5, 98, 239, 5, 5, 99, 239, 5, 5, 218, 244, 5, 5, 79, 158, 5, 5, 68, 169, 5, 5, 169, 174, 5, 5, 170, 174, 5, 5, 34, 181, 5, 5, 201, 188, 5, 5, 202, 188, 5, 5, 203, 188, 5, 5, 204, 188, 5, 5, 165, 196, 5, 5, 166, 196, 5, 5, 167, 196, 5, 5, 133, 204, 5, 5, 134, 204, 5, 5, 135, 204, 5, 5, 136, 204, 5, 5, 55, 212, 5, 5, 56, 212, 5, 5, 35, 219, 5, 5, 36, 219, 5, 5, 203, 225, 5, 5, 204, 225, 5, 5, 200, 235, 5, 5, 214, 247, 5, 5, 80, 158, 5, 5, 81, 158, 5, 5, 120, 159, 5, 5, 102, 161, 5, 5, 146, 164, 5, 5, 147, 164, 5, 5, 69, 169, 5, 5, 171, 174, 5, 5, 138, 204, 5, 5, 82, 158, 5, 5, 103, 161, 5, 5, 172, 174, 5, 5, 173, 174, 5, 5, 174, 174, 5, 5, 35, 181, 5, 5, 65, 231, 5, 5, 83, 158, 5, 5, 149, 164, 5, 5, 150, 164, 5, 5, 70, 169, 5, 5, 71, 169, 5, 5, 176, 174, 5, 5, 177, 174, 5, 5, 36, 181, 5, 5, 37, 181, 5, 5, 38, 181, 5, 5, 39, 181, 5, 5, 40, 181, 5, 5, 41, 181, 5, 5, 42, 181, 5, 5, 43, 181, 5, 5, 44, 181, 5, 5, 205, 188, 5, 5, 206, 188, 5, 5, 207, 188, 5, 5, 208, 188, 5, 5, 168, 196, 5, 5, 169, 196, 5, 5, 170, 196, 5, 5, 171, 196, 5, 5, 172, 196, 5, 5, 173, 196, 5, 5, 174, 196, 5, 5, 175, 196, 5, 5, 139, 204, 5, 5, 140, 204, 5, 5, 141, 204, 5, 5, 142, 204, 5, 5, 143, 204, 5, 5, 144, 204, 5, 5, 145, 204, 5, 5, 58, 212, 5, 5, 37, 219, 5, 5, 38, 219, 5, 5, 39, 219, 5, 5, 40, 219, 5, 5, 205, 225, 5, 5, 206, 225, 5, 5, 207, 225, 5, 5, 208, 225, 5, 5, 209, 225, 5, 5, 66, 231, 5, 5, 67, 231, 5, 5, 68, 231, 5, 5, 201, 235, 5, 5, 101, 239, 5, 5, 152, 246, 5, 5, 225, 249, 5, 5, 84, 158, 5, 5, 121, 159, 5, 5, 122, 159, 5, 5, 104, 161, 5, 5, 72, 169, 5, 5, 85, 158, 5, 5, 123, 159, 5, 5, 105, 161, 5, 5, 106, 161, 5, 5, 107, 161, 5, 5, 151, 164, 5, 5, 152, 164, 5, 5, 73, 169, 5, 5, 74, 169, 5, 5, 75, 169, 5, 5, 179, 174, 5, 5, 180, 174, 5, 5, 181, 174, 5, 5, 182, 174, 5, 5, 183, 174, 5, 5, 45, 181, 5, 5, 46, 181, 5, 5, 47, 181, 5, 5, 48, 181, 5, 5, 49, 181, 5, 5, 50, 181, 5, 5, 51, 181, 5, 5, 209, 188, 5, 5, 210, 188, 5, 5, 176, 196, 5, 5, 177, 196, 5, 5, 178, 196, 5, 5, 179, 196, 5, 5, 180, 196, 5, 5, 146, 204, 5, 5, 59, 212, 5, 5, 60, 212, 5, 5, 86, 158, 5, 5, 149, 157, 5, 5, 125, 159, 5, 5, 126, 159, 5, 5, 127, 159, 5, 5, 128, 159, 5, 5, 124, 159, 5, 5, 129, 159, 5, 5, 108, 161, 5, 5, 65, 160, 5, 5, 130, 159, 5, 5, 131, 159, 5, 5, 132, 159, 5, 5, 133, 159, 5, 5, 153, 164, 5, 5, 134, 159, 5, 5, 135, 159, 5, 5, 136, 159, 5, 5, 109, 161, 5, 5, 137, 159, 5, 5, 138, 159, 5, 5, 139, 159, 5, 5, 110, 161, 5, 5, 111, 161, 5, 5, 112, 161, 5, 5, 113, 161, 5, 5, 114, 161, 5, 5, 115, 161, 5, 5, 116, 161, 5, 5, 117, 161, 5, 5, 118, 161, 5, 5, 119, 161, 5, 5, 120, 161, 5, 5, 121, 161, 5, 5, 154, 164, 5, 5, 122, 161, 5, 5, 123, 161, 5, 5, 124, 161, 5, 5, 125, 161, 5, 5, 126, 161, 5, 5, 127, 161, 5, 5, 128, 161, 5, 5, 155, 164, 5, 5, 129, 161, 5, 5, 130, 161, 5, 5, 131, 161, 5, 5, 132, 161, 5, 5, 133, 161, 5, 5, 134, 161, 5, 5, 156, 164, 5, 5, 157, 164, 5, 5, 158, 164, 5, 5, 159, 164, 5, 5, 160, 164, 5, 5, 161, 164, 5, 5, 162, 164, 5, 5, 76, 169, 5, 5, 163, 164, 5, 5, 164, 164, 5, 5, 165, 164, 5, 5, 166, 164, 5, 5, 167, 164, 5, 5, 168, 164, 5, 5, 169, 164, 5, 5, 170, 164, 5, 5, 171, 164, 5, 5, 172, 164, 5, 5, 135, 161, 5, 5, 173, 164, 5, 5, 174, 164, 5, 5, 175, 164, 5, 5, 176, 164, 5, 5, 177, 164, 5, 5, 178, 164, 5, 5, 179, 164, 5, 5, 180, 164, 5, 5, 77, 169, 5, 5, 181, 164, 5, 5, 182, 164, 5, 5, 183, 164, 5, 5, 184, 164, 5, 5, 185, 164, 5, 5, 186, 164, 5, 5, 187, 164, 5, 5, 188, 164, 5, 5, 189, 164, 5, 5, 78, 169, 5, 5, 190, 164, 5, 5, 191, 164, 5, 5, 192, 164, 5, 5, 193, 164, 5, 5, 194, 164, 5, 5, 195, 164, 5, 5, 196, 164, 5, 5, 197, 164, 5, 5, 79, 169, 5, 5, 198, 164, 5, 5, 199, 164, 5, 5, 200, 164, 5, 5, 184, 174, 5, 5, 201, 164, 5, 5, 202, 164, 5, 5, 203, 164, 5, 5, 204, 164, 5, 5, 205, 164, 5, 5, 80, 169, 5, 5, 206, 164, 5, 5, 207, 164, 5, 5, 208, 164, 5, 5, 209, 164, 5, 5, 210, 164, 5, 5, 211, 164, 5, 5, 212, 164, 5, 5, 213, 164, 5, 5, 214, 164, 5, 5, 215, 164, 5, 5, 216, 164, 5, 5, 217, 164, 5, 5, 218, 164, 5, 5, 81, 169, 5, 5, 82, 169, 5, 5, 83, 169, 5, 5, 84, 169, 5, 5, 185, 174, 5, 5, 85, 169, 5, 5, 86, 169, 5, 5, 87, 169, 5, 5, 88, 169, 5, 5, 89, 169, 5, 5, 90, 169, 5, 5, 91, 169, 5, 5, 92, 169, 5, 5, 93, 169, 5, 5, 94, 169, 5, 5, 95, 169, 5, 5, 96, 169, 5, 5, 97, 169, 5, 5, 98, 169, 5, 5, 99, 169, 5, 5, 100, 169, 5, 5, 101, 169, 5, 5, 102, 169, 5, 5, 103, 169, 5, 5, 104, 169, 5, 5, 105, 169, 5, 5, 106, 169, 5, 5, 107, 169, 5, 5, 108, 169, 5, 5, 109, 169, 5, 5, 186, 174, 5, 5, 110, 169, 5, 5, 111, 169, 5, 5, 112, 169, 5, 5, 113, 169, 5, 5, 114, 169, 5, 5, 115, 169, 5, 5, 116, 169, 5, 5, 117, 169, 5, 5, 118, 169, 5, 5, 119, 169, 5, 5, 120, 169, 5, 5, 121, 169, 5, 5, 122, 169, 5, 5, 123, 169, 5, 5, 124, 169, 5, 5, 125, 169, 5, 5, 189, 174, 5, 5, 126, 169, 5, 5, 127, 169, 5, 5, 128, 169, 5, 5, 129, 169, 5, 5, 130, 169, 5, 5, 131, 169, 5, 5, 132, 169, 5, 5, 133, 169, 5, 5, 134, 169, 5, 5, 135, 169, 5, 5, 136, 169, 5, 5, 137, 169, 5, 5, 138, 169, 5, 5, 139, 169, 5, 5, 140, 169, 5, 5, 141, 169, 5, 5, 142, 169, 5, 5, 143, 169, 5, 5, 144, 169, 5, 5, 145, 169, 5, 5, 146, 169, 5, 5, 52, 181, 5, 5, 147, 169, 5, 5, 148, 169, 5, 5, 149, 169, 5, 5, 187, 174, 5, 5, 228, 175, 5, 5, 188, 174, 5, 5, 150, 169, 5, 5, 151, 169, 5, 5, 152, 169, 5, 5, 153, 169, 5, 5, 154, 169, 5, 5, 155, 169, 5, 5, 156, 169, 5, 5, 157, 169, 5, 5, 190, 174, 5, 5, 191, 174, 5, 5, 192, 174, 5, 5, 193, 174, 5, 5, 194, 174, 5, 5, 195, 174, 5, 5, 196, 174, 5, 5, 197, 174, 5, 5, 198, 174, 5, 5, 199, 174, 5, 5, 200, 174, 5, 5, 201, 174, 5, 5, 202, 174, 5, 5, 203, 174, 5, 5, 56, 181, 5, 5, 204, 174, 5, 5, 205, 174, 5, 5, 206, 174, 5, 5, 207, 174, 5, 5, 208, 174, 5, 5, 209, 174, 5, 5, 210, 174, 5, 5, 211, 174, 5, 5, 57, 181, 5, 5, 212, 174, 5, 5, 213, 174, 5, 5, 214, 174, 5, 5, 215, 174, 5, 5, 216, 174, 5, 5, 53, 181, 5, 5, 217, 174, 5, 5, 218, 174, 5, 5, 219, 174, 5, 5, 220, 174, 5, 5, 221, 174, 5, 5, 222, 174, 5, 5, 223, 174, 5, 5, 224, 174, 5, 5, 225, 174, 5, 5, 226, 174, 5, 5, 227, 174, 5, 5, 228, 174, 5, 5, 229, 174, 5, 5, 230, 174, 5, 5, 231, 174, 5, 5, 232, 174, 5, 5, 233, 174, 5, 5, 234, 174, 5, 5, 54, 181, 5, 5, 166, 169, 5, 5, 235, 174, 5, 5, 236, 174, 5, 5, 237, 174, 5, 5, 238, 174, 5, 5, 239, 174, 5, 5, 240, 174, 5, 5, 241, 174, 5, 5, 242, 174, 5, 5, 243, 174, 5, 5, 244, 174, 5, 5, 245, 174, 5, 5, 246, 174, 5, 5, 247, 174, 5, 5, 248, 174, 5, 5, 249, 174, 5, 5, 250, 174, 5, 5, 251, 174, 5, 5, 252, 174, 5, 5, 253, 174, 5, 5, 254, 174, 5, 5, 255, 174, 5, 5, 55, 181, 5, 5, 2, 175, 5, 5, 3, 175, 5, 5, 4, 175, 5, 5, 5, 175, 5, 5, 6, 175, 5, 5, 7, 175, 5, 5, 8, 175, 5, 5, 9, 175, 5, 5, 10, 175, 5, 5, 11, 175, 5, 5, 12, 175, 5, 5, 13, 175, 5, 5, 14, 175, 5, 5, 15, 175, 5, 5, 16, 175, 5, 5, 58, 181, 5, 5, 59, 181, 5, 5, 60, 181, 5, 5, 61, 181, 5, 5, 62, 181, 5, 5, 63, 181, 5, 5, 64, 181, 5, 5, 65, 181, 5, 5, 66, 181, 5, 5, 67, 181, 5, 5, 68, 181, 5, 5, 69, 181, 5, 5, 70, 181, 5, 5, 71, 181, 5, 5, 72, 181, 5, 5, 73, 181, 5, 5, 74, 181, 5, 5, 75, 181, 5, 5, 76, 181, 5, 5, 77, 181, 5, 5, 78, 181, 5, 5, 79, 181, 5, 5, 80, 181, 5, 5, 81, 181, 5, 5, 82, 181, 5, 5, 83, 181, 5, 5, 84, 181, 5, 5, 85, 181, 5, 5, 86, 181, 5, 5, 87, 181, 5, 5, 88, 181, 5, 5, 89, 181, 5, 5, 90, 181, 5, 5, 91, 181, 5, 5, 92, 181, 5, 5, 93, 181, 5, 5, 94, 181, 5, 5, 95, 181, 5, 5, 96, 181, 5, 5, 97, 181, 5, 5, 98, 181, 5, 5, 99, 181, 5, 5, 100, 181, 5, 5, 101, 181, 5, 5, 102, 181, 5, 5, 103, 181, 5, 5, 104, 181, 5, 5, 105, 181, 5, 5, 106, 181, 5, 5, 107, 181, 5, 5, 108, 181, 5, 5, 109, 181, 5, 5, 110, 181, 5, 5, 111, 181, 5, 5, 112, 181, 5, 5, 113, 181, 5, 5, 212, 188, 5, 5, 114, 181, 5, 5, 115, 181, 5, 5, 116, 181, 5, 5, 117, 181, 5, 5, 118, 181, 5, 5, 119, 181, 5, 5, 120, 181, 5, 5, 121, 181, 5, 5, 122, 181, 5, 5, 123, 181, 5, 5, 214, 188, 5, 5, 124, 181, 5, 5, 125, 181, 5, 5, 126, 181, 5, 5, 127, 181, 5, 5, 128, 181, 5, 5, 129, 181, 5, 5, 130, 181, 5, 5, 131, 181, 5, 5, 132, 181, 5, 5, 133, 181, 5, 5, 134, 181, 5, 5, 135, 181, 5, 5, 136, 181, 5, 5, 137, 181, 5, 5, 138, 181, 5, 5, 139, 181, 5, 5, 215, 188, 5, 5, 216, 188, 5, 5, 217, 188, 5, 5, 218, 188, 5, 5, 219, 188, 5, 5, 220, 188, 5, 5, 221, 188, 5, 5, 222, 188, 5, 5, 223, 188, 5, 5, 224, 188, 5, 5, 225, 188, 5, 5, 226, 188, 5, 5, 227, 188, 5, 5, 228, 188, 5, 5, 229, 188, 5, 5, 230, 188, 5, 5, 231, 188, 5, 5, 232, 188, 5, 5, 233, 188, 5, 5, 234, 188, 5, 5, 235, 188, 5, 5, 236, 188, 5, 5, 237, 188, 5, 5, 238, 188, 5, 5, 239, 188, 5, 5, 240, 188, 5, 5, 241, 188, 5, 5, 242, 188, 5, 5, 243, 188, 5, 5, 244, 188, 5, 5, 245, 188, 5, 5, 246, 188, 5, 5, 247, 188, 5, 5, 248, 188, 5, 5, 249, 188, 5, 5, 250, 188, 5, 5, 251, 188, 5, 5, 252, 188, 5, 5, 253, 188, 5, 5, 254, 188, 5, 5, 255, 188, 5, 5, 2, 189, 5, 5, 3, 189, 5, 5, 4, 189, 5, 5, 5, 189, 5, 5, 6, 189, 5, 5, 7, 189, 5, 5, 8, 189, 5, 5, 9, 189, 5, 5, 10, 189, 5, 5, 11, 189, 5, 5, 12, 189, 5, 5, 13, 189, 5, 5, 14, 189, 5, 5, 15, 189, 5, 5, 16, 189, 5, 5, 17, 189, 5, 5, 18, 189, 5, 5, 19, 189, 5, 5, 20, 189, 5, 5, 21, 189, 5, 5, 22, 189, 5, 5, 23, 189, 5, 5, 24, 189, 5, 5, 25, 189, 5, 5, 26, 189, 5, 5, 27, 189, 5, 5, 28, 189, 5, 5, 29, 189, 5, 5, 30, 189, 5, 5, 31, 189, 5, 5, 32, 189, 5, 5, 33, 189, 5, 5, 34, 189, 5, 5, 35, 189, 5, 5, 36, 189, 5, 5, 37, 189, 5, 5, 38, 189, 5, 5, 39, 189, 5, 5, 40, 189, 5, 5, 41, 189, 5, 5, 42, 189, 5, 5, 181, 196, 5, 5, 43, 189, 5, 5, 182, 196, 5, 5, 44, 189, 5, 5, 45, 189, 5, 5, 46, 189, 5, 5, 47, 189, 5, 5, 183, 196, 5, 5, 48, 189, 5, 5, 49, 189, 5, 5, 50, 189, 5, 5, 51, 189, 5, 5, 52, 189, 5, 5, 53, 189, 5, 5, 54, 189, 5, 5, 55, 189, 5, 5, 56, 189, 5, 5, 57, 189, 5, 5, 58, 189, 5, 5, 59, 189, 5, 5, 60, 189, 5, 5, 61, 189, 5, 5, 62, 189, 5, 5, 63, 189, 5, 5, 64, 189, 5, 5, 65, 189, 5, 5, 66, 189, 5, 5, 67, 189, 5, 5, 184, 196, 5, 5, 185, 196, 5, 5, 68, 189, 5, 5, 186, 196, 5, 5, 187, 196, 5, 5, 188, 196, 5, 5, 189, 196, 5, 5, 190, 196, 5, 5, 191, 196, 5, 5, 192, 196, 5, 5, 193, 196, 5, 5, 194, 196, 5, 5, 195, 196, 5, 5, 196, 196, 5, 5, 197, 196, 5, 5, 198, 196, 5, 5, 199, 196, 5, 5, 200, 196, 5, 5, 201, 196, 5, 5, 202, 196, 5, 5, 203, 196, 5, 5, 204, 196, 5, 5, 205, 196, 5, 5, 206, 196, 5, 5, 207, 196, 5, 5, 208, 196, 5, 5, 209, 196, 5, 5, 210, 196, 5, 5, 211, 196, 5, 5, 212, 196, 5, 5, 213, 196, 5, 5, 214, 196, 5, 5, 215, 196, 5, 5, 216, 196, 5, 5, 217, 196, 5, 5, 218, 196, 5, 5, 219, 196, 5, 5, 220, 196, 5, 5, 221, 196, 5, 5, 222, 196, 5, 5, 223, 196, 5, 5, 224, 196, 5, 5, 225, 196, 5, 5, 226, 196, 5, 5, 227, 196, 5, 5, 228, 196, 5, 5, 229, 196, 5, 5, 230, 196, 5, 5, 231, 196, 5, 5, 232, 196, 5, 5, 233, 196, 5, 5, 234, 196, 5, 5, 235, 196, 5, 5, 236, 196, 5, 5, 237, 196, 5, 5, 238, 196, 5, 5, 239, 196, 5, 5, 240, 196, 5, 5, 241, 196, 5, 5, 242, 196, 5, 5, 243, 196, 5, 5, 244, 196, 5, 5, 245, 196, 5, 5, 246, 196, 5, 5, 247, 196, 5, 5, 248, 196, 5, 5, 249, 196, 5, 5, 250, 196, 5, 5, 251, 196, 5, 5, 252, 196, 5, 5, 253, 196, 5, 5, 254, 196, 5, 5, 255, 196, 5, 5, 2, 197, 5, 5, 3, 197, 5, 5, 4, 197, 5, 5, 5, 197, 5, 5, 6, 197, 5, 5, 7, 197, 5, 5, 8, 197, 5, 5, 9, 197, 5, 5, 10, 197, 5, 5, 11, 197, 5, 5, 12, 197, 5, 5, 147, 204, 5, 5, 13, 197, 5, 5, 14, 197, 5, 5, 15, 197, 5, 5, 16, 197, 5, 5, 17, 197, 5, 5, 18, 197, 5, 5, 19, 197, 5, 5, 69, 189, 5, 5, 20, 197, 5, 5, 21, 197, 5, 5, 22, 197, 5, 5, 23, 197, 5, 5, 24, 197, 5, 5, 25, 197, 5, 5, 26, 197, 5, 5, 27, 197, 5, 5, 28, 197, 5, 5, 29, 197, 5, 5, 30, 197, 5, 5, 31, 197, 5, 5, 32, 197, 5, 5, 33, 197, 5, 5, 34, 197, 5, 5, 35, 197, 5, 5, 36, 197, 5, 5, 37, 197, 5, 5, 38, 197, 5, 5, 39, 197, 5, 5, 40, 197, 5, 5, 41, 197, 5, 5, 42, 197, 5, 5, 43, 197, 5, 5, 148, 204, 5, 5, 149, 204, 5, 5, 150, 204, 5, 5, 151, 204, 5, 5, 152, 204, 5, 5, 153, 204, 5, 5, 154, 204, 5, 5, 155, 204, 5, 5, 156, 204, 5, 5, 157, 204, 5, 5, 158, 204, 5, 5, 159, 204, 5, 5, 160, 204, 5, 5, 161, 204, 5, 5, 162, 204, 5, 5, 163, 204, 5, 5, 164, 204, 5, 5, 165, 204, 5, 5, 166, 204, 5, 5, 167, 204, 5, 5, 168, 204, 5, 5, 169, 204, 5, 5, 170, 204, 5, 5, 171, 204, 5, 5, 172, 204, 5, 5, 173, 204, 5, 5, 174, 204, 5, 5, 175, 204, 5, 5, 176, 204, 5, 5, 177, 204, 5, 5, 178, 204, 5, 5, 179, 204, 5, 5, 180, 204, 5, 5, 181, 204, 5, 5, 182, 204, 5, 5, 183, 204, 5, 5, 184, 204, 5, 5, 185, 204, 5, 5, 186, 204, 5, 5, 187, 204, 5, 5, 188, 204, 5, 5, 189, 204, 5, 5, 190, 204, 5, 5, 191, 204, 5, 5, 192, 204, 5, 5, 193, 204, 5, 5, 194, 204, 5, 5, 195, 204, 5, 5, 196, 204, 5, 5, 197, 204, 5, 5, 198, 204, 5, 5, 199, 204, 5, 5, 200, 204, 5, 5, 201, 204, 5, 5, 202, 204, 5, 5, 203, 204, 5, 5, 204, 204, 5, 5, 205, 204, 5, 5, 206, 204, 5, 5, 207, 204, 5, 5, 208, 204, 5, 5, 209, 204, 5, 5, 44, 197, 5, 5, 62, 212, 5, 5, 210, 204, 5, 5, 61, 212, 5, 5, 211, 204, 5, 5, 212, 204, 5, 5, 213, 204, 5, 5, 214, 204, 5, 5, 215, 204, 5, 5, 216, 204, 5, 5, 41, 219, 5, 5, 217, 204, 5, 5, 218, 204, 5, 5, 219, 204, 5, 5, 220, 204, 5, 5, 247, 204, 5, 5, 221, 204, 5, 5, 222, 204, 5, 5, 223, 204, 5, 5, 45, 197, 5, 5, 224, 204, 5, 5, 225, 204, 5, 5, 226, 204, 5, 5, 227, 204, 5, 5, 228, 204, 5, 5, 229, 204, 5, 5, 230, 204, 5, 5, 231, 204, 5, 5, 232, 204, 5, 5, 233, 204, 5, 5, 234, 204, 5, 5, 235, 204, 5, 5, 63, 212, 5, 5, 64, 212, 5, 5, 65, 212, 5, 5, 66, 212, 5, 5, 67, 212, 5, 5, 68, 212, 5, 5, 69, 212, 5, 5, 70, 212, 5, 5, 71, 212, 5, 5, 72, 212, 5, 5, 73, 212, 5, 5, 74, 212, 5, 5, 75, 212, 5, 5, 76, 212, 5, 5, 77, 212, 5, 5, 78, 212, 5, 5, 79, 212, 5, 5, 80, 212, 5, 5, 81, 212, 5, 5, 82, 212, 5, 5, 83, 212, 5, 5, 42, 219, 5, 5, 84, 212, 5, 5, 85, 212, 5, 5, 86, 212, 5, 5, 87, 212, 5, 5, 88, 212, 5, 5, 89, 212, 5, 5, 90, 212, 5, 5, 91, 212, 5, 5, 92, 212, 5, 5, 93, 212, 5, 5, 47, 219, 5, 5, 94, 212, 5, 5, 95, 212, 5, 5, 96, 212, 5, 5, 97, 212, 5, 5, 43, 219, 5, 5, 98, 212, 5, 5, 99, 212, 5, 5, 236, 204, 5, 5, 100, 212, 5, 5, 101, 212, 5, 5, 102, 212, 5, 5, 103, 212, 5, 5, 104, 212, 5, 5, 105, 212, 5, 5, 106, 212, 5, 5, 107, 212, 5, 5, 108, 212, 5, 5, 109, 212, 5, 5, 110, 212, 5, 5, 111, 212, 5, 5, 112, 212, 5, 5, 113, 212, 5, 5, 114, 212, 5, 5, 115, 212, 5, 5, 116, 212, 5, 5, 117, 212, 5, 5, 44, 219, 5, 5, 118, 212, 5, 5, 119, 212, 5, 5, 120, 212, 5, 5, 121, 212, 5, 5, 122, 212, 5, 5, 123, 212, 5, 5, 124, 212, 5, 5, 125, 212, 5, 5, 126, 212, 5, 5, 127, 212, 5, 5, 128, 212, 5, 5, 129, 212, 5, 5, 130, 212, 5, 5, 131, 212, 5, 5, 132, 212, 5, 5, 133, 212, 5, 5, 134, 212, 5, 5, 135, 212, 5, 5, 136, 212, 5, 5, 137, 212, 5, 5, 138, 212, 5, 5, 139, 212, 5, 5, 48, 219, 5, 5, 140, 212, 5, 5, 45, 219, 5, 5, 141, 212, 5, 5, 46, 219, 5, 5, 142, 212, 5, 5, 143, 212, 5, 5, 144, 212, 5, 5, 145, 212, 5, 5, 146, 212, 5, 5, 147, 212, 5, 5, 148, 212, 5, 5, 149, 212, 5, 5, 150, 212, 5, 5, 151, 212, 5, 5, 152, 212, 5, 5, 153, 212, 5, 5, 165, 212, 5, 5, 49, 219, 5, 5, 50, 219, 5, 5, 51, 219, 5, 5, 52, 219, 5, 5, 53, 219, 5, 5, 54, 219, 5, 5, 55, 219, 5, 5, 56, 219, 5, 5, 57, 219, 5, 5, 58, 219, 5, 5, 59, 219, 5, 5, 60, 219, 5, 5, 61, 219, 5, 5, 62, 219, 5, 5, 63, 219, 5, 5, 210, 225, 5, 5, 64, 219, 5, 5, 65, 219, 5, 5, 66, 219, 5, 5, 67, 219, 5, 5, 68, 219, 5, 5, 69, 219, 5, 5, 70, 219, 5, 5, 71, 219, 5, 5, 72, 219, 5, 5, 73, 219, 5, 5, 74, 219, 5, 5, 75, 219, 5, 5, 76, 219, 5, 5, 77, 219, 5, 5, 78, 219, 5, 5, 79, 219, 5, 5, 80, 219, 5, 5, 81, 219, 5, 5, 82, 219, 5, 5, 83, 219, 5, 5, 166, 212, 5, 5, 84, 219, 5, 5, 85, 219, 5, 5, 86, 219, 5, 5, 87, 219, 5, 5, 88, 219, 5, 5, 89, 219, 5, 5, 90, 219, 5, 5, 91, 219, 5, 5, 92, 219, 5, 5, 93, 219, 5, 5, 94, 219, 5, 5, 95, 219, 5, 5, 71, 231, 5, 5, 96, 219, 5, 5, 97, 219, 5, 5, 211, 225, 5, 5, 98, 219, 5, 5, 99, 219, 5, 5, 100, 219, 5, 5, 101, 219, 5, 5, 102, 219, 5, 5, 103, 219, 5, 5, 104, 219, 5, 5, 105, 219, 5, 5, 106, 219, 5, 5, 107, 219, 5, 5, 108, 219, 5, 5, 109, 219, 5, 5, 110, 219, 5, 5, 111, 219, 5, 5, 112, 219, 5, 5, 113, 219, 5, 5, 114, 219, 5, 5, 115, 219, 5, 5, 116, 219, 5, 5, 117, 219, 5, 5, 118, 219, 5, 5, 212, 225, 5, 5, 119, 219, 5, 5, 120, 219, 5, 5, 121, 219, 5, 5, 122, 219, 5, 5, 213, 225, 5, 5, 214, 225, 5, 5, 215, 225, 5, 5, 216, 225, 5, 5, 217, 225, 5, 5, 218, 225, 5, 5, 219, 225, 5, 5, 220, 225, 5, 5, 221, 225, 5, 5, 222, 225, 5, 5, 223, 225, 5, 5, 69, 231, 5, 5, 224, 225, 5, 5, 225, 225, 5, 5, 226, 225, 5, 5, 227, 225, 5, 5, 228, 225, 5, 5, 229, 225, 5, 5, 230, 225, 5, 5, 231, 225, 5, 5, 232, 225, 5, 5, 233, 225, 5, 5, 234, 225, 5, 5, 235, 225, 5, 5, 236, 225, 5, 5, 237, 225, 5, 5, 238, 225, 5, 5, 239, 225, 5, 5, 240, 225, 5, 5, 241, 225, 5, 5, 242, 225, 5, 5, 243, 225, 5, 5, 244, 225, 5, 5, 245, 225, 5, 5, 246, 225, 5, 5, 247, 225, 5, 5, 248, 225, 5, 5, 249, 225, 5, 5, 250, 225, 5, 5, 251, 225, 5, 5, 136, 219, 5, 5, 252, 225, 5, 5, 253, 225, 5, 5, 254, 225, 5, 5, 255, 225, 5, 5, 2, 226, 5, 5, 70, 231, 5, 5, 3, 226, 5, 5, 4, 226, 5, 5, 5, 226, 5, 5, 123, 219, 5, 5, 6, 226, 5, 5, 7, 226, 5, 5, 8, 226, 5, 5, 72, 231, 5, 5, 73, 231, 5, 5, 9, 226, 5, 5, 74, 231, 5, 5, 75, 231, 5, 5, 76, 231, 5, 5, 77, 231, 5, 5, 78, 231, 5, 5, 79, 231, 5, 5, 80, 231, 5, 5, 81, 231, 5, 5, 82, 231, 5, 5, 83, 231, 5, 5, 84, 231, 5, 5, 85, 231, 5, 5, 86, 231, 5, 5, 87, 231, 5, 5, 88, 231, 5, 5, 89, 231, 5, 5, 90, 231, 5, 5, 91, 231, 5, 5, 92, 231, 5, 5, 93, 231, 5, 5, 94, 231, 5, 5, 95, 231, 5, 5, 96, 231, 5, 5, 97, 231, 5, 5, 98, 231, 5, 5, 99, 231, 5, 5, 100, 231, 5, 5, 101, 231, 5, 5, 102, 239, 5, 5, 102, 231, 5, 5, 103, 231, 5, 5, 104, 231, 5, 5, 202, 235, 5, 5, 105, 231, 5, 5, 203, 235, 5, 5, 204, 235, 5, 5, 205, 235, 5, 5, 206, 235, 5, 5, 207, 235, 5, 5, 208, 235, 5, 5, 209, 235, 5, 5, 210, 235, 5, 5, 211, 235, 5, 5, 212, 235, 5, 5, 213, 235, 5, 5, 23, 226, 5, 5, 214, 235, 5, 5, 215, 235, 5, 5, 216, 235, 5, 5, 217, 235, 5, 5, 218, 235, 5, 5, 219, 235, 5, 5, 220, 235, 5, 5, 221, 235, 5, 5, 222, 235, 5, 5, 223, 235, 5, 5, 224, 235, 5, 5, 225, 235, 5, 5, 226, 235, 5, 5, 227, 235, 5, 5, 228, 235, 5, 5, 229, 235, 5, 5, 103, 239, 5, 5, 104, 239, 5, 5, 105, 239, 5, 5, 106, 239, 5, 5, 107, 239, 5, 5, 108, 239, 5, 5, 109, 239, 5, 5, 110, 239, 5, 5, 111, 239, 5, 5, 112, 239, 5, 5, 113, 239, 5, 5, 114, 239, 5, 5, 115, 239, 5, 5, 116, 239, 5, 5, 117, 239, 5, 5, 118, 239, 5, 5, 119, 239, 5, 5, 236, 235, 5, 5, 120, 239, 5, 5, 121, 239, 5, 5, 122, 239, 5, 5, 121, 242, 5, 5, 123, 239, 5, 5, 124, 239, 5, 5, 125, 239, 5, 5, 126, 239, 5, 5, 122, 242, 5, 5, 123, 242, 5, 5, 124, 242, 5, 5, 125, 242, 5, 5, 126, 242, 5, 5, 127, 242, 5, 5, 128, 242, 5, 5, 129, 242, 5, 5, 130, 242, 5, 5, 131, 242, 5, 5, 132, 242, 5, 5, 133, 242, 5, 5, 134, 242, 5, 5, 135, 242, 5, 5, 136, 242, 5, 5, 137, 242, 5, 5, 138, 242, 5, 5, 139, 242, 5, 5, 140, 242, 5, 5, 145, 242, 5, 5, 219, 244, 5, 5, 220, 244, 5, 5, 221, 244, 5, 5, 222, 244, 5, 5, 223, 244, 5, 5, 224, 244, 5, 5, 225, 244, 5, 5, 226, 244, 5, 5, 227, 244, 5, 5, 228, 244, 5, 5, 229, 244, 5, 5, 226, 249, 5, 5, 230, 244, 5, 5, 231, 244, 5, 5, 153, 246, 5, 5, 154, 246, 5, 5, 215, 247, 5, 5, 155, 246, 5, 5, 156, 246, 5, 5, 157, 246, 5, 5, 158, 246, 5, 5, 159, 246, 5, 5, 216, 247, 5, 5, 217, 247, 5, 5, 218, 247, 5, 5, 219, 247, 5, 5, 202, 248, 5, 5, 203, 248, 5, 5, 204, 248, 5, 5, 205, 248, 5, 5, 206, 248, 5, 5, 118, 249, 5, 5, 119, 249, 5, 5, 227, 249, 5, 5, 28, 250, 5, 5, 228, 249, 5, 5, 29, 250, 5, 5, 30, 250, 5, 5, 153, 250, 5, 5, 159, 250, 5, 5, 87, 158, 5, 5, 88, 158, 5, 5, 140, 159, 5, 5, 136, 161, 5, 5, 137, 161, 5, 5, 138, 161, 5, 5, 139, 161, 5, 5, 140, 161, 5, 5, 141, 161, 5, 5, 222, 164, 5, 5, 223, 164, 5, 5, 224, 164, 5, 5, 225, 164, 5, 5, 226, 164, 5, 5, 227, 164, 5, 5, 228, 164, 5, 5, 229, 164, 5, 5, 230, 164, 5, 5, 231, 164, 5, 5, 232, 164, 5, 5, 233, 164, 5, 5, 234, 164, 5, 5, 168, 169, 5, 5, 169, 169, 5, 5, 170, 169, 5, 5, 171, 169, 5, 5, 172, 169, 5, 5, 173, 169, 5, 5, 167, 169, 5, 5, 174, 169, 5, 5, 175, 169, 5, 5, 176, 169, 5, 5, 177, 169, 5, 5, 178, 169, 5, 5, 179, 169, 5, 5, 180, 169, 5, 5, 181, 169, 5, 5, 182, 169, 5, 5, 183, 169, 5, 5, 184, 169, 5, 5, 185, 169, 5, 5, 186, 169, 5, 5, 187, 169, 5, 5, 188, 169, 5, 5, 189, 169, 5, 5, 190, 169, 5, 5, 191, 169, 5, 5, 192, 169, 5, 5, 193, 169, 5, 5, 194, 169, 5, 5, 195, 169, 5, 5, 196, 169, 5, 5, 23, 175, 5, 5, 24, 175, 5, 5, 25, 175, 5, 5, 26, 175, 5, 5, 27, 175, 5, 5, 28, 175, 5, 5, 29, 175, 5, 5, 30, 175, 5, 5, 31, 175, 5, 5, 32, 175, 5, 5, 33, 175, 5, 5, 34, 175, 5, 5, 35, 175, 5, 5, 36, 175, 5, 5, 37, 175, 5, 5, 38, 175, 5, 5, 39, 175, 5, 5, 40, 175, 5, 5, 41, 175, 5, 5, 42, 175, 5, 5, 43, 175, 5, 5, 44, 175, 5, 5, 45, 175, 5, 5, 46, 175, 5, 5, 47, 175, 5, 5, 48, 175, 5, 5, 49, 175, 5, 5, 50, 175, 5, 5, 51, 175, 5, 5, 52, 175, 5, 5, 53, 175, 5, 5, 54, 175, 5, 5, 55, 175, 5, 5, 56, 175, 5, 5, 57, 175, 5, 5, 58, 175, 5, 5, 59, 175, 5, 5, 143, 181, 5, 5, 144, 181, 5, 5, 145, 181, 5, 5, 146, 181, 5, 5, 147, 181, 5, 5, 148, 181, 5, 5, 149, 181, 5, 5, 150, 181, 5, 5, 151, 181, 5, 5, 152, 181, 5, 5, 153, 181, 5, 5, 154, 181, 5, 5, 155, 181, 5, 5, 156, 181, 5, 5, 157, 181, 5, 5, 158, 181, 5, 5, 159, 181, 5, 5, 160, 181, 5, 5, 161, 181, 5, 5, 162, 181, 5, 5, 163, 181, 5, 5, 164, 181, 5, 5, 165, 181, 5, 5, 166, 181, 5, 5, 167, 181, 5, 5, 168, 181, 5, 5, 169, 181, 5, 5, 170, 181, 5, 5, 171, 181, 5, 5, 172, 181, 5, 5, 173, 181, 5, 5, 174, 181, 5, 5, 175, 181, 5, 5, 176, 181, 5, 5, 177, 181, 5, 5, 178, 181, 5, 5, 179, 181, 5, 5, 180, 181, 5, 5, 181, 181, 5, 5, 182, 181, 5, 5, 183, 181, 5, 5, 184, 181, 5, 5, 185, 181, 5, 5, 84, 189, 5, 5, 85, 189, 5, 5, 86, 189, 5, 5, 87, 189, 5, 5, 88, 189, 5, 5, 89, 189, 5, 5, 90, 189, 5, 5, 91, 189, 5, 5, 92, 189, 5, 5, 93, 189, 5, 5, 94, 189, 5, 5, 95, 189, 5, 5, 60, 197, 5, 5, 96, 189, 5, 5, 97, 189, 5, 5, 98, 189, 5, 5, 99, 189, 5, 5, 100, 189, 5, 5, 101, 189, 5, 5, 102, 189, 5, 5, 103, 189, 5, 5, 104, 189, 5, 5, 105, 189, 5, 5, 106, 189, 5, 5, 107, 189, 5, 5, 108, 189, 5, 5, 109, 189, 5, 5, 110, 189, 5, 5, 111, 189, 5, 5, 112, 189, 5, 5, 113, 189, 5, 5, 114, 189, 5, 5, 115, 189, 5, 5, 116, 189, 5, 5, 117, 189, 5, 5, 118, 189, 5, 5, 119, 189, 5, 5, 128, 189, 5, 5, 120, 189, 5, 5, 121, 189, 5, 5, 122, 189, 5, 5, 123, 189, 5, 5, 61, 197, 5, 5, 62, 197, 5, 5, 63, 197, 5, 5, 64, 197, 5, 5, 65, 197, 5, 5, 66, 197, 5, 5, 67, 197, 5, 5, 68, 197, 5, 5, 69, 197, 5, 5, 70, 197, 5, 5, 71, 197, 5, 5, 72, 197, 5, 5, 73, 197, 5, 5, 74, 197, 5, 5, 75, 197, 5, 5, 76, 197, 5, 5, 77, 197, 5, 5, 78, 197, 5, 5, 79, 197, 5, 5, 80, 197, 5, 5, 81, 197, 5, 5, 82, 197, 5, 5, 83, 197, 5, 5, 84, 197, 5, 5, 85, 197, 5, 5, 86, 197, 5, 5, 87, 197, 5, 5, 88, 197, 5, 5, 89, 197, 5, 5, 90, 197, 5, 5, 91, 197, 5, 5, 92, 197, 5, 5, 93, 197, 5, 5, 94, 197, 5, 5, 95, 197, 5, 5, 96, 197, 5, 5, 97, 197, 5, 5, 98, 197, 5, 5, 99, 197, 5, 5, 100, 197, 5, 5, 248, 204, 5, 5, 249, 204, 5, 5, 250, 204, 5, 5, 251, 204, 5, 5, 252, 204, 5, 5, 253, 204, 5, 5, 254, 204, 5, 5, 255, 204, 5, 5, 2, 205, 5, 5, 3, 205, 5, 5, 4, 205, 5, 5, 5, 205, 5, 5, 6, 205, 5, 5, 7, 205, 5, 5, 8, 205, 5, 5, 9, 205, 5, 5, 10, 205, 5, 5, 11, 205, 5, 5, 12, 205, 5, 5, 13, 205, 5, 5, 167, 212, 5, 5, 14, 205, 5, 5, 15, 205, 5, 5, 16, 205, 5, 5, 17, 205, 5, 5, 18, 205, 5, 5, 168, 212, 5, 5, 19, 205, 5, 5, 20, 205, 5, 5, 21, 205, 5, 5, 22, 205, 5, 5, 23, 205, 5, 5, 24, 205, 5, 5, 25, 205, 5, 5, 26, 205, 5, 5, 27, 205, 5, 5, 28, 205, 5, 5, 29, 205, 5, 5, 30, 205, 5, 5, 31, 205, 5, 5, 32, 205, 5, 5, 33, 205, 5, 5, 34, 205, 5, 5, 35, 205, 5, 5, 36, 205, 5, 5, 101, 197, 5, 5, 37, 205, 5, 5, 38, 205, 5, 5, 39, 205, 5, 5, 40, 205, 5, 5, 41, 205, 5, 5, 42, 205, 5, 5, 43, 205, 5, 5, 44, 205, 5, 5, 45, 205, 5, 5, 46, 205, 5, 5, 169, 212, 5, 5, 47, 205, 5, 5, 170, 212, 5, 5, 171, 212, 5, 5, 172, 212, 5, 5, 173, 212, 5, 5, 174, 212, 5, 5, 175, 212, 5, 5, 176, 212, 5, 5, 177, 212, 5, 5, 178, 212, 5, 5, 179, 212, 5, 5, 180, 212, 5, 5, 181, 212, 5, 5, 182, 212, 5, 5, 183, 212, 5, 5, 184, 212, 5, 5, 185, 212, 5, 5, 186, 212, 5, 5, 187, 212, 5, 5, 188, 212, 5, 5, 189, 212, 5, 5, 190, 212, 5, 5, 191, 212, 5, 5, 192, 212, 5, 5, 193, 212, 5, 5, 194, 212, 5, 5, 195, 212, 5, 5, 196, 212, 5, 5, 197, 212, 5, 5, 198, 212, 5, 5, 199, 212, 5, 5, 200, 212, 5, 5, 138, 219, 5, 5, 139, 219, 5, 5, 140, 219, 5, 5, 141, 219, 5, 5, 142, 219, 5, 5, 143, 219, 5, 5, 144, 219, 5, 5, 145, 219, 5, 5, 146, 219, 5, 5, 147, 219, 5, 5, 148, 219, 5, 5, 149, 219, 5, 5, 137, 219, 5, 5, 150, 219, 5, 5, 151, 219, 5, 5, 152, 219, 5, 5, 153, 219, 5, 5, 154, 219, 5, 5, 155, 219, 5, 5, 156, 219, 5, 5, 157, 219, 5, 5, 158, 219, 5, 5, 159, 219, 5, 5, 160, 219, 5, 5, 161, 219, 5, 5, 162, 219, 5, 5, 163, 219, 5, 5, 164, 219, 5, 5, 24, 226, 5, 5, 25, 226, 5, 5, 26, 226, 5, 5, 27, 226, 5, 5, 28, 226, 5, 5, 29, 226, 5, 5, 30, 226, 5, 5, 31, 226, 5, 5, 32, 226, 5, 5, 33, 226, 5, 5, 34, 226, 5, 5, 35, 226, 5, 5, 36, 226, 5, 5, 37, 226, 5, 5, 38, 226, 5, 5, 39, 226, 5, 5, 40, 226, 5, 5, 41, 226, 5, 5, 42, 226, 5, 5, 43, 226, 5, 5, 44, 226, 5, 5, 45, 226, 5, 5, 46, 226, 5, 5, 47, 226, 5, 5, 48, 226, 5, 5, 49, 226, 5, 5, 50, 226, 5, 5, 51, 226, 5, 5, 52, 226, 5, 5, 53, 226, 5, 5, 54, 226, 5, 5, 55, 226, 5, 5, 56, 226, 5, 5, 57, 226, 5, 5, 58, 226, 5, 5, 59, 226, 5, 5, 60, 226, 5, 5, 61, 226, 5, 5, 62, 226, 5, 5, 63, 226, 5, 5, 64, 226, 5, 5, 116, 231, 5, 5, 117, 231, 5, 5, 118, 231, 5, 5, 119, 231, 5, 5, 120, 231, 5, 5, 121, 231, 5, 5, 122, 231, 5, 5, 123, 231, 5, 5, 124, 231, 5, 5, 125, 231, 5, 5, 126, 231, 5, 5, 127, 231, 5, 5, 128, 231, 5, 5, 129, 231, 5, 5, 130, 231, 5, 5, 131, 231, 5, 5, 132, 231, 5, 5, 133, 231, 5, 5, 134, 231, 5, 5, 135, 231, 5, 5, 136, 231, 5, 5, 137, 231, 5, 5, 138, 231, 5, 5, 139, 231, 5, 5, 140, 231, 5, 5, 237, 235, 5, 5, 238, 235, 5, 5, 239, 235, 5, 5, 240, 235, 5, 5, 241, 235, 5, 5, 242, 235, 5, 5, 243, 235, 5, 5, 244, 235, 5, 5, 245, 235, 5, 5, 246, 235, 5, 5, 133, 239, 5, 5, 247, 235, 5, 5, 134, 239, 5, 5, 135, 239, 5, 5, 136, 239, 5, 5, 137, 239, 5, 5, 138, 239, 5, 5, 139, 239, 5, 5, 140, 239, 5, 5, 146, 242, 5, 5, 141, 239, 5, 5, 142, 239, 5, 5, 143, 239, 5, 5, 147, 242, 5, 5, 148, 242, 5, 5, 149, 242, 5, 5, 150, 242, 5, 5, 151, 242, 5, 5, 152, 242, 5, 5, 144, 239, 5, 5, 153, 242, 5, 5, 154, 242, 5, 5, 155, 242, 5, 5, 234, 244, 5, 5, 235, 244, 5, 5, 236, 244, 5, 5, 163, 246, 5, 5, 239, 244, 5, 5, 164, 246, 5, 5, 165, 246, 5, 5, 166, 246, 5, 5, 226, 247, 5, 5, 227, 247, 5, 5, 208, 248, 5, 5, 120, 249, 5, 5, 121, 249, 5, 5, 122, 249, 5, 5, 91, 250, 5, 5, 121, 250, 5, 5, 165, 250, 5, 5, 89, 158, 5, 5, 91, 158, 5, 5, 202, 169, 5, 5, 203, 169, 5, 5, 63, 175, 5, 5, 64, 175, 5, 5, 65, 175, 5, 5, 189, 181, 5, 5, 114, 197, 5, 5, 208, 212, 5, 5, 176, 219, 5, 5, 146, 231, 5, 5, 92, 158, 5, 5, 142, 161, 5, 5, 204, 169, 5, 5, 190, 181, 5, 5, 58, 205, 5, 5, 93, 158, 5, 5, 67, 175, 5, 5, 129, 189, 5, 5, 209, 212, 5, 5, 95, 158, 5, 5, 205, 169, 5, 5, 68, 175, 5, 5, 191, 181, 5, 5, 59, 205, 5, 5, 210, 212, 5, 5, 177, 219, 5, 5, 147, 231, 5, 5, 96, 158, 5, 5, 206, 169, 5, 5, 69, 175, 5, 5, 70, 175, 5, 5, 115, 197, 5, 5, 116, 197, 5, 5, 117, 197, 5, 5, 60, 205, 5, 5, 61, 205, 5, 5, 62, 205, 5, 5, 63, 205, 5, 5, 64, 205, 5, 5, 211, 212, 5, 5, 212, 212, 5, 5, 178, 219, 5, 5, 179, 219, 5, 5, 180, 219, 5, 5, 148, 239, 5, 5, 97, 158, 5, 5, 119, 197, 5, 5, 100, 158, 5, 5, 101, 158, 5, 5, 143, 161, 5, 5, 144, 161, 5, 5, 145, 161, 5, 5, 236, 164, 5, 5, 237, 164, 5, 5, 238, 164, 5, 5, 239, 164, 5, 5, 240, 164, 5, 5, 207, 169, 5, 5, 208, 169, 5, 5, 209, 169, 5, 5, 210, 169, 5, 5, 211, 169, 5, 5, 212, 169, 5, 5, 213, 169, 5, 5, 214, 169, 5, 5, 71, 175, 5, 5, 72, 175, 5, 5, 73, 175, 5, 5, 74, 175, 5, 5, 75, 175, 5, 5, 76, 175, 5, 5, 77, 175, 5, 5, 78, 175, 5, 5, 79, 175, 5, 5, 194, 181, 5, 5, 195, 181, 5, 5, 196, 181, 5, 5, 197, 181, 5, 5, 198, 181, 5, 5, 130, 189, 5, 5, 131, 189, 5, 5, 132, 189, 5, 5, 133, 189, 5, 5, 134, 189, 5, 5, 120, 197, 5, 5, 135, 189, 5, 5, 121, 197, 5, 5, 122, 197, 5, 5, 123, 197, 5, 5, 124, 197, 5, 5, 125, 197, 5, 5, 126, 197, 5, 5, 127, 197, 5, 5, 128, 197, 5, 5, 129, 197, 5, 5, 130, 197, 5, 5, 65, 205, 5, 5, 66, 205, 5, 5, 67, 205, 5, 5, 68, 205, 5, 5, 69, 205, 5, 5, 70, 205, 5, 5, 214, 212, 5, 5, 215, 212, 5, 5, 216, 212, 5, 5, 217, 212, 5, 5, 218, 212, 5, 5, 219, 212, 5, 5, 181, 219, 5, 5, 182, 219, 5, 5, 183, 219, 5, 5, 184, 219, 5, 5, 75, 226, 5, 5, 76, 226, 5, 5, 77, 226, 5, 5, 78, 226, 5, 5, 148, 231, 5, 5, 149, 239, 5, 5, 150, 239, 5, 5, 151, 239, 5, 5, 152, 239, 5, 5, 153, 239, 5, 5, 154, 239, 5, 5, 160, 242, 5, 5, 161, 242, 5, 5, 167, 246, 5, 5, 123, 249, 5, 5, 33, 250, 5, 5, 102, 158, 5, 5, 150, 157, 5, 5, 141, 159, 5, 5, 142, 159, 5, 5, 143, 159, 5, 5, 146, 161, 5, 5, 147, 161, 5, 5, 148, 161, 5, 5, 149, 161, 5, 5, 150, 161, 5, 5, 242, 164, 5, 5, 151, 161, 5, 5, 152, 161, 5, 5, 243, 164, 5, 5, 244, 164, 5, 5, 245, 164, 5, 5, 246, 164, 5, 5, 247, 164, 5, 5, 248, 164, 5, 5, 249, 164, 5, 5, 217, 169, 5, 5, 250, 164, 5, 5, 251, 164, 5, 5, 252, 164, 5, 5, 253, 164, 5, 5, 254, 164, 5, 5, 255, 164, 5, 5, 2, 165, 5, 5, 3, 165, 5, 5, 218, 169, 5, 5, 80, 175, 5, 5, 219, 169, 5, 5, 220, 169, 5, 5, 221, 169, 5, 5, 222, 169, 5, 5, 223, 169, 5, 5, 224, 169, 5, 5, 225, 169, 5, 5, 226, 169, 5, 5, 227, 169, 5, 5, 228, 169, 5, 5, 229, 169, 5, 5, 230, 169, 5, 5, 231, 169, 5, 5, 232, 169, 5, 5, 233, 169, 5, 5, 234, 169, 5, 5, 235, 169, 5, 5, 236, 169, 5, 5, 237, 169, 5, 5, 238, 169, 5, 5, 81, 175, 5, 5, 82, 175, 5, 5, 83, 175, 5, 5, 84, 175, 5, 5, 85, 175, 5, 5, 86, 175, 5, 5, 87, 175, 5, 5, 88, 175, 5, 5, 89, 175, 5, 5, 90, 175, 5, 5, 91, 175, 5, 5, 92, 175, 5, 5, 93, 175, 5, 5, 94, 175, 5, 5, 95, 175, 5, 5, 96, 175, 5, 5, 97, 175, 5, 5, 98, 175, 5, 5, 99, 175, 5, 5, 100, 175, 5, 5, 199, 181, 5, 5, 200, 181, 5, 5, 201, 181, 5, 5, 202, 181, 5, 5, 203, 181, 5, 5, 204, 181, 5, 5, 205, 181, 5, 5, 206, 181, 5, 5, 207, 181, 5, 5, 208, 181, 5, 5, 209, 181, 5, 5, 210, 181, 5, 5, 141, 189, 5, 5, 211, 181, 5, 5, 212, 181, 5, 5, 213, 181, 5, 5, 214, 181, 5, 5, 142, 189, 5, 5, 143, 189, 5, 5, 133, 197, 5, 5, 144, 189, 5, 5, 145, 189, 5, 5, 146, 189, 5, 5, 147, 189, 5, 5, 134, 197, 5, 5, 135, 197, 5, 5, 148, 189, 5, 5, 149, 189, 5, 5, 150, 189, 5, 5, 151, 189, 5, 5, 152, 189, 5, 5, 136, 197, 5, 5, 153, 189, 5, 5, 154, 189, 5, 5, 155, 189, 5, 5, 156, 189, 5, 5, 157, 189, 5, 5, 158, 189, 5, 5, 159, 189, 5, 5, 160, 189, 5, 5, 161, 189, 5, 5, 162, 189, 5, 5, 163, 189, 5, 5, 164, 189, 5, 5, 165, 189, 5, 5, 166, 189, 5, 5, 167, 189, 5, 5, 138, 197, 5, 5, 139, 197, 5, 5, 140, 197, 5, 5, 141, 197, 5, 5, 142, 197, 5, 5, 143, 197, 5, 5, 144, 197, 5, 5, 145, 197, 5, 5, 168, 189, 5, 5, 137, 197, 5, 5, 146, 197, 5, 5, 147, 197, 5, 5, 73, 205, 5, 5, 148, 197, 5, 5, 149, 197, 5, 5, 150, 197, 5, 5, 151, 197, 5, 5, 152, 197, 5, 5, 153, 197, 5, 5, 154, 197, 5, 5, 155, 197, 5, 5, 74, 205, 5, 5, 156, 197, 5, 5, 157, 197, 5, 5, 76, 205, 5, 5, 77, 205, 5, 5, 78, 205, 5, 5, 79, 205, 5, 5, 80, 205, 5, 5, 81, 205, 5, 5, 82, 205, 5, 5, 75, 205, 5, 5, 83, 205, 5, 5, 220, 212, 5, 5, 221, 212, 5, 5, 84, 205, 5, 5, 85, 205, 5, 5, 86, 205, 5, 5, 87, 205, 5, 5, 88, 205, 5, 5, 89, 205, 5, 5, 187, 219, 5, 5, 222, 212, 5, 5, 223, 212, 5, 5, 186, 219, 5, 5, 224, 212, 5, 5, 225, 212, 5, 5, 226, 212, 5, 5, 227, 212, 5, 5, 228, 212, 5, 5, 229, 212, 5, 5, 230, 212, 5, 5, 188, 219, 5, 5, 189, 219, 5, 5, 190, 219, 5, 5, 191, 219, 5, 5, 192, 219, 5, 5, 193, 219, 5, 5, 194, 219, 5, 5, 195, 219, 5, 5, 196, 219, 5, 5, 197, 219, 5, 5, 198, 219, 5, 5, 199, 219, 5, 5, 200, 219, 5, 5, 79, 226, 5, 5, 201, 219, 5, 5, 80, 226, 5, 5, 81, 226, 5, 5, 82, 226, 5, 5, 83, 226, 5, 5, 84, 226, 5, 5, 85, 226, 5, 5, 86, 226, 5, 5, 87, 226, 5, 5, 88, 226, 5, 5, 150, 231, 5, 5, 151, 231, 5, 5, 152, 231, 5, 5, 153, 231, 5, 5, 154, 231, 5, 5, 155, 231, 5, 5, 156, 231, 5, 5, 252, 235, 5, 5, 253, 235, 5, 5, 254, 235, 5, 5, 155, 239, 5, 5, 156, 239, 5, 5, 157, 239, 5, 5, 163, 242, 5, 5, 164, 242, 5, 5, 165, 242, 5, 5, 240, 244, 5, 5, 168, 246, 5, 5, 169, 246, 5, 5, 232, 247, 5, 5, 233, 247, 5, 5, 234, 247, 5, 5, 144, 159, 5, 5, 102, 175, 5, 5, 216, 181, 5, 5, 169, 189, 5, 5, 170, 189, 5, 5, 145, 159, 5, 5, 146, 159, 5, 5, 103, 158, 5, 5, 147, 159, 5, 5, 148, 159, 5, 5, 154, 161, 5, 5, 155, 161, 5, 5, 156, 161, 5, 5, 157, 161, 5, 5, 5, 165, 5, 5, 6, 165, 5, 5, 7, 165, 5, 5, 8, 165, 5, 5, 9, 165, 5, 5, 10, 165, 5, 5, 11, 165, 5, 5, 12, 165, 5, 5, 13, 165, 5, 5, 14, 165, 5, 5, 241, 169, 5, 5, 242, 169, 5, 5, 243, 169, 5, 5, 244, 169, 5, 5, 245, 169, 5, 5, 246, 169, 5, 5, 247, 169, 5, 5, 248, 169, 5, 5, 249, 169, 5, 5, 250, 169, 5, 5, 251, 169, 5, 5, 252, 169, 5, 5, 253, 169, 5, 5, 254, 169, 5, 5, 255, 169, 5, 5, 2, 170, 5, 5, 3, 170, 5, 5, 4, 170, 5, 5, 5, 170, 5, 5, 6, 170, 5, 5, 7, 170, 5, 5, 8, 170, 5, 5, 104, 175, 5, 5, 105, 175, 5, 5, 106, 175, 5, 5, 107, 175, 5, 5, 108, 175, 5, 5, 109, 175, 5, 5, 110, 175, 5, 5, 111, 175, 5, 5, 217, 181, 5, 5, 112, 175, 5, 5, 218, 181, 5, 5, 113, 175, 5, 5, 114, 175, 5, 5, 115, 175, 5, 5, 116, 175, 5, 5, 117, 175, 5, 5, 118, 175, 5, 5, 119, 175, 5, 5, 120, 175, 5, 5, 121, 175, 5, 5, 122, 175, 5, 5, 123, 175, 5, 5, 124, 175, 5, 5, 125, 175, 5, 5, 126, 175, 5, 5, 127, 175, 5, 5, 128, 175, 5, 5, 129, 175, 5, 5, 130, 175, 5, 5, 131, 175, 5, 5, 132, 175, 5, 5, 133, 175, 5, 5, 220, 181, 5, 5, 221, 181, 5, 5, 222, 181, 5, 5, 223, 181, 5, 5, 224, 181, 5, 5, 225, 181, 5, 5, 226, 181, 5, 5, 227, 181, 5, 5, 228, 181, 5, 5, 229, 181, 5, 5, 230, 181, 5, 5, 231, 181, 5, 5, 232, 181, 5, 5, 233, 181, 5, 5, 234, 181, 5, 5, 235, 181, 5, 5, 236, 181, 5, 5, 237, 181, 5, 5, 238, 181, 5, 5, 239, 181, 5, 5, 240, 181, 5, 5, 241, 181, 5, 5, 242, 181, 5, 5, 243, 181, 5, 5, 244, 181, 5, 5, 245, 181, 5, 5, 246, 181, 5, 5, 247, 181, 5, 5, 248, 181, 5, 5, 249, 181, 5, 5, 250, 181, 5, 5, 251, 181, 5, 5, 252, 181, 5, 5, 171, 189, 5, 5, 172, 189, 5, 5, 173, 189, 5, 5, 174, 189, 5, 5, 159, 197, 5, 5, 175, 189, 5, 5, 9, 182, 5, 5, 176, 189, 5, 5, 177, 189, 5, 5, 178, 189, 5, 5, 179, 189, 5, 5, 180, 189, 5, 5, 181, 189, 5, 5, 182, 189, 5, 5, 183, 189, 5, 5, 184, 189, 5, 5, 185, 189, 5, 5, 186, 189, 5, 5, 187, 189, 5, 5, 188, 189, 5, 5, 189, 189, 5, 5, 190, 189, 5, 5, 253, 181, 5, 5, 191, 189, 5, 5, 192, 189, 5, 5, 193, 189, 5, 5, 194, 189, 5, 5, 195, 189, 5, 5, 196, 189, 5, 5, 197, 189, 5, 5, 198, 189, 5, 5, 199, 189, 5, 5, 200, 189, 5, 5, 160, 197, 5, 5, 161, 197, 5, 5, 162, 197, 5, 5, 163, 197, 5, 5, 164, 197, 5, 5, 165, 197, 5, 5, 166, 197, 5, 5, 167, 197, 5, 5, 168, 197, 5, 5, 169, 197, 5, 5, 94, 205, 5, 5, 170, 197, 5, 5, 171, 197, 5, 5, 172, 197, 5, 5, 173, 197, 5, 5, 174, 197, 5, 5, 175, 197, 5, 5, 176, 197, 5, 5, 177, 197, 5, 5, 92, 205, 5, 5, 178, 197, 5, 5, 179, 197, 5, 5, 180, 197, 5, 5, 181, 197, 5, 5, 182, 197, 5, 5, 183, 197, 5, 5, 184, 197, 5, 5, 185, 197, 5, 5, 186, 197, 5, 5, 187, 197, 5, 5, 188, 197, 5, 5, 189, 197, 5, 5, 190, 197, 5, 5, 191, 197, 5, 5, 192, 197, 5, 5, 193, 197, 5, 5, 194, 197, 5, 5, 195, 197, 5, 5, 196, 197, 5, 5, 197, 197, 5, 5, 198, 197, 5, 5, 95, 205, 5, 5, 96, 205, 5, 5, 97, 205, 5, 5, 98, 205, 5, 5, 99, 205, 5, 5, 100, 205, 5, 5, 101, 205, 5, 5, 102, 205, 5, 5, 103, 205, 5, 5, 104, 205, 5, 5, 105, 205, 5, 5, 106, 205, 5, 5, 107, 205, 5, 5, 108, 205, 5, 5, 109, 205, 5, 5, 110, 205, 5, 5, 111, 205, 5, 5, 112, 205, 5, 5, 113, 205, 5, 5, 114, 205, 5, 5, 115, 205, 5, 5, 116, 205, 5, 5, 117, 205, 5, 5, 118, 205, 5, 5, 119, 205, 5, 5, 120, 205, 5, 5, 121, 205, 5, 5, 122, 205, 5, 5, 123, 205, 5, 5, 124, 205, 5, 5, 125, 205, 5, 5, 126, 205, 5, 5, 127, 205, 5, 5, 128, 205, 5, 5, 129, 205, 5, 5, 231, 212, 5, 5, 232, 212, 5, 5, 233, 212, 5, 5, 234, 212, 5, 5, 235, 212, 5, 5, 236, 212, 5, 5, 237, 212, 5, 5, 238, 212, 5, 5, 239, 212, 5, 5, 203, 219, 5, 5, 240, 212, 5, 5, 241, 212, 5, 5, 204, 219, 5, 5, 242, 212, 5, 5, 243, 212, 5, 5, 145, 205, 5, 5, 244, 212, 5, 5, 245, 212, 5, 5, 246, 212, 5, 5, 247, 212, 5, 5, 248, 212, 5, 5, 249, 212, 5, 5, 250, 212, 5, 5, 251, 212, 5, 5, 252, 212, 5, 5, 205, 219, 5, 5, 206, 219, 5, 5, 207, 219, 5, 5, 208, 219, 5, 5, 209, 219, 5, 5, 210, 219, 5, 5, 89, 226, 5, 5, 211, 219, 5, 5, 212, 219, 5, 5, 213, 219, 5, 5, 214, 219, 5, 5, 215, 219, 5, 5, 216, 219, 5, 5, 217, 219, 5, 5, 218, 219, 5, 5, 219, 219, 5, 5, 220, 219, 5, 5, 221, 219, 5, 5, 222, 219, 5, 5, 223, 219, 5, 5, 90, 226, 5, 5, 224, 219, 5, 5, 91, 226, 5, 5, 157, 231, 5, 5, 92, 226, 5, 5, 93, 226, 5, 5, 225, 219, 5, 5, 94, 226, 5, 5, 95, 226, 5, 5, 96, 226, 5, 5, 158, 231, 5, 5, 97, 226, 5, 5, 98, 226, 5, 5, 99, 226, 5, 5, 100, 226, 5, 5, 101, 226, 5, 5, 102, 226, 5, 5, 103, 226, 5, 5, 104, 226, 5, 5, 105, 226, 5, 5, 106, 226, 5, 5, 121, 226, 5, 5, 107, 226, 5, 5, 108, 226, 5, 5, 160, 231, 5, 5, 161, 231, 5, 5, 255, 235, 5, 5, 162, 231, 5, 5, 163, 231, 5, 5, 164, 231, 5, 5, 165, 231, 5, 5, 166, 231, 5, 5, 167, 231, 5, 5, 168, 231, 5, 5, 169, 231, 5, 5, 170, 231, 5, 5, 171, 231, 5, 5, 172, 231, 5, 5, 173, 231, 5, 5, 174, 231, 5, 5, 2, 236, 5, 5, 3, 236, 5, 5, 158, 239, 5, 5, 4, 236, 5, 5, 5, 236, 5, 5, 166, 242, 5, 5, 6, 236, 5, 5, 7, 236, 5, 5, 159, 239, 5, 5, 8, 236, 5, 5, 9, 236, 5, 5, 10, 236, 5, 5, 11, 236, 5, 5, 12, 236, 5, 5, 160, 239, 5, 5, 161, 239, 5, 5, 162, 239, 5, 5, 163, 239, 5, 5, 164, 239, 5, 5, 165, 239, 5, 5, 166, 239, 5, 5, 167, 239, 5, 5, 168, 239, 5, 5, 167, 242, 5, 5, 168, 242, 5, 5, 169, 242, 5, 5, 170, 242, 5, 5, 171, 242, 5, 5, 172, 242, 5, 5, 173, 242, 5, 5, 241, 244, 5, 5, 242, 244, 5, 5, 170, 246, 5, 5, 243, 244, 5, 5, 171, 246, 5, 5, 172, 246, 5, 5, 173, 246, 5, 5, 235, 247, 5, 5, 209, 248, 5, 5, 149, 159, 5, 5, 14, 170, 5, 5, 10, 182, 5, 5, 11, 182, 5, 5, 209, 189, 5, 5, 146, 205, 5, 5, 122, 226, 5, 5, 177, 239, 5, 5, 174, 246, 5, 5, 210, 248, 5, 5, 150, 159, 5, 5, 20, 165, 5, 5, 15, 170, 5, 5, 16, 170, 5, 5, 140, 175, 5, 5, 141, 175, 5, 5, 142, 175, 5, 5, 143, 175, 5, 5, 144, 175, 5, 5, 145, 175, 5, 5, 146, 175, 5, 5, 147, 175, 5, 5, 148, 175, 5, 5, 12, 182, 5, 5, 13, 182, 5, 5, 14, 182, 5, 5, 211, 189, 5, 5, 212, 189, 5, 5, 213, 189, 5, 5, 221, 197, 5, 5, 222, 197, 5, 5, 223, 197, 5, 5, 224, 197, 5, 5, 147, 205, 5, 5, 148, 205, 5, 5, 149, 205, 5, 5, 11, 213, 5, 5, 150, 205, 5, 5, 12, 213, 5, 5, 13, 213, 5, 5, 14, 213, 5, 5, 15, 213, 5, 5, 16, 213, 5, 5, 243, 219, 5, 5, 244, 219, 5, 5, 245, 219, 5, 5, 124, 226, 5, 5, 125, 226, 5, 5, 126, 226, 5, 5, 127, 226, 5, 5, 128, 226, 5, 5, 183, 231, 5, 5, 184, 231, 5, 5, 185, 231, 5, 5, 186, 231, 5, 5, 22, 236, 5, 5, 23, 236, 5, 5, 24, 236, 5, 5, 178, 239, 5, 5, 246, 244, 5, 5, 151, 159, 5, 5, 17, 170, 5, 5, 149, 175, 5, 5, 215, 189, 5, 5, 216, 189, 5, 5, 151, 205, 5, 5, 152, 205, 5, 5, 152, 159, 5, 5, 150, 175, 5, 5, 15, 182, 5, 5, 217, 189, 5, 5, 218, 189, 5, 5, 225, 197, 5, 5, 226, 197, 5, 5, 227, 197, 5, 5, 18, 213, 5, 5, 153, 159, 5, 5, 154, 159, 5, 5, 161, 161, 5, 5, 21, 165, 5, 5, 22, 165, 5, 5, 151, 175, 5, 5, 152, 175, 5, 5, 228, 197, 5, 5, 155, 159, 5, 5, 156, 159, 5, 5, 157, 159, 5, 5, 158, 159, 5, 5, 159, 159, 5, 5, 160, 159, 5, 5, 162, 161, 5, 5, 23, 165, 5, 5, 24, 165, 5, 5, 25, 165, 5, 5, 26, 165, 5, 5, 18, 170, 5, 5, 27, 165, 5, 5, 19, 170, 5, 5, 20, 170, 5, 5, 21, 170, 5, 5, 22, 170, 5, 5, 23, 170, 5, 5, 24, 170, 5, 5, 25, 170, 5, 5, 26, 170, 5, 5, 27, 170, 5, 5, 153, 175, 5, 5, 154, 175, 5, 5, 155, 175, 5, 5, 156, 175, 5, 5, 157, 175, 5, 5, 158, 175, 5, 5, 159, 175, 5, 5, 160, 175, 5, 5, 161, 175, 5, 5, 162, 175, 5, 5, 163, 175, 5, 5, 164, 175, 5, 5, 165, 175, 5, 5, 166, 175, 5, 5, 16, 182, 5, 5, 17, 182, 5, 5, 18, 182, 5, 5, 19, 182, 5, 5, 20, 182, 5, 5, 21, 182, 5, 5, 22, 182, 5, 5, 23, 182, 5, 5, 24, 182, 5, 5, 25, 182, 5, 5, 26, 182, 5, 5, 27, 182, 5, 5, 53, 182, 5, 5, 219, 189, 5, 5, 220, 189, 5, 5, 221, 189, 5, 5, 222, 189, 5, 5, 223, 189, 5, 5, 224, 189, 5, 5, 225, 189, 5, 5, 144, 190, 5, 5, 226, 189, 5, 5, 230, 197, 5, 5, 231, 197, 5, 5, 232, 197, 5, 5, 233, 197, 5, 5, 234, 197, 5, 5, 235, 197, 5, 5, 227, 189, 5, 5, 239, 197, 5, 5, 236, 197, 5, 5, 237, 197, 5, 5, 238, 197, 5, 5, 153, 205, 5, 5, 154, 205, 5, 5, 155, 205, 5, 5, 156, 205, 5, 5, 157, 205, 5, 5, 158, 205, 5, 5, 19, 213, 5, 5, 20, 213, 5, 5, 21, 213, 5, 5, 248, 219, 5, 5, 249, 219, 5, 5, 129, 226, 5, 5, 130, 226, 5, 5, 131, 226, 5, 5, 188, 231, 5, 5, 189, 231, 5, 5, 25, 236, 5, 5, 179, 239, 5, 5, 180, 239, 5, 5, 177, 242, 5, 5, 178, 242, 5, 5, 175, 246, 5, 5, 161, 159, 5, 5, 28, 170, 5, 5, 28, 182, 5, 5, 240, 197, 5, 5, 241, 197, 5, 5, 22, 213, 5, 5, 23, 213, 5, 5, 164, 159, 5, 5, 29, 165, 5, 5, 30, 165, 5, 5, 31, 165, 5, 5, 32, 165, 5, 5, 33, 165, 5, 5, 29, 170, 5, 5, 30, 170, 5, 5, 31, 170, 5, 5, 32, 170, 5, 5, 33, 170, 5, 5, 34, 170, 5, 5, 35, 170, 5, 5, 36, 170, 5, 5, 37, 170, 5, 5, 38, 170, 5, 5, 169, 175, 5, 5, 170, 175, 5, 5, 171, 175, 5, 5, 172, 175, 5, 5, 173, 175, 5, 5, 174, 175, 5, 5, 175, 175, 5, 5, 176, 175, 5, 5, 177, 175, 5, 5, 178, 175, 5, 5, 179, 175, 5, 5, 180, 175, 5, 5, 181, 175, 5, 5, 182, 175, 5, 5, 29, 182, 5, 5, 30, 182, 5, 5, 31, 182, 5, 5, 32, 182, 5, 5, 33, 182, 5, 5, 228, 189, 5, 5, 34, 182, 5, 5, 35, 182, 5, 5, 36, 182, 5, 5, 37, 182, 5, 5, 183, 175, 5, 5, 38, 182, 5, 5, 39, 182, 5, 5, 40, 182, 5, 5, 41, 182, 5, 5, 42, 182, 5, 5, 43, 182, 5, 5, 44, 182, 5, 5, 45, 182, 5, 5, 46, 182, 5, 5, 47, 182, 5, 5, 48, 182, 5, 5, 49, 182, 5, 5, 50, 182, 5, 5, 51, 182, 5, 5, 52, 182, 5, 5, 229, 189, 5, 5, 230, 189, 5, 5, 231, 189, 5, 5, 232, 189, 5, 5, 233, 189, 5, 5, 234, 189, 5, 5, 235, 189, 5, 5, 236, 189, 5, 5, 237, 189, 5, 5, 238, 189, 5, 5, 239, 189, 5, 5, 240, 189, 5, 5, 241, 189, 5, 5, 242, 197, 5, 5, 243, 197, 5, 5, 244, 197, 5, 5, 245, 197, 5, 5, 246, 197, 5, 5, 247, 197, 5, 5, 248, 197, 5, 5, 249, 197, 5, 5, 250, 197, 5, 5, 251, 197, 5, 5, 252, 197, 5, 5, 253, 197, 5, 5, 254, 197, 5, 5, 255, 197, 5, 5, 2, 198, 5, 5, 3, 198, 5, 5, 4, 198, 5, 5, 5, 198, 5, 5, 6, 198, 5, 5, 7, 198, 5, 5, 8, 198, 5, 5, 162, 205, 5, 5, 163, 205, 5, 5, 164, 205, 5, 5, 165, 205, 5, 5, 166, 205, 5, 5, 167, 205, 5, 5, 168, 205, 5, 5, 169, 205, 5, 5, 170, 205, 5, 5, 171, 205, 5, 5, 172, 205, 5, 5, 173, 205, 5, 5, 174, 205, 5, 5, 175, 205, 5, 5, 176, 205, 5, 5, 177, 205, 5, 5, 178, 205, 5, 5, 179, 205, 5, 5, 180, 205, 5, 5, 181, 205, 5, 5, 182, 205, 5, 5, 183, 205, 5, 5, 184, 205, 5, 5, 185, 205, 5, 5, 186, 205, 5, 5, 187, 205, 5, 5, 188, 205, 5, 5, 24, 213, 5, 5, 25, 213, 5, 5, 26, 213, 5, 5, 27, 213, 5, 5, 28, 213, 5, 5, 29, 213, 5, 5, 30, 213, 5, 5, 31, 213, 5, 5, 191, 205, 5, 5, 192, 205, 5, 5, 32, 213, 5, 5, 33, 213, 5, 5, 34, 213, 5, 5, 35, 213, 5, 5, 36, 213, 5, 5, 37, 213, 5, 5, 38, 213, 5, 5, 39, 213, 5, 5, 251, 219, 5, 5, 252, 219, 5, 5, 253, 219, 5, 5, 254, 219, 5, 5, 255, 219, 5, 5, 2, 220, 5, 5, 250, 219, 5, 5, 3, 220, 5, 5, 4, 220, 5, 5, 5, 220, 5, 5, 6, 220, 5, 5, 7, 220, 5, 5, 8, 220, 5, 5, 9, 220, 5, 5, 40, 213, 5, 5, 10, 220, 5, 5, 11, 220, 5, 5, 12, 220, 5, 5, 13, 220, 5, 5, 133, 226, 5, 5, 134, 226, 5, 5, 135, 226, 5, 5, 136, 226, 5, 5, 137, 226, 5, 5, 138, 226, 5, 5, 139, 226, 5, 5, 140, 226, 5, 5, 141, 226, 5, 5, 142, 226, 5, 5, 143, 226, 5, 5, 144, 226, 5, 5, 145, 226, 5, 5, 146, 226, 5, 5, 147, 226, 5, 5, 148, 226, 5, 5, 149, 226, 5, 5, 150, 226, 5, 5, 151, 226, 5, 5, 152, 226, 5, 5, 191, 231, 5, 5, 192, 231, 5, 5, 193, 231, 5, 5, 194, 231, 5, 5, 195, 231, 5, 5, 196, 231, 5, 5, 197, 231, 5, 5, 198, 231, 5, 5, 199, 231, 5, 5, 200, 231, 5, 5, 153, 226, 5, 5, 201, 231, 5, 5, 202, 231, 5, 5, 203, 231, 5, 5, 204, 231, 5, 5, 26, 236, 5, 5, 27, 236, 5, 5, 28, 236, 5, 5, 29, 236, 5, 5, 30, 236, 5, 5, 31, 236, 5, 5, 32, 236, 5, 5, 33, 236, 5, 5, 34, 236, 5, 5, 35, 236, 5, 5, 36, 236, 5, 5, 37, 236, 5, 5, 38, 236, 5, 5, 39, 236, 5, 5, 40, 236, 5, 5, 41, 236, 5, 5, 181, 239, 5, 5, 182, 239, 5, 5, 183, 239, 5, 5, 179, 242, 5, 5, 184, 239, 5, 5, 42, 236, 5, 5, 180, 242, 5, 5, 181, 242, 5, 5, 247, 244, 5, 5, 248, 244, 5, 5, 249, 244, 5, 5, 250, 244, 5, 5, 251, 244, 5, 5, 176, 246, 5, 5, 177, 246, 5, 5, 178, 246, 5, 5, 237, 247, 5, 5, 238, 247, 5, 5, 211, 248, 5, 5, 212, 248, 5, 5, 230, 249, 5, 5, 93, 250, 5, 5, 138, 250, 5, 5, 165, 159, 5, 5, 39, 170, 5, 5, 185, 175, 5, 5, 186, 175, 5, 5, 187, 175, 5, 5, 12, 198, 5, 5, 13, 198, 5, 5, 166, 159, 5, 5, 163, 161, 5, 5, 164, 161, 5, 5, 34, 165, 5, 5, 35, 165, 5, 5, 36, 165, 5, 5, 37, 165, 5, 5, 40, 170, 5, 5, 188, 175, 5, 5, 189, 175, 5, 5, 190, 175, 5, 5, 191, 175, 5, 5, 242, 189, 5, 5, 54, 182, 5, 5, 55, 182, 5, 5, 56, 182, 5, 5, 57, 182, 5, 5, 243, 189, 5, 5, 244, 189, 5, 5, 245, 189, 5, 5, 246, 189, 5, 5, 14, 198, 5, 5, 15, 198, 5, 5, 16, 198, 5, 5, 17, 198, 5, 5, 18, 198, 5, 5, 193, 205, 5, 5, 194, 205, 5, 5, 195, 205, 5, 5, 18, 220, 5, 5, 19, 220, 5, 5, 20, 220, 5, 5, 21, 220, 5, 5, 22, 220, 5, 5, 155, 226, 5, 5, 156, 226, 5, 5, 157, 226, 5, 5, 207, 231, 5, 5, 208, 231, 5, 5, 209, 231, 5, 5, 210, 231, 5, 5, 43, 236, 5, 5, 44, 236, 5, 5, 45, 236, 5, 5, 185, 239, 5, 5, 182, 242, 5, 5, 183, 242, 5, 5, 253, 244, 5, 5, 180, 246, 5, 5, 167, 159, 5, 5, 41, 170, 5, 5, 59, 182, 5, 5, 60, 182, 5, 5, 247, 189, 5, 5, 19, 198, 5, 5, 20, 198, 5, 5, 198, 205, 5, 5, 41, 213, 5, 5, 42, 213, 5, 5, 43, 213, 5, 5, 44, 213, 5, 5, 26, 220, 5, 5, 158, 226, 5, 5, 211, 231, 5, 5, 47, 236, 5, 5, 184, 242, 5, 5, 168, 159, 5, 5, 38, 165, 5, 5, 39, 165, 5, 5, 42, 170, 5, 5, 192, 175, 5, 5, 193, 175, 5, 5, 194, 175, 5, 5, 195, 175, 5, 5, 196, 175, 5, 5, 197, 175, 5, 5, 62, 182, 5, 5, 63, 182, 5, 5, 64, 182, 5, 5, 65, 182, 5, 5, 66, 182, 5, 5, 67, 182, 5, 5, 68, 182, 5, 5, 69, 182, 5, 5, 70, 182, 5, 5, 248, 189, 5, 5, 249, 189, 5, 5, 250, 189, 5, 5, 251, 189, 5, 5, 252, 189, 5, 5, 253, 189, 5, 5, 254, 189, 5, 5, 21, 198, 5, 5, 22, 198, 5, 5, 255, 189, 5, 5, 23, 198, 5, 5, 199, 205, 5, 5, 200, 205, 5, 5, 201, 205, 5, 5, 45, 213, 5, 5, 46, 213, 5, 5, 47, 213, 5, 5, 48, 213, 5, 5, 27, 220, 5, 5, 159, 226, 5, 5, 160, 226, 5, 5, 161, 226, 5, 5, 212, 231, 5, 5, 213, 231, 5, 5, 214, 231, 5, 5, 49, 236, 5, 5, 50, 236, 5, 5, 186, 242, 5, 5, 169, 159, 5, 5, 40, 165, 5, 5, 43, 170, 5, 5, 44, 170, 5, 5, 45, 170, 5, 5, 46, 170, 5, 5, 47, 170, 5, 5, 48, 170, 5, 5, 198, 175, 5, 5, 199, 175, 5, 5, 200, 175, 5, 5, 201, 175, 5, 5, 202, 175, 5, 5, 203, 175, 5, 5, 204, 175, 5, 5, 205, 175, 5, 5, 206, 175, 5, 5, 207, 175, 5, 5, 208, 175, 5, 5, 209, 175, 5, 5, 210, 175, 5, 5, 211, 175, 5, 5, 212, 175, 5, 5, 213, 175, 5, 5, 214, 175, 5, 5, 215, 175, 5, 5, 216, 175, 5, 5, 217, 175, 5, 5, 218, 175, 5, 5, 219, 175, 5, 5, 220, 175, 5, 5, 221, 175, 5, 5, 73, 182, 5, 5, 74, 182, 5, 5, 75, 182, 5, 5, 76, 182, 5, 5, 77, 182, 5, 5, 78, 182, 5, 5, 79, 182, 5, 5, 80, 182, 5, 5, 81, 182, 5, 5, 82, 182, 5, 5, 83, 182, 5, 5, 84, 182, 5, 5, 85, 182, 5, 5, 86, 182, 5, 5, 87, 182, 5, 5, 88, 182, 5, 5, 89, 182, 5, 5, 90, 182, 5, 5, 91, 182, 5, 5, 92, 182, 5, 5, 93, 182, 5, 5, 94, 182, 5, 5, 95, 182, 5, 5, 4, 190, 5, 5, 5, 190, 5, 5, 96, 182, 5, 5, 97, 182, 5, 5, 98, 182, 5, 5, 99, 182, 5, 5, 100, 182, 5, 5, 101, 182, 5, 5, 6, 190, 5, 5, 7, 190, 5, 5, 8, 190, 5, 5, 9, 190, 5, 5, 10, 190, 5, 5, 11, 190, 5, 5, 12, 190, 5, 5, 13, 190, 5, 5, 14, 190, 5, 5, 15, 190, 5, 5, 16, 190, 5, 5, 17, 190, 5, 5, 18, 190, 5, 5, 19, 190, 5, 5, 20, 190, 5, 5, 21, 190, 5, 5, 22, 190, 5, 5, 23, 190, 5, 5, 102, 182, 5, 5, 27, 190, 5, 5, 24, 190, 5, 5, 25, 198, 5, 5, 26, 198, 5, 5, 27, 198, 5, 5, 28, 198, 5, 5, 29, 198, 5, 5, 30, 198, 5, 5, 31, 198, 5, 5, 32, 198, 5, 5, 33, 198, 5, 5, 34, 198, 5, 5, 35, 198, 5, 5, 36, 198, 5, 5, 37, 198, 5, 5, 38, 198, 5, 5, 39, 198, 5, 5, 40, 198, 5, 5, 203, 205, 5, 5, 204, 205, 5, 5, 205, 205, 5, 5, 206, 205, 5, 5, 207, 205, 5, 5, 208, 205, 5, 5, 209, 205, 5, 5, 210, 205, 5, 5, 211, 205, 5, 5, 212, 205, 5, 5, 213, 205, 5, 5, 214, 205, 5, 5, 215, 205, 5, 5, 216, 205, 5, 5, 217, 205, 5, 5, 235, 205, 5, 5, 218, 205, 5, 5, 219, 205, 5, 5, 220, 205, 5, 5, 221, 205, 5, 5, 222, 205, 5, 5, 223, 205, 5, 5, 224, 205, 5, 5, 225, 205, 5, 5, 226, 205, 5, 5, 227, 205, 5, 5, 228, 205, 5, 5, 229, 205, 5, 5, 49, 213, 5, 5, 50, 213, 5, 5, 51, 213, 5, 5, 52, 213, 5, 5, 53, 213, 5, 5, 54, 213, 5, 5, 55, 213, 5, 5, 56, 213, 5, 5, 57, 213, 5, 5, 58, 213, 5, 5, 59, 213, 5, 5, 236, 205, 5, 5, 60, 213, 5, 5, 61, 213, 5, 5, 62, 213, 5, 5, 63, 213, 5, 5, 64, 213, 5, 5, 65, 213, 5, 5, 66, 213, 5, 5, 67, 213, 5, 5, 68, 213, 5, 5, 69, 213, 5, 5, 70, 213, 5, 5, 71, 213, 5, 5, 72, 213, 5, 5, 28, 220, 5, 5, 29, 220, 5, 5, 30, 220, 5, 5, 31, 220, 5, 5, 32, 220, 5, 5, 33, 220, 5, 5, 34, 220, 5, 5, 35, 220, 5, 5, 36, 220, 5, 5, 37, 220, 5, 5, 38, 220, 5, 5, 39, 220, 5, 5, 40, 220, 5, 5, 163, 226, 5, 5, 164, 226, 5, 5, 165, 226, 5, 5, 166, 226, 5, 5, 167, 226, 5, 5, 168, 226, 5, 5, 169, 226, 5, 5, 170, 226, 5, 5, 171, 226, 5, 5, 172, 226, 5, 5, 173, 226, 5, 5, 174, 226, 5, 5, 175, 226, 5, 5, 176, 226, 5, 5, 177, 226, 5, 5, 178, 226, 5, 5, 216, 231, 5, 5, 183, 226, 5, 5, 217, 231, 5, 5, 218, 231, 5, 5, 219, 231, 5, 5, 220, 231, 5, 5, 221, 231, 5, 5, 222, 231, 5, 5, 223, 231, 5, 5, 224, 231, 5, 5, 225, 231, 5, 5, 226, 231, 5, 5, 227, 231, 5, 5, 228, 231, 5, 5, 229, 231, 5, 5, 230, 231, 5, 5, 231, 231, 5, 5, 232, 231, 5, 5, 233, 231, 5, 5, 234, 231, 5, 5, 51, 236, 5, 5, 52, 236, 5, 5, 53, 236, 5, 5, 54, 236, 5, 5, 55, 236, 5, 5, 56, 236, 5, 5, 57, 236, 5, 5, 58, 236, 5, 5, 59, 236, 5, 5, 60, 236, 5, 5, 61, 236, 5, 5, 187, 239, 5, 5, 188, 239, 5, 5, 189, 239, 5, 5, 190, 239, 5, 5, 191, 239, 5, 5, 192, 239, 5, 5, 193, 239, 5, 5, 194, 239, 5, 5, 187, 242, 5, 5, 188, 242, 5, 5, 189, 242, 5, 5, 190, 242, 5, 5, 191, 242, 5, 5, 255, 244, 5, 5, 2, 245, 5, 5, 3, 245, 5, 5, 4, 245, 5, 5, 239, 247, 5, 5, 214, 248, 5, 5, 216, 248, 5, 5, 215, 248, 5, 5, 125, 249, 5, 5, 126, 249, 5, 5, 231, 249, 5, 5, 170, 159, 5, 5, 224, 175, 5, 5, 105, 182, 5, 5, 43, 198, 5, 5, 44, 198, 5, 5, 237, 205, 5, 5, 127, 249, 5, 5, 171, 159, 5, 5, 41, 165, 5, 5, 49, 170, 5, 5, 50, 170, 5, 5, 225, 175, 5, 5, 226, 175, 5, 5, 227, 175, 5, 5, 106, 182, 5, 5, 28, 190, 5, 5, 29, 190, 5, 5, 45, 198, 5, 5, 46, 198, 5, 5, 238, 205, 5, 5, 238, 231, 5, 5, 239, 231, 5, 5, 198, 239, 5, 5, 192, 242, 5, 5, 172, 159, 5, 5, 42, 165, 5, 5, 43, 165, 5, 5, 44, 165, 5, 5, 51, 170, 5, 5, 52, 170, 5, 5, 53, 170, 5, 5, 54, 170, 5, 5, 55, 170, 5, 5, 56, 170, 5, 5, 57, 170, 5, 5, 58, 170, 5, 5, 59, 170, 5, 5, 60, 170, 5, 5, 61, 170, 5, 5, 229, 175, 5, 5, 230, 175, 5, 5, 231, 175, 5, 5, 232, 175, 5, 5, 233, 175, 5, 5, 234, 175, 5, 5, 235, 175, 5, 5, 236, 175, 5, 5, 237, 175, 5, 5, 238, 175, 5, 5, 239, 175, 5, 5, 240, 175, 5, 5, 241, 175, 5, 5, 242, 175, 5, 5, 243, 175, 5, 5, 244, 175, 5, 5, 245, 175, 5, 5, 246, 175, 5, 5, 247, 175, 5, 5, 248, 175, 5, 5, 249, 175, 5, 5, 250, 175, 5, 5, 251, 175, 5, 5, 252, 175, 5, 5, 253, 175, 5, 5, 254, 175, 5, 5, 255, 175, 5, 5, 107, 182, 5, 5, 108, 182, 5, 5, 109, 182, 5, 5, 110, 182, 5, 5, 111, 182, 5, 5, 112, 182, 5, 5, 113, 182, 5, 5, 114, 182, 5, 5, 115, 182, 5, 5, 30, 190, 5, 5, 116, 182, 5, 5, 117, 182, 5, 5, 118, 182, 5, 5, 119, 182, 5, 5, 120, 182, 5, 5, 121, 182, 5, 5, 122, 182, 5, 5, 123, 182, 5, 5, 124, 182, 5, 5, 125, 182, 5, 5, 126, 182, 5, 5, 127, 182, 5, 5, 128, 182, 5, 5, 129, 182, 5, 5, 130, 182, 5, 5, 131, 182, 5, 5, 132, 182, 5, 5, 133, 182, 5, 5, 134, 182, 5, 5, 135, 182, 5, 5, 136, 182, 5, 5, 137, 182, 5, 5, 138, 182, 5, 5, 139, 182, 5, 5, 140, 182, 5, 5, 141, 182, 5, 5, 142, 182, 5, 5, 31, 190, 5, 5, 32, 190, 5, 5, 33, 190, 5, 5, 34, 190, 5, 5, 35, 190, 5, 5, 36, 190, 5, 5, 37, 190, 5, 5, 38, 190, 5, 5, 39, 190, 5, 5, 40, 190, 5, 5, 41, 190, 5, 5, 42, 190, 5, 5, 43, 190, 5, 5, 44, 190, 5, 5, 45, 190, 5, 5, 46, 190, 5, 5, 47, 190, 5, 5, 48, 190, 5, 5, 49, 190, 5, 5, 50, 190, 5, 5, 51, 190, 5, 5, 52, 190, 5, 5, 53, 190, 5, 5, 54, 190, 5, 5, 55, 190, 5, 5, 56, 190, 5, 5, 47, 198, 5, 5, 48, 198, 5, 5, 49, 198, 5, 5, 50, 198, 5, 5, 51, 198, 5, 5, 52, 198, 5, 5, 53, 198, 5, 5, 54, 198, 5, 5, 55, 198, 5, 5, 56, 198, 5, 5, 57, 198, 5, 5, 58, 198, 5, 5, 59, 198, 5, 5, 60, 198, 5, 5, 61, 198, 5, 5, 62, 198, 5, 5, 63, 198, 5, 5, 64, 198, 5, 5, 65, 198, 5, 5, 66, 198, 5, 5, 67, 198, 5, 5, 68, 198, 5, 5, 69, 198, 5, 5, 70, 198, 5, 5, 71, 198, 5, 5, 72, 198, 5, 5, 73, 198, 5, 5, 74, 198, 5, 5, 239, 205, 5, 5, 240, 205, 5, 5, 241, 205, 5, 5, 242, 205, 5, 5, 243, 205, 5, 5, 244, 205, 5, 5, 76, 213, 5, 5, 245, 205, 5, 5, 246, 205, 5, 5, 247, 205, 5, 5, 248, 205, 5, 5, 249, 205, 5, 5, 250, 205, 5, 5, 251, 205, 5, 5, 252, 205, 5, 5, 253, 205, 5, 5, 254, 205, 5, 5, 255, 205, 5, 5, 2, 206, 5, 5, 3, 206, 5, 5, 4, 206, 5, 5, 5, 206, 5, 5, 6, 206, 5, 5, 7, 206, 5, 5, 8, 206, 5, 5, 9, 206, 5, 5, 10, 206, 5, 5, 11, 206, 5, 5, 12, 206, 5, 5, 13, 206, 5, 5, 14, 206, 5, 5, 15, 206, 5, 5, 16, 206, 5, 5, 17, 206, 5, 5, 18, 206, 5, 5, 19, 206, 5, 5, 20, 206, 5, 5, 77, 213, 5, 5, 78, 213, 5, 5, 79, 213, 5, 5, 80, 213, 5, 5, 81, 213, 5, 5, 82, 213, 5, 5, 83, 213, 5, 5, 84, 213, 5, 5, 85, 213, 5, 5, 86, 213, 5, 5, 87, 213, 5, 5, 88, 213, 5, 5, 89, 213, 5, 5, 90, 213, 5, 5, 91, 213, 5, 5, 92, 213, 5, 5, 93, 213, 5, 5, 94, 213, 5, 5, 95, 213, 5, 5, 21, 206, 5, 5, 96, 213, 5, 5, 97, 213, 5, 5, 98, 213, 5, 5, 99, 213, 5, 5, 100, 213, 5, 5, 101, 213, 5, 5, 102, 213, 5, 5, 103, 213, 5, 5, 104, 213, 5, 5, 43, 220, 5, 5, 44, 220, 5, 5, 45, 220, 5, 5, 46, 220, 5, 5, 47, 220, 5, 5, 48, 220, 5, 5, 49, 220, 5, 5, 105, 213, 5, 5, 50, 220, 5, 5, 51, 220, 5, 5, 52, 220, 5, 5, 53, 220, 5, 5, 54, 220, 5, 5, 55, 220, 5, 5, 56, 220, 5, 5, 57, 220, 5, 5, 58, 220, 5, 5, 59, 220, 5, 5, 60, 220, 5, 5, 61, 220, 5, 5, 62, 220, 5, 5, 63, 220, 5, 5, 64, 220, 5, 5, 65, 220, 5, 5, 66, 220, 5, 5, 67, 220, 5, 5, 68, 220, 5, 5, 69, 220, 5, 5, 184, 226, 5, 5, 70, 220, 5, 5, 71, 220, 5, 5, 72, 220, 5, 5, 186, 226, 5, 5, 187, 226, 5, 5, 185, 226, 5, 5, 188, 226, 5, 5, 189, 226, 5, 5, 190, 226, 5, 5, 191, 226, 5, 5, 192, 226, 5, 5, 193, 226, 5, 5, 194, 226, 5, 5, 73, 220, 5, 5, 195, 226, 5, 5, 196, 226, 5, 5, 197, 226, 5, 5, 198, 226, 5, 5, 199, 226, 5, 5, 200, 226, 5, 5, 201, 226, 5, 5, 202, 226, 5, 5, 203, 226, 5, 5, 204, 226, 5, 5, 240, 231, 5, 5, 241, 231, 5, 5, 242, 231, 5, 5, 243, 231, 5, 5, 244, 231, 5, 5, 245, 231, 5, 5, 246, 231, 5, 5, 247, 231, 5, 5, 248, 231, 5, 5, 249, 231, 5, 5, 250, 231, 5, 5, 251, 231, 5, 5, 252, 231, 5, 5, 253, 231, 5, 5, 254, 231, 5, 5, 255, 231, 5, 5, 2, 232, 5, 5, 3, 232, 5, 5, 4, 232, 5, 5, 5, 232, 5, 5, 6, 232, 5, 5, 7, 232, 5, 5, 8, 232, 5, 5, 63, 236, 5, 5, 64, 236, 5, 5, 65, 236, 5, 5, 66, 236, 5, 5, 67, 236, 5, 5, 68, 236, 5, 5, 69, 236, 5, 5, 70, 236, 5, 5, 71, 236, 5, 5, 72, 236, 5, 5, 73, 236, 5, 5, 74, 236, 5, 5, 75, 236, 5, 5, 76, 236, 5, 5, 77, 236, 5, 5, 78, 236, 5, 5, 79, 236, 5, 5, 199, 239, 5, 5, 200, 239, 5, 5, 201, 239, 5, 5, 202, 239, 5, 5, 203, 239, 5, 5, 204, 239, 5, 5, 205, 239, 5, 5, 206, 239, 5, 5, 207, 239, 5, 5, 208, 239, 5, 5, 209, 239, 5, 5, 193, 242, 5, 5, 194, 242, 5, 5, 195, 242, 5, 5, 196, 242, 5, 5, 197, 242, 5, 5, 198, 242, 5, 5, 199, 242, 5, 5, 200, 242, 5, 5, 201, 242, 5, 5, 202, 242, 5, 5, 203, 242, 5, 5, 5, 245, 5, 5, 6, 245, 5, 5, 7, 245, 5, 5, 8, 245, 5, 5, 9, 245, 5, 5, 10, 245, 5, 5, 11, 245, 5, 5, 12, 245, 5, 5, 181, 246, 5, 5, 240, 247, 5, 5, 241, 247, 5, 5, 217, 248, 5, 5, 128, 249, 5, 5, 173, 159, 5, 5, 105, 158, 5, 5, 166, 161, 5, 5, 46, 165, 5, 5, 63, 170, 5, 5, 64, 170, 5, 5, 65, 170, 5, 5, 66, 170, 5, 5, 67, 170, 5, 5, 68, 170, 5, 5, 5, 176, 5, 5, 6, 176, 5, 5, 7, 176, 5, 5, 8, 176, 5, 5, 9, 176, 5, 5, 10, 176, 5, 5, 11, 176, 5, 5, 12, 176, 5, 5, 13, 176, 5, 5, 14, 176, 5, 5, 15, 176, 5, 5, 147, 182, 5, 5, 148, 182, 5, 5, 149, 182, 5, 5, 150, 182, 5, 5, 151, 182, 5, 5, 152, 182, 5, 5, 153, 182, 5, 5, 154, 182, 5, 5, 155, 182, 5, 5, 156, 182, 5, 5, 157, 182, 5, 5, 158, 182, 5, 5, 159, 182, 5, 5, 160, 182, 5, 5, 161, 182, 5, 5, 162, 182, 5, 5, 163, 182, 5, 5, 164, 182, 5, 5, 58, 190, 5, 5, 165, 182, 5, 5, 59, 190, 5, 5, 60, 190, 5, 5, 61, 190, 5, 5, 76, 198, 5, 5, 62, 190, 5, 5, 63, 190, 5, 5, 64, 190, 5, 5, 65, 190, 5, 5, 66, 190, 5, 5, 67, 190, 5, 5, 68, 190, 5, 5, 69, 190, 5, 5, 70, 190, 5, 5, 77, 198, 5, 5, 78, 198, 5, 5, 79, 198, 5, 5, 80, 198, 5, 5, 81, 198, 5, 5, 82, 198, 5, 5, 83, 198, 5, 5, 84, 198, 5, 5, 85, 198, 5, 5, 24, 206, 5, 5, 25, 206, 5, 5, 26, 206, 5, 5, 27, 206, 5, 5, 28, 206, 5, 5, 29, 206, 5, 5, 30, 206, 5, 5, 31, 206, 5, 5, 32, 206, 5, 5, 33, 206, 5, 5, 34, 206, 5, 5, 86, 198, 5, 5, 35, 206, 5, 5, 36, 206, 5, 5, 107, 213, 5, 5, 108, 213, 5, 5, 109, 213, 5, 5, 110, 213, 5, 5, 111, 213, 5, 5, 112, 213, 5, 5, 113, 213, 5, 5, 114, 213, 5, 5, 115, 213, 5, 5, 116, 213, 5, 5, 117, 213, 5, 5, 118, 213, 5, 5, 119, 213, 5, 5, 120, 213, 5, 5, 121, 213, 5, 5, 122, 213, 5, 5, 123, 213, 5, 5, 124, 213, 5, 5, 125, 213, 5, 5, 77, 220, 5, 5, 78, 220, 5, 5, 79, 220, 5, 5, 80, 220, 5, 5, 81, 220, 5, 5, 82, 220, 5, 5, 83, 220, 5, 5, 84, 220, 5, 5, 85, 220, 5, 5, 86, 220, 5, 5, 213, 226, 5, 5, 214, 226, 5, 5, 215, 226, 5, 5, 13, 232, 5, 5, 14, 232, 5, 5, 216, 226, 5, 5, 15, 232, 5, 5, 16, 232, 5, 5, 80, 236, 5, 5, 81, 236, 5, 5, 82, 236, 5, 5, 83, 236, 5, 5, 211, 239, 5, 5, 212, 239, 5, 5, 204, 242, 5, 5, 184, 246, 5, 5, 185, 246, 5, 5, 242, 247, 5, 5, 219, 248, 5, 5, 220, 248, 5, 5, 175, 159, 5, 5, 19, 176, 5, 5, 20, 176, 5, 5, 74, 190, 5, 5, 89, 198, 5, 5, 38, 206, 5, 5, 176, 159, 5, 5, 47, 165, 5, 5, 48, 165, 5, 5, 49, 165, 5, 5, 50, 165, 5, 5, 51, 165, 5, 5, 69, 170, 5, 5, 70, 170, 5, 5, 71, 170, 5, 5, 72, 170, 5, 5, 73, 170, 5, 5, 74, 170, 5, 5, 75, 170, 5, 5, 21, 176, 5, 5, 22, 176, 5, 5, 23, 176, 5, 5, 24, 176, 5, 5, 25, 176, 5, 5, 26, 176, 5, 5, 27, 176, 5, 5, 28, 176, 5, 5, 29, 176, 5, 5, 30, 176, 5, 5, 31, 176, 5, 5, 32, 176, 5, 5, 33, 176, 5, 5, 170, 182, 5, 5, 171, 182, 5, 5, 172, 182, 5, 5, 173, 182, 5, 5, 174, 182, 5, 5, 175, 182, 5, 5, 176, 182, 5, 5, 177, 182, 5, 5, 178, 182, 5, 5, 179, 182, 5, 5, 180, 182, 5, 5, 181, 182, 5, 5, 182, 182, 5, 5, 183, 182, 5, 5, 184, 182, 5, 5, 185, 182, 5, 5, 186, 182, 5, 5, 187, 182, 5, 5, 188, 182, 5, 5, 189, 182, 5, 5, 190, 182, 5, 5, 191, 182, 5, 5, 192, 182, 5, 5, 193, 182, 5, 5, 194, 182, 5, 5, 75, 190, 5, 5, 76, 190, 5, 5, 77, 190, 5, 5, 78, 190, 5, 5, 79, 190, 5, 5, 80, 190, 5, 5, 81, 190, 5, 5, 82, 190, 5, 5, 83, 190, 5, 5, 84, 190, 5, 5, 85, 190, 5, 5, 86, 190, 5, 5, 87, 190, 5, 5, 88, 190, 5, 5, 90, 198, 5, 5, 91, 198, 5, 5, 92, 198, 5, 5, 93, 198, 5, 5, 94, 198, 5, 5, 95, 198, 5, 5, 96, 198, 5, 5, 89, 190, 5, 5, 97, 198, 5, 5, 98, 198, 5, 5, 99, 198, 5, 5, 100, 198, 5, 5, 101, 198, 5, 5, 102, 198, 5, 5, 103, 198, 5, 5, 104, 198, 5, 5, 40, 206, 5, 5, 41, 206, 5, 5, 42, 206, 5, 5, 43, 206, 5, 5, 44, 206, 5, 5, 45, 206, 5, 5, 46, 206, 5, 5, 47, 206, 5, 5, 48, 206, 5, 5, 49, 206, 5, 5, 50, 206, 5, 5, 51, 206, 5, 5, 52, 206, 5, 5, 53, 206, 5, 5, 54, 206, 5, 5, 55, 206, 5, 5, 56, 206, 5, 5, 57, 206, 5, 5, 58, 206, 5, 5, 59, 206, 5, 5, 60, 206, 5, 5, 61, 206, 5, 5, 62, 206, 5, 5, 126, 213, 5, 5, 127, 213, 5, 5, 128, 213, 5, 5, 129, 213, 5, 5, 130, 213, 5, 5, 131, 213, 5, 5, 132, 213, 5, 5, 133, 213, 5, 5, 134, 213, 5, 5, 135, 213, 5, 5, 136, 213, 5, 5, 137, 213, 5, 5, 138, 213, 5, 5, 139, 213, 5, 5, 88, 220, 5, 5, 142, 213, 5, 5, 89, 220, 5, 5, 90, 220, 5, 5, 91, 220, 5, 5, 92, 220, 5, 5, 93, 220, 5, 5, 94, 220, 5, 5, 95, 220, 5, 5, 96, 220, 5, 5, 97, 220, 5, 5, 98, 220, 5, 5, 99, 220, 5, 5, 100, 220, 5, 5, 101, 220, 5, 5, 102, 220, 5, 5, 218, 226, 5, 5, 219, 226, 5, 5, 220, 226, 5, 5, 221, 226, 5, 5, 222, 226, 5, 5, 18, 232, 5, 5, 140, 213, 5, 5, 223, 226, 5, 5, 224, 226, 5, 5, 225, 226, 5, 5, 226, 226, 5, 5, 227, 226, 5, 5, 228, 226, 5, 5, 229, 226, 5, 5, 230, 226, 5, 5, 234, 226, 5, 5, 19, 232, 5, 5, 20, 232, 5, 5, 21, 232, 5, 5, 22, 232, 5, 5, 23, 232, 5, 5, 24, 232, 5, 5, 25, 232, 5, 5, 26, 232, 5, 5, 27, 232, 5, 5, 28, 232, 5, 5, 29, 232, 5, 5, 85, 236, 5, 5, 86, 236, 5, 5, 87, 236, 5, 5, 88, 236, 5, 5, 89, 236, 5, 5, 214, 239, 5, 5, 215, 239, 5, 5, 216, 239, 5, 5, 217, 239, 5, 5, 218, 239, 5, 5, 219, 239, 5, 5, 220, 239, 5, 5, 221, 239, 5, 5, 205, 242, 5, 5, 206, 242, 5, 5, 207, 242, 5, 5, 208, 242, 5, 5, 186, 246, 5, 5, 187, 246, 5, 5, 222, 248, 5, 5, 221, 248, 5, 5, 177, 159, 5, 5, 167, 161, 5, 5, 53, 165, 5, 5, 54, 165, 5, 5, 76, 170, 5, 5, 77, 170, 5, 5, 78, 170, 5, 5, 79, 170, 5, 5, 37, 176, 5, 5, 38, 176, 5, 5, 39, 176, 5, 5, 40, 176, 5, 5, 41, 176, 5, 5, 42, 176, 5, 5, 43, 176, 5, 5, 44, 176, 5, 5, 195, 182, 5, 5, 196, 182, 5, 5, 197, 182, 5, 5, 198, 182, 5, 5, 199, 182, 5, 5, 200, 182, 5, 5, 201, 182, 5, 5, 202, 182, 5, 5, 203, 182, 5, 5, 204, 182, 5, 5, 205, 182, 5, 5, 94, 190, 5, 5, 95, 190, 5, 5, 96, 190, 5, 5, 97, 190, 5, 5, 98, 190, 5, 5, 99, 190, 5, 5, 100, 190, 5, 5, 106, 198, 5, 5, 107, 198, 5, 5, 108, 198, 5, 5, 109, 198, 5, 5, 101, 190, 5, 5, 110, 198, 5, 5, 111, 198, 5, 5, 112, 198, 5, 5, 66, 206, 5, 5, 67, 206, 5, 5, 68, 206, 5, 5, 69, 206, 5, 5, 70, 206, 5, 5, 71, 206, 5, 5, 72, 206, 5, 5, 73, 206, 5, 5, 74, 206, 5, 5, 75, 206, 5, 5, 143, 213, 5, 5, 144, 213, 5, 5, 145, 213, 5, 5, 146, 213, 5, 5, 147, 213, 5, 5, 148, 213, 5, 5, 107, 220, 5, 5, 108, 220, 5, 5, 109, 220, 5, 5, 110, 220, 5, 5, 111, 220, 5, 5, 112, 220, 5, 5, 113, 220, 5, 5, 235, 226, 5, 5, 236, 226, 5, 5, 237, 226, 5, 5, 238, 226, 5, 5, 239, 226, 5, 5, 240, 226, 5, 5, 241, 226, 5, 5, 242, 226, 5, 5, 243, 226, 5, 5, 31, 232, 5, 5, 32, 232, 5, 5, 33, 232, 5, 5, 34, 232, 5, 5, 35, 232, 5, 5, 17, 245, 5, 5, 92, 236, 5, 5, 93, 236, 5, 5, 224, 239, 5, 5, 209, 242, 5, 5, 18, 245, 5, 5, 19, 245, 5, 5, 188, 246, 5, 5, 178, 159, 5, 5, 55, 165, 5, 5, 56, 165, 5, 5, 81, 170, 5, 5, 82, 170, 5, 5, 47, 176, 5, 5, 48, 176, 5, 5, 49, 176, 5, 5, 50, 176, 5, 5, 51, 176, 5, 5, 52, 176, 5, 5, 53, 176, 5, 5, 54, 176, 5, 5, 207, 182, 5, 5, 208, 182, 5, 5, 209, 182, 5, 5, 210, 182, 5, 5, 211, 182, 5, 5, 212, 182, 5, 5, 213, 182, 5, 5, 64, 193, 5, 5, 65, 193, 5, 5, 102, 190, 5, 5, 113, 198, 5, 5, 114, 198, 5, 5, 115, 198, 5, 5, 116, 198, 5, 5, 117, 198, 5, 5, 118, 198, 5, 5, 77, 206, 5, 5, 78, 206, 5, 5, 79, 206, 5, 5, 80, 206, 5, 5, 150, 213, 5, 5, 151, 213, 5, 5, 155, 213, 5, 5, 152, 213, 5, 5, 153, 213, 5, 5, 245, 226, 5, 5, 37, 232, 5, 5, 38, 232, 5, 5, 39, 232, 5, 5, 94, 236, 5, 5, 210, 242, 5, 5, 211, 242, 5, 5, 189, 246, 5, 5, 168, 161, 5, 5, 84, 170, 5, 5, 85, 170, 5, 5, 55, 176, 5, 5, 56, 176, 5, 5, 57, 176, 5, 5, 58, 176, 5, 5, 59, 176, 5, 5, 60, 176, 5, 5, 61, 176, 5, 5, 62, 176, 5, 5, 234, 182, 5, 5, 215, 182, 5, 5, 216, 182, 5, 5, 217, 182, 5, 5, 218, 182, 5, 5, 219, 182, 5, 5, 220, 182, 5, 5, 221, 182, 5, 5, 222, 182, 5, 5, 223, 182, 5, 5, 224, 182, 5, 5, 225, 182, 5, 5, 226, 182, 5, 5, 227, 182, 5, 5, 228, 182, 5, 5, 229, 182, 5, 5, 230, 182, 5, 5, 231, 182, 5, 5, 103, 190, 5, 5, 104, 190, 5, 5, 105, 190, 5, 5, 106, 190, 5, 5, 107, 190, 5, 5, 108, 190, 5, 5, 109, 190, 5, 5, 110, 190, 5, 5, 111, 190, 5, 5, 112, 190, 5, 5, 113, 190, 5, 5, 114, 190, 5, 5, 115, 190, 5, 5, 116, 190, 5, 5, 117, 190, 5, 5, 118, 190, 5, 5, 119, 190, 5, 5, 120, 190, 5, 5, 121, 190, 5, 5, 122, 190, 5, 5, 123, 190, 5, 5, 124, 190, 5, 5, 125, 190, 5, 5, 126, 190, 5, 5, 127, 190, 5, 5, 128, 190, 5, 5, 129, 190, 5, 5, 130, 190, 5, 5, 131, 190, 5, 5, 132, 190, 5, 5, 133, 190, 5, 5, 134, 190, 5, 5, 135, 190, 5, 5, 136, 190, 5, 5, 137, 190, 5, 5, 138, 190, 5, 5, 139, 190, 5, 5, 140, 190, 5, 5, 141, 190, 5, 5, 142, 190, 5, 5, 143, 190, 5, 5, 119, 198, 5, 5, 120, 198, 5, 5, 121, 198, 5, 5, 122, 198, 5, 5, 123, 198, 5, 5, 124, 198, 5, 5, 125, 198, 5, 5, 126, 198, 5, 5, 127, 198, 5, 5, 128, 198, 5, 5, 129, 198, 5, 5, 130, 198, 5, 5, 131, 198, 5, 5, 132, 198, 5, 5, 133, 198, 5, 5, 134, 198, 5, 5, 135, 198, 5, 5, 136, 198, 5, 5, 137, 198, 5, 5, 138, 198, 5, 5, 139, 198, 5, 5, 140, 198, 5, 5, 141, 198, 5, 5, 142, 198, 5, 5, 143, 198, 5, 5, 144, 198, 5, 5, 145, 198, 5, 5, 146, 198, 5, 5, 147, 198, 5, 5, 148, 198, 5, 5, 149, 198, 5, 5, 83, 206, 5, 5, 84, 206, 5, 5, 85, 206, 5, 5, 86, 206, 5, 5, 87, 206, 5, 5, 88, 206, 5, 5, 89, 206, 5, 5, 90, 206, 5, 5, 91, 206, 5, 5, 92, 206, 5, 5, 93, 206, 5, 5, 94, 206, 5, 5, 95, 206, 5, 5, 96, 206, 5, 5, 153, 198, 5, 5, 97, 206, 5, 5, 98, 206, 5, 5, 99, 206, 5, 5, 100, 206, 5, 5, 101, 206, 5, 5, 102, 206, 5, 5, 103, 206, 5, 5, 104, 206, 5, 5, 156, 213, 5, 5, 105, 206, 5, 5, 106, 206, 5, 5, 107, 206, 5, 5, 108, 206, 5, 5, 109, 206, 5, 5, 110, 206, 5, 5, 111, 206, 5, 5, 112, 206, 5, 5, 113, 206, 5, 5, 114, 206, 5, 5, 115, 206, 5, 5, 157, 213, 5, 5, 158, 213, 5, 5, 159, 213, 5, 5, 160, 213, 5, 5, 161, 213, 5, 5, 162, 213, 5, 5, 163, 213, 5, 5, 164, 213, 5, 5, 165, 213, 5, 5, 166, 213, 5, 5, 167, 213, 5, 5, 168, 213, 5, 5, 169, 213, 5, 5, 170, 213, 5, 5, 171, 213, 5, 5, 172, 213, 5, 5, 173, 213, 5, 5, 174, 213, 5, 5, 175, 213, 5, 5, 176, 213, 5, 5, 177, 213, 5, 5, 178, 213, 5, 5, 179, 213, 5, 5, 180, 213, 5, 5, 181, 213, 5, 5, 182, 213, 5, 5, 183, 213, 5, 5, 184, 213, 5, 5, 185, 213, 5, 5, 186, 213, 5, 5, 187, 213, 5, 5, 188, 213, 5, 5, 189, 213, 5, 5, 190, 213, 5, 5, 191, 213, 5, 5, 192, 213, 5, 5, 193, 213, 5, 5, 194, 213, 5, 5, 195, 213, 5, 5, 196, 213, 5, 5, 197, 213, 5, 5, 198, 213, 5, 5, 199, 213, 5, 5, 114, 220, 5, 5, 115, 220, 5, 5, 116, 220, 5, 5, 117, 220, 5, 5, 118, 220, 5, 5, 119, 220, 5, 5, 120, 220, 5, 5, 121, 220, 5, 5, 122, 220, 5, 5, 123, 220, 5, 5, 124, 220, 5, 5, 125, 220, 5, 5, 216, 213, 5, 5, 126, 220, 5, 5, 127, 220, 5, 5, 128, 220, 5, 5, 129, 220, 5, 5, 130, 220, 5, 5, 131, 220, 5, 5, 132, 220, 5, 5, 116, 206, 5, 5, 133, 220, 5, 5, 134, 220, 5, 5, 135, 220, 5, 5, 136, 220, 5, 5, 137, 220, 5, 5, 138, 220, 5, 5, 139, 220, 5, 5, 140, 220, 5, 5, 247, 226, 5, 5, 141, 220, 5, 5, 142, 220, 5, 5, 143, 220, 5, 5, 144, 220, 5, 5, 145, 220, 5, 5, 146, 220, 5, 5, 147, 220, 5, 5, 148, 220, 5, 5, 149, 220, 5, 5, 150, 220, 5, 5, 248, 226, 5, 5, 249, 226, 5, 5, 250, 226, 5, 5, 251, 226, 5, 5, 252, 226, 5, 5, 253, 226, 5, 5, 254, 226, 5, 5, 255, 226, 5, 5, 2, 227, 5, 5, 3, 227, 5, 5, 4, 227, 5, 5, 5, 227, 5, 5, 6, 227, 5, 5, 7, 227, 5, 5, 8, 227, 5, 5, 9, 227, 5, 5, 10, 227, 5, 5, 11, 227, 5, 5, 12, 227, 5, 5, 13, 227, 5, 5, 14, 227, 5, 5, 15, 227, 5, 5, 16, 227, 5, 5, 17, 227, 5, 5, 18, 227, 5, 5, 19, 227, 5, 5, 20, 227, 5, 5, 21, 227, 5, 5, 42, 232, 5, 5, 43, 232, 5, 5, 44, 232, 5, 5, 45, 232, 5, 5, 46, 232, 5, 5, 47, 232, 5, 5, 48, 232, 5, 5, 49, 232, 5, 5, 50, 232, 5, 5, 33, 227, 5, 5, 51, 232, 5, 5, 52, 232, 5, 5, 53, 232, 5, 5, 54, 232, 5, 5, 55, 232, 5, 5, 56, 232, 5, 5, 57, 232, 5, 5, 58, 232, 5, 5, 59, 232, 5, 5, 60, 232, 5, 5, 61, 232, 5, 5, 62, 232, 5, 5, 63, 232, 5, 5, 64, 232, 5, 5, 65, 232, 5, 5, 66, 232, 5, 5, 67, 232, 5, 5, 68, 232, 5, 5, 69, 232, 5, 5, 70, 232, 5, 5, 71, 232, 5, 5, 72, 232, 5, 5, 73, 232, 5, 5, 22, 227, 5, 5, 74, 232, 5, 5, 75, 232, 5, 5, 76, 232, 5, 5, 41, 232, 5, 5, 77, 232, 5, 5, 78, 232, 5, 5, 89, 232, 5, 5, 95, 236, 5, 5, 96, 236, 5, 5, 97, 236, 5, 5, 98, 236, 5, 5, 99, 236, 5, 5, 100, 236, 5, 5, 101, 236, 5, 5, 102, 236, 5, 5, 103, 236, 5, 5, 104, 236, 5, 5, 105, 236, 5, 5, 106, 236, 5, 5, 107, 236, 5, 5, 108, 236, 5, 5, 109, 236, 5, 5, 110, 236, 5, 5, 111, 236, 5, 5, 112, 236, 5, 5, 113, 236, 5, 5, 225, 239, 5, 5, 114, 236, 5, 5, 115, 236, 5, 5, 116, 236, 5, 5, 117, 236, 5, 5, 118, 236, 5, 5, 119, 236, 5, 5, 226, 239, 5, 5, 227, 239, 5, 5, 228, 239, 5, 5, 229, 239, 5, 5, 230, 239, 5, 5, 231, 239, 5, 5, 232, 239, 5, 5, 233, 239, 5, 5, 234, 239, 5, 5, 235, 239, 5, 5, 236, 239, 5, 5, 237, 239, 5, 5, 238, 239, 5, 5, 239, 239, 5, 5, 240, 239, 5, 5, 241, 239, 5, 5, 212, 242, 5, 5, 213, 242, 5, 5, 214, 242, 5, 5, 215, 242, 5, 5, 216, 242, 5, 5, 217, 242, 5, 5, 218, 242, 5, 5, 219, 242, 5, 5, 220, 242, 5, 5, 221, 242, 5, 5, 222, 242, 5, 5, 223, 242, 5, 5, 224, 242, 5, 5, 21, 245, 5, 5, 22, 245, 5, 5, 23, 245, 5, 5, 24, 245, 5, 5, 25, 245, 5, 5, 225, 242, 5, 5, 20, 245, 5, 5, 190, 246, 5, 5, 191, 246, 5, 5, 192, 246, 5, 5, 193, 246, 5, 5, 194, 246, 5, 5, 195, 246, 5, 5, 196, 246, 5, 5, 243, 247, 5, 5, 197, 246, 5, 5, 198, 246, 5, 5, 199, 246, 5, 5, 244, 247, 5, 5, 245, 247, 5, 5, 246, 247, 5, 5, 247, 247, 5, 5, 248, 247, 5, 5, 249, 247, 5, 5, 250, 247, 5, 5, 129, 249, 5, 5, 223, 248, 5, 5, 130, 249, 5, 5, 131, 249, 5, 5, 132, 249, 5, 5, 133, 249, 5, 5, 232, 249, 5, 5, 233, 249, 5, 5, 140, 250, 5, 5, 160, 250, 5, 5, 169, 161, 5, 5, 86, 170, 5, 5, 87, 170, 5, 5, 88, 170, 5, 5, 65, 176, 5, 5, 66, 176, 5, 5, 67, 176, 5, 5, 68, 176, 5, 5, 69, 176, 5, 5, 70, 176, 5, 5, 71, 176, 5, 5, 72, 176, 5, 5, 73, 176, 5, 5, 74, 176, 5, 5, 75, 176, 5, 5, 76, 176, 5, 5, 235, 182, 5, 5, 236, 182, 5, 5, 237, 182, 5, 5, 238, 182, 5, 5, 239, 182, 5, 5, 240, 182, 5, 5, 241, 182, 5, 5, 242, 182, 5, 5, 243, 182, 5, 5, 244, 182, 5, 5, 245, 182, 5, 5, 246, 182, 5, 5, 247, 182, 5, 5, 248, 182, 5, 5, 249, 182, 5, 5, 145, 190, 5, 5, 146, 190, 5, 5, 147, 190, 5, 5, 148, 190, 5, 5, 149, 190, 5, 5, 150, 190, 5, 5, 151, 190, 5, 5, 152, 190, 5, 5, 153, 190, 5, 5, 154, 190, 5, 5, 155, 190, 5, 5, 156, 190, 5, 5, 154, 198, 5, 5, 155, 198, 5, 5, 156, 198, 5, 5, 157, 198, 5, 5, 158, 198, 5, 5, 157, 190, 5, 5, 159, 198, 5, 5, 160, 198, 5, 5, 161, 198, 5, 5, 162, 198, 5, 5, 163, 198, 5, 5, 164, 198, 5, 5, 165, 198, 5, 5, 166, 198, 5, 5, 167, 198, 5, 5, 168, 198, 5, 5, 127, 206, 5, 5, 128, 206, 5, 5, 129, 206, 5, 5, 130, 206, 5, 5, 131, 206, 5, 5, 132, 206, 5, 5, 133, 206, 5, 5, 134, 206, 5, 5, 217, 213, 5, 5, 218, 213, 5, 5, 219, 213, 5, 5, 220, 213, 5, 5, 221, 213, 5, 5, 222, 213, 5, 5, 223, 213, 5, 5, 224, 213, 5, 5, 225, 213, 5, 5, 226, 213, 5, 5, 135, 206, 5, 5, 227, 213, 5, 5, 170, 220, 5, 5, 171, 220, 5, 5, 172, 220, 5, 5, 173, 220, 5, 5, 174, 220, 5, 5, 175, 220, 5, 5, 176, 220, 5, 5, 177, 220, 5, 5, 178, 220, 5, 5, 179, 220, 5, 5, 180, 220, 5, 5, 181, 220, 5, 5, 182, 220, 5, 5, 34, 227, 5, 5, 35, 227, 5, 5, 36, 227, 5, 5, 37, 227, 5, 5, 38, 227, 5, 5, 39, 227, 5, 5, 40, 227, 5, 5, 41, 227, 5, 5, 42, 227, 5, 5, 43, 227, 5, 5, 90, 232, 5, 5, 91, 232, 5, 5, 92, 232, 5, 5, 93, 232, 5, 5, 94, 232, 5, 5, 95, 232, 5, 5, 96, 232, 5, 5, 97, 232, 5, 5, 98, 232, 5, 5, 99, 232, 5, 5, 125, 236, 5, 5, 126, 236, 5, 5, 127, 236, 5, 5, 128, 236, 5, 5, 129, 236, 5, 5, 100, 232, 5, 5, 246, 239, 5, 5, 247, 239, 5, 5, 248, 239, 5, 5, 249, 239, 5, 5, 250, 239, 5, 5, 226, 242, 5, 5, 227, 242, 5, 5, 228, 242, 5, 5, 201, 246, 5, 5, 28, 245, 5, 5, 235, 249, 5, 5, 202, 246, 5, 5, 252, 247, 5, 5, 134, 249, 5, 5, 34, 250, 5, 5, 170, 161, 5, 5, 171, 161, 5, 5, 57, 165, 5, 5, 58, 165, 5, 5, 89, 170, 5, 5, 90, 170, 5, 5, 91, 170, 5, 5, 92, 170, 5, 5, 78, 176, 5, 5, 79, 176, 5, 5, 80, 176, 5, 5, 81, 176, 5, 5, 82, 176, 5, 5, 83, 176, 5, 5, 84, 176, 5, 5, 85, 176, 5, 5, 86, 176, 5, 5, 87, 176, 5, 5, 250, 182, 5, 5, 251, 182, 5, 5, 252, 182, 5, 5, 253, 182, 5, 5, 254, 182, 5, 5, 255, 182, 5, 5, 2, 183, 5, 5, 3, 183, 5, 5, 4, 183, 5, 5, 5, 183, 5, 5, 6, 183, 5, 5, 7, 183, 5, 5, 8, 183, 5, 5, 9, 183, 5, 5, 10, 183, 5, 5, 11, 183, 5, 5, 12, 183, 5, 5, 13, 183, 5, 5, 14, 183, 5, 5, 15, 183, 5, 5, 16, 183, 5, 5, 17, 183, 5, 5, 18, 183, 5, 5, 19, 183, 5, 5, 20, 183, 5, 5, 21, 183, 5, 5, 22, 183, 5, 5, 23, 183, 5, 5, 24, 183, 5, 5, 25, 183, 5, 5, 159, 190, 5, 5, 160, 190, 5, 5, 170, 198, 5, 5, 171, 198, 5, 5, 161, 190, 5, 5, 162, 190, 5, 5, 163, 190, 5, 5, 164, 190, 5, 5, 165, 190, 5, 5, 166, 190, 5, 5, 167, 190, 5, 5, 168, 190, 5, 5, 169, 190, 5, 5, 170, 190, 5, 5, 171, 190, 5, 5, 172, 190, 5, 5, 173, 190, 5, 5, 174, 190, 5, 5, 175, 190, 5, 5, 176, 190, 5, 5, 177, 190, 5, 5, 178, 190, 5, 5, 179, 190, 5, 5, 180, 190, 5, 5, 181, 190, 5, 5, 182, 190, 5, 5, 183, 190, 5, 5, 184, 190, 5, 5, 185, 190, 5, 5, 186, 190, 5, 5, 187, 190, 5, 5, 188, 190, 5, 5, 189, 190, 5, 5, 190, 190, 5, 5, 191, 190, 5, 5, 192, 190, 5, 5, 193, 190, 5, 5, 172, 198, 5, 5, 173, 198, 5, 5, 174, 198, 5, 5, 175, 198, 5, 5, 176, 198, 5, 5, 177, 198, 5, 5, 178, 198, 5, 5, 179, 198, 5, 5, 180, 198, 5, 5, 181, 198, 5, 5, 182, 198, 5, 5, 183, 198, 5, 5, 184, 198, 5, 5, 185, 198, 5, 5, 140, 206, 5, 5, 186, 198, 5, 5, 187, 198, 5, 5, 188, 198, 5, 5, 189, 198, 5, 5, 190, 198, 5, 5, 191, 198, 5, 5, 192, 198, 5, 5, 193, 198, 5, 5, 194, 198, 5, 5, 195, 198, 5, 5, 196, 198, 5, 5, 197, 198, 5, 5, 198, 198, 5, 5, 199, 198, 5, 5, 200, 198, 5, 5, 201, 198, 5, 5, 202, 198, 5, 5, 203, 198, 5, 5, 204, 198, 5, 5, 205, 198, 5, 5, 206, 198, 5, 5, 207, 198, 5, 5, 208, 198, 5, 5, 209, 198, 5, 5, 210, 198, 5, 5, 211, 198, 5, 5, 212, 198, 5, 5, 213, 198, 5, 5, 141, 206, 5, 5, 142, 206, 5, 5, 143, 206, 5, 5, 144, 206, 5, 5, 145, 206, 5, 5, 146, 206, 5, 5, 214, 198, 5, 5, 147, 206, 5, 5, 148, 206, 5, 5, 149, 206, 5, 5, 150, 206, 5, 5, 151, 206, 5, 5, 152, 206, 5, 5, 153, 206, 5, 5, 154, 206, 5, 5, 155, 206, 5, 5, 156, 206, 5, 5, 157, 206, 5, 5, 158, 206, 5, 5, 159, 206, 5, 5, 160, 206, 5, 5, 161, 206, 5, 5, 162, 206, 5, 5, 163, 206, 5, 5, 164, 206, 5, 5, 165, 206, 5, 5, 166, 206, 5, 5, 167, 206, 5, 5, 168, 206, 5, 5, 169, 206, 5, 5, 230, 213, 5, 5, 170, 206, 5, 5, 171, 206, 5, 5, 172, 206, 5, 5, 173, 206, 5, 5, 174, 206, 5, 5, 232, 213, 5, 5, 233, 213, 5, 5, 234, 213, 5, 5, 235, 213, 5, 5, 236, 213, 5, 5, 237, 213, 5, 5, 238, 213, 5, 5, 239, 213, 5, 5, 240, 213, 5, 5, 241, 213, 5, 5, 242, 213, 5, 5, 243, 213, 5, 5, 244, 213, 5, 5, 245, 213, 5, 5, 246, 213, 5, 5, 247, 213, 5, 5, 248, 213, 5, 5, 249, 213, 5, 5, 250, 213, 5, 5, 251, 213, 5, 5, 252, 213, 5, 5, 253, 213, 5, 5, 254, 213, 5, 5, 255, 213, 5, 5, 2, 214, 5, 5, 3, 214, 5, 5, 4, 214, 5, 5, 5, 214, 5, 5, 6, 214, 5, 5, 7, 214, 5, 5, 8, 214, 5, 5, 9, 214, 5, 5, 10, 214, 5, 5, 11, 214, 5, 5, 12, 214, 5, 5, 13, 214, 5, 5, 14, 214, 5, 5, 15, 214, 5, 5, 16, 214, 5, 5, 17, 214, 5, 5, 18, 214, 5, 5, 19, 214, 5, 5, 20, 214, 5, 5, 21, 214, 5, 5, 22, 214, 5, 5, 23, 214, 5, 5, 24, 214, 5, 5, 25, 214, 5, 5, 26, 214, 5, 5, 27, 214, 5, 5, 28, 214, 5, 5, 29, 214, 5, 5, 231, 213, 5, 5, 30, 214, 5, 5, 31, 214, 5, 5, 186, 220, 5, 5, 32, 214, 5, 5, 33, 214, 5, 5, 187, 220, 5, 5, 188, 220, 5, 5, 189, 220, 5, 5, 190, 220, 5, 5, 191, 220, 5, 5, 192, 220, 5, 5, 193, 220, 5, 5, 194, 220, 5, 5, 195, 220, 5, 5, 196, 220, 5, 5, 197, 220, 5, 5, 198, 220, 5, 5, 199, 220, 5, 5, 200, 220, 5, 5, 201, 220, 5, 5, 202, 220, 5, 5, 203, 220, 5, 5, 204, 220, 5, 5, 205, 220, 5, 5, 206, 220, 5, 5, 207, 220, 5, 5, 208, 220, 5, 5, 209, 220, 5, 5, 210, 220, 5, 5, 211, 220, 5, 5, 212, 220, 5, 5, 213, 220, 5, 5, 214, 220, 5, 5, 215, 220, 5, 5, 216, 220, 5, 5, 217, 220, 5, 5, 218, 220, 5, 5, 219, 220, 5, 5, 220, 220, 5, 5, 221, 220, 5, 5, 222, 220, 5, 5, 223, 220, 5, 5, 224, 220, 5, 5, 225, 220, 5, 5, 226, 220, 5, 5, 227, 220, 5, 5, 228, 220, 5, 5, 229, 220, 5, 5, 230, 220, 5, 5, 231, 220, 5, 5, 232, 220, 5, 5, 233, 220, 5, 5, 234, 220, 5, 5, 235, 220, 5, 5, 236, 220, 5, 5, 45, 227, 5, 5, 46, 227, 5, 5, 47, 227, 5, 5, 48, 227, 5, 5, 49, 227, 5, 5, 50, 227, 5, 5, 51, 227, 5, 5, 52, 227, 5, 5, 53, 227, 5, 5, 54, 227, 5, 5, 55, 227, 5, 5, 56, 227, 5, 5, 57, 227, 5, 5, 58, 227, 5, 5, 59, 227, 5, 5, 60, 227, 5, 5, 61, 227, 5, 5, 62, 227, 5, 5, 63, 227, 5, 5, 64, 227, 5, 5, 65, 227, 5, 5, 66, 227, 5, 5, 67, 227, 5, 5, 68, 227, 5, 5, 69, 227, 5, 5, 70, 227, 5, 5, 71, 227, 5, 5, 72, 227, 5, 5, 73, 227, 5, 5, 74, 227, 5, 5, 75, 227, 5, 5, 76, 227, 5, 5, 77, 227, 5, 5, 102, 232, 5, 5, 103, 232, 5, 5, 104, 232, 5, 5, 105, 232, 5, 5, 106, 232, 5, 5, 107, 232, 5, 5, 108, 232, 5, 5, 109, 232, 5, 5, 110, 232, 5, 5, 111, 232, 5, 5, 112, 232, 5, 5, 113, 232, 5, 5, 114, 232, 5, 5, 115, 232, 5, 5, 116, 232, 5, 5, 117, 232, 5, 5, 118, 232, 5, 5, 119, 232, 5, 5, 120, 232, 5, 5, 121, 232, 5, 5, 122, 232, 5, 5, 123, 232, 5, 5, 124, 232, 5, 5, 125, 232, 5, 5, 126, 232, 5, 5, 127, 232, 5, 5, 128, 232, 5, 5, 129, 232, 5, 5, 130, 232, 5, 5, 131, 232, 5, 5, 132, 232, 5, 5, 139, 232, 5, 5, 133, 232, 5, 5, 134, 232, 5, 5, 252, 239, 5, 5, 135, 232, 5, 5, 136, 232, 5, 5, 132, 236, 5, 5, 133, 236, 5, 5, 134, 236, 5, 5, 135, 236, 5, 5, 136, 236, 5, 5, 137, 236, 5, 5, 138, 236, 5, 5, 139, 236, 5, 5, 140, 236, 5, 5, 141, 236, 5, 5, 142, 236, 5, 5, 143, 236, 5, 5, 144, 236, 5, 5, 145, 236, 5, 5, 146, 236, 5, 5, 147, 236, 5, 5, 148, 236, 5, 5, 149, 236, 5, 5, 150, 236, 5, 5, 151, 236, 5, 5, 152, 236, 5, 5, 153, 236, 5, 5, 154, 236, 5, 5, 155, 236, 5, 5, 253, 239, 5, 5, 156, 236, 5, 5, 254, 239, 5, 5, 255, 239, 5, 5, 2, 240, 5, 5, 3, 240, 5, 5, 4, 240, 5, 5, 5, 240, 5, 5, 6, 240, 5, 5, 7, 240, 5, 5, 8, 240, 5, 5, 157, 236, 5, 5, 9, 240, 5, 5, 10, 240, 5, 5, 11, 240, 5, 5, 12, 240, 5, 5, 13, 240, 5, 5, 14, 240, 5, 5, 15, 240, 5, 5, 16, 240, 5, 5, 17, 240, 5, 5, 230, 242, 5, 5, 231, 242, 5, 5, 232, 242, 5, 5, 233, 242, 5, 5, 234, 242, 5, 5, 235, 242, 5, 5, 236, 242, 5, 5, 237, 242, 5, 5, 238, 242, 5, 5, 29, 245, 5, 5, 30, 245, 5, 5, 31, 245, 5, 5, 32, 245, 5, 5, 33, 245, 5, 5, 34, 245, 5, 5, 35, 245, 5, 5, 36, 245, 5, 5, 37, 245, 5, 5, 38, 245, 5, 5, 39, 245, 5, 5, 40, 245, 5, 5, 41, 245, 5, 5, 203, 246, 5, 5, 204, 246, 5, 5, 253, 247, 5, 5, 254, 247, 5, 5, 255, 247, 5, 5, 2, 248, 5, 5, 225, 248, 5, 5, 135, 249, 5, 5, 136, 249, 5, 5, 137, 249, 5, 5, 138, 249, 5, 5, 35, 250, 5, 5, 36, 250, 5, 5, 123, 250, 5, 5, 151, 157, 5, 5, 179, 159, 5, 5, 172, 161, 5, 5, 173, 161, 5, 5, 174, 161, 5, 5, 175, 161, 5, 5, 176, 161, 5, 5, 177, 161, 5, 5, 178, 161, 5, 5, 179, 161, 5, 5, 180, 161, 5, 5, 181, 161, 5, 5, 182, 161, 5, 5, 60, 165, 5, 5, 61, 165, 5, 5, 62, 165, 5, 5, 63, 165, 5, 5, 64, 165, 5, 5, 65, 165, 5, 5, 66, 165, 5, 5, 67, 165, 5, 5, 68, 165, 5, 5, 69, 165, 5, 5, 59, 165, 5, 5, 70, 165, 5, 5, 71, 165, 5, 5, 72, 165, 5, 5, 73, 165, 5, 5, 74, 165, 5, 5, 75, 165, 5, 5, 76, 165, 5, 5, 77, 165, 5, 5, 94, 170, 5, 5, 95, 170, 5, 5, 96, 170, 5, 5, 97, 170, 5, 5, 98, 170, 5, 5, 99, 170, 5, 5, 100, 170, 5, 5, 101, 170, 5, 5, 102, 170, 5, 5, 103, 170, 5, 5, 104, 170, 5, 5, 105, 170, 5, 5, 106, 170, 5, 5, 107, 170, 5, 5, 108, 170, 5, 5, 109, 170, 5, 5, 110, 170, 5, 5, 111, 170, 5, 5, 91, 176, 5, 5, 92, 176, 5, 5, 93, 176, 5, 5, 94, 176, 5, 5, 95, 176, 5, 5, 96, 176, 5, 5, 97, 176, 5, 5, 98, 176, 5, 5, 99, 176, 5, 5, 100, 176, 5, 5, 101, 176, 5, 5, 102, 176, 5, 5, 103, 176, 5, 5, 104, 176, 5, 5, 105, 176, 5, 5, 27, 183, 5, 5, 28, 183, 5, 5, 29, 183, 5, 5, 30, 183, 5, 5, 31, 183, 5, 5, 32, 183, 5, 5, 33, 183, 5, 5, 34, 183, 5, 5, 35, 183, 5, 5, 196, 190, 5, 5, 197, 190, 5, 5, 198, 190, 5, 5, 199, 190, 5, 5, 200, 190, 5, 5, 201, 190, 5, 5, 202, 190, 5, 5, 203, 190, 5, 5, 204, 190, 5, 5, 205, 190, 5, 5, 206, 190, 5, 5, 207, 190, 5, 5, 208, 190, 5, 5, 209, 190, 5, 5, 210, 190, 5, 5, 211, 190, 5, 5, 212, 190, 5, 5, 213, 190, 5, 5, 214, 190, 5, 5, 215, 190, 5, 5, 216, 190, 5, 5, 217, 190, 5, 5, 218, 190, 5, 5, 219, 190, 5, 5, 220, 190, 5, 5, 216, 198, 5, 5, 217, 198, 5, 5, 218, 198, 5, 5, 219, 198, 5, 5, 220, 198, 5, 5, 221, 198, 5, 5, 222, 198, 5, 5, 223, 198, 5, 5, 224, 198, 5, 5, 225, 198, 5, 5, 226, 198, 5, 5, 227, 198, 5, 5, 228, 198, 5, 5, 229, 198, 5, 5, 230, 198, 5, 5, 231, 198, 5, 5, 232, 198, 5, 5, 233, 198, 5, 5, 234, 198, 5, 5, 235, 198, 5, 5, 236, 198, 5, 5, 237, 198, 5, 5, 238, 198, 5, 5, 177, 206, 5, 5, 178, 206, 5, 5, 179, 206, 5, 5, 180, 206, 5, 5, 181, 206, 5, 5, 182, 206, 5, 5, 183, 206, 5, 5, 184, 206, 5, 5, 185, 206, 5, 5, 186, 206, 5, 5, 187, 206, 5, 5, 188, 206, 5, 5, 41, 214, 5, 5, 42, 214, 5, 5, 43, 214, 5, 5, 44, 214, 5, 5, 45, 214, 5, 5, 46, 214, 5, 5, 47, 214, 5, 5, 248, 220, 5, 5, 249, 220, 5, 5, 250, 220, 5, 5, 251, 220, 5, 5, 79, 227, 5, 5, 80, 227, 5, 5, 81, 227, 5, 5, 82, 227, 5, 5, 83, 227, 5, 5, 19, 240, 5, 5, 183, 161, 5, 5, 112, 170, 5, 5, 106, 176, 5, 5, 36, 183, 5, 5, 37, 183, 5, 5, 221, 190, 5, 5, 38, 183, 5, 5, 222, 190, 5, 5, 239, 198, 5, 5, 240, 198, 5, 5, 241, 198, 5, 5, 48, 214, 5, 5, 49, 214, 5, 5, 84, 227, 5, 5, 140, 232, 5, 5, 141, 232, 5, 5, 142, 232, 5, 5, 163, 236, 5, 5, 164, 236, 5, 5, 165, 236, 5, 5, 20, 240, 5, 5, 21, 240, 5, 5, 243, 242, 5, 5, 42, 245, 5, 5, 207, 246, 5, 5, 208, 246, 5, 5, 226, 248, 5, 5, 184, 161, 5, 5, 180, 159, 5, 5, 107, 158, 5, 5, 113, 170, 5, 5, 78, 165, 5, 5, 114, 170, 5, 5, 115, 170, 5, 5, 108, 176, 5, 5, 116, 170, 5, 5, 109, 176, 5, 5, 40, 183, 5, 5, 41, 183, 5, 5, 42, 183, 5, 5, 43, 183, 5, 5, 44, 183, 5, 5, 45, 183, 5, 5, 46, 183, 5, 5, 47, 183, 5, 5, 225, 190, 5, 5, 242, 198, 5, 5, 243, 198, 5, 5, 244, 198, 5, 5, 189, 206, 5, 5, 190, 206, 5, 5, 191, 206, 5, 5, 192, 206, 5, 5, 193, 206, 5, 5, 194, 206, 5, 5, 195, 206, 5, 5, 196, 206, 5, 5, 50, 214, 5, 5, 51, 214, 5, 5, 52, 214, 5, 5, 197, 206, 5, 5, 53, 214, 5, 5, 54, 214, 5, 5, 252, 220, 5, 5, 253, 220, 5, 5, 254, 220, 5, 5, 255, 220, 5, 5, 85, 227, 5, 5, 86, 227, 5, 5, 87, 227, 5, 5, 88, 227, 5, 5, 143, 232, 5, 5, 144, 232, 5, 5, 145, 232, 5, 5, 167, 236, 5, 5, 146, 232, 5, 5, 168, 236, 5, 5, 22, 240, 5, 5, 23, 240, 5, 5, 24, 240, 5, 5, 25, 240, 5, 5, 210, 246, 5, 5, 227, 248, 5, 5, 228, 248, 5, 5, 185, 161, 5, 5, 117, 170, 5, 5, 118, 170, 5, 5, 110, 176, 5, 5, 111, 176, 5, 5, 112, 176, 5, 5, 48, 183, 5, 5, 113, 176, 5, 5, 49, 183, 5, 5, 50, 183, 5, 5, 51, 183, 5, 5, 226, 190, 5, 5, 52, 183, 5, 5, 53, 183, 5, 5, 54, 183, 5, 5, 55, 183, 5, 5, 227, 190, 5, 5, 228, 190, 5, 5, 229, 190, 5, 5, 230, 190, 5, 5, 231, 190, 5, 5, 232, 190, 5, 5, 245, 198, 5, 5, 246, 198, 5, 5, 247, 198, 5, 5, 199, 206, 5, 5, 200, 206, 5, 5, 201, 206, 5, 5, 202, 206, 5, 5, 203, 206, 5, 5, 204, 206, 5, 5, 205, 206, 5, 5, 206, 206, 5, 5, 55, 214, 5, 5, 2, 221, 5, 5, 3, 221, 5, 5, 4, 221, 5, 5, 5, 221, 5, 5, 6, 221, 5, 5, 89, 227, 5, 5, 90, 227, 5, 5, 169, 236, 5, 5, 170, 236, 5, 5, 171, 236, 5, 5, 26, 240, 5, 5, 27, 240, 5, 5, 28, 240, 5, 5, 29, 240, 5, 5, 245, 242, 5, 5, 43, 245, 5, 5, 44, 245, 5, 5, 189, 161, 5, 5, 115, 176, 5, 5, 116, 176, 5, 5, 56, 183, 5, 5, 57, 183, 5, 5, 58, 183, 5, 5, 59, 183, 5, 5, 60, 183, 5, 5, 61, 183, 5, 5, 62, 183, 5, 5, 233, 190, 5, 5, 234, 190, 5, 5, 235, 190, 5, 5, 236, 190, 5, 5, 237, 190, 5, 5, 238, 190, 5, 5, 239, 190, 5, 5, 240, 190, 5, 5, 241, 190, 5, 5, 242, 190, 5, 5, 243, 190, 5, 5, 244, 190, 5, 5, 250, 198, 5, 5, 251, 198, 5, 5, 252, 198, 5, 5, 253, 198, 5, 5, 254, 198, 5, 5, 255, 198, 5, 5, 2, 199, 5, 5, 3, 199, 5, 5, 207, 206, 5, 5, 208, 206, 5, 5, 209, 206, 5, 5, 56, 214, 5, 5, 57, 214, 5, 5, 58, 214, 5, 5, 59, 214, 5, 5, 60, 214, 5, 5, 61, 214, 5, 5, 62, 214, 5, 5, 64, 214, 5, 5, 7, 221, 5, 5, 8, 221, 5, 5, 9, 221, 5, 5, 10, 221, 5, 5, 11, 221, 5, 5, 12, 221, 5, 5, 13, 221, 5, 5, 14, 221, 5, 5, 91, 227, 5, 5, 92, 227, 5, 5, 93, 227, 5, 5, 94, 227, 5, 5, 148, 232, 5, 5, 149, 232, 5, 5, 150, 232, 5, 5, 151, 232, 5, 5, 152, 232, 5, 5, 172, 236, 5, 5, 173, 236, 5, 5, 174, 236, 5, 5, 175, 236, 5, 5, 176, 236, 5, 5, 153, 232, 5, 5, 30, 240, 5, 5, 31, 240, 5, 5, 246, 242, 5, 5, 247, 242, 5, 5, 190, 161, 5, 5, 108, 158, 5, 5, 191, 161, 5, 5, 67, 183, 5, 5, 119, 170, 5, 5, 68, 183, 5, 5, 119, 176, 5, 5, 245, 190, 5, 5, 246, 190, 5, 5, 69, 183, 5, 5, 4, 199, 5, 5, 192, 161, 5, 5, 120, 176, 5, 5, 121, 176, 5, 5, 122, 176, 5, 5, 123, 176, 5, 5, 124, 176, 5, 5, 193, 161, 5, 5, 120, 170, 5, 5, 125, 176, 5, 5, 70, 183, 5, 5, 71, 183, 5, 5, 72, 183, 5, 5, 73, 183, 5, 5, 74, 183, 5, 5, 247, 190, 5, 5, 248, 190, 5, 5, 249, 190, 5, 5, 250, 190, 5, 5, 251, 190, 5, 5, 252, 190, 5, 5, 5, 199, 5, 5, 210, 206, 5, 5, 211, 206, 5, 5, 65, 214, 5, 5, 66, 214, 5, 5, 67, 214, 5, 5, 16, 221, 5, 5, 17, 221, 5, 5, 95, 227, 5, 5, 96, 227, 5, 5, 97, 227, 5, 5, 157, 232, 5, 5, 158, 232, 5, 5, 180, 236, 5, 5, 181, 236, 5, 5, 248, 242, 5, 5, 46, 245, 5, 5, 211, 246, 5, 5, 212, 246, 5, 5, 194, 161, 5, 5, 79, 165, 5, 5, 121, 170, 5, 5, 126, 176, 5, 5, 127, 176, 5, 5, 75, 183, 5, 5, 76, 183, 5, 5, 77, 183, 5, 5, 78, 183, 5, 5, 79, 183, 5, 5, 80, 183, 5, 5, 81, 183, 5, 5, 82, 183, 5, 5, 83, 183, 5, 5, 84, 183, 5, 5, 85, 183, 5, 5, 253, 190, 5, 5, 254, 190, 5, 5, 255, 190, 5, 5, 2, 191, 5, 5, 3, 191, 5, 5, 4, 191, 5, 5, 5, 191, 5, 5, 6, 191, 5, 5, 7, 191, 5, 5, 8, 191, 5, 5, 9, 191, 5, 5, 6, 199, 5, 5, 7, 199, 5, 5, 8, 199, 5, 5, 9, 199, 5, 5, 10, 199, 5, 5, 11, 199, 5, 5, 12, 199, 5, 5, 212, 206, 5, 5, 213, 206, 5, 5, 214, 206, 5, 5, 215, 206, 5, 5, 68, 214, 5, 5, 69, 214, 5, 5, 70, 214, 5, 5, 71, 214, 5, 5, 72, 214, 5, 5, 73, 214, 5, 5, 74, 214, 5, 5, 13, 199, 5, 5, 75, 214, 5, 5, 76, 214, 5, 5, 77, 214, 5, 5, 19, 221, 5, 5, 20, 221, 5, 5, 21, 221, 5, 5, 22, 221, 5, 5, 23, 221, 5, 5, 24, 221, 5, 5, 25, 221, 5, 5, 26, 221, 5, 5, 99, 227, 5, 5, 100, 227, 5, 5, 103, 227, 5, 5, 159, 232, 5, 5, 160, 232, 5, 5, 161, 232, 5, 5, 162, 232, 5, 5, 163, 232, 5, 5, 164, 232, 5, 5, 182, 236, 5, 5, 183, 236, 5, 5, 184, 236, 5, 5, 32, 240, 5, 5, 249, 242, 5, 5, 250, 242, 5, 5, 251, 242, 5, 5, 252, 242, 5, 5, 213, 246, 5, 5, 214, 246, 5, 5, 195, 161, 5, 5, 109, 158, 5, 5, 86, 183, 5, 5, 87, 183, 5, 5, 122, 170, 5, 5, 216, 206, 5, 5, 217, 206, 5, 5, 218, 206, 5, 5, 80, 214, 5, 5, 81, 214, 5, 5, 197, 161, 5, 5, 183, 159, 5, 5, 198, 161, 5, 5, 199, 161, 5, 5, 200, 161, 5, 5, 201, 161, 5, 5, 123, 170, 5, 5, 80, 165, 5, 5, 81, 165, 5, 5, 82, 165, 5, 5, 83, 165, 5, 5, 84, 165, 5, 5, 85, 165, 5, 5, 86, 165, 5, 5, 87, 165, 5, 5, 88, 165, 5, 5, 89, 165, 5, 5, 90, 165, 5, 5, 91, 165, 5, 5, 92, 165, 5, 5, 93, 165, 5, 5, 94, 165, 5, 5, 95, 165, 5, 5, 96, 165, 5, 5, 124, 170, 5, 5, 125, 170, 5, 5, 126, 170, 5, 5, 127, 170, 5, 5, 128, 170, 5, 5, 129, 170, 5, 5, 130, 170, 5, 5, 131, 170, 5, 5, 132, 170, 5, 5, 133, 170, 5, 5, 134, 170, 5, 5, 135, 170, 5, 5, 136, 170, 5, 5, 137, 170, 5, 5, 138, 170, 5, 5, 139, 170, 5, 5, 140, 170, 5, 5, 141, 170, 5, 5, 142, 170, 5, 5, 143, 170, 5, 5, 144, 170, 5, 5, 145, 170, 5, 5, 146, 170, 5, 5, 147, 170, 5, 5, 148, 170, 5, 5, 149, 170, 5, 5, 150, 170, 5, 5, 151, 170, 5, 5, 152, 170, 5, 5, 153, 170, 5, 5, 154, 170, 5, 5, 155, 170, 5, 5, 156, 170, 5, 5, 129, 176, 5, 5, 130, 176, 5, 5, 131, 176, 5, 5, 132, 176, 5, 5, 133, 176, 5, 5, 134, 176, 5, 5, 135, 176, 5, 5, 136, 176, 5, 5, 137, 176, 5, 5, 138, 176, 5, 5, 139, 176, 5, 5, 140, 176, 5, 5, 141, 176, 5, 5, 142, 176, 5, 5, 143, 176, 5, 5, 144, 176, 5, 5, 145, 176, 5, 5, 146, 176, 5, 5, 15, 199, 5, 5, 147, 176, 5, 5, 148, 176, 5, 5, 149, 176, 5, 5, 150, 176, 5, 5, 151, 176, 5, 5, 152, 176, 5, 5, 153, 176, 5, 5, 154, 176, 5, 5, 155, 176, 5, 5, 156, 176, 5, 5, 157, 176, 5, 5, 158, 176, 5, 5, 159, 176, 5, 5, 160, 176, 5, 5, 161, 176, 5, 5, 162, 176, 5, 5, 163, 176, 5, 5, 164, 176, 5, 5, 165, 176, 5, 5, 166, 176, 5, 5, 167, 176, 5, 5, 168, 176, 5, 5, 169, 176, 5, 5, 10, 191, 5, 5, 88, 183, 5, 5, 89, 183, 5, 5, 90, 183, 5, 5, 91, 183, 5, 5, 92, 183, 5, 5, 93, 183, 5, 5, 94, 183, 5, 5, 95, 183, 5, 5, 96, 183, 5, 5, 97, 183, 5, 5, 98, 183, 5, 5, 99, 183, 5, 5, 100, 183, 5, 5, 101, 183, 5, 5, 102, 183, 5, 5, 103, 183, 5, 5, 104, 183, 5, 5, 16, 199, 5, 5, 105, 183, 5, 5, 106, 183, 5, 5, 107, 183, 5, 5, 108, 183, 5, 5, 109, 183, 5, 5, 110, 183, 5, 5, 111, 183, 5, 5, 112, 183, 5, 5, 113, 183, 5, 5, 114, 183, 5, 5, 170, 176, 5, 5, 115, 183, 5, 5, 116, 183, 5, 5, 117, 183, 5, 5, 118, 183, 5, 5, 119, 183, 5, 5, 120, 183, 5, 5, 121, 183, 5, 5, 122, 183, 5, 5, 123, 183, 5, 5, 124, 183, 5, 5, 17, 199, 5, 5, 11, 191, 5, 5, 12, 191, 5, 5, 13, 191, 5, 5, 14, 191, 5, 5, 15, 191, 5, 5, 16, 191, 5, 5, 17, 191, 5, 5, 18, 191, 5, 5, 19, 191, 5, 5, 20, 191, 5, 5, 21, 191, 5, 5, 18, 199, 5, 5, 22, 191, 5, 5, 23, 191, 5, 5, 24, 191, 5, 5, 25, 191, 5, 5, 26, 191, 5, 5, 27, 191, 5, 5, 28, 191, 5, 5, 29, 191, 5, 5, 30, 191, 5, 5, 31, 191, 5, 5, 32, 191, 5, 5, 33, 191, 5, 5, 34, 191, 5, 5, 35, 191, 5, 5, 36, 191, 5, 5, 37, 191, 5, 5, 38, 191, 5, 5, 39, 191, 5, 5, 40, 191, 5, 5, 41, 191, 5, 5, 42, 191, 5, 5, 43, 191, 5, 5, 44, 191, 5, 5, 45, 191, 5, 5, 19, 199, 5, 5, 20, 199, 5, 5, 21, 199, 5, 5, 22, 199, 5, 5, 23, 199, 5, 5, 24, 199, 5, 5, 25, 199, 5, 5, 26, 199, 5, 5, 27, 199, 5, 5, 28, 199, 5, 5, 29, 199, 5, 5, 30, 199, 5, 5, 31, 199, 5, 5, 32, 199, 5, 5, 33, 199, 5, 5, 34, 199, 5, 5, 35, 199, 5, 5, 36, 199, 5, 5, 37, 199, 5, 5, 38, 199, 5, 5, 39, 199, 5, 5, 40, 199, 5, 5, 41, 199, 5, 5, 82, 214, 5, 5, 42, 199, 5, 5, 43, 199, 5, 5, 44, 199, 5, 5, 45, 199, 5, 5, 46, 199, 5, 5, 47, 199, 5, 5, 48, 199, 5, 5, 49, 199, 5, 5, 50, 199, 5, 5, 51, 199, 5, 5, 221, 206, 5, 5, 222, 206, 5, 5, 223, 206, 5, 5, 224, 206, 5, 5, 225, 206, 5, 5, 226, 206, 5, 5, 227, 206, 5, 5, 228, 206, 5, 5, 229, 206, 5, 5, 230, 206, 5, 5, 231, 206, 5, 5, 232, 206, 5, 5, 233, 206, 5, 5, 234, 206, 5, 5, 235, 206, 5, 5, 236, 206, 5, 5, 237, 206, 5, 5, 238, 206, 5, 5, 239, 206, 5, 5, 240, 206, 5, 5, 241, 206, 5, 5, 242, 206, 5, 5, 243, 206, 5, 5, 244, 206, 5, 5, 245, 206, 5, 5, 58, 199, 5, 5, 246, 206, 5, 5, 247, 206, 5, 5, 248, 206, 5, 5, 249, 206, 5, 5, 250, 206, 5, 5, 251, 206, 5, 5, 252, 206, 5, 5, 253, 206, 5, 5, 254, 206, 5, 5, 255, 206, 5, 5, 83, 214, 5, 5, 84, 214, 5, 5, 85, 214, 5, 5, 86, 214, 5, 5, 87, 214, 5, 5, 88, 214, 5, 5, 89, 214, 5, 5, 90, 214, 5, 5, 91, 214, 5, 5, 92, 214, 5, 5, 93, 214, 5, 5, 94, 214, 5, 5, 95, 214, 5, 5, 96, 214, 5, 5, 97, 214, 5, 5, 98, 214, 5, 5, 99, 214, 5, 5, 104, 227, 5, 5, 100, 214, 5, 5, 28, 221, 5, 5, 29, 221, 5, 5, 30, 221, 5, 5, 31, 221, 5, 5, 32, 221, 5, 5, 33, 221, 5, 5, 34, 221, 5, 5, 35, 221, 5, 5, 36, 221, 5, 5, 37, 221, 5, 5, 38, 221, 5, 5, 39, 221, 5, 5, 40, 221, 5, 5, 41, 221, 5, 5, 42, 221, 5, 5, 43, 221, 5, 5, 44, 221, 5, 5, 45, 221, 5, 5, 182, 218, 5, 5, 166, 232, 5, 5, 106, 227, 5, 5, 107, 227, 5, 5, 108, 227, 5, 5, 109, 227, 5, 5, 110, 227, 5, 5, 111, 227, 5, 5, 112, 227, 5, 5, 113, 227, 5, 5, 114, 227, 5, 5, 115, 227, 5, 5, 116, 227, 5, 5, 117, 227, 5, 5, 118, 227, 5, 5, 119, 227, 5, 5, 120, 227, 5, 5, 121, 227, 5, 5, 122, 227, 5, 5, 126, 227, 5, 5, 167, 232, 5, 5, 127, 227, 5, 5, 168, 232, 5, 5, 169, 232, 5, 5, 170, 232, 5, 5, 171, 232, 5, 5, 172, 232, 5, 5, 173, 232, 5, 5, 174, 232, 5, 5, 175, 232, 5, 5, 176, 232, 5, 5, 177, 232, 5, 5, 178, 232, 5, 5, 179, 232, 5, 5, 180, 232, 5, 5, 181, 232, 5, 5, 182, 232, 5, 5, 183, 232, 5, 5, 184, 232, 5, 5, 33, 240, 5, 5, 185, 232, 5, 5, 185, 236, 5, 5, 186, 236, 5, 5, 187, 236, 5, 5, 188, 236, 5, 5, 189, 236, 5, 5, 190, 236, 5, 5, 191, 236, 5, 5, 34, 240, 5, 5, 35, 240, 5, 5, 253, 242, 5, 5, 36, 240, 5, 5, 37, 240, 5, 5, 254, 242, 5, 5, 255, 242, 5, 5, 2, 243, 5, 5, 3, 243, 5, 5, 47, 245, 5, 5, 215, 246, 5, 5, 216, 246, 5, 5, 139, 249, 5, 5, 140, 249, 5, 5, 3, 248, 5, 5, 202, 161, 5, 5, 158, 170, 5, 5, 159, 170, 5, 5, 59, 199, 5, 5, 108, 214, 5, 5, 188, 232, 5, 5, 189, 232, 5, 5, 203, 161, 5, 5, 97, 165, 5, 5, 129, 183, 5, 5, 130, 183, 5, 5, 60, 199, 5, 5, 61, 199, 5, 5, 62, 199, 5, 5, 49, 221, 5, 5, 128, 227, 5, 5, 204, 161, 5, 5, 172, 176, 5, 5, 64, 199, 5, 5, 65, 199, 5, 5, 66, 199, 5, 5, 67, 199, 5, 5, 68, 199, 5, 5, 109, 214, 5, 5, 129, 227, 5, 5, 205, 161, 5, 5, 160, 170, 5, 5, 161, 170, 5, 5, 173, 176, 5, 5, 132, 183, 5, 5, 133, 183, 5, 5, 52, 191, 5, 5, 69, 199, 5, 5, 70, 199, 5, 5, 3, 207, 5, 5, 131, 227, 5, 5, 110, 214, 5, 5, 130, 227, 5, 5, 132, 227, 5, 5, 196, 236, 5, 5, 42, 240, 5, 5, 207, 161, 5, 5, 162, 170, 5, 5, 163, 170, 5, 5, 164, 170, 5, 5, 134, 183, 5, 5, 53, 191, 5, 5, 71, 199, 5, 5, 111, 214, 5, 5, 112, 214, 5, 5, 113, 214, 5, 5, 50, 221, 5, 5, 51, 221, 5, 5, 133, 227, 5, 5, 197, 236, 5, 5, 43, 240, 5, 5, 208, 161, 5, 5, 72, 199, 5, 5, 4, 207, 5, 5, 114, 214, 5, 5, 209, 161, 5, 5, 165, 170, 5, 5, 174, 176, 5, 5, 175, 176, 5, 5, 176, 176, 5, 5, 177, 176, 5, 5, 135, 183, 5, 5, 136, 183, 5, 5, 137, 183, 5, 5, 138, 183, 5, 5, 139, 183, 5, 5, 140, 183, 5, 5, 141, 183, 5, 5, 142, 183, 5, 5, 143, 183, 5, 5, 144, 183, 5, 5, 145, 183, 5, 5, 146, 183, 5, 5, 147, 183, 5, 5, 54, 191, 5, 5, 55, 191, 5, 5, 56, 191, 5, 5, 57, 191, 5, 5, 58, 191, 5, 5, 59, 191, 5, 5, 60, 191, 5, 5, 61, 191, 5, 5, 62, 191, 5, 5, 63, 191, 5, 5, 73, 199, 5, 5, 74, 199, 5, 5, 75, 199, 5, 5, 76, 199, 5, 5, 5, 207, 5, 5, 6, 207, 5, 5, 7, 207, 5, 5, 8, 207, 5, 5, 9, 207, 5, 5, 10, 207, 5, 5, 11, 207, 5, 5, 12, 207, 5, 5, 13, 207, 5, 5, 14, 207, 5, 5, 115, 214, 5, 5, 116, 214, 5, 5, 117, 214, 5, 5, 118, 214, 5, 5, 52, 221, 5, 5, 53, 221, 5, 5, 54, 221, 5, 5, 55, 221, 5, 5, 56, 221, 5, 5, 57, 221, 5, 5, 58, 221, 5, 5, 134, 227, 5, 5, 135, 227, 5, 5, 136, 227, 5, 5, 137, 227, 5, 5, 138, 227, 5, 5, 191, 232, 5, 5, 192, 232, 5, 5, 193, 232, 5, 5, 194, 232, 5, 5, 198, 236, 5, 5, 199, 236, 5, 5, 200, 236, 5, 5, 44, 240, 5, 5, 45, 240, 5, 5, 46, 240, 5, 5, 47, 240, 5, 5, 48, 240, 5, 5, 5, 243, 5, 5, 6, 243, 5, 5, 7, 243, 5, 5, 8, 243, 5, 5, 48, 245, 5, 5, 218, 246, 5, 5, 5, 248, 5, 5, 229, 248, 5, 5, 210, 161, 5, 5, 98, 165, 5, 5, 166, 170, 5, 5, 195, 232, 5, 5, 211, 161, 5, 5, 149, 183, 5, 5, 66, 191, 5, 5, 78, 199, 5, 5, 49, 240, 5, 5, 230, 248, 5, 5, 212, 161, 5, 5, 153, 157, 5, 5, 184, 159, 5, 5, 214, 161, 5, 5, 215, 161, 5, 5, 216, 161, 5, 5, 217, 161, 5, 5, 218, 161, 5, 5, 219, 161, 5, 5, 220, 161, 5, 5, 221, 161, 5, 5, 99, 165, 5, 5, 100, 165, 5, 5, 101, 165, 5, 5, 102, 165, 5, 5, 103, 165, 5, 5, 104, 165, 5, 5, 105, 165, 5, 5, 106, 165, 5, 5, 107, 165, 5, 5, 108, 165, 5, 5, 109, 165, 5, 5, 110, 165, 5, 5, 111, 165, 5, 5, 112, 165, 5, 5, 113, 165, 5, 5, 114, 165, 5, 5, 115, 165, 5, 5, 178, 176, 5, 5, 116, 165, 5, 5, 117, 165, 5, 5, 118, 165, 5, 5, 167, 170, 5, 5, 168, 170, 5, 5, 169, 170, 5, 5, 170, 170, 5, 5, 171, 170, 5, 5, 172, 170, 5, 5, 173, 170, 5, 5, 174, 170, 5, 5, 175, 170, 5, 5, 176, 170, 5, 5, 177, 170, 5, 5, 178, 170, 5, 5, 179, 170, 5, 5, 180, 170, 5, 5, 181, 170, 5, 5, 182, 170, 5, 5, 183, 170, 5, 5, 184, 170, 5, 5, 185, 170, 5, 5, 186, 170, 5, 5, 187, 170, 5, 5, 188, 170, 5, 5, 189, 170, 5, 5, 190, 170, 5, 5, 191, 170, 5, 5, 192, 170, 5, 5, 193, 170, 5, 5, 194, 170, 5, 5, 195, 170, 5, 5, 196, 170, 5, 5, 197, 170, 5, 5, 198, 170, 5, 5, 199, 170, 5, 5, 200, 170, 5, 5, 201, 170, 5, 5, 150, 183, 5, 5, 202, 170, 5, 5, 203, 170, 5, 5, 204, 170, 5, 5, 226, 170, 5, 5, 205, 170, 5, 5, 206, 170, 5, 5, 207, 170, 5, 5, 208, 170, 5, 5, 209, 170, 5, 5, 210, 170, 5, 5, 211, 170, 5, 5, 212, 170, 5, 5, 213, 170, 5, 5, 214, 170, 5, 5, 215, 170, 5, 5, 216, 170, 5, 5, 217, 170, 5, 5, 218, 170, 5, 5, 219, 170, 5, 5, 220, 170, 5, 5, 179, 176, 5, 5, 180, 176, 5, 5, 181, 176, 5, 5, 182, 176, 5, 5, 183, 176, 5, 5, 184, 176, 5, 5, 185, 176, 5, 5, 186, 176, 5, 5, 187, 176, 5, 5, 188, 176, 5, 5, 189, 176, 5, 5, 190, 176, 5, 5, 191, 176, 5, 5, 192, 176, 5, 5, 193, 176, 5, 5, 194, 176, 5, 5, 195, 176, 5, 5, 196, 176, 5, 5, 197, 176, 5, 5, 198, 176, 5, 5, 199, 176, 5, 5, 200, 176, 5, 5, 201, 176, 5, 5, 202, 176, 5, 5, 203, 176, 5, 5, 204, 176, 5, 5, 205, 176, 5, 5, 206, 176, 5, 5, 207, 176, 5, 5, 208, 176, 5, 5, 209, 176, 5, 5, 210, 176, 5, 5, 211, 176, 5, 5, 212, 176, 5, 5, 213, 176, 5, 5, 214, 176, 5, 5, 215, 176, 5, 5, 216, 176, 5, 5, 217, 176, 5, 5, 218, 176, 5, 5, 219, 176, 5, 5, 220, 176, 5, 5, 221, 176, 5, 5, 222, 176, 5, 5, 223, 176, 5, 5, 224, 176, 5, 5, 225, 176, 5, 5, 226, 176, 5, 5, 227, 176, 5, 5, 228, 176, 5, 5, 229, 176, 5, 5, 230, 176, 5, 5, 231, 176, 5, 5, 232, 176, 5, 5, 233, 176, 5, 5, 234, 176, 5, 5, 154, 183, 5, 5, 235, 176, 5, 5, 236, 176, 5, 5, 237, 176, 5, 5, 238, 176, 5, 5, 239, 176, 5, 5, 240, 176, 5, 5, 241, 176, 5, 5, 242, 176, 5, 5, 243, 176, 5, 5, 79, 199, 5, 5, 244, 176, 5, 5, 245, 176, 5, 5, 246, 176, 5, 5, 155, 183, 5, 5, 156, 183, 5, 5, 157, 183, 5, 5, 158, 183, 5, 5, 159, 183, 5, 5, 160, 183, 5, 5, 161, 183, 5, 5, 69, 191, 5, 5, 162, 183, 5, 5, 163, 183, 5, 5, 164, 183, 5, 5, 165, 183, 5, 5, 166, 183, 5, 5, 70, 191, 5, 5, 167, 183, 5, 5, 168, 183, 5, 5, 169, 183, 5, 5, 170, 183, 5, 5, 171, 183, 5, 5, 172, 183, 5, 5, 173, 183, 5, 5, 174, 183, 5, 5, 175, 183, 5, 5, 176, 183, 5, 5, 177, 183, 5, 5, 178, 183, 5, 5, 179, 183, 5, 5, 180, 183, 5, 5, 181, 183, 5, 5, 182, 183, 5, 5, 183, 183, 5, 5, 184, 183, 5, 5, 185, 183, 5, 5, 186, 183, 5, 5, 187, 183, 5, 5, 188, 183, 5, 5, 252, 176, 5, 5, 80, 199, 5, 5, 189, 183, 5, 5, 190, 183, 5, 5, 221, 170, 5, 5, 191, 183, 5, 5, 192, 183, 5, 5, 193, 183, 5, 5, 194, 183, 5, 5, 195, 183, 5, 5, 196, 183, 5, 5, 197, 183, 5, 5, 81, 199, 5, 5, 198, 183, 5, 5, 199, 183, 5, 5, 200, 183, 5, 5, 201, 183, 5, 5, 202, 183, 5, 5, 203, 183, 5, 5, 204, 183, 5, 5, 205, 183, 5, 5, 206, 183, 5, 5, 207, 183, 5, 5, 208, 183, 5, 5, 209, 183, 5, 5, 242, 183, 5, 5, 210, 183, 5, 5, 211, 183, 5, 5, 212, 183, 5, 5, 213, 183, 5, 5, 214, 183, 5, 5, 67, 191, 5, 5, 215, 183, 5, 5, 216, 183, 5, 5, 217, 183, 5, 5, 218, 183, 5, 5, 219, 183, 5, 5, 220, 183, 5, 5, 221, 183, 5, 5, 222, 183, 5, 5, 223, 183, 5, 5, 224, 183, 5, 5, 225, 183, 5, 5, 226, 183, 5, 5, 227, 183, 5, 5, 228, 183, 5, 5, 229, 183, 5, 5, 230, 183, 5, 5, 231, 183, 5, 5, 68, 191, 5, 5, 232, 183, 5, 5, 233, 183, 5, 5, 234, 183, 5, 5, 235, 183, 5, 5, 71, 191, 5, 5, 72, 191, 5, 5, 73, 191, 5, 5, 74, 191, 5, 5, 75, 191, 5, 5, 76, 191, 5, 5, 77, 191, 5, 5, 78, 191, 5, 5, 79, 191, 5, 5, 80, 191, 5, 5, 81, 191, 5, 5, 82, 191, 5, 5, 83, 191, 5, 5, 84, 191, 5, 5, 85, 191, 5, 5, 86, 191, 5, 5, 87, 191, 5, 5, 88, 191, 5, 5, 89, 191, 5, 5, 90, 191, 5, 5, 91, 191, 5, 5, 92, 191, 5, 5, 93, 191, 5, 5, 94, 191, 5, 5, 95, 191, 5, 5, 96, 191, 5, 5, 97, 191, 5, 5, 98, 191, 5, 5, 99, 191, 5, 5, 100, 191, 5, 5, 101, 191, 5, 5, 102, 191, 5, 5, 103, 191, 5, 5, 104, 191, 5, 5, 105, 191, 5, 5, 106, 191, 5, 5, 107, 191, 5, 5, 108, 191, 5, 5, 109, 191, 5, 5, 110, 191, 5, 5, 111, 191, 5, 5, 112, 191, 5, 5, 86, 199, 5, 5, 113, 191, 5, 5, 114, 191, 5, 5, 115, 191, 5, 5, 116, 191, 5, 5, 117, 191, 5, 5, 118, 191, 5, 5, 119, 191, 5, 5, 120, 191, 5, 5, 121, 191, 5, 5, 122, 191, 5, 5, 123, 191, 5, 5, 124, 191, 5, 5, 125, 191, 5, 5, 126, 191, 5, 5, 127, 191, 5, 5, 128, 191, 5, 5, 129, 191, 5, 5, 130, 191, 5, 5, 150, 191, 5, 5, 131, 191, 5, 5, 132, 191, 5, 5, 133, 191, 5, 5, 134, 191, 5, 5, 135, 191, 5, 5, 136, 191, 5, 5, 137, 191, 5, 5, 138, 191, 5, 5, 139, 191, 5, 5, 140, 191, 5, 5, 141, 191, 5, 5, 142, 191, 5, 5, 143, 191, 5, 5, 16, 207, 5, 5, 144, 191, 5, 5, 145, 191, 5, 5, 87, 199, 5, 5, 88, 199, 5, 5, 89, 199, 5, 5, 90, 199, 5, 5, 91, 199, 5, 5, 92, 199, 5, 5, 93, 199, 5, 5, 94, 199, 5, 5, 95, 199, 5, 5, 96, 199, 5, 5, 97, 199, 5, 5, 98, 199, 5, 5, 99, 199, 5, 5, 100, 199, 5, 5, 101, 199, 5, 5, 102, 199, 5, 5, 103, 199, 5, 5, 104, 199, 5, 5, 105, 199, 5, 5, 106, 199, 5, 5, 107, 199, 5, 5, 108, 199, 5, 5, 109, 199, 5, 5, 110, 199, 5, 5, 111, 199, 5, 5, 112, 199, 5, 5, 113, 199, 5, 5, 20, 207, 5, 5, 114, 199, 5, 5, 115, 199, 5, 5, 116, 199, 5, 5, 117, 199, 5, 5, 118, 199, 5, 5, 119, 199, 5, 5, 120, 199, 5, 5, 121, 199, 5, 5, 122, 199, 5, 5, 123, 199, 5, 5, 124, 199, 5, 5, 125, 199, 5, 5, 126, 199, 5, 5, 127, 199, 5, 5, 128, 199, 5, 5, 129, 199, 5, 5, 130, 199, 5, 5, 131, 199, 5, 5, 132, 199, 5, 5, 133, 199, 5, 5, 134, 199, 5, 5, 135, 199, 5, 5, 136, 199, 5, 5, 137, 199, 5, 5, 138, 199, 5, 5, 139, 199, 5, 5, 140, 199, 5, 5, 141, 199, 5, 5, 142, 199, 5, 5, 143, 199, 5, 5, 144, 199, 5, 5, 145, 199, 5, 5, 146, 199, 5, 5, 147, 199, 5, 5, 148, 199, 5, 5, 149, 199, 5, 5, 150, 199, 5, 5, 151, 199, 5, 5, 152, 199, 5, 5, 153, 199, 5, 5, 154, 199, 5, 5, 155, 199, 5, 5, 156, 199, 5, 5, 157, 199, 5, 5, 158, 199, 5, 5, 159, 199, 5, 5, 160, 199, 5, 5, 161, 199, 5, 5, 162, 199, 5, 5, 163, 199, 5, 5, 164, 199, 5, 5, 165, 199, 5, 5, 166, 199, 5, 5, 167, 199, 5, 5, 168, 199, 5, 5, 169, 199, 5, 5, 170, 199, 5, 5, 171, 199, 5, 5, 172, 199, 5, 5, 173, 199, 5, 5, 174, 199, 5, 5, 175, 199, 5, 5, 176, 199, 5, 5, 177, 199, 5, 5, 178, 199, 5, 5, 179, 199, 5, 5, 180, 199, 5, 5, 181, 199, 5, 5, 182, 199, 5, 5, 183, 199, 5, 5, 184, 199, 5, 5, 185, 199, 5, 5, 186, 199, 5, 5, 187, 199, 5, 5, 188, 199, 5, 5, 21, 207, 5, 5, 189, 199, 5, 5, 190, 199, 5, 5, 22, 207, 5, 5, 23, 207, 5, 5, 24, 207, 5, 5, 25, 207, 5, 5, 39, 206, 5, 5, 26, 207, 5, 5, 27, 207, 5, 5, 28, 207, 5, 5, 29, 207, 5, 5, 30, 207, 5, 5, 31, 207, 5, 5, 32, 207, 5, 5, 33, 207, 5, 5, 34, 207, 5, 5, 35, 207, 5, 5, 36, 207, 5, 5, 205, 199, 5, 5, 37, 207, 5, 5, 38, 207, 5, 5, 39, 207, 5, 5, 40, 207, 5, 5, 41, 207, 5, 5, 42, 207, 5, 5, 43, 207, 5, 5, 44, 207, 5, 5, 45, 207, 5, 5, 46, 207, 5, 5, 47, 207, 5, 5, 48, 207, 5, 5, 49, 207, 5, 5, 50, 207, 5, 5, 51, 207, 5, 5, 52, 207, 5, 5, 53, 207, 5, 5, 54, 207, 5, 5, 55, 207, 5, 5, 56, 207, 5, 5, 57, 207, 5, 5, 58, 207, 5, 5, 59, 207, 5, 5, 60, 207, 5, 5, 61, 207, 5, 5, 62, 207, 5, 5, 63, 207, 5, 5, 64, 207, 5, 5, 65, 207, 5, 5, 66, 207, 5, 5, 191, 199, 5, 5, 67, 207, 5, 5, 68, 207, 5, 5, 69, 207, 5, 5, 70, 207, 5, 5, 71, 207, 5, 5, 72, 207, 5, 5, 73, 207, 5, 5, 74, 207, 5, 5, 75, 207, 5, 5, 76, 207, 5, 5, 77, 207, 5, 5, 78, 207, 5, 5, 79, 207, 5, 5, 80, 207, 5, 5, 81, 207, 5, 5, 82, 207, 5, 5, 83, 207, 5, 5, 84, 207, 5, 5, 85, 207, 5, 5, 86, 207, 5, 5, 87, 207, 5, 5, 88, 207, 5, 5, 89, 207, 5, 5, 90, 207, 5, 5, 91, 207, 5, 5, 92, 207, 5, 5, 93, 207, 5, 5, 94, 207, 5, 5, 95, 207, 5, 5, 96, 207, 5, 5, 97, 207, 5, 5, 98, 207, 5, 5, 99, 207, 5, 5, 100, 207, 5, 5, 101, 207, 5, 5, 102, 207, 5, 5, 103, 207, 5, 5, 104, 207, 5, 5, 105, 207, 5, 5, 106, 207, 5, 5, 107, 207, 5, 5, 108, 207, 5, 5, 109, 207, 5, 5, 110, 207, 5, 5, 111, 207, 5, 5, 112, 207, 5, 5, 113, 207, 5, 5, 114, 207, 5, 5, 115, 207, 5, 5, 116, 207, 5, 5, 60, 221, 5, 5, 117, 207, 5, 5, 118, 207, 5, 5, 119, 207, 5, 5, 120, 207, 5, 5, 121, 207, 5, 5, 130, 214, 5, 5, 131, 214, 5, 5, 132, 214, 5, 5, 133, 214, 5, 5, 134, 214, 5, 5, 135, 214, 5, 5, 136, 214, 5, 5, 137, 214, 5, 5, 138, 214, 5, 5, 139, 214, 5, 5, 140, 214, 5, 5, 141, 214, 5, 5, 142, 214, 5, 5, 143, 214, 5, 5, 144, 214, 5, 5, 145, 214, 5, 5, 146, 214, 5, 5, 147, 214, 5, 5, 148, 214, 5, 5, 149, 214, 5, 5, 150, 214, 5, 5, 151, 214, 5, 5, 152, 214, 5, 5, 153, 214, 5, 5, 154, 214, 5, 5, 155, 214, 5, 5, 156, 214, 5, 5, 157, 214, 5, 5, 158, 214, 5, 5, 159, 214, 5, 5, 160, 214, 5, 5, 161, 214, 5, 5, 162, 214, 5, 5, 163, 214, 5, 5, 164, 214, 5, 5, 165, 214, 5, 5, 166, 214, 5, 5, 167, 214, 5, 5, 168, 214, 5, 5, 169, 214, 5, 5, 201, 212, 5, 5, 170, 214, 5, 5, 171, 214, 5, 5, 172, 214, 5, 5, 173, 214, 5, 5, 174, 214, 5, 5, 175, 214, 5, 5, 176, 214, 5, 5, 177, 214, 5, 5, 178, 214, 5, 5, 179, 214, 5, 5, 180, 214, 5, 5, 181, 214, 5, 5, 134, 207, 5, 5, 182, 214, 5, 5, 183, 214, 5, 5, 135, 207, 5, 5, 184, 214, 5, 5, 185, 214, 5, 5, 186, 214, 5, 5, 187, 214, 5, 5, 188, 214, 5, 5, 189, 214, 5, 5, 190, 214, 5, 5, 191, 214, 5, 5, 192, 214, 5, 5, 193, 214, 5, 5, 194, 214, 5, 5, 195, 214, 5, 5, 196, 214, 5, 5, 197, 214, 5, 5, 198, 214, 5, 5, 199, 214, 5, 5, 200, 214, 5, 5, 201, 214, 5, 5, 202, 214, 5, 5, 203, 214, 5, 5, 204, 214, 5, 5, 139, 227, 5, 5, 205, 214, 5, 5, 62, 221, 5, 5, 206, 214, 5, 5, 207, 214, 5, 5, 208, 214, 5, 5, 209, 214, 5, 5, 225, 214, 5, 5, 210, 214, 5, 5, 63, 221, 5, 5, 64, 221, 5, 5, 65, 221, 5, 5, 66, 221, 5, 5, 67, 221, 5, 5, 68, 221, 5, 5, 69, 221, 5, 5, 70, 221, 5, 5, 71, 221, 5, 5, 72, 221, 5, 5, 136, 207, 5, 5, 73, 221, 5, 5, 74, 221, 5, 5, 75, 221, 5, 5, 76, 221, 5, 5, 77, 221, 5, 5, 78, 221, 5, 5, 79, 221, 5, 5, 80, 221, 5, 5, 81, 221, 5, 5, 82, 221, 5, 5, 83, 221, 5, 5, 84, 221, 5, 5, 85, 221, 5, 5, 86, 221, 5, 5, 87, 221, 5, 5, 88, 221, 5, 5, 89, 221, 5, 5, 90, 221, 5, 5, 91, 221, 5, 5, 92, 221, 5, 5, 93, 221, 5, 5, 137, 207, 5, 5, 94, 221, 5, 5, 95, 221, 5, 5, 96, 221, 5, 5, 97, 221, 5, 5, 98, 221, 5, 5, 99, 221, 5, 5, 100, 221, 5, 5, 101, 221, 5, 5, 102, 221, 5, 5, 103, 221, 5, 5, 104, 221, 5, 5, 105, 221, 5, 5, 106, 221, 5, 5, 107, 221, 5, 5, 108, 221, 5, 5, 109, 221, 5, 5, 110, 221, 5, 5, 111, 221, 5, 5, 112, 221, 5, 5, 113, 221, 5, 5, 114, 221, 5, 5, 115, 221, 5, 5, 116, 221, 5, 5, 117, 221, 5, 5, 118, 221, 5, 5, 119, 221, 5, 5, 120, 221, 5, 5, 121, 221, 5, 5, 122, 221, 5, 5, 123, 221, 5, 5, 124, 221, 5, 5, 125, 221, 5, 5, 126, 221, 5, 5, 127, 221, 5, 5, 128, 221, 5, 5, 129, 221, 5, 5, 130, 221, 5, 5, 131, 221, 5, 5, 132, 221, 5, 5, 133, 221, 5, 5, 134, 221, 5, 5, 135, 221, 5, 5, 136, 221, 5, 5, 137, 221, 5, 5, 138, 221, 5, 5, 139, 221, 5, 5, 140, 221, 5, 5, 141, 221, 5, 5, 142, 221, 5, 5, 143, 221, 5, 5, 144, 221, 5, 5, 145, 221, 5, 5, 146, 221, 5, 5, 159, 221, 5, 5, 145, 227, 5, 5, 146, 227, 5, 5, 147, 227, 5, 5, 148, 227, 5, 5, 149, 227, 5, 5, 150, 227, 5, 5, 151, 227, 5, 5, 152, 227, 5, 5, 153, 227, 5, 5, 154, 227, 5, 5, 155, 227, 5, 5, 156, 227, 5, 5, 157, 227, 5, 5, 158, 227, 5, 5, 159, 227, 5, 5, 160, 227, 5, 5, 161, 227, 5, 5, 160, 221, 5, 5, 162, 227, 5, 5, 163, 227, 5, 5, 164, 227, 5, 5, 165, 227, 5, 5, 166, 227, 5, 5, 167, 227, 5, 5, 168, 227, 5, 5, 202, 232, 5, 5, 169, 227, 5, 5, 170, 227, 5, 5, 171, 227, 5, 5, 172, 227, 5, 5, 173, 227, 5, 5, 174, 227, 5, 5, 175, 227, 5, 5, 176, 227, 5, 5, 177, 227, 5, 5, 178, 227, 5, 5, 179, 227, 5, 5, 180, 227, 5, 5, 181, 227, 5, 5, 182, 227, 5, 5, 183, 227, 5, 5, 184, 227, 5, 5, 185, 227, 5, 5, 186, 227, 5, 5, 187, 227, 5, 5, 188, 227, 5, 5, 189, 227, 5, 5, 190, 227, 5, 5, 191, 227, 5, 5, 192, 227, 5, 5, 193, 227, 5, 5, 194, 227, 5, 5, 195, 227, 5, 5, 196, 227, 5, 5, 197, 227, 5, 5, 198, 227, 5, 5, 203, 232, 5, 5, 204, 232, 5, 5, 205, 232, 5, 5, 206, 232, 5, 5, 207, 232, 5, 5, 208, 232, 5, 5, 209, 232, 5, 5, 210, 232, 5, 5, 211, 232, 5, 5, 212, 232, 5, 5, 213, 232, 5, 5, 214, 232, 5, 5, 215, 232, 5, 5, 216, 232, 5, 5, 217, 232, 5, 5, 218, 232, 5, 5, 219, 232, 5, 5, 220, 232, 5, 5, 221, 232, 5, 5, 222, 232, 5, 5, 223, 232, 5, 5, 224, 232, 5, 5, 211, 227, 5, 5, 225, 232, 5, 5, 226, 232, 5, 5, 227, 232, 5, 5, 228, 232, 5, 5, 229, 232, 5, 5, 230, 232, 5, 5, 231, 232, 5, 5, 232, 232, 5, 5, 233, 232, 5, 5, 234, 232, 5, 5, 235, 232, 5, 5, 236, 232, 5, 5, 237, 232, 5, 5, 238, 232, 5, 5, 239, 232, 5, 5, 240, 232, 5, 5, 241, 232, 5, 5, 242, 232, 5, 5, 243, 232, 5, 5, 244, 232, 5, 5, 245, 232, 5, 5, 246, 232, 5, 5, 247, 232, 5, 5, 248, 232, 5, 5, 249, 232, 5, 5, 250, 232, 5, 5, 251, 232, 5, 5, 252, 232, 5, 5, 211, 236, 5, 5, 253, 232, 5, 5, 254, 232, 5, 5, 255, 232, 5, 5, 8, 233, 5, 5, 2, 233, 5, 5, 9, 233, 5, 5, 212, 236, 5, 5, 213, 236, 5, 5, 214, 236, 5, 5, 215, 236, 5, 5, 216, 236, 5, 5, 217, 236, 5, 5, 218, 236, 5, 5, 219, 236, 5, 5, 220, 236, 5, 5, 221, 236, 5, 5, 222, 236, 5, 5, 223, 236, 5, 5, 224, 236, 5, 5, 225, 236, 5, 5, 226, 236, 5, 5, 227, 236, 5, 5, 228, 236, 5, 5, 229, 236, 5, 5, 230, 236, 5, 5, 231, 236, 5, 5, 232, 236, 5, 5, 233, 236, 5, 5, 234, 236, 5, 5, 235, 236, 5, 5, 236, 236, 5, 5, 237, 236, 5, 5, 238, 236, 5, 5, 239, 236, 5, 5, 240, 236, 5, 5, 241, 236, 5, 5, 242, 236, 5, 5, 243, 236, 5, 5, 244, 236, 5, 5, 53, 240, 5, 5, 245, 236, 5, 5, 246, 236, 5, 5, 49, 245, 5, 5, 54, 240, 5, 5, 55, 240, 5, 5, 56, 240, 5, 5, 57, 240, 5, 5, 58, 240, 5, 5, 59, 240, 5, 5, 60, 240, 5, 5, 61, 240, 5, 5, 62, 240, 5, 5, 63, 240, 5, 5, 64, 240, 5, 5, 65, 240, 5, 5, 66, 240, 5, 5, 67, 240, 5, 5, 68, 240, 5, 5, 69, 240, 5, 5, 70, 240, 5, 5, 71, 240, 5, 5, 72, 240, 5, 5, 73, 240, 5, 5, 74, 240, 5, 5, 75, 240, 5, 5, 76, 240, 5, 5, 77, 240, 5, 5, 78, 240, 5, 5, 16, 243, 5, 5, 79, 240, 5, 5, 80, 240, 5, 5, 81, 240, 5, 5, 82, 240, 5, 5, 83, 240, 5, 5, 84, 240, 5, 5, 85, 240, 5, 5, 17, 243, 5, 5, 89, 240, 5, 5, 90, 240, 5, 5, 18, 243, 5, 5, 19, 243, 5, 5, 20, 243, 5, 5, 21, 243, 5, 5, 22, 243, 5, 5, 23, 243, 5, 5, 24, 243, 5, 5, 25, 243, 5, 5, 26, 243, 5, 5, 27, 243, 5, 5, 28, 243, 5, 5, 29, 243, 5, 5, 30, 243, 5, 5, 31, 243, 5, 5, 32, 243, 5, 5, 33, 243, 5, 5, 34, 243, 5, 5, 35, 243, 5, 5, 36, 243, 5, 5, 37, 243, 5, 5, 38, 243, 5, 5, 39, 243, 5, 5, 40, 243, 5, 5, 41, 243, 5, 5, 42, 243, 5, 5, 50, 243, 5, 5, 43, 243, 5, 5, 44, 243, 5, 5, 52, 245, 5, 5, 53, 245, 5, 5, 54, 245, 5, 5, 55, 245, 5, 5, 56, 245, 5, 5, 57, 245, 5, 5, 51, 243, 5, 5, 58, 245, 5, 5, 59, 245, 5, 5, 60, 245, 5, 5, 61, 245, 5, 5, 62, 245, 5, 5, 63, 245, 5, 5, 45, 243, 5, 5, 64, 245, 5, 5, 9, 243, 5, 5, 65, 245, 5, 5, 66, 245, 5, 5, 67, 245, 5, 5, 68, 245, 5, 5, 69, 245, 5, 5, 70, 245, 5, 5, 71, 245, 5, 5, 222, 246, 5, 5, 72, 245, 5, 5, 73, 245, 5, 5, 74, 245, 5, 5, 52, 243, 5, 5, 7, 248, 5, 5, 223, 246, 5, 5, 224, 246, 5, 5, 225, 246, 5, 5, 226, 246, 5, 5, 227, 246, 5, 5, 228, 246, 5, 5, 8, 248, 5, 5, 9, 248, 5, 5, 10, 248, 5, 5, 11, 248, 5, 5, 12, 248, 5, 5, 13, 248, 5, 5, 14, 248, 5, 5, 15, 248, 5, 5, 16, 248, 5, 5, 17, 248, 5, 5, 141, 249, 5, 5, 232, 248, 5, 5, 236, 249, 5, 5, 233, 248, 5, 5, 142, 249, 5, 5, 143, 249, 5, 5, 144, 249, 5, 5, 145, 249, 5, 5, 37, 250, 5, 5, 124, 250, 5, 5, 96, 250, 5, 5, 222, 161, 5, 5, 227, 170, 5, 5, 228, 170, 5, 5, 253, 176, 5, 5, 243, 183, 5, 5, 149, 178, 5, 5, 244, 183, 5, 5, 245, 183, 5, 5, 152, 191, 5, 5, 153, 191, 5, 5, 154, 191, 5, 5, 155, 191, 5, 5, 156, 191, 5, 5, 157, 191, 5, 5, 206, 199, 5, 5, 138, 207, 5, 5, 207, 199, 5, 5, 139, 207, 5, 5, 140, 207, 5, 5, 226, 214, 5, 5, 227, 214, 5, 5, 161, 221, 5, 5, 213, 227, 5, 5, 214, 227, 5, 5, 215, 227, 5, 5, 216, 227, 5, 5, 10, 233, 5, 5, 11, 233, 5, 5, 251, 236, 5, 5, 237, 249, 5, 5, 223, 161, 5, 5, 121, 165, 5, 5, 229, 170, 5, 5, 230, 170, 5, 5, 231, 170, 5, 5, 232, 170, 5, 5, 233, 170, 5, 5, 234, 170, 5, 5, 254, 176, 5, 5, 255, 176, 5, 5, 2, 177, 5, 5, 3, 177, 5, 5, 4, 177, 5, 5, 5, 177, 5, 5, 6, 177, 5, 5, 7, 177, 5, 5, 8, 177, 5, 5, 9, 177, 5, 5, 10, 177, 5, 5, 11, 177, 5, 5, 12, 177, 5, 5, 13, 177, 5, 5, 14, 177, 5, 5, 15, 177, 5, 5, 16, 177, 5, 5, 246, 183, 5, 5, 247, 183, 5, 5, 248, 183, 5, 5, 249, 183, 5, 5, 208, 199, 5, 5, 250, 183, 5, 5, 251, 183, 5, 5, 252, 183, 5, 5, 253, 183, 5, 5, 254, 183, 5, 5, 255, 183, 5, 5, 2, 184, 5, 5, 3, 184, 5, 5, 4, 184, 5, 5, 5, 184, 5, 5, 6, 184, 5, 5, 7, 184, 5, 5, 8, 184, 5, 5, 9, 184, 5, 5, 10, 184, 5, 5, 11, 184, 5, 5, 12, 184, 5, 5, 13, 184, 5, 5, 14, 184, 5, 5, 15, 184, 5, 5, 16, 184, 5, 5, 17, 184, 5, 5, 18, 184, 5, 5, 19, 184, 5, 5, 20, 184, 5, 5, 21, 184, 5, 5, 22, 184, 5, 5, 23, 184, 5, 5, 24, 184, 5, 5, 25, 184, 5, 5, 26, 184, 5, 5, 27, 184, 5, 5, 28, 184, 5, 5, 29, 184, 5, 5, 158, 191, 5, 5, 30, 184, 5, 5, 159, 191, 5, 5, 160, 191, 5, 5, 161, 191, 5, 5, 162, 191, 5, 5, 163, 191, 5, 5, 164, 191, 5, 5, 165, 191, 5, 5, 166, 191, 5, 5, 167, 191, 5, 5, 168, 191, 5, 5, 169, 191, 5, 5, 170, 191, 5, 5, 171, 191, 5, 5, 172, 191, 5, 5, 173, 191, 5, 5, 174, 191, 5, 5, 175, 191, 5, 5, 176, 191, 5, 5, 177, 191, 5, 5, 178, 191, 5, 5, 179, 191, 5, 5, 180, 191, 5, 5, 181, 191, 5, 5, 182, 191, 5, 5, 183, 191, 5, 5, 184, 191, 5, 5, 185, 191, 5, 5, 186, 191, 5, 5, 187, 191, 5, 5, 188, 191, 5, 5, 189, 191, 5, 5, 190, 191, 5, 5, 191, 191, 5, 5, 192, 191, 5, 5, 193, 191, 5, 5, 209, 199, 5, 5, 210, 199, 5, 5, 211, 199, 5, 5, 212, 199, 5, 5, 213, 199, 5, 5, 214, 199, 5, 5, 141, 207, 5, 5, 215, 199, 5, 5, 216, 199, 5, 5, 217, 199, 5, 5, 218, 199, 5, 5, 219, 199, 5, 5, 220, 199, 5, 5, 221, 199, 5, 5, 222, 199, 5, 5, 223, 199, 5, 5, 224, 199, 5, 5, 225, 199, 5, 5, 226, 199, 5, 5, 227, 199, 5, 5, 228, 199, 5, 5, 229, 199, 5, 5, 230, 199, 5, 5, 231, 199, 5, 5, 232, 199, 5, 5, 233, 199, 5, 5, 234, 199, 5, 5, 235, 199, 5, 5, 236, 199, 5, 5, 237, 199, 5, 5, 238, 199, 5, 5, 239, 199, 5, 5, 240, 199, 5, 5, 241, 199, 5, 5, 242, 199, 5, 5, 243, 199, 5, 5, 244, 199, 5, 5, 142, 207, 5, 5, 143, 207, 5, 5, 144, 207, 5, 5, 145, 207, 5, 5, 146, 207, 5, 5, 147, 207, 5, 5, 148, 207, 5, 5, 149, 207, 5, 5, 150, 207, 5, 5, 151, 207, 5, 5, 152, 207, 5, 5, 153, 207, 5, 5, 154, 207, 5, 5, 155, 207, 5, 5, 156, 207, 5, 5, 157, 207, 5, 5, 158, 207, 5, 5, 159, 207, 5, 5, 160, 207, 5, 5, 161, 207, 5, 5, 162, 207, 5, 5, 163, 207, 5, 5, 164, 207, 5, 5, 165, 207, 5, 5, 166, 207, 5, 5, 167, 207, 5, 5, 168, 207, 5, 5, 169, 207, 5, 5, 229, 214, 5, 5, 230, 214, 5, 5, 170, 207, 5, 5, 171, 207, 5, 5, 172, 207, 5, 5, 173, 207, 5, 5, 174, 207, 5, 5, 232, 214, 5, 5, 233, 214, 5, 5, 234, 214, 5, 5, 235, 214, 5, 5, 236, 214, 5, 5, 237, 214, 5, 5, 238, 214, 5, 5, 239, 214, 5, 5, 240, 214, 5, 5, 241, 214, 5, 5, 242, 214, 5, 5, 243, 214, 5, 5, 244, 214, 5, 5, 245, 214, 5, 5, 246, 214, 5, 5, 247, 214, 5, 5, 248, 214, 5, 5, 249, 214, 5, 5, 250, 214, 5, 5, 231, 214, 5, 5, 251, 214, 5, 5, 252, 214, 5, 5, 253, 214, 5, 5, 254, 214, 5, 5, 255, 214, 5, 5, 2, 215, 5, 5, 3, 215, 5, 5, 4, 215, 5, 5, 5, 215, 5, 5, 6, 215, 5, 5, 7, 215, 5, 5, 8, 215, 5, 5, 9, 215, 5, 5, 10, 215, 5, 5, 11, 215, 5, 5, 12, 215, 5, 5, 13, 215, 5, 5, 14, 215, 5, 5, 15, 215, 5, 5, 16, 215, 5, 5, 17, 215, 5, 5, 18, 215, 5, 5, 19, 215, 5, 5, 20, 215, 5, 5, 21, 215, 5, 5, 22, 215, 5, 5, 175, 207, 5, 5, 23, 215, 5, 5, 24, 215, 5, 5, 25, 215, 5, 5, 26, 215, 5, 5, 27, 215, 5, 5, 162, 221, 5, 5, 177, 207, 5, 5, 163, 221, 5, 5, 164, 221, 5, 5, 165, 221, 5, 5, 166, 221, 5, 5, 167, 221, 5, 5, 168, 221, 5, 5, 169, 221, 5, 5, 28, 215, 5, 5, 170, 221, 5, 5, 171, 221, 5, 5, 172, 221, 5, 5, 173, 221, 5, 5, 174, 221, 5, 5, 175, 221, 5, 5, 176, 221, 5, 5, 177, 221, 5, 5, 178, 221, 5, 5, 179, 221, 5, 5, 180, 221, 5, 5, 181, 221, 5, 5, 182, 221, 5, 5, 183, 221, 5, 5, 184, 221, 5, 5, 185, 221, 5, 5, 186, 221, 5, 5, 187, 221, 5, 5, 188, 221, 5, 5, 189, 221, 5, 5, 190, 221, 5, 5, 29, 215, 5, 5, 191, 221, 5, 5, 192, 221, 5, 5, 193, 221, 5, 5, 194, 221, 5, 5, 195, 221, 5, 5, 196, 221, 5, 5, 197, 221, 5, 5, 198, 221, 5, 5, 199, 221, 5, 5, 200, 221, 5, 5, 201, 221, 5, 5, 202, 221, 5, 5, 203, 221, 5, 5, 217, 227, 5, 5, 204, 221, 5, 5, 205, 221, 5, 5, 206, 221, 5, 5, 207, 221, 5, 5, 208, 221, 5, 5, 209, 221, 5, 5, 210, 221, 5, 5, 218, 227, 5, 5, 214, 221, 5, 5, 219, 227, 5, 5, 220, 227, 5, 5, 221, 227, 5, 5, 222, 227, 5, 5, 223, 227, 5, 5, 224, 227, 5, 5, 225, 227, 5, 5, 226, 227, 5, 5, 227, 227, 5, 5, 228, 227, 5, 5, 229, 227, 5, 5, 230, 227, 5, 5, 231, 227, 5, 5, 232, 227, 5, 5, 233, 227, 5, 5, 234, 227, 5, 5, 235, 227, 5, 5, 236, 227, 5, 5, 237, 227, 5, 5, 238, 227, 5, 5, 239, 227, 5, 5, 240, 227, 5, 5, 241, 227, 5, 5, 242, 227, 5, 5, 243, 227, 5, 5, 244, 227, 5, 5, 245, 227, 5, 5, 246, 227, 5, 5, 247, 227, 5, 5, 248, 227, 5, 5, 249, 227, 5, 5, 250, 227, 5, 5, 251, 227, 5, 5, 252, 227, 5, 5, 253, 227, 5, 5, 254, 227, 5, 5, 255, 227, 5, 5, 2, 228, 5, 5, 3, 228, 5, 5, 12, 233, 5, 5, 13, 233, 5, 5, 14, 233, 5, 5, 15, 233, 5, 5, 16, 233, 5, 5, 17, 233, 5, 5, 18, 233, 5, 5, 19, 233, 5, 5, 20, 233, 5, 5, 21, 233, 5, 5, 22, 233, 5, 5, 23, 233, 5, 5, 24, 233, 5, 5, 25, 233, 5, 5, 26, 233, 5, 5, 27, 233, 5, 5, 28, 233, 5, 5, 29, 233, 5, 5, 30, 233, 5, 5, 31, 233, 5, 5, 32, 233, 5, 5, 33, 233, 5, 5, 34, 233, 5, 5, 35, 233, 5, 5, 36, 233, 5, 5, 37, 233, 5, 5, 38, 233, 5, 5, 39, 233, 5, 5, 40, 233, 5, 5, 41, 233, 5, 5, 42, 233, 5, 5, 43, 233, 5, 5, 44, 233, 5, 5, 45, 233, 5, 5, 46, 233, 5, 5, 47, 233, 5, 5, 48, 233, 5, 5, 49, 233, 5, 5, 50, 233, 5, 5, 51, 233, 5, 5, 52, 233, 5, 5, 254, 236, 5, 5, 255, 236, 5, 5, 92, 240, 5, 5, 2, 237, 5, 5, 253, 236, 5, 5, 3, 237, 5, 5, 4, 237, 5, 5, 5, 237, 5, 5, 6, 237, 5, 5, 7, 237, 5, 5, 8, 237, 5, 5, 58, 233, 5, 5, 9, 237, 5, 5, 10, 237, 5, 5, 211, 221, 5, 5, 11, 237, 5, 5, 12, 237, 5, 5, 13, 237, 5, 5, 14, 237, 5, 5, 15, 237, 5, 5, 16, 237, 5, 5, 17, 237, 5, 5, 18, 237, 5, 5, 19, 237, 5, 5, 20, 237, 5, 5, 21, 237, 5, 5, 22, 237, 5, 5, 23, 237, 5, 5, 24, 237, 5, 5, 25, 237, 5, 5, 26, 237, 5, 5, 27, 237, 5, 5, 28, 237, 5, 5, 29, 237, 5, 5, 30, 237, 5, 5, 93, 240, 5, 5, 94, 240, 5, 5, 95, 240, 5, 5, 96, 240, 5, 5, 97, 240, 5, 5, 98, 240, 5, 5, 99, 240, 5, 5, 100, 240, 5, 5, 101, 240, 5, 5, 102, 240, 5, 5, 103, 240, 5, 5, 91, 240, 5, 5, 104, 240, 5, 5, 105, 240, 5, 5, 106, 240, 5, 5, 107, 240, 5, 5, 108, 240, 5, 5, 109, 240, 5, 5, 110, 240, 5, 5, 111, 240, 5, 5, 112, 240, 5, 5, 113, 240, 5, 5, 114, 240, 5, 5, 115, 240, 5, 5, 31, 237, 5, 5, 116, 240, 5, 5, 53, 243, 5, 5, 54, 243, 5, 5, 55, 243, 5, 5, 56, 243, 5, 5, 57, 243, 5, 5, 58, 243, 5, 5, 59, 243, 5, 5, 60, 243, 5, 5, 61, 243, 5, 5, 62, 243, 5, 5, 80, 245, 5, 5, 81, 245, 5, 5, 82, 245, 5, 5, 83, 245, 5, 5, 117, 240, 5, 5, 84, 245, 5, 5, 85, 245, 5, 5, 86, 245, 5, 5, 87, 245, 5, 5, 88, 245, 5, 5, 89, 245, 5, 5, 232, 246, 5, 5, 233, 246, 5, 5, 234, 246, 5, 5, 235, 246, 5, 5, 92, 245, 5, 5, 236, 246, 5, 5, 93, 245, 5, 5, 237, 246, 5, 5, 20, 248, 5, 5, 21, 248, 5, 5, 22, 248, 5, 5, 23, 248, 5, 5, 24, 248, 5, 5, 25, 248, 5, 5, 26, 248, 5, 5, 19, 248, 5, 5, 236, 248, 5, 5, 237, 248, 5, 5, 238, 248, 5, 5, 239, 248, 5, 5, 240, 248, 5, 5, 241, 248, 5, 5, 146, 249, 5, 5, 238, 249, 5, 5, 38, 250, 5, 5, 39, 250, 5, 5, 40, 250, 5, 5, 224, 161, 5, 5, 17, 177, 5, 5, 18, 177, 5, 5, 31, 184, 5, 5, 32, 184, 5, 5, 200, 191, 5, 5, 250, 199, 5, 5, 251, 199, 5, 5, 252, 199, 5, 5, 253, 199, 5, 5, 94, 245, 5, 5, 242, 248, 5, 5, 225, 161, 5, 5, 20, 177, 5, 5, 21, 177, 5, 5, 33, 184, 5, 5, 201, 191, 5, 5, 202, 191, 5, 5, 203, 191, 5, 5, 204, 191, 5, 5, 205, 191, 5, 5, 254, 199, 5, 5, 255, 199, 5, 5, 2, 200, 5, 5, 178, 207, 5, 5, 179, 207, 5, 5, 215, 221, 5, 5, 216, 221, 5, 5, 217, 221, 5, 5, 218, 221, 5, 5, 6, 228, 5, 5, 7, 228, 5, 5, 8, 228, 5, 5, 9, 228, 5, 5, 243, 248, 5, 5, 226, 161, 5, 5, 185, 159, 5, 5, 236, 170, 5, 5, 23, 177, 5, 5, 24, 177, 5, 5, 237, 170, 5, 5, 25, 177, 5, 5, 26, 177, 5, 5, 27, 177, 5, 5, 28, 177, 5, 5, 35, 184, 5, 5, 36, 184, 5, 5, 37, 184, 5, 5, 38, 184, 5, 5, 39, 184, 5, 5, 40, 184, 5, 5, 41, 184, 5, 5, 42, 184, 5, 5, 43, 184, 5, 5, 44, 184, 5, 5, 45, 184, 5, 5, 46, 184, 5, 5, 47, 184, 5, 5, 48, 184, 5, 5, 49, 184, 5, 5, 50, 184, 5, 5, 51, 184, 5, 5, 52, 184, 5, 5, 53, 184, 5, 5, 54, 184, 5, 5, 55, 184, 5, 5, 56, 184, 5, 5, 57, 184, 5, 5, 58, 184, 5, 5, 59, 184, 5, 5, 60, 184, 5, 5, 61, 184, 5, 5, 206, 191, 5, 5, 207, 191, 5, 5, 208, 191, 5, 5, 209, 191, 5, 5, 210, 191, 5, 5, 211, 191, 5, 5, 212, 191, 5, 5, 213, 191, 5, 5, 214, 191, 5, 5, 215, 191, 5, 5, 216, 191, 5, 5, 217, 191, 5, 5, 218, 191, 5, 5, 219, 191, 5, 5, 220, 191, 5, 5, 221, 191, 5, 5, 222, 191, 5, 5, 223, 191, 5, 5, 224, 191, 5, 5, 225, 191, 5, 5, 226, 191, 5, 5, 227, 191, 5, 5, 228, 191, 5, 5, 229, 191, 5, 5, 230, 191, 5, 5, 231, 191, 5, 5, 232, 191, 5, 5, 233, 191, 5, 5, 234, 191, 5, 5, 235, 191, 5, 5, 236, 191, 5, 5, 237, 191, 5, 5, 238, 191, 5, 5, 239, 191, 5, 5, 240, 191, 5, 5, 241, 191, 5, 5, 242, 191, 5, 5, 243, 191, 5, 5, 244, 191, 5, 5, 247, 191, 5, 5, 245, 191, 5, 5, 3, 200, 5, 5, 4, 200, 5, 5, 5, 200, 5, 5, 6, 200, 5, 5, 7, 200, 5, 5, 8, 200, 5, 5, 9, 200, 5, 5, 10, 200, 5, 5, 11, 200, 5, 5, 12, 200, 5, 5, 13, 200, 5, 5, 14, 200, 5, 5, 15, 200, 5, 5, 16, 200, 5, 5, 17, 200, 5, 5, 18, 200, 5, 5, 19, 200, 5, 5, 20, 200, 5, 5, 21, 200, 5, 5, 22, 200, 5, 5, 23, 200, 5, 5, 24, 200, 5, 5, 25, 200, 5, 5, 26, 200, 5, 5, 27, 200, 5, 5, 180, 207, 5, 5, 181, 207, 5, 5, 182, 207, 5, 5, 183, 207, 5, 5, 184, 207, 5, 5, 185, 207, 5, 5, 186, 207, 5, 5, 187, 207, 5, 5, 188, 207, 5, 5, 189, 207, 5, 5, 190, 207, 5, 5, 191, 207, 5, 5, 192, 207, 5, 5, 28, 200, 5, 5, 193, 207, 5, 5, 194, 207, 5, 5, 195, 207, 5, 5, 196, 207, 5, 5, 197, 207, 5, 5, 198, 207, 5, 5, 199, 207, 5, 5, 200, 207, 5, 5, 201, 207, 5, 5, 202, 207, 5, 5, 30, 215, 5, 5, 203, 207, 5, 5, 204, 207, 5, 5, 205, 207, 5, 5, 220, 221, 5, 5, 31, 215, 5, 5, 32, 215, 5, 5, 33, 215, 5, 5, 34, 215, 5, 5, 35, 215, 5, 5, 36, 215, 5, 5, 37, 215, 5, 5, 38, 215, 5, 5, 39, 215, 5, 5, 40, 215, 5, 5, 41, 215, 5, 5, 42, 215, 5, 5, 43, 215, 5, 5, 44, 215, 5, 5, 45, 215, 5, 5, 46, 215, 5, 5, 47, 215, 5, 5, 48, 215, 5, 5, 49, 215, 5, 5, 50, 215, 5, 5, 51, 215, 5, 5, 52, 215, 5, 5, 53, 215, 5, 5, 54, 215, 5, 5, 55, 215, 5, 5, 56, 215, 5, 5, 30, 200, 5, 5, 57, 215, 5, 5, 58, 215, 5, 5, 59, 215, 5, 5, 221, 221, 5, 5, 222, 221, 5, 5, 223, 221, 5, 5, 224, 221, 5, 5, 225, 221, 5, 5, 226, 221, 5, 5, 227, 221, 5, 5, 228, 221, 5, 5, 229, 221, 5, 5, 230, 221, 5, 5, 231, 221, 5, 5, 232, 221, 5, 5, 233, 221, 5, 5, 234, 221, 5, 5, 235, 221, 5, 5, 236, 221, 5, 5, 237, 221, 5, 5, 238, 221, 5, 5, 239, 221, 5, 5, 240, 221, 5, 5, 241, 221, 5, 5, 61, 215, 5, 5, 242, 221, 5, 5, 243, 221, 5, 5, 244, 221, 5, 5, 12, 228, 5, 5, 13, 228, 5, 5, 14, 228, 5, 5, 15, 228, 5, 5, 16, 228, 5, 5, 17, 228, 5, 5, 18, 228, 5, 5, 19, 228, 5, 5, 20, 228, 5, 5, 21, 228, 5, 5, 22, 228, 5, 5, 23, 228, 5, 5, 24, 228, 5, 5, 25, 228, 5, 5, 26, 228, 5, 5, 27, 228, 5, 5, 28, 228, 5, 5, 29, 228, 5, 5, 30, 228, 5, 5, 31, 228, 5, 5, 32, 228, 5, 5, 59, 233, 5, 5, 33, 228, 5, 5, 60, 233, 5, 5, 61, 233, 5, 5, 62, 233, 5, 5, 63, 233, 5, 5, 64, 233, 5, 5, 65, 233, 5, 5, 66, 233, 5, 5, 67, 233, 5, 5, 68, 233, 5, 5, 69, 233, 5, 5, 70, 233, 5, 5, 71, 233, 5, 5, 72, 233, 5, 5, 73, 233, 5, 5, 74, 233, 5, 5, 75, 233, 5, 5, 76, 233, 5, 5, 35, 237, 5, 5, 36, 237, 5, 5, 37, 237, 5, 5, 38, 237, 5, 5, 39, 237, 5, 5, 40, 237, 5, 5, 41, 237, 5, 5, 42, 237, 5, 5, 43, 237, 5, 5, 44, 237, 5, 5, 45, 237, 5, 5, 46, 237, 5, 5, 79, 233, 5, 5, 47, 237, 5, 5, 77, 233, 5, 5, 48, 237, 5, 5, 118, 240, 5, 5, 119, 240, 5, 5, 120, 240, 5, 5, 121, 240, 5, 5, 122, 240, 5, 5, 123, 240, 5, 5, 124, 240, 5, 5, 125, 240, 5, 5, 126, 240, 5, 5, 127, 240, 5, 5, 128, 240, 5, 5, 129, 240, 5, 5, 130, 240, 5, 5, 65, 243, 5, 5, 66, 243, 5, 5, 67, 243, 5, 5, 68, 243, 5, 5, 69, 243, 5, 5, 70, 243, 5, 5, 95, 245, 5, 5, 96, 245, 5, 5, 97, 245, 5, 5, 98, 245, 5, 5, 99, 245, 5, 5, 100, 245, 5, 5, 239, 246, 5, 5, 240, 246, 5, 5, 241, 246, 5, 5, 242, 246, 5, 5, 27, 248, 5, 5, 28, 248, 5, 5, 244, 248, 5, 5, 29, 248, 5, 5, 245, 248, 5, 5, 148, 249, 5, 5, 149, 249, 5, 5, 150, 249, 5, 5, 151, 249, 5, 5, 152, 249, 5, 5, 41, 250, 5, 5, 227, 161, 5, 5, 228, 161, 5, 5, 229, 161, 5, 5, 31, 177, 5, 5, 248, 191, 5, 5, 31, 200, 5, 5, 32, 200, 5, 5, 206, 207, 5, 5, 51, 237, 5, 5, 131, 240, 5, 5, 132, 240, 5, 5, 30, 248, 5, 5, 153, 249, 5, 5, 123, 165, 5, 5, 32, 177, 5, 5, 65, 184, 5, 5, 66, 184, 5, 5, 249, 191, 5, 5, 250, 191, 5, 5, 251, 191, 5, 5, 252, 191, 5, 5, 253, 191, 5, 5, 254, 191, 5, 5, 34, 200, 5, 5, 72, 190, 5, 5, 35, 200, 5, 5, 36, 200, 5, 5, 33, 200, 5, 5, 37, 200, 5, 5, 207, 207, 5, 5, 208, 207, 5, 5, 62, 215, 5, 5, 63, 215, 5, 5, 64, 215, 5, 5, 65, 215, 5, 5, 66, 215, 5, 5, 248, 221, 5, 5, 249, 221, 5, 5, 250, 221, 5, 5, 251, 221, 5, 5, 38, 228, 5, 5, 39, 228, 5, 5, 40, 228, 5, 5, 253, 221, 5, 5, 41, 228, 5, 5, 82, 233, 5, 5, 83, 233, 5, 5, 84, 233, 5, 5, 85, 233, 5, 5, 86, 233, 5, 5, 52, 237, 5, 5, 53, 237, 5, 5, 54, 237, 5, 5, 55, 237, 5, 5, 133, 240, 5, 5, 134, 240, 5, 5, 135, 240, 5, 5, 136, 240, 5, 5, 137, 240, 5, 5, 71, 243, 5, 5, 72, 243, 5, 5, 73, 243, 5, 5, 101, 245, 5, 5, 102, 245, 5, 5, 243, 246, 5, 5, 244, 246, 5, 5, 154, 249, 5, 5, 111, 158, 5, 5, 230, 161, 5, 5, 124, 165, 5, 5, 238, 170, 5, 5, 239, 170, 5, 5, 16, 176, 5, 5, 33, 177, 5, 5, 34, 177, 5, 5, 35, 177, 5, 5, 67, 184, 5, 5, 2, 192, 5, 5, 38, 200, 5, 5, 39, 200, 5, 5, 210, 207, 5, 5, 67, 215, 5, 5, 254, 221, 5, 5, 255, 221, 5, 5, 125, 165, 5, 5, 36, 177, 5, 5, 37, 177, 5, 5, 3, 192, 5, 5, 4, 192, 5, 5, 5, 192, 5, 5, 6, 192, 5, 5, 7, 192, 5, 5, 40, 200, 5, 5, 41, 200, 5, 5, 211, 207, 5, 5, 42, 200, 5, 5, 43, 200, 5, 5, 212, 207, 5, 5, 213, 207, 5, 5, 214, 207, 5, 5, 215, 207, 5, 5, 216, 207, 5, 5, 217, 207, 5, 5, 218, 207, 5, 5, 219, 207, 5, 5, 220, 207, 5, 5, 68, 215, 5, 5, 69, 215, 5, 5, 70, 215, 5, 5, 71, 215, 5, 5, 2, 222, 5, 5, 3, 222, 5, 5, 4, 222, 5, 5, 5, 222, 5, 5, 6, 222, 5, 5, 43, 228, 5, 5, 87, 233, 5, 5, 88, 233, 5, 5, 57, 237, 5, 5, 138, 240, 5, 5, 139, 240, 5, 5, 74, 243, 5, 5, 75, 243, 5, 5, 76, 243, 5, 5, 103, 245, 5, 5, 246, 246, 5, 5, 247, 246, 5, 5, 31, 248, 5, 5, 32, 248, 5, 5, 155, 249, 5, 5, 128, 165, 5, 5, 129, 165, 5, 5, 38, 177, 5, 5, 39, 177, 5, 5, 40, 177, 5, 5, 41, 177, 5, 5, 42, 177, 5, 5, 43, 177, 5, 5, 44, 177, 5, 5, 68, 184, 5, 5, 69, 184, 5, 5, 70, 184, 5, 5, 71, 184, 5, 5, 72, 184, 5, 5, 73, 184, 5, 5, 74, 184, 5, 5, 75, 184, 5, 5, 76, 184, 5, 5, 77, 184, 5, 5, 78, 184, 5, 5, 79, 184, 5, 5, 80, 184, 5, 5, 81, 184, 5, 5, 82, 184, 5, 5, 83, 184, 5, 5, 84, 184, 5, 5, 85, 184, 5, 5, 9, 192, 5, 5, 10, 192, 5, 5, 11, 192, 5, 5, 12, 192, 5, 5, 13, 192, 5, 5, 14, 192, 5, 5, 15, 192, 5, 5, 16, 192, 5, 5, 17, 192, 5, 5, 18, 192, 5, 5, 19, 192, 5, 5, 20, 192, 5, 5, 21, 192, 5, 5, 22, 192, 5, 5, 23, 192, 5, 5, 24, 192, 5, 5, 25, 192, 5, 5, 26, 192, 5, 5, 27, 192, 5, 5, 28, 192, 5, 5, 29, 192, 5, 5, 30, 192, 5, 5, 31, 192, 5, 5, 32, 192, 5, 5, 33, 192, 5, 5, 44, 200, 5, 5, 45, 200, 5, 5, 46, 200, 5, 5, 47, 200, 5, 5, 48, 200, 5, 5, 49, 200, 5, 5, 50, 200, 5, 5, 51, 200, 5, 5, 52, 200, 5, 5, 53, 200, 5, 5, 221, 207, 5, 5, 222, 207, 5, 5, 54, 200, 5, 5, 55, 200, 5, 5, 56, 200, 5, 5, 57, 200, 5, 5, 58, 200, 5, 5, 59, 200, 5, 5, 60, 200, 5, 5, 61, 200, 5, 5, 62, 200, 5, 5, 63, 200, 5, 5, 64, 200, 5, 5, 65, 200, 5, 5, 66, 200, 5, 5, 67, 200, 5, 5, 68, 200, 5, 5, 69, 200, 5, 5, 70, 200, 5, 5, 71, 200, 5, 5, 72, 200, 5, 5, 73, 200, 5, 5, 74, 200, 5, 5, 75, 200, 5, 5, 76, 200, 5, 5, 77, 200, 5, 5, 78, 200, 5, 5, 79, 200, 5, 5, 80, 200, 5, 5, 81, 200, 5, 5, 82, 200, 5, 5, 83, 200, 5, 5, 84, 200, 5, 5, 85, 200, 5, 5, 86, 200, 5, 5, 223, 207, 5, 5, 224, 207, 5, 5, 225, 207, 5, 5, 226, 207, 5, 5, 227, 207, 5, 5, 228, 207, 5, 5, 229, 207, 5, 5, 230, 207, 5, 5, 231, 207, 5, 5, 232, 207, 5, 5, 233, 207, 5, 5, 234, 207, 5, 5, 235, 207, 5, 5, 236, 207, 5, 5, 237, 207, 5, 5, 238, 207, 5, 5, 239, 207, 5, 5, 240, 207, 5, 5, 241, 207, 5, 5, 242, 207, 5, 5, 243, 207, 5, 5, 244, 207, 5, 5, 245, 207, 5, 5, 246, 207, 5, 5, 247, 207, 5, 5, 248, 207, 5, 5, 249, 207, 5, 5, 250, 207, 5, 5, 251, 207, 5, 5, 252, 207, 5, 5, 253, 207, 5, 5, 254, 207, 5, 5, 255, 207, 5, 5, 2, 208, 5, 5, 3, 208, 5, 5, 4, 208, 5, 5, 5, 208, 5, 5, 6, 208, 5, 5, 7, 208, 5, 5, 8, 208, 5, 5, 9, 208, 5, 5, 10, 208, 5, 5, 72, 215, 5, 5, 73, 215, 5, 5, 74, 215, 5, 5, 75, 215, 5, 5, 76, 215, 5, 5, 77, 215, 5, 5, 78, 215, 5, 5, 79, 215, 5, 5, 80, 215, 5, 5, 81, 215, 5, 5, 7, 222, 5, 5, 82, 215, 5, 5, 83, 215, 5, 5, 84, 215, 5, 5, 85, 215, 5, 5, 86, 215, 5, 5, 87, 215, 5, 5, 88, 215, 5, 5, 89, 215, 5, 5, 90, 215, 5, 5, 91, 215, 5, 5, 11, 208, 5, 5, 92, 215, 5, 5, 93, 215, 5, 5, 94, 215, 5, 5, 95, 215, 5, 5, 96, 215, 5, 5, 97, 215, 5, 5, 98, 215, 5, 5, 99, 215, 5, 5, 100, 215, 5, 5, 101, 215, 5, 5, 102, 215, 5, 5, 103, 215, 5, 5, 104, 215, 5, 5, 105, 215, 5, 5, 8, 222, 5, 5, 9, 222, 5, 5, 10, 222, 5, 5, 11, 222, 5, 5, 12, 222, 5, 5, 13, 222, 5, 5, 14, 222, 5, 5, 15, 222, 5, 5, 16, 222, 5, 5, 17, 222, 5, 5, 18, 222, 5, 5, 19, 222, 5, 5, 20, 222, 5, 5, 21, 222, 5, 5, 22, 222, 5, 5, 23, 222, 5, 5, 24, 222, 5, 5, 25, 222, 5, 5, 26, 222, 5, 5, 27, 222, 5, 5, 28, 222, 5, 5, 29, 222, 5, 5, 30, 222, 5, 5, 31, 222, 5, 5, 32, 222, 5, 5, 33, 222, 5, 5, 34, 222, 5, 5, 35, 222, 5, 5, 36, 222, 5, 5, 37, 222, 5, 5, 38, 222, 5, 5, 39, 222, 5, 5, 40, 222, 5, 5, 41, 222, 5, 5, 42, 222, 5, 5, 43, 222, 5, 5, 44, 222, 5, 5, 45, 222, 5, 5, 46, 222, 5, 5, 47, 222, 5, 5, 48, 222, 5, 5, 49, 222, 5, 5, 50, 222, 5, 5, 51, 222, 5, 5, 58, 222, 5, 5, 45, 228, 5, 5, 46, 228, 5, 5, 47, 228, 5, 5, 48, 228, 5, 5, 49, 228, 5, 5, 50, 228, 5, 5, 51, 228, 5, 5, 52, 228, 5, 5, 53, 228, 5, 5, 54, 228, 5, 5, 55, 228, 5, 5, 56, 228, 5, 5, 57, 228, 5, 5, 52, 222, 5, 5, 58, 228, 5, 5, 59, 228, 5, 5, 60, 228, 5, 5, 61, 228, 5, 5, 62, 228, 5, 5, 63, 228, 5, 5, 64, 228, 5, 5, 65, 228, 5, 5, 66, 228, 5, 5, 67, 228, 5, 5, 68, 228, 5, 5, 69, 228, 5, 5, 70, 228, 5, 5, 71, 228, 5, 5, 59, 222, 5, 5, 72, 228, 5, 5, 73, 228, 5, 5, 74, 228, 5, 5, 75, 228, 5, 5, 76, 228, 5, 5, 77, 228, 5, 5, 78, 228, 5, 5, 79, 228, 5, 5, 80, 228, 5, 5, 81, 228, 5, 5, 82, 228, 5, 5, 90, 233, 5, 5, 91, 233, 5, 5, 92, 233, 5, 5, 93, 233, 5, 5, 94, 233, 5, 5, 95, 233, 5, 5, 96, 233, 5, 5, 97, 233, 5, 5, 98, 233, 5, 5, 99, 233, 5, 5, 100, 233, 5, 5, 101, 233, 5, 5, 102, 233, 5, 5, 103, 233, 5, 5, 104, 233, 5, 5, 105, 233, 5, 5, 85, 228, 5, 5, 106, 233, 5, 5, 107, 233, 5, 5, 108, 233, 5, 5, 109, 233, 5, 5, 110, 233, 5, 5, 111, 233, 5, 5, 112, 233, 5, 5, 113, 233, 5, 5, 114, 233, 5, 5, 115, 233, 5, 5, 116, 233, 5, 5, 117, 233, 5, 5, 118, 233, 5, 5, 119, 233, 5, 5, 59, 237, 5, 5, 60, 237, 5, 5, 61, 237, 5, 5, 62, 237, 5, 5, 63, 237, 5, 5, 64, 237, 5, 5, 65, 237, 5, 5, 66, 237, 5, 5, 67, 237, 5, 5, 68, 237, 5, 5, 69, 237, 5, 5, 70, 237, 5, 5, 71, 237, 5, 5, 72, 237, 5, 5, 73, 237, 5, 5, 74, 237, 5, 5, 75, 237, 5, 5, 76, 237, 5, 5, 77, 237, 5, 5, 78, 237, 5, 5, 79, 237, 5, 5, 80, 237, 5, 5, 81, 237, 5, 5, 82, 237, 5, 5, 83, 237, 5, 5, 84, 237, 5, 5, 85, 237, 5, 5, 86, 237, 5, 5, 140, 240, 5, 5, 141, 240, 5, 5, 142, 240, 5, 5, 143, 240, 5, 5, 144, 240, 5, 5, 145, 240, 5, 5, 104, 245, 5, 5, 146, 240, 5, 5, 88, 237, 5, 5, 147, 240, 5, 5, 148, 240, 5, 5, 149, 240, 5, 5, 150, 240, 5, 5, 151, 240, 5, 5, 77, 243, 5, 5, 152, 240, 5, 5, 153, 240, 5, 5, 154, 240, 5, 5, 155, 240, 5, 5, 156, 240, 5, 5, 157, 240, 5, 5, 158, 240, 5, 5, 159, 240, 5, 5, 160, 240, 5, 5, 161, 240, 5, 5, 162, 240, 5, 5, 163, 240, 5, 5, 164, 240, 5, 5, 165, 240, 5, 5, 166, 240, 5, 5, 78, 243, 5, 5, 79, 243, 5, 5, 80, 243, 5, 5, 81, 243, 5, 5, 82, 243, 5, 5, 83, 243, 5, 5, 84, 243, 5, 5, 85, 243, 5, 5, 86, 243, 5, 5, 87, 243, 5, 5, 88, 243, 5, 5, 89, 243, 5, 5, 90, 243, 5, 5, 91, 243, 5, 5, 92, 243, 5, 5, 93, 243, 5, 5, 94, 243, 5, 5, 95, 243, 5, 5, 96, 243, 5, 5, 97, 243, 5, 5, 98, 243, 5, 5, 99, 243, 5, 5, 105, 245, 5, 5, 106, 245, 5, 5, 107, 245, 5, 5, 108, 245, 5, 5, 109, 245, 5, 5, 110, 245, 5, 5, 111, 245, 5, 5, 112, 245, 5, 5, 113, 245, 5, 5, 114, 245, 5, 5, 115, 245, 5, 5, 248, 246, 5, 5, 249, 246, 5, 5, 250, 246, 5, 5, 251, 246, 5, 5, 252, 246, 5, 5, 253, 246, 5, 5, 254, 246, 5, 5, 255, 246, 5, 5, 4, 247, 5, 5, 33, 248, 5, 5, 34, 248, 5, 5, 35, 248, 5, 5, 36, 248, 5, 5, 37, 248, 5, 5, 38, 248, 5, 5, 39, 248, 5, 5, 40, 248, 5, 5, 41, 248, 5, 5, 42, 248, 5, 5, 247, 248, 5, 5, 248, 248, 5, 5, 249, 248, 5, 5, 250, 248, 5, 5, 251, 248, 5, 5, 252, 248, 5, 5, 156, 249, 5, 5, 157, 249, 5, 5, 158, 249, 5, 5, 239, 249, 5, 5, 240, 249, 5, 5, 42, 250, 5, 5, 43, 250, 5, 5, 44, 250, 5, 5, 125, 250, 5, 5, 46, 157, 5, 5, 112, 158, 5, 5, 113, 158, 5, 5, 114, 158, 5, 5, 115, 158, 5, 5, 116, 158, 5, 5, 187, 159, 5, 5, 188, 159, 5, 5, 189, 159, 5, 5, 190, 159, 5, 5, 191, 159, 5, 5, 192, 159, 5, 5, 193, 159, 5, 5, 194, 159, 5, 5, 195, 159, 5, 5, 196, 159, 5, 5, 197, 159, 5, 5, 198, 159, 5, 5, 231, 161, 5, 5, 232, 161, 5, 5, 233, 161, 5, 5, 234, 161, 5, 5, 235, 161, 5, 5, 236, 161, 5, 5, 237, 161, 5, 5, 238, 161, 5, 5, 239, 161, 5, 5, 240, 161, 5, 5, 241, 161, 5, 5, 242, 161, 5, 5, 243, 161, 5, 5, 244, 161, 5, 5, 245, 161, 5, 5, 130, 165, 5, 5, 131, 165, 5, 5, 132, 165, 5, 5, 133, 165, 5, 5, 134, 165, 5, 5, 135, 165, 5, 5, 136, 165, 5, 5, 137, 165, 5, 5, 138, 165, 5, 5, 139, 165, 5, 5, 140, 165, 5, 5, 141, 165, 5, 5, 142, 165, 5, 5, 143, 165, 5, 5, 144, 165, 5, 5, 145, 165, 5, 5, 146, 165, 5, 5, 147, 165, 5, 5, 240, 170, 5, 5, 241, 170, 5, 5, 242, 170, 5, 5, 243, 170, 5, 5, 244, 170, 5, 5, 245, 170, 5, 5, 246, 170, 5, 5, 247, 170, 5, 5, 248, 170, 5, 5, 249, 170, 5, 5, 250, 170, 5, 5, 251, 170, 5, 5, 252, 170, 5, 5, 253, 170, 5, 5, 254, 170, 5, 5, 255, 170, 5, 5, 2, 171, 5, 5, 3, 171, 5, 5, 4, 171, 5, 5, 5, 171, 5, 5, 6, 171, 5, 5, 7, 171, 5, 5, 8, 171, 5, 5, 46, 177, 5, 5, 47, 177, 5, 5, 48, 177, 5, 5, 49, 177, 5, 5, 50, 177, 5, 5, 51, 177, 5, 5, 52, 177, 5, 5, 53, 177, 5, 5, 54, 177, 5, 5, 55, 177, 5, 5, 56, 177, 5, 5, 57, 177, 5, 5, 58, 177, 5, 5, 87, 184, 5, 5, 88, 184, 5, 5, 89, 184, 5, 5, 90, 184, 5, 5, 91, 184, 5, 5, 92, 184, 5, 5, 93, 184, 5, 5, 94, 184, 5, 5, 95, 184, 5, 5, 96, 184, 5, 5, 97, 184, 5, 5, 98, 184, 5, 5, 99, 184, 5, 5, 100, 184, 5, 5, 101, 184, 5, 5, 102, 184, 5, 5, 103, 184, 5, 5, 104, 184, 5, 5, 105, 184, 5, 5, 106, 184, 5, 5, 37, 192, 5, 5, 38, 192, 5, 5, 39, 192, 5, 5, 40, 192, 5, 5, 41, 192, 5, 5, 42, 192, 5, 5, 43, 192, 5, 5, 44, 192, 5, 5, 45, 192, 5, 5, 46, 192, 5, 5, 47, 192, 5, 5, 48, 192, 5, 5, 49, 192, 5, 5, 50, 192, 5, 5, 51, 192, 5, 5, 52, 192, 5, 5, 53, 192, 5, 5, 54, 192, 5, 5, 55, 192, 5, 5, 56, 192, 5, 5, 90, 200, 5, 5, 91, 200, 5, 5, 92, 200, 5, 5, 93, 200, 5, 5, 94, 200, 5, 5, 95, 200, 5, 5, 96, 200, 5, 5, 97, 200, 5, 5, 98, 200, 5, 5, 13, 208, 5, 5, 14, 208, 5, 5, 15, 208, 5, 5, 16, 208, 5, 5, 17, 208, 5, 5, 109, 215, 5, 5, 110, 215, 5, 5, 111, 215, 5, 5, 112, 215, 5, 5, 113, 215, 5, 5, 114, 215, 5, 5, 60, 222, 5, 5, 61, 222, 5, 5, 62, 222, 5, 5, 170, 240, 5, 5, 148, 165, 5, 5, 107, 184, 5, 5, 57, 192, 5, 5, 58, 192, 5, 5, 59, 192, 5, 5, 18, 208, 5, 5, 115, 215, 5, 5, 63, 222, 5, 5, 124, 233, 5, 5, 125, 233, 5, 5, 126, 233, 5, 5, 89, 237, 5, 5, 171, 240, 5, 5, 5, 247, 5, 5, 43, 248, 5, 5, 149, 165, 5, 5, 108, 184, 5, 5, 109, 184, 5, 5, 60, 192, 5, 5, 19, 208, 5, 5, 20, 208, 5, 5, 64, 222, 5, 5, 65, 222, 5, 5, 66, 222, 5, 5, 127, 233, 5, 5, 90, 237, 5, 5, 105, 243, 5, 5, 159, 249, 5, 5, 45, 250, 5, 5, 97, 250, 5, 5, 150, 165, 5, 5, 9, 171, 5, 5, 110, 184, 5, 5, 61, 192, 5, 5, 62, 192, 5, 5, 63, 192, 5, 5, 64, 192, 5, 5, 65, 192, 5, 5, 66, 192, 5, 5, 99, 200, 5, 5, 100, 200, 5, 5, 101, 200, 5, 5, 102, 200, 5, 5, 22, 208, 5, 5, 23, 208, 5, 5, 24, 208, 5, 5, 25, 208, 5, 5, 26, 208, 5, 5, 116, 215, 5, 5, 117, 215, 5, 5, 118, 215, 5, 5, 119, 215, 5, 5, 86, 228, 5, 5, 69, 222, 5, 5, 87, 228, 5, 5, 88, 228, 5, 5, 128, 233, 5, 5, 129, 233, 5, 5, 130, 233, 5, 5, 131, 233, 5, 5, 132, 233, 5, 5, 91, 237, 5, 5, 92, 237, 5, 5, 173, 240, 5, 5, 172, 240, 5, 5, 151, 165, 5, 5, 111, 184, 5, 5, 112, 184, 5, 5, 113, 184, 5, 5, 69, 192, 5, 5, 70, 192, 5, 5, 104, 200, 5, 5, 105, 200, 5, 5, 106, 200, 5, 5, 107, 200, 5, 5, 108, 200, 5, 5, 109, 200, 5, 5, 27, 208, 5, 5, 28, 208, 5, 5, 29, 208, 5, 5, 30, 208, 5, 5, 31, 208, 5, 5, 32, 208, 5, 5, 33, 208, 5, 5, 120, 215, 5, 5, 121, 215, 5, 5, 122, 215, 5, 5, 70, 222, 5, 5, 71, 222, 5, 5, 91, 228, 5, 5, 92, 228, 5, 5, 93, 228, 5, 5, 94, 228, 5, 5, 133, 233, 5, 5, 134, 233, 5, 5, 135, 233, 5, 5, 94, 237, 5, 5, 95, 237, 5, 5, 96, 237, 5, 5, 174, 240, 5, 5, 161, 249, 5, 5, 46, 250, 5, 5, 152, 165, 5, 5, 59, 177, 5, 5, 60, 177, 5, 5, 61, 177, 5, 5, 114, 184, 5, 5, 115, 184, 5, 5, 116, 184, 5, 5, 117, 184, 5, 5, 71, 192, 5, 5, 72, 192, 5, 5, 73, 192, 5, 5, 74, 192, 5, 5, 75, 192, 5, 5, 76, 192, 5, 5, 77, 192, 5, 5, 78, 192, 5, 5, 79, 192, 5, 5, 80, 192, 5, 5, 111, 200, 5, 5, 112, 200, 5, 5, 113, 200, 5, 5, 35, 208, 5, 5, 114, 200, 5, 5, 115, 200, 5, 5, 116, 200, 5, 5, 117, 200, 5, 5, 118, 200, 5, 5, 119, 200, 5, 5, 120, 200, 5, 5, 121, 200, 5, 5, 122, 200, 5, 5, 123, 200, 5, 5, 124, 200, 5, 5, 125, 200, 5, 5, 126, 200, 5, 5, 127, 200, 5, 5, 128, 200, 5, 5, 36, 208, 5, 5, 37, 208, 5, 5, 38, 208, 5, 5, 39, 208, 5, 5, 40, 208, 5, 5, 41, 208, 5, 5, 42, 208, 5, 5, 43, 208, 5, 5, 44, 208, 5, 5, 45, 208, 5, 5, 46, 208, 5, 5, 47, 208, 5, 5, 48, 208, 5, 5, 123, 215, 5, 5, 124, 215, 5, 5, 125, 215, 5, 5, 126, 215, 5, 5, 127, 215, 5, 5, 128, 215, 5, 5, 129, 215, 5, 5, 130, 215, 5, 5, 131, 215, 5, 5, 132, 215, 5, 5, 72, 222, 5, 5, 73, 222, 5, 5, 74, 222, 5, 5, 75, 222, 5, 5, 76, 222, 5, 5, 77, 222, 5, 5, 78, 222, 5, 5, 79, 222, 5, 5, 80, 222, 5, 5, 81, 222, 5, 5, 82, 222, 5, 5, 83, 222, 5, 5, 84, 222, 5, 5, 85, 222, 5, 5, 86, 222, 5, 5, 87, 222, 5, 5, 88, 222, 5, 5, 89, 222, 5, 5, 90, 222, 5, 5, 91, 222, 5, 5, 93, 222, 5, 5, 95, 228, 5, 5, 96, 228, 5, 5, 97, 228, 5, 5, 98, 228, 5, 5, 99, 228, 5, 5, 100, 228, 5, 5, 101, 228, 5, 5, 102, 228, 5, 5, 136, 233, 5, 5, 137, 233, 5, 5, 138, 233, 5, 5, 139, 233, 5, 5, 140, 233, 5, 5, 141, 233, 5, 5, 142, 233, 5, 5, 143, 233, 5, 5, 97, 237, 5, 5, 98, 237, 5, 5, 99, 237, 5, 5, 141, 235, 5, 5, 100, 237, 5, 5, 101, 237, 5, 5, 102, 237, 5, 5, 103, 237, 5, 5, 175, 240, 5, 5, 176, 240, 5, 5, 177, 240, 5, 5, 178, 240, 5, 5, 179, 240, 5, 5, 180, 240, 5, 5, 181, 240, 5, 5, 107, 243, 5, 5, 108, 243, 5, 5, 109, 243, 5, 5, 117, 245, 5, 5, 118, 245, 5, 5, 119, 245, 5, 5, 120, 245, 5, 5, 121, 245, 5, 5, 6, 247, 5, 5, 7, 247, 5, 5, 8, 247, 5, 5, 9, 247, 5, 5, 44, 248, 5, 5, 45, 248, 5, 5, 253, 248, 5, 5, 162, 249, 5, 5, 117, 158, 5, 5, 246, 161, 5, 5, 247, 161, 5, 5, 248, 161, 5, 5, 153, 165, 5, 5, 154, 165, 5, 5, 10, 171, 5, 5, 11, 171, 5, 5, 12, 171, 5, 5, 13, 171, 5, 5, 14, 171, 5, 5, 15, 171, 5, 5, 16, 171, 5, 5, 17, 171, 5, 5, 18, 171, 5, 5, 19, 171, 5, 5, 20, 171, 5, 5, 21, 171, 5, 5, 22, 171, 5, 5, 62, 177, 5, 5, 63, 177, 5, 5, 64, 177, 5, 5, 65, 177, 5, 5, 66, 177, 5, 5, 67, 177, 5, 5, 68, 177, 5, 5, 69, 177, 5, 5, 70, 177, 5, 5, 71, 177, 5, 5, 72, 177, 5, 5, 73, 177, 5, 5, 119, 184, 5, 5, 120, 184, 5, 5, 121, 184, 5, 5, 122, 184, 5, 5, 123, 184, 5, 5, 124, 184, 5, 5, 125, 184, 5, 5, 126, 184, 5, 5, 127, 184, 5, 5, 128, 184, 5, 5, 129, 184, 5, 5, 81, 192, 5, 5, 82, 192, 5, 5, 83, 192, 5, 5, 84, 192, 5, 5, 131, 200, 5, 5, 132, 200, 5, 5, 133, 200, 5, 5, 134, 200, 5, 5, 135, 200, 5, 5, 136, 200, 5, 5, 137, 200, 5, 5, 138, 200, 5, 5, 139, 200, 5, 5, 140, 200, 5, 5, 141, 200, 5, 5, 52, 208, 5, 5, 53, 208, 5, 5, 136, 215, 5, 5, 137, 215, 5, 5, 138, 215, 5, 5, 139, 215, 5, 5, 94, 222, 5, 5, 104, 228, 5, 5, 105, 228, 5, 5, 106, 228, 5, 5, 107, 228, 5, 5, 145, 233, 5, 5, 146, 233, 5, 5, 122, 245, 5, 5, 155, 165, 5, 5, 85, 192, 5, 5, 86, 192, 5, 5, 87, 192, 5, 5, 54, 208, 5, 5, 55, 208, 5, 5, 56, 208, 5, 5, 140, 215, 5, 5, 108, 228, 5, 5, 95, 222, 5, 5, 109, 228, 5, 5, 147, 233, 5, 5, 156, 165, 5, 5, 249, 161, 5, 5, 74, 177, 5, 5, 75, 177, 5, 5, 76, 177, 5, 5, 77, 177, 5, 5, 130, 184, 5, 5, 131, 184, 5, 5, 132, 184, 5, 5, 88, 192, 5, 5, 89, 192, 5, 5, 90, 192, 5, 5, 91, 192, 5, 5, 92, 192, 5, 5, 93, 192, 5, 5, 94, 192, 5, 5, 142, 200, 5, 5, 143, 200, 5, 5, 144, 200, 5, 5, 145, 200, 5, 5, 146, 200, 5, 5, 147, 200, 5, 5, 148, 200, 5, 5, 149, 200, 5, 5, 150, 200, 5, 5, 151, 200, 5, 5, 152, 200, 5, 5, 153, 200, 5, 5, 57, 208, 5, 5, 58, 208, 5, 5, 59, 208, 5, 5, 60, 208, 5, 5, 61, 208, 5, 5, 62, 208, 5, 5, 63, 208, 5, 5, 64, 208, 5, 5, 65, 208, 5, 5, 141, 215, 5, 5, 142, 215, 5, 5, 143, 215, 5, 5, 144, 215, 5, 5, 145, 215, 5, 5, 146, 215, 5, 5, 96, 222, 5, 5, 97, 222, 5, 5, 98, 222, 5, 5, 99, 222, 5, 5, 100, 222, 5, 5, 101, 222, 5, 5, 102, 222, 5, 5, 103, 222, 5, 5, 104, 222, 5, 5, 105, 222, 5, 5, 110, 228, 5, 5, 111, 228, 5, 5, 112, 228, 5, 5, 148, 233, 5, 5, 104, 237, 5, 5, 182, 240, 5, 5, 183, 240, 5, 5, 184, 240, 5, 5, 185, 240, 5, 5, 110, 243, 5, 5, 123, 245, 5, 5, 124, 245, 5, 5, 46, 248, 5, 5, 242, 249, 5, 5, 157, 165, 5, 5, 79, 177, 5, 5, 134, 184, 5, 5, 135, 184, 5, 5, 136, 184, 5, 5, 137, 184, 5, 5, 97, 192, 5, 5, 98, 192, 5, 5, 99, 192, 5, 5, 68, 208, 5, 5, 100, 192, 5, 5, 101, 192, 5, 5, 102, 192, 5, 5, 103, 192, 5, 5, 104, 192, 5, 5, 105, 192, 5, 5, 106, 192, 5, 5, 107, 192, 5, 5, 156, 200, 5, 5, 157, 200, 5, 5, 158, 200, 5, 5, 159, 200, 5, 5, 160, 200, 5, 5, 161, 200, 5, 5, 162, 200, 5, 5, 163, 200, 5, 5, 164, 200, 5, 5, 165, 200, 5, 5, 166, 200, 5, 5, 67, 208, 5, 5, 167, 200, 5, 5, 168, 200, 5, 5, 169, 200, 5, 5, 170, 200, 5, 5, 171, 200, 5, 5, 172, 200, 5, 5, 173, 200, 5, 5, 174, 200, 5, 5, 175, 200, 5, 5, 176, 200, 5, 5, 177, 200, 5, 5, 178, 200, 5, 5, 179, 200, 5, 5, 180, 200, 5, 5, 69, 208, 5, 5, 70, 208, 5, 5, 71, 208, 5, 5, 72, 208, 5, 5, 73, 208, 5, 5, 74, 208, 5, 5, 75, 208, 5, 5, 76, 208, 5, 5, 77, 208, 5, 5, 78, 208, 5, 5, 79, 208, 5, 5, 80, 208, 5, 5, 81, 208, 5, 5, 82, 208, 5, 5, 83, 208, 5, 5, 84, 208, 5, 5, 85, 208, 5, 5, 86, 208, 5, 5, 87, 208, 5, 5, 88, 208, 5, 5, 89, 208, 5, 5, 90, 208, 5, 5, 181, 200, 5, 5, 91, 208, 5, 5, 92, 208, 5, 5, 93, 208, 5, 5, 94, 208, 5, 5, 95, 208, 5, 5, 96, 208, 5, 5, 147, 215, 5, 5, 148, 215, 5, 5, 149, 215, 5, 5, 150, 215, 5, 5, 151, 215, 5, 5, 152, 215, 5, 5, 153, 215, 5, 5, 154, 215, 5, 5, 155, 215, 5, 5, 156, 215, 5, 5, 157, 215, 5, 5, 158, 215, 5, 5, 159, 215, 5, 5, 160, 215, 5, 5, 161, 215, 5, 5, 162, 215, 5, 5, 163, 215, 5, 5, 164, 215, 5, 5, 165, 215, 5, 5, 106, 222, 5, 5, 107, 222, 5, 5, 108, 222, 5, 5, 109, 222, 5, 5, 110, 222, 5, 5, 111, 222, 5, 5, 112, 222, 5, 5, 113, 222, 5, 5, 114, 222, 5, 5, 115, 222, 5, 5, 116, 222, 5, 5, 117, 222, 5, 5, 118, 222, 5, 5, 119, 222, 5, 5, 120, 222, 5, 5, 121, 222, 5, 5, 122, 222, 5, 5, 123, 222, 5, 5, 124, 222, 5, 5, 125, 222, 5, 5, 126, 222, 5, 5, 127, 222, 5, 5, 128, 222, 5, 5, 129, 222, 5, 5, 130, 222, 5, 5, 131, 222, 5, 5, 132, 222, 5, 5, 133, 222, 5, 5, 144, 222, 5, 5, 134, 222, 5, 5, 135, 222, 5, 5, 136, 222, 5, 5, 137, 222, 5, 5, 114, 228, 5, 5, 115, 228, 5, 5, 116, 228, 5, 5, 117, 228, 5, 5, 118, 228, 5, 5, 119, 228, 5, 5, 120, 228, 5, 5, 145, 222, 5, 5, 121, 228, 5, 5, 122, 228, 5, 5, 138, 222, 5, 5, 123, 228, 5, 5, 124, 228, 5, 5, 125, 228, 5, 5, 126, 228, 5, 5, 127, 228, 5, 5, 128, 228, 5, 5, 129, 228, 5, 5, 130, 228, 5, 5, 131, 228, 5, 5, 132, 228, 5, 5, 133, 228, 5, 5, 149, 233, 5, 5, 150, 233, 5, 5, 151, 233, 5, 5, 152, 233, 5, 5, 153, 233, 5, 5, 154, 233, 5, 5, 155, 233, 5, 5, 156, 233, 5, 5, 157, 233, 5, 5, 158, 233, 5, 5, 159, 233, 5, 5, 160, 233, 5, 5, 161, 233, 5, 5, 162, 233, 5, 5, 105, 237, 5, 5, 106, 237, 5, 5, 107, 237, 5, 5, 108, 237, 5, 5, 109, 237, 5, 5, 110, 237, 5, 5, 111, 237, 5, 5, 112, 237, 5, 5, 113, 237, 5, 5, 114, 237, 5, 5, 115, 237, 5, 5, 116, 237, 5, 5, 117, 237, 5, 5, 118, 237, 5, 5, 119, 237, 5, 5, 120, 237, 5, 5, 121, 237, 5, 5, 122, 237, 5, 5, 123, 237, 5, 5, 124, 237, 5, 5, 186, 240, 5, 5, 129, 237, 5, 5, 187, 240, 5, 5, 188, 240, 5, 5, 189, 240, 5, 5, 190, 240, 5, 5, 125, 237, 5, 5, 191, 240, 5, 5, 192, 240, 5, 5, 193, 240, 5, 5, 194, 240, 5, 5, 195, 240, 5, 5, 196, 240, 5, 5, 197, 240, 5, 5, 198, 240, 5, 5, 199, 240, 5, 5, 200, 240, 5, 5, 201, 240, 5, 5, 202, 240, 5, 5, 203, 240, 5, 5, 204, 240, 5, 5, 205, 240, 5, 5, 206, 240, 5, 5, 207, 240, 5, 5, 126, 237, 5, 5, 111, 243, 5, 5, 112, 243, 5, 5, 113, 243, 5, 5, 114, 243, 5, 5, 115, 243, 5, 5, 116, 243, 5, 5, 212, 240, 5, 5, 117, 243, 5, 5, 118, 243, 5, 5, 125, 245, 5, 5, 126, 245, 5, 5, 127, 245, 5, 5, 128, 245, 5, 5, 129, 245, 5, 5, 130, 245, 5, 5, 10, 247, 5, 5, 11, 247, 5, 5, 12, 247, 5, 5, 13, 247, 5, 5, 14, 247, 5, 5, 15, 247, 5, 5, 16, 247, 5, 5, 17, 247, 5, 5, 47, 248, 5, 5, 48, 248, 5, 5, 18, 247, 5, 5, 49, 248, 5, 5, 50, 248, 5, 5, 254, 248, 5, 5, 255, 248, 5, 5, 2, 249, 5, 5, 3, 249, 5, 5, 164, 249, 5, 5, 165, 249, 5, 5, 166, 249, 5, 5, 167, 249, 5, 5, 168, 249, 5, 5, 243, 249, 5, 5, 244, 249, 5, 5, 98, 250, 5, 5, 47, 250, 5, 5, 48, 250, 5, 5, 159, 165, 5, 5, 139, 184, 5, 5, 108, 192, 5, 5, 109, 192, 5, 5, 110, 192, 5, 5, 186, 200, 5, 5, 100, 208, 5, 5, 101, 208, 5, 5, 168, 215, 5, 5, 169, 215, 5, 5, 170, 215, 5, 5, 146, 222, 5, 5, 147, 222, 5, 5, 148, 222, 5, 5, 149, 222, 5, 5, 150, 222, 5, 5, 151, 222, 5, 5, 152, 222, 5, 5, 136, 228, 5, 5, 137, 228, 5, 5, 130, 237, 5, 5, 131, 237, 5, 5, 132, 237, 5, 5, 213, 240, 5, 5, 214, 240, 5, 5, 215, 240, 5, 5, 216, 240, 5, 5, 120, 243, 5, 5, 133, 245, 5, 5, 4, 249, 5, 5, 49, 250, 5, 5, 160, 165, 5, 5, 23, 171, 5, 5, 80, 177, 5, 5, 81, 177, 5, 5, 140, 184, 5, 5, 141, 184, 5, 5, 142, 184, 5, 5, 143, 184, 5, 5, 144, 184, 5, 5, 145, 184, 5, 5, 146, 184, 5, 5, 147, 184, 5, 5, 113, 192, 5, 5, 114, 192, 5, 5, 115, 192, 5, 5, 116, 192, 5, 5, 117, 192, 5, 5, 118, 192, 5, 5, 119, 192, 5, 5, 120, 192, 5, 5, 121, 192, 5, 5, 122, 192, 5, 5, 123, 192, 5, 5, 124, 192, 5, 5, 125, 192, 5, 5, 126, 192, 5, 5, 187, 200, 5, 5, 188, 200, 5, 5, 189, 200, 5, 5, 190, 200, 5, 5, 191, 200, 5, 5, 192, 200, 5, 5, 193, 200, 5, 5, 194, 200, 5, 5, 195, 200, 5, 5, 103, 208, 5, 5, 196, 200, 5, 5, 197, 200, 5, 5, 198, 200, 5, 5, 199, 200, 5, 5, 200, 200, 5, 5, 201, 200, 5, 5, 202, 200, 5, 5, 203, 200, 5, 5, 204, 200, 5, 5, 205, 200, 5, 5, 206, 200, 5, 5, 207, 200, 5, 5, 208, 200, 5, 5, 209, 200, 5, 5, 210, 200, 5, 5, 211, 200, 5, 5, 104, 208, 5, 5, 105, 208, 5, 5, 106, 208, 5, 5, 107, 208, 5, 5, 108, 208, 5, 5, 109, 208, 5, 5, 110, 208, 5, 5, 111, 208, 5, 5, 112, 208, 5, 5, 113, 208, 5, 5, 114, 208, 5, 5, 115, 208, 5, 5, 116, 208, 5, 5, 117, 208, 5, 5, 118, 208, 5, 5, 171, 215, 5, 5, 172, 215, 5, 5, 173, 215, 5, 5, 174, 215, 5, 5, 175, 215, 5, 5, 176, 215, 5, 5, 177, 215, 5, 5, 178, 215, 5, 5, 179, 215, 5, 5, 154, 222, 5, 5, 155, 222, 5, 5, 156, 222, 5, 5, 157, 222, 5, 5, 158, 222, 5, 5, 159, 222, 5, 5, 160, 222, 5, 5, 161, 222, 5, 5, 162, 222, 5, 5, 163, 222, 5, 5, 164, 222, 5, 5, 165, 222, 5, 5, 166, 222, 5, 5, 167, 222, 5, 5, 168, 222, 5, 5, 169, 222, 5, 5, 170, 222, 5, 5, 171, 222, 5, 5, 172, 222, 5, 5, 173, 222, 5, 5, 174, 222, 5, 5, 175, 222, 5, 5, 176, 222, 5, 5, 138, 228, 5, 5, 139, 228, 5, 5, 140, 228, 5, 5, 141, 228, 5, 5, 142, 228, 5, 5, 143, 228, 5, 5, 144, 228, 5, 5, 145, 228, 5, 5, 146, 228, 5, 5, 147, 228, 5, 5, 148, 228, 5, 5, 149, 228, 5, 5, 150, 228, 5, 5, 151, 228, 5, 5, 152, 228, 5, 5, 153, 228, 5, 5, 166, 233, 5, 5, 167, 233, 5, 5, 168, 233, 5, 5, 169, 233, 5, 5, 170, 233, 5, 5, 171, 233, 5, 5, 172, 233, 5, 5, 173, 233, 5, 5, 174, 233, 5, 5, 134, 237, 5, 5, 135, 237, 5, 5, 136, 237, 5, 5, 137, 237, 5, 5, 138, 237, 5, 5, 139, 237, 5, 5, 140, 237, 5, 5, 217, 240, 5, 5, 218, 240, 5, 5, 219, 240, 5, 5, 220, 240, 5, 5, 221, 240, 5, 5, 222, 240, 5, 5, 223, 240, 5, 5, 224, 240, 5, 5, 121, 243, 5, 5, 122, 243, 5, 5, 123, 243, 5, 5, 124, 243, 5, 5, 125, 243, 5, 5, 126, 243, 5, 5, 134, 245, 5, 5, 135, 245, 5, 5, 136, 245, 5, 5, 137, 245, 5, 5, 138, 245, 5, 5, 19, 247, 5, 5, 20, 247, 5, 5, 21, 247, 5, 5, 52, 248, 5, 5, 53, 248, 5, 5, 50, 250, 5, 5, 118, 158, 5, 5, 199, 159, 5, 5, 250, 161, 5, 5, 161, 165, 5, 5, 162, 165, 5, 5, 163, 165, 5, 5, 24, 171, 5, 5, 25, 171, 5, 5, 26, 171, 5, 5, 27, 171, 5, 5, 28, 171, 5, 5, 82, 177, 5, 5, 83, 177, 5, 5, 84, 177, 5, 5, 85, 177, 5, 5, 86, 177, 5, 5, 87, 177, 5, 5, 88, 177, 5, 5, 89, 177, 5, 5, 90, 177, 5, 5, 91, 177, 5, 5, 92, 177, 5, 5, 148, 184, 5, 5, 149, 184, 5, 5, 150, 184, 5, 5, 151, 184, 5, 5, 152, 184, 5, 5, 153, 184, 5, 5, 154, 184, 5, 5, 155, 184, 5, 5, 133, 192, 5, 5, 134, 192, 5, 5, 135, 192, 5, 5, 215, 200, 5, 5, 216, 200, 5, 5, 217, 200, 5, 5, 218, 200, 5, 5, 219, 200, 5, 5, 220, 200, 5, 5, 221, 200, 5, 5, 222, 200, 5, 5, 119, 208, 5, 5, 120, 208, 5, 5, 121, 208, 5, 5, 122, 208, 5, 5, 123, 208, 5, 5, 124, 208, 5, 5, 182, 215, 5, 5, 183, 215, 5, 5, 184, 215, 5, 5, 180, 222, 5, 5, 155, 228, 5, 5, 156, 228, 5, 5, 164, 165, 5, 5, 223, 200, 5, 5, 224, 200, 5, 5, 125, 208, 5, 5, 126, 208, 5, 5, 127, 208, 5, 5, 185, 215, 5, 5, 186, 215, 5, 5, 187, 215, 5, 5, 181, 222, 5, 5, 157, 228, 5, 5, 158, 228, 5, 5, 159, 228, 5, 5, 160, 228, 5, 5, 161, 228, 5, 5, 162, 228, 5, 5, 177, 233, 5, 5, 143, 237, 5, 5, 225, 240, 5, 5, 127, 243, 5, 5, 139, 245, 5, 5, 165, 165, 5, 5, 156, 184, 5, 5, 128, 208, 5, 5, 182, 222, 5, 5, 226, 240, 5, 5, 166, 165, 5, 5, 120, 158, 5, 5, 200, 159, 5, 5, 251, 161, 5, 5, 252, 161, 5, 5, 253, 161, 5, 5, 254, 161, 5, 5, 255, 161, 5, 5, 2, 162, 5, 5, 168, 165, 5, 5, 169, 165, 5, 5, 170, 165, 5, 5, 171, 165, 5, 5, 172, 165, 5, 5, 173, 165, 5, 5, 174, 165, 5, 5, 175, 165, 5, 5, 176, 165, 5, 5, 177, 165, 5, 5, 178, 165, 5, 5, 179, 165, 5, 5, 29, 171, 5, 5, 30, 171, 5, 5, 31, 171, 5, 5, 32, 171, 5, 5, 33, 171, 5, 5, 34, 171, 5, 5, 35, 171, 5, 5, 36, 171, 5, 5, 37, 171, 5, 5, 38, 171, 5, 5, 39, 171, 5, 5, 40, 171, 5, 5, 41, 171, 5, 5, 42, 171, 5, 5, 43, 171, 5, 5, 44, 171, 5, 5, 45, 171, 5, 5, 46, 171, 5, 5, 47, 171, 5, 5, 48, 171, 5, 5, 49, 171, 5, 5, 50, 171, 5, 5, 93, 177, 5, 5, 94, 177, 5, 5, 95, 177, 5, 5, 96, 177, 5, 5, 97, 177, 5, 5, 98, 177, 5, 5, 99, 177, 5, 5, 100, 177, 5, 5, 101, 177, 5, 5, 102, 177, 5, 5, 103, 177, 5, 5, 104, 177, 5, 5, 51, 171, 5, 5, 105, 177, 5, 5, 106, 177, 5, 5, 107, 177, 5, 5, 108, 177, 5, 5, 109, 177, 5, 5, 110, 177, 5, 5, 111, 177, 5, 5, 157, 184, 5, 5, 158, 184, 5, 5, 159, 184, 5, 5, 160, 184, 5, 5, 161, 184, 5, 5, 162, 184, 5, 5, 163, 184, 5, 5, 164, 184, 5, 5, 165, 184, 5, 5, 166, 184, 5, 5, 167, 184, 5, 5, 168, 184, 5, 5, 169, 184, 5, 5, 170, 184, 5, 5, 171, 184, 5, 5, 172, 184, 5, 5, 173, 184, 5, 5, 174, 184, 5, 5, 175, 184, 5, 5, 176, 184, 5, 5, 177, 184, 5, 5, 178, 184, 5, 5, 179, 184, 5, 5, 136, 192, 5, 5, 137, 192, 5, 5, 138, 192, 5, 5, 139, 192, 5, 5, 140, 192, 5, 5, 141, 192, 5, 5, 142, 192, 5, 5, 143, 192, 5, 5, 144, 192, 5, 5, 145, 192, 5, 5, 146, 192, 5, 5, 147, 192, 5, 5, 148, 192, 5, 5, 149, 192, 5, 5, 150, 192, 5, 5, 151, 192, 5, 5, 152, 192, 5, 5, 153, 192, 5, 5, 154, 192, 5, 5, 155, 192, 5, 5, 156, 192, 5, 5, 157, 192, 5, 5, 158, 192, 5, 5, 159, 192, 5, 5, 160, 192, 5, 5, 161, 192, 5, 5, 162, 192, 5, 5, 163, 192, 5, 5, 164, 192, 5, 5, 226, 200, 5, 5, 227, 200, 5, 5, 228, 200, 5, 5, 229, 200, 5, 5, 230, 200, 5, 5, 231, 200, 5, 5, 232, 200, 5, 5, 233, 200, 5, 5, 234, 200, 5, 5, 235, 200, 5, 5, 236, 200, 5, 5, 237, 200, 5, 5, 238, 200, 5, 5, 239, 200, 5, 5, 240, 200, 5, 5, 241, 200, 5, 5, 242, 200, 5, 5, 243, 200, 5, 5, 244, 200, 5, 5, 245, 200, 5, 5, 129, 208, 5, 5, 130, 208, 5, 5, 131, 208, 5, 5, 132, 208, 5, 5, 133, 208, 5, 5, 134, 208, 5, 5, 135, 208, 5, 5, 136, 208, 5, 5, 137, 208, 5, 5, 138, 208, 5, 5, 139, 208, 5, 5, 140, 208, 5, 5, 141, 208, 5, 5, 142, 208, 5, 5, 143, 208, 5, 5, 144, 208, 5, 5, 145, 208, 5, 5, 146, 208, 5, 5, 147, 208, 5, 5, 148, 208, 5, 5, 149, 208, 5, 5, 150, 208, 5, 5, 151, 208, 5, 5, 152, 208, 5, 5, 153, 208, 5, 5, 154, 208, 5, 5, 155, 208, 5, 5, 156, 208, 5, 5, 188, 215, 5, 5, 189, 215, 5, 5, 190, 215, 5, 5, 191, 215, 5, 5, 192, 215, 5, 5, 193, 215, 5, 5, 194, 215, 5, 5, 195, 215, 5, 5, 196, 215, 5, 5, 197, 215, 5, 5, 198, 215, 5, 5, 199, 215, 5, 5, 200, 215, 5, 5, 201, 215, 5, 5, 183, 222, 5, 5, 184, 222, 5, 5, 185, 222, 5, 5, 186, 222, 5, 5, 187, 222, 5, 5, 188, 222, 5, 5, 189, 222, 5, 5, 190, 222, 5, 5, 191, 222, 5, 5, 192, 222, 5, 5, 193, 222, 5, 5, 194, 222, 5, 5, 163, 228, 5, 5, 195, 222, 5, 5, 164, 228, 5, 5, 165, 228, 5, 5, 166, 228, 5, 5, 196, 222, 5, 5, 167, 228, 5, 5, 168, 228, 5, 5, 169, 228, 5, 5, 170, 228, 5, 5, 171, 228, 5, 5, 178, 233, 5, 5, 179, 233, 5, 5, 180, 233, 5, 5, 181, 233, 5, 5, 182, 233, 5, 5, 183, 233, 5, 5, 184, 233, 5, 5, 185, 233, 5, 5, 186, 233, 5, 5, 172, 228, 5, 5, 144, 237, 5, 5, 145, 237, 5, 5, 187, 233, 5, 5, 227, 240, 5, 5, 228, 240, 5, 5, 229, 240, 5, 5, 128, 243, 5, 5, 140, 245, 5, 5, 54, 248, 5, 5, 55, 248, 5, 5, 183, 165, 5, 5, 201, 159, 5, 5, 202, 159, 5, 5, 4, 162, 5, 5, 180, 184, 5, 5, 5, 162, 5, 5, 6, 162, 5, 5, 7, 162, 5, 5, 8, 162, 5, 5, 9, 162, 5, 5, 10, 162, 5, 5, 11, 162, 5, 5, 12, 162, 5, 5, 184, 165, 5, 5, 185, 165, 5, 5, 186, 165, 5, 5, 187, 165, 5, 5, 188, 165, 5, 5, 189, 165, 5, 5, 190, 165, 5, 5, 191, 165, 5, 5, 192, 165, 5, 5, 193, 165, 5, 5, 194, 165, 5, 5, 195, 165, 5, 5, 196, 165, 5, 5, 167, 192, 5, 5, 197, 165, 5, 5, 53, 171, 5, 5, 54, 171, 5, 5, 55, 171, 5, 5, 56, 171, 5, 5, 57, 171, 5, 5, 58, 171, 5, 5, 59, 171, 5, 5, 60, 171, 5, 5, 61, 171, 5, 5, 62, 171, 5, 5, 63, 171, 5, 5, 64, 171, 5, 5, 65, 171, 5, 5, 66, 171, 5, 5, 67, 171, 5, 5, 112, 177, 5, 5, 113, 177, 5, 5, 114, 177, 5, 5, 115, 177, 5, 5, 116, 177, 5, 5, 117, 177, 5, 5, 118, 177, 5, 5, 119, 177, 5, 5, 120, 177, 5, 5, 121, 177, 5, 5, 122, 177, 5, 5, 123, 177, 5, 5, 124, 177, 5, 5, 125, 177, 5, 5, 126, 177, 5, 5, 127, 177, 5, 5, 158, 208, 5, 5, 128, 177, 5, 5, 129, 177, 5, 5, 130, 177, 5, 5, 131, 177, 5, 5, 132, 177, 5, 5, 202, 215, 5, 5, 133, 177, 5, 5, 168, 192, 5, 5, 134, 177, 5, 5, 181, 184, 5, 5, 182, 184, 5, 5, 183, 184, 5, 5, 184, 184, 5, 5, 185, 184, 5, 5, 186, 184, 5, 5, 187, 184, 5, 5, 188, 184, 5, 5, 189, 184, 5, 5, 190, 184, 5, 5, 191, 184, 5, 5, 192, 184, 5, 5, 193, 184, 5, 5, 194, 184, 5, 5, 195, 184, 5, 5, 196, 184, 5, 5, 197, 184, 5, 5, 198, 184, 5, 5, 169, 192, 5, 5, 170, 192, 5, 5, 171, 192, 5, 5, 172, 192, 5, 5, 173, 192, 5, 5, 174, 192, 5, 5, 175, 192, 5, 5, 176, 192, 5, 5, 177, 192, 5, 5, 137, 177, 5, 5, 178, 192, 5, 5, 179, 192, 5, 5, 180, 192, 5, 5, 248, 200, 5, 5, 197, 222, 5, 5, 181, 192, 5, 5, 182, 192, 5, 5, 249, 200, 5, 5, 177, 228, 5, 5, 250, 200, 5, 5, 251, 200, 5, 5, 183, 192, 5, 5, 252, 200, 5, 5, 253, 200, 5, 5, 254, 200, 5, 5, 255, 200, 5, 5, 2, 201, 5, 5, 3, 201, 5, 5, 4, 201, 5, 5, 5, 201, 5, 5, 6, 201, 5, 5, 7, 201, 5, 5, 8, 201, 5, 5, 9, 201, 5, 5, 10, 201, 5, 5, 159, 208, 5, 5, 160, 208, 5, 5, 161, 208, 5, 5, 162, 208, 5, 5, 163, 208, 5, 5, 164, 208, 5, 5, 165, 208, 5, 5, 166, 208, 5, 5, 167, 208, 5, 5, 168, 208, 5, 5, 169, 208, 5, 5, 170, 208, 5, 5, 171, 208, 5, 5, 203, 215, 5, 5, 204, 215, 5, 5, 205, 215, 5, 5, 206, 215, 5, 5, 207, 215, 5, 5, 208, 215, 5, 5, 209, 215, 5, 5, 210, 215, 5, 5, 211, 215, 5, 5, 212, 215, 5, 5, 213, 215, 5, 5, 214, 215, 5, 5, 215, 215, 5, 5, 216, 215, 5, 5, 198, 222, 5, 5, 199, 222, 5, 5, 147, 237, 5, 5, 200, 222, 5, 5, 201, 222, 5, 5, 202, 222, 5, 5, 12, 201, 5, 5, 203, 222, 5, 5, 204, 222, 5, 5, 205, 222, 5, 5, 206, 222, 5, 5, 207, 222, 5, 5, 208, 222, 5, 5, 178, 228, 5, 5, 179, 228, 5, 5, 180, 228, 5, 5, 181, 228, 5, 5, 182, 228, 5, 5, 190, 233, 5, 5, 191, 233, 5, 5, 148, 237, 5, 5, 149, 237, 5, 5, 150, 237, 5, 5, 151, 237, 5, 5, 152, 237, 5, 5, 231, 240, 5, 5, 232, 240, 5, 5, 129, 243, 5, 5, 233, 240, 5, 5, 130, 243, 5, 5, 141, 245, 5, 5, 142, 245, 5, 5, 143, 245, 5, 5, 22, 247, 5, 5, 23, 247, 5, 5, 200, 165, 5, 5, 138, 177, 5, 5, 139, 177, 5, 5, 199, 184, 5, 5, 200, 184, 5, 5, 201, 184, 5, 5, 202, 184, 5, 5, 203, 184, 5, 5, 204, 184, 5, 5, 205, 184, 5, 5, 186, 192, 5, 5, 187, 192, 5, 5, 188, 192, 5, 5, 189, 192, 5, 5, 190, 192, 5, 5, 191, 192, 5, 5, 192, 192, 5, 5, 193, 192, 5, 5, 194, 192, 5, 5, 195, 192, 5, 5, 196, 192, 5, 5, 197, 192, 5, 5, 13, 201, 5, 5, 14, 201, 5, 5, 15, 201, 5, 5, 16, 201, 5, 5, 17, 201, 5, 5, 18, 201, 5, 5, 19, 201, 5, 5, 172, 208, 5, 5, 173, 208, 5, 5, 174, 208, 5, 5, 175, 208, 5, 5, 176, 208, 5, 5, 177, 208, 5, 5, 178, 208, 5, 5, 179, 208, 5, 5, 180, 208, 5, 5, 181, 208, 5, 5, 182, 208, 5, 5, 183, 208, 5, 5, 217, 215, 5, 5, 218, 215, 5, 5, 219, 215, 5, 5, 220, 215, 5, 5, 221, 215, 5, 5, 222, 215, 5, 5, 223, 215, 5, 5, 224, 215, 5, 5, 225, 215, 5, 5, 226, 215, 5, 5, 227, 215, 5, 5, 228, 215, 5, 5, 229, 215, 5, 5, 230, 215, 5, 5, 210, 222, 5, 5, 211, 222, 5, 5, 212, 222, 5, 5, 213, 222, 5, 5, 214, 222, 5, 5, 215, 222, 5, 5, 216, 222, 5, 5, 217, 222, 5, 5, 218, 222, 5, 5, 219, 222, 5, 5, 220, 222, 5, 5, 221, 222, 5, 5, 222, 222, 5, 5, 183, 228, 5, 5, 184, 228, 5, 5, 185, 228, 5, 5, 186, 228, 5, 5, 187, 228, 5, 5, 188, 228, 5, 5, 189, 228, 5, 5, 190, 228, 5, 5, 191, 228, 5, 5, 192, 228, 5, 5, 193, 228, 5, 5, 192, 233, 5, 5, 193, 233, 5, 5, 194, 233, 5, 5, 195, 233, 5, 5, 196, 233, 5, 5, 197, 233, 5, 5, 198, 233, 5, 5, 199, 233, 5, 5, 200, 233, 5, 5, 201, 233, 5, 5, 202, 233, 5, 5, 203, 233, 5, 5, 204, 233, 5, 5, 153, 237, 5, 5, 154, 237, 5, 5, 155, 237, 5, 5, 156, 237, 5, 5, 157, 237, 5, 5, 158, 237, 5, 5, 159, 237, 5, 5, 160, 237, 5, 5, 234, 240, 5, 5, 235, 240, 5, 5, 236, 240, 5, 5, 237, 240, 5, 5, 238, 240, 5, 5, 131, 243, 5, 5, 132, 243, 5, 5, 133, 243, 5, 5, 134, 243, 5, 5, 135, 243, 5, 5, 136, 243, 5, 5, 137, 243, 5, 5, 144, 245, 5, 5, 145, 245, 5, 5, 146, 245, 5, 5, 56, 248, 5, 5, 5, 249, 5, 5, 6, 249, 5, 5, 7, 249, 5, 5, 8, 249, 5, 5, 169, 249, 5, 5, 9, 249, 5, 5, 245, 249, 5, 5, 246, 249, 5, 5, 51, 250, 5, 5, 201, 165, 5, 5, 70, 171, 5, 5, 200, 192, 5, 5, 21, 201, 5, 5, 22, 201, 5, 5, 138, 243, 5, 5, 202, 165, 5, 5, 140, 177, 5, 5, 201, 192, 5, 5, 23, 201, 5, 5, 161, 237, 5, 5, 71, 171, 5, 5, 72, 171, 5, 5, 141, 177, 5, 5, 142, 177, 5, 5, 206, 184, 5, 5, 207, 184, 5, 5, 208, 184, 5, 5, 209, 184, 5, 5, 210, 184, 5, 5, 211, 184, 5, 5, 212, 184, 5, 5, 213, 184, 5, 5, 214, 184, 5, 5, 215, 184, 5, 5, 216, 184, 5, 5, 217, 184, 5, 5, 218, 184, 5, 5, 219, 184, 5, 5, 202, 192, 5, 5, 203, 192, 5, 5, 204, 192, 5, 5, 205, 192, 5, 5, 206, 192, 5, 5, 207, 192, 5, 5, 208, 192, 5, 5, 209, 192, 5, 5, 210, 192, 5, 5, 211, 192, 5, 5, 212, 192, 5, 5, 213, 192, 5, 5, 214, 192, 5, 5, 215, 192, 5, 5, 216, 192, 5, 5, 217, 192, 5, 5, 218, 192, 5, 5, 219, 192, 5, 5, 220, 192, 5, 5, 221, 192, 5, 5, 222, 192, 5, 5, 223, 192, 5, 5, 224, 192, 5, 5, 225, 192, 5, 5, 226, 192, 5, 5, 227, 192, 5, 5, 24, 201, 5, 5, 25, 201, 5, 5, 26, 201, 5, 5, 27, 201, 5, 5, 28, 201, 5, 5, 29, 201, 5, 5, 30, 201, 5, 5, 31, 201, 5, 5, 32, 201, 5, 5, 33, 201, 5, 5, 34, 201, 5, 5, 35, 201, 5, 5, 36, 201, 5, 5, 37, 201, 5, 5, 38, 201, 5, 5, 39, 201, 5, 5, 40, 201, 5, 5, 41, 201, 5, 5, 42, 201, 5, 5, 43, 201, 5, 5, 44, 201, 5, 5, 45, 201, 5, 5, 46, 201, 5, 5, 47, 201, 5, 5, 48, 201, 5, 5, 49, 201, 5, 5, 50, 201, 5, 5, 51, 201, 5, 5, 52, 201, 5, 5, 53, 201, 5, 5, 54, 201, 5, 5, 55, 201, 5, 5, 56, 201, 5, 5, 57, 201, 5, 5, 58, 201, 5, 5, 59, 201, 5, 5, 60, 201, 5, 5, 61, 201, 5, 5, 62, 201, 5, 5, 63, 201, 5, 5, 64, 201, 5, 5, 65, 201, 5, 5, 66, 201, 5, 5, 67, 201, 5, 5, 68, 201, 5, 5, 69, 201, 5, 5, 70, 201, 5, 5, 71, 201, 5, 5, 232, 215, 5, 5, 185, 208, 5, 5, 186, 208, 5, 5, 187, 208, 5, 5, 188, 208, 5, 5, 189, 208, 5, 5, 190, 208, 5, 5, 191, 208, 5, 5, 192, 208, 5, 5, 193, 208, 5, 5, 194, 208, 5, 5, 195, 208, 5, 5, 196, 208, 5, 5, 197, 208, 5, 5, 198, 208, 5, 5, 199, 208, 5, 5, 200, 208, 5, 5, 201, 208, 5, 5, 202, 208, 5, 5, 203, 208, 5, 5, 204, 208, 5, 5, 205, 208, 5, 5, 206, 208, 5, 5, 207, 208, 5, 5, 208, 208, 5, 5, 209, 208, 5, 5, 210, 208, 5, 5, 211, 208, 5, 5, 212, 208, 5, 5, 213, 208, 5, 5, 214, 208, 5, 5, 215, 208, 5, 5, 216, 208, 5, 5, 217, 208, 5, 5, 218, 208, 5, 5, 219, 208, 5, 5, 220, 208, 5, 5, 221, 208, 5, 5, 222, 208, 5, 5, 223, 208, 5, 5, 224, 208, 5, 5, 225, 208, 5, 5, 226, 208, 5, 5, 227, 208, 5, 5, 228, 208, 5, 5, 229, 208, 5, 5, 230, 208, 5, 5, 231, 208, 5, 5, 232, 208, 5, 5, 233, 208, 5, 5, 234, 208, 5, 5, 235, 208, 5, 5, 236, 208, 5, 5, 237, 208, 5, 5, 238, 208, 5, 5, 239, 208, 5, 5, 240, 208, 5, 5, 241, 208, 5, 5, 242, 208, 5, 5, 243, 208, 5, 5, 244, 208, 5, 5, 245, 208, 5, 5, 246, 208, 5, 5, 247, 208, 5, 5, 248, 208, 5, 5, 249, 208, 5, 5, 250, 208, 5, 5, 251, 208, 5, 5, 252, 208, 5, 5, 253, 208, 5, 5, 254, 208, 5, 5, 255, 208, 5, 5, 233, 215, 5, 5, 234, 215, 5, 5, 235, 215, 5, 5, 236, 215, 5, 5, 237, 215, 5, 5, 238, 215, 5, 5, 239, 215, 5, 5, 47, 216, 5, 5, 240, 215, 5, 5, 241, 215, 5, 5, 242, 215, 5, 5, 243, 215, 5, 5, 244, 215, 5, 5, 245, 215, 5, 5, 246, 215, 5, 5, 247, 215, 5, 5, 248, 215, 5, 5, 249, 215, 5, 5, 250, 215, 5, 5, 251, 215, 5, 5, 252, 215, 5, 5, 253, 215, 5, 5, 254, 215, 5, 5, 255, 215, 5, 5, 2, 216, 5, 5, 3, 216, 5, 5, 2, 209, 5, 5, 4, 216, 5, 5, 5, 216, 5, 5, 6, 216, 5, 5, 7, 216, 5, 5, 8, 216, 5, 5, 9, 216, 5, 5, 10, 216, 5, 5, 11, 216, 5, 5, 12, 216, 5, 5, 13, 216, 5, 5, 14, 216, 5, 5, 15, 216, 5, 5, 16, 216, 5, 5, 17, 216, 5, 5, 18, 216, 5, 5, 19, 216, 5, 5, 20, 216, 5, 5, 21, 216, 5, 5, 22, 216, 5, 5, 23, 216, 5, 5, 24, 216, 5, 5, 25, 216, 5, 5, 26, 216, 5, 5, 27, 216, 5, 5, 28, 216, 5, 5, 29, 216, 5, 5, 30, 216, 5, 5, 31, 216, 5, 5, 32, 216, 5, 5, 33, 216, 5, 5, 34, 216, 5, 5, 35, 216, 5, 5, 36, 216, 5, 5, 37, 216, 5, 5, 223, 222, 5, 5, 224, 222, 5, 5, 225, 222, 5, 5, 226, 222, 5, 5, 227, 222, 5, 5, 228, 222, 5, 5, 229, 222, 5, 5, 230, 222, 5, 5, 231, 222, 5, 5, 232, 222, 5, 5, 233, 222, 5, 5, 234, 222, 5, 5, 235, 222, 5, 5, 236, 222, 5, 5, 237, 222, 5, 5, 238, 222, 5, 5, 239, 222, 5, 5, 240, 222, 5, 5, 241, 222, 5, 5, 242, 222, 5, 5, 243, 222, 5, 5, 244, 222, 5, 5, 245, 222, 5, 5, 246, 222, 5, 5, 247, 222, 5, 5, 196, 228, 5, 5, 248, 222, 5, 5, 249, 222, 5, 5, 250, 222, 5, 5, 251, 222, 5, 5, 252, 222, 5, 5, 253, 222, 5, 5, 254, 222, 5, 5, 255, 222, 5, 5, 2, 223, 5, 5, 3, 223, 5, 5, 4, 223, 5, 5, 5, 223, 5, 5, 6, 223, 5, 5, 7, 223, 5, 5, 8, 223, 5, 5, 9, 223, 5, 5, 10, 223, 5, 5, 11, 223, 5, 5, 12, 223, 5, 5, 13, 223, 5, 5, 14, 223, 5, 5, 15, 223, 5, 5, 16, 223, 5, 5, 17, 223, 5, 5, 18, 223, 5, 5, 19, 223, 5, 5, 20, 223, 5, 5, 21, 223, 5, 5, 22, 223, 5, 5, 23, 223, 5, 5, 24, 223, 5, 5, 25, 223, 5, 5, 26, 223, 5, 5, 27, 223, 5, 5, 46, 216, 5, 5, 28, 223, 5, 5, 29, 223, 5, 5, 30, 223, 5, 5, 31, 223, 5, 5, 32, 223, 5, 5, 33, 223, 5, 5, 34, 223, 5, 5, 35, 223, 5, 5, 197, 228, 5, 5, 198, 228, 5, 5, 199, 228, 5, 5, 200, 228, 5, 5, 201, 228, 5, 5, 202, 228, 5, 5, 203, 228, 5, 5, 204, 228, 5, 5, 205, 228, 5, 5, 206, 228, 5, 5, 207, 228, 5, 5, 208, 228, 5, 5, 209, 228, 5, 5, 210, 228, 5, 5, 211, 228, 5, 5, 212, 228, 5, 5, 213, 228, 5, 5, 214, 228, 5, 5, 215, 228, 5, 5, 216, 228, 5, 5, 217, 228, 5, 5, 218, 228, 5, 5, 219, 228, 5, 5, 220, 228, 5, 5, 221, 228, 5, 5, 222, 228, 5, 5, 223, 228, 5, 5, 224, 228, 5, 5, 225, 228, 5, 5, 226, 228, 5, 5, 227, 228, 5, 5, 228, 228, 5, 5, 229, 228, 5, 5, 230, 228, 5, 5, 231, 228, 5, 5, 232, 228, 5, 5, 233, 228, 5, 5, 234, 228, 5, 5, 235, 228, 5, 5, 236, 228, 5, 5, 237, 228, 5, 5, 238, 228, 5, 5, 239, 228, 5, 5, 240, 228, 5, 5, 241, 228, 5, 5, 242, 228, 5, 5, 243, 228, 5, 5, 244, 228, 5, 5, 245, 228, 5, 5, 206, 233, 5, 5, 246, 228, 5, 5, 247, 228, 5, 5, 248, 228, 5, 5, 249, 228, 5, 5, 250, 228, 5, 5, 251, 228, 5, 5, 252, 228, 5, 5, 253, 228, 5, 5, 254, 228, 5, 5, 255, 228, 5, 5, 2, 229, 5, 5, 3, 229, 5, 5, 4, 229, 5, 5, 5, 229, 5, 5, 6, 229, 5, 5, 7, 229, 5, 5, 8, 229, 5, 5, 9, 229, 5, 5, 10, 229, 5, 5, 11, 229, 5, 5, 12, 229, 5, 5, 13, 229, 5, 5, 14, 229, 5, 5, 15, 229, 5, 5, 16, 229, 5, 5, 17, 229, 5, 5, 18, 229, 5, 5, 19, 229, 5, 5, 20, 229, 5, 5, 21, 229, 5, 5, 207, 233, 5, 5, 22, 229, 5, 5, 208, 233, 5, 5, 209, 233, 5, 5, 210, 233, 5, 5, 211, 233, 5, 5, 212, 233, 5, 5, 213, 233, 5, 5, 214, 233, 5, 5, 215, 233, 5, 5, 216, 233, 5, 5, 217, 233, 5, 5, 218, 233, 5, 5, 219, 233, 5, 5, 220, 233, 5, 5, 221, 233, 5, 5, 222, 233, 5, 5, 223, 233, 5, 5, 224, 233, 5, 5, 225, 233, 5, 5, 226, 233, 5, 5, 227, 233, 5, 5, 228, 233, 5, 5, 229, 233, 5, 5, 230, 233, 5, 5, 231, 233, 5, 5, 232, 233, 5, 5, 233, 233, 5, 5, 234, 233, 5, 5, 235, 233, 5, 5, 236, 233, 5, 5, 237, 233, 5, 5, 238, 233, 5, 5, 239, 233, 5, 5, 240, 233, 5, 5, 241, 233, 5, 5, 242, 233, 5, 5, 243, 233, 5, 5, 244, 233, 5, 5, 245, 233, 5, 5, 246, 233, 5, 5, 247, 233, 5, 5, 248, 233, 5, 5, 249, 233, 5, 5, 250, 233, 5, 5, 251, 233, 5, 5, 252, 233, 5, 5, 253, 233, 5, 5, 254, 233, 5, 5, 255, 233, 5, 5, 2, 234, 5, 5, 36, 229, 5, 5, 3, 234, 5, 5, 4, 234, 5, 5, 5, 234, 5, 5, 6, 234, 5, 5, 7, 234, 5, 5, 8, 234, 5, 5, 9, 234, 5, 5, 10, 234, 5, 5, 11, 234, 5, 5, 12, 234, 5, 5, 13, 234, 5, 5, 14, 234, 5, 5, 15, 234, 5, 5, 162, 237, 5, 5, 163, 237, 5, 5, 164, 237, 5, 5, 165, 237, 5, 5, 166, 237, 5, 5, 167, 237, 5, 5, 168, 237, 5, 5, 169, 237, 5, 5, 170, 237, 5, 5, 171, 237, 5, 5, 172, 237, 5, 5, 173, 237, 5, 5, 174, 237, 5, 5, 175, 237, 5, 5, 176, 237, 5, 5, 177, 237, 5, 5, 178, 237, 5, 5, 179, 237, 5, 5, 180, 237, 5, 5, 181, 237, 5, 5, 182, 237, 5, 5, 183, 237, 5, 5, 184, 237, 5, 5, 185, 237, 5, 5, 186, 237, 5, 5, 32, 234, 5, 5, 187, 237, 5, 5, 188, 237, 5, 5, 189, 237, 5, 5, 190, 237, 5, 5, 191, 237, 5, 5, 192, 237, 5, 5, 193, 237, 5, 5, 240, 240, 5, 5, 194, 237, 5, 5, 195, 237, 5, 5, 196, 237, 5, 5, 197, 237, 5, 5, 198, 237, 5, 5, 33, 234, 5, 5, 199, 237, 5, 5, 200, 237, 5, 5, 201, 237, 5, 5, 202, 237, 5, 5, 203, 237, 5, 5, 204, 237, 5, 5, 205, 237, 5, 5, 206, 237, 5, 5, 207, 237, 5, 5, 208, 237, 5, 5, 209, 237, 5, 5, 210, 237, 5, 5, 211, 237, 5, 5, 212, 237, 5, 5, 213, 237, 5, 5, 214, 237, 5, 5, 241, 240, 5, 5, 242, 240, 5, 5, 243, 240, 5, 5, 244, 240, 5, 5, 245, 240, 5, 5, 246, 240, 5, 5, 247, 240, 5, 5, 248, 240, 5, 5, 249, 240, 5, 5, 250, 240, 5, 5, 251, 240, 5, 5, 252, 240, 5, 5, 253, 240, 5, 5, 254, 240, 5, 5, 255, 240, 5, 5, 2, 241, 5, 5, 3, 241, 5, 5, 4, 241, 5, 5, 5, 241, 5, 5, 6, 241, 5, 5, 7, 241, 5, 5, 8, 241, 5, 5, 9, 241, 5, 5, 10, 241, 5, 5, 11, 241, 5, 5, 12, 241, 5, 5, 13, 241, 5, 5, 14, 241, 5, 5, 15, 241, 5, 5, 16, 241, 5, 5, 17, 241, 5, 5, 18, 241, 5, 5, 19, 241, 5, 5, 20, 241, 5, 5, 21, 241, 5, 5, 22, 241, 5, 5, 23, 241, 5, 5, 24, 241, 5, 5, 25, 241, 5, 5, 26, 241, 5, 5, 27, 241, 5, 5, 28, 241, 5, 5, 29, 241, 5, 5, 30, 241, 5, 5, 31, 241, 5, 5, 32, 241, 5, 5, 33, 241, 5, 5, 34, 241, 5, 5, 35, 241, 5, 5, 36, 241, 5, 5, 37, 241, 5, 5, 139, 243, 5, 5, 147, 245, 5, 5, 140, 243, 5, 5, 141, 243, 5, 5, 142, 243, 5, 5, 143, 243, 5, 5, 38, 241, 5, 5, 144, 243, 5, 5, 145, 243, 5, 5, 146, 243, 5, 5, 147, 243, 5, 5, 148, 243, 5, 5, 149, 243, 5, 5, 150, 243, 5, 5, 151, 243, 5, 5, 152, 243, 5, 5, 153, 243, 5, 5, 154, 243, 5, 5, 155, 243, 5, 5, 156, 243, 5, 5, 157, 243, 5, 5, 158, 243, 5, 5, 159, 243, 5, 5, 160, 243, 5, 5, 161, 243, 5, 5, 162, 243, 5, 5, 163, 243, 5, 5, 164, 243, 5, 5, 165, 243, 5, 5, 166, 243, 5, 5, 167, 243, 5, 5, 168, 243, 5, 5, 169, 243, 5, 5, 170, 243, 5, 5, 171, 243, 5, 5, 172, 243, 5, 5, 173, 243, 5, 5, 174, 243, 5, 5, 175, 243, 5, 5, 176, 243, 5, 5, 177, 243, 5, 5, 178, 243, 5, 5, 179, 243, 5, 5, 180, 243, 5, 5, 181, 243, 5, 5, 182, 243, 5, 5, 183, 243, 5, 5, 184, 243, 5, 5, 185, 243, 5, 5, 186, 243, 5, 5, 187, 243, 5, 5, 188, 243, 5, 5, 189, 243, 5, 5, 190, 243, 5, 5, 148, 245, 5, 5, 149, 245, 5, 5, 150, 245, 5, 5, 151, 245, 5, 5, 152, 245, 5, 5, 153, 245, 5, 5, 202, 243, 5, 5, 154, 245, 5, 5, 155, 245, 5, 5, 156, 245, 5, 5, 157, 245, 5, 5, 158, 245, 5, 5, 159, 245, 5, 5, 160, 245, 5, 5, 161, 245, 5, 5, 162, 245, 5, 5, 163, 245, 5, 5, 164, 245, 5, 5, 165, 245, 5, 5, 203, 243, 5, 5, 166, 245, 5, 5, 167, 245, 5, 5, 168, 245, 5, 5, 169, 245, 5, 5, 170, 245, 5, 5, 24, 247, 5, 5, 25, 247, 5, 5, 26, 247, 5, 5, 27, 247, 5, 5, 28, 247, 5, 5, 29, 247, 5, 5, 30, 247, 5, 5, 31, 247, 5, 5, 32, 247, 5, 5, 33, 247, 5, 5, 34, 247, 5, 5, 35, 247, 5, 5, 36, 247, 5, 5, 37, 247, 5, 5, 38, 247, 5, 5, 39, 247, 5, 5, 40, 247, 5, 5, 41, 247, 5, 5, 42, 247, 5, 5, 57, 248, 5, 5, 58, 248, 5, 5, 59, 248, 5, 5, 60, 248, 5, 5, 61, 248, 5, 5, 62, 248, 5, 5, 63, 248, 5, 5, 64, 248, 5, 5, 65, 248, 5, 5, 66, 248, 5, 5, 67, 248, 5, 5, 68, 248, 5, 5, 69, 248, 5, 5, 70, 248, 5, 5, 71, 248, 5, 5, 72, 248, 5, 5, 73, 248, 5, 5, 74, 248, 5, 5, 43, 247, 5, 5, 10, 249, 5, 5, 11, 249, 5, 5, 12, 249, 5, 5, 13, 249, 5, 5, 14, 249, 5, 5, 170, 249, 5, 5, 171, 249, 5, 5, 172, 249, 5, 5, 173, 249, 5, 5, 174, 249, 5, 5, 175, 249, 5, 5, 176, 249, 5, 5, 247, 249, 5, 5, 248, 249, 5, 5, 249, 249, 5, 5, 250, 249, 5, 5, 251, 249, 5, 5, 252, 249, 5, 5, 253, 249, 5, 5, 52, 250, 5, 5, 53, 250, 5, 5, 54, 250, 5, 5, 55, 250, 5, 5, 100, 250, 5, 5, 101, 250, 5, 5, 102, 250, 5, 5, 103, 250, 5, 5, 126, 250, 5, 5, 127, 250, 5, 5, 203, 159, 5, 5, 13, 162, 5, 5, 14, 162, 5, 5, 203, 165, 5, 5, 204, 165, 5, 5, 205, 165, 5, 5, 206, 165, 5, 5, 207, 165, 5, 5, 73, 171, 5, 5, 74, 171, 5, 5, 75, 171, 5, 5, 76, 171, 5, 5, 77, 171, 5, 5, 78, 171, 5, 5, 79, 171, 5, 5, 80, 171, 5, 5, 81, 171, 5, 5, 82, 171, 5, 5, 83, 171, 5, 5, 143, 177, 5, 5, 144, 177, 5, 5, 145, 177, 5, 5, 146, 177, 5, 5, 147, 177, 5, 5, 148, 177, 5, 5, 149, 177, 5, 5, 150, 177, 5, 5, 151, 177, 5, 5, 152, 177, 5, 5, 153, 177, 5, 5, 154, 177, 5, 5, 155, 177, 5, 5, 156, 177, 5, 5, 157, 177, 5, 5, 158, 177, 5, 5, 159, 177, 5, 5, 160, 177, 5, 5, 161, 177, 5, 5, 162, 177, 5, 5, 163, 177, 5, 5, 164, 177, 5, 5, 165, 177, 5, 5, 166, 177, 5, 5, 220, 184, 5, 5, 221, 184, 5, 5, 222, 184, 5, 5, 223, 184, 5, 5, 224, 184, 5, 5, 225, 184, 5, 5, 226, 184, 5, 5, 227, 184, 5, 5, 228, 184, 5, 5, 229, 184, 5, 5, 230, 184, 5, 5, 231, 184, 5, 5, 232, 184, 5, 5, 233, 184, 5, 5, 234, 184, 5, 5, 235, 184, 5, 5, 236, 184, 5, 5, 237, 184, 5, 5, 238, 184, 5, 5, 239, 184, 5, 5, 240, 184, 5, 5, 241, 184, 5, 5, 242, 184, 5, 5, 243, 184, 5, 5, 244, 184, 5, 5, 245, 184, 5, 5, 246, 184, 5, 5, 247, 184, 5, 5, 248, 184, 5, 5, 249, 184, 5, 5, 250, 184, 5, 5, 230, 192, 5, 5, 231, 192, 5, 5, 232, 192, 5, 5, 233, 192, 5, 5, 234, 192, 5, 5, 235, 192, 5, 5, 236, 192, 5, 5, 237, 192, 5, 5, 238, 192, 5, 5, 239, 192, 5, 5, 240, 192, 5, 5, 241, 192, 5, 5, 242, 192, 5, 5, 243, 192, 5, 5, 244, 192, 5, 5, 245, 192, 5, 5, 246, 192, 5, 5, 247, 192, 5, 5, 248, 192, 5, 5, 249, 192, 5, 5, 250, 192, 5, 5, 251, 192, 5, 5, 252, 192, 5, 5, 253, 192, 5, 5, 254, 192, 5, 5, 255, 192, 5, 5, 2, 193, 5, 5, 3, 193, 5, 5, 4, 193, 5, 5, 5, 193, 5, 5, 6, 193, 5, 5, 7, 193, 5, 5, 8, 193, 5, 5, 9, 193, 5, 5, 10, 193, 5, 5, 11, 193, 5, 5, 12, 193, 5, 5, 13, 193, 5, 5, 14, 193, 5, 5, 15, 193, 5, 5, 16, 193, 5, 5, 76, 201, 5, 5, 77, 201, 5, 5, 78, 201, 5, 5, 79, 201, 5, 5, 80, 201, 5, 5, 81, 201, 5, 5, 82, 201, 5, 5, 83, 201, 5, 5, 84, 201, 5, 5, 85, 201, 5, 5, 86, 201, 5, 5, 87, 201, 5, 5, 88, 201, 5, 5, 89, 201, 5, 5, 90, 201, 5, 5, 91, 201, 5, 5, 92, 201, 5, 5, 93, 201, 5, 5, 94, 201, 5, 5, 95, 201, 5, 5, 96, 201, 5, 5, 97, 201, 5, 5, 98, 201, 5, 5, 99, 201, 5, 5, 100, 201, 5, 5, 101, 201, 5, 5, 102, 201, 5, 5, 103, 201, 5, 5, 104, 201, 5, 5, 105, 201, 5, 5, 8, 209, 5, 5, 9, 209, 5, 5, 10, 209, 5, 5, 11, 209, 5, 5, 12, 209, 5, 5, 13, 209, 5, 5, 14, 209, 5, 5, 15, 209, 5, 5, 16, 209, 5, 5, 17, 209, 5, 5, 18, 209, 5, 5, 19, 209, 5, 5, 20, 209, 5, 5, 21, 209, 5, 5, 22, 209, 5, 5, 23, 209, 5, 5, 24, 209, 5, 5, 25, 209, 5, 5, 26, 209, 5, 5, 27, 209, 5, 5, 28, 209, 5, 5, 29, 209, 5, 5, 30, 209, 5, 5, 31, 209, 5, 5, 32, 209, 5, 5, 33, 209, 5, 5, 34, 209, 5, 5, 35, 209, 5, 5, 48, 216, 5, 5, 49, 216, 5, 5, 50, 216, 5, 5, 51, 216, 5, 5, 52, 216, 5, 5, 53, 216, 5, 5, 54, 216, 5, 5, 55, 216, 5, 5, 56, 216, 5, 5, 57, 216, 5, 5, 58, 216, 5, 5, 59, 216, 5, 5, 60, 216, 5, 5, 61, 216, 5, 5, 62, 216, 5, 5, 63, 216, 5, 5, 64, 216, 5, 5, 65, 216, 5, 5, 66, 216, 5, 5, 67, 216, 5, 5, 48, 223, 5, 5, 49, 223, 5, 5, 50, 223, 5, 5, 51, 223, 5, 5, 52, 223, 5, 5, 53, 223, 5, 5, 54, 223, 5, 5, 55, 223, 5, 5, 56, 223, 5, 5, 57, 223, 5, 5, 58, 223, 5, 5, 59, 223, 5, 5, 60, 223, 5, 5, 61, 223, 5, 5, 62, 223, 5, 5, 63, 223, 5, 5, 37, 229, 5, 5, 38, 229, 5, 5, 39, 229, 5, 5, 40, 229, 5, 5, 41, 229, 5, 5, 42, 229, 5, 5, 43, 229, 5, 5, 44, 229, 5, 5, 45, 229, 5, 5, 46, 229, 5, 5, 47, 229, 5, 5, 34, 234, 5, 5, 35, 234, 5, 5, 36, 234, 5, 5, 37, 234, 5, 5, 38, 234, 5, 5, 39, 234, 5, 5, 40, 234, 5, 5, 41, 234, 5, 5, 42, 234, 5, 5, 43, 234, 5, 5, 44, 234, 5, 5, 228, 237, 5, 5, 229, 237, 5, 5, 230, 237, 5, 5, 231, 237, 5, 5, 232, 237, 5, 5, 233, 237, 5, 5, 52, 241, 5, 5, 204, 243, 5, 5, 205, 243, 5, 5, 49, 247, 5, 5, 50, 247, 5, 5, 84, 171, 5, 5, 85, 171, 5, 5, 17, 193, 5, 5, 18, 193, 5, 5, 106, 201, 5, 5, 64, 223, 5, 5, 53, 241, 5, 5, 51, 247, 5, 5, 204, 159, 5, 5, 86, 171, 5, 5, 167, 177, 5, 5, 168, 177, 5, 5, 251, 184, 5, 5, 252, 184, 5, 5, 253, 184, 5, 5, 19, 193, 5, 5, 20, 193, 5, 5, 21, 193, 5, 5, 22, 193, 5, 5, 23, 193, 5, 5, 107, 201, 5, 5, 108, 201, 5, 5, 109, 201, 5, 5, 110, 201, 5, 5, 111, 201, 5, 5, 112, 201, 5, 5, 113, 201, 5, 5, 114, 201, 5, 5, 115, 201, 5, 5, 116, 201, 5, 5, 117, 201, 5, 5, 118, 201, 5, 5, 119, 201, 5, 5, 37, 209, 5, 5, 38, 209, 5, 5, 39, 209, 5, 5, 40, 209, 5, 5, 41, 209, 5, 5, 42, 209, 5, 5, 43, 209, 5, 5, 44, 209, 5, 5, 45, 209, 5, 5, 68, 216, 5, 5, 69, 216, 5, 5, 70, 216, 5, 5, 71, 216, 5, 5, 72, 216, 5, 5, 73, 216, 5, 5, 74, 216, 5, 5, 75, 216, 5, 5, 76, 216, 5, 5, 77, 216, 5, 5, 65, 223, 5, 5, 66, 223, 5, 5, 67, 223, 5, 5, 68, 223, 5, 5, 69, 223, 5, 5, 70, 223, 5, 5, 71, 223, 5, 5, 72, 223, 5, 5, 73, 223, 5, 5, 74, 223, 5, 5, 48, 229, 5, 5, 49, 229, 5, 5, 46, 234, 5, 5, 50, 229, 5, 5, 51, 229, 5, 5, 52, 229, 5, 5, 53, 229, 5, 5, 54, 229, 5, 5, 55, 229, 5, 5, 56, 229, 5, 5, 57, 229, 5, 5, 45, 234, 5, 5, 58, 229, 5, 5, 59, 229, 5, 5, 47, 234, 5, 5, 48, 234, 5, 5, 49, 234, 5, 5, 50, 234, 5, 5, 51, 234, 5, 5, 52, 234, 5, 5, 53, 234, 5, 5, 54, 234, 5, 5, 55, 234, 5, 5, 56, 234, 5, 5, 62, 229, 5, 5, 57, 234, 5, 5, 58, 234, 5, 5, 234, 237, 5, 5, 235, 237, 5, 5, 236, 237, 5, 5, 237, 237, 5, 5, 238, 237, 5, 5, 239, 237, 5, 5, 240, 237, 5, 5, 241, 237, 5, 5, 242, 237, 5, 5, 54, 241, 5, 5, 55, 241, 5, 5, 56, 241, 5, 5, 57, 241, 5, 5, 58, 241, 5, 5, 206, 243, 5, 5, 207, 243, 5, 5, 208, 243, 5, 5, 209, 243, 5, 5, 178, 245, 5, 5, 179, 245, 5, 5, 180, 245, 5, 5, 181, 245, 5, 5, 182, 245, 5, 5, 52, 247, 5, 5, 155, 157, 5, 5, 121, 158, 5, 5, 205, 159, 5, 5, 15, 162, 5, 5, 16, 162, 5, 5, 17, 162, 5, 5, 18, 162, 5, 5, 19, 162, 5, 5, 208, 165, 5, 5, 209, 165, 5, 5, 210, 165, 5, 5, 211, 165, 5, 5, 212, 165, 5, 5, 213, 165, 5, 5, 214, 165, 5, 5, 215, 165, 5, 5, 87, 171, 5, 5, 88, 171, 5, 5, 169, 177, 5, 5, 170, 177, 5, 5, 171, 177, 5, 5, 172, 177, 5, 5, 173, 177, 5, 5, 174, 177, 5, 5, 175, 177, 5, 5, 176, 177, 5, 5, 177, 177, 5, 5, 2, 185, 5, 5, 3, 185, 5, 5, 4, 185, 5, 5, 5, 185, 5, 5, 25, 193, 5, 5, 26, 193, 5, 5, 27, 193, 5, 5, 28, 193, 5, 5, 29, 193, 5, 5, 30, 193, 5, 5, 31, 193, 5, 5, 32, 193, 5, 5, 33, 193, 5, 5, 34, 193, 5, 5, 120, 201, 5, 5, 121, 201, 5, 5, 122, 201, 5, 5, 123, 201, 5, 5, 124, 201, 5, 5, 46, 209, 5, 5, 47, 209, 5, 5, 48, 209, 5, 5, 49, 209, 5, 5, 78, 216, 5, 5, 63, 229, 5, 5, 89, 171, 5, 5, 156, 157, 5, 5, 206, 159, 5, 5, 207, 159, 5, 5, 20, 162, 5, 5, 21, 162, 5, 5, 22, 162, 5, 5, 23, 162, 5, 5, 24, 162, 5, 5, 216, 165, 5, 5, 217, 165, 5, 5, 218, 165, 5, 5, 219, 165, 5, 5, 220, 165, 5, 5, 221, 165, 5, 5, 222, 165, 5, 5, 223, 165, 5, 5, 224, 165, 5, 5, 225, 165, 5, 5, 226, 165, 5, 5, 227, 165, 5, 5, 228, 165, 5, 5, 229, 165, 5, 5, 230, 165, 5, 5, 231, 165, 5, 5, 232, 165, 5, 5, 233, 165, 5, 5, 90, 171, 5, 5, 91, 171, 5, 5, 92, 171, 5, 5, 93, 171, 5, 5, 94, 171, 5, 5, 95, 171, 5, 5, 96, 171, 5, 5, 97, 171, 5, 5, 98, 171, 5, 5, 99, 171, 5, 5, 100, 171, 5, 5, 101, 171, 5, 5, 102, 171, 5, 5, 103, 171, 5, 5, 104, 171, 5, 5, 105, 171, 5, 5, 106, 171, 5, 5, 107, 171, 5, 5, 108, 171, 5, 5, 178, 177, 5, 5, 179, 177, 5, 5, 180, 177, 5, 5, 181, 177, 5, 5, 182, 177, 5, 5, 183, 177, 5, 5, 184, 177, 5, 5, 185, 177, 5, 5, 186, 177, 5, 5, 187, 177, 5, 5, 188, 177, 5, 5, 189, 177, 5, 5, 6, 185, 5, 5, 7, 185, 5, 5, 8, 185, 5, 5, 9, 185, 5, 5, 25, 185, 5, 5, 10, 185, 5, 5, 11, 185, 5, 5, 12, 185, 5, 5, 13, 185, 5, 5, 14, 185, 5, 5, 15, 185, 5, 5, 16, 185, 5, 5, 17, 185, 5, 5, 18, 185, 5, 5, 19, 185, 5, 5, 20, 185, 5, 5, 21, 185, 5, 5, 22, 185, 5, 5, 23, 185, 5, 5, 24, 185, 5, 5, 35, 193, 5, 5, 36, 193, 5, 5, 37, 193, 5, 5, 38, 193, 5, 5, 39, 193, 5, 5, 40, 193, 5, 5, 41, 193, 5, 5, 42, 193, 5, 5, 125, 201, 5, 5, 43, 193, 5, 5, 44, 193, 5, 5, 45, 193, 5, 5, 46, 193, 5, 5, 47, 193, 5, 5, 48, 193, 5, 5, 49, 193, 5, 5, 50, 193, 5, 5, 126, 201, 5, 5, 51, 193, 5, 5, 127, 201, 5, 5, 128, 201, 5, 5, 129, 201, 5, 5, 130, 201, 5, 5, 131, 201, 5, 5, 132, 201, 5, 5, 133, 201, 5, 5, 134, 201, 5, 5, 135, 201, 5, 5, 136, 201, 5, 5, 137, 201, 5, 5, 138, 201, 5, 5, 139, 201, 5, 5, 140, 201, 5, 5, 141, 201, 5, 5, 142, 201, 5, 5, 143, 201, 5, 5, 144, 201, 5, 5, 50, 209, 5, 5, 145, 201, 5, 5, 51, 209, 5, 5, 52, 209, 5, 5, 53, 209, 5, 5, 54, 209, 5, 5, 55, 209, 5, 5, 56, 209, 5, 5, 57, 209, 5, 5, 58, 209, 5, 5, 79, 216, 5, 5, 80, 216, 5, 5, 81, 216, 5, 5, 82, 216, 5, 5, 83, 216, 5, 5, 84, 216, 5, 5, 85, 216, 5, 5, 86, 216, 5, 5, 87, 216, 5, 5, 76, 223, 5, 5, 77, 223, 5, 5, 78, 223, 5, 5, 79, 223, 5, 5, 64, 229, 5, 5, 65, 229, 5, 5, 66, 229, 5, 5, 67, 229, 5, 5, 68, 229, 5, 5, 69, 229, 5, 5, 59, 234, 5, 5, 60, 234, 5, 5, 61, 234, 5, 5, 62, 234, 5, 5, 63, 234, 5, 5, 64, 234, 5, 5, 65, 234, 5, 5, 244, 237, 5, 5, 59, 241, 5, 5, 211, 243, 5, 5, 109, 171, 5, 5, 71, 229, 5, 5, 66, 234, 5, 5, 110, 171, 5, 5, 27, 185, 5, 5, 28, 185, 5, 5, 29, 185, 5, 5, 30, 185, 5, 5, 31, 185, 5, 5, 55, 193, 5, 5, 56, 193, 5, 5, 148, 201, 5, 5, 149, 201, 5, 5, 150, 201, 5, 5, 151, 201, 5, 5, 152, 201, 5, 5, 153, 201, 5, 5, 154, 201, 5, 5, 155, 201, 5, 5, 61, 209, 5, 5, 62, 209, 5, 5, 63, 209, 5, 5, 88, 216, 5, 5, 64, 209, 5, 5, 65, 209, 5, 5, 66, 209, 5, 5, 89, 216, 5, 5, 90, 216, 5, 5, 91, 216, 5, 5, 81, 223, 5, 5, 72, 229, 5, 5, 73, 229, 5, 5, 67, 234, 5, 5, 245, 237, 5, 5, 246, 237, 5, 5, 247, 237, 5, 5, 248, 237, 5, 5, 249, 237, 5, 5, 250, 237, 5, 5, 251, 237, 5, 5, 252, 237, 5, 5, 253, 237, 5, 5, 254, 237, 5, 5, 60, 241, 5, 5, 2, 238, 5, 5, 61, 241, 5, 5, 183, 245, 5, 5, 22, 249, 5, 5, 23, 249, 5, 5, 106, 250, 5, 5, 112, 171, 5, 5, 57, 193, 5, 5, 58, 193, 5, 5, 59, 193, 5, 5, 156, 201, 5, 5, 157, 201, 5, 5, 158, 201, 5, 5, 159, 201, 5, 5, 160, 201, 5, 5, 161, 201, 5, 5, 162, 201, 5, 5, 163, 201, 5, 5, 67, 209, 5, 5, 68, 209, 5, 5, 69, 209, 5, 5, 70, 209, 5, 5, 71, 209, 5, 5, 72, 209, 5, 5, 73, 209, 5, 5, 74, 209, 5, 5, 75, 209, 5, 5, 76, 209, 5, 5, 77, 209, 5, 5, 93, 216, 5, 5, 94, 216, 5, 5, 95, 216, 5, 5, 82, 223, 5, 5, 83, 223, 5, 5, 84, 223, 5, 5, 85, 223, 5, 5, 86, 223, 5, 5, 87, 223, 5, 5, 88, 223, 5, 5, 89, 223, 5, 5, 90, 223, 5, 5, 74, 229, 5, 5, 75, 229, 5, 5, 76, 229, 5, 5, 77, 229, 5, 5, 78, 229, 5, 5, 79, 229, 5, 5, 80, 229, 5, 5, 81, 229, 5, 5, 82, 229, 5, 5, 83, 229, 5, 5, 84, 229, 5, 5, 85, 229, 5, 5, 86, 229, 5, 5, 70, 234, 5, 5, 71, 234, 5, 5, 72, 234, 5, 5, 73, 234, 5, 5, 74, 234, 5, 5, 75, 234, 5, 5, 76, 234, 5, 5, 77, 234, 5, 5, 78, 234, 5, 5, 82, 234, 5, 5, 3, 238, 5, 5, 4, 238, 5, 5, 5, 238, 5, 5, 6, 238, 5, 5, 63, 241, 5, 5, 64, 241, 5, 5, 65, 241, 5, 5, 66, 241, 5, 5, 67, 241, 5, 5, 68, 241, 5, 5, 69, 241, 5, 5, 70, 241, 5, 5, 212, 243, 5, 5, 213, 243, 5, 5, 214, 243, 5, 5, 215, 243, 5, 5, 184, 245, 5, 5, 216, 243, 5, 5, 217, 243, 5, 5, 185, 245, 5, 5, 186, 245, 5, 5, 187, 245, 5, 5, 188, 245, 5, 5, 189, 245, 5, 5, 190, 245, 5, 5, 191, 245, 5, 5, 53, 247, 5, 5, 54, 247, 5, 5, 55, 247, 5, 5, 56, 247, 5, 5, 57, 247, 5, 5, 83, 248, 5, 5, 24, 249, 5, 5, 25, 249, 5, 5, 26, 249, 5, 5, 27, 249, 5, 5, 28, 249, 5, 5, 29, 249, 5, 5, 30, 249, 5, 5, 179, 249, 5, 5, 3, 250, 5, 5, 58, 250, 5, 5, 59, 250, 5, 5, 60, 250, 5, 5, 61, 250, 5, 5, 129, 250, 5, 5, 174, 250, 5, 5, 113, 171, 5, 5, 114, 171, 5, 5, 166, 201, 5, 5, 167, 201, 5, 5, 79, 209, 5, 5, 80, 209, 5, 5, 100, 216, 5, 5, 101, 216, 5, 5, 102, 216, 5, 5, 92, 223, 5, 5, 92, 229, 5, 5, 93, 229, 5, 5, 8, 238, 5, 5, 115, 171, 5, 5, 168, 201, 5, 5, 93, 223, 5, 5, 74, 241, 5, 5, 192, 177, 5, 5, 116, 171, 5, 5, 103, 216, 5, 5, 94, 223, 5, 5, 94, 229, 5, 5, 192, 245, 5, 5, 86, 248, 5, 5, 193, 177, 5, 5, 63, 193, 5, 5, 169, 201, 5, 5, 170, 201, 5, 5, 171, 201, 5, 5, 172, 201, 5, 5, 173, 201, 5, 5, 174, 201, 5, 5, 175, 201, 5, 5, 81, 209, 5, 5, 82, 209, 5, 5, 83, 209, 5, 5, 84, 209, 5, 5, 85, 209, 5, 5, 86, 209, 5, 5, 87, 209, 5, 5, 88, 209, 5, 5, 104, 216, 5, 5, 105, 216, 5, 5, 106, 216, 5, 5, 107, 216, 5, 5, 108, 216, 5, 5, 109, 216, 5, 5, 110, 216, 5, 5, 111, 216, 5, 5, 112, 216, 5, 5, 113, 216, 5, 5, 114, 216, 5, 5, 115, 216, 5, 5, 116, 216, 5, 5, 95, 223, 5, 5, 96, 223, 5, 5, 97, 223, 5, 5, 98, 223, 5, 5, 99, 223, 5, 5, 100, 223, 5, 5, 101, 223, 5, 5, 102, 223, 5, 5, 103, 223, 5, 5, 104, 223, 5, 5, 105, 223, 5, 5, 106, 223, 5, 5, 95, 229, 5, 5, 96, 229, 5, 5, 97, 229, 5, 5, 98, 229, 5, 5, 99, 229, 5, 5, 100, 229, 5, 5, 101, 229, 5, 5, 84, 234, 5, 5, 85, 234, 5, 5, 86, 234, 5, 5, 87, 234, 5, 5, 88, 234, 5, 5, 89, 234, 5, 5, 90, 234, 5, 5, 91, 234, 5, 5, 9, 238, 5, 5, 10, 238, 5, 5, 11, 238, 5, 5, 12, 238, 5, 5, 13, 238, 5, 5, 14, 238, 5, 5, 15, 238, 5, 5, 16, 238, 5, 5, 17, 238, 5, 5, 18, 238, 5, 5, 19, 238, 5, 5, 20, 238, 5, 5, 21, 238, 5, 5, 22, 238, 5, 5, 23, 238, 5, 5, 75, 241, 5, 5, 76, 241, 5, 5, 77, 241, 5, 5, 78, 241, 5, 5, 79, 241, 5, 5, 80, 241, 5, 5, 81, 241, 5, 5, 221, 243, 5, 5, 222, 243, 5, 5, 223, 243, 5, 5, 224, 243, 5, 5, 193, 245, 5, 5, 194, 245, 5, 5, 195, 245, 5, 5, 196, 245, 5, 5, 59, 247, 5, 5, 60, 247, 5, 5, 61, 247, 5, 5, 62, 247, 5, 5, 87, 248, 5, 5, 88, 248, 5, 5, 31, 249, 5, 5, 32, 249, 5, 5, 33, 249, 5, 5, 4, 250, 5, 5, 143, 250, 5, 5, 194, 177, 5, 5, 176, 201, 5, 5, 117, 216, 5, 5, 118, 216, 5, 5, 107, 223, 5, 5, 108, 223, 5, 5, 109, 223, 5, 5, 102, 229, 5, 5, 92, 234, 5, 5, 93, 234, 5, 5, 94, 234, 5, 5, 24, 238, 5, 5, 25, 238, 5, 5, 26, 238, 5, 5, 27, 238, 5, 5, 28, 238, 5, 5, 225, 243, 5, 5, 82, 241, 5, 5, 83, 241, 5, 5, 84, 241, 5, 5, 85, 241, 5, 5, 226, 243, 5, 5, 197, 245, 5, 5, 198, 245, 5, 5, 63, 247, 5, 5, 34, 249, 5, 5, 35, 249, 5, 5, 122, 158, 5, 5, 236, 165, 5, 5, 195, 177, 5, 5, 177, 201, 5, 5, 89, 209, 5, 5, 90, 209, 5, 5, 119, 216, 5, 5, 196, 177, 5, 5, 91, 209, 5, 5, 110, 223, 5, 5, 103, 229, 5, 5, 95, 234, 5, 5, 86, 241, 5, 5, 197, 177, 5, 5, 92, 209, 5, 5, 93, 209, 5, 5, 120, 216, 5, 5, 121, 216, 5, 5, 104, 229, 5, 5, 30, 238, 5, 5, 31, 238, 5, 5, 87, 241, 5, 5, 88, 241, 5, 5, 227, 243, 5, 5, 228, 243, 5, 5, 229, 243, 5, 5, 89, 248, 5, 5, 198, 177, 5, 5, 66, 193, 5, 5, 67, 193, 5, 5, 68, 193, 5, 5, 178, 201, 5, 5, 179, 201, 5, 5, 180, 201, 5, 5, 181, 201, 5, 5, 94, 209, 5, 5, 95, 209, 5, 5, 96, 209, 5, 5, 97, 209, 5, 5, 98, 209, 5, 5, 99, 209, 5, 5, 100, 209, 5, 5, 101, 209, 5, 5, 102, 209, 5, 5, 103, 209, 5, 5, 104, 209, 5, 5, 124, 216, 5, 5, 125, 216, 5, 5, 126, 216, 5, 5, 127, 216, 5, 5, 128, 216, 5, 5, 123, 216, 5, 5, 129, 216, 5, 5, 112, 223, 5, 5, 113, 223, 5, 5, 114, 223, 5, 5, 115, 223, 5, 5, 116, 223, 5, 5, 117, 223, 5, 5, 118, 223, 5, 5, 119, 223, 5, 5, 120, 223, 5, 5, 105, 229, 5, 5, 106, 229, 5, 5, 121, 223, 5, 5, 122, 223, 5, 5, 123, 223, 5, 5, 124, 223, 5, 5, 125, 223, 5, 5, 126, 223, 5, 5, 127, 223, 5, 5, 107, 229, 5, 5, 108, 229, 5, 5, 109, 229, 5, 5, 110, 229, 5, 5, 111, 229, 5, 5, 112, 229, 5, 5, 113, 229, 5, 5, 114, 229, 5, 5, 115, 229, 5, 5, 116, 229, 5, 5, 117, 229, 5, 5, 118, 229, 5, 5, 119, 229, 5, 5, 120, 229, 5, 5, 121, 229, 5, 5, 122, 229, 5, 5, 123, 229, 5, 5, 34, 238, 5, 5, 33, 238, 5, 5, 96, 234, 5, 5, 97, 234, 5, 5, 98, 234, 5, 5, 99, 234, 5, 5, 100, 234, 5, 5, 101, 234, 5, 5, 102, 234, 5, 5, 103, 234, 5, 5, 104, 234, 5, 5, 105, 234, 5, 5, 106, 234, 5, 5, 35, 238, 5, 5, 36, 238, 5, 5, 37, 238, 5, 5, 38, 238, 5, 5, 39, 238, 5, 5, 40, 238, 5, 5, 41, 238, 5, 5, 42, 238, 5, 5, 43, 238, 5, 5, 44, 238, 5, 5, 45, 238, 5, 5, 89, 241, 5, 5, 90, 241, 5, 5, 91, 241, 5, 5, 92, 241, 5, 5, 93, 241, 5, 5, 94, 241, 5, 5, 95, 241, 5, 5, 96, 241, 5, 5, 97, 241, 5, 5, 230, 243, 5, 5, 231, 243, 5, 5, 232, 243, 5, 5, 233, 243, 5, 5, 234, 243, 5, 5, 201, 245, 5, 5, 202, 245, 5, 5, 203, 245, 5, 5, 204, 245, 5, 5, 205, 245, 5, 5, 64, 247, 5, 5, 65, 247, 5, 5, 66, 247, 5, 5, 90, 248, 5, 5, 91, 248, 5, 5, 92, 248, 5, 5, 93, 248, 5, 5, 37, 249, 5, 5, 180, 249, 5, 5, 181, 249, 5, 5, 62, 250, 5, 5, 63, 250, 5, 5, 25, 162, 5, 5, 117, 171, 5, 5, 118, 171, 5, 5, 199, 177, 5, 5, 200, 177, 5, 5, 201, 177, 5, 5, 202, 177, 5, 5, 32, 185, 5, 5, 33, 185, 5, 5, 34, 185, 5, 5, 35, 185, 5, 5, 36, 185, 5, 5, 37, 185, 5, 5, 38, 185, 5, 5, 39, 185, 5, 5, 40, 185, 5, 5, 69, 193, 5, 5, 70, 193, 5, 5, 71, 193, 5, 5, 72, 193, 5, 5, 182, 201, 5, 5, 183, 201, 5, 5, 184, 201, 5, 5, 185, 201, 5, 5, 186, 201, 5, 5, 187, 201, 5, 5, 188, 201, 5, 5, 106, 209, 5, 5, 107, 209, 5, 5, 108, 209, 5, 5, 109, 209, 5, 5, 110, 209, 5, 5, 111, 209, 5, 5, 112, 209, 5, 5, 131, 216, 5, 5, 128, 223, 5, 5, 129, 223, 5, 5, 130, 223, 5, 5, 131, 223, 5, 5, 132, 223, 5, 5, 133, 223, 5, 5, 127, 229, 5, 5, 128, 229, 5, 5, 129, 229, 5, 5, 130, 229, 5, 5, 46, 238, 5, 5, 47, 238, 5, 5, 99, 241, 5, 5, 235, 243, 5, 5, 206, 245, 5, 5, 94, 248, 5, 5, 203, 177, 5, 5, 189, 201, 5, 5, 190, 201, 5, 5, 113, 209, 5, 5, 114, 209, 5, 5, 132, 216, 5, 5, 133, 216, 5, 5, 134, 216, 5, 5, 135, 216, 5, 5, 136, 216, 5, 5, 134, 223, 5, 5, 135, 223, 5, 5, 131, 229, 5, 5, 132, 229, 5, 5, 110, 234, 5, 5, 111, 234, 5, 5, 48, 238, 5, 5, 49, 238, 5, 5, 50, 238, 5, 5, 100, 241, 5, 5, 101, 241, 5, 5, 102, 241, 5, 5, 103, 241, 5, 5, 104, 241, 5, 5, 105, 241, 5, 5, 236, 243, 5, 5, 237, 243, 5, 5, 238, 243, 5, 5, 239, 243, 5, 5, 207, 245, 5, 5, 208, 245, 5, 5, 209, 245, 5, 5, 210, 245, 5, 5, 211, 245, 5, 5, 212, 245, 5, 5, 67, 247, 5, 5, 64, 250, 5, 5, 65, 250, 5, 5, 123, 158, 5, 5, 237, 165, 5, 5, 204, 177, 5, 5, 205, 177, 5, 5, 206, 177, 5, 5, 192, 201, 5, 5, 115, 209, 5, 5, 141, 216, 5, 5, 140, 216, 5, 5, 142, 216, 5, 5, 136, 223, 5, 5, 134, 229, 5, 5, 135, 229, 5, 5, 207, 177, 5, 5, 213, 245, 5, 5, 66, 250, 5, 5, 157, 157, 5, 5, 208, 177, 5, 5, 209, 177, 5, 5, 73, 193, 5, 5, 41, 185, 5, 5, 42, 185, 5, 5, 43, 185, 5, 5, 74, 193, 5, 5, 75, 193, 5, 5, 193, 201, 5, 5, 194, 201, 5, 5, 195, 201, 5, 5, 196, 201, 5, 5, 197, 201, 5, 5, 116, 209, 5, 5, 198, 201, 5, 5, 117, 209, 5, 5, 199, 201, 5, 5, 200, 201, 5, 5, 118, 209, 5, 5, 201, 201, 5, 5, 120, 209, 5, 5, 121, 209, 5, 5, 122, 209, 5, 5, 123, 209, 5, 5, 124, 209, 5, 5, 143, 216, 5, 5, 125, 209, 5, 5, 137, 223, 5, 5, 126, 209, 5, 5, 127, 209, 5, 5, 128, 209, 5, 5, 129, 209, 5, 5, 130, 209, 5, 5, 145, 216, 5, 5, 146, 216, 5, 5, 147, 216, 5, 5, 148, 216, 5, 5, 149, 216, 5, 5, 150, 216, 5, 5, 151, 216, 5, 5, 152, 216, 5, 5, 138, 223, 5, 5, 153, 216, 5, 5, 139, 223, 5, 5, 140, 223, 5, 5, 154, 216, 5, 5, 141, 223, 5, 5, 155, 216, 5, 5, 156, 216, 5, 5, 136, 229, 5, 5, 143, 223, 5, 5, 144, 223, 5, 5, 145, 223, 5, 5, 146, 223, 5, 5, 147, 223, 5, 5, 148, 223, 5, 5, 149, 223, 5, 5, 150, 223, 5, 5, 151, 223, 5, 5, 138, 229, 5, 5, 139, 229, 5, 5, 140, 229, 5, 5, 137, 229, 5, 5, 141, 229, 5, 5, 142, 229, 5, 5, 143, 229, 5, 5, 144, 229, 5, 5, 145, 229, 5, 5, 146, 229, 5, 5, 147, 229, 5, 5, 116, 234, 5, 5, 148, 229, 5, 5, 149, 229, 5, 5, 150, 229, 5, 5, 151, 229, 5, 5, 117, 234, 5, 5, 118, 234, 5, 5, 119, 234, 5, 5, 120, 234, 5, 5, 51, 238, 5, 5, 121, 234, 5, 5, 122, 234, 5, 5, 123, 234, 5, 5, 124, 234, 5, 5, 125, 234, 5, 5, 153, 229, 5, 5, 126, 234, 5, 5, 55, 238, 5, 5, 127, 234, 5, 5, 56, 238, 5, 5, 57, 238, 5, 5, 58, 238, 5, 5, 59, 238, 5, 5, 60, 238, 5, 5, 61, 238, 5, 5, 62, 238, 5, 5, 63, 238, 5, 5, 64, 238, 5, 5, 65, 238, 5, 5, 129, 234, 5, 5, 107, 241, 5, 5, 109, 241, 5, 5, 110, 241, 5, 5, 130, 234, 5, 5, 111, 241, 5, 5, 112, 241, 5, 5, 113, 241, 5, 5, 240, 243, 5, 5, 241, 243, 5, 5, 242, 243, 5, 5, 243, 243, 5, 5, 244, 243, 5, 5, 214, 245, 5, 5, 245, 243, 5, 5, 246, 243, 5, 5, 247, 243, 5, 5, 248, 243, 5, 5, 68, 247, 5, 5, 69, 247, 5, 5, 215, 245, 5, 5, 216, 245, 5, 5, 217, 245, 5, 5, 249, 243, 5, 5, 70, 247, 5, 5, 71, 247, 5, 5, 95, 248, 5, 5, 38, 249, 5, 5, 182, 249, 5, 5, 183, 249, 5, 5, 67, 250, 5, 5, 68, 250, 5, 5, 144, 250, 5, 5, 158, 157, 5, 5, 208, 159, 5, 5, 209, 159, 5, 5, 26, 162, 5, 5, 27, 162, 5, 5, 238, 165, 5, 5, 239, 165, 5, 5, 240, 165, 5, 5, 241, 165, 5, 5, 242, 165, 5, 5, 243, 165, 5, 5, 244, 165, 5, 5, 119, 171, 5, 5, 120, 171, 5, 5, 121, 171, 5, 5, 122, 171, 5, 5, 123, 171, 5, 5, 124, 171, 5, 5, 211, 177, 5, 5, 212, 177, 5, 5, 213, 177, 5, 5, 214, 177, 5, 5, 215, 177, 5, 5, 216, 177, 5, 5, 217, 177, 5, 5, 218, 177, 5, 5, 44, 185, 5, 5, 45, 185, 5, 5, 46, 185, 5, 5, 47, 185, 5, 5, 48, 185, 5, 5, 49, 185, 5, 5, 76, 193, 5, 5, 77, 193, 5, 5, 78, 193, 5, 5, 79, 193, 5, 5, 202, 201, 5, 5, 203, 201, 5, 5, 131, 209, 5, 5, 204, 201, 5, 5, 205, 201, 5, 5, 132, 209, 5, 5, 133, 209, 5, 5, 134, 209, 5, 5, 135, 209, 5, 5, 136, 209, 5, 5, 157, 216, 5, 5, 158, 216, 5, 5, 152, 223, 5, 5, 153, 223, 5, 5, 184, 249, 5, 5, 219, 177, 5, 5, 80, 193, 5, 5, 131, 234, 5, 5, 221, 177, 5, 5, 137, 209, 5, 5, 159, 216, 5, 5, 160, 216, 5, 5, 161, 216, 5, 5, 154, 229, 5, 5, 155, 229, 5, 5, 156, 229, 5, 5, 132, 234, 5, 5, 133, 234, 5, 5, 134, 234, 5, 5, 66, 238, 5, 5, 67, 238, 5, 5, 114, 241, 5, 5, 115, 241, 5, 5, 250, 243, 5, 5, 218, 245, 5, 5, 96, 248, 5, 5, 69, 250, 5, 5, 50, 185, 5, 5, 206, 201, 5, 5, 207, 201, 5, 5, 138, 209, 5, 5, 139, 209, 5, 5, 140, 209, 5, 5, 141, 209, 5, 5, 142, 209, 5, 5, 143, 209, 5, 5, 144, 209, 5, 5, 163, 216, 5, 5, 164, 216, 5, 5, 165, 216, 5, 5, 166, 216, 5, 5, 167, 216, 5, 5, 168, 216, 5, 5, 169, 216, 5, 5, 170, 216, 5, 5, 171, 216, 5, 5, 172, 216, 5, 5, 173, 216, 5, 5, 174, 216, 5, 5, 175, 216, 5, 5, 176, 216, 5, 5, 177, 216, 5, 5, 178, 216, 5, 5, 179, 216, 5, 5, 180, 216, 5, 5, 154, 223, 5, 5, 155, 223, 5, 5, 156, 223, 5, 5, 157, 223, 5, 5, 158, 223, 5, 5, 159, 223, 5, 5, 160, 223, 5, 5, 161, 223, 5, 5, 162, 223, 5, 5, 163, 223, 5, 5, 164, 223, 5, 5, 165, 223, 5, 5, 166, 223, 5, 5, 167, 223, 5, 5, 168, 223, 5, 5, 169, 223, 5, 5, 170, 223, 5, 5, 171, 223, 5, 5, 172, 223, 5, 5, 173, 223, 5, 5, 174, 223, 5, 5, 175, 223, 5, 5, 176, 223, 5, 5, 177, 223, 5, 5, 178, 223, 5, 5, 157, 229, 5, 5, 158, 229, 5, 5, 159, 229, 5, 5, 160, 229, 5, 5, 161, 229, 5, 5, 162, 229, 5, 5, 163, 229, 5, 5, 164, 229, 5, 5, 165, 229, 5, 5, 166, 229, 5, 5, 167, 229, 5, 5, 168, 229, 5, 5, 169, 229, 5, 5, 170, 229, 5, 5, 171, 229, 5, 5, 172, 229, 5, 5, 173, 229, 5, 5, 174, 229, 5, 5, 156, 234, 5, 5, 137, 234, 5, 5, 138, 234, 5, 5, 139, 234, 5, 5, 140, 234, 5, 5, 141, 234, 5, 5, 142, 234, 5, 5, 143, 234, 5, 5, 144, 234, 5, 5, 145, 234, 5, 5, 146, 234, 5, 5, 147, 234, 5, 5, 148, 234, 5, 5, 149, 234, 5, 5, 150, 234, 5, 5, 151, 234, 5, 5, 152, 234, 5, 5, 68, 238, 5, 5, 69, 238, 5, 5, 70, 238, 5, 5, 71, 238, 5, 5, 72, 238, 5, 5, 73, 238, 5, 5, 74, 238, 5, 5, 75, 238, 5, 5, 76, 238, 5, 5, 77, 238, 5, 5, 78, 238, 5, 5, 79, 238, 5, 5, 80, 238, 5, 5, 81, 238, 5, 5, 82, 238, 5, 5, 83, 238, 5, 5, 117, 241, 5, 5, 118, 241, 5, 5, 119, 241, 5, 5, 120, 241, 5, 5, 121, 241, 5, 5, 122, 241, 5, 5, 123, 241, 5, 5, 124, 241, 5, 5, 125, 241, 5, 5, 126, 241, 5, 5, 127, 241, 5, 5, 128, 241, 5, 5, 129, 241, 5, 5, 130, 241, 5, 5, 131, 241, 5, 5, 132, 241, 5, 5, 133, 241, 5, 5, 134, 241, 5, 5, 135, 241, 5, 5, 136, 241, 5, 5, 137, 241, 5, 5, 251, 243, 5, 5, 252, 243, 5, 5, 253, 243, 5, 5, 254, 243, 5, 5, 255, 243, 5, 5, 2, 244, 5, 5, 3, 244, 5, 5, 4, 244, 5, 5, 5, 244, 5, 5, 6, 244, 5, 5, 7, 244, 5, 5, 8, 244, 5, 5, 9, 244, 5, 5, 10, 244, 5, 5, 11, 244, 5, 5, 12, 244, 5, 5, 219, 245, 5, 5, 220, 245, 5, 5, 221, 245, 5, 5, 222, 245, 5, 5, 223, 245, 5, 5, 224, 245, 5, 5, 225, 245, 5, 5, 226, 245, 5, 5, 227, 245, 5, 5, 228, 245, 5, 5, 229, 245, 5, 5, 230, 245, 5, 5, 231, 245, 5, 5, 232, 245, 5, 5, 233, 245, 5, 5, 73, 247, 5, 5, 74, 247, 5, 5, 75, 247, 5, 5, 76, 247, 5, 5, 77, 247, 5, 5, 78, 247, 5, 5, 79, 247, 5, 5, 80, 247, 5, 5, 81, 247, 5, 5, 82, 247, 5, 5, 83, 247, 5, 5, 84, 247, 5, 5, 85, 247, 5, 5, 86, 247, 5, 5, 97, 248, 5, 5, 98, 248, 5, 5, 99, 248, 5, 5, 100, 248, 5, 5, 101, 248, 5, 5, 102, 248, 5, 5, 103, 248, 5, 5, 39, 249, 5, 5, 40, 249, 5, 5, 41, 249, 5, 5, 5, 250, 5, 5, 6, 250, 5, 5, 7, 250, 5, 5, 8, 250, 5, 5, 70, 250, 5, 5, 10, 250, 5, 5, 71, 250, 5, 5, 72, 250, 5, 5, 107, 250, 5, 5, 108, 250, 5, 5, 130, 250, 5, 5, 145, 250, 5, 5, 159, 157, 5, 5, 210, 159, 5, 5, 28, 162, 5, 5, 29, 162, 5, 5, 30, 162, 5, 5, 245, 165, 5, 5, 246, 165, 5, 5, 247, 165, 5, 5, 248, 165, 5, 5, 125, 171, 5, 5, 126, 171, 5, 5, 127, 171, 5, 5, 128, 171, 5, 5, 129, 171, 5, 5, 130, 171, 5, 5, 131, 171, 5, 5, 132, 171, 5, 5, 133, 171, 5, 5, 134, 171, 5, 5, 135, 171, 5, 5, 136, 171, 5, 5, 222, 177, 5, 5, 223, 177, 5, 5, 224, 177, 5, 5, 225, 177, 5, 5, 226, 177, 5, 5, 227, 177, 5, 5, 228, 177, 5, 5, 229, 177, 5, 5, 230, 177, 5, 5, 51, 185, 5, 5, 52, 185, 5, 5, 53, 185, 5, 5, 54, 185, 5, 5, 55, 185, 5, 5, 56, 185, 5, 5, 81, 193, 5, 5, 82, 193, 5, 5, 83, 193, 5, 5, 84, 193, 5, 5, 85, 193, 5, 5, 86, 193, 5, 5, 87, 193, 5, 5, 208, 201, 5, 5, 209, 201, 5, 5, 210, 201, 5, 5, 211, 201, 5, 5, 212, 201, 5, 5, 145, 209, 5, 5, 146, 209, 5, 5, 147, 209, 5, 5, 148, 209, 5, 5, 183, 216, 5, 5, 184, 216, 5, 5, 185, 216, 5, 5, 180, 223, 5, 5, 157, 234, 5, 5, 140, 241, 5, 5, 19, 244, 5, 5, 20, 244, 5, 5, 57, 185, 5, 5, 213, 201, 5, 5, 149, 209, 5, 5, 150, 209, 5, 5, 151, 209, 5, 5, 152, 209, 5, 5, 153, 209, 5, 5, 186, 216, 5, 5, 187, 216, 5, 5, 188, 216, 5, 5, 181, 223, 5, 5, 182, 223, 5, 5, 183, 223, 5, 5, 184, 223, 5, 5, 185, 223, 5, 5, 186, 223, 5, 5, 176, 229, 5, 5, 177, 229, 5, 5, 178, 229, 5, 5, 179, 229, 5, 5, 180, 229, 5, 5, 158, 234, 5, 5, 159, 234, 5, 5, 182, 229, 5, 5, 91, 238, 5, 5, 92, 238, 5, 5, 141, 241, 5, 5, 142, 241, 5, 5, 21, 244, 5, 5, 143, 241, 5, 5, 22, 244, 5, 5, 23, 244, 5, 5, 24, 244, 5, 5, 25, 244, 5, 5, 26, 244, 5, 5, 27, 244, 5, 5, 28, 244, 5, 5, 235, 245, 5, 5, 236, 245, 5, 5, 237, 245, 5, 5, 88, 247, 5, 5, 104, 248, 5, 5, 89, 247, 5, 5, 105, 248, 5, 5, 106, 248, 5, 5, 42, 249, 5, 5, 188, 249, 5, 5, 11, 250, 5, 5, 58, 185, 5, 5, 88, 193, 5, 5, 190, 216, 5, 5, 187, 223, 5, 5, 93, 238, 5, 5, 90, 247, 5, 5, 107, 248, 5, 5, 59, 185, 5, 5, 214, 201, 5, 5, 155, 209, 5, 5, 156, 209, 5, 5, 191, 216, 5, 5, 192, 216, 5, 5, 193, 216, 5, 5, 194, 216, 5, 5, 195, 216, 5, 5, 196, 216, 5, 5, 197, 216, 5, 5, 198, 216, 5, 5, 188, 223, 5, 5, 189, 223, 5, 5, 183, 229, 5, 5, 190, 223, 5, 5, 191, 223, 5, 5, 192, 223, 5, 5, 193, 223, 5, 5, 194, 223, 5, 5, 195, 223, 5, 5, 196, 223, 5, 5, 184, 229, 5, 5, 185, 229, 5, 5, 186, 229, 5, 5, 187, 229, 5, 5, 188, 229, 5, 5, 189, 229, 5, 5, 190, 229, 5, 5, 160, 234, 5, 5, 161, 234, 5, 5, 162, 234, 5, 5, 163, 234, 5, 5, 164, 234, 5, 5, 165, 234, 5, 5, 166, 234, 5, 5, 94, 238, 5, 5, 95, 238, 5, 5, 96, 238, 5, 5, 97, 238, 5, 5, 191, 229, 5, 5, 98, 238, 5, 5, 144, 241, 5, 5, 145, 241, 5, 5, 146, 241, 5, 5, 147, 241, 5, 5, 148, 241, 5, 5, 149, 241, 5, 5, 150, 241, 5, 5, 29, 244, 5, 5, 30, 244, 5, 5, 31, 244, 5, 5, 32, 244, 5, 5, 238, 245, 5, 5, 239, 245, 5, 5, 240, 245, 5, 5, 241, 245, 5, 5, 242, 245, 5, 5, 92, 247, 5, 5, 93, 247, 5, 5, 94, 247, 5, 5, 95, 247, 5, 5, 91, 247, 5, 5, 108, 248, 5, 5, 109, 248, 5, 5, 110, 248, 5, 5, 43, 249, 5, 5, 44, 249, 5, 5, 189, 249, 5, 5, 74, 250, 5, 5, 60, 185, 5, 5, 201, 216, 5, 5, 197, 223, 5, 5, 192, 229, 5, 5, 101, 238, 5, 5, 33, 244, 5, 5, 98, 247, 5, 5, 45, 249, 5, 5, 46, 249, 5, 5, 75, 250, 5, 5, 61, 185, 5, 5, 76, 250, 5, 5, 131, 250, 5, 5, 62, 185, 5, 5, 193, 229, 5, 5, 167, 234, 5, 5, 102, 238, 5, 5, 103, 238, 5, 5, 153, 241, 5, 5, 34, 244, 5, 5, 243, 245, 5, 5, 244, 245, 5, 5, 99, 247, 5, 5, 63, 185, 5, 5, 157, 209, 5, 5, 202, 216, 5, 5, 203, 216, 5, 5, 204, 216, 5, 5, 205, 216, 5, 5, 206, 216, 5, 5, 198, 223, 5, 5, 199, 223, 5, 5, 200, 223, 5, 5, 201, 223, 5, 5, 194, 229, 5, 5, 168, 234, 5, 5, 169, 234, 5, 5, 104, 238, 5, 5, 105, 238, 5, 5, 106, 238, 5, 5, 107, 238, 5, 5, 108, 238, 5, 5, 109, 238, 5, 5, 35, 244, 5, 5, 245, 245, 5, 5, 246, 245, 5, 5, 247, 245, 5, 5, 248, 245, 5, 5, 100, 247, 5, 5, 101, 247, 5, 5, 47, 249, 5, 5, 48, 249, 5, 5, 49, 249, 5, 5, 89, 193, 5, 5, 158, 209, 5, 5, 159, 209, 5, 5, 160, 209, 5, 5, 161, 209, 5, 5, 208, 216, 5, 5, 209, 216, 5, 5, 210, 216, 5, 5, 211, 216, 5, 5, 203, 223, 5, 5, 204, 223, 5, 5, 205, 223, 5, 5, 206, 223, 5, 5, 207, 223, 5, 5, 208, 223, 5, 5, 209, 223, 5, 5, 210, 223, 5, 5, 211, 223, 5, 5, 212, 223, 5, 5, 213, 223, 5, 5, 214, 223, 5, 5, 215, 223, 5, 5, 216, 223, 5, 5, 217, 223, 5, 5, 218, 223, 5, 5, 219, 223, 5, 5, 220, 223, 5, 5, 221, 223, 5, 5, 222, 223, 5, 5, 223, 223, 5, 5, 224, 223, 5, 5, 225, 223, 5, 5, 195, 229, 5, 5, 196, 229, 5, 5, 197, 229, 5, 5, 198, 229, 5, 5, 199, 229, 5, 5, 200, 229, 5, 5, 201, 229, 5, 5, 202, 229, 5, 5, 203, 229, 5, 5, 204, 229, 5, 5, 205, 229, 5, 5, 206, 229, 5, 5, 170, 234, 5, 5, 207, 229, 5, 5, 208, 229, 5, 5, 209, 229, 5, 5, 210, 229, 5, 5, 211, 229, 5, 5, 212, 229, 5, 5, 213, 229, 5, 5, 214, 229, 5, 5, 215, 229, 5, 5, 216, 229, 5, 5, 217, 229, 5, 5, 218, 229, 5, 5, 219, 229, 5, 5, 220, 229, 5, 5, 221, 229, 5, 5, 222, 229, 5, 5, 223, 229, 5, 5, 224, 229, 5, 5, 171, 234, 5, 5, 172, 234, 5, 5, 173, 234, 5, 5, 174, 234, 5, 5, 175, 234, 5, 5, 176, 234, 5, 5, 177, 234, 5, 5, 178, 234, 5, 5, 179, 234, 5, 5, 180, 234, 5, 5, 225, 229, 5, 5, 181, 234, 5, 5, 182, 234, 5, 5, 183, 234, 5, 5, 184, 234, 5, 5, 185, 234, 5, 5, 186, 234, 5, 5, 187, 234, 5, 5, 188, 234, 5, 5, 189, 234, 5, 5, 190, 234, 5, 5, 191, 234, 5, 5, 192, 234, 5, 5, 193, 234, 5, 5, 194, 234, 5, 5, 195, 234, 5, 5, 196, 234, 5, 5, 197, 234, 5, 5, 111, 238, 5, 5, 112, 238, 5, 5, 113, 238, 5, 5, 114, 238, 5, 5, 115, 238, 5, 5, 198, 234, 5, 5, 116, 238, 5, 5, 117, 238, 5, 5, 118, 238, 5, 5, 119, 238, 5, 5, 120, 238, 5, 5, 121, 238, 5, 5, 122, 238, 5, 5, 123, 238, 5, 5, 124, 238, 5, 5, 125, 238, 5, 5, 154, 241, 5, 5, 126, 238, 5, 5, 127, 238, 5, 5, 128, 238, 5, 5, 129, 238, 5, 5, 130, 238, 5, 5, 131, 238, 5, 5, 132, 238, 5, 5, 133, 238, 5, 5, 202, 234, 5, 5, 134, 238, 5, 5, 135, 238, 5, 5, 136, 238, 5, 5, 137, 238, 5, 5, 138, 238, 5, 5, 155, 241, 5, 5, 156, 241, 5, 5, 157, 241, 5, 5, 158, 241, 5, 5, 159, 241, 5, 5, 160, 241, 5, 5, 161, 241, 5, 5, 162, 241, 5, 5, 163, 241, 5, 5, 164, 241, 5, 5, 165, 241, 5, 5, 166, 241, 5, 5, 167, 241, 5, 5, 168, 241, 5, 5, 169, 241, 5, 5, 170, 241, 5, 5, 171, 241, 5, 5, 172, 241, 5, 5, 173, 241, 5, 5, 174, 241, 5, 5, 175, 241, 5, 5, 176, 241, 5, 5, 177, 241, 5, 5, 178, 241, 5, 5, 179, 241, 5, 5, 180, 241, 5, 5, 181, 241, 5, 5, 182, 241, 5, 5, 183, 241, 5, 5, 184, 241, 5, 5, 185, 241, 5, 5, 186, 241, 5, 5, 187, 241, 5, 5, 188, 241, 5, 5, 37, 244, 5, 5, 38, 244, 5, 5, 39, 244, 5, 5, 40, 244, 5, 5, 192, 241, 5, 5, 36, 244, 5, 5, 41, 244, 5, 5, 139, 238, 5, 5, 42, 244, 5, 5, 43, 244, 5, 5, 44, 244, 5, 5, 45, 244, 5, 5, 46, 244, 5, 5, 47, 244, 5, 5, 48, 244, 5, 5, 49, 244, 5, 5, 50, 244, 5, 5, 51, 244, 5, 5, 52, 244, 5, 5, 53, 244, 5, 5, 54, 244, 5, 5, 55, 244, 5, 5, 56, 244, 5, 5, 57, 244, 5, 5, 58, 244, 5, 5, 59, 244, 5, 5, 60, 244, 5, 5, 61, 244, 5, 5, 62, 244, 5, 5, 63, 244, 5, 5, 64, 244, 5, 5, 65, 244, 5, 5, 66, 244, 5, 5, 67, 244, 5, 5, 68, 244, 5, 5, 69, 244, 5, 5, 70, 244, 5, 5, 71, 244, 5, 5, 249, 245, 5, 5, 250, 245, 5, 5, 251, 245, 5, 5, 252, 245, 5, 5, 72, 244, 5, 5, 253, 245, 5, 5, 254, 245, 5, 5, 255, 245, 5, 5, 2, 246, 5, 5, 3, 246, 5, 5, 4, 246, 5, 5, 5, 246, 5, 5, 6, 246, 5, 5, 7, 246, 5, 5, 8, 246, 5, 5, 9, 246, 5, 5, 10, 246, 5, 5, 11, 246, 5, 5, 12, 246, 5, 5, 13, 246, 5, 5, 14, 246, 5, 5, 102, 247, 5, 5, 103, 247, 5, 5, 104, 247, 5, 5, 105, 247, 5, 5, 106, 247, 5, 5, 107, 247, 5, 5, 108, 247, 5, 5, 109, 247, 5, 5, 110, 247, 5, 5, 111, 247, 5, 5, 112, 247, 5, 5, 113, 247, 5, 5, 114, 247, 5, 5, 115, 247, 5, 5, 116, 247, 5, 5, 75, 244, 5, 5, 117, 247, 5, 5, 118, 247, 5, 5, 119, 247, 5, 5, 120, 247, 5, 5, 121, 247, 5, 5, 122, 247, 5, 5, 123, 247, 5, 5, 124, 247, 5, 5, 130, 247, 5, 5, 111, 248, 5, 5, 112, 248, 5, 5, 113, 248, 5, 5, 114, 248, 5, 5, 115, 248, 5, 5, 116, 248, 5, 5, 117, 248, 5, 5, 118, 248, 5, 5, 119, 248, 5, 5, 120, 248, 5, 5, 121, 248, 5, 5, 122, 248, 5, 5, 123, 248, 5, 5, 124, 248, 5, 5, 125, 248, 5, 5, 126, 248, 5, 5, 127, 248, 5, 5, 128, 248, 5, 5, 51, 249, 5, 5, 52, 249, 5, 5, 53, 249, 5, 5, 54, 249, 5, 5, 55, 249, 5, 5, 56, 249, 5, 5, 57, 249, 5, 5, 58, 249, 5, 5, 59, 249, 5, 5, 60, 249, 5, 5, 61, 249, 5, 5, 62, 249, 5, 5, 190, 249, 5, 5, 63, 249, 5, 5, 130, 248, 5, 5, 64, 249, 5, 5, 191, 249, 5, 5, 192, 249, 5, 5, 193, 249, 5, 5, 194, 249, 5, 5, 66, 249, 5, 5, 12, 250, 5, 5, 13, 250, 5, 5, 14, 250, 5, 5, 15, 250, 5, 5, 16, 250, 5, 5, 17, 250, 5, 5, 77, 250, 5, 5, 78, 250, 5, 5, 132, 250, 5, 5, 147, 250, 5, 5, 166, 250, 5, 5, 137, 171, 5, 5, 64, 185, 5, 5, 90, 193, 5, 5, 216, 201, 5, 5, 217, 201, 5, 5, 218, 201, 5, 5, 219, 201, 5, 5, 220, 201, 5, 5, 162, 209, 5, 5, 163, 209, 5, 5, 164, 209, 5, 5, 165, 209, 5, 5, 166, 209, 5, 5, 167, 209, 5, 5, 168, 209, 5, 5, 169, 209, 5, 5, 170, 209, 5, 5, 171, 209, 5, 5, 172, 209, 5, 5, 173, 209, 5, 5, 174, 209, 5, 5, 213, 216, 5, 5, 214, 216, 5, 5, 215, 216, 5, 5, 216, 216, 5, 5, 217, 216, 5, 5, 218, 216, 5, 5, 219, 216, 5, 5, 220, 216, 5, 5, 221, 216, 5, 5, 222, 216, 5, 5, 223, 216, 5, 5, 224, 216, 5, 5, 225, 216, 5, 5, 226, 216, 5, 5, 227, 216, 5, 5, 229, 223, 5, 5, 230, 223, 5, 5, 231, 223, 5, 5, 232, 223, 5, 5, 233, 223, 5, 5, 234, 223, 5, 5, 235, 223, 5, 5, 236, 223, 5, 5, 237, 223, 5, 5, 238, 223, 5, 5, 239, 223, 5, 5, 240, 223, 5, 5, 241, 223, 5, 5, 231, 229, 5, 5, 232, 229, 5, 5, 233, 229, 5, 5, 234, 229, 5, 5, 235, 229, 5, 5, 236, 229, 5, 5, 237, 229, 5, 5, 238, 229, 5, 5, 239, 229, 5, 5, 240, 229, 5, 5, 241, 229, 5, 5, 242, 229, 5, 5, 243, 229, 5, 5, 244, 229, 5, 5, 245, 229, 5, 5, 203, 234, 5, 5, 204, 234, 5, 5, 218, 234, 5, 5, 205, 234, 5, 5, 206, 234, 5, 5, 207, 234, 5, 5, 208, 234, 5, 5, 209, 234, 5, 5, 210, 234, 5, 5, 211, 234, 5, 5, 212, 234, 5, 5, 213, 234, 5, 5, 214, 234, 5, 5, 215, 234, 5, 5, 216, 234, 5, 5, 217, 234, 5, 5, 142, 238, 5, 5, 143, 238, 5, 5, 144, 238, 5, 5, 145, 238, 5, 5, 146, 238, 5, 5, 147, 238, 5, 5, 148, 238, 5, 5, 193, 241, 5, 5, 194, 241, 5, 5, 195, 241, 5, 5, 196, 241, 5, 5, 197, 241, 5, 5, 198, 241, 5, 5, 199, 241, 5, 5, 200, 241, 5, 5, 201, 241, 5, 5, 76, 244, 5, 5, 77, 244, 5, 5, 78, 244, 5, 5, 79, 244, 5, 5, 21, 246, 5, 5, 22, 246, 5, 5, 23, 246, 5, 5, 24, 246, 5, 5, 131, 247, 5, 5, 91, 193, 5, 5, 221, 201, 5, 5, 175, 209, 5, 5, 176, 209, 5, 5, 177, 209, 5, 5, 178, 209, 5, 5, 179, 209, 5, 5, 232, 177, 5, 5, 180, 209, 5, 5, 181, 209, 5, 5, 182, 209, 5, 5, 183, 209, 5, 5, 228, 216, 5, 5, 229, 216, 5, 5, 230, 216, 5, 5, 231, 216, 5, 5, 232, 216, 5, 5, 233, 216, 5, 5, 242, 223, 5, 5, 243, 223, 5, 5, 244, 223, 5, 5, 245, 223, 5, 5, 246, 223, 5, 5, 247, 223, 5, 5, 248, 223, 5, 5, 249, 223, 5, 5, 250, 223, 5, 5, 251, 223, 5, 5, 252, 223, 5, 5, 253, 223, 5, 5, 254, 223, 5, 5, 255, 223, 5, 5, 2, 224, 5, 5, 3, 224, 5, 5, 4, 224, 5, 5, 5, 224, 5, 5, 6, 224, 5, 5, 246, 229, 5, 5, 7, 224, 5, 5, 8, 224, 5, 5, 9, 224, 5, 5, 10, 224, 5, 5, 247, 229, 5, 5, 248, 229, 5, 5, 249, 229, 5, 5, 250, 229, 5, 5, 251, 229, 5, 5, 252, 229, 5, 5, 253, 229, 5, 5, 254, 229, 5, 5, 255, 229, 5, 5, 2, 230, 5, 5, 3, 230, 5, 5, 4, 230, 5, 5, 5, 230, 5, 5, 219, 234, 5, 5, 6, 230, 5, 5, 7, 230, 5, 5, 8, 230, 5, 5, 9, 230, 5, 5, 10, 230, 5, 5, 11, 230, 5, 5, 12, 230, 5, 5, 13, 230, 5, 5, 14, 230, 5, 5, 15, 230, 5, 5, 16, 230, 5, 5, 17, 230, 5, 5, 18, 230, 5, 5, 19, 230, 5, 5, 20, 230, 5, 5, 21, 230, 5, 5, 220, 234, 5, 5, 221, 234, 5, 5, 222, 234, 5, 5, 223, 234, 5, 5, 224, 234, 5, 5, 225, 234, 5, 5, 226, 234, 5, 5, 227, 234, 5, 5, 228, 234, 5, 5, 229, 234, 5, 5, 230, 234, 5, 5, 231, 234, 5, 5, 232, 234, 5, 5, 233, 234, 5, 5, 234, 234, 5, 5, 235, 234, 5, 5, 236, 234, 5, 5, 237, 234, 5, 5, 238, 234, 5, 5, 239, 234, 5, 5, 240, 234, 5, 5, 241, 234, 5, 5, 242, 234, 5, 5, 243, 234, 5, 5, 244, 234, 5, 5, 245, 234, 5, 5, 246, 234, 5, 5, 247, 234, 5, 5, 248, 234, 5, 5, 149, 238, 5, 5, 150, 238, 5, 5, 151, 238, 5, 5, 152, 238, 5, 5, 153, 238, 5, 5, 154, 238, 5, 5, 155, 238, 5, 5, 156, 238, 5, 5, 157, 238, 5, 5, 158, 238, 5, 5, 159, 238, 5, 5, 160, 238, 5, 5, 161, 238, 5, 5, 162, 238, 5, 5, 163, 238, 5, 5, 164, 238, 5, 5, 165, 238, 5, 5, 166, 238, 5, 5, 167, 238, 5, 5, 168, 238, 5, 5, 169, 238, 5, 5, 170, 238, 5, 5, 171, 238, 5, 5, 202, 241, 5, 5, 172, 238, 5, 5, 173, 238, 5, 5, 58, 237, 5, 5, 174, 238, 5, 5, 203, 241, 5, 5, 252, 234, 5, 5, 204, 241, 5, 5, 205, 241, 5, 5, 206, 241, 5, 5, 207, 241, 5, 5, 208, 241, 5, 5, 209, 241, 5, 5, 210, 241, 5, 5, 211, 241, 5, 5, 212, 241, 5, 5, 213, 241, 5, 5, 214, 241, 5, 5, 215, 241, 5, 5, 216, 241, 5, 5, 217, 241, 5, 5, 218, 241, 5, 5, 219, 241, 5, 5, 220, 241, 5, 5, 221, 241, 5, 5, 222, 241, 5, 5, 223, 241, 5, 5, 224, 241, 5, 5, 225, 241, 5, 5, 226, 241, 5, 5, 227, 241, 5, 5, 228, 241, 5, 5, 229, 241, 5, 5, 230, 241, 5, 5, 231, 241, 5, 5, 232, 241, 5, 5, 233, 241, 5, 5, 234, 241, 5, 5, 235, 241, 5, 5, 236, 241, 5, 5, 237, 241, 5, 5, 238, 241, 5, 5, 239, 241, 5, 5, 240, 241, 5, 5, 241, 241, 5, 5, 242, 241, 5, 5, 243, 241, 5, 5, 80, 244, 5, 5, 244, 241, 5, 5, 81, 244, 5, 5, 82, 244, 5, 5, 83, 244, 5, 5, 84, 244, 5, 5, 85, 244, 5, 5, 86, 244, 5, 5, 87, 244, 5, 5, 88, 244, 5, 5, 89, 244, 5, 5, 90, 244, 5, 5, 91, 244, 5, 5, 92, 244, 5, 5, 93, 244, 5, 5, 94, 244, 5, 5, 95, 244, 5, 5, 96, 244, 5, 5, 97, 244, 5, 5, 98, 244, 5, 5, 99, 244, 5, 5, 100, 244, 5, 5, 101, 244, 5, 5, 102, 244, 5, 5, 103, 244, 5, 5, 104, 244, 5, 5, 105, 244, 5, 5, 106, 244, 5, 5, 25, 246, 5, 5, 26, 246, 5, 5, 27, 246, 5, 5, 28, 246, 5, 5, 29, 246, 5, 5, 30, 246, 5, 5, 31, 246, 5, 5, 32, 246, 5, 5, 33, 246, 5, 5, 34, 246, 5, 5, 35, 246, 5, 5, 36, 246, 5, 5, 37, 246, 5, 5, 38, 246, 5, 5, 39, 246, 5, 5, 40, 246, 5, 5, 41, 246, 5, 5, 42, 246, 5, 5, 43, 246, 5, 5, 113, 244, 5, 5, 44, 246, 5, 5, 45, 246, 5, 5, 46, 246, 5, 5, 47, 246, 5, 5, 48, 246, 5, 5, 49, 246, 5, 5, 50, 246, 5, 5, 51, 246, 5, 5, 52, 246, 5, 5, 53, 246, 5, 5, 54, 246, 5, 5, 132, 247, 5, 5, 55, 246, 5, 5, 56, 246, 5, 5, 57, 246, 5, 5, 58, 246, 5, 5, 133, 247, 5, 5, 134, 247, 5, 5, 135, 247, 5, 5, 136, 247, 5, 5, 137, 247, 5, 5, 138, 247, 5, 5, 139, 247, 5, 5, 140, 247, 5, 5, 141, 247, 5, 5, 142, 247, 5, 5, 143, 247, 5, 5, 144, 247, 5, 5, 145, 247, 5, 5, 146, 247, 5, 5, 147, 247, 5, 5, 148, 247, 5, 5, 125, 247, 5, 5, 131, 248, 5, 5, 132, 248, 5, 5, 133, 248, 5, 5, 134, 248, 5, 5, 135, 248, 5, 5, 136, 248, 5, 5, 137, 248, 5, 5, 138, 248, 5, 5, 152, 247, 5, 5, 139, 248, 5, 5, 140, 248, 5, 5, 141, 248, 5, 5, 142, 248, 5, 5, 143, 248, 5, 5, 144, 248, 5, 5, 145, 248, 5, 5, 146, 248, 5, 5, 147, 248, 5, 5, 148, 248, 5, 5, 149, 248, 5, 5, 153, 247, 5, 5, 150, 248, 5, 5, 151, 248, 5, 5, 152, 248, 5, 5, 68, 249, 5, 5, 67, 249, 5, 5, 153, 248, 5, 5, 154, 248, 5, 5, 69, 249, 5, 5, 70, 249, 5, 5, 71, 249, 5, 5, 72, 249, 5, 5, 73, 249, 5, 5, 74, 249, 5, 5, 75, 249, 5, 5, 76, 249, 5, 5, 77, 249, 5, 5, 78, 249, 5, 5, 79, 249, 5, 5, 80, 249, 5, 5, 81, 249, 5, 5, 82, 249, 5, 5, 196, 249, 5, 5, 197, 249, 5, 5, 198, 249, 5, 5, 199, 249, 5, 5, 200, 249, 5, 5, 201, 249, 5, 5, 202, 249, 5, 5, 203, 249, 5, 5, 18, 250, 5, 5, 19, 250, 5, 5, 79, 250, 5, 5, 80, 250, 5, 5, 81, 250, 5, 5, 110, 250, 5, 5, 111, 250, 5, 5, 112, 250, 5, 5, 133, 250, 5, 5, 134, 250, 5, 5, 148, 250, 5, 5, 149, 250, 5, 5, 211, 159, 5, 5, 249, 165, 5, 5, 250, 165, 5, 5, 138, 171, 5, 5, 139, 171, 5, 5, 140, 171, 5, 5, 233, 177, 5, 5, 234, 177, 5, 5, 235, 177, 5, 5, 236, 177, 5, 5, 237, 177, 5, 5, 65, 185, 5, 5, 66, 185, 5, 5, 67, 185, 5, 5, 68, 185, 5, 5, 69, 185, 5, 5, 70, 185, 5, 5, 71, 185, 5, 5, 72, 185, 5, 5, 73, 185, 5, 5, 74, 185, 5, 5, 75, 185, 5, 5, 76, 185, 5, 5, 77, 185, 5, 5, 92, 193, 5, 5, 93, 193, 5, 5, 94, 193, 5, 5, 95, 193, 5, 5, 96, 193, 5, 5, 97, 193, 5, 5, 98, 193, 5, 5, 99, 193, 5, 5, 100, 193, 5, 5, 222, 201, 5, 5, 223, 201, 5, 5, 224, 201, 5, 5, 225, 201, 5, 5, 226, 201, 5, 5, 227, 201, 5, 5, 228, 201, 5, 5, 229, 201, 5, 5, 230, 201, 5, 5, 184, 209, 5, 5, 185, 209, 5, 5, 186, 209, 5, 5, 187, 209, 5, 5, 188, 209, 5, 5, 189, 209, 5, 5, 190, 209, 5, 5, 191, 209, 5, 5, 192, 209, 5, 5, 193, 209, 5, 5, 194, 209, 5, 5, 195, 209, 5, 5, 234, 216, 5, 5, 235, 216, 5, 5, 236, 216, 5, 5, 15, 224, 5, 5, 237, 216, 5, 5, 238, 216, 5, 5, 239, 216, 5, 5, 240, 216, 5, 5, 16, 224, 5, 5, 17, 224, 5, 5, 18, 224, 5, 5, 19, 224, 5, 5, 20, 224, 5, 5, 21, 224, 5, 5, 22, 224, 5, 5, 23, 224, 5, 5, 23, 230, 5, 5, 24, 230, 5, 5, 25, 230, 5, 5, 26, 230, 5, 5, 253, 234, 5, 5, 254, 234, 5, 5, 255, 234, 5, 5, 2, 235, 5, 5, 175, 238, 5, 5, 176, 238, 5, 5, 177, 238, 5, 5, 178, 238, 5, 5, 248, 241, 5, 5, 249, 241, 5, 5, 154, 247, 5, 5, 155, 247, 5, 5, 101, 193, 5, 5, 24, 224, 5, 5, 27, 230, 5, 5, 250, 241, 5, 5, 114, 244, 5, 5, 62, 246, 5, 5, 63, 246, 5, 5, 87, 249, 5, 5, 88, 249, 5, 5, 28, 230, 5, 5, 102, 193, 5, 5, 196, 209, 5, 5, 197, 209, 5, 5, 198, 209, 5, 5, 25, 224, 5, 5, 26, 224, 5, 5, 29, 230, 5, 5, 30, 230, 5, 5, 31, 230, 5, 5, 32, 230, 5, 5, 3, 235, 5, 5, 4, 235, 5, 5, 5, 235, 5, 5, 179, 238, 5, 5, 180, 238, 5, 5, 181, 238, 5, 5, 182, 238, 5, 5, 183, 238, 5, 5, 251, 241, 5, 5, 252, 241, 5, 5, 253, 241, 5, 5, 254, 241, 5, 5, 255, 241, 5, 5, 2, 242, 5, 5, 3, 242, 5, 5, 115, 244, 5, 5, 116, 244, 5, 5, 117, 244, 5, 5, 118, 244, 5, 5, 64, 246, 5, 5, 65, 246, 5, 5, 156, 247, 5, 5, 159, 248, 5, 5, 89, 249, 5, 5, 205, 249, 5, 5, 114, 250, 5, 5, 157, 250, 5, 5, 167, 250, 5, 5, 103, 193, 5, 5, 251, 165, 5, 5, 241, 216, 5, 5, 28, 224, 5, 5, 29, 224, 5, 5, 30, 224, 5, 5, 31, 224, 5, 5, 34, 230, 5, 5, 35, 230, 5, 5, 36, 230, 5, 5, 7, 235, 5, 5, 8, 235, 5, 5, 189, 238, 5, 5, 190, 238, 5, 5, 4, 242, 5, 5, 5, 242, 5, 5, 119, 244, 5, 5, 158, 247, 5, 5, 135, 250, 5, 5, 104, 193, 5, 5, 32, 224, 5, 5, 38, 230, 5, 5, 105, 193, 5, 5, 242, 216, 5, 5, 243, 216, 5, 5, 33, 224, 5, 5, 192, 238, 5, 5, 6, 242, 5, 5, 121, 244, 5, 5, 160, 248, 5, 5, 231, 201, 5, 5, 232, 201, 5, 5, 39, 230, 5, 5, 40, 230, 5, 5, 9, 235, 5, 5, 10, 235, 5, 5, 11, 235, 5, 5, 193, 238, 5, 5, 194, 238, 5, 5, 206, 249, 5, 5, 233, 201, 5, 5, 34, 224, 5, 5, 12, 235, 5, 5, 159, 247, 5, 5, 234, 201, 5, 5, 106, 193, 5, 5, 36, 224, 5, 5, 41, 230, 5, 5, 42, 230, 5, 5, 43, 230, 5, 5, 44, 230, 5, 5, 45, 230, 5, 5, 165, 219, 5, 5, 13, 235, 5, 5, 14, 235, 5, 5, 15, 235, 5, 5, 16, 235, 5, 5, 17, 235, 5, 5, 195, 238, 5, 5, 196, 238, 5, 5, 197, 238, 5, 5, 8, 242, 5, 5, 9, 242, 5, 5, 124, 244, 5, 5, 125, 244, 5, 5, 126, 244, 5, 5, 127, 244, 5, 5, 128, 244, 5, 5, 129, 244, 5, 5, 130, 244, 5, 5, 67, 246, 5, 5, 68, 246, 5, 5, 69, 246, 5, 5, 70, 246, 5, 5, 71, 246, 5, 5, 160, 247, 5, 5, 161, 247, 5, 5, 161, 248, 5, 5, 162, 248, 5, 5, 163, 248, 5, 5, 207, 249, 5, 5, 21, 250, 5, 5, 82, 250, 5, 5, 115, 250, 5, 5, 235, 201, 5, 5, 46, 230, 5, 5, 19, 235, 5, 5, 10, 242, 5, 5, 200, 209, 5, 5, 141, 171, 5, 5, 20, 235, 5, 5, 198, 238, 5, 5, 199, 238, 5, 5, 200, 238, 5, 5, 11, 242, 5, 5, 12, 242, 5, 5, 72, 246, 5, 5, 164, 248, 5, 5, 165, 248, 5, 5, 208, 249, 5, 5, 209, 249, 5, 5, 22, 250, 5, 5, 236, 201, 5, 5, 201, 209, 5, 5, 132, 244, 5, 5, 202, 209, 5, 5, 37, 224, 5, 5, 38, 224, 5, 5, 39, 224, 5, 5, 47, 230, 5, 5, 203, 209, 5, 5, 204, 209, 5, 5, 202, 238, 5, 5, 203, 238, 5, 5, 14, 242, 5, 5, 73, 246, 5, 5, 74, 246, 5, 5, 75, 246, 5, 5, 76, 246, 5, 5, 166, 248, 5, 5, 210, 249, 5, 5, 90, 249, 5, 5, 211, 249, 5, 5, 205, 209, 5, 5, 142, 171, 5, 5, 21, 235, 5, 5, 22, 235, 5, 5, 23, 235, 5, 5, 204, 238, 5, 5, 205, 238, 5, 5, 206, 238, 5, 5, 207, 238, 5, 5, 208, 238, 5, 5, 209, 238, 5, 5, 210, 238, 5, 5, 211, 238, 5, 5, 15, 242, 5, 5, 133, 244, 5, 5, 134, 244, 5, 5, 135, 244, 5, 5, 77, 246, 5, 5, 162, 247, 5, 5, 163, 247, 5, 5, 164, 247, 5, 5, 165, 247, 5, 5, 167, 248, 5, 5, 168, 248, 5, 5, 169, 248, 5, 5, 170, 248, 5, 5, 116, 250, 5, 5, 244, 216, 5, 5, 48, 230, 5, 5, 49, 230, 5, 5, 24, 235, 5, 5, 25, 235, 5, 5, 16, 242, 5, 5, 17, 242, 5, 5, 166, 247, 5, 5, 172, 248, 5, 5, 173, 248, 5, 5, 91, 249, 5, 5, 92, 249, 5, 5, 212, 249, 5, 5, 83, 250, 5, 5, 172, 250, 5, 5, 245, 216, 5, 5, 26, 235, 5, 5, 212, 238, 5, 5, 18, 242, 5, 5, 78, 246, 5, 5, 174, 248, 5, 5, 31, 162, 5, 5, 40, 224, 5, 5, 41, 224, 5, 5, 50, 230, 5, 5, 28, 235, 5, 5, 213, 238, 5, 5, 19, 242, 5, 5, 20, 242, 5, 5, 21, 242, 5, 5, 137, 244, 5, 5, 138, 244, 5, 5, 139, 244, 5, 5, 79, 246, 5, 5, 140, 244, 5, 5, 141, 244, 5, 5, 142, 244, 5, 5, 143, 244, 5, 5, 144, 244, 5, 5, 29, 235, 5, 5, 145, 244, 5, 5, 80, 246, 5, 5, 81, 246, 5, 5, 82, 246, 5, 5, 83, 246, 5, 5, 84, 246, 5, 5, 85, 246, 5, 5, 167, 247, 5, 5, 168, 247, 5, 5, 169, 247, 5, 5, 175, 248, 5, 5, 176, 248, 5, 5, 177, 248, 5, 5, 178, 248, 5, 5, 179, 248, 5, 5, 93, 249, 5, 5, 94, 249, 5, 5, 95, 249, 5, 5, 96, 249, 5, 5, 97, 249, 5, 5, 98, 249, 5, 5, 213, 249, 5, 5, 214, 249, 5, 5, 215, 249, 5, 5, 216, 249, 5, 5, 117, 250, 5, 5, 118, 250, 5, 5, 170, 250, 5, 5, 143, 171, 5, 5, 78, 185, 5, 5, 107, 193, 5, 5, 237, 201, 5, 5, 206, 209, 5, 5, 207, 209, 5, 5, 208, 209, 5, 5, 209, 209, 5, 5, 246, 216, 5, 5, 247, 216, 5, 5, 42, 224, 5, 5, 43, 224, 5, 5, 30, 235, 5, 5, 31, 235, 5, 5, 51, 230, 5, 5, 214, 238, 5, 5, 22, 242, 5, 5, 24, 239, 5, 5, 146, 244, 5, 5, 87, 246, 5, 5, 171, 247, 5, 5, 172, 247, 5, 5, 173, 247, 5, 5, 162, 250, 5, 5, 168, 250, 5, 5, 176, 250, 5, 5, 212, 159, 5, 5, 108, 193, 5, 5, 109, 193, 5, 5, 52, 230, 5, 5, 88, 246, 5, 5, 119, 250, 5, 5, 252, 165, 5, 5, 32, 235, 5, 5, 89, 246, 5, 5, 174, 247, 5, 5, 218, 249, 5, 5, 24, 250, 5, 5, 25, 250, 5, 5, 166, 219, 5, 5, 86, 225, 5, 5, 143, 178, 5, 5, 3, 233, 5, 5, 78, 177, 5, 5, 212, 200, 5, 5, 23, 229, 5, 5, 234, 245, 5, 5, 73, 250, 5, 5, 3, 209, 5, 5, 77, 158, 5, 5, 27, 159, 5, 5, 215, 237, 5, 5, 144, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 166, 5, 5, 221, 163, 5, 5, 222, 163, 5, 5, 146, 191, 5, 5, 80, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 184, 5, 5, 94, 164, 5, 5, 160, 165, 5, 5, 42, 208, 5, 5, 213, 204, 5, 5, 38, 162, 5, 5, 247, 158, 5, 5, 52, 230, 5, 5, 52, 230, 5, 5, 175, 172, 5, 5, 71, 171, 5, 5, 211, 193, 5, 5, 79, 167, 5, 5, 31, 239, 5, 5, 249, 244, 5, 5, 24, 240, 5, 5, 15, 248, 5, 5, 28, 233, 5, 5, 48, 215, 5, 5, 54, 248, 5, 5, 218, 218, 5, 5, 216, 174, 5, 5, 164, 181, 5, 5, 232, 181, 5, 5, 41, 207, 5, 5, 176, 208, 5, 5, 173, 229, 5, 5, 239, 201, 5, 5, 173, 162, 5, 5, 214, 244, 5, 5, 236, 244, 5, 5, 72, 245, 5, 5, 149, 250, 5, 5, 233, 194, 5, 5, 94, 231, 5, 5, 241, 236, 5, 5, 66, 243, 5, 5, 129, 168, 5, 5, 37, 240, 5, 5, 84, 245, 5, 5, 23, 195, 5, 5, 161, 180, 5, 5, 78, 181, 5, 5, 208, 181, 5, 5, 129, 177, 5, 5, 49, 166, 5, 5, 122, 162, 5, 5, 183, 193, 5, 5, 20, 225, 5, 5, 70, 239, 5, 5, 148, 242, 5, 5, 161, 226, 5, 5, 190, 161, 5, 5, 31, 243, 5, 5, 138, 207, 5, 5, 85, 208, 5, 5, 184, 245, 5, 5, 215, 223, 5, 5, 67, 249, 5, 5, 4, 206, 5, 5, 30, 206, 5, 5, 236, 213, 5, 5, 98, 199, 5, 5, 210, 228, 5, 5, 102, 193, 5, 5, 47, 222, 5, 5, 247, 238, 5, 5, 205, 163, 5, 5, 198, 246, 5, 5, 214, 246, 5, 5, 238, 164, 5, 5, 58, 220, 5, 5, 36, 208, 5, 5, 70, 209, 5, 5, 74, 235, 5, 5, 228, 210, 5, 5, 234, 218, 5, 5, 9, 189, 5, 5, 97, 212, 5, 5, 164, 190, 5, 5, 116, 232, 5, 5, 179, 177, 5, 5, 174, 185, 5, 5, 198, 161, 5, 5, 30, 217, 5, 5, 104, 178, 5, 5, 53, 206, 5, 5, 12, 214, 5, 5, 137, 199, 5, 5, 45, 193, 5, 5, 250, 246, 5, 5, 174, 173, 5, 5, 218, 218, 5, 5, 77, 228, 5, 5, 175, 157, 5, 5, 224, 210, 5, 5, 109, 173, 5, 5, 169, 189, 5, 5, 227, 189, 5, 5, 221, 158, 5, 5, 252, 231, 5, 5, 176, 171, 5, 5, 38, 195, 5, 5, 164, 157, 5, 5, 112, 169, 5, 5, 151, 218, 5, 5, 20, 183, 5, 5, 204, 185, 5, 5, 172, 202, 5, 5, 209, 175, 5, 5, 53, 207, 5, 5, 101, 215, 5, 5, 203, 188, 5, 5, 165, 165, 5, 5, 188, 164, 5, 5, 195, 173, 5, 5, 200, 176, 5, 5, 227, 187, 5, 5, 223, 189, 5, 5, 154, 171, 5, 5, 102, 166, 5, 5, 101, 178, 5, 5, 104, 188, 5, 5, 129, 236, 5, 5, 98, 165, 5, 5, 43, 222, 5, 5, 23, 201, 5, 5, 76, 230, 5, 5, 222, 162, 5, 5, 109, 157, 5, 5, 22, 239, 5, 5, 121, 180, 5, 5, 208, 235, 5, 5, 201, 242, 5, 5, 67, 223, 5, 5, 130, 250, 5, 5, 3, 242, 5, 5, 34, 224, 5, 5, 32, 157, 5, 5, 78, 225, 5, 5, 197, 225, 5, 5, 21, 247, 5, 5, 6, 161, 5, 5, 48, 218, 5, 5, 193, 247, 5, 5, 111, 218, 5, 5, 115, 212, 5, 5, 2, 205, 5, 5, 220, 219, 5, 5, 75, 170, 5, 5, 217, 220, 5, 5, 159, 232, 5, 5, 170, 222, 5, 5, 70, 221, 5, 5, 160, 192, 5, 5, 209, 233, 5, 5, 94, 160, 5, 5, 99, 160, 5, 5, 68, 172, 5, 5, 147, 181, 5, 5, 20, 200, 5, 5, 101, 215, 5, 5, 55, 203, 5, 5, 34, 168, 5, 5, 192, 187, 5, 5, 64, 231, 5, 5, 237, 239, 5, 5, 252, 235, 5, 5, 164, 158, 5, 5, 13, 167, 5, 5, 224, 210, 5, 5, 165, 230, 5, 5, 57, 168, 5, 5, 104, 175, 5, 5, 203, 219, 5, 5, 227, 190, 5, 5, 2, 191, 5, 5, 191, 208, 5, 5, 69, 209, 5, 5, 30, 249, 5, 5, 128, 216, 5, 5, 54, 166, 5, 5, 82, 236, 5, 5, 133, 243, 5, 5, 66, 234, 5, 5, 55, 195, 5, 5, 12, 157, 5, 5, 225, 209, 5, 5, 205, 217, 5, 5, 142, 163, 5, 5, 115, 180, 5, 5, 218, 218, 5, 5, 48, 226, 5, 5, 193, 231, 5, 5, 83, 221, 5, 5, 171, 228, 5, 5, 51, 230, 5, 5, 232, 203, 5, 5, 225, 165, 5, 5, 36, 217, 5, 5, 239, 168, 5, 5, 119, 174, 5, 5, 251, 174, 5, 5, 163, 204, 5, 5, 253, 181, 5, 5, 21, 182, 5, 5, 62, 198, 5, 5, 2, 183, 5, 5, 97, 241, 5, 5, 218, 157, 5, 5, 81, 218, 5, 5, 48, 193, 5, 5, 51, 178, 5, 5, 8, 187, 5, 5, 25, 189, 5, 5, 174, 222, 5, 5, 104, 173, 5, 5, 111, 203, 5, 5, 170, 180, 5, 5, 169, 189, 5, 5, 136, 201, 5, 5, 142, 162, 5, 5, 142, 160, 5, 5, 211, 217, 5, 5, 209, 168, 5, 5, 106, 164, 5, 5, 143, 188, 5, 5, 136, 169, 5, 5, 188, 189, 5, 5, 253, 197, 5, 5, 85, 227, 5, 5, 185, 207, 5, 5, 202, 207, 5, 5, 202, 165, 5, 5, 2, 238, 5, 5, 194, 185, 5, 5, 193, 204, 5, 5, 185, 162, 5, 5, 50, 226, 5, 5, 97, 226, 5, 5, 19, 243, 5, 5, 77, 223, 5, 5, 124, 248, 5, 5, 159, 248, 5, 5, 12, 169, 5, 5, 248, 188, 5, 5, 188, 232, 5, 5, 178, 159, 5, 5, 113, 190, 5, 5, 145, 190, 5, 5, 217, 169, 5, 5, 191, 169, 5, 5, 162, 240, 5, 5, 193, 157, 5, 5, 185, 183, 5, 5, 130, 166, 5, 5, 236, 157, 5, 5, 79, 173, 5, 5, 138, 168, 5, 5, 41, 227, 5, 5, 220, 160, 5, 5, 218, 174, 5, 5, 169, 218, 5, 5, 152, 228, 5, 5, 225, 161, 5, 5, 181, 177, 5, 5, 123, 165, 5, 5, 22, 211, 5, 5, 85, 157, 5, 5, 67, 202, 5, 5, 35, 210, 5, 5, 27, 179, 5, 5, 168, 202, 5, 5, 247, 194, 5, 5, 241, 195, 5, 5, 144, 196, 5, 5, 34, 212, 5, 5, 69, 224, 5, 5, 168, 189, 5, 5, 63, 182, 5, 5, 166, 161, 5, 5, 162, 182, 5, 5, 61, 190, 5, 5, 115, 213, 5, 5, 80, 209, 5, 5, 225, 213, 5, 5, 189, 161, 5, 5, 212, 227, 5, 5, 50, 243, 5, 5, 194, 191, 5, 5, 59, 222, 5, 5, 95, 192, 5, 5, 3, 162, 5, 5, 242, 200, 5, 5, 183, 192, 5, 5, 36, 223, 5, 5, 24, 229, 5, 5, 59, 209, 5, 5, 199, 201, 5, 5, 127, 209, 5, 5, 150, 229, 5, 5, 33, 246, 5, 5, 189, 184, 5, 5, 71, 229, 5, 5, 159, 171, 5, 5, 238, 209, 5, 5, 111, 162, 5, 5, 11, 172, 5, 5, 51, 202, 5, 5, 162, 166, 5, 5, 232, 193, 5, 5, 53, 210, 5, 5, 89, 224, 5, 5, 74, 194, 5, 5, 35, 224, 5, 5, 210, 217, 5, 5, 123, 157, 5, 5, 250, 179, 5, 5, 103, 195, 5, 5, 46, 218, 5, 5, 29, 239, 5, 5, 13, 188, 5, 5, 5, 174, 5, 5, 2, 196, 5, 5, 108, 188, 5, 5, 91, 181, 5, 5, 68, 189, 5, 5, 114, 212, 5, 5, 101, 197, 5, 5, 91, 158, 5, 5, 173, 197, 5, 5, 9, 206, 5, 5, 63, 170, 5, 5, 10, 176, 5, 5, 9, 176, 5, 5, 148, 182, 5, 5, 154, 182, 5, 5, 161, 182, 5, 5, 113, 213, 5, 5, 114, 213, 5, 5, 99, 220, 5, 5, 42, 176, 5, 5, 116, 206, 5, 5, 217, 220, 5, 5, 46, 227, 5, 5, 126, 232, 5, 5, 197, 206, 5, 5, 119, 170, 5, 5, 130, 183, 5, 5, 153, 157, 5, 5, 153, 157, 5, 5, 191, 199, 5, 5, 232, 221, 5, 5, 72, 190, 5, 5, 80, 228, 5, 5, 81, 237, 5, 5, 127, 215, 5, 5, 177, 240, 5, 5, 120, 158, 5, 5, 242, 200, 5, 5, 61, 241, 5, 5, 229, 243, 5, 5, 121, 229, 5, 5, 228, 179, 5, 5, 55, 205, 5, 5, 133, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 165, 5, 5, 120, 162, 5, 5, 64, 160, 5, 5, 43, 166, 5, 5, 185, 158, 5, 5, 66, 224, 5, 5, 9, 172, 5, 5, 93, 157, 5, 5, 232, 193, 5, 5, 252, 185, 5, 5, 228, 193, 5, 5, 100, 202, 5, 5, 168, 202, 5, 5, 142, 217, 5, 5, 75, 167, 5, 5, 85, 167, 5, 5, 185, 186, 5, 5, 178, 224, 5, 5, 21, 211, 5, 5, 28, 211, 5, 5, 72, 187, 5, 5, 69, 203, 5, 5, 138, 187, 5, 5, 118, 203, 5, 5, 79, 203, 5, 5, 46, 218, 5, 5, 66, 211, 5, 5, 29, 239, 5, 5, 207, 230, 5, 5, 126, 195, 5, 5, 157, 203, 5, 5, 187, 195, 5, 5, 20, 188, 5, 5, 241, 195, 5, 5, 161, 180, 5, 5, 79, 188, 5, 5, 114, 164, 5, 5, 78, 158, 5, 5, 203, 188, 5, 5, 251, 174, 5, 5, 221, 204, 5, 5, 44, 197, 5, 5, 114, 212, 5, 5, 112, 239, 5, 5, 101, 197, 5, 5, 218, 231, 5, 5, 146, 231, 5, 5, 142, 159, 5, 5, 168, 189, 5, 5, 245, 212, 5, 5, 16, 213, 5, 5, 18, 170, 5, 5, 255, 219, 5, 5, 250, 219, 5, 5, 63, 182, 5, 5, 255, 189, 5, 5, 47, 170, 5, 5, 33, 198, 5, 5, 27, 190, 5, 5, 60, 220, 5, 5, 110, 220, 5, 5, 116, 206, 5, 5, 69, 176, 5, 5, 140, 206, 5, 5, 217, 220, 5, 5, 239, 198, 5, 5, 119, 170, 5, 5, 209, 183, 5, 5, 135, 199, 5, 5, 217, 227, 5, 5, 72, 233, 5, 5, 51, 237, 5, 5, 72, 190, 5, 5, 24, 222, 5, 5, 59, 222, 5, 5, 36, 222, 5, 5, 80, 228, 5, 5, 77, 228, 5, 5, 61, 228, 5, 5, 81, 237, 5, 5, 36, 248, 5, 5, 177, 240, 5, 5, 149, 228, 5, 5, 163, 228, 5, 5, 193, 233, 5, 5, 234, 215, 5, 5, 51, 193, 5, 5, 61, 241, 5, 5, 80, 209, 5, 5, 225, 243, 5, 5, 229, 243, 5, 5, 96, 209, 5, 5, 121, 229, 5, 5, 31, 244, 5, 5, 52, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 248, 5, 5, 214, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 32, 230, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 55, 158, 6, 5, 216, 160, 6, 5, 15, 158, 192, 0, 0, 0, 6, 5, 14, 157, 6, 5, 184, 160, 6, 5, 216, 207, 6, 5, 23, 158, 6, 5, 237, 159, 6, 5, 12, 174, 6, 5, 69, 197, 6, 5, 115, 180, 6, 5, 252, 171, 6, 5, 105, 173, 6, 5, 70, 160, 6, 5, 221, 203, 6, 5, 130, 162, 6, 5, 183, 190, 6, 5, 152, 159, 6, 5, 75, 192, 6, 5, 66, 163, 6, 5, 213, 162, 6, 5, 100, 212, 6, 5, 58, 164, 6, 5, 81, 180, 6, 5, 0, 128, 6, 5, 56, 157, 6, 5, 143, 208, 6, 5, 67, 159, 6, 5, 171, 157, 6, 5, 7, 159, 6, 5, 203, 173, 6, 5, 156, 165, 6, 5, 101, 159, 6, 5, 32, 206, 6, 5, 78, 170, 6, 5, 135, 160, 6, 5, 31, 197, 6, 5, 77, 161, 6, 5, 71, 158, 6, 5, 158, 159, 6, 5, 175, 193, 6, 5, 56, 186, 6, 5, 200, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 3, 63, 0, 197, 99, 63, 0, 197, 195, 63, 0, 197, 35, 64, 0, 197, 131, 64, 0, 197, 227, 64, 0, 197, 67, 65, 0, 197, 163, 65, 0, 197, 3, 66, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 76, 187, 6, 5, 3, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 157, 5, 5, 8, 157, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 138, 5, 5, 0, 139, 5, 5, 0, 140, 5, 5, 0, 141, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 142, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 144, 192, 0, 0, 0, 5, 5, 13, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 158, 192, 0, 0, 0, 5, 5, 183, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 178, 5, 5, 75, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 178, 192, 0, 0, 0, 5, 5, 78, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 185, 5, 5, 149, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 202, 5, 5, 30, 202, 5, 5, 31, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 209, 5, 5, 253, 209, 5, 5, 254, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 217, 192, 0, 0, 0, 5, 5, 25, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 230, 5, 5, 68, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 171, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 178, 5, 5, 83, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 224, 192, 0, 0, 0, 5, 5, 39, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 178, 5, 5, 108, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 230, 5, 5, 73, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 159, 5, 5, 21, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 162, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 166, 192, 0, 0, 0, 5, 5, 2, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 167, 5, 5, 6, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 167, 5, 5, 8, 167, 5, 5, 9, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 172, 5, 5, 107, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 178, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 178, 5, 5, 230, 178, 5, 5, 231, 178, 5, 5, 232, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 186, 5, 5, 37, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 186, 5, 5, 39, 186, 5, 5, 40, 186, 5, 5, 41, 186, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 186, 5, 5, 43, 186, 5, 5, 44, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 186, 192, 0, 0, 0, 5, 5, 46, 186, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 186, 192, 0, 0, 0, 5, 5, 49, 186, 5, 5, 50, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 194, 5, 5, 22, 194, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 194, 5, 5, 24, 194, 5, 5, 25, 194, 5, 5, 26, 194, 5, 5, 27, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 202, 192, 0, 0, 0, 5, 5, 128, 202, 5, 5, 129, 202, 5, 5, 130, 202, 5, 5, 131, 202, 5, 5, 132, 202, 5, 5, 133, 202, 5, 5, 134, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 202, 192, 0, 0, 0, 5, 5, 137, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 202, 5, 5, 141, 202, 5, 5, 142, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 210, 5, 5, 87, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 210, 5, 5, 90, 210, 5, 5, 91, 210, 5, 5, 92, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 210, 5, 5, 96, 210, 192, 0, 0, 0, 5, 5, 97, 210, 5, 5, 98, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 217, 5, 5, 115, 217, 5, 5, 116, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 224, 5, 5, 118, 224, 5, 5, 119, 224, 5, 5, 120, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 224, 5, 5, 122, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 230, 5, 5, 111, 230, 5, 5, 112, 230, 5, 5, 113, 230, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 235, 5, 5, 66, 235, 5, 5, 67, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 235, 5, 5, 69, 235, 5, 5, 70, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 238, 5, 5, 240, 238, 5, 5, 241, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 244, 192, 0, 0, 0, 5, 5, 170, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 167, 5, 5, 68, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 179, 5, 5, 30, 179, 5, 5, 31, 179, 5, 5, 32, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 194, 5, 5, 84, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 202, 5, 5, 194, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 217, 192, 0, 0, 0, 5, 5, 152, 217, 5, 5, 153, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 186, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 160, 192, 0, 0, 0, 5, 5, 214, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 163, 5, 5, 115, 163, 5, 5, 116, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 163, 192, 0, 0, 0, 5, 5, 118, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 167, 5, 5, 137, 167, 5, 5, 138, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 167, 5, 5, 140, 167, 5, 5, 141, 167, 192, 0, 0, 0, 5, 5, 142, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 172, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 179, 5, 5, 95, 179, 5, 5, 96, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 186, 5, 5, 209, 186, 5, 5, 210, 186, 5, 5, 211, 186, 5, 5, 212, 186, 5, 5, 213, 186, 192, 0, 0, 0, 5, 5, 214, 186, 5, 5, 215, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 194, 5, 5, 160, 194, 5, 5, 161, 194, 5, 5, 162, 194, 5, 5, 163, 194, 5, 5, 164, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 194, 5, 5, 168, 194, 5, 5, 169, 194, 5, 5, 170, 194, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 202, 5, 5, 242, 202, 192, 0, 0, 0, 5, 5, 243, 202, 5, 5, 244, 202, 5, 5, 245, 202, 5, 5, 246, 202, 5, 5, 247, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 210, 5, 5, 203, 210, 5, 5, 204, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 210, 5, 5, 207, 210, 5, 5, 208, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 217, 5, 5, 193, 217, 5, 5, 194, 217, 5, 5, 195, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 217, 5, 5, 199, 217, 5, 5, 200, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 224, 5, 5, 172, 224, 5, 5, 173, 224, 5, 5, 174, 224, 5, 5, 175, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 224, 5, 5, 177, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 230, 5, 5, 152, 230, 5, 5, 153, 230, 5, 5, 154, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 239, 5, 5, 6, 239, 5, 5, 7, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 242, 5, 5, 55, 242, 5, 5, 56, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 194, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 157, 5, 5, 121, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 194, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 160, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 179, 192, 0, 0, 0, 5, 5, 184, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 161, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 163, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 161, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 168, 5, 5, 83, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 180, 192, 0, 0, 0, 5, 5, 23, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 187, 5, 5, 157, 187, 5, 5, 158, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 187, 192, 0, 0, 0, 5, 5, 160, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 203, 5, 5, 123, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 218, 5, 5, 78, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 173, 192, 0, 0, 0, 5, 5, 236, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 195, 5, 5, 191, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 211, 5, 5, 163, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 225, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 225, 5, 5, 50, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 195, 5, 5, 222, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 180, 5, 5, 149, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 188, 5, 5, 72, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 196, 5, 5, 8, 196, 5, 5, 9, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 204, 192, 0, 0, 0, 5, 5, 9, 204, 5, 5, 10, 204, 5, 5, 11, 204, 5, 5, 12, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 225, 5, 5, 180, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 225, 5, 5, 98, 225, 5, 5, 99, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 235, 5, 5, 158, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 169, 5, 5, 48, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 169, 5, 5, 50, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 169, 192, 0, 0, 0, 5, 5, 52, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 181, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 212, 5, 5, 38, 212, 5, 5, 39, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 231, 192, 0, 0, 0, 5, 5, 55, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 242, 5, 5, 118, 242, 5, 5, 119, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 164, 5, 5, 220, 164, 5, 5, 221, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 169, 192, 0, 0, 0, 5, 5, 162, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 169, 5, 5, 164, 169, 5, 5, 165, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 175, 5, 5, 20, 175, 5, 5, 21, 175, 5, 5, 22, 175, 5, 5, 141, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 189, 5, 5, 75, 189, 5, 5, 76, 189, 5, 5, 77, 189, 5, 5, 78, 189, 5, 5, 79, 189, 5, 5, 80, 189, 5, 5, 81, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 197, 5, 5, 54, 197, 5, 5, 55, 197, 5, 5, 56, 197, 5, 5, 57, 197, 5, 5, 58, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 204, 5, 5, 242, 204, 5, 5, 243, 204, 5, 5, 244, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 212, 5, 5, 158, 212, 5, 5, 159, 212, 5, 5, 160, 212, 5, 5, 161, 212, 5, 5, 162, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 219, 5, 5, 130, 219, 5, 5, 131, 219, 5, 5, 132, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 226, 5, 5, 17, 226, 5, 5, 18, 226, 5, 5, 19, 226, 5, 5, 20, 226, 5, 5, 21, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 231, 5, 5, 111, 231, 5, 5, 112, 231, 5, 5, 113, 231, 5, 5, 114, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 235, 5, 5, 233, 235, 5, 5, 234, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 246, 5, 5, 161, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 247, 5, 5, 222, 247, 5, 5, 223, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 247, 192, 0, 0, 0, 5, 5, 207, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 169, 5, 5, 198, 169, 5, 5, 199, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 189, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 197, 5, 5, 108, 197, 5, 5, 109, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 197, 192, 0, 0, 0, 5, 5, 112, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 205, 192, 0, 0, 0, 5, 5, 52, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 205, 5, 5, 54, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 219, 5, 5, 169, 219, 5, 5, 170, 219, 5, 5, 171, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 226, 5, 5, 73, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 235, 5, 5, 250, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 239, 5, 5, 147, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 242, 5, 5, 159, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 244, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 161, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 165, 5, 5, 17, 165, 5, 5, 18, 165, 5, 5, 19, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 170, 5, 5, 11, 170, 5, 5, 12, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 175, 5, 5, 135, 175, 192, 0, 0, 0, 5, 5, 136, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 175, 5, 5, 139, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 182, 5, 5, 4, 182, 5, 5, 5, 182, 5, 5, 6, 182, 5, 5, 7, 182, 5, 5, 8, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 189, 5, 5, 205, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 189, 5, 5, 207, 189, 5, 5, 208, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 197, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 197, 5, 5, 207, 197, 5, 5, 208, 197, 5, 5, 209, 197, 5, 5, 210, 197, 5, 5, 211, 197, 5, 5, 212, 197, 5, 5, 213, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 197, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 197, 192, 0, 0, 0, 5, 5, 217, 197, 192, 0, 0, 0, 5, 5, 218, 197, 5, 5, 219, 197, 5, 5, 220, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 205, 5, 5, 134, 205, 5, 5, 135, 205, 5, 5, 136, 205, 5, 5, 137, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 205, 5, 5, 142, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 205, 192, 0, 0, 0, 5, 5, 144, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 212, 5, 5, 2, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 213, 5, 5, 4, 213, 5, 5, 5, 213, 5, 5, 6, 213, 192, 0, 0, 0, 5, 5, 7, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 213, 5, 5, 10, 213, 192, 0, 0, 0, 5, 5, 231, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 219, 5, 5, 233, 219, 5, 5, 234, 219, 5, 5, 235, 219, 5, 5, 236, 219, 5, 5, 237, 219, 5, 5, 238, 219, 5, 5, 239, 219, 5, 5, 240, 219, 5, 5, 241, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 226, 5, 5, 114, 226, 5, 5, 115, 226, 5, 5, 116, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 226, 5, 5, 119, 226, 5, 5, 120, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 231, 5, 5, 180, 231, 5, 5, 181, 231, 5, 5, 182, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 236, 5, 5, 15, 236, 5, 5, 16, 236, 5, 5, 17, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 236, 5, 5, 19, 236, 5, 5, 20, 236, 5, 5, 21, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 239, 5, 5, 172, 239, 5, 5, 173, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 239, 5, 5, 175, 239, 5, 5, 176, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 205, 5, 5, 161, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 205, 5, 5, 190, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 231, 192, 0, 0, 0, 5, 5, 206, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 190, 5, 5, 26, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 205, 5, 5, 234, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 213, 5, 5, 75, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 239, 192, 0, 0, 0, 5, 5, 197, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 176, 5, 5, 4, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 182, 192, 0, 0, 0, 5, 5, 146, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 226, 5, 5, 207, 226, 5, 5, 208, 226, 5, 5, 209, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 226, 5, 5, 212, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 158, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 206, 5, 5, 64, 206, 5, 5, 65, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 226, 5, 5, 40, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 213, 5, 5, 214, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 232, 5, 5, 122, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 236, 192, 0, 0, 0, 5, 5, 124, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 206, 5, 5, 138, 206, 5, 5, 139, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 220, 5, 5, 185, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 190, 5, 5, 195, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 220, 5, 5, 246, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 236, 5, 5, 160, 236, 5, 5, 161, 236, 5, 5, 162, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 242, 5, 5, 241, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 159, 5, 5, 182, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 161, 5, 5, 188, 161, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 176, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 183, 5, 5, 65, 183, 5, 5, 66, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 232, 5, 5, 156, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 161, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 183, 192, 0, 0, 0, 5, 5, 127, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 214, 5, 5, 106, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 236, 192, 0, 0, 0, 5, 5, 39, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 191, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 161, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 191, 5, 5, 65, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 214, 5, 5, 120, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 170, 192, 0, 0, 0, 5, 5, 223, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 176, 5, 5, 248, 176, 5, 5, 249, 176, 5, 5, 250, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 183, 5, 5, 239, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 199, 5, 5, 193, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 199, 5, 5, 196, 199, 5, 5, 197, 199, 5, 5, 198, 199, 5, 5, 199, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 199, 5, 5, 202, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 123, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 207, 5, 5, 125, 207, 5, 5, 126, 207, 5, 5, 127, 207, 5, 5, 128, 207, 5, 5, 129, 207, 5, 5, 130, 207, 5, 5, 131, 207, 5, 5, 132, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 214, 5, 5, 213, 214, 5, 5, 214, 214, 5, 5, 215, 214, 5, 5, 216, 214, 5, 5, 217, 214, 5, 5, 218, 214, 5, 5, 219, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 214, 5, 5, 223, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 221, 192, 0, 0, 0, 5, 5, 149, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 221, 5, 5, 152, 221, 5, 5, 153, 221, 5, 5, 154, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 227, 5, 5, 203, 227, 5, 5, 204, 227, 5, 5, 205, 227, 5, 5, 206, 227, 5, 5, 207, 227, 5, 5, 208, 227, 5, 5, 209, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 233, 5, 5, 5, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 233, 192, 0, 0, 0, 5, 5, 7, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 236, 5, 5, 249, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 240, 5, 5, 88, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 243, 5, 5, 48, 243, 5, 5, 49, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 245, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 245, 5, 5, 229, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 246, 5, 5, 231, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 248, 5, 5, 235, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 237, 192, 0, 0, 0, 5, 5, 34, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 221, 5, 5, 11, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 159, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 184, 5, 5, 63, 184, 5, 5, 64, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 228, 5, 5, 36, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 237, 5, 5, 50, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 12, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 233, 5, 5, 122, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 243, 5, 5, 103, 243, 5, 5, 104, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 200, 5, 5, 130, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 249, 192, 0, 0, 0, 5, 5, 241, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 208, 5, 5, 99, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 228, 5, 5, 135, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 233, 5, 5, 164, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 192, 192, 0, 0, 0, 5, 5, 112, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 200, 5, 5, 247, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 165, 5, 5, 199, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 192, 5, 5, 229, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 209, 5, 5, 6, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 216, 192, 0, 0, 0, 5, 5, 44, 216, 5, 5, 45, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 223, 192, 0, 0, 0, 5, 5, 44, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 223, 192, 0, 0, 0, 5, 5, 46, 223, 5, 5, 47, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 229, 5, 5, 27, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 229, 5, 5, 29, 229, 5, 5, 30, 229, 5, 5, 31, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 234, 5, 5, 20, 234, 5, 5, 21, 234, 5, 5, 22, 234, 5, 5, 23, 234, 5, 5, 24, 234, 5, 5, 25, 234, 5, 5, 26, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 234, 5, 5, 28, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 234, 5, 5, 30, 234, 5, 5, 31, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 237, 5, 5, 221, 237, 5, 5, 222, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 237, 5, 5, 226, 237, 192, 0, 0, 0, 5, 5, 227, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 241, 5, 5, 43, 241, 5, 5, 44, 241, 5, 5, 45, 241, 5, 5, 46, 241, 5, 5, 47, 241, 192, 0, 0, 0, 5, 5, 48, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 243, 5, 5, 193, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 194, 243, 5, 5, 195, 243, 5, 5, 196, 243, 192, 0, 0, 0, 5, 5, 197, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 243, 192, 0, 0, 0, 5, 5, 200, 243, 192, 0, 0, 0, 5, 5, 201, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 245, 5, 5, 173, 245, 5, 5, 174, 245, 5, 5, 175, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 247, 5, 5, 48, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 248, 192, 0, 0, 0, 5, 5, 81, 248, 5, 5, 82, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 249, 5, 5, 16, 249, 5, 5, 17, 249, 5, 5, 18, 249, 5, 5, 19, 249, 5, 5, 20, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 249, 5, 5, 2, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 250, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 250, 192, 0, 0, 0, 5, 5, 161, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 184, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 49, 157, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 165, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 177, 5, 5, 191, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 193, 5, 5, 54, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 201, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 216, 5, 5, 98, 216, 5, 5, 99, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 229, 5, 5, 89, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 229, 5, 5, 91, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 229, 5, 5, 126, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 234, 192, 0, 0, 0, 5, 5, 109, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 238, 192, 0, 0, 0, 5, 5, 90, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 244, 192, 0, 0, 0, 5, 5, 14, 244, 5, 5, 15, 244, 5, 5, 16, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 247, 5, 5, 150, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 248, 5, 5, 157, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 85, 249, 5, 5, 86, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 235, 192, 0, 0, 0, 5, 5, 185, 238, 192, 0, 0, 0, 5, 5, 186, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 239, 5, 5, 36, 162, 5, 5, 65, 157, 5, 5, 0, 133, 192, 0, 0, 0, 5, 5, 99, 162, 5, 5, 159, 171, 5, 5, 172, 171, 5, 5, 10, 178, 5, 5, 140, 185, 5, 5, 141, 193, 5, 5, 238, 209, 5, 5, 214, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 162, 5, 5, 98, 166, 5, 5, 153, 244, 5, 5, 104, 166, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 157, 5, 5, 70, 160, 192, 0, 0, 0, 5, 5, 227, 157, 5, 5, 91, 178, 5, 5, 205, 157, 5, 5, 195, 158, 5, 5, 120, 162, 192, 0, 0, 0, 5, 5, 27, 157, 5, 5, 90, 157, 5, 5, 147, 162, 5, 5, 131, 166, 5, 5, 142, 166, 5, 5, 175, 193, 5, 5, 37, 202, 192, 0, 0, 0, 5, 5, 9, 172, 5, 5, 11, 172, 5, 5, 51, 202, 5, 5, 93, 157, 5, 5, 218, 158, 5, 5, 219, 158, 5, 5, 221, 158, 5, 5, 226, 158, 5, 5, 162, 166, 5, 5, 189, 193, 5, 5, 171, 162, 5, 5, 23, 172, 5, 5, 144, 178, 5, 5, 144, 178, 5, 5, 144, 178, 192, 0, 0, 0, 5, 5, 138, 161, 5, 5, 13, 158, 5, 5, 152, 178, 192, 0, 0, 0, 5, 5, 253, 158, 5, 5, 5, 159, 5, 5, 133, 160, 5, 5, 37, 172, 5, 5, 212, 162, 5, 5, 228, 162, 5, 5, 201, 166, 5, 5, 41, 172, 5, 5, 175, 178, 5, 5, 200, 178, 5, 5, 250, 185, 5, 5, 10, 186, 5, 5, 208, 193, 5, 5, 208, 193, 5, 5, 228, 193, 5, 5, 246, 193, 5, 5, 254, 193, 5, 5, 69, 202, 5, 5, 103, 210, 5, 5, 53, 210, 5, 5, 104, 210, 5, 5, 85, 210, 5, 5, 106, 217, 5, 5, 236, 157, 5, 5, 181, 160, 5, 5, 159, 172, 5, 5, 86, 186, 5, 5, 111, 186, 5, 5, 116, 172, 5, 5, 33, 179, 5, 5, 62, 194, 5, 5, 136, 217, 192, 0, 0, 0, 5, 5, 68, 163, 5, 5, 127, 186, 5, 5, 70, 163, 5, 5, 184, 160, 5, 5, 159, 210, 5, 5, 137, 186, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 179, 5, 5, 62, 179, 5, 5, 74, 179, 5, 5, 185, 172, 5, 5, 189, 186, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 217, 5, 5, 253, 238, 5, 5, 253, 238, 192, 0, 0, 0, 5, 5, 225, 186, 5, 5, 254, 202, 5, 5, 224, 210, 5, 5, 9, 239, 192, 0, 0, 0, 5, 5, 135, 163, 5, 5, 2, 173, 5, 5, 21, 161, 5, 5, 119, 157, 192, 0, 0, 0, 5, 5, 242, 186, 5, 5, 123, 157, 5, 5, 216, 167, 5, 5, 173, 163, 192, 0, 0, 0, 5, 5, 221, 194, 192, 0, 0, 0, 5, 5, 30, 203, 5, 5, 251, 194, 5, 5, 235, 210, 5, 5, 167, 165, 5, 5, 36, 187, 192, 0, 0, 0, 5, 5, 254, 194, 5, 5, 172, 179, 5, 5, 3, 195, 5, 5, 253, 217, 5, 5, 203, 224, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 173, 5, 5, 53, 187, 5, 5, 56, 187, 5, 5, 23, 195, 192, 0, 0, 0, 5, 5, 138, 157, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 183, 5, 5, 5, 168, 5, 5, 5, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 163, 5, 5, 73, 187, 192, 0, 0, 0, 5, 5, 33, 195, 5, 5, 226, 163, 5, 5, 231, 163, 5, 5, 2, 164, 5, 5, 235, 179, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 179, 192, 0, 0, 0, 5, 5, 127, 187, 5, 5, 53, 211, 5, 5, 116, 203, 5, 5, 118, 203, 5, 5, 116, 203, 5, 5, 81, 211, 5, 5, 46, 218, 5, 5, 233, 224, 5, 5, 62, 218, 5, 5, 70, 218, 5, 5, 196, 230, 5, 5, 29, 239, 5, 5, 31, 239, 5, 5, 40, 161, 5, 5, 164, 187, 5, 5, 47, 161, 5, 5, 105, 168, 5, 5, 139, 168, 5, 5, 76, 180, 192, 0, 0, 0, 5, 5, 57, 180, 5, 5, 193, 173, 5, 5, 173, 187, 5, 5, 200, 187, 5, 5, 101, 180, 192, 0, 0, 0, 5, 5, 163, 203, 5, 5, 210, 203, 5, 5, 235, 187, 192, 0, 0, 0, 5, 5, 84, 218, 5, 5, 142, 211, 5, 5, 114, 218, 5, 5, 136, 211, 192, 0, 0, 0, 5, 5, 13, 188, 5, 5, 214, 203, 192, 0, 0, 0, 5, 5, 48, 188, 5, 5, 152, 180, 5, 5, 134, 180, 5, 5, 89, 225, 5, 5, 2, 196, 5, 5, 2, 204, 192, 0, 0, 0, 5, 5, 233, 171, 5, 5, 155, 185, 5, 5, 165, 193, 5, 5, 178, 211, 5, 5, 136, 170, 5, 5, 157, 170, 5, 5, 161, 180, 5, 5, 79, 188, 5, 5, 220, 206, 5, 5, 122, 164, 5, 5, 111, 164, 192, 0, 0, 0, 5, 5, 37, 169, 5, 5, 124, 174, 5, 5, 248, 168, 5, 5, 155, 174, 5, 5, 108, 188, 192, 0, 0, 0, 5, 5, 117, 188, 5, 5, 178, 180, 5, 5, 102, 196, 192, 0, 0, 0, 5, 5, 34, 204, 5, 5, 224, 211, 5, 5, 194, 218, 5, 5, 44, 231, 192, 0, 0, 0, 5, 5, 57, 231, 5, 5, 150, 246, 5, 5, 98, 161, 192, 0, 0, 0, 5, 5, 190, 225, 192, 0, 0, 0, 5, 5, 128, 204, 5, 5, 52, 212, 5, 5, 203, 188, 5, 5, 204, 188, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 212, 5, 5, 114, 161, 192, 0, 0, 0, 5, 5, 100, 169, 5, 5, 113, 169, 5, 5, 158, 164, 5, 5, 57, 181, 5, 5, 248, 174, 5, 5, 91, 181, 5, 5, 251, 174, 5, 5, 77, 181, 5, 5, 92, 181, 5, 5, 105, 181, 192, 0, 0, 0, 5, 5, 238, 174, 5, 5, 206, 196, 5, 5, 14, 197, 5, 5, 239, 204, 5, 5, 44, 197, 5, 5, 206, 204, 192, 0, 0, 0, 5, 5, 40, 189, 5, 5, 79, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 219, 5, 5, 132, 242, 5, 5, 112, 239, 5, 5, 109, 239, 192, 0, 0, 0, 5, 5, 226, 244, 5, 5, 231, 164, 5, 5, 225, 164, 5, 5, 37, 175, 192, 0, 0, 0, 5, 5, 252, 204, 192, 0, 0, 0, 5, 5, 140, 219, 192, 0, 0, 0, 5, 5, 121, 250, 5, 5, 146, 231, 5, 5, 62, 205, 192, 0, 0, 0, 5, 5, 120, 197, 5, 5, 217, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 239, 5, 5, 103, 158, 192, 0, 0, 0, 5, 5, 250, 169, 5, 5, 219, 181, 5, 5, 219, 181, 5, 5, 105, 205, 5, 5, 126, 205, 5, 5, 245, 212, 5, 5, 216, 219, 5, 5, 167, 239, 192, 0, 0, 0, 5, 5, 225, 197, 192, 0, 0, 0, 5, 5, 20, 170, 192, 0, 0, 0, 5, 5, 227, 189, 192, 0, 0, 0, 5, 5, 192, 205, 192, 0, 0, 0, 5, 5, 254, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 182, 5, 5, 90, 182, 5, 5, 90, 182, 5, 5, 33, 198, 5, 5, 41, 198, 5, 5, 32, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 43, 190, 5, 5, 4, 206, 5, 5, 60, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 182, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 213, 5, 5, 189, 182, 192, 0, 0, 0, 5, 5, 99, 220, 5, 5, 140, 213, 5, 5, 227, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 213, 5, 5, 201, 213, 192, 0, 0, 0, 5, 5, 138, 220, 5, 5, 247, 226, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 227, 192, 0, 0, 0, 5, 5, 100, 232, 5, 5, 125, 236, 5, 5, 78, 176, 192, 0, 0, 0, 5, 5, 193, 198, 5, 5, 241, 220, 5, 5, 21, 214, 5, 5, 231, 220, 5, 5, 130, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 227, 192, 0, 0, 0, 5, 5, 226, 190, 5, 5, 175, 236, 5, 5, 119, 170, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 199, 192, 0, 0, 0, 5, 5, 160, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 170, 5, 5, 109, 183, 192, 0, 0, 0, 5, 5, 24, 199, 5, 5, 206, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 183, 5, 5, 70, 199, 5, 5, 125, 208, 192, 0, 0, 0, 5, 5, 113, 165, 5, 5, 107, 165, 5, 5, 172, 170, 5, 5, 158, 162, 5, 5, 192, 170, 5, 5, 194, 170, 5, 5, 203, 170, 5, 5, 201, 176, 192, 0, 0, 0, 5, 5, 200, 176, 5, 5, 69, 191, 5, 5, 224, 183, 5, 5, 150, 191, 5, 5, 70, 191, 5, 5, 145, 191, 5, 5, 127, 199, 5, 5, 191, 199, 5, 5, 242, 183, 5, 5, 99, 199, 5, 5, 101, 199, 5, 5, 116, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 207, 5, 5, 74, 221, 5, 5, 108, 221, 192, 0, 0, 0, 5, 5, 181, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 245, 192, 0, 0, 0, 5, 5, 76, 245, 192, 0, 0, 0, 5, 5, 253, 176, 5, 5, 138, 207, 5, 5, 10, 233, 5, 5, 251, 236, 5, 5, 28, 184, 5, 5, 208, 199, 5, 5, 167, 207, 5, 5, 226, 199, 5, 5, 217, 227, 5, 5, 248, 214, 5, 5, 29, 215, 5, 5, 222, 227, 192, 0, 0, 0, 5, 5, 211, 221, 5, 5, 91, 240, 192, 0, 0, 0, 5, 5, 8, 228, 5, 5, 226, 161, 192, 0, 0, 0, 5, 5, 28, 200, 5, 5, 199, 207, 192, 0, 0, 0, 5, 5, 50, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 208, 5, 5, 61, 228, 5, 5, 36, 248, 5, 5, 150, 165, 192, 0, 0, 0, 5, 5, 77, 192, 5, 5, 128, 200, 5, 5, 253, 248, 5, 5, 131, 184, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 200, 5, 5, 68, 208, 5, 5, 86, 208, 192, 0, 0, 0, 5, 5, 146, 184, 5, 5, 149, 228, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 162, 5, 5, 137, 177, 5, 5, 165, 208, 192, 0, 0, 0, 5, 5, 206, 215, 5, 5, 195, 208, 5, 5, 5, 223, 5, 5, 6, 223, 5, 5, 47, 216, 5, 5, 38, 241, 5, 5, 171, 243, 192, 0, 0, 0, 5, 5, 107, 201, 192, 0, 0, 0, 5, 5, 46, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 201, 5, 5, 195, 224, 5, 5, 4, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 209, 5, 5, 96, 209, 5, 5, 124, 223, 192, 0, 0, 0, 5, 5, 41, 185, 192, 0, 0, 0, 5, 5, 151, 229, 5, 5, 115, 241, 5, 5, 175, 216, 5, 5, 147, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 244, 5, 5, 75, 244, 5, 5, 248, 223, 192, 0, 0, 0, 5, 5, 107, 244, 5, 5, 252, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 193, 192, 0, 0, 0, 5, 5, 235, 201, 5, 5, 141, 171, 5, 5, 72, 246, 5, 5, 37, 224, 5, 5, 203, 238, 5, 5, 244, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 134, 2, 4, 0, 0, 0, 42, 0, 134, 3, 4, 0, 0, 0, 50, 0, 134, 2, 4, 0, 0, 0, 50, 0, 134, 3, 4, 0, 0, 0, 58, 0, 134, 2, 4, 0, 0, 0, 58, 0, 134, 3, 4, 0, 0, 0, 66, 0, 134, 2, 4, 0, 0, 0, 66, 0, 134, 3, 4, 0, 0, 0, 68, 0, 134, 2, 4, 0, 0, 0, 68, 0, 134, 3, 4, 0, 0, 0, 70, 0, 134, 2, 4, 0, 0, 0, 70, 0, 134, 3, 4, 0, 0, 0, 82, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 2, 4, 0, 0, 0, 82, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 0, 128, 28, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 14, 157, 24, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 56, 157, 26, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 25, 159, 12, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 188, 157, 23, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 218, 157, 18, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 0, 153, 27, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 20, 157, 19, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 11, 157, 25, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 37, 157, 16, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 71, 158, 8, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 87, 158, 3, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 86, 158, 4, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 72, 158, 6, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 102, 157, 11, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 68, 158, 9, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 189, 180, 5, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 77, 161, 7, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 140, 160, 14, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 197, 181, 2, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 159, 162, 17, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 163, 158, 22, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 219, 166, 13, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 150, 167, 10, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 7, 160, 21, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 165, 166, 15, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 23, 160, 20, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 212, 55, 18, 254, 0, 0, 0, 0, 134, 3, 0, 42, 5, 162, 0, 0, 134, 2, 0, 42, 5, 162, 0, 0, 5, 3, 0, 42, 5, 162, 0, 0, 5, 2, 0, 42, 5, 162, 0, 0, 134, 3, 0, 50, 5, 162, 0, 0, 134, 2, 0, 50, 5, 162, 0, 0, 5, 3, 0, 50, 5, 162, 0, 0, 5, 2, 0, 50, 5, 162, 0, 0, 134, 3, 0, 58, 5, 162, 0, 0, 134, 2, 0, 58, 5, 162, 0, 0, 5, 3, 0, 58, 5, 162, 0, 0, 5, 2, 0, 58, 5, 162, 0, 0, 134, 3, 0, 68, 5, 160, 0, 0, 134, 2, 0, 68, 5, 160, 0, 0, 5, 3, 0, 68, 5, 160, 0, 0, 5, 2, 0, 68, 5, 160, 0, 0, 134, 3, 0, 82, 5, 162, 0, 0, 134, 2, 0, 82, 5, 162, 0, 0, 134, 74, 0, 82, 5, 162, 0, 0, 5, 3, 0, 82, 5, 162, 0, 0, 5, 2, 0, 82, 5, 162, 0, 0, 5, 74, 0, 82, 5, 162, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 134, 3, 0, 70, 5, 162, 0, 0, 134, 2, 0, 70, 5, 162, 0, 0, 5, 3, 0, 70, 5, 162, 0, 0, 5, 2, 0, 70, 5, 162, 0, 0, 134, 3, 0, 50, 5, 160, 0, 0, 134, 2, 0, 50, 5, 160, 0, 0, 5, 3, 0, 50, 5, 160, 0, 0, 5, 2, 0, 50, 5, 160, 0, 0, 134, 3, 0, 42, 5, 200, 0, 0, 134, 2, 0, 42, 5, 200, 0, 0, 5, 3, 0, 42, 5, 200, 0, 0, 5, 2, 0, 42, 5, 200, 0, 0, 134, 2, 0, 50, 5, 138, 0, 0, 5, 2, 0, 50, 5, 138, 0, 0, 134, 2, 0, 50, 5, 136, 0, 0, 5, 2, 0, 50, 5, 136, 0, 0, 134, 3, 0, 50, 5, 204, 0, 0, 134, 2, 0, 50, 5, 204, 0, 0, 5, 3, 0, 50, 5, 204, 0, 0, 5, 2, 0, 50, 5, 204, 0, 0, 134, 3, 0, 50, 5, 208, 0, 0, 134, 2, 0, 50, 5, 208, 0, 0, 5, 3, 0, 50, 5, 208, 0, 0, 5, 2, 0, 50, 5, 208, 0, 0, 134, 3, 0, 58, 5, 208, 0, 0, 134, 2, 0, 58, 5, 208, 0, 0, 5, 3, 0, 58, 5, 208, 0, 0, 5, 2, 0, 58, 5, 208, 0, 0, 134, 3, 0, 66, 5, 196, 0, 0, 134, 2, 0, 66, 5, 196, 0, 0, 5, 3, 0, 66, 5, 196, 0, 0, 5, 2, 0, 66, 5, 196, 0, 0, 134, 3, 0, 68, 5, 196, 0, 0, 134, 2, 0, 68, 5, 196, 0, 0, 5, 3, 0, 68, 5, 196, 0, 0, 5, 2, 0, 68, 5, 196, 0, 0, 134, 3, 0, 68, 5, 210, 0, 0, 134, 2, 0, 68, 5, 210, 0, 0, 5, 3, 0, 68, 5, 210, 0, 0, 5, 2, 0, 68, 5, 210, 0, 0, 134, 3, 0, 68, 5, 204, 0, 0, 134, 2, 0, 68, 5, 204, 0, 0, 5, 3, 0, 68, 5, 204, 0, 0, 5, 2, 0, 68, 5, 204, 0, 0, 134, 2, 0, 70, 5, 138, 0, 0, 5, 2, 0, 70, 5, 138, 0, 0, 134, 2, 0, 70, 5, 136, 0, 0, 5, 2, 0, 70, 5, 136, 0, 0, 134, 3, 0, 82, 5, 198, 0, 0, 134, 2, 0, 82, 5, 198, 0, 0, 134, 74, 0, 82, 5, 198, 0, 0, 5, 3, 0, 82, 5, 198, 0, 0, 5, 2, 0, 82, 5, 198, 0, 0, 5, 74, 0, 82, 5, 198, 0, 0, 134, 3, 0, 82, 5, 208, 0, 0, 134, 2, 0, 82, 5, 208, 0, 0, 134, 74, 0, 82, 5, 208, 0, 0, 5, 3, 0, 82, 5, 208, 0, 0, 5, 2, 0, 82, 5, 208, 0, 0, 5, 74, 0, 82, 5, 208, 0, 0, 134, 3, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 204, 0, 0, 134, 74, 0, 82, 5, 204, 0, 0, 5, 3, 0, 82, 5, 204, 0, 0, 5, 2, 0, 82, 5, 204, 0, 0, 5, 74, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 150, 0, 0, 5, 2, 0, 82, 5, 150, 0, 0, 134, 3, 0, 42, 5, 196, 0, 0, 134, 2, 0, 42, 5, 196, 0, 0, 5, 3, 0, 42, 5, 196, 0, 0, 5, 2, 0, 42, 5, 196, 0, 0, 134, 3, 0, 50, 5, 196, 0, 0, 134, 2, 0, 50, 5, 196, 0, 0, 5, 3, 0, 50, 5, 196, 0, 0, 5, 2, 0, 50, 5, 196, 0, 0, 134, 3, 0, 58, 5, 196, 0, 0, 134, 2, 0, 58, 5, 196, 0, 0, 5, 3, 0, 58, 5, 196, 0, 0, 5, 2, 0, 58, 5, 196, 0, 0, 134, 3, 0, 70, 5, 196, 0, 0, 134, 2, 0, 70, 5, 196, 0, 0, 5, 3, 0, 70, 5, 196, 0, 0, 5, 2, 0, 70, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 196, 0, 0, 134, 2, 0, 82, 5, 196, 0, 0, 134, 74, 0, 82, 5, 196, 0, 0, 5, 3, 0, 82, 5, 196, 0, 0, 5, 2, 0, 82, 5, 196, 0, 0, 5, 74, 0, 82, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 142, 9, 5, 5, 71, 171, 6, 5, 144, 9, 5, 5, 142, 9, 5, 5, 63, 170, 13, 5, 144, 9, 5, 5, 142, 9, 5, 5, 114, 184, 8, 5, 144, 9, 5, 5, 142, 9, 5, 5, 161, 182, 12, 5, 144, 9, 5, 5, 142, 9, 5, 5, 48, 213, 14, 5, 144, 9, 5, 5, 142, 9, 5, 5, 41, 208, 7, 5, 144, 9, 5, 5, 142, 9, 5, 5, 68, 190, 11, 5, 144, 9, 5, 5, 142, 9, 5, 5, 203, 161, 10, 5, 144, 9, 5, 5, 142, 9, 5, 5, 204, 161, 9, 5, 144, 9, 5, 5, 0, 21, 17, 5, 71, 158, 5, 5, 0, 23, 16, 5, 71, 158, 5, 5, 0, 25, 15, 5, 71, 158, 5, 5, 0, 27, 14, 5, 71, 158, 5, 5, 0, 29, 13, 5, 71, 158, 5, 5, 0, 31, 12, 5, 71, 158, 5, 5, 0, 33, 11, 5, 71, 158, 5, 5, 0, 35, 10, 5, 71, 158, 5, 5, 0, 37, 9, 5, 71, 158, 5, 5, 0, 21, 5, 5, 0, 19, 20, 5, 71, 158, 5, 5, 0, 21, 5, 5, 0, 21, 19, 5, 71, 158, 5, 5, 0, 21, 5, 5, 0, 23, 18, 5, 71, 158, 5, 5, 164, 158, 7, 5, 235, 166, 5, 5, 0, 19, 30, 5, 49, 175, 5, 5, 0, 21, 19, 5, 49, 175, 5, 5, 0, 23, 13, 5, 49, 175, 5, 5, 0, 25, 12, 5, 49, 175, 5, 5, 0, 27, 11, 5, 49, 175, 5, 5, 0, 29, 10, 5, 49, 175, 5, 5, 0, 31, 9, 5, 49, 175, 5, 5, 0, 33, 8, 5, 49, 175, 5, 5, 0, 35, 7, 5, 49, 175, 5, 5, 0, 37, 6, 5, 49, 175, 5, 5, 0, 21, 5, 5, 0, 19, 29, 5, 49, 175, 5, 5, 0, 21, 5, 5, 0, 21, 28, 5, 49, 175, 5, 5, 0, 21, 5, 5, 0, 23, 27, 5, 49, 175, 5, 5, 0, 21, 5, 5, 0, 25, 26, 5, 49, 175, 5, 5, 0, 21, 5, 5, 0, 27, 25, 5, 49, 175, 5, 5, 0, 21, 5, 5, 0, 29, 24, 5, 49, 175, 5, 5, 0, 21, 5, 5, 0, 31, 23, 5, 49, 175, 5, 5, 0, 21, 5, 5, 0, 33, 22, 5, 49, 175, 5, 5, 0, 21, 5, 5, 0, 35, 21, 5, 49, 175, 5, 5, 0, 21, 5, 5, 0, 37, 20, 5, 49, 175, 5, 5, 0, 23, 5, 5, 0, 19, 18, 5, 49, 175, 5, 5, 0, 23, 5, 5, 0, 21, 17, 5, 49, 175, 5, 5, 0, 23, 5, 5, 0, 23, 16, 5, 49, 175, 5, 5, 0, 23, 5, 5, 0, 25, 15, 5, 49, 175, 5, 5, 0, 23, 5, 5, 0, 27, 14, 5, 49, 175, 5, 5, 76, 159, 6, 5, 40, 161, 5, 5, 25, 174, 6, 5, 235, 166, 5, 5, 107, 157, 7, 5, 118, 159, 5, 5, 204, 168, 6, 5, 96, 169, 5, 5, 189, 180, 5, 5, 15, 161, 5, 5, 32, 160, 7, 5, 63, 170, 5, 5, 0, 21, 28, 5, 68, 158, 5, 5, 0, 23, 17, 5, 68, 158, 5, 5, 0, 25, 14, 5, 68, 158, 5, 5, 0, 27, 13, 5, 68, 158, 5, 5, 0, 29, 12, 5, 68, 158, 5, 5, 0, 31, 11, 5, 68, 158, 5, 5, 0, 33, 10, 5, 68, 158, 5, 5, 0, 35, 9, 5, 68, 158, 5, 5, 0, 37, 8, 5, 68, 158, 5, 5, 0, 21, 5, 5, 0, 19, 38, 5, 68, 158, 5, 5, 0, 21, 5, 5, 0, 21, 37, 5, 68, 158, 5, 5, 0, 21, 5, 5, 0, 23, 36, 5, 68, 158, 5, 5, 0, 21, 5, 5, 0, 25, 35, 5, 68, 158, 5, 5, 0, 21, 5, 5, 0, 27, 34, 5, 68, 158, 5, 5, 0, 21, 5, 5, 0, 29, 33, 5, 68, 158, 5, 5, 0, 21, 5, 5, 0, 31, 32, 5, 68, 158, 5, 5, 0, 21, 5, 5, 0, 33, 31, 5, 68, 158, 5, 5, 0, 21, 5, 5, 0, 35, 30, 5, 68, 158, 5, 5, 0, 21, 5, 5, 0, 37, 29, 5, 68, 158, 5, 5, 0, 23, 5, 5, 0, 19, 27, 5, 68, 158, 5, 5, 0, 23, 5, 5, 0, 21, 26, 5, 68, 158, 5, 5, 0, 23, 5, 5, 0, 23, 25, 5, 68, 158, 5, 5, 0, 23, 5, 5, 0, 25, 24, 5, 68, 158, 5, 5, 0, 23, 5, 5, 0, 27, 23, 5, 68, 158, 5, 5, 0, 23, 5, 5, 0, 29, 22, 5, 68, 158, 5, 5, 0, 23, 5, 5, 0, 31, 21, 5, 68, 158, 5, 5, 0, 23, 5, 5, 0, 33, 20, 5, 68, 158, 5, 5, 0, 23, 5, 5, 0, 35, 19, 5, 68, 158, 5, 5, 0, 23, 5, 5, 0, 37, 18, 5, 68, 158, 5, 5, 0, 25, 5, 5, 0, 19, 16, 5, 68, 158, 5, 5, 0, 25, 5, 5, 0, 21, 15, 5, 68, 158, 5, 5, 82, 10, 5, 5, 113, 159, 8, 5, 84, 10, 5, 5, 82, 10, 5, 5, 56, 157, 14, 5, 84, 10, 5, 5, 82, 10, 5, 5, 14, 157, 13, 5, 84, 10, 5, 5, 82, 10, 5, 5, 224, 160, 11, 5, 84, 10, 5, 5, 82, 10, 5, 5, 49, 175, 7, 5, 84, 10, 5, 5, 82, 10, 5, 5, 101, 159, 10, 5, 84, 10, 5, 5, 82, 10, 5, 5, 253, 189, 6, 5, 84, 10, 5, 5, 82, 10, 5, 5, 182, 193, 12, 5, 84, 10, 5, 5, 82, 10, 5, 5, 21, 188, 9, 5, 84, 10, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 0, 42, 156, 5, 5, 0, 12, 3, 7, 0, 12, 3, 198, 161, 64, 3, 198, 129, 65, 3, 0, 234, 134, 3, 0, 3, 198, 129, 1, 3, 0, 234, 134, 3, 4, 3, 0, 234, 134, 2, 0, 50, 156, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 134, 2, 12, 3, 0, 192, 198, 97, 64, 3, 0, 192, 198, 65, 65, 3, 0, 242, 134, 3, 0, 3, 0, 192, 198, 65, 1, 3, 0, 242, 134, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 58, 156, 5, 5, 0, 12, 3, 9, 0, 12, 3, 0, 192, 198, 161, 64, 3, 0, 192, 198, 129, 65, 3, 0, 250, 134, 3, 0, 3, 0, 192, 198, 129, 1, 3, 0, 250, 134, 3, 4, 3, 0, 250, 134, 2, 0, 66, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 0, 192, 198, 225, 64, 3, 0, 192, 198, 193, 65, 3, 255, 255, 0, 66, 134, 3, 0, 3, 0, 192, 198, 193, 1, 3, 255, 255, 0, 66, 134, 3, 4, 3, 255, 255, 0, 66, 134, 2, 0, 68, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 33, 64, 3, 1, 192, 198, 1, 65, 3, 255, 255, 0, 68, 134, 3, 0, 3, 1, 192, 198, 1, 1, 3, 255, 255, 0, 68, 134, 3, 4, 3, 255, 255, 0, 68, 134, 2, 0, 70, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 97, 64, 3, 1, 192, 198, 65, 65, 3, 255, 255, 0, 70, 134, 3, 0, 3, 1, 192, 198, 65, 1, 3, 255, 255, 0, 70, 134, 3, 4, 3, 255, 255, 0, 70, 134, 2, 0, 82, 156, 5, 7, 0, 12, 3, 14, 0, 12, 3, 1, 192, 198, 161, 64, 3, 1, 192, 198, 129, 65, 3, 255, 255, 0, 82, 134, 3, 68, 3, 255, 255, 0, 82, 134, 71, 0, 3, 1, 192, 198, 129, 1, 3, 255, 255, 0, 82, 134, 3, 4, 3, 255, 255, 0, 82, 134, 2, 8, 3, 197, 127, 0, 82, 134, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 0, 42, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 1, 192, 198, 225, 64, 3, 1, 192, 198, 193, 65, 3, 0, 234, 5, 3, 0, 3, 1, 192, 198, 193, 1, 3, 0, 234, 5, 3, 4, 3, 0, 234, 5, 2, 0, 50, 5, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 5, 2, 12, 3, 2, 192, 198, 33, 64, 3, 2, 192, 198, 1, 65, 3, 0, 242, 5, 3, 0, 3, 2, 192, 198, 1, 1, 3, 0, 242, 5, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 58, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 2, 192, 198, 97, 64, 3, 2, 192, 198, 65, 65, 3, 0, 250, 5, 3, 0, 3, 2, 192, 198, 65, 1, 3, 0, 250, 5, 3, 4, 3, 0, 250, 5, 2, 0, 66, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 161, 64, 3, 2, 192, 198, 129, 65, 3, 255, 255, 0, 66, 5, 3, 0, 3, 2, 192, 198, 129, 1, 3, 255, 255, 0, 66, 5, 3, 4, 3, 255, 255, 0, 66, 5, 2, 0, 68, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 225, 64, 3, 2, 192, 198, 193, 65, 3, 255, 255, 0, 68, 5, 3, 0, 3, 2, 192, 198, 193, 1, 3, 255, 255, 0, 68, 5, 3, 4, 3, 255, 255, 0, 68, 5, 2, 0, 70, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 3, 192, 198, 33, 64, 3, 3, 192, 198, 1, 65, 3, 255, 255, 0, 70, 5, 3, 0, 3, 3, 192, 198, 1, 1, 3, 255, 255, 0, 70, 5, 3, 4, 3, 255, 255, 0, 70, 5, 2, 0, 82, 5, 5, 7, 0, 12, 3, 14, 0, 12, 3, 3, 192, 198, 97, 64, 3, 3, 192, 198, 65, 65, 3, 255, 255, 0, 82, 5, 3, 68, 3, 255, 255, 0, 82, 5, 71, 0, 3, 3, 192, 198, 65, 1, 3, 255, 255, 0, 82, 5, 3, 4, 3, 255, 255, 0, 82, 5, 2, 8, 3, 197, 127, 0, 82, 5, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 82, 134, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 5, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 82, 5, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 42, 196, 162, 5, 0, 12, 3, 8, 0, 12, 3, 3, 192, 198, 194, 64, 3, 3, 192, 198, 130, 65, 3, 197, 162, 0, 3, 3, 192, 198, 130, 1, 3, 197, 162, 4, 3, 0, 192, 197, 98, 5, 42, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 66, 64, 3, 4, 192, 198, 2, 65, 3, 0, 192, 197, 162, 0, 3, 4, 192, 198, 2, 1, 3, 0, 192, 197, 162, 4, 3, 0, 192, 197, 226, 156, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 194, 64, 3, 4, 192, 198, 130, 65, 3, 1, 192, 197, 34, 0, 3, 4, 192, 198, 130, 1, 3, 1, 192, 197, 34, 4, 3, 1, 192, 197, 98, 5, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 66, 64, 3, 5, 192, 198, 2, 65, 3, 1, 192, 197, 162, 0, 3, 5, 192, 198, 2, 1, 3, 1, 192, 197, 162, 4, 3, 1, 192, 197, 226, 156, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 194, 64, 3, 5, 192, 198, 130, 65, 3, 2, 192, 197, 34, 0, 3, 5, 192, 198, 130, 1, 3, 2, 192, 197, 34, 4, 3, 2, 192, 197, 98, 5, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 66, 64, 3, 6, 192, 198, 2, 65, 3, 2, 192, 197, 162, 0, 3, 6, 192, 198, 2, 1, 3, 2, 192, 197, 162, 4, 3, 2, 192, 197, 226, 156, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 194, 64, 3, 6, 192, 198, 130, 65, 3, 3, 192, 197, 34, 0, 3, 6, 192, 198, 130, 1, 3, 3, 192, 197, 34, 4, 3, 3, 192, 197, 98, 5, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 7, 192, 198, 66, 64, 3, 7, 192, 198, 2, 65, 3, 3, 192, 197, 162, 0, 3, 7, 192, 198, 2, 1, 3, 3, 192, 197, 162, 4, 3, 3, 192, 197, 226, 156, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 4, 192, 197, 162, 12, 3, 7, 192, 198, 194, 64, 3, 7, 192, 198, 130, 65, 3, 4, 192, 197, 34, 0, 3, 7, 192, 198, 130, 1, 3, 4, 192, 197, 34, 4, 3, 4, 192, 197, 98, 5, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 5, 192, 197, 98, 12, 3, 8, 192, 198, 66, 64, 3, 8, 192, 198, 2, 65, 3, 4, 192, 197, 226, 0, 3, 8, 192, 198, 2, 1, 3, 4, 192, 197, 226, 4, 3, 5, 192, 197, 34, 156, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 8, 192, 198, 194, 64, 3, 8, 192, 198, 130, 65, 3, 5, 192, 197, 162, 0, 3, 8, 192, 198, 130, 1, 3, 5, 192, 197, 162, 4, 3, 5, 192, 197, 226, 5, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 9, 192, 198, 66, 64, 3, 9, 192, 198, 2, 65, 3, 6, 192, 197, 34, 0, 3, 9, 192, 198, 2, 1, 3, 6, 192, 197, 34, 4, 3, 6, 192, 197, 98, 156, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 34, 12, 3, 9, 192, 198, 194, 64, 3, 9, 192, 198, 130, 65, 3, 6, 192, 197, 162, 0, 3, 9, 192, 198, 130, 1, 3, 6, 192, 197, 162, 4, 3, 6, 192, 197, 226, 5, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 226, 12, 3, 10, 192, 198, 66, 64, 3, 10, 192, 198, 2, 65, 3, 7, 192, 197, 98, 0, 3, 10, 192, 198, 2, 1, 3, 7, 192, 197, 98, 4, 3, 7, 192, 197, 162, 156, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 10, 192, 198, 194, 64, 3, 10, 192, 198, 130, 65, 3, 8, 192, 197, 34, 0, 3, 10, 192, 198, 130, 1, 3, 8, 192, 197, 34, 4, 3, 8, 192, 197, 98, 5, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 66, 64, 3, 11, 192, 198, 2, 65, 3, 8, 192, 197, 162, 0, 3, 11, 192, 198, 2, 1, 3, 8, 192, 197, 162, 4, 3, 8, 192, 197, 226, 156, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 194, 64, 3, 11, 192, 198, 130, 65, 3, 9, 192, 197, 34, 0, 3, 11, 192, 198, 130, 1, 3, 9, 192, 197, 34, 4, 3, 9, 192, 197, 98, 5, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 66, 64, 3, 12, 192, 198, 2, 65, 3, 9, 192, 197, 162, 0, 3, 12, 192, 198, 2, 1, 3, 9, 192, 197, 162, 4, 3, 9, 192, 197, 226, 156, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 194, 64, 3, 12, 192, 198, 130, 65, 3, 10, 192, 197, 34, 0, 3, 12, 192, 198, 130, 1, 3, 10, 192, 197, 34, 4, 3, 10, 192, 197, 98, 5, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 66, 64, 3, 13, 192, 198, 2, 65, 3, 10, 192, 197, 162, 0, 3, 13, 192, 198, 2, 1, 3, 10, 192, 197, 162, 4, 3, 10, 192, 197, 226, 156, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 194, 64, 3, 13, 192, 198, 130, 65, 3, 12, 192, 197, 34, 0, 3, 13, 192, 198, 130, 1, 3, 12, 192, 197, 34, 4, 3, 12, 192, 197, 98, 5, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 66, 64, 3, 14, 192, 198, 2, 65, 3, 12, 192, 197, 162, 0, 3, 14, 192, 198, 2, 1, 3, 12, 192, 197, 162, 4, 3, 12, 192, 197, 226, 156, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 194, 64, 3, 14, 192, 198, 130, 65, 3, 13, 192, 197, 34, 0, 3, 14, 192, 198, 130, 1, 3, 13, 192, 197, 34, 4, 3, 13, 192, 197, 98, 5, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 66, 64, 3, 15, 192, 198, 2, 65, 3, 13, 192, 197, 162, 0, 3, 15, 192, 198, 2, 1, 3, 13, 192, 197, 162, 4, 3, 13, 192, 197, 226, 156, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 194, 64, 3, 15, 192, 198, 130, 65, 3, 14, 192, 197, 34, 0, 3, 15, 192, 198, 130, 1, 3, 14, 192, 197, 34, 4, 3, 14, 192, 197, 98, 5, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 66, 64, 3, 16, 192, 198, 2, 65, 3, 14, 192, 197, 162, 0, 3, 16, 192, 198, 2, 1, 3, 14, 192, 197, 162, 4, 3, 14, 192, 197, 226, 156, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 194, 64, 3, 16, 192, 198, 130, 65, 3, 15, 192, 197, 34, 0, 3, 16, 192, 198, 130, 1, 3, 15, 192, 197, 34, 4, 3, 15, 192, 197, 98, 5, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 66, 64, 3, 17, 192, 198, 2, 65, 3, 15, 192, 197, 162, 0, 3, 17, 192, 198, 2, 1, 3, 15, 192, 197, 162, 4, 3, 15, 192, 197, 226, 156, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 194, 64, 3, 17, 192, 198, 130, 65, 3, 16, 192, 197, 34, 0, 3, 17, 192, 198, 130, 1, 3, 16, 192, 197, 34, 4, 3, 16, 192, 197, 98, 5, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 66, 64, 3, 18, 192, 198, 2, 65, 3, 16, 192, 197, 162, 0, 3, 18, 192, 198, 2, 1, 3, 16, 192, 197, 162, 4, 3, 16, 192, 197, 226, 156, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 194, 64, 3, 18, 192, 198, 130, 65, 3, 17, 192, 197, 34, 0, 3, 18, 192, 198, 130, 1, 3, 17, 192, 197, 34, 4, 3, 17, 192, 197, 98, 5, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 66, 64, 3, 19, 192, 198, 2, 65, 3, 17, 192, 197, 162, 0, 3, 19, 192, 198, 2, 1, 3, 17, 192, 197, 162, 4, 3, 17, 192, 197, 226, 156, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 194, 64, 3, 19, 192, 198, 130, 65, 3, 18, 192, 197, 34, 0, 3, 19, 192, 198, 130, 1, 3, 18, 192, 197, 34, 4, 3, 18, 192, 197, 98, 5, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 20, 192, 198, 66, 64, 3, 20, 192, 198, 2, 65, 3, 18, 192, 197, 162, 0, 3, 20, 192, 198, 2, 1, 3, 18, 192, 197, 162, 4, 3, 18, 192, 197, 226, 156, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 20, 192, 197, 162, 12, 3, 20, 192, 198, 194, 64, 3, 20, 192, 198, 130, 65, 3, 20, 192, 197, 34, 0, 3, 20, 192, 198, 130, 1, 3, 20, 192, 197, 34, 4, 3, 20, 192, 197, 98, 5, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 21, 192, 197, 98, 12, 3, 21, 192, 198, 66, 64, 3, 21, 192, 198, 2, 65, 3, 20, 192, 197, 226, 0, 3, 21, 192, 198, 2, 1, 3, 20, 192, 197, 226, 4, 3, 21, 192, 197, 34, 156, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 34, 12, 3, 21, 192, 198, 194, 64, 3, 21, 192, 198, 130, 65, 3, 21, 192, 197, 162, 0, 3, 21, 192, 198, 130, 1, 3, 21, 192, 197, 162, 4, 3, 21, 192, 197, 226, 5, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 226, 12, 3, 22, 192, 198, 66, 64, 3, 22, 192, 198, 2, 65, 3, 22, 192, 197, 98, 0, 3, 22, 192, 198, 2, 1, 3, 22, 192, 197, 98, 4, 3, 22, 192, 197, 162, 156, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 23, 192, 197, 162, 12, 3, 22, 192, 198, 194, 64, 3, 22, 192, 198, 130, 65, 3, 23, 192, 197, 34, 0, 3, 22, 192, 198, 130, 1, 3, 23, 192, 197, 34, 4, 3, 23, 192, 197, 98, 5, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 24, 192, 197, 98, 12, 3, 23, 192, 198, 66, 64, 3, 23, 192, 198, 2, 65, 3, 23, 192, 197, 226, 0, 3, 23, 192, 198, 2, 1, 3, 23, 192, 197, 226, 4, 3, 24, 192, 197, 34, 156, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 23, 192, 198, 194, 64, 3, 23, 192, 198, 130, 65, 3, 25, 192, 197, 34, 0, 3, 23, 192, 198, 130, 1, 3, 25, 192, 197, 34, 4, 3, 25, 192, 197, 98, 5, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 66, 64, 3, 24, 192, 198, 2, 65, 3, 25, 192, 197, 162, 0, 3, 24, 192, 198, 2, 1, 3, 25, 192, 197, 162, 4, 3, 25, 192, 197, 226, 156, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 194, 64, 3, 24, 192, 198, 130, 65, 3, 26, 192, 197, 34, 0, 3, 24, 192, 198, 130, 1, 3, 26, 192, 197, 34, 4, 3, 26, 192, 197, 98, 5, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 66, 64, 3, 25, 192, 198, 2, 65, 3, 26, 192, 197, 162, 0, 3, 25, 192, 198, 2, 1, 3, 26, 192, 197, 162, 4, 3, 26, 192, 197, 226, 156, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 194, 64, 3, 25, 192, 198, 130, 65, 3, 27, 192, 197, 34, 0, 3, 25, 192, 198, 130, 1, 3, 27, 192, 197, 34, 4, 3, 27, 192, 197, 98, 5, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 66, 64, 3, 26, 192, 198, 2, 65, 3, 27, 192, 197, 162, 0, 3, 26, 192, 198, 2, 1, 3, 27, 192, 197, 162, 4, 3, 27, 192, 197, 226, 156, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 194, 64, 3, 26, 192, 198, 130, 65, 3, 28, 192, 197, 34, 0, 3, 26, 192, 198, 130, 1, 3, 28, 192, 197, 34, 4, 3, 28, 192, 197, 98, 5, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 66, 64, 3, 27, 192, 198, 2, 65, 3, 28, 192, 197, 162, 0, 3, 27, 192, 198, 2, 1, 3, 28, 192, 197, 162, 4, 3, 28, 192, 197, 226, 29, 0, 197, 35, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 227, 64, 3, 27, 192, 198, 131, 65, 3, 29, 192, 197, 131, 0, 3, 27, 192, 198, 131, 1, 3, 29, 192, 197, 131, 4, 3, 29, 192, 197, 227, 30, 0, 197, 67, 5, 0, 12, 3, 9, 0, 12, 3, 28, 192, 198, 163, 64, 3, 28, 192, 198, 67, 65, 3, 30, 192, 197, 163, 0, 3, 28, 192, 198, 67, 1, 3, 30, 192, 197, 163, 4, 3, 31, 192, 197, 3, 156, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 31, 192, 197, 226, 12, 3, 29, 192, 198, 66, 64, 3, 29, 192, 198, 2, 65, 3, 31, 192, 197, 98, 0, 3, 29, 192, 198, 2, 1, 3, 31, 192, 197, 98, 4, 3, 31, 192, 197, 162, 5, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 32, 192, 197, 162, 12, 3, 29, 192, 198, 194, 64, 3, 29, 192, 198, 130, 65, 3, 32, 192, 197, 34, 0, 3, 29, 192, 198, 130, 1, 3, 32, 192, 197, 34, 4, 3, 32, 192, 197, 98, 32, 0, 197, 227, 6, 0, 8, 3, 12, 0, 8, 3, 34, 192, 197, 3, 12, 3, 30, 192, 198, 99, 64, 3, 30, 192, 198, 3, 65, 3, 33, 192, 197, 67, 0, 3, 30, 192, 198, 3, 1, 3, 33, 192, 197, 67, 4, 3, 33, 192, 197, 163, 34, 0, 197, 99, 6, 0, 8, 3, 12, 0, 8, 3, 35, 192, 197, 131, 12, 3, 31, 192, 198, 35, 64, 3, 30, 192, 198, 195, 65, 3, 34, 192, 197, 195, 0, 3, 30, 192, 198, 195, 1, 3, 34, 192, 197, 195, 4, 3, 35, 192, 197, 35, 41, 0, 198, 161, 36, 0, 19, 40, 82, 0, 28, 40, 42, 0, 33, 40, 20, 0, 33, 40, 255, 255, 163, 250, 5, 5, 35, 40, 255, 255, 169, 250, 5, 5, 36, 40, 255, 255, 171, 250, 5, 5, 39, 40, 255, 255, 173, 250, 5, 5, 48, 40, 255, 255, 175, 250, 5, 5, 28, 40, 255, 255, 84, 250, 5, 5, 29, 40, 255, 255, 120, 250, 5, 5, 30, 40, 255, 255, 136, 250, 5, 5, 31, 40, 255, 255, 152, 250, 5, 5, 32, 40, 255, 255, 158, 250, 5, 5, 23, 40, 20, 0, 23, 40, 255, 255, 175, 247, 5, 5, 24, 40, 255, 255, 181, 248, 5, 5, 25, 40, 255, 255, 99, 249, 5, 5, 26, 40, 255, 255, 219, 249, 5, 5, 27, 40, 255, 255, 26, 250, 5, 5, 19, 40, 255, 255, 215, 238, 5, 5, 20, 40, 255, 255, 24, 242, 5, 5, 21, 40, 255, 255, 147, 244, 5, 5, 22, 40, 255, 255, 90, 246, 5, 5, 10, 40, 38, 0, 14, 40, 20, 0, 14, 40, 255, 255, 210, 209, 5, 5, 15, 40, 255, 255, 248, 216, 5, 5, 16, 40, 255, 255, 44, 224, 5, 5, 17, 40, 255, 255, 53, 230, 5, 5, 18, 40, 255, 255, 33, 235, 5, 5, 10, 40, 255, 255, 238, 177, 5, 5, 11, 40, 255, 255, 79, 185, 5, 5, 12, 40, 255, 255, 110, 193, 5, 5, 13, 40, 255, 255, 238, 201, 5, 5, 5, 40, 20, 0, 5, 40, 255, 255, 126, 158, 5, 5, 6, 40, 255, 255, 213, 159, 5, 5, 7, 40, 255, 255, 32, 162, 5, 5, 8, 40, 255, 255, 253, 165, 5, 5, 9, 40, 255, 255, 144, 171, 5, 5, 1, 40, 255, 255, 0, 127, 5, 5, 2, 40, 255, 255, 0, 150, 5, 5, 3, 40, 255, 255, 53, 157, 5, 5, 4, 40, 255, 255, 163, 157, 5, 5, 8, 0, 1, 40, 34, 40, 35, 40, 37, 40, 39, 40, 40, 40, 48, 40, 49, 40, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 28, 4, 185, 76, 185, 80, 185, 84, 30, 4, 186, 96, 185, 100, 185, 104, 32, 4, 34, 4, 36, 4, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 38, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 40, 4, 168, 60, 168, 64, 168, 68, 42, 4, 168, 76, 168, 80, 168, 84, 44, 4, 168, 96, 168, 100, 168, 104, 46, 4, 48, 4, 50, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 52, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 89, 56, 57, 56, 217, 57, 121, 58, 57, 58, 25, 58, 0, 8, 217, 66, 121, 72, 57, 72, 54, 4, 58, 74, 89, 88, 57, 88, 218, 89, 58, 90, 217, 68, 122, 114, 89, 120, 57, 120, 217, 121, 121, 122, 57, 122, 136, 14, 185, 122, 89, 152, 57, 152, 217, 153, 56, 4, 153, 169, 185, 176, 2, 8, 72, 56, 40, 56, 200, 57, 104, 58, 40, 58, 8, 58, 4, 8, 200, 66, 104, 72, 40, 72, 58, 4, 40, 74, 72, 88, 40, 88, 200, 89, 40, 90, 200, 68, 104, 114, 72, 120, 40, 120, 200, 121, 104, 122, 40, 122, 128, 14, 168, 122, 72, 152, 40, 152, 200, 153, 60, 4, 136, 169, 168, 176, 40, 170, 25, 56, 8, 56, 185, 57, 168, 57, 62, 4, 64, 4, 153, 65, 136, 65, 217, 65, 200, 65, 153, 66, 136, 66, 249, 65, 232, 65, 249, 69, 232, 69, 25, 71, 8, 71, 25, 72, 8, 72, 186, 73, 168, 73, 154, 74, 136, 74, 66, 4, 68, 4, 89, 72, 72, 72, 217, 81, 200, 81, 185, 81, 168, 81, 153, 82, 136, 82, 217, 82, 200, 82, 217, 85, 200, 85, 25, 87, 8, 87, 122, 90, 104, 90, 25, 88, 8, 88, 186, 89, 168, 89, 70, 4, 72, 4, 154, 90, 168, 92, 6, 8, 8, 8, 218, 97, 200, 97, 217, 102, 200, 102, 168, 132, 153, 105, 136, 105, 217, 106, 200, 106, 249, 105, 232, 105, 57, 107, 40, 107, 25, 107, 8, 107, 25, 112, 8, 112, 74, 4, 76, 4, 57, 112, 40, 112, 10, 8, 185, 116, 168, 116, 25, 120, 8, 120, 185, 121, 168, 121, 89, 122, 72, 122, 12, 8, 14, 8, 153, 137, 136, 137, 217, 138, 200, 138, 249, 137, 232, 137, 154, 141, 136, 141, 218, 141, 200, 141, 218, 142, 200, 142, 250, 141, 232, 141, 217, 146, 200, 146, 249, 145, 232, 145, 185, 148, 168, 148, 121, 154, 104, 154, 25, 152, 8, 152, 185, 153, 168, 153, 25, 154, 8, 154, 89, 154, 72, 154, 78, 4, 80, 4, 217, 161, 200, 161, 217, 169, 200, 169, 57, 170, 153, 173, 136, 173, 153, 174, 136, 174, 249, 173, 232, 173, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 72, 255, 5, 186, 88, 255, 5, 185, 108, 255, 5, 186, 112, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 255, 5, 168, 72, 255, 5, 168, 88, 255, 5, 168, 108, 255, 5, 168, 112, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 218, 73, 255, 5, 217, 152, 255, 5, 200, 73, 255, 5, 200, 152, 255, 5, 249, 58, 255, 5, 232, 58, 255, 5, 250, 74, 255, 5, 232, 74, 255, 5, 250, 90, 255, 5, 232, 90, 255, 5, 218, 114, 255, 5, 200, 114, 255, 5, 249, 154, 255, 5, 232, 154, 255, 1, 170, 170, 3, 0, 68, 0, 59, 0, 32, 0, 159, 251, 1, 16, 206, 41, 1, 96, 1, 0, 0, 96, 170, 170, 170, 170, 140, 77, 0, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 100, 0, 0, 0, 100, 1, 0, 0, 164, 32, 0, 0, 164, 32, 0, 0, 140, 40, 0, 0, 140, 40, 0, 0, 72, 45, 0, 0, 72, 45, 0, 0, 8, 69, 0, 0, 64, 73, 0, 0, 112, 77, 0, 0, 17, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 41, 127, 0, 0, 96, 125, 0, 0, 123, 44, 0, 118, 123, 125, 0, 0, 126, 169, 255, 0, 254, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 0, 249, 250, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 254, 255, 50, 105, 114, 84, 1, 0, 64, 8, 235, 0, 255, 255, 192, 0, 32, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 109, 0, 117, 0, 122, 0, 130, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 169, 0, 176, 0, 184, 0, 189, 0, 48, 0, 197, 0, 205, 0, 213, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 226, 0, 161, 0, 48, 0, 48, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 153, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 218, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 180, 1, 232, 1, 36, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 36, 0, 202, 193, 36, 0, 202, 226, 36, 0, 202, 3, 37, 0, 202, 36, 37, 0, 202, 69, 37, 0, 202, 102, 37, 0, 202, 135, 37, 0, 202, 168, 37, 0, 202, 201, 37, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 134, 2, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 38, 8, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 201, 6, 11, 0, 201, 70, 14, 0, 201, 134, 17, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 198, 20, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 70, 28, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 38, 31, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 198, 36, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 201, 166, 39, 0, 201, 230, 42, 0, 201, 38, 46, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 102, 49, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 134, 3, 0, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 226, 35, 0, 196, 160, 156, 46, 198, 65, 0, 0, 134, 3, 0, 50, 201, 230, 56, 0, 196, 150, 156, 50, 198, 129, 0, 0, 134, 3, 0, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 198, 65, 1, 0, 134, 3, 0, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 198, 129, 1, 0, 134, 3, 0, 82, 196, 142, 156, 82, 201, 198, 59, 0, 196, 136, 156, 90, 160, 5, 38, 93, 197, 34, 36, 0, 198, 193, 1, 0, 5, 3, 0, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 98, 36, 0, 196, 160, 5, 46, 198, 1, 2, 0, 5, 3, 0, 50, 201, 102, 63, 0, 196, 150, 5, 50, 198, 65, 2, 0, 5, 3, 0, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 198, 1, 3, 0, 5, 3, 0, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 198, 65, 3, 0, 5, 3, 0, 82, 196, 142, 5, 82, 201, 70, 66, 0, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 134, 2, 0, 42, 5, 2, 0, 42, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 69, 0, 201, 134, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 50, 5, 2, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 75, 0, 201, 70, 78, 0, 198, 97, 0, 0, 198, 33, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 58, 5, 2, 0, 58, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 81, 0, 201, 6, 84, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 68, 5, 3, 0, 68, 201, 230, 86, 0, 201, 198, 89, 0, 198, 33, 1, 0, 198, 225, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 70, 5, 2, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 82, 5, 2, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 92, 0, 201, 230, 95, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 99, 0, 201, 6, 102, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 104, 0, 201, 38, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 198, 225, 1, 0, 198, 161, 0, 0, 198, 97, 2, 0, 198, 97, 1, 0, 198, 33, 3, 0, 198, 161, 1, 0, 198, 97, 3, 0, 134, 70, 0, 82, 5, 70, 0, 82, 134, 71, 0, 82, 5, 71, 0, 82, 134, 72, 0, 82, 5, 72, 0, 82, 134, 73, 0, 82, 5, 73, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 111, 0, 201, 70, 114, 0, 197, 98, 8, 0, 197, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 1, 0, 198, 193, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 117, 0, 201, 6, 120, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 122, 0, 201, 198, 125, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 201, 166, 128, 0, 201, 134, 131, 0, 201, 102, 134, 0, 201, 70, 137, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 140, 0, 201, 6, 143, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 66, 5, 3, 0, 66, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 145, 0, 201, 198, 148, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 151, 0, 201, 134, 154, 0, 201, 102, 157, 0, 201, 70, 160, 0, 201, 38, 163, 0, 201, 6, 166, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 197, 226, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 168, 0, 201, 38, 172, 0, 201, 102, 175, 0, 201, 166, 178, 0, 201, 230, 181, 0, 201, 38, 185, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 24, 0, 197, 226, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 188, 0, 201, 70, 191, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 194, 0, 201, 6, 197, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 50, 5, 71, 0, 50, 134, 73, 0, 50, 5, 73, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 199, 0, 201, 198, 202, 0, 201, 166, 205, 0, 201, 134, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 5, 0, 197, 34, 6, 0, 198, 130, 8, 0, 198, 2, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 211, 0, 201, 70, 214, 0, 201, 38, 217, 0, 201, 102, 220, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 6, 0, 197, 98, 7, 0, 198, 130, 9, 0, 198, 2, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 223, 0, 201, 230, 226, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 32, 230, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 134, 2, 4, 0, 0, 0, 42, 0, 134, 3, 4, 0, 0, 0, 50, 0, 134, 2, 4, 0, 0, 0, 50, 0, 134, 3, 4, 0, 0, 0, 58, 0, 134, 2, 4, 0, 0, 0, 58, 0, 134, 3, 4, 0, 0, 0, 66, 0, 134, 2, 4, 0, 0, 0, 66, 0, 134, 3, 4, 0, 0, 0, 68, 0, 134, 2, 4, 0, 0, 0, 68, 0, 134, 3, 4, 0, 0, 0, 70, 0, 134, 2, 4, 0, 0, 0, 70, 0, 134, 3, 4, 0, 0, 0, 82, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 2, 4, 0, 0, 0, 82, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 5, 212, 55, 18, 254, 0, 0, 0, 0, 134, 3, 0, 42, 5, 162, 0, 0, 134, 2, 0, 42, 5, 162, 0, 0, 5, 3, 0, 42, 5, 162, 0, 0, 5, 2, 0, 42, 5, 162, 0, 0, 134, 3, 0, 50, 5, 162, 0, 0, 134, 2, 0, 50, 5, 162, 0, 0, 5, 3, 0, 50, 5, 162, 0, 0, 5, 2, 0, 50, 5, 162, 0, 0, 134, 3, 0, 58, 5, 162, 0, 0, 134, 2, 0, 58, 5, 162, 0, 0, 5, 3, 0, 58, 5, 162, 0, 0, 5, 2, 0, 58, 5, 162, 0, 0, 134, 3, 0, 68, 5, 160, 0, 0, 134, 2, 0, 68, 5, 160, 0, 0, 5, 3, 0, 68, 5, 160, 0, 0, 5, 2, 0, 68, 5, 160, 0, 0, 134, 3, 0, 82, 5, 162, 0, 0, 134, 2, 0, 82, 5, 162, 0, 0, 134, 74, 0, 82, 5, 162, 0, 0, 5, 3, 0, 82, 5, 162, 0, 0, 5, 2, 0, 82, 5, 162, 0, 0, 5, 74, 0, 82, 5, 162, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 134, 3, 0, 70, 5, 162, 0, 0, 134, 2, 0, 70, 5, 162, 0, 0, 5, 3, 0, 70, 5, 162, 0, 0, 5, 2, 0, 70, 5, 162, 0, 0, 134, 3, 0, 50, 5, 160, 0, 0, 134, 2, 0, 50, 5, 160, 0, 0, 5, 3, 0, 50, 5, 160, 0, 0, 5, 2, 0, 50, 5, 160, 0, 0, 134, 3, 0, 42, 5, 200, 0, 0, 134, 2, 0, 42, 5, 200, 0, 0, 5, 3, 0, 42, 5, 200, 0, 0, 5, 2, 0, 42, 5, 200, 0, 0, 134, 2, 0, 50, 5, 138, 0, 0, 5, 2, 0, 50, 5, 138, 0, 0, 134, 2, 0, 50, 5, 136, 0, 0, 5, 2, 0, 50, 5, 136, 0, 0, 134, 3, 0, 50, 5, 204, 0, 0, 134, 2, 0, 50, 5, 204, 0, 0, 5, 3, 0, 50, 5, 204, 0, 0, 5, 2, 0, 50, 5, 204, 0, 0, 134, 3, 0, 50, 5, 208, 0, 0, 134, 2, 0, 50, 5, 208, 0, 0, 5, 3, 0, 50, 5, 208, 0, 0, 5, 2, 0, 50, 5, 208, 0, 0, 134, 3, 0, 58, 5, 208, 0, 0, 134, 2, 0, 58, 5, 208, 0, 0, 5, 3, 0, 58, 5, 208, 0, 0, 5, 2, 0, 58, 5, 208, 0, 0, 134, 3, 0, 66, 5, 196, 0, 0, 134, 2, 0, 66, 5, 196, 0, 0, 5, 3, 0, 66, 5, 196, 0, 0, 5, 2, 0, 66, 5, 196, 0, 0, 134, 3, 0, 68, 5, 196, 0, 0, 134, 2, 0, 68, 5, 196, 0, 0, 5, 3, 0, 68, 5, 196, 0, 0, 5, 2, 0, 68, 5, 196, 0, 0, 134, 3, 0, 68, 5, 210, 0, 0, 134, 2, 0, 68, 5, 210, 0, 0, 5, 3, 0, 68, 5, 210, 0, 0, 5, 2, 0, 68, 5, 210, 0, 0, 134, 3, 0, 68, 5, 204, 0, 0, 134, 2, 0, 68, 5, 204, 0, 0, 5, 3, 0, 68, 5, 204, 0, 0, 5, 2, 0, 68, 5, 204, 0, 0, 134, 2, 0, 70, 5, 138, 0, 0, 5, 2, 0, 70, 5, 138, 0, 0, 134, 2, 0, 70, 5, 136, 0, 0, 5, 2, 0, 70, 5, 136, 0, 0, 134, 3, 0, 82, 5, 198, 0, 0, 134, 2, 0, 82, 5, 198, 0, 0, 134, 74, 0, 82, 5, 198, 0, 0, 5, 3, 0, 82, 5, 198, 0, 0, 5, 2, 0, 82, 5, 198, 0, 0, 5, 74, 0, 82, 5, 198, 0, 0, 134, 3, 0, 82, 5, 208, 0, 0, 134, 2, 0, 82, 5, 208, 0, 0, 134, 74, 0, 82, 5, 208, 0, 0, 5, 3, 0, 82, 5, 208, 0, 0, 5, 2, 0, 82, 5, 208, 0, 0, 5, 74, 0, 82, 5, 208, 0, 0, 134, 3, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 204, 0, 0, 134, 74, 0, 82, 5, 204, 0, 0, 5, 3, 0, 82, 5, 204, 0, 0, 5, 2, 0, 82, 5, 204, 0, 0, 5, 74, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 150, 0, 0, 5, 2, 0, 82, 5, 150, 0, 0, 134, 3, 0, 42, 5, 196, 0, 0, 134, 2, 0, 42, 5, 196, 0, 0, 5, 3, 0, 42, 5, 196, 0, 0, 5, 2, 0, 42, 5, 196, 0, 0, 134, 3, 0, 50, 5, 196, 0, 0, 134, 2, 0, 50, 5, 196, 0, 0, 5, 3, 0, 50, 5, 196, 0, 0, 5, 2, 0, 50, 5, 196, 0, 0, 134, 3, 0, 58, 5, 196, 0, 0, 134, 2, 0, 58, 5, 196, 0, 0, 5, 3, 0, 58, 5, 196, 0, 0, 5, 2, 0, 58, 5, 196, 0, 0, 134, 3, 0, 70, 5, 196, 0, 0, 134, 2, 0, 70, 5, 196, 0, 0, 5, 3, 0, 70, 5, 196, 0, 0, 5, 2, 0, 70, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 196, 0, 0, 134, 2, 0, 82, 5, 196, 0, 0, 134, 74, 0, 82, 5, 196, 0, 0, 5, 3, 0, 82, 5, 196, 0, 0, 5, 2, 0, 82, 5, 196, 0, 0, 5, 74, 0, 82, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 5, 5, 0, 19, 5, 5, 0, 21, 5, 5, 0, 23, 5, 5, 0, 25, 5, 5, 0, 27, 5, 5, 0, 29, 5, 5, 0, 31, 5, 5, 0, 33, 5, 5, 0, 35, 5, 5, 0, 37, 0, 42, 156, 5, 5, 0, 12, 3, 7, 0, 12, 3, 198, 161, 64, 3, 198, 129, 65, 3, 0, 234, 134, 3, 0, 3, 198, 129, 1, 3, 0, 234, 134, 3, 4, 3, 0, 234, 134, 2, 0, 50, 156, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 134, 2, 12, 3, 0, 192, 198, 97, 64, 3, 0, 192, 198, 65, 65, 3, 0, 242, 134, 3, 0, 3, 0, 192, 198, 65, 1, 3, 0, 242, 134, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 58, 156, 5, 5, 0, 12, 3, 9, 0, 12, 3, 0, 192, 198, 161, 64, 3, 0, 192, 198, 129, 65, 3, 0, 250, 134, 3, 0, 3, 0, 192, 198, 129, 1, 3, 0, 250, 134, 3, 4, 3, 0, 250, 134, 2, 0, 66, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 0, 192, 198, 225, 64, 3, 0, 192, 198, 193, 65, 3, 255, 255, 0, 66, 134, 3, 0, 3, 0, 192, 198, 193, 1, 3, 255, 255, 0, 66, 134, 3, 4, 3, 255, 255, 0, 66, 134, 2, 0, 68, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 33, 64, 3, 1, 192, 198, 1, 65, 3, 255, 255, 0, 68, 134, 3, 0, 3, 1, 192, 198, 1, 1, 3, 255, 255, 0, 68, 134, 3, 4, 3, 255, 255, 0, 68, 134, 2, 0, 70, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 97, 64, 3, 1, 192, 198, 65, 65, 3, 255, 255, 0, 70, 134, 3, 0, 3, 1, 192, 198, 65, 1, 3, 255, 255, 0, 70, 134, 3, 4, 3, 255, 255, 0, 70, 134, 2, 0, 82, 156, 5, 7, 0, 12, 3, 14, 0, 12, 3, 1, 192, 198, 161, 64, 3, 1, 192, 198, 129, 65, 3, 255, 255, 0, 82, 134, 3, 68, 3, 255, 255, 0, 82, 134, 71, 0, 3, 1, 192, 198, 129, 1, 3, 255, 255, 0, 82, 134, 3, 4, 3, 255, 255, 0, 82, 134, 2, 8, 3, 197, 127, 0, 82, 134, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 0, 42, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 1, 192, 198, 225, 64, 3, 1, 192, 198, 193, 65, 3, 0, 234, 5, 3, 0, 3, 1, 192, 198, 193, 1, 3, 0, 234, 5, 3, 4, 3, 0, 234, 5, 2, 0, 50, 5, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 5, 2, 12, 3, 2, 192, 198, 33, 64, 3, 2, 192, 198, 1, 65, 3, 0, 242, 5, 3, 0, 3, 2, 192, 198, 1, 1, 3, 0, 242, 5, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 58, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 2, 192, 198, 97, 64, 3, 2, 192, 198, 65, 65, 3, 0, 250, 5, 3, 0, 3, 2, 192, 198, 65, 1, 3, 0, 250, 5, 3, 4, 3, 0, 250, 5, 2, 0, 66, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 161, 64, 3, 2, 192, 198, 129, 65, 3, 255, 255, 0, 66, 5, 3, 0, 3, 2, 192, 198, 129, 1, 3, 255, 255, 0, 66, 5, 3, 4, 3, 255, 255, 0, 66, 5, 2, 0, 68, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 225, 64, 3, 2, 192, 198, 193, 65, 3, 255, 255, 0, 68, 5, 3, 0, 3, 2, 192, 198, 193, 1, 3, 255, 255, 0, 68, 5, 3, 4, 3, 255, 255, 0, 68, 5, 2, 0, 70, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 3, 192, 198, 33, 64, 3, 3, 192, 198, 1, 65, 3, 255, 255, 0, 70, 5, 3, 0, 3, 3, 192, 198, 1, 1, 3, 255, 255, 0, 70, 5, 3, 4, 3, 255, 255, 0, 70, 5, 2, 0, 82, 5, 5, 7, 0, 12, 3, 14, 0, 12, 3, 3, 192, 198, 97, 64, 3, 3, 192, 198, 65, 65, 3, 255, 255, 0, 82, 5, 3, 68, 3, 255, 255, 0, 82, 5, 71, 0, 3, 3, 192, 198, 65, 1, 3, 255, 255, 0, 82, 5, 3, 4, 3, 255, 255, 0, 82, 5, 2, 8, 3, 197, 127, 0, 82, 5, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 82, 134, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 5, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 82, 5, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 42, 196, 162, 5, 0, 12, 3, 8, 0, 12, 3, 3, 192, 198, 194, 64, 3, 3, 192, 198, 130, 65, 3, 197, 162, 0, 3, 3, 192, 198, 130, 1, 3, 197, 162, 4, 3, 0, 192, 197, 98, 5, 42, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 66, 64, 3, 4, 192, 198, 2, 65, 3, 0, 192, 197, 162, 0, 3, 4, 192, 198, 2, 1, 3, 0, 192, 197, 162, 4, 3, 0, 192, 197, 226, 156, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 194, 64, 3, 4, 192, 198, 130, 65, 3, 1, 192, 197, 34, 0, 3, 4, 192, 198, 130, 1, 3, 1, 192, 197, 34, 4, 3, 1, 192, 197, 98, 5, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 66, 64, 3, 5, 192, 198, 2, 65, 3, 1, 192, 197, 162, 0, 3, 5, 192, 198, 2, 1, 3, 1, 192, 197, 162, 4, 3, 1, 192, 197, 226, 156, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 194, 64, 3, 5, 192, 198, 130, 65, 3, 2, 192, 197, 34, 0, 3, 5, 192, 198, 130, 1, 3, 2, 192, 197, 34, 4, 3, 2, 192, 197, 98, 5, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 66, 64, 3, 6, 192, 198, 2, 65, 3, 2, 192, 197, 162, 0, 3, 6, 192, 198, 2, 1, 3, 2, 192, 197, 162, 4, 3, 2, 192, 197, 226, 156, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 194, 64, 3, 6, 192, 198, 130, 65, 3, 3, 192, 197, 34, 0, 3, 6, 192, 198, 130, 1, 3, 3, 192, 197, 34, 4, 3, 3, 192, 197, 98, 5, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 7, 192, 198, 66, 64, 3, 7, 192, 198, 2, 65, 3, 3, 192, 197, 162, 0, 3, 7, 192, 198, 2, 1, 3, 3, 192, 197, 162, 4, 3, 3, 192, 197, 226, 156, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 4, 192, 197, 162, 12, 3, 7, 192, 198, 194, 64, 3, 7, 192, 198, 130, 65, 3, 4, 192, 197, 34, 0, 3, 7, 192, 198, 130, 1, 3, 4, 192, 197, 34, 4, 3, 4, 192, 197, 98, 5, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 5, 192, 197, 98, 12, 3, 8, 192, 198, 66, 64, 3, 8, 192, 198, 2, 65, 3, 4, 192, 197, 226, 0, 3, 8, 192, 198, 2, 1, 3, 4, 192, 197, 226, 4, 3, 5, 192, 197, 34, 156, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 8, 192, 198, 194, 64, 3, 8, 192, 198, 130, 65, 3, 5, 192, 197, 162, 0, 3, 8, 192, 198, 130, 1, 3, 5, 192, 197, 162, 4, 3, 5, 192, 197, 226, 5, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 9, 192, 198, 66, 64, 3, 9, 192, 198, 2, 65, 3, 6, 192, 197, 34, 0, 3, 9, 192, 198, 2, 1, 3, 6, 192, 197, 34, 4, 3, 6, 192, 197, 98, 156, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 34, 12, 3, 9, 192, 198, 194, 64, 3, 9, 192, 198, 130, 65, 3, 6, 192, 197, 162, 0, 3, 9, 192, 198, 130, 1, 3, 6, 192, 197, 162, 4, 3, 6, 192, 197, 226, 5, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 226, 12, 3, 10, 192, 198, 66, 64, 3, 10, 192, 198, 2, 65, 3, 7, 192, 197, 98, 0, 3, 10, 192, 198, 2, 1, 3, 7, 192, 197, 98, 4, 3, 7, 192, 197, 162, 156, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 10, 192, 198, 194, 64, 3, 10, 192, 198, 130, 65, 3, 8, 192, 197, 34, 0, 3, 10, 192, 198, 130, 1, 3, 8, 192, 197, 34, 4, 3, 8, 192, 197, 98, 5, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 66, 64, 3, 11, 192, 198, 2, 65, 3, 8, 192, 197, 162, 0, 3, 11, 192, 198, 2, 1, 3, 8, 192, 197, 162, 4, 3, 8, 192, 197, 226, 156, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 194, 64, 3, 11, 192, 198, 130, 65, 3, 9, 192, 197, 34, 0, 3, 11, 192, 198, 130, 1, 3, 9, 192, 197, 34, 4, 3, 9, 192, 197, 98, 5, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 66, 64, 3, 12, 192, 198, 2, 65, 3, 9, 192, 197, 162, 0, 3, 12, 192, 198, 2, 1, 3, 9, 192, 197, 162, 4, 3, 9, 192, 197, 226, 156, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 194, 64, 3, 12, 192, 198, 130, 65, 3, 10, 192, 197, 34, 0, 3, 12, 192, 198, 130, 1, 3, 10, 192, 197, 34, 4, 3, 10, 192, 197, 98, 5, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 66, 64, 3, 13, 192, 198, 2, 65, 3, 10, 192, 197, 162, 0, 3, 13, 192, 198, 2, 1, 3, 10, 192, 197, 162, 4, 3, 10, 192, 197, 226, 156, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 194, 64, 3, 13, 192, 198, 130, 65, 3, 12, 192, 197, 34, 0, 3, 13, 192, 198, 130, 1, 3, 12, 192, 197, 34, 4, 3, 12, 192, 197, 98, 5, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 66, 64, 3, 14, 192, 198, 2, 65, 3, 12, 192, 197, 162, 0, 3, 14, 192, 198, 2, 1, 3, 12, 192, 197, 162, 4, 3, 12, 192, 197, 226, 156, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 194, 64, 3, 14, 192, 198, 130, 65, 3, 13, 192, 197, 34, 0, 3, 14, 192, 198, 130, 1, 3, 13, 192, 197, 34, 4, 3, 13, 192, 197, 98, 5, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 66, 64, 3, 15, 192, 198, 2, 65, 3, 13, 192, 197, 162, 0, 3, 15, 192, 198, 2, 1, 3, 13, 192, 197, 162, 4, 3, 13, 192, 197, 226, 156, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 194, 64, 3, 15, 192, 198, 130, 65, 3, 14, 192, 197, 34, 0, 3, 15, 192, 198, 130, 1, 3, 14, 192, 197, 34, 4, 3, 14, 192, 197, 98, 5, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 66, 64, 3, 16, 192, 198, 2, 65, 3, 14, 192, 197, 162, 0, 3, 16, 192, 198, 2, 1, 3, 14, 192, 197, 162, 4, 3, 14, 192, 197, 226, 156, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 194, 64, 3, 16, 192, 198, 130, 65, 3, 15, 192, 197, 34, 0, 3, 16, 192, 198, 130, 1, 3, 15, 192, 197, 34, 4, 3, 15, 192, 197, 98, 5, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 66, 64, 3, 17, 192, 198, 2, 65, 3, 15, 192, 197, 162, 0, 3, 17, 192, 198, 2, 1, 3, 15, 192, 197, 162, 4, 3, 15, 192, 197, 226, 156, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 194, 64, 3, 17, 192, 198, 130, 65, 3, 16, 192, 197, 34, 0, 3, 17, 192, 198, 130, 1, 3, 16, 192, 197, 34, 4, 3, 16, 192, 197, 98, 5, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 66, 64, 3, 18, 192, 198, 2, 65, 3, 16, 192, 197, 162, 0, 3, 18, 192, 198, 2, 1, 3, 16, 192, 197, 162, 4, 3, 16, 192, 197, 226, 156, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 194, 64, 3, 18, 192, 198, 130, 65, 3, 17, 192, 197, 34, 0, 3, 18, 192, 198, 130, 1, 3, 17, 192, 197, 34, 4, 3, 17, 192, 197, 98, 5, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 66, 64, 3, 19, 192, 198, 2, 65, 3, 17, 192, 197, 162, 0, 3, 19, 192, 198, 2, 1, 3, 17, 192, 197, 162, 4, 3, 17, 192, 197, 226, 156, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 194, 64, 3, 19, 192, 198, 130, 65, 3, 18, 192, 197, 34, 0, 3, 19, 192, 198, 130, 1, 3, 18, 192, 197, 34, 4, 3, 18, 192, 197, 98, 5, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 20, 192, 198, 66, 64, 3, 20, 192, 198, 2, 65, 3, 18, 192, 197, 162, 0, 3, 20, 192, 198, 2, 1, 3, 18, 192, 197, 162, 4, 3, 18, 192, 197, 226, 156, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 20, 192, 197, 162, 12, 3, 20, 192, 198, 194, 64, 3, 20, 192, 198, 130, 65, 3, 20, 192, 197, 34, 0, 3, 20, 192, 198, 130, 1, 3, 20, 192, 197, 34, 4, 3, 20, 192, 197, 98, 5, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 21, 192, 197, 98, 12, 3, 21, 192, 198, 66, 64, 3, 21, 192, 198, 2, 65, 3, 20, 192, 197, 226, 0, 3, 21, 192, 198, 2, 1, 3, 20, 192, 197, 226, 4, 3, 21, 192, 197, 34, 156, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 34, 12, 3, 21, 192, 198, 194, 64, 3, 21, 192, 198, 130, 65, 3, 21, 192, 197, 162, 0, 3, 21, 192, 198, 130, 1, 3, 21, 192, 197, 162, 4, 3, 21, 192, 197, 226, 5, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 226, 12, 3, 22, 192, 198, 66, 64, 3, 22, 192, 198, 2, 65, 3, 22, 192, 197, 98, 0, 3, 22, 192, 198, 2, 1, 3, 22, 192, 197, 98, 4, 3, 22, 192, 197, 162, 156, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 23, 192, 197, 162, 12, 3, 22, 192, 198, 194, 64, 3, 22, 192, 198, 130, 65, 3, 23, 192, 197, 34, 0, 3, 22, 192, 198, 130, 1, 3, 23, 192, 197, 34, 4, 3, 23, 192, 197, 98, 5, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 24, 192, 197, 98, 12, 3, 23, 192, 198, 66, 64, 3, 23, 192, 198, 2, 65, 3, 23, 192, 197, 226, 0, 3, 23, 192, 198, 2, 1, 3, 23, 192, 197, 226, 4, 3, 24, 192, 197, 34, 156, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 23, 192, 198, 194, 64, 3, 23, 192, 198, 130, 65, 3, 25, 192, 197, 34, 0, 3, 23, 192, 198, 130, 1, 3, 25, 192, 197, 34, 4, 3, 25, 192, 197, 98, 5, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 66, 64, 3, 24, 192, 198, 2, 65, 3, 25, 192, 197, 162, 0, 3, 24, 192, 198, 2, 1, 3, 25, 192, 197, 162, 4, 3, 25, 192, 197, 226, 156, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 194, 64, 3, 24, 192, 198, 130, 65, 3, 26, 192, 197, 34, 0, 3, 24, 192, 198, 130, 1, 3, 26, 192, 197, 34, 4, 3, 26, 192, 197, 98, 5, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 66, 64, 3, 25, 192, 198, 2, 65, 3, 26, 192, 197, 162, 0, 3, 25, 192, 198, 2, 1, 3, 26, 192, 197, 162, 4, 3, 26, 192, 197, 226, 156, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 194, 64, 3, 25, 192, 198, 130, 65, 3, 27, 192, 197, 34, 0, 3, 25, 192, 198, 130, 1, 3, 27, 192, 197, 34, 4, 3, 27, 192, 197, 98, 5, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 66, 64, 3, 26, 192, 198, 2, 65, 3, 27, 192, 197, 162, 0, 3, 26, 192, 198, 2, 1, 3, 27, 192, 197, 162, 4, 3, 27, 192, 197, 226, 156, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 194, 64, 3, 26, 192, 198, 130, 65, 3, 28, 192, 197, 34, 0, 3, 26, 192, 198, 130, 1, 3, 28, 192, 197, 34, 4, 3, 28, 192, 197, 98, 5, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 66, 64, 3, 27, 192, 198, 2, 65, 3, 28, 192, 197, 162, 0, 3, 27, 192, 198, 2, 1, 3, 28, 192, 197, 162, 4, 3, 28, 192, 197, 226, 29, 0, 197, 35, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 227, 64, 3, 27, 192, 198, 131, 65, 3, 29, 192, 197, 131, 0, 3, 27, 192, 198, 131, 1, 3, 29, 192, 197, 131, 4, 3, 29, 192, 197, 227, 30, 0, 197, 67, 5, 0, 12, 3, 9, 0, 12, 3, 28, 192, 198, 163, 64, 3, 28, 192, 198, 67, 65, 3, 30, 192, 197, 163, 0, 3, 28, 192, 198, 67, 1, 3, 30, 192, 197, 163, 4, 3, 31, 192, 197, 3, 156, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 31, 192, 197, 226, 12, 3, 29, 192, 198, 66, 64, 3, 29, 192, 198, 2, 65, 3, 31, 192, 197, 98, 0, 3, 29, 192, 198, 2, 1, 3, 31, 192, 197, 98, 4, 3, 31, 192, 197, 162, 5, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 32, 192, 197, 162, 12, 3, 29, 192, 198, 194, 64, 3, 29, 192, 198, 130, 65, 3, 32, 192, 197, 34, 0, 3, 29, 192, 198, 130, 1, 3, 32, 192, 197, 34, 4, 3, 32, 192, 197, 98, 32, 0, 197, 227, 6, 0, 8, 3, 12, 0, 8, 3, 34, 192, 197, 3, 12, 3, 30, 192, 198, 99, 64, 3, 30, 192, 198, 3, 65, 3, 33, 192, 197, 67, 0, 3, 30, 192, 198, 3, 1, 3, 33, 192, 197, 67, 4, 3, 33, 192, 197, 163, 34, 0, 197, 99, 6, 0, 8, 3, 12, 0, 8, 3, 35, 192, 197, 131, 12, 3, 31, 192, 198, 35, 64, 3, 30, 192, 198, 195, 65, 3, 34, 192, 197, 195, 0, 3, 30, 192, 198, 195, 1, 3, 34, 192, 197, 195, 4, 3, 35, 192, 197, 35, 31, 0, 198, 129, 0, 0, 0, 1, 12, 128, 107, 2, 117, 152, 75, 1, 254, 158, 187, 0, 7, 249, 74, 0, 196, 249, 38, 0, 10, 250, 20, 0, 10, 250, 255, 255, 25, 253, 193, 42, 30, 250, 255, 255, 195, 252, 193, 136, 60, 250, 255, 255, 135, 251, 193, 8, 149, 250, 255, 255, 11, 252, 193, 192, 206, 250, 255, 255, 224, 253, 193, 118, 196, 249, 255, 255, 224, 253, 193, 12, 233, 249, 255, 255, 98, 253, 193, 98, 247, 249, 255, 255, 152, 252, 193, 118, 8, 250, 255, 255, 15, 253, 193, 118, 64, 249, 16, 0, 64, 249, 255, 255, 209, 253, 193, 178, 113, 249, 255, 255, 78, 253, 193, 128, 129, 249, 255, 255, 106, 251, 193, 100, 138, 249, 255, 255, 42, 251, 193, 8, 7, 249, 255, 255, 224, 253, 193, 118, 8, 249, 255, 255, 224, 253, 193, 118, 10, 249, 255, 255, 98, 253, 193, 242, 52, 249, 255, 255, 198, 252, 193, 84, 127, 159, 75, 0, 159, 159, 57, 0, 159, 159, 255, 255, 224, 253, 193, 218, 160, 159, 255, 255, 224, 253, 193, 236, 126, 216, 4, 0, 2, 249, 255, 255, 70, 253, 193, 218, 9, 0, 210, 221, 20, 0, 210, 221, 255, 255, 45, 253, 193, 112, 21, 222, 255, 255, 213, 253, 193, 94, 23, 222, 255, 255, 217, 253, 193, 36, 24, 222, 255, 255, 217, 253, 193, 222, 28, 222, 255, 255, 219, 253, 193, 240, 29, 220, 255, 255, 34, 251, 193, 200, 117, 220, 255, 255, 130, 251, 193, 144, 120, 220, 255, 255, 135, 251, 193, 8, 144, 220, 255, 255, 160, 251, 193, 102, 196, 221, 255, 255, 16, 253, 193, 52, 127, 159, 255, 255, 223, 253, 193, 110, 141, 159, 255, 255, 224, 253, 193, 12, 153, 159, 255, 255, 224, 253, 193, 96, 156, 159, 255, 255, 224, 253, 193, 118, 59, 159, 16, 0, 59, 159, 255, 255, 219, 253, 193, 240, 74, 159, 255, 255, 220, 253, 193, 176, 80, 159, 255, 255, 220, 253, 193, 236, 82, 159, 255, 255, 220, 253, 193, 248, 254, 158, 255, 255, 217, 253, 193, 222, 14, 159, 255, 255, 218, 253, 193, 6, 19, 159, 255, 255, 218, 253, 193, 62, 32, 159, 255, 255, 218, 253, 193, 220, 50, 155, 70, 0, 165, 158, 34, 0, 205, 158, 16, 0, 205, 158, 255, 255, 214, 253, 193, 124, 209, 158, 255, 255, 214, 253, 193, 252, 249, 158, 255, 255, 217, 253, 193, 36, 253, 158, 255, 255, 217, 253, 193, 62, 165, 158, 255, 255, 211, 253, 193, 56, 166, 158, 255, 255, 212, 253, 193, 196, 187, 158, 255, 255, 213, 253, 193, 94, 195, 158, 255, 255, 213, 253, 193, 210, 229, 156, 16, 0, 229, 156, 255, 255, 195, 253, 193, 218, 31, 158, 255, 255, 206, 253, 193, 72, 117, 158, 255, 255, 208, 253, 193, 240, 127, 158, 255, 255, 209, 253, 193, 178, 50, 155, 255, 255, 180, 253, 193, 172, 60, 155, 255, 255, 181, 253, 193, 166, 90, 155, 255, 255, 184, 253, 193, 2, 124, 156, 255, 255, 193, 253, 193, 64, 153, 153, 34, 0, 216, 154, 16, 0, 216, 154, 255, 255, 175, 253, 193, 232, 223, 154, 255, 255, 176, 253, 193, 186, 37, 155, 255, 255, 180, 253, 193, 58, 47, 155, 255, 255, 180, 253, 193, 136, 153, 153, 255, 255, 164, 253, 193, 252, 172, 153, 255, 255, 165, 253, 193, 208, 108, 154, 255, 255, 171, 253, 193, 244, 168, 154, 255, 255, 173, 253, 193, 168, 222, 152, 16, 0, 222, 152, 255, 255, 157, 253, 193, 164, 223, 152, 255, 255, 157, 253, 193, 170, 99, 153, 255, 255, 163, 253, 193, 20, 150, 153, 255, 255, 164, 253, 193, 110, 117, 152, 255, 255, 153, 253, 193, 70, 168, 152, 255, 255, 154, 253, 193, 82, 206, 152, 255, 255, 156, 253, 193, 242, 219, 152, 255, 255, 157, 253, 193, 122, 100, 141, 142, 0, 127, 149, 70, 0, 94, 151, 34, 0, 230, 151, 16, 0, 230, 151, 255, 255, 147, 253, 193, 30, 237, 151, 255, 255, 147, 253, 193, 108, 243, 151, 255, 255, 147, 253, 193, 184, 1, 152, 255, 255, 148, 253, 193, 142, 94, 151, 255, 255, 140, 253, 193, 44, 98, 151, 255, 255, 140, 253, 193, 152, 105, 151, 255, 255, 141, 253, 193, 160, 203, 151, 255, 255, 145, 253, 193, 196, 182, 150, 16, 0, 182, 150, 255, 255, 131, 253, 193, 186, 185, 150, 255, 255, 131, 253, 193, 228, 232, 150, 255, 255, 134, 253, 193, 178, 81, 151, 255, 255, 139, 253, 193, 216, 127, 149, 255, 255, 120, 253, 193, 2, 128, 149, 255, 255, 120, 253, 193, 6, 232, 149, 255, 255, 124, 253, 193, 84, 28, 150, 255, 255, 125, 253, 193, 154, 181, 143, 34, 0, 204, 145, 16, 0, 204, 145, 255, 255, 98, 253, 193, 98, 209, 145, 255, 255, 98, 253, 193, 242, 133, 148, 255, 255, 114, 253, 193, 86, 119, 149, 255, 255, 118, 253, 193, 200, 181, 143, 255, 255, 78, 253, 193, 176, 145, 144, 255, 255, 87, 253, 193, 98, 73, 145, 255, 255, 93, 253, 193, 148, 198, 145, 255, 255, 98, 253, 193, 30, 202, 142, 16, 0, 202, 142, 255, 255, 70, 253, 193, 218, 102, 143, 255, 255, 76, 253, 193, 100, 155, 143, 255, 255, 77, 253, 193, 186, 176, 143, 255, 255, 78, 253, 193, 128, 100, 141, 255, 255, 54, 253, 193, 190, 112, 141, 255, 255, 55, 253, 193, 84, 179, 141, 255, 255, 59, 253, 193, 84, 171, 142, 255, 255, 68, 253, 193, 230, 107, 134, 70, 0, 0, 138, 34, 0, 85, 140, 16, 0, 85, 140, 255, 255, 45, 253, 193, 112, 120, 140, 255, 255, 47, 253, 193, 36, 157, 140, 255, 255, 48, 253, 193, 190, 29, 141, 255, 255, 53, 253, 193, 202, 0, 138, 255, 255, 30, 253, 193, 38, 160, 139, 255, 255, 40, 253, 193, 192, 55, 140, 255, 255, 43, 253, 193, 160, 70, 140, 255, 255, 44, 253, 193, 90, 126, 137, 16, 0, 126, 137, 255, 255, 24, 253, 193, 136, 139, 137, 255, 255, 25, 253, 193, 42, 193, 137, 255, 255, 27, 253, 193, 148, 210, 137, 255, 255, 27, 253, 193, 240, 107, 134, 255, 255, 254, 252, 193, 150, 64, 136, 255, 255, 14, 253, 193, 202, 76, 136, 255, 255, 15, 253, 193, 118, 99, 136, 255, 255, 16, 253, 193, 52, 252, 129, 34, 0, 110, 130, 16, 0, 110, 130, 255, 255, 222, 252, 193, 106, 114, 130, 255, 255, 222, 252, 193, 140, 120, 130, 255, 255, 223, 252, 193, 6, 77, 134, 255, 255, 252, 252, 193, 174, 252, 129, 255, 255, 217, 252, 193, 10, 12, 130, 255, 255, 218, 252, 193, 106, 27, 130, 255, 255, 219, 252, 193, 90, 31, 130, 255, 255, 219, 252, 193, 128, 137, 128, 16, 0, 137, 128, 255, 255, 203, 252, 193, 226, 227, 129, 255, 255, 215, 252, 193, 90, 234, 129, 255, 255, 215, 252, 193, 222, 243, 129, 255, 255, 216, 252, 193, 152, 12, 128, 255, 255, 198, 252, 193, 214, 18, 128, 255, 255, 199, 252, 193, 40, 51, 128, 255, 255, 200, 252, 193, 74, 127, 128, 255, 255, 203, 252, 193, 128, 54, 98, 30, 1, 132, 115, 142, 0, 226, 119, 70, 0, 115, 124, 34, 0, 81, 127, 16, 0, 81, 127, 255, 255, 189, 252, 193, 238, 138, 127, 255, 255, 193, 252, 193, 14, 189, 127, 255, 255, 195, 252, 193, 136, 1, 128, 255, 255, 198, 252, 193, 84, 115, 124, 255, 255, 168, 252, 193, 76, 248, 124, 255, 255, 174, 252, 193, 70, 159, 126, 255, 255, 185, 252, 193, 220, 54, 127, 255, 255, 188, 252, 193, 198, 190, 121, 16, 0, 190, 121, 255, 255, 140, 252, 193, 156, 116, 122, 255, 255, 148, 252, 193, 22, 203, 122, 255, 255, 152, 252, 193, 118, 249, 122, 255, 255, 154, 252, 193, 154, 226, 119, 255, 255, 124, 252, 193, 16, 243, 119, 255, 255, 125, 252, 193, 134, 58, 121, 255, 255, 134, 252, 193, 210, 184, 121, 255, 255, 140, 252, 193, 88, 139, 117, 34, 0, 174, 118, 16, 0, 174, 118, 255, 255, 109, 252, 193, 26, 191, 118, 255, 255, 110, 252, 193, 184, 238, 118, 255, 255, 113, 252, 193, 128, 219, 119, 255, 255, 123, 252, 193, 54, 139, 117, 255, 255, 98, 252, 193, 100, 146, 117, 255, 255, 98, 252, 193, 176, 118, 118, 255, 255, 106, 252, 193, 156, 125, 118, 255, 255, 106, 252, 193, 250, 24, 117, 16, 0, 24, 117, 255, 255, 92, 252, 193, 188, 31, 117, 255, 255, 93, 252, 193, 36, 40, 117, 255, 255, 93, 252, 193, 208, 48, 117, 255, 255, 94, 252, 193, 32, 132, 115, 255, 255, 79, 252, 193, 112, 137, 115, 255, 255, 79, 252, 193, 132, 220, 116, 255, 255, 89, 252, 193, 158, 230, 116, 255, 255, 90, 252, 193, 104, 179, 107, 70, 0, 42, 114, 34, 0, 71, 114, 16, 0, 71, 114, 255, 255, 66, 252, 193, 252, 89, 114, 255, 255, 68, 252, 193, 28, 91, 114, 255, 255, 68, 252, 193, 94, 172, 114, 255, 255, 72, 252, 193, 46, 42, 114, 255, 255, 64, 252, 193, 48, 54, 114, 255, 255, 65, 252, 193, 150, 59, 114, 255, 255, 65, 252, 193, 218, 63, 114, 255, 255, 66, 252, 193, 14, 15, 108, 16, 0, 15, 108, 255, 255, 20, 252, 193, 236, 20, 108, 255, 255, 21, 252, 193, 54, 52, 108, 255, 255, 21, 252, 193, 230, 107, 112, 255, 255, 48, 252, 193, 142, 179, 107, 255, 255, 14, 252, 193, 174, 203, 107, 255, 255, 16, 252, 193, 104, 212, 107, 255, 255, 16, 252, 193, 180, 219, 107, 255, 255, 17, 252, 193, 22, 224, 101, 34, 0, 40, 103, 16, 0, 40, 103, 255, 255, 235, 251, 193, 114, 32, 107, 255, 255, 6, 252, 193, 188, 98, 107, 255, 255, 9, 252, 193, 190, 121, 107, 255, 255, 11, 252, 193, 192, 224, 101, 255, 255, 222, 251, 193, 42, 229, 101, 255, 255, 222, 251, 193, 86, 240, 102, 255, 255, 232, 251, 193, 90, 8, 103, 255, 255, 233, 251, 193, 86, 135, 101, 16, 0, 135, 101, 255, 255, 217, 251, 193, 254, 151, 101, 255, 255, 218, 251, 193, 178, 164, 101, 255, 255, 219, 251, 193, 74, 185, 101, 255, 255, 220, 251, 193, 68, 54, 98, 255, 255, 190, 251, 193, 162, 75, 98, 255, 255, 191, 251, 193, 156, 47, 101, 255, 255, 212, 251, 193, 60, 52, 101, 255, 255, 212, 251, 193, 230, 215, 86, 142, 0, 219, 93, 70, 0, 254, 94, 34, 0, 97, 95, 16, 0, 97, 95, 255, 255, 165, 251, 193, 122, 115, 95, 255, 255, 166, 251, 193, 88, 195, 95, 255, 255, 169, 251, 193, 252, 8, 98, 255, 255, 188, 251, 193, 76, 254, 94, 255, 255, 160, 251, 193, 102, 11, 95, 255, 255, 161, 251, 193, 108, 19, 95, 255, 255, 161, 251, 193, 216, 80, 95, 255, 255, 164, 251, 193, 210, 114, 94, 16, 0, 114, 94, 255, 255, 153, 251, 193, 170, 122, 94, 255, 255, 154, 251, 193, 90, 127, 94, 255, 255, 154, 251, 193, 218, 244, 94, 255, 255, 159, 251, 193, 254, 219, 93, 255, 255, 147, 251, 193, 66, 229, 93, 255, 255, 147, 251, 193, 210, 241, 93, 255, 255, 148, 251, 193, 132, 254, 93, 255, 255, 149, 251, 193, 20, 80, 91, 34, 0, 34, 92, 16, 0, 34, 92, 255, 255, 130, 251, 193, 144, 56, 92, 255, 255, 131, 251, 193, 192, 110, 92, 255, 255, 135, 251, 193, 8, 113, 92, 255, 255, 135, 251, 193, 128, 80, 91, 255, 255, 119, 251, 193, 108, 128, 91, 255, 255, 121, 251, 193, 174, 248, 91, 255, 255, 128, 251, 193, 22, 15, 92, 255, 255, 129, 251, 193, 34, 10, 89, 16, 0, 10, 89, 255, 255, 100, 251, 193, 232, 21, 89, 255, 255, 101, 251, 193, 136, 39, 89, 255, 255, 102, 251, 193, 204, 115, 89, 255, 255, 106, 251, 193, 100, 215, 86, 255, 255, 83, 251, 193, 76, 31, 87, 255, 255, 85, 251, 193, 186, 235, 88, 255, 255, 99, 251, 193, 220, 2, 89, 255, 255, 100, 251, 193, 178, 224, 81, 70, 0, 65, 83, 34, 0, 130, 83, 16, 0, 130, 83, 255, 255, 50, 251, 193, 178, 182, 83, 255, 255, 53, 251, 193, 120, 200, 83, 255, 255, 54, 251, 193, 122, 227, 83, 255, 255, 56, 251, 193, 112, 65, 83, 255, 255, 47, 251, 193, 180, 92, 83, 255, 255, 49, 251, 193, 44, 100, 83, 255, 255, 209, 253, 193, 152, 105, 83, 255, 255, 49, 251, 193, 240, 249, 82, 16, 0, 249, 82, 255, 255, 44, 251, 193, 244, 21, 83, 255, 255, 45, 251, 193, 234, 26, 83, 255, 255, 46, 251, 193, 82, 56, 83, 255, 255, 47, 251, 193, 118, 224, 81, 255, 255, 33, 251, 193, 196, 245, 81, 255, 255, 34, 251, 193, 200, 0, 82, 255, 255, 35, 251, 193, 132, 155, 82, 255, 255, 42, 251, 193, 8, 160, 78, 34, 0, 107, 81, 16, 0, 107, 81, 255, 255, 28, 251, 193, 74, 130, 81, 255, 255, 29, 251, 193, 104, 150, 81, 255, 255, 30, 251, 193, 74, 171, 81, 255, 255, 31, 251, 193, 62, 160, 78, 255, 255, 8, 251, 193, 232, 186, 78, 255, 255, 10, 251, 193, 68, 63, 81, 255, 255, 26, 251, 193, 4, 101, 81, 255, 255, 27, 251, 193, 188, 63, 78, 16, 0, 63, 78, 255, 255, 4, 251, 193, 252, 89, 78, 255, 255, 6, 251, 193, 26, 133, 78, 255, 255, 7, 251, 193, 248, 140, 78, 255, 255, 8, 251, 193, 74, 0, 78, 255, 255, 2, 251, 193, 2, 40, 78, 255, 255, 4, 251, 193, 60, 44, 78, 255, 255, 66, 252, 193, 246, 54, 78, 255, 255, 4, 251, 193, 200, 26, 130, 246, 1, 0, 78, 1, 78, 40, 78, 41, 78, 44, 78, 45, 78, 54, 78, 55, 78, 63, 78, 64, 78, 89, 78, 90, 78, 133, 78, 134, 78, 140, 78, 141, 78, 160, 78, 161, 78, 186, 78, 187, 78, 63, 81, 64, 81, 101, 81, 102, 81, 107, 81, 108, 81, 130, 81, 131, 81, 150, 81, 151, 81, 171, 81, 172, 81, 224, 81, 225, 81, 245, 81, 246, 81, 0, 82, 1, 82, 155, 82, 156, 82, 249, 82, 250, 82, 21, 83, 22, 83, 26, 83, 27, 83, 56, 83, 57, 83, 65, 83, 66, 83, 92, 83, 93, 83, 100, 83, 101, 83, 105, 83, 106, 83, 130, 83, 131, 83, 182, 83, 183, 83, 200, 83, 201, 83, 227, 83, 228, 83, 215, 86, 216, 86, 31, 87, 32, 87, 235, 88, 236, 88, 2, 89, 3, 89, 10, 89, 11, 89, 21, 89, 22, 89, 39, 89, 40, 89, 115, 89, 116, 89, 80, 91, 81, 91, 128, 91, 129, 91, 248, 91, 249, 91, 15, 92, 16, 92, 34, 92, 35, 92, 56, 92, 57, 92, 110, 92, 111, 92, 113, 92, 114, 92, 219, 93, 220, 93, 229, 93, 230, 93, 241, 93, 242, 93, 254, 93, 255, 93, 114, 94, 115, 94, 122, 94, 123, 94, 127, 94, 128, 94, 244, 94, 245, 94, 254, 94, 255, 94, 11, 95, 12, 95, 19, 95, 20, 95, 80, 95, 81, 95, 97, 95, 98, 95, 115, 95, 116, 95, 195, 95, 196, 95, 8, 98, 9, 98, 54, 98, 55, 98, 75, 98, 76, 98, 47, 101, 48, 101, 52, 101, 53, 101, 135, 101, 136, 101, 151, 101, 152, 101, 164, 101, 165, 101, 185, 101, 186, 101, 224, 101, 225, 101, 229, 101, 230, 101, 240, 102, 241, 102, 8, 103, 9, 103, 40, 103, 41, 103, 32, 107, 33, 107, 98, 107, 99, 107, 121, 107, 122, 107, 179, 107, 180, 107, 203, 107, 204, 107, 212, 107, 213, 107, 219, 107, 220, 107, 15, 108, 16, 108, 20, 108, 21, 108, 52, 108, 53, 108, 107, 112, 108, 112, 42, 114, 43, 114, 54, 114, 55, 114, 59, 114, 60, 114, 63, 114, 64, 114, 71, 114, 72, 114, 89, 114, 90, 114, 91, 114, 92, 114, 172, 114, 173, 114, 132, 115, 133, 115, 137, 115, 138, 115, 220, 116, 221, 116, 230, 116, 231, 116, 24, 117, 25, 117, 31, 117, 32, 117, 40, 117, 41, 117, 48, 117, 49, 117, 139, 117, 140, 117, 146, 117, 147, 117, 118, 118, 119, 118, 125, 118, 126, 118, 174, 118, 175, 118, 191, 118, 192, 118, 238, 118, 239, 118, 219, 119, 220, 119, 226, 119, 227, 119, 243, 119, 244, 119, 58, 121, 59, 121, 184, 121, 185, 121, 190, 121, 191, 121, 116, 122, 117, 122, 203, 122, 204, 122, 249, 122, 250, 122, 115, 124, 116, 124, 248, 124, 249, 124, 159, 126, 160, 126, 54, 127, 55, 127, 81, 127, 82, 127, 138, 127, 139, 127, 189, 127, 190, 127, 1, 128, 2, 128, 12, 128, 13, 128, 18, 128, 19, 128, 51, 128, 52, 128, 127, 128, 128, 128, 137, 128, 138, 128, 227, 129, 228, 129, 234, 129, 235, 129, 243, 129, 244, 129, 252, 129, 253, 129, 12, 130, 13, 130, 27, 130, 28, 130, 31, 130, 32, 130, 110, 130, 111, 130, 114, 130, 115, 130, 120, 130, 121, 130, 77, 134, 78, 134, 107, 134, 108, 134, 64, 136, 65, 136, 76, 136, 77, 136, 99, 136, 100, 136, 126, 137, 127, 137, 139, 137, 140, 137, 193, 137, 194, 137, 210, 137, 211, 137, 0, 138, 1, 138, 160, 139, 161, 139, 55, 140, 56, 140, 70, 140, 71, 140, 85, 140, 86, 140, 120, 140, 121, 140, 157, 140, 158, 140, 29, 141, 30, 141, 100, 141, 101, 141, 112, 141, 113, 141, 179, 141, 180, 141, 171, 142, 172, 142, 202, 142, 203, 142, 102, 143, 103, 143, 155, 143, 156, 143, 176, 143, 177, 143, 181, 143, 182, 143, 145, 144, 146, 144, 73, 145, 74, 145, 198, 145, 199, 145, 204, 145, 205, 145, 209, 145, 210, 145, 133, 148, 134, 148, 119, 149, 120, 149, 127, 149, 129, 149, 232, 149, 233, 149, 28, 150, 29, 150, 182, 150, 183, 150, 185, 150, 186, 150, 232, 150, 233, 150, 81, 151, 82, 151, 94, 151, 95, 151, 98, 151, 99, 151, 105, 151, 106, 151, 203, 151, 204, 151, 230, 151, 231, 151, 237, 151, 238, 151, 243, 151, 244, 151, 1, 152, 2, 152, 117, 152, 118, 152, 168, 152, 169, 152, 206, 152, 207, 152, 219, 152, 220, 152, 222, 152, 224, 152, 99, 153, 100, 153, 150, 153, 151, 153, 153, 153, 154, 153, 172, 153, 173, 153, 108, 154, 109, 154, 168, 154, 169, 154, 216, 154, 217, 154, 223, 154, 224, 154, 37, 155, 38, 155, 47, 155, 48, 155, 50, 155, 51, 155, 60, 155, 61, 155, 90, 155, 91, 155, 124, 156, 125, 156, 229, 156, 230, 156, 31, 158, 32, 158, 117, 158, 118, 158, 127, 158, 128, 158, 165, 158, 167, 158, 187, 158, 188, 158, 195, 158, 196, 158, 205, 158, 206, 158, 209, 158, 210, 158, 249, 158, 250, 158, 253, 158, 255, 158, 14, 159, 15, 159, 19, 159, 20, 159, 32, 159, 33, 159, 59, 159, 60, 159, 74, 159, 75, 159, 80, 159, 81, 159, 82, 159, 83, 159, 127, 159, 128, 159, 141, 159, 142, 159, 153, 159, 154, 159, 156, 159, 157, 159, 159, 159, 161, 159, 126, 216, 127, 216, 2, 249, 3, 249, 7, 249, 9, 249, 10, 249, 11, 249, 52, 249, 53, 249, 64, 249, 65, 249, 113, 249, 114, 249, 129, 249, 130, 249, 138, 249, 139, 249, 196, 249, 197, 249, 233, 249, 234, 249, 247, 249, 248, 249, 8, 250, 9, 250, 10, 250, 11, 250, 30, 250, 31, 250, 60, 250, 61, 250, 149, 250, 150, 250, 206, 250, 207, 250, 2, 0, 29, 248, 2, 0, 30, 248, 2, 0, 117, 248, 2, 0, 118, 248, 2, 0, 120, 248, 2, 0, 121, 248, 2, 0, 144, 248, 2, 0, 145, 248, 2, 0, 196, 249, 2, 0, 197, 249, 2, 0, 210, 249, 2, 0, 211, 249, 2, 0, 21, 250, 2, 0, 22, 250, 2, 0, 23, 250, 2, 0, 25, 250, 2, 0, 28, 250, 2, 0, 29, 250, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 28, 4, 185, 76, 185, 80, 185, 84, 30, 4, 186, 96, 185, 100, 185, 104, 32, 4, 34, 4, 36, 4, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 38, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 40, 4, 168, 60, 168, 64, 168, 68, 42, 4, 168, 76, 168, 80, 168, 84, 44, 4, 168, 96, 168, 100, 168, 104, 46, 4, 48, 4, 50, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 52, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 89, 56, 57, 56, 217, 57, 121, 58, 57, 58, 25, 58, 0, 8, 217, 66, 121, 72, 57, 72, 54, 4, 58, 74, 89, 88, 57, 88, 218, 89, 58, 90, 217, 68, 122, 114, 89, 120, 57, 120, 217, 121, 121, 122, 57, 122, 136, 14, 185, 122, 89, 152, 57, 152, 217, 153, 56, 4, 153, 169, 185, 176, 2, 8, 72, 56, 40, 56, 200, 57, 104, 58, 40, 58, 8, 58, 4, 8, 200, 66, 104, 72, 40, 72, 58, 4, 40, 74, 72, 88, 40, 88, 200, 89, 40, 90, 200, 68, 104, 114, 72, 120, 40, 120, 200, 121, 104, 122, 40, 122, 128, 14, 168, 122, 72, 152, 40, 152, 200, 153, 60, 4, 136, 169, 168, 176, 40, 170, 25, 56, 8, 56, 185, 57, 168, 57, 62, 4, 64, 4, 153, 65, 136, 65, 217, 65, 200, 65, 153, 66, 136, 66, 249, 65, 232, 65, 249, 69, 232, 69, 25, 71, 8, 71, 25, 72, 8, 72, 186, 73, 168, 73, 154, 74, 136, 74, 66, 4, 68, 4, 89, 72, 72, 72, 217, 81, 200, 81, 185, 81, 168, 81, 153, 82, 136, 82, 217, 82, 200, 82, 217, 85, 200, 85, 25, 87, 8, 87, 122, 90, 104, 90, 25, 88, 8, 88, 186, 89, 168, 89, 70, 4, 72, 4, 154, 90, 168, 92, 6, 8, 8, 8, 218, 97, 200, 97, 217, 102, 200, 102, 168, 132, 153, 105, 136, 105, 217, 106, 200, 106, 249, 105, 232, 105, 57, 107, 40, 107, 25, 107, 8, 107, 25, 112, 8, 112, 74, 4, 76, 4, 57, 112, 40, 112, 10, 8, 185, 116, 168, 116, 25, 120, 8, 120, 185, 121, 168, 121, 89, 122, 72, 122, 12, 8, 14, 8, 153, 137, 136, 137, 217, 138, 200, 138, 249, 137, 232, 137, 154, 141, 136, 141, 218, 141, 200, 141, 218, 142, 200, 142, 250, 141, 232, 141, 217, 146, 200, 146, 249, 145, 232, 145, 185, 148, 168, 148, 121, 154, 104, 154, 25, 152, 8, 152, 185, 153, 168, 153, 25, 154, 8, 154, 89, 154, 72, 154, 78, 4, 80, 4, 217, 161, 200, 161, 217, 169, 200, 169, 57, 170, 153, 173, 136, 173, 153, 174, 136, 174, 249, 173, 232, 173, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 72, 255, 5, 186, 88, 255, 5, 185, 108, 255, 5, 186, 112, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 255, 5, 168, 72, 255, 5, 168, 88, 255, 5, 168, 108, 255, 5, 168, 112, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 218, 73, 255, 5, 217, 152, 255, 5, 200, 73, 255, 5, 200, 152, 255, 5, 249, 58, 255, 5, 232, 58, 255, 5, 250, 74, 255, 5, 232, 74, 255, 5, 250, 90, 255, 5, 232, 90, 255, 5, 218, 114, 255, 5, 200, 114, 255, 5, 249, 154, 255, 5, 232, 154, 255, 1, 3, 0, 68, 0, 59, 0, 32, 0, 175, 208, 2, 16, 13, 0, 0, 96, 1, 0, 0, 96, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 114, 114, 2, 0, 28, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 85, 67, 111, 108, 5, 0, 0, 0, 9, 104, 36, 8, 0, 0, 0, 0, 17, 0, 0, 0, 16, 32, 2, 15, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255, 68, 0, 0, 0, 100, 0, 0, 0, 100, 1, 0, 0, 172, 75, 2, 0, 172, 75, 2, 0, 28, 86, 2, 0, 28, 86, 2, 0, 120, 94, 2, 0, 120, 94, 2, 0, 28, 110, 2, 0, 38, 110, 2, 0, 86, 114, 2, 0, 17, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 41, 127, 0, 0, 96, 125, 0, 0, 123, 44, 0, 118, 123, 125, 0, 0, 126, 169, 255, 0, 254, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 0, 249, 250, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 254, 255, 50, 105, 114, 84, 1, 0, 28, 14, 224, 34, 64, 0, 192, 0, 96, 0, 0, 0, 8, 0, 16, 0, 24, 0, 48, 0, 56, 0, 64, 0, 72, 0, 80, 0, 88, 0, 96, 0, 104, 0, 109, 0, 117, 0, 122, 0, 130, 0, 137, 0, 145, 0, 48, 0, 56, 0, 48, 0, 56, 0, 152, 0, 160, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 184, 0, 191, 0, 199, 0, 204, 0, 48, 0, 212, 0, 220, 0, 228, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 236, 0, 244, 0, 252, 0, 4, 1, 12, 1, 20, 1, 28, 1, 36, 1, 44, 1, 52, 1, 60, 1, 48, 0, 67, 1, 69, 1, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 76, 1, 48, 0, 48, 0, 48, 0, 48, 0, 84, 1, 92, 1, 48, 0, 100, 1, 108, 1, 116, 1, 48, 0, 48, 0, 48, 0, 119, 1, 127, 1, 48, 0, 48, 0, 48, 0, 135, 1, 143, 1, 149, 1, 156, 1, 163, 1, 171, 1, 178, 1, 186, 1, 194, 1, 202, 1, 210, 1, 218, 1, 226, 1, 233, 1, 241, 1, 249, 1, 0, 2, 8, 2, 16, 2, 23, 2, 31, 2, 39, 2, 46, 2, 54, 2, 62, 2, 70, 2, 78, 2, 86, 2, 94, 2, 98, 2, 106, 2, 48, 0, 114, 2, 122, 2, 130, 2, 48, 0, 138, 2, 48, 0, 146, 2, 154, 2, 48, 0, 160, 2, 168, 2, 176, 2, 184, 2, 192, 2, 200, 2, 207, 2, 215, 2, 223, 2, 231, 2, 239, 2, 247, 2, 255, 2, 5, 3, 13, 3, 17, 3, 25, 3, 33, 3, 41, 3, 49, 3, 57, 3, 65, 3, 73, 3, 81, 3, 89, 3, 97, 3, 105, 3, 113, 3, 119, 3, 48, 0, 125, 3, 133, 3, 141, 3, 149, 3, 157, 3, 165, 3, 173, 3, 180, 3, 188, 3, 194, 3, 202, 3, 210, 3, 218, 3, 226, 3, 48, 0, 232, 3, 240, 3, 248, 3, 0, 4, 7, 4, 15, 4, 23, 4, 27, 4, 34, 4, 40, 4, 48, 4, 56, 4, 63, 4, 70, 4, 78, 4, 83, 4, 90, 4, 98, 4, 105, 4, 113, 4, 121, 4, 129, 4, 136, 4, 144, 4, 150, 4, 157, 4, 164, 4, 172, 4, 180, 4, 188, 4, 196, 4, 204, 4, 211, 4, 48, 0, 219, 4, 227, 4, 235, 4, 243, 4, 248, 4, 252, 4, 2, 5, 10, 5, 18, 5, 26, 5, 34, 5, 42, 5, 49, 5, 54, 5, 62, 5, 69, 5, 77, 5, 85, 5, 93, 5, 101, 5, 109, 5, 116, 5, 123, 5, 127, 5, 135, 5, 143, 5, 150, 5, 157, 5, 164, 5, 171, 5, 175, 5, 183, 5, 188, 5, 196, 5, 204, 5, 210, 5, 214, 5, 222, 5, 230, 5, 236, 5, 242, 5, 250, 5, 2, 6, 9, 6, 17, 6, 22, 6, 29, 6, 33, 6, 41, 6, 48, 6, 56, 6, 64, 6, 71, 6, 48, 0, 79, 6, 81, 6, 89, 6, 97, 6, 105, 6, 111, 6, 48, 0, 114, 6, 120, 6, 127, 6, 134, 6, 142, 6, 149, 6, 155, 6, 160, 6, 168, 6, 176, 6, 184, 6, 48, 0, 192, 6, 200, 6, 208, 6, 215, 6, 223, 6, 231, 6, 236, 6, 244, 6, 252, 6, 3, 7, 9, 7, 48, 0, 13, 7, 48, 0, 48, 0, 48, 0, 21, 7, 29, 7, 37, 7, 45, 7, 53, 7, 61, 7, 69, 7, 77, 7, 85, 7, 93, 7, 101, 7, 109, 7, 117, 7, 125, 7, 133, 7, 141, 7, 149, 7, 157, 7, 165, 7, 173, 7, 181, 7, 189, 7, 197, 7, 205, 7, 213, 7, 221, 7, 229, 7, 237, 7, 245, 7, 253, 7, 5, 8, 13, 8, 21, 8, 29, 8, 37, 8, 45, 8, 53, 8, 61, 8, 69, 8, 77, 8, 85, 8, 93, 8, 101, 8, 109, 8, 117, 8, 125, 8, 133, 8, 141, 8, 149, 8, 157, 8, 165, 8, 173, 8, 181, 8, 189, 8, 197, 8, 205, 8, 213, 8, 221, 8, 229, 8, 237, 8, 245, 8, 253, 8, 5, 9, 13, 9, 21, 9, 29, 9, 37, 9, 45, 9, 53, 9, 61, 9, 69, 9, 77, 9, 85, 9, 93, 9, 101, 9, 109, 9, 117, 9, 125, 9, 133, 9, 141, 9, 149, 9, 157, 9, 165, 9, 173, 9, 181, 9, 189, 9, 197, 9, 205, 9, 213, 9, 221, 9, 229, 9, 237, 9, 245, 9, 253, 9, 5, 10, 13, 10, 21, 10, 29, 10, 37, 10, 45, 10, 53, 10, 61, 10, 69, 10, 77, 10, 85, 10, 93, 10, 101, 10, 109, 10, 117, 10, 125, 10, 133, 10, 141, 10, 149, 10, 157, 10, 165, 10, 173, 10, 181, 10, 189, 10, 197, 10, 205, 10, 213, 10, 221, 10, 229, 10, 237, 10, 245, 10, 253, 10, 5, 11, 13, 11, 21, 11, 29, 11, 37, 11, 45, 11, 53, 11, 61, 11, 69, 11, 77, 11, 85, 11, 93, 11, 101, 11, 109, 11, 117, 11, 125, 11, 133, 11, 141, 11, 149, 11, 157, 11, 165, 11, 173, 11, 181, 11, 189, 11, 197, 11, 205, 11, 213, 11, 221, 11, 229, 11, 237, 11, 245, 11, 253, 11, 5, 12, 13, 12, 21, 12, 29, 12, 37, 12, 45, 12, 53, 12, 61, 12, 69, 12, 77, 12, 85, 12, 93, 12, 101, 12, 109, 12, 117, 12, 125, 12, 133, 12, 141, 12, 149, 12, 157, 12, 165, 12, 173, 12, 181, 12, 189, 12, 197, 12, 205, 12, 213, 12, 221, 12, 229, 12, 237, 12, 245, 12, 253, 12, 5, 13, 13, 13, 21, 13, 29, 13, 37, 13, 45, 13, 53, 13, 61, 13, 69, 13, 77, 13, 85, 13, 93, 13, 101, 13, 109, 13, 117, 13, 125, 13, 133, 13, 141, 13, 149, 13, 157, 13, 165, 13, 173, 13, 181, 13, 189, 13, 197, 13, 205, 13, 213, 13, 221, 13, 229, 13, 237, 13, 245, 13, 253, 13, 5, 14, 13, 14, 21, 14, 29, 14, 37, 14, 45, 14, 53, 14, 61, 14, 69, 14, 77, 14, 85, 14, 93, 14, 101, 14, 109, 14, 117, 14, 125, 14, 133, 14, 141, 14, 149, 14, 157, 14, 165, 14, 173, 14, 181, 14, 189, 14, 197, 14, 205, 14, 213, 14, 221, 14, 229, 14, 237, 14, 245, 14, 253, 14, 5, 15, 13, 15, 21, 15, 29, 15, 37, 15, 45, 15, 53, 15, 61, 15, 69, 15, 77, 15, 85, 15, 93, 15, 101, 15, 109, 15, 117, 15, 125, 15, 133, 15, 141, 15, 149, 15, 157, 15, 165, 15, 173, 15, 181, 15, 189, 15, 197, 15, 205, 15, 213, 15, 221, 15, 229, 15, 237, 15, 245, 15, 253, 15, 5, 16, 13, 16, 21, 16, 29, 16, 37, 16, 45, 16, 53, 16, 61, 16, 69, 16, 77, 16, 85, 16, 93, 16, 101, 16, 109, 16, 117, 16, 125, 16, 133, 16, 141, 16, 149, 16, 157, 16, 165, 16, 173, 16, 181, 16, 189, 16, 197, 16, 205, 16, 213, 16, 221, 16, 229, 16, 237, 16, 245, 16, 253, 16, 5, 17, 13, 17, 21, 17, 29, 17, 37, 17, 45, 17, 53, 17, 61, 17, 69, 17, 77, 17, 85, 17, 93, 17, 101, 17, 109, 17, 117, 17, 125, 17, 133, 17, 141, 17, 149, 17, 157, 17, 165, 17, 173, 17, 181, 17, 189, 17, 197, 17, 205, 17, 213, 17, 221, 17, 229, 17, 237, 17, 245, 17, 253, 17, 5, 18, 13, 18, 21, 18, 29, 18, 37, 18, 45, 18, 53, 18, 61, 18, 69, 18, 77, 18, 85, 18, 93, 18, 101, 18, 109, 18, 117, 18, 125, 18, 133, 18, 141, 18, 149, 18, 157, 18, 165, 18, 173, 18, 181, 18, 189, 18, 197, 18, 205, 18, 213, 18, 221, 18, 229, 18, 237, 18, 245, 18, 253, 18, 5, 19, 13, 19, 21, 19, 29, 19, 37, 19, 45, 19, 53, 19, 61, 19, 69, 19, 77, 19, 85, 19, 93, 19, 101, 19, 109, 19, 117, 19, 125, 19, 133, 19, 141, 19, 149, 19, 157, 19, 165, 19, 173, 19, 181, 19, 189, 19, 197, 19, 205, 19, 213, 19, 221, 19, 229, 19, 237, 19, 245, 19, 253, 19, 5, 20, 13, 20, 21, 20, 29, 20, 37, 20, 45, 20, 53, 20, 61, 20, 69, 20, 77, 20, 85, 20, 93, 20, 101, 20, 109, 20, 117, 20, 125, 20, 133, 20, 141, 20, 149, 20, 157, 20, 165, 20, 173, 20, 181, 20, 189, 20, 197, 20, 205, 20, 213, 20, 221, 20, 229, 20, 237, 20, 245, 20, 253, 20, 5, 21, 13, 21, 21, 21, 29, 21, 37, 21, 45, 21, 53, 21, 61, 21, 69, 21, 77, 21, 85, 21, 93, 21, 101, 21, 109, 21, 117, 21, 125, 21, 133, 21, 141, 21, 149, 21, 157, 21, 165, 21, 173, 21, 181, 21, 189, 21, 197, 21, 205, 21, 213, 21, 221, 21, 229, 21, 237, 21, 245, 21, 253, 21, 5, 22, 13, 22, 21, 22, 29, 22, 37, 22, 45, 22, 53, 22, 61, 22, 69, 22, 77, 22, 85, 22, 93, 22, 101, 22, 109, 22, 117, 22, 125, 22, 133, 22, 141, 22, 149, 22, 157, 22, 165, 22, 173, 22, 181, 22, 189, 22, 197, 22, 205, 22, 213, 22, 221, 22, 229, 22, 237, 22, 245, 22, 253, 22, 5, 23, 13, 23, 21, 23, 29, 23, 37, 23, 45, 23, 53, 23, 61, 23, 69, 23, 77, 23, 85, 23, 93, 23, 101, 23, 109, 23, 117, 23, 125, 23, 133, 23, 141, 23, 149, 23, 157, 23, 165, 23, 173, 23, 181, 23, 189, 23, 197, 23, 205, 23, 213, 23, 221, 23, 229, 23, 237, 23, 245, 23, 253, 23, 5, 24, 13, 24, 21, 24, 29, 24, 37, 24, 45, 24, 53, 24, 61, 24, 69, 24, 77, 24, 85, 24, 93, 24, 101, 24, 109, 24, 117, 24, 125, 24, 133, 24, 141, 24, 149, 24, 157, 24, 165, 24, 173, 24, 181, 24, 189, 24, 197, 24, 205, 24, 213, 24, 221, 24, 229, 24, 237, 24, 245, 24, 253, 24, 5, 25, 13, 25, 21, 25, 29, 25, 37, 25, 45, 25, 53, 25, 61, 25, 69, 25, 77, 25, 85, 25, 93, 25, 101, 25, 109, 25, 117, 25, 125, 25, 133, 25, 141, 25, 149, 25, 157, 25, 165, 25, 173, 25, 181, 25, 189, 25, 197, 25, 205, 25, 213, 25, 221, 25, 229, 25, 237, 25, 245, 25, 253, 25, 5, 26, 13, 26, 21, 26, 29, 26, 37, 26, 45, 26, 53, 26, 61, 26, 69, 26, 77, 26, 85, 26, 93, 26, 101, 26, 109, 26, 117, 26, 125, 26, 133, 26, 141, 26, 149, 26, 157, 26, 165, 26, 173, 26, 181, 26, 189, 26, 197, 26, 205, 26, 213, 26, 221, 26, 229, 26, 237, 26, 245, 26, 253, 26, 5, 27, 13, 27, 21, 27, 29, 27, 37, 27, 45, 27, 53, 27, 61, 27, 69, 27, 77, 27, 85, 27, 93, 27, 101, 27, 109, 27, 117, 27, 125, 27, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 193, 34, 176, 0, 48, 0, 48, 0, 168, 0, 201, 34, 209, 34, 215, 34, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 168, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 133, 27, 141, 27, 149, 27, 157, 27, 165, 27, 173, 27, 181, 27, 189, 27, 197, 27, 205, 27, 213, 27, 221, 27, 229, 27, 237, 27, 245, 27, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 252, 27, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 0, 128, 0, 192, 0, 0, 1, 64, 1, 128, 1, 180, 1, 232, 1, 36, 2, 192, 0, 192, 0, 96, 2, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 192, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 128, 8, 64, 0, 188, 8, 246, 8, 54, 9, 118, 9, 182, 9, 240, 9, 48, 10, 111, 10, 174, 10, 238, 10, 45, 11, 107, 11, 169, 11, 233, 11, 39, 12, 102, 12, 166, 12, 223, 12, 27, 13, 90, 13, 154, 13, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 64, 0, 218, 13, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 1, 28, 9, 28, 17, 28, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 24, 28, 48, 0, 30, 28, 48, 0, 35, 28, 48, 0, 48, 0, 48, 0, 48, 0, 41, 28, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 28, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 50, 28, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 57, 28, 63, 28, 48, 0, 66, 28, 48, 0, 48, 0, 72, 28, 48, 0, 48, 0, 48, 0, 48, 0, 78, 28, 82, 28, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 87, 28, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 93, 28, 101, 28, 107, 28, 113, 28, 117, 28, 124, 28, 131, 28, 136, 28, 144, 28, 48, 0, 150, 28, 48, 0, 158, 28, 48, 0, 163, 28, 48, 0, 171, 28, 177, 28, 48, 0, 184, 28, 48, 0, 189, 28, 48, 0, 193, 28, 201, 28, 204, 28, 48, 0, 212, 28, 220, 28, 48, 0, 48, 0, 227, 28, 48, 0, 232, 28, 48, 0, 240, 28, 48, 0, 48, 0, 48, 0, 246, 28, 254, 28, 48, 0, 48, 0, 48, 0, 6, 29, 13, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 15, 29, 48, 0, 48, 0, 48, 0, 23, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 26, 29, 32, 29, 48, 0, 48, 0, 39, 29, 43, 29, 50, 29, 48, 0, 48, 0, 58, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 61, 29, 48, 0, 48, 0, 69, 29, 48, 0, 48, 0, 48, 0, 75, 29, 78, 29, 86, 29, 48, 0, 90, 29, 98, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 106, 29, 48, 0, 48, 0, 114, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 122, 29, 48, 0, 48, 0, 128, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 132, 29, 48, 0, 48, 0, 48, 0, 137, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 145, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 151, 29, 48, 0, 48, 0, 48, 0, 154, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 162, 29, 48, 0, 48, 0, 48, 0, 48, 0, 169, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 172, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 178, 29, 48, 0, 48, 0, 48, 0, 48, 0, 183, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 189, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 196, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 201, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 208, 29, 48, 0, 48, 0, 48, 0, 214, 29, 48, 0, 217, 29, 225, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 230, 29, 238, 29, 48, 0, 241, 29, 249, 29, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 254, 29, 6, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 11, 30, 19, 30, 48, 0, 23, 30, 48, 0, 30, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 34, 30, 48, 0, 48, 0, 48, 0, 41, 30, 48, 0, 48, 0, 49, 30, 48, 0, 48, 0, 48, 0, 50, 30, 48, 0, 48, 0, 48, 0, 48, 0, 58, 30, 64, 30, 72, 30, 48, 0, 79, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 87, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 94, 30, 48, 0, 101, 30, 109, 30, 48, 0, 116, 30, 48, 0, 48, 0, 48, 0, 121, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 125, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 133, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 136, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 142, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 149, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 154, 30, 48, 0, 159, 30, 48, 0, 48, 0, 163, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 169, 30, 48, 0, 48, 0, 176, 30, 48, 0, 48, 0, 48, 0, 180, 30, 48, 0, 48, 0, 48, 0, 187, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 193, 30, 48, 0, 196, 30, 48, 0, 48, 0, 203, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 209, 30, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 217, 30, 48, 0, 48, 0, 48, 0, 48, 0, 225, 30, 48, 0, 48, 0, 229, 30, 48, 0, 48, 0, 48, 0, 236, 30, 48, 0, 244, 30, 246, 30, 254, 30, 3, 31, 48, 0, 11, 31, 15, 31, 20, 31, 48, 0, 48, 0, 26, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 32, 31, 48, 0, 36, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 41, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 31, 52, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 57, 31, 48, 0, 48, 0, 48, 0, 61, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 66, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 73, 31, 48, 0, 81, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 88, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 90, 31, 48, 0, 48, 0, 48, 0, 98, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 103, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 110, 31, 48, 0, 118, 31, 48, 0, 48, 0, 122, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 128, 31, 48, 0, 48, 0, 48, 0, 131, 31, 48, 0, 137, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 144, 31, 48, 0, 147, 31, 48, 0, 48, 0, 48, 0, 155, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 160, 31, 48, 0, 48, 0, 48, 0, 48, 0, 164, 31, 48, 0, 48, 0, 48, 0, 48, 0, 171, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 175, 31, 48, 0, 48, 0, 183, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 188, 31, 195, 31, 198, 31, 48, 0, 48, 0, 48, 0, 204, 31, 48, 0, 48, 0, 48, 0, 211, 31, 48, 0, 219, 31, 48, 0, 225, 31, 48, 0, 232, 31, 48, 0, 237, 31, 244, 31, 248, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 253, 31, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 4, 32, 48, 0, 48, 0, 10, 32, 48, 0, 48, 0, 18, 32, 48, 0, 48, 0, 48, 0, 26, 32, 48, 0, 48, 0, 48, 0, 34, 32, 48, 0, 48, 0, 48, 0, 48, 0, 39, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 45, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 52, 32, 48, 0, 48, 0, 60, 32, 48, 0, 66, 32, 48, 0, 48, 0, 48, 0, 71, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 76, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 83, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 87, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 92, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 98, 32, 48, 0, 48, 0, 48, 0, 48, 0, 104, 32, 48, 0, 48, 0, 108, 32, 48, 0, 115, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 123, 32, 48, 0, 48, 0, 130, 32, 48, 0, 138, 32, 141, 32, 149, 32, 154, 32, 48, 0, 48, 0, 48, 0, 48, 0, 161, 32, 48, 0, 167, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 171, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 176, 32, 183, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 185, 32, 48, 0, 192, 32, 198, 32, 48, 0, 206, 32, 48, 0, 48, 0, 48, 0, 48, 0, 211, 32, 216, 32, 48, 0, 224, 32, 48, 0, 48, 0, 48, 0, 227, 32, 48, 0, 234, 32, 48, 0, 240, 32, 48, 0, 48, 0, 48, 0, 247, 32, 48, 0, 48, 0, 254, 32, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 4, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 10, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 15, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 21, 33, 48, 0, 48, 0, 48, 0, 28, 33, 48, 0, 32, 33, 48, 0, 48, 0, 48, 0, 48, 0, 40, 33, 48, 0, 48, 0, 48, 0, 47, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 49, 33, 57, 33, 48, 0, 48, 0, 48, 0, 48, 0, 64, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 67, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 74, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 80, 33, 48, 0, 88, 33, 93, 33, 48, 0, 97, 33, 105, 33, 113, 33, 48, 0, 48, 0, 118, 33, 124, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 132, 33, 138, 33, 48, 0, 48, 0, 48, 0, 142, 33, 149, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 153, 33, 48, 0, 160, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 166, 33, 48, 0, 48, 0, 48, 0, 48, 0, 171, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 177, 33, 48, 0, 48, 0, 184, 33, 48, 0, 48, 0, 48, 0, 192, 33, 48, 0, 48, 0, 198, 33, 203, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 207, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 212, 33, 48, 0, 219, 33, 48, 0, 48, 0, 226, 33, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 232, 33, 48, 0, 48, 0, 238, 33, 48, 0, 48, 0, 48, 0, 245, 33, 247, 33, 48, 0, 254, 33, 48, 0, 6, 34, 48, 0, 48, 0, 48, 0, 12, 34, 48, 0, 18, 34, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 24, 34, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 31, 34, 48, 0, 39, 34, 43, 34, 49, 34, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 57, 34, 65, 34, 73, 34, 81, 34, 89, 34, 97, 34, 105, 34, 113, 34, 121, 34, 129, 34, 137, 34, 145, 34, 153, 34, 161, 34, 169, 34, 177, 34, 185, 34, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 255, 255, 255, 255, 203, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 4, 3, 5, 5, 6, 3, 5, 5, 8, 3, 5, 5, 10, 3, 5, 5, 12, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 0, 4, 5, 5, 90, 7, 5, 5, 120, 9, 5, 5, 150, 10, 5, 5, 138, 13, 5, 5, 152, 10, 5, 5, 144, 10, 5, 5, 106, 9, 5, 5, 142, 9, 5, 5, 144, 9, 5, 5, 128, 10, 5, 5, 121, 12, 5, 5, 0, 6, 5, 5, 14, 5, 5, 5, 0, 8, 5, 5, 138, 10, 202, 160, 41, 0, 202, 65, 39, 0, 202, 130, 39, 0, 202, 195, 39, 0, 202, 4, 40, 0, 202, 69, 40, 0, 202, 134, 40, 0, 202, 199, 40, 0, 202, 8, 41, 0, 202, 73, 41, 0, 5, 5, 54, 7, 5, 5, 42, 7, 5, 5, 129, 12, 5, 5, 131, 12, 5, 5, 133, 12, 5, 5, 98, 7, 5, 5, 126, 10, 201, 6, 0, 0, 156, 5, 0, 44, 156, 5, 0, 46, 156, 5, 0, 48, 201, 134, 2, 0, 156, 5, 0, 52, 156, 5, 0, 54, 156, 5, 0, 56, 201, 38, 8, 0, 156, 5, 0, 60, 156, 5, 0, 62, 156, 5, 0, 64, 201, 6, 11, 0, 201, 70, 14, 0, 201, 134, 17, 0, 156, 5, 0, 72, 160, 5, 0, 74, 156, 5, 0, 76, 156, 5, 0, 78, 156, 5, 0, 80, 201, 198, 20, 0, 156, 5, 0, 84, 160, 5, 0, 86, 156, 5, 0, 88, 156, 5, 0, 90, 156, 5, 0, 92, 5, 5, 146, 9, 5, 5, 140, 10, 5, 5, 148, 9, 5, 5, 10, 12, 5, 5, 10, 5, 5, 5, 4, 12, 201, 70, 28, 0, 5, 5, 0, 44, 5, 5, 0, 46, 5, 5, 0, 48, 201, 38, 31, 0, 5, 5, 0, 52, 5, 5, 0, 54, 5, 5, 0, 56, 201, 198, 36, 0, 5, 5, 0, 60, 5, 5, 0, 62, 5, 5, 0, 64, 201, 166, 39, 0, 201, 230, 42, 0, 201, 38, 46, 0, 5, 5, 0, 72, 5, 5, 0, 74, 5, 5, 0, 76, 5, 5, 0, 78, 5, 5, 0, 80, 201, 102, 49, 0, 5, 5, 0, 84, 5, 5, 0, 86, 5, 5, 0, 88, 5, 5, 0, 90, 5, 5, 0, 92, 5, 5, 150, 9, 5, 5, 137, 12, 5, 5, 152, 9, 5, 5, 141, 12, 0, 0, 0, 0, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 5, 5, 253, 255, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 0, 0, 134, 3, 0, 42, 196, 142, 156, 42, 196, 154, 156, 42, 196, 150, 156, 42, 196, 146, 156, 42, 197, 34, 66, 0, 196, 160, 156, 46, 198, 65, 0, 0, 134, 3, 0, 50, 201, 230, 56, 0, 196, 150, 156, 50, 198, 129, 0, 0, 134, 3, 0, 58, 196, 142, 156, 58, 196, 150, 156, 58, 160, 112, 0, 48, 196, 154, 156, 68, 198, 65, 1, 0, 134, 3, 0, 70, 196, 142, 156, 70, 196, 154, 156, 70, 196, 150, 156, 70, 5, 5, 127, 12, 196, 158, 156, 70, 198, 129, 1, 0, 134, 3, 0, 82, 196, 142, 156, 82, 201, 198, 59, 0, 196, 136, 156, 90, 160, 5, 38, 93, 197, 98, 66, 0, 198, 193, 1, 0, 5, 3, 0, 42, 196, 142, 5, 42, 196, 154, 5, 42, 196, 150, 5, 42, 196, 146, 5, 42, 197, 162, 66, 0, 196, 160, 5, 46, 198, 1, 2, 0, 5, 3, 0, 50, 201, 102, 63, 0, 196, 150, 5, 50, 198, 65, 2, 0, 5, 3, 0, 58, 196, 142, 5, 58, 196, 150, 5, 58, 5, 112, 0, 48, 196, 154, 5, 68, 198, 1, 3, 0, 5, 3, 0, 70, 196, 142, 5, 70, 196, 154, 5, 70, 196, 150, 5, 70, 5, 5, 125, 12, 196, 158, 5, 70, 198, 65, 3, 0, 5, 3, 0, 82, 196, 142, 5, 82, 201, 70, 66, 0, 196, 136, 5, 90, 5, 5, 38, 93, 196, 150, 5, 90, 134, 2, 0, 42, 5, 2, 0, 42, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 69, 0, 201, 134, 72, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 50, 5, 2, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 75, 0, 201, 70, 78, 0, 198, 97, 0, 0, 198, 33, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 58, 5, 2, 0, 58, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 81, 0, 201, 6, 84, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 68, 5, 3, 0, 68, 201, 230, 86, 0, 201, 198, 89, 0, 198, 33, 1, 0, 198, 225, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 70, 5, 2, 0, 70, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 2, 0, 82, 5, 2, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 92, 0, 201, 230, 95, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 99, 0, 201, 6, 102, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 104, 0, 201, 38, 108, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 33, 0, 0, 198, 225, 1, 0, 198, 161, 0, 0, 198, 97, 2, 0, 198, 97, 1, 0, 198, 33, 3, 0, 198, 161, 1, 0, 198, 97, 3, 0, 134, 70, 0, 82, 5, 70, 0, 82, 134, 71, 0, 82, 5, 71, 0, 82, 134, 72, 0, 82, 5, 72, 0, 82, 134, 73, 0, 82, 5, 73, 0, 82, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 111, 0, 201, 70, 114, 0, 197, 98, 8, 0, 197, 226, 8, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 198, 1, 1, 0, 198, 193, 2, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 117, 0, 201, 6, 120, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 194, 5, 222, 251, 192, 0, 0, 0, 194, 5, 220, 251, 194, 5, 221, 251, 5, 252, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 253, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 254, 0, 0, 5, 255, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 122, 0, 201, 198, 125, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 11, 0, 197, 98, 11, 0, 197, 162, 11, 0, 197, 226, 11, 0, 201, 166, 128, 0, 201, 134, 131, 0, 201, 102, 134, 0, 201, 70, 137, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 140, 0, 201, 6, 143, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 3, 0, 66, 5, 3, 0, 66, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 145, 0, 201, 198, 148, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 151, 0, 201, 134, 154, 0, 201, 102, 157, 0, 201, 70, 160, 0, 201, 38, 163, 0, 201, 6, 166, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 34, 19, 0, 197, 98, 19, 0, 197, 162, 19, 0, 197, 226, 19, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 168, 0, 201, 38, 172, 0, 201, 102, 175, 0, 201, 166, 178, 0, 201, 230, 181, 0, 201, 38, 185, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 24, 0, 197, 226, 24, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 188, 0, 201, 70, 191, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 38, 194, 0, 201, 6, 197, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 134, 71, 0, 50, 5, 71, 0, 50, 134, 73, 0, 50, 5, 73, 0, 50, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 230, 199, 0, 201, 198, 202, 0, 201, 166, 205, 0, 201, 134, 208, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 5, 0, 197, 34, 6, 0, 198, 130, 8, 0, 198, 2, 9, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 102, 211, 0, 201, 70, 214, 0, 201, 38, 217, 0, 201, 102, 220, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 6, 0, 197, 98, 7, 0, 198, 130, 9, 0, 198, 2, 10, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 166, 223, 0, 201, 230, 226, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 241, 192, 0, 0, 0, 5, 5, 80, 208, 192, 0, 0, 0, 5, 5, 124, 175, 192, 0, 0, 0, 5, 5, 148, 163, 192, 0, 0, 0, 5, 5, 199, 214, 5, 5, 200, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 230, 5, 5, 196, 248, 192, 0, 0, 0, 6, 5, 61, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 219, 192, 0, 0, 0, 5, 5, 45, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 227, 192, 0, 0, 0, 5, 5, 34, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 230, 192, 0, 0, 0, 5, 5, 54, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 225, 5, 5, 248, 225, 192, 0, 0, 0, 5, 5, 183, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 174, 5, 5, 7, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 38, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 210, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 9, 5, 82, 239, 6, 5, 128, 194, 6, 5, 188, 221, 6, 5, 64, 168, 7, 5, 209, 239, 6, 5, 114, 207, 9, 5, 67, 239, 6, 5, 210, 179, 7, 5, 83, 230, 6, 5, 25, 239, 6, 5, 217, 230, 7, 5, 0, 128, 6, 5, 81, 208, 6, 5, 85, 171, 6, 5, 50, 166, 6, 5, 123, 202, 6, 5, 149, 210, 6, 5, 125, 175, 6, 5, 22, 187, 6, 5, 248, 163, 6, 5, 220, 164, 6, 5, 145, 173, 6, 5, 104, 213, 8, 5, 225, 226, 6, 5, 149, 163, 6, 5, 125, 203, 6, 5, 26, 225, 6, 5, 101, 235, 6, 5, 93, 242, 6, 5, 142, 195, 6, 5, 255, 245, 7, 5, 237, 181, 6, 5, 22, 227, 6, 5, 140, 218, 6, 5, 246, 236, 6, 5, 165, 212, 6, 5, 47, 175, 7, 5, 212, 184, 6, 5, 204, 231, 6, 5, 173, 171, 6, 5, 34, 235, 6, 5, 201, 214, 6, 5, 26, 242, 6, 5, 183, 226, 6, 5, 185, 223, 6, 5, 37, 228, 6, 5, 45, 226, 6, 5, 186, 194, 6, 5, 124, 202, 6, 5, 159, 205, 6, 5, 151, 192, 6, 5, 139, 241, 6, 5, 172, 194, 6, 5, 204, 243, 6, 5, 216, 194, 6, 5, 254, 239, 6, 5, 187, 194, 6, 5, 146, 202, 6, 5, 38, 228, 6, 5, 79, 223, 6, 5, 241, 215, 6, 5, 130, 191, 6, 5, 53, 199, 7, 5, 14, 228, 6, 5, 55, 218, 6, 5, 212, 168, 7, 5, 19, 247, 6, 5, 197, 175, 6, 5, 161, 205, 6, 5, 146, 173, 6, 5, 12, 245, 7, 5, 8, 230, 6, 5, 188, 248, 8, 5, 197, 248, 7, 5, 80, 172, 6, 5, 164, 210, 6, 5, 141, 218, 6, 5, 73, 175, 6, 5, 59, 229, 6, 5, 13, 245, 6, 5, 221, 164, 6, 5, 19, 170, 6, 5, 23, 227, 6, 5, 21, 209, 7, 5, 208, 229, 7, 5, 178, 199, 6, 5, 219, 219, 6, 5, 186, 174, 6, 5, 155, 241, 6, 5, 57, 167, 6, 5, 123, 168, 6, 5, 240, 240, 6, 5, 46, 184, 6, 5, 100, 212, 6, 5, 132, 217, 6, 5, 48, 248, 6, 5, 103, 193, 6, 5, 152, 245, 6, 5, 102, 192, 7, 5, 8, 229, 6, 5, 24, 250, 6, 5, 71, 181, 6, 5, 22, 168, 6, 5, 255, 182, 6, 5, 36, 163, 6, 5, 155, 163, 6, 5, 234, 167, 6, 5, 249, 171, 6, 5, 83, 172, 6, 5, 20, 170, 6, 5, 10, 227, 6, 5, 227, 226, 6, 5, 30, 227, 6, 5, 33, 230, 6, 5, 83, 197, 6, 5, 55, 217, 6, 5, 26, 187, 6, 5, 165, 221, 6, 5, 68, 171, 6, 5, 86, 171, 6, 5, 230, 172, 6, 5, 88, 247, 6, 5, 15, 244, 6, 5, 6, 248, 6, 5, 193, 185, 6, 5, 26, 239, 6, 5, 117, 185, 6, 5, 53, 239, 6, 5, 51, 248, 6, 5, 56, 230, 6, 5, 179, 224, 6, 5, 189, 231, 6, 5, 178, 218, 6, 5, 157, 204, 6, 5, 177, 227, 6, 5, 64, 226, 6, 5, 254, 219, 6, 5, 172, 192, 6, 5, 222, 235, 6, 5, 46, 234, 6, 5, 217, 198, 6, 5, 166, 226, 6, 5, 78, 217, 6, 5, 139, 216, 6, 5, 87, 239, 6, 5, 22, 241, 6, 5, 78, 205, 6, 5, 27, 204, 6, 5, 157, 242, 6, 5, 35, 193, 6, 5, 198, 175, 6, 5, 12, 227, 6, 5, 184, 218, 6, 5, 201, 163, 6, 5, 84, 223, 7, 5, 116, 232, 6, 5, 217, 232, 6, 5, 137, 228, 6, 5, 174, 223, 6, 5, 244, 215, 6, 5, 183, 224, 6, 5, 250, 225, 6, 5, 22, 240, 6, 5, 71, 242, 6, 5, 234, 165, 6, 5, 250, 186, 7, 5, 164, 205, 6, 5, 35, 221, 6, 5, 163, 170, 6, 5, 195, 174, 6, 5, 43, 187, 6, 5, 81, 222, 6, 5, 8, 248, 6, 5, 89, 211, 6, 5, 153, 172, 6, 5, 217, 171, 6, 5, 165, 191, 6, 5, 12, 246, 6, 5, 149, 204, 6, 5, 118, 243, 6, 5, 88, 241, 6, 5, 199, 173, 6, 5, 154, 172, 6, 5, 241, 226, 6, 5, 17, 228, 6, 5, 41, 216, 6, 5, 34, 169, 6, 5, 44, 193, 6, 5, 4, 192, 6, 5, 143, 165, 6, 5, 202, 175, 6, 5, 41, 225, 6, 5, 170, 191, 6, 5, 13, 194, 6, 5, 189, 247, 6, 5, 33, 184, 6, 5, 231, 189, 6, 5, 18, 190, 6, 5, 182, 169, 6, 5, 23, 169, 6, 5, 49, 201, 6, 5, 110, 229, 6, 5, 207, 197, 6, 5, 174, 218, 6, 5, 210, 171, 6, 5, 204, 177, 6, 5, 57, 193, 6, 5, 114, 229, 6, 5, 219, 164, 6, 5, 218, 208, 6, 5, 77, 223, 6, 5, 248, 190, 6, 5, 241, 193, 6, 5, 233, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 5, 225, 226, 6, 5, 88, 184, 6, 5, 207, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 8, 5, 82, 239, 8, 5, 67, 239, 8, 5, 79, 236, 7, 5, 157, 235, 7, 5, 252, 228, 8, 5, 193, 222, 7, 5, 226, 213, 6, 5, 66, 203, 6, 5, 209, 239, 6, 5, 57, 166, 6, 5, 187, 177, 7, 5, 61, 181, 6, 5, 66, 175, 6, 5, 83, 230, 198, 131, 31, 0, 198, 227, 31, 0, 198, 67, 32, 0, 198, 163, 32, 0, 198, 3, 33, 0, 198, 99, 33, 0, 198, 195, 33, 0, 198, 35, 34, 0, 198, 131, 34, 0, 198, 227, 34, 0, 198, 67, 35, 0, 198, 163, 35, 0, 198, 3, 36, 0, 198, 99, 36, 0, 197, 227, 35, 0, 198, 195, 36, 0, 198, 35, 37, 0, 198, 131, 37, 0, 198, 227, 37, 0, 197, 67, 36, 0, 198, 67, 38, 0, 198, 163, 38, 0, 197, 163, 36, 0, 197, 3, 37, 0, 198, 3, 39, 0, 198, 99, 39, 0, 198, 195, 39, 0, 198, 35, 40, 0, 197, 99, 37, 0, 198, 131, 40, 0, 197, 195, 37, 0, 198, 227, 40, 0, 197, 35, 38, 0, 198, 67, 41, 0, 197, 131, 38, 0, 197, 227, 38, 0, 6, 5, 62, 247, 6, 5, 95, 242, 6, 5, 19, 247, 6, 5, 95, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 7, 5, 82, 239, 7, 5, 67, 239, 7, 5, 79, 236, 6, 5, 157, 235, 6, 5, 43, 245, 6, 5, 99, 188, 6, 5, 113, 208, 6, 5, 0, 128, 6, 5, 141, 204, 6, 5, 225, 226, 7, 5, 197, 248, 6, 5, 178, 199, 6, 5, 208, 229, 6, 5, 80, 172, 6, 5, 164, 205, 6, 5, 237, 181, 6, 5, 8, 230, 6, 5, 136, 221, 7, 5, 68, 242, 6, 5, 188, 227, 6, 5, 26, 172, 6, 5, 94, 179, 6, 5, 11, 234, 6, 5, 226, 221, 6, 5, 168, 185, 6, 5, 95, 171, 6, 5, 95, 183, 6, 5, 212, 184, 6, 5, 93, 227, 6, 5, 19, 242, 6, 5, 233, 243, 6, 5, 214, 221, 6, 5, 104, 216, 6, 5, 236, 214, 6, 5, 45, 214, 6, 5, 112, 221, 6, 5, 252, 228, 7, 5, 193, 222, 6, 5, 226, 213, 7, 5, 251, 232, 7, 5, 94, 242, 6, 5, 88, 239, 6, 5, 87, 233, 6, 5, 58, 217, 6, 5, 227, 204, 6, 5, 250, 208, 6, 5, 135, 231, 6, 5, 254, 213, 6, 5, 83, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 66, 39, 0, 197, 130, 39, 0, 197, 194, 39, 0, 197, 2, 40, 0, 197, 66, 40, 0, 197, 130, 40, 0, 197, 194, 40, 0, 197, 2, 41, 0, 197, 66, 41, 0, 197, 131, 41, 0, 197, 227, 41, 0, 197, 67, 42, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 162, 42, 0, 197, 226, 42, 0, 197, 34, 43, 0, 197, 98, 43, 0, 197, 162, 43, 0, 197, 226, 43, 0, 197, 34, 44, 0, 197, 98, 44, 0, 197, 162, 44, 0, 197, 226, 44, 0, 197, 35, 45, 0, 197, 131, 45, 0, 197, 227, 45, 0, 197, 67, 46, 0, 197, 163, 46, 0, 197, 3, 47, 0, 197, 99, 47, 0, 197, 195, 47, 0, 197, 35, 48, 0, 197, 131, 48, 0, 197, 227, 48, 0, 197, 67, 49, 0, 197, 163, 49, 0, 197, 3, 50, 0, 197, 99, 50, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 194, 50, 0, 197, 2, 51, 0, 197, 66, 51, 0, 197, 130, 51, 0, 197, 196, 51, 0, 197, 66, 52, 0, 197, 130, 52, 0, 197, 194, 52, 0, 197, 2, 53, 0, 197, 66, 53, 0, 197, 130, 53, 0, 197, 194, 53, 0, 197, 2, 54, 0, 197, 66, 54, 0, 197, 131, 54, 0, 197, 227, 54, 0, 197, 67, 55, 0, 197, 163, 55, 0, 197, 3, 56, 0, 197, 99, 56, 0, 197, 195, 56, 0, 197, 35, 57, 0, 197, 131, 57, 0, 197, 227, 57, 0, 197, 67, 58, 0, 197, 163, 58, 0, 197, 3, 59, 0, 197, 99, 59, 0, 197, 195, 59, 0, 197, 35, 60, 0, 197, 131, 60, 0, 197, 227, 60, 0, 197, 67, 61, 0, 197, 163, 61, 0, 197, 3, 62, 0, 197, 99, 62, 0, 192, 0, 0, 0, 5, 5, 212, 209, 5, 5, 118, 181, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 218, 192, 0, 0, 0, 5, 5, 5, 189, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 216, 5, 5, 208, 220, 5, 5, 76, 175, 5, 5, 80, 245, 5, 5, 51, 167, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 30, 169, 5, 5, 166, 210, 5, 5, 7, 240, 192, 0, 0, 0, 5, 5, 42, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 171, 5, 5, 54, 240, 5, 5, 53, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 208, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 180, 5, 5, 173, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 179, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 173, 192, 0, 0, 0, 5, 5, 143, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 212, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 236, 192, 0, 0, 0, 5, 5, 146, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 136, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 213, 192, 0, 0, 0, 5, 5, 164, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 163, 5, 5, 75, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 197, 198, 5, 5, 16, 248, 192, 0, 0, 0, 5, 5, 20, 194, 5, 5, 104, 240, 5, 5, 152, 217, 5, 5, 236, 194, 5, 5, 7, 186, 5, 5, 154, 180, 5, 5, 12, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 33, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 197, 5, 5, 249, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 217, 192, 0, 0, 0, 5, 5, 108, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 199, 5, 5, 36, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 243, 5, 5, 165, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 214, 192, 0, 0, 0, 5, 5, 223, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 196, 5, 5, 4, 199, 192, 0, 0, 0, 5, 5, 200, 213, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 202, 5, 5, 207, 199, 5, 5, 47, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 193, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 217, 192, 0, 0, 0, 5, 5, 20, 233, 5, 5, 144, 240, 5, 5, 192, 233, 192, 0, 0, 0, 5, 5, 149, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 164, 192, 0, 0, 0, 5, 5, 122, 247, 5, 5, 134, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 210, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 164, 192, 0, 0, 0, 5, 5, 117, 206, 192, 0, 0, 0, 5, 5, 138, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 220, 5, 5, 227, 179, 5, 5, 235, 206, 192, 0, 0, 0, 5, 5, 3, 233, 5, 5, 124, 166, 5, 5, 141, 203, 5, 5, 56, 238, 5, 5, 185, 232, 5, 5, 168, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 225, 5, 5, 231, 213, 5, 5, 102, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 194, 5, 5, 24, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 197, 192, 0, 0, 0, 5, 5, 17, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 166, 5, 5, 217, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 167, 192, 0, 0, 0, 5, 5, 142, 203, 5, 5, 94, 239, 5, 5, 73, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 191, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 234, 192, 0, 0, 0, 5, 5, 188, 212, 5, 5, 200, 173, 192, 0, 0, 0, 5, 5, 201, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 179, 5, 5, 65, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 188, 5, 5, 229, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 172, 192, 0, 0, 0, 5, 5, 116, 220, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 149, 238, 192, 0, 0, 0, 5, 5, 180, 210, 192, 0, 0, 0, 5, 5, 215, 234, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 230, 192, 0, 0, 0, 5, 5, 165, 169, 5, 5, 14, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 248, 5, 5, 38, 219, 5, 5, 132, 247, 5, 5, 239, 199, 5, 5, 28, 171, 5, 5, 41, 184, 5, 5, 230, 246, 5, 5, 53, 171, 5, 5, 9, 184, 5, 5, 221, 211, 5, 5, 158, 232, 5, 5, 213, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 197, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 22, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 189, 192, 0, 0, 0, 5, 5, 240, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 202, 5, 5, 222, 164, 192, 0, 0, 0, 5, 5, 114, 230, 5, 5, 30, 201, 5, 5, 6, 173, 5, 5, 160, 191, 5, 5, 22, 196, 5, 5, 207, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 225, 5, 5, 150, 170, 5, 5, 245, 190, 192, 0, 0, 0, 5, 5, 249, 243, 5, 5, 130, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 187, 187, 192, 0, 0, 0, 5, 5, 58, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 187, 5, 5, 26, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 233, 5, 5, 203, 174, 5, 5, 104, 209, 5, 5, 196, 163, 5, 5, 189, 223, 5, 5, 176, 233, 5, 5, 196, 170, 5, 5, 63, 198, 5, 5, 119, 213, 5, 5, 10, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 97, 223, 5, 5, 25, 210, 5, 5, 93, 200, 192, 0, 0, 0, 5, 5, 195, 247, 5, 5, 110, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 240, 5, 5, 120, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 185, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 193, 5, 5, 83, 188, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 231, 5, 5, 3, 165, 5, 5, 155, 245, 192, 0, 0, 0, 5, 5, 180, 187, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 196, 192, 0, 0, 0, 5, 5, 195, 197, 5, 5, 116, 243, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 215, 5, 5, 159, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 167, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 198, 192, 0, 0, 0, 5, 5, 71, 166, 5, 5, 220, 218, 5, 5, 248, 178, 5, 5, 128, 178, 5, 5, 10, 220, 192, 0, 0, 0, 5, 5, 6, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 203, 5, 5, 158, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 237, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 203, 5, 5, 99, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 199, 192, 0, 0, 0, 5, 5, 181, 241, 5, 5, 252, 233, 5, 5, 24, 194, 5, 5, 113, 205, 5, 5, 47, 205, 5, 5, 146, 175, 5, 5, 233, 205, 5, 5, 51, 169, 5, 5, 101, 200, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 210, 192, 0, 0, 0, 5, 5, 148, 182, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 186, 192, 0, 0, 0, 5, 5, 223, 233, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 214, 5, 5, 125, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 210, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 210, 192, 0, 0, 0, 5, 5, 93, 197, 5, 5, 81, 174, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 215, 5, 5, 36, 194, 5, 5, 104, 197, 5, 5, 122, 212, 5, 5, 126, 198, 5, 5, 210, 182, 5, 5, 85, 163, 5, 5, 79, 228, 5, 5, 238, 164, 5, 5, 19, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 195, 5, 5, 222, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 142, 186, 192, 0, 0, 0, 5, 5, 217, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 247, 5, 5, 117, 179, 5, 5, 158, 224, 5, 5, 137, 170, 5, 5, 11, 172, 5, 5, 230, 200, 5, 5, 2, 247, 5, 5, 181, 184, 5, 5, 231, 200, 5, 5, 32, 198, 5, 5, 170, 206, 5, 5, 104, 163, 5, 5, 155, 215, 5, 5, 94, 187, 5, 5, 234, 205, 192, 0, 0, 0, 5, 5, 219, 170, 5, 5, 98, 168, 5, 5, 228, 197, 5, 5, 47, 244, 192, 0, 0, 0, 5, 5, 224, 215, 5, 5, 188, 236, 5, 5, 100, 246, 5, 5, 200, 223, 5, 5, 4, 213, 5, 5, 243, 205, 5, 5, 139, 234, 5, 5, 166, 197, 5, 5, 87, 173, 5, 5, 224, 227, 5, 5, 145, 189, 192, 0, 0, 0, 5, 5, 150, 178, 5, 5, 149, 208, 5, 5, 10, 169, 5, 5, 235, 248, 5, 5, 126, 163, 192, 0, 0, 0, 5, 5, 135, 200, 5, 5, 61, 177, 5, 5, 81, 243, 5, 5, 106, 207, 5, 5, 113, 204, 5, 5, 121, 183, 5, 5, 214, 187, 5, 5, 157, 247, 5, 5, 235, 180, 5, 5, 68, 181, 5, 5, 52, 245, 5, 5, 182, 201, 5, 5, 190, 214, 5, 5, 254, 197, 192, 0, 0, 0, 5, 5, 17, 181, 5, 5, 89, 221, 192, 0, 0, 0, 5, 5, 37, 201, 5, 5, 230, 174, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 182, 192, 0, 0, 0, 5, 5, 175, 230, 5, 5, 25, 204, 192, 0, 0, 0, 5, 5, 23, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 92, 249, 5, 5, 125, 207, 5, 5, 109, 199, 192, 0, 0, 0, 5, 5, 158, 164, 5, 5, 86, 170, 192, 0, 0, 0, 5, 5, 185, 192, 5, 5, 53, 246, 192, 0, 0, 0, 5, 5, 255, 169, 5, 5, 171, 235, 5, 5, 241, 165, 5, 5, 189, 189, 5, 5, 163, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 191, 5, 5, 135, 219, 5, 5, 66, 191, 5, 5, 222, 166, 5, 5, 54, 231, 5, 5, 11, 210, 5, 5, 188, 187, 5, 5, 222, 194, 5, 5, 142, 215, 5, 5, 120, 213, 5, 5, 253, 215, 192, 0, 0, 0, 5, 5, 31, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 214, 5, 5, 195, 187, 5, 5, 34, 174, 5, 5, 179, 234, 5, 5, 37, 222, 5, 5, 0, 150, 5, 5, 15, 233, 5, 5, 141, 219, 5, 5, 26, 233, 5, 5, 110, 197, 5, 5, 70, 202, 192, 0, 0, 0, 5, 5, 207, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 181, 5, 5, 100, 215, 5, 5, 231, 242, 5, 5, 72, 180, 5, 5, 61, 179, 5, 5, 215, 176, 5, 5, 154, 207, 5, 5, 18, 239, 5, 5, 93, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 203, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 175, 192, 0, 0, 0, 5, 5, 92, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 205, 5, 5, 122, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 186, 5, 5, 127, 190, 5, 5, 202, 228, 192, 0, 0, 0, 5, 5, 142, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 227, 192, 0, 0, 0, 5, 5, 94, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 229, 5, 5, 111, 186, 192, 0, 0, 0, 5, 5, 217, 212, 192, 0, 0, 0, 5, 5, 132, 208, 192, 0, 0, 0, 5, 5, 193, 188, 5, 5, 245, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 87, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 169, 5, 5, 203, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 209, 192, 0, 0, 0, 5, 5, 129, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 250, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 235, 5, 5, 31, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 217, 192, 0, 0, 0, 5, 5, 118, 205, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 185, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 240, 5, 5, 119, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 245, 187, 192, 0, 0, 0, 5, 5, 120, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 156, 243, 192, 0, 0, 0, 5, 5, 208, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 174, 5, 5, 16, 200, 192, 0, 0, 0, 5, 5, 130, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 64, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 179, 5, 5, 110, 198, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 185, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 160, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 46, 181, 192, 0, 0, 0, 5, 5, 232, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 230, 192, 0, 0, 0, 5, 5, 87, 163, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 235, 5, 5, 156, 204, 5, 5, 221, 168, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 170, 192, 0, 0, 0, 5, 5, 207, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 175, 5, 5, 230, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 166, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 47, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 114, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 31, 242, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 204, 192, 5, 5, 43, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 179, 5, 5, 14, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 173, 5, 5, 111, 242, 5, 5, 204, 238, 5, 5, 77, 182, 5, 5, 228, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 217, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 213, 192, 0, 0, 0, 5, 5, 141, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 190, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 126, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 10, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 247, 238, 192, 0, 0, 0, 5, 5, 136, 212, 192, 0, 0, 0, 5, 5, 221, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 238, 192, 0, 0, 0, 5, 5, 128, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 193, 192, 0, 0, 0, 5, 5, 160, 204, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 221, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 185, 5, 5, 137, 230, 5, 5, 61, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 225, 5, 5, 62, 208, 192, 0, 0, 0, 5, 5, 81, 246, 5, 5, 111, 240, 5, 5, 147, 219, 5, 5, 184, 242, 192, 0, 0, 0, 5, 5, 83, 236, 5, 5, 210, 190, 5, 5, 200, 168, 5, 5, 223, 219, 5, 5, 97, 198, 5, 5, 118, 248, 5, 5, 106, 175, 5, 5, 240, 219, 5, 5, 177, 172, 5, 5, 162, 227, 5, 5, 133, 189, 5, 5, 67, 179, 5, 5, 179, 211, 5, 5, 208, 170, 5, 5, 107, 241, 5, 5, 11, 164, 5, 5, 51, 194, 5, 5, 126, 193, 5, 5, 116, 183, 5, 5, 186, 191, 192, 0, 0, 0, 5, 5, 3, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 179, 192, 0, 0, 0, 5, 5, 22, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 205, 200, 192, 0, 0, 0, 5, 5, 10, 236, 5, 5, 123, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 213, 5, 5, 254, 182, 192, 0, 0, 0, 5, 5, 56, 236, 5, 5, 94, 207, 192, 0, 0, 0, 5, 5, 240, 227, 5, 5, 164, 179, 5, 5, 78, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 229, 5, 5, 147, 189, 5, 5, 129, 185, 5, 5, 47, 206, 5, 5, 157, 186, 5, 5, 35, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 180, 5, 5, 109, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 158, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 230, 192, 0, 0, 0, 5, 5, 247, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 209, 5, 5, 239, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 68, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 174, 5, 5, 42, 239, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 181, 5, 5, 119, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 185, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 99, 184, 5, 5, 136, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 195, 5, 5, 151, 241, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 137, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 240, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 175, 247, 5, 5, 42, 242, 5, 5, 150, 177, 5, 5, 116, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 219, 5, 5, 18, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 25, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 202, 5, 5, 44, 222, 5, 5, 23, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 247, 5, 5, 20, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 20, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 247, 5, 5, 52, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 246, 170, 5, 5, 51, 168, 5, 5, 247, 170, 5, 5, 78, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 234, 175, 5, 5, 203, 173, 192, 0, 0, 0, 5, 5, 17, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 171, 192, 0, 0, 0, 5, 5, 21, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 35, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 239, 232, 5, 5, 195, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 187, 192, 0, 0, 0, 5, 5, 215, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 191, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 178, 192, 0, 0, 0, 5, 5, 84, 238, 192, 0, 0, 0, 5, 5, 82, 233, 5, 5, 193, 196, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 243, 192, 0, 0, 0, 5, 5, 33, 236, 5, 5, 218, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 208, 5, 5, 38, 203, 192, 0, 0, 0, 5, 5, 130, 215, 5, 5, 156, 170, 192, 0, 0, 0, 5, 5, 99, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 214, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 207, 192, 0, 0, 0, 5, 5, 120, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 3, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 174, 5, 5, 198, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 172, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 241, 5, 5, 170, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 61, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 183, 5, 5, 178, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 208, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 139, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 202, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 180, 192, 0, 0, 0, 5, 5, 183, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 170, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 28, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 231, 5, 5, 219, 248, 5, 5, 151, 207, 5, 5, 228, 195, 5, 5, 42, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 243, 5, 5, 154, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 141, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 133, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 249, 192, 0, 0, 0, 5, 5, 173, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 81, 195, 5, 5, 25, 194, 5, 5, 225, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 178, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 215, 192, 0, 0, 0, 5, 5, 170, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 234, 5, 5, 193, 210, 192, 0, 0, 0, 5, 5, 118, 228, 192, 0, 0, 0, 5, 5, 21, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 246, 192, 0, 0, 0, 5, 5, 151, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 173, 215, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 200, 192, 0, 0, 0, 5, 5, 241, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 176, 5, 5, 125, 245, 5, 5, 234, 209, 192, 0, 0, 0, 5, 5, 79, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 105, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 176, 192, 0, 0, 0, 5, 5, 112, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 176, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 246, 5, 5, 93, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 171, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 223, 5, 5, 87, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 213, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 237, 192, 0, 0, 0, 5, 5, 34, 164, 5, 5, 33, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 163, 5, 5, 202, 164, 192, 0, 0, 0, 5, 5, 243, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 174, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 222, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 224, 5, 5, 253, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 165, 5, 5, 24, 191, 192, 0, 0, 0, 5, 5, 27, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 186, 245, 5, 5, 221, 172, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 236, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 236, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 247, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 0, 153, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 228, 192, 0, 0, 0, 5, 5, 229, 194, 5, 5, 200, 239, 5, 5, 249, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 249, 192, 0, 0, 0, 5, 5, 197, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 8, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 80, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 39, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 9, 181, 192, 0, 0, 0, 5, 5, 180, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 73, 215, 192, 0, 0, 0, 5, 5, 107, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 82, 239, 5, 5, 187, 177, 5, 5, 117, 195, 5, 5, 113, 208, 5, 5, 251, 228, 5, 5, 225, 213, 5, 5, 103, 198, 5, 5, 231, 246, 5, 5, 40, 221, 5, 5, 79, 236, 5, 5, 252, 228, 5, 5, 226, 213, 5, 5, 194, 201, 5, 5, 118, 166, 5, 5, 255, 247, 5, 5, 190, 171, 5, 5, 231, 191, 5, 5, 52, 224, 5, 5, 53, 224, 5, 5, 113, 222, 5, 5, 93, 209, 5, 5, 195, 167, 5, 5, 26, 227, 5, 5, 25, 227, 5, 5, 213, 209, 5, 5, 57, 166, 5, 5, 78, 241, 5, 5, 75, 235, 5, 5, 222, 178, 5, 5, 103, 235, 5, 5, 85, 225, 5, 5, 118, 177, 5, 5, 214, 209, 5, 5, 55, 189, 5, 5, 119, 177, 5, 5, 69, 242, 5, 5, 56, 189, 5, 5, 156, 242, 5, 5, 83, 166, 5, 5, 123, 236, 5, 5, 128, 194, 5, 5, 119, 204, 5, 5, 202, 191, 5, 5, 220, 240, 5, 5, 56, 211, 5, 5, 193, 222, 5, 5, 125, 202, 5, 5, 203, 203, 5, 5, 185, 173, 5, 5, 96, 194, 5, 5, 71, 226, 5, 5, 124, 224, 5, 5, 220, 188, 5, 5, 29, 222, 5, 5, 188, 221, 5, 5, 75, 197, 5, 5, 184, 246, 5, 5, 225, 175, 5, 5, 115, 246, 5, 5, 189, 221, 5, 5, 115, 206, 5, 5, 29, 187, 5, 5, 20, 207, 5, 5, 64, 168, 5, 5, 56, 174, 5, 5, 126, 239, 5, 5, 250, 239, 5, 5, 11, 183, 5, 5, 133, 245, 5, 5, 142, 204, 5, 5, 143, 204, 5, 5, 249, 181, 5, 5, 158, 169, 5, 5, 252, 239, 5, 5, 83, 239, 5, 5, 54, 218, 5, 5, 240, 244, 5, 5, 105, 219, 5, 5, 214, 198, 5, 5, 120, 172, 5, 5, 13, 185, 5, 5, 159, 243, 5, 5, 156, 168, 5, 5, 106, 167, 5, 5, 157, 209, 5, 5, 40, 199, 5, 5, 194, 193, 5, 5, 92, 225, 5, 5, 98, 225, 5, 5, 209, 239, 5, 5, 202, 243, 5, 5, 48, 241, 5, 5, 124, 171, 5, 5, 141, 204, 5, 5, 249, 208, 5, 5, 70, 241, 5, 5, 45, 213, 5, 5, 38, 216, 5, 5, 232, 191, 5, 5, 144, 204, 5, 5, 227, 213, 5, 5, 50, 199, 5, 5, 58, 229, 5, 5, 223, 175, 5, 5, 11, 227, 5, 5, 201, 201, 5, 5, 127, 183, 5, 5, 254, 202, 5, 5, 39, 207, 5, 5, 228, 226, 5, 5, 46, 170, 5, 5, 215, 198, 5, 5, 168, 169, 5, 5, 186, 190, 5, 5, 105, 231, 5, 5, 213, 230, 5, 5, 57, 217, 5, 5, 214, 242, 5, 5, 153, 174, 5, 5, 122, 227, 5, 5, 229, 182, 5, 5, 114, 192, 5, 5, 221, 236, 5, 5, 170, 247, 5, 5, 3, 235, 5, 5, 165, 219, 5, 5, 120, 192, 5, 5, 216, 218, 5, 5, 219, 193, 5, 5, 122, 192, 5, 5, 188, 190, 5, 5, 8, 189, 5, 5, 102, 240, 5, 5, 114, 207, 5, 5, 31, 185, 5, 5, 60, 169, 5, 5, 2, 248, 5, 5, 62, 221, 5, 5, 35, 227, 5, 5, 39, 227, 5, 5, 67, 239, 5, 5, 201, 225, 5, 5, 154, 247, 5, 5, 101, 196, 5, 5, 153, 247, 5, 5, 126, 249, 5, 5, 51, 199, 5, 5, 154, 208, 5, 5, 43, 245, 5, 5, 114, 206, 5, 5, 156, 235, 5, 5, 20, 237, 5, 5, 173, 192, 5, 5, 174, 192, 5, 5, 21, 241, 5, 5, 242, 213, 5, 5, 24, 241, 5, 5, 160, 208, 5, 5, 26, 241, 5, 5, 51, 202, 5, 5, 210, 179, 5, 5, 78, 247, 5, 5, 229, 195, 5, 5, 52, 179, 5, 5, 236, 203, 5, 5, 193, 197, 5, 5, 6, 240, 5, 5, 121, 224, 5, 5, 192, 198, 5, 5, 62, 172, 5, 5, 138, 241, 5, 5, 76, 216, 5, 5, 78, 206, 5, 5, 145, 181, 5, 5, 68, 189, 5, 5, 77, 216, 5, 5, 81, 206, 5, 5, 86, 241, 5, 5, 200, 210, 5, 5, 77, 163, 5, 5, 113, 242, 5, 5, 71, 214, 5, 5, 13, 176, 5, 5, 125, 188, 5, 5, 92, 178, 5, 5, 169, 170, 5, 5, 83, 230, 5, 5, 84, 230, 5, 5, 39, 202, 5, 5, 250, 202, 5, 5, 79, 247, 5, 5, 253, 239, 5, 5, 177, 228, 5, 5, 85, 230, 5, 5, 11, 185, 5, 5, 188, 177, 5, 5, 20, 232, 5, 5, 182, 205, 5, 5, 211, 168, 5, 5, 16, 224, 5, 5, 0, 129, 5, 5, 33, 221, 5, 5, 160, 205, 5, 5, 204, 203, 5, 5, 51, 166, 5, 5, 173, 230, 5, 5, 74, 235, 5, 5, 148, 172, 5, 5, 87, 236, 5, 5, 191, 190, 5, 5, 52, 166, 5, 5, 110, 234, 5, 5, 166, 231, 5, 5, 27, 227, 5, 5, 30, 179, 5, 5, 41, 221, 5, 5, 188, 174, 5, 5, 34, 215, 5, 5, 35, 215, 5, 5, 250, 181, 5, 5, 127, 201, 5, 5, 161, 182, 5, 5, 108, 230, 5, 5, 51, 210, 5, 5, 132, 192, 5, 5, 131, 191, 5, 5, 65, 163, 5, 5, 77, 175, 5, 5, 174, 189, 5, 5, 211, 239, 5, 5, 5, 224, 5, 5, 2, 225, 5, 5, 197, 235, 5, 5, 5, 229, 5, 5, 127, 239, 5, 5, 81, 172, 5, 5, 178, 170, 5, 5, 109, 230, 5, 5, 67, 203, 5, 5, 6, 224, 5, 5, 57, 244, 5, 5, 114, 210, 5, 5, 239, 222, 5, 5, 23, 168, 5, 5, 187, 245, 5, 5, 45, 245, 5, 5, 77, 205, 5, 5, 86, 203, 5, 5, 209, 241, 5, 5, 187, 173, 5, 5, 111, 234, 5, 5, 111, 230, 5, 5, 82, 247, 5, 5, 127, 173, 5, 5, 148, 176, 5, 5, 166, 173, 5, 5, 194, 222, 5, 5, 250, 208, 5, 5, 245, 166, 5, 5, 155, 247, 5, 5, 96, 177, 5, 5, 203, 178, 5, 5, 83, 245, 5, 5, 8, 240, 5, 5, 17, 216, 5, 5, 230, 195, 5, 5, 86, 239, 5, 5, 41, 202, 5, 5, 54, 238, 5, 5, 46, 245, 5, 5, 151, 202, 5, 5, 59, 174, 5, 5, 121, 172, 5, 5, 236, 214, 5, 5, 207, 205, 5, 5, 196, 167, 5, 5, 2, 176, 5, 5, 18, 174, 5, 5, 110, 180, 5, 5, 240, 222, 5, 5, 7, 242, 5, 5, 179, 199, 5, 5, 66, 200, 5, 5, 3, 248, 5, 5, 229, 234, 5, 5, 52, 226, 5, 5, 88, 236, 5, 5, 48, 246, 5, 5, 53, 226, 5, 5, 173, 223, 5, 5, 241, 240, 5, 5, 178, 215, 5, 5, 228, 228, 5, 5, 237, 224, 5, 5, 192, 190, 5, 5, 112, 234, 5, 5, 254, 217, 5, 5, 19, 216, 5, 5, 49, 246, 5, 5, 207, 221, 5, 5, 27, 232, 5, 5, 82, 215, 5, 5, 146, 184, 5, 5, 66, 163, 5, 5, 253, 192, 5, 5, 180, 183, 5, 5, 201, 183, 5, 5, 49, 214, 5, 5, 99, 164, 5, 5, 242, 216, 5, 5, 84, 189, 5, 5, 43, 220, 5, 5, 136, 228, 5, 5, 151, 211, 5, 5, 210, 233, 5, 5, 175, 164, 5, 5, 36, 227, 5, 5, 2, 203, 5, 5, 197, 167, 5, 5, 11, 240, 5, 5, 160, 235, 5, 5, 213, 239, 5, 5, 63, 221, 5, 5, 160, 177, 5, 5, 60, 198, 5, 5, 175, 169, 5, 5, 21, 176, 5, 5, 208, 221, 5, 5, 120, 166, 5, 5, 152, 211, 5, 5, 225, 164, 5, 5, 205, 219, 5, 5, 202, 233, 5, 5, 117, 246, 5, 5, 149, 176, 5, 5, 209, 221, 5, 5, 252, 232, 5, 5, 96, 242, 5, 5, 58, 244, 5, 5, 203, 180, 5, 5, 133, 220, 5, 5, 85, 197, 5, 5, 248, 164, 5, 5, 20, 182, 5, 5, 178, 227, 5, 5, 159, 247, 5, 5, 12, 240, 5, 5, 62, 174, 5, 5, 255, 232, 5, 5, 54, 192, 5, 5, 136, 184, 5, 5, 163, 182, 5, 5, 181, 183, 5, 5, 40, 215, 5, 5, 198, 211, 5, 5, 210, 249, 5, 5, 153, 245, 5, 5, 58, 210, 5, 5, 109, 227, 5, 5, 245, 194, 5, 5, 65, 164, 5, 5, 247, 166, 5, 5, 41, 200, 5, 5, 152, 197, 5, 5, 194, 185, 5, 5, 66, 216, 5, 5, 161, 191, 5, 5, 22, 244, 5, 5, 157, 163, 5, 5, 138, 172, 5, 5, 49, 172, 5, 5, 4, 203, 5, 5, 70, 239, 5, 5, 84, 166, 5, 5, 52, 202, 5, 5, 166, 198, 5, 5, 174, 199, 5, 5, 255, 193, 5, 5, 54, 212, 5, 5, 12, 181, 5, 5, 28, 204, 5, 5, 211, 233, 5, 5, 23, 240, 5, 5, 13, 227, 5, 5, 142, 216, 5, 5, 138, 228, 5, 5, 3, 182, 5, 5, 171, 195, 5, 5, 100, 218, 5, 5, 212, 191, 5, 5, 35, 185, 5, 5, 136, 239, 5, 5, 60, 223, 5, 5, 42, 196, 5, 5, 155, 194, 5, 5, 35, 187, 5, 5, 113, 243, 5, 5, 40, 227, 5, 5, 70, 171, 5, 5, 132, 221, 5, 5, 245, 216, 5, 5, 97, 242, 5, 5, 199, 238, 5, 5, 3, 190, 5, 5, 84, 170, 5, 5, 28, 239, 5, 5, 196, 190, 5, 5, 5, 179, 5, 5, 165, 223, 5, 5, 4, 223, 5, 5, 3, 218, 5, 5, 193, 194, 5, 5, 2, 220, 5, 5, 91, 239, 5, 5, 183, 230, 5, 5, 30, 235, 5, 5, 184, 213, 5, 5, 161, 235, 5, 5, 45, 232, 5, 5, 59, 191, 5, 5, 90, 179, 5, 5, 29, 204, 5, 5, 159, 220, 5, 5, 232, 233, 5, 5, 158, 209, 5, 5, 81, 196, 5, 5, 209, 223, 5, 5, 137, 184, 5, 5, 188, 184, 5, 5, 184, 205, 5, 5, 55, 245, 5, 5, 25, 198, 5, 5, 96, 208, 5, 5, 137, 225, 5, 5, 237, 220, 5, 5, 8, 233, 5, 5, 57, 224, 5, 5, 201, 210, 5, 5, 60, 191, 5, 5, 237, 206, 5, 5, 131, 229, 5, 5, 166, 181, 5, 5, 206, 228, 5, 5, 94, 182, 5, 5, 73, 163, 5, 5, 98, 183, 5, 5, 124, 214, 5, 5, 238, 165, 5, 5, 89, 182, 5, 5, 9, 248, 5, 5, 115, 213, 5, 5, 151, 234, 5, 5, 155, 237, 5, 5, 2, 210, 5, 5, 222, 216, 5, 5, 179, 194, 5, 5, 24, 196, 5, 5, 56, 245, 5, 5, 42, 208, 5, 5, 40, 240, 5, 5, 158, 174, 5, 5, 41, 189, 5, 5, 232, 232, 5, 5, 190, 209, 5, 5, 44, 187, 5, 5, 251, 249, 5, 5, 19, 201, 5, 5, 145, 206, 5, 5, 168, 210, 5, 5, 101, 236, 5, 5, 244, 166, 5, 5, 139, 236, 5, 5, 79, 174, 5, 5, 181, 212, 5, 5, 251, 186, 5, 5, 159, 174, 5, 5, 158, 168, 5, 5, 18, 164, 5, 5, 171, 247, 5, 5, 169, 208, 5, 5, 186, 213, 5, 5, 25, 216, 5, 5, 237, 214, 5, 5, 10, 248, 5, 5, 228, 176, 5, 5, 175, 223, 5, 5, 18, 224, 5, 5, 56, 219, 5, 5, 217, 242, 5, 5, 173, 187, 5, 5, 45, 187, 5, 5, 36, 185, 5, 5, 110, 235, 5, 5, 15, 205, 5, 5, 241, 214, 5, 5, 162, 174, 5, 5, 191, 199, 5, 5, 57, 207, 5, 5, 215, 214, 5, 5, 209, 166, 5, 5, 85, 205, 5, 5, 189, 165, 5, 5, 211, 225, 5, 5, 203, 172, 5, 5, 255, 173, 5, 5, 32, 241, 5, 5, 229, 238, 5, 5, 209, 163, 5, 5, 63, 248, 5, 5, 252, 215, 5, 5, 233, 164, 5, 5, 44, 199, 5, 5, 239, 224, 5, 5, 73, 218, 5, 5, 87, 166, 5, 5, 163, 204, 5, 5, 164, 241, 5, 5, 232, 234, 5, 5, 174, 187, 5, 5, 202, 246, 5, 5, 37, 185, 5, 5, 116, 234, 5, 5, 143, 233, 5, 5, 205, 191, 5, 5, 69, 194, 5, 5, 210, 163, 5, 5, 99, 181, 5, 5, 69, 229, 5, 5, 70, 229, 5, 5, 179, 170, 5, 5, 156, 175, 5, 5, 244, 179, 5, 5, 137, 207, 5, 5, 29, 226, 5, 5, 167, 216, 5, 5, 152, 167, 5, 5, 112, 180, 5, 5, 52, 198, 5, 5, 224, 239, 5, 5, 118, 208, 5, 5, 217, 180, 5, 5, 156, 192, 5, 5, 150, 206, 5, 5, 219, 203, 5, 5, 249, 236, 5, 5, 40, 225, 5, 5, 146, 203, 5, 5, 174, 173, 5, 5, 104, 218, 5, 5, 53, 197, 5, 5, 243, 207, 5, 5, 88, 233, 5, 5, 58, 207, 5, 5, 169, 210, 5, 5, 152, 183, 5, 5, 198, 190, 5, 5, 8, 222, 5, 5, 169, 245, 5, 5, 121, 190, 5, 5, 93, 237, 5, 5, 160, 186, 5, 5, 20, 201, 5, 5, 231, 178, 5, 5, 192, 231, 5, 5, 130, 164, 5, 5, 57, 245, 5, 5, 59, 207, 5, 5, 18, 183, 5, 5, 17, 234, 5, 5, 17, 205, 5, 5, 198, 219, 5, 5, 53, 241, 5, 5, 105, 218, 5, 5, 150, 227, 5, 5, 93, 211, 5, 5, 126, 209, 5, 5, 96, 244, 5, 5, 58, 225, 5, 5, 204, 204, 5, 5, 222, 242, 5, 5, 29, 231, 5, 5, 248, 222, 5, 5, 115, 226, 5, 5, 75, 203, 5, 5, 177, 202, 5, 5, 63, 246, 5, 5, 15, 248, 5, 5, 91, 166, 5, 5, 236, 230, 5, 5, 164, 180, 5, 5, 226, 245, 5, 5, 100, 168, 5, 5, 36, 243, 5, 5, 204, 173, 5, 5, 113, 180, 5, 5, 193, 245, 5, 5, 212, 237, 5, 5, 12, 214, 5, 5, 182, 223, 5, 5, 38, 229, 5, 5, 173, 195, 5, 5, 232, 176, 5, 5, 13, 233, 5, 5, 131, 223, 5, 5, 147, 181, 5, 5, 215, 163, 5, 5, 63, 214, 5, 5, 39, 201, 5, 5, 222, 241, 5, 5, 138, 220, 5, 5, 58, 224, 5, 5, 127, 242, 5, 5, 44, 242, 5, 5, 92, 205, 5, 5, 232, 216, 5, 5, 67, 219, 5, 5, 158, 233, 5, 5, 207, 174, 5, 5, 208, 164, 5, 5, 217, 218, 5, 5, 126, 233, 5, 5, 199, 171, 5, 5, 67, 202, 5, 5, 227, 239, 5, 5, 64, 214, 5, 5, 228, 217, 5, 5, 4, 234, 5, 5, 156, 178, 5, 5, 239, 233, 5, 5, 173, 220, 5, 5, 187, 238, 5, 5, 204, 179, 5, 5, 205, 179, 5, 5, 216, 163, 5, 5, 221, 231, 5, 5, 228, 185, 5, 5, 151, 203, 5, 5, 64, 246, 5, 5, 132, 173, 5, 5, 8, 225, 5, 5, 224, 193, 5, 5, 67, 236, 5, 5, 237, 218, 5, 5, 156, 236, 5, 5, 174, 213, 5, 5, 213, 174, 5, 5, 98, 249, 5, 5, 99, 231, 5, 5, 78, 187, 5, 5, 152, 184, 5, 5, 184, 249, 5, 5, 38, 206, 5, 5, 48, 169, 5, 5, 161, 164, 5, 5, 127, 177, 5, 5, 70, 180, 5, 5, 2, 198, 5, 5, 156, 203, 5, 5, 213, 212, 5, 5, 95, 228, 5, 5, 173, 210, 5, 5, 152, 207, 5, 5, 118, 234, 5, 5, 219, 225, 5, 5, 89, 236, 5, 5, 221, 163, 5, 5, 221, 214, 5, 5, 7, 250, 5, 5, 168, 241, 5, 5, 41, 180, 5, 5, 206, 236, 5, 5, 67, 244, 5, 5, 115, 172, 5, 5, 92, 166, 5, 5, 23, 203, 5, 5, 74, 175, 5, 5, 48, 232, 5, 5, 115, 180, 5, 5, 101, 193, 5, 5, 12, 166, 5, 5, 189, 225, 5, 5, 157, 184, 5, 5, 69, 234, 5, 5, 122, 185, 5, 5, 211, 234, 5, 5, 215, 249, 5, 5, 61, 232, 5, 5, 132, 233, 5, 5, 181, 164, 5, 5, 113, 237, 5, 5, 146, 238, 5, 5, 58, 226, 5, 5, 25, 248, 5, 5, 200, 219, 5, 5, 223, 232, 5, 5, 232, 228, 5, 5, 135, 226, 5, 5, 168, 206, 5, 5, 101, 223, 5, 5, 148, 227, 5, 5, 136, 198, 5, 5, 12, 221, 5, 5, 102, 211, 5, 5, 48, 243, 5, 5, 251, 176, 5, 5, 80, 214, 5, 5, 232, 185, 5, 5, 228, 163, 5, 5, 93, 168, 5, 5, 190, 205, 5, 5, 196, 188, 5, 5, 26, 190, 5, 5, 109, 170, 5, 5, 85, 210, 5, 5, 55, 215, 5, 5, 43, 180, 5, 5, 168, 244, 5, 5, 24, 179, 5, 5, 143, 222, 5, 5, 106, 216, 5, 5, 99, 228, 5, 5, 165, 209, 5, 5, 107, 208, 5, 5, 90, 182, 5, 5, 79, 233, 5, 5, 228, 168, 5, 5, 234, 212, 5, 5, 178, 185, 5, 5, 32, 225, 5, 5, 168, 194, 5, 5, 226, 187, 5, 5, 135, 208, 5, 5, 66, 225, 5, 5, 86, 224, 5, 5, 89, 246, 5, 5, 239, 201, 5, 5, 56, 163, 5, 5, 99, 200, 5, 5, 68, 226, 5, 5, 254, 180, 5, 5, 43, 176, 5, 5, 51, 204, 5, 5, 173, 204, 5, 5, 126, 236, 5, 5, 135, 173, 5, 5, 215, 215, 5, 5, 252, 206, 5, 5, 244, 237, 5, 5, 252, 203, 5, 5, 112, 205, 5, 5, 188, 182, 5, 5, 27, 189, 5, 5, 103, 163, 5, 5, 94, 193, 5, 5, 238, 215, 5, 5, 169, 236, 5, 5, 216, 215, 5, 5, 19, 206, 5, 5, 14, 172, 5, 5, 104, 241, 5, 5, 237, 205, 5, 5, 93, 203, 5, 5, 200, 209, 5, 5, 47, 168, 5, 5, 216, 173, 5, 5, 63, 220, 5, 5, 71, 238, 5, 5, 13, 236, 5, 5, 186, 239, 5, 5, 67, 208, 5, 5, 194, 184, 5, 5, 90, 224, 5, 5, 128, 240, 5, 5, 108, 176, 5, 5, 129, 206, 5, 5, 113, 217, 5, 5, 90, 196, 5, 5, 49, 205, 5, 5, 227, 225, 5, 5, 245, 175, 5, 5, 57, 204, 5, 5, 149, 227, 5, 5, 51, 232, 5, 5, 103, 234, 5, 5, 24, 166, 5, 5, 227, 238, 5, 5, 194, 230, 5, 5, 124, 179, 5, 5, 37, 224, 5, 5, 214, 223, 5, 5, 29, 186, 5, 5, 192, 183, 5, 5, 203, 205, 5, 5, 189, 210, 5, 5, 232, 170, 5, 5, 73, 245, 5, 5, 121, 184, 5, 5, 154, 212, 5, 5, 193, 183, 5, 5, 21, 225, 5, 5, 206, 187, 5, 5, 128, 185, 5, 5, 80, 191, 5, 5, 17, 197, 5, 5, 59, 164, 5, 5, 146, 248, 5, 5, 164, 165, 5, 5, 137, 232, 5, 5, 133, 218, 5, 5, 192, 235, 5, 5, 19, 242, 5, 5, 231, 197, 5, 5, 228, 224, 5, 5, 230, 224, 5, 5, 116, 187, 5, 5, 143, 180, 5, 5, 91, 246, 5, 5, 33, 191, 5, 5, 195, 225, 5, 5, 105, 224, 5, 5, 151, 230, 5, 5, 97, 229, 5, 5, 141, 200, 5, 5, 144, 187, 5, 5, 105, 190, 5, 5, 140, 232, 5, 5, 158, 184, 5, 5, 119, 180, 5, 5, 10, 243, 5, 5, 108, 185, 5, 5, 140, 183, 5, 5, 25, 239, 5, 5, 78, 245, 5, 5, 159, 249, 5, 5, 122, 232, 5, 5, 26, 249, 5, 5, 29, 218, 5, 5, 147, 226, 5, 5, 225, 219, 5, 5, 31, 218, 5, 5, 37, 215, 5, 5, 153, 194, 5, 5, 132, 178, 5, 5, 49, 195, 5, 5, 133, 178, 5, 5, 192, 171, 5, 5, 242, 181, 5, 5, 3, 225, 5, 5, 27, 239, 5, 5, 134, 178, 5, 5, 29, 239, 5, 5, 163, 205, 5, 5, 243, 181, 5, 5, 162, 235, 5, 5, 215, 242, 5, 5, 218, 242, 5, 5, 16, 227, 192, 0, 0, 0, 5, 5, 83, 176, 5, 5, 72, 210, 5, 5, 183, 175, 5, 5, 81, 173, 5, 5, 27, 170, 5, 5, 159, 228, 5, 5, 185, 175, 192, 0, 0, 0, 5, 5, 102, 206, 5, 5, 12, 187, 5, 5, 81, 201, 5, 5, 217, 230, 5, 5, 80, 247, 5, 5, 40, 183, 5, 5, 52, 212, 5, 5, 58, 189, 5, 5, 172, 247, 5, 5, 0, 128, 5, 5, 188, 194, 5, 5, 99, 188, 5, 5, 166, 212, 5, 5, 161, 198, 5, 5, 16, 186, 5, 5, 230, 194, 5, 5, 62, 181, 5, 5, 64, 194, 5, 5, 162, 216, 5, 5, 55, 166, 5, 5, 161, 208, 5, 5, 42, 207, 5, 5, 145, 177, 5, 5, 107, 231, 5, 5, 77, 173, 5, 5, 61, 244, 5, 5, 202, 204, 5, 5, 22, 228, 5, 5, 193, 202, 5, 5, 103, 240, 5, 5, 216, 202, 5, 5, 147, 224, 5, 5, 81, 208, 5, 5, 55, 170, 5, 5, 74, 230, 5, 5, 41, 183, 5, 5, 27, 249, 5, 5, 43, 170, 5, 5, 177, 192, 5, 5, 75, 230, 5, 5, 230, 233, 5, 5, 82, 208, 5, 5, 231, 233, 5, 5, 42, 232, 5, 5, 122, 193, 5, 5, 91, 208, 5, 5, 57, 170, 5, 5, 48, 220, 5, 5, 59, 170, 5, 5, 85, 192, 5, 5, 229, 216, 5, 5, 200, 171, 5, 5, 85, 171, 5, 5, 95, 231, 5, 5, 158, 243, 5, 5, 45, 214, 5, 5, 169, 195, 5, 5, 13, 208, 5, 5, 187, 184, 5, 5, 131, 239, 5, 5, 34, 171, 5, 5, 46, 227, 5, 5, 67, 194, 5, 5, 224, 170, 5, 5, 227, 222, 5, 5, 60, 207, 5, 5, 250, 248, 5, 5, 31, 172, 5, 5, 148, 195, 5, 5, 221, 188, 5, 5, 208, 174, 5, 5, 46, 214, 5, 5, 116, 171, 5, 5, 50, 166, 5, 5, 223, 178, 5, 5, 105, 179, 5, 5, 179, 192, 5, 5, 235, 173, 5, 5, 54, 166, 5, 5, 57, 199, 5, 5, 148, 226, 5, 5, 118, 207, 5, 5, 58, 199, 5, 5, 23, 197, 5, 5, 72, 241, 5, 5, 159, 186, 5, 5, 77, 167, 5, 5, 63, 174, 5, 5, 250, 171, 5, 5, 4, 179, 5, 5, 137, 215, 5, 5, 178, 187, 5, 5, 79, 209, 5, 5, 195, 204, 5, 5, 142, 206, 5, 5, 132, 229, 5, 5, 236, 169, 5, 5, 198, 181, 5, 5, 119, 208, 5, 5, 81, 193, 5, 5, 162, 222, 5, 5, 94, 237, 5, 5, 151, 206, 5, 5, 42, 189, 5, 5, 136, 211, 5, 5, 76, 177, 5, 5, 100, 189, 5, 5, 14, 179, 5, 5, 157, 192, 5, 5, 22, 205, 5, 5, 124, 243, 5, 5, 56, 234, 5, 5, 169, 239, 5, 5, 79, 187, 5, 5, 144, 226, 5, 5, 51, 172, 5, 5, 168, 222, 5, 5, 212, 234, 5, 5, 123, 235, 5, 5, 72, 178, 5, 5, 238, 205, 5, 5, 9, 189, 5, 5, 10, 189, 5, 5, 122, 184, 5, 5, 254, 212, 5, 5, 254, 177, 5, 5, 123, 202, 5, 5, 3, 173, 5, 5, 4, 173, 5, 5, 5, 173, 5, 5, 252, 173, 5, 5, 190, 206, 5, 5, 186, 225, 5, 5, 129, 221, 5, 5, 188, 173, 5, 5, 110, 172, 5, 5, 55, 219, 5, 5, 60, 174, 5, 5, 189, 173, 5, 5, 169, 168, 5, 5, 193, 173, 5, 5, 83, 195, 5, 5, 40, 201, 5, 5, 89, 195, 5, 5, 123, 192, 5, 5, 136, 176, 5, 5, 199, 168, 5, 5, 149, 210, 5, 5, 28, 218, 5, 5, 78, 196, 5, 5, 179, 181, 5, 5, 96, 238, 5, 5, 147, 225, 5, 5, 197, 201, 5, 5, 93, 176, 5, 5, 75, 198, 5, 5, 79, 198, 5, 5, 66, 232, 5, 5, 125, 175, 5, 5, 72, 177, 5, 5, 126, 175, 5, 5, 105, 230, 5, 5, 106, 230, 5, 5, 121, 226, 5, 5, 69, 173, 5, 5, 94, 209, 5, 5, 255, 239, 5, 5, 198, 201, 5, 5, 161, 195, 5, 5, 165, 210, 5, 5, 32, 235, 5, 5, 154, 225, 5, 5, 47, 247, 5, 5, 202, 201, 5, 5, 3, 176, 5, 5, 138, 216, 5, 5, 131, 199, 5, 5, 185, 246, 5, 5, 119, 207, 5, 5, 166, 186, 5, 5, 199, 248, 5, 5, 176, 187, 5, 5, 31, 188, 5, 5, 234, 231, 5, 5, 180, 192, 5, 5, 139, 226, 5, 5, 64, 174, 5, 5, 149, 225, 5, 5, 247, 211, 5, 5, 191, 206, 5, 5, 40, 228, 5, 5, 251, 171, 5, 5, 85, 189, 5, 5, 196, 222, 5, 5, 78, 167, 5, 5, 124, 165, 5, 5, 131, 203, 5, 5, 132, 203, 5, 5, 12, 167, 5, 5, 30, 187, 5, 5, 41, 228, 5, 5, 125, 165, 5, 5, 122, 224, 5, 5, 116, 206, 5, 5, 104, 193, 5, 5, 212, 192, 5, 5, 155, 175, 5, 5, 141, 226, 5, 5, 102, 196, 5, 5, 4, 196, 5, 5, 97, 178, 5, 5, 71, 239, 5, 5, 186, 218, 5, 5, 171, 229, 5, 5, 108, 212, 5, 5, 123, 227, 5, 5, 212, 233, 5, 5, 50, 195, 5, 5, 135, 203, 5, 5, 30, 194, 5, 5, 213, 233, 5, 5, 31, 194, 5, 5, 84, 195, 5, 5, 98, 178, 5, 5, 157, 202, 5, 5, 214, 180, 5, 5, 124, 206, 5, 5, 226, 185, 5, 5, 120, 190, 5, 5, 241, 231, 5, 5, 249, 248, 5, 5, 193, 234, 5, 5, 45, 217, 5, 5, 51, 195, 5, 5, 234, 184, 5, 5, 119, 210, 5, 5, 124, 227, 5, 5, 142, 226, 5, 5, 123, 193, 5, 5, 80, 205, 5, 5, 194, 234, 5, 5, 169, 186, 5, 5, 152, 180, 5, 5, 204, 172, 5, 5, 37, 167, 5, 5, 127, 224, 5, 5, 174, 208, 5, 5, 143, 226, 5, 5, 193, 231, 5, 5, 191, 192, 5, 5, 173, 246, 5, 5, 42, 163, 5, 5, 213, 201, 5, 5, 54, 178, 5, 5, 110, 211, 5, 5, 89, 228, 5, 5, 233, 177, 5, 5, 86, 205, 5, 5, 170, 202, 5, 5, 43, 163, 5, 5, 123, 242, 5, 5, 61, 207, 5, 5, 165, 199, 5, 5, 45, 229, 5, 5, 23, 205, 5, 5, 63, 178, 5, 5, 157, 178, 5, 5, 255, 244, 5, 5, 124, 193, 5, 5, 209, 174, 5, 5, 43, 229, 5, 5, 93, 205, 5, 5, 145, 191, 5, 5, 29, 175, 5, 5, 90, 195, 5, 5, 145, 226, 5, 5, 49, 226, 5, 5, 134, 224, 5, 5, 111, 182, 5, 5, 27, 190, 5, 5, 187, 186, 5, 5, 190, 167, 5, 5, 60, 228, 5, 5, 68, 168, 5, 5, 137, 195, 5, 5, 45, 204, 5, 5, 114, 193, 5, 5, 134, 209, 5, 5, 163, 207, 5, 5, 145, 199, 5, 5, 72, 219, 5, 5, 77, 222, 5, 5, 136, 188, 5, 5, 84, 207, 5, 5, 222, 167, 5, 5, 54, 188, 5, 5, 46, 194, 5, 5, 58, 204, 5, 5, 47, 194, 5, 5, 116, 205, 5, 5, 117, 205, 5, 5, 55, 180, 5, 5, 168, 199, 5, 5, 217, 202, 5, 5, 128, 205, 5, 5, 143, 240, 5, 5, 129, 205, 5, 5, 31, 219, 5, 5, 106, 224, 5, 5, 70, 205, 5, 5, 86, 169, 5, 5, 239, 186, 5, 5, 203, 221, 5, 5, 22, 187, 5, 5, 18, 241, 5, 5, 107, 212, 5, 5, 97, 164, 5, 5, 190, 194, 5, 5, 253, 202, 5, 5, 81, 245, 5, 5, 174, 169, 5, 5, 177, 187, 5, 5, 208, 205, 5, 5, 240, 195, 5, 5, 250, 213, 5, 5, 143, 218, 5, 5, 3, 179, 5, 5, 210, 221, 5, 5, 147, 184, 5, 5, 133, 203, 5, 5, 199, 211, 5, 5, 254, 227, 5, 5, 13, 240, 5, 5, 131, 221, 5, 5, 95, 169, 5, 5, 31, 187, 5, 5, 209, 205, 5, 5, 167, 185, 5, 5, 168, 185, 5, 5, 240, 207, 5, 5, 144, 195, 5, 5, 23, 244, 5, 5, 136, 245, 5, 5, 213, 214, 5, 5, 85, 170, 5, 5, 250, 196, 5, 5, 137, 203, 5, 5, 179, 187, 5, 5, 86, 197, 5, 5, 41, 227, 5, 5, 52, 195, 5, 5, 214, 205, 5, 5, 41, 192, 5, 5, 74, 163, 5, 5, 255, 171, 5, 5, 87, 223, 5, 5, 103, 186, 5, 5, 252, 249, 5, 5, 253, 249, 5, 5, 195, 171, 5, 5, 53, 195, 5, 5, 186, 217, 5, 5, 244, 207, 5, 5, 111, 211, 5, 5, 9, 190, 5, 5, 132, 166, 5, 5, 7, 171, 5, 5, 90, 223, 5, 5, 86, 185, 5, 5, 111, 195, 5, 5, 201, 171, 5, 5, 18, 179, 5, 5, 6, 217, 5, 5, 9, 217, 5, 5, 163, 195, 5, 5, 102, 245, 5, 5, 68, 240, 5, 5, 189, 217, 5, 5, 123, 247, 5, 5, 46, 229, 5, 5, 175, 185, 5, 5, 94, 172, 5, 5, 28, 190, 5, 5, 94, 168, 5, 5, 78, 227, 5, 5, 232, 201, 5, 5, 233, 210, 5, 5, 60, 206, 5, 5, 234, 223, 5, 5, 110, 212, 5, 5, 107, 216, 5, 5, 119, 240, 5, 5, 165, 207, 5, 5, 241, 172, 5, 5, 223, 207, 5, 5, 189, 182, 5, 5, 85, 207, 5, 5, 246, 175, 5, 5, 27, 214, 5, 5, 184, 169, 5, 5, 193, 217, 5, 5, 195, 217, 5, 5, 101, 191, 5, 5, 111, 187, 5, 5, 201, 223, 5, 5, 152, 230, 5, 5, 112, 212, 5, 5, 248, 163, 5, 5, 244, 227, 5, 5, 127, 249, 5, 5, 120, 204, 5, 5, 41, 164, 5, 5, 53, 192, 5, 5, 79, 245, 5, 5, 128, 249, 5, 5, 18, 247, 5, 5, 112, 165, 5, 5, 233, 191, 5, 5, 234, 191, 5, 5, 249, 163, 5, 5, 37, 235, 5, 5, 213, 240, 5, 5, 32, 218, 5, 5, 133, 167, 5, 5, 193, 206, 5, 5, 174, 179, 5, 5, 162, 191, 5, 5, 222, 168, 5, 5, 206, 237, 5, 5, 19, 167, 5, 5, 101, 174, 5, 5, 203, 194, 5, 5, 30, 175, 5, 5, 166, 204, 5, 5, 128, 212, 5, 5, 220, 164, 5, 5, 130, 199, 5, 5, 195, 163, 5, 5, 68, 183, 5, 5, 111, 227, 5, 5, 145, 173, 5, 5, 154, 204, 5, 5, 128, 239, 5, 5, 208, 231, 5, 5, 50, 206, 5, 5, 231, 195, 5, 5, 51, 206, 5, 5, 248, 196, 5, 5, 216, 198, 5, 5, 183, 213, 5, 5, 153, 211, 5, 5, 229, 165, 5, 5, 2, 194, 5, 5, 97, 209, 5, 5, 172, 232, 5, 5, 252, 196, 5, 5, 181, 172, 5, 5, 228, 198, 5, 5, 17, 248, 5, 5, 14, 194, 5, 5, 153, 196, 5, 5, 90, 200, 5, 5, 242, 175, 5, 5, 44, 194, 5, 5, 128, 188, 5, 5, 137, 188, 5, 5, 60, 237, 5, 5, 255, 177, 5, 5, 180, 204, 5, 5, 193, 207, 5, 5, 104, 213, 5, 5, 20, 168, 5, 5, 149, 211, 5, 5, 88, 239, 5, 5, 250, 194, 5, 5, 219, 242, 5, 5, 217, 165, 5, 5, 208, 183, 5, 5, 167, 211, 5, 5, 225, 226, 5, 5, 251, 217, 5, 5, 50, 210, 5, 5, 88, 184, 5, 5, 207, 235, 5, 5, 216, 232, 5, 5, 7, 229, 5, 5, 44, 245, 5, 5, 64, 200, 5, 5, 98, 164, 5, 5, 31, 227, 5, 5, 105, 213, 5, 5, 232, 246, 5, 5, 107, 199, 5, 5, 251, 213, 5, 5, 233, 246, 5, 5, 181, 163, 5, 5, 218, 232, 5, 5, 5, 222, 5, 5, 254, 213, 5, 5, 227, 175, 5, 5, 179, 169, 5, 5, 99, 183, 5, 5, 229, 175, 5, 5, 68, 202, 5, 5, 92, 163, 5, 5, 204, 229, 5, 5, 149, 163, 5, 5, 19, 197, 5, 5, 227, 165, 5, 5, 110, 166, 5, 5, 132, 220, 5, 5, 244, 194, 5, 5, 183, 189, 5, 5, 70, 242, 5, 5, 228, 189, 5, 5, 175, 212, 5, 5, 127, 193, 5, 5, 190, 245, 5, 5, 65, 214, 5, 5, 125, 203, 5, 5, 126, 203, 5, 5, 114, 246, 5, 5, 13, 239, 5, 5, 129, 212, 5, 5, 56, 218, 5, 5, 44, 170, 5, 5, 233, 243, 5, 5, 222, 245, 5, 5, 255, 227, 5, 5, 44, 202, 5, 5, 12, 212, 5, 5, 185, 190, 5, 5, 61, 223, 5, 5, 230, 207, 5, 5, 51, 214, 5, 5, 246, 216, 5, 5, 185, 205, 5, 5, 13, 212, 5, 5, 95, 245, 5, 5, 55, 202, 5, 5, 213, 237, 5, 5, 94, 211, 5, 5, 227, 212, 5, 5, 108, 236, 5, 5, 26, 225, 5, 5, 47, 246, 5, 5, 187, 237, 5, 5, 124, 181, 5, 5, 23, 187, 5, 5, 126, 219, 5, 5, 104, 198, 5, 5, 24, 187, 5, 5, 8, 241, 5, 5, 222, 240, 5, 5, 20, 243, 5, 5, 186, 227, 5, 5, 199, 176, 5, 5, 93, 219, 5, 5, 114, 167, 5, 5, 242, 240, 5, 5, 87, 197, 5, 5, 244, 240, 5, 5, 187, 218, 5, 5, 233, 233, 5, 5, 116, 167, 5, 5, 160, 180, 5, 5, 168, 186, 5, 5, 189, 227, 5, 5, 47, 198, 5, 5, 131, 181, 5, 5, 24, 233, 5, 5, 195, 234, 5, 5, 205, 172, 5, 5, 37, 249, 5, 5, 240, 233, 5, 5, 43, 249, 5, 5, 42, 216, 5, 5, 223, 242, 5, 5, 80, 187, 5, 5, 153, 207, 5, 5, 155, 227, 5, 5, 128, 177, 5, 5, 157, 225, 5, 5, 167, 204, 5, 5, 191, 205, 5, 5, 103, 238, 5, 5, 23, 194, 5, 5, 54, 243, 5, 5, 124, 235, 5, 5, 87, 187, 5, 5, 33, 225, 5, 5, 27, 186, 5, 5, 102, 187, 5, 5, 194, 242, 5, 5, 255, 242, 5, 5, 83, 249, 5, 5, 101, 235, 5, 5, 189, 194, 5, 5, 217, 188, 5, 5, 32, 230, 5, 5, 245, 211, 5, 5, 246, 211, 5, 5, 49, 239, 5, 5, 116, 185, 5, 5, 231, 177, 5, 5, 179, 215, 5, 5, 114, 222, 5, 5, 81, 236, 5, 5, 61, 234, 5, 5, 62, 234, 5, 5, 63, 234, 5, 5, 66, 234, 5, 5, 67, 238, 5, 5, 107, 175, 5, 5, 93, 242, 5, 5, 124, 223, 5, 5, 40, 202, 5, 5, 67, 242, 5, 5, 236, 229, 5, 5, 39, 173, 5, 5, 12, 228, 5, 5, 203, 193, 5, 5, 0, 154, 5, 5, 112, 172, 5, 5, 233, 230, 5, 5, 37, 227, 5, 5, 64, 229, 5, 5, 24, 222, 5, 5, 237, 211, 5, 5, 20, 228, 5, 5, 235, 165, 5, 5, 250, 216, 5, 5, 69, 203, 5, 5, 83, 167, 5, 5, 65, 236, 5, 5, 42, 192, 5, 5, 123, 246, 5, 5, 66, 236, 5, 5, 47, 177, 5, 5, 18, 231, 5, 5, 92, 235, 5, 5, 142, 195, 5, 5, 31, 193, 5, 5, 37, 207, 5, 5, 175, 189, 5, 5, 121, 193, 5, 5, 127, 175, 5, 5, 145, 195, 5, 5, 142, 218, 5, 5, 82, 204, 5, 5, 224, 219, 5, 5, 0, 131, 5, 5, 189, 177, 5, 5, 37, 195, 5, 5, 106, 179, 5, 5, 80, 223, 5, 5, 9, 227, 5, 5, 94, 242, 5, 5, 244, 209, 5, 5, 186, 166, 5, 5, 79, 241, 5, 5, 241, 197, 5, 5, 104, 235, 5, 5, 37, 180, 5, 5, 58, 223, 5, 5, 14, 185, 5, 5, 73, 177, 5, 5, 199, 201, 5, 5, 7, 188, 5, 5, 107, 201, 5, 5, 125, 171, 5, 5, 180, 216, 5, 5, 182, 170, 5, 5, 3, 223, 5, 5, 203, 191, 5, 5, 92, 217, 5, 5, 141, 241, 5, 5, 167, 231, 5, 5, 84, 197, 5, 5, 42, 202, 5, 5, 97, 177, 5, 5, 35, 235, 5, 5, 162, 182, 5, 5, 26, 172, 5, 5, 46, 198, 5, 5, 27, 187, 5, 5, 239, 181, 5, 5, 94, 216, 5, 5, 60, 219, 5, 5, 228, 213, 5, 5, 71, 241, 5, 5, 57, 191, 5, 5, 223, 240, 5, 5, 61, 169, 5, 5, 185, 238, 5, 5, 164, 199, 5, 5, 89, 239, 5, 5, 15, 208, 5, 5, 54, 224, 5, 5, 21, 189, 5, 5, 127, 182, 5, 5, 161, 243, 5, 5, 193, 172, 5, 5, 226, 164, 5, 5, 11, 211, 5, 5, 12, 211, 5, 5, 205, 203, 5, 5, 121, 166, 5, 5, 231, 172, 5, 5, 33, 163, 5, 5, 182, 178, 5, 5, 70, 173, 5, 5, 152, 237, 5, 5, 74, 198, 5, 5, 128, 181, 5, 5, 241, 195, 5, 5, 227, 229, 5, 5, 254, 208, 5, 5, 131, 201, 5, 5, 58, 218, 5, 5, 207, 243, 5, 5, 14, 245, 5, 5, 15, 245, 5, 5, 252, 223, 5, 5, 232, 182, 5, 5, 80, 217, 5, 5, 171, 212, 5, 5, 26, 226, 5, 5, 181, 175, 5, 5, 48, 247, 5, 5, 44, 198, 5, 5, 108, 201, 5, 5, 16, 245, 5, 5, 39, 192, 5, 5, 49, 241, 5, 5, 41, 208, 5, 5, 58, 191, 5, 5, 191, 237, 5, 5, 159, 191, 5, 5, 190, 169, 5, 5, 69, 175, 5, 5, 255, 208, 5, 5, 88, 225, 5, 5, 17, 245, 5, 5, 40, 192, 5, 5, 19, 174, 5, 5, 83, 204, 5, 5, 109, 201, 5, 5, 59, 223, 5, 5, 9, 229, 5, 5, 233, 182, 5, 5, 128, 182, 5, 5, 150, 174, 5, 5, 14, 240, 5, 5, 70, 175, 5, 5, 186, 238, 5, 5, 32, 187, 5, 5, 245, 163, 5, 5, 29, 249, 5, 5, 139, 193, 5, 5, 69, 247, 5, 5, 25, 211, 5, 5, 246, 244, 5, 5, 195, 237, 5, 5, 186, 226, 5, 5, 231, 207, 5, 5, 103, 167, 5, 5, 51, 247, 5, 5, 6, 183, 5, 5, 17, 169, 5, 5, 176, 189, 5, 5, 59, 230, 5, 5, 10, 242, 5, 5, 202, 176, 5, 5, 3, 220, 5, 5, 42, 227, 5, 5, 44, 220, 5, 5, 7, 181, 5, 5, 110, 213, 5, 5, 24, 240, 5, 5, 30, 209, 5, 5, 170, 168, 5, 5, 174, 231, 5, 5, 254, 192, 5, 5, 155, 233, 5, 5, 118, 246, 5, 5, 226, 216, 5, 5, 133, 237, 5, 5, 46, 183, 5, 5, 119, 191, 5, 5, 226, 166, 5, 5, 25, 240, 5, 5, 122, 214, 5, 5, 139, 228, 5, 5, 218, 198, 5, 5, 54, 172, 5, 5, 18, 175, 5, 5, 248, 211, 5, 5, 17, 207, 5, 5, 76, 198, 5, 5, 210, 231, 5, 5, 4, 182, 5, 5, 51, 178, 5, 5, 46, 167, 5, 5, 13, 167, 5, 5, 126, 165, 5, 5, 67, 174, 5, 5, 255, 243, 5, 5, 88, 197, 5, 5, 233, 231, 5, 5, 89, 197, 5, 5, 180, 197, 5, 5, 158, 204, 5, 5, 249, 249, 5, 5, 191, 174, 5, 5, 4, 175, 5, 5, 45, 220, 5, 5, 154, 245, 5, 5, 239, 194, 5, 5, 255, 192, 5, 5, 240, 194, 5, 5, 22, 233, 5, 5, 123, 166, 5, 5, 227, 190, 5, 5, 224, 178, 5, 5, 109, 184, 5, 5, 91, 179, 5, 5, 108, 235, 5, 5, 186, 215, 5, 5, 188, 199, 5, 5, 35, 209, 5, 5, 72, 239, 5, 5, 199, 237, 5, 5, 159, 194, 5, 5, 108, 219, 5, 5, 116, 213, 5, 5, 142, 239, 5, 5, 175, 187, 5, 5, 108, 231, 5, 5, 126, 171, 5, 5, 31, 171, 5, 5, 157, 218, 5, 5, 215, 241, 5, 5, 211, 201, 5, 5, 49, 220, 5, 5, 140, 191, 5, 5, 133, 229, 5, 5, 128, 232, 5, 5, 214, 214, 5, 5, 181, 197, 5, 5, 255, 199, 5, 5, 43, 196, 5, 5, 247, 199, 5, 5, 175, 179, 5, 5, 87, 215, 5, 5, 200, 237, 5, 5, 148, 217, 5, 5, 238, 214, 5, 5, 140, 193, 5, 5, 25, 243, 5, 5, 195, 185, 5, 5, 92, 239, 5, 5, 21, 238, 5, 5, 151, 168, 5, 5, 185, 228, 5, 5, 167, 182, 5, 5, 112, 201, 5, 5, 36, 218, 5, 5, 52, 178, 5, 5, 166, 245, 5, 5, 71, 197, 5, 5, 32, 232, 5, 5, 100, 242, 5, 5, 67, 177, 5, 5, 220, 166, 5, 5, 78, 216, 5, 5, 22, 238, 5, 5, 171, 192, 5, 5, 253, 196, 5, 5, 10, 241, 5, 5, 5, 175, 5, 5, 125, 214, 5, 5, 7, 165, 5, 5, 69, 200, 5, 5, 65, 184, 5, 5, 103, 199, 5, 5, 177, 216, 5, 5, 82, 196, 5, 5, 78, 178, 5, 5, 76, 173, 5, 5, 161, 202, 5, 5, 189, 184, 5, 5, 82, 170, 5, 5, 50, 241, 5, 5, 247, 197, 5, 5, 38, 249, 5, 5, 40, 191, 5, 5, 47, 167, 5, 5, 193, 170, 5, 5, 142, 191, 5, 5, 141, 237, 5, 5, 5, 223, 5, 5, 4, 228, 5, 5, 165, 186, 5, 5, 230, 182, 5, 5, 219, 232, 5, 5, 103, 197, 5, 5, 9, 196, 5, 5, 128, 214, 5, 5, 191, 215, 5, 5, 171, 185, 5, 5, 44, 163, 5, 5, 133, 219, 5, 5, 66, 219, 5, 5, 69, 189, 5, 5, 0, 141, 5, 5, 127, 171, 5, 5, 184, 187, 5, 5, 250, 236, 5, 5, 93, 174, 5, 5, 113, 166, 5, 5, 120, 198, 5, 5, 193, 198, 5, 5, 230, 192, 5, 5, 176, 229, 5, 5, 200, 191, 5, 5, 108, 242, 5, 5, 30, 243, 5, 5, 9, 193, 5, 5, 39, 193, 5, 5, 246, 163, 5, 5, 81, 198, 5, 5, 200, 236, 5, 5, 96, 226, 5, 5, 55, 240, 5, 5, 23, 238, 5, 5, 44, 203, 5, 5, 185, 181, 5, 5, 91, 215, 5, 5, 203, 246, 5, 5, 54, 187, 5, 5, 186, 212, 5, 5, 68, 180, 5, 5, 10, 233, 5, 5, 8, 210, 5, 5, 177, 223, 5, 5, 19, 245, 5, 5, 80, 232, 5, 5, 11, 241, 5, 5, 182, 175, 5, 5, 6, 209, 5, 5, 168, 176, 5, 5, 16, 211, 5, 5, 47, 169, 5, 5, 109, 169, 5, 5, 233, 194, 5, 5, 193, 175, 5, 5, 250, 211, 5, 5, 172, 185, 5, 5, 59, 189, 5, 5, 223, 236, 5, 5, 81, 232, 5, 5, 213, 200, 5, 5, 147, 186, 5, 5, 126, 227, 5, 5, 214, 201, 5, 5, 253, 232, 5, 5, 171, 245, 5, 5, 249, 173, 5, 5, 222, 205, 5, 5, 45, 199, 5, 5, 42, 209, 5, 5, 23, 228, 5, 5, 15, 246, 5, 5, 172, 229, 5, 5, 42, 225, 5, 5, 36, 239, 5, 5, 68, 187, 5, 5, 83, 211, 5, 5, 230, 238, 5, 5, 244, 231, 5, 5, 51, 241, 5, 5, 92, 184, 5, 5, 147, 247, 5, 5, 101, 181, 5, 5, 64, 185, 5, 5, 151, 227, 5, 5, 197, 212, 5, 5, 66, 182, 5, 5, 229, 198, 5, 5, 34, 238, 5, 5, 213, 219, 5, 5, 86, 183, 5, 5, 199, 195, 5, 5, 33, 222, 5, 5, 34, 222, 5, 5, 230, 228, 5, 5, 233, 176, 5, 5, 195, 198, 5, 5, 223, 188, 5, 5, 137, 237, 5, 5, 18, 234, 5, 5, 43, 216, 5, 5, 130, 182, 5, 5, 59, 245, 5, 5, 62, 247, 5, 5, 238, 234, 5, 5, 152, 227, 5, 5, 45, 193, 5, 5, 9, 209, 5, 5, 10, 209, 5, 5, 180, 179, 5, 5, 29, 176, 5, 5, 30, 176, 5, 5, 93, 241, 5, 5, 183, 231, 5, 5, 237, 164, 5, 5, 239, 234, 5, 5, 22, 226, 5, 5, 109, 197, 5, 5, 12, 241, 5, 5, 12, 209, 5, 5, 137, 219, 5, 5, 155, 172, 5, 5, 60, 189, 5, 5, 92, 215, 5, 5, 247, 167, 5, 5, 153, 227, 5, 5, 6, 185, 5, 5, 245, 231, 5, 5, 97, 244, 5, 5, 131, 193, 5, 5, 156, 166, 5, 5, 166, 219, 5, 5, 226, 235, 5, 5, 138, 222, 5, 5, 236, 183, 5, 5, 142, 193, 5, 5, 86, 190, 5, 5, 128, 242, 5, 5, 234, 176, 5, 5, 63, 212, 5, 5, 129, 224, 5, 5, 150, 163, 5, 5, 217, 177, 5, 5, 102, 186, 5, 5, 220, 214, 5, 5, 241, 206, 5, 5, 71, 180, 5, 5, 96, 223, 5, 5, 165, 180, 5, 5, 225, 238, 5, 5, 131, 204, 5, 5, 35, 176, 5, 5, 241, 194, 5, 5, 241, 249, 5, 5, 127, 246, 5, 5, 105, 183, 5, 5, 96, 228, 5, 5, 81, 248, 5, 5, 144, 219, 5, 5, 239, 184, 5, 5, 105, 203, 5, 5, 28, 198, 5, 5, 107, 198, 5, 5, 36, 177, 5, 5, 16, 220, 5, 5, 213, 223, 5, 5, 215, 245, 5, 5, 161, 184, 5, 5, 82, 248, 5, 5, 131, 243, 5, 5, 222, 231, 5, 5, 147, 241, 5, 5, 139, 237, 5, 5, 203, 171, 5, 5, 5, 199, 5, 5, 167, 249, 5, 5, 66, 226, 5, 5, 83, 200, 5, 5, 221, 200, 5, 5, 222, 200, 5, 5, 80, 213, 5, 5, 79, 197, 5, 5, 223, 201, 5, 5, 155, 196, 5, 5, 228, 222, 5, 5, 20, 246, 5, 5, 156, 227, 5, 5, 14, 217, 5, 5, 41, 201, 5, 5, 66, 178, 5, 5, 238, 183, 5, 5, 100, 217, 5, 5, 73, 189, 5, 5, 83, 248, 5, 5, 124, 236, 5, 5, 12, 223, 5, 5, 163, 209, 5, 5, 39, 243, 5, 5, 238, 175, 5, 5, 105, 167, 5, 5, 67, 234, 5, 5, 181, 186, 5, 5, 52, 241, 5, 5, 68, 219, 5, 5, 227, 245, 5, 5, 145, 171, 5, 5, 160, 244, 5, 5, 96, 167, 5, 5, 115, 166, 5, 5, 118, 196, 5, 5, 127, 213, 5, 5, 84, 248, 5, 5, 157, 203, 5, 5, 230, 185, 5, 5, 30, 196, 5, 5, 84, 232, 5, 5, 75, 199, 5, 5, 173, 180, 5, 5, 174, 241, 5, 5, 158, 197, 5, 5, 136, 237, 5, 5, 20, 215, 5, 5, 37, 211, 5, 5, 231, 235, 5, 5, 216, 249, 5, 5, 160, 236, 5, 5, 183, 201, 5, 5, 18, 214, 5, 5, 61, 238, 5, 5, 208, 236, 5, 5, 62, 169, 5, 5, 135, 223, 5, 5, 199, 197, 5, 5, 12, 195, 5, 5, 9, 175, 5, 5, 116, 236, 5, 5, 171, 224, 5, 5, 219, 230, 5, 5, 43, 236, 5, 5, 144, 245, 5, 5, 233, 201, 5, 5, 127, 167, 5, 5, 3, 245, 5, 5, 154, 210, 5, 5, 79, 227, 5, 5, 174, 191, 5, 5, 122, 231, 5, 5, 115, 203, 5, 5, 143, 190, 5, 5, 116, 247, 5, 5, 159, 245, 5, 5, 184, 235, 5, 5, 15, 223, 5, 5, 220, 197, 5, 5, 245, 236, 192, 0, 0, 0, 5, 5, 182, 197, 5, 5, 228, 236, 5, 5, 234, 210, 5, 5, 243, 183, 5, 5, 81, 197, 5, 5, 120, 218, 5, 5, 11, 236, 5, 5, 222, 182, 5, 5, 208, 191, 5, 5, 226, 182, 5, 5, 18, 177, 5, 5, 28, 238, 5, 5, 86, 211, 5, 5, 159, 182, 5, 5, 52, 165, 5, 5, 104, 238, 5, 5, 105, 238, 5, 5, 129, 188, 5, 5, 27, 233, 5, 5, 123, 219, 5, 5, 124, 169, 5, 5, 76, 236, 5, 5, 68, 236, 5, 5, 31, 180, 5, 5, 178, 176, 5, 5, 136, 208, 5, 5, 97, 204, 5, 5, 156, 226, 5, 5, 253, 203, 5, 5, 98, 195, 5, 5, 44, 180, 5, 5, 62, 228, 5, 5, 36, 234, 5, 5, 31, 203, 5, 5, 38, 238, 5, 5, 152, 214, 5, 5, 95, 168, 5, 5, 235, 185, 5, 5, 120, 191, 5, 5, 59, 193, 5, 5, 153, 214, 5, 5, 235, 198, 5, 5, 100, 200, 5, 5, 148, 193, 5, 5, 189, 238, 5, 5, 59, 215, 5, 5, 252, 231, 5, 5, 15, 225, 5, 5, 205, 216, 5, 5, 180, 166, 5, 5, 55, 175, 5, 5, 63, 169, 5, 5, 50, 169, 5, 5, 16, 199, 5, 5, 163, 185, 5, 5, 236, 177, 5, 5, 121, 191, 5, 5, 53, 180, 5, 5, 75, 241, 5, 5, 184, 164, 5, 5, 110, 244, 5, 5, 16, 236, 5, 5, 98, 204, 5, 5, 107, 171, 5, 5, 225, 214, 5, 5, 104, 199, 5, 5, 171, 169, 5, 5, 68, 230, 5, 5, 23, 226, 5, 5, 146, 167, 5, 5, 179, 185, 5, 5, 227, 214, 5, 5, 244, 201, 5, 5, 199, 221, 5, 5, 244, 223, 5, 5, 164, 196, 5, 5, 32, 233, 5, 5, 158, 214, 5, 5, 130, 235, 5, 5, 227, 197, 5, 5, 176, 174, 5, 5, 232, 187, 5, 5, 167, 209, 5, 5, 245, 212, 5, 5, 228, 225, 5, 5, 139, 224, 5, 5, 47, 176, 5, 5, 208, 197, 5, 5, 20, 218, 5, 5, 185, 237, 5, 5, 80, 233, 5, 5, 242, 172, 5, 5, 20, 223, 5, 5, 13, 200, 5, 5, 135, 232, 5, 5, 132, 226, 5, 5, 158, 234, 5, 5, 48, 176, 5, 5, 128, 248, 5, 5, 133, 182, 5, 5, 134, 234, 5, 5, 102, 204, 5, 5, 57, 241, 5, 5, 246, 212, 5, 5, 52, 209, 5, 5, 223, 197, 5, 5, 138, 188, 5, 5, 207, 216, 5, 5, 124, 176, 5, 5, 28, 200, 5, 5, 190, 243, 5, 5, 216, 168, 5, 5, 109, 207, 5, 5, 240, 210, 5, 5, 239, 217, 5, 5, 246, 183, 5, 5, 181, 189, 5, 5, 131, 235, 5, 5, 248, 242, 5, 5, 228, 244, 5, 5, 13, 175, 5, 5, 89, 220, 5, 5, 140, 237, 5, 5, 66, 220, 5, 5, 242, 205, 5, 5, 196, 184, 5, 5, 113, 200, 5, 5, 89, 214, 5, 5, 56, 209, 5, 5, 252, 237, 5, 5, 87, 232, 5, 5, 113, 239, 5, 5, 97, 227, 5, 5, 105, 204, 5, 5, 106, 249, 5, 5, 32, 238, 5, 5, 224, 249, 5, 5, 177, 207, 5, 5, 92, 196, 5, 5, 38, 248, 5, 5, 97, 167, 5, 5, 168, 175, 5, 5, 124, 191, 5, 5, 70, 197, 5, 5, 190, 178, 5, 5, 74, 176, 5, 5, 37, 216, 5, 5, 5, 236, 5, 5, 226, 167, 5, 5, 32, 168, 5, 5, 153, 243, 5, 5, 34, 233, 5, 5, 123, 184, 5, 5, 187, 176, 5, 5, 93, 186, 5, 5, 41, 179, 5, 5, 170, 199, 5, 5, 197, 230, 5, 5, 213, 197, 5, 5, 233, 213, 5, 5, 120, 241, 5, 5, 60, 178, 5, 5, 49, 168, 5, 5, 39, 224, 5, 5, 221, 202, 5, 5, 248, 205, 5, 5, 232, 197, 5, 5, 236, 180, 5, 5, 22, 225, 5, 5, 27, 218, 5, 5, 160, 169, 5, 5, 224, 233, 5, 5, 239, 180, 5, 5, 245, 189, 5, 5, 130, 200, 5, 5, 123, 163, 5, 5, 20, 242, 5, 5, 4, 184, 5, 5, 195, 243, 5, 5, 90, 199, 5, 5, 161, 169, 5, 5, 125, 201, 5, 5, 157, 219, 5, 5, 218, 186, 5, 5, 69, 188, 5, 5, 205, 197, 5, 5, 128, 183, 5, 5, 174, 214, 5, 5, 79, 169, 5, 5, 74, 243, 5, 5, 121, 187, 5, 5, 200, 189, 5, 5, 251, 190, 5, 5, 82, 169, 5, 5, 60, 176, 5, 5, 231, 224, 5, 5, 146, 168, 5, 5, 35, 168, 5, 5, 112, 216, 5, 5, 215, 199, 5, 5, 83, 169, 5, 5, 158, 213, 5, 5, 95, 178, 5, 5, 38, 196, 5, 5, 200, 242, 5, 5, 108, 224, 5, 5, 135, 244, 5, 5, 163, 230, 5, 5, 158, 177, 5, 5, 238, 184, 5, 5, 85, 179, 5, 5, 182, 214, 5, 5, 194, 207, 5, 5, 16, 226, 5, 5, 159, 200, 5, 5, 226, 199, 5, 5, 153, 222, 5, 5, 13, 184, 5, 5, 183, 214, 5, 5, 229, 233, 5, 5, 232, 186, 5, 5, 153, 224, 5, 5, 222, 223, 5, 5, 152, 187, 5, 5, 204, 240, 5, 5, 88, 190, 5, 5, 130, 183, 5, 5, 230, 231, 5, 5, 138, 236, 5, 5, 100, 213, 5, 5, 171, 232, 5, 5, 8, 205, 5, 5, 231, 231, 5, 5, 204, 221, 5, 5, 56, 186, 5, 5, 21, 184, 5, 5, 126, 183, 5, 5, 85, 186, 5, 5, 9, 185, 5, 5, 255, 245, 5, 5, 38, 200, 5, 5, 109, 243, 5, 5, 245, 209, 5, 5, 157, 235, 5, 5, 97, 184, 5, 5, 12, 205, 5, 5, 39, 200, 5, 5, 20, 216, 5, 5, 110, 243, 5, 5, 94, 183, 5, 5, 108, 182, 5, 5, 109, 182, 5, 5, 204, 178, 5, 5, 232, 195, 5, 5, 247, 248, 5, 5, 92, 208, 5, 5, 99, 168, 5, 5, 129, 249, 5, 5, 38, 235, 5, 5, 246, 198, 5, 5, 40, 200, 5, 5, 30, 249, 5, 5, 153, 237, 5, 5, 158, 193, 5, 5, 240, 196, 5, 5, 39, 235, 5, 5, 155, 182, 5, 5, 196, 181, 5, 5, 4, 246, 5, 5, 193, 190, 5, 5, 159, 193, 5, 5, 117, 212, 5, 5, 10, 230, 5, 5, 87, 189, 5, 5, 79, 193, 5, 5, 160, 193, 5, 5, 102, 179, 5, 5, 161, 193, 5, 5, 197, 181, 5, 5, 101, 242, 5, 5, 162, 193, 5, 5, 168, 243, 5, 5, 21, 201, 5, 5, 246, 168, 5, 5, 13, 248, 5, 5, 82, 198, 5, 5, 39, 249, 5, 5, 199, 190, 5, 5, 46, 212, 5, 5, 18, 248, 5, 5, 95, 211, 5, 5, 163, 193, 5, 5, 57, 226, 5, 5, 21, 246, 5, 5, 44, 249, 5, 5, 47, 212, 5, 5, 14, 196, 5, 5, 222, 174, 5, 5, 51, 249, 5, 5, 52, 249, 5, 5, 43, 241, 5, 5, 214, 181, 5, 5, 222, 181, 5, 5, 223, 181, 5, 5, 112, 182, 5, 5, 112, 191, 5, 5, 103, 200, 5, 5, 145, 240, 5, 5, 184, 200, 5, 5, 167, 190, 5, 5, 178, 190, 5, 5, 237, 181, 5, 5, 19, 241, 5, 5, 238, 181, 5, 5, 164, 181, 5, 5, 40, 229, 5, 5, 244, 168, 5, 5, 255, 189, 5, 5, 100, 196, 5, 5, 221, 240, 5, 5, 43, 232, 5, 5, 3, 246, 5, 5, 132, 191, 5, 5, 50, 248, 5, 5, 241, 244, 5, 5, 208, 193, 5, 5, 24, 168, 5, 5, 129, 239, 5, 5, 66, 175, 5, 5, 52, 210, 5, 5, 53, 210, 5, 5, 233, 220, 5, 5, 19, 222, 5, 5, 94, 176, 5, 5, 78, 209, 5, 5, 239, 213, 5, 5, 129, 228, 5, 5, 20, 197, 5, 5, 27, 225, 5, 5, 156, 208, 5, 5, 229, 183, 5, 5, 96, 169, 5, 5, 206, 201, 5, 5, 40, 203, 5, 5, 145, 218, 5, 5, 146, 218, 5, 5, 83, 164, 5, 5, 185, 217, 5, 5, 15, 240, 5, 5, 254, 210, 5, 5, 191, 169, 5, 5, 16, 208, 5, 5, 97, 231, 5, 5, 138, 182, 5, 5, 147, 173, 5, 5, 128, 164, 5, 5, 129, 164, 5, 5, 226, 179, 5, 5, 170, 195, 5, 5, 246, 199, 5, 5, 2, 233, 5, 5, 242, 195, 5, 5, 250, 164, 5, 5, 118, 206, 5, 5, 222, 176, 5, 5, 75, 206, 5, 5, 153, 202, 5, 5, 79, 196, 5, 5, 200, 176, 5, 5, 76, 206, 5, 5, 192, 204, 5, 5, 241, 179, 5, 5, 33, 187, 5, 5, 15, 163, 5, 5, 87, 245, 5, 5, 92, 173, 5, 5, 89, 222, 5, 5, 169, 166, 5, 5, 101, 164, 5, 5, 50, 180, 5, 5, 196, 196, 5, 5, 157, 211, 5, 5, 23, 180, 5, 5, 61, 218, 5, 5, 21, 182, 5, 5, 111, 192, 5, 5, 171, 168, 5, 5, 161, 177, 5, 5, 128, 193, 5, 5, 147, 183, 5, 5, 108, 179, 5, 5, 200, 167, 5, 5, 83, 208, 5, 5, 59, 244, 5, 5, 149, 172, 5, 5, 141, 238, 5, 5, 4, 190, 5, 5, 216, 209, 5, 5, 64, 172, 5, 5, 38, 195, 5, 5, 79, 192, 5, 5, 81, 217, 5, 5, 0, 157, 5, 5, 33, 223, 5, 5, 187, 223, 5, 5, 88, 189, 5, 5, 212, 221, 5, 5, 192, 174, 5, 5, 219, 198, 5, 5, 188, 218, 5, 5, 28, 226, 5, 5, 226, 184, 5, 5, 27, 191, 5, 5, 28, 191, 5, 5, 185, 189, 5, 5, 142, 238, 5, 5, 80, 175, 5, 5, 14, 167, 5, 5, 24, 172, 5, 5, 144, 216, 5, 5, 6, 179, 5, 5, 162, 202, 5, 5, 153, 197, 5, 5, 86, 191, 5, 5, 169, 233, 5, 5, 63, 223, 5, 5, 119, 185, 5, 5, 214, 191, 5, 5, 114, 243, 5, 5, 48, 198, 5, 5, 120, 178, 5, 5, 228, 219, 5, 5, 80, 174, 5, 5, 160, 194, 5, 5, 159, 241, 5, 5, 79, 178, 5, 5, 80, 178, 5, 5, 5, 194, 5, 5, 143, 223, 5, 5, 27, 244, 5, 5, 165, 243, 5, 5, 123, 172, 5, 5, 83, 192, 5, 5, 34, 249, 5, 5, 22, 177, 5, 5, 2, 214, 5, 5, 197, 195, 5, 5, 244, 228, 5, 5, 16, 228, 5, 5, 202, 237, 5, 5, 98, 166, 5, 5, 164, 177, 5, 5, 139, 201, 5, 5, 227, 240, 5, 5, 44, 196, 5, 5, 50, 175, 5, 5, 248, 194, 5, 5, 97, 176, 5, 5, 85, 195, 5, 5, 191, 198, 5, 5, 64, 183, 5, 5, 228, 238, 5, 5, 120, 216, 5, 5, 192, 215, 5, 5, 95, 249, 5, 5, 141, 164, 5, 5, 28, 174, 5, 5, 18, 163, 5, 5, 56, 240, 5, 5, 236, 243, 5, 5, 121, 198, 5, 5, 255, 216, 5, 5, 30, 226, 5, 5, 221, 210, 5, 5, 231, 192, 5, 5, 24, 238, 5, 5, 187, 164, 5, 5, 162, 173, 5, 5, 14, 212, 5, 5, 255, 249, 5, 5, 44, 208, 5, 5, 15, 203, 5, 5, 230, 176, 5, 5, 164, 169, 5, 5, 139, 186, 5, 5, 233, 207, 5, 5, 93, 225, 5, 5, 53, 228, 5, 5, 169, 205, 5, 5, 134, 219, 5, 5, 185, 187, 5, 5, 186, 187, 5, 5, 133, 166, 5, 5, 99, 225, 5, 5, 157, 199, 5, 5, 134, 166, 5, 5, 243, 226, 5, 5, 187, 217, 5, 5, 141, 193, 5, 5, 85, 208, 5, 5, 73, 241, 5, 5, 93, 184, 5, 5, 213, 176, 5, 5, 71, 248, 5, 5, 138, 166, 5, 5, 231, 240, 5, 5, 64, 212, 5, 5, 25, 237, 5, 5, 248, 167, 5, 5, 96, 211, 5, 5, 206, 246, 5, 5, 63, 207, 5, 5, 217, 190, 5, 5, 77, 221, 5, 5, 54, 197, 5, 5, 176, 226, 5, 5, 233, 178, 5, 5, 96, 175, 5, 5, 39, 180, 5, 5, 231, 238, 5, 5, 26, 234, 5, 5, 215, 225, 5, 5, 188, 164, 5, 5, 174, 195, 5, 5, 108, 218, 5, 5, 83, 178, 5, 5, 69, 240, 5, 5, 109, 218, 5, 5, 70, 240, 5, 5, 235, 166, 5, 5, 221, 201, 5, 5, 163, 222, 5, 5, 183, 208, 5, 5, 34, 236, 5, 5, 64, 207, 5, 5, 156, 183, 5, 5, 11, 196, 5, 5, 61, 195, 5, 5, 69, 180, 5, 5, 198, 196, 5, 5, 209, 183, 5, 5, 205, 204, 5, 5, 121, 178, 5, 5, 188, 205, 5, 5, 193, 192, 5, 5, 72, 248, 5, 5, 214, 237, 5, 5, 154, 167, 5, 5, 83, 193, 5, 5, 245, 181, 5, 5, 161, 186, 5, 5, 183, 173, 5, 5, 249, 240, 5, 5, 172, 210, 5, 5, 244, 196, 5, 5, 245, 238, 5, 5, 225, 228, 5, 5, 107, 178, 5, 5, 69, 183, 5, 5, 188, 181, 5, 5, 111, 225, 5, 5, 132, 243, 5, 5, 8, 201, 5, 5, 28, 165, 5, 5, 190, 188, 5, 5, 145, 193, 5, 5, 37, 177, 5, 5, 139, 222, 5, 5, 54, 198, 5, 5, 22, 164, 5, 5, 23, 164, 5, 5, 190, 247, 5, 5, 154, 176, 5, 5, 30, 170, 5, 5, 106, 203, 5, 5, 22, 231, 5, 5, 96, 241, 5, 5, 247, 192, 5, 5, 165, 195, 5, 5, 91, 233, 5, 5, 191, 247, 5, 5, 42, 201, 5, 5, 224, 237, 5, 5, 169, 241, 5, 5, 40, 243, 5, 5, 49, 164, 5, 5, 172, 233, 5, 5, 205, 169, 5, 5, 30, 225, 5, 5, 23, 178, 5, 5, 35, 182, 5, 5, 238, 243, 5, 5, 240, 173, 5, 5, 250, 222, 5, 5, 224, 203, 5, 5, 175, 205, 5, 5, 211, 198, 5, 5, 198, 192, 5, 5, 80, 226, 5, 5, 26, 205, 5, 5, 210, 168, 5, 5, 162, 185, 5, 5, 100, 216, 5, 5, 35, 201, 5, 5, 151, 186, 5, 5, 166, 178, 5, 5, 176, 246, 5, 5, 101, 217, 5, 5, 188, 202, 5, 5, 78, 202, 5, 5, 86, 196, 5, 5, 169, 244, 5, 5, 37, 179, 5, 5, 118, 225, 5, 5, 13, 250, 5, 5, 92, 195, 5, 5, 161, 236, 5, 5, 162, 236, 5, 5, 249, 226, 5, 5, 103, 171, 5, 5, 44, 179, 5, 5, 125, 247, 5, 5, 119, 225, 5, 5, 215, 181, 5, 5, 75, 180, 5, 5, 23, 212, 5, 5, 230, 222, 5, 5, 88, 187, 5, 5, 231, 222, 5, 5, 165, 164, 5, 5, 2, 236, 5, 5, 189, 232, 5, 5, 122, 178, 5, 5, 85, 181, 5, 5, 111, 245, 5, 5, 120, 221, 5, 5, 190, 217, 5, 5, 175, 191, 5, 5, 250, 220, 5, 5, 62, 238, 5, 5, 204, 194, 5, 5, 181, 242, 5, 5, 178, 195, 5, 5, 86, 181, 5, 5, 53, 249, 5, 5, 255, 246, 5, 5, 81, 214, 5, 5, 102, 188, 5, 5, 204, 197, 5, 5, 133, 186, 5, 5, 16, 225, 5, 5, 163, 167, 5, 5, 198, 164, 5, 5, 200, 224, 5, 5, 36, 190, 5, 5, 235, 189, 5, 5, 169, 238, 5, 5, 179, 210, 5, 5, 218, 169, 5, 5, 125, 169, 5, 5, 118, 222, 5, 5, 252, 229, 5, 5, 103, 229, 5, 5, 255, 185, 5, 5, 50, 223, 5, 5, 138, 170, 5, 5, 149, 165, 5, 5, 173, 206, 5, 5, 251, 233, 5, 5, 153, 229, 5, 5, 4, 219, 5, 5, 13, 221, 5, 5, 189, 195, 5, 5, 218, 249, 5, 5, 176, 177, 5, 5, 214, 224, 5, 5, 124, 218, 5, 5, 137, 213, 5, 5, 149, 193, 5, 5, 77, 211, 5, 5, 232, 205, 5, 5, 255, 176, 5, 5, 233, 228, 5, 5, 99, 172, 5, 5, 214, 234, 5, 5, 55, 243, 5, 5, 46, 179, 5, 5, 194, 232, 5, 5, 132, 210, 5, 5, 59, 211, 5, 5, 49, 189, 5, 5, 169, 246, 5, 5, 102, 222, 5, 5, 141, 209, 5, 5, 195, 232, 5, 5, 208, 216, 5, 5, 100, 228, 5, 5, 101, 228, 5, 5, 8, 163, 5, 5, 232, 168, 5, 5, 91, 196, 5, 5, 250, 178, 5, 5, 20, 173, 5, 5, 30, 212, 5, 5, 139, 169, 5, 5, 186, 178, 5, 5, 187, 178, 5, 5, 68, 233, 5, 5, 7, 177, 5, 5, 51, 229, 5, 5, 115, 178, 5, 5, 116, 178, 5, 5, 254, 179, 5, 5, 137, 176, 5, 5, 59, 172, 5, 5, 108, 173, 5, 5, 57, 201, 5, 5, 255, 179, 5, 5, 51, 175, 5, 5, 106, 241, 5, 5, 253, 221, 5, 5, 155, 205, 5, 5, 153, 238, 5, 5, 65, 211, 5, 5, 253, 201, 5, 5, 143, 209, 5, 5, 202, 195, 5, 5, 146, 240, 5, 5, 12, 168, 5, 5, 75, 165, 5, 5, 177, 177, 5, 5, 22, 206, 5, 5, 77, 241, 5, 5, 225, 249, 5, 5, 66, 217, 5, 5, 6, 180, 5, 5, 72, 186, 5, 5, 89, 207, 5, 5, 249, 199, 5, 5, 112, 176, 5, 5, 162, 230, 5, 5, 191, 210, 5, 5, 197, 217, 5, 5, 229, 215, 5, 5, 93, 213, 5, 5, 169, 197, 5, 5, 75, 238, 5, 5, 237, 240, 5, 5, 149, 175, 5, 5, 233, 197, 5, 5, 24, 231, 5, 5, 6, 206, 5, 5, 131, 185, 5, 5, 36, 197, 5, 5, 201, 189, 5, 5, 198, 242, 5, 5, 15, 180, 5, 5, 109, 246, 5, 5, 250, 199, 5, 5, 34, 191, 5, 5, 35, 191, 5, 5, 21, 231, 5, 5, 122, 187, 5, 5, 252, 188, 5, 5, 164, 230, 5, 5, 166, 224, 5, 5, 207, 217, 5, 5, 205, 242, 5, 5, 109, 185, 5, 5, 30, 163, 5, 5, 182, 246, 5, 5, 22, 227, 5, 5, 86, 230, 5, 5, 106, 236, 5, 5, 183, 222, 5, 5, 184, 222, 5, 5, 10, 229, 5, 5, 90, 239, 5, 5, 176, 169, 5, 5, 33, 195, 5, 5, 219, 221, 5, 5, 187, 222, 5, 5, 251, 198, 5, 5, 255, 198, 5, 5, 228, 196, 5, 5, 102, 239, 5, 5, 6, 199, 5, 5, 15, 217, 5, 5, 233, 196, 5, 5, 28, 228, 5, 5, 217, 170, 5, 5, 69, 233, 5, 5, 30, 228, 5, 5, 114, 239, 5, 5, 140, 218, 5, 5, 30, 193, 5, 5, 202, 225, 5, 5, 52, 206, 5, 5, 236, 173, 5, 5, 202, 163, 5, 5, 203, 219, 5, 5, 239, 165, 5, 5, 246, 236, 5, 5, 116, 212, 5, 5, 89, 189, 5, 5, 198, 174, 5, 5, 196, 234, 5, 5, 230, 213, 5, 5, 50, 218, 5, 5, 119, 214, 5, 5, 60, 183, 5, 5, 236, 213, 5, 5, 141, 196, 5, 5, 165, 212, 5, 5, 218, 245, 5, 5, 89, 249, 5, 5, 45, 170, 5, 5, 141, 236, 5, 5, 49, 178, 5, 5, 50, 178, 5, 5, 83, 241, 5, 5, 108, 211, 5, 5, 220, 245, 5, 5, 87, 192, 5, 5, 88, 192, 5, 5, 43, 209, 5, 5, 24, 171, 5, 5, 25, 171, 5, 5, 187, 243, 5, 5, 183, 199, 5, 5, 215, 224, 5, 5, 47, 175, 5, 5, 19, 232, 5, 5, 61, 181, 5, 5, 132, 179, 5, 5, 17, 174, 5, 5, 202, 193, 5, 5, 140, 241, 5, 5, 254, 243, 5, 5, 171, 198, 5, 5, 23, 192, 5, 5, 184, 226, 5, 5, 148, 179, 5, 5, 133, 179, 5, 5, 211, 179, 5, 5, 211, 242, 5, 5, 223, 164, 5, 5, 130, 239, 5, 5, 39, 196, 5, 5, 255, 202, 5, 5, 61, 178, 5, 5, 133, 199, 5, 5, 16, 197, 5, 5, 160, 249, 5, 5, 3, 203, 5, 5, 0, 135, 5, 5, 2, 239, 5, 5, 115, 188, 5, 5, 209, 200, 5, 5, 150, 176, 5, 5, 216, 242, 5, 5, 26, 167, 5, 5, 241, 207, 5, 5, 162, 208, 5, 5, 22, 183, 5, 5, 253, 173, 5, 5, 6, 214, 5, 5, 129, 173, 5, 5, 146, 177, 5, 5, 89, 244, 5, 5, 114, 196, 5, 5, 117, 232, 5, 5, 211, 200, 5, 5, 36, 209, 5, 5, 77, 195, 5, 5, 62, 219, 5, 5, 115, 164, 5, 5, 41, 240, 5, 5, 36, 206, 5, 5, 203, 179, 5, 5, 186, 232, 5, 5, 124, 164, 5, 5, 187, 212, 5, 5, 90, 201, 5, 5, 183, 172, 5, 5, 77, 177, 5, 5, 13, 218, 5, 5, 177, 164, 5, 5, 171, 177, 5, 5, 144, 238, 5, 5, 168, 227, 5, 5, 124, 247, 5, 5, 74, 197, 5, 5, 145, 238, 5, 5, 112, 245, 5, 5, 147, 238, 5, 5, 39, 206, 5, 5, 130, 188, 5, 5, 71, 178, 5, 5, 116, 249, 5, 5, 41, 206, 5, 5, 92, 227, 5, 5, 137, 173, 5, 5, 208, 199, 5, 5, 91, 165, 5, 5, 168, 190, 5, 5, 96, 178, 5, 5, 212, 184, 5, 5, 141, 184, 5, 5, 200, 177, 5, 5, 12, 183, 5, 5, 54, 210, 5, 5, 189, 204, 5, 5, 32, 179, 5, 5, 145, 204, 5, 5, 179, 184, 5, 5, 162, 223, 5, 5, 239, 197, 5, 5, 38, 215, 5, 5, 47, 173, 5, 5, 126, 202, 5, 5, 179, 229, 5, 5, 181, 230, 5, 5, 152, 172, 5, 5, 106, 247, 5, 5, 129, 201, 5, 5, 169, 222, 5, 5, 189, 174, 5, 5, 18, 169, 5, 5, 226, 175, 5, 5, 119, 230, 5, 5, 20, 174, 5, 5, 141, 206, 5, 5, 165, 242, 5, 5, 194, 197, 5, 5, 60, 247, 5, 5, 197, 222, 5, 5, 150, 166, 5, 5, 31, 178, 5, 5, 154, 202, 5, 5, 243, 195, 5, 5, 241, 222, 5, 5, 142, 241, 5, 5, 210, 205, 5, 5, 130, 249, 5, 5, 165, 171, 5, 5, 232, 172, 5, 5, 122, 223, 5, 5, 200, 248, 5, 5, 171, 222, 5, 5, 44, 184, 5, 5, 22, 243, 5, 5, 234, 182, 5, 5, 242, 215, 5, 5, 93, 173, 5, 5, 227, 164, 5, 5, 160, 247, 5, 5, 54, 182, 5, 5, 190, 173, 5, 5, 234, 246, 5, 5, 158, 173, 5, 5, 47, 245, 5, 5, 52, 248, 5, 5, 209, 193, 5, 5, 35, 178, 5, 5, 0, 158, 5, 5, 145, 183, 5, 5, 31, 220, 5, 5, 25, 222, 5, 5, 207, 219, 5, 5, 19, 175, 5, 5, 16, 183, 5, 5, 91, 249, 5, 5, 216, 179, 5, 5, 84, 215, 5, 5, 101, 218, 5, 5, 144, 237, 5, 5, 250, 169, 5, 5, 97, 169, 5, 5, 116, 208, 5, 5, 254, 164, 5, 5, 140, 228, 5, 5, 95, 209, 5, 5, 145, 237, 5, 5, 90, 197, 5, 5, 227, 216, 5, 5, 122, 172, 5, 5, 64, 221, 5, 5, 223, 171, 5, 5, 102, 164, 5, 5, 65, 172, 5, 5, 22, 174, 5, 5, 90, 189, 5, 5, 170, 231, 5, 5, 171, 231, 5, 5, 14, 227, 5, 5, 76, 230, 5, 5, 43, 228, 5, 5, 2, 244, 5, 5, 105, 170, 5, 5, 196, 203, 5, 5, 2, 193, 5, 5, 163, 235, 5, 5, 164, 216, 5, 5, 52, 246, 5, 5, 109, 231, 5, 5, 43, 207, 5, 5, 44, 228, 5, 5, 128, 168, 5, 5, 127, 230, 5, 5, 160, 241, 5, 5, 7, 179, 5, 5, 8, 206, 5, 5, 67, 229, 5, 5, 56, 202, 5, 5, 215, 191, 5, 5, 95, 216, 5, 5, 110, 199, 5, 5, 214, 207, 5, 5, 239, 203, 5, 5, 84, 192, 5, 5, 196, 185, 5, 5, 200, 204, 5, 5, 201, 204, 5, 5, 143, 239, 5, 5, 91, 184, 5, 5, 103, 218, 5, 5, 212, 201, 5, 5, 215, 201, 5, 5, 187, 215, 5, 5, 198, 198, 5, 5, 161, 194, 5, 5, 19, 208, 5, 5, 40, 196, 5, 5, 26, 243, 5, 5, 50, 172, 5, 5, 181, 187, 5, 5, 249, 166, 5, 5, 203, 237, 5, 5, 102, 242, 5, 5, 166, 242, 5, 5, 167, 223, 5, 5, 144, 228, 5, 5, 115, 243, 5, 5, 236, 226, 5, 5, 6, 194, 5, 5, 56, 212, 5, 5, 110, 231, 5, 5, 91, 237, 5, 5, 223, 245, 5, 5, 140, 201, 5, 5, 151, 245, 5, 5, 227, 185, 5, 5, 29, 241, 5, 5, 17, 230, 5, 5, 240, 203, 5, 5, 157, 190, 5, 5, 159, 168, 5, 5, 193, 215, 5, 5, 5, 228, 5, 5, 254, 186, 5, 5, 94, 225, 5, 5, 59, 214, 5, 5, 194, 170, 5, 5, 29, 174, 5, 5, 202, 236, 5, 5, 199, 169, 5, 5, 59, 246, 5, 5, 57, 195, 5, 5, 252, 225, 5, 5, 253, 225, 5, 5, 167, 181, 5, 5, 102, 184, 5, 5, 148, 216, 5, 5, 101, 183, 5, 5, 177, 247, 5, 5, 239, 182, 5, 5, 38, 167, 5, 5, 33, 183, 5, 5, 215, 207, 5, 5, 150, 228, 5, 5, 204, 218, 5, 5, 83, 198, 5, 5, 231, 176, 5, 5, 172, 222, 5, 5, 156, 237, 5, 5, 137, 168, 5, 5, 95, 182, 5, 5, 194, 215, 5, 5, 197, 171, 5, 5, 20, 245, 5, 5, 170, 242, 5, 5, 60, 245, 5, 5, 25, 238, 5, 5, 168, 242, 5, 5, 178, 247, 5, 5, 174, 235, 5, 5, 179, 247, 5, 5, 139, 245, 5, 5, 55, 187, 5, 5, 90, 215, 5, 5, 206, 206, 5, 5, 239, 211, 5, 5, 92, 222, 5, 5, 122, 208, 5, 5, 93, 215, 5, 5, 35, 222, 5, 5, 234, 178, 5, 5, 240, 224, 5, 5, 12, 190, 5, 5, 46, 238, 5, 5, 147, 237, 5, 5, 148, 237, 5, 5, 229, 185, 5, 5, 176, 171, 5, 5, 77, 235, 5, 5, 48, 167, 5, 5, 242, 206, 5, 5, 179, 166, 5, 5, 5, 234, 5, 5, 108, 189, 5, 5, 207, 246, 5, 5, 183, 165, 5, 5, 136, 214, 5, 5, 100, 229, 5, 5, 11, 209, 5, 5, 12, 200, 5, 5, 56, 173, 5, 5, 184, 245, 5, 5, 249, 230, 5, 5, 246, 179, 5, 5, 156, 172, 5, 5, 226, 172, 5, 5, 152, 203, 5, 5, 63, 181, 5, 5, 157, 183, 5, 5, 65, 212, 5, 5, 156, 206, 5, 5, 248, 200, 5, 5, 86, 206, 5, 5, 76, 210, 5, 5, 168, 177, 5, 5, 169, 216, 5, 5, 72, 199, 5, 5, 174, 246, 5, 5, 41, 185, 5, 5, 20, 165, 5, 5, 125, 243, 5, 5, 10, 224, 5, 5, 79, 183, 5, 5, 210, 174, 5, 5, 157, 206, 5, 5, 200, 190, 5, 5, 246, 238, 5, 5, 20, 186, 5, 5, 199, 196, 5, 5, 175, 243, 5, 5, 34, 241, 5, 5, 207, 206, 5, 5, 69, 187, 5, 5, 148, 177, 5, 5, 94, 215, 5, 5, 158, 199, 5, 5, 139, 199, 5, 5, 98, 244, 5, 5, 78, 224, 5, 5, 190, 228, 5, 5, 149, 181, 5, 5, 100, 176, 5, 5, 223, 241, 5, 5, 106, 245, 5, 5, 120, 183, 5, 5, 3, 226, 5, 5, 76, 203, 5, 5, 206, 179, 5, 5, 16, 217, 5, 5, 85, 248, 5, 5, 22, 246, 5, 5, 244, 176, 5, 5, 35, 230, 5, 5, 240, 169, 5, 5, 239, 175, 5, 5, 31, 231, 5, 5, 205, 210, 5, 5, 14, 200, 5, 5, 196, 245, 5, 5, 127, 210, 5, 5, 81, 226, 5, 5, 148, 171, 5, 5, 214, 174, 5, 5, 198, 203, 5, 5, 158, 178, 5, 5, 176, 239, 5, 5, 251, 222, 5, 5, 206, 169, 5, 5, 43, 201, 5, 5, 178, 171, 5, 5, 206, 238, 5, 5, 103, 244, 5, 5, 102, 217, 5, 5, 107, 203, 5, 5, 229, 245, 5, 5, 5, 170, 5, 5, 99, 249, 5, 5, 86, 221, 5, 5, 224, 209, 5, 5, 70, 227, 5, 5, 72, 214, 5, 5, 56, 182, 5, 5, 192, 188, 5, 5, 67, 170, 5, 5, 80, 230, 5, 5, 114, 235, 5, 5, 101, 168, 5, 5, 128, 246, 5, 5, 142, 245, 5, 5, 169, 204, 5, 5, 7, 199, 5, 5, 135, 238, 5, 5, 127, 209, 5, 5, 24, 164, 5, 5, 194, 216, 5, 5, 207, 179, 5, 5, 225, 193, 5, 5, 160, 225, 5, 5, 175, 241, 5, 5, 41, 168, 5, 5, 51, 213, 5, 5, 54, 249, 5, 5, 225, 244, 5, 5, 64, 231, 5, 5, 220, 230, 5, 5, 16, 223, 5, 5, 43, 188, 5, 5, 243, 169, 5, 5, 61, 167, 5, 5, 136, 238, 5, 5, 20, 169, 5, 5, 90, 192, 5, 5, 160, 196, 5, 5, 235, 210, 5, 5, 90, 203, 5, 5, 32, 236, 5, 5, 178, 220, 5, 5, 55, 249, 5, 5, 116, 203, 5, 5, 65, 231, 5, 5, 36, 172, 5, 5, 106, 244, 5, 5, 79, 202, 5, 5, 163, 236, 5, 5, 34, 184, 5, 5, 106, 215, 5, 5, 154, 179, 5, 5, 120, 167, 5, 5, 113, 186, 5, 5, 71, 183, 5, 5, 9, 164, 5, 5, 63, 238, 5, 5, 42, 168, 5, 5, 140, 168, 5, 5, 120, 240, 5, 5, 79, 168, 5, 5, 113, 248, 5, 5, 88, 185, 5, 5, 140, 217, 5, 5, 103, 170, 5, 5, 107, 239, 5, 5, 14, 221, 5, 5, 215, 195, 5, 5, 219, 249, 5, 5, 215, 183, 5, 5, 197, 186, 5, 5, 179, 176, 5, 5, 230, 193, 5, 5, 143, 242, 5, 5, 194, 205, 5, 5, 119, 222, 5, 5, 17, 225, 5, 5, 253, 231, 5, 5, 67, 198, 5, 5, 124, 183, 5, 5, 211, 185, 5, 5, 70, 169, 5, 5, 124, 219, 5, 5, 79, 199, 5, 5, 80, 199, 5, 5, 150, 238, 5, 5, 125, 183, 5, 5, 60, 211, 5, 5, 64, 169, 5, 5, 67, 168, 5, 5, 24, 193, 5, 5, 69, 245, 5, 5, 168, 209, 5, 5, 59, 182, 5, 5, 123, 220, 5, 5, 35, 170, 5, 5, 109, 215, 5, 5, 110, 215, 5, 5, 132, 169, 5, 5, 233, 187, 5, 5, 139, 188, 5, 5, 147, 199, 5, 5, 235, 193, 5, 5, 125, 176, 5, 5, 127, 218, 5, 5, 209, 216, 5, 5, 112, 239, 5, 5, 148, 199, 5, 5, 247, 212, 5, 5, 165, 196, 5, 5, 21, 230, 5, 5, 248, 212, 5, 5, 58, 243, 5, 5, 91, 224, 5, 5, 3, 204, 5, 5, 246, 169, 5, 5, 66, 173, 5, 5, 243, 172, 5, 5, 63, 215, 5, 5, 147, 240, 5, 5, 114, 200, 5, 5, 106, 204, 5, 5, 238, 174, 5, 5, 98, 227, 5, 5, 76, 165, 5, 5, 106, 228, 5, 5, 37, 237, 5, 5, 66, 211, 5, 5, 177, 188, 5, 5, 185, 200, 5, 5, 8, 216, 5, 5, 38, 184, 5, 5, 253, 178, 5, 5, 148, 240, 5, 5, 70, 234, 5, 5, 73, 238, 5, 5, 103, 184, 5, 5, 124, 184, 5, 5, 21, 169, 5, 5, 53, 181, 5, 5, 40, 224, 5, 5, 249, 205, 5, 5, 193, 233, 5, 5, 237, 247, 5, 5, 143, 168, 5, 5, 86, 231, 5, 5, 198, 230, 5, 5, 21, 183, 5, 5, 149, 242, 5, 5, 126, 179, 5, 5, 127, 244, 5, 5, 81, 234, 5, 5, 40, 184, 5, 5, 230, 248, 5, 5, 181, 244, 5, 5, 184, 171, 5, 5, 115, 165, 5, 5, 22, 169, 5, 5, 200, 228, 5, 5, 175, 216, 5, 5, 221, 183, 5, 5, 2, 178, 5, 5, 97, 188, 5, 5, 20, 249, 5, 5, 77, 186, 5, 5, 75, 243, 5, 5, 239, 229, 5, 5, 154, 189, 5, 5, 76, 204, 5, 5, 104, 184, 5, 5, 79, 186, 5, 5, 105, 210, 5, 5, 136, 244, 5, 5, 241, 229, 5, 5, 144, 200, 5, 5, 91, 212, 5, 5, 83, 171, 5, 5, 235, 186, 5, 5, 169, 190, 5, 5, 206, 242, 5, 5, 182, 221, 5, 5, 82, 186, 5, 5, 204, 231, 5, 5, 127, 203, 5, 5, 116, 207, 5, 5, 117, 207, 5, 5, 62, 197, 5, 5, 178, 249, 5, 5, 19, 169, 5, 5, 188, 231, 5, 5, 29, 235, 5, 5, 66, 237, 5, 5, 65, 174, 5, 5, 199, 163, 5, 5, 106, 231, 5, 5, 211, 214, 5, 5, 22, 216, 5, 5, 20, 171, 5, 5, 164, 235, 5, 5, 103, 179, 5, 5, 250, 163, 5, 5, 158, 202, 5, 5, 3, 193, 5, 5, 142, 184, 5, 5, 58, 217, 5, 5, 119, 242, 5, 5, 132, 222, 5, 5, 184, 197, 5, 5, 158, 190, 5, 5, 68, 237, 5, 5, 45, 183, 5, 5, 128, 171, 5, 5, 78, 235, 5, 5, 77, 210, 5, 5, 101, 229, 5, 5, 95, 234, 5, 5, 232, 240, 5, 5, 123, 231, 5, 5, 189, 183, 5, 5, 45, 174, 5, 5, 130, 231, 5, 5, 198, 186, 5, 5, 67, 217, 5, 5, 143, 163, 5, 5, 199, 230, 5, 5, 10, 183, 5, 5, 8, 184, 5, 5, 10, 184, 5, 5, 137, 244, 5, 5, 170, 190, 5, 5, 173, 171, 5, 5, 108, 184, 5, 5, 96, 231, 5, 5, 31, 179, 5, 5, 250, 193, 5, 5, 194, 219, 5, 5, 130, 212, 5, 5, 4, 248, 5, 5, 15, 228, 5, 5, 198, 238, 5, 5, 180, 181, 5, 5, 122, 237, 5, 5, 188, 246, 5, 5, 57, 230, 5, 5, 210, 241, 5, 5, 132, 201, 5, 5, 132, 239, 5, 5, 119, 206, 5, 5, 155, 222, 5, 5, 89, 171, 5, 5, 191, 221, 5, 5, 95, 176, 5, 5, 137, 201, 5, 5, 87, 233, 5, 5, 66, 194, 5, 5, 46, 220, 5, 5, 214, 177, 5, 5, 201, 246, 5, 5, 137, 239, 5, 5, 15, 164, 5, 5, 233, 226, 5, 5, 234, 226, 5, 5, 175, 226, 5, 5, 183, 228, 5, 5, 54, 195, 5, 5, 95, 217, 5, 5, 47, 227, 5, 5, 103, 242, 5, 5, 212, 200, 5, 5, 154, 239, 5, 5, 43, 181, 5, 5, 17, 227, 5, 5, 188, 215, 5, 5, 197, 194, 5, 5, 100, 225, 5, 5, 120, 212, 5, 5, 198, 194, 5, 5, 129, 214, 5, 5, 39, 232, 5, 5, 112, 219, 5, 5, 20, 164, 5, 5, 197, 197, 5, 5, 31, 243, 5, 5, 130, 214, 5, 5, 16, 203, 5, 5, 187, 228, 5, 5, 186, 224, 5, 5, 53, 231, 5, 5, 87, 201, 5, 5, 93, 171, 5, 5, 149, 195, 5, 5, 183, 196, 5, 5, 9, 166, 5, 5, 149, 236, 5, 5, 19, 234, 5, 5, 132, 232, 5, 5, 178, 202, 5, 5, 254, 248, 5, 5, 179, 202, 5, 5, 176, 243, 5, 5, 96, 171, 5, 5, 150, 195, 5, 5, 97, 211, 5, 5, 16, 212, 5, 5, 177, 220, 5, 5, 98, 205, 5, 5, 215, 174, 5, 5, 115, 184, 5, 5, 93, 166, 5, 5, 174, 200, 5, 5, 6, 170, 5, 5, 5, 211, 5, 5, 91, 198, 5, 5, 86, 248, 5, 5, 247, 226, 5, 5, 116, 184, 5, 5, 226, 205, 5, 5, 118, 184, 5, 5, 247, 218, 5, 5, 26, 248, 5, 5, 25, 164, 5, 5, 184, 196, 5, 5, 119, 184, 5, 5, 7, 211, 5, 5, 126, 169, 5, 5, 152, 223, 5, 5, 79, 207, 5, 5, 125, 193, 5, 5, 8, 211, 5, 5, 236, 198, 5, 5, 118, 245, 5, 5, 227, 187, 5, 5, 253, 226, 5, 5, 120, 184, 5, 5, 201, 219, 5, 5, 193, 228, 5, 5, 70, 246, 5, 5, 47, 214, 5, 5, 185, 196, 5, 5, 104, 200, 5, 5, 234, 187, 5, 5, 70, 208, 5, 5, 186, 200, 5, 5, 164, 240, 5, 5, 195, 239, 5, 5, 35, 164, 5, 5, 215, 210, 5, 5, 178, 226, 5, 5, 36, 164, 5, 5, 225, 173, 5, 5, 34, 235, 5, 5, 130, 178, 5, 5, 158, 235, 5, 5, 209, 217, 5, 5, 136, 175, 5, 5, 85, 191, 5, 5, 135, 178, 5, 5, 19, 228, 5, 5, 187, 166, 5, 5, 195, 173, 5, 5, 115, 222, 5, 5, 30, 174, 5, 5, 190, 227, 5, 5, 58, 195, 5, 5, 9, 206, 5, 5, 14, 206, 5, 5, 117, 222, 5, 5, 125, 246, 5, 5, 67, 233, 5, 5, 230, 217, 5, 5, 145, 229, 5, 5, 143, 178, 5, 5, 148, 175, 5, 5, 201, 214, 5, 5, 128, 203, 5, 5, 251, 227, 5, 5, 50, 239, 5, 5, 51, 239, 5, 5, 52, 239, 5, 5, 126, 191, 5, 5, 190, 204, 5, 5, 60, 229, 5, 5, 178, 224, 5, 5, 253, 228, 5, 5, 254, 228, 5, 5, 155, 169, 5, 5, 123, 191, 5, 5, 7, 225, 5, 5, 22, 188, 5, 5, 151, 215, 5, 5, 152, 215, 5, 5, 245, 196, 5, 5, 26, 242, 5, 5, 73, 247, 5, 5, 27, 242, 5, 5, 18, 188, 5, 5, 19, 188, 5, 5, 156, 241, 5, 5, 186, 170, 5, 5, 74, 247, 5, 5, 75, 247, 5, 5, 76, 247, 5, 5, 128, 191, 5, 5, 157, 241, 5, 5, 100, 178, 5, 5, 152, 196, 5, 5, 229, 222, 5, 5, 170, 204, 5, 5, 124, 192, 5, 5, 52, 193, 5, 5, 125, 192, 5, 5, 79, 182, 5, 5, 129, 192, 5, 5, 130, 192, 5, 5, 183, 226, 5, 5, 205, 243, 5, 5, 57, 223, 5, 5, 115, 195, 5, 5, 146, 183, 5, 5, 183, 205, 5, 5, 50, 246, 5, 5, 42, 184, 5, 5, 234, 206, 5, 5, 36, 168, 5, 5, 135, 234, 5, 5, 107, 213, 5, 5, 207, 164, 5, 5, 194, 206, 5, 5, 210, 203, 5, 5, 72, 181, 5, 5, 158, 211, 5, 5, 212, 180, 5, 5, 211, 203, 5, 5, 250, 244, 5, 5, 94, 177, 5, 5, 190, 226, 5, 5, 18, 227, 5, 5, 173, 168, 5, 5, 216, 201, 5, 5, 60, 214, 5, 5, 215, 220, 5, 5, 61, 214, 5, 5, 153, 183, 5, 5, 109, 220, 5, 5, 189, 212, 5, 5, 60, 246, 5, 5, 126, 170, 5, 5, 149, 237, 5, 5, 9, 186, 5, 5, 184, 168, 5, 5, 220, 180, 5, 5, 207, 172, 5, 5, 108, 229, 5, 5, 73, 214, 5, 5, 205, 181, 5, 5, 71, 191, 5, 5, 74, 191, 5, 5, 86, 213, 5, 5, 137, 234, 5, 5, 76, 191, 5, 5, 86, 207, 5, 5, 85, 214, 5, 5, 97, 207, 5, 5, 112, 207, 5, 5, 248, 187, 5, 5, 188, 207, 5, 5, 123, 229, 5, 5, 162, 213, 5, 5, 185, 223, 5, 5, 137, 182, 5, 5, 199, 183, 5, 5, 37, 228, 5, 5, 165, 245, 5, 5, 36, 215, 5, 5, 25, 187, 5, 5, 188, 237, 5, 5, 177, 175, 5, 5, 149, 200, 5, 5, 226, 190, 5, 5, 9, 240, 5, 5, 251, 208, 5, 5, 112, 230, 5, 5, 84, 245, 5, 5, 115, 198, 5, 5, 134, 228, 5, 5, 5, 248, 5, 5, 148, 225, 5, 5, 21, 237, 5, 5, 252, 208, 5, 5, 113, 230, 5, 5, 201, 248, 5, 5, 84, 164, 5, 5, 211, 241, 5, 5, 14, 239, 5, 5, 243, 240, 5, 5, 88, 245, 5, 5, 134, 203, 5, 5, 192, 237, 5, 5, 45, 202, 5, 5, 59, 210, 5, 5, 94, 173, 5, 5, 189, 246, 5, 5, 157, 208, 5, 5, 106, 234, 5, 5, 115, 210, 5, 5, 158, 208, 5, 5, 164, 223, 5, 5, 206, 203, 5, 5, 154, 211, 5, 5, 203, 192, 5, 5, 180, 215, 5, 5, 139, 238, 5, 5, 17, 186, 5, 5, 137, 175, 5, 5, 0, 136, 5, 5, 4, 233, 5, 5, 5, 233, 5, 5, 60, 244, 5, 5, 44, 207, 5, 5, 184, 192, 5, 5, 40, 195, 5, 5, 67, 192, 5, 5, 59, 217, 5, 5, 170, 166, 5, 5, 37, 187, 5, 5, 21, 181, 5, 5, 159, 211, 5, 5, 159, 242, 5, 5, 68, 174, 5, 5, 13, 215, 5, 5, 68, 203, 5, 5, 168, 189, 5, 5, 22, 182, 5, 5, 237, 167, 5, 5, 143, 238, 5, 5, 81, 175, 5, 5, 25, 197, 5, 5, 204, 248, 5, 5, 160, 211, 5, 5, 64, 199, 5, 5, 191, 166, 5, 5, 224, 171, 5, 5, 238, 238, 5, 5, 22, 181, 5, 5, 91, 189, 5, 5, 34, 223, 5, 5, 209, 168, 5, 5, 226, 178, 5, 5, 119, 198, 5, 5, 103, 196, 5, 5, 14, 215, 5, 5, 49, 170, 5, 5, 165, 182, 5, 5, 60, 217, 5, 5, 26, 240, 5, 5, 3, 166, 5, 5, 92, 197, 5, 5, 0, 138, 5, 5, 133, 190, 5, 5, 204, 237, 5, 5, 199, 174, 5, 5, 217, 217, 5, 5, 23, 177, 5, 5, 8, 190, 5, 5, 7, 239, 5, 5, 30, 239, 5, 5, 216, 191, 5, 5, 42, 212, 5, 5, 8, 179, 5, 5, 144, 239, 5, 5, 67, 172, 5, 5, 237, 226, 5, 5, 200, 238, 5, 5, 9, 246, 5, 5, 168, 200, 5, 5, 194, 218, 5, 5, 92, 171, 5, 5, 252, 186, 5, 5, 163, 202, 5, 5, 168, 182, 5, 5, 10, 246, 5, 5, 104, 242, 5, 5, 99, 193, 5, 5, 187, 213, 5, 5, 253, 186, 5, 5, 161, 241, 5, 5, 84, 204, 5, 5, 70, 221, 5, 5, 159, 190, 5, 5, 241, 203, 5, 5, 157, 237, 5, 5, 158, 237, 5, 5, 64, 248, 5, 5, 7, 214, 5, 5, 101, 166, 5, 5, 192, 209, 5, 5, 118, 212, 5, 5, 201, 173, 5, 5, 202, 173, 5, 5, 49, 183, 5, 5, 255, 186, 5, 5, 39, 242, 5, 5, 195, 215, 5, 5, 147, 201, 5, 5, 138, 175, 5, 5, 151, 228, 5, 5, 95, 225, 5, 5, 199, 181, 5, 5, 232, 192, 5, 5, 45, 208, 5, 5, 248, 197, 5, 5, 191, 213, 5, 5, 169, 243, 5, 5, 14, 248, 5, 5, 140, 186, 5, 5, 188, 195, 5, 5, 173, 185, 5, 5, 38, 185, 5, 5, 141, 215, 5, 5, 15, 212, 5, 5, 55, 197, 5, 5, 167, 226, 5, 5, 168, 226, 5, 5, 59, 179, 5, 5, 224, 188, 5, 5, 140, 199, 5, 5, 208, 206, 5, 5, 42, 185, 5, 5, 184, 208, 5, 5, 234, 171, 5, 5, 200, 196, 5, 5, 201, 196, 5, 5, 220, 232, 5, 5, 84, 193, 5, 5, 210, 234, 5, 5, 250, 240, 5, 5, 251, 240, 5, 5, 205, 192, 5, 5, 201, 190, 5, 5, 202, 190, 5, 5, 150, 186, 5, 5, 141, 207, 5, 5, 55, 178, 5, 5, 78, 221, 5, 5, 143, 193, 5, 5, 177, 243, 5, 5, 235, 178, 5, 5, 85, 198, 5, 5, 79, 221, 5, 5, 65, 246, 5, 5, 192, 214, 5, 5, 249, 167, 5, 5, 129, 242, 5, 5, 95, 237, 5, 5, 153, 203, 5, 5, 178, 164, 5, 5, 221, 232, 5, 5, 12, 196, 5, 5, 236, 178, 5, 5, 111, 220, 5, 5, 85, 193, 5, 5, 178, 243, 5, 5, 205, 231, 5, 5, 26, 232, 5, 5, 44, 201, 5, 5, 192, 247, 5, 5, 217, 245, 5, 5, 36, 244, 5, 5, 207, 173, 5, 5, 23, 210, 5, 5, 37, 244, 5, 5, 166, 180, 5, 5, 230, 239, 5, 5, 239, 218, 5, 5, 71, 227, 5, 5, 40, 232, 5, 5, 224, 241, 5, 5, 225, 237, 5, 5, 234, 221, 5, 5, 166, 195, 5, 5, 92, 191, 5, 5, 227, 242, 5, 5, 241, 169, 5, 5, 92, 198, 5, 5, 224, 201, 5, 5, 225, 201, 5, 5, 223, 200, 5, 5, 150, 181, 5, 5, 48, 229, 5, 5, 207, 169, 5, 5, 174, 210, 5, 5, 107, 245, 5, 5, 193, 247, 5, 5, 92, 233, 5, 5, 22, 186, 5, 5, 44, 195, 5, 5, 177, 242, 5, 5, 178, 242, 5, 5, 71, 246, 5, 5, 93, 233, 5, 5, 147, 223, 5, 5, 26, 237, 5, 5, 60, 231, 5, 5, 75, 195, 5, 5, 206, 210, 5, 5, 194, 247, 5, 5, 245, 180, 5, 5, 253, 185, 5, 5, 216, 181, 5, 5, 123, 178, 5, 5, 228, 212, 5, 5, 126, 247, 5, 5, 119, 234, 5, 5, 103, 176, 5, 5, 66, 231, 5, 5, 164, 203, 5, 5, 93, 195, 5, 5, 44, 188, 5, 5, 113, 245, 5, 5, 100, 237, 5, 5, 132, 209, 5, 5, 124, 231, 5, 5, 28, 246, 5, 5, 182, 164, 5, 5, 129, 177, 5, 5, 184, 234, 5, 5, 155, 210, 5, 5, 14, 250, 5, 5, 244, 183, 5, 5, 185, 234, 5, 5, 137, 202, 5, 5, 112, 227, 5, 5, 248, 230, 5, 5, 117, 237, 5, 5, 101, 233, 5, 5, 62, 206, 5, 5, 228, 187, 5, 5, 216, 195, 5, 5, 136, 224, 5, 5, 51, 177, 5, 5, 105, 234, 5, 5, 205, 177, 5, 5, 192, 181, 5, 5, 2, 186, 5, 5, 51, 221, 5, 5, 119, 220, 5, 5, 120, 220, 5, 5, 106, 238, 5, 5, 37, 234, 5, 5, 177, 211, 5, 5, 40, 211, 5, 5, 216, 234, 5, 5, 31, 233, 5, 5, 144, 175, 5, 5, 145, 175, 5, 5, 54, 213, 5, 5, 114, 248, 5, 5, 5, 167, 5, 5, 243, 190, 5, 5, 109, 216, 5, 5, 138, 234, 5, 5, 57, 163, 5, 5, 210, 210, 5, 5, 4, 204, 5, 5, 244, 242, 5, 5, 181, 185, 5, 5, 145, 220, 5, 5, 233, 188, 5, 5, 235, 187, 5, 5, 236, 187, 5, 5, 178, 205, 5, 5, 138, 176, 5, 5, 117, 178, 5, 5, 70, 233, 5, 5, 103, 204, 5, 5, 48, 194, 5, 5, 191, 241, 5, 5, 5, 204, 5, 5, 192, 241, 5, 5, 169, 207, 5, 5, 90, 220, 5, 5, 149, 240, 5, 5, 68, 217, 5, 5, 57, 183, 5, 5, 109, 241, 5, 5, 110, 241, 5, 5, 189, 239, 5, 5, 252, 183, 5, 5, 159, 215, 5, 5, 100, 202, 5, 5, 90, 214, 5, 5, 47, 195, 5, 5, 2, 213, 5, 5, 12, 177, 5, 5, 154, 238, 5, 5, 33, 233, 5, 5, 171, 246, 5, 5, 196, 239, 5, 5, 87, 231, 5, 5, 150, 175, 5, 5, 173, 189, 5, 5, 226, 231, 5, 5, 40, 248, 5, 5, 231, 248, 5, 5, 223, 243, 5, 5, 230, 230, 5, 5, 191, 203, 5, 5, 117, 187, 5, 5, 243, 193, 5, 5, 252, 190, 5, 5, 253, 190, 5, 5, 136, 177, 5, 5, 91, 231, 5, 5, 32, 213, 5, 5, 12, 207, 5, 5, 109, 224, 5, 5, 220, 244, 5, 5, 112, 196, 5, 5, 201, 242, 5, 5, 254, 245, 5, 5, 63, 183, 5, 5, 92, 212, 5, 5, 4, 224, 5, 5, 13, 235, 5, 5, 171, 190, 5, 5, 140, 177, 5, 5, 141, 177, 5, 5, 28, 184, 5, 5, 207, 242, 5, 5, 204, 242, 5, 5, 13, 243, 5, 5, 144, 196, 5, 5, 14, 243, 5, 5, 45, 226, 5, 5, 77, 196, 5, 5, 46, 226, 5, 5, 253, 219, 5, 5, 29, 201, 5, 5, 77, 206, 5, 5, 214, 217, 5, 5, 238, 223, 5, 5, 239, 223, 5, 5, 204, 187, 5, 5, 186, 194, 5, 5, 251, 232, 5, 5, 186, 209, 5, 5, 38, 207, 5, 5, 217, 194, 5, 5, 40, 207, 5, 5, 247, 244, 5, 5, 13, 169, 5, 5, 14, 169, 5, 5, 169, 223, 5, 5, 24, 210, 5, 5, 30, 210, 5, 5, 124, 202, 5, 5, 210, 239, 5, 5, 155, 235, 5, 5, 0, 130, 5, 5, 59, 218, 5, 5, 208, 219, 5, 5, 96, 216, 5, 5, 145, 239, 5, 5, 186, 205, 5, 5, 4, 218, 5, 5, 242, 207, 5, 5, 0, 142, 5, 5, 14, 218, 5, 5, 159, 205, 5, 5, 57, 174, 5, 5, 207, 231, 5, 5, 244, 164, 5, 5, 28, 227, 5, 5, 119, 166, 5, 5, 190, 177, 5, 5, 200, 229, 5, 5, 238, 172, 5, 5, 227, 183, 5, 5, 185, 226, 5, 5, 71, 173, 5, 5, 164, 166, 5, 5, 147, 218, 5, 5, 172, 212, 5, 5, 59, 199, 5, 5, 22, 176, 5, 5, 5, 246, 5, 5, 43, 221, 5, 5, 111, 180, 5, 5, 82, 175, 5, 5, 98, 169, 5, 5, 248, 166, 5, 5, 165, 166, 5, 5, 246, 180, 5, 5, 37, 163, 5, 5, 116, 188, 5, 5, 189, 218, 5, 5, 35, 220, 5, 5, 70, 163, 5, 5, 190, 218, 5, 5, 229, 176, 5, 5, 104, 169, 5, 5, 42, 240, 5, 5, 43, 240, 5, 5, 174, 168, 5, 5, 81, 209, 5, 5, 232, 207, 5, 5, 184, 230, 5, 5, 201, 229, 5, 5, 86, 175, 5, 5, 114, 221, 5, 5, 209, 229, 5, 5, 193, 209, 5, 5, 166, 220, 5, 5, 195, 226, 5, 5, 121, 212, 5, 5, 47, 213, 5, 5, 142, 164, 5, 5, 94, 175, 5, 5, 215, 193, 5, 5, 19, 224, 5, 5, 185, 168, 5, 5, 47, 221, 5, 5, 102, 236, 5, 5, 175, 246, 5, 5, 97, 175, 5, 5, 16, 246, 5, 5, 9, 225, 5, 5, 154, 227, 5, 5, 185, 208, 5, 5, 246, 231, 5, 5, 164, 193, 5, 5, 68, 170, 5, 5, 24, 178, 5, 5, 29, 198, 5, 5, 118, 221, 5, 5, 195, 216, 5, 5, 101, 171, 5, 5, 23, 246, 5, 5, 197, 245, 5, 5, 114, 174, 5, 5, 87, 240, 5, 5, 147, 164, 5, 5, 189, 168, 5, 5, 70, 177, 5, 5, 205, 194, 5, 5, 62, 167, 5, 5, 92, 201, 5, 5, 155, 179, 5, 5, 104, 171, 5, 5, 91, 203, 5, 5, 134, 180, 5, 5, 24, 200, 5, 5, 213, 222, 5, 5, 63, 228, 5, 5, 139, 170, 5, 5, 100, 172, 5, 5, 150, 165, 5, 5, 166, 193, 5, 5, 254, 231, 5, 5, 101, 172, 5, 5, 148, 164, 5, 5, 52, 221, 5, 5, 130, 206, 5, 5, 140, 224, 5, 5, 135, 174, 5, 5, 14, 219, 5, 5, 240, 198, 5, 5, 244, 172, 5, 5, 133, 226, 5, 5, 53, 165, 5, 5, 114, 165, 5, 5, 59, 221, 5, 5, 117, 171, 5, 5, 144, 209, 5, 5, 75, 224, 5, 5, 109, 173, 5, 5, 238, 170, 5, 5, 150, 164, 5, 5, 41, 224, 5, 5, 138, 171, 5, 5, 177, 225, 5, 5, 188, 203, 5, 5, 167, 215, 5, 5, 43, 186, 5, 5, 151, 192, 5, 5, 167, 168, 5, 5, 63, 184, 5, 5, 191, 204, 5, 5, 81, 166, 5, 5, 85, 166, 5, 5, 165, 216, 5, 5, 161, 192, 5, 5, 139, 241, 5, 5, 208, 200, 5, 5, 95, 242, 5, 5, 13, 242, 5, 5, 133, 202, 5, 5, 172, 194, 5, 5, 21, 168, 5, 5, 125, 181, 5, 5, 21, 232, 5, 5, 173, 194, 5, 5, 170, 222, 5, 5, 67, 169, 5, 5, 135, 211, 5, 5, 251, 164, 5, 5, 217, 210, 5, 5, 205, 178, 5, 5, 130, 226, 5, 5, 254, 193, 5, 5, 9, 241, 5, 5, 167, 163, 5, 5, 208, 203, 5, 5, 243, 216, 5, 5, 184, 189, 5, 5, 48, 245, 5, 5, 192, 222, 5, 5, 23, 196, 5, 5, 92, 244, 5, 5, 203, 176, 5, 5, 15, 167, 5, 5, 162, 177, 5, 5, 224, 240, 5, 5, 166, 171, 5, 5, 213, 192, 5, 5, 214, 233, 5, 5, 154, 174, 5, 5, 166, 182, 5, 5, 115, 167, 5, 5, 194, 172, 5, 5, 69, 216, 5, 5, 218, 239, 5, 5, 195, 218, 5, 5, 13, 181, 5, 5, 196, 218, 5, 5, 239, 214, 5, 5, 36, 178, 5, 5, 11, 233, 5, 5, 131, 214, 5, 5, 200, 181, 5, 5, 12, 194, 5, 5, 26, 196, 5, 5, 195, 170, 5, 5, 146, 181, 5, 5, 75, 242, 5, 5, 102, 166, 5, 5, 86, 166, 5, 5, 141, 225, 5, 5, 43, 185, 5, 5, 21, 165, 5, 5, 69, 202, 5, 5, 203, 238, 5, 5, 141, 229, 5, 5, 213, 195, 5, 5, 214, 249, 5, 5, 55, 182, 5, 5, 96, 237, 5, 5, 142, 229, 5, 5, 128, 211, 5, 5, 87, 248, 5, 5, 19, 248, 5, 5, 169, 171, 5, 5, 41, 236, 5, 5, 243, 233, 5, 5, 45, 216, 5, 5, 209, 172, 5, 5, 171, 204, 5, 5, 234, 237, 5, 5, 228, 193, 5, 5, 103, 188, 5, 5, 158, 227, 5, 5, 126, 188, 5, 5, 114, 186, 5, 5, 44, 236, 5, 5, 248, 218, 5, 5, 50, 167, 5, 5, 130, 211, 5, 5, 174, 204, 5, 5, 175, 204, 5, 5, 195, 205, 5, 5, 107, 238, 5, 5, 62, 196, 5, 5, 236, 185, 5, 5, 241, 243, 5, 5, 23, 188, 5, 5, 105, 175, 5, 5, 37, 190, 5, 5, 121, 240, 5, 5, 172, 225, 5, 5, 92, 224, 5, 5, 224, 181, 5, 5, 133, 235, 5, 5, 6, 216, 5, 5, 170, 171, 5, 5, 34, 225, 5, 5, 70, 245, 5, 5, 215, 172, 5, 5, 175, 194, 5, 5, 36, 196, 5, 5, 93, 196, 5, 5, 77, 165, 5, 5, 67, 211, 5, 5, 91, 214, 5, 5, 12, 189, 5, 5, 13, 189, 5, 5, 249, 187, 5, 5, 203, 189, 5, 5, 252, 202, 5, 5, 221, 244, 5, 5, 74, 215, 5, 5, 139, 181, 5, 5, 235, 249, 5, 5, 236, 186, 5, 5, 142, 181, 5, 5, 204, 243, 5, 5, 210, 217, 5, 5, 160, 242, 5, 5, 144, 181, 5, 5, 164, 176, 5, 5, 219, 166, 5, 5, 81, 205, 5, 5, 42, 200, 5, 5, 17, 183, 5, 5, 45, 200, 5, 5, 216, 194, 5, 5, 89, 184, 5, 5, 76, 195, 5, 5, 228, 165, 5, 5, 10, 240, 5, 5, 26, 209, 5, 5, 208, 184, 5, 5, 128, 173, 5, 5, 18, 207, 5, 5, 220, 242, 5, 5, 44, 240, 5, 5, 187, 232, 5, 5, 54, 165, 5, 5, 254, 239, 5, 5, 84, 239, 5, 5, 68, 239, 5, 5, 80, 236, 5, 5, 32, 227, 5, 5, 69, 239, 5, 5, 72, 227, 5, 5, 80, 227, 5, 5, 187, 194, 5, 5, 95, 177, 5, 5, 206, 243, 5, 5, 52, 199, 5, 5, 58, 174, 5, 5, 128, 201, 5, 5, 242, 244, 5, 5, 76, 182, 5, 5, 30, 223, 5, 5, 53, 206, 5, 5, 16, 163, 5, 5, 181, 228, 5, 5, 223, 176, 5, 5, 9, 221, 5, 5, 124, 207, 5, 5, 149, 179, 5, 5, 155, 174, 5, 5, 204, 176, 5, 5, 35, 171, 5, 5, 85, 215, 5, 5, 248, 198, 5, 5, 225, 223, 5, 5, 148, 184, 5, 5, 143, 206, 5, 5, 212, 220, 5, 5, 146, 239, 5, 5, 73, 171, 5, 5, 43, 212, 5, 5, 172, 246, 5, 5, 227, 227, 5, 5, 235, 230, 5, 5, 96, 217, 5, 5, 152, 206, 5, 5, 78, 177, 5, 5, 10, 221, 5, 5, 57, 206, 5, 5, 57, 211, 5, 5, 155, 167, 5, 5, 31, 176, 5, 5, 58, 211, 5, 5, 29, 165, 5, 5, 30, 165, 5, 5, 196, 227, 5, 5, 36, 176, 5, 5, 36, 205, 5, 5, 91, 192, 5, 5, 210, 191, 5, 5, 117, 172, 5, 5, 55, 165, 5, 5, 138, 195, 5, 5, 154, 205, 5, 5, 136, 165, 5, 5, 159, 214, 5, 5, 49, 176, 5, 5, 152, 193, 5, 5, 65, 206, 5, 5, 173, 201, 5, 5, 43, 171, 5, 5, 157, 193, 5, 5, 181, 246, 5, 5, 202, 207, 5, 5, 146, 202, 5, 5, 147, 202, 5, 5, 207, 193, 5, 5, 68, 176, 5, 5, 5, 190, 5, 5, 6, 190, 5, 5, 124, 182, 5, 5, 80, 200, 5, 5, 240, 218, 5, 5, 91, 200, 5, 5, 92, 200, 5, 5, 190, 239, 5, 5, 191, 239, 5, 5, 201, 239, 5, 5, 202, 239, 5, 5, 193, 248, 5, 5, 195, 248, 5, 5, 38, 228, 5, 5, 140, 216, 5, 5, 20, 247, 5, 5, 164, 182, 5, 5, 27, 243, 5, 5, 28, 243, 5, 5, 65, 248, 5, 5, 6, 223, 5, 5, 20, 234, 5, 5, 145, 165, 5, 5, 79, 177, 5, 5, 10, 166, 5, 5, 157, 167, 5, 5, 15, 250, 5, 5, 85, 168, 5, 5, 15, 221, 5, 5, 214, 244, 5, 5, 28, 223, 5, 5, 79, 223, 5, 5, 20, 222, 5, 5, 53, 182, 5, 5, 46, 202, 5, 5, 168, 173, 5, 5, 198, 222, 5, 5, 16, 240, 5, 5, 83, 247, 5, 5, 186, 223, 5, 5, 230, 164, 5, 5, 151, 176, 5, 5, 92, 189, 5, 5, 69, 174, 5, 5, 89, 247, 5, 5, 65, 221, 5, 5, 149, 234, 5, 5, 90, 247, 5, 5, 144, 206, 5, 5, 87, 175, 5, 5, 182, 212, 5, 5, 5, 218, 5, 5, 167, 198, 5, 5, 28, 244, 5, 5, 232, 199, 5, 5, 87, 191, 5, 5, 49, 198, 5, 5, 93, 247, 5, 5, 138, 225, 5, 5, 205, 218, 5, 5, 223, 216, 5, 5, 153, 206, 5, 5, 201, 181, 5, 5, 76, 235, 5, 5, 57, 219, 5, 5, 39, 185, 5, 5, 79, 235, 5, 5, 58, 175, 5, 5, 210, 166, 5, 5, 78, 213, 5, 5, 239, 178, 5, 5, 181, 202, 5, 5, 10, 225, 5, 5, 219, 218, 5, 5, 80, 235, 5, 5, 9, 220, 5, 5, 44, 185, 5, 5, 73, 248, 5, 5, 66, 214, 5, 5, 223, 203, 5, 5, 94, 205, 5, 5, 73, 227, 5, 5, 77, 203, 5, 5, 243, 165, 5, 5, 45, 201, 5, 5, 216, 174, 5, 5, 231, 217, 5, 5, 72, 246, 5, 5, 121, 167, 5, 5, 176, 241, 5, 5, 238, 245, 5, 5, 229, 212, 5, 5, 90, 221, 5, 5, 96, 168, 5, 5, 178, 180, 5, 5, 63, 175, 5, 5, 93, 221, 5, 5, 96, 221, 5, 5, 130, 165, 5, 5, 64, 175, 5, 5, 158, 226, 5, 5, 197, 223, 5, 5, 59, 204, 5, 5, 115, 200, 5, 5, 62, 204, 5, 5, 33, 200, 5, 5, 227, 169, 5, 5, 43, 191, 5, 5, 58, 216, 5, 5, 62, 164, 5, 5, 222, 211, 5, 5, 241, 215, 5, 5, 36, 216, 5, 5, 245, 164, 5, 5, 2, 240, 5, 5, 15, 185, 5, 5, 87, 230, 5, 5, 128, 175, 5, 5, 211, 177, 5, 5, 228, 191, 5, 5, 155, 202, 5, 5, 97, 230, 5, 5, 89, 230, 5, 5, 156, 224, 5, 5, 22, 180, 5, 5, 92, 179, 5, 5, 93, 179, 5, 5, 103, 192, 5, 5, 23, 209, 5, 5, 33, 227, 5, 5, 33, 235, 5, 5, 180, 218, 5, 5, 107, 247, 5, 5, 183, 170, 5, 5, 167, 212, 5, 5, 7, 173, 5, 5, 93, 244, 5, 5, 97, 181, 5, 5, 225, 171, 5, 5, 49, 247, 5, 5, 199, 222, 5, 5, 149, 226, 5, 5, 90, 245, 5, 5, 47, 202, 5, 5, 49, 245, 5, 5, 111, 213, 5, 5, 41, 203, 5, 5, 8, 242, 5, 5, 235, 246, 5, 5, 40, 235, 5, 5, 88, 237, 5, 5, 80, 196, 5, 5, 53, 248, 5, 5, 230, 165, 5, 5, 181, 218, 5, 5, 159, 208, 5, 5, 230, 234, 5, 5, 180, 224, 5, 5, 134, 179, 5, 5, 139, 182, 5, 5, 116, 210, 5, 5, 90, 184, 5, 5, 4, 201, 5, 5, 34, 218, 5, 5, 48, 184, 5, 5, 9, 197, 5, 5, 41, 215, 5, 5, 243, 215, 5, 5, 209, 195, 5, 5, 220, 198, 5, 5, 106, 195, 5, 5, 130, 173, 5, 5, 231, 199, 5, 5, 136, 179, 5, 5, 109, 237, 5, 5, 50, 245, 5, 5, 194, 238, 5, 5, 37, 225, 5, 5, 140, 226, 5, 5, 45, 207, 5, 5, 27, 240, 5, 5, 16, 164, 5, 5, 226, 223, 5, 5, 226, 171, 5, 5, 228, 166, 5, 5, 6, 233, 5, 5, 167, 232, 5, 5, 80, 244, 5, 5, 46, 207, 5, 5, 103, 164, 5, 5, 151, 184, 5, 5, 47, 183, 5, 5, 66, 221, 5, 5, 166, 166, 5, 5, 125, 166, 5, 5, 247, 180, 5, 5, 69, 199, 5, 5, 65, 199, 5, 5, 47, 207, 5, 5, 20, 175, 5, 5, 117, 188, 5, 5, 111, 235, 5, 5, 17, 224, 5, 5, 227, 176, 5, 5, 88, 175, 5, 5, 138, 239, 5, 5, 183, 181, 5, 5, 36, 242, 5, 5, 25, 174, 5, 5, 57, 202, 5, 5, 135, 167, 5, 5, 166, 216, 5, 5, 93, 249, 5, 5, 149, 183, 5, 5, 204, 193, 5, 5, 70, 174, 5, 5, 112, 213, 5, 5, 255, 164, 5, 5, 11, 242, 5, 5, 96, 209, 5, 5, 157, 217, 5, 5, 43, 235, 5, 5, 58, 166, 5, 5, 84, 201, 5, 5, 247, 216, 5, 5, 204, 225, 5, 5, 2, 165, 5, 5, 134, 229, 5, 5, 183, 212, 5, 5, 230, 179, 5, 5, 27, 250, 5, 5, 125, 233, 5, 5, 136, 178, 5, 5, 154, 169, 5, 5, 153, 218, 5, 5, 28, 240, 5, 5, 48, 227, 5, 5, 123, 183, 5, 5, 218, 217, 5, 5, 49, 227, 5, 5, 117, 213, 5, 5, 197, 185, 5, 5, 199, 198, 5, 5, 254, 196, 5, 5, 87, 170, 5, 5, 158, 218, 5, 5, 3, 214, 5, 5, 187, 188, 5, 5, 14, 181, 5, 5, 85, 201, 5, 5, 24, 177, 5, 5, 249, 197, 5, 5, 64, 197, 5, 5, 7, 194, 5, 5, 200, 198, 5, 5, 184, 212, 5, 5, 30, 204, 5, 5, 138, 229, 5, 5, 195, 235, 5, 5, 224, 198, 5, 5, 220, 209, 5, 5, 81, 244, 5, 5, 75, 200, 5, 5, 43, 200, 5, 5, 88, 223, 5, 5, 45, 203, 5, 5, 147, 239, 5, 5, 37, 218, 5, 5, 206, 193, 5, 5, 22, 189, 5, 5, 2, 200, 5, 5, 194, 231, 5, 5, 251, 216, 5, 5, 65, 223, 5, 5, 255, 228, 5, 5, 217, 184, 5, 5, 170, 198, 5, 5, 3, 239, 5, 5, 55, 195, 5, 5, 9, 179, 5, 5, 74, 181, 5, 5, 199, 194, 5, 5, 44, 212, 5, 5, 190, 212, 5, 5, 82, 209, 5, 5, 207, 248, 5, 5, 137, 167, 5, 5, 198, 195, 5, 5, 57, 175, 5, 5, 76, 200, 5, 5, 207, 237, 5, 5, 234, 214, 5, 5, 216, 182, 5, 5, 122, 242, 5, 5, 86, 195, 5, 5, 234, 233, 5, 5, 65, 183, 5, 5, 181, 249, 5, 5, 197, 170, 5, 5, 5, 250, 5, 5, 2, 250, 5, 5, 251, 248, 5, 5, 208, 167, 5, 5, 225, 196, 5, 5, 129, 209, 5, 5, 209, 248, 5, 5, 74, 248, 5, 5, 202, 181, 5, 5, 220, 203, 5, 5, 198, 212, 5, 5, 138, 219, 5, 5, 23, 189, 5, 5, 218, 180, 5, 5, 122, 198, 5, 5, 250, 197, 5, 5, 101, 209, 5, 5, 221, 180, 5, 5, 135, 166, 5, 5, 57, 240, 5, 5, 170, 210, 5, 5, 56, 200, 5, 5, 191, 212, 5, 5, 212, 163, 5, 5, 106, 170, 5, 5, 99, 239, 5, 5, 194, 198, 5, 5, 110, 237, 5, 5, 45, 212, 5, 5, 139, 225, 5, 5, 104, 196, 5, 5, 98, 245, 5, 5, 99, 245, 5, 5, 14, 242, 5, 5, 172, 186, 5, 5, 70, 189, 5, 5, 218, 200, 5, 5, 44, 235, 5, 5, 71, 240, 5, 5, 210, 248, 5, 5, 56, 187, 5, 5, 98, 184, 5, 5, 66, 183, 5, 5, 215, 237, 5, 5, 17, 212, 5, 5, 136, 217, 5, 5, 73, 210, 5, 5, 100, 245, 5, 5, 4, 172, 5, 5, 81, 235, 5, 5, 184, 172, 5, 5, 186, 163, 5, 5, 67, 178, 5, 5, 240, 234, 5, 5, 43, 225, 5, 5, 173, 170, 5, 5, 103, 236, 5, 5, 183, 202, 5, 5, 104, 194, 5, 5, 99, 194, 5, 5, 170, 216, 5, 5, 157, 175, 5, 5, 123, 208, 5, 5, 56, 197, 5, 5, 103, 181, 5, 5, 203, 190, 5, 5, 214, 212, 5, 5, 175, 195, 5, 5, 131, 194, 5, 5, 210, 183, 5, 5, 112, 211, 5, 5, 21, 224, 5, 5, 183, 178, 5, 5, 177, 193, 5, 5, 83, 220, 5, 5, 87, 206, 5, 5, 208, 246, 5, 5, 6, 249, 5, 5, 174, 205, 5, 5, 189, 202, 5, 5, 21, 186, 5, 5, 75, 248, 5, 5, 199, 199, 5, 5, 117, 197, 5, 5, 66, 212, 5, 5, 247, 179, 5, 5, 222, 180, 5, 5, 225, 180, 5, 5, 237, 183, 5, 5, 94, 247, 5, 5, 254, 225, 5, 5, 230, 198, 5, 5, 249, 200, 5, 5, 199, 212, 5, 5, 29, 225, 5, 5, 254, 215, 5, 5, 17, 246, 5, 5, 84, 200, 5, 5, 226, 237, 5, 5, 226, 236, 5, 5, 128, 233, 5, 5, 206, 204, 5, 5, 8, 250, 5, 5, 169, 177, 5, 5, 65, 207, 5, 5, 84, 234, 5, 5, 112, 225, 5, 5, 60, 175, 5, 5, 222, 163, 5, 5, 105, 209, 5, 5, 75, 234, 5, 5, 32, 176, 5, 5, 100, 194, 5, 5, 109, 178, 5, 5, 70, 183, 5, 5, 185, 249, 5, 5, 82, 216, 5, 5, 94, 222, 5, 5, 38, 177, 5, 5, 46, 201, 5, 5, 116, 226, 5, 5, 139, 212, 5, 5, 14, 230, 5, 5, 122, 216, 5, 5, 244, 233, 5, 5, 218, 165, 5, 5, 7, 172, 5, 5, 94, 233, 5, 5, 167, 180, 5, 5, 188, 209, 5, 5, 28, 224, 5, 5, 229, 163, 5, 5, 103, 217, 5, 5, 230, 245, 5, 5, 171, 191, 5, 5, 86, 210, 5, 5, 73, 246, 5, 5, 100, 248, 5, 5, 196, 247, 5, 5, 31, 165, 5, 5, 104, 217, 5, 5, 224, 200, 5, 5, 10, 172, 5, 5, 32, 165, 5, 5, 105, 240, 5, 5, 204, 171, 5, 5, 9, 250, 5, 5, 112, 195, 5, 5, 101, 176, 5, 5, 133, 243, 5, 5, 227, 237, 5, 5, 196, 224, 5, 5, 70, 170, 5, 5, 86, 209, 5, 5, 65, 195, 5, 5, 206, 247, 5, 5, 64, 238, 5, 5, 108, 209, 5, 5, 228, 242, 5, 5, 162, 184, 5, 5, 142, 192, 5, 5, 186, 249, 5, 5, 129, 233, 5, 5, 254, 235, 5, 5, 162, 186, 5, 5, 133, 173, 5, 5, 233, 244, 5, 5, 156, 196, 5, 5, 161, 196, 5, 5, 28, 212, 5, 5, 206, 194, 5, 5, 144, 249, 5, 5, 164, 236, 5, 5, 171, 236, 5, 5, 202, 208, 5, 5, 177, 241, 5, 5, 114, 237, 5, 5, 98, 201, 5, 5, 80, 202, 5, 5, 53, 193, 5, 5, 80, 207, 5, 5, 146, 226, 5, 5, 211, 183, 5, 5, 19, 214, 5, 5, 94, 195, 5, 5, 109, 221, 5, 5, 16, 250, 5, 5, 48, 234, 5, 5, 17, 218, 5, 5, 215, 228, 5, 5, 96, 163, 5, 5, 113, 195, 5, 5, 102, 249, 5, 5, 133, 213, 5, 5, 26, 201, 5, 5, 18, 250, 5, 5, 69, 244, 5, 5, 89, 187, 5, 5, 19, 236, 5, 5, 156, 179, 5, 5, 141, 243, 5, 5, 182, 233, 5, 5, 21, 217, 5, 5, 177, 210, 5, 5, 145, 179, 5, 5, 36, 201, 5, 5, 188, 249, 5, 5, 216, 228, 5, 5, 52, 172, 5, 5, 237, 194, 5, 5, 197, 227, 5, 5, 84, 235, 5, 5, 69, 168, 5, 5, 102, 172, 5, 5, 103, 172, 5, 5, 167, 193, 5, 5, 110, 223, 5, 5, 86, 234, 5, 5, 79, 234, 5, 5, 77, 234, 5, 5, 217, 234, 5, 5, 15, 172, 5, 5, 98, 179, 5, 5, 17, 221, 5, 5, 219, 182, 5, 5, 151, 238, 5, 5, 253, 229, 5, 5, 140, 170, 5, 5, 108, 194, 5, 5, 31, 212, 5, 5, 86, 232, 5, 5, 176, 204, 5, 5, 105, 200, 5, 5, 91, 195, 5, 5, 131, 188, 5, 5, 195, 238, 5, 5, 118, 237, 5, 5, 239, 210, 5, 5, 244, 243, 5, 5, 98, 191, 5, 5, 234, 228, 5, 5, 137, 246, 5, 5, 113, 182, 5, 5, 111, 170, 5, 5, 91, 210, 5, 5, 201, 227, 5, 5, 220, 249, 5, 5, 142, 211, 5, 5, 217, 195, 5, 5, 8, 177, 5, 5, 128, 218, 5, 5, 237, 185, 5, 5, 7, 208, 5, 5, 141, 208, 5, 5, 137, 208, 5, 5, 129, 248, 5, 5, 201, 168, 5, 5, 229, 187, 5, 5, 97, 235, 5, 5, 18, 242, 5, 5, 159, 226, 5, 5, 15, 219, 5, 5, 220, 182, 5, 5, 67, 225, 5, 5, 53, 209, 5, 5, 178, 211, 5, 5, 164, 167, 5, 5, 241, 163, 5, 5, 117, 165, 5, 5, 155, 212, 5, 5, 6, 204, 5, 5, 196, 232, 5, 5, 111, 223, 5, 5, 140, 188, 5, 5, 204, 168, 5, 5, 166, 196, 5, 5, 106, 200, 5, 5, 170, 209, 5, 5, 128, 225, 5, 5, 245, 243, 5, 5, 246, 243, 5, 5, 88, 213, 5, 5, 3, 213, 5, 5, 50, 176, 5, 5, 2, 180, 5, 5, 91, 178, 5, 5, 145, 178, 5, 5, 144, 178, 5, 5, 177, 236, 5, 5, 178, 207, 5, 5, 255, 233, 5, 5, 160, 214, 5, 5, 245, 172, 5, 5, 136, 173, 5, 5, 182, 185, 5, 5, 212, 185, 5, 5, 160, 226, 5, 5, 69, 198, 5, 5, 57, 209, 5, 5, 111, 215, 5, 5, 16, 172, 5, 5, 131, 206, 5, 5, 12, 188, 5, 5, 71, 245, 5, 5, 87, 234, 5, 5, 170, 207, 5, 5, 159, 234, 5, 5, 223, 215, 5, 5, 32, 180, 5, 5, 33, 229, 5, 5, 230, 167, 5, 5, 150, 240, 5, 5, 231, 225, 5, 5, 64, 215, 5, 5, 58, 183, 5, 5, 52, 176, 5, 5, 33, 180, 5, 5, 136, 206, 5, 5, 103, 237, 5, 5, 148, 198, 5, 5, 63, 204, 5, 5, 144, 246, 5, 5, 119, 217, 5, 5, 2, 179, 5, 5, 249, 210, 5, 5, 243, 210, 5, 5, 90, 207, 5, 5, 49, 234, 5, 5, 204, 195, 5, 5, 92, 214, 5, 5, 128, 244, 5, 5, 155, 238, 5, 5, 81, 170, 5, 5, 151, 240, 5, 5, 14, 189, 5, 5, 236, 235, 5, 5, 71, 208, 5, 5, 236, 199, 5, 5, 176, 170, 5, 5, 73, 186, 5, 5, 74, 238, 5, 5, 15, 189, 5, 5, 152, 242, 5, 5, 68, 196, 5, 5, 234, 213, 5, 5, 116, 223, 5, 5, 41, 248, 5, 5, 250, 243, 5, 5, 72, 175, 5, 5, 15, 171, 5, 5, 76, 238, 5, 5, 239, 170, 5, 5, 148, 178, 5, 5, 225, 208, 5, 5, 92, 169, 5, 5, 33, 186, 5, 5, 177, 170, 5, 5, 42, 224, 5, 5, 32, 219, 5, 5, 174, 184, 5, 5, 175, 184, 5, 5, 150, 242, 5, 5, 76, 244, 5, 5, 124, 163, 5, 5, 39, 219, 5, 5, 40, 197, 5, 5, 18, 197, 5, 5, 91, 242, 5, 5, 55, 174, 5, 5, 70, 188, 5, 5, 139, 171, 5, 5, 134, 225, 5, 5, 152, 200, 5, 5, 160, 224, 5, 5, 17, 171, 5, 5, 78, 186, 5, 5, 241, 199, 5, 5, 147, 217, 5, 5, 169, 230, 5, 5, 162, 224, 5, 5, 232, 202, 5, 5, 104, 207, 5, 5, 160, 200, 5, 5, 209, 227, 5, 5, 205, 240, 5, 5, 212, 188, 5, 5, 118, 183, 5, 5, 57, 171, 5, 5, 120, 180, 5, 5, 203, 207, 5, 5, 210, 192, 5, 5, 211, 192, 5, 5, 191, 222, 5, 5, 130, 191, 5, 5, 198, 248, 5, 5, 82, 245, 5, 5, 188, 204, 5, 5, 181, 216, 5, 5, 128, 229, 5, 5, 48, 231, 5, 5, 106, 213, 5, 5, 86, 225, 5, 5, 183, 245, 5, 5, 209, 203, 5, 5, 136, 191, 5, 5, 196, 204, 5, 5, 26, 211, 5, 5, 186, 199, 5, 5, 27, 211, 5, 5, 134, 220, 5, 5, 15, 179, 5, 5, 124, 208, 5, 5, 48, 203, 5, 5, 33, 177, 5, 5, 54, 232, 5, 5, 51, 203, 5, 5, 135, 202, 5, 5, 84, 218, 5, 5, 167, 195, 5, 5, 81, 202, 5, 5, 124, 196, 5, 5, 242, 191, 5, 5, 134, 176, 5, 5, 143, 220, 5, 5, 41, 211, 5, 5, 150, 191, 5, 5, 45, 205, 5, 5, 176, 203, 5, 5, 115, 248, 5, 5, 46, 205, 5, 5, 245, 242, 5, 5, 43, 190, 5, 5, 241, 198, 5, 5, 147, 220, 5, 5, 146, 213, 5, 5, 151, 213, 5, 5, 246, 225, 5, 5, 112, 175, 5, 5, 225, 211, 5, 5, 53, 199, 5, 5, 54, 199, 5, 5, 55, 199, 5, 5, 189, 237, 5, 5, 38, 227, 5, 5, 210, 180, 5, 5, 47, 170, 5, 5, 66, 199, 5, 5, 38, 187, 5, 5, 160, 173, 5, 5, 220, 236, 5, 5, 215, 165, 5, 5, 165, 177, 5, 5, 84, 208, 5, 5, 245, 228, 5, 5, 155, 239, 5, 5, 225, 239, 5, 5, 90, 228, 5, 5, 73, 199, 5, 5, 160, 172, 5, 5, 229, 242, 5, 5, 14, 228, 5, 5, 35, 228, 5, 5, 7, 234, 5, 5, 59, 219, 5, 5, 246, 209, 5, 5, 16, 185, 5, 5, 214, 168, 5, 5, 0, 132, 5, 5, 46, 175, 5, 5, 172, 230, 5, 5, 40, 173, 5, 5, 226, 230, 5, 5, 44, 232, 5, 5, 252, 181, 5, 5, 42, 221, 5, 5, 93, 177, 5, 5, 226, 195, 5, 5, 139, 247, 5, 5, 3, 196, 5, 5, 133, 192, 5, 5, 135, 228, 5, 5, 125, 223, 5, 5, 253, 181, 5, 5, 32, 193, 5, 5, 146, 195, 5, 5, 85, 245, 5, 5, 66, 176, 5, 5, 55, 210, 5, 5, 99, 218, 5, 5, 115, 230, 5, 5, 52, 196, 5, 5, 160, 170, 5, 5, 30, 236, 5, 5, 14, 244, 5, 5, 49, 184, 5, 5, 100, 164, 5, 5, 181, 223, 5, 5, 26, 230, 5, 5, 173, 212, 5, 5, 117, 210, 5, 5, 67, 164, 5, 5, 42, 203, 5, 5, 161, 247, 5, 5, 66, 174, 5, 5, 140, 238, 5, 5, 174, 212, 5, 5, 198, 167, 5, 5, 148, 218, 5, 5, 182, 218, 5, 5, 193, 237, 5, 5, 67, 176, 5, 5, 220, 219, 5, 5, 89, 225, 5, 5, 156, 202, 5, 5, 212, 242, 5, 5, 10, 197, 5, 5, 231, 165, 5, 5, 223, 223, 5, 5, 192, 206, 5, 5, 50, 247, 5, 5, 247, 198, 5, 5, 202, 248, 5, 5, 122, 207, 5, 5, 11, 163, 5, 5, 14, 211, 5, 5, 4, 176, 5, 5, 104, 221, 5, 5, 161, 249, 5, 5, 190, 246, 5, 5, 3, 183, 5, 5, 17, 240, 5, 5, 61, 229, 5, 5, 254, 221, 5, 5, 39, 167, 5, 5, 212, 179, 5, 5, 189, 175, 5, 5, 233, 195, 5, 5, 127, 219, 5, 5, 40, 167, 5, 5, 151, 174, 5, 5, 9, 167, 5, 5, 0, 156, 5, 5, 130, 238, 5, 5, 235, 231, 5, 5, 110, 182, 5, 5, 133, 195, 5, 5, 189, 190, 5, 5, 73, 211, 5, 5, 208, 249, 5, 5, 60, 199, 5, 5, 42, 164, 5, 5, 59, 166, 5, 5, 154, 218, 5, 5, 136, 167, 5, 5, 228, 179, 5, 5, 126, 166, 5, 5, 201, 167, 5, 5, 109, 179, 5, 5, 212, 241, 5, 5, 210, 220, 5, 5, 61, 219, 5, 5, 3, 244, 5, 5, 43, 164, 5, 5, 77, 197, 5, 5, 182, 183, 5, 5, 84, 241, 5, 5, 205, 176, 5, 5, 86, 223, 5, 5, 202, 167, 5, 5, 5, 203, 5, 5, 91, 169, 5, 5, 251, 169, 5, 5, 167, 224, 5, 5, 225, 240, 5, 5, 9, 224, 5, 5, 161, 211, 5, 5, 253, 171, 5, 5, 205, 225, 5, 5, 6, 203, 5, 5, 71, 174, 5, 5, 94, 219, 5, 5, 192, 221, 5, 5, 228, 175, 5, 5, 206, 223, 5, 5, 66, 172, 5, 5, 61, 184, 5, 5, 227, 184, 5, 5, 156, 174, 5, 5, 10, 167, 5, 5, 104, 164, 5, 5, 208, 166, 5, 5, 216, 188, 5, 5, 223, 182, 5, 5, 197, 193, 5, 5, 118, 210, 5, 5, 48, 207, 5, 5, 53, 179, 5, 5, 0, 159, 5, 5, 5, 182, 5, 5, 6, 182, 5, 5, 131, 238, 5, 5, 195, 206, 5, 5, 6, 222, 5, 5, 80, 167, 5, 5, 209, 219, 5, 5, 170, 163, 5, 5, 168, 163, 5, 5, 206, 176, 5, 5, 183, 183, 5, 5, 49, 207, 5, 5, 53, 196, 5, 5, 29, 191, 5, 5, 13, 205, 5, 5, 75, 209, 5, 5, 211, 249, 5, 5, 18, 186, 5, 5, 110, 184, 5, 5, 38, 163, 5, 5, 237, 231, 5, 5, 69, 210, 5, 5, 163, 198, 5, 5, 54, 196, 5, 5, 50, 227, 5, 5, 143, 203, 5, 5, 106, 221, 5, 5, 100, 184, 5, 5, 219, 194, 5, 5, 221, 194, 5, 5, 59, 212, 5, 5, 222, 229, 5, 5, 31, 235, 5, 5, 212, 231, 5, 5, 120, 195, 5, 5, 148, 239, 5, 5, 4, 214, 5, 5, 235, 233, 5, 5, 3, 200, 5, 5, 129, 168, 5, 5, 79, 222, 5, 5, 238, 226, 5, 5, 224, 182, 5, 5, 152, 163, 5, 5, 36, 223, 5, 5, 130, 193, 5, 5, 197, 218, 5, 5, 55, 196, 5, 5, 81, 178, 5, 5, 82, 178, 5, 5, 159, 218, 5, 5, 102, 209, 5, 5, 239, 238, 5, 5, 209, 184, 5, 5, 238, 220, 5, 5, 163, 191, 5, 5, 85, 204, 5, 5, 47, 196, 5, 5, 10, 179, 5, 5, 225, 182, 5, 5, 15, 181, 5, 5, 182, 187, 5, 5, 63, 219, 5, 5, 61, 191, 5, 5, 25, 177, 5, 5, 137, 245, 5, 5, 130, 207, 5, 5, 218, 187, 5, 5, 21, 207, 5, 5, 207, 218, 5, 5, 160, 190, 5, 5, 30, 241, 5, 5, 168, 245, 5, 5, 54, 179, 5, 5, 5, 214, 5, 5, 48, 183, 5, 5, 82, 176, 5, 5, 31, 204, 5, 5, 71, 221, 5, 5, 130, 202, 5, 5, 4, 200, 5, 5, 88, 215, 5, 5, 184, 248, 5, 5, 26, 238, 5, 5, 8, 182, 5, 5, 155, 184, 5, 5, 197, 234, 5, 5, 78, 163, 5, 5, 233, 192, 5, 5, 204, 180, 5, 5, 241, 220, 5, 5, 101, 225, 5, 5, 198, 235, 5, 5, 199, 235, 5, 5, 245, 195, 5, 5, 238, 169, 5, 5, 41, 191, 5, 5, 200, 206, 5, 5, 153, 167, 5, 5, 18, 205, 5, 5, 58, 240, 5, 5, 168, 181, 5, 5, 56, 228, 5, 5, 231, 230, 5, 5, 204, 246, 5, 5, 8, 214, 5, 5, 130, 223, 5, 5, 237, 173, 5, 5, 35, 204, 5, 5, 58, 245, 5, 5, 46, 208, 5, 5, 164, 204, 5, 5, 209, 182, 5, 5, 226, 196, 5, 5, 192, 199, 5, 5, 203, 181, 5, 5, 9, 222, 5, 5, 41, 167, 5, 5, 63, 191, 5, 5, 0, 143, 5, 5, 123, 198, 5, 5, 228, 227, 5, 5, 224, 183, 5, 5, 216, 207, 5, 5, 25, 232, 5, 5, 139, 229, 5, 5, 62, 241, 5, 5, 138, 207, 5, 5, 114, 166, 5, 5, 194, 246, 5, 5, 136, 166, 5, 5, 84, 233, 5, 5, 59, 240, 5, 5, 186, 219, 5, 5, 64, 191, 5, 5, 40, 236, 5, 5, 9, 182, 5, 5, 165, 185, 5, 5, 78, 237, 5, 5, 143, 164, 5, 5, 19, 205, 5, 5, 214, 200, 5, 5, 139, 175, 5, 5, 167, 246, 5, 5, 236, 246, 5, 5, 226, 210, 5, 5, 187, 167, 5, 5, 184, 227, 5, 5, 194, 187, 5, 5, 235, 171, 5, 5, 165, 170, 5, 5, 166, 174, 5, 5, 160, 163, 5, 5, 66, 207, 5, 5, 135, 175, 5, 5, 185, 245, 5, 5, 35, 238, 5, 5, 234, 207, 5, 5, 216, 248, 5, 5, 127, 233, 5, 5, 168, 224, 5, 5, 31, 226, 5, 5, 154, 203, 5, 5, 187, 181, 5, 5, 131, 164, 5, 5, 242, 182, 5, 5, 73, 184, 5, 5, 232, 230, 5, 5, 243, 232, 5, 5, 194, 245, 5, 5, 200, 212, 5, 5, 50, 215, 5, 5, 105, 225, 5, 5, 126, 177, 5, 5, 31, 236, 5, 5, 190, 190, 5, 5, 137, 211, 5, 5, 194, 192, 5, 5, 56, 178, 5, 5, 24, 228, 5, 5, 102, 177, 5, 5, 42, 167, 5, 5, 214, 176, 5, 5, 37, 221, 5, 5, 22, 201, 5, 5, 131, 202, 5, 5, 152, 179, 5, 5, 73, 209, 5, 5, 186, 208, 5, 5, 211, 166, 5, 5, 74, 229, 5, 5, 80, 210, 5, 5, 109, 189, 5, 5, 54, 241, 5, 5, 35, 241, 5, 5, 142, 207, 5, 5, 80, 221, 5, 5, 61, 189, 5, 5, 132, 193, 5, 5, 72, 240, 5, 5, 196, 199, 5, 5, 93, 228, 5, 5, 107, 221, 5, 5, 216, 237, 5, 5, 21, 234, 5, 5, 40, 180, 5, 5, 194, 223, 5, 5, 56, 166, 5, 5, 100, 203, 5, 5, 223, 180, 5, 5, 65, 197, 5, 5, 73, 182, 5, 5, 224, 242, 5, 5, 198, 234, 5, 5, 11, 220, 5, 5, 209, 206, 5, 5, 119, 181, 5, 5, 125, 210, 5, 5, 206, 195, 5, 5, 153, 163, 5, 5, 160, 166, 5, 5, 101, 203, 5, 5, 230, 189, 5, 5, 195, 193, 5, 5, 57, 172, 5, 5, 251, 192, 5, 5, 221, 218, 5, 5, 11, 176, 5, 5, 29, 171, 5, 5, 94, 234, 5, 5, 17, 236, 5, 5, 101, 194, 5, 5, 191, 167, 5, 5, 100, 249, 5, 5, 164, 184, 5, 5, 27, 205, 5, 5, 87, 221, 5, 5, 132, 204, 5, 5, 28, 205, 5, 5, 197, 247, 5, 5, 182, 242, 5, 5, 119, 196, 5, 5, 112, 183, 5, 5, 116, 201, 5, 5, 36, 230, 5, 5, 39, 168, 5, 5, 231, 245, 5, 5, 255, 235, 5, 5, 118, 232, 5, 5, 105, 217, 5, 5, 149, 171, 5, 5, 168, 180, 5, 5, 225, 183, 5, 5, 132, 223, 5, 5, 74, 227, 5, 5, 130, 233, 5, 5, 249, 220, 5, 5, 103, 239, 5, 5, 232, 217, 5, 5, 242, 249, 5, 5, 197, 165, 5, 5, 38, 244, 5, 5, 225, 200, 5, 5, 230, 242, 5, 5, 134, 232, 5, 5, 233, 238, 5, 5, 196, 216, 5, 5, 38, 241, 5, 5, 198, 245, 5, 5, 34, 195, 5, 5, 19, 226, 5, 5, 60, 202, 5, 5, 226, 180, 5, 5, 236, 184, 5, 5, 242, 184, 5, 5, 197, 224, 5, 5, 78, 195, 5, 5, 133, 204, 5, 5, 135, 204, 5, 5, 210, 181, 5, 5, 108, 203, 5, 5, 15, 200, 5, 5, 176, 192, 5, 5, 180, 226, 5, 5, 148, 214, 5, 5, 39, 177, 5, 5, 244, 213, 5, 5, 45, 249, 5, 5, 128, 210, 5, 5, 64, 241, 5, 5, 133, 223, 5, 5, 69, 219, 5, 5, 187, 163, 5, 5, 170, 241, 5, 5, 232, 245, 5, 5, 206, 164, 5, 5, 66, 186, 5, 5, 65, 247, 5, 5, 17, 211, 5, 5, 69, 224, 5, 5, 146, 191, 5, 5, 254, 185, 5, 5, 131, 233, 5, 5, 218, 192, 5, 5, 42, 204, 5, 5, 92, 192, 5, 5, 20, 211, 5, 5, 67, 231, 5, 5, 24, 212, 5, 5, 11, 224, 5, 5, 20, 226, 5, 5, 117, 220, 5, 5, 82, 237, 5, 5, 71, 237, 5, 5, 97, 163, 5, 5, 221, 225, 5, 5, 68, 231, 5, 5, 166, 164, 5, 5, 172, 234, 5, 5, 20, 236, 5, 5, 13, 195, 5, 5, 178, 241, 5, 5, 141, 175, 5, 5, 85, 218, 5, 5, 153, 184, 5, 5, 231, 184, 5, 5, 217, 204, 5, 5, 45, 236, 5, 5, 46, 210, 5, 5, 26, 192, 5, 5, 153, 215, 5, 5, 186, 229, 5, 5, 117, 236, 5, 5, 227, 205, 5, 5, 132, 171, 5, 5, 235, 237, 5, 5, 36, 226, 5, 5, 165, 184, 5, 5, 61, 228, 5, 5, 62, 179, 5, 5, 98, 219, 5, 5, 77, 180, 5, 5, 63, 167, 5, 5, 73, 164, 5, 5, 10, 175, 5, 5, 90, 187, 5, 5, 157, 179, 5, 5, 11, 199, 5, 5, 249, 218, 5, 5, 145, 245, 5, 5, 116, 199, 5, 5, 92, 210, 5, 5, 66, 247, 5, 5, 75, 211, 5, 5, 87, 181, 5, 5, 180, 220, 5, 5, 236, 237, 5, 5, 20, 214, 5, 5, 168, 184, 5, 5, 74, 212, 5, 5, 150, 223, 5, 5, 116, 219, 5, 5, 181, 191, 5, 5, 63, 245, 5, 5, 8, 239, 5, 5, 198, 227, 5, 5, 227, 195, 5, 5, 199, 227, 5, 5, 84, 229, 5, 5, 164, 163, 5, 5, 179, 241, 5, 5, 36, 166, 5, 5, 46, 236, 5, 5, 234, 179, 5, 5, 213, 235, 5, 5, 137, 224, 5, 5, 210, 236, 5, 5, 137, 204, 5, 5, 157, 226, 5, 5, 126, 226, 5, 5, 196, 193, 5, 5, 73, 166, 5, 5, 245, 173, 5, 5, 197, 229, 5, 5, 2, 177, 5, 5, 49, 209, 5, 5, 50, 236, 5, 5, 189, 219, 5, 5, 174, 188, 5, 5, 68, 225, 5, 5, 19, 223, 5, 5, 109, 194, 5, 5, 38, 190, 5, 5, 145, 190, 5, 5, 3, 186, 5, 5, 134, 233, 5, 5, 239, 191, 5, 5, 81, 199, 5, 5, 73, 219, 5, 5, 136, 226, 5, 5, 126, 180, 5, 5, 146, 199, 5, 5, 218, 234, 5, 5, 9, 183, 5, 5, 72, 169, 5, 5, 17, 206, 5, 5, 231, 193, 5, 5, 212, 244, 5, 5, 125, 218, 5, 5, 112, 238, 5, 5, 16, 219, 5, 5, 247, 183, 5, 5, 141, 170, 5, 5, 159, 224, 5, 5, 139, 195, 5, 5, 150, 225, 5, 5, 202, 227, 5, 5, 114, 182, 5, 5, 52, 204, 5, 5, 66, 169, 5, 5, 73, 169, 5, 5, 150, 219, 5, 5, 97, 234, 5, 5, 251, 195, 5, 5, 151, 165, 5, 5, 63, 206, 5, 5, 142, 243, 5, 5, 99, 192, 5, 5, 93, 210, 5, 5, 24, 188, 5, 5, 93, 202, 5, 5, 111, 244, 5, 5, 110, 207, 5, 5, 61, 168, 5, 5, 59, 168, 5, 5, 166, 185, 5, 5, 191, 178, 5, 5, 220, 215, 5, 5, 23, 231, 5, 5, 49, 194, 5, 5, 136, 232, 5, 5, 130, 240, 5, 5, 112, 215, 5, 5, 71, 225, 5, 5, 72, 225, 5, 5, 200, 235, 5, 5, 56, 183, 5, 5, 187, 201, 5, 5, 134, 235, 5, 5, 143, 198, 5, 5, 183, 194, 5, 5, 12, 175, 5, 5, 81, 233, 5, 5, 79, 184, 5, 5, 11, 189, 5, 5, 110, 221, 5, 5, 27, 200, 5, 5, 190, 222, 5, 5, 60, 204, 5, 5, 140, 202, 5, 5, 31, 234, 5, 5, 14, 176, 5, 5, 15, 176, 5, 5, 198, 223, 5, 5, 58, 163, 5, 5, 184, 223, 5, 5, 111, 207, 5, 5, 177, 174, 5, 5, 221, 187, 5, 5, 137, 164, 5, 5, 178, 174, 5, 5, 201, 209, 5, 5, 59, 163, 5, 5, 174, 234, 5, 5, 51, 222, 5, 5, 145, 222, 5, 5, 92, 246, 5, 5, 217, 168, 5, 5, 21, 211, 5, 5, 188, 178, 5, 5, 80, 184, 5, 5, 121, 199, 5, 5, 28, 214, 5, 5, 182, 189, 5, 5, 61, 204, 5, 5, 5, 235, 5, 5, 74, 179, 5, 5, 149, 198, 5, 5, 202, 209, 5, 5, 182, 245, 5, 5, 51, 205, 5, 5, 145, 192, 5, 5, 226, 249, 5, 5, 92, 185, 5, 5, 136, 183, 5, 5, 241, 189, 5, 5, 107, 228, 5, 5, 52, 222, 5, 5, 5, 232, 5, 5, 219, 168, 5, 5, 12, 226, 5, 5, 5, 202, 5, 5, 85, 176, 5, 5, 237, 235, 5, 5, 32, 234, 5, 5, 121, 211, 5, 5, 120, 211, 5, 5, 239, 200, 5, 5, 121, 203, 5, 5, 245, 210, 5, 5, 76, 196, 5, 5, 249, 175, 5, 5, 34, 214, 5, 5, 243, 194, 5, 5, 31, 168, 5, 5, 154, 163, 5, 5, 156, 238, 5, 5, 91, 207, 5, 5, 111, 241, 5, 5, 18, 238, 5, 5, 30, 171, 5, 5, 70, 236, 5, 5, 44, 171, 5, 5, 141, 202, 5, 5, 127, 179, 5, 5, 61, 222, 5, 5, 151, 175, 5, 5, 161, 216, 5, 5, 76, 186, 5, 5, 225, 233, 5, 5, 33, 207, 5, 5, 68, 241, 5, 5, 216, 230, 5, 5, 126, 241, 5, 5, 121, 241, 5, 5, 194, 183, 5, 5, 207, 245, 5, 5, 106, 202, 5, 5, 40, 166, 5, 5, 125, 184, 5, 5, 155, 191, 5, 5, 40, 219, 5, 5, 26, 219, 5, 5, 69, 196, 5, 5, 80, 169, 5, 5, 139, 205, 5, 5, 38, 214, 5, 5, 209, 187, 5, 5, 237, 179, 5, 5, 165, 163, 5, 5, 72, 236, 5, 5, 246, 189, 5, 5, 116, 191, 5, 5, 28, 230, 5, 5, 157, 180, 5, 5, 56, 167, 5, 5, 77, 244, 5, 5, 155, 185, 5, 5, 226, 233, 5, 5, 96, 229, 5, 5, 138, 232, 5, 5, 85, 184, 5, 5, 168, 215, 5, 5, 77, 208, 5, 5, 172, 199, 5, 5, 123, 187, 5, 5, 251, 184, 5, 5, 207, 200, 5, 5, 194, 244, 5, 5, 204, 189, 5, 5, 36, 191, 5, 5, 101, 210, 5, 5, 108, 210, 5, 5, 139, 232, 5, 5, 106, 210, 5, 5, 44, 186, 5, 5, 75, 215, 5, 5, 138, 244, 5, 5, 230, 169, 5, 5, 165, 230, 5, 5, 77, 224, 5, 5, 235, 244, 5, 5, 9, 235, 5, 5, 40, 214, 5, 5, 210, 227, 5, 5, 109, 190, 5, 5, 79, 208, 5, 5, 58, 171, 5, 5, 237, 186, 5, 5, 141, 232, 5, 5, 172, 190, 5, 5, 238, 179, 5, 5, 65, 233, 5, 5, 153, 187, 5, 5, 142, 177, 5, 5, 150, 245, 5, 5, 88, 176, 5, 5, 79, 204, 5, 5, 204, 207, 5, 5, 83, 186, 5, 5, 163, 187, 5, 5, 138, 183, 5, 5, 55, 218, 5, 5, 29, 194, 5, 5, 8, 194, 5, 5, 128, 208, 5, 5, 243, 217, 5, 5, 212, 168, 5, 5, 213, 168, 5, 5, 13, 228, 5, 5, 118, 195, 5, 5, 9, 242, 5, 5, 229, 191, 5, 5, 215, 239, 5, 5, 191, 194, 5, 5, 104, 192, 5, 5, 68, 164, 5, 5, 181, 173, 5, 5, 115, 221, 5, 5, 193, 166, 5, 5, 124, 177, 5, 5, 147, 195, 5, 5, 2, 172, 5, 5, 96, 245, 5, 5, 80, 193, 5, 5, 105, 197, 5, 5, 237, 233, 5, 5, 216, 214, 5, 5, 75, 171, 5, 5, 212, 225, 5, 5, 166, 191, 5, 5, 169, 176, 5, 5, 7, 217, 5, 5, 91, 204, 5, 5, 5, 172, 5, 5, 188, 224, 5, 5, 168, 204, 5, 5, 155, 228, 5, 5, 64, 178, 5, 5, 20, 248, 5, 5, 93, 223, 5, 5, 102, 238, 5, 5, 171, 163, 5, 5, 8, 217, 5, 5, 92, 204, 5, 5, 65, 178, 5, 5, 170, 188, 5, 5, 239, 183, 5, 5, 22, 165, 5, 5, 31, 225, 5, 5, 151, 177, 5, 5, 57, 178, 5, 5, 88, 240, 5, 5, 140, 192, 5, 5, 104, 236, 5, 5, 45, 195, 5, 5, 41, 243, 5, 5, 184, 178, 5, 5, 226, 201, 5, 5, 220, 179, 5, 5, 222, 214, 5, 5, 68, 178, 5, 5, 46, 204, 5, 5, 169, 206, 5, 5, 41, 244, 5, 5, 203, 213, 5, 5, 238, 171, 5, 5, 147, 229, 5, 5, 39, 238, 5, 5, 135, 209, 5, 5, 40, 238, 5, 5, 111, 221, 5, 5, 183, 176, 5, 5, 255, 220, 5, 5, 48, 174, 5, 5, 156, 229, 5, 5, 237, 187, 5, 5, 181, 211, 5, 5, 51, 218, 5, 5, 239, 239, 5, 5, 64, 204, 5, 5, 132, 228, 5, 5, 65, 204, 5, 5, 62, 222, 5, 5, 165, 240, 5, 5, 178, 188, 5, 5, 83, 165, 5, 5, 223, 186, 5, 5, 231, 214, 5, 5, 232, 214, 5, 5, 19, 247, 5, 5, 56, 217, 5, 5, 163, 208, 5, 5, 175, 208, 5, 5, 187, 219, 5, 5, 13, 166, 5, 5, 143, 207, 5, 5, 188, 219, 5, 5, 109, 186, 5, 5, 185, 172, 5, 5, 72, 164, 5, 5, 74, 164, 5, 5, 30, 186, 5, 5, 43, 248, 5, 5, 49, 186, 5, 5, 113, 246, 5, 5, 197, 175, 5, 5, 11, 229, 5, 5, 21, 188, 5, 5, 72, 203, 5, 5, 2, 199, 5, 5, 13, 214, 5, 5, 78, 203, 5, 5, 23, 248, 5, 5, 181, 220, 5, 5, 99, 204, 5, 5, 204, 245, 5, 5, 52, 181, 5, 5, 215, 175, 5, 5, 161, 205, 5, 5, 81, 223, 5, 5, 163, 243, 5, 5, 157, 174, 5, 5, 28, 211, 5, 5, 107, 220, 5, 5, 200, 211, 5, 5, 26, 222, 5, 5, 112, 220, 5, 5, 165, 178, 5, 5, 201, 234, 5, 5, 115, 235, 5, 5, 4, 216, 5, 5, 40, 222, 5, 5, 49, 222, 5, 5, 241, 210, 5, 5, 238, 188, 5, 5, 63, 222, 5, 5, 232, 225, 5, 5, 179, 178, 5, 5, 205, 221, 5, 5, 146, 173, 5, 5, 125, 224, 5, 5, 173, 198, 5, 5, 166, 247, 5, 5, 191, 226, 5, 5, 250, 166, 5, 5, 37, 242, 5, 5, 253, 169, 5, 5, 117, 167, 5, 5, 176, 208, 5, 5, 78, 220, 5, 5, 25, 170, 5, 5, 65, 191, 5, 5, 251, 166, 5, 5, 213, 167, 5, 5, 38, 188, 5, 5, 33, 174, 5, 5, 175, 173, 5, 5, 137, 217, 5, 5, 88, 206, 5, 5, 89, 206, 5, 5, 187, 183, 5, 5, 222, 232, 5, 5, 234, 219, 5, 5, 231, 239, 5, 5, 45, 188, 5, 5, 234, 227, 5, 5, 103, 205, 5, 5, 207, 247, 5, 5, 235, 239, 5, 5, 209, 208, 5, 5, 27, 219, 5, 5, 246, 172, 5, 5, 72, 168, 5, 5, 248, 172, 5, 5, 98, 220, 5, 5, 97, 196, 5, 5, 50, 237, 5, 5, 248, 247, 5, 5, 12, 245, 5, 5, 148, 202, 5, 5, 164, 202, 5, 5, 184, 202, 5, 5, 202, 199, 5, 5, 8, 230, 5, 5, 20, 176, 5, 5, 155, 204, 5, 5, 144, 218, 5, 5, 68, 232, 5, 5, 252, 213, 5, 5, 11, 181, 5, 5, 211, 217, 5, 5, 241, 216, 5, 5, 118, 191, 5, 5, 233, 184, 5, 5, 152, 192, 5, 5, 118, 198, 5, 5, 107, 179, 5, 5, 224, 176, 5, 5, 182, 216, 5, 5, 123, 224, 5, 5, 230, 226, 5, 5, 24, 197, 5, 5, 16, 244, 5, 5, 235, 226, 5, 5, 108, 247, 5, 5, 227, 171, 5, 5, 228, 171, 5, 5, 154, 182, 5, 5, 77, 226, 5, 5, 51, 245, 5, 5, 133, 249, 5, 5, 203, 163, 5, 5, 15, 239, 5, 5, 23, 232, 5, 5, 86, 164, 5, 5, 138, 203, 5, 5, 197, 196, 5, 5, 13, 229, 5, 5, 67, 199, 5, 5, 170, 173, 5, 5, 244, 197, 5, 5, 32, 194, 5, 5, 238, 224, 5, 5, 94, 217, 5, 5, 27, 172, 5, 5, 245, 200, 5, 5, 73, 173, 5, 5, 255, 210, 5, 5, 221, 198, 5, 5, 29, 240, 5, 5, 176, 212, 5, 5, 246, 215, 5, 5, 161, 242, 5, 5, 24, 232, 5, 5, 171, 173, 5, 5, 243, 179, 5, 5, 207, 228, 5, 5, 52, 207, 5, 5, 29, 244, 5, 5, 129, 232, 5, 5, 63, 166, 5, 5, 119, 216, 5, 5, 223, 244, 5, 5, 160, 217, 5, 5, 194, 166, 5, 5, 213, 220, 5, 5, 95, 189, 5, 5, 79, 226, 5, 5, 245, 197, 5, 5, 254, 169, 5, 5, 241, 232, 5, 5, 105, 169, 5, 5, 240, 165, 5, 5, 252, 216, 5, 5, 246, 200, 5, 5, 210, 219, 5, 5, 141, 233, 5, 5, 51, 227, 5, 5, 52, 227, 5, 5, 59, 248, 5, 5, 200, 172, 5, 5, 26, 177, 5, 5, 51, 170, 5, 5, 203, 183, 5, 5, 28, 225, 5, 5, 253, 246, 5, 5, 228, 178, 5, 5, 45, 238, 5, 5, 64, 166, 5, 5, 16, 239, 5, 5, 50, 220, 5, 5, 229, 190, 5, 5, 140, 215, 5, 5, 28, 197, 5, 5, 44, 181, 5, 5, 237, 223, 5, 5, 244, 226, 5, 5, 88, 201, 5, 5, 89, 201, 5, 5, 149, 217, 5, 5, 115, 196, 5, 5, 188, 216, 5, 5, 36, 204, 5, 5, 216, 205, 5, 5, 208, 218, 5, 5, 217, 205, 5, 5, 246, 228, 5, 5, 170, 182, 5, 5, 181, 201, 5, 5, 32, 243, 5, 5, 219, 191, 5, 5, 80, 216, 5, 5, 223, 227, 5, 5, 202, 214, 5, 5, 89, 241, 5, 5, 111, 249, 5, 5, 7, 200, 5, 5, 86, 198, 5, 5, 127, 198, 5, 5, 53, 208, 5, 5, 209, 246, 5, 5, 199, 215, 5, 5, 202, 196, 5, 5, 56, 220, 5, 5, 201, 212, 5, 5, 102, 225, 5, 5, 42, 229, 5, 5, 103, 166, 5, 5, 139, 219, 5, 5, 140, 219, 5, 5, 103, 245, 5, 5, 128, 198, 5, 5, 81, 200, 5, 5, 253, 197, 5, 5, 189, 224, 5, 5, 210, 246, 5, 5, 105, 181, 5, 5, 41, 222, 5, 5, 40, 233, 5, 5, 37, 220, 5, 5, 249, 168, 5, 5, 127, 206, 5, 5, 215, 212, 5, 5, 80, 228, 5, 5, 188, 183, 5, 5, 216, 212, 5, 5, 114, 211, 5, 5, 14, 209, 5, 5, 93, 206, 5, 5, 16, 194, 5, 5, 108, 221, 5, 5, 89, 240, 5, 5, 241, 218, 5, 5, 248, 238, 5, 5, 214, 246, 5, 5, 226, 188, 5, 5, 74, 189, 5, 5, 244, 224, 5, 5, 95, 247, 5, 5, 203, 214, 5, 5, 145, 232, 5, 5, 227, 172, 5, 5, 106, 217, 5, 5, 248, 192, 5, 5, 177, 239, 5, 5, 204, 213, 5, 5, 113, 249, 5, 5, 19, 200, 5, 5, 234, 216, 5, 5, 12, 172, 5, 5, 125, 196, 5, 5, 55, 241, 5, 5, 226, 244, 5, 5, 111, 229, 5, 5, 76, 246, 5, 5, 109, 229, 5, 5, 117, 211, 5, 5, 71, 170, 5, 5, 106, 183, 5, 5, 37, 205, 5, 5, 182, 184, 5, 5, 249, 238, 5, 5, 42, 244, 5, 5, 83, 226, 5, 5, 182, 241, 5, 5, 231, 236, 5, 5, 235, 205, 5, 5, 40, 172, 5, 5, 53, 204, 5, 5, 99, 195, 5, 5, 28, 192, 5, 5, 127, 247, 5, 5, 46, 225, 5, 5, 50, 209, 5, 5, 8, 198, 5, 5, 56, 243, 5, 5, 95, 187, 5, 5, 68, 238, 5, 5, 199, 202, 5, 5, 212, 202, 5, 5, 174, 170, 5, 5, 147, 232, 5, 5, 86, 214, 5, 5, 9, 198, 5, 5, 104, 172, 5, 5, 133, 169, 5, 5, 53, 235, 5, 5, 216, 183, 5, 5, 22, 221, 5, 5, 107, 200, 5, 5, 54, 164, 5, 5, 144, 198, 5, 5, 142, 217, 5, 5, 61, 226, 5, 5, 240, 187, 5, 5, 65, 215, 5, 5, 53, 176, 5, 5, 132, 206, 5, 5, 62, 168, 5, 5, 239, 188, 5, 5, 134, 182, 5, 5, 90, 213, 5, 5, 152, 240, 5, 5, 218, 202, 5, 5, 103, 201, 5, 5, 110, 175, 5, 5, 112, 241, 5, 5, 113, 241, 5, 5, 104, 187, 5, 5, 7, 180, 5, 5, 196, 182, 5, 5, 207, 214, 5, 5, 216, 172, 5, 5, 197, 228, 5, 5, 244, 219, 5, 5, 13, 198, 5, 5, 166, 240, 5, 5, 86, 216, 5, 5, 132, 216, 5, 5, 172, 228, 5, 5, 69, 204, 5, 5, 60, 164, 5, 5, 182, 206, 5, 5, 71, 243, 5, 5, 77, 238, 5, 5, 122, 241, 5, 5, 202, 230, 5, 5, 116, 229, 5, 5, 240, 170, 5, 5, 200, 217, 5, 5, 251, 241, 5, 5, 12, 169, 5, 5, 124, 187, 5, 5, 209, 224, 5, 5, 41, 197, 5, 5, 62, 177, 5, 5, 15, 188, 5, 5, 82, 243, 5, 5, 220, 199, 5, 5, 209, 189, 5, 5, 33, 213, 5, 5, 92, 231, 5, 5, 3, 191, 5, 5, 137, 183, 5, 5, 130, 190, 5, 5, 173, 190, 5, 5, 226, 227, 5, 5, 121, 180, 5, 5, 16, 243, 5, 5, 184, 221, 5, 5, 188, 248, 5, 5, 189, 248, 5, 5, 150, 211, 5, 5, 80, 241, 5, 5, 246, 192, 5, 5, 82, 241, 5, 5, 222, 198, 5, 5, 94, 197, 5, 5, 71, 229, 5, 5, 35, 234, 5, 5, 34, 234, 5, 5, 56, 229, 5, 5, 136, 170, 5, 5, 133, 234, 5, 5, 136, 234, 5, 5, 227, 180, 5, 5, 39, 233, 5, 5, 85, 234, 5, 5, 17, 217, 5, 5, 94, 200, 5, 5, 217, 243, 5, 5, 111, 209, 5, 5, 88, 173, 5, 5, 18, 168, 5, 5, 197, 248, 5, 5, 68, 242, 5, 5, 26, 231, 5, 5, 150, 167, 5, 5, 96, 173, 5, 5, 72, 174, 5, 5, 96, 189, 5, 5, 180, 172, 5, 5, 201, 211, 5, 5, 215, 180, 5, 5, 218, 184, 5, 5, 45, 181, 5, 5, 181, 229, 5, 5, 242, 220, 5, 5, 130, 186, 5, 5, 129, 186, 5, 5, 25, 233, 5, 5, 33, 172, 5, 5, 34, 201, 5, 5, 111, 247, 5, 5, 131, 182, 5, 5, 240, 223, 5, 5, 227, 201, 5, 5, 129, 208, 5, 5, 107, 244, 5, 5, 98, 233, 5, 5, 112, 247, 5, 5, 197, 182, 5, 5, 138, 186, 5, 5, 223, 185, 5, 5, 243, 170, 5, 5, 4, 191, 5, 5, 80, 172, 5, 5, 132, 176, 5, 5, 116, 246, 5, 5, 94, 169, 5, 5, 122, 164, 5, 5, 81, 219, 5, 5, 126, 229, 5, 5, 127, 229, 5, 5, 82, 172, 5, 5, 130, 221, 5, 5, 88, 230, 5, 5, 0, 133, 5, 5, 245, 168, 5, 5, 76, 178, 5, 5, 77, 178, 5, 5, 129, 175, 5, 5, 28, 187, 5, 5, 252, 193, 5, 5, 203, 201, 5, 5, 122, 204, 5, 5, 224, 164, 5, 5, 8, 215, 5, 5, 87, 225, 5, 5, 208, 233, 5, 5, 119, 227, 5, 5, 227, 230, 5, 5, 219, 231, 5, 5, 53, 212, 5, 5, 60, 210, 5, 5, 162, 247, 5, 5, 105, 192, 5, 5, 248, 244, 5, 5, 126, 223, 5, 5, 42, 228, 5, 5, 213, 217, 5, 5, 8, 173, 5, 5, 89, 245, 5, 5, 168, 231, 5, 5, 249, 186, 5, 5, 163, 216, 5, 5, 9, 234, 5, 5, 23, 235, 5, 5, 120, 230, 5, 5, 139, 165, 5, 5, 2, 182, 5, 5, 225, 176, 5, 5, 44, 221, 5, 5, 187, 170, 5, 5, 82, 223, 5, 5, 18, 240, 5, 5, 236, 191, 5, 5, 192, 194, 5, 5, 32, 178, 5, 5, 167, 186, 5, 5, 2, 209, 5, 5, 129, 229, 5, 5, 181, 192, 5, 5, 20, 181, 5, 5, 73, 206, 5, 5, 130, 228, 5, 5, 251, 246, 5, 5, 34, 185, 5, 5, 186, 204, 5, 5, 188, 170, 5, 5, 17, 244, 5, 5, 43, 169, 5, 5, 155, 171, 5, 5, 165, 235, 5, 5, 33, 249, 5, 5, 174, 198, 5, 5, 195, 172, 5, 5, 182, 163, 5, 5, 139, 203, 5, 5, 227, 178, 5, 5, 24, 192, 5, 5, 213, 241, 5, 5, 43, 215, 5, 5, 187, 225, 5, 5, 78, 226, 5, 5, 158, 166, 5, 5, 65, 229, 5, 5, 135, 199, 5, 5, 247, 215, 5, 5, 55, 224, 5, 5, 213, 221, 5, 5, 56, 224, 5, 5, 89, 237, 5, 5, 87, 164, 5, 5, 90, 237, 5, 5, 48, 202, 5, 5, 191, 245, 5, 5, 211, 205, 5, 5, 80, 192, 5, 5, 209, 201, 5, 5, 21, 170, 5, 5, 238, 167, 5, 5, 4, 165, 5, 5, 91, 247, 5, 5, 19, 239, 5, 5, 150, 173, 5, 5, 97, 173, 5, 5, 30, 240, 5, 5, 73, 174, 5, 5, 96, 183, 5, 5, 177, 212, 5, 5, 68, 199, 5, 5, 226, 240, 5, 5, 191, 175, 5, 5, 13, 216, 5, 5, 211, 220, 5, 5, 143, 241, 5, 5, 218, 188, 5, 5, 8, 231, 5, 5, 182, 237, 5, 5, 194, 169, 5, 5, 227, 219, 5, 5, 176, 193, 5, 5, 62, 218, 5, 5, 41, 235, 5, 5, 179, 249, 5, 5, 53, 233, 5, 5, 192, 175, 5, 5, 66, 229, 5, 5, 69, 232, 5, 5, 119, 178, 5, 5, 39, 187, 5, 5, 81, 190, 5, 5, 14, 205, 5, 5, 90, 225, 5, 5, 130, 237, 5, 5, 29, 211, 5, 5, 194, 173, 5, 5, 97, 183, 5, 5, 123, 214, 5, 5, 45, 215, 5, 5, 6, 196, 5, 5, 175, 168, 5, 5, 111, 179, 5, 5, 75, 213, 5, 5, 160, 218, 5, 5, 198, 193, 5, 5, 126, 214, 5, 5, 88, 203, 5, 5, 9, 203, 5, 5, 70, 192, 5, 5, 251, 163, 5, 5, 106, 169, 5, 5, 46, 240, 5, 5, 87, 241, 5, 5, 90, 241, 5, 5, 53, 227, 5, 5, 230, 183, 5, 5, 231, 164, 5, 5, 101, 178, 5, 5, 149, 239, 5, 5, 97, 189, 5, 5, 65, 166, 5, 5, 185, 183, 5, 5, 228, 184, 5, 5, 95, 197, 5, 5, 106, 164, 5, 5, 12, 173, 5, 5, 204, 222, 5, 5, 89, 175, 5, 5, 170, 233, 5, 5, 62, 244, 5, 5, 26, 174, 5, 5, 158, 163, 5, 5, 101, 170, 5, 5, 115, 192, 5, 5, 117, 208, 5, 5, 77, 230, 5, 5, 34, 230, 5, 5, 60, 170, 5, 5, 246, 227, 5, 5, 92, 237, 5, 5, 173, 219, 5, 5, 188, 213, 5, 5, 105, 242, 5, 5, 145, 228, 5, 5, 34, 194, 5, 5, 64, 182, 5, 5, 109, 219, 5, 5, 100, 183, 5, 5, 112, 184, 5, 5, 3, 250, 5, 5, 209, 176, 5, 5, 198, 218, 5, 5, 64, 219, 5, 5, 144, 223, 5, 5, 27, 176, 5, 5, 6, 193, 5, 5, 130, 166, 5, 5, 161, 204, 5, 5, 98, 238, 5, 5, 82, 174, 5, 5, 16, 205, 5, 5, 0, 139, 5, 5, 102, 178, 5, 5, 253, 194, 5, 5, 23, 183, 5, 5, 220, 221, 5, 5, 8, 165, 5, 5, 84, 188, 5, 5, 210, 223, 5, 5, 47, 228, 5, 5, 172, 235, 5, 5, 208, 225, 5, 5, 230, 166, 5, 5, 43, 227, 5, 5, 199, 193, 5, 5, 65, 219, 5, 5, 216, 241, 5, 5, 54, 225, 5, 5, 162, 204, 5, 5, 54, 227, 5, 5, 68, 218, 5, 5, 85, 188, 5, 5, 196, 169, 5, 5, 170, 187, 5, 5, 50, 231, 5, 5, 110, 219, 5, 5, 95, 232, 5, 5, 140, 165, 5, 5, 135, 220, 5, 5, 199, 218, 5, 5, 230, 190, 5, 5, 11, 179, 5, 5, 190, 189, 5, 5, 16, 229, 5, 5, 47, 187, 5, 5, 19, 186, 5, 5, 254, 249, 5, 5, 135, 229, 5, 5, 222, 217, 5, 5, 224, 229, 5, 5, 38, 209, 5, 5, 167, 220, 5, 5, 120, 208, 5, 5, 57, 187, 5, 5, 156, 239, 5, 5, 70, 216, 5, 5, 243, 220, 5, 5, 58, 187, 5, 5, 224, 235, 5, 5, 106, 193, 5, 5, 162, 195, 5, 5, 118, 164, 5, 5, 98, 230, 5, 5, 217, 214, 5, 5, 159, 163, 5, 5, 99, 230, 5, 5, 88, 166, 5, 5, 112, 231, 5, 5, 20, 224, 5, 5, 60, 240, 5, 5, 215, 233, 5, 5, 230, 216, 5, 5, 136, 221, 5, 5, 87, 205, 5, 5, 36, 233, 5, 5, 33, 239, 5, 5, 57, 239, 5, 5, 76, 242, 5, 5, 128, 172, 5, 5, 223, 194, 5, 5, 122, 195, 5, 5, 199, 185, 5, 5, 79, 220, 5, 5, 189, 187, 5, 5, 120, 243, 5, 5, 82, 244, 5, 5, 106, 197, 5, 5, 169, 192, 5, 5, 61, 240, 5, 5, 58, 227, 5, 5, 168, 191, 5, 5, 33, 232, 5, 5, 161, 190, 5, 5, 94, 174, 5, 5, 147, 203, 5, 5, 201, 198, 5, 5, 37, 194, 5, 5, 178, 179, 5, 5, 163, 194, 5, 5, 13, 246, 5, 5, 3, 197, 5, 5, 185, 230, 5, 5, 241, 238, 5, 5, 201, 238, 5, 5, 245, 207, 5, 5, 157, 239, 5, 5, 10, 222, 5, 5, 10, 196, 5, 5, 209, 218, 5, 5, 135, 212, 5, 5, 171, 182, 5, 5, 115, 236, 5, 5, 114, 236, 5, 5, 172, 200, 5, 5, 239, 206, 5, 5, 165, 204, 5, 5, 83, 217, 5, 5, 106, 178, 5, 5, 90, 222, 5, 5, 180, 247, 5, 5, 131, 232, 192, 0, 0, 0, 5, 5, 95, 244, 5, 5, 235, 203, 5, 5, 114, 188, 5, 5, 136, 220, 5, 5, 229, 240, 5, 5, 19, 230, 5, 5, 168, 220, 5, 5, 98, 176, 5, 5, 121, 208, 5, 5, 159, 209, 5, 5, 138, 199, 5, 5, 38, 194, 5, 5, 37, 206, 5, 5, 173, 222, 5, 5, 223, 217, 5, 5, 203, 236, 5, 5, 130, 227, 5, 5, 174, 220, 5, 5, 185, 163, 5, 5, 133, 181, 5, 5, 56, 196, 5, 5, 158, 206, 5, 5, 207, 166, 5, 5, 132, 164, 5, 5, 102, 174, 5, 5, 250, 230, 5, 5, 211, 182, 5, 5, 144, 207, 5, 5, 202, 212, 5, 5, 110, 186, 5, 5, 87, 188, 5, 5, 205, 173, 5, 5, 125, 208, 5, 5, 54, 247, 5, 5, 24, 208, 5, 5, 138, 192, 5, 5, 150, 236, 5, 5, 43, 189, 5, 5, 15, 210, 5, 5, 170, 181, 5, 5, 78, 242, 5, 5, 201, 169, 5, 5, 145, 164, 5, 5, 42, 191, 5, 5, 140, 167, 5, 5, 175, 222, 5, 5, 235, 176, 5, 5, 98, 217, 5, 5, 206, 181, 5, 5, 82, 232, 5, 5, 99, 238, 5, 5, 86, 193, 5, 5, 23, 165, 5, 5, 171, 176, 5, 5, 87, 198, 5, 5, 180, 231, 5, 5, 80, 218, 5, 5, 134, 230, 5, 5, 218, 163, 5, 5, 238, 192, 5, 5, 24, 205, 5, 5, 219, 200, 5, 5, 211, 246, 5, 5, 156, 184, 5, 5, 19, 203, 5, 5, 25, 181, 5, 5, 185, 202, 5, 5, 137, 214, 5, 5, 67, 191, 5, 5, 6, 201, 5, 5, 230, 227, 5, 5, 109, 234, 5, 5, 103, 173, 5, 5, 97, 237, 5, 5, 22, 171, 5, 5, 24, 245, 5, 5, 173, 186, 5, 5, 174, 186, 5, 5, 203, 175, 5, 5, 3, 211, 5, 5, 208, 244, 5, 5, 204, 236, 5, 5, 210, 206, 5, 5, 153, 180, 5, 5, 67, 214, 5, 5, 229, 196, 5, 5, 38, 222, 5, 5, 75, 229, 5, 5, 70, 224, 5, 5, 57, 173, 5, 5, 66, 246, 5, 5, 159, 206, 5, 5, 175, 186, 5, 5, 11, 166, 5, 5, 240, 213, 5, 5, 150, 172, 5, 5, 181, 179, 5, 5, 222, 218, 5, 5, 45, 185, 5, 5, 120, 188, 5, 5, 25, 205, 5, 5, 11, 222, 5, 5, 110, 189, 5, 5, 182, 186, 5, 5, 44, 209, 5, 5, 94, 166, 5, 5, 206, 190, 5, 5, 46, 235, 5, 5, 175, 210, 5, 5, 179, 171, 5, 5, 195, 208, 5, 5, 196, 208, 5, 5, 27, 234, 5, 5, 147, 194, 5, 5, 80, 224, 5, 5, 61, 175, 5, 5, 186, 172, 5, 5, 213, 166, 5, 5, 162, 164, 5, 5, 163, 164, 5, 5, 252, 200, 5, 5, 95, 233, 5, 5, 113, 225, 5, 5, 71, 232, 5, 5, 75, 202, 5, 5, 81, 187, 5, 5, 158, 167, 5, 5, 88, 248, 5, 5, 89, 248, 5, 5, 88, 193, 5, 5, 58, 208, 5, 5, 20, 179, 5, 5, 73, 180, 5, 5, 196, 192, 5, 5, 96, 247, 5, 5, 245, 176, 5, 5, 202, 234, 5, 5, 16, 173, 5, 5, 61, 225, 5, 5, 140, 220, 5, 5, 15, 209, 5, 5, 7, 249, 5, 5, 232, 242, 5, 5, 90, 248, 5, 5, 48, 212, 5, 5, 90, 240, 5, 5, 110, 236, 5, 5, 103, 230, 5, 5, 32, 226, 5, 5, 152, 186, 5, 5, 130, 208, 5, 5, 13, 222, 5, 5, 115, 174, 5, 5, 9, 195, 5, 5, 50, 185, 5, 5, 101, 232, 5, 5, 102, 232, 5, 5, 235, 219, 5, 5, 71, 194, 5, 5, 84, 173, 5, 5, 104, 173, 5, 5, 160, 228, 5, 5, 115, 211, 5, 5, 163, 183, 5, 5, 215, 246, 5, 5, 181, 193, 5, 5, 20, 190, 5, 5, 219, 197, 5, 5, 109, 203, 5, 5, 232, 239, 5, 5, 24, 224, 5, 5, 24, 207, 5, 5, 245, 206, 5, 5, 114, 225, 5, 5, 245, 232, 5, 5, 45, 189, 5, 5, 35, 211, 5, 5, 114, 218, 5, 5, 33, 226, 5, 5, 233, 240, 5, 5, 213, 206, 5, 5, 188, 163, 5, 5, 248, 203, 5, 5, 42, 222, 5, 5, 125, 231, 5, 5, 14, 166, 5, 5, 159, 167, 5, 5, 219, 177, 5, 5, 190, 225, 5, 5, 50, 225, 5, 5, 180, 170, 5, 5, 159, 199, 5, 5, 29, 205, 5, 5, 226, 193, 5, 5, 128, 213, 5, 5, 244, 177, 5, 5, 176, 210, 5, 5, 179, 175, 5, 5, 39, 194, 5, 5, 186, 177, 5, 5, 92, 190, 5, 5, 82, 218, 5, 5, 113, 212, 5, 5, 55, 172, 5, 5, 253, 174, 5, 5, 252, 192, 5, 5, 192, 167, 5, 5, 114, 220, 5, 5, 216, 240, 5, 5, 57, 182, 5, 5, 111, 236, 5, 5, 85, 178, 5, 5, 56, 241, 5, 5, 223, 174, 5, 5, 77, 246, 5, 5, 239, 245, 5, 5, 168, 178, 5, 5, 79, 203, 5, 5, 99, 233, 5, 5, 218, 204, 5, 5, 178, 239, 5, 5, 217, 228, 5, 5, 52, 213, 5, 5, 49, 243, 5, 5, 238, 242, 5, 5, 59, 226, 5, 5, 219, 204, 5, 5, 84, 226, 5, 5, 27, 248, 5, 5, 123, 197, 5, 5, 71, 219, 5, 5, 202, 245, 5, 5, 110, 168, 5, 5, 210, 164, 5, 5, 148, 241, 5, 5, 203, 199, 5, 5, 202, 216, 5, 5, 240, 230, 5, 5, 43, 244, 5, 5, 243, 184, 5, 5, 183, 242, 5, 5, 127, 164, 5, 5, 20, 200, 5, 5, 126, 196, 5, 5, 225, 203, 5, 5, 127, 196, 5, 5, 119, 235, 5, 5, 210, 173, 5, 5, 246, 213, 5, 5, 58, 182, 5, 5, 223, 218, 5, 5, 104, 205, 5, 5, 96, 172, 5, 5, 72, 170, 5, 5, 192, 225, 5, 5, 76, 199, 5, 5, 12, 199, 5, 5, 197, 188, 5, 5, 153, 186, 5, 5, 152, 181, 5, 5, 107, 183, 5, 5, 209, 247, 5, 5, 54, 242, 5, 5, 213, 169, 5, 5, 116, 237, 5, 5, 171, 217, 5, 5, 172, 217, 5, 5, 70, 244, 5, 5, 182, 220, 5, 5, 111, 168, 5, 5, 99, 241, 5, 5, 82, 202, 5, 5, 166, 203, 5, 5, 100, 241, 5, 5, 191, 225, 5, 5, 214, 178, 5, 5, 198, 247, 5, 5, 119, 232, 5, 5, 240, 245, 5, 5, 214, 169, 5, 5, 229, 180, 5, 5, 83, 202, 5, 5, 167, 203, 5, 5, 95, 195, 5, 5, 225, 209, 5, 5, 170, 244, 5, 5, 53, 230, 5, 5, 51, 201, 5, 5, 233, 185, 5, 5, 24, 185, 5, 5, 114, 212, 5, 5, 117, 216, 5, 5, 152, 168, 5, 5, 113, 227, 5, 5, 243, 191, 5, 5, 250, 179, 5, 5, 68, 186, 5, 5, 3, 247, 5, 5, 210, 247, 5, 5, 226, 224, 5, 5, 49, 191, 5, 5, 25, 207, 5, 5, 226, 228, 5, 5, 238, 225, 5, 5, 57, 168, 5, 5, 82, 214, 5, 5, 82, 203, 5, 5, 122, 240, 5, 5, 121, 220, 5, 5, 130, 174, 5, 5, 169, 184, 5, 5, 108, 171, 5, 5, 118, 186, 5, 5, 71, 231, 5, 5, 60, 193, 5, 5, 114, 205, 5, 5, 26, 207, 5, 5, 39, 179, 5, 5, 226, 241, 5, 5, 188, 220, 5, 5, 154, 164, 5, 5, 140, 227, 5, 5, 60, 249, 5, 5, 187, 231, 5, 5, 41, 172, 5, 5, 172, 236, 5, 5, 92, 203, 5, 5, 183, 241, 5, 5, 177, 203, 5, 5, 101, 201, 5, 5, 163, 192, 5, 5, 187, 172, 5, 5, 118, 219, 5, 5, 133, 210, 5, 5, 52, 169, 5, 5, 83, 237, 5, 5, 61, 249, 5, 5, 83, 214, 5, 5, 72, 231, 5, 5, 254, 226, 5, 5, 88, 218, 5, 5, 219, 234, 5, 5, 149, 249, 5, 5, 156, 181, 5, 5, 49, 188, 5, 5, 73, 231, 5, 5, 80, 180, 5, 5, 29, 212, 5, 5, 190, 219, 5, 5, 125, 235, 5, 5, 52, 229, 5, 5, 65, 179, 5, 5, 19, 195, 5, 5, 235, 212, 5, 5, 61, 193, 5, 5, 138, 208, 5, 5, 29, 192, 5, 5, 30, 192, 5, 5, 74, 237, 5, 5, 64, 167, 5, 5, 160, 179, 5, 5, 182, 191, 5, 5, 236, 217, 5, 5, 131, 177, 5, 5, 87, 183, 5, 5, 87, 202, 5, 5, 187, 229, 5, 5, 95, 192, 5, 5, 39, 226, 5, 5, 42, 211, 5, 5, 153, 223, 5, 5, 156, 210, 5, 5, 234, 199, 5, 5, 219, 169, 5, 5, 24, 217, 5, 5, 209, 192, 5, 5, 10, 192, 5, 5, 50, 222, 5, 5, 39, 182, 5, 5, 210, 209, 5, 5, 85, 244, 5, 5, 132, 177, 5, 5, 83, 203, 5, 5, 180, 195, 5, 5, 45, 233, 5, 5, 180, 175, 5, 5, 45, 191, 5, 5, 18, 166, 5, 5, 149, 221, 5, 5, 125, 236, 5, 5, 60, 213, 5, 5, 245, 201, 5, 5, 141, 188, 5, 5, 108, 200, 5, 5, 223, 249, 5, 5, 185, 210, 5, 5, 187, 193, 5, 5, 250, 191, 5, 5, 251, 191, 5, 5, 116, 182, 5, 5, 149, 199, 5, 5, 142, 208, 5, 5, 112, 236, 5, 5, 221, 234, 5, 5, 165, 167, 5, 5, 88, 242, 5, 5, 19, 199, 5, 5, 42, 206, 5, 5, 84, 199, 5, 5, 237, 200, 5, 5, 50, 194, 5, 5, 248, 183, 5, 5, 131, 240, 5, 5, 13, 192, 5, 5, 219, 195, 5, 5, 236, 193, 5, 5, 237, 193, 5, 5, 39, 234, 5, 5, 146, 170, 5, 5, 200, 205, 5, 5, 158, 176, 5, 5, 181, 222, 5, 5, 26, 185, 5, 5, 134, 186, 5, 5, 202, 224, 5, 5, 54, 235, 5, 5, 202, 186, 5, 5, 253, 214, 5, 5, 119, 211, 5, 5, 255, 229, 5, 5, 21, 173, 5, 5, 215, 182, 5, 5, 111, 194, 5, 5, 2, 232, 5, 5, 178, 236, 5, 5, 124, 185, 5, 5, 237, 189, 5, 5, 52, 189, 5, 5, 113, 171, 5, 5, 239, 185, 5, 5, 245, 223, 5, 5, 179, 236, 5, 5, 22, 195, 5, 5, 151, 225, 5, 5, 87, 180, 5, 5, 155, 165, 5, 5, 44, 190, 5, 5, 138, 204, 5, 5, 177, 219, 5, 5, 75, 219, 5, 5, 90, 229, 5, 5, 23, 221, 5, 5, 113, 170, 5, 5, 74, 169, 5, 5, 36, 184, 5, 5, 87, 244, 5, 5, 33, 181, 5, 5, 166, 167, 5, 5, 246, 221, 5, 5, 141, 227, 5, 5, 5, 213, 5, 5, 83, 212, 5, 5, 205, 198, 5, 5, 231, 204, 5, 5, 55, 235, 5, 5, 251, 202, 5, 5, 105, 243, 5, 5, 62, 211, 5, 5, 89, 217, 5, 5, 117, 244, 5, 5, 80, 239, 5, 5, 244, 217, 5, 5, 131, 218, 5, 5, 172, 209, 5, 5, 28, 233, 5, 5, 91, 235, 5, 5, 253, 168, 5, 5, 159, 229, 5, 5, 151, 199, 5, 5, 171, 196, 5, 5, 199, 220, 5, 5, 71, 233, 5, 5, 228, 248, 5, 5, 108, 228, 5, 5, 6, 213, 5, 5, 88, 226, 5, 5, 178, 177, 5, 5, 136, 172, 5, 5, 146, 192, 5, 5, 76, 169, 5, 5, 74, 245, 5, 5, 146, 209, 5, 5, 22, 230, 5, 5, 29, 189, 5, 5, 61, 188, 5, 5, 173, 209, 5, 5, 225, 215, 5, 5, 45, 231, 5, 5, 22, 173, 5, 5, 125, 220, 5, 5, 45, 182, 5, 5, 205, 185, 5, 5, 154, 249, 5, 5, 231, 229, 5, 5, 192, 178, 5, 5, 74, 225, 5, 5, 93, 180, 5, 5, 233, 170, 5, 5, 5, 207, 5, 5, 131, 225, 5, 5, 189, 236, 5, 5, 179, 207, 5, 5, 180, 207, 5, 5, 179, 177, 5, 5, 116, 200, 5, 5, 254, 201, 5, 5, 159, 184, 5, 5, 110, 216, 5, 5, 60, 182, 5, 5, 134, 184, 5, 5, 255, 230, 5, 5, 152, 221, 5, 5, 198, 182, 5, 5, 199, 232, 5, 5, 112, 173, 5, 5, 156, 212, 5, 5, 81, 230, 5, 5, 206, 198, 5, 5, 139, 210, 5, 5, 25, 193, 5, 5, 96, 188, 5, 5, 215, 185, 5, 5, 14, 192, 5, 5, 175, 225, 5, 5, 165, 213, 5, 5, 57, 229, 5, 5, 206, 231, 5, 5, 109, 236, 5, 5, 101, 202, 5, 5, 187, 175, 5, 5, 188, 206, 5, 5, 242, 189, 5, 5, 156, 205, 5, 5, 239, 225, 5, 5, 70, 249, 5, 5, 77, 179, 5, 5, 94, 229, 5, 5, 24, 206, 5, 5, 11, 180, 5, 5, 17, 189, 5, 5, 198, 184, 5, 5, 224, 243, 5, 5, 62, 213, 5, 5, 40, 237, 5, 5, 67, 228, 5, 5, 51, 233, 5, 5, 145, 217, 5, 5, 78, 225, 5, 5, 164, 192, 5, 5, 6, 207, 5, 5, 52, 233, 5, 5, 167, 240, 5, 5, 250, 210, 5, 5, 255, 168, 5, 5, 195, 242, 5, 5, 94, 185, 5, 5, 220, 173, 5, 5, 61, 200, 5, 5, 114, 176, 5, 5, 222, 202, 5, 5, 41, 237, 5, 5, 144, 163, 5, 5, 206, 168, 5, 5, 132, 225, 5, 5, 196, 225, 5, 5, 255, 221, 5, 5, 53, 194, 5, 5, 107, 202, 5, 5, 202, 203, 5, 5, 85, 203, 5, 5, 123, 211, 5, 5, 195, 219, 5, 5, 52, 205, 5, 5, 68, 211, 5, 5, 170, 175, 5, 5, 242, 239, 5, 5, 187, 165, 5, 5, 106, 237, 5, 5, 176, 227, 5, 5, 18, 189, 5, 5, 171, 187, 5, 5, 156, 223, 5, 5, 244, 170, 5, 5, 196, 243, 5, 5, 197, 179, 5, 5, 129, 179, 5, 5, 185, 171, 5, 5, 228, 208, 5, 5, 118, 182, 5, 5, 29, 166, 5, 5, 213, 199, 5, 5, 227, 202, 5, 5, 100, 210, 5, 5, 196, 183, 5, 5, 127, 184, 5, 5, 119, 239, 5, 5, 36, 192, 5, 5, 182, 195, 5, 5, 251, 243, 5, 5, 90, 183, 5, 5, 134, 211, 5, 5, 252, 242, 5, 5, 229, 208, 5, 5, 120, 171, 5, 5, 247, 219, 5, 5, 56, 194, 5, 5, 91, 226, 5, 5, 14, 202, 5, 5, 138, 196, 5, 5, 185, 166, 5, 5, 144, 176, 5, 5, 178, 225, 5, 5, 211, 191, 5, 5, 186, 171, 5, 5, 23, 242, 5, 5, 42, 219, 5, 5, 97, 201, 5, 5, 102, 210, 5, 5, 134, 185, 5, 5, 98, 185, 5, 5, 217, 235, 5, 5, 248, 189, 5, 5, 125, 187, 5, 5, 12, 235, 5, 5, 104, 191, 5, 5, 140, 171, 5, 5, 136, 200, 5, 5, 178, 238, 5, 5, 205, 189, 5, 5, 33, 219, 5, 5, 18, 192, 5, 5, 4, 178, 5, 5, 78, 249, 5, 5, 126, 187, 5, 5, 222, 172, 5, 5, 66, 222, 5, 5, 73, 236, 5, 5, 161, 188, 5, 5, 74, 206, 5, 5, 180, 225, 5, 5, 148, 211, 5, 5, 161, 221, 5, 5, 210, 189, 5, 5, 202, 242, 5, 5, 133, 187, 5, 5, 160, 221, 5, 5, 233, 224, 5, 5, 192, 203, 5, 5, 10, 238, 5, 5, 135, 236, 5, 5, 243, 199, 5, 5, 11, 184, 5, 5, 164, 248, 5, 5, 6, 191, 5, 5, 80, 185, 5, 5, 117, 204, 5, 5, 169, 215, 5, 5, 249, 193, 5, 5, 35, 207, 5, 5, 179, 214, 5, 5, 157, 189, 5, 5, 142, 244, 5, 5, 3, 205, 5, 5, 229, 243, 5, 5, 66, 242, 5, 5, 202, 244, 5, 5, 61, 216, 5, 5, 206, 184, 5, 5, 136, 163, 5, 5, 112, 224, 5, 5, 50, 186, 5, 5, 36, 207, 5, 5, 243, 229, 5, 5, 213, 227, 5, 5, 45, 246, 5, 5, 95, 235, 5, 5, 94, 212, 5, 5, 226, 211, 5, 5, 132, 234, 5, 5, 187, 204, 5, 5, 169, 248, 5, 5, 111, 190, 5, 5, 159, 187, 5, 5, 14, 235, 5, 5, 174, 190, 5, 5, 90, 176, 5, 5, 207, 207, 5, 5, 108, 243, 5, 5, 86, 186, 5, 5, 60, 186, 5, 5, 185, 221, 5, 5, 112, 185, 5, 5, 21, 187, 5, 5, 32, 163, 5, 5, 134, 183, 5, 5, 175, 248, 5, 5, 162, 189, 5, 5, 185, 194, 5, 5, 164, 210, 5, 5, 209, 233, 5, 5, 153, 200, 5, 5, 248, 215, 5, 5, 163, 247, 5, 5, 31, 240, 5, 5, 64, 210, 5, 5, 165, 238, 5, 5, 185, 216, 5, 5, 227, 223, 5, 5, 209, 225, 5, 5, 39, 209, 5, 5, 109, 195, 5, 5, 62, 240, 5, 5, 139, 207, 5, 5, 121, 213, 5, 5, 2, 217, 5, 5, 78, 197, 5, 5, 76, 248, 5, 5, 154, 196, 5, 5, 105, 186, 5, 5, 189, 196, 5, 5, 183, 229, 5, 5, 203, 212, 5, 5, 27, 238, 5, 5, 104, 239, 5, 5, 131, 208, 5, 5, 242, 225, 5, 5, 71, 223, 5, 5, 207, 210, 5, 5, 190, 196, 5, 5, 177, 195, 5, 5, 191, 196, 5, 5, 179, 195, 5, 5, 67, 226, 5, 5, 160, 227, 5, 5, 111, 193, 5, 5, 139, 243, 5, 5, 5, 216, 5, 5, 247, 213, 5, 5, 211, 247, 5, 5, 182, 210, 5, 5, 154, 214, 5, 5, 105, 241, 5, 5, 151, 191, 5, 5, 5, 245, 5, 5, 47, 180, 5, 5, 239, 205, 5, 5, 171, 238, 5, 5, 242, 198, 5, 5, 231, 180, 5, 5, 155, 200, 5, 5, 213, 216, 5, 5, 98, 167, 5, 5, 91, 213, 5, 5, 228, 214, 5, 5, 243, 225, 5, 5, 204, 227, 5, 5, 109, 228, 5, 5, 71, 198, 5, 5, 233, 225, 5, 5, 168, 240, 5, 5, 7, 238, 5, 5, 242, 247, 5, 5, 15, 226, 5, 5, 163, 200, 5, 5, 141, 218, 5, 5, 112, 221, 5, 5, 201, 233, 5, 5, 122, 166, 5, 5, 53, 245, 5, 5, 164, 208, 5, 5, 127, 166, 5, 5, 128, 166, 5, 5, 214, 245, 5, 5, 53, 207, 5, 5, 121, 210, 5, 5, 39, 223, 5, 5, 228, 235, 5, 5, 72, 223, 5, 5, 233, 235, 5, 5, 232, 222, 5, 5, 28, 237, 5, 5, 29, 237, 5, 5, 96, 187, 5, 5, 207, 234, 5, 5, 223, 247, 5, 5, 105, 187, 5, 5, 244, 193, 5, 5, 73, 175, 5, 5, 190, 237, 5, 5, 154, 235, 5, 5, 193, 204, 5, 5, 128, 219, 5, 5, 99, 169, 5, 5, 100, 169, 5, 5, 144, 241, 5, 5, 101, 169, 5, 5, 150, 234, 5, 5, 5, 244, 5, 5, 98, 181, 5, 5, 17, 229, 5, 5, 90, 175, 5, 5, 229, 228, 5, 5, 3, 217, 5, 5, 7, 218, 5, 5, 72, 229, 5, 5, 73, 234, 5, 5, 145, 207, 5, 5, 84, 168, 5, 5, 84, 209, 5, 5, 16, 210, 5, 5, 151, 236, 5, 5, 113, 211, 5, 5, 165, 249, 5, 5, 189, 188, 5, 5, 91, 240, 5, 5, 235, 172, 5, 5, 115, 218, 5, 5, 97, 241, 5, 5, 76, 234, 5, 5, 253, 200, 5, 5, 240, 175, 5, 5, 84, 202, 5, 5, 48, 177, 5, 5, 189, 220, 5, 5, 171, 249, 5, 5, 6, 247, 5, 5, 65, 224, 5, 5, 37, 166, 5, 5, 232, 180, 5, 5, 240, 205, 5, 5, 237, 228, 5, 5, 191, 243, 5, 5, 85, 177, 5, 5, 177, 204, 5, 5, 117, 200, 5, 5, 17, 235, 5, 5, 153, 240, 5, 5, 250, 175, 5, 5, 44, 178, 5, 5, 25, 206, 5, 5, 205, 188, 5, 5, 42, 166, 5, 5, 5, 178, 5, 5, 158, 205, 5, 5, 4, 205, 5, 5, 59, 229, 5, 5, 166, 238, 5, 5, 164, 178, 5, 5, 221, 221, 5, 5, 121, 243, 5, 5, 138, 211, 5, 5, 73, 240, 5, 5, 131, 227, 5, 5, 196, 209, 5, 5, 35, 195, 5, 5, 196, 214, 5, 5, 15, 218, 5, 5, 173, 177, 5, 5, 57, 200, 5, 5, 58, 200, 5, 5, 54, 193, 5, 5, 136, 209, 5, 5, 241, 201, 5, 5, 132, 240, 5, 5, 172, 238, 5, 5, 60, 200, 5, 5, 176, 178, 5, 5, 120, 239, 5, 5, 187, 214, 5, 5, 13, 245, 5, 5, 95, 194, 5, 5, 61, 172, 5, 5, 234, 169, 5, 5, 235, 169, 5, 5, 44, 238, 5, 5, 197, 203, 5, 5, 239, 177, 5, 5, 101, 248, 5, 5, 221, 164, 5, 5, 246, 164, 5, 5, 9, 165, 5, 5, 243, 167, 5, 5, 244, 167, 5, 5, 13, 165, 5, 5, 99, 224, 5, 5, 19, 170, 5, 5, 216, 197, 5, 5, 15, 234, 5, 5, 239, 167, 5, 5, 217, 187, 5, 5, 10, 203, 5, 5, 77, 220, 5, 5, 253, 235, 5, 5, 89, 172, 5, 5, 65, 182, 5, 5, 224, 217, 5, 5, 58, 239, 5, 5, 55, 231, 5, 5, 143, 215, 5, 5, 240, 206, 5, 5, 58, 172, 5, 5, 218, 197, 5, 5, 17, 210, 5, 5, 204, 175, 5, 5, 132, 227, 5, 5, 27, 180, 5, 5, 237, 166, 5, 5, 246, 206, 5, 5, 58, 178, 5, 5, 242, 234, 5, 5, 209, 164, 5, 5, 82, 236, 5, 5, 107, 236, 5, 5, 73, 170, 5, 5, 3, 236, 5, 5, 85, 229, 5, 5, 150, 247, 5, 5, 69, 182, 5, 5, 124, 197, 5, 5, 105, 205, 5, 5, 66, 179, 5, 5, 84, 236, 5, 5, 51, 191, 5, 5, 60, 172, 5, 5, 36, 170, 5, 5, 199, 182, 5, 5, 101, 231, 5, 5, 35, 225, 5, 5, 254, 168, 5, 5, 84, 190, 5, 5, 92, 220, 5, 5, 36, 236, 5, 5, 93, 220, 5, 5, 245, 170, 5, 5, 249, 189, 5, 5, 227, 211, 5, 5, 66, 177, 5, 5, 23, 227, 5, 5, 147, 176, 5, 5, 222, 171, 5, 5, 120, 207, 5, 5, 190, 170, 5, 5, 21, 209, 5, 5, 58, 168, 5, 5, 13, 183, 5, 5, 24, 209, 5, 5, 130, 175, 5, 5, 42, 215, 5, 5, 47, 226, 5, 5, 74, 173, 5, 5, 24, 244, 5, 5, 43, 183, 5, 5, 48, 166, 5, 5, 83, 174, 5, 5, 146, 228, 5, 5, 229, 178, 5, 5, 91, 211, 5, 5, 40, 209, 5, 5, 122, 243, 5, 5, 192, 212, 5, 5, 198, 197, 5, 5, 64, 244, 5, 5, 202, 238, 5, 5, 33, 241, 5, 5, 62, 195, 5, 5, 98, 211, 5, 5, 39, 241, 5, 5, 240, 178, 5, 5, 37, 176, 5, 5, 93, 191, 5, 5, 116, 211, 5, 5, 71, 244, 5, 5, 117, 249, 5, 5, 118, 249, 5, 5, 208, 229, 5, 5, 218, 229, 5, 5, 103, 221, 5, 5, 53, 166, 5, 5, 247, 249, 5, 5, 92, 176, 5, 5, 219, 229, 5, 5, 17, 185, 5, 5, 200, 183, 5, 5, 152, 182, 5, 5, 45, 173, 5, 5, 251, 193, 5, 5, 126, 181, 5, 5, 57, 218, 5, 5, 251, 209, 5, 5, 5, 166, 5, 5, 22, 232, 5, 5, 191, 171, 5, 5, 4, 235, 5, 5, 65, 200, 5, 5, 74, 177, 5, 5, 114, 198, 5, 5, 163, 223, 5, 5, 21, 222, 5, 5, 70, 226, 5, 5, 186, 246, 5, 5, 48, 173, 5, 5, 48, 175, 5, 5, 168, 212, 5, 5, 254, 181, 5, 5, 184, 170, 5, 5, 249, 209, 5, 5, 25, 209, 5, 5, 86, 228, 5, 5, 154, 168, 5, 5, 116, 198, 5, 5, 56, 210, 5, 5, 243, 244, 5, 5, 244, 244, 5, 5, 255, 217, 5, 5, 159, 235, 5, 5, 211, 230, 5, 5, 218, 194, 5, 5, 7, 206, 5, 5, 31, 223, 5, 5, 245, 244, 5, 5, 247, 181, 5, 5, 185, 204, 5, 5, 49, 180, 5, 5, 60, 218, 5, 5, 149, 218, 5, 5, 61, 210, 5, 5, 87, 171, 5, 5, 33, 193, 5, 5, 77, 247, 5, 5, 120, 206, 5, 5, 122, 206, 5, 5, 7, 231, 5, 5, 18, 208, 5, 5, 133, 201, 5, 5, 135, 179, 5, 5, 101, 212, 5, 5, 49, 202, 5, 5, 232, 165, 5, 5, 233, 165, 5, 5, 153, 192, 5, 5, 61, 247, 5, 5, 195, 222, 5, 5, 148, 173, 5, 5, 35, 218, 5, 5, 123, 207, 5, 5, 41, 199, 5, 5, 47, 184, 5, 5, 27, 209, 5, 5, 95, 173, 5, 5, 244, 216, 5, 5, 248, 216, 5, 5, 15, 211, 5, 5, 133, 239, 5, 5, 188, 245, 5, 5, 131, 249, 5, 5, 31, 249, 5, 5, 190, 198, 5, 5, 213, 242, 5, 5, 181, 224, 5, 5, 182, 224, 5, 5, 23, 176, 5, 5, 29, 242, 5, 5, 206, 178, 5, 5, 61, 199, 5, 5, 185, 199, 5, 5, 114, 208, 5, 5, 84, 172, 5, 5, 216, 184, 5, 5, 192, 169, 5, 5, 21, 175, 5, 5, 193, 171, 5, 5, 88, 171, 5, 5, 150, 226, 5, 5, 107, 167, 5, 5, 228, 164, 5, 5, 120, 227, 5, 5, 150, 218, 5, 5, 246, 166, 5, 5, 79, 167, 5, 5, 88, 222, 5, 5, 209, 231, 5, 5, 55, 192, 5, 5, 32, 188, 5, 5, 193, 169, 5, 5, 236, 231, 5, 5, 191, 173, 5, 5, 164, 238, 5, 5, 34, 187, 5, 5, 194, 190, 5, 5, 113, 234, 5, 5, 192, 173, 5, 5, 6, 246, 5, 5, 62, 199, 5, 5, 102, 169, 5, 5, 252, 169, 5, 5, 130, 229, 5, 5, 19, 207, 5, 5, 130, 232, 5, 5, 7, 182, 5, 5, 23, 182, 5, 5, 24, 182, 5, 5, 91, 197, 5, 5, 40, 187, 5, 5, 71, 171, 5, 5, 139, 239, 5, 5, 113, 172, 5, 5, 196, 172, 5, 5, 33, 242, 5, 5, 73, 181, 5, 5, 191, 218, 5, 5, 221, 219, 5, 5, 4, 193, 5, 5, 76, 220, 5, 5, 162, 242, 5, 5, 109, 235, 5, 5, 26, 197, 5, 5, 94, 208, 5, 5, 196, 206, 5, 5, 52, 214, 5, 5, 253, 209, 5, 5, 33, 240, 5, 5, 7, 203, 5, 5, 200, 222, 5, 5, 57, 212, 5, 5, 171, 166, 5, 5, 68, 200, 5, 5, 90, 171, 5, 5, 116, 164, 5, 5, 238, 231, 5, 5, 166, 221, 5, 5, 21, 185, 5, 5, 12, 242, 5, 5, 5, 193, 5, 5, 138, 201, 5, 5, 112, 192, 5, 5, 139, 172, 5, 5, 50, 170, 5, 5, 166, 235, 5, 5, 223, 198, 5, 5, 172, 168, 5, 5, 203, 233, 5, 5, 50, 173, 5, 5, 63, 218, 5, 5, 143, 236, 5, 5, 138, 184, 5, 5, 52, 225, 5, 5, 93, 189, 5, 5, 27, 167, 5, 5, 39, 163, 5, 5, 31, 209, 5, 5, 167, 235, 5, 5, 150, 183, 5, 5, 236, 206, 5, 5, 208, 235, 5, 5, 214, 221, 5, 5, 14, 229, 5, 5, 142, 185, 5, 5, 158, 217, 5, 5, 126, 207, 5, 5, 74, 174, 5, 5, 81, 167, 5, 5, 254, 171, 5, 5, 139, 179, 5, 5, 4, 244, 5, 5, 128, 202, 5, 5, 250, 249, 5, 5, 98, 194, 5, 5, 195, 164, 5, 5, 61, 217, 5, 5, 228, 190, 5, 5, 186, 189, 5, 5, 24, 176, 5, 5, 132, 190, 5, 5, 53, 214, 5, 5, 172, 166, 5, 5, 239, 231, 5, 5, 79, 206, 5, 5, 166, 243, 5, 5, 58, 167, 5, 5, 144, 203, 5, 5, 47, 240, 5, 5, 5, 200, 5, 5, 44, 200, 5, 5, 46, 232, 5, 5, 96, 225, 5, 5, 117, 243, 5, 5, 11, 246, 5, 5, 50, 198, 5, 5, 82, 205, 5, 5, 30, 244, 5, 5, 183, 187, 5, 5, 175, 235, 5, 5, 165, 202, 5, 5, 31, 239, 5, 5, 145, 216, 5, 5, 84, 174, 5, 5, 201, 235, 5, 5, 223, 235, 5, 5, 161, 218, 5, 5, 234, 243, 5, 5, 21, 245, 5, 5, 76, 213, 5, 5, 121, 195, 5, 5, 134, 221, 5, 5, 54, 206, 5, 5, 134, 190, 5, 5, 137, 190, 5, 5, 240, 238, 5, 5, 12, 179, 5, 5, 216, 180, 5, 5, 88, 170, 5, 5, 143, 185, 5, 5, 96, 239, 5, 5, 74, 171, 5, 5, 58, 212, 5, 5, 166, 205, 5, 5, 195, 166, 5, 5, 58, 246, 5, 5, 189, 214, 5, 5, 56, 214, 5, 5, 141, 201, 5, 5, 253, 216, 5, 5, 145, 236, 5, 5, 255, 196, 5, 5, 176, 179, 5, 5, 103, 209, 5, 5, 50, 207, 5, 5, 54, 239, 5, 5, 4, 220, 5, 5, 218, 230, 5, 5, 176, 168, 5, 5, 219, 217, 5, 5, 38, 218, 5, 5, 200, 218, 5, 5, 162, 194, 5, 5, 169, 200, 5, 5, 29, 172, 5, 5, 175, 199, 5, 5, 138, 245, 5, 5, 83, 209, 5, 5, 221, 166, 5, 5, 251, 244, 5, 5, 164, 211, 5, 5, 34, 188, 5, 5, 48, 240, 5, 5, 11, 203, 5, 5, 147, 206, 5, 5, 151, 210, 5, 5, 10, 206, 5, 5, 242, 203, 5, 5, 160, 220, 5, 5, 192, 226, 5, 5, 27, 222, 5, 5, 236, 233, 5, 5, 124, 172, 5, 5, 71, 200, 5, 5, 166, 202, 5, 5, 35, 188, 5, 5, 126, 224, 5, 5, 5, 201, 5, 5, 43, 199, 5, 5, 190, 184, 5, 5, 220, 217, 5, 5, 215, 205, 5, 5, 190, 187, 5, 5, 12, 210, 5, 5, 61, 246, 5, 5, 174, 219, 5, 5, 47, 208, 5, 5, 84, 198, 5, 5, 144, 164, 5, 5, 198, 170, 5, 5, 30, 222, 5, 5, 40, 242, 5, 5, 193, 212, 5, 5, 79, 163, 5, 5, 199, 175, 5, 5, 215, 200, 5, 5, 148, 201, 5, 5, 63, 240, 5, 5, 247, 168, 5, 5, 206, 244, 5, 5, 64, 186, 5, 5, 251, 197, 5, 5, 141, 186, 5, 5, 106, 198, 5, 5, 2, 187, 5, 5, 215, 192, 5, 5, 95, 174, 5, 5, 22, 245, 5, 5, 188, 188, 5, 5, 97, 226, 5, 5, 238, 173, 5, 5, 64, 240, 5, 5, 66, 248, 5, 5, 172, 182, 5, 5, 174, 185, 5, 5, 188, 197, 5, 5, 218, 205, 5, 5, 17, 203, 5, 5, 152, 226, 5, 5, 99, 208, 5, 5, 239, 169, 5, 5, 251, 236, 5, 5, 56, 225, 5, 5, 252, 166, 5, 5, 197, 215, 5, 5, 209, 228, 5, 5, 204, 181, 5, 5, 241, 196, 5, 5, 160, 168, 5, 5, 231, 183, 5, 5, 124, 198, 5, 5, 84, 206, 5, 5, 132, 214, 5, 5, 191, 227, 5, 5, 72, 184, 5, 5, 186, 181, 5, 5, 4, 250, 5, 5, 218, 214, 5, 5, 95, 215, 5, 5, 169, 181, 5, 5, 159, 237, 5, 5, 146, 236, 5, 5, 129, 182, 5, 5, 217, 207, 5, 5, 107, 234, 5, 5, 219, 180, 5, 5, 59, 187, 5, 5, 210, 229, 5, 5, 176, 235, 5, 5, 145, 185, 5, 5, 211, 229, 5, 5, 150, 179, 5, 5, 241, 177, 5, 5, 207, 185, 5, 5, 40, 185, 5, 5, 119, 188, 5, 5, 14, 246, 5, 5, 170, 245, 5, 5, 138, 249, 5, 5, 216, 200, 5, 5, 170, 176, 5, 5, 212, 198, 5, 5, 41, 231, 5, 5, 88, 205, 5, 5, 36, 221, 5, 5, 225, 235, 5, 5, 103, 174, 5, 5, 102, 194, 5, 5, 171, 216, 5, 5, 25, 228, 5, 5, 225, 229, 5, 5, 252, 240, 5, 5, 255, 225, 5, 5, 48, 221, 5, 5, 94, 241, 5, 5, 57, 197, 5, 5, 195, 245, 5, 5, 88, 198, 5, 5, 11, 182, 5, 5, 237, 178, 5, 5, 111, 197, 5, 5, 172, 245, 5, 5, 211, 206, 5, 5, 192, 227, 5, 5, 44, 189, 5, 5, 250, 200, 5, 5, 60, 179, 5, 5, 12, 222, 5, 5, 170, 177, 5, 5, 106, 209, 5, 5, 59, 175, 5, 5, 249, 207, 5, 5, 118, 231, 5, 5, 204, 212, 5, 5, 193, 214, 5, 5, 187, 208, 5, 5, 46, 193, 5, 5, 178, 193, 5, 5, 131, 242, 5, 5, 225, 188, 5, 5, 114, 180, 5, 5, 12, 220, 5, 5, 188, 167, 5, 5, 255, 197, 5, 5, 241, 224, 5, 5, 76, 229, 5, 5, 126, 208, 5, 5, 155, 173, 5, 5, 111, 218, 5, 5, 13, 190, 5, 5, 80, 183, 5, 5, 243, 206, 5, 5, 182, 179, 5, 5, 82, 235, 5, 5, 146, 185, 5, 5, 175, 219, 5, 5, 186, 168, 5, 5, 175, 172, 5, 5, 98, 237, 5, 5, 104, 181, 5, 5, 37, 168, 5, 5, 33, 176, 5, 5, 77, 248, 5, 5, 158, 183, 5, 5, 148, 247, 5, 5, 14, 190, 5, 5, 158, 192, 5, 5, 97, 171, 5, 5, 160, 206, 5, 5, 111, 189, 5, 5, 204, 190, 5, 5, 179, 243, 5, 5, 241, 234, 5, 5, 204, 211, 5, 5, 233, 199, 5, 5, 78, 248, 5, 5, 74, 184, 5, 5, 156, 228, 5, 5, 144, 165, 5, 5, 100, 226, 5, 5, 232, 198, 5, 5, 255, 248, 5, 5, 46, 185, 5, 5, 23, 201, 5, 5, 99, 211, 5, 5, 132, 242, 5, 5, 152, 210, 5, 5, 64, 181, 5, 5, 158, 171, 5, 5, 169, 218, 5, 5, 211, 243, 5, 5, 98, 171, 5, 5, 133, 164, 5, 5, 2, 249, 5, 5, 63, 247, 5, 5, 237, 230, 5, 5, 157, 172, 5, 5, 100, 211, 5, 5, 3, 249, 5, 5, 43, 195, 5, 5, 186, 202, 5, 5, 193, 227, 5, 5, 4, 249, 5, 5, 251, 235, 5, 5, 15, 190, 5, 5, 197, 231, 5, 5, 243, 177, 5, 5, 66, 209, 5, 5, 95, 205, 5, 5, 202, 171, 5, 5, 38, 168, 5, 5, 164, 213, 5, 5, 187, 247, 5, 5, 5, 249, 5, 5, 191, 228, 5, 5, 211, 228, 5, 5, 37, 230, 5, 5, 226, 200, 5, 5, 194, 221, 5, 5, 30, 205, 5, 5, 180, 184, 5, 5, 199, 247, 5, 5, 26, 210, 5, 5, 151, 181, 5, 5, 207, 211, 5, 5, 39, 178, 5, 5, 17, 173, 5, 5, 70, 219, 5, 5, 93, 163, 5, 5, 199, 245, 5, 5, 174, 245, 5, 5, 246, 176, 5, 5, 233, 245, 5, 5, 254, 246, 5, 5, 188, 230, 5, 5, 74, 214, 5, 5, 245, 233, 5, 5, 129, 246, 5, 5, 118, 197, 5, 5, 206, 192, 5, 5, 233, 242, 5, 5, 157, 201, 5, 5, 169, 217, 5, 5, 78, 171, 5, 5, 46, 195, 5, 5, 31, 170, 5, 5, 104, 244, 5, 5, 234, 242, 5, 5, 49, 242, 5, 5, 117, 201, 5, 5, 159, 171, 5, 5, 39, 229, 5, 5, 242, 169, 5, 5, 35, 232, 5, 5, 9, 201, 5, 5, 27, 183, 5, 5, 17, 194, 5, 5, 98, 223, 5, 5, 228, 237, 5, 5, 223, 166, 5, 5, 208, 169, 5, 5, 194, 188, 5, 5, 45, 209, 5, 5, 46, 209, 5, 5, 209, 169, 5, 5, 82, 181, 5, 5, 57, 234, 5, 5, 24, 246, 5, 5, 68, 234, 5, 5, 102, 182, 5, 5, 205, 171, 5, 5, 85, 200, 5, 5, 120, 169, 5, 5, 197, 216, 5, 5, 76, 202, 5, 5, 101, 167, 5, 5, 211, 204, 5, 5, 31, 205, 5, 5, 8, 199, 5, 5, 2, 174, 5, 5, 46, 216, 5, 5, 92, 240, 5, 5, 239, 243, 5, 5, 141, 220, 5, 5, 248, 226, 5, 5, 110, 203, 5, 5, 62, 225, 5, 5, 47, 201, 5, 5, 42, 180, 5, 5, 200, 247, 5, 5, 33, 165, 5, 5, 8, 172, 5, 5, 198, 226, 5, 5, 189, 181, 5, 5, 25, 229, 5, 5, 10, 250, 5, 5, 247, 206, 5, 5, 21, 179, 5, 5, 125, 182, 5, 5, 43, 204, 5, 5, 47, 204, 5, 5, 27, 210, 5, 5, 137, 242, 5, 5, 52, 180, 5, 5, 239, 190, 5, 5, 200, 199, 5, 5, 46, 249, 5, 5, 113, 183, 5, 5, 108, 164, 5, 5, 83, 242, 5, 5, 70, 212, 5, 5, 177, 222, 5, 5, 75, 189, 5, 5, 81, 224, 5, 5, 179, 242, 5, 5, 103, 226, 5, 5, 240, 183, 5, 5, 126, 231, 5, 5, 177, 246, 5, 5, 199, 226, 5, 5, 127, 170, 5, 5, 164, 244, 5, 5, 7, 185, 5, 5, 157, 196, 5, 5, 13, 174, 5, 5, 99, 205, 5, 5, 18, 217, 5, 5, 231, 185, 5, 5, 25, 246, 5, 5, 240, 191, 5, 5, 241, 213, 5, 5, 165, 244, 5, 5, 174, 166, 5, 5, 228, 205, 5, 5, 50, 243, 5, 5, 78, 180, 5, 5, 56, 249, 5, 5, 230, 236, 5, 5, 57, 249, 5, 5, 127, 188, 5, 5, 225, 241, 5, 5, 23, 171, 5, 5, 164, 222, 5, 5, 120, 225, 5, 5, 66, 195, 5, 5, 144, 179, 5, 5, 45, 179, 5, 5, 146, 245, 5, 5, 28, 188, 5, 5, 60, 192, 5, 5, 21, 236, 5, 5, 37, 172, 5, 5, 117, 219, 5, 5, 250, 226, 5, 5, 106, 240, 5, 5, 222, 190, 5, 5, 35, 169, 5, 5, 250, 168, 5, 5, 241, 245, 5, 5, 91, 187, 5, 5, 13, 234, 5, 5, 114, 245, 5, 5, 230, 212, 5, 5, 4, 247, 5, 5, 38, 211, 5, 5, 250, 231, 5, 5, 201, 226, 5, 5, 165, 236, 5, 5, 37, 238, 5, 5, 236, 210, 5, 5, 47, 236, 5, 5, 145, 249, 5, 5, 21, 215, 5, 5, 140, 243, 5, 5, 69, 231, 5, 5, 25, 201, 5, 5, 166, 236, 5, 5, 242, 236, 5, 5, 212, 183, 5, 5, 38, 179, 5, 5, 202, 226, 5, 5, 221, 230, 5, 5, 29, 238, 5, 5, 90, 167, 5, 5, 222, 225, 5, 5, 161, 225, 5, 5, 110, 167, 5, 5, 117, 247, 5, 5, 120, 234, 5, 5, 133, 171, 5, 5, 176, 191, 5, 5, 130, 177, 5, 5, 4, 198, 5, 5, 100, 201, 5, 5, 134, 213, 5, 5, 127, 231, 5, 5, 180, 176, 5, 5, 250, 218, 5, 5, 154, 216, 5, 5, 171, 174, 5, 5, 168, 203, 5, 5, 117, 199, 5, 5, 149, 191, 5, 5, 185, 231, 5, 5, 158, 179, 5, 5, 138, 180, 5, 5, 255, 236, 5, 5, 41, 165, 5, 5, 95, 204, 5, 5, 95, 200, 5, 5, 136, 194, 5, 5, 185, 243, 5, 5, 9, 192, 5, 5, 241, 190, 5, 5, 242, 218, 5, 5, 51, 243, 5, 5, 200, 227, 5, 5, 129, 170, 5, 5, 171, 244, 5, 5, 105, 226, 5, 5, 94, 191, 5, 5, 91, 186, 5, 5, 164, 190, 5, 5, 235, 214, 5, 5, 15, 166, 5, 5, 235, 179, 5, 5, 102, 248, 5, 5, 10, 215, 5, 5, 82, 199, 5, 5, 56, 165, 5, 5, 152, 165, 5, 5, 6, 219, 5, 5, 64, 206, 5, 5, 155, 195, 5, 5, 222, 228, 5, 5, 235, 228, 5, 5, 157, 176, 5, 5, 109, 171, 5, 5, 108, 238, 5, 5, 236, 189, 5, 5, 49, 199, 5, 5, 237, 198, 5, 5, 17, 242, 5, 5, 138, 224, 5, 5, 65, 173, 5, 5, 221, 249, 5, 5, 137, 194, 5, 5, 130, 170, 5, 5, 131, 211, 5, 5, 220, 247, 5, 5, 97, 168, 5, 5, 200, 202, 5, 5, 4, 241, 5, 5, 242, 223, 5, 5, 139, 208, 5, 5, 87, 213, 5, 5, 201, 202, 5, 5, 39, 190, 5, 5, 238, 185, 5, 5, 244, 190, 5, 5, 196, 205, 5, 5, 168, 193, 5, 5, 85, 235, 5, 5, 10, 186, 5, 5, 129, 218, 5, 5, 247, 191, 5, 5, 61, 211, 5, 5, 188, 186, 5, 5, 242, 242, 5, 5, 38, 234, 5, 5, 100, 204, 5, 5, 50, 235, 5, 5, 107, 226, 5, 5, 115, 182, 5, 5, 170, 238, 5, 5, 137, 180, 5, 5, 76, 241, 5, 5, 55, 213, 5, 5, 110, 171, 5, 5, 81, 180, 5, 5, 127, 169, 5, 5, 236, 228, 5, 5, 140, 198, 5, 5, 132, 188, 5, 5, 69, 186, 5, 5, 147, 245, 5, 5, 51, 223, 5, 5, 127, 192, 5, 5, 246, 173, 5, 5, 141, 217, 5, 5, 108, 239, 5, 5, 142, 170, 5, 5, 202, 231, 5, 5, 218, 170, 5, 5, 218, 195, 5, 5, 146, 190, 5, 5, 145, 167, 5, 5, 154, 229, 5, 5, 38, 221, 5, 5, 18, 221, 5, 5, 189, 222, 5, 5, 25, 217, 5, 5, 235, 200, 5, 5, 184, 199, 5, 5, 115, 205, 5, 5, 144, 242, 5, 5, 254, 229, 5, 5, 230, 187, 5, 5, 225, 234, 5, 5, 179, 180, 5, 5, 86, 244, 5, 5, 20, 206, 5, 5, 86, 235, 5, 5, 215, 244, 5, 5, 167, 201, 5, 5, 237, 217, 5, 5, 155, 229, 5, 5, 110, 194, 5, 5, 174, 244, 5, 5, 155, 214, 5, 5, 153, 233, 5, 5, 246, 196, 5, 5, 26, 217, 5, 5, 201, 188, 5, 5, 7, 219, 5, 5, 30, 246, 5, 5, 45, 168, 5, 5, 130, 248, 5, 5, 104, 204, 5, 5, 177, 166, 5, 5, 109, 176, 5, 5, 109, 200, 5, 5, 182, 203, 5, 5, 72, 245, 5, 5, 163, 166, 5, 5, 94, 202, 5, 5, 55, 167, 5, 5, 32, 246, 5, 5, 180, 236, 5, 5, 136, 210, 5, 5, 137, 210, 5, 5, 249, 212, 5, 5, 54, 190, 5, 5, 144, 213, 5, 5, 21, 218, 5, 5, 217, 178, 5, 5, 58, 201, 5, 5, 17, 172, 5, 5, 44, 231, 5, 5, 181, 236, 5, 5, 204, 185, 5, 5, 196, 220, 5, 5, 87, 235, 5, 5, 133, 240, 5, 5, 178, 219, 5, 5, 179, 246, 5, 5, 102, 228, 5, 5, 3, 180, 5, 5, 246, 223, 5, 5, 241, 217, 5, 5, 167, 196, 5, 5, 58, 241, 5, 5, 10, 228, 5, 5, 225, 181, 5, 5, 153, 221, 5, 5, 204, 235, 5, 5, 209, 197, 5, 5, 64, 165, 5, 5, 65, 228, 5, 5, 93, 224, 5, 5, 94, 224, 5, 5, 115, 229, 5, 5, 194, 182, 5, 5, 218, 168, 5, 5, 234, 188, 5, 5, 33, 246, 5, 5, 242, 235, 5, 5, 235, 235, 5, 5, 129, 225, 5, 5, 110, 208, 5, 5, 130, 225, 5, 5, 150, 199, 5, 5, 7, 204, 5, 5, 213, 185, 5, 5, 199, 223, 5, 5, 144, 192, 5, 5, 27, 235, 5, 5, 188, 201, 5, 5, 135, 235, 5, 5, 157, 229, 5, 5, 147, 167, 5, 5, 99, 198, 5, 5, 152, 249, 5, 5, 104, 188, 5, 5, 189, 201, 5, 5, 136, 174, 5, 5, 10, 198, 5, 5, 136, 197, 5, 5, 113, 215, 5, 5, 119, 205, 5, 5, 66, 228, 5, 5, 147, 213, 5, 5, 186, 248, 5, 5, 238, 189, 5, 5, 28, 186, 5, 5, 140, 184, 5, 5, 232, 247, 5, 5, 16, 189, 5, 5, 214, 171, 5, 5, 72, 232, 5, 5, 75, 176, 5, 5, 240, 200, 5, 5, 6, 232, 5, 5, 93, 214, 5, 5, 136, 248, 5, 5, 13, 187, 5, 5, 99, 227, 5, 5, 69, 217, 5, 5, 139, 189, 5, 5, 242, 246, 5, 5, 146, 231, 5, 5, 227, 249, 5, 5, 118, 200, 5, 5, 98, 234, 5, 5, 203, 188, 5, 5, 180, 177, 5, 5, 114, 241, 5, 5, 157, 238, 5, 5, 187, 200, 5, 5, 200, 220, 5, 5, 96, 224, 5, 5, 151, 170, 5, 5, 16, 176, 5, 5, 54, 176, 5, 5, 155, 240, 5, 5, 38, 237, 5, 5, 48, 168, 5, 5, 92, 207, 5, 5, 75, 179, 5, 5, 246, 210, 5, 5, 255, 201, 5, 5, 58, 222, 5, 5, 146, 188, 5, 5, 197, 184, 5, 5, 154, 193, 5, 5, 244, 205, 5, 5, 110, 173, 5, 5, 238, 235, 5, 5, 102, 202, 5, 5, 6, 237, 5, 5, 119, 200, 5, 5, 194, 225, 5, 5, 78, 179, 5, 5, 104, 237, 5, 5, 207, 177, 5, 5, 239, 235, 5, 5, 200, 221, 5, 5, 70, 185, 5, 5, 25, 166, 5, 5, 147, 188, 5, 5, 82, 171, 5, 5, 216, 226, 5, 5, 160, 229, 5, 5, 118, 171, 5, 5, 139, 184, 5, 5, 184, 244, 5, 5, 185, 244, 5, 5, 241, 170, 5, 5, 251, 205, 5, 5, 226, 208, 5, 5, 84, 165, 5, 5, 223, 202, 5, 5, 235, 197, 5, 5, 200, 230, 5, 5, 252, 234, 5, 5, 208, 245, 5, 5, 167, 179, 5, 5, 225, 243, 5, 5, 226, 243, 5, 5, 146, 178, 5, 5, 194, 233, 5, 5, 210, 199, 5, 5, 183, 206, 5, 5, 94, 186, 5, 5, 73, 208, 5, 5, 50, 238, 5, 5, 235, 168, 5, 5, 64, 222, 5, 5, 37, 246, 5, 5, 26, 166, 5, 5, 68, 193, 5, 5, 144, 210, 5, 5, 186, 244, 5, 5, 49, 166, 5, 5, 65, 196, 5, 5, 220, 172, 5, 5, 130, 234, 5, 5, 162, 169, 5, 5, 140, 205, 5, 5, 105, 246, 5, 5, 154, 190, 5, 5, 151, 232, 5, 5, 103, 191, 5, 5, 118, 187, 5, 5, 21, 242, 5, 5, 88, 244, 5, 5, 247, 189, 5, 5, 194, 235, 5, 5, 28, 219, 5, 5, 191, 244, 5, 5, 3, 178, 5, 5, 104, 247, 5, 5, 34, 200, 5, 5, 105, 214, 5, 5, 72, 167, 5, 5, 201, 228, 5, 5, 166, 165, 5, 5, 102, 224, 5, 5, 143, 169, 5, 5, 71, 188, 5, 5, 247, 204, 5, 5, 11, 169, 5, 5, 244, 235, 5, 5, 133, 225, 5, 5, 73, 193, 5, 5, 31, 166, 5, 5, 216, 199, 5, 5, 233, 215, 5, 5, 206, 189, 5, 5, 23, 211, 5, 5, 157, 198, 5, 5, 195, 244, 5, 5, 90, 231, 5, 5, 127, 187, 5, 5, 184, 206, 5, 5, 178, 214, 5, 5, 196, 244, 5, 5, 17, 237, 5, 5, 110, 246, 5, 5, 107, 214, 5, 5, 242, 199, 5, 5, 87, 217, 5, 5, 159, 221, 5, 5, 254, 190, 5, 5, 77, 185, 5, 5, 152, 178, 5, 5, 31, 173, 5, 5, 35, 199, 5, 5, 78, 185, 5, 5, 166, 229, 5, 5, 180, 189, 5, 5, 199, 244, 5, 5, 54, 171, 5, 5, 238, 202, 5, 5, 209, 188, 5, 5, 145, 205, 5, 5, 200, 244, 5, 5, 141, 173, 5, 5, 253, 188, 5, 5, 192, 240, 5, 5, 255, 204, 5, 5, 236, 248, 5, 5, 110, 224, 5, 5, 118, 175, 5, 5, 153, 230, 5, 5, 66, 205, 5, 5, 45, 186, 5, 5, 32, 173, 5, 5, 4, 230, 5, 5, 23, 249, 5, 5, 70, 222, 5, 5, 227, 173, 5, 5, 211, 227, 5, 5, 137, 185, 5, 5, 51, 186, 5, 5, 94, 235, 5, 5, 223, 211, 5, 5, 236, 249, 5, 5, 147, 210, 5, 5, 140, 172, 5, 5, 118, 194, 5, 5, 40, 212, 5, 5, 97, 243, 5, 5, 19, 198, 5, 5, 236, 244, 5, 5, 205, 235, 5, 5, 159, 232, 5, 5, 175, 190, 5, 5, 87, 243, 5, 5, 238, 186, 5, 5, 84, 171, 5, 5, 126, 228, 5, 5, 239, 179, 5, 5, 89, 176, 5, 5, 80, 204, 5, 5, 154, 224, 5, 5, 237, 244, 5, 5, 22, 198, 5, 5, 31, 163, 5, 5, 183, 221, 5, 5, 84, 186, 5, 5, 57, 186, 5, 5, 139, 183, 5, 5, 183, 246, 5, 5, 179, 190, 5, 5, 250, 217, 5, 5, 177, 215, 5, 5, 99, 243, 5, 5, 168, 192, 5, 5, 104, 243, 5, 5, 181, 248, 5, 5, 178, 199, 5, 5, 138, 165, 5, 5, 129, 171, 5, 5, 154, 194, 5, 5, 121, 176, 5, 5, 252, 199, 5, 5, 120, 214, 5, 5, 121, 214, 5, 5, 150, 200, 5, 5, 110, 201, 5, 5, 86, 189, 5, 5, 77, 232, 5, 5, 137, 222, 5, 5, 146, 204, 5, 5, 106, 219, 5, 5, 50, 214, 5, 5, 83, 223, 5, 5, 23, 222, 5, 5, 29, 232, 5, 5, 30, 232, 5, 5, 92, 234, 5, 5, 18, 244, 5, 5, 32, 209, 5, 5, 201, 222, 5, 5, 98, 173, 5, 5, 52, 184, 5, 5, 95, 208, 5, 5, 22, 247, 5, 5, 192, 166, 5, 5, 34, 240, 5, 5, 187, 189, 5, 5, 27, 226, 5, 5, 203, 167, 5, 5, 107, 195, 5, 5, 82, 167, 5, 5, 163, 242, 5, 5, 108, 195, 5, 5, 130, 167, 5, 5, 86, 172, 5, 5, 253, 223, 5, 5, 20, 188, 5, 5, 33, 194, 5, 5, 236, 195, 5, 5, 207, 178, 5, 5, 156, 194, 5, 5, 249, 215, 5, 5, 192, 218, 5, 5, 157, 194, 5, 5, 158, 194, 5, 5, 54, 246, 5, 5, 82, 211, 5, 5, 4, 166, 5, 5, 23, 175, 5, 5, 189, 213, 5, 5, 72, 221, 5, 5, 168, 221, 5, 5, 42, 195, 5, 5, 229, 219, 5, 5, 85, 174, 5, 5, 0, 160, 5, 5, 57, 214, 5, 5, 103, 178, 5, 5, 177, 189, 5, 5, 7, 222, 5, 5, 161, 217, 5, 5, 54, 207, 5, 5, 38, 180, 5, 5, 28, 167, 5, 5, 97, 208, 5, 5, 17, 167, 5, 5, 112, 179, 5, 5, 113, 179, 5, 5, 66, 166, 5, 5, 63, 244, 5, 5, 156, 182, 5, 5, 61, 198, 5, 5, 222, 221, 5, 5, 111, 219, 5, 5, 147, 177, 5, 5, 119, 246, 5, 5, 239, 226, 5, 5, 186, 188, 5, 5, 89, 223, 5, 5, 86, 201, 5, 5, 74, 220, 5, 5, 225, 198, 5, 5, 180, 229, 5, 5, 90, 186, 5, 5, 132, 181, 5, 5, 37, 204, 5, 5, 4, 217, 5, 5, 202, 198, 5, 5, 103, 212, 5, 5, 191, 187, 5, 5, 217, 200, 5, 5, 32, 244, 5, 5, 242, 214, 5, 5, 243, 214, 5, 5, 144, 215, 5, 5, 170, 243, 5, 5, 254, 244, 5, 5, 6, 220, 5, 5, 165, 241, 5, 5, 59, 227, 5, 5, 225, 245, 5, 5, 173, 182, 5, 5, 130, 171, 5, 5, 34, 232, 5, 5, 110, 195, 5, 5, 114, 201, 5, 5, 208, 185, 5, 5, 192, 213, 5, 5, 170, 221, 5, 5, 150, 217, 5, 5, 74, 221, 5, 5, 198, 166, 5, 5, 124, 242, 5, 5, 46, 200, 5, 5, 164, 194, 5, 5, 190, 223, 5, 5, 8, 200, 5, 5, 123, 195, 5, 5, 236, 224, 5, 5, 13, 173, 5, 5, 229, 227, 5, 5, 91, 241, 5, 5, 77, 200, 192, 0, 0, 0, 5, 5, 125, 180, 5, 5, 219, 205, 5, 5, 15, 230, 5, 5, 219, 187, 5, 5, 205, 212, 5, 5, 104, 174, 5, 5, 101, 208, 5, 5, 68, 214, 5, 5, 248, 168, 5, 5, 134, 181, 5, 5, 39, 222, 5, 5, 171, 181, 5, 5, 196, 246, 5, 5, 190, 197, 5, 5, 141, 167, 5, 5, 131, 186, 5, 5, 42, 243, 5, 5, 10, 217, 5, 5, 206, 173, 5, 5, 94, 223, 5, 5, 58, 231, 5, 5, 3, 199, 5, 5, 206, 212, 5, 5, 77, 229, 5, 5, 155, 197, 5, 5, 188, 217, 5, 5, 29, 196, 5, 5, 218, 207, 5, 5, 138, 214, 5, 5, 207, 212, 5, 5, 133, 242, 5, 5, 129, 198, 5, 5, 188, 222, 5, 5, 54, 208, 5, 5, 236, 176, 5, 5, 69, 214, 5, 5, 71, 202, 5, 5, 104, 245, 5, 5, 134, 242, 5, 5, 107, 191, 5, 5, 89, 198, 5, 5, 37, 243, 5, 5, 220, 200, 5, 5, 172, 170, 5, 5, 244, 206, 5, 5, 158, 175, 5, 5, 223, 163, 5, 5, 105, 173, 5, 5, 25, 189, 5, 5, 204, 196, 5, 5, 24, 201, 5, 5, 132, 182, 5, 5, 218, 212, 5, 5, 243, 234, 5, 5, 28, 245, 5, 5, 118, 201, 5, 5, 255, 223, 5, 5, 168, 174, 5, 5, 200, 245, 5, 5, 249, 203, 5, 5, 47, 235, 5, 5, 3, 174, 5, 5, 190, 168, 5, 5, 140, 212, 5, 5, 238, 230, 5, 5, 219, 212, 5, 5, 141, 212, 5, 5, 30, 216, 5, 5, 230, 223, 5, 5, 43, 243, 5, 5, 44, 243, 5, 5, 93, 240, 5, 5, 155, 207, 5, 5, 91, 248, 5, 5, 207, 192, 5, 5, 66, 230, 5, 5, 3, 168, 5, 5, 48, 218, 5, 5, 114, 247, 5, 5, 26, 229, 5, 5, 44, 225, 5, 5, 232, 227, 5, 5, 46, 218, 5, 5, 75, 184, 5, 5, 219, 192, 5, 5, 168, 246, 5, 5, 199, 224, 5, 5, 159, 197, 5, 5, 148, 196, 5, 5, 233, 222, 5, 5, 169, 178, 5, 5, 176, 213, 5, 5, 21, 200, 5, 5, 4, 174, 5, 5, 198, 188, 5, 5, 107, 217, 5, 5, 124, 216, 5, 5, 52, 201, 5, 5, 48, 204, 5, 5, 220, 204, 5, 5, 42, 165, 5, 5, 108, 244, 5, 5, 198, 221, 5, 5, 78, 246, 5, 5, 104, 182, 5, 5, 81, 228, 5, 5, 236, 212, 5, 5, 183, 184, 5, 5, 184, 184, 5, 5, 83, 224, 5, 5, 141, 242, 5, 5, 105, 208, 5, 5, 108, 208, 5, 5, 103, 248, 5, 5, 9, 170, 5, 5, 138, 227, 5, 5, 134, 246, 5, 5, 86, 219, 5, 5, 31, 216, 5, 5, 148, 212, 5, 5, 38, 230, 5, 5, 215, 169, 5, 5, 232, 200, 5, 5, 22, 217, 5, 5, 238, 219, 5, 5, 242, 245, 5, 5, 18, 173, 5, 5, 31, 210, 5, 5, 30, 237, 5, 5, 44, 244, 5, 5, 200, 187, 5, 5, 196, 221, 5, 5, 117, 203, 5, 5, 48, 192, 5, 5, 112, 193, 5, 5, 255, 163, 5, 5, 13, 199, 5, 5, 114, 249, 5, 5, 177, 213, 5, 5, 114, 227, 5, 5, 81, 189, 5, 5, 200, 165, 5, 5, 96, 192, 5, 5, 97, 182, 5, 5, 82, 180, 5, 5, 2, 224, 5, 5, 64, 228, 5, 5, 5, 239, 5, 5, 105, 163, 5, 5, 94, 201, 5, 5, 25, 214, 5, 5, 138, 213, 5, 5, 119, 245, 5, 5, 237, 212, 5, 5, 193, 249, 5, 5, 133, 197, 5, 5, 161, 197, 5, 5, 238, 212, 5, 5, 150, 249, 5, 5, 45, 218, 5, 5, 8, 183, 5, 5, 83, 228, 5, 5, 163, 212, 5, 5, 244, 241, 5, 5, 191, 217, 5, 5, 111, 171, 5, 5, 133, 188, 5, 5, 175, 244, 5, 5, 66, 245, 5, 5, 74, 231, 5, 5, 208, 194, 5, 5, 106, 243, 5, 5, 84, 211, 5, 5, 29, 188, 5, 5, 239, 212, 5, 5, 65, 165, 5, 5, 156, 165, 5, 5, 56, 235, 5, 5, 45, 190, 5, 5, 232, 204, 5, 5, 104, 229, 5, 5, 134, 240, 5, 5, 240, 185, 5, 5, 167, 167, 5, 5, 5, 237, 5, 5, 135, 240, 5, 5, 129, 180, 5, 5, 171, 207, 5, 5, 106, 233, 5, 5, 194, 249, 5, 5, 85, 199, 5, 5, 187, 239, 5, 5, 17, 219, 5, 5, 114, 238, 5, 5, 138, 246, 5, 5, 93, 188, 5, 5, 145, 198, 5, 5, 173, 238, 5, 5, 16, 230, 5, 5, 109, 208, 5, 5, 147, 170, 5, 5, 247, 196, 5, 5, 238, 228, 5, 5, 18, 235, 5, 5, 200, 232, 5, 5, 237, 204, 5, 5, 7, 213, 5, 5, 8, 213, 5, 5, 9, 213, 5, 5, 156, 240, 5, 5, 229, 214, 5, 5, 115, 223, 5, 5, 64, 201, 5, 5, 144, 224, 5, 5, 115, 241, 5, 5, 8, 180, 5, 5, 71, 230, 5, 5, 65, 243, 5, 5, 120, 215, 5, 5, 147, 209, 5, 5, 72, 208, 5, 5, 126, 176, 5, 5, 219, 178, 5, 5, 169, 228, 5, 5, 9, 204, 5, 5, 113, 173, 5, 5, 141, 235, 5, 5, 241, 187, 5, 5, 137, 248, 5, 5, 240, 188, 5, 5, 200, 182, 5, 5, 239, 227, 5, 5, 89, 173, 5, 5, 23, 173, 5, 5, 66, 243, 5, 5, 245, 217, 5, 5, 92, 186, 5, 5, 247, 169, 5, 5, 128, 180, 5, 5, 157, 240, 5, 5, 133, 206, 5, 5, 175, 170, 5, 5, 189, 206, 5, 5, 66, 204, 5, 5, 187, 244, 5, 5, 148, 248, 5, 5, 169, 240, 5, 5, 70, 217, 5, 5, 34, 186, 5, 5, 146, 179, 5, 5, 88, 232, 5, 5, 99, 234, 5, 5, 42, 237, 5, 5, 17, 213, 5, 5, 35, 212, 5, 5, 65, 235, 5, 5, 150, 188, 5, 5, 62, 200, 5, 5, 178, 221, 5, 5, 98, 214, 5, 5, 146, 189, 5, 5, 249, 245, 5, 5, 170, 240, 5, 5, 36, 214, 5, 5, 245, 219, 5, 5, 124, 200, 5, 5, 53, 175, 5, 5, 199, 184, 5, 5, 35, 186, 5, 5, 203, 230, 5, 5, 164, 215, 5, 5, 126, 195, 5, 5, 201, 217, 5, 5, 252, 205, 5, 5, 43, 224, 5, 5, 171, 175, 5, 5, 252, 241, 5, 5, 172, 197, 5, 5, 95, 186, 5, 5, 168, 165, 5, 5, 88, 231, 5, 5, 128, 187, 5, 5, 147, 169, 5, 5, 61, 164, 5, 5, 244, 230, 5, 5, 105, 191, 5, 5, 252, 184, 5, 5, 101, 238, 5, 5, 24, 218, 5, 5, 42, 197, 5, 5, 194, 229, 5, 5, 27, 188, 5, 5, 134, 187, 5, 5, 211, 189, 5, 5, 195, 207, 5, 5, 17, 188, 5, 5, 83, 243, 5, 5, 34, 213, 5, 5, 108, 214, 5, 5, 7, 191, 5, 5, 133, 241, 5, 5, 104, 234, 5, 5, 166, 230, 5, 5, 239, 248, 5, 5, 96, 186, 5, 5, 96, 235, 5, 5, 200, 207, 5, 5, 174, 226, 5, 5, 119, 194, 5, 5, 107, 207, 5, 5, 204, 223, 5, 5, 61, 171, 5, 5, 122, 180, 5, 5, 98, 186, 5, 5, 186, 221, 5, 5, 87, 186, 5, 5, 163, 189, 5, 5, 22, 235, 5, 5, 183, 248, 5, 5, 219, 219, 5, 5, 251, 219, 5, 5, 159, 166, 5, 5, 67, 221, 5, 5, 18, 167, 5, 5, 55, 225, 5, 5, 35, 249, 5, 5, 57, 238, 5, 5, 131, 246, 5, 5, 162, 198, 5, 5, 172, 207, 5, 5, 183, 207, 5, 5, 186, 174, 5, 5, 60, 241, 5, 5, 17, 163, 5, 5, 19, 177, 5, 5, 66, 241, 5, 5, 155, 241, 5, 5, 233, 232, 5, 5, 251, 229, 5, 5, 61, 239, 5, 5, 57, 167, 5, 5, 131, 226, 5, 5, 254, 194, 5, 5, 173, 232, 5, 5, 49, 177, 5, 5, 43, 211, 5, 5, 243, 249, 5, 5, 69, 211, 5, 5, 123, 168, 5, 5, 88, 164, 5, 5, 84, 167, 5, 5, 236, 223, 5, 5, 212, 204, 5, 5, 214, 166, 5, 5, 246, 177, 5, 5, 125, 226, 5, 5, 212, 247, 5, 5, 87, 219, 5, 5, 201, 165, 5, 5, 50, 177, 5, 5, 155, 164, 5, 5, 106, 163, 5, 5, 127, 226, 5, 5, 89, 242, 5, 5, 121, 242, 5, 5, 6, 178, 5, 5, 240, 240, 5, 5, 63, 225, 5, 5, 46, 184, 5, 5, 60, 184, 5, 5, 153, 168, 5, 5, 123, 204, 5, 5, 83, 170, 5, 5, 33, 179, 5, 5, 63, 172, 5, 5, 169, 185, 5, 5, 121, 230, 5, 5, 181, 170, 5, 5, 151, 173, 5, 5, 22, 170, 5, 5, 87, 172, 5, 5, 186, 192, 5, 5, 91, 245, 5, 5, 24, 243, 5, 5, 138, 191, 5, 5, 204, 163, 5, 5, 173, 235, 5, 5, 83, 205, 5, 5, 37, 193, 5, 5, 106, 242, 5, 5, 141, 191, 5, 5, 18, 229, 5, 5, 68, 172, 5, 5, 210, 176, 5, 5, 70, 210, 5, 5, 109, 212, 5, 5, 60, 212, 5, 5, 195, 231, 5, 5, 94, 179, 5, 5, 194, 212, 5, 5, 201, 170, 5, 5, 247, 195, 5, 5, 78, 210, 5, 5, 61, 245, 5, 5, 87, 193, 5, 5, 220, 212, 5, 5, 176, 186, 5, 5, 184, 186, 5, 5, 49, 167, 5, 5, 228, 201, 5, 5, 197, 192, 5, 5, 116, 218, 5, 5, 119, 164, 5, 5, 71, 212, 5, 5, 104, 226, 5, 5, 247, 177, 5, 5, 71, 207, 5, 5, 27, 203, 5, 5, 221, 204, 5, 5, 211, 173, 5, 5, 102, 168, 5, 5, 14, 195, 5, 5, 250, 206, 5, 5, 128, 195, 5, 5, 168, 225, 5, 5, 139, 213, 5, 5, 233, 163, 5, 5, 147, 190, 5, 5, 228, 203, 5, 5, 25, 169, 5, 5, 85, 236, 5, 5, 139, 246, 5, 5, 37, 170, 5, 5, 193, 178, 5, 5, 201, 182, 5, 5, 211, 209, 5, 5, 67, 206, 5, 5, 166, 213, 5, 5, 129, 187, 5, 5, 7, 178, 5, 5, 211, 187, 5, 5, 218, 166, 5, 5, 75, 168, 5, 5, 132, 163, 5, 5, 35, 213, 5, 5, 14, 224, 5, 5, 46, 246, 5, 5, 146, 196, 5, 5, 15, 224, 5, 5, 100, 212, 5, 5, 99, 212, 5, 5, 0, 155, 5, 5, 46, 173, 5, 5, 247, 209, 5, 5, 127, 202, 5, 5, 8, 234, 5, 5, 22, 222, 5, 5, 237, 238, 5, 5, 133, 191, 5, 5, 185, 222, 5, 5, 174, 194, 5, 5, 42, 169, 5, 5, 30, 242, 5, 5, 234, 195, 5, 5, 67, 163, 5, 5, 45, 198, 5, 5, 25, 241, 5, 5, 162, 243, 5, 5, 154, 200, 5, 5, 186, 222, 5, 5, 162, 249, 5, 5, 11, 197, 5, 5, 50, 184, 5, 5, 163, 176, 5, 5, 88, 211, 5, 5, 242, 222, 5, 5, 85, 172, 5, 5, 200, 163, 5, 5, 204, 167, 5, 5, 238, 206, 5, 5, 140, 239, 5, 5, 15, 229, 5, 5, 16, 167, 5, 5, 185, 213, 5, 5, 25, 182, 5, 5, 249, 198, 5, 5, 94, 189, 5, 5, 197, 172, 5, 5, 240, 167, 5, 5, 184, 183, 5, 5, 214, 241, 5, 5, 98, 242, 5, 5, 68, 192, 5, 5, 82, 217, 5, 5, 197, 206, 5, 5, 25, 176, 5, 5, 71, 163, 5, 5, 20, 196, 5, 5, 138, 215, 5, 5, 111, 184, 5, 5, 170, 200, 5, 5, 168, 198, 5, 5, 32, 204, 5, 5, 96, 197, 5, 5, 230, 219, 5, 5, 58, 202, 5, 5, 6, 218, 5, 5, 48, 228, 5, 5, 55, 179, 5, 5, 51, 231, 5, 5, 21, 228, 5, 5, 169, 182, 5, 5, 198, 185, 5, 5, 240, 177, 5, 5, 190, 213, 5, 5, 193, 226, 5, 5, 83, 196, 5, 5, 73, 221, 5, 5, 60, 248, 5, 5, 67, 237, 5, 5, 181, 247, 5, 5, 14, 165, 5, 5, 199, 170, 5, 5, 195, 212, 5, 5, 246, 207, 5, 5, 170, 186, 5, 5, 193, 213, 5, 5, 171, 243, 5, 5, 57, 237, 5, 5, 106, 186, 5, 5, 213, 163, 5, 5, 210, 218, 5, 5, 171, 242, 5, 5, 127, 227, 5, 5, 60, 187, 5, 5, 125, 198, 5, 5, 145, 215, 5, 5, 90, 206, 5, 5, 212, 166, 5, 5, 161, 172, 5, 5, 139, 214, 5, 5, 172, 163, 5, 5, 188, 208, 5, 5, 159, 183, 5, 5, 146, 165, 5, 5, 240, 243, 5, 5, 47, 185, 5, 5, 196, 187, 5, 5, 208, 204, 5, 5, 33, 211, 5, 5, 203, 196, 5, 5, 138, 242, 5, 5, 179, 193, 5, 5, 144, 233, 5, 5, 39, 171, 5, 5, 242, 224, 5, 5, 100, 239, 5, 5, 224, 218, 5, 5, 81, 221, 5, 5, 253, 240, 5, 5, 8, 171, 5, 5, 6, 234, 5, 5, 152, 234, 5, 5, 169, 227, 5, 5, 220, 187, 5, 5, 149, 177, 5, 5, 90, 190, 5, 5, 9, 199, 5, 5, 96, 233, 5, 5, 40, 194, 5, 5, 74, 246, 5, 5, 208, 173, 5, 5, 175, 245, 5, 5, 47, 249, 5, 5, 123, 216, 5, 5, 140, 221, 5, 5, 17, 170, 5, 5, 132, 246, 5, 5, 50, 226, 5, 5, 209, 215, 5, 5, 103, 182, 5, 5, 40, 241, 5, 5, 52, 183, 5, 5, 245, 213, 5, 5, 24, 203, 5, 5, 30, 198, 5, 5, 198, 165, 5, 5, 50, 242, 5, 5, 55, 242, 5, 5, 210, 169, 5, 5, 148, 223, 5, 5, 180, 241, 5, 5, 72, 237, 5, 5, 98, 163, 5, 5, 38, 172, 5, 5, 118, 199, 5, 5, 58, 249, 5, 5, 48, 236, 5, 5, 49, 169, 5, 5, 59, 249, 5, 5, 71, 175, 5, 5, 116, 248, 5, 5, 203, 226, 5, 5, 221, 197, 5, 5, 87, 211, 5, 5, 107, 240, 5, 5, 183, 220, 5, 5, 121, 234, 5, 5, 229, 197, 5, 5, 143, 170, 5, 5, 174, 206, 5, 5, 43, 206, 5, 5, 128, 169, 5, 5, 19, 221, 5, 5, 87, 224, 5, 5, 110, 238, 5, 5, 109, 238, 5, 5, 224, 197, 5, 5, 220, 234, 5, 5, 138, 164, 5, 5, 173, 207, 5, 5, 66, 165, 5, 5, 67, 165, 5, 5, 59, 201, 5, 5, 233, 168, 5, 5, 235, 188, 5, 5, 33, 217, 5, 5, 195, 182, 5, 5, 247, 241, 5, 5, 238, 187, 5, 5, 190, 229, 5, 5, 162, 214, 5, 5, 36, 228, 5, 5, 189, 178, 5, 5, 107, 204, 5, 5, 190, 191, 5, 5, 10, 208, 5, 5, 252, 177, 5, 5, 120, 200, 5, 5, 94, 196, 5, 5, 160, 215, 5, 5, 94, 214, 5, 5, 48, 179, 5, 5, 162, 215, 5, 5, 198, 217, 5, 5, 126, 184, 5, 5, 144, 168, 5, 5, 211, 199, 5, 5, 89, 183, 5, 5, 16, 171, 5, 5, 210, 187, 5, 5, 76, 183, 5, 5, 177, 194, 5, 5, 33, 228, 5, 5, 207, 189, 5, 5, 50, 179, 5, 5, 235, 215, 5, 5, 55, 171, 5, 5, 154, 230, 5, 5, 161, 200, 5, 5, 77, 183, 5, 5, 110, 190, 5, 5, 174, 215, 5, 5, 247, 208, 5, 5, 205, 207, 5, 5, 132, 217, 5, 5, 167, 171, 5, 5, 113, 231, 5, 5, 89, 191, 5, 5, 194, 189, 5, 5, 48, 248, 5, 5, 140, 236, 5, 5, 81, 247, 5, 5, 248, 209, 5, 5, 127, 191, 5, 5, 191, 177, 5, 5, 19, 185, 5, 5, 0, 134, 5, 5, 204, 201, 5, 5, 134, 201, 5, 5, 226, 176, 5, 5, 72, 226, 5, 5, 107, 192, 5, 5, 147, 204, 5, 5, 164, 247, 5, 5, 3, 209, 5, 5, 165, 247, 5, 5, 38, 225, 5, 5, 31, 169, 5, 5, 194, 194, 5, 5, 54, 245, 5, 5, 23, 174, 5, 5, 23, 247, 5, 5, 212, 203, 5, 5, 245, 240, 5, 5, 7, 166, 5, 5, 236, 165, 5, 5, 159, 164, 5, 5, 205, 248, 5, 5, 127, 207, 5, 5, 161, 170, 5, 5, 128, 207, 5, 5, 192, 246, 5, 5, 197, 204, 5, 5, 195, 169, 5, 5, 5, 176, 5, 5, 135, 168, 5, 5, 55, 246, 5, 5, 167, 200, 5, 5, 182, 215, 5, 5, 30, 211, 5, 5, 98, 189, 5, 5, 91, 175, 5, 5, 49, 240, 5, 5, 224, 238, 5, 5, 177, 168, 5, 5, 166, 177, 5, 5, 86, 174, 5, 5, 133, 217, 5, 5, 77, 213, 5, 5, 41, 163, 5, 5, 204, 233, 5, 5, 71, 192, 5, 5, 70, 203, 5, 5, 247, 227, 5, 5, 196, 166, 5, 5, 171, 233, 5, 5, 255, 194, 5, 5, 78, 230, 5, 5, 19, 229, 5, 5, 147, 228, 5, 5, 150, 239, 5, 5, 234, 232, 5, 5, 12, 203, 5, 5, 229, 171, 5, 5, 49, 228, 5, 5, 86, 188, 5, 5, 10, 165, 5, 5, 161, 220, 5, 5, 162, 220, 5, 5, 131, 207, 5, 5, 141, 172, 5, 5, 231, 190, 5, 5, 170, 205, 5, 5, 87, 204, 5, 5, 89, 205, 5, 5, 61, 187, 5, 5, 180, 194, 5, 5, 49, 215, 5, 5, 7, 220, 5, 5, 224, 194, 5, 5, 125, 242, 5, 5, 15, 215, 5, 5, 33, 244, 5, 5, 231, 216, 5, 5, 138, 190, 5, 5, 147, 236, 5, 5, 137, 221, 5, 5, 222, 210, 5, 5, 172, 243, 5, 5, 225, 217, 5, 5, 21, 164, 5, 5, 59, 239, 5, 5, 99, 216, 5, 5, 166, 241, 5, 5, 194, 213, 5, 5, 178, 198, 5, 5, 216, 193, 5, 5, 153, 226, 5, 5, 5, 217, 5, 5, 69, 164, 5, 5, 253, 166, 5, 5, 218, 185, 5, 5, 69, 176, 5, 5, 232, 244, 5, 5, 9, 200, 5, 5, 29, 247, 5, 5, 160, 237, 5, 5, 106, 225, 5, 5, 237, 176, 5, 5, 62, 245, 5, 5, 25, 245, 5, 5, 103, 225, 5, 5, 55, 208, 5, 5, 202, 169, 5, 5, 219, 163, 5, 5, 172, 181, 5, 5, 200, 215, 5, 5, 216, 225, 5, 5, 90, 198, 5, 5, 138, 217, 5, 5, 172, 242, 5, 5, 18, 210, 5, 5, 165, 217, 5, 5, 112, 186, 5, 5, 3, 187, 5, 5, 17, 215, 5, 5, 105, 174, 5, 5, 36, 188, 5, 5, 254, 240, 5, 5, 208, 212, 5, 5, 112, 189, 5, 5, 177, 186, 5, 5, 223, 205, 5, 5, 171, 188, 5, 5, 225, 236, 5, 5, 244, 236, 5, 5, 12, 174, 5, 5, 197, 246, 5, 5, 172, 177, 5, 5, 139, 168, 5, 5, 115, 220, 5, 5, 229, 235, 5, 5, 236, 171, 5, 5, 92, 248, 5, 5, 214, 206, 5, 5, 170, 224, 5, 5, 51, 185, 5, 5, 31, 247, 5, 5, 49, 229, 5, 5, 26, 246, 5, 5, 106, 181, 5, 5, 220, 225, 5, 5, 246, 232, 5, 5, 190, 164, 5, 5, 64, 225, 5, 5, 46, 199, 5, 5, 197, 208, 5, 5, 238, 237, 5, 5, 205, 196, 5, 5, 245, 224, 5, 5, 198, 208, 5, 5, 191, 164, 5, 5, 216, 246, 5, 5, 21, 190, 5, 5, 83, 235, 5, 5, 86, 194, 5, 5, 235, 242, 5, 5, 81, 177, 5, 5, 224, 163, 5, 5, 227, 188, 5, 5, 227, 210, 5, 5, 4, 168, 5, 5, 161, 166, 5, 5, 19, 212, 5, 5, 43, 222, 5, 5, 228, 210, 5, 5, 142, 172, 5, 5, 5, 206, 5, 5, 142, 212, 5, 5, 26, 178, 5, 5, 226, 203, 5, 5, 10, 201, 5, 5, 28, 248, 5, 5, 74, 170, 5, 5, 216, 169, 5, 5, 85, 226, 5, 5, 108, 217, 5, 5, 174, 180, 5, 5, 125, 216, 5, 5, 104, 175, 5, 5, 39, 230, 5, 5, 239, 171, 5, 5, 222, 204, 5, 5, 79, 246, 5, 5, 32, 231, 5, 5, 233, 200, 5, 5, 21, 214, 5, 5, 51, 226, 5, 5, 38, 205, 5, 5, 140, 222, 5, 5, 45, 225, 5, 5, 199, 188, 5, 5, 75, 212, 5, 5, 205, 213, 5, 5, 170, 178, 5, 5, 101, 249, 5, 5, 255, 240, 5, 5, 72, 183, 5, 5, 14, 199, 5, 5, 109, 244, 5, 5, 213, 247, 5, 5, 53, 201, 5, 5, 13, 231, 5, 5, 232, 235, 5, 5, 50, 188, 5, 5, 208, 226, 5, 5, 75, 231, 5, 5, 232, 236, 5, 5, 184, 241, 5, 5, 7, 247, 5, 5, 67, 245, 5, 5, 190, 220, 5, 5, 236, 205, 5, 5, 178, 244, 5, 5, 36, 169, 5, 5, 161, 179, 5, 5, 55, 188, 5, 5, 83, 180, 5, 5, 97, 187, 5, 5, 119, 186, 5, 5, 232, 193, 5, 5, 251, 220, 5, 5, 44, 211, 5, 5, 173, 234, 5, 5, 166, 207, 5, 5, 222, 219, 5, 5, 185, 241, 5, 5, 69, 238, 5, 5, 19, 166, 5, 5, 91, 229, 5, 5, 18, 225, 5, 5, 216, 196, 5, 5, 121, 222, 5, 5, 57, 235, 5, 5, 201, 205, 5, 5, 115, 239, 5, 5, 226, 234, 5, 5, 58, 235, 5, 5, 219, 208, 5, 5, 203, 186, 5, 5, 216, 244, 5, 5, 233, 236, 5, 5, 38, 210, 5, 5, 143, 217, 5, 5, 115, 238, 5, 5, 175, 188, 5, 5, 167, 170, 5, 5, 25, 221, 5, 5, 192, 243, 5, 5, 163, 199, 5, 5, 118, 244, 5, 5, 145, 246, 5, 5, 59, 190, 5, 5, 35, 245, 5, 5, 127, 176, 5, 5, 22, 215, 5, 5, 201, 232, 5, 5, 246, 217, 5, 5, 212, 211, 5, 5, 115, 176, 5, 5, 241, 188, 5, 5, 242, 187, 5, 5, 157, 212, 5, 5, 190, 236, 5, 5, 65, 201, 5, 5, 238, 193, 5, 5, 234, 168, 5, 5, 134, 206, 5, 5, 24, 173, 5, 5, 245, 205, 5, 5, 62, 188, 5, 5, 2, 202, 5, 5, 151, 200, 5, 5, 137, 206, 5, 5, 78, 238, 5, 5, 92, 165, 5, 5, 100, 234, 5, 5, 215, 211, 5, 5, 73, 232, 5, 5, 76, 176, 5, 5, 70, 204, 5, 5, 151, 194, 5, 5, 35, 180, 5, 5, 125, 200, 5, 5, 191, 200, 5, 5, 243, 235, 5, 5, 43, 237, 5, 5, 93, 181, 5, 5, 240, 225, 5, 5, 243, 247, 5, 5, 253, 205, 5, 5, 208, 189, 5, 5, 30, 166, 5, 5, 105, 229, 5, 5, 68, 247, 5, 5, 35, 233, 5, 5, 38, 186, 5, 5, 99, 213, 5, 5, 156, 231, 5, 5, 146, 217, 5, 5, 37, 231, 5, 5, 210, 245, 5, 5, 252, 191, 5, 5, 99, 185, 5, 5, 8, 178, 5, 5, 130, 187, 5, 5, 43, 219, 5, 5, 48, 230, 5, 5, 225, 186, 5, 5, 154, 232, 5, 5, 158, 212, 5, 5, 240, 180, 5, 5, 246, 193, 5, 5, 14, 237, 5, 5, 255, 184, 5, 5, 8, 191, 5, 5, 212, 189, 5, 5, 135, 187, 5, 5, 157, 232, 5, 5, 97, 186, 5, 5, 143, 244, 5, 5, 59, 171, 5, 5, 62, 216, 5, 5, 162, 212, 5, 5, 120, 194, 5, 5, 175, 175, 5, 5, 162, 232, 5, 5, 202, 200, 5, 5, 103, 193, 5, 5, 72, 163, 5, 5, 31, 177, 5, 5, 80, 163, 5, 5, 74, 199, 5, 5, 121, 218, 5, 5, 80, 168, 5, 5, 113, 164, 5, 5, 156, 230, 5, 5, 164, 187, 5, 5, 152, 245, 192, 0, 0, 0, 5, 5, 67, 216, 5, 5, 65, 210, 5, 5, 89, 164, 5, 5, 99, 167, 5, 5, 172, 173, 5, 5, 6, 176, 5, 5, 133, 247, 5, 5, 167, 238, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 167, 245, 5, 5, 252, 198, 5, 5, 101, 189, 5, 5, 159, 239, 5, 5, 187, 168, 5, 5, 177, 233, 5, 5, 161, 163, 5, 5, 220, 207, 5, 5, 11, 225, 5, 5, 13, 223, 192, 0, 0, 0, 5, 5, 104, 176, 5, 5, 9, 171, 5, 5, 141, 166, 5, 5, 97, 222, 5, 5, 195, 168, 5, 5, 202, 165, 5, 5, 62, 220, 5, 5, 191, 220, 192, 0, 0, 0, 5, 5, 183, 233, 5, 5, 119, 244, 5, 5, 54, 209, 5, 5, 114, 215, 5, 5, 5, 186, 5, 5, 13, 177, 5, 5, 175, 238, 5, 5, 234, 170, 5, 5, 125, 222, 5, 5, 200, 164, 5, 5, 32, 189, 5, 5, 210, 232, 5, 5, 75, 245, 5, 5, 50, 168, 5, 5, 252, 175, 5, 5, 135, 247, 5, 5, 133, 244, 5, 5, 8, 243, 5, 5, 102, 192, 5, 5, 83, 175, 5, 5, 178, 228, 5, 5, 80, 181, 5, 5, 81, 181, 5, 5, 95, 198, 5, 5, 13, 225, 5, 5, 8, 229, 5, 5, 109, 211, 5, 5, 152, 228, 5, 5, 130, 224, 5, 5, 131, 224, 5, 5, 251, 230, 5, 5, 27, 229, 5, 5, 130, 236, 5, 5, 166, 228, 5, 5, 24, 250, 5, 5, 199, 229, 5, 5, 2, 190, 5, 5, 152, 174, 5, 5, 248, 249, 5, 5, 186, 164, 5, 5, 254, 173, 5, 5, 117, 184, 5, 5, 71, 181, 5, 5, 28, 242, 5, 5, 66, 203, 5, 5, 133, 228, 5, 5, 82, 219, 5, 5, 159, 177, 5, 5, 61, 174, 5, 5, 95, 183, 5, 5, 123, 177, 5, 5, 157, 168, 5, 5, 129, 181, 5, 5, 136, 199, 5, 5, 165, 181, 5, 5, 236, 220, 5, 5, 31, 232, 5, 5, 222, 170, 5, 5, 5, 165, 5, 5, 165, 208, 5, 5, 100, 188, 5, 5, 215, 217, 5, 5, 33, 188, 5, 5, 39, 225, 5, 5, 69, 172, 5, 5, 135, 249, 5, 5, 52, 173, 5, 5, 87, 174, 5, 5, 214, 192, 5, 5, 75, 181, 5, 5, 214, 203, 5, 5, 215, 203, 5, 5, 102, 212, 5, 5, 120, 246, 5, 5, 88, 174, 5, 5, 76, 181, 5, 5, 70, 172, 5, 5, 53, 178, 5, 5, 86, 167, 5, 5, 11, 206, 5, 5, 140, 245, 5, 5, 25, 175, 5, 5, 102, 183, 5, 5, 37, 188, 5, 5, 125, 164, 5, 5, 216, 220, 5, 5, 213, 225, 5, 5, 72, 172, 5, 5, 73, 172, 5, 5, 238, 233, 5, 5, 79, 181, 5, 5, 222, 191, 5, 5, 24, 165, 5, 5, 26, 175, 5, 5, 225, 218, 5, 5, 218, 237, 5, 5, 189, 208, 5, 5, 108, 191, 5, 5, 54, 167, 5, 5, 215, 240, 5, 5, 240, 172, 5, 5, 141, 199, 5, 5, 171, 227, 5, 5, 201, 247, 5, 5, 75, 172, 5, 5, 59, 208, 5, 5, 74, 240, 5, 5, 39, 188, 5, 5, 172, 227, 5, 5, 40, 177, 5, 5, 25, 224, 5, 5, 142, 199, 5, 5, 73, 176, 5, 5, 96, 222, 5, 5, 234, 201, 5, 5, 220, 246, 5, 5, 15, 206, 5, 5, 125, 225, 5, 5, 47, 225, 5, 5, 153, 182, 5, 5, 90, 185, 5, 5, 246, 201, 5, 5, 138, 223, 5, 5, 63, 188, 5, 5, 71, 177, 5, 5, 123, 182, 5, 5, 33, 189, 5, 5, 29, 206, 5, 5, 31, 206, 5, 5, 46, 224, 5, 5, 52, 168, 5, 5, 63, 177, 5, 5, 65, 177, 5, 5, 22, 168, 5, 5, 140, 203, 5, 5, 28, 176, 5, 5, 79, 229, 5, 5, 80, 229, 5, 5, 8, 219, 5, 5, 185, 239, 5, 5, 255, 182, 5, 5, 15, 183, 5, 5, 192, 177, 5, 5, 229, 164, 5, 5, 97, 203, 5, 5, 224, 187, 5, 5, 187, 192, 5, 5, 139, 191, 5, 5, 159, 204, 5, 5, 36, 220, 5, 5, 229, 213, 5, 5, 88, 228, 5, 5, 183, 216, 5, 5, 221, 184, 5, 5, 41, 187, 5, 5, 25, 244, 5, 5, 221, 224, 5, 5, 38, 242, 5, 5, 0, 140, 5, 5, 216, 203, 5, 5, 132, 207, 5, 5, 170, 208, 5, 5, 170, 213, 5, 5, 231, 234, 5, 5, 11, 165, 5, 5, 50, 240, 5, 5, 48, 187, 5, 5, 142, 233, 5, 5, 122, 226, 5, 5, 196, 173, 5, 5, 223, 221, 5, 5, 29, 167, 5, 5, 245, 167, 5, 5, 118, 192, 5, 5, 6, 195, 5, 5, 156, 233, 5, 5, 46, 217, 5, 5, 74, 218, 5, 5, 8, 176, 5, 5, 217, 220, 5, 5, 125, 172, 5, 5, 165, 218, 5, 5, 132, 180, 5, 5, 201, 206, 5, 5, 61, 202, 5, 5, 206, 172, 5, 5, 202, 206, 5, 5, 54, 228, 5, 5, 18, 203, 5, 5, 135, 217, 5, 5, 113, 219, 5, 5, 89, 166, 5, 5, 232, 183, 5, 5, 117, 221, 5, 5, 212, 249, 5, 5, 154, 206, 5, 5, 67, 212, 5, 5, 133, 180, 5, 5, 157, 182, 5, 5, 165, 239, 5, 5, 103, 203, 5, 5, 67, 246, 5, 5, 50, 200, 5, 5, 231, 179, 5, 5, 65, 244, 5, 5, 95, 223, 5, 5, 226, 218, 5, 5, 164, 198, 5, 5, 13, 241, 5, 5, 7, 170, 5, 5, 209, 175, 5, 5, 165, 206, 5, 5, 141, 214, 5, 5, 217, 182, 5, 5, 190, 181, 5, 5, 205, 170, 5, 5, 27, 168, 5, 5, 142, 214, 5, 5, 58, 237, 5, 5, 39, 174, 5, 5, 82, 187, 5, 5, 243, 218, 5, 5, 181, 234, 5, 5, 69, 178, 5, 5, 108, 245, 5, 5, 136, 227, 5, 5, 176, 185, 5, 5, 27, 228, 5, 5, 227, 200, 5, 5, 101, 215, 5, 5, 108, 240, 5, 5, 183, 164, 5, 5, 50, 221, 5, 5, 87, 194, 5, 5, 251, 179, 5, 5, 213, 172, 5, 5, 24, 169, 5, 5, 229, 188, 5, 5, 17, 223, 5, 5, 194, 202, 5, 5, 109, 181, 5, 5, 208, 238, 5, 5, 102, 223, 5, 5, 43, 165, 5, 5, 44, 165, 5, 5, 240, 171, 5, 5, 90, 193, 5, 5, 124, 178, 5, 5, 150, 237, 5, 5, 80, 246, 5, 5, 151, 247, 5, 5, 246, 234, 5, 5, 15, 241, 5, 5, 176, 221, 5, 5, 155, 234, 5, 5, 241, 175, 5, 5, 218, 228, 5, 5, 235, 222, 5, 5, 107, 223, 5, 5, 117, 248, 5, 5, 34, 198, 5, 5, 214, 173, 5, 5, 244, 184, 5, 5, 45, 244, 5, 5, 201, 224, 5, 5, 218, 181, 5, 5, 29, 248, 5, 5, 150, 193, 5, 5, 33, 247, 5, 5, 236, 200, 5, 5, 34, 196, 5, 5, 84, 203, 5, 5, 143, 243, 5, 5, 123, 240, 5, 5, 11, 186, 5, 5, 35, 236, 5, 5, 174, 207, 5, 5, 112, 223, 5, 5, 251, 212, 5, 5, 75, 194, 5, 5, 136, 240, 5, 5, 8, 247, 5, 5, 95, 202, 5, 5, 128, 226, 5, 5, 76, 164, 5, 5, 110, 200, 5, 5, 56, 188, 5, 5, 220, 223, 5, 5, 31, 228, 5, 5, 224, 184, 5, 5, 133, 177, 5, 5, 52, 175, 5, 5, 133, 165, 5, 5, 236, 179, 5, 5, 55, 221, 5, 5, 159, 165, 5, 5, 224, 228, 5, 5, 163, 234, 5, 5, 125, 190, 5, 5, 158, 240, 5, 5, 148, 233, 5, 5, 12, 224, 5, 5, 56, 221, 5, 5, 202, 219, 5, 5, 78, 236, 5, 5, 240, 235, 5, 5, 11, 212, 5, 5, 112, 177, 5, 5, 13, 186, 5, 5, 14, 186, 5, 5, 140, 169, 5, 5, 247, 210, 5, 5, 222, 243, 5, 5, 217, 244, 5, 5, 67, 201, 5, 5, 144, 174, 5, 5, 246, 187, 5, 5, 249, 190, 5, 5, 175, 209, 5, 5, 65, 188, 5, 5, 184, 185, 5, 5, 124, 215, 5, 5, 218, 172, 5, 5, 251, 175, 5, 5, 247, 243, 5, 5, 170, 197, 5, 5, 42, 238, 5, 5, 81, 164, 5, 5, 125, 215, 5, 5, 76, 194, 5, 5, 57, 194, 5, 5, 210, 184, 5, 5, 153, 248, 5, 5, 41, 246, 5, 5, 178, 240, 5, 5, 231, 249, 5, 5, 33, 168, 5, 5, 133, 185, 5, 5, 119, 187, 5, 5, 120, 229, 5, 5, 59, 176, 5, 5, 19, 189, 5, 5, 183, 177, 5, 5, 20, 189, 5, 5, 75, 185, 5, 5, 134, 165, 5, 5, 233, 202, 5, 5, 25, 223, 5, 5, 79, 244, 5, 5, 155, 217, 5, 5, 122, 203, 5, 5, 102, 221, 5, 5, 163, 169, 5, 5, 145, 187, 5, 5, 227, 199, 5, 5, 83, 185, 5, 5, 25, 202, 5, 5, 139, 177, 5, 5, 156, 217, 5, 5, 222, 244, 5, 5, 231, 243, 5, 5, 229, 211, 5, 5, 240, 249, 5, 5, 240, 179, 5, 5, 143, 177, 5, 5, 131, 190, 5, 5, 182, 190, 5, 5, 183, 190, 5, 5, 36, 163, 5, 5, 40, 163, 5, 5, 9, 194, 5, 5, 0, 161, 5, 5, 114, 172, 5, 5, 122, 176, 5, 5, 116, 172, 5, 5, 155, 163, 5, 5, 156, 163, 5, 5, 89, 209, 5, 5, 50, 202, 5, 5, 78, 232, 5, 5, 79, 232, 5, 5, 56, 170, 5, 5, 67, 175, 5, 5, 153, 166, 5, 5, 99, 203, 5, 5, 38, 201, 5, 5, 212, 193, 5, 5, 205, 233, 5, 5, 102, 189, 5, 5, 255, 191, 5, 5, 110, 169, 5, 5, 62, 202, 5, 5, 38, 204, 5, 5, 189, 167, 5, 5, 5, 192, 5, 5, 36, 238, 5, 5, 163, 237, 5, 5, 3, 198, 5, 5, 133, 198, 5, 5, 34, 165, 5, 5, 217, 246, 5, 5, 29, 224, 5, 5, 178, 210, 5, 5, 231, 212, 5, 5, 41, 238, 5, 5, 206, 214, 5, 5, 11, 198, 5, 5, 102, 201, 5, 5, 12, 198, 5, 5, 7, 232, 5, 5, 228, 234, 5, 5, 14, 198, 5, 5, 209, 214, 5, 5, 123, 241, 5, 5, 183, 166, 5, 5, 15, 198, 5, 5, 72, 204, 5, 5, 82, 238, 5, 5, 136, 216, 5, 5, 105, 201, 5, 5, 136, 187, 5, 5, 90, 168, 5, 5, 149, 197, 5, 5, 114, 204, 5, 5, 234, 167, 5, 5, 134, 192, 5, 5, 30, 167, 5, 5, 53, 220, 5, 5, 25, 208, 5, 5, 28, 210, 5, 5, 25, 235, 5, 5, 25, 212, 5, 5, 74, 219, 5, 5, 62, 193, 5, 5, 30, 208, 5, 5, 31, 208, 5, 5, 64, 220, 5, 5, 77, 219, 5, 5, 69, 193, 5, 5, 218, 219, 5, 5, 9, 178, 5, 5, 249, 171, 5, 5, 4, 209, 5, 5, 155, 244, 5, 5, 167, 247, 5, 5, 184, 163, 5, 5, 211, 219, 5, 5, 205, 222, 5, 5, 100, 167, 5, 5, 97, 197, 5, 5, 157, 244, 5, 5, 107, 197, 5, 5, 65, 240, 5, 5, 45, 163, 5, 5, 205, 246, 5, 5, 108, 197, 5, 5, 20, 239, 5, 5, 110, 220, 5, 5, 169, 242, 5, 5, 203, 204, 5, 5, 112, 197, 5, 5, 149, 247, 5, 5, 105, 196, 5, 5, 59, 173, 5, 5, 238, 191, 5, 5, 159, 175, 5, 5, 59, 167, 5, 5, 169, 174, 5, 5, 29, 210, 5, 5, 44, 229, 5, 5, 160, 175, 5, 5, 29, 190, 5, 5, 118, 220, 5, 5, 229, 170, 5, 5, 199, 186, 5, 5, 197, 205, 5, 5, 27, 217, 5, 5, 227, 204, 5, 5, 66, 167, 5, 5, 113, 194, 5, 5, 217, 238, 5, 5, 199, 189, 5, 5, 240, 216, 5, 5, 26, 220, 5, 5, 116, 176, 5, 5, 220, 238, 5, 5, 74, 193, 5, 5, 137, 187, 5, 5, 83, 172, 5, 5, 193, 177, 5, 5, 154, 192, 5, 5, 184, 216, 5, 5, 191, 170, 5, 5, 109, 247, 5, 5, 102, 218, 5, 5, 33, 209, 5, 5, 85, 249, 5, 5, 77, 181, 5, 5, 40, 216, 5, 5, 200, 178, 5, 5, 251, 215, 5, 5, 118, 213, 5, 5, 85, 167, 5, 5, 197, 173, 5, 5, 208, 178, 5, 5, 230, 171, 5, 5, 30, 172, 5, 5, 36, 229, 5, 5, 55, 227, 5, 5, 136, 249, 5, 5, 196, 171, 5, 5, 52, 167, 5, 5, 173, 173, 5, 5, 156, 171, 5, 5, 230, 175, 5, 5, 197, 169, 5, 5, 61, 170, 5, 5, 186, 195, 5, 5, 189, 215, 5, 5, 136, 195, 5, 5, 60, 227, 5, 5, 7, 244, 5, 5, 75, 221, 5, 5, 218, 241, 5, 5, 153, 228, 5, 5, 193, 199, 5, 5, 49, 175, 5, 5, 218, 220, 5, 5, 30, 197, 5, 5, 203, 206, 5, 5, 210, 228, 5, 5, 160, 239, 5, 5, 37, 229, 5, 5, 2, 170, 5, 5, 111, 169, 5, 5, 224, 221, 5, 5, 169, 220, 5, 5, 170, 220, 5, 5, 175, 171, 5, 5, 61, 227, 5, 5, 252, 248, 5, 5, 32, 177, 5, 5, 206, 183, 5, 5, 198, 231, 5, 5, 199, 231, 5, 5, 254, 223, 5, 5, 96, 219, 5, 5, 163, 217, 5, 5, 69, 166, 5, 5, 76, 171, 5, 5, 234, 190, 5, 5, 252, 236, 5, 5, 177, 182, 5, 5, 32, 171, 5, 5, 69, 177, 5, 5, 238, 176, 5, 5, 4, 183, 5, 5, 34, 172, 5, 5, 166, 217, 5, 5, 10, 223, 5, 5, 31, 197, 5, 5, 250, 207, 5, 5, 89, 170, 5, 5, 246, 220, 5, 5, 54, 181, 5, 5, 35, 244, 5, 5, 225, 242, 5, 5, 115, 169, 5, 5, 249, 222, 5, 5, 112, 169, 5, 5, 184, 219, 5, 5, 82, 221, 5, 5, 211, 169, 5, 5, 207, 236, 5, 5, 8, 228, 5, 5, 134, 198, 5, 5, 228, 200, 5, 5, 247, 176, 5, 5, 143, 225, 5, 5, 182, 234, 5, 5, 251, 207, 5, 5, 164, 237, 5, 5, 128, 170, 5, 5, 205, 215, 5, 5, 221, 212, 5, 5, 243, 196, 5, 5, 66, 185, 5, 5, 32, 205, 5, 5, 82, 228, 5, 5, 110, 181, 5, 5, 148, 194, 5, 5, 221, 246, 5, 5, 163, 186, 5, 5, 81, 227, 5, 5, 149, 212, 5, 5, 201, 187, 5, 5, 2, 241, 5, 5, 98, 206, 5, 5, 91, 221, 5, 5, 190, 186, 5, 5, 67, 185, 5, 5, 31, 237, 5, 5, 2, 208, 5, 5, 215, 229, 5, 5, 2, 237, 5, 5, 235, 177, 5, 5, 45, 165, 5, 5, 43, 168, 5, 5, 97, 172, 5, 5, 254, 200, 5, 5, 213, 183, 5, 5, 30, 190, 5, 5, 109, 240, 5, 5, 169, 203, 5, 5, 22, 234, 5, 5, 39, 220, 5, 5, 221, 247, 5, 5, 2, 201, 5, 5, 53, 169, 5, 5, 232, 213, 5, 5, 159, 216, 5, 5, 25, 200, 5, 5, 150, 194, 5, 5, 37, 232, 5, 5, 118, 226, 5, 5, 228, 204, 5, 5, 11, 170, 5, 5, 27, 178, 5, 5, 35, 198, 5, 5, 111, 217, 5, 5, 120, 181, 5, 5, 130, 196, 5, 5, 11, 192, 5, 5, 14, 231, 5, 5, 76, 170, 5, 5, 28, 217, 5, 5, 132, 172, 5, 5, 205, 245, 5, 5, 151, 171, 5, 5, 59, 224, 5, 5, 163, 196, 5, 5, 33, 171, 5, 5, 130, 247, 5, 5, 156, 195, 5, 5, 110, 176, 5, 5, 173, 224, 5, 5, 23, 195, 5, 5, 69, 236, 5, 5, 180, 213, 5, 5, 152, 212, 5, 5, 136, 169, 5, 5, 45, 172, 5, 5, 114, 170, 5, 5, 217, 229, 5, 5, 8, 232, 5, 5, 57, 221, 5, 5, 159, 240, 5, 5, 86, 177, 5, 5, 140, 195, 5, 5, 141, 169, 5, 5, 232, 229, 5, 5, 64, 235, 5, 5, 225, 185, 5, 5, 22, 223, 5, 5, 115, 170, 5, 5, 86, 168, 5, 5, 75, 225, 5, 5, 27, 194, 5, 5, 235, 170, 5, 5, 244, 200, 5, 5, 39, 231, 5, 5, 63, 168, 5, 5, 19, 213, 5, 5, 177, 209, 5, 5, 236, 168, 5, 5, 202, 221, 5, 5, 141, 176, 5, 5, 199, 228, 5, 5, 234, 229, 5, 5, 14, 188, 5, 5, 202, 223, 5, 5, 126, 215, 5, 5, 191, 195, 5, 5, 124, 241, 5, 5, 37, 217, 5, 5, 204, 182, 5, 5, 96, 170, 5, 5, 245, 188, 5, 5, 54, 194, 5, 5, 133, 205, 5, 5, 127, 241, 5, 5, 83, 238, 5, 5, 131, 200, 5, 5, 96, 220, 5, 5, 56, 205, 5, 5, 75, 193, 5, 5, 248, 219, 5, 5, 218, 211, 5, 5, 229, 169, 5, 5, 61, 224, 5, 5, 37, 236, 5, 5, 135, 184, 5, 5, 204, 217, 5, 5, 254, 241, 5, 5, 217, 199, 5, 5, 249, 170, 5, 5, 187, 171, 5, 5, 147, 168, 5, 5, 188, 171, 5, 5, 138, 187, 5, 5, 43, 197, 5, 5, 196, 207, 5, 5, 128, 217, 5, 5, 189, 171, 5, 5, 228, 199, 5, 5, 215, 189, 5, 5, 254, 170, 5, 5, 11, 191, 5, 5, 122, 194, 5, 5, 135, 170, 5, 5, 102, 213, 5, 5, 237, 225, 5, 5, 123, 180, 5, 5, 193, 195, 5, 5, 206, 221, 5, 5, 20, 170, 5, 5, 167, 205, 5, 5, 222, 188, 5, 5, 93, 248, 5, 5, 185, 229, 5, 5, 251, 231, 5, 5, 208, 207, 5, 5, 10, 227, 5, 5, 216, 239, 5, 5, 184, 228, 5, 5, 64, 218, 5, 5, 27, 198, 5, 5, 186, 228, 5, 5, 205, 244, 5, 5, 22, 207, 5, 5, 13, 220, 5, 5, 39, 204, 5, 5, 183, 234, 5, 5, 163, 178, 5, 5, 47, 238, 5, 5, 71, 204, 5, 5, 202, 232, 5, 5, 194, 248, 5, 5, 28, 163, 5, 5, 227, 226, 5, 5, 213, 177, 5, 5, 28, 209, 5, 5, 207, 201, 5, 5, 172, 231, 5, 5, 113, 192, 5, 5, 92, 245, 5, 5, 129, 219, 5, 5, 5, 196, 5, 5, 188, 192, 5, 5, 113, 213, 5, 5, 9, 173, 5, 5, 27, 197, 5, 5, 96, 176, 5, 5, 32, 169, 5, 5, 125, 227, 5, 5, 231, 175, 5, 5, 133, 207, 5, 5, 49, 187, 5, 5, 27, 174, 5, 5, 231, 171, 5, 5, 183, 237, 5, 5, 189, 199, 5, 5, 211, 195, 5, 5, 162, 218, 5, 5, 37, 209, 5, 5, 172, 195, 5, 5, 217, 203, 5, 5, 8, 166, 5, 5, 205, 237, 5, 5, 31, 241, 5, 5, 206, 167, 5, 5, 130, 219, 5, 5, 167, 242, 5, 5, 22, 237, 5, 5, 116, 222, 5, 5, 176, 223, 5, 5, 209, 178, 5, 5, 53, 167, 5, 5, 29, 243, 5, 5, 4, 206, 5, 5, 198, 173, 5, 5, 129, 172, 5, 5, 113, 169, 5, 5, 97, 219, 5, 5, 204, 206, 5, 5, 67, 248, 5, 5, 7, 195, 5, 5, 30, 182, 5, 5, 31, 182, 5, 5, 212, 176, 5, 5, 199, 219, 5, 5, 171, 220, 5, 5, 208, 237, 5, 5, 96, 174, 5, 5, 74, 172, 5, 5, 225, 221, 5, 5, 235, 184, 5, 5, 195, 165, 5, 5, 149, 184, 5, 5, 161, 168, 5, 5, 139, 167, 5, 5, 103, 189, 5, 5, 31, 167, 5, 5, 23, 185, 5, 5, 199, 166, 5, 5, 46, 163, 5, 5, 200, 166, 5, 5, 154, 228, 5, 5, 220, 231, 5, 5, 58, 238, 5, 5, 62, 187, 5, 5, 235, 190, 5, 5, 174, 182, 5, 5, 28, 250, 5, 5, 63, 187, 5, 5, 48, 197, 5, 5, 188, 225, 5, 5, 246, 195, 5, 5, 68, 212, 5, 5, 141, 221, 5, 5, 4, 197, 5, 5, 220, 193, 5, 5, 219, 237, 5, 5, 50, 183, 5, 5, 85, 209, 5, 5, 16, 190, 5, 5, 68, 246, 5, 5, 59, 238, 5, 5, 206, 191, 5, 5, 201, 215, 5, 5, 151, 216, 5, 5, 173, 242, 5, 5, 19, 179, 5, 5, 142, 167, 5, 5, 209, 212, 5, 5, 219, 185, 5, 5, 153, 201, 5, 5, 184, 229, 5, 5, 196, 213, 5, 5, 130, 209, 5, 5, 147, 211, 5, 5, 18, 246, 5, 5, 162, 209, 5, 5, 75, 240, 5, 5, 248, 195, 5, 5, 143, 214, 5, 5, 20, 212, 5, 5, 157, 224, 5, 5, 115, 186, 5, 5, 119, 201, 5, 5, 202, 247, 5, 5, 144, 214, 5, 5, 201, 213, 5, 5, 214, 170, 5, 5, 140, 190, 5, 5, 11, 250, 5, 5, 179, 223, 5, 5, 195, 223, 5, 5, 201, 245, 5, 5, 40, 188, 5, 5, 224, 244, 5, 5, 206, 170, 5, 5, 21, 212, 5, 5, 45, 243, 5, 5, 137, 227, 5, 5, 230, 196, 5, 5, 94, 248, 5, 5, 99, 223, 5, 5, 160, 199, 5, 5, 232, 189, 5, 5, 212, 224, 5, 5, 33, 205, 5, 5, 223, 184, 5, 5, 101, 237, 5, 5, 46, 222, 5, 5, 250, 195, 5, 5, 160, 167, 5, 5, 142, 242, 5, 5, 98, 222, 5, 5, 58, 197, 5, 5, 121, 225, 5, 5, 203, 208, 5, 5, 145, 233, 5, 5, 139, 211, 5, 5, 230, 188, 5, 5, 29, 208, 5, 5, 54, 163, 5, 5, 222, 177, 5, 5, 241, 171, 5, 5, 83, 177, 5, 5, 223, 204, 5, 5, 160, 197, 5, 5, 31, 190, 5, 5, 65, 238, 5, 5, 32, 237, 5, 5, 26, 212, 5, 5, 154, 186, 5, 5, 189, 163, 5, 5, 186, 243, 5, 5, 140, 178, 5, 5, 64, 245, 5, 5, 204, 208, 5, 5, 218, 190, 5, 5, 222, 246, 5, 5, 153, 177, 5, 5, 55, 183, 5, 5, 231, 163, 5, 5, 47, 209, 5, 5, 213, 224, 5, 5, 33, 231, 5, 5, 186, 242, 5, 5, 52, 177, 5, 5, 224, 177, 5, 5, 248, 177, 5, 5, 40, 182, 5, 5, 178, 203, 5, 5, 112, 244, 5, 5, 221, 165, 5, 5, 68, 195, 5, 5, 57, 165, 5, 5, 135, 246, 5, 5, 188, 229, 5, 5, 192, 220, 5, 5, 173, 178, 5, 5, 209, 213, 5, 5, 107, 176, 5, 5, 177, 180, 5, 5, 74, 183, 5, 5, 193, 167, 5, 5, 48, 205, 5, 5, 3, 177, 5, 5, 45, 180, 5, 5, 154, 223, 5, 5, 122, 181, 5, 5, 51, 209, 5, 5, 221, 178, 5, 5, 215, 173, 5, 5, 174, 217, 5, 5, 32, 212, 5, 5, 33, 212, 5, 5, 37, 169, 5, 5, 210, 194, 5, 5, 81, 184, 5, 5, 183, 236, 5, 5, 171, 237, 5, 5, 184, 233, 5, 5, 57, 188, 5, 5, 136, 235, 5, 5, 89, 180, 5, 5, 170, 164, 5, 5, 122, 199, 5, 5, 223, 167, 5, 5, 93, 246, 5, 5, 118, 236, 5, 5, 125, 185, 5, 5, 175, 234, 5, 5, 91, 181, 5, 5, 211, 213, 5, 5, 252, 212, 5, 5, 142, 188, 5, 5, 67, 167, 5, 5, 34, 246, 5, 5, 173, 249, 5, 5, 125, 178, 5, 5, 152, 219, 5, 5, 24, 195, 5, 5, 237, 184, 5, 5, 122, 222, 5, 5, 143, 211, 5, 5, 140, 194, 5, 5, 126, 222, 5, 5, 97, 224, 5, 5, 58, 209, 5, 5, 118, 238, 5, 5, 148, 167, 5, 5, 105, 188, 5, 5, 243, 189, 5, 5, 190, 195, 5, 5, 137, 226, 5, 5, 216, 224, 5, 5, 220, 243, 5, 5, 126, 185, 5, 5, 160, 165, 5, 5, 59, 209, 5, 5, 77, 169, 5, 5, 60, 209, 5, 5, 222, 234, 5, 5, 113, 233, 5, 5, 144, 211, 5, 5, 13, 226, 5, 5, 224, 190, 5, 5, 6, 202, 5, 5, 112, 228, 5, 5, 185, 185, 5, 5, 216, 211, 5, 5, 203, 232, 5, 5, 142, 176, 5, 5, 134, 205, 5, 5, 152, 213, 5, 5, 246, 188, 5, 5, 228, 177, 5, 5, 12, 180, 5, 5, 68, 201, 5, 5, 69, 167, 5, 5, 227, 231, 5, 5, 150, 209, 5, 5, 160, 176, 5, 5, 113, 187, 5, 5, 157, 205, 5, 5, 12, 204, 5, 5, 167, 213, 5, 5, 39, 221, 5, 5, 178, 209, 5, 5, 194, 178, 5, 5, 57, 205, 5, 5, 154, 248, 5, 5, 108, 222, 5, 5, 143, 197, 5, 5, 72, 195, 5, 5, 10, 232, 5, 5, 96, 185, 5, 5, 48, 195, 5, 5, 197, 225, 5, 5, 128, 241, 5, 5, 37, 212, 5, 5, 117, 176, 5, 5, 244, 239, 5, 5, 30, 206, 5, 5, 231, 167, 5, 5, 232, 167, 5, 5, 155, 248, 5, 5, 131, 168, 5, 5, 9, 238, 5, 5, 86, 238, 5, 5, 32, 195, 5, 5, 251, 204, 5, 5, 161, 248, 5, 5, 38, 231, 5, 5, 250, 170, 5, 5, 33, 167, 5, 5, 197, 233, 5, 5, 133, 163, 5, 5, 91, 244, 5, 5, 58, 169, 5, 5, 227, 233, 5, 5, 129, 215, 5, 5, 44, 197, 5, 5, 101, 185, 5, 5, 136, 185, 5, 5, 46, 219, 5, 5, 139, 187, 5, 5, 140, 187, 5, 5, 33, 173, 5, 5, 41, 212, 5, 5, 34, 167, 5, 5, 144, 244, 5, 5, 146, 187, 5, 5, 12, 191, 5, 5, 13, 191, 5, 5, 150, 169, 5, 5, 137, 163, 5, 5, 244, 229, 5, 5, 123, 194, 5, 5, 53, 186, 5, 5, 143, 232, 5, 5, 209, 242, 5, 5, 30, 227, 5, 5, 24, 227, 5, 5, 248, 186, 5, 5, 177, 230, 5, 5, 188, 227, 5, 5, 206, 248, 5, 5, 168, 235, 5, 5, 166, 208, 5, 5, 34, 179, 5, 5, 44, 169, 5, 5, 58, 214, 5, 5, 145, 241, 5, 5, 46, 215, 5, 5, 171, 208, 5, 5, 172, 208, 5, 5, 163, 218, 5, 5, 176, 164, 5, 5, 137, 178, 5, 5, 243, 222, 5, 5, 128, 230, 5, 5, 97, 239, 5, 5, 245, 226, 5, 5, 109, 242, 5, 5, 211, 218, 5, 5, 24, 181, 5, 5, 97, 174, 5, 5, 204, 174, 5, 5, 94, 171, 5, 5, 235, 232, 5, 5, 75, 218, 5, 5, 61, 237, 5, 5, 3, 170, 5, 5, 12, 233, 5, 5, 166, 211, 5, 5, 70, 199, 5, 5, 226, 221, 5, 5, 179, 228, 5, 5, 23, 237, 5, 5, 173, 233, 5, 5, 212, 223, 5, 5, 37, 171, 5, 5, 68, 191, 5, 5, 21, 248, 5, 5, 71, 216, 5, 5, 29, 245, 5, 5, 16, 181, 5, 5, 92, 168, 5, 5, 228, 221, 5, 5, 15, 194, 5, 5, 197, 213, 5, 5, 81, 218, 5, 5, 187, 202, 5, 5, 45, 192, 5, 5, 175, 220, 5, 5, 46, 192, 5, 5, 212, 229, 5, 5, 224, 205, 5, 5, 213, 228, 5, 5, 223, 191, 5, 5, 231, 196, 5, 5, 248, 176, 5, 5, 140, 175, 5, 5, 201, 199, 5, 5, 187, 179, 5, 5, 205, 208, 5, 5, 91, 193, 5, 5, 107, 194, 5, 5, 41, 233, 5, 5, 128, 189, 5, 5, 32, 190, 5, 5, 74, 166, 5, 5, 229, 205, 5, 5, 142, 175, 5, 5, 122, 218, 5, 5, 22, 190, 5, 5, 84, 224, 5, 5, 190, 163, 5, 5, 168, 219, 5, 5, 26, 200, 5, 5, 85, 242, 5, 5, 140, 213, 5, 5, 144, 243, 5, 5, 139, 231, 5, 5, 205, 199, 5, 5, 193, 220, 5, 5, 131, 174, 5, 5, 103, 249, 5, 5, 34, 245, 5, 5, 156, 215, 5, 5, 46, 244, 5, 5, 89, 218, 5, 5, 109, 239, 5, 5, 220, 169, 5, 5, 126, 235, 5, 5, 4, 177, 5, 5, 247, 163, 5, 5, 53, 222, 5, 5, 197, 220, 5, 5, 20, 250, 5, 5, 96, 202, 5, 5, 51, 192, 5, 5, 90, 180, 5, 5, 137, 235, 5, 5, 54, 169, 5, 5, 71, 179, 5, 5, 255, 174, 5, 5, 120, 217, 5, 5, 221, 208, 5, 5, 138, 248, 5, 5, 94, 213, 5, 5, 7, 202, 5, 5, 191, 235, 5, 5, 100, 224, 5, 5, 57, 176, 5, 5, 58, 194, 5, 5, 47, 237, 5, 5, 14, 187, 5, 5, 202, 184, 5, 5, 51, 171, 5, 5, 154, 175, 5, 5, 141, 187, 5, 5, 157, 230, 5, 5, 242, 248, 5, 5, 201, 180, 5, 5, 163, 232, 5, 5, 161, 185, 5, 5, 33, 230, 5, 5, 11, 248, 5, 5, 173, 247, 5, 5, 178, 186, 5, 5, 76, 214, 5, 5, 237, 210, 5, 5, 83, 197, 5, 5, 181, 181, 5, 5, 12, 215, 5, 5, 107, 235, 5, 5, 90, 230, 5, 5, 182, 181, 5, 5, 173, 231, 5, 5, 142, 223, 5, 5, 135, 192, 5, 5, 35, 240, 5, 5, 44, 215, 5, 5, 60, 166, 5, 5, 64, 184, 5, 5, 221, 209, 5, 5, 222, 209, 5, 5, 226, 222, 5, 5, 131, 173, 5, 5, 246, 197, 5, 5, 137, 249, 5, 5, 2, 195, 5, 5, 157, 171, 5, 5, 69, 218, 5, 5, 82, 206, 5, 5, 232, 164, 5, 5, 70, 218, 5, 5, 61, 248, 5, 5, 95, 171, 5, 5, 27, 196, 5, 5, 107, 164, 5, 5, 209, 167, 5, 5, 154, 183, 5, 5, 64, 187, 5, 5, 41, 242, 5, 5, 213, 232, 5, 5, 210, 167, 5, 5, 81, 163, 5, 5, 104, 189, 5, 5, 114, 169, 5, 5, 145, 225, 5, 5, 66, 184, 5, 5, 223, 210, 5, 5, 8, 244, 5, 5, 242, 232, 5, 5, 212, 218, 5, 5, 76, 218, 5, 5, 99, 229, 5, 5, 62, 207, 5, 5, 179, 231, 5, 5, 176, 199, 5, 5, 217, 201, 5, 5, 57, 225, 5, 5, 178, 182, 5, 5, 227, 218, 5, 5, 177, 199, 5, 5, 113, 197, 5, 5, 127, 243, 5, 5, 119, 231, 5, 5, 228, 218, 5, 5, 104, 203, 5, 5, 101, 230, 5, 5, 37, 178, 5, 5, 166, 239, 5, 5, 142, 221, 5, 5, 82, 200, 5, 5, 192, 184, 5, 5, 218, 174, 5, 5, 222, 212, 5, 5, 124, 195, 5, 5, 116, 186, 5, 5, 40, 174, 5, 5, 16, 218, 5, 5, 213, 229, 5, 5, 69, 191, 5, 5, 232, 196, 5, 5, 141, 192, 5, 5, 94, 206, 5, 5, 169, 180, 5, 5, 115, 225, 5, 5, 212, 181, 5, 5, 233, 227, 5, 5, 214, 229, 5, 5, 44, 241, 5, 5, 219, 190, 5, 5, 33, 190, 5, 5, 92, 193, 5, 5, 247, 232, 5, 5, 104, 230, 5, 5, 167, 222, 5, 5, 167, 164, 5, 5, 173, 163, 5, 5, 235, 201, 5, 5, 86, 218, 5, 5, 251, 218, 5, 5, 234, 196, 5, 5, 155, 186, 5, 5, 161, 167, 5, 5, 15, 195, 5, 5, 75, 166, 5, 5, 30, 224, 5, 5, 42, 233, 5, 5, 105, 248, 5, 5, 132, 236, 5, 5, 110, 191, 5, 5, 50, 216, 5, 5, 110, 239, 5, 5, 141, 213, 5, 5, 222, 165, 5, 5, 202, 202, 5, 5, 132, 174, 5, 5, 46, 168, 5, 5, 170, 184, 5, 5, 119, 203, 5, 5, 236, 222, 5, 5, 103, 233, 5, 5, 236, 216, 5, 5, 69, 225, 5, 5, 163, 175, 5, 5, 56, 247, 5, 5, 115, 215, 5, 5, 140, 231, 5, 5, 131, 248, 5, 5, 213, 202, 5, 5, 34, 217, 5, 5, 228, 220, 5, 5, 18, 219, 5, 5, 166, 175, 5, 5, 94, 203, 5, 5, 247, 201, 5, 5, 32, 192, 5, 5, 33, 192, 5, 5, 64, 193, 5, 5, 77, 231, 5, 5, 34, 237, 5, 5, 103, 231, 5, 5, 219, 202, 5, 5, 220, 195, 5, 5, 107, 172, 5, 5, 88, 234, 5, 5, 226, 169, 5, 5, 34, 219, 5, 5, 203, 202, 5, 5, 55, 190, 5, 5, 134, 236, 5, 5, 3, 202, 5, 5, 218, 244, 5, 5, 57, 247, 5, 5, 232, 209, 5, 5, 241, 235, 5, 5, 167, 197, 5, 5, 160, 240, 5, 5, 70, 201, 5, 5, 118, 209, 5, 5, 142, 202, 5, 5, 44, 237, 5, 5, 168, 214, 5, 5, 237, 168, 5, 5, 13, 204, 5, 5, 14, 222, 5, 5, 238, 222, 5, 5, 54, 233, 5, 5, 81, 191, 5, 5, 48, 237, 5, 5, 203, 184, 5, 5, 245, 235, 5, 5, 132, 200, 5, 5, 150, 230, 5, 5, 177, 184, 5, 5, 162, 248, 5, 5, 132, 168, 5, 5, 234, 202, 5, 5, 85, 182, 5, 5, 58, 247, 5, 5, 80, 225, 5, 5, 218, 199, 5, 5, 45, 197, 5, 5, 84, 191, 5, 5, 173, 165, 5, 5, 252, 235, 5, 5, 158, 230, 5, 5, 17, 222, 5, 5, 242, 186, 5, 5, 15, 235, 5, 5, 55, 217, 5, 5, 135, 245, 5, 5, 124, 204, 5, 5, 132, 212, 5, 5, 178, 212, 5, 5, 133, 212, 5, 5, 51, 197, 5, 5, 143, 247, 5, 5, 148, 228, 5, 5, 125, 206, 5, 5, 234, 241, 5, 5, 48, 226, 5, 5, 158, 222, 5, 5, 184, 181, 5, 5, 170, 185, 5, 5, 98, 209, 5, 5, 196, 219, 5, 5, 219, 241, 5, 5, 216, 165, 5, 5, 8, 164, 5, 5, 220, 241, 5, 5, 70, 166, 5, 5, 141, 245, 5, 5, 171, 221, 5, 5, 88, 204, 5, 5, 194, 209, 5, 5, 99, 177, 5, 5, 2, 245, 5, 5, 222, 193, 5, 5, 167, 241, 5, 5, 229, 218, 5, 5, 123, 226, 5, 5, 239, 241, 5, 5, 48, 181, 5, 5, 93, 204, 5, 5, 124, 226, 5, 5, 103, 208, 5, 5, 145, 214, 5, 5, 107, 225, 5, 5, 153, 195, 5, 5, 16, 235, 5, 5, 176, 245, 5, 5, 42, 176, 5, 5, 16, 196, 5, 5, 16, 195, 5, 5, 47, 222, 5, 5, 203, 216, 5, 5, 133, 236, 5, 5, 72, 194, 5, 5, 106, 196, 5, 5, 211, 175, 5, 5, 78, 222, 5, 5, 192, 217, 5, 5, 179, 245, 5, 5, 148, 245, 5, 5, 45, 241, 5, 5, 222, 247, 5, 5, 81, 207, 5, 5, 153, 212, 5, 5, 193, 241, 5, 5, 194, 241, 5, 5, 51, 181, 5, 5, 247, 223, 5, 5, 34, 248, 5, 5, 92, 181, 5, 5, 113, 177, 5, 5, 93, 207, 5, 5, 25, 188, 5, 5, 10, 213, 5, 5, 123, 245, 5, 5, 108, 196, 5, 5, 129, 226, 5, 5, 216, 219, 5, 5, 194, 196, 5, 5, 192, 196, 5, 5, 250, 190, 5, 5, 79, 225, 5, 5, 253, 234, 5, 5, 81, 168, 5, 5, 93, 232, 5, 5, 21, 235, 5, 5, 205, 209, 5, 5, 159, 212, 5, 5, 219, 175, 5, 5, 94, 232, 5, 5, 38, 191, 5, 5, 123, 209, 5, 5, 26, 187, 5, 5, 203, 225, 5, 5, 110, 227, 5, 5, 194, 174, 5, 5, 195, 210, 5, 5, 210, 225, 5, 5, 142, 201, 5, 5, 173, 208, 5, 5, 23, 198, 5, 5, 55, 229, 5, 5, 143, 173, 5, 5, 136, 229, 5, 5, 168, 171, 5, 5, 238, 211, 5, 5, 137, 220, 5, 5, 227, 221, 5, 5, 105, 189, 5, 5, 236, 190, 5, 5, 90, 166, 5, 5, 155, 206, 5, 5, 164, 206, 5, 5, 11, 221, 5, 5, 179, 163, 5, 5, 182, 235, 5, 5, 60, 208, 5, 5, 158, 201, 5, 5, 180, 182, 5, 5, 112, 237, 5, 5, 166, 206, 5, 5, 106, 177, 5, 5, 110, 240, 5, 5, 148, 229, 5, 5, 171, 206, 5, 5, 241, 211, 5, 5, 179, 203, 5, 5, 164, 168, 5, 5, 159, 178, 5, 5, 249, 227, 5, 5, 134, 222, 5, 5, 140, 234, 5, 5, 128, 176, 5, 5, 28, 235, 5, 5, 216, 245, 5, 5, 186, 206, 5, 5, 184, 195, 5, 5, 187, 206, 5, 5, 165, 221, 5, 5, 167, 221, 5, 5, 22, 185, 5, 5, 151, 167, 5, 5, 174, 247, 5, 5, 37, 223, 5, 5, 116, 192, 5, 5, 192, 170, 5, 5, 169, 221, 5, 5, 252, 246, 5, 5, 22, 178, 5, 5, 218, 201, 5, 5, 191, 214, 5, 5, 0, 144, 5, 5, 62, 237, 5, 5, 63, 202, 5, 5, 2, 211, 5, 5, 232, 219, 5, 5, 79, 237, 5, 5, 247, 240, 5, 5, 91, 222, 5, 5, 40, 249, 5, 5, 71, 199, 5, 5, 179, 198, 5, 5, 219, 214, 5, 5, 108, 234, 5, 5, 15, 165, 5, 5, 234, 164, 5, 5, 20, 205, 5, 5, 228, 239, 5, 5, 238, 178, 5, 5, 57, 228, 5, 5, 23, 229, 5, 5, 8, 175, 5, 5, 172, 176, 5, 5, 173, 221, 5, 5, 243, 182, 5, 5, 11, 223, 5, 5, 12, 193, 5, 5, 70, 187, 5, 5, 107, 209, 5, 5, 6, 172, 5, 5, 254, 163, 5, 5, 26, 181, 5, 5, 179, 235, 5, 5, 106, 174, 5, 5, 241, 233, 5, 5, 134, 164, 5, 5, 255, 166, 5, 5, 27, 175, 5, 5, 181, 231, 5, 5, 239, 176, 5, 5, 113, 189, 5, 5, 247, 231, 5, 5, 144, 184, 5, 5, 107, 174, 5, 5, 74, 192, 5, 5, 15, 173, 5, 5, 20, 203, 5, 5, 139, 192, 5, 5, 60, 173, 5, 5, 19, 227, 5, 5, 53, 170, 5, 5, 189, 166, 5, 5, 244, 180, 5, 5, 209, 204, 5, 5, 137, 212, 5, 5, 237, 190, 5, 5, 25, 172, 5, 5, 196, 165, 5, 5, 141, 190, 5, 5, 41, 194, 5, 5, 172, 211, 5, 5, 43, 223, 5, 5, 134, 243, 5, 5, 241, 241, 5, 5, 150, 215, 5, 5, 239, 164, 5, 5, 143, 212, 5, 5, 57, 196, 5, 5, 133, 176, 5, 5, 197, 214, 5, 5, 176, 205, 5, 5, 72, 212, 5, 5, 81, 237, 5, 5, 189, 230, 5, 5, 131, 172, 5, 5, 5, 197, 5, 5, 235, 221, 5, 5, 212, 182, 5, 5, 229, 201, 5, 5, 31, 175, 5, 5, 185, 198, 5, 5, 246, 233, 5, 5, 252, 222, 5, 5, 154, 195, 5, 5, 52, 185, 5, 5, 35, 165, 5, 5, 216, 227, 5, 5, 71, 176, 5, 5, 88, 221, 5, 5, 248, 233, 5, 5, 43, 174, 5, 5, 146, 249, 5, 5, 219, 181, 5, 5, 162, 166, 5, 5, 191, 186, 5, 5, 120, 186, 5, 5, 27, 207, 5, 5, 88, 194, 5, 5, 39, 205, 5, 5, 96, 198, 5, 5, 183, 182, 5, 5, 206, 213, 5, 5, 252, 218, 5, 5, 116, 225, 5, 5, 63, 237, 5, 5, 82, 227, 5, 5, 241, 221, 5, 5, 248, 232, 5, 5, 204, 214, 5, 5, 235, 227, 5, 5, 153, 181, 5, 5, 249, 233, 5, 5, 187, 242, 5, 5, 49, 192, 5, 5, 88, 196, 5, 5, 126, 192, 5, 5, 31, 224, 5, 5, 50, 197, 5, 5, 208, 192, 5, 5, 147, 249, 5, 5, 142, 237, 5, 5, 87, 210, 5, 5, 205, 214, 5, 5, 40, 205, 5, 5, 44, 167, 5, 5, 56, 185, 5, 5, 105, 232, 5, 5, 240, 164, 5, 5, 58, 165, 5, 5, 59, 165, 5, 5, 209, 191, 5, 5, 119, 179, 5, 5, 201, 193, 5, 5, 152, 247, 5, 5, 229, 204, 5, 5, 164, 175, 5, 5, 22, 193, 5, 5, 48, 223, 5, 5, 95, 221, 5, 5, 140, 211, 5, 5, 163, 227, 5, 5, 40, 220, 5, 5, 40, 190, 5, 5, 107, 163, 5, 5, 242, 201, 5, 5, 231, 188, 5, 5, 64, 237, 5, 5, 63, 208, 5, 5, 133, 174, 5, 5, 89, 219, 5, 5, 23, 193, 5, 5, 59, 197, 5, 5, 134, 210, 5, 5, 94, 210, 5, 5, 64, 208, 5, 5, 40, 226, 5, 5, 89, 194, 5, 5, 12, 249, 5, 5, 253, 233, 5, 5, 224, 232, 5, 5, 141, 163, 5, 5, 255, 231, 5, 5, 112, 209, 5, 5, 70, 182, 5, 5, 98, 190, 5, 5, 243, 175, 5, 5, 156, 214, 5, 5, 242, 230, 5, 5, 120, 205, 5, 5, 114, 217, 5, 5, 203, 165, 5, 5, 84, 237, 5, 5, 53, 216, 5, 5, 158, 215, 5, 5, 202, 168, 5, 5, 198, 220, 5, 5, 129, 216, 5, 5, 20, 199, 5, 5, 188, 239, 5, 5, 244, 221, 5, 5, 192, 248, 5, 5, 86, 226, 5, 5, 99, 191, 5, 5, 7, 245, 5, 5, 251, 178, 5, 5, 191, 229, 5, 5, 97, 202, 5, 5, 170, 203, 5, 5, 60, 201, 5, 5, 130, 216, 5, 5, 13, 170, 5, 5, 67, 173, 5, 5, 62, 226, 5, 5, 146, 222, 5, 5, 103, 168, 5, 5, 217, 173, 5, 5, 177, 221, 5, 5, 61, 201, 5, 5, 116, 209, 5, 5, 36, 198, 5, 5, 228, 209, 5, 5, 161, 171, 5, 5, 186, 210, 5, 5, 102, 193, 5, 5, 168, 196, 5, 5, 115, 227, 5, 5, 4, 186, 5, 5, 155, 249, 5, 5, 139, 197, 5, 5, 94, 180, 5, 5, 229, 248, 5, 5, 13, 224, 5, 5, 15, 192, 5, 5, 190, 172, 5, 5, 243, 230, 5, 5, 98, 221, 5, 5, 63, 192, 5, 5, 253, 183, 5, 5, 190, 210, 5, 5, 208, 214, 5, 5, 19, 235, 5, 5, 32, 191, 5, 5, 171, 167, 5, 5, 30, 178, 5, 5, 106, 187, 5, 5, 79, 165, 5, 5, 59, 222, 5, 5, 176, 225, 5, 5, 218, 227, 5, 5, 53, 223, 5, 5, 249, 221, 5, 5, 48, 211, 5, 5, 246, 190, 5, 5, 31, 186, 5, 5, 239, 204, 5, 5, 147, 166, 5, 5, 213, 186, 5, 5, 126, 200, 5, 5, 85, 165, 5, 5, 188, 176, 5, 5, 66, 235, 5, 5, 151, 242, 5, 5, 175, 167, 5, 5, 89, 234, 5, 5, 148, 222, 5, 5, 13, 168, 5, 5, 87, 168, 5, 5, 188, 175, 5, 5, 149, 248, 5, 5, 136, 171, 5, 5, 117, 182, 5, 5, 9, 232, 5, 5, 219, 227, 5, 5, 55, 194, 5, 5, 197, 239, 5, 5, 88, 199, 5, 5, 146, 224, 5, 5, 159, 195, 5, 5, 164, 234, 5, 5, 207, 168, 5, 5, 89, 232, 5, 5, 8, 202, 5, 5, 28, 194, 5, 5, 192, 236, 5, 5, 6, 186, 5, 5, 3, 234, 5, 5, 60, 190, 5, 5, 83, 184, 5, 5, 214, 236, 5, 5, 20, 235, 5, 5, 117, 177, 5, 5, 216, 236, 5, 5, 27, 185, 5, 5, 177, 178, 5, 5, 82, 189, 5, 5, 169, 214, 5, 5, 127, 163, 5, 5, 121, 171, 5, 5, 220, 227, 5, 5, 118, 176, 5, 5, 250, 187, 5, 5, 253, 175, 5, 5, 184, 177, 5, 5, 181, 174, 5, 5, 58, 205, 5, 5, 18, 172, 5, 5, 175, 196, 5, 5, 116, 175, 5, 5, 17, 204, 5, 5, 129, 176, 5, 5, 74, 201, 5, 5, 85, 237, 5, 5, 186, 185, 5, 5, 124, 232, 5, 5, 175, 214, 5, 5, 68, 190, 5, 5, 106, 227, 5, 5, 125, 232, 5, 5, 70, 209, 5, 5, 216, 166, 5, 5, 71, 209, 5, 5, 217, 166, 5, 5, 147, 192, 5, 5, 101, 207, 5, 5, 46, 178, 5, 5, 69, 190, 5, 5, 199, 242, 5, 5, 142, 163, 5, 5, 77, 176, 5, 5, 15, 236, 5, 5, 4, 222, 5, 5, 255, 190, 5, 5, 103, 210, 5, 5, 162, 188, 5, 5, 148, 166, 5, 5, 70, 220, 5, 5, 79, 185, 5, 5, 117, 227, 5, 5, 46, 186, 5, 5, 178, 196, 5, 5, 249, 247, 5, 5, 237, 248, 5, 5, 236, 197, 5, 5, 206, 220, 5, 5, 131, 179, 5, 5, 241, 180, 5, 5, 12, 184, 5, 5, 47, 224, 5, 5, 114, 202, 5, 5, 219, 240, 5, 5, 72, 209, 5, 5, 146, 180, 5, 5, 154, 222, 5, 5, 72, 220, 5, 5, 254, 172, 5, 5, 75, 236, 5, 5, 71, 220, 5, 5, 198, 210, 5, 5, 74, 222, 5, 5, 148, 180, 5, 5, 78, 190, 5, 5, 216, 189, 5, 5, 72, 205, 5, 5, 72, 182, 5, 5, 203, 244, 5, 5, 172, 248, 5, 5, 84, 185, 5, 5, 15, 191, 5, 5, 128, 209, 5, 5, 168, 188, 5, 5, 54, 186, 5, 5, 112, 210, 5, 5, 243, 248, 5, 5, 224, 222, 5, 5, 230, 211, 5, 5, 169, 188, 5, 5, 213, 165, 5, 5, 181, 178, 5, 5, 63, 233, 5, 5, 243, 186, 5, 5, 153, 235, 5, 5, 117, 190, 5, 5, 204, 244, 5, 5, 245, 248, 5, 5, 76, 222, 5, 5, 182, 248, 5, 5, 68, 171, 5, 5, 165, 176, 5, 5, 10, 173, 5, 5, 141, 228, 5, 5, 131, 219, 5, 5, 149, 228, 5, 5, 214, 184, 5, 5, 98, 197, 5, 5, 144, 185, 5, 5, 47, 215, 5, 5, 175, 231, 5, 5, 151, 183, 5, 5, 36, 235, 5, 5, 58, 221, 5, 5, 71, 210, 5, 5, 204, 219, 5, 5, 235, 164, 5, 5, 91, 164, 5, 5, 101, 245, 5, 5, 128, 227, 5, 5, 212, 195, 5, 5, 52, 230, 5, 5, 125, 173, 5, 5, 16, 165, 5, 5, 233, 234, 5, 5, 253, 243, 5, 5, 136, 219, 5, 5, 67, 223, 5, 5, 147, 179, 5, 5, 98, 199, 5, 5, 0, 145, 5, 5, 71, 187, 5, 5, 121, 192, 5, 5, 67, 207, 5, 5, 202, 166, 5, 5, 116, 169, 5, 5, 144, 234, 5, 5, 80, 220, 5, 5, 57, 220, 5, 5, 179, 186, 5, 5, 152, 236, 5, 5, 55, 181, 5, 5, 72, 187, 5, 5, 223, 212, 5, 5, 157, 236, 5, 5, 159, 201, 5, 5, 181, 182, 5, 5, 128, 231, 5, 5, 242, 233, 5, 5, 220, 248, 5, 5, 17, 220, 5, 5, 228, 188, 5, 5, 178, 222, 5, 5, 162, 163, 5, 5, 221, 185, 5, 5, 134, 173, 5, 5, 38, 239, 5, 5, 173, 211, 5, 5, 178, 197, 5, 5, 47, 189, 5, 5, 210, 215, 5, 5, 120, 174, 5, 5, 48, 189, 5, 5, 96, 234, 5, 5, 99, 206, 5, 5, 6, 187, 5, 5, 223, 248, 5, 5, 41, 190, 5, 5, 253, 206, 5, 5, 210, 208, 5, 5, 247, 234, 5, 5, 175, 163, 5, 5, 20, 221, 5, 5, 232, 188, 5, 5, 146, 233, 5, 5, 103, 206, 5, 5, 183, 193, 5, 5, 161, 199, 5, 5, 90, 236, 5, 5, 91, 236, 5, 5, 91, 180, 5, 5, 223, 165, 5, 5, 42, 230, 5, 5, 218, 171, 5, 5, 37, 198, 5, 5, 237, 216, 5, 5, 147, 233, 5, 5, 21, 199, 5, 5, 121, 205, 5, 5, 123, 232, 5, 5, 188, 233, 5, 5, 204, 186, 5, 5, 95, 214, 5, 5, 52, 174, 5, 5, 171, 184, 5, 5, 242, 163, 5, 5, 66, 193, 5, 5, 11, 215, 5, 5, 16, 192, 5, 5, 95, 180, 5, 5, 47, 210, 5, 5, 96, 203, 5, 5, 33, 234, 5, 5, 182, 222, 5, 5, 98, 180, 5, 5, 45, 171, 5, 5, 93, 236, 5, 5, 139, 173, 5, 5, 63, 232, 5, 5, 222, 195, 5, 5, 68, 206, 5, 5, 78, 169, 5, 5, 94, 236, 5, 5, 95, 236, 5, 5, 176, 184, 5, 5, 25, 213, 5, 5, 54, 189, 5, 5, 69, 206, 5, 5, 98, 196, 5, 5, 145, 163, 5, 5, 196, 200, 5, 5, 167, 229, 5, 5, 152, 233, 5, 5, 236, 215, 5, 5, 178, 184, 5, 5, 120, 182, 5, 5, 17, 184, 5, 5, 147, 187, 5, 5, 21, 233, 5, 5, 196, 176, 5, 5, 20, 184, 5, 5, 59, 181, 5, 5, 100, 186, 5, 5, 86, 171, 5, 5, 106, 235, 5, 5, 125, 204, 5, 5, 108, 213, 5, 5, 195, 194, 5, 5, 105, 221, 5, 5, 127, 204, 5, 5, 99, 242, 5, 5, 167, 202, 5, 5, 168, 223, 5, 5, 51, 220, 5, 5, 221, 217, 5, 5, 191, 248, 5, 5, 144, 201, 5, 5, 144, 247, 5, 5, 99, 197, 5, 5, 193, 246, 5, 5, 129, 230, 5, 5, 53, 247, 5, 5, 26, 247, 5, 5, 13, 210, 5, 5, 240, 182, 5, 5, 114, 231, 5, 5, 218, 179, 5, 5, 54, 184, 5, 5, 228, 172, 5, 5, 171, 202, 5, 5, 73, 229, 5, 5, 98, 226, 5, 5, 211, 167, 5, 5, 244, 220, 5, 5, 129, 227, 5, 5, 149, 201, 5, 5, 167, 218, 5, 5, 64, 202, 5, 5, 78, 173, 5, 5, 139, 249, 5, 5, 130, 230, 5, 5, 9, 176, 5, 5, 171, 205, 5, 5, 148, 236, 5, 5, 176, 173, 5, 5, 224, 236, 5, 5, 234, 234, 5, 5, 150, 204, 5, 5, 213, 231, 5, 5, 34, 163, 5, 5, 187, 205, 5, 5, 35, 174, 5, 5, 230, 218, 5, 5, 206, 233, 5, 5, 184, 231, 5, 5, 22, 224, 5, 5, 154, 201, 5, 5, 214, 231, 5, 5, 147, 185, 5, 5, 124, 213, 5, 5, 108, 174, 5, 5, 70, 214, 5, 5, 157, 228, 5, 5, 49, 163, 5, 5, 229, 221, 5, 5, 168, 211, 5, 5, 114, 189, 5, 5, 230, 221, 5, 5, 6, 228, 5, 5, 159, 192, 5, 5, 66, 244, 5, 5, 109, 174, 5, 5, 33, 182, 5, 5, 219, 220, 5, 5, 98, 175, 5, 5, 217, 225, 5, 5, 196, 226, 5, 5, 211, 222, 5, 5, 96, 215, 5, 5, 236, 232, 5, 5, 86, 208, 5, 5, 109, 164, 5, 5, 205, 211, 5, 5, 117, 169, 5, 5, 143, 229, 5, 5, 37, 233, 5, 5, 49, 197, 5, 5, 91, 206, 5, 5, 138, 230, 5, 5, 186, 198, 5, 5, 77, 214, 5, 5, 160, 203, 5, 5, 144, 221, 5, 5, 26, 224, 5, 5, 134, 193, 5, 5, 163, 163, 5, 5, 156, 207, 5, 5, 32, 197, 5, 5, 10, 199, 5, 5, 217, 233, 5, 5, 175, 200, 5, 5, 220, 192, 5, 5, 159, 179, 5, 5, 161, 203, 5, 5, 31, 196, 5, 5, 44, 204, 5, 5, 73, 212, 5, 5, 230, 191, 5, 5, 142, 190, 5, 5, 170, 217, 5, 5, 179, 164, 5, 5, 206, 215, 5, 5, 116, 174, 5, 5, 254, 191, 5, 5, 22, 179, 5, 5, 62, 231, 5, 5, 56, 181, 5, 5, 135, 243, 5, 5, 120, 185, 5, 5, 78, 214, 5, 5, 252, 207, 5, 5, 19, 217, 5, 5, 224, 191, 5, 5, 41, 177, 5, 5, 213, 182, 5, 5, 116, 235, 5, 5, 58, 206, 5, 5, 72, 216, 5, 5, 86, 200, 5, 5, 157, 207, 5, 5, 117, 218, 5, 5, 41, 205, 5, 5, 3, 208, 5, 5, 18, 223, 5, 5, 207, 171, 5, 5, 224, 220, 5, 5, 73, 191, 5, 5, 117, 225, 5, 5, 32, 210, 5, 5, 86, 229, 5, 5, 153, 164, 5, 5, 214, 182, 5, 5, 149, 214, 5, 5, 179, 200, 5, 5, 208, 210, 5, 5, 241, 192, 5, 5, 9, 215, 5, 5, 175, 180, 5, 5, 225, 179, 5, 5, 22, 214, 5, 5, 162, 201, 5, 5, 135, 213, 5, 5, 121, 174, 5, 5, 135, 181, 5, 5, 3, 237, 5, 5, 142, 178, 5, 5, 235, 196, 5, 5, 44, 174, 5, 5, 100, 206, 5, 5, 77, 199, 5, 5, 87, 218, 5, 5, 188, 242, 5, 5, 106, 208, 5, 5, 242, 173, 5, 5, 195, 202, 5, 5, 43, 217, 5, 5, 148, 230, 5, 5, 104, 233, 5, 5, 172, 224, 5, 5, 89, 178, 5, 5, 98, 187, 5, 5, 96, 191, 5, 5, 50, 189, 5, 5, 34, 224, 5, 5, 105, 212, 5, 5, 9, 228, 5, 5, 211, 208, 5, 5, 212, 208, 5, 5, 165, 222, 5, 5, 213, 208, 5, 5, 223, 246, 5, 5, 183, 210, 5, 5, 218, 215, 5, 5, 29, 228, 5, 5, 31, 246, 5, 5, 17, 209, 5, 5, 188, 179, 5, 5, 224, 246, 5, 5, 199, 192, 5, 5, 100, 247, 5, 5, 185, 164, 5, 5, 99, 222, 5, 5, 24, 234, 5, 5, 184, 193, 5, 5, 248, 234, 5, 5, 211, 190, 5, 5, 92, 188, 5, 5, 18, 209, 5, 5, 144, 220, 5, 5, 60, 165, 5, 5, 8, 226, 5, 5, 134, 189, 5, 5, 180, 171, 5, 5, 140, 208, 5, 5, 113, 209, 5, 5, 90, 181, 5, 5, 105, 233, 5, 5, 138, 194, 5, 5, 106, 232, 5, 5, 240, 212, 5, 5, 141, 231, 5, 5, 173, 216, 5, 5, 63, 189, 5, 5, 198, 205, 5, 5, 164, 172, 5, 5, 252, 230, 5, 5, 243, 171, 5, 5, 120, 248, 5, 5, 135, 233, 5, 5, 19, 173, 5, 5, 97, 191, 5, 5, 29, 217, 5, 5, 120, 244, 5, 5, 3, 229, 5, 5, 67, 209, 5, 5, 35, 217, 5, 5, 54, 216, 5, 5, 233, 204, 5, 5, 69, 195, 5, 5, 221, 215, 5, 5, 34, 231, 5, 5, 181, 171, 5, 5, 248, 201, 5, 5, 175, 178, 5, 5, 170, 226, 5, 5, 9, 177, 5, 5, 245, 171, 5, 5, 152, 171, 5, 5, 64, 249, 5, 5, 87, 214, 5, 5, 30, 164, 5, 5, 139, 235, 5, 5, 229, 209, 5, 5, 204, 165, 5, 5, 206, 200, 5, 5, 224, 192, 5, 5, 59, 235, 5, 5, 212, 171, 5, 5, 140, 246, 5, 5, 231, 174, 5, 5, 95, 246, 5, 5, 214, 179, 5, 5, 62, 192, 5, 5, 162, 171, 5, 5, 29, 214, 5, 5, 202, 188, 5, 5, 108, 233, 5, 5, 248, 165, 5, 5, 195, 249, 5, 5, 146, 243, 5, 5, 180, 180, 5, 5, 115, 193, 5, 5, 19, 219, 5, 5, 196, 249, 5, 5, 73, 225, 5, 5, 95, 224, 5, 5, 109, 175, 5, 5, 212, 213, 5, 5, 65, 249, 5, 5, 137, 233, 5, 5, 210, 216, 5, 5, 234, 244, 5, 5, 170, 246, 5, 5, 225, 192, 5, 5, 115, 217, 5, 5, 182, 244, 5, 5, 246, 205, 5, 5, 161, 240, 5, 5, 105, 222, 5, 5, 219, 183, 5, 5, 149, 164, 5, 5, 67, 193, 5, 5, 68, 167, 5, 5, 68, 220, 5, 5, 240, 204, 5, 5, 163, 233, 5, 5, 85, 212, 5, 5, 2, 230, 5, 5, 120, 249, 5, 5, 213, 213, 5, 5, 223, 234, 5, 5, 11, 213, 5, 5, 80, 231, 5, 5, 165, 179, 5, 5, 239, 174, 5, 5, 156, 249, 5, 5, 176, 224, 5, 5, 35, 192, 5, 5, 224, 230, 5, 5, 26, 199, 5, 5, 52, 232, 5, 5, 139, 180, 5, 5, 226, 215, 5, 5, 191, 236, 5, 5, 230, 220, 5, 5, 149, 233, 5, 5, 166, 179, 5, 5, 106, 201, 5, 5, 30, 234, 5, 5, 86, 165, 5, 5, 11, 174, 5, 5, 165, 234, 5, 5, 214, 186, 5, 5, 217, 236, 5, 5, 251, 242, 5, 5, 95, 213, 5, 5, 151, 233, 5, 5, 95, 185, 5, 5, 11, 208, 5, 5, 192, 210, 5, 5, 153, 170, 5, 5, 134, 218, 5, 5, 82, 191, 5, 5, 108, 172, 5, 5, 88, 168, 5, 5, 151, 188, 5, 5, 46, 171, 5, 5, 179, 217, 5, 5, 138, 233, 5, 5, 9, 202, 5, 5, 68, 228, 5, 5, 45, 237, 5, 5, 26, 173, 5, 5, 102, 191, 5, 5, 193, 164, 5, 5, 117, 239, 5, 5, 23, 236, 5, 5, 97, 170, 5, 5, 197, 241, 5, 5, 78, 211, 5, 5, 13, 201, 5, 5, 239, 215, 5, 5, 235, 202, 5, 5, 167, 227, 5, 5, 27, 215, 5, 5, 72, 230, 5, 5, 180, 217, 5, 5, 49, 237, 5, 5, 153, 209, 5, 5, 205, 232, 5, 5, 254, 232, 5, 5, 96, 218, 5, 5, 116, 228, 5, 5, 96, 236, 5, 5, 248, 188, 5, 5, 156, 248, 5, 5, 249, 172, 5, 5, 251, 187, 5, 5, 14, 226, 5, 5, 73, 233, 5, 5, 141, 205, 5, 5, 31, 230, 5, 5, 148, 224, 5, 5, 4, 231, 5, 5, 28, 215, 5, 5, 133, 200, 5, 5, 153, 199, 5, 5, 59, 233, 5, 5, 26, 213, 5, 5, 80, 211, 5, 5, 209, 249, 5, 5, 54, 175, 5, 5, 34, 229, 5, 5, 137, 200, 5, 5, 157, 213, 5, 5, 246, 235, 5, 5, 63, 205, 5, 5, 32, 206, 5, 5, 197, 200, 5, 5, 75, 232, 5, 5, 70, 235, 5, 5, 110, 214, 5, 5, 74, 204, 5, 5, 96, 165, 5, 5, 61, 176, 5, 5, 182, 240, 5, 5, 207, 184, 5, 5, 51, 237, 5, 5, 183, 240, 5, 5, 222, 227, 5, 5, 217, 216, 5, 5, 239, 202, 5, 5, 33, 166, 5, 5, 162, 210, 5, 5, 47, 186, 5, 5, 241, 168, 5, 5, 205, 217, 5, 5, 61, 233, 5, 5, 241, 208, 5, 5, 184, 167, 5, 5, 3, 242, 5, 5, 151, 169, 5, 5, 159, 185, 5, 5, 41, 214, 5, 5, 62, 233, 5, 5, 46, 197, 5, 5, 24, 242, 5, 5, 40, 217, 5, 5, 102, 185, 5, 5, 76, 215, 5, 5, 113, 224, 5, 5, 118, 204, 5, 5, 217, 189, 5, 5, 114, 224, 5, 5, 149, 244, 5, 5, 14, 234, 5, 5, 167, 230, 5, 5, 79, 215, 5, 5, 30, 233, 5, 5, 64, 233, 5, 5, 156, 190, 5, 5, 244, 186, 5, 5, 176, 175, 5, 5, 89, 186, 5, 5, 114, 185, 5, 5, 214, 188, 5, 5, 102, 235, 5, 5, 121, 204, 5, 5, 140, 247, 5, 5, 130, 201, 5, 5, 42, 220, 5, 5, 39, 215, 5, 5, 134, 191, 5, 5, 190, 248, 5, 5, 43, 202, 5, 5, 187, 246, 5, 5, 21, 197, 5, 5, 152, 202, 5, 5, 116, 230, 5, 5, 51, 246, 5, 5, 132, 249, 5, 5, 135, 201, 5, 5, 95, 226, 5, 5, 199, 167, 5, 5, 121, 227, 5, 5, 182, 192, 5, 5, 235, 182, 5, 5, 122, 230, 5, 5, 140, 233, 5, 5, 195, 190, 5, 5, 72, 173, 5, 5, 151, 218, 5, 5, 21, 247, 5, 5, 169, 173, 5, 5, 211, 221, 5, 5, 234, 220, 5, 5, 51, 184, 5, 5, 62, 229, 5, 5, 183, 215, 5, 5, 155, 192, 5, 5, 54, 214, 5, 5, 75, 174, 5, 5, 185, 188, 5, 5, 230, 232, 5, 5, 142, 228, 5, 5, 114, 213, 5, 5, 65, 218, 5, 5, 202, 222, 5, 5, 47, 220, 5, 5, 105, 164, 5, 5, 76, 174, 5, 5, 206, 225, 5, 5, 2, 228, 5, 5, 36, 240, 5, 5, 80, 206, 5, 5, 84, 175, 5, 5, 152, 164, 5, 5, 52, 231, 5, 5, 145, 203, 5, 5, 25, 196, 5, 5, 29, 230, 5, 5, 27, 177, 5, 5, 177, 198, 5, 5, 72, 200, 5, 5, 253, 191, 5, 5, 162, 217, 5, 5, 55, 206, 5, 5, 135, 190, 5, 5, 134, 207, 5, 5, 33, 204, 5, 5, 208, 182, 5, 5, 234, 192, 5, 5, 133, 214, 5, 5, 247, 207, 5, 5, 16, 215, 5, 5, 122, 213, 5, 5, 9, 237, 5, 5, 151, 179, 5, 5, 172, 202, 5, 5, 162, 180, 5, 5, 222, 201, 5, 5, 11, 217, 5, 5, 115, 189, 5, 5, 99, 244, 5, 5, 12, 217, 5, 5, 13, 209, 5, 5, 158, 172, 5, 5, 2, 226, 5, 5, 2, 229, 5, 5, 132, 194, 5, 5, 32, 229, 5, 5, 19, 246, 5, 5, 177, 171, 5, 5, 26, 228, 5, 5, 189, 164, 5, 5, 23, 224, 5, 5, 184, 179, 5, 5, 88, 188, 5, 5, 104, 212, 5, 5, 89, 233, 5, 5, 139, 220, 5, 5, 212, 246, 5, 5, 90, 191, 5, 5, 93, 222, 5, 5, 120, 231, 5, 5, 64, 195, 5, 5, 47, 216, 5, 5, 213, 204, 5, 5, 206, 171, 5, 5, 67, 186, 5, 5, 170, 180, 5, 5, 160, 171, 5, 5, 230, 201, 5, 5, 112, 249, 5, 5, 87, 200, 5, 5, 117, 235, 5, 5, 84, 178, 5, 5, 167, 178, 5, 5, 244, 165, 5, 5, 207, 215, 5, 5, 58, 192, 5, 5, 95, 222, 5, 5, 204, 200, 5, 5, 249, 176, 5, 5, 72, 191, 5, 5, 199, 165, 5, 5, 237, 171, 5, 5, 48, 249, 5, 5, 230, 205, 5, 5, 224, 174, 5, 5, 222, 230, 5, 5, 225, 220, 5, 5, 5, 174, 5, 5, 213, 234, 5, 5, 27, 192, 5, 5, 85, 224, 5, 5, 192, 186, 5, 5, 112, 240, 5, 5, 224, 204, 5, 5, 17, 166, 5, 5, 70, 168, 5, 5, 144, 170, 5, 5, 89, 185, 5, 5, 113, 244, 5, 5, 211, 236, 5, 5, 92, 170, 5, 5, 22, 236, 5, 5, 30, 214, 5, 5, 239, 187, 5, 5, 104, 228, 5, 5, 197, 232, 5, 5, 23, 206, 5, 5, 161, 210, 5, 5, 148, 209, 5, 5, 188, 200, 5, 5, 67, 204, 5, 5, 60, 233, 5, 5, 230, 172, 5, 5, 55, 214, 5, 5, 189, 192, 5, 5, 233, 172, 5, 5, 8, 212, 5, 5, 234, 172, 5, 5, 65, 209, 5, 5, 50, 163, 5, 5, 191, 168, 5, 5, 101, 216, 5, 5, 252, 219, 5, 5, 200, 192, 5, 5, 114, 244, 5, 5, 122, 244, 5, 5, 145, 211, 5, 5, 235, 213, 5, 5, 114, 194, 5, 5, 74, 233, 5, 5, 14, 180, 5, 5, 131, 234, 5, 5, 62, 209, 5, 5, 136, 247, 5, 5, 139, 244, 5, 5, 103, 185, 5, 5, 21, 180, 5, 5, 218, 189, 5, 5, 124, 194, 5, 5, 88, 247, 5, 5, 87, 247, 5, 5, 178, 192, 5, 5, 92, 247, 5, 5, 105, 198, 5, 5, 89, 190, 5, 5, 85, 190, 5, 5, 89, 174, 5, 5, 36, 171, 5, 5, 126, 172, 5, 5, 10, 193, 5, 5, 193, 221, 5, 5, 205, 206, 5, 5, 26, 170, 5, 5, 41, 193, 5, 5, 233, 171, 5, 5, 192, 192, 5, 5, 19, 163, 5, 5, 133, 193, 5, 5, 171, 180, 5, 5, 253, 207, 5, 5, 117, 174, 5, 5, 219, 228, 5, 5, 239, 242, 5, 5, 239, 219, 5, 5, 43, 233, 5, 5, 135, 193, 5, 5, 48, 222, 5, 5, 106, 248, 5, 5, 253, 218, 5, 5, 235, 238, 5, 5, 133, 172, 5, 5, 70, 186, 5, 5, 112, 229, 5, 5, 127, 235, 5, 5, 8, 168, 5, 5, 55, 169, 5, 5, 94, 188, 5, 5, 24, 163, 5, 5, 135, 172, 5, 5, 209, 186, 5, 5, 248, 223, 5, 5, 146, 246, 5, 5, 80, 165, 5, 5, 224, 202, 5, 5, 204, 232, 5, 5, 163, 226, 5, 5, 98, 188, 5, 5, 10, 202, 5, 5, 12, 208, 5, 5, 123, 171, 5, 5, 249, 219, 5, 5, 102, 190, 5, 5, 17, 168, 5, 5, 30, 202, 5, 5, 36, 202, 5, 5, 176, 190, 5, 5, 15, 244, 5, 5, 72, 171, 5, 5, 31, 211, 5, 5, 24, 175, 5, 5, 237, 169, 5, 5, 31, 244, 5, 5, 77, 242, 5, 5, 74, 242, 5, 5, 100, 173, 5, 5, 0, 146, 5, 5, 2, 192, 5, 5, 83, 244, 5, 5, 42, 193, 5, 5, 32, 211, 5, 5, 174, 232, 5, 5, 3, 192, 5, 5, 117, 189, 5, 5, 76, 240, 5, 5, 231, 221, 5, 5, 152, 176, 5, 5, 244, 214, 5, 5, 74, 211, 5, 5, 170, 239, 5, 5, 208, 215, 5, 5, 63, 231, 5, 5, 123, 212, 5, 5, 124, 212, 5, 5, 76, 211, 5, 5, 180, 200, 5, 5, 209, 236, 5, 5, 211, 215, 5, 5, 113, 240, 5, 5, 90, 244, 5, 5, 45, 211, 5, 5, 138, 210, 5, 5, 224, 247, 5, 5, 226, 192, 5, 5, 183, 203, 5, 5, 56, 180, 5, 5, 71, 249, 5, 5, 12, 213, 5, 5, 28, 173, 5, 5, 70, 228, 5, 5, 118, 173, 5, 5, 71, 228, 5, 5, 182, 188, 5, 5, 100, 185, 5, 5, 228, 192, 5, 5, 85, 183, 5, 5, 85, 211, 5, 5, 163, 224, 5, 5, 6, 248, 5, 5, 232, 194, 5, 5, 51, 240, 5, 5, 154, 226, 5, 5, 115, 247, 5, 5, 79, 173, 5, 5, 150, 201, 5, 5, 91, 223, 5, 5, 92, 223, 5, 5, 235, 234, 5, 5, 110, 174, 5, 5, 198, 213, 5, 5, 126, 164, 5, 5, 77, 240, 5, 5, 229, 184, 5, 5, 78, 240, 5, 5, 214, 167, 5, 5, 118, 189, 5, 5, 101, 188, 5, 5, 231, 218, 5, 5, 206, 211, 5, 5, 48, 213, 5, 5, 17, 214, 5, 5, 73, 216, 5, 5, 224, 212, 5, 5, 129, 213, 5, 5, 36, 195, 5, 5, 197, 209, 5, 5, 88, 200, 5, 5, 17, 200, 5, 5, 150, 214, 5, 5, 161, 227, 5, 5, 163, 201, 5, 5, 18, 206, 5, 5, 181, 176, 5, 5, 249, 234, 5, 5, 188, 172, 5, 5, 165, 175, 5, 5, 164, 227, 5, 5, 108, 223, 5, 5, 245, 221, 5, 5, 50, 205, 5, 5, 116, 217, 5, 5, 113, 223, 5, 5, 104, 168, 5, 5, 109, 233, 5, 5, 200, 246, 5, 5, 29, 200, 5, 5, 38, 198, 5, 5, 140, 197, 5, 5, 168, 197, 5, 5, 151, 198, 5, 5, 119, 238, 5, 5, 73, 168, 5, 5, 171, 240, 5, 5, 152, 188, 5, 5, 40, 198, 5, 5, 161, 238, 5, 5, 249, 188, 5, 5, 104, 167, 5, 5, 206, 209, 5, 5, 122, 238, 5, 5, 250, 172, 5, 5, 172, 240, 5, 5, 138, 200, 5, 5, 126, 217, 5, 5, 173, 175, 5, 5, 2, 242, 5, 5, 193, 185, 5, 5, 192, 185, 5, 5, 119, 195, 5, 5, 64, 170, 5, 5, 164, 219, 5, 5, 177, 208, 5, 5, 72, 192, 5, 5, 75, 192, 5, 5, 76, 192, 5, 5, 68, 177, 5, 5, 42, 177, 5, 5, 26, 239, 5, 5, 173, 229, 5, 5, 28, 231, 5, 5, 24, 183, 5, 5, 25, 183, 5, 5, 155, 178, 5, 5, 117, 185, 5, 5, 130, 181, 5, 5, 176, 231, 5, 5, 217, 192, 5, 5, 7, 224, 5, 5, 252, 197, 5, 5, 140, 249, 5, 5, 20, 163, 5, 5, 215, 167, 5, 5, 167, 239, 5, 5, 180, 235, 5, 5, 251, 211, 5, 5, 21, 203, 5, 5, 68, 207, 5, 5, 166, 199, 5, 5, 162, 225, 5, 5, 209, 185, 5, 5, 220, 190, 5, 5, 88, 202, 5, 5, 54, 180, 5, 5, 191, 238, 5, 5, 241, 185, 5, 5, 88, 183, 5, 5, 46, 206, 5, 5, 128, 167, 5, 5, 90, 219, 5, 5, 245, 185, 5, 5, 15, 202, 5, 5, 216, 185, 5, 5, 221, 199, 5, 5, 25, 242, 5, 5, 153, 169, 5, 5, 244, 199, 5, 5, 53, 239, 5, 5, 19, 240, 5, 5, 195, 177, 5, 5, 61, 241, 5, 5, 6, 175, 5, 5, 111, 237, 5, 5, 224, 210, 5, 5, 141, 249, 5, 5, 69, 223, 5, 5, 232, 175, 5, 5, 233, 175, 5, 5, 151, 201, 5, 5, 235, 192, 5, 5, 106, 218, 5, 5, 87, 167, 5, 5, 233, 183, 5, 5, 235, 175, 5, 5, 220, 220, 5, 5, 193, 223, 5, 5, 119, 189, 5, 5, 83, 221, 5, 5, 79, 242, 5, 5, 157, 245, 5, 5, 225, 187, 5, 5, 238, 190, 5, 5, 112, 218, 5, 5, 114, 184, 5, 5, 18, 181, 5, 5, 39, 239, 5, 5, 41, 241, 5, 5, 249, 180, 5, 5, 109, 193, 5, 5, 42, 217, 5, 5, 124, 188, 5, 5, 5, 198, 5, 5, 50, 229, 5, 5, 202, 187, 5, 5, 155, 168, 5, 5, 104, 206, 5, 5, 82, 207, 5, 5, 241, 164, 5, 5, 220, 176, 5, 5, 169, 193, 5, 5, 34, 247, 5, 5, 30, 217, 5, 5, 162, 168, 5, 5, 51, 235, 5, 5, 230, 177, 5, 5, 168, 183, 5, 5, 157, 181, 5, 5, 30, 207, 5, 5, 60, 235, 5, 5, 107, 196, 5, 5, 215, 188, 5, 5, 169, 196, 5, 5, 61, 235, 5, 5, 143, 188, 5, 5, 131, 247, 5, 5, 172, 196, 5, 5, 148, 188, 5, 5, 153, 188, 5, 5, 67, 235, 5, 5, 120, 238, 5, 5, 29, 229, 5, 5, 119, 237, 5, 5, 138, 181, 5, 5, 176, 196, 5, 5, 5, 184, 5, 5, 135, 218, 5, 5, 255, 175, 5, 5, 129, 184, 5, 5, 92, 209, 5, 5, 198, 183, 5, 5, 140, 181, 5, 5, 141, 181, 5, 5, 16, 191, 5, 5, 51, 248, 5, 5, 47, 248, 5, 5, 233, 219, 5, 5, 177, 235, 5, 5, 144, 236, 5, 5, 114, 240, 5, 5, 167, 236, 5, 5, 185, 235, 5, 5, 241, 219, 5, 5, 242, 219, 5, 5, 56, 230, 5, 5, 3, 240, 5, 5, 10, 185, 5, 5, 205, 201, 5, 5, 250, 209, 5, 5, 194, 195, 5, 5, 51, 234, 5, 5, 137, 191, 5, 5, 68, 163, 5, 5, 210, 200, 5, 5, 31, 201, 5, 5, 217, 239, 5, 5, 123, 230, 5, 5, 212, 214, 5, 5, 212, 230, 5, 5, 34, 220, 5, 5, 90, 249, 5, 5, 33, 178, 5, 5, 183, 192, 5, 5, 49, 231, 5, 5, 108, 192, 5, 5, 128, 223, 5, 5, 189, 245, 5, 5, 4, 225, 5, 5, 36, 193, 5, 5, 66, 218, 5, 5, 77, 198, 5, 5, 24, 174, 5, 5, 174, 172, 5, 5, 99, 173, 5, 5, 229, 166, 5, 5, 131, 167, 5, 5, 198, 204, 5, 5, 161, 173, 5, 5, 156, 222, 5, 5, 34, 242, 5, 5, 237, 182, 5, 5, 11, 239, 5, 5, 195, 195, 5, 5, 60, 230, 5, 5, 196, 194, 5, 5, 57, 248, 5, 5, 52, 247, 5, 5, 158, 241, 5, 5, 167, 208, 5, 5, 241, 167, 5, 5, 150, 210, 5, 5, 179, 212, 5, 5, 180, 212, 5, 5, 198, 172, 5, 5, 196, 195, 5, 5, 123, 206, 5, 5, 137, 179, 5, 5, 205, 228, 5, 5, 225, 222, 5, 5, 45, 221, 5, 5, 255, 213, 5, 5, 208, 228, 5, 5, 121, 246, 5, 5, 52, 220, 5, 5, 28, 177, 5, 5, 7, 176, 5, 5, 201, 172, 5, 5, 0, 162, 5, 5, 75, 163, 5, 5, 202, 211, 5, 5, 78, 181, 5, 5, 206, 163, 5, 5, 105, 193, 5, 5, 104, 179, 5, 5, 177, 231, 5, 5, 7, 196, 5, 5, 71, 218, 5, 5, 204, 183, 5, 5, 178, 168, 5, 5, 200, 231, 5, 5, 160, 174, 5, 5, 132, 167, 5, 5, 163, 220, 5, 5, 89, 215, 5, 5, 9, 233, 5, 5, 231, 166, 5, 5, 71, 203, 5, 5, 41, 229, 5, 5, 72, 218, 5, 5, 252, 163, 5, 5, 71, 172, 5, 5, 165, 211, 5, 5, 250, 198, 5, 5, 3, 195, 5, 5, 64, 223, 5, 5, 235, 243, 5, 5, 186, 216, 5, 5, 6, 244, 5, 5, 232, 190, 5, 5, 13, 179, 5, 5, 246, 194, 5, 5, 191, 189, 5, 5, 148, 206, 5, 5, 150, 184, 5, 5, 126, 242, 5, 5, 108, 167, 5, 5, 48, 196, 5, 5, 161, 239, 5, 5, 165, 194, 5, 5, 189, 197, 5, 5, 143, 191, 5, 5, 16, 179, 5, 5, 7, 223, 5, 5, 245, 203, 5, 5, 39, 218, 5, 5, 40, 218, 5, 5, 34, 239, 5, 5, 242, 238, 5, 5, 203, 198, 5, 5, 108, 168, 5, 5, 142, 183, 5, 5, 201, 231, 5, 5, 217, 193, 5, 5, 104, 225, 5, 5, 47, 181, 5, 5, 77, 218, 5, 5, 236, 234, 5, 5, 200, 169, 5, 5, 9, 214, 5, 5, 237, 234, 5, 5, 10, 214, 5, 5, 181, 169, 5, 5, 180, 169, 5, 5, 65, 202, 5, 5, 118, 214, 5, 5, 101, 184, 5, 5, 85, 196, 5, 5, 210, 235, 5, 5, 188, 232, 5, 5, 178, 208, 5, 5, 67, 183, 5, 5, 77, 171, 5, 5, 191, 184, 5, 5, 163, 190, 5, 5, 237, 246, 5, 5, 89, 163, 5, 5, 55, 247, 5, 5, 213, 246, 5, 5, 245, 214, 5, 5, 40, 204, 5, 5, 161, 206, 5, 5, 80, 242, 5, 5, 196, 198, 5, 5, 192, 234, 5, 5, 197, 187, 5, 5, 58, 228, 5, 5, 173, 181, 5, 5, 203, 169, 5, 5, 101, 226, 5, 5, 212, 228, 5, 5, 153, 210, 5, 5, 68, 175, 5, 5, 219, 207, 5, 5, 153, 234, 5, 5, 246, 214, 5, 5, 27, 216, 5, 5, 12, 182, 5, 5, 11, 167, 5, 5, 108, 225, 5, 5, 35, 183, 5, 5, 223, 168, 5, 5, 205, 175, 5, 5, 13, 182, 5, 5, 43, 184, 5, 5, 84, 183, 5, 5, 26, 168, 5, 5, 47, 193, 5, 5, 91, 190, 5, 5, 73, 187, 5, 5, 172, 188, 5, 5, 49, 221, 5, 5, 244, 234, 5, 5, 111, 203, 5, 5, 62, 189, 5, 5, 206, 229, 5, 5, 5, 168, 5, 5, 147, 165, 5, 5, 208, 190, 5, 5, 109, 168, 5, 5, 121, 185, 5, 5, 158, 196, 5, 5, 34, 226, 5, 5, 38, 176, 5, 5, 107, 181, 5, 5, 36, 183, 5, 5, 95, 206, 5, 5, 7, 183, 5, 5, 241, 184, 5, 5, 98, 241, 5, 5, 139, 242, 5, 5, 139, 230, 5, 5, 214, 228, 5, 5, 4, 226, 5, 5, 174, 174, 5, 5, 170, 174, 5, 5, 215, 206, 5, 5, 176, 172, 5, 5, 36, 211, 5, 5, 239, 246, 5, 5, 23, 179, 5, 5, 6, 168, 5, 5, 165, 193, 5, 5, 97, 233, 5, 5, 220, 177, 5, 5, 203, 245, 5, 5, 244, 169, 5, 5, 167, 183, 5, 5, 141, 222, 5, 5, 103, 223, 5, 5, 58, 234, 5, 5, 96, 190, 5, 5, 188, 238, 5, 5, 252, 176, 5, 5, 209, 238, 5, 5, 126, 216, 5, 5, 73, 183, 5, 5, 149, 229, 5, 5, 226, 229, 5, 5, 114, 183, 5, 5, 189, 249, 5, 5, 234, 222, 5, 5, 58, 230, 5, 5, 239, 237, 5, 5, 4, 236, 5, 5, 220, 181, 5, 5, 149, 241, 5, 5, 107, 205, 5, 5, 82, 246, 5, 5, 49, 204, 5, 5, 203, 247, 5, 5, 28, 203, 5, 5, 171, 178, 5, 5, 46, 165, 5, 5, 14, 225, 5, 5, 225, 174, 5, 5, 212, 215, 5, 5, 214, 183, 5, 5, 208, 171, 5, 5, 160, 245, 5, 5, 135, 180, 5, 5, 91, 182, 5, 5, 156, 164, 5, 5, 157, 210, 5, 5, 75, 191, 5, 5, 161, 245, 5, 5, 77, 236, 5, 5, 84, 180, 5, 5, 173, 236, 5, 5, 100, 222, 5, 5, 183, 191, 5, 5, 124, 240, 5, 5, 108, 163, 5, 5, 231, 187, 5, 5, 89, 202, 5, 5, 9, 168, 5, 5, 26, 214, 5, 5, 12, 192, 5, 5, 79, 191, 5, 5, 38, 166, 5, 5, 182, 238, 5, 5, 19, 225, 5, 5, 46, 190, 5, 5, 170, 193, 5, 5, 111, 167, 5, 5, 21, 226, 5, 5, 157, 165, 5, 5, 44, 206, 5, 5, 49, 174, 5, 5, 92, 180, 5, 5, 75, 169, 5, 5, 253, 212, 5, 5, 123, 222, 5, 5, 100, 191, 5, 5, 8, 204, 5, 5, 227, 244, 5, 5, 52, 191, 5, 5, 21, 219, 5, 5, 88, 217, 5, 5, 90, 226, 5, 5, 30, 189, 5, 5, 202, 182, 5, 5, 172, 167, 5, 5, 220, 183, 5, 5, 24, 226, 5, 5, 243, 187, 5, 5, 250, 234, 5, 5, 239, 193, 5, 5, 165, 214, 5, 5, 130, 180, 5, 5, 25, 173, 5, 5, 132, 218, 5, 5, 10, 204, 5, 5, 110, 228, 5, 5, 244, 198, 5, 5, 251, 234, 5, 5, 46, 231, 5, 5, 55, 216, 5, 5, 15, 237, 5, 5, 138, 173, 5, 5, 129, 244, 5, 5, 69, 228, 5, 5, 2, 222, 5, 5, 18, 176, 5, 5, 96, 196, 5, 5, 200, 184, 5, 5, 150, 182, 5, 5, 154, 188, 5, 5, 87, 165, 5, 5, 230, 249, 5, 5, 184, 207, 5, 5, 234, 225, 5, 5, 173, 240, 5, 5, 239, 207, 5, 5, 250, 184, 5, 5, 180, 188, 5, 5, 24, 236, 5, 5, 151, 182, 5, 5, 70, 193, 5, 5, 231, 208, 5, 5, 255, 234, 5, 5, 43, 166, 5, 5, 202, 217, 5, 5, 83, 183, 5, 5, 211, 245, 5, 5, 192, 232, 5, 5, 234, 215, 5, 5, 169, 165, 5, 5, 176, 216, 5, 5, 186, 196, 5, 5, 253, 184, 5, 5, 153, 242, 5, 5, 213, 189, 5, 5, 222, 199, 5, 5, 217, 231, 5, 5, 129, 190, 5, 5, 228, 211, 5, 5, 193, 232, 5, 5, 177, 190, 5, 5, 179, 183, 5, 5, 218, 231, 5, 5, 179, 224, 5, 5, 66, 210, 5, 5, 192, 245, 5, 5, 182, 194, 5, 5, 177, 232, 5, 5, 247, 188, 5, 5, 176, 194, 5, 5, 189, 231, 5, 5, 26, 204, 5, 5, 234, 183, 5, 5, 63, 224, 5, 5, 192, 202, 5, 5, 8, 192, 5, 5, 64, 224, 5, 5, 182, 171, 5, 5, 254, 183, 5, 5, 178, 218, 5, 5, 201, 218, 5, 5, 172, 191, 5, 5, 100, 205, 5, 5, 43, 177, 5, 5, 83, 218, 5, 5, 248, 214, 5, 5, 120, 179, 5, 5, 201, 220, 5, 5, 157, 204, 5, 5, 184, 215, 5, 5, 168, 247, 5, 5, 129, 223, 5, 5, 221, 241, 5, 5, 182, 247, 5, 5, 155, 226, 5, 5, 130, 213, 5, 5, 131, 213, 5, 5, 172, 204, 5, 5, 233, 247, 5, 5, 32, 248, 5, 5, 174, 216, 5, 5, 31, 207, 5, 5, 179, 204, 5, 5, 33, 216, 5, 5, 177, 227, 5, 5, 183, 227, 5, 5, 214, 227, 5, 5, 148, 204, 5, 5, 62, 227, 5, 5, 232, 179, 5, 5, 81, 229, 5, 5, 89, 227, 5, 5, 111, 181, 5, 5, 46, 180, 5, 5, 7, 169, 5, 5, 8, 169, 5, 5, 91, 194, 5, 5, 154, 199, 5, 5, 121, 181, 5, 5, 64, 226, 5, 5, 229, 229, 5, 5, 207, 213, 5, 5, 68, 245, 5, 5, 254, 219, 5, 5, 131, 175, 5, 5, 54, 226, 5, 5, 50, 228, 5, 5, 219, 239, 5, 5, 68, 173, 5, 5, 155, 166, 5, 5, 140, 179, 5, 5, 14, 173, 5, 5, 92, 164, 5, 5, 55, 226, 5, 5, 180, 198, 5, 5, 177, 173, 5, 5, 70, 164, 5, 5, 236, 164, 5, 5, 82, 190, 5, 5, 208, 222, 5, 5, 90, 205, 5, 5, 117, 234, 5, 5, 120, 189, 5, 5, 174, 221, 5, 5, 248, 231, 5, 5, 108, 178, 5, 5, 90, 163, 5, 5, 97, 215, 5, 5, 201, 191, 5, 5, 56, 226, 5, 5, 199, 213, 5, 5, 195, 189, 5, 5, 144, 212, 5, 5, 119, 167, 5, 5, 225, 212, 5, 5, 51, 196, 5, 5, 122, 174, 5, 5, 85, 232, 5, 5, 243, 173, 5, 5, 193, 186, 5, 5, 236, 227, 5, 5, 214, 247, 5, 5, 121, 186, 5, 5, 175, 181, 5, 5, 107, 248, 5, 5, 83, 246, 5, 5, 109, 163, 5, 5, 12, 171, 5, 5, 94, 184, 5, 5, 221, 206, 5, 5, 62, 201, 5, 5, 18, 228, 5, 5, 70, 195, 5, 5, 249, 165, 5, 5, 105, 172, 5, 5, 54, 177, 5, 5, 224, 175, 5, 5, 172, 164, 5, 5, 139, 223, 5, 5, 162, 240, 5, 5, 57, 236, 5, 5, 123, 234, 5, 5, 42, 234, 5, 5, 246, 185, 5, 5, 114, 175, 5, 5, 90, 217, 5, 5, 253, 241, 5, 5, 164, 226, 5, 5, 146, 176, 5, 5, 78, 176, 5, 5, 71, 211, 5, 5, 250, 189, 5, 5, 245, 239, 5, 5, 113, 202, 5, 5, 146, 205, 5, 5, 223, 199, 5, 5, 252, 170, 5, 5, 242, 208, 5, 5, 252, 189, 5, 5, 219, 189, 5, 5, 117, 224, 5, 5, 248, 229, 5, 5, 172, 192, 5, 5, 40, 189, 5, 5, 199, 204, 5, 5, 243, 204, 5, 5, 222, 235, 5, 5, 33, 243, 5, 5, 111, 174, 5, 5, 163, 168, 5, 5, 76, 243, 5, 5, 95, 243, 5, 5, 46, 234, 5, 5, 55, 234, 5, 5, 4, 240, 5, 5, 20, 185, 5, 5, 127, 181, 5, 5, 237, 203, 5, 5, 55, 238, 5, 5, 14, 183, 5, 5, 19, 181, 5, 5, 238, 203, 5, 5, 130, 203, 5, 5, 149, 167, 5, 5, 191, 246, 5, 5, 20, 240, 5, 5, 205, 223, 5, 5, 174, 171, 5, 5, 69, 171, 5, 5, 109, 192, 5, 5, 62, 210, 5, 5, 54, 248, 5, 5, 55, 248, 5, 5, 245, 227, 5, 5, 127, 212, 5, 5, 34, 178, 5, 5, 63, 199, 5, 5, 5, 209, 5, 5, 189, 170, 5, 5, 190, 231, 5, 5, 73, 200, 5, 5, 247, 236, 5, 5, 183, 218, 5, 5, 39, 216, 5, 5, 242, 167, 5, 5, 77, 174, 5, 5, 141, 182, 5, 5, 56, 246, 5, 5, 18, 245, 5, 5, 67, 218, 5, 5, 34, 209, 5, 5, 45, 228, 5, 5, 24, 247, 5, 5, 167, 210, 5, 5, 91, 230, 5, 5, 78, 174, 5, 5, 135, 195, 5, 5, 213, 203, 5, 5, 188, 189, 5, 5, 249, 216, 5, 5, 210, 201, 5, 5, 219, 210, 5, 5, 168, 208, 5, 5, 164, 242, 5, 5, 75, 173, 5, 5, 0, 137, 5, 5, 9, 231, 5, 5, 250, 215, 5, 5, 159, 202, 5, 5, 101, 199, 5, 5, 102, 199, 5, 5, 152, 173, 5, 5, 93, 245, 5, 5, 129, 207, 5, 5, 69, 192, 5, 5, 156, 218, 5, 5, 134, 249, 5, 5, 220, 210, 5, 5, 132, 238, 5, 5, 155, 225, 5, 5, 58, 170, 5, 5, 246, 240, 5, 5, 199, 172, 5, 5, 180, 230, 5, 5, 175, 198, 5, 5, 42, 235, 5, 5, 164, 243, 5, 5, 72, 242, 5, 5, 237, 165, 5, 5, 37, 240, 5, 5, 125, 209, 5, 5, 57, 246, 5, 5, 42, 187, 5, 5, 25, 168, 5, 5, 196, 237, 5, 5, 185, 215, 5, 5, 5, 225, 5, 5, 114, 234, 5, 5, 215, 221, 5, 5, 129, 236, 5, 5, 161, 180, 5, 5, 94, 249, 5, 5, 79, 230, 5, 5, 99, 189, 5, 5, 114, 179, 5, 5, 248, 227, 5, 5, 166, 176, 5, 5, 146, 171, 5, 5, 126, 211, 5, 5, 50, 187, 5, 5, 25, 250, 5, 5, 4, 195, 5, 5, 88, 172, 5, 5, 207, 163, 5, 5, 253, 163, 5, 5, 73, 192, 5, 5, 232, 171, 5, 5, 220, 239, 5, 5, 221, 239, 5, 5, 55, 207, 5, 5, 65, 168, 5, 5, 234, 230, 5, 5, 21, 196, 5, 5, 113, 184, 5, 5, 186, 183, 5, 5, 76, 163, 5, 5, 67, 166, 5, 5, 51, 228, 5, 5, 7, 215, 5, 5, 217, 241, 5, 5, 48, 215, 5, 5, 123, 164, 5, 5, 145, 201, 5, 5, 94, 244, 5, 5, 95, 219, 5, 5, 230, 178, 5, 5, 199, 206, 5, 5, 29, 177, 5, 5, 226, 183, 5, 5, 117, 192, 5, 5, 226, 198, 5, 5, 179, 168, 5, 5, 198, 169, 5, 5, 90, 174, 5, 5, 20, 229, 5, 5, 7, 193, 5, 5, 12, 165, 5, 5, 122, 246, 5, 5, 91, 174, 5, 5, 28, 222, 5, 5, 62, 170, 5, 5, 53, 173, 5, 5, 13, 203, 5, 5, 23, 170, 5, 5, 24, 170, 5, 5, 2, 163, 5, 5, 174, 233, 5, 5, 107, 169, 5, 5, 111, 231, 5, 5, 211, 176, 5, 5, 38, 223, 5, 5, 168, 202, 5, 5, 83, 206, 5, 5, 233, 190, 5, 5, 99, 235, 5, 5, 30, 184, 5, 5, 49, 249, 5, 5, 62, 217, 5, 5, 158, 244, 5, 5, 134, 212, 5, 5, 169, 191, 5, 5, 32, 172, 5, 5, 65, 187, 5, 5, 56, 231, 5, 5, 237, 243, 5, 5, 175, 192, 5, 5, 171, 210, 5, 5, 216, 223, 5, 5, 187, 224, 5, 5, 68, 248, 5, 5, 211, 197, 5, 5, 196, 231, 5, 5, 192, 187, 5, 5, 26, 245, 5, 5, 173, 202, 5, 5, 218, 193, 5, 5, 216, 233, 5, 5, 21, 205, 5, 5, 175, 233, 5, 5, 86, 192, 5, 5, 166, 194, 5, 5, 200, 170, 5, 5, 145, 223, 5, 5, 246, 203, 5, 5, 247, 203, 5, 5, 98, 174, 5, 5, 183, 247, 5, 5, 138, 221, 5, 5, 115, 231, 5, 5, 12, 206, 5, 5, 47, 200, 5, 5, 123, 243, 5, 5, 146, 223, 5, 5, 130, 172, 5, 5, 47, 231, 5, 5, 186, 230, 5, 5, 151, 226, 5, 5, 215, 170, 5, 5, 175, 182, 5, 5, 245, 222, 5, 5, 67, 210, 5, 5, 172, 221, 5, 5, 226, 217, 5, 5, 173, 200, 5, 5, 31, 174, 5, 5, 61, 212, 5, 5, 220, 191, 5, 5, 7, 175, 5, 5, 96, 206, 5, 5, 168, 216, 5, 5, 65, 226, 5, 5, 47, 234, 5, 5, 85, 206, 5, 5, 35, 239, 5, 5, 243, 238, 5, 5, 160, 209, 5, 5, 40, 223, 5, 5, 100, 230, 5, 5, 91, 205, 5, 5, 163, 180, 5, 5, 33, 201, 5, 5, 183, 168, 5, 5, 66, 187, 5, 5, 172, 205, 5, 5, 200, 185, 5, 5, 140, 229, 5, 5, 174, 222, 5, 5, 28, 175, 5, 5, 46, 203, 5, 5, 20, 230, 5, 5, 17, 165, 5, 5, 243, 231, 5, 5, 161, 209, 5, 5, 78, 200, 5, 5, 174, 202, 5, 5, 99, 176, 5, 5, 185, 248, 5, 5, 57, 231, 5, 5, 247, 200, 5, 5, 149, 216, 5, 5, 139, 190, 5, 5, 159, 244, 5, 5, 227, 217, 5, 5, 220, 205, 5, 5, 69, 237, 5, 5, 128, 243, 5, 5, 169, 169, 5, 5, 152, 201, 5, 5, 54, 220, 5, 5, 238, 241, 5, 5, 38, 178, 5, 5, 69, 246, 5, 5, 180, 186, 5, 5, 206, 175, 5, 5, 36, 174, 5, 5, 102, 230, 5, 5, 180, 243, 5, 5, 114, 197, 5, 5, 218, 164, 5, 5, 139, 166, 5, 5, 166, 249, 5, 5, 173, 176, 5, 5, 208, 181, 5, 5, 253, 236, 5, 5, 254, 236, 5, 5, 109, 225, 5, 5, 190, 224, 5, 5, 27, 245, 5, 5, 128, 165, 5, 5, 210, 212, 5, 5, 239, 192, 5, 5, 74, 187, 5, 5, 224, 168, 5, 5, 232, 221, 5, 5, 118, 169, 5, 5, 75, 187, 5, 5, 176, 222, 5, 5, 244, 232, 5, 5, 14, 182, 5, 5, 19, 210, 5, 5, 134, 227, 5, 5, 205, 236, 5, 5, 191, 224, 5, 5, 156, 167, 5, 5, 23, 207, 5, 5, 204, 169, 5, 5, 225, 170, 5, 5, 172, 216, 5, 5, 92, 206, 5, 5, 178, 223, 5, 5, 158, 228, 5, 5, 26, 208, 5, 5, 180, 242, 5, 5, 148, 181, 5, 5, 45, 242, 5, 5, 199, 234, 5, 5, 84, 194, 5, 5, 130, 198, 5, 5, 81, 242, 5, 5, 200, 234, 5, 5, 49, 203, 5, 5, 85, 247, 5, 5, 153, 236, 5, 5, 55, 184, 5, 5, 231, 227, 5, 5, 202, 215, 5, 5, 143, 186, 5, 5, 112, 174, 5, 5, 161, 237, 5, 5, 119, 169, 5, 5, 64, 247, 5, 5, 155, 203, 5, 5, 104, 183, 5, 5, 92, 172, 5, 5, 176, 195, 5, 5, 48, 185, 5, 5, 121, 188, 5, 5, 246, 226, 5, 5, 173, 245, 5, 5, 248, 181, 5, 5, 51, 215, 5, 5, 197, 199, 5, 5, 46, 242, 5, 5, 161, 244, 5, 5, 100, 244, 5, 5, 235, 194, 5, 5, 102, 226, 5, 5, 213, 170, 5, 5, 216, 170, 5, 5, 218, 233, 5, 5, 218, 246, 5, 5, 97, 206, 5, 5, 250, 176, 5, 5, 208, 211, 5, 5, 241, 178, 5, 5, 214, 204, 5, 5, 103, 232, 5, 5, 14, 193, 5, 5, 230, 184, 5, 5, 24, 190, 5, 5, 248, 206, 5, 5, 133, 246, 5, 5, 27, 208, 5, 5, 241, 183, 5, 5, 206, 196, 5, 5, 119, 197, 5, 5, 225, 168, 5, 5, 36, 232, 5, 5, 25, 192, 5, 5, 182, 193, 5, 5, 118, 174, 5, 5, 209, 190, 5, 5, 246, 224, 5, 5, 27, 224, 5, 5, 99, 237, 5, 5, 37, 226, 5, 5, 142, 220, 5, 5, 166, 170, 5, 5, 28, 234, 5, 5, 3, 163, 5, 5, 185, 186, 5, 5, 213, 181, 5, 5, 53, 163, 5, 5, 135, 198, 5, 5, 50, 164, 5, 5, 18, 211, 5, 5, 235, 207, 5, 5, 226, 212, 5, 5, 229, 210, 5, 5, 216, 176, 5, 5, 112, 203, 5, 5, 226, 168, 5, 5, 102, 176, 5, 5, 189, 205, 5, 5, 164, 209, 5, 5, 115, 179, 5, 5, 221, 192, 5, 5, 115, 199, 5, 5, 15, 193, 5, 5, 124, 189, 5, 5, 162, 172, 5, 5, 230, 210, 5, 5, 207, 238, 5, 5, 97, 247, 5, 5, 192, 164, 5, 5, 38, 220, 5, 5, 140, 242, 5, 5, 216, 206, 5, 5, 215, 204, 5, 5, 7, 189, 5, 5, 28, 180, 5, 5, 82, 229, 5, 5, 83, 181, 5, 5, 161, 175, 5, 5, 47, 199, 5, 5, 199, 208, 5, 5, 120, 197, 5, 5, 245, 234, 5, 5, 186, 179, 5, 5, 82, 226, 5, 5, 36, 165, 5, 5, 12, 225, 5, 5, 176, 200, 5, 5, 211, 172, 5, 5, 53, 185, 5, 5, 133, 208, 5, 5, 226, 170, 5, 5, 192, 168, 5, 5, 234, 245, 5, 5, 39, 176, 5, 5, 157, 227, 5, 5, 177, 200, 5, 5, 236, 242, 5, 5, 171, 239, 5, 5, 27, 181, 5, 5, 200, 208, 5, 5, 219, 246, 5, 5, 247, 233, 5, 5, 28, 183, 5, 5, 7, 221, 5, 5, 67, 182, 5, 5, 134, 204, 5, 5, 250, 180, 5, 5, 93, 190, 5, 5, 113, 165, 5, 5, 217, 240, 5, 5, 227, 170, 5, 5, 151, 163, 5, 5, 35, 167, 5, 5, 221, 177, 5, 5, 166, 244, 5, 5, 172, 244, 5, 5, 167, 244, 5, 5, 146, 214, 5, 5, 212, 235, 5, 5, 226, 209, 5, 5, 17, 195, 5, 5, 105, 216, 5, 5, 240, 246, 5, 5, 30, 248, 5, 5, 215, 247, 5, 5, 226, 174, 5, 5, 200, 188, 5, 5, 109, 217, 5, 5, 110, 217, 5, 5, 115, 183, 5, 5, 250, 233, 5, 5, 177, 245, 5, 5, 117, 226, 5, 5, 151, 214, 5, 5, 204, 247, 5, 5, 220, 165, 5, 5, 75, 170, 5, 5, 210, 238, 5, 5, 240, 237, 5, 5, 144, 190, 5, 5, 173, 244, 5, 5, 60, 196, 5, 5, 61, 196, 5, 5, 16, 206, 5, 5, 209, 171, 5, 5, 17, 233, 5, 5, 18, 233, 5, 5, 214, 232, 5, 5, 26, 164, 5, 5, 41, 230, 5, 5, 82, 213, 5, 5, 101, 241, 5, 5, 211, 238, 5, 5, 210, 211, 5, 5, 225, 204, 5, 5, 123, 174, 5, 5, 95, 191, 5, 5, 101, 206, 5, 5, 102, 167, 5, 5, 212, 173, 5, 5, 164, 201, 5, 5, 165, 201, 5, 5, 33, 198, 5, 5, 52, 243, 5, 5, 191, 181, 5, 5, 185, 219, 5, 5, 131, 231, 5, 5, 51, 216, 5, 5, 145, 230, 5, 5, 177, 191, 5, 5, 74, 209, 5, 5, 118, 211, 5, 5, 79, 171, 5, 5, 54, 201, 5, 5, 162, 228, 5, 5, 227, 168, 5, 5, 244, 191, 5, 5, 249, 178, 5, 5, 60, 220, 5, 5, 108, 205, 5, 5, 84, 246, 5, 5, 99, 163, 5, 5, 243, 245, 5, 5, 157, 166, 5, 5, 196, 202, 5, 5, 15, 199, 5, 5, 190, 232, 5, 5, 29, 203, 5, 5, 172, 178, 5, 5, 243, 241, 5, 5, 4, 237, 5, 5, 49, 235, 5, 5, 76, 212, 5, 5, 244, 245, 5, 5, 184, 220, 5, 5, 128, 196, 5, 5, 154, 181, 5, 5, 255, 200, 5, 5, 83, 213, 5, 5, 204, 226, 5, 5, 48, 209, 5, 5, 24, 186, 5, 5, 100, 233, 5, 5, 150, 241, 5, 5, 9, 249, 5, 5, 217, 169, 5, 5, 115, 249, 5, 5, 150, 229, 5, 5, 253, 176, 5, 5, 142, 222, 5, 5, 73, 194, 5, 5, 82, 230, 5, 5, 48, 217, 5, 5, 135, 224, 5, 5, 96, 195, 5, 5, 162, 196, 5, 5, 105, 199, 5, 5, 40, 206, 5, 5, 234, 185, 5, 5, 85, 246, 5, 5, 224, 166, 5, 5, 120, 242, 5, 5, 52, 236, 5, 5, 242, 243, 5, 5, 209, 226, 5, 5, 108, 226, 5, 5, 255, 226, 5, 5, 119, 249, 5, 5, 194, 220, 5, 5, 145, 186, 5, 5, 191, 230, 5, 5, 230, 170, 5, 5, 99, 187, 5, 5, 10, 212, 5, 5, 65, 237, 5, 5, 62, 249, 5, 5, 100, 187, 5, 5, 29, 207, 5, 5, 241, 212, 5, 5, 168, 164, 5, 5, 169, 225, 5, 5, 224, 216, 5, 5, 226, 181, 5, 5, 51, 188, 5, 5, 206, 199, 5, 5, 154, 177, 5, 5, 184, 210, 5, 5, 215, 232, 5, 5, 203, 166, 5, 5, 186, 234, 5, 5, 13, 249, 5, 5, 170, 225, 5, 5, 121, 248, 5, 5, 75, 196, 5, 5, 65, 167, 5, 5, 229, 168, 5, 5, 230, 168, 5, 5, 251, 182, 5, 5, 189, 229, 5, 5, 56, 213, 5, 5, 107, 173, 5, 5, 151, 249, 5, 5, 94, 221, 5, 5, 230, 204, 5, 5, 90, 202, 5, 5, 241, 230, 5, 5, 122, 234, 5, 5, 6, 239, 5, 5, 42, 171, 5, 5, 212, 197, 5, 5, 75, 237, 5, 5, 195, 220, 5, 5, 42, 172, 5, 5, 53, 236, 5, 5, 31, 217, 5, 5, 46, 188, 5, 5, 57, 213, 5, 5, 63, 193, 5, 5, 122, 186, 5, 5, 76, 231, 5, 5, 129, 247, 5, 5, 248, 191, 5, 5, 206, 234, 5, 5, 210, 226, 5, 5, 85, 180, 5, 5, 122, 190, 5, 5, 223, 230, 5, 5, 212, 236, 5, 5, 112, 217, 5, 5, 234, 163, 5, 5, 227, 241, 5, 5, 45, 194, 5, 5, 61, 165, 5, 5, 136, 233, 5, 5, 139, 194, 5, 5, 23, 233, 5, 5, 31, 181, 5, 5, 254, 233, 5, 5, 8, 167, 5, 5, 25, 186, 5, 5, 55, 176, 5, 5, 204, 202, 5, 5, 205, 186, 5, 5, 167, 228, 5, 5, 135, 186, 5, 5, 122, 248, 5, 5, 178, 189, 5, 5, 176, 244, 5, 5, 129, 169, 5, 5, 110, 177, 5, 5, 34, 181, 5, 5, 54, 170, 5, 5, 160, 182, 5, 5, 229, 225, 5, 5, 168, 167, 5, 5, 214, 238, 5, 5, 144, 188, 5, 5, 62, 235, 5, 5, 89, 213, 5, 5, 196, 168, 5, 5, 230, 209, 5, 5, 202, 205, 5, 5, 109, 226, 5, 5, 184, 203, 5, 5, 35, 246, 5, 5, 74, 182, 5, 5, 40, 234, 5, 5, 132, 248, 5, 5, 137, 240, 5, 5, 164, 231, 5, 5, 13, 188, 5, 5, 69, 165, 5, 5, 239, 189, 5, 5, 44, 217, 5, 5, 143, 166, 5, 5, 26, 221, 5, 5, 91, 185, 5, 5, 63, 211, 5, 5, 148, 170, 5, 5, 190, 242, 5, 5, 136, 189, 5, 5, 197, 202, 5, 5, 158, 165, 5, 5, 141, 194, 5, 5, 109, 198, 5, 5, 184, 176, 5, 5, 184, 236, 5, 5, 47, 190, 5, 5, 203, 227, 5, 5, 239, 228, 5, 5, 185, 176, 5, 5, 134, 171, 5, 5, 194, 217, 5, 5, 149, 170, 5, 5, 113, 163, 5, 5, 10, 177, 5, 5, 187, 234, 5, 5, 179, 219, 5, 5, 168, 228, 5, 5, 175, 217, 5, 5, 141, 246, 5, 5, 22, 199, 5, 5, 116, 238, 5, 5, 81, 171, 5, 5, 242, 185, 5, 5, 160, 234, 5, 5, 215, 222, 5, 5, 29, 234, 5, 5, 181, 166, 5, 5, 45, 206, 5, 5, 114, 171, 5, 5, 63, 235, 5, 5, 37, 184, 5, 5, 111, 200, 5, 5, 8, 208, 5, 5, 193, 243, 5, 5, 122, 205, 5, 5, 62, 184, 5, 5, 92, 229, 5, 5, 147, 243, 5, 5, 171, 193, 5, 5, 203, 224, 5, 5, 86, 199, 5, 5, 142, 227, 5, 5, 157, 195, 5, 5, 187, 210, 5, 5, 26, 169, 5, 5, 179, 232, 5, 5, 28, 232, 5, 5, 64, 211, 5, 5, 186, 175, 5, 5, 176, 188, 5, 5, 28, 189, 5, 5, 158, 195, 5, 5, 48, 238, 5, 5, 70, 165, 5, 5, 210, 186, 5, 5, 101, 246, 5, 5, 103, 202, 5, 5, 140, 210, 5, 5, 53, 229, 5, 5, 247, 172, 5, 5, 236, 170, 5, 5, 192, 238, 5, 5, 145, 224, 5, 5, 156, 177, 5, 5, 22, 218, 5, 5, 11, 204, 5, 5, 2, 231, 5, 5, 3, 231, 5, 5, 127, 185, 5, 5, 234, 247, 5, 5, 174, 209, 5, 5, 173, 225, 5, 5, 124, 199, 5, 5, 241, 204, 5, 5, 172, 169, 5, 5, 157, 249, 5, 5, 4, 164, 5, 5, 63, 242, 5, 5, 213, 211, 5, 5, 61, 190, 5, 5, 23, 230, 5, 5, 121, 200, 5, 5, 253, 237, 5, 5, 185, 180, 5, 5, 191, 172, 5, 5, 181, 207, 5, 5, 48, 233, 5, 5, 145, 172, 5, 5, 195, 230, 5, 5, 114, 173, 5, 5, 173, 196, 5, 5, 233, 229, 5, 5, 253, 230, 5, 5, 17, 241, 5, 5, 214, 216, 5, 5, 240, 174, 5, 5, 182, 207, 5, 5, 113, 176, 5, 5, 36, 245, 5, 5, 255, 178, 5, 5, 142, 235, 5, 5, 166, 214, 5, 5, 149, 213, 5, 5, 247, 190, 5, 5, 9, 247, 5, 5, 11, 228, 5, 5, 222, 208, 5, 5, 242, 204, 5, 5, 198, 249, 5, 5, 76, 237, 5, 5, 149, 189, 5, 5, 150, 248, 5, 5, 216, 213, 5, 5, 134, 247, 5, 5, 108, 202, 5, 5, 175, 201, 5, 5, 193, 235, 5, 5, 201, 184, 5, 5, 130, 185, 5, 5, 123, 217, 5, 5, 200, 249, 5, 5, 151, 248, 5, 5, 63, 213, 5, 5, 16, 198, 5, 5, 12, 164, 5, 5, 214, 197, 5, 5, 79, 238, 5, 5, 250, 245, 5, 5, 127, 200, 5, 5, 128, 200, 5, 5, 225, 202, 5, 5, 195, 233, 5, 5, 56, 216, 5, 5, 244, 246, 5, 5, 137, 171, 5, 5, 174, 240, 5, 5, 158, 186, 5, 5, 26, 206, 5, 5, 101, 234, 5, 5, 173, 228, 5, 5, 70, 211, 5, 5, 155, 188, 5, 5, 28, 195, 5, 5, 73, 249, 5, 5, 38, 175, 5, 5, 237, 180, 5, 5, 61, 180, 5, 5, 53, 217, 5, 5, 88, 165, 5, 5, 94, 220, 5, 5, 77, 237, 5, 5, 67, 215, 5, 5, 27, 173, 5, 5, 208, 177, 5, 5, 99, 214, 5, 5, 71, 193, 5, 5, 100, 214, 5, 5, 118, 229, 5, 5, 135, 205, 5, 5, 215, 197, 5, 5, 123, 201, 5, 5, 216, 235, 5, 5, 9, 216, 5, 5, 199, 217, 5, 5, 250, 241, 5, 5, 180, 163, 5, 5, 71, 236, 5, 5, 119, 229, 5, 5, 203, 217, 5, 5, 151, 178, 5, 5, 145, 168, 5, 5, 106, 246, 5, 5, 128, 184, 5, 5, 44, 224, 5, 5, 166, 169, 5, 5, 204, 230, 5, 5, 82, 168, 5, 5, 130, 179, 5, 5, 229, 202, 5, 5, 74, 232, 5, 5, 207, 224, 5, 5, 205, 220, 5, 5, 64, 239, 5, 5, 197, 183, 5, 5, 192, 244, 5, 5, 37, 192, 5, 5, 93, 235, 5, 5, 176, 214, 5, 5, 232, 208, 5, 5, 137, 172, 5, 5, 59, 205, 5, 5, 38, 217, 5, 5, 139, 196, 5, 5, 111, 202, 5, 5, 226, 165, 5, 5, 115, 177, 5, 5, 122, 171, 5, 5, 39, 186, 5, 5, 254, 205, 5, 5, 126, 234, 5, 5, 163, 171, 5, 5, 160, 212, 5, 5, 120, 209, 5, 5, 165, 215, 5, 5, 3, 188, 5, 5, 193, 238, 5, 5, 127, 215, 5, 5, 195, 236, 5, 5, 54, 191, 5, 5, 184, 240, 5, 5, 39, 217, 5, 5, 48, 214, 5, 5, 226, 186, 5, 5, 185, 240, 5, 5, 240, 184, 5, 5, 135, 185, 5, 5, 108, 204, 5, 5, 193, 176, 5, 5, 176, 218, 5, 5, 193, 163, 5, 5, 144, 180, 5, 5, 255, 241, 5, 5, 148, 169, 5, 5, 242, 200, 5, 5, 170, 165, 5, 5, 251, 172, 5, 5, 74, 236, 5, 5, 16, 180, 5, 5, 75, 182, 5, 5, 161, 212, 5, 5, 181, 209, 5, 5, 156, 246, 5, 5, 75, 188, 5, 5, 139, 200, 5, 5, 196, 238, 5, 5, 38, 192, 5, 5, 204, 249, 5, 5, 66, 164, 5, 5, 142, 187, 5, 5, 121, 229, 5, 5, 181, 225, 5, 5, 52, 238, 5, 5, 35, 189, 5, 5, 76, 232, 5, 5, 129, 217, 5, 5, 24, 211, 5, 5, 82, 185, 5, 5, 224, 199, 5, 5, 71, 182, 5, 5, 131, 245, 5, 5, 5, 231, 5, 5, 6, 231, 5, 5, 243, 208, 5, 5, 209, 198, 5, 5, 214, 189, 5, 5, 136, 236, 5, 5, 87, 182, 5, 5, 253, 170, 5, 5, 207, 249, 5, 5, 208, 168, 5, 5, 45, 248, 5, 5, 206, 217, 5, 5, 240, 202, 5, 5, 89, 208, 5, 5, 130, 217, 5, 5, 85, 169, 5, 5, 49, 210, 5, 5, 137, 236, 5, 5, 90, 208, 5, 5, 15, 174, 5, 5, 14, 184, 5, 5, 147, 163, 5, 5, 155, 230, 5, 5, 200, 240, 5, 5, 171, 215, 5, 5, 250, 247, 5, 5, 14, 207, 5, 5, 211, 188, 5, 5, 184, 188, 5, 5, 230, 243, 5, 5, 72, 211, 5, 5, 145, 244, 5, 5, 9, 191, 5, 5, 223, 179, 5, 5, 111, 246, 5, 5, 240, 248, 5, 5, 158, 189, 5, 5, 224, 211, 5, 5, 205, 241, 5, 5, 36, 173, 5, 5, 232, 169, 5, 5, 159, 198, 5, 5, 113, 196, 5, 5, 52, 186, 5, 5, 243, 202, 5, 5, 119, 176, 5, 5, 158, 170, 5, 5, 160, 185, 5, 5, 106, 185, 5, 5, 36, 200, 5, 5, 230, 173, 5, 5, 98, 218, 5, 5, 161, 246, 5, 5, 142, 196, 5, 5, 155, 220, 5, 5, 245, 199, 5, 5, 240, 186, 5, 5, 247, 202, 5, 5, 63, 171, 5, 5, 139, 185, 5, 5, 251, 199, 5, 5, 113, 190, 5, 5, 32, 202, 5, 5, 143, 196, 5, 5, 80, 190, 5, 5, 9, 205, 5, 5, 218, 235, 5, 5, 151, 180, 5, 5, 113, 185, 5, 5, 106, 212, 5, 5, 188, 214, 5, 5, 209, 240, 5, 5, 180, 190, 5, 5, 170, 170, 5, 5, 122, 165, 5, 5, 217, 198, 5, 5, 42, 199, 5, 5, 229, 189, 5, 5, 222, 184, 5, 5, 88, 191, 5, 5, 112, 235, 5, 5, 134, 214, 5, 5, 122, 210, 5, 5, 218, 225, 5, 5, 233, 198, 5, 5, 191, 216, 5, 5, 180, 234, 5, 5, 113, 174, 5, 5, 192, 216, 5, 5, 199, 216, 5, 5, 234, 189, 5, 5, 48, 199, 5, 5, 216, 247, 5, 5, 6, 198, 5, 5, 255, 203, 5, 5, 83, 207, 5, 5, 172, 193, 5, 5, 57, 164, 5, 5, 192, 242, 5, 5, 148, 220, 5, 5, 149, 220, 5, 5, 109, 196, 5, 5, 27, 166, 5, 5, 154, 213, 5, 5, 165, 229, 5, 5, 166, 226, 5, 5, 252, 209, 5, 5, 75, 177, 5, 5, 129, 202, 5, 5, 255, 209, 5, 5, 196, 177, 5, 5, 187, 226, 5, 5, 169, 213, 5, 5, 135, 207, 5, 5, 132, 219, 5, 5, 179, 227, 5, 5, 145, 247, 5, 5, 80, 198, 5, 5, 178, 231, 5, 5, 146, 201, 5, 5, 6, 200, 5, 5, 223, 170, 5, 5, 204, 191, 5, 5, 248, 236, 5, 5, 171, 213, 5, 5, 217, 223, 5, 5, 240, 226, 5, 5, 222, 239, 5, 5, 33, 169, 5, 5, 97, 216, 5, 5, 154, 173, 5, 5, 210, 237, 5, 5, 0, 147, 5, 5, 70, 223, 5, 5, 81, 210, 5, 5, 27, 247, 5, 5, 28, 247, 5, 5, 11, 231, 5, 5, 160, 164, 5, 5, 246, 167, 5, 5, 211, 248, 5, 5, 212, 248, 5, 5, 22, 208, 5, 5, 179, 208, 5, 5, 176, 182, 5, 5, 209, 243, 5, 5, 180, 208, 5, 5, 74, 234, 5, 5, 41, 249, 5, 5, 140, 207, 5, 5, 48, 200, 5, 5, 225, 210, 5, 5, 181, 208, 5, 5, 246, 222, 5, 5, 248, 240, 5, 5, 217, 197, 5, 5, 90, 172, 5, 5, 84, 247, 5, 5, 101, 173, 5, 5, 102, 173, 5, 5, 181, 198, 5, 5, 200, 194, 5, 5, 70, 232, 5, 5, 205, 174, 5, 5, 61, 230, 5, 5, 221, 203, 5, 5, 99, 174, 5, 5, 8, 223, 5, 5, 194, 175, 5, 5, 47, 164, 5, 5, 146, 215, 5, 5, 160, 183, 5, 5, 96, 179, 5, 5, 223, 209, 5, 5, 47, 242, 5, 5, 114, 219, 5, 5, 188, 168, 5, 5, 41, 223, 5, 5, 112, 242, 5, 5, 115, 197, 5, 5, 64, 198, 5, 5, 69, 207, 5, 5, 76, 187, 5, 5, 211, 174, 5, 5, 64, 230, 5, 5, 84, 219, 5, 5, 77, 192, 5, 5, 251, 167, 5, 5, 252, 167, 5, 5, 98, 215, 5, 5, 233, 221, 5, 5, 80, 177, 5, 5, 129, 165, 5, 5, 72, 166, 5, 5, 13, 193, 5, 5, 81, 220, 5, 5, 169, 211, 5, 5, 180, 227, 5, 5, 253, 180, 5, 5, 121, 189, 5, 5, 48, 193, 5, 5, 34, 176, 5, 5, 49, 193, 5, 5, 162, 244, 5, 5, 77, 187, 5, 5, 59, 225, 5, 5, 174, 211, 5, 5, 91, 170, 5, 5, 173, 191, 5, 5, 219, 233, 5, 5, 52, 200, 5, 5, 53, 200, 5, 5, 207, 170, 5, 5, 219, 174, 5, 5, 39, 244, 5, 5, 143, 245, 5, 5, 198, 187, 5, 5, 145, 221, 5, 5, 105, 239, 5, 5, 102, 215, 5, 5, 58, 196, 5, 5, 250, 203, 5, 5, 83, 187, 5, 5, 96, 240, 5, 5, 193, 168, 5, 5, 134, 208, 5, 5, 73, 197, 5, 5, 181, 227, 5, 5, 172, 239, 5, 5, 98, 247, 5, 5, 121, 169, 5, 5, 145, 212, 5, 5, 109, 209, 5, 5, 19, 194, 5, 5, 146, 212, 5, 5, 244, 218, 5, 5, 108, 170, 5, 5, 222, 185, 5, 5, 145, 219, 5, 5, 52, 203, 5, 5, 53, 183, 5, 5, 118, 235, 5, 5, 201, 208, 5, 5, 127, 216, 5, 5, 227, 203, 5, 5, 33, 210, 5, 5, 237, 227, 5, 5, 17, 250, 5, 5, 56, 203, 5, 5, 98, 182, 5, 5, 180, 223, 5, 5, 174, 163, 5, 5, 169, 237, 5, 5, 137, 198, 5, 5, 113, 229, 5, 5, 139, 217, 5, 5, 213, 173, 5, 5, 220, 228, 5, 5, 221, 228, 5, 5, 172, 174, 5, 5, 213, 215, 5, 5, 122, 219, 5, 5, 32, 245, 5, 5, 124, 174, 5, 5, 194, 186, 5, 5, 123, 186, 5, 5, 47, 165, 5, 5, 163, 225, 5, 5, 10, 249, 5, 5, 84, 242, 5, 5, 171, 203, 5, 5, 44, 176, 5, 5, 189, 242, 5, 5, 155, 181, 5, 5, 174, 177, 5, 5, 99, 182, 5, 5, 54, 200, 5, 5, 178, 245, 5, 5, 90, 218, 5, 5, 102, 237, 5, 5, 165, 172, 5, 5, 222, 206, 5, 5, 112, 171, 5, 5, 215, 208, 5, 5, 216, 208, 5, 5, 123, 248, 5, 5, 65, 208, 5, 5, 245, 184, 5, 5, 13, 171, 5, 5, 46, 211, 5, 5, 128, 235, 5, 5, 242, 212, 5, 5, 212, 190, 5, 5, 101, 187, 5, 5, 53, 177, 5, 5, 32, 181, 5, 5, 189, 179, 5, 5, 213, 196, 5, 5, 98, 198, 5, 5, 141, 198, 5, 5, 124, 248, 5, 5, 169, 164, 5, 5, 178, 172, 5, 5, 10, 168, 5, 5, 246, 245, 5, 5, 185, 178, 5, 5, 125, 240, 5, 5, 14, 249, 5, 5, 243, 236, 5, 5, 81, 212, 5, 5, 158, 210, 5, 5, 15, 231, 5, 5, 169, 183, 5, 5, 103, 211, 5, 5, 136, 246, 5, 5, 64, 189, 5, 5, 185, 193, 5, 5, 178, 246, 5, 5, 243, 178, 5, 5, 245, 237, 5, 5, 94, 164, 5, 5, 5, 177, 5, 5, 101, 247, 5, 5, 78, 234, 5, 5, 72, 244, 5, 5, 177, 244, 5, 5, 151, 193, 5, 5, 88, 224, 5, 5, 225, 177, 5, 5, 246, 184, 5, 5, 26, 195, 5, 5, 172, 203, 5, 5, 248, 213, 5, 5, 158, 181, 5, 5, 79, 170, 5, 5, 211, 216, 5, 5, 183, 171, 5, 5, 226, 247, 5, 5, 120, 203, 5, 5, 2, 227, 5, 5, 117, 217, 5, 5, 63, 201, 5, 5, 246, 242, 5, 5, 205, 165, 5, 5, 43, 230, 5, 5, 247, 245, 5, 5, 232, 174, 5, 5, 66, 249, 5, 5, 14, 170, 5, 5, 142, 246, 5, 5, 137, 174, 5, 5, 193, 230, 5, 5, 31, 214, 5, 5, 61, 242, 5, 5, 39, 210, 5, 5, 38, 170, 5, 5, 181, 213, 5, 5, 121, 244, 5, 5, 212, 226, 5, 5, 171, 226, 5, 5, 57, 180, 5, 5, 150, 221, 5, 5, 110, 233, 5, 5, 181, 180, 5, 5, 233, 174, 5, 5, 67, 249, 5, 5, 132, 196, 5, 5, 231, 170, 5, 5, 249, 184, 5, 5, 251, 177, 5, 5, 23, 199, 5, 5, 231, 209, 5, 5, 55, 177, 5, 5, 103, 187, 5, 5, 180, 245, 5, 5, 121, 249, 5, 5, 242, 211, 5, 5, 117, 183, 5, 5, 243, 185, 5, 5, 87, 226, 5, 5, 78, 231, 5, 5, 180, 244, 5, 5, 21, 206, 5, 5, 114, 164, 5, 5, 125, 186, 5, 5, 123, 167, 5, 5, 143, 235, 5, 5, 13, 213, 5, 5, 222, 233, 5, 5, 14, 213, 5, 5, 72, 249, 5, 5, 119, 247, 5, 5, 149, 188, 5, 5, 58, 236, 5, 5, 79, 164, 5, 5, 81, 231, 5, 5, 82, 231, 5, 5, 104, 202, 5, 5, 8, 245, 5, 5, 26, 215, 5, 5, 152, 198, 5, 5, 248, 210, 5, 5, 192, 239, 5, 5, 213, 164, 5, 5, 125, 199, 5, 5, 96, 180, 5, 5, 240, 239, 5, 5, 43, 178, 5, 5, 30, 183, 5, 5, 141, 197, 5, 5, 27, 199, 5, 5, 240, 193, 5, 5, 38, 169, 5, 5, 46, 172, 5, 5, 163, 240, 5, 5, 40, 247, 5, 5, 183, 244, 5, 5, 100, 179, 5, 5, 217, 222, 5, 5, 124, 234, 5, 5, 39, 236, 5, 5, 69, 209, 5, 5, 131, 170, 5, 5, 60, 181, 5, 5, 240, 228, 5, 5, 103, 227, 5, 5, 43, 234, 5, 5, 24, 223, 5, 5, 16, 177, 5, 5, 121, 238, 5, 5, 62, 190, 5, 5, 151, 246, 5, 5, 36, 219, 5, 5, 99, 180, 5, 5, 206, 224, 5, 5, 74, 168, 5, 5, 217, 211, 5, 5, 14, 168, 5, 5, 238, 247, 5, 5, 136, 205, 5, 5, 101, 190, 5, 5, 247, 185, 5, 5, 10, 211, 5, 5, 218, 222, 5, 5, 227, 243, 5, 5, 27, 206, 5, 5, 203, 229, 5, 5, 43, 247, 5, 5, 170, 214, 5, 5, 245, 246, 5, 5, 156, 219, 5, 5, 181, 219, 5, 5, 27, 169, 5, 5, 28, 169, 5, 5, 155, 193, 5, 5, 66, 188, 5, 5, 40, 170, 5, 5, 20, 213, 5, 5, 68, 235, 5, 5, 216, 186, 5, 5, 132, 170, 5, 5, 171, 214, 5, 5, 51, 225, 5, 5, 29, 221, 5, 5, 220, 170, 5, 5, 113, 216, 5, 5, 144, 169, 5, 5, 29, 233, 5, 5, 145, 235, 5, 5, 236, 209, 5, 5, 101, 179, 5, 5, 136, 218, 5, 5, 177, 167, 5, 5, 178, 167, 5, 5, 73, 204, 5, 5, 219, 211, 5, 5, 118, 165, 5, 5, 252, 187, 5, 5, 73, 167, 5, 5, 26, 194, 5, 5, 97, 213, 5, 5, 143, 202, 5, 5, 128, 222, 5, 5, 75, 201, 5, 5, 179, 172, 5, 5, 187, 185, 5, 5, 186, 207, 5, 5, 187, 207, 5, 5, 134, 200, 5, 5, 197, 243, 5, 5, 103, 224, 5, 5, 18, 204, 5, 5, 117, 228, 5, 5, 59, 183, 5, 5, 177, 214, 5, 5, 37, 245, 5, 5, 173, 226, 5, 5, 247, 217, 5, 5, 146, 235, 5, 5, 241, 225, 5, 5, 81, 225, 5, 5, 79, 176, 5, 5, 16, 187, 5, 5, 111, 214, 5, 5, 119, 228, 5, 5, 246, 239, 5, 5, 138, 206, 5, 5, 40, 175, 5, 5, 107, 224, 5, 5, 63, 209, 5, 5, 164, 233, 5, 5, 88, 216, 5, 5, 207, 227, 5, 5, 128, 190, 5, 5, 252, 210, 5, 5, 197, 244, 5, 5, 221, 223, 5, 5, 131, 187, 5, 5, 58, 232, 5, 5, 127, 217, 5, 5, 163, 188, 5, 5, 179, 221, 5, 5, 13, 232, 5, 5, 249, 213, 5, 5, 221, 170, 5, 5, 112, 214, 5, 5, 244, 208, 5, 5, 93, 231, 5, 5, 67, 205, 5, 5, 19, 171, 5, 5, 237, 197, 5, 5, 205, 230, 5, 5, 225, 199, 5, 5, 71, 222, 5, 5, 194, 203, 5, 5, 148, 168, 5, 5, 82, 197, 5, 5, 141, 171, 5, 5, 37, 173, 5, 5, 104, 185, 5, 5, 193, 203, 5, 5, 4, 185, 5, 5, 21, 172, 5, 5, 233, 186, 5, 5, 120, 226, 5, 5, 148, 187, 5, 5, 242, 209, 5, 5, 18, 184, 5, 5, 220, 189, 5, 5, 47, 178, 5, 5, 184, 214, 5, 5, 164, 221, 5, 5, 17, 191, 5, 5, 149, 187, 5, 5, 18, 191, 5, 5, 231, 173, 5, 5, 59, 241, 5, 5, 205, 164, 5, 5, 131, 183, 5, 5, 78, 193, 5, 5, 229, 207, 5, 5, 150, 244, 5, 5, 168, 229, 5, 5, 41, 213, 5, 5, 82, 234, 5, 5, 231, 211, 5, 5, 96, 212, 5, 5, 48, 178, 5, 5, 83, 234, 5, 5, 125, 170, 5, 5, 234, 211, 5, 5, 195, 203, 5, 5, 187, 221, 5, 5, 18, 222, 5, 5, 78, 217, 5, 5, 32, 201, 5, 5, 58, 184, 5, 5, 232, 166, 5, 5, 219, 184, 5, 5, 28, 216, 5, 5, 253, 222, 5, 5, 183, 169, 5, 5, 76, 239, 5, 5, 242, 194, 5, 5, 142, 171, 5, 5, 163, 213, 5, 5, 139, 216, 5, 5, 221, 242, 5, 5, 187, 195, 5, 5, 96, 249, 5, 5, 7, 212, 5, 5, 122, 189, 5, 5, 167, 217, 5, 5, 144, 229, 5, 5, 99, 215, 5, 5, 218, 182, 5, 5, 102, 216, 5, 5, 113, 203, 5, 5, 107, 215, 5, 5, 3, 241, 5, 5, 24, 199, 5, 5, 143, 246, 5, 5, 167, 175, 5, 5, 161, 226, 5, 5, 147, 246, 5, 5, 169, 175, 5, 5, 161, 222, 5, 5, 207, 198, 5, 5, 232, 211, 5, 5, 87, 239, 5, 5, 85, 239, 5, 5, 112, 166, 5, 5, 136, 192, 5, 5, 176, 247, 5, 5, 182, 165, 5, 5, 158, 223, 5, 5, 151, 239, 5, 5, 52, 228, 5, 5, 222, 224, 5, 5, 32, 174, 5, 5, 130, 194, 5, 5, 80, 173, 5, 5, 196, 229, 5, 5, 148, 203, 5, 5, 241, 182, 5, 5, 209, 222, 5, 5, 10, 176, 5, 5, 66, 240, 5, 5, 247, 222, 5, 5, 210, 222, 5, 5, 222, 203, 5, 5, 168, 218, 5, 5, 11, 214, 5, 5, 62, 230, 5, 5, 78, 218, 5, 5, 131, 230, 5, 5, 203, 210, 5, 5, 173, 205, 5, 5, 23, 208, 5, 5, 42, 249, 5, 5, 4, 170, 5, 5, 218, 223, 5, 5, 133, 238, 5, 5, 32, 184, 5, 5, 10, 200, 5, 5, 63, 230, 5, 5, 22, 203, 5, 5, 34, 182, 5, 5, 170, 189, 5, 5, 99, 175, 5, 5, 48, 164, 5, 5, 20, 167, 5, 5, 240, 241, 5, 5, 16, 233, 5, 5, 25, 165, 5, 5, 7, 228, 5, 5, 26, 180, 5, 5, 70, 207, 5, 5, 156, 197, 5, 5, 64, 217, 5, 5, 18, 215, 5, 5, 221, 220, 5, 5, 168, 239, 5, 5, 167, 166, 5, 5, 51, 163, 5, 5, 153, 176, 5, 5, 158, 245, 5, 5, 212, 174, 5, 5, 133, 194, 5, 5, 232, 218, 5, 5, 233, 218, 5, 5, 65, 230, 5, 5, 88, 167, 5, 5, 79, 240, 5, 5, 65, 170, 5, 5, 15, 182, 5, 5, 245, 182, 5, 5, 57, 192, 5, 5, 168, 217, 5, 5, 142, 219, 5, 5, 170, 211, 5, 5, 220, 163, 5, 5, 134, 194, 5, 5, 49, 213, 5, 5, 223, 183, 5, 5, 91, 163, 5, 5, 52, 163, 5, 5, 119, 174, 5, 5, 73, 223, 5, 5, 74, 223, 5, 5, 32, 196, 5, 5, 140, 230, 5, 5, 59, 206, 5, 5, 53, 203, 5, 5, 101, 205, 5, 5, 94, 163, 5, 5, 162, 203, 5, 5, 40, 239, 5, 5, 147, 191, 5, 5, 190, 230, 5, 5, 146, 221, 5, 5, 227, 193, 5, 5, 136, 243, 5, 5, 12, 234, 5, 5, 199, 187, 5, 5, 249, 194, 5, 5, 178, 216, 5, 5, 179, 222, 5, 5, 72, 176, 5, 5, 200, 216, 5, 5, 207, 196, 5, 5, 207, 195, 5, 5, 35, 184, 5, 5, 151, 229, 5, 5, 57, 203, 5, 5, 236, 196, 5, 5, 123, 225, 5, 5, 7, 187, 5, 5, 222, 207, 5, 5, 163, 228, 5, 5, 43, 167, 5, 5, 178, 191, 5, 5, 115, 240, 5, 5, 108, 248, 5, 5, 226, 220, 5, 5, 41, 188, 5, 5, 34, 210, 5, 5, 125, 212, 5, 5, 198, 202, 5, 5, 116, 240, 5, 5, 116, 166, 5, 5, 180, 222, 5, 5, 216, 229, 5, 5, 139, 227, 5, 5, 126, 212, 5, 5, 8, 187, 5, 5, 134, 188, 5, 5, 173, 188, 5, 5, 33, 196, 5, 5, 42, 205, 5, 5, 229, 172, 5, 5, 73, 224, 5, 5, 62, 165, 5, 5, 214, 196, 5, 5, 190, 179, 5, 5, 104, 249, 5, 5, 135, 189, 5, 5, 75, 223, 5, 5, 248, 224, 5, 5, 35, 224, 5, 5, 59, 178, 5, 5, 184, 165, 5, 5, 65, 189, 5, 5, 6, 229, 5, 5, 240, 166, 5, 5, 241, 166, 5, 5, 166, 172, 5, 5, 15, 249, 5, 5, 123, 190, 5, 5, 186, 193, 5, 5, 243, 242, 5, 5, 249, 177, 5, 5, 230, 180, 5, 5, 10, 219, 5, 5, 223, 206, 5, 5, 236, 239, 5, 5, 205, 202, 5, 5, 118, 218, 5, 5, 137, 193, 5, 5, 208, 195, 5, 5, 68, 209, 5, 5, 233, 180, 5, 5, 182, 180, 5, 5, 234, 174, 5, 5, 172, 226, 5, 5, 88, 214, 5, 5, 224, 165, 5, 5, 56, 177, 5, 5, 217, 196, 5, 5, 160, 178, 5, 5, 23, 215, 5, 5, 24, 215, 5, 5, 164, 197, 5, 5, 105, 249, 5, 5, 3, 164, 5, 5, 31, 164, 5, 5, 235, 174, 5, 5, 227, 247, 5, 5, 126, 182, 5, 5, 247, 242, 5, 5, 30, 200, 5, 5, 237, 163, 5, 5, 193, 225, 5, 5, 77, 191, 5, 5, 36, 167, 5, 5, 247, 175, 5, 5, 174, 249, 5, 5, 40, 179, 5, 5, 64, 192, 5, 5, 15, 175, 5, 5, 237, 199, 5, 5, 83, 231, 5, 5, 107, 249, 5, 5, 225, 230, 5, 5, 31, 183, 5, 5, 111, 208, 5, 5, 234, 236, 5, 5, 80, 164, 5, 5, 100, 182, 5, 5, 78, 223, 5, 5, 119, 236, 5, 5, 39, 184, 5, 5, 123, 244, 5, 5, 232, 203, 5, 5, 97, 210, 5, 5, 238, 199, 5, 5, 37, 196, 5, 5, 156, 188, 5, 5, 32, 186, 5, 5, 217, 186, 5, 5, 170, 219, 5, 5, 218, 226, 5, 5, 83, 191, 5, 5, 175, 240, 5, 5, 21, 177, 5, 5, 101, 214, 5, 5, 68, 215, 5, 5, 152, 246, 5, 5, 188, 165, 5, 5, 44, 234, 5, 5, 11, 202, 5, 5, 79, 211, 5, 5, 113, 236, 5, 5, 6, 164, 5, 5, 57, 216, 5, 5, 116, 165, 5, 5, 147, 174, 5, 5, 60, 205, 5, 5, 150, 222, 5, 5, 61, 205, 5, 5, 2, 235, 5, 5, 112, 202, 5, 5, 254, 175, 5, 5, 228, 231, 5, 5, 29, 173, 5, 5, 128, 163, 5, 5, 114, 216, 5, 5, 15, 216, 5, 5, 131, 165, 5, 5, 24, 230, 5, 5, 133, 170, 5, 5, 40, 186, 5, 5, 137, 238, 5, 5, 14, 232, 5, 5, 59, 194, 5, 5, 53, 234, 5, 5, 52, 237, 5, 5, 204, 184, 5, 5, 76, 224, 5, 5, 183, 188, 5, 5, 97, 165, 5, 5, 181, 205, 5, 5, 80, 176, 5, 5, 122, 229, 5, 5, 36, 180, 5, 5, 100, 165, 5, 5, 48, 186, 5, 5, 242, 168, 5, 5, 206, 230, 5, 5, 177, 218, 5, 5, 149, 178, 5, 5, 124, 229, 5, 5, 162, 245, 5, 5, 108, 227, 5, 5, 166, 163, 5, 5, 42, 214, 5, 5, 138, 163, 5, 5, 234, 224, 5, 5, 85, 185, 5, 5, 19, 191, 5, 5, 72, 213, 5, 5, 80, 215, 5, 5, 55, 186, 5, 5, 171, 219, 5, 5, 121, 175, 5, 5, 108, 207, 5, 5, 164, 232, 5, 5, 224, 226, 5, 5, 75, 205, 5, 5, 93, 167, 5, 5, 210, 240, 5, 5, 101, 186, 5, 5, 22, 241, 5, 5, 169, 212, 5, 5, 170, 212, 5, 5, 235, 241, 5, 5, 250, 173, 5, 5, 249, 179, 5, 5, 254, 174, 5, 5, 16, 174, 5, 5, 247, 174, 5, 5, 27, 163, 5, 5, 147, 197, 5, 5, 33, 202, 5, 5, 38, 202, 5, 5, 78, 205, 5, 5, 68, 194, 5, 5, 242, 165, 5, 5, 34, 243, 5, 5, 223, 193, 5, 5, 146, 207, 5, 5, 119, 168, 5, 5, 66, 170, 5, 5, 99, 171, 5, 5, 100, 171, 5, 5, 131, 171, 5, 5, 66, 227, 5, 5, 183, 235, 5, 5, 71, 224, 5, 5, 94, 190, 5, 5, 158, 207, 5, 5, 106, 171, 5, 5, 57, 181, 5, 5, 135, 188, 5, 5, 245, 241, 5, 5, 11, 219, 5, 5, 32, 208, 5, 5, 58, 213, 5, 5, 84, 228, 5, 5, 248, 245, 5, 5, 145, 213, 5, 5, 113, 181, 5, 5, 235, 247, 5, 5, 74, 186, 5, 5, 254, 237, 5, 5, 28, 178, 5, 5, 211, 210, 5, 5, 129, 167, 5, 5, 226, 202, 5, 5, 47, 172, 5, 5, 188, 244, 5, 5, 100, 192, 5, 5, 190, 211, 5, 5, 150, 220, 5, 5, 255, 205, 5, 5, 77, 194, 5, 5, 130, 176, 5, 5, 143, 205, 5, 5, 103, 190, 5, 5, 5, 212, 5, 5, 253, 204, 5, 5, 43, 246, 5, 5, 197, 207, 5, 5, 194, 211, 5, 5, 106, 190, 5, 5, 80, 186, 5, 5, 203, 228, 5, 5, 197, 176, 5, 5, 83, 194, 5, 5, 76, 205, 5, 5, 65, 194, 5, 5, 23, 243, 5, 5, 211, 193, 5, 5, 91, 171, 5, 5, 56, 227, 5, 5, 68, 224, 5, 5, 63, 186, 5, 5, 136, 207, 5, 5, 175, 202, 5, 5, 50, 213, 5, 5, 174, 176, 5, 5, 108, 181, 5, 5, 217, 247, 5, 5, 97, 192, 5, 5, 241, 205, 5, 5, 253, 211, 5, 5, 27, 204, 5, 5, 3, 210, 5, 5, 168, 205, 5, 5, 145, 234, 5, 5, 147, 207, 5, 5, 234, 218, 5, 5, 8, 224, 5, 5, 72, 202, 5, 5, 16, 193, 5, 5, 40, 176, 5, 5, 132, 231, 5, 5, 217, 176, 5, 5, 231, 228, 5, 5, 143, 199, 5, 5, 77, 212, 5, 5, 179, 191, 5, 5, 83, 227, 5, 5, 199, 203, 5, 5, 21, 194, 5, 5, 207, 194, 5, 5, 223, 225, 5, 5, 200, 203, 5, 5, 27, 201, 5, 5, 36, 210, 5, 5, 128, 216, 5, 5, 174, 236, 5, 5, 170, 183, 5, 5, 249, 201, 5, 5, 48, 190, 5, 5, 22, 219, 5, 5, 76, 219, 5, 5, 81, 165, 5, 5, 133, 216, 5, 5, 30, 199, 5, 5, 241, 228, 5, 5, 213, 194, 5, 5, 44, 219, 5, 5, 72, 217, 5, 5, 236, 225, 5, 5, 36, 213, 5, 5, 205, 239, 5, 5, 155, 187, 5, 5, 201, 207, 5, 5, 39, 213, 5, 5, 89, 243, 5, 5, 43, 213, 5, 5, 157, 242, 5, 5, 158, 242, 5, 5, 215, 177, 5, 5, 200, 174, 5, 5, 4, 210, 5, 5, 5, 210, 5, 5, 86, 204, 5, 5, 113, 201, 5, 5, 169, 202, 5, 5, 239, 172, 5, 5, 8, 218, 5, 5, 100, 177, 5, 5, 185, 201, 5, 5, 219, 223, 5, 5, 202, 179, 5, 5, 189, 216, 5, 5, 149, 203, 5, 5, 162, 239, 5, 5, 132, 230, 5, 5, 9, 218, 5, 5, 173, 243, 5, 5, 91, 228, 5, 5, 41, 209, 5, 5, 10, 182, 5, 5, 176, 202, 5, 5, 10, 218, 5, 5, 174, 243, 5, 5, 162, 237, 5, 5, 82, 173, 5, 5, 36, 241, 5, 5, 146, 241, 5, 5, 126, 237, 5, 5, 192, 228, 5, 5, 181, 243, 5, 5, 255, 215, 5, 5, 148, 207, 5, 5, 194, 214, 5, 5, 5, 183, 5, 5, 192, 224, 5, 5, 48, 242, 5, 5, 170, 218, 5, 5, 41, 218, 5, 5, 178, 173, 5, 5, 29, 216, 5, 5, 229, 223, 5, 5, 194, 227, 5, 5, 52, 215, 5, 5, 202, 235, 5, 5, 166, 222, 5, 5, 233, 216, 5, 5, 80, 240, 5, 5, 81, 240, 5, 5, 158, 236, 5, 5, 14, 223, 5, 5, 80, 197, 5, 5, 161, 228, 5, 5, 121, 197, 5, 5, 20, 217, 5, 5, 222, 220, 5, 5, 236, 221, 5, 5, 119, 221, 5, 5, 89, 192, 5, 5, 133, 231, 5, 5, 186, 231, 5, 5, 84, 220, 5, 5, 50, 193, 5, 5, 220, 174, 5, 5, 34, 205, 5, 5, 44, 177, 5, 5, 125, 189, 5, 5, 218, 176, 5, 5, 84, 244, 5, 5, 84, 187, 5, 5, 54, 183, 5, 5, 89, 167, 5, 5, 59, 220, 5, 5, 160, 192, 5, 5, 97, 240, 5, 5, 72, 207, 5, 5, 242, 241, 5, 5, 115, 219, 5, 5, 173, 239, 5, 5, 174, 239, 5, 5, 240, 211, 5, 5, 237, 219, 5, 5, 194, 168, 5, 5, 37, 165, 5, 5, 49, 218, 5, 5, 175, 211, 5, 5, 4, 163, 5, 5, 32, 175, 5, 5, 237, 232, 5, 5, 153, 179, 5, 5, 197, 221, 5, 5, 179, 233, 5, 5, 146, 219, 5, 5, 12, 250, 5, 5, 235, 216, 5, 5, 234, 217, 5, 5, 117, 240, 5, 5, 93, 201, 5, 5, 125, 197, 5, 5, 84, 227, 5, 5, 151, 223, 5, 5, 224, 214, 5, 5, 205, 226, 5, 5, 169, 198, 5, 5, 171, 223, 5, 5, 93, 192, 5, 5, 22, 194, 5, 5, 78, 212, 5, 5, 96, 200, 5, 5, 173, 203, 5, 5, 144, 199, 5, 5, 225, 191, 5, 5, 74, 216, 5, 5, 245, 245, 5, 5, 164, 228, 5, 5, 61, 220, 5, 5, 184, 182, 5, 5, 41, 171, 5, 5, 86, 220, 5, 5, 56, 172, 5, 5, 241, 237, 5, 5, 22, 200, 5, 5, 185, 242, 5, 5, 42, 218, 5, 5, 136, 193, 5, 5, 78, 239, 5, 5, 95, 166, 5, 5, 49, 181, 5, 5, 179, 239, 5, 5, 123, 185, 5, 5, 147, 221, 5, 5, 7, 197, 5, 5, 41, 196, 5, 5, 33, 245, 5, 5, 109, 248, 5, 5, 136, 180, 5, 5, 206, 202, 5, 5, 12, 219, 5, 5, 146, 230, 5, 5, 156, 234, 5, 5, 136, 186, 5, 5, 170, 237, 5, 5, 14, 197, 5, 5, 94, 238, 5, 5, 90, 227, 5, 5, 176, 181, 5, 5, 51, 176, 5, 5, 235, 163, 5, 5, 89, 224, 5, 5, 117, 242, 5, 5, 252, 195, 5, 5, 198, 209, 5, 5, 209, 210, 5, 5, 174, 229, 5, 5, 213, 238, 5, 5, 33, 248, 5, 5, 226, 214, 5, 5, 124, 225, 5, 5, 229, 203, 5, 5, 219, 215, 5, 5, 6, 245, 5, 5, 120, 245, 5, 5, 50, 192, 5, 5, 128, 237, 5, 5, 106, 166, 5, 5, 102, 200, 5, 5, 175, 206, 5, 5, 177, 229, 5, 5, 252, 220, 5, 5, 178, 229, 5, 5, 250, 177, 5, 5, 162, 199, 5, 5, 48, 225, 5, 5, 207, 229, 5, 5, 185, 203, 5, 5, 71, 195, 5, 5, 182, 211, 5, 5, 88, 235, 5, 5, 198, 214, 5, 5, 35, 237, 5, 5, 102, 247, 5, 5, 116, 215, 5, 5, 189, 172, 5, 5, 21, 223, 5, 5, 72, 179, 5, 5, 138, 240, 5, 5, 217, 183, 5, 5, 194, 243, 5, 5, 111, 177, 5, 5, 29, 168, 5, 5, 55, 222, 5, 5, 141, 224, 5, 5, 174, 224, 5, 5, 160, 222, 5, 5, 214, 202, 5, 5, 144, 208, 5, 5, 4, 180, 5, 5, 103, 222, 5, 5, 96, 246, 5, 5, 255, 206, 5, 5, 133, 211, 5, 5, 252, 178, 5, 5, 124, 221, 5, 5, 3, 232, 5, 5, 107, 232, 5, 5, 95, 210, 5, 5, 56, 222, 5, 5, 78, 189, 5, 5, 123, 205, 5, 5, 230, 225, 5, 5, 230, 197, 5, 5, 223, 190, 5, 5, 194, 228, 5, 5, 185, 165, 5, 5, 248, 199, 5, 5, 114, 168, 5, 5, 228, 247, 5, 5, 58, 177, 5, 5, 215, 216, 5, 5, 122, 168, 5, 5, 100, 227, 5, 5, 121, 217, 5, 5, 101, 227, 5, 5, 28, 201, 5, 5, 152, 199, 5, 5, 255, 237, 5, 5, 254, 222, 5, 5, 14, 177, 5, 5, 35, 214, 5, 5, 139, 174, 5, 5, 252, 168, 5, 5, 161, 181, 5, 5, 130, 205, 5, 5, 19, 209, 5, 5, 139, 248, 5, 5, 147, 231, 5, 5, 127, 222, 5, 5, 92, 213, 5, 5, 122, 200, 5, 5, 149, 243, 5, 5, 218, 238, 5, 5, 121, 215, 5, 5, 108, 183, 5, 5, 205, 224, 5, 5, 251, 173, 5, 5, 154, 221, 5, 5, 48, 244, 5, 5, 67, 243, 5, 5, 66, 201, 5, 5, 122, 217, 5, 5, 191, 191, 5, 5, 172, 184, 5, 5, 145, 208, 5, 5, 95, 170, 5, 5, 117, 241, 5, 5, 148, 246, 5, 5, 179, 216, 5, 5, 142, 180, 5, 5, 27, 220, 5, 5, 113, 228, 5, 5, 54, 205, 5, 5, 184, 166, 5, 5, 174, 196, 5, 5, 96, 201, 5, 5, 212, 199, 5, 5, 156, 191, 5, 5, 189, 244, 5, 5, 47, 171, 5, 5, 210, 214, 5, 5, 119, 171, 5, 5, 96, 213, 5, 5, 50, 211, 5, 5, 177, 224, 5, 5, 86, 217, 5, 5, 189, 180, 5, 5, 193, 236, 5, 5, 173, 164, 5, 5, 55, 223, 5, 5, 98, 210, 5, 5, 104, 227, 5, 5, 48, 206, 5, 5, 74, 249, 5, 5, 102, 214, 5, 5, 171, 197, 5, 5, 168, 179, 5, 5, 198, 241, 5, 5, 199, 241, 5, 5, 58, 219, 5, 5, 244, 247, 5, 5, 167, 165, 5, 5, 98, 235, 5, 5, 146, 211, 5, 5, 219, 186, 5, 5, 81, 169, 5, 5, 145, 169, 5, 5, 242, 228, 5, 5, 158, 219, 5, 5, 172, 171, 5, 5, 62, 205, 5, 5, 11, 232, 5, 5, 123, 203, 5, 5, 157, 188, 5, 5, 248, 185, 5, 5, 102, 234, 5, 5, 177, 238, 5, 5, 152, 194, 5, 5, 65, 213, 5, 5, 67, 222, 5, 5, 123, 238, 5, 5, 124, 238, 5, 5, 205, 205, 5, 5, 159, 219, 5, 5, 203, 239, 5, 5, 245, 198, 5, 5, 71, 206, 5, 5, 118, 170, 5, 5, 251, 223, 5, 5, 158, 198, 5, 5, 126, 199, 5, 5, 149, 224, 5, 5, 216, 216, 5, 5, 206, 232, 5, 5, 249, 235, 5, 5, 29, 213, 5, 5, 78, 219, 5, 5, 153, 178, 5, 5, 127, 221, 5, 5, 61, 183, 5, 5, 41, 186, 5, 5, 179, 237, 5, 5, 140, 244, 5, 5, 189, 207, 5, 5, 18, 202, 5, 5, 83, 233, 5, 5, 75, 204, 5, 5, 146, 163, 5, 5, 140, 200, 5, 5, 151, 222, 5, 5, 40, 245, 5, 5, 169, 232, 5, 5, 92, 219, 5, 5, 6, 227, 5, 5, 207, 209, 5, 5, 17, 180, 5, 5, 170, 232, 5, 5, 3, 169, 5, 5, 35, 229, 5, 5, 131, 217, 5, 5, 90, 232, 5, 5, 20, 180, 5, 5, 86, 176, 5, 5, 53, 237, 5, 5, 170, 215, 5, 5, 22, 202, 5, 5, 112, 204, 5, 5, 140, 206, 5, 5, 152, 220, 5, 5, 129, 183, 5, 5, 123, 239, 5, 5, 87, 238, 5, 5, 99, 220, 5, 5, 53, 168, 5, 5, 63, 200, 5, 5, 156, 199, 5, 5, 193, 240, 5, 5, 194, 240, 5, 5, 120, 228, 5, 5, 170, 230, 5, 5, 93, 183, 5, 5, 163, 210, 5, 5, 110, 222, 5, 5, 86, 179, 5, 5, 92, 199, 5, 5, 30, 220, 5, 5, 238, 197, 5, 5, 88, 238, 5, 5, 146, 244, 5, 5, 147, 205, 5, 5, 166, 248, 5, 5, 73, 205, 5, 5, 147, 200, 5, 5, 12, 178, 5, 5, 162, 219, 5, 5, 182, 217, 5, 5, 160, 232, 5, 5, 138, 185, 5, 5, 204, 228, 5, 5, 162, 246, 5, 5, 155, 224, 5, 5, 160, 187, 5, 5, 207, 239, 5, 5, 2, 166, 5, 5, 163, 219, 5, 5, 90, 243, 5, 5, 13, 238, 5, 5, 50, 224, 5, 5, 165, 246, 5, 5, 51, 224, 5, 5, 5, 242, 5, 5, 118, 224, 5, 5, 171, 230, 5, 5, 232, 243, 5, 5, 58, 186, 5, 5, 235, 224, 5, 5, 43, 214, 5, 5, 22, 184, 5, 5, 164, 200, 5, 5, 165, 232, 5, 5, 212, 240, 5, 5, 91, 176, 5, 5, 106, 220, 5, 5, 100, 243, 5, 5, 21, 178, 5, 5, 155, 242, 5, 5, 149, 202, 5, 5, 210, 177, 5, 5, 187, 174, 5, 5, 107, 230, 5, 5, 196, 201, 5, 5, 129, 203, 5, 5, 184, 201, 5, 5, 201, 179, 5, 5, 168, 230, 5, 5, 85, 228, 5, 5, 22, 209, 5, 5, 251, 181, 5, 5, 252, 217, 5, 5, 5, 240, 5, 5, 253, 217, 5, 5, 150, 202, 5, 5, 110, 230, 5, 5, 35, 206, 5, 5, 67, 200, 5, 5, 163, 238, 5, 5, 41, 207, 5, 5, 23, 241, 5, 5, 2, 183, 5, 5, 225, 216, 5, 5, 151, 237, 5, 5, 221, 190, 5, 5, 33, 218, 5, 5, 121, 237, 5, 5, 248, 173, 5, 5, 187, 227, 5, 5, 167, 173, 5, 5, 121, 207, 5, 5, 113, 221, 5, 5, 34, 193, 5, 5, 76, 197, 5, 5, 168, 168, 5, 5, 229, 232, 5, 5, 231, 226, 5, 5, 47, 218, 5, 5, 107, 219, 5, 5, 142, 236, 5, 5, 209, 220, 5, 5, 201, 176, 5, 5, 255, 219, 5, 5, 167, 233, 5, 5, 155, 211, 5, 5, 226, 219, 5, 5, 252, 164, 5, 5, 21, 240, 5, 5, 134, 239, 5, 5, 251, 196, 5, 5, 118, 185, 5, 5, 44, 227, 5, 5, 129, 193, 5, 5, 188, 226, 5, 5, 53, 202, 5, 5, 254, 199, 5, 5, 91, 225, 5, 5, 133, 221, 5, 5, 143, 228, 5, 5, 137, 199, 5, 5, 26, 176, 5, 5, 81, 192, 5, 5, 55, 212, 5, 5, 3, 194, 5, 5, 216, 217, 5, 5, 38, 240, 5, 5, 68, 221, 5, 5, 213, 191, 5, 5, 68, 216, 5, 5, 166, 223, 5, 5, 18, 201, 5, 5, 228, 216, 5, 5, 5, 220, 5, 5, 218, 203, 5, 5, 252, 244, 5, 5, 12, 248, 5, 5, 191, 209, 5, 5, 97, 245, 5, 5, 43, 192, 5, 5, 107, 242, 5, 5, 74, 200, 5, 5, 12, 197, 5, 5, 175, 229, 5, 5, 123, 237, 5, 5, 93, 238, 5, 5, 127, 211, 5, 5, 139, 221, 5, 5, 98, 232, 5, 5, 160, 184, 5, 5, 242, 177, 5, 5, 31, 222, 5, 5, 182, 172, 5, 5, 59, 195, 5, 5, 62, 246, 5, 5, 184, 247, 5, 5, 205, 229, 5, 5, 188, 228, 5, 5, 101, 177, 5, 5, 128, 224, 5, 5, 71, 189, 5, 5, 159, 222, 5, 5, 24, 237, 5, 5, 245, 179, 5, 5, 189, 228, 5, 5, 67, 240, 5, 5, 90, 170, 5, 5, 193, 224, 5, 5, 35, 177, 5, 5, 91, 201, 5, 5, 96, 205, 5, 5, 15, 214, 5, 5, 84, 217, 5, 5, 95, 241, 5, 5, 126, 246, 5, 5, 220, 237, 5, 5, 79, 248, 5, 5, 99, 217, 5, 5, 79, 224, 5, 5, 121, 231, 5, 5, 205, 238, 5, 5, 38, 243, 5, 5, 240, 176, 5, 5, 40, 171, 5, 5, 120, 168, 5, 5, 193, 216, 5, 5, 68, 169, 5, 5, 84, 176, 5, 5, 159, 236, 5, 5, 79, 214, 5, 5, 171, 241, 5, 5, 164, 164, 5, 5, 75, 227, 5, 5, 82, 210, 5, 5, 102, 171, 5, 5, 192, 205, 5, 5, 110, 170, 5, 5, 148, 219, 5, 5, 43, 205, 5, 5, 171, 171, 5, 5, 253, 179, 5, 5, 168, 232, 5, 5, 166, 209, 5, 5, 26, 186, 5, 5, 251, 168, 5, 5, 167, 207, 5, 5, 59, 243, 5, 5, 160, 210, 5, 5, 88, 220, 5, 5, 232, 224, 5, 5, 35, 193, 5, 5, 74, 210, 5, 5, 155, 201, 5, 5, 173, 213, 5, 5, 73, 202, 5, 5, 166, 201, 5, 5, 68, 198, 5, 5, 120, 201, 5, 5, 21, 213, 5, 5, 22, 213, 5, 5, 171, 199, 5, 5, 253, 187, 5, 5, 111, 198, 5, 5, 13, 178, 5, 5, 20, 191, 5, 5, 198, 175, 5, 5, 13, 206, 5, 5, 7, 209, 5, 5, 67, 227, 5, 5, 9, 187, 5, 5, 123, 176, 5, 5, 180, 246, 5, 5, 212, 164, 5, 5, 158, 229, 5, 5, 230, 215, 5, 5, 222, 173, 5, 5, 47, 219, 5, 5, 54, 219, 5, 5, 101, 243, 5, 5, 103, 243, 5, 5, 12, 227, 5, 5, 207, 225, 5, 5, 11, 200, 5, 5, 142, 182, 5, 5, 82, 240, 5, 5, 143, 182, 5, 5, 83, 240, 5, 5, 210, 204, 5, 5, 0, 151, 5, 5, 55, 198, 5, 5, 230, 237, 5, 5, 158, 225, 5, 5, 103, 216, 5, 5, 234, 200, 5, 5, 226, 204, 5, 5, 200, 195, 5, 5, 226, 191, 5, 5, 76, 207, 5, 5, 46, 174, 5, 5, 243, 212, 5, 5, 20, 166, 5, 5, 225, 197, 5, 5, 140, 248, 5, 5, 155, 221, 5, 5, 34, 203, 5, 5, 115, 173, 5, 5, 23, 213, 5, 5, 121, 163, 5, 5, 11, 247, 5, 5, 192, 200, 5, 5, 28, 166, 5, 5, 190, 176, 5, 5, 115, 233, 5, 5, 120, 173, 5, 5, 186, 240, 5, 5, 184, 218, 5, 5, 44, 164, 5, 5, 211, 223, 5, 5, 244, 238, 5, 5, 253, 167, 5, 5, 246, 182, 5, 5, 216, 167, 5, 5, 78, 192, 5, 5, 248, 182, 5, 5, 114, 242, 5, 5, 82, 177, 5, 5, 122, 169, 5, 5, 186, 235, 5, 5, 249, 214, 5, 5, 181, 200, 5, 5, 209, 196, 5, 5, 126, 197, 5, 5, 222, 197, 5, 5, 123, 169, 5, 5, 142, 198, 5, 5, 77, 170, 5, 5, 200, 186, 5, 5, 171, 183, 5, 5, 243, 164, 5, 5, 39, 248, 5, 5, 35, 203, 5, 5, 106, 182, 5, 5, 18, 170, 5, 5, 15, 168, 5, 5, 24, 213, 5, 5, 4, 238, 5, 5, 100, 207, 5, 5, 146, 169, 5, 5, 152, 225, 5, 5, 18, 180, 5, 5, 165, 200, 5, 5, 211, 207, 5, 5, 201, 163, 5, 5, 164, 220, 5, 5, 36, 249, 5, 5, 201, 174, 5, 5, 11, 234, 5, 5, 234, 194, 5, 5, 95, 179, 5, 5, 163, 239, 5, 5, 184, 198, 5, 5, 238, 246, 5, 5, 138, 168, 5, 5, 198, 199, 5, 5, 61, 173, 5, 5, 233, 179, 5, 5, 103, 194, 5, 5, 249, 231, 5, 5, 113, 218, 5, 5, 75, 239, 5, 5, 237, 221, 5, 5, 76, 227, 5, 5, 38, 165, 5, 5, 134, 231, 5, 5, 77, 239, 5, 5, 42, 194, 5, 5, 121, 168, 5, 5, 219, 165, 5, 5, 170, 169, 5, 5, 102, 175, 5, 5, 47, 229, 5, 5, 33, 197, 5, 5, 212, 172, 5, 5, 251, 180, 5, 5, 175, 239, 5, 5, 44, 223, 5, 5, 69, 170, 5, 5, 157, 197, 5, 5, 39, 165, 5, 5, 34, 190, 5, 5, 26, 189, 5, 5, 97, 200, 5, 5, 227, 191, 5, 5, 112, 168, 5, 5, 135, 231, 5, 5, 80, 203, 5, 5, 23, 217, 5, 5, 55, 232, 5, 5, 50, 204, 5, 5, 245, 191, 5, 5, 176, 232, 5, 5, 109, 205, 5, 5, 115, 244, 5, 5, 18, 218, 5, 5, 253, 220, 5, 5, 174, 227, 5, 5, 21, 166, 5, 5, 22, 166, 5, 5, 37, 210, 5, 5, 175, 227, 5, 5, 75, 226, 5, 5, 178, 232, 5, 5, 23, 220, 5, 5, 69, 185, 5, 5, 148, 232, 5, 5, 221, 233, 5, 5, 175, 224, 5, 5, 248, 228, 5, 5, 114, 181, 5, 5, 242, 166, 5, 5, 227, 192, 5, 5, 117, 215, 5, 5, 185, 169, 5, 5, 124, 205, 5, 5, 36, 237, 5, 5, 236, 174, 5, 5, 48, 180, 5, 5, 89, 235, 5, 5, 90, 235, 5, 5, 23, 219, 5, 5, 250, 201, 5, 5, 54, 221, 5, 5, 29, 178, 5, 5, 247, 205, 5, 5, 43, 218, 5, 5, 119, 226, 5, 5, 175, 249, 5, 5, 33, 164, 5, 5, 38, 232, 5, 5, 71, 185, 5, 5, 10, 174, 5, 5, 129, 234, 5, 5, 12, 202, 5, 5, 54, 229, 5, 5, 176, 240, 5, 5, 149, 222, 5, 5, 243, 174, 5, 5, 101, 192, 5, 5, 14, 236, 5, 5, 12, 232, 5, 5, 254, 187, 5, 5, 179, 240, 5, 5, 177, 163, 5, 5, 219, 224, 5, 5, 246, 246, 5, 5, 41, 219, 5, 5, 109, 222, 5, 5, 171, 165, 5, 5, 124, 249, 5, 5, 211, 232, 5, 5, 62, 176, 5, 5, 155, 232, 5, 5, 77, 243, 5, 5, 15, 169, 5, 5, 121, 228, 5, 5, 249, 246, 5, 5, 201, 244, 5, 5, 2, 206, 5, 5, 165, 192, 5, 5, 133, 215, 5, 5, 180, 232, 5, 5, 105, 165, 5, 5, 14, 178, 5, 5, 106, 229, 5, 5, 88, 243, 5, 5, 249, 228, 5, 5, 183, 217, 5, 5, 44, 191, 5, 5, 167, 192, 5, 5, 182, 232, 5, 5, 198, 163, 5, 5, 156, 220, 5, 5, 190, 174, 5, 5, 28, 249, 5, 5, 231, 194, 5, 5, 10, 234, 5, 5, 240, 231, 5, 5, 86, 215, 5, 5, 169, 163, 5, 5, 46, 221, 5, 5, 187, 199, 5, 5, 193, 218, 5, 5, 51, 173, 5, 5, 229, 179, 5, 5, 136, 168, 5, 5, 214, 165, 5, 5, 82, 192, 5, 5, 216, 221, 5, 5, 97, 194, 5, 5, 74, 239, 5, 5, 84, 205, 5, 5, 117, 164, 5, 5, 57, 227, 5, 5, 248, 180, 5, 5, 35, 194, 5, 5, 29, 197, 5, 5, 92, 175, 5, 5, 63, 170, 5, 5, 202, 172, 5, 5, 154, 197, 5, 5, 152, 239, 5, 5, 53, 232, 5, 5, 213, 218, 5, 5, 73, 203, 5, 5, 79, 200, 5, 5, 116, 231, 5, 5, 24, 189, 5, 5, 10, 190, 5, 5, 175, 232, 5, 5, 117, 231, 5, 5, 221, 191, 5, 5, 221, 205, 5, 5, 20, 210, 5, 5, 247, 220, 5, 5, 65, 185, 5, 5, 170, 227, 5, 5, 221, 174, 5, 5, 25, 178, 5, 5, 231, 201, 5, 5, 102, 229, 5, 5, 247, 228, 5, 5, 220, 233, 5, 5, 40, 165, 5, 5, 18, 220, 5, 5, 222, 192, 5, 5, 238, 166, 5, 5, 12, 176, 5, 5, 68, 185, 5, 5, 6, 174, 5, 5, 101, 222, 5, 5, 228, 174, 5, 5, 144, 222, 5, 5, 12, 236, 5, 5, 4, 232, 5, 5, 68, 243, 5, 5, 149, 232, 5, 5, 122, 249, 5, 5, 209, 232, 5, 5, 114, 228, 5, 5, 190, 244, 5, 5, 166, 192, 5, 5, 84, 223, 5, 5, 211, 212, 5, 5, 195, 227, 5, 5, 111, 183, 5, 5, 185, 182, 5, 5, 136, 213, 5, 5, 65, 225, 5, 5, 162, 197, 5, 5, 76, 225, 5, 5, 169, 219, 5, 5, 214, 213, 5, 5, 100, 180, 5, 5, 116, 232, 5, 5, 115, 232, 5, 5, 51, 187, 5, 5, 128, 204, 5, 5, 202, 174, 5, 5, 231, 219, 5, 5, 137, 192, 5, 5, 8, 209, 5, 5, 92, 228, 5, 5, 138, 212, 5, 5, 210, 243, 5, 5, 148, 215, 5, 5, 157, 233, 5, 5, 149, 207, 5, 5, 4, 211, 5, 5, 42, 223, 5, 5, 159, 233, 5, 5, 224, 224, 5, 5, 225, 224, 5, 5, 45, 177, 5, 5, 217, 206, 5, 5, 232, 223, 5, 5, 155, 176, 5, 5, 132, 213, 5, 5, 85, 220, 5, 5, 159, 207, 5, 5, 221, 248, 5, 5, 176, 211, 5, 5, 85, 202, 5, 5, 45, 223, 5, 5, 164, 225, 5, 5, 113, 193, 5, 5, 85, 217, 5, 5, 136, 231, 5, 5, 29, 181, 5, 5, 86, 178, 5, 5, 203, 187, 5, 5, 143, 192, 5, 5, 213, 236, 5, 5, 157, 234, 5, 5, 59, 213, 5, 5, 243, 219, 5, 5, 175, 236, 5, 5, 221, 243, 5, 5, 2, 207, 5, 5, 125, 205, 5, 5, 34, 212, 5, 5, 127, 180, 5, 5, 9, 226, 5, 5, 227, 234, 5, 5, 49, 190, 5, 5, 254, 211, 5, 5, 111, 176, 5, 5, 233, 209, 5, 5, 148, 231, 5, 5, 186, 180, 5, 5, 188, 211, 5, 5, 117, 223, 5, 5, 77, 201, 5, 5, 180, 209, 5, 5, 154, 209, 5, 5, 109, 204, 5, 5, 91, 232, 5, 5, 242, 180, 5, 5, 66, 239, 5, 5, 142, 232, 5, 5, 144, 232, 5, 5, 217, 232, 5, 5, 154, 166, 5, 5, 45, 164, 5, 5, 28, 196, 5, 5, 8, 195, 5, 5, 201, 178, 5, 5, 150, 207, 5, 5, 37, 174, 5, 5, 211, 224, 5, 5, 44, 205, 5, 5, 182, 173, 5, 5, 171, 218, 5, 5, 36, 179, 5, 5, 217, 248, 5, 5, 21, 163, 5, 5, 190, 208, 5, 5, 218, 248, 5, 5, 34, 211, 5, 5, 68, 182, 5, 5, 116, 179, 5, 5, 98, 240, 5, 5, 76, 184, 5, 5, 126, 189, 5, 5, 8, 170, 5, 5, 5, 163, 5, 5, 20, 177, 5, 5, 15, 196, 5, 5, 36, 182, 5, 5, 25, 203, 5, 5, 160, 233, 5, 5, 25, 167, 5, 5, 76, 209, 5, 5, 238, 221, 5, 5, 218, 206, 5, 5, 152, 177, 5, 5, 119, 218, 5, 5, 41, 174, 5, 5, 60, 167, 5, 5, 73, 207, 5, 5, 59, 228, 5, 5, 140, 163, 5, 5, 166, 183, 5, 5, 74, 207, 5, 5, 85, 187, 5, 5, 170, 192, 5, 5, 180, 239, 5, 5, 236, 201, 5, 5, 110, 178, 5, 5, 154, 215, 5, 5, 251, 203, 5, 5, 111, 178, 5, 5, 148, 221, 5, 5, 79, 212, 5, 5, 49, 196, 5, 5, 80, 222, 5, 5, 43, 194, 5, 5, 150, 212, 5, 5, 149, 196, 5, 5, 75, 216, 5, 5, 104, 223, 5, 5, 35, 190, 5, 5, 113, 168, 5, 5, 254, 218, 5, 5, 58, 203, 5, 5, 58, 181, 5, 5, 23, 234, 5, 5, 102, 205, 5, 5, 64, 179, 5, 5, 133, 209, 5, 5, 48, 165, 5, 5, 56, 215, 5, 5, 112, 178, 5, 5, 237, 201, 5, 5, 254, 206, 5, 5, 207, 202, 5, 5, 88, 229, 5, 5, 227, 181, 5, 5, 225, 225, 5, 5, 176, 206, 5, 5, 245, 183, 5, 5, 157, 214, 5, 5, 142, 166, 5, 5, 65, 217, 5, 5, 26, 235, 5, 5, 79, 172, 5, 5, 89, 229, 5, 5, 51, 189, 5, 5, 19, 250, 5, 5, 54, 204, 5, 5, 36, 224, 5, 5, 137, 209, 5, 5, 93, 170, 5, 5, 73, 179, 5, 5, 126, 205, 5, 5, 220, 208, 5, 5, 181, 245, 5, 5, 97, 246, 5, 5, 245, 225, 5, 5, 186, 203, 5, 5, 98, 202, 5, 5, 249, 183, 5, 5, 224, 206, 5, 5, 225, 206, 5, 5, 213, 190, 5, 5, 50, 190, 5, 5, 164, 186, 5, 5, 235, 199, 5, 5, 87, 207, 5, 5, 52, 223, 5, 5, 226, 246, 5, 5, 84, 212, 5, 5, 156, 180, 5, 5, 114, 163, 5, 5, 225, 232, 5, 5, 117, 209, 5, 5, 237, 239, 5, 5, 161, 234, 5, 5, 111, 233, 5, 5, 25, 234, 5, 5, 112, 233, 5, 5, 194, 167, 5, 5, 24, 219, 5, 5, 97, 221, 5, 5, 155, 177, 5, 5, 130, 218, 5, 5, 236, 247, 5, 5, 73, 178, 5, 5, 220, 178, 5, 5, 69, 226, 5, 5, 22, 250, 5, 5, 237, 222, 5, 5, 15, 177, 5, 5, 101, 219, 5, 5, 217, 224, 5, 5, 25, 226, 5, 5, 127, 205, 5, 5, 118, 193, 5, 5, 97, 180, 5, 5, 32, 207, 5, 5, 140, 174, 5, 5, 226, 232, 5, 5, 59, 177, 5, 5, 115, 168, 5, 5, 44, 230, 5, 5, 173, 184, 5, 5, 187, 180, 5, 5, 159, 223, 5, 5, 92, 182, 5, 5, 55, 205, 5, 5, 152, 175, 5, 5, 177, 234, 5, 5, 148, 208, 5, 5, 79, 179, 5, 5, 51, 211, 5, 5, 84, 184, 5, 5, 135, 177, 5, 5, 190, 180, 5, 5, 109, 202, 5, 5, 3, 184, 5, 5, 70, 167, 5, 5, 30, 188, 5, 5, 152, 232, 5, 5, 93, 165, 5, 5, 165, 226, 5, 5, 63, 190, 5, 5, 255, 187, 5, 5, 166, 234, 5, 5, 62, 180, 5, 5, 117, 175, 5, 5, 194, 236, 5, 5, 98, 213, 5, 5, 151, 196, 5, 5, 16, 202, 5, 5, 137, 218, 5, 5, 53, 211, 5, 5, 191, 176, 5, 5, 74, 167, 5, 5, 116, 233, 5, 5, 158, 188, 5, 5, 203, 164, 5, 5, 64, 232, 5, 5, 86, 184, 5, 5, 132, 165, 5, 5, 86, 182, 5, 5, 10, 207, 5, 5, 131, 176, 5, 5, 143, 234, 5, 5, 71, 215, 5, 5, 30, 173, 5, 5, 182, 225, 5, 5, 221, 222, 5, 5, 196, 178, 5, 5, 62, 163, 5, 5, 169, 234, 5, 5, 170, 234, 5, 5, 190, 207, 5, 5, 191, 207, 5, 5, 34, 189, 5, 5, 43, 179, 5, 5, 155, 209, 5, 5, 113, 207, 5, 5, 4, 169, 5, 5, 222, 187, 5, 5, 197, 178, 5, 5, 8, 235, 5, 5, 37, 197, 5, 5, 92, 232, 5, 5, 41, 175, 5, 5, 101, 165, 5, 5, 102, 165, 5, 5, 180, 221, 5, 5, 102, 207, 5, 5, 183, 225, 5, 5, 208, 209, 5, 5, 202, 178, 5, 5, 48, 224, 5, 5, 26, 202, 5, 5, 77, 245, 5, 5, 241, 248, 5, 5, 87, 184, 5, 5, 36, 189, 5, 5, 212, 187, 5, 5, 139, 218, 5, 5, 156, 187, 5, 5, 51, 219, 5, 5, 115, 224, 5, 5, 185, 225, 5, 5, 180, 178, 5, 5, 163, 246, 5, 5, 21, 191, 5, 5, 37, 189, 5, 5, 78, 215, 5, 5, 166, 246, 5, 5, 57, 233, 5, 5, 59, 186, 5, 5, 117, 214, 5, 5, 160, 230, 5, 5, 206, 235, 5, 5, 23, 184, 5, 5, 89, 179, 5, 5, 233, 211, 5, 5, 234, 203, 5, 5, 10, 235, 5, 5, 191, 234, 5, 5, 103, 213, 5, 5, 147, 196, 5, 5, 212, 207, 5, 5, 38, 189, 5, 5, 137, 228, 5, 5, 201, 194, 5, 5, 236, 175, 5, 5, 144, 173, 5, 5, 171, 211, 5, 5, 205, 180, 5, 5, 87, 178, 5, 5, 88, 178, 5, 5, 209, 194, 5, 5, 124, 186, 5, 5, 149, 230, 5, 5, 124, 190, 5, 5, 49, 238, 5, 5, 251, 201, 5, 5, 3, 207, 5, 5, 116, 180, 5, 5, 68, 197, 5, 5, 214, 185, 5, 5, 249, 242, 5, 5, 16, 170, 5, 5, 223, 195, 5, 5, 191, 211, 5, 5, 249, 185, 5, 5, 217, 185, 5, 5, 94, 178, 5, 5, 138, 218, 5, 5, 107, 243, 5, 5, 208, 180, 5, 5, 174, 175, 5, 5, 238, 244, 5, 5, 178, 248, 5, 5, 174, 223, 5, 5, 27, 241, 5, 5, 10, 194, 5, 5, 20, 208, 5, 5, 124, 246, 5, 5, 213, 248, 5, 5, 26, 216, 5, 5, 95, 175, 5, 5, 97, 217, 5, 5, 55, 173, 5, 5, 133, 230, 5, 5, 55, 228, 5, 5, 13, 197, 5, 5, 78, 229, 5, 5, 144, 182, 5, 5, 194, 224, 5, 5, 100, 175, 5, 5, 221, 237, 5, 5, 247, 182, 5, 5, 191, 208, 5, 5, 28, 170, 5, 5, 30, 231, 5, 5, 136, 230, 5, 5, 126, 210, 5, 5, 133, 222, 5, 5, 115, 201, 5, 5, 234, 198, 5, 5, 209, 211, 5, 5, 34, 197, 5, 5, 219, 176, 5, 5, 127, 189, 5, 5, 103, 175, 5, 5, 100, 238, 5, 5, 223, 220, 5, 5, 62, 173, 5, 5, 8, 197, 5, 5, 68, 244, 5, 5, 143, 167, 5, 5, 225, 163, 5, 5, 17, 193, 5, 5, 18, 193, 5, 5, 21, 167, 5, 5, 239, 221, 5, 5, 100, 231, 5, 5, 231, 237, 5, 5, 6, 163, 5, 5, 33, 220, 5, 5, 172, 218, 5, 5, 172, 241, 5, 5, 10, 195, 5, 5, 99, 240, 5, 5, 245, 218, 5, 5, 85, 227, 5, 5, 197, 168, 5, 5, 41, 239, 5, 5, 226, 194, 5, 5, 251, 206, 5, 5, 96, 204, 5, 5, 167, 194, 5, 5, 128, 197, 5, 5, 97, 195, 5, 5, 80, 212, 5, 5, 20, 220, 5, 5, 49, 232, 5, 5, 255, 218, 5, 5, 173, 227, 5, 5, 77, 189, 5, 5, 126, 248, 5, 5, 238, 227, 5, 5, 59, 242, 5, 5, 241, 246, 5, 5, 218, 243, 5, 5, 151, 219, 5, 5, 225, 246, 5, 5, 175, 174, 5, 5, 104, 211, 5, 5, 24, 220, 5, 5, 172, 183, 5, 5, 156, 186, 5, 5, 153, 219, 5, 5, 146, 220, 5, 5, 79, 189, 5, 5, 145, 231, 5, 5, 31, 200, 5, 5, 103, 247, 5, 5, 10, 226, 5, 5, 188, 193, 5, 5, 181, 195, 5, 5, 238, 239, 5, 5, 170, 167, 5, 5, 188, 210, 5, 5, 142, 194, 5, 5, 82, 184, 5, 5, 203, 168, 5, 5, 90, 194, 5, 5, 238, 163, 5, 5, 214, 190, 5, 5, 215, 166, 5, 5, 53, 189, 5, 5, 36, 231, 5, 5, 45, 230, 5, 5, 105, 202, 5, 5, 49, 244, 5, 5, 122, 215, 5, 5, 63, 226, 5, 5, 60, 234, 5, 5, 89, 226, 5, 5, 192, 191, 5, 5, 64, 242, 5, 5, 122, 201, 5, 5, 93, 229, 5, 5, 241, 174, 5, 5, 149, 231, 5, 5, 141, 174, 5, 5, 10, 247, 5, 5, 139, 164, 5, 5, 127, 220, 5, 5, 239, 247, 5, 5, 12, 247, 5, 5, 169, 179, 5, 5, 72, 193, 5, 5, 203, 220, 5, 5, 217, 213, 5, 5, 75, 249, 5, 5, 64, 190, 5, 5, 19, 204, 5, 5, 249, 223, 5, 5, 136, 222, 5, 5, 154, 246, 5, 5, 14, 201, 5, 5, 91, 217, 5, 5, 161, 219, 5, 5, 110, 204, 5, 5, 151, 220, 5, 5, 108, 166, 5, 5, 206, 185, 5, 5, 121, 173, 5, 5, 252, 172, 5, 5, 251, 188, 5, 5, 199, 191, 5, 5, 247, 235, 5, 5, 185, 195, 5, 5, 198, 200, 5, 5, 248, 239, 5, 5, 115, 202, 5, 5, 111, 222, 5, 5, 48, 239, 5, 5, 167, 248, 5, 5, 148, 205, 5, 5, 126, 201, 5, 5, 107, 185, 5, 5, 7, 167, 5, 5, 157, 187, 5, 5, 162, 187, 5, 5, 221, 189, 5, 5, 39, 189, 5, 5, 172, 223, 5, 5, 20, 241, 5, 5, 253, 193, 5, 5, 93, 217, 5, 5, 78, 175, 5, 5, 124, 230, 5, 5, 131, 222, 5, 5, 197, 237, 5, 5, 197, 190, 5, 5, 27, 231, 5, 5, 111, 201, 5, 5, 8, 193, 5, 5, 5, 195, 5, 5, 192, 189, 5, 5, 32, 220, 5, 5, 164, 218, 5, 5, 53, 240, 5, 5, 227, 198, 5, 5, 214, 220, 5, 5, 52, 187, 5, 5, 163, 241, 5, 5, 92, 211, 5, 5, 63, 227, 5, 5, 47, 232, 5, 5, 214, 218, 5, 5, 89, 204, 5, 5, 8, 220, 5, 5, 62, 212, 5, 5, 11, 190, 5, 5, 90, 204, 5, 5, 143, 219, 5, 5, 167, 174, 5, 5, 72, 189, 5, 5, 77, 184, 5, 5, 226, 163, 5, 5, 18, 200, 5, 5, 135, 194, 5, 5, 99, 247, 5, 5, 46, 189, 5, 5, 5, 226, 5, 5, 129, 231, 5, 5, 59, 234, 5, 5, 125, 174, 5, 5, 86, 202, 5, 5, 5, 247, 5, 5, 87, 229, 5, 5, 3, 167, 5, 5, 63, 249, 5, 5, 210, 213, 5, 5, 78, 184, 5, 5, 51, 190, 5, 5, 155, 219, 5, 5, 242, 188, 5, 5, 244, 215, 5, 5, 19, 193, 5, 5, 180, 233, 5, 5, 181, 233, 5, 5, 44, 168, 5, 5, 44, 233, 5, 5, 247, 165, 5, 5, 247, 184, 5, 5, 248, 184, 5, 5, 189, 233, 5, 5, 51, 217, 5, 5, 112, 164, 5, 5, 250, 165, 5, 5, 251, 165, 5, 5, 252, 165, 5, 5, 52, 217, 5, 5, 253, 165, 5, 5, 82, 164, 5, 5, 198, 233, 5, 5, 254, 165, 5, 5, 255, 165, 5, 5, 183, 224, 5, 5, 214, 230, 5, 5, 193, 184, 5, 5, 195, 184, 5, 5, 150, 224, 5, 5, 250, 225, 5, 5, 249, 225, 5, 5, 214, 240, 5, 5, 176, 230, 5, 5, 193, 165, 5, 5, 194, 165, 5, 5, 229, 226, 5, 5, 228, 230, 5, 5, 223, 187, 5, 5, 17, 175, 5, 5, 67, 224, 5, 5, 110, 192, 5, 5, 63, 210, 5, 5, 141, 247, 5, 5, 142, 247, 5, 5, 29, 209, 5, 5, 2, 218, 5, 5, 135, 239, 5, 5, 192, 193, 5, 5, 177, 169, 5, 5, 115, 208, 5, 5, 211, 231, 5, 5, 110, 247, 5, 5, 244, 181, 5, 5, 157, 222, 5, 5, 156, 244, 5, 5, 213, 180, 5, 5, 180, 249, 5, 5, 212, 205, 5, 5, 176, 198, 5, 5, 28, 241, 5, 5, 41, 173, 5, 5, 94, 245, 5, 5, 22, 175, 5, 5, 154, 237, 5, 5, 183, 197, 5, 5, 172, 219, 5, 5, 255, 222, 5, 5, 213, 205, 5, 5, 84, 249, 5, 5, 8, 246, 5, 5, 118, 188, 5, 5, 35, 223, 5, 5, 188, 223, 5, 5, 205, 183, 5, 5, 23, 181, 5, 5, 202, 218, 5, 5, 153, 239, 5, 5, 98, 208, 5, 5, 14, 203, 5, 5, 185, 224, 5, 5, 93, 175, 5, 5, 55, 239, 5, 5, 167, 176, 5, 5, 197, 166, 5, 5, 217, 221, 5, 5, 30, 177, 5, 5, 242, 231, 5, 5, 177, 179, 5, 5, 137, 229, 5, 5, 29, 182, 5, 5, 6, 212, 5, 5, 149, 206, 5, 5, 49, 200, 5, 5, 17, 179, 5, 5, 110, 242, 5, 5, 38, 171, 5, 5, 196, 164, 5, 5, 219, 201, 5, 5, 19, 183, 5, 5, 164, 239, 5, 5, 150, 203, 5, 5, 82, 222, 5, 5, 193, 187, 5, 5, 11, 218, 5, 5, 96, 182, 5, 5, 124, 237, 5, 5, 64, 227, 5, 5, 179, 179, 5, 5, 118, 167, 5, 5, 53, 198, 5, 5, 207, 183, 5, 5, 211, 178, 5, 5, 100, 208, 5, 5, 151, 217, 5, 5, 12, 218, 5, 5, 104, 166, 5, 5, 15, 242, 5, 5, 140, 214, 5, 5, 21, 210, 5, 5, 224, 179, 5, 5, 175, 221, 5, 5, 22, 210, 5, 5, 241, 176, 5, 5, 242, 176, 5, 5, 209, 181, 5, 5, 162, 206, 5, 5, 224, 180, 5, 5, 207, 175, 5, 5, 229, 239, 5, 5, 176, 219, 5, 5, 158, 182, 5, 5, 181, 194, 5, 5, 105, 245, 5, 5, 68, 227, 5, 5, 142, 225, 5, 5, 154, 236, 5, 5, 83, 232, 5, 5, 119, 212, 5, 5, 239, 173, 5, 5, 123, 188, 5, 5, 241, 236, 5, 5, 51, 200, 5, 5, 4, 187, 5, 5, 100, 193, 5, 5, 54, 185, 5, 5, 136, 164, 5, 5, 203, 234, 5, 5, 160, 207, 5, 5, 197, 164, 5, 5, 229, 200, 5, 5, 75, 175, 5, 5, 25, 190, 5, 5, 51, 242, 5, 5, 19, 220, 5, 5, 225, 205, 5, 5, 96, 248, 5, 5, 244, 225, 5, 5, 121, 196, 5, 5, 235, 245, 5, 5, 228, 180, 5, 5, 100, 240, 5, 5, 33, 175, 5, 5, 86, 249, 5, 5, 95, 190, 5, 5, 211, 164, 5, 5, 166, 184, 5, 5, 218, 247, 5, 5, 105, 176, 5, 5, 11, 237, 5, 5, 215, 178, 5, 5, 33, 237, 5, 5, 240, 242, 5, 5, 60, 226, 5, 5, 46, 223, 5, 5, 176, 180, 5, 5, 110, 248, 5, 5, 251, 226, 5, 5, 185, 220, 5, 5, 56, 242, 5, 5, 190, 249, 5, 5, 242, 237, 5, 5, 245, 165, 5, 5, 193, 193, 5, 5, 243, 237, 5, 5, 208, 213, 5, 5, 55, 201, 5, 5, 35, 210, 5, 5, 162, 175, 5, 5, 34, 175, 5, 5, 29, 246, 5, 5, 122, 183, 5, 5, 181, 239, 5, 5, 98, 192, 5, 5, 186, 241, 5, 5, 66, 224, 5, 5, 52, 188, 5, 5, 19, 218, 5, 5, 68, 179, 5, 5, 6, 177, 5, 5, 49, 223, 5, 5, 87, 249, 5, 5, 176, 236, 5, 5, 69, 179, 5, 5, 159, 210, 5, 5, 65, 169, 5, 5, 187, 241, 5, 5, 112, 194, 5, 5, 27, 221, 5, 5, 117, 238, 5, 5, 93, 227, 5, 5, 228, 233, 5, 5, 114, 223, 5, 5, 185, 236, 5, 5, 62, 232, 5, 5, 125, 219, 5, 5, 218, 178, 5, 5, 11, 177, 5, 5, 244, 185, 5, 5, 54, 223, 5, 5, 176, 234, 5, 5, 243, 188, 5, 5, 72, 233, 5, 5, 30, 230, 5, 5, 96, 210, 5, 5, 153, 217, 5, 5, 141, 248, 5, 5, 193, 239, 5, 5, 124, 245, 5, 5, 244, 187, 5, 5, 98, 207, 5, 5, 105, 227, 5, 5, 89, 165, 5, 5, 154, 241, 5, 5, 186, 169, 5, 5, 103, 214, 5, 5, 46, 237, 5, 5, 187, 197, 5, 5, 95, 220, 5, 5, 140, 180, 5, 5, 65, 239, 5, 5, 164, 171, 5, 5, 206, 165, 5, 5, 210, 165, 5, 5, 232, 184, 5, 5, 227, 186, 5, 5, 79, 249, 5, 5, 206, 241, 5, 5, 114, 190, 5, 5, 19, 187, 5, 5, 22, 240, 5, 5, 143, 181, 5, 5, 135, 176, 5, 5, 253, 208, 5, 5, 213, 249, 5, 5, 39, 228, 5, 5, 73, 198, 5, 5, 158, 247, 5, 5, 185, 170, 5, 5, 182, 230, 5, 5, 131, 212, 5, 5, 250, 246, 5, 5, 22, 197, 5, 5, 21, 174, 5, 5, 235, 195, 5, 5, 6, 166, 5, 5, 149, 173, 5, 5, 141, 216, 5, 5, 236, 182, 5, 5, 245, 215, 5, 5, 182, 228, 5, 5, 140, 164, 5, 5, 32, 249, 5, 5, 24, 235, 5, 5, 180, 199, 5, 5, 253, 213, 5, 5, 146, 164, 5, 5, 249, 244, 5, 5, 51, 207, 5, 5, 35, 242, 5, 5, 78, 198, 5, 5, 110, 179, 5, 5, 219, 209, 5, 5, 6, 165, 5, 5, 205, 167, 5, 5, 61, 166, 5, 5, 3, 228, 5, 5, 205, 163, 5, 5, 156, 245, 5, 5, 207, 176, 5, 5, 96, 232, 5, 5, 85, 241, 5, 5, 219, 188, 5, 5, 2, 197, 5, 5, 213, 193, 5, 5, 135, 221, 5, 5, 194, 226, 5, 5, 8, 196, 5, 5, 62, 248, 5, 5, 217, 191, 5, 5, 100, 197, 5, 5, 99, 209, 5, 5, 203, 218, 5, 5, 59, 202, 5, 5, 171, 200, 5, 5, 51, 198, 5, 5, 146, 216, 5, 5, 243, 203, 5, 5, 46, 213, 5, 5, 229, 193, 5, 5, 154, 184, 5, 5, 104, 186, 5, 5, 43, 203, 5, 5, 84, 196, 5, 5, 116, 221, 5, 5, 148, 186, 5, 5, 182, 249, 5, 5, 174, 242, 5, 5, 97, 225, 5, 5, 200, 175, 5, 5, 196, 212, 5, 5, 48, 191, 5, 5, 163, 174, 5, 5, 23, 245, 5, 5, 100, 174, 5, 5, 44, 192, 5, 5, 240, 197, 5, 5, 107, 186, 5, 5, 47, 203, 5, 5, 236, 192, 5, 5, 48, 208, 5, 5, 207, 244, 5, 5, 82, 163, 5, 5, 123, 213, 5, 5, 214, 163, 5, 5, 67, 187, 5, 5, 142, 249, 5, 5, 140, 166, 5, 5, 195, 214, 5, 5, 127, 208, 5, 5, 255, 167, 5, 5, 101, 211, 5, 5, 144, 193, 5, 5, 14, 220, 5, 5, 248, 179, 5, 5, 99, 232, 5, 5, 180, 168, 5, 5, 49, 185, 5, 5, 162, 183, 5, 5, 169, 224, 5, 5, 52, 242, 5, 5, 144, 166, 5, 5, 44, 216, 5, 5, 237, 175, 5, 5, 249, 206, 5, 5, 228, 249, 5, 5, 131, 209, 5, 5, 106, 239, 5, 5, 184, 175, 5, 5, 237, 242, 5, 5, 212, 169, 5, 5, 239, 230, 5, 5, 227, 163, 5, 5, 232, 237, 5, 5, 83, 229, 5, 5, 255, 207, 5, 5, 24, 248, 5, 5, 187, 249, 5, 5, 31, 198, 5, 5, 122, 196, 5, 5, 48, 216, 5, 5, 49, 216, 5, 5, 49, 236, 5, 5, 79, 180, 5, 5, 39, 172, 5, 5, 232, 212, 5, 5, 215, 230, 5, 5, 224, 225, 5, 5, 137, 231, 5, 5, 104, 232, 5, 5, 102, 241, 5, 5, 4, 245, 5, 5, 52, 216, 5, 5, 148, 249, 5, 5, 7, 198, 5, 5, 222, 249, 5, 5, 242, 164, 5, 5, 78, 170, 5, 5, 243, 223, 5, 5, 47, 174, 5, 5, 11, 188, 5, 5, 188, 243, 5, 5, 120, 222, 5, 5, 83, 199, 5, 5, 138, 209, 5, 5, 146, 242, 5, 5, 21, 221, 5, 5, 145, 170, 5, 5, 139, 209, 5, 5, 238, 216, 5, 5, 139, 176, 5, 5, 94, 165, 5, 5, 242, 217, 5, 5, 72, 165, 5, 5, 198, 232, 5, 5, 27, 246, 5, 5, 125, 221, 5, 5, 80, 170, 5, 5, 105, 228, 5, 5, 236, 188, 5, 5, 182, 166, 5, 5, 248, 175, 5, 5, 237, 170, 5, 5, 118, 241, 5, 5, 52, 234, 5, 5, 95, 196, 5, 5, 219, 173, 5, 5, 242, 170, 5, 5, 109, 232, 5, 5, 38, 197, 5, 5, 181, 188, 5, 5, 153, 232, 5, 5, 104, 224, 5, 5, 22, 242, 5, 5, 19, 202, 5, 5, 78, 243, 5, 5, 111, 224, 5, 5, 189, 234, 5, 5, 155, 189, 5, 5, 162, 200, 5, 5, 38, 213, 5, 5, 228, 173, 5, 5, 161, 232, 5, 5, 158, 187, 5, 5, 71, 242, 5, 5, 197, 177, 5, 5, 6, 210, 5, 5, 32, 222, 5, 5, 254, 166, 5, 5, 55, 220, 5, 5, 226, 239, 5, 5, 119, 192, 5, 5, 185, 247, 5, 5, 151, 204, 5, 5, 226, 242, 5, 5, 38, 233, 5, 5, 29, 170, 5, 5, 248, 220, 5, 5, 13, 217, 5, 5, 208, 175, 5, 5, 176, 220, 5, 5, 83, 173, 5, 5, 109, 249, 5, 5, 252, 174, 5, 5, 183, 249, 5, 5, 142, 179, 5, 5, 65, 181, 5, 5, 77, 209, 5, 5, 37, 182, 5, 5, 154, 234, 5, 5, 65, 198, 5, 5, 20, 193, 5, 5, 131, 236, 5, 5, 118, 172, 5, 5, 32, 224, 5, 5, 50, 232, 5, 5, 53, 172, 5, 5, 210, 185, 5, 5, 7, 226, 5, 5, 33, 224, 5, 5, 116, 242, 5, 5, 186, 182, 5, 5, 175, 218, 5, 5, 57, 215, 5, 5, 61, 206, 5, 5, 126, 225, 5, 5, 243, 243, 5, 5, 228, 181, 5, 5, 101, 204, 5, 5, 221, 169, 5, 5, 35, 196, 5, 5, 59, 237, 5, 5, 149, 185, 5, 5, 231, 168, 5, 5, 46, 233, 5, 5, 191, 197, 5, 5, 57, 243, 5, 5, 217, 227, 5, 5, 105, 184, 5, 5, 36, 246, 5, 5, 52, 190, 5, 5, 71, 186, 5, 5, 147, 242, 5, 5, 192, 179, 5, 5, 233, 166, 5, 5, 124, 220, 5, 5, 110, 226, 5, 5, 5, 180, 5, 5, 47, 233, 5, 5, 104, 222, 5, 5, 162, 234, 5, 5, 218, 196, 5, 5, 188, 180, 5, 5, 123, 215, 5, 5, 237, 177, 5, 5, 28, 199, 5, 5, 239, 216, 5, 5, 160, 216, 5, 5, 34, 180, 5, 5, 41, 210, 5, 5, 111, 226, 5, 5, 199, 249, 5, 5, 205, 166, 5, 5, 29, 195, 5, 5, 59, 236, 5, 5, 48, 171, 5, 5, 86, 212, 5, 5, 60, 224, 5, 5, 178, 234, 5, 5, 201, 249, 5, 5, 26, 250, 5, 5, 21, 239, 5, 5, 121, 219, 5, 5, 192, 197, 5, 5, 101, 180, 5, 5, 70, 206, 5, 5, 89, 168, 5, 5, 220, 224, 5, 5, 158, 248, 5, 5, 220, 186, 5, 5, 65, 232, 5, 5, 188, 185, 5, 5, 121, 239, 5, 5, 72, 206, 5, 5, 109, 166, 5, 5, 111, 204, 5, 5, 30, 213, 5, 5, 19, 180, 5, 5, 119, 172, 5, 5, 205, 184, 5, 5, 195, 240, 5, 5, 17, 187, 5, 5, 103, 207, 5, 5, 84, 243, 5, 5, 196, 240, 5, 5, 106, 184, 5, 5, 208, 230, 5, 5, 208, 217, 5, 5, 49, 224, 5, 5, 91, 243, 5, 5, 159, 189, 5, 5, 64, 171, 5, 5, 65, 171, 5, 5, 107, 184, 5, 5, 35, 216, 5, 5, 115, 204, 5, 5, 221, 227, 5, 5, 67, 171, 5, 5, 102, 243, 5, 5, 234, 165, 5, 5, 16, 234, 5, 5, 69, 227, 5, 5, 115, 242, 5, 5, 77, 227, 5, 5, 107, 227, 5, 5, 250, 186, 5, 5, 244, 222, 5, 5, 74, 241, 5, 5, 76, 189, 5, 5, 221, 186, 5, 5, 164, 205, 5, 5, 165, 205, 5, 5, 7, 210, 5, 5, 223, 239, 5, 5, 10, 188, 5, 5, 132, 175, 5, 5, 212, 219, 5, 5, 198, 177, 5, 5, 201, 166, 5, 5, 14, 210, 5, 5, 0, 148, 5, 5, 164, 174, 5, 5, 172, 220, 5, 5, 107, 218, 5, 5, 0, 149, 5, 5, 187, 190, 5, 5, 165, 174, 5, 5, 20, 183, 5, 5, 103, 177, 5, 5, 94, 228, 5, 5, 187, 209, 5, 5, 152, 195, 5, 5, 74, 226, 5, 5, 182, 231, 5, 5, 43, 173, 5, 5, 112, 199, 5, 5, 113, 199, 5, 5, 131, 198, 5, 5, 195, 192, 5, 5, 192, 208, 5, 5, 203, 170, 5, 5, 11, 230, 5, 5, 243, 176, 5, 5, 181, 235, 5, 5, 125, 213, 5, 5, 84, 240, 5, 5, 208, 223, 5, 5, 197, 226, 5, 5, 241, 181, 5, 5, 212, 212, 5, 5, 215, 184, 5, 5, 79, 210, 5, 5, 48, 210, 5, 5, 97, 205, 5, 5, 40, 168, 5, 5, 65, 241, 5, 5, 177, 205, 5, 5, 13, 163, 5, 5, 156, 173, 5, 5, 198, 224, 5, 5, 153, 216, 5, 5, 196, 175, 5, 5, 222, 248, 5, 5, 83, 210, 5, 5, 42, 174, 5, 5, 217, 167, 5, 5, 249, 182, 5, 5, 2, 216, 5, 5, 166, 237, 5, 5, 161, 207, 5, 5, 213, 178, 5, 5, 59, 192, 5, 5, 212, 243, 5, 5, 129, 210, 5, 5, 93, 164, 5, 5, 211, 235, 5, 5, 92, 230, 5, 5, 233, 223, 5, 5, 56, 184, 5, 5, 85, 173, 5, 5, 169, 249, 5, 5, 136, 202, 5, 5, 231, 210, 5, 5, 218, 167, 5, 5, 146, 193, 5, 5, 160, 201, 5, 5, 183, 243, 5, 5, 28, 208, 5, 5, 200, 226, 5, 5, 101, 240, 5, 5, 212, 222, 5, 5, 81, 213, 5, 5, 241, 191, 5, 5, 12, 230, 5, 5, 182, 199, 5, 5, 143, 179, 5, 5, 238, 195, 5, 5, 110, 249, 5, 5, 83, 190, 5, 5, 233, 237, 5, 5, 70, 247, 5, 5, 70, 178, 5, 5, 142, 231, 5, 5, 190, 183, 5, 5, 221, 181, 5, 5, 86, 227, 5, 5, 242, 171, 5, 5, 21, 193, 5, 5, 18, 195, 5, 5, 129, 189, 5, 5, 76, 166, 5, 5, 187, 235, 5, 5, 55, 193, 5, 5, 100, 163, 5, 5, 219, 167, 5, 5, 111, 248, 5, 5, 188, 235, 5, 5, 249, 232, 5, 5, 105, 166, 5, 5, 57, 242, 5, 5, 88, 181, 5, 5, 81, 203, 5, 5, 186, 220, 5, 5, 20, 227, 5, 5, 87, 227, 5, 5, 123, 218, 5, 5, 77, 207, 5, 5, 72, 224, 5, 5, 206, 226, 5, 5, 207, 226, 5, 5, 173, 217, 5, 5, 214, 219, 5, 5, 51, 164, 5, 5, 129, 197, 5, 5, 49, 165, 5, 5, 28, 229, 5, 5, 165, 225, 5, 5, 252, 226, 5, 5, 101, 163, 5, 5, 242, 221, 5, 5, 105, 223, 5, 5, 215, 231, 5, 5, 175, 166, 5, 5, 187, 182, 5, 5, 131, 210, 5, 5, 126, 174, 5, 5, 197, 219, 5, 5, 90, 188, 5, 5, 88, 210, 5, 5, 127, 174, 5, 5, 92, 187, 5, 5, 224, 248, 5, 5, 220, 167, 5, 5, 10, 244, 5, 5, 110, 164, 5, 5, 55, 163, 5, 5, 174, 203, 5, 5, 61, 192, 5, 5, 152, 229, 5, 5, 92, 221, 5, 5, 78, 172, 5, 5, 84, 213, 5, 5, 85, 213, 5, 5, 254, 176, 5, 5, 30, 203, 5, 5, 98, 172, 5, 5, 30, 180, 5, 5, 182, 200, 5, 5, 234, 239, 5, 5, 121, 235, 5, 5, 35, 197, 5, 5, 247, 194, 5, 5, 197, 163, 5, 5, 110, 205, 5, 5, 190, 182, 5, 5, 155, 216, 5, 5, 169, 201, 5, 5, 55, 204, 5, 5, 76, 223, 5, 5, 79, 239, 5, 5, 148, 190, 5, 5, 77, 166, 5, 5, 91, 227, 5, 5, 94, 170, 5, 5, 32, 203, 5, 5, 189, 243, 5, 5, 33, 208, 5, 5, 22, 220, 5, 5, 181, 226, 5, 5, 83, 216, 5, 5, 191, 182, 5, 5, 130, 169, 5, 5, 150, 185, 5, 5, 243, 201, 5, 5, 127, 248, 5, 5, 32, 217, 5, 5, 94, 230, 5, 5, 85, 233, 5, 5, 13, 219, 5, 5, 151, 212, 5, 5, 97, 228, 5, 5, 109, 223, 5, 5, 157, 215, 5, 5, 156, 216, 5, 5, 82, 212, 5, 5, 221, 167, 5, 5, 255, 180, 5, 5, 151, 221, 5, 5, 108, 216, 5, 5, 43, 172, 5, 5, 45, 196, 5, 5, 188, 241, 5, 5, 60, 215, 5, 5, 108, 215, 5, 5, 251, 214, 5, 5, 34, 208, 5, 5, 136, 223, 5, 5, 203, 185, 5, 5, 91, 202, 5, 5, 28, 168, 5, 5, 192, 230, 5, 5, 80, 171, 5, 5, 111, 239, 5, 5, 145, 243, 5, 5, 169, 194, 5, 5, 236, 238, 5, 5, 122, 177, 5, 5, 86, 242, 5, 5, 234, 235, 5, 5, 129, 195, 5, 5, 135, 210, 5, 5, 165, 190, 5, 5, 196, 235, 5, 5, 30, 238, 5, 5, 107, 177, 5, 5, 146, 198, 5, 5, 16, 231, 5, 5, 95, 227, 5, 5, 253, 195, 5, 5, 40, 210, 5, 5, 163, 214, 5, 5, 154, 219, 5, 5, 25, 215, 5, 5, 191, 232, 5, 5, 183, 180, 5, 5, 208, 234, 5, 5, 116, 193, 5, 5, 190, 201, 5, 5, 216, 222, 5, 5, 208, 179, 5, 5, 78, 191, 5, 5, 224, 169, 5, 5, 126, 186, 5, 5, 227, 246, 5, 5, 7, 216, 5, 5, 153, 249, 5, 5, 239, 163, 5, 5, 122, 245, 5, 5, 186, 236, 5, 5, 133, 248, 5, 5, 98, 224, 5, 5, 226, 177, 5, 5, 115, 163, 5, 5, 147, 198, 5, 5, 61, 203, 5, 5, 172, 201, 5, 5, 6, 235, 5, 5, 218, 173, 5, 5, 74, 224, 5, 5, 228, 246, 5, 5, 25, 219, 5, 5, 140, 235, 5, 5, 118, 217, 5, 5, 123, 199, 5, 5, 35, 248, 5, 5, 35, 181, 5, 5, 227, 194, 5, 5, 57, 222, 5, 5, 114, 191, 5, 5, 157, 216, 5, 5, 9, 211, 5, 5, 223, 228, 5, 5, 100, 198, 5, 5, 115, 191, 5, 5, 67, 241, 5, 5, 174, 225, 5, 5, 208, 232, 5, 5, 226, 206, 5, 5, 222, 215, 5, 5, 2, 181, 5, 5, 210, 170, 5, 5, 9, 169, 5, 5, 206, 186, 5, 5, 92, 167, 5, 5, 17, 231, 5, 5, 127, 225, 5, 5, 52, 192, 5, 5, 10, 187, 5, 5, 99, 179, 5, 5, 100, 166, 5, 5, 247, 221, 5, 5, 8, 221, 5, 5, 193, 181, 5, 5, 95, 188, 5, 5, 49, 233, 5, 5, 95, 207, 5, 5, 36, 225, 5, 5, 88, 249, 5, 5, 131, 205, 5, 5, 201, 192, 5, 5, 114, 177, 5, 5, 194, 179, 5, 5, 20, 225, 5, 5, 215, 190, 5, 5, 189, 193, 5, 5, 140, 189, 5, 5, 192, 163, 5, 5, 56, 190, 5, 5, 212, 186, 5, 5, 49, 211, 5, 5, 45, 167, 5, 5, 238, 207, 5, 5, 246, 171, 5, 5, 50, 233, 5, 5, 173, 167, 5, 5, 250, 238, 5, 5, 227, 167, 5, 5, 227, 215, 5, 5, 235, 240, 5, 5, 85, 222, 5, 5, 153, 185, 5, 5, 27, 195, 5, 5, 60, 197, 5, 5, 76, 179, 5, 5, 219, 196, 5, 5, 253, 177, 5, 5, 44, 183, 5, 5, 41, 226, 5, 5, 150, 231, 5, 5, 99, 221, 5, 5, 121, 164, 5, 5, 255, 183, 5, 5, 150, 233, 5, 5, 112, 226, 5, 5, 9, 180, 5, 5, 227, 177, 5, 5, 223, 208, 5, 5, 141, 210, 5, 5, 106, 222, 5, 5, 4, 202, 5, 5, 142, 248, 5, 5, 204, 205, 5, 5, 92, 194, 5, 5, 41, 170, 5, 5, 249, 224, 5, 5, 115, 181, 5, 5, 15, 213, 5, 5, 204, 188, 5, 5, 195, 179, 5, 5, 95, 193, 5, 5, 209, 234, 5, 5, 161, 229, 5, 5, 202, 220, 5, 5, 57, 190, 5, 5, 14, 171, 5, 5, 58, 190, 5, 5, 106, 199, 5, 5, 186, 165, 5, 5, 125, 191, 5, 5, 59, 185, 5, 5, 203, 195, 5, 5, 184, 173, 5, 5, 134, 245, 5, 5, 32, 183, 5, 5, 243, 246, 5, 5, 150, 232, 5, 5, 99, 199, 5, 5, 65, 175, 5, 5, 66, 215, 5, 5, 105, 168, 5, 5, 169, 199, 5, 5, 80, 189, 5, 5, 147, 172, 5, 5, 168, 170, 5, 5, 102, 195, 5, 5, 124, 244, 5, 5, 161, 176, 5, 5, 206, 188, 5, 5, 156, 193, 5, 5, 163, 215, 5, 5, 45, 178, 5, 5, 231, 181, 5, 5, 38, 246, 5, 5, 114, 233, 5, 5, 244, 174, 5, 5, 46, 230, 5, 5, 110, 202, 5, 5, 5, 238, 5, 5, 37, 208, 5, 5, 218, 224, 5, 5, 191, 180, 5, 5, 91, 219, 5, 5, 89, 199, 5, 5, 50, 244, 5, 5, 178, 178, 5, 5, 218, 213, 5, 5, 240, 247, 5, 5, 255, 195, 5, 5, 134, 216, 5, 5, 71, 201, 5, 5, 103, 246, 5, 5, 245, 174, 5, 5, 217, 219, 5, 5, 140, 223, 5, 5, 119, 209, 5, 5, 219, 226, 5, 5, 124, 201, 5, 5, 136, 196, 5, 5, 117, 181, 5, 5, 98, 170, 5, 5, 151, 209, 5, 5, 152, 209, 5, 5, 41, 198, 5, 5, 209, 179, 5, 5, 69, 235, 5, 5, 193, 200, 5, 5, 125, 241, 5, 5, 248, 171, 5, 5, 137, 205, 5, 5, 161, 178, 5, 5, 138, 205, 5, 5, 107, 237, 5, 5, 137, 196, 5, 5, 31, 199, 5, 5, 124, 217, 5, 5, 93, 178, 5, 5, 189, 203, 5, 5, 204, 220, 5, 5, 207, 165, 5, 5, 219, 222, 5, 5, 154, 231, 5, 5, 255, 214, 5, 5, 69, 241, 5, 5, 248, 169, 5, 5, 225, 166, 5, 5, 33, 238, 5, 5, 233, 203, 5, 5, 196, 210, 5, 5, 228, 169, 5, 5, 236, 236, 5, 5, 39, 175, 5, 5, 174, 164, 5, 5, 220, 213, 5, 5, 159, 188, 5, 5, 237, 236, 5, 5, 114, 195, 5, 5, 72, 188, 5, 5, 201, 241, 5, 5, 130, 241, 5, 5, 91, 183, 5, 5, 120, 247, 5, 5, 89, 231, 5, 5, 104, 180, 5, 5, 238, 236, 5, 5, 54, 211, 5, 5, 120, 187, 5, 5, 193, 229, 5, 5, 43, 226, 5, 5, 129, 163, 5, 5, 75, 167, 5, 5, 16, 175, 5, 5, 215, 164, 5, 5, 121, 236, 5, 5, 202, 192, 5, 5, 155, 231, 5, 5, 10, 245, 5, 5, 193, 244, 5, 5, 104, 201, 5, 5, 39, 181, 5, 5, 73, 188, 5, 5, 103, 195, 5, 5, 86, 237, 5, 5, 147, 227, 5, 5, 60, 236, 5, 5, 247, 246, 5, 5, 17, 198, 5, 5, 5, 221, 5, 5, 6, 221, 5, 5, 127, 186, 5, 5, 180, 240, 5, 5, 77, 249, 5, 5, 118, 180, 5, 5, 6, 184, 5, 5, 66, 213, 5, 5, 37, 203, 5, 5, 158, 191, 5, 5, 39, 169, 5, 5, 226, 207, 5, 5, 131, 237, 5, 5, 240, 232, 5, 5, 239, 236, 5, 5, 237, 213, 5, 5, 14, 174, 5, 5, 71, 247, 5, 5, 228, 182, 5, 5, 251, 189, 5, 5, 240, 236, 5, 5, 179, 238, 5, 5, 228, 232, 5, 5, 119, 182, 5, 5, 3, 215, 5, 5, 117, 194, 5, 5, 181, 217, 5, 5, 207, 188, 5, 5, 34, 228, 5, 5, 159, 238, 5, 5, 134, 170, 5, 5, 149, 169, 5, 5, 104, 190, 5, 5, 98, 165, 5, 5, 157, 246, 5, 5, 76, 188, 5, 5, 194, 176, 5, 5, 97, 236, 5, 5, 139, 233, 5, 5, 204, 239, 5, 5, 70, 190, 5, 5, 126, 238, 5, 5, 2, 196, 5, 5, 55, 211, 5, 5, 224, 234, 5, 5, 152, 208, 5, 5, 24, 225, 5, 5, 63, 180, 5, 5, 154, 170, 5, 5, 233, 249, 5, 5, 151, 224, 5, 5, 223, 173, 5, 5, 185, 206, 5, 5, 172, 165, 5, 5, 163, 229, 5, 5, 15, 186, 5, 5, 29, 215, 5, 5, 71, 235, 5, 5, 2, 191, 5, 5, 156, 232, 5, 5, 144, 205, 5, 5, 45, 234, 5, 5, 228, 186, 5, 5, 238, 213, 5, 5, 31, 213, 5, 5, 224, 195, 5, 5, 5, 230, 5, 5, 204, 164, 5, 5, 60, 221, 5, 5, 197, 210, 5, 5, 82, 225, 5, 5, 75, 190, 5, 5, 127, 199, 5, 5, 116, 202, 5, 5, 243, 168, 5, 5, 142, 200, 5, 5, 81, 211, 5, 5, 178, 166, 5, 5, 254, 188, 5, 5, 248, 235, 5, 5, 30, 215, 5, 5, 98, 236, 5, 5, 83, 225, 5, 5, 179, 196, 5, 5, 149, 235, 5, 5, 110, 188, 5, 5, 62, 183, 5, 5, 79, 201, 5, 5, 66, 168, 5, 5, 54, 237, 5, 5, 34, 173, 5, 5, 182, 209, 5, 5, 50, 212, 5, 5, 53, 244, 5, 5, 64, 180, 5, 5, 115, 216, 5, 5, 198, 207, 5, 5, 20, 204, 5, 5, 75, 233, 5, 5, 2, 188, 5, 5, 122, 209, 5, 5, 189, 185, 5, 5, 154, 178, 5, 5, 16, 216, 5, 5, 126, 232, 5, 5, 23, 202, 5, 5, 68, 205, 5, 5, 222, 222, 5, 5, 145, 176, 5, 5, 239, 240, 5, 5, 230, 244, 5, 5, 127, 178, 5, 5, 199, 207, 5, 5, 92, 183, 5, 5, 127, 232, 5, 5, 5, 169, 5, 5, 4, 181, 5, 5, 35, 173, 5, 5, 61, 221, 5, 5, 209, 177, 5, 5, 122, 228, 5, 5, 82, 195, 5, 5, 2, 205, 5, 5, 224, 172, 5, 5, 55, 237, 5, 5, 253, 189, 5, 5, 228, 207, 5, 5, 145, 200, 5, 5, 168, 248, 5, 5, 166, 188, 5, 5, 72, 222, 5, 5, 156, 209, 5, 5, 149, 205, 5, 5, 73, 222, 5, 5, 105, 185, 5, 5, 106, 165, 5, 5, 5, 181, 5, 5, 199, 200, 5, 5, 134, 241, 5, 5, 75, 178, 5, 5, 191, 193, 5, 5, 81, 176, 5, 5, 105, 207, 5, 5, 122, 173, 5, 5, 42, 175, 5, 5, 244, 163, 5, 5, 201, 240, 5, 5, 89, 238, 5, 5, 124, 233, 5, 5, 26, 218, 5, 5, 116, 177, 5, 5, 252, 221, 5, 5, 210, 198, 5, 5, 112, 222, 5, 5, 31, 202, 5, 5, 19, 184, 5, 5, 150, 197, 5, 5, 229, 199, 5, 5, 107, 211, 5, 5, 34, 166, 5, 5, 148, 244, 5, 5, 181, 196, 5, 5, 131, 184, 5, 5, 218, 216, 5, 5, 150, 205, 5, 5, 151, 205, 5, 5, 199, 210, 5, 5, 161, 223, 5, 5, 53, 219, 5, 5, 143, 171, 5, 5, 241, 186, 5, 5, 110, 185, 5, 5, 34, 202, 5, 5, 66, 233, 5, 5, 47, 197, 5, 5, 250, 228, 5, 5, 186, 167, 5, 5, 5, 185, 5, 5, 16, 178, 5, 5, 195, 229, 5, 5, 18, 226, 5, 5, 106, 191, 5, 5, 179, 165, 5, 5, 63, 164, 5, 5, 254, 189, 5, 5, 240, 215, 5, 5, 195, 196, 5, 5, 22, 191, 5, 5, 15, 238, 5, 5, 222, 189, 5, 5, 11, 216, 5, 5, 152, 205, 5, 5, 99, 186, 5, 5, 139, 163, 5, 5, 10, 205, 5, 5, 6, 242, 5, 5, 119, 224, 5, 5, 64, 216, 5, 5, 153, 205, 5, 5, 44, 213, 5, 5, 125, 194, 5, 5, 127, 234, 5, 5, 24, 184, 5, 5, 140, 185, 5, 5, 11, 235, 5, 5, 235, 211, 5, 5, 94, 167, 5, 5, 118, 190, 5, 5, 58, 233, 5, 5, 181, 190, 5, 5, 67, 232, 5, 5, 26, 184, 5, 5, 213, 207, 5, 5, 124, 180, 5, 5, 125, 229, 5, 5, 61, 186, 5, 5, 162, 205, 5, 5, 122, 191, 5, 5, 212, 239, 5, 5, 158, 220, 5, 5, 194, 177, 5, 5, 206, 219, 5, 5, 168, 166, 5, 5, 9, 188, 5, 5, 240, 181, 5, 5, 68, 210, 5, 5, 73, 226, 5, 5, 46, 228, 5, 5, 54, 202, 5, 5, 11, 173, 5, 5, 98, 177, 5, 5, 162, 170, 5, 5, 213, 184, 5, 5, 26, 244, 5, 5, 207, 223, 5, 5, 147, 216, 5, 5, 237, 191, 5, 5, 131, 166, 5, 5, 138, 179, 5, 5, 56, 207, 5, 5, 210, 178, 5, 5, 228, 223, 5, 5, 206, 222, 5, 5, 238, 182, 5, 5, 208, 163, 5, 5, 190, 192, 5, 5, 90, 164, 5, 5, 120, 210, 5, 5, 236, 241, 5, 5, 202, 210, 5, 5, 21, 208, 5, 5, 253, 244, 5, 5, 56, 192, 5, 5, 237, 195, 5, 5, 153, 173, 5, 5, 181, 199, 5, 5, 217, 179, 5, 5, 53, 184, 5, 5, 12, 163, 5, 5, 69, 248, 5, 5, 123, 210, 5, 5, 76, 221, 5, 5, 124, 210, 5, 5, 43, 193, 5, 5, 47, 163, 5, 5, 146, 237, 5, 5, 188, 166, 5, 5, 137, 166, 5, 5, 83, 163, 5, 5, 214, 248, 5, 5, 55, 233, 5, 5, 91, 172, 5, 5, 25, 180, 5, 5, 74, 203, 5, 5, 167, 177, 5, 5, 43, 242, 5, 5, 252, 180, 5, 5, 84, 163, 5, 5, 106, 189, 5, 5, 182, 229, 5, 5, 75, 210, 5, 5, 52, 170, 5, 5, 46, 164, 5, 5, 65, 227, 5, 5, 164, 217, 5, 5, 35, 179, 5, 5, 18, 165, 5, 5, 155, 183, 5, 5, 212, 167, 5, 5, 62, 178, 5, 5, 152, 216, 5, 5, 127, 195, 5, 5, 201, 185, 5, 5, 60, 239, 5, 5, 204, 170, 5, 5, 37, 241, 5, 5, 82, 242, 5, 5, 110, 225, 5, 5, 50, 203, 5, 5, 63, 241, 5, 5, 51, 183, 5, 5, 235, 218, 5, 5, 70, 176, 5, 5, 179, 182, 5, 5, 70, 191, 5, 5, 104, 177, 5, 5, 129, 243, 5, 5, 88, 195, 5, 5, 85, 219, 5, 5, 143, 221, 5, 5, 79, 213, 5, 5, 218, 177, 5, 5, 121, 177, 5, 5, 53, 215, 5, 5, 114, 199, 5, 5, 69, 212, 5, 5, 135, 227, 5, 5, 72, 197, 5, 5, 105, 177, 5, 5, 207, 191, 5, 5, 35, 172, 5, 5, 85, 221, 5, 5, 227, 235, 5, 5, 41, 204, 5, 5, 101, 239, 5, 5, 132, 224, 5, 5, 179, 226, 5, 5, 51, 180, 5, 5, 232, 238, 5, 5, 182, 243, 5, 5, 187, 230, 5, 5, 240, 221, 5, 5, 177, 185, 5, 5, 6, 169, 5, 5, 30, 245, 5, 5, 55, 185, 5, 5, 97, 179, 5, 5, 195, 188, 5, 5, 249, 195, 5, 5, 147, 214, 5, 5, 227, 236, 5, 5, 5, 187, 5, 5, 207, 232, 5, 5, 159, 225, 5, 5, 147, 193, 5, 5, 47, 192, 5, 5, 167, 237, 5, 5, 19, 215, 5, 5, 204, 234, 5, 5, 109, 191, 5, 5, 209, 173, 5, 5, 3, 216, 5, 5, 89, 188, 5, 5, 79, 195, 5, 5, 35, 205, 5, 5, 12, 231, 5, 5, 155, 180, 5, 5, 117, 186, 5, 5, 6, 211, 5, 5, 219, 206, 5, 5, 135, 237, 5, 5, 39, 211, 5, 5, 167, 219, 5, 5, 167, 184, 5, 5, 205, 234, 5, 5, 33, 170, 5, 5, 120, 164, 5, 5, 207, 208, 5, 5, 62, 175, 5, 5, 67, 195, 5, 5, 210, 196, 5, 5, 247, 224, 5, 5, 233, 212, 5, 5, 93, 193, 5, 5, 97, 190, 5, 5, 38, 226, 5, 5, 84, 222, 5, 5, 193, 205, 5, 5, 2, 219, 5, 5, 58, 215, 5, 5, 236, 207, 5, 5, 167, 199, 5, 5, 239, 166, 5, 5, 252, 179, 5, 5, 223, 177, 5, 5, 111, 205, 5, 5, 78, 207, 5, 5, 11, 171, 5, 5, 138, 231, 5, 5, 114, 209, 5, 5, 116, 244, 5, 5, 100, 195, 5, 5, 47, 211, 5, 5, 129, 235, 5, 5, 246, 237, 5, 5, 137, 223, 5, 5, 140, 209, 5, 5, 214, 222, 5, 5, 174, 178, 5, 5, 54, 236, 5, 5, 56, 201, 5, 5, 183, 200, 5, 5, 31, 238, 5, 5, 41, 178, 5, 5, 245, 169, 5, 5, 12, 186, 5, 5, 143, 231, 5, 5, 214, 172, 5, 5, 222, 169, 5, 5, 137, 169, 5, 5, 2, 221, 5, 5, 116, 163, 5, 5, 187, 191, 5, 5, 250, 183, 5, 5, 117, 180, 5, 5, 225, 207, 5, 5, 251, 183, 5, 5, 227, 182, 5, 5, 58, 188, 5, 5, 34, 192, 5, 5, 171, 164, 5, 5, 139, 240, 5, 5, 33, 203, 5, 5, 23, 166, 5, 5, 79, 231, 5, 5, 161, 165, 5, 5, 60, 180, 5, 5, 152, 170, 5, 5, 100, 190, 5, 5, 199, 164, 5, 5, 229, 249, 5, 5, 180, 206, 5, 5, 159, 176, 5, 5, 227, 232, 5, 5, 176, 217, 5, 5, 64, 188, 5, 5, 101, 224, 5, 5, 185, 207, 5, 5, 26, 188, 5, 5, 239, 168, 5, 5, 244, 189, 5, 5, 147, 178, 5, 5, 36, 186, 5, 5, 2, 169, 5, 5, 7, 235, 5, 5, 52, 211, 5, 5, 143, 176, 5, 5, 214, 199, 5, 5, 97, 185, 5, 5, 194, 200, 5, 5, 68, 222, 5, 5, 160, 188, 5, 5, 181, 240, 5, 5, 160, 223, 5, 5, 174, 165, 5, 5, 2, 185, 5, 5, 116, 224, 5, 5, 63, 216, 5, 5, 35, 221, 5, 5, 6, 225, 5, 5, 152, 204, 5, 5, 134, 238, 5, 5, 46, 177, 5, 5, 183, 211, 5, 5, 16, 188, 5, 5, 60, 171, 5, 5, 34, 221, 5, 5, 163, 170, 5, 5, 45, 169, 5, 5, 223, 229, 5, 5, 77, 228, 5, 5, 194, 199, 5, 5, 164, 170, 5, 5, 175, 242, 5, 5, 26, 165, 5, 5, 132, 198, 5, 5, 27, 165, 5, 5, 131, 228, 5, 5, 80, 195, 5, 5, 239, 195, 5, 5, 180, 164, 5, 5, 161, 201, 5, 5, 42, 231, 5, 5, 105, 236, 5, 5, 103, 215, 5, 5, 104, 215, 5, 5, 216, 204, 5, 5, 9, 172, 5, 5, 175, 213, 5, 5, 220, 229, 5, 5, 210, 175, 5, 5, 88, 219, 5, 5, 81, 183, 5, 5, 87, 220, 5, 5, 144, 167, 5, 5, 224, 213, 5, 5, 130, 189, 5, 5, 246, 165, 5, 5, 50, 165, 5, 5, 43, 231, 5, 5, 70, 238, 5, 5, 74, 194, 5, 5, 184, 191, 5, 5, 185, 191, 5, 5, 134, 172, 5, 5, 226, 225, 5, 5, 186, 201, 5, 5, 233, 193, 5, 5, 13, 172, 5, 5, 221, 235, 5, 5, 237, 196, 5, 5, 146, 186, 5, 5, 53, 191, 5, 5, 159, 181, 5, 5, 165, 227, 5, 5, 4, 207, 5, 5, 226, 248, 5, 5, 227, 248, 5, 5, 142, 224, 5, 5, 255, 211, 5, 5, 31, 189, 5, 5, 250, 224, 5, 5, 225, 227, 5, 5, 238, 196, 5, 5, 148, 242, 5, 5, 41, 247, 5, 5, 193, 242, 5, 5, 2, 238, 5, 5, 3, 201, 5, 5, 143, 248, 5, 5, 42, 247, 5, 5, 111, 216, 5, 5, 5, 164, 5, 5, 191, 201, 5, 5, 3, 212, 5, 5, 230, 241, 5, 5, 44, 247, 5, 5, 96, 164, 5, 5, 251, 238, 5, 5, 39, 246, 5, 5, 154, 243, 5, 5, 66, 196, 5, 5, 36, 212, 5, 5, 37, 186, 5, 5, 238, 177, 5, 5, 115, 212, 5, 5, 221, 173, 5, 5, 94, 181, 5, 5, 7, 184, 5, 5, 83, 179, 5, 5, 104, 195, 5, 5, 144, 197, 5, 5, 38, 212, 5, 5, 138, 226, 5, 5, 78, 194, 5, 5, 217, 175, 5, 5, 20, 209, 5, 5, 110, 196, 5, 5, 106, 180, 5, 5, 80, 194, 5, 5, 83, 168, 5, 5, 192, 195, 5, 5, 159, 213, 5, 5, 143, 200, 5, 5, 152, 224, 5, 5, 54, 168, 5, 5, 120, 176, 5, 5, 200, 200, 5, 5, 87, 179, 5, 5, 46, 247, 5, 5, 42, 179, 5, 5, 159, 170, 5, 5, 221, 229, 5, 5, 75, 228, 5, 5, 21, 243, 5, 5, 117, 198, 5, 5, 247, 164, 5, 5, 59, 247, 5, 5, 134, 226, 5, 5, 40, 231, 5, 5, 7, 246, 5, 5, 83, 215, 5, 5, 136, 201, 5, 5, 194, 204, 5, 5, 252, 171, 5, 5, 210, 195, 5, 5, 171, 170, 5, 5, 83, 219, 5, 5, 78, 183, 5, 5, 214, 193, 5, 5, 25, 247, 5, 5, 56, 179, 5, 5, 3, 172, 5, 5, 47, 191, 5, 5, 87, 195, 5, 5, 127, 172, 5, 5, 164, 191, 5, 5, 101, 197, 5, 5, 227, 196, 5, 5, 130, 204, 5, 5, 215, 248, 5, 5, 108, 186, 5, 5, 234, 177, 5, 5, 80, 248, 5, 5, 136, 242, 5, 5, 243, 224, 5, 5, 126, 213, 5, 5, 30, 247, 5, 5, 251, 200, 5, 5, 176, 242, 5, 5, 222, 237, 5, 5, 133, 224, 5, 5, 23, 186, 5, 5, 252, 211, 5, 5, 89, 200, 5, 5, 59, 196, 5, 5, 22, 212, 5, 5, 130, 197, 5, 5, 89, 181, 5, 5, 35, 175, 5, 5, 9, 212, 5, 5, 108, 198, 5, 5, 190, 200, 5, 5, 195, 174, 5, 5, 185, 174, 5, 5, 18, 185, 5, 5, 131, 178, 5, 5, 21, 216, 5, 5, 57, 210, 5, 5, 86, 245, 5, 5, 235, 191, 5, 5, 179, 218, 5, 5, 111, 243, 5, 5, 19, 244, 5, 5, 190, 175, 5, 5, 194, 237, 5, 5, 12, 229, 5, 5, 85, 164, 5, 5, 234, 166, 5, 5, 244, 195, 5, 5, 163, 249, 5, 5, 25, 231, 5, 5, 152, 218, 5, 5, 236, 167, 5, 5, 121, 206, 5, 5, 159, 173, 5, 5, 20, 244, 5, 5, 112, 243, 5, 5, 235, 220, 5, 5, 98, 203, 5, 5, 53, 225, 5, 5, 198, 237, 5, 5, 162, 211, 5, 5, 208, 176, 5, 5, 231, 232, 5, 5, 7, 233, 5, 5, 163, 177, 5, 5, 169, 189, 5, 5, 134, 237, 5, 5, 26, 182, 5, 5, 27, 182, 5, 5, 183, 163, 5, 5, 62, 166, 5, 5, 196, 174, 5, 5, 160, 202, 5, 5, 7, 190, 5, 5, 30, 191, 5, 5, 184, 224, 5, 5, 143, 216, 5, 5, 104, 178, 5, 5, 8, 186, 5, 5, 108, 169, 5, 5, 56, 206, 5, 5, 68, 229, 5, 5, 105, 178, 5, 5, 190, 215, 5, 5, 32, 239, 5, 5, 11, 194, 5, 5, 146, 247, 5, 5, 218, 191, 5, 5, 76, 228, 5, 5, 49, 208, 5, 5, 195, 209, 5, 5, 150, 216, 5, 5, 99, 226, 5, 5, 206, 174, 5, 5, 19, 165, 5, 5, 195, 213, 5, 5, 78, 228, 5, 5, 22, 229, 5, 5, 215, 218, 5, 5, 215, 168, 5, 5, 195, 175, 5, 5, 97, 249, 5, 5, 245, 220, 5, 5, 156, 225, 5, 5, 198, 215, 5, 5, 178, 175, 5, 5, 235, 183, 5, 5, 164, 249, 5, 5, 147, 215, 5, 5, 236, 166, 5, 5, 208, 172, 5, 5, 100, 232, 5, 5, 85, 240, 5, 5, 138, 178, 5, 5, 205, 190, 5, 5, 130, 243, 5, 5, 212, 206, 5, 5, 35, 226, 5, 5, 195, 224, 5, 5, 2, 168, 5, 5, 123, 189, 5, 5, 185, 179, 5, 5, 203, 215, 5, 5, 17, 190, 5, 5, 24, 229, 5, 5, 149, 215, 5, 5, 137, 243, 5, 5, 195, 221, 5, 5, 40, 244, 5, 5, 179, 230, 5, 5, 202, 213, 5, 5, 169, 226, 5, 5, 46, 243, 5, 5, 138, 243, 5, 5, 146, 229, 5, 5, 156, 176, 5, 5, 205, 247, 5, 5, 240, 190, 5, 5, 236, 245, 5, 5, 237, 245, 5, 5, 242, 183, 5, 5, 139, 178, 5, 5, 10, 237, 5, 5, 234, 238, 5, 5, 48, 201, 5, 5, 114, 203, 5, 5, 12, 237, 5, 5, 213, 243, 5, 5, 246, 191, 5, 5, 241, 242, 5, 5, 23, 200, 5, 5, 180, 191, 5, 5, 170, 249, 5, 5, 115, 245, 5, 5, 129, 196, 5, 5, 66, 238, 5, 5, 142, 213, 5, 5, 86, 180, 5, 5, 208, 202, 5, 5, 53, 221, 5, 5, 147, 175, 5, 5, 111, 238, 5, 5, 143, 213, 5, 5, 219, 243, 5, 5, 220, 235, 5, 5, 27, 230, 5, 5, 237, 188, 5, 5, 80, 182, 5, 5, 140, 176, 5, 5, 68, 204, 5, 5, 39, 237, 5, 5, 13, 237, 5, 5, 158, 238, 5, 5, 161, 215, 5, 5, 116, 173, 5, 5, 195, 183, 5, 5, 45, 239, 5, 5, 13, 202, 5, 5, 153, 175, 5, 5, 64, 213, 5, 5, 228, 243, 5, 5, 175, 237, 5, 5, 35, 200, 5, 5, 37, 191, 5, 5, 37, 213, 5, 5, 43, 187, 5, 5, 108, 187, 5, 5, 114, 187, 5, 5, 81, 222, 5, 5, 253, 198, 5, 5, 79, 218, 5, 5, 80, 237, 5, 5, 248, 207, 5, 5, 103, 183, 5, 5, 86, 240, 5, 5, 18, 212, 5, 5, 47, 243, 5, 5, 232, 210, 5, 5, 84, 210, 5, 5, 44, 218, 5, 5, 14, 241, 5, 5, 77, 202, 5, 5, 89, 193, 5, 5, 178, 200, 5, 5, 3, 219, 5, 5, 94, 192, 5, 5, 4, 208, 5, 5, 185, 233, 5, 5, 217, 249, 5, 5, 220, 206, 5, 5, 166, 225, 5, 5, 239, 198, 5, 5, 224, 231, 5, 5, 149, 190, 5, 5, 229, 247, 5, 5, 38, 224, 5, 5, 87, 177, 5, 5, 8, 237, 5, 5, 156, 198, 5, 5, 212, 245, 5, 5, 238, 229, 5, 5, 116, 194, 5, 5, 179, 225, 5, 5, 229, 231, 5, 5, 232, 249, 5, 5, 17, 202, 5, 5, 27, 213, 5, 5, 45, 224, 5, 5, 109, 188, 5, 5, 222, 186, 5, 5, 109, 183, 5, 5, 73, 217, 5, 5, 232, 231, 5, 5, 121, 202, 5, 5, 122, 202, 5, 5, 8, 248, 5, 5, 188, 247, 5, 5, 75, 217, 5, 5, 231, 182, 5, 5, 236, 172, 5, 5, 230, 235, 5, 5, 93, 172, 5, 5, 32, 247, 5, 5, 86, 173, 5, 5, 109, 167, 5, 5, 143, 249, 5, 5, 86, 187, 5, 5, 106, 223, 5, 5, 11, 244, 5, 5, 131, 189, 5, 5, 87, 185, 5, 5, 226, 238, 5, 5, 10, 164, 5, 5, 116, 245, 5, 5, 175, 177, 5, 5, 106, 176, 5, 5, 78, 199, 5, 5, 117, 245, 5, 5, 108, 177, 5, 5, 206, 216, 5, 5, 209, 202, 5, 5, 106, 172, 5, 5, 204, 224, 5, 5, 164, 214, 5, 5, 120, 219, 5, 5, 160, 181, 5, 5, 3, 221, 5, 5, 6, 167, 5, 5, 225, 169, 5, 5, 137, 189, 5, 5, 146, 208, 5, 5, 25, 220, 5, 5, 209, 199, 5, 5, 166, 227, 5, 5, 167, 172, 5, 5, 174, 201, 5, 5, 91, 220, 5, 5, 150, 243, 5, 5, 173, 183, 5, 5, 250, 221, 5, 5, 149, 182, 5, 5, 244, 188, 5, 5, 179, 189, 5, 5, 26, 179, 5, 5, 130, 244, 5, 5, 126, 245, 5, 5, 150, 189, 5, 5, 237, 229, 5, 5, 151, 189, 5, 5, 219, 213, 5, 5, 176, 201, 5, 5, 155, 243, 5, 5, 187, 169, 5, 5, 188, 169, 5, 5, 176, 249, 5, 5, 107, 188, 5, 5, 27, 171, 5, 5, 32, 166, 5, 5, 128, 245, 5, 5, 158, 246, 5, 5, 70, 196, 5, 5, 199, 243, 5, 5, 67, 213, 5, 5, 187, 240, 5, 5, 51, 238, 5, 5, 63, 176, 5, 5, 150, 180, 5, 5, 99, 236, 5, 5, 165, 248, 5, 5, 76, 190, 5, 5, 10, 191, 5, 5, 119, 175, 5, 5, 118, 202, 5, 5, 113, 167, 5, 5, 55, 244, 5, 5, 29, 163, 5, 5, 233, 167, 5, 5, 44, 246, 5, 5, 229, 173, 5, 5, 161, 213, 5, 5, 244, 202, 5, 5, 167, 169, 5, 5, 255, 170, 5, 5, 2, 171, 5, 5, 111, 185, 5, 5, 165, 187, 5, 5, 230, 199, 5, 5, 53, 238, 5, 5, 225, 172, 5, 5, 122, 175, 5, 5, 23, 191, 5, 5, 160, 189, 5, 5, 90, 238, 5, 5, 232, 173, 5, 5, 169, 187, 5, 5, 38, 164, 5, 5, 179, 197, 5, 5, 175, 197, 5, 5, 177, 197, 5, 5, 97, 166, 5, 5, 89, 211, 5, 5, 90, 211, 5, 5, 167, 206, 5, 5, 66, 181, 5, 5, 187, 220, 5, 5, 172, 206, 5, 5, 70, 225, 5, 5, 141, 211, 5, 5, 177, 206, 5, 5, 179, 206, 5, 5, 181, 177, 5, 5, 181, 206, 5, 5, 70, 181, 5, 5, 153, 172, 5, 5, 163, 172, 5, 5, 130, 195, 5, 5, 52, 171, 5, 5, 217, 171, 5, 5, 216, 171, 5, 5, 53, 164, 5, 5, 108, 241, 5, 5, 116, 181, 5, 5, 146, 200, 5, 5, 135, 241, 5, 5, 165, 191, 5, 5, 199, 177, 5, 5, 149, 223, 5, 5, 130, 210, 5, 5, 141, 230, 5, 5, 175, 176, 5, 5, 40, 178, 5, 5, 109, 245, 5, 5, 142, 230, 5, 5, 238, 210, 5, 5, 231, 205, 5, 5, 49, 217, 5, 5, 57, 184, 5, 5, 14, 163, 5, 5, 215, 243, 5, 5, 203, 235, 5, 5, 250, 182, 5, 5, 131, 169, 5, 5, 238, 232, 5, 5, 36, 175, 5, 5, 111, 164, 5, 5, 126, 240, 5, 5, 246, 241, 5, 5, 191, 179, 5, 5, 236, 163, 5, 5, 59, 203, 5, 5, 170, 201, 5, 5, 22, 167, 5, 5, 12, 244, 5, 5, 99, 166, 5, 5, 148, 243, 5, 5, 188, 191, 5, 5, 193, 179, 5, 5, 187, 203, 5, 5, 32, 214, 5, 5, 215, 238, 5, 5, 216, 238, 5, 5, 165, 198, 5, 5, 228, 194, 5, 5, 88, 209, 5, 5, 37, 175, 5, 5, 171, 209, 5, 5, 137, 181, 5, 5, 116, 170, 5, 5, 229, 244, 5, 5, 7, 237, 5, 5, 38, 181, 5, 5, 203, 209, 5, 5, 177, 217, 5, 5, 67, 197, 5, 5, 194, 164, 5, 5, 80, 179, 5, 5, 4, 229, 5, 5, 80, 166, 5, 5, 67, 196, 5, 5, 230, 206, 5, 5, 8, 185, 5, 5, 106, 214, 5, 5, 47, 230, 5, 5, 151, 164, 5, 5, 23, 239, 5, 5, 237, 209, 5, 5, 238, 209, 5, 5, 145, 197, 5, 5, 230, 214, 5, 5, 109, 172, 5, 5, 231, 206, 5, 5, 248, 204, 5, 5, 209, 165, 5, 5, 162, 176, 5, 5, 249, 204, 5, 5, 14, 247, 5, 5, 171, 179, 5, 5, 65, 192, 5, 5, 84, 179, 5, 5, 243, 163, 5, 5, 39, 214, 5, 5, 76, 167, 5, 5, 196, 191, 5, 5, 103, 165, 5, 5, 71, 196, 5, 5, 65, 180, 5, 5, 251, 185, 5, 5, 61, 194, 5, 5, 183, 209, 5, 5, 54, 217, 5, 5, 27, 202, 5, 5, 6, 205, 5, 5, 33, 206, 5, 5, 164, 224, 5, 5, 43, 175, 5, 5, 94, 199, 5, 5, 175, 215, 5, 5, 113, 210, 5, 5, 18, 178, 5, 5, 163, 245, 5, 5, 11, 205, 5, 5, 62, 186, 5, 5, 12, 246, 5, 5, 143, 230, 5, 5, 134, 174, 5, 5, 12, 170, 5, 5, 111, 212, 5, 5, 189, 191, 5, 5, 98, 246, 5, 5, 204, 209, 5, 5, 102, 198, 5, 5, 49, 225, 5, 5, 73, 196, 5, 5, 54, 230, 5, 5, 203, 249, 5, 5, 205, 227, 5, 5, 107, 246, 5, 5, 195, 191, 5, 5, 178, 163, 5, 5, 172, 179, 5, 5, 66, 192, 5, 5, 205, 249, 5, 5, 19, 192, 5, 5, 104, 165, 5, 5, 112, 246, 5, 5, 56, 237, 5, 5, 15, 178, 5, 5, 164, 245, 5, 5, 19, 178, 5, 5, 2, 246, 5, 5, 125, 230, 5, 5, 92, 174, 5, 5, 94, 198, 5, 5, 86, 246, 5, 5, 191, 249, 5, 5, 162, 179, 5, 5, 149, 204, 5, 5, 153, 204, 5, 5, 62, 215, 5, 5, 97, 214, 5, 5, 69, 215, 5, 5, 20, 202, 5, 5, 118, 243, 5, 5, 223, 231, 5, 5, 192, 249, 5, 5, 250, 227, 5, 5, 25, 185, 5, 5, 174, 167, 5, 5, 76, 201, 5, 5, 131, 244, 5, 5, 206, 249, 5, 5, 181, 167, 5, 5, 222, 238, 5, 5, 157, 243, 5, 5, 90, 216, 5, 5, 95, 199, 5, 5, 88, 241, 5, 5, 202, 177, 5, 5, 129, 211, 5, 5, 116, 196, 5, 5, 104, 216, 5, 5, 230, 229, 5, 5, 66, 198, 5, 5, 201, 216, 5, 5, 182, 239, 5, 5, 204, 216, 5, 5, 184, 237, 5, 5, 127, 237, 5, 5, 150, 196, 5, 5, 208, 208, 5, 5, 187, 198, 5, 5, 112, 248, 5, 5, 198, 246, 5, 5, 75, 164, 5, 5, 216, 178, 5, 5, 182, 176, 5, 5, 244, 175, 5, 5, 91, 167, 5, 5, 176, 166, 5, 5, 172, 189, 5, 5, 196, 223, 5, 5, 128, 206, 5, 5, 151, 185, 5, 5, 138, 197, 5, 5, 142, 209, 5, 5, 248, 237, 5, 5, 172, 237, 5, 5, 99, 246, 5, 5, 33, 214, 5, 5, 63, 196, 5, 5, 195, 228, 5, 5, 194, 239, 5, 5, 198, 228, 5, 5, 185, 197, 5, 5, 129, 178, 5, 5, 36, 248, 5, 5, 165, 168, 5, 5, 152, 185, 5, 5, 179, 174, 5, 5, 62, 203, 5, 5, 215, 179, 5, 5, 123, 200, 5, 5, 134, 196, 5, 5, 63, 203, 5, 5, 87, 190, 5, 5, 177, 181, 5, 5, 77, 225, 5, 5, 219, 244, 5, 5, 123, 249, 5, 5, 29, 199, 5, 5, 153, 198, 5, 5, 135, 206, 5, 5, 81, 182, 5, 5, 82, 182, 5, 5, 142, 168, 5, 5, 72, 185, 5, 5, 83, 182, 5, 5, 158, 231, 5, 5, 157, 231, 5, 5, 42, 226, 5, 5, 229, 177, 5, 5, 73, 185, 5, 5, 13, 180, 5, 5, 154, 198, 5, 5, 99, 210, 5, 5, 30, 195, 5, 5, 254, 234, 5, 5, 154, 217, 5, 5, 212, 210, 5, 5, 198, 239, 5, 5, 7, 236, 5, 5, 194, 180, 5, 5, 176, 237, 5, 5, 8, 238, 5, 5, 196, 242, 5, 5, 67, 247, 5, 5, 183, 195, 5, 5, 245, 249, 5, 5, 129, 222, 5, 5, 197, 242, 5, 5, 166, 215, 5, 5, 34, 216, 5, 5, 247, 239, 5, 5, 108, 249, 5, 5, 122, 236, 5, 5, 137, 177, 5, 5, 138, 177, 5, 5, 49, 206, 5, 5, 111, 196, 5, 5, 157, 185, 5, 5, 190, 185, 5, 5, 91, 168, 5, 5, 221, 245, 5, 5, 121, 170, 5, 5, 171, 234, 5, 5, 207, 241, 5, 5, 20, 198, 5, 5, 184, 209, 5, 5, 98, 193, 5, 5, 25, 218, 5, 5, 11, 243, 5, 5, 148, 200, 5, 5, 165, 224, 5, 5, 209, 230, 5, 5, 4, 171, 5, 5, 35, 166, 5, 5, 176, 215, 5, 5, 150, 168, 5, 5, 223, 189, 5, 5, 88, 186, 5, 5, 25, 184, 5, 5, 97, 212, 5, 5, 81, 241, 5, 5, 201, 177, 5, 5, 125, 211, 5, 5, 62, 198, 5, 5, 98, 216, 5, 5, 228, 229, 5, 5, 187, 216, 5, 5, 190, 216, 5, 5, 195, 246, 5, 5, 82, 193, 5, 5, 212, 178, 5, 5, 182, 208, 5, 5, 71, 164, 5, 5, 125, 237, 5, 5, 183, 198, 5, 5, 70, 248, 5, 5, 196, 189, 5, 5, 171, 189, 5, 5, 173, 166, 5, 5, 126, 206, 5, 5, 163, 203, 5, 5, 54, 203, 5, 5, 174, 181, 5, 5, 122, 197, 5, 5, 209, 244, 5, 5, 104, 208, 5, 5, 11, 195, 5, 5, 183, 239, 5, 5, 141, 168, 5, 5, 98, 200, 5, 5, 78, 182, 5, 5, 138, 198, 5, 5, 210, 244, 5, 5, 211, 244, 5, 5, 21, 195, 5, 5, 184, 180, 5, 5, 244, 249, 5, 5, 249, 237, 5, 5, 124, 222, 5, 5, 191, 242, 5, 5, 173, 237, 5, 5, 2, 184, 5, 5, 104, 170, 5, 5, 134, 177, 5, 5, 120, 236, 5, 5, 18, 198, 5, 5, 156, 185, 5, 5, 161, 224, 5, 5, 207, 230, 5, 5, 149, 168, 5, 5, 95, 212, 5, 5, 199, 173, 5, 5, 148, 165, 5, 5, 138, 193, 5, 5, 128, 174, 5, 5, 178, 213, 5, 5, 122, 220, 5, 5, 153, 165, 5, 5, 214, 235, 5, 5, 7, 163, 5, 5, 121, 179, 5, 5, 205, 187, 5, 5, 117, 193, 5, 5, 178, 217, 5, 5, 241, 227, 5, 5, 99, 207, 5, 5, 165, 165, 5, 5, 147, 235, 5, 5, 108, 246, 5, 5, 52, 244, 5, 5, 204, 241, 5, 5, 62, 236, 5, 5, 105, 195, 5, 5, 63, 236, 5, 5, 253, 172, 5, 5, 77, 188, 5, 5, 69, 213, 5, 5, 111, 188, 5, 5, 76, 168, 5, 5, 77, 168, 5, 5, 79, 188, 5, 5, 175, 165, 5, 5, 176, 165, 5, 5, 177, 165, 5, 5, 4, 188, 5, 5, 192, 165, 5, 5, 250, 235, 5, 5, 233, 173, 5, 5, 6, 215, 5, 5, 186, 173, 5, 5, 21, 244, 5, 5, 108, 220, 5, 5, 141, 165, 5, 5, 209, 235, 5, 5, 75, 207, 5, 5, 122, 235, 5, 5, 55, 236, 5, 5, 189, 241, 5, 5, 53, 188, 5, 5, 71, 168, 5, 5, 162, 165, 5, 5, 163, 165, 5, 5, 154, 172, 5, 5, 255, 172, 5, 5, 173, 172, 5, 5, 151, 172, 5, 5, 241, 226, 5, 5, 242, 226, 5, 5, 64, 234, 5, 5, 220, 201, 5, 5, 216, 177, 5, 5, 178, 235, 5, 5, 16, 182, 5, 5, 82, 220, 5, 5, 70, 237, 5, 5, 81, 216, 5, 5, 145, 182, 5, 5, 144, 230, 5, 5, 97, 248, 5, 5, 5, 208, 5, 5, 100, 223, 5, 5, 216, 243, 5, 5, 63, 173, 5, 5, 64, 173, 5, 5, 73, 237, 5, 5, 214, 243, 5, 5, 221, 179, 5, 5, 184, 239, 5, 5, 19, 233, 5, 5, 51, 165, 5, 5, 201, 203, 5, 5, 163, 179, 5, 5, 91, 188, 5, 5, 190, 233, 5, 5, 8, 181, 5, 5, 189, 235, 5, 5, 27, 164, 5, 5, 88, 227, 5, 5, 113, 178, 5, 5, 200, 197, 5, 5, 147, 230, 5, 5, 112, 181, 5, 5, 56, 204, 5, 5, 60, 203, 5, 5, 78, 166, 5, 5, 190, 241, 5, 5, 192, 182, 5, 5, 191, 233, 5, 5, 84, 216, 5, 5, 73, 244, 5, 5, 9, 208, 5, 5, 62, 239, 5, 5, 61, 243, 5, 5, 33, 185, 5, 5, 244, 212, 5, 5, 71, 234, 5, 5, 60, 163, 5, 5, 37, 183, 5, 5, 250, 237, 5, 5, 145, 166, 5, 5, 69, 208, 5, 5, 214, 175, 5, 5, 187, 236, 5, 5, 230, 247, 5, 5, 96, 227, 5, 5, 196, 241, 5, 5, 12, 201, 5, 5, 190, 193, 5, 5, 102, 227, 5, 5, 132, 205, 5, 5, 107, 222, 5, 5, 79, 166, 5, 5, 228, 215, 5, 5, 146, 166, 5, 5, 119, 241, 5, 5, 10, 180, 5, 5, 169, 172, 5, 5, 28, 221, 5, 5, 149, 246, 5, 5, 93, 194, 5, 5, 3, 238, 5, 5, 185, 184, 5, 5, 202, 249, 5, 5, 32, 199, 5, 5, 72, 201, 5, 5, 10, 181, 5, 5, 129, 200, 5, 5, 244, 204, 5, 5, 42, 198, 5, 5, 80, 238, 5, 5, 102, 180, 5, 5, 90, 173, 5, 5, 153, 246, 5, 5, 76, 193, 5, 5, 141, 223, 5, 5, 129, 237, 5, 5, 105, 180, 5, 5, 130, 163, 5, 5, 17, 192, 5, 5, 155, 213, 5, 5, 177, 196, 5, 5, 108, 188, 5, 5, 61, 236, 5, 5, 198, 179, 5, 5, 131, 241, 5, 5, 158, 249, 5, 5, 84, 169, 5, 5, 107, 180, 5, 5, 119, 170, 5, 5, 90, 165, 5, 5, 163, 248, 5, 5, 4, 215, 5, 5, 206, 205, 5, 5, 100, 236, 5, 5, 180, 196, 5, 5, 152, 222, 5, 5, 123, 228, 5, 5, 119, 223, 5, 5, 64, 176, 5, 5, 197, 240, 5, 5, 24, 202, 5, 5, 25, 230, 5, 5, 84, 225, 5, 5, 237, 249, 5, 5, 173, 179, 5, 5, 160, 246, 5, 5, 91, 216, 5, 5, 100, 220, 5, 5, 91, 173, 5, 5, 203, 197, 5, 5, 3, 171, 5, 5, 92, 243, 5, 5, 90, 169, 5, 5, 120, 224, 5, 5, 92, 216, 5, 5, 119, 190, 5, 5, 166, 232, 5, 5, 135, 183, 5, 5, 226, 226, 5, 5, 212, 177, 5, 5, 200, 201, 5, 5, 255, 181, 5, 5, 67, 180, 5, 5, 140, 182, 5, 5, 109, 213, 5, 5, 126, 230, 5, 5, 56, 248, 5, 5, 85, 223, 5, 5, 49, 173, 5, 5, 208, 243, 5, 5, 79, 205, 5, 5, 45, 227, 5, 5, 17, 164, 5, 5, 169, 235, 5, 5, 99, 178, 5, 5, 141, 239, 5, 5, 56, 239, 5, 5, 18, 230, 5, 5, 79, 216, 5, 5, 102, 197, 5, 5, 32, 185, 5, 5, 34, 204, 5, 5, 185, 212, 5, 5, 68, 166, 5, 5, 48, 163, 5, 5, 201, 175, 5, 5, 211, 237, 5, 5, 186, 247, 5, 5, 34, 183, 5, 5, 50, 208, 5, 5, 180, 193, 5, 5, 7, 201, 5, 5, 204, 215, 5, 5, 85, 194, 5, 5, 134, 223, 5, 5, 159, 196, 5, 5, 56, 193, 5, 5, 42, 236, 5, 5, 82, 224, 5, 5, 103, 241, 5, 5, 69, 169, 5, 5, 102, 163, 5, 5, 47, 188, 5, 5, 250, 214, 5, 5, 199, 205, 5, 5, 112, 170, 5, 5, 92, 236, 5, 5, 147, 222, 5, 5, 133, 183, 5, 5, 17, 228, 5, 5, 117, 196, 5, 5, 173, 193, 5, 5, 41, 216, 5, 5, 106, 173, 5, 5, 111, 163, 5, 5, 191, 183, 5, 5, 63, 165, 5, 5, 119, 163, 5, 5, 230, 181, 5, 5, 70, 198, 5, 5, 170, 172, 5, 5, 245, 204, 5, 5, 219, 238, 5, 5, 85, 238, 5, 5, 248, 174, 5, 5, 72, 215, 5, 5, 125, 249, 5, 5, 10, 216, 5, 5, 123, 173, 5, 5, 133, 168, 5, 5, 12, 216, 5, 5, 34, 169, 5, 5, 98, 248, 5, 5, 241, 173, 5, 5, 139, 198, 5, 5, 176, 176, 5, 5, 38, 182, 5, 5, 149, 219, 5, 5, 47, 223, 5, 5, 235, 217, 5, 5, 243, 221, 5, 5, 91, 218, 5, 5, 4, 167, 5, 5, 32, 216, 5, 5, 13, 230, 5, 5, 215, 235, 5, 5, 172, 249, 5, 5, 35, 247, 5, 5, 126, 218, 5, 5, 45, 176, 5, 5, 50, 191, 5, 5, 60, 242, 5, 5, 112, 163, 5, 5, 28, 164, 5, 5, 168, 207, 5, 5, 42, 182, 5, 5, 127, 240, 5, 5, 180, 211, 5, 5, 36, 247, 5, 5, 184, 211, 5, 5, 87, 208, 5, 5, 190, 166, 5, 5, 215, 219, 5, 5, 18, 249, 5, 5, 243, 166, 5, 5, 65, 220, 5, 5, 88, 207, 5, 5, 248, 221, 5, 5, 145, 184, 5, 5, 227, 206, 5, 5, 224, 167, 5, 5, 184, 232, 5, 5, 95, 203, 5, 5, 138, 189, 5, 5, 229, 220, 5, 5, 123, 179, 5, 5, 237, 174, 5, 5, 74, 244, 5, 5, 21, 227, 5, 5, 73, 165, 5, 5, 43, 182, 5, 5, 44, 182, 5, 5, 190, 235, 5, 5, 59, 188, 5, 5, 57, 169, 5, 5, 116, 168, 5, 5, 196, 179, 5, 5, 29, 219, 5, 5, 84, 231, 5, 5, 141, 180, 5, 5, 25, 179, 5, 5, 196, 217, 5, 5, 49, 212, 5, 5, 170, 228, 5, 5, 88, 208, 5, 5, 63, 239, 5, 5, 201, 197, 5, 5, 120, 163, 5, 5, 156, 221, 5, 5, 152, 243, 5, 5, 150, 190, 5, 5, 75, 220, 5, 5, 58, 176, 5, 5, 104, 214, 5, 5, 68, 188, 5, 5, 7, 207, 5, 5, 120, 237, 5, 5, 213, 210, 5, 5, 211, 170, 5, 5, 128, 186, 5, 5, 155, 198, 5, 5, 232, 181, 5, 5, 125, 217, 5, 5, 101, 182, 5, 5, 74, 208, 5, 5, 186, 237, 5, 5, 144, 225, 5, 5, 135, 216, 5, 5, 43, 238, 5, 5, 65, 190, 5, 5, 86, 222, 5, 5, 29, 220, 5, 5, 206, 227, 5, 5, 118, 168, 5, 5, 221, 196, 5, 5, 199, 179, 5, 5, 60, 185, 5, 5, 117, 233, 5, 5, 73, 195, 5, 5, 233, 208, 5, 5, 234, 208, 5, 5, 73, 243, 5, 5, 171, 172, 5, 5, 26, 236, 5, 5, 253, 242, 5, 5, 197, 191, 5, 5, 231, 241, 5, 5, 129, 245, 5, 5, 125, 168, 5, 5, 72, 235, 5, 5, 126, 168, 5, 5, 145, 210, 5, 5, 172, 172, 5, 5, 78, 201, 5, 5, 146, 210, 5, 5, 173, 199, 5, 5, 246, 247, 5, 5, 197, 180, 5, 5, 89, 212, 5, 5, 222, 213, 5, 5, 119, 233, 5, 5, 140, 196, 5, 5, 49, 230, 5, 5, 148, 235, 5, 5, 120, 193, 5, 5, 49, 182, 5, 5, 247, 193, 5, 5, 64, 236, 5, 5, 107, 210, 5, 5, 136, 225, 5, 5, 48, 219, 5, 5, 78, 188, 5, 5, 183, 167, 5, 5, 145, 180, 5, 5, 70, 213, 5, 5, 50, 234, 5, 5, 10, 178, 5, 5, 85, 243, 5, 5, 80, 249, 5, 5, 112, 232, 5, 5, 27, 236, 5, 5, 124, 228, 5, 5, 234, 186, 5, 5, 49, 219, 5, 5, 242, 229, 5, 5, 77, 190, 5, 5, 71, 213, 5, 5, 107, 190, 5, 5, 30, 221, 5, 5, 152, 169, 5, 5, 160, 238, 5, 5, 72, 234, 5, 5, 178, 165, 5, 5, 73, 235, 5, 5, 195, 211, 5, 5, 107, 165, 5, 5, 50, 219, 5, 5, 170, 248, 5, 5, 219, 216, 5, 5, 128, 199, 5, 5, 63, 163, 5, 5, 196, 236, 5, 5, 185, 214, 5, 5, 2, 189, 5, 5, 154, 220, 5, 5, 198, 178, 5, 5, 82, 188, 5, 5, 51, 182, 5, 5, 142, 234, 5, 5, 185, 177, 5, 5, 22, 204, 5, 5, 6, 181, 5, 5, 93, 243, 5, 5, 115, 190, 5, 5, 102, 220, 5, 5, 112, 206, 5, 5, 206, 240, 5, 5, 136, 241, 5, 5, 18, 182, 5, 5, 134, 168, 5, 5, 73, 220, 5, 5, 98, 243, 5, 5, 25, 191, 5, 5, 56, 191, 5, 5, 149, 180, 5, 5, 65, 216, 5, 5, 249, 202, 5, 5, 249, 229, 5, 5, 16, 207, 5, 5, 74, 213, 5, 5, 166, 200, 5, 5, 245, 186, 5, 5, 181, 165, 5, 5, 29, 169, 5, 5, 49, 248, 5, 5, 19, 182, 5, 5, 212, 217, 5, 5, 32, 223, 5, 5, 156, 211, 5, 5, 9, 230, 5, 5, 69, 163, 5, 5, 46, 191, 5, 5, 183, 232, 5, 5, 15, 227, 5, 5, 170, 235, 5, 5, 197, 174, 5, 5, 198, 206, 5, 5, 97, 232, 5, 5, 218, 221, 5, 5, 28, 182, 5, 5, 143, 184, 5, 5, 87, 203, 5, 5, 39, 240, 5, 5, 85, 175, 5, 5, 127, 214, 5, 5, 46, 169, 5, 5, 119, 243, 5, 5, 244, 203, 5, 5, 111, 199, 5, 5, 136, 190, 5, 5, 196, 197, 5, 5, 107, 168, 5, 5, 142, 165, 5, 5, 171, 186, 5, 5, 140, 225, 5, 5, 35, 243, 5, 5, 121, 216, 5, 5, 204, 210, 5, 5, 51, 208, 5, 5, 193, 208, 5, 5, 194, 208, 5, 5, 63, 195, 5, 5, 83, 222, 5, 5, 90, 233, 5, 5, 155, 236, 5, 5, 65, 234, 5, 5, 124, 168, 5, 5, 246, 218, 5, 5, 123, 196, 5, 5, 18, 236, 5, 5, 110, 245, 5, 5, 148, 238, 5, 5, 48, 188, 5, 5, 89, 210, 5, 5, 98, 228, 5, 5, 154, 165, 5, 5, 99, 190, 5, 5, 52, 235, 5, 5, 143, 224, 5, 5, 69, 220, 5, 5, 236, 202, 5, 5, 240, 229, 5, 5, 59, 216, 5, 5, 44, 193, 5, 5, 75, 246, 5, 5, 87, 246, 5, 5, 88, 246, 5, 5, 219, 247, 5, 5, 162, 192, 5, 5, 118, 240, 5, 5, 12, 239, 5, 5, 213, 179, 5, 5, 230, 203, 5, 5, 55, 164, 5, 5, 240, 163, 5, 5, 162, 233, 5, 5, 206, 180, 5, 5, 221, 176, 5, 5, 17, 196, 5, 5, 186, 197, 5, 5, 149, 209, 5, 5, 39, 198, 5, 5, 50, 196, 5, 5, 193, 191, 5, 5, 93, 182, 5, 5, 244, 192, 5, 5, 117, 168, 5, 5, 95, 165, 5, 5, 31, 195, 5, 5, 87, 209, 5, 5, 247, 247, 5, 5, 18, 237, 5, 5, 250, 185, 5, 5, 134, 163, 5, 5, 181, 214, 5, 5, 157, 164, 5, 5, 135, 163, 5, 5, 165, 233, 5, 5, 187, 196, 5, 5, 44, 166, 5, 5, 87, 169, 5, 5, 5, 188, 5, 5, 252, 185, 5, 5, 186, 214, 5, 5, 17, 178, 5, 5, 181, 232, 5, 5, 19, 237, 5, 5, 209, 180, 5, 5, 46, 166, 5, 5, 188, 196, 5, 5, 224, 189, 5, 5, 4, 192, 5, 5, 7, 192, 5, 5, 199, 209, 5, 5, 116, 195, 5, 5, 189, 209, 5, 5, 191, 185, 5, 5, 38, 236, 5, 5, 143, 165, 5, 5, 208, 196, 5, 5, 211, 196, 5, 5, 177, 176, 5, 5, 179, 173, 5, 5, 252, 214, 5, 5, 67, 230, 5, 5, 34, 170, 5, 5, 46, 176, 5, 5, 215, 196, 5, 5, 39, 166, 5, 5, 143, 172, 5, 5, 36, 181, 5, 5, 225, 167, 5, 5, 151, 231, 5, 5, 144, 172, 5, 5, 70, 230, 5, 5, 234, 180, 5, 5, 56, 164, 5, 5, 74, 165, 5, 5, 39, 170, 5, 5, 138, 174, 5, 5, 44, 239, 5, 5, 85, 231, 5, 5, 186, 211, 5, 5, 211, 194, 5, 5, 254, 214, 5, 5, 64, 196, 5, 5, 220, 202, 5, 5, 176, 167, 5, 5, 3, 222, 5, 5, 242, 227, 5, 5, 218, 236, 5, 5, 238, 180, 5, 5, 115, 187, 5, 5, 41, 166, 5, 5, 118, 233, 5, 5, 192, 176, 5, 5, 179, 167, 5, 5, 108, 237, 5, 5, 100, 221, 5, 5, 88, 212, 5, 5, 120, 233, 5, 5, 235, 229, 5, 5, 64, 205, 5, 5, 62, 182, 5, 5, 36, 199, 5, 5, 254, 184, 5, 5, 139, 204, 5, 5, 245, 208, 5, 5, 164, 188, 5, 5, 231, 220, 5, 5, 45, 166, 5, 5, 185, 167, 5, 5, 59, 169, 5, 5, 86, 236, 5, 5, 122, 170, 5, 5, 123, 170, 5, 5, 128, 236, 5, 5, 220, 216, 5, 5, 213, 187, 5, 5, 110, 210, 5, 5, 109, 210, 5, 5, 132, 183, 5, 5, 201, 200, 5, 5, 72, 196, 5, 5, 132, 184, 5, 5, 47, 166, 5, 5, 215, 187, 5, 5, 161, 230, 5, 5, 202, 175, 5, 5, 213, 175, 5, 5, 82, 183, 5, 5, 192, 201, 5, 5, 156, 213, 5, 5, 220, 175, 5, 5, 112, 198, 5, 5, 221, 175, 5, 5, 222, 175, 5, 5, 140, 204, 5, 5, 41, 225, 5, 5, 179, 248, 5, 5, 180, 248, 5, 5, 170, 191, 5, 5, 70, 243, 5, 5, 180, 174, 5, 5, 251, 210, 5, 5, 245, 193, 5, 5, 121, 233, 5, 5, 112, 188, 5, 5, 248, 193, 5, 5, 243, 228, 5, 5, 171, 248, 5, 5, 13, 194, 5, 5, 10, 170, 5, 5, 210, 202, 5, 5, 217, 208, 5, 5, 129, 191, 5, 5, 131, 196, 5, 5, 11, 201, 5, 5, 9, 163, 5, 5, 204, 166, 5, 5, 15, 170, 5, 5, 212, 216, 5, 5, 250, 242, 5, 5, 172, 214, 5, 5, 66, 189, 5, 5, 159, 248, 5, 5, 84, 182, 5, 5, 150, 208, 5, 5, 105, 247, 5, 5, 67, 189, 5, 5, 155, 246, 5, 5, 131, 192, 5, 5, 27, 223, 5, 5, 78, 168, 5, 5, 108, 165, 5, 5, 88, 169, 5, 5, 144, 202, 5, 5, 221, 216, 5, 5, 62, 224, 5, 5, 17, 243, 5, 5, 103, 220, 5, 5, 189, 247, 5, 5, 134, 175, 5, 5, 93, 230, 5, 5, 175, 203, 5, 5, 35, 163, 5, 5, 171, 201, 5, 5, 17, 182, 5, 5, 109, 177, 5, 5, 139, 202, 5, 5, 36, 217, 5, 5, 174, 237, 5, 5, 251, 237, 5, 5, 143, 227, 5, 5, 189, 198, 5, 5, 147, 182, 5, 5, 138, 169, 5, 5, 231, 203, 5, 5, 196, 228, 5, 5, 95, 164, 5, 5, 68, 249, 5, 5, 11, 168, 5, 5, 240, 189, 5, 5, 37, 247, 5, 5, 25, 199, 5, 5, 198, 189, 5, 5, 216, 231, 5, 5, 164, 173, 5, 5, 111, 173, 5, 5, 253, 182, 5, 5, 62, 242, 5, 5, 127, 168, 5, 5, 156, 169, 5, 5, 142, 197, 5, 5, 215, 213, 5, 5, 187, 211, 5, 5, 101, 198, 5, 5, 228, 167, 5, 5, 141, 189, 5, 5, 46, 182, 5, 5, 61, 163, 5, 5, 42, 210, 5, 5, 205, 168, 5, 5, 142, 174, 5, 5, 82, 165, 5, 5, 207, 233, 5, 5, 150, 246, 5, 5, 228, 206, 5, 5, 88, 177, 5, 5, 26, 163, 5, 5, 65, 242, 5, 5, 145, 194, 5, 5, 229, 167, 5, 5, 68, 184, 5, 5, 131, 216, 5, 5, 125, 179, 5, 5, 58, 164, 5, 5, 242, 174, 5, 5, 102, 219, 5, 5, 96, 207, 5, 5, 27, 193, 5, 5, 116, 227, 5, 5, 27, 179, 5, 5, 123, 175, 5, 5, 81, 179, 5, 5, 190, 203, 5, 5, 95, 229, 5, 5, 58, 198, 5, 5, 87, 216, 5, 5, 46, 239, 5, 5, 252, 238, 5, 5, 40, 246, 5, 5, 246, 219, 5, 5, 157, 221, 5, 5, 248, 243, 5, 5, 207, 187, 5, 5, 151, 190, 5, 5, 205, 182, 5, 5, 207, 180, 5, 5, 177, 240, 5, 5, 96, 166, 5, 5, 104, 246, 5, 5, 14, 204, 5, 5, 19, 196, 5, 5, 242, 193, 5, 5, 70, 215, 5, 5, 194, 191, 5, 5, 55, 200, 5, 5, 224, 185, 5, 5, 2, 175, 5, 5, 131, 195, 5, 5, 32, 215, 5, 5, 74, 178, 5, 5, 38, 208, 5, 5, 195, 180, 5, 5, 215, 171, 5, 5, 243, 227, 5, 5, 103, 219, 5, 5, 219, 236, 5, 5, 214, 210, 5, 5, 245, 247, 5, 5, 38, 183, 5, 5, 160, 219, 5, 5, 146, 194, 5, 5, 245, 192, 5, 5, 199, 236, 5, 5, 38, 245, 5, 5, 44, 210, 5, 5, 72, 228, 5, 5, 220, 168, 5, 5, 241, 200, 5, 5, 40, 181, 5, 5, 15, 187, 5, 5, 145, 227, 5, 5, 146, 227, 5, 5, 132, 195, 5, 5, 248, 170, 5, 5, 146, 225, 5, 5, 172, 187, 5, 5, 121, 232, 5, 5, 168, 213, 5, 5, 23, 250, 5, 5, 175, 183, 5, 5, 159, 231, 5, 5, 237, 208, 5, 5, 101, 221, 5, 5, 89, 216, 5, 5, 39, 183, 5, 5, 113, 226, 5, 5, 237, 202, 5, 5, 91, 177, 5, 5, 121, 209, 5, 5, 97, 193, 5, 5, 41, 220, 5, 5, 246, 178, 5, 5, 61, 185, 5, 5, 223, 172, 5, 5, 176, 183, 5, 5, 188, 240, 5, 5, 222, 196, 5, 5, 71, 190, 5, 5, 181, 204, 5, 5, 253, 224, 5, 5, 107, 206, 5, 5, 216, 190, 5, 5, 153, 189, 5, 5, 110, 232, 5, 5, 229, 186, 5, 5, 18, 171, 5, 5, 122, 233, 5, 5, 45, 219, 5, 5, 71, 184, 5, 5, 100, 199, 5, 5, 187, 248, 5, 5, 17, 177, 5, 5, 221, 226, 5, 5, 175, 228, 5, 5, 17, 201, 5, 5, 198, 180, 5, 5, 43, 198, 5, 5, 137, 216, 5, 5, 162, 221, 5, 5, 3, 185, 5, 5, 123, 233, 5, 5, 57, 232, 5, 5, 211, 165, 5, 5, 212, 165, 5, 5, 243, 200, 5, 5, 90, 212, 5, 5, 59, 232, 5, 5, 252, 245, 5, 5, 253, 245, 5, 5, 251, 247, 5, 5, 92, 226, 5, 5, 50, 230, 5, 5, 64, 177, 5, 5, 80, 201, 5, 5, 210, 188, 5, 5, 3, 243, 5, 5, 239, 209, 5, 5, 240, 209, 5, 5, 69, 205, 5, 5, 217, 164, 5, 5, 11, 238, 5, 5, 54, 244, 5, 5, 250, 174, 5, 5, 9, 236, 5, 5, 148, 192, 5, 5, 182, 213, 5, 5, 63, 182, 5, 5, 37, 199, 5, 5, 118, 227, 5, 5, 245, 230, 5, 5, 184, 217, 5, 5, 16, 247, 5, 5, 194, 210, 5, 5, 21, 198, 5, 5, 11, 245, 5, 5, 165, 173, 5, 5, 28, 236, 5, 5, 80, 188, 5, 5, 40, 169, 5, 5, 8, 227, 5, 5, 222, 226, 5, 5, 81, 194, 5, 5, 163, 231, 5, 5, 147, 180, 5, 5, 51, 179, 5, 5, 208, 241, 5, 5, 181, 237, 5, 5, 246, 249, 5, 5, 148, 210, 5, 5, 246, 208, 5, 5, 17, 247, 5, 5, 246, 230, 5, 5, 227, 228, 5, 5, 167, 188, 5, 5, 127, 238, 5, 5, 30, 185, 5, 5, 37, 200, 5, 5, 22, 172, 5, 5, 245, 202, 5, 5, 42, 181, 5, 5, 196, 211, 5, 5, 7, 205, 5, 5, 176, 228, 5, 5, 124, 170, 5, 5, 73, 213, 5, 5, 45, 210, 5, 5, 191, 165, 5, 5, 246, 202, 5, 5, 242, 202, 5, 5, 181, 221, 5, 5, 34, 206, 5, 5, 5, 215, 5, 5, 130, 222, 5, 5, 238, 249, 5, 5, 31, 221, 5, 5, 225, 195, 5, 5, 77, 217, 5, 5, 120, 165, 5, 5, 173, 248, 5, 5, 197, 211, 5, 5, 116, 216, 5, 5, 64, 163, 5, 5, 81, 204, 5, 5, 248, 217, 5, 5, 197, 236, 5, 5, 82, 201, 5, 5, 77, 233, 5, 5, 127, 228, 5, 5, 128, 228, 5, 5, 2, 173, 5, 5, 62, 194, 5, 5, 3, 189, 5, 5, 249, 217, 5, 5, 154, 171, 5, 5, 101, 213, 5, 5, 212, 232, 5, 5, 118, 216, 5, 5, 142, 173, 5, 5, 82, 194, 5, 5, 59, 198, 5, 5, 99, 196, 5, 5, 60, 232, 5, 5, 29, 236, 5, 5, 104, 220, 5, 5, 150, 192, 5, 5, 63, 194, 5, 5, 231, 244, 5, 5, 20, 187, 5, 5, 25, 225, 5, 5, 164, 185, 5, 5, 169, 229, 5, 5, 91, 238, 5, 5, 210, 230, 5, 5, 248, 202, 5, 5, 41, 217, 5, 5, 96, 199, 5, 5, 170, 229, 5, 5, 167, 187, 5, 5, 216, 187, 5, 5, 168, 187, 5, 5, 144, 171, 5, 5, 207, 220, 5, 5, 93, 216, 5, 5, 17, 238, 5, 5, 225, 189, 5, 5, 126, 194, 5, 5, 246, 186, 5, 5, 81, 215, 5, 5, 169, 247, 5, 5, 133, 175, 5, 5, 132, 202, 5, 5, 53, 242, 5, 5, 146, 182, 5, 5, 233, 189, 5, 5, 163, 173, 5, 5, 0, 152, 5, 5, 131, 197, 5, 5, 22, 163, 5, 5, 198, 168, 5, 5, 67, 184, 5, 5, 197, 189, 5, 5, 58, 242, 5, 5, 99, 219, 5, 5, 227, 174, 5, 5, 23, 163, 5, 5, 52, 164, 5, 5, 56, 198, 5, 5, 7, 168, 5, 5, 118, 179, 5, 5, 234, 193, 5, 5, 181, 203, 5, 5, 125, 195, 5, 5, 90, 246, 5, 5, 43, 239, 5, 5, 193, 182, 5, 5, 56, 232, 5, 5, 57, 198, 5, 5, 89, 196, 5, 5, 211, 202, 5, 5, 2, 204, 5, 5, 61, 215, 5, 5, 100, 219, 5, 5, 85, 216, 5, 5, 238, 217, 5, 5, 243, 192, 5, 5, 207, 186, 5, 5, 145, 188, 5, 5, 234, 204, 5, 5, 11, 187, 5, 5, 4, 227, 5, 5, 37, 181, 5, 5, 143, 194, 5, 5, 144, 227, 5, 5, 238, 200, 5, 5, 35, 208, 5, 5, 215, 202, 5, 5, 21, 250, 5, 5, 106, 211, 5, 5, 142, 189, 5, 5, 224, 208, 5, 5, 108, 232, 5, 5, 168, 172, 5, 5, 220, 196, 5, 5, 251, 224, 5, 5, 96, 193, 5, 5, 174, 183, 5, 5, 69, 184, 5, 5, 89, 177, 5, 5, 105, 206, 5, 5, 171, 228, 5, 5, 215, 226, 5, 5, 152, 231, 5, 5, 140, 173, 5, 5, 60, 177, 5, 5, 214, 164, 5, 5, 23, 225, 5, 5, 192, 180, 5, 5, 13, 247, 5, 5, 251, 245, 5, 5, 6, 236, 5, 5, 6, 238, 5, 5, 235, 209, 5, 5, 246, 174, 5, 5, 73, 201, 5, 5, 87, 212, 5, 5, 28, 206, 5, 5, 208, 165, 5, 5, 25, 236, 5, 5, 125, 238, 5, 5, 235, 208, 5, 5, 49, 179, 5, 5, 79, 194, 5, 5, 202, 241, 5, 5, 124, 167, 5, 5, 250, 204, 5, 5, 29, 185, 5, 5, 190, 165, 5, 5, 76, 217, 5, 5, 119, 165, 5, 5, 120, 170, 5, 5, 20, 172, 5, 5, 234, 249, 5, 5, 159, 246, 5, 5, 68, 213, 5, 5, 60, 194, 5, 5, 125, 228, 5, 5, 255, 188, 5, 5, 76, 233, 5, 5, 93, 199, 5, 5, 149, 192, 5, 5, 18, 187, 5, 5, 101, 220, 5, 5, 94, 194, 5, 5, 33, 184, 5, 5, 233, 239, 5, 5, 129, 174, 5, 5, 93, 187, 5, 5, 136, 204, 5, 5, 107, 166, 5, 5, 53, 243, 5, 5, 161, 174, 5, 5, 84, 177, 5, 5, 238, 201, 5, 5, 7, 174, 5, 5, 229, 230, 5, 5, 128, 192, 5, 5, 211, 226, 5, 5, 8, 174, 5, 5, 44, 172, 5, 5, 29, 164, 5, 5, 16, 249, 5, 5, 93, 218, 5, 5, 87, 199, 5, 5, 242, 210, 5, 5, 50, 174, 5, 5, 77, 164, 5, 5, 38, 247, 5, 5, 235, 204, 5, 5, 213, 226, 5, 5, 134, 248, 5, 5, 237, 172, 5, 5, 153, 241, 5, 5, 175, 207, 5, 5, 176, 207, 5, 5, 30, 168, 5, 5, 156, 200, 5, 5, 4, 221, 5, 5, 32, 164, 5, 5, 62, 243, 5, 5, 234, 240, 5, 5, 126, 221, 5, 5, 157, 173, 5, 5, 9, 174, 5, 5, 39, 247, 5, 5, 174, 238, 5, 5, 111, 175, 5, 5, 153, 191, 5, 5, 196, 230, 5, 5, 143, 189, 5, 5, 135, 171, 5, 5, 143, 174, 5, 5, 47, 182, 5, 5, 247, 171, 5, 5, 110, 187, 5, 5, 225, 165, 5, 5, 30, 219, 5, 5, 154, 191, 5, 5, 19, 249, 5, 5, 196, 233, 5, 5, 214, 211, 5, 5, 167, 214, 5, 5, 23, 223, 5, 5, 56, 176, 5, 5, 229, 206, 5, 5, 228, 241, 5, 5, 28, 193, 5, 5, 244, 178, 5, 5, 144, 248, 5, 5, 13, 244, 5, 5, 102, 231, 5, 5, 236, 240, 5, 5, 3, 181, 5, 5, 145, 248, 5, 5, 123, 181, 5, 5, 125, 244, 5, 5, 126, 178, 5, 5, 9, 245, 5, 5, 47, 239, 5, 5, 119, 193, 5, 5, 81, 238, 5, 5, 94, 218, 5, 5, 72, 243, 5, 5, 208, 198, 5, 5, 173, 214, 5, 5, 64, 203, 5, 5, 208, 187, 5, 5, 158, 221, 5, 5, 51, 244, 5, 5, 193, 180, 5, 5, 177, 201, 5, 5, 152, 190, 5, 5, 201, 230, 5, 5, 99, 170, 5, 5, 157, 191, 5, 5, 96, 230, 5, 5, 15, 204, 5, 5, 2, 215, 5, 5, 28, 220, 5, 5, 95, 218, 5, 5, 153, 190, 5, 5, 213, 198, 5, 5, 96, 184, 5, 5, 19, 238, 5, 5, 166, 190, 5, 5, 65, 203, 5, 5, 230, 202, 5, 5, 233, 181, 5, 5, 158, 200, 5, 5, 61, 182, 5, 5, 117, 166, 5, 5, 39, 245, 5, 5, 227, 207, 5, 5, 160, 248, 5, 5, 131, 163, 5, 5, 75, 208, 5, 5, 23, 218, 5, 5, 216, 164, 5, 5, 28, 213, 5, 5, 76, 208, 5, 5, 9, 207, 5, 5, 194, 181, 5, 5, 106, 206, 5, 5, 196, 180, 5, 5, 177, 237, 5, 5, 178, 237, 5, 5, 15, 197, 5, 5, 34, 199, 5, 5, 76, 245, 5, 5, 174, 228, 5, 5, 76, 185, 5, 5, 116, 204, 5, 5, 95, 167, 5, 5, 72, 190, 5, 5, 16, 168, 5, 5, 98, 229, 5, 5, 148, 174, 5, 5, 221, 238, 5, 5, 69, 222, 5, 5, 182, 167, 5, 5, 253, 210, 5, 5, 104, 210, 5, 5, 194, 163, 5, 5, 92, 177, 5, 5, 73, 190, 5, 5, 39, 212, 5, 5, 254, 204, 5, 5, 11, 207, 5, 5, 246, 181, 5, 5, 238, 240, 5, 5, 22, 249, 5, 5, 238, 208, 5, 5, 230, 186, 5, 5, 132, 241, 5, 5, 87, 222, 5, 5, 61, 197, 5, 5, 118, 178, 5, 5, 223, 196, 5, 5, 31, 229, 5, 5, 239, 208, 5, 5, 108, 206, 5, 5, 189, 240, 5, 5, 190, 240, 5, 5, 109, 206, 5, 5, 160, 231, 5, 5, 62, 185, 5, 5, 247, 178, 5, 5, 153, 208, 5, 5, 114, 226, 5, 5, 229, 192, 5, 5, 232, 206, 5, 5, 192, 207, 5, 5, 218, 240, 5, 5, 86, 233, 5, 5, 21, 202, 5, 5, 164, 229, 5, 5, 134, 244, 5, 5, 120, 223, 5, 5, 153, 171, 5, 5, 51, 230, 5, 5, 223, 238, 5, 5, 241, 209, 5, 5, 199, 180, 5, 5, 38, 199, 5, 5, 200, 180, 5, 5, 12, 238, 5, 5, 124, 203, 5, 5, 42, 170, 5, 5, 149, 174, 5, 5, 93, 226, 5, 5, 234, 181, 5, 5, 4, 243, 5, 5, 148, 197, 5, 5, 81, 249, 5, 5, 106, 168, 5, 5, 224, 196, 5, 5, 231, 169, 5, 5, 39, 199, 5, 5, 141, 244, 5, 5, 76, 226, 5, 5, 132, 245, 5, 5, 13, 207, 5, 5, 28, 179, 5, 5, 125, 234, 5, 5, 180, 173, 5, 5, 173, 197, 5, 5, 147, 244, 5, 5, 82, 249, 5, 5, 77, 215, 5, 5, 121, 247, 5, 5, 223, 226, 5, 5, 174, 197, 5, 5, 184, 225, 5, 5, 109, 180, 5, 5, 223, 213, 5, 5, 247, 230, 5, 5, 81, 188, 5, 5, 119, 202, 5, 5, 29, 193, 5, 5, 5, 205, 5, 5, 87, 237, 5, 5, 160, 198, 5, 5, 199, 233, 5, 5, 200, 233, 5, 5, 202, 240, 5, 5, 4, 242, 5, 5, 86, 243, 5, 5, 28, 202, 5, 5, 150, 187, 5, 5, 95, 181, 5, 5, 160, 195, 5, 5, 158, 180, 5, 5, 159, 180, 5, 5, 203, 240, 5, 5, 235, 181, 5, 5, 41, 169, 5, 5, 21, 204, 5, 5, 20, 192, 5, 5, 96, 181, 5, 5, 210, 224, 5, 5, 120, 202, 5, 5, 121, 182, 5, 5, 183, 219, 5, 5, 128, 238, 5, 5, 233, 241, 5, 5, 124, 239, 5, 5, 180, 238, 5, 5, 121, 223, 5, 5, 52, 219, 5, 5, 113, 188, 5, 5, 239, 249, 5, 5, 55, 191, 5, 5, 109, 165, 5, 5, 245, 229, 5, 5, 75, 222, 5, 5, 253, 247, 5, 5, 42, 245, 5, 5, 206, 207, 5, 5, 201, 243, 5, 5, 202, 180, 5, 5, 151, 235, 5, 5, 23, 204, 5, 5, 207, 240, 5, 5, 129, 199, 5, 5, 110, 165, 5, 5, 151, 244, 5, 5, 198, 236, 5, 5, 83, 201, 5, 5, 38, 173, 5, 5, 24, 204, 5, 5, 6, 188, 5, 5, 94, 243, 5, 5, 22, 192, 5, 5, 184, 204, 5, 5, 134, 215, 5, 5, 135, 215, 5, 5, 236, 181, 5, 5, 173, 169, 5, 5, 78, 233, 5, 5, 174, 248, 5, 5, 152, 244, 5, 5, 79, 190, 5, 5, 122, 182, 5, 5, 136, 215, 5, 5, 74, 217, 5, 5, 208, 240, 5, 5, 55, 168, 5, 5, 199, 225, 5, 5, 116, 190, 5, 5, 42, 213, 5, 5, 208, 239, 5, 5, 37, 202, 5, 5, 18, 232, 5, 5, 254, 247, 5, 5, 105, 220, 5, 5, 137, 241, 5, 5, 56, 244, 5, 5, 56, 168, 5, 5, 133, 184, 5, 5, 97, 199, 5, 5, 66, 171, 5, 5, 153, 244, 5, 5, 6, 171, 5, 5, 198, 176, 5, 5, 244, 248, 5, 5, 176, 248, 5, 5, 141, 185, 5, 5, 64, 164, 5, 5, 226, 189, 5, 5, 176, 197, 5, 5, 26, 191, 5, 5, 250, 229, 5, 5, 246, 248, 5, 5, 154, 244, 5, 5, 127, 194, 5, 5, 236, 211, 5, 5, 247, 186, 5, 5, 184, 190, 5, 5, 29, 184, 5, 5, 126, 204, 5, 5, 208, 201, 5, 5, 248, 248, 5, 5, 28, 172, 5, 5, 189, 226, 5, 5, 168, 238, 5, 5, 228, 240, 5, 5, 115, 234, 5, 5, 19, 164, 5, 5, 240, 220, 5, 5, 11, 193, 5, 5, 232, 178, 5, 5, 193, 189, 5, 5, 230, 240, 5, 5, 135, 214, 5, 5, 9, 244, 5, 5, 107, 189, 5, 5, 9, 223, 5, 5, 203, 211, 5, 5, 253, 248, 5, 5, 63, 217, 5, 5, 32, 182, 5, 5, 113, 235, 5, 5, 236, 218, 5, 5, 37, 239, 5, 5, 108, 193, 5, 5, 247, 214, 5, 5, 204, 198, 5, 5, 15, 220, 5, 5, 144, 191, 5, 5, 162, 190, 5, 5, 156, 201, 5, 5, 31, 245, 5, 5, 95, 163, 5, 5, 186, 186, 5, 5, 221, 207, 5, 5, 51, 193, 5, 5, 168, 237, 5, 5, 99, 248, 5, 5, 105, 215, 5, 5, 172, 180, 5, 5, 65, 245, 5, 5, 27, 212, 5, 5, 150, 171, 5, 5, 212, 238, 5, 5, 212, 196, 5, 5, 191, 163, 5, 5, 162, 167, 5, 5, 90, 210, 5, 5, 106, 226, 5, 5, 223, 192, 5, 5, 11, 249, 5, 5, 168, 236, 5, 5, 18, 199, 5, 5, 134, 197, 5, 5, 247, 237, 5, 5, 65, 193, 5, 5, 227, 209, 5, 5, 187, 233, 5, 5, 223, 169, 5, 5, 121, 245, 5, 5, 140, 240, 5, 5, 248, 241, 5, 5, 118, 247, 5, 5, 60, 188, 5, 5, 99, 202, 5, 5, 141, 240, 5, 5, 236, 204, 5, 5, 165, 197, 5, 5, 116, 239, 5, 5, 126, 244, 5, 5, 180, 219, 5, 5, 106, 188, 5, 5, 247, 187, 5, 5, 16, 204, 5, 5, 178, 204, 5, 5, 152, 248, 5, 5, 67, 190, 5, 5, 195, 200, 5, 5, 97, 220, 5, 5, 132, 244, 5, 5, 91, 199, 5, 5, 251, 170, 5, 5, 121, 194, 5, 5, 246, 229, 5, 5, 231, 189, 5, 5, 179, 205, 5, 5, 144, 189, 5, 5, 65, 205, 5, 5, 131, 215, 5, 5, 190, 234, 5, 5, 71, 205, 5, 5, 74, 205, 5, 5, 208, 242, 5, 5, 188, 234, 5, 5, 18, 190, 5, 5, 16, 242, 5, 5, 146, 234, 5, 5, 138, 202, 5, 5, 23, 167, 5, 5, 147, 234, 5, 5, 24, 167, 5, 5, 251, 221, 5, 5, 36, 208, 5, 5, 201, 221, 5, 5, 246, 204, 5, 5, 49, 171, 5, 5, 50, 171, 5, 5, 44, 248, 5, 5, 74, 188, 5, 5, 208, 224, 5, 5, 39, 208, 5, 5, 250, 188, 5, 5, 177, 183, 5, 5, 240, 208, 5, 5, 74, 190, 5, 5, 182, 204, 5, 5, 40, 208, 5, 5, 110, 206, 5, 5, 132, 187, 5, 5, 60, 216, 5, 5, 132, 215, 5, 5, 39, 203, 5, 5, 56, 171, 5, 5, 151, 187, 5, 5, 212, 227, 5, 5, 32, 221, 5, 5, 4, 189, 5, 5, 113, 206, 5, 5, 248, 208, 5, 5, 165, 189, 5, 5, 210, 242, 5, 5, 148, 234, 5, 5, 182, 169, 5, 5, 178, 169, 5, 5, 135, 197, 5, 5, 3, 224, 5, 5, 51, 174, 5, 5, 219, 171, 5, 5, 213, 171, 5, 5, 53, 174, 5, 5, 32, 167, 5, 5, 2, 212, 5, 5, 189, 211, 5, 5, 100, 170, 5, 5, 54, 174, 5, 5, 232, 215, 5, 5, 63, 185, 5, 5, 192, 211, 5, 5, 221, 171, 5, 5, 123, 223, 5, 5, 234, 173, 5, 5, 38, 174, 5, 5, 185, 211, 5, 5, 220, 171, 5, 5, 23, 169, 5, 5, 159, 169, 5, 5, 71, 169, 5, 5, 32, 200, 5, 5, 157, 169, 5, 5, 111, 232, 5, 5, 186, 184, 5, 5, 124, 173, 5, 5, 49, 201, 5, 5, 50, 201, 5, 5, 180, 205, 5, 5, 170, 194, 5, 5, 69, 181, 5, 5, 222, 179, 5, 5, 178, 201, 5, 5, 155, 199, 5, 5, 39, 197, 5, 5, 180, 201, 5, 5, 110, 229, 5, 5, 208, 186, 5, 5, 70, 184, 5, 5, 29, 223, 5, 5, 207, 197, 5, 5, 206, 197, 5, 5, 142, 240, 5, 5, 142, 210, 5, 5, 17, 176, 5, 5, 150, 213, 5, 5, 136, 182, 5, 5, 142, 169, 5, 5, 135, 169, 5, 5, 143, 210, 5, 5, 115, 175, 5, 5, 235, 225, 5, 5, 90, 242, 5, 5, 157, 177, 5, 5, 122, 239, 5, 5, 221, 213, 5, 5, 254, 242, 5, 5, 193, 211, 5, 5, 249, 169, 5, 5, 5, 243, 5, 5, 124, 211, 5, 5, 238, 248, 5, 5, 231, 186, 5, 5, 87, 176, 5, 5, 11, 178, 5, 5, 90, 234, 5, 5, 154, 242, 5, 5, 203, 242, 5, 5, 9, 243, 5, 5, 19, 176, 5, 5, 254, 238, 5, 5, 232, 220, 5, 5, 120, 175, 5, 5, 91, 234, 5, 5, 125, 239, 5, 5, 233, 169, 5, 5, 131, 220, 5, 5, 19, 243, 5, 5, 20, 178, 5, 5, 227, 189, 5, 5, 174, 218, 5, 5, 126, 173, 5, 5, 146, 174, 5, 5, 182, 174, 5, 5, 210, 171, 5, 5, 194, 171, 5, 5, 76, 249, 5, 5, 168, 234, 5, 5, 4, 212, 5, 5, 250, 223, 5, 5, 149, 245, 5, 5, 163, 221, 5, 5, 97, 218, 5, 5, 5, 171, 5, 5, 129, 238, 5, 5, 121, 165, 5, 5, 52, 182, 5, 5, 111, 165, 5, 5, 50, 249, 5, 5, 241, 223, 5, 5, 50, 182, 5, 5, 204, 177, 5, 5, 115, 171, 5, 5, 29, 183, 5, 5, 206, 177, 5, 5, 153, 231, 5, 5, 57, 193, 5, 5, 58, 193, 5, 5, 245, 178, 5, 5, 119, 173, 5, 5, 200, 179, 5, 5, 24, 249, 5, 5, 19, 168, 5, 5, 255, 224, 5, 5, 21, 192, 5, 5, 64, 209, 5, 5, 25, 249, 5, 5, 66, 180, 5, 5, 131, 180, 5, 5, 114, 229, 5, 5, 107, 229, 5, 5, 117, 173, 5, 5, 219, 172, 5, 5, 45, 247, 5, 5, 10, 163, 5, 5, 90, 177, 5, 5, 48, 182, 5, 5, 220, 222, 5, 5, 220, 211, 5, 5, 30, 229, 5, 5, 5, 227, 5, 5, 118, 242, 5, 5, 7, 227, 5, 5, 162, 181, 5, 5, 41, 245, 5, 5, 95, 184, 5, 5, 111, 206, 5, 5, 16, 201, 5, 5, 15, 207, 5, 5, 12, 243, 5, 5, 195, 181, 5, 5, 152, 235, 5, 5, 40, 213, 5, 5, 237, 215, 5, 5, 15, 243, 5, 5, 115, 185, 5, 5, 219, 164, 5, 5, 229, 241, 5, 5, 43, 210, 5, 5, 72, 198, 5, 5, 127, 245, 5, 5, 130, 245, 5, 5, 24, 198, 5, 5, 116, 214, 5, 5, 14, 238, 5, 5, 79, 219, 5, 5, 31, 215, 5, 5, 137, 247, 5, 5, 80, 219, 5, 5, 211, 184, 5, 5, 141, 183, 5, 5, 218, 208, 5, 5, 192, 219, 5, 5, 231, 202, 5, 5, 161, 231, 5, 5, 29, 202, 5, 5, 35, 202, 5, 5, 155, 208, 5, 5, 252, 201, 5, 5, 77, 223, 5, 5, 227, 224, 5, 5, 229, 224, 5, 5, 146, 197, 5, 5, 5, 241, 5, 5, 200, 243, 5, 5, 113, 214, 5, 5, 7, 164, 5, 5, 15, 232, 5, 5, 114, 214, 5, 5, 215, 223, 5, 5, 26, 223, 5, 5, 6, 243, 5, 5, 34, 207, 5, 5, 41, 181, 5, 5, 156, 189, 5, 5, 152, 189, 5, 5, 153, 225, 5, 5, 93, 212, 5, 5, 115, 214, 5, 5, 205, 195, 5, 5, 15, 184, 5, 5, 183, 204, 5, 5, 232, 241, 5, 5, 17, 226, 5, 5, 239, 196, 5, 5, 46, 248, 5, 5, 198, 225, 5, 5, 249, 239, 5, 5, 178, 183, 5, 5, 17, 232, 5, 5, 113, 232, 5, 5, 244, 211, 5, 5, 177, 249, 5, 5, 18, 243, 5, 5, 183, 238, 5, 5, 16, 238, 5, 5, 213, 245, 5, 5, 211, 240, 5, 5, 166, 233, 5, 5, 114, 232, 5, 5, 144, 177, 5, 5, 200, 225, 5, 5, 3, 206, 5, 5, 46, 241, 5, 5, 62, 223, 5, 5, 223, 224, 5, 5, 116, 197, 5, 5, 184, 243, 5, 5, 28, 207, 5, 5, 132, 189, 5, 5, 2, 164, 5, 5, 30, 181, 5, 5, 144, 231, 5, 5, 201, 195, 5, 5, 37, 248, 5, 5, 11, 226, 5, 5, 243, 211, 5, 5, 209, 245, 5, 5, 248, 190, 5, 5, 125, 167, 5, 5, 214, 194, 5, 5, 126, 167, 5, 5, 7, 243, 5, 5, 14, 191, 5, 5, 39, 191, 5, 5, 215, 194, 5, 5, 168, 195, 5, 5, 44, 175, 5, 5, 167, 189, 5, 5, 45, 175, 5, 5, 225, 190, 5, 5, 202, 194, 5, 5, 164, 195, 5, 5, 241, 193, 5, 5, 243, 209, 5, 5, 123, 165, 5, 5, 210, 193, 5, 5, 233, 248, 5, 5, 44, 226, 5, 5, 151, 197, 5, 5, 209, 207, 5, 5, 44, 214, 5, 5, 177, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 209, 5, 5, 246, 192, 5, 5, 174, 223, 5, 5, 80, 203, 5, 5, 117, 199, 5, 5, 71, 226, 5, 5, 37, 207, 5, 5, 241, 193, 5, 5, 241, 193, 5, 5, 36, 209, 5, 5, 164, 205, 5, 5, 239, 184, 5, 5, 22, 183, 5, 5, 78, 186, 5, 5, 83, 185, 5, 5, 102, 190, 5, 5, 113, 190, 5, 5, 101, 190, 5, 5, 123, 190, 5, 5, 114, 190, 5, 5, 26, 185, 5, 5, 134, 190, 5, 5, 208, 185, 5, 5, 138, 190, 5, 5, 144, 190, 5, 5, 210, 185, 5, 5, 150, 190, 5, 5, 188, 190, 5, 5, 185, 190, 5, 5, 50, 186, 5, 5, 96, 186, 5, 5, 52, 186, 5, 5, 184, 190, 5, 5, 22, 186, 5, 5, 94, 186, 5, 5, 39, 186, 5, 5, 48, 186, 5, 5, 227, 184, 5, 5, 253, 184, 5, 5, 4, 185, 5, 5, 114, 186, 5, 5, 129, 186, 5, 5, 141, 186, 5, 5, 106, 186, 5, 5, 104, 186, 5, 5, 35, 185, 5, 5, 159, 186, 5, 5, 175, 185, 5, 5, 241, 189, 5, 5, 248, 189, 5, 5, 211, 189, 5, 5, 199, 189, 5, 5, 193, 185, 5, 5, 214, 189, 5, 5, 234, 189, 5, 5, 35, 190, 5, 5, 76, 190, 5, 5, 240, 189, 5, 5, 79, 190, 5, 5, 31, 190, 5, 5, 32, 190, 5, 5, 96, 191, 5, 5, 24, 190, 5, 5, 56, 190, 5, 5, 18, 190, 5, 5, 223, 190, 5, 5, 34, 191, 5, 5, 208, 184, 5, 5, 15, 191, 5, 5, 16, 191, 5, 5, 169, 185, 5, 5, 125, 185, 5, 5, 34, 190, 5, 5, 87, 185, 5, 5, 131, 185, 5, 5, 74, 191, 5, 5, 239, 185, 5, 5, 146, 185, 5, 5, 10, 186, 5, 5, 147, 185, 5, 5, 82, 191, 5, 5, 8, 186, 5, 5, 86, 185, 5, 5, 10, 185, 5, 5, 10, 189, 5, 5, 100, 189, 5, 5, 155, 186, 5, 5, 134, 189, 5, 5, 124, 189, 5, 5, 123, 189, 5, 5, 12, 178, 5, 5, 223, 182, 5, 5, 26, 185, 5, 5, 172, 184, 5, 5, 225, 175, 5, 5, 120, 184, 5, 5, 151, 184, 5, 5, 89, 191, 5, 5, 74, 240, 5, 5, 195, 163, 5, 5, 69, 167, 5, 5, 238, 165, 5, 5, 216, 174, 5, 5, 118, 166, 5, 5, 90, 171, 5, 5, 156, 229, 5, 5, 224, 236, 5, 5, 62, 234, 5, 5, 2, 236, 5, 5, 36, 229, 5, 5, 101, 241, 5, 5, 177, 229, 5, 5, 131, 227, 5, 5, 183, 224, 5, 5, 181, 224, 5, 5, 238, 226, 5, 5, 234, 230, 5, 5, 216, 237, 5, 5, 218, 237, 5, 5, 68, 189, 5, 5, 58, 189, 5, 5, 42, 189, 5, 5, 43, 189, 5, 5, 54, 189, 5, 5, 40, 189, 5, 5, 78, 189, 5, 5, 76, 189, 5, 5, 111, 187, 5, 5, 58, 191, 5, 5, 212, 184, 5, 5, 203, 189, 5, 5, 65, 191, 5, 5, 103, 191, 5, 5, 139, 187, 5, 5, 53, 191, 5, 5, 245, 186, 5, 5, 132, 187, 5, 5, 208, 186, 5, 5, 22, 187, 5, 5, 104, 187, 5, 5, 105, 187, 5, 5, 157, 187, 5, 5, 63, 184, 5, 5, 140, 188, 5, 5, 212, 188, 5, 5, 79, 184, 5, 5, 132, 188, 5, 5, 198, 188, 5, 5, 175, 188, 5, 5, 64, 184, 5, 5, 202, 188, 5, 5, 153, 188, 5, 5, 82, 184, 5, 5, 144, 188, 5, 5, 123, 188, 5, 5, 206, 188, 5, 5, 176, 187, 5, 5, 177, 187, 5, 5, 25, 243, 5, 5, 191, 187, 5, 5, 199, 187, 5, 5, 177, 229, 5, 5, 126, 188, 5, 5, 90, 184, 5, 5, 73, 184, 5, 5, 205, 188, 5, 5, 162, 188, 5, 5, 210, 187, 5, 5, 174, 189, 5, 5, 87, 189, 5, 5, 120, 184, 5, 5, 173, 189, 5, 5, 117, 188, 5, 5, 98, 189, 5, 5, 178, 244, 5, 5, 117, 189, 5, 5, 119, 189, 5, 5, 129, 189, 5, 5, 131, 189, 5, 5, 160, 189, 5, 5, 172, 189, 5, 5, 35, 187, 5, 5, 14, 187, 5, 5, 17, 187, 5, 5, 114, 187, 5, 5, 226, 237, 5, 5, 31, 185, 5, 5, 226, 187, 5, 5, 234, 187, 5, 5, 42, 184, 5, 5, 21, 188, 5, 5, 26, 185, 5, 5, 241, 187, 5, 5, 246, 187, 5, 5, 13, 188, 5, 5, 244, 187, 5, 5, 248, 190, 5, 5, 113, 249, 5, 5, 25, 231, 5, 5, 54, 188, 5, 5, 55, 224, 5, 5, 84, 188, 5, 5, 34, 188, 5, 5, 28, 188, 5, 5, 36, 188, 5, 5, 37, 188, 5, 5, 40, 188, 5, 5, 54, 184, 5, 5, 157, 185, 5, 5, 99, 188, 5, 5, 43, 190, 5, 5, 17, 190, 5, 5, 198, 190, 5, 5, 202, 190, 5, 5, 204, 190, 5, 5, 214, 190, 5, 5, 87, 191, 5, 5, 89, 187, 5, 5, 57, 187, 5, 5, 89, 191, 5, 5, 240, 190, 5, 5, 30, 187, 5, 5, 27, 187, 5, 5, 76, 191, 5, 5, 29, 240, 5, 5, 249, 186, 5, 5, 173, 186, 5, 5, 150, 183, 5, 5, 3, 187, 5, 5, 82, 187, 5, 5, 209, 186, 5, 5, 7, 187, 5, 5, 8, 187, 5, 5, 250, 186, 5, 5, 222, 186, 5, 5, 208, 183, 5, 5, 212, 183, 5, 5, 21, 189, 5, 5, 240, 188, 5, 5, 241, 188, 5, 5, 35, 189, 5, 5, 237, 188, 5, 5, 3, 189, 5, 5, 4, 189, 5, 5, 218, 188, 5, 5, 225, 188, 5, 5, 247, 188, 5, 5, 26, 187, 5, 5, 70, 187, 5, 5, 71, 187, 5, 5, 186, 222, 5, 5, 192, 218, 5, 5, 6, 227, 5, 5, 177, 228, 5, 5, 146, 223, 5, 5, 212, 233, 5, 5, 94, 209, 5, 5, 36, 178, 5, 5, 53, 179, 5, 5, 95, 180, 5, 5, 194, 219, 5, 5, 12, 179, 5, 5, 54, 164, 5, 5, 141, 174, 5, 5, 139, 216, 5, 5, 56, 206, 5, 5, 78, 205, 5, 5, 62, 196, 5, 5, 78, 245, 5, 5, 75, 199, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 222, 192, 0, 0, 0, 5, 5, 114, 211, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 212, 5, 5, 140, 221, 5, 5, 65, 240, 5, 5, 248, 186, 5, 5, 179, 228, 5, 5, 71, 216, 5, 5, 131, 174, 5, 5, 172, 206, 5, 5, 103, 206, 5, 5, 6, 248, 192, 0, 0, 0, 5, 5, 49, 210, 192, 0, 0, 0, 5, 5, 154, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 240, 5, 5, 184, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 173, 5, 5, 189, 235, 5, 5, 93, 194, 5, 5, 174, 197, 5, 5, 107, 186, 5, 5, 108, 187, 5, 5, 55, 245, 5, 5, 126, 236, 5, 5, 192, 171, 5, 5, 195, 171, 5, 5, 233, 210, 5, 5, 181, 163, 5, 5, 79, 197, 5, 5, 44, 180, 5, 5, 56, 209, 5, 5, 210, 168, 5, 5, 139, 169, 5, 5, 137, 234, 5, 5, 185, 223, 5, 5, 56, 200, 5, 5, 91, 195, 5, 5, 196, 232, 5, 5, 134, 225, 5, 5, 5, 172, 5, 5, 164, 202, 5, 5, 109, 229, 5, 5, 201, 169, 5, 5, 188, 197, 5, 5, 194, 221, 5, 5, 140, 198, 5, 5, 196, 221, 5, 5, 251, 219, 5, 5, 246, 232, 5, 5, 189, 163, 5, 5, 188, 227, 5, 5, 163, 218, 5, 5, 172, 208, 5, 5, 109, 242, 5, 5, 235, 232, 5, 5, 226, 221, 5, 5, 205, 199, 5, 5, 193, 220, 5, 5, 64, 193, 5, 5, 184, 181, 5, 5, 170, 203, 5, 5, 202, 188, 5, 5, 246, 205, 5, 5, 26, 173, 5, 5, 112, 229, 5, 5, 164, 219, 5, 5, 63, 224, 5, 5, 55, 234, 5, 5, 55, 234, 5, 5, 185, 219, 5, 5, 164, 197, 5, 5, 66, 227, 5, 5, 117, 241, 5, 5, 205, 205, 5, 5, 21, 166, 5, 5, 211, 232, 5, 5, 249, 225, 5, 5, 100, 240, 5, 5, 109, 183, 5, 5, 90, 216, 5, 5, 142, 168, 5, 5, 76, 200, 192, 0, 0, 0, 5, 5, 91, 194, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 83, 166, 5, 5, 23, 197, 5, 5, 52, 212, 5, 5, 142, 216, 5, 5, 147, 226, 5, 5, 216, 202, 5, 5, 252, 249, 5, 5, 244, 227, 5, 5, 79, 197, 5, 5, 180, 179, 5, 5, 83, 200, 5, 5, 159, 245, 5, 5, 230, 222, 5, 5, 108, 173, 5, 5, 216, 242, 5, 5, 115, 164, 5, 5, 20, 165, 5, 5, 193, 233, 5, 5, 107, 238, 5, 5, 121, 240, 5, 5, 20, 234, 5, 5, 176, 241, 5, 5, 94, 247, 5, 5, 216, 228, 5, 5, 100, 248, 5, 5, 196, 232, 5, 5, 151, 238, 5, 5, 134, 225, 5, 5, 112, 175, 5, 5, 197, 247, 5, 5, 45, 236, 5, 5, 73, 166, 5, 5, 102, 238, 5, 5, 114, 211, 5, 5, 129, 186, 5, 5, 111, 247, 5, 5, 44, 221, 5, 5, 73, 175, 5, 5, 131, 227, 5, 5, 34, 188, 5, 5, 185, 243, 5, 5, 127, 231, 5, 5, 140, 198, 5, 5, 184, 206, 5, 5, 196, 221, 5, 5, 177, 209, 5, 5, 183, 207, 5, 5, 46, 173, 5, 5, 140, 221, 5, 5, 251, 220, 5, 5, 183, 233, 5, 5, 136, 199, 5, 5, 75, 194, 5, 5, 8, 247, 5, 5, 65, 240, 5, 5, 44, 229, 5, 5, 102, 218, 5, 5, 251, 207, 5, 5, 184, 219, 5, 5, 91, 181, 5, 5, 51, 181, 5, 5, 170, 203, 5, 5, 144, 185, 5, 5, 159, 179, 5, 5, 202, 188, 5, 5, 191, 168, 5, 5, 164, 219, 5, 5, 33, 201, 5, 5, 115, 199, 5, 5, 121, 249, 5, 5, 79, 211, 5, 5, 247, 174, 5, 5, 66, 227, 5, 5, 111, 177, 5, 5, 154, 221, 5, 5, 133, 211, 5, 5, 117, 241, 5, 5, 172, 184, 5, 5, 139, 248, 5, 5, 205, 205, 5, 5, 2, 166, 5, 5, 211, 232, 5, 5, 93, 229, 5, 5, 54, 223, 5, 5, 59, 236, 5, 5, 155, 216, 5, 5, 195, 221, 5, 5, 109, 183, 5, 5, 172, 206, 5, 5, 178, 163, 5, 5, 90, 216, 5, 5, 184, 237, 5, 5, 142, 168, 5, 5, 231, 220, 5, 5, 241, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 238, 5, 5, 125, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 201, 32, 230, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 14, 228, 6, 5, 188, 231, 6, 5, 236, 229, 192, 0, 0, 0, 6, 5, 67, 239, 6, 5, 49, 178, 6, 5, 199, 203, 6, 5, 61, 181, 6, 5, 236, 203, 6, 5, 223, 244, 6, 5, 28, 245, 6, 5, 21, 188, 6, 5, 119, 210, 6, 5, 49, 198, 6, 5, 42, 232, 6, 5, 4, 216, 6, 5, 149, 225, 6, 5, 211, 222, 6, 5, 8, 229, 6, 5, 61, 173, 6, 5, 10, 229, 6, 5, 26, 226, 6, 5, 242, 242, 6, 5, 212, 179, 6, 5, 114, 166, 6, 5, 82, 239, 6, 5, 79, 236, 6, 5, 56, 242, 6, 5, 251, 232, 6, 5, 193, 222, 6, 5, 94, 242, 6, 5, 159, 218, 6, 5, 116, 232, 6, 5, 46, 175, 6, 5, 229, 205, 6, 5, 51, 197, 6, 5, 84, 197, 6, 5, 127, 170, 6, 5, 68, 242, 6, 5, 197, 248, 6, 5, 133, 228, 6, 5, 145, 191, 6, 5, 160, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 197, 195, 62, 0, 197, 35, 63, 0, 197, 131, 63, 0, 197, 227, 63, 0, 197, 67, 64, 0, 197, 163, 64, 0, 197, 3, 65, 0, 197, 99, 65, 0, 197, 195, 65, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 6, 5, 58, 175, 6, 5, 37, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 239, 5, 5, 250, 232, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 243, 192, 0, 0, 0, 5, 5, 129, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 44, 173, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 218, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 251, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 177, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 255, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 181, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 207, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 219, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 209, 192, 0, 0, 0, 5, 5, 190, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 79, 175, 5, 5, 219, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 120, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 93, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 69, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 100, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 198, 5, 5, 57, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 107, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 171, 192, 0, 0, 0, 5, 5, 195, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 125, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 91, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 101, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 77, 172, 192, 0, 0, 0, 5, 5, 144, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 165, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 245, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 216, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 176, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 195, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 115, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 11, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 70, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 26, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 4, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 132, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 147, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 72, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 90, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 27, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 212, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 224, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 195, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 62, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 89, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 229, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 51, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 225, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 71, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 238, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 63, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 98, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 54, 215, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 209, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 53, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 218, 192, 0, 0, 0, 5, 5, 202, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 17, 199, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 24, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 65, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 75, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 127, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 111, 166, 192, 0, 0, 0, 5, 5, 24, 216, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 169, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 106, 205, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 183, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 177, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 230, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 189, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 7, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 237, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 45, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 143, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 166, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 103, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 41, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 199, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 238, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 240, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 249, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 138, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 2, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 117, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 244, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 15, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 113, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 52, 240, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 217, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 206, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 14, 175, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 228, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 59, 200, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 116, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 18, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 13, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 112, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 192, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 202, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 250, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 182, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 134, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 203, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 34, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 95, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 180, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 211, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 78, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 179, 213, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 200, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 214, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 55, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 154, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 16, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 119, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 228, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 243, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 214, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 146, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 130, 184, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 174, 193, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 227, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 76, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 148, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 188, 198, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 58, 185, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 198, 229, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 186, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 135, 225, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 108, 180, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 252, 227, 192, 0, 0, 0, 5, 5, 254, 167, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 168, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 122, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 185, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 246, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 221, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 32, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 222, 5, 5, 19, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 241, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 128, 220, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 210, 207, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 37, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 129, 204, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 144, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 19, 211, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 60, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 243, 5, 5, 105, 237, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 155, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 223, 222, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 150, 235, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 163, 206, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 159, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 121, 201, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 118, 163, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 232, 248, 192, 0, 0, 0, 5, 5, 199, 239, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 233, 217, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 226, 197, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 67, 188, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 201, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 174, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 157, 223, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 66, 190, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 153, 220, 5, 5, 15, 247, 5, 5, 241, 202, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 187, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 171, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 161, 233, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 166, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 193, 219, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 42, 178, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 109, 187, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 36, 203, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 212, 170, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 152, 191, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 220, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 254, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 226, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 74, 195, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 94, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 189, 169, 5, 5, 23, 172, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 145, 196, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 194, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 178, 181, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 227, 208, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 231, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 88, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 6, 241, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 96, 243, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 183, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 29, 187, 5, 5, 184, 246, 5, 5, 126, 239, 192, 0, 0, 0, 5, 5, 181, 183, 5, 5, 55, 245, 5, 5, 94, 182, 5, 5, 87, 166, 5, 5, 221, 231, 5, 5, 221, 163, 5, 5, 126, 236, 5, 5, 106, 216, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 192, 171, 5, 5, 243, 181, 5, 5, 81, 201, 5, 5, 42, 207, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 40, 183, 5, 5, 42, 232, 192, 0, 0, 0, 5, 5, 95, 231, 5, 5, 250, 248, 5, 5, 51, 166, 5, 5, 223, 178, 5, 5, 23, 197, 192, 0, 0, 0, 5, 5, 149, 210, 5, 5, 105, 230, 5, 5, 235, 167, 5, 5, 50, 195, 5, 5, 120, 190, 5, 5, 145, 191, 5, 5, 134, 224, 192, 0, 0, 0, 5, 5, 252, 249, 5, 5, 195, 171, 5, 5, 233, 210, 5, 5, 244, 227, 5, 5, 249, 163, 5, 5, 37, 235, 5, 5, 195, 163, 5, 5, 64, 200, 5, 5, 181, 163, 5, 5, 92, 163, 5, 5, 44, 202, 5, 5, 55, 202, 5, 5, 94, 211, 5, 5, 94, 211, 5, 5, 94, 211, 192, 0, 0, 0, 5, 5, 252, 199, 5, 5, 40, 202, 5, 5, 66, 236, 192, 0, 0, 0, 5, 5, 82, 204, 5, 5, 80, 223, 5, 5, 141, 241, 5, 5, 186, 215, 5, 5, 171, 212, 5, 5, 88, 225, 5, 5, 3, 220, 5, 5, 199, 237, 5, 5, 127, 171, 5, 5, 68, 180, 5, 5, 9, 209, 5, 5, 92, 215, 5, 5, 96, 228, 5, 5, 96, 228, 5, 5, 83, 200, 5, 5, 12, 223, 5, 5, 68, 219, 5, 5, 174, 241, 5, 5, 222, 181, 5, 5, 44, 180, 5, 5, 223, 181, 5, 5, 223, 197, 5, 5, 97, 167, 5, 5, 94, 209, 5, 5, 183, 222, 5, 5, 93, 225, 5, 5, 108, 218, 5, 5, 245, 181, 5, 5, 144, 216, 5, 5, 172, 233, 5, 5, 49, 164, 5, 5, 7, 177, 192, 0, 0, 0, 5, 5, 176, 169, 5, 5, 255, 198, 5, 5, 236, 173, 5, 5, 49, 178, 5, 5, 24, 171, 5, 5, 168, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 215, 201, 5, 5, 177, 247, 5, 5, 95, 182, 5, 5, 128, 168, 5, 5, 210, 174, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 21, 230, 5, 5, 77, 186, 5, 5, 77, 186, 192, 0, 0, 0, 5, 5, 254, 248, 5, 5, 247, 218, 5, 5, 120, 184, 5, 5, 35, 164, 192, 0, 0, 0, 5, 5, 19, 228, 5, 5, 9, 206, 5, 5, 68, 176, 5, 5, 26, 242, 192, 0, 0, 0, 5, 5, 205, 181, 5, 5, 185, 223, 5, 5, 14, 215, 5, 5, 59, 210, 192, 0, 0, 0, 5, 5, 227, 242, 192, 0, 0, 0, 5, 5, 129, 177, 5, 5, 124, 231, 5, 5, 136, 224, 5, 5, 214, 217, 5, 5, 238, 223, 192, 0, 0, 0, 5, 5, 14, 218, 5, 5, 209, 229, 5, 5, 68, 170, 5, 5, 109, 173, 5, 5, 150, 170, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 86, 166, 5, 5, 21, 165, 5, 5, 141, 229, 5, 5, 114, 186, 192, 0, 0, 0, 5, 5, 216, 194, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 247, 5, 5, 149, 179, 5, 5, 149, 179, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 140, 216, 5, 5, 79, 177, 192, 0, 0, 0, 5, 5, 239, 178, 5, 5, 97, 230, 5, 5, 180, 218, 5, 5, 9, 197, 5, 5, 251, 248, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 56, 200, 192, 0, 0, 0, 5, 5, 183, 178, 5, 5, 182, 233, 5, 5, 36, 201, 5, 5, 216, 228, 5, 5, 36, 201, 5, 5, 237, 185, 5, 5, 196, 232, 5, 5, 223, 215, 5, 5, 136, 173, 5, 5, 87, 234, 5, 5, 239, 170, 5, 5, 134, 225, 5, 5, 78, 186, 5, 5, 86, 225, 5, 5, 48, 203, 5, 5, 3, 196, 5, 5, 43, 164, 5, 5, 0, 159, 5, 5, 216, 207, 192, 0, 0, 0, 5, 5, 204, 246, 5, 5, 129, 168, 5, 5, 184, 227, 5, 5, 31, 236, 5, 5, 60, 202, 192, 0, 0, 0, 5, 5, 227, 205, 5, 5, 182, 242, 5, 5, 224, 242, 192, 0, 0, 0, 5, 5, 72, 169, 5, 5, 63, 206, 5, 5, 27, 200, 5, 5, 52, 204, 192, 0, 0, 0, 5, 5, 5, 172, 5, 5, 169, 206, 192, 0, 0, 0, 5, 5, 184, 202, 5, 5, 71, 229, 5, 5, 216, 205, 5, 5, 100, 246, 5, 5, 109, 229, 5, 5, 2, 247, 192, 0, 0, 0, 5, 5, 59, 170, 5, 5, 200, 171, 5, 5, 39, 233, 5, 5, 235, 205, 5, 5, 237, 182, 5, 5, 24, 180, 5, 5, 129, 186, 5, 5, 111, 247, 5, 5, 98, 233, 5, 5, 2, 209, 5, 5, 139, 165, 192, 0, 0, 0, 5, 5, 109, 199, 5, 5, 199, 193, 5, 5, 209, 201, 5, 5, 114, 236, 5, 5, 201, 169, 192, 0, 0, 0, 5, 5, 99, 238, 5, 5, 118, 164, 5, 5, 125, 231, 192, 0, 0, 0, 5, 5, 71, 219, 5, 5, 183, 241, 5, 5, 251, 191, 5, 5, 176, 227, 192, 0, 0, 0, 5, 5, 33, 219, 5, 5, 206, 239, 5, 5, 209, 233, 192, 0, 0, 0, 5, 5, 213, 216, 192, 0, 0, 0, 5, 5, 28, 237, 5, 5, 6, 247, 5, 5, 131, 227, 5, 5, 196, 209, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 48, 173, 192, 0, 0, 0, 5, 5, 162, 242, 5, 5, 116, 164, 5, 5, 61, 210, 5, 5, 21, 245, 5, 5, 221, 166, 5, 5, 188, 197, 5, 5, 34, 188, 5, 5, 251, 197, 5, 5, 218, 205, 5, 5, 231, 183, 192, 0, 0, 0, 5, 5, 176, 168, 5, 5, 206, 192, 5, 5, 137, 242, 5, 5, 31, 191, 5, 5, 127, 231, 5, 5, 130, 177, 192, 0, 0, 0, 5, 5, 132, 242, 5, 5, 246, 223, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 110, 173, 5, 5, 236, 248, 5, 5, 184, 206, 5, 5, 195, 244, 192, 0, 0, 0, 5, 5, 147, 210, 5, 5, 29, 232, 5, 5, 137, 222, 5, 5, 38, 180, 192, 0, 0, 0, 5, 5, 169, 178, 192, 0, 0, 0, 5, 5, 56, 235, 192, 0, 0, 0, 5, 5, 22, 235, 5, 5, 183, 207, 5, 5, 87, 219, 192, 0, 0, 0, 5, 5, 220, 212, 5, 5, 233, 163, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 50, 179, 5, 5, 81, 247, 192, 0, 0, 0, 5, 5, 205, 248, 5, 5, 181, 168, 5, 5, 181, 168, 5, 5, 104, 175, 5, 5, 213, 247, 5, 5, 251, 220, 5, 5, 233, 236, 5, 5, 158, 212, 192, 0, 0, 0, 5, 5, 251, 230, 192, 0, 0, 0, 5, 5, 31, 232, 192, 0, 0, 0, 5, 5, 74, 240, 192, 0, 0, 0, 5, 5, 29, 248, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 102, 218, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 169, 220, 5, 5, 170, 220, 5, 5, 170, 220, 5, 5, 251, 207, 5, 5, 159, 203, 5, 5, 173, 224, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 151, 216, 5, 5, 31, 190, 5, 5, 91, 181, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 235, 232, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 131, 174, 5, 5, 99, 229, 192, 0, 0, 0, 5, 5, 64, 193, 5, 5, 203, 202, 5, 5, 57, 247, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 164, 168, 5, 5, 200, 168, 192, 0, 0, 0, 5, 5, 146, 222, 5, 5, 177, 221, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 242, 163, 192, 0, 0, 0, 5, 5, 69, 206, 5, 5, 94, 236, 5, 5, 167, 202, 192, 0, 0, 0, 5, 5, 179, 164, 5, 5, 151, 241, 5, 5, 141, 231, 5, 5, 137, 233, 5, 5, 23, 236, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 248, 223, 192, 0, 0, 0, 5, 5, 83, 244, 5, 5, 122, 238, 5, 5, 164, 219, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 162, 168, 192, 0, 0, 0, 5, 5, 67, 235, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 57, 248, 5, 5, 236, 234, 192, 0, 0, 0, 5, 5, 5, 168, 5, 5, 225, 244, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 182, 247, 5, 5, 131, 213, 5, 5, 181, 233, 192, 0, 0, 0, 5, 5, 5, 209, 5, 5, 54, 248, 5, 5, 67, 218, 5, 5, 167, 185, 5, 5, 101, 199, 5, 5, 152, 173, 5, 5, 246, 240, 5, 5, 21, 196, 192, 0, 0, 0, 5, 5, 234, 230, 5, 5, 216, 223, 5, 5, 57, 231, 5, 5, 155, 203, 5, 5, 26, 245, 5, 5, 213, 170, 5, 5, 216, 176, 5, 5, 185, 219, 5, 5, 183, 168, 5, 5, 248, 206, 5, 5, 27, 208, 5, 5, 28, 234, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 196, 168, 5, 5, 202, 205, 5, 5, 187, 234, 192, 0, 0, 0, 5, 5, 253, 230, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 222, 184, 5, 5, 234, 189, 5, 5, 109, 196, 5, 5, 154, 213, 5, 5, 8, 223, 5, 5, 81, 210, 5, 5, 10, 249, 5, 5, 193, 168, 5, 5, 121, 249, 5, 5, 53, 177, 5, 5, 150, 221, 5, 5, 222, 233, 192, 0, 0, 0, 5, 5, 26, 194, 5, 5, 88, 216, 192, 0, 0, 0, 5, 5, 161, 222, 5, 5, 87, 239, 192, 0, 0, 0, 5, 5, 41, 188, 5, 5, 216, 229, 192, 0, 0, 0, 5, 5, 243, 242, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 124, 225, 5, 5, 139, 248, 5, 5, 2, 166, 5, 5, 12, 227, 192, 0, 0, 0, 5, 5, 103, 194, 5, 5, 39, 165, 5, 5, 167, 192, 5, 5, 8, 209, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 5, 163, 5, 5, 44, 205, 5, 5, 113, 168, 192, 0, 0, 0, 5, 5, 133, 230, 5, 5, 93, 229, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 253, 208, 5, 5, 180, 168, 5, 5, 137, 231, 192, 0, 0, 0, 5, 5, 243, 223, 5, 5, 100, 163, 5, 5, 118, 217, 5, 5, 123, 199, 5, 5, 77, 166, 5, 5, 81, 211, 5, 5, 126, 232, 192, 0, 0, 0, 5, 5, 80, 195, 192, 0, 0, 0, 5, 5, 225, 227, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 84, 210, 5, 5, 2, 213, 5, 5, 176, 249, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 104, 165, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 184, 237, 5, 5, 184, 237, 5, 5, 165, 168, 192, 0, 0, 0, 5, 5, 220, 201, 192, 0, 0, 0, 5, 5, 3, 238, 5, 5, 125, 249, 5, 5, 28, 164, 5, 5, 101, 182, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 231, 220, 5, 5, 242, 202, 5, 5, 235, 204, 192, 0, 0, 0, 5, 5, 117, 202, 5, 5, 16, 168, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 5, 5, 23, 169, 192, 0, 0, 0, 5, 5, 174, 218, 5, 5, 194, 171, 5, 5, 97, 218, 5, 5, 115, 171, 5, 5, 119, 173, 5, 5, 219, 164, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 204, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 1, 0, 0, 205, 2, 0, 0, 205, 1, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 192, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 134, 2, 4, 0, 0, 0, 42, 0, 134, 3, 4, 0, 0, 0, 50, 0, 134, 2, 4, 0, 0, 0, 50, 0, 134, 3, 4, 0, 0, 0, 58, 0, 134, 2, 4, 0, 0, 0, 58, 0, 134, 3, 4, 0, 0, 0, 66, 0, 134, 2, 4, 0, 0, 0, 66, 0, 134, 3, 4, 0, 0, 0, 68, 0, 134, 2, 4, 0, 0, 0, 68, 0, 134, 3, 4, 0, 0, 0, 70, 0, 134, 2, 4, 0, 0, 0, 70, 0, 134, 3, 4, 0, 0, 0, 82, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 2, 4, 0, 0, 0, 82, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 162, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 160, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 200, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 66, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 210, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 68, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 198, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 208, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 204, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 42, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 50, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 58, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 70, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 134, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 3, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 2, 4, 0, 0, 0, 82, 0, 5, 0, 190, 0, 0, 0, 0, 0, 5, 0, 196, 0, 0, 0, 0, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 82, 239, 28, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 67, 239, 24, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 79, 236, 26, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 157, 235, 12, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 43, 245, 23, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 99, 188, 18, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 113, 208, 27, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 0, 128, 19, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 141, 204, 25, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 225, 226, 16, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 197, 248, 8, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 178, 199, 3, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 208, 229, 4, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 80, 172, 6, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 237, 181, 11, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 8, 230, 9, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 136, 221, 5, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 68, 242, 7, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 26, 172, 14, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 94, 179, 2, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 168, 185, 17, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 77, 175, 22, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 218, 198, 13, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 58, 217, 10, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 250, 208, 21, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 254, 213, 15, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 0, 0, 142, 9, 0, 5, 0, 5, 0, 0, 236, 214, 20, 15, 0, 5, 0, 0, 144, 9, 0, 5, 0, 5, 212, 55, 18, 254, 0, 0, 0, 0, 134, 3, 0, 42, 5, 162, 0, 0, 134, 2, 0, 42, 5, 162, 0, 0, 5, 3, 0, 42, 5, 162, 0, 0, 5, 2, 0, 42, 5, 162, 0, 0, 134, 3, 0, 50, 5, 162, 0, 0, 134, 2, 0, 50, 5, 162, 0, 0, 5, 3, 0, 50, 5, 162, 0, 0, 5, 2, 0, 50, 5, 162, 0, 0, 134, 3, 0, 58, 5, 162, 0, 0, 134, 2, 0, 58, 5, 162, 0, 0, 5, 3, 0, 58, 5, 162, 0, 0, 5, 2, 0, 58, 5, 162, 0, 0, 134, 3, 0, 68, 5, 160, 0, 0, 134, 2, 0, 68, 5, 160, 0, 0, 5, 3, 0, 68, 5, 160, 0, 0, 5, 2, 0, 68, 5, 160, 0, 0, 134, 3, 0, 82, 5, 162, 0, 0, 134, 2, 0, 82, 5, 162, 0, 0, 134, 74, 0, 82, 5, 162, 0, 0, 5, 3, 0, 82, 5, 162, 0, 0, 5, 2, 0, 82, 5, 162, 0, 0, 5, 74, 0, 82, 5, 162, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 134, 3, 0, 70, 5, 162, 0, 0, 134, 2, 0, 70, 5, 162, 0, 0, 5, 3, 0, 70, 5, 162, 0, 0, 5, 2, 0, 70, 5, 162, 0, 0, 134, 3, 0, 50, 5, 160, 0, 0, 134, 2, 0, 50, 5, 160, 0, 0, 5, 3, 0, 50, 5, 160, 0, 0, 5, 2, 0, 50, 5, 160, 0, 0, 134, 3, 0, 42, 5, 200, 0, 0, 134, 2, 0, 42, 5, 200, 0, 0, 5, 3, 0, 42, 5, 200, 0, 0, 5, 2, 0, 42, 5, 200, 0, 0, 134, 2, 0, 50, 5, 138, 0, 0, 5, 2, 0, 50, 5, 138, 0, 0, 134, 2, 0, 50, 5, 136, 0, 0, 5, 2, 0, 50, 5, 136, 0, 0, 134, 3, 0, 50, 5, 204, 0, 0, 134, 2, 0, 50, 5, 204, 0, 0, 5, 3, 0, 50, 5, 204, 0, 0, 5, 2, 0, 50, 5, 204, 0, 0, 134, 3, 0, 50, 5, 208, 0, 0, 134, 2, 0, 50, 5, 208, 0, 0, 5, 3, 0, 50, 5, 208, 0, 0, 5, 2, 0, 50, 5, 208, 0, 0, 134, 3, 0, 58, 5, 208, 0, 0, 134, 2, 0, 58, 5, 208, 0, 0, 5, 3, 0, 58, 5, 208, 0, 0, 5, 2, 0, 58, 5, 208, 0, 0, 134, 3, 0, 66, 5, 196, 0, 0, 134, 2, 0, 66, 5, 196, 0, 0, 5, 3, 0, 66, 5, 196, 0, 0, 5, 2, 0, 66, 5, 196, 0, 0, 134, 3, 0, 68, 5, 196, 0, 0, 134, 2, 0, 68, 5, 196, 0, 0, 5, 3, 0, 68, 5, 196, 0, 0, 5, 2, 0, 68, 5, 196, 0, 0, 134, 3, 0, 68, 5, 210, 0, 0, 134, 2, 0, 68, 5, 210, 0, 0, 5, 3, 0, 68, 5, 210, 0, 0, 5, 2, 0, 68, 5, 210, 0, 0, 134, 3, 0, 68, 5, 204, 0, 0, 134, 2, 0, 68, 5, 204, 0, 0, 5, 3, 0, 68, 5, 204, 0, 0, 5, 2, 0, 68, 5, 204, 0, 0, 134, 2, 0, 70, 5, 138, 0, 0, 5, 2, 0, 70, 5, 138, 0, 0, 134, 2, 0, 70, 5, 136, 0, 0, 5, 2, 0, 70, 5, 136, 0, 0, 134, 3, 0, 82, 5, 198, 0, 0, 134, 2, 0, 82, 5, 198, 0, 0, 134, 74, 0, 82, 5, 198, 0, 0, 5, 3, 0, 82, 5, 198, 0, 0, 5, 2, 0, 82, 5, 198, 0, 0, 5, 74, 0, 82, 5, 198, 0, 0, 134, 3, 0, 82, 5, 208, 0, 0, 134, 2, 0, 82, 5, 208, 0, 0, 134, 74, 0, 82, 5, 208, 0, 0, 5, 3, 0, 82, 5, 208, 0, 0, 5, 2, 0, 82, 5, 208, 0, 0, 5, 74, 0, 82, 5, 208, 0, 0, 134, 3, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 204, 0, 0, 134, 74, 0, 82, 5, 204, 0, 0, 5, 3, 0, 82, 5, 204, 0, 0, 5, 2, 0, 82, 5, 204, 0, 0, 5, 74, 0, 82, 5, 204, 0, 0, 134, 2, 0, 82, 5, 150, 0, 0, 5, 2, 0, 82, 5, 150, 0, 0, 134, 3, 0, 42, 5, 196, 0, 0, 134, 2, 0, 42, 5, 196, 0, 0, 5, 3, 0, 42, 5, 196, 0, 0, 5, 2, 0, 42, 5, 196, 0, 0, 134, 3, 0, 50, 5, 196, 0, 0, 134, 2, 0, 50, 5, 196, 0, 0, 5, 3, 0, 50, 5, 196, 0, 0, 5, 2, 0, 50, 5, 196, 0, 0, 134, 3, 0, 58, 5, 196, 0, 0, 134, 2, 0, 58, 5, 196, 0, 0, 5, 3, 0, 58, 5, 196, 0, 0, 5, 2, 0, 58, 5, 196, 0, 0, 134, 3, 0, 70, 5, 196, 0, 0, 134, 2, 0, 70, 5, 196, 0, 0, 5, 3, 0, 70, 5, 196, 0, 0, 5, 2, 0, 70, 5, 196, 0, 0, 156, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 70, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 196, 0, 0, 134, 2, 0, 82, 5, 196, 0, 0, 134, 74, 0, 82, 5, 196, 0, 0, 5, 3, 0, 82, 5, 196, 0, 0, 5, 2, 0, 82, 5, 196, 0, 0, 5, 74, 0, 82, 5, 196, 0, 0, 156, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 134, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 3, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 2, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 74, 0, 82, 5, 190, 0, 0, 5, 196, 0, 0, 5, 5, 142, 9, 5, 5, 164, 205, 6, 5, 144, 9, 5, 5, 142, 9, 5, 5, 188, 227, 13, 5, 144, 9, 5, 5, 142, 9, 5, 5, 11, 234, 8, 5, 144, 9, 5, 5, 142, 9, 5, 5, 226, 221, 12, 5, 144, 9, 5, 5, 142, 9, 5, 5, 227, 204, 14, 5, 144, 9, 5, 5, 142, 9, 5, 5, 135, 231, 7, 5, 144, 9, 5, 5, 142, 9, 5, 5, 187, 202, 11, 5, 144, 9, 5, 5, 142, 9, 5, 5, 189, 231, 10, 5, 144, 9, 5, 5, 142, 9, 5, 5, 178, 218, 9, 5, 144, 9, 5, 5, 0, 21, 17, 5, 197, 248, 5, 5, 0, 23, 16, 5, 197, 248, 5, 5, 0, 25, 15, 5, 197, 248, 5, 5, 0, 27, 14, 5, 197, 248, 5, 5, 0, 29, 13, 5, 197, 248, 5, 5, 0, 31, 12, 5, 197, 248, 5, 5, 0, 33, 11, 5, 197, 248, 5, 5, 0, 35, 10, 5, 197, 248, 5, 5, 0, 37, 9, 5, 197, 248, 5, 5, 0, 21, 5, 5, 0, 19, 20, 5, 197, 248, 5, 5, 0, 21, 5, 5, 0, 21, 19, 5, 197, 248, 5, 5, 0, 21, 5, 5, 0, 23, 18, 5, 197, 248, 5, 5, 0, 19, 30, 5, 147, 177, 5, 5, 0, 21, 19, 5, 147, 177, 5, 5, 0, 23, 13, 5, 147, 177, 5, 5, 0, 25, 12, 5, 147, 177, 5, 5, 0, 27, 11, 5, 147, 177, 5, 5, 0, 29, 10, 5, 147, 177, 5, 5, 0, 31, 9, 5, 147, 177, 5, 5, 0, 33, 8, 5, 147, 177, 5, 5, 0, 35, 7, 5, 147, 177, 5, 5, 0, 37, 6, 5, 147, 177, 5, 5, 0, 21, 5, 5, 0, 19, 29, 5, 147, 177, 5, 5, 0, 21, 5, 5, 0, 21, 28, 5, 147, 177, 5, 5, 0, 21, 5, 5, 0, 23, 27, 5, 147, 177, 5, 5, 0, 21, 5, 5, 0, 25, 26, 5, 147, 177, 5, 5, 0, 21, 5, 5, 0, 27, 25, 5, 147, 177, 5, 5, 0, 21, 5, 5, 0, 29, 24, 5, 147, 177, 5, 5, 0, 21, 5, 5, 0, 31, 23, 5, 147, 177, 5, 5, 0, 21, 5, 5, 0, 33, 22, 5, 147, 177, 5, 5, 0, 21, 5, 5, 0, 35, 21, 5, 147, 177, 5, 5, 0, 21, 5, 5, 0, 37, 20, 5, 147, 177, 5, 5, 0, 23, 5, 5, 0, 19, 18, 5, 147, 177, 5, 5, 0, 23, 5, 5, 0, 21, 17, 5, 147, 177, 5, 5, 0, 23, 5, 5, 0, 23, 16, 5, 147, 177, 5, 5, 0, 23, 5, 5, 0, 25, 15, 5, 147, 177, 5, 5, 0, 23, 5, 5, 0, 27, 14, 5, 147, 177, 5, 5, 167, 168, 6, 5, 86, 225, 5, 5, 210, 219, 6, 5, 89, 197, 5, 5, 47, 175, 7, 5, 112, 221, 5, 5, 27, 172, 6, 5, 191, 218, 5, 5, 136, 221, 5, 5, 32, 227, 5, 5, 66, 200, 7, 5, 188, 227, 5, 5, 0, 21, 28, 5, 8, 230, 5, 5, 0, 23, 17, 5, 8, 230, 5, 5, 0, 25, 14, 5, 8, 230, 5, 5, 0, 27, 13, 5, 8, 230, 5, 5, 0, 29, 12, 5, 8, 230, 5, 5, 0, 31, 11, 5, 8, 230, 5, 5, 0, 33, 10, 5, 8, 230, 5, 5, 0, 35, 9, 5, 8, 230, 5, 5, 0, 37, 8, 5, 8, 230, 5, 5, 0, 21, 5, 5, 0, 19, 38, 5, 8, 230, 5, 5, 0, 21, 5, 5, 0, 21, 37, 5, 8, 230, 5, 5, 0, 21, 5, 5, 0, 23, 36, 5, 8, 230, 5, 5, 0, 21, 5, 5, 0, 25, 35, 5, 8, 230, 5, 5, 0, 21, 5, 5, 0, 27, 34, 5, 8, 230, 5, 5, 0, 21, 5, 5, 0, 29, 33, 5, 8, 230, 5, 5, 0, 21, 5, 5, 0, 31, 32, 5, 8, 230, 5, 5, 0, 21, 5, 5, 0, 33, 31, 5, 8, 230, 5, 5, 0, 21, 5, 5, 0, 35, 30, 5, 8, 230, 5, 5, 0, 21, 5, 5, 0, 37, 29, 5, 8, 230, 5, 5, 0, 23, 5, 5, 0, 19, 27, 5, 8, 230, 5, 5, 0, 23, 5, 5, 0, 21, 26, 5, 8, 230, 5, 5, 0, 23, 5, 5, 0, 23, 25, 5, 8, 230, 5, 5, 0, 23, 5, 5, 0, 25, 24, 5, 8, 230, 5, 5, 0, 23, 5, 5, 0, 27, 23, 5, 8, 230, 5, 5, 0, 23, 5, 5, 0, 29, 22, 5, 8, 230, 5, 5, 0, 23, 5, 5, 0, 31, 21, 5, 8, 230, 5, 5, 0, 23, 5, 5, 0, 33, 20, 5, 8, 230, 5, 5, 0, 23, 5, 5, 0, 35, 19, 5, 8, 230, 5, 5, 0, 23, 5, 5, 0, 37, 18, 5, 8, 230, 5, 5, 0, 25, 5, 5, 0, 19, 16, 5, 8, 230, 5, 5, 0, 25, 5, 5, 0, 21, 15, 5, 8, 230, 5, 5, 82, 10, 5, 5, 122, 164, 8, 5, 84, 10, 5, 5, 82, 10, 5, 5, 79, 236, 14, 5, 84, 10, 5, 5, 82, 10, 5, 5, 67, 239, 13, 5, 84, 10, 5, 5, 82, 10, 5, 5, 198, 238, 11, 5, 84, 10, 5, 5, 82, 10, 5, 5, 147, 177, 7, 5, 84, 10, 5, 5, 82, 10, 5, 5, 46, 175, 10, 5, 84, 10, 5, 5, 82, 10, 5, 5, 159, 175, 6, 5, 84, 10, 5, 5, 82, 10, 5, 5, 46, 229, 12, 5, 84, 10, 5, 5, 82, 10, 5, 5, 171, 163, 9, 5, 84, 10, 160, 112, 0, 42, 162, 5, 0, 50, 5, 112, 0, 78, 19, 5, 0, 78, 5, 112, 0, 42, 19, 5, 0, 50, 0, 42, 156, 5, 5, 0, 12, 3, 7, 0, 12, 3, 198, 161, 64, 3, 198, 129, 65, 3, 0, 234, 134, 3, 0, 3, 198, 129, 1, 3, 0, 234, 134, 3, 4, 3, 0, 234, 134, 2, 0, 50, 156, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 134, 2, 12, 3, 0, 192, 198, 97, 64, 3, 0, 192, 198, 65, 65, 3, 0, 242, 134, 3, 0, 3, 0, 192, 198, 65, 1, 3, 0, 242, 134, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 58, 156, 5, 5, 0, 12, 3, 9, 0, 12, 3, 0, 192, 198, 161, 64, 3, 0, 192, 198, 129, 65, 3, 0, 250, 134, 3, 0, 3, 0, 192, 198, 129, 1, 3, 0, 250, 134, 3, 4, 3, 0, 250, 134, 2, 0, 66, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 0, 192, 198, 225, 64, 3, 0, 192, 198, 193, 65, 3, 255, 255, 0, 66, 134, 3, 0, 3, 0, 192, 198, 193, 1, 3, 255, 255, 0, 66, 134, 3, 4, 3, 255, 255, 0, 66, 134, 2, 0, 68, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 33, 64, 3, 1, 192, 198, 1, 65, 3, 255, 255, 0, 68, 134, 3, 0, 3, 1, 192, 198, 1, 1, 3, 255, 255, 0, 68, 134, 3, 4, 3, 255, 255, 0, 68, 134, 2, 0, 70, 156, 5, 5, 0, 12, 3, 10, 0, 12, 3, 1, 192, 198, 97, 64, 3, 1, 192, 198, 65, 65, 3, 255, 255, 0, 70, 134, 3, 0, 3, 1, 192, 198, 65, 1, 3, 255, 255, 0, 70, 134, 3, 4, 3, 255, 255, 0, 70, 134, 2, 0, 82, 156, 5, 7, 0, 12, 3, 14, 0, 12, 3, 1, 192, 198, 161, 64, 3, 1, 192, 198, 129, 65, 3, 255, 255, 0, 82, 134, 3, 68, 3, 255, 255, 0, 82, 134, 71, 0, 3, 1, 192, 198, 129, 1, 3, 255, 255, 0, 82, 134, 3, 4, 3, 255, 255, 0, 82, 134, 2, 8, 3, 197, 127, 0, 82, 134, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 0, 42, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 1, 192, 198, 225, 64, 3, 1, 192, 198, 193, 65, 3, 0, 234, 5, 3, 0, 3, 1, 192, 198, 193, 1, 3, 0, 234, 5, 3, 4, 3, 0, 234, 5, 2, 0, 50, 5, 5, 6, 0, 4, 3, 12, 0, 4, 3, 0, 242, 5, 2, 12, 3, 2, 192, 198, 33, 64, 3, 2, 192, 198, 1, 65, 3, 0, 242, 5, 3, 0, 3, 2, 192, 198, 1, 1, 3, 0, 242, 5, 3, 2, 3, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 58, 5, 5, 5, 0, 12, 3, 9, 0, 12, 3, 2, 192, 198, 97, 64, 3, 2, 192, 198, 65, 65, 3, 0, 250, 5, 3, 0, 3, 2, 192, 198, 65, 1, 3, 0, 250, 5, 3, 4, 3, 0, 250, 5, 2, 0, 66, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 161, 64, 3, 2, 192, 198, 129, 65, 3, 255, 255, 0, 66, 5, 3, 0, 3, 2, 192, 198, 129, 1, 3, 255, 255, 0, 66, 5, 3, 4, 3, 255, 255, 0, 66, 5, 2, 0, 68, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 2, 192, 198, 225, 64, 3, 2, 192, 198, 193, 65, 3, 255, 255, 0, 68, 5, 3, 0, 3, 2, 192, 198, 193, 1, 3, 255, 255, 0, 68, 5, 3, 4, 3, 255, 255, 0, 68, 5, 2, 0, 70, 5, 5, 5, 0, 12, 3, 10, 0, 12, 3, 3, 192, 198, 33, 64, 3, 3, 192, 198, 1, 65, 3, 255, 255, 0, 70, 5, 3, 0, 3, 3, 192, 198, 1, 1, 3, 255, 255, 0, 70, 5, 3, 4, 3, 255, 255, 0, 70, 5, 2, 0, 82, 5, 5, 7, 0, 12, 3, 14, 0, 12, 3, 3, 192, 198, 97, 64, 3, 3, 192, 198, 65, 65, 3, 255, 255, 0, 82, 5, 3, 68, 3, 255, 255, 0, 82, 5, 71, 0, 3, 3, 192, 198, 65, 1, 3, 255, 255, 0, 82, 5, 3, 4, 3, 255, 255, 0, 82, 5, 2, 8, 3, 197, 127, 0, 82, 5, 74, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 134, 72, 64, 3, 0, 242, 134, 73, 65, 3, 0, 242, 134, 71, 0, 3, 0, 242, 134, 73, 1, 3, 0, 242, 134, 71, 4, 3, 0, 242, 134, 70, 0, 82, 134, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 134, 72, 64, 3, 255, 255, 0, 82, 134, 73, 65, 3, 255, 255, 0, 82, 134, 71, 0, 3, 255, 255, 0, 82, 134, 73, 1, 3, 255, 255, 0, 82, 134, 71, 4, 3, 255, 255, 0, 82, 134, 70, 5, 50, 196, 142, 5, 0, 12, 3, 9, 0, 12, 3, 0, 242, 5, 72, 64, 3, 0, 242, 5, 73, 65, 3, 0, 242, 5, 71, 0, 3, 0, 242, 5, 73, 1, 3, 0, 242, 5, 71, 4, 3, 0, 242, 5, 70, 0, 82, 5, 74, 5, 0, 12, 3, 12, 0, 12, 3, 255, 255, 0, 82, 5, 72, 64, 3, 255, 255, 0, 82, 5, 73, 65, 3, 255, 255, 0, 82, 5, 71, 0, 3, 255, 255, 0, 82, 5, 73, 1, 3, 255, 255, 0, 82, 5, 71, 4, 3, 255, 255, 0, 82, 5, 70, 156, 42, 196, 162, 5, 0, 12, 3, 8, 0, 12, 3, 3, 192, 198, 194, 64, 3, 3, 192, 198, 130, 65, 3, 197, 162, 0, 3, 3, 192, 198, 130, 1, 3, 197, 162, 4, 3, 0, 192, 197, 98, 5, 42, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 66, 64, 3, 4, 192, 198, 2, 65, 3, 0, 192, 197, 162, 0, 3, 4, 192, 198, 2, 1, 3, 0, 192, 197, 162, 4, 3, 0, 192, 197, 226, 156, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 4, 192, 198, 194, 64, 3, 4, 192, 198, 130, 65, 3, 1, 192, 197, 34, 0, 3, 4, 192, 198, 130, 1, 3, 1, 192, 197, 34, 4, 3, 1, 192, 197, 98, 5, 50, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 66, 64, 3, 5, 192, 198, 2, 65, 3, 1, 192, 197, 162, 0, 3, 5, 192, 198, 2, 1, 3, 1, 192, 197, 162, 4, 3, 1, 192, 197, 226, 156, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 5, 192, 198, 194, 64, 3, 5, 192, 198, 130, 65, 3, 2, 192, 197, 34, 0, 3, 5, 192, 198, 130, 1, 3, 2, 192, 197, 34, 4, 3, 2, 192, 197, 98, 5, 58, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 66, 64, 3, 6, 192, 198, 2, 65, 3, 2, 192, 197, 162, 0, 3, 6, 192, 198, 2, 1, 3, 2, 192, 197, 162, 4, 3, 2, 192, 197, 226, 156, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 6, 192, 198, 194, 64, 3, 6, 192, 198, 130, 65, 3, 3, 192, 197, 34, 0, 3, 6, 192, 198, 130, 1, 3, 3, 192, 197, 34, 4, 3, 3, 192, 197, 98, 5, 68, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 7, 192, 198, 66, 64, 3, 7, 192, 198, 2, 65, 3, 3, 192, 197, 162, 0, 3, 7, 192, 198, 2, 1, 3, 3, 192, 197, 162, 4, 3, 3, 192, 197, 226, 156, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 4, 192, 197, 162, 12, 3, 7, 192, 198, 194, 64, 3, 7, 192, 198, 130, 65, 3, 4, 192, 197, 34, 0, 3, 7, 192, 198, 130, 1, 3, 4, 192, 197, 34, 4, 3, 4, 192, 197, 98, 5, 82, 196, 162, 6, 0, 8, 3, 12, 0, 8, 3, 5, 192, 197, 98, 12, 3, 8, 192, 198, 66, 64, 3, 8, 192, 198, 2, 65, 3, 4, 192, 197, 226, 0, 3, 8, 192, 198, 2, 1, 3, 4, 192, 197, 226, 4, 3, 5, 192, 197, 34, 156, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 8, 192, 198, 194, 64, 3, 8, 192, 198, 130, 65, 3, 5, 192, 197, 162, 0, 3, 8, 192, 198, 130, 1, 3, 5, 192, 197, 162, 4, 3, 5, 192, 197, 226, 5, 70, 196, 190, 5, 0, 12, 3, 9, 0, 12, 3, 9, 192, 198, 66, 64, 3, 9, 192, 198, 2, 65, 3, 6, 192, 197, 34, 0, 3, 9, 192, 198, 2, 1, 3, 6, 192, 197, 34, 4, 3, 6, 192, 197, 98, 156, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 34, 12, 3, 9, 192, 198, 194, 64, 3, 9, 192, 198, 130, 65, 3, 6, 192, 197, 162, 0, 3, 9, 192, 198, 130, 1, 3, 6, 192, 197, 162, 4, 3, 6, 192, 197, 226, 5, 82, 196, 190, 6, 0, 8, 3, 12, 0, 8, 3, 7, 192, 197, 226, 12, 3, 10, 192, 198, 66, 64, 3, 10, 192, 198, 2, 65, 3, 7, 192, 197, 98, 0, 3, 10, 192, 198, 2, 1, 3, 7, 192, 197, 98, 4, 3, 7, 192, 197, 162, 156, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 10, 192, 198, 194, 64, 3, 10, 192, 198, 130, 65, 3, 8, 192, 197, 34, 0, 3, 10, 192, 198, 130, 1, 3, 8, 192, 197, 34, 4, 3, 8, 192, 197, 98, 5, 70, 196, 162, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 66, 64, 3, 11, 192, 198, 2, 65, 3, 8, 192, 197, 162, 0, 3, 11, 192, 198, 2, 1, 3, 8, 192, 197, 162, 4, 3, 8, 192, 197, 226, 156, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 11, 192, 198, 194, 64, 3, 11, 192, 198, 130, 65, 3, 9, 192, 197, 34, 0, 3, 11, 192, 198, 130, 1, 3, 9, 192, 197, 34, 4, 3, 9, 192, 197, 98, 5, 50, 196, 160, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 66, 64, 3, 12, 192, 198, 2, 65, 3, 9, 192, 197, 162, 0, 3, 12, 192, 198, 2, 1, 3, 9, 192, 197, 162, 4, 3, 9, 192, 197, 226, 156, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 12, 192, 198, 194, 64, 3, 12, 192, 198, 130, 65, 3, 10, 192, 197, 34, 0, 3, 12, 192, 198, 130, 1, 3, 10, 192, 197, 34, 4, 3, 10, 192, 197, 98, 5, 42, 196, 200, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 66, 64, 3, 13, 192, 198, 2, 65, 3, 10, 192, 197, 162, 0, 3, 13, 192, 198, 2, 1, 3, 10, 192, 197, 162, 4, 3, 10, 192, 197, 226, 156, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 13, 192, 198, 194, 64, 3, 13, 192, 198, 130, 65, 3, 12, 192, 197, 34, 0, 3, 13, 192, 198, 130, 1, 3, 12, 192, 197, 34, 4, 3, 12, 192, 197, 98, 5, 50, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 66, 64, 3, 14, 192, 198, 2, 65, 3, 12, 192, 197, 162, 0, 3, 14, 192, 198, 2, 1, 3, 12, 192, 197, 162, 4, 3, 12, 192, 197, 226, 156, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 14, 192, 198, 194, 64, 3, 14, 192, 198, 130, 65, 3, 13, 192, 197, 34, 0, 3, 14, 192, 198, 130, 1, 3, 13, 192, 197, 34, 4, 3, 13, 192, 197, 98, 5, 50, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 66, 64, 3, 15, 192, 198, 2, 65, 3, 13, 192, 197, 162, 0, 3, 15, 192, 198, 2, 1, 3, 13, 192, 197, 162, 4, 3, 13, 192, 197, 226, 156, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 15, 192, 198, 194, 64, 3, 15, 192, 198, 130, 65, 3, 14, 192, 197, 34, 0, 3, 15, 192, 198, 130, 1, 3, 14, 192, 197, 34, 4, 3, 14, 192, 197, 98, 5, 58, 196, 208, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 66, 64, 3, 16, 192, 198, 2, 65, 3, 14, 192, 197, 162, 0, 3, 16, 192, 198, 2, 1, 3, 14, 192, 197, 162, 4, 3, 14, 192, 197, 226, 156, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 16, 192, 198, 194, 64, 3, 16, 192, 198, 130, 65, 3, 15, 192, 197, 34, 0, 3, 16, 192, 198, 130, 1, 3, 15, 192, 197, 34, 4, 3, 15, 192, 197, 98, 5, 66, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 66, 64, 3, 17, 192, 198, 2, 65, 3, 15, 192, 197, 162, 0, 3, 17, 192, 198, 2, 1, 3, 15, 192, 197, 162, 4, 3, 15, 192, 197, 226, 156, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 17, 192, 198, 194, 64, 3, 17, 192, 198, 130, 65, 3, 16, 192, 197, 34, 0, 3, 17, 192, 198, 130, 1, 3, 16, 192, 197, 34, 4, 3, 16, 192, 197, 98, 5, 68, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 66, 64, 3, 18, 192, 198, 2, 65, 3, 16, 192, 197, 162, 0, 3, 18, 192, 198, 2, 1, 3, 16, 192, 197, 162, 4, 3, 16, 192, 197, 226, 156, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 18, 192, 198, 194, 64, 3, 18, 192, 198, 130, 65, 3, 17, 192, 197, 34, 0, 3, 18, 192, 198, 130, 1, 3, 17, 192, 197, 34, 4, 3, 17, 192, 197, 98, 5, 68, 196, 210, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 66, 64, 3, 19, 192, 198, 2, 65, 3, 17, 192, 197, 162, 0, 3, 19, 192, 198, 2, 1, 3, 17, 192, 197, 162, 4, 3, 17, 192, 197, 226, 156, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 19, 192, 198, 194, 64, 3, 19, 192, 198, 130, 65, 3, 18, 192, 197, 34, 0, 3, 19, 192, 198, 130, 1, 3, 18, 192, 197, 34, 4, 3, 18, 192, 197, 98, 5, 68, 196, 204, 5, 0, 12, 3, 9, 0, 12, 3, 20, 192, 198, 66, 64, 3, 20, 192, 198, 2, 65, 3, 18, 192, 197, 162, 0, 3, 20, 192, 198, 2, 1, 3, 18, 192, 197, 162, 4, 3, 18, 192, 197, 226, 156, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 20, 192, 197, 162, 12, 3, 20, 192, 198, 194, 64, 3, 20, 192, 198, 130, 65, 3, 20, 192, 197, 34, 0, 3, 20, 192, 198, 130, 1, 3, 20, 192, 197, 34, 4, 3, 20, 192, 197, 98, 5, 82, 196, 198, 6, 0, 8, 3, 12, 0, 8, 3, 21, 192, 197, 98, 12, 3, 21, 192, 198, 66, 64, 3, 21, 192, 198, 2, 65, 3, 20, 192, 197, 226, 0, 3, 21, 192, 198, 2, 1, 3, 20, 192, 197, 226, 4, 3, 21, 192, 197, 34, 156, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 34, 12, 3, 21, 192, 198, 194, 64, 3, 21, 192, 198, 130, 65, 3, 21, 192, 197, 162, 0, 3, 21, 192, 198, 130, 1, 3, 21, 192, 197, 162, 4, 3, 21, 192, 197, 226, 5, 82, 196, 208, 6, 0, 8, 3, 12, 0, 8, 3, 22, 192, 197, 226, 12, 3, 22, 192, 198, 66, 64, 3, 22, 192, 198, 2, 65, 3, 22, 192, 197, 98, 0, 3, 22, 192, 198, 2, 1, 3, 22, 192, 197, 98, 4, 3, 22, 192, 197, 162, 156, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 23, 192, 197, 162, 12, 3, 22, 192, 198, 194, 64, 3, 22, 192, 198, 130, 65, 3, 23, 192, 197, 34, 0, 3, 22, 192, 198, 130, 1, 3, 23, 192, 197, 34, 4, 3, 23, 192, 197, 98, 5, 82, 196, 204, 6, 0, 8, 3, 12, 0, 8, 3, 24, 192, 197, 98, 12, 3, 23, 192, 198, 66, 64, 3, 23, 192, 198, 2, 65, 3, 23, 192, 197, 226, 0, 3, 23, 192, 198, 2, 1, 3, 23, 192, 197, 226, 4, 3, 24, 192, 197, 34, 156, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 23, 192, 198, 194, 64, 3, 23, 192, 198, 130, 65, 3, 25, 192, 197, 34, 0, 3, 23, 192, 198, 130, 1, 3, 25, 192, 197, 34, 4, 3, 25, 192, 197, 98, 5, 42, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 66, 64, 3, 24, 192, 198, 2, 65, 3, 25, 192, 197, 162, 0, 3, 24, 192, 198, 2, 1, 3, 25, 192, 197, 162, 4, 3, 25, 192, 197, 226, 156, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 24, 192, 198, 194, 64, 3, 24, 192, 198, 130, 65, 3, 26, 192, 197, 34, 0, 3, 24, 192, 198, 130, 1, 3, 26, 192, 197, 34, 4, 3, 26, 192, 197, 98, 5, 50, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 66, 64, 3, 25, 192, 198, 2, 65, 3, 26, 192, 197, 162, 0, 3, 25, 192, 198, 2, 1, 3, 26, 192, 197, 162, 4, 3, 26, 192, 197, 226, 156, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 25, 192, 198, 194, 64, 3, 25, 192, 198, 130, 65, 3, 27, 192, 197, 34, 0, 3, 25, 192, 198, 130, 1, 3, 27, 192, 197, 34, 4, 3, 27, 192, 197, 98, 5, 58, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 66, 64, 3, 26, 192, 198, 2, 65, 3, 27, 192, 197, 162, 0, 3, 26, 192, 198, 2, 1, 3, 27, 192, 197, 162, 4, 3, 27, 192, 197, 226, 156, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 26, 192, 198, 194, 64, 3, 26, 192, 198, 130, 65, 3, 28, 192, 197, 34, 0, 3, 26, 192, 198, 130, 1, 3, 28, 192, 197, 34, 4, 3, 28, 192, 197, 98, 5, 70, 196, 196, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 66, 64, 3, 27, 192, 198, 2, 65, 3, 28, 192, 197, 162, 0, 3, 27, 192, 198, 2, 1, 3, 28, 192, 197, 162, 4, 3, 28, 192, 197, 226, 29, 0, 197, 35, 5, 0, 12, 3, 9, 0, 12, 3, 27, 192, 198, 227, 64, 3, 27, 192, 198, 131, 65, 3, 29, 192, 197, 131, 0, 3, 27, 192, 198, 131, 1, 3, 29, 192, 197, 131, 4, 3, 29, 192, 197, 227, 30, 0, 197, 67, 5, 0, 12, 3, 9, 0, 12, 3, 28, 192, 198, 163, 64, 3, 28, 192, 198, 67, 65, 3, 30, 192, 197, 163, 0, 3, 28, 192, 198, 67, 1, 3, 30, 192, 197, 163, 4, 3, 31, 192, 197, 3, 156, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 31, 192, 197, 226, 12, 3, 29, 192, 198, 66, 64, 3, 29, 192, 198, 2, 65, 3, 31, 192, 197, 98, 0, 3, 29, 192, 198, 2, 1, 3, 31, 192, 197, 98, 4, 3, 31, 192, 197, 162, 5, 82, 196, 196, 6, 0, 8, 3, 12, 0, 8, 3, 32, 192, 197, 162, 12, 3, 29, 192, 198, 194, 64, 3, 29, 192, 198, 130, 65, 3, 32, 192, 197, 34, 0, 3, 29, 192, 198, 130, 1, 3, 32, 192, 197, 34, 4, 3, 32, 192, 197, 98, 32, 0, 197, 227, 6, 0, 8, 3, 12, 0, 8, 3, 34, 192, 197, 3, 12, 3, 30, 192, 198, 99, 64, 3, 30, 192, 198, 3, 65, 3, 33, 192, 197, 67, 0, 3, 30, 192, 198, 3, 1, 3, 33, 192, 197, 67, 4, 3, 33, 192, 197, 163, 34, 0, 197, 99, 6, 0, 8, 3, 12, 0, 8, 3, 35, 192, 197, 131, 12, 3, 31, 192, 198, 35, 64, 3, 30, 192, 198, 195, 65, 3, 34, 192, 197, 195, 0, 3, 30, 192, 198, 195, 1, 3, 34, 192, 197, 195, 4, 3, 35, 192, 197, 35, 41, 0, 198, 161, 35, 0, 23, 49, 78, 0, 33, 49, 38, 0, 37, 49, 20, 0, 37, 49, 255, 255, 22, 239, 5, 5, 38, 49, 255, 255, 24, 239, 5, 5, 39, 49, 255, 255, 81, 239, 5, 5, 40, 49, 255, 255, 239, 244, 5, 5, 41, 49, 255, 255, 138, 247, 5, 5, 33, 49, 255, 255, 162, 238, 5, 5, 34, 49, 255, 255, 197, 238, 5, 5, 35, 49, 255, 255, 255, 238, 5, 5, 36, 49, 255, 255, 10, 239, 5, 5, 27, 49, 20, 0, 27, 49, 255, 255, 138, 237, 5, 5, 28, 49, 255, 255, 143, 237, 5, 5, 30, 49, 255, 255, 20, 238, 5, 5, 31, 49, 255, 255, 92, 238, 5, 5, 32, 49, 255, 255, 95, 238, 5, 5, 23, 49, 255, 255, 104, 231, 5, 5, 24, 49, 255, 255, 154, 233, 5, 5, 25, 49, 255, 255, 100, 235, 5, 5, 26, 49, 255, 255, 132, 237, 5, 5, 14, 49, 38, 0, 18, 49, 20, 0, 18, 49, 255, 255, 164, 212, 5, 5, 19, 49, 255, 255, 53, 218, 5, 5, 20, 49, 255, 255, 2, 223, 5, 5, 21, 49, 255, 255, 182, 226, 5, 5, 22, 49, 255, 255, 6, 230, 5, 5, 14, 49, 255, 255, 238, 194, 5, 5, 15, 49, 255, 255, 69, 197, 5, 5, 16, 49, 255, 255, 193, 201, 5, 5, 17, 49, 255, 255, 112, 208, 5, 5, 9, 49, 20, 0, 9, 49, 255, 255, 3, 175, 5, 5, 10, 49, 255, 255, 29, 179, 5, 5, 11, 49, 255, 255, 221, 182, 5, 5, 12, 49, 255, 255, 225, 184, 5, 5, 13, 49, 255, 255, 117, 191, 5, 5, 5, 49, 255, 255, 0, 127, 5, 5, 6, 49, 255, 255, 149, 166, 5, 5, 7, 49, 255, 255, 16, 169, 5, 5, 8, 49, 255, 255, 111, 172, 5, 5, 4, 0, 6, 49, 29, 49, 30, 49, 42, 49, 5, 2, 64, 12, 32, 14, 192, 14, 232, 14, 24, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 8, 12, 16, 12, 24, 12, 32, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 12, 168, 12, 16, 13, 184, 13, 216, 14, 192, 13, 176, 13, 216, 12, 72, 13, 80, 13, 152, 13, 112, 14, 144, 12, 96, 12, 208, 12, 160, 13, 168, 16, 168, 20, 168, 24, 168, 28, 168, 32, 168, 36, 168, 40, 168, 44, 168, 48, 168, 52, 160, 12, 152, 12, 144, 14, 152, 14, 160, 14, 184, 12, 144, 13, 26, 4, 185, 60, 185, 64, 185, 68, 28, 4, 185, 76, 185, 80, 185, 84, 30, 4, 186, 96, 185, 100, 185, 104, 32, 4, 34, 4, 36, 4, 185, 124, 185, 128, 185, 136, 186, 140, 185, 144, 38, 4, 185, 156, 185, 160, 185, 164, 185, 168, 185, 172, 88, 13, 168, 13, 96, 13, 56, 14, 80, 12, 40, 14, 40, 4, 168, 60, 168, 64, 168, 68, 42, 4, 168, 76, 168, 80, 168, 84, 44, 4, 168, 96, 168, 100, 168, 104, 46, 4, 48, 4, 50, 4, 168, 124, 168, 128, 168, 136, 168, 140, 168, 144, 52, 4, 168, 156, 168, 160, 168, 164, 168, 168, 168, 172, 104, 13, 176, 14, 112, 13, 192, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 176, 12, 208, 14, 224, 14, 200, 14, 232, 14, 184, 14, 128, 13, 72, 14, 96, 14, 170, 56, 56, 13, 168, 14, 0, 0, 104, 14, 64, 14, 88, 14, 120, 14, 169, 24, 169, 28, 48, 14, 1, 0, 136, 13, 1, 0, 80, 14, 169, 20, 170, 120, 64, 13, 1, 0, 1, 0, 1, 0, 192, 12, 89, 56, 57, 56, 217, 57, 121, 58, 57, 58, 25, 58, 0, 8, 217, 66, 121, 72, 57, 72, 54, 4, 58, 74, 89, 88, 57, 88, 218, 89, 58, 90, 217, 68, 122, 114, 89, 120, 57, 120, 217, 121, 121, 122, 57, 122, 136, 14, 185, 122, 89, 152, 57, 152, 217, 153, 56, 4, 153, 169, 185, 176, 2, 8, 72, 56, 40, 56, 200, 57, 104, 58, 40, 58, 8, 58, 4, 8, 200, 66, 104, 72, 40, 72, 58, 4, 40, 74, 72, 88, 40, 88, 200, 89, 40, 90, 200, 68, 104, 114, 72, 120, 40, 120, 200, 121, 104, 122, 40, 122, 128, 14, 168, 122, 72, 152, 40, 152, 200, 153, 60, 4, 136, 169, 168, 176, 40, 170, 25, 56, 8, 56, 185, 57, 168, 57, 62, 4, 64, 4, 153, 65, 136, 65, 217, 65, 200, 65, 153, 66, 136, 66, 249, 65, 232, 65, 249, 69, 232, 69, 25, 71, 8, 71, 25, 72, 8, 72, 186, 73, 168, 73, 154, 74, 136, 74, 66, 4, 68, 4, 89, 72, 72, 72, 217, 81, 200, 81, 185, 81, 168, 81, 153, 82, 136, 82, 217, 82, 200, 82, 217, 85, 200, 85, 25, 87, 8, 87, 122, 90, 104, 90, 25, 88, 8, 88, 186, 89, 168, 89, 70, 4, 72, 4, 154, 90, 168, 92, 6, 8, 8, 8, 218, 97, 200, 97, 217, 102, 200, 102, 168, 132, 153, 105, 136, 105, 217, 106, 200, 106, 249, 105, 232, 105, 57, 107, 40, 107, 25, 107, 8, 107, 25, 112, 8, 112, 74, 4, 76, 4, 57, 112, 40, 112, 10, 8, 185, 116, 168, 116, 25, 120, 8, 120, 185, 121, 168, 121, 89, 122, 72, 122, 12, 8, 14, 8, 153, 137, 136, 137, 217, 138, 200, 138, 249, 137, 232, 137, 154, 141, 136, 141, 218, 141, 200, 141, 218, 142, 200, 142, 250, 141, 232, 141, 217, 146, 200, 146, 249, 145, 232, 145, 185, 148, 168, 148, 121, 154, 104, 154, 25, 152, 8, 152, 185, 153, 168, 153, 25, 154, 8, 154, 89, 154, 72, 154, 78, 4, 80, 4, 217, 161, 200, 161, 217, 169, 200, 169, 57, 170, 153, 173, 136, 173, 153, 174, 136, 174, 249, 173, 232, 173, 232, 140, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 65, 12, 66, 12, 65, 12, 65, 12, 65, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 104, 12, 105, 12, 112, 12, 120, 12, 128, 12, 136, 12, 120, 13, 88, 12, 224, 12, 232, 12, 240, 12, 248, 12, 24, 13, 32, 13, 40, 13, 48, 13, 216, 13, 224, 13, 232, 13, 240, 13, 209, 12, 16, 8, 1, 0, 248, 13, 48, 12, 56, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 66, 12, 200, 13, 208, 13, 0, 14, 18, 8, 1, 0, 8, 14, 20, 8, 1, 0, 16, 14, 0, 13, 8, 13, 24, 14, 22, 8, 200, 12, 72, 12, 32, 14, 217, 56, 187, 72, 200, 140, 169, 140, 200, 56, 169, 72, 187, 88, 187, 96, 169, 88, 169, 96, 169, 180, 169, 112, 217, 120, 187, 72, 200, 120, 169, 72, 209, 12, 209, 12, 1, 14, 1, 14, 9, 14, 9, 14, 169, 12, 169, 12, 255, 5, 0, 0, 255, 5, 185, 56, 255, 5, 186, 72, 255, 5, 186, 88, 255, 5, 185, 108, 255, 5, 186, 112, 255, 5, 185, 120, 255, 5, 185, 152, 255, 5, 168, 56, 255, 5, 168, 72, 255, 5, 168, 88, 255, 5, 168, 108, 255, 5, 168, 112, 255, 5, 168, 120, 255, 5, 168, 152, 255, 5, 218, 73, 255, 5, 217, 152, 255, 5, 200, 73, 255, 5, 200, 152, 255, 5, 249, 58, 255, 5, 232, 58, 255, 5, 250, 74, 255, 5, 232, 74, 255, 5, 250, 90, 255, 5, 232, 90, 255, 5, 218, 114, 255, 5, 200, 114, 255, 5, 249, 154, 255, 5, 232, 154, 255, 1, 170, 170, 3, 0, 68, 0, 59, 0, 32, 0, 27, 228, 2, 16, 251, 184, 0, 96, 1, 0, 0, 96, 8, 0, 51, 0, 83, 0, 91, 0, 109, 0, 101, 0, 116, 0, 123, 0, 130, 0, 170, 170, 148, 48, 1, 32, 6, 0, 0, 96, 250, 140, 1, 32, 153, 251, 1, 32, 233, 155, 1, 80, 169, 208, 2, 32, 19, 228, 2, 32, 185, 128, 3, 32, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 190, 128, 3, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 72, 0, 75, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23, 0, 0, 32, 7, 0, 0, 0, 15, 0, 0, 0, 27, 0, 0, 0, 27, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 23, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 115, 0, 100, 101, 102, 97, 117, 108, 116, 0, 170, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 115, 0, 116, 0, 114, 0, 111, 0, 107, 0, 101, 0, 0, 0, 1, 0, 51, 0, 6, 0, 2, 0, 32, 0, 40, 0, 170, 170, 1, 0, 0, 96, 13, 0, 0, 80, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 80, 7, 0, 0, 0, 9, 0, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, 0, 0, 95, 95, 95, 0, 0, 0, 1, 0, 32, 0, 0, 0, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 77, 0, 79, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 83, 0, 71, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 18, 0, 0, 0, 18, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 18, 0, 0, 0, 37, 37, 65, 76, 73, 65, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 84, 0, 87, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 80, 7, 0, 0, 0, 10, 0, 0, 0, 15, 0, 0, 0, 15, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 15, 0, 0, 0, 86, 101, 114, 115, 105, 111, 110, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 1, 0, 32, 0, 1, 0, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 247, 2, 0, 32, 7, 0, 0, 0, 53, 1, 0, 0, 249, 2, 0, 0, 249, 2, 0, 0, 43, 1, 0, 0, 1, 0, 0, 0, 54, 1, 0, 0, 99, 111, 100, 101, 77, 97, 112, 0, 65, 68, 80, 0, 65, 69, 68, 0, 65, 70, 65, 0, 65, 70, 78, 0, 65, 76, 75, 0, 65, 76, 76, 0, 65, 77, 68, 0, 65, 78, 71, 0, 65, 79, 65, 0, 65, 79, 75, 0, 65, 79, 78, 0, 65, 79, 82, 0, 65, 82, 65, 0, 65, 82, 80, 0, 65, 82, 83, 0, 65, 82, 89, 0, 65, 84, 83, 0, 65, 85, 68, 0, 65, 87, 71, 0, 65, 89, 77, 0, 65, 90, 77, 0, 65, 90, 78, 0, 66, 65, 68, 0, 66, 65, 77, 0, 66, 66, 68, 0, 66, 68, 84, 0, 66, 69, 67, 0, 66, 69, 70, 0, 66, 69, 76, 0, 66, 71, 74, 0, 66, 71, 75, 0, 66, 71, 76, 0, 66, 71, 78, 0, 66, 72, 68, 0, 66, 73, 70, 0, 66, 77, 68, 0, 66, 78, 68, 0, 66, 79, 66, 0, 66, 79, 80, 0, 66, 79, 86, 0, 66, 82, 66, 0, 66, 82, 67, 0, 66, 82, 69, 0, 66, 82, 76, 0, 66, 82, 78, 0, 66, 82, 82, 0, 66, 83, 68, 0, 66, 84, 78, 0, 66, 85, 75, 0, 66, 87, 80, 0, 66, 89, 66, 0, 66, 89, 78, 0, 66, 89, 82, 0, 66, 90, 68, 0, 67, 65, 68, 0, 67, 68, 70, 0, 67, 72, 67, 0, 67, 72, 69, 0, 67, 72, 70, 0, 67, 72, 87, 0, 67, 76, 70, 0, 67, 76, 80, 0, 67, 78, 89, 0, 67, 79, 80, 0, 67, 79, 85, 0, 67, 82, 67, 0, 67, 83, 68, 0, 67, 83, 74, 0, 67, 83, 75, 0, 67, 85, 67, 0, 67, 85, 80, 0, 67, 86, 69, 0, 67, 89, 80, 0, 67, 90, 75, 0, 68, 68, 77, 0, 68, 69, 77, 0, 68, 74, 70, 0, 68, 75, 75, 0, 68, 79, 80, 0, 68, 90, 68, 0, 69, 67, 83, 0, 69, 67, 86, 0, 69, 69, 75, 0, 69, 71, 80, 0, 69, 82, 78, 0, 69, 83, 65, 0, 69, 83, 66, 0, 69, 83, 80, 0, 69, 84, 66, 0, 69, 85, 82, 0, 70, 73, 77, 0, 70, 74, 68, 0, 70, 75, 80, 0, 70, 82, 70, 0, 71, 66, 80, 0, 71, 69, 75, 0, 71, 69, 76, 0, 71, 72, 67, 0, 71, 72, 80, 0, 71, 72, 83, 0, 71, 73, 80, 0, 71, 77, 68, 0, 71, 78, 69, 0, 71, 78, 70, 0, 71, 78, 83, 0, 71, 81, 69, 0, 71, 82, 68, 0, 71, 84, 81, 0, 71, 87, 69, 0, 71, 87, 80, 0, 71, 89, 68, 0, 72, 75, 68, 0, 72, 78, 76, 0, 72, 82, 68, 0, 72, 82, 75, 0, 72, 84, 71, 0, 72, 85, 70, 0, 73, 68, 82, 0, 73, 69, 80, 0, 73, 76, 80, 0, 73, 76, 82, 0, 73, 76, 83, 0, 73, 78, 82, 0, 73, 81, 68, 0, 73, 82, 82, 0, 73, 83, 74, 0, 73, 83, 75, 0, 73, 84, 76, 0, 74, 77, 68, 0, 74, 79, 68, 0, 74, 80, 89, 0, 75, 69, 83, 0, 75, 71, 83, 0, 75, 72, 82, 0, 75, 77, 70, 0, 75, 80, 87, 0, 75, 82, 87, 0, 75, 87, 68, 0, 75, 89, 68, 0, 75, 90, 84, 0, 76, 65, 74, 0, 76, 65, 75, 0, 76, 66, 80, 0, 76, 75, 82, 0, 76, 82, 68, 0, 76, 83, 76, 0, 76, 83, 77, 0, 76, 84, 76, 0, 76, 84, 84, 0, 76, 85, 67, 0, 76, 85, 70, 0, 76, 85, 76, 0, 76, 86, 76, 0, 76, 86, 82, 0, 76, 89, 68, 0, 77, 65, 68, 0, 77, 68, 76, 0, 77, 71, 65, 0, 77, 71, 70, 0, 77, 75, 68, 0, 77, 76, 70, 0, 77, 77, 75, 0, 77, 78, 84, 0, 77, 79, 80, 0, 77, 82, 79, 0, 77, 82, 85, 0, 77, 84, 76, 0, 77, 84, 80, 0, 77, 85, 82, 0, 77, 86, 81, 0, 77, 86, 82, 0, 77, 87, 75, 0, 77, 88, 78, 0, 77, 88, 80, 0, 77, 88, 86, 0, 77, 89, 82, 0, 77, 90, 69, 0, 77, 90, 77, 0, 77, 90, 78, 0, 78, 65, 68, 0, 78, 71, 78, 0, 78, 73, 67, 0, 78, 73, 79, 0, 78, 76, 71, 0, 78, 79, 75, 0, 78, 80, 82, 0, 78, 90, 68, 0, 79, 77, 82, 0, 80, 65, 66, 0, 80, 69, 72, 0, 80, 69, 73, 0, 80, 69, 78, 0, 80, 69, 83, 0, 80, 71, 75, 0, 80, 72, 80, 0, 80, 75, 82, 0, 80, 76, 78, 0, 80, 76, 90, 0, 80, 84, 69, 0, 80, 89, 71, 0, 81, 65, 82, 0, 82, 72, 68, 0, 82, 79, 75, 0, 82, 79, 76, 0, 82, 79, 78, 0, 82, 83, 68, 0, 82, 85, 66, 0, 82, 85, 82, 0, 82, 87, 70, 0, 83, 65, 82, 0, 83, 66, 68, 0, 83, 67, 82, 0, 83, 68, 68, 0, 83, 68, 71, 0, 83, 68, 80, 0, 83, 69, 75, 0, 83, 71, 68, 0, 83, 72, 80, 0, 83, 73, 84, 0, 83, 75, 75, 0, 83, 76, 76, 0, 83, 79, 83, 0, 83, 82, 68, 0, 83, 82, 71, 0, 83, 83, 80, 0, 83, 84, 68, 0, 83, 84, 78, 0, 83, 85, 82, 0, 83, 86, 67, 0, 83, 89, 80, 0, 83, 90, 76, 0, 84, 72, 66, 0, 84, 74, 82, 0, 84, 74, 83, 0, 84, 77, 77, 0, 84, 77, 84, 0, 84, 78, 68, 0, 84, 79, 80, 0, 84, 80, 69, 0, 84, 82, 76, 0, 84, 82, 89, 0, 84, 84, 68, 0, 84, 87, 68, 0, 84, 90, 83, 0, 85, 65, 72, 0, 85, 65, 75, 0, 85, 71, 83, 0, 85, 71, 87, 0, 85, 71, 88, 0, 85, 83, 68, 0, 85, 83, 78, 0, 85, 83, 83, 0, 85, 89, 73, 0, 85, 89, 78, 0, 85, 89, 80, 0, 85, 89, 85, 0, 85, 89, 87, 0, 85, 90, 83, 0, 86, 69, 66, 0, 86, 69, 70, 0, 86, 69, 83, 0, 86, 78, 67, 0, 86, 78, 68, 0, 86, 85, 86, 0, 87, 83, 84, 0, 88, 65, 70, 0, 88, 65, 71, 0, 88, 65, 85, 0, 88, 66, 65, 0, 88, 66, 66, 0, 88, 66, 67, 0, 88, 66, 68, 0, 88, 67, 68, 0, 88, 68, 82, 0, 88, 69, 85, 0, 88, 79, 70, 0, 88, 80, 68, 0, 88, 80, 70, 0, 88, 80, 84, 0, 88, 83, 85, 0, 88, 84, 83, 0, 88, 85, 65, 0, 88, 88, 88, 0, 89, 68, 68, 0, 89, 69, 82, 0, 89, 85, 68, 0, 89, 85, 77, 0, 89, 85, 78, 0, 90, 65, 76, 0, 90, 65, 82, 0, 90, 77, 75, 0, 90, 77, 87, 0, 90, 82, 78, 0, 90, 82, 90, 0, 90, 87, 67, 0, 90, 87, 68, 0, 90, 87, 76, 0, 90, 87, 78, 0, 90, 87, 82, 0, 0, 0, 170, 170, 43, 1, 40, 0, 44, 0, 48, 0, 52, 0, 56, 0, 60, 0, 64, 0, 68, 0, 72, 0, 76, 0, 80, 0, 84, 0, 88, 0, 92, 0, 96, 0, 100, 0, 104, 0, 108, 0, 112, 0, 116, 0, 120, 0, 124, 0, 128, 0, 132, 0, 136, 0, 140, 0, 144, 0, 148, 0, 152, 0, 156, 0, 160, 0, 164, 0, 168, 0, 172, 0, 176, 0, 180, 0, 184, 0, 188, 0, 192, 0, 196, 0, 200, 0, 204, 0, 208, 0, 212, 0, 216, 0, 220, 0, 224, 0, 228, 0, 232, 0, 236, 0, 240, 0, 244, 0, 248, 0, 252, 0, 0, 1, 4, 1, 8, 1, 12, 1, 16, 1, 20, 1, 24, 1, 28, 1, 32, 1, 36, 1, 40, 1, 44, 1, 48, 1, 52, 1, 56, 1, 60, 1, 64, 1, 68, 1, 72, 1, 76, 1, 80, 1, 84, 1, 88, 1, 92, 1, 96, 1, 100, 1, 104, 1, 108, 1, 112, 1, 116, 1, 120, 1, 124, 1, 128, 1, 132, 1, 136, 1, 140, 1, 144, 1, 148, 1, 152, 1, 156, 1, 160, 1, 164, 1, 168, 1, 172, 1, 176, 1, 180, 1, 184, 1, 188, 1, 192, 1, 196, 1, 200, 1, 204, 1, 208, 1, 212, 1, 216, 1, 220, 1, 224, 1, 228, 1, 232, 1, 236, 1, 240, 1, 244, 1, 248, 1, 252, 1, 0, 2, 4, 2, 8, 2, 12, 2, 16, 2, 20, 2, 24, 2, 28, 2, 32, 2, 36, 2, 40, 2, 44, 2, 48, 2, 52, 2, 56, 2, 60, 2, 64, 2, 68, 2, 72, 2, 76, 2, 80, 2, 84, 2, 88, 2, 92, 2, 96, 2, 100, 2, 104, 2, 108, 2, 112, 2, 116, 2, 120, 2, 124, 2, 128, 2, 132, 2, 136, 2, 140, 2, 144, 2, 148, 2, 152, 2, 156, 2, 160, 2, 164, 2, 168, 2, 172, 2, 176, 2, 180, 2, 184, 2, 188, 2, 192, 2, 196, 2, 200, 2, 204, 2, 208, 2, 212, 2, 216, 2, 220, 2, 224, 2, 228, 2, 232, 2, 236, 2, 240, 2, 244, 2, 248, 2, 252, 2, 0, 3, 4, 3, 8, 3, 12, 3, 16, 3, 20, 3, 24, 3, 28, 3, 32, 3, 36, 3, 40, 3, 44, 3, 48, 3, 52, 3, 56, 3, 60, 3, 64, 3, 68, 3, 72, 3, 76, 3, 80, 3, 84, 3, 88, 3, 92, 3, 96, 3, 100, 3, 104, 3, 108, 3, 112, 3, 116, 3, 120, 3, 124, 3, 128, 3, 132, 3, 136, 3, 140, 3, 144, 3, 148, 3, 152, 3, 156, 3, 160, 3, 164, 3, 168, 3, 172, 3, 176, 3, 180, 3, 184, 3, 188, 3, 192, 3, 196, 3, 200, 3, 204, 3, 208, 3, 212, 3, 216, 3, 220, 3, 224, 3, 228, 3, 232, 3, 236, 3, 240, 3, 244, 3, 248, 3, 252, 3, 0, 4, 4, 4, 8, 4, 12, 4, 16, 4, 20, 4, 24, 4, 28, 4, 32, 4, 36, 4, 40, 4, 44, 4, 48, 4, 52, 4, 56, 4, 60, 4, 64, 4, 68, 4, 72, 4, 76, 4, 80, 4, 84, 4, 88, 4, 92, 4, 96, 4, 100, 4, 104, 4, 108, 4, 112, 4, 116, 4, 120, 4, 124, 4, 128, 4, 132, 4, 136, 4, 140, 4, 144, 4, 148, 4, 152, 4, 156, 4, 160, 4, 164, 4, 168, 4, 172, 4, 176, 4, 180, 4, 184, 4, 188, 4, 192, 4, 196, 4, 200, 4, 204, 4, 208, 4, 20, 0, 0, 112, 16, 3, 0, 112, 4, 0, 0, 112, 203, 3, 0, 112, 8, 0, 0, 112, 8, 0, 0, 112, 51, 0, 0, 112, 20, 2, 0, 112, 205, 3, 0, 112, 24, 0, 0, 112, 24, 0, 0, 112, 214, 3, 0, 112, 32, 0, 0, 112, 32, 0, 0, 112, 32, 0, 0, 112, 32, 0, 0, 112, 40, 0, 0, 112, 36, 0, 0, 112, 21, 2, 0, 112, 177, 3, 0, 112, 31, 0, 0, 112, 176, 3, 0, 112, 70, 0, 0, 112, 209, 3, 0, 112, 52, 0, 0, 112, 50, 0, 0, 112, 225, 3, 0, 112, 56, 0, 0, 112, 224, 3, 0, 112, 100, 0, 0, 112, 100, 0, 0, 112, 100, 0, 0, 112, 207, 3, 0, 112, 48, 0, 0, 112, 108, 0, 0, 112, 60, 0, 0, 112, 96, 0, 0, 112, 68, 0, 0, 112, 68, 0, 0, 112, 216, 3, 0, 112, 76, 0, 0, 112, 76, 0, 0, 112, 76, 0, 0, 112, 218, 3, 0, 112, 76, 0, 0, 112, 219, 3, 0, 112, 44, 0, 0, 112, 64, 0, 0, 112, 104, 0, 0, 112, 72, 0, 0, 112, 112, 0, 0, 112, 165, 3, 0, 112, 206, 3, 0, 112, 84, 0, 0, 112, 124, 0, 0, 112, 208, 3, 0, 112, 180, 3, 0, 112, 179, 3, 0, 112, 244, 2, 0, 112, 180, 3, 0, 112, 222, 3, 0, 112, 152, 0, 0, 112, 156, 0, 0, 112, 170, 0, 0, 112, 202, 3, 0, 112, 188, 0, 0, 112, 123, 3, 0, 112, 203, 0, 0, 112, 200, 0, 0, 112, 163, 3, 0, 112, 192, 0, 0, 112, 132, 0, 0, 112, 196, 0, 0, 112, 203, 0, 0, 112, 22, 1, 0, 112, 20, 1, 0, 112, 6, 1, 0, 112, 208, 0, 0, 112, 214, 0, 0, 112, 12, 0, 0, 112, 218, 0, 0, 112, 215, 3, 0, 112, 233, 0, 0, 112, 50, 3, 0, 112, 232, 0, 0, 112, 228, 3, 0, 112, 227, 3, 0, 112, 212, 2, 0, 112, 230, 0, 0, 112, 210, 3, 0, 112, 246, 0, 0, 112, 242, 0, 0, 112, 238, 0, 0, 112, 250, 0, 0, 112, 58, 3, 0, 112, 12, 1, 0, 112, 213, 3, 0, 112, 32, 1, 0, 112, 171, 3, 0, 112, 168, 3, 0, 112, 36, 1, 0, 112, 14, 1, 0, 112, 68, 1, 0, 112, 68, 1, 0, 112, 68, 1, 0, 112, 226, 0, 0, 112, 44, 1, 0, 112, 64, 1, 0, 112, 112, 2, 0, 112, 112, 2, 0, 112, 72, 1, 0, 112, 88, 1, 0, 112, 84, 1, 0, 112, 191, 0, 0, 112, 191, 0, 0, 112, 76, 1, 0, 112, 92, 1, 0, 112, 104, 1, 0, 112, 116, 1, 0, 112, 120, 1, 0, 112, 120, 1, 0, 112, 120, 1, 0, 112, 100, 1, 0, 112, 112, 1, 0, 112, 108, 1, 0, 112, 96, 1, 0, 112, 96, 1, 0, 112, 124, 1, 0, 112, 132, 1, 0, 112, 144, 1, 0, 112, 136, 1, 0, 112, 148, 1, 0, 112, 161, 1, 0, 112, 116, 0, 0, 112, 174, 0, 0, 112, 152, 1, 0, 112, 154, 1, 0, 112, 158, 1, 0, 112, 136, 0, 0, 112, 142, 1, 0, 112, 162, 1, 0, 112, 162, 1, 0, 112, 166, 1, 0, 112, 144, 0, 0, 112, 174, 1, 0, 112, 170, 1, 0, 112, 170, 1, 0, 112, 184, 1, 0, 112, 184, 1, 0, 112, 221, 3, 0, 112, 186, 1, 0, 112, 220, 3, 0, 112, 172, 1, 0, 112, 172, 1, 0, 112, 178, 1, 0, 112, 248, 1, 0, 112, 242, 1, 0, 112, 201, 3, 0, 112, 194, 1, 0, 112, 39, 3, 0, 112, 210, 1, 0, 112, 104, 0, 0, 112, 240, 1, 0, 112, 190, 1, 0, 112, 222, 1, 0, 112, 161, 3, 0, 112, 214, 1, 0, 112, 214, 1, 0, 112, 224, 1, 0, 112, 206, 1, 0, 112, 206, 1, 0, 112, 198, 1, 0, 112, 228, 1, 0, 112, 228, 1, 0, 112, 211, 3, 0, 112, 202, 1, 0, 112, 252, 1, 0, 112, 252, 1, 0, 112, 175, 3, 0, 112, 4, 2, 0, 112, 54, 2, 0, 112, 46, 2, 0, 112, 46, 2, 0, 112, 16, 2, 0, 112, 66, 2, 0, 112, 12, 2, 0, 112, 42, 2, 0, 112, 0, 2, 0, 112, 78, 2, 0, 112, 92, 2, 0, 112, 92, 2, 0, 112, 92, 2, 0, 112, 92, 2, 0, 112, 86, 2, 0, 112, 96, 2, 0, 112, 74, 2, 0, 112, 217, 3, 0, 112, 104, 2, 0, 112, 108, 2, 0, 112, 88, 2, 0, 112, 122, 2, 0, 112, 204, 2, 0, 112, 130, 2, 0, 112, 130, 2, 0, 112, 178, 3, 0, 112, 173, 3, 0, 112, 131, 2, 0, 112, 42, 3, 0, 112, 134, 2, 0, 112, 170, 2, 0, 112, 90, 0, 0, 112, 178, 2, 0, 112, 224, 2, 0, 112, 170, 3, 0, 112, 224, 2, 0, 112, 240, 2, 0, 112, 190, 2, 0, 112, 142, 2, 0, 112, 193, 2, 0, 112, 191, 2, 0, 112, 182, 2, 0, 112, 194, 2, 0, 112, 200, 3, 0, 112, 228, 2, 0, 112, 216, 2, 0, 112, 166, 2, 0, 112, 162, 3, 0, 112, 42, 3, 0, 112, 222, 0, 0, 112, 248, 2, 0, 112, 236, 2, 0, 112, 252, 2, 0, 112, 250, 2, 0, 112, 204, 3, 0, 112, 27, 3, 0, 112, 166, 3, 0, 112, 20, 3, 0, 112, 8, 3, 0, 112, 114, 2, 0, 112, 24, 3, 0, 112, 181, 3, 0, 112, 12, 3, 0, 112, 133, 3, 0, 112, 66, 3, 0, 112, 212, 3, 0, 112, 36, 3, 0, 112, 32, 3, 0, 112, 32, 3, 0, 112, 32, 3, 0, 112, 72, 3, 0, 112, 229, 3, 0, 112, 230, 3, 0, 112, 172, 3, 0, 112, 90, 3, 0, 112, 90, 3, 0, 112, 90, 3, 0, 112, 159, 3, 0, 112, 92, 3, 0, 112, 94, 3, 0, 112, 169, 3, 0, 112, 160, 3, 0, 112, 192, 2, 0, 112, 192, 2, 0, 112, 36, 2, 0, 112, 114, 3, 0, 112, 182, 3, 0, 112, 193, 3, 0, 112, 191, 3, 0, 112, 187, 3, 0, 112, 188, 3, 0, 112, 189, 3, 0, 112, 190, 3, 0, 112, 183, 3, 0, 112, 192, 3, 0, 112, 186, 3, 0, 112, 184, 3, 0, 112, 196, 3, 0, 112, 185, 3, 0, 112, 194, 3, 0, 112, 226, 3, 0, 112, 195, 3, 0, 112, 197, 3, 0, 112, 231, 3, 0, 112, 208, 2, 0, 112, 118, 3, 0, 112, 122, 3, 0, 112, 123, 3, 0, 112, 122, 3, 0, 112, 223, 3, 0, 112, 198, 2, 0, 112, 126, 3, 0, 112, 199, 3, 0, 112, 180, 0, 0, 112, 180, 0, 0, 112, 204, 2, 0, 112, 204, 2, 0, 112, 164, 3, 0, 112, 174, 3, 0, 112, 167, 3, 0, 112, 1, 0, 32, 0, 54, 1, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 108, 18, 0, 32, 7, 0, 0, 0, 92, 1, 0, 0, 113, 18, 0, 0, 113, 18, 0, 0, 156, 0, 0, 0, 1, 0, 0, 0, 7, 12, 0, 0, 108, 111, 99, 97, 108, 101, 115, 0, 97, 102, 0, 97, 109, 0, 97, 114, 0, 97, 122, 0, 98, 103, 0, 98, 110, 0, 98, 115, 0, 99, 97, 0, 99, 99, 112, 0, 99, 104, 114, 0, 99, 115, 0, 99, 121, 0, 100, 97, 0, 100, 101, 0, 101, 101, 0, 101, 108, 0, 101, 110, 0, 101, 115, 95, 67, 79, 0, 101, 116, 0, 101, 117, 0, 102, 97, 0, 102, 105, 0, 102, 105, 108, 0, 102, 114, 0, 103, 108, 0, 103, 115, 119, 0, 103, 117, 0, 104, 101, 0, 104, 105, 0, 104, 117, 0, 104, 121, 0, 105, 100, 0, 105, 115, 0, 105, 116, 0, 106, 97, 0, 107, 97, 0, 107, 107, 0, 107, 109, 0, 107, 110, 0, 107, 111, 0, 107, 121, 0, 108, 111, 0, 108, 116, 0, 108, 118, 0, 109, 107, 0, 109, 108, 0, 109, 110, 0, 109, 114, 0, 109, 115, 0, 109, 121, 0, 110, 98, 0, 110, 101, 0, 110, 108, 0, 112, 97, 0, 112, 108, 0, 112, 116, 0, 114, 111, 0, 114, 111, 111, 116, 0, 114, 117, 0, 115, 105, 0, 115, 107, 0, 115, 108, 0, 115, 113, 0, 115, 114, 0, 115, 118, 0, 116, 97, 0, 116, 101, 0, 116, 104, 0, 116, 114, 0, 117, 107, 0, 117, 114, 0, 117, 122, 0, 118, 105, 0, 121, 117, 101, 0, 122, 104, 0, 122, 117, 0, 108, 111, 99, 97, 108, 101, 115, 95, 115, 101, 108, 101, 99, 116, 105, 111, 110, 0, 114, 117, 108, 101, 115, 0, 115, 101, 116, 49, 0, 98, 101, 102, 111, 114, 101, 0, 102, 114, 111, 109, 0, 112, 109, 0, 115, 101, 116, 49, 48, 0, 97, 102, 116, 101, 114, 110, 111, 111, 110, 49, 0, 101, 118, 101, 110, 105, 110, 103, 49, 0, 109, 105, 100, 110, 105, 103, 104, 116, 0, 97, 116, 0, 109, 111, 114, 110, 105, 110, 103, 49, 0, 110, 105, 103, 104, 116, 49, 0, 110, 111, 111, 110, 0, 115, 101, 116, 49, 48, 48, 0, 109, 111, 114, 110, 105, 110, 103, 50, 0, 115, 101, 116, 49, 48, 49, 0, 115, 101, 116, 49, 48, 50, 0, 115, 101, 116, 49, 48, 51, 0, 115, 101, 116, 49, 48, 52, 0, 115, 101, 116, 49, 48, 53, 0, 115, 101, 116, 49, 48, 54, 0, 115, 101, 116, 49, 48, 55, 0, 115, 101, 116, 49, 48, 56, 0, 115, 101, 116, 49, 48, 57, 0, 115, 101, 116, 49, 49, 0, 115, 101, 116, 49, 49, 48, 0, 115, 101, 116, 49, 49, 49, 0, 97, 102, 116, 101, 114, 110, 111, 111, 110, 50, 0, 110, 105, 103, 104, 116, 50, 0, 115, 101, 116, 49, 49, 50, 0, 115, 101, 116, 49, 49, 51, 0, 115, 101, 116, 49, 49, 52, 0, 115, 101, 116, 49, 49, 53, 0, 115, 101, 116, 49, 49, 54, 0, 115, 101, 116, 49, 49, 55, 0, 115, 101, 116, 49, 49, 56, 0, 115, 101, 116, 49, 49, 57, 0, 101, 118, 101, 110, 105, 110, 103, 50, 0, 115, 101, 116, 49, 50, 0, 115, 101, 116, 49, 50, 48, 0, 115, 101, 116, 49, 50, 49, 0, 115, 101, 116, 49, 50, 50, 0, 115, 101, 116, 49, 50, 51, 0, 115, 101, 116, 49, 50, 52, 0, 115, 101, 116, 49, 50, 53, 0, 115, 101, 116, 49, 50, 54, 0, 115, 101, 116, 49, 50, 55, 0, 115, 101, 116, 49, 50, 56, 0, 115, 101, 116, 49, 50, 57, 0, 115, 101, 116, 49, 51, 0, 115, 101, 116, 49, 51, 48, 0, 115, 101, 116, 49, 51, 49, 0, 115, 101, 116, 49, 51, 50, 0, 115, 101, 116, 49, 51, 51, 0, 115, 101, 116, 49, 51, 52, 0, 115, 101, 116, 49, 51, 53, 0, 115, 101, 116, 49, 51, 54, 0, 115, 101, 116, 49, 51, 55, 0, 115, 101, 116, 49, 51, 56, 0, 115, 101, 116, 49, 51, 57, 0, 115, 101, 116, 49, 52, 0, 115, 101, 116, 49, 52, 48, 0, 115, 101, 116, 49, 52, 49, 0, 115, 101, 116, 49, 52, 50, 0, 115, 101, 116, 49, 52, 51, 0, 115, 101, 116, 49, 52, 52, 0, 115, 101, 116, 49, 52, 53, 0, 115, 101, 116, 49, 52, 54, 0, 115, 101, 116, 49, 52, 55, 0, 115, 101, 116, 49, 52, 56, 0, 115, 101, 116, 49, 52, 57, 0, 115, 101, 116, 49, 53, 0, 115, 101, 116, 49, 53, 48, 0, 115, 101, 116, 49, 53, 49, 0, 115, 101, 116, 49, 53, 50, 0, 115, 101, 116, 49, 53, 51, 0, 115, 101, 116, 49, 53, 52, 0, 115, 101, 116, 49, 53, 53, 0, 115, 101, 116, 49, 53, 54, 0, 115, 101, 116, 49, 54, 0, 115, 101, 116, 49, 55, 0, 115, 101, 116, 49, 56, 0, 115, 101, 116, 49, 57, 0, 115, 101, 116, 50, 0, 115, 101, 116, 50, 48, 0, 115, 101, 116, 50, 49, 0, 115, 101, 116, 50, 50, 0, 115, 101, 116, 50, 51, 0, 115, 101, 116, 50, 52, 0, 115, 101, 116, 50, 53, 0, 115, 101, 116, 50, 54, 0, 115, 101, 116, 50, 55, 0, 115, 101, 116, 50, 56, 0, 115, 101, 116, 50, 57, 0, 115, 101, 116, 51, 0, 115, 101, 116, 51, 48, 0, 115, 101, 116, 51, 49, 0, 115, 101, 116, 51, 50, 0, 115, 101, 116, 51, 51, 0, 115, 101, 116, 51, 52, 0, 115, 101, 116, 51, 53, 0, 115, 101, 116, 51, 54, 0, 115, 101, 116, 51, 55, 0, 115, 101, 116, 51, 56, 0, 115, 101, 116, 51, 57, 0, 115, 101, 116, 52, 0, 115, 101, 116, 52, 48, 0, 115, 101, 116, 52, 49, 0, 115, 101, 116, 52, 50, 0, 115, 101, 116, 52, 51, 0, 115, 101, 116, 52, 52, 0, 115, 101, 116, 52, 53, 0, 115, 101, 116, 52, 54, 0, 115, 101, 116, 52, 55, 0, 115, 101, 116, 52, 56, 0, 115, 101, 116, 52, 57, 0, 115, 101, 116, 53, 0, 115, 101, 116, 53, 48, 0, 115, 101, 116, 53, 49, 0, 115, 101, 116, 53, 50, 0, 115, 101, 116, 53, 51, 0, 115, 101, 116, 53, 52, 0, 115, 101, 116, 53, 53, 0, 115, 101, 116, 53, 54, 0, 115, 101, 116, 53, 55, 0, 115, 101, 116, 53, 56, 0, 115, 101, 116, 53, 57, 0, 115, 101, 116, 54, 0, 115, 101, 116, 54, 48, 0, 115, 101, 116, 54, 49, 0, 115, 101, 116, 54, 50, 0, 115, 101, 116, 54, 51, 0, 115, 101, 116, 54, 52, 0, 115, 101, 116, 54, 53, 0, 115, 101, 116, 54, 54, 0, 115, 101, 116, 54, 55, 0, 115, 101, 116, 54, 56, 0, 115, 101, 116, 54, 57, 0, 115, 101, 116, 55, 0, 115, 101, 116, 55, 48, 0, 115, 101, 116, 55, 49, 0, 115, 101, 116, 55, 50, 0, 115, 101, 116, 55, 51, 0, 115, 101, 116, 55, 52, 0, 115, 101, 116, 55, 53, 0, 115, 101, 116, 55, 54, 0, 115, 101, 116, 55, 55, 0, 115, 101, 116, 55, 56, 0, 115, 101, 116, 55, 57, 0, 115, 101, 116, 56, 0, 115, 101, 116, 56, 48, 0, 115, 101, 116, 56, 49, 0, 115, 101, 116, 56, 50, 0, 115, 101, 116, 56, 51, 0, 115, 101, 116, 56, 52, 0, 115, 101, 116, 56, 53, 0, 115, 101, 116, 56, 54, 0, 115, 101, 116, 56, 55, 0, 115, 101, 116, 56, 56, 0, 115, 101, 116, 56, 57, 0, 115, 101, 116, 57, 0, 115, 101, 116, 57, 48, 0, 115, 101, 116, 57, 49, 0, 115, 101, 116, 57, 50, 0, 115, 101, 116, 57, 51, 0, 115, 101, 116, 57, 52, 0, 115, 101, 116, 57, 53, 0, 115, 101, 116, 57, 54, 0, 115, 101, 116, 57, 55, 0, 115, 101, 116, 57, 56, 0, 115, 101, 116, 57, 57, 0, 170, 170, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 0, 0, 49, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 48, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 48, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 48, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 48, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 48, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 53, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 53, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 53, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 53, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 53, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 53, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 53, 0, 54, 0, 0, 0, 79, 0, 40, 0, 43, 0, 46, 0, 49, 0, 52, 0, 55, 0, 58, 0, 61, 0, 64, 0, 68, 0, 72, 0, 75, 0, 78, 0, 81, 0, 84, 0, 87, 0, 90, 0, 43, 1, 93, 0, 99, 0, 102, 0, 105, 0, 108, 0, 111, 0, 115, 0, 118, 0, 121, 0, 125, 0, 128, 0, 131, 0, 69, 0, 134, 0, 137, 0, 140, 0, 143, 0, 146, 0, 149, 0, 152, 0, 155, 0, 158, 0, 161, 0, 164, 0, 167, 0, 170, 0, 173, 0, 176, 0, 179, 0, 182, 0, 185, 0, 188, 0, 191, 0, 194, 0, 197, 0, 200, 0, 203, 0, 206, 0, 209, 0, 212, 0, 215, 0, 218, 0, 223, 0, 226, 0, 229, 0, 232, 0, 235, 0, 238, 0, 241, 0, 122, 0, 244, 0, 247, 0, 250, 0, 253, 0, 0, 1, 3, 1, 6, 1, 9, 1, 12, 1, 16, 1, 19, 1, 11, 0, 26, 2, 14, 2, 246, 1, 6, 1, 190, 0, 30, 1, 226, 0, 190, 0, 74, 2, 42, 1, 86, 2, 31, 0, 26, 0, 80, 2, 114, 1, 6, 0, 214, 0, 220, 0, 168, 1, 62, 2, 132, 1, 162, 1, 234, 1, 244, 0, 208, 0, 21, 0, 84, 1, 20, 2, 78, 1, 24, 1, 174, 1, 126, 1, 222, 1, 196, 0, 232, 0, 192, 1, 56, 2, 252, 1, 38, 2, 156, 1, 198, 1, 2, 2, 216, 1, 60, 1, 66, 1, 12, 1, 150, 1, 68, 2, 90, 1, 228, 1, 204, 1, 36, 0, 96, 1, 16, 0, 102, 1, 54, 1, 202, 0, 238, 0, 1, 0, 250, 0, 108, 1, 48, 1, 36, 1, 120, 1, 18, 1, 41, 0, 44, 2, 138, 1, 144, 1, 210, 1, 240, 1, 0, 1, 72, 1, 8, 2, 32, 2, 186, 1, 180, 1, 50, 2, 79, 0, 40, 0, 43, 0, 46, 0, 49, 0, 52, 0, 55, 0, 58, 0, 61, 0, 64, 0, 68, 0, 72, 0, 75, 0, 78, 0, 81, 0, 84, 0, 87, 0, 90, 0, 43, 1, 93, 0, 99, 0, 102, 0, 105, 0, 108, 0, 111, 0, 115, 0, 118, 0, 121, 0, 125, 0, 128, 0, 131, 0, 69, 0, 134, 0, 137, 0, 140, 0, 143, 0, 146, 0, 149, 0, 152, 0, 155, 0, 158, 0, 161, 0, 164, 0, 167, 0, 170, 0, 173, 0, 176, 0, 179, 0, 182, 0, 185, 0, 188, 0, 191, 0, 194, 0, 197, 0, 200, 0, 203, 0, 206, 0, 209, 0, 212, 0, 215, 0, 218, 0, 223, 0, 226, 0, 229, 0, 232, 0, 235, 0, 238, 0, 241, 0, 122, 0, 244, 0, 247, 0, 250, 0, 253, 0, 0, 1, 3, 1, 6, 1, 9, 1, 12, 1, 16, 1, 19, 1, 104, 2, 56, 4, 249, 3, 186, 3, 253, 2, 218, 2, 206, 2, 176, 2, 218, 2, 84, 4, 232, 2, 98, 4, 128, 2, 122, 2, 91, 4, 39, 3, 98, 2, 164, 2, 170, 2, 214, 3, 63, 4, 46, 3, 221, 3, 21, 4, 194, 2, 158, 2, 116, 2, 88, 3, 0, 4, 81, 3, 200, 2, 228, 3, 53, 3, 7, 4, 146, 2, 182, 2, 165, 3, 60, 3, 193, 3, 35, 4, 144, 3, 172, 3, 200, 3, 242, 3, 25, 3, 32, 3, 4, 3, 137, 3, 70, 4, 95, 3, 14, 4, 77, 4, 134, 2, 102, 3, 110, 2, 109, 3, 246, 2, 152, 2, 188, 2, 92, 2, 11, 3, 116, 3, 239, 2, 225, 2, 67, 3, 212, 2, 140, 2, 42, 4, 123, 3, 130, 3, 235, 3, 179, 3, 18, 3, 74, 3, 207, 3, 28, 4, 158, 3, 151, 3, 49, 4, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 2, 0, 51, 1, 58, 1, 82, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 2, 0, 51, 1, 58, 1, 106, 0, 64, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 64, 0, 130, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 2, 0, 51, 1, 58, 1, 148, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 148, 0, 2, 0, 51, 1, 58, 1, 70, 0, 130, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 2, 0, 51, 1, 58, 1, 148, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 148, 0, 2, 0, 51, 1, 58, 1, 70, 0, 130, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 106, 0, 64, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 118, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 2, 0, 51, 1, 58, 1, 160, 0, 70, 0, 2, 0, 51, 1, 58, 1, 118, 0, 160, 0, 2, 0, 51, 1, 58, 1, 70, 0, 130, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 106, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 154, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 154, 0, 2, 0, 51, 1, 58, 1, 94, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 178, 0, 46, 0, 2, 0, 51, 1, 58, 1, 124, 0, 178, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 124, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 112, 0, 2, 0, 51, 1, 58, 1, 70, 0, 142, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 142, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 148, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 148, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 124, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 124, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 124, 0, 88, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 124, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 124, 0, 52, 0, 2, 0, 51, 1, 58, 1, 64, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 70, 0, 124, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 124, 0, 88, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 124, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 64, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 112, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 76, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 88, 0, 2, 0, 51, 1, 58, 1, 130, 0, 94, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 130, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 88, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 118, 0, 2, 0, 51, 1, 58, 1, 64, 0, 142, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 142, 0, 58, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 118, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 100, 0, 136, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 136, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 172, 0, 118, 0, 2, 0, 51, 1, 58, 1, 52, 0, 172, 0, 2, 0, 51, 1, 58, 1, 94, 0, 52, 0, 2, 0, 51, 1, 58, 1, 64, 0, 136, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 136, 0, 94, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 112, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 166, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 166, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 112, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 166, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 166, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 88, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 154, 0, 94, 0, 2, 0, 51, 1, 58, 1, 70, 0, 154, 0, 2, 0, 51, 1, 58, 1, 124, 0, 46, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 124, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 64, 0, 136, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 136, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 52, 0, 2, 0, 51, 1, 58, 1, 76, 0, 94, 0, 2, 0, 51, 1, 58, 1, 160, 0, 64, 0, 2, 0, 51, 1, 58, 1, 46, 0, 160, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 178, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 178, 0, 2, 0, 51, 1, 58, 1, 64, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 160, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 2, 0, 51, 1, 58, 1, 160, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 130, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 154, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 154, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 154, 0, 52, 0, 2, 0, 51, 1, 58, 1, 106, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 100, 0, 154, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 148, 0, 64, 0, 2, 0, 51, 1, 58, 1, 46, 0, 148, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 2, 0, 51, 1, 58, 1, 64, 0, 70, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 112, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 124, 0, 46, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 2, 0, 51, 1, 58, 1, 82, 0, 124, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 124, 0, 88, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 124, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 112, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 64, 0, 136, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 142, 0, 58, 0, 2, 0, 51, 1, 58, 1, 136, 0, 142, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 136, 0, 130, 0, 2, 0, 51, 1, 58, 1, 64, 0, 136, 0, 2, 0, 51, 1, 58, 1, 172, 0, 106, 0, 2, 0, 51, 1, 58, 1, 52, 0, 172, 0, 2, 0, 51, 1, 58, 1, 106, 0, 58, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 118, 0, 2, 0, 51, 1, 58, 1, 142, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 142, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 2, 0, 51, 1, 58, 1, 82, 0, 76, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 88, 0, 2, 0, 51, 1, 58, 1, 184, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 184, 0, 2, 0, 51, 1, 58, 1, 70, 0, 94, 0, 2, 0, 51, 1, 58, 1, 112, 0, 106, 0, 2, 0, 51, 1, 58, 1, 94, 0, 112, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 112, 0, 2, 0, 51, 1, 58, 1, 76, 0, 94, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 118, 0, 2, 0, 51, 1, 58, 1, 76, 0, 94, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 88, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 82, 0, 46, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 118, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 100, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 130, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 118, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 160, 0, 70, 0, 2, 0, 51, 1, 58, 1, 118, 0, 160, 0, 2, 0, 51, 1, 58, 1, 70, 0, 130, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 64, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 64, 0, 130, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 148, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 148, 0, 2, 0, 51, 1, 58, 1, 70, 0, 130, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 148, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 148, 0, 2, 0, 51, 1, 58, 1, 70, 0, 130, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 64, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 154, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 154, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 124, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 124, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 124, 0, 88, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 124, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 124, 0, 52, 0, 2, 0, 51, 1, 58, 1, 64, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 70, 0, 124, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 124, 0, 88, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 124, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 64, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 88, 0, 2, 0, 51, 1, 58, 1, 130, 0, 94, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 130, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 88, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 118, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 64, 0, 142, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 142, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 178, 0, 46, 0, 2, 0, 51, 1, 58, 1, 124, 0, 178, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 124, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 148, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 148, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 112, 0, 2, 0, 51, 1, 58, 1, 70, 0, 142, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 142, 0, 58, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 118, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 100, 0, 136, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 136, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 172, 0, 118, 0, 2, 0, 51, 1, 58, 1, 52, 0, 172, 0, 2, 0, 51, 1, 58, 1, 94, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 64, 0, 136, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 136, 0, 94, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 154, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 100, 0, 154, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 154, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 154, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 118, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 148, 0, 64, 0, 2, 0, 51, 1, 58, 1, 46, 0, 148, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 2, 0, 51, 1, 58, 1, 64, 0, 70, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 112, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 166, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 166, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 112, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 166, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 166, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 88, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 154, 0, 94, 0, 2, 0, 51, 1, 58, 1, 70, 0, 154, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 64, 0, 136, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 136, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 88, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 112, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 124, 0, 88, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 124, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 172, 0, 106, 0, 2, 0, 51, 1, 58, 1, 52, 0, 172, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 58, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 118, 0, 2, 0, 51, 1, 58, 1, 142, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 142, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 112, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 52, 0, 2, 0, 51, 1, 58, 1, 76, 0, 94, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 160, 0, 64, 0, 2, 0, 51, 1, 58, 1, 46, 0, 160, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 178, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 178, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 64, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 160, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 160, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 130, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 112, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 64, 0, 136, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 142, 0, 58, 0, 2, 0, 51, 1, 58, 1, 136, 0, 142, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 88, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 136, 0, 130, 0, 2, 0, 51, 1, 58, 1, 64, 0, 136, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 2, 0, 51, 1, 58, 1, 82, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 88, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 88, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 184, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 184, 0, 2, 0, 51, 1, 58, 1, 70, 0, 94, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 112, 0, 106, 0, 2, 0, 51, 1, 58, 1, 94, 0, 112, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 118, 0, 2, 0, 51, 1, 58, 1, 76, 0, 94, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 46, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 118, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 100, 0, 70, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 1, 0, 101, 1, 46, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 2, 0, 51, 1, 58, 1, 82, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 64, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 118, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 118, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 52, 0, 112, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 106, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 106, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 106, 0, 64, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 106, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 94, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 94, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 1, 0, 101, 1, 58, 0, 2, 0, 51, 1, 58, 1, 106, 0, 100, 0, 2, 0, 51, 1, 58, 1, 46, 0, 106, 0, 2, 0, 51, 1, 58, 1, 100, 0, 58, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 112, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 76, 0, 2, 0, 51, 1, 58, 1, 124, 0, 46, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 124, 0, 2, 0, 51, 1, 58, 1, 124, 0, 46, 0, 2, 0, 51, 1, 58, 1, 46, 0, 58, 0, 2, 0, 51, 1, 58, 1, 82, 0, 124, 0, 2, 0, 51, 1, 58, 1, 112, 0, 46, 0, 2, 0, 51, 1, 58, 1, 94, 0, 112, 0, 2, 0, 51, 1, 58, 1, 76, 0, 94, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 82, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 130, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 100, 0, 2, 0, 51, 1, 58, 1, 100, 0, 130, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 82, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 58, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 70, 0, 2, 0, 51, 1, 58, 1, 70, 0, 76, 0, 2, 0, 51, 1, 58, 1, 52, 0, 46, 0, 2, 0, 51, 1, 58, 1, 76, 0, 52, 0, 2, 0, 51, 1, 58, 1, 46, 0, 64, 0, 2, 0, 51, 1, 58, 1, 64, 0, 76, 0, 170, 170, 2, 0, 43, 0, 63, 1, 170, 170, 167, 5, 0, 80, 172, 5, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 177, 5, 0, 80, 182, 5, 0, 80, 187, 5, 0, 80, 190, 5, 0, 80, 195, 5, 0, 80, 200, 5, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 203, 5, 0, 80, 208, 5, 0, 80, 213, 5, 0, 80, 218, 5, 0, 80, 223, 5, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 228, 5, 0, 80, 233, 5, 0, 80, 238, 5, 0, 80, 243, 5, 0, 80, 248, 5, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 253, 5, 0, 80, 2, 6, 0, 80, 7, 6, 0, 80, 12, 6, 0, 80, 17, 6, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 22, 6, 0, 80, 27, 6, 0, 80, 32, 6, 0, 80, 37, 6, 0, 80, 42, 6, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 47, 6, 0, 80, 52, 6, 0, 80, 57, 6, 0, 80, 62, 6, 0, 80, 67, 6, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 72, 6, 0, 80, 77, 6, 0, 80, 82, 6, 0, 80, 87, 6, 0, 80, 92, 6, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 97, 6, 0, 80, 102, 6, 0, 80, 107, 6, 0, 80, 112, 6, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 117, 6, 0, 80, 122, 6, 0, 80, 127, 6, 0, 80, 132, 6, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 137, 6, 0, 80, 142, 6, 0, 80, 147, 6, 0, 80, 152, 6, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 157, 6, 0, 80, 162, 6, 0, 80, 167, 6, 0, 80, 172, 6, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 177, 6, 0, 80, 182, 6, 0, 80, 187, 6, 0, 80, 190, 6, 0, 80, 195, 6, 0, 80, 200, 6, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 203, 6, 0, 80, 208, 6, 0, 80, 213, 6, 0, 80, 218, 6, 0, 80, 6, 0, 72, 1, 224, 1, 104, 1, 132, 1, 113, 1, 235, 1, 170, 170, 223, 6, 0, 80, 228, 6, 0, 80, 233, 6, 0, 80, 238, 6, 0, 80, 243, 6, 0, 80, 248, 6, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 253, 6, 0, 80, 2, 7, 0, 80, 7, 7, 0, 80, 12, 7, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 17, 7, 0, 80, 22, 7, 0, 80, 27, 7, 0, 80, 32, 7, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 37, 7, 0, 80, 42, 7, 0, 80, 47, 7, 0, 80, 52, 7, 0, 80, 57, 7, 0, 80, 5, 0, 72, 1, 224, 1, 83, 1, 104, 1, 113, 1, 62, 7, 0, 80, 67, 7, 0, 80, 72, 7, 0, 80, 77, 7, 0, 80, 82, 7, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 87, 7, 0, 80, 92, 7, 0, 80, 97, 7, 0, 80, 102, 7, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 107, 7, 0, 80, 112, 7, 0, 80, 117, 7, 0, 80, 122, 7, 0, 80, 127, 7, 0, 80, 132, 7, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 137, 7, 0, 80, 142, 7, 0, 80, 147, 7, 0, 80, 152, 7, 0, 80, 6, 0, 72, 1, 83, 1, 42, 2, 104, 1, 132, 1, 113, 1, 170, 170, 157, 7, 0, 80, 162, 7, 0, 80, 167, 7, 0, 80, 172, 7, 0, 80, 177, 7, 0, 80, 182, 7, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 170, 170, 187, 7, 0, 80, 192, 7, 0, 80, 197, 7, 0, 80, 200, 7, 0, 80, 205, 7, 0, 80, 210, 7, 0, 80, 5, 0, 72, 1, 224, 1, 83, 1, 104, 1, 113, 1, 215, 7, 0, 80, 220, 7, 0, 80, 225, 7, 0, 80, 230, 7, 0, 80, 235, 7, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 240, 7, 0, 80, 245, 7, 0, 80, 250, 7, 0, 80, 255, 7, 0, 80, 6, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 235, 1, 170, 170, 4, 8, 0, 80, 9, 8, 0, 80, 14, 8, 0, 80, 19, 8, 0, 80, 24, 8, 0, 80, 29, 8, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 42, 2, 104, 1, 132, 1, 113, 1, 34, 8, 0, 80, 39, 8, 0, 80, 44, 8, 0, 80, 49, 8, 0, 80, 54, 8, 0, 80, 59, 8, 0, 80, 64, 8, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 69, 8, 0, 80, 74, 8, 0, 80, 79, 8, 0, 80, 84, 8, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 42, 2, 104, 1, 132, 1, 113, 1, 89, 8, 0, 80, 94, 8, 0, 80, 99, 8, 0, 80, 104, 8, 0, 80, 109, 8, 0, 80, 114, 8, 0, 80, 119, 8, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 124, 8, 0, 80, 129, 8, 0, 80, 134, 8, 0, 80, 139, 8, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 144, 8, 0, 80, 149, 8, 0, 80, 154, 8, 0, 80, 159, 8, 0, 80, 164, 8, 0, 80, 169, 8, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 174, 8, 0, 80, 179, 8, 0, 80, 184, 8, 0, 80, 189, 8, 0, 80, 194, 8, 0, 80, 199, 8, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 113, 1, 235, 1, 204, 8, 0, 80, 209, 8, 0, 80, 214, 8, 0, 80, 219, 8, 0, 80, 224, 8, 0, 80, 5, 0, 83, 1, 104, 1, 132, 1, 113, 1, 120, 1, 229, 8, 0, 80, 234, 8, 0, 80, 239, 8, 0, 80, 244, 8, 0, 80, 249, 8, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 252, 8, 0, 80, 1, 9, 0, 80, 6, 9, 0, 80, 11, 9, 0, 80, 16, 9, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 21, 9, 0, 80, 26, 9, 0, 80, 31, 9, 0, 80, 36, 9, 0, 80, 41, 9, 0, 80, 46, 9, 0, 80, 6, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 235, 1, 170, 170, 51, 9, 0, 80, 56, 9, 0, 80, 61, 9, 0, 80, 66, 9, 0, 80, 71, 9, 0, 80, 76, 9, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 81, 9, 0, 80, 86, 9, 0, 80, 91, 9, 0, 80, 96, 9, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 101, 9, 0, 80, 106, 9, 0, 80, 111, 9, 0, 80, 116, 9, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 121, 9, 0, 80, 126, 9, 0, 80, 131, 9, 0, 80, 136, 9, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 141, 9, 0, 80, 146, 9, 0, 80, 151, 9, 0, 80, 156, 9, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 161, 9, 0, 80, 166, 9, 0, 80, 171, 9, 0, 80, 176, 9, 0, 80, 181, 9, 0, 80, 6, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 235, 1, 170, 170, 186, 9, 0, 80, 191, 9, 0, 80, 196, 9, 0, 80, 201, 9, 0, 80, 206, 9, 0, 80, 211, 9, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 42, 2, 104, 1, 113, 1, 170, 170, 216, 9, 0, 80, 221, 9, 0, 80, 226, 9, 0, 80, 231, 9, 0, 80, 236, 9, 0, 80, 241, 9, 0, 80, 4, 0, 83, 1, 132, 1, 113, 1, 120, 1, 170, 170, 246, 9, 0, 80, 251, 9, 0, 80, 0, 10, 0, 80, 5, 10, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 8, 10, 0, 80, 13, 10, 0, 80, 18, 10, 0, 80, 23, 10, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 235, 1, 28, 10, 0, 80, 33, 10, 0, 80, 38, 10, 0, 80, 43, 10, 0, 80, 48, 10, 0, 80, 53, 10, 0, 80, 58, 10, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 113, 1, 235, 1, 170, 170, 63, 10, 0, 80, 68, 10, 0, 80, 73, 10, 0, 80, 78, 10, 0, 80, 83, 10, 0, 80, 88, 10, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 93, 10, 0, 80, 98, 10, 0, 80, 103, 10, 0, 80, 108, 10, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 113, 10, 0, 80, 118, 10, 0, 80, 123, 10, 0, 80, 128, 10, 0, 80, 133, 10, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 138, 10, 0, 80, 143, 10, 0, 80, 148, 10, 0, 80, 153, 10, 0, 80, 158, 10, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 163, 10, 0, 80, 168, 10, 0, 80, 173, 10, 0, 80, 178, 10, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 183, 10, 0, 80, 188, 10, 0, 80, 193, 10, 0, 80, 198, 10, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 203, 10, 0, 80, 208, 10, 0, 80, 213, 10, 0, 80, 218, 10, 0, 80, 223, 10, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 228, 10, 0, 80, 233, 10, 0, 80, 238, 10, 0, 80, 243, 10, 0, 80, 248, 10, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 253, 10, 0, 80, 2, 11, 0, 80, 7, 11, 0, 80, 12, 11, 0, 80, 15, 11, 0, 80, 20, 11, 0, 80, 25, 11, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 30, 11, 0, 80, 35, 11, 0, 80, 40, 11, 0, 80, 45, 11, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 50, 11, 0, 80, 55, 11, 0, 80, 60, 11, 0, 80, 65, 11, 0, 80, 70, 11, 0, 80, 75, 11, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 80, 11, 0, 80, 85, 11, 0, 80, 90, 11, 0, 80, 95, 11, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 100, 11, 0, 80, 105, 11, 0, 80, 110, 11, 0, 80, 115, 11, 0, 80, 2, 0, 72, 1, 104, 1, 170, 170, 120, 11, 0, 80, 125, 11, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 130, 11, 0, 80, 135, 11, 0, 80, 140, 11, 0, 80, 145, 11, 0, 80, 150, 11, 0, 80, 155, 11, 0, 80, 3, 0, 72, 1, 83, 1, 104, 1, 160, 11, 0, 80, 165, 11, 0, 80, 170, 11, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 175, 11, 0, 80, 180, 11, 0, 80, 185, 11, 0, 80, 188, 11, 0, 80, 193, 11, 0, 80, 198, 11, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 201, 11, 0, 80, 206, 11, 0, 80, 211, 11, 0, 80, 214, 11, 0, 80, 219, 11, 0, 80, 224, 11, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 227, 11, 0, 80, 232, 11, 0, 80, 237, 11, 0, 80, 240, 11, 0, 80, 245, 11, 0, 80, 250, 11, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 253, 11, 0, 80, 2, 12, 0, 80, 7, 12, 0, 80, 10, 12, 0, 80, 15, 12, 0, 80, 20, 12, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 23, 12, 0, 80, 28, 12, 0, 80, 33, 12, 0, 80, 36, 12, 0, 80, 41, 12, 0, 80, 46, 12, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 49, 12, 0, 80, 54, 12, 0, 80, 59, 12, 0, 80, 62, 12, 0, 80, 67, 12, 0, 80, 72, 12, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 170, 170, 75, 12, 0, 80, 80, 12, 0, 80, 85, 12, 0, 80, 88, 12, 0, 80, 93, 12, 0, 80, 98, 12, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 103, 12, 0, 80, 108, 12, 0, 80, 113, 12, 0, 80, 116, 12, 0, 80, 121, 12, 0, 80, 126, 12, 0, 80, 131, 12, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 134, 12, 0, 80, 139, 12, 0, 80, 144, 12, 0, 80, 147, 12, 0, 80, 152, 12, 0, 80, 157, 12, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 160, 12, 0, 80, 165, 12, 0, 80, 170, 12, 0, 80, 173, 12, 0, 80, 178, 12, 0, 80, 183, 12, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 186, 12, 0, 80, 191, 12, 0, 80, 196, 12, 0, 80, 199, 12, 0, 80, 204, 12, 0, 80, 209, 12, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 212, 12, 0, 80, 217, 12, 0, 80, 222, 12, 0, 80, 225, 12, 0, 80, 230, 12, 0, 80, 235, 12, 0, 80, 240, 12, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 243, 12, 0, 80, 248, 12, 0, 80, 253, 12, 0, 80, 0, 13, 0, 80, 5, 13, 0, 80, 10, 13, 0, 80, 15, 13, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 18, 13, 0, 80, 23, 13, 0, 80, 28, 13, 0, 80, 31, 13, 0, 80, 36, 13, 0, 80, 41, 13, 0, 80, 46, 13, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 49, 13, 0, 80, 54, 13, 0, 80, 59, 13, 0, 80, 62, 13, 0, 80, 67, 13, 0, 80, 72, 13, 0, 80, 77, 13, 0, 80, 5, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 80, 13, 0, 80, 85, 13, 0, 80, 90, 13, 0, 80, 93, 13, 0, 80, 98, 13, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 103, 13, 0, 80, 108, 13, 0, 80, 113, 13, 0, 80, 116, 13, 0, 80, 121, 13, 0, 80, 126, 13, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 129, 13, 0, 80, 134, 13, 0, 80, 139, 13, 0, 80, 142, 13, 0, 80, 147, 13, 0, 80, 152, 13, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 113, 1, 170, 170, 155, 13, 0, 80, 160, 13, 0, 80, 165, 13, 0, 80, 170, 13, 0, 80, 173, 13, 0, 80, 178, 13, 0, 80, 5, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 183, 13, 0, 80, 188, 13, 0, 80, 193, 13, 0, 80, 196, 13, 0, 80, 201, 13, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 206, 13, 0, 80, 211, 13, 0, 80, 216, 13, 0, 80, 221, 13, 0, 80, 226, 13, 0, 80, 231, 13, 0, 80, 5, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 236, 13, 0, 80, 241, 13, 0, 80, 246, 13, 0, 80, 249, 13, 0, 80, 254, 13, 0, 80, 8, 0, 72, 1, 83, 1, 42, 2, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 170, 170, 3, 14, 0, 80, 8, 14, 0, 80, 13, 14, 0, 80, 18, 14, 0, 80, 21, 14, 0, 80, 26, 14, 0, 80, 31, 14, 0, 80, 36, 14, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 39, 14, 0, 80, 44, 14, 0, 80, 49, 14, 0, 80, 54, 14, 0, 80, 57, 14, 0, 80, 62, 14, 0, 80, 67, 14, 0, 80, 5, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 70, 14, 0, 80, 75, 14, 0, 80, 80, 14, 0, 80, 83, 14, 0, 80, 88, 14, 0, 80, 8, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 235, 1, 120, 1, 170, 170, 93, 14, 0, 80, 98, 14, 0, 80, 103, 14, 0, 80, 106, 14, 0, 80, 111, 14, 0, 80, 116, 14, 0, 80, 121, 14, 0, 80, 126, 14, 0, 80, 5, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 129, 14, 0, 80, 134, 14, 0, 80, 139, 14, 0, 80, 142, 14, 0, 80, 147, 14, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 152, 14, 0, 80, 157, 14, 0, 80, 162, 14, 0, 80, 167, 14, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 172, 14, 0, 80, 177, 14, 0, 80, 182, 14, 0, 80, 185, 14, 0, 80, 190, 14, 0, 80, 195, 14, 0, 80, 200, 14, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 203, 14, 0, 80, 208, 14, 0, 80, 213, 14, 0, 80, 216, 14, 0, 80, 221, 14, 0, 80, 226, 14, 0, 80, 6, 0, 72, 1, 224, 1, 104, 1, 132, 1, 113, 1, 235, 1, 170, 170, 229, 14, 0, 80, 234, 14, 0, 80, 239, 14, 0, 80, 244, 14, 0, 80, 249, 14, 0, 80, 254, 14, 0, 80, 9, 0, 72, 1, 224, 1, 83, 1, 42, 2, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 3, 15, 0, 80, 8, 15, 0, 80, 13, 15, 0, 80, 18, 15, 0, 80, 23, 15, 0, 80, 26, 15, 0, 80, 31, 15, 0, 80, 36, 15, 0, 80, 41, 15, 0, 80, 5, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 44, 15, 0, 80, 49, 15, 0, 80, 54, 15, 0, 80, 57, 15, 0, 80, 62, 15, 0, 80, 9, 0, 72, 1, 224, 1, 83, 1, 42, 2, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 67, 15, 0, 80, 72, 15, 0, 80, 77, 15, 0, 80, 82, 15, 0, 80, 87, 15, 0, 80, 90, 15, 0, 80, 95, 15, 0, 80, 100, 15, 0, 80, 105, 15, 0, 80, 5, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 108, 15, 0, 80, 113, 15, 0, 80, 118, 15, 0, 80, 121, 15, 0, 80, 126, 15, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 131, 15, 0, 80, 136, 15, 0, 80, 141, 15, 0, 80, 144, 15, 0, 80, 149, 15, 0, 80, 154, 15, 0, 80, 159, 15, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 162, 15, 0, 80, 167, 15, 0, 80, 172, 15, 0, 80, 175, 15, 0, 80, 180, 15, 0, 80, 185, 15, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 113, 1, 170, 170, 188, 15, 0, 80, 193, 15, 0, 80, 198, 15, 0, 80, 203, 15, 0, 80, 206, 15, 0, 80, 211, 15, 0, 80, 8, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 235, 1, 120, 1, 170, 170, 216, 15, 0, 80, 221, 15, 0, 80, 226, 15, 0, 80, 229, 15, 0, 80, 234, 15, 0, 80, 239, 15, 0, 80, 244, 15, 0, 80, 249, 15, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 252, 15, 0, 80, 1, 16, 0, 80, 6, 16, 0, 80, 11, 16, 0, 80, 14, 16, 0, 80, 19, 16, 0, 80, 24, 16, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 29, 16, 0, 80, 34, 16, 0, 80, 39, 16, 0, 80, 44, 16, 0, 80, 47, 16, 0, 80, 52, 16, 0, 80, 57, 16, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 235, 1, 120, 1, 62, 16, 0, 80, 67, 16, 0, 80, 72, 16, 0, 80, 75, 16, 0, 80, 80, 16, 0, 80, 85, 16, 0, 80, 90, 16, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 93, 16, 0, 80, 98, 16, 0, 80, 103, 16, 0, 80, 106, 16, 0, 80, 111, 16, 0, 80, 116, 16, 0, 80, 121, 16, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 124, 16, 0, 80, 129, 16, 0, 80, 134, 16, 0, 80, 137, 16, 0, 80, 142, 16, 0, 80, 147, 16, 0, 80, 8, 0, 72, 1, 224, 1, 83, 1, 42, 2, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 150, 16, 0, 80, 155, 16, 0, 80, 160, 16, 0, 80, 165, 16, 0, 80, 170, 16, 0, 80, 173, 16, 0, 80, 178, 16, 0, 80, 183, 16, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 186, 16, 0, 80, 191, 16, 0, 80, 196, 16, 0, 80, 199, 16, 0, 80, 204, 16, 0, 80, 209, 16, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 212, 16, 0, 80, 217, 16, 0, 80, 222, 16, 0, 80, 225, 16, 0, 80, 230, 16, 0, 80, 235, 16, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 238, 16, 0, 80, 243, 16, 0, 80, 248, 16, 0, 80, 253, 16, 0, 80, 2, 17, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 7, 17, 0, 80, 12, 17, 0, 80, 17, 17, 0, 80, 22, 17, 0, 80, 25, 17, 0, 80, 30, 17, 0, 80, 35, 17, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 40, 17, 0, 80, 45, 17, 0, 80, 50, 17, 0, 80, 53, 17, 0, 80, 58, 17, 0, 80, 63, 17, 0, 80, 68, 17, 0, 80, 8, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 170, 170, 71, 17, 0, 80, 76, 17, 0, 80, 81, 17, 0, 80, 86, 17, 0, 80, 89, 17, 0, 80, 94, 17, 0, 80, 99, 17, 0, 80, 104, 17, 0, 80, 8, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 235, 1, 120, 1, 170, 170, 107, 17, 0, 80, 112, 17, 0, 80, 117, 17, 0, 80, 120, 17, 0, 80, 125, 17, 0, 80, 130, 17, 0, 80, 135, 17, 0, 80, 140, 17, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 143, 17, 0, 80, 148, 17, 0, 80, 153, 17, 0, 80, 156, 17, 0, 80, 161, 17, 0, 80, 166, 17, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 169, 17, 0, 80, 174, 17, 0, 80, 179, 17, 0, 80, 182, 17, 0, 80, 187, 17, 0, 80, 192, 17, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 195, 17, 0, 80, 200, 17, 0, 80, 205, 17, 0, 80, 208, 17, 0, 80, 213, 17, 0, 80, 218, 17, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 235, 1, 221, 17, 0, 80, 226, 17, 0, 80, 231, 17, 0, 80, 236, 17, 0, 80, 241, 17, 0, 80, 246, 17, 0, 80, 251, 17, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 113, 1, 235, 1, 0, 18, 0, 80, 5, 18, 0, 80, 10, 18, 0, 80, 15, 18, 0, 80, 18, 18, 0, 80, 23, 18, 0, 80, 28, 18, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 33, 18, 0, 80, 38, 18, 0, 80, 43, 18, 0, 80, 46, 18, 0, 80, 51, 18, 0, 80, 56, 18, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 59, 18, 0, 80, 64, 18, 0, 80, 69, 18, 0, 80, 72, 18, 0, 80, 77, 18, 0, 80, 82, 18, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 170, 170, 85, 18, 0, 80, 90, 18, 0, 80, 95, 18, 0, 80, 98, 18, 0, 80, 103, 18, 0, 80, 108, 18, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 113, 18, 0, 80, 118, 18, 0, 80, 123, 18, 0, 80, 126, 18, 0, 80, 131, 18, 0, 80, 136, 18, 0, 80, 7, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 120, 1, 139, 18, 0, 80, 144, 18, 0, 80, 149, 18, 0, 80, 152, 18, 0, 80, 157, 18, 0, 80, 162, 18, 0, 80, 167, 18, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 18, 0, 80, 175, 18, 0, 80, 180, 18, 0, 80, 185, 18, 0, 80, 190, 18, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 195, 18, 0, 80, 200, 18, 0, 80, 205, 18, 0, 80, 208, 18, 0, 80, 213, 18, 0, 80, 218, 18, 0, 80, 7, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 221, 18, 0, 80, 226, 18, 0, 80, 231, 18, 0, 80, 236, 18, 0, 80, 239, 18, 0, 80, 244, 18, 0, 80, 249, 18, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 113, 1, 120, 1, 170, 170, 254, 18, 0, 80, 3, 19, 0, 80, 8, 19, 0, 80, 11, 19, 0, 80, 16, 19, 0, 80, 21, 19, 0, 80, 3, 0, 72, 1, 104, 1, 120, 1, 24, 19, 0, 80, 29, 19, 0, 80, 34, 19, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 37, 19, 0, 80, 42, 19, 0, 80, 47, 19, 0, 80, 52, 19, 0, 80, 57, 19, 0, 80, 62, 19, 0, 80, 5, 0, 72, 1, 83, 1, 92, 1, 104, 1, 120, 1, 67, 19, 0, 80, 72, 19, 0, 80, 77, 19, 0, 80, 80, 19, 0, 80, 85, 19, 0, 80, 2, 0, 43, 0, 63, 1, 170, 170, 88, 19, 0, 80, 93, 19, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 170, 170, 98, 19, 0, 80, 103, 19, 0, 80, 108, 19, 0, 80, 111, 19, 0, 80, 116, 19, 0, 80, 121, 19, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 126, 19, 0, 80, 131, 19, 0, 80, 136, 19, 0, 80, 141, 19, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 146, 19, 0, 80, 151, 19, 0, 80, 156, 19, 0, 80, 161, 19, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 166, 19, 0, 80, 171, 19, 0, 80, 176, 19, 0, 80, 181, 19, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 92, 1, 104, 1, 113, 1, 170, 170, 186, 19, 0, 80, 191, 19, 0, 80, 196, 19, 0, 80, 201, 19, 0, 80, 204, 19, 0, 80, 209, 19, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 214, 19, 0, 80, 219, 19, 0, 80, 224, 19, 0, 80, 229, 19, 0, 80, 234, 19, 0, 80, 239, 19, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 244, 19, 0, 80, 249, 19, 0, 80, 254, 19, 0, 80, 3, 20, 0, 80, 8, 20, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 13, 20, 0, 80, 18, 20, 0, 80, 23, 20, 0, 80, 28, 20, 0, 80, 33, 20, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 38, 20, 0, 80, 43, 20, 0, 80, 48, 20, 0, 80, 53, 20, 0, 80, 58, 20, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 63, 20, 0, 80, 68, 20, 0, 80, 73, 20, 0, 80, 78, 20, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 83, 20, 0, 80, 88, 20, 0, 80, 93, 20, 0, 80, 98, 20, 0, 80, 6, 0, 72, 1, 83, 1, 92, 1, 104, 1, 132, 1, 113, 1, 170, 170, 103, 20, 0, 80, 108, 20, 0, 80, 113, 20, 0, 80, 116, 20, 0, 80, 121, 20, 0, 80, 126, 20, 0, 80, 5, 0, 72, 1, 83, 1, 104, 1, 132, 1, 113, 1, 131, 20, 0, 80, 136, 20, 0, 80, 141, 20, 0, 80, 146, 20, 0, 80, 151, 20, 0, 80, 4, 0, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 156, 20, 0, 80, 161, 20, 0, 80, 166, 20, 0, 80, 171, 20, 0, 80, 3, 0, 83, 1, 132, 1, 113, 1, 176, 20, 0, 80, 181, 20, 0, 80, 186, 20, 0, 80, 6, 0, 72, 1, 224, 1, 83, 1, 104, 1, 132, 1, 113, 1, 170, 170, 191, 20, 0, 80, 196, 20, 0, 80, 201, 20, 0, 80, 206, 20, 0, 80, 211, 20, 0, 80, 216, 20, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 221, 20, 0, 80, 226, 20, 0, 80, 231, 20, 0, 80, 236, 20, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 241, 20, 0, 80, 246, 20, 0, 80, 251, 20, 0, 80, 0, 21, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 5, 21, 0, 80, 10, 21, 0, 80, 15, 21, 0, 80, 20, 21, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 25, 21, 0, 80, 30, 21, 0, 80, 35, 21, 0, 80, 40, 21, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 45, 21, 0, 80, 50, 21, 0, 80, 55, 21, 0, 80, 60, 21, 0, 80, 4, 0, 72, 1, 83, 1, 104, 1, 113, 1, 170, 170, 65, 21, 0, 80, 70, 21, 0, 80, 75, 21, 0, 80, 80, 21, 0, 80, 156, 0, 46, 1, 66, 1, 125, 1, 141, 1, 148, 1, 155, 1, 162, 1, 169, 1, 176, 1, 183, 1, 190, 1, 197, 1, 204, 1, 210, 1, 217, 1, 242, 1, 249, 1, 0, 2, 7, 2, 14, 2, 21, 2, 28, 2, 35, 2, 51, 2, 57, 2, 64, 2, 71, 2, 78, 2, 85, 2, 92, 2, 99, 2, 106, 2, 113, 2, 120, 2, 127, 2, 133, 2, 140, 2, 147, 2, 154, 2, 161, 2, 168, 2, 175, 2, 182, 2, 189, 2, 196, 2, 203, 2, 209, 2, 216, 2, 223, 2, 230, 2, 237, 2, 244, 2, 251, 2, 2, 3, 9, 3, 16, 3, 23, 3, 29, 3, 36, 3, 43, 3, 50, 3, 57, 3, 64, 3, 71, 3, 78, 3, 84, 3, 90, 3, 96, 3, 102, 3, 107, 3, 113, 3, 119, 3, 125, 3, 131, 3, 137, 3, 143, 3, 149, 3, 155, 3, 161, 3, 167, 3, 172, 3, 178, 3, 184, 3, 190, 3, 196, 3, 202, 3, 208, 3, 214, 3, 220, 3, 226, 3, 232, 3, 237, 3, 243, 3, 249, 3, 255, 3, 5, 4, 11, 4, 17, 4, 23, 4, 29, 4, 35, 4, 41, 4, 46, 4, 52, 4, 58, 4, 64, 4, 70, 4, 76, 4, 82, 4, 88, 4, 94, 4, 100, 4, 106, 4, 111, 4, 117, 4, 123, 4, 129, 4, 135, 4, 141, 4, 147, 4, 153, 4, 159, 4, 165, 4, 171, 4, 176, 4, 182, 4, 188, 4, 194, 4, 200, 4, 206, 4, 212, 4, 218, 4, 224, 4, 230, 4, 236, 4, 241, 4, 247, 4, 253, 4, 3, 5, 9, 5, 15, 5, 21, 5, 27, 5, 33, 5, 39, 5, 45, 5, 50, 5, 56, 5, 62, 5, 68, 5, 74, 5, 80, 5, 86, 5, 92, 5, 98, 5, 104, 5, 170, 170, 7, 12, 0, 32, 11, 12, 0, 32, 21, 12, 0, 32, 29, 12, 0, 32, 37, 12, 0, 32, 45, 12, 0, 32, 53, 12, 0, 32, 61, 12, 0, 32, 69, 12, 0, 32, 76, 12, 0, 32, 83, 12, 0, 32, 90, 12, 0, 32, 97, 12, 0, 32, 107, 12, 0, 32, 114, 12, 0, 32, 124, 12, 0, 32, 131, 12, 0, 32, 138, 12, 0, 32, 146, 12, 0, 32, 154, 12, 0, 32, 161, 12, 0, 32, 171, 12, 0, 32, 178, 12, 0, 32, 188, 12, 0, 32, 198, 12, 0, 32, 206, 12, 0, 32, 213, 12, 0, 32, 223, 12, 0, 32, 234, 12, 0, 32, 241, 12, 0, 32, 252, 12, 0, 32, 3, 13, 0, 32, 13, 13, 0, 32, 23, 13, 0, 32, 31, 13, 0, 32, 39, 13, 0, 32, 47, 13, 0, 32, 57, 13, 0, 32, 67, 13, 0, 32, 74, 13, 0, 32, 81, 13, 0, 32, 88, 13, 0, 32, 95, 13, 0, 32, 103, 13, 0, 32, 113, 13, 0, 32, 123, 13, 0, 32, 130, 13, 0, 32, 137, 13, 0, 32, 148, 13, 0, 32, 158, 13, 0, 32, 165, 13, 0, 32, 173, 13, 0, 32, 181, 13, 0, 32, 188, 13, 0, 32, 195, 13, 0, 32, 203, 13, 0, 32, 211, 13, 0, 32, 222, 13, 0, 32, 229, 13, 0, 32, 239, 13, 0, 32, 246, 13, 0, 32, 253, 13, 0, 32, 1, 14, 0, 32, 11, 14, 0, 32, 16, 14, 0, 32, 26, 14, 0, 32, 36, 14, 0, 32, 46, 14, 0, 32, 56, 14, 0, 32, 66, 14, 0, 32, 76, 14, 0, 32, 86, 14, 0, 32, 97, 14, 0, 32, 107, 14, 0, 32, 117, 14, 0, 32, 127, 14, 0, 32, 138, 14, 0, 32, 149, 14, 0, 32, 160, 14, 0, 32, 171, 14, 0, 32, 179, 14, 0, 32, 189, 14, 0, 32, 199, 14, 0, 32, 209, 14, 0, 32, 217, 14, 0, 32, 227, 14, 0, 32, 235, 14, 0, 32, 248, 14, 0, 32, 3, 15, 0, 32, 11, 15, 0, 32, 24, 15, 0, 32, 32, 15, 0, 32, 39, 15, 0, 32, 50, 15, 0, 32, 60, 15, 0, 32, 70, 15, 0, 32, 84, 15, 0, 32, 92, 15, 0, 32, 106, 15, 0, 32, 114, 15, 0, 32, 125, 15, 0, 32, 135, 15, 0, 32, 145, 15, 0, 32, 158, 15, 0, 32, 169, 15, 0, 32, 180, 15, 0, 32, 191, 15, 0, 32, 202, 15, 0, 32, 212, 15, 0, 32, 225, 15, 0, 32, 235, 15, 0, 32, 245, 15, 0, 32, 253, 15, 0, 32, 8, 16, 0, 32, 19, 16, 0, 32, 32, 16, 0, 32, 45, 16, 0, 32, 55, 16, 0, 32, 65, 16, 0, 32, 75, 16, 0, 32, 86, 16, 0, 32, 97, 16, 0, 32, 107, 16, 0, 32, 117, 16, 0, 32, 127, 16, 0, 32, 137, 16, 0, 32, 148, 16, 0, 32, 156, 16, 0, 32, 166, 16, 0, 32, 177, 16, 0, 32, 187, 16, 0, 32, 192, 16, 0, 32, 202, 16, 0, 32, 210, 16, 0, 32, 214, 16, 0, 32, 224, 16, 0, 32, 231, 16, 0, 32, 238, 16, 0, 32, 245, 16, 0, 32, 255, 16, 0, 32, 9, 17, 0, 32, 17, 17, 0, 32, 25, 17, 0, 32, 33, 17, 0, 32, 40, 17, 0, 32, 47, 17, 0, 32, 57, 17, 0, 32, 65, 17, 0, 32, 72, 17, 0, 32, 77, 17, 0, 32, 87, 17, 0, 32, 94, 17, 0, 32, 101, 17, 0, 32, 108, 17, 0, 32, 115, 17, 0, 32, 122, 17, 0, 32, 3, 0, 32, 0, 22, 1, 40, 1, 105, 4, 0, 80, 8, 5, 0, 80, 129, 17, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 124, 0, 0, 32, 7, 0, 0, 0, 53, 0, 0, 0, 126, 0, 0, 0, 126, 0, 0, 0, 54, 0, 0, 0, 1, 0, 0, 0, 124, 0, 0, 0, 103, 101, 110, 100, 101, 114, 76, 105, 115, 116, 0, 97, 102, 0, 97, 114, 0, 98, 103, 0, 98, 110, 0, 99, 97, 0, 99, 115, 0, 100, 97, 0, 100, 101, 0, 101, 108, 0, 101, 110, 0, 101, 115, 0, 101, 116, 0, 101, 117, 0, 102, 97, 0, 102, 105, 0, 102, 105, 108, 0, 102, 114, 0, 103, 117, 0, 104, 101, 0, 104, 105, 0, 104, 114, 0, 104, 117, 0, 105, 100, 0, 105, 115, 0, 105, 116, 0, 106, 97, 0, 107, 110, 0, 107, 111, 0, 108, 116, 0, 108, 118, 0, 109, 108, 0, 109, 114, 0, 109, 115, 0, 110, 108, 0, 110, 111, 0, 112, 108, 0, 112, 116, 0, 114, 111, 0, 114, 117, 0, 115, 107, 0, 115, 108, 0, 115, 114, 0, 115, 118, 0, 115, 119, 0, 116, 97, 0, 116, 101, 0, 116, 104, 0, 116, 114, 0, 117, 107, 0, 117, 114, 0, 118, 105, 0, 122, 104, 0, 122, 104, 95, 72, 97, 110, 116, 0, 122, 117, 0, 170, 0, 0, 110, 0, 101, 0, 117, 0, 116, 0, 114, 0, 97, 0, 108, 0, 0, 0, 109, 0, 97, 0, 108, 0, 101, 0, 84, 0, 97, 0, 105, 0, 110, 0, 116, 0, 115, 0, 0, 0, 109, 0, 105, 0, 120, 0, 101, 0, 100, 0, 78, 0, 101, 0, 117, 0, 116, 0, 114, 0, 97, 0, 108, 0, 0, 0, 54, 0, 43, 0, 46, 0, 49, 0, 52, 0, 55, 0, 58, 0, 61, 0, 64, 0, 67, 0, 70, 0, 73, 0, 76, 0, 79, 0, 82, 0, 85, 0, 88, 0, 92, 0, 95, 0, 98, 0, 101, 0, 104, 0, 107, 0, 110, 0, 113, 0, 116, 0, 119, 0, 122, 0, 125, 0, 128, 0, 131, 0, 134, 0, 137, 0, 140, 0, 143, 0, 146, 0, 149, 0, 152, 0, 155, 0, 158, 0, 161, 0, 164, 0, 167, 0, 170, 0, 173, 0, 176, 0, 179, 0, 182, 0, 185, 0, 188, 0, 191, 0, 194, 0, 197, 0, 200, 0, 208, 0, 1, 0, 9, 0, 1, 0, 1, 0, 9, 0, 9, 0, 1, 0, 1, 0, 20, 0, 1, 0, 9, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 9, 0, 1, 0, 9, 0, 9, 0, 9, 0, 1, 0, 1, 0, 20, 0, 9, 0, 1, 0, 1, 0, 1, 0, 9, 0, 9, 0, 1, 0, 9, 0, 1, 0, 9, 0, 1, 0, 9, 0, 9, 0, 9, 0, 9, 0, 9, 0, 9, 0, 9, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 9, 0, 9, 0, 1, 0, 9, 0, 9, 0, 1, 0, 1, 0, 32, 0, 33, 0, 0, 80, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 32, 7, 0, 0, 0, 11, 0, 0, 0, 14, 0, 0, 0, 14, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 12, 0, 0, 0, 83, 116, 97, 110, 100, 97, 114, 100, 73, 67, 85, 0, 0, 0, 170, 170, 1, 0, 32, 0, 0, 0, 0, 128, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 0, 0, 80, 7, 0, 0, 0, 14, 0, 0, 0, 22, 0, 0, 0, 22, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 22, 0, 0, 0, 68, 97, 116, 97, 86, 101, 114, 115, 105, 111, 110, 0, 73, 67, 85, 86, 101, 114, 115, 105, 111, 110, 0, 170, 0, 0, 54, 0, 54, 0, 46, 0, 49, 0, 46, 0, 48, 0, 46, 0, 48, 0, 0, 0, 2, 0, 32, 0, 44, 0, 1, 0, 1, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 168, 6, 0, 32, 7, 0, 0, 0, 22, 3, 0, 0, 178, 6, 0, 0, 178, 6, 0, 0, 47, 1, 0, 0, 1, 0, 0, 0, 33, 6, 0, 0, 98, 99, 112, 84, 121, 112, 101, 65, 108, 105, 97, 115, 0, 99, 97, 0, 105, 115, 108, 97, 109, 105, 99, 99, 0, 107, 101, 121, 73, 110, 102, 111, 0, 100, 101, 112, 114, 101, 99, 97, 116, 101, 100, 0, 107, 104, 0, 118, 116, 0, 118, 97, 108, 117, 101, 84, 121, 112, 101, 0, 104, 48, 0, 120, 48, 0, 107, 101, 121, 77, 97, 112, 0, 99, 97, 108, 101, 110, 100, 97, 114, 0, 99, 111, 108, 97, 108, 116, 101, 114, 110, 97, 116, 101, 0, 99, 111, 108, 98, 97, 99, 107, 119, 97, 114, 100, 115, 0, 99, 111, 108, 99, 97, 115, 101, 102, 105, 114, 115, 116, 0, 99, 111, 108, 99, 97, 115, 101, 108, 101, 118, 101, 108, 0, 99, 111, 108, 104, 105, 114, 97, 103, 97, 110, 97, 113, 117, 97, 116, 101, 114, 110, 97, 114, 121, 0, 99, 111, 108, 108, 97, 116, 105, 111, 110, 0, 99, 111, 108, 110, 111, 114, 109, 97, 108, 105, 122, 97, 116, 105, 111, 110, 0, 99, 111, 108, 110, 117, 109, 101, 114, 105, 99, 0, 99, 111, 108, 114, 101, 111, 114, 100, 101, 114, 0, 99, 111, 108, 115, 116, 114, 101, 110, 103, 116, 104, 0, 99, 117, 114, 114, 101, 110, 99, 121, 0, 100, 48, 0, 102, 119, 0, 104, 111, 117, 114, 115, 0, 105, 48, 0, 107, 48, 0, 107, 118, 0, 108, 98, 0, 108, 119, 0, 109, 48, 0, 109, 101, 97, 115, 117, 114, 101, 0, 110, 117, 109, 98, 101, 114, 115, 0, 115, 48, 0, 116, 48, 0, 116, 105, 109, 101, 122, 111, 110, 101, 0, 118, 97, 114, 105, 97, 98, 108, 101, 116, 111, 112, 0, 116, 121, 112, 101, 65, 108, 105, 97, 115, 0, 113, 117, 97, 114, 116, 101, 114, 110, 97, 114, 121, 0, 116, 121, 112, 101, 73, 110, 102, 111, 0, 99, 111, 0, 100, 105, 114, 101, 99, 116, 0, 99, 97, 109, 116, 114, 0, 116, 121, 112, 101, 77, 97, 112, 0, 98, 117, 100, 100, 104, 105, 115, 116, 0, 99, 104, 105, 110, 101, 115, 101, 0, 99, 111, 112, 116, 105, 99, 0, 100, 97, 110, 103, 105, 0, 101, 116, 104, 105, 111, 112, 105, 99, 0, 101, 116, 104, 105, 111, 112, 105, 99, 45, 97, 109, 101, 116, 101, 45, 97, 108, 101, 109, 0, 103, 114, 101, 103, 111, 114, 105, 97, 110, 0, 104, 101, 98, 114, 101, 119, 0, 105, 110, 100, 105, 97, 110, 0, 105, 115, 108, 97, 109, 105, 99, 0, 105, 115, 108, 97, 109, 105, 99, 45, 99, 105, 118, 105, 108, 0, 105, 115, 108, 97, 109, 105, 99, 45, 114, 103, 115, 97, 0, 105, 115, 108, 97, 109, 105, 99, 45, 116, 98, 108, 97, 0, 105, 115, 108, 97, 109, 105, 99, 45, 117, 109, 97, 108, 113, 117, 114, 97, 0, 105, 115, 111, 56, 54, 48, 49, 0, 106, 97, 112, 97, 110, 101, 115, 101, 0, 112, 101, 114, 115, 105, 97, 110, 0, 114, 111, 99, 0, 97, 99, 99, 111, 117, 110, 116, 0, 115, 116, 97, 110, 100, 97, 114, 100, 0, 110, 111, 110, 45, 105, 103, 110, 111, 114, 97, 98, 108, 101, 0, 115, 104, 105, 102, 116, 101, 100, 0, 121, 101, 115, 0, 108, 111, 119, 101, 114, 0, 117, 112, 112, 101, 114, 0, 98, 105, 103, 53, 104, 97, 110, 0, 99, 111, 109, 112, 97, 116, 0, 100, 105, 99, 116, 105, 111, 110, 97, 114, 121, 0, 100, 117, 99, 101, 116, 0, 101, 109, 111, 106, 105, 0, 103, 98, 50, 51, 49, 50, 104, 97, 110, 0, 112, 104, 111, 110, 101, 98, 111, 111, 107, 0, 112, 104, 111, 110, 101, 116, 105, 99, 0, 112, 105, 110, 121, 105, 110, 0, 114, 101, 102, 111, 114, 109, 101, 100, 0, 115, 101, 97, 114, 99, 104, 0, 115, 101, 97, 114, 99, 104, 106, 108, 0, 115, 116, 114, 111, 107, 101, 0, 116, 114, 97, 100, 105, 116, 105, 111, 110, 97, 108, 0, 117, 110, 105, 104, 97, 110, 0, 122, 104, 117, 121, 105, 110, 0, 82, 69, 79, 82, 68, 69, 82, 95, 67, 79, 68, 69, 0, 100, 105, 103, 105, 116, 0, 112, 117, 110, 99, 116, 0, 115, 112, 97, 99, 101, 0, 115, 121, 109, 98, 111, 108, 0, 105, 100, 101, 110, 116, 105, 99, 97, 108, 0, 112, 114, 105, 109, 97, 114, 121, 0, 115, 101, 99, 111, 110, 100, 97, 114, 121, 0, 116, 101, 114, 116, 105, 97, 114, 121, 0, 97, 100, 112, 0, 97, 101, 100, 0, 97, 102, 97, 0, 97, 102, 110, 0, 97, 108, 107, 0, 97, 109, 100, 0, 97, 110, 103, 0, 97, 111, 97, 0, 97, 111, 107, 0, 97, 111, 110, 0, 97, 111, 114, 0, 97, 114, 108, 0, 97, 114, 109, 0, 97, 114, 112, 0, 97, 114, 115, 0, 97, 116, 115, 0, 97, 117, 100, 0, 97, 119, 103, 0, 97, 122, 109, 0, 97, 122, 110, 0, 98, 97, 100, 0, 98, 97, 109, 0, 98, 97, 110, 0, 98, 98, 100, 0, 98, 100, 116, 0, 98, 101, 99, 0, 98, 101, 102, 0, 98, 101, 108, 0, 98, 103, 108, 0, 98, 103, 109, 0, 98, 103, 110, 0, 98, 103, 111, 0, 98, 104, 100, 0, 98, 105, 102, 0, 98, 109, 100, 0, 98, 110, 100, 0, 98, 111, 98, 0, 98, 111, 112, 0, 98, 111, 118, 0, 98, 114, 98, 0, 98, 114, 99, 0, 98, 114, 101, 0, 98, 114, 108, 0, 98, 114, 110, 0, 98, 114, 114, 0, 98, 114, 122, 0, 98, 115, 100, 0, 98, 116, 110, 0, 98, 117, 107, 0, 98, 119, 112, 0, 98, 121, 98, 0, 98, 121, 110, 0, 98, 121, 114, 0, 98, 122, 100, 0, 99, 97, 100, 0, 99, 100, 102, 0, 99, 104, 101, 0, 99, 104, 102, 0, 99, 104, 119, 0, 99, 108, 101, 0, 99, 108, 102, 0, 99, 108, 112, 0, 99, 110, 104, 0, 99, 110, 120, 0, 99, 110, 121, 0, 99, 111, 112, 0, 99, 111, 117, 0, 99, 114, 99, 0, 99, 115, 100, 0, 99, 115, 107, 0, 99, 117, 99, 0, 99, 117, 112, 0, 99, 118, 101, 0, 99, 121, 112, 0, 99, 122, 107, 0, 100, 100, 109, 0, 100, 101, 109, 0, 100, 106, 102, 0, 100, 107, 107, 0, 100, 111, 112, 0, 100, 122, 100, 0, 101, 99, 115, 0, 101, 99, 118, 0, 101, 101, 107, 0, 101, 103, 112, 0, 101, 114, 110, 0, 101, 115, 97, 0, 101, 115, 98, 0, 101, 115, 112, 0, 101, 116, 98, 0, 101, 117, 114, 0, 102, 105, 109, 0, 102, 106, 100, 0, 102, 107, 112, 0, 102, 114, 102, 0, 103, 98, 112, 0, 103, 101, 107, 0, 103, 101, 108, 0, 103, 104, 99, 0, 103, 104, 115, 0, 103, 105, 112, 0, 103, 109, 100, 0, 103, 110, 102, 0, 103, 110, 115, 0, 103, 113, 101, 0, 103, 114, 100, 0, 103, 116, 113, 0, 103, 119, 101, 0, 103, 119, 112, 0, 103, 121, 100, 0, 104, 107, 100, 0, 104, 110, 108, 0, 104, 114, 107, 0, 104, 116, 103, 0, 104, 117, 102, 0, 105, 100, 114, 0, 105, 101, 112, 0, 105, 108, 112, 0, 105, 108, 114, 0, 105, 108, 115, 0, 105, 110, 114, 0, 105, 113, 100, 0, 105, 114, 114, 0, 105, 115, 106, 0, 105, 115, 107, 0, 105, 116, 108, 0, 106, 109, 100, 0, 106, 111, 100, 0, 106, 112, 121, 0, 107, 101, 115, 0, 107, 103, 115, 0, 107, 104, 114, 0, 107, 109, 102, 0, 107, 112, 119, 0, 107, 114, 104, 0, 107, 114, 111, 0, 107, 114, 119, 0, 107, 119, 100, 0, 107, 121, 100, 0, 107, 122, 116, 0, 108, 97, 107, 0, 108, 98, 112, 0, 108, 107, 114, 0, 108, 114, 100, 0, 108, 115, 108, 0, 108, 116, 108, 0, 108, 116, 116, 0, 108, 117, 99, 0, 108, 117, 102, 0, 108, 117, 108, 0, 108, 118, 108, 0, 108, 118, 114, 0, 108, 121, 100, 0, 109, 97, 100, 0, 109, 97, 102, 0, 109, 99, 102, 0, 109, 100, 99, 0, 109, 100, 108, 0, 109, 103, 97, 0, 109, 103, 102, 0, 109, 107, 100, 0, 109, 107, 110, 0, 109, 108, 102, 0, 109, 109, 107, 0, 109, 110, 116, 0, 109, 111, 112, 0, 109, 114, 111, 0, 109, 114, 117, 0, 109, 116, 108, 0, 109, 116, 112, 0, 109, 117, 114, 0, 109, 118, 112, 0, 109, 118, 114, 0, 109, 119, 107, 0, 109, 120, 110, 0, 109, 120, 112, 0, 109, 120, 118, 0, 109, 121, 114, 0, 109, 122, 101, 0, 109, 122, 109, 0, 109, 122, 110, 0, 110, 97, 100, 0, 110, 103, 110, 0, 110, 105, 99, 0, 110, 105, 111, 0, 110, 108, 103, 0, 110, 111, 107, 0, 110, 112, 114, 0, 110, 122, 100, 0, 111, 109, 114, 0, 112, 97, 98, 0, 112, 101, 105, 0, 112, 101, 110, 0, 112, 101, 115, 0, 112, 103, 107, 0, 112, 104, 112, 0, 112, 107, 114, 0, 112, 108, 110, 0, 112, 108, 122, 0, 112, 116, 101, 0, 112, 121, 103, 0, 113, 97, 114, 0, 114, 104, 100, 0, 114, 111, 108, 0, 114, 111, 110, 0, 114, 115, 100, 0, 114, 117, 98, 0, 114, 117, 114, 0, 114, 119, 102, 0, 115, 97, 114, 0, 115, 98, 100, 0, 115, 99, 114, 0, 115, 100, 100, 0, 115, 100, 103, 0, 115, 100, 112, 0, 115, 101, 107, 0, 115, 103, 100, 0, 115, 104, 112, 0, 115, 105, 116, 0, 115, 107, 107, 0, 115, 108, 108, 0, 115, 111, 115, 0, 115, 114, 100, 0, 115, 114, 103, 0, 115, 115, 112, 0, 115, 116, 100, 0, 115, 116, 110, 0, 115, 117, 114, 0, 115, 118, 99, 0, 115, 121, 112, 0, 115, 122, 108, 0, 116, 104, 98, 0, 116, 106, 114, 0, 116, 106, 115, 0, 116, 109, 109, 0, 116, 109, 116, 0, 116, 110, 100, 0, 116, 112, 101, 0, 116, 114, 108, 0, 116, 114, 121, 0, 116, 116, 100, 0, 116, 119, 100, 0, 116, 122, 115, 0, 117, 97, 104, 0, 117, 97, 107, 0, 117, 103, 115, 0, 117, 103, 120, 0, 117, 115, 100, 0, 117, 115, 110, 0, 117, 115, 115, 0, 117, 121, 105, 0, 117, 121, 112, 0, 117, 121, 117, 0, 117, 121, 119, 0, 117, 122, 115, 0, 118, 101, 98, 0, 118, 101, 102, 0, 118, 101, 115, 0, 118, 110, 100, 0, 118, 110, 110, 0, 118, 117, 118, 0, 119, 115, 116, 0, 120, 97, 102, 0, 120, 97, 103, 0, 120, 97, 117, 0, 120, 98, 97, 0, 120, 98, 98, 0, 120, 98, 99, 0, 120, 98, 100, 0, 120, 99, 100, 0, 120, 100, 114, 0, 120, 101, 117, 0, 120, 102, 111, 0, 120, 102, 117, 0, 120, 111, 102, 0, 120, 112, 100, 0, 120, 112, 102, 0, 120, 112, 116, 0, 120, 114, 101, 0, 120, 115, 117, 0, 120, 116, 115, 0, 120, 117, 97, 0, 120, 120, 120, 0, 121, 100, 100, 0, 121, 101, 114, 0, 121, 117, 100, 0, 121, 117, 109, 0, 121, 117, 110, 0, 121, 117, 114, 0, 122, 97, 108, 0, 122, 97, 114, 0, 122, 109, 107, 0, 122, 109, 119, 0, 122, 114, 110, 0, 122, 114, 122, 0, 122, 119, 100, 0, 122, 119, 108, 0, 122, 119, 114, 0, 97, 99, 99, 101, 110, 116, 115, 0, 97, 115, 99, 105, 105, 0, 99, 97, 115, 101, 102, 111, 108, 100, 0, 102, 99, 99, 0, 102, 99, 100, 0, 102, 117, 108, 108, 119, 105, 100, 116, 104, 0, 104, 97, 108, 102, 119, 105, 100, 116, 104, 0, 104, 101, 120, 0, 110, 97, 109, 101, 0, 110, 102, 99, 0, 110, 102, 100, 0, 110, 102, 107, 99, 0, 110, 102, 107, 100, 0, 110, 117, 108, 108, 0, 110, 117, 109, 101, 114, 105, 99, 80, 105, 110, 121, 105, 110, 0, 112, 117, 98, 108, 105, 115, 104, 105, 110, 103, 0, 114, 101, 109, 111, 118, 101, 0, 116, 105, 116, 108, 101, 0, 122, 97, 119, 103, 121, 105, 0, 100, 101, 102, 97, 117, 108, 116, 0, 116, 101, 120, 116, 0, 102, 114, 105, 0, 109, 111, 110, 0, 115, 97, 116, 0, 115, 117, 110, 0, 116, 104, 117, 0, 116, 117, 101, 0, 119, 101, 100, 0, 104, 121, 98, 114, 105, 100, 0, 104, 49, 49, 0, 104, 49, 50, 0, 104, 50, 51, 0, 104, 50, 52, 0, 104, 97, 110, 100, 119, 114, 105, 116, 0, 119, 117, 98, 105, 0, 49, 48, 49, 107, 101, 121, 0, 49, 48, 50, 107, 101, 121, 0, 54, 48, 48, 100, 112, 105, 0, 55, 54, 56, 100, 112, 105, 0, 97, 110, 100, 114, 111, 105, 100, 0, 97, 122, 101, 114, 116, 121, 0, 99, 104, 114, 111, 109, 101, 111, 115, 0, 99, 111, 108, 101, 109, 97, 107, 0, 100, 118, 111, 114, 97, 107, 0, 100, 118, 111, 114, 97, 107, 108, 0, 100, 118, 111, 114, 97, 107, 114, 0, 101, 108, 50, 50, 48, 0, 101, 108, 51, 49, 57, 0, 101, 120, 116, 101, 110, 100, 101, 100, 0, 103, 111, 111, 103, 108, 101, 118, 107, 0, 105, 115, 105, 114, 105, 0, 108, 101, 103, 97, 99, 121, 0, 108, 116, 49, 50, 48, 53, 0, 108, 116, 49, 53, 56, 50, 0, 110, 117, 116, 97, 97, 113, 0, 111, 115, 120, 0, 112, 97, 116, 116, 97, 0, 113, 119, 101, 114, 116, 121, 0, 113, 119, 101, 114, 116, 122, 0, 116, 97, 57, 57, 0, 118, 97, 114, 0, 118, 105, 113, 114, 0, 119, 105, 110, 100, 111, 119, 115, 0, 108, 111, 111, 115, 101, 0, 110, 111, 114, 109, 97, 108, 0, 115, 116, 114, 105, 99, 116, 0, 98, 114, 101, 97, 107, 97, 108, 108, 0, 107, 101, 101, 112, 97, 108, 108, 0, 97, 108, 97, 108, 111, 99, 0, 98, 117, 99, 107, 119, 97, 108, 116, 0, 99, 115, 115, 0, 100, 105, 110, 0, 101, 119, 116, 115, 0, 103, 111, 115, 116, 0, 105, 97, 115, 116, 0, 105, 115, 111, 0, 106, 97, 118, 97, 0, 109, 99, 115, 116, 0, 109, 110, 115, 0, 110, 97, 109, 101, 115, 0, 112, 101, 114, 99, 101, 110, 116, 0, 112, 101, 114, 108, 0, 112, 108, 97, 105, 110, 0, 115, 97, 116, 116, 115, 0, 117, 110, 103, 101, 103, 110, 0, 117, 110, 105, 99, 111, 100, 101, 0, 120, 109, 108, 0, 120, 109, 108, 49, 48, 0, 105, 109, 112, 101, 114, 105, 97, 108, 0, 109, 101, 116, 114, 105, 99, 0, 117, 115, 115, 121, 115, 116, 101, 109, 0, 97, 100, 108, 109, 0, 97, 104, 111, 109, 0, 97, 114, 97, 98, 0, 97, 114, 97, 98, 101, 120, 116, 0, 97, 114, 109, 110, 0, 97, 114, 109, 110, 108, 111, 119, 0, 98, 97, 108, 105, 0, 98, 101, 110, 103, 0, 98, 104, 107, 115, 0, 98, 114, 97, 104, 0, 99, 97, 107, 109, 0, 99, 104, 97, 109, 0, 99, 121, 114, 108, 0, 100, 101, 118, 97, 0, 100, 105, 97, 107, 0, 101, 116, 104, 105, 0, 102, 105, 110, 97, 110, 99, 101, 0, 102, 117, 108, 108, 119, 105, 100, 101, 0, 103, 101, 111, 114, 0, 103, 111, 110, 103, 0, 103, 111, 110, 109, 0, 103, 114, 101, 107, 0, 103, 114, 101, 107, 108, 111, 119, 0, 103, 117, 106, 114, 0, 103, 117, 114, 117, 0, 104, 97, 110, 105, 100, 97, 121, 115, 0, 104, 97, 110, 105, 100, 101, 99, 0, 104, 97, 110, 115, 0, 104, 97, 110, 115, 102, 105, 110, 0, 104, 97, 110, 116, 0, 104, 97, 110, 116, 102, 105, 110, 0, 104, 101, 98, 114, 0, 104, 109, 110, 103, 0, 104, 109, 110, 112, 0, 106, 112, 97, 110, 0, 106, 112, 97, 110, 102, 105, 110, 0, 106, 112, 97, 110, 121, 101, 97, 114, 0, 107, 97, 108, 105, 0, 107, 104, 109, 114, 0, 107, 110, 100, 97, 0, 108, 97, 110, 97, 0, 108, 97, 110, 97, 116, 104, 97, 109, 0, 108, 97, 111, 111, 0, 108, 97, 116, 110, 0, 108, 101, 112, 99, 0, 108, 105, 109, 98, 0, 109, 97, 116, 104, 98, 111, 108, 100, 0, 109, 97, 116, 104, 100, 98, 108, 0, 109, 97, 116, 104, 109, 111, 110, 111, 0, 109, 97, 116, 104, 115, 97, 110, 98, 0, 109, 97, 116, 104, 115, 97, 110, 115, 0, 109, 108, 121, 109, 0, 109, 111, 100, 105, 0, 109, 111, 110, 103, 0, 109, 114, 111, 111, 0, 109, 116, 101, 105, 0, 109, 121, 109, 114, 0, 109, 121, 109, 114, 115, 104, 97, 110, 0, 109, 121, 109, 114, 116, 108, 110, 103, 0, 110, 97, 116, 105, 118, 101, 0, 110, 101, 119, 97, 0, 110, 107, 111, 111, 0, 111, 108, 99, 107, 0, 111, 114, 121, 97, 0, 111, 115, 109, 97, 0, 114, 111, 104, 103, 0, 114, 111, 109, 97, 110, 0, 114, 111, 109, 97, 110, 108, 111, 119, 0, 115, 97, 117, 114, 0, 115, 104, 114, 100, 0, 115, 105, 110, 100, 0, 115, 105, 110, 104, 0, 115, 111, 114, 97, 0, 115, 117, 110, 100, 0, 116, 97, 107, 114, 0, 116, 97, 108, 117, 0, 116, 97, 109, 108, 0, 116, 97, 109, 108, 100, 101, 99, 0, 116, 101, 108, 117, 0, 116, 104, 97, 105, 0, 116, 105, 98, 116, 0, 116, 105, 114, 104, 0, 118, 97, 105, 105, 0, 119, 97, 114, 97, 0, 119, 99, 104, 111, 0, 82, 71, 95, 75, 69, 89, 95, 86, 65, 76, 85, 69, 0, 83, 85, 66, 68, 73, 86, 73, 83, 73, 79, 78, 95, 67, 79, 68, 69, 0, 110, 111, 110, 101, 0, 112, 111, 115, 105, 120, 0, 67, 79, 68, 69, 80, 79, 73, 78, 84, 83, 0, 80, 82, 73, 86, 65, 84, 69, 95, 85, 83, 69, 0, 170, 170, 170, 0, 0, 99, 0, 97, 0, 0, 0, 99, 0, 111, 0, 0, 0, 99, 0, 117, 0, 0, 0, 104, 0, 99, 0, 0, 0, 107, 0, 97, 0, 0, 0, 107, 0, 98, 0, 0, 0, 107, 0, 99, 0, 0, 0, 107, 0, 102, 0, 0, 0, 107, 0, 104, 0, 0, 0, 107, 0, 107, 0, 0, 0, 107, 0, 110, 0, 0, 0, 107, 0, 114, 0, 0, 0, 107, 0, 115, 0, 0, 0, 109, 0, 115, 0, 0, 0, 110, 0, 117, 0, 0, 0, 116, 0, 122, 0, 0, 0, 118, 0, 116, 0, 0, 0, 97, 0, 110, 0, 121, 0, 0, 0, 99, 0, 49, 0, 49, 0, 0, 0, 116, 0, 114, 0, 117, 0, 101, 0, 0, 0, 100, 0, 105, 0, 99, 0, 116, 0, 0, 0, 116, 0, 114, 0, 97, 0, 100, 0, 0, 0, 102, 0, 97, 0, 108, 0, 115, 0, 101, 0, 0, 0, 102, 0, 119, 0, 105, 0, 100, 0, 116, 0, 104, 0, 0, 0, 103, 0, 98, 0, 50, 0, 51, 0, 49, 0, 50, 0, 0, 0, 104, 0, 119, 0, 105, 0, 100, 0, 116, 0, 104, 0, 0, 0, 108, 0, 101, 0, 118, 0, 101, 0, 108, 0, 49, 0, 0, 0, 108, 0, 101, 0, 118, 0, 101, 0, 108, 0, 50, 0, 0, 0, 108, 0, 101, 0, 118, 0, 101, 0, 108, 0, 51, 0, 0, 0, 108, 0, 101, 0, 118, 0, 101, 0, 108, 0, 52, 0, 0, 0, 115, 0, 105, 0, 110, 0, 103, 0, 108, 0, 101, 0, 0, 0, 110, 0, 112, 0, 105, 0, 110, 0, 121, 0, 105, 0, 110, 0, 0, 0, 112, 0, 117, 0, 98, 0, 108, 0, 105, 0, 115, 0, 104, 0, 0, 0, 101, 0, 116, 0, 104, 0, 105, 0, 111, 0, 97, 0, 97, 0, 0, 0, 103, 0, 114, 0, 101, 0, 103, 0, 111, 0, 114, 0, 121, 0, 0, 0, 105, 0, 100, 0, 101, 0, 110, 0, 116, 0, 105, 0, 99, 0, 0, 0, 112, 0, 104, 0, 111, 0, 110, 0, 101, 0, 98, 0, 107, 0, 0, 0, 109, 0, 117, 0, 108, 0, 116, 0, 105, 0, 112, 0, 108, 0, 101, 0, 0, 0, 99, 0, 104, 0, 97, 0, 114, 0, 110, 0, 97, 0, 109, 0, 101, 0, 0, 0, 110, 0, 111, 0, 105, 0, 103, 0, 110, 0, 111, 0, 114, 0, 101, 0, 0, 0, 112, 0, 114, 0, 112, 0, 114, 0, 110, 0, 97, 0, 109, 0, 101, 0, 0, 0, 116, 0, 114, 0, 97, 0, 100, 0, 105, 0, 116, 0, 105, 0, 111, 0, 0, 0, 117, 0, 107, 0, 115, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 0, 0, 113, 0, 117, 0, 97, 0, 116, 0, 101, 0, 114, 0, 110, 0, 97, 0, 114, 0, 121, 0, 0, 0, 105, 0, 110, 0, 99, 0, 114, 0, 101, 0, 109, 0, 101, 0, 110, 0, 116, 0, 97, 0, 108, 0, 0, 0, 105, 0, 115, 0, 108, 0, 97, 0, 109, 0, 105, 0, 99, 0, 45, 0, 99, 0, 105, 0, 118, 0, 105, 0, 108, 0, 0, 0, 1, 0, 48, 0, 6, 1, 2, 0, 76, 0, 79, 0, 60, 0, 60, 0, 5, 0, 45, 0, 92, 0, 74, 6, 79, 0, 95, 0, 250, 0, 130, 0, 185, 0, 185, 0, 52, 0, 35, 0, 105, 0, 166, 5, 114, 0, 127, 0, 140, 0, 153, 0, 166, 0, 188, 0, 198, 0, 215, 0, 226, 0, 237, 0, 249, 0, 2, 1, 28, 10, 5, 1, 92, 0, 8, 1, 14, 1, 17, 1, 20, 1, 23, 1, 26, 1, 29, 1, 32, 1, 40, 1, 182, 6, 48, 1, 22, 7, 30, 7, 51, 1, 54, 1, 104, 10, 63, 1, 95, 0, 1, 0, 0, 0, 13, 0, 16, 0, 22, 0, 19, 0, 25, 0, 4, 0, 28, 0, 31, 0, 34, 0, 37, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 40, 0, 43, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 46, 0, 0, 0, 49, 0, 0, 0, 1, 0, 85, 1, 239, 0, 1, 0, 109, 1, 60, 0, 1, 0, 116, 1, 60, 0, 18, 0, 130, 1, 139, 1, 147, 1, 154, 1, 160, 1, 169, 1, 189, 1, 199, 1, 206, 1, 213, 1, 221, 1, 235, 1, 248, 1, 5, 2, 22, 2, 30, 2, 39, 2, 47, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 153, 0, 161, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 51, 2, 59, 2, 0, 0, 0, 0, 2, 0, 68, 2, 82, 2, 203, 0, 0, 0, 2, 0, 68, 11, 90, 2, 75, 0, 60, 0, 3, 0, 94, 2, 68, 11, 100, 2, 0, 0, 75, 0, 0, 0, 2, 0, 68, 11, 90, 2, 75, 0, 60, 0, 2, 0, 68, 11, 90, 2, 75, 0, 60, 0, 19, 0, 106, 2, 114, 2, 121, 2, 109, 1, 132, 2, 138, 2, 135, 10, 144, 2, 154, 2, 164, 2, 173, 2, 180, 2, 189, 2, 196, 2, 59, 2, 205, 2, 212, 2, 224, 2, 231, 2, 0, 0, 0, 0, 65, 0, 0, 0, 0, 0, 0, 0, 0, 0, 88, 0, 177, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 70, 0, 0, 0, 0, 0, 2, 0, 68, 11, 90, 2, 75, 0, 60, 0, 2, 0, 68, 11, 90, 2, 75, 0, 60, 0, 6, 0, 238, 2, 249, 0, 251, 2, 1, 3, 7, 3, 13, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 20, 3, 30, 3, 177, 0, 38, 3, 48, 3, 169, 0, 102, 0, 123, 0, 109, 0, 116, 0, 47, 1, 57, 3, 61, 3, 65, 3, 69, 3, 73, 3, 145, 9, 77, 3, 81, 3, 85, 3, 89, 3, 93, 3, 97, 3, 12, 12, 101, 3, 105, 3, 109, 3, 113, 3, 117, 3, 121, 3, 125, 3, 129, 3, 133, 3, 137, 3, 141, 3, 145, 3, 149, 3, 153, 3, 157, 3, 161, 3, 165, 3, 169, 3, 173, 3, 177, 3, 181, 3, 185, 3, 189, 3, 193, 3, 197, 3, 201, 3, 16, 3, 205, 3, 209, 3, 213, 3, 217, 3, 221, 3, 225, 3, 229, 3, 233, 3, 237, 3, 241, 3, 245, 3, 249, 3, 253, 3, 1, 4, 5, 4, 9, 4, 13, 4, 17, 4, 21, 4, 25, 4, 29, 4, 33, 4, 37, 4, 41, 4, 45, 4, 49, 4, 53, 4, 57, 4, 61, 4, 65, 4, 69, 4, 73, 4, 77, 4, 81, 4, 85, 4, 89, 4, 93, 4, 97, 4, 101, 4, 105, 4, 109, 4, 113, 4, 117, 4, 121, 4, 125, 4, 129, 4, 133, 4, 137, 4, 141, 4, 145, 4, 149, 4, 153, 4, 157, 4, 161, 4, 165, 4, 169, 4, 173, 4, 177, 4, 181, 4, 185, 4, 189, 4, 193, 4, 197, 4, 201, 4, 205, 4, 209, 4, 213, 4, 217, 4, 221, 4, 225, 4, 229, 4, 233, 4, 237, 4, 241, 4, 245, 4, 195, 11, 249, 4, 253, 4, 1, 5, 5, 5, 9, 5, 13, 5, 17, 5, 21, 5, 25, 5, 29, 5, 33, 5, 37, 5, 41, 5, 45, 5, 49, 5, 53, 5, 57, 5, 61, 5, 65, 5, 69, 5, 73, 5, 77, 5, 81, 5, 85, 5, 89, 5, 93, 5, 97, 5, 101, 5, 105, 5, 109, 5, 113, 5, 117, 5, 121, 5, 125, 5, 129, 5, 133, 5, 137, 5, 141, 5, 145, 5, 149, 5, 153, 5, 157, 5, 161, 5, 165, 5, 169, 5, 173, 5, 177, 5, 181, 5, 185, 5, 189, 5, 193, 5, 197, 5, 201, 5, 205, 5, 209, 5, 213, 5, 217, 5, 221, 5, 225, 5, 229, 5, 233, 5, 237, 5, 241, 5, 245, 5, 249, 5, 253, 5, 1, 6, 5, 6, 9, 6, 13, 6, 17, 6, 21, 6, 25, 6, 29, 6, 33, 6, 37, 6, 41, 6, 45, 6, 49, 6, 53, 6, 57, 6, 61, 6, 65, 6, 69, 6, 73, 6, 77, 6, 81, 6, 85, 6, 89, 6, 93, 6, 97, 6, 101, 6, 105, 6, 109, 6, 113, 6, 117, 6, 121, 6, 125, 6, 129, 6, 133, 6, 137, 6, 141, 6, 145, 6, 149, 6, 153, 6, 157, 6, 161, 6, 165, 6, 169, 6, 173, 6, 177, 6, 181, 6, 185, 6, 189, 6, 193, 6, 197, 6, 201, 6, 205, 6, 209, 6, 213, 6, 217, 6, 221, 6, 225, 6, 229, 6, 233, 6, 71, 1, 237, 6, 241, 6, 245, 6, 249, 6, 253, 6, 1, 7, 5, 7, 9, 7, 13, 7, 17, 7, 21, 7, 25, 7, 29, 7, 33, 7, 37, 7, 41, 7, 45, 7, 49, 7, 53, 7, 57, 7, 61, 7, 65, 7, 69, 7, 73, 7, 77, 7, 81, 7, 85, 7, 89, 7, 93, 7, 97, 7, 101, 7, 105, 7, 109, 7, 113, 7, 117, 7, 121, 7, 125, 7, 129, 7, 133, 7, 137, 7, 141, 7, 145, 7, 149, 7, 153, 7, 157, 7, 161, 7, 165, 7, 169, 7, 173, 7, 177, 7, 181, 7, 185, 7, 189, 7, 193, 7, 197, 7, 201, 7, 205, 7, 209, 7, 213, 7, 217, 7, 221, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 22, 0, 225, 7, 233, 7, 239, 7, 251, 2, 248, 7, 252, 7, 0, 8, 10, 8, 20, 8, 94, 2, 24, 8, 29, 8, 33, 8, 37, 8, 42, 8, 47, 8, 52, 8, 66, 8, 77, 8, 84, 8, 100, 2, 90, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 81, 0, 95, 0, 0, 0, 0, 0, 194, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 137, 0, 145, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 97, 8, 138, 2, 105, 8, 0, 0, 0, 0, 0, 0, 7, 0, 110, 8, 114, 8, 118, 8, 122, 8, 126, 8, 130, 8, 134, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 138, 8, 0, 0, 4, 0, 145, 8, 149, 8, 153, 8, 157, 8, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 161, 8, 173, 2, 215, 11, 170, 8, 0, 0, 0, 0, 0, 0, 0, 0, 29, 0, 175, 8, 182, 8, 189, 8, 196, 8, 203, 8, 211, 8, 218, 8, 227, 8, 235, 8, 242, 8, 250, 8, 2, 9, 8, 9, 14, 9, 23, 9, 32, 9, 38, 9, 45, 9, 52, 9, 59, 9, 66, 9, 70, 9, 76, 9, 83, 9, 90, 9, 215, 11, 95, 9, 99, 9, 104, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 249, 0, 1, 3, 7, 3, 13, 3, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 112, 9, 118, 9, 125, 9, 0, 0, 0, 0, 0, 0, 3, 0, 132, 9, 141, 9, 118, 9, 0, 0, 0, 0, 0, 0, 22, 0, 149, 9, 177, 3, 156, 9, 203, 6, 165, 9, 169, 9, 173, 9, 178, 9, 183, 9, 188, 9, 192, 9, 197, 9, 202, 9, 206, 9, 212, 9, 220, 9, 225, 9, 231, 9, 237, 9, 244, 9, 252, 9, 0, 10, 0, 0, 0, 0, 0, 0, 56, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 212, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 6, 10, 15, 10, 22, 10, 230, 0, 0, 0, 0, 0, 87, 0, 31, 10, 36, 10, 41, 10, 46, 10, 54, 10, 59, 10, 67, 10, 72, 10, 77, 10, 82, 10, 87, 10, 92, 10, 97, 10, 102, 10, 107, 10, 112, 10, 117, 10, 125, 10, 134, 10, 139, 10, 144, 10, 149, 10, 154, 10, 162, 10, 167, 10, 172, 10, 181, 10, 189, 10, 194, 10, 202, 10, 207, 10, 215, 10, 220, 10, 225, 10, 192, 9, 230, 10, 235, 10, 243, 10, 252, 10, 1, 11, 6, 11, 11, 11, 16, 11, 25, 11, 30, 11, 35, 11, 40, 11, 45, 11, 54, 11, 62, 11, 71, 11, 80, 11, 89, 11, 94, 11, 99, 11, 104, 11, 109, 11, 114, 11, 119, 11, 128, 11, 137, 11, 144, 11, 149, 11, 154, 11, 159, 11, 164, 11, 169, 11, 174, 11, 180, 11, 189, 11, 194, 11, 199, 11, 204, 11, 209, 11, 214, 11, 219, 11, 224, 11, 229, 11, 234, 11, 242, 11, 247, 11, 252, 11, 1, 12, 212, 2, 6, 12, 11, 12, 16, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 221, 0, 0, 0, 0, 0, 0, 0, 1, 0, 21, 12, 0, 0, 6, 0, 225, 7, 233, 7, 20, 8, 52, 8, 66, 8, 90, 8, 0, 0, 0, 0, 0, 0, 137, 0, 145, 0, 0, 0, 1, 0, 34, 12, 0, 0, 2, 0, 51, 12, 59, 2, 0, 0, 0, 0, 1, 0, 215, 11, 0, 0, 1, 0, 56, 12, 0, 0, 1, 0, 62, 12, 0, 0, 1, 0, 73, 12, 0, 0, 170, 170, 38, 0, 0, 0, 47, 0, 73, 0, 67, 0, 85, 0, 68, 0, 65, 0, 84, 0, 65, 0, 47, 0, 116, 0, 105, 0, 109, 0, 101, 0, 122, 0, 111, 0, 110, 0, 101, 0, 84, 0, 121, 0, 112, 0, 101, 0, 115, 0, 47, 0, 98, 0, 99, 0, 112, 0, 84, 0, 121, 0, 112, 0, 101, 0, 65, 0, 108, 0, 105, 0, 97, 0, 115, 0, 47, 0, 116, 0, 122, 0, 0, 0, 170, 170, 2, 0, 45, 0, 87, 9, 170, 170, 20, 1, 0, 80, 33, 6, 0, 48, 2, 0, 65, 0, 82, 0, 170, 170, 23, 1, 0, 80, 28, 1, 0, 80, 41, 0, 0, 0, 47, 0, 73, 0, 67, 0, 85, 0, 68, 0, 65, 0, 84, 0, 65, 0, 47, 0, 116, 0, 105, 0, 109, 0, 101, 0, 122, 0, 111, 0, 110, 0, 101, 0, 84, 0, 121, 0, 112, 0, 101, 0, 115, 0, 47, 0, 116, 0, 121, 0, 112, 0, 101, 0, 65, 0, 108, 0, 105, 0, 97, 0, 115, 0, 47, 0, 116, 0, 105, 0, 109, 0, 101, 0, 122, 0, 111, 0, 110, 0, 101, 0, 0, 0, 2, 0, 237, 0, 54, 1, 170, 170, 110, 1, 0, 80, 62, 6, 0, 48, 2, 0, 106, 1, 87, 9, 170, 170, 113, 1, 0, 80, 116, 1, 0, 80, 1, 0, 65, 0, 88, 6, 0, 32, 39, 0, 0, 0, 47, 0, 73, 0, 67, 0, 85, 0, 68, 0, 65, 0, 84, 0, 65, 0, 47, 0, 116, 0, 105, 0, 109, 0, 101, 0, 122, 0, 111, 0, 110, 0, 101, 0, 84, 0, 121, 0, 112, 0, 101, 0, 115, 0, 47, 0, 116, 0, 121, 0, 112, 0, 101, 0, 77, 0, 97, 0, 112, 0, 47, 0, 116, 0, 105, 0, 109, 0, 101, 0, 122, 0, 111, 0, 110, 0, 101, 0, 0, 0, 35, 0, 105, 0, 166, 5, 114, 0, 127, 0, 140, 0, 153, 0, 166, 0, 188, 0, 198, 0, 215, 0, 226, 0, 237, 0, 249, 0, 2, 1, 28, 10, 5, 1, 92, 0, 8, 1, 14, 1, 17, 1, 20, 1, 23, 1, 26, 1, 29, 1, 32, 1, 40, 1, 182, 6, 48, 1, 22, 7, 30, 7, 51, 1, 54, 1, 104, 10, 63, 1, 95, 0, 119, 1, 0, 80, 156, 1, 0, 80, 161, 1, 0, 80, 166, 1, 0, 80, 171, 1, 0, 80, 178, 1, 0, 80, 183, 1, 0, 80, 188, 1, 0, 80, 227, 1, 0, 80, 232, 1, 0, 80, 237, 1, 0, 80, 250, 1, 0, 80, 5, 2, 0, 80, 100, 4, 0, 80, 145, 4, 0, 80, 152, 4, 0, 80, 167, 4, 0, 80, 170, 4, 0, 80, 179, 4, 0, 80, 188, 4, 0, 80, 247, 4, 0, 80, 0, 5, 0, 80, 7, 5, 0, 80, 14, 5, 0, 80, 59, 5, 0, 80, 66, 5, 0, 80, 241, 5, 0, 80, 244, 5, 0, 80, 1, 6, 0, 80, 4, 6, 0, 80, 9, 6, 0, 80, 94, 6, 0, 48, 12, 6, 0, 80, 15, 6, 0, 80, 18, 6, 0, 80, 6, 0, 32, 0, 57, 0, 98, 0, 75, 1, 97, 1, 122, 1, 170, 170, 54, 6, 0, 32, 58, 6, 0, 32, 39, 1, 0, 80, 84, 6, 0, 32, 92, 6, 0, 32, 115, 6, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 195, 37, 0, 32, 7, 0, 0, 0, 34, 0, 0, 0, 199, 37, 0, 0, 199, 37, 0, 0, 5, 0, 0, 0, 1, 0, 0, 0, 41, 24, 0, 0, 108, 105, 107, 101, 108, 121, 0, 108, 97, 110, 103, 117, 97, 103, 101, 65, 108, 105, 97, 115, 101, 115, 0, 114, 101, 103, 105, 111, 110, 65, 108, 105, 97, 115, 101, 115, 0, 116, 114, 105, 101, 0, 108, 115, 114, 115, 0, 109, 97, 116, 99, 104, 0, 114, 101, 103, 105, 111, 110, 84, 111, 80, 97, 114, 116, 105, 116, 105, 111, 110, 115, 0, 112, 97, 114, 116, 105, 116, 105, 111, 110, 115, 0, 112, 97, 114, 97, 100, 105, 103, 109, 115, 0, 100, 105, 115, 116, 97, 110, 99, 101, 115, 0, 170, 0, 0, 46, 0, 0, 0, 90, 0, 90, 0, 0, 0, 73, 0, 78, 0, 0, 0, 73, 0, 68, 0, 0, 0, 82, 0, 85, 0, 0, 0, 67, 0, 78, 0, 0, 0, 67, 0, 77, 0, 0, 0, 80, 0, 72, 0, 0, 0, 73, 0, 82, 0, 0, 0, 78, 0, 80, 0, 0, 0, 80, 0, 75, 0, 0, 0, 67, 0, 65, 0, 0, 0, 85, 0, 83, 0, 0, 0, 78, 0, 71, 0, 0, 0, 84, 0, 82, 0, 0, 0, 84, 0, 90, 0, 0, 0, 68, 0, 69, 0, 0, 0, 73, 0, 84, 0, 0, 0, 75, 0, 69, 0, 0, 0, 71, 0, 66, 0, 0, 0, 90, 0, 65, 0, 0, 0, 77, 0, 76, 0, 0, 0, 83, 0, 68, 0, 0, 0, 84, 0, 72, 0, 0, 0, 85, 0, 71, 0, 0, 0, 69, 0, 84, 0, 0, 0, 70, 0, 82, 0, 0, 0, 77, 0, 77, 0, 0, 0, 77, 0, 89, 0, 0, 0, 67, 0, 68, 0, 0, 0, 69, 0, 83, 0, 0, 0, 71, 0, 72, 0, 0, 0, 77, 0, 90, 0, 0, 0, 70, 0, 77, 0, 0, 0, 73, 0, 81, 0, 0, 0, 77, 0, 65, 0, 0, 0, 77, 0, 88, 0, 0, 0, 83, 0, 69, 0, 0, 0, 83, 0, 78, 0, 0, 0, 65, 0, 70, 0, 0, 0, 66, 0, 68, 0, 0, 0, 66, 0, 82, 0, 0, 0, 67, 0, 73, 0, 0, 0, 71, 0, 69, 0, 0, 0, 71, 0, 78, 0, 0, 0, 71, 0, 82, 0, 0, 0, 77, 0, 78, 0, 0, 0, 78, 0, 69, 0, 0, 0, 78, 0, 76, 0, 0, 0, 85, 0, 65, 0, 0, 0, 67, 0, 72, 0, 0, 0, 78, 0, 65, 0, 0, 0, 84, 0, 71, 0, 0, 0, 85, 0, 90, 0, 0, 0, 86, 0, 78, 0, 0, 0, 90, 0, 87, 0, 0, 0, 110, 0, 108, 0, 0, 0, 65, 0, 90, 0, 0, 0, 66, 0, 70, 0, 0, 0, 68, 0, 90, 0, 0, 0, 69, 0, 71, 0, 0, 0, 69, 0, 82, 0, 0, 0, 70, 0, 73, 0, 0, 0, 70, 0, 74, 0, 0, 0, 73, 0, 76, 0, 0, 0, 74, 0, 80, 0, 0, 0, 75, 0, 77, 0, 0, 0, 76, 0, 65, 0, 0, 0, 78, 0, 79, 0, 0, 0, 80, 0, 76, 0, 0, 0, 82, 0, 83, 0, 0, 0, 83, 0, 65, 0, 0, 0, 83, 0, 76, 0, 0, 0, 83, 0, 89, 0, 0, 0, 84, 0, 87, 0, 0, 0, 85, 0, 77, 0, 0, 0, 89, 0, 69, 0, 0, 0, 65, 0, 76, 0, 0, 0, 65, 0, 79, 0, 0, 0, 65, 0, 81, 0, 0, 0, 65, 0, 85, 0, 0, 0, 66, 0, 69, 0, 0, 0, 66, 0, 74, 0, 0, 0, 66, 0, 79, 0, 0, 0, 66, 0, 84, 0, 0, 0, 67, 0, 87, 0, 0, 0, 67, 0, 89, 0, 0, 0, 71, 0, 70, 0, 0, 0, 71, 0, 81, 0, 0, 0, 75, 0, 72, 0, 0, 0, 75, 0, 82, 0, 0, 0, 75, 0, 90, 0, 0, 0, 76, 0, 66, 0, 0, 0, 76, 0, 82, 0, 0, 0, 77, 0, 68, 0, 0, 0, 77, 0, 85, 0, 0, 0, 77, 0, 87, 0, 0, 0, 80, 0, 65, 0, 0, 0, 80, 0, 70, 0, 0, 0, 80, 0, 71, 0, 0, 0, 82, 0, 79, 0, 0, 0, 83, 0, 82, 0, 0, 0, 84, 0, 76, 0, 0, 0, 84, 0, 78, 0, 0, 0, 87, 0, 70, 0, 0, 0, 88, 0, 75, 0, 0, 0, 89, 0, 84, 0, 0, 0, 102, 0, 102, 0, 0, 0, 65, 0, 77, 0, 0, 0, 65, 0, 84, 0, 0, 0, 65, 0, 87, 0, 0, 0, 66, 0, 65, 0, 0, 0, 66, 0, 71, 0, 0, 0, 66, 0, 78, 0, 0, 0, 67, 0, 70, 0, 0, 0, 67, 0, 76, 0, 0, 0, 67, 0, 79, 0, 0, 0, 67, 0, 86, 0, 0, 0, 67, 0, 90, 0, 0, 0, 68, 0, 75, 0, 0, 0, 69, 0, 67, 0, 0, 0, 69, 0, 69, 0, 0, 0, 71, 0, 65, 0, 0, 0, 71, 0, 84, 0, 0, 0, 71, 0, 85, 0, 0, 0, 71, 0, 87, 0, 0, 0, 72, 0, 75, 0, 0, 0, 72, 0, 85, 0, 0, 0, 73, 0, 69, 0, 0, 0, 74, 0, 79, 0, 0, 0, 75, 0, 73, 0, 0, 0, 76, 0, 84, 0, 0, 0, 76, 0, 85, 0, 0, 0, 76, 0, 86, 0, 0, 0, 77, 0, 75, 0, 0, 0, 77, 0, 79, 0, 0, 0, 77, 0, 82, 0, 0, 0, 77, 0, 86, 0, 0, 0, 80, 0, 69, 0, 0, 0, 80, 0, 84, 0, 0, 0, 82, 0, 69, 0, 0, 0, 83, 0, 67, 0, 0, 0, 83, 0, 75, 0, 0, 0, 83, 0, 79, 0, 0, 0, 84, 0, 74, 0, 0, 0, 86, 0, 65, 0, 0, 0, 86, 0, 85, 0, 0, 0, 90, 0, 77, 0, 0, 0, 100, 0, 118, 0, 0, 0, 106, 0, 119, 0, 0, 0, 108, 0, 118, 0, 0, 0, 115, 0, 99, 0, 0, 0, 115, 0, 118, 0, 0, 0, 116, 0, 108, 0, 0, 0, 48, 0, 51, 0, 0, 0, 48, 0, 53, 0, 0, 0, 48, 0, 50, 0, 0, 0, 48, 0, 52, 0, 0, 0, 49, 0, 54, 0, 0, 0, 65, 0, 68, 0, 0, 0, 65, 0, 69, 0, 0, 0, 65, 0, 78, 0, 0, 0, 65, 0, 82, 0, 0, 0, 65, 0, 83, 0, 0, 0, 65, 0, 88, 0, 0, 0, 66, 0, 72, 0, 0, 0, 66, 0, 73, 0, 0, 0, 66, 0, 76, 0, 0, 0, 66, 0, 81, 0, 0, 0, 66, 0, 85, 0, 0, 0, 66, 0, 86, 0, 0, 0, 66, 0, 89, 0, 0, 0, 67, 0, 67, 0, 0, 0, 67, 0, 71, 0, 0, 0, 67, 0, 80, 0, 0, 0, 67, 0, 82, 0, 0, 0, 67, 0, 83, 0, 0, 0, 67, 0, 84, 0, 0, 0, 67, 0, 85, 0, 0, 0, 68, 0, 68, 0, 0, 0, 68, 0, 71, 0, 0, 0, 68, 0, 74, 0, 0, 0, 68, 0, 79, 0, 0, 0, 68, 0, 89, 0, 0, 0, 69, 0, 65, 0, 0, 0, 69, 0, 72, 0, 0, 0, 69, 0, 85, 0, 0, 0, 69, 0, 90, 0, 0, 0, 70, 0, 79, 0, 0, 0, 70, 0, 81, 0, 0, 0, 70, 0, 88, 0, 0, 0, 71, 0, 76, 0, 0, 0, 71, 0, 77, 0, 0, 0, 71, 0, 80, 0, 0, 0, 71, 0, 83, 0, 0, 0, 72, 0, 77, 0, 0, 0, 72, 0, 78, 0, 0, 0, 72, 0, 82, 0, 0, 0, 72, 0, 84, 0, 0, 0, 72, 0, 86, 0, 0, 0, 73, 0, 67, 0, 0, 0, 73, 0, 77, 0, 0, 0, 73, 0, 83, 0, 0, 0, 74, 0, 69, 0, 0, 0, 74, 0, 77, 0, 0, 0, 74, 0, 84, 0, 0, 0, 75, 0, 71, 0, 0, 0, 75, 0, 80, 0, 0, 0, 75, 0, 87, 0, 0, 0, 76, 0, 73, 0, 0, 0, 76, 0, 75, 0, 0, 0, 76, 0, 83, 0, 0, 0, 76, 0, 89, 0, 0, 0, 77, 0, 67, 0, 0, 0, 77, 0, 69, 0, 0, 0, 77, 0, 70, 0, 0, 0, 77, 0, 71, 0, 0, 0, 77, 0, 72, 0, 0, 0, 77, 0, 73, 0, 0, 0, 77, 0, 81, 0, 0, 0, 77, 0, 84, 0, 0, 0, 78, 0, 67, 0, 0, 0, 78, 0, 72, 0, 0, 0, 78, 0, 73, 0, 0, 0, 78, 0, 81, 0, 0, 0, 78, 0, 82, 0, 0, 0, 78, 0, 84, 0, 0, 0, 78, 0, 85, 0, 0, 0, 78, 0, 90, 0, 0, 0, 79, 0, 77, 0, 0, 0, 80, 0, 67, 0, 0, 0, 80, 0, 77, 0, 0, 0, 80, 0, 82, 0, 0, 0, 80, 0, 83, 0, 0, 0, 80, 0, 85, 0, 0, 0, 80, 0, 87, 0, 0, 0, 80, 0, 89, 0, 0, 0, 80, 0, 90, 0, 0, 0, 81, 0, 65, 0, 0, 0, 81, 0, 85, 0, 0, 0, 82, 0, 72, 0, 0, 0, 82, 0, 87, 0, 0, 0, 83, 0, 66, 0, 0, 0, 83, 0, 73, 0, 0, 0, 83, 0, 74, 0, 0, 0, 83, 0, 77, 0, 0, 0, 83, 0, 83, 0, 0, 0, 83, 0, 84, 0, 0, 0, 83, 0, 85, 0, 0, 0, 83, 0, 86, 0, 0, 0, 83, 0, 88, 0, 0, 0, 84, 0, 68, 0, 0, 0, 84, 0, 70, 0, 0, 0, 84, 0, 75, 0, 0, 0, 84, 0, 77, 0, 0, 0, 84, 0, 79, 0, 0, 0, 84, 0, 80, 0, 0, 0, 84, 0, 86, 0, 0, 0, 85, 0, 75, 0, 0, 0, 85, 0, 89, 0, 0, 0, 86, 0, 68, 0, 0, 0, 86, 0, 69, 0, 0, 0, 87, 0, 75, 0, 0, 0, 87, 0, 83, 0, 0, 0, 89, 0, 68, 0, 0, 0, 89, 0, 85, 0, 0, 0, 90, 0, 82, 0, 0, 0, 98, 0, 110, 0, 0, 0, 98, 0, 115, 0, 0, 0, 99, 0, 118, 0, 0, 0, 102, 0, 106, 0, 0, 0, 110, 0, 118, 0, 0, 0, 116, 0, 119, 0, 0, 0, 118, 0, 101, 0, 0, 0, 118, 0, 105, 0, 0, 0, 118, 0, 111, 0, 0, 0, 122, 0, 117, 0, 0, 0, 110, 0, 102, 0, 114, 0, 0, 0, 120, 0, 101, 0, 115, 0, 0, 0, 108, 0, 122, 0, 104, 0, 0, 0, 119, 0, 97, 0, 114, 0, 0, 0, 109, 0, 101, 0, 110, 0, 0, 0, 48, 0, 48, 0, 49, 0, 0, 0, 98, 0, 107, 0, 117, 0, 0, 0, 103, 0, 121, 0, 105, 0, 0, 0, 120, 0, 115, 0, 114, 0, 0, 0, 98, 0, 115, 0, 113, 0, 0, 0, 109, 0, 100, 0, 101, 0, 0, 0, 115, 0, 109, 0, 115, 0, 0, 0, 120, 0, 115, 0, 97, 0, 0, 0, 121, 0, 97, 0, 122, 0, 0, 0, 100, 0, 116, 0, 112, 0, 0, 0, 107, 0, 104, 0, 97, 0, 0, 0, 109, 0, 102, 0, 97, 0, 0, 0, 116, 0, 107, 0, 107, 0, 0, 0, 116, 0, 115, 0, 119, 0, 0, 0, 98, 0, 104, 0, 111, 0, 0, 0, 100, 0, 111, 0, 105, 0, 0, 0, 101, 0, 109, 0, 107, 0, 0, 0, 109, 0, 97, 0, 110, 0, 0, 0, 113, 0, 117, 0, 122, 0, 0, 0, 116, 0, 115, 0, 100, 0, 0, 0, 118, 0, 114, 0, 111, 0, 0, 0, 120, 0, 109, 0, 110, 0, 0, 0, 121, 0, 107, 0, 111, 0, 0, 0, 109, 0, 105, 0, 115, 0, 0, 0, 112, 0, 112, 0, 97, 0, 0, 0, 122, 0, 122, 0, 97, 0, 0, 0, 99, 0, 106, 0, 118, 0, 0, 0, 101, 0, 107, 0, 121, 0, 0, 0, 102, 0, 105, 0, 116, 0, 0, 0, 110, 0, 97, 0, 107, 0, 0, 0, 111, 0, 107, 0, 118, 0, 0, 0, 112, 0, 110, 0, 98, 0, 0, 0, 115, 0, 116, 0, 107, 0, 0, 0, 115, 0, 117, 0, 107, 0, 0, 0, 115, 0, 117, 0, 114, 0, 0, 0, 116, 0, 98, 0, 111, 0, 0, 0, 119, 0, 101, 0, 108, 0, 0, 0, 119, 0, 106, 0, 97, 0, 0, 0, 121, 0, 117, 0, 103, 0, 0, 0, 122, 0, 110, 0, 101, 0, 0, 0, 102, 0, 105, 0, 108, 0, 0, 0, 107, 0, 100, 0, 116, 0, 0, 0, 111, 0, 121, 0, 98, 0, 0, 0, 112, 0, 97, 0, 112, 0, 0, 0, 118, 0, 97, 0, 106, 0, 0, 0, 104, 0, 111, 0, 99, 0, 0, 0, 114, 0, 107, 0, 105, 0, 0, 0, 116, 0, 109, 0, 104, 0, 0, 0, 121, 0, 97, 0, 109, 0, 0, 0, 99, 0, 113, 0, 117, 0, 0, 0, 104, 0, 104, 0, 121, 0, 0, 0, 104, 0, 110, 0, 111, 0, 0, 0, 107, 0, 97, 0, 97, 0, 0, 0, 107, 0, 100, 0, 122, 0, 0, 0, 108, 0, 111, 0, 114, 0, 0, 0, 108, 0, 116, 0, 103, 0, 0, 0, 109, 0, 101, 0, 117, 0, 0, 0, 109, 0, 105, 0, 119, 0, 0, 0, 110, 0, 99, 0, 117, 0, 0, 0, 110, 0, 104, 0, 101, 0, 0, 0, 110, 0, 105, 0, 110, 0, 0, 0, 110, 0, 107, 0, 103, 0, 0, 0, 110, 0, 109, 0, 103, 0, 0, 0, 110, 0, 115, 0, 111, 0, 0, 0, 112, 0, 107, 0, 97, 0, 0, 0, 114, 0, 103, 0, 110, 0, 0, 0, 115, 0, 97, 0, 102, 0, 0, 0, 115, 0, 99, 0, 97, 0, 0, 0, 115, 0, 99, 0, 115, 0, 0, 0, 115, 0, 105, 0, 100, 0, 0, 0, 115, 0, 107, 0, 115, 0, 0, 0, 115, 0, 112, 0, 115, 0, 0, 0, 116, 0, 98, 0, 103, 0, 0, 0, 116, 0, 99, 0, 121, 0, 0, 0, 116, 0, 101, 0, 116, 0, 0, 0, 116, 0, 104, 0, 114, 0, 0, 0, 116, 0, 105, 0, 107, 0, 0, 0, 116, 0, 107, 0, 114, 0, 0, 0, 116, 0, 109, 0, 121, 0, 0, 0, 117, 0, 116, 0, 114, 0, 0, 0, 119, 0, 105, 0, 117, 0, 0, 0, 119, 0, 115, 0, 107, 0, 0, 0, 120, 0, 99, 0, 114, 0, 0, 0, 121, 0, 97, 0, 121, 0, 0, 0, 122, 0, 109, 0, 105, 0, 0, 0, 122, 0, 111, 0, 109, 0, 0, 0, 122, 0, 115, 0, 109, 0, 0, 0, 52, 0, 49, 0, 57, 0, 0, 0, 98, 0, 97, 0, 108, 0, 0, 0, 98, 0, 97, 0, 110, 0, 0, 0, 98, 0, 98, 0, 99, 0, 0, 0, 98, 0, 102, 0, 121, 0, 0, 0, 98, 0, 105, 0, 107, 0, 0, 0, 98, 0, 117, 0, 97, 0, 0, 0, 98, 0, 117, 0, 103, 0, 0, 0, 99, 0, 104, 0, 109, 0, 0, 0, 99, 0, 109, 0, 103, 0, 0, 0, 100, 0, 101, 0, 110, 0, 0, 0, 101, 0, 109, 0, 97, 0, 0, 0, 103, 0, 98, 0, 97, 0, 0, 0, 103, 0, 111, 0, 110, 0, 0, 0, 103, 0, 114, 0, 98, 0, 0, 0, 103, 0, 114, 0, 99, 0, 0, 0, 103, 0, 118, 0, 114, 0, 0, 0, 104, 0, 105, 0, 102, 0, 0, 0, 104, 0, 110, 0, 110, 0, 0, 0, 107, 0, 98, 0, 100, 0, 0, 0, 107, 0, 108, 0, 110, 0, 0, 0, 107, 0, 111, 0, 107, 0, 0, 0, 107, 0, 112, 0, 101, 0, 0, 0, 107, 0, 116, 0, 114, 0, 0, 0, 107, 0, 122, 0, 106, 0, 0, 0, 107, 0, 122, 0, 116, 0, 0, 0, 108, 0, 97, 0, 104, 0, 0, 0, 108, 0, 101, 0, 122, 0, 0, 0, 108, 0, 105, 0, 102, 0, 0, 0, 108, 0, 117, 0, 121, 0, 0, 0, 109, 0, 97, 0, 105, 0, 0, 0, 109, 0, 97, 0, 107, 0, 0, 0, 109, 0, 110, 0, 105, 0, 0, 0, 109, 0, 110, 0, 119, 0, 0, 0, 109, 0, 114, 0, 121, 0, 0, 0, 109, 0, 119, 0, 114, 0, 0, 0, 110, 0, 101, 0, 119, 0, 0, 0, 112, 0, 97, 0, 108, 0, 0, 0, 112, 0, 105, 0, 106, 0, 0, 0, 114, 0, 97, 0, 106, 0, 0, 0, 114, 0, 101, 0, 106, 0, 0, 0, 114, 0, 104, 0, 103, 0, 0, 0, 114, 0, 105, 0, 102, 0, 0, 0, 115, 0, 97, 0, 116, 0, 0, 0, 115, 0, 107, 0, 114, 0, 0, 0, 115, 0, 111, 0, 103, 0, 0, 0, 115, 0, 114, 0, 120, 0, 0, 0, 115, 0, 121, 0, 108, 0, 0, 0, 115, 0, 121, 0, 114, 0, 0, 0, 116, 0, 97, 0, 106, 0, 0, 0, 116, 0, 98, 0, 119, 0, 0, 0, 116, 0, 100, 0, 117, 0, 0, 0, 116, 0, 115, 0, 102, 0, 0, 0, 116, 0, 118, 0, 100, 0, 0, 0, 117, 0, 111, 0, 107, 0, 0, 0, 120, 0, 109, 0, 114, 0, 0, 0, 121, 0, 117, 0, 101, 0, 0, 0, 97, 0, 108, 0, 98, 0, 0, 0, 97, 0, 115, 0, 101, 0, 0, 0, 98, 0, 107, 0, 109, 0, 0, 0, 98, 0, 120, 0, 104, 0, 0, 0, 100, 0, 116, 0, 121, 0, 0, 0, 103, 0, 107, 0, 110, 0, 0, 0, 104, 0, 108, 0, 117, 0, 0, 0, 107, 0, 98, 0, 109, 0, 0, 0, 107, 0, 102, 0, 121, 0, 0, 0, 107, 0, 104, 0, 116, 0, 0, 0, 107, 0, 104, 0, 122, 0, 0, 0, 107, 0, 107, 0, 106, 0, 0, 0, 107, 0, 114, 0, 109, 0, 0, 0, 107, 0, 116, 0, 111, 0, 0, 0, 108, 0, 108, 0, 110, 0, 0, 0, 108, 0, 111, 0, 106, 0, 0, 0, 109, 0, 98, 0, 104, 0, 0, 0, 109, 0, 101, 0, 101, 0, 0, 0, 109, 0, 102, 0, 111, 0, 0, 0, 109, 0, 107, 0, 119, 0, 0, 0, 110, 0, 99, 0, 101, 0, 0, 0, 110, 0, 103, 0, 108, 0, 0, 0, 110, 0, 103, 0, 118, 0, 0, 0, 110, 0, 105, 0, 105, 0, 0, 0, 110, 0, 115, 0, 110, 0, 0, 0, 111, 0, 116, 0, 97, 0, 0, 0, 112, 0, 110, 0, 110, 0, 0, 0, 114, 0, 119, 0, 111, 0, 0, 0, 115, 0, 98, 0, 101, 0, 0, 0, 115, 0, 114, 0, 110, 0, 0, 0, 116, 0, 102, 0, 105, 0, 0, 0, 116, 0, 103, 0, 117, 0, 0, 0, 116, 0, 105, 0, 103, 0, 0, 0, 116, 0, 105, 0, 111, 0, 0, 0, 116, 0, 107, 0, 108, 0, 0, 0, 116, 0, 116, 0, 101, 0, 0, 0, 116, 0, 116, 0, 115, 0, 0, 0, 116, 0, 116, 0, 116, 0, 0, 0, 117, 0, 103, 0, 97, 0, 0, 0, 117, 0, 108, 0, 105, 0, 0, 0, 117, 0, 114, 0, 119, 0, 0, 0, 119, 0, 97, 0, 101, 0, 0, 0, 119, 0, 97, 0, 110, 0, 0, 0, 119, 0, 98, 0, 114, 0, 0, 0, 119, 0, 101, 0, 111, 0, 0, 0, 119, 0, 106, 0, 105, 0, 0, 0, 119, 0, 109, 0, 111, 0, 0, 0, 119, 0, 115, 0, 103, 0, 0, 0, 119, 0, 119, 0, 97, 0, 0, 0, 120, 0, 98, 0, 105, 0, 0, 0, 120, 0, 99, 0, 104, 0, 0, 0, 120, 0, 99, 0, 111, 0, 0, 0, 120, 0, 108, 0, 97, 0, 0, 0, 120, 0, 110, 0, 97, 0, 0, 0, 120, 0, 110, 0, 114, 0, 0, 0, 120, 0, 110, 0, 121, 0, 0, 0, 120, 0, 115, 0, 105, 0, 0, 0, 120, 0, 115, 0, 108, 0, 0, 0, 121, 0, 97, 0, 115, 0, 0, 0, 121, 0, 97, 0, 118, 0, 0, 0, 121, 0, 98, 0, 97, 0, 0, 0, 121, 0, 108, 0, 103, 0, 0, 0, 121, 0, 109, 0, 108, 0, 0, 0, 121, 0, 109, 0, 116, 0, 0, 0, 121, 0, 111, 0, 115, 0, 0, 0, 121, 0, 115, 0, 115, 0, 0, 0, 122, 0, 105, 0, 97, 0, 0, 0, 48, 0, 51, 0, 52, 0, 0, 0, 53, 0, 56, 0, 50, 0, 0, 0, 56, 0, 56, 0, 54, 0, 0, 0, 56, 0, 57, 0, 48, 0, 0, 0, 48, 0, 54, 0, 50, 0, 0, 0, 49, 0, 55, 0, 50, 0, 0, 0, 50, 0, 48, 0, 48, 0, 0, 0, 50, 0, 51, 0, 48, 0, 0, 0, 50, 0, 56, 0, 48, 0, 0, 0, 53, 0, 51, 0, 50, 0, 0, 0, 55, 0, 51, 0, 54, 0, 0, 0, 56, 0, 51, 0, 48, 0, 0, 0, 97, 0, 97, 0, 105, 0, 0, 0, 97, 0, 97, 0, 107, 0, 0, 0, 97, 0, 97, 0, 109, 0, 0, 0, 97, 0, 97, 0, 115, 0, 0, 0, 97, 0, 97, 0, 117, 0, 0, 0, 97, 0, 98, 0, 105, 0, 0, 0, 97, 0, 98, 0, 113, 0, 0, 0, 97, 0, 98, 0, 114, 0, 0, 0, 97, 0, 98, 0, 116, 0, 0, 0, 97, 0, 98, 0, 121, 0, 0, 0, 97, 0, 99, 0, 100, 0, 0, 0, 97, 0, 99, 0, 101, 0, 0, 0, 97, 0, 99, 0, 104, 0, 0, 0, 97, 0, 99, 0, 110, 0, 0, 0, 97, 0, 100, 0, 97, 0, 0, 0, 97, 0, 100, 0, 101, 0, 0, 0, 97, 0, 100, 0, 106, 0, 0, 0, 97, 0, 100, 0, 112, 0, 0, 0, 97, 0, 100, 0, 120, 0, 0, 0, 97, 0, 100, 0, 121, 0, 0, 0, 97, 0, 100, 0, 122, 0, 0, 0, 97, 0, 101, 0, 98, 0, 0, 0, 97, 0, 101, 0, 121, 0, 0, 0, 97, 0, 103, 0, 99, 0, 0, 0, 97, 0, 103, 0, 100, 0, 0, 0, 97, 0, 103, 0, 103, 0, 0, 0, 97, 0, 103, 0, 109, 0, 0, 0, 97, 0, 103, 0, 111, 0, 0, 0, 97, 0, 103, 0, 113, 0, 0, 0, 97, 0, 104, 0, 97, 0, 0, 0, 97, 0, 104, 0, 108, 0, 0, 0, 97, 0, 104, 0, 111, 0, 0, 0, 97, 0, 106, 0, 103, 0, 0, 0, 97, 0, 106, 0, 117, 0, 0, 0, 97, 0, 107, 0, 107, 0, 0, 0, 97, 0, 108, 0, 97, 0, 0, 0, 97, 0, 108, 0, 110, 0, 0, 0, 97, 0, 108, 0, 115, 0, 0, 0, 97, 0, 108, 0, 116, 0, 0, 0, 97, 0, 109, 0, 109, 0, 0, 0, 97, 0, 109, 0, 110, 0, 0, 0, 97, 0, 109, 0, 112, 0, 0, 0, 97, 0, 110, 0, 99, 0, 0, 0, 97, 0, 110, 0, 107, 0, 0, 0, 97, 0, 110, 0, 110, 0, 0, 0, 97, 0, 110, 0, 121, 0, 0, 0, 97, 0, 111, 0, 103, 0, 0, 0, 97, 0, 111, 0, 106, 0, 0, 0, 97, 0, 111, 0, 109, 0, 0, 0, 97, 0, 111, 0, 122, 0, 0, 0, 97, 0, 112, 0, 99, 0, 0, 0, 97, 0, 112, 0, 100, 0, 0, 0, 97, 0, 112, 0, 101, 0, 0, 0, 97, 0, 112, 0, 114, 0, 0, 0, 97, 0, 112, 0, 115, 0, 0, 0, 97, 0, 112, 0, 122, 0, 0, 0, 97, 0, 114, 0, 104, 0, 0, 0, 97, 0, 114, 0, 109, 0, 0, 0, 97, 0, 114, 0, 110, 0, 0, 0, 97, 0, 114, 0, 111, 0, 0, 0, 97, 0, 114, 0, 113, 0, 0, 0, 97, 0, 114, 0, 115, 0, 0, 0, 97, 0, 114, 0, 121, 0, 0, 0, 97, 0, 114, 0, 122, 0, 0, 0, 97, 0, 115, 0, 97, 0, 0, 0, 97, 0, 115, 0, 100, 0, 0, 0, 97, 0, 115, 0, 103, 0, 0, 0, 97, 0, 115, 0, 111, 0, 0, 0, 97, 0, 115, 0, 116, 0, 0, 0, 97, 0, 116, 0, 97, 0, 0, 0, 97, 0, 116, 0, 103, 0, 0, 0, 97, 0, 116, 0, 106, 0, 0, 0, 97, 0, 117, 0, 101, 0, 0, 0, 97, 0, 117, 0, 121, 0, 0, 0, 97, 0, 118, 0, 108, 0, 0, 0, 97, 0, 118, 0, 110, 0, 0, 0, 97, 0, 118, 0, 117, 0, 0, 0, 97, 0, 119, 0, 97, 0, 0, 0, 97, 0, 119, 0, 98, 0, 0, 0, 97, 0, 119, 0, 111, 0, 0, 0, 97, 0, 119, 0, 120, 0, 0, 0, 97, 0, 121, 0, 98, 0, 0, 0, 97, 0, 121, 0, 114, 0, 0, 0, 97, 0, 121, 0, 120, 0, 0, 0, 97, 0, 122, 0, 106, 0, 0, 0, 98, 0, 97, 0, 112, 0, 0, 0, 98, 0, 97, 0, 113, 0, 0, 0, 98, 0, 97, 0, 114, 0, 0, 0, 98, 0, 97, 0, 115, 0, 0, 0, 98, 0, 97, 0, 118, 0, 0, 0, 98, 0, 97, 0, 120, 0, 0, 0, 98, 0, 98, 0, 97, 0, 0, 0, 98, 0, 98, 0, 98, 0, 0, 0, 98, 0, 98, 0, 100, 0, 0, 0, 98, 0, 98, 0, 106, 0, 0, 0, 98, 0, 98, 0, 112, 0, 0, 0, 98, 0, 98, 0, 114, 0, 0, 0, 98, 0, 99, 0, 99, 0, 0, 0, 98, 0, 99, 0, 102, 0, 0, 0, 98, 0, 99, 0, 103, 0, 0, 0, 98, 0, 99, 0, 104, 0, 0, 0, 98, 0, 99, 0, 105, 0, 0, 0, 98, 0, 99, 0, 108, 0, 0, 0, 98, 0, 99, 0, 109, 0, 0, 0, 98, 0, 99, 0, 110, 0, 0, 0, 98, 0, 99, 0, 111, 0, 0, 0, 98, 0, 99, 0, 113, 0, 0, 0, 98, 0, 99, 0, 117, 0, 0, 0, 98, 0, 100, 0, 100, 0, 0, 0, 98, 0, 101, 0, 102, 0, 0, 0, 98, 0, 101, 0, 104, 0, 0, 0, 98, 0, 101, 0, 106, 0, 0, 0, 98, 0, 101, 0, 109, 0, 0, 0, 98, 0, 101, 0, 116, 0, 0, 0, 98, 0, 101, 0, 119, 0, 0, 0, 98, 0, 101, 0, 120, 0, 0, 0, 98, 0, 101, 0, 122, 0, 0, 0, 98, 0, 102, 0, 100, 0, 0, 0, 98, 0, 102, 0, 113, 0, 0, 0, 98, 0, 102, 0, 116, 0, 0, 0, 98, 0, 103, 0, 99, 0, 0, 0, 98, 0, 103, 0, 109, 0, 0, 0, 98, 0, 103, 0, 110, 0, 0, 0, 98, 0, 103, 0, 120, 0, 0, 0, 98, 0, 104, 0, 98, 0, 0, 0, 98, 0, 104, 0, 103, 0, 0, 0, 98, 0, 104, 0, 105, 0, 0, 0, 98, 0, 104, 0, 108, 0, 0, 0, 98, 0, 104, 0, 121, 0, 0, 0, 98, 0, 105, 0, 98, 0, 0, 0, 98, 0, 105, 0, 103, 0, 0, 0, 98, 0, 105, 0, 109, 0, 0, 0, 98, 0, 105, 0, 110, 0, 0, 0, 98, 0, 105, 0, 111, 0, 0, 0, 98, 0, 105, 0, 113, 0, 0, 0, 98, 0, 106, 0, 100, 0, 0, 0, 98, 0, 106, 0, 104, 0, 0, 0, 98, 0, 106, 0, 105, 0, 0, 0, 98, 0, 106, 0, 106, 0, 0, 0, 98, 0, 106, 0, 110, 0, 0, 0, 98, 0, 106, 0, 111, 0, 0, 0, 98, 0, 106, 0, 114, 0, 0, 0, 98, 0, 106, 0, 116, 0, 0, 0, 98, 0, 106, 0, 122, 0, 0, 0, 98, 0, 107, 0, 99, 0, 0, 0, 98, 0, 107, 0, 113, 0, 0, 0, 98, 0, 107, 0, 118, 0, 0, 0, 98, 0, 108, 0, 116, 0, 0, 0, 98, 0, 109, 0, 102, 0, 0, 0, 98, 0, 109, 0, 104, 0, 0, 0, 98, 0, 109, 0, 107, 0, 0, 0, 98, 0, 109, 0, 113, 0, 0, 0, 98, 0, 109, 0, 117, 0, 0, 0, 98, 0, 110, 0, 99, 0, 0, 0, 98, 0, 110, 0, 103, 0, 0, 0, 98, 0, 110, 0, 109, 0, 0, 0, 98, 0, 110, 0, 112, 0, 0, 0, 98, 0, 111, 0, 106, 0, 0, 0, 98, 0, 111, 0, 109, 0, 0, 0, 98, 0, 111, 0, 110, 0, 0, 0, 98, 0, 112, 0, 121, 0, 0, 0, 98, 0, 113, 0, 99, 0, 0, 0, 98, 0, 113, 0, 105, 0, 0, 0, 98, 0, 113, 0, 112, 0, 0, 0, 98, 0, 113, 0, 118, 0, 0, 0, 98, 0, 114, 0, 97, 0, 0, 0, 98, 0, 114, 0, 104, 0, 0, 0, 98, 0, 114, 0, 120, 0, 0, 0, 98, 0, 114, 0, 122, 0, 0, 0, 98, 0, 115, 0, 106, 0, 0, 0, 98, 0, 115, 0, 115, 0, 0, 0, 98, 0, 115, 0, 116, 0, 0, 0, 98, 0, 116, 0, 111, 0, 0, 0, 98, 0, 116, 0, 116, 0, 0, 0, 98, 0, 116, 0, 118, 0, 0, 0, 98, 0, 117, 0, 99, 0, 0, 0, 98, 0, 117, 0, 100, 0, 0, 0, 98, 0, 117, 0, 107, 0, 0, 0, 98, 0, 117, 0, 109, 0, 0, 0, 98, 0, 117, 0, 111, 0, 0, 0, 98, 0, 117, 0, 114, 0, 0, 0, 98, 0, 117, 0, 115, 0, 0, 0, 98, 0, 117, 0, 117, 0, 0, 0, 98, 0, 118, 0, 98, 0, 0, 0, 98, 0, 119, 0, 100, 0, 0, 0, 98, 0, 119, 0, 114, 0, 0, 0, 98, 0, 120, 0, 107, 0, 0, 0, 98, 0, 120, 0, 114, 0, 0, 0, 98, 0, 121, 0, 101, 0, 0, 0, 98, 0, 121, 0, 110, 0, 0, 0, 98, 0, 121, 0, 114, 0, 0, 0, 98, 0, 121, 0, 115, 0, 0, 0, 98, 0, 121, 0, 118, 0, 0, 0, 98, 0, 121, 0, 120, 0, 0, 0, 98, 0, 122, 0, 97, 0, 0, 0, 98, 0, 122, 0, 101, 0, 0, 0, 98, 0, 122, 0, 102, 0, 0, 0, 98, 0, 122, 0, 104, 0, 0, 0, 98, 0, 122, 0, 119, 0, 0, 0, 99, 0, 97, 0, 110, 0, 0, 0, 99, 0, 97, 0, 120, 0, 0, 0, 99, 0, 98, 0, 106, 0, 0, 0, 99, 0, 99, 0, 104, 0, 0, 0, 99, 0, 99, 0, 112, 0, 0, 0, 99, 0, 99, 0, 113, 0, 0, 0, 99, 0, 101, 0, 98, 0, 0, 0, 99, 0, 102, 0, 97, 0, 0, 0, 99, 0, 103, 0, 103, 0, 0, 0, 99, 0, 104, 0, 105, 0, 0, 0, 99, 0, 104, 0, 107, 0, 0, 0, 99, 0, 104, 0, 112, 0, 0, 0, 99, 0, 104, 0, 114, 0, 0, 0, 99, 0, 105, 0, 99, 0, 0, 0, 99, 0, 105, 0, 114, 0, 0, 0, 99, 0, 106, 0, 97, 0, 0, 0, 99, 0, 106, 0, 109, 0, 0, 0, 99, 0, 106, 0, 114, 0, 0, 0, 99, 0, 107, 0, 97, 0, 0, 0, 99, 0, 107, 0, 98, 0, 0, 0, 99, 0, 107, 0, 108, 0, 0, 0, 99, 0, 107, 0, 111, 0, 0, 0, 99, 0, 107, 0, 121, 0, 0, 0, 99, 0, 108, 0, 97, 0, 0, 0, 99, 0, 108, 0, 100, 0, 0, 0, 99, 0, 109, 0, 101, 0, 0, 0, 99, 0, 109, 0, 107, 0, 0, 0, 99, 0, 109, 0, 110, 0, 0, 0, 99, 0, 109, 0, 114, 0, 0, 0, 99, 0, 111, 0, 112, 0, 0, 0, 99, 0, 111, 0, 121, 0, 0, 0, 99, 0, 112, 0, 115, 0, 0, 0, 99, 0, 114, 0, 104, 0, 0, 0, 99, 0, 114, 0, 106, 0, 0, 0, 99, 0, 114, 0, 107, 0, 0, 0, 99, 0, 114, 0, 108, 0, 0, 0, 99, 0, 114, 0, 109, 0, 0, 0, 99, 0, 114, 0, 115, 0, 0, 0, 99, 0, 115, 0, 98, 0, 0, 0, 99, 0, 115, 0, 119, 0, 0, 0, 99, 0, 116, 0, 100, 0, 0, 0, 99, 0, 119, 0, 100, 0, 0, 0, 99, 0, 122, 0, 101, 0, 0, 0, 100, 0, 97, 0, 100, 0, 0, 0, 100, 0, 97, 0, 102, 0, 0, 0, 100, 0, 97, 0, 103, 0, 0, 0, 100, 0, 97, 0, 104, 0, 0, 0, 100, 0, 97, 0, 107, 0, 0, 0, 100, 0, 97, 0, 114, 0, 0, 0, 100, 0, 97, 0, 118, 0, 0, 0, 100, 0, 98, 0, 100, 0, 0, 0, 100, 0, 98, 0, 113, 0, 0, 0, 100, 0, 99, 0, 99, 0, 0, 0, 100, 0, 100, 0, 110, 0, 0, 0, 100, 0, 101, 0, 100, 0, 0, 0, 100, 0, 101, 0, 108, 0, 0, 0, 100, 0, 101, 0, 118, 0, 0, 0, 100, 0, 103, 0, 97, 0, 0, 0, 100, 0, 103, 0, 104, 0, 0, 0, 100, 0, 103, 0, 105, 0, 0, 0, 100, 0, 103, 0, 108, 0, 0, 0, 100, 0, 103, 0, 111, 0, 0, 0, 100, 0, 103, 0, 114, 0, 0, 0, 100, 0, 103, 0, 122, 0, 0, 0, 100, 0, 104, 0, 100, 0, 0, 0, 100, 0, 105, 0, 97, 0, 0, 0, 100, 0, 105, 0, 102, 0, 0, 0, 100, 0, 105, 0, 107, 0, 0, 0, 100, 0, 105, 0, 110, 0, 0, 0, 100, 0, 105, 0, 113, 0, 0, 0, 100, 0, 105, 0, 116, 0, 0, 0, 100, 0, 106, 0, 101, 0, 0, 0, 100, 0, 110, 0, 106, 0, 0, 0, 100, 0, 111, 0, 98, 0, 0, 0, 100, 0, 111, 0, 112, 0, 0, 0, 100, 0, 111, 0, 119, 0, 0, 0, 100, 0, 114, 0, 104, 0, 0, 0, 100, 0, 114, 0, 105, 0, 0, 0, 100, 0, 114, 0, 108, 0, 0, 0, 100, 0, 114, 0, 115, 0, 0, 0, 100, 0, 115, 0, 98, 0, 0, 0, 100, 0, 116, 0, 109, 0, 0, 0, 100, 0, 116, 0, 115, 0, 0, 0, 100, 0, 117, 0, 97, 0, 0, 0, 100, 0, 117, 0, 99, 0, 0, 0, 100, 0, 117, 0, 100, 0, 0, 0, 100, 0, 117, 0, 103, 0, 0, 0, 100, 0, 117, 0, 116, 0, 0, 0, 100, 0, 117, 0, 122, 0, 0, 0, 100, 0, 118, 0, 97, 0, 0, 0, 100, 0, 119, 0, 119, 0, 0, 0, 100, 0, 121, 0, 111, 0, 0, 0, 100, 0, 121, 0, 117, 0, 0, 0, 100, 0, 122, 0, 103, 0, 0, 0, 101, 0, 98, 0, 117, 0, 0, 0, 101, 0, 102, 0, 105, 0, 0, 0, 101, 0, 103, 0, 108, 0, 0, 0, 101, 0, 103, 0, 121, 0, 0, 0, 101, 0, 107, 0, 97, 0, 0, 0, 101, 0, 107, 0, 107, 0, 0, 0, 101, 0, 109, 0, 105, 0, 0, 0, 101, 0, 110, 0, 110, 0, 0, 0, 101, 0, 110, 0, 113, 0, 0, 0, 101, 0, 114, 0, 105, 0, 0, 0, 101, 0, 115, 0, 103, 0, 0, 0, 101, 0, 115, 0, 107, 0, 0, 0, 101, 0, 115, 0, 117, 0, 0, 0, 101, 0, 116, 0, 114, 0, 0, 0, 101, 0, 116, 0, 116, 0, 0, 0, 101, 0, 116, 0, 117, 0, 0, 0, 101, 0, 116, 0, 120, 0, 0, 0, 101, 0, 119, 0, 111, 0, 0, 0, 101, 0, 120, 0, 116, 0, 0, 0, 102, 0, 97, 0, 97, 0, 0, 0, 102, 0, 97, 0, 98, 0, 0, 0, 102, 0, 97, 0, 103, 0, 0, 0, 102, 0, 97, 0, 105, 0, 0, 0, 102, 0, 97, 0, 110, 0, 0, 0, 102, 0, 97, 0, 116, 0, 0, 0, 102, 0, 102, 0, 105, 0, 0, 0, 102, 0, 102, 0, 109, 0, 0, 0, 102, 0, 105, 0, 97, 0, 0, 0, 102, 0, 108, 0, 114, 0, 0, 0, 102, 0, 109, 0, 112, 0, 0, 0, 102, 0, 111, 0, 100, 0, 0, 0, 102, 0, 111, 0, 110, 0, 0, 0, 102, 0, 111, 0, 114, 0, 0, 0, 102, 0, 112, 0, 101, 0, 0, 0, 102, 0, 113, 0, 115, 0, 0, 0, 102, 0, 114, 0, 99, 0, 0, 0, 102, 0, 114, 0, 101, 0, 0, 0, 102, 0, 114, 0, 112, 0, 0, 0, 102, 0, 114, 0, 114, 0, 0, 0, 102, 0, 114, 0, 115, 0, 0, 0, 102, 0, 117, 0, 98, 0, 0, 0, 102, 0, 117, 0, 99, 0, 0, 0, 102, 0, 117, 0, 100, 0, 0, 0, 102, 0, 117, 0, 101, 0, 0, 0, 102, 0, 117, 0, 102, 0, 0, 0, 102, 0, 117, 0, 104, 0, 0, 0, 102, 0, 117, 0, 113, 0, 0, 0, 102, 0, 117, 0, 114, 0, 0, 0, 102, 0, 117, 0, 118, 0, 0, 0, 102, 0, 117, 0, 121, 0, 0, 0, 102, 0, 118, 0, 114, 0, 0, 0, 103, 0, 97, 0, 97, 0, 0, 0, 103, 0, 97, 0, 102, 0, 0, 0, 103, 0, 97, 0, 103, 0, 0, 0, 103, 0, 97, 0, 104, 0, 0, 0, 103, 0, 97, 0, 106, 0, 0, 0, 103, 0, 97, 0, 108, 0, 0, 0, 103, 0, 97, 0, 110, 0, 0, 0, 103, 0, 97, 0, 118, 0, 0, 0, 103, 0, 97, 0, 119, 0, 0, 0, 103, 0, 97, 0, 121, 0, 0, 0, 103, 0, 97, 0, 122, 0, 0, 0, 103, 0, 98, 0, 102, 0, 0, 0, 103, 0, 98, 0, 109, 0, 0, 0, 103, 0, 98, 0, 111, 0, 0, 0, 103, 0, 98, 0, 121, 0, 0, 0, 103, 0, 98, 0, 122, 0, 0, 0, 103, 0, 99, 0, 114, 0, 0, 0, 103, 0, 100, 0, 101, 0, 0, 0, 103, 0, 100, 0, 106, 0, 0, 0, 103, 0, 100, 0, 110, 0, 0, 0, 103, 0, 100, 0, 114, 0, 0, 0, 103, 0, 101, 0, 98, 0, 0, 0, 103, 0, 101, 0, 106, 0, 0, 0, 103, 0, 101, 0, 108, 0, 0, 0, 103, 0, 101, 0, 111, 0, 0, 0, 103, 0, 101, 0, 114, 0, 0, 0, 103, 0, 101, 0, 122, 0, 0, 0, 103, 0, 102, 0, 107, 0, 0, 0, 103, 0, 102, 0, 120, 0, 0, 0, 103, 0, 103, 0, 110, 0, 0, 0, 103, 0, 104, 0, 115, 0, 0, 0, 103, 0, 105, 0, 108, 0, 0, 0, 103, 0, 105, 0, 109, 0, 0, 0, 103, 0, 106, 0, 107, 0, 0, 0, 103, 0, 106, 0, 110, 0, 0, 0, 103, 0, 106, 0, 117, 0, 0, 0, 103, 0, 107, 0, 112, 0, 0, 0, 103, 0, 108, 0, 107, 0, 0, 0, 103, 0, 109, 0, 109, 0, 0, 0, 103, 0, 109, 0, 118, 0, 0, 0, 103, 0, 110, 0, 100, 0, 0, 0, 103, 0, 110, 0, 103, 0, 0, 0, 103, 0, 110, 0, 111, 0, 0, 0, 103, 0, 111, 0, 100, 0, 0, 0, 103, 0, 111, 0, 102, 0, 0, 0, 103, 0, 111, 0, 105, 0, 0, 0, 103, 0, 111, 0, 109, 0, 0, 0, 103, 0, 111, 0, 114, 0, 0, 0, 103, 0, 111, 0, 115, 0, 0, 0, 103, 0, 111, 0, 116, 0, 0, 0, 103, 0, 114, 0, 101, 0, 0, 0, 103, 0, 114, 0, 116, 0, 0, 0, 103, 0, 114, 0, 119, 0, 0, 0, 103, 0, 115, 0, 119, 0, 0, 0, 103, 0, 116, 0, 105, 0, 0, 0, 103, 0, 117, 0, 98, 0, 0, 0, 103, 0, 117, 0, 99, 0, 0, 0, 103, 0, 117, 0, 100, 0, 0, 0, 103, 0, 117, 0, 103, 0, 0, 0, 103, 0, 117, 0, 114, 0, 0, 0, 103, 0, 117, 0, 118, 0, 0, 0, 103, 0, 117, 0, 119, 0, 0, 0, 103, 0, 117, 0, 120, 0, 0, 0, 103, 0, 117, 0, 122, 0, 0, 0, 103, 0, 118, 0, 102, 0, 0, 0, 103, 0, 118, 0, 115, 0, 0, 0, 103, 0, 119, 0, 99, 0, 0, 0, 103, 0, 119, 0, 105, 0, 0, 0, 103, 0, 119, 0, 116, 0, 0, 0, 103, 0, 121, 0, 97, 0, 0, 0, 104, 0, 97, 0, 107, 0, 0, 0, 104, 0, 97, 0, 122, 0, 0, 0, 104, 0, 98, 0, 98, 0, 0, 0, 104, 0, 100, 0, 110, 0, 0, 0, 104, 0, 100, 0, 121, 0, 0, 0, 104, 0, 101, 0, 97, 0, 0, 0, 104, 0, 105, 0, 97, 0, 0, 0, 104, 0, 105, 0, 103, 0, 0, 0, 104, 0, 105, 0, 104, 0, 0, 0, 104, 0, 105, 0, 108, 0, 0, 0, 104, 0, 105, 0, 109, 0, 0, 0, 104, 0, 108, 0, 97, 0, 0, 0, 104, 0, 108, 0, 101, 0, 0, 0, 104, 0, 109, 0, 100, 0, 0, 0, 104, 0, 109, 0, 110, 0, 0, 0, 104, 0, 109, 0, 116, 0, 0, 0, 104, 0, 110, 0, 100, 0, 0, 0, 104, 0, 110, 0, 101, 0, 0, 0, 104, 0, 110, 0, 106, 0, 0, 0, 104, 0, 111, 0, 116, 0, 0, 0, 104, 0, 114, 0, 114, 0, 0, 0, 104, 0, 115, 0, 98, 0, 0, 0, 104, 0, 115, 0, 110, 0, 0, 0, 104, 0, 117, 0, 105, 0, 0, 0, 104, 0, 117, 0, 119, 0, 0, 0, 105, 0, 97, 0, 110, 0, 0, 0, 105, 0, 97, 0, 114, 0, 0, 0, 105, 0, 98, 0, 97, 0, 0, 0, 105, 0, 98, 0, 98, 0, 0, 0, 105, 0, 98, 0, 105, 0, 0, 0, 105, 0, 98, 0, 121, 0, 0, 0, 105, 0, 99, 0, 97, 0, 0, 0, 105, 0, 99, 0, 101, 0, 0, 0, 105, 0, 99, 0, 104, 0, 0, 0, 105, 0, 100, 0, 105, 0, 0, 0, 105, 0, 100, 0, 117, 0, 0, 0, 105, 0, 102, 0, 101, 0, 0, 0, 105, 0, 103, 0, 98, 0, 0, 0, 105, 0, 103, 0, 101, 0, 0, 0, 105, 0, 106, 0, 106, 0, 0, 0, 105, 0, 107, 0, 101, 0, 0, 0, 105, 0, 107, 0, 107, 0, 0, 0, 105, 0, 107, 0, 116, 0, 0, 0, 105, 0, 107, 0, 119, 0, 0, 0, 105, 0, 107, 0, 120, 0, 0, 0, 105, 0, 108, 0, 111, 0, 0, 0, 105, 0, 108, 0, 119, 0, 0, 0, 105, 0, 109, 0, 111, 0, 0, 0, 105, 0, 111, 0, 117, 0, 0, 0, 105, 0, 114, 0, 105, 0, 0, 0, 105, 0, 119, 0, 109, 0, 0, 0, 105, 0, 119, 0, 115, 0, 0, 0, 105, 0, 122, 0, 104, 0, 0, 0, 105, 0, 122, 0, 105, 0, 0, 0, 106, 0, 97, 0, 98, 0, 0, 0, 106, 0, 97, 0, 108, 0, 0, 0, 106, 0, 97, 0, 109, 0, 0, 0, 106, 0, 98, 0, 111, 0, 0, 0, 106, 0, 98, 0, 117, 0, 0, 0, 106, 0, 101, 0, 103, 0, 0, 0, 106, 0, 101, 0, 110, 0, 0, 0, 106, 0, 103, 0, 107, 0, 0, 0, 106, 0, 103, 0, 111, 0, 0, 0, 106, 0, 105, 0, 98, 0, 0, 0, 106, 0, 109, 0, 99, 0, 0, 0, 106, 0, 109, 0, 108, 0, 0, 0, 106, 0, 114, 0, 97, 0, 0, 0, 106, 0, 114, 0, 98, 0, 0, 0, 106, 0, 117, 0, 116, 0, 0, 0, 107, 0, 97, 0, 98, 0, 0, 0, 107, 0, 97, 0, 99, 0, 0, 0, 107, 0, 97, 0, 100, 0, 0, 0, 107, 0, 97, 0, 105, 0, 0, 0, 107, 0, 97, 0, 106, 0, 0, 0, 107, 0, 97, 0, 107, 0, 0, 0, 107, 0, 97, 0, 109, 0, 0, 0, 107, 0, 97, 0, 111, 0, 0, 0, 107, 0, 98, 0, 112, 0, 0, 0, 107, 0, 98, 0, 113, 0, 0, 0, 107, 0, 98, 0, 120, 0, 0, 0, 107, 0, 98, 0, 121, 0, 0, 0, 107, 0, 99, 0, 103, 0, 0, 0, 107, 0, 99, 0, 107, 0, 0, 0, 107, 0, 99, 0, 108, 0, 0, 0, 107, 0, 99, 0, 116, 0, 0, 0, 107, 0, 100, 0, 101, 0, 0, 0, 107, 0, 100, 0, 104, 0, 0, 0, 107, 0, 100, 0, 108, 0, 0, 0, 107, 0, 101, 0, 97, 0, 0, 0, 107, 0, 101, 0, 110, 0, 0, 0, 107, 0, 101, 0, 122, 0, 0, 0, 107, 0, 102, 0, 111, 0, 0, 0, 107, 0, 102, 0, 114, 0, 0, 0, 107, 0, 103, 0, 99, 0, 0, 0, 107, 0, 103, 0, 101, 0, 0, 0, 107, 0, 103, 0, 102, 0, 0, 0, 107, 0, 103, 0, 104, 0, 0, 0, 107, 0, 103, 0, 112, 0, 0, 0, 107, 0, 104, 0, 98, 0, 0, 0, 107, 0, 104, 0, 107, 0, 0, 0, 107, 0, 104, 0, 110, 0, 0, 0, 107, 0, 104, 0, 113, 0, 0, 0, 107, 0, 104, 0, 115, 0, 0, 0, 107, 0, 104, 0, 119, 0, 0, 0, 107, 0, 105, 0, 106, 0, 0, 0, 107, 0, 105, 0, 117, 0, 0, 0, 107, 0, 105, 0, 119, 0, 0, 0, 107, 0, 106, 0, 100, 0, 0, 0, 107, 0, 106, 0, 103, 0, 0, 0, 107, 0, 106, 0, 115, 0, 0, 0, 107, 0, 106, 0, 121, 0, 0, 0, 107, 0, 107, 0, 99, 0, 0, 0, 107, 0, 108, 0, 113, 0, 0, 0, 107, 0, 108, 0, 116, 0, 0, 0, 107, 0, 108, 0, 120, 0, 0, 0, 107, 0, 109, 0, 98, 0, 0, 0, 107, 0, 109, 0, 104, 0, 0, 0, 107, 0, 109, 0, 108, 0, 0, 0, 107, 0, 109, 0, 111, 0, 0, 0, 107, 0, 109, 0, 114, 0, 0, 0, 107, 0, 109, 0, 115, 0, 0, 0, 107, 0, 109, 0, 117, 0, 0, 0, 107, 0, 109, 0, 119, 0, 0, 0, 107, 0, 110, 0, 99, 0, 0, 0, 107, 0, 110, 0, 102, 0, 0, 0, 107, 0, 110, 0, 103, 0, 0, 0, 107, 0, 110, 0, 110, 0, 0, 0, 107, 0, 110, 0, 112, 0, 0, 0, 107, 0, 111, 0, 105, 0, 0, 0, 107, 0, 111, 0, 106, 0, 0, 0, 107, 0, 111, 0, 108, 0, 0, 0, 107, 0, 111, 0, 115, 0, 0, 0, 107, 0, 111, 0, 122, 0, 0, 0, 107, 0, 112, 0, 102, 0, 0, 0, 107, 0, 112, 0, 111, 0, 0, 0, 107, 0, 112, 0, 114, 0, 0, 0, 107, 0, 112, 0, 118, 0, 0, 0, 107, 0, 112, 0, 120, 0, 0, 0, 107, 0, 113, 0, 98, 0, 0, 0, 107, 0, 113, 0, 102, 0, 0, 0, 107, 0, 113, 0, 115, 0, 0, 0, 107, 0, 113, 0, 121, 0, 0, 0, 107, 0, 114, 0, 99, 0, 0, 0, 107, 0, 114, 0, 105, 0, 0, 0, 107, 0, 114, 0, 106, 0, 0, 0, 107, 0, 114, 0, 108, 0, 0, 0, 107, 0, 114, 0, 115, 0, 0, 0, 107, 0, 114, 0, 117, 0, 0, 0, 107, 0, 115, 0, 98, 0, 0, 0, 107, 0, 115, 0, 100, 0, 0, 0, 107, 0, 115, 0, 102, 0, 0, 0, 107, 0, 115, 0, 104, 0, 0, 0, 107, 0, 115, 0, 106, 0, 0, 0, 107, 0, 115, 0, 114, 0, 0, 0, 107, 0, 116, 0, 98, 0, 0, 0, 107, 0, 116, 0, 109, 0, 0, 0, 107, 0, 116, 0, 122, 0, 0, 0, 107, 0, 117, 0, 98, 0, 0, 0, 107, 0, 117, 0, 100, 0, 0, 0, 107, 0, 117, 0, 101, 0, 0, 0, 107, 0, 117, 0, 106, 0, 0, 0, 107, 0, 117, 0, 109, 0, 0, 0, 107, 0, 117, 0, 110, 0, 0, 0, 107, 0, 117, 0, 112, 0, 0, 0, 107, 0, 117, 0, 115, 0, 0, 0, 107, 0, 118, 0, 103, 0, 0, 0, 107, 0, 118, 0, 114, 0, 0, 0, 107, 0, 118, 0, 115, 0, 0, 0, 107, 0, 118, 0, 120, 0, 0, 0, 107, 0, 119, 0, 106, 0, 0, 0, 107, 0, 119, 0, 111, 0, 0, 0, 107, 0, 119, 0, 113, 0, 0, 0, 107, 0, 119, 0, 118, 0, 0, 0, 107, 0, 120, 0, 97, 0, 0, 0, 107, 0, 120, 0, 99, 0, 0, 0, 107, 0, 120, 0, 101, 0, 0, 0, 107, 0, 120, 0, 109, 0, 0, 0, 107, 0, 120, 0, 112, 0, 0, 0, 107, 0, 120, 0, 119, 0, 0, 0, 107, 0, 120, 0, 122, 0, 0, 0, 107, 0, 121, 0, 101, 0, 0, 0, 107, 0, 121, 0, 120, 0, 0, 0, 107, 0, 122, 0, 114, 0, 0, 0, 108, 0, 97, 0, 98, 0, 0, 0, 108, 0, 97, 0, 100, 0, 0, 0, 108, 0, 97, 0, 106, 0, 0, 0, 108, 0, 97, 0, 115, 0, 0, 0, 108, 0, 98, 0, 101, 0, 0, 0, 108, 0, 98, 0, 107, 0, 0, 0, 108, 0, 98, 0, 117, 0, 0, 0, 108, 0, 98, 0, 119, 0, 0, 0, 108, 0, 99, 0, 109, 0, 0, 0, 108, 0, 99, 0, 112, 0, 0, 0, 108, 0, 99, 0, 113, 0, 0, 0, 108, 0, 100, 0, 98, 0, 0, 0, 108, 0, 101, 0, 100, 0, 0, 0, 108, 0, 101, 0, 101, 0, 0, 0, 108, 0, 101, 0, 109, 0, 0, 0, 108, 0, 101, 0, 112, 0, 0, 0, 108, 0, 101, 0, 113, 0, 0, 0, 108, 0, 101, 0, 117, 0, 0, 0, 108, 0, 103, 0, 103, 0, 0, 0, 108, 0, 105, 0, 97, 0, 0, 0, 108, 0, 105, 0, 100, 0, 0, 0, 108, 0, 105, 0, 103, 0, 0, 0, 108, 0, 105, 0, 104, 0, 0, 0, 108, 0, 105, 0, 105, 0, 0, 0, 108, 0, 105, 0, 106, 0, 0, 0, 108, 0, 105, 0, 115, 0, 0, 0, 108, 0, 106, 0, 112, 0, 0, 0, 108, 0, 107, 0, 105, 0, 0, 0, 108, 0, 107, 0, 116, 0, 0, 0, 108, 0, 108, 0, 101, 0, 0, 0, 108, 0, 108, 0, 111, 0, 0, 0, 108, 0, 109, 0, 109, 0, 0, 0, 108, 0, 109, 0, 110, 0, 0, 0, 108, 0, 109, 0, 111, 0, 0, 0, 108, 0, 109, 0, 112, 0, 0, 0, 108, 0, 110, 0, 115, 0, 0, 0, 108, 0, 110, 0, 117, 0, 0, 0, 108, 0, 111, 0, 107, 0, 0, 0, 108, 0, 111, 0, 108, 0, 0, 0, 108, 0, 111, 0, 115, 0, 0, 0, 108, 0, 111, 0, 122, 0, 0, 0, 108, 0, 114, 0, 99, 0, 0, 0, 108, 0, 114, 0, 114, 0, 0, 0, 108, 0, 117, 0, 97, 0, 0, 0, 108, 0, 117, 0, 111, 0, 0, 0, 108, 0, 117, 0, 122, 0, 0, 0, 108, 0, 118, 0, 115, 0, 0, 0, 108, 0, 119, 0, 108, 0, 0, 0, 108, 0, 122, 0, 122, 0, 0, 0, 109, 0, 97, 0, 99, 0, 0, 0, 109, 0, 97, 0, 100, 0, 0, 0, 109, 0, 97, 0, 102, 0, 0, 0, 109, 0, 97, 0, 103, 0, 0, 0, 109, 0, 97, 0, 111, 0, 0, 0, 109, 0, 97, 0, 115, 0, 0, 0, 109, 0, 97, 0, 119, 0, 0, 0, 109, 0, 97, 0, 121, 0, 0, 0, 109, 0, 97, 0, 122, 0, 0, 0, 109, 0, 98, 0, 111, 0, 0, 0, 109, 0, 98, 0, 113, 0, 0, 0, 109, 0, 98, 0, 117, 0, 0, 0, 109, 0, 98, 0, 119, 0, 0, 0, 109, 0, 99, 0, 105, 0, 0, 0, 109, 0, 99, 0, 112, 0, 0, 0, 109, 0, 99, 0, 113, 0, 0, 0, 109, 0, 99, 0, 114, 0, 0, 0, 109, 0, 99, 0, 117, 0, 0, 0, 109, 0, 100, 0, 97, 0, 0, 0, 109, 0, 100, 0, 102, 0, 0, 0, 109, 0, 100, 0, 104, 0, 0, 0, 109, 0, 100, 0, 106, 0, 0, 0, 109, 0, 100, 0, 114, 0, 0, 0, 109, 0, 100, 0, 120, 0, 0, 0, 109, 0, 101, 0, 100, 0, 0, 0, 109, 0, 101, 0, 103, 0, 0, 0, 109, 0, 101, 0, 107, 0, 0, 0, 109, 0, 101, 0, 114, 0, 0, 0, 109, 0, 101, 0, 116, 0, 0, 0, 109, 0, 102, 0, 101, 0, 0, 0, 109, 0, 102, 0, 110, 0, 0, 0, 109, 0, 102, 0, 113, 0, 0, 0, 109, 0, 103, 0, 104, 0, 0, 0, 109, 0, 103, 0, 108, 0, 0, 0, 109, 0, 103, 0, 111, 0, 0, 0, 109, 0, 103, 0, 112, 0, 0, 0, 109, 0, 103, 0, 121, 0, 0, 0, 109, 0, 104, 0, 105, 0, 0, 0, 109, 0, 104, 0, 108, 0, 0, 0, 109, 0, 104, 0, 114, 0, 0, 0, 109, 0, 105, 0, 102, 0, 0, 0, 109, 0, 105, 0, 110, 0, 0, 0, 109, 0, 107, 0, 105, 0, 0, 0, 109, 0, 107, 0, 108, 0, 0, 0, 109, 0, 107, 0, 112, 0, 0, 0, 109, 0, 108, 0, 101, 0, 0, 0, 109, 0, 108, 0, 112, 0, 0, 0, 109, 0, 108, 0, 115, 0, 0, 0, 109, 0, 109, 0, 111, 0, 0, 0, 109, 0, 109, 0, 117, 0, 0, 0, 109, 0, 109, 0, 120, 0, 0, 0, 109, 0, 110, 0, 97, 0, 0, 0, 109, 0, 110, 0, 102, 0, 0, 0, 109, 0, 110, 0, 107, 0, 0, 0, 109, 0, 111, 0, 97, 0, 0, 0, 109, 0, 111, 0, 101, 0, 0, 0, 109, 0, 111, 0, 104, 0, 0, 0, 109, 0, 111, 0, 109, 0, 0, 0, 109, 0, 111, 0, 115, 0, 0, 0, 109, 0, 111, 0, 120, 0, 0, 0, 109, 0, 112, 0, 112, 0, 0, 0, 109, 0, 112, 0, 115, 0, 0, 0, 109, 0, 112, 0, 116, 0, 0, 0, 109, 0, 112, 0, 120, 0, 0, 0, 109, 0, 113, 0, 108, 0, 0, 0, 109, 0, 114, 0, 100, 0, 0, 0, 109, 0, 114, 0, 106, 0, 0, 0, 109, 0, 114, 0, 111, 0, 0, 0, 109, 0, 115, 0, 116, 0, 0, 0, 109, 0, 116, 0, 99, 0, 0, 0, 109, 0, 116, 0, 102, 0, 0, 0, 109, 0, 116, 0, 105, 0, 0, 0, 109, 0, 116, 0, 109, 0, 0, 0, 109, 0, 116, 0, 114, 0, 0, 0, 109, 0, 117, 0, 97, 0, 0, 0, 109, 0, 117, 0, 112, 0, 0, 0, 109, 0, 117, 0, 114, 0, 0, 0, 109, 0, 117, 0, 115, 0, 0, 0, 109, 0, 118, 0, 97, 0, 0, 0, 109, 0, 118, 0, 110, 0, 0, 0, 109, 0, 118, 0, 121, 0, 0, 0, 109, 0, 119, 0, 106, 0, 0, 0, 109, 0, 119, 0, 107, 0, 0, 0, 109, 0, 119, 0, 118, 0, 0, 0, 109, 0, 119, 0, 119, 0, 0, 0, 109, 0, 120, 0, 99, 0, 0, 0, 109, 0, 120, 0, 109, 0, 0, 0, 109, 0, 121, 0, 100, 0, 0, 0, 109, 0, 121, 0, 107, 0, 0, 0, 109, 0, 121, 0, 109, 0, 0, 0, 109, 0, 121, 0, 116, 0, 0, 0, 109, 0, 121, 0, 118, 0, 0, 0, 109, 0, 121, 0, 119, 0, 0, 0, 109, 0, 121, 0, 120, 0, 0, 0, 109, 0, 121, 0, 122, 0, 0, 0, 109, 0, 122, 0, 107, 0, 0, 0, 109, 0, 122, 0, 109, 0, 0, 0, 109, 0, 122, 0, 110, 0, 0, 0, 109, 0, 122, 0, 112, 0, 0, 0, 109, 0, 122, 0, 119, 0, 0, 0, 109, 0, 122, 0, 122, 0, 0, 0, 110, 0, 97, 0, 99, 0, 0, 0, 110, 0, 97, 0, 100, 0, 0, 0, 110, 0, 97, 0, 102, 0, 0, 0, 110, 0, 97, 0, 110, 0, 0, 0, 110, 0, 97, 0, 112, 0, 0, 0, 110, 0, 97, 0, 113, 0, 0, 0, 110, 0, 97, 0, 115, 0, 0, 0, 110, 0, 98, 0, 114, 0, 0, 0, 110, 0, 99, 0, 97, 0, 0, 0, 110, 0, 99, 0, 102, 0, 0, 0, 110, 0, 99, 0, 104, 0, 0, 0, 110, 0, 99, 0, 111, 0, 0, 0, 110, 0, 99, 0, 112, 0, 0, 0, 110, 0, 100, 0, 99, 0, 0, 0, 110, 0, 100, 0, 115, 0, 0, 0, 110, 0, 101, 0, 98, 0, 0, 0, 110, 0, 101, 0, 120, 0, 0, 0, 110, 0, 103, 0, 97, 0, 0, 0, 110, 0, 103, 0, 98, 0, 0, 0, 110, 0, 103, 0, 116, 0, 0, 0, 110, 0, 104, 0, 98, 0, 0, 0, 110, 0, 104, 0, 119, 0, 0, 0, 110, 0, 105, 0, 102, 0, 0, 0, 110, 0, 105, 0, 106, 0, 0, 0, 110, 0, 105, 0, 117, 0, 0, 0, 110, 0, 105, 0, 121, 0, 0, 0, 110, 0, 105, 0, 122, 0, 0, 0, 110, 0, 106, 0, 111, 0, 0, 0, 110, 0, 107, 0, 111, 0, 0, 0, 110, 0, 109, 0, 122, 0, 0, 0, 110, 0, 110, 0, 102, 0, 0, 0, 110, 0, 110, 0, 104, 0, 0, 0, 110, 0, 110, 0, 107, 0, 0, 0, 110, 0, 110, 0, 109, 0, 0, 0, 110, 0, 110, 0, 112, 0, 0, 0, 110, 0, 110, 0, 115, 0, 0, 0, 110, 0, 110, 0, 120, 0, 0, 0, 110, 0, 111, 0, 100, 0, 0, 0, 110, 0, 111, 0, 101, 0, 0, 0, 110, 0, 111, 0, 110, 0, 0, 0, 110, 0, 111, 0, 112, 0, 0, 0, 110, 0, 111, 0, 117, 0, 0, 0, 110, 0, 112, 0, 105, 0, 0, 0, 110, 0, 113, 0, 111, 0, 0, 0, 110, 0, 114, 0, 98, 0, 0, 0, 110, 0, 115, 0, 107, 0, 0, 0, 110, 0, 115, 0, 115, 0, 0, 0, 110, 0, 116, 0, 109, 0, 0, 0, 110, 0, 116, 0, 114, 0, 0, 0, 110, 0, 116, 0, 115, 0, 0, 0, 110, 0, 117, 0, 105, 0, 0, 0, 110, 0, 117, 0, 110, 0, 0, 0, 110, 0, 117, 0, 112, 0, 0, 0, 110, 0, 117, 0, 115, 0, 0, 0, 110, 0, 117, 0, 118, 0, 0, 0, 110, 0, 117, 0, 120, 0, 0, 0, 110, 0, 119, 0, 98, 0, 0, 0, 110, 0, 120, 0, 113, 0, 0, 0, 110, 0, 120, 0, 114, 0, 0, 0, 110, 0, 121, 0, 99, 0, 0, 0, 110, 0, 121, 0, 109, 0, 0, 0, 110, 0, 121, 0, 110, 0, 0, 0, 110, 0, 122, 0, 105, 0, 0, 0, 111, 0, 103, 0, 99, 0, 0, 0, 111, 0, 106, 0, 103, 0, 0, 0, 111, 0, 107, 0, 114, 0, 0, 0, 111, 0, 110, 0, 110, 0, 0, 0, 111, 0, 110, 0, 115, 0, 0, 0, 111, 0, 112, 0, 97, 0, 0, 0, 111, 0, 112, 0, 109, 0, 0, 0, 111, 0, 114, 0, 111, 0, 0, 0, 111, 0, 114, 0, 117, 0, 0, 0, 111, 0, 114, 0, 121, 0, 0, 0, 111, 0, 115, 0, 97, 0, 0, 0, 111, 0, 116, 0, 107, 0, 0, 0, 111, 0, 117, 0, 110, 0, 0, 0, 111, 0, 122, 0, 109, 0, 0, 0, 112, 0, 97, 0, 103, 0, 0, 0, 112, 0, 97, 0, 109, 0, 0, 0, 112, 0, 97, 0, 117, 0, 0, 0, 112, 0, 98, 0, 105, 0, 0, 0, 112, 0, 98, 0, 117, 0, 0, 0, 112, 0, 99, 0, 100, 0, 0, 0, 112, 0, 99, 0, 109, 0, 0, 0, 112, 0, 99, 0, 114, 0, 0, 0, 112, 0, 100, 0, 99, 0, 0, 0, 112, 0, 100, 0, 116, 0, 0, 0, 112, 0, 101, 0, 100, 0, 0, 0, 112, 0, 101, 0, 114, 0, 0, 0, 112, 0, 101, 0, 115, 0, 0, 0, 112, 0, 101, 0, 120, 0, 0, 0, 112, 0, 102, 0, 108, 0, 0, 0, 112, 0, 104, 0, 108, 0, 0, 0, 112, 0, 104, 0, 110, 0, 0, 0, 112, 0, 104, 0, 114, 0, 0, 0, 112, 0, 105, 0, 108, 0, 0, 0, 112, 0, 105, 0, 112, 0, 0, 0, 112, 0, 107, 0, 111, 0, 0, 0, 112, 0, 108, 0, 97, 0, 0, 0, 112, 0, 108, 0, 116, 0, 0, 0, 112, 0, 109, 0, 99, 0, 0, 0, 112, 0, 109, 0, 115, 0, 0, 0, 112, 0, 109, 0, 117, 0, 0, 0, 112, 0, 110, 0, 103, 0, 0, 0, 112, 0, 110, 0, 116, 0, 0, 0, 112, 0, 111, 0, 110, 0, 0, 0, 112, 0, 112, 0, 111, 0, 0, 0, 112, 0, 112, 0, 114, 0, 0, 0, 112, 0, 114, 0, 97, 0, 0, 0, 112, 0, 114, 0, 100, 0, 0, 0, 112, 0, 114, 0, 103, 0, 0, 0, 112, 0, 114, 0, 121, 0, 0, 0, 112, 0, 115, 0, 115, 0, 0, 0, 112, 0, 116, 0, 112, 0, 0, 0, 112, 0, 117, 0, 98, 0, 0, 0, 112, 0, 117, 0, 117, 0, 0, 0, 112, 0, 117, 0, 122, 0, 0, 0, 112, 0, 119, 0, 97, 0, 0, 0, 113, 0, 117, 0, 99, 0, 0, 0, 113, 0, 117, 0, 103, 0, 0, 0, 113, 0, 117, 0, 104, 0, 0, 0, 114, 0, 97, 0, 111, 0, 0, 0, 114, 0, 97, 0, 113, 0, 0, 0, 114, 0, 97, 0, 115, 0, 0, 0, 114, 0, 99, 0, 102, 0, 0, 0, 114, 0, 101, 0, 108, 0, 0, 0, 114, 0, 101, 0, 115, 0, 0, 0, 114, 0, 105, 0, 97, 0, 0, 0, 114, 0, 106, 0, 115, 0, 0, 0, 114, 0, 107, 0, 116, 0, 0, 0, 114, 0, 109, 0, 102, 0, 0, 0, 114, 0, 109, 0, 111, 0, 0, 0, 114, 0, 109, 0, 116, 0, 0, 0, 114, 0, 109, 0, 117, 0, 0, 0, 114, 0, 109, 0, 120, 0, 0, 0, 114, 0, 109, 0, 121, 0, 0, 0, 114, 0, 110, 0, 97, 0, 0, 0, 114, 0, 110, 0, 103, 0, 0, 0, 114, 0, 111, 0, 98, 0, 0, 0, 114, 0, 111, 0, 102, 0, 0, 0, 114, 0, 111, 0, 109, 0, 0, 0, 114, 0, 114, 0, 111, 0, 0, 0, 114, 0, 116, 0, 109, 0, 0, 0, 114, 0, 117, 0, 101, 0, 0, 0, 114, 0, 117, 0, 103, 0, 0, 0, 114, 0, 117, 0, 109, 0, 0, 0, 114, 0, 119, 0, 107, 0, 0, 0, 114, 0, 121, 0, 117, 0, 0, 0, 115, 0, 97, 0, 104, 0, 0, 0, 115, 0, 97, 0, 113, 0, 0, 0, 115, 0, 97, 0, 115, 0, 0, 0, 115, 0, 97, 0, 118, 0, 0, 0, 115, 0, 97, 0, 122, 0, 0, 0, 115, 0, 98, 0, 97, 0, 0, 0, 115, 0, 98, 0, 112, 0, 0, 0, 115, 0, 99, 0, 99, 0, 0, 0, 115, 0, 99, 0, 107, 0, 0, 0, 115, 0, 99, 0, 108, 0, 0, 0, 115, 0, 99, 0, 110, 0, 0, 0, 115, 0, 99, 0, 111, 0, 0, 0, 115, 0, 99, 0, 114, 0, 0, 0, 115, 0, 100, 0, 99, 0, 0, 0, 115, 0, 100, 0, 104, 0, 0, 0, 115, 0, 101, 0, 102, 0, 0, 0, 115, 0, 101, 0, 104, 0, 0, 0, 115, 0, 101, 0, 105, 0, 0, 0, 115, 0, 101, 0, 115, 0, 0, 0, 115, 0, 103, 0, 97, 0, 0, 0, 115, 0, 103, 0, 115, 0, 0, 0, 115, 0, 103, 0, 119, 0, 0, 0, 115, 0, 103, 0, 122, 0, 0, 0, 115, 0, 104, 0, 105, 0, 0, 0, 115, 0, 104, 0, 107, 0, 0, 0, 115, 0, 104, 0, 110, 0, 0, 0, 115, 0, 105, 0, 103, 0, 0, 0, 115, 0, 105, 0, 108, 0, 0, 0, 115, 0, 105, 0, 109, 0, 0, 0, 115, 0, 106, 0, 114, 0, 0, 0, 115, 0, 107, 0, 99, 0, 0, 0, 115, 0, 107, 0, 107, 0, 0, 0, 115, 0, 108, 0, 100, 0, 0, 0, 115, 0, 108, 0, 105, 0, 0, 0, 115, 0, 108, 0, 108, 0, 0, 0, 115, 0, 108, 0, 111, 0, 0, 0, 115, 0, 108, 0, 121, 0, 0, 0, 115, 0, 109, 0, 106, 0, 0, 0, 115, 0, 109, 0, 110, 0, 0, 0, 115, 0, 109, 0, 112, 0, 0, 0, 115, 0, 109, 0, 113, 0, 0, 0, 115, 0, 110, 0, 99, 0, 0, 0, 115, 0, 110, 0, 107, 0, 0, 0, 115, 0, 110, 0, 112, 0, 0, 0, 115, 0, 110, 0, 120, 0, 0, 0, 115, 0, 110, 0, 121, 0, 0, 0, 115, 0, 110, 0, 122, 0, 0, 0, 115, 0, 111, 0, 107, 0, 0, 0, 115, 0, 111, 0, 113, 0, 0, 0, 115, 0, 111, 0, 117, 0, 0, 0, 115, 0, 111, 0, 121, 0, 0, 0, 115, 0, 112, 0, 100, 0, 0, 0, 115, 0, 112, 0, 108, 0, 0, 0, 115, 0, 112, 0, 121, 0, 0, 0, 115, 0, 114, 0, 98, 0, 0, 0, 115, 0, 114, 0, 99, 0, 0, 0, 115, 0, 114, 0, 114, 0, 0, 0, 115, 0, 115, 0, 100, 0, 0, 0, 115, 0, 115, 0, 103, 0, 0, 0, 115, 0, 115, 0, 121, 0, 0, 0, 115, 0, 116, 0, 113, 0, 0, 0, 115, 0, 117, 0, 97, 0, 0, 0, 115, 0, 117, 0, 101, 0, 0, 0, 115, 0, 117, 0, 106, 0, 0, 0, 115, 0, 117, 0, 115, 0, 0, 0, 115, 0, 119, 0, 98, 0, 0, 0, 115, 0, 119, 0, 99, 0, 0, 0, 115, 0, 119, 0, 103, 0, 0, 0, 115, 0, 119, 0, 104, 0, 0, 0, 115, 0, 119, 0, 112, 0, 0, 0, 115, 0, 119, 0, 118, 0, 0, 0, 115, 0, 120, 0, 110, 0, 0, 0, 115, 0, 120, 0, 119, 0, 0, 0, 115, 0, 122, 0, 108, 0, 0, 0, 116, 0, 97, 0, 110, 0, 0, 0, 116, 0, 97, 0, 113, 0, 0, 0, 116, 0, 98, 0, 99, 0, 0, 0, 116, 0, 98, 0, 100, 0, 0, 0, 116, 0, 98, 0, 102, 0, 0, 0, 116, 0, 98, 0, 122, 0, 0, 0, 116, 0, 99, 0, 105, 0, 0, 0, 116, 0, 100, 0, 100, 0, 0, 0, 116, 0, 100, 0, 102, 0, 0, 0, 116, 0, 100, 0, 103, 0, 0, 0, 116, 0, 100, 0, 104, 0, 0, 0, 116, 0, 101, 0, 100, 0, 0, 0, 116, 0, 101, 0, 109, 0, 0, 0, 116, 0, 101, 0, 111, 0, 0, 0, 116, 0, 103, 0, 99, 0, 0, 0, 116, 0, 103, 0, 111, 0, 0, 0, 116, 0, 104, 0, 99, 0, 0, 0, 116, 0, 104, 0, 108, 0, 0, 0, 116, 0, 104, 0, 113, 0, 0, 0, 116, 0, 104, 0, 120, 0, 0, 0, 116, 0, 105, 0, 98, 0, 0, 0, 116, 0, 105, 0, 101, 0, 0, 0, 116, 0, 105, 0, 102, 0, 0, 0, 116, 0, 105, 0, 109, 0, 0, 0, 116, 0, 105, 0, 118, 0, 0, 0, 116, 0, 107, 0, 116, 0, 0, 0, 116, 0, 108, 0, 102, 0, 0, 0, 116, 0, 108, 0, 119, 0, 0, 0, 116, 0, 108, 0, 120, 0, 0, 0, 116, 0, 108, 0, 121, 0, 0, 0, 116, 0, 109, 0, 112, 0, 0, 0, 116, 0, 110, 0, 101, 0, 0, 0, 116, 0, 110, 0, 104, 0, 0, 0, 116, 0, 111, 0, 102, 0, 0, 0, 116, 0, 111, 0, 103, 0, 0, 0, 116, 0, 111, 0, 113, 0, 0, 0, 116, 0, 112, 0, 105, 0, 0, 0, 116, 0, 112, 0, 109, 0, 0, 0, 116, 0, 112, 0, 111, 0, 0, 0, 116, 0, 112, 0, 122, 0, 0, 0, 116, 0, 113, 0, 111, 0, 0, 0, 116, 0, 114, 0, 117, 0, 0, 0, 116, 0, 114, 0, 118, 0, 0, 0, 116, 0, 114, 0, 119, 0, 0, 0, 116, 0, 115, 0, 103, 0, 0, 0, 116, 0, 115, 0, 106, 0, 0, 0, 116, 0, 116, 0, 100, 0, 0, 0, 116, 0, 116, 0, 106, 0, 0, 0, 116, 0, 116, 0, 113, 0, 0, 0, 116, 0, 116, 0, 114, 0, 0, 0, 116, 0, 117, 0, 104, 0, 0, 0, 116, 0, 117, 0, 108, 0, 0, 0, 116, 0, 117, 0, 109, 0, 0, 0, 116, 0, 117, 0, 113, 0, 0, 0, 116, 0, 118, 0, 108, 0, 0, 0, 116, 0, 118, 0, 117, 0, 0, 0, 116, 0, 119, 0, 104, 0, 0, 0, 116, 0, 119, 0, 109, 0, 0, 0, 116, 0, 119, 0, 113, 0, 0, 0, 116, 0, 120, 0, 103, 0, 0, 0, 116, 0, 121, 0, 97, 0, 0, 0, 116, 0, 121, 0, 106, 0, 0, 0, 116, 0, 121, 0, 118, 0, 0, 0, 116, 0, 122, 0, 109, 0, 0, 0, 117, 0, 98, 0, 117, 0, 0, 0, 117, 0, 100, 0, 109, 0, 0, 0, 117, 0, 109, 0, 98, 0, 0, 0, 117, 0, 109, 0, 117, 0, 0, 0, 117, 0, 110, 0, 120, 0, 0, 0, 117, 0, 114, 0, 105, 0, 0, 0, 117, 0, 114, 0, 116, 0, 0, 0, 117, 0, 115, 0, 97, 0, 0, 0, 117, 0, 118, 0, 104, 0, 0, 0, 117, 0, 118, 0, 108, 0, 0, 0, 117, 0, 122, 0, 110, 0, 0, 0, 118, 0, 97, 0, 103, 0, 0, 0, 118, 0, 97, 0, 105, 0, 0, 0, 118, 0, 97, 0, 110, 0, 0, 0, 118, 0, 101, 0, 99, 0, 0, 0, 118, 0, 101, 0, 112, 0, 0, 0, 118, 0, 105, 0, 99, 0, 0, 0, 118, 0, 105, 0, 118, 0, 0, 0, 118, 0, 108, 0, 115, 0, 0, 0, 118, 0, 109, 0, 102, 0, 0, 0, 118, 0, 109, 0, 119, 0, 0, 0, 118, 0, 111, 0, 116, 0, 0, 0, 118, 0, 117, 0, 110, 0, 0, 0, 118, 0, 117, 0, 116, 0, 0, 0, 119, 0, 97, 0, 106, 0, 0, 0, 119, 0, 97, 0, 108, 0, 0, 0, 119, 0, 97, 0, 119, 0, 0, 0, 119, 0, 98, 0, 112, 0, 0, 0, 119, 0, 98, 0, 113, 0, 0, 0, 119, 0, 99, 0, 105, 0, 0, 0, 119, 0, 101, 0, 114, 0, 0, 0, 119, 0, 103, 0, 105, 0, 0, 0, 119, 0, 104, 0, 103, 0, 0, 0, 119, 0, 105, 0, 98, 0, 0, 0, 119, 0, 105, 0, 118, 0, 0, 0, 119, 0, 108, 0, 115, 0, 0, 0, 119, 0, 110, 0, 99, 0, 0, 0, 119, 0, 110, 0, 105, 0, 0, 0, 119, 0, 110, 0, 117, 0, 0, 0, 119, 0, 111, 0, 98, 0, 0, 0, 119, 0, 111, 0, 115, 0, 0, 0, 119, 0, 114, 0, 115, 0, 0, 0, 119, 0, 116, 0, 109, 0, 0, 0, 119, 0, 117, 0, 117, 0, 0, 0, 119, 0, 117, 0, 118, 0, 0, 0, 120, 0, 97, 0, 118, 0, 0, 0, 120, 0, 98, 0, 97, 0, 0, 0, 120, 0, 105, 0, 97, 0, 0, 0, 120, 0, 107, 0, 104, 0, 0, 0, 120, 0, 108, 0, 99, 0, 0, 0, 120, 0, 108, 0, 100, 0, 0, 0, 120, 0, 109, 0, 102, 0, 0, 0, 120, 0, 111, 0, 103, 0, 0, 0, 120, 0, 111, 0, 110, 0, 0, 0, 120, 0, 112, 0, 101, 0, 0, 0, 120, 0, 112, 0, 114, 0, 0, 0, 120, 0, 114, 0, 98, 0, 0, 0, 120, 0, 115, 0, 106, 0, 0, 0, 120, 0, 115, 0, 109, 0, 0, 0, 120, 0, 119, 0, 101, 0, 0, 0, 121, 0, 97, 0, 111, 0, 0, 0, 121, 0, 97, 0, 112, 0, 0, 0, 121, 0, 97, 0, 116, 0, 0, 0, 121, 0, 98, 0, 98, 0, 0, 0, 121, 0, 98, 0, 100, 0, 0, 0, 121, 0, 98, 0, 121, 0, 0, 0, 121, 0, 100, 0, 100, 0, 0, 0, 121, 0, 101, 0, 114, 0, 0, 0, 121, 0, 103, 0, 114, 0, 0, 0, 121, 0, 103, 0, 119, 0, 0, 0, 121, 0, 108, 0, 101, 0, 0, 0, 121, 0, 108, 0, 108, 0, 0, 0, 121, 0, 109, 0, 97, 0, 0, 0, 121, 0, 111, 0, 110, 0, 0, 0, 121, 0, 114, 0, 98, 0, 0, 0, 121, 0, 114, 0, 101, 0, 0, 0, 121, 0, 117, 0, 97, 0, 0, 0, 121, 0, 117, 0, 106, 0, 0, 0, 121, 0, 117, 0, 116, 0, 0, 0, 121, 0, 117, 0, 117, 0, 0, 0, 121, 0, 117, 0, 119, 0, 0, 0, 122, 0, 97, 0, 103, 0, 0, 0, 122, 0, 97, 0, 105, 0, 0, 0, 122, 0, 97, 0, 112, 0, 0, 0, 122, 0, 100, 0, 106, 0, 0, 0, 122, 0, 101, 0, 97, 0, 0, 0, 122, 0, 103, 0, 104, 0, 0, 0, 122, 0, 104, 0, 120, 0, 0, 0, 122, 0, 107, 0, 116, 0, 0, 0, 122, 0, 108, 0, 109, 0, 0, 0, 122, 0, 121, 0, 98, 0, 0, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 68, 0, 101, 0, 118, 0, 97, 0, 0, 0, 67, 0, 121, 0, 114, 0, 108, 0, 0, 0, 69, 0, 116, 0, 104, 0, 105, 0, 0, 0, 72, 0, 97, 0, 110, 0, 116, 0, 0, 0, 84, 0, 104, 0, 97, 0, 105, 0, 0, 0, 66, 0, 101, 0, 110, 0, 103, 0, 0, 0, 72, 0, 101, 0, 98, 0, 114, 0, 0, 0, 67, 0, 97, 0, 110, 0, 115, 0, 0, 0, 72, 0, 97, 0, 110, 0, 115, 0, 0, 0, 67, 0, 111, 0, 112, 0, 116, 0, 0, 0, 77, 0, 121, 0, 109, 0, 114, 0, 0, 0, 83, 0, 117, 0, 110, 0, 100, 0, 0, 0, 71, 0, 114, 0, 101, 0, 107, 0, 0, 0, 77, 0, 97, 0, 114, 0, 99, 0, 0, 0, 84, 0, 101, 0, 108, 0, 117, 0, 0, 0, 78, 0, 115, 0, 104, 0, 117, 0, 0, 0, 84, 0, 102, 0, 110, 0, 103, 0, 0, 0, 84, 0, 105, 0, 98, 0, 116, 0, 0, 0, 75, 0, 110, 0, 100, 0, 97, 0, 0, 0, 82, 0, 117, 0, 110, 0, 114, 0, 0, 0, 87, 0, 99, 0, 104, 0, 111, 0, 0, 0, 65, 0, 118, 0, 115, 0, 116, 0, 0, 0, 68, 0, 117, 0, 112, 0, 108, 0, 0, 0, 71, 0, 117, 0, 114, 0, 117, 0, 0, 0, 75, 0, 116, 0, 104, 0, 105, 0, 0, 0, 76, 0, 105, 0, 115, 0, 117, 0, 0, 0, 80, 0, 114, 0, 116, 0, 105, 0, 0, 0, 71, 0, 101, 0, 111, 0, 114, 0, 0, 0, 75, 0, 97, 0, 110, 0, 97, 0, 0, 0, 75, 0, 111, 0, 114, 0, 101, 0, 0, 0, 76, 0, 97, 0, 111, 0, 111, 0, 0, 0, 78, 0, 107, 0, 111, 0, 111, 0, 0, 0, 84, 0, 97, 0, 109, 0, 108, 0, 0, 0, 66, 0, 114, 0, 97, 0, 105, 0, 0, 0, 67, 0, 104, 0, 97, 0, 109, 0, 0, 0, 67, 0, 112, 0, 114, 0, 116, 0, 0, 0, 71, 0, 108, 0, 97, 0, 103, 0, 0, 0, 73, 0, 116, 0, 97, 0, 108, 0, 0, 0, 74, 0, 97, 0, 109, 0, 111, 0, 0, 0, 75, 0, 97, 0, 108, 0, 105, 0, 0, 0, 75, 0, 104, 0, 111, 0, 106, 0, 0, 0, 77, 0, 111, 0, 110, 0, 103, 0, 0, 0, 77, 0, 114, 0, 111, 0, 111, 0, 0, 0, 79, 0, 103, 0, 97, 0, 109, 0, 0, 0, 79, 0, 115, 0, 109, 0, 97, 0, 0, 0, 80, 0, 104, 0, 97, 0, 103, 0, 0, 0, 83, 0, 97, 0, 114, 0, 98, 0, 0, 0, 83, 0, 104, 0, 97, 0, 119, 0, 0, 0, 83, 0, 105, 0, 100, 0, 100, 0, 0, 0, 83, 0, 105, 0, 110, 0, 104, 0, 0, 0, 84, 0, 97, 0, 118, 0, 116, 0, 0, 0, 88, 0, 112, 0, 101, 0, 111, 0, 0, 0, 71, 0, 111, 0, 116, 0, 104, 0, 0, 0, 72, 0, 117, 0, 110, 0, 103, 0, 0, 0, 77, 0, 117, 0, 108, 0, 116, 0, 0, 0, 83, 0, 111, 0, 103, 0, 100, 0, 0, 0, 83, 0, 111, 0, 121, 0, 111, 0, 0, 0, 83, 0, 121, 0, 108, 0, 111, 0, 0, 0, 65, 0, 100, 0, 108, 0, 109, 0, 0, 0, 65, 0, 103, 0, 104, 0, 98, 0, 0, 0, 65, 0, 104, 0, 111, 0, 109, 0, 0, 0, 65, 0, 114, 0, 109, 0, 105, 0, 0, 0, 65, 0, 114, 0, 109, 0, 110, 0, 0, 0, 66, 0, 97, 0, 108, 0, 105, 0, 0, 0, 66, 0, 97, 0, 109, 0, 117, 0, 0, 0, 66, 0, 97, 0, 115, 0, 115, 0, 0, 0, 66, 0, 97, 0, 116, 0, 107, 0, 0, 0, 66, 0, 104, 0, 107, 0, 115, 0, 0, 0, 66, 0, 111, 0, 112, 0, 111, 0, 0, 0, 66, 0, 114, 0, 97, 0, 104, 0, 0, 0, 66, 0, 117, 0, 103, 0, 105, 0, 0, 0, 66, 0, 117, 0, 104, 0, 100, 0, 0, 0, 67, 0, 97, 0, 107, 0, 109, 0, 0, 0, 67, 0, 97, 0, 114, 0, 105, 0, 0, 0, 67, 0, 104, 0, 101, 0, 114, 0, 0, 0, 67, 0, 104, 0, 114, 0, 115, 0, 0, 0, 68, 0, 105, 0, 97, 0, 107, 0, 0, 0, 68, 0, 111, 0, 103, 0, 114, 0, 0, 0, 69, 0, 103, 0, 121, 0, 112, 0, 0, 0, 69, 0, 108, 0, 98, 0, 97, 0, 0, 0, 69, 0, 108, 0, 121, 0, 109, 0, 0, 0, 71, 0, 111, 0, 110, 0, 103, 0, 0, 0, 71, 0, 111, 0, 110, 0, 109, 0, 0, 0, 71, 0, 114, 0, 97, 0, 110, 0, 0, 0, 71, 0, 117, 0, 106, 0, 114, 0, 0, 0, 72, 0, 97, 0, 110, 0, 98, 0, 0, 0, 72, 0, 97, 0, 110, 0, 103, 0, 0, 0, 72, 0, 97, 0, 110, 0, 105, 0, 0, 0, 72, 0, 97, 0, 110, 0, 111, 0, 0, 0, 72, 0, 97, 0, 116, 0, 114, 0, 0, 0, 72, 0, 105, 0, 114, 0, 97, 0, 0, 0, 72, 0, 108, 0, 117, 0, 119, 0, 0, 0, 72, 0, 109, 0, 110, 0, 103, 0, 0, 0, 72, 0, 109, 0, 110, 0, 112, 0, 0, 0, 74, 0, 97, 0, 118, 0, 97, 0, 0, 0, 74, 0, 112, 0, 97, 0, 110, 0, 0, 0, 75, 0, 104, 0, 97, 0, 114, 0, 0, 0, 75, 0, 104, 0, 109, 0, 114, 0, 0, 0, 75, 0, 105, 0, 116, 0, 115, 0, 0, 0, 76, 0, 97, 0, 110, 0, 97, 0, 0, 0, 76, 0, 101, 0, 112, 0, 99, 0, 0, 0, 76, 0, 105, 0, 109, 0, 98, 0, 0, 0, 76, 0, 105, 0, 110, 0, 97, 0, 0, 0, 76, 0, 105, 0, 110, 0, 98, 0, 0, 0, 76, 0, 121, 0, 99, 0, 105, 0, 0, 0, 76, 0, 121, 0, 100, 0, 105, 0, 0, 0, 77, 0, 97, 0, 104, 0, 106, 0, 0, 0, 77, 0, 97, 0, 107, 0, 97, 0, 0, 0, 77, 0, 97, 0, 110, 0, 100, 0, 0, 0, 77, 0, 97, 0, 110, 0, 105, 0, 0, 0, 77, 0, 101, 0, 100, 0, 102, 0, 0, 0, 77, 0, 101, 0, 110, 0, 100, 0, 0, 0, 77, 0, 101, 0, 114, 0, 99, 0, 0, 0, 77, 0, 101, 0, 114, 0, 111, 0, 0, 0, 77, 0, 108, 0, 121, 0, 109, 0, 0, 0, 77, 0, 111, 0, 100, 0, 105, 0, 0, 0, 77, 0, 116, 0, 101, 0, 105, 0, 0, 0, 78, 0, 97, 0, 110, 0, 100, 0, 0, 0, 78, 0, 97, 0, 114, 0, 98, 0, 0, 0, 78, 0, 98, 0, 97, 0, 116, 0, 0, 0, 78, 0, 101, 0, 119, 0, 97, 0, 0, 0, 79, 0, 108, 0, 99, 0, 107, 0, 0, 0, 79, 0, 114, 0, 107, 0, 104, 0, 0, 0, 79, 0, 114, 0, 121, 0, 97, 0, 0, 0, 79, 0, 115, 0, 103, 0, 101, 0, 0, 0, 80, 0, 97, 0, 108, 0, 109, 0, 0, 0, 80, 0, 97, 0, 117, 0, 99, 0, 0, 0, 80, 0, 101, 0, 114, 0, 109, 0, 0, 0, 80, 0, 104, 0, 108, 0, 105, 0, 0, 0, 80, 0, 104, 0, 108, 0, 112, 0, 0, 0, 80, 0, 104, 0, 110, 0, 120, 0, 0, 0, 80, 0, 108, 0, 114, 0, 100, 0, 0, 0, 82, 0, 106, 0, 110, 0, 103, 0, 0, 0, 82, 0, 111, 0, 104, 0, 103, 0, 0, 0, 83, 0, 97, 0, 109, 0, 114, 0, 0, 0, 83, 0, 97, 0, 117, 0, 114, 0, 0, 0, 83, 0, 103, 0, 110, 0, 119, 0, 0, 0, 83, 0, 104, 0, 114, 0, 100, 0, 0, 0, 83, 0, 105, 0, 110, 0, 100, 0, 0, 0, 83, 0, 111, 0, 103, 0, 111, 0, 0, 0, 83, 0, 111, 0, 114, 0, 97, 0, 0, 0, 83, 0, 121, 0, 114, 0, 99, 0, 0, 0, 84, 0, 97, 0, 103, 0, 98, 0, 0, 0, 84, 0, 97, 0, 107, 0, 114, 0, 0, 0, 84, 0, 97, 0, 108, 0, 101, 0, 0, 0, 84, 0, 97, 0, 108, 0, 117, 0, 0, 0, 84, 0, 97, 0, 110, 0, 103, 0, 0, 0, 84, 0, 103, 0, 108, 0, 103, 0, 0, 0, 84, 0, 104, 0, 97, 0, 97, 0, 0, 0, 84, 0, 105, 0, 114, 0, 104, 0, 0, 0, 85, 0, 103, 0, 97, 0, 114, 0, 0, 0, 86, 0, 97, 0, 105, 0, 105, 0, 0, 0, 87, 0, 97, 0, 114, 0, 97, 0, 0, 0, 88, 0, 115, 0, 117, 0, 120, 0, 0, 0, 89, 0, 101, 0, 122, 0, 105, 0, 0, 0, 89, 0, 105, 0, 105, 0, 105, 0, 0, 0, 90, 0, 97, 0, 110, 0, 98, 0, 0, 0, 115, 0, 107, 0, 105, 0, 112, 0, 0, 0, 48, 0, 49, 0, 50, 0, 51, 0, 52, 0, 54, 0, 0, 0, 115, 0, 99, 0, 114, 0, 105, 0, 112, 0, 116, 0, 0, 0, 48, 0, 49, 0, 50, 0, 51, 0, 52, 0, 53, 0, 54, 0, 0, 0, 82, 1, 215, 6, 219, 6, 19, 7, 40, 4, 83, 7, 83, 14, 147, 5, 100, 3, 99, 7, 100, 3, 80, 25, 76, 3, 179, 7, 28, 4, 211, 7, 131, 21, 239, 7, 183, 15, 23, 8, 160, 4, 27, 8, 71, 19, 31, 8, 116, 3, 39, 8, 52, 4, 83, 8, 179, 4, 103, 8, 195, 4, 179, 8, 91, 8, 212, 5, 139, 3, 235, 8, 35, 11, 159, 9, 140, 4, 183, 9, 35, 5, 187, 9, 199, 4, 255, 9, 11, 4, 15, 10, 72, 3, 47, 10, 203, 17, 51, 10, 91, 10, 75, 10, 111, 5, 83, 10, 91, 6, 87, 10, 72, 3, 99, 10, 71, 5, 23, 4, 91, 20, 143, 10, 156, 4, 147, 10, 100, 4, 223, 10, 143, 3, 235, 10, 59, 5, 247, 10, 251, 10, 255, 10, 183, 3, 3, 11, 243, 10, 27, 11, 168, 3, 71, 11, 168, 0, 119, 11, 124, 4, 147, 3, 151, 3, 143, 11, 132, 4, 195, 11, 200, 3, 243, 11, 64, 3, 7, 12, 65, 1, 75, 12, 203, 10, 87, 12, 168, 4, 99, 12, 231, 4, 143, 12, 84, 4, 147, 12, 104, 3, 159, 12, 3, 4, 163, 12, 239, 4, 215, 12, 227, 4, 247, 12, 228, 3, 7, 13, 103, 19, 23, 13, 88, 4, 31, 13, 75, 11, 67, 13, 223, 4, 83, 13, 126, 24, 91, 13, 127, 13, 111, 13, 103, 5, 151, 13, 35, 14, 187, 13, 139, 19, 199, 13, 176, 3, 231, 13, 148, 4, 255, 13, 67, 12, 68, 4, 104, 4, 56, 4, 64, 4, 51, 14, 251, 3, 72, 6, 92, 3, 191, 1, 188, 3, 187, 14, 19, 22, 199, 14, 27, 15, 211, 14, 168, 3, 35, 15, 88, 3, 51, 15, 136, 4, 59, 15, 72, 4, 63, 15, 3, 5, 75, 15, 247, 15, 103, 15, 204, 3, 195, 5, 31, 9, 11, 5, 119, 3, 227, 15, 119, 12, 243, 15, 19, 4, 3, 16, 131, 5, 15, 5, 119, 3, 19, 5, 119, 3, 55, 16, 51, 9, 127, 16, 99, 20, 155, 16, 199, 18, 159, 16, 147, 20, 219, 16, 194, 1, 231, 16, 148, 3, 247, 16, 164, 4, 3, 17, 108, 3, 75, 17, 35, 10, 131, 17, 207, 4, 187, 17, 151, 3, 76, 6, 164, 3, 247, 17, 55, 5, 19, 18, 75, 5, 43, 18, 3, 4, 67, 18, 135, 7, 79, 18, 55, 5, 127, 18, 111, 6, 171, 18, 39, 4, 7, 19, 151, 18, 11, 19, 235, 5, 32, 4, 208, 3, 35, 19, 240, 3, 63, 19, 71, 5, 123, 19, 208, 5, 155, 19, 44, 4, 167, 19, 3, 4, 191, 19, 112, 4, 203, 19, 23, 7, 219, 19, 128, 3, 223, 19, 128, 3, 7, 20, 76, 4, 11, 20, 167, 13, 19, 20, 243, 19, 207, 3, 23, 5, 179, 3, 191, 4, 39, 20, 75, 16, 55, 20, 25, 25, 75, 20, 67, 20, 155, 3, 24, 4, 151, 20, 171, 20, 191, 20, 164, 3, 95, 4, 119, 13, 231, 20, 96, 3, 251, 20, 128, 4, 71, 21, 251, 3, 87, 21, 152, 4, 159, 21, 255, 4, 167, 21, 197, 1, 219, 21, 136, 3, 123, 5, 119, 3, 51, 22, 139, 22, 63, 22, 251, 3, 67, 22, 224, 3, 71, 22, 103, 20, 131, 3, 215, 22, 203, 1, 243, 3, 95, 22, 67, 6, 107, 22, 231, 22, 111, 22, 111, 14, 127, 5, 115, 5, 179, 22, 15, 4, 48, 3, 200, 3, 255, 22, 199, 10, 135, 5, 219, 4, 27, 23, 156, 3, 227, 3, 120, 4, 171, 23, 239, 9, 175, 23, 3, 7, 179, 23, 91, 23, 203, 23, 7, 5, 215, 23, 199, 21, 119, 6, 215, 4, 243, 23, 11, 4, 251, 23, 92, 3, 19, 24, 203, 16, 143, 6, 7, 18, 147, 6, 167, 4, 47, 24, 235, 3, 59, 24, 63, 24, 171, 4, 108, 3, 91, 24, 184, 3, 180, 18, 0, 0, 0, 0, 0, 0, 122, 27, 134, 27, 0, 0, 36, 4, 95, 24, 75, 0, 207, 6, 95, 24, 3, 0, 211, 6, 95, 24, 3, 0, 223, 6, 95, 24, 3, 0, 102, 24, 110, 24, 129, 0, 227, 6, 95, 24, 3, 0, 231, 6, 110, 24, 3, 0, 235, 6, 95, 24, 93, 0, 239, 6, 95, 24, 3, 0, 243, 6, 95, 24, 3, 0, 247, 6, 95, 24, 3, 0, 251, 6, 95, 24, 9, 0, 255, 6, 95, 24, 72, 0, 7, 7, 95, 24, 93, 0, 11, 7, 95, 24, 3, 0, 15, 7, 95, 24, 3, 0, 40, 4, 190, 24, 252, 0, 27, 7, 110, 24, 12, 0, 31, 7, 95, 24, 3, 0, 56, 6, 210, 24, 24, 0, 35, 7, 100, 24, 53, 1, 39, 7, 95, 24, 3, 0, 92, 4, 95, 24, 60, 0, 43, 7, 95, 24, 3, 0, 47, 7, 95, 24, 3, 0, 51, 7, 95, 24, 3, 0, 55, 7, 95, 24, 3, 0, 59, 7, 95, 24, 3, 0, 63, 7, 95, 24, 18, 0, 67, 7, 95, 24, 3, 0, 71, 7, 95, 24, 3, 0, 75, 7, 149, 25, 6, 0, 79, 7, 95, 24, 3, 0, 200, 3, 95, 24, 93, 0, 87, 7, 102, 27, 102, 0, 91, 7, 95, 24, 3, 0, 100, 3, 95, 24, 231, 0, 45, 25, 95, 24, 3, 0, 95, 7, 95, 24, 59, 1, 103, 7, 110, 24, 12, 0, 20, 4, 115, 24, 75, 0, 107, 7, 95, 24, 3, 0, 111, 7, 95, 24, 3, 0, 40, 25, 95, 24, 39, 0, 115, 7, 95, 24, 3, 0, 60, 6, 95, 24, 90, 0, 119, 7, 95, 24, 3, 0, 123, 7, 95, 24, 3, 0, 127, 7, 95, 24, 3, 0, 131, 7, 95, 24, 3, 0, 139, 7, 95, 24, 3, 0, 143, 7, 95, 24, 3, 0, 147, 7, 95, 24, 9, 0, 151, 7, 100, 24, 3, 0, 155, 7, 100, 24, 156, 0, 159, 7, 95, 24, 3, 0, 163, 7, 95, 24, 3, 0, 167, 7, 95, 24, 3, 0, 171, 7, 95, 24, 3, 0, 76, 3, 100, 24, 180, 0, 171, 24, 154, 25, 24, 0, 171, 24, 188, 26, 131, 1, 171, 24, 218, 26, 219, 0, 175, 7, 95, 24, 3, 0, 28, 4, 159, 25, 68, 1, 183, 7, 95, 24, 89, 1, 187, 7, 95, 24, 249, 0, 191, 7, 100, 24, 177, 0, 195, 7, 100, 24, 213, 0, 199, 7, 100, 24, 105, 0, 203, 7, 100, 24, 180, 0, 124, 6, 130, 24, 6, 0, 207, 7, 95, 24, 45, 0, 151, 5, 17, 27, 36, 0, 215, 7, 95, 24, 3, 0, 219, 7, 95, 24, 3, 0, 223, 7, 95, 24, 90, 0, 227, 7, 95, 24, 3, 0, 231, 7, 95, 24, 3, 0, 235, 7, 95, 24, 33, 0, 243, 7, 95, 24, 3, 0, 128, 6, 110, 24, 12, 0, 247, 7, 100, 24, 3, 0, 251, 7, 95, 24, 3, 0, 100, 25, 95, 24, 3, 0, 255, 7, 95, 24, 3, 0, 3, 8, 105, 24, 6, 0, 7, 8, 95, 24, 3, 0, 11, 8, 95, 24, 3, 0, 15, 8, 95, 24, 3, 0, 160, 4, 95, 24, 249, 0, 19, 8, 95, 24, 3, 0, 116, 3, 95, 24, 171, 0, 116, 3, 100, 24, 102, 0, 116, 3, 100, 24, 24, 0, 116, 3, 110, 24, 12, 0, 132, 6, 110, 24, 12, 0, 179, 4, 100, 24, 30, 0, 183, 4, 95, 24, 9, 0, 35, 8, 105, 24, 27, 0, 52, 4, 95, 24, 90, 0, 43, 8, 95, 24, 71, 1, 47, 8, 95, 24, 18, 0, 51, 8, 95, 24, 3, 0, 55, 8, 169, 25, 18, 0, 59, 8, 95, 24, 3, 0, 63, 8, 95, 24, 3, 0, 187, 4, 95, 24, 9, 0, 67, 8, 95, 24, 3, 0, 71, 8, 95, 24, 18, 0, 75, 8, 95, 24, 3, 0, 79, 8, 95, 24, 3, 0, 87, 8, 95, 24, 3, 0, 95, 8, 95, 24, 3, 0, 99, 8, 95, 24, 126, 0, 195, 4, 95, 24, 21, 0, 107, 8, 95, 24, 3, 0, 111, 8, 95, 24, 3, 0, 115, 8, 95, 24, 3, 0, 119, 8, 115, 24, 3, 0, 123, 8, 95, 24, 3, 0, 127, 8, 95, 24, 3, 0, 4, 6, 110, 24, 1, 2, 131, 8, 95, 24, 3, 0, 135, 8, 95, 24, 3, 0, 139, 8, 100, 24, 66, 0, 143, 8, 95, 24, 185, 1, 147, 8, 95, 24, 3, 0, 151, 8, 95, 24, 9, 0, 155, 8, 95, 24, 3, 0, 159, 8, 95, 24, 45, 0, 163, 8, 95, 24, 18, 0, 167, 8, 9, 25, 6, 0, 171, 8, 100, 24, 30, 0, 191, 4, 105, 24, 6, 0, 116, 4, 110, 24, 80, 1, 175, 8, 105, 24, 6, 0, 183, 8, 100, 24, 30, 0, 187, 8, 165, 24, 42, 0, 139, 3, 105, 24, 6, 0, 191, 8, 105, 24, 6, 0, 195, 8, 95, 24, 3, 0, 199, 8, 105, 24, 6, 0, 203, 8, 95, 24, 3, 0, 207, 8, 95, 24, 3, 0, 88, 6, 95, 24, 182, 1, 211, 8, 95, 24, 3, 0, 215, 8, 95, 24, 3, 0, 219, 8, 95, 24, 3, 0, 223, 8, 95, 24, 39, 0, 227, 8, 95, 24, 3, 0, 231, 8, 95, 24, 3, 0, 239, 8, 95, 24, 3, 0, 243, 8, 115, 24, 3, 0, 247, 8, 105, 24, 6, 0, 251, 8, 95, 24, 9, 0, 255, 8, 95, 24, 3, 0, 3, 9, 95, 24, 3, 0, 7, 9, 95, 24, 114, 0, 11, 9, 95, 24, 3, 0, 15, 9, 95, 24, 3, 0, 155, 5, 95, 24, 18, 0, 19, 9, 95, 24, 3, 0, 87, 3, 95, 24, 21, 0, 23, 9, 95, 24, 3, 0, 27, 9, 99, 25, 162, 0, 176, 5, 95, 24, 63, 0, 35, 9, 95, 24, 3, 0, 39, 9, 95, 24, 3, 0, 43, 9, 95, 24, 63, 0, 47, 9, 95, 24, 3, 0, 33, 3, 130, 24, 120, 0, 55, 9, 95, 24, 3, 0, 59, 9, 95, 24, 3, 0, 63, 9, 95, 24, 3, 0, 224, 3, 190, 24, 15, 0, 67, 9, 95, 24, 3, 0, 71, 9, 95, 24, 3, 0, 75, 9, 95, 24, 3, 0, 79, 9, 130, 24, 6, 0, 83, 9, 95, 24, 3, 0, 87, 9, 100, 24, 24, 0, 91, 9, 95, 24, 3, 0, 95, 9, 95, 24, 126, 0, 64, 6, 95, 24, 78, 0, 99, 9, 105, 24, 6, 0, 103, 9, 100, 24, 30, 0, 107, 9, 105, 24, 6, 0, 111, 9, 95, 24, 3, 0, 36, 3, 95, 24, 77, 1, 115, 9, 95, 24, 3, 0, 99, 3, 174, 25, 23, 1, 119, 9, 95, 24, 18, 0, 123, 9, 115, 24, 3, 0, 127, 9, 95, 24, 21, 0, 131, 9, 95, 24, 3, 0, 135, 9, 105, 24, 30, 0, 199, 4, 110, 24, 12, 0, 139, 9, 95, 24, 62, 1, 143, 9, 95, 24, 3, 0, 203, 4, 95, 24, 9, 0, 147, 9, 95, 24, 3, 0, 151, 9, 95, 24, 18, 0, 155, 9, 95, 24, 3, 0, 140, 4, 155, 24, 81, 0, 163, 9, 95, 24, 3, 0, 167, 9, 95, 24, 3, 0, 171, 9, 95, 24, 8, 1, 175, 9, 95, 24, 3, 0, 179, 9, 95, 24, 3, 0, 159, 5, 95, 24, 3, 0, 35, 5, 95, 24, 54, 0, 191, 9, 95, 24, 3, 0, 195, 9, 115, 24, 183, 0, 199, 9, 95, 24, 3, 0, 203, 9, 95, 24, 3, 0, 207, 9, 95, 24, 18, 0, 211, 9, 95, 24, 3, 0, 215, 9, 95, 24, 3, 0, 219, 9, 95, 24, 63, 0, 223, 9, 95, 24, 3, 0, 227, 9, 95, 24, 3, 0, 231, 9, 95, 24, 3, 0, 96, 4, 95, 24, 90, 0, 235, 9, 95, 24, 3, 0, 243, 9, 95, 24, 3, 0, 247, 9, 95, 24, 39, 0, 251, 9, 209, 25, 120, 0, 228, 5, 110, 24, 12, 0, 3, 10, 95, 24, 21, 0, 7, 10, 95, 24, 3, 0, 11, 10, 95, 24, 72, 0, 92, 6, 95, 24, 116, 1, 72, 3, 145, 24, 15, 0, 72, 3, 120, 24, 240, 0, 72, 3, 120, 24, 83, 1, 72, 3, 120, 24, 5, 1, 72, 3, 120, 24, 122, 1, 72, 3, 120, 24, 9, 0, 72, 3, 120, 24, 149, 1, 72, 3, 120, 24, 84, 0, 72, 3, 120, 24, 38, 1, 72, 3, 120, 24, 21, 0, 72, 3, 120, 24, 35, 1, 72, 3, 120, 24, 47, 1, 72, 3, 120, 24, 69, 0, 72, 3, 120, 24, 222, 0, 72, 3, 120, 24, 57, 0, 72, 3, 120, 24, 36, 0, 72, 3, 120, 24, 162, 0, 72, 3, 189, 25, 222, 0, 72, 3, 18, 26, 222, 0, 19, 10, 95, 24, 99, 0, 207, 4, 110, 24, 12, 0, 206, 24, 95, 24, 36, 0, 23, 10, 95, 24, 33, 0, 27, 10, 219, 25, 36, 0, 31, 10, 95, 24, 36, 0, 39, 10, 100, 24, 11, 1, 43, 10, 19, 25, 162, 0, 187, 3, 95, 24, 3, 0, 55, 10, 100, 24, 102, 0, 59, 10, 95, 24, 3, 0, 63, 10, 95, 24, 3, 0, 67, 10, 95, 24, 3, 0, 71, 10, 95, 24, 3, 0, 111, 5, 42, 27, 102, 0, 79, 10, 95, 24, 3, 0, 211, 4, 129, 25, 138, 0, 96, 6, 95, 24, 78, 0, 95, 10, 150, 24, 180, 0, 103, 10, 95, 24, 21, 0, 156, 4, 140, 24, 33, 0, 107, 10, 110, 24, 147, 0, 111, 10, 140, 24, 33, 0, 115, 10, 140, 24, 33, 0, 119, 10, 140, 24, 33, 0, 123, 10, 140, 24, 33, 0, 127, 10, 95, 24, 167, 1, 100, 4, 95, 24, 98, 1, 131, 10, 95, 24, 207, 0, 135, 10, 140, 24, 33, 0, 139, 10, 223, 26, 81, 0, 60, 4, 110, 24, 12, 0, 60, 4, 29, 25, 80, 1, 39, 3, 110, 24, 12, 0, 120, 4, 95, 24, 57, 0, 197, 24, 95, 24, 101, 1, 151, 10, 95, 24, 3, 0, 155, 10, 95, 24, 3, 0, 159, 10, 95, 24, 3, 0, 163, 10, 95, 24, 3, 0, 167, 10, 95, 24, 36, 0, 171, 10, 110, 24, 12, 0, 175, 10, 95, 24, 54, 0, 179, 10, 95, 24, 3, 0, 183, 10, 95, 24, 3, 0, 187, 10, 100, 24, 6, 0, 191, 10, 95, 24, 3, 0, 104, 3, 95, 24, 48, 0, 195, 10, 95, 24, 3, 0, 215, 4, 95, 24, 33, 0, 207, 10, 95, 24, 3, 0, 211, 10, 95, 24, 3, 0, 215, 10, 95, 24, 3, 0, 219, 10, 100, 24, 3, 0, 143, 3, 100, 24, 6, 0, 227, 10, 95, 24, 33, 0, 231, 10, 95, 24, 3, 0, 59, 5, 105, 24, 6, 0, 239, 10, 95, 24, 3, 0, 183, 3, 95, 24, 42, 0, 7, 11, 95, 24, 141, 0, 11, 11, 95, 24, 126, 0, 15, 11, 95, 24, 3, 0, 19, 11, 95, 24, 3, 0, 23, 11, 95, 24, 3, 0, 168, 3, 110, 24, 138, 0, 168, 3, 54, 25, 15, 0, 31, 11, 95, 24, 3, 0, 39, 11, 115, 24, 3, 0, 43, 11, 95, 24, 48, 0, 47, 11, 95, 24, 63, 0, 119, 3, 95, 24, 84, 0, 51, 11, 95, 24, 3, 0, 163, 5, 105, 24, 27, 0, 55, 11, 95, 24, 18, 0, 59, 11, 95, 24, 3, 0, 63, 11, 95, 24, 3, 0, 67, 11, 95, 24, 3, 0, 168, 0, 95, 24, 144, 0, 188, 1, 77, 27, 155, 1, 79, 11, 95, 24, 3, 0, 83, 11, 95, 24, 3, 0, 87, 11, 95, 24, 114, 0, 91, 11, 95, 24, 174, 0, 95, 11, 95, 24, 3, 0, 99, 11, 95, 24, 54, 0, 216, 5, 95, 24, 93, 0, 103, 11, 95, 24, 39, 0, 107, 11, 95, 24, 51, 0, 111, 11, 239, 25, 180, 0, 115, 11, 95, 24, 3, 0, 124, 4, 95, 24, 107, 1, 191, 3, 44, 25, 81, 0, 228, 3, 165, 24, 135, 0, 219, 4, 95, 24, 3, 0, 123, 11, 95, 24, 3, 0, 151, 3, 95, 24, 64, 2, 151, 3, 4, 25, 132, 0, 80, 3, 95, 24, 36, 0, 80, 3, 84, 25, 57, 0, 127, 11, 95, 24, 3, 0, 131, 11, 95, 24, 3, 0, 68, 6, 95, 24, 83, 3, 135, 11, 95, 24, 3, 0, 68, 3, 95, 24, 90, 0, 139, 11, 3, 26, 6, 0, 132, 4, 95, 24, 36, 0, 147, 11, 95, 24, 36, 0, 151, 11, 95, 24, 3, 0, 155, 11, 34, 25, 51, 0, 159, 11, 95, 24, 3, 0, 163, 11, 95, 24, 3, 0, 167, 11, 95, 24, 18, 0, 171, 11, 95, 24, 90, 0, 128, 3, 100, 24, 24, 0, 175, 11, 95, 24, 3, 0, 179, 11, 95, 24, 3, 0, 183, 11, 95, 24, 3, 0, 187, 11, 95, 24, 3, 0, 191, 11, 95, 24, 8, 1, 65, 1, 95, 24, 114, 0, 65, 1, 139, 25, 132, 0, 199, 11, 95, 24, 3, 0, 203, 11, 95, 24, 63, 0, 12, 6, 95, 24, 186, 0, 207, 11, 100, 24, 66, 0, 243, 3, 95, 24, 21, 0, 195, 3, 95, 24, 111, 0, 42, 3, 95, 24, 189, 0, 211, 11, 95, 24, 3, 0, 215, 11, 95, 24, 3, 0, 220, 5, 95, 24, 52, 2, 219, 11, 95, 24, 3, 0, 223, 11, 95, 24, 246, 0, 227, 11, 95, 24, 3, 0, 231, 11, 95, 24, 3, 0, 235, 11, 95, 24, 3, 0, 64, 3, 95, 24, 78, 0, 239, 11, 95, 24, 36, 0, 247, 11, 95, 24, 78, 0, 251, 11, 95, 24, 48, 0, 255, 11, 95, 24, 48, 0, 3, 12, 100, 24, 18, 0, 11, 12, 95, 24, 56, 1, 15, 12, 95, 24, 3, 0, 19, 12, 95, 24, 132, 0, 23, 12, 95, 24, 3, 0, 27, 12, 95, 24, 141, 0, 31, 12, 95, 24, 51, 0, 35, 12, 95, 24, 39, 0, 39, 12, 95, 24, 3, 0, 43, 12, 95, 24, 66, 0, 180, 5, 95, 24, 144, 0, 44, 6, 95, 24, 128, 1, 47, 12, 95, 24, 93, 0, 51, 12, 95, 24, 3, 0, 55, 12, 95, 24, 26, 1, 59, 12, 95, 24, 3, 0, 63, 12, 95, 24, 3, 0, 65, 25, 95, 24, 3, 0, 71, 12, 145, 24, 15, 0, 79, 12, 95, 24, 3, 0, 83, 12, 95, 24, 9, 0, 168, 4, 95, 24, 75, 0, 223, 4, 95, 24, 3, 0, 91, 12, 95, 24, 3, 0, 95, 12, 105, 24, 6, 0, 231, 4, 95, 24, 3, 0, 103, 12, 95, 24, 3, 0, 107, 12, 100, 24, 24, 0, 111, 12, 95, 24, 5, 1, 126, 25, 95, 24, 57, 0, 115, 12, 95, 24, 3, 0, 123, 12, 95, 24, 3, 0, 127, 12, 95, 24, 3, 0, 131, 12, 95, 24, 3, 0, 135, 12, 95, 24, 3, 0, 139, 12, 95, 24, 3, 0, 84, 4, 240, 24, 129, 0, 151, 12, 115, 24, 75, 0, 155, 12, 95, 24, 3, 0, 239, 4, 105, 24, 27, 0, 167, 12, 95, 24, 3, 0, 171, 12, 95, 24, 134, 1, 175, 12, 95, 24, 3, 0, 179, 12, 100, 24, 30, 0, 183, 12, 95, 24, 3, 0, 187, 12, 100, 24, 30, 0, 167, 5, 95, 24, 3, 0, 191, 12, 95, 24, 3, 0, 232, 5, 95, 24, 90, 0, 195, 12, 100, 24, 24, 0, 199, 12, 95, 24, 3, 0, 203, 12, 115, 24, 3, 0, 88, 4, 95, 24, 196, 2, 207, 12, 95, 24, 3, 0, 211, 12, 95, 24, 3, 0, 227, 4, 175, 24, 6, 0, 219, 12, 95, 24, 3, 0, 223, 12, 115, 24, 3, 0, 227, 12, 95, 24, 3, 0, 231, 12, 105, 24, 6, 0, 235, 12, 95, 24, 9, 0, 239, 12, 95, 24, 144, 0, 243, 12, 109, 25, 147, 0, 235, 4, 24, 25, 2, 1, 235, 4, 108, 26, 135, 0, 251, 12, 130, 24, 6, 0, 255, 12, 95, 24, 3, 0, 3, 13, 95, 24, 150, 0, 16, 6, 13, 26, 6, 0, 11, 13, 95, 24, 123, 0, 15, 13, 95, 24, 92, 1, 19, 13, 95, 24, 3, 0, 27, 13, 95, 24, 93, 0, 35, 13, 95, 24, 3, 0, 39, 13, 95, 24, 3, 0, 43, 13, 95, 24, 54, 0, 236, 5, 95, 24, 91, 2, 47, 13, 95, 24, 3, 0, 51, 13, 95, 24, 3, 0, 55, 13, 100, 24, 3, 0, 59, 13, 95, 24, 33, 0, 63, 13, 100, 24, 3, 0, 91, 3, 95, 24, 3, 0, 124, 3, 95, 24, 39, 0, 124, 3, 100, 24, 66, 0, 124, 3, 100, 24, 18, 0, 75, 25, 95, 24, 3, 0, 71, 13, 145, 24, 15, 0, 20, 25, 95, 24, 3, 0, 85, 25, 95, 24, 36, 0, 75, 13, 100, 24, 117, 0, 79, 13, 95, 24, 3, 0, 87, 13, 115, 24, 3, 0, 64, 4, 135, 24, 192, 0, 27, 4, 95, 24, 3, 0, 227, 24, 105, 24, 6, 0, 95, 13, 95, 24, 3, 0, 243, 4, 95, 24, 189, 0, 99, 13, 95, 24, 3, 0, 103, 13, 95, 24, 3, 0, 107, 13, 95, 24, 21, 0, 103, 5, 105, 24, 6, 0, 115, 13, 95, 24, 3, 0, 171, 5, 48, 26, 42, 0, 123, 13, 248, 26, 15, 0, 131, 13, 95, 24, 3, 0, 135, 13, 100, 24, 30, 0, 139, 13, 105, 24, 6, 0, 143, 13, 53, 26, 201, 0, 247, 4, 95, 24, 21, 0, 31, 4, 100, 24, 30, 0, 207, 24, 95, 24, 41, 1, 7, 4, 105, 24, 6, 0, 50, 25, 105, 24, 6, 0, 147, 13, 95, 24, 3, 0, 128, 4, 95, 24, 79, 2, 155, 13, 95, 24, 48, 0, 159, 13, 145, 24, 15, 0, 184, 5, 95, 24, 82, 2, 182, 24, 95, 24, 125, 1, 163, 13, 95, 24, 3, 0, 188, 5, 95, 24, 153, 0, 156, 6, 95, 24, 83, 3, 171, 13, 95, 24, 3, 0, 175, 13, 95, 24, 3, 0, 179, 13, 95, 24, 84, 0, 183, 13, 95, 24, 39, 0, 191, 13, 95, 24, 3, 0, 195, 13, 95, 24, 3, 0, 176, 3, 95, 24, 94, 2, 203, 13, 95, 24, 3, 0, 104, 4, 95, 24, 9, 0, 90, 25, 95, 24, 3, 0, 207, 13, 95, 24, 3, 0, 211, 13, 95, 24, 3, 0, 215, 13, 95, 24, 156, 0, 20, 6, 95, 24, 39, 0, 219, 13, 95, 24, 3, 0, 223, 13, 95, 24, 3, 0, 240, 5, 112, 27, 15, 0, 227, 13, 95, 24, 3, 0, 148, 4, 140, 24, 33, 0, 235, 13, 95, 24, 3, 0, 239, 13, 95, 24, 33, 0, 243, 13, 95, 24, 3, 0, 247, 13, 95, 24, 3, 0, 251, 13, 95, 24, 21, 0, 3, 14, 95, 24, 3, 0, 68, 4, 95, 24, 9, 0, 95, 25, 110, 24, 12, 0, 24, 6, 95, 24, 83, 3, 7, 14, 95, 24, 3, 0, 11, 14, 95, 24, 3, 0, 196, 3, 95, 24, 51, 0, 56, 4, 135, 24, 192, 0, 15, 14, 95, 24, 3, 0, 19, 14, 95, 24, 3, 0, 23, 14, 95, 24, 12, 0, 27, 14, 95, 24, 3, 0, 232, 3, 68, 26, 195, 0, 31, 14, 95, 24, 3, 0, 39, 14, 95, 24, 100, 2, 43, 14, 95, 24, 83, 3, 47, 14, 95, 24, 3, 0, 55, 14, 95, 24, 3, 0, 59, 14, 95, 24, 3, 0, 63, 14, 95, 24, 18, 0, 92, 3, 135, 24, 83, 3, 67, 14, 95, 24, 3, 0, 71, 14, 95, 24, 45, 0, 75, 14, 105, 24, 27, 0, 79, 14, 95, 24, 3, 0, 87, 14, 95, 24, 101, 1, 188, 3, 95, 24, 9, 0, 191, 1, 95, 24, 9, 0, 35, 4, 110, 24, 159, 0, 91, 14, 95, 24, 177, 0, 95, 14, 95, 24, 81, 0, 99, 14, 95, 24, 3, 0, 103, 14, 95, 24, 3, 0, 107, 14, 95, 24, 39, 0, 115, 14, 95, 24, 54, 0, 119, 14, 95, 24, 63, 0, 251, 4, 110, 24, 12, 0, 175, 5, 95, 24, 3, 0, 123, 14, 95, 24, 3, 0, 127, 14, 95, 24, 3, 0, 131, 14, 95, 24, 3, 0, 135, 14, 100, 24, 141, 0, 139, 14, 95, 24, 39, 0, 143, 14, 95, 24, 165, 0, 147, 14, 95, 24, 3, 0, 151, 14, 95, 24, 3, 0, 155, 14, 95, 24, 45, 0, 159, 14, 100, 24, 156, 0, 163, 14, 95, 24, 3, 0, 247, 3, 125, 24, 69, 0, 167, 14, 95, 24, 95, 1, 171, 14, 95, 24, 18, 0, 175, 14, 95, 24, 3, 0, 179, 14, 95, 24, 126, 0, 183, 14, 105, 24, 6, 0, 179, 5, 105, 24, 6, 0, 72, 4, 95, 24, 87, 0, 191, 14, 95, 24, 9, 0, 195, 14, 95, 24, 3, 0, 203, 14, 95, 24, 123, 0, 123, 3, 95, 24, 6, 0, 207, 14, 62, 27, 15, 0, 215, 14, 105, 24, 6, 0, 219, 14, 95, 24, 63, 0, 223, 14, 95, 24, 3, 0, 183, 5, 155, 24, 6, 0, 227, 14, 100, 24, 30, 0, 187, 5, 95, 24, 3, 0, 12, 4, 95, 24, 54, 0, 231, 14, 95, 24, 3, 0, 235, 14, 95, 24, 42, 0, 239, 14, 95, 24, 3, 0, 192, 5, 95, 24, 153, 0, 243, 14, 95, 24, 3, 0, 247, 14, 255, 24, 201, 0, 251, 14, 95, 24, 3, 0, 255, 14, 95, 24, 3, 0, 132, 3, 110, 24, 17, 1, 132, 3, 100, 24, 117, 0, 132, 3, 100, 24, 15, 0, 132, 3, 100, 24, 24, 0, 132, 3, 100, 24, 138, 0, 3, 15, 95, 24, 3, 0, 191, 5, 95, 24, 18, 0, 28, 6, 95, 24, 61, 2, 255, 4, 95, 24, 54, 0, 7, 15, 95, 24, 3, 0, 11, 15, 95, 24, 3, 0, 15, 15, 95, 24, 3, 0, 156, 5, 78, 26, 11, 1, 19, 15, 95, 24, 234, 0, 23, 15, 95, 24, 3, 0, 31, 15, 95, 24, 3, 0, 88, 3, 95, 24, 42, 0, 88, 3, 100, 24, 20, 1, 88, 3, 100, 24, 102, 0, 88, 3, 107, 27, 129, 0, 39, 15, 95, 24, 3, 0, 43, 15, 95, 24, 3, 0, 47, 15, 95, 24, 3, 0, 168, 5, 195, 24, 6, 0, 136, 4, 95, 24, 3, 0, 55, 15, 95, 24, 119, 1, 3, 5, 105, 24, 6, 0, 67, 15, 95, 24, 3, 0, 172, 3, 250, 24, 14, 1, 71, 15, 110, 24, 12, 0, 79, 15, 95, 24, 3, 0, 83, 15, 95, 24, 99, 0, 87, 15, 95, 24, 3, 0, 7, 5, 95, 24, 23, 1, 91, 15, 95, 24, 3, 0, 95, 15, 95, 24, 3, 0, 99, 15, 95, 24, 3, 0, 204, 3, 110, 24, 12, 0, 107, 15, 95, 24, 3, 0, 111, 15, 95, 24, 3, 0, 115, 15, 95, 24, 3, 0, 119, 15, 95, 24, 3, 0, 123, 15, 115, 24, 3, 0, 127, 15, 110, 24, 12, 0, 131, 15, 95, 24, 216, 0, 135, 15, 95, 24, 21, 0, 139, 15, 95, 24, 12, 0, 143, 15, 95, 24, 3, 0, 147, 15, 105, 24, 6, 0, 108, 4, 100, 24, 6, 0, 151, 15, 95, 24, 45, 0, 155, 15, 95, 24, 3, 0, 159, 15, 95, 24, 18, 0, 163, 15, 95, 24, 48, 0, 167, 15, 95, 24, 3, 0, 171, 15, 95, 24, 3, 0, 175, 15, 115, 24, 3, 0, 179, 15, 95, 24, 3, 0, 199, 5, 95, 24, 3, 0, 11, 5, 95, 24, 84, 0, 187, 15, 95, 24, 3, 0, 191, 15, 95, 24, 3, 0, 195, 15, 95, 24, 3, 0, 199, 15, 95, 24, 3, 0, 203, 15, 110, 24, 12, 0, 207, 15, 95, 24, 3, 0, 211, 15, 95, 24, 3, 0, 215, 15, 95, 24, 3, 0, 219, 15, 95, 24, 3, 0, 223, 15, 95, 24, 9, 0, 231, 15, 100, 24, 30, 0, 224, 5, 95, 24, 57, 0, 235, 15, 95, 24, 3, 0, 239, 15, 95, 24, 3, 0, 19, 4, 95, 24, 3, 0, 251, 15, 95, 24, 3, 0, 255, 15, 115, 24, 3, 0, 131, 5, 95, 24, 3, 0, 7, 16, 125, 24, 69, 0, 11, 16, 100, 24, 30, 0, 15, 16, 95, 24, 3, 0, 19, 16, 95, 24, 3, 0, 192, 3, 110, 24, 106, 2, 192, 3, 100, 24, 15, 0, 192, 3, 95, 24, 42, 0, 23, 16, 95, 24, 3, 0, 27, 16, 95, 24, 3, 0, 15, 5, 95, 24, 84, 0, 31, 16, 95, 24, 3, 0, 19, 5, 95, 24, 84, 0, 100, 6, 95, 24, 179, 1, 35, 16, 103, 26, 135, 0, 39, 16, 135, 24, 192, 0, 30, 25, 95, 24, 45, 0, 23, 5, 100, 24, 30, 0, 43, 16, 95, 24, 72, 0, 47, 16, 95, 24, 3, 0, 148, 5, 95, 24, 140, 1, 51, 16, 110, 24, 12, 0, 59, 16, 95, 24, 3, 0, 63, 16, 95, 24, 9, 0, 67, 16, 95, 24, 3, 0, 71, 16, 125, 24, 15, 0, 79, 16, 95, 24, 3, 0, 83, 16, 95, 24, 3, 0, 87, 16, 95, 24, 3, 0, 91, 16, 95, 24, 3, 0, 95, 16, 93, 26, 6, 0, 99, 16, 95, 24, 3, 0, 103, 16, 95, 24, 3, 0, 27, 5, 110, 24, 12, 0, 136, 6, 95, 24, 72, 0, 107, 16, 95, 24, 3, 0, 48, 6, 95, 24, 144, 0, 111, 16, 95, 24, 3, 0, 115, 16, 95, 24, 3, 0, 31, 5, 105, 24, 27, 0, 31, 5, 98, 26, 6, 0, 119, 16, 95, 24, 3, 0, 123, 16, 95, 24, 3, 0, 131, 16, 95, 24, 51, 0, 135, 16, 230, 24, 15, 0, 139, 16, 95, 24, 9, 0, 143, 16, 100, 24, 24, 0, 147, 16, 95, 24, 36, 0, 151, 16, 95, 24, 3, 0, 203, 5, 95, 24, 3, 0, 163, 16, 175, 24, 6, 0, 167, 16, 95, 24, 51, 0, 171, 16, 95, 24, 3, 0, 204, 5, 95, 24, 87, 0, 175, 16, 95, 24, 3, 0, 179, 16, 95, 24, 3, 0, 136, 25, 255, 24, 201, 0, 207, 5, 95, 24, 3, 0, 183, 16, 95, 24, 3, 0, 187, 16, 95, 24, 87, 0, 43, 4, 95, 24, 3, 0, 191, 16, 95, 24, 3, 0, 195, 16, 95, 24, 185, 1, 199, 16, 100, 24, 24, 0, 121, 25, 95, 24, 137, 1, 47, 4, 95, 24, 143, 1, 172, 5, 95, 24, 87, 0, 207, 16, 95, 24, 87, 0, 211, 16, 95, 24, 54, 0, 215, 16, 100, 24, 24, 0, 194, 1, 95, 24, 143, 1, 223, 16, 125, 24, 69, 0, 71, 3, 145, 24, 15, 0, 227, 16, 95, 24, 42, 0, 148, 3, 110, 24, 146, 1, 235, 16, 95, 24, 9, 0, 239, 16, 95, 24, 18, 0, 243, 16, 105, 24, 6, 0, 39, 5, 105, 24, 6, 0, 43, 5, 95, 24, 9, 0, 164, 4, 95, 24, 172, 2, 251, 16, 95, 24, 54, 0, 255, 16, 95, 24, 3, 0, 108, 3, 95, 24, 84, 0, 108, 3, 100, 24, 4, 2, 108, 3, 100, 24, 9, 0, 7, 17, 95, 24, 108, 0, 211, 5, 95, 24, 3, 0, 11, 17, 95, 24, 3, 0, 15, 17, 95, 24, 3, 0, 19, 17, 95, 24, 3, 0, 23, 17, 95, 24, 3, 0, 27, 17, 95, 24, 3, 0, 31, 17, 95, 24, 3, 0, 35, 17, 95, 24, 3, 0, 39, 17, 95, 24, 3, 0, 43, 17, 95, 24, 3, 0, 47, 17, 95, 24, 3, 0, 103, 3, 100, 24, 3, 0, 51, 17, 110, 24, 12, 0, 55, 17, 95, 24, 21, 0, 59, 17, 95, 24, 3, 0, 63, 17, 95, 24, 9, 0, 67, 17, 115, 24, 3, 0, 71, 17, 95, 24, 3, 0, 215, 5, 95, 24, 3, 0, 79, 17, 95, 24, 3, 0, 79, 3, 95, 24, 216, 0, 83, 17, 95, 24, 54, 0, 87, 17, 95, 24, 3, 0, 51, 4, 95, 24, 3, 0, 127, 3, 100, 24, 69, 0, 91, 17, 95, 24, 29, 1, 95, 17, 95, 24, 3, 0, 219, 5, 95, 24, 3, 0, 99, 17, 95, 24, 3, 0, 76, 4, 95, 24, 136, 2, 103, 17, 95, 24, 96, 0, 107, 17, 95, 24, 3, 0, 111, 17, 95, 24, 18, 0, 115, 17, 105, 24, 27, 0, 119, 17, 95, 24, 45, 0, 16, 4, 95, 24, 139, 2, 123, 17, 95, 24, 3, 0, 127, 17, 95, 24, 3, 0, 135, 17, 95, 24, 3, 0, 139, 17, 95, 24, 9, 0, 175, 3, 38, 26, 102, 0, 175, 3, 143, 26, 39, 0, 55, 4, 95, 24, 3, 0, 143, 17, 100, 24, 3, 0, 147, 17, 95, 24, 3, 0, 151, 17, 95, 24, 3, 0, 223, 5, 95, 24, 3, 0, 140, 6, 163, 26, 6, 0, 155, 17, 95, 24, 3, 0, 159, 17, 95, 24, 3, 0, 163, 17, 95, 24, 66, 0, 167, 17, 95, 24, 3, 0, 171, 17, 95, 24, 3, 0, 175, 17, 95, 24, 3, 0, 179, 17, 95, 24, 3, 0, 183, 17, 95, 24, 3, 0, 47, 5, 130, 24, 6, 0, 51, 5, 155, 24, 81, 0, 164, 3, 95, 24, 44, 1, 191, 17, 95, 24, 3, 0, 195, 17, 95, 24, 33, 0, 199, 17, 95, 24, 33, 0, 207, 17, 95, 24, 174, 0, 211, 17, 95, 24, 3, 0, 215, 17, 95, 24, 3, 0, 219, 17, 95, 24, 3, 0, 223, 17, 95, 24, 3, 0, 227, 17, 95, 24, 3, 0, 231, 17, 95, 24, 3, 0, 157, 24, 105, 24, 6, 0, 235, 17, 105, 24, 27, 0, 239, 17, 110, 24, 12, 0, 243, 17, 59, 25, 120, 0, 144, 6, 95, 24, 148, 2, 251, 17, 95, 24, 3, 0, 255, 17, 95, 24, 3, 0, 3, 18, 95, 24, 3, 0, 11, 18, 105, 24, 6, 0, 15, 18, 95, 24, 18, 0, 75, 5, 105, 24, 6, 0, 23, 18, 95, 24, 3, 0, 27, 18, 95, 24, 36, 0, 31, 18, 95, 24, 3, 0, 35, 18, 95, 24, 3, 0, 39, 18, 100, 24, 30, 0, 47, 18, 95, 24, 63, 0, 51, 18, 95, 24, 9, 0, 55, 18, 58, 26, 36, 0, 59, 18, 95, 24, 165, 0, 63, 18, 95, 24, 3, 0, 71, 18, 95, 24, 3, 0, 75, 18, 115, 24, 3, 0, 83, 18, 110, 24, 12, 0, 87, 18, 95, 24, 3, 0, 91, 18, 95, 24, 72, 0, 95, 18, 133, 26, 24, 0, 99, 18, 95, 24, 3, 0, 103, 18, 95, 24, 3, 0, 107, 18, 100, 24, 24, 0, 111, 18, 95, 24, 3, 0, 115, 18, 95, 24, 3, 0, 119, 18, 95, 24, 3, 0, 104, 6, 95, 24, 163, 2, 123, 18, 95, 24, 3, 0, 131, 18, 95, 24, 3, 0, 199, 3, 95, 24, 3, 0, 135, 18, 145, 24, 15, 0, 139, 18, 95, 24, 51, 0, 143, 18, 95, 24, 153, 0, 147, 18, 95, 24, 3, 0, 208, 3, 95, 24, 204, 0, 155, 18, 95, 24, 3, 0, 227, 5, 95, 24, 3, 0, 159, 18, 95, 24, 3, 0, 163, 18, 95, 24, 108, 0, 167, 18, 95, 24, 3, 0, 59, 4, 95, 24, 3, 0, 162, 24, 95, 24, 165, 0, 175, 18, 95, 24, 96, 0, 179, 18, 95, 24, 48, 0, 240, 3, 105, 24, 27, 0, 183, 18, 95, 24, 3, 0, 63, 5, 105, 24, 27, 0, 187, 18, 95, 24, 3, 0, 63, 3, 95, 24, 3, 0, 116, 25, 95, 24, 153, 0, 191, 18, 95, 24, 3, 0, 195, 18, 95, 24, 3, 0, 231, 5, 95, 24, 96, 0, 203, 18, 95, 24, 3, 0, 63, 4, 95, 24, 108, 0, 207, 18, 95, 24, 108, 0, 211, 18, 95, 24, 3, 0, 239, 5, 95, 24, 3, 0, 215, 18, 95, 24, 9, 0, 67, 4, 95, 24, 3, 0, 219, 18, 95, 24, 169, 2, 223, 18, 95, 24, 3, 0, 227, 18, 95, 24, 3, 0, 231, 18, 95, 24, 6, 0, 71, 4, 95, 24, 3, 0, 235, 18, 95, 24, 3, 0, 75, 4, 95, 24, 18, 0, 239, 18, 95, 24, 3, 0, 252, 5, 95, 24, 204, 0, 243, 18, 95, 24, 3, 0, 247, 18, 95, 24, 18, 0, 251, 18, 95, 24, 3, 0, 255, 18, 95, 24, 3, 0, 3, 19, 205, 24, 6, 0, 32, 4, 95, 24, 204, 0, 15, 19, 88, 26, 69, 0, 19, 19, 105, 24, 6, 0, 23, 19, 200, 24, 111, 0, 27, 19, 95, 24, 3, 0, 31, 19, 95, 24, 3, 0, 39, 19, 4, 25, 132, 0, 108, 6, 95, 24, 60, 0, 43, 19, 95, 24, 3, 0, 47, 19, 140, 24, 33, 0, 243, 5, 95, 24, 3, 0, 79, 4, 95, 24, 60, 0, 51, 19, 95, 24, 3, 0, 55, 19, 95, 24, 3, 0, 59, 19, 95, 24, 3, 0, 67, 19, 95, 24, 3, 0, 75, 19, 95, 24, 3, 0, 79, 19, 95, 24, 226, 2, 83, 19, 95, 24, 3, 0, 87, 19, 95, 24, 3, 0, 45, 3, 95, 24, 36, 0, 91, 19, 95, 24, 3, 0, 95, 19, 95, 24, 15, 0, 99, 19, 95, 24, 3, 0, 112, 6, 95, 24, 32, 1, 107, 19, 95, 24, 45, 0, 111, 19, 95, 24, 72, 0, 115, 19, 95, 24, 93, 0, 8, 4, 95, 24, 78, 0, 119, 19, 95, 24, 3, 0, 127, 19, 95, 24, 3, 0, 203, 3, 95, 24, 3, 0, 55, 25, 95, 24, 3, 0, 131, 19, 95, 24, 3, 0, 135, 19, 95, 24, 3, 0, 143, 19, 95, 24, 3, 0, 44, 4, 208, 26, 6, 0, 147, 19, 95, 24, 3, 0, 151, 19, 100, 24, 3, 0, 148, 6, 110, 24, 129, 0, 159, 19, 213, 26, 36, 0, 247, 5, 100, 24, 3, 0, 163, 19, 203, 26, 138, 0, 171, 19, 95, 24, 3, 0, 180, 3, 220, 24, 6, 0, 180, 3, 100, 24, 30, 0, 175, 19, 95, 24, 21, 0, 67, 5, 233, 26, 24, 0, 67, 5, 238, 26, 15, 0, 179, 19, 95, 24, 21, 0, 255, 3, 95, 24, 74, 1, 183, 19, 95, 24, 193, 2, 187, 19, 95, 24, 3, 0, 112, 4, 100, 24, 117, 0, 195, 19, 95, 24, 78, 0, 199, 19, 95, 24, 39, 0, 207, 19, 95, 24, 36, 0, 211, 19, 95, 24, 33, 0, 215, 19, 95, 24, 3, 0, 105, 25, 104, 25, 24, 0, 227, 19, 95, 24, 3, 0, 231, 19, 95, 24, 48, 0, 235, 19, 100, 24, 3, 0, 239, 19, 243, 26, 20, 1, 247, 19, 95, 24, 3, 0, 251, 19, 95, 24, 3, 0, 83, 4, 194, 25, 6, 0, 255, 19, 95, 24, 54, 0, 217, 24, 95, 24, 207, 0, 3, 20, 95, 24, 3, 0, 15, 20, 95, 24, 51, 0, 23, 20, 95, 24, 3, 0, 251, 5, 95, 24, 3, 0, 27, 20, 165, 24, 135, 0, 31, 20, 95, 24, 99, 0, 179, 3, 105, 24, 6, 0, 35, 20, 95, 24, 3, 0, 43, 20, 73, 26, 30, 0, 47, 20, 100, 24, 24, 0, 51, 20, 95, 24, 83, 3, 59, 20, 95, 24, 3, 0, 152, 24, 95, 24, 123, 0, 63, 20, 95, 24, 3, 0, 71, 20, 95, 24, 110, 1, 79, 20, 95, 24, 3, 0, 24, 4, 95, 24, 158, 1, 83, 20, 95, 24, 113, 1, 87, 20, 95, 24, 104, 1, 15, 25, 95, 24, 3, 0, 95, 20, 95, 24, 3, 0, 107, 20, 95, 24, 164, 1, 79, 5, 95, 24, 9, 0, 111, 20, 95, 24, 3, 0, 115, 20, 95, 24, 3, 0, 87, 4, 95, 24, 51, 0, 83, 5, 100, 24, 81, 0, 119, 20, 95, 24, 6, 0, 87, 5, 185, 24, 105, 0, 87, 5, 95, 24, 144, 0, 123, 20, 105, 24, 27, 0, 127, 20, 130, 24, 120, 0, 196, 5, 95, 24, 150, 0, 131, 20, 95, 24, 186, 0, 135, 20, 95, 24, 150, 0, 139, 20, 100, 24, 24, 0, 143, 20, 95, 24, 111, 0, 8, 6, 95, 24, 242, 1, 155, 20, 95, 24, 3, 0, 159, 20, 95, 24, 96, 0, 163, 20, 95, 24, 9, 0, 167, 20, 95, 24, 45, 0, 60, 25, 95, 24, 3, 0, 175, 20, 95, 24, 3, 0, 179, 20, 95, 24, 189, 0, 222, 24, 110, 24, 12, 0, 183, 20, 110, 24, 147, 0, 187, 20, 95, 24, 214, 2, 52, 6, 95, 24, 211, 2, 195, 20, 95, 24, 45, 0, 255, 5, 95, 24, 3, 0, 199, 20, 245, 24, 195, 0, 112, 3, 105, 24, 6, 0, 91, 4, 95, 24, 93, 0, 203, 20, 110, 24, 12, 0, 207, 20, 95, 24, 54, 0, 211, 20, 95, 24, 9, 0, 91, 5, 95, 24, 6, 0, 215, 20, 95, 24, 114, 0, 219, 20, 12, 27, 6, 0, 223, 20, 95, 24, 3, 0, 3, 6, 95, 24, 3, 0, 227, 20, 95, 24, 45, 0, 197, 1, 95, 24, 51, 0, 96, 3, 110, 24, 210, 0, 96, 3, 95, 24, 12, 0, 96, 3, 95, 24, 130, 2, 96, 3, 95, 24, 44, 1, 96, 3, 95, 24, 42, 0, 235, 20, 105, 24, 6, 0, 239, 20, 100, 24, 3, 0, 243, 20, 95, 24, 51, 0, 247, 20, 95, 24, 57, 0, 99, 4, 95, 24, 33, 0, 160, 3, 100, 24, 30, 0, 160, 3, 105, 24, 6, 0, 160, 3, 49, 25, 6, 0, 160, 3, 27, 27, 6, 0, 255, 20, 95, 24, 51, 0, 3, 21, 100, 24, 24, 0, 152, 5, 95, 24, 204, 0, 7, 21, 95, 24, 126, 0, 11, 21, 95, 24, 96, 0, 15, 21, 95, 24, 108, 0, 19, 21, 95, 24, 63, 0, 80, 6, 95, 24, 86, 1, 23, 21, 64, 25, 128, 1, 27, 21, 95, 24, 137, 1, 31, 21, 115, 24, 3, 0, 35, 21, 95, 24, 3, 0, 39, 21, 185, 24, 105, 0, 43, 21, 95, 24, 3, 0, 47, 21, 155, 24, 81, 0, 181, 24, 100, 24, 3, 0, 116, 6, 94, 25, 118, 2, 103, 4, 95, 24, 75, 0, 51, 21, 95, 24, 3, 0, 55, 21, 95, 24, 3, 0, 59, 21, 95, 24, 3, 0, 63, 21, 95, 24, 3, 0, 152, 4, 95, 24, 170, 1, 67, 21, 95, 24, 3, 0, 95, 5, 100, 24, 30, 0, 107, 4, 95, 24, 3, 0, 120, 6, 95, 24, 217, 2, 75, 21, 95, 24, 3, 0, 79, 21, 95, 24, 207, 0, 83, 21, 95, 24, 3, 0, 91, 21, 95, 24, 9, 0, 172, 4, 95, 24, 21, 3, 70, 25, 95, 24, 111, 0, 95, 21, 95, 24, 111, 0, 99, 21, 95, 24, 186, 0, 103, 21, 7, 27, 192, 0, 107, 21, 95, 24, 3, 0, 107, 3, 95, 24, 186, 0, 244, 5, 95, 24, 165, 0, 111, 21, 95, 24, 3, 0, 115, 21, 95, 24, 63, 0, 119, 21, 95, 24, 3, 0, 123, 21, 95, 24, 3, 0, 127, 21, 95, 24, 3, 0, 80, 4, 95, 24, 173, 1, 99, 5, 124, 25, 159, 0, 135, 21, 95, 24, 3, 0, 139, 21, 95, 24, 3, 0, 143, 21, 125, 24, 69, 0, 147, 21, 95, 24, 3, 0, 151, 21, 95, 24, 3, 0, 155, 21, 95, 24, 3, 0, 111, 4, 95, 24, 3, 0, 163, 21, 37, 27, 6, 0, 7, 6, 95, 24, 47, 1, 171, 21, 95, 24, 114, 0, 152, 6, 95, 24, 60, 0, 175, 21, 95, 24, 3, 0, 179, 21, 95, 24, 3, 0, 183, 21, 95, 24, 183, 0, 212, 24, 95, 24, 60, 0, 211, 3, 95, 24, 3, 0, 187, 21, 95, 24, 48, 0, 232, 24, 95, 24, 9, 0, 191, 21, 95, 24, 3, 0, 195, 21, 95, 24, 3, 0, 215, 3, 95, 24, 45, 0, 219, 3, 95, 24, 3, 0, 203, 21, 95, 24, 132, 0, 200, 1, 95, 24, 111, 0, 136, 3, 95, 24, 45, 0, 207, 21, 100, 24, 62, 1, 211, 21, 95, 24, 87, 0, 215, 21, 95, 24, 48, 0, 223, 21, 95, 24, 3, 0, 227, 21, 105, 24, 6, 0, 231, 21, 95, 24, 9, 0, 235, 21, 95, 24, 3, 0, 107, 5, 130, 24, 120, 0, 239, 21, 95, 24, 207, 0, 248, 5, 9, 25, 6, 0, 115, 5, 105, 24, 27, 0, 35, 25, 95, 24, 3, 0, 243, 21, 95, 24, 3, 0, 247, 21, 95, 24, 3, 0, 251, 21, 95, 24, 3, 0, 255, 21, 95, 24, 3, 0, 3, 22, 95, 24, 3, 0, 115, 4, 95, 24, 3, 0, 223, 3, 95, 24, 3, 0, 119, 5, 95, 24, 21, 0, 7, 22, 95, 24, 3, 0, 11, 22, 95, 24, 3, 0, 119, 4, 195, 24, 6, 0, 15, 22, 57, 27, 15, 0, 23, 22, 105, 24, 27, 0, 27, 22, 105, 24, 27, 0, 123, 5, 95, 24, 84, 0, 32, 6, 175, 24, 6, 0, 31, 22, 95, 24, 3, 0, 35, 22, 95, 24, 216, 0, 39, 22, 95, 24, 72, 0, 123, 4, 95, 24, 50, 1, 11, 6, 95, 24, 3, 0, 48, 4, 110, 24, 176, 1, 48, 4, 100, 24, 30, 0, 43, 22, 95, 24, 3, 0, 47, 22, 95, 24, 3, 0, 15, 6, 95, 24, 3, 0, 111, 25, 125, 24, 69, 0, 55, 22, 105, 24, 27, 0, 59, 22, 105, 24, 27, 0, 127, 4, 105, 24, 27, 0, 237, 24, 115, 24, 75, 0, 75, 22, 95, 24, 3, 0, 19, 6, 115, 24, 183, 0, 131, 4, 95, 24, 3, 0, 79, 22, 95, 24, 3, 0, 23, 6, 95, 24, 3, 0, 83, 22, 95, 24, 39, 0, 212, 3, 95, 24, 250, 2, 27, 6, 95, 24, 247, 2, 135, 4, 95, 24, 171, 0, 87, 22, 105, 24, 27, 0, 203, 1, 95, 24, 21, 0, 91, 22, 95, 24, 3, 0, 99, 22, 95, 24, 3, 0, 103, 22, 95, 24, 171, 0, 15, 4, 95, 24, 141, 0, 139, 4, 95, 24, 3, 0, 97, 24, 95, 24, 60, 0, 115, 22, 95, 24, 3, 0, 200, 5, 95, 24, 253, 2, 119, 22, 95, 24, 3, 0, 123, 22, 95, 24, 32, 1, 127, 22, 95, 24, 3, 0, 131, 22, 95, 24, 41, 1, 135, 22, 95, 24, 3, 0, 143, 22, 95, 24, 3, 0, 147, 22, 95, 24, 3, 0, 144, 4, 95, 24, 42, 0, 151, 22, 95, 24, 42, 0, 155, 22, 95, 24, 222, 0, 159, 22, 100, 24, 3, 0, 36, 6, 95, 24, 60, 0, 159, 3, 165, 24, 135, 0, 127, 5, 105, 24, 27, 0, 163, 22, 95, 24, 21, 0, 167, 22, 190, 24, 252, 0, 135, 3, 95, 24, 3, 0, 40, 6, 110, 24, 12, 0, 171, 22, 95, 24, 3, 0, 31, 6, 95, 24, 3, 0, 175, 22, 95, 24, 72, 0, 183, 22, 95, 24, 3, 0, 35, 6, 125, 24, 69, 0, 39, 6, 95, 24, 171, 0, 187, 22, 95, 24, 3, 0, 191, 22, 95, 24, 3, 0, 195, 22, 95, 24, 32, 1, 199, 22, 95, 24, 3, 0, 203, 22, 95, 24, 3, 3, 207, 22, 95, 24, 3, 0, 211, 22, 95, 24, 3, 0, 219, 22, 95, 24, 141, 0, 223, 22, 67, 27, 15, 0, 164, 5, 95, 24, 38, 1, 227, 22, 95, 24, 3, 0, 235, 22, 110, 24, 12, 0, 239, 22, 95, 24, 105, 0, 243, 22, 95, 24, 3, 0, 247, 22, 110, 24, 12, 0, 236, 3, 100, 24, 15, 0, 236, 3, 110, 24, 17, 1, 236, 3, 110, 24, 138, 0, 43, 6, 87, 27, 219, 0, 216, 3, 110, 24, 147, 0, 47, 6, 95, 24, 99, 0, 251, 22, 95, 24, 234, 0, 80, 3, 95, 24, 39, 0, 80, 3, 95, 24, 240, 0, 68, 3, 95, 24, 108, 0, 136, 3, 95, 24, 87, 0, 80, 3, 95, 24, 60, 0, 68, 3, 95, 24, 22, 2, 80, 3, 95, 24, 41, 1, 80, 3, 95, 24, 116, 1, 156, 3, 95, 24, 159, 0, 76, 3, 100, 24, 213, 0, 80, 3, 95, 24, 57, 0, 68, 3, 95, 24, 175, 4, 255, 3, 95, 24, 255, 0, 76, 3, 100, 24, 66, 0, 76, 3, 100, 24, 228, 0, 96, 4, 95, 24, 221, 1, 76, 3, 100, 24, 224, 1, 128, 3, 100, 24, 117, 0, 152, 24, 95, 24, 234, 0, 161, 24, 95, 24, 237, 0, 68, 3, 95, 24, 230, 1, 172, 4, 95, 24, 233, 1, 104, 3, 95, 24, 71, 1, 168, 0, 95, 24, 74, 1, 200, 1, 95, 24, 236, 1, 168, 0, 95, 24, 243, 0, 64, 3, 95, 24, 174, 0, 76, 3, 100, 24, 239, 1, 64, 3, 95, 24, 246, 0, 64, 3, 95, 24, 245, 1, 108, 3, 95, 24, 83, 1, 68, 3, 95, 24, 249, 0, 255, 3, 95, 24, 248, 1, 161, 24, 95, 24, 254, 1, 64, 3, 95, 24, 86, 1, 64, 3, 95, 24, 7, 2, 104, 3, 95, 24, 150, 0, 64, 3, 95, 24, 126, 0, 68, 3, 95, 24, 89, 1, 64, 3, 95, 24, 18, 0, 68, 3, 95, 24, 92, 1, 161, 24, 95, 24, 10, 2, 68, 3, 95, 24, 13, 2, 152, 24, 95, 24, 95, 1, 228, 3, 165, 24, 2, 1, 36, 4, 95, 24, 31, 2, 68, 3, 95, 24, 34, 2, 76, 3, 100, 24, 177, 0, 68, 3, 95, 24, 40, 2, 68, 3, 95, 24, 104, 1, 76, 3, 100, 24, 43, 2, 237, 24, 115, 24, 183, 0, 104, 3, 95, 24, 49, 2, 64, 3, 95, 24, 110, 1, 64, 3, 95, 24, 5, 1, 64, 3, 95, 24, 132, 0, 64, 3, 95, 24, 67, 2, 68, 3, 95, 24, 8, 1, 161, 24, 95, 24, 70, 2, 68, 3, 95, 24, 113, 1, 152, 24, 95, 24, 119, 1, 161, 24, 95, 24, 73, 2, 68, 3, 95, 24, 76, 2, 68, 3, 95, 24, 88, 2, 76, 3, 100, 24, 102, 0, 76, 3, 100, 24, 131, 1, 136, 3, 95, 24, 54, 0, 76, 3, 100, 24, 198, 0, 172, 3, 250, 24, 109, 2, 76, 3, 100, 24, 112, 2, 222, 24, 110, 24, 17, 1, 76, 3, 100, 24, 20, 1, 104, 3, 95, 24, 115, 2, 212, 24, 95, 24, 121, 2, 64, 3, 95, 24, 140, 1, 76, 3, 100, 24, 124, 2, 76, 3, 100, 24, 105, 0, 64, 3, 95, 24, 127, 2, 164, 3, 95, 24, 26, 1, 64, 3, 95, 24, 133, 2, 64, 3, 95, 24, 145, 2, 76, 3, 100, 24, 152, 1, 152, 24, 95, 24, 96, 0, 92, 4, 95, 24, 153, 0, 64, 3, 95, 24, 151, 2, 124, 3, 95, 24, 141, 0, 68, 3, 95, 24, 157, 2, 76, 3, 100, 24, 175, 2, 68, 3, 95, 24, 35, 1, 68, 3, 95, 24, 158, 1, 64, 3, 95, 24, 38, 1, 220, 3, 100, 24, 30, 0, 64, 3, 95, 24, 181, 2, 68, 3, 95, 24, 184, 2, 76, 3, 100, 24, 187, 2, 152, 24, 95, 24, 161, 1, 76, 3, 100, 24, 202, 2, 80, 3, 95, 24, 28, 2, 64, 3, 95, 24, 164, 1, 64, 3, 95, 24, 167, 1, 208, 3, 95, 24, 220, 2, 196, 3, 95, 24, 223, 2, 64, 3, 95, 24, 114, 0, 168, 0, 95, 24, 47, 1, 152, 24, 95, 24, 229, 2, 68, 3, 95, 24, 235, 2, 76, 3, 100, 24, 219, 0, 64, 3, 95, 24, 241, 2, 64, 3, 95, 24, 244, 2, 64, 3, 95, 24, 156, 0, 152, 24, 95, 24, 50, 1, 76, 3, 100, 24, 53, 1, 136, 3, 95, 24, 72, 0, 68, 3, 95, 24, 9, 3, 196, 3, 95, 24, 179, 1, 54, 3, 95, 24, 162, 0, 68, 3, 95, 24, 15, 3, 64, 3, 95, 24, 56, 1, 100, 3, 95, 24, 59, 1, 64, 3, 95, 24, 62, 1, 27, 5, 144, 25, 12, 0, 108, 4, 100, 24, 57, 0, 220, 3, 100, 24, 6, 0, 220, 3, 100, 24, 29, 1, 124, 3, 100, 24, 39, 0, 128, 3, 100, 24, 176, 1, 116, 3, 100, 24, 42, 0, 183, 4, 164, 25, 9, 0, 187, 4, 179, 25, 9, 0, 112, 3, 184, 25, 6, 0, 64, 3, 14, 25, 78, 0, 203, 4, 199, 25, 9, 0, 87, 3, 204, 25, 21, 0, 155, 4, 214, 25, 42, 0, 95, 6, 224, 25, 159, 0, 148, 3, 110, 24, 231, 0, 96, 3, 110, 24, 77, 1, 148, 3, 110, 24, 135, 0, 216, 3, 110, 24, 26, 1, 116, 4, 110, 24, 44, 1, 216, 3, 110, 24, 170, 1, 251, 4, 110, 24, 42, 0, 96, 3, 110, 24, 59, 1, 240, 3, 105, 24, 252, 0, 243, 4, 105, 24, 189, 0, 139, 3, 105, 24, 29, 1, 188, 1, 229, 25, 155, 1, 143, 3, 234, 25, 6, 0, 64, 3, 215, 24, 78, 0, 100, 3, 244, 25, 231, 0, 171, 24, 249, 25, 24, 0, 79, 6, 254, 25, 6, 0, 112, 3, 8, 26, 6, 0, 172, 3, 23, 26, 14, 1, 72, 3, 28, 26, 15, 0, 247, 4, 33, 26, 21, 0, 92, 3, 135, 24, 33, 0, 92, 3, 135, 24, 57, 0, 92, 3, 135, 24, 111, 0, 92, 3, 135, 24, 147, 0, 92, 3, 135, 24, 36, 0, 232, 3, 43, 26, 195, 0, 182, 24, 114, 25, 125, 1, 172, 3, 39, 25, 14, 1, 188, 3, 63, 26, 9, 0, 232, 3, 245, 24, 195, 0, 83, 24, 83, 26, 15, 0, 139, 3, 225, 24, 6, 0, 80, 3, 95, 24, 75, 0, 212, 3, 95, 24, 117, 0, 88, 3, 95, 24, 68, 1, 184, 3, 95, 24, 15, 0, 144, 4, 95, 24, 2, 1, 64, 3, 95, 24, 177, 0, 88, 3, 95, 24, 129, 0, 212, 3, 95, 24, 24, 0, 64, 3, 95, 24, 198, 0, 64, 3, 95, 24, 105, 0, 100, 3, 95, 24, 146, 1, 152, 24, 95, 24, 149, 1, 64, 3, 95, 24, 152, 1, 64, 3, 95, 24, 219, 0, 64, 3, 95, 24, 53, 1, 217, 24, 95, 24, 147, 0, 183, 23, 113, 26, 42, 0, 187, 23, 118, 26, 42, 0, 227, 24, 123, 26, 6, 0, 43, 5, 128, 26, 9, 0, 167, 3, 138, 26, 15, 0, 224, 3, 170, 24, 15, 0, 79, 3, 148, 26, 216, 0, 139, 5, 153, 26, 66, 0, 139, 5, 158, 26, 66, 0, 157, 24, 168, 26, 6, 0, 47, 5, 173, 26, 6, 0, 95, 5, 119, 25, 30, 0, 51, 5, 155, 24, 69, 0, 112, 3, 178, 26, 6, 0, 103, 6, 183, 26, 213, 0, 63, 5, 193, 26, 27, 0, 79, 24, 180, 24, 15, 0, 91, 5, 198, 26, 6, 0, 80, 4, 69, 25, 173, 1, 204, 3, 228, 26, 12, 0, 71, 3, 74, 25, 15, 0, 207, 23, 235, 24, 24, 0, 79, 5, 253, 26, 9, 0, 83, 5, 2, 27, 81, 0, 111, 3, 79, 25, 228, 0, 112, 3, 22, 27, 6, 0, 112, 3, 89, 25, 6, 0, 99, 5, 32, 27, 159, 0, 232, 24, 160, 24, 9, 0, 107, 5, 134, 25, 120, 0, 119, 5, 47, 27, 21, 0, 143, 3, 52, 27, 6, 0, 75, 24, 185, 24, 105, 0, 243, 3, 72, 27, 21, 0, 247, 3, 125, 24, 11, 1, 247, 3, 125, 24, 201, 0, 39, 5, 82, 27, 6, 0, 35, 23, 92, 27, 23, 1, 7, 4, 97, 27, 6, 0, 211, 4, 117, 27, 138, 0, 201, 24, 130, 24, 6, 0, 201, 24, 105, 24, 27, 0, 3, 23, 130, 24, 6, 0, 135, 5, 95, 24, 3, 0, 7, 23, 95, 24, 3, 0, 11, 23, 95, 24, 3, 0, 51, 6, 95, 24, 3, 0, 15, 23, 95, 24, 3, 0, 143, 4, 95, 24, 3, 0, 19, 23, 95, 24, 3, 0, 23, 23, 95, 24, 3, 0, 156, 3, 100, 24, 117, 0, 156, 3, 110, 24, 15, 0, 31, 23, 95, 24, 3, 0, 39, 23, 95, 24, 3, 0, 51, 3, 95, 24, 60, 0, 43, 23, 95, 24, 51, 0, 47, 23, 95, 24, 12, 0, 51, 23, 95, 24, 238, 2, 55, 23, 95, 24, 3, 0, 59, 23, 95, 24, 243, 0, 63, 23, 95, 24, 48, 0, 67, 23, 95, 24, 96, 0, 57, 3, 95, 24, 83, 3, 71, 23, 95, 24, 12, 0, 163, 3, 95, 24, 107, 1, 75, 23, 95, 24, 45, 0, 79, 23, 95, 24, 3, 0, 84, 6, 95, 24, 243, 0, 55, 6, 95, 24, 150, 0, 83, 23, 95, 24, 3, 0, 87, 23, 115, 24, 75, 0, 59, 6, 95, 24, 3, 0, 75, 3, 95, 24, 21, 0, 95, 23, 95, 24, 240, 0, 99, 23, 175, 24, 6, 0, 63, 6, 105, 24, 6, 0, 103, 23, 95, 24, 3, 0, 107, 23, 95, 24, 3, 0, 111, 23, 95, 24, 3, 0, 115, 23, 95, 24, 3, 0, 119, 23, 95, 24, 3, 0, 147, 4, 95, 24, 3, 0, 123, 23, 95, 24, 3, 0, 231, 3, 95, 24, 3, 0, 71, 6, 95, 24, 3, 0, 127, 23, 95, 24, 56, 1, 75, 6, 95, 24, 3, 0, 131, 23, 95, 24, 3, 0, 135, 23, 100, 24, 198, 0, 139, 23, 95, 24, 3, 0, 0, 6, 95, 24, 114, 0, 143, 23, 95, 24, 3, 0, 147, 23, 95, 24, 3, 0, 151, 23, 95, 24, 3, 0, 151, 4, 95, 24, 3, 0, 155, 23, 105, 24, 6, 0, 159, 23, 145, 24, 15, 0, 163, 23, 95, 24, 3, 0, 83, 6, 95, 24, 3, 0, 167, 23, 95, 24, 123, 0, 87, 6, 95, 24, 3, 0, 67, 3, 95, 24, 3, 0, 160, 5, 95, 24, 60, 0, 99, 6, 95, 24, 3, 0, 191, 23, 240, 24, 129, 0, 107, 6, 105, 24, 6, 0, 195, 23, 95, 24, 72, 0, 199, 23, 95, 24, 3, 0, 211, 23, 95, 24, 3, 0, 115, 6, 95, 24, 3, 0, 219, 23, 95, 24, 3, 0, 95, 3, 105, 24, 27, 0, 223, 23, 95, 24, 3, 0, 227, 23, 95, 24, 96, 0, 231, 23, 95, 24, 99, 0, 123, 6, 95, 24, 3, 0, 235, 23, 95, 24, 3, 0, 127, 6, 95, 24, 18, 0, 159, 4, 95, 24, 3, 0, 115, 3, 95, 24, 3, 0, 131, 6, 95, 24, 3, 0, 239, 23, 95, 24, 18, 0, 247, 23, 95, 24, 3, 0, 255, 23, 95, 24, 3, 0, 3, 24, 95, 24, 3, 0, 7, 24, 95, 24, 3, 0, 171, 3, 95, 24, 3, 0, 11, 24, 95, 24, 3, 0, 135, 6, 95, 24, 3, 0, 15, 24, 95, 24, 3, 0, 139, 6, 95, 24, 3, 0, 131, 25, 95, 24, 39, 0, 23, 24, 95, 24, 3, 0, 27, 24, 95, 24, 3, 0, 31, 24, 95, 24, 3, 0, 111, 24, 95, 24, 123, 0, 151, 6, 95, 24, 3, 0, 35, 24, 95, 24, 108, 0, 143, 5, 120, 24, 122, 1, 143, 5, 145, 24, 15, 0, 39, 24, 95, 24, 3, 0, 43, 24, 95, 24, 3, 0, 51, 24, 95, 24, 3, 0, 55, 24, 95, 24, 66, 0, 67, 24, 100, 24, 198, 0, 71, 24, 95, 24, 144, 0, 155, 6, 95, 24, 3, 0, 87, 24, 95, 24, 156, 0, 163, 4, 95, 24, 84, 0, 239, 3, 95, 24, 3, 0, 60, 3, 95, 24, 60, 0, 76, 0, 175, 6, 159, 6, 179, 6, 12, 0, 183, 6, 98, 1, 187, 6, 75, 0, 191, 6, 48, 0, 195, 6, 255, 0, 163, 6, 99, 0, 199, 6, 66, 0, 203, 6, 97, 2, 167, 6, 228, 0, 171, 6, 210, 0, 227, 1, 255, 0, 251, 1, 81, 0, 16, 2, 210, 0, 19, 2, 134, 1, 25, 2, 48, 0, 37, 2, 246, 0, 55, 2, 237, 0, 58, 2, 78, 0, 85, 2, 174, 0, 103, 2, 225, 0, 142, 2, 225, 0, 154, 2, 182, 1, 160, 2, 237, 0, 166, 2, 213, 0, 178, 2, 99, 0, 190, 2, 225, 0, 199, 2, 35, 1, 205, 2, 46, 2, 208, 2, 165, 0, 232, 2, 12, 0, 0, 3, 50, 1, 6, 3, 57, 0, 12, 3, 162, 0, 18, 3, 225, 0, 24, 3, 228, 0, 27, 3, 210, 0, 30, 3, 87, 0, 18, 0, 68, 3, 95, 24, 90, 0, 152, 24, 95, 24, 123, 0, 68, 3, 95, 24, 175, 4, 80, 3, 95, 24, 57, 0, 152, 24, 95, 24, 161, 1, 80, 3, 95, 24, 36, 0, 15, 0, 1, 0, 173, 6, 85, 3, 165, 6, 207, 1, 161, 6, 210, 1, 169, 6, 141, 27, 206, 1, 218, 1, 212, 1, 209, 1, 215, 1, 127, 27, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 114, 38, 0, 0, 0, 26, 109, 204, 20, 116, 196, 239, 119, 194, 105, 119, 162, 185, 120, 164, 84, 121, 164, 196, 122, 15, 109, 127, 122, 91, 122, 48, 225, 173, 151, 232, 46, 245, 175, 59, 16, 225, 165, 57, 2, 42, 60, 66, 204, 17, 72, 17, 97, 110, 1, 226, 163, 253, 244, 163, 248, 11, 77, 36, 84, 14, 84, 202, 184, 85, 42, 86, 16, 206, 163, 251, 16, 211, 163, 250, 77, 202, 193, 80, 42, 83, 16, 210, 163, 246, 2, 193, 163, 245, 198, 163, 243, 200, 163, 244, 71, 212, 206, 71, 50, 72, 216, 30, 35, 73, 16, 196, 163, 240, 1, 194, 163, 249, 198, 163, 238, 109, 50, 110, 54, 115, 58, 121, 16, 226, 173, 151, 16, 233, 175, 57, 16, 229, 175, 58, 16, 237, 34, 2, 42, 169, 12, 67, 182, 6, 73, 16, 196, 169, 14, 104, 23, 104, 50, 105, 54, 107, 58, 108, 16, 237, 175, 56, 16, 248, 173, 180, 16, 225, 175, 55, 16, 244, 173, 146, 97, 50, 100, 54, 101, 58, 103, 16, 232, 173, 196, 16, 231, 175, 52, 16, 234, 175, 53, 16, 225, 175, 54, 0, 18, 109, 70, 116, 27, 116, 52, 117, 56, 119, 66, 225, 173, 232, 239, 173, 255, 16, 237, 175, 4, 1, 245, 175, 5, 246, 175, 6, 16, 225, 175, 7, 109, 60, 110, 64, 111, 80, 114, 90, 115, 1, 231, 173, 131, 235, 175, 3, 16, 239, 173, 251, 2, 227, 173, 252, 233, 173, 253, 245, 173, 254, 1, 226, 175, 0, 243, 175, 1, 16, 243, 175, 2, 103, 38, 103, 54, 104, 58, 105, 62, 106, 78, 108, 16, 243, 173, 250, 16, 233, 173, 243, 16, 231, 173, 244, 2, 226, 173, 245, 245, 173, 246, 246, 173, 247, 1, 225, 173, 248, 233, 173, 249, 97, 56, 98, 84, 99, 100, 101, 1, 236, 165, 32, 242, 173, 242, 4, 229, 173, 233, 234, 173, 234, 236, 173, 235, 238, 173, 236, 242, 173, 237, 2, 240, 173, 238, 241, 173, 239, 242, 173, 240, 16, 233, 173, 241, 12, 110, 61, 114, 33, 114, 46, 115, 50, 119, 78, 232, 175, 11, 16, 226, 175, 17, 4, 225, 173, 188, 233, 175, 18, 236, 165, 47, 237, 175, 19, 242, 175, 20, 16, 229, 175, 21, 110, 52, 111, 62, 112, 1, 229, 167, 141, 242, 173, 185, 1, 225, 173, 178, 242, 175, 14, 1, 231, 175, 15, 238, 175, 16, 101, 29, 101, 58, 108, 62, 109, 2, 230, 175, 13, 238, 173, 168, 242, 173, 171, 16, 243, 175, 10, 2, 225, 175, 12, 227, 173, 164, 228, 173, 165, 97, 52, 98, 56, 99, 1, 239, 173, 114, 242, 173, 113, 16, 246, 175, 8, 16, 233, 175, 9, 13, 109, 73, 115, 48, 115, 48, 117, 52, 233, 167, 51, 239, 175, 40, 16, 243, 175, 45, 4, 225, 175, 46, 229, 50, 234, 175, 49, 244, 175, 50, 247, 175, 51, 1, 42, 46, 72, 18, 97, 110, 243, 175, 48, 1, 42, 175, 47, 67, 16, 206, 175, 48, 109, 58, 111, 62, 114, 2, 226, 175, 42, 229, 175, 43, 236, 175, 44, 16, 236, 175, 39, 16, 238, 175, 41, 101, 32, 101, 62, 103, 66, 107, 76, 108, 2, 229, 175, 36, 231, 175, 37, 236, 175, 38, 16, 242, 175, 32, 1, 242, 175, 33, 247, 175, 34, 16, 239, 175, 35, 97, 46, 98, 96, 100, 16, 228, 167, 51, 7, 244, 12, 244, 175, 25, 246, 175, 26, 249, 175, 27, 250, 175, 28, 237, 167, 182, 239, 175, 22, 240, 175, 23, 243, 175, 24, 2, 225, 175, 29, 226, 175, 30, 249, 175, 31, 116, 162, 82, 117, 164, 208, 118, 10, 111, 31, 229, 9, 229, 173, 219, 233, 173, 95, 239, 173, 227, 111, 52, 114, 56, 117, 1, 238, 173, 230, 244, 173, 231, 16, 244, 173, 228, 16, 239, 173, 229, 97, 60, 101, 76, 105, 86, 108, 96, 109, 1, 230, 173, 225, 247, 173, 226, 2, 231, 173, 217, 233, 173, 201, 238, 173, 218, 1, 227, 173, 220, 240, 173, 221, 1, 227, 173, 222, 246, 173, 223, 16, 243, 173, 224, 0, 38, 117, 118, 233, 31, 242, 14, 242, 171, 197, 243, 171, 201, 244, 171, 207, 247, 103, 249, 171, 223, 233, 171, 170, 235, 171, 177, 236, 171, 181, 238, 171, 187, 239, 171, 189, 122, 36, 122, 54, 225, 171, 137, 229, 171, 155, 231, 46, 232, 171, 166, 16, 237, 171, 226, 1, 42, 46, 65, 18, 114, 97, 226, 171, 162, 1, 42, 171, 161, 80, 16, 203, 171, 162, 117, 62, 118, 84, 119, 100, 120, 204, 225, 121, 1, 225, 171, 224, 246, 171, 225, 3, 232, 171, 214, 236, 171, 215, 237, 171, 216, 241, 171, 217, 2, 228, 167, 185, 236, 171, 218, 245, 171, 219, 1, 232, 171, 220, 241, 171, 221, 107, 125, 112, 73, 112, 94, 113, 110, 114, 114, 115, 130, 116, 6, 241, 12, 241, 171, 185, 242, 171, 211, 243, 171, 212, 244, 171, 213, 228, 171, 208, 229, 171, 209, 234, 171, 210, 2, 233, 171, 193, 237, 171, 194, 250, 171, 195, 16, 239, 171, 196, 2, 245, 171, 198, 246, 171, 199, 247, 171, 200, 4, 228, 171, 202, 230, 171, 203, 231, 171, 204, 234, 171, 205, 247, 171, 206, 107, 66, 108, 82, 109, 98, 110, 108, 111, 2, 230, 171, 190, 231, 171, 191, 241, 171, 192, 2, 236, 171, 178, 242, 171, 179, 244, 171, 180, 2, 230, 171, 182, 248, 171, 183, 249, 171, 184, 1, 232, 171, 185, 249, 171, 186, 16, 232, 171, 188, 101, 70, 101, 94, 102, 116, 103, 120, 104, 136, 105, 6, 235, 12, 235, 171, 173, 237, 171, 174, 239, 171, 175, 246, 171, 176, 226, 163, 177, 230, 171, 171, 231, 171, 172, 3, 228, 171, 156, 237, 171, 157, 239, 171, 158, 244, 171, 159, 16, 233, 171, 160, 2, 227, 171, 163, 239, 171, 164, 245, 171, 165, 2, 236, 171, 167, 241, 171, 168, 242, 171, 169, 97, 68, 98, 90, 99, 134, 100, 3, 228, 171, 151, 231, 171, 152, 232, 171, 153, 245, 171, 154, 3, 234, 171, 138, 236, 171, 139, 238, 171, 140, 241, 171, 141, 6, 231, 12, 231, 171, 145, 239, 171, 146, 247, 171, 147, 250, 171, 148, 227, 171, 142, 228, 171, 143, 230, 171, 144, 1, 233, 171, 149, 249, 171, 150, 15, 115, 95, 231, 44, 231, 66, 235, 171, 233, 242, 173, 71, 250, 1, 42, 134, 65, 18, 114, 97, 226, 173, 215, 1, 42, 46, 67, 18, 121, 114, 236, 171, 230, 2, 42, 171, 229, 75, 42, 77, 16, 206, 171, 231, 16, 218, 171, 230, 115, 96, 116, 100, 118, 104, 122, 16, 238, 1, 42, 46, 65, 18, 114, 97, 226, 173, 215, 2, 42, 171, 244, 65, 42, 67, 16, 206, 173, 216, 16, 198, 173, 215, 16, 225, 173, 211, 16, 242, 173, 212, 1, 232, 173, 213, 236, 173, 214, 109, 50, 109, 62, 110, 66, 111, 112, 114, 2, 233, 173, 208, 244, 173, 209, 247, 173, 210, 16, 226, 171, 235, 1, 242, 38, 248, 173, 206, 1, 42, 46, 68, 18, 101, 118, 225, 173, 205, 1, 42, 173, 204, 78, 16, 208, 173, 205, 16, 235, 173, 207, 98, 50, 100, 54, 103, 58, 108, 16, 233, 171, 234, 16, 245, 171, 227, 16, 237, 171, 228, 16, 225, 171, 232, 112, 195, 149, 112, 164, 249, 113, 166, 215, 114, 166, 228, 115, 0, 40, 120, 132, 236, 57, 242, 41, 245, 9, 245, 171, 120, 246, 171, 126, 247, 171, 127, 242, 44, 243, 171, 113, 244, 171, 117, 34, 3, 42, 171, 43, 77, 164, 191, 82, 42, 84, 16, 210, 171, 47, 1, 207, 171, 46, 213, 171, 44, 236, 171, 83, 237, 171, 88, 238, 171, 95, 239, 171, 101, 241, 109, 228, 41, 228, 56, 229, 171, 59, 231, 171, 64, 233, 171, 73, 235, 171, 79, 3, 42, 171, 53, 68, 50, 75, 58, 83, 18, 105, 110, 228, 171, 56, 18, 101, 118, 225, 171, 54, 18, 104, 111, 234, 171, 55, 120, 52, 121, 62, 122, 72, 225, 171, 31, 227, 171, 42, 1, 238, 171, 133, 247, 171, 134, 1, 236, 171, 135, 242, 165, 12, 16, 236, 171, 136, 108, 173, 114, 79, 114, 88, 115, 116, 116, 132, 117, 142, 119, 5, 232, 9, 232, 171, 127, 240, 171, 131, 246, 171, 132, 226, 171, 128, 227, 171, 129, 231, 171, 130, 4, 226, 171, 110, 227, 171, 42, 238, 171, 111, 242, 171, 112, 248, 165, 241, 2, 228, 171, 114, 231, 171, 115, 249, 171, 116, 1, 235, 171, 118, 241, 171, 119, 4, 225, 171, 121, 229, 171, 122, 235, 171, 123, 242, 171, 124, 243, 171, 125, 108, 74, 109, 102, 110, 140, 111, 162, 67, 112, 3, 228, 171, 107, 236, 171, 108, 243, 171, 109, 249, 167, 116, 4, 228, 171, 84, 233, 171, 85, 236, 171, 86, 239, 171, 79, 249, 171, 87, 5, 240, 9, 240, 171, 92, 241, 171, 93, 243, 171, 94, 225, 171, 89, 234, 171, 90, 238, 171, 91, 4, 227, 171, 96, 235, 171, 97, 240, 171, 98, 248, 171, 99, 249, 171, 100, 4, 231, 171, 102, 235, 171, 103, 241, 171, 104, 245, 171, 105, 249, 171, 106, 103, 62, 103, 66, 104, 88, 105, 110, 106, 132, 107, 2, 227, 171, 80, 242, 171, 81, 243, 171, 82, 3, 225, 171, 65, 243, 171, 66, 247, 171, 67, 250, 171, 68, 3, 233, 171, 69, 235, 171, 70, 238, 171, 71, 245, 171, 72, 3, 228, 171, 74, 231, 171, 75, 236, 171, 76, 237, 171, 77, 16, 242, 171, 78, 97, 74, 98, 118, 99, 134, 100, 162, 107, 101, 3, 230, 171, 60, 232, 171, 61, 233, 171, 62, 243, 171, 63, 6, 243, 12, 243, 171, 35, 244, 171, 36, 246, 171, 37, 250, 171, 38, 230, 171, 32, 232, 171, 33, 241, 171, 34, 2, 225, 171, 39, 229, 171, 40, 240, 171, 41, 6, 238, 12, 238, 171, 50, 239, 171, 51, 242, 165, 255, 243, 171, 52, 227, 44, 235, 171, 48, 236, 171, 49, 34, 5, 82, 17, 82, 46, 83, 50, 84, 16, 210, 171, 47, 16, 207, 171, 46, 16, 213, 171, 44, 42, 171, 43, 49, 42, 77, 16, 197, 171, 45, 17, 55, 178, 171, 44, 1, 227, 171, 57, 232, 171, 58, 0, 22, 110, 99, 117, 43, 236, 9, 236, 169, 234, 243, 169, 219, 244, 169, 247, 117, 76, 119, 80, 225, 1, 42, 46, 65, 18, 114, 97, 226, 169, 211, 1, 42, 169, 210, 80, 16, 203, 169, 211, 16, 245, 169, 249, 16, 225, 169, 250, 114, 23, 114, 46, 115, 62, 116, 16, 240, 169, 248, 2, 225, 169, 243, 228, 169, 244, 231, 169, 245, 16, 243, 169, 246, 110, 52, 111, 74, 112, 1, 225, 169, 241, 239, 169, 242, 3, 226, 167, 202, 231, 169, 237, 238, 169, 238, 244, 169, 239, 16, 238, 169, 240, 102, 48, 107, 23, 107, 46, 108, 56, 109, 16, 243, 169, 236, 1, 225, 169, 232, 239, 169, 233, 1, 225, 169, 235, 244, 169, 45, 102, 52, 104, 56, 105, 1, 236, 169, 230, 240, 169, 231, 16, 236, 169, 227, 1, 236, 169, 228, 238, 169, 229, 97, 78, 98, 126, 99, 136, 100, 146, 101, 4, 228, 169, 224, 239, 169, 225, 242, 165, 112, 243, 165, 112, 248, 169, 226, 4, 231, 169, 212, 236, 50, 237, 169, 215, 240, 169, 216, 245, 169, 217, 1, 42, 169, 213, 80, 18, 104, 108, 240, 169, 214, 1, 233, 169, 218, 245, 169, 219, 1, 228, 169, 220, 237, 169, 221, 1, 227, 169, 222, 244, 169, 223, 1, 117, 38, 245, 169, 251, 2, 227, 169, 252, 231, 169, 253, 250, 169, 251, 0, 20, 111, 69, 121, 23, 239, 9, 239, 169, 74, 245, 171, 24, 247, 171, 27, 121, 44, 237, 171, 11, 238, 171, 16, 16, 245, 171, 30, 111, 60, 114, 76, 116, 80, 117, 84, 119, 1, 235, 171, 28, 239, 171, 29, 2, 226, 171, 19, 230, 171, 20, 239, 171, 21, 16, 239, 171, 22, 16, 237, 171, 23, 2, 229, 171, 25, 231, 171, 26, 237, 169, 74, 105, 52, 105, 60, 106, 86, 107, 90, 109, 94, 110, 1, 225, 171, 17, 231, 171, 18, 1, 225, 171, 6, 230, 34, 1, 42, 171, 7, 78, 16, 204, 171, 8, 16, 243, 171, 9, 16, 244, 171, 10, 3, 230, 171, 12, 239, 171, 13, 244, 171, 14, 245, 171, 15, 97, 54, 99, 70, 101, 74, 103, 90, 104, 16, 231, 171, 5, 2, 233, 169, 254, 234, 169, 95, 239, 169, 255, 16, 230, 171, 0, 2, 234, 171, 1, 236, 171, 2, 243, 171, 3, 16, 238, 171, 4, 109, 162, 87, 110, 166, 68, 111, 11, 116, 29, 237, 9, 237, 165, 161, 242, 169, 202, 243, 169, 205, 116, 42, 122, 52, 227, 169, 194, 1, 225, 169, 207, 235, 169, 208, 16, 237, 169, 209, 112, 23, 112, 46, 114, 50, 115, 16, 225, 169, 206, 16, 237, 169, 201, 2, 239, 169, 203, 245, 169, 204, 249, 169, 202, 103, 58, 107, 62, 110, 2, 231, 169, 198, 238, 169, 199, 243, 169, 200, 16, 227, 169, 195, 1, 242, 169, 196, 246, 169, 197, 0, 34, 116, 159, 233, 38, 239, 15, 239, 169, 74, 242, 169, 85, 243, 164, 146, 244, 169, 89, 249, 163, 206, 233, 169, 9, 235, 169, 3, 236, 169, 63, 238, 1, 42, 186, 189, 77, 18, 111, 110, 231, 165, 64, 120, 61, 120, 52, 121, 62, 122, 100, 231, 169, 45, 232, 169, 51, 1, 227, 169, 104, 237, 169, 105, 5, 247, 9, 247, 169, 109, 248, 169, 110, 250, 169, 111, 235, 169, 106, 237, 169, 107, 246, 169, 108, 5, 240, 9, 240, 169, 115, 247, 169, 116, 250, 169, 117, 235, 169, 112, 237, 169, 113, 238, 169, 114, 116, 68, 117, 90, 118, 112, 119, 3, 235, 169, 101, 242, 165, 55, 246, 169, 102, 247, 169, 103, 3, 227, 169, 90, 230, 169, 91, 233, 169, 92, 242, 169, 93, 3, 225, 169, 94, 240, 169, 95, 242, 169, 96, 243, 169, 97, 2, 225, 169, 98, 238, 169, 99, 249, 169, 100, 105, 133, 110, 68, 110, 66, 111, 94, 112, 122, 113, 144, 114, 2, 228, 169, 86, 234, 169, 87, 239, 169, 88, 4, 225, 169, 70, 230, 169, 71, 233, 169, 72, 235, 182, 14, 247, 169, 73, 4, 225, 169, 75, 229, 169, 76, 232, 169, 77, 243, 169, 78, 248, 169, 79, 3, 240, 169, 80, 243, 169, 81, 244, 169, 82, 248, 169, 83, 16, 236, 169, 84, 105, 62, 107, 104, 108, 126, 109, 2, 239, 169, 67, 245, 169, 68, 248, 169, 69, 3, 230, 169, 54, 238, 169, 55, 243, 38, 247, 169, 58, 1, 42, 169, 56, 77, 18, 101, 100, 230, 169, 57, 3, 233, 169, 59, 236, 169, 60, 240, 169, 61, 247, 169, 62, 2, 229, 169, 64, 240, 169, 65, 243, 169, 66, 101, 73, 101, 62, 102, 106, 103, 134, 104, 2, 233, 169, 52, 236, 169, 53, 242, 163, 255, 6, 238, 12, 238, 169, 36, 242, 169, 37, 244, 169, 38, 245, 169, 39, 228, 169, 33, 229, 169, 34, 235, 169, 35, 4, 225, 169, 40, 229, 169, 41, 238, 169, 42, 239, 169, 43, 241, 169, 44, 4, 232, 169, 46, 236, 169, 47, 239, 169, 48, 240, 169, 49, 249, 169, 50, 97, 94, 98, 162, 86, 99, 162, 99, 100, 6, 232, 12, 232, 169, 29, 234, 169, 30, 242, 169, 31, 248, 169, 32, 225, 169, 26, 229, 169, 27, 230, 169, 28, 11, 238, 35, 247, 24, 247, 169, 11, 249, 38, 250, 169, 15, 34, 2, 42, 169, 12, 67, 42, 73, 16, 196, 169, 14, 16, 195, 169, 13, 238, 180, 31, 239, 169, 9, 243, 169, 10, 231, 9, 231, 169, 6, 233, 169, 7, 235, 169, 8, 227, 169, 3, 228, 169, 4, 230, 169, 5, 4, 232, 169, 16, 239, 169, 17, 241, 169, 18, 245, 169, 19, 247, 169, 20, 4, 233, 169, 21, 240, 169, 22, 241, 169, 23, 242, 169, 24, 245, 169, 25, 0, 33, 116, 95, 228, 29, 238, 15, 238, 169, 160, 239, 169, 166, 242, 169, 173, 246, 169, 186, 249, 169, 190, 228, 169, 133, 229, 169, 136, 231, 169, 141, 236, 165, 76, 121, 21, 121, 48, 122, 58, 225, 169, 118, 226, 169, 126, 1, 237, 169, 191, 238, 169, 192, 16, 233, 169, 193, 116, 56, 117, 66, 119, 94, 120, 1, 241, 169, 188, 242, 169, 189, 1, 237, 169, 179, 242, 169, 180, 4, 233, 169, 181, 240, 169, 182, 243, 169, 183, 246, 169, 184, 248, 169, 185, 16, 226, 169, 187, 106, 93, 111, 50, 111, 72, 112, 100, 113, 104, 114, 108, 115, 3, 235, 169, 175, 238, 169, 176, 239, 169, 177, 243, 169, 178, 4, 228, 169, 167, 229, 169, 168, 238, 169, 169, 240, 169, 170, 245, 169, 171, 16, 233, 169, 136, 16, 239, 169, 172, 16, 226, 169, 174, 106, 74, 107, 78, 109, 88, 110, 4, 230, 169, 161, 232, 169, 162, 235, 169, 163, 237, 169, 164, 240, 169, 165, 16, 239, 169, 155, 1, 231, 169, 156, 239, 169, 157, 1, 231, 169, 158, 250, 169, 159, 102, 55, 102, 90, 103, 94, 104, 110, 105, 6, 238, 12, 238, 169, 151, 245, 169, 152, 249, 169, 153, 250, 169, 154, 230, 169, 148, 233, 169, 149, 234, 169, 150, 16, 242, 169, 140, 2, 225, 169, 142, 226, 169, 143, 236, 169, 144, 2, 226, 169, 145, 229, 169, 146, 247, 169, 147, 97, 62, 99, 106, 100, 144, 101, 2, 226, 169, 137, 247, 169, 138, 248, 169, 139, 6, 238, 12, 238, 169, 122, 240, 169, 123, 241, 169, 124, 243, 169, 125, 227, 169, 119, 230, 169, 120, 235, 169, 121, 5, 232, 9, 232, 169, 130, 239, 169, 131, 245, 169, 132, 225, 169, 127, 229, 169, 128, 230, 169, 129, 1, 227, 169, 134, 243, 169, 135, 102, 199, 71, 105, 196, 141, 105, 164, 22, 106, 164, 196, 107, 166, 14, 108, 0, 27, 116, 74, 231, 23, 239, 12, 239, 167, 241, 244, 167, 249, 245, 167, 251, 246, 167, 255, 231, 167, 219, 233, 167, 221, 238, 167, 238, 119, 21, 119, 48, 122, 52, 225, 167, 198, 226, 167, 205, 16, 236, 169, 0, 1, 232, 169, 1, 250, 169, 2, 116, 46, 117, 50, 118, 16, 243, 167, 255, 16, 231, 167, 250, 3, 225, 167, 252, 239, 167, 253, 249, 163, 213, 250, 167, 254, 106, 74, 109, 49, 109, 50, 110, 66, 111, 76, 114, 16, 227, 167, 248, 2, 238, 167, 235, 239, 167, 236, 240, 167, 237, 1, 243, 167, 239, 245, 167, 240, 5, 242, 9, 242, 167, 245, 243, 167, 246, 250, 167, 247, 234, 167, 242, 235, 167, 243, 236, 167, 244, 106, 52, 107, 56, 108, 1, 229, 167, 233, 238, 167, 234, 16, 240, 167, 230, 1, 233, 167, 231, 244, 167, 232, 100, 72, 100, 108, 101, 112, 103, 156, 105, 6, 231, 12, 231, 167, 226, 232, 167, 227, 234, 167, 228, 243, 167, 229, 225, 167, 222, 228, 167, 223, 230, 1, 42, 167, 224, 76, 18, 105, 109, 226, 167, 225, 16, 226, 167, 211, 6, 240, 12, 240, 167, 215, 241, 167, 216, 245, 167, 217, 250, 167, 218, 228, 167, 212, 229, 167, 213, 237, 167, 214, 16, 231, 167, 220, 97, 52, 98, 90, 99, 1, 237, 167, 209, 240, 167, 210, 5, 232, 9, 232, 167, 202, 234, 167, 203, 243, 167, 204, 226, 167, 199, 228, 167, 200, 231, 167, 201, 2, 229, 167, 206, 245, 167, 207, 247, 167, 208, 0, 25, 119, 57, 235, 23, 243, 12, 243, 167, 13, 244, 167, 37, 245, 167, 25, 247, 167, 38, 235, 165, 104, 238, 167, 32, 239, 167, 34, 228, 9, 228, 167, 15, 231, 167, 20, 233, 167, 23, 119, 42, 122, 52, 225, 167, 6, 1, 237, 167, 39, 243, 167, 40, 1, 232, 167, 41, 233, 167, 42, 106, 54, 109, 23, 109, 50, 110, 54, 111, 58, 114, 16, 233, 167, 36, 16, 239, 167, 31, 16, 232, 167, 33, 16, 245, 167, 35, 106, 46, 107, 50, 108, 16, 239, 167, 30, 16, 234, 167, 24, 4, 229, 167, 25, 235, 167, 26, 244, 167, 27, 247, 167, 28, 248, 167, 29, 100, 26, 100, 52, 102, 68, 103, 1, 226, 167, 21, 229, 167, 22, 2, 228, 167, 16, 233, 167, 17, 245, 167, 18, 16, 229, 167, 19, 97, 58, 98, 68, 99, 2, 225, 167, 12, 229, 167, 13, 232, 167, 14, 1, 238, 167, 7, 242, 167, 8, 2, 225, 167, 9, 226, 167, 10, 249, 167, 11, 11, 114, 26, 233, 9, 233, 167, 51, 246, 167, 57, 247, 167, 58, 114, 42, 117, 46, 225, 167, 43, 16, 225, 167, 55, 16, 244, 167, 56, 103, 23, 103, 52, 105, 62, 109, 1, 227, 167, 53, 236, 167, 54, 1, 235, 167, 49, 239, 167, 50, 16, 226, 167, 52, 97, 46, 98, 56, 101, 16, 238, 167, 48, 1, 226, 167, 44, 237, 167, 45, 1, 239, 167, 46, 245, 167, 47, 0, 40, 117, 192, 231, 235, 116, 242, 63, 246, 40, 246, 167, 145, 247, 167, 179, 249, 2, 42, 50, 65, 76, 76, 18, 97, 116, 238, 167, 192, 2, 42, 167, 190, 67, 42, 84, 16, 210, 167, 192, 16, 206, 167, 191, 18, 114, 97, 226, 167, 191, 242, 167, 132, 243, 167, 157, 245, 2, 42, 164, 94, 65, 164, 100, 89, 18, 101, 122, 233, 167, 127, 235, 56, 236, 167, 115, 237, 167, 120, 238, 167, 131, 239, 167, 136, 1, 42, 46, 65, 18, 114, 97, 226, 167, 110, 4, 42, 167, 108, 65, 50, 67, 54, 73, 58, 77, 16, 206, 167, 112, 16, 198, 167, 109, 16, 206, 167, 110, 16, 210, 167, 111, 122, 24, 122, 56, 225, 165, 176, 231, 167, 87, 233, 167, 99, 234, 167, 103, 2, 234, 167, 195, 242, 167, 196, 244, 167, 197, 117, 60, 118, 110, 119, 126, 120, 142, 121, 1, 229, 167, 193, 248, 167, 194, 7, 237, 12, 237, 167, 172, 238, 167, 173, 240, 167, 174, 243, 167, 175, 226, 167, 168, 228, 167, 169, 229, 167, 170, 234, 167, 171, 2, 231, 167, 176, 242, 167, 177, 248, 167, 178, 2, 234, 167, 180, 239, 167, 181, 241, 167, 182, 6, 237, 12, 237, 167, 186, 240, 167, 187, 247, 167, 188, 250, 167, 189, 225, 167, 183, 227, 167, 184, 229, 167, 185, 107, 192, 213, 112, 99, 112, 74, 113, 112, 114, 134, 115, 162, 69, 116, 3, 226, 167, 164, 237, 167, 165, 239, 167, 166, 242, 167, 167, 5, 242, 9, 242, 167, 144, 246, 167, 145, 248, 167, 146, 229, 167, 141, 230, 167, 142, 239, 167, 143, 3, 226, 167, 147, 230, 167, 148, 243, 167, 149, 249, 167, 150, 5, 236, 9, 236, 167, 154, 243, 167, 155, 245, 167, 156, 227, 167, 151, 233, 167, 152, 234, 167, 153, 5, 232, 9, 232, 167, 161, 234, 167, 162, 242, 167, 163, 226, 167, 158, 228, 167, 159, 230, 167, 160, 107, 80, 108, 90, 109, 112, 110, 162, 87, 111, 4, 233, 167, 137, 235, 167, 134, 236, 167, 138, 243, 167, 139, 250, 167, 140, 1, 227, 167, 113, 234, 167, 114, 3, 238, 167, 116, 241, 167, 117, 244, 167, 118, 248, 167, 119, 6, 242, 38, 242, 50, 243, 167, 128, 245, 167, 129, 247, 167, 130, 2, 42, 50, 65, 64, 89, 18, 101, 122, 233, 167, 127, 1, 42, 167, 124, 76, 16, 194, 167, 125, 18, 114, 97, 226, 167, 126, 226, 167, 121, 232, 167, 122, 239, 167, 123, 4, 227, 167, 132, 230, 167, 133, 231, 167, 87, 238, 167, 134, 240, 167, 135, 102, 82, 102, 72, 103, 88, 104, 104, 105, 160, 106, 3, 228, 167, 104, 231, 167, 105, 243, 167, 106, 249, 167, 107, 2, 239, 167, 84, 242, 167, 85, 249, 167, 86, 2, 229, 167, 88, 230, 167, 89, 240, 167, 90, 8, 241, 15, 241, 167, 94, 243, 167, 95, 244, 167, 96, 247, 167, 97, 250, 167, 98, 225, 167, 91, 226, 167, 92, 235, 172, 128, 238, 167, 93, 2, 234, 167, 100, 245, 167, 101, 247, 167, 102, 97, 68, 98, 118, 99, 156, 100, 162, 72, 101, 2, 225, 167, 81, 238, 167, 82, 250, 167, 83, 7, 233, 12, 233, 167, 63, 234, 167, 64, 237, 167, 65, 239, 167, 66, 225, 167, 59, 226, 167, 60, 227, 167, 61, 228, 167, 62, 5, 241, 9, 241, 167, 70, 248, 167, 71, 249, 167, 72, 228, 167, 67, 237, 167, 68, 240, 167, 69, 3, 231, 167, 73, 235, 167, 74, 236, 167, 75, 244, 167, 76, 3, 229, 167, 77, 232, 167, 78, 236, 167, 79, 244, 167, 80, 102, 162, 181, 103, 164, 106, 104, 0, 19, 117, 57, 242, 15, 242, 165, 255, 244, 167, 2, 245, 167, 3, 249, 163, 66, 250, 167, 5, 117, 54, 225, 58, 229, 165, 233, 233, 165, 235, 239, 165, 251, 16, 233, 167, 4, 34, 3, 42, 165, 223, 55, 46, 67, 52, 83, 16, 196, 165, 224, 17, 51, 182, 165, 224, 16, 205, 165, 225, 108, 56, 108, 60, 109, 70, 110, 80, 111, 108, 115, 1, 226, 167, 0, 238, 167, 1, 1, 225, 165, 242, 245, 165, 243, 1, 228, 165, 244, 244, 165, 245, 4, 228, 165, 246, 229, 165, 247, 234, 165, 248, 238, 165, 249, 239, 165, 250, 2, 227, 165, 252, 234, 165, 253, 244, 165, 254, 97, 88, 98, 116, 100, 120, 104, 124, 105, 5, 232, 9, 232, 165, 239, 236, 165, 240, 237, 165, 241, 225, 165, 236, 230, 165, 237, 231, 165, 238, 4, 231, 165, 226, 235, 165, 227, 237, 165, 228, 247, 165, 229, 250, 165, 230, 16, 226, 165, 231, 16, 249, 165, 232, 16, 249, 165, 234, 0, 17, 117, 83, 233, 15, 233, 165, 122, 234, 165, 126, 239, 165, 129, 242, 165, 135, 249, 165, 150, 117, 66, 118, 148, 225, 165, 112, 230, 1, 42, 165, 118, 65, 18, 100, 108, 237, 165, 119, 9, 232, 15, 232, 165, 144, 241, 165, 145, 242, 165, 146, 246, 165, 147, 249, 165, 148, 226, 165, 140, 227, 50, 228, 165, 141, 229, 165, 142, 230, 165, 143, 1, 42, 165, 118, 65, 18, 100, 108, 237, 165, 119, 16, 242, 165, 149, 109, 47, 109, 78, 111, 82, 112, 98, 113, 102, 114, 4, 227, 165, 136, 229, 165, 135, 240, 165, 137, 242, 165, 138, 243, 165, 139, 16, 240, 165, 128, 2, 228, 165, 130, 238, 165, 131, 242, 165, 132, 16, 229, 165, 133, 16, 243, 165, 134, 97, 50, 102, 86, 105, 96, 108, 16, 242, 165, 127, 5, 233, 8, 233, 165, 116, 238, 165, 117, 244, 103, 225, 165, 113, 226, 165, 114, 231, 165, 115, 1, 233, 165, 120, 237, 165, 121, 2, 225, 165, 123, 236, 165, 124, 244, 165, 125, 0, 26, 110, 159, 121, 29, 236, 12, 236, 165, 188, 238, 165, 192, 245, 165, 208, 246, 165, 216, 121, 44, 225, 165, 151, 228, 165, 169, 1, 225, 165, 162, 233, 165, 222, 115, 58, 115, 62, 117, 66, 118, 116, 119, 2, 227, 165, 219, 233, 165, 220, 244, 165, 221, 16, 247, 165, 207, 7, 242, 12, 242, 165, 212, 247, 165, 213, 248, 165, 214, 250, 165, 215, 226, 165, 209, 227, 165, 210, 228, 165, 211, 231, 165, 192, 2, 230, 165, 217, 242, 165, 179, 243, 165, 218, 110, 90, 111, 106, 114, 4, 226, 165, 165, 227, 50, 229, 165, 91, 244, 165, 205, 247, 165, 206, 1, 42, 165, 203, 76, 18, 105, 110, 226, 165, 204, 2, 228, 165, 193, 231, 165, 194, 239, 165, 195, 7, 238, 12, 238, 165, 195, 242, 165, 200, 243, 165, 201, 244, 165, 202, 228, 165, 196, 230, 165, 197, 233, 165, 198, 237, 165, 199, 103, 57, 106, 35, 106, 56, 107, 72, 108, 82, 109, 1, 237, 165, 190, 246, 165, 191, 2, 235, 165, 183, 238, 165, 184, 245, 165, 185, 1, 238, 165, 186, 240, 165, 187, 16, 235, 165, 189, 103, 52, 104, 56, 105, 1, 236, 165, 181, 237, 165, 182, 16, 238, 165, 179, 16, 243, 165, 180, 100, 40, 100, 46, 101, 62, 102, 16, 235, 165, 178, 2, 229, 165, 170, 238, 165, 171, 242, 165, 172, 5, 239, 9, 239, 165, 176, 242, 165, 45, 250, 165, 177, 226, 165, 173, 234, 165, 174, 236, 165, 175, 97, 46, 98, 108, 99, 16, 242, 165, 168, 9, 237, 15, 237, 165, 157, 238, 165, 158, 247, 165, 159, 249, 165, 160, 250, 165, 161, 225, 165, 152, 230, 165, 153, 231, 165, 154, 232, 165, 155, 234, 165, 156, 5, 239, 9, 239, 165, 165, 249, 165, 166, 250, 165, 167, 225, 165, 162, 230, 165, 163, 237, 165, 164, 99, 194, 239, 99, 162, 157, 100, 164, 249, 101, 0, 17, 119, 45, 238, 25, 238, 56, 239, 165, 100, 243, 165, 102, 244, 165, 89, 245, 163, 102, 1, 42, 165, 96, 83, 18, 104, 97, 247, 165, 97, 119, 48, 120, 52, 229, 165, 84, 236, 165, 91, 16, 239, 165, 110, 16, 244, 165, 111, 109, 70, 109, 72, 110, 122, 114, 132, 115, 136, 116, 3, 242, 165, 106, 244, 165, 107, 245, 165, 108, 248, 165, 109, 2, 225, 165, 92, 233, 165, 93, 235, 1, 42, 46, 78, 18, 107, 111, 239, 165, 95, 1, 42, 165, 94, 71, 16, 206, 165, 95, 1, 238, 165, 98, 241, 165, 99, 16, 233, 165, 101, 2, 231, 165, 103, 235, 165, 104, 245, 165, 105, 98, 62, 102, 66, 103, 70, 107, 2, 225, 165, 88, 235, 165, 89, 249, 165, 90, 16, 245, 165, 83, 16, 233, 165, 85, 1, 236, 165, 86, 249, 165, 87, 0, 27, 114, 92, 232, 33, 243, 22, 243, 165, 25, 245, 44, 246, 165, 31, 249, 165, 32, 1, 42, 165, 29, 71, 18, 108, 97, 231, 165, 30, 232, 163, 234, 239, 165, 15, 242, 165, 18, 119, 18, 119, 48, 122, 52, 225, 163, 225, 229, 163, 230, 16, 228, 165, 18, 16, 229, 165, 25, 114, 46, 115, 84, 116, 16, 228, 165, 28, 5, 236, 9, 236, 165, 22, 237, 165, 23, 243, 165, 24, 232, 165, 19, 234, 165, 20, 235, 165, 21, 1, 226, 165, 26, 247, 165, 27, 105, 82, 108, 48, 108, 50, 109, 60, 111, 108, 112, 16, 243, 165, 17, 1, 225, 165, 11, 228, 165, 12, 2, 229, 165, 13, 231, 165, 14, 238, 2, 42, 162, 99, 66, 162, 180, 72, 17, 97, 110, 1, 226, 163, 253, 244, 163, 248, 16, 240, 165, 16, 105, 64, 106, 68, 107, 3, 226, 165, 7, 236, 165, 8, 239, 165, 9, 249, 165, 10, 16, 227, 165, 3, 2, 225, 165, 4, 237, 165, 5, 246, 165, 6, 101, 149, 101, 162, 134, 102, 162, 135, 103, 162, 136, 104, 5, 239, 9, 239, 165, 0, 240, 165, 1, 242, 165, 2, 233, 44, 235, 163, 254, 237, 163, 255, 2, 42, 60, 66, 162, 95, 72, 17, 97, 110, 1, 226, 163, 253, 244, 163, 248, 11, 77, 51, 84, 23, 84, 46, 85, 56, 86, 16, 196, 163, 251, 1, 200, 163, 247, 215, 163, 248, 1, 203, 163, 249, 211, 163, 250, 77, 46, 80, 56, 83, 16, 210, 163, 246, 1, 207, 163, 241, 217, 163, 242, 2, 198, 163, 243, 200, 163, 244, 218, 163, 245, 71, 14, 71, 48, 72, 180, 100, 73, 16, 196, 163, 240, 16, 198, 163, 238, 42, 163, 235, 65, 42, 66, 16, 206, 163, 237, 16, 213, 163, 236, 18, 111, 112, 239, 163, 252, 16, 226, 163, 231, 16, 225, 163, 232, 16, 231, 163, 233, 97, 52, 98, 56, 99, 1, 232, 163, 228, 240, 163, 229, 16, 238, 163, 226, 16, 234, 163, 227, 0, 22, 114, 111, 121, 28, 229, 8, 229, 165, 45, 246, 165, 77, 250, 69, 121, 42, 122, 52, 225, 165, 33, 1, 239, 165, 80, 245, 165, 81, 16, 231, 165, 82, 117, 29, 117, 46, 118, 74, 119, 16, 247, 165, 79, 4, 225, 165, 72, 227, 165, 73, 228, 165, 74, 231, 165, 75, 244, 165, 76, 16, 225, 165, 78, 114, 64, 115, 116, 116, 3, 237, 165, 68, 240, 165, 69, 243, 165, 70, 249, 165, 71, 2, 232, 44, 233, 165, 65, 243, 165, 66, 1, 42, 46, 77, 18, 111, 110, 231, 165, 64, 1, 42, 165, 63, 67, 16, 206, 165, 64, 16, 226, 165, 67, 103, 68, 106, 26, 106, 64, 110, 68, 111, 3, 226, 165, 60, 233, 165, 52, 240, 165, 61, 247, 165, 62, 16, 229, 165, 58, 16, 234, 165, 59, 103, 52, 104, 96, 105, 1, 225, 165, 56, 241, 165, 57, 6, 236, 12, 236, 165, 51, 239, 165, 52, 242, 165, 53, 250, 165, 54, 225, 165, 48, 232, 165, 49, 233, 165, 50, 16, 228, 165, 55, 97, 60, 98, 104, 99, 114, 100, 118, 101, 1, 228, 165, 46, 238, 165, 47, 6, 232, 12, 232, 165, 37, 235, 165, 38, 242, 165, 39, 246, 165, 40, 228, 165, 34, 230, 165, 35, 231, 165, 36, 1, 228, 165, 41, 241, 165, 42, 16, 227, 165, 43, 16, 238, 165, 44, 42, 164, 247, 97, 190, 157, 98, 0, 35, 115, 165, 229, 29, 237, 15, 237, 163, 168, 238, 163, 173, 239, 163, 177, 242, 163, 186, 243, 163, 191, 229, 163, 124, 231, 163, 137, 232, 163, 141, 233, 163, 147, 119, 65, 119, 50, 120, 60, 121, 76, 122, 114, 225, 163, 98, 1, 228, 163, 210, 242, 163, 211, 2, 232, 163, 212, 235, 163, 213, 242, 163, 199, 5, 243, 9, 243, 163, 217, 246, 163, 218, 248, 163, 219, 229, 163, 214, 238, 163, 215, 242, 163, 216, 4, 225, 163, 220, 229, 163, 221, 230, 163, 222, 232, 163, 223, 247, 163, 224, 115, 50, 116, 72, 117, 88, 118, 16, 226, 163, 209, 3, 234, 163, 192, 241, 163, 193, 243, 163, 194, 244, 163, 195, 2, 239, 163, 196, 244, 163, 197, 246, 163, 198, 9, 237, 15, 237, 163, 204, 239, 163, 205, 242, 163, 206, 243, 163, 207, 245, 163, 208, 225, 163, 199, 227, 163, 200, 228, 163, 201, 231, 163, 202, 235, 163, 203, 106, 125, 110, 59, 110, 72, 111, 88, 112, 104, 113, 108, 114, 3, 225, 163, 187, 232, 163, 188, 248, 163, 189, 250, 163, 190, 2, 231, 163, 174, 237, 163, 175, 240, 163, 176, 2, 234, 163, 178, 237, 163, 179, 238, 163, 180, 16, 249, 163, 181, 3, 227, 163, 182, 233, 163, 183, 240, 163, 184, 246, 163, 185, 106, 70, 107, 120, 108, 164, 171, 109, 3, 232, 163, 169, 235, 163, 170, 241, 163, 171, 245, 163, 172, 7, 239, 12, 239, 163, 158, 242, 163, 159, 244, 163, 160, 250, 163, 161, 232, 163, 154, 233, 163, 155, 234, 163, 156, 238, 163, 157, 4, 227, 163, 162, 237, 163, 163, 241, 163, 164, 245, 163, 165, 246, 163, 166, 101, 106, 101, 98, 102, 148, 103, 162, 68, 104, 162, 75, 105, 6, 237, 12, 237, 163, 150, 238, 163, 151, 239, 163, 152, 241, 163, 153, 226, 163, 148, 231, 163, 149, 235, 163, 117, 7, 244, 12, 244, 163, 129, 247, 163, 130, 248, 163, 131, 250, 163, 132, 230, 163, 125, 232, 163, 126, 234, 163, 127, 237, 163, 128, 3, 228, 163, 133, 241, 163, 134, 244, 163, 135, 249, 163, 136, 2, 227, 163, 138, 238, 163, 139, 248, 163, 140, 5, 236, 9, 236, 163, 145, 239, 163, 141, 249, 163, 146, 226, 163, 142, 231, 163, 143, 233, 163, 144, 97, 50, 98, 100, 99, 144, 100, 16, 228, 163, 123, 7, 242, 12, 242, 163, 103, 243, 163, 104, 246, 163, 105, 248, 163, 106, 236, 163, 99, 238, 163, 100, 240, 163, 101, 241, 163, 102, 6, 228, 12, 228, 163, 110, 234, 163, 111, 240, 163, 112, 242, 163, 113, 225, 163, 107, 226, 163, 108, 227, 163, 109, 9, 237, 15, 237, 163, 118, 238, 163, 119, 239, 163, 120, 241, 163, 121, 245, 163, 122, 227, 163, 99, 230, 163, 114, 232, 163, 115, 233, 163, 116, 236, 163, 117, 0, 24, 77, 194, 101, 84, 192, 202, 87, 54, 87, 54, 88, 78, 89, 100, 90, 18, 97, 110, 226, 173, 203, 1, 97, 44, 99, 17, 104, 239, 169, 165, 17, 114, 225, 173, 202, 1, 112, 42, 115, 17, 117, 248, 105, 17, 101, 239, 169, 225, 1, 101, 44, 105, 17, 105, 233, 167, 23, 17, 122, 233, 167, 127, 84, 52, 85, 162, 135, 86, 18, 97, 105, 233, 173, 201, 5, 103, 70, 103, 74, 104, 80, 105, 1, 98, 42, 114, 16, 232, 173, 200, 16, 244, 1, 42, 163, 177, 66, 16, 212, 69, 17, 108, 231, 173, 197, 16, 97, 1, 225, 88, 233, 4, 42, 171, 166, 67, 50, 75, 54, 76, 58, 84, 16, 200, 171, 166, 16, 206, 167, 210, 16, 200, 173, 198, 16, 193, 173, 199, 1, 42, 165, 77, 77, 16, 214, 165, 77, 97, 48, 101, 124, 102, 17, 110, 231, 173, 196, 5, 109, 17, 109, 46, 110, 50, 118, 16, 244, 163, 167, 16, 236, 171, 137, 16, 231, 171, 222, 103, 52, 107, 56, 108, 1, 229, 171, 151, 245, 167, 92, 16, 226, 173, 194, 16, 242, 173, 195, 17, 108, 245, 171, 155, 18, 103, 97, 242, 171, 232, 80, 192, 195, 80, 162, 114, 82, 162, 168, 83, 6, 105, 67, 105, 66, 111, 106, 117, 144, 121, 1, 108, 42, 114, 16, 227, 165, 12, 16, 239, 173, 193, 1, 100, 62, 110, 1, 228, 171, 56, 232, 1, 42, 171, 73, 76, 16, 203, 171, 73, 16, 228, 173, 190, 2, 103, 46, 114, 56, 121, 16, 239, 165, 14, 1, 228, 171, 102, 239, 173, 191, 16, 225, 171, 110, 17, 110, 228, 173, 192, 97, 62, 103, 94, 104, 1, 97, 42, 114, 16, 228, 173, 189, 16, 247, 165, 97, 2, 109, 46, 114, 50, 117, 16, 242, 171, 38, 16, 242, 171, 92, 16, 226, 173, 188, 17, 110, 247, 163, 75, 4, 97, 56, 101, 74, 104, 80, 108, 118, 114, 17, 116, 233, 173, 185, 1, 108, 42, 117, 16, 227, 165, 28, 16, 237, 161, 17, 114, 237, 173, 183, 2, 97, 46, 108, 50, 110, 16, 248, 169, 229, 16, 231, 173, 184, 1, 233, 169, 213, 240, 169, 214, 17, 114, 228, 165, 244, 2, 106, 48, 111, 54, 117, 17, 110, 242, 169, 169, 17, 110, 231, 173, 186, 17, 104, 231, 173, 187, 77, 124, 78, 162, 164, 79, 3, 103, 66, 108, 72, 114, 78, 115, 1, 103, 42, 109, 16, 225, 173, 182, 16, 229, 169, 206, 17, 97, 237, 171, 65, 17, 99, 235, 173, 181, 1, 107, 42, 121, 16, 225, 169, 202, 16, 232, 169, 208, 7, 114, 46, 114, 90, 116, 96, 117, 102, 121, 17, 109, 242, 3, 42, 163, 206, 73, 46, 77, 50, 84, 16, 200, 173, 176, 16, 206, 167, 96, 16, 205, 163, 206, 17, 111, 239, 169, 88, 17, 101, 233, 173, 174, 17, 108, 244, 173, 175, 97, 66, 101, 116, 108, 154, 111, 1, 100, 42, 110, 16, 231, 165, 64, 16, 233, 173, 173, 3, 104, 50, 107, 54, 110, 58, 114, 16, 227, 173, 169, 16, 234, 173, 166, 16, 225, 173, 167, 1, 228, 169, 111, 233, 173, 168, 2, 100, 52, 110, 56, 114, 1, 227, 173, 171, 239, 173, 172, 16, 230, 169, 57, 16, 228, 173, 170, 17, 121, 237, 169, 63, 4, 97, 56, 98, 76, 101, 80, 107, 86, 115, 17, 104, 245, 173, 180, 1, 110, 42, 114, 16, 226, 173, 178, 16, 228, 173, 177, 17, 97, 244, 159, 17, 119, 225, 173, 179, 17, 111, 239, 165, 95, 71, 196, 21, 74, 195, 41, 74, 166, 186, 75, 166, 211, 76, 3, 97, 70, 101, 166, 150, 105, 166, 152, 121, 1, 99, 42, 100, 16, 233, 173, 165, 16, 233, 173, 164, 2, 110, 166, 115, 111, 166, 116, 116, 16, 238, 0, 26, 75, 193, 57, 84, 91, 87, 20, 87, 172, 105, 88, 172, 109, 89, 48, 90, 1, 193, 171, 240, 215, 171, 95, 16, 212, 173, 99, 84, 64, 85, 130, 86, 3, 193, 173, 94, 197, 173, 96, 206, 173, 95, 213, 163, 147, 11, 206, 20, 214, 9, 214, 171, 218, 215, 171, 199, 218, 171, 127, 206, 173, 162, 207, 171, 189, 210, 171, 197, 203, 197, 205, 203, 171, 178, 204, 173, 90, 205, 171, 177, 4, 193, 173, 163, 199, 173, 92, 211, 165, 96, 217, 173, 93, 218, 171, 244, 78, 130, 78, 138, 80, 162, 74, 82, 162, 109, 83, 12, 206, 23, 212, 12, 212, 173, 84, 213, 167, 154, 214, 173, 85, 217, 173, 161, 206, 173, 82, 207, 171, 101, 210, 173, 83, 202, 9, 202, 173, 80, 203, 171, 79, 205, 173, 81, 195, 173, 79, 197, 171, 126, 201, 171, 83, 6, 199, 12, 199, 171, 236, 201, 173, 66, 204, 165, 76, 207, 169, 126, 193, 173, 63, 195, 173, 64, 197, 173, 65, 10, 204, 20, 212, 9, 212, 173, 75, 215, 169, 217, 217, 165, 192, 204, 169, 234, 205, 173, 72, 210, 173, 73, 193, 173, 68, 197, 173, 69, 198, 173, 70, 199, 171, 193, 200, 165, 124, 3, 197, 173, 78, 207, 169, 74, 213, 167, 154, 215, 171, 27, 75, 158, 76, 162, 67, 77, 0, 16, 205, 29, 212, 15, 212, 169, 89, 213, 169, 41, 216, 171, 238, 217, 169, 12, 218, 173, 62, 205, 167, 61, 207, 173, 159, 209, 173, 60, 210, 173, 160, 198, 12, 198, 173, 59, 199, 169, 45, 203, 173, 158, 204, 163, 168, 193, 173, 157, 195, 173, 57, 196, 173, 58, 197, 171, 45, 1, 197, 173, 46, 205, 173, 156, 4, 201, 173, 52, 211, 173, 53, 212, 167, 249, 213, 173, 54, 214, 167, 255, 67, 189, 70, 95, 70, 76, 71, 92, 72, 162, 72, 73, 4, 195, 173, 43, 196, 167, 15, 210, 173, 155, 211, 167, 13, 212, 167, 37, 2, 201, 165, 122, 207, 165, 129, 210, 165, 135, 12, 206, 23, 211, 12, 211, 173, 38, 212, 173, 39, 213, 171, 243, 215, 173, 40, 206, 173, 35, 208, 173, 36, 209, 173, 37, 198, 8, 198, 173, 34, 200, 103, 204, 167, 115, 193, 173, 33, 194, 171, 246, 197, 173, 154, 4, 205, 173, 41, 206, 173, 42, 210, 165, 255, 212, 167, 2, 213, 167, 3, 67, 82, 68, 162, 66, 69, 5, 211, 9, 211, 165, 102, 212, 173, 148, 218, 173, 32, 193, 173, 28, 195, 173, 29, 197, 165, 89, 15, 207, 26, 214, 12, 214, 173, 23, 215, 171, 248, 217, 173, 152, 218, 165, 25, 207, 173, 20, 208, 173, 21, 210, 173, 22, 213, 171, 241, 201, 198, 139, 201, 173, 17, 204, 173, 18, 205, 173, 19, 206, 173, 151, 5, 203, 9, 203, 165, 33, 207, 173, 26, 218, 173, 153, 197, 165, 45, 199, 173, 77, 202, 173, 25, 52, 94, 52, 174, 134, 65, 120, 66, 11, 206, 20, 210, 9, 210, 169, 247, 213, 167, 61, 214, 173, 13, 206, 173, 10, 207, 173, 11, 209, 173, 12, 201, 9, 201, 171, 16, 202, 173, 8, 204, 173, 9, 193, 163, 191, 197, 173, 5, 198, 173, 6, 12, 210, 23, 213, 12, 213, 171, 237, 215, 173, 3, 216, 173, 4, 218, 163, 94, 210, 173, 0, 211, 173, 1, 212, 173, 2, 205, 9, 205, 173, 150, 207, 171, 254, 209, 171, 255, 196, 171, 251, 198, 173, 149, 204, 109, 42, 165, 96, 49, 44, 50, 17, 51, 176, 173, 148, 17, 55, 178, 167, 154, 16, 225, 169, 167, 16, 239, 1, 42, 167, 241, 76, 16, 193, 167, 241, 17, 112, 227, 167, 215, 2, 109, 46, 110, 50, 115, 16, 245, 167, 229, 16, 226, 167, 225, 1, 225, 167, 199, 226, 165, 204, 1, 97, 58, 112, 17, 97, 238, 1, 42, 167, 43, 74, 16, 208, 167, 43, 1, 109, 42, 118, 16, 225, 173, 144, 16, 239, 173, 143, 5, 110, 30, 110, 48, 111, 54, 116, 17, 104, 233, 173, 147, 17, 100, 225, 167, 131, 17, 114, 229, 1, 42, 167, 136, 75, 1, 208, 173, 48, 210, 167, 136, 97, 48, 104, 68, 105, 17, 116, 243, 173, 146, 1, 108, 42, 110, 16, 225, 173, 145, 16, 233, 165, 90, 2, 97, 46, 109, 50, 111, 16, 234, 171, 55, 16, 242, 169, 243, 16, 242, 1, 42, 167, 120, 75, 16, 200, 167, 120, 71, 52, 72, 162, 92, 73, 18, 116, 97, 236, 165, 107, 4, 101, 70, 108, 90, 111, 96, 114, 122, 117, 1, 106, 42, 114, 16, 245, 169, 210, 16, 242, 165, 208, 17, 111, 242, 1, 42, 165, 176, 71, 16, 197, 165, 176, 17, 97, 231, 165, 30, 1, 110, 42, 116, 16, 232, 165, 202, 1, 231, 173, 131, 237, 165, 103, 1, 97, 80, 101, 16, 235, 3, 42, 165, 91, 67, 46, 71, 50, 84, 16, 210, 163, 140, 16, 217, 173, 24, 16, 210, 165, 91, 16, 238, 173, 132, 5, 108, 24, 108, 48, 109, 54, 117, 17, 110, 231, 173, 142, 17, 117, 247, 165, 243, 16, 110, 1, 231, 165, 248, 240, 169, 103, 97, 48, 101, 156, 105, 17, 114, 225, 173, 141, 1, 110, 42, 116, 16, 242, 169, 56, 5, 239, 36, 239, 173, 135, 243, 76, 244, 3, 42, 163, 248, 72, 46, 77, 50, 84, 16, 215, 163, 248, 16, 203, 163, 239, 16, 207, 163, 241, 1, 42, 163, 235, 67, 16, 206, 163, 235, 226, 163, 253, 231, 173, 133, 233, 173, 134, 17, 98, 242, 5, 73, 23, 73, 58, 83, 62, 85, 2, 193, 173, 139, 203, 173, 137, 211, 173, 140, 16, 204, 165, 233, 16, 197, 173, 138, 42, 165, 233, 67, 42, 71, 16, 194, 173, 137, 16, 193, 173, 136, 67, 193, 6, 67, 112, 68, 162, 191, 69, 2, 103, 64, 108, 70, 116, 17, 104, 233, 1, 42, 117, 69, 1, 210, 173, 31, 212, 117, 17, 121, 240, 165, 87, 1, 98, 42, 121, 16, 237, 173, 130, 16, 225, 173, 129, 4, 97, 162, 104, 104, 162, 119, 111, 162, 134, 112, 162, 136, 121, 17, 114, 236, 10, 77, 51, 84, 20, 84, 46, 85, 56, 88, 16, 203, 173, 122, 1, 202, 171, 161, 210, 173, 121, 16, 193, 171, 233, 77, 46, 82, 62, 83, 16, 203, 173, 120, 2, 196, 173, 118, 203, 169, 3, 206, 165, 63, 2, 207, 173, 119, 211, 171, 43, 213, 171, 24, 42, 171, 24, 65, 56, 66, 60, 71, 76, 75, 1, 199, 167, 190, 218, 173, 50, 16, 204, 173, 115, 2, 193, 173, 116, 199, 163, 137, 217, 163, 124, 1, 197, 45, 210, 173, 117, 2, 107, 46, 110, 50, 114, 16, 233, 173, 113, 16, 237, 163, 229, 16, 243, 165, 18, 2, 97, 46, 101, 50, 114, 16, 243, 173, 114, 16, 237, 165, 5, 16, 242, 165, 2, 17, 112, 244, 165, 16, 17, 114, 244, 165, 203, 3, 101, 52, 105, 134, 111, 140, 117, 17, 112, 236, 173, 128, 17, 118, 225, 6, 73, 23, 73, 50, 77, 54, 78, 58, 80, 16, 203, 163, 198, 16, 206, 165, 235, 16, 213, 173, 125, 16, 208, 169, 136, 42, 165, 235, 66, 42, 70, 16, 202, 173, 124, 16, 212, 173, 123, 17, 97, 235, 173, 126, 17, 103, 242, 173, 127, 42, 162, 84, 65, 170, 49, 66, 5, 111, 31, 111, 62, 114, 68, 117, 1, 103, 42, 104, 16, 228, 173, 112, 16, 233, 173, 111, 17, 112, 239, 163, 252, 16, 97, 1, 232, 169, 232, 233, 173, 110, 97, 48, 101, 92, 104, 17, 107, 243, 173, 109, 3, 108, 50, 109, 54, 115, 58, 116, 16, 235, 173, 108, 16, 233, 173, 107, 16, 245, 163, 106, 16, 243, 163, 193, 17, 110, 231, 1, 42, 163, 173, 66, 16, 196, 163, 173, 0, 33, 74, 193, 192, 82, 192, 204, 86, 56, 86, 60, 87, 88, 88, 98, 89, 102, 90, 1, 210, 171, 239, 215, 171, 95, 4, 193, 173, 94, 196, 173, 95, 197, 173, 96, 206, 173, 95, 213, 163, 147, 1, 198, 173, 97, 211, 171, 88, 16, 203, 173, 98, 3, 196, 171, 250, 197, 171, 250, 212, 173, 99, 213, 171, 43, 82, 70, 83, 108, 84, 162, 87, 85, 3, 193, 171, 233, 199, 173, 92, 217, 173, 93, 218, 171, 244, 5, 211, 9, 211, 171, 43, 213, 171, 24, 215, 171, 27, 197, 173, 78, 200, 171, 95, 207, 169, 74, 14, 205, 26, 212, 12, 212, 173, 84, 213, 171, 24, 214, 173, 85, 217, 173, 86, 205, 173, 81, 206, 173, 82, 207, 171, 101, 210, 173, 83, 197, 12, 197, 171, 126, 201, 171, 83, 202, 173, 80, 203, 171, 79, 193, 171, 245, 195, 173, 79, 196, 171, 249, 14, 205, 26, 210, 12, 210, 171, 197, 214, 171, 218, 215, 163, 248, 218, 171, 127, 205, 171, 177, 206, 173, 91, 207, 171, 189, 208, 173, 90, 200, 12, 200, 171, 166, 202, 171, 161, 203, 171, 178, 204, 173, 90, 196, 173, 87, 198, 173, 88, 199, 173, 89, 78, 100, 78, 64, 79, 168, 150, 80, 120, 81, 2, 193, 173, 76, 207, 173, 77, 213, 171, 246, 9, 204, 15, 204, 165, 76, 207, 169, 126, 208, 169, 136, 209, 171, 255, 212, 171, 245, 193, 173, 63, 195, 173, 64, 197, 173, 65, 200, 163, 147, 201, 173, 66, 13, 205, 23, 212, 12, 212, 173, 75, 215, 169, 217, 217, 165, 192, 218, 173, 68, 205, 173, 72, 210, 173, 73, 211, 173, 74, 199, 12, 199, 171, 193, 200, 165, 124, 203, 173, 71, 204, 169, 234, 193, 173, 68, 197, 173, 69, 198, 173, 70, 74, 162, 72, 75, 162, 76, 76, 162, 100, 77, 0, 18, 206, 32, 213, 15, 213, 169, 41, 214, 165, 77, 216, 171, 238, 217, 169, 12, 218, 173, 62, 206, 165, 63, 207, 163, 241, 209, 173, 60, 210, 173, 61, 212, 169, 89, 198, 15, 198, 173, 59, 199, 169, 45, 203, 169, 3, 204, 163, 168, 205, 163, 206, 193, 173, 56, 195, 173, 57, 196, 173, 58, 197, 171, 45, 1, 207, 173, 45, 208, 167, 43, 7, 208, 12, 208, 173, 48, 210, 167, 136, 215, 173, 49, 218, 173, 50, 197, 173, 46, 199, 167, 190, 200, 167, 120, 205, 173, 47, 8, 211, 15, 211, 173, 53, 212, 167, 249, 213, 173, 54, 214, 167, 255, 217, 173, 55, 193, 167, 241, 194, 173, 51, 201, 173, 52, 203, 171, 73, 65, 193, 96, 69, 152, 69, 106, 70, 162, 65, 71, 162, 78, 72, 162, 117, 73, 7, 209, 12, 209, 173, 44, 210, 165, 112, 211, 167, 13, 212, 167, 37, 195, 173, 43, 196, 167, 15, 204, 165, 233, 206, 165, 235, 9, 210, 14, 210, 173, 31, 211, 165, 102, 212, 117, 213, 171, 246, 218, 173, 32, 193, 173, 28, 195, 173, 29, 197, 165, 89, 199, 155, 200, 173, 30, 4, 201, 165, 122, 207, 165, 129, 209, 171, 255, 210, 165, 135, 216, 165, 135, 11, 208, 20, 211, 9, 211, 173, 38, 212, 173, 39, 215, 173, 40, 208, 173, 36, 209, 173, 37, 210, 165, 91, 200, 8, 200, 103, 204, 167, 115, 206, 173, 35, 193, 173, 33, 197, 165, 176, 198, 173, 34, 6, 210, 12, 210, 165, 255, 212, 167, 2, 213, 167, 3, 214, 173, 6, 203, 163, 239, 205, 173, 41, 206, 173, 42, 65, 94, 66, 162, 76, 67, 162, 131, 68, 6, 203, 12, 203, 165, 33, 207, 173, 26, 217, 173, 8, 218, 173, 27, 196, 165, 45, 197, 165, 45, 202, 173, 25, 13, 209, 23, 212, 12, 212, 173, 2, 215, 173, 3, 216, 173, 4, 218, 163, 94, 209, 171, 255, 210, 173, 0, 211, 173, 1, 204, 11, 204, 109, 205, 163, 66, 206, 171, 248, 207, 171, 254, 196, 171, 251, 197, 171, 252, 198, 171, 253, 0, 16, 204, 28, 210, 14, 210, 169, 247, 212, 69, 213, 163, 206, 214, 173, 13, 217, 163, 124, 204, 173, 9, 206, 173, 10, 207, 173, 11, 209, 173, 12, 199, 12, 199, 163, 137, 200, 173, 7, 201, 171, 16, 202, 173, 8, 193, 163, 191, 196, 163, 173, 197, 173, 5, 198, 173, 6, 0, 16, 207, 29, 213, 15, 213, 171, 241, 214, 173, 23, 215, 171, 248, 217, 173, 24, 218, 165, 25, 207, 173, 20, 208, 173, 21, 210, 173, 22, 211, 171, 43, 201, 12, 201, 173, 17, 204, 173, 18, 205, 173, 19, 206, 163, 235, 196, 171, 239, 198, 173, 14, 199, 173, 15, 200, 173, 16, 52, 34, 52, 66, 53, 72, 55, 78, 56, 1, 56, 42, 57, 16, 176, 171, 43, 16, 182, 171, 250, 17, 49, 185, 171, 247, 17, 51, 178, 171, 248, 17, 51, 182, 171, 249, 42, 165, 96, 48, 80, 49, 162, 107, 50, 2, 48, 46, 51, 56, 56, 16, 176, 165, 45, 1, 176, 165, 25, 178, 171, 236, 16, 176, 117, 5, 51, 35, 51, 52, 53, 74, 54, 1, 177, 171, 88, 178, 165, 235, 3, 176, 163, 235, 180, 165, 235, 181, 167, 15, 185, 167, 37, 2, 179, 171, 237, 180, 171, 242, 183, 171, 243, 48, 52, 49, 74, 50, 1, 177, 165, 96, 185, 171, 241, 3, 178, 171, 236, 179, 165, 96, 181, 169, 247, 185, 171, 237, 6, 181, 11, 181, 155, 183, 171, 239, 184, 171, 240, 185, 165, 96, 177, 171, 236, 179, 171, 238, 180, 171, 127, 2, 52, 46, 53, 62, 55, 16, 178, 171, 24, 2, 178, 163, 235, 179, 171, 244, 181, 171, 245, 3, 176, 171, 24, 177, 171, 24, 180, 171, 246, 181, 165, 45, 4, 100, 54, 103, 60, 104, 66, 114, 70, 118, 17, 115, 244, 75, 17, 108, 237, 165, 119, 17, 104, 226, 173, 100, 17, 111, 237, 99, 1, 97, 60, 109, 1, 233, 157, 238, 1, 42, 163, 66, 65, 16, 205, 163, 66, 16, 226, 0, 19, 76, 98, 81, 49, 81, 60, 83, 64, 84, 80, 85, 106, 89, 1, 197, 171, 250, 212, 171, 128, 16, 193, 173, 76, 2, 193, 171, 245, 196, 171, 249, 217, 173, 86, 4, 199, 145, 200, 169, 40, 202, 173, 105, 206, 173, 91, 210, 173, 106, 16, 203, 173, 101, 76, 60, 77, 70, 78, 98, 79, 102, 80, 1, 203, 173, 71, 211, 173, 74, 1, 194, 173, 51, 217, 173, 55, 4, 193, 173, 56, 205, 171, 5, 206, 167, 112, 210, 173, 61, 213, 173, 103, 16, 199, 173, 104, 16, 205, 173, 67, 69, 46, 69, 66, 71, 74, 73, 78, 74, 100, 75, 2, 200, 165, 4, 205, 173, 47, 215, 173, 49, 1, 199, 155, 200, 173, 30, 16, 194, 173, 101, 3, 196, 169, 14, 206, 173, 102, 209, 173, 44, 210, 165, 112, 16, 207, 173, 45, 42, 155, 65, 50, 66, 60, 67, 70, 68, 16, 218, 173, 27, 1, 197, 171, 252, 198, 171, 253, 1, 200, 173, 7, 213, 171, 5, 1, 195, 169, 13, 206, 171, 229, 0, 33, 117, 136, 230, 43, 242, 33, 242, 155, 243, 163, 73, 246, 163, 83, 249, 163, 92, 250, 1, 42, 46, 65, 18, 114, 97, 226, 163, 96, 2, 42, 163, 94, 73, 138, 82, 16, 213, 163, 97, 230, 81, 235, 103, 237, 117, 238, 127, 122, 45, 122, 44, 225, 37, 226, 45, 229, 75, 16, 234, 1, 42, 46, 65, 18, 114, 97, 226, 163, 96, 3, 42, 163, 94, 49, 46, 73, 52, 83, 16, 213, 163, 97, 17, 55, 178, 163, 97, 1, 209, 163, 95, 210, 163, 96, 117, 56, 118, 60, 119, 82, 121, 1, 226, 163, 93, 242, 163, 92, 16, 249, 163, 82, 3, 236, 163, 84, 238, 163, 85, 244, 163, 86, 245, 163, 87, 3, 225, 163, 88, 226, 163, 89, 239, 163, 90, 248, 163, 91, 107, 150, 111, 105, 111, 68, 112, 78, 114, 104, 115, 162, 80, 116, 2, 225, 163, 79, 231, 163, 80, 234, 163, 81, 2, 234, 137, 237, 139, 250, 141, 5, 242, 6, 242, 149, 243, 151, 250, 153, 227, 143, 228, 145, 229, 147, 9, 239, 15, 239, 163, 68, 241, 163, 69, 243, 163, 70, 249, 163, 71, 250, 163, 72, 226, 155, 227, 50, 232, 163, 65, 237, 163, 66, 238, 163, 67, 2, 42, 157, 78, 44, 80, 18, 97, 108, 237, 161, 18, 98, 97, 244, 159, 4, 225, 163, 74, 229, 163, 75, 231, 163, 76, 239, 163, 77, 244, 163, 78, 107, 60, 108, 62, 109, 88, 110, 3, 227, 129, 235, 131, 238, 133, 249, 135, 16, 235, 105, 5, 238, 6, 238, 113, 243, 109, 244, 115, 225, 107, 226, 109, 233, 111, 3, 237, 119, 238, 121, 239, 123, 240, 125, 101, 37, 101, 48, 103, 54, 104, 80, 106, 16, 231, 101, 1, 226, 77, 249, 79, 5, 237, 6, 237, 89, 239, 91, 241, 93, 227, 83, 228, 85, 231, 87, 2, 225, 95, 236, 97, 239, 99, 97, 72, 98, 82, 99, 100, 100, 5, 240, 6, 240, 69, 249, 71, 250, 73, 225, 63, 229, 65, 234, 67, 2, 233, 39, 235, 41, 245, 43, 4, 233, 47, 241, 49, 242, 51, 244, 53, 249, 55, 2, 228, 57, 229, 59, 232, 61, 170, 170, 4, 0, 39, 0, 74, 0, 55, 0, 69, 0, 170, 170, 149, 27, 0, 144, 232, 28, 0, 144, 157, 47, 0, 144, 43, 24, 0, 16, 4, 0, 0, 0, 80, 0, 0, 0, 50, 0, 0, 0, 4, 0, 0, 0, 3, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 141, 6, 0, 0, 0, 0, 8, 9, 10, 0, 2, 0, 0, 0, 11, 0, 9, 0, 2, 1, 9, 0, 1, 1, 10, 0, 10, 0, 0, 0, 0, 0, 0, 0, 10, 12, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 11, 0, 0, 0, 11, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 1, 0, 1, 0, 0, 0, 0, 13, 1, 0, 0, 13, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 0, 2, 0, 0, 1, 1, 2, 1, 0, 1, 2, 3, 1, 1, 0, 2, 1, 0, 1, 1, 2, 0, 1, 1, 1, 1, 1, 1, 1, 0, 2, 2, 1, 2, 0, 2, 2, 2, 1, 1, 2, 1, 0, 1, 2, 2, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 2, 1, 1, 2, 1, 0, 2, 1, 0, 2, 1, 2, 1, 1, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 1, 0, 2, 0, 1, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 13, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 5, 0, 2, 1, 2, 1, 1, 1, 0, 0, 2, 1, 1, 0, 2, 1, 1, 2, 2, 3, 0, 1, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 0, 1, 2, 0, 0, 0, 1, 0, 2, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 2, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 2, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 2, 1, 1, 1, 2, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 4, 0, 4, 0, 1, 1, 1, 2, 1, 3, 0, 0, 1, 1, 1, 1, 6, 3, 2, 4, 2, 1, 1, 1, 1, 2, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 2, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 2, 1, 1, 1, 0, 0, 1, 1, 2, 1, 0, 0, 0, 7, 1, 1, 0, 0, 1, 0, 2, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 2, 1, 1, 1, 2, 0, 2, 1, 1, 1, 0, 2, 1, 0, 1, 1, 1, 0, 1, 1, 1, 1, 4, 1, 1, 0, 1, 0, 2, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 3, 14, 0, 0, 0, 0, 7, 0, 0, 0, 0, 0, 2, 1, 1, 0, 2, 0, 2, 0, 2, 0, 7, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 235, 8, 0, 0, 0, 25, 110, 195, 217, 116, 193, 111, 119, 179, 119, 162, 105, 120, 162, 115, 121, 162, 131, 122, 3, 108, 52, 109, 174, 14, 232, 52, 245, 17, 101, 238, 53, 18, 237, 109, 243, 53, 17, 122, 232, 32, 1, 42, 133, 72, 17, 97, 110, 1, 243, 152, 244, 18, 72, 97, 110, 1, 243, 71, 244, 32, 7, 179, 20, 179, 42, 180, 38, 181, 54, 182, 5, 179, 199, 112, 179, 41, 180, 41, 182, 41, 16, 181, 41, 42, 43, 176, 38, 177, 34, 178, 5, 179, 199, 92, 179, 41, 180, 41, 182, 41, 5, 179, 199, 82, 179, 41, 180, 41, 182, 41, 19, 72, 97, 110, 244, 63, 1, 117, 42, 239, 17, 102, 242, 53, 18, 245, 122, 232, 53, 1, 109, 42, 232, 17, 101, 238, 53, 2, 237, 162, 65, 246, 34, 247, 17, 109, 231, 53, 3, 105, 50, 117, 56, 233, 62, 239, 17, 101, 238, 53, 18, 232, 121, 233, 53, 18, 229, 122, 232, 53, 17, 101, 238, 52, 1, 42, 133, 72, 22, 101, 98, 242, 76, 97, 116, 238, 53, 116, 74, 117, 162, 142, 118, 1, 107, 44, 114, 18, 239, 101, 244, 53, 1, 235, 34, 244, 17, 109, 243, 53, 14, 225, 75, 235, 14, 235, 162, 128, 238, 162, 178, 239, 162, 175, 244, 17, 114, 245, 53, 225, 78, 229, 108, 231, 168, 162, 233, 17, 101, 238, 52, 1, 42, 133, 69, 22, 116, 104, 233, 76, 97, 116, 238, 53, 17, 101, 238, 52, 1, 42, 133, 84, 22, 97, 109, 236, 76, 97, 116, 238, 53, 17, 101, 238, 52, 1, 42, 133, 84, 22, 101, 108, 245, 76, 97, 116, 238, 53, 108, 28, 108, 54, 109, 170, 72, 117, 54, 120, 18, 249, 109, 231, 53, 18, 232, 101, 238, 53, 1, 237, 162, 87, 249, 18, 107, 108, 238, 53, 100, 50, 101, 162, 217, 107, 18, 231, 109, 231, 53, 18, 248, 109, 231, 53, 4, 114, 84, 122, 90, 231, 172, 210, 242, 162, 139, 250, 17, 114, 245, 52, 1, 42, 133, 76, 22, 97, 116, 238, 67, 121, 114, 236, 53, 18, 235, 109, 243, 53, 18, 243, 117, 250, 53, 113, 193, 97, 113, 162, 172, 114, 164, 73, 115, 0, 17, 232, 85, 241, 45, 241, 48, 242, 52, 244, 40, 245, 166, 232, 247, 17, 101, 238, 53, 16, 115, 1, 232, 41, 242, 32, 2, 42, 133, 67, 52, 76, 22, 97, 116, 238, 67, 121, 114, 236, 43, 22, 121, 114, 236, 76, 97, 116, 238, 43, 232, 50, 233, 66, 238, 34, 239, 17, 101, 238, 53, 2, 98, 170, 73, 104, 34, 115, 16, 242, 41, 17, 101, 238, 52, 1, 42, 133, 83, 22, 105, 110, 232, 76, 97, 116, 238, 53, 112, 46, 112, 82, 114, 88, 115, 94, 225, 168, 196, 228, 17, 101, 238, 52, 1, 42, 133, 65, 22, 114, 97, 226, 76, 97, 116, 238, 53, 18, 246, 111, 242, 53, 18, 239, 115, 227, 53, 1, 232, 172, 235, 249, 17, 97, 225, 41, 100, 52, 103, 64, 104, 72, 107, 18, 231, 109, 231, 53, 1, 227, 34, 238, 17, 115, 227, 53, 19, 227, 107, 108, 238, 53, 18, 245, 97, 242, 53, 4, 117, 58, 118, 144, 119, 162, 100, 120, 162, 109, 245, 17, 101, 243, 53, 12, 236, 22, 243, 11, 243, 42, 247, 38, 248, 34, 249, 17, 113, 245, 53, 236, 38, 240, 34, 242, 17, 113, 245, 53, 231, 9, 231, 38, 232, 34, 235, 17, 113, 245, 53, 226, 38, 228, 34, 230, 17, 113, 245, 53, 13, 237, 22, 240, 11, 240, 42, 243, 38, 247, 34, 250, 17, 113, 245, 53, 237, 38, 238, 34, 239, 17, 113, 245, 53, 232, 11, 232, 42, 233, 38, 234, 34, 236, 17, 113, 245, 53, 225, 38, 227, 34, 229, 17, 113, 245, 53, 3, 225, 42, 227, 38, 232, 34, 243, 17, 113, 245, 53, 10, 239, 20, 244, 9, 244, 38, 245, 34, 247, 17, 113, 245, 53, 239, 38, 240, 34, 242, 17, 113, 245, 53, 225, 46, 227, 42, 232, 38, 236, 34, 238, 17, 113, 245, 53, 17, 113, 245, 53, 3, 97, 166, 80, 237, 48, 238, 162, 129, 247, 17, 102, 242, 53, 17, 100, 229, 53, 110, 162, 107, 111, 162, 207, 112, 8, 107, 77, 107, 170, 34, 115, 90, 225, 104, 243, 134, 244, 17, 112, 244, 162, 128, 7, 179, 195, 168, 179, 168, 172, 180, 168, 169, 181, 168, 166, 182, 1, 177, 41, 182, 41, 1, 229, 162, 207, 244, 17, 112, 243, 53, 17, 101, 238, 52, 1, 42, 133, 71, 22, 117, 114, 245, 76, 97, 116, 238, 53, 17, 101, 238, 52, 1, 42, 133, 65, 22, 114, 97, 226, 76, 97, 116, 238, 53, 98, 54, 99, 172, 175, 101, 54, 103, 18, 225, 97, 242, 53, 18, 244, 112, 243, 53, 18, 236, 109, 243, 53, 10, 121, 68, 238, 23, 238, 46, 239, 56, 249, 17, 101, 238, 53, 16, 110, 1, 226, 53, 239, 53, 1, 100, 120, 110, 16, 226, 35, 121, 72, 226, 86, 229, 17, 101, 238, 52, 1, 42, 133, 68, 22, 101, 118, 225, 76, 97, 116, 238, 53, 1, 228, 164, 184, 238, 17, 101, 238, 53, 1, 100, 44, 110, 1, 238, 53, 239, 35, 16, 225, 49, 97, 166, 129, 104, 52, 105, 58, 108, 66, 115, 18, 239, 101, 238, 53, 18, 228, 103, 238, 53, 19, 241, 107, 108, 238, 53, 19, 229, 108, 117, 249, 53, 4, 107, 84, 114, 92, 227, 164, 11, 237, 162, 181, 242, 17, 101, 238, 52, 1, 42, 133, 79, 22, 114, 121, 225, 76, 97, 116, 238, 53, 19, 233, 107, 108, 238, 53, 2, 227, 164, 253, 238, 34, 243, 17, 109, 243, 53, 103, 194, 76, 106, 193, 146, 106, 162, 141, 107, 162, 154, 108, 162, 253, 109, 0, 17, 231, 62, 238, 30, 238, 162, 153, 242, 164, 175, 243, 162, 123, 244, 162, 112, 249, 17, 101, 238, 52, 1, 42, 133, 77, 22, 121, 109, 242, 76, 97, 116, 238, 53, 231, 162, 174, 233, 162, 88, 235, 68, 236, 17, 101, 238, 52, 1, 42, 133, 77, 22, 108, 121, 237, 76, 97, 116, 238, 53, 17, 98, 231, 53, 110, 34, 110, 58, 113, 64, 115, 70, 117, 164, 237, 118, 18, 230, 109, 238, 53, 18, 240, 122, 232, 53, 18, 231, 109, 243, 53, 1, 232, 170, 123, 233, 17, 109, 243, 53, 97, 52, 101, 58, 102, 64, 105, 18, 238, 109, 243, 53, 18, 248, 109, 243, 53, 18, 239, 109, 243, 53, 2, 225, 70, 226, 66, 229, 17, 101, 238, 53, 1, 97, 42, 246, 17, 105, 228, 53, 1, 235, 34, 248, 17, 109, 243, 53, 9, 235, 51, 235, 46, 237, 50, 238, 80, 245, 110, 249, 17, 114, 245, 53, 17, 101, 238, 52, 1, 42, 133, 75, 22, 104, 109, 242, 76, 97, 116, 238, 53, 17, 101, 238, 52, 1, 42, 133, 75, 22, 110, 100, 225, 76, 97, 116, 238, 53, 17, 116, 242, 53, 114, 54, 118, 60, 120, 72, 225, 78, 231, 17, 102, 242, 53, 18, 233, 101, 238, 53, 1, 226, 34, 242, 17, 109, 243, 53, 18, 228, 109, 243, 53, 17, 101, 238, 52, 1, 42, 133, 71, 22, 101, 111, 242, 76, 97, 116, 238, 53, 14, 117, 59, 226, 30, 226, 80, 231, 162, 145, 238, 168, 169, 239, 17, 101, 238, 52, 1, 42, 133, 76, 22, 97, 111, 239, 76, 97, 116, 238, 53, 17, 100, 229, 41, 117, 50, 119, 56, 122, 64, 225, 17, 105, 244, 53, 18, 225, 102, 242, 53, 19, 231, 108, 117, 249, 53, 18, 232, 122, 232, 53, 111, 42, 111, 72, 114, 78, 115, 92, 116, 2, 231, 48, 239, 34, 243, 18, 108, 117, 249, 53, 17, 108, 246, 53, 18, 250, 101, 238, 53, 1, 233, 34, 237, 18, 108, 117, 249, 53, 19, 237, 108, 117, 249, 53, 99, 70, 105, 82, 107, 2, 226, 38, 239, 34, 243, 18, 108, 117, 249, 53, 18, 108, 117, 249, 53, 1, 229, 34, 230, 17, 109, 243, 53, 18, 247, 109, 243, 53, 103, 106, 104, 162, 110, 105, 5, 228, 10, 228, 168, 47, 231, 34, 243, 17, 101, 238, 53, 100, 46, 107, 54, 225, 17, 101, 238, 53, 19, 225, 108, 117, 249, 53, 18, 244, 105, 245, 53, 10, 117, 30, 236, 11, 236, 162, 227, 238, 162, 224, 245, 17, 104, 233, 53, 117, 46, 225, 34, 228, 17, 101, 238, 53, 1, 233, 34, 238, 17, 103, 238, 53, 97, 56, 107, 76, 110, 84, 111, 90, 115, 18, 247, 100, 229, 41, 2, 225, 162, 80, 238, 164, 87, 248, 17, 111, 237, 53, 19, 240, 107, 112, 229, 53, 18, 247, 103, 238, 53, 19, 237, 107, 111, 235, 53, 6, 225, 35, 225, 132, 242, 74, 244, 166, 173, 249, 17, 114, 245, 52, 1, 42, 133, 65, 22, 114, 109, 238, 67, 121, 114, 236, 53, 1, 98, 40, 115, 16, 232, 41, 16, 243, 41, 97, 48, 106, 74, 115, 18, 238, 122, 232, 53, 2, 229, 48, 235, 164, 11, 247, 17, 101, 238, 53, 17, 111, 237, 53, 18, 233, 109, 243, 53, 100, 192, 224, 100, 136, 101, 162, 72, 102, 3, 102, 50, 117, 56, 239, 106, 249, 17, 110, 236, 53, 18, 237, 102, 230, 53, 6, 232, 11, 232, 42, 233, 38, 241, 34, 246, 17, 102, 230, 53, 226, 38, 229, 34, 230, 17, 102, 230, 53, 17, 102, 230, 53, 17, 100, 225, 53, 2, 116, 52, 117, 58, 225, 16, 110, 1, 226, 49, 239, 49, 18, 249, 110, 229, 53, 18, 240, 109, 243, 53, 7, 239, 70, 239, 162, 159, 243, 46, 244, 150, 245, 17, 101, 243, 53, 17, 101, 243, 162, 128, 7, 179, 12, 179, 66, 180, 62, 181, 58, 182, 1, 177, 41, 182, 41, 42, 43, 176, 38, 177, 78, 178, 4, 176, 41, 178, 41, 179, 41, 180, 41, 181, 41, 4, 176, 41, 178, 41, 179, 41, 180, 41, 181, 41, 1, 177, 41, 182, 41, 17, 102, 233, 53, 110, 152, 121, 160, 229, 162, 85, 238, 17, 101, 238, 162, 128, 7, 179, 23, 179, 98, 180, 48, 181, 90, 182, 1, 178, 41, 182, 41, 4, 176, 39, 177, 39, 179, 39, 180, 39, 181, 39, 42, 43, 176, 48, 177, 44, 178, 1, 178, 41, 182, 41, 4, 176, 41, 177, 41, 179, 41, 180, 39, 181, 41, 19, 226, 107, 108, 238, 53, 19, 239, 107, 108, 238, 53, 97, 162, 86, 98, 164, 95, 99, 10, 111, 44, 122, 22, 122, 48, 239, 164, 127, 249, 17, 101, 238, 53, 1, 232, 34, 239, 17, 122, 232, 53, 17, 122, 232, 53, 111, 48, 112, 54, 114, 18, 243, 102, 242, 53, 18, 225, 109, 243, 53, 18, 248, 122, 232, 53, 100, 56, 101, 62, 104, 70, 106, 76, 107, 18, 226, 97, 242, 53, 18, 239, 122, 232, 53, 19, 226, 102, 105, 236, 53, 18, 242, 101, 238, 53, 18, 249, 122, 232, 53, 0, 21, 117, 140, 226, 101, 242, 85, 242, 74, 249, 162, 76, 250, 17, 114, 245, 52, 1, 42, 133, 76, 22, 97, 116, 238, 67, 121, 114, 236, 53, 17, 97, 242, 162, 128, 7, 179, 19, 179, 60, 180, 38, 181, 34, 182, 5, 180, 34, 180, 41, 181, 41, 182, 41, 16, 179, 41, 42, 43, 176, 38, 177, 34, 178, 5, 180, 15, 180, 41, 181, 41, 182, 41, 5, 180, 6, 180, 41, 181, 41, 182, 41, 176, 41, 177, 41, 178, 41, 17, 101, 243, 53, 226, 162, 224, 230, 42, 235, 17, 101, 238, 53, 17, 110, 236, 53, 117, 164, 10, 118, 52, 121, 58, 122, 78, 225, 18, 115, 115, 249, 41, 18, 236, 97, 242, 53, 3, 232, 42, 236, 38, 238, 34, 240, 17, 97, 242, 53, 18, 226, 97, 250, 53, 102, 53, 108, 30, 108, 62, 112, 68, 114, 3, 241, 42, 243, 38, 249, 34, 250, 17, 97, 242, 53, 18, 238, 115, 241, 53, 1, 227, 34, 228, 17, 97, 242, 53, 102, 48, 105, 54, 106, 18, 240, 97, 242, 53, 18, 226, 97, 242, 53, 19, 233, 115, 121, 242, 53, 97, 62, 98, 78, 99, 90, 100, 144, 101, 1, 226, 34, 227, 17, 97, 242, 53, 2, 229, 38, 239, 50, 244, 17, 115, 241, 53, 1, 232, 34, 246, 17, 97, 242, 53, 5, 247, 9, 247, 38, 248, 34, 249, 17, 97, 242, 53, 232, 46, 237, 34, 241, 17, 97, 242, 53, 17, 101, 238, 53, 17, 97, 242, 53, 18, 230, 97, 242, 53, 12, 118, 66, 232, 42, 232, 62, 238, 66, 242, 96, 243, 1, 104, 40, 115, 16, 232, 41, 16, 242, 41, 17, 104, 233, 53, 17, 101, 238, 52, 1, 42, 133, 66, 22, 101, 110, 231, 76, 97, 116, 238, 53, 17, 102, 242, 53, 118, 46, 122, 58, 229, 17, 114, 245, 53, 1, 229, 34, 245, 17, 109, 243, 53, 18, 227, 109, 231, 53, 106, 27, 106, 48, 109, 68, 116, 18, 234, 109, 243, 53, 1, 238, 42, 241, 17, 109, 231, 53, 17, 109, 243, 53, 18, 237, 109, 231, 53, 98, 48, 101, 54, 104, 18, 242, 109, 231, 53, 18, 250, 97, 242, 53, 18, 237, 101, 238, 53, 170, 5, 0, 125, 0, 115, 0, 104, 0, 85, 0, 69, 0, 208, 33, 0, 224, 234, 47, 0, 144, 253, 47, 0, 144, 215, 33, 0, 16, 127, 35, 0, 16, 2, 0, 32, 0, 79, 0, 170, 170, 201, 33, 0, 32, 187, 37, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 88, 73, 0, 80, 7, 0, 0, 0, 170, 8, 0, 0, 148, 52, 0, 0, 148, 52, 0, 0, 61, 7, 0, 0, 1, 0, 0, 0, 148, 52, 0, 0, 97, 97, 105, 0, 97, 97, 107, 0, 97, 97, 117, 0, 97, 98, 105, 0, 97, 98, 113, 0, 97, 98, 114, 0, 97, 98, 116, 0, 97, 98, 121, 0, 97, 99, 100, 0, 97, 99, 101, 0, 97, 99, 104, 0, 97, 100, 97, 0, 97, 100, 101, 0, 97, 100, 106, 0, 97, 100, 112, 0, 97, 100, 121, 0, 97, 100, 122, 0, 97, 101, 98, 0, 97, 101, 121, 0, 97, 103, 99, 0, 97, 103, 100, 0, 97, 103, 103, 0, 97, 103, 109, 0, 97, 103, 111, 0, 97, 103, 113, 0, 97, 104, 97, 0, 97, 104, 108, 0, 97, 104, 111, 0, 97, 106, 103, 0, 97, 107, 107, 0, 97, 108, 97, 0, 97, 108, 110, 0, 97, 108, 116, 0, 97, 109, 109, 0, 97, 109, 110, 0, 97, 109, 112, 0, 97, 110, 99, 0, 97, 110, 107, 0, 97, 110, 110, 0, 97, 110, 121, 0, 97, 111, 106, 0, 97, 111, 109, 0, 97, 111, 122, 0, 97, 112, 99, 0, 97, 112, 100, 0, 97, 112, 101, 0, 97, 112, 114, 0, 97, 112, 115, 0, 97, 112, 122, 0, 97, 114, 99, 95, 78, 98, 97, 116, 0, 97, 114, 99, 95, 80, 97, 108, 109, 0, 97, 114, 104, 0, 97, 114, 110, 0, 97, 114, 111, 0, 97, 114, 113, 0, 97, 114, 115, 0, 97, 114, 121, 0, 97, 114, 122, 0, 97, 115, 97, 0, 97, 115, 101, 0, 97, 115, 103, 0, 97, 115, 111, 0, 97, 115, 116, 0, 97, 116, 97, 0, 97, 116, 103, 0, 97, 116, 106, 0, 97, 117, 121, 0, 97, 118, 108, 0, 97, 118, 110, 0, 97, 118, 117, 0, 97, 119, 97, 0, 97, 119, 98, 0, 97, 119, 111, 0, 97, 119, 120, 0, 97, 121, 98, 0, 97, 122, 95, 65, 114, 97, 98, 0, 97, 122, 95, 73, 81, 0, 97, 122, 95, 73, 82, 0, 97, 122, 95, 82, 85, 0, 98, 97, 108, 0, 98, 97, 110, 0, 98, 97, 112, 0, 98, 97, 114, 0, 98, 97, 115, 0, 98, 97, 118, 0, 98, 97, 120, 0, 98, 98, 97, 0, 98, 98, 98, 0, 98, 98, 99, 0, 98, 98, 100, 0, 98, 98, 106, 0, 98, 98, 112, 0, 98, 98, 114, 0, 98, 99, 102, 0, 98, 99, 104, 0, 98, 99, 105, 0, 98, 99, 109, 0, 98, 99, 110, 0, 98, 99, 111, 0, 98, 99, 113, 0, 98, 99, 117, 0, 98, 100, 100, 0, 98, 101, 102, 0, 98, 101, 104, 0, 98, 101, 106, 0, 98, 101, 109, 0, 98, 101, 116, 0, 98, 101, 119, 0, 98, 101, 120, 0, 98, 101, 122, 0, 98, 102, 100, 0, 98, 102, 113, 0, 98, 102, 116, 0, 98, 102, 121, 0, 98, 103, 99, 0, 98, 103, 110, 0, 98, 103, 120, 0, 98, 104, 98, 0, 98, 104, 103, 0, 98, 104, 105, 0, 98, 104, 108, 0, 98, 104, 111, 0, 98, 104, 121, 0, 98, 105, 98, 0, 98, 105, 103, 0, 98, 105, 107, 0, 98, 105, 109, 0, 98, 105, 110, 0, 98, 105, 111, 0, 98, 105, 113, 0, 98, 106, 104, 0, 98, 106, 105, 0, 98, 106, 106, 0, 98, 106, 110, 0, 98, 106, 111, 0, 98, 106, 114, 0, 98, 106, 116, 0, 98, 106, 122, 0, 98, 107, 99, 0, 98, 107, 109, 0, 98, 107, 113, 0, 98, 107, 117, 0, 98, 107, 118, 0, 98, 108, 116, 0, 98, 109, 104, 0, 98, 109, 107, 0, 98, 109, 113, 0, 98, 109, 117, 0, 98, 110, 0, 98, 110, 103, 0, 98, 110, 109, 0, 98, 110, 112, 0, 98, 111, 106, 0, 98, 111, 109, 0, 98, 111, 110, 0, 98, 112, 121, 0, 98, 113, 99, 0, 98, 113, 105, 0, 98, 113, 112, 0, 98, 113, 118, 0, 98, 114, 97, 0, 98, 114, 104, 0, 98, 114, 120, 0, 98, 114, 122, 0, 98, 115, 0, 98, 115, 106, 0, 98, 115, 113, 0, 98, 115, 115, 0, 98, 115, 116, 0, 98, 116, 111, 0, 98, 116, 116, 0, 98, 116, 118, 0, 98, 117, 97, 0, 98, 117, 99, 0, 98, 117, 100, 0, 98, 117, 103, 0, 98, 117, 107, 0, 98, 117, 109, 0, 98, 117, 111, 0, 98, 117, 115, 0, 98, 117, 117, 0, 98, 118, 98, 0, 98, 119, 100, 0, 98, 119, 114, 0, 98, 120, 104, 0, 98, 121, 101, 0, 98, 121, 110, 0, 98, 121, 114, 0, 98, 121, 115, 0, 98, 121, 118, 0, 98, 121, 120, 0, 98, 122, 97, 0, 98, 122, 101, 0, 98, 122, 102, 0, 98, 122, 104, 0, 98, 122, 119, 0, 99, 97, 110, 0, 99, 98, 106, 0, 99, 99, 104, 0, 99, 99, 112, 0, 99, 101, 98, 0, 99, 102, 97, 0, 99, 103, 103, 0, 99, 104, 107, 0, 99, 104, 109, 0, 99, 104, 112, 0, 99, 104, 114, 0, 99, 105, 99, 0, 99, 106, 97, 0, 99, 106, 109, 0, 99, 106, 118, 0, 99, 107, 98, 0, 99, 107, 108, 0, 99, 107, 111, 0, 99, 107, 121, 0, 99, 108, 97, 0, 99, 109, 101, 0, 99, 109, 103, 0, 99, 111, 112, 0, 99, 112, 115, 0, 99, 114, 104, 0, 99, 114, 106, 0, 99, 114, 107, 0, 99, 114, 108, 0, 99, 114, 109, 0, 99, 114, 115, 0, 99, 115, 98, 0, 99, 115, 119, 0, 99, 116, 100, 0, 99, 117, 95, 71, 108, 97, 103, 0, 99, 118, 0, 100, 97, 100, 0, 100, 97, 102, 0, 100, 97, 103, 0, 100, 97, 104, 0, 100, 97, 107, 0, 100, 97, 114, 0, 100, 97, 118, 0, 100, 98, 100, 0, 100, 98, 113, 0, 100, 99, 99, 0, 100, 100, 110, 0, 100, 101, 100, 0, 100, 101, 110, 0, 100, 103, 97, 0, 100, 103, 104, 0, 100, 103, 105, 0, 100, 103, 108, 0, 100, 103, 114, 0, 100, 103, 122, 0, 100, 105, 97, 0, 100, 106, 101, 0, 100, 110, 106, 0, 100, 111, 98, 0, 100, 111, 105, 0, 100, 111, 112, 0, 100, 111, 119, 0, 100, 114, 104, 0, 100, 114, 105, 0, 100, 114, 115, 0, 100, 115, 98, 0, 100, 116, 109, 0, 100, 116, 112, 0, 100, 116, 115, 0, 100, 116, 121, 0, 100, 117, 97, 0, 100, 117, 99, 0, 100, 117, 100, 0, 100, 117, 103, 0, 100, 118, 0, 100, 118, 97, 0, 100, 119, 119, 0, 100, 121, 111, 0, 100, 121, 117, 0, 100, 122, 103, 0, 101, 98, 117, 0, 101, 102, 105, 0, 101, 103, 108, 0, 101, 103, 121, 0, 101, 107, 97, 0, 101, 107, 121, 0, 101, 109, 97, 0, 101, 109, 105, 0, 101, 110, 95, 83, 104, 97, 119, 0, 101, 110, 110, 0, 101, 110, 113, 0, 101, 114, 105, 0, 101, 115, 103, 0, 101, 115, 117, 0, 101, 116, 114, 0, 101, 116, 116, 0, 101, 116, 117, 0, 101, 116, 120, 0, 101, 119, 111, 0, 101, 120, 116, 0, 102, 97, 97, 0, 102, 97, 98, 0, 102, 97, 103, 0, 102, 97, 105, 0, 102, 97, 110, 0, 102, 102, 0, 102, 102, 95, 65, 100, 108, 109, 0, 102, 102, 105, 0, 102, 102, 109, 0, 102, 105, 97, 0, 102, 105, 108, 0, 102, 105, 116, 0, 102, 106, 0, 102, 108, 114, 0, 102, 109, 112, 0, 102, 111, 100, 0, 102, 111, 110, 0, 102, 111, 114, 0, 102, 112, 101, 0, 102, 113, 115, 0, 102, 114, 99, 0, 102, 114, 112, 0, 102, 114, 114, 0, 102, 114, 115, 0, 102, 117, 98, 0, 102, 117, 100, 0, 102, 117, 101, 0, 102, 117, 102, 0, 102, 117, 104, 0, 102, 117, 113, 0, 102, 117, 114, 0, 102, 117, 118, 0, 102, 117, 121, 0, 102, 118, 114, 0, 103, 97, 97, 0, 103, 97, 102, 0, 103, 97, 103, 0, 103, 97, 104, 0, 103, 97, 106, 0, 103, 97, 110, 0, 103, 97, 119, 0, 103, 97, 121, 0, 103, 98, 97, 0, 103, 98, 102, 0, 103, 98, 109, 0, 103, 98, 121, 0, 103, 98, 122, 0, 103, 99, 114, 0, 103, 100, 101, 0, 103, 100, 110, 0, 103, 100, 114, 0, 103, 101, 98, 0, 103, 101, 106, 0, 103, 101, 108, 0, 103, 101, 122, 0, 103, 102, 107, 0, 103, 103, 110, 0, 103, 104, 115, 0, 103, 105, 108, 0, 103, 105, 109, 0, 103, 106, 107, 0, 103, 106, 110, 0, 103, 106, 117, 0, 103, 107, 110, 0, 103, 107, 112, 0, 103, 108, 107, 0, 103, 109, 109, 0, 103, 109, 118, 0, 103, 110, 100, 0, 103, 110, 103, 0, 103, 111, 100, 0, 103, 111, 102, 0, 103, 111, 105, 0, 103, 111, 109, 0, 103, 111, 110, 0, 103, 111, 114, 0, 103, 111, 115, 0, 103, 111, 116, 0, 103, 114, 98, 0, 103, 114, 99, 0, 103, 114, 99, 95, 76, 105, 110, 98, 0, 103, 114, 116, 0, 103, 114, 119, 0, 103, 115, 119, 0, 103, 117, 98, 0, 103, 117, 99, 0, 103, 117, 100, 0, 103, 117, 114, 0, 103, 117, 119, 0, 103, 117, 120, 0, 103, 117, 122, 0, 103, 118, 0, 103, 118, 102, 0, 103, 118, 114, 0, 103, 118, 115, 0, 103, 119, 99, 0, 103, 119, 105, 0, 103, 119, 116, 0, 103, 121, 105, 0, 104, 97, 95, 67, 77, 0, 104, 97, 95, 83, 68, 0, 104, 97, 107, 0, 104, 97, 122, 0, 104, 98, 98, 0, 104, 100, 121, 0, 104, 104, 121, 0, 104, 105, 97, 0, 104, 105, 102, 0, 104, 105, 103, 0, 104, 105, 104, 0, 104, 105, 108, 0, 104, 108, 97, 0, 104, 108, 117, 0, 104, 109, 100, 0, 104, 109, 116, 0, 104, 110, 100, 0, 104, 110, 101, 0, 104, 110, 106, 0, 104, 110, 110, 0, 104, 110, 111, 0, 104, 111, 99, 0, 104, 111, 116, 0, 104, 115, 98, 0, 104, 115, 110, 0, 104, 117, 105, 0, 105, 97, 110, 0, 105, 97, 114, 0, 105, 98, 97, 0, 105, 98, 98, 0, 105, 98, 121, 0, 105, 99, 97, 0, 105, 99, 104, 0, 105, 100, 105, 0, 105, 100, 117, 0, 105, 102, 101, 0, 105, 103, 98, 0, 105, 103, 101, 0, 105, 106, 106, 0, 105, 107, 107, 0, 105, 107, 116, 0, 105, 107, 119, 0, 105, 107, 120, 0, 105, 108, 111, 0, 105, 109, 111, 0, 105, 111, 117, 0, 105, 114, 105, 0, 105, 119, 109, 0, 105, 119, 115, 0, 105, 122, 104, 0, 105, 122, 105, 0, 106, 97, 98, 0, 106, 97, 109, 0, 106, 98, 111, 0, 106, 98, 117, 0, 106, 101, 110, 0, 106, 103, 107, 0, 106, 103, 111, 0, 106, 105, 98, 0, 106, 109, 99, 0, 106, 109, 108, 0, 106, 114, 97, 0, 106, 117, 116, 0, 106, 119, 0, 107, 97, 97, 0, 107, 97, 98, 0, 107, 97, 99, 0, 107, 97, 100, 0, 107, 97, 105, 0, 107, 97, 106, 0, 107, 97, 109, 0, 107, 97, 111, 0, 107, 98, 100, 0, 107, 98, 109, 0, 107, 98, 112, 0, 107, 98, 113, 0, 107, 98, 120, 0, 107, 98, 121, 0, 107, 99, 103, 0, 107, 99, 107, 0, 107, 99, 108, 0, 107, 99, 116, 0, 107, 100, 101, 0, 107, 100, 104, 0, 107, 100, 108, 0, 107, 100, 116, 0, 107, 101, 97, 0, 107, 101, 110, 0, 107, 101, 122, 0, 107, 102, 111, 0, 107, 102, 114, 0, 107, 102, 121, 0, 107, 103, 101, 0, 107, 103, 102, 0, 107, 103, 112, 0, 107, 104, 97, 0, 107, 104, 98, 0, 107, 104, 110, 0, 107, 104, 113, 0, 107, 104, 115, 0, 107, 104, 116, 0, 107, 104, 119, 0, 107, 104, 122, 0, 107, 105, 106, 0, 107, 105, 117, 0, 107, 105, 119, 0, 107, 106, 100, 0, 107, 106, 103, 0, 107, 106, 115, 0, 107, 106, 121, 0, 107, 107, 95, 65, 70, 0, 107, 107, 95, 65, 114, 97, 98, 0, 107, 107, 95, 67, 78, 0, 107, 107, 95, 73, 82, 0, 107, 107, 95, 77, 78, 0, 107, 107, 99, 0, 107, 107, 106, 0, 107, 108, 110, 0, 107, 108, 113, 0, 107, 108, 116, 0, 107, 108, 120, 0, 107, 109, 98, 0, 107, 109, 104, 0, 107, 109, 111, 0, 107, 109, 115, 0, 107, 109, 117, 0, 107, 109, 119, 0, 107, 110, 102, 0, 107, 110, 112, 0, 107, 111, 105, 0, 107, 111, 107, 0, 107, 111, 108, 0, 107, 111, 115, 0, 107, 111, 122, 0, 107, 112, 101, 0, 107, 112, 102, 0, 107, 112, 111, 0, 107, 112, 114, 0, 107, 112, 120, 0, 107, 113, 98, 0, 107, 113, 102, 0, 107, 113, 115, 0, 107, 113, 121, 0, 107, 114, 99, 0, 107, 114, 105, 0, 107, 114, 106, 0, 107, 114, 108, 0, 107, 114, 115, 0, 107, 114, 117, 0, 107, 115, 98, 0, 107, 115, 100, 0, 107, 115, 102, 0, 107, 115, 104, 0, 107, 115, 106, 0, 107, 115, 114, 0, 107, 116, 98, 0, 107, 116, 109, 0, 107, 116, 111, 0, 107, 116, 114, 0, 107, 117, 95, 65, 114, 97, 98, 0, 107, 117, 95, 76, 66, 0, 107, 117, 95, 89, 101, 122, 105, 0, 107, 117, 98, 0, 107, 117, 100, 0, 107, 117, 101, 0, 107, 117, 106, 0, 107, 117, 109, 0, 107, 117, 110, 0, 107, 117, 112, 0, 107, 117, 115, 0, 107, 118, 103, 0, 107, 118, 114, 0, 107, 118, 120, 0, 107, 119, 106, 0, 107, 119, 111, 0, 107, 119, 113, 0, 107, 120, 97, 0, 107, 120, 99, 0, 107, 120, 101, 0, 107, 120, 109, 0, 107, 120, 112, 0, 107, 120, 119, 0, 107, 120, 122, 0, 107, 121, 95, 65, 114, 97, 98, 0, 107, 121, 95, 67, 78, 0, 107, 121, 95, 76, 97, 116, 110, 0, 107, 121, 95, 84, 82, 0, 107, 121, 101, 0, 107, 121, 120, 0, 107, 122, 106, 0, 107, 122, 114, 0, 107, 122, 116, 0, 108, 97, 98, 0, 108, 97, 100, 0, 108, 97, 104, 0, 108, 97, 106, 0, 108, 97, 115, 0, 108, 98, 0, 108, 98, 101, 0, 108, 98, 117, 0, 108, 98, 119, 0, 108, 99, 109, 0, 108, 99, 112, 0, 108, 100, 98, 0, 108, 101, 100, 0, 108, 101, 101, 0, 108, 101, 109, 0, 108, 101, 112, 0, 108, 101, 113, 0, 108, 101, 117, 0, 108, 101, 122, 0, 108, 103, 103, 0, 108, 105, 97, 0, 108, 105, 100, 0, 108, 105, 102, 0, 108, 105, 102, 95, 76, 105, 109, 98, 0, 108, 105, 103, 0, 108, 105, 104, 0, 108, 105, 106, 0, 108, 105, 115, 0, 108, 106, 112, 0, 108, 107, 105, 0, 108, 107, 116, 0, 108, 108, 101, 0, 108, 108, 110, 0, 108, 109, 110, 0, 108, 109, 111, 0, 108, 109, 112, 0, 108, 110, 115, 0, 108, 110, 117, 0, 108, 111, 106, 0, 108, 111, 107, 0, 108, 111, 108, 0, 108, 111, 114, 0, 108, 111, 115, 0, 108, 111, 122, 0, 108, 114, 99, 0, 108, 116, 103, 0, 108, 117, 97, 0, 108, 117, 111, 0, 108, 117, 121, 0, 108, 117, 122, 0, 108, 118, 0, 108, 119, 108, 0, 108, 122, 104, 0, 108, 122, 122, 0, 109, 97, 100, 0, 109, 97, 102, 0, 109, 97, 103, 0, 109, 97, 105, 0, 109, 97, 107, 0, 109, 97, 110, 0, 109, 97, 110, 95, 71, 78, 0, 109, 97, 110, 95, 78, 107, 111, 111, 0, 109, 97, 115, 0, 109, 97, 119, 0, 109, 97, 122, 0, 109, 98, 104, 0, 109, 98, 111, 0, 109, 98, 113, 0, 109, 98, 117, 0, 109, 98, 119, 0, 109, 99, 105, 0, 109, 99, 112, 0, 109, 99, 113, 0, 109, 99, 114, 0, 109, 99, 117, 0, 109, 100, 97, 0, 109, 100, 101, 0, 109, 100, 102, 0, 109, 100, 104, 0, 109, 100, 106, 0, 109, 100, 114, 0, 109, 100, 120, 0, 109, 101, 100, 0, 109, 101, 101, 0, 109, 101, 107, 0, 109, 101, 110, 0, 109, 101, 114, 0, 109, 101, 116, 0, 109, 101, 117, 0, 109, 102, 97, 0, 109, 102, 101, 0, 109, 102, 110, 0, 109, 102, 111, 0, 109, 102, 113, 0, 109, 103, 104, 0, 109, 103, 108, 0, 109, 103, 111, 0, 109, 103, 112, 0, 109, 103, 121, 0, 109, 104, 105, 0, 109, 104, 108, 0, 109, 105, 102, 0, 109, 105, 110, 0, 109, 105, 115, 0, 109, 105, 115, 95, 77, 101, 100, 102, 0, 109, 105, 119, 0, 109, 107, 105, 0, 109, 107, 108, 0, 109, 107, 112, 0, 109, 107, 119, 0, 109, 108, 101, 0, 109, 108, 112, 0, 109, 108, 115, 0, 109, 109, 111, 0, 109, 109, 117, 0, 109, 109, 120, 0, 109, 110, 95, 67, 78, 0, 109, 110, 95, 77, 111, 110, 103, 0, 109, 110, 97, 0, 109, 110, 102, 0, 109, 110, 105, 0, 109, 110, 119, 0, 109, 111, 97, 0, 109, 111, 101, 0, 109, 111, 104, 0, 109, 111, 115, 0, 109, 111, 120, 0, 109, 112, 112, 0, 109, 112, 115, 0, 109, 112, 116, 0, 109, 112, 120, 0, 109, 113, 108, 0, 109, 114, 100, 0, 109, 114, 106, 0, 109, 114, 111, 0, 109, 115, 95, 67, 67, 0, 109, 115, 95, 73, 68, 0, 109, 116, 99, 0, 109, 116, 102, 0, 109, 116, 105, 0, 109, 116, 114, 0, 109, 117, 97, 0, 109, 117, 114, 0, 109, 117, 115, 0, 109, 118, 97, 0, 109, 118, 110, 0, 109, 118, 121, 0, 109, 119, 107, 0, 109, 119, 114, 0, 109, 119, 118, 0, 109, 119, 119, 0, 109, 120, 99, 0, 109, 120, 109, 0, 109, 121, 107, 0, 109, 121, 109, 0, 109, 121, 118, 0, 109, 121, 119, 0, 109, 121, 120, 0, 109, 121, 122, 0, 109, 122, 107, 0, 109, 122, 109, 0, 109, 122, 110, 0, 109, 122, 112, 0, 109, 122, 119, 0, 109, 122, 122, 0, 110, 97, 99, 0, 110, 97, 102, 0, 110, 97, 107, 0, 110, 97, 110, 0, 110, 97, 112, 0, 110, 97, 113, 0, 110, 97, 115, 0, 110, 99, 97, 0, 110, 99, 101, 0, 110, 99, 102, 0, 110, 99, 104, 0, 110, 99, 111, 0, 110, 99, 117, 0, 110, 100, 99, 0, 110, 100, 115, 0, 110, 101, 98, 0, 110, 101, 119, 0, 110, 101, 120, 0, 110, 102, 114, 0, 110, 103, 97, 0, 110, 103, 98, 0, 110, 103, 108, 0, 110, 104, 98, 0, 110, 104, 101, 0, 110, 104, 119, 0, 110, 105, 102, 0, 110, 105, 105, 0, 110, 105, 106, 0, 110, 105, 110, 0, 110, 105, 117, 0, 110, 105, 121, 0, 110, 105, 122, 0, 110, 106, 111, 0, 110, 107, 103, 0, 110, 107, 111, 0, 110, 108, 0, 110, 109, 103, 0, 110, 109, 122, 0, 110, 110, 102, 0, 110, 110, 104, 0, 110, 110, 107, 0, 110, 110, 109, 0, 110, 110, 112, 0, 110, 111, 100, 0, 110, 111, 101, 0, 110, 111, 110, 0, 110, 111, 112, 0, 110, 111, 117, 0, 110, 113, 111, 0, 110, 114, 98, 0, 110, 115, 107, 0, 110, 115, 110, 0, 110, 115, 111, 0, 110, 115, 115, 0, 110, 116, 109, 0, 110, 116, 114, 0, 110, 117, 105, 0, 110, 117, 112, 0, 110, 117, 115, 0, 110, 117, 118, 0, 110, 117, 120, 0, 110, 118, 0, 110, 119, 98, 0, 110, 120, 113, 0, 110, 120, 114, 0, 110, 121, 109, 0, 110, 121, 110, 0, 110, 122, 105, 0, 111, 103, 99, 0, 111, 107, 114, 0, 111, 107, 118, 0, 111, 110, 110, 0, 111, 110, 115, 0, 111, 112, 109, 0, 111, 114, 111, 0, 111, 114, 117, 0, 111, 115, 97, 0, 111, 116, 97, 0, 111, 116, 107, 0, 111, 122, 109, 0, 112, 97, 95, 65, 114, 97, 98, 0, 112, 97, 95, 80, 75, 0, 112, 97, 103, 0, 112, 97, 108, 0, 112, 97, 108, 95, 80, 104, 108, 112, 0, 112, 97, 109, 0, 112, 97, 112, 0, 112, 97, 117, 0, 112, 98, 105, 0, 112, 99, 100, 0, 112, 99, 109, 0, 112, 100, 99, 0, 112, 100, 116, 0, 112, 101, 100, 0, 112, 101, 120, 0, 112, 102, 108, 0, 112, 104, 108, 0, 112, 104, 110, 0, 112, 105, 108, 0, 112, 105, 112, 0, 112, 107, 97, 0, 112, 107, 111, 0, 112, 108, 97, 0, 112, 109, 115, 0, 112, 110, 103, 0, 112, 110, 110, 0, 112, 110, 116, 0, 112, 111, 110, 0, 112, 112, 97, 0, 112, 112, 111, 0, 112, 114, 97, 0, 112, 114, 100, 0, 112, 114, 103, 0, 112, 115, 115, 0, 112, 116, 112, 0, 112, 117, 117, 0, 112, 119, 97, 0, 113, 117, 0, 113, 117, 99, 0, 113, 117, 103, 0, 114, 97, 106, 0, 114, 97, 111, 0, 114, 99, 102, 0, 114, 101, 106, 0, 114, 101, 108, 0, 114, 101, 115, 0, 114, 103, 110, 0, 114, 104, 103, 0, 114, 105, 97, 0, 114, 105, 102, 0, 114, 105, 102, 95, 78, 76, 0, 114, 106, 115, 0, 114, 107, 116, 0, 114, 109, 102, 0, 114, 109, 111, 0, 114, 109, 116, 0, 114, 109, 117, 0, 114, 110, 97, 0, 114, 110, 103, 0, 114, 111, 98, 0, 114, 111, 102, 0, 114, 114, 111, 0, 114, 116, 109, 0, 114, 117, 101, 0, 114, 117, 103, 0, 114, 119, 107, 0, 114, 119, 111, 0, 114, 121, 117, 0, 115, 97, 102, 0, 115, 97, 104, 0, 115, 97, 113, 0, 115, 97, 115, 0, 115, 97, 116, 0, 115, 97, 118, 0, 115, 97, 122, 0, 115, 98, 97, 0, 115, 98, 101, 0, 115, 98, 112, 0, 115, 99, 0, 115, 99, 107, 0, 115, 99, 108, 0, 115, 99, 110, 0, 115, 99, 111, 0, 115, 99, 115, 0, 115, 100, 95, 68, 101, 118, 97, 0, 115, 100, 95, 75, 104, 111, 106, 0, 115, 100, 95, 83, 105, 110, 100, 0, 115, 100, 99, 0, 115, 100, 104, 0, 115, 101, 102, 0, 115, 101, 104, 0, 115, 101, 105, 0, 115, 101, 115, 0, 115, 103, 97, 0, 115, 103, 115, 0, 115, 103, 119, 0, 115, 103, 122, 0, 115, 104, 105, 0, 115, 104, 107, 0, 115, 104, 110, 0, 115, 105, 100, 0, 115, 105, 103, 0, 115, 105, 108, 0, 115, 105, 109, 0, 115, 106, 114, 0, 115, 107, 99, 0, 115, 107, 114, 0, 115, 107, 115, 0, 115, 108, 0, 115, 108, 100, 0, 115, 108, 105, 0, 115, 108, 108, 0, 115, 108, 121, 0, 115, 109, 106, 0, 115, 109, 110, 0, 115, 109, 112, 0, 115, 109, 113, 0, 115, 109, 115, 0, 115, 110, 99, 0, 115, 110, 107, 0, 115, 110, 112, 0, 115, 110, 120, 0, 115, 110, 121, 0, 115, 111, 103, 0, 115, 111, 107, 0, 115, 111, 113, 0, 115, 111, 117, 0, 115, 111, 121, 0, 115, 112, 100, 0, 115, 112, 108, 0, 115, 112, 115, 0, 115, 114, 95, 77, 69, 0, 115, 114, 95, 82, 79, 0, 115, 114, 95, 82, 85, 0, 115, 114, 95, 84, 82, 0, 115, 114, 98, 0, 115, 114, 110, 0, 115, 114, 114, 0, 115, 114, 120, 0, 115, 115, 100, 0, 115, 115, 103, 0, 115, 115, 121, 0, 115, 116, 107, 0, 115, 116, 113, 0, 115, 117, 97, 0, 115, 117, 101, 0, 115, 117, 107, 0, 115, 117, 114, 0, 115, 117, 115, 0, 115, 118, 0, 115, 119, 98, 0, 115, 119, 99, 0, 115, 119, 103, 0, 115, 119, 112, 0, 115, 119, 118, 0, 115, 120, 110, 0, 115, 120, 119, 0, 115, 121, 108, 0, 115, 121, 114, 0, 115, 122, 108, 0, 116, 97, 106, 0, 116, 97, 110, 0, 116, 97, 113, 0, 116, 98, 99, 0, 116, 98, 100, 0, 116, 98, 102, 0, 116, 98, 103, 0, 116, 98, 111, 0, 116, 98, 119, 0, 116, 98, 122, 0, 116, 99, 105, 0, 116, 99, 121, 0, 116, 100, 100, 0, 116, 100, 103, 0, 116, 100, 104, 0, 116, 100, 117, 0, 116, 101, 100, 0, 116, 101, 109, 0, 116, 101, 111, 0, 116, 101, 116, 0, 116, 102, 105, 0, 116, 103, 95, 65, 114, 97, 98, 0, 116, 103, 95, 80, 75, 0, 116, 103, 99, 0, 116, 103, 111, 0, 116, 103, 117, 0, 116, 104, 108, 0, 116, 104, 113, 0, 116, 104, 114, 0, 116, 105, 102, 0, 116, 105, 103, 0, 116, 105, 107, 0, 116, 105, 109, 0, 116, 105, 111, 0, 116, 105, 118, 0, 116, 107, 108, 0, 116, 107, 114, 0, 116, 107, 116, 0, 116, 108, 0, 116, 108, 102, 0, 116, 108, 120, 0, 116, 108, 121, 0, 116, 109, 104, 0, 116, 109, 121, 0, 116, 110, 104, 0, 116, 111, 102, 0, 116, 111, 103, 0, 116, 111, 113, 0, 116, 112, 105, 0, 116, 112, 109, 0, 116, 112, 122, 0, 116, 113, 111, 0, 116, 114, 117, 0, 116, 114, 118, 0, 116, 114, 119, 0, 116, 115, 100, 0, 116, 115, 102, 0, 116, 115, 103, 0, 116, 115, 106, 0, 116, 115, 119, 0, 116, 116, 100, 0, 116, 116, 101, 0, 116, 116, 106, 0, 116, 116, 114, 0, 116, 116, 115, 0, 116, 116, 116, 0, 116, 117, 104, 0, 116, 117, 108, 0, 116, 117, 109, 0, 116, 117, 113, 0, 116, 118, 100, 0, 116, 118, 108, 0, 116, 118, 117, 0, 116, 119, 104, 0, 116, 119, 113, 0, 116, 120, 103, 0, 116, 121, 97, 0, 116, 121, 118, 0, 116, 122, 109, 0, 117, 98, 117, 0, 117, 100, 109, 0, 117, 103, 95, 67, 121, 114, 108, 0, 117, 103, 95, 75, 90, 0, 117, 103, 95, 77, 78, 0, 117, 103, 97, 0, 117, 108, 105, 0, 117, 109, 98, 0, 117, 110, 100, 95, 48, 48, 50, 0, 117, 110, 100, 95, 48, 48, 51, 0, 117, 110, 100, 95, 48, 48, 53, 0, 117, 110, 100, 95, 48, 48, 57, 0, 117, 110, 100, 95, 48, 49, 49, 0, 117, 110, 100, 95, 48, 49, 51, 0, 117, 110, 100, 95, 48, 49, 52, 0, 117, 110, 100, 95, 48, 49, 53, 0, 117, 110, 100, 95, 48, 49, 55, 0, 117, 110, 100, 95, 48, 49, 56, 0, 117, 110, 100, 95, 48, 49, 57, 0, 117, 110, 100, 95, 48, 50, 49, 0, 117, 110, 100, 95, 48, 50, 57, 0, 117, 110, 100, 95, 48, 51, 48, 0, 117, 110, 100, 95, 48, 51, 52, 0, 117, 110, 100, 95, 48, 51, 53, 0, 117, 110, 100, 95, 48, 51, 57, 0, 117, 110, 100, 95, 48, 53, 51, 0, 117, 110, 100, 95, 48, 53, 52, 0, 117, 110, 100, 95, 48, 53, 55, 0, 117, 110, 100, 95, 48, 54, 49, 0, 117, 110, 100, 95, 49, 52, 50, 0, 117, 110, 100, 95, 49, 52, 51, 0, 117, 110, 100, 95, 49, 52, 53, 0, 117, 110, 100, 95, 49, 53, 48, 0, 117, 110, 100, 95, 49, 53, 49, 0, 117, 110, 100, 95, 49, 53, 52, 0, 117, 110, 100, 95, 49, 53, 53, 0, 117, 110, 100, 95, 50, 48, 50, 0, 117, 110, 100, 95, 52, 49, 57, 0, 117, 110, 100, 95, 65, 68, 0, 117, 110, 100, 95, 65, 69, 0, 117, 110, 100, 95, 65, 70, 0, 117, 110, 100, 95, 65, 76, 0, 117, 110, 100, 95, 65, 77, 0, 117, 110, 100, 95, 65, 79, 0, 117, 110, 100, 95, 65, 81, 0, 117, 110, 100, 95, 65, 82, 0, 117, 110, 100, 95, 65, 83, 0, 117, 110, 100, 95, 65, 84, 0, 117, 110, 100, 95, 65, 87, 0, 117, 110, 100, 95, 65, 88, 0, 117, 110, 100, 95, 65, 90, 0, 117, 110, 100, 95, 65, 100, 108, 109, 0, 117, 110, 100, 95, 65, 103, 104, 98, 0, 117, 110, 100, 95, 65, 104, 111, 109, 0, 117, 110, 100, 95, 65, 114, 97, 98, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 67, 67, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 67, 78, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 71, 66, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 73, 68, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 73, 78, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 75, 72, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 77, 77, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 77, 78, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 77, 85, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 78, 71, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 80, 75, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 84, 71, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 84, 72, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 84, 74, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 84, 82, 0, 117, 110, 100, 95, 65, 114, 97, 98, 95, 89, 84, 0, 117, 110, 100, 95, 65, 114, 109, 105, 0, 117, 110, 100, 95, 65, 114, 109, 110, 0, 117, 110, 100, 95, 65, 118, 115, 116, 0, 117, 110, 100, 95, 66, 65, 0, 117, 110, 100, 95, 66, 68, 0, 117, 110, 100, 95, 66, 69, 0, 117, 110, 100, 95, 66, 70, 0, 117, 110, 100, 95, 66, 71, 0, 117, 110, 100, 95, 66, 72, 0, 117, 110, 100, 95, 66, 73, 0, 117, 110, 100, 95, 66, 74, 0, 117, 110, 100, 95, 66, 76, 0, 117, 110, 100, 95, 66, 78, 0, 117, 110, 100, 95, 66, 79, 0, 117, 110, 100, 95, 66, 81, 0, 117, 110, 100, 95, 66, 82, 0, 117, 110, 100, 95, 66, 84, 0, 117, 110, 100, 95, 66, 86, 0, 117, 110, 100, 95, 66, 89, 0, 117, 110, 100, 95, 66, 97, 108, 105, 0, 117, 110, 100, 95, 66, 97, 109, 117, 0, 117, 110, 100, 95, 66, 97, 115, 115, 0, 117, 110, 100, 95, 66, 97, 116, 107, 0, 117, 110, 100, 95, 66, 101, 110, 103, 0, 117, 110, 100, 95, 66, 104, 107, 115, 0, 117, 110, 100, 95, 66, 111, 112, 111, 0, 117, 110, 100, 95, 66, 114, 97, 104, 0, 117, 110, 100, 95, 66, 114, 97, 105, 0, 117, 110, 100, 95, 66, 117, 103, 105, 0, 117, 110, 100, 95, 66, 117, 104, 100, 0, 117, 110, 100, 95, 67, 68, 0, 117, 110, 100, 95, 67, 70, 0, 117, 110, 100, 95, 67, 71, 0, 117, 110, 100, 95, 67, 72, 0, 117, 110, 100, 95, 67, 73, 0, 117, 110, 100, 95, 67, 76, 0, 117, 110, 100, 95, 67, 77, 0, 117, 110, 100, 95, 67, 78, 0, 117, 110, 100, 95, 67, 79, 0, 117, 110, 100, 95, 67, 80, 0, 117, 110, 100, 95, 67, 82, 0, 117, 110, 100, 95, 67, 85, 0, 117, 110, 100, 95, 67, 86, 0, 117, 110, 100, 95, 67, 87, 0, 117, 110, 100, 95, 67, 89, 0, 117, 110, 100, 95, 67, 90, 0, 117, 110, 100, 95, 67, 97, 107, 109, 0, 117, 110, 100, 95, 67, 97, 110, 115, 0, 117, 110, 100, 95, 67, 97, 114, 105, 0, 117, 110, 100, 95, 67, 104, 97, 109, 0, 117, 110, 100, 95, 67, 104, 101, 114, 0, 117, 110, 100, 95, 67, 104, 114, 115, 0, 117, 110, 100, 95, 67, 111, 112, 116, 0, 117, 110, 100, 95, 67, 112, 114, 116, 0, 117, 110, 100, 95, 67, 121, 114, 108, 0, 117, 110, 100, 95, 67, 121, 114, 108, 95, 65, 76, 0, 117, 110, 100, 95, 67, 121, 114, 108, 95, 66, 65, 0, 117, 110, 100, 95, 67, 121, 114, 108, 95, 71, 69, 0, 117, 110, 100, 95, 67, 121, 114, 108, 95, 71, 82, 0, 117, 110, 100, 95, 67, 121, 114, 108, 95, 77, 68, 0, 117, 110, 100, 95, 67, 121, 114, 108, 95, 82, 79, 0, 117, 110, 100, 95, 67, 121, 114, 108, 95, 83, 75, 0, 117, 110, 100, 95, 67, 121, 114, 108, 95, 84, 82, 0, 117, 110, 100, 95, 67, 121, 114, 108, 95, 88, 75, 0, 117, 110, 100, 95, 68, 69, 0, 117, 110, 100, 95, 68, 74, 0, 117, 110, 100, 95, 68, 75, 0, 117, 110, 100, 95, 68, 79, 0, 117, 110, 100, 95, 68, 90, 0, 117, 110, 100, 95, 68, 101, 118, 97, 0, 117, 110, 100, 95, 68, 101, 118, 97, 95, 66, 84, 0, 117, 110, 100, 95, 68, 101, 118, 97, 95, 70, 74, 0, 117, 110, 100, 95, 68, 101, 118, 97, 95, 77, 85, 0, 117, 110, 100, 95, 68, 101, 118, 97, 95, 80, 75, 0, 117, 110, 100, 95, 68, 105, 97, 107, 0, 117, 110, 100, 95, 68, 111, 103, 114, 0, 117, 110, 100, 95, 68, 117, 112, 108, 0, 117, 110, 100, 95, 69, 65, 0, 117, 110, 100, 95, 69, 67, 0, 117, 110, 100, 95, 69, 69, 0, 117, 110, 100, 95, 69, 71, 0, 117, 110, 100, 95, 69, 72, 0, 117, 110, 100, 95, 69, 82, 0, 117, 110, 100, 95, 69, 83, 0, 117, 110, 100, 95, 69, 84, 0, 117, 110, 100, 95, 69, 85, 0, 117, 110, 100, 95, 69, 90, 0, 117, 110, 100, 95, 69, 103, 121, 112, 0, 117, 110, 100, 95, 69, 108, 98, 97, 0, 117, 110, 100, 95, 69, 108, 121, 109, 0, 117, 110, 100, 95, 69, 116, 104, 105, 0, 117, 110, 100, 95, 70, 73, 0, 117, 110, 100, 95, 70, 79, 0, 117, 110, 100, 95, 70, 82, 0, 117, 110, 100, 95, 71, 65, 0, 117, 110, 100, 95, 71, 69, 0, 117, 110, 100, 95, 71, 70, 0, 117, 110, 100, 95, 71, 72, 0, 117, 110, 100, 95, 71, 76, 0, 117, 110, 100, 95, 71, 78, 0, 117, 110, 100, 95, 71, 80, 0, 117, 110, 100, 95, 71, 81, 0, 117, 110, 100, 95, 71, 82, 0, 117, 110, 100, 95, 71, 83, 0, 117, 110, 100, 95, 71, 84, 0, 117, 110, 100, 95, 71, 87, 0, 117, 110, 100, 95, 71, 101, 111, 114, 0, 117, 110, 100, 95, 71, 108, 97, 103, 0, 117, 110, 100, 95, 71, 111, 110, 103, 0, 117, 110, 100, 95, 71, 111, 110, 109, 0, 117, 110, 100, 95, 71, 111, 116, 104, 0, 117, 110, 100, 95, 71, 114, 97, 110, 0, 117, 110, 100, 95, 71, 114, 101, 107, 0, 117, 110, 100, 95, 71, 114, 101, 107, 95, 84, 82, 0, 117, 110, 100, 95, 71, 117, 106, 114, 0, 117, 110, 100, 95, 71, 117, 114, 117, 0, 117, 110, 100, 95, 72, 75, 0, 117, 110, 100, 95, 72, 77, 0, 117, 110, 100, 95, 72, 78, 0, 117, 110, 100, 95, 72, 82, 0, 117, 110, 100, 95, 72, 84, 0, 117, 110, 100, 95, 72, 85, 0, 117, 110, 100, 95, 72, 97, 110, 98, 0, 117, 110, 100, 95, 72, 97, 110, 103, 0, 117, 110, 100, 95, 72, 97, 110, 105, 0, 117, 110, 100, 95, 72, 97, 110, 111, 0, 117, 110, 100, 95, 72, 97, 110, 115, 0, 117, 110, 100, 95, 72, 97, 110, 116, 0, 117, 110, 100, 95, 72, 97, 116, 114, 0, 117, 110, 100, 95, 72, 101, 98, 114, 0, 117, 110, 100, 95, 72, 101, 98, 114, 95, 67, 65, 0, 117, 110, 100, 95, 72, 101, 98, 114, 95, 71, 66, 0, 117, 110, 100, 95, 72, 101, 98, 114, 95, 83, 69, 0, 117, 110, 100, 95, 72, 101, 98, 114, 95, 85, 65, 0, 117, 110, 100, 95, 72, 101, 98, 114, 95, 85, 83, 0, 117, 110, 100, 95, 72, 105, 114, 97, 0, 117, 110, 100, 95, 72, 108, 117, 119, 0, 117, 110, 100, 95, 72, 109, 110, 103, 0, 117, 110, 100, 95, 72, 109, 110, 112, 0, 117, 110, 100, 95, 72, 117, 110, 103, 0, 117, 110, 100, 95, 73, 67, 0, 117, 110, 100, 95, 73, 68, 0, 117, 110, 100, 95, 73, 76, 0, 117, 110, 100, 95, 73, 78, 0, 117, 110, 100, 95, 73, 81, 0, 117, 110, 100, 95, 73, 82, 0, 117, 110, 100, 95, 73, 83, 0, 117, 110, 100, 95, 73, 84, 0, 117, 110, 100, 95, 73, 116, 97, 108, 0, 117, 110, 100, 95, 74, 79, 0, 117, 110, 100, 95, 74, 80, 0, 117, 110, 100, 95, 74, 97, 109, 111, 0, 117, 110, 100, 95, 74, 97, 118, 97, 0, 117, 110, 100, 95, 74, 112, 97, 110, 0, 117, 110, 100, 95, 75, 69, 0, 117, 110, 100, 95, 75, 71, 0, 117, 110, 100, 95, 75, 72, 0, 117, 110, 100, 95, 75, 77, 0, 117, 110, 100, 95, 75, 80, 0, 117, 110, 100, 95, 75, 82, 0, 117, 110, 100, 95, 75, 87, 0, 117, 110, 100, 95, 75, 90, 0, 117, 110, 100, 95, 75, 97, 108, 105, 0, 117, 110, 100, 95, 75, 97, 110, 97, 0, 117, 110, 100, 95, 75, 104, 97, 114, 0, 117, 110, 100, 95, 75, 104, 109, 114, 0, 117, 110, 100, 95, 75, 104, 111, 106, 0, 117, 110, 100, 95, 75, 105, 116, 115, 0, 117, 110, 100, 95, 75, 110, 100, 97, 0, 117, 110, 100, 95, 75, 111, 114, 101, 0, 117, 110, 100, 95, 75, 116, 104, 105, 0, 117, 110, 100, 95, 76, 65, 0, 117, 110, 100, 95, 76, 66, 0, 117, 110, 100, 95, 76, 73, 0, 117, 110, 100, 95, 76, 75, 0, 117, 110, 100, 95, 76, 83, 0, 117, 110, 100, 95, 76, 84, 0, 117, 110, 100, 95, 76, 85, 0, 117, 110, 100, 95, 76, 86, 0, 117, 110, 100, 95, 76, 89, 0, 117, 110, 100, 95, 76, 97, 110, 97, 0, 117, 110, 100, 95, 76, 97, 111, 111, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 65, 70, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 65, 77, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 67, 78, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 67, 89, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 68, 90, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 69, 84, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 71, 69, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 73, 82, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 75, 77, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 77, 65, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 77, 75, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 77, 77, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 77, 79, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 77, 82, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 82, 85, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 83, 89, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 84, 78, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 84, 87, 0, 117, 110, 100, 95, 76, 97, 116, 110, 95, 85, 65, 0, 117, 110, 100, 95, 76, 101, 112, 99, 0, 117, 110, 100, 95, 76, 105, 109, 98, 0, 117, 110, 100, 95, 76, 105, 110, 97, 0, 117, 110, 100, 95, 76, 105, 110, 98, 0, 117, 110, 100, 95, 76, 105, 115, 117, 0, 117, 110, 100, 95, 76, 121, 99, 105, 0, 117, 110, 100, 95, 76, 121, 100, 105, 0, 117, 110, 100, 95, 77, 65, 0, 117, 110, 100, 95, 77, 67, 0, 117, 110, 100, 95, 77, 68, 0, 117, 110, 100, 95, 77, 69, 0, 117, 110, 100, 95, 77, 70, 0, 117, 110, 100, 95, 77, 71, 0, 117, 110, 100, 95, 77, 75, 0, 117, 110, 100, 95, 77, 76, 0, 117, 110, 100, 95, 77, 77, 0, 117, 110, 100, 95, 77, 78, 0, 117, 110, 100, 95, 77, 79, 0, 117, 110, 100, 95, 77, 81, 0, 117, 110, 100, 95, 77, 82, 0, 117, 110, 100, 95, 77, 84, 0, 117, 110, 100, 95, 77, 85, 0, 117, 110, 100, 95, 77, 86, 0, 117, 110, 100, 95, 77, 88, 0, 117, 110, 100, 95, 77, 89, 0, 117, 110, 100, 95, 77, 90, 0, 117, 110, 100, 95, 77, 97, 104, 106, 0, 117, 110, 100, 95, 77, 97, 107, 97, 0, 117, 110, 100, 95, 77, 97, 110, 100, 0, 117, 110, 100, 95, 77, 97, 110, 105, 0, 117, 110, 100, 95, 77, 97, 114, 99, 0, 117, 110, 100, 95, 77, 101, 100, 102, 0, 117, 110, 100, 95, 77, 101, 110, 100, 0, 117, 110, 100, 95, 77, 101, 114, 99, 0, 117, 110, 100, 95, 77, 101, 114, 111, 0, 117, 110, 100, 95, 77, 108, 121, 109, 0, 117, 110, 100, 95, 77, 111, 100, 105, 0, 117, 110, 100, 95, 77, 111, 110, 103, 0, 117, 110, 100, 95, 77, 114, 111, 111, 0, 117, 110, 100, 95, 77, 116, 101, 105, 0, 117, 110, 100, 95, 77, 117, 108, 116, 0, 117, 110, 100, 95, 77, 121, 109, 114, 0, 117, 110, 100, 95, 77, 121, 109, 114, 95, 73, 78, 0, 117, 110, 100, 95, 77, 121, 109, 114, 95, 84, 72, 0, 117, 110, 100, 95, 78, 65, 0, 117, 110, 100, 95, 78, 67, 0, 117, 110, 100, 95, 78, 69, 0, 117, 110, 100, 95, 78, 73, 0, 117, 110, 100, 95, 78, 76, 0, 117, 110, 100, 95, 78, 79, 0, 117, 110, 100, 95, 78, 80, 0, 117, 110, 100, 95, 78, 97, 110, 100, 0, 117, 110, 100, 95, 78, 97, 114, 98, 0, 117, 110, 100, 95, 78, 98, 97, 116, 0, 117, 110, 100, 95, 78, 101, 119, 97, 0, 117, 110, 100, 95, 78, 107, 111, 111, 0, 117, 110, 100, 95, 78, 115, 104, 117, 0, 117, 110, 100, 95, 79, 77, 0, 117, 110, 100, 95, 79, 103, 97, 109, 0, 117, 110, 100, 95, 79, 108, 99, 107, 0, 117, 110, 100, 95, 79, 114, 107, 104, 0, 117, 110, 100, 95, 79, 114, 121, 97, 0, 117, 110, 100, 95, 79, 115, 103, 101, 0, 117, 110, 100, 95, 79, 115, 109, 97, 0, 117, 110, 100, 95, 80, 65, 0, 117, 110, 100, 95, 80, 69, 0, 117, 110, 100, 95, 80, 70, 0, 117, 110, 100, 95, 80, 71, 0, 117, 110, 100, 95, 80, 72, 0, 117, 110, 100, 95, 80, 75, 0, 117, 110, 100, 95, 80, 76, 0, 117, 110, 100, 95, 80, 77, 0, 117, 110, 100, 95, 80, 82, 0, 117, 110, 100, 95, 80, 83, 0, 117, 110, 100, 95, 80, 84, 0, 117, 110, 100, 95, 80, 87, 0, 117, 110, 100, 95, 80, 89, 0, 117, 110, 100, 95, 80, 97, 108, 109, 0, 117, 110, 100, 95, 80, 97, 117, 99, 0, 117, 110, 100, 95, 80, 101, 114, 109, 0, 117, 110, 100, 95, 80, 104, 97, 103, 0, 117, 110, 100, 95, 80, 104, 108, 105, 0, 117, 110, 100, 95, 80, 104, 108, 112, 0, 117, 110, 100, 95, 80, 104, 110, 120, 0, 117, 110, 100, 95, 80, 108, 114, 100, 0, 117, 110, 100, 95, 80, 114, 116, 105, 0, 117, 110, 100, 95, 81, 65, 0, 117, 110, 100, 95, 81, 79, 0, 117, 110, 100, 95, 82, 69, 0, 117, 110, 100, 95, 82, 79, 0, 117, 110, 100, 95, 82, 83, 0, 117, 110, 100, 95, 82, 85, 0, 117, 110, 100, 95, 82, 87, 0, 117, 110, 100, 95, 82, 106, 110, 103, 0, 117, 110, 100, 95, 82, 111, 104, 103, 0, 117, 110, 100, 95, 82, 117, 110, 114, 0, 117, 110, 100, 95, 83, 65, 0, 117, 110, 100, 95, 83, 67, 0, 117, 110, 100, 95, 83, 68, 0, 117, 110, 100, 95, 83, 69, 0, 117, 110, 100, 95, 83, 73, 0, 117, 110, 100, 95, 83, 74, 0, 117, 110, 100, 95, 83, 75, 0, 117, 110, 100, 95, 83, 77, 0, 117, 110, 100, 95, 83, 78, 0, 117, 110, 100, 95, 83, 79, 0, 117, 110, 100, 95, 83, 82, 0, 117, 110, 100, 95, 83, 84, 0, 117, 110, 100, 95, 83, 86, 0, 117, 110, 100, 95, 83, 89, 0, 117, 110, 100, 95, 83, 97, 109, 114, 0, 117, 110, 100, 95, 83, 97, 114, 98, 0, 117, 110, 100, 95, 83, 97, 117, 114, 0, 117, 110, 100, 95, 83, 103, 110, 119, 0, 117, 110, 100, 95, 83, 104, 97, 119, 0, 117, 110, 100, 95, 83, 104, 114, 100, 0, 117, 110, 100, 95, 83, 105, 100, 100, 0, 117, 110, 100, 95, 83, 105, 110, 100, 0, 117, 110, 100, 95, 83, 105, 110, 104, 0, 117, 110, 100, 95, 83, 111, 103, 100, 0, 117, 110, 100, 95, 83, 111, 103, 111, 0, 117, 110, 100, 95, 83, 111, 114, 97, 0, 117, 110, 100, 95, 83, 111, 121, 111, 0, 117, 110, 100, 95, 83, 117, 110, 100, 0, 117, 110, 100, 95, 83, 121, 108, 111, 0, 117, 110, 100, 95, 83, 121, 114, 99, 0, 117, 110, 100, 95, 84, 68, 0, 117, 110, 100, 95, 84, 70, 0, 117, 110, 100, 95, 84, 71, 0, 117, 110, 100, 95, 84, 72, 0, 117, 110, 100, 95, 84, 74, 0, 117, 110, 100, 95, 84, 75, 0, 117, 110, 100, 95, 84, 76, 0, 117, 110, 100, 95, 84, 77, 0, 117, 110, 100, 95, 84, 78, 0, 117, 110, 100, 95, 84, 79, 0, 117, 110, 100, 95, 84, 82, 0, 117, 110, 100, 95, 84, 86, 0, 117, 110, 100, 95, 84, 87, 0, 117, 110, 100, 95, 84, 90, 0, 117, 110, 100, 95, 84, 97, 103, 98, 0, 117, 110, 100, 95, 84, 97, 107, 114, 0, 117, 110, 100, 95, 84, 97, 108, 101, 0, 117, 110, 100, 95, 84, 97, 108, 117, 0, 117, 110, 100, 95, 84, 97, 109, 108, 0, 117, 110, 100, 95, 84, 97, 110, 103, 0, 117, 110, 100, 95, 84, 97, 118, 116, 0, 117, 110, 100, 95, 84, 101, 108, 117, 0, 117, 110, 100, 95, 84, 102, 110, 103, 0, 117, 110, 100, 95, 84, 103, 108, 103, 0, 117, 110, 100, 95, 84, 104, 97, 97, 0, 117, 110, 100, 95, 84, 104, 97, 105, 0, 117, 110, 100, 95, 84, 104, 97, 105, 95, 67, 78, 0, 117, 110, 100, 95, 84, 104, 97, 105, 95, 75, 72, 0, 117, 110, 100, 95, 84, 104, 97, 105, 95, 76, 65, 0, 117, 110, 100, 95, 84, 105, 98, 116, 0, 117, 110, 100, 95, 84, 105, 114, 104, 0, 117, 110, 100, 95, 85, 65, 0, 117, 110, 100, 95, 85, 71, 0, 117, 110, 100, 95, 85, 89, 0, 117, 110, 100, 95, 85, 90, 0, 117, 110, 100, 95, 85, 103, 97, 114, 0, 117, 110, 100, 95, 86, 65, 0, 117, 110, 100, 95, 86, 69, 0, 117, 110, 100, 95, 86, 78, 0, 117, 110, 100, 95, 86, 85, 0, 117, 110, 100, 95, 86, 97, 105, 105, 0, 117, 110, 100, 95, 87, 70, 0, 117, 110, 100, 95, 87, 83, 0, 117, 110, 100, 95, 87, 97, 114, 97, 0, 117, 110, 100, 95, 87, 99, 104, 111, 0, 117, 110, 100, 95, 88, 75, 0, 117, 110, 100, 95, 88, 112, 101, 111, 0, 117, 110, 100, 95, 88, 115, 117, 120, 0, 117, 110, 100, 95, 89, 69, 0, 117, 110, 100, 95, 89, 84, 0, 117, 110, 100, 95, 89, 101, 122, 105, 0, 117, 110, 100, 95, 89, 105, 105, 105, 0, 117, 110, 100, 95, 90, 87, 0, 117, 110, 100, 95, 90, 97, 110, 98, 0, 117, 110, 114, 95, 68, 101, 118, 97, 0, 117, 110, 114, 95, 78, 80, 0, 117, 110, 120, 0, 117, 111, 107, 0, 117, 114, 105, 0, 117, 114, 116, 0, 117, 114, 119, 0, 117, 115, 97, 0, 117, 116, 114, 0, 117, 118, 104, 0, 117, 118, 108, 0, 117, 122, 95, 65, 70, 0, 117, 122, 95, 65, 114, 97, 98, 0, 117, 122, 95, 67, 78, 0, 118, 97, 103, 0, 118, 97, 105, 0, 118, 97, 110, 0, 118, 101, 0, 118, 101, 99, 0, 118, 101, 112, 0, 118, 105, 0, 118, 105, 99, 0, 118, 105, 118, 0, 118, 108, 115, 0, 118, 109, 102, 0, 118, 109, 119, 0, 118, 111, 0, 118, 111, 116, 0, 118, 114, 111, 0, 118, 117, 110, 0, 118, 117, 116, 0, 119, 97, 101, 0, 119, 97, 106, 0, 119, 97, 108, 0, 119, 97, 110, 0, 119, 97, 114, 0, 119, 98, 112, 0, 119, 98, 113, 0, 119, 98, 114, 0, 119, 99, 105, 0, 119, 101, 114, 0, 119, 103, 105, 0, 119, 104, 103, 0, 119, 105, 98, 0, 119, 105, 117, 0, 119, 105, 118, 0, 119, 106, 97, 0, 119, 106, 105, 0, 119, 108, 115, 0, 119, 109, 111, 0, 119, 110, 99, 0, 119, 110, 105, 0, 119, 110, 117, 0, 119, 111, 98, 0, 119, 111, 115, 0, 119, 114, 115, 0, 119, 115, 103, 0, 119, 115, 107, 0, 119, 116, 109, 0, 119, 117, 117, 0, 119, 117, 118, 0, 119, 119, 97, 0, 120, 97, 118, 0, 120, 98, 105, 0, 120, 99, 111, 0, 120, 99, 114, 0, 120, 101, 115, 0, 120, 108, 97, 0, 120, 108, 99, 0, 120, 108, 100, 0, 120, 109, 102, 0, 120, 109, 110, 0, 120, 109, 114, 0, 120, 110, 97, 0, 120, 110, 114, 0, 120, 111, 103, 0, 120, 111, 110, 0, 120, 112, 114, 0, 120, 114, 98, 0, 120, 115, 97, 0, 120, 115, 105, 0, 120, 115, 109, 0, 120, 115, 114, 0, 120, 119, 101, 0, 121, 97, 109, 0, 121, 97, 111, 0, 121, 97, 112, 0, 121, 97, 115, 0, 121, 97, 116, 0, 121, 97, 118, 0, 121, 97, 121, 0, 121, 97, 122, 0, 121, 98, 97, 0, 121, 98, 98, 0, 121, 98, 121, 0, 121, 101, 114, 0, 121, 103, 114, 0, 121, 103, 119, 0, 121, 107, 111, 0, 121, 108, 101, 0, 121, 108, 103, 0, 121, 108, 108, 0, 121, 109, 108, 0, 121, 111, 110, 0, 121, 114, 98, 0, 121, 114, 101, 0, 121, 115, 115, 0, 121, 117, 97, 0, 121, 117, 101, 0, 121, 117, 101, 95, 67, 78, 0, 121, 117, 101, 95, 72, 97, 110, 115, 0, 121, 117, 106, 0, 121, 117, 116, 0, 121, 117, 119, 0, 122, 97, 103, 0, 122, 100, 106, 0, 122, 101, 97, 0, 122, 103, 104, 0, 122, 104, 95, 65, 85, 0, 122, 104, 95, 66, 78, 0, 122, 104, 95, 66, 111, 112, 111, 0, 122, 104, 95, 71, 66, 0, 122, 104, 95, 71, 70, 0, 122, 104, 95, 72, 75, 0, 122, 104, 95, 72, 97, 110, 98, 0, 122, 104, 95, 72, 97, 110, 116, 0, 122, 104, 95, 73, 68, 0, 122, 104, 95, 77, 79, 0, 122, 104, 95, 77, 89, 0, 122, 104, 95, 80, 65, 0, 122, 104, 95, 80, 70, 0, 122, 104, 95, 80, 72, 0, 122, 104, 95, 83, 82, 0, 122, 104, 95, 84, 72, 0, 122, 104, 95, 84, 87, 0, 122, 104, 95, 85, 83, 0, 122, 104, 95, 86, 78, 0, 122, 104, 120, 0, 122, 105, 97, 0, 122, 107, 116, 0, 122, 108, 109, 0, 122, 109, 105, 0, 122, 110, 101, 0, 122, 117, 0, 122, 122, 97, 0, 170, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 114, 0, 117, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 69, 0, 71, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 84, 0, 87, 0, 0, 0, 97, 0, 109, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 69, 0, 84, 0, 0, 0, 98, 0, 110, 0, 95, 0, 66, 0, 101, 0, 110, 0, 103, 0, 95, 0, 66, 0, 68, 0, 0, 0, 100, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 100, 0, 118, 0, 95, 0, 84, 0, 104, 0, 97, 0, 97, 0, 95, 0, 77, 0, 86, 0, 0, 0, 101, 0, 108, 0, 95, 0, 71, 0, 114, 0, 101, 0, 107, 0, 95, 0, 71, 0, 82, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 104, 0, 101, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 95, 0, 73, 0, 76, 0, 0, 0, 104, 0, 121, 0, 95, 0, 65, 0, 114, 0, 109, 0, 110, 0, 95, 0, 65, 0, 77, 0, 0, 0, 105, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 105, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 106, 0, 97, 0, 95, 0, 74, 0, 112, 0, 97, 0, 110, 0, 95, 0, 74, 0, 80, 0, 0, 0, 107, 0, 97, 0, 95, 0, 71, 0, 101, 0, 111, 0, 114, 0, 95, 0, 71, 0, 69, 0, 0, 0, 107, 0, 109, 0, 95, 0, 75, 0, 104, 0, 109, 0, 114, 0, 95, 0, 75, 0, 72, 0, 0, 0, 107, 0, 111, 0, 95, 0, 75, 0, 111, 0, 114, 0, 101, 0, 95, 0, 75, 0, 82, 0, 0, 0, 108, 0, 111, 0, 95, 0, 76, 0, 97, 0, 111, 0, 111, 0, 95, 0, 76, 0, 65, 0, 0, 0, 109, 0, 110, 0, 95, 0, 77, 0, 111, 0, 110, 0, 103, 0, 95, 0, 67, 0, 78, 0, 0, 0, 109, 0, 121, 0, 95, 0, 77, 0, 121, 0, 109, 0, 114, 0, 95, 0, 77, 0, 77, 0, 0, 0, 112, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 82, 0, 0, 0, 115, 0, 105, 0, 95, 0, 83, 0, 105, 0, 110, 0, 104, 0, 95, 0, 76, 0, 75, 0, 0, 0, 115, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 87, 0, 83, 0, 0, 0, 115, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 116, 0, 104, 0, 95, 0, 84, 0, 104, 0, 97, 0, 105, 0, 95, 0, 84, 0, 72, 0, 0, 0, 117, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 117, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 90, 0, 0, 0, 97, 0, 98, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 71, 0, 69, 0, 0, 0, 97, 0, 101, 0, 95, 0, 65, 0, 118, 0, 115, 0, 116, 0, 95, 0, 73, 0, 82, 0, 0, 0, 97, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 72, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 83, 0, 65, 0, 0, 0, 97, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 97, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 90, 0, 0, 0, 98, 0, 101, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 66, 0, 89, 0, 0, 0, 98, 0, 103, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 66, 0, 71, 0, 0, 0, 98, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 86, 0, 85, 0, 0, 0, 98, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 76, 0, 0, 0, 98, 0, 111, 0, 95, 0, 84, 0, 105, 0, 98, 0, 116, 0, 95, 0, 67, 0, 78, 0, 0, 0, 98, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 65, 0, 0, 0, 99, 0, 114, 0, 95, 0, 67, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 65, 0, 0, 0, 99, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 90, 0, 0, 0, 99, 0, 117, 0, 95, 0, 71, 0, 108, 0, 97, 0, 103, 0, 95, 0, 66, 0, 71, 0, 0, 0, 100, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 75, 0, 0, 0, 100, 0, 122, 0, 95, 0, 84, 0, 105, 0, 98, 0, 116, 0, 95, 0, 66, 0, 84, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 85, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 66, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 104, 0, 97, 0, 119, 0, 95, 0, 71, 0, 66, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 85, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 83, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 88, 0, 0, 0, 101, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 69, 0, 0, 0, 102, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 102, 0, 102, 0, 95, 0, 65, 0, 100, 0, 108, 0, 109, 0, 95, 0, 71, 0, 78, 0, 0, 0, 102, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 73, 0, 0, 0, 102, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 79, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 82, 0, 0, 0, 103, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 89, 0, 0, 0, 103, 0, 117, 0, 95, 0, 71, 0, 117, 0, 106, 0, 114, 0, 95, 0, 73, 0, 78, 0, 0, 0, 104, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 72, 0, 82, 0, 0, 0, 104, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 72, 0, 84, 0, 0, 0, 104, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 72, 0, 85, 0, 0, 0, 105, 0, 105, 0, 95, 0, 89, 0, 105, 0, 105, 0, 105, 0, 95, 0, 67, 0, 78, 0, 0, 0, 105, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 83, 0, 0, 0, 107, 0, 107, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 67, 0, 78, 0, 0, 0, 107, 0, 107, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 77, 0, 78, 0, 0, 0, 107, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 76, 0, 0, 0, 107, 0, 110, 0, 95, 0, 75, 0, 110, 0, 100, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 117, 0, 95, 0, 89, 0, 101, 0, 122, 0, 105, 0, 95, 0, 71, 0, 69, 0, 0, 0, 107, 0, 121, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 67, 0, 78, 0, 0, 0, 107, 0, 121, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 75, 0, 71, 0, 0, 0, 107, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 82, 0, 0, 0, 108, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 76, 0, 84, 0, 0, 0, 108, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 76, 0, 86, 0, 0, 0, 109, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 71, 0, 0, 0, 109, 0, 107, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 77, 0, 75, 0, 0, 0, 109, 0, 108, 0, 95, 0, 77, 0, 108, 0, 121, 0, 109, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 110, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 77, 0, 78, 0, 0, 0, 109, 0, 115, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 67, 0, 67, 0, 0, 0, 109, 0, 115, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 68, 0, 0, 0, 109, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 89, 0, 0, 0, 109, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 84, 0, 0, 0, 110, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 79, 0, 0, 0, 110, 0, 101, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 110, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 76, 0, 0, 0, 111, 0, 114, 0, 95, 0, 79, 0, 114, 0, 121, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 112, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 112, 0, 97, 0, 95, 0, 71, 0, 117, 0, 114, 0, 117, 0, 95, 0, 73, 0, 78, 0, 0, 0, 112, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 76, 0, 0, 0, 114, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 73, 0, 0, 0, 114, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 79, 0, 0, 0, 114, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 87, 0, 0, 0, 115, 0, 100, 0, 95, 0, 75, 0, 104, 0, 111, 0, 106, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 100, 0, 95, 0, 83, 0, 105, 0, 110, 0, 100, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 75, 0, 0, 0, 115, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 73, 0, 0, 0, 115, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 87, 0, 0, 0, 115, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 79, 0, 0, 0, 115, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 76, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 83, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 69, 0, 0, 0, 115, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 69, 0, 0, 0, 115, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 68, 0, 0, 0, 116, 0, 97, 0, 95, 0, 84, 0, 97, 0, 109, 0, 108, 0, 95, 0, 73, 0, 78, 0, 0, 0, 116, 0, 101, 0, 95, 0, 84, 0, 101, 0, 108, 0, 117, 0, 95, 0, 73, 0, 78, 0, 0, 0, 116, 0, 103, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 116, 0, 103, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 84, 0, 74, 0, 0, 0, 116, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 77, 0, 0, 0, 116, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 79, 0, 0, 0, 116, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 82, 0, 0, 0, 117, 0, 103, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 67, 0, 78, 0, 0, 0, 117, 0, 103, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 75, 0, 90, 0, 0, 0, 117, 0, 107, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 85, 0, 65, 0, 0, 0, 117, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 65, 0, 70, 0, 0, 0, 118, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 86, 0, 78, 0, 0, 0, 122, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 78, 0, 0, 0, 122, 0, 104, 0, 95, 0, 66, 0, 111, 0, 112, 0, 111, 0, 95, 0, 84, 0, 87, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 98, 0, 95, 0, 84, 0, 87, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 72, 0, 75, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 77, 0, 79, 0, 0, 0, 97, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 74, 0, 0, 0, 97, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 84, 0, 0, 0, 97, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 65, 0, 0, 0, 97, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 97, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 83, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 65, 0, 69, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 66, 0, 72, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 68, 0, 90, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 69, 0, 72, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 81, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 74, 0, 79, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 75, 0, 77, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 75, 0, 87, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 76, 0, 66, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 76, 0, 89, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 77, 0, 65, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 77, 0, 82, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 79, 0, 77, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 83, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 81, 0, 65, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 83, 0, 68, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 83, 0, 89, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 84, 0, 78, 0, 0, 0, 97, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 89, 0, 69, 0, 0, 0, 97, 0, 115, 0, 95, 0, 66, 0, 101, 0, 110, 0, 103, 0, 95, 0, 73, 0, 78, 0, 0, 0, 97, 0, 118, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 97, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 79, 0, 0, 0, 97, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 81, 0, 0, 0, 97, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 84, 0, 82, 0, 0, 0, 97, 0, 122, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 98, 0, 97, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 98, 0, 103, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 79, 0, 0, 0, 98, 0, 111, 0, 95, 0, 77, 0, 97, 0, 114, 0, 99, 0, 95, 0, 67, 0, 78, 0, 0, 0, 98, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 82, 0, 0, 0, 99, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 68, 0, 0, 0, 99, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 83, 0, 0, 0, 99, 0, 101, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 99, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 85, 0, 0, 0, 99, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 82, 0, 0, 0, 99, 0, 117, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 99, 0, 118, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 99, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 66, 0, 0, 0, 100, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 84, 0, 0, 0, 100, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 72, 0, 0, 0, 100, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 90, 0, 0, 0, 100, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 76, 0, 73, 0, 0, 0, 100, 0, 118, 0, 95, 0, 68, 0, 105, 0, 97, 0, 107, 0, 95, 0, 77, 0, 86, 0, 0, 0, 101, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 72, 0, 0, 0, 101, 0, 108, 0, 95, 0, 71, 0, 114, 0, 101, 0, 107, 0, 95, 0, 67, 0, 89, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 71, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 84, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 85, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 71, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 82, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 79, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 76, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 79, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 82, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 79, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 65, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 67, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 81, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 84, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 72, 0, 78, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 67, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 73, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 65, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 69, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 82, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 86, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 89, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 86, 0, 69, 0, 0, 0, 101, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 83, 0, 0, 0, 102, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 65, 0, 70, 0, 0, 0, 102, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 84, 0, 74, 0, 0, 0, 102, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 78, 0, 0, 0, 102, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 74, 0, 0, 0, 102, 0, 114, 0, 95, 0, 66, 0, 114, 0, 97, 0, 105, 0, 95, 0, 70, 0, 82, 0, 0, 0, 102, 0, 114, 0, 95, 0, 68, 0, 117, 0, 112, 0, 108, 0, 95, 0, 70, 0, 82, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 70, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 74, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 76, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 70, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 71, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 73, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 90, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 65, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 70, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 78, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 80, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 77, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 76, 0, 85, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 65, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 67, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 70, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 81, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 82, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 67, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 70, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 77, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 69, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 67, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 78, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 89, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 68, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 70, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 71, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 78, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 87, 0, 70, 0, 0, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 89, 0, 84, 0, 0, 0, 102, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 76, 0, 0, 0, 103, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 69, 0, 0, 0, 103, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 66, 0, 0, 0, 103, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 83, 0, 0, 0, 103, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 77, 0, 0, 0, 104, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 67, 0, 77, 0, 0, 0, 104, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 78, 0, 71, 0, 0, 0, 104, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 83, 0, 68, 0, 0, 0, 104, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 69, 0, 0, 0, 104, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 104, 0, 105, 0, 95, 0, 77, 0, 97, 0, 104, 0, 106, 0, 95, 0, 73, 0, 78, 0, 0, 0, 104, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 71, 0, 0, 0, 104, 0, 117, 0, 95, 0, 72, 0, 117, 0, 110, 0, 103, 0, 95, 0, 72, 0, 85, 0, 0, 0, 104, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 65, 0, 0, 0, 105, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 105, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 105, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 77, 0, 0, 0, 105, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 86, 0, 65, 0, 0, 0, 105, 0, 117, 0, 95, 0, 67, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 65, 0, 0, 0, 105, 0, 119, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 95, 0, 73, 0, 76, 0, 0, 0, 106, 0, 97, 0, 95, 0, 72, 0, 105, 0, 114, 0, 97, 0, 95, 0, 74, 0, 80, 0, 0, 0, 106, 0, 97, 0, 95, 0, 75, 0, 97, 0, 110, 0, 97, 0, 95, 0, 74, 0, 80, 0, 0, 0, 106, 0, 105, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 95, 0, 85, 0, 65, 0, 0, 0, 106, 0, 118, 0, 95, 0, 74, 0, 97, 0, 118, 0, 97, 0, 95, 0, 73, 0, 68, 0, 0, 0, 106, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 106, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 107, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 68, 0, 0, 0, 107, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 107, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 65, 0, 0, 0, 107, 0, 107, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 65, 0, 70, 0, 0, 0, 107, 0, 107, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 107, 0, 107, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 75, 0, 90, 0, 0, 0, 107, 0, 111, 0, 95, 0, 72, 0, 97, 0, 110, 0, 103, 0, 95, 0, 75, 0, 82, 0, 0, 0, 107, 0, 111, 0, 95, 0, 74, 0, 97, 0, 109, 0, 111, 0, 95, 0, 75, 0, 82, 0, 0, 0, 107, 0, 111, 0, 95, 0, 75, 0, 111, 0, 114, 0, 101, 0, 95, 0, 75, 0, 80, 0, 0, 0, 107, 0, 115, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 71, 0, 66, 0, 0, 0, 107, 0, 115, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 117, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 81, 0, 0, 0, 107, 0, 117, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 76, 0, 66, 0, 0, 0, 107, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 77, 0, 0, 0, 107, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 69, 0, 0, 0, 107, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 82, 0, 0, 0, 107, 0, 118, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 107, 0, 118, 0, 95, 0, 80, 0, 101, 0, 114, 0, 109, 0, 95, 0, 82, 0, 85, 0, 0, 0, 107, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 66, 0, 0, 0, 108, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 86, 0, 65, 0, 0, 0, 108, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 76, 0, 85, 0, 0, 0, 108, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 108, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 76, 0, 0, 0, 108, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 68, 0, 0, 0, 108, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 68, 0, 0, 0, 109, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 72, 0, 0, 0, 109, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 90, 0, 0, 0, 109, 0, 107, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 65, 0, 76, 0, 0, 0, 109, 0, 107, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 71, 0, 82, 0, 0, 0, 109, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 79, 0, 0, 0, 109, 0, 114, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 114, 0, 95, 0, 77, 0, 111, 0, 100, 0, 105, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 78, 0, 0, 0, 110, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 82, 0, 0, 0, 110, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 74, 0, 0, 0, 110, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 87, 0, 0, 0, 110, 0, 101, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 66, 0, 84, 0, 0, 0, 110, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 65, 0, 0, 0, 110, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 87, 0, 0, 0, 110, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 69, 0, 0, 0, 110, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 82, 0, 0, 0, 110, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 79, 0, 0, 0, 110, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 79, 0, 0, 0, 110, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 110, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 110, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 87, 0, 0, 0, 111, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 82, 0, 0, 0, 111, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 84, 0, 0, 0, 111, 0, 115, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 71, 0, 69, 0, 0, 0, 112, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 65, 0, 0, 0, 112, 0, 115, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 65, 0, 70, 0, 0, 0, 112, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 79, 0, 0, 0, 112, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 86, 0, 0, 0, 112, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 87, 0, 0, 0, 112, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 79, 0, 0, 0, 112, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 90, 0, 0, 0, 112, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 84, 0, 0, 0, 112, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 84, 0, 0, 0, 112, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 76, 0, 0, 0, 113, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 69, 0, 0, 0, 114, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 72, 0, 0, 0, 114, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 68, 0, 0, 0, 114, 0, 117, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 75, 0, 90, 0, 0, 0, 115, 0, 97, 0, 95, 0, 66, 0, 104, 0, 107, 0, 115, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 97, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 97, 0, 95, 0, 71, 0, 114, 0, 97, 0, 110, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 97, 0, 95, 0, 78, 0, 97, 0, 110, 0, 100, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 97, 0, 95, 0, 83, 0, 104, 0, 114, 0, 100, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 97, 0, 95, 0, 83, 0, 105, 0, 100, 0, 100, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 115, 0, 100, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 115, 0, 100, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 79, 0, 0, 0, 115, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 70, 0, 0, 0, 115, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 83, 0, 0, 0, 115, 0, 111, 0, 95, 0, 79, 0, 115, 0, 109, 0, 97, 0, 95, 0, 83, 0, 79, 0, 0, 0, 115, 0, 113, 0, 95, 0, 69, 0, 108, 0, 98, 0, 97, 0, 95, 0, 65, 0, 76, 0, 0, 0, 115, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 75, 0, 0, 0, 115, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 88, 0, 75, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 66, 0, 65, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 88, 0, 75, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 79, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 85, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 82, 0, 0, 0, 115, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 115, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 76, 0, 83, 0, 0, 0, 115, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 115, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 115, 0, 117, 0, 95, 0, 83, 0, 117, 0, 110, 0, 100, 0, 95, 0, 73, 0, 68, 0, 0, 0, 115, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 88, 0, 0, 0, 115, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 115, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 116, 0, 105, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 69, 0, 82, 0, 0, 0, 116, 0, 105, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 69, 0, 84, 0, 0, 0, 116, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 70, 0, 0, 0, 116, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 82, 0, 0, 0, 116, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 116, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 116, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 89, 0, 0, 0, 116, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 116, 0, 116, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 116, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 70, 0, 0, 0, 117, 0, 103, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 77, 0, 78, 0, 0, 0, 117, 0, 107, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 77, 0, 68, 0, 0, 0, 117, 0, 107, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 83, 0, 75, 0, 0, 0, 117, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 78, 0, 0, 0, 117, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 77, 0, 85, 0, 0, 0, 117, 0, 122, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 67, 0, 78, 0, 0, 0, 118, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 119, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 69, 0, 0, 0, 119, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 78, 0, 0, 0, 120, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 121, 0, 105, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 95, 0, 67, 0, 65, 0, 0, 0, 121, 0, 105, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 95, 0, 71, 0, 66, 0, 0, 0, 121, 0, 105, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 95, 0, 83, 0, 69, 0, 0, 0, 121, 0, 105, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 95, 0, 85, 0, 65, 0, 0, 0, 121, 0, 105, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 95, 0, 85, 0, 83, 0, 0, 0, 121, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 105, 0, 95, 0, 67, 0, 78, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 65, 0, 85, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 66, 0, 78, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 71, 0, 66, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 71, 0, 70, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 73, 0, 68, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 77, 0, 89, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 80, 0, 65, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 80, 0, 70, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 80, 0, 72, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 83, 0, 82, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 84, 0, 72, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 85, 0, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 86, 0, 78, 0, 0, 0, 122, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 108, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 98, 0, 104, 0, 105, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 97, 0, 110, 0, 95, 0, 78, 0, 107, 0, 111, 0, 111, 0, 95, 0, 71, 0, 78, 0, 0, 0, 97, 0, 104, 0, 111, 0, 95, 0, 65, 0, 104, 0, 111, 0, 109, 0, 95, 0, 73, 0, 78, 0, 0, 0, 97, 0, 107, 0, 107, 0, 95, 0, 88, 0, 115, 0, 117, 0, 120, 0, 95, 0, 73, 0, 81, 0, 0, 0, 97, 0, 112, 0, 100, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 84, 0, 71, 0, 0, 0, 97, 0, 114, 0, 99, 0, 95, 0, 65, 0, 114, 0, 109, 0, 105, 0, 95, 0, 73, 0, 82, 0, 0, 0, 97, 0, 114, 0, 99, 0, 95, 0, 78, 0, 98, 0, 97, 0, 116, 0, 95, 0, 74, 0, 79, 0, 0, 0, 97, 0, 114, 0, 99, 0, 95, 0, 80, 0, 97, 0, 108, 0, 109, 0, 95, 0, 83, 0, 89, 0, 0, 0, 97, 0, 115, 0, 101, 0, 95, 0, 83, 0, 103, 0, 110, 0, 119, 0, 95, 0, 85, 0, 83, 0, 0, 0, 98, 0, 97, 0, 120, 0, 95, 0, 66, 0, 97, 0, 109, 0, 117, 0, 95, 0, 67, 0, 77, 0, 0, 0, 98, 0, 103, 0, 120, 0, 95, 0, 71, 0, 114, 0, 101, 0, 107, 0, 95, 0, 84, 0, 82, 0, 0, 0, 98, 0, 108, 0, 116, 0, 95, 0, 84, 0, 97, 0, 118, 0, 116, 0, 95, 0, 86, 0, 78, 0, 0, 0, 98, 0, 115, 0, 113, 0, 95, 0, 66, 0, 97, 0, 115, 0, 115, 0, 95, 0, 76, 0, 82, 0, 0, 0, 98, 0, 116, 0, 118, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 80, 0, 75, 0, 0, 0, 99, 0, 99, 0, 112, 0, 95, 0, 67, 0, 97, 0, 107, 0, 109, 0, 95, 0, 66, 0, 68, 0, 0, 0, 99, 0, 104, 0, 114, 0, 95, 0, 67, 0, 104, 0, 101, 0, 114, 0, 95, 0, 85, 0, 83, 0, 0, 0, 99, 0, 106, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 75, 0, 72, 0, 0, 0, 99, 0, 106, 0, 109, 0, 95, 0, 67, 0, 104, 0, 97, 0, 109, 0, 95, 0, 86, 0, 78, 0, 0, 0, 99, 0, 109, 0, 103, 0, 95, 0, 83, 0, 111, 0, 121, 0, 111, 0, 95, 0, 77, 0, 78, 0, 0, 0, 99, 0, 111, 0, 112, 0, 95, 0, 67, 0, 111, 0, 112, 0, 116, 0, 95, 0, 69, 0, 71, 0, 0, 0, 99, 0, 116, 0, 100, 0, 95, 0, 80, 0, 97, 0, 117, 0, 99, 0, 95, 0, 77, 0, 77, 0, 0, 0, 101, 0, 103, 0, 121, 0, 95, 0, 69, 0, 103, 0, 121, 0, 112, 0, 95, 0, 69, 0, 71, 0, 0, 0, 101, 0, 107, 0, 121, 0, 95, 0, 75, 0, 97, 0, 108, 0, 105, 0, 95, 0, 77, 0, 77, 0, 0, 0, 101, 0, 115, 0, 103, 0, 95, 0, 71, 0, 111, 0, 110, 0, 109, 0, 95, 0, 73, 0, 78, 0, 0, 0, 101, 0, 116, 0, 116, 0, 95, 0, 73, 0, 116, 0, 97, 0, 108, 0, 95, 0, 73, 0, 84, 0, 0, 0, 102, 0, 105, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 103, 0, 111, 0, 116, 0, 95, 0, 71, 0, 111, 0, 116, 0, 104, 0, 95, 0, 85, 0, 65, 0, 0, 0, 103, 0, 114, 0, 99, 0, 95, 0, 67, 0, 112, 0, 114, 0, 116, 0, 95, 0, 67, 0, 89, 0, 0, 0, 103, 0, 114, 0, 99, 0, 95, 0, 76, 0, 105, 0, 110, 0, 98, 0, 95, 0, 71, 0, 82, 0, 0, 0, 104, 0, 108, 0, 117, 0, 95, 0, 72, 0, 108, 0, 117, 0, 119, 0, 95, 0, 84, 0, 82, 0, 0, 0, 104, 0, 109, 0, 100, 0, 95, 0, 80, 0, 108, 0, 114, 0, 100, 0, 95, 0, 67, 0, 78, 0, 0, 0, 104, 0, 110, 0, 106, 0, 95, 0, 72, 0, 109, 0, 110, 0, 103, 0, 95, 0, 76, 0, 65, 0, 0, 0, 107, 0, 97, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 77, 0, 0, 0, 107, 0, 104, 0, 98, 0, 95, 0, 84, 0, 97, 0, 108, 0, 117, 0, 95, 0, 67, 0, 78, 0, 0, 0, 107, 0, 104, 0, 116, 0, 95, 0, 77, 0, 121, 0, 109, 0, 114, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 114, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 85, 0, 0, 0, 108, 0, 97, 0, 98, 0, 95, 0, 76, 0, 105, 0, 110, 0, 97, 0, 95, 0, 71, 0, 82, 0, 0, 0, 108, 0, 99, 0, 112, 0, 95, 0, 84, 0, 104, 0, 97, 0, 105, 0, 95, 0, 67, 0, 78, 0, 0, 0, 108, 0, 101, 0, 112, 0, 95, 0, 76, 0, 101, 0, 112, 0, 99, 0, 95, 0, 73, 0, 78, 0, 0, 0, 108, 0, 105, 0, 102, 0, 95, 0, 76, 0, 105, 0, 109, 0, 98, 0, 95, 0, 73, 0, 78, 0, 0, 0, 108, 0, 105, 0, 115, 0, 95, 0, 76, 0, 105, 0, 115, 0, 117, 0, 95, 0, 67, 0, 78, 0, 0, 0, 109, 0, 102, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 84, 0, 72, 0, 0, 0, 109, 0, 102, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 85, 0, 0, 0, 109, 0, 105, 0, 115, 0, 95, 0, 72, 0, 97, 0, 116, 0, 114, 0, 95, 0, 73, 0, 81, 0, 0, 0, 109, 0, 105, 0, 115, 0, 95, 0, 77, 0, 101, 0, 100, 0, 102, 0, 95, 0, 78, 0, 71, 0, 0, 0, 109, 0, 114, 0, 111, 0, 95, 0, 77, 0, 114, 0, 111, 0, 111, 0, 95, 0, 66, 0, 68, 0, 0, 0, 109, 0, 119, 0, 119, 0, 95, 0, 72, 0, 109, 0, 110, 0, 112, 0, 95, 0, 85, 0, 83, 0, 0, 0, 109, 0, 121, 0, 122, 0, 95, 0, 77, 0, 97, 0, 110, 0, 100, 0, 95, 0, 73, 0, 82, 0, 0, 0, 110, 0, 110, 0, 112, 0, 95, 0, 87, 0, 99, 0, 104, 0, 111, 0, 95, 0, 73, 0, 78, 0, 0, 0, 110, 0, 111, 0, 100, 0, 95, 0, 76, 0, 97, 0, 110, 0, 97, 0, 95, 0, 84, 0, 72, 0, 0, 0, 110, 0, 111, 0, 110, 0, 95, 0, 82, 0, 117, 0, 110, 0, 114, 0, 95, 0, 83, 0, 69, 0, 0, 0, 111, 0, 115, 0, 97, 0, 95, 0, 79, 0, 115, 0, 103, 0, 101, 0, 95, 0, 85, 0, 83, 0, 0, 0, 111, 0, 116, 0, 107, 0, 95, 0, 79, 0, 114, 0, 107, 0, 104, 0, 95, 0, 77, 0, 78, 0, 0, 0, 112, 0, 97, 0, 108, 0, 95, 0, 80, 0, 104, 0, 108, 0, 105, 0, 95, 0, 73, 0, 82, 0, 0, 0, 112, 0, 97, 0, 108, 0, 95, 0, 80, 0, 104, 0, 108, 0, 112, 0, 95, 0, 67, 0, 78, 0, 0, 0, 112, 0, 97, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 87, 0, 0, 0, 112, 0, 101, 0, 111, 0, 95, 0, 88, 0, 112, 0, 101, 0, 111, 0, 95, 0, 73, 0, 82, 0, 0, 0, 112, 0, 104, 0, 110, 0, 95, 0, 80, 0, 104, 0, 110, 0, 120, 0, 95, 0, 76, 0, 66, 0, 0, 0, 112, 0, 107, 0, 97, 0, 95, 0, 66, 0, 114, 0, 97, 0, 104, 0, 95, 0, 73, 0, 78, 0, 0, 0, 112, 0, 114, 0, 97, 0, 95, 0, 75, 0, 104, 0, 97, 0, 114, 0, 95, 0, 80, 0, 75, 0, 0, 0, 114, 0, 104, 0, 103, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 77, 0, 77, 0, 0, 0, 115, 0, 97, 0, 122, 0, 95, 0, 83, 0, 97, 0, 117, 0, 114, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 103, 0, 97, 0, 95, 0, 79, 0, 103, 0, 97, 0, 109, 0, 95, 0, 73, 0, 69, 0, 0, 0, 115, 0, 109, 0, 112, 0, 95, 0, 83, 0, 97, 0, 109, 0, 114, 0, 95, 0, 73, 0, 76, 0, 0, 0, 115, 0, 111, 0, 103, 0, 95, 0, 83, 0, 111, 0, 103, 0, 100, 0, 95, 0, 85, 0, 90, 0, 0, 0, 115, 0, 114, 0, 98, 0, 95, 0, 83, 0, 111, 0, 114, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 119, 0, 98, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 89, 0, 84, 0, 0, 0, 115, 0, 121, 0, 114, 0, 95, 0, 83, 0, 121, 0, 114, 0, 99, 0, 95, 0, 73, 0, 81, 0, 0, 0, 116, 0, 100, 0, 100, 0, 95, 0, 84, 0, 97, 0, 108, 0, 101, 0, 95, 0, 67, 0, 78, 0, 0, 0, 116, 0, 107, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 75, 0, 0, 0, 116, 0, 112, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 71, 0, 0, 0, 116, 0, 114, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 87, 0, 0, 0, 116, 0, 118, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 86, 0, 0, 0, 116, 0, 120, 0, 103, 0, 95, 0, 84, 0, 97, 0, 110, 0, 103, 0, 95, 0, 67, 0, 78, 0, 0, 0, 117, 0, 103, 0, 97, 0, 95, 0, 85, 0, 103, 0, 97, 0, 114, 0, 95, 0, 83, 0, 89, 0, 0, 0, 117, 0, 110, 0, 114, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 118, 0, 97, 0, 105, 0, 95, 0, 86, 0, 97, 0, 105, 0, 105, 0, 95, 0, 76, 0, 82, 0, 0, 0, 119, 0, 115, 0, 103, 0, 95, 0, 71, 0, 111, 0, 110, 0, 103, 0, 95, 0, 73, 0, 78, 0, 0, 0, 120, 0, 99, 0, 111, 0, 95, 0, 67, 0, 104, 0, 114, 0, 115, 0, 95, 0, 85, 0, 90, 0, 0, 0, 120, 0, 99, 0, 114, 0, 95, 0, 67, 0, 97, 0, 114, 0, 105, 0, 95, 0, 84, 0, 82, 0, 0, 0, 120, 0, 108, 0, 99, 0, 95, 0, 76, 0, 121, 0, 99, 0, 105, 0, 95, 0, 84, 0, 82, 0, 0, 0, 120, 0, 108, 0, 100, 0, 95, 0, 76, 0, 121, 0, 100, 0, 105, 0, 95, 0, 84, 0, 82, 0, 0, 0, 120, 0, 109, 0, 110, 0, 95, 0, 77, 0, 97, 0, 110, 0, 105, 0, 95, 0, 67, 0, 78, 0, 0, 0, 120, 0, 109, 0, 114, 0, 95, 0, 77, 0, 101, 0, 114, 0, 99, 0, 95, 0, 83, 0, 68, 0, 0, 0, 120, 0, 110, 0, 97, 0, 95, 0, 78, 0, 97, 0, 114, 0, 98, 0, 95, 0, 83, 0, 65, 0, 0, 0, 120, 0, 112, 0, 114, 0, 95, 0, 80, 0, 114, 0, 116, 0, 105, 0, 95, 0, 73, 0, 82, 0, 0, 0, 120, 0, 115, 0, 97, 0, 95, 0, 83, 0, 97, 0, 114, 0, 98, 0, 95, 0, 89, 0, 69, 0, 0, 0, 121, 0, 117, 0, 101, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 122, 0, 103, 0, 104, 0, 95, 0, 84, 0, 102, 0, 110, 0, 103, 0, 95, 0, 77, 0, 65, 0, 0, 0, 122, 0, 104, 0, 120, 0, 95, 0, 78, 0, 115, 0, 104, 0, 117, 0, 95, 0, 67, 0, 78, 0, 0, 0, 122, 0, 107, 0, 116, 0, 95, 0, 75, 0, 105, 0, 116, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 109, 0, 105, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 111, 0, 107, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 97, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 97, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 97, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 98, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 98, 0, 113, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 98, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 72, 0, 0, 0, 97, 0, 98, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 98, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 99, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 99, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 97, 0, 99, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 97, 0, 100, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 72, 0, 0, 0, 97, 0, 100, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 100, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 100, 0, 112, 0, 95, 0, 84, 0, 105, 0, 98, 0, 116, 0, 95, 0, 66, 0, 84, 0, 0, 0, 97, 0, 100, 0, 121, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 97, 0, 100, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 101, 0, 98, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 84, 0, 78, 0, 0, 0, 97, 0, 101, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 103, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 103, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 103, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 103, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 103, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 103, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 97, 0, 104, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 104, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 106, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 108, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 108, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 108, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 88, 0, 75, 0, 0, 0, 97, 0, 108, 0, 116, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 97, 0, 109, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 109, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 109, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 97, 0, 109, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 110, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 110, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 110, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 110, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 111, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 111, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 111, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 97, 0, 112, 0, 99, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 112, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 112, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 112, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 112, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 114, 0, 99, 0, 95, 0, 69, 0, 108, 0, 121, 0, 109, 0, 95, 0, 73, 0, 82, 0, 0, 0, 97, 0, 114, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 114, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 76, 0, 0, 0, 97, 0, 114, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 79, 0, 0, 0, 97, 0, 114, 0, 113, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 68, 0, 90, 0, 0, 0, 97, 0, 114, 0, 115, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 83, 0, 65, 0, 0, 0, 97, 0, 114, 0, 121, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 77, 0, 65, 0, 0, 0, 97, 0, 114, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 69, 0, 71, 0, 0, 0, 97, 0, 115, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 97, 0, 115, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 115, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 115, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 83, 0, 0, 0, 97, 0, 116, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 116, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 116, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 97, 0, 117, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 118, 0, 108, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 118, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 118, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 118, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 119, 0, 97, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 97, 0, 119, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 119, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 119, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 97, 0, 121, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 97, 0, 108, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 98, 0, 97, 0, 110, 0, 95, 0, 66, 0, 97, 0, 108, 0, 105, 0, 95, 0, 73, 0, 68, 0, 0, 0, 98, 0, 97, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 98, 0, 97, 0, 112, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 98, 0, 97, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 84, 0, 0, 0, 98, 0, 97, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 98, 0, 97, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 98, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 98, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 98, 0, 99, 0, 95, 0, 66, 0, 97, 0, 116, 0, 107, 0, 95, 0, 73, 0, 68, 0, 0, 0, 98, 0, 98, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 98, 0, 98, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 98, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 98, 0, 98, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 98, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 99, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 99, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 99, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 73, 0, 0, 0, 98, 0, 99, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 99, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 99, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 99, 0, 113, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 99, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 100, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 101, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 101, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 101, 0, 106, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 83, 0, 68, 0, 0, 0, 98, 0, 101, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 77, 0, 0, 0, 98, 0, 101, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 101, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 98, 0, 101, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 101, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 98, 0, 102, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 98, 0, 102, 0, 113, 0, 95, 0, 84, 0, 97, 0, 109, 0, 108, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 102, 0, 116, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 98, 0, 102, 0, 121, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 103, 0, 99, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 103, 0, 110, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 98, 0, 104, 0, 98, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 104, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 104, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 104, 0, 111, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 104, 0, 111, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 77, 0, 85, 0, 0, 0, 98, 0, 104, 0, 111, 0, 95, 0, 75, 0, 116, 0, 104, 0, 105, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 104, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 105, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 105, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 105, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 98, 0, 105, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 105, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 98, 0, 105, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 105, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 106, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 106, 0, 105, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 106, 0, 106, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 106, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 98, 0, 106, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 106, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 106, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 78, 0, 0, 0, 98, 0, 106, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 107, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 107, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 98, 0, 107, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 107, 0, 117, 0, 95, 0, 66, 0, 117, 0, 104, 0, 100, 0, 95, 0, 80, 0, 72, 0, 0, 0, 98, 0, 107, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 98, 0, 107, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 109, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 109, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 109, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 76, 0, 0, 0, 98, 0, 109, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 110, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 110, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 110, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 111, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 111, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 111, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 112, 0, 121, 0, 95, 0, 66, 0, 101, 0, 110, 0, 103, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 113, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 113, 0, 105, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 98, 0, 113, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 113, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 73, 0, 0, 0, 98, 0, 114, 0, 97, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 114, 0, 104, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 98, 0, 114, 0, 120, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 114, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 115, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 115, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 98, 0, 115, 0, 116, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 116, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 98, 0, 116, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 117, 0, 97, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 98, 0, 117, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 89, 0, 84, 0, 0, 0, 98, 0, 117, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 117, 0, 103, 0, 95, 0, 66, 0, 117, 0, 103, 0, 105, 0, 95, 0, 73, 0, 68, 0, 0, 0, 98, 0, 117, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 98, 0, 117, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 117, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 98, 0, 117, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 117, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 117, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 118, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 81, 0, 0, 0, 98, 0, 119, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 119, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 120, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 121, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 121, 0, 110, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 69, 0, 82, 0, 0, 0, 98, 0, 121, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 121, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 121, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 98, 0, 121, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 122, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 122, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 76, 0, 0, 0, 98, 0, 122, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 122, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 98, 0, 122, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 99, 0, 97, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 99, 0, 98, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 99, 0, 99, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 99, 0, 101, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 99, 0, 102, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 99, 0, 103, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 99, 0, 104, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 77, 0, 0, 0, 99, 0, 104, 0, 109, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 99, 0, 104, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 99, 0, 104, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 99, 0, 105, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 99, 0, 106, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 99, 0, 107, 0, 98, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 81, 0, 0, 0, 99, 0, 107, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 99, 0, 107, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 99, 0, 107, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 99, 0, 108, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 99, 0, 109, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 99, 0, 109, 0, 103, 0, 95, 0, 90, 0, 97, 0, 110, 0, 98, 0, 95, 0, 77, 0, 78, 0, 0, 0, 99, 0, 112, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 99, 0, 114, 0, 104, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 85, 0, 65, 0, 0, 0, 99, 0, 114, 0, 106, 0, 95, 0, 67, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 65, 0, 0, 0, 99, 0, 114, 0, 107, 0, 95, 0, 67, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 65, 0, 0, 0, 99, 0, 114, 0, 108, 0, 95, 0, 67, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 65, 0, 0, 0, 99, 0, 114, 0, 109, 0, 95, 0, 67, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 65, 0, 0, 0, 99, 0, 114, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 67, 0, 0, 0, 99, 0, 115, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 76, 0, 0, 0, 99, 0, 115, 0, 119, 0, 95, 0, 67, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 65, 0, 0, 0, 100, 0, 97, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 97, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 97, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 97, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 97, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 100, 0, 97, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 100, 0, 97, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 100, 0, 98, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 98, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 99, 0, 99, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 78, 0, 0, 0, 100, 0, 100, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 101, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 100, 0, 103, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 103, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 103, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 103, 0, 108, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 103, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 100, 0, 103, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 105, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 106, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 69, 0, 0, 0, 100, 0, 110, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 73, 0, 0, 0, 100, 0, 111, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 111, 0, 105, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 78, 0, 0, 0, 100, 0, 111, 0, 105, 0, 95, 0, 68, 0, 111, 0, 103, 0, 114, 0, 95, 0, 73, 0, 78, 0, 0, 0, 100, 0, 111, 0, 105, 0, 95, 0, 84, 0, 97, 0, 107, 0, 114, 0, 95, 0, 73, 0, 78, 0, 0, 0, 100, 0, 111, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 111, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 114, 0, 104, 0, 95, 0, 77, 0, 111, 0, 110, 0, 103, 0, 95, 0, 67, 0, 78, 0, 0, 0, 100, 0, 114, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 114, 0, 115, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 115, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 100, 0, 116, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 76, 0, 0, 0, 100, 0, 116, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 89, 0, 0, 0, 100, 0, 116, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 116, 0, 121, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 100, 0, 117, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 100, 0, 117, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 117, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 117, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 118, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 119, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 100, 0, 121, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 78, 0, 0, 0, 100, 0, 121, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 70, 0, 0, 0, 100, 0, 122, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 101, 0, 98, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 101, 0, 102, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 101, 0, 103, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 101, 0, 107, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 101, 0, 109, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 101, 0, 109, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 101, 0, 110, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 101, 0, 110, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 101, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 101, 0, 114, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 52, 0, 49, 0, 57, 0, 0, 0, 101, 0, 115, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 101, 0, 116, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 101, 0, 116, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 101, 0, 116, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 101, 0, 119, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 101, 0, 120, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 83, 0, 0, 0, 102, 0, 97, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 97, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 97, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 97, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 97, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 81, 0, 0, 0, 102, 0, 102, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 102, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 76, 0, 0, 0, 102, 0, 105, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 83, 0, 68, 0, 0, 0, 102, 0, 105, 0, 108, 0, 95, 0, 84, 0, 103, 0, 108, 0, 103, 0, 95, 0, 80, 0, 72, 0, 0, 0, 102, 0, 105, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 69, 0, 0, 0, 102, 0, 108, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 109, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 111, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 111, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 74, 0, 0, 0, 102, 0, 111, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 112, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 113, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 114, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 102, 0, 114, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 82, 0, 0, 0, 102, 0, 114, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 102, 0, 114, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 102, 0, 117, 0, 98, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 67, 0, 77, 0, 0, 0, 102, 0, 117, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 87, 0, 70, 0, 0, 0, 102, 0, 117, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 117, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 78, 0, 0, 0, 102, 0, 117, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 117, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 69, 0, 0, 0, 102, 0, 117, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 102, 0, 117, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 102, 0, 117, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 102, 0, 118, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 68, 0, 0, 0, 103, 0, 97, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 72, 0, 0, 0, 103, 0, 97, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 97, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 68, 0, 0, 0, 103, 0, 97, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 97, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 97, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 97, 0, 110, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 103, 0, 97, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 97, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 103, 0, 98, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 98, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 98, 0, 109, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 103, 0, 98, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 98, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 103, 0, 99, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 70, 0, 0, 0, 103, 0, 100, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 100, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 100, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 101, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 101, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 101, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 101, 0, 122, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 69, 0, 84, 0, 0, 0, 103, 0, 102, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 103, 0, 110, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 103, 0, 104, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 105, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 73, 0, 0, 0, 103, 0, 105, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 106, 0, 107, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 103, 0, 106, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 106, 0, 117, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 103, 0, 107, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 107, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 108, 0, 107, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 103, 0, 109, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 109, 0, 118, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 110, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 110, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 111, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 111, 0, 102, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 111, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 111, 0, 109, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 103, 0, 111, 0, 110, 0, 95, 0, 84, 0, 101, 0, 108, 0, 117, 0, 95, 0, 73, 0, 78, 0, 0, 0, 103, 0, 111, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 103, 0, 111, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 76, 0, 0, 0, 103, 0, 114, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 114, 0, 116, 0, 95, 0, 66, 0, 101, 0, 110, 0, 103, 0, 95, 0, 73, 0, 78, 0, 0, 0, 103, 0, 114, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 115, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 72, 0, 0, 0, 103, 0, 117, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 82, 0, 0, 0, 103, 0, 117, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 79, 0, 0, 0, 103, 0, 117, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 117, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 72, 0, 0, 0, 103, 0, 117, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 117, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 117, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 103, 0, 118, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 118, 0, 114, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 103, 0, 118, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 119, 0, 99, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 119, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 103, 0, 119, 0, 116, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 103, 0, 121, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 97, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 97, 0, 107, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 104, 0, 97, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 97, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 104, 0, 97, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 65, 0, 70, 0, 0, 0, 104, 0, 98, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 100, 0, 121, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 104, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 105, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 105, 0, 102, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 70, 0, 74, 0, 0, 0, 104, 0, 105, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 74, 0, 0, 0, 104, 0, 105, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 105, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 105, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 104, 0, 108, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 109, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 110, 0, 100, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 104, 0, 110, 0, 101, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 104, 0, 110, 0, 110, 0, 95, 0, 72, 0, 97, 0, 110, 0, 111, 0, 95, 0, 80, 0, 72, 0, 0, 0, 104, 0, 110, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 104, 0, 110, 0, 111, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 104, 0, 111, 0, 99, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 104, 0, 111, 0, 99, 0, 95, 0, 87, 0, 97, 0, 114, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 104, 0, 111, 0, 106, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 104, 0, 111, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 104, 0, 115, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 104, 0, 115, 0, 110, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 104, 0, 117, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 105, 0, 97, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 97, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 98, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 89, 0, 0, 0, 105, 0, 98, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 105, 0, 98, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 99, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 99, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 100, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 100, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 100, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 102, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 71, 0, 0, 0, 105, 0, 103, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 103, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 106, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 107, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 107, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 105, 0, 107, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 107, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 108, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 105, 0, 109, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 110, 0, 104, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 105, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 105, 0, 111, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 114, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 119, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 119, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 105, 0, 122, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 85, 0, 0, 0, 105, 0, 122, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 106, 0, 97, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 106, 0, 97, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 74, 0, 77, 0, 0, 0, 106, 0, 98, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 106, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 106, 0, 103, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 106, 0, 103, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 106, 0, 105, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 106, 0, 109, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 106, 0, 109, 0, 108, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 106, 0, 114, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 106, 0, 117, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 75, 0, 0, 0, 107, 0, 97, 0, 97, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 85, 0, 90, 0, 0, 0, 107, 0, 97, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 90, 0, 0, 0, 107, 0, 97, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 97, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 97, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 107, 0, 97, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 107, 0, 97, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 76, 0, 0, 0, 107, 0, 98, 0, 100, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 107, 0, 98, 0, 100, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 84, 0, 82, 0, 0, 0, 107, 0, 98, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 98, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 98, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 98, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 98, 0, 121, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 78, 0, 69, 0, 0, 0, 107, 0, 99, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 107, 0, 99, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 87, 0, 0, 0, 107, 0, 99, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 99, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 100, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 107, 0, 100, 0, 104, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 84, 0, 71, 0, 0, 0, 107, 0, 100, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 100, 0, 116, 0, 95, 0, 84, 0, 104, 0, 97, 0, 105, 0, 95, 0, 75, 0, 72, 0, 0, 0, 107, 0, 100, 0, 116, 0, 95, 0, 84, 0, 104, 0, 97, 0, 105, 0, 95, 0, 76, 0, 65, 0, 0, 0, 107, 0, 100, 0, 116, 0, 95, 0, 84, 0, 104, 0, 97, 0, 105, 0, 95, 0, 84, 0, 72, 0, 0, 0, 107, 0, 101, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 86, 0, 0, 0, 107, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 107, 0, 101, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 102, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 73, 0, 0, 0, 107, 0, 102, 0, 114, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 102, 0, 121, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 103, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 107, 0, 103, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 103, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 82, 0, 0, 0, 107, 0, 104, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 104, 0, 110, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 104, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 76, 0, 0, 0, 107, 0, 104, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 104, 0, 119, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 107, 0, 104, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 105, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 105, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 82, 0, 0, 0, 107, 0, 105, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 106, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 106, 0, 103, 0, 95, 0, 76, 0, 97, 0, 111, 0, 111, 0, 95, 0, 76, 0, 65, 0, 0, 0, 107, 0, 106, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 106, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 107, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 107, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 107, 0, 108, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 107, 0, 108, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 108, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 108, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 109, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 79, 0, 0, 0, 107, 0, 109, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 109, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 109, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 109, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 109, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 110, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 87, 0, 0, 0, 107, 0, 110, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 111, 0, 105, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 107, 0, 111, 0, 107, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 111, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 111, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 77, 0, 0, 0, 107, 0, 111, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 112, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 76, 0, 82, 0, 0, 0, 107, 0, 112, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 112, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 112, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 112, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 113, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 113, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 113, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 113, 0, 121, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 114, 0, 99, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 107, 0, 114, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 76, 0, 0, 0, 107, 0, 114, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 107, 0, 114, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 114, 0, 117, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 115, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 107, 0, 115, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 115, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 107, 0, 115, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 107, 0, 115, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 116, 0, 98, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 116, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 116, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 116, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 89, 0, 0, 0, 107, 0, 117, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 117, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 117, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 117, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 117, 0, 109, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 107, 0, 117, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 117, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 117, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 118, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 118, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 107, 0, 118, 0, 120, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 107, 0, 119, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 119, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 119, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 120, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 120, 0, 99, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 120, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 120, 0, 109, 0, 95, 0, 84, 0, 104, 0, 97, 0, 105, 0, 95, 0, 84, 0, 72, 0, 0, 0, 107, 0, 120, 0, 112, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 107, 0, 120, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 120, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 121, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 121, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 122, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 89, 0, 0, 0, 107, 0, 122, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 107, 0, 122, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 89, 0, 0, 0, 108, 0, 97, 0, 100, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 95, 0, 73, 0, 76, 0, 0, 0, 108, 0, 97, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 108, 0, 97, 0, 104, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 108, 0, 97, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 108, 0, 97, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 98, 0, 101, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 108, 0, 98, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 98, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 108, 0, 99, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 100, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 101, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 101, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 101, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 101, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 101, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 101, 0, 122, 0, 95, 0, 65, 0, 103, 0, 104, 0, 98, 0, 95, 0, 82, 0, 85, 0, 0, 0, 108, 0, 101, 0, 122, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 108, 0, 103, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 105, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 105, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 105, 0, 102, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 108, 0, 105, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 105, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 105, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 108, 0, 106, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 108, 0, 107, 0, 105, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 108, 0, 107, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 108, 0, 108, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 108, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 109, 0, 110, 0, 95, 0, 84, 0, 101, 0, 108, 0, 117, 0, 95, 0, 73, 0, 78, 0, 0, 0, 108, 0, 109, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 108, 0, 109, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 110, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 110, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 111, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 111, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 111, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 68, 0, 0, 0, 108, 0, 111, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 111, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 108, 0, 111, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 77, 0, 0, 0, 108, 0, 114, 0, 99, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 108, 0, 116, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 76, 0, 86, 0, 0, 0, 108, 0, 117, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 68, 0, 0, 0, 108, 0, 117, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 108, 0, 117, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 108, 0, 117, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 108, 0, 119, 0, 108, 0, 95, 0, 84, 0, 104, 0, 97, 0, 105, 0, 95, 0, 84, 0, 72, 0, 0, 0, 108, 0, 122, 0, 104, 0, 95, 0, 80, 0, 104, 0, 97, 0, 103, 0, 95, 0, 67, 0, 78, 0, 0, 0, 108, 0, 122, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 82, 0, 0, 0, 109, 0, 97, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 109, 0, 97, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 109, 0, 97, 0, 103, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 97, 0, 105, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 97, 0, 105, 0, 95, 0, 84, 0, 105, 0, 114, 0, 104, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 97, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 109, 0, 97, 0, 107, 0, 95, 0, 77, 0, 97, 0, 107, 0, 97, 0, 95, 0, 73, 0, 68, 0, 0, 0, 109, 0, 97, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 77, 0, 0, 0, 109, 0, 97, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 109, 0, 97, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 97, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 88, 0, 0, 0, 109, 0, 98, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 98, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 98, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 98, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 98, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 99, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 99, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 99, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 99, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 99, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 100, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 100, 0, 101, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 100, 0, 102, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 109, 0, 100, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 109, 0, 100, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 100, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 109, 0, 100, 0, 120, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 101, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 101, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 101, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 76, 0, 0, 0, 109, 0, 101, 0, 110, 0, 95, 0, 77, 0, 101, 0, 110, 0, 100, 0, 95, 0, 83, 0, 76, 0, 0, 0, 109, 0, 101, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 109, 0, 101, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 101, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 102, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 102, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 102, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 103, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 90, 0, 0, 0, 109, 0, 103, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 103, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 109, 0, 103, 0, 112, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 109, 0, 103, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 109, 0, 104, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 104, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 105, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 105, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 107, 0, 105, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 107, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 107, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 107, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 108, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 108, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 108, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 68, 0, 0, 0, 109, 0, 109, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 109, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 109, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 110, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 110, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 110, 0, 105, 0, 95, 0, 66, 0, 101, 0, 110, 0, 103, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 110, 0, 105, 0, 95, 0, 77, 0, 116, 0, 101, 0, 105, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 110, 0, 119, 0, 95, 0, 77, 0, 121, 0, 109, 0, 114, 0, 95, 0, 77, 0, 77, 0, 0, 0, 109, 0, 110, 0, 119, 0, 95, 0, 77, 0, 121, 0, 109, 0, 114, 0, 95, 0, 84, 0, 72, 0, 0, 0, 109, 0, 111, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 111, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 109, 0, 111, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 109, 0, 111, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 70, 0, 0, 0, 109, 0, 111, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 112, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 112, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 112, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 112, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 113, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 114, 0, 100, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 109, 0, 114, 0, 106, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 109, 0, 116, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 116, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 116, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 116, 0, 114, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 117, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 109, 0, 117, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 117, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 109, 0, 118, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 118, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 118, 0, 121, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 109, 0, 119, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 76, 0, 0, 0, 109, 0, 119, 0, 114, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 119, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 109, 0, 120, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 87, 0, 0, 0, 109, 0, 120, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 121, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 121, 0, 109, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 121, 0, 118, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 109, 0, 121, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 121, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 109, 0, 122, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 122, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 122, 0, 110, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 109, 0, 122, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 122, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 109, 0, 122, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 97, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 97, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 97, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 97, 0, 110, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 110, 0, 97, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 110, 0, 97, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 65, 0, 0, 0, 110, 0, 97, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 99, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 99, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 99, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 99, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 88, 0, 0, 0, 110, 0, 99, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 99, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 100, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 90, 0, 0, 0, 110, 0, 100, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 110, 0, 101, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 101, 0, 119, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 110, 0, 101, 0, 119, 0, 95, 0, 78, 0, 101, 0, 119, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 110, 0, 101, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 102, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 103, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 103, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 103, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 90, 0, 0, 0, 110, 0, 104, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 104, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 88, 0, 0, 0, 110, 0, 104, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 88, 0, 0, 0, 110, 0, 105, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 105, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 105, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 110, 0, 105, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 105, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 85, 0, 0, 0, 110, 0, 105, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 105, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 106, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 78, 0, 0, 0, 110, 0, 107, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 107, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 109, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 110, 0, 109, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 110, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 110, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 110, 0, 110, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 110, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 111, 0, 101, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 110, 0, 111, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 111, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 113, 0, 111, 0, 95, 0, 78, 0, 107, 0, 111, 0, 111, 0, 95, 0, 71, 0, 78, 0, 0, 0, 110, 0, 114, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 115, 0, 107, 0, 95, 0, 67, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 65, 0, 0, 0, 110, 0, 115, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 115, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 110, 0, 115, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 116, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 116, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 117, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 117, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 117, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 83, 0, 0, 0, 110, 0, 117, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 117, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 119, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 120, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 78, 0, 0, 0, 110, 0, 120, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 110, 0, 121, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 110, 0, 121, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 110, 0, 122, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 72, 0, 0, 0, 111, 0, 103, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 111, 0, 107, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 111, 0, 110, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 111, 0, 110, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 111, 0, 110, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 111, 0, 112, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 111, 0, 114, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 111, 0, 114, 0, 117, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 111, 0, 116, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 111, 0, 122, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 97, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 112, 0, 97, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 112, 0, 97, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 87, 0, 0, 0, 112, 0, 97, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 81, 0, 0, 0, 112, 0, 97, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 87, 0, 0, 0, 112, 0, 98, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 99, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 82, 0, 0, 0, 112, 0, 99, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 112, 0, 100, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 112, 0, 100, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 112, 0, 101, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 101, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 102, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 112, 0, 104, 0, 108, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 105, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 105, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 107, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 112, 0, 108, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 109, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 112, 0, 110, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 110, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 110, 0, 116, 0, 95, 0, 71, 0, 114, 0, 101, 0, 107, 0, 95, 0, 71, 0, 82, 0, 0, 0, 112, 0, 111, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 77, 0, 0, 0, 112, 0, 112, 0, 97, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 112, 0, 112, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 114, 0, 100, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 112, 0, 115, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 116, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 112, 0, 117, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 65, 0, 0, 0, 112, 0, 119, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 113, 0, 117, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 84, 0, 0, 0, 113, 0, 117, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 67, 0, 0, 0, 114, 0, 97, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 114, 0, 97, 0, 106, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 114, 0, 97, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 114, 0, 99, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 69, 0, 0, 0, 114, 0, 101, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 114, 0, 101, 0, 106, 0, 95, 0, 82, 0, 106, 0, 110, 0, 103, 0, 95, 0, 73, 0, 68, 0, 0, 0, 114, 0, 101, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 114, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 114, 0, 103, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 114, 0, 104, 0, 103, 0, 95, 0, 82, 0, 111, 0, 104, 0, 103, 0, 95, 0, 77, 0, 77, 0, 0, 0, 114, 0, 105, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 78, 0, 0, 0, 114, 0, 105, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 76, 0, 0, 0, 114, 0, 105, 0, 102, 0, 95, 0, 84, 0, 102, 0, 110, 0, 103, 0, 95, 0, 77, 0, 65, 0, 0, 0, 114, 0, 106, 0, 115, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 114, 0, 107, 0, 116, 0, 95, 0, 66, 0, 101, 0, 110, 0, 103, 0, 95, 0, 66, 0, 68, 0, 0, 0, 114, 0, 109, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 73, 0, 0, 0, 114, 0, 109, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 72, 0, 0, 0, 114, 0, 109, 0, 116, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 114, 0, 109, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 69, 0, 0, 0, 114, 0, 110, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 114, 0, 110, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 90, 0, 0, 0, 114, 0, 111, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 114, 0, 111, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 114, 0, 111, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 114, 0, 114, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 114, 0, 116, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 74, 0, 0, 0, 114, 0, 117, 0, 101, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 85, 0, 65, 0, 0, 0, 114, 0, 117, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 66, 0, 0, 0, 114, 0, 119, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 114, 0, 119, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 114, 0, 121, 0, 117, 0, 95, 0, 75, 0, 97, 0, 110, 0, 97, 0, 95, 0, 74, 0, 80, 0, 0, 0, 115, 0, 97, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 72, 0, 0, 0, 115, 0, 97, 0, 104, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 115, 0, 97, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 115, 0, 97, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 115, 0, 97, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 97, 0, 116, 0, 95, 0, 79, 0, 108, 0, 99, 0, 107, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 97, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 78, 0, 0, 0, 115, 0, 98, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 98, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 98, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 115, 0, 99, 0, 107, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 99, 0, 108, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 99, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 115, 0, 99, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 66, 0, 0, 0, 115, 0, 99, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 115, 0, 100, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 115, 0, 100, 0, 104, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 115, 0, 101, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 73, 0, 0, 0, 115, 0, 101, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 90, 0, 0, 0, 115, 0, 101, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 88, 0, 0, 0, 115, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 76, 0, 0, 0, 115, 0, 103, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 76, 0, 84, 0, 0, 0, 115, 0, 103, 0, 119, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 103, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 104, 0, 105, 0, 95, 0, 84, 0, 102, 0, 110, 0, 103, 0, 95, 0, 77, 0, 65, 0, 0, 0, 115, 0, 104, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 104, 0, 110, 0, 95, 0, 77, 0, 121, 0, 109, 0, 114, 0, 95, 0, 77, 0, 77, 0, 0, 0, 115, 0, 104, 0, 117, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 105, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 84, 0, 0, 0, 115, 0, 105, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 105, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 105, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 106, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 107, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 107, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 115, 0, 107, 0, 114, 0, 95, 0, 77, 0, 117, 0, 108, 0, 116, 0, 95, 0, 80, 0, 75, 0, 0, 0, 115, 0, 107, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 108, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 108, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 76, 0, 0, 0, 115, 0, 108, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 108, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 115, 0, 109, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 69, 0, 0, 0, 115, 0, 109, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 69, 0, 0, 0, 115, 0, 109, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 73, 0, 0, 0, 115, 0, 109, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 109, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 73, 0, 0, 0, 115, 0, 110, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 110, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 76, 0, 0, 0, 115, 0, 110, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 110, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 110, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 111, 0, 103, 0, 95, 0, 83, 0, 111, 0, 103, 0, 111, 0, 95, 0, 85, 0, 90, 0, 0, 0, 115, 0, 111, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 111, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 111, 0, 117, 0, 95, 0, 84, 0, 104, 0, 97, 0, 105, 0, 95, 0, 84, 0, 72, 0, 0, 0, 115, 0, 111, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 112, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 112, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 112, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 114, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 82, 0, 0, 0, 115, 0, 114, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 78, 0, 0, 0, 115, 0, 114, 0, 120, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 115, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 115, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 115, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 82, 0, 0, 0, 115, 0, 116, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 116, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 115, 0, 117, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 117, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 117, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 115, 0, 117, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 117, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 78, 0, 0, 0, 115, 0, 119, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 68, 0, 0, 0, 115, 0, 119, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 115, 0, 119, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 119, 0, 118, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 120, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 115, 0, 120, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 115, 0, 121, 0, 108, 0, 95, 0, 66, 0, 101, 0, 110, 0, 103, 0, 95, 0, 66, 0, 68, 0, 0, 0, 115, 0, 121, 0, 108, 0, 95, 0, 83, 0, 121, 0, 108, 0, 111, 0, 95, 0, 66, 0, 68, 0, 0, 0, 115, 0, 122, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 76, 0, 0, 0, 116, 0, 97, 0, 106, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 116, 0, 97, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 97, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 97, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 98, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 98, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 98, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 98, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 98, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 98, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 116, 0, 98, 0, 119, 0, 95, 0, 84, 0, 97, 0, 103, 0, 98, 0, 95, 0, 80, 0, 72, 0, 0, 0, 116, 0, 98, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 99, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 99, 0, 121, 0, 95, 0, 75, 0, 110, 0, 100, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 116, 0, 100, 0, 103, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 116, 0, 100, 0, 104, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 116, 0, 100, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 89, 0, 0, 0, 116, 0, 101, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 101, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 76, 0, 0, 0, 116, 0, 101, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 116, 0, 101, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 76, 0, 0, 0, 116, 0, 102, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 103, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 103, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 103, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 104, 0, 108, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 116, 0, 104, 0, 113, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 116, 0, 104, 0, 114, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 116, 0, 105, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 105, 0, 103, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 69, 0, 82, 0, 0, 0, 116, 0, 105, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 105, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 105, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 105, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 116, 0, 107, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 90, 0, 0, 0, 116, 0, 107, 0, 116, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 116, 0, 108, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 108, 0, 120, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 108, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 90, 0, 0, 0, 116, 0, 109, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 69, 0, 0, 0, 116, 0, 109, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 110, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 111, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 111, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 87, 0, 0, 0, 116, 0, 111, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 112, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 112, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 113, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 114, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 82, 0, 0, 0, 116, 0, 114, 0, 119, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 115, 0, 100, 0, 95, 0, 71, 0, 114, 0, 101, 0, 107, 0, 95, 0, 71, 0, 82, 0, 0, 0, 116, 0, 115, 0, 102, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 116, 0, 115, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 116, 0, 115, 0, 106, 0, 95, 0, 84, 0, 105, 0, 98, 0, 116, 0, 95, 0, 66, 0, 84, 0, 0, 0, 116, 0, 115, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 116, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 116, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 116, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 116, 0, 116, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 116, 0, 115, 0, 95, 0, 84, 0, 104, 0, 97, 0, 105, 0, 95, 0, 84, 0, 72, 0, 0, 0, 116, 0, 116, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 90, 0, 0, 0, 116, 0, 117, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 117, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 117, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 87, 0, 0, 0, 116, 0, 117, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 118, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 118, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 119, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 119, 0, 113, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 69, 0, 0, 0, 116, 0, 121, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 116, 0, 121, 0, 118, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 116, 0, 122, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 65, 0, 0, 0, 117, 0, 98, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 117, 0, 100, 0, 109, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 85, 0, 0, 0, 117, 0, 108, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 77, 0, 0, 0, 117, 0, 109, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 79, 0, 0, 0, 117, 0, 110, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 81, 0, 0, 0, 117, 0, 110, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 86, 0, 0, 0, 117, 0, 110, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 80, 0, 0, 0, 117, 0, 110, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 71, 0, 83, 0, 0, 0, 117, 0, 110, 0, 100, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 72, 0, 77, 0, 0, 0, 117, 0, 110, 0, 114, 0, 95, 0, 66, 0, 101, 0, 110, 0, 103, 0, 95, 0, 73, 0, 78, 0, 0, 0, 117, 0, 110, 0, 120, 0, 95, 0, 66, 0, 101, 0, 110, 0, 103, 0, 95, 0, 73, 0, 78, 0, 0, 0, 117, 0, 111, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 117, 0, 114, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 117, 0, 114, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 117, 0, 114, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 117, 0, 115, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 117, 0, 116, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 117, 0, 118, 0, 104, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 117, 0, 118, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 118, 0, 97, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 118, 0, 97, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 118, 0, 101, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 118, 0, 101, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 85, 0, 0, 0, 118, 0, 105, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 88, 0, 0, 0, 118, 0, 105, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 118, 0, 108, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 69, 0, 0, 0, 118, 0, 109, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 118, 0, 109, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 90, 0, 0, 0, 118, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 118, 0, 111, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 85, 0, 0, 0, 118, 0, 114, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 69, 0, 69, 0, 0, 0, 118, 0, 117, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 118, 0, 117, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 97, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 72, 0, 0, 0, 119, 0, 97, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 97, 0, 108, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 95, 0, 69, 0, 84, 0, 0, 0, 119, 0, 97, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 97, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 119, 0, 98, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 85, 0, 0, 0, 119, 0, 98, 0, 113, 0, 95, 0, 84, 0, 101, 0, 108, 0, 117, 0, 95, 0, 73, 0, 78, 0, 0, 0, 119, 0, 98, 0, 114, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 119, 0, 99, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 101, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 103, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 104, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 105, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 105, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 105, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 106, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 106, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 108, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 87, 0, 70, 0, 0, 0, 119, 0, 109, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 110, 0, 99, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 110, 0, 105, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 75, 0, 77, 0, 0, 0, 119, 0, 110, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 111, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 111, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 114, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 115, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 116, 0, 109, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 119, 0, 117, 0, 117, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 119, 0, 117, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 119, 0, 119, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 120, 0, 97, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 82, 0, 0, 0, 120, 0, 98, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 120, 0, 101, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 120, 0, 108, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 120, 0, 109, 0, 102, 0, 95, 0, 71, 0, 101, 0, 111, 0, 114, 0, 95, 0, 71, 0, 69, 0, 0, 0, 120, 0, 109, 0, 114, 0, 95, 0, 77, 0, 101, 0, 114, 0, 111, 0, 95, 0, 83, 0, 68, 0, 0, 0, 120, 0, 110, 0, 114, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 120, 0, 111, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 120, 0, 111, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 120, 0, 114, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 120, 0, 115, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 120, 0, 115, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 120, 0, 115, 0, 114, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 95, 0, 78, 0, 80, 0, 0, 0, 120, 0, 119, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 97, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 97, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 90, 0, 0, 0, 121, 0, 97, 0, 112, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 70, 0, 77, 0, 0, 0, 121, 0, 97, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 97, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 97, 0, 118, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 121, 0, 97, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 97, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 98, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 98, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 121, 0, 98, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 101, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 103, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 103, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 105, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 121, 0, 107, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 108, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 108, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 108, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 109, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 111, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 114, 0, 98, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 114, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 114, 0, 108, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 82, 0, 0, 0, 121, 0, 115, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 117, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 88, 0, 0, 0, 121, 0, 117, 0, 101, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 72, 0, 75, 0, 0, 0, 121, 0, 117, 0, 106, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 117, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 121, 0, 117, 0, 119, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 122, 0, 97, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 68, 0, 0, 0, 122, 0, 100, 0, 106, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 75, 0, 77, 0, 0, 0, 122, 0, 101, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 78, 0, 76, 0, 0, 0, 122, 0, 105, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 122, 0, 108, 0, 109, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 71, 0, 0, 0, 122, 0, 109, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 89, 0, 0, 0, 122, 0, 110, 0, 101, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 90, 0, 90, 0, 0, 0, 122, 0, 122, 0, 97, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 84, 0, 82, 0, 0, 0, 106, 0, 98, 0, 111, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 112, 0, 114, 0, 103, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 61, 7, 106, 7, 32, 0, 36, 0, 40, 0, 98, 32, 44, 0, 48, 0, 52, 0, 56, 0, 60, 0, 64, 0, 68, 0, 72, 0, 76, 0, 80, 0, 84, 0, 88, 0, 92, 0, 96, 0, 173, 32, 100, 0, 104, 0, 149, 14, 108, 0, 112, 0, 116, 0, 120, 0, 124, 0, 128, 0, 132, 0, 136, 0, 140, 0, 144, 0, 66, 12, 148, 0, 152, 0, 241, 24, 156, 0, 160, 0, 129, 33, 164, 0, 168, 0, 158, 24, 172, 0, 185, 32, 176, 0, 180, 0, 184, 0, 188, 0, 192, 0, 196, 0, 200, 0, 204, 0, 208, 0, 212, 0, 216, 0, 220, 0, 224, 0, 189, 32, 95, 27, 228, 0, 237, 0, 246, 0, 250, 0, 254, 0, 2, 1, 6, 1, 10, 1, 14, 1, 141, 33, 18, 1, 22, 1, 26, 1, 30, 1, 34, 1, 38, 1, 42, 1, 46, 1, 50, 1, 149, 33, 54, 1, 58, 1, 9, 31, 62, 1, 66, 1, 70, 1, 74, 1, 78, 1, 153, 33, 82, 1, 157, 33, 86, 1, 94, 1, 100, 1, 106, 1, 161, 33, 112, 1, 116, 1, 120, 1, 124, 1, 128, 1, 132, 1, 136, 1, 140, 1, 144, 1, 148, 1, 152, 1, 156, 1, 160, 1, 164, 1, 168, 1, 172, 1, 176, 1, 180, 1, 184, 1, 188, 1, 192, 1, 196, 1, 200, 1, 181, 14, 204, 1, 208, 1, 212, 1, 216, 1, 220, 1, 224, 1, 228, 1, 232, 1, 236, 1, 240, 1, 244, 1, 248, 1, 46, 16, 252, 1, 0, 2, 4, 2, 8, 2, 12, 2, 16, 2, 20, 2, 24, 2, 28, 2, 45, 33, 32, 2, 36, 2, 40, 2, 44, 2, 48, 2, 52, 2, 56, 2, 60, 2, 64, 2, 68, 2, 72, 2, 76, 2, 80, 2, 84, 2, 88, 2, 92, 2, 96, 2, 100, 2, 104, 2, 108, 2, 112, 2, 142, 7, 116, 2, 120, 2, 124, 2, 128, 2, 132, 2, 135, 2, 139, 2, 143, 2, 50, 16, 147, 2, 151, 2, 155, 2, 159, 2, 163, 2, 167, 2, 171, 2, 175, 2, 201, 32, 179, 2, 183, 2, 187, 2, 191, 2, 195, 2, 198, 2, 202, 2, 206, 2, 210, 2, 214, 2, 218, 2, 222, 2, 226, 2, 230, 2, 234, 2, 238, 2, 242, 2, 246, 2, 250, 2, 254, 2, 2, 3, 6, 3, 10, 3, 14, 3, 18, 3, 22, 3, 26, 3, 30, 3, 34, 3, 38, 3, 42, 3, 46, 3, 50, 3, 54, 3, 58, 3, 62, 3, 86, 12, 66, 3, 70, 3, 74, 3, 78, 3, 90, 12, 82, 3, 86, 3, 90, 3, 98, 12, 94, 3, 98, 3, 214, 31, 102, 3, 106, 3, 110, 3, 114, 3, 118, 3, 122, 3, 126, 3, 130, 3, 134, 3, 138, 3, 142, 3, 146, 3, 150, 3, 49, 33, 154, 3, 158, 3, 53, 33, 162, 3, 166, 3, 170, 3, 174, 3, 178, 3, 182, 3, 208, 14, 186, 3, 190, 3, 194, 3, 106, 12, 198, 3, 206, 3, 66, 16, 40, 25, 209, 3, 213, 3, 217, 3, 221, 3, 225, 3, 229, 3, 233, 3, 237, 3, 241, 3, 245, 3, 249, 3, 199, 10, 253, 3, 1, 4, 5, 4, 9, 4, 13, 4, 17, 4, 21, 4, 25, 4, 29, 4, 33, 4, 37, 4, 41, 4, 45, 4, 49, 4, 53, 4, 57, 4, 61, 4, 65, 4, 69, 4, 73, 4, 77, 4, 81, 4, 85, 4, 89, 4, 93, 4, 97, 4, 101, 4, 105, 4, 108, 4, 112, 4, 116, 4, 120, 4, 97, 0, 124, 4, 128, 4, 227, 10, 132, 4, 136, 4, 140, 4, 144, 4, 148, 4, 50, 14, 152, 4, 156, 4, 235, 10, 160, 4, 168, 4, 172, 4, 94, 16, 176, 4, 57, 33, 180, 4, 184, 4, 98, 16, 188, 4, 192, 4, 196, 4, 200, 4, 247, 10, 204, 4, 208, 4, 251, 10, 212, 4, 216, 4, 220, 4, 224, 4, 228, 4, 232, 4, 235, 4, 243, 4, 247, 4, 102, 16, 251, 4, 255, 4, 3, 5, 7, 5, 10, 5, 14, 5, 7, 11, 18, 5, 22, 5, 26, 5, 30, 5, 34, 5, 130, 12, 38, 5, 42, 5, 46, 5, 50, 5, 54, 5, 58, 5, 62, 5, 66, 5, 70, 5, 74, 5, 78, 5, 82, 5, 86, 5, 90, 5, 214, 7, 115, 17, 94, 5, 98, 5, 102, 5, 106, 5, 110, 5, 81, 28, 114, 5, 118, 5, 122, 5, 126, 5, 130, 5, 134, 5, 138, 5, 142, 5, 146, 5, 51, 30, 150, 5, 154, 5, 158, 5, 162, 5, 166, 5, 170, 5, 174, 5, 178, 5, 182, 5, 186, 5, 190, 5, 194, 5, 198, 5, 202, 5, 206, 5, 210, 5, 214, 5, 142, 12, 218, 5, 222, 5, 226, 5, 58, 14, 230, 5, 234, 5, 238, 5, 242, 5, 246, 5, 250, 5, 254, 5, 2, 6, 6, 6, 10, 6, 14, 6, 18, 6, 22, 6, 31, 6, 35, 6, 39, 6, 128, 16, 43, 6, 47, 6, 51, 6, 55, 6, 59, 6, 63, 6, 67, 6, 71, 6, 74, 6, 78, 6, 82, 6, 86, 6, 90, 6, 94, 6, 98, 6, 230, 7, 102, 6, 108, 6, 253, 28, 114, 6, 43, 21, 164, 4, 118, 6, 122, 6, 126, 6, 150, 12, 130, 6, 58, 25, 134, 6, 138, 6, 142, 6, 146, 6, 150, 6, 154, 6, 158, 6, 162, 6, 166, 6, 170, 6, 174, 6, 178, 6, 182, 6, 186, 6, 215, 31, 190, 6, 223, 14, 194, 6, 140, 16, 198, 6, 202, 6, 250, 7, 66, 28, 206, 6, 131, 6, 2, 8, 141, 34, 210, 6, 214, 6, 218, 6, 222, 6, 226, 6, 230, 6, 234, 6, 32, 15, 23, 30, 238, 6, 242, 6, 246, 6, 148, 16, 250, 6, 254, 6, 162, 12, 2, 7, 152, 16, 6, 7, 10, 7, 14, 7, 18, 7, 22, 7, 26, 7, 170, 12, 41, 30, 160, 16, 30, 7, 34, 7, 51, 11, 4, 5, 225, 32, 64, 11, 38, 7, 42, 7, 46, 7, 50, 7, 233, 32, 54, 7, 58, 7, 62, 7, 66, 7, 70, 7, 74, 7, 78, 7, 237, 32, 82, 7, 86, 7, 90, 7, 94, 7, 98, 7, 123, 3, 102, 7, 211, 13, 105, 7, 109, 7, 113, 7, 117, 7, 121, 7, 125, 7, 129, 7, 133, 7, 137, 7, 141, 7, 145, 7, 149, 7, 153, 7, 157, 7, 161, 7, 165, 7, 169, 7, 173, 7, 177, 7, 181, 7, 185, 7, 189, 7, 193, 7, 197, 7, 201, 7, 205, 7, 209, 7, 213, 7, 190, 12, 217, 7, 221, 7, 225, 7, 229, 7, 233, 7, 237, 7, 241, 7, 245, 7, 249, 7, 253, 7, 1, 8, 68, 11, 5, 8, 9, 8, 13, 8, 70, 8, 17, 8, 21, 8, 25, 8, 29, 8, 7, 7, 33, 8, 39, 8, 47, 8, 53, 8, 59, 8, 65, 8, 69, 8, 168, 16, 73, 8, 77, 8, 81, 8, 85, 8, 97, 2, 89, 8, 93, 8, 97, 8, 101, 8, 105, 8, 109, 8, 211, 5, 113, 8, 117, 8, 185, 33, 121, 8, 125, 8, 129, 8, 133, 8, 137, 8, 141, 8, 145, 8, 149, 8, 153, 8, 157, 8, 161, 8, 165, 8, 169, 8, 173, 8, 172, 16, 177, 8, 181, 8, 185, 8, 189, 8, 193, 8, 197, 8, 60, 15, 201, 8, 205, 8, 209, 8, 213, 8, 217, 8, 221, 8, 225, 8, 229, 8, 233, 8, 237, 8, 105, 2, 241, 8, 249, 8, 255, 8, 7, 9, 11, 9, 15, 9, 19, 9, 23, 9, 27, 9, 31, 9, 35, 9, 80, 13, 39, 9, 43, 9, 47, 9, 80, 11, 51, 9, 55, 9, 59, 9, 63, 9, 67, 9, 71, 9, 75, 9, 79, 9, 83, 9, 87, 9, 149, 4, 91, 9, 99, 9, 105, 9, 113, 9, 119, 9, 123, 9, 127, 9, 131, 9, 135, 9, 61, 33, 139, 9, 143, 9, 202, 3, 147, 9, 151, 9, 155, 9, 159, 9, 162, 9, 166, 9, 170, 9, 174, 9, 178, 9, 182, 9, 186, 9, 190, 9, 194, 9, 198, 9, 202, 9, 206, 9, 210, 9, 193, 33, 214, 9, 119, 17, 218, 9, 222, 9, 226, 9, 230, 9, 239, 9, 243, 9, 247, 9, 251, 9, 255, 9, 3, 10, 7, 10, 11, 10, 15, 10, 19, 10, 23, 10, 27, 10, 16, 10, 31, 10, 35, 10, 96, 30, 39, 10, 43, 10, 47, 10, 51, 10, 55, 10, 59, 10, 63, 10, 186, 27, 67, 10, 159, 6, 71, 10, 75, 10, 79, 10, 83, 10, 87, 10, 90, 10, 94, 10, 98, 10, 102, 10, 106, 10, 110, 10, 114, 10, 118, 10, 122, 10, 126, 10, 133, 10, 142, 10, 146, 10, 150, 10, 154, 10, 158, 10, 162, 10, 166, 10, 170, 10, 174, 10, 178, 10, 182, 10, 186, 10, 190, 10, 194, 10, 198, 10, 202, 10, 206, 10, 210, 10, 214, 10, 218, 10, 222, 10, 226, 10, 230, 10, 234, 10, 238, 10, 242, 10, 246, 10, 250, 10, 254, 10, 2, 11, 6, 11, 10, 11, 201, 12, 14, 11, 18, 11, 22, 11, 26, 11, 30, 11, 195, 16, 34, 11, 38, 11, 153, 34, 42, 11, 46, 11, 50, 11, 54, 11, 63, 11, 121, 2, 67, 11, 71, 11, 75, 11, 79, 11, 201, 33, 83, 11, 87, 11, 91, 11, 95, 11, 99, 11, 103, 11, 77, 33, 107, 11, 113, 11, 121, 11, 125, 11, 129, 11, 133, 11, 245, 32, 137, 11, 141, 11, 145, 11, 149, 11, 153, 11, 157, 11, 161, 11, 165, 11, 169, 11, 173, 11, 195, 27, 177, 11, 181, 11, 185, 11, 99, 15, 189, 11, 195, 11, 97, 14, 201, 11, 205, 11, 209, 11, 213, 11, 217, 11, 221, 11, 225, 11, 229, 11, 233, 11, 237, 11, 241, 11, 245, 11, 249, 11, 253, 11, 1, 12, 5, 12, 199, 16, 9, 12, 13, 12, 17, 12, 21, 12, 25, 12, 29, 12, 33, 12, 37, 12, 41, 12, 45, 12, 49, 12, 53, 12, 85, 33, 57, 12, 61, 12, 65, 12, 69, 12, 73, 12, 77, 12, 81, 12, 147, 26, 85, 12, 89, 12, 93, 12, 97, 12, 101, 12, 105, 12, 87, 30, 109, 12, 113, 12, 157, 34, 117, 12, 121, 12, 125, 12, 129, 12, 71, 24, 133, 12, 137, 12, 141, 12, 145, 12, 149, 12, 153, 12, 157, 12, 161, 12, 165, 12, 169, 12, 173, 12, 177, 12, 181, 12, 185, 12, 189, 12, 193, 12, 197, 12, 200, 12, 204, 12, 231, 13, 208, 12, 212, 12, 216, 12, 220, 12, 224, 12, 187, 6, 228, 12, 232, 12, 236, 12, 240, 12, 244, 12, 248, 12, 89, 33, 252, 12, 0, 13, 4, 13, 8, 13, 12, 13, 16, 13, 20, 13, 24, 13, 28, 13, 32, 13, 36, 13, 40, 13, 44, 13, 47, 13, 51, 13, 55, 13, 119, 15, 59, 13, 63, 13, 67, 13, 191, 6, 71, 13, 75, 13, 79, 13, 225, 18, 117, 11, 83, 13, 87, 13, 91, 13, 52, 10, 95, 13, 99, 13, 9, 33, 103, 13, 107, 13, 111, 13, 115, 13, 243, 13, 119, 13, 127, 13, 133, 13, 137, 13, 141, 13, 150, 13, 154, 13, 158, 13, 162, 13, 166, 13, 170, 13, 174, 13, 178, 13, 182, 13, 230, 31, 186, 13, 190, 13, 194, 13, 198, 13, 202, 13, 206, 13, 210, 13, 214, 13, 60, 22, 218, 13, 222, 13, 226, 13, 230, 13, 234, 13, 238, 13, 242, 13, 246, 13, 250, 13, 254, 13, 2, 14, 151, 15, 6, 14, 71, 21, 10, 14, 14, 14, 18, 14, 22, 14, 25, 14, 29, 14, 133, 20, 33, 14, 37, 14, 41, 14, 45, 14, 49, 14, 53, 14, 57, 14, 61, 14, 65, 14, 69, 14, 73, 14, 80, 14, 84, 14, 245, 28, 88, 14, 92, 14, 96, 14, 100, 14, 183, 15, 104, 14, 108, 14, 161, 32, 112, 14, 116, 14, 167, 27, 120, 14, 124, 14, 108, 23, 128, 14, 132, 14, 68, 32, 136, 14, 140, 14, 144, 14, 109, 33, 148, 14, 152, 14, 156, 14, 160, 14, 164, 14, 168, 14, 172, 14, 176, 14, 180, 14, 184, 14, 188, 14, 191, 14, 195, 14, 199, 14, 203, 14, 207, 14, 247, 16, 211, 14, 219, 14, 227, 14, 235, 14, 239, 14, 23, 1, 243, 14, 247, 14, 251, 14, 255, 14, 17, 33, 3, 15, 7, 15, 11, 15, 15, 15, 19, 15, 23, 15, 27, 15, 65, 28, 113, 33, 31, 15, 35, 15, 39, 15, 43, 15, 47, 15, 21, 33, 51, 15, 55, 15, 59, 15, 63, 15, 66, 15, 70, 15, 74, 15, 78, 15, 117, 33, 126, 28, 82, 15, 86, 15, 90, 15, 94, 15, 98, 15, 5, 13, 102, 15, 106, 15, 110, 15, 114, 15, 118, 15, 9, 13, 122, 15, 126, 15, 130, 15, 134, 15, 138, 15, 142, 15, 146, 15, 150, 15, 203, 2, 121, 33, 154, 15, 160, 15, 166, 15, 172, 15, 178, 15, 182, 15, 186, 15, 190, 15, 217, 33, 194, 15, 198, 15, 202, 15, 197, 19, 206, 15, 210, 15, 156, 26, 214, 15, 218, 15, 222, 15, 226, 15, 230, 15, 234, 15, 7, 17, 237, 15, 241, 15, 245, 15, 249, 15, 253, 15, 1, 16, 5, 16, 9, 16, 13, 16, 17, 16, 108, 13, 21, 16, 135, 24, 25, 16, 29, 16, 33, 16, 37, 16, 41, 16, 45, 16, 49, 16, 53, 16, 57, 16, 61, 16, 65, 16, 69, 16, 73, 16, 77, 16, 81, 16, 15, 17, 85, 16, 89, 16, 93, 16, 97, 16, 101, 16, 68, 10, 105, 16, 113, 16, 119, 16, 123, 16, 127, 16, 60, 23, 131, 16, 135, 16, 139, 16, 43, 29, 143, 16, 147, 16, 151, 16, 155, 16, 159, 16, 163, 16, 207, 15, 167, 16, 171, 16, 175, 16, 179, 16, 182, 16, 186, 16, 190, 16, 194, 16, 198, 16, 110, 9, 202, 16, 234, 8, 206, 16, 210, 16, 214, 16, 218, 16, 222, 16, 226, 16, 230, 16, 76, 32, 234, 16, 238, 16, 242, 16, 27, 17, 246, 16, 250, 16, 254, 16, 2, 17, 6, 17, 31, 17, 10, 17, 14, 17, 18, 17, 22, 17, 26, 17, 30, 17, 34, 17, 38, 17, 42, 17, 46, 17, 50, 17, 54, 17, 58, 17, 62, 17, 66, 17, 70, 17, 86, 4, 74, 17, 78, 17, 82, 17, 86, 17, 90, 17, 133, 14, 94, 17, 102, 17, 108, 17, 114, 17, 223, 15, 118, 17, 122, 17, 86, 30, 126, 17, 134, 17, 142, 17, 150, 17, 158, 17, 166, 17, 174, 17, 182, 17, 190, 17, 198, 17, 206, 17, 214, 17, 222, 17, 230, 17, 238, 17, 246, 17, 254, 17, 6, 18, 14, 18, 22, 18, 30, 18, 38, 18, 46, 18, 54, 18, 62, 18, 70, 18, 78, 18, 86, 18, 94, 18, 102, 18, 110, 18, 117, 18, 124, 18, 131, 18, 138, 18, 145, 18, 152, 18, 159, 18, 166, 18, 173, 18, 180, 18, 187, 18, 194, 18, 201, 18, 210, 18, 219, 18, 228, 18, 237, 18, 249, 18, 5, 19, 17, 19, 29, 19, 41, 19, 53, 19, 65, 19, 77, 19, 89, 19, 101, 19, 113, 19, 125, 19, 137, 19, 149, 19, 161, 19, 173, 19, 182, 19, 191, 19, 200, 19, 207, 19, 214, 19, 221, 19, 228, 19, 235, 19, 242, 19, 249, 19, 0, 20, 7, 20, 14, 20, 21, 20, 28, 20, 35, 20, 42, 20, 49, 20, 56, 20, 65, 20, 74, 20, 83, 20, 92, 20, 101, 20, 110, 20, 119, 20, 128, 20, 137, 20, 146, 20, 155, 20, 162, 20, 169, 20, 176, 20, 183, 20, 190, 20, 197, 20, 204, 20, 211, 20, 218, 20, 225, 20, 232, 20, 239, 20, 246, 20, 253, 20, 4, 21, 11, 21, 20, 21, 29, 21, 38, 21, 47, 21, 56, 21, 65, 21, 74, 21, 83, 21, 92, 21, 104, 21, 116, 21, 128, 21, 140, 21, 152, 21, 164, 21, 176, 21, 188, 21, 200, 21, 207, 21, 214, 21, 221, 21, 228, 21, 235, 21, 244, 21, 0, 22, 12, 22, 24, 22, 36, 22, 45, 22, 54, 22, 63, 22, 70, 22, 77, 22, 84, 22, 91, 22, 98, 22, 105, 22, 112, 22, 119, 22, 126, 22, 133, 22, 142, 22, 151, 22, 160, 22, 169, 22, 176, 22, 183, 22, 190, 22, 197, 22, 204, 22, 211, 22, 218, 22, 225, 22, 232, 22, 239, 22, 246, 22, 253, 22, 4, 23, 11, 23, 18, 23, 27, 23, 36, 23, 45, 23, 54, 23, 63, 23, 72, 23, 81, 23, 93, 23, 102, 23, 111, 23, 118, 23, 125, 23, 132, 23, 139, 23, 146, 23, 153, 23, 162, 23, 171, 23, 180, 23, 189, 23, 198, 23, 207, 23, 216, 23, 225, 23, 237, 23, 249, 23, 5, 24, 17, 24, 29, 24, 38, 24, 47, 24, 56, 24, 65, 24, 74, 24, 81, 24, 88, 24, 95, 24, 102, 24, 109, 24, 116, 24, 123, 24, 130, 24, 139, 24, 146, 24, 153, 24, 162, 24, 171, 24, 180, 24, 187, 24, 194, 24, 201, 24, 208, 24, 215, 24, 222, 24, 229, 24, 236, 24, 245, 24, 254, 24, 7, 25, 16, 25, 25, 25, 34, 25, 43, 25, 52, 25, 61, 25, 68, 25, 75, 25, 82, 25, 89, 25, 96, 25, 103, 25, 110, 25, 117, 25, 124, 25, 133, 25, 142, 25, 154, 25, 166, 25, 178, 25, 190, 25, 202, 25, 214, 25, 226, 25, 238, 25, 250, 25, 6, 26, 18, 26, 30, 26, 42, 26, 54, 26, 66, 26, 78, 26, 90, 26, 102, 26, 114, 26, 123, 26, 132, 26, 141, 26, 150, 26, 159, 26, 168, 26, 177, 26, 184, 26, 191, 26, 198, 26, 205, 26, 212, 26, 219, 26, 226, 26, 233, 26, 240, 26, 247, 26, 254, 26, 5, 27, 12, 27, 19, 27, 26, 27, 33, 27, 40, 27, 47, 27, 54, 27, 63, 27, 72, 27, 81, 27, 90, 27, 99, 27, 108, 27, 117, 27, 126, 27, 135, 27, 144, 27, 153, 27, 162, 27, 171, 27, 180, 27, 189, 27, 198, 27, 210, 27, 222, 27, 229, 27, 236, 27, 243, 27, 250, 27, 1, 28, 8, 28, 15, 28, 24, 28, 33, 28, 42, 28, 51, 28, 60, 28, 69, 28, 76, 28, 85, 28, 94, 28, 103, 28, 112, 28, 121, 28, 130, 28, 137, 28, 144, 28, 151, 28, 158, 28, 165, 28, 172, 28, 179, 28, 186, 28, 193, 28, 200, 28, 207, 28, 214, 28, 221, 28, 230, 28, 239, 28, 248, 28, 1, 29, 10, 29, 19, 29, 28, 29, 37, 29, 46, 29, 53, 29, 60, 29, 67, 29, 74, 29, 81, 29, 88, 29, 95, 29, 104, 29, 113, 29, 122, 29, 129, 29, 136, 29, 143, 29, 150, 29, 157, 29, 164, 29, 171, 29, 178, 29, 185, 29, 192, 29, 199, 29, 206, 29, 213, 29, 220, 29, 229, 29, 238, 29, 247, 29, 0, 30, 9, 30, 18, 30, 27, 30, 36, 30, 45, 30, 54, 30, 63, 30, 72, 30, 81, 30, 90, 30, 99, 30, 108, 30, 115, 30, 122, 30, 129, 30, 136, 30, 143, 30, 150, 30, 157, 30, 164, 30, 171, 30, 178, 30, 185, 30, 192, 30, 199, 30, 206, 30, 215, 30, 224, 30, 233, 30, 242, 30, 251, 30, 4, 31, 13, 31, 22, 31, 31, 31, 40, 31, 49, 31, 58, 31, 70, 31, 82, 31, 94, 31, 103, 31, 112, 31, 119, 31, 126, 31, 133, 31, 140, 31, 149, 31, 156, 31, 163, 31, 170, 31, 177, 31, 186, 31, 193, 31, 200, 31, 209, 31, 218, 31, 225, 31, 234, 31, 243, 31, 250, 31, 1, 32, 10, 32, 19, 32, 26, 32, 118, 29, 35, 32, 44, 32, 51, 32, 55, 32, 227, 15, 59, 32, 63, 32, 67, 32, 71, 32, 75, 32, 79, 32, 83, 32, 84, 10, 87, 32, 93, 32, 101, 32, 107, 32, 111, 32, 115, 32, 119, 32, 122, 32, 126, 32, 130, 32, 133, 32, 137, 32, 141, 32, 145, 32, 149, 32, 153, 32, 156, 32, 160, 32, 164, 32, 168, 32, 37, 33, 172, 32, 176, 32, 180, 32, 184, 32, 188, 32, 192, 32, 196, 32, 200, 32, 204, 32, 208, 32, 212, 32, 216, 32, 220, 32, 224, 32, 228, 32, 232, 32, 236, 32, 240, 32, 244, 32, 248, 32, 252, 32, 0, 33, 141, 14, 4, 33, 8, 33, 12, 33, 16, 33, 20, 33, 24, 33, 28, 33, 32, 33, 36, 33, 40, 33, 44, 33, 48, 33, 52, 33, 56, 33, 19, 3, 60, 33, 64, 33, 68, 33, 72, 33, 76, 33, 80, 33, 84, 33, 88, 33, 92, 33, 96, 33, 100, 33, 104, 33, 108, 33, 112, 33, 116, 33, 120, 33, 124, 33, 128, 33, 132, 33, 136, 33, 140, 33, 144, 33, 148, 33, 152, 33, 156, 33, 160, 33, 164, 33, 168, 33, 172, 33, 176, 33, 180, 33, 99, 6, 184, 33, 188, 33, 192, 33, 196, 33, 200, 33, 78, 30, 204, 33, 208, 33, 212, 33, 98, 17, 216, 33, 220, 33, 224, 33, 228, 33, 235, 33, 244, 33, 248, 33, 252, 33, 164, 34, 0, 34, 4, 34, 8, 34, 12, 34, 95, 10, 16, 34, 22, 34, 28, 34, 36, 34, 42, 34, 48, 34, 54, 34, 62, 34, 70, 34, 76, 34, 82, 34, 88, 34, 94, 34, 100, 34, 106, 34, 112, 34, 118, 34, 124, 34, 130, 34, 136, 34, 140, 34, 144, 34, 148, 34, 152, 34, 156, 34, 160, 34, 163, 34, 52, 5, 54, 21, 66, 21, 78, 21, 53, 1, 90, 21, 102, 21, 114, 21, 126, 21, 138, 21, 150, 21, 162, 21, 174, 21, 186, 21, 198, 21, 210, 21, 222, 21, 234, 21, 246, 21, 64, 1, 2, 22, 14, 22, 74, 5, 26, 22, 38, 22, 50, 22, 62, 22, 74, 22, 86, 22, 98, 22, 110, 22, 242, 16, 122, 22, 75, 1, 254, 16, 134, 22, 146, 22, 158, 22, 170, 22, 45, 0, 182, 22, 194, 22, 206, 22, 218, 22, 85, 5, 230, 22, 242, 22, 254, 22, 10, 23, 22, 23, 34, 23, 46, 23, 58, 23, 10, 17, 70, 23, 82, 23, 94, 23, 106, 23, 23, 0, 22, 17, 34, 17, 46, 17, 130, 23, 142, 23, 154, 23, 166, 23, 178, 23, 190, 23, 202, 23, 49, 6, 214, 23, 58, 17, 226, 23, 238, 23, 250, 23, 6, 24, 18, 24, 30, 24, 42, 24, 60, 6, 54, 24, 66, 24, 78, 24, 90, 24, 102, 24, 114, 24, 126, 24, 138, 24, 71, 6, 150, 24, 108, 1, 97, 1, 82, 6, 97, 1, 104, 6, 115, 6, 162, 24, 186, 24, 198, 24, 210, 24, 222, 24, 234, 24, 70, 17, 246, 24, 2, 25, 26, 25, 38, 25, 50, 25, 62, 25, 74, 25, 86, 25, 98, 25, 110, 25, 122, 25, 134, 25, 146, 25, 158, 25, 170, 25, 182, 25, 119, 1, 194, 25, 206, 25, 218, 25, 230, 25, 242, 25, 254, 25, 10, 26, 22, 26, 34, 26, 46, 26, 58, 26, 70, 26, 130, 1, 82, 26, 94, 26, 82, 17, 106, 26, 118, 26, 218, 16, 130, 26, 142, 26, 178, 26, 141, 1, 190, 26, 202, 26, 214, 26, 226, 26, 238, 26, 250, 26, 6, 27, 18, 27, 30, 27, 42, 27, 54, 27, 66, 27, 78, 27, 90, 27, 102, 27, 114, 27, 126, 27, 138, 27, 162, 27, 174, 27, 94, 17, 152, 1, 186, 27, 198, 27, 210, 27, 222, 27, 56, 0, 234, 27, 246, 27, 2, 28, 163, 1, 14, 28, 26, 28, 38, 28, 50, 28, 62, 28, 74, 28, 86, 28, 98, 28, 148, 6, 110, 28, 122, 28, 134, 28, 146, 28, 174, 1, 158, 28, 106, 17, 170, 28, 182, 28, 194, 28, 206, 28, 118, 17, 218, 28, 230, 28, 242, 28, 10, 29, 22, 29, 34, 29, 46, 29, 58, 29, 70, 29, 82, 29, 94, 29, 106, 29, 118, 29, 130, 29, 142, 29, 154, 29, 166, 29, 178, 29, 190, 29, 202, 29, 214, 29, 226, 29, 238, 29, 250, 29, 170, 6, 6, 30, 18, 30, 30, 30, 130, 17, 181, 6, 42, 30, 54, 30, 66, 30, 192, 6, 78, 30, 90, 30, 102, 30, 114, 30, 142, 17, 126, 30, 154, 17, 166, 17, 138, 30, 150, 30, 162, 30, 174, 30, 186, 30, 198, 30, 210, 30, 178, 17, 203, 6, 190, 17, 234, 30, 185, 1, 246, 30, 2, 31, 14, 31, 26, 31, 38, 31, 50, 31, 196, 1, 62, 31, 74, 31, 202, 17, 214, 6, 207, 1, 225, 6, 236, 6, 218, 1, 86, 31, 98, 31, 110, 31, 122, 31, 134, 31, 146, 31, 158, 31, 170, 31, 182, 31, 194, 31, 206, 31, 67, 0, 218, 31, 230, 31, 242, 31, 254, 31, 10, 32, 22, 32, 34, 32, 46, 32, 58, 32, 70, 32, 82, 32, 94, 32, 106, 32, 142, 32, 154, 32, 166, 32, 178, 32, 190, 32, 202, 32, 214, 32, 226, 32, 238, 32, 250, 32, 6, 33, 18, 33, 30, 33, 42, 33, 78, 0, 54, 33, 66, 33, 78, 33, 90, 33, 229, 1, 102, 33, 114, 33, 46, 7, 126, 33, 138, 33, 214, 17, 150, 33, 226, 17, 89, 0, 162, 33, 174, 33, 1, 0, 6, 2, 186, 33, 198, 33, 210, 33, 222, 33, 28, 2, 238, 17, 246, 33, 50, 2, 2, 34, 250, 17, 14, 34, 26, 34, 76, 8, 38, 34, 50, 34, 61, 2, 62, 34, 74, 34, 86, 34, 98, 34, 110, 34, 109, 8, 72, 2, 122, 34, 134, 34, 83, 2, 146, 34, 6, 18, 170, 34, 120, 8, 182, 34, 194, 34, 94, 2, 206, 34, 218, 34, 230, 34, 242, 34, 254, 34, 105, 2, 10, 35, 22, 35, 34, 35, 46, 35, 58, 35, 70, 35, 82, 35, 94, 35, 106, 35, 118, 35, 130, 35, 142, 35, 154, 35, 166, 35, 249, 9, 4, 10, 178, 35, 190, 35, 202, 35, 214, 35, 226, 35, 238, 35, 250, 35, 6, 36, 18, 36, 30, 36, 42, 36, 54, 36, 66, 36, 78, 36, 90, 36, 15, 10, 102, 36, 114, 36, 126, 36, 138, 36, 150, 36, 162, 36, 174, 36, 186, 36, 198, 36, 210, 36, 222, 36, 234, 36, 246, 36, 2, 37, 14, 37, 26, 37, 38, 37, 26, 10, 50, 37, 62, 37, 74, 37, 116, 2, 86, 37, 98, 37, 110, 37, 122, 37, 134, 37, 146, 37, 158, 37, 170, 37, 182, 37, 18, 18, 194, 37, 30, 18, 42, 18, 206, 37, 218, 37, 230, 37, 127, 2, 242, 37, 254, 37, 10, 38, 22, 38, 34, 38, 46, 38, 58, 38, 37, 10, 70, 38, 82, 38, 94, 38, 106, 38, 118, 38, 130, 38, 142, 38, 92, 10, 48, 10, 70, 10, 154, 38, 166, 38, 178, 38, 190, 38, 202, 38, 214, 38, 226, 38, 111, 0, 238, 38, 219, 16, 250, 38, 18, 39, 30, 39, 42, 39, 54, 39, 66, 39, 54, 18, 66, 18, 78, 39, 90, 39, 102, 39, 78, 18, 126, 39, 138, 39, 114, 10, 150, 39, 174, 39, 186, 39, 138, 2, 198, 39, 210, 39, 149, 2, 160, 2, 222, 39, 122, 0, 136, 10, 234, 39, 246, 39, 2, 40, 14, 40, 26, 40, 38, 40, 50, 40, 62, 40, 133, 0, 74, 40, 86, 40, 98, 40, 110, 40, 147, 10, 122, 40, 134, 40, 171, 2, 146, 40, 158, 10, 158, 40, 170, 40, 182, 40, 194, 40, 206, 40, 218, 40, 31, 21, 230, 40, 242, 40, 254, 40, 10, 41, 182, 2, 144, 0, 191, 10, 202, 10, 22, 41, 34, 41, 46, 41, 58, 41, 155, 0, 70, 41, 82, 41, 62, 73, 94, 41, 106, 41, 118, 41, 130, 41, 235, 10, 142, 41, 154, 41, 166, 41, 178, 41, 190, 41, 1, 11, 12, 11, 166, 0, 202, 41, 214, 41, 90, 18, 226, 41, 238, 41, 250, 41, 6, 42, 18, 42, 30, 42, 54, 42, 66, 42, 78, 42, 90, 42, 102, 42, 114, 42, 126, 42, 138, 42, 150, 42, 162, 42, 174, 42, 186, 42, 222, 42, 234, 42, 246, 42, 2, 43, 14, 43, 26, 43, 38, 43, 23, 11, 50, 43, 62, 43, 74, 43, 86, 43, 102, 18, 98, 43, 110, 43, 122, 43, 114, 18, 134, 43, 146, 43, 34, 11, 158, 43, 170, 43, 182, 43, 45, 11, 194, 43, 206, 43, 218, 43, 230, 43, 78, 11, 56, 11, 193, 2, 193, 2, 67, 11, 204, 2, 242, 43, 254, 43, 215, 2, 10, 44, 22, 44, 34, 44, 46, 44, 177, 0, 58, 44, 70, 44, 82, 44, 94, 44, 106, 44, 118, 44, 226, 2, 130, 44, 142, 44, 188, 0, 154, 44, 166, 44, 178, 44, 190, 44, 202, 44, 214, 44, 226, 44, 238, 44, 250, 44, 6, 45, 18, 45, 30, 45, 42, 45, 54, 45, 43, 21, 66, 45, 78, 45, 90, 45, 126, 18, 102, 45, 114, 45, 133, 11, 126, 45, 138, 45, 150, 45, 162, 45, 174, 45, 186, 45, 198, 45, 210, 45, 222, 45, 234, 45, 188, 11, 144, 11, 155, 11, 237, 2, 246, 45, 2, 46, 14, 46, 26, 46, 38, 46, 50, 46, 62, 46, 74, 46, 199, 11, 86, 46, 98, 46, 110, 46, 221, 11, 122, 46, 134, 46, 146, 46, 158, 46, 170, 46, 182, 46, 194, 46, 206, 46, 218, 46, 230, 46, 3, 3, 248, 2, 248, 2, 14, 3, 14, 3, 242, 46, 254, 46, 10, 47, 22, 47, 34, 47, 232, 11, 138, 18, 46, 47, 58, 47, 70, 47, 82, 47, 94, 47, 243, 11, 106, 47, 118, 47, 130, 47, 142, 47, 150, 18, 154, 47, 166, 47, 178, 47, 190, 47, 162, 18, 202, 47, 214, 47, 238, 47, 254, 11, 250, 47, 9, 12, 6, 48, 18, 48, 30, 48, 174, 18, 42, 48, 54, 48, 66, 48, 186, 18, 78, 48, 90, 48, 102, 48, 114, 48, 126, 48, 138, 48, 150, 48, 162, 48, 20, 12, 174, 48, 186, 48, 199, 0, 198, 48, 210, 48, 222, 48, 234, 48, 246, 48, 2, 49, 14, 49, 25, 3, 26, 49, 31, 12, 38, 49, 50, 49, 62, 49, 74, 49, 36, 3, 86, 49, 206, 16, 110, 49, 122, 49, 134, 49, 146, 49, 158, 49, 182, 49, 206, 49, 230, 16, 230, 16, 218, 49, 230, 49, 242, 49, 254, 49, 10, 50, 22, 50, 34, 50, 46, 50, 58, 50, 70, 50, 82, 50, 94, 50, 106, 50, 118, 50, 130, 50, 142, 50, 154, 50, 166, 50, 178, 50, 190, 50, 202, 50, 214, 50, 226, 50, 238, 50, 6, 51, 18, 51, 30, 51, 198, 18, 210, 18, 42, 51, 54, 51, 66, 51, 47, 3, 78, 51, 90, 51, 102, 51, 114, 51, 126, 51, 42, 12, 138, 51, 150, 51, 53, 12, 162, 51, 30, 21, 222, 18, 234, 18, 174, 51, 58, 3, 186, 51, 198, 51, 210, 51, 222, 51, 69, 3, 234, 51, 246, 51, 2, 52, 14, 52, 26, 52, 38, 52, 80, 3, 210, 0, 210, 0, 50, 52, 62, 52, 74, 52, 98, 52, 86, 12, 122, 52, 134, 52, 146, 52, 158, 52, 170, 52, 182, 52, 194, 52, 206, 52, 218, 52, 230, 52, 97, 12, 242, 52, 254, 52, 246, 18, 113, 3, 91, 3, 102, 3, 124, 3, 10, 53, 22, 53, 34, 53, 46, 53, 58, 53, 70, 53, 82, 53, 94, 53, 106, 53, 118, 53, 130, 53, 142, 53, 154, 53, 2, 19, 166, 53, 178, 53, 221, 0, 190, 53, 202, 53, 214, 53, 226, 53, 238, 53, 14, 19, 250, 53, 6, 54, 18, 54, 30, 54, 42, 54, 54, 54, 130, 12, 66, 54, 78, 54, 90, 54, 102, 54, 114, 54, 126, 54, 138, 54, 135, 3, 150, 54, 162, 54, 174, 54, 186, 54, 198, 54, 210, 54, 152, 12, 222, 54, 234, 54, 146, 3, 246, 54, 2, 55, 26, 55, 38, 55, 174, 12, 50, 55, 62, 55, 74, 55, 86, 55, 98, 55, 110, 55, 122, 55, 134, 55, 146, 55, 158, 55, 170, 55, 182, 55, 194, 55, 206, 55, 218, 55, 230, 55, 157, 3, 242, 55, 254, 55, 218, 12, 10, 56, 22, 56, 34, 56, 46, 56, 26, 19, 229, 12, 38, 19, 58, 56, 50, 19, 70, 56, 82, 56, 94, 56, 240, 12, 106, 56, 118, 56, 130, 56, 142, 56, 154, 56, 166, 56, 178, 56, 190, 56, 202, 56, 214, 56, 226, 56, 238, 56, 251, 12, 250, 56, 6, 57, 18, 57, 6, 13, 30, 57, 42, 57, 54, 57, 17, 13, 66, 57, 42, 21, 78, 57, 28, 13, 90, 57, 102, 57, 114, 57, 126, 57, 168, 3, 138, 57, 150, 57, 39, 13, 62, 19, 162, 57, 74, 19, 174, 57, 190, 3, 179, 3, 179, 3, 186, 57, 86, 19, 98, 19, 198, 57, 210, 57, 110, 19, 246, 57, 2, 58, 14, 58, 26, 58, 38, 58, 50, 58, 122, 19, 62, 58, 74, 58, 86, 58, 134, 19, 98, 58, 110, 58, 146, 19, 122, 58, 201, 3, 134, 58, 146, 58, 158, 58, 170, 58, 182, 58, 194, 58, 206, 58, 218, 58, 158, 19, 230, 58, 75, 73, 61, 13, 242, 58, 232, 0, 254, 58, 10, 59, 22, 59, 160, 13, 34, 59, 46, 59, 58, 59, 70, 59, 82, 59, 94, 59, 106, 59, 130, 59, 142, 59, 154, 59, 170, 19, 178, 59, 202, 59, 190, 59, 214, 59, 226, 59, 171, 13, 238, 59, 250, 59, 6, 60, 18, 60, 212, 3, 30, 60, 42, 60, 223, 3, 54, 60, 66, 60, 78, 60, 90, 60, 102, 60, 12, 0, 114, 60, 126, 60, 234, 3, 138, 60, 150, 60, 162, 60, 215, 13, 174, 60, 186, 60, 198, 60, 210, 60, 222, 60, 246, 60, 182, 19, 2, 61, 14, 61, 26, 61, 14, 14, 38, 61, 50, 61, 62, 61, 74, 61, 86, 61, 25, 14, 36, 14, 245, 3, 0, 4, 98, 61, 110, 61, 47, 14, 122, 61, 134, 61, 146, 61, 158, 61, 58, 14, 194, 19, 170, 61, 182, 61, 194, 61, 206, 61, 218, 61, 230, 61, 242, 61, 243, 0, 254, 61, 10, 62, 22, 62, 34, 62, 46, 62, 11, 4, 58, 62, 70, 62, 94, 62, 22, 4, 106, 62, 118, 62, 130, 62, 142, 62, 254, 0, 154, 62, 166, 62, 178, 62, 206, 19, 190, 62, 202, 62, 33, 4, 214, 62, 226, 62, 238, 62, 250, 62, 6, 63, 44, 4, 218, 19, 30, 63, 42, 63, 54, 63, 66, 63, 78, 63, 90, 63, 102, 63, 55, 4, 66, 4, 77, 4, 146, 14, 157, 14, 168, 14, 230, 19, 114, 63, 126, 63, 138, 63, 179, 14, 150, 63, 162, 63, 174, 63, 201, 14, 186, 63, 198, 63, 212, 14, 210, 63, 222, 63, 234, 63, 246, 63, 2, 64, 88, 4, 9, 1, 242, 19, 14, 64, 26, 64, 38, 64, 50, 64, 62, 64, 74, 64, 86, 64, 254, 19, 110, 64, 110, 4, 122, 64, 134, 64, 146, 64, 158, 64, 170, 64, 182, 64, 194, 64, 206, 64, 218, 64, 230, 64, 254, 64, 10, 65, 22, 65, 10, 20, 34, 65, 46, 65, 58, 65, 121, 4, 70, 65, 82, 65, 94, 65, 106, 65, 118, 65, 143, 4, 132, 4, 132, 4, 130, 65, 142, 65, 154, 65, 20, 1, 166, 65, 178, 65, 190, 65, 22, 15, 202, 65, 214, 65, 226, 65, 238, 65, 250, 65, 6, 66, 154, 4, 22, 20, 18, 66, 30, 66, 55, 15, 42, 66, 54, 66, 66, 66, 78, 66, 90, 66, 66, 15, 102, 66, 165, 4, 114, 66, 126, 66, 138, 66, 34, 20, 150, 66, 162, 66, 174, 66, 176, 4, 186, 66, 46, 20, 198, 66, 88, 15, 210, 66, 222, 66, 234, 66, 246, 66, 2, 67, 99, 15, 14, 67, 26, 67, 38, 67, 50, 67, 62, 67, 74, 67, 86, 67, 98, 67, 110, 67, 122, 67, 134, 67, 58, 20, 146, 67, 158, 67, 170, 67, 70, 20, 110, 15, 182, 67, 194, 67, 206, 67, 218, 67, 230, 67, 187, 4, 198, 4, 198, 4, 121, 15, 82, 20, 209, 4, 242, 67, 254, 67, 1, 0, 100, 0, 1, 0, 232, 0, 240, 1, 100, 0, 39, 2, 9, 1, 23, 0, 99, 4, 112, 7, 1, 0, 1, 0, 17, 2, 207, 16, 219, 16, 133, 0, 144, 0, 240, 1, 101, 7, 90, 7, 254, 0, 207, 16, 42, 1, 86, 1, 12, 0, 12, 0, 251, 1, 67, 0, 100, 0, 234, 33, 159, 6, 96, 5, 87, 8, 55, 4, 122, 0, 72, 13, 10, 68, 123, 7, 69, 14, 247, 6, 185, 12, 234, 14, 108, 1, 72, 2, 226, 47, 242, 16, 23, 0, 91, 3, 187, 4, 122, 11, 102, 3, 154, 15, 154, 17, 170, 19, 204, 2, 165, 15, 59, 10, 31, 1, 10, 17, 198, 18, 98, 8, 93, 6, 242, 19, 22, 17, 122, 0, 64, 1, 174, 1, 56, 0, 196, 12, 153, 8, 130, 1, 107, 5, 212, 3, 164, 8, 175, 8, 119, 12, 134, 7, 222, 57, 232, 0, 229, 1, 22, 68, 119, 1, 174, 24, 70, 17, 106, 17, 14, 25, 56, 0, 204, 13, 253, 4, 146, 19, 131, 8, 254, 28, 150, 27, 99, 4, 186, 8, 197, 8, 2, 7, 208, 8, 145, 7, 219, 8, 207, 16, 156, 7, 34, 68, 167, 7, 17, 2, 83, 13, 234, 57, 57, 7, 196, 1, 130, 17, 185, 1, 142, 20, 166, 17, 142, 17, 130, 20, 190, 17, 30, 18, 12, 0, 64, 12, 124, 14, 53, 1, 75, 12, 132, 15, 126, 6, 143, 15, 42, 42, 135, 14, 67, 0, 41, 5, 218, 1, 178, 7, 118, 5, 219, 16, 163, 12, 6, 39, 154, 26, 118, 17, 35, 7, 118, 32, 142, 8, 189, 7, 200, 7, 50, 2, 23, 0, 129, 5, 11, 15, 28, 2, 45, 0, 251, 1, 13, 7, 214, 17, 91, 14, 118, 23, 45, 0, 83, 2, 94, 2, 105, 2, 241, 8, 166, 0, 252, 8, 75, 1, 215, 2, 7, 9, 18, 9, 211, 7, 89, 0, 46, 68, 222, 7, 94, 13, 166, 0, 207, 1, 118, 20, 238, 17, 18, 18, 226, 13, 89, 0, 82, 17, 127, 2, 190, 3, 19, 5, 58, 68, 233, 7, 138, 2, 149, 2, 160, 2, 8, 5, 89, 11, 41, 16, 114, 39, 207, 16, 34, 0, 222, 18, 111, 0, 231, 15, 242, 15, 253, 15, 8, 16, 19, 16, 213, 10, 54, 18, 78, 18, 2, 19, 125, 10, 244, 7, 133, 0, 111, 0, 219, 16, 140, 5, 61, 2, 182, 2, 144, 0, 250, 17, 151, 5, 155, 0, 100, 11, 246, 10, 155, 0, 245, 14, 3, 3, 177, 0, 162, 5, 111, 11, 188, 0, 173, 5, 193, 13, 226, 17, 224, 10, 158, 19, 177, 0, 245, 3, 18, 21, 226, 2, 188, 0, 166, 26, 199, 0, 184, 5, 24, 7, 243, 0, 190, 14, 25, 3, 40, 9, 36, 3, 195, 5, 38, 19, 199, 0, 33, 15, 166, 11, 242, 4, 77, 15, 230, 8, 79, 7, 177, 11, 44, 15, 29, 9, 51, 9, 102, 14, 90, 18, 105, 13, 95, 9, 126, 18, 172, 9, 216, 9, 46, 20, 50, 13, 162, 18, 174, 18, 138, 18, 42, 18, 186, 18, 154, 20, 166, 20, 206, 5, 62, 9, 182, 13, 77, 4, 73, 9, 47, 3, 58, 3, 152, 1, 221, 0, 80, 3, 30, 5, 84, 9, 217, 5, 124, 3, 210, 18, 78, 0, 39, 2, 113, 3, 116, 13, 103, 10, 194, 49, 14, 19, 178, 20, 137, 6, 234, 18, 250, 50, 190, 20, 10, 71, 69, 3, 108, 12, 210, 0, 246, 18, 86, 52, 82, 62, 221, 0, 114, 18, 110, 52, 63, 5, 106, 9, 81, 10, 255, 7, 157, 3, 135, 3, 146, 3, 237, 13, 202, 20, 34, 17, 14, 55, 230, 16, 6, 21, 228, 5, 194, 19, 234, 60, 74, 19, 168, 3, 62, 19, 80, 14, 10, 8, 21, 8, 117, 9, 34, 20, 6, 18, 31, 1, 201, 3, 128, 9, 32, 8, 239, 5, 127, 13, 110, 19, 116, 2, 46, 17, 202, 17, 210, 11, 98, 49, 86, 19, 98, 19, 134, 19, 66, 18, 214, 20, 250, 5, 68, 7, 139, 9, 223, 3, 66, 4, 12, 0, 234, 3, 118, 59, 166, 59, 50, 19, 86, 1, 150, 9, 5, 6, 88, 4, 22, 4, 141, 12, 11, 4, 169, 10, 161, 9, 44, 4, 207, 12, 138, 13, 43, 8, 16, 6, 206, 19, 226, 20, 182, 19, 58, 17, 6, 2, 248, 13, 3, 14, 0, 4, 243, 0, 218, 19, 18, 63, 230, 19, 178, 17, 223, 14, 98, 64, 254, 19, 183, 9, 194, 9, 205, 9, 20, 1, 143, 4, 22, 20, 149, 13, 154, 4, 27, 6, 165, 4, 176, 4, 58, 20, 34, 0, 9, 1, 242, 64, 130, 32, 10, 20, 102, 18, 110, 4, 70, 20, 94, 17, 121, 4, 250, 20, 158, 34, 78, 0, 20, 1, 150, 18, 198, 42, 210, 42, 163, 1, 170, 49, 209, 4, 0, 15, 54, 8, 42, 1, 82, 20, 180, 10, 65, 8, 231, 4, 141, 1, 106, 20, 227, 9, 254, 0, 162, 39, 26, 19, 113, 14, 122, 19, 254, 16, 38, 6, 238, 9, 237, 2, 171, 2, 33, 4, 222, 30, 70, 68, 94, 20, 94, 20, 82, 68, 94, 68, 31, 1, 106, 68, 118, 68, 130, 68, 142, 68, 154, 68, 166, 68, 178, 68, 42, 1, 220, 4, 220, 4, 176, 15, 190, 68, 106, 20, 202, 68, 187, 15, 214, 68, 226, 68, 231, 4, 238, 68, 250, 68, 6, 69, 18, 69, 30, 69, 42, 69, 54, 69, 66, 69, 78, 69, 90, 69, 198, 15, 102, 69, 114, 69, 126, 69, 138, 69, 150, 69, 162, 69, 174, 69, 186, 69, 198, 69, 210, 69, 222, 69, 234, 69, 246, 69, 2, 70, 14, 70, 26, 70, 38, 70, 50, 70, 62, 70, 74, 70, 86, 70, 98, 70, 209, 15, 110, 70, 122, 70, 134, 70, 118, 20, 146, 70, 158, 70, 170, 70, 182, 70, 194, 70, 206, 70, 218, 70, 130, 20, 142, 20, 230, 70, 220, 15, 242, 70, 154, 20, 166, 20, 254, 70, 178, 20, 190, 20, 202, 20, 22, 71, 34, 71, 46, 71, 214, 20, 58, 71, 226, 20, 70, 71, 82, 71, 94, 71, 106, 71, 118, 71, 130, 71, 142, 71, 154, 71, 166, 71, 178, 71, 190, 71, 202, 71, 214, 71, 226, 71, 238, 71, 250, 71, 6, 72, 18, 72, 30, 72, 42, 72, 54, 72, 66, 72, 78, 72, 90, 72, 30, 16, 102, 72, 114, 72, 126, 72, 138, 72, 150, 72, 162, 72, 174, 72, 238, 20, 238, 20, 186, 72, 198, 72, 210, 72, 242, 4, 222, 72, 234, 72, 246, 72, 250, 20, 207, 16, 52, 16, 63, 16, 253, 4, 74, 16, 85, 16, 19, 5, 8, 5, 34, 0, 96, 16, 30, 5, 107, 16, 118, 16, 129, 16, 140, 16, 151, 16, 162, 16, 34, 0, 173, 16, 184, 16, 6, 21, 2, 73, 18, 21, 14, 73, 26, 73, 38, 73, 195, 16, 50, 73, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 118, 40, 0, 32, 7, 0, 0, 0, 143, 8, 0, 0, 123, 40, 0, 0, 123, 40, 0, 0, 173, 1, 0, 0, 1, 0, 0, 0, 132, 32, 0, 0, 109, 97, 112, 84, 105, 109, 101, 122, 111, 110, 101, 115, 0, 65, 99, 114, 101, 0, 48, 48, 49, 0, 65, 102, 103, 104, 97, 110, 105, 115, 116, 97, 110, 0, 65, 102, 114, 105, 99, 97, 95, 67, 101, 110, 116, 114, 97, 108, 0, 66, 73, 0, 66, 87, 0, 67, 68, 0, 77, 87, 0, 82, 87, 0, 90, 77, 0, 90, 87, 0, 65, 102, 114, 105, 99, 97, 95, 69, 97, 115, 116, 101, 114, 110, 0, 68, 74, 0, 69, 82, 0, 69, 84, 0, 75, 77, 0, 77, 71, 0, 83, 79, 0, 84, 90, 0, 85, 71, 0, 89, 84, 0, 65, 102, 114, 105, 99, 97, 95, 70, 97, 114, 87, 101, 115, 116, 101, 114, 110, 0, 65, 102, 114, 105, 99, 97, 95, 83, 111, 117, 116, 104, 101, 114, 110, 0, 76, 83, 0, 83, 90, 0, 65, 102, 114, 105, 99, 97, 95, 87, 101, 115, 116, 101, 114, 110, 0, 65, 79, 0, 66, 74, 0, 67, 70, 0, 67, 71, 0, 67, 77, 0, 71, 65, 0, 71, 81, 0, 78, 69, 0, 84, 68, 0, 65, 107, 116, 121, 117, 98, 105, 110, 115, 107, 0, 65, 108, 97, 115, 107, 97, 0, 65, 108, 97, 115, 107, 97, 95, 72, 97, 119, 97, 105, 105, 0, 65, 109, 97, 122, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 95, 67, 101, 110, 116, 114, 97, 108, 0, 66, 90, 0, 67, 65, 0, 67, 82, 0, 71, 84, 0, 72, 78, 0, 77, 88, 0, 83, 86, 0, 65, 109, 101, 114, 105, 99, 97, 95, 69, 97, 115, 116, 101, 114, 110, 0, 66, 83, 0, 72, 84, 0, 74, 77, 0, 75, 89, 0, 80, 65, 0, 65, 109, 101, 114, 105, 99, 97, 95, 77, 111, 117, 110, 116, 97, 105, 110, 0, 65, 109, 101, 114, 105, 99, 97, 95, 80, 97, 99, 105, 102, 105, 99, 0, 65, 114, 97, 98, 105, 97, 110, 0, 66, 72, 0, 73, 81, 0, 75, 87, 0, 81, 65, 0, 89, 69, 0, 65, 114, 103, 101, 110, 116, 105, 110, 97, 0, 65, 114, 103, 101, 110, 116, 105, 110, 97, 95, 87, 101, 115, 116, 101, 114, 110, 0, 65, 114, 109, 101, 110, 105, 97, 0, 65, 115, 104, 107, 104, 97, 98, 97, 100, 0, 65, 116, 108, 97, 110, 116, 105, 99, 0, 65, 71, 0, 65, 73, 0, 65, 87, 0, 66, 66, 0, 66, 77, 0, 66, 81, 0, 67, 87, 0, 68, 77, 0, 71, 68, 0, 71, 76, 0, 71, 80, 0, 75, 78, 0, 76, 67, 0, 77, 70, 0, 77, 81, 0, 77, 83, 0, 80, 82, 0, 83, 88, 0, 84, 84, 0, 86, 67, 0, 86, 71, 0, 86, 73, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 95, 67, 101, 110, 116, 114, 97, 108, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 95, 67, 101, 110, 116, 114, 97, 108, 87, 101, 115, 116, 101, 114, 110, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 95, 69, 97, 115, 116, 101, 114, 110, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 95, 87, 101, 115, 116, 101, 114, 110, 0, 65, 122, 101, 114, 98, 97, 105, 106, 97, 110, 0, 66, 97, 110, 103, 108, 97, 100, 101, 115, 104, 0, 66, 101, 114, 105, 110, 103, 0, 66, 104, 117, 116, 97, 110, 0, 66, 111, 108, 105, 118, 105, 97, 0, 66, 111, 114, 110, 101, 111, 0, 66, 114, 97, 115, 105, 108, 105, 97, 0, 66, 114, 105, 116, 105, 115, 104, 0, 67, 104, 97, 109, 111, 114, 114, 111, 0, 71, 85, 0, 67, 104, 105, 108, 101, 0, 67, 104, 105, 110, 97, 0, 67, 111, 108, 111, 109, 98, 105, 97, 0, 67, 111, 111, 107, 0, 67, 117, 98, 97, 0, 68, 97, 99, 99, 97, 0, 68, 111, 109, 105, 110, 105, 99, 97, 110, 0, 68, 117, 116, 99, 104, 95, 71, 117, 105, 97, 110, 97, 0, 69, 97, 115, 116, 95, 84, 105, 109, 111, 114, 0, 69, 99, 117, 97, 100, 111, 114, 0, 69, 117, 114, 111, 112, 101, 95, 67, 101, 110, 116, 114, 97, 108, 0, 65, 68, 0, 65, 76, 0, 65, 84, 0, 66, 65, 0, 66, 69, 0, 67, 72, 0, 67, 90, 0, 68, 69, 0, 68, 75, 0, 69, 83, 0, 71, 73, 0, 72, 82, 0, 72, 85, 0, 73, 84, 0, 76, 73, 0, 76, 85, 0, 77, 67, 0, 77, 69, 0, 77, 75, 0, 78, 76, 0, 78, 79, 0, 80, 76, 0, 82, 83, 0, 83, 69, 0, 83, 73, 0, 83, 75, 0, 83, 77, 0, 84, 78, 0, 86, 65, 0, 88, 75, 0, 69, 117, 114, 111, 112, 101, 95, 69, 97, 115, 116, 101, 114, 110, 0, 65, 88, 0, 66, 71, 0, 67, 89, 0, 69, 71, 0, 70, 73, 0, 71, 82, 0, 74, 79, 0, 76, 66, 0, 83, 89, 0, 69, 117, 114, 111, 112, 101, 95, 70, 117, 114, 116, 104, 101, 114, 95, 69, 97, 115, 116, 101, 114, 110, 0, 82, 85, 0, 69, 117, 114, 111, 112, 101, 95, 87, 101, 115, 116, 101, 114, 110, 0, 70, 79, 0, 70, 97, 108, 107, 108, 97, 110, 100, 0, 70, 114, 101, 110, 99, 104, 95, 71, 117, 105, 97, 110, 97, 0, 70, 114, 101, 110, 99, 104, 95, 83, 111, 117, 116, 104, 101, 114, 110, 0, 70, 114, 117, 110, 122, 101, 0, 66, 70, 0, 67, 73, 0, 71, 66, 0, 71, 72, 0, 71, 77, 0, 71, 78, 0, 73, 69, 0, 77, 76, 0, 77, 82, 0, 83, 72, 0, 83, 76, 0, 83, 78, 0, 84, 71, 0, 71, 105, 108, 98, 101, 114, 116, 95, 73, 115, 108, 97, 110, 100, 115, 0, 71, 114, 101, 101, 110, 108, 97, 110, 100, 95, 67, 101, 110, 116, 114, 97, 108, 0, 71, 114, 101, 101, 110, 108, 97, 110, 100, 95, 69, 97, 115, 116, 101, 114, 110, 0, 71, 114, 101, 101, 110, 108, 97, 110, 100, 95, 87, 101, 115, 116, 101, 114, 110, 0, 71, 117, 108, 102, 0, 79, 77, 0, 72, 97, 119, 97, 105, 105, 95, 65, 108, 101, 117, 116, 105, 97, 110, 0, 73, 110, 100, 105, 97, 0, 76, 75, 0, 73, 110, 100, 105, 97, 110, 95, 79, 99, 101, 97, 110, 0, 73, 110, 100, 111, 99, 104, 105, 110, 97, 0, 75, 72, 0, 76, 65, 0, 73, 110, 100, 111, 110, 101, 115, 105, 97, 95, 67, 101, 110, 116, 114, 97, 108, 0, 73, 110, 100, 111, 110, 101, 115, 105, 97, 95, 69, 97, 115, 116, 101, 114, 110, 0, 73, 110, 100, 111, 110, 101, 115, 105, 97, 95, 87, 101, 115, 116, 101, 114, 110, 0, 73, 114, 97, 110, 0, 73, 114, 105, 115, 104, 0, 73, 115, 114, 97, 101, 108, 0, 74, 97, 112, 97, 110, 0, 75, 97, 122, 97, 107, 104, 115, 116, 97, 110, 95, 69, 97, 115, 116, 101, 114, 110, 0, 75, 97, 122, 97, 107, 104, 115, 116, 97, 110, 95, 87, 101, 115, 116, 101, 114, 110, 0, 75, 105, 122, 105, 108, 111, 114, 100, 97, 0, 75, 111, 114, 101, 97, 0, 75, 117, 121, 98, 121, 115, 104, 101, 118, 0, 75, 121, 114, 103, 121, 115, 116, 97, 110, 0, 76, 97, 110, 107, 97, 0, 76, 105, 98, 101, 114, 105, 97, 0, 76, 105, 110, 101, 95, 73, 115, 108, 97, 110, 100, 115, 0, 77, 97, 108, 97, 121, 97, 0, 77, 97, 108, 97, 121, 115, 105, 97, 0, 77, 97, 114, 115, 104, 97, 108, 108, 95, 73, 115, 108, 97, 110, 100, 115, 0, 77, 101, 120, 105, 99, 111, 95, 78, 111, 114, 116, 104, 119, 101, 115, 116, 0, 77, 101, 120, 105, 99, 111, 95, 80, 97, 99, 105, 102, 105, 99, 0, 77, 111, 110, 103, 111, 108, 105, 97, 0, 77, 121, 97, 110, 109, 97, 114, 0, 78, 101, 112, 97, 108, 0, 78, 101, 119, 95, 67, 97, 108, 101, 100, 111, 110, 105, 97, 0, 78, 101, 119, 95, 90, 101, 97, 108, 97, 110, 100, 0, 65, 81, 0, 78, 101, 119, 102, 111, 117, 110, 100, 108, 97, 110, 100, 0, 78, 111, 114, 116, 104, 95, 77, 97, 114, 105, 97, 110, 97, 0, 80, 97, 107, 105, 115, 116, 97, 110, 0, 80, 97, 112, 117, 97, 95, 78, 101, 119, 95, 71, 117, 105, 110, 101, 97, 0, 80, 97, 114, 97, 103, 117, 97, 121, 0, 80, 101, 114, 117, 0, 80, 104, 105, 108, 105, 112, 112, 105, 110, 101, 115, 0, 80, 104, 111, 101, 110, 105, 120, 95, 73, 115, 108, 97, 110, 100, 115, 0, 80, 105, 101, 114, 114, 101, 95, 77, 105, 113, 117, 101, 108, 111, 110, 0, 83, 97, 109, 111, 97, 0, 83, 101, 121, 99, 104, 101, 108, 108, 101, 115, 0, 83, 104, 101, 118, 99, 104, 101, 110, 107, 111, 0, 83, 111, 108, 111, 109, 111, 110, 0, 83, 117, 114, 105, 110, 97, 109, 101, 0, 83, 118, 101, 114, 100, 108, 111, 118, 115, 107, 0, 84, 97, 106, 105, 107, 105, 115, 116, 97, 110, 0, 84, 111, 107, 101, 108, 97, 117, 0, 84, 111, 110, 103, 97, 0, 84, 117, 114, 107, 101, 121, 0, 84, 117, 114, 107, 109, 101, 110, 105, 115, 116, 97, 110, 0, 84, 117, 118, 97, 108, 117, 0, 85, 114, 97, 108, 115, 107, 0, 85, 114, 117, 103, 117, 97, 121, 0, 85, 122, 98, 101, 107, 105, 115, 116, 97, 110, 0, 86, 97, 110, 117, 97, 116, 117, 0, 86, 101, 110, 101, 122, 117, 101, 108, 97, 0, 89, 117, 107, 111, 110, 0, 109, 101, 116, 97, 122, 111, 110, 101, 73, 110, 102, 111, 0, 65, 102, 114, 105, 99, 97, 58, 65, 98, 105, 100, 106, 97, 110, 0, 65, 102, 114, 105, 99, 97, 58, 65, 99, 99, 114, 97, 0, 65, 102, 114, 105, 99, 97, 58, 65, 100, 100, 105, 115, 95, 65, 98, 97, 98, 97, 0, 65, 102, 114, 105, 99, 97, 58, 65, 108, 103, 105, 101, 114, 115, 0, 65, 102, 114, 105, 99, 97, 58, 65, 115, 109, 101, 114, 97, 0, 65, 102, 114, 105, 99, 97, 58, 66, 97, 109, 97, 107, 111, 0, 65, 102, 114, 105, 99, 97, 58, 66, 97, 110, 103, 117, 105, 0, 65, 102, 114, 105, 99, 97, 58, 66, 97, 110, 106, 117, 108, 0, 65, 102, 114, 105, 99, 97, 58, 66, 105, 115, 115, 97, 117, 0, 65, 102, 114, 105, 99, 97, 58, 66, 108, 97, 110, 116, 121, 114, 101, 0, 65, 102, 114, 105, 99, 97, 58, 66, 114, 97, 122, 122, 97, 118, 105, 108, 108, 101, 0, 65, 102, 114, 105, 99, 97, 58, 66, 117, 106, 117, 109, 98, 117, 114, 97, 0, 65, 102, 114, 105, 99, 97, 58, 67, 97, 105, 114, 111, 0, 65, 102, 114, 105, 99, 97, 58, 67, 97, 115, 97, 98, 108, 97, 110, 99, 97, 0, 65, 102, 114, 105, 99, 97, 58, 67, 101, 117, 116, 97, 0, 65, 102, 114, 105, 99, 97, 58, 67, 111, 110, 97, 107, 114, 121, 0, 65, 102, 114, 105, 99, 97, 58, 68, 97, 107, 97, 114, 0, 65, 102, 114, 105, 99, 97, 58, 68, 97, 114, 95, 101, 115, 95, 83, 97, 108, 97, 97, 109, 0, 65, 102, 114, 105, 99, 97, 58, 68, 106, 105, 98, 111, 117, 116, 105, 0, 65, 102, 114, 105, 99, 97, 58, 68, 111, 117, 97, 108, 97, 0, 65, 102, 114, 105, 99, 97, 58, 69, 108, 95, 65, 97, 105, 117, 110, 0, 65, 102, 114, 105, 99, 97, 58, 70, 114, 101, 101, 116, 111, 119, 110, 0, 65, 102, 114, 105, 99, 97, 58, 71, 97, 98, 111, 114, 111, 110, 101, 0, 65, 102, 114, 105, 99, 97, 58, 72, 97, 114, 97, 114, 101, 0, 65, 102, 114, 105, 99, 97, 58, 74, 111, 104, 97, 110, 110, 101, 115, 98, 117, 114, 103, 0, 65, 102, 114, 105, 99, 97, 58, 74, 117, 98, 97, 0, 65, 102, 114, 105, 99, 97, 58, 75, 97, 109, 112, 97, 108, 97, 0, 65, 102, 114, 105, 99, 97, 58, 75, 104, 97, 114, 116, 111, 117, 109, 0, 65, 102, 114, 105, 99, 97, 58, 75, 105, 103, 97, 108, 105, 0, 65, 102, 114, 105, 99, 97, 58, 75, 105, 110, 115, 104, 97, 115, 97, 0, 65, 102, 114, 105, 99, 97, 58, 76, 97, 103, 111, 115, 0, 65, 102, 114, 105, 99, 97, 58, 76, 105, 98, 114, 101, 118, 105, 108, 108, 101, 0, 65, 102, 114, 105, 99, 97, 58, 76, 111, 109, 101, 0, 65, 102, 114, 105, 99, 97, 58, 76, 117, 97, 110, 100, 97, 0, 65, 102, 114, 105, 99, 97, 58, 76, 117, 98, 117, 109, 98, 97, 115, 104, 105, 0, 65, 102, 114, 105, 99, 97, 58, 76, 117, 115, 97, 107, 97, 0, 65, 102, 114, 105, 99, 97, 58, 77, 97, 108, 97, 98, 111, 0, 65, 102, 114, 105, 99, 97, 58, 77, 97, 112, 117, 116, 111, 0, 65, 102, 114, 105, 99, 97, 58, 77, 97, 115, 101, 114, 117, 0, 65, 102, 114, 105, 99, 97, 58, 77, 98, 97, 98, 97, 110, 101, 0, 65, 102, 114, 105, 99, 97, 58, 77, 111, 103, 97, 100, 105, 115, 104, 117, 0, 65, 102, 114, 105, 99, 97, 58, 77, 111, 110, 114, 111, 118, 105, 97, 0, 65, 102, 114, 105, 99, 97, 58, 78, 97, 105, 114, 111, 98, 105, 0, 65, 102, 114, 105, 99, 97, 58, 78, 100, 106, 97, 109, 101, 110, 97, 0, 65, 102, 114, 105, 99, 97, 58, 78, 105, 97, 109, 101, 121, 0, 65, 102, 114, 105, 99, 97, 58, 78, 111, 117, 97, 107, 99, 104, 111, 116, 116, 0, 65, 102, 114, 105, 99, 97, 58, 79, 117, 97, 103, 97, 100, 111, 117, 103, 111, 117, 0, 65, 102, 114, 105, 99, 97, 58, 80, 111, 114, 116, 111, 45, 78, 111, 118, 111, 0, 65, 102, 114, 105, 99, 97, 58, 83, 97, 111, 95, 84, 111, 109, 101, 0, 65, 102, 114, 105, 99, 97, 58, 84, 114, 105, 112, 111, 108, 105, 0, 65, 102, 114, 105, 99, 97, 58, 84, 117, 110, 105, 115, 0, 65, 102, 114, 105, 99, 97, 58, 87, 105, 110, 100, 104, 111, 101, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 100, 97, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 110, 99, 104, 111, 114, 97, 103, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 110, 103, 117, 105, 108, 108, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 110, 116, 105, 103, 117, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 97, 103, 117, 97, 105, 110, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 76, 97, 95, 82, 105, 111, 106, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 82, 105, 111, 95, 71, 97, 108, 108, 101, 103, 111, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 83, 97, 108, 116, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 83, 97, 110, 95, 74, 117, 97, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 83, 97, 110, 95, 76, 117, 105, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 84, 117, 99, 117, 109, 97, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 85, 115, 104, 117, 97, 105, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 117, 98, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 115, 117, 110, 99, 105, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 97, 104, 105, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 97, 104, 105, 97, 95, 66, 97, 110, 100, 101, 114, 97, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 97, 114, 98, 97, 100, 111, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 101, 108, 101, 109, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 101, 108, 105, 122, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 108, 97, 110, 99, 45, 83, 97, 98, 108, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 111, 97, 95, 86, 105, 115, 116, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 111, 103, 111, 116, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 111, 105, 115, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 117, 101, 110, 111, 115, 95, 65, 105, 114, 101, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 109, 98, 114, 105, 100, 103, 101, 95, 66, 97, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 109, 112, 111, 95, 71, 114, 97, 110, 100, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 110, 99, 117, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 114, 97, 99, 97, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 116, 97, 109, 97, 114, 99, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 121, 101, 110, 110, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 121, 109, 97, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 104, 105, 99, 97, 103, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 104, 105, 104, 117, 97, 104, 117, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 111, 114, 97, 108, 95, 72, 97, 114, 98, 111, 117, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 111, 114, 100, 111, 98, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 111, 115, 116, 97, 95, 82, 105, 99, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 114, 101, 115, 116, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 117, 105, 97, 98, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 117, 114, 97, 99, 97, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 97, 110, 109, 97, 114, 107, 115, 104, 97, 118, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 97, 119, 115, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 97, 119, 115, 111, 110, 95, 67, 114, 101, 101, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 101, 110, 118, 101, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 101, 116, 114, 111, 105, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 111, 109, 105, 110, 105, 99, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 69, 100, 109, 111, 110, 116, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 69, 105, 114, 117, 110, 101, 112, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 69, 108, 95, 83, 97, 108, 118, 97, 100, 111, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 70, 111, 114, 116, 95, 78, 101, 108, 115, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 70, 111, 114, 116, 97, 108, 101, 122, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 108, 97, 99, 101, 95, 66, 97, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 111, 100, 116, 104, 97, 98, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 111, 111, 115, 101, 95, 66, 97, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 114, 97, 110, 100, 95, 84, 117, 114, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 114, 101, 110, 97, 100, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 117, 97, 100, 101, 108, 111, 117, 112, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 117, 97, 116, 101, 109, 97, 108, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 117, 97, 121, 97, 113, 117, 105, 108, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 117, 121, 97, 110, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 72, 97, 108, 105, 102, 97, 120, 0, 65, 109, 101, 114, 105, 99, 97, 58, 72, 97, 118, 97, 110, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 72, 101, 114, 109, 111, 115, 105, 108, 108, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 75, 110, 111, 120, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 77, 97, 114, 101, 110, 103, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 80, 101, 116, 101, 114, 115, 98, 117, 114, 103, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 84, 101, 108, 108, 95, 67, 105, 116, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 86, 101, 118, 97, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 86, 105, 110, 99, 101, 110, 110, 101, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 87, 105, 110, 97, 109, 97, 99, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 112, 111, 108, 105, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 117, 118, 105, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 113, 97, 108, 117, 105, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 74, 97, 109, 97, 105, 99, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 74, 117, 106, 117, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 74, 117, 110, 101, 97, 117, 0, 65, 109, 101, 114, 105, 99, 97, 58, 75, 101, 110, 116, 117, 99, 107, 121, 58, 77, 111, 110, 116, 105, 99, 101, 108, 108, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 75, 114, 97, 108, 101, 110, 100, 105, 106, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 76, 97, 95, 80, 97, 122, 0, 65, 109, 101, 114, 105, 99, 97, 58, 76, 105, 109, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 76, 111, 115, 95, 65, 110, 103, 101, 108, 101, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 76, 111, 117, 105, 115, 118, 105, 108, 108, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 76, 111, 119, 101, 114, 95, 80, 114, 105, 110, 99, 101, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 99, 101, 105, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 110, 97, 103, 117, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 110, 97, 117, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 114, 105, 103, 111, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 114, 116, 105, 110, 105, 113, 117, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 116, 97, 109, 111, 114, 111, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 122, 97, 116, 108, 97, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 101, 110, 100, 111, 122, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 101, 110, 111, 109, 105, 110, 101, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 101, 114, 105, 100, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 101, 116, 108, 97, 107, 97, 116, 108, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 101, 120, 105, 99, 111, 95, 67, 105, 116, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 105, 113, 117, 101, 108, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 111, 110, 99, 116, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 111, 110, 116, 101, 114, 114, 101, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 111, 110, 116, 101, 118, 105, 100, 101, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 111, 110, 116, 115, 101, 114, 114, 97, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 97, 115, 115, 97, 117, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 101, 119, 95, 89, 111, 114, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 105, 112, 105, 103, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 111, 109, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 111, 114, 111, 110, 104, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 111, 114, 116, 104, 95, 68, 97, 107, 111, 116, 97, 58, 66, 101, 117, 108, 97, 104, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 111, 114, 116, 104, 95, 68, 97, 107, 111, 116, 97, 58, 67, 101, 110, 116, 101, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 111, 114, 116, 104, 95, 68, 97, 107, 111, 116, 97, 58, 78, 101, 119, 95, 83, 97, 108, 101, 109, 0, 65, 109, 101, 114, 105, 99, 97, 58, 79, 106, 105, 110, 97, 103, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 97, 110, 97, 109, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 97, 110, 103, 110, 105, 114, 116, 117, 110, 103, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 97, 114, 97, 109, 97, 114, 105, 98, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 104, 111, 101, 110, 105, 120, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 111, 114, 116, 45, 97, 117, 45, 80, 114, 105, 110, 99, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 111, 114, 116, 95, 111, 102, 95, 83, 112, 97, 105, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 111, 114, 116, 111, 95, 86, 101, 108, 104, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 117, 101, 114, 116, 111, 95, 82, 105, 99, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 117, 110, 116, 97, 95, 65, 114, 101, 110, 97, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 97, 105, 110, 121, 95, 82, 105, 118, 101, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 97, 110, 107, 105, 110, 95, 73, 110, 108, 101, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 101, 99, 105, 102, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 101, 103, 105, 110, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 101, 115, 111, 108, 117, 116, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 105, 111, 95, 66, 114, 97, 110, 99, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 97, 110, 116, 97, 95, 73, 115, 97, 98, 101, 108, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 97, 110, 116, 97, 114, 101, 109, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 97, 110, 116, 105, 97, 103, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 97, 110, 116, 111, 95, 68, 111, 109, 105, 110, 103, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 97, 111, 95, 80, 97, 117, 108, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 99, 111, 114, 101, 115, 98, 121, 115, 117, 110, 100, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 105, 116, 107, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 66, 97, 114, 116, 104, 101, 108, 101, 109, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 74, 111, 104, 110, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 75, 105, 116, 116, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 76, 117, 99, 105, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 84, 104, 111, 109, 97, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 86, 105, 110, 99, 101, 110, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 119, 105, 102, 116, 95, 67, 117, 114, 114, 101, 110, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 101, 103, 117, 99, 105, 103, 97, 108, 112, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 104, 117, 108, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 104, 117, 110, 100, 101, 114, 95, 66, 97, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 105, 106, 117, 97, 110, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 111, 114, 111, 110, 116, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 111, 114, 116, 111, 108, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 86, 97, 110, 99, 111, 117, 118, 101, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 87, 104, 105, 116, 101, 104, 111, 114, 115, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 87, 105, 110, 110, 105, 112, 101, 103, 0, 65, 109, 101, 114, 105, 99, 97, 58, 89, 97, 107, 117, 116, 97, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 89, 101, 108, 108, 111, 119, 107, 110, 105, 102, 101, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 67, 97, 115, 101, 121, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 68, 97, 118, 105, 115, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 68, 117, 109, 111, 110, 116, 68, 85, 114, 118, 105, 108, 108, 101, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 77, 97, 99, 113, 117, 97, 114, 105, 101, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 77, 97, 119, 115, 111, 110, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 77, 99, 77, 117, 114, 100, 111, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 80, 97, 108, 109, 101, 114, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 82, 111, 116, 104, 101, 114, 97, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 83, 121, 111, 119, 97, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 84, 114, 111, 108, 108, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 86, 111, 115, 116, 111, 107, 0, 65, 114, 99, 116, 105, 99, 58, 76, 111, 110, 103, 121, 101, 97, 114, 98, 121, 101, 110, 0, 65, 115, 105, 97, 58, 65, 100, 101, 110, 0, 65, 115, 105, 97, 58, 65, 108, 109, 97, 116, 121, 0, 65, 115, 105, 97, 58, 65, 109, 109, 97, 110, 0, 65, 115, 105, 97, 58, 65, 110, 97, 100, 121, 114, 0, 65, 115, 105, 97, 58, 65, 113, 116, 97, 117, 0, 65, 115, 105, 97, 58, 65, 113, 116, 111, 98, 101, 0, 65, 115, 105, 97, 58, 65, 115, 104, 103, 97, 98, 97, 116, 0, 65, 115, 105, 97, 58, 65, 116, 121, 114, 97, 117, 0, 65, 115, 105, 97, 58, 66, 97, 103, 104, 100, 97, 100, 0, 65, 115, 105, 97, 58, 66, 97, 104, 114, 97, 105, 110, 0, 65, 115, 105, 97, 58, 66, 97, 107, 117, 0, 65, 115, 105, 97, 58, 66, 97, 110, 103, 107, 111, 107, 0, 65, 115, 105, 97, 58, 66, 101, 105, 114, 117, 116, 0, 65, 115, 105, 97, 58, 66, 105, 115, 104, 107, 101, 107, 0, 65, 115, 105, 97, 58, 66, 114, 117, 110, 101, 105, 0, 65, 115, 105, 97, 58, 67, 97, 108, 99, 117, 116, 116, 97, 0, 65, 115, 105, 97, 58, 67, 104, 105, 116, 97, 0, 65, 115, 105, 97, 58, 67, 104, 111, 105, 98, 97, 108, 115, 97, 110, 0, 65, 115, 105, 97, 58, 67, 111, 108, 111, 109, 98, 111, 0, 65, 115, 105, 97, 58, 68, 97, 109, 97, 115, 99, 117, 115, 0, 65, 115, 105, 97, 58, 68, 104, 97, 107, 97, 0, 65, 115, 105, 97, 58, 68, 105, 108, 105, 0, 65, 115, 105, 97, 58, 68, 117, 98, 97, 105, 0, 65, 115, 105, 97, 58, 68, 117, 115, 104, 97, 110, 98, 101, 0, 65, 115, 105, 97, 58, 70, 97, 109, 97, 103, 117, 115, 116, 97, 0, 65, 115, 105, 97, 58, 71, 97, 122, 97, 0, 65, 115, 105, 97, 58, 72, 101, 98, 114, 111, 110, 0, 65, 115, 105, 97, 58, 72, 111, 110, 103, 95, 75, 111, 110, 103, 0, 65, 115, 105, 97, 58, 72, 111, 118, 100, 0, 65, 115, 105, 97, 58, 73, 114, 107, 117, 116, 115, 107, 0, 65, 115, 105, 97, 58, 74, 97, 107, 97, 114, 116, 97, 0, 65, 115, 105, 97, 58, 74, 97, 121, 97, 112, 117, 114, 97, 0, 65, 115, 105, 97, 58, 74, 101, 114, 117, 115, 97, 108, 101, 109, 0, 65, 115, 105, 97, 58, 75, 97, 98, 117, 108, 0, 65, 115, 105, 97, 58, 75, 97, 109, 99, 104, 97, 116, 107, 97, 0, 65, 115, 105, 97, 58, 75, 97, 114, 97, 99, 104, 105, 0, 65, 115, 105, 97, 58, 75, 97, 116, 109, 97, 110, 100, 117, 0, 65, 115, 105, 97, 58, 75, 104, 97, 110, 100, 121, 103, 97, 0, 65, 115, 105, 97, 58, 75, 114, 97, 115, 110, 111, 121, 97, 114, 115, 107, 0, 65, 115, 105, 97, 58, 75, 117, 97, 108, 97, 95, 76, 117, 109, 112, 117, 114, 0, 65, 115, 105, 97, 58, 75, 117, 99, 104, 105, 110, 103, 0, 65, 115, 105, 97, 58, 75, 117, 119, 97, 105, 116, 0, 65, 115, 105, 97, 58, 77, 97, 99, 97, 117, 0, 65, 115, 105, 97, 58, 77, 97, 103, 97, 100, 97, 110, 0, 65, 115, 105, 97, 58, 77, 97, 107, 97, 115, 115, 97, 114, 0, 65, 115, 105, 97, 58, 77, 97, 110, 105, 108, 97, 0, 65, 115, 105, 97, 58, 77, 117, 115, 99, 97, 116, 0, 65, 115, 105, 97, 58, 78, 105, 99, 111, 115, 105, 97, 0, 65, 115, 105, 97, 58, 78, 111, 118, 111, 107, 117, 122, 110, 101, 116, 115, 107, 0, 65, 115, 105, 97, 58, 78, 111, 118, 111, 115, 105, 98, 105, 114, 115, 107, 0, 65, 115, 105, 97, 58, 79, 109, 115, 107, 0, 65, 115, 105, 97, 58, 79, 114, 97, 108, 0, 65, 115, 105, 97, 58, 80, 104, 110, 111, 109, 95, 80, 101, 110, 104, 0, 65, 115, 105, 97, 58, 80, 111, 110, 116, 105, 97, 110, 97, 107, 0, 65, 115, 105, 97, 58, 80, 121, 111, 110, 103, 121, 97, 110, 103, 0, 65, 115, 105, 97, 58, 81, 97, 116, 97, 114, 0, 65, 115, 105, 97, 58, 81, 111, 115, 116, 97, 110, 97, 121, 0, 65, 115, 105, 97, 58, 81, 121, 122, 121, 108, 111, 114, 100, 97, 0, 65, 115, 105, 97, 58, 82, 97, 110, 103, 111, 111, 110, 0, 65, 115, 105, 97, 58, 82, 105, 121, 97, 100, 104, 0, 65, 115, 105, 97, 58, 83, 97, 105, 103, 111, 110, 0, 65, 115, 105, 97, 58, 83, 97, 107, 104, 97, 108, 105, 110, 0, 65, 115, 105, 97, 58, 83, 97, 109, 97, 114, 107, 97, 110, 100, 0, 65, 115, 105, 97, 58, 83, 101, 111, 117, 108, 0, 65, 115, 105, 97, 58, 83, 104, 97, 110, 103, 104, 97, 105, 0, 65, 115, 105, 97, 58, 83, 105, 110, 103, 97, 112, 111, 114, 101, 0, 65, 115, 105, 97, 58, 83, 114, 101, 100, 110, 101, 107, 111, 108, 121, 109, 115, 107, 0, 65, 115, 105, 97, 58, 84, 97, 105, 112, 101, 105, 0, 65, 115, 105, 97, 58, 84, 97, 115, 104, 107, 101, 110, 116, 0, 65, 115, 105, 97, 58, 84, 98, 105, 108, 105, 115, 105, 0, 65, 115, 105, 97, 58, 84, 101, 104, 114, 97, 110, 0, 65, 115, 105, 97, 58, 84, 104, 105, 109, 112, 104, 117, 0, 65, 115, 105, 97, 58, 84, 111, 107, 121, 111, 0, 65, 115, 105, 97, 58, 85, 108, 97, 97, 110, 98, 97, 97, 116, 97, 114, 0, 65, 115, 105, 97, 58, 85, 114, 117, 109, 113, 105, 0, 65, 115, 105, 97, 58, 85, 115, 116, 45, 78, 101, 114, 97, 0, 65, 115, 105, 97, 58, 86, 105, 101, 110, 116, 105, 97, 110, 101, 0, 65, 115, 105, 97, 58, 86, 108, 97, 100, 105, 118, 111, 115, 116, 111, 107, 0, 65, 115, 105, 97, 58, 89, 97, 107, 117, 116, 115, 107, 0, 65, 115, 105, 97, 58, 89, 101, 107, 97, 116, 101, 114, 105, 110, 98, 117, 114, 103, 0, 65, 115, 105, 97, 58, 89, 101, 114, 101, 118, 97, 110, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 65, 122, 111, 114, 101, 115, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 66, 101, 114, 109, 117, 100, 97, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 67, 97, 110, 97, 114, 121, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 67, 97, 112, 101, 95, 86, 101, 114, 100, 101, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 70, 97, 101, 114, 111, 101, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 77, 97, 100, 101, 105, 114, 97, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 82, 101, 121, 107, 106, 97, 118, 105, 107, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 83, 111, 117, 116, 104, 95, 71, 101, 111, 114, 103, 105, 97, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 83, 116, 95, 72, 101, 108, 101, 110, 97, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 83, 116, 97, 110, 108, 101, 121, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 65, 100, 101, 108, 97, 105, 100, 101, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 66, 114, 105, 115, 98, 97, 110, 101, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 66, 114, 111, 107, 101, 110, 95, 72, 105, 108, 108, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 67, 117, 114, 114, 105, 101, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 68, 97, 114, 119, 105, 110, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 69, 117, 99, 108, 97, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 72, 111, 98, 97, 114, 116, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 76, 105, 110, 100, 101, 109, 97, 110, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 76, 111, 114, 100, 95, 72, 111, 119, 101, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 77, 101, 108, 98, 111, 117, 114, 110, 101, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 80, 101, 114, 116, 104, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 83, 121, 100, 110, 101, 121, 0, 67, 83, 84, 54, 67, 68, 84, 0, 69, 83, 84, 53, 69, 68, 84, 0, 69, 116, 99, 58, 71, 77, 84, 0, 69, 117, 114, 111, 112, 101, 58, 65, 109, 115, 116, 101, 114, 100, 97, 109, 0, 69, 117, 114, 111, 112, 101, 58, 65, 110, 100, 111, 114, 114, 97, 0, 69, 117, 114, 111, 112, 101, 58, 65, 115, 116, 114, 97, 107, 104, 97, 110, 0, 69, 117, 114, 111, 112, 101, 58, 65, 116, 104, 101, 110, 115, 0, 69, 117, 114, 111, 112, 101, 58, 66, 101, 108, 103, 114, 97, 100, 101, 0, 69, 117, 114, 111, 112, 101, 58, 66, 101, 114, 108, 105, 110, 0, 69, 117, 114, 111, 112, 101, 58, 66, 114, 97, 116, 105, 115, 108, 97, 118, 97, 0, 69, 117, 114, 111, 112, 101, 58, 66, 114, 117, 115, 115, 101, 108, 115, 0, 69, 117, 114, 111, 112, 101, 58, 66, 117, 99, 104, 97, 114, 101, 115, 116, 0, 69, 117, 114, 111, 112, 101, 58, 66, 117, 100, 97, 112, 101, 115, 116, 0, 69, 117, 114, 111, 112, 101, 58, 66, 117, 115, 105, 110, 103, 101, 110, 0, 69, 117, 114, 111, 112, 101, 58, 67, 104, 105, 115, 105, 110, 97, 117, 0, 69, 117, 114, 111, 112, 101, 58, 67, 111, 112, 101, 110, 104, 97, 103, 101, 110, 0, 69, 117, 114, 111, 112, 101, 58, 68, 117, 98, 108, 105, 110, 0, 69, 117, 114, 111, 112, 101, 58, 71, 105, 98, 114, 97, 108, 116, 97, 114, 0, 69, 117, 114, 111, 112, 101, 58, 71, 117, 101, 114, 110, 115, 101, 121, 0, 69, 117, 114, 111, 112, 101, 58, 72, 101, 108, 115, 105, 110, 107, 105, 0, 69, 117, 114, 111, 112, 101, 58, 73, 115, 108, 101, 95, 111, 102, 95, 77, 97, 110, 0, 69, 117, 114, 111, 112, 101, 58, 73, 115, 116, 97, 110, 98, 117, 108, 0, 69, 117, 114, 111, 112, 101, 58, 74, 101, 114, 115, 101, 121, 0, 69, 117, 114, 111, 112, 101, 58, 75, 97, 108, 105, 110, 105, 110, 103, 114, 97, 100, 0, 69, 117, 114, 111, 112, 101, 58, 75, 105, 101, 118, 0, 69, 117, 114, 111, 112, 101, 58, 76, 105, 115, 98, 111, 110, 0, 69, 117, 114, 111, 112, 101, 58, 76, 106, 117, 98, 108, 106, 97, 110, 97, 0, 69, 117, 114, 111, 112, 101, 58, 76, 111, 110, 100, 111, 110, 0, 69, 117, 114, 111, 112, 101, 58, 76, 117, 120, 101, 109, 98, 111, 117, 114, 103, 0, 69, 117, 114, 111, 112, 101, 58, 77, 97, 100, 114, 105, 100, 0, 69, 117, 114, 111, 112, 101, 58, 77, 97, 108, 116, 97, 0, 69, 117, 114, 111, 112, 101, 58, 77, 97, 114, 105, 101, 104, 97, 109, 110, 0, 69, 117, 114, 111, 112, 101, 58, 77, 105, 110, 115, 107, 0, 69, 117, 114, 111, 112, 101, 58, 77, 111, 110, 97, 99, 111, 0, 69, 117, 114, 111, 112, 101, 58, 77, 111, 115, 99, 111, 119, 0, 69, 117, 114, 111, 112, 101, 58, 79, 115, 108, 111, 0, 69, 117, 114, 111, 112, 101, 58, 80, 97, 114, 105, 115, 0, 69, 117, 114, 111, 112, 101, 58, 80, 111, 100, 103, 111, 114, 105, 99, 97, 0, 69, 117, 114, 111, 112, 101, 58, 80, 114, 97, 103, 117, 101, 0, 69, 117, 114, 111, 112, 101, 58, 82, 105, 103, 97, 0, 69, 117, 114, 111, 112, 101, 58, 82, 111, 109, 101, 0, 69, 117, 114, 111, 112, 101, 58, 83, 97, 109, 97, 114, 97, 0, 69, 117, 114, 111, 112, 101, 58, 83, 97, 110, 95, 77, 97, 114, 105, 110, 111, 0, 69, 117, 114, 111, 112, 101, 58, 83, 97, 114, 97, 106, 101, 118, 111, 0, 69, 117, 114, 111, 112, 101, 58, 83, 97, 114, 97, 116, 111, 118, 0, 69, 117, 114, 111, 112, 101, 58, 83, 105, 109, 102, 101, 114, 111, 112, 111, 108, 0, 69, 117, 114, 111, 112, 101, 58, 83, 107, 111, 112, 106, 101, 0, 69, 117, 114, 111, 112, 101, 58, 83, 111, 102, 105, 97, 0, 69, 117, 114, 111, 112, 101, 58, 83, 116, 111, 99, 107, 104, 111, 108, 109, 0, 69, 117, 114, 111, 112, 101, 58, 84, 97, 108, 108, 105, 110, 110, 0, 69, 117, 114, 111, 112, 101, 58, 84, 105, 114, 97, 110, 101, 0, 69, 117, 114, 111, 112, 101, 58, 85, 108, 121, 97, 110, 111, 118, 115, 107, 0, 69, 117, 114, 111, 112, 101, 58, 85, 122, 104, 103, 111, 114, 111, 100, 0, 69, 117, 114, 111, 112, 101, 58, 86, 97, 100, 117, 122, 0, 69, 117, 114, 111, 112, 101, 58, 86, 97, 116, 105, 99, 97, 110, 0, 69, 117, 114, 111, 112, 101, 58, 86, 105, 101, 110, 110, 97, 0, 69, 117, 114, 111, 112, 101, 58, 86, 105, 108, 110, 105, 117, 115, 0, 69, 117, 114, 111, 112, 101, 58, 86, 111, 108, 103, 111, 103, 114, 97, 100, 0, 69, 117, 114, 111, 112, 101, 58, 87, 97, 114, 115, 97, 119, 0, 69, 117, 114, 111, 112, 101, 58, 90, 97, 103, 114, 101, 98, 0, 69, 117, 114, 111, 112, 101, 58, 90, 97, 112, 111, 114, 111, 122, 104, 121, 101, 0, 69, 117, 114, 111, 112, 101, 58, 90, 117, 114, 105, 99, 104, 0, 73, 110, 100, 105, 97, 110, 58, 65, 110, 116, 97, 110, 97, 110, 97, 114, 105, 118, 111, 0, 73, 110, 100, 105, 97, 110, 58, 67, 104, 97, 103, 111, 115, 0, 73, 110, 100, 105, 97, 110, 58, 67, 104, 114, 105, 115, 116, 109, 97, 115, 0, 73, 110, 100, 105, 97, 110, 58, 67, 111, 99, 111, 115, 0, 73, 110, 100, 105, 97, 110, 58, 67, 111, 109, 111, 114, 111, 0, 73, 110, 100, 105, 97, 110, 58, 75, 101, 114, 103, 117, 101, 108, 101, 110, 0, 73, 110, 100, 105, 97, 110, 58, 77, 97, 104, 101, 0, 73, 110, 100, 105, 97, 110, 58, 77, 97, 108, 100, 105, 118, 101, 115, 0, 73, 110, 100, 105, 97, 110, 58, 77, 97, 117, 114, 105, 116, 105, 117, 115, 0, 73, 110, 100, 105, 97, 110, 58, 77, 97, 121, 111, 116, 116, 101, 0, 73, 110, 100, 105, 97, 110, 58, 82, 101, 117, 110, 105, 111, 110, 0, 77, 83, 84, 55, 77, 68, 84, 0, 80, 83, 84, 56, 80, 68, 84, 0, 80, 97, 99, 105, 102, 105, 99, 58, 65, 112, 105, 97, 0, 80, 97, 99, 105, 102, 105, 99, 58, 65, 117, 99, 107, 108, 97, 110, 100, 0, 80, 97, 99, 105, 102, 105, 99, 58, 66, 111, 117, 103, 97, 105, 110, 118, 105, 108, 108, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 67, 104, 97, 116, 104, 97, 109, 0, 80, 97, 99, 105, 102, 105, 99, 58, 69, 97, 115, 116, 101, 114, 0, 80, 97, 99, 105, 102, 105, 99, 58, 69, 102, 97, 116, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 69, 110, 100, 101, 114, 98, 117, 114, 121, 0, 80, 97, 99, 105, 102, 105, 99, 58, 70, 97, 107, 97, 111, 102, 111, 0, 80, 97, 99, 105, 102, 105, 99, 58, 70, 105, 106, 105, 0, 80, 97, 99, 105, 102, 105, 99, 58, 70, 117, 110, 97, 102, 117, 116, 105, 0, 80, 97, 99, 105, 102, 105, 99, 58, 71, 97, 108, 97, 112, 97, 103, 111, 115, 0, 80, 97, 99, 105, 102, 105, 99, 58, 71, 97, 109, 98, 105, 101, 114, 0, 80, 97, 99, 105, 102, 105, 99, 58, 71, 117, 97, 100, 97, 108, 99, 97, 110, 97, 108, 0, 80, 97, 99, 105, 102, 105, 99, 58, 71, 117, 97, 109, 0, 80, 97, 99, 105, 102, 105, 99, 58, 72, 111, 110, 111, 108, 117, 108, 117, 0, 80, 97, 99, 105, 102, 105, 99, 58, 74, 111, 104, 110, 115, 116, 111, 110, 0, 80, 97, 99, 105, 102, 105, 99, 58, 75, 105, 114, 105, 116, 105, 109, 97, 116, 105, 0, 80, 97, 99, 105, 102, 105, 99, 58, 75, 111, 115, 114, 97, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 75, 119, 97, 106, 97, 108, 101, 105, 110, 0, 80, 97, 99, 105, 102, 105, 99, 58, 77, 97, 106, 117, 114, 111, 0, 80, 97, 99, 105, 102, 105, 99, 58, 77, 97, 114, 113, 117, 101, 115, 97, 115, 0, 80, 97, 99, 105, 102, 105, 99, 58, 77, 105, 100, 119, 97, 121, 0, 80, 97, 99, 105, 102, 105, 99, 58, 78, 97, 117, 114, 117, 0, 80, 97, 99, 105, 102, 105, 99, 58, 78, 105, 117, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 78, 111, 114, 102, 111, 108, 107, 0, 80, 97, 99, 105, 102, 105, 99, 58, 78, 111, 117, 109, 101, 97, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 97, 103, 111, 95, 80, 97, 103, 111, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 97, 108, 97, 117, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 105, 116, 99, 97, 105, 114, 110, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 111, 110, 97, 112, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 111, 114, 116, 95, 77, 111, 114, 101, 115, 98, 121, 0, 80, 97, 99, 105, 102, 105, 99, 58, 82, 97, 114, 111, 116, 111, 110, 103, 97, 0, 80, 97, 99, 105, 102, 105, 99, 58, 83, 97, 105, 112, 97, 110, 0, 80, 97, 99, 105, 102, 105, 99, 58, 84, 97, 104, 105, 116, 105, 0, 80, 97, 99, 105, 102, 105, 99, 58, 84, 97, 114, 97, 119, 97, 0, 80, 97, 99, 105, 102, 105, 99, 58, 84, 111, 110, 103, 97, 116, 97, 112, 117, 0, 80, 97, 99, 105, 102, 105, 99, 58, 84, 114, 117, 107, 0, 80, 97, 99, 105, 102, 105, 99, 58, 87, 97, 107, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 87, 97, 108, 108, 105, 115, 0, 112, 114, 105, 109, 97, 114, 121, 90, 111, 110, 101, 115, 0, 67, 76, 0, 67, 78, 0, 69, 67, 0, 77, 72, 0, 77, 89, 0, 78, 90, 0, 80, 84, 0, 85, 65, 0, 85, 90, 0, 0, 0, 71, 0, 77, 0, 84, 0, 0, 0, 65, 0, 99, 0, 114, 0, 101, 0, 0, 0, 71, 0, 117, 0, 108, 0, 102, 0, 0, 0, 67, 0, 111, 0, 111, 0, 107, 0, 0, 0, 67, 0, 117, 0, 98, 0, 97, 0, 0, 0, 73, 0, 114, 0, 97, 0, 110, 0, 0, 0, 80, 0, 101, 0, 114, 0, 117, 0, 0, 0, 89, 0, 117, 0, 107, 0, 111, 0, 110, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 0, 0, 67, 0, 104, 0, 105, 0, 108, 0, 101, 0, 0, 0, 75, 0, 111, 0, 114, 0, 101, 0, 97, 0, 0, 0, 67, 0, 104, 0, 105, 0, 110, 0, 97, 0, 0, 0, 83, 0, 97, 0, 109, 0, 111, 0, 97, 0, 0, 0, 68, 0, 97, 0, 99, 0, 99, 0, 97, 0, 0, 0, 73, 0, 114, 0, 105, 0, 115, 0, 104, 0, 0, 0, 74, 0, 97, 0, 112, 0, 97, 0, 110, 0, 0, 0, 76, 0, 97, 0, 110, 0, 107, 0, 97, 0, 0, 0, 78, 0, 101, 0, 112, 0, 97, 0, 108, 0, 0, 0, 84, 0, 111, 0, 110, 0, 103, 0, 97, 0, 0, 0, 65, 0, 109, 0, 97, 0, 122, 0, 111, 0, 110, 0, 0, 0, 65, 0, 108, 0, 97, 0, 115, 0, 107, 0, 97, 0, 0, 0, 66, 0, 101, 0, 114, 0, 105, 0, 110, 0, 103, 0, 0, 0, 73, 0, 115, 0, 114, 0, 97, 0, 101, 0, 108, 0, 0, 0, 84, 0, 117, 0, 114, 0, 107, 0, 101, 0, 121, 0, 0, 0, 66, 0, 104, 0, 117, 0, 116, 0, 97, 0, 110, 0, 0, 0, 66, 0, 111, 0, 114, 0, 110, 0, 101, 0, 111, 0, 0, 0, 70, 0, 114, 0, 117, 0, 110, 0, 122, 0, 101, 0, 0, 0, 77, 0, 97, 0, 108, 0, 97, 0, 121, 0, 97, 0, 0, 0, 84, 0, 117, 0, 118, 0, 97, 0, 108, 0, 117, 0, 0, 0, 85, 0, 114, 0, 97, 0, 108, 0, 115, 0, 107, 0, 0, 0, 65, 0, 114, 0, 97, 0, 98, 0, 105, 0, 97, 0, 110, 0, 0, 0, 66, 0, 114, 0, 105, 0, 116, 0, 105, 0, 115, 0, 104, 0, 0, 0, 69, 0, 99, 0, 117, 0, 97, 0, 100, 0, 111, 0, 114, 0, 0, 0, 65, 0, 114, 0, 109, 0, 101, 0, 110, 0, 105, 0, 97, 0, 0, 0, 66, 0, 111, 0, 108, 0, 105, 0, 118, 0, 105, 0, 97, 0, 0, 0, 76, 0, 105, 0, 98, 0, 101, 0, 114, 0, 105, 0, 97, 0, 0, 0, 77, 0, 121, 0, 97, 0, 110, 0, 109, 0, 97, 0, 114, 0, 0, 0, 83, 0, 111, 0, 108, 0, 111, 0, 109, 0, 111, 0, 110, 0, 0, 0, 84, 0, 111, 0, 107, 0, 101, 0, 108, 0, 97, 0, 117, 0, 0, 0, 85, 0, 114, 0, 117, 0, 103, 0, 117, 0, 97, 0, 121, 0, 0, 0, 86, 0, 97, 0, 110, 0, 117, 0, 97, 0, 116, 0, 117, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 0, 0, 66, 0, 114, 0, 97, 0, 115, 0, 105, 0, 108, 0, 105, 0, 97, 0, 0, 0, 67, 0, 104, 0, 97, 0, 109, 0, 111, 0, 114, 0, 114, 0, 111, 0, 0, 0, 77, 0, 97, 0, 108, 0, 97, 0, 121, 0, 115, 0, 105, 0, 97, 0, 0, 0, 77, 0, 111, 0, 110, 0, 103, 0, 111, 0, 108, 0, 105, 0, 97, 0, 0, 0, 67, 0, 111, 0, 108, 0, 111, 0, 109, 0, 98, 0, 105, 0, 97, 0, 0, 0, 70, 0, 97, 0, 108, 0, 107, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 80, 0, 97, 0, 107, 0, 105, 0, 115, 0, 116, 0, 97, 0, 110, 0, 0, 0, 80, 0, 97, 0, 114, 0, 97, 0, 103, 0, 117, 0, 97, 0, 121, 0, 0, 0, 83, 0, 117, 0, 114, 0, 105, 0, 110, 0, 97, 0, 109, 0, 101, 0, 0, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 0, 0, 73, 0, 110, 0, 100, 0, 111, 0, 99, 0, 104, 0, 105, 0, 110, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 107, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 79, 0, 114, 0, 97, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 118, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 79, 0, 109, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 104, 0, 107, 0, 104, 0, 97, 0, 98, 0, 97, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 100, 0, 101, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 105, 0, 108, 0, 105, 0, 0, 0, 68, 0, 111, 0, 109, 0, 105, 0, 110, 0, 105, 0, 99, 0, 97, 0, 110, 0, 0, 0, 75, 0, 105, 0, 122, 0, 105, 0, 108, 0, 111, 0, 114, 0, 100, 0, 97, 0, 0, 0, 75, 0, 117, 0, 121, 0, 98, 0, 121, 0, 115, 0, 104, 0, 101, 0, 118, 0, 0, 0, 75, 0, 121, 0, 114, 0, 103, 0, 121, 0, 115, 0, 116, 0, 97, 0, 110, 0, 0, 0, 86, 0, 101, 0, 110, 0, 101, 0, 122, 0, 117, 0, 101, 0, 108, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 113, 0, 116, 0, 97, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 104, 0, 97, 0, 107, 0, 97, 0, 0, 0, 69, 0, 97, 0, 115, 0, 116, 0, 95, 0, 84, 0, 105, 0, 109, 0, 111, 0, 114, 0, 0, 0, 85, 0, 122, 0, 98, 0, 101, 0, 107, 0, 105, 0, 115, 0, 116, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 99, 0, 97, 0, 117, 0, 0, 0, 65, 0, 107, 0, 116, 0, 121, 0, 117, 0, 98, 0, 105, 0, 110, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 109, 0, 109, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 117, 0, 98, 0, 97, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 98, 0, 117, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 81, 0, 97, 0, 116, 0, 97, 0, 114, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 101, 0, 111, 0, 117, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 111, 0, 107, 0, 121, 0, 111, 0, 0, 0, 65, 0, 122, 0, 101, 0, 114, 0, 98, 0, 97, 0, 105, 0, 106, 0, 97, 0, 110, 0, 0, 0, 66, 0, 97, 0, 110, 0, 103, 0, 108, 0, 97, 0, 100, 0, 101, 0, 115, 0, 104, 0, 0, 0, 83, 0, 101, 0, 121, 0, 99, 0, 104, 0, 101, 0, 108, 0, 108, 0, 101, 0, 115, 0, 0, 0, 83, 0, 104, 0, 101, 0, 118, 0, 99, 0, 104, 0, 101, 0, 110, 0, 107, 0, 111, 0, 0, 0, 83, 0, 118, 0, 101, 0, 114, 0, 100, 0, 108, 0, 111, 0, 118, 0, 115, 0, 107, 0, 0, 0, 84, 0, 97, 0, 106, 0, 105, 0, 107, 0, 105, 0, 115, 0, 116, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 113, 0, 116, 0, 111, 0, 98, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 108, 0, 109, 0, 97, 0, 116, 0, 121, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 110, 0, 97, 0, 100, 0, 121, 0, 114, 0, 0, 0, 78, 0, 101, 0, 119, 0, 95, 0, 90, 0, 101, 0, 97, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 114, 0, 117, 0, 110, 0, 101, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 97, 0, 105, 0, 112, 0, 101, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 114, 0, 117, 0, 109, 0, 113, 0, 105, 0, 0, 0, 65, 0, 102, 0, 103, 0, 104, 0, 97, 0, 110, 0, 105, 0, 115, 0, 116, 0, 97, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 109, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 101, 0, 105, 0, 114, 0, 117, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 117, 0, 119, 0, 97, 0, 105, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 110, 0, 105, 0, 108, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 117, 0, 115, 0, 99, 0, 97, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 82, 0, 105, 0, 121, 0, 97, 0, 100, 0, 104, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 101, 0, 104, 0, 114, 0, 97, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 75, 0, 105, 0, 101, 0, 118, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 79, 0, 115, 0, 108, 0, 111, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 82, 0, 111, 0, 109, 0, 101, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 104, 0, 101, 0, 0, 0, 80, 0, 104, 0, 105, 0, 108, 0, 105, 0, 112, 0, 112, 0, 105, 0, 110, 0, 101, 0, 115, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 97, 0, 107, 0, 117, 0, 116, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 103, 0, 97, 0, 100, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 114, 0, 97, 0, 99, 0, 104, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 98, 0, 105, 0, 108, 0, 105, 0, 115, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 101, 0, 114, 0, 101, 0, 118, 0, 97, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 117, 0, 97, 0, 109, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 73, 0, 114, 0, 107, 0, 117, 0, 116, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 105, 0, 115, 0, 104, 0, 107, 0, 101, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 111, 0, 108, 0, 111, 0, 109, 0, 98, 0, 111, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 117, 0, 99, 0, 104, 0, 105, 0, 110, 0, 103, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 111, 0, 99, 0, 111, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 65, 0, 112, 0, 105, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 70, 0, 105, 0, 106, 0, 105, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 105, 0, 117, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 114, 0, 117, 0, 107, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 87, 0, 97, 0, 107, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 99, 0, 99, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 105, 0, 114, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 107, 0, 97, 0, 114, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 97, 0, 103, 0, 111, 0, 115, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 117, 0, 110, 0, 105, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 100, 0, 97, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 105, 0, 109, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 103, 0, 104, 0, 100, 0, 97, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 104, 0, 114, 0, 97, 0, 105, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 110, 0, 103, 0, 107, 0, 111, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 74, 0, 97, 0, 107, 0, 97, 0, 114, 0, 116, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 78, 0, 105, 0, 99, 0, 111, 0, 115, 0, 105, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 82, 0, 97, 0, 110, 0, 103, 0, 111, 0, 111, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 104, 0, 105, 0, 109, 0, 112, 0, 104, 0, 117, 0, 0, 0, 68, 0, 117, 0, 116, 0, 99, 0, 104, 0, 95, 0, 71, 0, 117, 0, 105, 0, 97, 0, 110, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 97, 0, 108, 0, 116, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 105, 0, 110, 0, 115, 0, 107, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 80, 0, 97, 0, 114, 0, 105, 0, 115, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 111, 0, 102, 0, 105, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 97, 0, 100, 0, 117, 0, 122, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 95, 0, 79, 0, 99, 0, 101, 0, 97, 0, 110, 0, 0, 0, 76, 0, 105, 0, 110, 0, 101, 0, 95, 0, 73, 0, 115, 0, 108, 0, 97, 0, 110, 0, 100, 0, 115, 0, 0, 0, 78, 0, 101, 0, 119, 0, 102, 0, 111, 0, 117, 0, 110, 0, 100, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 84, 0, 117, 0, 114, 0, 107, 0, 109, 0, 101, 0, 110, 0, 105, 0, 115, 0, 116, 0, 97, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 111, 0, 115, 0, 99, 0, 111, 0, 119, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 97, 0, 115, 0, 104, 0, 107, 0, 101, 0, 110, 0, 116, 0, 0, 0, 65, 0, 108, 0, 97, 0, 115, 0, 107, 0, 97, 0, 95, 0, 72, 0, 97, 0, 119, 0, 97, 0, 105, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 117, 0, 115, 0, 104, 0, 97, 0, 110, 0, 98, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 109, 0, 97, 0, 114, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 115, 0, 104, 0, 103, 0, 97, 0, 98, 0, 97, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 104, 0, 97, 0, 110, 0, 103, 0, 104, 0, 97, 0, 105, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 101, 0, 114, 0, 108, 0, 105, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 68, 0, 117, 0, 98, 0, 108, 0, 105, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 111, 0, 110, 0, 100, 0, 111, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 97, 0, 100, 0, 114, 0, 105, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 97, 0, 107, 0, 104, 0, 97, 0, 108, 0, 105, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 97, 0, 117, 0, 114, 0, 117, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 97, 0, 108, 0, 97, 0, 117, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 115, 0, 109, 0, 101, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 109, 0, 97, 0, 107, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 110, 0, 103, 0, 117, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 110, 0, 106, 0, 117, 0, 108, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 111, 0, 117, 0, 97, 0, 108, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 114, 0, 97, 0, 114, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 105, 0, 103, 0, 97, 0, 108, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 117, 0, 97, 0, 110, 0, 100, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 117, 0, 115, 0, 97, 0, 107, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 108, 0, 97, 0, 98, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 112, 0, 117, 0, 116, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 115, 0, 101, 0, 114, 0, 117, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 105, 0, 97, 0, 109, 0, 101, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 117, 0, 98, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 104, 0, 117, 0, 108, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 97, 0, 108, 0, 99, 0, 117, 0, 116, 0, 116, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 97, 0, 109, 0, 97, 0, 115, 0, 99, 0, 117, 0, 115, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 74, 0, 97, 0, 121, 0, 97, 0, 112, 0, 117, 0, 114, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 116, 0, 109, 0, 97, 0, 110, 0, 100, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 107, 0, 97, 0, 115, 0, 115, 0, 97, 0, 114, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 116, 0, 104, 0, 101, 0, 110, 0, 115, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 105, 0, 115, 0, 98, 0, 111, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 111, 0, 110, 0, 97, 0, 99, 0, 111, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 80, 0, 114, 0, 97, 0, 103, 0, 117, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 107, 0, 111, 0, 112, 0, 106, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 84, 0, 105, 0, 114, 0, 97, 0, 110, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 105, 0, 101, 0, 110, 0, 110, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 87, 0, 97, 0, 114, 0, 115, 0, 97, 0, 119, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 90, 0, 97, 0, 103, 0, 114, 0, 101, 0, 98, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 90, 0, 117, 0, 114, 0, 105, 0, 99, 0, 104, 0, 0, 0, 70, 0, 114, 0, 101, 0, 110, 0, 99, 0, 104, 0, 95, 0, 71, 0, 117, 0, 105, 0, 97, 0, 110, 0, 97, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 104, 0, 97, 0, 103, 0, 111, 0, 115, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 111, 0, 109, 0, 111, 0, 114, 0, 111, 0, 0, 0, 78, 0, 101, 0, 119, 0, 95, 0, 67, 0, 97, 0, 108, 0, 101, 0, 100, 0, 111, 0, 110, 0, 105, 0, 97, 0, 0, 0, 78, 0, 111, 0, 114, 0, 116, 0, 104, 0, 95, 0, 77, 0, 97, 0, 114, 0, 105, 0, 97, 0, 110, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 69, 0, 102, 0, 97, 0, 116, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 95, 0, 67, 0, 101, 0, 110, 0, 116, 0, 114, 0, 97, 0, 108, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 95, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 95, 0, 67, 0, 101, 0, 110, 0, 116, 0, 114, 0, 97, 0, 108, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 95, 0, 87, 0, 101, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 95, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 95, 0, 87, 0, 101, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 95, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 81, 0, 121, 0, 122, 0, 121, 0, 108, 0, 111, 0, 114, 0, 100, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 97, 0, 109, 0, 97, 0, 114, 0, 107, 0, 97, 0, 110, 0, 100, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 77, 0, 97, 0, 106, 0, 117, 0, 114, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 83, 0, 97, 0, 105, 0, 112, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 110, 0, 103, 0, 95, 0, 75, 0, 111, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 109, 0, 99, 0, 104, 0, 97, 0, 116, 0, 107, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 80, 0, 121, 0, 111, 0, 110, 0, 103, 0, 121, 0, 97, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 105, 0, 110, 0, 103, 0, 97, 0, 112, 0, 111, 0, 114, 0, 101, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 82, 0, 101, 0, 117, 0, 110, 0, 105, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 121, 0, 97, 0, 110, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 75, 0, 111, 0, 115, 0, 114, 0, 97, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 111, 0, 110, 0, 97, 0, 112, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 97, 0, 104, 0, 105, 0, 116, 0, 105, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 87, 0, 97, 0, 108, 0, 108, 0, 105, 0, 115, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 98, 0, 105, 0, 100, 0, 106, 0, 97, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 110, 0, 97, 0, 107, 0, 114, 0, 121, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 97, 0, 109, 0, 112, 0, 97, 0, 108, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 98, 0, 97, 0, 98, 0, 97, 0, 110, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 97, 0, 105, 0, 114, 0, 111, 0, 98, 0, 105, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 101, 0, 108, 0, 105, 0, 122, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 111, 0, 103, 0, 111, 0, 116, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 121, 0, 109, 0, 97, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 101, 0, 110, 0, 118, 0, 101, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 118, 0, 97, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 117, 0, 110, 0, 101, 0, 97, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 97, 0, 95, 0, 80, 0, 97, 0, 122, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 110, 0, 97, 0, 117, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 97, 0, 115, 0, 115, 0, 97, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 97, 0, 110, 0, 97, 0, 109, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 74, 0, 101, 0, 114, 0, 117, 0, 115, 0, 97, 0, 108, 0, 101, 0, 109, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 86, 0, 105, 0, 101, 0, 110, 0, 116, 0, 105, 0, 97, 0, 110, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 110, 0, 100, 0, 111, 0, 114, 0, 114, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 97, 0, 116, 0, 105, 0, 99, 0, 97, 0, 110, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 121, 0, 111, 0, 116, 0, 116, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 111, 0, 117, 0, 109, 0, 101, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 97, 0, 114, 0, 97, 0, 119, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 95, 0, 67, 0, 101, 0, 110, 0, 116, 0, 114, 0, 97, 0, 108, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 95, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 95, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 95, 0, 83, 0, 111, 0, 117, 0, 116, 0, 104, 0, 101, 0, 114, 0, 110, 0, 0, 0, 72, 0, 97, 0, 119, 0, 97, 0, 105, 0, 105, 0, 95, 0, 65, 0, 108, 0, 101, 0, 117, 0, 116, 0, 105, 0, 97, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 101, 0, 108, 0, 103, 0, 114, 0, 97, 0, 100, 0, 101, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 65, 0, 122, 0, 111, 0, 114, 0, 101, 0, 115, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 104, 0, 111, 0, 105, 0, 98, 0, 97, 0, 108, 0, 115, 0, 97, 0, 110, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 108, 0, 100, 0, 105, 0, 118, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 111, 0, 110, 0, 104, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 67, 0, 104, 0, 97, 0, 116, 0, 104, 0, 97, 0, 109, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 97, 0, 109, 0, 98, 0, 105, 0, 101, 0, 114, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 111, 0, 114, 0, 102, 0, 111, 0, 108, 0, 107, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 108, 0, 97, 0, 110, 0, 116, 0, 121, 0, 114, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 106, 0, 105, 0, 98, 0, 111, 0, 117, 0, 116, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 108, 0, 95, 0, 65, 0, 97, 0, 105, 0, 117, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 70, 0, 114, 0, 101, 0, 101, 0, 116, 0, 111, 0, 119, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 97, 0, 98, 0, 111, 0, 114, 0, 111, 0, 110, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 105, 0, 110, 0, 115, 0, 104, 0, 97, 0, 115, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 114, 0, 111, 0, 118, 0, 105, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 100, 0, 106, 0, 97, 0, 109, 0, 101, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 116, 0, 105, 0, 103, 0, 117, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 114, 0, 97, 0, 99, 0, 97, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 121, 0, 101, 0, 110, 0, 110, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 104, 0, 105, 0, 99, 0, 97, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 117, 0, 114, 0, 97, 0, 99, 0, 97, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 111, 0, 100, 0, 116, 0, 104, 0, 97, 0, 98, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 114, 0, 101, 0, 110, 0, 97, 0, 100, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 108, 0, 105, 0, 102, 0, 97, 0, 120, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 97, 0, 109, 0, 97, 0, 105, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 114, 0, 105, 0, 103, 0, 111, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 105, 0, 106, 0, 117, 0, 97, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 111, 0, 114, 0, 111, 0, 110, 0, 116, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 111, 0, 114, 0, 116, 0, 111, 0, 108, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 89, 0, 97, 0, 107, 0, 117, 0, 116, 0, 97, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 80, 0, 104, 0, 110, 0, 111, 0, 109, 0, 95, 0, 80, 0, 101, 0, 110, 0, 104, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 67, 0, 97, 0, 110, 0, 97, 0, 114, 0, 121, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 70, 0, 97, 0, 101, 0, 114, 0, 111, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 69, 0, 117, 0, 99, 0, 108, 0, 97, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 80, 0, 101, 0, 114, 0, 116, 0, 104, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 114, 0, 117, 0, 115, 0, 115, 0, 101, 0, 108, 0, 115, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 117, 0, 100, 0, 97, 0, 112, 0, 101, 0, 115, 0, 116, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 72, 0, 101, 0, 108, 0, 115, 0, 105, 0, 110, 0, 107, 0, 105, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 73, 0, 115, 0, 116, 0, 97, 0, 110, 0, 98, 0, 117, 0, 108, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 114, 0, 97, 0, 106, 0, 101, 0, 118, 0, 111, 0, 0, 0, 70, 0, 114, 0, 101, 0, 110, 0, 99, 0, 104, 0, 95, 0, 83, 0, 111, 0, 117, 0, 116, 0, 104, 0, 101, 0, 114, 0, 110, 0, 0, 0, 71, 0, 105, 0, 108, 0, 98, 0, 101, 0, 114, 0, 116, 0, 95, 0, 73, 0, 115, 0, 108, 0, 97, 0, 110, 0, 100, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 70, 0, 97, 0, 107, 0, 97, 0, 111, 0, 102, 0, 111, 0, 0, 0, 80, 0, 104, 0, 111, 0, 101, 0, 110, 0, 105, 0, 120, 0, 95, 0, 73, 0, 115, 0, 108, 0, 97, 0, 110, 0, 100, 0, 115, 0, 0, 0, 80, 0, 105, 0, 101, 0, 114, 0, 114, 0, 101, 0, 95, 0, 77, 0, 105, 0, 113, 0, 117, 0, 101, 0, 108, 0, 111, 0, 110, 0, 0, 0, 49, 0, 57, 0, 55, 0, 48, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 57, 0, 57, 0, 57, 0, 57, 0, 45, 0, 49, 0, 50, 0, 45, 0, 51, 0, 49, 0, 32, 0, 50, 0, 51, 0, 58, 0, 53, 0, 57, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 95, 0, 77, 0, 111, 0, 117, 0, 110, 0, 116, 0, 97, 0, 105, 0, 110, 0, 0, 0, 50, 0, 48, 0, 48, 0, 48, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 57, 0, 32, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 49, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 49, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 51, 0, 45, 0, 49, 0, 49, 0, 45, 0, 51, 0, 48, 0, 32, 0, 48, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 54, 0, 45, 0, 48, 0, 52, 0, 45, 0, 48, 0, 50, 0, 32, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 57, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 53, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 51, 0, 45, 0, 51, 0, 48, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 48, 0, 54, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 51, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 48, 0, 32, 0, 49, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 48, 0, 54, 0, 45, 0, 50, 0, 48, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 48, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 51, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 48, 0, 32, 0, 49, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 49, 0, 50, 0, 45, 0, 49, 0, 53, 0, 32, 0, 49, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 48, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 50, 0, 45, 0, 48, 0, 53, 0, 45, 0, 51, 0, 49, 0, 32, 0, 50, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 52, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 54, 0, 32, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 52, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 55, 0, 32, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 49, 0, 45, 0, 49, 0, 50, 0, 45, 0, 50, 0, 51, 0, 32, 0, 48, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 51, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 48, 0, 32, 0, 48, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 52, 0, 45, 0, 48, 0, 51, 0, 45, 0, 49, 0, 54, 0, 32, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 48, 0, 45, 0, 48, 0, 54, 0, 45, 0, 51, 0, 48, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 51, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 51, 0, 45, 0, 48, 0, 51, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 53, 0, 45, 0, 48, 0, 55, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 56, 0, 45, 0, 51, 0, 49, 0, 32, 0, 49, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 48, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 50, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 56, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 53, 0, 45, 0, 49, 0, 50, 0, 45, 0, 51, 0, 49, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 56, 0, 45, 0, 48, 0, 52, 0, 45, 0, 48, 0, 53, 0, 32, 0, 48, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 57, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 49, 0, 32, 0, 48, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 48, 0, 45, 0, 48, 0, 49, 0, 45, 0, 49, 0, 53, 0, 32, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 48, 0, 45, 0, 49, 0, 50, 0, 45, 0, 50, 0, 50, 0, 32, 0, 49, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 49, 0, 45, 0, 48, 0, 52, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 48, 0, 53, 0, 45, 0, 51, 0, 49, 0, 32, 0, 48, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 48, 0, 55, 0, 45, 0, 50, 0, 53, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 48, 0, 32, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 55, 0, 45, 0, 48, 0, 51, 0, 45, 0, 49, 0, 49, 0, 32, 0, 48, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 55, 0, 45, 0, 49, 0, 49, 0, 45, 0, 48, 0, 52, 0, 32, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 56, 0, 45, 0, 48, 0, 54, 0, 45, 0, 50, 0, 52, 0, 32, 0, 48, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 49, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 55, 0, 32, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 51, 0, 45, 0, 49, 0, 49, 0, 45, 0, 49, 0, 48, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 114, 0, 97, 0, 115, 0, 110, 0, 111, 0, 121, 0, 97, 0, 114, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 86, 0, 108, 0, 97, 0, 100, 0, 105, 0, 118, 0, 111, 0, 115, 0, 116, 0, 111, 0, 107, 0, 0, 0, 49, 0, 57, 0, 57, 0, 50, 0, 45, 0, 48, 0, 49, 0, 45, 0, 49, 0, 57, 0, 32, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 52, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 53, 0, 32, 0, 49, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 54, 0, 45, 0, 49, 0, 50, 0, 45, 0, 48, 0, 51, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 78, 0, 111, 0, 118, 0, 111, 0, 115, 0, 105, 0, 98, 0, 105, 0, 114, 0, 115, 0, 107, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 111, 0, 108, 0, 103, 0, 111, 0, 103, 0, 114, 0, 97, 0, 100, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 115, 0, 101, 0, 121, 0, 0, 0, 49, 0, 57, 0, 55, 0, 49, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 53, 0, 32, 0, 49, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 49, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 53, 0, 32, 0, 49, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 50, 0, 45, 0, 48, 0, 52, 0, 45, 0, 51, 0, 48, 0, 32, 0, 48, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 50, 0, 45, 0, 48, 0, 53, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 48, 0, 58, 0, 52, 0, 53, 0, 0, 0, 49, 0, 57, 0, 55, 0, 50, 0, 45, 0, 48, 0, 56, 0, 45, 0, 51, 0, 48, 0, 32, 0, 48, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 51, 0, 45, 0, 48, 0, 52, 0, 45, 0, 50, 0, 57, 0, 32, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 51, 0, 45, 0, 48, 0, 53, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 51, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 56, 0, 32, 0, 48, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 52, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 55, 0, 32, 0, 48, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 53, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 53, 0, 45, 0, 48, 0, 50, 0, 45, 0, 49, 0, 54, 0, 32, 0, 48, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 53, 0, 45, 0, 49, 0, 49, 0, 45, 0, 50, 0, 48, 0, 32, 0, 48, 0, 51, 0, 58, 0, 51, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 54, 0, 45, 0, 48, 0, 52, 0, 45, 0, 49, 0, 52, 0, 32, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 54, 0, 45, 0, 48, 0, 53, 0, 45, 0, 48, 0, 50, 0, 32, 0, 49, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 54, 0, 45, 0, 48, 0, 57, 0, 45, 0, 50, 0, 54, 0, 32, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 55, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 48, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 55, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 48, 0, 32, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 56, 0, 45, 0, 48, 0, 54, 0, 45, 0, 50, 0, 56, 0, 32, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 57, 0, 45, 0, 48, 0, 52, 0, 45, 0, 50, 0, 57, 0, 32, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 55, 0, 57, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 53, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 48, 0, 45, 0, 48, 0, 52, 0, 45, 0, 50, 0, 55, 0, 32, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 48, 0, 45, 0, 48, 0, 53, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 48, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 54, 0, 32, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 49, 0, 45, 0, 48, 0, 50, 0, 45, 0, 50, 0, 56, 0, 32, 0, 49, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 49, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 57, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 49, 0, 45, 0, 48, 0, 51, 0, 45, 0, 51, 0, 49, 0, 32, 0, 49, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 49, 0, 45, 0, 48, 0, 53, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 49, 0, 45, 0, 48, 0, 57, 0, 45, 0, 51, 0, 48, 0, 32, 0, 49, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 49, 0, 45, 0, 49, 0, 50, 0, 45, 0, 51, 0, 49, 0, 32, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 49, 0, 45, 0, 49, 0, 50, 0, 45, 0, 51, 0, 49, 0, 32, 0, 49, 0, 54, 0, 58, 0, 51, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 49, 0, 45, 0, 49, 0, 50, 0, 45, 0, 51, 0, 49, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 50, 0, 45, 0, 48, 0, 51, 0, 45, 0, 51, 0, 49, 0, 32, 0, 49, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 50, 0, 45, 0, 48, 0, 53, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 50, 0, 45, 0, 49, 0, 50, 0, 45, 0, 48, 0, 50, 0, 32, 0, 48, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 51, 0, 45, 0, 48, 0, 51, 0, 45, 0, 51, 0, 49, 0, 32, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 53, 0, 45, 0, 48, 0, 52, 0, 45, 0, 49, 0, 57, 0, 32, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 53, 0, 45, 0, 49, 0, 50, 0, 45, 0, 51, 0, 49, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 54, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 55, 0, 45, 0, 48, 0, 57, 0, 45, 0, 51, 0, 48, 0, 32, 0, 49, 0, 56, 0, 58, 0, 51, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 55, 0, 45, 0, 49, 0, 50, 0, 45, 0, 51, 0, 49, 0, 32, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 56, 0, 45, 0, 48, 0, 52, 0, 45, 0, 48, 0, 51, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 49, 0, 0, 0, 49, 0, 57, 0, 56, 0, 56, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 48, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 49, 0, 0, 0, 49, 0, 57, 0, 56, 0, 57, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 53, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 48, 0, 45, 0, 48, 0, 51, 0, 45, 0, 48, 0, 52, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 48, 0, 45, 0, 48, 0, 51, 0, 45, 0, 49, 0, 52, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 48, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 48, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 48, 0, 45, 0, 48, 0, 53, 0, 45, 0, 48, 0, 51, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 48, 0, 45, 0, 48, 0, 53, 0, 45, 0, 48, 0, 53, 0, 32, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 48, 0, 45, 0, 48, 0, 54, 0, 45, 0, 51, 0, 48, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 51, 0, 45, 0, 51, 0, 49, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 52, 0, 45, 0, 48, 0, 56, 0, 32, 0, 50, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 54, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 56, 0, 45, 0, 50, 0, 57, 0, 32, 0, 50, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 56, 0, 45, 0, 51, 0, 48, 0, 32, 0, 50, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 57, 0, 45, 0, 48, 0, 56, 0, 32, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 57, 0, 45, 0, 50, 0, 50, 0, 32, 0, 50, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 48, 0, 57, 0, 45, 0, 50, 0, 57, 0, 32, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 54, 0, 32, 0, 50, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 55, 0, 32, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 49, 0, 50, 0, 45, 0, 49, 0, 53, 0, 32, 0, 50, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 50, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 50, 0, 45, 0, 48, 0, 49, 0, 45, 0, 49, 0, 56, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 50, 0, 45, 0, 48, 0, 57, 0, 45, 0, 50, 0, 52, 0, 32, 0, 48, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 50, 0, 45, 0, 48, 0, 57, 0, 45, 0, 50, 0, 55, 0, 32, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 50, 0, 45, 0, 48, 0, 57, 0, 45, 0, 50, 0, 55, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 50, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 53, 0, 32, 0, 48, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 51, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 51, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 56, 0, 32, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 51, 0, 45, 0, 48, 0, 56, 0, 45, 0, 50, 0, 49, 0, 32, 0, 49, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 52, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 48, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 52, 0, 45, 0, 48, 0, 52, 0, 45, 0, 51, 0, 48, 0, 32, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 53, 0, 45, 0, 48, 0, 52, 0, 45, 0, 48, 0, 50, 0, 32, 0, 48, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 54, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 54, 0, 45, 0, 48, 0, 51, 0, 45, 0, 51, 0, 49, 0, 32, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 54, 0, 45, 0, 48, 0, 53, 0, 45, 0, 50, 0, 52, 0, 32, 0, 49, 0, 56, 0, 58, 0, 51, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 54, 0, 45, 0, 48, 0, 57, 0, 45, 0, 50, 0, 57, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 55, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 55, 0, 45, 0, 48, 0, 51, 0, 45, 0, 51, 0, 48, 0, 32, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 55, 0, 45, 0, 49, 0, 48, 0, 45, 0, 48, 0, 51, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 56, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 57, 0, 32, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 56, 0, 45, 0, 48, 0, 56, 0, 45, 0, 48, 0, 50, 0, 32, 0, 48, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 57, 0, 45, 0, 49, 0, 48, 0, 45, 0, 48, 0, 51, 0, 32, 0, 48, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 57, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 49, 0, 32, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 57, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 49, 0, 32, 0, 48, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 57, 0, 45, 0, 49, 0, 50, 0, 45, 0, 49, 0, 57, 0, 32, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 48, 0, 45, 0, 48, 0, 51, 0, 45, 0, 48, 0, 51, 0, 32, 0, 48, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 48, 0, 45, 0, 48, 0, 57, 0, 45, 0, 49, 0, 54, 0, 32, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 48, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 57, 0, 32, 0, 48, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 48, 0, 45, 0, 49, 0, 49, 0, 45, 0, 48, 0, 53, 0, 32, 0, 48, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 48, 0, 45, 0, 49, 0, 50, 0, 45, 0, 48, 0, 51, 0, 32, 0, 48, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 49, 0, 45, 0, 48, 0, 52, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 51, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 54, 0, 32, 0, 48, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 51, 0, 45, 0, 49, 0, 50, 0, 45, 0, 51, 0, 49, 0, 32, 0, 49, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 48, 0, 54, 0, 45, 0, 49, 0, 51, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 48, 0, 32, 0, 50, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 54, 0, 45, 0, 48, 0, 52, 0, 45, 0, 49, 0, 52, 0, 32, 0, 49, 0, 56, 0, 58, 0, 51, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 54, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 57, 0, 32, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 56, 0, 45, 0, 48, 0, 49, 0, 45, 0, 50, 0, 49, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 56, 0, 45, 0, 48, 0, 54, 0, 45, 0, 50, 0, 52, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 57, 0, 45, 0, 49, 0, 48, 0, 45, 0, 49, 0, 55, 0, 32, 0, 49, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 48, 0, 45, 0, 48, 0, 51, 0, 45, 0, 48, 0, 52, 0, 32, 0, 49, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 48, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 55, 0, 32, 0, 49, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 48, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 55, 0, 32, 0, 49, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 48, 0, 45, 0, 48, 0, 52, 0, 45, 0, 48, 0, 51, 0, 32, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 48, 0, 45, 0, 48, 0, 52, 0, 45, 0, 48, 0, 52, 0, 32, 0, 48, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 48, 0, 45, 0, 49, 0, 49, 0, 45, 0, 48, 0, 55, 0, 32, 0, 48, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 49, 0, 45, 0, 48, 0, 57, 0, 45, 0, 49, 0, 50, 0, 32, 0, 49, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 49, 0, 45, 0, 48, 0, 57, 0, 45, 0, 49, 0, 50, 0, 32, 0, 49, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 49, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 55, 0, 32, 0, 49, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 50, 0, 45, 0, 48, 0, 50, 0, 45, 0, 50, 0, 49, 0, 32, 0, 49, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 50, 0, 45, 0, 49, 0, 49, 0, 45, 0, 49, 0, 48, 0, 32, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 51, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 53, 0, 32, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 52, 0, 45, 0, 48, 0, 51, 0, 45, 0, 51, 0, 48, 0, 32, 0, 48, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 52, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 53, 0, 32, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 52, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 53, 0, 32, 0, 49, 0, 57, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 52, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 53, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 52, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 54, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 53, 0, 45, 0, 48, 0, 50, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 53, 0, 45, 0, 48, 0, 51, 0, 45, 0, 48, 0, 56, 0, 32, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 53, 0, 45, 0, 48, 0, 56, 0, 45, 0, 49, 0, 52, 0, 32, 0, 49, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 53, 0, 45, 0, 49, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 53, 0, 45, 0, 49, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 54, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 54, 0, 32, 0, 49, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 54, 0, 45, 0, 48, 0, 51, 0, 45, 0, 50, 0, 54, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 54, 0, 45, 0, 48, 0, 57, 0, 45, 0, 48, 0, 54, 0, 32, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 54, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 49, 0, 32, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 55, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 51, 0, 32, 0, 50, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 55, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 49, 0, 32, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 56, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 56, 0, 45, 0, 48, 0, 51, 0, 45, 0, 49, 0, 48, 0, 32, 0, 49, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 56, 0, 45, 0, 48, 0, 51, 0, 45, 0, 49, 0, 49, 0, 32, 0, 48, 0, 55, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 56, 0, 45, 0, 48, 0, 53, 0, 45, 0, 48, 0, 52, 0, 32, 0, 49, 0, 53, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 56, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 55, 0, 32, 0, 50, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 56, 0, 45, 0, 49, 0, 48, 0, 45, 0, 50, 0, 56, 0, 32, 0, 48, 0, 50, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 56, 0, 45, 0, 49, 0, 49, 0, 45, 0, 48, 0, 52, 0, 32, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 56, 0, 45, 0, 49, 0, 50, 0, 45, 0, 50, 0, 48, 0, 32, 0, 49, 0, 56, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 57, 0, 45, 0, 48, 0, 49, 0, 45, 0, 48, 0, 49, 0, 32, 0, 48, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 57, 0, 45, 0, 48, 0, 49, 0, 45, 0, 50, 0, 48, 0, 32, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 105, 0, 97, 0, 103, 0, 111, 0, 0, 0, 77, 0, 97, 0, 114, 0, 115, 0, 104, 0, 97, 0, 108, 0, 108, 0, 95, 0, 73, 0, 115, 0, 108, 0, 97, 0, 110, 0, 100, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 65, 0, 117, 0, 99, 0, 107, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 80, 0, 97, 0, 112, 0, 117, 0, 97, 0, 95, 0, 78, 0, 101, 0, 119, 0, 95, 0, 71, 0, 117, 0, 105, 0, 110, 0, 101, 0, 97, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 104, 0, 114, 0, 105, 0, 115, 0, 116, 0, 109, 0, 97, 0, 115, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 117, 0, 114, 0, 105, 0, 116, 0, 105, 0, 117, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 105, 0, 116, 0, 99, 0, 97, 0, 105, 0, 114, 0, 110, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 118, 0, 105, 0, 115, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 121, 0, 111, 0, 119, 0, 97, 0, 0, 0, 49, 0, 57, 0, 55, 0, 53, 0, 45, 0, 48, 0, 54, 0, 45, 0, 49, 0, 50, 0, 32, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 56, 0, 51, 0, 45, 0, 49, 0, 49, 0, 45, 0, 51, 0, 48, 0, 32, 0, 49, 0, 48, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 49, 0, 45, 0, 49, 0, 48, 0, 45, 0, 48, 0, 54, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 49, 0, 57, 0, 57, 0, 50, 0, 45, 0, 49, 0, 48, 0, 45, 0, 49, 0, 56, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 48, 0, 53, 0, 45, 0, 50, 0, 51, 0, 32, 0, 48, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 48, 0, 53, 0, 45, 0, 51, 0, 48, 0, 32, 0, 48, 0, 51, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 52, 0, 45, 0, 48, 0, 57, 0, 45, 0, 50, 0, 54, 0, 32, 0, 48, 0, 52, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 48, 0, 53, 0, 45, 0, 49, 0, 48, 0, 45, 0, 51, 0, 48, 0, 32, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 52, 0, 45, 0, 49, 0, 50, 0, 45, 0, 50, 0, 55, 0, 32, 0, 49, 0, 54, 0, 58, 0, 48, 0, 48, 0, 0, 0, 50, 0, 48, 0, 49, 0, 54, 0, 45, 0, 48, 0, 57, 0, 45, 0, 48, 0, 55, 0, 32, 0, 50, 0, 49, 0, 58, 0, 48, 0, 48, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 117, 0, 106, 0, 117, 0, 109, 0, 98, 0, 117, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 103, 0, 97, 0, 100, 0, 105, 0, 115, 0, 104, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 103, 0, 117, 0, 105, 0, 108, 0, 108, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 115, 0, 117, 0, 110, 0, 99, 0, 105, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 114, 0, 98, 0, 97, 0, 100, 0, 111, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 111, 0, 109, 0, 105, 0, 110, 0, 105, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 100, 0, 109, 0, 111, 0, 110, 0, 116, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 122, 0, 97, 0, 116, 0, 108, 0, 97, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 105, 0, 113, 0, 117, 0, 101, 0, 108, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 101, 0, 119, 0, 95, 0, 89, 0, 111, 0, 114, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 74, 0, 111, 0, 104, 0, 110, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 75, 0, 105, 0, 116, 0, 116, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 76, 0, 117, 0, 99, 0, 105, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 87, 0, 105, 0, 110, 0, 110, 0, 105, 0, 112, 0, 101, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 108, 0, 97, 0, 97, 0, 110, 0, 98, 0, 97, 0, 97, 0, 116, 0, 97, 0, 114, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 66, 0, 101, 0, 114, 0, 109, 0, 117, 0, 100, 0, 97, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 83, 0, 116, 0, 97, 0, 110, 0, 108, 0, 101, 0, 121, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 83, 0, 121, 0, 100, 0, 110, 0, 101, 0, 121, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 109, 0, 115, 0, 116, 0, 101, 0, 114, 0, 100, 0, 97, 0, 109, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 117, 0, 99, 0, 104, 0, 97, 0, 114, 0, 101, 0, 115, 0, 116, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 71, 0, 105, 0, 98, 0, 114, 0, 97, 0, 108, 0, 116, 0, 97, 0, 114, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 106, 0, 117, 0, 98, 0, 108, 0, 106, 0, 97, 0, 110, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 97, 0, 114, 0, 105, 0, 101, 0, 104, 0, 97, 0, 109, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 80, 0, 111, 0, 100, 0, 103, 0, 111, 0, 114, 0, 105, 0, 99, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 116, 0, 111, 0, 99, 0, 107, 0, 104, 0, 111, 0, 108, 0, 109, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 75, 0, 101, 0, 114, 0, 103, 0, 117, 0, 101, 0, 108, 0, 101, 0, 110, 0, 0, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 95, 0, 78, 0, 111, 0, 114, 0, 116, 0, 104, 0, 119, 0, 101, 0, 115, 0, 116, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 70, 0, 117, 0, 110, 0, 97, 0, 102, 0, 117, 0, 116, 0, 105, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 72, 0, 111, 0, 110, 0, 111, 0, 108, 0, 117, 0, 108, 0, 117, 0, 0, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 95, 0, 87, 0, 101, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 95, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 95, 0, 87, 0, 101, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 95, 0, 67, 0, 101, 0, 110, 0, 116, 0, 114, 0, 97, 0, 108, 0, 0, 0, 73, 0, 110, 0, 100, 0, 111, 0, 110, 0, 101, 0, 115, 0, 105, 0, 97, 0, 95, 0, 67, 0, 101, 0, 110, 0, 116, 0, 114, 0, 97, 0, 108, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 95, 0, 70, 0, 97, 0, 114, 0, 87, 0, 101, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 97, 0, 121, 0, 97, 0, 113, 0, 117, 0, 105, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 117, 0, 97, 0, 108, 0, 97, 0, 95, 0, 76, 0, 117, 0, 109, 0, 112, 0, 117, 0, 114, 0, 0, 0, 71, 0, 114, 0, 101, 0, 101, 0, 110, 0, 108, 0, 97, 0, 110, 0, 100, 0, 95, 0, 87, 0, 101, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 73, 0, 110, 0, 100, 0, 111, 0, 110, 0, 101, 0, 115, 0, 105, 0, 97, 0, 95, 0, 87, 0, 101, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 111, 0, 111, 0, 115, 0, 101, 0, 95, 0, 66, 0, 97, 0, 121, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 97, 0, 108, 0, 97, 0, 112, 0, 97, 0, 103, 0, 111, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 75, 0, 119, 0, 97, 0, 106, 0, 97, 0, 108, 0, 101, 0, 105, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 77, 0, 97, 0, 114, 0, 113, 0, 117, 0, 101, 0, 115, 0, 97, 0, 115, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 119, 0, 115, 0, 111, 0, 110, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 86, 0, 111, 0, 115, 0, 116, 0, 111, 0, 107, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 105, 0, 98, 0, 114, 0, 101, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 117, 0, 98, 0, 117, 0, 109, 0, 98, 0, 97, 0, 115, 0, 104, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 117, 0, 97, 0, 107, 0, 99, 0, 104, 0, 111, 0, 116, 0, 116, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 111, 0, 45, 0, 78, 0, 111, 0, 118, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 99, 0, 104, 0, 111, 0, 114, 0, 97, 0, 103, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 97, 0, 116, 0, 101, 0, 109, 0, 97, 0, 108, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 111, 0, 95, 0, 80, 0, 97, 0, 117, 0, 108, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 84, 0, 104, 0, 111, 0, 109, 0, 97, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 86, 0, 97, 0, 110, 0, 99, 0, 111, 0, 117, 0, 118, 0, 101, 0, 114, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 114, 0, 97, 0, 116, 0, 105, 0, 115, 0, 108, 0, 97, 0, 118, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 67, 0, 111, 0, 112, 0, 101, 0, 110, 0, 104, 0, 97, 0, 103, 0, 101, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 117, 0, 120, 0, 101, 0, 109, 0, 98, 0, 111, 0, 117, 0, 114, 0, 103, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 110, 0, 95, 0, 77, 0, 97, 0, 114, 0, 105, 0, 110, 0, 111, 0, 0, 0, 71, 0, 114, 0, 101, 0, 101, 0, 110, 0, 108, 0, 97, 0, 110, 0, 100, 0, 95, 0, 67, 0, 101, 0, 110, 0, 116, 0, 114, 0, 97, 0, 108, 0, 0, 0, 71, 0, 114, 0, 101, 0, 101, 0, 110, 0, 108, 0, 97, 0, 110, 0, 100, 0, 95, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 73, 0, 110, 0, 100, 0, 111, 0, 110, 0, 101, 0, 115, 0, 105, 0, 97, 0, 95, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 69, 0, 110, 0, 100, 0, 101, 0, 114, 0, 98, 0, 117, 0, 114, 0, 121, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 97, 0, 103, 0, 111, 0, 95, 0, 80, 0, 97, 0, 103, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 82, 0, 97, 0, 114, 0, 111, 0, 116, 0, 111, 0, 110, 0, 103, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 111, 0, 110, 0, 103, 0, 97, 0, 116, 0, 97, 0, 112, 0, 117, 0, 0, 0, 75, 0, 97, 0, 122, 0, 97, 0, 107, 0, 104, 0, 115, 0, 116, 0, 97, 0, 110, 0, 95, 0, 87, 0, 101, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 75, 0, 97, 0, 122, 0, 97, 0, 107, 0, 104, 0, 115, 0, 116, 0, 97, 0, 110, 0, 95, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 101, 0, 107, 0, 97, 0, 116, 0, 101, 0, 114, 0, 105, 0, 110, 0, 98, 0, 117, 0, 114, 0, 103, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 97, 0, 114, 0, 97, 0, 109, 0, 97, 0, 114, 0, 105, 0, 98, 0, 111, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 111, 0, 116, 0, 104, 0, 101, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 100, 0, 100, 0, 105, 0, 115, 0, 95, 0, 65, 0, 98, 0, 97, 0, 98, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 114, 0, 97, 0, 122, 0, 122, 0, 97, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 79, 0, 117, 0, 97, 0, 103, 0, 97, 0, 100, 0, 111, 0, 117, 0, 103, 0, 111, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 115, 0, 116, 0, 97, 0, 95, 0, 82, 0, 105, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 97, 0, 100, 0, 101, 0, 108, 0, 111, 0, 117, 0, 112, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 101, 0, 114, 0, 109, 0, 111, 0, 115, 0, 105, 0, 108, 0, 108, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 114, 0, 97, 0, 108, 0, 101, 0, 110, 0, 100, 0, 105, 0, 106, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 114, 0, 116, 0, 105, 0, 110, 0, 105, 0, 113, 0, 117, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 101, 0, 118, 0, 105, 0, 100, 0, 101, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 115, 0, 101, 0, 114, 0, 114, 0, 97, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 105, 0, 111, 0, 95, 0, 66, 0, 114, 0, 97, 0, 110, 0, 99, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 86, 0, 105, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 99, 0, 77, 0, 117, 0, 114, 0, 100, 0, 111, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 82, 0, 101, 0, 121, 0, 107, 0, 106, 0, 97, 0, 118, 0, 105, 0, 107, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 83, 0, 116, 0, 95, 0, 72, 0, 101, 0, 108, 0, 101, 0, 110, 0, 97, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 65, 0, 100, 0, 101, 0, 108, 0, 97, 0, 105, 0, 100, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 75, 0, 97, 0, 108, 0, 105, 0, 110, 0, 105, 0, 110, 0, 103, 0, 114, 0, 97, 0, 100, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 75, 0, 105, 0, 114, 0, 105, 0, 116, 0, 105, 0, 109, 0, 97, 0, 116, 0, 105, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 67, 0, 97, 0, 112, 0, 101, 0, 95, 0, 86, 0, 101, 0, 114, 0, 100, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 76, 0, 111, 0, 114, 0, 100, 0, 95, 0, 72, 0, 111, 0, 119, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 111, 0, 104, 0, 97, 0, 110, 0, 110, 0, 101, 0, 115, 0, 98, 0, 117, 0, 114, 0, 103, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 108, 0, 95, 0, 83, 0, 97, 0, 108, 0, 118, 0, 97, 0, 100, 0, 111, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 115, 0, 95, 0, 65, 0, 110, 0, 103, 0, 101, 0, 108, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 95, 0, 67, 0, 105, 0, 116, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 117, 0, 101, 0, 114, 0, 116, 0, 111, 0, 95, 0, 82, 0, 105, 0, 99, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 101, 0, 103, 0, 117, 0, 99, 0, 105, 0, 103, 0, 97, 0, 108, 0, 112, 0, 97, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 65, 0, 110, 0, 116, 0, 97, 0, 110, 0, 97, 0, 110, 0, 97, 0, 114, 0, 105, 0, 118, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 117, 0, 97, 0, 100, 0, 97, 0, 108, 0, 99, 0, 97, 0, 110, 0, 97, 0, 108, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 99, 0, 111, 0, 114, 0, 101, 0, 115, 0, 98, 0, 121, 0, 115, 0, 117, 0, 110, 0, 100, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 99, 0, 113, 0, 117, 0, 97, 0, 114, 0, 105, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 114, 0, 95, 0, 101, 0, 115, 0, 95, 0, 83, 0, 97, 0, 108, 0, 97, 0, 97, 0, 109, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 117, 0, 101, 0, 110, 0, 111, 0, 115, 0, 95, 0, 65, 0, 105, 0, 114, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 97, 0, 95, 0, 73, 0, 115, 0, 97, 0, 98, 0, 101, 0, 108, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 95, 0, 77, 0, 111, 0, 114, 0, 101, 0, 115, 0, 98, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 119, 0, 101, 0, 114, 0, 95, 0, 80, 0, 114, 0, 105, 0, 110, 0, 99, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 95, 0, 111, 0, 102, 0, 95, 0, 83, 0, 112, 0, 97, 0, 105, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 111, 0, 95, 0, 68, 0, 111, 0, 109, 0, 105, 0, 110, 0, 103, 0, 111, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 95, 0, 70, 0, 117, 0, 114, 0, 116, 0, 104, 0, 101, 0, 114, 0, 95, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 83, 0, 111, 0, 117, 0, 116, 0, 104, 0, 95, 0, 71, 0, 101, 0, 111, 0, 114, 0, 103, 0, 105, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 45, 0, 97, 0, 117, 0, 45, 0, 80, 0, 114, 0, 105, 0, 110, 0, 99, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 95, 0, 67, 0, 101, 0, 110, 0, 116, 0, 114, 0, 97, 0, 108, 0, 87, 0, 101, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 117, 0, 109, 0, 111, 0, 110, 0, 116, 0, 68, 0, 85, 0, 114, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 95, 0, 76, 0, 117, 0, 105, 0, 115, 0, 0, 0, 1, 0, 50, 0, 134, 33, 1, 0, 50, 0, 78, 2, 8, 0, 50, 0, 81, 0, 84, 0, 87, 0, 90, 0, 93, 0, 96, 0, 99, 0, 4, 7, 244, 27, 20, 12, 19, 31, 212, 11, 204, 6, 232, 6, 190, 6, 10, 0, 50, 0, 117, 0, 120, 0, 123, 0, 126, 0, 129, 0, 132, 0, 135, 0, 138, 0, 141, 0, 246, 9, 228, 11, 120, 6, 200, 32, 56, 8, 190, 34, 5, 28, 16, 35, 216, 9, 215, 10, 1, 0, 50, 0, 244, 11, 3, 0, 50, 0, 178, 0, 181, 0, 70, 34, 18, 7, 231, 9, 11, 0, 50, 0, 199, 0, 202, 0, 87, 0, 205, 0, 208, 0, 211, 0, 214, 0, 217, 0, 220, 0, 223, 0, 163, 4, 218, 6, 55, 31, 36, 12, 148, 6, 219, 32, 176, 6, 1, 31, 246, 6, 32, 7, 68, 12, 1, 0, 50, 0, 188, 2, 1, 0, 50, 0, 80, 10, 1, 0, 50, 0, 73, 31, 1, 0, 50, 0, 200, 2, 1, 0, 50, 0, 110, 10, 8, 0, 50, 0, 25, 1, 28, 1, 31, 1, 34, 1, 37, 1, 40, 1, 43, 1, 132, 12, 5, 10, 209, 28, 1, 33, 91, 31, 170, 34, 130, 34, 90, 34, 7, 0, 50, 0, 62, 1, 28, 1, 65, 1, 68, 1, 71, 1, 74, 1, 141, 28, 125, 10, 4, 13, 212, 35, 212, 12, 35, 10, 140, 10, 3, 0, 50, 0, 28, 1, 40, 1, 50, 10, 90, 28, 39, 33, 3, 0, 50, 0, 28, 1, 40, 1, 110, 34, 145, 31, 244, 12, 1, 0, 50, 0, 212, 2, 1, 0, 50, 0, 59, 4, 1, 0, 50, 0, 246, 1, 1, 0, 50, 0, 188, 2, 6, 0, 50, 0, 118, 1, 121, 1, 124, 1, 127, 1, 130, 1, 88, 3, 228, 4, 215, 4, 52, 3, 89, 2, 176, 1, 1, 0, 50, 0, 37, 35, 1, 0, 50, 0, 30, 36, 1, 0, 50, 0, 224, 3, 1, 0, 50, 0, 250, 5, 23, 0, 50, 0, 188, 1, 191, 1, 194, 1, 197, 1, 200, 1, 203, 1, 206, 1, 209, 1, 212, 1, 215, 1, 218, 1, 221, 1, 224, 1, 227, 1, 230, 1, 233, 1, 236, 1, 239, 1, 242, 1, 245, 1, 248, 1, 251, 1, 196, 12, 84, 12, 22, 28, 46, 7, 56, 28, 243, 28, 58, 33, 148, 12, 73, 28, 180, 12, 60, 7, 20, 33, 175, 28, 192, 28, 228, 12, 77, 33, 115, 33, 150, 34, 100, 35, 122, 35, 153, 33, 20, 13, 127, 31, 1, 0, 50, 0, 229, 33, 1, 0, 50, 0, 100, 13, 1, 0, 50, 0, 21, 29, 1, 0, 50, 0, 116, 13, 1, 0, 50, 0, 126, 1, 1, 0, 50, 0, 100, 11, 1, 0, 50, 0, 126, 1, 1, 0, 50, 0, 1, 2, 1, 0, 50, 0, 189, 4, 1, 0, 50, 0, 37, 5, 1, 0, 50, 0, 95, 10, 1, 0, 50, 0, 33, 4, 1, 0, 50, 0, 109, 31, 1, 0, 50, 0, 50, 6, 1, 0, 50, 0, 236, 2, 1, 0, 50, 0, 30, 34, 1, 0, 50, 0, 118, 17, 2, 0, 50, 0, 154, 2, 6, 9, 237, 3, 1, 0, 50, 0, 164, 11, 1, 0, 50, 0, 177, 26, 1, 0, 50, 0, 8, 6, 1, 0, 50, 0, 116, 11, 1, 0, 50, 0, 245, 26, 1, 0, 50, 0, 46, 4, 1, 0, 50, 0, 20, 10, 1, 0, 50, 0, 69, 32, 1, 0, 50, 0, 65, 10, 1, 0, 50, 0, 1, 2, 1, 0, 50, 0, 40, 27, 1, 0, 50, 0, 144, 35, 1, 0, 50, 0, 4, 36, 1, 0, 50, 0, 222, 5, 1, 0, 50, 0, 162, 32, 1, 0, 50, 0, 186, 1, 1, 0, 50, 0, 111, 9, 1, 0, 50, 0, 77, 30, 32, 0, 50, 0, 251, 2, 254, 2, 1, 3, 4, 3, 7, 3, 10, 3, 13, 3, 16, 3, 19, 3, 22, 3, 25, 3, 28, 3, 31, 3, 34, 3, 37, 3, 40, 3, 43, 3, 46, 3, 49, 3, 95, 27, 52, 3, 55, 3, 58, 3, 61, 3, 64, 3, 67, 3, 70, 3, 73, 3, 76, 3, 79, 3, 82, 3, 89, 5, 185, 10, 214, 7, 228, 7, 196, 13, 132, 13, 14, 8, 186, 7, 22, 6, 181, 31, 64, 6, 72, 29, 0, 8, 148, 13, 136, 3, 115, 5, 199, 31, 172, 7, 123, 29, 200, 7, 63, 5, 38, 29, 124, 3, 242, 7, 84, 11, 140, 29, 89, 29, 163, 31, 217, 31, 176, 4, 200, 10, 84, 11, 10, 0, 50, 0, 100, 3, 103, 3, 106, 3, 109, 3, 112, 3, 115, 3, 118, 3, 121, 3, 124, 3, 55, 29, 106, 29, 102, 5, 11, 5, 137, 4, 164, 13, 144, 7, 56, 2, 40, 3, 88, 7, 2, 0, 50, 0, 150, 3, 76, 5, 248, 33, 2, 0, 50, 0, 168, 3, 68, 13, 84, 13, 1, 0, 50, 0, 4, 29, 1, 0, 50, 0, 72, 4, 1, 0, 50, 0, 116, 12, 1, 0, 50, 0, 157, 29, 1, 0, 50, 0, 7, 4, 14, 0, 50, 0, 217, 3, 220, 3, 223, 3, 226, 3, 229, 3, 232, 3, 235, 3, 238, 3, 241, 3, 244, 3, 247, 3, 250, 3, 253, 3, 191, 33, 238, 32, 186, 9, 50, 6, 124, 4, 162, 6, 201, 9, 36, 6, 134, 6, 37, 31, 210, 33, 4, 12, 150, 4, 28, 3, 1, 0, 50, 0, 167, 30, 1, 0, 50, 0, 180, 11, 1, 0, 50, 0, 211, 3, 1, 0, 50, 0, 245, 10, 1, 0, 50, 0, 149, 30, 1, 0, 50, 0, 230, 34, 1, 0, 50, 0, 230, 34, 1, 0, 50, 0, 164, 12, 1, 0, 50, 0, 237, 3, 2, 0, 50, 0, 75, 4, 67, 2, 76, 3, 1, 0, 50, 0, 96, 9, 1, 0, 50, 0, 208, 29, 1, 0, 50, 0, 21, 9, 1, 0, 50, 0, 146, 1, 2, 0, 50, 0, 100, 4, 74, 7, 20, 4, 1, 0, 50, 0, 42, 8, 3, 0, 50, 0, 126, 4, 129, 4, 241, 4, 52, 13, 170, 10, 1, 0, 50, 0, 130, 7, 1, 0, 50, 0, 102, 7, 1, 0, 50, 0, 254, 4, 1, 0, 50, 0, 100, 3, 1, 0, 50, 0, 36, 6, 1, 0, 50, 0, 250, 3, 1, 0, 50, 0, 155, 10, 1, 0, 50, 0, 111, 2, 1, 0, 50, 0, 36, 9, 1, 0, 50, 0, 198, 3, 1, 0, 50, 0, 200, 2, 1, 0, 50, 0, 188, 2, 1, 0, 50, 0, 217, 8, 1, 0, 50, 0, 100, 2, 1, 0, 50, 0, 126, 9, 1, 0, 50, 0, 255, 16, 1, 0, 50, 0, 236, 5, 1, 0, 50, 0, 185, 30, 1, 0, 50, 0, 7, 4, 1, 0, 50, 0, 20, 4, 1, 0, 50, 0, 52, 12, 1, 0, 50, 0, 11, 34, 1, 0, 50, 0, 50, 34, 1, 0, 50, 0, 34, 2, 1, 0, 50, 0, 251, 34, 1, 0, 50, 0, 185, 3, 1, 0, 50, 0, 95, 30, 1, 0, 50, 0, 33, 4, 1, 0, 50, 0, 132, 11, 1, 0, 50, 0, 203, 30, 1, 0, 50, 0, 247, 8, 1, 0, 50, 0, 6, 27, 1, 0, 50, 0, 221, 30, 1, 0, 50, 0, 58, 35, 1, 0, 50, 0, 107, 28, 1, 0, 50, 0, 226, 28, 1, 0, 50, 0, 180, 5, 1, 0, 50, 0, 24, 5, 1, 0, 50, 0, 92, 6, 1, 0, 50, 0, 116, 7, 1, 0, 50, 0, 230, 10, 2, 0, 50, 0, 169, 5, 211, 26, 172, 33, 1, 0, 50, 0, 158, 28, 1, 0, 50, 0, 85, 4, 1, 0, 50, 0, 196, 11, 1, 0, 50, 0, 148, 11, 1, 0, 50, 0, 6, 9, 1, 0, 50, 0, 84, 17, 1, 0, 50, 0, 156, 1, 1, 0, 50, 0, 136, 1, 1, 0, 50, 0, 198, 3, 1, 0, 50, 0, 106, 6, 1, 0, 50, 0, 79, 35, 1, 0, 50, 0, 39, 28, 1, 0, 50, 0, 202, 4, 1, 0, 50, 0, 64, 3, 1, 0, 50, 0, 33, 32, 1, 0, 50, 0, 124, 28, 1, 0, 50, 0, 23, 27, 1, 0, 50, 0, 141, 9, 1, 0, 50, 0, 51, 9, 1, 0, 50, 0, 217, 8, 1, 0, 50, 0, 81, 9, 1, 0, 50, 0, 181, 32, 1, 0, 50, 0, 78, 6, 1, 0, 50, 0, 236, 5, 1, 0, 50, 0, 232, 8, 1, 0, 50, 0, 51, 32, 1, 0, 50, 0, 148, 3, 1, 0, 50, 0, 246, 1, 1, 0, 50, 0, 66, 9, 1, 0, 50, 0, 210, 34, 1, 0, 50, 0, 189, 35, 1, 0, 50, 0, 162, 32, 1, 0, 50, 0, 143, 32, 1, 0, 50, 0, 57, 27, 1, 0, 50, 0, 156, 9, 1, 0, 50, 0, 248, 2, 1, 0, 50, 0, 222, 5, 1, 0, 50, 0, 194, 5, 1, 0, 50, 0, 211, 3, 1, 0, 50, 0, 244, 13, 1, 0, 50, 0, 87, 32, 1, 0, 50, 0, 98, 4, 1, 0, 50, 0, 180, 13, 1, 0, 50, 0, 250, 5, 1, 0, 50, 0, 191, 29, 1, 0, 50, 0, 136, 1, 1, 0, 50, 0, 96, 33, 1, 0, 50, 0, 4, 3, 1, 0, 50, 0, 194, 5, 1, 0, 50, 0, 98, 8, 1, 0, 50, 0, 100, 12, 1, 0, 50, 0, 16, 17, 1, 0, 50, 0, 101, 17, 1, 0, 50, 0, 239, 30, 1, 0, 50, 0, 111, 4, 1, 0, 50, 0, 171, 9, 1, 0, 50, 0, 172, 3, 1, 0, 50, 0, 143, 32, 1, 0, 50, 0, 224, 3, 1, 0, 50, 0, 36, 13, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 172, 8, 3, 0, 187, 8, 36, 14, 134, 18, 3, 0, 112, 8, 134, 18, 202, 18, 3, 0, 187, 8, 202, 18, 65, 19, 3, 0, 112, 8, 65, 19, 53, 14, 1, 0, 172, 8, 1, 0, 1, 0, 1, 0, 157, 8, 1, 0, 1, 0, 3, 0, 59, 30, 36, 14, 32, 18, 3, 0, 1, 0, 32, 18, 53, 14, 1, 0, 142, 8, 1, 0, 157, 8, 1, 0, 142, 8, 1, 0, 127, 8, 3, 0, 187, 8, 36, 14, 137, 15, 3, 0, 112, 8, 137, 15, 235, 19, 3, 0, 187, 8, 235, 19, 92, 26, 3, 0, 187, 8, 36, 14, 137, 15, 3, 0, 112, 8, 137, 15, 53, 14, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 172, 8, 1, 0, 172, 8, 1, 0, 157, 8, 3, 0, 59, 30, 36, 14, 83, 18, 3, 0, 187, 8, 83, 18, 92, 26, 1, 0, 1, 0, 1, 0, 142, 8, 1, 0, 142, 8, 1, 0, 52, 11, 3, 0, 142, 8, 36, 14, 68, 16, 3, 0, 172, 8, 68, 16, 53, 14, 1, 0, 172, 8, 3, 0, 142, 8, 36, 14, 68, 16, 3, 0, 172, 8, 68, 16, 246, 25, 3, 0, 142, 8, 246, 25, 53, 14, 1, 0, 142, 8, 1, 0, 157, 8, 1, 0, 157, 8, 1, 0, 157, 8, 1, 0, 1, 0, 1, 0, 157, 8, 1, 0, 142, 8, 1, 0, 142, 8, 1, 0, 157, 8, 1, 0, 142, 8, 1, 0, 52, 11, 1, 0, 52, 11, 1, 0, 172, 8, 3, 0, 224, 0, 36, 14, 186, 17, 3, 0, 1, 0, 186, 17, 53, 14, 1, 0, 172, 8, 1, 0, 157, 8, 1, 0, 157, 8, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 157, 8, 3, 0, 1, 0, 36, 14, 7, 26, 3, 0, 157, 8, 7, 26, 143, 26, 3, 0, 1, 0, 143, 26, 53, 14, 3, 0, 127, 8, 36, 14, 133, 19, 3, 0, 112, 8, 133, 19, 149, 20, 3, 0, 127, 8, 149, 20, 130, 22, 3, 0, 112, 8, 130, 22, 181, 22, 3, 0, 127, 8, 181, 22, 213, 24, 3, 0, 112, 8, 213, 24, 230, 24, 3, 0, 127, 8, 230, 24, 53, 14, 1, 0, 112, 8, 3, 0, 52, 11, 36, 14, 132, 20, 3, 0, 142, 8, 132, 20, 28, 22, 3, 0, 157, 8, 28, 22, 229, 25, 3, 0, 142, 8, 229, 25, 53, 14, 3, 0, 121, 0, 36, 14, 206, 14, 3, 0, 68, 11, 91, 27, 53, 14, 3, 0, 208, 5, 36, 14, 1, 15, 3, 0, 35, 0, 1, 15, 121, 14, 3, 0, 114, 0, 121, 14, 53, 14, 1, 0, 16, 1, 1, 0, 16, 1, 1, 0, 25, 1, 3, 0, 106, 1, 36, 14, 171, 15, 3, 0, 225, 29, 171, 15, 205, 15, 3, 0, 106, 1, 205, 15, 189, 14, 3, 0, 225, 29, 189, 14, 223, 14, 3, 0, 106, 1, 223, 14, 53, 14, 3, 0, 106, 1, 36, 14, 189, 14, 3, 0, 225, 29, 189, 14, 223, 14, 3, 0, 106, 1, 223, 14, 53, 14, 3, 0, 106, 1, 36, 14, 188, 15, 3, 0, 106, 1, 239, 15, 53, 14, 3, 0, 106, 1, 36, 14, 171, 15, 3, 0, 225, 29, 171, 15, 205, 15, 3, 0, 106, 1, 205, 15, 119, 16, 3, 0, 225, 29, 119, 16, 136, 16, 3, 0, 106, 1, 136, 16, 53, 14, 3, 0, 106, 1, 36, 14, 115, 20, 3, 0, 225, 29, 115, 20, 234, 20, 3, 0, 106, 1, 234, 20, 232, 22, 3, 0, 225, 29, 232, 22, 44, 23, 3, 0, 106, 1, 44, 23, 119, 16, 3, 0, 225, 29, 119, 16, 136, 16, 3, 0, 106, 1, 136, 16, 248, 23, 3, 0, 225, 29, 248, 23, 53, 14, 3, 0, 106, 1, 36, 14, 188, 15, 3, 0, 106, 1, 239, 15, 189, 14, 3, 0, 225, 29, 189, 14, 180, 23, 3, 0, 106, 1, 180, 23, 53, 14, 3, 0, 106, 1, 36, 14, 159, 27, 3, 0, 106, 1, 223, 14, 53, 14, 1, 0, 16, 1, 1, 0, 88, 1, 1, 0, 25, 1, 3, 0, 36, 11, 36, 14, 240, 14, 3, 0, 70, 14, 240, 14, 111, 24, 3, 0, 4, 11, 111, 24, 53, 14, 1, 0, 16, 1, 1, 0, 25, 1, 1, 0, 4, 11, 1, 0, 16, 1, 1, 0, 107, 0, 1, 0, 61, 1, 1, 0, 70, 14, 1, 0, 106, 1, 3, 0, 70, 14, 36, 14, 10, 23, 3, 0, 4, 11, 10, 23, 87, 14, 3, 0, 20, 11, 87, 14, 95, 23, 3, 0, 4, 11, 95, 23, 129, 23, 3, 0, 70, 14, 129, 23, 53, 14, 1, 0, 107, 0, 3, 0, 4, 11, 36, 14, 103, 15, 3, 0, 20, 11, 103, 15, 215, 22, 3, 0, 4, 11, 215, 22, 76, 25, 3, 0, 20, 11, 76, 25, 53, 14, 1, 0, 236, 1, 3, 0, 106, 1, 36, 14, 188, 15, 3, 0, 106, 1, 239, 15, 189, 14, 3, 0, 225, 29, 189, 14, 223, 14, 3, 0, 106, 1, 223, 14, 53, 14, 1, 0, 28, 8, 1, 0, 20, 11, 1, 0, 4, 11, 3, 0, 4, 11, 36, 14, 34, 16, 3, 0, 202, 8, 34, 16, 53, 14, 1, 0, 20, 11, 3, 0, 106, 1, 36, 14, 188, 15, 3, 0, 106, 1, 239, 15, 53, 14, 1, 0, 4, 11, 1, 0, 70, 14, 1, 0, 107, 0, 1, 0, 16, 1, 3, 0, 113, 30, 36, 14, 79, 22, 3, 0, 1, 0, 79, 22, 53, 14, 3, 0, 35, 0, 36, 14, 254, 17, 3, 0, 36, 11, 254, 17, 53, 14, 3, 0, 36, 11, 36, 14, 203, 17, 3, 0, 70, 14, 203, 17, 53, 14, 1, 0, 70, 14, 1, 0, 20, 11, 1, 0, 16, 1, 1, 0, 70, 14, 3, 0, 5, 0, 36, 14, 204, 16, 3, 0, 107, 0, 204, 16, 238, 16, 3, 0, 5, 0, 238, 16, 53, 14, 1, 0, 4, 11, 3, 0, 36, 11, 36, 14, 93, 25, 3, 0, 70, 14, 93, 25, 53, 14, 1, 0, 25, 1, 1, 0, 16, 1, 1, 0, 113, 30, 3, 0, 16, 1, 36, 14, 47, 20, 3, 0, 157, 30, 47, 20, 64, 20, 3, 0, 16, 1, 64, 20, 53, 14, 3, 0, 20, 11, 36, 14, 127, 25, 3, 0, 16, 1, 127, 25, 41, 26, 3, 0, 20, 11, 41, 26, 53, 14, 1, 0, 16, 1, 1, 0, 16, 1, 1, 0, 4, 11, 1, 0, 200, 0, 1, 0, 104, 9, 1, 0, 16, 1, 1, 0, 20, 0, 3, 0, 36, 11, 36, 14, 240, 14, 3, 0, 202, 8, 240, 14, 53, 14, 3, 0, 4, 11, 36, 14, 97, 21, 3, 0, 20, 11, 97, 21, 138, 14, 3, 0, 4, 11, 138, 14, 53, 14, 3, 0, 20, 11, 36, 14, 69, 15, 3, 0, 4, 11, 69, 15, 86, 15, 3, 0, 20, 11, 86, 15, 53, 14, 3, 0, 4, 11, 36, 14, 151, 18, 3, 0, 20, 11, 151, 18, 138, 14, 3, 0, 4, 11, 138, 14, 187, 16, 3, 0, 20, 11, 187, 16, 53, 14, 3, 0, 20, 11, 36, 14, 138, 14, 3, 0, 4, 11, 138, 14, 53, 14, 1, 0, 20, 11, 3, 0, 20, 11, 36, 14, 138, 14, 3, 0, 4, 11, 138, 14, 187, 16, 3, 0, 20, 11, 187, 16, 53, 14, 3, 0, 20, 11, 36, 14, 138, 14, 3, 0, 4, 11, 138, 14, 170, 16, 3, 0, 20, 11, 170, 16, 53, 14, 1, 0, 20, 11, 3, 0, 36, 11, 36, 14, 185, 18, 3, 0, 70, 14, 185, 18, 53, 14, 3, 0, 20, 11, 36, 14, 51, 16, 3, 0, 4, 11, 51, 16, 87, 14, 3, 0, 20, 11, 87, 14, 53, 14, 1, 0, 20, 11, 3, 0, 106, 1, 36, 14, 98, 20, 3, 0, 106, 1, 108, 27, 53, 14, 3, 0, 36, 11, 36, 14, 219, 18, 3, 0, 35, 0, 219, 18, 253, 18, 3, 0, 36, 11, 253, 18, 120, 15, 3, 0, 35, 0, 120, 15, 121, 14, 3, 0, 114, 0, 121, 14, 53, 14, 3, 0, 4, 11, 36, 14, 87, 14, 3, 0, 20, 11, 87, 14, 53, 14, 1, 0, 16, 1, 1, 0, 216, 0, 1, 0, 30, 0, 1, 0, 36, 11, 3, 0, 20, 11, 36, 14, 69, 15, 3, 0, 4, 11, 69, 15, 86, 15, 3, 0, 20, 11, 86, 15, 53, 14, 1, 0, 16, 1, 1, 0, 25, 1, 3, 0, 4, 11, 36, 14, 237, 17, 3, 0, 20, 11, 237, 17, 49, 18, 3, 0, 4, 11, 49, 18, 131, 21, 3, 0, 20, 11, 131, 21, 165, 21, 3, 0, 4, 11, 165, 21, 233, 21, 3, 0, 20, 11, 233, 21, 147, 22, 3, 0, 4, 11, 147, 22, 53, 14, 1, 0, 107, 0, 1, 0, 16, 1, 1, 0, 16, 1, 1, 0, 4, 11, 3, 0, 36, 11, 36, 14, 240, 14, 3, 0, 202, 8, 240, 14, 53, 14, 3, 0, 106, 1, 36, 14, 98, 20, 3, 0, 106, 1, 125, 27, 142, 27, 3, 0, 106, 1, 176, 27, 53, 14, 3, 0, 20, 11, 36, 14, 220, 17, 3, 0, 4, 11, 220, 17, 53, 14, 3, 0, 4, 11, 36, 14, 103, 15, 3, 0, 20, 11, 103, 15, 184, 19, 3, 0, 4, 11, 184, 19, 53, 14, 3, 0, 36, 11, 36, 14, 144, 25, 3, 0, 114, 0, 144, 25, 109, 26, 3, 0, 36, 11, 109, 26, 160, 26, 3, 0, 114, 0, 160, 26, 53, 14, 1, 0, 4, 11, 3, 0, 16, 1, 36, 14, 236, 18, 3, 0, 20, 14, 236, 18, 53, 14, 1, 0, 16, 1, 1, 0, 4, 11, 1, 0, 0, 1, 1, 0, 16, 1, 1, 0, 20, 11, 1, 0, 20, 11, 1, 0, 20, 11, 3, 0, 121, 0, 36, 14, 206, 14, 3, 0, 35, 0, 206, 14, 121, 14, 3, 0, 114, 0, 121, 14, 53, 14, 1, 0, 156, 11, 3, 0, 70, 14, 36, 14, 128, 24, 3, 0, 4, 11, 128, 24, 53, 14, 3, 0, 70, 14, 36, 14, 216, 21, 3, 0, 4, 11, 216, 21, 53, 14, 3, 0, 70, 14, 36, 14, 146, 23, 3, 0, 4, 11, 146, 23, 53, 14, 3, 0, 4, 11, 36, 14, 34, 16, 3, 0, 70, 14, 34, 16, 53, 14, 1, 0, 20, 11, 3, 0, 16, 1, 36, 14, 62, 22, 3, 0, 20, 11, 62, 22, 51, 16, 3, 0, 4, 11, 51, 16, 87, 14, 3, 0, 20, 11, 87, 14, 53, 14, 3, 0, 50, 5, 36, 14, 66, 18, 3, 0, 97, 1, 66, 18, 53, 14, 1, 0, 70, 14, 1, 0, 20, 11, 1, 0, 16, 1, 1, 0, 107, 0, 1, 0, 16, 1, 3, 0, 47, 0, 36, 14, 67, 17, 1, 0, 4, 11, 3, 0, 4, 11, 36, 14, 87, 14, 3, 0, 20, 11, 87, 14, 102, 16, 3, 0, 4, 11, 102, 16, 53, 14, 1, 0, 25, 1, 1, 0, 4, 11, 3, 0, 4, 11, 36, 14, 87, 14, 3, 0, 20, 11, 87, 14, 102, 16, 3, 0, 4, 11, 102, 16, 231, 23, 3, 0, 20, 11, 231, 23, 170, 16, 3, 0, 4, 11, 170, 16, 53, 14, 3, 0, 5, 0, 36, 14, 204, 16, 3, 0, 107, 0, 204, 16, 238, 16, 3, 0, 5, 0, 238, 16, 53, 14, 1, 0, 174, 29, 3, 0, 107, 0, 36, 14, 9, 24, 3, 0, 25, 1, 9, 24, 53, 14, 1, 0, 47, 0, 3, 0, 196, 1, 36, 14, 15, 18, 3, 0, 16, 1, 15, 18, 78, 23, 3, 0, 20, 11, 78, 23, 112, 23, 3, 0, 16, 1, 112, 23, 53, 14, 1, 0, 25, 1, 3, 0, 235, 31, 36, 14, 31, 19, 3, 0, 253, 31, 31, 19, 53, 14, 3, 0, 36, 11, 36, 14, 120, 15, 3, 0, 35, 0, 120, 15, 121, 14, 3, 0, 114, 0, 121, 14, 53, 14, 1, 0, 16, 1, 1, 0, 154, 5, 1, 0, 16, 1, 1, 0, 16, 1, 1, 0, 16, 1, 1, 0, 16, 1, 3, 0, 70, 14, 36, 14, 169, 17, 3, 0, 4, 11, 169, 17, 53, 14, 1, 0, 4, 11, 1, 0, 16, 1, 1, 0, 20, 11, 1, 0, 36, 11, 1, 0, 20, 11, 1, 0, 16, 1, 1, 0, 36, 11, 1, 0, 36, 11, 1, 0, 4, 11, 3, 0, 35, 0, 36, 14, 121, 14, 3, 0, 114, 0, 121, 14, 53, 14, 1, 0, 70, 14, 3, 0, 5, 30, 36, 14, 26, 24, 3, 0, 129, 17, 26, 24, 43, 24, 3, 0, 5, 30, 43, 24, 179, 24, 3, 0, 129, 17, 179, 24, 196, 24, 3, 0, 5, 30, 196, 24, 212, 25, 3, 0, 129, 17, 212, 25, 24, 26, 3, 0, 5, 30, 24, 26, 53, 14, 1, 0, 51, 27, 1, 0, 15, 36, 3, 0, 243, 29, 36, 14, 94, 24, 3, 0, 6, 35, 94, 24, 53, 14, 1, 0, 232, 30, 1, 0, 224, 2, 3, 0, 106, 1, 36, 14, 167, 19, 3, 0, 47, 0, 167, 19, 67, 17, 1, 0, 192, 32, 1, 0, 68, 27, 1, 0, 1, 0, 1, 0, 250, 30, 1, 0, 112, 8, 1, 0, 184, 0, 3, 0, 205, 2, 36, 14, 197, 23, 3, 0, 124, 32, 197, 23, 53, 14, 1, 0, 127, 8, 3, 0, 217, 2, 36, 14, 60, 24, 3, 0, 190, 3, 60, 24, 50, 17, 3, 0, 217, 2, 50, 17, 53, 14, 3, 0, 155, 2, 36, 14, 18, 15, 3, 0, 251, 1, 18, 15, 35, 15, 3, 0, 105, 32, 35, 15, 53, 14, 3, 0, 45, 2, 36, 14, 18, 15, 3, 0, 193, 2, 18, 15, 153, 16, 3, 0, 105, 32, 193, 27, 53, 14, 3, 0, 166, 1, 36, 14, 80, 21, 3, 0, 167, 5, 80, 21, 53, 14, 3, 0, 105, 32, 35, 15, 53, 14, 1, 0, 184, 0, 3, 0, 10, 0, 36, 14, 52, 15, 3, 0, 184, 0, 52, 15, 53, 14, 3, 0, 131, 1, 36, 14, 251, 20, 3, 0, 122, 2, 251, 20, 53, 14, 1, 0, 116, 1, 1, 0, 127, 8, 3, 0, 156, 0, 36, 14, 12, 21, 3, 0, 226, 1, 12, 21, 53, 14, 1, 0, 241, 2, 1, 0, 41, 0, 3, 0, 177, 3, 36, 14, 8, 25, 3, 0, 255, 3, 8, 25, 161, 25, 3, 0, 177, 3, 161, 25, 53, 14, 3, 0, 52, 1, 36, 14, 201, 19, 3, 0, 121, 11, 201, 19, 53, 14, 3, 0, 41, 0, 36, 14, 113, 22, 3, 0, 89, 0, 113, 22, 214, 23, 3, 0, 41, 0, 214, 23, 53, 14, 1, 0, 127, 8, 3, 0, 71, 0, 36, 14, 135, 17, 3, 0, 133, 2, 135, 17, 53, 14, 3, 0, 12, 2, 36, 14, 100, 18, 3, 0, 41, 30, 100, 18, 61, 23, 3, 0, 12, 2, 61, 23, 53, 14, 1, 0, 10, 0, 3, 0, 227, 5, 36, 14, 29, 21, 3, 0, 177, 2, 29, 21, 53, 14, 3, 0, 127, 8, 36, 14, 227, 27, 3, 0, 128, 0, 36, 14, 17, 16, 3, 0, 127, 8, 17, 16, 53, 14, 3, 0, 128, 0, 36, 14, 17, 16, 3, 0, 127, 8, 17, 16, 53, 14, 1, 0, 26, 9, 1, 0, 151, 1, 1, 0, 255, 3, 1, 0, 131, 30, 1, 0, 15, 32, 1, 0, 128, 0, 1, 0, 16, 3, 1, 0, 41, 9, 3, 0, 203, 3, 36, 14, 152, 17, 3, 0, 79, 1, 152, 17, 53, 14, 1, 0, 95, 0, 3, 0, 177, 3, 36, 14, 163, 23, 3, 0, 21, 17, 163, 23, 162, 24, 3, 0, 177, 3, 162, 24, 53, 14, 1, 0, 4, 17, 3, 0, 163, 0, 36, 14, 116, 19, 3, 0, 43, 1, 116, 19, 53, 14, 3, 0, 149, 0, 36, 14, 99, 19, 3, 0, 43, 1, 99, 19, 53, 14, 1, 0, 184, 0, 3, 0, 39, 2, 36, 14, 27, 23, 3, 0, 59, 0, 27, 23, 53, 14, 1, 0, 190, 3, 1, 0, 41, 30, 1, 0, 160, 3, 1, 0, 10, 0, 1, 0, 127, 8, 3, 0, 4, 17, 36, 14, 77, 24, 3, 0, 89, 17, 77, 24, 25, 25, 3, 0, 4, 17, 25, 25, 53, 14, 1, 0, 89, 17, 1, 0, 161, 1, 3, 0, 177, 0, 36, 14, 114, 21, 3, 0, 141, 1, 114, 21, 35, 15, 3, 0, 105, 32, 35, 15, 53, 14, 1, 0, 116, 1, 3, 0, 41, 30, 36, 14, 30, 20, 3, 0, 131, 30, 30, 20, 53, 14, 3, 0, 53, 0, 36, 14, 110, 25, 3, 0, 56, 9, 110, 25, 58, 26, 3, 0, 53, 0, 58, 26, 53, 14, 3, 0, 10, 0, 36, 14, 52, 15, 3, 0, 184, 0, 52, 15, 53, 14, 3, 0, 124, 32, 153, 16, 53, 14, 3, 0, 206, 1, 36, 14, 18, 15, 3, 0, 222, 8, 18, 15, 153, 16, 3, 0, 124, 32, 153, 16, 126, 26, 3, 0, 105, 32, 126, 26, 53, 14, 1, 0, 232, 0, 1, 0, 184, 0, 3, 0, 116, 1, 74, 27, 53, 14, 1, 0, 83, 6, 3, 0, 237, 8, 36, 14, 82, 19, 3, 0, 199, 5, 82, 19, 150, 19, 3, 0, 237, 8, 150, 19, 222, 15, 3, 0, 23, 2, 222, 15, 53, 14, 1, 0, 53, 0, 1, 0, 59, 0, 1, 0, 71, 9, 3, 0, 190, 3, 36, 14, 50, 17, 1, 0, 253, 2, 3, 0, 199, 5, 36, 14, 222, 15, 3, 0, 23, 2, 222, 15, 53, 14, 3, 0, 216, 3, 36, 14, 217, 20, 3, 0, 204, 35, 217, 20, 53, 14, 1, 0, 25, 0, 3, 0, 41, 0, 36, 14, 13, 20, 3, 0, 142, 0, 13, 20, 53, 14, 1, 0, 83, 0, 1, 0, 52, 1, 1, 0, 9, 3, 3, 0, 177, 3, 36, 14, 48, 19, 3, 0, 190, 3, 48, 19, 145, 24, 3, 0, 21, 17, 145, 24, 53, 14, 1, 0, 116, 1, 1, 0, 21, 17, 1, 0, 177, 3, 3, 0, 166, 2, 36, 14, 148, 21, 3, 0, 148, 32, 148, 21, 53, 14, 3, 0, 229, 3, 36, 14, 46, 21, 3, 0, 208, 0, 46, 21, 53, 14, 3, 0, 109, 11, 36, 14, 199, 21, 3, 0, 187, 8, 199, 21, 250, 21, 3, 0, 109, 11, 250, 21, 53, 14, 1, 0, 16, 1, 1, 0, 187, 8, 1, 0, 39, 34, 1, 0, 187, 8, 1, 0, 187, 8, 1, 0, 1, 0, 1, 0, 198, 35, 1, 0, 1, 0, 1, 0, 70, 1, 1, 0, 23, 30, 1, 0, 243, 29, 1, 0, 23, 30, 1, 0, 243, 29, 1, 0, 23, 30, 1, 0, 235, 35, 1, 0, 243, 29, 1, 0, 243, 29, 3, 0, 243, 29, 36, 14, 14, 19, 3, 0, 60, 34, 14, 19, 53, 14, 1, 0, 243, 29, 1, 0, 5, 30, 1, 0, 243, 29, 1, 0, 4, 11, 1, 0, 20, 11, 1, 0, 1, 0, 1, 0, 112, 8, 1, 0, 112, 8, 3, 0, 187, 5, 0, 16, 178, 25, 1, 0, 127, 8, 1, 0, 112, 8, 1, 0, 112, 8, 1, 0, 112, 8, 1, 0, 112, 8, 1, 0, 127, 8, 1, 0, 112, 8, 1, 0, 112, 8, 3, 0, 187, 5, 36, 14, 166, 20, 3, 0, 127, 8, 166, 20, 53, 14, 1, 0, 112, 8, 3, 0, 77, 0, 36, 14, 104, 14, 3, 0, 1, 0, 104, 14, 53, 14, 1, 0, 112, 8, 3, 0, 192, 0, 36, 14, 104, 14, 3, 0, 1, 0, 104, 14, 53, 14, 1, 0, 127, 8, 3, 0, 192, 0, 36, 14, 104, 14, 3, 0, 1, 0, 104, 14, 53, 14, 3, 0, 127, 8, 36, 14, 168, 18, 3, 0, 135, 0, 168, 18, 218, 19, 3, 0, 127, 8, 218, 19, 195, 25, 3, 0, 135, 0, 195, 25, 53, 14, 3, 0, 192, 0, 36, 14, 104, 14, 3, 0, 1, 0, 104, 14, 53, 14, 3, 0, 187, 5, 36, 14, 155, 14, 3, 0, 127, 8, 155, 14, 221, 16, 3, 0, 166, 35, 221, 16, 42, 25, 3, 0, 127, 8, 42, 25, 53, 14, 3, 0, 187, 5, 36, 14, 183, 20, 3, 0, 127, 8, 183, 20, 53, 14, 3, 0, 112, 8, 36, 14, 117, 18, 3, 0, 187, 8, 117, 18, 182, 21, 3, 0, 112, 8, 182, 21, 96, 22, 3, 0, 187, 8, 96, 22, 53, 14, 1, 0, 112, 8, 3, 0, 192, 0, 36, 14, 104, 14, 3, 0, 1, 0, 104, 14, 53, 14, 1, 0, 112, 8, 1, 0, 112, 8, 1, 0, 112, 8, 1, 0, 127, 8, 3, 0, 187, 5, 36, 14, 172, 14, 3, 0, 127, 8, 172, 14, 221, 16, 3, 0, 166, 35, 221, 16, 59, 25, 3, 0, 187, 5, 59, 25, 53, 14, 1, 0, 112, 8, 3, 0, 187, 5, 36, 14, 172, 14, 3, 0, 127, 8, 172, 14, 33, 17, 3, 0, 187, 5, 33, 17, 53, 14, 1, 0, 112, 8, 1, 0, 112, 8, 1, 0, 112, 8, 1, 0, 112, 8, 3, 0, 187, 5, 36, 14, 155, 14, 3, 0, 127, 8, 155, 14, 53, 14, 1, 0, 112, 8, 3, 0, 216, 1, 36, 14, 81, 20, 3, 0, 187, 5, 81, 20, 172, 14, 3, 0, 127, 8, 172, 14, 63, 21, 3, 0, 243, 5, 63, 21, 53, 14, 1, 0, 112, 8, 1, 0, 112, 8, 3, 0, 187, 5, 0, 16, 67, 17, 3, 0, 187, 5, 36, 14, 154, 15, 3, 0, 127, 8, 154, 15, 45, 22, 3, 0, 187, 5, 45, 22, 164, 22, 3, 0, 127, 8, 164, 22, 247, 24, 3, 0, 187, 5, 247, 24, 53, 14, 1, 0, 112, 8, 1, 0, 127, 8, 1, 0, 112, 8, 3, 0, 187, 5, 36, 14, 155, 14, 3, 0, 127, 8, 155, 14, 53, 14, 1, 0, 112, 8, 3, 0, 187, 5, 33, 17, 178, 25, 3, 0, 187, 5, 36, 14, 154, 15, 3, 0, 112, 8, 154, 15, 200, 20, 3, 0, 127, 8, 200, 20, 53, 14, 1, 0, 112, 8, 1, 0, 112, 8, 1, 0, 112, 8, 3, 0, 187, 5, 36, 14, 155, 14, 3, 0, 127, 8, 155, 14, 198, 22, 3, 0, 112, 8, 198, 22, 249, 22, 3, 0, 127, 8, 249, 22, 53, 14, 3, 0, 108, 17, 36, 14, 0, 16, 3, 0, 187, 5, 0, 16, 75, 26, 3, 0, 108, 17, 75, 26, 53, 14, 1, 0, 112, 8, 1, 0, 112, 8, 3, 0, 187, 5, 36, 14, 172, 14, 3, 0, 127, 8, 172, 14, 53, 14, 1, 0, 112, 8, 1, 0, 172, 8, 1, 0, 128, 5, 1, 0, 252, 26, 1, 0, 53, 4, 1, 0, 172, 8, 1, 0, 212, 13, 1, 0, 144, 2, 1, 0, 139, 11, 1, 0, 13, 27, 1, 0, 172, 8, 1, 0, 88, 9, 1, 0, 70, 14, 1, 0, 36, 11, 1, 0, 67, 4, 1, 0, 224, 2, 3, 0, 228, 26, 36, 14, 210, 27, 1, 0, 172, 11, 1, 0, 119, 9, 1, 0, 8, 1, 1, 0, 4, 14, 1, 0, 248, 0, 1, 0, 80, 4, 1, 0, 170, 0, 3, 0, 200, 0, 36, 14, 252, 19, 3, 0, 175, 30, 252, 19, 53, 14, 1, 0, 188, 11, 1, 0, 240, 0, 3, 0, 245, 3, 36, 14, 85, 16, 3, 0, 34, 1, 85, 16, 53, 14, 3, 0, 208, 5, 36, 14, 1, 15, 3, 0, 68, 11, 1, 15, 53, 14, 3, 0, 208, 5, 36, 14, 1, 15, 3, 0, 68, 11, 1, 15, 53, 14, 1, 0, 141, 5, 1, 0, 134, 9, 3, 0, 193, 30, 36, 14, 11, 22, 3, 0, 194, 26, 11, 22, 53, 14, 1, 0, 194, 26, 1, 0, 211, 30, 3, 0, 121, 0, 36, 14, 206, 14, 3, 0, 65, 0, 206, 14, 53, 14, 1, 0, 100, 6, 1, 0, 93, 4, 1, 0, 204, 11, 1, 0, 70, 8, 3, 0, 121, 0, 36, 14, 206, 14, 3, 0, 65, 0, 206, 14, 53, 14, 1, 0, 114, 6, 1, 0, 31, 27, 1, 0, 149, 9, 1, 0, 228, 26, 1, 0, 15, 0, 3, 0, 84, 8, 36, 14, 85, 16, 3, 0, 34, 1, 85, 16, 53, 14, 1, 0, 164, 9, 1, 0, 228, 13, 1, 0, 101, 0, 1, 0, 106, 4, 1, 0, 119, 4, 1, 0, 179, 9, 11, 0, 33, 34, 36, 34, 16, 3, 39, 34, 22, 3, 42, 34, 45, 34, 48, 34, 51, 34, 54, 34, 57, 34, 177, 26, 8, 6, 22, 6, 77, 30, 64, 6, 247, 8, 95, 30, 211, 26, 158, 7, 112, 3, 194, 5, 170, 170, 191, 0, 45, 0, 54, 0, 66, 0, 102, 0, 144, 0, 162, 0, 184, 0, 226, 0, 237, 0, 244, 0, 152, 21, 2, 1, 9, 1, 46, 1, 77, 1, 94, 1, 175, 21, 170, 31, 187, 21, 198, 21, 110, 1, 133, 1, 143, 1, 161, 1, 169, 1, 179, 1, 254, 1, 16, 2, 41, 2, 59, 2, 77, 2, 196, 25, 6, 22, 88, 2, 99, 2, 106, 2, 113, 2, 121, 2, 128, 2, 137, 2, 54, 22, 245, 25, 177, 20, 145, 2, 221, 31, 157, 2, 163, 2, 91, 22, 17, 31, 34, 31, 169, 2, 178, 2, 183, 2, 188, 2, 194, 20, 194, 2, 211, 20, 166, 22, 204, 2, 217, 2, 237, 31, 228, 2, 236, 2, 85, 3, 127, 3, 153, 3, 171, 3, 44, 32, 180, 3, 194, 3, 210, 3, 94, 27, 74, 32, 92, 32, 67, 26, 0, 4, 180, 13, 16, 4, 34, 4, 52, 4, 128, 32, 70, 4, 32, 14, 78, 4, 217, 22, 232, 22, 94, 4, 103, 4, 116, 4, 132, 4, 150, 4, 168, 4, 186, 4, 191, 4, 242, 22, 197, 4, 204, 4, 52, 23, 67, 23, 210, 4, 229, 4, 248, 4, 2, 5, 194, 32, 108, 23, 8, 5, 209, 32, 18, 5, 28, 5, 34, 5, 42, 5, 11, 27, 168, 23, 237, 20, 179, 23, 55, 5, 62, 5, 90, 31, 242, 32, 71, 5, 106, 31, 2, 21, 88, 5, 105, 5, 120, 5, 80, 29, 129, 5, 19, 33, 137, 5, 143, 5, 157, 5, 172, 5, 33, 33, 46, 33, 86, 17, 185, 5, 5, 24, 22, 24, 32, 24, 199, 5, 95, 33, 208, 5, 225, 5, 234, 5, 239, 5, 251, 5, 11, 6, 109, 33, 126, 33, 73, 24, 113, 24, 138, 31, 57, 21, 165, 24, 174, 29, 179, 24, 27, 6, 33, 6, 44, 6, 219, 24, 55, 6, 61, 26, 63, 6, 72, 6, 76, 21, 195, 33, 253, 24, 83, 6, 9, 25, 23, 25, 94, 6, 102, 6, 243, 33, 108, 6, 115, 6, 128, 6, 135, 6, 142, 6, 89, 25, 150, 6, 161, 6, 169, 6, 130, 25, 162, 30, 110, 21, 0, 34, 13, 34, 147, 25, 160, 25, 179, 25, 179, 6, 57, 36, 0, 80, 60, 36, 0, 80, 63, 36, 0, 80, 80, 36, 0, 80, 101, 36, 0, 80, 104, 36, 0, 80, 111, 36, 0, 80, 134, 36, 0, 80, 137, 36, 0, 80, 140, 36, 0, 80, 143, 36, 0, 80, 146, 36, 0, 80, 149, 36, 0, 80, 166, 36, 0, 80, 181, 36, 0, 80, 188, 36, 0, 80, 195, 36, 0, 80, 198, 36, 0, 80, 201, 36, 0, 80, 204, 36, 0, 80, 207, 36, 0, 80, 220, 36, 0, 80, 223, 36, 0, 80, 226, 36, 0, 80, 229, 36, 0, 80, 232, 36, 0, 80, 23, 37, 0, 80, 26, 37, 0, 80, 29, 37, 0, 80, 32, 37, 0, 80, 35, 37, 0, 80, 38, 37, 0, 80, 41, 37, 0, 80, 44, 37, 0, 80, 47, 37, 0, 80, 50, 37, 0, 80, 53, 37, 0, 80, 56, 37, 0, 80, 59, 37, 0, 80, 62, 37, 0, 80, 65, 37, 0, 80, 68, 37, 0, 80, 71, 37, 0, 80, 74, 37, 0, 80, 79, 37, 0, 80, 82, 37, 0, 80, 85, 37, 0, 80, 88, 37, 0, 80, 91, 37, 0, 80, 94, 37, 0, 80, 97, 37, 0, 80, 100, 37, 0, 80, 103, 37, 0, 80, 106, 37, 0, 80, 109, 37, 0, 80, 112, 37, 0, 80, 115, 37, 0, 80, 118, 37, 0, 80, 121, 37, 0, 80, 124, 37, 0, 80, 127, 37, 0, 80, 130, 37, 0, 80, 133, 37, 0, 80, 198, 37, 0, 80, 219, 37, 0, 80, 224, 37, 0, 80, 229, 37, 0, 80, 232, 37, 0, 80, 235, 37, 0, 80, 238, 37, 0, 80, 241, 37, 0, 80, 244, 37, 0, 80, 17, 38, 0, 80, 20, 38, 0, 80, 23, 38, 0, 80, 26, 38, 0, 80, 29, 38, 0, 80, 32, 38, 0, 80, 35, 38, 0, 80, 38, 38, 0, 80, 41, 38, 0, 80, 44, 38, 0, 80, 49, 38, 0, 80, 52, 38, 0, 80, 55, 38, 0, 80, 58, 38, 0, 80, 61, 38, 0, 80, 66, 38, 0, 80, 69, 38, 0, 80, 76, 38, 0, 80, 79, 38, 0, 80, 82, 38, 0, 80, 85, 38, 0, 80, 88, 38, 0, 80, 91, 38, 0, 80, 94, 38, 0, 80, 97, 38, 0, 80, 100, 38, 0, 80, 103, 38, 0, 80, 106, 38, 0, 80, 109, 38, 0, 80, 112, 38, 0, 80, 115, 38, 0, 80, 118, 38, 0, 80, 121, 38, 0, 80, 124, 38, 0, 80, 127, 38, 0, 80, 130, 38, 0, 80, 133, 38, 0, 80, 136, 38, 0, 80, 139, 38, 0, 80, 142, 38, 0, 80, 145, 38, 0, 80, 148, 38, 0, 80, 151, 38, 0, 80, 154, 38, 0, 80, 157, 38, 0, 80, 160, 38, 0, 80, 163, 38, 0, 80, 166, 38, 0, 80, 169, 38, 0, 80, 172, 38, 0, 80, 175, 38, 0, 80, 178, 38, 0, 80, 181, 38, 0, 80, 184, 38, 0, 80, 187, 38, 0, 80, 190, 38, 0, 80, 193, 38, 0, 80, 196, 38, 0, 80, 199, 38, 0, 80, 204, 38, 0, 80, 207, 38, 0, 80, 210, 38, 0, 80, 213, 38, 0, 80, 216, 38, 0, 80, 219, 38, 0, 80, 222, 38, 0, 80, 225, 38, 0, 80, 228, 38, 0, 80, 231, 38, 0, 80, 234, 38, 0, 80, 237, 38, 0, 80, 240, 38, 0, 80, 243, 38, 0, 80, 246, 38, 0, 80, 249, 38, 0, 80, 252, 38, 0, 80, 255, 38, 0, 80, 2, 39, 0, 80, 5, 39, 0, 80, 8, 39, 0, 80, 11, 39, 0, 80, 14, 39, 0, 80, 17, 39, 0, 80, 20, 39, 0, 80, 23, 39, 0, 80, 26, 39, 0, 80, 29, 39, 0, 80, 32, 39, 0, 80, 35, 39, 0, 80, 38, 39, 0, 80, 41, 39, 0, 80, 44, 39, 0, 80, 47, 39, 0, 80, 50, 39, 0, 80, 53, 39, 0, 80, 56, 39, 0, 80, 59, 39, 0, 80, 62, 39, 0, 80, 65, 39, 0, 80, 68, 39, 0, 80, 71, 39, 0, 80, 74, 39, 0, 80, 77, 39, 0, 80, 80, 39, 0, 80, 83, 39, 0, 80, 86, 39, 0, 80, 89, 39, 0, 80, 92, 39, 0, 80, 95, 39, 0, 80, 98, 39, 0, 80, 101, 39, 0, 80, 104, 39, 0, 80, 107, 39, 0, 80, 110, 39, 0, 80, 113, 39, 0, 80, 116, 39, 0, 80, 119, 39, 0, 80, 122, 39, 0, 80, 125, 39, 0, 80, 1, 0, 0, 0, 128, 39, 0, 144, 1, 0, 0, 0, 130, 39, 0, 144, 1, 0, 0, 0, 132, 39, 0, 144, 4, 0, 0, 0, 134, 39, 0, 144, 138, 39, 0, 144, 142, 39, 0, 144, 146, 39, 0, 144, 1, 0, 0, 0, 150, 39, 0, 144, 1, 0, 0, 0, 152, 39, 0, 144, 1, 0, 0, 0, 154, 39, 0, 144, 1, 0, 0, 0, 156, 39, 0, 144, 2, 0, 0, 0, 158, 39, 0, 144, 162, 39, 0, 144, 1, 0, 0, 0, 166, 39, 0, 144, 1, 0, 0, 0, 168, 39, 0, 144, 1, 0, 0, 0, 170, 39, 0, 144, 1, 0, 0, 0, 172, 39, 0, 144, 3, 0, 0, 0, 174, 39, 0, 144, 178, 39, 0, 144, 182, 39, 0, 144, 2, 0, 0, 0, 186, 39, 0, 144, 190, 39, 0, 144, 1, 0, 0, 0, 194, 39, 0, 144, 1, 0, 0, 0, 196, 39, 0, 144, 1, 0, 0, 0, 198, 39, 0, 144, 1, 0, 0, 0, 200, 39, 0, 144, 1, 0, 0, 0, 202, 39, 0, 144, 2, 0, 0, 0, 204, 39, 0, 144, 208, 39, 0, 144, 1, 0, 0, 0, 212, 39, 0, 144, 1, 0, 0, 0, 214, 39, 0, 144, 1, 0, 0, 0, 216, 39, 0, 144, 1, 0, 0, 0, 218, 39, 0, 144, 2, 0, 0, 0, 220, 39, 0, 144, 224, 39, 0, 144, 1, 0, 0, 0, 228, 39, 0, 144, 3, 0, 0, 0, 230, 39, 0, 144, 234, 39, 0, 144, 238, 39, 0, 144, 1, 0, 0, 0, 242, 39, 0, 144, 1, 0, 0, 0, 244, 39, 0, 144, 1, 0, 0, 0, 246, 39, 0, 144, 1, 0, 0, 0, 248, 39, 0, 144, 1, 0, 0, 0, 250, 39, 0, 144, 1, 0, 0, 0, 252, 39, 0, 144, 1, 0, 0, 0, 254, 39, 0, 144, 1, 0, 0, 0, 0, 40, 0, 144, 1, 0, 0, 0, 2, 40, 0, 144, 1, 0, 0, 0, 4, 40, 0, 144, 1, 0, 0, 0, 6, 40, 0, 144, 1, 0, 0, 0, 8, 40, 0, 144, 1, 0, 0, 0, 10, 40, 0, 144, 2, 0, 0, 0, 12, 40, 0, 144, 16, 40, 0, 144, 1, 0, 0, 0, 20, 40, 0, 144, 1, 0, 0, 0, 22, 40, 0, 144, 1, 0, 0, 0, 24, 40, 0, 144, 1, 0, 0, 0, 26, 40, 0, 144, 1, 0, 0, 0, 28, 40, 0, 144, 1, 0, 0, 0, 30, 40, 0, 144, 3, 0, 0, 0, 32, 40, 0, 144, 36, 40, 0, 144, 40, 40, 0, 144, 7, 0, 0, 0, 44, 40, 0, 144, 48, 40, 0, 144, 52, 40, 0, 144, 56, 40, 0, 144, 60, 40, 0, 144, 64, 40, 0, 144, 68, 40, 0, 144, 1, 0, 0, 0, 72, 40, 0, 144, 4, 0, 0, 0, 74, 40, 0, 144, 78, 40, 0, 144, 82, 40, 0, 144, 86, 40, 0, 144, 2, 0, 0, 0, 90, 40, 0, 144, 94, 40, 0, 144, 3, 0, 0, 0, 98, 40, 0, 144, 102, 40, 0, 144, 106, 40, 0, 144, 1, 0, 0, 0, 110, 40, 0, 144, 1, 0, 0, 0, 112, 40, 0, 144, 1, 0, 0, 0, 114, 40, 0, 144, 5, 0, 0, 0, 116, 40, 0, 144, 120, 40, 0, 144, 124, 40, 0, 144, 128, 40, 0, 144, 132, 40, 0, 144, 3, 0, 0, 0, 136, 40, 0, 144, 140, 40, 0, 144, 144, 40, 0, 144, 2, 0, 0, 0, 148, 40, 0, 144, 152, 40, 0, 144, 5, 0, 0, 0, 156, 40, 0, 144, 160, 40, 0, 144, 164, 40, 0, 144, 168, 40, 0, 144, 172, 40, 0, 144, 8, 0, 0, 0, 176, 40, 0, 144, 180, 40, 0, 144, 184, 40, 0, 144, 188, 40, 0, 144, 192, 40, 0, 144, 196, 40, 0, 144, 200, 40, 0, 144, 204, 40, 0, 144, 4, 0, 0, 0, 208, 40, 0, 144, 212, 40, 0, 144, 216, 40, 0, 144, 220, 40, 0, 144, 2, 0, 0, 0, 224, 40, 0, 144, 228, 40, 0, 144, 1, 0, 0, 0, 232, 40, 0, 144, 1, 0, 0, 0, 234, 40, 0, 144, 1, 0, 0, 0, 236, 40, 0, 144, 3, 0, 0, 0, 238, 40, 0, 144, 242, 40, 0, 144, 246, 40, 0, 144, 1, 0, 0, 0, 250, 40, 0, 144, 1, 0, 0, 0, 252, 40, 0, 144, 1, 0, 0, 0, 254, 40, 0, 144, 1, 0, 0, 0, 0, 41, 0, 144, 1, 0, 0, 0, 2, 41, 0, 144, 1, 0, 0, 0, 4, 41, 0, 144, 1, 0, 0, 0, 6, 41, 0, 144, 1, 0, 0, 0, 8, 41, 0, 144, 5, 0, 0, 0, 10, 41, 0, 144, 14, 41, 0, 144, 18, 41, 0, 144, 22, 41, 0, 144, 26, 41, 0, 144, 1, 0, 0, 0, 30, 41, 0, 144, 4, 0, 0, 0, 32, 41, 0, 144, 36, 41, 0, 144, 40, 41, 0, 144, 44, 41, 0, 144, 1, 0, 0, 0, 48, 41, 0, 144, 4, 0, 0, 0, 50, 41, 0, 144, 54, 41, 0, 144, 58, 41, 0, 144, 62, 41, 0, 144, 1, 0, 0, 0, 66, 41, 0, 144, 1, 0, 0, 0, 68, 41, 0, 144, 1, 0, 0, 0, 70, 41, 0, 144, 2, 0, 0, 0, 72, 41, 0, 144, 76, 41, 0, 144, 1, 0, 0, 0, 80, 41, 0, 144, 2, 0, 0, 0, 82, 41, 0, 144, 86, 41, 0, 144, 1, 0, 0, 0, 90, 41, 0, 144, 1, 0, 0, 0, 92, 41, 0, 144, 1, 0, 0, 0, 94, 41, 0, 144, 1, 0, 0, 0, 96, 41, 0, 144, 2, 0, 0, 0, 98, 41, 0, 144, 102, 41, 0, 144, 2, 0, 0, 0, 106, 41, 0, 144, 110, 41, 0, 144, 2, 0, 0, 0, 114, 41, 0, 144, 118, 41, 0, 144, 1, 0, 0, 0, 122, 41, 0, 144, 1, 0, 0, 0, 124, 41, 0, 144, 1, 0, 0, 0, 126, 41, 0, 144, 1, 0, 0, 0, 128, 41, 0, 144, 3, 0, 0, 0, 130, 41, 0, 144, 134, 41, 0, 144, 138, 41, 0, 144, 1, 0, 0, 0, 142, 41, 0, 144, 2, 0, 0, 0, 144, 41, 0, 144, 148, 41, 0, 144, 1, 0, 0, 0, 152, 41, 0, 144, 1, 0, 0, 0, 154, 41, 0, 144, 1, 0, 0, 0, 156, 41, 0, 144, 3, 0, 0, 0, 158, 41, 0, 144, 162, 41, 0, 144, 166, 41, 0, 144, 3, 0, 0, 0, 170, 41, 0, 144, 174, 41, 0, 144, 178, 41, 0, 144, 1, 0, 0, 0, 182, 41, 0, 144, 1, 0, 0, 0, 184, 41, 0, 144, 1, 0, 0, 0, 186, 41, 0, 144, 1, 0, 0, 0, 188, 41, 0, 144, 1, 0, 0, 0, 190, 41, 0, 144, 1, 0, 0, 0, 192, 41, 0, 144, 1, 0, 0, 0, 194, 41, 0, 144, 2, 0, 0, 0, 196, 41, 0, 144, 200, 41, 0, 144, 3, 0, 0, 0, 204, 41, 0, 144, 208, 41, 0, 144, 212, 41, 0, 144, 3, 0, 0, 0, 216, 41, 0, 144, 220, 41, 0, 144, 224, 41, 0, 144, 4, 0, 0, 0, 228, 41, 0, 144, 232, 41, 0, 144, 236, 41, 0, 144, 240, 41, 0, 144, 2, 0, 0, 0, 244, 41, 0, 144, 248, 41, 0, 144, 1, 0, 0, 0, 252, 41, 0, 144, 3, 0, 0, 0, 254, 41, 0, 144, 2, 42, 0, 144, 6, 42, 0, 144, 3, 0, 0, 0, 10, 42, 0, 144, 14, 42, 0, 144, 18, 42, 0, 144, 1, 0, 0, 0, 22, 42, 0, 144, 2, 0, 0, 0, 24, 42, 0, 144, 28, 42, 0, 144, 3, 0, 0, 0, 32, 42, 0, 144, 36, 42, 0, 144, 40, 42, 0, 144, 1, 0, 0, 0, 44, 42, 0, 144, 2, 0, 0, 0, 46, 42, 0, 144, 50, 42, 0, 144, 5, 0, 0, 0, 54, 42, 0, 144, 58, 42, 0, 144, 62, 42, 0, 144, 66, 42, 0, 144, 70, 42, 0, 144, 2, 0, 0, 0, 74, 42, 0, 144, 78, 42, 0, 144, 1, 0, 0, 0, 82, 42, 0, 144, 1, 0, 0, 0, 84, 42, 0, 144, 1, 0, 0, 0, 86, 42, 0, 144, 1, 0, 0, 0, 88, 42, 0, 144, 3, 0, 0, 0, 90, 42, 0, 144, 94, 42, 0, 144, 98, 42, 0, 144, 1, 0, 0, 0, 102, 42, 0, 144, 1, 0, 0, 0, 104, 42, 0, 144, 7, 0, 0, 0, 106, 42, 0, 144, 110, 42, 0, 144, 114, 42, 0, 144, 118, 42, 0, 144, 122, 42, 0, 144, 126, 42, 0, 144, 130, 42, 0, 144, 1, 0, 0, 0, 134, 42, 0, 144, 1, 0, 0, 0, 136, 42, 0, 144, 1, 0, 0, 0, 138, 42, 0, 144, 1, 0, 0, 0, 140, 42, 0, 144, 2, 0, 0, 0, 142, 42, 0, 144, 146, 42, 0, 144, 3, 0, 0, 0, 150, 42, 0, 144, 154, 42, 0, 144, 158, 42, 0, 144, 2, 0, 0, 0, 162, 42, 0, 144, 166, 42, 0, 144, 3, 0, 0, 0, 170, 42, 0, 144, 174, 42, 0, 144, 178, 42, 0, 144, 4, 0, 0, 0, 182, 42, 0, 144, 186, 42, 0, 144, 190, 42, 0, 144, 194, 42, 0, 144, 1, 0, 0, 0, 198, 42, 0, 144, 2, 0, 0, 0, 200, 42, 0, 144, 204, 42, 0, 144, 1, 0, 0, 0, 208, 42, 0, 144, 1, 0, 0, 0, 210, 42, 0, 144, 1, 0, 0, 0, 212, 42, 0, 144, 1, 0, 0, 0, 214, 42, 0, 144, 1, 0, 0, 0, 216, 42, 0, 144, 1, 0, 0, 0, 218, 42, 0, 144, 1, 0, 0, 0, 220, 42, 0, 144, 3, 0, 0, 0, 222, 42, 0, 144, 226, 42, 0, 144, 230, 42, 0, 144, 1, 0, 0, 0, 234, 42, 0, 144, 2, 0, 0, 0, 236, 42, 0, 144, 240, 42, 0, 144, 2, 0, 0, 0, 244, 42, 0, 144, 248, 42, 0, 144, 2, 0, 0, 0, 252, 42, 0, 144, 0, 43, 0, 144, 2, 0, 0, 0, 4, 43, 0, 144, 8, 43, 0, 144, 1, 0, 0, 0, 12, 43, 0, 144, 4, 0, 0, 0, 14, 43, 0, 144, 18, 43, 0, 144, 22, 43, 0, 144, 26, 43, 0, 144, 2, 0, 0, 0, 30, 43, 0, 144, 34, 43, 0, 144, 1, 0, 0, 0, 38, 43, 0, 144, 1, 0, 0, 0, 40, 43, 0, 144, 1, 0, 0, 0, 42, 43, 0, 144, 1, 0, 0, 0, 44, 43, 0, 144, 1, 0, 0, 0, 46, 43, 0, 144, 1, 0, 0, 0, 48, 43, 0, 144, 1, 0, 0, 0, 52, 43, 0, 144, 3, 0, 0, 0, 54, 43, 0, 144, 58, 43, 0, 144, 62, 43, 0, 144, 1, 0, 0, 0, 66, 43, 0, 144, 1, 0, 0, 0, 68, 43, 0, 144, 5, 0, 0, 0, 70, 43, 0, 144, 74, 43, 0, 144, 78, 43, 0, 144, 82, 43, 0, 144, 86, 43, 0, 144, 3, 0, 0, 0, 90, 43, 0, 144, 94, 43, 0, 144, 98, 43, 0, 144, 1, 0, 0, 0, 102, 43, 0, 144, 2, 0, 0, 0, 104, 43, 0, 144, 108, 43, 0, 144, 1, 0, 0, 0, 112, 43, 0, 144, 4, 0, 0, 0, 114, 43, 0, 144, 118, 43, 0, 144, 122, 43, 0, 144, 126, 43, 0, 144, 1, 0, 0, 0, 130, 43, 0, 144, 2, 0, 0, 0, 132, 43, 0, 144, 136, 43, 0, 144, 3, 0, 0, 0, 140, 43, 0, 144, 144, 43, 0, 144, 148, 43, 0, 144, 1, 0, 0, 0, 152, 43, 0, 144, 1, 0, 0, 0, 154, 43, 0, 144, 1, 0, 0, 0, 156, 43, 0, 144, 1, 0, 0, 0, 158, 43, 0, 144, 1, 0, 0, 0, 160, 43, 0, 144, 1, 0, 0, 0, 162, 43, 0, 144, 2, 0, 0, 0, 164, 43, 0, 144, 168, 43, 0, 144, 1, 0, 0, 0, 172, 43, 0, 144, 1, 0, 0, 0, 174, 43, 0, 144, 1, 0, 0, 0, 176, 43, 0, 144, 1, 0, 0, 0, 178, 43, 0, 144, 1, 0, 0, 0, 180, 43, 0, 144, 1, 0, 0, 0, 182, 43, 0, 144, 1, 0, 0, 0, 184, 43, 0, 144, 1, 0, 0, 0, 186, 43, 0, 144, 1, 0, 0, 0, 188, 43, 0, 144, 2, 0, 0, 0, 190, 43, 0, 144, 194, 43, 0, 144, 1, 0, 0, 0, 198, 43, 0, 144, 7, 0, 0, 0, 200, 43, 0, 144, 204, 43, 0, 144, 208, 43, 0, 144, 212, 43, 0, 144, 216, 43, 0, 144, 220, 43, 0, 144, 224, 43, 0, 144, 1, 0, 0, 0, 228, 43, 0, 144, 1, 0, 0, 0, 230, 43, 0, 144, 2, 0, 0, 0, 232, 43, 0, 144, 236, 43, 0, 144, 1, 0, 0, 0, 240, 43, 0, 144, 1, 0, 0, 0, 242, 43, 0, 144, 2, 0, 0, 0, 244, 43, 0, 144, 248, 43, 0, 144, 1, 0, 0, 0, 252, 43, 0, 144, 1, 0, 0, 0, 254, 43, 0, 144, 1, 0, 0, 0, 0, 44, 0, 144, 1, 0, 0, 0, 2, 44, 0, 144, 1, 0, 0, 0, 4, 44, 0, 144, 1, 0, 0, 0, 6, 44, 0, 144, 2, 0, 0, 0, 8, 44, 0, 144, 12, 44, 0, 144, 1, 0, 0, 0, 16, 44, 0, 144, 3, 0, 0, 0, 18, 44, 0, 144, 22, 44, 0, 144, 26, 44, 0, 144, 3, 0, 0, 0, 30, 44, 0, 144, 34, 44, 0, 144, 38, 44, 0, 144, 3, 0, 0, 0, 42, 44, 0, 144, 46, 44, 0, 144, 50, 44, 0, 144, 2, 0, 0, 0, 54, 44, 0, 144, 58, 44, 0, 144, 1, 0, 0, 0, 62, 44, 0, 144, 1, 0, 0, 0, 66, 44, 0, 144, 2, 0, 0, 0, 68, 44, 0, 144, 72, 44, 0, 144, 2, 0, 0, 0, 76, 44, 0, 144, 80, 44, 0, 144, 1, 0, 0, 0, 84, 44, 0, 144, 1, 0, 0, 0, 86, 44, 0, 144, 2, 0, 0, 0, 88, 44, 0, 144, 92, 44, 0, 144, 1, 0, 0, 0, 96, 44, 0, 144, 1, 0, 0, 0, 98, 44, 0, 144, 3, 0, 0, 0, 100, 44, 0, 144, 104, 44, 0, 144, 108, 44, 0, 144, 2, 0, 0, 0, 112, 44, 0, 144, 116, 44, 0, 144, 3, 0, 0, 0, 120, 44, 0, 144, 124, 44, 0, 144, 128, 44, 0, 144, 1, 0, 0, 0, 132, 44, 0, 144, 2, 0, 0, 0, 134, 44, 0, 144, 138, 44, 0, 144, 3, 0, 0, 0, 142, 44, 0, 144, 146, 44, 0, 144, 150, 44, 0, 144, 1, 0, 0, 0, 154, 44, 0, 144, 2, 0, 0, 0, 156, 44, 0, 144, 160, 44, 0, 144, 1, 0, 0, 0, 164, 44, 0, 144, 2, 0, 0, 0, 168, 44, 0, 144, 172, 44, 0, 144, 2, 0, 0, 0, 176, 44, 0, 144, 180, 44, 0, 144, 1, 0, 0, 0, 184, 44, 0, 144, 1, 0, 0, 0, 186, 44, 0, 144, 1, 0, 0, 0, 188, 44, 0, 144, 1, 0, 0, 0, 190, 44, 0, 144, 1, 0, 0, 0, 192, 44, 0, 144, 1, 0, 0, 0, 194, 44, 0, 144, 1, 0, 0, 0, 196, 44, 0, 144, 1, 0, 0, 0, 198, 44, 0, 144, 2, 0, 0, 0, 200, 44, 0, 144, 204, 44, 0, 144, 1, 0, 0, 0, 208, 44, 0, 144, 3, 0, 0, 0, 210, 44, 0, 144, 214, 44, 0, 144, 218, 44, 0, 144, 1, 0, 0, 0, 222, 44, 0, 144, 2, 0, 0, 0, 224, 44, 0, 144, 228, 44, 0, 144, 2, 0, 0, 0, 232, 44, 0, 144, 236, 44, 0, 144, 1, 0, 0, 0, 240, 44, 0, 144, 2, 0, 0, 0, 242, 44, 0, 144, 246, 44, 0, 144, 1, 0, 0, 0, 250, 44, 0, 144, 1, 0, 0, 0, 252, 44, 0, 144, 1, 0, 0, 0, 254, 44, 0, 144, 1, 0, 0, 0, 0, 45, 0, 144, 1, 0, 0, 0, 2, 45, 0, 144, 3, 0, 0, 0, 4, 45, 0, 144, 8, 45, 0, 144, 12, 45, 0, 144, 1, 0, 0, 0, 16, 45, 0, 144, 1, 0, 0, 0, 18, 45, 0, 144, 3, 0, 0, 0, 20, 45, 0, 144, 24, 45, 0, 144, 28, 45, 0, 144, 1, 0, 0, 0, 32, 45, 0, 144, 2, 0, 0, 0, 34, 45, 0, 144, 38, 45, 0, 144, 3, 0, 0, 0, 42, 45, 0, 144, 46, 45, 0, 144, 50, 45, 0, 144, 2, 0, 0, 0, 54, 45, 0, 144, 58, 45, 0, 144, 1, 0, 0, 0, 62, 45, 0, 144, 4, 0, 0, 0, 66, 45, 0, 144, 70, 45, 0, 144, 74, 45, 0, 144, 78, 45, 0, 144, 1, 0, 0, 0, 82, 45, 0, 144, 1, 0, 0, 0, 84, 45, 0, 144, 1, 0, 0, 0, 86, 45, 0, 144, 1, 0, 0, 0, 90, 45, 0, 144, 4, 0, 0, 0, 92, 45, 0, 144, 96, 45, 0, 144, 100, 45, 0, 144, 104, 45, 0, 144, 1, 0, 0, 0, 108, 45, 0, 144, 1, 0, 0, 0, 110, 45, 0, 144, 1, 0, 0, 0, 112, 45, 0, 144, 1, 0, 0, 0, 114, 45, 0, 144, 1, 0, 0, 0, 118, 45, 0, 144, 2, 0, 0, 0, 120, 45, 0, 144, 124, 45, 0, 144, 2, 0, 0, 0, 128, 45, 0, 144, 132, 45, 0, 144, 1, 0, 0, 0, 136, 45, 0, 144, 2, 0, 0, 0, 138, 45, 0, 144, 142, 45, 0, 144, 1, 0, 0, 0, 146, 45, 0, 144, 1, 0, 0, 0, 148, 45, 0, 144, 1, 0, 0, 0, 150, 45, 0, 144, 3, 0, 0, 0, 152, 45, 0, 144, 156, 45, 0, 144, 160, 45, 0, 144, 1, 0, 0, 0, 164, 45, 0, 144, 1, 0, 0, 0, 166, 45, 0, 144, 1, 0, 0, 0, 168, 45, 0, 144, 2, 0, 0, 0, 170, 45, 0, 144, 174, 45, 0, 144, 2, 0, 0, 0, 178, 45, 0, 144, 182, 45, 0, 144, 3, 0, 0, 0, 186, 45, 0, 144, 190, 45, 0, 144, 194, 45, 0, 144, 1, 0, 0, 0, 198, 45, 0, 144, 1, 0, 0, 0, 200, 45, 0, 144, 1, 0, 0, 0, 202, 45, 0, 144, 1, 0, 0, 0, 204, 45, 0, 144, 1, 0, 0, 0, 206, 45, 0, 144, 1, 0, 0, 0, 208, 45, 0, 144, 1, 0, 0, 0, 210, 45, 0, 144, 1, 0, 0, 0, 212, 45, 0, 144, 1, 0, 0, 0, 214, 45, 0, 144, 1, 0, 0, 0, 216, 45, 0, 144, 1, 0, 0, 0, 218, 45, 0, 144, 1, 0, 0, 0, 220, 45, 0, 144, 1, 0, 0, 0, 222, 45, 0, 144, 1, 0, 0, 0, 224, 45, 0, 144, 1, 0, 0, 0, 226, 45, 0, 144, 1, 0, 0, 0, 228, 45, 0, 144, 1, 0, 0, 0, 230, 45, 0, 144, 2, 0, 0, 0, 232, 45, 0, 144, 236, 45, 0, 144, 1, 0, 0, 0, 240, 45, 0, 144, 1, 0, 0, 0, 242, 45, 0, 144, 1, 0, 0, 0, 244, 45, 0, 144, 1, 0, 0, 0, 246, 45, 0, 144, 1, 0, 0, 0, 248, 45, 0, 144, 1, 0, 0, 0, 250, 45, 0, 144, 1, 0, 0, 0, 252, 45, 0, 144, 1, 0, 0, 0, 254, 45, 0, 144, 1, 0, 0, 0, 0, 46, 0, 144, 1, 0, 0, 0, 4, 46, 0, 144, 1, 0, 0, 0, 6, 46, 0, 144, 1, 0, 0, 0, 8, 46, 0, 144, 1, 0, 0, 0, 10, 46, 0, 144, 1, 0, 0, 0, 12, 46, 0, 144, 1, 0, 0, 0, 14, 46, 0, 144, 1, 0, 0, 0, 16, 46, 0, 144, 1, 0, 0, 0, 18, 46, 0, 144, 2, 0, 0, 0, 20, 46, 0, 144, 24, 46, 0, 144, 1, 0, 0, 0, 28, 46, 0, 144, 2, 0, 0, 0, 30, 46, 0, 144, 34, 46, 0, 144, 1, 0, 0, 0, 38, 46, 0, 144, 2, 0, 0, 0, 40, 46, 0, 144, 44, 46, 0, 144, 1, 0, 0, 0, 48, 46, 0, 144, 2, 0, 0, 0, 50, 46, 0, 144, 54, 46, 0, 144, 4, 0, 0, 0, 58, 46, 0, 144, 62, 46, 0, 144, 66, 46, 0, 144, 70, 46, 0, 144, 2, 0, 0, 0, 74, 46, 0, 144, 78, 46, 0, 144, 4, 0, 0, 0, 82, 46, 0, 144, 86, 46, 0, 144, 90, 46, 0, 144, 94, 46, 0, 144, 2, 0, 0, 0, 98, 46, 0, 144, 102, 46, 0, 144, 4, 0, 0, 0, 106, 46, 0, 144, 110, 46, 0, 144, 114, 46, 0, 144, 118, 46, 0, 144, 1, 0, 0, 0, 122, 46, 0, 144, 2, 0, 0, 0, 124, 46, 0, 144, 128, 46, 0, 144, 1, 0, 0, 0, 132, 46, 0, 144, 1, 0, 0, 0, 134, 46, 0, 144, 1, 0, 0, 0, 136, 46, 0, 144, 1, 0, 0, 0, 138, 46, 0, 144, 4, 0, 0, 0, 140, 46, 0, 144, 144, 46, 0, 144, 148, 46, 0, 144, 152, 46, 0, 144, 1, 0, 0, 0, 156, 46, 0, 144, 3, 0, 0, 0, 158, 46, 0, 144, 162, 46, 0, 144, 166, 46, 0, 144, 1, 0, 0, 0, 170, 46, 0, 144, 1, 0, 0, 0, 172, 46, 0, 144, 1, 0, 0, 0, 174, 46, 0, 144, 1, 0, 0, 0, 176, 46, 0, 144, 2, 0, 0, 0, 178, 46, 0, 144, 182, 46, 0, 144, 1, 0, 0, 0, 186, 46, 0, 144, 4, 0, 0, 0, 188, 46, 0, 144, 192, 46, 0, 144, 196, 46, 0, 144, 200, 46, 0, 144, 1, 0, 0, 0, 204, 46, 0, 144, 1, 0, 0, 0, 206, 46, 0, 144, 1, 0, 0, 0, 208, 46, 0, 144, 5, 0, 0, 0, 212, 46, 0, 144, 216, 46, 0, 144, 220, 46, 0, 144, 224, 46, 0, 144, 228, 46, 0, 144, 1, 0, 0, 0, 232, 46, 0, 144, 1, 0, 0, 0, 234, 46, 0, 144, 1, 0, 0, 0, 236, 46, 0, 144, 2, 0, 0, 0, 238, 46, 0, 144, 242, 46, 0, 144, 1, 0, 0, 0, 246, 46, 0, 144, 1, 0, 0, 0, 248, 46, 0, 144, 3, 0, 0, 0, 252, 46, 0, 144, 0, 47, 0, 144, 4, 47, 0, 144, 1, 0, 0, 0, 8, 47, 0, 144, 1, 0, 0, 0, 10, 47, 0, 144, 1, 0, 0, 0, 12, 47, 0, 144, 4, 0, 0, 0, 14, 47, 0, 144, 18, 47, 0, 144, 22, 47, 0, 144, 26, 47, 0, 144, 3, 0, 0, 0, 30, 47, 0, 144, 34, 47, 0, 144, 38, 47, 0, 144, 1, 0, 0, 0, 42, 47, 0, 144, 1, 0, 0, 0, 44, 47, 0, 144, 2, 0, 0, 0, 46, 47, 0, 144, 50, 47, 0, 144, 1, 0, 0, 0, 54, 47, 0, 144, 1, 0, 0, 0, 56, 47, 0, 144, 1, 0, 0, 0, 58, 47, 0, 144, 1, 0, 0, 0, 60, 47, 0, 144, 1, 0, 0, 0, 62, 47, 0, 144, 1, 0, 0, 0, 64, 47, 0, 144, 1, 0, 0, 0, 66, 47, 0, 144, 1, 0, 0, 0, 68, 47, 0, 144, 1, 0, 0, 0, 70, 47, 0, 144, 1, 0, 0, 0, 72, 47, 0, 144, 1, 0, 0, 0, 74, 47, 0, 144, 1, 0, 0, 0, 76, 47, 0, 144, 1, 0, 0, 0, 78, 47, 0, 144, 1, 0, 0, 0, 80, 47, 0, 144, 1, 0, 0, 0, 82, 47, 0, 144, 1, 0, 0, 0, 84, 47, 0, 144, 1, 0, 0, 0, 86, 47, 0, 144, 1, 0, 0, 0, 90, 47, 0, 144, 1, 0, 0, 0, 92, 47, 0, 144, 1, 0, 0, 0, 94, 47, 0, 144, 1, 0, 0, 0, 96, 47, 0, 144, 1, 0, 0, 0, 98, 47, 0, 144, 1, 0, 0, 0, 100, 47, 0, 144, 1, 0, 0, 0, 102, 47, 0, 144, 2, 0, 0, 0, 104, 47, 0, 144, 108, 47, 0, 144, 1, 0, 0, 0, 112, 47, 0, 144, 1, 0, 0, 0, 114, 47, 0, 144, 2, 0, 0, 0, 116, 47, 0, 144, 120, 47, 0, 144, 2, 0, 0, 0, 124, 47, 0, 144, 128, 47, 0, 144, 2, 0, 0, 0, 132, 47, 0, 144, 136, 47, 0, 144, 1, 0, 0, 0, 140, 47, 0, 144, 1, 0, 0, 0, 142, 47, 0, 144, 2, 0, 0, 0, 144, 47, 0, 144, 148, 47, 0, 144, 1, 0, 0, 0, 152, 47, 0, 144, 1, 0, 0, 0, 154, 47, 0, 144, 2, 0, 0, 0, 156, 47, 0, 144, 160, 47, 0, 144, 1, 0, 0, 0, 164, 47, 0, 144, 1, 0, 0, 0, 166, 47, 0, 144, 1, 0, 0, 0, 168, 47, 0, 144, 1, 0, 0, 0, 170, 47, 0, 144, 2, 0, 0, 0, 172, 47, 0, 144, 176, 47, 0, 144, 1, 0, 0, 0, 180, 47, 0, 144, 1, 0, 0, 0, 182, 47, 0, 144, 1, 0, 0, 0, 184, 47, 0, 144, 1, 0, 0, 0, 186, 47, 0, 144, 1, 0, 0, 0, 188, 47, 0, 144, 2, 0, 0, 0, 190, 47, 0, 144, 194, 47, 0, 144, 1, 0, 0, 0, 198, 47, 0, 144, 1, 0, 0, 0, 200, 47, 0, 144, 1, 0, 0, 0, 202, 47, 0, 144, 1, 0, 0, 0, 204, 47, 0, 144, 1, 0, 0, 0, 206, 47, 0, 144, 1, 0, 0, 0, 208, 47, 0, 144, 173, 1, 198, 6, 213, 6, 226, 6, 245, 6, 4, 7, 18, 7, 32, 7, 46, 7, 60, 7, 74, 7, 90, 7, 109, 7, 126, 7, 139, 7, 157, 7, 170, 7, 185, 7, 198, 7, 219, 7, 235, 7, 249, 7, 9, 8, 25, 8, 41, 8, 55, 8, 75, 8, 87, 8, 102, 8, 118, 8, 132, 8, 148, 8, 161, 8, 179, 8, 191, 8, 205, 8, 223, 8, 237, 8, 251, 8, 9, 9, 23, 9, 38, 9, 55, 9, 71, 9, 86, 9, 102, 9, 116, 9, 134, 9, 153, 9, 171, 9, 187, 9, 202, 9, 215, 9, 231, 9, 244, 9, 6, 10, 23, 10, 39, 10, 57, 10, 84, 10, 115, 10, 139, 10, 166, 10, 193, 10, 219, 10, 245, 10, 3, 11, 20, 11, 34, 11, 57, 11, 74, 11, 88, 11, 103, 11, 124, 11, 142, 11, 157, 11, 171, 11, 192, 11, 214, 11, 235, 11, 250, 11, 10, 12, 28, 12, 44, 12, 59, 12, 75, 12, 93, 12, 115, 12, 131, 12, 150, 12, 166, 12, 181, 12, 197, 12, 218, 12, 233, 12, 254, 12, 13, 13, 29, 13, 46, 13, 63, 13, 80, 13, 100, 13, 120, 13, 138, 13, 156, 13, 172, 13, 190, 13, 209, 13, 225, 13, 244, 13, 6, 14, 24, 14, 39, 14, 55, 14, 70, 14, 89, 14, 110, 14, 134, 14, 161, 14, 187, 14, 209, 14, 235, 14, 3, 15, 24, 15, 39, 15, 55, 15, 71, 15, 85, 15, 100, 15, 128, 15, 147, 15, 162, 15, 175, 15, 195, 15, 214, 15, 236, 15, 251, 15, 11, 16, 26, 16, 42, 16, 61, 16, 79, 16, 96, 16, 112, 16, 130, 16, 145, 16, 164, 16, 184, 16, 201, 16, 217, 16, 235, 16, 254, 16, 17, 17, 32, 17, 49, 17, 65, 17, 78, 17, 94, 17, 122, 17, 150, 17, 181, 17, 197, 17, 212, 17, 232, 17, 251, 17, 11, 18, 34, 18, 56, 18, 76, 18, 96, 18, 117, 18, 137, 18, 158, 18, 173, 18, 188, 18, 205, 18, 224, 18, 245, 18, 6, 19, 23, 19, 45, 19, 63, 19, 84, 19, 98, 19, 120, 19, 137, 19, 154, 19, 171, 19, 189, 19, 208, 19, 230, 19, 250, 19, 8, 20, 28, 20, 44, 20, 60, 20, 76, 20, 94, 20, 113, 20, 130, 20, 146, 20, 166, 20, 183, 20, 200, 20, 226, 20, 247, 20, 9, 21, 28, 21, 46, 21, 65, 21, 82, 21, 99, 21, 117, 21, 137, 21, 147, 21, 159, 21, 170, 21, 182, 21, 193, 21, 205, 21, 219, 21, 231, 21, 244, 21, 1, 22, 11, 22, 24, 22, 36, 22, 49, 22, 61, 22, 75, 22, 86, 22, 102, 22, 115, 22, 129, 22, 140, 22, 150, 22, 161, 22, 175, 22, 190, 22, 200, 22, 212, 22, 227, 22, 237, 22, 250, 22, 7, 23, 21, 23, 36, 23, 47, 23, 62, 23, 75, 23, 89, 23, 103, 23, 120, 23, 138, 23, 151, 23, 163, 23, 174, 23, 187, 23, 201, 23, 213, 23, 225, 23, 238, 23, 0, 24, 17, 24, 27, 24, 37, 24, 53, 24, 68, 24, 83, 24, 94, 24, 108, 24, 123, 24, 136, 24, 148, 24, 160, 24, 174, 24, 189, 24, 200, 24, 214, 24, 229, 24, 248, 24, 4, 25, 18, 25, 31, 25, 43, 25, 56, 25, 67, 25, 84, 25, 96, 25, 110, 25, 125, 25, 142, 25, 155, 25, 174, 25, 187, 25, 203, 25, 220, 25, 236, 25, 0, 26, 16, 26, 33, 26, 52, 26, 75, 26, 94, 26, 111, 26, 130, 26, 149, 26, 171, 26, 188, 26, 205, 26, 221, 26, 238, 26, 1, 27, 21, 27, 41, 27, 57, 27, 74, 27, 82, 27, 90, 27, 98, 27, 115, 27, 130, 27, 147, 27, 161, 27, 177, 27, 191, 27, 209, 27, 225, 27, 242, 27, 2, 28, 18, 28, 34, 28, 52, 28, 66, 28, 83, 28, 99, 28, 115, 28, 134, 28, 150, 28, 164, 28, 183, 28, 195, 28, 209, 28, 226, 28, 240, 28, 2, 29, 16, 29, 29, 29, 46, 29, 59, 29, 73, 29, 87, 29, 99, 29, 112, 29, 129, 29, 143, 29, 155, 29, 167, 29, 181, 29, 199, 29, 215, 29, 230, 29, 248, 29, 6, 30, 19, 30, 36, 30, 51, 30, 65, 30, 82, 30, 98, 30, 111, 30, 126, 30, 140, 30, 155, 30, 172, 30, 186, 30, 200, 30, 218, 30, 232, 30, 252, 30, 10, 31, 27, 31, 40, 31, 54, 31, 71, 31, 83, 31, 99, 31, 116, 31, 131, 31, 146, 31, 154, 31, 162, 31, 175, 31, 192, 31, 213, 31, 229, 31, 244, 31, 2, 32, 20, 32, 36, 32, 49, 32, 66, 32, 84, 32, 100, 32, 120, 32, 133, 32, 150, 32, 167, 32, 186, 32, 201, 32, 219, 32, 234, 32, 252, 32, 11, 33, 25, 33, 38, 33, 54, 33, 69, 33, 87, 33, 101, 33, 118, 33, 133, 33, 154, 33, 172, 33, 187, 33, 202, 33, 217, 33, 235, 33, 248, 33, 5, 34, 163, 33, 0, 128, 165, 33, 0, 128, 167, 33, 0, 128, 169, 33, 0, 128, 174, 33, 0, 128, 176, 33, 0, 128, 178, 33, 0, 128, 180, 33, 0, 128, 182, 33, 0, 128, 185, 33, 0, 128, 187, 33, 0, 128, 189, 33, 0, 128, 191, 33, 0, 128, 193, 33, 0, 128, 197, 33, 0, 128, 200, 33, 0, 128, 202, 33, 0, 128, 204, 33, 0, 128, 206, 33, 0, 128, 208, 33, 0, 128, 210, 33, 0, 128, 213, 33, 0, 128, 215, 33, 0, 128, 217, 33, 0, 128, 219, 33, 0, 128, 221, 33, 0, 128, 224, 33, 0, 128, 226, 33, 0, 128, 230, 33, 0, 128, 232, 33, 0, 128, 234, 33, 0, 128, 236, 33, 0, 128, 238, 33, 0, 128, 240, 33, 0, 128, 242, 33, 0, 128, 244, 33, 0, 128, 246, 33, 0, 128, 248, 33, 0, 128, 250, 33, 0, 128, 252, 33, 0, 128, 254, 33, 0, 128, 0, 34, 0, 128, 3, 34, 0, 128, 5, 34, 0, 128, 7, 34, 0, 128, 9, 34, 0, 128, 11, 34, 0, 128, 13, 34, 0, 128, 15, 34, 0, 128, 19, 34, 0, 128, 27, 34, 0, 128, 29, 34, 0, 128, 34, 34, 0, 128, 37, 34, 0, 128, 41, 34, 0, 128, 43, 34, 0, 128, 45, 34, 0, 128, 47, 34, 0, 128, 53, 34, 0, 128, 57, 34, 0, 128, 60, 34, 0, 128, 66, 34, 0, 128, 75, 34, 0, 128, 80, 34, 0, 128, 83, 34, 0, 128, 85, 34, 0, 128, 87, 34, 0, 128, 89, 34, 0, 128, 93, 34, 0, 128, 95, 34, 0, 128, 97, 34, 0, 128, 99, 34, 0, 128, 101, 34, 0, 128, 103, 34, 0, 128, 105, 34, 0, 128, 107, 34, 0, 128, 109, 34, 0, 128, 115, 34, 0, 128, 117, 34, 0, 128, 122, 34, 0, 128, 124, 34, 0, 128, 129, 34, 0, 128, 131, 34, 0, 128, 133, 34, 0, 128, 135, 34, 0, 128, 138, 34, 0, 128, 140, 34, 0, 128, 143, 34, 0, 128, 145, 34, 0, 128, 147, 34, 0, 128, 149, 34, 0, 128, 151, 34, 0, 128, 154, 34, 0, 128, 157, 34, 0, 128, 160, 34, 0, 128, 162, 34, 0, 128, 164, 34, 0, 128, 166, 34, 0, 128, 168, 34, 0, 128, 172, 34, 0, 128, 174, 34, 0, 128, 177, 34, 0, 128, 179, 34, 0, 128, 181, 34, 0, 128, 183, 34, 0, 128, 187, 34, 0, 128, 191, 34, 0, 128, 193, 34, 0, 128, 195, 34, 0, 128, 197, 34, 0, 128, 199, 34, 0, 128, 201, 34, 0, 128, 203, 34, 0, 128, 205, 34, 0, 128, 208, 34, 0, 128, 212, 34, 0, 128, 216, 34, 0, 128, 221, 34, 0, 128, 224, 34, 0, 128, 226, 34, 0, 128, 230, 34, 0, 128, 234, 34, 0, 128, 236, 34, 0, 128, 239, 34, 0, 128, 243, 34, 0, 128, 245, 34, 0, 128, 248, 34, 0, 128, 254, 34, 0, 128, 1, 35, 0, 128, 3, 35, 0, 128, 5, 35, 0, 128, 7, 35, 0, 128, 9, 35, 0, 128, 13, 35, 0, 128, 15, 35, 0, 128, 17, 35, 0, 128, 25, 35, 0, 128, 27, 35, 0, 128, 29, 35, 0, 128, 31, 35, 0, 128, 33, 35, 0, 128, 36, 35, 0, 128, 40, 35, 0, 128, 43, 35, 0, 128, 47, 35, 0, 128, 52, 35, 0, 128, 54, 35, 0, 128, 57, 35, 0, 128, 59, 35, 0, 128, 61, 35, 0, 128, 63, 35, 0, 128, 65, 35, 0, 128, 67, 35, 0, 128, 69, 35, 0, 128, 71, 35, 0, 128, 75, 35, 0, 128, 77, 35, 0, 128, 80, 35, 0, 128, 83, 35, 0, 128, 86, 35, 0, 128, 89, 35, 0, 128, 91, 35, 0, 128, 96, 35, 0, 128, 99, 35, 0, 128, 101, 35, 0, 128, 103, 35, 0, 128, 105, 35, 0, 128, 107, 35, 0, 128, 109, 35, 0, 128, 111, 35, 0, 128, 113, 35, 0, 128, 117, 35, 0, 128, 119, 35, 0, 128, 121, 35, 0, 128, 127, 35, 0, 128, 131, 35, 0, 128, 133, 35, 0, 128, 136, 35, 0, 128, 138, 35, 0, 128, 143, 35, 0, 128, 145, 35, 0, 128, 148, 35, 0, 128, 152, 35, 0, 128, 154, 35, 0, 128, 156, 35, 0, 128, 158, 35, 0, 128, 160, 35, 0, 128, 162, 35, 0, 128, 164, 35, 0, 128, 167, 35, 0, 128, 169, 35, 0, 128, 171, 35, 0, 128, 173, 35, 0, 128, 175, 35, 0, 128, 177, 35, 0, 128, 179, 35, 0, 128, 181, 35, 0, 128, 183, 35, 0, 128, 185, 35, 0, 128, 188, 35, 0, 128, 190, 35, 0, 128, 198, 35, 0, 128, 200, 35, 0, 128, 202, 35, 0, 128, 205, 35, 0, 128, 207, 35, 0, 128, 209, 35, 0, 128, 212, 35, 0, 128, 214, 35, 0, 128, 216, 35, 0, 128, 218, 35, 0, 128, 220, 35, 0, 128, 222, 35, 0, 128, 224, 35, 0, 128, 227, 35, 0, 128, 229, 35, 0, 128, 233, 35, 0, 128, 237, 35, 0, 128, 241, 35, 0, 128, 244, 35, 0, 128, 246, 35, 0, 128, 248, 35, 0, 128, 251, 35, 0, 128, 254, 35, 0, 128, 0, 36, 0, 128, 2, 36, 0, 128, 5, 36, 0, 128, 7, 36, 0, 128, 9, 36, 0, 128, 13, 36, 0, 128, 16, 36, 0, 128, 20, 36, 0, 128, 22, 36, 0, 128, 25, 36, 0, 128, 29, 36, 0, 128, 31, 36, 0, 128, 34, 36, 0, 128, 36, 36, 0, 128, 39, 36, 0, 128, 42, 36, 0, 128, 44, 36, 0, 128, 46, 36, 0, 128, 48, 36, 0, 128, 50, 36, 0, 128, 52, 36, 0, 128, 54, 36, 0, 128, 56, 36, 0, 128, 58, 36, 0, 128, 61, 36, 0, 128, 63, 36, 0, 128, 67, 36, 0, 128, 69, 36, 0, 128, 72, 36, 0, 128, 75, 36, 0, 128, 77, 36, 0, 128, 80, 36, 0, 128, 82, 36, 0, 128, 84, 36, 0, 128, 86, 36, 0, 128, 88, 36, 0, 128, 90, 36, 0, 128, 94, 36, 0, 128, 96, 36, 0, 128, 98, 36, 0, 128, 102, 36, 0, 128, 104, 36, 0, 128, 107, 36, 0, 128, 111, 36, 0, 128, 114, 36, 0, 128, 116, 36, 0, 128, 121, 36, 0, 128, 123, 36, 0, 128, 125, 36, 0, 128, 127, 36, 0, 128, 129, 36, 0, 128, 134, 36, 0, 128, 136, 36, 0, 128, 138, 36, 0, 128, 140, 36, 0, 128, 142, 36, 0, 128, 144, 36, 0, 128, 147, 36, 0, 128, 150, 36, 0, 128, 152, 36, 0, 128, 155, 36, 0, 128, 157, 36, 0, 128, 159, 36, 0, 128, 161, 36, 0, 128, 165, 36, 0, 128, 167, 36, 0, 128, 169, 36, 0, 128, 171, 36, 0, 128, 174, 36, 0, 128, 177, 36, 0, 128, 181, 36, 0, 128, 183, 36, 0, 128, 185, 36, 0, 128, 187, 36, 0, 128, 189, 36, 0, 128, 191, 36, 0, 128, 193, 36, 0, 128, 195, 36, 0, 128, 197, 36, 0, 128, 199, 36, 0, 128, 201, 36, 0, 128, 203, 36, 0, 128, 205, 36, 0, 128, 207, 36, 0, 128, 209, 36, 0, 128, 211, 36, 0, 128, 213, 36, 0, 128, 215, 36, 0, 128, 218, 36, 0, 128, 220, 36, 0, 128, 222, 36, 0, 128, 224, 36, 0, 128, 226, 36, 0, 128, 228, 36, 0, 128, 230, 36, 0, 128, 232, 36, 0, 128, 234, 36, 0, 128, 236, 36, 0, 128, 238, 36, 0, 128, 240, 36, 0, 128, 242, 36, 0, 128, 244, 36, 0, 128, 246, 36, 0, 128, 248, 36, 0, 128, 250, 36, 0, 128, 252, 36, 0, 128, 255, 36, 0, 128, 1, 37, 0, 128, 4, 37, 0, 128, 6, 37, 0, 128, 9, 37, 0, 128, 11, 37, 0, 128, 14, 37, 0, 128, 19, 37, 0, 128, 22, 37, 0, 128, 27, 37, 0, 128, 30, 37, 0, 128, 35, 37, 0, 128, 37, 37, 0, 128, 40, 37, 0, 128, 42, 37, 0, 128, 44, 37, 0, 128, 46, 37, 0, 128, 48, 37, 0, 128, 53, 37, 0, 128, 55, 37, 0, 128, 59, 37, 0, 128, 61, 37, 0, 128, 63, 37, 0, 128, 65, 37, 0, 128, 67, 37, 0, 128, 70, 37, 0, 128, 72, 37, 0, 128, 77, 37, 0, 128, 79, 37, 0, 128, 81, 37, 0, 128, 83, 37, 0, 128, 89, 37, 0, 128, 91, 37, 0, 128, 93, 37, 0, 128, 95, 37, 0, 128, 98, 37, 0, 128, 100, 37, 0, 128, 102, 37, 0, 128, 106, 37, 0, 128, 108, 37, 0, 128, 110, 37, 0, 128, 112, 37, 0, 128, 117, 37, 0, 128, 121, 37, 0, 128, 123, 37, 0, 128, 125, 37, 0, 128, 128, 37, 0, 128, 130, 37, 0, 128, 132, 37, 0, 128, 134, 37, 0, 128, 136, 37, 0, 128, 138, 37, 0, 128, 140, 37, 0, 128, 142, 37, 0, 128, 144, 37, 0, 128, 146, 37, 0, 128, 148, 37, 0, 128, 150, 37, 0, 128, 152, 37, 0, 128, 154, 37, 0, 128, 156, 37, 0, 128, 158, 37, 0, 128, 160, 37, 0, 128, 162, 37, 0, 128, 164, 37, 0, 128, 166, 37, 0, 128, 168, 37, 0, 128, 170, 37, 0, 128, 172, 37, 0, 128, 174, 37, 0, 128, 176, 37, 0, 128, 179, 37, 0, 128, 181, 37, 0, 128, 183, 37, 0, 128, 186, 37, 0, 128, 189, 37, 0, 128, 192, 37, 0, 128, 194, 37, 0, 128, 196, 37, 0, 128, 199, 37, 0, 128, 201, 37, 0, 128, 203, 37, 0, 128, 206, 37, 0, 128, 208, 37, 0, 128, 210, 37, 0, 128, 212, 37, 0, 128, 214, 37, 0, 128, 217, 37, 0, 128, 219, 37, 0, 128, 221, 37, 0, 128, 223, 37, 0, 128, 225, 37, 0, 128, 227, 37, 0, 128, 230, 37, 0, 128, 232, 37, 0, 128, 234, 37, 0, 128, 236, 37, 0, 128, 238, 37, 0, 128, 240, 37, 0, 128, 3, 0, 32, 0, 185, 6, 20, 34, 132, 32, 0, 32, 242, 37, 0, 32, 210, 47, 0, 80, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 159, 33, 0, 32, 7, 0, 0, 0, 40, 5, 0, 0, 163, 33, 0, 0, 163, 33, 0, 0, 128, 2, 0, 0, 1, 0, 0, 0, 121, 26, 0, 0, 97, 108, 105, 97, 115, 0, 108, 97, 110, 103, 117, 97, 103, 101, 0, 97, 97, 95, 83, 65, 65, 72, 79, 0, 114, 101, 97, 115, 111, 110, 0, 114, 101, 112, 108, 97, 99, 101, 109, 101, 110, 116, 0, 97, 97, 109, 0, 97, 97, 114, 0, 97, 98, 107, 0, 97, 100, 112, 0, 97, 102, 114, 0, 97, 106, 117, 0, 97, 107, 97, 0, 97, 108, 98, 0, 97, 108, 115, 0, 97, 109, 104, 0, 97, 114, 97, 0, 97, 114, 98, 0, 97, 114, 103, 0, 97, 114, 109, 0, 97, 114, 116, 95, 108, 111, 106, 98, 97, 110, 0, 97, 115, 100, 0, 97, 115, 109, 0, 97, 117, 101, 0, 97, 118, 97, 0, 97, 118, 101, 0, 97, 121, 109, 0, 97, 121, 114, 0, 97, 121, 120, 0, 97, 122, 95, 65, 90, 0, 97, 122, 101, 0, 97, 122, 106, 0, 98, 97, 107, 0, 98, 97, 109, 0, 98, 97, 113, 0, 98, 99, 99, 0, 98, 99, 108, 0, 98, 101, 108, 0, 98, 101, 110, 0, 98, 103, 109, 0, 98, 104, 0, 98, 105, 104, 0, 98, 105, 115, 0, 98, 106, 100, 0, 98, 111, 100, 0, 98, 111, 115, 0, 98, 114, 101, 0, 98, 115, 95, 66, 65, 0, 98, 117, 108, 0, 98, 117, 114, 0, 98, 120, 107, 0, 98, 120, 114, 0, 99, 97, 116, 0, 99, 99, 113, 0, 99, 101, 108, 45, 103, 97, 117, 108, 105, 115, 104, 0, 99, 101, 115, 0, 99, 104, 97, 0, 99, 104, 101, 0, 99, 104, 105, 0, 99, 104, 117, 0, 99, 104, 118, 0, 99, 106, 114, 0, 99, 107, 97, 0, 99, 108, 100, 0, 99, 109, 107, 0, 99, 109, 110, 0, 99, 110, 114, 0, 99, 111, 114, 0, 99, 111, 115, 0, 99, 111, 121, 0, 99, 113, 117, 0, 99, 114, 101, 0, 99, 119, 100, 0, 99, 121, 109, 0, 99, 122, 101, 0, 100, 97, 110, 0, 100, 101, 117, 0, 100, 103, 111, 0, 100, 104, 100, 0, 100, 105, 107, 0, 100, 105, 113, 0, 100, 105, 116, 0, 100, 105, 118, 0, 100, 114, 104, 0, 100, 114, 119, 0, 100, 117, 116, 0, 100, 122, 111, 0, 101, 107, 107, 0, 101, 108, 108, 0, 101, 109, 107, 0, 101, 110, 103, 0, 101, 112, 111, 0, 101, 115, 107, 0, 101, 115, 116, 0, 101, 117, 115, 0, 101, 119, 101, 0, 102, 97, 111, 0, 102, 97, 115, 0, 102, 97, 116, 0, 102, 105, 106, 0, 102, 105, 110, 0, 102, 114, 97, 0, 102, 114, 101, 0, 102, 114, 121, 0, 102, 117, 99, 0, 102, 117, 108, 0, 103, 97, 118, 0, 103, 97, 122, 0, 103, 98, 111, 0, 103, 101, 111, 0, 103, 101, 114, 0, 103, 102, 120, 0, 103, 103, 110, 0, 103, 108, 97, 0, 103, 108, 101, 0, 103, 108, 103, 0, 103, 108, 118, 0, 103, 110, 111, 0, 103, 114, 101, 0, 103, 114, 110, 0, 103, 116, 105, 0, 103, 117, 103, 0, 103, 117, 106, 0, 103, 117, 118, 0, 103, 121, 97, 0, 104, 97, 95, 76, 97, 116, 110, 95, 71, 72, 0, 104, 97, 95, 76, 97, 116, 110, 95, 78, 69, 0, 104, 97, 95, 76, 97, 116, 110, 95, 78, 71, 0, 104, 97, 116, 0, 104, 97, 117, 0, 104, 98, 115, 0, 104, 100, 110, 0, 104, 101, 97, 0, 104, 101, 98, 0, 104, 101, 114, 0, 104, 105, 109, 0, 104, 105, 110, 0, 104, 109, 111, 0, 104, 114, 114, 0, 104, 114, 118, 0, 104, 117, 110, 0, 104, 121, 101, 0, 105, 45, 100, 101, 102, 97, 117, 108, 116, 0, 105, 45, 101, 110, 111, 99, 104, 105, 97, 110, 0, 105, 45, 109, 105, 110, 103, 111, 0, 105, 95, 97, 109, 105, 0, 105, 95, 98, 110, 110, 0, 105, 95, 104, 97, 107, 0, 105, 95, 107, 108, 105, 110, 103, 111, 110, 0, 105, 95, 108, 117, 120, 0, 105, 95, 110, 97, 118, 97, 106, 111, 0, 105, 95, 112, 119, 110, 0, 105, 95, 116, 97, 111, 0, 105, 95, 116, 97, 121, 0, 105, 95, 116, 115, 117, 0, 105, 98, 105, 0, 105, 98, 111, 0, 105, 99, 101, 0, 105, 100, 111, 0, 105, 105, 105, 0, 105, 107, 101, 0, 105, 107, 117, 0, 105, 108, 101, 0, 105, 108, 119, 0, 105, 110, 97, 0, 105, 110, 100, 0, 105, 112, 107, 0, 105, 115, 108, 0, 105, 116, 97, 0, 105, 119, 0, 106, 97, 118, 0, 106, 101, 103, 0, 106, 112, 110, 0, 106, 119, 0, 107, 97, 108, 0, 107, 97, 110, 0, 107, 97, 115, 0, 107, 97, 116, 0, 107, 97, 117, 0, 107, 97, 122, 0, 107, 103, 99, 0, 107, 103, 104, 0, 107, 104, 107, 0, 107, 104, 109, 0, 107, 105, 107, 0, 107, 105, 110, 0, 107, 105, 114, 0, 107, 107, 95, 67, 121, 114, 108, 95, 75, 90, 0, 107, 109, 114, 0, 107, 110, 99, 0, 107, 110, 103, 0, 107, 110, 110, 0, 107, 111, 106, 0, 107, 111, 109, 0, 107, 111, 110, 0, 107, 111, 114, 0, 107, 112, 118, 0, 107, 114, 109, 0, 107, 115, 95, 65, 114, 97, 98, 95, 73, 78, 0, 107, 116, 114, 0, 107, 117, 97, 0, 107, 117, 114, 0, 107, 118, 115, 0, 107, 119, 113, 0, 107, 120, 101, 0, 107, 121, 95, 67, 121, 114, 108, 95, 75, 71, 0, 107, 122, 106, 0, 107, 122, 116, 0, 108, 97, 111, 0, 108, 97, 116, 0, 108, 97, 118, 0, 108, 98, 107, 0, 108, 105, 105, 0, 108, 105, 109, 0, 108, 105, 110, 0, 108, 105, 116, 0, 108, 108, 111, 0, 108, 109, 109, 0, 108, 116, 122, 0, 108, 117, 98, 0, 108, 117, 103, 0, 108, 118, 115, 0, 109, 97, 99, 0, 109, 97, 104, 0, 109, 97, 108, 0, 109, 97, 111, 0, 109, 97, 114, 0, 109, 97, 121, 0, 109, 101, 103, 0, 109, 104, 114, 0, 109, 107, 100, 0, 109, 108, 103, 0, 109, 108, 116, 0, 109, 110, 95, 67, 121, 114, 108, 95, 77, 78, 0, 109, 110, 107, 0, 109, 111, 108, 0, 109, 111, 110, 0, 109, 114, 105, 0, 109, 115, 95, 76, 97, 116, 110, 95, 66, 78, 0, 109, 115, 95, 76, 97, 116, 110, 95, 77, 89, 0, 109, 115, 95, 76, 97, 116, 110, 95, 83, 71, 0, 109, 115, 97, 0, 109, 115, 116, 0, 109, 117, 112, 0, 109, 119, 106, 0, 109, 121, 97, 0, 109, 121, 100, 0, 109, 121, 116, 0, 110, 97, 100, 0, 110, 97, 117, 0, 110, 97, 118, 0, 110, 98, 108, 0, 110, 99, 112, 0, 110, 100, 101, 0, 110, 100, 111, 0, 110, 101, 112, 0, 110, 108, 100, 0, 110, 110, 111, 0, 110, 110, 115, 0, 110, 110, 120, 0, 110, 111, 95, 66, 79, 75, 77, 65, 76, 0, 110, 111, 95, 78, 89, 78, 79, 82, 83, 75, 0, 110, 111, 95, 98, 111, 107, 0, 110, 111, 95, 110, 121, 110, 0, 110, 111, 98, 0, 110, 111, 114, 0, 110, 112, 105, 0, 110, 116, 115, 0, 110, 121, 97, 0, 111, 99, 105, 0, 111, 106, 103, 0, 111, 106, 105, 0, 111, 114, 105, 0, 111, 114, 109, 0, 111, 115, 115, 0, 111, 117, 110, 0, 112, 97, 95, 73, 78, 0, 112, 97, 95, 80, 75, 0, 112, 97, 110, 0, 112, 98, 117, 0, 112, 99, 114, 0, 112, 101, 114, 0, 112, 101, 115, 0, 112, 108, 105, 0, 112, 108, 116, 0, 112, 109, 99, 0, 112, 109, 117, 0, 112, 110, 98, 0, 112, 111, 108, 0, 112, 111, 114, 0, 112, 112, 97, 0, 112, 112, 114, 0, 112, 114, 115, 0, 112, 114, 121, 0, 112, 117, 115, 0, 112, 117, 122, 0, 113, 117, 101, 0, 113, 117, 122, 0, 114, 109, 121, 0, 114, 111, 104, 0, 114, 111, 110, 0, 114, 117, 109, 0, 114, 117, 110, 0, 114, 117, 115, 0, 115, 97, 103, 0, 115, 97, 110, 0, 115, 99, 97, 0, 115, 99, 99, 0, 115, 99, 114, 0, 115, 103, 110, 95, 66, 69, 95, 70, 82, 0, 115, 103, 110, 95, 66, 69, 95, 78, 76, 0, 115, 103, 110, 95, 67, 72, 95, 68, 69, 0, 115, 104, 105, 95, 77, 65, 0, 115, 105, 110, 0, 115, 107, 107, 0, 115, 108, 107, 0, 115, 108, 111, 0, 115, 108, 118, 0, 115, 109, 101, 0, 115, 109, 111, 0, 115, 110, 97, 0, 115, 110, 100, 0, 115, 111, 109, 0, 115, 111, 116, 0, 115, 112, 97, 0, 115, 112, 121, 0, 115, 113, 105, 0, 115, 114, 95, 66, 65, 0, 115, 114, 95, 77, 69, 0, 115, 114, 95, 82, 83, 0, 115, 114, 95, 88, 75, 0, 115, 114, 99, 0, 115, 114, 100, 0, 115, 114, 112, 0, 115, 115, 119, 0, 115, 117, 110, 0, 115, 119, 97, 0, 115, 119, 99, 0, 115, 119, 101, 0, 115, 119, 104, 0, 116, 97, 104, 0, 116, 97, 109, 0, 116, 97, 116, 0, 116, 100, 117, 0, 116, 101, 108, 0, 116, 103, 107, 0, 116, 103, 108, 0, 116, 104, 97, 0, 116, 104, 99, 0, 116, 104, 120, 0, 116, 105, 98, 0, 116, 105, 101, 0, 116, 105, 114, 0, 116, 107, 107, 0, 116, 108, 0, 116, 108, 119, 0, 116, 109, 112, 0, 116, 110, 101, 0, 116, 110, 102, 0, 116, 111, 110, 0, 116, 115, 102, 0, 116, 115, 110, 0, 116, 115, 111, 0, 116, 116, 113, 0, 116, 117, 107, 0, 116, 117, 114, 0, 116, 119, 0, 116, 119, 105, 0, 116, 122, 109, 95, 76, 97, 116, 110, 95, 77, 65, 0, 117, 103, 95, 65, 114, 97, 98, 95, 67, 78, 0, 117, 105, 103, 0, 117, 107, 114, 0, 117, 109, 117, 0, 117, 111, 107, 0, 117, 114, 100, 0, 117, 122, 95, 65, 70, 0, 117, 122, 95, 85, 90, 0, 117, 122, 98, 0, 117, 122, 110, 0, 118, 97, 105, 95, 76, 82, 0, 118, 101, 110, 0, 118, 105, 101, 0, 118, 111, 108, 0, 119, 101, 108, 0, 119, 108, 110, 0, 119, 111, 108, 0, 120, 98, 97, 0, 120, 104, 111, 0, 120, 105, 97, 0, 120, 107, 104, 0, 120, 112, 101, 0, 120, 115, 106, 0, 120, 115, 108, 0, 121, 98, 100, 0, 121, 100, 100, 0, 121, 105, 100, 0, 121, 109, 97, 0, 121, 109, 116, 0, 121, 111, 114, 0, 121, 111, 115, 0, 121, 117, 101, 95, 67, 78, 0, 121, 117, 101, 95, 72, 75, 0, 121, 117, 117, 0, 122, 97, 105, 0, 122, 104, 45, 109, 105, 110, 0, 122, 104, 95, 67, 78, 0, 122, 104, 95, 72, 75, 0, 122, 104, 95, 77, 79, 0, 122, 104, 95, 83, 71, 0, 122, 104, 95, 84, 87, 0, 122, 104, 95, 103, 117, 111, 121, 117, 0, 122, 104, 95, 104, 97, 107, 107, 97, 0, 122, 104, 95, 109, 105, 110, 95, 110, 97, 110, 0, 122, 104, 95, 120, 105, 97, 110, 103, 0, 122, 104, 97, 0, 122, 104, 111, 0, 122, 115, 109, 0, 122, 117, 108, 0, 122, 121, 98, 0, 115, 99, 114, 105, 112, 116, 0, 81, 97, 97, 105, 0, 115, 117, 98, 100, 105, 118, 105, 115, 105, 111, 110, 0, 99, 110, 49, 49, 0, 99, 110, 49, 50, 0, 99, 110, 49, 51, 0, 99, 110, 49, 52, 0, 99, 110, 49, 53, 0, 99, 110, 50, 49, 0, 99, 110, 50, 50, 0, 99, 110, 50, 51, 0, 99, 110, 51, 49, 0, 99, 110, 51, 50, 0, 99, 110, 51, 51, 0, 99, 110, 51, 52, 0, 99, 110, 51, 53, 0, 99, 110, 51, 54, 0, 99, 110, 51, 55, 0, 99, 110, 52, 49, 0, 99, 110, 52, 50, 0, 99, 110, 52, 51, 0, 99, 110, 52, 52, 0, 99, 110, 52, 53, 0, 99, 110, 52, 54, 0, 99, 110, 53, 48, 0, 99, 110, 53, 49, 0, 99, 110, 53, 50, 0, 99, 110, 53, 51, 0, 99, 110, 53, 52, 0, 99, 110, 54, 49, 0, 99, 110, 54, 50, 0, 99, 110, 54, 51, 0, 99, 110, 54, 52, 0, 99, 110, 54, 53, 0, 99, 110, 55, 49, 0, 99, 110, 57, 49, 0, 99, 110, 57, 50, 0, 99, 122, 49, 48, 97, 0, 99, 122, 49, 48, 98, 0, 99, 122, 49, 48, 99, 0, 99, 122, 49, 48, 100, 0, 99, 122, 49, 48, 101, 0, 99, 122, 49, 48, 102, 0, 99, 122, 54, 49, 49, 0, 99, 122, 54, 49, 50, 0, 99, 122, 54, 49, 51, 0, 99, 122, 54, 49, 52, 0, 99, 122, 54, 49, 53, 0, 99, 122, 54, 50, 49, 0, 99, 122, 54, 50, 50, 0, 99, 122, 54, 50, 51, 0, 99, 122, 54, 50, 52, 0, 99, 122, 54, 50, 54, 0, 99, 122, 54, 50, 55, 0, 99, 122, 106, 99, 0, 99, 122, 106, 109, 0, 99, 122, 107, 97, 0, 99, 122, 107, 114, 0, 99, 122, 108, 105, 0, 99, 122, 109, 111, 0, 99, 122, 111, 108, 0, 99, 122, 112, 97, 0, 99, 122, 112, 108, 0, 99, 122, 112, 114, 0, 99, 122, 115, 116, 0, 99, 122, 117, 115, 0, 99, 122, 118, 121, 0, 99, 122, 122, 108, 0, 101, 115, 99, 110, 0, 102, 105, 48, 49, 0, 102, 114, 98, 0, 102, 114, 98, 108, 0, 102, 114, 99, 0, 102, 114, 99, 112, 0, 102, 114, 100, 0, 102, 114, 102, 0, 102, 114, 103, 0, 102, 114, 103, 102, 0, 102, 114, 103, 112, 0, 102, 114, 104, 0, 102, 114, 105, 0, 102, 114, 106, 0, 102, 114, 107, 0, 102, 114, 108, 0, 102, 114, 109, 0, 102, 114, 109, 102, 0, 102, 114, 109, 113, 0, 102, 114, 110, 0, 102, 114, 110, 99, 0, 102, 114, 111, 0, 102, 114, 112, 0, 102, 114, 112, 102, 0, 102, 114, 112, 109, 0, 102, 114, 113, 0, 102, 114, 114, 0, 102, 114, 114, 101, 0, 102, 114, 115, 0, 102, 114, 116, 0, 102, 114, 116, 102, 0, 102, 114, 117, 0, 102, 114, 118, 0, 102, 114, 119, 102, 0, 102, 114, 121, 116, 0, 108, 97, 120, 110, 0, 108, 117, 100, 0, 108, 117, 108, 0, 109, 114, 110, 107, 99, 0, 110, 108, 97, 119, 0, 110, 108, 99, 119, 0, 110, 108, 115, 120, 0, 110, 111, 50, 51, 0, 110, 122, 110, 0, 110, 122, 115, 0, 111, 109, 98, 97, 0, 111, 109, 115, 104, 0, 112, 108, 100, 115, 0, 112, 108, 107, 112, 0, 112, 108, 108, 98, 0, 112, 108, 108, 100, 0, 112, 108, 108, 117, 0, 112, 108, 109, 97, 0, 112, 108, 109, 122, 0, 112, 108, 111, 112, 0, 112, 108, 112, 100, 0, 112, 108, 112, 107, 0, 112, 108, 112, 109, 0, 112, 108, 115, 107, 0, 112, 108, 115, 108, 0, 112, 108, 119, 110, 0, 112, 108, 119, 112, 0, 112, 108, 122, 112, 0, 115, 104, 116, 97, 0, 116, 116, 101, 116, 111, 0, 116, 116, 114, 99, 109, 0, 116, 116, 119, 116, 111, 0, 116, 119, 107, 104, 113, 0, 116, 119, 116, 110, 113, 0, 116, 119, 116, 112, 113, 0, 116, 119, 116, 120, 113, 0, 117, 115, 97, 115, 0, 117, 115, 103, 117, 0, 117, 115, 109, 112, 0, 117, 115, 112, 114, 0, 117, 115, 117, 109, 0, 117, 115, 118, 105, 0, 116, 101, 114, 114, 105, 116, 111, 114, 121, 0, 48, 48, 52, 0, 48, 48, 56, 0, 48, 49, 48, 0, 48, 49, 50, 0, 48, 49, 54, 0, 48, 50, 48, 0, 48, 50, 52, 0, 48, 50, 56, 0, 48, 51, 49, 0, 48, 51, 50, 0, 48, 51, 54, 0, 48, 52, 48, 0, 48, 52, 52, 0, 48, 52, 56, 0, 48, 53, 48, 0, 48, 53, 49, 0, 48, 53, 50, 0, 48, 53, 54, 0, 48, 54, 48, 0, 48, 54, 50, 0, 48, 54, 52, 0, 48, 54, 56, 0, 48, 55, 48, 0, 48, 55, 50, 0, 48, 55, 52, 0, 48, 55, 54, 0, 48, 56, 52, 0, 48, 56, 54, 0, 48, 57, 48, 0, 48, 57, 50, 0, 48, 57, 54, 0, 49, 48, 48, 0, 49, 48, 52, 0, 49, 48, 56, 0, 49, 49, 50, 0, 49, 49, 54, 0, 49, 50, 48, 0, 49, 50, 52, 0, 49, 51, 50, 0, 49, 51, 54, 0, 49, 52, 48, 0, 49, 52, 52, 0, 49, 52, 56, 0, 49, 53, 50, 0, 49, 53, 54, 0, 49, 53, 56, 0, 49, 54, 50, 0, 49, 54, 54, 0, 49, 55, 48, 0, 49, 55, 50, 0, 49, 55, 52, 0, 49, 55, 53, 0, 49, 55, 56, 0, 49, 56, 48, 0, 49, 56, 52, 0, 49, 56, 56, 0, 49, 57, 49, 0, 49, 57, 50, 0, 49, 57, 54, 0, 50, 48, 48, 0, 50, 48, 51, 0, 50, 48, 52, 0, 50, 48, 56, 0, 50, 49, 50, 0, 50, 49, 52, 0, 50, 49, 56, 0, 50, 50, 50, 0, 50, 50, 54, 0, 50, 51, 48, 0, 50, 51, 49, 0, 50, 51, 50, 0, 50, 51, 51, 0, 50, 51, 52, 0, 50, 51, 56, 0, 50, 51, 57, 0, 50, 52, 50, 0, 50, 52, 54, 0, 50, 52, 56, 0, 50, 52, 57, 0, 50, 53, 48, 0, 50, 53, 52, 0, 50, 53, 56, 0, 50, 54, 48, 0, 50, 54, 50, 0, 50, 54, 54, 0, 50, 54, 56, 0, 50, 55, 48, 0, 50, 55, 53, 0, 50, 55, 54, 0, 50, 55, 56, 0, 50, 56, 48, 0, 50, 56, 56, 0, 50, 57, 50, 0, 50, 57, 54, 0, 51, 48, 48, 0, 51, 48, 52, 0, 51, 48, 56, 0, 51, 49, 50, 0, 51, 49, 54, 0, 51, 50, 48, 0, 51, 50, 52, 0, 51, 50, 56, 0, 51, 51, 50, 0, 51, 51, 52, 0, 51, 51, 54, 0, 51, 52, 48, 0, 51, 52, 52, 0, 51, 52, 56, 0, 51, 53, 50, 0, 51, 53, 54, 0, 51, 54, 48, 0, 51, 54, 52, 0, 51, 54, 56, 0, 51, 55, 50, 0, 51, 55, 54, 0, 51, 56, 48, 0, 51, 56, 52, 0, 51, 56, 56, 0, 51, 57, 50, 0, 51, 57, 56, 0, 52, 48, 48, 0, 52, 48, 52, 0, 52, 48, 56, 0, 52, 49, 48, 0, 52, 49, 52, 0, 52, 49, 55, 0, 52, 49, 56, 0, 52, 50, 50, 0, 52, 50, 54, 0, 52, 50, 56, 0, 52, 51, 48, 0, 52, 51, 52, 0, 52, 51, 56, 0, 52, 52, 48, 0, 52, 52, 50, 0, 52, 52, 54, 0, 52, 53, 48, 0, 52, 53, 52, 0, 52, 53, 56, 0, 52, 54, 50, 0, 52, 54, 54, 0, 52, 55, 48, 0, 52, 55, 52, 0, 52, 55, 56, 0, 52, 56, 48, 0, 52, 56, 52, 0, 52, 57, 50, 0, 52, 57, 54, 0, 52, 57, 56, 0, 52, 57, 57, 0, 53, 48, 48, 0, 53, 48, 52, 0, 53, 48, 56, 0, 53, 49, 50, 0, 53, 49, 54, 0, 53, 50, 48, 0, 53, 50, 52, 0, 53, 50, 56, 0, 53, 51, 48, 0, 53, 51, 49, 0, 53, 51, 50, 0, 53, 51, 51, 0, 53, 51, 52, 0, 53, 51, 53, 0, 53, 51, 54, 0, 53, 52, 48, 0, 53, 52, 56, 0, 53, 53, 52, 0, 53, 53, 56, 0, 53, 54, 50, 0, 53, 54, 54, 0, 53, 55, 48, 0, 53, 55, 52, 0, 53, 55, 56, 0, 53, 56, 48, 0, 53, 56, 49, 0, 53, 56, 50, 0, 53, 56, 51, 0, 53, 56, 52, 0, 53, 56, 53, 0, 53, 56, 54, 0, 53, 57, 49, 0, 53, 57, 56, 0, 54, 48, 48, 0, 54, 48, 52, 0, 54, 48, 56, 0, 54, 49, 54, 0, 54, 50, 48, 0, 54, 51, 48, 0, 54, 51, 52, 0, 54, 51, 56, 0, 54, 52, 50, 0, 54, 52, 51, 0, 54, 52, 54, 0, 54, 53, 50, 0, 54, 53, 52, 0, 54, 53, 57, 0, 54, 54, 48, 0, 54, 54, 50, 0, 54, 54, 51, 0, 54, 54, 54, 0, 54, 55, 48, 0, 54, 55, 52, 0, 54, 55, 56, 0, 54, 56, 50, 0, 54, 56, 54, 0, 54, 56, 56, 0, 54, 57, 48, 0, 54, 57, 52, 0, 55, 48, 50, 0, 55, 48, 51, 0, 55, 48, 52, 0, 55, 48, 53, 0, 55, 48, 54, 0, 55, 49, 48, 0, 55, 49, 54, 0, 55, 50, 48, 0, 55, 50, 52, 0, 55, 50, 56, 0, 55, 50, 57, 0, 55, 51, 50, 0, 55, 51, 54, 0, 55, 52, 48, 0, 55, 52, 52, 0, 55, 52, 56, 0, 55, 53, 50, 0, 55, 53, 54, 0, 55, 54, 48, 0, 55, 54, 50, 0, 55, 54, 52, 0, 55, 54, 56, 0, 55, 55, 50, 0, 55, 55, 54, 0, 55, 56, 48, 0, 55, 56, 52, 0, 55, 56, 56, 0, 55, 57, 50, 0, 55, 57, 53, 0, 55, 57, 54, 0, 55, 57, 56, 0, 56, 48, 48, 0, 56, 48, 52, 0, 56, 48, 55, 0, 56, 49, 48, 0, 56, 49, 56, 0, 56, 50, 54, 0, 56, 51, 48, 0, 56, 51, 49, 0, 56, 51, 50, 0, 56, 51, 51, 0, 56, 51, 52, 0, 56, 52, 48, 0, 56, 53, 48, 0, 56, 53, 52, 0, 56, 53, 56, 0, 56, 54, 48, 0, 56, 54, 50, 0, 56, 55, 54, 0, 56, 56, 50, 0, 56, 56, 54, 0, 56, 56, 55, 0, 56, 57, 48, 0, 56, 57, 49, 0, 56, 57, 52, 0, 57, 53, 56, 0, 57, 53, 57, 0, 57, 54, 48, 0, 57, 54, 50, 0, 57, 54, 51, 0, 57, 54, 52, 0, 57, 54, 53, 0, 57, 54, 54, 0, 57, 54, 55, 0, 57, 54, 56, 0, 57, 54, 57, 0, 57, 55, 48, 0, 57, 55, 49, 0, 57, 55, 50, 0, 57, 55, 51, 0, 57, 55, 52, 0, 57, 55, 53, 0, 57, 55, 54, 0, 57, 55, 55, 0, 57, 55, 56, 0, 57, 55, 57, 0, 57, 56, 48, 0, 57, 56, 49, 0, 57, 56, 50, 0, 57, 56, 51, 0, 57, 56, 52, 0, 57, 56, 53, 0, 57, 56, 54, 0, 57, 56, 55, 0, 57, 56, 56, 0, 57, 56, 57, 0, 57, 57, 48, 0, 57, 57, 49, 0, 57, 57, 50, 0, 57, 57, 51, 0, 57, 57, 52, 0, 57, 57, 53, 0, 57, 57, 54, 0, 57, 57, 55, 0, 57, 57, 56, 0, 57, 57, 57, 0, 65, 65, 65, 0, 65, 66, 87, 0, 65, 70, 71, 0, 65, 71, 79, 0, 65, 73, 65, 0, 65, 76, 65, 0, 65, 76, 66, 0, 65, 78, 84, 0, 65, 82, 69, 0, 65, 82, 71, 0, 65, 82, 77, 0, 65, 83, 67, 0, 65, 83, 77, 0, 65, 84, 65, 0, 65, 84, 70, 0, 65, 84, 71, 0, 65, 85, 83, 0, 65, 85, 84, 0, 65, 90, 69, 0, 66, 68, 73, 0, 66, 69, 76, 0, 66, 69, 78, 0, 66, 69, 83, 0, 66, 70, 65, 0, 66, 71, 68, 0, 66, 71, 82, 0, 66, 72, 82, 0, 66, 72, 83, 0, 66, 73, 72, 0, 66, 76, 77, 0, 66, 76, 82, 0, 66, 76, 90, 0, 66, 77, 85, 0, 66, 79, 76, 0, 66, 82, 65, 0, 66, 82, 66, 0, 66, 82, 78, 0, 66, 84, 78, 0, 66, 85, 0, 66, 85, 82, 0, 66, 86, 84, 0, 66, 87, 65, 0, 67, 65, 70, 0, 67, 65, 78, 0, 67, 67, 75, 0, 67, 72, 69, 0, 67, 72, 76, 0, 67, 72, 78, 0, 67, 73, 86, 0, 67, 77, 82, 0, 67, 79, 68, 0, 67, 79, 71, 0, 67, 79, 75, 0, 67, 79, 76, 0, 67, 79, 77, 0, 67, 80, 84, 0, 67, 80, 86, 0, 67, 82, 73, 0, 67, 83, 0, 67, 85, 66, 0, 67, 85, 87, 0, 67, 88, 82, 0, 67, 89, 77, 0, 67, 89, 80, 0, 67, 90, 69, 0, 68, 68, 82, 0, 68, 69, 85, 0, 68, 71, 65, 0, 68, 74, 73, 0, 68, 77, 65, 0, 68, 78, 75, 0, 68, 79, 77, 0, 68, 89, 0, 68, 90, 65, 0, 69, 67, 85, 0, 69, 71, 89, 0, 69, 82, 73, 0, 69, 83, 72, 0, 69, 83, 80, 0, 69, 83, 84, 0, 69, 84, 72, 0, 70, 73, 78, 0, 70, 74, 73, 0, 70, 76, 75, 0, 70, 81, 0, 70, 82, 65, 0, 70, 82, 79, 0, 70, 83, 77, 0, 70, 88, 0, 70, 88, 88, 0, 71, 65, 66, 0, 71, 66, 82, 0, 71, 69, 79, 0, 71, 71, 89, 0, 71, 72, 65, 0, 71, 73, 66, 0, 71, 73, 78, 0, 71, 76, 80, 0, 71, 77, 66, 0, 71, 78, 66, 0, 71, 78, 81, 0, 71, 82, 67, 0, 71, 82, 68, 0, 71, 82, 76, 0, 71, 84, 77, 0, 71, 85, 70, 0, 71, 85, 77, 0, 71, 85, 89, 0, 72, 75, 71, 0, 72, 77, 68, 0, 72, 78, 68, 0, 72, 82, 86, 0, 72, 84, 73, 0, 72, 85, 78, 0, 72, 86, 0, 73, 68, 78, 0, 73, 77, 78, 0, 73, 78, 68, 0, 73, 79, 84, 0, 73, 82, 76, 0, 73, 82, 78, 0, 73, 82, 81, 0, 73, 83, 76, 0, 73, 83, 82, 0, 73, 84, 65, 0, 74, 65, 77, 0, 74, 69, 89, 0, 74, 79, 82, 0, 74, 80, 78, 0, 74, 84, 0, 75, 65, 90, 0, 75, 69, 78, 0, 75, 71, 90, 0, 75, 72, 77, 0, 75, 73, 82, 0, 75, 78, 65, 0, 75, 79, 82, 0, 75, 87, 84, 0, 76, 65, 79, 0, 76, 66, 78, 0, 76, 66, 82, 0, 76, 66, 89, 0, 76, 67, 65, 0, 76, 73, 69, 0, 76, 75, 65, 0, 76, 83, 79, 0, 76, 84, 85, 0, 76, 85, 88, 0, 76, 86, 65, 0, 77, 65, 67, 0, 77, 65, 70, 0, 77, 65, 82, 0, 77, 67, 79, 0, 77, 68, 65, 0, 77, 68, 71, 0, 77, 68, 86, 0, 77, 69, 88, 0, 77, 72, 76, 0, 77, 75, 68, 0, 77, 76, 73, 0, 77, 76, 84, 0, 77, 77, 82, 0, 77, 78, 69, 0, 77, 78, 71, 0, 77, 78, 80, 0, 77, 79, 90, 0, 77, 82, 84, 0, 77, 83, 82, 0, 77, 84, 81, 0, 77, 85, 83, 0, 77, 87, 73, 0, 77, 89, 83, 0, 77, 89, 84, 0, 78, 65, 77, 0, 78, 67, 76, 0, 78, 69, 82, 0, 78, 70, 75, 0, 78, 71, 65, 0, 78, 72, 0, 78, 73, 67, 0, 78, 73, 85, 0, 78, 76, 68, 0, 78, 79, 82, 0, 78, 80, 76, 0, 78, 82, 85, 0, 78, 84, 90, 0, 78, 90, 76, 0, 79, 77, 78, 0, 80, 65, 75, 0, 80, 65, 78, 0, 80, 67, 0, 80, 67, 78, 0, 80, 69, 82, 0, 80, 72, 76, 0, 80, 76, 87, 0, 80, 78, 71, 0, 80, 79, 76, 0, 80, 82, 73, 0, 80, 82, 75, 0, 80, 82, 84, 0, 80, 82, 89, 0, 80, 83, 69, 0, 80, 85, 0, 80, 89, 70, 0, 80, 90, 0, 81, 65, 84, 0, 81, 77, 77, 0, 81, 78, 78, 0, 81, 80, 80, 0, 81, 81, 81, 0, 81, 82, 82, 0, 81, 83, 83, 0, 81, 84, 84, 0, 81, 85, 0, 81, 85, 85, 0, 81, 86, 86, 0, 81, 87, 87, 0, 81, 88, 88, 0, 81, 89, 89, 0, 81, 90, 90, 0, 82, 69, 85, 0, 82, 72, 0, 82, 79, 85, 0, 82, 85, 83, 0, 82, 87, 65, 0, 83, 65, 85, 0, 83, 67, 71, 0, 83, 68, 78, 0, 83, 69, 78, 0, 83, 71, 80, 0, 83, 71, 83, 0, 83, 72, 78, 0, 83, 74, 77, 0, 83, 76, 66, 0, 83, 76, 69, 0, 83, 76, 86, 0, 83, 77, 82, 0, 83, 79, 77, 0, 83, 80, 77, 0, 83, 82, 66, 0, 83, 83, 68, 0, 83, 84, 80, 0, 83, 85, 0, 83, 85, 78, 0, 83, 85, 82, 0, 83, 86, 75, 0, 83, 86, 78, 0, 83, 87, 69, 0, 83, 87, 90, 0, 83, 88, 77, 0, 83, 89, 67, 0, 83, 89, 82, 0, 84, 65, 65, 0, 84, 67, 65, 0, 84, 67, 68, 0, 84, 71, 79, 0, 84, 72, 65, 0, 84, 74, 75, 0, 84, 75, 76, 0, 84, 75, 77, 0, 84, 76, 83, 0, 84, 77, 80, 0, 84, 79, 78, 0, 84, 84, 79, 0, 84, 85, 78, 0, 84, 85, 82, 0, 84, 85, 86, 0, 84, 87, 78, 0, 84, 90, 65, 0, 85, 71, 65, 0, 85, 75, 0, 85, 75, 82, 0, 85, 77, 73, 0, 85, 82, 89, 0, 85, 83, 65, 0, 85, 90, 66, 0, 86, 65, 84, 0, 86, 67, 84, 0, 86, 68, 0, 86, 69, 78, 0, 86, 71, 66, 0, 86, 73, 82, 0, 86, 78, 77, 0, 86, 85, 84, 0, 87, 75, 0, 87, 76, 70, 0, 87, 83, 77, 0, 88, 65, 65, 0, 88, 66, 66, 0, 88, 67, 67, 0, 88, 68, 68, 0, 88, 69, 69, 0, 88, 70, 70, 0, 88, 71, 71, 0, 88, 72, 72, 0, 88, 73, 73, 0, 88, 74, 74, 0, 88, 75, 75, 0, 88, 76, 76, 0, 88, 77, 77, 0, 88, 78, 78, 0, 88, 79, 79, 0, 88, 80, 80, 0, 88, 81, 81, 0, 88, 82, 82, 0, 88, 83, 83, 0, 88, 84, 84, 0, 88, 85, 85, 0, 88, 86, 86, 0, 88, 87, 87, 0, 88, 88, 88, 0, 88, 89, 89, 0, 88, 90, 90, 0, 89, 68, 0, 89, 69, 77, 0, 89, 77, 68, 0, 89, 85, 0, 89, 85, 71, 0, 90, 65, 70, 0, 90, 65, 82, 0, 90, 77, 66, 0, 90, 82, 0, 90, 87, 69, 0, 90, 90, 90, 0, 118, 97, 114, 105, 97, 110, 116, 0, 65, 65, 76, 65, 78, 68, 0, 72, 69, 80, 76, 79, 67, 0, 80, 79, 76, 89, 84, 79, 78, 73, 0, 97, 114, 101, 118, 101, 108, 97, 0, 97, 114, 101, 118, 109, 100, 97, 0, 104, 101, 112, 108, 111, 99, 0, 100, 101, 102, 97, 117, 108, 116, 67, 111, 110, 116, 101, 110, 116, 0, 0, 0, 85, 0, 77, 0, 0, 0, 89, 0, 69, 0, 0, 0, 110, 0, 98, 0, 0, 0, 65, 0, 88, 0, 0, 0, 84, 0, 76, 0, 0, 0, 114, 0, 111, 0, 0, 0, 122, 0, 104, 0, 0, 0, 65, 0, 81, 0, 0, 0, 65, 0, 83, 0, 0, 0, 65, 0, 87, 0, 0, 0, 66, 0, 70, 0, 0, 0, 66, 0, 74, 0, 0, 0, 66, 0, 76, 0, 0, 0, 67, 0, 87, 0, 0, 0, 69, 0, 85, 0, 0, 0, 71, 0, 70, 0, 0, 0, 71, 0, 80, 0, 0, 0, 71, 0, 85, 0, 0, 0, 75, 0, 73, 0, 0, 0, 77, 0, 70, 0, 0, 0, 77, 0, 80, 0, 0, 0, 77, 0, 81, 0, 0, 0, 78, 0, 67, 0, 0, 0, 80, 0, 65, 0, 0, 0, 80, 0, 70, 0, 0, 0, 80, 0, 77, 0, 0, 0, 80, 0, 82, 0, 0, 0, 82, 0, 69, 0, 0, 0, 83, 0, 68, 0, 0, 0, 83, 0, 88, 0, 0, 0, 86, 0, 73, 0, 0, 0, 86, 0, 85, 0, 0, 0, 87, 0, 70, 0, 0, 0, 89, 0, 84, 0, 0, 0, 102, 0, 97, 0, 0, 0, 104, 0, 121, 0, 0, 0, 109, 0, 115, 0, 0, 0, 115, 0, 113, 0, 0, 0, 121, 0, 105, 0, 0, 0, 65, 0, 65, 0, 0, 0, 65, 0, 68, 0, 0, 0, 65, 0, 69, 0, 0, 0, 65, 0, 71, 0, 0, 0, 65, 0, 73, 0, 0, 0, 65, 0, 79, 0, 0, 0, 65, 0, 82, 0, 0, 0, 65, 0, 84, 0, 0, 0, 66, 0, 66, 0, 0, 0, 66, 0, 72, 0, 0, 0, 66, 0, 77, 0, 0, 0, 66, 0, 79, 0, 0, 0, 66, 0, 83, 0, 0, 0, 66, 0, 86, 0, 0, 0, 66, 0, 87, 0, 0, 0, 66, 0, 90, 0, 0, 0, 67, 0, 67, 0, 0, 0, 67, 0, 71, 0, 0, 0, 67, 0, 73, 0, 0, 0, 67, 0, 75, 0, 0, 0, 67, 0, 79, 0, 0, 0, 67, 0, 80, 0, 0, 0, 67, 0, 82, 0, 0, 0, 67, 0, 85, 0, 0, 0, 67, 0, 88, 0, 0, 0, 67, 0, 89, 0, 0, 0, 68, 0, 74, 0, 0, 0, 68, 0, 77, 0, 0, 0, 68, 0, 79, 0, 0, 0, 69, 0, 67, 0, 0, 0, 69, 0, 71, 0, 0, 0, 69, 0, 72, 0, 0, 0, 70, 0, 74, 0, 0, 0, 70, 0, 75, 0, 0, 0, 70, 0, 77, 0, 0, 0, 71, 0, 65, 0, 0, 0, 71, 0, 68, 0, 0, 0, 71, 0, 73, 0, 0, 0, 71, 0, 77, 0, 0, 0, 71, 0, 81, 0, 0, 0, 71, 0, 83, 0, 0, 0, 71, 0, 87, 0, 0, 0, 71, 0, 89, 0, 0, 0, 72, 0, 77, 0, 0, 0, 72, 0, 78, 0, 0, 0, 72, 0, 84, 0, 0, 0, 73, 0, 79, 0, 0, 0, 74, 0, 69, 0, 0, 0, 74, 0, 77, 0, 0, 0, 74, 0, 79, 0, 0, 0, 75, 0, 77, 0, 0, 0, 75, 0, 78, 0, 0, 0, 75, 0, 80, 0, 0, 0, 75, 0, 87, 0, 0, 0, 75, 0, 89, 0, 0, 0, 76, 0, 66, 0, 0, 0, 76, 0, 67, 0, 0, 0, 76, 0, 73, 0, 0, 0, 76, 0, 83, 0, 0, 0, 76, 0, 89, 0, 0, 0, 77, 0, 67, 0, 0, 0, 77, 0, 68, 0, 0, 0, 77, 0, 72, 0, 0, 0, 77, 0, 82, 0, 0, 0, 77, 0, 83, 0, 0, 0, 77, 0, 88, 0, 0, 0, 78, 0, 70, 0, 0, 0, 78, 0, 73, 0, 0, 0, 78, 0, 82, 0, 0, 0, 78, 0, 85, 0, 0, 0, 79, 0, 77, 0, 0, 0, 80, 0, 71, 0, 0, 0, 80, 0, 78, 0, 0, 0, 80, 0, 83, 0, 0, 0, 80, 0, 84, 0, 0, 0, 81, 0, 65, 0, 0, 0, 81, 0, 77, 0, 0, 0, 81, 0, 78, 0, 0, 0, 81, 0, 80, 0, 0, 0, 81, 0, 81, 0, 0, 0, 81, 0, 82, 0, 0, 0, 81, 0, 83, 0, 0, 0, 81, 0, 84, 0, 0, 0, 81, 0, 86, 0, 0, 0, 81, 0, 87, 0, 0, 0, 81, 0, 88, 0, 0, 0, 81, 0, 89, 0, 0, 0, 81, 0, 90, 0, 0, 0, 83, 0, 65, 0, 0, 0, 83, 0, 66, 0, 0, 0, 83, 0, 67, 0, 0, 0, 83, 0, 72, 0, 0, 0, 83, 0, 74, 0, 0, 0, 83, 0, 76, 0, 0, 0, 83, 0, 77, 0, 0, 0, 83, 0, 82, 0, 0, 0, 83, 0, 84, 0, 0, 0, 83, 0, 86, 0, 0, 0, 83, 0, 89, 0, 0, 0, 83, 0, 90, 0, 0, 0, 84, 0, 65, 0, 0, 0, 84, 0, 67, 0, 0, 0, 84, 0, 68, 0, 0, 0, 84, 0, 75, 0, 0, 0, 84, 0, 78, 0, 0, 0, 84, 0, 84, 0, 0, 0, 84, 0, 86, 0, 0, 0, 85, 0, 89, 0, 0, 0, 86, 0, 65, 0, 0, 0, 86, 0, 67, 0, 0, 0, 86, 0, 69, 0, 0, 0, 86, 0, 71, 0, 0, 0, 87, 0, 83, 0, 0, 0, 88, 0, 65, 0, 0, 0, 88, 0, 66, 0, 0, 0, 88, 0, 67, 0, 0, 0, 88, 0, 68, 0, 0, 0, 88, 0, 69, 0, 0, 0, 88, 0, 70, 0, 0, 0, 88, 0, 71, 0, 0, 0, 88, 0, 72, 0, 0, 0, 88, 0, 73, 0, 0, 0, 88, 0, 74, 0, 0, 0, 88, 0, 76, 0, 0, 0, 88, 0, 77, 0, 0, 0, 88, 0, 78, 0, 0, 0, 88, 0, 79, 0, 0, 0, 88, 0, 80, 0, 0, 0, 88, 0, 81, 0, 0, 0, 88, 0, 82, 0, 0, 0, 88, 0, 83, 0, 0, 0, 88, 0, 84, 0, 0, 0, 88, 0, 85, 0, 0, 0, 88, 0, 86, 0, 0, 0, 88, 0, 87, 0, 0, 0, 88, 0, 88, 0, 0, 0, 88, 0, 89, 0, 0, 0, 88, 0, 90, 0, 0, 0, 90, 0, 90, 0, 0, 0, 97, 0, 114, 0, 0, 0, 97, 0, 122, 0, 0, 0, 99, 0, 114, 0, 0, 0, 99, 0, 115, 0, 0, 0, 100, 0, 101, 0, 0, 0, 101, 0, 116, 0, 0, 0, 101, 0, 117, 0, 0, 0, 102, 0, 102, 0, 0, 0, 102, 0, 114, 0, 0, 0, 103, 0, 110, 0, 0, 0, 105, 0, 100, 0, 0, 0, 105, 0, 115, 0, 0, 0, 105, 0, 117, 0, 0, 0, 106, 0, 118, 0, 0, 0, 107, 0, 97, 0, 0, 0, 107, 0, 103, 0, 0, 0, 107, 0, 114, 0, 0, 0, 107, 0, 117, 0, 0, 0, 107, 0, 118, 0, 0, 0, 108, 0, 98, 0, 0, 0, 108, 0, 118, 0, 0, 0, 109, 0, 103, 0, 0, 0, 109, 0, 107, 0, 0, 0, 109, 0, 121, 0, 0, 0, 110, 0, 101, 0, 0, 0, 110, 0, 108, 0, 0, 0, 110, 0, 118, 0, 0, 0, 111, 0, 106, 0, 0, 0, 112, 0, 115, 0, 0, 0, 113, 0, 117, 0, 0, 0, 115, 0, 107, 0, 0, 0, 115, 0, 114, 0, 0, 0, 115, 0, 119, 0, 0, 0, 65, 0, 67, 0, 0, 0, 68, 0, 71, 0, 0, 0, 73, 0, 67, 0, 0, 0, 97, 0, 97, 0, 0, 0, 97, 0, 98, 0, 0, 0, 97, 0, 101, 0, 0, 0, 97, 0, 118, 0, 0, 0, 98, 0, 101, 0, 0, 0, 98, 0, 103, 0, 0, 0, 98, 0, 105, 0, 0, 0, 98, 0, 109, 0, 0, 0, 98, 0, 110, 0, 0, 0, 99, 0, 97, 0, 0, 0, 99, 0, 101, 0, 0, 0, 99, 0, 111, 0, 0, 0, 99, 0, 117, 0, 0, 0, 99, 0, 118, 0, 0, 0, 100, 0, 97, 0, 0, 0, 100, 0, 118, 0, 0, 0, 101, 0, 101, 0, 0, 0, 102, 0, 105, 0, 0, 0, 102, 0, 111, 0, 0, 0, 103, 0, 97, 0, 0, 0, 103, 0, 108, 0, 0, 0, 103, 0, 117, 0, 0, 0, 104, 0, 116, 0, 0, 0, 104, 0, 117, 0, 0, 0, 104, 0, 122, 0, 0, 0, 105, 0, 97, 0, 0, 0, 105, 0, 101, 0, 0, 0, 105, 0, 103, 0, 0, 0, 105, 0, 111, 0, 0, 0, 105, 0, 116, 0, 0, 0, 106, 0, 97, 0, 0, 0, 107, 0, 106, 0, 0, 0, 107, 0, 107, 0, 0, 0, 107, 0, 108, 0, 0, 0, 107, 0, 109, 0, 0, 0, 107, 0, 110, 0, 0, 0, 107, 0, 115, 0, 0, 0, 107, 0, 119, 0, 0, 0, 107, 0, 121, 0, 0, 0, 108, 0, 97, 0, 0, 0, 108, 0, 103, 0, 0, 0, 108, 0, 105, 0, 0, 0, 108, 0, 111, 0, 0, 0, 108, 0, 117, 0, 0, 0, 109, 0, 116, 0, 0, 0, 110, 0, 97, 0, 0, 0, 110, 0, 100, 0, 0, 0, 110, 0, 114, 0, 0, 0, 111, 0, 99, 0, 0, 0, 111, 0, 115, 0, 0, 0, 112, 0, 105, 0, 0, 0, 112, 0, 108, 0, 0, 0, 112, 0, 116, 0, 0, 0, 114, 0, 110, 0, 0, 0, 114, 0, 119, 0, 0, 0, 115, 0, 97, 0, 0, 0, 115, 0, 101, 0, 0, 0, 115, 0, 103, 0, 0, 0, 115, 0, 105, 0, 0, 0, 115, 0, 108, 0, 0, 0, 115, 0, 109, 0, 0, 0, 115, 0, 111, 0, 0, 0, 115, 0, 116, 0, 0, 0, 115, 0, 118, 0, 0, 0, 116, 0, 97, 0, 0, 0, 116, 0, 101, 0, 0, 0, 116, 0, 103, 0, 0, 0, 116, 0, 104, 0, 0, 0, 116, 0, 105, 0, 0, 0, 116, 0, 107, 0, 0, 0, 116, 0, 111, 0, 0, 0, 116, 0, 114, 0, 0, 0, 116, 0, 115, 0, 0, 0, 116, 0, 116, 0, 0, 0, 116, 0, 121, 0, 0, 0, 117, 0, 107, 0, 0, 0, 117, 0, 114, 0, 0, 0, 118, 0, 101, 0, 0, 0, 118, 0, 105, 0, 0, 0, 118, 0, 111, 0, 0, 0, 119, 0, 97, 0, 0, 0, 119, 0, 111, 0, 0, 0, 120, 0, 104, 0, 0, 0, 121, 0, 111, 0, 0, 0, 122, 0, 117, 0, 0, 0, 100, 0, 116, 0, 112, 0, 0, 0, 107, 0, 97, 0, 107, 0, 0, 0, 111, 0, 121, 0, 98, 0, 0, 0, 118, 0, 97, 0, 106, 0, 0, 0, 98, 0, 104, 0, 111, 0, 0, 0, 114, 0, 107, 0, 105, 0, 0, 0, 97, 0, 109, 0, 105, 0, 0, 0, 98, 0, 105, 0, 107, 0, 0, 0, 100, 0, 101, 0, 108, 0, 0, 0, 100, 0, 117, 0, 122, 0, 0, 0, 106, 0, 98, 0, 111, 0, 0, 0, 107, 0, 100, 0, 122, 0, 0, 0, 112, 0, 104, 0, 114, 0, 0, 0, 116, 0, 97, 0, 121, 0, 0, 0, 122, 0, 111, 0, 109, 0, 0, 0, 122, 0, 122, 0, 97, 0, 0, 0, 102, 0, 105, 0, 108, 0, 0, 0, 104, 0, 97, 0, 107, 0, 0, 0, 109, 0, 97, 0, 110, 0, 0, 0, 109, 0, 114, 0, 121, 0, 0, 0, 112, 0, 105, 0, 106, 0, 0, 0, 98, 0, 102, 0, 121, 0, 0, 0, 99, 0, 109, 0, 114, 0, 0, 0, 100, 0, 101, 0, 110, 0, 0, 0, 103, 0, 98, 0, 97, 0, 0, 0, 107, 0, 109, 0, 108, 0, 0, 0, 110, 0, 97, 0, 110, 0, 0, 0, 110, 0, 98, 0, 114, 0, 0, 0, 110, 0, 103, 0, 118, 0, 0, 0, 111, 0, 112, 0, 97, 0, 0, 0, 114, 0, 97, 0, 115, 0, 0, 0, 116, 0, 109, 0, 104, 0, 0, 0, 116, 0, 115, 0, 117, 0, 0, 0, 119, 0, 101, 0, 111, 0, 0, 0, 120, 0, 99, 0, 104, 0, 0, 0, 120, 0, 110, 0, 121, 0, 0, 0, 121, 0, 97, 0, 109, 0, 0, 0, 121, 0, 117, 0, 103, 0, 0, 0, 97, 0, 97, 0, 115, 0, 0, 0, 97, 0, 99, 0, 110, 0, 0, 0, 97, 0, 100, 0, 120, 0, 0, 0, 97, 0, 111, 0, 103, 0, 0, 0, 98, 0, 97, 0, 108, 0, 0, 0, 98, 0, 99, 0, 103, 0, 0, 0, 98, 0, 109, 0, 102, 0, 0, 0, 98, 0, 110, 0, 99, 0, 0, 0, 98, 0, 110, 0, 110, 0, 0, 0, 98, 0, 117, 0, 97, 0, 0, 0, 99, 0, 97, 0, 120, 0, 0, 0, 99, 0, 104, 0, 109, 0, 0, 0, 99, 0, 105, 0, 114, 0, 0, 0, 100, 0, 101, 0, 118, 0, 0, 0, 100, 0, 105, 0, 102, 0, 0, 0, 100, 0, 105, 0, 110, 0, 0, 0, 100, 0, 111, 0, 105, 0, 0, 0, 100, 0, 114, 0, 108, 0, 0, 0, 101, 0, 109, 0, 97, 0, 0, 0, 103, 0, 97, 0, 108, 0, 0, 0, 103, 0, 100, 0, 106, 0, 0, 0, 103, 0, 111, 0, 110, 0, 0, 0, 103, 0, 114, 0, 98, 0, 0, 0, 103, 0, 118, 0, 114, 0, 0, 0, 104, 0, 97, 0, 105, 0, 0, 0, 104, 0, 108, 0, 101, 0, 0, 0, 104, 0, 109, 0, 110, 0, 0, 0, 104, 0, 115, 0, 110, 0, 0, 0, 104, 0, 117, 0, 119, 0, 0, 0, 104, 0, 121, 0, 119, 0, 0, 0, 106, 0, 97, 0, 108, 0, 0, 0, 106, 0, 114, 0, 98, 0, 0, 0, 107, 0, 108, 0, 110, 0, 0, 0, 107, 0, 111, 0, 107, 0, 0, 0, 107, 0, 112, 0, 101, 0, 0, 0, 107, 0, 116, 0, 122, 0, 0, 0, 107, 0, 119, 0, 118, 0, 0, 0, 108, 0, 97, 0, 104, 0, 0, 0, 108, 0, 99, 0, 113, 0, 0, 0, 108, 0, 114, 0, 114, 0, 0, 0, 108, 0, 117, 0, 121, 0, 0, 0, 109, 0, 111, 0, 109, 0, 0, 0, 109, 0, 116, 0, 109, 0, 0, 0, 109, 0, 119, 0, 114, 0, 0, 0, 110, 0, 103, 0, 116, 0, 0, 0, 110, 0, 117, 0, 110, 0, 0, 0, 110, 0, 121, 0, 99, 0, 0, 0, 112, 0, 114, 0, 116, 0, 0, 0, 112, 0, 117, 0, 98, 0, 0, 0, 112, 0, 119, 0, 110, 0, 0, 0, 113, 0, 117, 0, 104, 0, 0, 0, 114, 0, 97, 0, 106, 0, 0, 0, 114, 0, 97, 0, 113, 0, 0, 0, 114, 0, 109, 0, 120, 0, 0, 0, 114, 0, 111, 0, 109, 0, 0, 0, 115, 0, 102, 0, 98, 0, 0, 0, 115, 0, 103, 0, 103, 0, 0, 0, 115, 0, 110, 0, 122, 0, 0, 0, 115, 0, 114, 0, 120, 0, 0, 0, 115, 0, 115, 0, 121, 0, 0, 0, 115, 0, 117, 0, 106, 0, 0, 0, 115, 0, 121, 0, 114, 0, 0, 0, 116, 0, 97, 0, 106, 0, 0, 0, 116, 0, 97, 0, 111, 0, 0, 0, 116, 0, 100, 0, 102, 0, 0, 0, 116, 0, 108, 0, 104, 0, 0, 0, 116, 0, 112, 0, 111, 0, 0, 0, 116, 0, 118, 0, 100, 0, 0, 0, 116, 0, 119, 0, 109, 0, 0, 0, 116, 0, 121, 0, 106, 0, 0, 0, 118, 0, 103, 0, 116, 0, 0, 0, 119, 0, 97, 0, 119, 0, 0, 0, 122, 0, 97, 0, 112, 0, 0, 0, 99, 0, 110, 0, 109, 0, 110, 0, 0, 0, 99, 0, 110, 0, 104, 0, 101, 0, 0, 0, 99, 0, 110, 0, 115, 0, 99, 0, 0, 0, 99, 0, 110, 0, 102, 0, 106, 0, 0, 0, 99, 0, 110, 0, 103, 0, 100, 0, 0, 0, 99, 0, 110, 0, 104, 0, 97, 0, 0, 0, 99, 0, 110, 0, 104, 0, 105, 0, 0, 0, 99, 0, 110, 0, 108, 0, 110, 0, 0, 0, 99, 0, 110, 0, 115, 0, 100, 0, 0, 0, 99, 0, 110, 0, 115, 0, 110, 0, 0, 0, 90, 0, 105, 0, 110, 0, 104, 0, 0, 0, 99, 0, 110, 0, 97, 0, 104, 0, 0, 0, 99, 0, 110, 0, 98, 0, 106, 0, 0, 0, 99, 0, 110, 0, 99, 0, 113, 0, 0, 0, 99, 0, 110, 0, 103, 0, 115, 0, 0, 0, 99, 0, 110, 0, 103, 0, 120, 0, 0, 0, 99, 0, 110, 0, 103, 0, 122, 0, 0, 0, 99, 0, 110, 0, 104, 0, 98, 0, 0, 0, 99, 0, 110, 0, 104, 0, 108, 0, 0, 0, 99, 0, 110, 0, 104, 0, 110, 0, 0, 0, 99, 0, 110, 0, 106, 0, 108, 0, 0, 0, 99, 0, 110, 0, 106, 0, 115, 0, 0, 0, 99, 0, 110, 0, 106, 0, 120, 0, 0, 0, 99, 0, 110, 0, 110, 0, 120, 0, 0, 0, 99, 0, 110, 0, 113, 0, 104, 0, 0, 0, 99, 0, 110, 0, 115, 0, 104, 0, 0, 0, 99, 0, 110, 0, 115, 0, 120, 0, 0, 0, 99, 0, 110, 0, 116, 0, 106, 0, 0, 0, 99, 0, 110, 0, 120, 0, 106, 0, 0, 0, 99, 0, 110, 0, 120, 0, 122, 0, 0, 0, 99, 0, 110, 0, 121, 0, 110, 0, 0, 0, 99, 0, 110, 0, 122, 0, 106, 0, 0, 0, 99, 0, 122, 0, 49, 0, 48, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 0, 0, 99, 0, 122, 0, 51, 0, 49, 0, 0, 0, 99, 0, 122, 0, 51, 0, 50, 0, 0, 0, 99, 0, 122, 0, 52, 0, 49, 0, 0, 0, 99, 0, 122, 0, 52, 0, 50, 0, 0, 0, 99, 0, 122, 0, 53, 0, 49, 0, 0, 0, 99, 0, 122, 0, 53, 0, 50, 0, 0, 0, 99, 0, 122, 0, 53, 0, 51, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 0, 0, 99, 0, 122, 0, 55, 0, 49, 0, 0, 0, 99, 0, 122, 0, 55, 0, 50, 0, 0, 0, 99, 0, 122, 0, 56, 0, 48, 0, 0, 0, 108, 0, 97, 0, 120, 0, 115, 0, 0, 0, 110, 0, 111, 0, 53, 0, 48, 0, 0, 0, 112, 0, 108, 0, 48, 0, 50, 0, 0, 0, 112, 0, 108, 0, 48, 0, 52, 0, 0, 0, 112, 0, 108, 0, 48, 0, 54, 0, 0, 0, 112, 0, 108, 0, 48, 0, 56, 0, 0, 0, 112, 0, 108, 0, 49, 0, 48, 0, 0, 0, 112, 0, 108, 0, 49, 0, 50, 0, 0, 0, 112, 0, 108, 0, 49, 0, 52, 0, 0, 0, 112, 0, 108, 0, 49, 0, 54, 0, 0, 0, 112, 0, 108, 0, 49, 0, 56, 0, 0, 0, 112, 0, 108, 0, 50, 0, 48, 0, 0, 0, 112, 0, 108, 0, 50, 0, 50, 0, 0, 0, 112, 0, 108, 0, 50, 0, 52, 0, 0, 0, 112, 0, 108, 0, 50, 0, 54, 0, 0, 0, 112, 0, 108, 0, 50, 0, 56, 0, 0, 0, 112, 0, 108, 0, 51, 0, 48, 0, 0, 0, 112, 0, 108, 0, 51, 0, 50, 0, 0, 0, 82, 0, 83, 0, 32, 0, 77, 0, 69, 0, 0, 0, 102, 0, 114, 0, 95, 0, 70, 0, 82, 0, 0, 0, 102, 0, 114, 0, 103, 0, 101, 0, 115, 0, 0, 0, 83, 0, 65, 0, 32, 0, 73, 0, 81, 0, 0, 0, 102, 0, 114, 0, 110, 0, 97, 0, 113, 0, 0, 0, 102, 0, 114, 0, 110, 0, 111, 0, 114, 0, 0, 0, 107, 0, 107, 0, 95, 0, 75, 0, 90, 0, 0, 0, 107, 0, 121, 0, 95, 0, 75, 0, 71, 0, 0, 0, 109, 0, 110, 0, 95, 0, 77, 0, 78, 0, 0, 0, 109, 0, 115, 0, 95, 0, 77, 0, 89, 0, 0, 0, 109, 0, 121, 0, 95, 0, 77, 0, 77, 0, 0, 0, 115, 0, 119, 0, 95, 0, 67, 0, 68, 0, 0, 0, 65, 0, 81, 0, 32, 0, 84, 0, 70, 0, 0, 0, 99, 0, 121, 0, 95, 0, 71, 0, 66, 0, 0, 0, 102, 0, 97, 0, 95, 0, 97, 0, 102, 0, 0, 0, 115, 0, 110, 0, 95, 0, 90, 0, 87, 0, 0, 0, 116, 0, 119, 0, 116, 0, 110, 0, 110, 0, 0, 0, 74, 0, 69, 0, 32, 0, 71, 0, 71, 0, 0, 0, 97, 0, 107, 0, 95, 0, 71, 0, 72, 0, 0, 0, 98, 0, 101, 0, 95, 0, 66, 0, 89, 0, 0, 0, 98, 0, 103, 0, 95, 0, 66, 0, 71, 0, 0, 0, 99, 0, 115, 0, 95, 0, 67, 0, 90, 0, 0, 0, 100, 0, 97, 0, 95, 0, 68, 0, 75, 0, 0, 0, 100, 0, 118, 0, 95, 0, 77, 0, 86, 0, 0, 0, 100, 0, 122, 0, 95, 0, 66, 0, 84, 0, 0, 0, 101, 0, 108, 0, 95, 0, 71, 0, 82, 0, 0, 0, 101, 0, 116, 0, 95, 0, 69, 0, 69, 0, 0, 0, 101, 0, 117, 0, 95, 0, 69, 0, 83, 0, 0, 0, 102, 0, 111, 0, 95, 0, 70, 0, 79, 0, 0, 0, 102, 0, 114, 0, 97, 0, 114, 0, 97, 0, 0, 0, 102, 0, 114, 0, 98, 0, 102, 0, 99, 0, 0, 0, 102, 0, 114, 0, 104, 0, 100, 0, 102, 0, 0, 0, 102, 0, 114, 0, 111, 0, 99, 0, 99, 0, 0, 0, 102, 0, 121, 0, 95, 0, 78, 0, 76, 0, 0, 0, 103, 0, 97, 0, 95, 0, 73, 0, 69, 0, 0, 0, 103, 0, 110, 0, 95, 0, 80, 0, 89, 0, 0, 0, 103, 0, 118, 0, 95, 0, 73, 0, 77, 0, 0, 0, 104, 0, 97, 0, 95, 0, 78, 0, 71, 0, 0, 0, 104, 0, 101, 0, 95, 0, 73, 0, 76, 0, 0, 0, 104, 0, 114, 0, 95, 0, 72, 0, 82, 0, 0, 0, 104, 0, 117, 0, 95, 0, 72, 0, 85, 0, 0, 0, 104, 0, 121, 0, 95, 0, 65, 0, 77, 0, 0, 0, 105, 0, 115, 0, 95, 0, 73, 0, 83, 0, 0, 0, 105, 0, 116, 0, 95, 0, 73, 0, 84, 0, 0, 0, 105, 0, 117, 0, 95, 0, 67, 0, 65, 0, 0, 0, 106, 0, 97, 0, 95, 0, 74, 0, 80, 0, 0, 0, 106, 0, 118, 0, 95, 0, 73, 0, 68, 0, 0, 0, 107, 0, 108, 0, 95, 0, 71, 0, 76, 0, 0, 0, 107, 0, 109, 0, 95, 0, 75, 0, 72, 0, 0, 0, 107, 0, 111, 0, 95, 0, 75, 0, 82, 0, 0, 0, 107, 0, 115, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 117, 0, 95, 0, 84, 0, 82, 0, 0, 0, 108, 0, 98, 0, 95, 0, 76, 0, 85, 0, 0, 0, 108, 0, 111, 0, 95, 0, 76, 0, 65, 0, 0, 0, 108, 0, 116, 0, 95, 0, 76, 0, 84, 0, 0, 0, 108, 0, 118, 0, 95, 0, 76, 0, 86, 0, 0, 0, 109, 0, 103, 0, 95, 0, 77, 0, 71, 0, 0, 0, 109, 0, 105, 0, 95, 0, 78, 0, 90, 0, 0, 0, 109, 0, 107, 0, 95, 0, 77, 0, 75, 0, 0, 0, 109, 0, 115, 0, 95, 0, 66, 0, 78, 0, 0, 0, 109, 0, 116, 0, 95, 0, 77, 0, 84, 0, 0, 0, 110, 0, 101, 0, 95, 0, 78, 0, 80, 0, 0, 0, 110, 0, 110, 0, 95, 0, 78, 0, 79, 0, 0, 0, 110, 0, 121, 0, 95, 0, 77, 0, 87, 0, 0, 0, 111, 0, 115, 0, 95, 0, 71, 0, 69, 0, 0, 0, 112, 0, 115, 0, 95, 0, 65, 0, 70, 0, 0, 0, 112, 0, 116, 0, 95, 0, 66, 0, 82, 0, 0, 0, 113, 0, 117, 0, 95, 0, 80, 0, 69, 0, 0, 0, 114, 0, 110, 0, 95, 0, 66, 0, 73, 0, 0, 0, 114, 0, 111, 0, 95, 0, 82, 0, 79, 0, 0, 0, 114, 0, 119, 0, 95, 0, 82, 0, 87, 0, 0, 0, 115, 0, 103, 0, 95, 0, 67, 0, 70, 0, 0, 0, 115, 0, 105, 0, 95, 0, 76, 0, 75, 0, 0, 0, 115, 0, 107, 0, 95, 0, 83, 0, 75, 0, 0, 0, 115, 0, 108, 0, 95, 0, 83, 0, 73, 0, 0, 0, 115, 0, 111, 0, 95, 0, 83, 0, 79, 0, 0, 0, 115, 0, 113, 0, 95, 0, 65, 0, 76, 0, 0, 0, 115, 0, 114, 0, 95, 0, 77, 0, 69, 0, 0, 0, 115, 0, 118, 0, 95, 0, 83, 0, 69, 0, 0, 0, 116, 0, 103, 0, 95, 0, 84, 0, 74, 0, 0, 0, 116, 0, 104, 0, 95, 0, 84, 0, 72, 0, 0, 0, 116, 0, 107, 0, 95, 0, 84, 0, 77, 0, 0, 0, 116, 0, 111, 0, 95, 0, 84, 0, 79, 0, 0, 0, 116, 0, 116, 0, 116, 0, 111, 0, 98, 0, 0, 0, 117, 0, 103, 0, 95, 0, 67, 0, 78, 0, 0, 0, 117, 0, 107, 0, 95, 0, 85, 0, 65, 0, 0, 0, 117, 0, 114, 0, 95, 0, 80, 0, 75, 0, 0, 0, 119, 0, 97, 0, 95, 0, 66, 0, 69, 0, 0, 0, 121, 0, 111, 0, 95, 0, 78, 0, 71, 0, 0, 0, 122, 0, 117, 0, 95, 0, 90, 0, 65, 0, 0, 0, 67, 0, 90, 0, 32, 0, 83, 0, 75, 0, 0, 0, 97, 0, 97, 0, 95, 0, 69, 0, 84, 0, 0, 0, 97, 0, 102, 0, 95, 0, 90, 0, 65, 0, 0, 0, 97, 0, 109, 0, 95, 0, 69, 0, 84, 0, 0, 0, 97, 0, 110, 0, 95, 0, 69, 0, 83, 0, 0, 0, 97, 0, 115, 0, 95, 0, 73, 0, 78, 0, 0, 0, 98, 0, 97, 0, 95, 0, 82, 0, 85, 0, 0, 0, 98, 0, 109, 0, 95, 0, 77, 0, 76, 0, 0, 0, 98, 0, 110, 0, 95, 0, 66, 0, 68, 0, 0, 0, 98, 0, 111, 0, 95, 0, 67, 0, 78, 0, 0, 0, 98, 0, 114, 0, 95, 0, 70, 0, 82, 0, 0, 0, 99, 0, 97, 0, 95, 0, 69, 0, 83, 0, 0, 0, 99, 0, 101, 0, 95, 0, 82, 0, 85, 0, 0, 0, 99, 0, 111, 0, 95, 0, 70, 0, 82, 0, 0, 0, 99, 0, 117, 0, 95, 0, 82, 0, 85, 0, 0, 0, 99, 0, 118, 0, 95, 0, 82, 0, 85, 0, 0, 0, 99, 0, 122, 0, 49, 0, 49, 0, 48, 0, 0, 0, 99, 0, 122, 0, 49, 0, 49, 0, 49, 0, 0, 0, 99, 0, 122, 0, 49, 0, 49, 0, 50, 0, 0, 0, 99, 0, 122, 0, 49, 0, 49, 0, 51, 0, 0, 0, 99, 0, 122, 0, 49, 0, 49, 0, 52, 0, 0, 0, 99, 0, 122, 0, 49, 0, 49, 0, 53, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 50, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 51, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 52, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 53, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 49, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 50, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 51, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 52, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 54, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 55, 0, 0, 0, 99, 0, 122, 0, 54, 0, 54, 0, 51, 0, 0, 0, 100, 0, 101, 0, 95, 0, 68, 0, 69, 0, 0, 0, 101, 0, 101, 0, 95, 0, 71, 0, 72, 0, 0, 0, 101, 0, 110, 0, 95, 0, 85, 0, 83, 0, 0, 0, 101, 0, 115, 0, 95, 0, 69, 0, 83, 0, 0, 0, 102, 0, 97, 0, 95, 0, 65, 0, 70, 0, 0, 0, 102, 0, 97, 0, 95, 0, 73, 0, 82, 0, 0, 0, 102, 0, 105, 0, 95, 0, 70, 0, 73, 0, 0, 0, 102, 0, 114, 0, 98, 0, 114, 0, 101, 0, 0, 0, 102, 0, 114, 0, 99, 0, 111, 0, 114, 0, 0, 0, 102, 0, 114, 0, 99, 0, 118, 0, 108, 0, 0, 0, 102, 0, 114, 0, 105, 0, 100, 0, 102, 0, 0, 0, 102, 0, 114, 0, 112, 0, 97, 0, 99, 0, 0, 0, 102, 0, 114, 0, 112, 0, 100, 0, 108, 0, 0, 0, 103, 0, 100, 0, 95, 0, 71, 0, 66, 0, 0, 0, 103, 0, 108, 0, 95, 0, 69, 0, 83, 0, 0, 0, 103, 0, 117, 0, 95, 0, 73, 0, 78, 0, 0, 0, 104, 0, 97, 0, 95, 0, 71, 0, 72, 0, 0, 0, 104, 0, 97, 0, 95, 0, 78, 0, 69, 0, 0, 0, 104, 0, 105, 0, 95, 0, 73, 0, 78, 0, 0, 0, 105, 0, 100, 0, 95, 0, 73, 0, 68, 0, 0, 0, 105, 0, 103, 0, 95, 0, 78, 0, 71, 0, 0, 0, 105, 0, 105, 0, 95, 0, 67, 0, 78, 0, 0, 0, 107, 0, 97, 0, 95, 0, 71, 0, 69, 0, 0, 0, 107, 0, 105, 0, 95, 0, 75, 0, 69, 0, 0, 0, 107, 0, 110, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 119, 0, 95, 0, 71, 0, 66, 0, 0, 0, 108, 0, 103, 0, 95, 0, 85, 0, 71, 0, 0, 0, 108, 0, 110, 0, 95, 0, 67, 0, 68, 0, 0, 0, 108, 0, 117, 0, 95, 0, 67, 0, 68, 0, 0, 0, 109, 0, 108, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 114, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 115, 0, 95, 0, 83, 0, 71, 0, 0, 0, 110, 0, 98, 0, 95, 0, 78, 0, 79, 0, 0, 0, 110, 0, 100, 0, 95, 0, 90, 0, 87, 0, 0, 0, 110, 0, 108, 0, 95, 0, 78, 0, 76, 0, 0, 0, 110, 0, 114, 0, 95, 0, 90, 0, 65, 0, 0, 0, 111, 0, 99, 0, 95, 0, 70, 0, 82, 0, 0, 0, 111, 0, 109, 0, 95, 0, 69, 0, 84, 0, 0, 0, 111, 0, 114, 0, 95, 0, 73, 0, 78, 0, 0, 0, 112, 0, 108, 0, 95, 0, 80, 0, 76, 0, 0, 0, 114, 0, 109, 0, 95, 0, 67, 0, 72, 0, 0, 0, 114, 0, 117, 0, 95, 0, 82, 0, 85, 0, 0, 0, 115, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 99, 0, 95, 0, 73, 0, 84, 0, 0, 0, 115, 0, 100, 0, 95, 0, 80, 0, 75, 0, 0, 0, 115, 0, 101, 0, 95, 0, 78, 0, 79, 0, 0, 0, 115, 0, 115, 0, 95, 0, 90, 0, 65, 0, 0, 0, 115, 0, 116, 0, 95, 0, 90, 0, 65, 0, 0, 0, 115, 0, 119, 0, 95, 0, 84, 0, 90, 0, 0, 0, 116, 0, 97, 0, 95, 0, 73, 0, 78, 0, 0, 0, 116, 0, 101, 0, 95, 0, 73, 0, 78, 0, 0, 0, 116, 0, 105, 0, 95, 0, 69, 0, 84, 0, 0, 0, 116, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 116, 0, 114, 0, 95, 0, 84, 0, 82, 0, 0, 0, 116, 0, 115, 0, 95, 0, 90, 0, 65, 0, 0, 0, 116, 0, 116, 0, 95, 0, 82, 0, 85, 0, 0, 0, 116, 0, 116, 0, 109, 0, 114, 0, 99, 0, 0, 0, 116, 0, 119, 0, 107, 0, 104, 0, 104, 0, 0, 0, 116, 0, 119, 0, 110, 0, 119, 0, 116, 0, 0, 0, 116, 0, 119, 0, 116, 0, 120, 0, 103, 0, 0, 0, 118, 0, 101, 0, 95, 0, 90, 0, 65, 0, 0, 0, 118, 0, 105, 0, 95, 0, 86, 0, 78, 0, 0, 0, 119, 0, 111, 0, 95, 0, 83, 0, 78, 0, 0, 0, 120, 0, 104, 0, 95, 0, 90, 0, 65, 0, 0, 0, 108, 0, 101, 0, 103, 0, 97, 0, 99, 0, 121, 0, 0, 0, 110, 0, 100, 0, 115, 0, 95, 0, 68, 0, 69, 0, 0, 0, 98, 0, 108, 0, 116, 0, 95, 0, 86, 0, 78, 0, 0, 0, 103, 0, 101, 0, 122, 0, 95, 0, 69, 0, 84, 0, 0, 0, 116, 0, 114, 0, 118, 0, 95, 0, 84, 0, 87, 0, 0, 0, 116, 0, 122, 0, 109, 0, 95, 0, 77, 0, 65, 0, 0, 0, 97, 0, 114, 0, 110, 0, 95, 0, 67, 0, 76, 0, 0, 0, 98, 0, 101, 0, 109, 0, 95, 0, 90, 0, 77, 0, 0, 0, 98, 0, 101, 0, 122, 0, 95, 0, 84, 0, 90, 0, 0, 0, 98, 0, 114, 0, 120, 0, 95, 0, 73, 0, 78, 0, 0, 0, 99, 0, 99, 0, 112, 0, 95, 0, 66, 0, 68, 0, 0, 0, 100, 0, 121, 0, 111, 0, 95, 0, 83, 0, 78, 0, 0, 0, 102, 0, 105, 0, 108, 0, 95, 0, 80, 0, 72, 0, 0, 0, 103, 0, 115, 0, 119, 0, 95, 0, 67, 0, 72, 0, 0, 0, 103, 0, 117, 0, 122, 0, 95, 0, 75, 0, 69, 0, 0, 0, 104, 0, 97, 0, 119, 0, 95, 0, 85, 0, 83, 0, 0, 0, 105, 0, 102, 0, 101, 0, 95, 0, 84, 0, 71, 0, 0, 0, 107, 0, 97, 0, 98, 0, 95, 0, 68, 0, 90, 0, 0, 0, 107, 0, 101, 0, 97, 0, 95, 0, 67, 0, 86, 0, 0, 0, 109, 0, 102, 0, 101, 0, 95, 0, 77, 0, 85, 0, 0, 0, 109, 0, 103, 0, 104, 0, 95, 0, 77, 0, 90, 0, 0, 0, 109, 0, 121, 0, 118, 0, 95, 0, 82, 0, 85, 0, 0, 0, 109, 0, 122, 0, 110, 0, 95, 0, 73, 0, 82, 0, 0, 0, 110, 0, 97, 0, 113, 0, 95, 0, 78, 0, 65, 0, 0, 0, 110, 0, 113, 0, 111, 0, 95, 0, 71, 0, 78, 0, 0, 0, 110, 0, 117, 0, 115, 0, 95, 0, 83, 0, 83, 0, 0, 0, 113, 0, 117, 0, 99, 0, 95, 0, 71, 0, 84, 0, 0, 0, 115, 0, 101, 0, 115, 0, 95, 0, 77, 0, 76, 0, 0, 0, 115, 0, 109, 0, 115, 0, 95, 0, 70, 0, 73, 0, 0, 0, 115, 0, 115, 0, 121, 0, 95, 0, 69, 0, 82, 0, 0, 0, 115, 0, 121, 0, 114, 0, 95, 0, 73, 0, 81, 0, 0, 0, 115, 0, 122, 0, 108, 0, 95, 0, 80, 0, 76, 0, 0, 0, 116, 0, 101, 0, 111, 0, 95, 0, 85, 0, 71, 0, 0, 0, 116, 0, 119, 0, 113, 0, 95, 0, 78, 0, 69, 0, 0, 0, 119, 0, 98, 0, 112, 0, 95, 0, 65, 0, 85, 0, 0, 0, 121, 0, 97, 0, 118, 0, 95, 0, 67, 0, 77, 0, 0, 0, 97, 0, 103, 0, 113, 0, 95, 0, 67, 0, 77, 0, 0, 0, 97, 0, 114, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 97, 0, 115, 0, 97, 0, 95, 0, 84, 0, 90, 0, 0, 0, 97, 0, 115, 0, 116, 0, 95, 0, 69, 0, 83, 0, 0, 0, 98, 0, 97, 0, 115, 0, 95, 0, 67, 0, 77, 0, 0, 0, 98, 0, 103, 0, 110, 0, 95, 0, 80, 0, 75, 0, 0, 0, 98, 0, 115, 0, 115, 0, 95, 0, 67, 0, 77, 0, 0, 0, 98, 0, 121, 0, 110, 0, 95, 0, 69, 0, 82, 0, 0, 0, 99, 0, 99, 0, 104, 0, 95, 0, 78, 0, 71, 0, 0, 0, 99, 0, 101, 0, 98, 0, 95, 0, 80, 0, 72, 0, 0, 0, 99, 0, 103, 0, 103, 0, 95, 0, 85, 0, 71, 0, 0, 0, 99, 0, 104, 0, 114, 0, 95, 0, 85, 0, 83, 0, 0, 0, 99, 0, 105, 0, 99, 0, 95, 0, 85, 0, 83, 0, 0, 0, 99, 0, 107, 0, 98, 0, 95, 0, 73, 0, 81, 0, 0, 0, 100, 0, 97, 0, 118, 0, 95, 0, 75, 0, 69, 0, 0, 0, 100, 0, 106, 0, 101, 0, 95, 0, 78, 0, 69, 0, 0, 0, 100, 0, 115, 0, 98, 0, 95, 0, 68, 0, 69, 0, 0, 0, 100, 0, 117, 0, 97, 0, 95, 0, 67, 0, 77, 0, 0, 0, 101, 0, 98, 0, 117, 0, 95, 0, 75, 0, 69, 0, 0, 0, 101, 0, 111, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 101, 0, 119, 0, 111, 0, 95, 0, 67, 0, 77, 0, 0, 0, 102, 0, 117, 0, 114, 0, 95, 0, 73, 0, 84, 0, 0, 0, 103, 0, 97, 0, 97, 0, 95, 0, 71, 0, 72, 0, 0, 0, 104, 0, 115, 0, 98, 0, 95, 0, 68, 0, 69, 0, 0, 0, 105, 0, 97, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 105, 0, 111, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 106, 0, 103, 0, 111, 0, 95, 0, 67, 0, 77, 0, 0, 0, 106, 0, 109, 0, 99, 0, 95, 0, 84, 0, 90, 0, 0, 0, 107, 0, 97, 0, 106, 0, 95, 0, 78, 0, 71, 0, 0, 0, 107, 0, 97, 0, 109, 0, 95, 0, 75, 0, 69, 0, 0, 0, 107, 0, 99, 0, 103, 0, 95, 0, 78, 0, 71, 0, 0, 0, 107, 0, 100, 0, 101, 0, 95, 0, 84, 0, 90, 0, 0, 0, 107, 0, 101, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 107, 0, 104, 0, 113, 0, 95, 0, 77, 0, 76, 0, 0, 0, 107, 0, 107, 0, 106, 0, 95, 0, 67, 0, 77, 0, 0, 0, 107, 0, 108, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 107, 0, 111, 0, 107, 0, 95, 0, 73, 0, 78, 0, 0, 0, 107, 0, 112, 0, 101, 0, 95, 0, 76, 0, 82, 0, 0, 0, 107, 0, 115, 0, 98, 0, 95, 0, 84, 0, 90, 0, 0, 0, 107, 0, 115, 0, 102, 0, 95, 0, 67, 0, 77, 0, 0, 0, 107, 0, 115, 0, 104, 0, 95, 0, 68, 0, 69, 0, 0, 0, 108, 0, 97, 0, 103, 0, 95, 0, 84, 0, 90, 0, 0, 0, 108, 0, 107, 0, 116, 0, 95, 0, 85, 0, 83, 0, 0, 0, 108, 0, 114, 0, 99, 0, 95, 0, 73, 0, 82, 0, 0, 0, 108, 0, 117, 0, 111, 0, 95, 0, 75, 0, 69, 0, 0, 0, 108, 0, 117, 0, 121, 0, 95, 0, 75, 0, 69, 0, 0, 0, 109, 0, 97, 0, 115, 0, 95, 0, 75, 0, 69, 0, 0, 0, 109, 0, 101, 0, 114, 0, 95, 0, 75, 0, 69, 0, 0, 0, 109, 0, 103, 0, 111, 0, 95, 0, 67, 0, 77, 0, 0, 0, 109, 0, 110, 0, 105, 0, 95, 0, 73, 0, 78, 0, 0, 0, 109, 0, 111, 0, 104, 0, 95, 0, 67, 0, 65, 0, 0, 0, 109, 0, 117, 0, 97, 0, 95, 0, 67, 0, 77, 0, 0, 0, 109, 0, 117, 0, 115, 0, 95, 0, 85, 0, 83, 0, 0, 0, 110, 0, 109, 0, 103, 0, 95, 0, 67, 0, 77, 0, 0, 0, 110, 0, 110, 0, 104, 0, 95, 0, 67, 0, 77, 0, 0, 0, 110, 0, 115, 0, 111, 0, 95, 0, 90, 0, 65, 0, 0, 0, 110, 0, 121, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 111, 0, 115, 0, 97, 0, 95, 0, 85, 0, 83, 0, 0, 0, 114, 0, 111, 0, 102, 0, 95, 0, 84, 0, 90, 0, 0, 0, 114, 0, 119, 0, 107, 0, 95, 0, 84, 0, 90, 0, 0, 0, 115, 0, 97, 0, 104, 0, 95, 0, 82, 0, 85, 0, 0, 0, 115, 0, 97, 0, 113, 0, 95, 0, 75, 0, 69, 0, 0, 0, 115, 0, 98, 0, 112, 0, 95, 0, 84, 0, 90, 0, 0, 0, 115, 0, 99, 0, 110, 0, 95, 0, 73, 0, 84, 0, 0, 0, 115, 0, 100, 0, 104, 0, 95, 0, 73, 0, 82, 0, 0, 0, 115, 0, 101, 0, 104, 0, 95, 0, 77, 0, 90, 0, 0, 0, 115, 0, 105, 0, 100, 0, 95, 0, 69, 0, 84, 0, 0, 0, 115, 0, 109, 0, 97, 0, 95, 0, 83, 0, 69, 0, 0, 0, 115, 0, 109, 0, 106, 0, 95, 0, 83, 0, 69, 0, 0, 0, 115, 0, 109, 0, 110, 0, 95, 0, 70, 0, 73, 0, 0, 0, 116, 0, 105, 0, 103, 0, 95, 0, 69, 0, 82, 0, 0, 0, 118, 0, 111, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 118, 0, 117, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 119, 0, 97, 0, 101, 0, 95, 0, 67, 0, 72, 0, 0, 0, 119, 0, 97, 0, 108, 0, 95, 0, 69, 0, 84, 0, 0, 0, 120, 0, 111, 0, 103, 0, 95, 0, 85, 0, 71, 0, 0, 0, 121, 0, 105, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 122, 0, 103, 0, 104, 0, 95, 0, 77, 0, 65, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 112, 0, 97, 0, 95, 0, 71, 0, 117, 0, 114, 0, 117, 0, 0, 0, 117, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 48, 0, 51, 0, 52, 0, 32, 0, 49, 0, 52, 0, 51, 0, 0, 0, 65, 0, 76, 0, 65, 0, 76, 0, 67, 0, 57, 0, 55, 0, 0, 0, 80, 0, 79, 0, 76, 0, 89, 0, 84, 0, 79, 0, 78, 0, 0, 0, 97, 0, 108, 0, 97, 0, 108, 0, 99, 0, 57, 0, 55, 0, 0, 0, 97, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 98, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 102, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 106, 0, 98, 0, 111, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 112, 0, 114, 0, 103, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 0, 0, 115, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 0, 0, 111, 0, 118, 0, 101, 0, 114, 0, 108, 0, 111, 0, 110, 0, 103, 0, 0, 0, 67, 0, 87, 0, 32, 0, 83, 0, 88, 0, 32, 0, 66, 0, 81, 0, 0, 0, 118, 0, 97, 0, 105, 0, 95, 0, 86, 0, 97, 0, 105, 0, 105, 0, 0, 0, 115, 0, 104, 0, 105, 0, 95, 0, 84, 0, 102, 0, 110, 0, 103, 0, 0, 0, 121, 0, 117, 0, 101, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 0, 0, 111, 0, 109, 0, 98, 0, 106, 0, 32, 0, 111, 0, 109, 0, 98, 0, 115, 0, 0, 0, 111, 0, 109, 0, 115, 0, 106, 0, 32, 0, 111, 0, 109, 0, 115, 0, 115, 0, 0, 0, 100, 0, 101, 0, 112, 0, 114, 0, 101, 0, 99, 0, 97, 0, 116, 0, 101, 0, 100, 0, 0, 0, 97, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 65, 0, 90, 0, 0, 0, 98, 0, 115, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 66, 0, 65, 0, 0, 0, 117, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 85, 0, 90, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 112, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 80, 0, 75, 0, 0, 0, 112, 0, 97, 0, 95, 0, 71, 0, 117, 0, 114, 0, 117, 0, 95, 0, 73, 0, 78, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 82, 0, 83, 0, 0, 0, 117, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 65, 0, 70, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 84, 0, 87, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 72, 0, 75, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 77, 0, 79, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 88, 0, 75, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 82, 0, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 83, 0, 71, 0, 0, 0, 97, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 73, 0, 82, 0, 0, 0, 97, 0, 122, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 65, 0, 90, 0, 0, 0, 98, 0, 109, 0, 95, 0, 78, 0, 107, 0, 111, 0, 111, 0, 95, 0, 77, 0, 76, 0, 0, 0, 98, 0, 115, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 66, 0, 65, 0, 0, 0, 101, 0, 110, 0, 95, 0, 68, 0, 115, 0, 114, 0, 116, 0, 95, 0, 85, 0, 83, 0, 0, 0, 102, 0, 102, 0, 95, 0, 65, 0, 100, 0, 108, 0, 109, 0, 95, 0, 71, 0, 78, 0, 0, 0, 102, 0, 102, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 83, 0, 78, 0, 0, 0, 104, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 78, 0, 71, 0, 0, 0, 105, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 109, 0, 110, 0, 95, 0, 77, 0, 111, 0, 110, 0, 103, 0, 95, 0, 67, 0, 78, 0, 0, 0, 109, 0, 115, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 95, 0, 77, 0, 89, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 66, 0, 65, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 69, 0, 0, 0, 115, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 73, 0, 68, 0, 0, 0, 117, 0, 122, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 95, 0, 85, 0, 90, 0, 0, 0, 70, 0, 77, 0, 32, 0, 77, 0, 72, 0, 32, 0, 77, 0, 80, 0, 32, 0, 80, 0, 87, 0, 0, 0, 115, 0, 104, 0, 105, 0, 95, 0, 84, 0, 102, 0, 110, 0, 103, 0, 95, 0, 77, 0, 65, 0, 0, 0, 118, 0, 97, 0, 105, 0, 95, 0, 86, 0, 97, 0, 105, 0, 105, 0, 95, 0, 76, 0, 82, 0, 0, 0, 121, 0, 117, 0, 101, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 95, 0, 67, 0, 78, 0, 0, 0, 121, 0, 117, 0, 101, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 72, 0, 75, 0, 0, 0, 115, 0, 104, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 77, 0, 65, 0, 0, 0, 118, 0, 97, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 95, 0, 76, 0, 82, 0, 0, 0, 110, 0, 97, 0, 110, 0, 45, 0, 120, 0, 45, 0, 122, 0, 104, 0, 45, 0, 109, 0, 105, 0, 110, 0, 0, 0, 109, 0, 97, 0, 99, 0, 114, 0, 111, 0, 108, 0, 97, 0, 110, 0, 103, 0, 117, 0, 97, 0, 103, 0, 101, 0, 0, 0, 98, 0, 105, 0, 98, 0, 108, 0, 105, 0, 111, 0, 103, 0, 114, 0, 97, 0, 112, 0, 104, 0, 105, 0, 99, 0, 0, 0, 115, 0, 101, 0, 101, 0, 45, 0, 120, 0, 45, 0, 105, 0, 45, 0, 109, 0, 105, 0, 110, 0, 103, 0, 111, 0, 0, 0, 101, 0, 110, 0, 45, 0, 120, 0, 45, 0, 105, 0, 45, 0, 100, 0, 101, 0, 102, 0, 97, 0, 117, 0, 108, 0, 116, 0, 0, 0, 108, 0, 117, 0, 101, 0, 99, 0, 32, 0, 108, 0, 117, 0, 103, 0, 114, 0, 32, 0, 108, 0, 117, 0, 114, 0, 109, 0, 0, 0, 109, 0, 114, 0, 49, 0, 51, 0, 32, 0, 109, 0, 114, 0, 49, 0, 52, 0, 32, 0, 109, 0, 114, 0, 49, 0, 53, 0, 0, 0, 117, 0, 110, 0, 100, 0, 45, 0, 120, 0, 45, 0, 105, 0, 45, 0, 101, 0, 110, 0, 111, 0, 99, 0, 104, 0, 105, 0, 97, 0, 110, 0, 0, 0, 82, 0, 83, 0, 32, 0, 77, 0, 69, 0, 32, 0, 83, 0, 73, 0, 32, 0, 72, 0, 82, 0, 32, 0, 77, 0, 75, 0, 32, 0, 66, 0, 65, 0, 0, 0, 120, 0, 116, 0, 103, 0, 45, 0, 120, 0, 45, 0, 99, 0, 101, 0, 108, 0, 45, 0, 103, 0, 97, 0, 117, 0, 108, 0, 105, 0, 115, 0, 104, 0, 0, 0, 108, 0, 117, 0, 99, 0, 97, 0, 32, 0, 108, 0, 117, 0, 101, 0, 115, 0, 32, 0, 108, 0, 117, 0, 108, 0, 117, 0, 32, 0, 108, 0, 117, 0, 109, 0, 101, 0, 0, 0, 108, 0, 117, 0, 99, 0, 108, 0, 32, 0, 108, 0, 117, 0, 100, 0, 105, 0, 32, 0, 108, 0, 117, 0, 114, 0, 100, 0, 32, 0, 108, 0, 117, 0, 118, 0, 100, 0, 32, 0, 108, 0, 117, 0, 119, 0, 105, 0, 0, 0, 82, 0, 85, 0, 32, 0, 65, 0, 77, 0, 32, 0, 65, 0, 90, 0, 32, 0, 66, 0, 89, 0, 32, 0, 71, 0, 69, 0, 32, 0, 75, 0, 71, 0, 32, 0, 75, 0, 90, 0, 32, 0, 77, 0, 68, 0, 32, 0, 84, 0, 74, 0, 32, 0, 84, 0, 77, 0, 32, 0, 85, 0, 65, 0, 32, 0, 85, 0, 90, 0, 0, 0, 41, 220, 110, 0, 122, 0, 99, 0, 97, 0, 110, 0, 32, 0, 110, 0, 122, 0, 109, 0, 98, 0, 104, 0, 32, 0, 110, 0, 122, 0, 110, 0, 115, 0, 110, 0, 32, 0, 110, 0, 122, 0, 111, 0, 116, 0, 97, 0, 32, 0, 110, 0, 122, 0, 115, 0, 116, 0, 108, 0, 32, 0, 110, 0, 122, 0, 116, 0, 97, 0, 115, 0, 32, 0, 110, 0, 122, 0, 119, 0, 116, 0, 99, 0, 0, 0, 44, 220, 82, 0, 85, 0, 32, 0, 65, 0, 77, 0, 32, 0, 65, 0, 90, 0, 32, 0, 66, 0, 89, 0, 32, 0, 69, 0, 69, 0, 32, 0, 71, 0, 69, 0, 32, 0, 75, 0, 90, 0, 32, 0, 75, 0, 71, 0, 32, 0, 76, 0, 86, 0, 32, 0, 76, 0, 84, 0, 32, 0, 77, 0, 68, 0, 32, 0, 84, 0, 74, 0, 32, 0, 84, 0, 77, 0, 32, 0, 85, 0, 65, 0, 32, 0, 85, 0, 90, 0, 0, 0, 53, 220, 110, 0, 122, 0, 97, 0, 117, 0, 107, 0, 32, 0, 110, 0, 122, 0, 98, 0, 111, 0, 112, 0, 32, 0, 110, 0, 122, 0, 103, 0, 105, 0, 115, 0, 32, 0, 110, 0, 122, 0, 104, 0, 107, 0, 98, 0, 32, 0, 110, 0, 122, 0, 109, 0, 119, 0, 116, 0, 32, 0, 110, 0, 122, 0, 110, 0, 116, 0, 108, 0, 32, 0, 110, 0, 122, 0, 116, 0, 107, 0, 105, 0, 32, 0, 110, 0, 122, 0, 119, 0, 103, 0, 110, 0, 32, 0, 110, 0, 122, 0, 119, 0, 107, 0, 111, 0, 0, 0, 2, 0, 56, 0, 63, 0, 183, 14, 6, 5, 2, 0, 56, 0, 63, 0, 183, 14, 26, 4, 2, 0, 56, 0, 63, 0, 118, 14, 131, 2, 2, 0, 56, 0, 63, 0, 118, 14, 134, 2, 2, 0, 56, 0, 63, 0, 183, 14, 175, 3, 2, 0, 56, 0, 63, 0, 118, 14, 213, 6, 2, 0, 56, 0, 63, 0, 98, 16, 150, 4, 2, 0, 56, 0, 63, 0, 118, 14, 135, 3, 2, 0, 56, 0, 63, 0, 112, 16, 112, 0, 2, 0, 56, 0, 63, 0, 98, 16, 112, 0, 2, 0, 56, 0, 63, 0, 118, 14, 19, 4, 2, 0, 56, 0, 63, 0, 118, 14, 23, 2, 2, 0, 56, 0, 63, 0, 98, 16, 23, 2, 2, 0, 56, 0, 63, 0, 118, 14, 235, 3, 2, 0, 56, 0, 63, 0, 112, 16, 106, 0, 2, 0, 56, 0, 63, 0, 183, 14, 170, 3, 2, 0, 56, 0, 63, 0, 183, 14, 254, 4, 2, 0, 56, 0, 63, 0, 118, 14, 251, 3, 2, 0, 56, 0, 63, 0, 183, 14, 166, 4, 2, 0, 56, 0, 63, 0, 118, 14, 140, 2, 2, 0, 56, 0, 63, 0, 118, 14, 137, 2, 2, 0, 56, 0, 63, 0, 118, 14, 183, 3, 2, 0, 56, 0, 63, 0, 98, 16, 183, 3, 2, 0, 56, 0, 63, 0, 183, 14, 206, 4, 2, 0, 56, 0, 63, 0, 224, 10, 194, 14, 2, 0, 56, 0, 63, 0, 118, 14, 26, 2, 2, 0, 56, 0, 63, 0, 98, 16, 26, 2, 2, 0, 56, 0, 63, 0, 118, 14, 227, 3, 2, 0, 56, 0, 63, 0, 118, 14, 152, 2, 2, 0, 56, 0, 63, 0, 112, 16, 41, 2, 2, 0, 56, 0, 63, 0, 98, 16, 42, 4, 2, 0, 56, 0, 63, 0, 98, 16, 158, 3, 2, 0, 56, 0, 63, 0, 118, 14, 143, 2, 2, 0, 56, 0, 63, 0, 118, 14, 155, 2, 2, 0, 56, 0, 63, 0, 183, 14, 46, 4, 2, 0, 56, 0, 63, 0, 98, 16, 146, 3, 2, 0, 56, 0, 63, 0, 118, 14, 146, 3, 2, 0, 56, 0, 63, 0, 118, 14, 149, 2, 2, 0, 56, 0, 63, 0, 183, 14, 94, 4, 2, 0, 56, 0, 63, 0, 118, 14, 171, 3, 2, 0, 56, 0, 63, 0, 118, 14, 170, 14, 2, 0, 56, 0, 63, 0, 118, 14, 239, 3, 2, 0, 56, 0, 63, 0, 224, 10, 205, 14, 2, 0, 56, 0, 63, 0, 118, 14, 146, 2, 2, 0, 56, 0, 63, 0, 112, 16, 92, 2, 2, 0, 56, 0, 63, 0, 98, 16, 186, 4, 2, 0, 56, 0, 63, 0, 98, 16, 62, 4, 2, 0, 56, 0, 63, 0, 118, 14, 158, 2, 2, 0, 56, 0, 63, 0, 183, 14, 150, 3, 2, 0, 56, 0, 63, 0, 224, 10, 220, 16, 2, 0, 56, 0, 63, 0, 118, 14, 32, 2, 2, 0, 56, 0, 63, 0, 118, 14, 11, 4, 2, 0, 56, 0, 63, 0, 118, 14, 161, 2, 2, 0, 56, 0, 63, 0, 112, 16, 19, 0, 2, 0, 56, 0, 63, 0, 118, 14, 167, 2, 2, 0, 56, 0, 63, 0, 118, 14, 170, 2, 2, 0, 56, 0, 63, 0, 183, 14, 190, 4, 2, 0, 56, 0, 63, 0, 183, 14, 218, 3, 2, 0, 56, 0, 63, 0, 98, 16, 14, 5, 2, 0, 56, 0, 63, 0, 183, 14, 10, 4, 2, 0, 56, 0, 63, 0, 98, 16, 19, 0, 2, 0, 56, 0, 63, 0, 224, 10, 76, 8, 2, 0, 56, 0, 63, 0, 118, 14, 242, 2, 2, 0, 56, 0, 63, 0, 118, 14, 164, 2, 2, 0, 56, 0, 63, 0, 183, 14, 210, 3, 2, 0, 56, 0, 63, 0, 183, 14, 226, 4, 2, 0, 56, 0, 63, 0, 118, 14, 29, 2, 2, 0, 56, 0, 63, 0, 98, 16, 29, 2, 2, 0, 56, 0, 63, 0, 118, 14, 228, 10, 2, 0, 56, 0, 63, 0, 112, 16, 32, 2, 2, 0, 56, 0, 63, 0, 118, 14, 173, 2, 2, 0, 56, 0, 63, 0, 118, 14, 35, 2, 2, 0, 56, 0, 63, 0, 98, 16, 90, 4, 2, 0, 56, 0, 63, 0, 98, 16, 198, 4, 2, 0, 56, 0, 63, 0, 98, 16, 86, 4, 2, 0, 56, 0, 63, 0, 98, 16, 190, 3, 2, 0, 56, 0, 63, 0, 183, 14, 82, 4, 2, 0, 56, 0, 63, 0, 118, 14, 176, 2, 2, 0, 56, 0, 63, 0, 183, 14, 64, 5, 2, 0, 56, 0, 63, 0, 183, 14, 210, 6, 2, 0, 56, 0, 63, 0, 112, 16, 98, 2, 2, 0, 56, 0, 63, 0, 118, 14, 175, 3, 2, 0, 56, 0, 63, 0, 98, 16, 38, 2, 2, 0, 56, 0, 63, 0, 118, 14, 163, 3, 2, 0, 56, 0, 63, 0, 98, 16, 202, 3, 2, 0, 56, 0, 63, 0, 118, 14, 223, 3, 2, 0, 56, 0, 63, 0, 118, 14, 7, 4, 2, 0, 56, 0, 63, 0, 98, 16, 159, 3, 2, 0, 56, 0, 63, 0, 118, 14, 38, 2, 2, 0, 56, 0, 63, 0, 118, 14, 41, 2, 2, 0, 56, 0, 63, 0, 118, 14, 179, 2, 2, 0, 56, 0, 63, 0, 118, 14, 185, 2, 2, 0, 56, 0, 63, 0, 118, 14, 103, 0, 2, 0, 56, 0, 63, 0, 98, 16, 135, 3, 2, 0, 56, 0, 63, 0, 118, 14, 79, 5, 2, 0, 56, 0, 63, 0, 118, 14, 182, 2, 2, 0, 56, 0, 63, 0, 118, 14, 47, 2, 2, 0, 56, 0, 63, 0, 112, 16, 47, 2, 2, 0, 56, 0, 63, 0, 118, 14, 215, 3, 2, 0, 56, 0, 63, 0, 98, 16, 44, 2, 2, 0, 56, 0, 63, 0, 118, 14, 44, 2, 2, 0, 56, 0, 63, 0, 183, 14, 78, 4, 2, 0, 56, 0, 63, 0, 98, 16, 187, 3, 2, 0, 56, 0, 63, 0, 98, 16, 114, 4, 2, 0, 56, 0, 63, 0, 112, 16, 65, 2, 2, 0, 56, 0, 63, 0, 112, 16, 35, 2, 2, 0, 56, 0, 63, 0, 183, 14, 142, 3, 2, 0, 56, 0, 63, 0, 183, 14, 118, 4, 2, 0, 56, 0, 63, 0, 118, 14, 84, 5, 2, 0, 56, 0, 63, 0, 118, 14, 188, 2, 2, 0, 56, 0, 63, 0, 118, 14, 191, 2, 2, 0, 56, 0, 63, 0, 118, 14, 243, 3, 2, 0, 56, 0, 63, 0, 98, 16, 110, 4, 2, 0, 56, 0, 63, 0, 112, 16, 163, 3, 2, 0, 56, 0, 63, 0, 118, 14, 50, 2, 2, 0, 56, 0, 63, 0, 183, 14, 210, 4, 2, 0, 56, 0, 63, 0, 98, 16, 50, 2, 2, 0, 56, 0, 63, 0, 118, 14, 194, 2, 2, 0, 56, 0, 63, 0, 183, 14, 166, 3, 2, 0, 56, 0, 63, 0, 98, 16, 226, 3, 2, 0, 56, 0, 63, 0, 224, 10, 192, 9, 2, 0, 56, 0, 63, 0, 224, 10, 198, 9, 2, 0, 56, 0, 63, 0, 224, 10, 92, 7, 2, 0, 56, 0, 63, 0, 118, 14, 197, 2, 2, 0, 56, 0, 63, 0, 118, 14, 89, 5, 2, 0, 56, 0, 63, 0, 118, 14, 254, 13, 2, 0, 56, 0, 63, 0, 98, 16, 122, 4, 2, 0, 56, 0, 63, 0, 98, 16, 130, 4, 2, 0, 56, 0, 63, 0, 118, 14, 69, 5, 2, 0, 56, 0, 63, 0, 118, 14, 203, 2, 2, 0, 56, 0, 63, 0, 98, 16, 2, 5, 2, 0, 56, 0, 63, 0, 118, 14, 94, 5, 2, 0, 56, 0, 63, 0, 118, 14, 147, 3, 2, 0, 56, 0, 63, 0, 183, 14, 146, 4, 2, 0, 56, 0, 63, 0, 118, 14, 179, 3, 2, 0, 56, 0, 63, 0, 118, 14, 200, 2, 2, 0, 56, 0, 63, 0, 118, 14, 106, 0, 2, 0, 56, 0, 63, 0, 224, 10, 140, 16, 2, 0, 56, 0, 63, 0, 224, 10, 185, 16, 2, 0, 56, 0, 63, 0, 224, 10, 126, 16, 2, 0, 56, 0, 63, 0, 183, 14, 154, 3, 2, 0, 56, 0, 63, 0, 183, 14, 58, 4, 2, 0, 56, 0, 63, 0, 183, 14, 198, 3, 2, 0, 56, 0, 63, 0, 183, 14, 30, 5, 2, 0, 56, 0, 63, 0, 183, 14, 80, 2, 2, 0, 56, 0, 63, 0, 183, 14, 101, 2, 2, 0, 56, 0, 63, 0, 183, 14, 222, 4, 2, 0, 56, 0, 63, 0, 183, 14, 22, 5, 2, 0, 56, 0, 63, 0, 183, 14, 182, 3, 2, 0, 56, 0, 63, 0, 183, 14, 2, 4, 2, 0, 56, 0, 63, 0, 183, 14, 246, 3, 2, 0, 56, 0, 63, 0, 118, 14, 212, 2, 2, 0, 56, 0, 63, 0, 112, 16, 56, 2, 2, 0, 56, 0, 63, 0, 118, 14, 215, 2, 2, 0, 56, 0, 63, 0, 118, 14, 142, 14, 2, 0, 56, 0, 63, 0, 98, 16, 59, 2, 2, 0, 56, 0, 63, 0, 118, 14, 59, 2, 2, 0, 56, 0, 63, 0, 118, 14, 209, 2, 2, 0, 56, 0, 63, 0, 183, 14, 102, 4, 2, 0, 56, 0, 63, 0, 183, 14, 53, 2, 2, 0, 56, 0, 63, 0, 118, 14, 206, 2, 2, 0, 56, 0, 63, 0, 118, 14, 53, 2, 2, 0, 56, 0, 63, 0, 118, 14, 159, 3, 2, 0, 56, 0, 63, 0, 118, 14, 56, 2, 2, 0, 56, 0, 63, 0, 118, 14, 218, 2, 2, 0, 56, 0, 63, 0, 183, 14, 69, 5, 2, 0, 56, 0, 63, 0, 118, 14, 62, 2, 2, 0, 56, 0, 63, 0, 183, 14, 138, 3, 2, 0, 56, 0, 63, 0, 183, 14, 115, 0, 2, 0, 56, 0, 63, 0, 118, 14, 221, 2, 2, 0, 56, 0, 63, 0, 183, 14, 62, 2, 2, 0, 56, 0, 63, 0, 118, 14, 230, 2, 2, 0, 56, 0, 63, 0, 118, 14, 236, 2, 2, 0, 56, 0, 63, 0, 118, 14, 239, 2, 2, 0, 56, 0, 63, 0, 118, 14, 65, 2, 2, 0, 56, 0, 63, 0, 118, 14, 71, 2, 2, 0, 56, 0, 63, 0, 118, 14, 227, 2, 2, 0, 56, 0, 63, 0, 183, 14, 26, 5, 2, 0, 56, 0, 63, 0, 183, 14, 230, 3, 2, 0, 56, 0, 63, 0, 98, 16, 64, 5, 2, 0, 56, 0, 63, 0, 118, 14, 233, 2, 2, 0, 56, 0, 63, 0, 118, 14, 151, 3, 2, 0, 56, 0, 63, 0, 118, 14, 37, 3, 2, 0, 56, 0, 63, 0, 118, 14, 245, 2, 2, 0, 56, 0, 63, 0, 224, 10, 162, 6, 2, 0, 56, 0, 63, 0, 98, 16, 74, 2, 2, 0, 56, 0, 63, 0, 98, 16, 71, 2, 2, 0, 56, 0, 63, 0, 98, 16, 68, 2, 2, 0, 56, 0, 63, 0, 98, 16, 158, 4, 2, 0, 56, 0, 63, 0, 183, 14, 170, 4, 2, 0, 56, 0, 63, 0, 118, 14, 77, 2, 2, 0, 56, 0, 63, 0, 118, 14, 68, 2, 2, 0, 56, 0, 63, 0, 118, 14, 204, 17, 2, 0, 56, 0, 63, 0, 98, 16, 77, 2, 2, 0, 56, 0, 63, 0, 183, 14, 50, 4, 2, 0, 56, 0, 63, 0, 224, 10, 170, 7, 2, 0, 56, 0, 63, 0, 183, 14, 130, 3, 2, 0, 56, 0, 63, 0, 118, 14, 224, 2, 2, 0, 56, 0, 63, 0, 118, 14, 74, 2, 2, 0, 56, 0, 63, 0, 183, 14, 106, 4, 2, 0, 56, 0, 63, 0, 183, 14, 18, 4, 2, 0, 56, 0, 63, 0, 183, 14, 38, 5, 2, 0, 56, 0, 63, 0, 224, 10, 168, 6, 2, 0, 56, 0, 63, 0, 183, 14, 130, 3, 2, 0, 56, 0, 63, 0, 183, 14, 130, 3, 2, 0, 56, 0, 63, 0, 118, 14, 1, 3, 2, 0, 56, 0, 63, 0, 118, 14, 248, 2, 2, 0, 56, 0, 63, 0, 118, 14, 83, 2, 2, 0, 56, 0, 63, 0, 98, 16, 54, 4, 2, 0, 56, 0, 63, 0, 183, 14, 234, 4, 2, 0, 56, 0, 63, 0, 118, 14, 254, 2, 2, 0, 56, 0, 63, 0, 118, 14, 99, 5, 2, 0, 56, 0, 63, 0, 118, 14, 152, 16, 2, 0, 56, 0, 63, 0, 183, 14, 202, 4, 2, 0, 56, 0, 63, 0, 183, 14, 238, 4, 2, 0, 56, 0, 63, 0, 118, 14, 80, 2, 2, 0, 56, 0, 63, 0, 118, 14, 4, 3, 2, 0, 56, 0, 63, 0, 118, 14, 251, 2, 2, 0, 56, 0, 63, 0, 98, 16, 83, 2, 2, 0, 56, 0, 63, 0, 112, 16, 89, 2, 2, 0, 56, 0, 63, 0, 118, 14, 255, 3, 2, 0, 56, 0, 63, 0, 118, 14, 231, 3, 2, 0, 56, 0, 63, 0, 112, 16, 155, 3, 2, 0, 56, 0, 63, 0, 118, 14, 219, 3, 2, 0, 56, 0, 63, 0, 112, 16, 109, 0, 2, 0, 56, 0, 63, 0, 183, 14, 74, 4, 2, 0, 56, 0, 63, 0, 98, 16, 70, 4, 2, 0, 56, 0, 63, 0, 118, 14, 89, 2, 2, 0, 56, 0, 63, 0, 118, 14, 86, 2, 2, 0, 56, 0, 63, 0, 118, 14, 7, 3, 2, 0, 56, 0, 63, 0, 224, 10, 174, 6, 2, 0, 56, 0, 63, 0, 98, 16, 202, 3, 2, 0, 56, 0, 63, 0, 183, 14, 16, 0, 2, 0, 56, 0, 63, 0, 118, 14, 16, 0, 2, 0, 56, 0, 63, 0, 118, 14, 64, 5, 2, 0, 56, 0, 63, 0, 118, 14, 155, 3, 2, 0, 56, 0, 63, 0, 224, 10, 224, 7, 2, 0, 56, 0, 63, 0, 224, 10, 180, 6, 2, 0, 56, 0, 63, 0, 224, 10, 26, 10, 2, 0, 56, 0, 63, 0, 118, 14, 109, 0, 2, 0, 56, 0, 63, 0, 183, 14, 206, 3, 2, 0, 56, 0, 63, 0, 98, 16, 230, 4, 2, 0, 56, 0, 63, 0, 183, 14, 142, 3, 2, 0, 56, 0, 63, 0, 118, 14, 92, 2, 2, 0, 56, 0, 63, 0, 183, 14, 38, 4, 2, 0, 56, 0, 63, 0, 183, 14, 206, 3, 2, 0, 56, 0, 63, 0, 183, 14, 14, 4, 2, 0, 56, 0, 63, 0, 118, 14, 10, 3, 2, 0, 56, 0, 63, 0, 118, 14, 101, 2, 2, 0, 56, 0, 63, 0, 118, 14, 16, 3, 2, 0, 56, 0, 63, 0, 183, 14, 174, 3, 2, 0, 56, 0, 63, 0, 118, 14, 13, 3, 2, 0, 56, 0, 63, 0, 118, 14, 124, 14, 2, 0, 56, 0, 63, 0, 118, 14, 95, 2, 2, 0, 56, 0, 63, 0, 118, 14, 98, 2, 2, 0, 56, 0, 63, 0, 118, 14, 225, 6, 2, 0, 56, 0, 63, 0, 183, 14, 238, 3, 2, 0, 56, 0, 63, 0, 183, 14, 242, 3, 2, 0, 56, 0, 63, 0, 224, 10, 7, 0, 2, 0, 56, 0, 63, 0, 183, 14, 7, 0, 2, 0, 56, 0, 63, 0, 183, 14, 225, 6, 2, 0, 56, 0, 63, 0, 183, 14, 7, 0, 2, 0, 56, 0, 63, 0, 183, 14, 225, 6, 2, 0, 56, 0, 63, 0, 118, 14, 7, 0, 2, 0, 56, 0, 63, 0, 118, 14, 7, 0, 2, 0, 56, 0, 63, 0, 98, 16, 95, 2, 2, 0, 56, 0, 63, 0, 183, 14, 210, 3, 2, 0, 56, 0, 63, 0, 118, 14, 15, 4, 2, 0, 56, 0, 63, 0, 118, 14, 19, 3, 2, 0, 56, 0, 63, 0, 98, 16, 104, 2, 2, 0, 56, 0, 63, 0, 118, 14, 104, 2, 2, 0, 56, 0, 63, 0, 118, 14, 159, 6, 2, 0, 56, 0, 63, 0, 118, 14, 187, 3, 2, 0, 56, 0, 63, 0, 98, 16, 159, 6, 2, 0, 56, 0, 63, 0, 118, 14, 22, 3, 2, 0, 56, 0, 63, 0, 183, 14, 142, 3, 2, 0, 56, 0, 63, 0, 224, 10, 249, 14, 2, 0, 56, 0, 63, 0, 224, 10, 238, 14, 2, 0, 56, 0, 63, 0, 118, 14, 247, 3, 2, 0, 56, 0, 63, 0, 98, 16, 107, 2, 2, 0, 56, 0, 63, 0, 183, 14, 34, 4, 2, 0, 56, 0, 63, 0, 112, 16, 103, 0, 2, 0, 56, 0, 63, 0, 98, 16, 103, 0, 2, 0, 56, 0, 63, 0, 118, 14, 25, 3, 2, 0, 56, 0, 63, 0, 98, 16, 86, 2, 2, 0, 56, 0, 63, 0, 183, 14, 138, 4, 2, 0, 56, 0, 63, 0, 183, 14, 178, 3, 2, 0, 56, 0, 63, 0, 98, 16, 174, 4, 2, 0, 56, 0, 63, 0, 118, 14, 28, 3, 2, 0, 56, 0, 63, 0, 118, 14, 31, 3, 2, 0, 56, 0, 63, 0, 183, 14, 214, 3, 2, 0, 56, 0, 63, 0, 183, 14, 178, 4, 2, 0, 56, 0, 63, 0, 118, 14, 120, 9, 2, 0, 56, 0, 63, 0, 183, 14, 214, 4, 2, 0, 56, 0, 63, 0, 118, 14, 107, 2, 2, 0, 56, 0, 63, 0, 183, 14, 218, 4, 2, 0, 56, 0, 63, 0, 118, 14, 110, 2, 2, 0, 56, 0, 63, 0, 98, 16, 110, 2, 2, 0, 56, 0, 63, 0, 98, 16, 242, 4, 2, 0, 56, 0, 63, 0, 118, 14, 167, 16, 2, 0, 56, 0, 63, 0, 118, 14, 16, 0, 2, 0, 56, 0, 63, 0, 112, 16, 16, 0, 2, 0, 56, 0, 63, 0, 118, 14, 34, 3, 2, 0, 56, 0, 63, 0, 118, 14, 11, 14, 2, 0, 56, 0, 63, 0, 118, 14, 46, 3, 2, 0, 56, 0, 63, 0, 118, 14, 40, 3, 2, 0, 56, 0, 63, 0, 183, 14, 126, 4, 2, 0, 56, 0, 63, 0, 183, 14, 116, 2, 2, 0, 56, 0, 63, 0, 183, 14, 179, 3, 2, 0, 56, 0, 63, 0, 183, 14, 246, 4, 2, 0, 56, 0, 63, 0, 183, 14, 50, 5, 2, 0, 56, 0, 63, 0, 183, 14, 250, 4, 2, 0, 56, 0, 63, 0, 224, 10, 254, 13, 2, 0, 56, 0, 63, 0, 224, 10, 13, 16, 2, 0, 56, 0, 63, 0, 118, 14, 49, 3, 2, 0, 56, 0, 63, 0, 183, 14, 138, 3, 2, 0, 56, 0, 63, 0, 118, 14, 113, 2, 2, 0, 56, 0, 63, 0, 112, 16, 113, 2, 2, 0, 56, 0, 63, 0, 118, 14, 52, 3, 2, 0, 56, 0, 63, 0, 118, 14, 43, 3, 2, 0, 56, 0, 63, 0, 118, 14, 55, 3, 2, 0, 56, 0, 63, 0, 118, 14, 109, 5, 2, 0, 56, 0, 63, 0, 118, 14, 104, 5, 2, 0, 56, 0, 63, 0, 118, 14, 58, 3, 2, 0, 56, 0, 63, 0, 118, 14, 61, 3, 2, 0, 56, 0, 63, 0, 118, 14, 141, 6, 2, 0, 56, 0, 63, 0, 98, 16, 154, 4, 2, 0, 56, 0, 63, 0, 118, 14, 112, 0, 2, 0, 56, 0, 63, 0, 224, 10, 213, 15, 2, 0, 56, 0, 63, 0, 224, 10, 224, 15, 2, 0, 56, 0, 63, 0, 224, 10, 4, 15, 2, 0, 56, 0, 63, 0, 224, 10, 59, 15, 2, 0, 56, 0, 63, 0, 98, 16, 74, 5, 2, 0, 56, 0, 63, 0, 118, 14, 74, 5, 2, 0, 56, 0, 63, 0, 118, 14, 116, 2, 2, 0, 56, 0, 63, 0, 118, 14, 180, 14, 2, 0, 56, 0, 63, 0, 118, 14, 3, 4, 2, 0, 56, 0, 63, 0, 118, 14, 119, 2, 2, 0, 56, 0, 63, 0, 118, 14, 192, 6, 2, 0, 56, 0, 63, 0, 118, 14, 64, 3, 2, 0, 56, 0, 63, 0, 98, 16, 119, 2, 2, 0, 56, 0, 63, 0, 118, 14, 97, 3, 2, 0, 56, 0, 63, 0, 118, 14, 67, 3, 2, 0, 56, 0, 63, 0, 118, 14, 94, 3, 2, 0, 56, 0, 63, 0, 183, 14, 130, 3, 2, 0, 56, 0, 63, 0, 118, 14, 70, 3, 2, 0, 56, 0, 63, 0, 118, 14, 73, 3, 2, 0, 56, 0, 63, 0, 118, 14, 194, 3, 2, 0, 56, 0, 63, 0, 118, 14, 76, 3, 2, 0, 56, 0, 63, 0, 183, 14, 34, 5, 2, 0, 56, 0, 63, 0, 183, 14, 138, 3, 2, 0, 56, 0, 63, 0, 112, 16, 171, 3, 2, 0, 56, 0, 63, 0, 183, 14, 250, 3, 2, 0, 56, 0, 63, 0, 118, 14, 79, 3, 2, 0, 56, 0, 63, 0, 183, 14, 42, 5, 2, 0, 56, 0, 63, 0, 224, 10, 194, 3, 2, 0, 56, 0, 63, 0, 183, 14, 6, 4, 2, 0, 56, 0, 63, 0, 183, 14, 46, 5, 2, 0, 56, 0, 63, 0, 183, 14, 134, 3, 2, 0, 56, 0, 63, 0, 183, 14, 210, 6, 2, 0, 56, 0, 63, 0, 118, 14, 85, 3, 2, 0, 56, 0, 63, 0, 183, 14, 18, 5, 2, 0, 56, 0, 63, 0, 118, 14, 19, 14, 2, 0, 56, 0, 63, 0, 118, 14, 91, 3, 2, 0, 56, 0, 63, 0, 98, 16, 254, 3, 2, 0, 56, 0, 63, 0, 118, 14, 82, 3, 2, 0, 56, 0, 63, 0, 118, 14, 88, 3, 2, 0, 56, 0, 63, 0, 98, 16, 135, 3, 2, 0, 56, 0, 63, 0, 118, 14, 135, 3, 2, 0, 56, 0, 63, 0, 224, 10, 3, 11, 2, 0, 56, 0, 63, 0, 224, 10, 118, 8, 2, 0, 56, 0, 63, 0, 118, 14, 23, 4, 2, 0, 56, 0, 63, 0, 118, 14, 100, 3, 2, 0, 56, 0, 63, 0, 98, 16, 162, 3, 2, 0, 56, 0, 63, 0, 183, 14, 98, 4, 2, 0, 56, 0, 63, 0, 118, 14, 103, 3, 2, 0, 56, 0, 63, 0, 224, 10, 15, 15, 2, 0, 56, 0, 63, 0, 224, 10, 216, 14, 2, 0, 56, 0, 63, 0, 118, 14, 167, 3, 2, 0, 56, 0, 63, 0, 98, 16, 167, 3, 2, 0, 56, 0, 63, 0, 224, 10, 25, 16, 2, 0, 56, 0, 63, 0, 118, 14, 106, 3, 2, 0, 56, 0, 63, 0, 118, 14, 109, 3, 2, 0, 56, 0, 63, 0, 118, 14, 112, 3, 2, 0, 56, 0, 63, 0, 112, 16, 228, 10, 2, 0, 56, 0, 63, 0, 118, 14, 115, 3, 2, 0, 56, 0, 63, 0, 118, 14, 118, 3, 2, 0, 56, 0, 63, 0, 183, 14, 66, 4, 2, 0, 56, 0, 63, 0, 118, 14, 121, 3, 2, 0, 56, 0, 63, 0, 183, 14, 30, 4, 2, 0, 56, 0, 63, 0, 183, 14, 54, 5, 2, 0, 56, 0, 63, 0, 98, 16, 162, 4, 2, 0, 56, 0, 63, 0, 183, 14, 10, 5, 2, 0, 56, 0, 63, 0, 98, 16, 222, 3, 2, 0, 56, 0, 63, 0, 183, 14, 150, 3, 2, 0, 56, 0, 63, 0, 98, 16, 115, 0, 2, 0, 56, 0, 63, 0, 118, 14, 115, 0, 2, 0, 56, 0, 63, 0, 183, 14, 182, 4, 2, 0, 56, 0, 63, 0, 183, 14, 194, 4, 2, 0, 56, 0, 63, 0, 118, 14, 124, 3, 2, 0, 56, 0, 63, 0, 183, 14, 186, 3, 2, 0, 56, 0, 63, 0, 224, 10, 37, 16, 2, 0, 56, 0, 63, 0, 224, 10, 49, 16, 2, 0, 56, 0, 63, 0, 183, 14, 22, 4, 2, 0, 56, 0, 63, 0, 98, 16, 58, 5, 2, 0, 56, 0, 63, 0, 224, 10, 85, 16, 2, 0, 56, 0, 63, 0, 224, 10, 227, 14, 2, 0, 56, 0, 63, 0, 224, 10, 37, 15, 2, 0, 56, 0, 63, 0, 224, 10, 48, 15, 2, 0, 56, 0, 63, 0, 224, 10, 81, 15, 2, 0, 56, 0, 63, 0, 224, 10, 26, 15, 2, 0, 56, 0, 63, 0, 183, 14, 19, 0, 2, 0, 56, 0, 63, 0, 183, 14, 198, 3, 2, 0, 56, 0, 63, 0, 183, 14, 234, 3, 2, 0, 56, 0, 63, 0, 183, 14, 134, 4, 2, 0, 56, 0, 63, 0, 118, 14, 191, 3, 2, 0, 56, 0, 63, 0, 118, 14, 19, 0, 2, 0, 56, 0, 63, 0, 98, 16, 109, 0, 2, 0, 56, 0, 63, 0, 118, 14, 127, 3, 2, 0, 56, 0, 63, 0, 98, 16, 191, 3, 2, 0, 56, 0, 63, 0, 183, 14, 112, 5, 2, 0, 56, 0, 63, 0, 183, 14, 122, 5, 2, 0, 56, 0, 63, 0, 183, 14, 197, 5, 2, 0, 56, 0, 63, 0, 183, 14, 67, 5, 2, 0, 56, 0, 63, 0, 183, 14, 192, 5, 2, 0, 56, 0, 63, 0, 183, 14, 62, 5, 2, 0, 56, 0, 63, 0, 183, 14, 97, 5, 2, 0, 56, 0, 63, 0, 183, 14, 162, 5, 2, 0, 56, 0, 63, 0, 183, 14, 152, 5, 2, 0, 56, 0, 63, 0, 183, 14, 187, 5, 2, 0, 56, 0, 63, 0, 183, 14, 167, 5, 2, 0, 56, 0, 63, 0, 183, 14, 217, 5, 2, 0, 56, 0, 63, 0, 183, 14, 117, 5, 2, 0, 56, 0, 63, 0, 183, 14, 77, 5, 2, 0, 56, 0, 63, 0, 183, 14, 172, 5, 2, 0, 56, 0, 63, 0, 183, 14, 102, 5, 2, 0, 56, 0, 63, 0, 183, 14, 87, 5, 2, 0, 56, 0, 63, 0, 183, 14, 147, 5, 2, 0, 56, 0, 63, 0, 183, 14, 157, 5, 2, 0, 56, 0, 63, 0, 183, 14, 82, 5, 2, 0, 56, 0, 63, 0, 183, 14, 137, 5, 2, 0, 56, 0, 63, 0, 183, 14, 92, 5, 2, 0, 56, 0, 63, 0, 183, 14, 127, 5, 2, 0, 56, 0, 63, 0, 183, 14, 72, 5, 2, 0, 56, 0, 63, 0, 183, 14, 142, 5, 2, 0, 56, 0, 63, 0, 183, 14, 212, 5, 2, 0, 56, 0, 63, 0, 183, 14, 207, 5, 2, 0, 56, 0, 63, 0, 183, 14, 107, 5, 2, 0, 56, 0, 63, 0, 183, 14, 132, 5, 2, 0, 56, 0, 63, 0, 183, 14, 182, 5, 2, 0, 56, 0, 63, 0, 183, 14, 177, 5, 2, 0, 56, 0, 63, 0, 183, 14, 202, 5, 2, 0, 56, 0, 63, 0, 118, 14, 0, 11, 2, 0, 56, 0, 63, 0, 118, 14, 45, 15, 2, 0, 56, 0, 63, 0, 118, 14, 56, 15, 2, 0, 56, 0, 63, 0, 183, 14, 250, 8, 2, 0, 56, 0, 63, 0, 183, 14, 0, 9, 2, 0, 56, 0, 63, 0, 183, 14, 6, 9, 2, 0, 56, 0, 63, 0, 183, 14, 12, 9, 2, 0, 56, 0, 63, 0, 183, 14, 18, 9, 2, 0, 56, 0, 63, 0, 183, 14, 24, 9, 2, 0, 56, 0, 63, 0, 183, 14, 90, 9, 2, 0, 56, 0, 63, 0, 183, 14, 30, 9, 2, 0, 56, 0, 63, 0, 183, 14, 36, 9, 2, 0, 56, 0, 63, 0, 183, 14, 42, 9, 2, 0, 56, 0, 63, 0, 183, 14, 48, 9, 2, 0, 56, 0, 63, 0, 183, 14, 54, 9, 2, 0, 56, 0, 63, 0, 183, 14, 60, 9, 2, 0, 56, 0, 63, 0, 183, 14, 66, 9, 2, 0, 56, 0, 63, 0, 183, 14, 72, 9, 2, 0, 56, 0, 63, 0, 183, 14, 78, 9, 2, 0, 56, 0, 63, 0, 183, 14, 84, 9, 2, 0, 56, 0, 63, 0, 183, 14, 232, 5, 2, 0, 56, 0, 63, 0, 183, 14, 16, 6, 2, 0, 56, 0, 63, 0, 183, 14, 242, 5, 2, 0, 56, 0, 63, 0, 183, 14, 1, 6, 2, 0, 56, 0, 63, 0, 183, 14, 252, 5, 2, 0, 56, 0, 63, 0, 183, 14, 31, 6, 2, 0, 56, 0, 63, 0, 183, 14, 21, 6, 2, 0, 56, 0, 63, 0, 183, 14, 6, 6, 2, 0, 56, 0, 63, 0, 183, 14, 237, 5, 2, 0, 56, 0, 63, 0, 183, 14, 222, 5, 2, 0, 56, 0, 63, 0, 183, 14, 227, 5, 2, 0, 56, 0, 63, 0, 183, 14, 247, 5, 2, 0, 56, 0, 63, 0, 183, 14, 11, 6, 2, 0, 56, 0, 63, 0, 183, 14, 26, 6, 2, 0, 56, 0, 63, 0, 118, 14, 128, 2, 2, 0, 56, 0, 63, 0, 118, 14, 10, 0, 2, 0, 56, 0, 63, 0, 183, 14, 138, 6, 2, 0, 56, 0, 63, 0, 183, 14, 150, 6, 2, 0, 56, 0, 63, 0, 118, 14, 37, 0, 2, 0, 56, 0, 63, 0, 183, 14, 44, 7, 2, 0, 56, 0, 63, 0, 118, 14, 181, 0, 2, 0, 56, 0, 63, 0, 183, 14, 50, 7, 2, 0, 56, 0, 63, 0, 183, 14, 138, 9, 2, 0, 56, 0, 63, 0, 183, 14, 150, 9, 2, 0, 56, 0, 63, 0, 183, 14, 138, 6, 2, 0, 56, 0, 63, 0, 118, 14, 46, 0, 2, 0, 56, 0, 63, 0, 118, 14, 49, 0, 2, 0, 56, 0, 63, 0, 183, 14, 144, 9, 2, 0, 56, 0, 63, 0, 183, 14, 50, 7, 2, 0, 56, 0, 63, 0, 183, 14, 156, 9, 2, 0, 56, 0, 63, 0, 183, 14, 62, 7, 2, 0, 56, 0, 63, 0, 183, 14, 150, 6, 2, 0, 56, 0, 63, 0, 183, 14, 138, 6, 2, 0, 56, 0, 63, 0, 118, 14, 58, 0, 2, 0, 56, 0, 63, 0, 118, 14, 64, 0, 2, 0, 56, 0, 63, 0, 183, 14, 62, 7, 2, 0, 56, 0, 63, 0, 118, 14, 67, 0, 2, 0, 56, 0, 63, 0, 183, 14, 56, 7, 2, 0, 56, 0, 63, 0, 183, 14, 156, 6, 2, 0, 56, 0, 63, 0, 118, 14, 73, 0, 2, 0, 56, 0, 63, 0, 118, 14, 76, 0, 2, 0, 56, 0, 63, 0, 183, 14, 156, 6, 2, 0, 56, 0, 63, 0, 183, 14, 168, 9, 2, 0, 56, 0, 63, 0, 118, 14, 82, 0, 2, 0, 56, 0, 63, 0, 183, 14, 56, 7, 2, 0, 56, 0, 63, 0, 183, 14, 150, 6, 2, 0, 56, 0, 63, 0, 118, 14, 201, 6, 2, 0, 56, 0, 63, 0, 183, 14, 162, 9, 2, 0, 56, 0, 63, 0, 183, 14, 44, 7, 2, 0, 56, 0, 63, 0, 118, 14, 97, 0, 2, 0, 56, 0, 63, 0, 118, 14, 100, 0, 2, 0, 56, 0, 63, 0, 183, 14, 36, 6, 2, 0, 56, 0, 63, 0, 183, 14, 2, 17, 2, 0, 56, 0, 63, 0, 183, 14, 155, 16, 2, 0, 56, 0, 63, 0, 183, 14, 238, 16, 2, 0, 56, 0, 63, 0, 183, 14, 170, 16, 2, 0, 56, 0, 63, 0, 118, 14, 28, 0, 2, 0, 56, 0, 63, 0, 118, 14, 40, 0, 2, 0, 56, 0, 63, 0, 118, 14, 88, 0, 2, 0, 56, 0, 63, 0, 183, 14, 41, 6, 2, 0, 56, 0, 63, 0, 183, 14, 152, 17, 2, 0, 56, 0, 63, 0, 183, 14, 63, 17, 2, 0, 56, 0, 63, 0, 183, 14, 163, 14, 2, 0, 56, 0, 63, 0, 183, 14, 173, 14, 2, 0, 56, 0, 63, 0, 183, 14, 46, 6, 2, 0, 56, 0, 63, 0, 183, 14, 51, 6, 2, 0, 56, 0, 63, 0, 183, 14, 61, 6, 2, 0, 56, 0, 63, 0, 183, 14, 66, 6, 2, 0, 56, 0, 63, 0, 183, 14, 56, 6, 2, 0, 56, 0, 63, 0, 183, 14, 71, 6, 2, 0, 56, 0, 63, 0, 183, 14, 76, 6, 2, 0, 56, 0, 63, 0, 183, 14, 81, 6, 2, 0, 56, 0, 63, 0, 183, 14, 91, 6, 2, 0, 56, 0, 63, 0, 183, 14, 86, 6, 2, 0, 56, 0, 63, 0, 183, 14, 96, 6, 2, 0, 56, 0, 63, 0, 183, 14, 106, 6, 2, 0, 56, 0, 63, 0, 183, 14, 101, 6, 2, 0, 56, 0, 63, 0, 183, 14, 111, 6, 2, 0, 56, 0, 63, 0, 183, 14, 116, 6, 2, 0, 56, 0, 63, 0, 183, 14, 121, 6, 2, 0, 56, 0, 63, 0, 118, 14, 162, 1, 2, 0, 56, 0, 63, 0, 183, 14, 112, 8, 2, 0, 56, 0, 63, 0, 183, 14, 176, 10, 2, 0, 56, 0, 63, 0, 183, 14, 112, 8, 2, 0, 56, 0, 63, 0, 183, 14, 182, 10, 2, 0, 56, 0, 63, 0, 183, 14, 222, 6, 2, 0, 56, 0, 63, 0, 183, 14, 188, 10, 2, 0, 56, 0, 63, 0, 183, 14, 194, 10, 2, 0, 56, 0, 63, 0, 118, 14, 25, 0, 2, 0, 56, 0, 63, 0, 118, 14, 52, 0, 2, 0, 56, 0, 63, 0, 118, 14, 61, 0, 2, 0, 56, 0, 63, 0, 118, 14, 79, 0, 2, 0, 56, 0, 63, 0, 118, 14, 1, 0, 2, 0, 56, 0, 63, 0, 118, 14, 91, 0, 2, 0, 56, 0, 63, 0, 118, 14, 7, 8, 2, 0, 56, 0, 63, 0, 118, 14, 73, 8, 2, 0, 56, 0, 63, 0, 118, 14, 22, 0, 2, 0, 56, 0, 63, 0, 118, 14, 91, 11, 2, 0, 56, 0, 63, 0, 118, 14, 25, 0, 2, 0, 56, 0, 63, 0, 118, 14, 121, 0, 2, 0, 56, 0, 63, 0, 118, 14, 133, 0, 2, 0, 56, 0, 63, 0, 118, 14, 127, 0, 2, 0, 56, 0, 63, 0, 118, 14, 202, 14, 2, 0, 56, 0, 63, 0, 118, 14, 136, 0, 2, 0, 56, 0, 63, 0, 118, 14, 210, 11, 2, 0, 56, 0, 63, 0, 118, 14, 139, 0, 2, 0, 56, 0, 63, 0, 118, 14, 154, 0, 2, 0, 56, 0, 63, 0, 118, 14, 145, 0, 2, 0, 56, 0, 63, 0, 118, 14, 42, 11, 2, 0, 56, 0, 63, 0, 118, 14, 119, 7, 2, 0, 56, 0, 63, 0, 118, 14, 142, 0, 2, 0, 56, 0, 63, 0, 118, 14, 139, 8, 2, 0, 56, 0, 63, 0, 118, 14, 148, 0, 2, 0, 56, 0, 63, 0, 183, 14, 22, 14, 2, 0, 56, 0, 63, 0, 118, 14, 17, 7, 2, 0, 56, 0, 63, 0, 118, 14, 151, 0, 2, 0, 56, 0, 63, 0, 118, 14, 213, 14, 2, 0, 56, 0, 63, 0, 118, 14, 160, 0, 2, 0, 56, 0, 63, 0, 118, 14, 157, 0, 2, 0, 56, 0, 63, 0, 118, 14, 13, 8, 2, 0, 56, 0, 63, 0, 118, 14, 163, 0, 2, 0, 56, 0, 63, 0, 118, 14, 0, 1, 2, 0, 56, 0, 63, 0, 118, 14, 129, 1, 2, 0, 56, 0, 63, 0, 118, 14, 195, 1, 2, 0, 56, 0, 63, 0, 118, 14, 227, 7, 2, 0, 56, 0, 63, 0, 118, 14, 249, 6, 2, 0, 56, 0, 63, 0, 118, 14, 189, 6, 2, 0, 56, 0, 63, 0, 118, 14, 25, 8, 2, 0, 56, 0, 63, 0, 118, 14, 243, 6, 2, 0, 56, 0, 63, 0, 118, 14, 161, 7, 2, 0, 56, 0, 63, 0, 118, 14, 217, 11, 2, 0, 56, 0, 63, 0, 118, 14, 137, 7, 2, 0, 56, 0, 63, 0, 118, 14, 98, 11, 2, 0, 56, 0, 63, 0, 118, 14, 24, 1, 2, 0, 56, 0, 63, 0, 118, 14, 43, 8, 2, 0, 56, 0, 63, 0, 118, 14, 49, 8, 2, 0, 56, 0, 63, 0, 118, 14, 168, 1, 2, 0, 56, 0, 63, 0, 118, 14, 14, 11, 2, 0, 56, 0, 63, 0, 118, 14, 235, 14, 2, 0, 56, 0, 63, 0, 118, 14, 0, 11, 2, 0, 56, 0, 63, 0, 118, 14, 190, 0, 2, 0, 56, 0, 63, 0, 118, 14, 166, 0, 2, 0, 56, 0, 63, 0, 118, 14, 178, 0, 2, 0, 56, 0, 63, 0, 183, 14, 27, 17, 2, 0, 56, 0, 63, 0, 118, 14, 12, 1, 2, 0, 56, 0, 63, 0, 118, 14, 100, 0, 2, 0, 56, 0, 63, 0, 118, 14, 169, 0, 2, 0, 56, 0, 63, 0, 118, 14, 195, 6, 2, 0, 56, 0, 63, 0, 118, 14, 175, 0, 2, 0, 56, 0, 63, 0, 118, 14, 184, 0, 2, 0, 56, 0, 63, 0, 118, 14, 107, 7, 2, 0, 56, 0, 63, 0, 118, 14, 187, 0, 2, 0, 56, 0, 63, 0, 118, 14, 193, 0, 2, 0, 56, 0, 63, 0, 183, 14, 154, 8, 2, 0, 56, 0, 63, 0, 118, 14, 255, 6, 2, 0, 56, 0, 63, 0, 118, 14, 34, 0, 2, 0, 56, 0, 63, 0, 118, 14, 5, 7, 2, 0, 56, 0, 63, 0, 118, 14, 199, 0, 2, 0, 56, 0, 63, 0, 118, 14, 202, 0, 2, 0, 56, 0, 63, 0, 118, 14, 205, 0, 2, 0, 56, 0, 63, 0, 118, 14, 153, 1, 2, 0, 56, 0, 63, 0, 118, 14, 235, 0, 2, 0, 56, 0, 63, 0, 183, 14, 249, 10, 2, 0, 56, 0, 63, 0, 118, 14, 249, 10, 2, 0, 56, 0, 63, 0, 118, 14, 175, 11, 2, 0, 56, 0, 63, 0, 118, 14, 29, 7, 2, 0, 56, 0, 63, 0, 118, 14, 41, 7, 2, 0, 56, 0, 63, 0, 118, 14, 217, 0, 2, 0, 56, 0, 63, 0, 118, 14, 238, 0, 2, 0, 56, 0, 63, 0, 118, 14, 214, 0, 2, 0, 56, 0, 63, 0, 118, 14, 168, 11, 2, 0, 56, 0, 63, 0, 118, 14, 10, 0, 2, 0, 56, 0, 63, 0, 118, 14, 135, 6, 2, 0, 56, 0, 63, 0, 118, 14, 135, 6, 2, 0, 56, 0, 63, 0, 118, 14, 46, 0, 2, 0, 56, 0, 63, 0, 118, 14, 73, 0, 2, 0, 56, 0, 63, 0, 118, 14, 201, 6, 2, 0, 56, 0, 63, 0, 118, 14, 196, 0, 2, 0, 56, 0, 63, 0, 118, 14, 223, 0, 2, 0, 56, 0, 63, 0, 118, 14, 1, 8, 2, 0, 56, 0, 63, 0, 118, 14, 232, 0, 2, 0, 56, 0, 63, 0, 118, 14, 81, 1, 2, 0, 56, 0, 63, 0, 118, 14, 235, 10, 2, 0, 56, 0, 63, 0, 118, 14, 235, 10, 2, 0, 56, 0, 63, 0, 183, 14, 235, 10, 2, 0, 56, 0, 63, 0, 118, 14, 237, 6, 2, 0, 56, 0, 63, 0, 118, 14, 229, 0, 2, 0, 56, 0, 63, 0, 118, 14, 55, 0, 2, 0, 56, 0, 63, 0, 118, 14, 23, 7, 2, 0, 56, 0, 63, 0, 118, 14, 155, 7, 2, 0, 56, 0, 63, 0, 118, 14, 226, 0, 2, 0, 56, 0, 63, 0, 118, 14, 49, 0, 2, 0, 56, 0, 63, 0, 118, 14, 52, 0, 2, 0, 56, 0, 63, 0, 118, 14, 154, 11, 2, 0, 56, 0, 63, 0, 118, 14, 140, 11, 2, 0, 56, 0, 63, 0, 118, 14, 244, 0, 2, 0, 56, 0, 63, 0, 118, 14, 253, 0, 2, 0, 56, 0, 63, 0, 118, 14, 247, 0, 2, 0, 56, 0, 63, 0, 118, 14, 186, 1, 2, 0, 56, 0, 63, 0, 118, 14, 250, 0, 2, 0, 56, 0, 63, 0, 118, 14, 45, 15, 2, 0, 56, 0, 63, 0, 118, 14, 113, 7, 2, 0, 56, 0, 63, 0, 118, 14, 125, 7, 2, 0, 56, 0, 63, 0, 118, 14, 35, 11, 2, 0, 56, 0, 63, 0, 118, 14, 149, 7, 2, 0, 56, 0, 63, 0, 118, 14, 126, 11, 2, 0, 56, 0, 63, 0, 118, 14, 182, 11, 2, 0, 56, 0, 63, 0, 118, 14, 77, 7, 2, 0, 56, 0, 63, 0, 118, 14, 101, 7, 2, 0, 56, 0, 63, 0, 118, 14, 131, 7, 2, 0, 56, 0, 63, 0, 118, 14, 172, 0, 2, 0, 56, 0, 63, 0, 118, 14, 6, 1, 2, 0, 56, 0, 63, 0, 118, 14, 143, 7, 2, 0, 56, 0, 63, 0, 118, 14, 165, 6, 2, 0, 56, 0, 63, 0, 118, 14, 9, 1, 2, 0, 56, 0, 63, 0, 118, 14, 70, 11, 2, 0, 56, 0, 63, 0, 118, 14, 18, 1, 2, 0, 56, 0, 63, 0, 118, 14, 167, 7, 2, 0, 56, 0, 63, 0, 118, 14, 21, 1, 2, 0, 56, 0, 63, 0, 118, 14, 171, 6, 2, 0, 56, 0, 63, 0, 118, 14, 191, 7, 2, 0, 56, 0, 63, 0, 118, 14, 27, 1, 2, 0, 56, 0, 63, 0, 118, 14, 36, 1, 2, 0, 56, 0, 63, 0, 118, 14, 203, 7, 2, 0, 56, 0, 63, 0, 118, 14, 82, 16, 2, 0, 56, 0, 63, 0, 118, 14, 39, 1, 2, 0, 56, 0, 63, 0, 118, 14, 33, 1, 2, 0, 56, 0, 63, 0, 118, 14, 197, 7, 2, 0, 56, 0, 63, 0, 118, 14, 185, 7, 2, 0, 56, 0, 63, 0, 118, 14, 56, 15, 2, 0, 56, 0, 63, 0, 118, 14, 209, 7, 2, 0, 56, 0, 63, 0, 118, 14, 251, 7, 2, 0, 56, 0, 63, 0, 118, 14, 183, 6, 2, 0, 56, 0, 63, 0, 118, 14, 11, 7, 2, 0, 56, 0, 63, 0, 118, 14, 161, 11, 2, 0, 56, 0, 63, 0, 118, 14, 233, 7, 2, 0, 56, 0, 63, 0, 118, 14, 64, 0, 2, 0, 56, 0, 63, 0, 118, 14, 51, 1, 2, 0, 56, 0, 63, 0, 118, 14, 105, 11, 2, 0, 56, 0, 63, 0, 118, 14, 57, 1, 2, 0, 56, 0, 63, 0, 118, 14, 42, 1, 2, 0, 56, 0, 63, 0, 118, 14, 177, 6, 2, 0, 56, 0, 63, 0, 118, 14, 45, 1, 2, 0, 56, 0, 63, 0, 118, 14, 79, 8, 2, 0, 56, 0, 63, 0, 118, 14, 54, 1, 2, 0, 56, 0, 63, 0, 118, 14, 70, 16, 2, 0, 56, 0, 63, 0, 118, 14, 112, 11, 2, 0, 56, 0, 63, 0, 118, 14, 72, 1, 2, 0, 56, 0, 63, 0, 118, 14, 133, 11, 2, 0, 56, 0, 63, 0, 118, 14, 66, 1, 2, 0, 56, 0, 63, 0, 118, 14, 239, 7, 2, 0, 56, 0, 63, 0, 118, 14, 71, 7, 2, 0, 56, 0, 63, 0, 118, 14, 127, 14, 2, 0, 56, 0, 63, 0, 118, 14, 40, 0, 2, 0, 56, 0, 63, 0, 183, 14, 127, 14, 2, 0, 56, 0, 63, 0, 118, 14, 28, 0, 2, 0, 56, 0, 63, 0, 118, 14, 88, 0, 2, 0, 56, 0, 63, 0, 118, 14, 133, 14, 2, 0, 56, 0, 63, 0, 118, 14, 144, 6, 2, 0, 56, 0, 63, 0, 118, 14, 67, 0, 2, 0, 56, 0, 63, 0, 118, 14, 94, 0, 2, 0, 56, 0, 63, 0, 118, 14, 215, 7, 2, 0, 56, 0, 63, 0, 118, 14, 63, 1, 2, 0, 56, 0, 63, 0, 118, 14, 203, 11, 2, 0, 56, 0, 63, 0, 118, 14, 145, 8, 2, 0, 56, 0, 63, 0, 118, 14, 69, 1, 2, 0, 56, 0, 63, 0, 118, 14, 60, 1, 2, 0, 56, 0, 63, 0, 118, 14, 245, 7, 2, 0, 56, 0, 63, 0, 118, 14, 61, 0, 2, 0, 56, 0, 63, 0, 118, 14, 1, 0, 2, 0, 56, 0, 63, 0, 183, 14, 1, 16, 2, 0, 56, 0, 63, 0, 118, 14, 220, 0, 2, 0, 56, 0, 63, 0, 118, 14, 48, 1, 2, 0, 56, 0, 63, 0, 118, 14, 10, 16, 2, 0, 56, 0, 63, 0, 118, 14, 133, 8, 2, 0, 56, 0, 63, 0, 118, 14, 70, 0, 2, 0, 56, 0, 63, 0, 118, 14, 75, 1, 2, 0, 56, 0, 63, 0, 118, 14, 83, 7, 2, 0, 56, 0, 63, 0, 118, 14, 19, 8, 2, 0, 56, 0, 63, 0, 118, 14, 56, 11, 2, 0, 56, 0, 63, 0, 118, 14, 78, 1, 2, 0, 56, 0, 63, 0, 118, 14, 189, 11, 2, 0, 56, 0, 63, 0, 118, 14, 84, 1, 2, 0, 56, 0, 63, 0, 118, 14, 241, 0, 2, 0, 56, 0, 63, 0, 118, 14, 13, 0, 2, 0, 56, 0, 63, 0, 118, 14, 79, 0, 2, 0, 56, 0, 63, 0, 118, 14, 87, 1, 2, 0, 56, 0, 63, 0, 118, 14, 82, 0, 2, 0, 56, 0, 63, 0, 118, 14, 31, 8, 2, 0, 56, 0, 63, 0, 118, 14, 119, 11, 2, 0, 56, 0, 63, 0, 118, 14, 37, 8, 2, 0, 56, 0, 63, 0, 118, 14, 37, 0, 2, 0, 56, 0, 63, 0, 118, 14, 135, 1, 2, 0, 56, 0, 63, 0, 118, 14, 15, 1, 2, 0, 56, 0, 63, 0, 118, 14, 130, 0, 2, 0, 56, 0, 63, 0, 118, 14, 30, 1, 2, 0, 56, 0, 63, 0, 118, 14, 58, 0, 2, 0, 56, 0, 63, 0, 118, 14, 76, 0, 2, 0, 56, 0, 63, 0, 118, 14, 189, 1, 2, 0, 56, 0, 63, 0, 118, 14, 144, 1, 2, 0, 56, 0, 63, 0, 118, 14, 150, 1, 2, 0, 56, 0, 63, 0, 118, 14, 126, 1, 2, 0, 56, 0, 63, 0, 118, 14, 49, 11, 2, 0, 56, 0, 63, 0, 118, 14, 78, 15, 2, 0, 56, 0, 63, 0, 118, 14, 132, 1, 2, 0, 56, 0, 63, 0, 118, 14, 141, 1, 2, 0, 56, 0, 63, 0, 118, 14, 89, 15, 2, 0, 56, 0, 63, 0, 118, 14, 55, 8, 2, 0, 56, 0, 63, 0, 118, 14, 242, 10, 2, 0, 56, 0, 63, 0, 118, 14, 61, 8, 2, 0, 56, 0, 63, 0, 118, 14, 67, 8, 2, 0, 56, 0, 63, 0, 118, 14, 151, 8, 2, 0, 56, 0, 63, 0, 118, 14, 219, 6, 2, 0, 56, 0, 63, 0, 118, 14, 4, 0, 2, 0, 56, 0, 63, 0, 118, 14, 35, 7, 2, 0, 56, 0, 63, 0, 118, 14, 147, 11, 2, 0, 56, 0, 63, 0, 118, 14, 85, 0, 2, 0, 56, 0, 63, 0, 118, 14, 211, 0, 2, 0, 56, 0, 63, 0, 183, 14, 85, 0, 2, 0, 56, 0, 63, 0, 118, 14, 147, 1, 2, 0, 56, 0, 63, 0, 118, 14, 138, 1, 2, 0, 56, 0, 63, 0, 118, 14, 159, 1, 2, 0, 56, 0, 63, 0, 118, 14, 85, 8, 2, 0, 56, 0, 63, 0, 118, 14, 63, 11, 2, 0, 56, 0, 63, 0, 118, 14, 156, 1, 2, 0, 56, 0, 63, 0, 118, 14, 91, 8, 2, 0, 56, 0, 63, 0, 118, 14, 97, 8, 2, 0, 56, 0, 63, 0, 118, 14, 84, 11, 2, 0, 56, 0, 63, 0, 118, 14, 171, 1, 2, 0, 56, 0, 63, 0, 118, 14, 109, 8, 2, 0, 56, 0, 63, 0, 118, 14, 177, 1, 2, 0, 56, 0, 63, 0, 118, 14, 124, 0, 2, 0, 56, 0, 63, 0, 118, 14, 174, 1, 2, 0, 56, 0, 63, 0, 118, 14, 179, 7, 2, 0, 56, 0, 63, 0, 118, 14, 103, 8, 2, 0, 56, 0, 63, 0, 118, 14, 165, 1, 2, 0, 56, 0, 63, 0, 118, 14, 180, 1, 2, 0, 56, 0, 63, 0, 118, 14, 196, 11, 2, 0, 56, 0, 63, 0, 118, 14, 127, 8, 2, 0, 56, 0, 63, 0, 118, 14, 221, 7, 2, 0, 56, 0, 63, 0, 118, 14, 106, 17, 2, 0, 56, 0, 63, 0, 118, 14, 208, 0, 2, 0, 56, 0, 63, 0, 118, 14, 207, 6, 2, 0, 56, 0, 63, 0, 183, 14, 228, 6, 2, 0, 56, 0, 63, 0, 118, 14, 231, 6, 2, 0, 56, 0, 63, 0, 118, 14, 3, 1, 2, 0, 56, 0, 63, 0, 118, 14, 89, 7, 2, 0, 56, 0, 63, 0, 118, 14, 28, 11, 2, 0, 56, 0, 63, 0, 118, 14, 77, 11, 2, 0, 56, 0, 63, 0, 118, 14, 91, 0, 2, 0, 56, 0, 63, 0, 118, 14, 31, 0, 2, 0, 56, 0, 63, 0, 118, 14, 183, 1, 2, 0, 56, 0, 63, 0, 118, 14, 224, 14, 2, 0, 56, 0, 63, 0, 118, 14, 192, 1, 2, 0, 56, 0, 63, 0, 118, 14, 97, 0, 2, 0, 56, 0, 63, 0, 118, 14, 198, 1, 2, 0, 56, 0, 63, 0, 183, 14, 4, 0, 2, 0, 56, 0, 63, 0, 118, 14, 4, 0, 2, 0, 56, 0, 63, 0, 183, 14, 202, 16, 2, 0, 56, 0, 63, 0, 118, 14, 126, 6, 2, 0, 56, 0, 63, 0, 118, 14, 21, 11, 2, 0, 56, 0, 63, 0, 118, 14, 118, 0, 2, 0, 56, 0, 63, 0, 118, 14, 90, 1, 2, 0, 56, 0, 63, 0, 118, 14, 93, 1, 2, 0, 56, 0, 63, 0, 118, 14, 96, 1, 2, 0, 56, 0, 63, 0, 118, 14, 99, 1, 2, 0, 56, 0, 63, 0, 118, 14, 102, 1, 2, 0, 56, 0, 63, 0, 118, 14, 105, 1, 2, 0, 56, 0, 63, 0, 118, 14, 108, 1, 2, 0, 56, 0, 63, 0, 118, 14, 43, 0, 2, 0, 56, 0, 63, 0, 118, 14, 111, 1, 2, 0, 56, 0, 63, 0, 118, 14, 114, 1, 2, 0, 56, 0, 63, 0, 118, 14, 117, 1, 2, 0, 56, 0, 63, 0, 118, 14, 120, 1, 2, 0, 56, 0, 63, 0, 118, 14, 123, 1, 2, 0, 56, 0, 63, 0, 118, 14, 201, 1, 2, 0, 56, 0, 63, 0, 118, 14, 204, 1, 2, 0, 56, 0, 63, 0, 118, 14, 207, 1, 2, 0, 56, 0, 63, 0, 118, 14, 210, 1, 2, 0, 56, 0, 63, 0, 118, 14, 213, 1, 2, 0, 56, 0, 63, 0, 118, 14, 216, 1, 2, 0, 56, 0, 63, 0, 118, 14, 219, 1, 2, 0, 56, 0, 63, 0, 118, 14, 222, 1, 2, 0, 56, 0, 63, 0, 118, 14, 225, 1, 2, 0, 56, 0, 63, 0, 118, 14, 228, 1, 2, 0, 56, 0, 63, 0, 118, 14, 67, 15, 2, 0, 56, 0, 63, 0, 118, 14, 231, 1, 2, 0, 56, 0, 63, 0, 118, 14, 234, 1, 2, 0, 56, 0, 63, 0, 118, 14, 237, 1, 2, 0, 56, 0, 63, 0, 118, 14, 240, 1, 2, 0, 56, 0, 63, 0, 118, 14, 243, 1, 2, 0, 56, 0, 63, 0, 118, 14, 246, 1, 2, 0, 56, 0, 63, 0, 118, 14, 249, 1, 2, 0, 56, 0, 63, 0, 118, 14, 252, 1, 2, 0, 56, 0, 63, 0, 118, 14, 255, 1, 2, 0, 56, 0, 63, 0, 118, 14, 2, 2, 2, 0, 56, 0, 63, 0, 118, 14, 5, 2, 2, 0, 56, 0, 63, 0, 118, 14, 8, 2, 2, 0, 56, 0, 63, 0, 118, 14, 11, 2, 2, 0, 56, 0, 63, 0, 118, 14, 14, 2, 2, 0, 56, 0, 63, 0, 118, 14, 17, 2, 2, 0, 56, 0, 63, 0, 118, 14, 20, 2, 2, 0, 56, 0, 63, 0, 118, 14, 118, 0, 2, 0, 56, 0, 63, 0, 118, 14, 28, 0, 2, 0, 56, 0, 63, 0, 118, 14, 7, 8, 2, 0, 56, 0, 63, 0, 118, 14, 133, 0, 2, 0, 56, 0, 63, 0, 118, 14, 130, 0, 2, 0, 56, 0, 63, 0, 118, 14, 10, 0, 2, 0, 56, 0, 63, 0, 118, 14, 73, 8, 2, 0, 56, 0, 63, 0, 183, 14, 127, 14, 2, 0, 56, 0, 63, 0, 118, 14, 121, 0, 2, 0, 56, 0, 63, 0, 118, 14, 127, 14, 2, 0, 56, 0, 63, 0, 118, 14, 124, 0, 2, 0, 56, 0, 63, 0, 118, 14, 136, 0, 2, 0, 56, 0, 63, 0, 118, 14, 119, 7, 2, 0, 56, 0, 63, 0, 118, 14, 122, 2, 2, 0, 56, 0, 63, 0, 118, 14, 25, 0, 2, 0, 56, 0, 63, 0, 118, 14, 22, 0, 2, 0, 56, 0, 63, 0, 118, 14, 201, 6, 2, 0, 56, 0, 63, 0, 118, 14, 127, 0, 2, 0, 56, 0, 63, 0, 118, 14, 210, 11, 2, 0, 56, 0, 63, 0, 118, 14, 139, 0, 2, 0, 56, 0, 63, 0, 118, 14, 202, 14, 2, 0, 56, 0, 63, 0, 118, 14, 25, 8, 2, 0, 56, 0, 63, 0, 118, 14, 139, 8, 2, 0, 56, 0, 63, 0, 118, 14, 34, 0, 2, 0, 56, 0, 63, 0, 118, 14, 133, 14, 2, 0, 56, 0, 63, 0, 118, 14, 31, 0, 2, 0, 56, 0, 63, 0, 118, 14, 42, 11, 2, 0, 56, 0, 63, 0, 118, 14, 249, 6, 2, 0, 56, 0, 63, 0, 118, 14, 145, 0, 2, 0, 56, 0, 63, 0, 118, 14, 154, 0, 2, 0, 56, 0, 63, 0, 118, 14, 213, 14, 2, 0, 56, 0, 63, 0, 118, 14, 37, 0, 2, 0, 56, 0, 63, 0, 118, 14, 243, 6, 2, 0, 56, 0, 63, 0, 118, 14, 163, 0, 2, 0, 56, 0, 63, 0, 118, 14, 148, 0, 2, 0, 56, 0, 63, 0, 118, 14, 151, 0, 2, 0, 56, 0, 63, 0, 118, 14, 13, 8, 2, 0, 56, 0, 63, 0, 118, 14, 142, 0, 2, 0, 56, 0, 63, 0, 118, 14, 227, 7, 2, 0, 56, 0, 63, 0, 118, 14, 17, 7, 2, 0, 56, 0, 63, 0, 183, 14, 189, 6, 2, 0, 56, 0, 63, 0, 118, 14, 189, 6, 2, 0, 56, 0, 63, 0, 118, 14, 157, 0, 2, 0, 56, 0, 63, 0, 118, 14, 160, 0, 2, 0, 56, 0, 63, 0, 118, 14, 43, 8, 2, 0, 56, 0, 63, 0, 118, 14, 137, 7, 2, 0, 56, 0, 63, 0, 118, 14, 166, 0, 2, 0, 56, 0, 63, 0, 118, 14, 63, 11, 2, 0, 56, 0, 63, 0, 118, 14, 14, 11, 2, 0, 56, 0, 63, 0, 118, 14, 235, 14, 2, 0, 56, 0, 63, 0, 118, 14, 172, 0, 2, 0, 56, 0, 63, 0, 118, 14, 217, 11, 2, 0, 56, 0, 63, 0, 118, 14, 195, 6, 2, 0, 56, 0, 63, 0, 118, 14, 169, 0, 2, 0, 56, 0, 63, 0, 118, 14, 175, 0, 2, 0, 56, 0, 63, 0, 118, 14, 178, 0, 2, 0, 56, 0, 63, 0, 118, 14, 12, 1, 2, 0, 56, 0, 63, 0, 118, 14, 181, 0, 2, 0, 56, 0, 63, 0, 118, 14, 98, 11, 2, 0, 56, 0, 63, 0, 118, 14, 184, 0, 2, 0, 56, 0, 63, 0, 183, 14, 126, 6, 2, 0, 56, 0, 63, 0, 183, 14, 55, 0, 2, 0, 56, 0, 63, 0, 118, 14, 187, 0, 2, 0, 56, 0, 63, 0, 118, 14, 40, 0, 2, 0, 56, 0, 63, 0, 118, 14, 190, 0, 2, 0, 56, 0, 63, 0, 118, 14, 24, 1, 2, 0, 56, 0, 63, 0, 118, 14, 193, 0, 2, 0, 56, 0, 63, 0, 118, 14, 255, 6, 2, 0, 56, 0, 63, 0, 183, 14, 235, 10, 2, 0, 56, 0, 63, 0, 118, 14, 235, 10, 2, 0, 56, 0, 63, 0, 118, 14, 235, 10, 2, 0, 56, 0, 63, 0, 118, 14, 125, 2, 2, 0, 56, 0, 63, 0, 118, 14, 196, 0, 2, 0, 56, 0, 63, 0, 118, 14, 199, 0, 2, 0, 56, 0, 63, 0, 118, 14, 5, 7, 2, 0, 56, 0, 63, 0, 118, 14, 202, 0, 2, 0, 56, 0, 63, 0, 183, 14, 34, 0, 2, 0, 56, 0, 63, 0, 118, 14, 91, 11, 2, 0, 56, 0, 63, 0, 118, 14, 205, 0, 2, 0, 56, 0, 63, 0, 118, 14, 208, 0, 2, 0, 56, 0, 63, 0, 118, 14, 175, 11, 2, 0, 56, 0, 63, 0, 118, 14, 211, 0, 2, 0, 56, 0, 63, 0, 118, 14, 35, 7, 2, 0, 56, 0, 63, 0, 118, 14, 29, 7, 2, 0, 56, 0, 63, 0, 118, 14, 249, 10, 2, 0, 56, 0, 63, 0, 118, 14, 168, 11, 2, 0, 56, 0, 63, 0, 118, 14, 214, 0, 2, 0, 56, 0, 63, 0, 118, 14, 217, 0, 2, 0, 56, 0, 63, 0, 183, 14, 198, 6, 2, 0, 56, 0, 63, 0, 118, 14, 135, 6, 2, 0, 56, 0, 63, 0, 118, 14, 41, 7, 2, 0, 56, 0, 63, 0, 118, 14, 220, 0, 2, 0, 56, 0, 63, 0, 183, 14, 135, 6, 2, 0, 56, 0, 63, 0, 118, 14, 135, 6, 2, 0, 56, 0, 63, 0, 118, 14, 223, 0, 2, 0, 56, 0, 63, 0, 118, 14, 207, 6, 2, 0, 56, 0, 63, 0, 118, 14, 1, 8, 2, 0, 56, 0, 63, 0, 118, 14, 231, 6, 2, 0, 56, 0, 63, 0, 118, 14, 237, 6, 2, 0, 56, 0, 63, 0, 118, 14, 229, 0, 2, 0, 56, 0, 63, 0, 118, 14, 140, 11, 2, 0, 56, 0, 63, 0, 118, 14, 49, 0, 2, 0, 56, 0, 63, 0, 118, 14, 232, 0, 2, 0, 56, 0, 63, 0, 118, 14, 241, 0, 2, 0, 56, 0, 63, 0, 118, 14, 235, 0, 2, 0, 56, 0, 63, 0, 118, 14, 23, 7, 2, 0, 56, 0, 63, 0, 118, 14, 226, 0, 2, 0, 56, 0, 63, 0, 118, 14, 155, 7, 2, 0, 56, 0, 63, 0, 118, 14, 154, 11, 2, 0, 56, 0, 63, 0, 118, 14, 46, 0, 2, 0, 56, 0, 63, 0, 118, 14, 52, 0, 2, 0, 56, 0, 63, 0, 118, 14, 244, 0, 2, 0, 56, 0, 63, 0, 118, 14, 45, 15, 2, 0, 56, 0, 63, 0, 118, 14, 247, 0, 2, 0, 56, 0, 63, 0, 118, 14, 250, 0, 2, 0, 56, 0, 63, 0, 118, 14, 107, 7, 2, 0, 56, 0, 63, 0, 118, 14, 253, 0, 2, 0, 56, 0, 63, 0, 118, 14, 113, 7, 2, 0, 56, 0, 63, 0, 183, 14, 31, 0, 2, 0, 56, 0, 63, 0, 118, 14, 149, 7, 2, 0, 56, 0, 63, 0, 118, 14, 89, 7, 2, 0, 56, 0, 63, 0, 118, 14, 35, 11, 2, 0, 56, 0, 63, 0, 118, 14, 0, 1, 2, 0, 56, 0, 63, 0, 118, 14, 77, 7, 2, 0, 56, 0, 63, 0, 118, 14, 126, 11, 2, 0, 56, 0, 63, 0, 118, 14, 182, 11, 2, 0, 56, 0, 63, 0, 118, 14, 125, 7, 2, 0, 56, 0, 63, 0, 118, 14, 101, 7, 2, 0, 56, 0, 63, 0, 118, 14, 131, 7, 2, 0, 56, 0, 63, 0, 118, 14, 6, 1, 2, 0, 56, 0, 63, 0, 118, 14, 3, 1, 2, 0, 56, 0, 63, 0, 118, 14, 9, 1, 2, 0, 56, 0, 63, 0, 118, 14, 143, 7, 2, 0, 56, 0, 63, 0, 183, 14, 1, 0, 2, 0, 56, 0, 63, 0, 118, 14, 165, 6, 2, 0, 56, 0, 63, 0, 118, 14, 70, 11, 2, 0, 56, 0, 63, 0, 118, 14, 171, 6, 2, 0, 56, 0, 63, 0, 118, 14, 161, 7, 2, 0, 56, 0, 63, 0, 118, 14, 55, 0, 2, 0, 56, 0, 63, 0, 118, 14, 15, 1, 2, 0, 56, 0, 63, 0, 118, 14, 167, 7, 2, 0, 56, 0, 63, 0, 118, 14, 21, 1, 2, 0, 56, 0, 63, 0, 118, 14, 191, 7, 2, 0, 56, 0, 63, 0, 118, 14, 27, 1, 2, 0, 56, 0, 63, 0, 118, 14, 82, 16, 2, 0, 56, 0, 63, 0, 118, 14, 39, 1, 2, 0, 56, 0, 63, 0, 118, 14, 30, 1, 2, 0, 56, 0, 63, 0, 118, 14, 33, 1, 2, 0, 56, 0, 63, 0, 118, 14, 49, 8, 2, 0, 56, 0, 63, 0, 118, 14, 36, 1, 2, 0, 56, 0, 63, 0, 118, 14, 197, 7, 2, 0, 56, 0, 63, 0, 118, 14, 185, 7, 2, 0, 56, 0, 63, 0, 118, 14, 203, 7, 2, 0, 56, 0, 63, 0, 118, 14, 56, 15, 2, 0, 56, 0, 63, 0, 118, 14, 58, 0, 2, 0, 56, 0, 63, 0, 118, 14, 70, 16, 2, 0, 56, 0, 63, 0, 118, 14, 42, 1, 2, 0, 56, 0, 63, 0, 118, 14, 45, 1, 2, 0, 56, 0, 63, 0, 118, 14, 209, 7, 2, 0, 56, 0, 63, 0, 118, 14, 11, 7, 2, 0, 56, 0, 63, 0, 118, 14, 57, 1, 2, 0, 56, 0, 63, 0, 118, 14, 48, 1, 2, 0, 56, 0, 63, 0, 183, 14, 1, 0, 2, 0, 56, 0, 63, 0, 118, 14, 221, 7, 2, 0, 56, 0, 63, 0, 118, 14, 161, 11, 2, 0, 56, 0, 63, 0, 118, 14, 233, 7, 2, 0, 56, 0, 63, 0, 118, 14, 189, 6, 2, 0, 56, 0, 63, 0, 118, 14, 79, 8, 2, 0, 56, 0, 63, 0, 118, 14, 177, 6, 2, 0, 56, 0, 63, 0, 118, 14, 61, 0, 2, 0, 56, 0, 63, 0, 118, 14, 112, 11, 2, 0, 56, 0, 63, 0, 118, 14, 51, 1, 2, 0, 56, 0, 63, 0, 118, 14, 54, 1, 2, 0, 56, 0, 63, 0, 118, 14, 64, 0, 2, 0, 56, 0, 63, 0, 118, 14, 105, 11, 2, 0, 56, 0, 63, 0, 118, 14, 251, 7, 2, 0, 56, 0, 63, 0, 118, 14, 183, 6, 2, 0, 56, 0, 63, 0, 118, 14, 100, 0, 2, 0, 56, 0, 63, 0, 118, 14, 133, 11, 2, 0, 56, 0, 63, 0, 118, 14, 67, 0, 2, 0, 56, 0, 63, 0, 118, 14, 203, 11, 2, 0, 56, 0, 63, 0, 118, 14, 60, 1, 2, 0, 56, 0, 63, 0, 118, 14, 145, 8, 2, 0, 56, 0, 63, 0, 183, 14, 94, 0, 2, 0, 56, 0, 63, 0, 118, 14, 63, 1, 2, 0, 56, 0, 63, 0, 118, 14, 69, 1, 2, 0, 56, 0, 63, 0, 118, 14, 71, 7, 2, 0, 56, 0, 63, 0, 118, 14, 245, 7, 2, 0, 56, 0, 63, 0, 118, 14, 239, 7, 2, 0, 56, 0, 63, 0, 183, 14, 22, 0, 2, 0, 56, 0, 63, 0, 118, 14, 66, 1, 2, 0, 56, 0, 63, 0, 183, 14, 144, 6, 2, 0, 56, 0, 63, 0, 118, 14, 144, 6, 2, 0, 56, 0, 63, 0, 118, 14, 215, 7, 2, 0, 56, 0, 63, 0, 118, 14, 72, 1, 2, 0, 56, 0, 63, 0, 118, 14, 133, 8, 2, 0, 56, 0, 63, 0, 118, 14, 70, 0, 2, 0, 56, 0, 63, 0, 183, 14, 1, 16, 2, 0, 56, 0, 63, 0, 118, 14, 78, 1, 2, 0, 56, 0, 63, 0, 118, 14, 19, 8, 2, 0, 56, 0, 63, 0, 118, 14, 56, 11, 2, 0, 56, 0, 63, 0, 118, 14, 10, 16, 2, 0, 56, 0, 63, 0, 118, 14, 75, 1, 2, 0, 56, 0, 63, 0, 118, 14, 189, 11, 2, 0, 56, 0, 63, 0, 118, 14, 79, 0, 2, 0, 56, 0, 63, 0, 118, 14, 18, 1, 2, 0, 56, 0, 63, 0, 118, 14, 84, 1, 2, 0, 56, 0, 63, 0, 118, 14, 83, 7, 2, 0, 56, 0, 63, 0, 118, 14, 81, 1, 2, 0, 56, 0, 63, 0, 183, 14, 1, 0, 2, 0, 56, 0, 63, 0, 118, 14, 73, 0, 2, 0, 56, 0, 63, 0, 183, 14, 70, 0, 2, 0, 56, 0, 63, 0, 118, 14, 87, 1, 2, 0, 56, 0, 63, 0, 118, 14, 90, 1, 2, 0, 56, 0, 63, 0, 118, 14, 93, 1, 2, 0, 56, 0, 63, 0, 118, 14, 96, 1, 2, 0, 56, 0, 63, 0, 118, 14, 99, 1, 2, 0, 56, 0, 63, 0, 118, 14, 102, 1, 2, 0, 56, 0, 63, 0, 118, 14, 105, 1, 2, 0, 56, 0, 63, 0, 118, 14, 108, 1, 2, 0, 56, 0, 63, 0, 183, 14, 43, 0, 2, 0, 56, 0, 63, 0, 118, 14, 43, 0, 2, 0, 56, 0, 63, 0, 118, 14, 111, 1, 2, 0, 56, 0, 63, 0, 118, 14, 114, 1, 2, 0, 56, 0, 63, 0, 118, 14, 117, 1, 2, 0, 56, 0, 63, 0, 118, 14, 120, 1, 2, 0, 56, 0, 63, 0, 118, 14, 123, 1, 2, 0, 56, 0, 63, 0, 118, 14, 82, 0, 2, 0, 56, 0, 63, 0, 183, 14, 219, 6, 2, 0, 56, 0, 63, 0, 118, 14, 31, 8, 2, 0, 56, 0, 63, 0, 118, 14, 119, 11, 2, 0, 56, 0, 63, 0, 118, 14, 37, 8, 2, 0, 56, 0, 63, 0, 118, 14, 126, 1, 2, 0, 56, 0, 63, 0, 118, 14, 126, 6, 2, 0, 56, 0, 63, 0, 118, 14, 85, 0, 2, 0, 56, 0, 63, 0, 118, 14, 49, 11, 2, 0, 56, 0, 63, 0, 118, 14, 89, 15, 2, 0, 56, 0, 63, 0, 118, 14, 238, 0, 2, 0, 56, 0, 63, 0, 118, 14, 135, 1, 2, 0, 56, 0, 63, 0, 118, 14, 138, 1, 2, 0, 56, 0, 63, 0, 118, 14, 129, 1, 2, 0, 56, 0, 63, 0, 118, 14, 141, 1, 2, 0, 56, 0, 63, 0, 118, 14, 153, 1, 2, 0, 56, 0, 63, 0, 118, 14, 144, 1, 2, 0, 56, 0, 63, 0, 118, 14, 67, 8, 2, 0, 56, 0, 63, 0, 118, 14, 76, 0, 2, 0, 56, 0, 63, 0, 118, 14, 78, 15, 2, 0, 56, 0, 63, 0, 118, 14, 147, 11, 2, 0, 56, 0, 63, 0, 118, 14, 150, 1, 2, 0, 56, 0, 63, 0, 183, 14, 106, 17, 2, 0, 56, 0, 63, 0, 118, 14, 106, 17, 2, 0, 56, 0, 63, 0, 118, 14, 147, 1, 2, 0, 56, 0, 63, 0, 118, 14, 55, 8, 2, 0, 56, 0, 63, 0, 118, 14, 61, 8, 2, 0, 56, 0, 63, 0, 118, 14, 85, 8, 2, 0, 56, 0, 63, 0, 118, 14, 159, 1, 2, 0, 56, 0, 63, 0, 118, 14, 88, 0, 2, 0, 56, 0, 63, 0, 118, 14, 132, 1, 2, 0, 56, 0, 63, 0, 118, 14, 156, 1, 2, 0, 56, 0, 63, 0, 118, 14, 162, 1, 2, 0, 56, 0, 63, 0, 118, 14, 165, 1, 2, 0, 56, 0, 63, 0, 118, 14, 168, 1, 2, 0, 56, 0, 63, 0, 118, 14, 84, 11, 2, 0, 56, 0, 63, 0, 118, 14, 97, 8, 2, 0, 56, 0, 63, 0, 118, 14, 91, 8, 2, 0, 56, 0, 63, 0, 118, 14, 171, 1, 2, 0, 56, 0, 63, 0, 118, 14, 103, 8, 2, 0, 56, 0, 63, 0, 118, 14, 13, 0, 2, 0, 56, 0, 63, 0, 118, 14, 13, 0, 2, 0, 56, 0, 63, 0, 118, 14, 109, 8, 2, 0, 56, 0, 63, 0, 183, 14, 13, 0, 2, 0, 56, 0, 63, 0, 118, 14, 177, 1, 2, 0, 56, 0, 63, 0, 118, 14, 174, 1, 2, 0, 56, 0, 63, 0, 118, 14, 179, 7, 2, 0, 56, 0, 63, 0, 118, 14, 180, 1, 2, 0, 56, 0, 63, 0, 118, 14, 0, 11, 2, 0, 56, 0, 63, 0, 118, 14, 28, 11, 2, 0, 56, 0, 63, 0, 118, 14, 196, 11, 2, 0, 56, 0, 63, 0, 183, 14, 207, 6, 2, 0, 56, 0, 63, 0, 118, 14, 127, 8, 2, 0, 56, 0, 63, 0, 118, 14, 1, 0, 2, 0, 56, 0, 63, 0, 118, 14, 183, 1, 2, 0, 56, 0, 63, 0, 118, 14, 77, 11, 2, 0, 56, 0, 63, 0, 118, 14, 224, 14, 2, 0, 56, 0, 63, 0, 118, 14, 186, 1, 2, 0, 56, 0, 63, 0, 118, 14, 189, 1, 2, 0, 56, 0, 63, 0, 183, 14, 242, 10, 2, 0, 56, 0, 63, 0, 118, 14, 192, 1, 2, 0, 56, 0, 63, 0, 118, 14, 195, 1, 2, 0, 56, 0, 63, 0, 118, 14, 91, 0, 2, 0, 56, 0, 63, 0, 118, 14, 242, 10, 2, 0, 56, 0, 63, 0, 118, 14, 94, 0, 2, 0, 56, 0, 63, 0, 183, 14, 1, 0, 2, 0, 56, 0, 63, 0, 118, 14, 97, 0, 2, 0, 56, 0, 63, 0, 118, 14, 198, 1, 2, 0, 56, 0, 63, 0, 118, 14, 201, 1, 2, 0, 56, 0, 63, 0, 118, 14, 204, 1, 2, 0, 56, 0, 63, 0, 118, 14, 207, 1, 2, 0, 56, 0, 63, 0, 118, 14, 210, 1, 2, 0, 56, 0, 63, 0, 118, 14, 213, 1, 2, 0, 56, 0, 63, 0, 118, 14, 216, 1, 2, 0, 56, 0, 63, 0, 118, 14, 219, 1, 2, 0, 56, 0, 63, 0, 118, 14, 222, 1, 2, 0, 56, 0, 63, 0, 118, 14, 225, 1, 2, 0, 56, 0, 63, 0, 118, 14, 228, 1, 2, 0, 56, 0, 63, 0, 118, 14, 67, 15, 2, 0, 56, 0, 63, 0, 118, 14, 231, 1, 2, 0, 56, 0, 63, 0, 118, 14, 234, 1, 2, 0, 56, 0, 63, 0, 118, 14, 237, 1, 2, 0, 56, 0, 63, 0, 118, 14, 240, 1, 2, 0, 56, 0, 63, 0, 118, 14, 243, 1, 2, 0, 56, 0, 63, 0, 118, 14, 246, 1, 2, 0, 56, 0, 63, 0, 118, 14, 249, 1, 2, 0, 56, 0, 63, 0, 118, 14, 252, 1, 2, 0, 56, 0, 63, 0, 118, 14, 255, 1, 2, 0, 56, 0, 63, 0, 118, 14, 2, 2, 2, 0, 56, 0, 63, 0, 118, 14, 5, 2, 2, 0, 56, 0, 63, 0, 118, 14, 8, 2, 2, 0, 56, 0, 63, 0, 118, 14, 11, 2, 2, 0, 56, 0, 63, 0, 118, 14, 14, 2, 2, 0, 56, 0, 63, 0, 118, 14, 17, 2, 2, 0, 56, 0, 63, 0, 183, 14, 4, 0, 2, 0, 56, 0, 63, 0, 118, 14, 4, 0, 2, 0, 56, 0, 63, 0, 118, 14, 4, 0, 2, 0, 56, 0, 63, 0, 183, 14, 126, 6, 2, 0, 56, 0, 63, 0, 118, 14, 126, 6, 2, 0, 56, 0, 63, 0, 118, 14, 151, 8, 2, 0, 56, 0, 63, 0, 118, 14, 195, 6, 2, 0, 56, 0, 63, 0, 118, 14, 21, 11, 2, 0, 56, 0, 63, 0, 183, 14, 195, 6, 2, 0, 56, 0, 63, 0, 118, 14, 219, 6, 2, 0, 56, 0, 63, 0, 118, 14, 20, 2, 2, 0, 56, 0, 63, 0, 183, 14, 10, 0, 2, 0, 56, 0, 63, 0, 183, 14, 30, 14, 2, 0, 56, 0, 63, 0, 183, 14, 38, 14, 2, 0, 56, 0, 63, 0, 183, 14, 106, 0, 2, 0, 56, 0, 63, 0, 183, 14, 142, 4, 2, 0, 56, 0, 63, 0, 183, 14, 46, 14, 37, 1, 160, 8, 166, 8, 220, 11, 234, 6, 172, 8, 178, 8, 227, 11, 10, 11, 184, 8, 234, 11, 241, 11, 92, 15, 103, 15, 54, 14, 194, 14, 190, 8, 248, 11, 240, 6, 17, 11, 24, 11, 246, 6, 255, 11, 238, 10, 196, 8, 114, 15, 202, 8, 208, 8, 214, 8, 31, 11, 125, 15, 62, 14, 205, 14, 6, 12, 13, 12, 220, 8, 20, 12, 38, 11, 226, 8, 27, 12, 34, 12, 41, 12, 48, 12, 55, 12, 232, 8, 252, 6, 238, 8, 244, 8, 204, 6, 2, 7, 62, 12, 96, 9, 69, 12, 76, 12, 83, 12, 8, 7, 45, 11, 14, 7, 90, 12, 102, 9, 20, 7, 136, 15, 108, 9, 97, 12, 114, 9, 26, 7, 32, 7, 104, 12, 126, 9, 147, 15, 70, 14, 158, 15, 132, 9, 52, 11, 38, 7, 132, 6, 111, 12, 68, 7, 74, 7, 118, 12, 174, 9, 245, 10, 180, 9, 80, 7, 59, 11, 186, 9, 66, 11, 86, 7, 169, 15, 92, 7, 73, 11, 98, 7, 204, 9, 104, 7, 125, 12, 110, 7, 116, 7, 132, 12, 210, 9, 80, 11, 216, 9, 222, 9, 139, 12, 122, 7, 128, 7, 134, 7, 180, 15, 140, 7, 78, 14, 146, 12, 153, 12, 146, 7, 228, 9, 87, 11, 160, 12, 167, 12, 174, 12, 181, 12, 94, 11, 188, 12, 195, 12, 234, 9, 162, 6, 202, 12, 152, 7, 209, 12, 158, 7, 240, 9, 164, 7, 216, 12, 223, 12, 170, 7, 230, 12, 237, 12, 244, 12, 176, 7, 246, 9, 168, 6, 251, 12, 182, 7, 252, 9, 2, 13, 2, 10, 188, 7, 9, 13, 194, 7, 8, 10, 16, 13, 23, 13, 200, 7, 30, 13, 37, 13, 101, 11, 206, 7, 108, 11, 44, 13, 212, 7, 218, 7, 14, 10, 174, 6, 191, 15, 51, 13, 58, 13, 20, 10, 202, 15, 180, 6, 230, 7, 65, 13, 72, 13, 186, 6, 115, 11, 122, 11, 129, 11, 32, 10, 38, 10, 231, 10, 236, 7, 44, 10, 79, 13, 242, 7, 86, 13, 136, 11, 50, 10, 93, 13, 143, 11, 248, 7, 100, 13, 56, 10, 62, 10, 68, 10, 254, 7, 107, 13, 238, 14, 6, 14, 249, 14, 74, 10, 86, 14, 4, 8, 10, 8, 16, 8, 150, 11, 80, 10, 22, 8, 28, 8, 114, 13, 86, 10, 34, 8, 121, 13, 92, 10, 128, 13, 135, 13, 142, 13, 98, 10, 149, 13, 104, 10, 156, 13, 110, 10, 163, 13, 157, 11, 40, 8, 61, 16, 145, 14, 13, 16, 46, 8, 170, 13, 52, 8, 58, 8, 177, 13, 184, 13, 191, 13, 164, 11, 216, 6, 64, 8, 70, 8, 94, 14, 4, 15, 70, 15, 116, 10, 171, 11, 122, 10, 102, 14, 235, 15, 82, 8, 128, 10, 178, 11, 185, 11, 134, 10, 140, 10, 192, 11, 88, 8, 94, 8, 146, 10, 198, 13, 100, 8, 152, 10, 106, 8, 158, 10, 252, 10, 164, 10, 170, 10, 199, 11, 3, 11, 118, 8, 124, 8, 130, 8, 15, 15, 246, 15, 14, 14, 216, 14, 73, 16, 136, 14, 25, 16, 200, 10, 206, 10, 205, 13, 212, 13, 136, 8, 219, 13, 226, 13, 206, 11, 212, 10, 218, 10, 233, 13, 213, 11, 240, 13, 142, 8, 37, 16, 154, 14, 49, 16, 247, 13, 110, 14, 227, 14, 26, 15, 148, 8, 170, 170, 164, 1, 47, 0, 75, 0, 79, 0, 83, 0, 87, 0, 91, 0, 95, 0, 99, 0, 103, 0, 107, 0, 111, 0, 115, 0, 119, 0, 123, 0, 127, 0, 131, 0, 142, 0, 146, 0, 150, 0, 154, 0, 158, 0, 162, 0, 166, 0, 170, 0, 174, 0, 180, 0, 184, 0, 188, 0, 192, 0, 196, 0, 200, 0, 204, 0, 208, 0, 212, 0, 216, 0, 220, 0, 223, 0, 227, 0, 231, 0, 235, 0, 239, 0, 243, 0, 247, 0, 253, 0, 1, 1, 5, 1, 9, 1, 13, 1, 17, 1, 21, 1, 33, 1, 37, 1, 41, 1, 45, 1, 49, 1, 53, 1, 57, 1, 61, 1, 65, 1, 69, 1, 73, 1, 77, 1, 81, 1, 85, 1, 89, 1, 93, 1, 97, 1, 101, 1, 105, 1, 109, 1, 113, 1, 117, 1, 121, 1, 125, 1, 129, 1, 133, 1, 137, 1, 141, 1, 145, 1, 149, 1, 153, 1, 157, 1, 161, 1, 165, 1, 169, 1, 173, 1, 177, 1, 181, 1, 185, 1, 189, 1, 193, 1, 197, 1, 201, 1, 205, 1, 209, 1, 213, 1, 217, 1, 221, 1, 225, 1, 229, 1, 233, 1, 237, 1, 241, 1, 245, 1, 249, 1, 253, 1, 1, 2, 5, 2, 9, 2, 13, 2, 17, 2, 21, 2, 25, 2, 29, 2, 33, 2, 37, 2, 41, 2, 45, 2, 49, 2, 53, 2, 57, 2, 68, 2, 79, 2, 90, 2, 94, 2, 98, 2, 102, 2, 106, 2, 110, 2, 114, 2, 118, 2, 122, 2, 126, 2, 130, 2, 134, 2, 138, 2, 142, 2, 146, 2, 156, 2, 167, 2, 175, 2, 181, 2, 187, 2, 193, 2, 203, 2, 209, 2, 218, 2, 224, 2, 230, 2, 236, 2, 242, 2, 246, 2, 250, 2, 254, 2, 2, 3, 6, 3, 10, 3, 14, 3, 18, 3, 198, 5, 22, 3, 26, 3, 30, 3, 34, 3, 38, 3, 42, 3, 45, 3, 49, 3, 5, 5, 53, 3, 57, 3, 60, 3, 64, 3, 68, 3, 72, 3, 76, 3, 80, 3, 84, 3, 88, 3, 92, 3, 96, 3, 100, 3, 104, 3, 108, 3, 112, 3, 123, 3, 127, 3, 131, 3, 135, 3, 139, 3, 143, 3, 147, 3, 151, 3, 155, 3, 159, 3, 163, 3, 174, 3, 178, 3, 182, 3, 186, 3, 190, 3, 194, 3, 198, 3, 209, 3, 213, 3, 217, 3, 221, 3, 225, 3, 229, 3, 233, 3, 237, 3, 241, 3, 245, 3, 249, 3, 253, 3, 1, 4, 5, 4, 9, 4, 13, 4, 17, 4, 21, 4, 25, 4, 29, 4, 33, 4, 37, 4, 41, 4, 45, 4, 49, 4, 53, 4, 57, 4, 61, 4, 72, 4, 245, 8, 76, 4, 80, 4, 84, 4, 88, 4, 99, 4, 110, 4, 121, 4, 125, 4, 129, 4, 133, 4, 137, 4, 141, 4, 145, 4, 149, 4, 153, 4, 157, 4, 161, 4, 165, 4, 169, 4, 173, 4, 177, 4, 181, 4, 185, 4, 189, 4, 193, 4, 186, 4, 197, 4, 207, 4, 218, 4, 225, 4, 232, 4, 236, 4, 240, 4, 244, 4, 248, 4, 252, 4, 0, 5, 4, 5, 8, 5, 12, 5, 151, 10, 16, 5, 20, 5, 24, 5, 30, 5, 36, 5, 40, 5, 44, 5, 48, 5, 52, 5, 56, 5, 60, 5, 64, 5, 68, 5, 72, 5, 76, 5, 80, 5, 84, 5, 88, 5, 92, 5, 96, 5, 100, 5, 104, 5, 108, 5, 112, 5, 116, 5, 120, 5, 124, 5, 128, 5, 132, 5, 136, 5, 140, 5, 144, 5, 148, 5, 152, 5, 156, 5, 160, 5, 170, 5, 180, 5, 241, 9, 190, 5, 197, 5, 201, 5, 205, 5, 209, 5, 213, 5, 217, 5, 221, 5, 225, 5, 229, 5, 233, 5, 237, 5, 241, 5, 245, 5, 249, 5, 253, 5, 3, 6, 9, 6, 15, 6, 21, 6, 25, 6, 29, 6, 33, 6, 37, 6, 41, 6, 45, 6, 49, 6, 53, 6, 57, 6, 61, 6, 65, 6, 69, 6, 73, 6, 77, 6, 81, 6, 85, 6, 89, 6, 93, 6, 97, 6, 101, 6, 105, 6, 109, 6, 113, 6, 116, 6, 120, 6, 124, 6, 128, 6, 132, 6, 136, 6, 140, 6, 144, 6, 148, 6, 152, 6, 156, 6, 160, 6, 163, 6, 167, 6, 179, 6, 190, 6, 194, 6, 198, 6, 202, 6, 206, 6, 210, 6, 216, 6, 222, 6, 226, 6, 230, 6, 237, 6, 241, 6, 245, 6, 249, 6, 253, 6, 1, 7, 5, 7, 9, 7, 13, 7, 17, 7, 21, 7, 25, 7, 29, 7, 33, 7, 37, 7, 41, 7, 45, 7, 49, 7, 53, 7, 57, 7, 61, 7, 68, 7, 75, 7, 79, 7, 83, 7, 90, 7, 96, 7, 102, 7, 108, 7, 114, 7, 120, 7, 129, 7, 138, 7, 149, 7, 158, 7, 162, 7, 166, 7, 170, 7, 174, 7, 170, 170, 207, 17, 0, 80, 212, 17, 0, 80, 217, 17, 0, 80, 222, 17, 0, 80, 227, 17, 0, 80, 232, 17, 0, 80, 237, 17, 0, 80, 242, 17, 0, 80, 247, 17, 0, 80, 252, 17, 0, 80, 1, 18, 0, 80, 6, 18, 0, 80, 11, 18, 0, 80, 16, 18, 0, 80, 21, 18, 0, 80, 26, 18, 0, 80, 31, 18, 0, 80, 36, 18, 0, 80, 41, 18, 0, 80, 46, 18, 0, 80, 51, 18, 0, 80, 56, 18, 0, 80, 61, 18, 0, 80, 66, 18, 0, 80, 71, 18, 0, 80, 76, 18, 0, 80, 81, 18, 0, 80, 86, 18, 0, 80, 91, 18, 0, 80, 96, 18, 0, 80, 101, 18, 0, 80, 106, 18, 0, 80, 111, 18, 0, 80, 116, 18, 0, 80, 121, 18, 0, 80, 126, 18, 0, 80, 131, 18, 0, 80, 136, 18, 0, 80, 141, 18, 0, 80, 146, 18, 0, 80, 151, 18, 0, 80, 156, 18, 0, 80, 161, 18, 0, 80, 166, 18, 0, 80, 171, 18, 0, 80, 176, 18, 0, 80, 181, 18, 0, 80, 186, 18, 0, 80, 191, 18, 0, 80, 196, 18, 0, 80, 201, 18, 0, 80, 206, 18, 0, 80, 211, 18, 0, 80, 216, 18, 0, 80, 221, 18, 0, 80, 226, 18, 0, 80, 231, 18, 0, 80, 236, 18, 0, 80, 241, 18, 0, 80, 246, 18, 0, 80, 251, 18, 0, 80, 0, 19, 0, 80, 5, 19, 0, 80, 10, 19, 0, 80, 15, 19, 0, 80, 20, 19, 0, 80, 25, 19, 0, 80, 30, 19, 0, 80, 35, 19, 0, 80, 40, 19, 0, 80, 45, 19, 0, 80, 50, 19, 0, 80, 55, 19, 0, 80, 60, 19, 0, 80, 65, 19, 0, 80, 70, 19, 0, 80, 75, 19, 0, 80, 80, 19, 0, 80, 85, 19, 0, 80, 90, 19, 0, 80, 95, 19, 0, 80, 100, 19, 0, 80, 105, 19, 0, 80, 110, 19, 0, 80, 115, 19, 0, 80, 120, 19, 0, 80, 125, 19, 0, 80, 130, 19, 0, 80, 135, 19, 0, 80, 140, 19, 0, 80, 145, 19, 0, 80, 150, 19, 0, 80, 155, 19, 0, 80, 160, 19, 0, 80, 165, 19, 0, 80, 170, 19, 0, 80, 175, 19, 0, 80, 180, 19, 0, 80, 185, 19, 0, 80, 190, 19, 0, 80, 195, 19, 0, 80, 200, 19, 0, 80, 205, 19, 0, 80, 210, 19, 0, 80, 215, 19, 0, 80, 220, 19, 0, 80, 225, 19, 0, 80, 230, 19, 0, 80, 235, 19, 0, 80, 240, 19, 0, 80, 245, 19, 0, 80, 250, 19, 0, 80, 255, 19, 0, 80, 4, 20, 0, 80, 9, 20, 0, 80, 14, 20, 0, 80, 19, 20, 0, 80, 24, 20, 0, 80, 29, 20, 0, 80, 34, 20, 0, 80, 39, 20, 0, 80, 44, 20, 0, 80, 49, 20, 0, 80, 54, 20, 0, 80, 59, 20, 0, 80, 64, 20, 0, 80, 69, 20, 0, 80, 74, 20, 0, 80, 79, 20, 0, 80, 84, 20, 0, 80, 89, 20, 0, 80, 94, 20, 0, 80, 99, 20, 0, 80, 104, 20, 0, 80, 109, 20, 0, 80, 114, 20, 0, 80, 119, 20, 0, 80, 124, 20, 0, 80, 129, 20, 0, 80, 134, 20, 0, 80, 139, 20, 0, 80, 144, 20, 0, 80, 149, 20, 0, 80, 154, 20, 0, 80, 159, 20, 0, 80, 164, 20, 0, 80, 169, 20, 0, 80, 174, 20, 0, 80, 179, 20, 0, 80, 184, 20, 0, 80, 189, 20, 0, 80, 194, 20, 0, 80, 199, 20, 0, 80, 204, 20, 0, 80, 209, 20, 0, 80, 214, 20, 0, 80, 219, 20, 0, 80, 224, 20, 0, 80, 229, 20, 0, 80, 234, 20, 0, 80, 239, 20, 0, 80, 244, 20, 0, 80, 249, 20, 0, 80, 254, 20, 0, 80, 3, 21, 0, 80, 8, 21, 0, 80, 13, 21, 0, 80, 18, 21, 0, 80, 23, 21, 0, 80, 28, 21, 0, 80, 33, 21, 0, 80, 38, 21, 0, 80, 43, 21, 0, 80, 48, 21, 0, 80, 53, 21, 0, 80, 58, 21, 0, 80, 63, 21, 0, 80, 68, 21, 0, 80, 73, 21, 0, 80, 78, 21, 0, 80, 83, 21, 0, 80, 88, 21, 0, 80, 93, 21, 0, 80, 98, 21, 0, 80, 103, 21, 0, 80, 108, 21, 0, 80, 113, 21, 0, 80, 118, 21, 0, 80, 123, 21, 0, 80, 128, 21, 0, 80, 133, 21, 0, 80, 138, 21, 0, 80, 143, 21, 0, 80, 148, 21, 0, 80, 153, 21, 0, 80, 158, 21, 0, 80, 163, 21, 0, 80, 168, 21, 0, 80, 173, 21, 0, 80, 178, 21, 0, 80, 183, 21, 0, 80, 188, 21, 0, 80, 193, 21, 0, 80, 198, 21, 0, 80, 203, 21, 0, 80, 208, 21, 0, 80, 213, 21, 0, 80, 218, 21, 0, 80, 223, 21, 0, 80, 228, 21, 0, 80, 233, 21, 0, 80, 238, 21, 0, 80, 243, 21, 0, 80, 248, 21, 0, 80, 253, 21, 0, 80, 2, 22, 0, 80, 7, 22, 0, 80, 12, 22, 0, 80, 17, 22, 0, 80, 22, 22, 0, 80, 27, 22, 0, 80, 32, 22, 0, 80, 37, 22, 0, 80, 42, 22, 0, 80, 47, 22, 0, 80, 52, 22, 0, 80, 57, 22, 0, 80, 62, 22, 0, 80, 67, 22, 0, 80, 72, 22, 0, 80, 77, 22, 0, 80, 82, 22, 0, 80, 87, 22, 0, 80, 92, 22, 0, 80, 97, 22, 0, 80, 102, 22, 0, 80, 107, 22, 0, 80, 112, 22, 0, 80, 117, 22, 0, 80, 122, 22, 0, 80, 127, 22, 0, 80, 132, 22, 0, 80, 137, 22, 0, 80, 142, 22, 0, 80, 147, 22, 0, 80, 152, 22, 0, 80, 157, 22, 0, 80, 162, 22, 0, 80, 167, 22, 0, 80, 172, 22, 0, 80, 177, 22, 0, 80, 182, 22, 0, 80, 187, 22, 0, 80, 192, 22, 0, 80, 197, 22, 0, 80, 202, 22, 0, 80, 207, 22, 0, 80, 212, 22, 0, 80, 217, 22, 0, 80, 222, 22, 0, 80, 227, 22, 0, 80, 232, 22, 0, 80, 237, 22, 0, 80, 242, 22, 0, 80, 247, 22, 0, 80, 252, 22, 0, 80, 1, 23, 0, 80, 6, 23, 0, 80, 11, 23, 0, 80, 16, 23, 0, 80, 21, 23, 0, 80, 26, 23, 0, 80, 31, 23, 0, 80, 36, 23, 0, 80, 41, 23, 0, 80, 46, 23, 0, 80, 51, 23, 0, 80, 56, 23, 0, 80, 61, 23, 0, 80, 66, 23, 0, 80, 71, 23, 0, 80, 76, 23, 0, 80, 81, 23, 0, 80, 86, 23, 0, 80, 91, 23, 0, 80, 96, 23, 0, 80, 101, 23, 0, 80, 106, 23, 0, 80, 111, 23, 0, 80, 116, 23, 0, 80, 121, 23, 0, 80, 126, 23, 0, 80, 131, 23, 0, 80, 136, 23, 0, 80, 141, 23, 0, 80, 146, 23, 0, 80, 151, 23, 0, 80, 156, 23, 0, 80, 161, 23, 0, 80, 166, 23, 0, 80, 171, 23, 0, 80, 176, 23, 0, 80, 181, 23, 0, 80, 186, 23, 0, 80, 191, 23, 0, 80, 196, 23, 0, 80, 201, 23, 0, 80, 206, 23, 0, 80, 211, 23, 0, 80, 216, 23, 0, 80, 221, 23, 0, 80, 226, 23, 0, 80, 231, 23, 0, 80, 236, 23, 0, 80, 241, 23, 0, 80, 246, 23, 0, 80, 251, 23, 0, 80, 0, 24, 0, 80, 5, 24, 0, 80, 10, 24, 0, 80, 15, 24, 0, 80, 20, 24, 0, 80, 25, 24, 0, 80, 30, 24, 0, 80, 35, 24, 0, 80, 40, 24, 0, 80, 45, 24, 0, 80, 50, 24, 0, 80, 55, 24, 0, 80, 60, 24, 0, 80, 65, 24, 0, 80, 70, 24, 0, 80, 75, 24, 0, 80, 80, 24, 0, 80, 85, 24, 0, 80, 90, 24, 0, 80, 95, 24, 0, 80, 100, 24, 0, 80, 105, 24, 0, 80, 110, 24, 0, 80, 115, 24, 0, 80, 120, 24, 0, 80, 125, 24, 0, 80, 130, 24, 0, 80, 135, 24, 0, 80, 140, 24, 0, 80, 145, 24, 0, 80, 150, 24, 0, 80, 155, 24, 0, 80, 160, 24, 0, 80, 165, 24, 0, 80, 170, 24, 0, 80, 175, 24, 0, 80, 180, 24, 0, 80, 185, 24, 0, 80, 190, 24, 0, 80, 195, 24, 0, 80, 200, 24, 0, 80, 205, 24, 0, 80, 210, 24, 0, 80, 215, 24, 0, 80, 220, 24, 0, 80, 225, 24, 0, 80, 230, 24, 0, 80, 235, 24, 0, 80, 240, 24, 0, 80, 245, 24, 0, 80, 250, 24, 0, 80, 255, 24, 0, 80, 4, 25, 0, 80, 9, 25, 0, 80, 14, 25, 0, 80, 19, 25, 0, 80, 24, 25, 0, 80, 29, 25, 0, 80, 34, 25, 0, 80, 39, 25, 0, 80, 44, 25, 0, 80, 49, 25, 0, 80, 54, 25, 0, 80, 59, 25, 0, 80, 64, 25, 0, 80, 69, 25, 0, 80, 74, 25, 0, 80, 79, 25, 0, 80, 84, 25, 0, 80, 89, 25, 0, 80, 94, 25, 0, 80, 99, 25, 0, 80, 104, 25, 0, 80, 109, 25, 0, 80, 114, 25, 0, 80, 119, 25, 0, 80, 124, 25, 0, 80, 129, 25, 0, 80, 134, 25, 0, 80, 139, 25, 0, 80, 144, 25, 0, 80, 149, 25, 0, 80, 154, 25, 0, 80, 159, 25, 0, 80, 164, 25, 0, 80, 169, 25, 0, 80, 174, 25, 0, 80, 179, 25, 0, 80, 184, 25, 0, 80, 189, 25, 0, 80, 194, 25, 0, 80, 199, 25, 0, 80, 204, 25, 0, 80, 209, 25, 0, 80, 214, 25, 0, 80, 219, 25, 0, 80, 224, 25, 0, 80, 229, 25, 0, 80, 234, 25, 0, 80, 239, 25, 0, 80, 244, 25, 0, 80, 249, 25, 0, 80, 254, 25, 0, 80, 1, 0, 185, 7, 3, 26, 0, 80, 145, 0, 202, 7, 207, 7, 212, 7, 217, 7, 222, 7, 227, 7, 232, 7, 237, 7, 242, 7, 247, 7, 252, 7, 1, 8, 6, 8, 11, 8, 16, 8, 21, 8, 26, 8, 31, 8, 36, 8, 41, 8, 46, 8, 51, 8, 56, 8, 61, 8, 66, 8, 71, 8, 76, 8, 81, 8, 86, 8, 91, 8, 96, 8, 101, 8, 106, 8, 111, 8, 116, 8, 122, 8, 128, 8, 134, 8, 140, 8, 146, 8, 152, 8, 158, 8, 164, 8, 170, 8, 176, 8, 182, 8, 188, 8, 194, 8, 200, 8, 206, 8, 212, 8, 218, 8, 223, 8, 228, 8, 233, 8, 238, 8, 243, 8, 248, 8, 253, 8, 2, 9, 7, 9, 12, 9, 17, 9, 22, 9, 27, 9, 32, 9, 37, 9, 217, 1, 42, 9, 46, 9, 51, 9, 55, 9, 60, 9, 221, 1, 64, 9, 68, 9, 72, 9, 77, 9, 82, 9, 86, 9, 90, 9, 94, 9, 98, 9, 102, 9, 106, 9, 111, 9, 116, 9, 120, 9, 125, 9, 129, 9, 133, 9, 138, 9, 143, 9, 147, 9, 151, 9, 156, 9, 160, 9, 164, 9, 169, 9, 173, 9, 177, 9, 182, 9, 187, 9, 192, 9, 9, 4, 196, 9, 200, 9, 206, 9, 211, 9, 216, 9, 221, 9, 226, 9, 230, 9, 234, 9, 239, 9, 244, 9, 249, 9, 254, 9, 3, 10, 8, 10, 13, 10, 18, 10, 23, 10, 28, 10, 33, 10, 38, 10, 43, 10, 48, 10, 53, 10, 58, 10, 63, 10, 68, 10, 73, 10, 79, 10, 85, 10, 91, 10, 97, 10, 103, 10, 109, 10, 115, 10, 120, 10, 125, 10, 130, 10, 135, 10, 140, 10, 8, 26, 0, 80, 13, 26, 0, 80, 18, 26, 0, 80, 23, 26, 0, 80, 28, 26, 0, 80, 33, 26, 0, 80, 38, 26, 0, 80, 43, 26, 0, 80, 48, 26, 0, 80, 53, 26, 0, 80, 58, 26, 0, 80, 63, 26, 0, 80, 68, 26, 0, 80, 73, 26, 0, 80, 78, 26, 0, 80, 83, 26, 0, 80, 88, 26, 0, 80, 93, 26, 0, 80, 98, 26, 0, 80, 103, 26, 0, 80, 108, 26, 0, 80, 113, 26, 0, 80, 118, 26, 0, 80, 123, 26, 0, 80, 128, 26, 0, 80, 133, 26, 0, 80, 138, 26, 0, 80, 143, 26, 0, 80, 148, 26, 0, 80, 153, 26, 0, 80, 158, 26, 0, 80, 163, 26, 0, 80, 168, 26, 0, 80, 173, 26, 0, 80, 178, 26, 0, 80, 183, 26, 0, 80, 188, 26, 0, 80, 193, 26, 0, 80, 198, 26, 0, 80, 203, 26, 0, 80, 208, 26, 0, 80, 213, 26, 0, 80, 218, 26, 0, 80, 223, 26, 0, 80, 228, 26, 0, 80, 233, 26, 0, 80, 238, 26, 0, 80, 243, 26, 0, 80, 248, 26, 0, 80, 253, 26, 0, 80, 2, 27, 0, 80, 7, 27, 0, 80, 12, 27, 0, 80, 17, 27, 0, 80, 22, 27, 0, 80, 27, 27, 0, 80, 32, 27, 0, 80, 37, 27, 0, 80, 42, 27, 0, 80, 47, 27, 0, 80, 52, 27, 0, 80, 57, 27, 0, 80, 62, 27, 0, 80, 67, 27, 0, 80, 72, 27, 0, 80, 77, 27, 0, 80, 82, 27, 0, 80, 87, 27, 0, 80, 92, 27, 0, 80, 97, 27, 0, 80, 102, 27, 0, 80, 107, 27, 0, 80, 112, 27, 0, 80, 117, 27, 0, 80, 122, 27, 0, 80, 127, 27, 0, 80, 132, 27, 0, 80, 137, 27, 0, 80, 142, 27, 0, 80, 147, 27, 0, 80, 152, 27, 0, 80, 157, 27, 0, 80, 162, 27, 0, 80, 167, 27, 0, 80, 172, 27, 0, 80, 177, 27, 0, 80, 182, 27, 0, 80, 187, 27, 0, 80, 192, 27, 0, 80, 197, 27, 0, 80, 202, 27, 0, 80, 207, 27, 0, 80, 212, 27, 0, 80, 217, 27, 0, 80, 222, 27, 0, 80, 227, 27, 0, 80, 232, 27, 0, 80, 237, 27, 0, 80, 242, 27, 0, 80, 247, 27, 0, 80, 252, 27, 0, 80, 1, 28, 0, 80, 6, 28, 0, 80, 11, 28, 0, 80, 16, 28, 0, 80, 21, 28, 0, 80, 26, 28, 0, 80, 31, 28, 0, 80, 36, 28, 0, 80, 41, 28, 0, 80, 46, 28, 0, 80, 51, 28, 0, 80, 56, 28, 0, 80, 61, 28, 0, 80, 66, 28, 0, 80, 71, 28, 0, 80, 76, 28, 0, 80, 81, 28, 0, 80, 86, 28, 0, 80, 91, 28, 0, 80, 96, 28, 0, 80, 101, 28, 0, 80, 106, 28, 0, 80, 111, 28, 0, 80, 116, 28, 0, 80, 121, 28, 0, 80, 126, 28, 0, 80, 131, 28, 0, 80, 136, 28, 0, 80, 141, 28, 0, 80, 146, 28, 0, 80, 151, 28, 0, 80, 156, 28, 0, 80, 161, 28, 0, 80, 166, 28, 0, 80, 171, 28, 0, 80, 176, 28, 0, 80, 181, 28, 0, 80, 186, 28, 0, 80, 191, 28, 0, 80, 196, 28, 0, 80, 201, 28, 0, 80, 206, 28, 0, 80, 211, 28, 0, 80, 216, 28, 0, 80, 128, 2, 155, 10, 159, 10, 163, 10, 167, 10, 171, 10, 175, 10, 179, 10, 183, 10, 187, 10, 191, 10, 195, 10, 199, 10, 203, 10, 207, 10, 211, 10, 215, 10, 219, 10, 223, 10, 227, 10, 231, 10, 235, 10, 239, 10, 243, 10, 247, 10, 251, 10, 255, 10, 3, 11, 7, 11, 11, 11, 15, 11, 19, 11, 23, 11, 27, 11, 31, 11, 35, 11, 39, 11, 43, 11, 47, 11, 51, 11, 55, 11, 59, 11, 63, 11, 67, 11, 71, 11, 75, 11, 79, 11, 83, 11, 87, 11, 91, 11, 95, 11, 99, 11, 103, 11, 107, 11, 111, 11, 115, 11, 119, 11, 123, 11, 127, 11, 131, 11, 135, 11, 139, 11, 143, 11, 147, 11, 151, 11, 155, 11, 159, 11, 163, 11, 167, 11, 171, 11, 175, 11, 179, 11, 183, 11, 187, 11, 191, 11, 195, 11, 199, 11, 203, 11, 207, 11, 211, 11, 215, 11, 219, 11, 223, 11, 227, 11, 231, 11, 235, 11, 239, 11, 243, 11, 247, 11, 251, 11, 255, 11, 3, 12, 7, 12, 11, 12, 15, 12, 19, 12, 23, 12, 27, 12, 31, 12, 35, 12, 39, 12, 43, 12, 47, 12, 51, 12, 55, 12, 59, 12, 63, 12, 67, 12, 71, 12, 75, 12, 79, 12, 83, 12, 87, 12, 91, 12, 95, 12, 99, 12, 103, 12, 107, 12, 111, 12, 115, 12, 119, 12, 123, 12, 127, 12, 131, 12, 135, 12, 139, 12, 143, 12, 147, 12, 151, 12, 155, 12, 159, 12, 163, 12, 167, 12, 171, 12, 175, 12, 179, 12, 183, 12, 187, 12, 191, 12, 195, 12, 199, 12, 203, 12, 207, 12, 211, 12, 215, 12, 219, 12, 223, 12, 227, 12, 231, 12, 235, 12, 239, 12, 243, 12, 247, 12, 251, 12, 255, 12, 3, 13, 7, 13, 11, 13, 15, 13, 19, 13, 23, 13, 27, 13, 31, 13, 35, 13, 39, 13, 43, 13, 47, 13, 51, 13, 55, 13, 59, 13, 63, 13, 67, 13, 71, 13, 75, 13, 79, 13, 83, 13, 87, 13, 91, 13, 95, 13, 99, 13, 103, 13, 107, 13, 111, 13, 115, 13, 119, 13, 123, 13, 127, 13, 160, 8, 131, 13, 135, 13, 202, 8, 208, 8, 139, 13, 143, 13, 147, 13, 151, 13, 155, 13, 159, 13, 163, 13, 167, 13, 171, 13, 175, 13, 179, 13, 183, 13, 187, 13, 191, 13, 195, 13, 199, 13, 203, 13, 207, 13, 211, 13, 215, 13, 219, 13, 223, 13, 227, 13, 231, 13, 235, 13, 239, 13, 243, 13, 247, 13, 251, 13, 255, 13, 3, 14, 7, 14, 11, 14, 15, 14, 19, 14, 23, 14, 27, 14, 31, 14, 35, 14, 39, 14, 43, 14, 47, 14, 51, 14, 55, 14, 59, 14, 63, 14, 67, 14, 71, 14, 75, 14, 79, 14, 83, 14, 87, 14, 91, 14, 95, 14, 99, 14, 103, 14, 107, 14, 111, 14, 115, 14, 119, 14, 123, 14, 127, 14, 131, 14, 135, 14, 139, 14, 143, 14, 147, 14, 151, 14, 155, 14, 159, 14, 163, 14, 167, 14, 171, 14, 175, 14, 179, 14, 183, 14, 187, 14, 191, 14, 195, 14, 199, 14, 203, 14, 207, 14, 211, 14, 215, 14, 219, 14, 223, 14, 227, 14, 231, 14, 235, 14, 239, 14, 243, 14, 247, 14, 251, 14, 255, 14, 3, 15, 7, 15, 11, 15, 15, 15, 19, 15, 23, 15, 27, 15, 31, 15, 35, 15, 39, 15, 43, 15, 47, 15, 51, 15, 55, 15, 59, 15, 63, 15, 67, 15, 71, 15, 75, 15, 79, 15, 83, 15, 87, 15, 91, 15, 95, 15, 99, 15, 103, 15, 107, 15, 111, 15, 115, 15, 119, 15, 76, 18, 102, 20, 123, 15, 127, 15, 131, 15, 135, 15, 139, 15, 143, 15, 147, 15, 151, 15, 155, 15, 159, 15, 163, 15, 167, 15, 171, 15, 175, 15, 179, 15, 183, 15, 187, 15, 191, 15, 195, 15, 199, 15, 203, 15, 207, 15, 211, 15, 215, 15, 219, 15, 223, 15, 227, 15, 231, 15, 235, 15, 239, 15, 243, 15, 247, 15, 250, 15, 254, 15, 2, 16, 6, 16, 10, 16, 14, 16, 18, 16, 22, 16, 26, 16, 30, 16, 34, 16, 38, 16, 42, 16, 46, 16, 50, 16, 54, 16, 58, 16, 62, 16, 66, 16, 70, 16, 165, 19, 73, 16, 77, 16, 81, 16, 85, 16, 89, 16, 93, 16, 215, 19, 97, 16, 101, 16, 105, 16, 109, 16, 113, 16, 117, 16, 121, 16, 125, 16, 128, 16, 132, 16, 136, 16, 140, 16, 144, 16, 148, 16, 152, 16, 156, 16, 160, 16, 164, 16, 168, 16, 172, 16, 175, 16, 179, 16, 183, 16, 187, 16, 190, 16, 194, 16, 198, 16, 202, 16, 206, 16, 210, 16, 214, 16, 218, 16, 222, 16, 226, 16, 230, 16, 234, 16, 238, 16, 242, 16, 246, 16, 250, 16, 254, 16, 2, 17, 6, 17, 10, 17, 14, 17, 18, 17, 22, 17, 26, 17, 30, 17, 34, 17, 37, 17, 41, 17, 45, 17, 49, 17, 53, 17, 57, 17, 61, 17, 65, 17, 69, 17, 73, 17, 77, 17, 81, 17, 85, 17, 89, 17, 93, 17, 96, 17, 100, 17, 104, 17, 108, 17, 112, 17, 116, 17, 120, 17, 124, 17, 128, 17, 132, 17, 136, 17, 140, 17, 144, 17, 148, 17, 152, 17, 156, 17, 160, 17, 164, 17, 168, 17, 172, 17, 176, 17, 180, 17, 184, 17, 188, 17, 192, 17, 196, 17, 200, 17, 204, 17, 145, 19, 208, 17, 212, 17, 216, 17, 220, 17, 224, 17, 228, 17, 232, 17, 236, 17, 240, 17, 244, 17, 248, 17, 252, 17, 0, 18, 4, 18, 8, 18, 12, 18, 16, 18, 20, 18, 24, 18, 28, 18, 32, 18, 35, 18, 39, 18, 43, 18, 47, 18, 51, 18, 235, 16, 55, 18, 124, 15, 59, 18, 63, 18, 67, 18, 71, 18, 75, 18, 79, 18, 82, 18, 86, 18, 90, 18, 94, 18, 98, 18, 102, 18, 106, 18, 110, 18, 114, 18, 118, 18, 122, 18, 126, 18, 129, 18, 133, 18, 136, 18, 140, 18, 144, 18, 148, 18, 152, 18, 156, 18, 160, 18, 164, 18, 168, 18, 171, 18, 175, 18, 179, 18, 183, 18, 187, 18, 191, 18, 195, 18, 199, 18, 202, 18, 206, 18, 210, 18, 214, 18, 218, 18, 222, 18, 226, 18, 230, 18, 234, 18, 238, 18, 242, 18, 246, 18, 250, 18, 254, 18, 2, 19, 6, 19, 10, 19, 14, 19, 18, 19, 22, 19, 26, 19, 29, 19, 33, 19, 37, 19, 41, 19, 45, 19, 49, 19, 53, 19, 57, 19, 61, 19, 65, 19, 69, 19, 73, 19, 77, 19, 81, 19, 85, 19, 89, 19, 93, 19, 97, 19, 101, 19, 105, 19, 23, 19, 109, 19, 113, 19, 117, 19, 121, 19, 125, 19, 129, 19, 133, 19, 137, 19, 140, 19, 144, 19, 148, 19, 152, 19, 156, 19, 160, 19, 164, 19, 168, 19, 171, 19, 175, 19, 179, 19, 183, 19, 187, 19, 191, 19, 194, 19, 198, 19, 202, 19, 206, 19, 210, 19, 214, 19, 218, 19, 222, 19, 226, 19, 230, 19, 234, 19, 238, 19, 242, 19, 246, 19, 250, 19, 254, 19, 2, 20, 6, 20, 10, 20, 14, 20, 18, 20, 22, 20, 26, 20, 30, 20, 34, 20, 38, 20, 42, 20, 46, 20, 50, 20, 53, 20, 57, 20, 61, 20, 64, 20, 68, 20, 72, 20, 76, 20, 80, 20, 83, 20, 87, 20, 170, 170, 221, 28, 0, 80, 226, 28, 0, 80, 231, 28, 0, 80, 236, 28, 0, 80, 241, 28, 0, 80, 246, 28, 0, 80, 251, 28, 0, 80, 0, 29, 0, 80, 5, 29, 0, 80, 10, 29, 0, 80, 15, 29, 0, 80, 20, 29, 0, 80, 25, 29, 0, 80, 30, 29, 0, 80, 35, 29, 0, 80, 40, 29, 0, 80, 45, 29, 0, 80, 50, 29, 0, 80, 55, 29, 0, 80, 60, 29, 0, 80, 65, 29, 0, 80, 70, 29, 0, 80, 75, 29, 0, 80, 80, 29, 0, 80, 85, 29, 0, 80, 90, 29, 0, 80, 95, 29, 0, 80, 100, 29, 0, 80, 105, 29, 0, 80, 110, 29, 0, 80, 115, 29, 0, 80, 120, 29, 0, 80, 125, 29, 0, 80, 130, 29, 0, 80, 135, 29, 0, 80, 140, 29, 0, 80, 145, 29, 0, 80, 150, 29, 0, 80, 155, 29, 0, 80, 160, 29, 0, 80, 165, 29, 0, 80, 170, 29, 0, 80, 175, 29, 0, 80, 180, 29, 0, 80, 185, 29, 0, 80, 190, 29, 0, 80, 195, 29, 0, 80, 200, 29, 0, 80, 205, 29, 0, 80, 210, 29, 0, 80, 215, 29, 0, 80, 220, 29, 0, 80, 225, 29, 0, 80, 230, 29, 0, 80, 235, 29, 0, 80, 240, 29, 0, 80, 245, 29, 0, 80, 250, 29, 0, 80, 255, 29, 0, 80, 4, 30, 0, 80, 9, 30, 0, 80, 14, 30, 0, 80, 19, 30, 0, 80, 24, 30, 0, 80, 29, 30, 0, 80, 34, 30, 0, 80, 39, 30, 0, 80, 44, 30, 0, 80, 49, 30, 0, 80, 54, 30, 0, 80, 59, 30, 0, 80, 64, 30, 0, 80, 69, 30, 0, 80, 74, 30, 0, 80, 79, 30, 0, 80, 84, 30, 0, 80, 89, 30, 0, 80, 94, 30, 0, 80, 99, 30, 0, 80, 104, 30, 0, 80, 109, 30, 0, 80, 114, 30, 0, 80, 119, 30, 0, 80, 124, 30, 0, 80, 129, 30, 0, 80, 134, 30, 0, 80, 139, 30, 0, 80, 144, 30, 0, 80, 149, 30, 0, 80, 154, 30, 0, 80, 159, 30, 0, 80, 164, 30, 0, 80, 169, 30, 0, 80, 174, 30, 0, 80, 179, 30, 0, 80, 184, 30, 0, 80, 189, 30, 0, 80, 194, 30, 0, 80, 199, 30, 0, 80, 204, 30, 0, 80, 209, 30, 0, 80, 214, 30, 0, 80, 219, 30, 0, 80, 224, 30, 0, 80, 229, 30, 0, 80, 234, 30, 0, 80, 239, 30, 0, 80, 244, 30, 0, 80, 249, 30, 0, 80, 254, 30, 0, 80, 3, 31, 0, 80, 8, 31, 0, 80, 13, 31, 0, 80, 18, 31, 0, 80, 23, 31, 0, 80, 28, 31, 0, 80, 33, 31, 0, 80, 38, 31, 0, 80, 43, 31, 0, 80, 48, 31, 0, 80, 53, 31, 0, 80, 58, 31, 0, 80, 63, 31, 0, 80, 68, 31, 0, 80, 73, 31, 0, 80, 78, 31, 0, 80, 83, 31, 0, 80, 88, 31, 0, 80, 93, 31, 0, 80, 98, 31, 0, 80, 103, 31, 0, 80, 108, 31, 0, 80, 113, 31, 0, 80, 118, 31, 0, 80, 123, 31, 0, 80, 128, 31, 0, 80, 133, 31, 0, 80, 138, 31, 0, 80, 143, 31, 0, 80, 148, 31, 0, 80, 153, 31, 0, 80, 158, 31, 0, 80, 163, 31, 0, 80, 168, 31, 0, 80, 173, 31, 0, 80, 178, 31, 0, 80, 183, 31, 0, 80, 188, 31, 0, 80, 193, 31, 0, 80, 198, 31, 0, 80, 203, 31, 0, 80, 208, 31, 0, 80, 213, 31, 0, 80, 218, 31, 0, 80, 223, 31, 0, 80, 228, 31, 0, 80, 233, 31, 0, 80, 238, 31, 0, 80, 243, 31, 0, 80, 248, 31, 0, 80, 253, 31, 0, 80, 2, 32, 0, 80, 7, 32, 0, 80, 12, 32, 0, 80, 17, 32, 0, 80, 22, 32, 0, 80, 27, 32, 0, 80, 32, 32, 0, 80, 37, 32, 0, 80, 42, 32, 0, 80, 47, 32, 0, 80, 52, 32, 0, 80, 57, 32, 0, 80, 62, 32, 0, 80, 67, 32, 0, 80, 72, 32, 0, 80, 77, 32, 0, 80, 82, 32, 0, 80, 87, 32, 0, 80, 92, 32, 0, 80, 97, 32, 0, 80, 102, 32, 0, 80, 107, 32, 0, 80, 112, 32, 0, 80, 117, 32, 0, 80, 122, 32, 0, 80, 127, 32, 0, 80, 132, 32, 0, 80, 137, 32, 0, 80, 142, 32, 0, 80, 147, 32, 0, 80, 152, 32, 0, 80, 157, 32, 0, 80, 162, 32, 0, 80, 167, 32, 0, 80, 172, 32, 0, 80, 177, 32, 0, 80, 182, 32, 0, 80, 187, 32, 0, 80, 192, 32, 0, 80, 197, 32, 0, 80, 202, 32, 0, 80, 207, 32, 0, 80, 212, 32, 0, 80, 217, 32, 0, 80, 222, 32, 0, 80, 227, 32, 0, 80, 232, 32, 0, 80, 237, 32, 0, 80, 242, 32, 0, 80, 247, 32, 0, 80, 252, 32, 0, 80, 1, 33, 0, 80, 6, 33, 0, 80, 11, 33, 0, 80, 16, 33, 0, 80, 21, 33, 0, 80, 26, 33, 0, 80, 31, 33, 0, 80, 36, 33, 0, 80, 41, 33, 0, 80, 46, 33, 0, 80, 51, 33, 0, 80, 56, 33, 0, 80, 61, 33, 0, 80, 66, 33, 0, 80, 71, 33, 0, 80, 76, 33, 0, 80, 81, 33, 0, 80, 86, 33, 0, 80, 91, 33, 0, 80, 96, 33, 0, 80, 101, 33, 0, 80, 106, 33, 0, 80, 111, 33, 0, 80, 116, 33, 0, 80, 121, 33, 0, 80, 126, 33, 0, 80, 131, 33, 0, 80, 136, 33, 0, 80, 141, 33, 0, 80, 146, 33, 0, 80, 151, 33, 0, 80, 156, 33, 0, 80, 161, 33, 0, 80, 166, 33, 0, 80, 171, 33, 0, 80, 176, 33, 0, 80, 181, 33, 0, 80, 186, 33, 0, 80, 191, 33, 0, 80, 196, 33, 0, 80, 201, 33, 0, 80, 206, 33, 0, 80, 211, 33, 0, 80, 216, 33, 0, 80, 221, 33, 0, 80, 226, 33, 0, 80, 231, 33, 0, 80, 236, 33, 0, 80, 241, 33, 0, 80, 246, 33, 0, 80, 251, 33, 0, 80, 0, 34, 0, 80, 5, 34, 0, 80, 10, 34, 0, 80, 15, 34, 0, 80, 20, 34, 0, 80, 25, 34, 0, 80, 30, 34, 0, 80, 35, 34, 0, 80, 40, 34, 0, 80, 45, 34, 0, 80, 50, 34, 0, 80, 55, 34, 0, 80, 60, 34, 0, 80, 65, 34, 0, 80, 70, 34, 0, 80, 75, 34, 0, 80, 80, 34, 0, 80, 85, 34, 0, 80, 90, 34, 0, 80, 95, 34, 0, 80, 100, 34, 0, 80, 105, 34, 0, 80, 110, 34, 0, 80, 115, 34, 0, 80, 120, 34, 0, 80, 125, 34, 0, 80, 130, 34, 0, 80, 135, 34, 0, 80, 140, 34, 0, 80, 145, 34, 0, 80, 150, 34, 0, 80, 155, 34, 0, 80, 160, 34, 0, 80, 165, 34, 0, 80, 170, 34, 0, 80, 175, 34, 0, 80, 180, 34, 0, 80, 185, 34, 0, 80, 190, 34, 0, 80, 195, 34, 0, 80, 200, 34, 0, 80, 205, 34, 0, 80, 210, 34, 0, 80, 215, 34, 0, 80, 220, 34, 0, 80, 225, 34, 0, 80, 230, 34, 0, 80, 235, 34, 0, 80, 240, 34, 0, 80, 245, 34, 0, 80, 250, 34, 0, 80, 255, 34, 0, 80, 4, 35, 0, 80, 9, 35, 0, 80, 14, 35, 0, 80, 19, 35, 0, 80, 24, 35, 0, 80, 29, 35, 0, 80, 34, 35, 0, 80, 39, 35, 0, 80, 44, 35, 0, 80, 49, 35, 0, 80, 54, 35, 0, 80, 59, 35, 0, 80, 64, 35, 0, 80, 69, 35, 0, 80, 74, 35, 0, 80, 79, 35, 0, 80, 84, 35, 0, 80, 89, 35, 0, 80, 94, 35, 0, 80, 99, 35, 0, 80, 104, 35, 0, 80, 109, 35, 0, 80, 114, 35, 0, 80, 119, 35, 0, 80, 124, 35, 0, 80, 129, 35, 0, 80, 134, 35, 0, 80, 139, 35, 0, 80, 144, 35, 0, 80, 149, 35, 0, 80, 154, 35, 0, 80, 159, 35, 0, 80, 164, 35, 0, 80, 169, 35, 0, 80, 174, 35, 0, 80, 179, 35, 0, 80, 184, 35, 0, 80, 189, 35, 0, 80, 194, 35, 0, 80, 199, 35, 0, 80, 204, 35, 0, 80, 209, 35, 0, 80, 214, 35, 0, 80, 219, 35, 0, 80, 224, 35, 0, 80, 229, 35, 0, 80, 234, 35, 0, 80, 239, 35, 0, 80, 244, 35, 0, 80, 249, 35, 0, 80, 254, 35, 0, 80, 3, 36, 0, 80, 8, 36, 0, 80, 13, 36, 0, 80, 18, 36, 0, 80, 23, 36, 0, 80, 28, 36, 0, 80, 33, 36, 0, 80, 38, 36, 0, 80, 43, 36, 0, 80, 48, 36, 0, 80, 53, 36, 0, 80, 58, 36, 0, 80, 63, 36, 0, 80, 68, 36, 0, 80, 73, 36, 0, 80, 78, 36, 0, 80, 83, 36, 0, 80, 88, 36, 0, 80, 93, 36, 0, 80, 98, 36, 0, 80, 103, 36, 0, 80, 108, 36, 0, 80, 113, 36, 0, 80, 118, 36, 0, 80, 123, 36, 0, 80, 128, 36, 0, 80, 133, 36, 0, 80, 138, 36, 0, 80, 143, 36, 0, 80, 148, 36, 0, 80, 153, 36, 0, 80, 158, 36, 0, 80, 163, 36, 0, 80, 168, 36, 0, 80, 173, 36, 0, 80, 178, 36, 0, 80, 183, 36, 0, 80, 188, 36, 0, 80, 193, 36, 0, 80, 198, 36, 0, 80, 203, 36, 0, 80, 208, 36, 0, 80, 213, 36, 0, 80, 218, 36, 0, 80, 223, 36, 0, 80, 228, 36, 0, 80, 233, 36, 0, 80, 238, 36, 0, 80, 243, 36, 0, 80, 248, 36, 0, 80, 253, 36, 0, 80, 2, 37, 0, 80, 7, 37, 0, 80, 12, 37, 0, 80, 17, 37, 0, 80, 22, 37, 0, 80, 27, 37, 0, 80, 32, 37, 0, 80, 37, 37, 0, 80, 42, 37, 0, 80, 47, 37, 0, 80, 52, 37, 0, 80, 57, 37, 0, 80, 62, 37, 0, 80, 67, 37, 0, 80, 72, 37, 0, 80, 77, 37, 0, 80, 82, 37, 0, 80, 87, 37, 0, 80, 92, 37, 0, 80, 97, 37, 0, 80, 102, 37, 0, 80, 107, 37, 0, 80, 112, 37, 0, 80, 117, 37, 0, 80, 122, 37, 0, 80, 127, 37, 0, 80, 132, 37, 0, 80, 137, 37, 0, 80, 142, 37, 0, 80, 147, 37, 0, 80, 152, 37, 0, 80, 157, 37, 0, 80, 162, 37, 0, 80, 167, 37, 0, 80, 172, 37, 0, 80, 177, 37, 0, 80, 182, 37, 0, 80, 187, 37, 0, 80, 192, 37, 0, 80, 197, 37, 0, 80, 202, 37, 0, 80, 207, 37, 0, 80, 212, 37, 0, 80, 217, 37, 0, 80, 222, 37, 0, 80, 227, 37, 0, 80, 232, 37, 0, 80, 237, 37, 0, 80, 242, 37, 0, 80, 247, 37, 0, 80, 252, 37, 0, 80, 1, 38, 0, 80, 6, 38, 0, 80, 11, 38, 0, 80, 16, 38, 0, 80, 21, 38, 0, 80, 26, 38, 0, 80, 31, 38, 0, 80, 36, 38, 0, 80, 41, 38, 0, 80, 46, 38, 0, 80, 51, 38, 0, 80, 56, 38, 0, 80, 61, 38, 0, 80, 66, 38, 0, 80, 71, 38, 0, 80, 76, 38, 0, 80, 81, 38, 0, 80, 86, 38, 0, 80, 91, 38, 0, 80, 96, 38, 0, 80, 101, 38, 0, 80, 106, 38, 0, 80, 111, 38, 0, 80, 116, 38, 0, 80, 121, 38, 0, 80, 126, 38, 0, 80, 131, 38, 0, 80, 136, 38, 0, 80, 141, 38, 0, 80, 146, 38, 0, 80, 151, 38, 0, 80, 156, 38, 0, 80, 161, 38, 0, 80, 166, 38, 0, 80, 171, 38, 0, 80, 176, 38, 0, 80, 181, 38, 0, 80, 186, 38, 0, 80, 191, 38, 0, 80, 196, 38, 0, 80, 201, 38, 0, 80, 206, 38, 0, 80, 211, 38, 0, 80, 216, 38, 0, 80, 221, 38, 0, 80, 226, 38, 0, 80, 231, 38, 0, 80, 236, 38, 0, 80, 241, 38, 0, 80, 246, 38, 0, 80, 251, 38, 0, 80, 0, 39, 0, 80, 5, 39, 0, 80, 10, 39, 0, 80, 15, 39, 0, 80, 20, 39, 0, 80, 25, 39, 0, 80, 30, 39, 0, 80, 35, 39, 0, 80, 40, 39, 0, 80, 45, 39, 0, 80, 50, 39, 0, 80, 55, 39, 0, 80, 60, 39, 0, 80, 65, 39, 0, 80, 70, 39, 0, 80, 75, 39, 0, 80, 80, 39, 0, 80, 85, 39, 0, 80, 90, 39, 0, 80, 95, 39, 0, 80, 100, 39, 0, 80, 105, 39, 0, 80, 110, 39, 0, 80, 115, 39, 0, 80, 120, 39, 0, 80, 125, 39, 0, 80, 130, 39, 0, 80, 135, 39, 0, 80, 140, 39, 0, 80, 145, 39, 0, 80, 150, 39, 0, 80, 155, 39, 0, 80, 160, 39, 0, 80, 165, 39, 0, 80, 170, 39, 0, 80, 175, 39, 0, 80, 180, 39, 0, 80, 185, 39, 0, 80, 190, 39, 0, 80, 195, 39, 0, 80, 200, 39, 0, 80, 205, 39, 0, 80, 210, 39, 0, 80, 215, 39, 0, 80, 220, 39, 0, 80, 225, 39, 0, 80, 230, 39, 0, 80, 235, 39, 0, 80, 240, 39, 0, 80, 245, 39, 0, 80, 250, 39, 0, 80, 255, 39, 0, 80, 4, 40, 0, 80, 9, 40, 0, 80, 14, 40, 0, 80, 19, 40, 0, 80, 24, 40, 0, 80, 29, 40, 0, 80, 34, 40, 0, 80, 39, 40, 0, 80, 44, 40, 0, 80, 49, 40, 0, 80, 54, 40, 0, 80, 59, 40, 0, 80, 64, 40, 0, 80, 69, 40, 0, 80, 74, 40, 0, 80, 79, 40, 0, 80, 84, 40, 0, 80, 89, 40, 0, 80, 94, 40, 0, 80, 99, 40, 0, 80, 104, 40, 0, 80, 109, 40, 0, 80, 114, 40, 0, 80, 119, 40, 0, 80, 124, 40, 0, 80, 129, 40, 0, 80, 134, 40, 0, 80, 139, 40, 0, 80, 144, 40, 0, 80, 149, 40, 0, 80, 154, 40, 0, 80, 159, 40, 0, 80, 164, 40, 0, 80, 169, 40, 0, 80, 174, 40, 0, 80, 179, 40, 0, 80, 184, 40, 0, 80, 189, 40, 0, 80, 194, 40, 0, 80, 199, 40, 0, 80, 204, 40, 0, 80, 209, 40, 0, 80, 214, 40, 0, 80, 219, 40, 0, 80, 224, 40, 0, 80, 229, 40, 0, 80, 234, 40, 0, 80, 239, 40, 0, 80, 244, 40, 0, 80, 249, 40, 0, 80, 254, 40, 0, 80, 3, 41, 0, 80, 8, 41, 0, 80, 13, 41, 0, 80, 18, 41, 0, 80, 23, 41, 0, 80, 28, 41, 0, 80, 33, 41, 0, 80, 38, 41, 0, 80, 43, 41, 0, 80, 48, 41, 0, 80, 53, 41, 0, 80, 58, 41, 0, 80, 63, 41, 0, 80, 68, 41, 0, 80, 73, 41, 0, 80, 78, 41, 0, 80, 83, 41, 0, 80, 88, 41, 0, 80, 6, 0, 99, 20, 106, 20, 113, 20, 122, 20, 130, 20, 138, 20, 170, 170, 93, 41, 0, 80, 98, 41, 0, 80, 103, 41, 0, 80, 108, 41, 0, 80, 113, 41, 0, 80, 118, 41, 0, 80, 5, 0, 38, 0, 178, 7, 190, 7, 145, 10, 91, 20, 121, 26, 0, 32, 240, 28, 0, 32, 242, 28, 0, 32, 204, 29, 0, 32, 141, 33, 0, 32, 2, 0, 32, 0, 145, 20, 170, 170, 151, 33, 0, 32, 123, 41, 0, 144, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 162, 5, 0, 32, 7, 0, 0, 0, 143, 0, 0, 0, 164, 5, 0, 0, 164, 5, 0, 0, 85, 0, 0, 0, 1, 0, 0, 0, 121, 3, 0, 0, 110, 117, 109, 98, 101, 114, 105, 110, 103, 83, 121, 115, 116, 101, 109, 115, 0, 97, 100, 108, 109, 0, 97, 108, 103, 111, 114, 105, 116, 104, 109, 105, 99, 0, 100, 101, 115, 99, 0, 114, 97, 100, 105, 120, 0, 97, 104, 111, 109, 0, 97, 114, 97, 98, 0, 97, 114, 97, 98, 101, 120, 116, 0, 97, 114, 109, 110, 0, 97, 114, 109, 110, 108, 111, 119, 0, 98, 97, 108, 105, 0, 98, 101, 110, 103, 0, 98, 104, 107, 115, 0, 98, 114, 97, 104, 0, 99, 97, 107, 109, 0, 99, 104, 97, 109, 0, 99, 121, 114, 108, 0, 100, 101, 118, 97, 0, 100, 105, 97, 107, 0, 101, 116, 104, 105, 0, 102, 117, 108, 108, 119, 105, 100, 101, 0, 103, 101, 111, 114, 0, 103, 111, 110, 103, 0, 103, 111, 110, 109, 0, 103, 114, 101, 107, 0, 103, 114, 101, 107, 108, 111, 119, 0, 103, 117, 106, 114, 0, 103, 117, 114, 117, 0, 104, 97, 110, 105, 100, 97, 121, 115, 0, 104, 97, 110, 105, 100, 101, 99, 0, 104, 97, 110, 115, 0, 104, 97, 110, 115, 102, 105, 110, 0, 104, 97, 110, 116, 0, 104, 97, 110, 116, 102, 105, 110, 0, 104, 101, 98, 114, 0, 104, 109, 110, 103, 0, 104, 109, 110, 112, 0, 106, 97, 118, 97, 0, 106, 112, 97, 110, 0, 106, 112, 97, 110, 102, 105, 110, 0, 106, 112, 97, 110, 121, 101, 97, 114, 0, 107, 97, 108, 105, 0, 107, 104, 109, 114, 0, 107, 110, 100, 97, 0, 108, 97, 110, 97, 0, 108, 97, 110, 97, 116, 104, 97, 109, 0, 108, 97, 111, 111, 0, 108, 97, 116, 110, 0, 108, 101, 112, 99, 0, 108, 105, 109, 98, 0, 109, 97, 116, 104, 98, 111, 108, 100, 0, 109, 97, 116, 104, 100, 98, 108, 0, 109, 97, 116, 104, 109, 111, 110, 111, 0, 109, 97, 116, 104, 115, 97, 110, 98, 0, 109, 97, 116, 104, 115, 97, 110, 115, 0, 109, 108, 121, 109, 0, 109, 111, 100, 105, 0, 109, 111, 110, 103, 0, 109, 114, 111, 111, 0, 109, 116, 101, 105, 0, 109, 121, 109, 114, 0, 109, 121, 109, 114, 115, 104, 97, 110, 0, 109, 121, 109, 114, 116, 108, 110, 103, 0, 110, 101, 119, 97, 0, 110, 107, 111, 111, 0, 111, 108, 99, 107, 0, 111, 114, 121, 97, 0, 111, 115, 109, 97, 0, 114, 111, 104, 103, 0, 114, 111, 109, 97, 110, 0, 114, 111, 109, 97, 110, 108, 111, 119, 0, 115, 97, 117, 114, 0, 115, 101, 103, 109, 101, 110, 116, 0, 115, 104, 114, 100, 0, 115, 105, 110, 100, 0, 115, 105, 110, 104, 0, 115, 111, 114, 97, 0, 115, 117, 110, 100, 0, 116, 97, 107, 114, 0, 116, 97, 108, 117, 0, 116, 97, 109, 108, 0, 116, 97, 109, 108, 100, 101, 99, 0, 116, 101, 108, 117, 0, 116, 104, 97, 105, 0, 116, 105, 98, 116, 0, 116, 105, 114, 104, 0, 118, 97, 105, 105, 0, 119, 97, 114, 97, 0, 119, 99, 104, 111, 0, 0, 0, 37, 0, 116, 0, 97, 0, 109, 0, 105, 0, 108, 0, 0, 0, 37, 0, 104, 0, 101, 0, 98, 0, 114, 0, 101, 0, 119, 0, 0, 0, 37, 0, 101, 0, 116, 0, 104, 0, 105, 0, 111, 0, 112, 0, 105, 0, 99, 0, 0, 0, 37, 0, 103, 0, 101, 0, 111, 0, 114, 0, 103, 0, 105, 0, 97, 0, 110, 0, 0, 0, 48, 0, 49, 0, 50, 0, 51, 0, 52, 0, 53, 0, 54, 0, 55, 0, 56, 0, 57, 0, 0, 0, 96, 6, 97, 6, 98, 6, 99, 6, 100, 6, 101, 6, 102, 6, 103, 6, 104, 6, 105, 6, 0, 0, 240, 6, 241, 6, 242, 6, 243, 6, 244, 6, 245, 6, 246, 6, 247, 6, 248, 6, 249, 6, 0, 0, 192, 7, 193, 7, 194, 7, 195, 7, 196, 7, 197, 7, 198, 7, 199, 7, 200, 7, 201, 7, 0, 0, 102, 9, 103, 9, 104, 9, 105, 9, 106, 9, 107, 9, 108, 9, 109, 9, 110, 9, 111, 9, 0, 0, 230, 9, 231, 9, 232, 9, 233, 9, 234, 9, 235, 9, 236, 9, 237, 9, 238, 9, 239, 9, 0, 0, 102, 10, 103, 10, 104, 10, 105, 10, 106, 10, 107, 10, 108, 10, 109, 10, 110, 10, 111, 10, 0, 0, 230, 10, 231, 10, 232, 10, 233, 10, 234, 10, 235, 10, 236, 10, 237, 10, 238, 10, 239, 10, 0, 0, 102, 11, 103, 11, 104, 11, 105, 11, 106, 11, 107, 11, 108, 11, 109, 11, 110, 11, 111, 11, 0, 0, 230, 11, 231, 11, 232, 11, 233, 11, 234, 11, 235, 11, 236, 11, 237, 11, 238, 11, 239, 11, 0, 0, 102, 12, 103, 12, 104, 12, 105, 12, 106, 12, 107, 12, 108, 12, 109, 12, 110, 12, 111, 12, 0, 0, 230, 12, 231, 12, 232, 12, 233, 12, 234, 12, 235, 12, 236, 12, 237, 12, 238, 12, 239, 12, 0, 0, 102, 13, 103, 13, 104, 13, 105, 13, 106, 13, 107, 13, 108, 13, 109, 13, 110, 13, 111, 13, 0, 0, 230, 13, 231, 13, 232, 13, 233, 13, 234, 13, 235, 13, 236, 13, 237, 13, 238, 13, 239, 13, 0, 0, 80, 14, 81, 14, 82, 14, 83, 14, 84, 14, 85, 14, 86, 14, 87, 14, 88, 14, 89, 14, 0, 0, 208, 14, 209, 14, 210, 14, 211, 14, 212, 14, 213, 14, 214, 14, 215, 14, 216, 14, 217, 14, 0, 0, 32, 15, 33, 15, 34, 15, 35, 15, 36, 15, 37, 15, 38, 15, 39, 15, 40, 15, 41, 15, 0, 0, 64, 16, 65, 16, 66, 16, 67, 16, 68, 16, 69, 16, 70, 16, 71, 16, 72, 16, 73, 16, 0, 0, 144, 16, 145, 16, 146, 16, 147, 16, 148, 16, 149, 16, 150, 16, 151, 16, 152, 16, 153, 16, 0, 0, 224, 23, 225, 23, 226, 23, 227, 23, 228, 23, 229, 23, 230, 23, 231, 23, 232, 23, 233, 23, 0, 0, 16, 24, 17, 24, 18, 24, 19, 24, 20, 24, 21, 24, 22, 24, 23, 24, 24, 24, 25, 24, 0, 0, 70, 25, 71, 25, 72, 25, 73, 25, 74, 25, 75, 25, 76, 25, 77, 25, 78, 25, 79, 25, 0, 0, 208, 25, 209, 25, 210, 25, 211, 25, 212, 25, 213, 25, 214, 25, 215, 25, 216, 25, 217, 25, 0, 0, 128, 26, 129, 26, 130, 26, 131, 26, 132, 26, 133, 26, 134, 26, 135, 26, 136, 26, 137, 26, 0, 0, 144, 26, 145, 26, 146, 26, 147, 26, 148, 26, 149, 26, 150, 26, 151, 26, 152, 26, 153, 26, 0, 0, 80, 27, 81, 27, 82, 27, 83, 27, 84, 27, 85, 27, 86, 27, 87, 27, 88, 27, 89, 27, 0, 0, 176, 27, 177, 27, 178, 27, 179, 27, 180, 27, 181, 27, 182, 27, 183, 27, 184, 27, 185, 27, 0, 0, 64, 28, 65, 28, 66, 28, 67, 28, 68, 28, 69, 28, 70, 28, 71, 28, 72, 28, 73, 28, 0, 0, 80, 28, 81, 28, 82, 28, 83, 28, 84, 28, 85, 28, 86, 28, 87, 28, 88, 28, 89, 28, 0, 0, 7, 48, 0, 78, 140, 78, 9, 78, 219, 86, 148, 78, 109, 81, 3, 78, 107, 81, 93, 78, 0, 0, 32, 166, 33, 166, 34, 166, 35, 166, 36, 166, 37, 166, 38, 166, 39, 166, 40, 166, 41, 166, 0, 0, 208, 168, 209, 168, 210, 168, 211, 168, 212, 168, 213, 168, 214, 168, 215, 168, 216, 168, 217, 168, 0, 0, 0, 169, 1, 169, 2, 169, 3, 169, 4, 169, 5, 169, 6, 169, 7, 169, 8, 169, 9, 169, 0, 0, 208, 169, 209, 169, 210, 169, 211, 169, 212, 169, 213, 169, 214, 169, 215, 169, 216, 169, 217, 169, 0, 0, 240, 169, 241, 169, 242, 169, 243, 169, 244, 169, 245, 169, 246, 169, 247, 169, 248, 169, 249, 169, 0, 0, 80, 170, 81, 170, 82, 170, 83, 170, 84, 170, 85, 170, 86, 170, 87, 170, 88, 170, 89, 170, 0, 0, 240, 171, 241, 171, 242, 171, 243, 171, 244, 171, 245, 171, 246, 171, 247, 171, 248, 171, 249, 171, 0, 0, 16, 255, 17, 255, 18, 255, 19, 255, 20, 255, 21, 255, 22, 255, 23, 255, 24, 255, 25, 255, 0, 0, 37, 0, 103, 0, 114, 0, 101, 0, 101, 0, 107, 0, 45, 0, 108, 0, 111, 0, 119, 0, 101, 0, 114, 0, 0, 0, 37, 0, 103, 0, 114, 0, 101, 0, 101, 0, 107, 0, 45, 0, 117, 0, 112, 0, 112, 0, 101, 0, 114, 0, 0, 0, 37, 0, 114, 0, 111, 0, 109, 0, 97, 0, 110, 0, 45, 0, 108, 0, 111, 0, 119, 0, 101, 0, 114, 0, 0, 0, 37, 0, 114, 0, 111, 0, 109, 0, 97, 0, 110, 0, 45, 0, 117, 0, 112, 0, 112, 0, 101, 0, 114, 0, 0, 0, 37, 0, 97, 0, 114, 0, 109, 0, 101, 0, 110, 0, 105, 0, 97, 0, 110, 0, 45, 0, 108, 0, 111, 0, 119, 0, 101, 0, 114, 0, 0, 0, 37, 0, 97, 0, 114, 0, 109, 0, 101, 0, 110, 0, 105, 0, 97, 0, 110, 0, 45, 0, 117, 0, 112, 0, 112, 0, 101, 0, 114, 0, 0, 0, 37, 0, 99, 0, 121, 0, 114, 0, 105, 0, 108, 0, 108, 0, 105, 0, 99, 0, 45, 0, 108, 0, 111, 0, 119, 0, 101, 0, 114, 0, 0, 0, 1, 216, 160, 220, 1, 216, 161, 220, 1, 216, 162, 220, 1, 216, 163, 220, 1, 216, 164, 220, 1, 216, 165, 220, 1, 216, 166, 220, 1, 216, 167, 220, 1, 216, 168, 220, 1, 216, 169, 220, 0, 0, 3, 216, 48, 221, 3, 216, 49, 221, 3, 216, 50, 221, 3, 216, 51, 221, 3, 216, 52, 221, 3, 216, 53, 221, 3, 216, 54, 221, 3, 216, 55, 221, 3, 216, 56, 221, 3, 216, 57, 221, 0, 0, 4, 216, 102, 220, 4, 216, 103, 220, 4, 216, 104, 220, 4, 216, 105, 220, 4, 216, 106, 220, 4, 216, 107, 220, 4, 216, 108, 220, 4, 216, 109, 220, 4, 216, 110, 220, 4, 216, 111, 220, 0, 0, 4, 216, 240, 220, 4, 216, 241, 220, 4, 216, 242, 220, 4, 216, 243, 220, 4, 216, 244, 220, 4, 216, 245, 220, 4, 216, 246, 220, 4, 216, 247, 220, 4, 216, 248, 220, 4, 216, 249, 220, 0, 0, 4, 216, 54, 221, 4, 216, 55, 221, 4, 216, 56, 221, 4, 216, 57, 221, 4, 216, 58, 221, 4, 216, 59, 221, 4, 216, 60, 221, 4, 216, 61, 221, 4, 216, 62, 221, 4, 216, 63, 221, 0, 0, 4, 216, 208, 221, 4, 216, 209, 221, 4, 216, 210, 221, 4, 216, 211, 221, 4, 216, 212, 221, 4, 216, 213, 221, 4, 216, 214, 221, 4, 216, 215, 221, 4, 216, 216, 221, 4, 216, 217, 221, 0, 0, 4, 216, 240, 222, 4, 216, 241, 222, 4, 216, 242, 222, 4, 216, 243, 222, 4, 216, 244, 222, 4, 216, 245, 222, 4, 216, 246, 222, 4, 216, 247, 222, 4, 216, 248, 222, 4, 216, 249, 222, 0, 0, 5, 216, 80, 220, 5, 216, 81, 220, 5, 216, 82, 220, 5, 216, 83, 220, 5, 216, 84, 220, 5, 216, 85, 220, 5, 216, 86, 220, 5, 216, 87, 220, 5, 216, 88, 220, 5, 216, 89, 220, 0, 0, 5, 216, 208, 220, 5, 216, 209, 220, 5, 216, 210, 220, 5, 216, 211, 220, 5, 216, 212, 220, 5, 216, 213, 220, 5, 216, 214, 220, 5, 216, 215, 220, 5, 216, 216, 220, 5, 216, 217, 220, 0, 0, 5, 216, 80, 222, 5, 216, 81, 222, 5, 216, 82, 222, 5, 216, 83, 222, 5, 216, 84, 222, 5, 216, 85, 222, 5, 216, 86, 222, 5, 216, 87, 222, 5, 216, 88, 222, 5, 216, 89, 222, 0, 0, 5, 216, 192, 222, 5, 216, 193, 222, 5, 216, 194, 222, 5, 216, 195, 222, 5, 216, 196, 222, 5, 216, 197, 222, 5, 216, 198, 222, 5, 216, 199, 222, 5, 216, 200, 222, 5, 216, 201, 222, 0, 0, 5, 216, 48, 223, 5, 216, 49, 223, 5, 216, 50, 223, 5, 216, 51, 223, 5, 216, 52, 223, 5, 216, 53, 223, 5, 216, 54, 223, 5, 216, 55, 223, 5, 216, 56, 223, 5, 216, 57, 223, 0, 0, 6, 216, 224, 220, 6, 216, 225, 220, 6, 216, 226, 220, 6, 216, 227, 220, 6, 216, 228, 220, 6, 216, 229, 220, 6, 216, 230, 220, 6, 216, 231, 220, 6, 216, 232, 220, 6, 216, 233, 220, 0, 0, 6, 216, 80, 221, 6, 216, 81, 221, 6, 216, 82, 221, 6, 216, 83, 221, 6, 216, 84, 221, 6, 216, 85, 221, 6, 216, 86, 221, 6, 216, 87, 221, 6, 216, 88, 221, 6, 216, 89, 221, 0, 0, 7, 216, 80, 220, 7, 216, 81, 220, 7, 216, 82, 220, 7, 216, 83, 220, 7, 216, 84, 220, 7, 216, 85, 220, 7, 216, 86, 220, 7, 216, 87, 220, 7, 216, 88, 220, 7, 216, 89, 220, 0, 0, 7, 216, 80, 221, 7, 216, 81, 221, 7, 216, 82, 221, 7, 216, 83, 221, 7, 216, 84, 221, 7, 216, 85, 221, 7, 216, 86, 221, 7, 216, 87, 221, 7, 216, 88, 221, 7, 216, 89, 221, 0, 0, 7, 216, 160, 221, 7, 216, 161, 221, 7, 216, 162, 221, 7, 216, 163, 221, 7, 216, 164, 221, 7, 216, 165, 221, 7, 216, 166, 221, 7, 216, 167, 221, 7, 216, 168, 221, 7, 216, 169, 221, 0, 0, 26, 216, 96, 222, 26, 216, 97, 222, 26, 216, 98, 222, 26, 216, 99, 222, 26, 216, 100, 222, 26, 216, 101, 222, 26, 216, 102, 222, 26, 216, 103, 222, 26, 216, 104, 222, 26, 216, 105, 222, 0, 0, 26, 216, 80, 223, 26, 216, 81, 223, 26, 216, 82, 223, 26, 216, 83, 223, 26, 216, 84, 223, 26, 216, 85, 223, 26, 216, 86, 223, 26, 216, 87, 223, 26, 216, 88, 223, 26, 216, 89, 223, 0, 0, 53, 216, 206, 223, 53, 216, 207, 223, 53, 216, 208, 223, 53, 216, 209, 223, 53, 216, 210, 223, 53, 216, 211, 223, 53, 216, 212, 223, 53, 216, 213, 223, 53, 216, 214, 223, 53, 216, 215, 223, 0, 0, 53, 216, 216, 223, 53, 216, 217, 223, 53, 216, 218, 223, 53, 216, 219, 223, 53, 216, 220, 223, 53, 216, 221, 223, 53, 216, 222, 223, 53, 216, 223, 223, 53, 216, 224, 223, 53, 216, 225, 223, 0, 0, 53, 216, 226, 223, 53, 216, 227, 223, 53, 216, 228, 223, 53, 216, 229, 223, 53, 216, 230, 223, 53, 216, 231, 223, 53, 216, 232, 223, 53, 216, 233, 223, 53, 216, 234, 223, 53, 216, 235, 223, 0, 0, 53, 216, 236, 223, 53, 216, 237, 223, 53, 216, 238, 223, 53, 216, 239, 223, 53, 216, 240, 223, 53, 216, 241, 223, 53, 216, 242, 223, 53, 216, 243, 223, 53, 216, 244, 223, 53, 216, 245, 223, 0, 0, 53, 216, 246, 223, 53, 216, 247, 223, 53, 216, 248, 223, 53, 216, 249, 223, 53, 216, 250, 223, 53, 216, 251, 223, 53, 216, 252, 223, 53, 216, 253, 223, 53, 216, 254, 223, 53, 216, 255, 223, 0, 0, 56, 216, 64, 221, 56, 216, 65, 221, 56, 216, 66, 221, 56, 216, 67, 221, 56, 216, 68, 221, 56, 216, 69, 221, 56, 216, 70, 221, 56, 216, 71, 221, 56, 216, 72, 221, 56, 216, 73, 221, 0, 0, 56, 216, 240, 222, 56, 216, 241, 222, 56, 216, 242, 222, 56, 216, 243, 222, 56, 216, 244, 222, 56, 216, 245, 222, 56, 216, 246, 222, 56, 216, 247, 222, 56, 216, 248, 222, 56, 216, 249, 222, 0, 0, 58, 216, 80, 221, 58, 216, 81, 221, 58, 216, 82, 221, 58, 216, 83, 221, 58, 216, 84, 221, 58, 216, 85, 221, 58, 216, 86, 221, 58, 216, 87, 221, 58, 216, 88, 221, 58, 216, 89, 221, 0, 0, 62, 216, 240, 223, 62, 216, 241, 223, 62, 216, 242, 223, 62, 216, 243, 223, 62, 216, 244, 223, 62, 216, 245, 223, 62, 216, 246, 223, 62, 216, 247, 223, 62, 216, 248, 223, 62, 216, 249, 223, 0, 0, 106, 0, 97, 0, 47, 0, 83, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 82, 0, 117, 0, 108, 0, 101, 0, 115, 0, 47, 0, 37, 0, 115, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 45, 0, 99, 0, 97, 0, 114, 0, 100, 0, 105, 0, 110, 0, 97, 0, 108, 0, 0, 0, 122, 0, 104, 0, 47, 0, 83, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 82, 0, 117, 0, 108, 0, 101, 0, 115, 0, 47, 0, 37, 0, 115, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 45, 0, 99, 0, 97, 0, 114, 0, 100, 0, 105, 0, 110, 0, 97, 0, 108, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 47, 0, 83, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 82, 0, 117, 0, 108, 0, 101, 0, 115, 0, 47, 0, 37, 0, 115, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 45, 0, 99, 0, 97, 0, 114, 0, 100, 0, 105, 0, 110, 0, 97, 0, 108, 0, 0, 0, 41, 220, 122, 0, 104, 0, 47, 0, 83, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 82, 0, 117, 0, 108, 0, 101, 0, 115, 0, 47, 0, 37, 0, 115, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 45, 0, 110, 0, 117, 0, 109, 0, 98, 0, 101, 0, 114, 0, 105, 0, 110, 0, 103, 0, 45, 0, 100, 0, 97, 0, 121, 0, 115, 0, 0, 0, 45, 220, 106, 0, 97, 0, 47, 0, 83, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 82, 0, 117, 0, 108, 0, 101, 0, 115, 0, 47, 0, 37, 0, 115, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 45, 0, 99, 0, 97, 0, 114, 0, 100, 0, 105, 0, 110, 0, 97, 0, 108, 0, 45, 0, 102, 0, 105, 0, 110, 0, 97, 0, 110, 0, 99, 0, 105, 0, 97, 0, 108, 0, 0, 0, 45, 220, 122, 0, 104, 0, 47, 0, 83, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 82, 0, 117, 0, 108, 0, 101, 0, 115, 0, 47, 0, 37, 0, 115, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 45, 0, 99, 0, 97, 0, 114, 0, 100, 0, 105, 0, 110, 0, 97, 0, 108, 0, 45, 0, 102, 0, 105, 0, 110, 0, 97, 0, 110, 0, 99, 0, 105, 0, 97, 0, 108, 0, 0, 0, 46, 220, 106, 0, 97, 0, 47, 0, 83, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 82, 0, 117, 0, 108, 0, 101, 0, 115, 0, 47, 0, 37, 0, 115, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 45, 0, 110, 0, 117, 0, 109, 0, 98, 0, 101, 0, 114, 0, 105, 0, 110, 0, 103, 0, 45, 0, 121, 0, 101, 0, 97, 0, 114, 0, 45, 0, 108, 0, 97, 0, 116, 0, 110, 0, 0, 0, 50, 220, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 47, 0, 83, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 82, 0, 117, 0, 108, 0, 101, 0, 115, 0, 47, 0, 37, 0, 115, 0, 112, 0, 101, 0, 108, 0, 108, 0, 111, 0, 117, 0, 116, 0, 45, 0, 99, 0, 97, 0, 114, 0, 100, 0, 105, 0, 110, 0, 97, 0, 108, 0, 45, 0, 102, 0, 105, 0, 110, 0, 97, 0, 110, 0, 99, 0, 105, 0, 97, 0, 108, 0, 0, 0, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 76, 4, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 17, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 47, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 58, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 10, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 250, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 55, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 91, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 80, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 84, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 126, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 165, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 26, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 80, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 59, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 16, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 187, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 26, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 122, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 101, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 211, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 198, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 113, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 102, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 231, 4, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 99, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 154, 4, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 65, 5, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 190, 4, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 160, 5, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 8, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 164, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 34, 4, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 143, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 118, 4, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 18, 5, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 112, 5, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 132, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 245, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 157, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 33, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 44, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 201, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 36, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 77, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 11, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 185, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 206, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 13, 4, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 248, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 227, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 168, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 231, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 0, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 143, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 176, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 223, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 234, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 154, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 189, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 69, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 88, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 124, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 42, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 63, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 237, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 224, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 121, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 97, 4, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 147, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 168, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 179, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 105, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 66, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 252, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 22, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 1, 0, 0, 112, 1, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 135, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 146, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 190, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 212, 0, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 210, 2, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 110, 1, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 38, 3, 0, 96, 10, 0, 0, 112, 3, 0, 54, 0, 66, 0, 71, 0, 0, 0, 0, 112, 55, 4, 0, 96, 10, 0, 0, 112, 85, 0, 49, 0, 77, 0, 82, 0, 87, 0, 95, 0, 100, 0, 108, 0, 113, 0, 118, 0, 123, 0, 128, 0, 133, 0, 138, 0, 143, 0, 148, 0, 153, 0, 158, 0, 167, 0, 172, 0, 177, 0, 182, 0, 187, 0, 195, 0, 200, 0, 205, 0, 214, 0, 222, 0, 227, 0, 235, 0, 240, 0, 248, 0, 253, 0, 2, 1, 7, 1, 12, 1, 17, 1, 25, 1, 34, 1, 39, 1, 44, 1, 49, 1, 54, 1, 63, 1, 68, 1, 73, 1, 78, 1, 83, 1, 92, 1, 100, 1, 109, 1, 118, 1, 127, 1, 132, 1, 137, 1, 142, 1, 147, 1, 152, 1, 157, 1, 166, 1, 175, 1, 180, 1, 185, 1, 190, 1, 195, 1, 200, 1, 205, 1, 211, 1, 220, 1, 225, 1, 233, 1, 238, 1, 243, 1, 248, 1, 253, 1, 2, 2, 7, 2, 12, 2, 17, 2, 25, 2, 30, 2, 35, 2, 40, 2, 45, 2, 50, 2, 55, 2, 121, 3, 0, 32, 126, 3, 0, 32, 131, 3, 0, 32, 136, 3, 0, 32, 141, 3, 0, 32, 146, 3, 0, 32, 151, 3, 0, 32, 156, 3, 0, 32, 161, 3, 0, 32, 166, 3, 0, 32, 171, 3, 0, 32, 176, 3, 0, 32, 181, 3, 0, 32, 186, 3, 0, 32, 191, 3, 0, 32, 196, 3, 0, 32, 201, 3, 0, 32, 206, 3, 0, 32, 211, 3, 0, 32, 216, 3, 0, 32, 221, 3, 0, 32, 226, 3, 0, 32, 231, 3, 0, 32, 236, 3, 0, 32, 241, 3, 0, 32, 246, 3, 0, 32, 251, 3, 0, 32, 0, 4, 0, 32, 5, 4, 0, 32, 10, 4, 0, 32, 15, 4, 0, 32, 20, 4, 0, 32, 25, 4, 0, 32, 30, 4, 0, 32, 35, 4, 0, 32, 40, 4, 0, 32, 45, 4, 0, 32, 50, 4, 0, 32, 55, 4, 0, 32, 60, 4, 0, 32, 65, 4, 0, 32, 70, 4, 0, 32, 75, 4, 0, 32, 80, 4, 0, 32, 85, 4, 0, 32, 90, 4, 0, 32, 95, 4, 0, 32, 100, 4, 0, 32, 105, 4, 0, 32, 110, 4, 0, 32, 115, 4, 0, 32, 120, 4, 0, 32, 125, 4, 0, 32, 130, 4, 0, 32, 135, 4, 0, 32, 140, 4, 0, 32, 145, 4, 0, 32, 150, 4, 0, 32, 155, 4, 0, 32, 160, 4, 0, 32, 165, 4, 0, 32, 170, 4, 0, 32, 175, 4, 0, 32, 180, 4, 0, 32, 185, 4, 0, 32, 190, 4, 0, 32, 195, 4, 0, 32, 200, 4, 0, 32, 205, 4, 0, 32, 210, 4, 0, 32, 215, 4, 0, 32, 220, 4, 0, 32, 225, 4, 0, 32, 230, 4, 0, 32, 235, 4, 0, 32, 240, 4, 0, 32, 245, 4, 0, 32, 250, 4, 0, 32, 255, 4, 0, 32, 4, 5, 0, 32, 9, 5, 0, 32, 14, 5, 0, 32, 19, 5, 0, 32, 24, 5, 0, 32, 29, 5, 0, 32, 1, 0, 32, 0, 34, 5, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 59, 3, 0, 32, 7, 0, 0, 0, 106, 0, 0, 0, 63, 3, 0, 0, 63, 3, 0, 0, 88, 0, 0, 0, 1, 0, 0, 0, 96, 2, 0, 0, 108, 111, 99, 97, 108, 101, 115, 0, 97, 102, 0, 97, 107, 0, 97, 109, 0, 97, 110, 0, 97, 114, 0, 97, 115, 0, 97, 122, 0, 98, 101, 0, 98, 103, 0, 98, 110, 0, 98, 115, 0, 99, 97, 0, 99, 115, 0, 99, 121, 0, 100, 97, 0, 100, 101, 0, 101, 108, 0, 101, 110, 0, 101, 116, 0, 101, 117, 0, 102, 97, 0, 102, 105, 0, 102, 105, 108, 0, 102, 114, 0, 103, 97, 0, 103, 108, 0, 103, 115, 119, 0, 103, 117, 0, 104, 101, 0, 104, 105, 0, 104, 114, 0, 104, 117, 0, 104, 121, 0, 105, 97, 0, 105, 100, 0, 105, 111, 0, 105, 115, 0, 105, 116, 0, 106, 97, 0, 107, 97, 0, 107, 107, 0, 107, 109, 0, 107, 110, 0, 107, 111, 0, 107, 121, 0, 108, 111, 0, 108, 116, 0, 108, 118, 0, 109, 107, 0, 109, 108, 0, 109, 110, 0, 109, 114, 0, 109, 115, 0, 109, 121, 0, 110, 98, 0, 110, 101, 0, 110, 108, 0, 111, 114, 0, 112, 97, 0, 112, 108, 0, 112, 115, 0, 112, 116, 0, 114, 111, 0, 114, 117, 0, 115, 99, 0, 115, 99, 110, 0, 115, 100, 0, 115, 105, 0, 115, 107, 0, 115, 108, 0, 115, 113, 0, 115, 114, 0, 115, 118, 0, 116, 97, 0, 116, 101, 0, 116, 104, 0, 116, 107, 0, 116, 114, 0, 117, 103, 0, 117, 107, 0, 117, 114, 0, 117, 122, 0, 118, 105, 0, 121, 117, 101, 0, 122, 104, 0, 122, 117, 0, 114, 117, 108, 101, 115, 0, 115, 101, 116, 48, 48, 0, 115, 101, 116, 48, 49, 0, 115, 101, 116, 48, 50, 0, 115, 101, 116, 48, 51, 0, 115, 101, 116, 48, 52, 0, 115, 101, 116, 48, 53, 0, 115, 101, 116, 48, 54, 0, 115, 101, 116, 48, 55, 0, 115, 101, 116, 48, 56, 0, 115, 101, 116, 48, 57, 0, 115, 101, 116, 49, 48, 0, 115, 101, 116, 49, 49, 0, 115, 101, 116, 49, 50, 0, 115, 101, 116, 49, 51, 0, 115, 101, 116, 49, 52, 0, 115, 101, 116, 49, 53, 0, 115, 101, 116, 49, 54, 0, 115, 101, 116, 49, 55, 0, 115, 101, 116, 49, 56, 0, 170, 170, 0, 0, 111, 0, 110, 0, 101, 0, 0, 0, 102, 0, 101, 0, 119, 0, 0, 0, 116, 0, 119, 0, 111, 0, 0, 0, 109, 0, 97, 0, 110, 0, 121, 0, 0, 0, 122, 0, 101, 0, 114, 0, 111, 0, 0, 0, 111, 0, 116, 0, 104, 0, 101, 0, 114, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 56, 0, 0, 0, 88, 0, 40, 0, 43, 0, 46, 0, 49, 0, 52, 0, 55, 0, 58, 0, 61, 0, 64, 0, 67, 0, 70, 0, 73, 0, 76, 0, 79, 0, 82, 0, 85, 0, 88, 0, 91, 0, 49, 1, 94, 0, 97, 0, 100, 0, 103, 0, 106, 0, 110, 0, 113, 0, 116, 0, 119, 0, 123, 0, 126, 0, 129, 0, 132, 0, 135, 0, 138, 0, 141, 0, 144, 0, 147, 0, 150, 0, 153, 0, 156, 0, 159, 0, 162, 0, 165, 0, 168, 0, 171, 0, 174, 0, 177, 0, 180, 0, 183, 0, 186, 0, 189, 0, 192, 0, 195, 0, 198, 0, 201, 0, 204, 0, 207, 0, 210, 0, 213, 0, 216, 0, 219, 0, 222, 0, 225, 0, 228, 0, 231, 0, 234, 0, 237, 0, 241, 0, 244, 0, 247, 0, 250, 0, 253, 0, 0, 1, 3, 1, 120, 0, 6, 1, 9, 1, 12, 1, 15, 1, 18, 1, 21, 1, 24, 1, 27, 1, 30, 1, 33, 1, 36, 1, 40, 1, 43, 1, 35, 0, 59, 0, 41, 0, 35, 0, 137, 0, 41, 0, 29, 0, 65, 0, 35, 0, 41, 0, 71, 0, 35, 0, 77, 0, 131, 0, 53, 0, 29, 0, 29, 0, 35, 0, 35, 0, 35, 0, 35, 0, 59, 0, 35, 0, 53, 0, 41, 0, 125, 0, 29, 0, 29, 0, 41, 0, 119, 0, 41, 0, 71, 0, 29, 0, 41, 0, 35, 0, 47, 0, 35, 0, 53, 0, 29, 0, 47, 0, 83, 0, 29, 0, 47, 0, 41, 0, 47, 0, 29, 0, 47, 0, 65, 0, 101, 0, 95, 0, 29, 0, 29, 0, 41, 0, 47, 0, 47, 0, 35, 0, 29, 0, 29, 0, 59, 0, 53, 0, 77, 0, 41, 0, 41, 0, 107, 0, 65, 0, 29, 0, 29, 0, 59, 0, 89, 0, 77, 0, 113, 0, 29, 0, 71, 0, 35, 0, 29, 0, 29, 0, 29, 0, 47, 0, 29, 0, 29, 0, 29, 0, 65, 0, 35, 0, 29, 0, 47, 0, 47, 0, 47, 0, 41, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 1, 0, 1, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 23, 0, 1, 0, 3, 0, 23, 0, 1, 0, 23, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 1, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 23, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 1, 0, 1, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 1, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 1, 0, 1, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 23, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 1, 0, 23, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 1, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 1, 0, 23, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 23, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 18, 0, 18, 0, 23, 0, 3, 0, 18, 0, 1, 0, 1, 0, 3, 0, 18, 0, 23, 0, 23, 0, 3, 0, 1, 0, 18, 0, 23, 0, 3, 0, 1, 0, 1, 0, 1, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 23, 0, 18, 0, 23, 0, 3, 0, 23, 0, 1, 0, 1, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 5, 0, 5, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 5, 0, 1, 0, 5, 0, 3, 0, 5, 0, 5, 0, 5, 0, 3, 0, 5, 0, 23, 0, 23, 0, 3, 0, 23, 0, 5, 0, 5, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 1, 0, 1, 0, 3, 0, 1, 0, 5, 0, 5, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 5, 0, 1, 0, 1, 0, 3, 0, 5, 0, 5, 0, 5, 0, 3, 0, 5, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 1, 0, 3, 0, 23, 0, 5, 0, 5, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 1, 0, 5, 0, 3, 0, 1, 0, 9, 0, 9, 0, 3, 0, 1, 0, 5, 0, 5, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 9, 0, 1, 0, 5, 0, 3, 0, 9, 0, 9, 0, 9, 0, 3, 0, 9, 0, 5, 0, 5, 0, 3, 0, 9, 0, 23, 0, 23, 0, 3, 0, 5, 0, 1, 0, 5, 0, 3, 0, 5, 0, 9, 0, 9, 0, 3, 0, 5, 0, 5, 0, 5, 0, 3, 0, 5, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 5, 0, 3, 0, 23, 0, 9, 0, 9, 0, 3, 0, 23, 0, 5, 0, 5, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 9, 0, 23, 0, 3, 0, 1, 0, 13, 0, 13, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 9, 0, 13, 0, 23, 0, 3, 0, 9, 0, 23, 0, 23, 0, 3, 0, 13, 0, 13, 0, 13, 0, 3, 0, 13, 0, 23, 0, 13, 0, 3, 0, 23, 0, 1, 0, 23, 0, 3, 0, 23, 0, 9, 0, 23, 0, 3, 0, 23, 0, 13, 0, 13, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 5, 0, 5, 0, 3, 0, 1, 0, 13, 0, 13, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 5, 0, 5, 0, 5, 0, 3, 0, 5, 0, 13, 0, 13, 0, 3, 0, 5, 0, 23, 0, 23, 0, 3, 0, 13, 0, 1, 0, 1, 0, 3, 0, 13, 0, 5, 0, 5, 0, 3, 0, 13, 0, 13, 0, 13, 0, 3, 0, 13, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 1, 0, 3, 0, 23, 0, 5, 0, 5, 0, 3, 0, 23, 0, 13, 0, 13, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 1, 0, 1, 0, 3, 0, 1, 0, 5, 0, 5, 0, 3, 0, 1, 0, 13, 0, 13, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 5, 0, 1, 0, 1, 0, 3, 0, 5, 0, 5, 0, 5, 0, 3, 0, 5, 0, 13, 0, 13, 0, 3, 0, 5, 0, 23, 0, 23, 0, 3, 0, 13, 0, 1, 0, 1, 0, 3, 0, 13, 0, 5, 0, 5, 0, 3, 0, 13, 0, 13, 0, 13, 0, 3, 0, 13, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 1, 0, 3, 0, 23, 0, 5, 0, 5, 0, 3, 0, 23, 0, 13, 0, 13, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 1, 0, 9, 0, 9, 0, 3, 0, 1, 0, 5, 0, 5, 0, 3, 0, 1, 0, 13, 0, 13, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 9, 0, 5, 0, 5, 0, 3, 0, 9, 0, 13, 0, 13, 0, 3, 0, 9, 0, 23, 0, 23, 0, 3, 0, 5, 0, 5, 0, 5, 0, 3, 0, 5, 0, 13, 0, 13, 0, 3, 0, 5, 0, 23, 0, 23, 0, 3, 0, 13, 0, 13, 0, 13, 0, 3, 0, 13, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 1, 0, 3, 0, 23, 0, 9, 0, 9, 0, 3, 0, 23, 0, 5, 0, 5, 0, 3, 0, 23, 0, 13, 0, 13, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 18, 0, 1, 0, 1, 0, 3, 0, 18, 0, 9, 0, 9, 0, 3, 0, 18, 0, 5, 0, 5, 0, 3, 0, 18, 0, 13, 0, 13, 0, 3, 0, 18, 0, 23, 0, 23, 0, 3, 0, 1, 0, 9, 0, 9, 0, 3, 0, 1, 0, 5, 0, 5, 0, 3, 0, 1, 0, 13, 0, 13, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 9, 0, 5, 0, 5, 0, 3, 0, 9, 0, 13, 0, 13, 0, 3, 0, 9, 0, 23, 0, 23, 0, 3, 0, 5, 0, 13, 0, 13, 0, 3, 0, 5, 0, 23, 0, 23, 0, 3, 0, 13, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 1, 0, 3, 0, 23, 0, 9, 0, 9, 0, 3, 0, 23, 0, 5, 0, 5, 0, 3, 0, 23, 0, 13, 0, 13, 0, 3, 0, 23, 0, 23, 0, 23, 0, 3, 0, 18, 0, 1, 0, 18, 0, 3, 0, 18, 0, 9, 0, 18, 0, 3, 0, 18, 0, 5, 0, 5, 0, 3, 0, 18, 0, 13, 0, 13, 0, 3, 0, 18, 0, 23, 0, 23, 0, 3, 0, 1, 0, 9, 0, 23, 0, 3, 0, 1, 0, 5, 0, 5, 0, 3, 0, 1, 0, 13, 0, 13, 0, 3, 0, 1, 0, 23, 0, 23, 0, 3, 0, 9, 0, 5, 0, 5, 0, 3, 0, 9, 0, 13, 0, 13, 0, 3, 0, 9, 0, 23, 0, 23, 0, 3, 0, 5, 0, 5, 0, 5, 0, 3, 0, 5, 0, 13, 0, 13, 0, 3, 0, 5, 0, 23, 0, 23, 0, 3, 0, 13, 0, 5, 0, 5, 0, 3, 0, 13, 0, 13, 0, 13, 0, 3, 0, 13, 0, 23, 0, 23, 0, 3, 0, 23, 0, 1, 0, 23, 0, 3, 0, 23, 0, 9, 0, 23, 0, 3, 0, 23, 0, 5, 0, 5, 0, 3, 0, 23, 0, 13, 0, 13, 0, 3, 0, 23, 0, 23, 0, 23, 0, 1, 0, 0, 0, 64, 1, 0, 144, 3, 0, 0, 0, 68, 1, 0, 144, 72, 1, 0, 144, 76, 1, 0, 144, 3, 0, 0, 0, 80, 1, 0, 144, 84, 1, 0, 144, 88, 1, 0, 144, 3, 0, 0, 0, 92, 1, 0, 144, 96, 1, 0, 144, 100, 1, 0, 144, 3, 0, 0, 0, 104, 1, 0, 144, 108, 1, 0, 144, 112, 1, 0, 144, 4, 0, 0, 0, 116, 1, 0, 144, 120, 1, 0, 144, 124, 1, 0, 144, 128, 1, 0, 144, 4, 0, 0, 0, 132, 1, 0, 144, 136, 1, 0, 144, 140, 1, 0, 144, 144, 1, 0, 144, 4, 0, 0, 0, 148, 1, 0, 144, 152, 1, 0, 144, 156, 1, 0, 144, 160, 1, 0, 144, 4, 0, 0, 0, 164, 1, 0, 144, 168, 1, 0, 144, 172, 1, 0, 144, 176, 1, 0, 144, 9, 0, 0, 0, 180, 1, 0, 144, 184, 1, 0, 144, 188, 1, 0, 144, 192, 1, 0, 144, 196, 1, 0, 144, 200, 1, 0, 144, 204, 1, 0, 144, 208, 1, 0, 144, 212, 1, 0, 144, 7, 0, 0, 0, 216, 1, 0, 144, 220, 1, 0, 144, 224, 1, 0, 144, 228, 1, 0, 144, 232, 1, 0, 144, 236, 1, 0, 144, 240, 1, 0, 144, 9, 0, 0, 0, 244, 1, 0, 144, 248, 1, 0, 144, 252, 1, 0, 144, 0, 2, 0, 144, 4, 2, 0, 144, 8, 2, 0, 144, 12, 2, 0, 144, 16, 2, 0, 144, 20, 2, 0, 144, 16, 0, 0, 0, 24, 2, 0, 144, 28, 2, 0, 144, 32, 2, 0, 144, 36, 2, 0, 144, 40, 2, 0, 144, 44, 2, 0, 144, 48, 2, 0, 144, 52, 2, 0, 144, 56, 2, 0, 144, 60, 2, 0, 144, 64, 2, 0, 144, 68, 2, 0, 144, 72, 2, 0, 144, 76, 2, 0, 144, 80, 2, 0, 144, 84, 2, 0, 144, 11, 0, 0, 0, 88, 2, 0, 144, 92, 2, 0, 144, 96, 2, 0, 144, 100, 2, 0, 144, 104, 2, 0, 144, 108, 2, 0, 144, 112, 2, 0, 144, 116, 2, 0, 144, 120, 2, 0, 144, 124, 2, 0, 144, 128, 2, 0, 144, 14, 0, 0, 0, 132, 2, 0, 144, 136, 2, 0, 144, 140, 2, 0, 144, 144, 2, 0, 144, 148, 2, 0, 144, 152, 2, 0, 144, 156, 2, 0, 144, 160, 2, 0, 144, 164, 2, 0, 144, 168, 2, 0, 144, 172, 2, 0, 144, 176, 2, 0, 144, 180, 2, 0, 144, 184, 2, 0, 144, 16, 0, 0, 0, 188, 2, 0, 144, 192, 2, 0, 144, 196, 2, 0, 144, 200, 2, 0, 144, 204, 2, 0, 144, 208, 2, 0, 144, 212, 2, 0, 144, 216, 2, 0, 144, 220, 2, 0, 144, 224, 2, 0, 144, 228, 2, 0, 144, 232, 2, 0, 144, 236, 2, 0, 144, 240, 2, 0, 144, 244, 2, 0, 144, 248, 2, 0, 144, 17, 0, 0, 0, 252, 2, 0, 144, 0, 3, 0, 144, 4, 3, 0, 144, 8, 3, 0, 144, 12, 3, 0, 144, 16, 3, 0, 144, 20, 3, 0, 144, 24, 3, 0, 144, 28, 3, 0, 144, 32, 3, 0, 144, 36, 3, 0, 144, 40, 3, 0, 144, 44, 3, 0, 144, 48, 3, 0, 144, 52, 3, 0, 144, 56, 3, 0, 144, 60, 3, 0, 144, 20, 0, 0, 0, 64, 3, 0, 144, 68, 3, 0, 144, 72, 3, 0, 144, 76, 3, 0, 144, 80, 3, 0, 144, 84, 3, 0, 144, 88, 3, 0, 144, 92, 3, 0, 144, 96, 3, 0, 144, 100, 3, 0, 144, 104, 3, 0, 144, 108, 3, 0, 144, 112, 3, 0, 144, 116, 3, 0, 144, 120, 3, 0, 144, 124, 3, 0, 144, 128, 3, 0, 144, 132, 3, 0, 144, 136, 3, 0, 144, 140, 3, 0, 144, 23, 0, 0, 0, 144, 3, 0, 144, 148, 3, 0, 144, 152, 3, 0, 144, 156, 3, 0, 144, 160, 3, 0, 144, 164, 3, 0, 144, 168, 3, 0, 144, 172, 3, 0, 144, 176, 3, 0, 144, 180, 3, 0, 144, 184, 3, 0, 144, 188, 3, 0, 144, 192, 3, 0, 144, 196, 3, 0, 144, 200, 3, 0, 144, 204, 3, 0, 144, 208, 3, 0, 144, 212, 3, 0, 144, 216, 3, 0, 144, 220, 3, 0, 144, 224, 3, 0, 144, 228, 3, 0, 144, 232, 3, 0, 144, 19, 0, 52, 1, 58, 1, 64, 1, 70, 1, 76, 1, 82, 1, 88, 1, 94, 1, 100, 1, 106, 1, 112, 1, 118, 1, 124, 1, 130, 1, 136, 1, 142, 1, 148, 1, 154, 1, 160, 1, 96, 2, 0, 128, 98, 2, 0, 128, 102, 2, 0, 128, 106, 2, 0, 128, 110, 2, 0, 128, 114, 2, 0, 128, 119, 2, 0, 128, 124, 2, 0, 128, 129, 2, 0, 128, 134, 2, 0, 128, 144, 2, 0, 128, 152, 2, 0, 128, 162, 2, 0, 128, 179, 2, 0, 128, 191, 2, 0, 128, 206, 2, 0, 128, 223, 2, 0, 128, 241, 2, 0, 128, 6, 3, 0, 128, 2, 0, 32, 0, 46, 1, 170, 170, 143, 0, 0, 80, 30, 3, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 111, 32, 0, 32, 7, 0, 0, 0, 15, 1, 0, 0, 116, 32, 0, 0, 116, 32, 0, 0, 211, 0, 0, 0, 1, 0, 0, 0, 22, 32, 0, 0, 108, 111, 99, 97, 108, 101, 115, 0, 97, 102, 0, 97, 107, 0, 97, 109, 0, 97, 110, 0, 97, 114, 0, 97, 114, 115, 0, 97, 115, 97, 0, 97, 115, 116, 0, 97, 122, 0, 98, 101, 0, 98, 101, 109, 0, 98, 101, 122, 0, 98, 103, 0, 98, 104, 111, 0, 98, 109, 0, 98, 110, 0, 98, 114, 0, 98, 114, 120, 0, 98, 115, 0, 99, 97, 0, 99, 101, 0, 99, 101, 98, 0, 99, 103, 103, 0, 99, 104, 114, 0, 99, 107, 98, 0, 99, 115, 0, 99, 121, 0, 100, 97, 0, 100, 115, 98, 0, 100, 118, 0, 100, 122, 0, 101, 101, 0, 101, 108, 0, 101, 110, 0, 101, 116, 0, 101, 117, 0, 102, 97, 0, 102, 102, 0, 102, 105, 0, 102, 105, 108, 0, 102, 111, 0, 102, 114, 0, 102, 117, 114, 0, 102, 121, 0, 103, 97, 0, 103, 100, 0, 103, 108, 0, 103, 115, 119, 0, 103, 117, 0, 103, 117, 119, 0, 103, 118, 0, 104, 97, 0, 104, 97, 119, 0, 104, 101, 0, 104, 115, 98, 0, 104, 117, 0, 104, 121, 0, 105, 97, 0, 105, 100, 0, 105, 105, 0, 105, 110, 0, 105, 111, 0, 105, 115, 0, 105, 116, 0, 105, 117, 0, 105, 119, 0, 106, 97, 0, 106, 98, 111, 0, 106, 103, 111, 0, 106, 105, 0, 106, 109, 99, 0, 106, 118, 0, 106, 119, 0, 107, 97, 0, 107, 97, 98, 0, 107, 97, 106, 0, 107, 99, 103, 0, 107, 100, 101, 0, 107, 101, 97, 0, 107, 107, 0, 107, 107, 106, 0, 107, 108, 0, 107, 109, 0, 107, 110, 0, 107, 111, 0, 107, 115, 0, 107, 115, 98, 0, 107, 115, 104, 0, 107, 117, 0, 107, 119, 0, 107, 121, 0, 108, 97, 103, 0, 108, 98, 0, 108, 103, 0, 108, 107, 116, 0, 108, 110, 0, 108, 111, 0, 108, 116, 0, 108, 118, 0, 109, 97, 115, 0, 109, 103, 0, 109, 103, 111, 0, 109, 107, 0, 109, 108, 0, 109, 111, 0, 109, 114, 0, 109, 116, 0, 109, 121, 0, 110, 97, 104, 0, 110, 97, 113, 0, 110, 98, 0, 110, 100, 0, 110, 108, 0, 110, 110, 0, 110, 110, 104, 0, 110, 111, 0, 110, 113, 111, 0, 110, 114, 0, 110, 115, 111, 0, 110, 121, 110, 0, 111, 109, 0, 111, 114, 0, 111, 115, 0, 111, 115, 97, 0, 112, 97, 0, 112, 97, 112, 0, 112, 108, 0, 112, 114, 103, 0, 112, 115, 0, 112, 116, 0, 112, 116, 95, 80, 84, 0, 114, 109, 0, 114, 111, 102, 0, 114, 111, 111, 116, 0, 114, 117, 0, 114, 119, 107, 0, 115, 97, 104, 0, 115, 97, 113, 0, 115, 99, 0, 115, 99, 110, 0, 115, 100, 0, 115, 100, 104, 0, 115, 101, 0, 115, 101, 104, 0, 115, 101, 115, 0, 115, 103, 0, 115, 104, 105, 0, 115, 105, 0, 115, 107, 0, 115, 108, 0, 115, 109, 97, 0, 115, 109, 105, 0, 115, 109, 106, 0, 115, 109, 110, 0, 115, 109, 115, 0, 115, 110, 0, 115, 113, 0, 115, 114, 0, 115, 115, 0, 115, 115, 121, 0, 115, 117, 0, 115, 118, 0, 115, 121, 114, 0, 116, 97, 0, 116, 101, 0, 116, 101, 111, 0, 116, 104, 0, 116, 105, 0, 116, 105, 103, 0, 116, 107, 0, 116, 108, 0, 116, 110, 0, 116, 111, 0, 116, 114, 0, 116, 115, 0, 116, 122, 109, 0, 117, 103, 0, 117, 107, 0, 117, 122, 0, 118, 101, 0, 118, 105, 0, 118, 111, 0, 118, 117, 110, 0, 119, 97, 0, 119, 97, 101, 0, 120, 104, 0, 120, 111, 103, 0, 121, 105, 0, 121, 111, 0, 121, 117, 101, 0, 122, 104, 0, 122, 117, 0, 108, 111, 99, 97, 108, 101, 115, 95, 111, 114, 100, 105, 110, 97, 108, 115, 0, 114, 117, 108, 101, 115, 0, 115, 101, 116, 48, 0, 111, 116, 104, 101, 114, 0, 115, 101, 116, 49, 0, 111, 110, 101, 0, 115, 101, 116, 49, 48, 0, 115, 101, 116, 49, 49, 0, 115, 101, 116, 49, 50, 0, 115, 101, 116, 49, 51, 0, 122, 101, 114, 111, 0, 115, 101, 116, 49, 52, 0, 115, 101, 116, 49, 53, 0, 115, 101, 116, 49, 54, 0, 116, 119, 111, 0, 115, 101, 116, 49, 55, 0, 102, 101, 119, 0, 115, 101, 116, 49, 56, 0, 115, 101, 116, 49, 57, 0, 115, 101, 116, 50, 0, 115, 101, 116, 50, 48, 0, 115, 101, 116, 50, 49, 0, 115, 101, 116, 50, 50, 0, 115, 101, 116, 50, 51, 0, 109, 97, 110, 121, 0, 115, 101, 116, 50, 52, 0, 115, 101, 116, 50, 53, 0, 115, 101, 116, 50, 54, 0, 115, 101, 116, 50, 55, 0, 115, 101, 116, 50, 56, 0, 115, 101, 116, 50, 57, 0, 115, 101, 116, 51, 0, 115, 101, 116, 51, 48, 0, 115, 101, 116, 51, 49, 0, 115, 101, 116, 51, 50, 0, 115, 101, 116, 51, 51, 0, 115, 101, 116, 51, 52, 0, 115, 101, 116, 51, 53, 0, 115, 101, 116, 51, 54, 0, 115, 101, 116, 51, 55, 0, 115, 101, 116, 51, 56, 0, 115, 101, 116, 51, 57, 0, 115, 101, 116, 52, 0, 115, 101, 116, 52, 48, 0, 115, 101, 116, 52, 49, 0, 115, 101, 116, 52, 50, 0, 115, 101, 116, 52, 51, 0, 115, 101, 116, 52, 52, 0, 115, 101, 116, 52, 53, 0, 115, 101, 116, 52, 54, 0, 115, 101, 116, 52, 55, 0, 115, 101, 116, 52, 56, 0, 115, 101, 116, 52, 57, 0, 115, 101, 116, 53, 0, 115, 101, 116, 53, 48, 0, 115, 101, 116, 53, 49, 0, 115, 101, 116, 53, 50, 0, 115, 101, 116, 53, 51, 0, 115, 101, 116, 53, 52, 0, 115, 101, 116, 53, 53, 0, 115, 101, 116, 53, 54, 0, 115, 101, 116, 53, 55, 0, 115, 101, 116, 53, 56, 0, 115, 101, 116, 54, 0, 115, 101, 116, 55, 0, 115, 101, 116, 56, 0, 115, 101, 116, 57, 0, 170, 170, 170, 0, 0, 115, 0, 101, 0, 116, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 49, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 50, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 53, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 51, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 54, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 56, 0, 0, 0, 115, 0, 101, 0, 116, 0, 52, 0, 57, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 48, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 49, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 50, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 51, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 52, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 55, 0, 0, 0, 115, 0, 101, 0, 116, 0, 53, 0, 56, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 52, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 54, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 54, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 0, 0, 105, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 50, 0, 44, 0, 51, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 51, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 44, 0, 51, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 51, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 44, 0, 53, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 53, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 46, 0, 46, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 126, 0, 52, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 51, 0, 44, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 44, 0, 32, 0, 52, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 53, 0, 44, 0, 54, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 53, 0, 44, 0, 32, 0, 54, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 44, 0, 49, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 49, 0, 49, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 50, 0, 44, 0, 49, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 49, 0, 50, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 51, 0, 44, 0, 49, 0, 51, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 44, 0, 32, 0, 49, 0, 51, 0, 0, 0, 105, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 0, 0, 105, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 48, 0, 44, 0, 55, 0, 44, 0, 56, 0, 44, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 55, 0, 126, 0, 57, 0, 0, 0, 105, 0, 32, 0, 61, 0, 32, 0, 50, 0, 46, 0, 46, 0, 52, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 52, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 44, 0, 53, 0, 44, 0, 55, 0, 46, 0, 46, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 53, 0, 44, 0, 32, 0, 55, 0, 126, 0, 57, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 44, 0, 53, 0, 44, 0, 55, 0, 44, 0, 56, 0, 44, 0, 57, 0, 44, 0, 49, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 53, 0, 44, 0, 32, 0, 55, 0, 126, 0, 49, 0, 48, 0, 0, 0, 105, 0, 32, 0, 61, 0, 32, 0, 48, 0, 44, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 53, 0, 0, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 126, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 51, 0, 126, 0, 49, 0, 57, 0, 44, 0, 32, 0, 50, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 44, 0, 32, 0, 38, 32, 0, 0, 105, 0, 32, 0, 61, 0, 32, 0, 48, 0, 46, 0, 46, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 53, 0, 0, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 49, 0, 44, 0, 56, 0, 44, 0, 56, 0, 48, 0, 44, 0, 56, 0, 48, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 56, 0, 44, 0, 32, 0, 49, 0, 49, 0, 44, 0, 32, 0, 56, 0, 48, 0, 44, 0, 32, 0, 56, 0, 48, 0, 48, 0, 0, 0, 46, 220, 105, 0, 32, 0, 61, 0, 32, 0, 48, 0, 44, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 54, 0, 0, 0, 50, 220, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 50, 220, 110, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 50, 220, 110, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 50, 220, 110, 0, 32, 0, 61, 0, 32, 0, 51, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 50, 220, 110, 0, 32, 0, 61, 0, 32, 0, 54, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 54, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 52, 220, 32, 0, 32, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 52, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 126, 0, 49, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 52, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 126, 0, 53, 0, 44, 0, 32, 0, 55, 0, 44, 0, 32, 0, 56, 0, 44, 0, 32, 0, 49, 0, 49, 0, 126, 0, 49, 0, 53, 0, 44, 0, 32, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 56, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 53, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 48, 0, 126, 0, 50, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 53, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 48, 0, 126, 0, 51, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 53, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 49, 0, 126, 0, 51, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 55, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 126, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 55, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 51, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 55, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 52, 0, 126, 0, 49, 0, 56, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 55, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 53, 0, 126, 0, 49, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 55, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 54, 0, 126, 0, 50, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 55, 220, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 111, 0, 114, 0, 32, 0, 116, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 61, 0, 32, 0, 48, 0, 44, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 54, 0, 0, 0, 56, 220, 105, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 126, 0, 48, 0, 46, 0, 48, 0, 52, 0, 0, 0, 56, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 126, 0, 50, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 56, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 126, 0, 50, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 57, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 126, 0, 50, 0, 44, 0, 32, 0, 52, 0, 126, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 58, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 44, 0, 32, 0, 52, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 58, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 53, 0, 44, 0, 32, 0, 55, 0, 126, 0, 50, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 58, 220, 102, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 32, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 60, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 126, 0, 52, 0, 44, 0, 32, 0, 54, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 60, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 51, 0, 126, 0, 54, 0, 44, 0, 32, 0, 57, 0, 126, 0, 49, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 61, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 44, 0, 32, 0, 51, 0, 44, 0, 32, 0, 53, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 62, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 52, 0, 126, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 126, 0, 50, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 64, 220, 105, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 111, 0, 114, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 46, 0, 46, 0, 50, 0, 48, 0, 44, 0, 52, 0, 48, 0, 44, 0, 54, 0, 48, 0, 44, 0, 56, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 126, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 0, 0, 65, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 126, 0, 55, 0, 44, 0, 32, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 73, 220, 32, 0, 32, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 73, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 49, 0, 49, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 55, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 73, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 49, 0, 50, 0, 44, 0, 32, 0, 50, 0, 50, 0, 44, 0, 32, 0, 51, 0, 50, 0, 44, 0, 32, 0, 52, 0, 50, 0, 44, 0, 32, 0, 53, 0, 50, 0, 44, 0, 32, 0, 54, 0, 50, 0, 44, 0, 32, 0, 55, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 0, 0, 75, 220, 110, 0, 32, 0, 61, 0, 32, 0, 53, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 53, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 53, 0, 44, 0, 32, 0, 50, 0, 48, 0, 53, 0, 44, 0, 32, 0, 51, 0, 48, 0, 53, 0, 44, 0, 32, 0, 52, 0, 48, 0, 53, 0, 44, 0, 32, 0, 53, 0, 48, 0, 53, 0, 44, 0, 32, 0, 54, 0, 48, 0, 53, 0, 44, 0, 32, 0, 55, 0, 48, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 53, 0, 44, 0, 32, 0, 38, 32, 0, 0, 76, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 126, 0, 53, 0, 44, 0, 32, 0, 55, 0, 44, 0, 32, 0, 56, 0, 44, 0, 32, 0, 49, 0, 49, 0, 126, 0, 49, 0, 53, 0, 44, 0, 32, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 56, 0, 44, 0, 32, 0, 50, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 76, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 50, 0, 48, 0, 49, 0, 44, 0, 32, 0, 51, 0, 48, 0, 49, 0, 44, 0, 32, 0, 52, 0, 48, 0, 49, 0, 44, 0, 32, 0, 53, 0, 48, 0, 49, 0, 44, 0, 32, 0, 54, 0, 48, 0, 49, 0, 44, 0, 32, 0, 55, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 76, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 50, 0, 48, 0, 50, 0, 44, 0, 32, 0, 51, 0, 48, 0, 50, 0, 44, 0, 32, 0, 52, 0, 48, 0, 50, 0, 44, 0, 32, 0, 53, 0, 48, 0, 50, 0, 44, 0, 32, 0, 54, 0, 48, 0, 50, 0, 44, 0, 32, 0, 55, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 0, 0, 78, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 54, 0, 44, 0, 57, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 54, 0, 44, 0, 32, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 57, 0, 44, 0, 32, 0, 50, 0, 54, 0, 44, 0, 32, 0, 50, 0, 57, 0, 44, 0, 32, 0, 51, 0, 54, 0, 44, 0, 32, 0, 51, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 54, 0, 44, 0, 32, 0, 38, 32, 0, 0, 78, 220, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 44, 0, 49, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 49, 0, 49, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 78, 220, 110, 0, 32, 0, 61, 0, 32, 0, 50, 0, 44, 0, 49, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 49, 0, 50, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 79, 220, 118, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 32, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 81, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 51, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 51, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 44, 0, 32, 0, 50, 0, 51, 0, 44, 0, 32, 0, 51, 0, 51, 0, 44, 0, 32, 0, 52, 0, 51, 0, 44, 0, 32, 0, 53, 0, 51, 0, 44, 0, 32, 0, 54, 0, 51, 0, 44, 0, 32, 0, 55, 0, 51, 0, 44, 0, 32, 0, 56, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 44, 0, 32, 0, 38, 32, 0, 0, 81, 220, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 55, 0, 49, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 81, 220, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 50, 0, 50, 0, 44, 0, 32, 0, 51, 0, 50, 0, 44, 0, 32, 0, 52, 0, 50, 0, 44, 0, 32, 0, 53, 0, 50, 0, 44, 0, 32, 0, 54, 0, 50, 0, 44, 0, 32, 0, 55, 0, 50, 0, 44, 0, 32, 0, 56, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 0, 0, 81, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 55, 0, 49, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 81, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 50, 0, 50, 0, 44, 0, 32, 0, 51, 0, 50, 0, 44, 0, 32, 0, 52, 0, 50, 0, 44, 0, 32, 0, 53, 0, 50, 0, 44, 0, 32, 0, 54, 0, 50, 0, 44, 0, 32, 0, 55, 0, 50, 0, 44, 0, 32, 0, 56, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 0, 0, 81, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 52, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 52, 0, 44, 0, 32, 0, 50, 0, 52, 0, 44, 0, 32, 0, 51, 0, 52, 0, 44, 0, 32, 0, 52, 0, 52, 0, 44, 0, 32, 0, 53, 0, 52, 0, 44, 0, 32, 0, 54, 0, 52, 0, 44, 0, 32, 0, 55, 0, 52, 0, 44, 0, 32, 0, 56, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 52, 0, 44, 0, 32, 0, 38, 32, 0, 0, 82, 220, 110, 0, 32, 0, 61, 0, 32, 0, 48, 0, 46, 0, 46, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 84, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 57, 0, 44, 0, 32, 0, 50, 0, 50, 0, 126, 0, 50, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 50, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 50, 0, 126, 0, 49, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 50, 0, 44, 0, 32, 0, 38, 32, 0, 0, 85, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 57, 0, 44, 0, 32, 0, 50, 0, 57, 0, 44, 0, 32, 0, 51, 0, 48, 0, 44, 0, 32, 0, 51, 0, 57, 0, 44, 0, 32, 0, 52, 0, 57, 0, 44, 0, 32, 0, 53, 0, 57, 0, 44, 0, 32, 0, 54, 0, 57, 0, 44, 0, 32, 0, 55, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 91, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 55, 0, 49, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 93, 220, 105, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 111, 0, 114, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 54, 0, 32, 0, 111, 0, 114, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 52, 0, 48, 0, 44, 0, 54, 0, 48, 0, 44, 0, 57, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 54, 0, 44, 0, 32, 0, 49, 0, 54, 0, 44, 0, 32, 0, 50, 0, 54, 0, 44, 0, 32, 0, 51, 0, 54, 0, 44, 0, 32, 0, 52, 0, 48, 0, 44, 0, 32, 0, 52, 0, 54, 0, 44, 0, 32, 0, 53, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 54, 0, 44, 0, 32, 0, 38, 32, 0, 0, 93, 220, 110, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 98, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 52, 0, 126, 0, 49, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 102, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 51, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 102, 220, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 46, 0, 46, 0, 52, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 46, 0, 46, 0, 52, 0, 44, 0, 50, 0, 49, 0, 46, 0, 46, 0, 50, 0, 52, 0, 44, 0, 52, 0, 49, 0, 46, 0, 46, 0, 52, 0, 52, 0, 44, 0, 54, 0, 49, 0, 46, 0, 46, 0, 54, 0, 52, 0, 44, 0, 56, 0, 49, 0, 46, 0, 46, 0, 56, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 126, 0, 52, 0, 44, 0, 32, 0, 50, 0, 49, 0, 126, 0, 50, 0, 52, 0, 44, 0, 32, 0, 52, 0, 49, 0, 126, 0, 52, 0, 52, 0, 44, 0, 32, 0, 54, 0, 49, 0, 126, 0, 54, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 104, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 46, 0, 46, 0, 52, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 50, 0, 46, 0, 46, 0, 49, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 52, 0, 44, 0, 32, 0, 50, 0, 50, 0, 126, 0, 50, 0, 52, 0, 44, 0, 32, 0, 51, 0, 50, 0, 126, 0, 51, 0, 52, 0, 44, 0, 32, 0, 52, 0, 50, 0, 126, 0, 52, 0, 52, 0, 44, 0, 32, 0, 53, 0, 50, 0, 126, 0, 53, 0, 52, 0, 44, 0, 32, 0, 54, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 0, 0, 108, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 48, 0, 44, 0, 50, 0, 48, 0, 44, 0, 52, 0, 48, 0, 44, 0, 54, 0, 48, 0, 44, 0, 56, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 48, 0, 44, 0, 32, 0, 52, 0, 48, 0, 44, 0, 32, 0, 54, 0, 48, 0, 44, 0, 32, 0, 56, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 111, 220, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 44, 0, 50, 0, 44, 0, 53, 0, 44, 0, 55, 0, 44, 0, 56, 0, 32, 0, 111, 0, 114, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 48, 0, 44, 0, 53, 0, 48, 0, 44, 0, 55, 0, 48, 0, 44, 0, 56, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 44, 0, 32, 0, 53, 0, 44, 0, 32, 0, 55, 0, 44, 0, 32, 0, 56, 0, 44, 0, 32, 0, 49, 0, 49, 0, 44, 0, 32, 0, 49, 0, 50, 0, 44, 0, 32, 0, 49, 0, 53, 0, 44, 0, 32, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 56, 0, 44, 0, 32, 0, 50, 0, 48, 0, 126, 0, 50, 0, 50, 0, 44, 0, 32, 0, 50, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 113, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 46, 0, 46, 0, 49, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 48, 0, 44, 0, 32, 0, 51, 0, 48, 0, 44, 0, 32, 0, 52, 0, 48, 0, 44, 0, 32, 0, 53, 0, 48, 0, 44, 0, 32, 0, 54, 0, 48, 0, 44, 0, 32, 0, 55, 0, 48, 0, 44, 0, 32, 0, 56, 0, 48, 0, 44, 0, 32, 0, 57, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 114, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 51, 0, 46, 0, 46, 0, 49, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 126, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 51, 0, 126, 0, 49, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 57, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 117, 220, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 55, 0, 44, 0, 56, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 55, 0, 44, 0, 49, 0, 56, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 55, 0, 44, 0, 32, 0, 56, 0, 44, 0, 32, 0, 50, 0, 55, 0, 44, 0, 32, 0, 50, 0, 56, 0, 44, 0, 32, 0, 51, 0, 55, 0, 44, 0, 32, 0, 51, 0, 56, 0, 44, 0, 32, 0, 52, 0, 55, 0, 44, 0, 32, 0, 52, 0, 56, 0, 44, 0, 32, 0, 53, 0, 55, 0, 44, 0, 32, 0, 53, 0, 56, 0, 44, 0, 32, 0, 54, 0, 55, 0, 44, 0, 32, 0, 54, 0, 56, 0, 44, 0, 32, 0, 55, 0, 55, 0, 44, 0, 32, 0, 55, 0, 56, 0, 44, 0, 32, 0, 56, 0, 55, 0, 44, 0, 32, 0, 56, 0, 56, 0, 44, 0, 32, 0, 49, 0, 48, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 55, 0, 44, 0, 32, 0, 38, 32, 0, 0, 117, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 44, 0, 50, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 44, 0, 49, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 50, 0, 50, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 51, 0, 50, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 52, 0, 50, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 53, 0, 50, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 54, 0, 50, 0, 44, 0, 32, 0, 55, 0, 49, 0, 44, 0, 32, 0, 55, 0, 50, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 56, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 117, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 44, 0, 51, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 50, 0, 44, 0, 49, 0, 51, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 51, 0, 44, 0, 32, 0, 50, 0, 50, 0, 44, 0, 32, 0, 50, 0, 51, 0, 44, 0, 32, 0, 51, 0, 50, 0, 44, 0, 32, 0, 51, 0, 51, 0, 44, 0, 32, 0, 52, 0, 50, 0, 44, 0, 32, 0, 52, 0, 51, 0, 44, 0, 32, 0, 53, 0, 50, 0, 44, 0, 32, 0, 53, 0, 51, 0, 44, 0, 32, 0, 54, 0, 50, 0, 44, 0, 32, 0, 54, 0, 51, 0, 44, 0, 32, 0, 55, 0, 50, 0, 44, 0, 32, 0, 55, 0, 51, 0, 44, 0, 32, 0, 56, 0, 50, 0, 44, 0, 32, 0, 56, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 0, 0, 119, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 49, 0, 46, 0, 46, 0, 57, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 49, 0, 126, 0, 50, 0, 54, 0, 44, 0, 32, 0, 49, 0, 49, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 119, 220, 110, 0, 32, 0, 61, 0, 32, 0, 50, 0, 46, 0, 46, 0, 49, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 49, 0, 48, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 57, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 48, 0, 0, 0, 121, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 53, 0, 126, 0, 56, 0, 44, 0, 32, 0, 49, 0, 48, 0, 126, 0, 50, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 123, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 50, 0, 46, 0, 48, 0, 126, 0, 51, 0, 46, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 123, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 126, 0, 49, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 123, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 50, 0, 46, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 123, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 52, 0, 44, 0, 32, 0, 54, 0, 44, 0, 32, 0, 57, 0, 44, 0, 32, 0, 49, 0, 52, 0, 44, 0, 32, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 57, 0, 44, 0, 32, 0, 50, 0, 52, 0, 44, 0, 32, 0, 50, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 52, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 52, 0, 44, 0, 32, 0, 48, 0, 46, 0, 54, 0, 44, 0, 32, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 52, 0, 44, 0, 32, 0, 49, 0, 46, 0, 54, 0, 44, 0, 32, 0, 49, 0, 46, 0, 57, 0, 44, 0, 32, 0, 50, 0, 46, 0, 52, 0, 44, 0, 32, 0, 50, 0, 46, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 52, 0, 44, 0, 32, 0, 38, 32, 0, 0, 123, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 49, 0, 46, 0, 46, 0, 49, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 49, 0, 126, 0, 49, 0, 57, 0, 44, 0, 32, 0, 49, 0, 49, 0, 49, 0, 126, 0, 49, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 126, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 126, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 126, 220, 110, 0, 32, 0, 61, 0, 32, 0, 48, 0, 44, 0, 49, 0, 32, 0, 111, 0, 114, 0, 32, 0, 105, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 102, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 48, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 131, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 126, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 126, 0, 51, 0, 46, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 131, 220, 110, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 46, 0, 46, 0, 49, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 126, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 126, 0, 49, 0, 48, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 132, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 132, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 50, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 56, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 132, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 111, 0, 114, 0, 32, 0, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 53, 0, 46, 0, 46, 0, 57, 0, 32, 0, 111, 0, 114, 0, 32, 0, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 49, 0, 46, 0, 46, 0, 49, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 53, 0, 126, 0, 49, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 133, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 49, 0, 126, 0, 50, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 57, 0, 44, 0, 32, 0, 50, 0, 46, 0, 49, 0, 126, 0, 50, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 133, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 48, 0, 126, 0, 51, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 133, 220, 110, 0, 32, 0, 61, 0, 32, 0, 51, 0, 46, 0, 46, 0, 54, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 126, 0, 54, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 134, 220, 110, 0, 32, 0, 61, 0, 32, 0, 51, 0, 46, 0, 46, 0, 49, 0, 48, 0, 44, 0, 49, 0, 51, 0, 46, 0, 46, 0, 49, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 126, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 51, 0, 126, 0, 49, 0, 57, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 57, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 57, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 48, 0, 0, 0, 135, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 126, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 135, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 51, 0, 126, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 136, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 126, 0, 50, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 136, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 48, 0, 126, 0, 51, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 136, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 126, 0, 49, 0, 48, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 136, 220, 118, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 46, 0, 46, 0, 49, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 126, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 138, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 52, 0, 44, 0, 32, 0, 53, 0, 44, 0, 32, 0, 55, 0, 126, 0, 50, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 139, 220, 110, 0, 32, 0, 61, 0, 32, 0, 55, 0, 46, 0, 46, 0, 49, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 55, 0, 126, 0, 49, 0, 48, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 57, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 57, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 57, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 57, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 48, 0, 48, 0, 48, 0, 0, 0, 140, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 126, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 50, 0, 126, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 50, 0, 126, 0, 49, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 141, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 46, 0, 46, 0, 57, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 46, 0, 46, 0, 49, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 57, 0, 44, 0, 32, 0, 50, 0, 50, 0, 126, 0, 50, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 57, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 141, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 54, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 57, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 54, 0, 44, 0, 32, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 44, 0, 32, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 57, 0, 44, 0, 32, 0, 50, 0, 48, 0, 44, 0, 32, 0, 50, 0, 54, 0, 44, 0, 32, 0, 50, 0, 57, 0, 44, 0, 32, 0, 51, 0, 48, 0, 44, 0, 32, 0, 51, 0, 54, 0, 44, 0, 32, 0, 51, 0, 57, 0, 44, 0, 32, 0, 52, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 142, 220, 116, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 32, 0, 111, 0, 114, 0, 32, 0, 116, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 55, 0, 49, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 143, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 51, 0, 44, 0, 50, 0, 51, 0, 44, 0, 52, 0, 51, 0, 44, 0, 54, 0, 51, 0, 44, 0, 56, 0, 51, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 44, 0, 32, 0, 50, 0, 51, 0, 44, 0, 32, 0, 52, 0, 51, 0, 44, 0, 32, 0, 54, 0, 51, 0, 44, 0, 32, 0, 56, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 51, 0, 44, 0, 32, 0, 49, 0, 50, 0, 51, 0, 44, 0, 32, 0, 49, 0, 52, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 145, 220, 110, 0, 32, 0, 61, 0, 32, 0, 48, 0, 46, 0, 46, 0, 49, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 61, 0, 32, 0, 49, 0, 49, 0, 46, 0, 46, 0, 57, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 44, 0, 32, 0, 49, 0, 49, 0, 126, 0, 50, 0, 52, 0, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 57, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 52, 0, 46, 0, 48, 0, 0, 0, 146, 220, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 51, 0, 44, 0, 52, 0, 32, 0, 111, 0, 114, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 50, 0, 48, 0, 48, 0, 44, 0, 51, 0, 48, 0, 48, 0, 44, 0, 52, 0, 48, 0, 48, 0, 44, 0, 53, 0, 48, 0, 48, 0, 44, 0, 54, 0, 48, 0, 48, 0, 44, 0, 55, 0, 48, 0, 48, 0, 44, 0, 56, 0, 48, 0, 48, 0, 44, 0, 57, 0, 48, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 44, 0, 32, 0, 52, 0, 44, 0, 32, 0, 49, 0, 51, 0, 44, 0, 32, 0, 49, 0, 52, 0, 44, 0, 32, 0, 50, 0, 51, 0, 44, 0, 32, 0, 50, 0, 52, 0, 44, 0, 32, 0, 51, 0, 51, 0, 44, 0, 32, 0, 51, 0, 52, 0, 44, 0, 32, 0, 52, 0, 51, 0, 44, 0, 32, 0, 52, 0, 52, 0, 44, 0, 32, 0, 53, 0, 51, 0, 44, 0, 32, 0, 53, 0, 52, 0, 44, 0, 32, 0, 54, 0, 51, 0, 44, 0, 32, 0, 54, 0, 52, 0, 44, 0, 32, 0, 55, 0, 51, 0, 44, 0, 32, 0, 55, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 44, 0, 32, 0, 38, 32, 0, 0, 146, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 48, 0, 46, 0, 46, 0, 49, 0, 32, 0, 111, 0, 114, 0, 32, 0, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 53, 0, 46, 0, 46, 0, 57, 0, 32, 0, 111, 0, 114, 0, 32, 0, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 50, 0, 46, 0, 46, 0, 49, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 53, 0, 126, 0, 49, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 149, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 53, 0, 126, 0, 49, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 48, 0, 46, 0, 53, 0, 126, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 53, 0, 126, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 53, 0, 126, 0, 50, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 150, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 44, 0, 55, 0, 49, 0, 44, 0, 57, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 150, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 50, 0, 44, 0, 55, 0, 50, 0, 44, 0, 57, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 50, 0, 50, 0, 44, 0, 32, 0, 51, 0, 50, 0, 44, 0, 32, 0, 52, 0, 50, 0, 44, 0, 32, 0, 53, 0, 50, 0, 44, 0, 32, 0, 54, 0, 50, 0, 44, 0, 32, 0, 56, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 154, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 55, 0, 49, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 55, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 157, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 50, 0, 126, 0, 49, 0, 54, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 158, 220, 110, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 44, 0, 50, 0, 49, 0, 44, 0, 52, 0, 49, 0, 44, 0, 54, 0, 49, 0, 44, 0, 56, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 50, 0, 49, 0, 44, 0, 32, 0, 49, 0, 52, 0, 49, 0, 44, 0, 32, 0, 49, 0, 54, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 50, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 54, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 158, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 46, 0, 46, 0, 49, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 55, 0, 49, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 55, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 163, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 111, 0, 114, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 50, 0, 48, 0, 49, 0, 44, 0, 32, 0, 51, 0, 48, 0, 49, 0, 44, 0, 32, 0, 52, 0, 48, 0, 49, 0, 44, 0, 32, 0, 53, 0, 48, 0, 49, 0, 44, 0, 32, 0, 54, 0, 48, 0, 49, 0, 44, 0, 32, 0, 55, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 44, 0, 32, 0, 50, 0, 46, 0, 49, 0, 44, 0, 32, 0, 51, 0, 46, 0, 49, 0, 44, 0, 32, 0, 52, 0, 46, 0, 49, 0, 44, 0, 32, 0, 53, 0, 46, 0, 49, 0, 44, 0, 32, 0, 54, 0, 46, 0, 49, 0, 44, 0, 32, 0, 55, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 163, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 111, 0, 114, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 50, 0, 48, 0, 50, 0, 44, 0, 32, 0, 51, 0, 48, 0, 50, 0, 44, 0, 32, 0, 52, 0, 48, 0, 50, 0, 44, 0, 32, 0, 53, 0, 48, 0, 50, 0, 44, 0, 32, 0, 54, 0, 48, 0, 50, 0, 44, 0, 32, 0, 55, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 50, 0, 44, 0, 32, 0, 49, 0, 46, 0, 50, 0, 44, 0, 32, 0, 50, 0, 46, 0, 50, 0, 44, 0, 32, 0, 51, 0, 46, 0, 50, 0, 44, 0, 32, 0, 52, 0, 46, 0, 50, 0, 44, 0, 32, 0, 53, 0, 46, 0, 50, 0, 44, 0, 32, 0, 54, 0, 46, 0, 50, 0, 44, 0, 32, 0, 55, 0, 46, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 50, 0, 44, 0, 32, 0, 38, 32, 0, 0, 165, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 46, 0, 46, 0, 52, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 50, 0, 46, 0, 46, 0, 49, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 52, 0, 44, 0, 32, 0, 50, 0, 50, 0, 126, 0, 50, 0, 52, 0, 44, 0, 32, 0, 51, 0, 50, 0, 126, 0, 51, 0, 52, 0, 44, 0, 32, 0, 52, 0, 50, 0, 126, 0, 52, 0, 52, 0, 44, 0, 32, 0, 53, 0, 50, 0, 126, 0, 53, 0, 52, 0, 44, 0, 32, 0, 54, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 171, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 48, 0, 48, 0, 126, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 50, 0, 48, 0, 48, 0, 126, 0, 50, 0, 48, 0, 50, 0, 44, 0, 32, 0, 51, 0, 48, 0, 48, 0, 126, 0, 51, 0, 48, 0, 50, 0, 44, 0, 32, 0, 52, 0, 48, 0, 48, 0, 126, 0, 52, 0, 48, 0, 50, 0, 44, 0, 32, 0, 53, 0, 48, 0, 48, 0, 126, 0, 53, 0, 48, 0, 50, 0, 44, 0, 32, 0, 54, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 126, 0, 48, 0, 46, 0, 57, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 126, 0, 49, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 175, 220, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 126, 0, 50, 0, 48, 0, 44, 0, 32, 0, 51, 0, 48, 0, 44, 0, 32, 0, 52, 0, 48, 0, 44, 0, 32, 0, 53, 0, 48, 0, 44, 0, 32, 0, 54, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 189, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 51, 0, 46, 0, 46, 0, 52, 0, 44, 0, 57, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 48, 0, 46, 0, 46, 0, 49, 0, 57, 0, 44, 0, 55, 0, 48, 0, 46, 0, 46, 0, 55, 0, 57, 0, 44, 0, 57, 0, 48, 0, 46, 0, 46, 0, 57, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 44, 0, 32, 0, 52, 0, 44, 0, 32, 0, 57, 0, 44, 0, 32, 0, 50, 0, 51, 0, 44, 0, 32, 0, 50, 0, 52, 0, 44, 0, 32, 0, 50, 0, 57, 0, 44, 0, 32, 0, 51, 0, 51, 0, 44, 0, 32, 0, 51, 0, 52, 0, 44, 0, 32, 0, 51, 0, 57, 0, 44, 0, 32, 0, 52, 0, 51, 0, 44, 0, 32, 0, 52, 0, 52, 0, 44, 0, 32, 0, 52, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 57, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 57, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 51, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 195, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 32, 0, 111, 0, 114, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 55, 0, 49, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 44, 0, 32, 0, 50, 0, 46, 0, 49, 0, 44, 0, 32, 0, 51, 0, 46, 0, 49, 0, 44, 0, 32, 0, 52, 0, 46, 0, 49, 0, 44, 0, 32, 0, 53, 0, 46, 0, 49, 0, 44, 0, 32, 0, 54, 0, 46, 0, 49, 0, 44, 0, 32, 0, 55, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 198, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 51, 0, 46, 0, 46, 0, 52, 0, 32, 0, 111, 0, 114, 0, 32, 0, 118, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 44, 0, 32, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 52, 0, 44, 0, 32, 0, 50, 0, 48, 0, 51, 0, 44, 0, 32, 0, 50, 0, 48, 0, 52, 0, 44, 0, 32, 0, 51, 0, 48, 0, 51, 0, 44, 0, 32, 0, 51, 0, 48, 0, 52, 0, 44, 0, 32, 0, 52, 0, 48, 0, 51, 0, 44, 0, 32, 0, 52, 0, 48, 0, 52, 0, 44, 0, 32, 0, 53, 0, 48, 0, 51, 0, 44, 0, 32, 0, 53, 0, 48, 0, 52, 0, 44, 0, 32, 0, 54, 0, 48, 0, 51, 0, 44, 0, 32, 0, 54, 0, 48, 0, 52, 0, 44, 0, 32, 0, 55, 0, 48, 0, 51, 0, 44, 0, 32, 0, 55, 0, 48, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 53, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 200, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 53, 0, 46, 0, 46, 0, 57, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 49, 0, 46, 0, 46, 0, 49, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 53, 0, 126, 0, 49, 0, 57, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 55, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 46, 0, 48, 0, 44, 0, 32, 0, 57, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 225, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 32, 0, 111, 0, 114, 0, 32, 0, 118, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 49, 0, 32, 0, 111, 0, 114, 0, 32, 0, 118, 0, 32, 0, 33, 0, 61, 0, 32, 0, 50, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 49, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 51, 0, 49, 0, 44, 0, 32, 0, 52, 0, 49, 0, 44, 0, 32, 0, 53, 0, 49, 0, 44, 0, 32, 0, 54, 0, 49, 0, 44, 0, 32, 0, 55, 0, 49, 0, 44, 0, 32, 0, 56, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 49, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 46, 0, 49, 0, 44, 0, 32, 0, 50, 0, 46, 0, 49, 0, 44, 0, 32, 0, 51, 0, 46, 0, 49, 0, 44, 0, 32, 0, 52, 0, 46, 0, 49, 0, 44, 0, 32, 0, 53, 0, 46, 0, 49, 0, 44, 0, 32, 0, 54, 0, 46, 0, 49, 0, 44, 0, 32, 0, 55, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 49, 0, 44, 0, 32, 0, 38, 32, 0, 0, 225, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 46, 0, 46, 0, 52, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 50, 0, 46, 0, 46, 0, 49, 0, 52, 0, 32, 0, 111, 0, 114, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 46, 0, 46, 0, 52, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 49, 0, 50, 0, 46, 0, 46, 0, 49, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 126, 0, 52, 0, 44, 0, 32, 0, 50, 0, 50, 0, 126, 0, 50, 0, 52, 0, 44, 0, 32, 0, 51, 0, 50, 0, 126, 0, 51, 0, 52, 0, 44, 0, 32, 0, 52, 0, 50, 0, 126, 0, 52, 0, 52, 0, 44, 0, 32, 0, 53, 0, 50, 0, 126, 0, 53, 0, 52, 0, 44, 0, 32, 0, 54, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 50, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 50, 0, 126, 0, 48, 0, 46, 0, 52, 0, 44, 0, 32, 0, 49, 0, 46, 0, 50, 0, 126, 0, 49, 0, 46, 0, 52, 0, 44, 0, 32, 0, 50, 0, 46, 0, 50, 0, 126, 0, 50, 0, 46, 0, 52, 0, 44, 0, 32, 0, 51, 0, 46, 0, 50, 0, 126, 0, 51, 0, 46, 0, 52, 0, 44, 0, 32, 0, 52, 0, 46, 0, 50, 0, 126, 0, 52, 0, 46, 0, 52, 0, 44, 0, 32, 0, 53, 0, 46, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 50, 0, 44, 0, 32, 0, 38, 32, 0, 0, 235, 220, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 49, 0, 46, 0, 46, 0, 49, 0, 57, 0, 32, 0, 111, 0, 114, 0, 32, 0, 118, 0, 32, 0, 61, 0, 32, 0, 50, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 49, 0, 46, 0, 46, 0, 49, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 126, 0, 50, 0, 48, 0, 44, 0, 32, 0, 51, 0, 48, 0, 44, 0, 32, 0, 52, 0, 48, 0, 44, 0, 32, 0, 53, 0, 48, 0, 44, 0, 32, 0, 54, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 49, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 51, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 53, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 54, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 247, 220, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 51, 0, 46, 0, 46, 0, 52, 0, 32, 0, 111, 0, 114, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 51, 0, 46, 0, 46, 0, 52, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 51, 0, 44, 0, 32, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 52, 0, 44, 0, 32, 0, 50, 0, 48, 0, 51, 0, 44, 0, 32, 0, 50, 0, 48, 0, 52, 0, 44, 0, 32, 0, 51, 0, 48, 0, 51, 0, 44, 0, 32, 0, 51, 0, 48, 0, 52, 0, 44, 0, 32, 0, 52, 0, 48, 0, 51, 0, 44, 0, 32, 0, 52, 0, 48, 0, 52, 0, 44, 0, 32, 0, 53, 0, 48, 0, 51, 0, 44, 0, 32, 0, 53, 0, 48, 0, 52, 0, 44, 0, 32, 0, 54, 0, 48, 0, 51, 0, 44, 0, 32, 0, 54, 0, 48, 0, 52, 0, 44, 0, 32, 0, 55, 0, 48, 0, 51, 0, 44, 0, 32, 0, 55, 0, 48, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 51, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 51, 0, 44, 0, 32, 0, 48, 0, 46, 0, 52, 0, 44, 0, 32, 0, 49, 0, 46, 0, 51, 0, 44, 0, 32, 0, 49, 0, 46, 0, 52, 0, 44, 0, 32, 0, 50, 0, 46, 0, 51, 0, 44, 0, 32, 0, 50, 0, 46, 0, 52, 0, 44, 0, 32, 0, 51, 0, 46, 0, 51, 0, 44, 0, 32, 0, 51, 0, 46, 0, 52, 0, 44, 0, 32, 0, 52, 0, 46, 0, 51, 0, 44, 0, 32, 0, 52, 0, 46, 0, 52, 0, 44, 0, 32, 0, 53, 0, 46, 0, 51, 0, 44, 0, 32, 0, 53, 0, 46, 0, 52, 0, 44, 0, 32, 0, 54, 0, 46, 0, 51, 0, 44, 0, 32, 0, 54, 0, 46, 0, 52, 0, 44, 0, 32, 0, 55, 0, 46, 0, 51, 0, 44, 0, 32, 0, 55, 0, 46, 0, 52, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 51, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 51, 0, 44, 0, 32, 0, 38, 32, 0, 0, 19, 221, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 50, 0, 44, 0, 50, 0, 50, 0, 44, 0, 52, 0, 50, 0, 44, 0, 54, 0, 50, 0, 44, 0, 56, 0, 50, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 46, 0, 50, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 52, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 54, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 56, 0, 48, 0, 48, 0, 48, 0, 48, 0, 32, 0, 111, 0, 114, 0, 32, 0, 110, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 110, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 32, 0, 61, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 50, 0, 44, 0, 32, 0, 50, 0, 50, 0, 44, 0, 32, 0, 52, 0, 50, 0, 44, 0, 32, 0, 54, 0, 50, 0, 44, 0, 32, 0, 56, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 44, 0, 32, 0, 49, 0, 50, 0, 50, 0, 44, 0, 32, 0, 49, 0, 52, 0, 50, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 52, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 54, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 56, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 50, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 52, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 29, 221, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 61, 0, 32, 0, 49, 0, 44, 0, 50, 0, 44, 0, 51, 0, 32, 0, 111, 0, 114, 0, 32, 0, 118, 0, 32, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 105, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 52, 0, 44, 0, 54, 0, 44, 0, 57, 0, 32, 0, 111, 0, 114, 0, 32, 0, 118, 0, 32, 0, 33, 0, 61, 0, 32, 0, 48, 0, 32, 0, 97, 0, 110, 0, 100, 0, 32, 0, 102, 0, 32, 0, 37, 0, 32, 0, 49, 0, 48, 0, 32, 0, 33, 0, 61, 0, 32, 0, 52, 0, 44, 0, 54, 0, 44, 0, 57, 0, 32, 0, 64, 0, 105, 0, 110, 0, 116, 0, 101, 0, 103, 0, 101, 0, 114, 0, 32, 0, 48, 0, 126, 0, 51, 0, 44, 0, 32, 0, 53, 0, 44, 0, 32, 0, 55, 0, 44, 0, 32, 0, 56, 0, 44, 0, 32, 0, 49, 0, 48, 0, 126, 0, 49, 0, 51, 0, 44, 0, 32, 0, 49, 0, 53, 0, 44, 0, 32, 0, 49, 0, 55, 0, 44, 0, 32, 0, 49, 0, 56, 0, 44, 0, 32, 0, 50, 0, 48, 0, 44, 0, 32, 0, 50, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 44, 0, 32, 0, 38, 32, 32, 0, 64, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 97, 0, 108, 0, 32, 0, 48, 0, 46, 0, 48, 0, 126, 0, 48, 0, 46, 0, 51, 0, 44, 0, 32, 0, 48, 0, 46, 0, 53, 0, 44, 0, 32, 0, 48, 0, 46, 0, 55, 0, 44, 0, 32, 0, 48, 0, 46, 0, 56, 0, 44, 0, 32, 0, 49, 0, 46, 0, 48, 0, 126, 0, 49, 0, 46, 0, 51, 0, 44, 0, 32, 0, 49, 0, 46, 0, 53, 0, 44, 0, 32, 0, 49, 0, 46, 0, 55, 0, 44, 0, 32, 0, 49, 0, 46, 0, 56, 0, 44, 0, 32, 0, 50, 0, 46, 0, 48, 0, 44, 0, 32, 0, 50, 0, 46, 0, 49, 0, 44, 0, 32, 0, 49, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 49, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 48, 0, 46, 0, 48, 0, 44, 0, 32, 0, 38, 32, 0, 0, 211, 0, 40, 0, 43, 0, 46, 0, 49, 0, 52, 0, 55, 0, 113, 1, 59, 0, 63, 0, 67, 0, 70, 0, 73, 0, 77, 0, 81, 0, 84, 0, 88, 0, 91, 0, 4, 1, 94, 0, 97, 0, 101, 0, 104, 0, 107, 0, 110, 0, 114, 0, 118, 0, 122, 0, 126, 0, 129, 0, 132, 0, 40, 1, 135, 0, 139, 0, 142, 0, 145, 0, 148, 0, 151, 0, 87, 2, 15, 2, 154, 0, 157, 0, 160, 0, 163, 0, 166, 0, 169, 0, 173, 0, 176, 0, 179, 0, 183, 0, 186, 0, 189, 0, 192, 0, 195, 0, 199, 0, 202, 0, 206, 0, 209, 0, 212, 0, 216, 0, 22, 2, 119, 0, 219, 0, 223, 0, 226, 0, 229, 0, 232, 0, 97, 2, 235, 0, 238, 0, 241, 0, 244, 0, 247, 0, 250, 0, 253, 0, 0, 1, 3, 1, 7, 1, 11, 1, 14, 1, 18, 1, 21, 1, 24, 1, 27, 1, 31, 1, 35, 1, 39, 1, 43, 1, 47, 1, 50, 1, 54, 1, 57, 1, 60, 1, 63, 1, 66, 1, 69, 1, 73, 1, 77, 1, 80, 1, 83, 1, 86, 1, 90, 1, 93, 1, 96, 1, 100, 1, 103, 1, 106, 1, 109, 1, 112, 1, 116, 1, 119, 1, 123, 1, 126, 1, 47, 2, 129, 1, 132, 1, 51, 2, 135, 1, 138, 1, 141, 1, 145, 1, 149, 1, 152, 1, 214, 2, 155, 1, 158, 1, 161, 1, 165, 1, 168, 1, 172, 1, 175, 1, 65, 3, 179, 1, 183, 1, 186, 1, 189, 1, 192, 1, 196, 1, 199, 1, 203, 1, 206, 1, 210, 1, 213, 1, 216, 1, 222, 1, 243, 2, 225, 1, 229, 1, 234, 1, 237, 1, 241, 1, 245, 1, 249, 1, 252, 1, 0, 2, 3, 2, 7, 2, 10, 2, 14, 2, 18, 2, 74, 1, 21, 2, 25, 2, 28, 2, 31, 2, 34, 2, 38, 2, 42, 2, 46, 2, 50, 2, 54, 2, 176, 1, 57, 2, 60, 2, 63, 2, 66, 2, 64, 0, 70, 2, 73, 2, 196, 0, 76, 2, 80, 2, 83, 2, 86, 2, 90, 2, 93, 2, 96, 2, 100, 2, 103, 2, 106, 2, 109, 2, 112, 2, 115, 2, 118, 2, 122, 2, 125, 2, 180, 0, 128, 2, 131, 2, 134, 2, 137, 2, 140, 2, 144, 2, 147, 2, 9, 3, 151, 2, 154, 2, 158, 2, 161, 2, 164, 2, 168, 2, 171, 2, 1, 0, 16, 0, 21, 0, 1, 0, 123, 0, 123, 0, 21, 0, 1, 0, 11, 0, 1, 0, 189, 0, 1, 0, 1, 0, 1, 0, 16, 0, 6, 0, 21, 0, 6, 0, 207, 0, 1, 0, 69, 0, 11, 0, 1, 0, 75, 0, 1, 0, 1, 0, 1, 0, 111, 0, 231, 0, 46, 0, 11, 0, 99, 0, 1, 0, 6, 0, 1, 0, 1, 0, 11, 0, 1, 0, 1, 0, 11, 0, 1, 0, 21, 0, 26, 0, 11, 0, 75, 0, 1, 0, 26, 0, 1, 0, 11, 0, 213, 0, 171, 0, 11, 0, 1, 0, 21, 0, 16, 0, 219, 0, 1, 0, 1, 0, 105, 0, 21, 0, 69, 0, 99, 0, 1, 0, 26, 0, 11, 0, 6, 0, 6, 0, 6, 0, 6, 0, 11, 0, 141, 0, 11, 0, 63, 0, 105, 0, 6, 0, 6, 0, 1, 0, 11, 0, 1, 0, 6, 0, 6, 0, 1, 0, 26, 0, 1, 0, 1, 0, 6, 0, 6, 0, 1, 0, 1, 0, 1, 0, 6, 0, 21, 0, 6, 0, 1, 0, 1, 0, 159, 0, 1, 0, 225, 0, 1, 0, 153, 0, 1, 0, 1, 0, 6, 0, 16, 0, 6, 0, 195, 0, 87, 0, 1, 0, 16, 0, 1, 0, 147, 0, 1, 0, 1, 0, 93, 0, 1, 0, 6, 0, 201, 0, 6, 0, 1, 0, 63, 0, 1, 0, 1, 0, 1, 0, 11, 0, 1, 0, 1, 0, 1, 0, 6, 0, 1, 0, 16, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 6, 0, 16, 0, 1, 0, 183, 0, 87, 0, 1, 0, 31, 0, 11, 0, 1, 0, 93, 0, 1, 0, 6, 0, 117, 0, 1, 0, 6, 0, 1, 0, 11, 0, 11, 0, 1, 0, 1, 0, 63, 0, 1, 0, 6, 0, 6, 0, 69, 0, 165, 0, 36, 0, 111, 0, 177, 0, 63, 0, 63, 0, 63, 0, 63, 0, 63, 0, 1, 0, 1, 0, 1, 0, 69, 0, 1, 0, 1, 0, 1, 0, 6, 0, 11, 0, 11, 0, 1, 0, 1, 0, 1, 0, 1, 0, 6, 0, 16, 0, 1, 0, 1, 0, 75, 0, 1, 0, 6, 0, 1, 0, 1, 0, 41, 0, 1, 0, 117, 0, 11, 0, 1, 0, 1, 0, 6, 0, 1, 0, 1, 0, 16, 0, 1, 0, 6, 0, 1, 0, 1, 0, 11, 0, 6, 0, 6, 0, 6, 0, 21, 0, 98, 0, 40, 0, 46, 0, 49, 0, 52, 0, 113, 1, 67, 0, 70, 0, 81, 0, 91, 0, 101, 0, 104, 0, 107, 0, 126, 0, 129, 0, 132, 0, 40, 1, 135, 0, 148, 0, 151, 0, 15, 2, 154, 0, 157, 0, 160, 0, 166, 0, 169, 0, 176, 0, 183, 0, 186, 0, 189, 0, 192, 0, 195, 0, 199, 0, 216, 0, 22, 2, 119, 0, 219, 0, 223, 0, 226, 0, 229, 0, 232, 0, 238, 0, 244, 0, 247, 0, 253, 0, 0, 1, 24, 1, 47, 1, 57, 1, 60, 1, 63, 1, 80, 1, 83, 1, 103, 1, 106, 1, 109, 1, 123, 1, 126, 1, 47, 2, 129, 1, 132, 1, 51, 2, 138, 1, 149, 1, 214, 2, 155, 1, 186, 1, 196, 1, 203, 1, 206, 1, 210, 1, 213, 1, 243, 2, 229, 1, 234, 1, 249, 1, 252, 1, 0, 2, 74, 1, 25, 2, 28, 2, 31, 2, 57, 2, 60, 2, 73, 2, 196, 0, 80, 2, 83, 2, 90, 2, 100, 2, 103, 2, 112, 2, 125, 2, 180, 0, 128, 2, 134, 2, 164, 2, 168, 2, 171, 2, 51, 0, 51, 0, 51, 0, 51, 0, 135, 0, 71, 1, 255, 0, 51, 0, 135, 0, 51, 0, 59, 1, 51, 0, 51, 0, 83, 1, 51, 0, 51, 0, 51, 0, 51, 0, 41, 1, 51, 0, 51, 0, 51, 0, 51, 0, 51, 0, 57, 0, 57, 0, 51, 0, 57, 0, 53, 1, 51, 0, 51, 0, 129, 0, 51, 0, 129, 0, 51, 0, 51, 0, 243, 0, 57, 0, 51, 0, 51, 0, 51, 0, 51, 0, 81, 0, 51, 0, 51, 0, 23, 1, 17, 1, 51, 0, 51, 0, 51, 0, 35, 1, 51, 0, 57, 0, 51, 0, 51, 0, 65, 1, 51, 0, 51, 0, 57, 0, 47, 1, 57, 0, 51, 0, 51, 0, 249, 0, 51, 0, 77, 1, 51, 0, 51, 0, 51, 0, 51, 0, 51, 0, 57, 0, 51, 0, 51, 0, 81, 0, 81, 0, 51, 0, 51, 0, 51, 0, 51, 0, 51, 0, 29, 1, 51, 0, 237, 0, 51, 0, 51, 0, 51, 0, 51, 0, 11, 1, 57, 0, 51, 0, 5, 1, 51, 0, 51, 0, 57, 0, 51, 0, 51, 0, 51, 0, 1, 0, 202, 2, 216, 24, 2, 0, 213, 2, 202, 2, 161, 7, 85, 25, 2, 0, 213, 2, 202, 2, 132, 38, 195, 43, 2, 0, 213, 2, 202, 2, 176, 49, 216, 36, 2, 0, 213, 2, 202, 2, 200, 56, 210, 25, 3, 0, 213, 2, 202, 2, 241, 2, 7, 52, 93, 16, 205, 53, 3, 0, 213, 2, 202, 2, 241, 2, 208, 3, 91, 24, 52, 4, 3, 0, 213, 2, 202, 2, 241, 2, 0, 4, 214, 28, 52, 4, 3, 0, 213, 2, 202, 2, 8, 3, 0, 4, 14, 33, 104, 4, 3, 0, 18, 3, 213, 2, 202, 2, 103, 23, 161, 7, 104, 30, 3, 0, 18, 3, 213, 2, 202, 2, 53, 35, 122, 2, 221, 5, 3, 0, 18, 3, 213, 2, 202, 2, 234, 52, 176, 49, 96, 41, 2, 0, 213, 2, 202, 2, 49, 3, 91, 24, 4, 0, 18, 3, 213, 2, 202, 2, 8, 3, 253, 31, 38, 13, 33, 34, 118, 13, 4, 0, 18, 3, 213, 2, 202, 2, 8, 3, 117, 50, 58, 12, 246, 6, 136, 12, 4, 0, 18, 3, 213, 2, 202, 2, 8, 3, 186, 54, 162, 45, 96, 41, 71, 46, 4, 0, 63, 3, 213, 2, 202, 2, 8, 3, 162, 20, 122, 2, 137, 18, 149, 2, 4, 0, 18, 3, 63, 3, 213, 2, 202, 2, 204, 2, 198, 13, 122, 2, 246, 6, 4, 0, 18, 3, 63, 3, 213, 2, 202, 2, 89, 19, 204, 40, 122, 2, 190, 10, 4, 0, 18, 3, 63, 3, 213, 2, 202, 2, 236, 46, 61, 51, 39, 43, 4, 5, 4, 0, 18, 3, 63, 3, 213, 2, 202, 2, 102, 37, 2, 9, 2, 45, 64, 48, 4, 0, 18, 3, 63, 3, 213, 2, 202, 2, 81, 28, 79, 26, 0, 4, 239, 30, 4, 0, 18, 3, 63, 3, 213, 2, 202, 2, 89, 19, 226, 29, 10, 17, 190, 10, 2, 0, 213, 2, 202, 2, 128, 3, 91, 24, 5, 0, 18, 3, 63, 3, 213, 2, 202, 2, 8, 3, 241, 48, 198, 17, 247, 41, 224, 23, 143, 42, 5, 0, 18, 3, 63, 3, 213, 2, 202, 2, 8, 3, 118, 31, 75, 36, 0, 4, 151, 33, 104, 4, 5, 0, 18, 3, 63, 3, 213, 2, 202, 2, 8, 3, 195, 19, 198, 13, 9, 11, 88, 3, 84, 11, 6, 0, 18, 3, 63, 3, 213, 2, 202, 2, 8, 3, 241, 2, 20, 39, 98, 44, 0, 4, 37, 18, 179, 55, 52, 4, 6, 0, 18, 3, 63, 3, 213, 2, 202, 2, 8, 3, 241, 2, 21, 21, 238, 22, 0, 4, 147, 47, 104, 4, 52, 4, 6, 0, 18, 3, 63, 3, 213, 2, 202, 2, 8, 3, 241, 2, 156, 4, 208, 4, 0, 4, 191, 35, 104, 4, 52, 4, 1, 0, 202, 2, 58, 5, 2, 0, 213, 2, 202, 2, 0, 22, 132, 6, 2, 0, 213, 2, 202, 2, 89, 1, 75, 6, 2, 0, 213, 2, 202, 2, 218, 1, 62, 9, 2, 0, 213, 2, 202, 2, 122, 2, 204, 26, 2, 0, 213, 2, 202, 2, 240, 1, 246, 6, 2, 0, 18, 3, 202, 2, 119, 22, 138, 8, 2, 0, 18, 3, 202, 2, 23, 14, 79, 8, 2, 0, 18, 3, 202, 2, 214, 12, 236, 11, 2, 0, 63, 3, 202, 2, 245, 37, 112, 5, 2, 0, 63, 3, 202, 2, 168, 3, 123, 10, 3, 0, 63, 3, 213, 2, 202, 2, 57, 10, 157, 1, 20, 6, 3, 0, 63, 3, 213, 2, 202, 2, 182, 15, 89, 1, 186, 9, 3, 0, 63, 3, 213, 2, 202, 2, 159, 11, 241, 18, 47, 7, 4, 0, 18, 3, 213, 2, 202, 2, 8, 3, 23, 14, 16, 15, 189, 6, 99, 15, 2, 0, 213, 2, 202, 2, 76, 27, 92, 29, 4, 0, 18, 3, 213, 2, 202, 2, 8, 3, 106, 1, 89, 1, 246, 6, 174, 1, 4, 0, 18, 3, 213, 2, 202, 2, 8, 3, 98, 2, 50, 2, 249, 9, 74, 2, 4, 0, 18, 3, 213, 2, 202, 2, 8, 3, 106, 1, 196, 1, 246, 6, 140, 1, 4, 0, 63, 3, 213, 2, 202, 2, 8, 3, 137, 21, 106, 14, 124, 9, 189, 14, 4, 0, 18, 3, 63, 3, 213, 2, 202, 2, 56, 40, 103, 17, 49, 20, 179, 16, 5, 0, 18, 3, 63, 3, 213, 2, 202, 2, 8, 3, 106, 1, 123, 1, 89, 1, 198, 8, 174, 1, 5, 0, 18, 3, 63, 3, 213, 2, 202, 2, 8, 3, 106, 1, 123, 1, 12, 3, 21, 8, 174, 1, 5, 0, 18, 3, 63, 3, 213, 2, 202, 2, 8, 3, 106, 1, 123, 1, 236, 2, 219, 7, 174, 1, 6, 0, 18, 3, 63, 3, 213, 2, 202, 2, 8, 3, 241, 2, 6, 2, 28, 2, 89, 1, 166, 5, 140, 1, 176, 2, 2, 0, 213, 2, 202, 2, 9, 16, 214, 28, 2, 0, 213, 2, 202, 2, 165, 39, 171, 34, 2, 0, 213, 2, 202, 2, 0, 4, 133, 32, 2, 0, 213, 2, 202, 2, 104, 7, 204, 27, 59, 0, 197, 2, 208, 2, 217, 2, 223, 2, 229, 2, 235, 2, 246, 2, 252, 2, 2, 3, 12, 3, 22, 3, 28, 3, 34, 3, 39, 3, 45, 3, 51, 3, 57, 3, 68, 3, 74, 3, 80, 3, 86, 3, 92, 3, 98, 3, 104, 3, 109, 3, 115, 3, 121, 3, 127, 3, 133, 3, 139, 3, 145, 3, 151, 3, 157, 3, 163, 3, 169, 3, 174, 3, 180, 3, 186, 3, 192, 3, 198, 3, 204, 3, 210, 3, 216, 3, 222, 3, 228, 3, 234, 3, 239, 3, 245, 3, 251, 3, 1, 4, 7, 4, 13, 4, 19, 4, 25, 4, 31, 4, 37, 4, 42, 4, 47, 4, 52, 4, 83, 60, 0, 80, 86, 60, 0, 80, 91, 60, 0, 80, 96, 60, 0, 80, 101, 60, 0, 80, 106, 60, 0, 80, 113, 60, 0, 80, 120, 60, 0, 80, 127, 60, 0, 80, 134, 60, 0, 80, 141, 60, 0, 80, 148, 60, 0, 80, 155, 60, 0, 80, 160, 60, 0, 80, 169, 60, 0, 80, 178, 60, 0, 80, 187, 60, 0, 80, 196, 60, 0, 80, 205, 60, 0, 80, 214, 60, 0, 80, 223, 60, 0, 80, 232, 60, 0, 80, 241, 60, 0, 80, 250, 60, 0, 80, 255, 60, 0, 80, 10, 61, 0, 80, 21, 61, 0, 80, 32, 61, 0, 80, 45, 61, 0, 80, 58, 61, 0, 80, 71, 61, 0, 80, 74, 61, 0, 80, 79, 61, 0, 80, 84, 61, 0, 80, 89, 61, 0, 80, 94, 61, 0, 80, 99, 61, 0, 80, 104, 61, 0, 80, 109, 61, 0, 80, 114, 61, 0, 80, 119, 61, 0, 80, 124, 61, 0, 80, 131, 61, 0, 80, 138, 61, 0, 80, 145, 61, 0, 80, 154, 61, 0, 80, 159, 61, 0, 80, 168, 61, 0, 80, 177, 61, 0, 80, 186, 61, 0, 80, 195, 61, 0, 80, 204, 61, 0, 80, 215, 61, 0, 80, 226, 61, 0, 80, 237, 61, 0, 80, 250, 61, 0, 80, 255, 61, 0, 80, 4, 62, 0, 80, 9, 62, 0, 80, 3, 0, 32, 0, 174, 2, 191, 2, 231, 57, 0, 80, 142, 59, 0, 80, 22, 32, 0, 32, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 162, 231, 0, 32, 7, 0, 0, 0, 136, 6, 0, 0, 193, 231, 0, 0, 193, 231, 0, 0, 9, 3, 0, 0, 1, 0, 0, 0, 173, 158, 0, 0, 99, 97, 108, 101, 110, 100, 97, 114, 68, 97, 116, 97, 0, 98, 117, 100, 100, 104, 105, 115, 116, 0, 101, 114, 97, 115, 0, 115, 116, 97, 114, 116, 0, 115, 121, 115, 116, 101, 109, 0, 99, 104, 105, 110, 101, 115, 101, 0, 99, 111, 112, 116, 105, 99, 0, 101, 110, 100, 0, 100, 97, 110, 103, 105, 0, 101, 116, 104, 105, 111, 112, 105, 99, 0, 101, 116, 104, 105, 111, 112, 105, 99, 45, 97, 109, 101, 116, 101, 45, 97, 108, 101, 109, 0, 103, 114, 101, 103, 111, 114, 105, 97, 110, 0, 104, 101, 98, 114, 101, 119, 0, 105, 110, 100, 105, 97, 110, 0, 105, 115, 108, 97, 109, 105, 99, 0, 105, 115, 108, 97, 109, 105, 99, 45, 99, 105, 118, 105, 108, 0, 105, 115, 108, 97, 109, 105, 99, 45, 114, 103, 115, 97, 0, 105, 115, 108, 97, 109, 105, 99, 45, 116, 98, 108, 97, 0, 105, 115, 108, 97, 109, 105, 99, 45, 117, 109, 97, 108, 113, 117, 114, 97, 0, 106, 97, 112, 97, 110, 101, 115, 101, 0, 49, 48, 48, 0, 49, 48, 49, 0, 49, 48, 50, 0, 49, 48, 51, 0, 49, 48, 52, 0, 49, 48, 53, 0, 49, 48, 54, 0, 49, 48, 55, 0, 49, 48, 56, 0, 49, 48, 57, 0, 49, 49, 48, 0, 49, 49, 49, 0, 49, 49, 50, 0, 49, 49, 51, 0, 49, 49, 52, 0, 49, 49, 53, 0, 49, 49, 54, 0, 49, 49, 55, 0, 49, 49, 56, 0, 49, 49, 57, 0, 49, 50, 48, 0, 49, 50, 49, 0, 49, 50, 50, 0, 49, 50, 51, 0, 49, 50, 52, 0, 49, 50, 53, 0, 49, 50, 54, 0, 49, 50, 55, 0, 49, 50, 56, 0, 49, 50, 57, 0, 49, 51, 48, 0, 49, 51, 49, 0, 49, 51, 50, 0, 49, 51, 51, 0, 49, 51, 52, 0, 49, 51, 53, 0, 49, 51, 54, 0, 49, 51, 55, 0, 49, 51, 56, 0, 49, 51, 57, 0, 49, 52, 48, 0, 49, 52, 49, 0, 49, 52, 50, 0, 49, 52, 51, 0, 49, 52, 52, 0, 49, 52, 53, 0, 49, 52, 54, 0, 49, 52, 55, 0, 49, 52, 56, 0, 49, 52, 57, 0, 49, 53, 49, 0, 49, 53, 50, 0, 49, 53, 51, 0, 49, 53, 52, 0, 49, 53, 53, 0, 49, 53, 54, 0, 49, 53, 55, 0, 49, 53, 56, 0, 49, 53, 57, 0, 49, 54, 48, 0, 49, 54, 49, 0, 49, 54, 50, 0, 49, 54, 51, 0, 49, 54, 52, 0, 49, 54, 53, 0, 49, 54, 54, 0, 49, 54, 55, 0, 49, 54, 56, 0, 49, 54, 57, 0, 49, 55, 48, 0, 49, 55, 49, 0, 49, 55, 50, 0, 49, 55, 51, 0, 49, 55, 52, 0, 49, 55, 53, 0, 49, 55, 54, 0, 49, 55, 55, 0, 49, 55, 56, 0, 49, 55, 57, 0, 49, 56, 48, 0, 49, 56, 49, 0, 49, 56, 50, 0, 49, 56, 51, 0, 49, 56, 52, 0, 49, 56, 53, 0, 49, 56, 54, 0, 49, 56, 55, 0, 49, 56, 56, 0, 49, 56, 57, 0, 49, 57, 48, 0, 49, 57, 49, 0, 49, 57, 50, 0, 49, 57, 51, 0, 49, 57, 52, 0, 49, 57, 53, 0, 49, 57, 54, 0, 49, 57, 55, 0, 49, 57, 56, 0, 49, 57, 57, 0, 50, 48, 48, 0, 50, 48, 49, 0, 50, 48, 50, 0, 50, 48, 51, 0, 50, 48, 52, 0, 50, 48, 53, 0, 50, 48, 54, 0, 50, 48, 55, 0, 50, 48, 56, 0, 50, 48, 57, 0, 50, 49, 48, 0, 50, 49, 49, 0, 50, 49, 50, 0, 50, 49, 51, 0, 50, 49, 52, 0, 50, 49, 53, 0, 50, 49, 54, 0, 50, 49, 55, 0, 50, 49, 56, 0, 50, 49, 57, 0, 50, 50, 48, 0, 50, 50, 49, 0, 50, 50, 50, 0, 50, 50, 51, 0, 50, 50, 52, 0, 50, 50, 53, 0, 50, 50, 54, 0, 50, 50, 55, 0, 50, 50, 56, 0, 50, 50, 57, 0, 50, 51, 48, 0, 50, 51, 49, 0, 50, 51, 50, 0, 50, 51, 51, 0, 50, 51, 52, 0, 50, 51, 53, 0, 50, 51, 54, 0, 112, 101, 114, 115, 105, 97, 110, 0, 114, 111, 99, 0, 99, 97, 108, 101, 110, 100, 97, 114, 80, 114, 101, 102, 101, 114, 101, 110, 99, 101, 68, 97, 116, 97, 0, 65, 69, 0, 65, 70, 0, 66, 72, 0, 67, 78, 0, 67, 88, 0, 68, 74, 0, 68, 90, 0, 69, 71, 0, 69, 72, 0, 69, 82, 0, 69, 84, 0, 73, 76, 0, 73, 81, 0, 73, 82, 0, 74, 79, 0, 74, 80, 0, 75, 77, 0, 75, 82, 0, 75, 87, 0, 76, 66, 0, 76, 89, 0, 77, 65, 0, 77, 79, 0, 77, 82, 0, 79, 77, 0, 80, 83, 0, 81, 65, 0, 83, 65, 0, 83, 68, 0, 83, 71, 0, 83, 89, 0, 84, 68, 0, 84, 72, 0, 84, 78, 0, 89, 69, 0, 99, 108, 100, 114, 86, 101, 114, 115, 105, 111, 110, 0, 99, 111, 100, 101, 77, 97, 112, 112, 105, 110, 103, 115, 0, 99, 111, 100, 101, 77, 97, 112, 112, 105, 110, 103, 115, 67, 117, 114, 114, 101, 110, 99, 121, 0, 105, 100, 86, 97, 108, 105, 100, 105, 116, 121, 0, 99, 117, 114, 114, 101, 110, 99, 121, 0, 100, 101, 112, 114, 101, 99, 97, 116, 101, 100, 0, 114, 101, 103, 117, 108, 97, 114, 0, 117, 110, 107, 110, 111, 119, 110, 0, 108, 97, 110, 103, 117, 97, 103, 101, 0, 112, 114, 105, 118, 97, 116, 101, 95, 117, 115, 101, 0, 114, 101, 115, 101, 114, 118, 101, 100, 0, 115, 112, 101, 99, 105, 97, 108, 0, 109, 97, 99, 114, 111, 114, 101, 103, 105, 111, 110, 0, 115, 99, 114, 105, 112, 116, 0, 115, 117, 98, 100, 105, 118, 105, 115, 105, 111, 110, 0, 117, 110, 105, 116, 0, 108, 97, 110, 103, 117, 97, 103, 101, 68, 97, 116, 97, 0, 112, 114, 105, 109, 97, 114, 121, 0, 115, 99, 114, 105, 112, 116, 115, 0, 115, 101, 99, 111, 110, 100, 97, 114, 121, 0, 116, 101, 114, 114, 105, 116, 111, 114, 105, 101, 115, 0, 97, 98, 113, 0, 97, 98, 114, 0, 97, 99, 101, 0, 97, 99, 104, 0, 97, 100, 97, 0, 97, 100, 121, 0, 97, 101, 98, 0, 97, 103, 113, 0, 97, 105, 105, 0, 97, 105, 110, 0, 97, 107, 107, 0, 97, 107, 122, 0, 97, 108, 101, 0, 97, 108, 110, 0, 97, 108, 116, 0, 97, 109, 111, 0, 97, 110, 103, 0, 97, 110, 112, 0, 97, 111, 122, 0, 97, 114, 99, 0, 97, 114, 110, 0, 97, 114, 111, 0, 97, 114, 112, 0, 97, 114, 113, 0, 97, 114, 115, 0, 97, 114, 119, 0, 97, 114, 122, 0, 97, 115, 97, 0, 97, 115, 116, 0, 97, 116, 106, 0, 97, 118, 107, 0, 97, 119, 97, 0, 98, 97, 108, 0, 98, 97, 110, 0, 98, 97, 112, 0, 98, 97, 114, 0, 98, 97, 115, 0, 98, 97, 120, 0, 98, 98, 99, 0, 98, 98, 106, 0, 98, 99, 105, 0, 98, 101, 106, 0, 98, 101, 109, 0, 98, 101, 119, 0, 98, 101, 122, 0, 98, 102, 100, 0, 98, 102, 113, 0, 98, 102, 116, 0, 98, 102, 121, 0, 98, 103, 0, 98, 103, 99, 0, 98, 103, 110, 0, 98, 103, 120, 0, 98, 104, 98, 0, 98, 104, 105, 0, 98, 104, 107, 0, 98, 104, 111, 0, 98, 105, 0, 98, 105, 107, 0, 98, 105, 110, 0, 98, 106, 106, 0, 98, 106, 110, 0, 98, 106, 116, 0, 98, 107, 109, 0, 98, 107, 117, 0, 98, 108, 116, 0, 98, 109, 113, 0, 98, 110, 0, 98, 111, 0, 98, 112, 121, 0, 98, 113, 105, 0, 98, 113, 118, 0, 98, 114, 97, 0, 98, 114, 104, 0, 98, 114, 120, 0, 98, 115, 0, 98, 115, 99, 0, 98, 115, 115, 0, 98, 116, 111, 0, 98, 116, 118, 0, 98, 117, 97, 0, 98, 117, 99, 0, 98, 117, 103, 0, 98, 117, 109, 0, 98, 118, 98, 0, 98, 121, 110, 0, 98, 121, 118, 0, 98, 122, 101, 0, 98, 122, 120, 0, 99, 97, 100, 0, 99, 97, 114, 0, 99, 97, 121, 0, 99, 99, 104, 0, 99, 99, 112, 0, 99, 101, 98, 0, 99, 103, 103, 0, 99, 104, 107, 0, 99, 104, 109, 0, 99, 104, 110, 0, 99, 104, 111, 0, 99, 104, 112, 0, 99, 104, 114, 0, 99, 104, 121, 0, 99, 105, 99, 0, 99, 106, 97, 0, 99, 106, 109, 0, 99, 106, 115, 0, 99, 107, 98, 0, 99, 107, 116, 0, 99, 111, 112, 0, 99, 112, 115, 0, 99, 114, 104, 0, 99, 114, 106, 0, 99, 114, 107, 0, 99, 114, 108, 0, 99, 114, 109, 0, 99, 114, 115, 0, 99, 115, 98, 0, 99, 115, 119, 0, 99, 116, 100, 0, 99, 117, 0, 99, 118, 0, 100, 97, 107, 0, 100, 97, 114, 0, 100, 97, 118, 0, 100, 99, 99, 0, 100, 101, 108, 0, 100, 101, 110, 0, 100, 103, 114, 0, 100, 105, 110, 0, 100, 106, 101, 0, 100, 110, 103, 0, 100, 110, 106, 0, 100, 111, 105, 0, 100, 115, 98, 0, 100, 116, 109, 0, 100, 116, 112, 0, 100, 116, 121, 0, 100, 117, 97, 0, 100, 117, 109, 0, 100, 118, 0, 100, 121, 111, 0, 100, 121, 117, 0, 100, 122, 0, 101, 98, 117, 0, 101, 102, 105, 0, 101, 103, 108, 0, 101, 103, 121, 0, 101, 107, 97, 0, 101, 107, 121, 0, 101, 110, 109, 0, 101, 115, 117, 0, 101, 116, 116, 0, 101, 118, 110, 0, 101, 119, 111, 0, 101, 120, 116, 0, 102, 97, 110, 0, 102, 102, 0, 102, 102, 109, 0, 102, 105, 97, 0, 102, 105, 108, 0, 102, 105, 116, 0, 102, 106, 0, 102, 111, 110, 0, 102, 114, 99, 0, 102, 114, 109, 0, 102, 114, 111, 0, 102, 114, 112, 0, 102, 114, 114, 0, 102, 114, 115, 0, 102, 117, 100, 0, 102, 117, 113, 0, 102, 117, 114, 0, 102, 117, 118, 0, 102, 118, 114, 0, 103, 97, 97, 0, 103, 97, 103, 0, 103, 97, 110, 0, 103, 97, 121, 0, 103, 98, 97, 0, 103, 98, 109, 0, 103, 98, 122, 0, 103, 99, 114, 0, 103, 100, 0, 103, 101, 122, 0, 103, 105, 108, 0, 103, 106, 107, 0, 103, 106, 117, 0, 103, 108, 100, 0, 103, 108, 107, 0, 103, 109, 104, 0, 103, 111, 104, 0, 103, 111, 109, 0, 103, 111, 110, 0, 103, 111, 114, 0, 103, 111, 115, 0, 103, 111, 116, 0, 103, 114, 98, 0, 103, 114, 99, 0, 103, 114, 116, 0, 103, 115, 119, 0, 103, 117, 0, 103, 117, 98, 0, 103, 117, 99, 0, 103, 117, 114, 0, 103, 117, 122, 0, 103, 118, 0, 103, 118, 114, 0, 103, 119, 105, 0, 104, 97, 105, 0, 104, 97, 107, 0, 104, 97, 119, 0, 104, 97, 122, 0, 104, 105, 102, 0, 104, 105, 108, 0, 104, 105, 116, 0, 104, 109, 100, 0, 104, 109, 110, 0, 104, 110, 100, 0, 104, 110, 101, 0, 104, 110, 106, 0, 104, 110, 110, 0, 104, 110, 111, 0, 104, 111, 99, 0, 104, 111, 106, 0, 104, 111, 112, 0, 104, 115, 98, 0, 104, 115, 110, 0, 104, 117, 0, 104, 117, 112, 0, 104, 122, 0, 105, 98, 97, 0, 105, 98, 98, 0, 105, 102, 101, 0, 105, 107, 116, 0, 105, 108, 111, 0, 105, 110, 104, 0, 105, 122, 104, 0, 106, 97, 109, 0, 106, 103, 111, 0, 106, 109, 99, 0, 106, 109, 108, 0, 106, 112, 114, 0, 106, 114, 98, 0, 106, 117, 116, 0, 106, 118, 0, 107, 97, 97, 0, 107, 97, 98, 0, 107, 97, 99, 0, 107, 97, 106, 0, 107, 97, 109, 0, 107, 97, 111, 0, 107, 98, 100, 0, 107, 99, 97, 0, 107, 99, 103, 0, 107, 99, 107, 0, 107, 100, 101, 0, 107, 100, 116, 0, 107, 101, 97, 0, 107, 102, 111, 0, 107, 102, 114, 0, 107, 102, 121, 0, 107, 103, 0, 107, 103, 101, 0, 107, 103, 112, 0, 107, 104, 97, 0, 107, 104, 98, 0, 107, 104, 110, 0, 107, 104, 113, 0, 107, 104, 116, 0, 107, 104, 119, 0, 107, 105, 117, 0, 107, 106, 103, 0, 107, 106, 104, 0, 107, 107, 106, 0, 107, 108, 110, 0, 107, 109, 98, 0, 107, 110, 0, 107, 110, 102, 0, 107, 111, 105, 0, 107, 111, 107, 0, 107, 111, 115, 0, 107, 112, 101, 0, 107, 112, 121, 0, 107, 114, 99, 0, 107, 114, 105, 0, 107, 114, 106, 0, 107, 114, 108, 0, 107, 114, 117, 0, 107, 115, 0, 107, 115, 98, 0, 107, 115, 102, 0, 107, 115, 104, 0, 107, 117, 109, 0, 107, 117, 116, 0, 107, 118, 0, 107, 118, 114, 0, 107, 118, 120, 0, 107, 119, 0, 107, 120, 109, 0, 107, 120, 112, 0, 107, 121, 117, 0, 108, 97, 98, 0, 108, 97, 100, 0, 108, 97, 103, 0, 108, 97, 104, 0, 108, 97, 106, 0, 108, 97, 109, 0, 108, 98, 0, 108, 98, 101, 0, 108, 98, 119, 0, 108, 99, 112, 0, 108, 101, 112, 0, 108, 101, 122, 0, 108, 102, 110, 0, 108, 105, 102, 0, 108, 105, 106, 0, 108, 105, 115, 0, 108, 105, 118, 0, 108, 106, 112, 0, 108, 107, 105, 0, 108, 107, 116, 0, 108, 109, 110, 0, 108, 109, 111, 0, 108, 111, 108, 0, 108, 111, 122, 0, 108, 114, 99, 0, 108, 116, 103, 0, 108, 117, 0, 108, 117, 97, 0, 108, 117, 105, 0, 108, 117, 110, 0, 108, 117, 111, 0, 108, 117, 115, 0, 108, 117, 116, 0, 108, 117, 121, 0, 108, 117, 122, 0, 108, 118, 0, 108, 119, 108, 0, 108, 122, 104, 0, 108, 122, 122, 0, 109, 97, 100, 0, 109, 97, 102, 0, 109, 97, 103, 0, 109, 97, 105, 0, 109, 97, 107, 0, 109, 97, 110, 0, 109, 97, 115, 0, 109, 97, 122, 0, 109, 100, 102, 0, 109, 100, 104, 0, 109, 100, 114, 0, 109, 100, 116, 0, 109, 101, 110, 0, 109, 101, 114, 0, 109, 101, 121, 0, 109, 102, 97, 0, 109, 102, 101, 0, 109, 102, 118, 0, 109, 103, 104, 0, 109, 103, 111, 0, 109, 103, 112, 0, 109, 103, 121, 0, 109, 105, 110, 0, 109, 107, 0, 109, 108, 115, 0, 109, 110, 99, 0, 109, 110, 105, 0, 109, 110, 115, 0, 109, 110, 119, 0, 109, 111, 101, 0, 109, 111, 104, 0, 109, 111, 115, 0, 109, 114, 100, 0, 109, 114, 106, 0, 109, 114, 111, 0, 109, 116, 114, 0, 109, 117, 97, 0, 109, 117, 115, 0, 109, 118, 121, 0, 109, 119, 107, 0, 109, 119, 108, 0, 109, 119, 114, 0, 109, 119, 118, 0, 109, 120, 99, 0, 109, 121, 0, 109, 121, 118, 0, 109, 121, 120, 0, 109, 121, 122, 0, 109, 122, 110, 0, 110, 97, 110, 0, 110, 97, 112, 0, 110, 99, 104, 0, 110, 100, 99, 0, 110, 100, 115, 0, 110, 101, 119, 0, 110, 103, 108, 0, 110, 104, 101, 0, 110, 104, 119, 0, 110, 105, 97, 0, 110, 105, 106, 0, 110, 105, 117, 0, 110, 106, 111, 0, 110, 109, 103, 0, 110, 110, 104, 0, 110, 111, 100, 0, 110, 111, 101, 0, 110, 111, 103, 0, 110, 111, 110, 0, 110, 111, 118, 0, 110, 113, 111, 0, 110, 115, 107, 0, 110, 115, 111, 0, 110, 117, 115, 0, 110, 118, 0, 110, 120, 113, 0, 110, 121, 0, 110, 121, 109, 0, 110, 121, 110, 0, 110, 121, 111, 0, 110, 122, 105, 0, 111, 115, 97, 0, 111, 115, 99, 0, 111, 116, 107, 0, 112, 97, 0, 112, 97, 103, 0, 112, 97, 108, 0, 112, 97, 109, 0, 112, 97, 112, 0, 112, 97, 117, 0, 112, 99, 100, 0, 112, 99, 109, 0, 112, 100, 99, 0, 112, 100, 116, 0, 112, 101, 111, 0, 112, 102, 108, 0, 112, 104, 110, 0, 112, 107, 111, 0, 112, 108, 0, 112, 109, 115, 0, 112, 110, 116, 0, 112, 111, 110, 0, 112, 114, 100, 0, 112, 114, 103, 0, 112, 114, 111, 0, 112, 117, 117, 0, 113, 117, 0, 113, 117, 99, 0, 113, 117, 103, 0, 114, 97, 106, 0, 114, 97, 112, 0, 114, 97, 114, 0, 114, 99, 102, 0, 114, 101, 106, 0, 114, 103, 110, 0, 114, 105, 97, 0, 114, 105, 102, 0, 114, 106, 115, 0, 114, 107, 116, 0, 114, 109, 102, 0, 114, 109, 111, 0, 114, 109, 116, 0, 114, 109, 117, 0, 114, 110, 103, 0, 114, 111, 98, 0, 114, 111, 102, 0, 114, 111, 109, 0, 114, 116, 109, 0, 114, 117, 101, 0, 114, 117, 103, 0, 114, 117, 112, 0, 114, 119, 107, 0, 114, 121, 117, 0, 115, 97, 100, 0, 115, 97, 102, 0, 115, 97, 104, 0, 115, 97, 109, 0, 115, 97, 113, 0, 115, 97, 116, 0, 115, 97, 118, 0, 115, 97, 122, 0, 115, 98, 112, 0, 115, 99, 107, 0, 115, 99, 111, 0, 115, 99, 115, 0, 115, 100, 99, 0, 115, 100, 104, 0, 115, 101, 101, 0, 115, 101, 102, 0, 115, 101, 104, 0, 115, 101, 105, 0, 115, 101, 108, 0, 115, 101, 115, 0, 115, 103, 115, 0, 115, 104, 105, 0, 115, 104, 110, 0, 115, 105, 100, 0, 115, 107, 114, 0, 115, 108, 105, 0, 115, 108, 121, 0, 115, 109, 97, 0, 115, 109, 106, 0, 115, 109, 110, 0, 115, 109, 112, 0, 115, 109, 115, 0, 115, 110, 102, 0, 115, 110, 107, 0, 115, 111, 117, 0, 115, 113, 0, 115, 114, 98, 0, 115, 114, 110, 0, 115, 114, 114, 0, 115, 114, 120, 0, 115, 115, 121, 0, 115, 116, 113, 0, 115, 117, 107, 0, 115, 117, 115, 0, 115, 118, 0, 115, 119, 98, 0, 115, 119, 103, 0, 115, 119, 118, 0, 115, 120, 110, 0, 115, 121, 105, 0, 115, 121, 108, 0, 115, 121, 114, 0, 115, 122, 108, 0, 116, 97, 98, 0, 116, 97, 106, 0, 116, 98, 119, 0, 116, 99, 121, 0, 116, 100, 100, 0, 116, 100, 103, 0, 116, 100, 104, 0, 116, 101, 0, 116, 101, 111, 0, 116, 101, 114, 0, 116, 101, 116, 0, 116, 104, 0, 116, 104, 108, 0, 116, 104, 113, 0, 116, 104, 114, 0, 116, 105, 0, 116, 105, 103, 0, 116, 105, 118, 0, 116, 107, 108, 0, 116, 107, 114, 0, 116, 107, 116, 0, 116, 108, 105, 0, 116, 108, 121, 0, 116, 109, 104, 0, 116, 110, 114, 0, 116, 111, 103, 0, 116, 112, 105, 0, 116, 114, 117, 0, 116, 114, 118, 0, 116, 115, 100, 0, 116, 115, 103, 0, 116, 115, 105, 0, 116, 115, 106, 0, 116, 116, 106, 0, 116, 116, 115, 0, 116, 116, 116, 0, 116, 117, 109, 0, 116, 118, 108, 0, 116, 119, 113, 0, 116, 121, 118, 0, 116, 122, 109, 0, 117, 100, 101, 0, 117, 100, 109, 0, 117, 103, 97, 0, 117, 108, 105, 0, 117, 109, 98, 0, 117, 110, 100, 0, 117, 110, 114, 0, 117, 110, 120, 0, 118, 97, 105, 0, 118, 101, 0, 118, 101, 99, 0, 118, 101, 112, 0, 118, 105, 0, 118, 105, 99, 0, 118, 108, 115, 0, 118, 109, 102, 0, 118, 109, 119, 0, 118, 111, 116, 0, 118, 114, 111, 0, 118, 117, 110, 0, 119, 97, 101, 0, 119, 97, 114, 0, 119, 97, 115, 0, 119, 98, 112, 0, 119, 98, 113, 0, 119, 98, 114, 0, 119, 110, 105, 0, 119, 116, 109, 0, 119, 117, 117, 0, 120, 97, 108, 0, 120, 97, 118, 0, 120, 99, 114, 0, 120, 104, 0, 120, 108, 99, 0, 120, 108, 100, 0, 120, 109, 102, 0, 120, 109, 110, 0, 120, 109, 114, 0, 120, 110, 97, 0, 120, 110, 114, 0, 120, 111, 103, 0, 120, 112, 114, 0, 120, 115, 97, 0, 120, 115, 114, 0, 120, 117, 109, 0, 121, 97, 111, 0, 121, 97, 112, 0, 121, 97, 118, 0, 121, 98, 98, 0, 121, 114, 107, 0, 121, 117, 97, 0, 121, 117, 101, 0, 122, 97, 103, 0, 122, 97, 112, 0, 122, 100, 106, 0, 122, 101, 97, 0, 122, 101, 110, 0, 122, 103, 104, 0, 122, 109, 105, 0, 122, 117, 0, 122, 117, 110, 0, 122, 122, 97, 0, 108, 97, 110, 103, 117, 97, 103, 101, 77, 97, 116, 99, 104, 105, 110, 103, 0, 119, 114, 105, 116, 116, 101, 110, 0, 108, 97, 110, 103, 117, 97, 103, 101, 77, 97, 116, 99, 104, 105, 110, 103, 73, 110, 102, 111, 0, 109, 97, 116, 99, 104, 86, 97, 114, 105, 97, 98, 108, 101, 0, 97, 109, 101, 114, 105, 99, 97, 115, 0, 99, 110, 115, 97, 114, 0, 101, 110, 85, 83, 0, 109, 97, 103, 104, 114, 101, 98, 0, 112, 97, 114, 97, 100, 105, 103, 109, 76, 111, 99, 97, 108, 101, 115, 0, 108, 97, 110, 103, 117, 97, 103, 101, 77, 97, 116, 99, 104, 105, 110, 103, 78, 101, 119, 0, 109, 101, 97, 115, 117, 114, 101, 109, 101, 110, 116, 68, 97, 116, 97, 0, 77, 101, 97, 115, 117, 114, 101, 109, 101, 110, 116, 83, 121, 115, 116, 101, 109, 0, 80, 97, 112, 101, 114, 83, 105, 122, 101, 0, 66, 83, 0, 77, 101, 97, 115, 117, 114, 101, 109, 101, 110, 116, 83, 121, 115, 116, 101, 109, 67, 97, 116, 101, 103, 111, 114, 121, 0, 116, 101, 109, 112, 101, 114, 97, 116, 117, 114, 101, 0, 66, 90, 0, 67, 76, 0, 67, 79, 0, 67, 82, 0, 71, 66, 0, 71, 84, 0, 75, 89, 0, 76, 82, 0, 77, 77, 0, 77, 88, 0, 78, 73, 0, 80, 65, 0, 80, 72, 0, 80, 82, 0, 80, 87, 0, 83, 86, 0, 86, 69, 0, 112, 97, 114, 101, 110, 116, 76, 111, 99, 97, 108, 101, 115, 0, 101, 110, 95, 48, 48, 49, 0, 101, 110, 95, 49, 53, 48, 0, 101, 115, 95, 52, 49, 57, 0, 112, 116, 95, 80, 84, 0, 114, 111, 111, 116, 0, 122, 104, 95, 72, 97, 110, 116, 95, 72, 75, 0, 115, 117, 98, 100, 105, 118, 105, 115, 105, 111, 110, 67, 111, 110, 116, 97, 105, 110, 109, 101, 110, 116, 0, 65, 68, 0, 65, 71, 0, 65, 76, 0, 65, 77, 0, 65, 79, 0, 65, 82, 0, 65, 84, 0, 65, 85, 0, 65, 90, 0, 66, 65, 0, 66, 66, 0, 66, 68, 0, 66, 69, 0, 66, 70, 0, 66, 71, 0, 66, 73, 0, 66, 74, 0, 66, 78, 0, 66, 81, 0, 66, 84, 0, 66, 87, 0, 66, 89, 0, 67, 68, 0, 67, 70, 0, 67, 71, 0, 67, 73, 0, 67, 77, 0, 67, 85, 0, 67, 86, 0, 67, 89, 0, 67, 90, 0, 68, 69, 0, 68, 75, 0, 68, 77, 0, 68, 79, 0, 69, 69, 0, 70, 73, 0, 70, 74, 0, 70, 77, 0, 70, 82, 0, 71, 65, 0, 71, 68, 0, 71, 69, 0, 71, 72, 0, 71, 76, 0, 71, 77, 0, 71, 78, 0, 71, 82, 0, 71, 87, 0, 71, 89, 0, 72, 78, 0, 72, 82, 0, 72, 84, 0, 72, 85, 0, 73, 68, 0, 73, 69, 0, 73, 83, 0, 74, 77, 0, 75, 69, 0, 75, 71, 0, 75, 72, 0, 75, 73, 0, 75, 78, 0, 75, 80, 0, 75, 90, 0, 76, 65, 0, 76, 67, 0, 76, 73, 0, 76, 75, 0, 76, 83, 0, 76, 84, 0, 76, 85, 0, 76, 86, 0, 77, 67, 0, 77, 68, 0, 77, 69, 0, 77, 71, 0, 77, 72, 0, 77, 75, 0, 77, 76, 0, 77, 78, 0, 77, 84, 0, 77, 85, 0, 77, 86, 0, 77, 87, 0, 77, 89, 0, 77, 90, 0, 78, 65, 0, 78, 69, 0, 78, 71, 0, 78, 76, 0, 78, 79, 0, 78, 80, 0, 78, 82, 0, 78, 90, 0, 80, 71, 0, 80, 75, 0, 80, 76, 0, 80, 89, 0, 82, 79, 0, 82, 83, 0, 82, 85, 0, 82, 87, 0, 83, 66, 0, 83, 67, 0, 83, 69, 0, 83, 72, 0, 83, 73, 0, 83, 75, 0, 83, 76, 0, 83, 77, 0, 83, 78, 0, 83, 79, 0, 83, 82, 0, 83, 83, 0, 83, 84, 0, 83, 90, 0, 84, 71, 0, 84, 74, 0, 84, 76, 0, 84, 77, 0, 84, 79, 0, 84, 82, 0, 84, 84, 0, 84, 86, 0, 84, 90, 0, 85, 65, 0, 85, 71, 0, 85, 77, 0, 85, 89, 0, 85, 90, 0, 86, 67, 0, 86, 78, 0, 86, 85, 0, 87, 70, 0, 87, 83, 0, 90, 77, 0, 90, 87, 0, 97, 122, 110, 120, 0, 98, 100, 97, 0, 98, 100, 98, 0, 98, 100, 99, 0, 98, 100, 100, 0, 98, 100, 101, 0, 98, 100, 102, 0, 98, 100, 103, 0, 98, 100, 104, 0, 98, 101, 118, 108, 103, 0, 98, 101, 119, 97, 108, 0, 98, 102, 48, 49, 0, 98, 102, 48, 50, 0, 98, 102, 48, 51, 0, 98, 102, 48, 52, 0, 98, 102, 48, 53, 0, 98, 102, 48, 54, 0, 98, 102, 48, 55, 0, 98, 102, 48, 56, 0, 98, 102, 48, 57, 0, 98, 102, 49, 48, 0, 98, 102, 49, 49, 0, 98, 102, 49, 50, 0, 98, 102, 49, 51, 0, 99, 118, 98, 0, 99, 118, 115, 0, 99, 122, 50, 48, 0, 99, 122, 51, 49, 0, 99, 122, 51, 50, 0, 99, 122, 52, 49, 0, 99, 122, 52, 50, 0, 99, 122, 53, 49, 0, 99, 122, 53, 50, 0, 99, 122, 53, 51, 0, 99, 122, 54, 51, 0, 99, 122, 54, 52, 0, 99, 122, 55, 49, 0, 99, 122, 55, 50, 0, 99, 122, 56, 48, 0, 100, 111, 51, 51, 0, 100, 111, 51, 52, 0, 100, 111, 51, 53, 0, 100, 111, 51, 54, 0, 100, 111, 51, 55, 0, 100, 111, 51, 56, 0, 100, 111, 51, 57, 0, 100, 111, 52, 48, 0, 100, 111, 52, 49, 0, 100, 111, 52, 50, 0, 101, 115, 97, 110, 0, 101, 115, 97, 114, 0, 101, 115, 97, 115, 0, 101, 115, 99, 98, 0, 101, 115, 99, 108, 0, 101, 115, 99, 109, 0, 101, 115, 99, 110, 0, 101, 115, 99, 116, 0, 101, 115, 101, 120, 0, 101, 115, 103, 97, 0, 101, 115, 105, 98, 0, 101, 115, 109, 99, 0, 101, 115, 109, 100, 0, 101, 115, 110, 99, 0, 101, 115, 112, 118, 0, 101, 115, 114, 105, 0, 101, 115, 118, 99, 0, 102, 106, 99, 0, 102, 106, 101, 0, 102, 106, 110, 0, 102, 106, 119, 0, 102, 114, 97, 114, 97, 0, 102, 114, 98, 102, 99, 0, 102, 114, 98, 114, 101, 0, 102, 114, 99, 111, 114, 0, 102, 114, 99, 118, 108, 0, 102, 114, 103, 101, 115, 0, 102, 114, 103, 117, 97, 0, 102, 114, 104, 100, 102, 0, 102, 114, 105, 100, 102, 0, 102, 114, 108, 114, 101, 0, 102, 114, 109, 97, 121, 0, 102, 114, 110, 97, 113, 0, 102, 114, 110, 111, 114, 0, 102, 114, 111, 99, 99, 0, 102, 114, 112, 97, 99, 0, 102, 114, 112, 100, 108, 0, 103, 98, 101, 110, 103, 0, 103, 98, 110, 105, 114, 0, 103, 98, 115, 99, 116, 0, 103, 98, 119, 108, 115, 0, 103, 110, 98, 0, 103, 110, 100, 0, 103, 110, 102, 0, 103, 110, 107, 0, 103, 110, 108, 0, 103, 110, 109, 0, 103, 110, 110, 0, 103, 113, 99, 0, 103, 113, 105, 0, 103, 119, 108, 0, 103, 119, 110, 0, 103, 119, 115, 0, 105, 100, 106, 119, 0, 105, 100, 107, 97, 0, 105, 100, 109, 108, 0, 105, 100, 110, 117, 0, 105, 100, 112, 112, 0, 105, 100, 115, 108, 0, 105, 100, 115, 109, 0, 105, 101, 99, 0, 105, 101, 108, 0, 105, 101, 109, 0, 105, 101, 117, 0, 105, 116, 50, 49, 0, 105, 116, 50, 51, 0, 105, 116, 50, 53, 0, 105, 116, 51, 50, 0, 105, 116, 51, 52, 0, 105, 116, 52, 50, 0, 105, 116, 52, 53, 0, 105, 116, 53, 50, 0, 105, 116, 53, 53, 0, 105, 116, 53, 55, 0, 105, 116, 54, 50, 0, 105, 116, 54, 53, 0, 105, 116, 54, 55, 0, 105, 116, 55, 50, 0, 105, 116, 55, 53, 0, 105, 116, 55, 55, 0, 105, 116, 55, 56, 0, 105, 116, 56, 50, 0, 105, 116, 56, 56, 0, 107, 110, 107, 0, 107, 110, 110, 0, 108, 107, 49, 0, 108, 107, 50, 0, 108, 107, 51, 0, 108, 107, 52, 0, 108, 107, 53, 0, 108, 107, 54, 0, 108, 107, 55, 0, 108, 107, 56, 0, 108, 107, 57, 0, 109, 97, 48, 49, 0, 109, 97, 48, 50, 0, 109, 97, 48, 51, 0, 109, 97, 48, 52, 0, 109, 97, 48, 53, 0, 109, 97, 48, 54, 0, 109, 97, 48, 55, 0, 109, 97, 48, 56, 0, 109, 97, 48, 57, 0, 109, 97, 49, 48, 0, 109, 97, 49, 49, 0, 109, 97, 49, 50, 0, 109, 104, 108, 0, 109, 104, 116, 0, 109, 119, 99, 0, 109, 119, 110, 0, 109, 119, 115, 0, 110, 112, 49, 0, 110, 112, 50, 0, 110, 112, 51, 0, 110, 112, 52, 0, 110, 112, 53, 0, 112, 104, 48, 49, 0, 112, 104, 48, 50, 0, 112, 104, 48, 51, 0, 112, 104, 48, 53, 0, 112, 104, 48, 54, 0, 112, 104, 48, 55, 0, 112, 104, 48, 56, 0, 112, 104, 48, 57, 0, 112, 104, 49, 48, 0, 112, 104, 49, 49, 0, 112, 104, 49, 50, 0, 112, 104, 49, 51, 0, 112, 104, 49, 52, 0, 112, 104, 49, 53, 0, 112, 104, 52, 48, 0, 112, 104, 52, 49, 0, 114, 115, 107, 109, 0, 114, 115, 118, 111, 0, 117, 103, 99, 0, 117, 103, 101, 0, 117, 103, 110, 0, 117, 103, 119, 0, 116, 101, 114, 114, 105, 116, 111, 114, 121, 67, 111, 110, 116, 97, 105, 110, 109, 101, 110, 116, 0, 48, 48, 50, 0, 48, 48, 53, 0, 48, 48, 57, 0, 48, 49, 49, 0, 48, 49, 51, 0, 48, 49, 52, 0, 48, 49, 53, 0, 48, 49, 55, 0, 48, 49, 56, 0, 48, 49, 57, 0, 48, 50, 49, 0, 48, 50, 57, 0, 48, 51, 48, 0, 48, 51, 52, 0, 48, 51, 53, 0, 48, 51, 57, 0, 48, 53, 51, 0, 48, 53, 52, 0, 48, 53, 55, 0, 48, 54, 49, 0, 81, 79, 0, 99, 111, 110, 116, 97, 105, 110, 101, 100, 71, 114, 111, 117, 112, 105, 110, 103, 115, 0, 103, 114, 111, 117, 112, 105, 110, 103, 0, 48, 48, 51, 0, 69, 85, 0, 69, 90, 0, 85, 78, 0, 116, 101, 114, 114, 105, 116, 111, 114, 121, 73, 110, 102, 111, 0, 65, 67, 0, 112, 111, 112, 117, 108, 97, 116, 105, 111, 110, 83, 104, 97, 114, 101, 70, 0, 116, 101, 114, 114, 105, 116, 111, 114, 121, 70, 0, 111, 102, 102, 105, 99, 105, 97, 108, 83, 116, 97, 116, 117, 115, 0, 119, 114, 105, 116, 105, 110, 103, 83, 104, 97, 114, 101, 70, 0, 107, 107, 95, 65, 114, 97, 98, 0, 117, 122, 95, 65, 114, 97, 98, 0, 65, 73, 0, 108, 105, 116, 101, 114, 97, 99, 121, 83, 104, 97, 114, 101, 70, 0, 65, 81, 0, 65, 83, 0, 122, 104, 95, 72, 97, 110, 116, 0, 65, 87, 0, 65, 88, 0, 97, 122, 95, 67, 121, 114, 108, 0, 98, 115, 95, 67, 121, 114, 108, 0, 115, 114, 95, 76, 97, 116, 110, 0, 102, 102, 95, 65, 100, 108, 109, 0, 66, 76, 0, 66, 77, 0, 109, 115, 95, 65, 114, 97, 98, 0, 66, 86, 0, 105, 117, 95, 76, 97, 116, 110, 0, 67, 67, 0, 67, 75, 0, 104, 97, 95, 65, 114, 97, 98, 0, 107, 121, 95, 65, 114, 97, 98, 0, 109, 110, 95, 77, 111, 110, 103, 0, 117, 122, 95, 67, 121, 114, 108, 0, 121, 117, 101, 95, 72, 97, 110, 115, 0, 67, 80, 0, 67, 87, 0, 68, 71, 0, 69, 65, 0, 70, 75, 0, 70, 79, 0, 71, 70, 0, 71, 71, 0, 71, 73, 0, 109, 97, 110, 95, 78, 107, 111, 111, 0, 71, 80, 0, 71, 83, 0, 71, 85, 0, 72, 77, 0, 73, 67, 0, 73, 77, 0, 115, 100, 95, 68, 101, 118, 97, 0, 73, 79, 0, 97, 122, 95, 65, 114, 97, 98, 0, 74, 69, 0, 117, 103, 95, 67, 121, 114, 108, 0, 107, 117, 95, 65, 114, 97, 98, 0, 118, 97, 105, 95, 76, 97, 116, 110, 0, 114, 105, 102, 95, 76, 97, 116, 110, 0, 115, 104, 105, 95, 76, 97, 116, 110, 0, 77, 70, 0, 98, 109, 95, 78, 107, 111, 111, 0, 77, 80, 0, 77, 81, 0, 77, 83, 0, 78, 67, 0, 78, 70, 0, 117, 110, 114, 95, 68, 101, 118, 97, 0, 78, 85, 0, 80, 70, 0, 112, 97, 95, 65, 114, 97, 98, 0, 116, 103, 95, 65, 114, 97, 98, 0, 80, 77, 0, 80, 78, 0, 82, 69, 0, 83, 74, 0, 83, 88, 0, 84, 65, 0, 84, 67, 0, 84, 70, 0, 84, 75, 0, 107, 121, 95, 76, 97, 116, 110, 0, 86, 65, 0, 86, 71, 0, 86, 73, 0, 88, 75, 0, 89, 84, 0, 90, 90, 0, 116, 105, 109, 101, 68, 97, 116, 97, 0, 97, 108, 108, 111, 119, 101, 100, 0, 112, 114, 101, 102, 101, 114, 114, 101, 100, 0, 97, 102, 95, 90, 65, 0, 97, 114, 95, 48, 48, 49, 0, 99, 97, 95, 69, 83, 0, 101, 115, 95, 66, 79, 0, 101, 115, 95, 66, 82, 0, 101, 115, 95, 69, 67, 0, 101, 115, 95, 69, 83, 0, 101, 115, 95, 71, 81, 0, 101, 115, 95, 80, 69, 0, 102, 114, 95, 67, 65, 0, 103, 108, 95, 69, 83, 0, 103, 117, 95, 73, 78, 0, 104, 105, 95, 73, 78, 0, 105, 116, 95, 67, 72, 0, 105, 116, 95, 73, 84, 0, 107, 110, 95, 73, 78, 0, 109, 108, 95, 73, 78, 0, 109, 114, 95, 73, 78, 0, 112, 97, 95, 73, 78, 0, 116, 97, 95, 73, 78, 0, 116, 101, 95, 73, 78, 0, 122, 117, 95, 90, 65, 0, 117, 110, 105, 116, 80, 114, 101, 102, 101, 114, 101, 110, 99, 101, 68, 97, 116, 97, 0, 97, 114, 101, 97, 45, 108, 97, 110, 100, 45, 97, 103, 114, 105, 99, 117, 108, 116, 0, 97, 114, 101, 97, 45, 108, 97, 110, 100, 45, 99, 111, 109, 109, 101, 114, 99, 108, 0, 97, 114, 101, 97, 45, 108, 97, 110, 100, 45, 114, 101, 115, 105, 100, 110, 116, 108, 0, 99, 111, 110, 99, 101, 110, 116, 114, 45, 98, 108, 111, 111, 100, 45, 103, 108, 117, 99, 111, 115, 101, 0, 99, 111, 110, 115, 117, 109, 112, 116, 105, 111, 110, 45, 118, 101, 104, 105, 99, 108, 101, 45, 102, 117, 101, 108, 0, 100, 117, 114, 97, 116, 105, 111, 110, 45, 109, 117, 115, 105, 99, 45, 116, 114, 97, 99, 107, 0, 100, 117, 114, 97, 116, 105, 111, 110, 45, 112, 101, 114, 115, 111, 110, 45, 97, 103, 101, 0, 100, 117, 114, 97, 116, 105, 111, 110, 45, 116, 118, 45, 112, 114, 111, 103, 114, 97, 109, 0, 101, 110, 101, 114, 103, 121, 45, 102, 111, 111, 100, 0, 101, 110, 101, 114, 103, 121, 45, 112, 101, 114, 115, 111, 110, 45, 117, 115, 97, 103, 101, 0, 108, 101, 110, 103, 116, 104, 45, 112, 101, 114, 115, 111, 110, 0, 108, 101, 110, 103, 116, 104, 45, 112, 101, 114, 115, 111, 110, 45, 115, 109, 97, 108, 108, 0, 108, 101, 110, 103, 116, 104, 45, 114, 97, 105, 110, 102, 97, 108, 108, 0, 108, 101, 110, 103, 116, 104, 45, 114, 111, 97, 100, 0, 108, 101, 110, 103, 116, 104, 45, 114, 111, 97, 100, 45, 115, 109, 97, 108, 108, 0, 108, 101, 110, 103, 116, 104, 45, 115, 110, 111, 119, 102, 97, 108, 108, 0, 108, 101, 110, 103, 116, 104, 45, 118, 101, 104, 105, 99, 108, 101, 0, 108, 101, 110, 103, 116, 104, 45, 118, 105, 115, 105, 98, 108, 116, 121, 0, 108, 101, 110, 103, 116, 104, 45, 118, 105, 115, 105, 98, 108, 116, 121, 45, 115, 109, 97, 108, 108, 0, 109, 97, 115, 115, 45, 112, 101, 114, 115, 111, 110, 0, 109, 97, 115, 115, 45, 112, 101, 114, 115, 111, 110, 45, 115, 109, 97, 108, 108, 0, 112, 114, 101, 115, 115, 117, 114, 101, 45, 98, 97, 114, 111, 109, 116, 114, 99, 0, 115, 112, 101, 101, 100, 45, 114, 111, 97, 100, 45, 116, 114, 97, 118, 101, 108, 0, 115, 112, 101, 101, 100, 45, 119, 105, 110, 100, 0, 116, 101, 109, 112, 101, 114, 97, 116, 117, 114, 101, 45, 112, 101, 114, 115, 111, 110, 0, 116, 101, 109, 112, 101, 114, 97, 116, 117, 114, 101, 45, 119, 101, 97, 116, 104, 101, 114, 0, 118, 111, 108, 117, 109, 101, 45, 118, 101, 104, 105, 99, 108, 101, 45, 102, 117, 101, 108, 0, 108, 101, 110, 103, 116, 104, 45, 112, 101, 114, 115, 111, 110, 45, 105, 110, 102, 111, 114, 109, 97, 108, 0, 99, 111, 110, 115, 117, 109, 112, 116, 105, 111, 110, 45, 118, 101, 104, 105, 99, 108, 101, 45, 102, 117, 101, 108, 45, 105, 110, 102, 111, 114, 109, 97, 108, 0, 108, 101, 110, 103, 116, 104, 45, 112, 101, 114, 115, 111, 110, 45, 115, 109, 97, 108, 108, 45, 105, 110, 102, 111, 114, 109, 97, 108, 0, 108, 101, 110, 103, 116, 104, 45, 114, 111, 97, 100, 45, 105, 110, 102, 111, 114, 109, 97, 108, 0, 119, 101, 101, 107, 68, 97, 116, 97, 0, 65, 78, 0, 119, 101, 101, 107, 68, 97, 116, 97, 37, 118, 97, 114, 105, 97, 110, 116, 0, 119, 101, 101, 107, 79, 102, 80, 114, 101, 102, 101, 114, 101, 110, 99, 101, 0, 105, 119, 0, 122, 104, 95, 84, 87, 0, 170, 0, 0, 104, 0, 66, 0, 0, 0, 73, 0, 68, 0, 0, 0, 69, 0, 84, 0, 0, 0, 68, 0, 90, 0, 0, 0, 102, 0, 102, 0, 0, 0, 67, 0, 90, 0, 0, 0, 84, 0, 82, 0, 0, 0, 66, 0, 71, 0, 0, 0, 67, 0, 73, 0, 0, 0, 68, 0, 74, 0, 0, 0, 72, 0, 85, 0, 0, 0, 73, 0, 81, 0, 0, 0, 75, 0, 90, 0, 0, 0, 77, 0, 76, 0, 0, 0, 66, 0, 70, 0, 0, 0, 66, 0, 73, 0, 0, 0, 69, 0, 71, 0, 0, 0, 71, 0, 69, 0, 0, 0, 78, 0, 79, 0, 0, 0, 83, 0, 89, 0, 0, 0, 88, 0, 75, 0, 0, 0, 89, 0, 69, 0, 0, 0, 65, 0, 69, 0, 0, 0, 66, 0, 74, 0, 0, 0, 72, 0, 84, 0, 0, 0, 74, 0, 79, 0, 0, 0, 75, 0, 87, 0, 0, 0, 76, 0, 89, 0, 0, 0, 77, 0, 72, 0, 0, 0, 81, 0, 65, 0, 0, 0, 84, 0, 74, 0, 0, 0, 85, 0, 90, 0, 0, 0, 65, 0, 83, 0, 0, 0, 65, 0, 87, 0, 0, 0, 66, 0, 72, 0, 0, 0, 66, 0, 81, 0, 0, 0, 66, 0, 84, 0, 0, 0, 67, 0, 87, 0, 0, 0, 71, 0, 85, 0, 0, 0, 73, 0, 83, 0, 0, 0, 74, 0, 80, 0, 0, 0, 75, 0, 72, 0, 0, 0, 77, 0, 67, 0, 0, 0, 77, 0, 69, 0, 0, 0, 77, 0, 86, 0, 0, 0, 78, 0, 67, 0, 0, 0, 65, 0, 81, 0, 0, 0, 65, 0, 88, 0, 0, 0, 66, 0, 76, 0, 0, 0, 66, 0, 85, 0, 0, 0, 66, 0, 86, 0, 0, 0, 69, 0, 65, 0, 0, 0, 70, 0, 88, 0, 0, 0, 71, 0, 76, 0, 0, 0, 77, 0, 81, 0, 0, 0, 80, 0, 83, 0, 0, 0, 81, 0, 79, 0, 0, 0, 81, 0, 85, 0, 0, 0, 84, 0, 87, 0, 0, 0, 90, 0, 82, 0, 0, 0, 67, 0, 80, 0, 0, 0, 69, 0, 90, 0, 0, 0, 81, 0, 77, 0, 0, 0, 81, 0, 78, 0, 0, 0, 81, 0, 80, 0, 0, 0, 81, 0, 82, 0, 0, 0, 81, 0, 83, 0, 0, 0, 81, 0, 84, 0, 0, 0, 81, 0, 86, 0, 0, 0, 81, 0, 87, 0, 0, 0, 81, 0, 88, 0, 0, 0, 81, 0, 89, 0, 0, 0, 81, 0, 90, 0, 0, 0, 88, 0, 65, 0, 0, 0, 88, 0, 66, 0, 0, 0, 88, 0, 67, 0, 0, 0, 88, 0, 68, 0, 0, 0, 88, 0, 69, 0, 0, 0, 88, 0, 70, 0, 0, 0, 88, 0, 71, 0, 0, 0, 88, 0, 72, 0, 0, 0, 88, 0, 73, 0, 0, 0, 88, 0, 74, 0, 0, 0, 88, 0, 76, 0, 0, 0, 88, 0, 79, 0, 0, 0, 88, 0, 81, 0, 0, 0, 88, 0, 83, 0, 0, 0, 88, 0, 84, 0, 0, 0, 88, 0, 85, 0, 0, 0, 88, 0, 87, 0, 0, 0, 88, 0, 89, 0, 0, 0, 88, 0, 90, 0, 0, 0, 104, 0, 122, 0, 0, 0, 107, 0, 106, 0, 0, 0, 122, 0, 104, 0, 98, 0, 0, 0, 122, 0, 101, 0, 110, 0, 0, 0, 105, 0, 115, 0, 48, 0, 0, 0, 78, 0, 82, 0, 85, 0, 0, 0, 85, 0, 83, 0, 78, 0, 0, 0, 80, 0, 67, 0, 78, 0, 0, 0, 86, 0, 73, 0, 82, 0, 0, 0, 108, 0, 117, 0, 121, 0, 0, 0, 120, 0, 122, 0, 104, 0, 0, 0, 65, 0, 82, 0, 83, 0, 0, 0, 68, 0, 77, 0, 65, 0, 0, 0, 77, 0, 82, 0, 79, 0, 0, 0, 69, 0, 84, 0, 72, 0, 0, 0, 73, 0, 83, 0, 75, 0, 0, 0, 88, 0, 67, 0, 68, 0, 0, 0, 78, 0, 80, 0, 76, 0, 0, 0, 88, 0, 69, 0, 69, 0, 0, 0, 75, 0, 72, 0, 82, 0, 0, 0, 107, 0, 108, 0, 110, 0, 0, 0, 83, 0, 73, 0, 84, 0, 0, 0, 90, 0, 65, 0, 70, 0, 0, 0, 122, 0, 104, 0, 105, 0, 0, 0, 78, 0, 71, 0, 78, 0, 0, 0, 88, 0, 85, 0, 65, 0, 0, 0, 77, 0, 76, 0, 84, 0, 0, 0, 77, 0, 78, 0, 69, 0, 0, 0, 83, 0, 76, 0, 86, 0, 0, 0, 82, 0, 87, 0, 70, 0, 0, 0, 83, 0, 84, 0, 78, 0, 0, 0, 66, 0, 71, 0, 82, 0, 0, 0, 70, 0, 76, 0, 75, 0, 0, 0, 84, 0, 75, 0, 77, 0, 0, 0, 88, 0, 66, 0, 65, 0, 0, 0, 88, 0, 66, 0, 68, 0, 0, 0, 121, 0, 100, 0, 97, 0, 0, 0, 70, 0, 75, 0, 80, 0, 0, 0, 72, 0, 84, 0, 71, 0, 0, 0, 78, 0, 65, 0, 68, 0, 0, 0, 84, 0, 84, 0, 68, 0, 0, 0, 77, 0, 76, 0, 73, 0, 0, 0, 77, 0, 78, 0, 80, 0, 0, 0, 77, 0, 89, 0, 84, 0, 0, 0, 83, 0, 86, 0, 78, 0, 0, 0, 122, 0, 111, 0, 109, 0, 0, 0, 88, 0, 80, 0, 70, 0, 0, 0, 72, 0, 75, 0, 71, 0, 0, 0, 75, 0, 65, 0, 90, 0, 0, 0, 76, 0, 66, 0, 89, 0, 0, 0, 76, 0, 83, 0, 79, 0, 0, 0, 76, 0, 86, 0, 65, 0, 0, 0, 77, 0, 67, 0, 70, 0, 0, 0, 77, 0, 83, 0, 82, 0, 0, 0, 83, 0, 76, 0, 66, 0, 0, 0, 83, 0, 79, 0, 77, 0, 0, 0, 85, 0, 75, 0, 82, 0, 0, 0, 88, 0, 77, 0, 77, 0, 0, 0, 88, 0, 82, 0, 69, 0, 0, 0, 90, 0, 77, 0, 75, 0, 0, 0, 75, 0, 77, 0, 70, 0, 0, 0, 76, 0, 89, 0, 68, 0, 0, 0, 77, 0, 78, 0, 84, 0, 0, 0, 85, 0, 89, 0, 85, 0, 0, 0, 88, 0, 83, 0, 85, 0, 0, 0, 113, 0, 120, 0, 104, 0, 0, 0, 65, 0, 76, 0, 65, 0, 0, 0, 71, 0, 84, 0, 77, 0, 0, 0, 71, 0, 85, 0, 77, 0, 0, 0, 73, 0, 83, 0, 74, 0, 0, 0, 76, 0, 66, 0, 78, 0, 0, 0, 77, 0, 71, 0, 70, 0, 0, 0, 78, 0, 65, 0, 77, 0, 0, 0, 79, 0, 77, 0, 78, 0, 0, 0, 83, 0, 67, 0, 71, 0, 0, 0, 83, 0, 77, 0, 82, 0, 0, 0, 85, 0, 71, 0, 65, 0, 0, 0, 85, 0, 83, 0, 65, 0, 0, 0, 87, 0, 83, 0, 77, 0, 0, 0, 88, 0, 70, 0, 79, 0, 0, 0, 89, 0, 77, 0, 68, 0, 0, 0, 90, 0, 65, 0, 76, 0, 0, 0, 119, 0, 121, 0, 105, 0, 0, 0, 88, 0, 88, 0, 88, 0, 0, 0, 78, 0, 73, 0, 67, 0, 0, 0, 65, 0, 84, 0, 70, 0, 0, 0, 69, 0, 67, 0, 83, 0, 0, 0, 73, 0, 84, 0, 65, 0, 0, 0, 75, 0, 72, 0, 77, 0, 0, 0, 80, 0, 65, 0, 78, 0, 0, 0, 83, 0, 71, 0, 80, 0, 0, 0, 83, 0, 80, 0, 77, 0, 0, 0, 83, 0, 84, 0, 80, 0, 0, 0, 84, 0, 77, 0, 80, 0, 0, 0, 85, 0, 71, 0, 83, 0, 0, 0, 88, 0, 69, 0, 85, 0, 0, 0, 89, 0, 68, 0, 68, 0, 0, 0, 99, 0, 106, 0, 118, 0, 0, 0, 105, 0, 102, 0, 121, 0, 0, 0, 110, 0, 122, 0, 117, 0, 0, 0, 116, 0, 109, 0, 108, 0, 0, 0, 117, 0, 102, 0, 105, 0, 0, 0, 117, 0, 105, 0, 115, 0, 0, 0, 117, 0, 116, 0, 114, 0, 0, 0, 118, 0, 97, 0, 97, 0, 0, 0, 118, 0, 101, 0, 111, 0, 0, 0, 118, 0, 101, 0, 117, 0, 0, 0, 119, 0, 105, 0, 116, 0, 0, 0, 119, 0, 110, 0, 121, 0, 0, 0, 120, 0, 97, 0, 121, 0, 0, 0, 120, 0, 98, 0, 103, 0, 0, 0, 120, 0, 103, 0, 100, 0, 0, 0, 120, 0, 117, 0, 103, 0, 0, 0, 120, 0, 117, 0, 114, 0, 0, 0, 120, 0, 119, 0, 111, 0, 0, 0, 121, 0, 101, 0, 101, 0, 0, 0, 121, 0, 103, 0, 117, 0, 0, 0, 121, 0, 105, 0, 97, 0, 0, 0, 121, 0, 107, 0, 103, 0, 0, 0, 121, 0, 107, 0, 121, 0, 0, 0, 121, 0, 108, 0, 103, 0, 0, 0, 121, 0, 109, 0, 116, 0, 0, 0, 121, 0, 114, 0, 119, 0, 0, 0, 122, 0, 98, 0, 101, 0, 0, 0, 122, 0, 107, 0, 107, 0, 0, 0, 122, 0, 111, 0, 99, 0, 0, 0, 122, 0, 116, 0, 101, 0, 0, 0, 122, 0, 116, 0, 103, 0, 0, 0, 57, 0, 54, 0, 55, 0, 0, 0, 66, 0, 84, 0, 78, 0, 0, 0, 67, 0, 72, 0, 69, 0, 0, 0, 77, 0, 75, 0, 68, 0, 0, 0, 90, 0, 65, 0, 82, 0, 0, 0, 99, 0, 107, 0, 98, 0, 0, 0, 102, 0, 105, 0, 108, 0, 0, 0, 103, 0, 113, 0, 105, 0, 0, 0, 103, 0, 115, 0, 119, 0, 0, 0, 105, 0, 108, 0, 109, 0, 0, 0, 107, 0, 110, 0, 110, 0, 0, 0, 117, 0, 103, 0, 101, 0, 0, 0, 77, 0, 88, 0, 78, 0, 0, 0, 77, 0, 88, 0, 86, 0, 0, 0, 77, 0, 65, 0, 67, 0, 0, 0, 84, 0, 85, 0, 78, 0, 0, 0, 88, 0, 65, 0, 65, 0, 0, 0, 90, 0, 90, 0, 90, 0, 0, 0, 49, 0, 52, 0, 53, 0, 0, 0, 50, 0, 51, 0, 56, 0, 0, 0, 50, 0, 52, 0, 50, 0, 0, 0, 50, 0, 54, 0, 50, 0, 0, 0, 50, 0, 55, 0, 48, 0, 0, 0, 50, 0, 57, 0, 50, 0, 0, 0, 51, 0, 52, 0, 48, 0, 0, 0, 51, 0, 52, 0, 52, 0, 0, 0, 51, 0, 52, 0, 56, 0, 0, 0, 51, 0, 53, 0, 50, 0, 0, 0, 51, 0, 53, 0, 54, 0, 0, 0, 51, 0, 54, 0, 48, 0, 0, 0, 51, 0, 54, 0, 52, 0, 0, 0, 51, 0, 54, 0, 56, 0, 0, 0, 51, 0, 55, 0, 54, 0, 0, 0, 51, 0, 56, 0, 56, 0, 0, 0, 51, 0, 57, 0, 50, 0, 0, 0, 51, 0, 57, 0, 56, 0, 0, 0, 52, 0, 48, 0, 48, 0, 0, 0, 52, 0, 51, 0, 52, 0, 0, 0, 52, 0, 52, 0, 48, 0, 0, 0, 52, 0, 52, 0, 54, 0, 0, 0, 52, 0, 53, 0, 52, 0, 0, 0, 52, 0, 53, 0, 56, 0, 0, 0, 52, 0, 54, 0, 50, 0, 0, 0, 52, 0, 55, 0, 56, 0, 0, 0, 52, 0, 56, 0, 48, 0, 0, 0, 52, 0, 56, 0, 52, 0, 0, 0, 52, 0, 57, 0, 54, 0, 0, 0, 52, 0, 57, 0, 56, 0, 0, 0, 53, 0, 48, 0, 52, 0, 0, 0, 53, 0, 49, 0, 54, 0, 0, 0, 53, 0, 52, 0, 56, 0, 0, 0, 53, 0, 53, 0, 52, 0, 0, 0, 53, 0, 53, 0, 56, 0, 0, 0, 53, 0, 54, 0, 54, 0, 0, 0, 53, 0, 55, 0, 56, 0, 0, 0, 53, 0, 56, 0, 54, 0, 0, 0, 53, 0, 57, 0, 56, 0, 0, 0, 54, 0, 48, 0, 48, 0, 0, 0, 54, 0, 48, 0, 52, 0, 0, 0, 54, 0, 48, 0, 56, 0, 0, 0, 54, 0, 50, 0, 54, 0, 0, 0, 54, 0, 53, 0, 52, 0, 0, 0, 54, 0, 55, 0, 56, 0, 0, 0, 54, 0, 56, 0, 50, 0, 0, 0, 54, 0, 57, 0, 48, 0, 0, 0, 54, 0, 57, 0, 52, 0, 0, 0, 55, 0, 48, 0, 50, 0, 0, 0, 55, 0, 48, 0, 52, 0, 0, 0, 55, 0, 48, 0, 54, 0, 0, 0, 55, 0, 49, 0, 48, 0, 0, 0, 55, 0, 50, 0, 56, 0, 0, 0, 55, 0, 52, 0, 56, 0, 0, 0, 55, 0, 53, 0, 50, 0, 0, 0, 55, 0, 53, 0, 54, 0, 0, 0, 55, 0, 54, 0, 48, 0, 0, 0, 55, 0, 54, 0, 52, 0, 0, 0, 55, 0, 55, 0, 54, 0, 0, 0, 55, 0, 56, 0, 48, 0, 0, 0, 55, 0, 56, 0, 52, 0, 0, 0, 55, 0, 56, 0, 56, 0, 0, 0, 56, 0, 48, 0, 48, 0, 0, 0, 56, 0, 48, 0, 55, 0, 0, 0, 56, 0, 49, 0, 56, 0, 0, 0, 56, 0, 50, 0, 54, 0, 0, 0, 56, 0, 51, 0, 52, 0, 0, 0, 56, 0, 52, 0, 48, 0, 0, 0, 56, 0, 53, 0, 56, 0, 0, 0, 56, 0, 54, 0, 48, 0, 0, 0, 56, 0, 56, 0, 50, 0, 0, 0, 56, 0, 57, 0, 49, 0, 0, 0, 57, 0, 53, 0, 56, 0, 0, 0, 57, 0, 53, 0, 57, 0, 0, 0, 57, 0, 54, 0, 48, 0, 0, 0, 57, 0, 54, 0, 49, 0, 0, 0, 57, 0, 54, 0, 50, 0, 0, 0, 57, 0, 54, 0, 51, 0, 0, 0, 57, 0, 54, 0, 52, 0, 0, 0, 57, 0, 54, 0, 53, 0, 0, 0, 57, 0, 54, 0, 56, 0, 0, 0, 57, 0, 54, 0, 57, 0, 0, 0, 57, 0, 55, 0, 48, 0, 0, 0, 57, 0, 55, 0, 49, 0, 0, 0, 57, 0, 55, 0, 50, 0, 0, 0, 57, 0, 55, 0, 51, 0, 0, 0, 57, 0, 55, 0, 52, 0, 0, 0, 57, 0, 55, 0, 53, 0, 0, 0, 57, 0, 55, 0, 54, 0, 0, 0, 57, 0, 55, 0, 55, 0, 0, 0, 57, 0, 55, 0, 56, 0, 0, 0, 57, 0, 55, 0, 57, 0, 0, 0, 57, 0, 56, 0, 48, 0, 0, 0, 57, 0, 56, 0, 49, 0, 0, 0, 57, 0, 56, 0, 52, 0, 0, 0, 57, 0, 56, 0, 53, 0, 0, 0, 57, 0, 56, 0, 54, 0, 0, 0, 57, 0, 57, 0, 55, 0, 0, 0, 57, 0, 57, 0, 57, 0, 0, 0, 65, 0, 69, 0, 68, 0, 0, 0, 65, 0, 70, 0, 78, 0, 0, 0, 65, 0, 76, 0, 76, 0, 0, 0, 65, 0, 77, 0, 68, 0, 0, 0, 65, 0, 78, 0, 71, 0, 0, 0, 65, 0, 79, 0, 65, 0, 0, 0, 65, 0, 85, 0, 68, 0, 0, 0, 65, 0, 87, 0, 71, 0, 0, 0, 65, 0, 90, 0, 78, 0, 0, 0, 66, 0, 65, 0, 77, 0, 0, 0, 66, 0, 66, 0, 68, 0, 0, 0, 66, 0, 68, 0, 84, 0, 0, 0, 66, 0, 69, 0, 76, 0, 0, 0, 66, 0, 71, 0, 78, 0, 0, 0, 66, 0, 72, 0, 68, 0, 0, 0, 66, 0, 73, 0, 70, 0, 0, 0, 66, 0, 77, 0, 68, 0, 0, 0, 66, 0, 78, 0, 68, 0, 0, 0, 66, 0, 79, 0, 66, 0, 0, 0, 66, 0, 79, 0, 76, 0, 0, 0, 66, 0, 79, 0, 86, 0, 0, 0, 66, 0, 82, 0, 76, 0, 0, 0, 66, 0, 82, 0, 78, 0, 0, 0, 66, 0, 83, 0, 68, 0, 0, 0, 66, 0, 87, 0, 80, 0, 0, 0, 66, 0, 89, 0, 82, 0, 0, 0, 66, 0, 90, 0, 68, 0, 0, 0, 67, 0, 65, 0, 68, 0, 0, 0, 67, 0, 68, 0, 70, 0, 0, 0, 67, 0, 72, 0, 70, 0, 0, 0, 67, 0, 72, 0, 87, 0, 0, 0, 67, 0, 76, 0, 80, 0, 0, 0, 67, 0, 78, 0, 72, 0, 0, 0, 67, 0, 78, 0, 89, 0, 0, 0, 67, 0, 79, 0, 80, 0, 0, 0, 67, 0, 79, 0, 85, 0, 0, 0, 67, 0, 82, 0, 67, 0, 0, 0, 67, 0, 85, 0, 67, 0, 0, 0, 67, 0, 85, 0, 80, 0, 0, 0, 67, 0, 86, 0, 69, 0, 0, 0, 67, 0, 89, 0, 80, 0, 0, 0, 67, 0, 90, 0, 75, 0, 0, 0, 68, 0, 74, 0, 70, 0, 0, 0, 68, 0, 75, 0, 75, 0, 0, 0, 68, 0, 79, 0, 80, 0, 0, 0, 68, 0, 90, 0, 68, 0, 0, 0, 69, 0, 71, 0, 80, 0, 0, 0, 69, 0, 82, 0, 78, 0, 0, 0, 69, 0, 83, 0, 80, 0, 0, 0, 69, 0, 84, 0, 66, 0, 0, 0, 69, 0, 85, 0, 82, 0, 0, 0, 70, 0, 74, 0, 68, 0, 0, 0, 71, 0, 66, 0, 80, 0, 0, 0, 71, 0, 69, 0, 76, 0, 0, 0, 71, 0, 72, 0, 83, 0, 0, 0, 71, 0, 73, 0, 80, 0, 0, 0, 71, 0, 77, 0, 68, 0, 0, 0, 71, 0, 78, 0, 70, 0, 0, 0, 71, 0, 82, 0, 68, 0, 0, 0, 71, 0, 84, 0, 81, 0, 0, 0, 71, 0, 89, 0, 68, 0, 0, 0, 72, 0, 75, 0, 68, 0, 0, 0, 72, 0, 78, 0, 76, 0, 0, 0, 72, 0, 82, 0, 75, 0, 0, 0, 72, 0, 85, 0, 70, 0, 0, 0, 73, 0, 68, 0, 82, 0, 0, 0, 73, 0, 76, 0, 83, 0, 0, 0, 73, 0, 78, 0, 82, 0, 0, 0, 73, 0, 81, 0, 68, 0, 0, 0, 73, 0, 82, 0, 82, 0, 0, 0, 74, 0, 77, 0, 68, 0, 0, 0, 74, 0, 79, 0, 68, 0, 0, 0, 74, 0, 80, 0, 89, 0, 0, 0, 75, 0, 69, 0, 83, 0, 0, 0, 75, 0, 71, 0, 83, 0, 0, 0, 75, 0, 80, 0, 87, 0, 0, 0, 75, 0, 82, 0, 87, 0, 0, 0, 75, 0, 87, 0, 68, 0, 0, 0, 75, 0, 89, 0, 68, 0, 0, 0, 75, 0, 90, 0, 84, 0, 0, 0, 76, 0, 65, 0, 75, 0, 0, 0, 76, 0, 66, 0, 80, 0, 0, 0, 76, 0, 75, 0, 82, 0, 0, 0, 76, 0, 82, 0, 68, 0, 0, 0, 76, 0, 83, 0, 76, 0, 0, 0, 76, 0, 84, 0, 76, 0, 0, 0, 77, 0, 65, 0, 68, 0, 0, 0, 77, 0, 65, 0, 70, 0, 0, 0, 77, 0, 68, 0, 76, 0, 0, 0, 77, 0, 71, 0, 65, 0, 0, 0, 77, 0, 77, 0, 75, 0, 0, 0, 77, 0, 79, 0, 80, 0, 0, 0, 77, 0, 82, 0, 85, 0, 0, 0, 77, 0, 85, 0, 82, 0, 0, 0, 77, 0, 86, 0, 82, 0, 0, 0, 77, 0, 87, 0, 75, 0, 0, 0, 77, 0, 89, 0, 82, 0, 0, 0, 77, 0, 90, 0, 78, 0, 0, 0, 78, 0, 73, 0, 79, 0, 0, 0, 78, 0, 79, 0, 75, 0, 0, 0, 78, 0, 80, 0, 82, 0, 0, 0, 78, 0, 90, 0, 68, 0, 0, 0, 79, 0, 77, 0, 82, 0, 0, 0, 80, 0, 65, 0, 66, 0, 0, 0, 80, 0, 69, 0, 78, 0, 0, 0, 80, 0, 71, 0, 75, 0, 0, 0, 80, 0, 72, 0, 80, 0, 0, 0, 80, 0, 75, 0, 82, 0, 0, 0, 80, 0, 76, 0, 78, 0, 0, 0, 80, 0, 89, 0, 71, 0, 0, 0, 81, 0, 65, 0, 82, 0, 0, 0, 81, 0, 85, 0, 85, 0, 0, 0, 82, 0, 79, 0, 78, 0, 0, 0, 82, 0, 83, 0, 68, 0, 0, 0, 82, 0, 85, 0, 66, 0, 0, 0, 83, 0, 65, 0, 82, 0, 0, 0, 83, 0, 66, 0, 68, 0, 0, 0, 83, 0, 67, 0, 82, 0, 0, 0, 83, 0, 68, 0, 71, 0, 0, 0, 83, 0, 69, 0, 75, 0, 0, 0, 83, 0, 71, 0, 68, 0, 0, 0, 83, 0, 72, 0, 80, 0, 0, 0, 83, 0, 76, 0, 76, 0, 0, 0, 83, 0, 79, 0, 83, 0, 0, 0, 83, 0, 82, 0, 68, 0, 0, 0, 83, 0, 83, 0, 80, 0, 0, 0, 83, 0, 84, 0, 68, 0, 0, 0, 83, 0, 85, 0, 82, 0, 0, 0, 83, 0, 89, 0, 80, 0, 0, 0, 83, 0, 90, 0, 76, 0, 0, 0, 84, 0, 72, 0, 66, 0, 0, 0, 84, 0, 74, 0, 83, 0, 0, 0, 84, 0, 77, 0, 84, 0, 0, 0, 84, 0, 78, 0, 68, 0, 0, 0, 84, 0, 79, 0, 80, 0, 0, 0, 84, 0, 82, 0, 89, 0, 0, 0, 84, 0, 87, 0, 68, 0, 0, 0, 84, 0, 90, 0, 83, 0, 0, 0, 85, 0, 65, 0, 72, 0, 0, 0, 85, 0, 71, 0, 88, 0, 0, 0, 85, 0, 83, 0, 68, 0, 0, 0, 85, 0, 89, 0, 73, 0, 0, 0, 85, 0, 89, 0, 87, 0, 0, 0, 85, 0, 90, 0, 83, 0, 0, 0, 86, 0, 69, 0, 70, 0, 0, 0, 86, 0, 69, 0, 83, 0, 0, 0, 86, 0, 78, 0, 68, 0, 0, 0, 86, 0, 85, 0, 86, 0, 0, 0, 87, 0, 83, 0, 84, 0, 0, 0, 88, 0, 65, 0, 70, 0, 0, 0, 88, 0, 65, 0, 71, 0, 0, 0, 88, 0, 65, 0, 85, 0, 0, 0, 88, 0, 66, 0, 66, 0, 0, 0, 88, 0, 68, 0, 82, 0, 0, 0, 88, 0, 79, 0, 70, 0, 0, 0, 88, 0, 80, 0, 68, 0, 0, 0, 88, 0, 80, 0, 84, 0, 0, 0, 88, 0, 84, 0, 83, 0, 0, 0, 89, 0, 69, 0, 82, 0, 0, 0, 90, 0, 77, 0, 87, 0, 0, 0, 97, 0, 106, 0, 112, 0, 0, 0, 97, 0, 114, 0, 113, 0, 0, 0, 97, 0, 114, 0, 115, 0, 0, 0, 97, 0, 114, 0, 121, 0, 0, 0, 97, 0, 114, 0, 122, 0, 0, 0, 98, 0, 106, 0, 113, 0, 0, 0, 99, 0, 103, 0, 50, 0, 0, 0, 99, 0, 103, 0, 53, 0, 0, 0, 99, 0, 104, 0, 114, 0, 0, 0, 99, 0, 106, 0, 121, 0, 0, 0, 99, 0, 114, 0, 97, 0, 0, 0, 99, 0, 114, 0, 99, 0, 0, 0, 99, 0, 114, 0, 108, 0, 0, 0, 99, 0, 114, 0, 112, 0, 0, 0, 99, 0, 114, 0, 115, 0, 0, 0, 99, 0, 118, 0, 98, 0, 0, 0, 99, 0, 118, 0, 115, 0, 0, 0, 99, 0, 122, 0, 104, 0, 0, 0, 100, 0, 116, 0, 121, 0, 0, 0, 101, 0, 99, 0, 121, 0, 0, 0, 101, 0, 103, 0, 99, 0, 0, 0, 101, 0, 115, 0, 98, 0, 0, 0, 101, 0, 115, 0, 99, 0, 0, 0, 101, 0, 115, 0, 112, 0, 0, 0, 101, 0, 115, 0, 118, 0, 0, 0, 101, 0, 121, 0, 111, 0, 0, 0, 102, 0, 102, 0, 109, 0, 0, 0, 102, 0, 106, 0, 99, 0, 0, 0, 102, 0, 106, 0, 101, 0, 0, 0, 102, 0, 106, 0, 110, 0, 0, 0, 102, 0, 106, 0, 114, 0, 0, 0, 102, 0, 106, 0, 119, 0, 0, 0, 102, 0, 117, 0, 98, 0, 0, 0, 103, 0, 97, 0, 97, 0, 0, 0, 103, 0, 109, 0, 98, 0, 0, 0, 103, 0, 109, 0, 117, 0, 0, 0, 103, 0, 109, 0, 119, 0, 0, 0, 103, 0, 110, 0, 98, 0, 0, 0, 103, 0, 110, 0, 99, 0, 0, 0, 103, 0, 110, 0, 100, 0, 0, 0, 103, 0, 110, 0, 102, 0, 0, 0, 103, 0, 110, 0, 107, 0, 0, 0, 103, 0, 110, 0, 108, 0, 0, 0, 103, 0, 110, 0, 109, 0, 0, 0, 103, 0, 110, 0, 110, 0, 0, 0, 103, 0, 113, 0, 99, 0, 0, 0, 103, 0, 114, 0, 109, 0, 0, 0, 103, 0, 119, 0, 108, 0, 0, 0, 103, 0, 119, 0, 115, 0, 0, 0, 104, 0, 106, 0, 105, 0, 0, 0, 104, 0, 115, 0, 110, 0, 0, 0, 105, 0, 101, 0, 99, 0, 0, 0, 105, 0, 101, 0, 100, 0, 0, 0, 105, 0, 101, 0, 103, 0, 0, 0, 105, 0, 101, 0, 108, 0, 0, 0, 105, 0, 101, 0, 109, 0, 0, 0, 105, 0, 101, 0, 117, 0, 0, 0, 105, 0, 108, 0, 100, 0, 0, 0, 105, 0, 108, 0, 122, 0, 0, 0, 107, 0, 103, 0, 106, 0, 0, 0, 107, 0, 103, 0, 116, 0, 0, 0, 107, 0, 103, 0, 121, 0, 0, 0, 107, 0, 104, 0, 49, 0, 0, 0, 107, 0, 104, 0, 50, 0, 0, 0, 107, 0, 105, 0, 103, 0, 0, 0, 107, 0, 105, 0, 112, 0, 0, 0, 107, 0, 109, 0, 97, 0, 0, 0, 107, 0, 109, 0, 103, 0, 0, 0, 107, 0, 109, 0, 109, 0, 0, 0, 107, 0, 110, 0, 107, 0, 0, 0, 107, 0, 114, 0, 105, 0, 0, 0, 108, 0, 107, 0, 49, 0, 0, 0, 108, 0, 107, 0, 50, 0, 0, 0, 108, 0, 107, 0, 51, 0, 0, 0, 108, 0, 107, 0, 52, 0, 0, 0, 108, 0, 107, 0, 53, 0, 0, 0, 108, 0, 107, 0, 54, 0, 0, 0, 108, 0, 107, 0, 55, 0, 0, 0, 108, 0, 107, 0, 56, 0, 0, 0, 108, 0, 107, 0, 57, 0, 0, 0, 108, 0, 111, 0, 122, 0, 0, 0, 108, 0, 114, 0, 105, 0, 0, 0, 108, 0, 115, 0, 97, 0, 0, 0, 108, 0, 115, 0, 103, 0, 0, 0, 108, 0, 116, 0, 115, 0, 0, 0, 109, 0, 102, 0, 101, 0, 0, 0, 109, 0, 103, 0, 97, 0, 0, 0, 109, 0, 103, 0, 100, 0, 0, 0, 109, 0, 103, 0, 102, 0, 0, 0, 109, 0, 103, 0, 109, 0, 0, 0, 109, 0, 104, 0, 108, 0, 0, 0, 109, 0, 104, 0, 116, 0, 0, 0, 109, 0, 108, 0, 49, 0, 0, 0, 109, 0, 110, 0, 49, 0, 0, 0, 109, 0, 119, 0, 99, 0, 0, 0, 109, 0, 119, 0, 110, 0, 0, 0, 109, 0, 119, 0, 115, 0, 0, 0, 109, 0, 122, 0, 103, 0, 0, 0, 109, 0, 122, 0, 105, 0, 0, 0, 109, 0, 122, 0, 108, 0, 0, 0, 109, 0, 122, 0, 110, 0, 0, 0, 110, 0, 115, 0, 111, 0, 0, 0, 110, 0, 122, 0, 115, 0, 0, 0, 111, 0, 114, 0, 99, 0, 0, 0, 112, 0, 97, 0, 49, 0, 0, 0, 112, 0, 99, 0, 109, 0, 0, 0, 112, 0, 121, 0, 49, 0, 0, 0, 113, 0, 118, 0, 97, 0, 0, 0, 113, 0, 118, 0, 99, 0, 0, 0, 113, 0, 118, 0, 101, 0, 0, 0, 113, 0, 118, 0, 115, 0, 0, 0, 113, 0, 118, 0, 119, 0, 0, 0, 113, 0, 119, 0, 99, 0, 0, 0, 113, 0, 119, 0, 104, 0, 0, 0, 113, 0, 120, 0, 97, 0, 0, 0, 113, 0, 120, 0, 99, 0, 0, 0, 113, 0, 120, 0, 108, 0, 0, 0, 113, 0, 120, 0, 119, 0, 0, 0, 114, 0, 111, 0, 98, 0, 0, 0, 115, 0, 101, 0, 107, 0, 0, 0, 115, 0, 108, 0, 110, 0, 0, 0, 115, 0, 108, 0, 115, 0, 0, 0, 115, 0, 108, 0, 119, 0, 0, 0, 115, 0, 115, 0, 121, 0, 0, 0, 115, 0, 116, 0, 112, 0, 0, 0, 116, 0, 103, 0, 99, 0, 0, 0, 116, 0, 103, 0, 107, 0, 0, 0, 116, 0, 103, 0, 109, 0, 0, 0, 116, 0, 103, 0, 112, 0, 0, 0, 116, 0, 103, 0, 115, 0, 0, 0, 116, 0, 104, 0, 115, 0, 0, 0, 116, 0, 108, 0, 104, 0, 0, 0, 116, 0, 109, 0, 100, 0, 0, 0, 117, 0, 103, 0, 99, 0, 0, 0, 117, 0, 103, 0, 110, 0, 0, 0, 117, 0, 103, 0, 119, 0, 0, 0, 117, 0, 122, 0, 115, 0, 0, 0, 118, 0, 114, 0, 111, 0, 0, 0, 67, 0, 88, 0, 82, 0, 0, 0, 77, 0, 88, 0, 80, 0, 0, 0, 83, 0, 88, 0, 77, 0, 0, 0, 88, 0, 81, 0, 81, 0, 0, 0, 106, 0, 105, 0, 111, 0, 0, 0, 107, 0, 111, 0, 106, 0, 0, 0, 115, 0, 100, 0, 122, 0, 0, 0, 116, 0, 100, 0, 118, 0, 0, 0, 117, 0, 98, 0, 105, 0, 0, 0, 117, 0, 112, 0, 105, 0, 0, 0, 119, 0, 104, 0, 117, 0, 0, 0, 119, 0, 115, 0, 107, 0, 0, 0, 120, 0, 99, 0, 101, 0, 0, 0, 120, 0, 99, 0, 114, 0, 0, 0, 120, 0, 105, 0, 105, 0, 0, 0, 120, 0, 110, 0, 100, 0, 0, 0, 120, 0, 118, 0, 101, 0, 0, 0, 121, 0, 101, 0, 108, 0, 0, 0, 121, 0, 107, 0, 105, 0, 0, 0, 121, 0, 111, 0, 115, 0, 0, 0, 121, 0, 115, 0, 103, 0, 0, 0, 121, 0, 115, 0, 108, 0, 0, 0, 121, 0, 116, 0, 108, 0, 0, 0, 121, 0, 116, 0, 119, 0, 0, 0, 121, 0, 116, 0, 121, 0, 0, 0, 122, 0, 99, 0, 97, 0, 0, 0, 122, 0, 99, 0, 104, 0, 0, 0, 122, 0, 105, 0, 119, 0, 0, 0, 122, 0, 119, 0, 97, 0, 0, 0, 122, 0, 122, 0, 97, 0, 0, 0, 50, 0, 51, 0, 57, 0, 0, 0, 50, 0, 52, 0, 54, 0, 0, 0, 50, 0, 52, 0, 56, 0, 0, 0, 50, 0, 52, 0, 57, 0, 0, 0, 50, 0, 53, 0, 48, 0, 0, 0, 50, 0, 53, 0, 52, 0, 0, 0, 50, 0, 53, 0, 56, 0, 0, 0, 50, 0, 54, 0, 48, 0, 0, 0, 50, 0, 54, 0, 54, 0, 0, 0, 50, 0, 54, 0, 56, 0, 0, 0, 50, 0, 55, 0, 53, 0, 0, 0, 50, 0, 55, 0, 54, 0, 0, 0, 50, 0, 55, 0, 56, 0, 0, 0, 50, 0, 56, 0, 56, 0, 0, 0, 50, 0, 57, 0, 54, 0, 0, 0, 51, 0, 48, 0, 48, 0, 0, 0, 51, 0, 51, 0, 54, 0, 0, 0, 51, 0, 55, 0, 50, 0, 0, 0, 51, 0, 56, 0, 48, 0, 0, 0, 51, 0, 56, 0, 52, 0, 0, 0, 52, 0, 51, 0, 56, 0, 0, 0, 52, 0, 52, 0, 50, 0, 0, 0, 52, 0, 53, 0, 48, 0, 0, 0, 52, 0, 54, 0, 54, 0, 0, 0, 52, 0, 55, 0, 48, 0, 0, 0, 52, 0, 55, 0, 52, 0, 0, 0, 52, 0, 57, 0, 50, 0, 0, 0, 52, 0, 57, 0, 57, 0, 0, 0, 53, 0, 48, 0, 48, 0, 0, 0, 53, 0, 48, 0, 56, 0, 0, 0, 53, 0, 50, 0, 48, 0, 0, 0, 53, 0, 50, 0, 56, 0, 0, 0, 53, 0, 51, 0, 48, 0, 0, 0, 53, 0, 51, 0, 53, 0, 0, 0, 53, 0, 51, 0, 54, 0, 0, 0, 53, 0, 52, 0, 48, 0, 0, 0, 53, 0, 54, 0, 50, 0, 0, 0, 53, 0, 55, 0, 48, 0, 0, 0, 53, 0, 55, 0, 52, 0, 0, 0, 53, 0, 56, 0, 48, 0, 0, 0, 53, 0, 56, 0, 49, 0, 0, 0, 53, 0, 56, 0, 51, 0, 0, 0, 53, 0, 56, 0, 52, 0, 0, 0, 53, 0, 56, 0, 53, 0, 0, 0, 53, 0, 57, 0, 48, 0, 0, 0, 53, 0, 57, 0, 49, 0, 0, 0, 54, 0, 49, 0, 50, 0, 0, 0, 54, 0, 49, 0, 54, 0, 0, 0, 54, 0, 50, 0, 48, 0, 0, 0, 54, 0, 50, 0, 52, 0, 0, 0, 54, 0, 51, 0, 48, 0, 0, 0, 54, 0, 51, 0, 56, 0, 0, 0, 54, 0, 53, 0, 50, 0, 0, 0, 54, 0, 53, 0, 57, 0, 0, 0, 54, 0, 54, 0, 48, 0, 0, 0, 54, 0, 54, 0, 50, 0, 0, 0, 54, 0, 54, 0, 51, 0, 0, 0, 54, 0, 54, 0, 54, 0, 0, 0, 54, 0, 55, 0, 48, 0, 0, 0, 54, 0, 55, 0, 52, 0, 0, 0, 54, 0, 56, 0, 54, 0, 0, 0, 54, 0, 56, 0, 56, 0, 0, 0, 55, 0, 48, 0, 51, 0, 0, 0, 55, 0, 48, 0, 53, 0, 0, 0, 55, 0, 49, 0, 54, 0, 0, 0, 55, 0, 50, 0, 48, 0, 0, 0, 55, 0, 50, 0, 57, 0, 0, 0, 55, 0, 51, 0, 50, 0, 0, 0, 55, 0, 52, 0, 48, 0, 0, 0, 55, 0, 52, 0, 52, 0, 0, 0, 55, 0, 54, 0, 50, 0, 0, 0, 55, 0, 54, 0, 56, 0, 0, 0, 55, 0, 55, 0, 50, 0, 0, 0, 55, 0, 57, 0, 50, 0, 0, 0, 55, 0, 57, 0, 53, 0, 0, 0, 55, 0, 57, 0, 54, 0, 0, 0, 55, 0, 57, 0, 56, 0, 0, 0, 56, 0, 49, 0, 48, 0, 0, 0, 56, 0, 51, 0, 49, 0, 0, 0, 56, 0, 51, 0, 50, 0, 0, 0, 56, 0, 51, 0, 51, 0, 0, 0, 56, 0, 53, 0, 48, 0, 0, 0, 56, 0, 53, 0, 52, 0, 0, 0, 56, 0, 54, 0, 50, 0, 0, 0, 56, 0, 55, 0, 54, 0, 0, 0, 56, 0, 56, 0, 54, 0, 0, 0, 56, 0, 56, 0, 55, 0, 0, 0, 56, 0, 57, 0, 52, 0, 0, 0, 57, 0, 50, 0, 55, 0, 0, 0, 57, 0, 50, 0, 56, 0, 0, 0, 57, 0, 50, 0, 57, 0, 0, 0, 57, 0, 51, 0, 48, 0, 0, 0, 57, 0, 51, 0, 49, 0, 0, 0, 57, 0, 51, 0, 52, 0, 0, 0, 57, 0, 51, 0, 54, 0, 0, 0, 57, 0, 51, 0, 55, 0, 0, 0, 57, 0, 51, 0, 56, 0, 0, 0, 57, 0, 52, 0, 48, 0, 0, 0, 57, 0, 52, 0, 49, 0, 0, 0, 57, 0, 52, 0, 52, 0, 0, 0, 57, 0, 52, 0, 54, 0, 0, 0, 57, 0, 52, 0, 55, 0, 0, 0, 57, 0, 52, 0, 56, 0, 0, 0, 57, 0, 52, 0, 57, 0, 0, 0, 57, 0, 53, 0, 48, 0, 0, 0, 57, 0, 53, 0, 49, 0, 0, 0, 57, 0, 53, 0, 50, 0, 0, 0, 57, 0, 53, 0, 51, 0, 0, 0, 57, 0, 53, 0, 53, 0, 0, 0, 57, 0, 53, 0, 54, 0, 0, 0, 57, 0, 53, 0, 55, 0, 0, 0, 57, 0, 54, 0, 54, 0, 0, 0, 57, 0, 56, 0, 50, 0, 0, 0, 57, 0, 56, 0, 51, 0, 0, 0, 57, 0, 56, 0, 55, 0, 0, 0, 57, 0, 56, 0, 56, 0, 0, 0, 57, 0, 56, 0, 57, 0, 0, 0, 57, 0, 57, 0, 49, 0, 0, 0, 57, 0, 57, 0, 50, 0, 0, 0, 57, 0, 57, 0, 51, 0, 0, 0, 57, 0, 57, 0, 53, 0, 0, 0, 57, 0, 57, 0, 56, 0, 0, 0, 65, 0, 65, 0, 65, 0, 0, 0, 65, 0, 66, 0, 87, 0, 0, 0, 65, 0, 68, 0, 80, 0, 0, 0, 65, 0, 70, 0, 65, 0, 0, 0, 65, 0, 70, 0, 71, 0, 0, 0, 65, 0, 71, 0, 79, 0, 0, 0, 65, 0, 73, 0, 65, 0, 0, 0, 65, 0, 76, 0, 66, 0, 0, 0, 65, 0, 76, 0, 75, 0, 0, 0, 65, 0, 78, 0, 68, 0, 0, 0, 65, 0, 78, 0, 84, 0, 0, 0, 65, 0, 79, 0, 75, 0, 0, 0, 65, 0, 79, 0, 78, 0, 0, 0, 65, 0, 79, 0, 82, 0, 0, 0, 65, 0, 82, 0, 65, 0, 0, 0, 65, 0, 82, 0, 69, 0, 0, 0, 65, 0, 82, 0, 71, 0, 0, 0, 65, 0, 82, 0, 77, 0, 0, 0, 65, 0, 82, 0, 80, 0, 0, 0, 65, 0, 83, 0, 77, 0, 0, 0, 65, 0, 84, 0, 65, 0, 0, 0, 65, 0, 84, 0, 71, 0, 0, 0, 65, 0, 84, 0, 83, 0, 0, 0, 65, 0, 85, 0, 83, 0, 0, 0, 65, 0, 85, 0, 84, 0, 0, 0, 65, 0, 90, 0, 69, 0, 0, 0, 65, 0, 90, 0, 77, 0, 0, 0, 66, 0, 65, 0, 68, 0, 0, 0, 66, 0, 65, 0, 78, 0, 0, 0, 66, 0, 68, 0, 73, 0, 0, 0, 66, 0, 69, 0, 67, 0, 0, 0, 66, 0, 69, 0, 70, 0, 0, 0, 66, 0, 69, 0, 78, 0, 0, 0, 66, 0, 69, 0, 83, 0, 0, 0, 66, 0, 70, 0, 65, 0, 0, 0, 66, 0, 71, 0, 68, 0, 0, 0, 66, 0, 71, 0, 79, 0, 0, 0, 66, 0, 72, 0, 82, 0, 0, 0, 66, 0, 72, 0, 83, 0, 0, 0, 66, 0, 73, 0, 72, 0, 0, 0, 66, 0, 76, 0, 77, 0, 0, 0, 66, 0, 76, 0, 82, 0, 0, 0, 66, 0, 76, 0, 90, 0, 0, 0, 66, 0, 77, 0, 85, 0, 0, 0, 66, 0, 79, 0, 80, 0, 0, 0, 66, 0, 82, 0, 65, 0, 0, 0, 66, 0, 82, 0, 66, 0, 0, 0, 66, 0, 82, 0, 69, 0, 0, 0, 66, 0, 82, 0, 82, 0, 0, 0, 66, 0, 82, 0, 90, 0, 0, 0, 66, 0, 85, 0, 75, 0, 0, 0, 66, 0, 85, 0, 82, 0, 0, 0, 66, 0, 86, 0, 84, 0, 0, 0, 66, 0, 87, 0, 65, 0, 0, 0, 66, 0, 89, 0, 66, 0, 0, 0, 66, 0, 89, 0, 78, 0, 0, 0, 67, 0, 65, 0, 70, 0, 0, 0, 67, 0, 65, 0, 78, 0, 0, 0, 67, 0, 67, 0, 75, 0, 0, 0, 67, 0, 72, 0, 76, 0, 0, 0, 67, 0, 72, 0, 78, 0, 0, 0, 67, 0, 73, 0, 86, 0, 0, 0, 67, 0, 76, 0, 70, 0, 0, 0, 67, 0, 77, 0, 82, 0, 0, 0, 67, 0, 78, 0, 88, 0, 0, 0, 67, 0, 79, 0, 68, 0, 0, 0, 67, 0, 79, 0, 71, 0, 0, 0, 67, 0, 79, 0, 75, 0, 0, 0, 67, 0, 79, 0, 76, 0, 0, 0, 67, 0, 79, 0, 77, 0, 0, 0, 67, 0, 80, 0, 86, 0, 0, 0, 67, 0, 82, 0, 73, 0, 0, 0, 67, 0, 83, 0, 68, 0, 0, 0, 67, 0, 83, 0, 75, 0, 0, 0, 67, 0, 85, 0, 66, 0, 0, 0, 67, 0, 85, 0, 87, 0, 0, 0, 67, 0, 89, 0, 77, 0, 0, 0, 67, 0, 90, 0, 69, 0, 0, 0, 68, 0, 68, 0, 77, 0, 0, 0, 68, 0, 68, 0, 82, 0, 0, 0, 68, 0, 69, 0, 77, 0, 0, 0, 68, 0, 69, 0, 85, 0, 0, 0, 68, 0, 74, 0, 73, 0, 0, 0, 68, 0, 78, 0, 75, 0, 0, 0, 68, 0, 79, 0, 77, 0, 0, 0, 68, 0, 90, 0, 65, 0, 0, 0, 69, 0, 67, 0, 85, 0, 0, 0, 69, 0, 67, 0, 86, 0, 0, 0, 69, 0, 69, 0, 75, 0, 0, 0, 69, 0, 71, 0, 89, 0, 0, 0, 69, 0, 82, 0, 73, 0, 0, 0, 69, 0, 83, 0, 72, 0, 0, 0, 69, 0, 83, 0, 84, 0, 0, 0, 70, 0, 73, 0, 77, 0, 0, 0, 70, 0, 73, 0, 78, 0, 0, 0, 70, 0, 74, 0, 73, 0, 0, 0, 70, 0, 82, 0, 65, 0, 0, 0, 70, 0, 82, 0, 70, 0, 0, 0, 70, 0, 82, 0, 79, 0, 0, 0, 70, 0, 83, 0, 77, 0, 0, 0, 70, 0, 88, 0, 88, 0, 0, 0, 71, 0, 65, 0, 66, 0, 0, 0, 71, 0, 66, 0, 82, 0, 0, 0, 71, 0, 69, 0, 75, 0, 0, 0, 71, 0, 69, 0, 79, 0, 0, 0, 71, 0, 71, 0, 89, 0, 0, 0, 71, 0, 72, 0, 65, 0, 0, 0, 71, 0, 72, 0, 67, 0, 0, 0, 71, 0, 73, 0, 66, 0, 0, 0, 71, 0, 73, 0, 78, 0, 0, 0, 71, 0, 76, 0, 80, 0, 0, 0, 71, 0, 77, 0, 66, 0, 0, 0, 71, 0, 78, 0, 66, 0, 0, 0, 71, 0, 78, 0, 81, 0, 0, 0, 71, 0, 78, 0, 83, 0, 0, 0, 71, 0, 81, 0, 69, 0, 0, 0, 71, 0, 82, 0, 67, 0, 0, 0, 71, 0, 82, 0, 76, 0, 0, 0, 71, 0, 85, 0, 70, 0, 0, 0, 71, 0, 85, 0, 89, 0, 0, 0, 71, 0, 87, 0, 69, 0, 0, 0, 71, 0, 87, 0, 80, 0, 0, 0, 72, 0, 77, 0, 68, 0, 0, 0, 72, 0, 78, 0, 68, 0, 0, 0, 72, 0, 82, 0, 68, 0, 0, 0, 72, 0, 82, 0, 86, 0, 0, 0, 72, 0, 84, 0, 73, 0, 0, 0, 72, 0, 85, 0, 78, 0, 0, 0, 73, 0, 68, 0, 78, 0, 0, 0, 73, 0, 69, 0, 80, 0, 0, 0, 73, 0, 76, 0, 80, 0, 0, 0, 73, 0, 76, 0, 82, 0, 0, 0, 73, 0, 77, 0, 78, 0, 0, 0, 73, 0, 78, 0, 68, 0, 0, 0, 73, 0, 79, 0, 84, 0, 0, 0, 73, 0, 82, 0, 76, 0, 0, 0, 73, 0, 82, 0, 78, 0, 0, 0, 73, 0, 82, 0, 81, 0, 0, 0, 73, 0, 83, 0, 76, 0, 0, 0, 73, 0, 83, 0, 82, 0, 0, 0, 73, 0, 84, 0, 76, 0, 0, 0, 74, 0, 65, 0, 77, 0, 0, 0, 74, 0, 69, 0, 89, 0, 0, 0, 74, 0, 79, 0, 82, 0, 0, 0, 74, 0, 80, 0, 78, 0, 0, 0, 75, 0, 69, 0, 78, 0, 0, 0, 75, 0, 71, 0, 90, 0, 0, 0, 75, 0, 73, 0, 82, 0, 0, 0, 75, 0, 78, 0, 65, 0, 0, 0, 75, 0, 79, 0, 82, 0, 0, 0, 75, 0, 82, 0, 72, 0, 0, 0, 75, 0, 82, 0, 79, 0, 0, 0, 75, 0, 87, 0, 84, 0, 0, 0, 76, 0, 65, 0, 79, 0, 0, 0, 76, 0, 66, 0, 82, 0, 0, 0, 76, 0, 67, 0, 65, 0, 0, 0, 76, 0, 73, 0, 69, 0, 0, 0, 76, 0, 75, 0, 65, 0, 0, 0, 76, 0, 84, 0, 84, 0, 0, 0, 76, 0, 84, 0, 85, 0, 0, 0, 76, 0, 85, 0, 67, 0, 0, 0, 76, 0, 85, 0, 70, 0, 0, 0, 76, 0, 85, 0, 76, 0, 0, 0, 76, 0, 85, 0, 88, 0, 0, 0, 76, 0, 86, 0, 76, 0, 0, 0, 76, 0, 86, 0, 82, 0, 0, 0, 77, 0, 65, 0, 82, 0, 0, 0, 77, 0, 67, 0, 79, 0, 0, 0, 77, 0, 68, 0, 65, 0, 0, 0, 77, 0, 68, 0, 67, 0, 0, 0, 77, 0, 68, 0, 71, 0, 0, 0, 77, 0, 68, 0, 86, 0, 0, 0, 77, 0, 69, 0, 88, 0, 0, 0, 77, 0, 72, 0, 76, 0, 0, 0, 77, 0, 75, 0, 78, 0, 0, 0, 77, 0, 76, 0, 70, 0, 0, 0, 77, 0, 77, 0, 82, 0, 0, 0, 77, 0, 78, 0, 71, 0, 0, 0, 77, 0, 79, 0, 90, 0, 0, 0, 77, 0, 82, 0, 84, 0, 0, 0, 77, 0, 84, 0, 76, 0, 0, 0, 77, 0, 84, 0, 80, 0, 0, 0, 77, 0, 84, 0, 81, 0, 0, 0, 77, 0, 85, 0, 83, 0, 0, 0, 77, 0, 86, 0, 80, 0, 0, 0, 77, 0, 87, 0, 73, 0, 0, 0, 77, 0, 89, 0, 83, 0, 0, 0, 77, 0, 90, 0, 69, 0, 0, 0, 77, 0, 90, 0, 77, 0, 0, 0, 78, 0, 67, 0, 76, 0, 0, 0, 78, 0, 69, 0, 82, 0, 0, 0, 78, 0, 70, 0, 75, 0, 0, 0, 78, 0, 71, 0, 65, 0, 0, 0, 78, 0, 73, 0, 85, 0, 0, 0, 78, 0, 76, 0, 68, 0, 0, 0, 78, 0, 76, 0, 71, 0, 0, 0, 78, 0, 79, 0, 82, 0, 0, 0, 78, 0, 84, 0, 90, 0, 0, 0, 78, 0, 90, 0, 76, 0, 0, 0, 80, 0, 65, 0, 75, 0, 0, 0, 80, 0, 69, 0, 73, 0, 0, 0, 80, 0, 69, 0, 82, 0, 0, 0, 80, 0, 69, 0, 83, 0, 0, 0, 80, 0, 72, 0, 76, 0, 0, 0, 80, 0, 76, 0, 87, 0, 0, 0, 80, 0, 76, 0, 90, 0, 0, 0, 80, 0, 78, 0, 71, 0, 0, 0, 80, 0, 79, 0, 76, 0, 0, 0, 80, 0, 82, 0, 73, 0, 0, 0, 80, 0, 82, 0, 75, 0, 0, 0, 80, 0, 82, 0, 84, 0, 0, 0, 80, 0, 82, 0, 89, 0, 0, 0, 80, 0, 83, 0, 69, 0, 0, 0, 80, 0, 84, 0, 69, 0, 0, 0, 80, 0, 89, 0, 70, 0, 0, 0, 81, 0, 65, 0, 84, 0, 0, 0, 81, 0, 77, 0, 77, 0, 0, 0, 81, 0, 78, 0, 78, 0, 0, 0, 81, 0, 79, 0, 79, 0, 0, 0, 81, 0, 80, 0, 80, 0, 0, 0, 81, 0, 81, 0, 81, 0, 0, 0, 81, 0, 82, 0, 82, 0, 0, 0, 81, 0, 83, 0, 83, 0, 0, 0, 81, 0, 84, 0, 84, 0, 0, 0, 81, 0, 86, 0, 86, 0, 0, 0, 81, 0, 87, 0, 87, 0, 0, 0, 81, 0, 88, 0, 88, 0, 0, 0, 81, 0, 89, 0, 89, 0, 0, 0, 81, 0, 90, 0, 90, 0, 0, 0, 82, 0, 69, 0, 85, 0, 0, 0, 82, 0, 72, 0, 68, 0, 0, 0, 82, 0, 79, 0, 76, 0, 0, 0, 82, 0, 79, 0, 85, 0, 0, 0, 82, 0, 85, 0, 82, 0, 0, 0, 82, 0, 85, 0, 83, 0, 0, 0, 82, 0, 87, 0, 65, 0, 0, 0, 83, 0, 65, 0, 85, 0, 0, 0, 83, 0, 68, 0, 68, 0, 0, 0, 83, 0, 68, 0, 78, 0, 0, 0, 83, 0, 68, 0, 80, 0, 0, 0, 83, 0, 69, 0, 78, 0, 0, 0, 83, 0, 71, 0, 83, 0, 0, 0, 83, 0, 72, 0, 78, 0, 0, 0, 83, 0, 74, 0, 77, 0, 0, 0, 83, 0, 75, 0, 75, 0, 0, 0, 83, 0, 76, 0, 69, 0, 0, 0, 83, 0, 82, 0, 66, 0, 0, 0, 83, 0, 82, 0, 71, 0, 0, 0, 83, 0, 83, 0, 68, 0, 0, 0, 83, 0, 85, 0, 78, 0, 0, 0, 83, 0, 86, 0, 67, 0, 0, 0, 83, 0, 86, 0, 75, 0, 0, 0, 83, 0, 87, 0, 69, 0, 0, 0, 83, 0, 87, 0, 90, 0, 0, 0, 83, 0, 89, 0, 67, 0, 0, 0, 83, 0, 89, 0, 82, 0, 0, 0, 84, 0, 67, 0, 65, 0, 0, 0, 84, 0, 67, 0, 68, 0, 0, 0, 84, 0, 71, 0, 79, 0, 0, 0, 84, 0, 72, 0, 65, 0, 0, 0, 84, 0, 74, 0, 75, 0, 0, 0, 84, 0, 74, 0, 82, 0, 0, 0, 84, 0, 75, 0, 76, 0, 0, 0, 84, 0, 76, 0, 83, 0, 0, 0, 84, 0, 77, 0, 77, 0, 0, 0, 84, 0, 79, 0, 78, 0, 0, 0, 84, 0, 80, 0, 69, 0, 0, 0, 84, 0, 82, 0, 76, 0, 0, 0, 84, 0, 84, 0, 79, 0, 0, 0, 84, 0, 85, 0, 82, 0, 0, 0, 84, 0, 85, 0, 86, 0, 0, 0, 84, 0, 87, 0, 78, 0, 0, 0, 84, 0, 90, 0, 65, 0, 0, 0, 85, 0, 65, 0, 75, 0, 0, 0, 85, 0, 77, 0, 73, 0, 0, 0, 85, 0, 82, 0, 89, 0, 0, 0, 85, 0, 83, 0, 83, 0, 0, 0, 85, 0, 89, 0, 80, 0, 0, 0, 85, 0, 90, 0, 66, 0, 0, 0, 86, 0, 65, 0, 84, 0, 0, 0, 86, 0, 67, 0, 84, 0, 0, 0, 86, 0, 69, 0, 66, 0, 0, 0, 86, 0, 69, 0, 78, 0, 0, 0, 86, 0, 71, 0, 66, 0, 0, 0, 86, 0, 78, 0, 77, 0, 0, 0, 86, 0, 78, 0, 78, 0, 0, 0, 86, 0, 85, 0, 84, 0, 0, 0, 87, 0, 76, 0, 70, 0, 0, 0, 88, 0, 66, 0, 67, 0, 0, 0, 88, 0, 67, 0, 67, 0, 0, 0, 88, 0, 68, 0, 68, 0, 0, 0, 88, 0, 70, 0, 70, 0, 0, 0, 88, 0, 70, 0, 85, 0, 0, 0, 88, 0, 71, 0, 71, 0, 0, 0, 88, 0, 72, 0, 72, 0, 0, 0, 88, 0, 73, 0, 73, 0, 0, 0, 88, 0, 74, 0, 74, 0, 0, 0, 88, 0, 75, 0, 75, 0, 0, 0, 88, 0, 76, 0, 76, 0, 0, 0, 88, 0, 78, 0, 78, 0, 0, 0, 88, 0, 79, 0, 79, 0, 0, 0, 88, 0, 80, 0, 80, 0, 0, 0, 88, 0, 82, 0, 82, 0, 0, 0, 88, 0, 83, 0, 83, 0, 0, 0, 88, 0, 84, 0, 84, 0, 0, 0, 88, 0, 85, 0, 85, 0, 0, 0, 88, 0, 86, 0, 86, 0, 0, 0, 88, 0, 87, 0, 87, 0, 0, 0, 88, 0, 89, 0, 89, 0, 0, 0, 88, 0, 90, 0, 90, 0, 0, 0, 89, 0, 69, 0, 77, 0, 0, 0, 89, 0, 85, 0, 68, 0, 0, 0, 89, 0, 85, 0, 71, 0, 0, 0, 89, 0, 85, 0, 82, 0, 0, 0, 90, 0, 77, 0, 66, 0, 0, 0, 90, 0, 82, 0, 78, 0, 0, 0, 90, 0, 82, 0, 90, 0, 0, 0, 90, 0, 87, 0, 68, 0, 0, 0, 90, 0, 87, 0, 69, 0, 0, 0, 90, 0, 87, 0, 76, 0, 0, 0, 90, 0, 87, 0, 82, 0, 0, 0, 97, 0, 97, 0, 101, 0, 0, 0, 97, 0, 97, 0, 109, 0, 0, 0, 97, 0, 97, 0, 111, 0, 0, 0, 97, 0, 97, 0, 122, 0, 0, 0, 97, 0, 98, 0, 104, 0, 0, 0, 97, 0, 98, 0, 118, 0, 0, 0, 97, 0, 99, 0, 109, 0, 0, 0, 97, 0, 99, 0, 113, 0, 0, 0, 97, 0, 99, 0, 119, 0, 0, 0, 97, 0, 99, 0, 120, 0, 0, 0, 97, 0, 99, 0, 121, 0, 0, 0, 97, 0, 100, 0, 102, 0, 0, 0, 97, 0, 100, 0, 108, 0, 0, 0, 97, 0, 100, 0, 112, 0, 0, 0, 97, 0, 101, 0, 98, 0, 0, 0, 97, 0, 101, 0, 117, 0, 0, 0, 97, 0, 101, 0, 119, 0, 0, 0, 97, 0, 102, 0, 98, 0, 0, 0, 97, 0, 102, 0, 107, 0, 0, 0, 97, 0, 102, 0, 122, 0, 0, 0, 97, 0, 103, 0, 112, 0, 0, 0, 97, 0, 105, 0, 115, 0, 0, 0, 97, 0, 105, 0, 116, 0, 0, 0, 97, 0, 106, 0, 97, 0, 0, 0, 97, 0, 106, 0, 103, 0, 0, 0, 97, 0, 106, 0, 105, 0, 0, 0, 97, 0, 106, 0, 110, 0, 0, 0, 97, 0, 106, 0, 116, 0, 0, 0, 97, 0, 106, 0, 117, 0, 0, 0, 97, 0, 106, 0, 119, 0, 0, 0, 97, 0, 106, 0, 122, 0, 0, 0, 97, 0, 108, 0, 110, 0, 0, 0, 97, 0, 108, 0, 115, 0, 0, 0, 97, 0, 111, 0, 120, 0, 0, 0, 97, 0, 111, 0, 122, 0, 0, 0, 97, 0, 112, 0, 99, 0, 0, 0, 97, 0, 112, 0, 100, 0, 0, 0, 97, 0, 113, 0, 97, 0, 0, 0, 97, 0, 113, 0, 103, 0, 0, 0, 97, 0, 113, 0, 112, 0, 0, 0, 97, 0, 113, 0, 114, 0, 0, 0, 97, 0, 113, 0, 116, 0, 0, 0, 97, 0, 113, 0, 122, 0, 0, 0, 97, 0, 114, 0, 102, 0, 0, 0, 97, 0, 114, 0, 103, 0, 0, 0, 97, 0, 114, 0, 104, 0, 0, 0, 97, 0, 114, 0, 106, 0, 0, 0, 97, 0, 114, 0, 108, 0, 0, 0, 97, 0, 114, 0, 112, 0, 0, 0, 97, 0, 114, 0, 114, 0, 0, 0, 97, 0, 114, 0, 117, 0, 0, 0, 97, 0, 114, 0, 118, 0, 0, 0, 97, 0, 114, 0, 119, 0, 0, 0, 97, 0, 114, 0, 120, 0, 0, 0, 97, 0, 115, 0, 100, 0, 0, 0, 97, 0, 116, 0, 49, 0, 0, 0, 97, 0, 116, 0, 50, 0, 0, 0, 97, 0, 116, 0, 51, 0, 0, 0, 97, 0, 116, 0, 52, 0, 0, 0, 97, 0, 116, 0, 53, 0, 0, 0, 97, 0, 116, 0, 54, 0, 0, 0, 97, 0, 116, 0, 55, 0, 0, 0, 97, 0, 116, 0, 56, 0, 0, 0, 97, 0, 116, 0, 57, 0, 0, 0, 97, 0, 117, 0, 101, 0, 0, 0, 97, 0, 117, 0, 122, 0, 0, 0, 97, 0, 118, 0, 98, 0, 0, 0, 97, 0, 118, 0, 100, 0, 0, 0, 97, 0, 118, 0, 108, 0, 0, 0, 97, 0, 119, 0, 107, 0, 0, 0, 97, 0, 120, 0, 98, 0, 0, 0, 97, 0, 120, 0, 103, 0, 0, 0, 97, 0, 120, 0, 120, 0, 0, 0, 97, 0, 121, 0, 104, 0, 0, 0, 97, 0, 121, 0, 108, 0, 0, 0, 97, 0, 121, 0, 110, 0, 0, 0, 97, 0, 121, 0, 112, 0, 0, 0, 97, 0, 121, 0, 114, 0, 0, 0, 97, 0, 121, 0, 120, 0, 0, 0, 97, 0, 122, 0, 98, 0, 0, 0, 97, 0, 122, 0, 103, 0, 0, 0, 97, 0, 122, 0, 106, 0, 0, 0, 97, 0, 122, 0, 116, 0, 0, 0, 97, 0, 122, 0, 122, 0, 0, 0, 98, 0, 98, 0, 122, 0, 0, 0, 98, 0, 99, 0, 99, 0, 0, 0, 98, 0, 99, 0, 108, 0, 0, 0, 98, 0, 100, 0, 97, 0, 0, 0, 98, 0, 100, 0, 98, 0, 0, 0, 98, 0, 100, 0, 99, 0, 0, 0, 98, 0, 100, 0, 100, 0, 0, 0, 98, 0, 100, 0, 101, 0, 0, 0, 98, 0, 100, 0, 104, 0, 0, 0, 98, 0, 103, 0, 109, 0, 0, 0, 98, 0, 104, 0, 107, 0, 0, 0, 98, 0, 104, 0, 111, 0, 0, 0, 98, 0, 104, 0, 114, 0, 0, 0, 98, 0, 106, 0, 100, 0, 0, 0, 98, 0, 106, 0, 110, 0, 0, 0, 98, 0, 107, 0, 97, 0, 0, 0, 98, 0, 107, 0, 98, 0, 0, 0, 98, 0, 109, 0, 109, 0, 0, 0, 98, 0, 109, 0, 121, 0, 0, 0, 98, 0, 109, 0, 122, 0, 0, 0, 98, 0, 111, 0, 98, 0, 0, 0, 98, 0, 111, 0, 99, 0, 0, 0, 98, 0, 111, 0, 110, 0, 0, 0, 98, 0, 111, 0, 111, 0, 0, 0, 98, 0, 111, 0, 115, 0, 0, 0, 98, 0, 111, 0, 116, 0, 0, 0, 98, 0, 112, 0, 100, 0, 0, 0, 98, 0, 116, 0, 97, 0, 0, 0, 98, 0, 116, 0, 98, 0, 0, 0, 98, 0, 116, 0, 106, 0, 0, 0, 98, 0, 116, 0, 108, 0, 0, 0, 98, 0, 118, 0, 101, 0, 0, 0, 98, 0, 118, 0, 117, 0, 0, 0, 98, 0, 120, 0, 107, 0, 0, 0, 98, 0, 120, 0, 114, 0, 0, 0, 98, 0, 120, 0, 115, 0, 0, 0, 98, 0, 120, 0, 120, 0, 0, 0, 98, 0, 120, 0, 122, 0, 0, 0, 98, 0, 121, 0, 121, 0, 0, 0, 98, 0, 121, 0, 122, 0, 0, 0, 98, 0, 122, 0, 99, 0, 0, 0, 99, 0, 98, 0, 101, 0, 0, 0, 99, 0, 98, 0, 103, 0, 0, 0, 99, 0, 98, 0, 104, 0, 0, 0, 99, 0, 98, 0, 121, 0, 0, 0, 99, 0, 99, 0, 97, 0, 0, 0, 99, 0, 99, 0, 106, 0, 0, 0, 99, 0, 99, 0, 113, 0, 0, 0, 99, 0, 100, 0, 97, 0, 0, 0, 99, 0, 100, 0, 111, 0, 0, 0, 99, 0, 101, 0, 103, 0, 0, 0, 99, 0, 101, 0, 110, 0, 0, 0, 99, 0, 101, 0, 116, 0, 0, 0, 99, 0, 101, 0, 121, 0, 0, 0, 99, 0, 102, 0, 97, 0, 0, 0, 99, 0, 102, 0, 100, 0, 0, 0, 99, 0, 102, 0, 103, 0, 0, 0, 99, 0, 102, 0, 109, 0, 0, 0, 99, 0, 103, 0, 55, 0, 0, 0, 99, 0, 103, 0, 56, 0, 0, 0, 99, 0, 103, 0, 57, 0, 0, 0, 99, 0, 103, 0, 99, 0, 0, 0, 99, 0, 103, 0, 103, 0, 0, 0, 99, 0, 103, 0, 107, 0, 0, 0, 99, 0, 105, 0, 104, 0, 0, 0, 99, 0, 105, 0, 107, 0, 0, 0, 99, 0, 105, 0, 112, 0, 0, 0, 99, 0, 105, 0, 114, 0, 0, 0, 99, 0, 105, 0, 119, 0, 0, 0, 99, 0, 105, 0, 121, 0, 0, 0, 99, 0, 106, 0, 97, 0, 0, 0, 99, 0, 106, 0, 107, 0, 0, 0, 99, 0, 106, 0, 114, 0, 0, 0, 99, 0, 106, 0, 115, 0, 0, 0, 99, 0, 107, 0, 97, 0, 0, 0, 99, 0, 107, 0, 104, 0, 0, 0, 99, 0, 107, 0, 108, 0, 0, 0, 99, 0, 108, 0, 99, 0, 0, 0, 99, 0, 108, 0, 101, 0, 0, 0, 99, 0, 108, 0, 111, 0, 0, 0, 99, 0, 108, 0, 119, 0, 0, 0, 99, 0, 108, 0, 121, 0, 0, 0, 99, 0, 109, 0, 101, 0, 0, 0, 99, 0, 109, 0, 105, 0, 0, 0, 99, 0, 109, 0, 107, 0, 0, 0, 99, 0, 110, 0, 114, 0, 0, 0, 99, 0, 111, 0, 121, 0, 0, 0, 99, 0, 111, 0, 122, 0, 0, 0, 99, 0, 112, 0, 105, 0, 0, 0, 99, 0, 112, 0, 115, 0, 0, 0, 99, 0, 112, 0, 117, 0, 0, 0, 99, 0, 112, 0, 120, 0, 0, 0, 99, 0, 113, 0, 100, 0, 0, 0, 99, 0, 113, 0, 117, 0, 0, 0, 99, 0, 114, 0, 103, 0, 0, 0, 99, 0, 114, 0, 104, 0, 0, 0, 99, 0, 116, 0, 97, 0, 0, 0, 99, 0, 116, 0, 122, 0, 0, 0, 99, 0, 117, 0, 109, 0, 0, 0, 99, 0, 118, 0, 103, 0, 0, 0, 99, 0, 118, 0, 110, 0, 0, 0, 99, 0, 119, 0, 100, 0, 0, 0, 99, 0, 119, 0, 101, 0, 0, 0, 99, 0, 119, 0, 103, 0, 0, 0, 99, 0, 119, 0, 116, 0, 0, 0, 99, 0, 121, 0, 111, 0, 0, 0, 99, 0, 122, 0, 107, 0, 0, 0, 99, 0, 122, 0, 111, 0, 0, 0, 99, 0, 122, 0, 116, 0, 0, 0, 100, 0, 97, 0, 97, 0, 0, 0, 100, 0, 97, 0, 102, 0, 0, 0, 100, 0, 97, 0, 112, 0, 0, 0, 100, 0, 99, 0, 99, 0, 0, 0, 100, 0, 100, 0, 97, 0, 0, 0, 100, 0, 100, 0, 103, 0, 0, 0, 100, 0, 101, 0, 122, 0, 0, 0, 100, 0, 104, 0, 100, 0, 0, 0, 100, 0, 104, 0, 103, 0, 0, 0, 100, 0, 105, 0, 107, 0, 0, 0, 100, 0, 105, 0, 113, 0, 0, 0, 100, 0, 105, 0, 117, 0, 0, 0, 100, 0, 106, 0, 108, 0, 0, 0, 100, 0, 106, 0, 114, 0, 0, 0, 100, 0, 106, 0, 117, 0, 0, 0, 100, 0, 107, 0, 107, 0, 0, 0, 100, 0, 107, 0, 120, 0, 0, 0, 100, 0, 108, 0, 103, 0, 0, 0, 100, 0, 108, 0, 107, 0, 0, 0, 100, 0, 109, 0, 103, 0, 0, 0, 100, 0, 110, 0, 97, 0, 0, 0, 100, 0, 110, 0, 103, 0, 0, 0, 100, 0, 110, 0, 121, 0, 0, 0, 100, 0, 114, 0, 103, 0, 0, 0, 100, 0, 114, 0, 104, 0, 0, 0, 100, 0, 114, 0, 108, 0, 0, 0, 100, 0, 114, 0, 119, 0, 0, 0, 100, 0, 115, 0, 101, 0, 0, 0, 100, 0, 115, 0, 113, 0, 0, 0, 100, 0, 116, 0, 100, 0, 0, 0, 100, 0, 116, 0, 107, 0, 0, 0, 100, 0, 117, 0, 106, 0, 0, 0, 100, 0, 117, 0, 112, 0, 0, 0, 100, 0, 118, 0, 97, 0, 0, 0, 100, 0, 119, 0, 97, 0, 0, 0, 100, 0, 119, 0, 108, 0, 0, 0, 100, 0, 119, 0, 117, 0, 0, 0, 100, 0, 119, 0, 119, 0, 0, 0, 100, 0, 121, 0, 100, 0, 0, 0, 100, 0, 121, 0, 103, 0, 0, 0, 100, 0, 121, 0, 105, 0, 0, 0, 100, 0, 121, 0, 117, 0, 0, 0, 100, 0, 121, 0, 121, 0, 0, 0, 100, 0, 122, 0, 97, 0, 0, 0, 100, 0, 122, 0, 100, 0, 0, 0, 100, 0, 122, 0, 101, 0, 0, 0, 100, 0, 122, 0, 103, 0, 0, 0, 100, 0, 122, 0, 108, 0, 0, 0, 100, 0, 122, 0, 110, 0, 0, 0, 101, 0, 97, 0, 97, 0, 0, 0, 101, 0, 98, 0, 103, 0, 0, 0, 101, 0, 98, 0, 107, 0, 0, 0, 101, 0, 98, 0, 117, 0, 0, 0, 101, 0, 99, 0, 97, 0, 0, 0, 101, 0, 99, 0, 98, 0, 0, 0, 101, 0, 99, 0, 99, 0, 0, 0, 101, 0, 99, 0, 100, 0, 0, 0, 101, 0, 99, 0, 102, 0, 0, 0, 101, 0, 99, 0, 103, 0, 0, 0, 101, 0, 99, 0, 104, 0, 0, 0, 101, 0, 99, 0, 105, 0, 0, 0, 101, 0, 99, 0, 108, 0, 0, 0, 101, 0, 99, 0, 109, 0, 0, 0, 101, 0, 99, 0, 112, 0, 0, 0, 101, 0, 99, 0, 114, 0, 0, 0, 101, 0, 99, 0, 115, 0, 0, 0, 101, 0, 99, 0, 116, 0, 0, 0, 101, 0, 99, 0, 117, 0, 0, 0, 101, 0, 99, 0, 120, 0, 0, 0, 101, 0, 99, 0, 122, 0, 0, 0, 101, 0, 101, 0, 101, 0, 0, 0, 101, 0, 102, 0, 97, 0, 0, 0, 101, 0, 102, 0, 101, 0, 0, 0, 101, 0, 102, 0, 105, 0, 0, 0, 101, 0, 103, 0, 97, 0, 0, 0, 101, 0, 103, 0, 108, 0, 0, 0, 101, 0, 103, 0, 111, 0, 0, 0, 101, 0, 105, 0, 112, 0, 0, 0, 101, 0, 105, 0, 118, 0, 0, 0, 101, 0, 107, 0, 99, 0, 0, 0, 101, 0, 107, 0, 103, 0, 0, 0, 101, 0, 107, 0, 105, 0, 0, 0, 101, 0, 107, 0, 114, 0, 0, 0, 101, 0, 108, 0, 101, 0, 0, 0, 101, 0, 108, 0, 111, 0, 0, 0, 101, 0, 108, 0, 112, 0, 0, 0, 101, 0, 108, 0, 120, 0, 0, 0, 101, 0, 109, 0, 101, 0, 0, 0, 101, 0, 109, 0, 103, 0, 0, 0, 101, 0, 109, 0, 105, 0, 0, 0, 101, 0, 109, 0, 107, 0, 0, 0, 101, 0, 109, 0, 112, 0, 0, 0, 101, 0, 109, 0, 115, 0, 0, 0, 101, 0, 109, 0, 117, 0, 0, 0, 101, 0, 110, 0, 98, 0, 0, 0, 101, 0, 110, 0, 104, 0, 0, 0, 101, 0, 111, 0, 116, 0, 0, 0, 101, 0, 112, 0, 105, 0, 0, 0, 101, 0, 115, 0, 113, 0, 0, 0, 101, 0, 116, 0, 120, 0, 0, 0, 101, 0, 116, 0, 122, 0, 0, 0, 101, 0, 118, 0, 101, 0, 0, 0, 101, 0, 118, 0, 104, 0, 0, 0, 101, 0, 118, 0, 110, 0, 0, 0, 101, 0, 119, 0, 111, 0, 0, 0, 101, 0, 120, 0, 116, 0, 0, 0, 101, 0, 121, 0, 97, 0, 0, 0, 101, 0, 122, 0, 97, 0, 0, 0, 101, 0, 122, 0, 101, 0, 0, 0, 102, 0, 97, 0, 100, 0, 0, 0, 102, 0, 97, 0, 112, 0, 0, 0, 102, 0, 97, 0, 114, 0, 0, 0, 102, 0, 97, 0, 116, 0, 0, 0, 102, 0, 97, 0, 117, 0, 0, 0, 102, 0, 98, 0, 108, 0, 0, 0, 102, 0, 99, 0, 115, 0, 0, 0, 102, 0, 102, 0, 105, 0, 0, 0, 102, 0, 103, 0, 114, 0, 0, 0, 102, 0, 105, 0, 97, 0, 0, 0, 102, 0, 105, 0, 101, 0, 0, 0, 102, 0, 105, 0, 112, 0, 0, 0, 102, 0, 105, 0, 114, 0, 0, 0, 102, 0, 105, 0, 119, 0, 0, 0, 102, 0, 107, 0, 107, 0, 0, 0, 102, 0, 107, 0, 118, 0, 0, 0, 102, 0, 108, 0, 97, 0, 0, 0, 102, 0, 108, 0, 108, 0, 0, 0, 102, 0, 108, 0, 114, 0, 0, 0, 102, 0, 108, 0, 121, 0, 0, 0, 102, 0, 109, 0, 117, 0, 0, 0, 102, 0, 110, 0, 98, 0, 0, 0, 102, 0, 110, 0, 105, 0, 0, 0, 102, 0, 111, 0, 100, 0, 0, 0, 102, 0, 111, 0, 105, 0, 0, 0, 102, 0, 111, 0, 120, 0, 0, 0, 102, 0, 112, 0, 101, 0, 0, 0, 102, 0, 113, 0, 115, 0, 0, 0, 102, 0, 114, 0, 99, 0, 0, 0, 102, 0, 114, 0, 107, 0, 0, 0, 102, 0, 114, 0, 109, 0, 0, 0, 102, 0, 114, 0, 110, 0, 0, 0, 102, 0, 115, 0, 108, 0, 0, 0, 102, 0, 115, 0, 115, 0, 0, 0, 102, 0, 117, 0, 99, 0, 0, 0, 102, 0, 117, 0, 101, 0, 0, 0, 102, 0, 117, 0, 102, 0, 0, 0, 102, 0, 117, 0, 104, 0, 0, 0, 102, 0, 117, 0, 105, 0, 0, 0, 102, 0, 117, 0, 113, 0, 0, 0, 102, 0, 117, 0, 121, 0, 0, 0, 102, 0, 118, 0, 114, 0, 0, 0, 102, 0, 119, 0, 97, 0, 0, 0, 102, 0, 119, 0, 101, 0, 0, 0, 103, 0, 97, 0, 49, 0, 0, 0, 103, 0, 97, 0, 50, 0, 0, 0, 103, 0, 97, 0, 51, 0, 0, 0, 103, 0, 97, 0, 52, 0, 0, 0, 103, 0, 97, 0, 53, 0, 0, 0, 103, 0, 97, 0, 54, 0, 0, 0, 103, 0, 97, 0, 55, 0, 0, 0, 103, 0, 97, 0, 56, 0, 0, 0, 103, 0, 97, 0, 57, 0, 0, 0, 103, 0, 97, 0, 118, 0, 0, 0, 103, 0, 97, 0, 120, 0, 0, 0, 103, 0, 97, 0, 122, 0, 0, 0, 103, 0, 98, 0, 99, 0, 0, 0, 103, 0, 99, 0, 108, 0, 0, 0, 103, 0, 99, 0, 110, 0, 0, 0, 103, 0, 99, 0, 116, 0, 0, 0, 103, 0, 100, 0, 120, 0, 0, 0, 103, 0, 101, 0, 113, 0, 0, 0, 103, 0, 102, 0, 107, 0, 0, 0, 103, 0, 102, 0, 116, 0, 0, 0, 103, 0, 102, 0, 120, 0, 0, 0, 103, 0, 103, 0, 103, 0, 0, 0, 103, 0, 103, 0, 114, 0, 0, 0, 103, 0, 103, 0, 119, 0, 0, 0, 103, 0, 104, 0, 99, 0, 0, 0, 103, 0, 104, 0, 101, 0, 0, 0, 103, 0, 104, 0, 104, 0, 0, 0, 103, 0, 105, 0, 111, 0, 0, 0, 103, 0, 106, 0, 107, 0, 0, 0, 103, 0, 106, 0, 114, 0, 0, 0, 103, 0, 106, 0, 117, 0, 0, 0, 103, 0, 107, 0, 97, 0, 0, 0, 103, 0, 107, 0, 112, 0, 0, 0, 103, 0, 107, 0, 117, 0, 0, 0, 103, 0, 108, 0, 111, 0, 0, 0, 103, 0, 108, 0, 114, 0, 0, 0, 103, 0, 108, 0, 117, 0, 0, 0, 103, 0, 108, 0, 119, 0, 0, 0, 103, 0, 108, 0, 121, 0, 0, 0, 103, 0, 109, 0, 108, 0, 0, 0, 103, 0, 109, 0, 109, 0, 0, 0, 103, 0, 109, 0, 113, 0, 0, 0, 103, 0, 110, 0, 111, 0, 0, 0, 103, 0, 110, 0, 122, 0, 0, 0, 103, 0, 111, 0, 109, 0, 0, 0, 103, 0, 112, 0, 97, 0, 0, 0, 103, 0, 112, 0, 101, 0, 0, 0, 103, 0, 112, 0, 110, 0, 0, 0, 103, 0, 113, 0, 97, 0, 0, 0, 103, 0, 113, 0, 110, 0, 0, 0, 103, 0, 113, 0, 114, 0, 0, 0, 103, 0, 113, 0, 117, 0, 0, 0, 103, 0, 114, 0, 97, 0, 0, 0, 103, 0, 114, 0, 98, 0, 0, 0, 103, 0, 114, 0, 99, 0, 0, 0, 103, 0, 114, 0, 100, 0, 0, 0, 103, 0, 114, 0, 102, 0, 0, 0, 103, 0, 114, 0, 103, 0, 0, 0, 103, 0, 114, 0, 104, 0, 0, 0, 103, 0, 114, 0, 106, 0, 0, 0, 103, 0, 114, 0, 107, 0, 0, 0, 103, 0, 114, 0, 108, 0, 0, 0, 103, 0, 115, 0, 101, 0, 0, 0, 103, 0, 115, 0, 103, 0, 0, 0, 103, 0, 115, 0, 115, 0, 0, 0, 103, 0, 116, 0, 105, 0, 0, 0, 103, 0, 116, 0, 117, 0, 0, 0, 103, 0, 117, 0, 103, 0, 0, 0, 103, 0, 117, 0, 110, 0, 0, 0, 103, 0, 117, 0, 122, 0, 0, 0, 103, 0, 118, 0, 97, 0, 0, 0, 103, 0, 118, 0, 99, 0, 0, 0, 103, 0, 118, 0, 106, 0, 0, 0, 103, 0, 118, 0, 121, 0, 0, 0, 103, 0, 119, 0, 114, 0, 0, 0, 103, 0, 120, 0, 120, 0, 0, 0, 103, 0, 121, 0, 97, 0, 0, 0, 103, 0, 121, 0, 98, 0, 0, 0, 103, 0, 121, 0, 105, 0, 0, 0, 103, 0, 121, 0, 114, 0, 0, 0, 103, 0, 121, 0, 121, 0, 0, 0, 103, 0, 122, 0, 105, 0, 0, 0, 103, 0, 122, 0, 110, 0, 0, 0, 104, 0, 97, 0, 101, 0, 0, 0, 104, 0, 97, 0, 107, 0, 0, 0, 104, 0, 98, 0, 117, 0, 0, 0, 104, 0, 99, 0, 97, 0, 0, 0, 104, 0, 99, 0, 104, 0, 0, 0, 104, 0, 100, 0, 110, 0, 0, 0, 104, 0, 100, 0, 115, 0, 0, 0, 104, 0, 100, 0, 121, 0, 0, 0, 104, 0, 101, 0, 97, 0, 0, 0, 104, 0, 101, 0, 100, 0, 0, 0, 104, 0, 101, 0, 109, 0, 0, 0, 104, 0, 103, 0, 109, 0, 0, 0, 104, 0, 103, 0, 119, 0, 0, 0, 104, 0, 104, 0, 105, 0, 0, 0, 104, 0, 104, 0, 114, 0, 0, 0, 104, 0, 104, 0, 121, 0, 0, 0, 104, 0, 105, 0, 109, 0, 0, 0, 104, 0, 105, 0, 111, 0, 0, 0, 104, 0, 105, 0, 114, 0, 0, 0, 104, 0, 105, 0, 116, 0, 0, 0, 104, 0, 107, 0, 97, 0, 0, 0, 104, 0, 107, 0, 101, 0, 0, 0, 104, 0, 107, 0, 107, 0, 0, 0, 104, 0, 107, 0, 110, 0, 0, 0, 104, 0, 110, 0, 97, 0, 0, 0, 104, 0, 110, 0, 117, 0, 0, 0, 104, 0, 112, 0, 111, 0, 0, 0, 104, 0, 112, 0, 115, 0, 0, 0, 104, 0, 114, 0, 97, 0, 0, 0, 104, 0, 114, 0, 99, 0, 0, 0, 104, 0, 114, 0, 101, 0, 0, 0, 104, 0, 114, 0, 107, 0, 0, 0, 104, 0, 114, 0, 109, 0, 0, 0, 104, 0, 114, 0, 114, 0, 0, 0, 104, 0, 114, 0, 122, 0, 0, 0, 104, 0, 115, 0, 98, 0, 0, 0, 104, 0, 115, 0, 108, 0, 0, 0, 104, 0, 115, 0, 115, 0, 0, 0, 104, 0, 116, 0, 111, 0, 0, 0, 104, 0, 116, 0, 115, 0, 0, 0, 104, 0, 116, 0, 117, 0, 0, 0, 104, 0, 116, 0, 120, 0, 0, 0, 104, 0, 118, 0, 99, 0, 0, 0, 104, 0, 118, 0, 101, 0, 0, 0, 104, 0, 118, 0, 107, 0, 0, 0, 104, 0, 118, 0, 110, 0, 0, 0, 104, 0, 118, 0, 118, 0, 0, 0, 104, 0, 119, 0, 97, 0, 0, 0, 104, 0, 119, 0, 99, 0, 0, 0, 104, 0, 119, 0, 111, 0, 0, 0, 104, 0, 121, 0, 97, 0, 0, 0, 105, 0, 97, 0, 105, 0, 0, 0, 105, 0, 97, 0, 112, 0, 0, 0, 105, 0, 97, 0, 114, 0, 0, 0, 105, 0, 98, 0, 105, 0, 0, 0, 105, 0, 98, 0, 117, 0, 0, 0, 105, 0, 98, 0, 121, 0, 0, 0, 105, 0, 99, 0, 104, 0, 0, 0, 105, 0, 99, 0, 108, 0, 0, 0, 105, 0, 99, 0, 114, 0, 0, 0, 105, 0, 100, 0, 97, 0, 0, 0, 105, 0, 100, 0, 105, 0, 0, 0, 105, 0, 102, 0, 107, 0, 0, 0, 105, 0, 102, 0, 109, 0, 0, 0, 105, 0, 103, 0, 98, 0, 0, 0, 105, 0, 103, 0, 101, 0, 0, 0, 105, 0, 103, 0, 103, 0, 0, 0, 105, 0, 103, 0, 115, 0, 0, 0, 105, 0, 103, 0, 119, 0, 0, 0, 105, 0, 104, 0, 98, 0, 0, 0, 105, 0, 104, 0, 105, 0, 0, 0, 105, 0, 104, 0, 112, 0, 0, 0, 105, 0, 104, 0, 119, 0, 0, 0, 105, 0, 105, 0, 110, 0, 0, 0, 105, 0, 105, 0, 114, 0, 0, 0, 105, 0, 106, 0, 99, 0, 0, 0, 105, 0, 106, 0, 101, 0, 0, 0, 105, 0, 106, 0, 106, 0, 0, 0, 105, 0, 106, 0, 115, 0, 0, 0, 105, 0, 107, 0, 101, 0, 0, 0, 105, 0, 107, 0, 116, 0, 0, 0, 105, 0, 107, 0, 122, 0, 0, 0, 105, 0, 108, 0, 107, 0, 0, 0, 105, 0, 108, 0, 108, 0, 0, 0, 105, 0, 108, 0, 119, 0, 0, 0, 105, 0, 109, 0, 101, 0, 0, 0, 105, 0, 109, 0, 108, 0, 0, 0, 105, 0, 110, 0, 106, 0, 0, 0, 105, 0, 110, 0, 122, 0, 0, 0, 105, 0, 111, 0, 114, 0, 0, 0, 105, 0, 111, 0, 117, 0, 0, 0, 105, 0, 112, 0, 105, 0, 0, 0, 105, 0, 112, 0, 111, 0, 0, 0, 105, 0, 113, 0, 117, 0, 0, 0, 105, 0, 113, 0, 119, 0, 0, 0, 105, 0, 114, 0, 101, 0, 0, 0, 105, 0, 114, 0, 114, 0, 0, 0, 105, 0, 114, 0, 117, 0, 0, 0, 105, 0, 115, 0, 49, 0, 0, 0, 105, 0, 115, 0, 50, 0, 0, 0, 105, 0, 115, 0, 51, 0, 0, 0, 105, 0, 115, 0, 52, 0, 0, 0, 105, 0, 115, 0, 53, 0, 0, 0, 105, 0, 115, 0, 54, 0, 0, 0, 105, 0, 115, 0, 55, 0, 0, 0, 105, 0, 115, 0, 56, 0, 0, 0, 105, 0, 115, 0, 107, 0, 0, 0, 105, 0, 115, 0, 114, 0, 0, 0, 105, 0, 116, 0, 105, 0, 0, 0, 105, 0, 116, 0, 111, 0, 0, 0, 105, 0, 117, 0, 109, 0, 0, 0, 105, 0, 118, 0, 118, 0, 0, 0, 105, 0, 119, 0, 107, 0, 0, 0, 105, 0, 119, 0, 109, 0, 0, 0, 105, 0, 119, 0, 111, 0, 0, 0, 105, 0, 119, 0, 115, 0, 0, 0, 105, 0, 120, 0, 99, 0, 0, 0, 105, 0, 120, 0, 108, 0, 0, 0, 105, 0, 121, 0, 97, 0, 0, 0, 105, 0, 121, 0, 111, 0, 0, 0, 105, 0, 121, 0, 120, 0, 0, 0, 105, 0, 122, 0, 104, 0, 0, 0, 105, 0, 122, 0, 105, 0, 0, 0, 105, 0, 122, 0, 114, 0, 0, 0, 106, 0, 97, 0, 104, 0, 0, 0, 106, 0, 97, 0, 120, 0, 0, 0, 106, 0, 98, 0, 101, 0, 0, 0, 106, 0, 98, 0, 114, 0, 0, 0, 106, 0, 98, 0, 119, 0, 0, 0, 106, 0, 100, 0, 97, 0, 0, 0, 106, 0, 100, 0, 103, 0, 0, 0, 106, 0, 100, 0, 116, 0, 0, 0, 106, 0, 101, 0, 98, 0, 0, 0, 106, 0, 101, 0, 101, 0, 0, 0, 106, 0, 101, 0, 103, 0, 0, 0, 106, 0, 101, 0, 114, 0, 0, 0, 106, 0, 103, 0, 101, 0, 0, 0, 106, 0, 103, 0, 107, 0, 0, 0, 106, 0, 103, 0, 111, 0, 0, 0, 106, 0, 104, 0, 105, 0, 0, 0, 106, 0, 104, 0, 115, 0, 0, 0, 106, 0, 105, 0, 113, 0, 0, 0, 106, 0, 105, 0, 121, 0, 0, 0, 106, 0, 106, 0, 101, 0, 0, 0, 106, 0, 106, 0, 114, 0, 0, 0, 106, 0, 107, 0, 97, 0, 0, 0, 106, 0, 107, 0, 109, 0, 0, 0, 106, 0, 107, 0, 114, 0, 0, 0, 106, 0, 107, 0, 117, 0, 0, 0, 106, 0, 108, 0, 101, 0, 0, 0, 106, 0, 108, 0, 115, 0, 0, 0, 106, 0, 109, 0, 105, 0, 0, 0, 106, 0, 109, 0, 108, 0, 0, 0, 106, 0, 109, 0, 110, 0, 0, 0, 106, 0, 110, 0, 97, 0, 0, 0, 106, 0, 110, 0, 100, 0, 0, 0, 106, 0, 110, 0, 108, 0, 0, 0, 106, 0, 110, 0, 115, 0, 0, 0, 106, 0, 111, 0, 100, 0, 0, 0, 106, 0, 111, 0, 103, 0, 0, 0, 106, 0, 112, 0, 97, 0, 0, 0, 106, 0, 112, 0, 114, 0, 0, 0, 106, 0, 112, 0, 120, 0, 0, 0, 106, 0, 113, 0, 114, 0, 0, 0, 106, 0, 114, 0, 114, 0, 0, 0, 106, 0, 115, 0, 108, 0, 0, 0, 106, 0, 117, 0, 119, 0, 0, 0, 106, 0, 117, 0, 121, 0, 0, 0, 106, 0, 118, 0, 100, 0, 0, 0, 106, 0, 118, 0, 110, 0, 0, 0, 106, 0, 119, 0, 105, 0, 0, 0, 106, 0, 121, 0, 97, 0, 0, 0, 106, 0, 121, 0, 101, 0, 0, 0, 106, 0, 121, 0, 121, 0, 0, 0, 107, 0, 98, 0, 102, 0, 0, 0, 107, 0, 100, 0, 118, 0, 0, 0, 107, 0, 103, 0, 99, 0, 0, 0, 107, 0, 103, 0, 104, 0, 0, 0, 107, 0, 103, 0, 111, 0, 0, 0, 107, 0, 104, 0, 51, 0, 0, 0, 107, 0, 104, 0, 52, 0, 0, 0, 107, 0, 104, 0, 53, 0, 0, 0, 107, 0, 104, 0, 54, 0, 0, 0, 107, 0, 104, 0, 55, 0, 0, 0, 107, 0, 104, 0, 56, 0, 0, 0, 107, 0, 104, 0, 57, 0, 0, 0, 107, 0, 104, 0, 107, 0, 0, 0, 107, 0, 109, 0, 114, 0, 0, 0, 107, 0, 110, 0, 99, 0, 0, 0, 107, 0, 110, 0, 103, 0, 0, 0, 107, 0, 111, 0, 97, 0, 0, 0, 107, 0, 111, 0, 107, 0, 0, 0, 107, 0, 111, 0, 120, 0, 0, 0, 107, 0, 112, 0, 101, 0, 0, 0, 107, 0, 112, 0, 112, 0, 0, 0, 107, 0, 114, 0, 109, 0, 0, 0, 107, 0, 116, 0, 114, 0, 0, 0, 107, 0, 118, 0, 98, 0, 0, 0, 107, 0, 118, 0, 114, 0, 0, 0, 107, 0, 118, 0, 115, 0, 0, 0, 107, 0, 119, 0, 113, 0, 0, 0, 107, 0, 120, 0, 100, 0, 0, 0, 107, 0, 120, 0, 101, 0, 0, 0, 107, 0, 120, 0, 102, 0, 0, 0, 107, 0, 122, 0, 104, 0, 0, 0, 107, 0, 122, 0, 106, 0, 0, 0, 107, 0, 122, 0, 116, 0, 0, 0, 108, 0, 97, 0, 117, 0, 0, 0, 108, 0, 98, 0, 107, 0, 0, 0, 108, 0, 99, 0, 102, 0, 0, 0, 108, 0, 99, 0, 104, 0, 0, 0, 108, 0, 99, 0, 115, 0, 0, 0, 108, 0, 100, 0, 100, 0, 0, 0, 108, 0, 101, 0, 103, 0, 0, 0, 108, 0, 102, 0, 97, 0, 0, 0, 108, 0, 102, 0, 110, 0, 0, 0, 108, 0, 103, 0, 122, 0, 0, 0, 108, 0, 104, 0, 112, 0, 0, 0, 108, 0, 105, 0, 105, 0, 0, 0, 108, 0, 105, 0, 119, 0, 0, 0, 108, 0, 106, 0, 97, 0, 0, 0, 108, 0, 106, 0, 101, 0, 0, 0, 108, 0, 106, 0, 105, 0, 0, 0, 108, 0, 106, 0, 108, 0, 0, 0, 108, 0, 106, 0, 112, 0, 0, 0, 108, 0, 107, 0, 98, 0, 0, 0, 108, 0, 107, 0, 115, 0, 0, 0, 108, 0, 107, 0, 121, 0, 0, 0, 108, 0, 108, 0, 111, 0, 0, 0, 108, 0, 108, 0, 115, 0, 0, 0, 108, 0, 108, 0, 120, 0, 0, 0, 108, 0, 110, 0, 115, 0, 0, 0, 108, 0, 110, 0, 117, 0, 0, 0, 108, 0, 110, 0, 122, 0, 0, 0, 108, 0, 112, 0, 97, 0, 0, 0, 108, 0, 112, 0, 101, 0, 0, 0, 108, 0, 114, 0, 97, 0, 0, 0, 108, 0, 114, 0, 99, 0, 0, 0, 108, 0, 114, 0, 103, 0, 0, 0, 108, 0, 114, 0, 114, 0, 0, 0, 108, 0, 114, 0, 116, 0, 0, 0, 108, 0, 114, 0, 118, 0, 0, 0, 108, 0, 114, 0, 122, 0, 0, 0, 108, 0, 115, 0, 99, 0, 0, 0, 108, 0, 115, 0, 100, 0, 0, 0, 108, 0, 115, 0, 101, 0, 0, 0, 108, 0, 115, 0, 102, 0, 0, 0, 108, 0, 115, 0, 106, 0, 0, 0, 108, 0, 115, 0, 118, 0, 0, 0, 108, 0, 115, 0, 121, 0, 0, 0, 108, 0, 116, 0, 99, 0, 0, 0, 108, 0, 116, 0, 103, 0, 0, 0, 108, 0, 116, 0, 111, 0, 0, 0, 108, 0, 116, 0, 117, 0, 0, 0, 108, 0, 117, 0, 100, 0, 0, 0, 108, 0, 117, 0, 103, 0, 0, 0, 108, 0, 117, 0, 108, 0, 0, 0, 108, 0, 118, 0, 97, 0, 0, 0, 108, 0, 118, 0, 107, 0, 0, 0, 108, 0, 118, 0, 117, 0, 0, 0, 108, 0, 119, 0, 97, 0, 0, 0, 108, 0, 119, 0, 101, 0, 0, 0, 108, 0, 119, 0, 103, 0, 0, 0, 108, 0, 119, 0, 111, 0, 0, 0, 108, 0, 119, 0, 119, 0, 0, 0, 108, 0, 121, 0, 97, 0, 0, 0, 108, 0, 121, 0, 103, 0, 0, 0, 108, 0, 121, 0, 110, 0, 0, 0, 108, 0, 122, 0, 108, 0, 0, 0, 108, 0, 122, 0, 110, 0, 0, 0, 108, 0, 122, 0, 122, 0, 0, 0, 109, 0, 97, 0, 120, 0, 0, 0, 109, 0, 97, 0, 122, 0, 0, 0, 109, 0, 101, 0, 103, 0, 0, 0, 109, 0, 101, 0, 111, 0, 0, 0, 109, 0, 102, 0, 97, 0, 0, 0, 109, 0, 102, 0, 98, 0, 0, 0, 109, 0, 103, 0, 116, 0, 0, 0, 109, 0, 103, 0, 117, 0, 0, 0, 109, 0, 103, 0, 120, 0, 0, 0, 109, 0, 104, 0, 104, 0, 0, 0, 109, 0, 104, 0, 114, 0, 0, 0, 109, 0, 105, 0, 110, 0, 0, 0, 109, 0, 105, 0, 115, 0, 0, 0, 109, 0, 106, 0, 97, 0, 0, 0, 109, 0, 108, 0, 50, 0, 0, 0, 109, 0, 108, 0, 51, 0, 0, 0, 109, 0, 108, 0, 52, 0, 0, 0, 109, 0, 108, 0, 53, 0, 0, 0, 109, 0, 108, 0, 54, 0, 0, 0, 109, 0, 108, 0, 55, 0, 0, 0, 109, 0, 108, 0, 56, 0, 0, 0, 109, 0, 108, 0, 57, 0, 0, 0, 109, 0, 108, 0, 100, 0, 0, 0, 109, 0, 108, 0, 122, 0, 0, 0, 109, 0, 110, 0, 107, 0, 0, 0, 109, 0, 110, 0, 116, 0, 0, 0, 109, 0, 111, 0, 97, 0, 0, 0, 109, 0, 111, 0, 102, 0, 0, 0, 109, 0, 111, 0, 109, 0, 0, 0, 109, 0, 113, 0, 103, 0, 0, 0, 109, 0, 115, 0, 105, 0, 0, 0, 109, 0, 115, 0, 116, 0, 0, 0, 109, 0, 117, 0, 105, 0, 0, 0, 109, 0, 117, 0, 112, 0, 0, 0, 109, 0, 118, 0, 102, 0, 0, 0, 109, 0, 119, 0, 100, 0, 0, 0, 109, 0, 119, 0, 106, 0, 0, 0, 109, 0, 119, 0, 122, 0, 0, 0, 109, 0, 121, 0, 100, 0, 0, 0, 109, 0, 121, 0, 105, 0, 0, 0, 109, 0, 121, 0, 113, 0, 0, 0, 109, 0, 121, 0, 116, 0, 0, 0, 109, 0, 122, 0, 98, 0, 0, 0, 109, 0, 122, 0, 112, 0, 0, 0, 109, 0, 122, 0, 113, 0, 0, 0, 109, 0, 122, 0, 115, 0, 0, 0, 109, 0, 122, 0, 116, 0, 0, 0, 110, 0, 98, 0, 120, 0, 0, 0, 110, 0, 98, 0, 121, 0, 0, 0, 110, 0, 99, 0, 120, 0, 0, 0, 110, 0, 99, 0, 122, 0, 0, 0, 110, 0, 101, 0, 49, 0, 0, 0, 110, 0, 101, 0, 50, 0, 0, 0, 110, 0, 101, 0, 51, 0, 0, 0, 110, 0, 101, 0, 52, 0, 0, 0, 110, 0, 101, 0, 53, 0, 0, 0, 110, 0, 101, 0, 54, 0, 0, 0, 110, 0, 101, 0, 55, 0, 0, 0, 110, 0, 101, 0, 56, 0, 0, 0, 110, 0, 102, 0, 100, 0, 0, 0, 110, 0, 102, 0, 108, 0, 0, 0, 110, 0, 102, 0, 117, 0, 0, 0, 110, 0, 104, 0, 100, 0, 0, 0, 110, 0, 105, 0, 113, 0, 0, 0, 110, 0, 106, 0, 100, 0, 0, 0, 110, 0, 107, 0, 122, 0, 0, 0, 110, 0, 108, 0, 99, 0, 0, 0, 110, 0, 108, 0, 113, 0, 0, 0, 110, 0, 108, 0, 114, 0, 0, 0, 110, 0, 110, 0, 115, 0, 0, 0, 110, 0, 111, 0, 97, 0, 0, 0, 110, 0, 111, 0, 111, 0, 0, 0, 110, 0, 112, 0, 49, 0, 0, 0, 110, 0, 112, 0, 50, 0, 0, 0, 110, 0, 112, 0, 51, 0, 0, 0, 110, 0, 112, 0, 52, 0, 0, 0, 110, 0, 112, 0, 53, 0, 0, 0, 110, 0, 112, 0, 108, 0, 0, 0, 110, 0, 112, 0, 115, 0, 0, 0, 110, 0, 112, 0, 117, 0, 0, 0, 110, 0, 113, 0, 103, 0, 0, 0, 110, 0, 113, 0, 113, 0, 0, 0, 110, 0, 113, 0, 121, 0, 0, 0, 110, 0, 114, 0, 105, 0, 0, 0, 110, 0, 114, 0, 112, 0, 0, 0, 110, 0, 114, 0, 114, 0, 0, 0, 110, 0, 114, 0, 120, 0, 0, 0, 110, 0, 114, 0, 122, 0, 0, 0, 110, 0, 116, 0, 109, 0, 0, 0, 110, 0, 116, 0, 115, 0, 0, 0, 110, 0, 118, 0, 104, 0, 0, 0, 110, 0, 118, 0, 109, 0, 0, 0, 110, 0, 118, 0, 111, 0, 0, 0, 110, 0, 119, 0, 101, 0, 0, 0, 110, 0, 119, 0, 103, 0, 0, 0, 110, 0, 119, 0, 105, 0, 0, 0, 110, 0, 119, 0, 111, 0, 0, 0, 110, 0, 119, 0, 114, 0, 0, 0, 110, 0, 120, 0, 97, 0, 0, 0, 110, 0, 120, 0, 103, 0, 0, 0, 110, 0, 120, 0, 105, 0, 0, 0, 110, 0, 120, 0, 117, 0, 0, 0, 110, 0, 120, 0, 120, 0, 0, 0, 110, 0, 121, 0, 100, 0, 0, 0, 110, 0, 122, 0, 100, 0, 0, 0, 110, 0, 122, 0, 105, 0, 0, 0, 110, 0, 122, 0, 107, 0, 0, 0, 110, 0, 122, 0, 109, 0, 0, 0, 110, 0, 122, 0, 110, 0, 0, 0, 111, 0, 97, 0, 97, 0, 0, 0, 111, 0, 97, 0, 99, 0, 0, 0, 111, 0, 97, 0, 118, 0, 0, 0, 111, 0, 98, 0, 105, 0, 0, 0, 111, 0, 98, 0, 111, 0, 0, 0, 111, 0, 99, 0, 97, 0, 0, 0, 111, 0, 99, 0, 104, 0, 0, 0, 111, 0, 99, 0, 111, 0, 0, 0, 111, 0, 99, 0, 117, 0, 0, 0, 111, 0, 100, 0, 97, 0, 0, 0, 111, 0, 100, 0, 107, 0, 0, 0, 111, 0, 102, 0, 111, 0, 0, 0, 111, 0, 102, 0, 115, 0, 0, 0, 111, 0, 102, 0, 117, 0, 0, 0, 111, 0, 103, 0, 103, 0, 0, 0, 111, 0, 105, 0, 97, 0, 0, 0, 111, 0, 105, 0, 110, 0, 0, 0, 111, 0, 106, 0, 103, 0, 0, 0, 111, 0, 106, 0, 112, 0, 0, 0, 111, 0, 107, 0, 105, 0, 0, 0, 111, 0, 107, 0, 120, 0, 0, 0, 111, 0, 108, 0, 97, 0, 0, 0, 111, 0, 108, 0, 107, 0, 0, 0, 111, 0, 108, 0, 111, 0, 0, 0, 111, 0, 108, 0, 114, 0, 0, 0, 111, 0, 109, 0, 101, 0, 0, 0, 111, 0, 109, 0, 103, 0, 0, 0, 111, 0, 109, 0, 105, 0, 0, 0, 111, 0, 111, 0, 100, 0, 0, 0, 111, 0, 111, 0, 103, 0, 0, 0, 111, 0, 112, 0, 97, 0, 0, 0, 111, 0, 112, 0, 107, 0, 0, 0, 111, 0, 112, 0, 109, 0, 0, 0, 111, 0, 112, 0, 121, 0, 0, 0, 111, 0, 114, 0, 110, 0, 0, 0, 111, 0, 114, 0, 115, 0, 0, 0, 111, 0, 114, 0, 121, 0, 0, 0, 111, 0, 114, 0, 122, 0, 0, 0, 111, 0, 115, 0, 99, 0, 0, 0, 111, 0, 115, 0, 105, 0, 0, 0, 111, 0, 115, 0, 120, 0, 0, 0, 111, 0, 116, 0, 105, 0, 0, 0, 111, 0, 117, 0, 101, 0, 0, 0, 111, 0, 117, 0, 105, 0, 0, 0, 111, 0, 117, 0, 109, 0, 0, 0, 111, 0, 117, 0, 110, 0, 0, 0, 111, 0, 118, 0, 100, 0, 0, 0, 111, 0, 119, 0, 105, 0, 0, 0, 111, 0, 119, 0, 108, 0, 0, 0, 111, 0, 121, 0, 98, 0, 0, 0, 111, 0, 121, 0, 100, 0, 0, 0, 111, 0, 121, 0, 109, 0, 0, 0, 111, 0, 121, 0, 121, 0, 0, 0, 111, 0, 122, 0, 109, 0, 0, 0, 112, 0, 97, 0, 50, 0, 0, 0, 112, 0, 97, 0, 51, 0, 0, 0, 112, 0, 97, 0, 52, 0, 0, 0, 112, 0, 97, 0, 53, 0, 0, 0, 112, 0, 97, 0, 54, 0, 0, 0, 112, 0, 97, 0, 55, 0, 0, 0, 112, 0, 97, 0, 56, 0, 0, 0, 112, 0, 97, 0, 57, 0, 0, 0, 112, 0, 98, 0, 116, 0, 0, 0, 112, 0, 98, 0, 117, 0, 0, 0, 112, 0, 98, 0, 118, 0, 0, 0, 112, 0, 98, 0, 121, 0, 0, 0, 112, 0, 98, 0, 122, 0, 0, 0, 112, 0, 99, 0, 112, 0, 0, 0, 112, 0, 99, 0, 114, 0, 0, 0, 112, 0, 99, 0, 119, 0, 0, 0, 112, 0, 100, 0, 97, 0, 0, 0, 112, 0, 100, 0, 99, 0, 0, 0, 112, 0, 100, 0, 105, 0, 0, 0, 112, 0, 101, 0, 108, 0, 0, 0, 112, 0, 101, 0, 115, 0, 0, 0, 112, 0, 101, 0, 118, 0, 0, 0, 112, 0, 102, 0, 97, 0, 0, 0, 112, 0, 102, 0, 101, 0, 0, 0, 112, 0, 102, 0, 108, 0, 0, 0, 112, 0, 103, 0, 100, 0, 0, 0, 112, 0, 103, 0, 103, 0, 0, 0, 112, 0, 103, 0, 105, 0, 0, 0, 112, 0, 103, 0, 110, 0, 0, 0, 112, 0, 103, 0, 115, 0, 0, 0, 112, 0, 103, 0, 117, 0, 0, 0, 112, 0, 103, 0, 121, 0, 0, 0, 112, 0, 103, 0, 122, 0, 0, 0, 112, 0, 104, 0, 97, 0, 0, 0, 112, 0, 104, 0, 100, 0, 0, 0, 112, 0, 106, 0, 116, 0, 0, 0, 112, 0, 108, 0, 116, 0, 0, 0, 112, 0, 109, 0, 99, 0, 0, 0, 112, 0, 109, 0, 117, 0, 0, 0, 112, 0, 110, 0, 98, 0, 0, 0, 112, 0, 111, 0, 99, 0, 0, 0, 112, 0, 111, 0, 100, 0, 0, 0, 112, 0, 111, 0, 107, 0, 0, 0, 112, 0, 112, 0, 97, 0, 0, 0, 112, 0, 112, 0, 101, 0, 0, 0, 112, 0, 112, 0, 105, 0, 0, 0, 112, 0, 112, 0, 114, 0, 0, 0, 112, 0, 113, 0, 97, 0, 0, 0, 112, 0, 113, 0, 101, 0, 0, 0, 112, 0, 113, 0, 109, 0, 0, 0, 112, 0, 113, 0, 119, 0, 0, 0, 112, 0, 114, 0, 98, 0, 0, 0, 112, 0, 114, 0, 115, 0, 0, 0, 112, 0, 114, 0, 121, 0, 0, 0, 112, 0, 114, 0, 122, 0, 0, 0, 112, 0, 115, 0, 116, 0, 0, 0, 112, 0, 115, 0, 119, 0, 0, 0, 112, 0, 115, 0, 121, 0, 0, 0, 112, 0, 116, 0, 97, 0, 0, 0, 112, 0, 116, 0, 121, 0, 0, 0, 112, 0, 117, 0, 107, 0, 0, 0, 112, 0, 117, 0, 109, 0, 0, 0, 112, 0, 117, 0, 122, 0, 0, 0, 112, 0, 119, 0, 103, 0, 0, 0, 112, 0, 119, 0, 105, 0, 0, 0, 112, 0, 119, 0, 114, 0, 0, 0, 112, 0, 119, 0, 119, 0, 0, 0, 112, 0, 120, 0, 109, 0, 0, 0, 112, 0, 121, 0, 50, 0, 0, 0, 112, 0, 121, 0, 51, 0, 0, 0, 112, 0, 121, 0, 52, 0, 0, 0, 112, 0, 121, 0, 53, 0, 0, 0, 112, 0, 121, 0, 54, 0, 0, 0, 112, 0, 121, 0, 55, 0, 0, 0, 112, 0, 121, 0, 56, 0, 0, 0, 112, 0, 121, 0, 57, 0, 0, 0, 112, 0, 121, 0, 101, 0, 0, 0, 112, 0, 121, 0, 115, 0, 0, 0, 112, 0, 121, 0, 117, 0, 0, 0, 112, 0, 122, 0, 110, 0, 0, 0, 113, 0, 102, 0, 122, 0, 0, 0, 113, 0, 117, 0, 98, 0, 0, 0, 113, 0, 117, 0, 100, 0, 0, 0, 113, 0, 117, 0, 102, 0, 0, 0, 113, 0, 117, 0, 103, 0, 0, 0, 113, 0, 117, 0, 104, 0, 0, 0, 113, 0, 117, 0, 107, 0, 0, 0, 113, 0, 117, 0, 108, 0, 0, 0, 113, 0, 117, 0, 112, 0, 0, 0, 113, 0, 117, 0, 114, 0, 0, 0, 113, 0, 117, 0, 119, 0, 0, 0, 113, 0, 117, 0, 120, 0, 0, 0, 113, 0, 117, 0, 121, 0, 0, 0, 113, 0, 117, 0, 122, 0, 0, 0, 113, 0, 118, 0, 104, 0, 0, 0, 113, 0, 118, 0, 105, 0, 0, 0, 113, 0, 118, 0, 106, 0, 0, 0, 113, 0, 118, 0, 108, 0, 0, 0, 113, 0, 118, 0, 109, 0, 0, 0, 113, 0, 118, 0, 110, 0, 0, 0, 113, 0, 118, 0, 111, 0, 0, 0, 113, 0, 118, 0, 112, 0, 0, 0, 113, 0, 118, 0, 122, 0, 0, 0, 113, 0, 119, 0, 101, 0, 0, 0, 113, 0, 119, 0, 109, 0, 0, 0, 113, 0, 119, 0, 115, 0, 0, 0, 113, 0, 120, 0, 110, 0, 0, 0, 113, 0, 120, 0, 111, 0, 0, 0, 113, 0, 120, 0, 112, 0, 0, 0, 113, 0, 120, 0, 114, 0, 0, 0, 113, 0, 120, 0, 116, 0, 0, 0, 113, 0, 120, 0, 117, 0, 0, 0, 113, 0, 121, 0, 97, 0, 0, 0, 113, 0, 121, 0, 112, 0, 0, 0, 114, 0, 97, 0, 103, 0, 0, 0, 114, 0, 98, 0, 98, 0, 0, 0, 114, 0, 98, 0, 112, 0, 0, 0, 114, 0, 99, 0, 102, 0, 0, 0, 114, 0, 101, 0, 103, 0, 0, 0, 114, 0, 101, 0, 121, 0, 0, 0, 114, 0, 103, 0, 97, 0, 0, 0, 114, 0, 103, 0, 101, 0, 0, 0, 114, 0, 103, 0, 110, 0, 0, 0, 114, 0, 103, 0, 117, 0, 0, 0, 114, 0, 104, 0, 103, 0, 0, 0, 114, 0, 104, 0, 112, 0, 0, 0, 114, 0, 105, 0, 97, 0, 0, 0, 114, 0, 105, 0, 102, 0, 0, 0, 114, 0, 105, 0, 114, 0, 0, 0, 114, 0, 106, 0, 103, 0, 0, 0, 114, 0, 106, 0, 105, 0, 0, 0, 114, 0, 106, 0, 115, 0, 0, 0, 114, 0, 107, 0, 109, 0, 0, 0, 114, 0, 107, 0, 119, 0, 0, 0, 114, 0, 109, 0, 114, 0, 0, 0, 114, 0, 109, 0, 122, 0, 0, 0, 114, 0, 110, 0, 97, 0, 0, 0, 114, 0, 110, 0, 100, 0, 0, 0, 114, 0, 110, 0, 103, 0, 0, 0, 114, 0, 110, 0, 108, 0, 0, 0, 114, 0, 110, 0, 110, 0, 0, 0, 114, 0, 110, 0, 112, 0, 0, 0, 114, 0, 110, 0, 114, 0, 0, 0, 114, 0, 110, 0, 119, 0, 0, 0, 114, 0, 111, 0, 99, 0, 0, 0, 114, 0, 111, 0, 114, 0, 0, 0, 114, 0, 111, 0, 117, 0, 0, 0, 114, 0, 111, 0, 119, 0, 0, 0, 114, 0, 112, 0, 110, 0, 0, 0, 114, 0, 112, 0, 116, 0, 0, 0, 114, 0, 114, 0, 116, 0, 0, 0, 114, 0, 115, 0, 98, 0, 0, 0, 114, 0, 116, 0, 99, 0, 0, 0, 114, 0, 116, 0, 104, 0, 0, 0, 114, 0, 116, 0, 109, 0, 0, 0, 114, 0, 116, 0, 115, 0, 0, 0, 114, 0, 116, 0, 119, 0, 0, 0, 114, 0, 117, 0, 107, 0, 0, 0, 114, 0, 119, 0, 107, 0, 0, 0, 114, 0, 119, 0, 109, 0, 0, 0, 114, 0, 119, 0, 111, 0, 0, 0, 114, 0, 119, 0, 114, 0, 0, 0, 114, 0, 120, 0, 100, 0, 0, 0, 114, 0, 120, 0, 119, 0, 0, 0, 114, 0, 121, 0, 110, 0, 0, 0, 114, 0, 121, 0, 115, 0, 0, 0, 114, 0, 121, 0, 117, 0, 0, 0, 114, 0, 122, 0, 104, 0, 0, 0, 115, 0, 97, 0, 111, 0, 0, 0, 115, 0, 100, 0, 110, 0, 0, 0, 115, 0, 100, 0, 120, 0, 0, 0, 115, 0, 101, 0, 100, 0, 0, 0, 115, 0, 101, 0, 104, 0, 0, 0, 115, 0, 101, 0, 105, 0, 0, 0, 115, 0, 101, 0, 109, 0, 0, 0, 115, 0, 101, 0, 111, 0, 0, 0, 115, 0, 101, 0, 115, 0, 0, 0, 115, 0, 101, 0, 117, 0, 0, 0, 115, 0, 101, 0, 121, 0, 0, 0, 115, 0, 101, 0, 122, 0, 0, 0, 115, 0, 102, 0, 98, 0, 0, 0, 115, 0, 102, 0, 101, 0, 0, 0, 115, 0, 102, 0, 109, 0, 0, 0, 115, 0, 102, 0, 115, 0, 0, 0, 115, 0, 102, 0, 119, 0, 0, 0, 115, 0, 103, 0, 108, 0, 0, 0, 115, 0, 103, 0, 111, 0, 0, 0, 115, 0, 103, 0, 112, 0, 0, 0, 115, 0, 106, 0, 119, 0, 0, 0, 115, 0, 107, 0, 103, 0, 0, 0, 115, 0, 107, 0, 107, 0, 0, 0, 115, 0, 112, 0, 103, 0, 0, 0, 115, 0, 112, 0, 105, 0, 0, 0, 115, 0, 112, 0, 120, 0, 0, 0, 115, 0, 112, 0, 121, 0, 0, 0, 115, 0, 113, 0, 97, 0, 0, 0, 115, 0, 113, 0, 104, 0, 0, 0, 115, 0, 114, 0, 111, 0, 0, 0, 115, 0, 115, 0, 104, 0, 0, 0, 115, 0, 117, 0, 101, 0, 0, 0, 115, 0, 117, 0, 103, 0, 0, 0, 115, 0, 118, 0, 107, 0, 0, 0, 115, 0, 118, 0, 109, 0, 0, 0, 115, 0, 118, 0, 114, 0, 0, 0, 115, 0, 118, 0, 120, 0, 0, 0, 115, 0, 119, 0, 98, 0, 0, 0, 115, 0, 119, 0, 99, 0, 0, 0, 115, 0, 119, 0, 104, 0, 0, 0, 115, 0, 120, 0, 101, 0, 0, 0, 115, 0, 120, 0, 103, 0, 0, 0, 115, 0, 120, 0, 117, 0, 0, 0, 115, 0, 120, 0, 119, 0, 0, 0, 115, 0, 121, 0, 105, 0, 0, 0, 115, 0, 121, 0, 114, 0, 0, 0, 115, 0, 122, 0, 103, 0, 0, 0, 115, 0, 122, 0, 108, 0, 0, 0, 115, 0, 122, 0, 110, 0, 0, 0, 115, 0, 122, 0, 112, 0, 0, 0, 115, 0, 122, 0, 115, 0, 0, 0, 115, 0, 122, 0, 121, 0, 0, 0, 116, 0, 100, 0, 117, 0, 0, 0, 116, 0, 100, 0, 120, 0, 0, 0, 116, 0, 101, 0, 99, 0, 0, 0, 116, 0, 101, 0, 107, 0, 0, 0, 116, 0, 102, 0, 116, 0, 0, 0, 116, 0, 103, 0, 103, 0, 0, 0, 116, 0, 104, 0, 99, 0, 0, 0, 116, 0, 104, 0, 120, 0, 0, 0, 116, 0, 105, 0, 97, 0, 0, 0, 116, 0, 106, 0, 103, 0, 0, 0, 116, 0, 106, 0, 115, 0, 0, 0, 116, 0, 106, 0, 119, 0, 0, 0, 116, 0, 107, 0, 103, 0, 0, 0, 116, 0, 107, 0, 107, 0, 0, 0, 116, 0, 107, 0, 122, 0, 0, 0, 116, 0, 108, 0, 119, 0, 0, 0, 116, 0, 109, 0, 97, 0, 0, 0, 116, 0, 109, 0, 109, 0, 0, 0, 116, 0, 109, 0, 112, 0, 0, 0, 116, 0, 109, 0, 119, 0, 0, 0, 116, 0, 111, 0, 101, 0, 0, 0, 116, 0, 113, 0, 98, 0, 0, 0, 116, 0, 113, 0, 119, 0, 0, 0, 116, 0, 115, 0, 102, 0, 0, 0, 116, 0, 116, 0, 113, 0, 0, 0, 116, 0, 117, 0, 121, 0, 0, 0, 116, 0, 120, 0, 101, 0, 0, 0, 116, 0, 120, 0, 121, 0, 0, 0, 116, 0, 121, 0, 97, 0, 0, 0, 116, 0, 121, 0, 101, 0, 0, 0, 116, 0, 121, 0, 108, 0, 0, 0, 116, 0, 121, 0, 110, 0, 0, 0, 116, 0, 121, 0, 112, 0, 0, 0, 116, 0, 121, 0, 120, 0, 0, 0, 116, 0, 121, 0, 122, 0, 0, 0, 116, 0, 122, 0, 104, 0, 0, 0, 116, 0, 122, 0, 106, 0, 0, 0, 116, 0, 122, 0, 120, 0, 0, 0, 117, 0, 97, 0, 114, 0, 0, 0, 117, 0, 98, 0, 121, 0, 0, 0, 117, 0, 100, 0, 103, 0, 0, 0, 117, 0, 104, 0, 97, 0, 0, 0, 117, 0, 104, 0, 110, 0, 0, 0, 117, 0, 105, 0, 118, 0, 0, 0, 117, 0, 106, 0, 105, 0, 0, 0, 117, 0, 107, 0, 115, 0, 0, 0, 117, 0, 107, 0, 119, 0, 0, 0, 117, 0, 107, 0, 121, 0, 0, 0, 117, 0, 108, 0, 105, 0, 0, 0, 117, 0, 108, 0, 119, 0, 0, 0, 117, 0, 109, 0, 103, 0, 0, 0, 117, 0, 109, 0, 117, 0, 0, 0, 117, 0, 110, 0, 97, 0, 0, 0, 117, 0, 110, 0, 101, 0, 0, 0, 117, 0, 110, 0, 112, 0, 0, 0, 117, 0, 110, 0, 117, 0, 0, 0, 117, 0, 110, 0, 120, 0, 0, 0, 117, 0, 110, 0, 122, 0, 0, 0, 117, 0, 111, 0, 107, 0, 0, 0, 117, 0, 112, 0, 118, 0, 0, 0, 117, 0, 114, 0, 107, 0, 0, 0, 117, 0, 115, 0, 112, 0, 0, 0, 117, 0, 115, 0, 117, 0, 0, 0, 117, 0, 116, 0, 104, 0, 0, 0, 117, 0, 116, 0, 112, 0, 0, 0, 117, 0, 116, 0, 117, 0, 0, 0, 117, 0, 117, 0, 114, 0, 0, 0, 117, 0, 117, 0, 117, 0, 0, 0, 117, 0, 118, 0, 104, 0, 0, 0, 117, 0, 118, 0, 108, 0, 0, 0, 117, 0, 121, 0, 97, 0, 0, 0, 117, 0, 122, 0, 110, 0, 0, 0, 118, 0, 97, 0, 121, 0, 0, 0, 118, 0, 98, 0, 98, 0, 0, 0, 118, 0, 98, 0, 107, 0, 0, 0, 118, 0, 101, 0, 97, 0, 0, 0, 118, 0, 101, 0, 98, 0, 0, 0, 118, 0, 101, 0, 99, 0, 0, 0, 118, 0, 101, 0, 100, 0, 0, 0, 118, 0, 101, 0, 101, 0, 0, 0, 118, 0, 101, 0, 102, 0, 0, 0, 118, 0, 101, 0, 103, 0, 0, 0, 118, 0, 101, 0, 104, 0, 0, 0, 118, 0, 101, 0, 105, 0, 0, 0, 118, 0, 101, 0, 106, 0, 0, 0, 118, 0, 101, 0, 107, 0, 0, 0, 118, 0, 101, 0, 108, 0, 0, 0, 118, 0, 101, 0, 109, 0, 0, 0, 118, 0, 101, 0, 112, 0, 0, 0, 118, 0, 101, 0, 115, 0, 0, 0, 118, 0, 101, 0, 116, 0, 0, 0, 118, 0, 101, 0, 118, 0, 0, 0, 118, 0, 101, 0, 119, 0, 0, 0, 118, 0, 101, 0, 120, 0, 0, 0, 118, 0, 101, 0, 121, 0, 0, 0, 118, 0, 101, 0, 122, 0, 0, 0, 118, 0, 103, 0, 114, 0, 0, 0, 118, 0, 103, 0, 116, 0, 0, 0, 118, 0, 105, 0, 118, 0, 0, 0, 118, 0, 107, 0, 97, 0, 0, 0, 118, 0, 107, 0, 107, 0, 0, 0, 118, 0, 107, 0, 116, 0, 0, 0, 118, 0, 108, 0, 112, 0, 0, 0, 118, 0, 108, 0, 115, 0, 0, 0, 118, 0, 110, 0, 107, 0, 0, 0, 118, 0, 110, 0, 109, 0, 0, 0, 118, 0, 110, 0, 112, 0, 0, 0, 118, 0, 111, 0, 116, 0, 0, 0, 118, 0, 114, 0, 97, 0, 0, 0, 118, 0, 115, 0, 105, 0, 0, 0, 118, 0, 116, 0, 111, 0, 0, 0, 118, 0, 117, 0, 116, 0, 0, 0, 118, 0, 119, 0, 97, 0, 0, 0, 119, 0, 99, 0, 105, 0, 0, 0, 119, 0, 100, 0, 100, 0, 0, 0, 119, 0, 100, 0, 103, 0, 0, 0, 119, 0, 100, 0, 121, 0, 0, 0, 119, 0, 101, 0, 97, 0, 0, 0, 119, 0, 101, 0, 119, 0, 0, 0, 119, 0, 102, 0, 103, 0, 0, 0, 119, 0, 103, 0, 103, 0, 0, 0, 119, 0, 103, 0, 105, 0, 0, 0, 119, 0, 103, 0, 111, 0, 0, 0, 119, 0, 103, 0, 117, 0, 0, 0, 119, 0, 103, 0, 119, 0, 0, 0, 119, 0, 103, 0, 121, 0, 0, 0, 119, 0, 104, 0, 103, 0, 0, 0, 119, 0, 104, 0, 107, 0, 0, 0, 119, 0, 105, 0, 114, 0, 0, 0, 119, 0, 105, 0, 119, 0, 0, 0, 119, 0, 105, 0, 121, 0, 0, 0, 119, 0, 106, 0, 105, 0, 0, 0, 119, 0, 107, 0, 100, 0, 0, 0, 119, 0, 107, 0, 121, 0, 0, 0, 119, 0, 108, 0, 97, 0, 0, 0, 119, 0, 108, 0, 99, 0, 0, 0, 119, 0, 108, 0, 101, 0, 0, 0, 119, 0, 110, 0, 103, 0, 0, 0, 119, 0, 111, 0, 119, 0, 0, 0, 119, 0, 111, 0, 121, 0, 0, 0, 119, 0, 112, 0, 99, 0, 0, 0, 119, 0, 114, 0, 100, 0, 0, 0, 119, 0, 115, 0, 103, 0, 0, 0, 119, 0, 115, 0, 105, 0, 0, 0, 119, 0, 116, 0, 102, 0, 0, 0, 119, 0, 116, 0, 107, 0, 0, 0, 119, 0, 116, 0, 109, 0, 0, 0, 119, 0, 116, 0, 119, 0, 0, 0, 119, 0, 117, 0, 100, 0, 0, 0, 119, 0, 117, 0, 104, 0, 0, 0, 119, 0, 117, 0, 114, 0, 0, 0, 119, 0, 117, 0, 117, 0, 0, 0, 119, 0, 119, 0, 111, 0, 0, 0, 119, 0, 119, 0, 114, 0, 0, 0, 119, 0, 119, 0, 119, 0, 0, 0, 119, 0, 120, 0, 97, 0, 0, 0, 119, 0, 120, 0, 119, 0, 0, 0, 119, 0, 121, 0, 109, 0, 0, 0, 119, 0, 121, 0, 114, 0, 0, 0, 119, 0, 121, 0, 121, 0, 0, 0, 120, 0, 97, 0, 103, 0, 0, 0, 120, 0, 98, 0, 97, 0, 0, 0, 120, 0, 98, 0, 114, 0, 0, 0, 120, 0, 98, 0, 119, 0, 0, 0, 120, 0, 98, 0, 120, 0, 0, 0, 120, 0, 98, 0, 121, 0, 0, 0, 120, 0, 99, 0, 121, 0, 0, 0, 120, 0, 100, 0, 97, 0, 0, 0, 120, 0, 100, 0, 99, 0, 0, 0, 120, 0, 100, 0, 107, 0, 0, 0, 120, 0, 100, 0, 109, 0, 0, 0, 120, 0, 100, 0, 111, 0, 0, 0, 120, 0, 100, 0, 121, 0, 0, 0, 120, 0, 101, 0, 98, 0, 0, 0, 120, 0, 101, 0, 100, 0, 0, 0, 120, 0, 101, 0, 103, 0, 0, 0, 120, 0, 101, 0, 112, 0, 0, 0, 120, 0, 102, 0, 97, 0, 0, 0, 120, 0, 103, 0, 105, 0, 0, 0, 120, 0, 103, 0, 114, 0, 0, 0, 120, 0, 103, 0, 117, 0, 0, 0, 120, 0, 103, 0, 119, 0, 0, 0, 120, 0, 104, 0, 97, 0, 0, 0, 120, 0, 104, 0, 114, 0, 0, 0, 120, 0, 105, 0, 97, 0, 0, 0, 120, 0, 105, 0, 98, 0, 0, 0, 120, 0, 105, 0, 108, 0, 0, 0, 120, 0, 105, 0, 110, 0, 0, 0, 120, 0, 105, 0, 112, 0, 0, 0, 120, 0, 105, 0, 118, 0, 0, 0, 120, 0, 105, 0, 121, 0, 0, 0, 120, 0, 106, 0, 98, 0, 0, 0, 120, 0, 106, 0, 116, 0, 0, 0, 120, 0, 107, 0, 104, 0, 0, 0, 120, 0, 108, 0, 103, 0, 0, 0, 120, 0, 108, 0, 105, 0, 0, 0, 120, 0, 108, 0, 115, 0, 0, 0, 120, 0, 108, 0, 117, 0, 0, 0, 120, 0, 108, 0, 121, 0, 0, 0, 120, 0, 109, 0, 109, 0, 0, 0, 120, 0, 109, 0, 118, 0, 0, 0, 120, 0, 109, 0, 119, 0, 0, 0, 120, 0, 110, 0, 107, 0, 0, 0, 120, 0, 111, 0, 103, 0, 0, 0, 120, 0, 111, 0, 105, 0, 0, 0, 120, 0, 111, 0, 107, 0, 0, 0, 120, 0, 111, 0, 114, 0, 0, 0, 120, 0, 111, 0, 119, 0, 0, 0, 120, 0, 112, 0, 97, 0, 0, 0, 120, 0, 112, 0, 99, 0, 0, 0, 120, 0, 112, 0, 101, 0, 0, 0, 120, 0, 112, 0, 103, 0, 0, 0, 120, 0, 112, 0, 121, 0, 0, 0, 120, 0, 113, 0, 97, 0, 0, 0, 120, 0, 113, 0, 116, 0, 0, 0, 120, 0, 114, 0, 103, 0, 0, 0, 120, 0, 114, 0, 105, 0, 0, 0, 120, 0, 114, 0, 119, 0, 0, 0, 120, 0, 115, 0, 106, 0, 0, 0, 120, 0, 115, 0, 108, 0, 0, 0, 120, 0, 115, 0, 121, 0, 0, 0, 120, 0, 117, 0, 100, 0, 0, 0, 120, 0, 117, 0, 106, 0, 0, 0, 120, 0, 118, 0, 105, 0, 0, 0, 120, 0, 118, 0, 115, 0, 0, 0, 120, 0, 119, 0, 97, 0, 0, 0, 120, 0, 119, 0, 103, 0, 0, 0, 120, 0, 119, 0, 114, 0, 0, 0, 120, 0, 119, 0, 116, 0, 0, 0, 120, 0, 119, 0, 119, 0, 0, 0, 120, 0, 120, 0, 98, 0, 0, 0, 120, 0, 120, 0, 107, 0, 0, 0, 120, 0, 120, 0, 109, 0, 0, 0, 120, 0, 120, 0, 114, 0, 0, 0, 120, 0, 120, 0, 116, 0, 0, 0, 120, 0, 121, 0, 116, 0, 0, 0, 120, 0, 121, 0, 121, 0, 0, 0, 120, 0, 122, 0, 109, 0, 0, 0, 120, 0, 122, 0, 112, 0, 0, 0, 121, 0, 98, 0, 100, 0, 0, 0, 121, 0, 98, 0, 101, 0, 0, 0, 121, 0, 99, 0, 104, 0, 0, 0, 121, 0, 99, 0, 110, 0, 0, 0, 121, 0, 99, 0, 112, 0, 0, 0, 121, 0, 100, 0, 100, 0, 0, 0, 121, 0, 100, 0, 103, 0, 0, 0, 121, 0, 100, 0, 107, 0, 0, 0, 121, 0, 100, 0, 115, 0, 0, 0, 121, 0, 101, 0, 97, 0, 0, 0, 121, 0, 101, 0, 99, 0, 0, 0, 121, 0, 101, 0, 110, 0, 0, 0, 121, 0, 101, 0, 121, 0, 0, 0, 121, 0, 103, 0, 97, 0, 0, 0, 121, 0, 103, 0, 105, 0, 0, 0, 121, 0, 103, 0, 112, 0, 0, 0, 121, 0, 103, 0, 119, 0, 0, 0, 121, 0, 104, 0, 100, 0, 0, 0, 121, 0, 104, 0, 115, 0, 0, 0, 121, 0, 105, 0, 104, 0, 0, 0, 121, 0, 105, 0, 120, 0, 0, 0, 121, 0, 105, 0, 121, 0, 0, 0, 121, 0, 105, 0, 122, 0, 0, 0, 121, 0, 107, 0, 97, 0, 0, 0, 121, 0, 107, 0, 114, 0, 0, 0, 121, 0, 108, 0, 101, 0, 0, 0, 121, 0, 108, 0, 105, 0, 0, 0, 121, 0, 108, 0, 114, 0, 0, 0, 121, 0, 108, 0, 117, 0, 0, 0, 121, 0, 108, 0, 121, 0, 0, 0, 121, 0, 109, 0, 120, 0, 0, 0, 121, 0, 109, 0, 122, 0, 0, 0, 121, 0, 110, 0, 97, 0, 0, 0, 121, 0, 110, 0, 103, 0, 0, 0, 121, 0, 110, 0, 104, 0, 0, 0, 121, 0, 110, 0, 115, 0, 0, 0, 121, 0, 110, 0, 117, 0, 0, 0, 121, 0, 111, 0, 98, 0, 0, 0, 121, 0, 111, 0, 103, 0, 0, 0, 121, 0, 111, 0, 105, 0, 0, 0, 121, 0, 111, 0, 116, 0, 0, 0, 121, 0, 112, 0, 107, 0, 0, 0, 121, 0, 112, 0, 122, 0, 0, 0, 121, 0, 114, 0, 101, 0, 0, 0, 121, 0, 114, 0, 105, 0, 0, 0, 121, 0, 114, 0, 115, 0, 0, 0, 121, 0, 114, 0, 121, 0, 0, 0, 121, 0, 115, 0, 121, 0, 0, 0, 121, 0, 116, 0, 112, 0, 0, 0, 121, 0, 117, 0, 101, 0, 0, 0, 121, 0, 117, 0, 117, 0, 0, 0, 121, 0, 118, 0, 97, 0, 0, 0, 121, 0, 118, 0, 116, 0, 0, 0, 121, 0, 119, 0, 103, 0, 0, 0, 121, 0, 119, 0, 108, 0, 0, 0, 121, 0, 119, 0, 110, 0, 0, 0, 121, 0, 119, 0, 119, 0, 0, 0, 121, 0, 120, 0, 97, 0, 0, 0, 121, 0, 120, 0, 103, 0, 0, 0, 121, 0, 120, 0, 117, 0, 0, 0, 121, 0, 120, 0, 121, 0, 0, 0, 121, 0, 121, 0, 114, 0, 0, 0, 121, 0, 121, 0, 117, 0, 0, 0, 121, 0, 121, 0, 122, 0, 0, 0, 121, 0, 122, 0, 103, 0, 0, 0, 121, 0, 122, 0, 107, 0, 0, 0, 122, 0, 98, 0, 99, 0, 0, 0, 122, 0, 98, 0, 108, 0, 0, 0, 122, 0, 98, 0, 116, 0, 0, 0, 122, 0, 98, 0, 119, 0, 0, 0, 122, 0, 100, 0, 106, 0, 0, 0, 122, 0, 101, 0, 97, 0, 0, 0, 122, 0, 103, 0, 104, 0, 0, 0, 122, 0, 103, 0, 114, 0, 0, 0, 122, 0, 104, 0, 100, 0, 0, 0, 122, 0, 104, 0, 110, 0, 0, 0, 122, 0, 105, 0, 122, 0, 0, 0, 122, 0, 107, 0, 100, 0, 0, 0, 122, 0, 107, 0, 122, 0, 0, 0, 122, 0, 108, 0, 101, 0, 0, 0, 122, 0, 108, 0, 106, 0, 0, 0, 122, 0, 108, 0, 109, 0, 0, 0, 122, 0, 108, 0, 113, 0, 0, 0, 122, 0, 108, 0, 115, 0, 0, 0, 122, 0, 108, 0, 119, 0, 0, 0, 122, 0, 110, 0, 107, 0, 0, 0, 122, 0, 110, 0, 115, 0, 0, 0, 122, 0, 111, 0, 104, 0, 0, 0, 122, 0, 111, 0, 111, 0, 0, 0, 122, 0, 113, 0, 101, 0, 0, 0, 122, 0, 114, 0, 97, 0, 0, 0, 122, 0, 114, 0, 103, 0, 0, 0, 122, 0, 114, 0, 115, 0, 0, 0, 122, 0, 117, 0, 97, 0, 0, 0, 122, 0, 117, 0, 104, 0, 0, 0, 122, 0, 117, 0, 121, 0, 0, 0, 122, 0, 120, 0, 120, 0, 0, 0, 122, 0, 121, 0, 98, 0, 0, 0, 122, 0, 121, 0, 103, 0, 0, 0, 122, 0, 121, 0, 106, 0, 0, 0, 122, 0, 121, 0, 110, 0, 0, 0, 122, 0, 121, 0, 112, 0, 0, 0, 122, 0, 122, 0, 106, 0, 0, 0, 116, 0, 122, 0, 49, 0, 48, 0, 0, 0, 69, 0, 71, 0, 126, 0, 72, 0, 0, 0, 102, 0, 114, 0, 57, 0, 48, 0, 0, 0, 115, 0, 121, 0, 113, 0, 117, 0, 0, 0, 122, 0, 119, 0, 109, 0, 115, 0, 0, 0, 98, 0, 119, 0, 102, 0, 114, 0, 0, 0, 109, 0, 119, 0, 114, 0, 117, 0, 0, 0, 67, 0, 97, 0, 110, 0, 115, 0, 0, 0, 109, 0, 119, 0, 109, 0, 103, 0, 0, 0, 116, 0, 104, 0, 57, 0, 54, 0, 0, 0, 103, 0, 121, 0, 101, 0, 115, 0, 0, 0, 84, 0, 104, 0, 97, 0, 105, 0, 0, 0, 71, 0, 114, 0, 101, 0, 107, 0, 0, 0, 84, 0, 105, 0, 98, 0, 116, 0, 0, 0, 103, 0, 121, 0, 100, 0, 101, 0, 0, 0, 84, 0, 102, 0, 110, 0, 103, 0, 0, 0, 115, 0, 121, 0, 105, 0, 100, 0, 0, 0, 109, 0, 119, 0, 110, 0, 98, 0, 0, 0, 115, 0, 122, 0, 115, 0, 104, 0, 0, 0, 109, 0, 118, 0, 110, 0, 111, 0, 0, 0, 83, 0, 121, 0, 114, 0, 99, 0, 0, 0, 98, 0, 121, 0, 104, 0, 114, 0, 0, 0, 103, 0, 119, 0, 98, 0, 115, 0, 0, 0, 104, 0, 117, 0, 112, 0, 115, 0, 0, 0, 108, 0, 121, 0, 110, 0, 108, 0, 0, 0, 97, 0, 122, 0, 109, 0, 105, 0, 0, 0, 97, 0, 122, 0, 115, 0, 114, 0, 0, 0, 107, 0, 119, 0, 107, 0, 117, 0, 0, 0, 108, 0, 114, 0, 109, 0, 121, 0, 0, 0, 117, 0, 121, 0, 116, 0, 97, 0, 0, 0, 117, 0, 122, 0, 115, 0, 97, 0, 0, 0, 117, 0, 122, 0, 115, 0, 117, 0, 0, 0, 82, 0, 117, 0, 110, 0, 114, 0, 0, 0, 98, 0, 122, 0, 115, 0, 99, 0, 0, 0, 109, 0, 119, 0, 110, 0, 101, 0, 0, 0, 122, 0, 119, 0, 109, 0, 110, 0, 0, 0, 66, 0, 117, 0, 103, 0, 105, 0, 0, 0, 73, 0, 116, 0, 97, 0, 108, 0, 0, 0, 102, 0, 114, 0, 57, 0, 50, 0, 0, 0, 83, 0, 121, 0, 108, 0, 111, 0, 0, 0, 97, 0, 122, 0, 103, 0, 97, 0, 0, 0, 97, 0, 122, 0, 108, 0, 97, 0, 0, 0, 98, 0, 115, 0, 115, 0, 119, 0, 0, 0, 98, 0, 121, 0, 98, 0, 114, 0, 0, 0, 98, 0, 122, 0, 99, 0, 121, 0, 0, 0, 99, 0, 118, 0, 115, 0, 100, 0, 0, 0, 101, 0, 116, 0, 97, 0, 102, 0, 0, 0, 103, 0, 113, 0, 107, 0, 110, 0, 0, 0, 104, 0, 117, 0, 107, 0, 109, 0, 0, 0, 108, 0, 117, 0, 114, 0, 109, 0, 0, 0, 117, 0, 121, 0, 99, 0, 111, 0, 0, 0, 117, 0, 121, 0, 115, 0, 111, 0, 0, 0, 117, 0, 121, 0, 116, 0, 116, 0, 0, 0, 117, 0, 122, 0, 115, 0, 105, 0, 0, 0, 117, 0, 122, 0, 116, 0, 107, 0, 0, 0, 117, 0, 122, 0, 116, 0, 111, 0, 0, 0, 83, 0, 97, 0, 114, 0, 98, 0, 0, 0, 99, 0, 110, 0, 121, 0, 110, 0, 0, 0, 100, 0, 101, 0, 115, 0, 108, 0, 0, 0, 100, 0, 101, 0, 115, 0, 116, 0, 0, 0, 103, 0, 101, 0, 115, 0, 122, 0, 0, 0, 105, 0, 101, 0, 115, 0, 111, 0, 0, 0, 105, 0, 113, 0, 119, 0, 97, 0, 0, 0, 108, 0, 97, 0, 99, 0, 104, 0, 0, 0, 110, 0, 108, 0, 122, 0, 104, 0, 0, 0, 110, 0, 112, 0, 103, 0, 97, 0, 0, 0, 121, 0, 101, 0, 115, 0, 97, 0, 0, 0, 121, 0, 101, 0, 115, 0, 104, 0, 0, 0, 97, 0, 122, 0, 110, 0, 97, 0, 0, 0, 97, 0, 122, 0, 115, 0, 109, 0, 0, 0, 98, 0, 119, 0, 107, 0, 108, 0, 0, 0, 98, 0, 119, 0, 107, 0, 119, 0, 0, 0, 115, 0, 118, 0, 115, 0, 118, 0, 0, 0, 117, 0, 122, 0, 110, 0, 103, 0, 0, 0, 67, 0, 104, 0, 97, 0, 109, 0, 0, 0, 75, 0, 97, 0, 108, 0, 105, 0, 0, 0, 75, 0, 97, 0, 110, 0, 97, 0, 0, 0, 80, 0, 104, 0, 97, 0, 103, 0, 0, 0, 83, 0, 97, 0, 109, 0, 114, 0, 0, 0, 88, 0, 115, 0, 117, 0, 120, 0, 0, 0, 66, 0, 111, 0, 112, 0, 111, 0, 0, 0, 67, 0, 111, 0, 112, 0, 116, 0, 0, 0, 82, 0, 106, 0, 110, 0, 103, 0, 0, 0, 83, 0, 111, 0, 114, 0, 97, 0, 0, 0, 83, 0, 111, 0, 121, 0, 111, 0, 0, 0, 86, 0, 97, 0, 105, 0, 105, 0, 0, 0, 97, 0, 101, 0, 114, 0, 107, 0, 0, 0, 97, 0, 101, 0, 117, 0, 113, 0, 0, 0, 97, 0, 117, 0, 119, 0, 97, 0, 0, 0, 98, 0, 105, 0, 98, 0, 114, 0, 0, 0, 98, 0, 105, 0, 107, 0, 105, 0, 0, 0, 98, 0, 105, 0, 109, 0, 97, 0, 0, 0, 98, 0, 105, 0, 109, 0, 121, 0, 0, 0, 98, 0, 106, 0, 97, 0, 113, 0, 0, 0, 98, 0, 110, 0, 116, 0, 117, 0, 0, 0, 98, 0, 115, 0, 101, 0, 103, 0, 0, 0, 98, 0, 115, 0, 114, 0, 99, 0, 0, 0, 98, 0, 119, 0, 108, 0, 111, 0, 0, 0, 98, 0, 119, 0, 110, 0, 119, 0, 0, 0, 99, 0, 100, 0, 105, 0, 116, 0, 0, 0, 99, 0, 100, 0, 107, 0, 108, 0, 0, 0, 99, 0, 102, 0, 109, 0, 112, 0, 0, 0, 99, 0, 102, 0, 115, 0, 101, 0, 0, 0, 99, 0, 104, 0, 116, 0, 105, 0, 0, 0, 99, 0, 105, 0, 108, 0, 103, 0, 0, 0, 99, 0, 105, 0, 118, 0, 98, 0, 0, 0, 99, 0, 105, 0, 122, 0, 122, 0, 0, 0, 99, 0, 108, 0, 114, 0, 109, 0, 0, 0, 99, 0, 108, 0, 118, 0, 115, 0, 0, 0, 99, 0, 110, 0, 108, 0, 110, 0, 0, 0, 99, 0, 110, 0, 110, 0, 120, 0, 0, 0, 99, 0, 122, 0, 107, 0, 97, 0, 0, 0, 99, 0, 122, 0, 112, 0, 97, 0, 0, 0, 99, 0, 122, 0, 115, 0, 116, 0, 0, 0, 100, 0, 101, 0, 116, 0, 104, 0, 0, 0, 100, 0, 106, 0, 111, 0, 98, 0, 0, 0, 101, 0, 103, 0, 109, 0, 110, 0, 0, 0, 101, 0, 115, 0, 101, 0, 120, 0, 0, 0, 101, 0, 115, 0, 103, 0, 99, 0, 0, 0, 101, 0, 115, 0, 104, 0, 117, 0, 0, 0, 101, 0, 115, 0, 112, 0, 118, 0, 0, 0, 102, 0, 114, 0, 57, 0, 53, 0, 0, 0, 103, 0, 101, 0, 107, 0, 97, 0, 0, 0, 103, 0, 108, 0, 115, 0, 109, 0, 0, 0, 103, 0, 110, 0, 102, 0, 97, 0, 0, 0, 103, 0, 110, 0, 102, 0, 114, 0, 0, 0, 103, 0, 110, 0, 108, 0, 97, 0, 0, 0, 103, 0, 110, 0, 108, 0, 111, 0, 0, 0, 103, 0, 110, 0, 112, 0, 105, 0, 0, 0, 103, 0, 116, 0, 106, 0, 97, 0, 0, 0, 103, 0, 116, 0, 106, 0, 117, 0, 0, 0, 103, 0, 116, 0, 122, 0, 97, 0, 0, 0, 103, 0, 119, 0, 99, 0, 97, 0, 0, 0, 103, 0, 119, 0, 111, 0, 105, 0, 0, 0, 103, 0, 121, 0, 117, 0, 116, 0, 0, 0, 104, 0, 110, 0, 99, 0, 112, 0, 0, 0, 104, 0, 117, 0, 100, 0, 101, 0, 0, 0, 104, 0, 117, 0, 100, 0, 117, 0, 0, 0, 104, 0, 117, 0, 103, 0, 121, 0, 0, 0, 104, 0, 117, 0, 109, 0, 105, 0, 0, 0, 104, 0, 117, 0, 110, 0, 107, 0, 0, 0, 104, 0, 117, 0, 115, 0, 107, 0, 0, 0, 104, 0, 117, 0, 115, 0, 110, 0, 0, 0, 104, 0, 117, 0, 118, 0, 101, 0, 0, 0, 105, 0, 100, 0, 103, 0, 111, 0, 0, 0, 105, 0, 100, 0, 106, 0, 119, 0, 0, 0, 105, 0, 100, 0, 112, 0, 112, 0, 0, 0, 105, 0, 101, 0, 99, 0, 101, 0, 0, 0, 105, 0, 101, 0, 99, 0, 119, 0, 0, 0, 105, 0, 101, 0, 107, 0, 101, 0, 0, 0, 105, 0, 101, 0, 107, 0, 107, 0, 0, 0, 105, 0, 101, 0, 107, 0, 121, 0, 0, 0, 105, 0, 101, 0, 108, 0, 107, 0, 0, 0, 105, 0, 101, 0, 108, 0, 109, 0, 0, 0, 105, 0, 108, 0, 104, 0, 97, 0, 0, 0, 105, 0, 110, 0, 116, 0, 103, 0, 0, 0, 105, 0, 110, 0, 116, 0, 114, 0, 0, 0, 105, 0, 116, 0, 98, 0, 110, 0, 0, 0, 105, 0, 116, 0, 102, 0, 105, 0, 0, 0, 105, 0, 116, 0, 102, 0, 114, 0, 0, 0, 105, 0, 116, 0, 112, 0, 97, 0, 0, 0, 105, 0, 116, 0, 116, 0, 110, 0, 0, 0, 107, 0, 119, 0, 106, 0, 97, 0, 0, 0, 108, 0, 97, 0, 97, 0, 116, 0, 0, 0, 108, 0, 97, 0, 118, 0, 105, 0, 0, 0, 108, 0, 97, 0, 120, 0, 101, 0, 0, 0, 108, 0, 114, 0, 103, 0, 107, 0, 0, 0, 108, 0, 114, 0, 114, 0, 105, 0, 0, 0, 108, 0, 117, 0, 101, 0, 115, 0, 0, 0, 108, 0, 117, 0, 108, 0, 117, 0, 0, 0, 108, 0, 121, 0, 110, 0, 113, 0, 0, 0, 108, 0, 121, 0, 119, 0, 97, 0, 0, 0, 109, 0, 99, 0, 103, 0, 97, 0, 0, 0, 109, 0, 99, 0, 106, 0, 101, 0, 0, 0, 109, 0, 99, 0, 108, 0, 97, 0, 0, 0, 109, 0, 99, 0, 109, 0, 97, 0, 0, 0, 109, 0, 99, 0, 109, 0, 99, 0, 0, 0, 109, 0, 99, 0, 109, 0, 103, 0, 0, 0, 109, 0, 99, 0, 109, 0, 111, 0, 0, 0, 109, 0, 100, 0, 104, 0, 105, 0, 0, 0, 109, 0, 100, 0, 114, 0, 105, 0, 0, 0, 109, 0, 117, 0, 98, 0, 108, 0, 0, 0, 109, 0, 117, 0, 98, 0, 114, 0, 0, 0, 109, 0, 117, 0, 114, 0, 114, 0, 0, 0, 109, 0, 119, 0, 108, 0, 105, 0, 0, 0, 109, 0, 119, 0, 110, 0, 105, 0, 0, 0, 109, 0, 119, 0, 110, 0, 107, 0, 0, 0, 109, 0, 119, 0, 110, 0, 117, 0, 0, 0, 109, 0, 119, 0, 115, 0, 97, 0, 0, 0, 110, 0, 103, 0, 97, 0, 110, 0, 0, 0, 110, 0, 103, 0, 98, 0, 111, 0, 0, 0, 110, 0, 103, 0, 99, 0, 114, 0, 0, 0, 110, 0, 103, 0, 106, 0, 105, 0, 0, 0, 110, 0, 103, 0, 114, 0, 105, 0, 0, 0, 110, 0, 103, 0, 116, 0, 97, 0, 0, 0, 110, 0, 105, 0, 99, 0, 97, 0, 0, 0, 110, 0, 112, 0, 107, 0, 111, 0, 0, 0, 110, 0, 112, 0, 110, 0, 97, 0, 0, 0, 110, 0, 112, 0, 114, 0, 97, 0, 0, 0, 110, 0, 112, 0, 115, 0, 97, 0, 0, 0, 110, 0, 112, 0, 115, 0, 101, 0, 0, 0, 111, 0, 109, 0, 122, 0, 97, 0, 0, 0, 112, 0, 107, 0, 103, 0, 98, 0, 0, 0, 112, 0, 108, 0, 108, 0, 98, 0, 0, 0, 114, 0, 111, 0, 97, 0, 114, 0, 0, 0, 114, 0, 117, 0, 98, 0, 97, 0, 0, 0, 114, 0, 117, 0, 98, 0, 117, 0, 0, 0, 114, 0, 117, 0, 100, 0, 97, 0, 0, 0, 114, 0, 117, 0, 115, 0, 97, 0, 0, 0, 114, 0, 117, 0, 116, 0, 97, 0, 0, 0, 115, 0, 100, 0, 100, 0, 119, 0, 0, 0, 115, 0, 100, 0, 103, 0, 122, 0, 0, 0, 115, 0, 100, 0, 110, 0, 114, 0, 0, 0, 115, 0, 107, 0, 112, 0, 118, 0, 0, 0, 115, 0, 107, 0, 122, 0, 105, 0, 0, 0, 115, 0, 108, 0, 110, 0, 119, 0, 0, 0, 115, 0, 110, 0, 108, 0, 103, 0, 0, 0, 115, 0, 111, 0, 98, 0, 114, 0, 0, 0, 115, 0, 111, 0, 103, 0, 101, 0, 0, 0, 115, 0, 111, 0, 115, 0, 97, 0, 0, 0, 115, 0, 114, 0, 115, 0, 105, 0, 0, 0, 115, 0, 114, 0, 119, 0, 97, 0, 0, 0, 115, 0, 115, 0, 101, 0, 99, 0, 0, 0, 115, 0, 115, 0, 101, 0, 119, 0, 0, 0, 115, 0, 115, 0, 106, 0, 103, 0, 0, 0, 115, 0, 121, 0, 104, 0, 97, 0, 0, 0, 116, 0, 100, 0, 107, 0, 97, 0, 0, 0, 116, 0, 106, 0, 103, 0, 98, 0, 0, 0, 116, 0, 108, 0, 98, 0, 97, 0, 0, 0, 116, 0, 108, 0, 118, 0, 105, 0, 0, 0, 117, 0, 115, 0, 99, 0, 97, 0, 0, 0, 117, 0, 115, 0, 100, 0, 99, 0, 0, 0, 117, 0, 115, 0, 108, 0, 97, 0, 0, 0, 117, 0, 115, 0, 110, 0, 104, 0, 0, 0, 117, 0, 121, 0, 99, 0, 108, 0, 0, 0, 117, 0, 121, 0, 109, 0, 97, 0, 0, 0, 117, 0, 121, 0, 114, 0, 110, 0, 0, 0, 119, 0, 102, 0, 117, 0, 118, 0, 0, 0, 119, 0, 115, 0, 118, 0, 115, 0, 0, 0, 121, 0, 101, 0, 104, 0, 117, 0, 0, 0, 121, 0, 101, 0, 106, 0, 97, 0, 0, 0, 121, 0, 101, 0, 114, 0, 97, 0, 0, 0, 121, 0, 101, 0, 115, 0, 117, 0, 0, 0, 122, 0, 97, 0, 101, 0, 99, 0, 0, 0, 122, 0, 119, 0, 104, 0, 97, 0, 0, 0, 49, 0, 57, 0, 57, 0, 52, 0, 0, 0, 83, 0, 103, 0, 110, 0, 119, 0, 0, 0, 97, 0, 122, 0, 98, 0, 97, 0, 0, 0, 97, 0, 122, 0, 110, 0, 118, 0, 0, 0, 98, 0, 119, 0, 106, 0, 119, 0, 0, 0, 98, 0, 119, 0, 115, 0, 101, 0, 0, 0, 98, 0, 121, 0, 104, 0, 111, 0, 0, 0, 98, 0, 121, 0, 118, 0, 105, 0, 0, 0, 99, 0, 110, 0, 102, 0, 106, 0, 0, 0, 99, 0, 117, 0, 57, 0, 57, 0, 0, 0, 99, 0, 118, 0, 116, 0, 115, 0, 0, 0, 99, 0, 122, 0, 53, 0, 49, 0, 0, 0, 99, 0, 122, 0, 53, 0, 50, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 0, 0, 99, 0, 122, 0, 55, 0, 50, 0, 0, 0, 99, 0, 122, 0, 56, 0, 48, 0, 0, 0, 103, 0, 101, 0, 115, 0, 106, 0, 0, 0, 103, 0, 121, 0, 99, 0, 117, 0, 0, 0, 103, 0, 121, 0, 112, 0, 116, 0, 0, 0, 104, 0, 117, 0, 99, 0, 115, 0, 0, 0, 104, 0, 117, 0, 104, 0, 101, 0, 0, 0, 104, 0, 117, 0, 107, 0, 118, 0, 0, 0, 105, 0, 116, 0, 109, 0, 115, 0, 0, 0, 108, 0, 121, 0, 106, 0, 97, 0, 0, 0, 109, 0, 119, 0, 116, 0, 104, 0, 0, 0, 114, 0, 111, 0, 99, 0, 118, 0, 0, 0, 114, 0, 115, 0, 118, 0, 111, 0, 0, 0, 115, 0, 118, 0, 115, 0, 115, 0, 0, 0, 115, 0, 122, 0, 108, 0, 117, 0, 0, 0, 117, 0, 122, 0, 102, 0, 97, 0, 0, 0, 117, 0, 122, 0, 106, 0, 105, 0, 0, 0, 99, 0, 122, 0, 107, 0, 114, 0, 0, 0, 109, 0, 116, 0, 53, 0, 48, 0, 0, 0, 109, 0, 116, 0, 54, 0, 48, 0, 0, 0, 65, 0, 103, 0, 104, 0, 98, 0, 0, 0, 65, 0, 114, 0, 109, 0, 105, 0, 0, 0, 65, 0, 118, 0, 115, 0, 116, 0, 0, 0, 66, 0, 97, 0, 108, 0, 105, 0, 0, 0, 66, 0, 97, 0, 109, 0, 117, 0, 0, 0, 66, 0, 97, 0, 116, 0, 107, 0, 0, 0, 66, 0, 117, 0, 104, 0, 100, 0, 0, 0, 67, 0, 97, 0, 107, 0, 109, 0, 0, 0, 67, 0, 97, 0, 114, 0, 105, 0, 0, 0, 67, 0, 104, 0, 101, 0, 114, 0, 0, 0, 67, 0, 112, 0, 114, 0, 116, 0, 0, 0, 68, 0, 117, 0, 112, 0, 108, 0, 0, 0, 69, 0, 103, 0, 121, 0, 112, 0, 0, 0, 69, 0, 108, 0, 98, 0, 97, 0, 0, 0, 71, 0, 111, 0, 116, 0, 104, 0, 0, 0, 71, 0, 114, 0, 97, 0, 110, 0, 0, 0, 71, 0, 117, 0, 106, 0, 114, 0, 0, 0, 72, 0, 97, 0, 110, 0, 105, 0, 0, 0, 72, 0, 97, 0, 110, 0, 111, 0, 0, 0, 72, 0, 109, 0, 110, 0, 103, 0, 0, 0, 74, 0, 97, 0, 118, 0, 97, 0, 0, 0, 74, 0, 112, 0, 97, 0, 110, 0, 0, 0, 75, 0, 111, 0, 114, 0, 101, 0, 0, 0, 76, 0, 97, 0, 110, 0, 97, 0, 0, 0, 76, 0, 101, 0, 112, 0, 99, 0, 0, 0, 76, 0, 105, 0, 109, 0, 98, 0, 0, 0, 76, 0, 105, 0, 115, 0, 117, 0, 0, 0, 76, 0, 121, 0, 99, 0, 105, 0, 0, 0, 76, 0, 121, 0, 100, 0, 105, 0, 0, 0, 77, 0, 97, 0, 104, 0, 106, 0, 0, 0, 77, 0, 97, 0, 110, 0, 100, 0, 0, 0, 77, 0, 97, 0, 110, 0, 105, 0, 0, 0, 77, 0, 101, 0, 110, 0, 100, 0, 0, 0, 77, 0, 101, 0, 114, 0, 99, 0, 0, 0, 77, 0, 111, 0, 100, 0, 105, 0, 0, 0, 77, 0, 114, 0, 111, 0, 111, 0, 0, 0, 77, 0, 116, 0, 101, 0, 105, 0, 0, 0, 78, 0, 97, 0, 114, 0, 98, 0, 0, 0, 78, 0, 98, 0, 97, 0, 116, 0, 0, 0, 79, 0, 103, 0, 97, 0, 109, 0, 0, 0, 79, 0, 108, 0, 99, 0, 107, 0, 0, 0, 79, 0, 114, 0, 107, 0, 104, 0, 0, 0, 79, 0, 115, 0, 103, 0, 101, 0, 0, 0, 79, 0, 115, 0, 109, 0, 97, 0, 0, 0, 80, 0, 97, 0, 108, 0, 109, 0, 0, 0, 80, 0, 101, 0, 114, 0, 109, 0, 0, 0, 80, 0, 104, 0, 108, 0, 105, 0, 0, 0, 80, 0, 104, 0, 108, 0, 112, 0, 0, 0, 80, 0, 104, 0, 110, 0, 120, 0, 0, 0, 80, 0, 108, 0, 114, 0, 100, 0, 0, 0, 80, 0, 114, 0, 116, 0, 105, 0, 0, 0, 83, 0, 97, 0, 117, 0, 114, 0, 0, 0, 83, 0, 104, 0, 114, 0, 100, 0, 0, 0, 83, 0, 105, 0, 100, 0, 100, 0, 0, 0, 83, 0, 117, 0, 110, 0, 100, 0, 0, 0, 84, 0, 97, 0, 103, 0, 98, 0, 0, 0, 84, 0, 97, 0, 107, 0, 114, 0, 0, 0, 84, 0, 97, 0, 108, 0, 101, 0, 0, 0, 84, 0, 97, 0, 108, 0, 117, 0, 0, 0, 84, 0, 97, 0, 118, 0, 116, 0, 0, 0, 84, 0, 103, 0, 108, 0, 103, 0, 0, 0, 84, 0, 104, 0, 97, 0, 97, 0, 0, 0, 84, 0, 105, 0, 114, 0, 104, 0, 0, 0, 85, 0, 103, 0, 97, 0, 114, 0, 0, 0, 87, 0, 97, 0, 114, 0, 97, 0, 0, 0, 88, 0, 112, 0, 101, 0, 111, 0, 0, 0, 89, 0, 105, 0, 105, 0, 105, 0, 0, 0, 97, 0, 101, 0, 97, 0, 106, 0, 0, 0, 97, 0, 101, 0, 97, 0, 122, 0, 0, 0, 97, 0, 101, 0, 100, 0, 117, 0, 0, 0, 97, 0, 101, 0, 102, 0, 117, 0, 0, 0, 97, 0, 101, 0, 115, 0, 104, 0, 0, 0, 97, 0, 109, 0, 97, 0, 103, 0, 0, 0, 97, 0, 109, 0, 97, 0, 118, 0, 0, 0, 97, 0, 109, 0, 101, 0, 114, 0, 0, 0, 97, 0, 109, 0, 103, 0, 114, 0, 0, 0, 97, 0, 109, 0, 107, 0, 116, 0, 0, 0, 97, 0, 109, 0, 108, 0, 111, 0, 0, 0, 97, 0, 109, 0, 115, 0, 104, 0, 0, 0, 97, 0, 109, 0, 115, 0, 117, 0, 0, 0, 97, 0, 109, 0, 116, 0, 118, 0, 0, 0, 97, 0, 109, 0, 118, 0, 100, 0, 0, 0, 97, 0, 117, 0, 110, 0, 116, 0, 0, 0, 97, 0, 117, 0, 115, 0, 97, 0, 0, 0, 97, 0, 122, 0, 110, 0, 120, 0, 0, 0, 97, 0, 122, 0, 115, 0, 97, 0, 0, 0, 97, 0, 122, 0, 120, 0, 97, 0, 0, 0, 97, 0, 122, 0, 121, 0, 101, 0, 0, 0, 98, 0, 104, 0, 49, 0, 55, 0, 0, 0, 98, 0, 105, 0, 98, 0, 98, 0, 0, 0, 98, 0, 105, 0, 99, 0, 97, 0, 0, 0, 98, 0, 105, 0, 99, 0, 105, 0, 0, 0, 98, 0, 105, 0, 103, 0, 105, 0, 0, 0, 98, 0, 105, 0, 107, 0, 114, 0, 0, 0, 98, 0, 105, 0, 107, 0, 121, 0, 0, 0, 98, 0, 105, 0, 109, 0, 117, 0, 0, 0, 98, 0, 105, 0, 109, 0, 119, 0, 0, 0, 98, 0, 105, 0, 110, 0, 103, 0, 0, 0, 98, 0, 105, 0, 114, 0, 109, 0, 0, 0, 98, 0, 105, 0, 114, 0, 116, 0, 0, 0, 98, 0, 105, 0, 114, 0, 121, 0, 0, 0, 98, 0, 106, 0, 98, 0, 111, 0, 0, 0, 98, 0, 106, 0, 99, 0, 111, 0, 0, 0, 98, 0, 106, 0, 100, 0, 111, 0, 0, 0, 98, 0, 106, 0, 107, 0, 111, 0, 0, 0, 98, 0, 106, 0, 108, 0, 105, 0, 0, 0, 98, 0, 106, 0, 109, 0, 111, 0, 0, 0, 98, 0, 106, 0, 111, 0, 117, 0, 0, 0, 98, 0, 106, 0, 112, 0, 108, 0, 0, 0, 98, 0, 106, 0, 122, 0, 111, 0, 0, 0, 98, 0, 110, 0, 98, 0, 101, 0, 0, 0, 98, 0, 110, 0, 98, 0, 109, 0, 0, 0, 98, 0, 110, 0, 116, 0, 101, 0, 0, 0, 98, 0, 113, 0, 98, 0, 111, 0, 0, 0, 98, 0, 113, 0, 115, 0, 97, 0, 0, 0, 98, 0, 113, 0, 115, 0, 101, 0, 0, 0, 98, 0, 114, 0, 97, 0, 99, 0, 0, 0, 98, 0, 114, 0, 97, 0, 112, 0, 0, 0, 98, 0, 114, 0, 98, 0, 97, 0, 0, 0, 98, 0, 114, 0, 99, 0, 101, 0, 0, 0, 98, 0, 114, 0, 100, 0, 102, 0, 0, 0, 98, 0, 114, 0, 101, 0, 115, 0, 0, 0, 98, 0, 114, 0, 103, 0, 111, 0, 0, 0, 98, 0, 114, 0, 109, 0, 97, 0, 0, 0, 98, 0, 114, 0, 109, 0, 103, 0, 0, 0, 98, 0, 114, 0, 112, 0, 101, 0, 0, 0, 98, 0, 114, 0, 112, 0, 105, 0, 0, 0, 98, 0, 114, 0, 112, 0, 114, 0, 0, 0, 98, 0, 114, 0, 114, 0, 106, 0, 0, 0, 98, 0, 114, 0, 115, 0, 99, 0, 0, 0, 98, 0, 114, 0, 115, 0, 101, 0, 0, 0, 98, 0, 114, 0, 115, 0, 112, 0, 0, 0, 98, 0, 114, 0, 116, 0, 111, 0, 0, 0, 98, 0, 115, 0, 97, 0, 107, 0, 0, 0, 98, 0, 115, 0, 98, 0, 105, 0, 0, 0, 98, 0, 115, 0, 98, 0, 112, 0, 0, 0, 98, 0, 115, 0, 98, 0, 121, 0, 0, 0, 98, 0, 115, 0, 99, 0, 101, 0, 0, 0, 98, 0, 115, 0, 99, 0, 105, 0, 0, 0, 98, 0, 115, 0, 99, 0, 107, 0, 0, 0, 98, 0, 115, 0, 99, 0, 111, 0, 0, 0, 98, 0, 115, 0, 99, 0, 115, 0, 0, 0, 98, 0, 115, 0, 101, 0, 120, 0, 0, 0, 98, 0, 115, 0, 102, 0, 112, 0, 0, 0, 98, 0, 115, 0, 104, 0, 105, 0, 0, 0, 98, 0, 115, 0, 104, 0, 116, 0, 0, 0, 98, 0, 115, 0, 105, 0, 110, 0, 0, 0, 98, 0, 115, 0, 108, 0, 105, 0, 0, 0, 98, 0, 115, 0, 109, 0, 99, 0, 0, 0, 98, 0, 115, 0, 109, 0, 103, 0, 0, 0, 98, 0, 115, 0, 109, 0, 105, 0, 0, 0, 98, 0, 115, 0, 110, 0, 101, 0, 0, 0, 98, 0, 115, 0, 110, 0, 115, 0, 0, 0, 98, 0, 115, 0, 114, 0, 105, 0, 0, 0, 98, 0, 115, 0, 115, 0, 97, 0, 0, 0, 98, 0, 115, 0, 115, 0, 101, 0, 0, 0, 98, 0, 115, 0, 115, 0, 111, 0, 0, 0, 98, 0, 115, 0, 115, 0, 115, 0, 0, 0, 98, 0, 115, 0, 119, 0, 103, 0, 0, 0, 98, 0, 116, 0, 103, 0, 97, 0, 0, 0, 98, 0, 116, 0, 116, 0, 121, 0, 0, 0, 98, 0, 119, 0, 99, 0, 101, 0, 0, 0, 98, 0, 119, 0, 99, 0, 104, 0, 0, 0, 98, 0, 119, 0, 103, 0, 97, 0, 0, 0, 98, 0, 119, 0, 103, 0, 104, 0, 0, 0, 98, 0, 119, 0, 107, 0, 103, 0, 0, 0, 98, 0, 119, 0, 110, 0, 101, 0, 0, 0, 98, 0, 119, 0, 115, 0, 116, 0, 0, 0, 98, 0, 121, 0, 104, 0, 109, 0, 0, 0, 98, 0, 121, 0, 109, 0, 97, 0, 0, 0, 98, 0, 121, 0, 109, 0, 105, 0, 0, 0, 98, 0, 122, 0, 98, 0, 122, 0, 0, 0, 98, 0, 122, 0, 111, 0, 119, 0, 0, 0, 99, 0, 97, 0, 97, 0, 98, 0, 0, 0, 99, 0, 97, 0, 98, 0, 99, 0, 0, 0, 99, 0, 97, 0, 109, 0, 98, 0, 0, 0, 99, 0, 97, 0, 110, 0, 98, 0, 0, 0, 99, 0, 97, 0, 110, 0, 108, 0, 0, 0, 99, 0, 97, 0, 111, 0, 110, 0, 0, 0, 99, 0, 97, 0, 112, 0, 101, 0, 0, 0, 99, 0, 97, 0, 113, 0, 99, 0, 0, 0, 99, 0, 97, 0, 115, 0, 107, 0, 0, 0, 99, 0, 97, 0, 121, 0, 116, 0, 0, 0, 99, 0, 100, 0, 98, 0, 99, 0, 0, 0, 99, 0, 100, 0, 98, 0, 117, 0, 0, 0, 99, 0, 100, 0, 101, 0, 113, 0, 0, 0, 99, 0, 100, 0, 104, 0, 117, 0, 0, 0, 99, 0, 100, 0, 107, 0, 99, 0, 0, 0, 99, 0, 100, 0, 107, 0, 101, 0, 0, 0, 99, 0, 100, 0, 107, 0, 103, 0, 0, 0, 99, 0, 100, 0, 107, 0, 110, 0, 0, 0, 99, 0, 100, 0, 107, 0, 115, 0, 0, 0, 99, 0, 100, 0, 108, 0, 111, 0, 0, 0, 99, 0, 100, 0, 108, 0, 117, 0, 0, 0, 99, 0, 100, 0, 109, 0, 97, 0, 0, 0, 99, 0, 100, 0, 110, 0, 107, 0, 0, 0, 99, 0, 100, 0, 110, 0, 117, 0, 0, 0, 99, 0, 100, 0, 115, 0, 97, 0, 0, 0, 99, 0, 100, 0, 115, 0, 107, 0, 0, 0, 99, 0, 100, 0, 115, 0, 117, 0, 0, 0, 99, 0, 100, 0, 116, 0, 97, 0, 0, 0, 99, 0, 100, 0, 116, 0, 111, 0, 0, 0, 99, 0, 100, 0, 116, 0, 117, 0, 0, 0, 99, 0, 102, 0, 97, 0, 99, 0, 0, 0, 99, 0, 102, 0, 98, 0, 98, 0, 0, 0, 99, 0, 102, 0, 98, 0, 107, 0, 0, 0, 99, 0, 102, 0, 104, 0, 107, 0, 0, 0, 99, 0, 102, 0, 104, 0, 109, 0, 0, 0, 99, 0, 102, 0, 104, 0, 115, 0, 0, 0, 99, 0, 102, 0, 107, 0, 98, 0, 0, 0, 99, 0, 102, 0, 107, 0, 103, 0, 0, 0, 99, 0, 102, 0, 108, 0, 98, 0, 0, 0, 99, 0, 102, 0, 109, 0, 98, 0, 0, 0, 99, 0, 102, 0, 110, 0, 109, 0, 0, 0, 99, 0, 102, 0, 111, 0, 112, 0, 0, 0, 99, 0, 102, 0, 117, 0, 107, 0, 0, 0, 99, 0, 102, 0, 118, 0, 107, 0, 0, 0, 99, 0, 104, 0, 97, 0, 103, 0, 0, 0, 99, 0, 104, 0, 97, 0, 105, 0, 0, 0, 99, 0, 104, 0, 97, 0, 114, 0, 0, 0, 99, 0, 104, 0, 98, 0, 101, 0, 0, 0, 99, 0, 104, 0, 98, 0, 108, 0, 0, 0, 99, 0, 104, 0, 98, 0, 115, 0, 0, 0, 99, 0, 104, 0, 102, 0, 114, 0, 0, 0, 99, 0, 104, 0, 103, 0, 101, 0, 0, 0, 99, 0, 104, 0, 103, 0, 108, 0, 0, 0, 99, 0, 104, 0, 103, 0, 114, 0, 0, 0, 99, 0, 104, 0, 106, 0, 117, 0, 0, 0, 99, 0, 104, 0, 108, 0, 117, 0, 0, 0, 99, 0, 104, 0, 110, 0, 101, 0, 0, 0, 99, 0, 104, 0, 110, 0, 119, 0, 0, 0, 99, 0, 104, 0, 111, 0, 119, 0, 0, 0, 99, 0, 104, 0, 115, 0, 111, 0, 0, 0, 99, 0, 104, 0, 115, 0, 122, 0, 0, 0, 99, 0, 104, 0, 116, 0, 103, 0, 0, 0, 99, 0, 104, 0, 117, 0, 114, 0, 0, 0, 99, 0, 104, 0, 118, 0, 100, 0, 0, 0, 99, 0, 104, 0, 118, 0, 115, 0, 0, 0, 99, 0, 105, 0, 97, 0, 98, 0, 0, 0, 99, 0, 105, 0, 98, 0, 115, 0, 0, 0, 99, 0, 105, 0, 99, 0, 109, 0, 0, 0, 99, 0, 105, 0, 100, 0, 110, 0, 0, 0, 99, 0, 105, 0, 103, 0, 100, 0, 0, 0, 99, 0, 105, 0, 108, 0, 99, 0, 0, 0, 99, 0, 105, 0, 109, 0, 103, 0, 0, 0, 99, 0, 105, 0, 115, 0, 109, 0, 0, 0, 99, 0, 105, 0, 115, 0, 118, 0, 0, 0, 99, 0, 105, 0, 119, 0, 114, 0, 0, 0, 99, 0, 105, 0, 121, 0, 109, 0, 0, 0, 99, 0, 108, 0, 97, 0, 105, 0, 0, 0, 99, 0, 108, 0, 97, 0, 110, 0, 0, 0, 99, 0, 108, 0, 97, 0, 112, 0, 0, 0, 99, 0, 108, 0, 97, 0, 114, 0, 0, 0, 99, 0, 108, 0, 97, 0, 116, 0, 0, 0, 99, 0, 108, 0, 98, 0, 105, 0, 0, 0, 99, 0, 108, 0, 99, 0, 111, 0, 0, 0, 99, 0, 108, 0, 108, 0, 105, 0, 0, 0, 99, 0, 108, 0, 108, 0, 108, 0, 0, 0, 99, 0, 108, 0, 108, 0, 114, 0, 0, 0, 99, 0, 108, 0, 109, 0, 97, 0, 0, 0, 99, 0, 108, 0, 109, 0, 108, 0, 0, 0, 99, 0, 108, 0, 110, 0, 98, 0, 0, 0, 99, 0, 108, 0, 116, 0, 97, 0, 0, 0, 99, 0, 109, 0, 97, 0, 100, 0, 0, 0, 99, 0, 109, 0, 99, 0, 101, 0, 0, 0, 99, 0, 109, 0, 101, 0, 110, 0, 0, 0, 99, 0, 109, 0, 101, 0, 115, 0, 0, 0, 99, 0, 109, 0, 108, 0, 116, 0, 0, 0, 99, 0, 109, 0, 110, 0, 111, 0, 0, 0, 99, 0, 109, 0, 110, 0, 119, 0, 0, 0, 99, 0, 109, 0, 111, 0, 117, 0, 0, 0, 99, 0, 109, 0, 115, 0, 117, 0, 0, 0, 99, 0, 109, 0, 115, 0, 119, 0, 0, 0, 99, 0, 110, 0, 97, 0, 104, 0, 0, 0, 99, 0, 110, 0, 98, 0, 106, 0, 0, 0, 99, 0, 110, 0, 99, 0, 113, 0, 0, 0, 99, 0, 110, 0, 103, 0, 100, 0, 0, 0, 99, 0, 110, 0, 103, 0, 115, 0, 0, 0, 99, 0, 110, 0, 103, 0, 120, 0, 0, 0, 99, 0, 110, 0, 103, 0, 122, 0, 0, 0, 99, 0, 110, 0, 104, 0, 101, 0, 0, 0, 99, 0, 110, 0, 104, 0, 105, 0, 0, 0, 99, 0, 110, 0, 104, 0, 110, 0, 0, 0, 99, 0, 110, 0, 106, 0, 108, 0, 0, 0, 99, 0, 110, 0, 106, 0, 115, 0, 0, 0, 99, 0, 110, 0, 106, 0, 120, 0, 0, 0, 99, 0, 110, 0, 109, 0, 111, 0, 0, 0, 99, 0, 110, 0, 110, 0, 109, 0, 0, 0, 99, 0, 110, 0, 113, 0, 104, 0, 0, 0, 99, 0, 110, 0, 115, 0, 104, 0, 0, 0, 99, 0, 110, 0, 115, 0, 110, 0, 0, 0, 99, 0, 110, 0, 115, 0, 120, 0, 0, 0, 99, 0, 110, 0, 116, 0, 106, 0, 0, 0, 99, 0, 110, 0, 116, 0, 119, 0, 0, 0, 99, 0, 110, 0, 120, 0, 106, 0, 0, 0, 99, 0, 110, 0, 120, 0, 122, 0, 0, 0, 99, 0, 110, 0, 122, 0, 106, 0, 0, 0, 99, 0, 111, 0, 100, 0, 99, 0, 0, 0, 99, 0, 114, 0, 115, 0, 106, 0, 0, 0, 99, 0, 117, 0, 48, 0, 49, 0, 0, 0, 99, 0, 118, 0, 98, 0, 114, 0, 0, 0, 99, 0, 118, 0, 98, 0, 118, 0, 0, 0, 99, 0, 118, 0, 99, 0, 97, 0, 0, 0, 99, 0, 118, 0, 99, 0, 102, 0, 0, 0, 99, 0, 118, 0, 99, 0, 114, 0, 0, 0, 99, 0, 118, 0, 109, 0, 97, 0, 0, 0, 99, 0, 118, 0, 109, 0, 111, 0, 0, 0, 99, 0, 118, 0, 112, 0, 97, 0, 0, 0, 99, 0, 118, 0, 112, 0, 110, 0, 0, 0, 99, 0, 118, 0, 112, 0, 114, 0, 0, 0, 99, 0, 118, 0, 114, 0, 98, 0, 0, 0, 99, 0, 118, 0, 114, 0, 103, 0, 0, 0, 99, 0, 118, 0, 114, 0, 115, 0, 0, 0, 99, 0, 118, 0, 115, 0, 102, 0, 0, 0, 99, 0, 118, 0, 115, 0, 111, 0, 0, 0, 99, 0, 118, 0, 115, 0, 115, 0, 0, 0, 99, 0, 118, 0, 115, 0, 118, 0, 0, 0, 99, 0, 118, 0, 116, 0, 97, 0, 0, 0, 99, 0, 122, 0, 49, 0, 48, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 0, 0, 99, 0, 122, 0, 51, 0, 49, 0, 0, 0, 99, 0, 122, 0, 51, 0, 50, 0, 0, 0, 99, 0, 122, 0, 52, 0, 49, 0, 0, 0, 99, 0, 122, 0, 52, 0, 50, 0, 0, 0, 99, 0, 122, 0, 53, 0, 51, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 0, 0, 99, 0, 122, 0, 55, 0, 49, 0, 0, 0, 100, 0, 101, 0, 98, 0, 98, 0, 0, 0, 100, 0, 101, 0, 98, 0, 101, 0, 0, 0, 100, 0, 101, 0, 98, 0, 119, 0, 0, 0, 100, 0, 101, 0, 98, 0, 121, 0, 0, 0, 100, 0, 101, 0, 104, 0, 98, 0, 0, 0, 100, 0, 101, 0, 104, 0, 101, 0, 0, 0, 100, 0, 101, 0, 104, 0, 104, 0, 0, 0, 100, 0, 101, 0, 109, 0, 118, 0, 0, 0, 100, 0, 101, 0, 110, 0, 105, 0, 0, 0, 100, 0, 101, 0, 110, 0, 119, 0, 0, 0, 100, 0, 101, 0, 114, 0, 112, 0, 0, 0, 100, 0, 101, 0, 115, 0, 104, 0, 0, 0, 100, 0, 101, 0, 115, 0, 110, 0, 0, 0, 100, 0, 106, 0, 116, 0, 97, 0, 0, 0, 101, 0, 101, 0, 51, 0, 55, 0, 0, 0, 101, 0, 101, 0, 51, 0, 57, 0, 0, 0, 101, 0, 101, 0, 52, 0, 52, 0, 0, 0, 101, 0, 101, 0, 52, 0, 57, 0, 0, 0, 101, 0, 101, 0, 53, 0, 49, 0, 0, 0, 101, 0, 101, 0, 53, 0, 55, 0, 0, 0, 101, 0, 101, 0, 53, 0, 57, 0, 0, 0, 101, 0, 101, 0, 54, 0, 53, 0, 0, 0, 101, 0, 101, 0, 54, 0, 55, 0, 0, 0, 101, 0, 101, 0, 55, 0, 48, 0, 0, 0, 101, 0, 101, 0, 55, 0, 52, 0, 0, 0, 101, 0, 101, 0, 55, 0, 56, 0, 0, 0, 101, 0, 101, 0, 56, 0, 50, 0, 0, 0, 101, 0, 101, 0, 56, 0, 52, 0, 0, 0, 101, 0, 101, 0, 56, 0, 54, 0, 0, 0, 101, 0, 103, 0, 98, 0, 97, 0, 0, 0, 101, 0, 103, 0, 98, 0, 104, 0, 0, 0, 101, 0, 103, 0, 100, 0, 107, 0, 0, 0, 101, 0, 103, 0, 100, 0, 116, 0, 0, 0, 101, 0, 103, 0, 103, 0, 104, 0, 0, 0, 101, 0, 103, 0, 103, 0, 122, 0, 0, 0, 101, 0, 103, 0, 105, 0, 115, 0, 0, 0, 101, 0, 103, 0, 106, 0, 115, 0, 0, 0, 101, 0, 103, 0, 107, 0, 98, 0, 0, 0, 101, 0, 103, 0, 107, 0, 110, 0, 0, 0, 101, 0, 103, 0, 108, 0, 120, 0, 0, 0, 101, 0, 103, 0, 109, 0, 116, 0, 0, 0, 101, 0, 114, 0, 97, 0, 110, 0, 0, 0, 101, 0, 114, 0, 100, 0, 107, 0, 0, 0, 101, 0, 114, 0, 100, 0, 117, 0, 0, 0, 101, 0, 114, 0, 103, 0, 98, 0, 0, 0, 101, 0, 114, 0, 109, 0, 97, 0, 0, 0, 101, 0, 114, 0, 115, 0, 107, 0, 0, 0, 101, 0, 115, 0, 97, 0, 98, 0, 0, 0, 101, 0, 115, 0, 97, 0, 108, 0, 0, 0, 101, 0, 115, 0, 97, 0, 110, 0, 0, 0, 101, 0, 115, 0, 97, 0, 118, 0, 0, 0, 101, 0, 115, 0, 98, 0, 97, 0, 0, 0, 101, 0, 115, 0, 98, 0, 105, 0, 0, 0, 101, 0, 115, 0, 98, 0, 117, 0, 0, 0, 101, 0, 115, 0, 99, 0, 101, 0, 0, 0, 101, 0, 115, 0, 99, 0, 110, 0, 0, 0, 101, 0, 115, 0, 99, 0, 111, 0, 0, 0, 101, 0, 115, 0, 103, 0, 97, 0, 0, 0, 101, 0, 115, 0, 103, 0, 105, 0, 0, 0, 101, 0, 115, 0, 103, 0, 114, 0, 0, 0, 101, 0, 115, 0, 103, 0, 117, 0, 0, 0, 101, 0, 115, 0, 105, 0, 98, 0, 0, 0, 101, 0, 115, 0, 108, 0, 101, 0, 0, 0, 101, 0, 115, 0, 108, 0, 111, 0, 0, 0, 101, 0, 115, 0, 108, 0, 117, 0, 0, 0, 101, 0, 115, 0, 109, 0, 97, 0, 0, 0, 101, 0, 115, 0, 109, 0, 108, 0, 0, 0, 101, 0, 115, 0, 109, 0, 117, 0, 0, 0, 101, 0, 115, 0, 110, 0, 97, 0, 0, 0, 101, 0, 115, 0, 110, 0, 99, 0, 0, 0, 101, 0, 115, 0, 111, 0, 114, 0, 0, 0, 101, 0, 115, 0, 112, 0, 109, 0, 0, 0, 101, 0, 115, 0, 112, 0, 111, 0, 0, 0, 101, 0, 115, 0, 114, 0, 105, 0, 0, 0, 101, 0, 115, 0, 115, 0, 97, 0, 0, 0, 101, 0, 115, 0, 115, 0, 101, 0, 0, 0, 101, 0, 115, 0, 115, 0, 103, 0, 0, 0, 101, 0, 115, 0, 115, 0, 111, 0, 0, 0, 101, 0, 115, 0, 115, 0, 115, 0, 0, 0, 101, 0, 115, 0, 116, 0, 111, 0, 0, 0, 101, 0, 115, 0, 118, 0, 97, 0, 0, 0, 101, 0, 115, 0, 118, 0, 99, 0, 0, 0, 101, 0, 115, 0, 118, 0, 105, 0, 0, 0, 101, 0, 115, 0, 122, 0, 97, 0, 0, 0, 101, 0, 116, 0, 97, 0, 97, 0, 0, 0, 101, 0, 116, 0, 97, 0, 109, 0, 0, 0, 101, 0, 116, 0, 98, 0, 101, 0, 0, 0, 101, 0, 116, 0, 100, 0, 100, 0, 0, 0, 101, 0, 116, 0, 103, 0, 97, 0, 0, 0, 101, 0, 116, 0, 104, 0, 97, 0, 0, 0, 101, 0, 116, 0, 111, 0, 114, 0, 0, 0, 101, 0, 116, 0, 116, 0, 105, 0, 0, 0, 102, 0, 105, 0, 48, 0, 49, 0, 0, 0, 102, 0, 114, 0, 98, 0, 108, 0, 0, 0, 102, 0, 114, 0, 99, 0, 112, 0, 0, 0, 102, 0, 114, 0, 103, 0, 102, 0, 0, 0, 102, 0, 114, 0, 103, 0, 112, 0, 0, 0, 102, 0, 114, 0, 109, 0, 102, 0, 0, 0, 102, 0, 114, 0, 109, 0, 113, 0, 0, 0, 102, 0, 114, 0, 110, 0, 99, 0, 0, 0, 102, 0, 114, 0, 112, 0, 102, 0, 0, 0, 102, 0, 114, 0, 112, 0, 109, 0, 0, 0, 102, 0, 114, 0, 114, 0, 101, 0, 0, 0, 102, 0, 114, 0, 116, 0, 102, 0, 0, 0, 102, 0, 114, 0, 119, 0, 102, 0, 0, 0, 102, 0, 114, 0, 121, 0, 116, 0, 0, 0, 103, 0, 100, 0, 49, 0, 48, 0, 0, 0, 103, 0, 101, 0, 97, 0, 98, 0, 0, 0, 103, 0, 101, 0, 97, 0, 106, 0, 0, 0, 103, 0, 101, 0, 103, 0, 117, 0, 0, 0, 103, 0, 101, 0, 105, 0, 109, 0, 0, 0, 103, 0, 101, 0, 107, 0, 107, 0, 0, 0, 103, 0, 101, 0, 109, 0, 109, 0, 0, 0, 103, 0, 101, 0, 114, 0, 108, 0, 0, 0, 103, 0, 101, 0, 116, 0, 98, 0, 0, 0, 103, 0, 104, 0, 97, 0, 97, 0, 0, 0, 103, 0, 104, 0, 97, 0, 104, 0, 0, 0, 103, 0, 104, 0, 98, 0, 97, 0, 0, 0, 103, 0, 104, 0, 99, 0, 112, 0, 0, 0, 103, 0, 104, 0, 101, 0, 112, 0, 0, 0, 103, 0, 104, 0, 110, 0, 112, 0, 0, 0, 103, 0, 104, 0, 116, 0, 118, 0, 0, 0, 103, 0, 104, 0, 117, 0, 101, 0, 0, 0, 103, 0, 104, 0, 117, 0, 119, 0, 0, 0, 103, 0, 104, 0, 119, 0, 112, 0, 0, 0, 103, 0, 108, 0, 97, 0, 118, 0, 0, 0, 103, 0, 108, 0, 107, 0, 117, 0, 0, 0, 103, 0, 108, 0, 113, 0, 101, 0, 0, 0, 103, 0, 108, 0, 113, 0, 116, 0, 0, 0, 103, 0, 110, 0, 98, 0, 107, 0, 0, 0, 103, 0, 110, 0, 99, 0, 111, 0, 0, 0, 103, 0, 110, 0, 100, 0, 98, 0, 0, 0, 103, 0, 110, 0, 100, 0, 105, 0, 0, 0, 103, 0, 110, 0, 100, 0, 108, 0, 0, 0, 103, 0, 110, 0, 100, 0, 117, 0, 0, 0, 103, 0, 110, 0, 102, 0, 111, 0, 0, 0, 103, 0, 110, 0, 103, 0, 97, 0, 0, 0, 103, 0, 110, 0, 103, 0, 117, 0, 0, 0, 103, 0, 110, 0, 107, 0, 115, 0, 0, 0, 103, 0, 110, 0, 108, 0, 101, 0, 0, 0, 103, 0, 110, 0, 110, 0, 122, 0, 0, 0, 103, 0, 110, 0, 115, 0, 105, 0, 0, 0, 103, 0, 110, 0, 116, 0, 101, 0, 0, 0, 103, 0, 110, 0, 116, 0, 111, 0, 0, 0, 103, 0, 110, 0, 121, 0, 111, 0, 0, 0, 103, 0, 113, 0, 97, 0, 110, 0, 0, 0, 103, 0, 113, 0, 98, 0, 110, 0, 0, 0, 103, 0, 113, 0, 98, 0, 115, 0, 0, 0, 103, 0, 113, 0, 99, 0, 115, 0, 0, 0, 103, 0, 113, 0, 108, 0, 105, 0, 0, 0, 103, 0, 113, 0, 119, 0, 110, 0, 0, 0, 103, 0, 114, 0, 54, 0, 57, 0, 0, 0, 103, 0, 116, 0, 97, 0, 118, 0, 0, 0, 103, 0, 116, 0, 98, 0, 118, 0, 0, 0, 103, 0, 116, 0, 99, 0, 109, 0, 0, 0, 103, 0, 116, 0, 99, 0, 113, 0, 0, 0, 103, 0, 116, 0, 101, 0, 115, 0, 0, 0, 103, 0, 116, 0, 103, 0, 117, 0, 0, 0, 103, 0, 116, 0, 104, 0, 117, 0, 0, 0, 103, 0, 116, 0, 105, 0, 122, 0, 0, 0, 103, 0, 116, 0, 112, 0, 101, 0, 0, 0, 103, 0, 116, 0, 112, 0, 114, 0, 0, 0, 103, 0, 116, 0, 113, 0, 99, 0, 0, 0, 103, 0, 116, 0, 113, 0, 122, 0, 0, 0, 103, 0, 116, 0, 114, 0, 101, 0, 0, 0, 103, 0, 116, 0, 115, 0, 97, 0, 0, 0, 103, 0, 116, 0, 115, 0, 109, 0, 0, 0, 103, 0, 116, 0, 115, 0, 111, 0, 0, 0, 103, 0, 116, 0, 115, 0, 114, 0, 0, 0, 103, 0, 116, 0, 115, 0, 117, 0, 0, 0, 103, 0, 116, 0, 116, 0, 111, 0, 0, 0, 103, 0, 119, 0, 98, 0, 97, 0, 0, 0, 103, 0, 119, 0, 103, 0, 97, 0, 0, 0, 103, 0, 119, 0, 113, 0, 117, 0, 0, 0, 103, 0, 119, 0, 116, 0, 111, 0, 0, 0, 103, 0, 121, 0, 98, 0, 97, 0, 0, 0, 103, 0, 121, 0, 101, 0, 98, 0, 0, 0, 103, 0, 121, 0, 109, 0, 97, 0, 0, 0, 103, 0, 121, 0, 112, 0, 109, 0, 0, 0, 103, 0, 121, 0, 117, 0, 100, 0, 0, 0, 104, 0, 110, 0, 97, 0, 116, 0, 0, 0, 104, 0, 110, 0, 99, 0, 104, 0, 0, 0, 104, 0, 110, 0, 99, 0, 114, 0, 0, 0, 104, 0, 110, 0, 101, 0, 112, 0, 0, 0, 104, 0, 110, 0, 102, 0, 109, 0, 0, 0, 104, 0, 110, 0, 103, 0, 100, 0, 0, 0, 104, 0, 110, 0, 105, 0, 98, 0, 0, 0, 104, 0, 110, 0, 105, 0, 110, 0, 0, 0, 104, 0, 110, 0, 108, 0, 101, 0, 0, 0, 104, 0, 110, 0, 108, 0, 112, 0, 0, 0, 104, 0, 110, 0, 111, 0, 99, 0, 0, 0, 104, 0, 110, 0, 111, 0, 108, 0, 0, 0, 104, 0, 110, 0, 115, 0, 98, 0, 0, 0, 104, 0, 110, 0, 118, 0, 97, 0, 0, 0, 104, 0, 110, 0, 121, 0, 111, 0, 0, 0, 104, 0, 116, 0, 99, 0, 101, 0, 0, 0, 104, 0, 116, 0, 103, 0, 97, 0, 0, 0, 104, 0, 116, 0, 110, 0, 105, 0, 0, 0, 104, 0, 116, 0, 110, 0, 111, 0, 0, 0, 104, 0, 116, 0, 111, 0, 117, 0, 0, 0, 104, 0, 117, 0, 98, 0, 97, 0, 0, 0, 104, 0, 117, 0, 98, 0, 99, 0, 0, 0, 104, 0, 117, 0, 98, 0, 101, 0, 0, 0, 104, 0, 117, 0, 98, 0, 107, 0, 0, 0, 104, 0, 117, 0, 98, 0, 117, 0, 0, 0, 104, 0, 117, 0, 98, 0, 122, 0, 0, 0, 104, 0, 117, 0, 101, 0, 103, 0, 0, 0, 104, 0, 117, 0, 101, 0, 114, 0, 0, 0, 104, 0, 117, 0, 102, 0, 101, 0, 0, 0, 104, 0, 117, 0, 103, 0, 115, 0, 0, 0, 104, 0, 117, 0, 104, 0, 98, 0, 0, 0, 104, 0, 117, 0, 104, 0, 118, 0, 0, 0, 104, 0, 117, 0, 106, 0, 110, 0, 0, 0, 104, 0, 117, 0, 107, 0, 101, 0, 0, 0, 104, 0, 117, 0, 110, 0, 111, 0, 0, 0, 104, 0, 117, 0, 110, 0, 121, 0, 0, 0, 104, 0, 117, 0, 112, 0, 101, 0, 0, 0, 104, 0, 117, 0, 115, 0, 100, 0, 0, 0, 104, 0, 117, 0, 115, 0, 102, 0, 0, 0, 104, 0, 117, 0, 115, 0, 104, 0, 0, 0, 104, 0, 117, 0, 115, 0, 122, 0, 0, 0, 104, 0, 117, 0, 116, 0, 98, 0, 0, 0, 104, 0, 117, 0, 116, 0, 111, 0, 0, 0, 104, 0, 117, 0, 118, 0, 97, 0, 0, 0, 104, 0, 117, 0, 118, 0, 109, 0, 0, 0, 104, 0, 117, 0, 122, 0, 97, 0, 0, 0, 104, 0, 117, 0, 122, 0, 101, 0, 0, 0, 105, 0, 100, 0, 97, 0, 99, 0, 0, 0, 105, 0, 100, 0, 98, 0, 101, 0, 0, 0, 105, 0, 100, 0, 98, 0, 116, 0, 0, 0, 105, 0, 100, 0, 106, 0, 105, 0, 0, 0, 105, 0, 100, 0, 106, 0, 107, 0, 0, 0, 105, 0, 100, 0, 106, 0, 116, 0, 0, 0, 105, 0, 100, 0, 107, 0, 105, 0, 0, 0, 105, 0, 100, 0, 108, 0, 97, 0, 0, 0, 105, 0, 100, 0, 109, 0, 97, 0, 0, 0, 105, 0, 100, 0, 109, 0, 108, 0, 0, 0, 105, 0, 100, 0, 109, 0, 117, 0, 0, 0, 105, 0, 100, 0, 110, 0, 98, 0, 0, 0, 105, 0, 100, 0, 114, 0, 105, 0, 0, 0, 105, 0, 100, 0, 115, 0, 103, 0, 0, 0, 105, 0, 100, 0, 121, 0, 111, 0, 0, 0, 105, 0, 101, 0, 100, 0, 108, 0, 0, 0, 105, 0, 101, 0, 108, 0, 100, 0, 0, 0, 105, 0, 101, 0, 108, 0, 104, 0, 0, 0, 105, 0, 101, 0, 108, 0, 115, 0, 0, 0, 105, 0, 101, 0, 109, 0, 104, 0, 0, 0, 105, 0, 101, 0, 111, 0, 121, 0, 0, 0, 105, 0, 101, 0, 114, 0, 110, 0, 0, 0, 105, 0, 101, 0, 116, 0, 97, 0, 0, 0, 105, 0, 101, 0, 119, 0, 100, 0, 0, 0, 105, 0, 101, 0, 119, 0, 104, 0, 0, 0, 105, 0, 108, 0, 106, 0, 109, 0, 0, 0, 105, 0, 108, 0, 116, 0, 97, 0, 0, 0, 105, 0, 110, 0, 97, 0, 110, 0, 0, 0, 105, 0, 110, 0, 97, 0, 112, 0, 0, 0, 105, 0, 110, 0, 98, 0, 114, 0, 0, 0, 105, 0, 110, 0, 99, 0, 116, 0, 0, 0, 105, 0, 110, 0, 100, 0, 100, 0, 0, 0, 105, 0, 110, 0, 100, 0, 108, 0, 0, 0, 105, 0, 110, 0, 100, 0, 110, 0, 0, 0, 105, 0, 110, 0, 103, 0, 97, 0, 0, 0, 105, 0, 110, 0, 103, 0, 106, 0, 0, 0, 105, 0, 110, 0, 104, 0, 112, 0, 0, 0, 105, 0, 110, 0, 104, 0, 114, 0, 0, 0, 105, 0, 110, 0, 106, 0, 104, 0, 0, 0, 105, 0, 110, 0, 106, 0, 107, 0, 0, 0, 105, 0, 110, 0, 107, 0, 97, 0, 0, 0, 105, 0, 110, 0, 107, 0, 108, 0, 0, 0, 105, 0, 110, 0, 108, 0, 100, 0, 0, 0, 105, 0, 110, 0, 109, 0, 104, 0, 0, 0, 105, 0, 110, 0, 109, 0, 108, 0, 0, 0, 105, 0, 110, 0, 109, 0, 110, 0, 0, 0, 105, 0, 110, 0, 109, 0, 112, 0, 0, 0, 105, 0, 110, 0, 109, 0, 122, 0, 0, 0, 105, 0, 110, 0, 110, 0, 108, 0, 0, 0, 105, 0, 110, 0, 111, 0, 114, 0, 0, 0, 105, 0, 110, 0, 112, 0, 98, 0, 0, 0, 105, 0, 110, 0, 112, 0, 121, 0, 0, 0, 105, 0, 110, 0, 114, 0, 106, 0, 0, 0, 105, 0, 110, 0, 115, 0, 107, 0, 0, 0, 105, 0, 110, 0, 116, 0, 110, 0, 0, 0, 105, 0, 110, 0, 117, 0, 112, 0, 0, 0, 105, 0, 110, 0, 117, 0, 116, 0, 0, 0, 105, 0, 110, 0, 119, 0, 98, 0, 0, 0, 105, 0, 113, 0, 97, 0, 110, 0, 0, 0, 105, 0, 113, 0, 97, 0, 114, 0, 0, 0, 105, 0, 113, 0, 98, 0, 103, 0, 0, 0, 105, 0, 113, 0, 100, 0, 97, 0, 0, 0, 105, 0, 113, 0, 100, 0, 105, 0, 0, 0, 105, 0, 113, 0, 100, 0, 113, 0, 0, 0, 105, 0, 113, 0, 107, 0, 97, 0, 0, 0, 105, 0, 113, 0, 107, 0, 105, 0, 0, 0, 105, 0, 113, 0, 109, 0, 97, 0, 0, 0, 105, 0, 113, 0, 109, 0, 117, 0, 0, 0, 105, 0, 113, 0, 110, 0, 97, 0, 0, 0, 105, 0, 113, 0, 110, 0, 105, 0, 0, 0, 105, 0, 113, 0, 113, 0, 97, 0, 0, 0, 105, 0, 113, 0, 115, 0, 100, 0, 0, 0, 105, 0, 113, 0, 115, 0, 117, 0, 0, 0, 105, 0, 116, 0, 50, 0, 49, 0, 0, 0, 105, 0, 116, 0, 50, 0, 51, 0, 0, 0, 105, 0, 116, 0, 50, 0, 53, 0, 0, 0, 105, 0, 116, 0, 51, 0, 50, 0, 0, 0, 105, 0, 116, 0, 51, 0, 52, 0, 0, 0, 105, 0, 116, 0, 51, 0, 54, 0, 0, 0, 105, 0, 116, 0, 52, 0, 50, 0, 0, 0, 105, 0, 116, 0, 52, 0, 53, 0, 0, 0, 105, 0, 116, 0, 53, 0, 50, 0, 0, 0, 105, 0, 116, 0, 53, 0, 53, 0, 0, 0, 105, 0, 116, 0, 53, 0, 55, 0, 0, 0, 105, 0, 116, 0, 54, 0, 50, 0, 0, 0, 105, 0, 116, 0, 54, 0, 53, 0, 0, 0, 105, 0, 116, 0, 54, 0, 55, 0, 0, 0, 105, 0, 116, 0, 55, 0, 50, 0, 0, 0, 105, 0, 116, 0, 55, 0, 53, 0, 0, 0, 105, 0, 116, 0, 56, 0, 50, 0, 0, 0, 105, 0, 116, 0, 56, 0, 56, 0, 0, 0, 105, 0, 116, 0, 97, 0, 103, 0, 0, 0, 105, 0, 116, 0, 97, 0, 116, 0, 0, 0, 105, 0, 116, 0, 97, 0, 118, 0, 0, 0, 105, 0, 116, 0, 98, 0, 97, 0, 0, 0, 105, 0, 116, 0, 98, 0, 103, 0, 0, 0, 105, 0, 116, 0, 98, 0, 105, 0, 0, 0, 105, 0, 116, 0, 98, 0, 108, 0, 0, 0, 105, 0, 116, 0, 98, 0, 122, 0, 0, 0, 105, 0, 116, 0, 99, 0, 101, 0, 0, 0, 105, 0, 116, 0, 99, 0, 108, 0, 0, 0, 105, 0, 116, 0, 99, 0, 122, 0, 0, 0, 105, 0, 116, 0, 101, 0, 110, 0, 0, 0, 105, 0, 116, 0, 102, 0, 99, 0, 0, 0, 105, 0, 116, 0, 102, 0, 101, 0, 0, 0, 105, 0, 116, 0, 102, 0, 103, 0, 0, 0, 105, 0, 116, 0, 102, 0, 109, 0, 0, 0, 105, 0, 116, 0, 103, 0, 101, 0, 0, 0, 105, 0, 116, 0, 103, 0, 114, 0, 0, 0, 105, 0, 116, 0, 105, 0, 109, 0, 0, 0, 105, 0, 116, 0, 105, 0, 115, 0, 0, 0, 105, 0, 116, 0, 107, 0, 114, 0, 0, 0, 105, 0, 116, 0, 108, 0, 99, 0, 0, 0, 105, 0, 116, 0, 108, 0, 101, 0, 0, 0, 105, 0, 116, 0, 108, 0, 105, 0, 0, 0, 105, 0, 116, 0, 108, 0, 111, 0, 0, 0, 105, 0, 116, 0, 109, 0, 101, 0, 0, 0, 105, 0, 116, 0, 109, 0, 105, 0, 0, 0, 105, 0, 116, 0, 110, 0, 97, 0, 0, 0, 105, 0, 116, 0, 110, 0, 111, 0, 0, 0, 105, 0, 116, 0, 110, 0, 117, 0, 0, 0, 105, 0, 116, 0, 111, 0, 114, 0, 0, 0, 105, 0, 116, 0, 112, 0, 103, 0, 0, 0, 105, 0, 116, 0, 112, 0, 105, 0, 0, 0, 105, 0, 116, 0, 112, 0, 114, 0, 0, 0, 105, 0, 116, 0, 112, 0, 122, 0, 0, 0, 105, 0, 116, 0, 114, 0, 97, 0, 0, 0, 105, 0, 116, 0, 114, 0, 99, 0, 0, 0, 105, 0, 116, 0, 114, 0, 101, 0, 0, 0, 105, 0, 116, 0, 114, 0, 103, 0, 0, 0, 105, 0, 116, 0, 114, 0, 105, 0, 0, 0, 105, 0, 116, 0, 115, 0, 97, 0, 0, 0, 105, 0, 116, 0, 115, 0, 105, 0, 0, 0, 105, 0, 116, 0, 115, 0, 118, 0, 0, 0, 105, 0, 116, 0, 116, 0, 97, 0, 0, 0, 105, 0, 116, 0, 116, 0, 101, 0, 0, 0, 105, 0, 116, 0, 116, 0, 118, 0, 0, 0, 105, 0, 116, 0, 118, 0, 101, 0, 0, 0, 105, 0, 116, 0, 118, 0, 105, 0, 0, 0, 105, 0, 116, 0, 118, 0, 118, 0, 0, 0, 106, 0, 111, 0, 97, 0, 106, 0, 0, 0, 106, 0, 111, 0, 97, 0, 109, 0, 0, 0, 106, 0, 111, 0, 97, 0, 113, 0, 0, 0, 106, 0, 111, 0, 97, 0, 116, 0, 0, 0, 106, 0, 111, 0, 97, 0, 122, 0, 0, 0, 106, 0, 111, 0, 98, 0, 97, 0, 0, 0, 106, 0, 111, 0, 105, 0, 114, 0, 0, 0, 106, 0, 111, 0, 106, 0, 97, 0, 0, 0, 106, 0, 111, 0, 107, 0, 97, 0, 0, 0, 106, 0, 111, 0, 109, 0, 97, 0, 0, 0, 106, 0, 111, 0, 109, 0, 100, 0, 0, 0, 106, 0, 111, 0, 109, 0, 110, 0, 0, 0, 107, 0, 103, 0, 103, 0, 98, 0, 0, 0, 107, 0, 103, 0, 103, 0, 111, 0, 0, 0, 107, 0, 110, 0, 49, 0, 53, 0, 0, 0, 107, 0, 112, 0, 49, 0, 48, 0, 0, 0, 107, 0, 114, 0, 49, 0, 49, 0, 0, 0, 107, 0, 114, 0, 53, 0, 48, 0, 0, 0, 107, 0, 119, 0, 97, 0, 104, 0, 0, 0, 107, 0, 119, 0, 102, 0, 97, 0, 0, 0, 107, 0, 119, 0, 104, 0, 97, 0, 0, 0, 107, 0, 119, 0, 109, 0, 117, 0, 0, 0, 108, 0, 97, 0, 104, 0, 111, 0, 0, 0, 108, 0, 97, 0, 107, 0, 104, 0, 0, 0, 108, 0, 97, 0, 108, 0, 109, 0, 0, 0, 108, 0, 97, 0, 108, 0, 112, 0, 0, 0, 108, 0, 97, 0, 111, 0, 117, 0, 0, 0, 108, 0, 97, 0, 112, 0, 104, 0, 0, 0, 108, 0, 97, 0, 115, 0, 108, 0, 0, 0, 108, 0, 97, 0, 115, 0, 118, 0, 0, 0, 108, 0, 97, 0, 118, 0, 116, 0, 0, 0, 108, 0, 97, 0, 120, 0, 97, 0, 0, 0, 108, 0, 97, 0, 120, 0, 105, 0, 0, 0, 108, 0, 97, 0, 120, 0, 115, 0, 0, 0, 108, 0, 98, 0, 97, 0, 107, 0, 0, 0, 108, 0, 98, 0, 97, 0, 115, 0, 0, 0, 108, 0, 98, 0, 98, 0, 97, 0, 0, 0, 108, 0, 98, 0, 106, 0, 97, 0, 0, 0, 108, 0, 98, 0, 106, 0, 108, 0, 0, 0, 108, 0, 98, 0, 110, 0, 97, 0, 0, 0, 108, 0, 114, 0, 98, 0, 103, 0, 0, 0, 108, 0, 114, 0, 98, 0, 109, 0, 0, 0, 108, 0, 114, 0, 99, 0, 109, 0, 0, 0, 108, 0, 114, 0, 103, 0, 98, 0, 0, 0, 108, 0, 114, 0, 103, 0, 103, 0, 0, 0, 108, 0, 114, 0, 103, 0, 112, 0, 0, 0, 108, 0, 114, 0, 108, 0, 111, 0, 0, 0, 108, 0, 114, 0, 109, 0, 103, 0, 0, 0, 108, 0, 114, 0, 109, 0, 111, 0, 0, 0, 108, 0, 114, 0, 110, 0, 105, 0, 0, 0, 108, 0, 114, 0, 114, 0, 103, 0, 0, 0, 108, 0, 114, 0, 115, 0, 105, 0, 0, 0, 108, 0, 116, 0, 54, 0, 48, 0, 0, 0, 108, 0, 116, 0, 97, 0, 108, 0, 0, 0, 108, 0, 116, 0, 107, 0, 108, 0, 0, 0, 108, 0, 116, 0, 107, 0, 117, 0, 0, 0, 108, 0, 116, 0, 109, 0, 114, 0, 0, 0, 108, 0, 116, 0, 112, 0, 110, 0, 0, 0, 108, 0, 116, 0, 115, 0, 97, 0, 0, 0, 108, 0, 116, 0, 116, 0, 97, 0, 0, 0, 108, 0, 116, 0, 116, 0, 101, 0, 0, 0, 108, 0, 116, 0, 117, 0, 116, 0, 0, 0, 108, 0, 116, 0, 118, 0, 108, 0, 0, 0, 108, 0, 117, 0, 99, 0, 97, 0, 0, 0, 108, 0, 117, 0, 99, 0, 108, 0, 0, 0, 108, 0, 117, 0, 100, 0, 105, 0, 0, 0, 108, 0, 117, 0, 101, 0, 99, 0, 0, 0, 108, 0, 117, 0, 103, 0, 114, 0, 0, 0, 108, 0, 117, 0, 109, 0, 101, 0, 0, 0, 108, 0, 117, 0, 114, 0, 100, 0, 0, 0, 108, 0, 117, 0, 118, 0, 100, 0, 0, 0, 108, 0, 117, 0, 119, 0, 105, 0, 0, 0, 108, 0, 121, 0, 98, 0, 97, 0, 0, 0, 108, 0, 121, 0, 98, 0, 117, 0, 0, 0, 108, 0, 121, 0, 100, 0, 114, 0, 0, 0, 108, 0, 121, 0, 103, 0, 116, 0, 0, 0, 108, 0, 121, 0, 106, 0, 103, 0, 0, 0, 108, 0, 121, 0, 106, 0, 105, 0, 0, 0, 108, 0, 121, 0, 106, 0, 117, 0, 0, 0, 108, 0, 121, 0, 107, 0, 102, 0, 0, 0, 108, 0, 121, 0, 109, 0, 98, 0, 0, 0, 108, 0, 121, 0, 109, 0, 113, 0, 0, 0, 108, 0, 121, 0, 115, 0, 98, 0, 0, 0, 108, 0, 121, 0, 115, 0, 114, 0, 0, 0, 108, 0, 121, 0, 116, 0, 98, 0, 0, 0, 108, 0, 121, 0, 119, 0, 100, 0, 0, 0, 108, 0, 121, 0, 119, 0, 115, 0, 0, 0, 108, 0, 121, 0, 122, 0, 97, 0, 0, 0, 109, 0, 99, 0, 99, 0, 108, 0, 0, 0, 109, 0, 99, 0, 99, 0, 111, 0, 0, 0, 109, 0, 99, 0, 102, 0, 111, 0, 0, 0, 109, 0, 99, 0, 109, 0, 117, 0, 0, 0, 109, 0, 99, 0, 112, 0, 104, 0, 0, 0, 109, 0, 99, 0, 115, 0, 100, 0, 0, 0, 109, 0, 99, 0, 115, 0, 114, 0, 0, 0, 109, 0, 99, 0, 118, 0, 114, 0, 0, 0, 109, 0, 100, 0, 97, 0, 110, 0, 0, 0, 109, 0, 100, 0, 98, 0, 97, 0, 0, 0, 109, 0, 100, 0, 98, 0, 100, 0, 0, 0, 109, 0, 100, 0, 99, 0, 97, 0, 0, 0, 109, 0, 100, 0, 100, 0, 111, 0, 0, 0, 109, 0, 100, 0, 100, 0, 114, 0, 0, 0, 109, 0, 100, 0, 100, 0, 117, 0, 0, 0, 109, 0, 100, 0, 101, 0, 100, 0, 0, 0, 109, 0, 100, 0, 102, 0, 97, 0, 0, 0, 109, 0, 100, 0, 102, 0, 108, 0, 0, 0, 109, 0, 100, 0, 103, 0, 97, 0, 0, 0, 109, 0, 100, 0, 103, 0, 108, 0, 0, 0, 109, 0, 100, 0, 105, 0, 97, 0, 0, 0, 109, 0, 100, 0, 108, 0, 101, 0, 0, 0, 109, 0, 100, 0, 110, 0, 105, 0, 0, 0, 109, 0, 100, 0, 111, 0, 99, 0, 0, 0, 109, 0, 100, 0, 111, 0, 114, 0, 0, 0, 109, 0, 100, 0, 114, 0, 101, 0, 0, 0, 109, 0, 100, 0, 115, 0, 100, 0, 0, 0, 109, 0, 100, 0, 115, 0, 105, 0, 0, 0, 109, 0, 100, 0, 115, 0, 116, 0, 0, 0, 109, 0, 100, 0, 115, 0, 118, 0, 0, 0, 109, 0, 100, 0, 116, 0, 97, 0, 0, 0, 109, 0, 100, 0, 116, 0, 101, 0, 0, 0, 109, 0, 100, 0, 117, 0, 110, 0, 0, 0, 109, 0, 107, 0, 49, 0, 54, 0, 0, 0, 109, 0, 107, 0, 51, 0, 48, 0, 0, 0, 109, 0, 107, 0, 54, 0, 57, 0, 0, 0, 109, 0, 107, 0, 55, 0, 56, 0, 0, 0, 109, 0, 107, 0, 56, 0, 53, 0, 0, 0, 109, 0, 108, 0, 49, 0, 48, 0, 0, 0, 109, 0, 117, 0, 97, 0, 103, 0, 0, 0, 109, 0, 117, 0, 99, 0, 99, 0, 0, 0, 109, 0, 117, 0, 99, 0, 117, 0, 0, 0, 109, 0, 117, 0, 102, 0, 108, 0, 0, 0, 109, 0, 117, 0, 103, 0, 112, 0, 0, 0, 109, 0, 117, 0, 109, 0, 111, 0, 0, 0, 109, 0, 117, 0, 112, 0, 97, 0, 0, 0, 109, 0, 117, 0, 112, 0, 108, 0, 0, 0, 109, 0, 117, 0, 112, 0, 117, 0, 0, 0, 109, 0, 117, 0, 112, 0, 119, 0, 0, 0, 109, 0, 117, 0, 113, 0, 98, 0, 0, 0, 109, 0, 117, 0, 114, 0, 111, 0, 0, 0, 109, 0, 117, 0, 115, 0, 97, 0, 0, 0, 109, 0, 117, 0, 118, 0, 112, 0, 0, 0, 109, 0, 118, 0, 49, 0, 55, 0, 0, 0, 109, 0, 118, 0, 50, 0, 48, 0, 0, 0, 109, 0, 119, 0, 98, 0, 97, 0, 0, 0, 109, 0, 119, 0, 98, 0, 108, 0, 0, 0, 109, 0, 119, 0, 99, 0, 107, 0, 0, 0, 109, 0, 119, 0, 99, 0, 114, 0, 0, 0, 109, 0, 119, 0, 99, 0, 116, 0, 0, 0, 109, 0, 119, 0, 100, 0, 101, 0, 0, 0, 109, 0, 119, 0, 100, 0, 111, 0, 0, 0, 109, 0, 119, 0, 108, 0, 107, 0, 0, 0, 109, 0, 119, 0, 109, 0, 99, 0, 0, 0, 109, 0, 119, 0, 109, 0, 117, 0, 0, 0, 109, 0, 119, 0, 109, 0, 119, 0, 0, 0, 109, 0, 119, 0, 109, 0, 122, 0, 0, 0, 109, 0, 119, 0, 110, 0, 115, 0, 0, 0, 109, 0, 119, 0, 112, 0, 104, 0, 0, 0, 109, 0, 119, 0, 122, 0, 111, 0, 0, 0, 110, 0, 97, 0, 99, 0, 97, 0, 0, 0, 110, 0, 97, 0, 101, 0, 114, 0, 0, 0, 110, 0, 97, 0, 104, 0, 97, 0, 0, 0, 110, 0, 97, 0, 107, 0, 97, 0, 0, 0, 110, 0, 97, 0, 107, 0, 101, 0, 0, 0, 110, 0, 97, 0, 107, 0, 104, 0, 0, 0, 110, 0, 97, 0, 107, 0, 117, 0, 0, 0, 110, 0, 97, 0, 107, 0, 119, 0, 0, 0, 110, 0, 97, 0, 111, 0, 100, 0, 0, 0, 110, 0, 97, 0, 111, 0, 104, 0, 0, 0, 110, 0, 97, 0, 111, 0, 110, 0, 0, 0, 110, 0, 97, 0, 111, 0, 119, 0, 0, 0, 110, 0, 103, 0, 97, 0, 98, 0, 0, 0, 110, 0, 103, 0, 97, 0, 100, 0, 0, 0, 110, 0, 103, 0, 97, 0, 107, 0, 0, 0, 110, 0, 103, 0, 98, 0, 97, 0, 0, 0, 110, 0, 103, 0, 98, 0, 101, 0, 0, 0, 110, 0, 103, 0, 98, 0, 121, 0, 0, 0, 110, 0, 103, 0, 100, 0, 101, 0, 0, 0, 110, 0, 103, 0, 101, 0, 98, 0, 0, 0, 110, 0, 103, 0, 101, 0, 100, 0, 0, 0, 110, 0, 103, 0, 101, 0, 107, 0, 0, 0, 110, 0, 103, 0, 101, 0, 110, 0, 0, 0, 110, 0, 103, 0, 102, 0, 99, 0, 0, 0, 110, 0, 103, 0, 103, 0, 111, 0, 0, 0, 110, 0, 103, 0, 105, 0, 109, 0, 0, 0, 110, 0, 103, 0, 107, 0, 116, 0, 0, 0, 110, 0, 103, 0, 107, 0, 119, 0, 0, 0, 110, 0, 103, 0, 108, 0, 97, 0, 0, 0, 110, 0, 103, 0, 110, 0, 97, 0, 0, 0, 110, 0, 103, 0, 110, 0, 105, 0, 0, 0, 110, 0, 103, 0, 111, 0, 103, 0, 0, 0, 110, 0, 103, 0, 111, 0, 110, 0, 0, 0, 110, 0, 103, 0, 111, 0, 115, 0, 0, 0, 110, 0, 103, 0, 111, 0, 121, 0, 0, 0, 110, 0, 103, 0, 112, 0, 108, 0, 0, 0, 110, 0, 103, 0, 115, 0, 111, 0, 0, 0, 110, 0, 103, 0, 121, 0, 111, 0, 0, 0, 110, 0, 103, 0, 122, 0, 97, 0, 0, 0, 110, 0, 105, 0, 97, 0, 110, 0, 0, 0, 110, 0, 105, 0, 97, 0, 115, 0, 0, 0, 110, 0, 105, 0, 98, 0, 111, 0, 0, 0, 110, 0, 105, 0, 99, 0, 105, 0, 0, 0, 110, 0, 105, 0, 99, 0, 111, 0, 0, 0, 110, 0, 105, 0, 101, 0, 115, 0, 0, 0, 110, 0, 105, 0, 103, 0, 114, 0, 0, 0, 110, 0, 105, 0, 106, 0, 105, 0, 0, 0, 110, 0, 105, 0, 108, 0, 101, 0, 0, 0, 110, 0, 105, 0, 109, 0, 100, 0, 0, 0, 110, 0, 105, 0, 109, 0, 110, 0, 0, 0, 110, 0, 105, 0, 110, 0, 115, 0, 0, 0, 110, 0, 105, 0, 114, 0, 105, 0, 0, 0, 110, 0, 105, 0, 115, 0, 106, 0, 0, 0, 110, 0, 108, 0, 97, 0, 119, 0, 0, 0, 110, 0, 108, 0, 99, 0, 119, 0, 0, 0, 110, 0, 108, 0, 100, 0, 114, 0, 0, 0, 110, 0, 108, 0, 102, 0, 108, 0, 0, 0, 110, 0, 108, 0, 102, 0, 114, 0, 0, 0, 110, 0, 108, 0, 103, 0, 101, 0, 0, 0, 110, 0, 108, 0, 103, 0, 114, 0, 0, 0, 110, 0, 108, 0, 108, 0, 105, 0, 0, 0, 110, 0, 108, 0, 110, 0, 98, 0, 0, 0, 110, 0, 108, 0, 110, 0, 104, 0, 0, 0, 110, 0, 108, 0, 111, 0, 118, 0, 0, 0, 110, 0, 108, 0, 115, 0, 120, 0, 0, 0, 110, 0, 108, 0, 117, 0, 116, 0, 0, 0, 110, 0, 108, 0, 122, 0, 101, 0, 0, 0, 110, 0, 112, 0, 98, 0, 97, 0, 0, 0, 110, 0, 112, 0, 98, 0, 104, 0, 0, 0, 110, 0, 112, 0, 100, 0, 104, 0, 0, 0, 110, 0, 112, 0, 106, 0, 97, 0, 0, 0, 110, 0, 112, 0, 107, 0, 97, 0, 0, 0, 110, 0, 112, 0, 108, 0, 117, 0, 0, 0, 110, 0, 112, 0, 109, 0, 97, 0, 0, 0, 110, 0, 112, 0, 109, 0, 101, 0, 0, 0, 111, 0, 109, 0, 98, 0, 106, 0, 0, 0, 111, 0, 109, 0, 98, 0, 115, 0, 0, 0, 111, 0, 109, 0, 98, 0, 117, 0, 0, 0, 111, 0, 109, 0, 100, 0, 97, 0, 0, 0, 111, 0, 109, 0, 109, 0, 97, 0, 0, 0, 111, 0, 109, 0, 109, 0, 117, 0, 0, 0, 111, 0, 109, 0, 115, 0, 106, 0, 0, 0, 111, 0, 109, 0, 115, 0, 115, 0, 0, 0, 111, 0, 109, 0, 119, 0, 117, 0, 0, 0, 111, 0, 109, 0, 122, 0, 117, 0, 0, 0, 112, 0, 97, 0, 49, 0, 48, 0, 0, 0, 112, 0, 97, 0, 101, 0, 109, 0, 0, 0, 112, 0, 97, 0, 107, 0, 121, 0, 0, 0, 112, 0, 97, 0, 110, 0, 98, 0, 0, 0, 112, 0, 107, 0, 98, 0, 97, 0, 0, 0, 112, 0, 107, 0, 105, 0, 115, 0, 0, 0, 112, 0, 107, 0, 106, 0, 107, 0, 0, 0, 112, 0, 107, 0, 107, 0, 112, 0, 0, 0, 112, 0, 107, 0, 112, 0, 98, 0, 0, 0, 112, 0, 107, 0, 115, 0, 100, 0, 0, 0, 112, 0, 107, 0, 116, 0, 97, 0, 0, 0, 112, 0, 108, 0, 48, 0, 50, 0, 0, 0, 112, 0, 108, 0, 48, 0, 52, 0, 0, 0, 112, 0, 108, 0, 48, 0, 54, 0, 0, 0, 112, 0, 108, 0, 48, 0, 56, 0, 0, 0, 112, 0, 108, 0, 49, 0, 48, 0, 0, 0, 112, 0, 108, 0, 49, 0, 50, 0, 0, 0, 112, 0, 108, 0, 49, 0, 52, 0, 0, 0, 112, 0, 108, 0, 49, 0, 54, 0, 0, 0, 112, 0, 108, 0, 49, 0, 56, 0, 0, 0, 112, 0, 108, 0, 50, 0, 48, 0, 0, 0, 112, 0, 108, 0, 50, 0, 50, 0, 0, 0, 112, 0, 108, 0, 50, 0, 52, 0, 0, 0, 112, 0, 108, 0, 50, 0, 54, 0, 0, 0, 112, 0, 108, 0, 50, 0, 56, 0, 0, 0, 112, 0, 108, 0, 51, 0, 48, 0, 0, 0, 112, 0, 108, 0, 51, 0, 50, 0, 0, 0, 112, 0, 116, 0, 50, 0, 48, 0, 0, 0, 112, 0, 116, 0, 51, 0, 48, 0, 0, 0, 112, 0, 121, 0, 49, 0, 57, 0, 0, 0, 113, 0, 97, 0, 100, 0, 97, 0, 0, 0, 113, 0, 97, 0, 107, 0, 104, 0, 0, 0, 113, 0, 97, 0, 109, 0, 115, 0, 0, 0, 113, 0, 97, 0, 114, 0, 97, 0, 0, 0, 113, 0, 97, 0, 115, 0, 104, 0, 0, 0, 113, 0, 97, 0, 117, 0, 115, 0, 0, 0, 113, 0, 97, 0, 119, 0, 97, 0, 0, 0, 113, 0, 97, 0, 122, 0, 97, 0, 0, 0, 114, 0, 111, 0, 97, 0, 98, 0, 0, 0, 114, 0, 111, 0, 97, 0, 103, 0, 0, 0, 114, 0, 111, 0, 98, 0, 99, 0, 0, 0, 114, 0, 111, 0, 98, 0, 104, 0, 0, 0, 114, 0, 111, 0, 98, 0, 110, 0, 0, 0, 114, 0, 111, 0, 98, 0, 114, 0, 0, 0, 114, 0, 111, 0, 98, 0, 116, 0, 0, 0, 114, 0, 111, 0, 98, 0, 118, 0, 0, 0, 114, 0, 111, 0, 98, 0, 122, 0, 0, 0, 114, 0, 111, 0, 99, 0, 106, 0, 0, 0, 114, 0, 111, 0, 99, 0, 108, 0, 0, 0, 114, 0, 111, 0, 100, 0, 98, 0, 0, 0, 114, 0, 111, 0, 100, 0, 106, 0, 0, 0, 114, 0, 111, 0, 103, 0, 106, 0, 0, 0, 114, 0, 111, 0, 103, 0, 108, 0, 0, 0, 114, 0, 111, 0, 103, 0, 114, 0, 0, 0, 114, 0, 111, 0, 104, 0, 100, 0, 0, 0, 114, 0, 111, 0, 104, 0, 114, 0, 0, 0, 114, 0, 111, 0, 105, 0, 102, 0, 0, 0, 114, 0, 111, 0, 105, 0, 108, 0, 0, 0, 114, 0, 111, 0, 105, 0, 115, 0, 0, 0, 114, 0, 111, 0, 109, 0, 104, 0, 0, 0, 114, 0, 111, 0, 109, 0, 109, 0, 0, 0, 114, 0, 111, 0, 109, 0, 115, 0, 0, 0, 114, 0, 111, 0, 110, 0, 116, 0, 0, 0, 114, 0, 111, 0, 111, 0, 116, 0, 0, 0, 114, 0, 111, 0, 112, 0, 104, 0, 0, 0, 114, 0, 111, 0, 115, 0, 98, 0, 0, 0, 114, 0, 111, 0, 115, 0, 106, 0, 0, 0, 114, 0, 111, 0, 115, 0, 109, 0, 0, 0, 114, 0, 111, 0, 115, 0, 118, 0, 0, 0, 114, 0, 111, 0, 116, 0, 114, 0, 0, 0, 114, 0, 111, 0, 118, 0, 108, 0, 0, 0, 114, 0, 111, 0, 118, 0, 110, 0, 0, 0, 114, 0, 111, 0, 118, 0, 115, 0, 0, 0, 114, 0, 115, 0, 107, 0, 109, 0, 0, 0, 114, 0, 117, 0, 97, 0, 100, 0, 0, 0, 114, 0, 117, 0, 97, 0, 108, 0, 0, 0, 114, 0, 117, 0, 99, 0, 101, 0, 0, 0, 114, 0, 117, 0, 99, 0, 117, 0, 0, 0, 114, 0, 117, 0, 105, 0, 110, 0, 0, 0, 114, 0, 117, 0, 107, 0, 111, 0, 0, 0, 114, 0, 117, 0, 107, 0, 114, 0, 0, 0, 114, 0, 117, 0, 109, 0, 101, 0, 0, 0, 114, 0, 117, 0, 109, 0, 111, 0, 0, 0, 114, 0, 117, 0, 115, 0, 101, 0, 0, 0, 114, 0, 117, 0, 116, 0, 121, 0, 0, 0, 114, 0, 117, 0, 117, 0, 100, 0, 0, 0, 115, 0, 97, 0, 49, 0, 52, 0, 0, 0, 115, 0, 98, 0, 99, 0, 101, 0, 0, 0, 115, 0, 98, 0, 99, 0, 104, 0, 0, 0, 115, 0, 98, 0, 99, 0, 116, 0, 0, 0, 115, 0, 98, 0, 103, 0, 117, 0, 0, 0, 115, 0, 98, 0, 105, 0, 115, 0, 0, 0, 115, 0, 98, 0, 114, 0, 98, 0, 0, 0, 115, 0, 98, 0, 116, 0, 101, 0, 0, 0, 115, 0, 98, 0, 119, 0, 101, 0, 0, 0, 115, 0, 100, 0, 100, 0, 99, 0, 0, 0, 115, 0, 100, 0, 100, 0, 101, 0, 0, 0, 115, 0, 100, 0, 100, 0, 110, 0, 0, 0, 115, 0, 100, 0, 100, 0, 115, 0, 0, 0, 115, 0, 100, 0, 103, 0, 100, 0, 0, 0, 115, 0, 100, 0, 103, 0, 107, 0, 0, 0, 115, 0, 100, 0, 107, 0, 97, 0, 0, 0, 115, 0, 100, 0, 107, 0, 104, 0, 0, 0, 115, 0, 100, 0, 107, 0, 110, 0, 0, 0, 115, 0, 100, 0, 107, 0, 115, 0, 0, 0, 115, 0, 100, 0, 110, 0, 98, 0, 0, 0, 115, 0, 100, 0, 110, 0, 111, 0, 0, 0, 115, 0, 100, 0, 110, 0, 119, 0, 0, 0, 115, 0, 100, 0, 114, 0, 115, 0, 0, 0, 115, 0, 100, 0, 115, 0, 105, 0, 0, 0, 115, 0, 101, 0, 98, 0, 100, 0, 0, 0, 115, 0, 104, 0, 97, 0, 99, 0, 0, 0, 115, 0, 104, 0, 104, 0, 108, 0, 0, 0, 115, 0, 104, 0, 116, 0, 97, 0, 0, 0, 115, 0, 107, 0, 98, 0, 99, 0, 0, 0, 115, 0, 107, 0, 98, 0, 108, 0, 0, 0, 115, 0, 107, 0, 107, 0, 105, 0, 0, 0, 115, 0, 107, 0, 110, 0, 105, 0, 0, 0, 115, 0, 107, 0, 116, 0, 97, 0, 0, 0, 115, 0, 107, 0, 116, 0, 99, 0, 0, 0, 115, 0, 110, 0, 100, 0, 98, 0, 0, 0, 115, 0, 110, 0, 100, 0, 107, 0, 0, 0, 115, 0, 110, 0, 102, 0, 107, 0, 0, 0, 115, 0, 110, 0, 107, 0, 97, 0, 0, 0, 115, 0, 110, 0, 107, 0, 108, 0, 0, 0, 115, 0, 110, 0, 109, 0, 116, 0, 0, 0, 115, 0, 110, 0, 115, 0, 101, 0, 0, 0, 115, 0, 110, 0, 115, 0, 108, 0, 0, 0, 115, 0, 110, 0, 116, 0, 99, 0, 0, 0, 115, 0, 110, 0, 116, 0, 104, 0, 0, 0, 115, 0, 110, 0, 122, 0, 103, 0, 0, 0, 115, 0, 111, 0, 97, 0, 119, 0, 0, 0, 115, 0, 111, 0, 98, 0, 107, 0, 0, 0, 115, 0, 111, 0, 98, 0, 110, 0, 0, 0, 115, 0, 111, 0, 98, 0, 121, 0, 0, 0, 115, 0, 111, 0, 103, 0, 97, 0, 0, 0, 115, 0, 111, 0, 104, 0, 105, 0, 0, 0, 115, 0, 111, 0, 106, 0, 100, 0, 0, 0, 115, 0, 111, 0, 106, 0, 104, 0, 0, 0, 115, 0, 111, 0, 109, 0, 117, 0, 0, 0, 115, 0, 111, 0, 110, 0, 117, 0, 0, 0, 115, 0, 111, 0, 115, 0, 100, 0, 0, 0, 115, 0, 111, 0, 115, 0, 104, 0, 0, 0, 115, 0, 111, 0, 115, 0, 111, 0, 0, 0, 115, 0, 111, 0, 116, 0, 111, 0, 0, 0, 115, 0, 111, 0, 119, 0, 111, 0, 0, 0, 115, 0, 114, 0, 98, 0, 114, 0, 0, 0, 115, 0, 114, 0, 99, 0, 109, 0, 0, 0, 115, 0, 114, 0, 99, 0, 114, 0, 0, 0, 115, 0, 114, 0, 109, 0, 97, 0, 0, 0, 115, 0, 114, 0, 110, 0, 105, 0, 0, 0, 115, 0, 114, 0, 112, 0, 109, 0, 0, 0, 115, 0, 114, 0, 112, 0, 114, 0, 0, 0, 115, 0, 114, 0, 115, 0, 97, 0, 0, 0, 115, 0, 115, 0, 98, 0, 110, 0, 0, 0, 115, 0, 115, 0, 98, 0, 119, 0, 0, 0, 115, 0, 115, 0, 101, 0, 101, 0, 0, 0, 115, 0, 115, 0, 108, 0, 107, 0, 0, 0, 115, 0, 115, 0, 110, 0, 117, 0, 0, 0, 115, 0, 115, 0, 117, 0, 121, 0, 0, 0, 115, 0, 115, 0, 119, 0, 114, 0, 0, 0, 115, 0, 118, 0, 97, 0, 104, 0, 0, 0, 115, 0, 118, 0, 99, 0, 97, 0, 0, 0, 115, 0, 118, 0, 99, 0, 104, 0, 0, 0, 115, 0, 118, 0, 99, 0, 117, 0, 0, 0, 115, 0, 118, 0, 108, 0, 105, 0, 0, 0, 115, 0, 118, 0, 109, 0, 111, 0, 0, 0, 115, 0, 118, 0, 112, 0, 97, 0, 0, 0, 115, 0, 118, 0, 115, 0, 97, 0, 0, 0, 115, 0, 118, 0, 115, 0, 109, 0, 0, 0, 115, 0, 118, 0, 115, 0, 111, 0, 0, 0, 115, 0, 118, 0, 117, 0, 110, 0, 0, 0, 115, 0, 118, 0, 117, 0, 115, 0, 0, 0, 115, 0, 121, 0, 100, 0, 105, 0, 0, 0, 115, 0, 121, 0, 100, 0, 114, 0, 0, 0, 115, 0, 121, 0, 100, 0, 121, 0, 0, 0, 115, 0, 121, 0, 104, 0, 105, 0, 0, 0, 115, 0, 121, 0, 108, 0, 97, 0, 0, 0, 115, 0, 121, 0, 114, 0, 97, 0, 0, 0, 115, 0, 121, 0, 114, 0, 100, 0, 0, 0, 115, 0, 121, 0, 115, 0, 117, 0, 0, 0, 115, 0, 121, 0, 116, 0, 97, 0, 0, 0, 115, 0, 122, 0, 104, 0, 104, 0, 0, 0, 115, 0, 122, 0, 109, 0, 97, 0, 0, 0, 116, 0, 100, 0, 98, 0, 97, 0, 0, 0, 116, 0, 100, 0, 98, 0, 103, 0, 0, 0, 116, 0, 100, 0, 98, 0, 111, 0, 0, 0, 116, 0, 100, 0, 99, 0, 98, 0, 0, 0, 116, 0, 100, 0, 101, 0, 101, 0, 0, 0, 116, 0, 100, 0, 101, 0, 111, 0, 0, 0, 116, 0, 100, 0, 103, 0, 114, 0, 0, 0, 116, 0, 100, 0, 104, 0, 108, 0, 0, 0, 116, 0, 100, 0, 108, 0, 99, 0, 0, 0, 116, 0, 100, 0, 108, 0, 111, 0, 0, 0, 116, 0, 100, 0, 108, 0, 114, 0, 0, 0, 116, 0, 100, 0, 109, 0, 97, 0, 0, 0, 116, 0, 100, 0, 109, 0, 99, 0, 0, 0, 116, 0, 100, 0, 109, 0, 101, 0, 0, 0, 116, 0, 100, 0, 109, 0, 111, 0, 0, 0, 116, 0, 100, 0, 110, 0, 100, 0, 0, 0, 116, 0, 100, 0, 111, 0, 100, 0, 0, 0, 116, 0, 100, 0, 115, 0, 97, 0, 0, 0, 116, 0, 100, 0, 115, 0, 105, 0, 0, 0, 116, 0, 100, 0, 116, 0, 97, 0, 0, 0, 116, 0, 100, 0, 116, 0, 105, 0, 0, 0, 116, 0, 100, 0, 119, 0, 102, 0, 0, 0, 116, 0, 106, 0, 100, 0, 117, 0, 0, 0, 116, 0, 106, 0, 107, 0, 116, 0, 0, 0, 116, 0, 106, 0, 114, 0, 97, 0, 0, 0, 116, 0, 106, 0, 115, 0, 117, 0, 0, 0, 116, 0, 108, 0, 97, 0, 108, 0, 0, 0, 116, 0, 108, 0, 97, 0, 110, 0, 0, 0, 116, 0, 108, 0, 98, 0, 111, 0, 0, 0, 116, 0, 108, 0, 99, 0, 111, 0, 0, 0, 116, 0, 108, 0, 100, 0, 105, 0, 0, 0, 116, 0, 108, 0, 101, 0, 114, 0, 0, 0, 116, 0, 108, 0, 108, 0, 97, 0, 0, 0, 116, 0, 108, 0, 108, 0, 105, 0, 0, 0, 116, 0, 108, 0, 109, 0, 102, 0, 0, 0, 116, 0, 108, 0, 109, 0, 116, 0, 0, 0, 116, 0, 108, 0, 111, 0, 101, 0, 0, 0, 116, 0, 110, 0, 54, 0, 49, 0, 0, 0, 117, 0, 97, 0, 48, 0, 53, 0, 0, 0, 117, 0, 97, 0, 48, 0, 55, 0, 0, 0, 117, 0, 97, 0, 48, 0, 57, 0, 0, 0, 117, 0, 97, 0, 49, 0, 50, 0, 0, 0, 117, 0, 97, 0, 49, 0, 52, 0, 0, 0, 117, 0, 97, 0, 49, 0, 56, 0, 0, 0, 117, 0, 97, 0, 50, 0, 49, 0, 0, 0, 117, 0, 97, 0, 50, 0, 51, 0, 0, 0, 117, 0, 97, 0, 50, 0, 54, 0, 0, 0, 117, 0, 97, 0, 51, 0, 48, 0, 0, 0, 117, 0, 97, 0, 51, 0, 50, 0, 0, 0, 117, 0, 97, 0, 51, 0, 53, 0, 0, 0, 117, 0, 97, 0, 52, 0, 48, 0, 0, 0, 117, 0, 97, 0, 52, 0, 51, 0, 0, 0, 117, 0, 97, 0, 52, 0, 54, 0, 0, 0, 117, 0, 97, 0, 52, 0, 56, 0, 0, 0, 117, 0, 97, 0, 53, 0, 49, 0, 0, 0, 117, 0, 97, 0, 53, 0, 51, 0, 0, 0, 117, 0, 97, 0, 53, 0, 54, 0, 0, 0, 117, 0, 97, 0, 53, 0, 57, 0, 0, 0, 117, 0, 97, 0, 54, 0, 49, 0, 0, 0, 117, 0, 97, 0, 54, 0, 51, 0, 0, 0, 117, 0, 97, 0, 54, 0, 53, 0, 0, 0, 117, 0, 97, 0, 54, 0, 56, 0, 0, 0, 117, 0, 97, 0, 55, 0, 49, 0, 0, 0, 117, 0, 97, 0, 55, 0, 52, 0, 0, 0, 117, 0, 97, 0, 55, 0, 55, 0, 0, 0, 117, 0, 109, 0, 54, 0, 55, 0, 0, 0, 117, 0, 109, 0, 55, 0, 49, 0, 0, 0, 117, 0, 109, 0, 55, 0, 54, 0, 0, 0, 117, 0, 109, 0, 55, 0, 57, 0, 0, 0, 117, 0, 109, 0, 56, 0, 49, 0, 0, 0, 117, 0, 109, 0, 56, 0, 52, 0, 0, 0, 117, 0, 109, 0, 56, 0, 54, 0, 0, 0, 117, 0, 109, 0, 56, 0, 57, 0, 0, 0, 117, 0, 109, 0, 57, 0, 53, 0, 0, 0, 117, 0, 115, 0, 97, 0, 115, 0, 0, 0, 117, 0, 115, 0, 97, 0, 122, 0, 0, 0, 117, 0, 115, 0, 99, 0, 111, 0, 0, 0, 117, 0, 115, 0, 99, 0, 116, 0, 0, 0, 117, 0, 115, 0, 100, 0, 101, 0, 0, 0, 117, 0, 115, 0, 102, 0, 108, 0, 0, 0, 117, 0, 115, 0, 103, 0, 97, 0, 0, 0, 117, 0, 115, 0, 103, 0, 117, 0, 0, 0, 117, 0, 115, 0, 104, 0, 105, 0, 0, 0, 117, 0, 115, 0, 105, 0, 97, 0, 0, 0, 117, 0, 115, 0, 105, 0, 100, 0, 0, 0, 117, 0, 115, 0, 105, 0, 108, 0, 0, 0, 117, 0, 115, 0, 105, 0, 110, 0, 0, 0, 117, 0, 115, 0, 107, 0, 115, 0, 0, 0, 117, 0, 115, 0, 107, 0, 121, 0, 0, 0, 117, 0, 115, 0, 109, 0, 97, 0, 0, 0, 117, 0, 115, 0, 109, 0, 105, 0, 0, 0, 117, 0, 115, 0, 109, 0, 112, 0, 0, 0, 117, 0, 115, 0, 110, 0, 106, 0, 0, 0, 117, 0, 115, 0, 110, 0, 109, 0, 0, 0, 117, 0, 115, 0, 110, 0, 118, 0, 0, 0, 117, 0, 115, 0, 110, 0, 121, 0, 0, 0, 117, 0, 115, 0, 111, 0, 104, 0, 0, 0, 117, 0, 115, 0, 111, 0, 107, 0, 0, 0, 117, 0, 115, 0, 111, 0, 114, 0, 0, 0, 117, 0, 115, 0, 112, 0, 97, 0, 0, 0, 117, 0, 115, 0, 112, 0, 114, 0, 0, 0, 117, 0, 115, 0, 114, 0, 105, 0, 0, 0, 117, 0, 115, 0, 116, 0, 110, 0, 0, 0, 117, 0, 115, 0, 116, 0, 120, 0, 0, 0, 117, 0, 115, 0, 117, 0, 109, 0, 0, 0, 117, 0, 115, 0, 117, 0, 116, 0, 0, 0, 117, 0, 115, 0, 118, 0, 97, 0, 0, 0, 117, 0, 115, 0, 118, 0, 105, 0, 0, 0, 117, 0, 115, 0, 118, 0, 116, 0, 0, 0, 117, 0, 115, 0, 119, 0, 97, 0, 0, 0, 117, 0, 115, 0, 119, 0, 105, 0, 0, 0, 117, 0, 115, 0, 119, 0, 118, 0, 0, 0, 117, 0, 115, 0, 119, 0, 121, 0, 0, 0, 117, 0, 121, 0, 99, 0, 97, 0, 0, 0, 117, 0, 121, 0, 100, 0, 117, 0, 0, 0, 117, 0, 121, 0, 102, 0, 100, 0, 0, 0, 117, 0, 121, 0, 102, 0, 115, 0, 0, 0, 117, 0, 121, 0, 108, 0, 97, 0, 0, 0, 117, 0, 121, 0, 109, 0, 111, 0, 0, 0, 117, 0, 121, 0, 112, 0, 97, 0, 0, 0, 117, 0, 121, 0, 114, 0, 118, 0, 0, 0, 117, 0, 121, 0, 115, 0, 97, 0, 0, 0, 117, 0, 121, 0, 115, 0, 106, 0, 0, 0, 117, 0, 122, 0, 97, 0, 110, 0, 0, 0, 117, 0, 122, 0, 98, 0, 117, 0, 0, 0, 117, 0, 122, 0, 110, 0, 119, 0, 0, 0, 117, 0, 122, 0, 113, 0, 97, 0, 0, 0, 117, 0, 122, 0, 113, 0, 114, 0, 0, 0, 117, 0, 122, 0, 120, 0, 111, 0, 0, 0, 118, 0, 110, 0, 48, 0, 57, 0, 0, 0, 118, 0, 110, 0, 49, 0, 56, 0, 0, 0, 118, 0, 110, 0, 51, 0, 57, 0, 0, 0, 118, 0, 110, 0, 52, 0, 57, 0, 0, 0, 118, 0, 110, 0, 54, 0, 49, 0, 0, 0, 118, 0, 110, 0, 54, 0, 51, 0, 0, 0, 118, 0, 110, 0, 99, 0, 116, 0, 0, 0, 118, 0, 110, 0, 100, 0, 110, 0, 0, 0, 118, 0, 110, 0, 104, 0, 110, 0, 0, 0, 118, 0, 110, 0, 104, 0, 112, 0, 0, 0, 118, 0, 110, 0, 115, 0, 103, 0, 0, 0, 119, 0, 102, 0, 97, 0, 108, 0, 0, 0, 119, 0, 102, 0, 115, 0, 103, 0, 0, 0, 119, 0, 115, 0, 97, 0, 97, 0, 0, 0, 119, 0, 115, 0, 97, 0, 108, 0, 0, 0, 119, 0, 115, 0, 97, 0, 116, 0, 0, 0, 119, 0, 115, 0, 102, 0, 97, 0, 0, 0, 119, 0, 115, 0, 103, 0, 101, 0, 0, 0, 119, 0, 115, 0, 103, 0, 105, 0, 0, 0, 119, 0, 115, 0, 112, 0, 97, 0, 0, 0, 119, 0, 115, 0, 115, 0, 97, 0, 0, 0, 119, 0, 115, 0, 116, 0, 117, 0, 0, 0, 119, 0, 115, 0, 118, 0, 102, 0, 0, 0, 121, 0, 101, 0, 97, 0, 98, 0, 0, 0, 121, 0, 101, 0, 97, 0, 100, 0, 0, 0, 121, 0, 101, 0, 97, 0, 109, 0, 0, 0, 121, 0, 101, 0, 98, 0, 97, 0, 0, 0, 121, 0, 101, 0, 100, 0, 97, 0, 0, 0, 121, 0, 101, 0, 100, 0, 104, 0, 0, 0, 121, 0, 101, 0, 104, 0, 100, 0, 0, 0, 121, 0, 101, 0, 104, 0, 106, 0, 0, 0, 121, 0, 101, 0, 105, 0, 98, 0, 0, 0, 121, 0, 101, 0, 108, 0, 97, 0, 0, 0, 121, 0, 101, 0, 109, 0, 97, 0, 0, 0, 121, 0, 101, 0, 109, 0, 114, 0, 0, 0, 121, 0, 101, 0, 109, 0, 119, 0, 0, 0, 121, 0, 101, 0, 115, 0, 100, 0, 0, 0, 121, 0, 101, 0, 115, 0, 110, 0, 0, 0, 121, 0, 101, 0, 116, 0, 97, 0, 0, 0, 122, 0, 97, 0, 102, 0, 115, 0, 0, 0, 122, 0, 97, 0, 103, 0, 116, 0, 0, 0, 122, 0, 97, 0, 108, 0, 112, 0, 0, 0, 122, 0, 97, 0, 109, 0, 112, 0, 0, 0, 122, 0, 97, 0, 110, 0, 99, 0, 0, 0, 122, 0, 97, 0, 110, 0, 108, 0, 0, 0, 122, 0, 97, 0, 110, 0, 119, 0, 0, 0, 122, 0, 97, 0, 119, 0, 99, 0, 0, 0, 122, 0, 109, 0, 49, 0, 48, 0, 0, 0, 122, 0, 119, 0, 98, 0, 117, 0, 0, 0, 122, 0, 119, 0, 109, 0, 97, 0, 0, 0, 122, 0, 119, 0, 109, 0, 99, 0, 0, 0, 122, 0, 119, 0, 109, 0, 101, 0, 0, 0, 122, 0, 119, 0, 109, 0, 105, 0, 0, 0, 49, 0, 57, 0, 48, 0, 49, 0, 0, 0, 49, 0, 57, 0, 57, 0, 54, 0, 0, 0, 66, 0, 104, 0, 107, 0, 115, 0, 0, 0, 72, 0, 105, 0, 114, 0, 97, 0, 0, 0, 72, 0, 109, 0, 110, 0, 112, 0, 0, 0, 72, 0, 114, 0, 107, 0, 116, 0, 0, 0, 72, 0, 117, 0, 110, 0, 103, 0, 0, 0, 77, 0, 97, 0, 114, 0, 99, 0, 0, 0, 77, 0, 101, 0, 114, 0, 111, 0, 0, 0, 83, 0, 111, 0, 103, 0, 111, 0, 0, 0, 97, 0, 108, 0, 109, 0, 109, 0, 0, 0, 97, 0, 108, 0, 115, 0, 104, 0, 0, 0, 97, 0, 108, 0, 115, 0, 107, 0, 0, 0, 98, 0, 106, 0, 97, 0, 107, 0, 0, 0, 98, 0, 114, 0, 114, 0, 111, 0, 0, 0, 99, 0, 104, 0, 115, 0, 104, 0, 0, 0, 99, 0, 110, 0, 104, 0, 107, 0, 0, 0, 99, 0, 118, 0, 115, 0, 108, 0, 0, 0, 100, 0, 106, 0, 97, 0, 114, 0, 0, 0, 101, 0, 115, 0, 99, 0, 98, 0, 0, 0, 103, 0, 101, 0, 115, 0, 107, 0, 0, 0, 103, 0, 110, 0, 98, 0, 102, 0, 0, 0, 104, 0, 117, 0, 115, 0, 115, 0, 0, 0, 105, 0, 100, 0, 115, 0, 98, 0, 0, 0, 105, 0, 100, 0, 115, 0, 108, 0, 0, 0, 105, 0, 101, 0, 99, 0, 110, 0, 0, 0, 105, 0, 101, 0, 99, 0, 111, 0, 0, 0, 105, 0, 116, 0, 109, 0, 98, 0, 0, 0, 105, 0, 116, 0, 112, 0, 99, 0, 0, 0, 105, 0, 116, 0, 118, 0, 97, 0, 0, 0, 109, 0, 100, 0, 99, 0, 114, 0, 0, 0, 109, 0, 119, 0, 107, 0, 114, 0, 0, 0, 111, 0, 109, 0, 115, 0, 104, 0, 0, 0, 112, 0, 108, 0, 108, 0, 117, 0, 0, 0, 115, 0, 121, 0, 104, 0, 108, 0, 0, 0, 77, 0, 117, 0, 108, 0, 116, 0, 0, 0, 99, 0, 122, 0, 108, 0, 105, 0, 0, 0, 99, 0, 122, 0, 109, 0, 111, 0, 0, 0, 99, 0, 122, 0, 112, 0, 108, 0, 0, 0, 100, 0, 122, 0, 51, 0, 50, 0, 0, 0, 100, 0, 122, 0, 51, 0, 54, 0, 0, 0, 100, 0, 122, 0, 52, 0, 52, 0, 0, 0, 100, 0, 122, 0, 52, 0, 56, 0, 0, 0, 102, 0, 114, 0, 56, 0, 52, 0, 0, 0, 102, 0, 114, 0, 56, 0, 53, 0, 0, 0, 102, 0, 114, 0, 57, 0, 52, 0, 0, 0, 103, 0, 119, 0, 98, 0, 109, 0, 0, 0, 105, 0, 116, 0, 98, 0, 111, 0, 0, 0, 109, 0, 116, 0, 54, 0, 56, 0, 0, 0, 109, 0, 119, 0, 107, 0, 115, 0, 0, 0, 109, 0, 119, 0, 109, 0, 104, 0, 0, 0, 116, 0, 114, 0, 56, 0, 49, 0, 0, 0, 116, 0, 122, 0, 49, 0, 50, 0, 0, 0, 116, 0, 122, 0, 49, 0, 53, 0, 0, 0, 116, 0, 122, 0, 49, 0, 57, 0, 0, 0, 116, 0, 122, 0, 50, 0, 48, 0, 0, 0, 117, 0, 121, 0, 114, 0, 111, 0, 0, 0, 68, 0, 74, 0, 126, 0, 75, 0, 0, 0, 51, 0, 54, 0, 46, 0, 49, 0, 0, 0, 65, 0, 67, 0, 126, 0, 71, 0, 0, 0, 65, 0, 76, 0, 126, 0, 77, 0, 0, 0, 65, 0, 81, 0, 126, 0, 85, 0, 0, 0, 65, 0, 87, 0, 126, 0, 88, 0, 0, 0, 65, 0, 104, 0, 111, 0, 109, 0, 0, 0, 66, 0, 65, 0, 126, 0, 66, 0, 0, 0, 66, 0, 68, 0, 126, 0, 74, 0, 0, 0, 66, 0, 76, 0, 126, 0, 79, 0, 0, 0, 66, 0, 81, 0, 126, 0, 84, 0, 0, 0, 66, 0, 86, 0, 126, 0, 87, 0, 0, 0, 66, 0, 89, 0, 126, 0, 90, 0, 0, 0, 66, 0, 97, 0, 115, 0, 115, 0, 0, 0, 67, 0, 67, 0, 126, 0, 68, 0, 0, 0, 67, 0, 70, 0, 126, 0, 73, 0, 0, 0, 67, 0, 75, 0, 126, 0, 80, 0, 0, 0, 67, 0, 85, 0, 126, 0, 90, 0, 0, 0, 67, 0, 104, 0, 114, 0, 115, 0, 0, 0, 68, 0, 105, 0, 97, 0, 107, 0, 0, 0, 68, 0, 111, 0, 103, 0, 114, 0, 0, 0, 69, 0, 82, 0, 126, 0, 84, 0, 0, 0, 69, 0, 108, 0, 121, 0, 109, 0, 0, 0, 70, 0, 73, 0, 126, 0, 75, 0, 0, 0, 71, 0, 65, 0, 126, 0, 66, 0, 0, 0, 71, 0, 68, 0, 126, 0, 73, 0, 0, 0, 71, 0, 76, 0, 126, 0, 78, 0, 0, 0, 71, 0, 80, 0, 126, 0, 85, 0, 0, 0, 71, 0, 111, 0, 110, 0, 103, 0, 0, 0, 71, 0, 111, 0, 110, 0, 109, 0, 0, 0, 72, 0, 77, 0, 126, 0, 78, 0, 0, 0, 72, 0, 84, 0, 126, 0, 85, 0, 0, 0, 72, 0, 97, 0, 110, 0, 98, 0, 0, 0, 72, 0, 97, 0, 110, 0, 103, 0, 0, 0, 72, 0, 97, 0, 116, 0, 114, 0, 0, 0, 72, 0, 108, 0, 117, 0, 119, 0, 0, 0, 73, 0, 67, 0, 126, 0, 69, 0, 0, 0, 73, 0, 76, 0, 126, 0, 79, 0, 0, 0, 73, 0, 81, 0, 126, 0, 84, 0, 0, 0, 74, 0, 79, 0, 126, 0, 80, 0, 0, 0, 74, 0, 97, 0, 109, 0, 111, 0, 0, 0, 75, 0, 71, 0, 126, 0, 73, 0, 0, 0, 75, 0, 77, 0, 126, 0, 78, 0, 0, 0, 75, 0, 89, 0, 126, 0, 90, 0, 0, 0, 75, 0, 104, 0, 97, 0, 114, 0, 0, 0, 75, 0, 105, 0, 116, 0, 115, 0, 0, 0, 75, 0, 116, 0, 104, 0, 105, 0, 0, 0, 76, 0, 65, 0, 126, 0, 67, 0, 0, 0, 76, 0, 82, 0, 126, 0, 86, 0, 0, 0, 76, 0, 105, 0, 110, 0, 97, 0, 0, 0, 76, 0, 105, 0, 110, 0, 98, 0, 0, 0, 77, 0, 67, 0, 126, 0, 72, 0, 0, 0, 77, 0, 75, 0, 126, 0, 90, 0, 0, 0, 77, 0, 97, 0, 107, 0, 97, 0, 0, 0, 77, 0, 101, 0, 100, 0, 102, 0, 0, 0, 78, 0, 69, 0, 126, 0, 71, 0, 0, 0, 78, 0, 79, 0, 126, 0, 80, 0, 0, 0, 78, 0, 97, 0, 110, 0, 100, 0, 0, 0, 78, 0, 101, 0, 119, 0, 97, 0, 0, 0, 78, 0, 115, 0, 104, 0, 117, 0, 0, 0, 80, 0, 69, 0, 126, 0, 72, 0, 0, 0, 80, 0, 75, 0, 126, 0, 78, 0, 0, 0, 80, 0, 82, 0, 126, 0, 84, 0, 0, 0, 80, 0, 97, 0, 117, 0, 99, 0, 0, 0, 81, 0, 77, 0, 126, 0, 78, 0, 0, 0, 81, 0, 80, 0, 126, 0, 84, 0, 0, 0, 81, 0, 86, 0, 126, 0, 90, 0, 0, 0, 81, 0, 97, 0, 97, 0, 103, 0, 0, 0, 81, 0, 97, 0, 97, 0, 104, 0, 0, 0, 81, 0, 97, 0, 97, 0, 105, 0, 0, 0, 82, 0, 111, 0, 104, 0, 103, 0, 0, 0, 83, 0, 65, 0, 126, 0, 69, 0, 0, 0, 83, 0, 71, 0, 126, 0, 79, 0, 0, 0, 83, 0, 82, 0, 126, 0, 84, 0, 0, 0, 83, 0, 88, 0, 126, 0, 90, 0, 0, 0, 83, 0, 111, 0, 103, 0, 100, 0, 0, 0, 84, 0, 67, 0, 126, 0, 68, 0, 0, 0, 84, 0, 70, 0, 126, 0, 72, 0, 0, 0, 84, 0, 74, 0, 126, 0, 79, 0, 0, 0, 84, 0, 86, 0, 126, 0, 87, 0, 0, 0, 84, 0, 97, 0, 110, 0, 103, 0, 0, 0, 85, 0, 89, 0, 126, 0, 90, 0, 0, 0, 87, 0, 99, 0, 104, 0, 111, 0, 0, 0, 88, 0, 65, 0, 126, 0, 66, 0, 0, 0, 88, 0, 67, 0, 126, 0, 74, 0, 0, 0, 88, 0, 76, 0, 126, 0, 90, 0, 0, 0, 89, 0, 101, 0, 122, 0, 105, 0, 0, 0, 90, 0, 97, 0, 110, 0, 98, 0, 0, 0, 90, 0, 105, 0, 110, 0, 104, 0, 0, 0, 90, 0, 109, 0, 116, 0, 104, 0, 0, 0, 90, 0, 115, 0, 121, 0, 101, 0, 0, 0, 90, 0, 115, 0, 121, 0, 109, 0, 0, 0, 90, 0, 120, 0, 120, 0, 120, 0, 0, 0, 90, 0, 121, 0, 121, 0, 121, 0, 0, 0, 90, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 100, 0, 48, 0, 50, 0, 0, 0, 97, 0, 100, 0, 48, 0, 51, 0, 0, 0, 97, 0, 100, 0, 48, 0, 52, 0, 0, 0, 97, 0, 100, 0, 48, 0, 53, 0, 0, 0, 97, 0, 100, 0, 48, 0, 54, 0, 0, 0, 97, 0, 100, 0, 48, 0, 55, 0, 0, 0, 97, 0, 100, 0, 48, 0, 56, 0, 0, 0, 97, 0, 103, 0, 48, 0, 51, 0, 0, 0, 97, 0, 103, 0, 48, 0, 52, 0, 0, 0, 97, 0, 103, 0, 48, 0, 53, 0, 0, 0, 97, 0, 103, 0, 48, 0, 54, 0, 0, 0, 97, 0, 103, 0, 48, 0, 55, 0, 0, 0, 97, 0, 103, 0, 48, 0, 56, 0, 0, 0, 97, 0, 103, 0, 49, 0, 48, 0, 0, 0, 97, 0, 103, 0, 49, 0, 49, 0, 0, 0, 97, 0, 108, 0, 48, 0, 49, 0, 0, 0, 97, 0, 108, 0, 48, 0, 50, 0, 0, 0, 97, 0, 108, 0, 48, 0, 51, 0, 0, 0, 97, 0, 108, 0, 48, 0, 52, 0, 0, 0, 97, 0, 108, 0, 48, 0, 53, 0, 0, 0, 97, 0, 108, 0, 48, 0, 54, 0, 0, 0, 97, 0, 108, 0, 48, 0, 55, 0, 0, 0, 97, 0, 108, 0, 48, 0, 56, 0, 0, 0, 97, 0, 108, 0, 48, 0, 57, 0, 0, 0, 97, 0, 108, 0, 49, 0, 48, 0, 0, 0, 97, 0, 108, 0, 49, 0, 49, 0, 0, 0, 97, 0, 108, 0, 49, 0, 50, 0, 0, 0, 97, 0, 108, 0, 98, 0, 114, 0, 0, 0, 97, 0, 108, 0, 98, 0, 117, 0, 0, 0, 97, 0, 108, 0, 100, 0, 105, 0, 0, 0, 97, 0, 108, 0, 100, 0, 108, 0, 0, 0, 97, 0, 108, 0, 100, 0, 114, 0, 0, 0, 97, 0, 108, 0, 100, 0, 118, 0, 0, 0, 97, 0, 108, 0, 101, 0, 108, 0, 0, 0, 97, 0, 108, 0, 101, 0, 114, 0, 0, 0, 97, 0, 108, 0, 102, 0, 114, 0, 0, 0, 97, 0, 108, 0, 103, 0, 106, 0, 0, 0, 97, 0, 108, 0, 103, 0, 114, 0, 0, 0, 97, 0, 108, 0, 104, 0, 97, 0, 0, 0, 97, 0, 108, 0, 107, 0, 111, 0, 0, 0, 97, 0, 108, 0, 107, 0, 114, 0, 0, 0, 97, 0, 108, 0, 107, 0, 117, 0, 0, 0, 97, 0, 108, 0, 108, 0, 98, 0, 0, 0, 97, 0, 108, 0, 108, 0, 101, 0, 0, 0, 97, 0, 108, 0, 108, 0, 117, 0, 0, 0, 97, 0, 108, 0, 109, 0, 107, 0, 0, 0, 97, 0, 108, 0, 109, 0, 114, 0, 0, 0, 97, 0, 108, 0, 109, 0, 116, 0, 0, 0, 97, 0, 108, 0, 112, 0, 103, 0, 0, 0, 97, 0, 108, 0, 112, 0, 117, 0, 0, 0, 97, 0, 108, 0, 115, 0, 114, 0, 0, 0, 97, 0, 108, 0, 116, 0, 101, 0, 0, 0, 97, 0, 108, 0, 116, 0, 112, 0, 0, 0, 97, 0, 108, 0, 116, 0, 114, 0, 0, 0, 97, 0, 108, 0, 118, 0, 108, 0, 0, 0, 98, 0, 97, 0, 49, 0, 48, 0, 0, 0, 98, 0, 98, 0, 48, 0, 49, 0, 0, 0, 98, 0, 98, 0, 48, 0, 50, 0, 0, 0, 98, 0, 98, 0, 48, 0, 51, 0, 0, 0, 98, 0, 98, 0, 48, 0, 52, 0, 0, 0, 98, 0, 98, 0, 48, 0, 53, 0, 0, 0, 98, 0, 98, 0, 48, 0, 54, 0, 0, 0, 98, 0, 98, 0, 48, 0, 55, 0, 0, 0, 98, 0, 98, 0, 48, 0, 56, 0, 0, 0, 98, 0, 98, 0, 48, 0, 57, 0, 0, 0, 98, 0, 98, 0, 49, 0, 48, 0, 0, 0, 98, 0, 98, 0, 49, 0, 49, 0, 0, 0, 98, 0, 100, 0, 48, 0, 49, 0, 0, 0, 98, 0, 100, 0, 48, 0, 50, 0, 0, 0, 98, 0, 100, 0, 48, 0, 51, 0, 0, 0, 98, 0, 100, 0, 48, 0, 52, 0, 0, 0, 98, 0, 100, 0, 48, 0, 53, 0, 0, 0, 98, 0, 100, 0, 48, 0, 54, 0, 0, 0, 98, 0, 100, 0, 48, 0, 55, 0, 0, 0, 98, 0, 100, 0, 48, 0, 56, 0, 0, 0, 98, 0, 100, 0, 48, 0, 57, 0, 0, 0, 98, 0, 100, 0, 49, 0, 48, 0, 0, 0, 98, 0, 100, 0, 49, 0, 49, 0, 0, 0, 98, 0, 100, 0, 49, 0, 50, 0, 0, 0, 98, 0, 100, 0, 49, 0, 51, 0, 0, 0, 98, 0, 100, 0, 49, 0, 52, 0, 0, 0, 98, 0, 100, 0, 49, 0, 53, 0, 0, 0, 98, 0, 100, 0, 49, 0, 54, 0, 0, 0, 98, 0, 100, 0, 49, 0, 55, 0, 0, 0, 98, 0, 100, 0, 49, 0, 56, 0, 0, 0, 98, 0, 100, 0, 49, 0, 57, 0, 0, 0, 98, 0, 100, 0, 50, 0, 48, 0, 0, 0, 98, 0, 100, 0, 50, 0, 49, 0, 0, 0, 98, 0, 100, 0, 50, 0, 50, 0, 0, 0, 98, 0, 100, 0, 50, 0, 51, 0, 0, 0, 98, 0, 100, 0, 50, 0, 52, 0, 0, 0, 98, 0, 100, 0, 50, 0, 53, 0, 0, 0, 98, 0, 100, 0, 50, 0, 54, 0, 0, 0, 98, 0, 100, 0, 50, 0, 55, 0, 0, 0, 98, 0, 100, 0, 50, 0, 56, 0, 0, 0, 98, 0, 100, 0, 50, 0, 57, 0, 0, 0, 98, 0, 100, 0, 51, 0, 48, 0, 0, 0, 98, 0, 100, 0, 51, 0, 49, 0, 0, 0, 98, 0, 100, 0, 51, 0, 50, 0, 0, 0, 98, 0, 100, 0, 51, 0, 51, 0, 0, 0, 98, 0, 100, 0, 51, 0, 52, 0, 0, 0, 98, 0, 100, 0, 51, 0, 53, 0, 0, 0, 98, 0, 100, 0, 51, 0, 54, 0, 0, 0, 98, 0, 100, 0, 51, 0, 55, 0, 0, 0, 98, 0, 100, 0, 51, 0, 56, 0, 0, 0, 98, 0, 100, 0, 51, 0, 57, 0, 0, 0, 98, 0, 100, 0, 52, 0, 48, 0, 0, 0, 98, 0, 100, 0, 52, 0, 49, 0, 0, 0, 98, 0, 100, 0, 52, 0, 50, 0, 0, 0, 98, 0, 100, 0, 52, 0, 51, 0, 0, 0, 98, 0, 100, 0, 52, 0, 52, 0, 0, 0, 98, 0, 100, 0, 52, 0, 53, 0, 0, 0, 98, 0, 100, 0, 52, 0, 54, 0, 0, 0, 98, 0, 100, 0, 52, 0, 55, 0, 0, 0, 98, 0, 100, 0, 52, 0, 56, 0, 0, 0, 98, 0, 100, 0, 52, 0, 57, 0, 0, 0, 98, 0, 100, 0, 53, 0, 48, 0, 0, 0, 98, 0, 100, 0, 53, 0, 49, 0, 0, 0, 98, 0, 100, 0, 53, 0, 50, 0, 0, 0, 98, 0, 100, 0, 53, 0, 51, 0, 0, 0, 98, 0, 100, 0, 53, 0, 52, 0, 0, 0, 98, 0, 100, 0, 53, 0, 53, 0, 0, 0, 98, 0, 100, 0, 53, 0, 54, 0, 0, 0, 98, 0, 100, 0, 53, 0, 55, 0, 0, 0, 98, 0, 100, 0, 53, 0, 56, 0, 0, 0, 98, 0, 100, 0, 53, 0, 57, 0, 0, 0, 98, 0, 100, 0, 54, 0, 48, 0, 0, 0, 98, 0, 100, 0, 54, 0, 49, 0, 0, 0, 98, 0, 100, 0, 54, 0, 50, 0, 0, 0, 98, 0, 100, 0, 54, 0, 51, 0, 0, 0, 98, 0, 100, 0, 54, 0, 52, 0, 0, 0, 98, 0, 102, 0, 48, 0, 49, 0, 0, 0, 98, 0, 102, 0, 48, 0, 50, 0, 0, 0, 98, 0, 102, 0, 48, 0, 51, 0, 0, 0, 98, 0, 102, 0, 48, 0, 52, 0, 0, 0, 98, 0, 102, 0, 48, 0, 53, 0, 0, 0, 98, 0, 102, 0, 48, 0, 54, 0, 0, 0, 98, 0, 102, 0, 48, 0, 55, 0, 0, 0, 98, 0, 102, 0, 48, 0, 56, 0, 0, 0, 98, 0, 102, 0, 48, 0, 57, 0, 0, 0, 98, 0, 102, 0, 49, 0, 48, 0, 0, 0, 98, 0, 102, 0, 49, 0, 49, 0, 0, 0, 98, 0, 102, 0, 49, 0, 50, 0, 0, 0, 98, 0, 102, 0, 49, 0, 51, 0, 0, 0, 98, 0, 103, 0, 48, 0, 49, 0, 0, 0, 98, 0, 103, 0, 48, 0, 50, 0, 0, 0, 98, 0, 103, 0, 48, 0, 51, 0, 0, 0, 98, 0, 103, 0, 48, 0, 52, 0, 0, 0, 98, 0, 103, 0, 48, 0, 53, 0, 0, 0, 98, 0, 103, 0, 48, 0, 54, 0, 0, 0, 98, 0, 103, 0, 48, 0, 55, 0, 0, 0, 98, 0, 103, 0, 48, 0, 56, 0, 0, 0, 98, 0, 103, 0, 48, 0, 57, 0, 0, 0, 98, 0, 103, 0, 49, 0, 48, 0, 0, 0, 98, 0, 103, 0, 49, 0, 49, 0, 0, 0, 98, 0, 103, 0, 49, 0, 50, 0, 0, 0, 98, 0, 103, 0, 49, 0, 51, 0, 0, 0, 98, 0, 103, 0, 49, 0, 52, 0, 0, 0, 98, 0, 103, 0, 49, 0, 53, 0, 0, 0, 98, 0, 103, 0, 49, 0, 54, 0, 0, 0, 98, 0, 103, 0, 49, 0, 55, 0, 0, 0, 98, 0, 103, 0, 49, 0, 56, 0, 0, 0, 98, 0, 103, 0, 49, 0, 57, 0, 0, 0, 98, 0, 103, 0, 50, 0, 48, 0, 0, 0, 98, 0, 103, 0, 50, 0, 49, 0, 0, 0, 98, 0, 103, 0, 50, 0, 50, 0, 0, 0, 98, 0, 103, 0, 50, 0, 51, 0, 0, 0, 98, 0, 103, 0, 50, 0, 52, 0, 0, 0, 98, 0, 103, 0, 50, 0, 53, 0, 0, 0, 98, 0, 103, 0, 50, 0, 54, 0, 0, 0, 98, 0, 103, 0, 50, 0, 55, 0, 0, 0, 98, 0, 103, 0, 50, 0, 56, 0, 0, 0, 98, 0, 104, 0, 49, 0, 51, 0, 0, 0, 98, 0, 104, 0, 49, 0, 52, 0, 0, 0, 98, 0, 104, 0, 49, 0, 53, 0, 0, 0, 98, 0, 104, 0, 49, 0, 54, 0, 0, 0, 98, 0, 105, 0, 98, 0, 108, 0, 0, 0, 98, 0, 105, 0, 98, 0, 109, 0, 0, 0, 98, 0, 106, 0, 97, 0, 108, 0, 0, 0, 98, 0, 114, 0, 97, 0, 108, 0, 0, 0, 98, 0, 114, 0, 97, 0, 109, 0, 0, 0, 98, 0, 114, 0, 109, 0, 115, 0, 0, 0, 98, 0, 114, 0, 109, 0, 116, 0, 0, 0, 98, 0, 114, 0, 112, 0, 97, 0, 0, 0, 98, 0, 114, 0, 112, 0, 98, 0, 0, 0, 98, 0, 114, 0, 114, 0, 110, 0, 0, 0, 98, 0, 114, 0, 114, 0, 114, 0, 0, 0, 98, 0, 114, 0, 114, 0, 115, 0, 0, 0, 98, 0, 115, 0, 110, 0, 111, 0, 0, 0, 98, 0, 115, 0, 110, 0, 112, 0, 0, 0, 98, 0, 116, 0, 49, 0, 49, 0, 0, 0, 98, 0, 116, 0, 49, 0, 50, 0, 0, 0, 98, 0, 116, 0, 49, 0, 51, 0, 0, 0, 98, 0, 116, 0, 49, 0, 52, 0, 0, 0, 98, 0, 116, 0, 49, 0, 53, 0, 0, 0, 98, 0, 116, 0, 50, 0, 49, 0, 0, 0, 98, 0, 116, 0, 50, 0, 50, 0, 0, 0, 98, 0, 116, 0, 50, 0, 51, 0, 0, 0, 98, 0, 116, 0, 50, 0, 52, 0, 0, 0, 98, 0, 116, 0, 51, 0, 49, 0, 0, 0, 98, 0, 116, 0, 51, 0, 50, 0, 0, 0, 98, 0, 116, 0, 51, 0, 51, 0, 0, 0, 98, 0, 116, 0, 51, 0, 52, 0, 0, 0, 98, 0, 116, 0, 52, 0, 49, 0, 0, 0, 98, 0, 116, 0, 52, 0, 50, 0, 0, 0, 98, 0, 116, 0, 52, 0, 51, 0, 0, 0, 98, 0, 116, 0, 52, 0, 52, 0, 0, 0, 98, 0, 116, 0, 52, 0, 53, 0, 0, 0, 98, 0, 119, 0, 115, 0, 111, 0, 0, 0, 98, 0, 119, 0, 115, 0, 112, 0, 0, 0, 99, 0, 97, 0, 110, 0, 115, 0, 0, 0, 99, 0, 97, 0, 110, 0, 116, 0, 0, 0, 99, 0, 97, 0, 110, 0, 117, 0, 0, 0, 99, 0, 100, 0, 98, 0, 110, 0, 0, 0, 99, 0, 100, 0, 104, 0, 107, 0, 0, 0, 99, 0, 100, 0, 104, 0, 108, 0, 0, 0, 99, 0, 100, 0, 107, 0, 97, 0, 0, 0, 99, 0, 100, 0, 107, 0, 119, 0, 0, 0, 99, 0, 100, 0, 109, 0, 110, 0, 0, 0, 99, 0, 100, 0, 109, 0, 111, 0, 0, 0, 99, 0, 100, 0, 111, 0, 114, 0, 0, 0, 99, 0, 103, 0, 49, 0, 49, 0, 0, 0, 99, 0, 103, 0, 49, 0, 50, 0, 0, 0, 99, 0, 103, 0, 49, 0, 51, 0, 0, 0, 99, 0, 103, 0, 49, 0, 52, 0, 0, 0, 99, 0, 103, 0, 49, 0, 53, 0, 0, 0, 99, 0, 103, 0, 49, 0, 54, 0, 0, 0, 99, 0, 104, 0, 115, 0, 103, 0, 0, 0, 99, 0, 104, 0, 122, 0, 103, 0, 0, 0, 99, 0, 104, 0, 122, 0, 104, 0, 0, 0, 99, 0, 110, 0, 55, 0, 49, 0, 0, 0, 99, 0, 110, 0, 104, 0, 97, 0, 0, 0, 99, 0, 110, 0, 104, 0, 98, 0, 0, 0, 99, 0, 110, 0, 104, 0, 108, 0, 0, 0, 99, 0, 110, 0, 115, 0, 99, 0, 0, 0, 99, 0, 110, 0, 115, 0, 100, 0, 0, 0, 99, 0, 117, 0, 48, 0, 51, 0, 0, 0, 99, 0, 117, 0, 48, 0, 52, 0, 0, 0, 99, 0, 117, 0, 48, 0, 53, 0, 0, 0, 99, 0, 117, 0, 48, 0, 54, 0, 0, 0, 99, 0, 117, 0, 48, 0, 55, 0, 0, 0, 99, 0, 117, 0, 48, 0, 56, 0, 0, 0, 99, 0, 117, 0, 48, 0, 57, 0, 0, 0, 99, 0, 117, 0, 49, 0, 48, 0, 0, 0, 99, 0, 117, 0, 49, 0, 49, 0, 0, 0, 99, 0, 117, 0, 49, 0, 50, 0, 0, 0, 99, 0, 117, 0, 49, 0, 51, 0, 0, 0, 99, 0, 117, 0, 49, 0, 52, 0, 0, 0, 99, 0, 117, 0, 49, 0, 53, 0, 0, 0, 99, 0, 117, 0, 49, 0, 54, 0, 0, 0, 99, 0, 118, 0, 115, 0, 109, 0, 0, 0, 99, 0, 121, 0, 48, 0, 49, 0, 0, 0, 99, 0, 121, 0, 48, 0, 50, 0, 0, 0, 99, 0, 121, 0, 48, 0, 51, 0, 0, 0, 99, 0, 121, 0, 48, 0, 52, 0, 0, 0, 99, 0, 121, 0, 48, 0, 53, 0, 0, 0, 99, 0, 121, 0, 48, 0, 54, 0, 0, 0, 99, 0, 122, 0, 106, 0, 99, 0, 0, 0, 99, 0, 122, 0, 106, 0, 109, 0, 0, 0, 99, 0, 122, 0, 111, 0, 108, 0, 0, 0, 99, 0, 122, 0, 112, 0, 114, 0, 0, 0, 99, 0, 122, 0, 117, 0, 115, 0, 0, 0, 99, 0, 122, 0, 118, 0, 121, 0, 0, 0, 99, 0, 122, 0, 122, 0, 108, 0, 0, 0, 100, 0, 106, 0, 97, 0, 115, 0, 0, 0, 100, 0, 106, 0, 100, 0, 105, 0, 0, 0, 100, 0, 106, 0, 100, 0, 106, 0, 0, 0, 100, 0, 107, 0, 56, 0, 49, 0, 0, 0, 100, 0, 107, 0, 56, 0, 50, 0, 0, 0, 100, 0, 107, 0, 56, 0, 51, 0, 0, 0, 100, 0, 107, 0, 56, 0, 52, 0, 0, 0, 100, 0, 107, 0, 56, 0, 53, 0, 0, 0, 100, 0, 109, 0, 48, 0, 50, 0, 0, 0, 100, 0, 109, 0, 48, 0, 51, 0, 0, 0, 100, 0, 109, 0, 48, 0, 52, 0, 0, 0, 100, 0, 109, 0, 48, 0, 53, 0, 0, 0, 100, 0, 109, 0, 48, 0, 54, 0, 0, 0, 100, 0, 109, 0, 48, 0, 55, 0, 0, 0, 100, 0, 109, 0, 48, 0, 56, 0, 0, 0, 100, 0, 109, 0, 48, 0, 57, 0, 0, 0, 100, 0, 109, 0, 49, 0, 48, 0, 0, 0, 100, 0, 109, 0, 49, 0, 49, 0, 0, 0, 100, 0, 111, 0, 48, 0, 49, 0, 0, 0, 100, 0, 111, 0, 48, 0, 50, 0, 0, 0, 100, 0, 111, 0, 48, 0, 51, 0, 0, 0, 100, 0, 111, 0, 48, 0, 52, 0, 0, 0, 100, 0, 111, 0, 48, 0, 53, 0, 0, 0, 100, 0, 111, 0, 48, 0, 54, 0, 0, 0, 100, 0, 111, 0, 48, 0, 55, 0, 0, 0, 100, 0, 111, 0, 48, 0, 56, 0, 0, 0, 100, 0, 111, 0, 48, 0, 57, 0, 0, 0, 100, 0, 111, 0, 49, 0, 48, 0, 0, 0, 100, 0, 111, 0, 49, 0, 49, 0, 0, 0, 100, 0, 111, 0, 49, 0, 50, 0, 0, 0, 100, 0, 111, 0, 49, 0, 51, 0, 0, 0, 100, 0, 111, 0, 49, 0, 52, 0, 0, 0, 100, 0, 111, 0, 49, 0, 53, 0, 0, 0, 100, 0, 111, 0, 49, 0, 54, 0, 0, 0, 100, 0, 111, 0, 49, 0, 55, 0, 0, 0, 100, 0, 111, 0, 49, 0, 56, 0, 0, 0, 100, 0, 111, 0, 49, 0, 57, 0, 0, 0, 100, 0, 111, 0, 50, 0, 48, 0, 0, 0, 100, 0, 111, 0, 50, 0, 49, 0, 0, 0, 100, 0, 111, 0, 50, 0, 50, 0, 0, 0, 100, 0, 111, 0, 50, 0, 51, 0, 0, 0, 100, 0, 111, 0, 50, 0, 52, 0, 0, 0, 100, 0, 111, 0, 50, 0, 53, 0, 0, 0, 100, 0, 111, 0, 50, 0, 54, 0, 0, 0, 100, 0, 111, 0, 50, 0, 55, 0, 0, 0, 100, 0, 111, 0, 50, 0, 56, 0, 0, 0, 100, 0, 111, 0, 50, 0, 57, 0, 0, 0, 100, 0, 111, 0, 51, 0, 48, 0, 0, 0, 100, 0, 111, 0, 51, 0, 49, 0, 0, 0, 100, 0, 111, 0, 51, 0, 50, 0, 0, 0, 100, 0, 111, 0, 51, 0, 51, 0, 0, 0, 100, 0, 111, 0, 51, 0, 52, 0, 0, 0, 100, 0, 111, 0, 51, 0, 53, 0, 0, 0, 100, 0, 111, 0, 51, 0, 54, 0, 0, 0, 100, 0, 111, 0, 51, 0, 55, 0, 0, 0, 100, 0, 111, 0, 51, 0, 56, 0, 0, 0, 100, 0, 111, 0, 51, 0, 57, 0, 0, 0, 100, 0, 111, 0, 52, 0, 48, 0, 0, 0, 100, 0, 111, 0, 52, 0, 49, 0, 0, 0, 100, 0, 111, 0, 52, 0, 50, 0, 0, 0, 100, 0, 122, 0, 48, 0, 49, 0, 0, 0, 100, 0, 122, 0, 48, 0, 50, 0, 0, 0, 100, 0, 122, 0, 48, 0, 51, 0, 0, 0, 100, 0, 122, 0, 48, 0, 52, 0, 0, 0, 100, 0, 122, 0, 48, 0, 53, 0, 0, 0, 100, 0, 122, 0, 48, 0, 54, 0, 0, 0, 100, 0, 122, 0, 48, 0, 55, 0, 0, 0, 100, 0, 122, 0, 48, 0, 56, 0, 0, 0, 100, 0, 122, 0, 48, 0, 57, 0, 0, 0, 100, 0, 122, 0, 49, 0, 48, 0, 0, 0, 100, 0, 122, 0, 49, 0, 49, 0, 0, 0, 100, 0, 122, 0, 49, 0, 50, 0, 0, 0, 100, 0, 122, 0, 49, 0, 51, 0, 0, 0, 100, 0, 122, 0, 49, 0, 52, 0, 0, 0, 100, 0, 122, 0, 49, 0, 53, 0, 0, 0, 100, 0, 122, 0, 49, 0, 54, 0, 0, 0, 100, 0, 122, 0, 49, 0, 55, 0, 0, 0, 100, 0, 122, 0, 49, 0, 56, 0, 0, 0, 100, 0, 122, 0, 49, 0, 57, 0, 0, 0, 100, 0, 122, 0, 50, 0, 48, 0, 0, 0, 100, 0, 122, 0, 50, 0, 49, 0, 0, 0, 100, 0, 122, 0, 50, 0, 50, 0, 0, 0, 100, 0, 122, 0, 50, 0, 51, 0, 0, 0, 100, 0, 122, 0, 50, 0, 52, 0, 0, 0, 100, 0, 122, 0, 50, 0, 53, 0, 0, 0, 100, 0, 122, 0, 50, 0, 54, 0, 0, 0, 100, 0, 122, 0, 50, 0, 55, 0, 0, 0, 100, 0, 122, 0, 50, 0, 56, 0, 0, 0, 100, 0, 122, 0, 50, 0, 57, 0, 0, 0, 100, 0, 122, 0, 51, 0, 48, 0, 0, 0, 100, 0, 122, 0, 51, 0, 49, 0, 0, 0, 100, 0, 122, 0, 51, 0, 51, 0, 0, 0, 100, 0, 122, 0, 51, 0, 52, 0, 0, 0, 100, 0, 122, 0, 51, 0, 53, 0, 0, 0, 100, 0, 122, 0, 51, 0, 55, 0, 0, 0, 100, 0, 122, 0, 51, 0, 56, 0, 0, 0, 100, 0, 122, 0, 51, 0, 57, 0, 0, 0, 100, 0, 122, 0, 52, 0, 48, 0, 0, 0, 100, 0, 122, 0, 52, 0, 49, 0, 0, 0, 100, 0, 122, 0, 52, 0, 50, 0, 0, 0, 100, 0, 122, 0, 52, 0, 51, 0, 0, 0, 100, 0, 122, 0, 52, 0, 53, 0, 0, 0, 100, 0, 122, 0, 52, 0, 54, 0, 0, 0, 100, 0, 122, 0, 52, 0, 55, 0, 0, 0, 101, 0, 99, 0, 115, 0, 100, 0, 0, 0, 101, 0, 99, 0, 115, 0, 101, 0, 0, 0, 101, 0, 115, 0, 97, 0, 114, 0, 0, 0, 101, 0, 115, 0, 97, 0, 115, 0, 0, 0, 101, 0, 115, 0, 99, 0, 97, 0, 0, 0, 101, 0, 115, 0, 99, 0, 99, 0, 0, 0, 101, 0, 115, 0, 99, 0, 108, 0, 0, 0, 101, 0, 115, 0, 99, 0, 109, 0, 0, 0, 101, 0, 115, 0, 99, 0, 114, 0, 0, 0, 101, 0, 115, 0, 99, 0, 115, 0, 0, 0, 101, 0, 115, 0, 99, 0, 116, 0, 0, 0, 101, 0, 115, 0, 99, 0, 117, 0, 0, 0, 101, 0, 115, 0, 109, 0, 99, 0, 0, 0, 101, 0, 115, 0, 109, 0, 100, 0, 0, 0, 101, 0, 115, 0, 116, 0, 101, 0, 0, 0, 101, 0, 115, 0, 116, 0, 102, 0, 0, 0, 101, 0, 116, 0, 115, 0, 110, 0, 0, 0, 101, 0, 116, 0, 115, 0, 111, 0, 0, 0, 102, 0, 105, 0, 48, 0, 50, 0, 0, 0, 102, 0, 105, 0, 48, 0, 51, 0, 0, 0, 102, 0, 105, 0, 48, 0, 52, 0, 0, 0, 102, 0, 105, 0, 48, 0, 53, 0, 0, 0, 102, 0, 105, 0, 48, 0, 54, 0, 0, 0, 102, 0, 105, 0, 48, 0, 55, 0, 0, 0, 102, 0, 105, 0, 48, 0, 56, 0, 0, 0, 102, 0, 105, 0, 48, 0, 57, 0, 0, 0, 102, 0, 105, 0, 49, 0, 48, 0, 0, 0, 102, 0, 105, 0, 49, 0, 49, 0, 0, 0, 102, 0, 105, 0, 49, 0, 50, 0, 0, 0, 102, 0, 105, 0, 49, 0, 51, 0, 0, 0, 102, 0, 105, 0, 49, 0, 52, 0, 0, 0, 102, 0, 105, 0, 49, 0, 53, 0, 0, 0, 102, 0, 105, 0, 49, 0, 54, 0, 0, 0, 102, 0, 105, 0, 49, 0, 55, 0, 0, 0, 102, 0, 105, 0, 49, 0, 56, 0, 0, 0, 102, 0, 105, 0, 49, 0, 57, 0, 0, 0, 102, 0, 106, 0, 48, 0, 49, 0, 0, 0, 102, 0, 106, 0, 48, 0, 50, 0, 0, 0, 102, 0, 106, 0, 48, 0, 51, 0, 0, 0, 102, 0, 106, 0, 48, 0, 52, 0, 0, 0, 102, 0, 106, 0, 48, 0, 53, 0, 0, 0, 102, 0, 106, 0, 48, 0, 54, 0, 0, 0, 102, 0, 106, 0, 48, 0, 55, 0, 0, 0, 102, 0, 106, 0, 48, 0, 56, 0, 0, 0, 102, 0, 106, 0, 48, 0, 57, 0, 0, 0, 102, 0, 106, 0, 49, 0, 48, 0, 0, 0, 102, 0, 106, 0, 49, 0, 49, 0, 0, 0, 102, 0, 106, 0, 49, 0, 50, 0, 0, 0, 102, 0, 106, 0, 49, 0, 51, 0, 0, 0, 102, 0, 106, 0, 49, 0, 52, 0, 0, 0, 102, 0, 114, 0, 48, 0, 49, 0, 0, 0, 102, 0, 114, 0, 48, 0, 50, 0, 0, 0, 102, 0, 114, 0, 48, 0, 51, 0, 0, 0, 102, 0, 114, 0, 48, 0, 52, 0, 0, 0, 102, 0, 114, 0, 48, 0, 53, 0, 0, 0, 102, 0, 114, 0, 48, 0, 54, 0, 0, 0, 102, 0, 114, 0, 48, 0, 55, 0, 0, 0, 102, 0, 114, 0, 48, 0, 56, 0, 0, 0, 102, 0, 114, 0, 48, 0, 57, 0, 0, 0, 102, 0, 114, 0, 49, 0, 48, 0, 0, 0, 102, 0, 114, 0, 49, 0, 49, 0, 0, 0, 102, 0, 114, 0, 49, 0, 50, 0, 0, 0, 102, 0, 114, 0, 49, 0, 51, 0, 0, 0, 102, 0, 114, 0, 49, 0, 52, 0, 0, 0, 102, 0, 114, 0, 49, 0, 53, 0, 0, 0, 102, 0, 114, 0, 49, 0, 54, 0, 0, 0, 102, 0, 114, 0, 49, 0, 55, 0, 0, 0, 102, 0, 114, 0, 49, 0, 56, 0, 0, 0, 102, 0, 114, 0, 49, 0, 57, 0, 0, 0, 102, 0, 114, 0, 50, 0, 49, 0, 0, 0, 102, 0, 114, 0, 50, 0, 50, 0, 0, 0, 102, 0, 114, 0, 50, 0, 51, 0, 0, 0, 102, 0, 114, 0, 50, 0, 52, 0, 0, 0, 102, 0, 114, 0, 50, 0, 53, 0, 0, 0, 102, 0, 114, 0, 50, 0, 54, 0, 0, 0, 102, 0, 114, 0, 50, 0, 55, 0, 0, 0, 102, 0, 114, 0, 50, 0, 56, 0, 0, 0, 102, 0, 114, 0, 50, 0, 57, 0, 0, 0, 102, 0, 114, 0, 50, 0, 97, 0, 0, 0, 102, 0, 114, 0, 50, 0, 98, 0, 0, 0, 102, 0, 114, 0, 51, 0, 48, 0, 0, 0, 102, 0, 114, 0, 51, 0, 49, 0, 0, 0, 102, 0, 114, 0, 51, 0, 50, 0, 0, 0, 102, 0, 114, 0, 51, 0, 51, 0, 0, 0, 102, 0, 114, 0, 51, 0, 52, 0, 0, 0, 102, 0, 114, 0, 51, 0, 53, 0, 0, 0, 102, 0, 114, 0, 51, 0, 54, 0, 0, 0, 102, 0, 114, 0, 51, 0, 55, 0, 0, 0, 102, 0, 114, 0, 51, 0, 56, 0, 0, 0, 102, 0, 114, 0, 51, 0, 57, 0, 0, 0, 102, 0, 114, 0, 52, 0, 48, 0, 0, 0, 102, 0, 114, 0, 52, 0, 49, 0, 0, 0, 102, 0, 114, 0, 52, 0, 50, 0, 0, 0, 102, 0, 114, 0, 52, 0, 51, 0, 0, 0, 102, 0, 114, 0, 52, 0, 52, 0, 0, 0, 102, 0, 114, 0, 52, 0, 53, 0, 0, 0, 102, 0, 114, 0, 52, 0, 54, 0, 0, 0, 102, 0, 114, 0, 52, 0, 55, 0, 0, 0, 102, 0, 114, 0, 52, 0, 56, 0, 0, 0, 102, 0, 114, 0, 52, 0, 57, 0, 0, 0, 102, 0, 114, 0, 53, 0, 48, 0, 0, 0, 102, 0, 114, 0, 53, 0, 49, 0, 0, 0, 102, 0, 114, 0, 53, 0, 50, 0, 0, 0, 102, 0, 114, 0, 53, 0, 51, 0, 0, 0, 102, 0, 114, 0, 53, 0, 52, 0, 0, 0, 102, 0, 114, 0, 53, 0, 53, 0, 0, 0, 102, 0, 114, 0, 53, 0, 54, 0, 0, 0, 102, 0, 114, 0, 53, 0, 55, 0, 0, 0, 102, 0, 114, 0, 53, 0, 56, 0, 0, 0, 102, 0, 114, 0, 53, 0, 57, 0, 0, 0, 102, 0, 114, 0, 54, 0, 48, 0, 0, 0, 102, 0, 114, 0, 54, 0, 49, 0, 0, 0, 102, 0, 114, 0, 54, 0, 50, 0, 0, 0, 102, 0, 114, 0, 54, 0, 51, 0, 0, 0, 102, 0, 114, 0, 54, 0, 52, 0, 0, 0, 102, 0, 114, 0, 54, 0, 53, 0, 0, 0, 102, 0, 114, 0, 54, 0, 54, 0, 0, 0, 102, 0, 114, 0, 54, 0, 55, 0, 0, 0, 102, 0, 114, 0, 54, 0, 56, 0, 0, 0, 102, 0, 114, 0, 54, 0, 57, 0, 0, 0, 102, 0, 114, 0, 55, 0, 48, 0, 0, 0, 102, 0, 114, 0, 55, 0, 49, 0, 0, 0, 102, 0, 114, 0, 55, 0, 50, 0, 0, 0, 102, 0, 114, 0, 55, 0, 51, 0, 0, 0, 102, 0, 114, 0, 55, 0, 52, 0, 0, 0, 102, 0, 114, 0, 55, 0, 53, 0, 0, 0, 102, 0, 114, 0, 55, 0, 54, 0, 0, 0, 102, 0, 114, 0, 55, 0, 55, 0, 0, 0, 102, 0, 114, 0, 55, 0, 56, 0, 0, 0, 102, 0, 114, 0, 55, 0, 57, 0, 0, 0, 102, 0, 114, 0, 56, 0, 48, 0, 0, 0, 102, 0, 114, 0, 56, 0, 49, 0, 0, 0, 102, 0, 114, 0, 56, 0, 50, 0, 0, 0, 102, 0, 114, 0, 56, 0, 51, 0, 0, 0, 102, 0, 114, 0, 56, 0, 54, 0, 0, 0, 102, 0, 114, 0, 56, 0, 55, 0, 0, 0, 102, 0, 114, 0, 56, 0, 56, 0, 0, 0, 102, 0, 114, 0, 56, 0, 57, 0, 0, 0, 102, 0, 114, 0, 57, 0, 49, 0, 0, 0, 102, 0, 114, 0, 57, 0, 51, 0, 0, 0, 103, 0, 100, 0, 48, 0, 49, 0, 0, 0, 103, 0, 100, 0, 48, 0, 50, 0, 0, 0, 103, 0, 100, 0, 48, 0, 51, 0, 0, 0, 103, 0, 100, 0, 48, 0, 52, 0, 0, 0, 103, 0, 100, 0, 48, 0, 53, 0, 0, 0, 103, 0, 100, 0, 48, 0, 54, 0, 0, 0, 103, 0, 108, 0, 113, 0, 97, 0, 0, 0, 103, 0, 110, 0, 98, 0, 101, 0, 0, 0, 103, 0, 110, 0, 107, 0, 97, 0, 0, 0, 103, 0, 110, 0, 107, 0, 98, 0, 0, 0, 103, 0, 110, 0, 107, 0, 100, 0, 0, 0, 103, 0, 110, 0, 107, 0, 101, 0, 0, 0, 103, 0, 110, 0, 107, 0, 110, 0, 0, 0, 103, 0, 110, 0, 107, 0, 111, 0, 0, 0, 103, 0, 110, 0, 109, 0, 99, 0, 0, 0, 103, 0, 110, 0, 109, 0, 100, 0, 0, 0, 103, 0, 110, 0, 109, 0, 108, 0, 0, 0, 103, 0, 110, 0, 109, 0, 109, 0, 0, 0, 103, 0, 114, 0, 48, 0, 49, 0, 0, 0, 103, 0, 114, 0, 97, 0, 49, 0, 0, 0, 103, 0, 119, 0, 98, 0, 108, 0, 0, 0, 104, 0, 110, 0, 99, 0, 108, 0, 0, 0, 104, 0, 110, 0, 99, 0, 109, 0, 0, 0, 104, 0, 114, 0, 48, 0, 49, 0, 0, 0, 104, 0, 114, 0, 48, 0, 50, 0, 0, 0, 104, 0, 114, 0, 48, 0, 51, 0, 0, 0, 104, 0, 114, 0, 48, 0, 52, 0, 0, 0, 104, 0, 114, 0, 48, 0, 53, 0, 0, 0, 104, 0, 114, 0, 48, 0, 54, 0, 0, 0, 104, 0, 114, 0, 48, 0, 55, 0, 0, 0, 104, 0, 114, 0, 48, 0, 56, 0, 0, 0, 104, 0, 114, 0, 48, 0, 57, 0, 0, 0, 104, 0, 114, 0, 49, 0, 48, 0, 0, 0, 104, 0, 114, 0, 49, 0, 49, 0, 0, 0, 104, 0, 114, 0, 49, 0, 50, 0, 0, 0, 104, 0, 114, 0, 49, 0, 51, 0, 0, 0, 104, 0, 114, 0, 49, 0, 52, 0, 0, 0, 104, 0, 114, 0, 49, 0, 53, 0, 0, 0, 104, 0, 114, 0, 49, 0, 54, 0, 0, 0, 104, 0, 114, 0, 49, 0, 55, 0, 0, 0, 104, 0, 114, 0, 49, 0, 56, 0, 0, 0, 104, 0, 114, 0, 49, 0, 57, 0, 0, 0, 104, 0, 114, 0, 50, 0, 48, 0, 0, 0, 104, 0, 114, 0, 50, 0, 49, 0, 0, 0, 104, 0, 116, 0, 110, 0, 100, 0, 0, 0, 104, 0, 116, 0, 110, 0, 101, 0, 0, 0, 104, 0, 116, 0, 115, 0, 100, 0, 0, 0, 104, 0, 116, 0, 115, 0, 101, 0, 0, 0, 104, 0, 117, 0, 115, 0, 111, 0, 0, 0, 104, 0, 117, 0, 115, 0, 116, 0, 0, 0, 105, 0, 100, 0, 98, 0, 97, 0, 0, 0, 105, 0, 100, 0, 98, 0, 98, 0, 0, 0, 105, 0, 100, 0, 106, 0, 97, 0, 0, 0, 105, 0, 100, 0, 106, 0, 98, 0, 0, 0, 105, 0, 100, 0, 107, 0, 97, 0, 0, 0, 105, 0, 100, 0, 107, 0, 98, 0, 0, 0, 105, 0, 100, 0, 107, 0, 114, 0, 0, 0, 105, 0, 100, 0, 107, 0, 115, 0, 0, 0, 105, 0, 100, 0, 107, 0, 116, 0, 0, 0, 105, 0, 100, 0, 107, 0, 117, 0, 0, 0, 105, 0, 100, 0, 110, 0, 116, 0, 0, 0, 105, 0, 100, 0, 110, 0, 117, 0, 0, 0, 105, 0, 100, 0, 112, 0, 97, 0, 0, 0, 105, 0, 100, 0, 112, 0, 98, 0, 0, 0, 105, 0, 100, 0, 115, 0, 97, 0, 0, 0, 105, 0, 100, 0, 115, 0, 109, 0, 0, 0, 105, 0, 100, 0, 115, 0, 110, 0, 0, 0, 105, 0, 100, 0, 115, 0, 114, 0, 0, 0, 105, 0, 100, 0, 115, 0, 115, 0, 0, 0, 105, 0, 100, 0, 115, 0, 116, 0, 0, 0, 105, 0, 100, 0, 115, 0, 117, 0, 0, 0, 105, 0, 101, 0, 109, 0, 110, 0, 0, 0, 105, 0, 101, 0, 109, 0, 111, 0, 0, 0, 105, 0, 101, 0, 119, 0, 119, 0, 0, 0, 105, 0, 101, 0, 119, 0, 120, 0, 0, 0, 105, 0, 110, 0, 97, 0, 114, 0, 0, 0, 105, 0, 110, 0, 97, 0, 115, 0, 0, 0, 105, 0, 113, 0, 98, 0, 97, 0, 0, 0, 105, 0, 113, 0, 98, 0, 98, 0, 0, 0, 105, 0, 114, 0, 48, 0, 49, 0, 0, 0, 105, 0, 114, 0, 48, 0, 50, 0, 0, 0, 105, 0, 114, 0, 48, 0, 51, 0, 0, 0, 105, 0, 114, 0, 48, 0, 52, 0, 0, 0, 105, 0, 114, 0, 48, 0, 53, 0, 0, 0, 105, 0, 114, 0, 48, 0, 54, 0, 0, 0, 105, 0, 114, 0, 48, 0, 55, 0, 0, 0, 105, 0, 114, 0, 48, 0, 56, 0, 0, 0, 105, 0, 114, 0, 49, 0, 48, 0, 0, 0, 105, 0, 114, 0, 49, 0, 49, 0, 0, 0, 105, 0, 114, 0, 49, 0, 50, 0, 0, 0, 105, 0, 114, 0, 49, 0, 51, 0, 0, 0, 105, 0, 114, 0, 49, 0, 52, 0, 0, 0, 105, 0, 114, 0, 49, 0, 53, 0, 0, 0, 105, 0, 114, 0, 49, 0, 54, 0, 0, 0, 105, 0, 114, 0, 49, 0, 55, 0, 0, 0, 105, 0, 114, 0, 49, 0, 56, 0, 0, 0, 105, 0, 114, 0, 49, 0, 57, 0, 0, 0, 105, 0, 114, 0, 50, 0, 48, 0, 0, 0, 105, 0, 114, 0, 50, 0, 49, 0, 0, 0, 105, 0, 114, 0, 50, 0, 50, 0, 0, 0, 105, 0, 114, 0, 50, 0, 51, 0, 0, 0, 105, 0, 114, 0, 50, 0, 52, 0, 0, 0, 105, 0, 114, 0, 50, 0, 53, 0, 0, 0, 105, 0, 114, 0, 50, 0, 54, 0, 0, 0, 105, 0, 114, 0, 50, 0, 55, 0, 0, 0, 105, 0, 114, 0, 50, 0, 56, 0, 0, 0, 105, 0, 114, 0, 50, 0, 57, 0, 0, 0, 105, 0, 114, 0, 51, 0, 48, 0, 0, 0, 105, 0, 114, 0, 51, 0, 49, 0, 0, 0, 105, 0, 114, 0, 51, 0, 50, 0, 0, 0, 105, 0, 116, 0, 55, 0, 55, 0, 0, 0, 105, 0, 116, 0, 55, 0, 56, 0, 0, 0, 105, 0, 116, 0, 97, 0, 110, 0, 0, 0, 105, 0, 116, 0, 97, 0, 111, 0, 0, 0, 105, 0, 116, 0, 97, 0, 112, 0, 0, 0, 105, 0, 116, 0, 97, 0, 113, 0, 0, 0, 105, 0, 116, 0, 97, 0, 114, 0, 0, 0, 105, 0, 116, 0, 98, 0, 114, 0, 0, 0, 105, 0, 116, 0, 98, 0, 115, 0, 0, 0, 105, 0, 116, 0, 98, 0, 116, 0, 0, 0, 105, 0, 116, 0, 99, 0, 97, 0, 0, 0, 105, 0, 116, 0, 99, 0, 98, 0, 0, 0, 105, 0, 116, 0, 99, 0, 104, 0, 0, 0, 105, 0, 116, 0, 99, 0, 110, 0, 0, 0, 105, 0, 116, 0, 99, 0, 111, 0, 0, 0, 105, 0, 116, 0, 99, 0, 114, 0, 0, 0, 105, 0, 116, 0, 99, 0, 115, 0, 0, 0, 105, 0, 116, 0, 99, 0, 116, 0, 0, 0, 105, 0, 116, 0, 103, 0, 111, 0, 0, 0, 105, 0, 116, 0, 108, 0, 116, 0, 0, 0, 105, 0, 116, 0, 108, 0, 117, 0, 0, 0, 105, 0, 116, 0, 109, 0, 99, 0, 0, 0, 105, 0, 116, 0, 109, 0, 110, 0, 0, 0, 105, 0, 116, 0, 109, 0, 111, 0, 0, 0, 105, 0, 116, 0, 109, 0, 116, 0, 0, 0, 105, 0, 116, 0, 111, 0, 103, 0, 0, 0, 105, 0, 116, 0, 111, 0, 116, 0, 0, 0, 105, 0, 116, 0, 112, 0, 100, 0, 0, 0, 105, 0, 116, 0, 112, 0, 101, 0, 0, 0, 105, 0, 116, 0, 112, 0, 111, 0, 0, 0, 105, 0, 116, 0, 112, 0, 116, 0, 0, 0, 105, 0, 116, 0, 112, 0, 117, 0, 0, 0, 105, 0, 116, 0, 112, 0, 118, 0, 0, 0, 105, 0, 116, 0, 114, 0, 109, 0, 0, 0, 105, 0, 116, 0, 114, 0, 110, 0, 0, 0, 105, 0, 116, 0, 114, 0, 111, 0, 0, 0, 105, 0, 116, 0, 115, 0, 100, 0, 0, 0, 105, 0, 116, 0, 115, 0, 111, 0, 0, 0, 105, 0, 116, 0, 115, 0, 112, 0, 0, 0, 105, 0, 116, 0, 115, 0, 114, 0, 0, 0, 105, 0, 116, 0, 115, 0, 115, 0, 0, 0, 105, 0, 116, 0, 116, 0, 111, 0, 0, 0, 105, 0, 116, 0, 116, 0, 112, 0, 0, 0, 105, 0, 116, 0, 116, 0, 114, 0, 0, 0, 105, 0, 116, 0, 117, 0, 100, 0, 0, 0, 105, 0, 116, 0, 118, 0, 98, 0, 0, 0, 105, 0, 116, 0, 118, 0, 99, 0, 0, 0, 105, 0, 116, 0, 118, 0, 114, 0, 0, 0, 105, 0, 116, 0, 118, 0, 116, 0, 0, 0, 106, 0, 109, 0, 48, 0, 49, 0, 0, 0, 106, 0, 109, 0, 48, 0, 50, 0, 0, 0, 106, 0, 109, 0, 48, 0, 51, 0, 0, 0, 106, 0, 109, 0, 48, 0, 52, 0, 0, 0, 106, 0, 109, 0, 48, 0, 53, 0, 0, 0, 106, 0, 109, 0, 48, 0, 54, 0, 0, 0, 106, 0, 109, 0, 48, 0, 55, 0, 0, 0, 106, 0, 109, 0, 48, 0, 56, 0, 0, 0, 106, 0, 109, 0, 48, 0, 57, 0, 0, 0, 106, 0, 109, 0, 49, 0, 48, 0, 0, 0, 106, 0, 109, 0, 49, 0, 49, 0, 0, 0, 106, 0, 109, 0, 49, 0, 50, 0, 0, 0, 106, 0, 109, 0, 49, 0, 51, 0, 0, 0, 106, 0, 109, 0, 49, 0, 52, 0, 0, 0, 106, 0, 112, 0, 48, 0, 49, 0, 0, 0, 106, 0, 112, 0, 48, 0, 50, 0, 0, 0, 106, 0, 112, 0, 48, 0, 51, 0, 0, 0, 106, 0, 112, 0, 48, 0, 52, 0, 0, 0, 106, 0, 112, 0, 48, 0, 53, 0, 0, 0, 106, 0, 112, 0, 48, 0, 54, 0, 0, 0, 106, 0, 112, 0, 48, 0, 55, 0, 0, 0, 106, 0, 112, 0, 48, 0, 56, 0, 0, 0, 106, 0, 112, 0, 48, 0, 57, 0, 0, 0, 106, 0, 112, 0, 49, 0, 48, 0, 0, 0, 106, 0, 112, 0, 49, 0, 49, 0, 0, 0, 106, 0, 112, 0, 49, 0, 50, 0, 0, 0, 106, 0, 112, 0, 49, 0, 51, 0, 0, 0, 106, 0, 112, 0, 49, 0, 52, 0, 0, 0, 106, 0, 112, 0, 49, 0, 53, 0, 0, 0, 106, 0, 112, 0, 49, 0, 54, 0, 0, 0, 106, 0, 112, 0, 49, 0, 55, 0, 0, 0, 106, 0, 112, 0, 49, 0, 56, 0, 0, 0, 106, 0, 112, 0, 49, 0, 57, 0, 0, 0, 106, 0, 112, 0, 50, 0, 48, 0, 0, 0, 106, 0, 112, 0, 50, 0, 49, 0, 0, 0, 106, 0, 112, 0, 50, 0, 50, 0, 0, 0, 106, 0, 112, 0, 50, 0, 51, 0, 0, 0, 106, 0, 112, 0, 50, 0, 52, 0, 0, 0, 106, 0, 112, 0, 50, 0, 53, 0, 0, 0, 106, 0, 112, 0, 50, 0, 54, 0, 0, 0, 106, 0, 112, 0, 50, 0, 55, 0, 0, 0, 106, 0, 112, 0, 50, 0, 56, 0, 0, 0, 106, 0, 112, 0, 50, 0, 57, 0, 0, 0, 106, 0, 112, 0, 51, 0, 48, 0, 0, 0, 106, 0, 112, 0, 51, 0, 49, 0, 0, 0, 106, 0, 112, 0, 51, 0, 50, 0, 0, 0, 106, 0, 112, 0, 51, 0, 51, 0, 0, 0, 106, 0, 112, 0, 51, 0, 52, 0, 0, 0, 106, 0, 112, 0, 51, 0, 53, 0, 0, 0, 106, 0, 112, 0, 51, 0, 54, 0, 0, 0, 106, 0, 112, 0, 51, 0, 55, 0, 0, 0, 106, 0, 112, 0, 51, 0, 56, 0, 0, 0, 106, 0, 112, 0, 51, 0, 57, 0, 0, 0, 106, 0, 112, 0, 52, 0, 48, 0, 0, 0, 106, 0, 112, 0, 52, 0, 49, 0, 0, 0, 106, 0, 112, 0, 52, 0, 50, 0, 0, 0, 106, 0, 112, 0, 52, 0, 51, 0, 0, 0, 106, 0, 112, 0, 52, 0, 52, 0, 0, 0, 106, 0, 112, 0, 52, 0, 53, 0, 0, 0, 106, 0, 112, 0, 52, 0, 54, 0, 0, 0, 106, 0, 112, 0, 52, 0, 55, 0, 0, 0, 107, 0, 101, 0, 48, 0, 49, 0, 0, 0, 107, 0, 101, 0, 48, 0, 50, 0, 0, 0, 107, 0, 101, 0, 48, 0, 51, 0, 0, 0, 107, 0, 101, 0, 48, 0, 52, 0, 0, 0, 107, 0, 101, 0, 48, 0, 53, 0, 0, 0, 107, 0, 101, 0, 48, 0, 54, 0, 0, 0, 107, 0, 101, 0, 48, 0, 55, 0, 0, 0, 107, 0, 101, 0, 48, 0, 56, 0, 0, 0, 107, 0, 101, 0, 48, 0, 57, 0, 0, 0, 107, 0, 101, 0, 49, 0, 48, 0, 0, 0, 107, 0, 101, 0, 49, 0, 49, 0, 0, 0, 107, 0, 101, 0, 49, 0, 50, 0, 0, 0, 107, 0, 101, 0, 49, 0, 51, 0, 0, 0, 107, 0, 101, 0, 49, 0, 52, 0, 0, 0, 107, 0, 101, 0, 49, 0, 53, 0, 0, 0, 107, 0, 101, 0, 49, 0, 54, 0, 0, 0, 107, 0, 101, 0, 49, 0, 55, 0, 0, 0, 107, 0, 101, 0, 49, 0, 56, 0, 0, 0, 107, 0, 101, 0, 49, 0, 57, 0, 0, 0, 107, 0, 101, 0, 50, 0, 48, 0, 0, 0, 107, 0, 101, 0, 50, 0, 49, 0, 0, 0, 107, 0, 101, 0, 50, 0, 50, 0, 0, 0, 107, 0, 101, 0, 50, 0, 51, 0, 0, 0, 107, 0, 101, 0, 50, 0, 52, 0, 0, 0, 107, 0, 101, 0, 50, 0, 53, 0, 0, 0, 107, 0, 101, 0, 50, 0, 54, 0, 0, 0, 107, 0, 101, 0, 50, 0, 55, 0, 0, 0, 107, 0, 101, 0, 50, 0, 56, 0, 0, 0, 107, 0, 101, 0, 50, 0, 57, 0, 0, 0, 107, 0, 101, 0, 51, 0, 48, 0, 0, 0, 107, 0, 101, 0, 51, 0, 49, 0, 0, 0, 107, 0, 101, 0, 51, 0, 50, 0, 0, 0, 107, 0, 101, 0, 51, 0, 51, 0, 0, 0, 107, 0, 101, 0, 51, 0, 52, 0, 0, 0, 107, 0, 101, 0, 51, 0, 53, 0, 0, 0, 107, 0, 101, 0, 51, 0, 54, 0, 0, 0, 107, 0, 101, 0, 51, 0, 55, 0, 0, 0, 107, 0, 101, 0, 51, 0, 56, 0, 0, 0, 107, 0, 101, 0, 51, 0, 57, 0, 0, 0, 107, 0, 101, 0, 52, 0, 48, 0, 0, 0, 107, 0, 101, 0, 52, 0, 49, 0, 0, 0, 107, 0, 101, 0, 52, 0, 50, 0, 0, 0, 107, 0, 101, 0, 52, 0, 51, 0, 0, 0, 107, 0, 101, 0, 52, 0, 52, 0, 0, 0, 107, 0, 101, 0, 52, 0, 53, 0, 0, 0, 107, 0, 101, 0, 52, 0, 54, 0, 0, 0, 107, 0, 101, 0, 52, 0, 55, 0, 0, 0, 107, 0, 104, 0, 49, 0, 48, 0, 0, 0, 107, 0, 104, 0, 49, 0, 49, 0, 0, 0, 107, 0, 104, 0, 49, 0, 50, 0, 0, 0, 107, 0, 104, 0, 49, 0, 51, 0, 0, 0, 107, 0, 104, 0, 49, 0, 52, 0, 0, 0, 107, 0, 104, 0, 49, 0, 53, 0, 0, 0, 107, 0, 104, 0, 49, 0, 54, 0, 0, 0, 107, 0, 104, 0, 49, 0, 55, 0, 0, 0, 107, 0, 104, 0, 49, 0, 56, 0, 0, 0, 107, 0, 104, 0, 49, 0, 57, 0, 0, 0, 107, 0, 104, 0, 50, 0, 48, 0, 0, 0, 107, 0, 104, 0, 50, 0, 49, 0, 0, 0, 107, 0, 104, 0, 50, 0, 50, 0, 0, 0, 107, 0, 104, 0, 50, 0, 51, 0, 0, 0, 107, 0, 104, 0, 50, 0, 52, 0, 0, 0, 107, 0, 104, 0, 50, 0, 53, 0, 0, 0, 107, 0, 110, 0, 48, 0, 49, 0, 0, 0, 107, 0, 110, 0, 48, 0, 50, 0, 0, 0, 107, 0, 110, 0, 48, 0, 51, 0, 0, 0, 107, 0, 110, 0, 48, 0, 52, 0, 0, 0, 107, 0, 110, 0, 48, 0, 53, 0, 0, 0, 107, 0, 110, 0, 48, 0, 54, 0, 0, 0, 107, 0, 110, 0, 48, 0, 55, 0, 0, 0, 107, 0, 110, 0, 48, 0, 56, 0, 0, 0, 107, 0, 110, 0, 48, 0, 57, 0, 0, 0, 107, 0, 110, 0, 49, 0, 48, 0, 0, 0, 107, 0, 110, 0, 49, 0, 49, 0, 0, 0, 107, 0, 110, 0, 49, 0, 50, 0, 0, 0, 107, 0, 110, 0, 49, 0, 51, 0, 0, 0, 107, 0, 112, 0, 48, 0, 49, 0, 0, 0, 107, 0, 112, 0, 48, 0, 50, 0, 0, 0, 107, 0, 112, 0, 48, 0, 51, 0, 0, 0, 107, 0, 112, 0, 48, 0, 52, 0, 0, 0, 107, 0, 112, 0, 48, 0, 53, 0, 0, 0, 107, 0, 112, 0, 48, 0, 54, 0, 0, 0, 107, 0, 112, 0, 48, 0, 55, 0, 0, 0, 107, 0, 112, 0, 48, 0, 56, 0, 0, 0, 107, 0, 112, 0, 48, 0, 57, 0, 0, 0, 107, 0, 112, 0, 49, 0, 51, 0, 0, 0, 107, 0, 112, 0, 49, 0, 52, 0, 0, 0, 107, 0, 114, 0, 50, 0, 54, 0, 0, 0, 107, 0, 114, 0, 50, 0, 55, 0, 0, 0, 107, 0, 114, 0, 50, 0, 56, 0, 0, 0, 107, 0, 114, 0, 50, 0, 57, 0, 0, 0, 107, 0, 114, 0, 51, 0, 48, 0, 0, 0, 107, 0, 114, 0, 51, 0, 49, 0, 0, 0, 107, 0, 114, 0, 52, 0, 49, 0, 0, 0, 107, 0, 114, 0, 52, 0, 50, 0, 0, 0, 107, 0, 114, 0, 52, 0, 51, 0, 0, 0, 107, 0, 114, 0, 52, 0, 52, 0, 0, 0, 107, 0, 114, 0, 52, 0, 53, 0, 0, 0, 107, 0, 114, 0, 52, 0, 54, 0, 0, 0, 107, 0, 114, 0, 52, 0, 55, 0, 0, 0, 107, 0, 114, 0, 52, 0, 56, 0, 0, 0, 107, 0, 114, 0, 52, 0, 57, 0, 0, 0, 108, 0, 97, 0, 98, 0, 107, 0, 0, 0, 108, 0, 97, 0, 98, 0, 108, 0, 0, 0, 108, 0, 97, 0, 120, 0, 110, 0, 0, 0, 108, 0, 98, 0, 98, 0, 104, 0, 0, 0, 108, 0, 98, 0, 98, 0, 105, 0, 0, 0, 108, 0, 99, 0, 48, 0, 49, 0, 0, 0, 108, 0, 99, 0, 48, 0, 50, 0, 0, 0, 108, 0, 99, 0, 48, 0, 51, 0, 0, 0, 108, 0, 99, 0, 48, 0, 53, 0, 0, 0, 108, 0, 99, 0, 48, 0, 54, 0, 0, 0, 108, 0, 99, 0, 48, 0, 55, 0, 0, 0, 108, 0, 99, 0, 48, 0, 56, 0, 0, 0, 108, 0, 99, 0, 49, 0, 48, 0, 0, 0, 108, 0, 99, 0, 49, 0, 49, 0, 0, 0, 108, 0, 99, 0, 49, 0, 50, 0, 0, 0, 108, 0, 105, 0, 48, 0, 49, 0, 0, 0, 108, 0, 105, 0, 48, 0, 50, 0, 0, 0, 108, 0, 105, 0, 48, 0, 51, 0, 0, 0, 108, 0, 105, 0, 48, 0, 52, 0, 0, 0, 108, 0, 105, 0, 48, 0, 53, 0, 0, 0, 108, 0, 105, 0, 48, 0, 54, 0, 0, 0, 108, 0, 105, 0, 48, 0, 55, 0, 0, 0, 108, 0, 105, 0, 48, 0, 56, 0, 0, 0, 108, 0, 105, 0, 48, 0, 57, 0, 0, 0, 108, 0, 105, 0, 49, 0, 48, 0, 0, 0, 108, 0, 105, 0, 49, 0, 49, 0, 0, 0, 108, 0, 107, 0, 49, 0, 49, 0, 0, 0, 108, 0, 107, 0, 49, 0, 50, 0, 0, 0, 108, 0, 107, 0, 49, 0, 51, 0, 0, 0, 108, 0, 107, 0, 50, 0, 49, 0, 0, 0, 108, 0, 107, 0, 50, 0, 50, 0, 0, 0, 108, 0, 107, 0, 50, 0, 51, 0, 0, 0, 108, 0, 107, 0, 51, 0, 49, 0, 0, 0, 108, 0, 107, 0, 51, 0, 50, 0, 0, 0, 108, 0, 107, 0, 51, 0, 51, 0, 0, 0, 108, 0, 107, 0, 52, 0, 49, 0, 0, 0, 108, 0, 107, 0, 52, 0, 50, 0, 0, 0, 108, 0, 107, 0, 52, 0, 51, 0, 0, 0, 108, 0, 107, 0, 52, 0, 52, 0, 0, 0, 108, 0, 107, 0, 52, 0, 53, 0, 0, 0, 108, 0, 107, 0, 53, 0, 49, 0, 0, 0, 108, 0, 107, 0, 53, 0, 50, 0, 0, 0, 108, 0, 107, 0, 53, 0, 51, 0, 0, 0, 108, 0, 107, 0, 54, 0, 49, 0, 0, 0, 108, 0, 107, 0, 54, 0, 50, 0, 0, 0, 108, 0, 107, 0, 55, 0, 49, 0, 0, 0, 108, 0, 107, 0, 55, 0, 50, 0, 0, 0, 108, 0, 107, 0, 56, 0, 49, 0, 0, 0, 108, 0, 107, 0, 56, 0, 50, 0, 0, 0, 108, 0, 107, 0, 57, 0, 49, 0, 0, 0, 108, 0, 107, 0, 57, 0, 50, 0, 0, 0, 108, 0, 116, 0, 48, 0, 49, 0, 0, 0, 108, 0, 116, 0, 48, 0, 50, 0, 0, 0, 108, 0, 116, 0, 48, 0, 51, 0, 0, 0, 108, 0, 116, 0, 48, 0, 52, 0, 0, 0, 108, 0, 116, 0, 48, 0, 53, 0, 0, 0, 108, 0, 116, 0, 48, 0, 54, 0, 0, 0, 108, 0, 116, 0, 48, 0, 55, 0, 0, 0, 108, 0, 116, 0, 48, 0, 56, 0, 0, 0, 108, 0, 116, 0, 48, 0, 57, 0, 0, 0, 108, 0, 116, 0, 49, 0, 48, 0, 0, 0, 108, 0, 116, 0, 49, 0, 49, 0, 0, 0, 108, 0, 116, 0, 49, 0, 50, 0, 0, 0, 108, 0, 116, 0, 49, 0, 51, 0, 0, 0, 108, 0, 116, 0, 49, 0, 52, 0, 0, 0, 108, 0, 116, 0, 49, 0, 53, 0, 0, 0, 108, 0, 116, 0, 49, 0, 54, 0, 0, 0, 108, 0, 116, 0, 49, 0, 55, 0, 0, 0, 108, 0, 116, 0, 49, 0, 56, 0, 0, 0, 108, 0, 116, 0, 49, 0, 57, 0, 0, 0, 108, 0, 116, 0, 50, 0, 48, 0, 0, 0, 108, 0, 116, 0, 50, 0, 49, 0, 0, 0, 108, 0, 116, 0, 50, 0, 50, 0, 0, 0, 108, 0, 116, 0, 50, 0, 51, 0, 0, 0, 108, 0, 116, 0, 50, 0, 52, 0, 0, 0, 108, 0, 116, 0, 50, 0, 53, 0, 0, 0, 108, 0, 116, 0, 50, 0, 54, 0, 0, 0, 108, 0, 116, 0, 50, 0, 55, 0, 0, 0, 108, 0, 116, 0, 50, 0, 56, 0, 0, 0, 108, 0, 116, 0, 50, 0, 57, 0, 0, 0, 108, 0, 116, 0, 51, 0, 48, 0, 0, 0, 108, 0, 116, 0, 51, 0, 49, 0, 0, 0, 108, 0, 116, 0, 51, 0, 50, 0, 0, 0, 108, 0, 116, 0, 51, 0, 51, 0, 0, 0, 108, 0, 116, 0, 51, 0, 52, 0, 0, 0, 108, 0, 116, 0, 51, 0, 53, 0, 0, 0, 108, 0, 116, 0, 51, 0, 54, 0, 0, 0, 108, 0, 116, 0, 51, 0, 55, 0, 0, 0, 108, 0, 116, 0, 51, 0, 56, 0, 0, 0, 108, 0, 116, 0, 51, 0, 57, 0, 0, 0, 108, 0, 116, 0, 52, 0, 48, 0, 0, 0, 108, 0, 116, 0, 52, 0, 49, 0, 0, 0, 108, 0, 116, 0, 52, 0, 50, 0, 0, 0, 108, 0, 116, 0, 52, 0, 51, 0, 0, 0, 108, 0, 116, 0, 52, 0, 52, 0, 0, 0, 108, 0, 116, 0, 52, 0, 53, 0, 0, 0, 108, 0, 116, 0, 52, 0, 54, 0, 0, 0, 108, 0, 116, 0, 52, 0, 55, 0, 0, 0, 108, 0, 116, 0, 52, 0, 56, 0, 0, 0, 108, 0, 116, 0, 52, 0, 57, 0, 0, 0, 108, 0, 116, 0, 53, 0, 48, 0, 0, 0, 108, 0, 116, 0, 53, 0, 49, 0, 0, 0, 108, 0, 116, 0, 53, 0, 50, 0, 0, 0, 108, 0, 116, 0, 53, 0, 51, 0, 0, 0, 108, 0, 116, 0, 53, 0, 52, 0, 0, 0, 108, 0, 116, 0, 53, 0, 53, 0, 0, 0, 108, 0, 116, 0, 53, 0, 54, 0, 0, 0, 108, 0, 116, 0, 53, 0, 55, 0, 0, 0, 108, 0, 116, 0, 53, 0, 56, 0, 0, 0, 108, 0, 116, 0, 53, 0, 57, 0, 0, 0, 108, 0, 121, 0, 109, 0, 105, 0, 0, 0, 108, 0, 121, 0, 109, 0, 106, 0, 0, 0, 109, 0, 97, 0, 48, 0, 49, 0, 0, 0, 109, 0, 97, 0, 48, 0, 50, 0, 0, 0, 109, 0, 97, 0, 48, 0, 51, 0, 0, 0, 109, 0, 97, 0, 48, 0, 52, 0, 0, 0, 109, 0, 97, 0, 48, 0, 53, 0, 0, 0, 109, 0, 97, 0, 48, 0, 54, 0, 0, 0, 109, 0, 97, 0, 48, 0, 55, 0, 0, 0, 109, 0, 97, 0, 48, 0, 56, 0, 0, 0, 109, 0, 97, 0, 48, 0, 57, 0, 0, 0, 109, 0, 97, 0, 49, 0, 48, 0, 0, 0, 109, 0, 97, 0, 49, 0, 49, 0, 0, 0, 109, 0, 97, 0, 49, 0, 50, 0, 0, 0, 109, 0, 99, 0, 115, 0, 111, 0, 0, 0, 109, 0, 99, 0, 115, 0, 112, 0, 0, 0, 109, 0, 100, 0, 98, 0, 114, 0, 0, 0, 109, 0, 100, 0, 98, 0, 115, 0, 0, 0, 109, 0, 100, 0, 99, 0, 108, 0, 0, 0, 109, 0, 100, 0, 99, 0, 109, 0, 0, 0, 109, 0, 100, 0, 99, 0, 115, 0, 0, 0, 109, 0, 100, 0, 99, 0, 116, 0, 0, 0, 109, 0, 100, 0, 99, 0, 117, 0, 0, 0, 109, 0, 100, 0, 115, 0, 110, 0, 0, 0, 109, 0, 100, 0, 115, 0, 111, 0, 0, 0, 109, 0, 101, 0, 48, 0, 49, 0, 0, 0, 109, 0, 101, 0, 48, 0, 50, 0, 0, 0, 109, 0, 101, 0, 48, 0, 51, 0, 0, 0, 109, 0, 101, 0, 48, 0, 52, 0, 0, 0, 109, 0, 101, 0, 48, 0, 53, 0, 0, 0, 109, 0, 101, 0, 48, 0, 54, 0, 0, 0, 109, 0, 101, 0, 48, 0, 55, 0, 0, 0, 109, 0, 101, 0, 48, 0, 56, 0, 0, 0, 109, 0, 101, 0, 48, 0, 57, 0, 0, 0, 109, 0, 101, 0, 49, 0, 48, 0, 0, 0, 109, 0, 101, 0, 49, 0, 49, 0, 0, 0, 109, 0, 101, 0, 49, 0, 50, 0, 0, 0, 109, 0, 101, 0, 49, 0, 51, 0, 0, 0, 109, 0, 101, 0, 49, 0, 52, 0, 0, 0, 109, 0, 101, 0, 49, 0, 53, 0, 0, 0, 109, 0, 101, 0, 49, 0, 54, 0, 0, 0, 109, 0, 101, 0, 49, 0, 55, 0, 0, 0, 109, 0, 101, 0, 49, 0, 56, 0, 0, 0, 109, 0, 101, 0, 49, 0, 57, 0, 0, 0, 109, 0, 101, 0, 50, 0, 48, 0, 0, 0, 109, 0, 101, 0, 50, 0, 49, 0, 0, 0, 109, 0, 101, 0, 50, 0, 50, 0, 0, 0, 109, 0, 101, 0, 50, 0, 51, 0, 0, 0, 109, 0, 107, 0, 48, 0, 49, 0, 0, 0, 109, 0, 107, 0, 48, 0, 50, 0, 0, 0, 109, 0, 107, 0, 48, 0, 51, 0, 0, 0, 109, 0, 107, 0, 48, 0, 52, 0, 0, 0, 109, 0, 107, 0, 48, 0, 53, 0, 0, 0, 109, 0, 107, 0, 48, 0, 54, 0, 0, 0, 109, 0, 107, 0, 48, 0, 55, 0, 0, 0, 109, 0, 107, 0, 48, 0, 56, 0, 0, 0, 109, 0, 107, 0, 48, 0, 57, 0, 0, 0, 109, 0, 107, 0, 49, 0, 48, 0, 0, 0, 109, 0, 107, 0, 49, 0, 49, 0, 0, 0, 109, 0, 107, 0, 49, 0, 50, 0, 0, 0, 109, 0, 107, 0, 49, 0, 51, 0, 0, 0, 109, 0, 107, 0, 49, 0, 52, 0, 0, 0, 109, 0, 107, 0, 49, 0, 53, 0, 0, 0, 109, 0, 107, 0, 49, 0, 55, 0, 0, 0, 109, 0, 107, 0, 49, 0, 56, 0, 0, 0, 109, 0, 107, 0, 49, 0, 57, 0, 0, 0, 109, 0, 107, 0, 50, 0, 48, 0, 0, 0, 109, 0, 107, 0, 50, 0, 49, 0, 0, 0, 109, 0, 107, 0, 50, 0, 50, 0, 0, 0, 109, 0, 107, 0, 50, 0, 51, 0, 0, 0, 109, 0, 107, 0, 50, 0, 52, 0, 0, 0, 109, 0, 107, 0, 50, 0, 53, 0, 0, 0, 109, 0, 107, 0, 50, 0, 54, 0, 0, 0, 109, 0, 107, 0, 50, 0, 55, 0, 0, 0, 109, 0, 107, 0, 51, 0, 49, 0, 0, 0, 109, 0, 107, 0, 51, 0, 50, 0, 0, 0, 109, 0, 107, 0, 51, 0, 51, 0, 0, 0, 109, 0, 107, 0, 51, 0, 52, 0, 0, 0, 109, 0, 107, 0, 51, 0, 53, 0, 0, 0, 109, 0, 107, 0, 51, 0, 54, 0, 0, 0, 109, 0, 107, 0, 51, 0, 55, 0, 0, 0, 109, 0, 107, 0, 52, 0, 48, 0, 0, 0, 109, 0, 107, 0, 52, 0, 49, 0, 0, 0, 109, 0, 107, 0, 52, 0, 50, 0, 0, 0, 109, 0, 107, 0, 52, 0, 51, 0, 0, 0, 109, 0, 107, 0, 52, 0, 52, 0, 0, 0, 109, 0, 107, 0, 52, 0, 53, 0, 0, 0, 109, 0, 107, 0, 52, 0, 54, 0, 0, 0, 109, 0, 107, 0, 52, 0, 55, 0, 0, 0, 109, 0, 107, 0, 52, 0, 56, 0, 0, 0, 109, 0, 107, 0, 52, 0, 57, 0, 0, 0, 109, 0, 107, 0, 53, 0, 48, 0, 0, 0, 109, 0, 107, 0, 53, 0, 49, 0, 0, 0, 109, 0, 107, 0, 53, 0, 50, 0, 0, 0, 109, 0, 107, 0, 53, 0, 51, 0, 0, 0, 109, 0, 107, 0, 53, 0, 52, 0, 0, 0, 109, 0, 107, 0, 53, 0, 53, 0, 0, 0, 109, 0, 107, 0, 53, 0, 54, 0, 0, 0, 109, 0, 107, 0, 53, 0, 55, 0, 0, 0, 109, 0, 107, 0, 53, 0, 56, 0, 0, 0, 109, 0, 107, 0, 53, 0, 57, 0, 0, 0, 109, 0, 107, 0, 54, 0, 48, 0, 0, 0, 109, 0, 107, 0, 54, 0, 49, 0, 0, 0, 109, 0, 107, 0, 54, 0, 50, 0, 0, 0, 109, 0, 107, 0, 54, 0, 51, 0, 0, 0, 109, 0, 107, 0, 54, 0, 52, 0, 0, 0, 109, 0, 107, 0, 54, 0, 53, 0, 0, 0, 109, 0, 107, 0, 54, 0, 54, 0, 0, 0, 109, 0, 107, 0, 54, 0, 55, 0, 0, 0, 109, 0, 107, 0, 54, 0, 56, 0, 0, 0, 109, 0, 107, 0, 55, 0, 48, 0, 0, 0, 109, 0, 107, 0, 55, 0, 49, 0, 0, 0, 109, 0, 107, 0, 55, 0, 50, 0, 0, 0, 109, 0, 107, 0, 55, 0, 51, 0, 0, 0, 109, 0, 107, 0, 55, 0, 52, 0, 0, 0, 109, 0, 107, 0, 55, 0, 53, 0, 0, 0, 109, 0, 107, 0, 55, 0, 54, 0, 0, 0, 109, 0, 107, 0, 55, 0, 55, 0, 0, 0, 109, 0, 107, 0, 55, 0, 57, 0, 0, 0, 109, 0, 107, 0, 56, 0, 48, 0, 0, 0, 109, 0, 107, 0, 56, 0, 49, 0, 0, 0, 109, 0, 107, 0, 56, 0, 50, 0, 0, 0, 109, 0, 107, 0, 56, 0, 51, 0, 0, 0, 109, 0, 107, 0, 56, 0, 52, 0, 0, 0, 109, 0, 109, 0, 48, 0, 49, 0, 0, 0, 109, 0, 109, 0, 48, 0, 50, 0, 0, 0, 109, 0, 109, 0, 48, 0, 51, 0, 0, 0, 109, 0, 109, 0, 48, 0, 52, 0, 0, 0, 109, 0, 109, 0, 48, 0, 53, 0, 0, 0, 109, 0, 109, 0, 48, 0, 54, 0, 0, 0, 109, 0, 109, 0, 48, 0, 55, 0, 0, 0, 109, 0, 109, 0, 49, 0, 49, 0, 0, 0, 109, 0, 109, 0, 49, 0, 50, 0, 0, 0, 109, 0, 109, 0, 49, 0, 51, 0, 0, 0, 109, 0, 109, 0, 49, 0, 52, 0, 0, 0, 109, 0, 109, 0, 49, 0, 53, 0, 0, 0, 109, 0, 109, 0, 49, 0, 54, 0, 0, 0, 109, 0, 109, 0, 49, 0, 55, 0, 0, 0, 109, 0, 109, 0, 49, 0, 56, 0, 0, 0, 109, 0, 114, 0, 48, 0, 49, 0, 0, 0, 109, 0, 114, 0, 48, 0, 50, 0, 0, 0, 109, 0, 114, 0, 48, 0, 51, 0, 0, 0, 109, 0, 114, 0, 48, 0, 52, 0, 0, 0, 109, 0, 114, 0, 48, 0, 53, 0, 0, 0, 109, 0, 114, 0, 48, 0, 54, 0, 0, 0, 109, 0, 114, 0, 48, 0, 55, 0, 0, 0, 109, 0, 114, 0, 48, 0, 56, 0, 0, 0, 109, 0, 114, 0, 48, 0, 57, 0, 0, 0, 109, 0, 114, 0, 49, 0, 48, 0, 0, 0, 109, 0, 114, 0, 49, 0, 49, 0, 0, 0, 109, 0, 114, 0, 49, 0, 50, 0, 0, 0, 109, 0, 114, 0, 49, 0, 51, 0, 0, 0, 109, 0, 114, 0, 49, 0, 52, 0, 0, 0, 109, 0, 114, 0, 49, 0, 53, 0, 0, 0, 109, 0, 116, 0, 48, 0, 49, 0, 0, 0, 109, 0, 116, 0, 48, 0, 50, 0, 0, 0, 109, 0, 116, 0, 48, 0, 51, 0, 0, 0, 109, 0, 116, 0, 48, 0, 52, 0, 0, 0, 109, 0, 116, 0, 48, 0, 53, 0, 0, 0, 109, 0, 116, 0, 48, 0, 54, 0, 0, 0, 109, 0, 116, 0, 48, 0, 55, 0, 0, 0, 109, 0, 116, 0, 48, 0, 56, 0, 0, 0, 109, 0, 116, 0, 48, 0, 57, 0, 0, 0, 109, 0, 116, 0, 49, 0, 48, 0, 0, 0, 109, 0, 116, 0, 49, 0, 49, 0, 0, 0, 109, 0, 116, 0, 49, 0, 50, 0, 0, 0, 109, 0, 116, 0, 49, 0, 51, 0, 0, 0, 109, 0, 116, 0, 49, 0, 52, 0, 0, 0, 109, 0, 116, 0, 49, 0, 53, 0, 0, 0, 109, 0, 116, 0, 49, 0, 54, 0, 0, 0, 109, 0, 116, 0, 49, 0, 55, 0, 0, 0, 109, 0, 116, 0, 49, 0, 56, 0, 0, 0, 109, 0, 116, 0, 49, 0, 57, 0, 0, 0, 109, 0, 116, 0, 50, 0, 48, 0, 0, 0, 109, 0, 116, 0, 50, 0, 49, 0, 0, 0, 109, 0, 116, 0, 50, 0, 50, 0, 0, 0, 109, 0, 116, 0, 50, 0, 51, 0, 0, 0, 109, 0, 116, 0, 50, 0, 52, 0, 0, 0, 109, 0, 116, 0, 50, 0, 53, 0, 0, 0, 109, 0, 116, 0, 50, 0, 54, 0, 0, 0, 109, 0, 116, 0, 50, 0, 55, 0, 0, 0, 109, 0, 116, 0, 50, 0, 56, 0, 0, 0, 109, 0, 116, 0, 50, 0, 57, 0, 0, 0, 109, 0, 116, 0, 51, 0, 48, 0, 0, 0, 109, 0, 116, 0, 51, 0, 49, 0, 0, 0, 109, 0, 116, 0, 51, 0, 50, 0, 0, 0, 109, 0, 116, 0, 51, 0, 51, 0, 0, 0, 109, 0, 116, 0, 51, 0, 52, 0, 0, 0, 109, 0, 116, 0, 51, 0, 53, 0, 0, 0, 109, 0, 116, 0, 51, 0, 54, 0, 0, 0, 109, 0, 116, 0, 51, 0, 55, 0, 0, 0, 109, 0, 116, 0, 51, 0, 56, 0, 0, 0, 109, 0, 116, 0, 51, 0, 57, 0, 0, 0, 109, 0, 116, 0, 52, 0, 48, 0, 0, 0, 109, 0, 116, 0, 52, 0, 49, 0, 0, 0, 109, 0, 116, 0, 52, 0, 50, 0, 0, 0, 109, 0, 116, 0, 52, 0, 51, 0, 0, 0, 109, 0, 116, 0, 52, 0, 52, 0, 0, 0, 109, 0, 116, 0, 52, 0, 53, 0, 0, 0, 109, 0, 116, 0, 52, 0, 54, 0, 0, 0, 109, 0, 116, 0, 52, 0, 55, 0, 0, 0, 109, 0, 116, 0, 52, 0, 56, 0, 0, 0, 109, 0, 116, 0, 52, 0, 57, 0, 0, 0, 109, 0, 116, 0, 53, 0, 49, 0, 0, 0, 109, 0, 116, 0, 53, 0, 50, 0, 0, 0, 109, 0, 116, 0, 53, 0, 51, 0, 0, 0, 109, 0, 116, 0, 53, 0, 52, 0, 0, 0, 109, 0, 116, 0, 53, 0, 53, 0, 0, 0, 109, 0, 116, 0, 53, 0, 54, 0, 0, 0, 109, 0, 116, 0, 53, 0, 55, 0, 0, 0, 109, 0, 116, 0, 53, 0, 56, 0, 0, 0, 109, 0, 116, 0, 53, 0, 57, 0, 0, 0, 109, 0, 116, 0, 54, 0, 49, 0, 0, 0, 109, 0, 116, 0, 54, 0, 50, 0, 0, 0, 109, 0, 116, 0, 54, 0, 51, 0, 0, 0, 109, 0, 116, 0, 54, 0, 52, 0, 0, 0, 109, 0, 116, 0, 54, 0, 53, 0, 0, 0, 109, 0, 116, 0, 54, 0, 54, 0, 0, 0, 109, 0, 116, 0, 54, 0, 55, 0, 0, 0, 109, 0, 118, 0, 48, 0, 48, 0, 0, 0, 109, 0, 118, 0, 48, 0, 49, 0, 0, 0, 109, 0, 118, 0, 48, 0, 50, 0, 0, 0, 109, 0, 118, 0, 48, 0, 51, 0, 0, 0, 109, 0, 118, 0, 48, 0, 52, 0, 0, 0, 109, 0, 118, 0, 48, 0, 53, 0, 0, 0, 109, 0, 118, 0, 48, 0, 55, 0, 0, 0, 109, 0, 118, 0, 48, 0, 56, 0, 0, 0, 109, 0, 118, 0, 49, 0, 50, 0, 0, 0, 109, 0, 118, 0, 49, 0, 51, 0, 0, 0, 109, 0, 118, 0, 49, 0, 52, 0, 0, 0, 109, 0, 118, 0, 50, 0, 51, 0, 0, 0, 109, 0, 118, 0, 50, 0, 52, 0, 0, 0, 109, 0, 118, 0, 50, 0, 53, 0, 0, 0, 109, 0, 118, 0, 50, 0, 54, 0, 0, 0, 109, 0, 118, 0, 50, 0, 55, 0, 0, 0, 109, 0, 118, 0, 50, 0, 56, 0, 0, 0, 109, 0, 118, 0, 50, 0, 57, 0, 0, 0, 109, 0, 118, 0, 99, 0, 101, 0, 0, 0, 109, 0, 118, 0, 110, 0, 99, 0, 0, 0, 109, 0, 118, 0, 115, 0, 99, 0, 0, 0, 109, 0, 118, 0, 115, 0, 117, 0, 0, 0, 109, 0, 118, 0, 117, 0, 110, 0, 0, 0, 109, 0, 118, 0, 117, 0, 115, 0, 0, 0, 109, 0, 121, 0, 48, 0, 49, 0, 0, 0, 109, 0, 121, 0, 48, 0, 50, 0, 0, 0, 109, 0, 121, 0, 48, 0, 51, 0, 0, 0, 109, 0, 121, 0, 48, 0, 52, 0, 0, 0, 109, 0, 121, 0, 48, 0, 53, 0, 0, 0, 109, 0, 121, 0, 48, 0, 54, 0, 0, 0, 109, 0, 121, 0, 48, 0, 55, 0, 0, 0, 109, 0, 121, 0, 48, 0, 56, 0, 0, 0, 109, 0, 121, 0, 48, 0, 57, 0, 0, 0, 109, 0, 121, 0, 49, 0, 48, 0, 0, 0, 109, 0, 121, 0, 49, 0, 49, 0, 0, 0, 109, 0, 121, 0, 49, 0, 50, 0, 0, 0, 109, 0, 121, 0, 49, 0, 51, 0, 0, 0, 109, 0, 121, 0, 49, 0, 52, 0, 0, 0, 109, 0, 121, 0, 49, 0, 53, 0, 0, 0, 109, 0, 121, 0, 49, 0, 54, 0, 0, 0, 110, 0, 97, 0, 111, 0, 115, 0, 0, 0, 110, 0, 97, 0, 111, 0, 116, 0, 0, 0, 110, 0, 103, 0, 107, 0, 100, 0, 0, 0, 110, 0, 103, 0, 107, 0, 101, 0, 0, 0, 110, 0, 103, 0, 107, 0, 110, 0, 0, 0, 110, 0, 103, 0, 107, 0, 111, 0, 0, 0, 110, 0, 105, 0, 109, 0, 115, 0, 0, 0, 110, 0, 105, 0, 109, 0, 116, 0, 0, 0, 110, 0, 111, 0, 48, 0, 49, 0, 0, 0, 110, 0, 111, 0, 48, 0, 50, 0, 0, 0, 110, 0, 111, 0, 48, 0, 51, 0, 0, 0, 110, 0, 111, 0, 48, 0, 52, 0, 0, 0, 110, 0, 111, 0, 48, 0, 53, 0, 0, 0, 110, 0, 111, 0, 48, 0, 54, 0, 0, 0, 110, 0, 111, 0, 48, 0, 55, 0, 0, 0, 110, 0, 111, 0, 48, 0, 56, 0, 0, 0, 110, 0, 111, 0, 48, 0, 57, 0, 0, 0, 110, 0, 111, 0, 49, 0, 48, 0, 0, 0, 110, 0, 111, 0, 49, 0, 49, 0, 0, 0, 110, 0, 111, 0, 49, 0, 50, 0, 0, 0, 110, 0, 111, 0, 49, 0, 52, 0, 0, 0, 110, 0, 111, 0, 49, 0, 53, 0, 0, 0, 110, 0, 111, 0, 49, 0, 56, 0, 0, 0, 110, 0, 111, 0, 49, 0, 57, 0, 0, 0, 110, 0, 111, 0, 50, 0, 48, 0, 0, 0, 110, 0, 111, 0, 50, 0, 49, 0, 0, 0, 110, 0, 111, 0, 50, 0, 50, 0, 0, 0, 110, 0, 111, 0, 53, 0, 48, 0, 0, 0, 110, 0, 112, 0, 112, 0, 49, 0, 0, 0, 110, 0, 112, 0, 112, 0, 50, 0, 0, 0, 110, 0, 112, 0, 112, 0, 51, 0, 0, 0, 110, 0, 112, 0, 112, 0, 52, 0, 0, 0, 110, 0, 112, 0, 112, 0, 53, 0, 0, 0, 110, 0, 112, 0, 112, 0, 54, 0, 0, 0, 110, 0, 112, 0, 112, 0, 55, 0, 0, 0, 110, 0, 114, 0, 48, 0, 49, 0, 0, 0, 110, 0, 114, 0, 48, 0, 50, 0, 0, 0, 110, 0, 114, 0, 48, 0, 51, 0, 0, 0, 110, 0, 114, 0, 48, 0, 52, 0, 0, 0, 110, 0, 114, 0, 48, 0, 53, 0, 0, 0, 110, 0, 114, 0, 48, 0, 54, 0, 0, 0, 110, 0, 114, 0, 48, 0, 55, 0, 0, 0, 110, 0, 114, 0, 48, 0, 56, 0, 0, 0, 110, 0, 114, 0, 48, 0, 57, 0, 0, 0, 110, 0, 114, 0, 49, 0, 48, 0, 0, 0, 110, 0, 114, 0, 49, 0, 49, 0, 0, 0, 110, 0, 114, 0, 49, 0, 50, 0, 0, 0, 110, 0, 114, 0, 49, 0, 51, 0, 0, 0, 110, 0, 114, 0, 49, 0, 52, 0, 0, 0, 111, 0, 109, 0, 98, 0, 97, 0, 0, 0, 112, 0, 104, 0, 48, 0, 48, 0, 0, 0, 112, 0, 104, 0, 48, 0, 49, 0, 0, 0, 112, 0, 104, 0, 48, 0, 50, 0, 0, 0, 112, 0, 104, 0, 48, 0, 51, 0, 0, 0, 112, 0, 104, 0, 48, 0, 53, 0, 0, 0, 112, 0, 104, 0, 48, 0, 54, 0, 0, 0, 112, 0, 104, 0, 48, 0, 55, 0, 0, 0, 112, 0, 104, 0, 48, 0, 56, 0, 0, 0, 112, 0, 104, 0, 48, 0, 57, 0, 0, 0, 112, 0, 104, 0, 49, 0, 48, 0, 0, 0, 112, 0, 104, 0, 49, 0, 49, 0, 0, 0, 112, 0, 104, 0, 49, 0, 50, 0, 0, 0, 112, 0, 104, 0, 49, 0, 51, 0, 0, 0, 112, 0, 104, 0, 49, 0, 52, 0, 0, 0, 112, 0, 104, 0, 49, 0, 53, 0, 0, 0, 112, 0, 104, 0, 52, 0, 48, 0, 0, 0, 112, 0, 104, 0, 52, 0, 49, 0, 0, 0, 112, 0, 108, 0, 100, 0, 115, 0, 0, 0, 112, 0, 108, 0, 107, 0, 112, 0, 0, 0, 112, 0, 108, 0, 108, 0, 100, 0, 0, 0, 112, 0, 108, 0, 109, 0, 97, 0, 0, 0, 112, 0, 108, 0, 109, 0, 122, 0, 0, 0, 112, 0, 108, 0, 111, 0, 112, 0, 0, 0, 112, 0, 108, 0, 112, 0, 100, 0, 0, 0, 112, 0, 108, 0, 112, 0, 107, 0, 0, 0, 112, 0, 108, 0, 112, 0, 109, 0, 0, 0, 112, 0, 108, 0, 119, 0, 110, 0, 0, 0, 112, 0, 108, 0, 119, 0, 112, 0, 0, 0, 112, 0, 108, 0, 122, 0, 112, 0, 0, 0, 112, 0, 116, 0, 48, 0, 49, 0, 0, 0, 112, 0, 116, 0, 48, 0, 50, 0, 0, 0, 112, 0, 116, 0, 48, 0, 51, 0, 0, 0, 112, 0, 116, 0, 48, 0, 52, 0, 0, 0, 112, 0, 116, 0, 48, 0, 53, 0, 0, 0, 112, 0, 116, 0, 48, 0, 54, 0, 0, 0, 112, 0, 116, 0, 48, 0, 55, 0, 0, 0, 112, 0, 116, 0, 48, 0, 56, 0, 0, 0, 112, 0, 116, 0, 48, 0, 57, 0, 0, 0, 112, 0, 116, 0, 49, 0, 48, 0, 0, 0, 112, 0, 116, 0, 49, 0, 49, 0, 0, 0, 112, 0, 116, 0, 49, 0, 50, 0, 0, 0, 112, 0, 116, 0, 49, 0, 51, 0, 0, 0, 112, 0, 116, 0, 49, 0, 52, 0, 0, 0, 112, 0, 116, 0, 49, 0, 53, 0, 0, 0, 112, 0, 116, 0, 49, 0, 54, 0, 0, 0, 112, 0, 116, 0, 49, 0, 55, 0, 0, 0, 112, 0, 116, 0, 49, 0, 56, 0, 0, 0, 112, 0, 121, 0, 49, 0, 48, 0, 0, 0, 112, 0, 121, 0, 49, 0, 49, 0, 0, 0, 112, 0, 121, 0, 49, 0, 50, 0, 0, 0, 112, 0, 121, 0, 49, 0, 51, 0, 0, 0, 112, 0, 121, 0, 49, 0, 52, 0, 0, 0, 112, 0, 121, 0, 49, 0, 53, 0, 0, 0, 112, 0, 121, 0, 49, 0, 54, 0, 0, 0, 114, 0, 111, 0, 99, 0, 115, 0, 0, 0, 114, 0, 111, 0, 99, 0, 116, 0, 0, 0, 114, 0, 111, 0, 116, 0, 108, 0, 0, 0, 114, 0, 111, 0, 116, 0, 109, 0, 0, 0, 114, 0, 115, 0, 48, 0, 48, 0, 0, 0, 114, 0, 115, 0, 48, 0, 49, 0, 0, 0, 114, 0, 115, 0, 48, 0, 50, 0, 0, 0, 114, 0, 115, 0, 48, 0, 51, 0, 0, 0, 114, 0, 115, 0, 48, 0, 52, 0, 0, 0, 114, 0, 115, 0, 48, 0, 53, 0, 0, 0, 114, 0, 115, 0, 48, 0, 54, 0, 0, 0, 114, 0, 115, 0, 48, 0, 55, 0, 0, 0, 114, 0, 115, 0, 48, 0, 56, 0, 0, 0, 114, 0, 115, 0, 48, 0, 57, 0, 0, 0, 114, 0, 115, 0, 49, 0, 48, 0, 0, 0, 114, 0, 115, 0, 49, 0, 49, 0, 0, 0, 114, 0, 115, 0, 49, 0, 50, 0, 0, 0, 114, 0, 115, 0, 49, 0, 51, 0, 0, 0, 114, 0, 115, 0, 49, 0, 52, 0, 0, 0, 114, 0, 115, 0, 49, 0, 53, 0, 0, 0, 114, 0, 115, 0, 49, 0, 54, 0, 0, 0, 114, 0, 115, 0, 49, 0, 55, 0, 0, 0, 114, 0, 115, 0, 49, 0, 56, 0, 0, 0, 114, 0, 115, 0, 49, 0, 57, 0, 0, 0, 114, 0, 115, 0, 50, 0, 48, 0, 0, 0, 114, 0, 115, 0, 50, 0, 49, 0, 0, 0, 114, 0, 115, 0, 50, 0, 50, 0, 0, 0, 114, 0, 115, 0, 50, 0, 51, 0, 0, 0, 114, 0, 115, 0, 50, 0, 52, 0, 0, 0, 114, 0, 115, 0, 50, 0, 53, 0, 0, 0, 114, 0, 115, 0, 50, 0, 54, 0, 0, 0, 114, 0, 115, 0, 50, 0, 55, 0, 0, 0, 114, 0, 115, 0, 50, 0, 56, 0, 0, 0, 114, 0, 115, 0, 50, 0, 57, 0, 0, 0, 114, 0, 117, 0, 107, 0, 98, 0, 0, 0, 114, 0, 117, 0, 107, 0, 99, 0, 0, 0, 114, 0, 117, 0, 107, 0, 107, 0, 0, 0, 114, 0, 117, 0, 107, 0, 108, 0, 0, 0, 114, 0, 119, 0, 48, 0, 49, 0, 0, 0, 114, 0, 119, 0, 48, 0, 50, 0, 0, 0, 114, 0, 119, 0, 48, 0, 51, 0, 0, 0, 114, 0, 119, 0, 48, 0, 52, 0, 0, 0, 114, 0, 119, 0, 48, 0, 53, 0, 0, 0, 115, 0, 97, 0, 48, 0, 49, 0, 0, 0, 115, 0, 97, 0, 48, 0, 50, 0, 0, 0, 115, 0, 97, 0, 48, 0, 51, 0, 0, 0, 115, 0, 97, 0, 48, 0, 52, 0, 0, 0, 115, 0, 97, 0, 48, 0, 53, 0, 0, 0, 115, 0, 97, 0, 48, 0, 54, 0, 0, 0, 115, 0, 97, 0, 48, 0, 55, 0, 0, 0, 115, 0, 97, 0, 48, 0, 56, 0, 0, 0, 115, 0, 97, 0, 48, 0, 57, 0, 0, 0, 115, 0, 97, 0, 49, 0, 48, 0, 0, 0, 115, 0, 97, 0, 49, 0, 49, 0, 0, 0, 115, 0, 97, 0, 49, 0, 50, 0, 0, 0, 115, 0, 98, 0, 109, 0, 107, 0, 0, 0, 115, 0, 98, 0, 109, 0, 108, 0, 0, 0, 115, 0, 99, 0, 48, 0, 49, 0, 0, 0, 115, 0, 99, 0, 48, 0, 50, 0, 0, 0, 115, 0, 99, 0, 48, 0, 51, 0, 0, 0, 115, 0, 99, 0, 48, 0, 52, 0, 0, 0, 115, 0, 99, 0, 48, 0, 53, 0, 0, 0, 115, 0, 99, 0, 48, 0, 54, 0, 0, 0, 115, 0, 99, 0, 48, 0, 55, 0, 0, 0, 115, 0, 99, 0, 48, 0, 56, 0, 0, 0, 115, 0, 99, 0, 48, 0, 57, 0, 0, 0, 115, 0, 99, 0, 49, 0, 48, 0, 0, 0, 115, 0, 99, 0, 49, 0, 49, 0, 0, 0, 115, 0, 99, 0, 49, 0, 50, 0, 0, 0, 115, 0, 99, 0, 49, 0, 51, 0, 0, 0, 115, 0, 99, 0, 49, 0, 52, 0, 0, 0, 115, 0, 99, 0, 49, 0, 53, 0, 0, 0, 115, 0, 99, 0, 49, 0, 54, 0, 0, 0, 115, 0, 99, 0, 49, 0, 55, 0, 0, 0, 115, 0, 99, 0, 49, 0, 56, 0, 0, 0, 115, 0, 99, 0, 49, 0, 57, 0, 0, 0, 115, 0, 99, 0, 50, 0, 48, 0, 0, 0, 115, 0, 99, 0, 50, 0, 49, 0, 0, 0, 115, 0, 99, 0, 50, 0, 50, 0, 0, 0, 115, 0, 99, 0, 50, 0, 51, 0, 0, 0, 115, 0, 99, 0, 50, 0, 52, 0, 0, 0, 115, 0, 99, 0, 50, 0, 53, 0, 0, 0, 115, 0, 101, 0, 97, 0, 98, 0, 0, 0, 115, 0, 101, 0, 97, 0, 99, 0, 0, 0, 115, 0, 103, 0, 48, 0, 49, 0, 0, 0, 115, 0, 103, 0, 48, 0, 50, 0, 0, 0, 115, 0, 103, 0, 48, 0, 51, 0, 0, 0, 115, 0, 103, 0, 48, 0, 52, 0, 0, 0, 115, 0, 103, 0, 48, 0, 53, 0, 0, 0, 115, 0, 109, 0, 48, 0, 49, 0, 0, 0, 115, 0, 109, 0, 48, 0, 50, 0, 0, 0, 115, 0, 109, 0, 48, 0, 51, 0, 0, 0, 115, 0, 109, 0, 48, 0, 52, 0, 0, 0, 115, 0, 109, 0, 48, 0, 53, 0, 0, 0, 115, 0, 109, 0, 48, 0, 54, 0, 0, 0, 115, 0, 109, 0, 48, 0, 55, 0, 0, 0, 115, 0, 109, 0, 48, 0, 56, 0, 0, 0, 115, 0, 109, 0, 48, 0, 57, 0, 0, 0, 115, 0, 110, 0, 107, 0, 100, 0, 0, 0, 115, 0, 110, 0, 107, 0, 101, 0, 0, 0, 115, 0, 121, 0, 104, 0, 109, 0, 0, 0, 116, 0, 104, 0, 49, 0, 48, 0, 0, 0, 116, 0, 104, 0, 49, 0, 49, 0, 0, 0, 116, 0, 104, 0, 49, 0, 50, 0, 0, 0, 116, 0, 104, 0, 49, 0, 51, 0, 0, 0, 116, 0, 104, 0, 49, 0, 52, 0, 0, 0, 116, 0, 104, 0, 49, 0, 53, 0, 0, 0, 116, 0, 104, 0, 49, 0, 54, 0, 0, 0, 116, 0, 104, 0, 49, 0, 55, 0, 0, 0, 116, 0, 104, 0, 49, 0, 56, 0, 0, 0, 116, 0, 104, 0, 49, 0, 57, 0, 0, 0, 116, 0, 104, 0, 50, 0, 48, 0, 0, 0, 116, 0, 104, 0, 50, 0, 49, 0, 0, 0, 116, 0, 104, 0, 50, 0, 50, 0, 0, 0, 116, 0, 104, 0, 50, 0, 51, 0, 0, 0, 116, 0, 104, 0, 50, 0, 52, 0, 0, 0, 116, 0, 104, 0, 50, 0, 53, 0, 0, 0, 116, 0, 104, 0, 50, 0, 54, 0, 0, 0, 116, 0, 104, 0, 50, 0, 55, 0, 0, 0, 116, 0, 104, 0, 51, 0, 48, 0, 0, 0, 116, 0, 104, 0, 51, 0, 49, 0, 0, 0, 116, 0, 104, 0, 51, 0, 50, 0, 0, 0, 116, 0, 104, 0, 51, 0, 51, 0, 0, 0, 116, 0, 104, 0, 51, 0, 52, 0, 0, 0, 116, 0, 104, 0, 51, 0, 53, 0, 0, 0, 116, 0, 104, 0, 51, 0, 54, 0, 0, 0, 116, 0, 104, 0, 51, 0, 55, 0, 0, 0, 116, 0, 104, 0, 51, 0, 56, 0, 0, 0, 116, 0, 104, 0, 51, 0, 57, 0, 0, 0, 116, 0, 104, 0, 52, 0, 48, 0, 0, 0, 116, 0, 104, 0, 52, 0, 49, 0, 0, 0, 116, 0, 104, 0, 52, 0, 50, 0, 0, 0, 116, 0, 104, 0, 52, 0, 51, 0, 0, 0, 116, 0, 104, 0, 52, 0, 52, 0, 0, 0, 116, 0, 104, 0, 52, 0, 53, 0, 0, 0, 116, 0, 104, 0, 52, 0, 54, 0, 0, 0, 116, 0, 104, 0, 52, 0, 55, 0, 0, 0, 116, 0, 104, 0, 52, 0, 56, 0, 0, 0, 116, 0, 104, 0, 52, 0, 57, 0, 0, 0, 116, 0, 104, 0, 53, 0, 48, 0, 0, 0, 116, 0, 104, 0, 53, 0, 49, 0, 0, 0, 116, 0, 104, 0, 53, 0, 50, 0, 0, 0, 116, 0, 104, 0, 53, 0, 51, 0, 0, 0, 116, 0, 104, 0, 53, 0, 52, 0, 0, 0, 116, 0, 104, 0, 53, 0, 53, 0, 0, 0, 116, 0, 104, 0, 53, 0, 54, 0, 0, 0, 116, 0, 104, 0, 53, 0, 55, 0, 0, 0, 116, 0, 104, 0, 53, 0, 56, 0, 0, 0, 116, 0, 104, 0, 54, 0, 48, 0, 0, 0, 116, 0, 104, 0, 54, 0, 49, 0, 0, 0, 116, 0, 104, 0, 54, 0, 50, 0, 0, 0, 116, 0, 104, 0, 54, 0, 51, 0, 0, 0, 116, 0, 104, 0, 54, 0, 52, 0, 0, 0, 116, 0, 104, 0, 54, 0, 53, 0, 0, 0, 116, 0, 104, 0, 54, 0, 54, 0, 0, 0, 116, 0, 104, 0, 54, 0, 55, 0, 0, 0, 116, 0, 104, 0, 55, 0, 48, 0, 0, 0, 116, 0, 104, 0, 55, 0, 49, 0, 0, 0, 116, 0, 104, 0, 55, 0, 50, 0, 0, 0, 116, 0, 104, 0, 55, 0, 51, 0, 0, 0, 116, 0, 104, 0, 55, 0, 52, 0, 0, 0, 116, 0, 104, 0, 55, 0, 53, 0, 0, 0, 116, 0, 104, 0, 55, 0, 54, 0, 0, 0, 116, 0, 104, 0, 55, 0, 55, 0, 0, 0, 116, 0, 104, 0, 56, 0, 48, 0, 0, 0, 116, 0, 104, 0, 56, 0, 49, 0, 0, 0, 116, 0, 104, 0, 56, 0, 50, 0, 0, 0, 116, 0, 104, 0, 56, 0, 51, 0, 0, 0, 116, 0, 104, 0, 56, 0, 52, 0, 0, 0, 116, 0, 104, 0, 56, 0, 53, 0, 0, 0, 116, 0, 104, 0, 56, 0, 54, 0, 0, 0, 116, 0, 104, 0, 57, 0, 48, 0, 0, 0, 116, 0, 104, 0, 57, 0, 49, 0, 0, 0, 116, 0, 104, 0, 57, 0, 50, 0, 0, 0, 116, 0, 104, 0, 57, 0, 51, 0, 0, 0, 116, 0, 104, 0, 57, 0, 52, 0, 0, 0, 116, 0, 104, 0, 57, 0, 53, 0, 0, 0, 116, 0, 110, 0, 49, 0, 49, 0, 0, 0, 116, 0, 110, 0, 49, 0, 50, 0, 0, 0, 116, 0, 110, 0, 49, 0, 51, 0, 0, 0, 116, 0, 110, 0, 49, 0, 52, 0, 0, 0, 116, 0, 110, 0, 50, 0, 49, 0, 0, 0, 116, 0, 110, 0, 50, 0, 50, 0, 0, 0, 116, 0, 110, 0, 50, 0, 51, 0, 0, 0, 116, 0, 110, 0, 51, 0, 49, 0, 0, 0, 116, 0, 110, 0, 51, 0, 50, 0, 0, 0, 116, 0, 110, 0, 51, 0, 51, 0, 0, 0, 116, 0, 110, 0, 51, 0, 52, 0, 0, 0, 116, 0, 110, 0, 52, 0, 49, 0, 0, 0, 116, 0, 110, 0, 52, 0, 50, 0, 0, 0, 116, 0, 110, 0, 52, 0, 51, 0, 0, 0, 116, 0, 110, 0, 53, 0, 49, 0, 0, 0, 116, 0, 110, 0, 53, 0, 50, 0, 0, 0, 116, 0, 110, 0, 53, 0, 51, 0, 0, 0, 116, 0, 110, 0, 55, 0, 49, 0, 0, 0, 116, 0, 110, 0, 55, 0, 50, 0, 0, 0, 116, 0, 110, 0, 55, 0, 51, 0, 0, 0, 116, 0, 110, 0, 56, 0, 49, 0, 0, 0, 116, 0, 110, 0, 56, 0, 50, 0, 0, 0, 116, 0, 110, 0, 56, 0, 51, 0, 0, 0, 116, 0, 111, 0, 48, 0, 49, 0, 0, 0, 116, 0, 111, 0, 48, 0, 50, 0, 0, 0, 116, 0, 111, 0, 48, 0, 51, 0, 0, 0, 116, 0, 111, 0, 48, 0, 52, 0, 0, 0, 116, 0, 111, 0, 48, 0, 53, 0, 0, 0, 116, 0, 114, 0, 48, 0, 49, 0, 0, 0, 116, 0, 114, 0, 48, 0, 50, 0, 0, 0, 116, 0, 114, 0, 48, 0, 51, 0, 0, 0, 116, 0, 114, 0, 48, 0, 52, 0, 0, 0, 116, 0, 114, 0, 48, 0, 53, 0, 0, 0, 116, 0, 114, 0, 48, 0, 54, 0, 0, 0, 116, 0, 114, 0, 48, 0, 55, 0, 0, 0, 116, 0, 114, 0, 48, 0, 56, 0, 0, 0, 116, 0, 114, 0, 48, 0, 57, 0, 0, 0, 116, 0, 114, 0, 49, 0, 48, 0, 0, 0, 116, 0, 114, 0, 49, 0, 49, 0, 0, 0, 116, 0, 114, 0, 49, 0, 50, 0, 0, 0, 116, 0, 114, 0, 49, 0, 51, 0, 0, 0, 116, 0, 114, 0, 49, 0, 52, 0, 0, 0, 116, 0, 114, 0, 49, 0, 53, 0, 0, 0, 116, 0, 114, 0, 49, 0, 54, 0, 0, 0, 116, 0, 114, 0, 49, 0, 55, 0, 0, 0, 116, 0, 114, 0, 49, 0, 56, 0, 0, 0, 116, 0, 114, 0, 49, 0, 57, 0, 0, 0, 116, 0, 114, 0, 50, 0, 48, 0, 0, 0, 116, 0, 114, 0, 50, 0, 49, 0, 0, 0, 116, 0, 114, 0, 50, 0, 50, 0, 0, 0, 116, 0, 114, 0, 50, 0, 51, 0, 0, 0, 116, 0, 114, 0, 50, 0, 52, 0, 0, 0, 116, 0, 114, 0, 50, 0, 53, 0, 0, 0, 116, 0, 114, 0, 50, 0, 54, 0, 0, 0, 116, 0, 114, 0, 50, 0, 55, 0, 0, 0, 116, 0, 114, 0, 50, 0, 56, 0, 0, 0, 116, 0, 114, 0, 50, 0, 57, 0, 0, 0, 116, 0, 114, 0, 51, 0, 48, 0, 0, 0, 116, 0, 114, 0, 51, 0, 49, 0, 0, 0, 116, 0, 114, 0, 51, 0, 50, 0, 0, 0, 116, 0, 114, 0, 51, 0, 51, 0, 0, 0, 116, 0, 114, 0, 51, 0, 52, 0, 0, 0, 116, 0, 114, 0, 51, 0, 53, 0, 0, 0, 116, 0, 114, 0, 51, 0, 54, 0, 0, 0, 116, 0, 114, 0, 51, 0, 55, 0, 0, 0, 116, 0, 114, 0, 51, 0, 56, 0, 0, 0, 116, 0, 114, 0, 51, 0, 57, 0, 0, 0, 116, 0, 114, 0, 52, 0, 48, 0, 0, 0, 116, 0, 114, 0, 52, 0, 49, 0, 0, 0, 116, 0, 114, 0, 52, 0, 50, 0, 0, 0, 116, 0, 114, 0, 52, 0, 51, 0, 0, 0, 116, 0, 114, 0, 52, 0, 52, 0, 0, 0, 116, 0, 114, 0, 52, 0, 53, 0, 0, 0, 116, 0, 114, 0, 52, 0, 54, 0, 0, 0, 116, 0, 114, 0, 52, 0, 55, 0, 0, 0, 116, 0, 114, 0, 52, 0, 56, 0, 0, 0, 116, 0, 114, 0, 52, 0, 57, 0, 0, 0, 116, 0, 114, 0, 53, 0, 48, 0, 0, 0, 116, 0, 114, 0, 53, 0, 49, 0, 0, 0, 116, 0, 114, 0, 53, 0, 50, 0, 0, 0, 116, 0, 114, 0, 53, 0, 51, 0, 0, 0, 116, 0, 114, 0, 53, 0, 52, 0, 0, 0, 116, 0, 114, 0, 53, 0, 53, 0, 0, 0, 116, 0, 114, 0, 53, 0, 54, 0, 0, 0, 116, 0, 114, 0, 53, 0, 55, 0, 0, 0, 116, 0, 114, 0, 53, 0, 56, 0, 0, 0, 116, 0, 114, 0, 53, 0, 57, 0, 0, 0, 116, 0, 114, 0, 54, 0, 48, 0, 0, 0, 116, 0, 114, 0, 54, 0, 49, 0, 0, 0, 116, 0, 114, 0, 54, 0, 50, 0, 0, 0, 116, 0, 114, 0, 54, 0, 51, 0, 0, 0, 116, 0, 114, 0, 54, 0, 52, 0, 0, 0, 116, 0, 114, 0, 54, 0, 53, 0, 0, 0, 116, 0, 114, 0, 54, 0, 54, 0, 0, 0, 116, 0, 114, 0, 54, 0, 55, 0, 0, 0, 116, 0, 114, 0, 54, 0, 56, 0, 0, 0, 116, 0, 114, 0, 54, 0, 57, 0, 0, 0, 116, 0, 114, 0, 55, 0, 48, 0, 0, 0, 116, 0, 114, 0, 55, 0, 49, 0, 0, 0, 116, 0, 114, 0, 55, 0, 50, 0, 0, 0, 116, 0, 114, 0, 55, 0, 51, 0, 0, 0, 116, 0, 114, 0, 55, 0, 52, 0, 0, 0, 116, 0, 114, 0, 55, 0, 53, 0, 0, 0, 116, 0, 114, 0, 55, 0, 54, 0, 0, 0, 116, 0, 114, 0, 55, 0, 55, 0, 0, 0, 116, 0, 114, 0, 55, 0, 56, 0, 0, 0, 116, 0, 114, 0, 55, 0, 57, 0, 0, 0, 116, 0, 114, 0, 56, 0, 48, 0, 0, 0, 116, 0, 122, 0, 48, 0, 49, 0, 0, 0, 116, 0, 122, 0, 48, 0, 50, 0, 0, 0, 116, 0, 122, 0, 48, 0, 51, 0, 0, 0, 116, 0, 122, 0, 48, 0, 52, 0, 0, 0, 116, 0, 122, 0, 48, 0, 53, 0, 0, 0, 116, 0, 122, 0, 48, 0, 54, 0, 0, 0, 116, 0, 122, 0, 48, 0, 55, 0, 0, 0, 116, 0, 122, 0, 48, 0, 56, 0, 0, 0, 116, 0, 122, 0, 48, 0, 57, 0, 0, 0, 116, 0, 122, 0, 49, 0, 49, 0, 0, 0, 116, 0, 122, 0, 49, 0, 51, 0, 0, 0, 116, 0, 122, 0, 49, 0, 52, 0, 0, 0, 116, 0, 122, 0, 49, 0, 54, 0, 0, 0, 116, 0, 122, 0, 49, 0, 55, 0, 0, 0, 116, 0, 122, 0, 49, 0, 56, 0, 0, 0, 116, 0, 122, 0, 50, 0, 49, 0, 0, 0, 116, 0, 122, 0, 50, 0, 50, 0, 0, 0, 116, 0, 122, 0, 50, 0, 51, 0, 0, 0, 116, 0, 122, 0, 50, 0, 52, 0, 0, 0, 116, 0, 122, 0, 50, 0, 53, 0, 0, 0, 116, 0, 122, 0, 50, 0, 54, 0, 0, 0, 116, 0, 122, 0, 50, 0, 55, 0, 0, 0, 116, 0, 122, 0, 50, 0, 56, 0, 0, 0, 116, 0, 122, 0, 50, 0, 57, 0, 0, 0, 116, 0, 122, 0, 51, 0, 48, 0, 0, 0, 116, 0, 122, 0, 51, 0, 49, 0, 0, 0, 117, 0, 115, 0, 97, 0, 108, 0, 0, 0, 117, 0, 115, 0, 109, 0, 100, 0, 0, 0, 117, 0, 115, 0, 109, 0, 101, 0, 0, 0, 117, 0, 115, 0, 109, 0, 110, 0, 0, 0, 117, 0, 115, 0, 109, 0, 115, 0, 0, 0, 117, 0, 115, 0, 109, 0, 116, 0, 0, 0, 117, 0, 115, 0, 110, 0, 99, 0, 0, 0, 117, 0, 115, 0, 110, 0, 100, 0, 0, 0, 117, 0, 115, 0, 110, 0, 101, 0, 0, 0, 117, 0, 115, 0, 115, 0, 99, 0, 0, 0, 117, 0, 115, 0, 115, 0, 100, 0, 0, 0, 118, 0, 99, 0, 48, 0, 49, 0, 0, 0, 118, 0, 99, 0, 48, 0, 50, 0, 0, 0, 118, 0, 99, 0, 48, 0, 51, 0, 0, 0, 118, 0, 99, 0, 48, 0, 52, 0, 0, 0, 118, 0, 99, 0, 48, 0, 53, 0, 0, 0, 118, 0, 99, 0, 48, 0, 54, 0, 0, 0, 118, 0, 110, 0, 48, 0, 49, 0, 0, 0, 118, 0, 110, 0, 48, 0, 50, 0, 0, 0, 118, 0, 110, 0, 48, 0, 51, 0, 0, 0, 118, 0, 110, 0, 48, 0, 52, 0, 0, 0, 118, 0, 110, 0, 48, 0, 53, 0, 0, 0, 118, 0, 110, 0, 48, 0, 54, 0, 0, 0, 118, 0, 110, 0, 48, 0, 55, 0, 0, 0, 118, 0, 110, 0, 49, 0, 51, 0, 0, 0, 118, 0, 110, 0, 49, 0, 52, 0, 0, 0, 118, 0, 110, 0, 50, 0, 48, 0, 0, 0, 118, 0, 110, 0, 50, 0, 49, 0, 0, 0, 118, 0, 110, 0, 50, 0, 50, 0, 0, 0, 118, 0, 110, 0, 50, 0, 51, 0, 0, 0, 118, 0, 110, 0, 50, 0, 52, 0, 0, 0, 118, 0, 110, 0, 50, 0, 53, 0, 0, 0, 118, 0, 110, 0, 50, 0, 54, 0, 0, 0, 118, 0, 110, 0, 50, 0, 55, 0, 0, 0, 118, 0, 110, 0, 50, 0, 56, 0, 0, 0, 118, 0, 110, 0, 50, 0, 57, 0, 0, 0, 118, 0, 110, 0, 51, 0, 48, 0, 0, 0, 118, 0, 110, 0, 51, 0, 49, 0, 0, 0, 118, 0, 110, 0, 51, 0, 50, 0, 0, 0, 118, 0, 110, 0, 51, 0, 51, 0, 0, 0, 118, 0, 110, 0, 51, 0, 52, 0, 0, 0, 118, 0, 110, 0, 51, 0, 53, 0, 0, 0, 118, 0, 110, 0, 51, 0, 54, 0, 0, 0, 118, 0, 110, 0, 51, 0, 55, 0, 0, 0, 118, 0, 110, 0, 52, 0, 48, 0, 0, 0, 118, 0, 110, 0, 52, 0, 49, 0, 0, 0, 118, 0, 110, 0, 52, 0, 51, 0, 0, 0, 118, 0, 110, 0, 52, 0, 52, 0, 0, 0, 118, 0, 110, 0, 52, 0, 53, 0, 0, 0, 118, 0, 110, 0, 52, 0, 54, 0, 0, 0, 118, 0, 110, 0, 52, 0, 55, 0, 0, 0, 118, 0, 110, 0, 53, 0, 48, 0, 0, 0, 118, 0, 110, 0, 53, 0, 49, 0, 0, 0, 118, 0, 110, 0, 53, 0, 50, 0, 0, 0, 118, 0, 110, 0, 53, 0, 51, 0, 0, 0, 118, 0, 110, 0, 53, 0, 52, 0, 0, 0, 118, 0, 110, 0, 53, 0, 53, 0, 0, 0, 118, 0, 110, 0, 53, 0, 54, 0, 0, 0, 118, 0, 110, 0, 53, 0, 55, 0, 0, 0, 118, 0, 110, 0, 53, 0, 56, 0, 0, 0, 118, 0, 110, 0, 53, 0, 57, 0, 0, 0, 118, 0, 110, 0, 54, 0, 54, 0, 0, 0, 118, 0, 110, 0, 54, 0, 55, 0, 0, 0, 118, 0, 110, 0, 54, 0, 56, 0, 0, 0, 118, 0, 110, 0, 54, 0, 57, 0, 0, 0, 118, 0, 110, 0, 55, 0, 48, 0, 0, 0, 118, 0, 110, 0, 55, 0, 49, 0, 0, 0, 118, 0, 110, 0, 55, 0, 50, 0, 0, 0, 118, 0, 110, 0, 55, 0, 51, 0, 0, 0, 122, 0, 109, 0, 48, 0, 49, 0, 0, 0, 122, 0, 109, 0, 48, 0, 50, 0, 0, 0, 122, 0, 109, 0, 48, 0, 51, 0, 0, 0, 122, 0, 109, 0, 48, 0, 52, 0, 0, 0, 122, 0, 109, 0, 48, 0, 53, 0, 0, 0, 122, 0, 109, 0, 48, 0, 54, 0, 0, 0, 122, 0, 109, 0, 48, 0, 55, 0, 0, 0, 122, 0, 109, 0, 48, 0, 56, 0, 0, 0, 122, 0, 109, 0, 48, 0, 57, 0, 0, 0, 122, 0, 119, 0, 109, 0, 118, 0, 0, 0, 122, 0, 119, 0, 109, 0, 119, 0, 0, 0, 101, 0, 110, 0, 95, 0, 73, 0, 78, 0, 0, 0, 97, 0, 122, 0, 122, 0, 97, 0, 114, 0, 0, 0, 101, 0, 110, 0, 95, 0, 68, 0, 69, 0, 0, 0, 101, 0, 110, 0, 95, 0, 80, 0, 72, 0, 0, 0, 101, 0, 110, 0, 95, 0, 90, 0, 65, 0, 0, 0, 101, 0, 110, 0, 95, 0, 80, 0, 75, 0, 0, 0, 101, 0, 110, 0, 95, 0, 85, 0, 71, 0, 0, 0, 101, 0, 110, 0, 95, 0, 71, 0, 66, 0, 0, 0, 101, 0, 110, 0, 95, 0, 75, 0, 69, 0, 0, 0, 101, 0, 110, 0, 95, 0, 78, 0, 71, 0, 0, 0, 48, 0, 53, 0, 51, 0, 126, 0, 52, 0, 0, 0, 101, 0, 110, 0, 95, 0, 78, 0, 76, 0, 0, 0, 112, 0, 116, 0, 95, 0, 70, 0, 82, 0, 0, 0, 101, 0, 110, 0, 95, 0, 65, 0, 84, 0, 0, 0, 101, 0, 110, 0, 95, 0, 66, 0, 69, 0, 0, 0, 112, 0, 116, 0, 95, 0, 77, 0, 90, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 73, 0, 0, 0, 101, 0, 110, 0, 95, 0, 84, 0, 90, 0, 0, 0, 110, 0, 122, 0, 119, 0, 103, 0, 110, 0, 0, 0, 112, 0, 116, 0, 95, 0, 67, 0, 72, 0, 0, 0, 112, 0, 116, 0, 95, 0, 76, 0, 85, 0, 0, 0, 112, 0, 116, 0, 95, 0, 80, 0, 84, 0, 0, 0, 101, 0, 110, 0, 95, 0, 67, 0, 65, 0, 0, 0, 101, 0, 110, 0, 95, 0, 67, 0, 89, 0, 0, 0, 101, 0, 110, 0, 95, 0, 68, 0, 75, 0, 0, 0, 101, 0, 110, 0, 95, 0, 70, 0, 73, 0, 0, 0, 101, 0, 110, 0, 95, 0, 71, 0, 72, 0, 0, 0, 101, 0, 110, 0, 95, 0, 77, 0, 84, 0, 0, 0, 101, 0, 110, 0, 95, 0, 78, 0, 65, 0, 0, 0, 108, 0, 117, 0, 110, 0, 97, 0, 114, 0, 0, 0, 116, 0, 119, 0, 104, 0, 115, 0, 113, 0, 0, 0, 101, 0, 110, 0, 95, 0, 69, 0, 82, 0, 0, 0, 101, 0, 110, 0, 95, 0, 70, 0, 74, 0, 0, 0, 101, 0, 110, 0, 95, 0, 70, 0, 77, 0, 0, 0, 101, 0, 110, 0, 95, 0, 73, 0, 69, 0, 0, 0, 101, 0, 110, 0, 95, 0, 77, 0, 85, 0, 0, 0, 101, 0, 110, 0, 95, 0, 77, 0, 89, 0, 0, 0, 101, 0, 110, 0, 95, 0, 82, 0, 87, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 68, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 69, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 71, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 76, 0, 0, 0, 101, 0, 110, 0, 95, 0, 86, 0, 85, 0, 0, 0, 101, 0, 110, 0, 95, 0, 90, 0, 77, 0, 0, 0, 101, 0, 110, 0, 95, 0, 90, 0, 87, 0, 0, 0, 112, 0, 116, 0, 95, 0, 65, 0, 79, 0, 0, 0, 112, 0, 116, 0, 95, 0, 71, 0, 81, 0, 0, 0, 116, 0, 119, 0, 116, 0, 110, 0, 110, 0, 0, 0, 42, 0, 95, 0, 42, 0, 95, 0, 42, 0, 0, 0, 97, 0, 111, 0, 108, 0, 117, 0, 97, 0, 0, 0, 101, 0, 110, 0, 95, 0, 67, 0, 77, 0, 0, 0, 101, 0, 110, 0, 95, 0, 73, 0, 76, 0, 0, 0, 101, 0, 110, 0, 95, 0, 74, 0, 77, 0, 0, 0, 101, 0, 110, 0, 95, 0, 77, 0, 71, 0, 0, 0, 101, 0, 110, 0, 95, 0, 77, 0, 87, 0, 0, 0, 101, 0, 110, 0, 95, 0, 78, 0, 82, 0, 0, 0, 101, 0, 110, 0, 95, 0, 78, 0, 90, 0, 0, 0, 101, 0, 110, 0, 95, 0, 80, 0, 71, 0, 0, 0, 101, 0, 110, 0, 95, 0, 80, 0, 87, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 67, 0, 0, 0, 101, 0, 110, 0, 95, 0, 87, 0, 83, 0, 0, 0, 101, 0, 115, 0, 95, 0, 66, 0, 79, 0, 0, 0, 101, 0, 115, 0, 95, 0, 69, 0, 67, 0, 0, 0, 101, 0, 115, 0, 95, 0, 80, 0, 89, 0, 0, 0, 103, 0, 98, 0, 101, 0, 115, 0, 115, 0, 0, 0, 103, 0, 98, 0, 121, 0, 111, 0, 114, 0, 0, 0, 103, 0, 98, 0, 122, 0, 101, 0, 116, 0, 0, 0, 107, 0, 122, 0, 121, 0, 117, 0, 122, 0, 0, 0, 109, 0, 97, 0, 98, 0, 101, 0, 109, 0, 0, 0, 109, 0, 97, 0, 116, 0, 97, 0, 122, 0, 0, 0, 112, 0, 116, 0, 95, 0, 66, 0, 82, 0, 0, 0, 103, 0, 98, 0, 115, 0, 103, 0, 99, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 57, 0, 0, 0, 109, 0, 97, 0, 109, 0, 97, 0, 114, 0, 0, 0, 112, 0, 104, 0, 116, 0, 97, 0, 114, 0, 0, 0, 114, 0, 117, 0, 115, 0, 97, 0, 114, 0, 0, 0, 114, 0, 117, 0, 121, 0, 97, 0, 114, 0, 0, 0, 97, 0, 102, 0, 116, 0, 97, 0, 107, 0, 0, 0, 97, 0, 111, 0, 117, 0, 105, 0, 103, 0, 0, 0, 98, 0, 101, 0, 119, 0, 104, 0, 116, 0, 0, 0, 101, 0, 110, 0, 95, 0, 66, 0, 87, 0, 0, 0, 101, 0, 110, 0, 95, 0, 68, 0, 77, 0, 0, 0, 101, 0, 110, 0, 95, 0, 71, 0, 77, 0, 0, 0, 101, 0, 110, 0, 95, 0, 71, 0, 89, 0, 0, 0, 101, 0, 110, 0, 95, 0, 75, 0, 73, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 83, 0, 0, 0, 101, 0, 110, 0, 95, 0, 78, 0, 85, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 83, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 90, 0, 0, 0, 101, 0, 110, 0, 95, 0, 84, 0, 79, 0, 0, 0, 101, 0, 110, 0, 95, 0, 84, 0, 84, 0, 0, 0, 101, 0, 110, 0, 95, 0, 84, 0, 86, 0, 0, 0, 101, 0, 110, 0, 95, 0, 86, 0, 67, 0, 0, 0, 101, 0, 115, 0, 95, 0, 65, 0, 82, 0, 0, 0, 101, 0, 115, 0, 95, 0, 66, 0, 90, 0, 0, 0, 101, 0, 115, 0, 95, 0, 67, 0, 76, 0, 0, 0, 101, 0, 115, 0, 95, 0, 67, 0, 82, 0, 0, 0, 101, 0, 115, 0, 95, 0, 68, 0, 79, 0, 0, 0, 101, 0, 115, 0, 95, 0, 71, 0, 84, 0, 0, 0, 101, 0, 115, 0, 95, 0, 77, 0, 88, 0, 0, 0, 101, 0, 115, 0, 95, 0, 78, 0, 73, 0, 0, 0, 101, 0, 115, 0, 95, 0, 80, 0, 65, 0, 0, 0, 101, 0, 115, 0, 95, 0, 80, 0, 69, 0, 0, 0, 101, 0, 115, 0, 95, 0, 83, 0, 86, 0, 0, 0, 101, 0, 115, 0, 95, 0, 86, 0, 69, 0, 0, 0, 103, 0, 98, 0, 118, 0, 103, 0, 108, 0, 0, 0, 103, 0, 98, 0, 119, 0, 108, 0, 110, 0, 0, 0, 107, 0, 122, 0, 115, 0, 104, 0, 121, 0, 0, 0, 107, 0, 122, 0, 122, 0, 104, 0, 97, 0, 0, 0, 109, 0, 104, 0, 110, 0, 109, 0, 107, 0, 0, 0, 109, 0, 104, 0, 117, 0, 116, 0, 105, 0, 0, 0, 112, 0, 116, 0, 95, 0, 67, 0, 86, 0, 0, 0, 112, 0, 116, 0, 95, 0, 71, 0, 87, 0, 0, 0, 112, 0, 116, 0, 95, 0, 84, 0, 76, 0, 0, 0, 114, 0, 117, 0, 122, 0, 97, 0, 98, 0, 0, 0, 116, 0, 116, 0, 115, 0, 102, 0, 111, 0, 0, 0, 99, 0, 111, 0, 98, 0, 111, 0, 108, 0, 0, 0, 103, 0, 98, 0, 103, 0, 119, 0, 110, 0, 0, 0, 109, 0, 97, 0, 101, 0, 115, 0, 109, 0, 0, 0, 109, 0, 104, 0, 107, 0, 105, 0, 108, 0, 0, 0, 109, 0, 120, 0, 110, 0, 108, 0, 101, 0, 0, 0, 109, 0, 120, 0, 118, 0, 101, 0, 114, 0, 0, 0, 112, 0, 101, 0, 116, 0, 117, 0, 109, 0, 0, 0, 112, 0, 104, 0, 98, 0, 111, 0, 104, 0, 0, 0, 112, 0, 104, 0, 99, 0, 101, 0, 98, 0, 0, 0, 112, 0, 104, 0, 115, 0, 108, 0, 101, 0, 0, 0, 112, 0, 119, 0, 50, 0, 49, 0, 52, 0, 0, 0, 49, 0, 53, 0, 52, 0, 126, 0, 53, 0, 0, 0, 101, 0, 110, 0, 95, 0, 65, 0, 71, 0, 0, 0, 101, 0, 110, 0, 95, 0, 65, 0, 73, 0, 0, 0, 101, 0, 110, 0, 95, 0, 65, 0, 85, 0, 0, 0, 101, 0, 110, 0, 95, 0, 66, 0, 66, 0, 0, 0, 101, 0, 110, 0, 95, 0, 66, 0, 83, 0, 0, 0, 101, 0, 110, 0, 95, 0, 67, 0, 67, 0, 0, 0, 101, 0, 110, 0, 95, 0, 71, 0, 68, 0, 0, 0, 101, 0, 110, 0, 95, 0, 71, 0, 73, 0, 0, 0, 101, 0, 110, 0, 95, 0, 73, 0, 77, 0, 0, 0, 101, 0, 110, 0, 95, 0, 74, 0, 69, 0, 0, 0, 101, 0, 110, 0, 95, 0, 75, 0, 78, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 67, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 82, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 66, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 88, 0, 0, 0, 101, 0, 110, 0, 95, 0, 84, 0, 75, 0, 0, 0, 101, 0, 110, 0, 95, 0, 86, 0, 71, 0, 0, 0, 101, 0, 115, 0, 95, 0, 67, 0, 79, 0, 0, 0, 101, 0, 115, 0, 95, 0, 67, 0, 85, 0, 0, 0, 101, 0, 115, 0, 95, 0, 72, 0, 78, 0, 0, 0, 101, 0, 115, 0, 95, 0, 80, 0, 82, 0, 0, 0, 101, 0, 115, 0, 95, 0, 85, 0, 89, 0, 0, 0, 103, 0, 98, 0, 119, 0, 108, 0, 118, 0, 0, 0, 108, 0, 118, 0, 48, 0, 48, 0, 53, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 49, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 51, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 49, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 52, 0, 0, 0, 112, 0, 101, 0, 112, 0, 105, 0, 117, 0, 0, 0, 112, 0, 115, 0, 114, 0, 98, 0, 104, 0, 0, 0, 112, 0, 116, 0, 95, 0, 77, 0, 79, 0, 0, 0, 112, 0, 116, 0, 95, 0, 83, 0, 84, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 54, 0, 0, 0, 114, 0, 117, 0, 115, 0, 97, 0, 107, 0, 0, 0, 114, 0, 117, 0, 115, 0, 109, 0, 111, 0, 0, 0, 97, 0, 102, 0, 98, 0, 100, 0, 103, 0, 0, 0, 97, 0, 102, 0, 103, 0, 104, 0, 97, 0, 0, 0, 97, 0, 102, 0, 106, 0, 111, 0, 119, 0, 0, 0, 97, 0, 111, 0, 122, 0, 97, 0, 105, 0, 0, 0, 97, 0, 122, 0, 98, 0, 97, 0, 108, 0, 0, 0, 97, 0, 122, 0, 105, 0, 109, 0, 105, 0, 0, 0, 97, 0, 122, 0, 111, 0, 103, 0, 117, 0, 0, 0, 97, 0, 122, 0, 113, 0, 117, 0, 115, 0, 0, 0, 98, 0, 101, 0, 119, 0, 108, 0, 103, 0, 0, 0, 98, 0, 102, 0, 98, 0, 97, 0, 122, 0, 0, 0, 98, 0, 102, 0, 115, 0, 101, 0, 110, 0, 0, 0, 98, 0, 102, 0, 122, 0, 105, 0, 114, 0, 0, 0, 99, 0, 111, 0, 103, 0, 117, 0, 118, 0, 0, 0, 99, 0, 111, 0, 110, 0, 115, 0, 97, 0, 0, 0, 99, 0, 111, 0, 115, 0, 97, 0, 112, 0, 0, 0, 102, 0, 114, 0, 97, 0, 114, 0, 97, 0, 0, 0, 102, 0, 114, 0, 103, 0, 101, 0, 115, 0, 0, 0, 102, 0, 114, 0, 108, 0, 114, 0, 101, 0, 0, 0, 103, 0, 98, 0, 99, 0, 108, 0, 100, 0, 0, 0, 103, 0, 98, 0, 99, 0, 109, 0, 110, 0, 0, 0, 103, 0, 98, 0, 100, 0, 98, 0, 121, 0, 0, 0, 103, 0, 98, 0, 100, 0, 101, 0, 118, 0, 0, 0, 103, 0, 98, 0, 100, 0, 117, 0, 100, 0, 0, 0, 103, 0, 98, 0, 101, 0, 114, 0, 119, 0, 0, 0, 103, 0, 98, 0, 102, 0, 108, 0, 110, 0, 0, 0, 103, 0, 98, 0, 103, 0, 114, 0, 101, 0, 0, 0, 103, 0, 98, 0, 104, 0, 110, 0, 115, 0, 0, 0, 103, 0, 98, 0, 105, 0, 111, 0, 119, 0, 0, 0, 103, 0, 98, 0, 107, 0, 104, 0, 108, 0, 0, 0, 103, 0, 98, 0, 108, 0, 99, 0, 101, 0, 0, 0, 103, 0, 98, 0, 108, 0, 110, 0, 100, 0, 0, 0, 103, 0, 98, 0, 109, 0, 117, 0, 108, 0, 0, 0, 103, 0, 98, 0, 110, 0, 119, 0, 109, 0, 0, 0, 103, 0, 98, 0, 114, 0, 100, 0, 98, 0, 0, 0, 103, 0, 98, 0, 115, 0, 116, 0, 121, 0, 0, 0, 103, 0, 98, 0, 119, 0, 100, 0, 117, 0, 0, 0, 103, 0, 98, 0, 119, 0, 111, 0, 107, 0, 0, 0, 107, 0, 122, 0, 97, 0, 108, 0, 97, 0, 0, 0, 108, 0, 118, 0, 108, 0, 112, 0, 120, 0, 0, 0, 108, 0, 118, 0, 118, 0, 101, 0, 110, 0, 0, 0, 109, 0, 97, 0, 99, 0, 104, 0, 116, 0, 0, 0, 109, 0, 97, 0, 105, 0, 110, 0, 101, 0, 0, 0, 109, 0, 97, 0, 110, 0, 97, 0, 100, 0, 0, 0, 109, 0, 97, 0, 115, 0, 101, 0, 102, 0, 0, 0, 109, 0, 97, 0, 115, 0, 101, 0, 116, 0, 0, 0, 109, 0, 104, 0, 97, 0, 114, 0, 110, 0, 0, 0, 109, 0, 104, 0, 101, 0, 98, 0, 111, 0, 0, 0, 109, 0, 120, 0, 99, 0, 111, 0, 97, 0, 0, 0, 109, 0, 120, 0, 103, 0, 114, 0, 111, 0, 0, 0, 109, 0, 120, 0, 104, 0, 105, 0, 100, 0, 0, 0, 110, 0, 122, 0, 98, 0, 111, 0, 112, 0, 0, 0, 112, 0, 103, 0, 101, 0, 98, 0, 114, 0, 0, 0, 112, 0, 104, 0, 100, 0, 97, 0, 111, 0, 0, 0, 112, 0, 104, 0, 103, 0, 117, 0, 105, 0, 0, 0, 112, 0, 104, 0, 105, 0, 102, 0, 117, 0, 0, 0, 112, 0, 104, 0, 105, 0, 108, 0, 105, 0, 0, 0, 112, 0, 104, 0, 105, 0, 108, 0, 115, 0, 0, 0, 112, 0, 104, 0, 105, 0, 115, 0, 97, 0, 0, 0, 112, 0, 115, 0, 103, 0, 122, 0, 97, 0, 0, 0, 112, 0, 119, 0, 48, 0, 48, 0, 50, 0, 0, 0, 112, 0, 119, 0, 48, 0, 48, 0, 52, 0, 0, 0, 112, 0, 119, 0, 48, 0, 49, 0, 48, 0, 0, 0, 112, 0, 119, 0, 48, 0, 53, 0, 48, 0, 0, 0, 112, 0, 119, 0, 49, 0, 48, 0, 48, 0, 0, 0, 112, 0, 119, 0, 49, 0, 53, 0, 48, 0, 0, 0, 112, 0, 119, 0, 50, 0, 49, 0, 50, 0, 0, 0, 112, 0, 119, 0, 50, 0, 49, 0, 56, 0, 0, 0, 112, 0, 119, 0, 50, 0, 50, 0, 50, 0, 0, 0, 114, 0, 117, 0, 97, 0, 114, 0, 107, 0, 0, 0, 114, 0, 117, 0, 105, 0, 114, 0, 107, 0, 0, 0, 114, 0, 117, 0, 107, 0, 97, 0, 109, 0, 0, 0, 114, 0, 117, 0, 107, 0, 100, 0, 97, 0, 0, 0, 114, 0, 117, 0, 107, 0, 103, 0, 110, 0, 0, 0, 114, 0, 117, 0, 111, 0, 114, 0, 101, 0, 0, 0, 114, 0, 117, 0, 115, 0, 118, 0, 101, 0, 0, 0, 114, 0, 117, 0, 118, 0, 111, 0, 114, 0, 0, 0, 116, 0, 119, 0, 110, 0, 97, 0, 110, 0, 0, 0, 118, 0, 117, 0, 115, 0, 101, 0, 101, 0, 0, 0, 97, 0, 122, 0, 122, 0, 97, 0, 110, 0, 0, 0, 101, 0, 110, 0, 95, 0, 66, 0, 77, 0, 0, 0, 101, 0, 110, 0, 95, 0, 67, 0, 75, 0, 0, 0, 101, 0, 110, 0, 95, 0, 67, 0, 88, 0, 0, 0, 101, 0, 110, 0, 95, 0, 68, 0, 71, 0, 0, 0, 101, 0, 110, 0, 95, 0, 70, 0, 75, 0, 0, 0, 101, 0, 110, 0, 95, 0, 71, 0, 71, 0, 0, 0, 101, 0, 110, 0, 95, 0, 73, 0, 79, 0, 0, 0, 101, 0, 110, 0, 95, 0, 75, 0, 89, 0, 0, 0, 101, 0, 110, 0, 95, 0, 77, 0, 83, 0, 0, 0, 101, 0, 110, 0, 95, 0, 78, 0, 70, 0, 0, 0, 101, 0, 110, 0, 95, 0, 80, 0, 78, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 72, 0, 0, 0, 101, 0, 110, 0, 95, 0, 84, 0, 67, 0, 0, 0, 108, 0, 118, 0, 100, 0, 103, 0, 118, 0, 0, 0, 109, 0, 120, 0, 116, 0, 97, 0, 109, 0, 0, 0, 110, 0, 122, 0, 119, 0, 107, 0, 111, 0, 0, 0, 112, 0, 104, 0, 115, 0, 117, 0, 107, 0, 0, 0, 112, 0, 104, 0, 122, 0, 97, 0, 115, 0, 0, 0, 118, 0, 117, 0, 116, 0, 97, 0, 101, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 50, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 51, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 56, 0, 0, 0, 99, 0, 122, 0, 51, 0, 50, 0, 52, 0, 0, 0, 99, 0, 122, 0, 52, 0, 50, 0, 50, 0, 0, 0, 99, 0, 122, 0, 52, 0, 50, 0, 54, 0, 0, 0, 99, 0, 122, 0, 53, 0, 49, 0, 50, 0, 0, 0, 99, 0, 122, 0, 53, 0, 50, 0, 52, 0, 0, 0, 99, 0, 122, 0, 53, 0, 51, 0, 51, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 52, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 51, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 54, 0, 0, 0, 103, 0, 98, 0, 115, 0, 116, 0, 115, 0, 0, 0, 108, 0, 118, 0, 48, 0, 48, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 52, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 55, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 56, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 48, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 55, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 49, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 54, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 50, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 54, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 50, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 48, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 48, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 49, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 50, 0, 0, 0, 117, 0, 103, 0, 50, 0, 51, 0, 50, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 48, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 56, 0, 0, 0, 117, 0, 103, 0, 51, 0, 51, 0, 50, 0, 0, 0, 117, 0, 103, 0, 52, 0, 48, 0, 52, 0, 0, 0, 117, 0, 103, 0, 52, 0, 48, 0, 56, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 48, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 52, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 55, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 56, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 57, 0, 0, 0, 117, 0, 103, 0, 52, 0, 51, 0, 48, 0, 0, 0, 97, 0, 102, 0, 98, 0, 100, 0, 115, 0, 0, 0, 97, 0, 102, 0, 98, 0, 103, 0, 108, 0, 0, 0, 97, 0, 102, 0, 100, 0, 97, 0, 121, 0, 0, 0, 97, 0, 102, 0, 102, 0, 114, 0, 97, 0, 0, 0, 97, 0, 102, 0, 102, 0, 121, 0, 98, 0, 0, 0, 97, 0, 102, 0, 103, 0, 104, 0, 111, 0, 0, 0, 97, 0, 102, 0, 104, 0, 101, 0, 108, 0, 0, 0, 97, 0, 102, 0, 104, 0, 101, 0, 114, 0, 0, 0, 97, 0, 102, 0, 107, 0, 97, 0, 98, 0, 0, 0, 97, 0, 102, 0, 107, 0, 97, 0, 110, 0, 0, 0, 97, 0, 102, 0, 107, 0, 97, 0, 112, 0, 0, 0, 97, 0, 102, 0, 107, 0, 100, 0, 122, 0, 0, 0, 97, 0, 102, 0, 107, 0, 104, 0, 111, 0, 0, 0, 97, 0, 102, 0, 107, 0, 110, 0, 114, 0, 0, 0, 97, 0, 102, 0, 108, 0, 97, 0, 103, 0, 0, 0, 97, 0, 102, 0, 108, 0, 111, 0, 103, 0, 0, 0, 97, 0, 102, 0, 110, 0, 97, 0, 110, 0, 0, 0, 97, 0, 102, 0, 110, 0, 105, 0, 109, 0, 0, 0, 97, 0, 102, 0, 110, 0, 117, 0, 114, 0, 0, 0, 97, 0, 102, 0, 112, 0, 97, 0, 110, 0, 0, 0, 97, 0, 102, 0, 112, 0, 97, 0, 114, 0, 0, 0, 97, 0, 102, 0, 112, 0, 105, 0, 97, 0, 0, 0, 97, 0, 102, 0, 112, 0, 107, 0, 97, 0, 0, 0, 97, 0, 102, 0, 115, 0, 97, 0, 109, 0, 0, 0, 97, 0, 102, 0, 115, 0, 97, 0, 114, 0, 0, 0, 97, 0, 102, 0, 117, 0, 114, 0, 117, 0, 0, 0, 97, 0, 102, 0, 119, 0, 97, 0, 114, 0, 0, 0, 97, 0, 102, 0, 122, 0, 97, 0, 98, 0, 0, 0, 97, 0, 111, 0, 98, 0, 103, 0, 111, 0, 0, 0, 97, 0, 111, 0, 98, 0, 103, 0, 117, 0, 0, 0, 97, 0, 111, 0, 98, 0, 105, 0, 101, 0, 0, 0, 97, 0, 111, 0, 99, 0, 97, 0, 98, 0, 0, 0, 97, 0, 111, 0, 99, 0, 99, 0, 117, 0, 0, 0, 97, 0, 111, 0, 99, 0, 117, 0, 115, 0, 0, 0, 97, 0, 111, 0, 104, 0, 117, 0, 97, 0, 0, 0, 97, 0, 111, 0, 104, 0, 117, 0, 105, 0, 0, 0, 97, 0, 111, 0, 108, 0, 110, 0, 111, 0, 0, 0, 97, 0, 111, 0, 108, 0, 115, 0, 117, 0, 0, 0, 97, 0, 111, 0, 109, 0, 97, 0, 108, 0, 0, 0, 97, 0, 111, 0, 109, 0, 111, 0, 120, 0, 0, 0, 97, 0, 111, 0, 110, 0, 97, 0, 109, 0, 0, 0, 97, 0, 117, 0, 97, 0, 99, 0, 116, 0, 0, 0, 97, 0, 117, 0, 110, 0, 115, 0, 119, 0, 0, 0, 97, 0, 117, 0, 113, 0, 108, 0, 100, 0, 0, 0, 97, 0, 117, 0, 116, 0, 97, 0, 115, 0, 0, 0, 97, 0, 117, 0, 118, 0, 105, 0, 99, 0, 0, 0, 97, 0, 122, 0, 97, 0, 98, 0, 115, 0, 0, 0, 97, 0, 122, 0, 97, 0, 103, 0, 97, 0, 0, 0, 97, 0, 122, 0, 97, 0, 103, 0, 99, 0, 0, 0, 97, 0, 122, 0, 97, 0, 103, 0, 109, 0, 0, 0, 97, 0, 122, 0, 97, 0, 103, 0, 115, 0, 0, 0, 97, 0, 122, 0, 97, 0, 103, 0, 117, 0, 0, 0, 97, 0, 122, 0, 97, 0, 115, 0, 116, 0, 0, 0, 97, 0, 122, 0, 98, 0, 97, 0, 98, 0, 0, 0, 97, 0, 122, 0, 98, 0, 97, 0, 114, 0, 0, 0, 97, 0, 122, 0, 98, 0, 101, 0, 121, 0, 0, 0, 97, 0, 122, 0, 98, 0, 105, 0, 108, 0, 0, 0, 97, 0, 122, 0, 99, 0, 97, 0, 98, 0, 0, 0, 97, 0, 122, 0, 99, 0, 97, 0, 108, 0, 0, 0, 97, 0, 122, 0, 99, 0, 117, 0, 108, 0, 0, 0, 97, 0, 122, 0, 100, 0, 97, 0, 115, 0, 0, 0, 97, 0, 122, 0, 102, 0, 117, 0, 122, 0, 0, 0, 97, 0, 122, 0, 103, 0, 97, 0, 100, 0, 0, 0, 97, 0, 122, 0, 103, 0, 111, 0, 114, 0, 0, 0, 97, 0, 122, 0, 103, 0, 111, 0, 121, 0, 0, 0, 97, 0, 122, 0, 103, 0, 121, 0, 103, 0, 0, 0, 97, 0, 122, 0, 104, 0, 97, 0, 99, 0, 0, 0, 97, 0, 122, 0, 105, 0, 115, 0, 109, 0, 0, 0, 97, 0, 122, 0, 107, 0, 97, 0, 108, 0, 0, 0, 97, 0, 122, 0, 107, 0, 97, 0, 110, 0, 0, 0, 97, 0, 122, 0, 107, 0, 117, 0, 114, 0, 0, 0, 97, 0, 122, 0, 108, 0, 97, 0, 99, 0, 0, 0, 97, 0, 122, 0, 108, 0, 97, 0, 110, 0, 0, 0, 97, 0, 122, 0, 108, 0, 101, 0, 114, 0, 0, 0, 97, 0, 122, 0, 109, 0, 97, 0, 115, 0, 0, 0, 97, 0, 122, 0, 110, 0, 101, 0, 102, 0, 0, 0, 97, 0, 122, 0, 111, 0, 114, 0, 100, 0, 0, 0, 97, 0, 122, 0, 113, 0, 97, 0, 98, 0, 0, 0, 97, 0, 122, 0, 113, 0, 97, 0, 120, 0, 0, 0, 97, 0, 122, 0, 113, 0, 97, 0, 122, 0, 0, 0, 97, 0, 122, 0, 113, 0, 98, 0, 97, 0, 0, 0, 97, 0, 122, 0, 113, 0, 98, 0, 105, 0, 0, 0, 97, 0, 122, 0, 113, 0, 111, 0, 98, 0, 0, 0, 97, 0, 122, 0, 115, 0, 97, 0, 98, 0, 0, 0, 97, 0, 122, 0, 115, 0, 97, 0, 100, 0, 0, 0, 97, 0, 122, 0, 115, 0, 97, 0, 104, 0, 0, 0, 97, 0, 122, 0, 115, 0, 97, 0, 114, 0, 0, 0, 97, 0, 122, 0, 115, 0, 97, 0, 116, 0, 0, 0, 97, 0, 122, 0, 115, 0, 98, 0, 110, 0, 0, 0, 97, 0, 122, 0, 115, 0, 105, 0, 121, 0, 0, 0, 97, 0, 122, 0, 115, 0, 107, 0, 114, 0, 0, 0, 97, 0, 122, 0, 115, 0, 109, 0, 105, 0, 0, 0, 97, 0, 122, 0, 115, 0, 109, 0, 120, 0, 0, 0, 97, 0, 122, 0, 115, 0, 117, 0, 115, 0, 0, 0, 97, 0, 122, 0, 116, 0, 97, 0, 114, 0, 0, 0, 97, 0, 122, 0, 116, 0, 111, 0, 118, 0, 0, 0, 97, 0, 122, 0, 117, 0, 99, 0, 97, 0, 0, 0, 97, 0, 122, 0, 120, 0, 97, 0, 99, 0, 0, 0, 97, 0, 122, 0, 120, 0, 99, 0, 105, 0, 0, 0, 97, 0, 122, 0, 120, 0, 105, 0, 122, 0, 0, 0, 97, 0, 122, 0, 120, 0, 118, 0, 100, 0, 0, 0, 97, 0, 122, 0, 121, 0, 97, 0, 114, 0, 0, 0, 97, 0, 122, 0, 121, 0, 101, 0, 118, 0, 0, 0, 98, 0, 97, 0, 98, 0, 105, 0, 104, 0, 0, 0, 98, 0, 97, 0, 98, 0, 114, 0, 99, 0, 0, 0, 98, 0, 97, 0, 115, 0, 114, 0, 112, 0, 0, 0, 98, 0, 101, 0, 98, 0, 114, 0, 117, 0, 0, 0, 98, 0, 101, 0, 118, 0, 97, 0, 110, 0, 0, 0, 98, 0, 101, 0, 118, 0, 98, 0, 114, 0, 0, 0, 98, 0, 101, 0, 118, 0, 108, 0, 103, 0, 0, 0, 98, 0, 101, 0, 118, 0, 108, 0, 105, 0, 0, 0, 98, 0, 101, 0, 118, 0, 111, 0, 118, 0, 0, 0, 98, 0, 101, 0, 118, 0, 119, 0, 118, 0, 0, 0, 98, 0, 101, 0, 119, 0, 97, 0, 108, 0, 0, 0, 98, 0, 101, 0, 119, 0, 98, 0, 114, 0, 0, 0, 98, 0, 101, 0, 119, 0, 108, 0, 120, 0, 0, 0, 98, 0, 101, 0, 119, 0, 110, 0, 97, 0, 0, 0, 98, 0, 102, 0, 98, 0, 103, 0, 114, 0, 0, 0, 98, 0, 102, 0, 98, 0, 108, 0, 103, 0, 0, 0, 98, 0, 102, 0, 98, 0, 108, 0, 107, 0, 0, 0, 98, 0, 102, 0, 99, 0, 111, 0, 109, 0, 0, 0, 98, 0, 102, 0, 103, 0, 97, 0, 110, 0, 0, 0, 98, 0, 102, 0, 103, 0, 110, 0, 97, 0, 0, 0, 98, 0, 102, 0, 103, 0, 111, 0, 117, 0, 0, 0, 98, 0, 102, 0, 104, 0, 111, 0, 117, 0, 0, 0, 98, 0, 102, 0, 105, 0, 111, 0, 98, 0, 0, 0, 98, 0, 102, 0, 107, 0, 97, 0, 100, 0, 0, 0, 98, 0, 102, 0, 107, 0, 101, 0, 110, 0, 0, 0, 98, 0, 102, 0, 107, 0, 109, 0, 100, 0, 0, 0, 98, 0, 102, 0, 107, 0, 109, 0, 112, 0, 0, 0, 98, 0, 102, 0, 107, 0, 111, 0, 112, 0, 0, 0, 98, 0, 102, 0, 107, 0, 111, 0, 119, 0, 0, 0, 98, 0, 102, 0, 108, 0, 101, 0, 114, 0, 0, 0, 98, 0, 102, 0, 108, 0, 111, 0, 114, 0, 0, 0, 98, 0, 102, 0, 109, 0, 111, 0, 117, 0, 0, 0, 98, 0, 102, 0, 110, 0, 97, 0, 109, 0, 0, 0, 98, 0, 102, 0, 110, 0, 97, 0, 111, 0, 0, 0, 98, 0, 102, 0, 110, 0, 97, 0, 121, 0, 0, 0, 98, 0, 102, 0, 110, 0, 111, 0, 117, 0, 0, 0, 98, 0, 102, 0, 111, 0, 117, 0, 98, 0, 0, 0, 98, 0, 102, 0, 111, 0, 117, 0, 100, 0, 0, 0, 98, 0, 102, 0, 112, 0, 97, 0, 115, 0, 0, 0, 98, 0, 102, 0, 112, 0, 111, 0, 110, 0, 0, 0, 98, 0, 102, 0, 115, 0, 105, 0, 115, 0, 0, 0, 98, 0, 102, 0, 115, 0, 109, 0, 116, 0, 0, 0, 98, 0, 102, 0, 115, 0, 110, 0, 103, 0, 0, 0, 98, 0, 102, 0, 115, 0, 111, 0, 109, 0, 0, 0, 98, 0, 102, 0, 115, 0, 111, 0, 114, 0, 0, 0, 98, 0, 102, 0, 116, 0, 97, 0, 112, 0, 0, 0, 98, 0, 102, 0, 116, 0, 117, 0, 105, 0, 0, 0, 98, 0, 102, 0, 121, 0, 97, 0, 103, 0, 0, 0, 98, 0, 102, 0, 121, 0, 97, 0, 116, 0, 0, 0, 98, 0, 102, 0, 122, 0, 111, 0, 110, 0, 0, 0, 98, 0, 102, 0, 122, 0, 111, 0, 117, 0, 0, 0, 98, 0, 122, 0, 99, 0, 122, 0, 108, 0, 0, 0, 98, 0, 122, 0, 116, 0, 111, 0, 108, 0, 0, 0, 99, 0, 102, 0, 98, 0, 103, 0, 102, 0, 0, 0, 99, 0, 103, 0, 98, 0, 122, 0, 118, 0, 0, 0, 99, 0, 111, 0, 97, 0, 109, 0, 97, 0, 0, 0, 99, 0, 111, 0, 97, 0, 110, 0, 116, 0, 0, 0, 99, 0, 111, 0, 97, 0, 114, 0, 97, 0, 0, 0, 99, 0, 111, 0, 97, 0, 116, 0, 108, 0, 0, 0, 99, 0, 111, 0, 98, 0, 111, 0, 121, 0, 0, 0, 99, 0, 111, 0, 99, 0, 97, 0, 108, 0, 0, 0, 99, 0, 111, 0, 99, 0, 97, 0, 113, 0, 0, 0, 99, 0, 111, 0, 99, 0, 97, 0, 115, 0, 0, 0, 99, 0, 111, 0, 99, 0, 97, 0, 117, 0, 0, 0, 99, 0, 111, 0, 99, 0, 101, 0, 115, 0, 0, 0, 99, 0, 111, 0, 99, 0, 104, 0, 111, 0, 0, 0, 99, 0, 111, 0, 99, 0, 111, 0, 114, 0, 0, 0, 99, 0, 111, 0, 99, 0, 117, 0, 110, 0, 0, 0, 99, 0, 111, 0, 103, 0, 117, 0, 97, 0, 0, 0, 99, 0, 111, 0, 104, 0, 117, 0, 105, 0, 0, 0, 99, 0, 111, 0, 108, 0, 97, 0, 103, 0, 0, 0, 99, 0, 111, 0, 109, 0, 97, 0, 103, 0, 0, 0, 99, 0, 111, 0, 109, 0, 101, 0, 116, 0, 0, 0, 99, 0, 111, 0, 110, 0, 97, 0, 114, 0, 0, 0, 99, 0, 111, 0, 112, 0, 117, 0, 116, 0, 0, 0, 99, 0, 111, 0, 113, 0, 117, 0, 105, 0, 0, 0, 99, 0, 111, 0, 114, 0, 105, 0, 115, 0, 0, 0, 99, 0, 111, 0, 115, 0, 97, 0, 110, 0, 0, 0, 99, 0, 111, 0, 115, 0, 117, 0, 99, 0, 0, 0, 99, 0, 111, 0, 116, 0, 111, 0, 108, 0, 0, 0, 99, 0, 111, 0, 118, 0, 97, 0, 99, 0, 0, 0, 99, 0, 111, 0, 118, 0, 97, 0, 117, 0, 0, 0, 99, 0, 111, 0, 118, 0, 105, 0, 100, 0, 0, 0, 101, 0, 99, 0, 114, 0, 126, 0, 115, 0, 0, 0, 101, 0, 103, 0, 97, 0, 108, 0, 120, 0, 0, 0, 101, 0, 103, 0, 97, 0, 115, 0, 110, 0, 0, 0, 101, 0, 103, 0, 97, 0, 115, 0, 116, 0, 0, 0, 101, 0, 103, 0, 98, 0, 110, 0, 115, 0, 0, 0, 101, 0, 103, 0, 102, 0, 121, 0, 109, 0, 0, 0, 101, 0, 103, 0, 107, 0, 102, 0, 115, 0, 0, 0, 101, 0, 103, 0, 109, 0, 110, 0, 102, 0, 0, 0, 101, 0, 103, 0, 112, 0, 116, 0, 115, 0, 0, 0, 101, 0, 103, 0, 115, 0, 104, 0, 103, 0, 0, 0, 101, 0, 103, 0, 115, 0, 104, 0, 114, 0, 0, 0, 101, 0, 103, 0, 115, 0, 105, 0, 110, 0, 0, 0, 101, 0, 103, 0, 115, 0, 117, 0, 122, 0, 0, 0, 101, 0, 103, 0, 119, 0, 97, 0, 100, 0, 0, 0, 102, 0, 109, 0, 107, 0, 115, 0, 97, 0, 0, 0, 102, 0, 109, 0, 112, 0, 110, 0, 105, 0, 0, 0, 102, 0, 109, 0, 116, 0, 114, 0, 107, 0, 0, 0, 102, 0, 109, 0, 121, 0, 97, 0, 112, 0, 0, 0, 102, 0, 114, 0, 98, 0, 102, 0, 99, 0, 0, 0, 102, 0, 114, 0, 98, 0, 114, 0, 101, 0, 0, 0, 102, 0, 114, 0, 99, 0, 111, 0, 114, 0, 0, 0, 102, 0, 114, 0, 99, 0, 118, 0, 108, 0, 0, 0, 102, 0, 114, 0, 103, 0, 117, 0, 97, 0, 0, 0, 102, 0, 114, 0, 104, 0, 100, 0, 102, 0, 0, 0, 102, 0, 114, 0, 105, 0, 100, 0, 102, 0, 0, 0, 102, 0, 114, 0, 109, 0, 97, 0, 121, 0, 0, 0, 102, 0, 114, 0, 110, 0, 97, 0, 113, 0, 0, 0, 102, 0, 114, 0, 110, 0, 111, 0, 114, 0, 0, 0, 102, 0, 114, 0, 111, 0, 99, 0, 99, 0, 0, 0, 102, 0, 114, 0, 112, 0, 97, 0, 99, 0, 0, 0, 102, 0, 114, 0, 112, 0, 100, 0, 108, 0, 0, 0, 103, 0, 98, 0, 97, 0, 103, 0, 98, 0, 0, 0, 103, 0, 98, 0, 97, 0, 103, 0, 121, 0, 0, 0, 103, 0, 98, 0, 97, 0, 110, 0, 100, 0, 0, 0, 103, 0, 98, 0, 97, 0, 110, 0, 110, 0, 0, 0, 103, 0, 98, 0, 97, 0, 110, 0, 115, 0, 0, 0, 103, 0, 98, 0, 98, 0, 97, 0, 115, 0, 0, 0, 103, 0, 98, 0, 98, 0, 98, 0, 100, 0, 0, 0, 103, 0, 98, 0, 98, 0, 101, 0, 110, 0, 0, 0, 103, 0, 98, 0, 98, 0, 101, 0, 120, 0, 0, 0, 103, 0, 98, 0, 98, 0, 102, 0, 115, 0, 0, 0, 103, 0, 98, 0, 98, 0, 103, 0, 101, 0, 0, 0, 103, 0, 98, 0, 98, 0, 103, 0, 119, 0, 0, 0, 103, 0, 98, 0, 98, 0, 105, 0, 114, 0, 0, 0, 103, 0, 98, 0, 98, 0, 107, 0, 109, 0, 0, 0, 103, 0, 98, 0, 98, 0, 109, 0, 104, 0, 0, 0, 103, 0, 98, 0, 98, 0, 110, 0, 101, 0, 0, 0, 103, 0, 98, 0, 98, 0, 110, 0, 104, 0, 0, 0, 103, 0, 98, 0, 98, 0, 110, 0, 115, 0, 0, 0, 103, 0, 98, 0, 98, 0, 111, 0, 108, 0, 0, 0, 103, 0, 98, 0, 98, 0, 112, 0, 108, 0, 0, 0, 103, 0, 98, 0, 98, 0, 114, 0, 121, 0, 0, 0, 103, 0, 98, 0, 98, 0, 115, 0, 116, 0, 0, 0, 103, 0, 98, 0, 98, 0, 117, 0, 114, 0, 0, 0, 103, 0, 98, 0, 99, 0, 97, 0, 109, 0, 0, 0, 103, 0, 98, 0, 99, 0, 97, 0, 121, 0, 0, 0, 103, 0, 98, 0, 99, 0, 98, 0, 102, 0, 0, 0, 103, 0, 98, 0, 99, 0, 99, 0, 103, 0, 0, 0, 103, 0, 98, 0, 99, 0, 103, 0, 110, 0, 0, 0, 103, 0, 98, 0, 99, 0, 104, 0, 101, 0, 0, 0, 103, 0, 98, 0, 99, 0, 104, 0, 119, 0, 0, 0, 103, 0, 98, 0, 99, 0, 108, 0, 107, 0, 0, 0, 103, 0, 98, 0, 99, 0, 109, 0, 97, 0, 0, 0, 103, 0, 98, 0, 99, 0, 109, 0, 100, 0, 0, 0, 103, 0, 98, 0, 99, 0, 111, 0, 110, 0, 0, 0, 103, 0, 98, 0, 99, 0, 111, 0, 118, 0, 0, 0, 103, 0, 98, 0, 99, 0, 114, 0, 102, 0, 0, 0, 103, 0, 98, 0, 99, 0, 114, 0, 121, 0, 0, 0, 103, 0, 98, 0, 99, 0, 119, 0, 121, 0, 0, 0, 103, 0, 98, 0, 100, 0, 97, 0, 108, 0, 0, 0, 103, 0, 98, 0, 100, 0, 101, 0, 110, 0, 0, 0, 103, 0, 98, 0, 100, 0, 101, 0, 114, 0, 0, 0, 103, 0, 98, 0, 100, 0, 103, 0, 121, 0, 0, 0, 103, 0, 98, 0, 100, 0, 111, 0, 114, 0, 0, 0, 103, 0, 98, 0, 100, 0, 114, 0, 115, 0, 0, 0, 103, 0, 98, 0, 100, 0, 117, 0, 114, 0, 0, 0, 103, 0, 98, 0, 101, 0, 97, 0, 108, 0, 0, 0, 103, 0, 98, 0, 101, 0, 97, 0, 119, 0, 0, 0, 103, 0, 98, 0, 101, 0, 97, 0, 121, 0, 0, 0, 103, 0, 98, 0, 101, 0, 100, 0, 104, 0, 0, 0, 103, 0, 98, 0, 101, 0, 100, 0, 117, 0, 0, 0, 103, 0, 98, 0, 101, 0, 108, 0, 110, 0, 0, 0, 103, 0, 98, 0, 101, 0, 108, 0, 115, 0, 0, 0, 103, 0, 98, 0, 101, 0, 114, 0, 121, 0, 0, 0, 103, 0, 98, 0, 101, 0, 115, 0, 120, 0, 0, 0, 103, 0, 98, 0, 102, 0, 97, 0, 108, 0, 0, 0, 103, 0, 98, 0, 102, 0, 105, 0, 102, 0, 0, 0, 103, 0, 98, 0, 102, 0, 109, 0, 111, 0, 0, 0, 103, 0, 98, 0, 103, 0, 97, 0, 116, 0, 0, 0, 103, 0, 98, 0, 103, 0, 98, 0, 110, 0, 0, 0, 103, 0, 98, 0, 103, 0, 108, 0, 103, 0, 0, 0, 103, 0, 98, 0, 103, 0, 108, 0, 115, 0, 0, 0, 103, 0, 98, 0, 104, 0, 97, 0, 118, 0, 0, 0, 103, 0, 98, 0, 104, 0, 99, 0, 107, 0, 0, 0, 103, 0, 98, 0, 104, 0, 101, 0, 102, 0, 0, 0, 103, 0, 98, 0, 104, 0, 105, 0, 108, 0, 0, 0, 103, 0, 98, 0, 104, 0, 108, 0, 100, 0, 0, 0, 103, 0, 98, 0, 104, 0, 109, 0, 102, 0, 0, 0, 103, 0, 98, 0, 104, 0, 112, 0, 108, 0, 0, 0, 103, 0, 98, 0, 104, 0, 114, 0, 116, 0, 0, 0, 103, 0, 98, 0, 104, 0, 114, 0, 119, 0, 0, 0, 103, 0, 98, 0, 104, 0, 114, 0, 121, 0, 0, 0, 103, 0, 98, 0, 105, 0, 111, 0, 115, 0, 0, 0, 103, 0, 98, 0, 105, 0, 115, 0, 108, 0, 0, 0, 103, 0, 98, 0, 105, 0, 118, 0, 99, 0, 0, 0, 103, 0, 98, 0, 107, 0, 101, 0, 99, 0, 0, 0, 103, 0, 98, 0, 107, 0, 101, 0, 110, 0, 0, 0, 103, 0, 98, 0, 107, 0, 105, 0, 114, 0, 0, 0, 103, 0, 98, 0, 107, 0, 116, 0, 116, 0, 0, 0, 103, 0, 98, 0, 107, 0, 119, 0, 108, 0, 0, 0, 103, 0, 98, 0, 108, 0, 97, 0, 110, 0, 0, 0, 103, 0, 98, 0, 108, 0, 98, 0, 99, 0, 0, 0, 103, 0, 98, 0, 108, 0, 98, 0, 104, 0, 0, 0, 103, 0, 98, 0, 108, 0, 100, 0, 115, 0, 0, 0, 103, 0, 98, 0, 108, 0, 101, 0, 99, 0, 0, 0, 103, 0, 98, 0, 108, 0, 101, 0, 119, 0, 0, 0, 103, 0, 98, 0, 108, 0, 105, 0, 110, 0, 0, 0, 103, 0, 98, 0, 108, 0, 105, 0, 118, 0, 0, 0, 103, 0, 98, 0, 108, 0, 117, 0, 116, 0, 0, 0, 103, 0, 98, 0, 109, 0, 97, 0, 110, 0, 0, 0, 103, 0, 98, 0, 109, 0, 100, 0, 98, 0, 0, 0, 103, 0, 98, 0, 109, 0, 100, 0, 119, 0, 0, 0, 103, 0, 98, 0, 109, 0, 101, 0, 97, 0, 0, 0, 103, 0, 98, 0, 109, 0, 105, 0, 107, 0, 0, 0, 103, 0, 98, 0, 109, 0, 108, 0, 110, 0, 0, 0, 103, 0, 98, 0, 109, 0, 111, 0, 110, 0, 0, 0, 103, 0, 98, 0, 109, 0, 114, 0, 116, 0, 0, 0, 103, 0, 98, 0, 109, 0, 114, 0, 121, 0, 0, 0, 103, 0, 98, 0, 109, 0, 116, 0, 121, 0, 0, 0, 103, 0, 98, 0, 110, 0, 97, 0, 121, 0, 0, 0, 103, 0, 98, 0, 110, 0, 98, 0, 108, 0, 0, 0, 103, 0, 98, 0, 110, 0, 101, 0, 108, 0, 0, 0, 103, 0, 98, 0, 110, 0, 101, 0, 116, 0, 0, 0, 103, 0, 98, 0, 110, 0, 102, 0, 107, 0, 0, 0, 103, 0, 98, 0, 110, 0, 103, 0, 109, 0, 0, 0, 103, 0, 98, 0, 110, 0, 105, 0, 114, 0, 0, 0, 103, 0, 98, 0, 110, 0, 108, 0, 107, 0, 0, 0, 103, 0, 98, 0, 110, 0, 108, 0, 110, 0, 0, 0, 103, 0, 98, 0, 110, 0, 109, 0, 100, 0, 0, 0, 103, 0, 98, 0, 110, 0, 115, 0, 109, 0, 0, 0, 103, 0, 98, 0, 110, 0, 116, 0, 104, 0, 0, 0, 103, 0, 98, 0, 110, 0, 116, 0, 108, 0, 0, 0, 103, 0, 98, 0, 110, 0, 116, 0, 116, 0, 0, 0, 103, 0, 98, 0, 110, 0, 116, 0, 121, 0, 0, 0, 103, 0, 98, 0, 110, 0, 119, 0, 112, 0, 0, 0, 103, 0, 98, 0, 110, 0, 121, 0, 107, 0, 0, 0, 103, 0, 98, 0, 111, 0, 108, 0, 100, 0, 0, 0, 103, 0, 98, 0, 111, 0, 114, 0, 107, 0, 0, 0, 103, 0, 98, 0, 111, 0, 120, 0, 102, 0, 0, 0, 103, 0, 98, 0, 112, 0, 101, 0, 109, 0, 0, 0, 103, 0, 98, 0, 112, 0, 107, 0, 110, 0, 0, 0, 103, 0, 98, 0, 112, 0, 108, 0, 121, 0, 0, 0, 103, 0, 98, 0, 112, 0, 111, 0, 108, 0, 0, 0, 103, 0, 98, 0, 112, 0, 111, 0, 114, 0, 0, 0, 103, 0, 98, 0, 112, 0, 111, 0, 119, 0, 0, 0, 103, 0, 98, 0, 112, 0, 116, 0, 101, 0, 0, 0, 103, 0, 98, 0, 114, 0, 99, 0, 99, 0, 0, 0, 103, 0, 98, 0, 114, 0, 99, 0, 104, 0, 0, 0, 103, 0, 98, 0, 114, 0, 99, 0, 116, 0, 0, 0, 103, 0, 98, 0, 114, 0, 100, 0, 103, 0, 0, 0, 103, 0, 98, 0, 114, 0, 102, 0, 119, 0, 0, 0, 103, 0, 98, 0, 114, 0, 105, 0, 99, 0, 0, 0, 103, 0, 98, 0, 114, 0, 111, 0, 116, 0, 0, 0, 103, 0, 98, 0, 114, 0, 117, 0, 116, 0, 0, 0, 103, 0, 98, 0, 115, 0, 97, 0, 119, 0, 0, 0, 103, 0, 98, 0, 115, 0, 97, 0, 121, 0, 0, 0, 103, 0, 98, 0, 115, 0, 99, 0, 98, 0, 0, 0, 103, 0, 98, 0, 115, 0, 99, 0, 116, 0, 0, 0, 103, 0, 98, 0, 115, 0, 102, 0, 107, 0, 0, 0, 103, 0, 98, 0, 115, 0, 102, 0, 116, 0, 0, 0, 103, 0, 98, 0, 115, 0, 104, 0, 102, 0, 0, 0, 103, 0, 98, 0, 115, 0, 104, 0, 110, 0, 0, 0, 103, 0, 98, 0, 115, 0, 104, 0, 114, 0, 0, 0, 103, 0, 98, 0, 115, 0, 107, 0, 112, 0, 0, 0, 103, 0, 98, 0, 115, 0, 108, 0, 107, 0, 0, 0, 103, 0, 98, 0, 115, 0, 110, 0, 100, 0, 0, 0, 103, 0, 98, 0, 115, 0, 111, 0, 115, 0, 0, 0, 103, 0, 98, 0, 115, 0, 114, 0, 121, 0, 0, 0, 103, 0, 98, 0, 115, 0, 116, 0, 101, 0, 0, 0, 103, 0, 98, 0, 115, 0, 116, 0, 110, 0, 0, 0, 103, 0, 98, 0, 115, 0, 119, 0, 97, 0, 0, 0, 103, 0, 98, 0, 115, 0, 119, 0, 100, 0, 0, 0, 103, 0, 98, 0, 115, 0, 119, 0, 107, 0, 0, 0, 103, 0, 98, 0, 116, 0, 97, 0, 109, 0, 0, 0, 103, 0, 98, 0, 116, 0, 102, 0, 119, 0, 0, 0, 103, 0, 98, 0, 116, 0, 104, 0, 114, 0, 0, 0, 103, 0, 98, 0, 116, 0, 111, 0, 98, 0, 0, 0, 103, 0, 98, 0, 116, 0, 111, 0, 102, 0, 0, 0, 103, 0, 98, 0, 116, 0, 114, 0, 102, 0, 0, 0, 103, 0, 98, 0, 116, 0, 119, 0, 104, 0, 0, 0, 103, 0, 98, 0, 117, 0, 107, 0, 109, 0, 0, 0, 103, 0, 98, 0, 119, 0, 97, 0, 114, 0, 0, 0, 103, 0, 98, 0, 119, 0, 98, 0, 107, 0, 0, 0, 103, 0, 98, 0, 119, 0, 102, 0, 116, 0, 0, 0, 103, 0, 98, 0, 119, 0, 103, 0, 110, 0, 0, 0, 103, 0, 98, 0, 119, 0, 105, 0, 108, 0, 0, 0, 103, 0, 98, 0, 119, 0, 107, 0, 102, 0, 0, 0, 103, 0, 98, 0, 119, 0, 108, 0, 108, 0, 0, 0, 103, 0, 98, 0, 119, 0, 108, 0, 115, 0, 0, 0, 103, 0, 98, 0, 119, 0, 110, 0, 100, 0, 0, 0, 103, 0, 98, 0, 119, 0, 110, 0, 109, 0, 0, 0, 103, 0, 98, 0, 119, 0, 111, 0, 114, 0, 0, 0, 103, 0, 98, 0, 119, 0, 114, 0, 108, 0, 0, 0, 103, 0, 98, 0, 119, 0, 114, 0, 116, 0, 0, 0, 103, 0, 98, 0, 119, 0, 114, 0, 120, 0, 0, 0, 103, 0, 98, 0, 119, 0, 115, 0, 109, 0, 0, 0, 103, 0, 98, 0, 119, 0, 115, 0, 120, 0, 0, 0, 103, 0, 109, 0, 108, 0, 126, 0, 110, 0, 0, 0, 107, 0, 122, 0, 97, 0, 107, 0, 109, 0, 0, 0, 107, 0, 122, 0, 97, 0, 107, 0, 116, 0, 0, 0, 107, 0, 122, 0, 97, 0, 108, 0, 109, 0, 0, 0, 107, 0, 122, 0, 97, 0, 115, 0, 116, 0, 0, 0, 107, 0, 122, 0, 97, 0, 116, 0, 121, 0, 0, 0, 107, 0, 122, 0, 98, 0, 97, 0, 121, 0, 0, 0, 107, 0, 122, 0, 107, 0, 97, 0, 114, 0, 0, 0, 107, 0, 122, 0, 107, 0, 117, 0, 115, 0, 0, 0, 107, 0, 122, 0, 107, 0, 122, 0, 121, 0, 0, 0, 107, 0, 122, 0, 109, 0, 97, 0, 110, 0, 0, 0, 107, 0, 122, 0, 112, 0, 97, 0, 118, 0, 0, 0, 107, 0, 122, 0, 115, 0, 101, 0, 118, 0, 0, 0, 107, 0, 122, 0, 118, 0, 111, 0, 115, 0, 0, 0, 107, 0, 122, 0, 122, 0, 97, 0, 112, 0, 0, 0, 108, 0, 118, 0, 49, 0, 49, 0, 48, 0, 0, 0, 108, 0, 118, 0, 106, 0, 101, 0, 108, 0, 0, 0, 108, 0, 118, 0, 106, 0, 107, 0, 98, 0, 0, 0, 108, 0, 118, 0, 106, 0, 117, 0, 114, 0, 0, 0, 108, 0, 118, 0, 114, 0, 101, 0, 122, 0, 0, 0, 108, 0, 118, 0, 114, 0, 105, 0, 120, 0, 0, 0, 108, 0, 118, 0, 118, 0, 109, 0, 114, 0, 0, 0, 109, 0, 97, 0, 97, 0, 103, 0, 100, 0, 0, 0, 109, 0, 97, 0, 97, 0, 111, 0, 117, 0, 0, 0, 109, 0, 97, 0, 97, 0, 115, 0, 122, 0, 0, 0, 109, 0, 97, 0, 97, 0, 122, 0, 105, 0, 0, 0, 109, 0, 97, 0, 98, 0, 111, 0, 100, 0, 0, 0, 109, 0, 97, 0, 98, 0, 111, 0, 109, 0, 0, 0, 109, 0, 97, 0, 98, 0, 114, 0, 114, 0, 0, 0, 109, 0, 97, 0, 99, 0, 97, 0, 115, 0, 0, 0, 109, 0, 97, 0, 99, 0, 104, 0, 101, 0, 0, 0, 109, 0, 97, 0, 99, 0, 104, 0, 105, 0, 0, 0, 109, 0, 97, 0, 100, 0, 114, 0, 105, 0, 0, 0, 109, 0, 97, 0, 101, 0, 114, 0, 114, 0, 0, 0, 109, 0, 97, 0, 101, 0, 115, 0, 105, 0, 0, 0, 109, 0, 97, 0, 102, 0, 97, 0, 104, 0, 0, 0, 109, 0, 97, 0, 102, 0, 101, 0, 115, 0, 0, 0, 109, 0, 97, 0, 102, 0, 105, 0, 103, 0, 0, 0, 109, 0, 97, 0, 102, 0, 113, 0, 104, 0, 0, 0, 109, 0, 97, 0, 104, 0, 97, 0, 106, 0, 0, 0, 109, 0, 97, 0, 104, 0, 97, 0, 111, 0, 0, 0, 109, 0, 97, 0, 104, 0, 111, 0, 99, 0, 0, 0, 109, 0, 97, 0, 105, 0, 102, 0, 114, 0, 0, 0, 109, 0, 97, 0, 106, 0, 100, 0, 105, 0, 0, 0, 109, 0, 97, 0, 106, 0, 114, 0, 97, 0, 0, 0, 109, 0, 97, 0, 107, 0, 101, 0, 110, 0, 0, 0, 109, 0, 97, 0, 107, 0, 101, 0, 115, 0, 0, 0, 109, 0, 97, 0, 107, 0, 104, 0, 101, 0, 0, 0, 109, 0, 97, 0, 108, 0, 97, 0, 97, 0, 0, 0, 109, 0, 97, 0, 108, 0, 97, 0, 114, 0, 0, 0, 109, 0, 97, 0, 109, 0, 100, 0, 102, 0, 0, 0, 109, 0, 97, 0, 109, 0, 101, 0, 100, 0, 0, 0, 109, 0, 97, 0, 109, 0, 101, 0, 107, 0, 0, 0, 109, 0, 97, 0, 109, 0, 105, 0, 100, 0, 0, 0, 109, 0, 97, 0, 109, 0, 111, 0, 104, 0, 0, 0, 109, 0, 97, 0, 109, 0, 111, 0, 117, 0, 0, 0, 109, 0, 97, 0, 110, 0, 111, 0, 117, 0, 0, 0, 109, 0, 97, 0, 111, 0, 117, 0, 97, 0, 0, 0, 109, 0, 97, 0, 111, 0, 117, 0, 100, 0, 0, 0, 109, 0, 97, 0, 111, 0, 117, 0, 106, 0, 0, 0, 109, 0, 97, 0, 111, 0, 117, 0, 122, 0, 0, 0, 109, 0, 97, 0, 114, 0, 97, 0, 98, 0, 0, 0, 109, 0, 97, 0, 114, 0, 101, 0, 104, 0, 0, 0, 109, 0, 97, 0, 115, 0, 97, 0, 102, 0, 0, 0, 109, 0, 97, 0, 115, 0, 97, 0, 108, 0, 0, 0, 109, 0, 97, 0, 115, 0, 105, 0, 98, 0, 0, 0, 109, 0, 97, 0, 115, 0, 105, 0, 102, 0, 0, 0, 109, 0, 97, 0, 115, 0, 107, 0, 104, 0, 0, 0, 109, 0, 97, 0, 116, 0, 97, 0, 102, 0, 0, 0, 109, 0, 97, 0, 116, 0, 97, 0, 105, 0, 0, 0, 109, 0, 97, 0, 116, 0, 97, 0, 111, 0, 0, 0, 109, 0, 97, 0, 116, 0, 97, 0, 114, 0, 0, 0, 109, 0, 97, 0, 116, 0, 97, 0, 116, 0, 0, 0, 109, 0, 97, 0, 116, 0, 101, 0, 116, 0, 0, 0, 109, 0, 97, 0, 116, 0, 105, 0, 110, 0, 0, 0, 109, 0, 97, 0, 116, 0, 105, 0, 122, 0, 0, 0, 109, 0, 97, 0, 116, 0, 110, 0, 103, 0, 0, 0, 109, 0, 97, 0, 116, 0, 110, 0, 116, 0, 0, 0, 109, 0, 97, 0, 121, 0, 117, 0, 115, 0, 0, 0, 109, 0, 97, 0, 122, 0, 97, 0, 103, 0, 0, 0, 109, 0, 104, 0, 97, 0, 117, 0, 114, 0, 0, 0, 109, 0, 104, 0, 101, 0, 110, 0, 105, 0, 0, 0, 109, 0, 104, 0, 106, 0, 97, 0, 98, 0, 0, 0, 109, 0, 104, 0, 106, 0, 97, 0, 108, 0, 0, 0, 109, 0, 104, 0, 107, 0, 119, 0, 97, 0, 0, 0, 109, 0, 104, 0, 108, 0, 97, 0, 101, 0, 0, 0, 109, 0, 104, 0, 108, 0, 105, 0, 98, 0, 0, 0, 109, 0, 104, 0, 108, 0, 105, 0, 107, 0, 0, 0, 109, 0, 104, 0, 109, 0, 97, 0, 106, 0, 0, 0, 109, 0, 104, 0, 109, 0, 97, 0, 108, 0, 0, 0, 109, 0, 104, 0, 109, 0, 101, 0, 106, 0, 0, 0, 109, 0, 104, 0, 109, 0, 105, 0, 108, 0, 0, 0, 109, 0, 104, 0, 110, 0, 109, 0, 117, 0, 0, 0, 109, 0, 104, 0, 114, 0, 111, 0, 110, 0, 0, 0, 109, 0, 104, 0, 117, 0, 106, 0, 97, 0, 0, 0, 109, 0, 104, 0, 119, 0, 116, 0, 104, 0, 0, 0, 109, 0, 104, 0, 119, 0, 116, 0, 106, 0, 0, 0, 109, 0, 108, 0, 98, 0, 107, 0, 111, 0, 0, 0, 109, 0, 110, 0, 48, 0, 51, 0, 53, 0, 0, 0, 109, 0, 110, 0, 48, 0, 51, 0, 55, 0, 0, 0, 109, 0, 110, 0, 48, 0, 51, 0, 57, 0, 0, 0, 109, 0, 110, 0, 48, 0, 52, 0, 49, 0, 0, 0, 109, 0, 110, 0, 48, 0, 52, 0, 51, 0, 0, 0, 109, 0, 110, 0, 48, 0, 52, 0, 57, 0, 0, 0, 109, 0, 110, 0, 48, 0, 53, 0, 49, 0, 0, 0, 109, 0, 110, 0, 48, 0, 53, 0, 51, 0, 0, 0, 109, 0, 110, 0, 48, 0, 53, 0, 53, 0, 0, 0, 109, 0, 110, 0, 48, 0, 53, 0, 55, 0, 0, 0, 109, 0, 110, 0, 48, 0, 53, 0, 57, 0, 0, 0, 109, 0, 110, 0, 48, 0, 54, 0, 49, 0, 0, 0, 109, 0, 110, 0, 48, 0, 54, 0, 55, 0, 0, 0, 109, 0, 110, 0, 48, 0, 54, 0, 57, 0, 0, 0, 109, 0, 110, 0, 48, 0, 55, 0, 49, 0, 0, 0, 109, 0, 110, 0, 48, 0, 55, 0, 51, 0, 0, 0, 109, 0, 118, 0, 109, 0, 108, 0, 101, 0, 0, 0, 109, 0, 120, 0, 97, 0, 103, 0, 117, 0, 0, 0, 109, 0, 120, 0, 98, 0, 99, 0, 110, 0, 0, 0, 109, 0, 120, 0, 98, 0, 99, 0, 115, 0, 0, 0, 109, 0, 120, 0, 99, 0, 97, 0, 109, 0, 0, 0, 109, 0, 120, 0, 99, 0, 104, 0, 104, 0, 0, 0, 109, 0, 120, 0, 99, 0, 104, 0, 112, 0, 0, 0, 109, 0, 120, 0, 99, 0, 109, 0, 120, 0, 0, 0, 109, 0, 120, 0, 99, 0, 111, 0, 108, 0, 0, 0, 109, 0, 120, 0, 100, 0, 117, 0, 114, 0, 0, 0, 109, 0, 120, 0, 103, 0, 117, 0, 97, 0, 0, 0, 109, 0, 120, 0, 106, 0, 97, 0, 108, 0, 0, 0, 109, 0, 120, 0, 109, 0, 101, 0, 120, 0, 0, 0, 109, 0, 120, 0, 109, 0, 105, 0, 99, 0, 0, 0, 109, 0, 120, 0, 109, 0, 111, 0, 114, 0, 0, 0, 109, 0, 120, 0, 110, 0, 97, 0, 121, 0, 0, 0, 109, 0, 120, 0, 111, 0, 97, 0, 120, 0, 0, 0, 109, 0, 120, 0, 112, 0, 117, 0, 101, 0, 0, 0, 109, 0, 120, 0, 113, 0, 117, 0, 101, 0, 0, 0, 109, 0, 120, 0, 114, 0, 111, 0, 111, 0, 0, 0, 109, 0, 120, 0, 115, 0, 105, 0, 110, 0, 0, 0, 109, 0, 120, 0, 115, 0, 108, 0, 112, 0, 0, 0, 109, 0, 120, 0, 115, 0, 111, 0, 110, 0, 0, 0, 109, 0, 120, 0, 116, 0, 97, 0, 98, 0, 0, 0, 109, 0, 120, 0, 116, 0, 108, 0, 97, 0, 0, 0, 109, 0, 120, 0, 121, 0, 117, 0, 99, 0, 0, 0, 109, 0, 120, 0, 122, 0, 97, 0, 99, 0, 0, 0, 109, 0, 122, 0, 109, 0, 112, 0, 109, 0, 0, 0, 110, 0, 122, 0, 97, 0, 117, 0, 107, 0, 0, 0, 110, 0, 122, 0, 99, 0, 97, 0, 110, 0, 0, 0, 110, 0, 122, 0, 99, 0, 105, 0, 116, 0, 0, 0, 110, 0, 122, 0, 103, 0, 105, 0, 115, 0, 0, 0, 110, 0, 122, 0, 104, 0, 107, 0, 98, 0, 0, 0, 110, 0, 122, 0, 109, 0, 98, 0, 104, 0, 0, 0, 110, 0, 122, 0, 109, 0, 119, 0, 116, 0, 0, 0, 110, 0, 122, 0, 110, 0, 115, 0, 110, 0, 0, 0, 110, 0, 122, 0, 110, 0, 116, 0, 108, 0, 0, 0, 110, 0, 122, 0, 111, 0, 116, 0, 97, 0, 0, 0, 110, 0, 122, 0, 115, 0, 116, 0, 108, 0, 0, 0, 110, 0, 122, 0, 116, 0, 97, 0, 115, 0, 0, 0, 110, 0, 122, 0, 116, 0, 107, 0, 105, 0, 0, 0, 110, 0, 122, 0, 119, 0, 116, 0, 99, 0, 0, 0, 111, 0, 116, 0, 104, 0, 101, 0, 114, 0, 0, 0, 112, 0, 101, 0, 97, 0, 109, 0, 97, 0, 0, 0, 112, 0, 101, 0, 97, 0, 110, 0, 99, 0, 0, 0, 112, 0, 101, 0, 97, 0, 112, 0, 117, 0, 0, 0, 112, 0, 101, 0, 97, 0, 114, 0, 101, 0, 0, 0, 112, 0, 101, 0, 97, 0, 121, 0, 97, 0, 0, 0, 112, 0, 101, 0, 99, 0, 97, 0, 106, 0, 0, 0, 112, 0, 101, 0, 99, 0, 97, 0, 108, 0, 0, 0, 112, 0, 101, 0, 99, 0, 117, 0, 115, 0, 0, 0, 112, 0, 101, 0, 104, 0, 117, 0, 99, 0, 0, 0, 112, 0, 101, 0, 104, 0, 117, 0, 118, 0, 0, 0, 112, 0, 101, 0, 105, 0, 99, 0, 97, 0, 0, 0, 112, 0, 101, 0, 106, 0, 117, 0, 110, 0, 0, 0, 112, 0, 101, 0, 108, 0, 105, 0, 109, 0, 0, 0, 112, 0, 101, 0, 108, 0, 109, 0, 97, 0, 0, 0, 112, 0, 101, 0, 108, 0, 111, 0, 114, 0, 0, 0, 112, 0, 101, 0, 109, 0, 100, 0, 100, 0, 0, 0, 112, 0, 101, 0, 109, 0, 111, 0, 113, 0, 0, 0, 112, 0, 101, 0, 112, 0, 97, 0, 115, 0, 0, 0, 112, 0, 101, 0, 112, 0, 117, 0, 110, 0, 0, 0, 112, 0, 101, 0, 115, 0, 97, 0, 109, 0, 0, 0, 112, 0, 101, 0, 116, 0, 97, 0, 99, 0, 0, 0, 112, 0, 101, 0, 117, 0, 99, 0, 97, 0, 0, 0, 112, 0, 103, 0, 99, 0, 112, 0, 107, 0, 0, 0, 112, 0, 103, 0, 99, 0, 112, 0, 109, 0, 0, 0, 112, 0, 103, 0, 101, 0, 104, 0, 103, 0, 0, 0, 112, 0, 103, 0, 101, 0, 112, 0, 119, 0, 0, 0, 112, 0, 103, 0, 101, 0, 115, 0, 119, 0, 0, 0, 112, 0, 103, 0, 103, 0, 112, 0, 107, 0, 0, 0, 112, 0, 103, 0, 104, 0, 108, 0, 97, 0, 0, 0, 112, 0, 103, 0, 106, 0, 119, 0, 107, 0, 0, 0, 112, 0, 103, 0, 109, 0, 98, 0, 97, 0, 0, 0, 112, 0, 103, 0, 109, 0, 114, 0, 108, 0, 0, 0, 112, 0, 103, 0, 110, 0, 99, 0, 100, 0, 0, 0, 112, 0, 103, 0, 110, 0, 105, 0, 107, 0, 0, 0, 112, 0, 103, 0, 110, 0, 112, 0, 112, 0, 0, 0, 112, 0, 103, 0, 110, 0, 115, 0, 98, 0, 0, 0, 112, 0, 103, 0, 115, 0, 97, 0, 110, 0, 0, 0, 112, 0, 103, 0, 115, 0, 104, 0, 109, 0, 0, 0, 112, 0, 103, 0, 119, 0, 98, 0, 107, 0, 0, 0, 112, 0, 103, 0, 119, 0, 104, 0, 109, 0, 0, 0, 112, 0, 103, 0, 119, 0, 112, 0, 100, 0, 0, 0, 112, 0, 104, 0, 97, 0, 98, 0, 114, 0, 0, 0, 112, 0, 104, 0, 97, 0, 103, 0, 110, 0, 0, 0, 112, 0, 104, 0, 97, 0, 103, 0, 115, 0, 0, 0, 112, 0, 104, 0, 97, 0, 107, 0, 108, 0, 0, 0, 112, 0, 104, 0, 97, 0, 108, 0, 98, 0, 0, 0, 112, 0, 104, 0, 97, 0, 110, 0, 116, 0, 0, 0, 112, 0, 104, 0, 97, 0, 112, 0, 97, 0, 0, 0, 112, 0, 104, 0, 97, 0, 117, 0, 114, 0, 0, 0, 112, 0, 104, 0, 98, 0, 97, 0, 110, 0, 0, 0, 112, 0, 104, 0, 98, 0, 97, 0, 115, 0, 0, 0, 112, 0, 104, 0, 98, 0, 101, 0, 110, 0, 0, 0, 112, 0, 104, 0, 98, 0, 105, 0, 108, 0, 0, 0, 112, 0, 104, 0, 98, 0, 116, 0, 103, 0, 0, 0, 112, 0, 104, 0, 98, 0, 116, 0, 110, 0, 0, 0, 112, 0, 104, 0, 99, 0, 97, 0, 103, 0, 0, 0, 112, 0, 104, 0, 99, 0, 97, 0, 112, 0, 0, 0, 112, 0, 104, 0, 99, 0, 97, 0, 118, 0, 0, 0, 112, 0, 104, 0, 99, 0, 111, 0, 109, 0, 0, 0, 112, 0, 104, 0, 100, 0, 97, 0, 115, 0, 0, 0, 112, 0, 104, 0, 100, 0, 97, 0, 118, 0, 0, 0, 112, 0, 104, 0, 100, 0, 105, 0, 110, 0, 0, 0, 112, 0, 104, 0, 100, 0, 118, 0, 111, 0, 0, 0, 112, 0, 104, 0, 101, 0, 97, 0, 115, 0, 0, 0, 112, 0, 104, 0, 105, 0, 108, 0, 110, 0, 0, 0, 112, 0, 104, 0, 107, 0, 97, 0, 108, 0, 0, 0, 112, 0, 104, 0, 108, 0, 97, 0, 103, 0, 0, 0, 112, 0, 104, 0, 108, 0, 97, 0, 110, 0, 0, 0, 112, 0, 104, 0, 108, 0, 97, 0, 115, 0, 0, 0, 112, 0, 104, 0, 108, 0, 101, 0, 121, 0, 0, 0, 112, 0, 104, 0, 108, 0, 117, 0, 110, 0, 0, 0, 112, 0, 104, 0, 109, 0, 97, 0, 100, 0, 0, 0, 112, 0, 104, 0, 109, 0, 97, 0, 103, 0, 0, 0, 112, 0, 104, 0, 109, 0, 97, 0, 115, 0, 0, 0, 112, 0, 104, 0, 109, 0, 100, 0, 99, 0, 0, 0, 112, 0, 104, 0, 109, 0, 100, 0, 114, 0, 0, 0, 112, 0, 104, 0, 109, 0, 111, 0, 117, 0, 0, 0, 112, 0, 104, 0, 109, 0, 115, 0, 99, 0, 0, 0, 112, 0, 104, 0, 109, 0, 115, 0, 114, 0, 0, 0, 112, 0, 104, 0, 110, 0, 99, 0, 111, 0, 0, 0, 112, 0, 104, 0, 110, 0, 101, 0, 99, 0, 0, 0, 112, 0, 104, 0, 110, 0, 101, 0, 114, 0, 0, 0, 112, 0, 104, 0, 110, 0, 115, 0, 97, 0, 0, 0, 112, 0, 104, 0, 110, 0, 117, 0, 101, 0, 0, 0, 112, 0, 104, 0, 110, 0, 117, 0, 118, 0, 0, 0, 112, 0, 104, 0, 112, 0, 108, 0, 119, 0, 0, 0, 112, 0, 104, 0, 113, 0, 117, 0, 101, 0, 0, 0, 112, 0, 104, 0, 113, 0, 117, 0, 105, 0, 0, 0, 112, 0, 104, 0, 114, 0, 105, 0, 122, 0, 0, 0, 112, 0, 104, 0, 114, 0, 111, 0, 109, 0, 0, 0, 112, 0, 104, 0, 115, 0, 97, 0, 114, 0, 0, 0, 112, 0, 104, 0, 115, 0, 99, 0, 111, 0, 0, 0, 112, 0, 104, 0, 115, 0, 105, 0, 103, 0, 0, 0, 112, 0, 104, 0, 115, 0, 108, 0, 117, 0, 0, 0, 112, 0, 104, 0, 115, 0, 111, 0, 114, 0, 0, 0, 112, 0, 104, 0, 115, 0, 117, 0, 110, 0, 0, 0, 112, 0, 104, 0, 115, 0, 117, 0, 114, 0, 0, 0, 112, 0, 104, 0, 116, 0, 97, 0, 119, 0, 0, 0, 112, 0, 104, 0, 119, 0, 115, 0, 97, 0, 0, 0, 112, 0, 104, 0, 122, 0, 97, 0, 110, 0, 0, 0, 112, 0, 104, 0, 122, 0, 109, 0, 98, 0, 0, 0, 112, 0, 104, 0, 122, 0, 115, 0, 105, 0, 0, 0, 112, 0, 115, 0, 98, 0, 116, 0, 104, 0, 0, 0, 112, 0, 115, 0, 100, 0, 101, 0, 98, 0, 0, 0, 112, 0, 115, 0, 104, 0, 98, 0, 110, 0, 0, 0, 112, 0, 115, 0, 106, 0, 114, 0, 104, 0, 0, 0, 112, 0, 115, 0, 107, 0, 121, 0, 115, 0, 0, 0, 112, 0, 115, 0, 110, 0, 98, 0, 115, 0, 0, 0, 112, 0, 115, 0, 110, 0, 103, 0, 122, 0, 0, 0, 112, 0, 115, 0, 113, 0, 113, 0, 97, 0, 0, 0, 112, 0, 115, 0, 114, 0, 102, 0, 104, 0, 0, 0, 112, 0, 115, 0, 115, 0, 108, 0, 116, 0, 0, 0, 112, 0, 115, 0, 116, 0, 98, 0, 115, 0, 0, 0, 112, 0, 115, 0, 116, 0, 107, 0, 109, 0, 0, 0, 112, 0, 119, 0, 50, 0, 50, 0, 52, 0, 0, 0, 112, 0, 119, 0, 51, 0, 53, 0, 48, 0, 0, 0, 112, 0, 119, 0, 51, 0, 55, 0, 48, 0, 0, 0, 112, 0, 121, 0, 97, 0, 115, 0, 117, 0, 0, 0, 114, 0, 117, 0, 97, 0, 108, 0, 116, 0, 0, 0, 114, 0, 117, 0, 97, 0, 109, 0, 117, 0, 0, 0, 114, 0, 117, 0, 97, 0, 115, 0, 116, 0, 0, 0, 114, 0, 117, 0, 98, 0, 101, 0, 108, 0, 0, 0, 114, 0, 117, 0, 98, 0, 114, 0, 121, 0, 0, 0, 114, 0, 117, 0, 99, 0, 104, 0, 101, 0, 0, 0, 114, 0, 117, 0, 99, 0, 104, 0, 117, 0, 0, 0, 114, 0, 117, 0, 105, 0, 118, 0, 97, 0, 0, 0, 114, 0, 117, 0, 107, 0, 101, 0, 109, 0, 0, 0, 114, 0, 117, 0, 107, 0, 103, 0, 100, 0, 0, 0, 114, 0, 117, 0, 107, 0, 104, 0, 97, 0, 0, 0, 114, 0, 117, 0, 107, 0, 104, 0, 109, 0, 0, 0, 114, 0, 117, 0, 107, 0, 105, 0, 114, 0, 0, 0, 114, 0, 117, 0, 107, 0, 108, 0, 117, 0, 0, 0, 114, 0, 117, 0, 107, 0, 111, 0, 115, 0, 0, 0, 114, 0, 117, 0, 107, 0, 114, 0, 115, 0, 0, 0, 114, 0, 117, 0, 107, 0, 121, 0, 97, 0, 0, 0, 114, 0, 117, 0, 108, 0, 101, 0, 110, 0, 0, 0, 114, 0, 117, 0, 108, 0, 105, 0, 112, 0, 0, 0, 114, 0, 117, 0, 109, 0, 97, 0, 103, 0, 0, 0, 114, 0, 117, 0, 109, 0, 111, 0, 115, 0, 0, 0, 114, 0, 117, 0, 109, 0, 111, 0, 119, 0, 0, 0, 114, 0, 117, 0, 109, 0, 117, 0, 114, 0, 0, 0, 114, 0, 117, 0, 110, 0, 101, 0, 110, 0, 0, 0, 114, 0, 117, 0, 110, 0, 103, 0, 114, 0, 0, 0, 114, 0, 117, 0, 110, 0, 105, 0, 122, 0, 0, 0, 114, 0, 117, 0, 110, 0, 118, 0, 115, 0, 0, 0, 114, 0, 117, 0, 111, 0, 109, 0, 115, 0, 0, 0, 114, 0, 117, 0, 111, 0, 114, 0, 108, 0, 0, 0, 114, 0, 117, 0, 112, 0, 101, 0, 114, 0, 0, 0, 114, 0, 117, 0, 112, 0, 110, 0, 122, 0, 0, 0, 114, 0, 117, 0, 112, 0, 114, 0, 105, 0, 0, 0, 114, 0, 117, 0, 112, 0, 115, 0, 107, 0, 0, 0, 114, 0, 117, 0, 114, 0, 111, 0, 115, 0, 0, 0, 114, 0, 117, 0, 114, 0, 121, 0, 97, 0, 0, 0, 114, 0, 117, 0, 115, 0, 97, 0, 109, 0, 0, 0, 114, 0, 117, 0, 115, 0, 112, 0, 101, 0, 0, 0, 114, 0, 117, 0, 115, 0, 116, 0, 97, 0, 0, 0, 114, 0, 117, 0, 116, 0, 97, 0, 109, 0, 0, 0, 114, 0, 117, 0, 116, 0, 111, 0, 109, 0, 0, 0, 114, 0, 117, 0, 116, 0, 117, 0, 108, 0, 0, 0, 114, 0, 117, 0, 116, 0, 118, 0, 101, 0, 0, 0, 114, 0, 117, 0, 116, 0, 121, 0, 117, 0, 0, 0, 114, 0, 117, 0, 117, 0, 108, 0, 121, 0, 0, 0, 114, 0, 117, 0, 118, 0, 103, 0, 103, 0, 0, 0, 114, 0, 117, 0, 118, 0, 108, 0, 97, 0, 0, 0, 114, 0, 117, 0, 118, 0, 108, 0, 103, 0, 0, 0, 114, 0, 117, 0, 121, 0, 97, 0, 110, 0, 0, 0, 114, 0, 117, 0, 121, 0, 101, 0, 118, 0, 0, 0, 116, 0, 116, 0, 97, 0, 114, 0, 105, 0, 0, 0, 116, 0, 116, 0, 99, 0, 104, 0, 97, 0, 0, 0, 116, 0, 116, 0, 99, 0, 116, 0, 116, 0, 0, 0, 116, 0, 116, 0, 100, 0, 109, 0, 110, 0, 0, 0, 116, 0, 116, 0, 109, 0, 114, 0, 99, 0, 0, 0, 116, 0, 116, 0, 112, 0, 101, 0, 100, 0, 0, 0, 116, 0, 116, 0, 112, 0, 111, 0, 115, 0, 0, 0, 116, 0, 116, 0, 112, 0, 114, 0, 116, 0, 0, 0, 116, 0, 116, 0, 112, 0, 116, 0, 102, 0, 0, 0, 116, 0, 116, 0, 115, 0, 103, 0, 101, 0, 0, 0, 116, 0, 116, 0, 115, 0, 105, 0, 112, 0, 0, 0, 116, 0, 116, 0, 115, 0, 106, 0, 108, 0, 0, 0, 116, 0, 116, 0, 116, 0, 111, 0, 98, 0, 0, 0, 116, 0, 116, 0, 116, 0, 117, 0, 112, 0, 0, 0, 116, 0, 118, 0, 102, 0, 117, 0, 110, 0, 0, 0, 116, 0, 118, 0, 110, 0, 105, 0, 116, 0, 0, 0, 116, 0, 118, 0, 110, 0, 107, 0, 102, 0, 0, 0, 116, 0, 118, 0, 110, 0, 107, 0, 108, 0, 0, 0, 116, 0, 118, 0, 110, 0, 109, 0, 97, 0, 0, 0, 116, 0, 118, 0, 110, 0, 109, 0, 103, 0, 0, 0, 116, 0, 118, 0, 110, 0, 117, 0, 105, 0, 0, 0, 116, 0, 118, 0, 118, 0, 97, 0, 105, 0, 0, 0, 116, 0, 119, 0, 99, 0, 104, 0, 97, 0, 0, 0, 116, 0, 119, 0, 99, 0, 121, 0, 105, 0, 0, 0, 116, 0, 119, 0, 99, 0, 121, 0, 113, 0, 0, 0, 116, 0, 119, 0, 104, 0, 115, 0, 122, 0, 0, 0, 116, 0, 119, 0, 104, 0, 117, 0, 97, 0, 0, 0, 116, 0, 119, 0, 105, 0, 108, 0, 97, 0, 0, 0, 116, 0, 119, 0, 107, 0, 101, 0, 101, 0, 0, 0, 116, 0, 119, 0, 107, 0, 104, 0, 104, 0, 0, 0, 116, 0, 119, 0, 107, 0, 105, 0, 110, 0, 0, 0, 116, 0, 119, 0, 108, 0, 105, 0, 101, 0, 0, 0, 116, 0, 119, 0, 109, 0, 105, 0, 97, 0, 0, 0, 116, 0, 119, 0, 110, 0, 119, 0, 116, 0, 0, 0, 116, 0, 119, 0, 112, 0, 101, 0, 110, 0, 0, 0, 116, 0, 119, 0, 112, 0, 105, 0, 102, 0, 0, 0, 116, 0, 119, 0, 116, 0, 97, 0, 111, 0, 0, 0, 116, 0, 119, 0, 116, 0, 112, 0, 101, 0, 0, 0, 116, 0, 119, 0, 116, 0, 116, 0, 116, 0, 0, 0, 116, 0, 119, 0, 116, 0, 120, 0, 103, 0, 0, 0, 116, 0, 119, 0, 121, 0, 117, 0, 110, 0, 0, 0, 118, 0, 117, 0, 109, 0, 97, 0, 112, 0, 0, 0, 118, 0, 117, 0, 112, 0, 97, 0, 109, 0, 0, 0, 118, 0, 117, 0, 115, 0, 97, 0, 109, 0, 0, 0, 118, 0, 117, 0, 116, 0, 111, 0, 98, 0, 0, 0, 97, 0, 108, 0, 117, 0, 107, 0, 117, 0, 0, 0, 97, 0, 111, 0, 99, 0, 110, 0, 111, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 52, 0, 0, 0, 99, 0, 122, 0, 51, 0, 49, 0, 50, 0, 0, 0, 99, 0, 122, 0, 51, 0, 49, 0, 54, 0, 0, 0, 99, 0, 122, 0, 53, 0, 51, 0, 49, 0, 0, 0, 99, 0, 122, 0, 53, 0, 51, 0, 50, 0, 0, 0, 99, 0, 122, 0, 53, 0, 51, 0, 52, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 50, 0, 0, 0, 99, 0, 122, 0, 55, 0, 50, 0, 52, 0, 0, 0, 99, 0, 122, 0, 56, 0, 48, 0, 52, 0, 0, 0, 103, 0, 98, 0, 97, 0, 114, 0, 109, 0, 0, 0, 103, 0, 98, 0, 98, 0, 100, 0, 102, 0, 0, 0, 103, 0, 98, 0, 100, 0, 103, 0, 110, 0, 0, 0, 103, 0, 98, 0, 100, 0, 114, 0, 121, 0, 0, 0, 103, 0, 98, 0, 101, 0, 110, 0, 102, 0, 0, 0, 103, 0, 98, 0, 102, 0, 101, 0, 114, 0, 0, 0, 103, 0, 98, 0, 108, 0, 115, 0, 98, 0, 0, 0, 108, 0, 118, 0, 48, 0, 48, 0, 51, 0, 0, 0, 108, 0, 118, 0, 48, 0, 48, 0, 56, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 50, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 53, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 54, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 48, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 52, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 56, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 49, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 50, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 52, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 53, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 54, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 48, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 52, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 56, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 49, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 50, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 51, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 52, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 54, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 48, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 52, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 56, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 48, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 50, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 52, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 54, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 52, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 54, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 48, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 50, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 54, 0, 0, 0, 111, 0, 115, 0, 111, 0, 106, 0, 115, 0, 0, 0, 112, 0, 115, 0, 106, 0, 101, 0, 110, 0, 0, 0, 112, 0, 119, 0, 50, 0, 50, 0, 54, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 50, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 48, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 48, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 50, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 51, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 49, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 53, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 50, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 54, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 53, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 54, 0, 0, 0, 117, 0, 103, 0, 50, 0, 51, 0, 48, 0, 0, 0, 117, 0, 103, 0, 50, 0, 51, 0, 49, 0, 0, 0, 117, 0, 103, 0, 50, 0, 51, 0, 51, 0, 0, 0, 117, 0, 103, 0, 50, 0, 51, 0, 52, 0, 0, 0, 117, 0, 103, 0, 51, 0, 48, 0, 52, 0, 0, 0, 117, 0, 103, 0, 51, 0, 48, 0, 56, 0, 0, 0, 117, 0, 103, 0, 51, 0, 51, 0, 52, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 56, 0, 0, 0, 109, 0, 97, 0, 115, 0, 105, 0, 107, 0, 0, 0, 115, 0, 111, 0, 116, 0, 97, 0, 118, 0, 0, 0, 49, 0, 52, 0, 50, 0, 126, 0, 51, 0, 0, 0, 48, 0, 48, 0, 49, 0, 126, 0, 51, 0, 0, 0, 48, 0, 49, 0, 51, 0, 126, 0, 53, 0, 0, 0, 48, 0, 49, 0, 55, 0, 126, 0, 57, 0, 0, 0, 48, 0, 51, 0, 52, 0, 126, 0, 53, 0, 0, 0, 49, 0, 53, 0, 48, 0, 126, 0, 49, 0, 0, 0, 65, 0, 82, 0, 76, 0, 126, 0, 77, 0, 0, 0, 66, 0, 71, 0, 76, 0, 126, 0, 77, 0, 0, 0, 66, 0, 82, 0, 66, 0, 126, 0, 67, 0, 0, 0, 67, 0, 76, 0, 69, 0, 126, 0, 70, 0, 0, 0, 69, 0, 83, 0, 65, 0, 126, 0, 66, 0, 0, 0, 72, 0, 75, 0, 43, 0, 77, 0, 79, 0, 0, 0, 88, 0, 66, 0, 65, 0, 126, 0, 68, 0, 0, 0, 89, 0, 85, 0, 77, 0, 126, 0, 78, 0, 0, 0, 97, 0, 97, 0, 97, 0, 126, 0, 105, 0, 0, 0, 97, 0, 97, 0, 107, 0, 126, 0, 108, 0, 0, 0, 97, 0, 97, 0, 110, 0, 126, 0, 113, 0, 0, 0, 97, 0, 97, 0, 115, 0, 126, 0, 120, 0, 0, 0, 97, 0, 98, 0, 97, 0, 126, 0, 106, 0, 0, 0, 97, 0, 98, 0, 108, 0, 126, 0, 122, 0, 0, 0, 97, 0, 99, 0, 97, 0, 126, 0, 98, 0, 0, 0, 97, 0, 99, 0, 100, 0, 126, 0, 102, 0, 0, 0, 97, 0, 99, 0, 104, 0, 126, 0, 105, 0, 0, 0, 97, 0, 99, 0, 107, 0, 126, 0, 110, 0, 0, 0, 97, 0, 99, 0, 112, 0, 126, 0, 122, 0, 0, 0, 97, 0, 100, 0, 97, 0, 126, 0, 98, 0, 0, 0, 97, 0, 100, 0, 100, 0, 126, 0, 106, 0, 0, 0, 97, 0, 100, 0, 110, 0, 126, 0, 111, 0, 0, 0, 97, 0, 100, 0, 113, 0, 126, 0, 117, 0, 0, 0, 97, 0, 100, 0, 119, 0, 126, 0, 122, 0, 0, 0, 97, 0, 101, 0, 97, 0, 126, 0, 101, 0, 0, 0, 97, 0, 101, 0, 107, 0, 126, 0, 110, 0, 0, 0, 97, 0, 101, 0, 113, 0, 126, 0, 115, 0, 0, 0, 97, 0, 101, 0, 121, 0, 126, 0, 122, 0, 0, 0, 97, 0, 102, 0, 97, 0, 126, 0, 98, 0, 0, 0, 97, 0, 102, 0, 98, 0, 97, 0, 108, 0, 0, 0, 97, 0, 102, 0, 98, 0, 97, 0, 109, 0, 0, 0, 97, 0, 102, 0, 100, 0, 126, 0, 101, 0, 0, 0, 97, 0, 102, 0, 103, 0, 126, 0, 105, 0, 0, 0, 97, 0, 102, 0, 110, 0, 126, 0, 112, 0, 0, 0, 97, 0, 102, 0, 115, 0, 126, 0, 117, 0, 0, 0, 97, 0, 103, 0, 97, 0, 126, 0, 111, 0, 0, 0, 97, 0, 103, 0, 113, 0, 126, 0, 122, 0, 0, 0, 97, 0, 104, 0, 97, 0, 126, 0, 98, 0, 0, 0, 97, 0, 104, 0, 103, 0, 126, 0, 105, 0, 0, 0, 97, 0, 104, 0, 107, 0, 126, 0, 112, 0, 0, 0, 97, 0, 104, 0, 114, 0, 126, 0, 116, 0, 0, 0, 97, 0, 105, 0, 97, 0, 126, 0, 114, 0, 0, 0, 97, 0, 105, 0, 119, 0, 126, 0, 121, 0, 0, 0, 97, 0, 107, 0, 98, 0, 126, 0, 109, 0, 0, 0, 97, 0, 107, 0, 111, 0, 126, 0, 122, 0, 0, 0, 97, 0, 108, 0, 99, 0, 126, 0, 114, 0, 0, 0, 97, 0, 108, 0, 116, 0, 126, 0, 122, 0, 0, 0, 97, 0, 109, 0, 97, 0, 126, 0, 99, 0, 0, 0, 97, 0, 109, 0, 101, 0, 126, 0, 103, 0, 0, 0, 97, 0, 109, 0, 105, 0, 126, 0, 122, 0, 0, 0, 97, 0, 110, 0, 97, 0, 126, 0, 122, 0, 0, 0, 97, 0, 111, 0, 97, 0, 126, 0, 110, 0, 0, 0, 97, 0, 111, 0, 99, 0, 110, 0, 110, 0, 0, 0, 97, 0, 111, 0, 114, 0, 126, 0, 117, 0, 0, 0, 97, 0, 112, 0, 97, 0, 126, 0, 122, 0, 0, 0, 97, 0, 113, 0, 99, 0, 126, 0, 100, 0, 0, 0, 97, 0, 113, 0, 108, 0, 126, 0, 110, 0, 0, 0, 97, 0, 114, 0, 97, 0, 126, 0, 104, 0, 0, 0, 97, 0, 114, 0, 99, 0, 126, 0, 101, 0, 0, 0, 97, 0, 114, 0, 104, 0, 126, 0, 108, 0, 0, 0, 97, 0, 114, 0, 106, 0, 126, 0, 110, 0, 0, 0, 97, 0, 114, 0, 110, 0, 126, 0, 122, 0, 0, 0, 97, 0, 114, 0, 112, 0, 126, 0, 122, 0, 0, 0, 97, 0, 115, 0, 97, 0, 126, 0, 99, 0, 0, 0, 97, 0, 115, 0, 101, 0, 126, 0, 108, 0, 0, 0, 97, 0, 115, 0, 110, 0, 126, 0, 122, 0, 0, 0, 97, 0, 116, 0, 49, 0, 126, 0, 57, 0, 0, 0, 97, 0, 116, 0, 97, 0, 126, 0, 101, 0, 0, 0, 97, 0, 116, 0, 103, 0, 126, 0, 122, 0, 0, 0, 97, 0, 117, 0, 97, 0, 126, 0, 100, 0, 0, 0, 97, 0, 117, 0, 102, 0, 126, 0, 117, 0, 0, 0, 97, 0, 117, 0, 119, 0, 126, 0, 122, 0, 0, 0, 97, 0, 118, 0, 107, 0, 126, 0, 111, 0, 0, 0, 97, 0, 118, 0, 115, 0, 126, 0, 118, 0, 0, 0, 97, 0, 119, 0, 97, 0, 126, 0, 101, 0, 0, 0, 97, 0, 119, 0, 103, 0, 126, 0, 105, 0, 0, 0, 97, 0, 119, 0, 109, 0, 126, 0, 111, 0, 0, 0, 97, 0, 119, 0, 114, 0, 126, 0, 121, 0, 0, 0, 97, 0, 120, 0, 107, 0, 126, 0, 109, 0, 0, 0, 97, 0, 121, 0, 97, 0, 126, 0, 101, 0, 0, 0, 97, 0, 121, 0, 103, 0, 126, 0, 105, 0, 0, 0, 97, 0, 121, 0, 107, 0, 126, 0, 108, 0, 0, 0, 97, 0, 121, 0, 110, 0, 126, 0, 113, 0, 0, 0, 97, 0, 121, 0, 115, 0, 126, 0, 117, 0, 0, 0, 97, 0, 121, 0, 121, 0, 126, 0, 122, 0, 0, 0, 97, 0, 122, 0, 97, 0, 126, 0, 100, 0, 0, 0, 97, 0, 122, 0, 109, 0, 126, 0, 111, 0, 0, 0, 97, 0, 122, 0, 115, 0, 97, 0, 107, 0, 0, 0, 97, 0, 122, 0, 115, 0, 97, 0, 108, 0, 0, 0, 97, 0, 122, 0, 122, 0, 97, 0, 113, 0, 0, 0, 98, 0, 97, 0, 97, 0, 126, 0, 106, 0, 0, 0, 98, 0, 97, 0, 110, 0, 126, 0, 112, 0, 0, 0, 98, 0, 97, 0, 114, 0, 108, 0, 97, 0, 0, 0, 98, 0, 97, 0, 114, 0, 126, 0, 121, 0, 0, 0, 98, 0, 98, 0, 97, 0, 126, 0, 122, 0, 0, 0, 98, 0, 99, 0, 97, 0, 126, 0, 98, 0, 0, 0, 98, 0, 99, 0, 100, 0, 126, 0, 107, 0, 0, 0, 98, 0, 99, 0, 109, 0, 126, 0, 119, 0, 0, 0, 98, 0, 99, 0, 121, 0, 126, 0, 122, 0, 0, 0, 98, 0, 100, 0, 97, 0, 126, 0, 104, 0, 0, 0, 98, 0, 100, 0, 97, 0, 126, 0, 122, 0, 0, 0, 98, 0, 101, 0, 97, 0, 126, 0, 107, 0, 0, 0, 98, 0, 101, 0, 111, 0, 126, 0, 122, 0, 0, 0, 98, 0, 102, 0, 97, 0, 126, 0, 117, 0, 0, 0, 98, 0, 102, 0, 98, 0, 97, 0, 108, 0, 0, 0, 98, 0, 102, 0, 98, 0, 97, 0, 109, 0, 0, 0, 98, 0, 102, 0, 98, 0, 97, 0, 110, 0, 0, 0, 98, 0, 102, 0, 107, 0, 111, 0, 115, 0, 0, 0, 98, 0, 102, 0, 107, 0, 111, 0, 116, 0, 0, 0, 98, 0, 102, 0, 119, 0, 126, 0, 122, 0, 0, 0, 98, 0, 103, 0, 97, 0, 126, 0, 103, 0, 0, 0, 98, 0, 103, 0, 105, 0, 126, 0, 108, 0, 0, 0, 98, 0, 103, 0, 110, 0, 126, 0, 122, 0, 0, 0, 98, 0, 104, 0, 97, 0, 126, 0, 106, 0, 0, 0, 98, 0, 104, 0, 108, 0, 126, 0, 122, 0, 0, 0, 98, 0, 105, 0, 97, 0, 126, 0, 103, 0, 0, 0, 98, 0, 105, 0, 106, 0, 126, 0, 114, 0, 0, 0, 98, 0, 105, 0, 115, 0, 107, 0, 101, 0, 0, 0, 98, 0, 105, 0, 116, 0, 126, 0, 122, 0, 0, 0, 98, 0, 106, 0, 97, 0, 126, 0, 99, 0, 0, 0, 98, 0, 106, 0, 101, 0, 126, 0, 112, 0, 0, 0, 98, 0, 106, 0, 114, 0, 126, 0, 122, 0, 0, 0, 98, 0, 107, 0, 99, 0, 126, 0, 100, 0, 0, 0, 98, 0, 107, 0, 102, 0, 126, 0, 122, 0, 0, 0, 98, 0, 108, 0, 97, 0, 126, 0, 116, 0, 0, 0, 98, 0, 108, 0, 118, 0, 126, 0, 122, 0, 0, 0, 98, 0, 109, 0, 97, 0, 126, 0, 120, 0, 0, 0, 98, 0, 110, 0, 97, 0, 126, 0, 103, 0, 0, 0, 98, 0, 110, 0, 105, 0, 126, 0, 122, 0, 0, 0, 98, 0, 111, 0, 97, 0, 126, 0, 98, 0, 0, 0, 98, 0, 111, 0, 98, 0, 126, 0, 99, 0, 0, 0, 98, 0, 111, 0, 101, 0, 126, 0, 114, 0, 0, 0, 98, 0, 111, 0, 110, 0, 126, 0, 112, 0, 0, 0, 98, 0, 111, 0, 111, 0, 110, 0, 116, 0, 0, 0, 98, 0, 111, 0, 115, 0, 126, 0, 116, 0, 0, 0, 98, 0, 111, 0, 116, 0, 126, 0, 122, 0, 0, 0, 98, 0, 112, 0, 97, 0, 126, 0, 98, 0, 0, 0, 98, 0, 112, 0, 103, 0, 126, 0, 122, 0, 0, 0, 98, 0, 113, 0, 97, 0, 126, 0, 100, 0, 0, 0, 98, 0, 113, 0, 102, 0, 126, 0, 122, 0, 0, 0, 98, 0, 114, 0, 97, 0, 126, 0, 100, 0, 0, 0, 98, 0, 114, 0, 102, 0, 126, 0, 122, 0, 0, 0, 98, 0, 115, 0, 97, 0, 126, 0, 99, 0, 0, 0, 98, 0, 115, 0, 101, 0, 126, 0, 121, 0, 0, 0, 98, 0, 116, 0, 99, 0, 126, 0, 107, 0, 0, 0, 98, 0, 116, 0, 109, 0, 126, 0, 122, 0, 0, 0, 98, 0, 117, 0, 97, 0, 126, 0, 107, 0, 0, 0, 98, 0, 117, 0, 109, 0, 126, 0, 113, 0, 0, 0, 98, 0, 117, 0, 115, 0, 126, 0, 122, 0, 0, 0, 98, 0, 118, 0, 97, 0, 126, 0, 114, 0, 0, 0, 98, 0, 118, 0, 116, 0, 126, 0, 122, 0, 0, 0, 98, 0, 119, 0, 97, 0, 126, 0, 117, 0, 0, 0, 98, 0, 119, 0, 119, 0, 126, 0, 122, 0, 0, 0, 98, 0, 120, 0, 97, 0, 126, 0, 106, 0, 0, 0, 98, 0, 120, 0, 108, 0, 126, 0, 113, 0, 0, 0, 98, 0, 120, 0, 117, 0, 126, 0, 119, 0, 0, 0, 98, 0, 121, 0, 97, 0, 126, 0, 116, 0, 0, 0, 98, 0, 121, 0, 118, 0, 126, 0, 120, 0, 0, 0, 98, 0, 122, 0, 97, 0, 126, 0, 122, 0, 0, 0, 99, 0, 97, 0, 97, 0, 126, 0, 115, 0, 0, 0, 99, 0, 97, 0, 117, 0, 126, 0, 122, 0, 0, 0, 99, 0, 98, 0, 97, 0, 126, 0, 100, 0, 0, 0, 99, 0, 98, 0, 105, 0, 126, 0, 108, 0, 0, 0, 99, 0, 98, 0, 110, 0, 126, 0, 111, 0, 0, 0, 99, 0, 98, 0, 113, 0, 126, 0, 119, 0, 0, 0, 99, 0, 99, 0, 99, 0, 126, 0, 101, 0, 0, 0, 99, 0, 99, 0, 103, 0, 126, 0, 104, 0, 0, 0, 99, 0, 99, 0, 108, 0, 126, 0, 112, 0, 0, 0, 99, 0, 99, 0, 114, 0, 126, 0, 115, 0, 0, 0, 99, 0, 100, 0, 99, 0, 126, 0, 106, 0, 0, 0, 99, 0, 100, 0, 109, 0, 126, 0, 111, 0, 0, 0, 99, 0, 100, 0, 114, 0, 126, 0, 115, 0, 0, 0, 99, 0, 100, 0, 121, 0, 126, 0, 122, 0, 0, 0, 99, 0, 101, 0, 97, 0, 126, 0, 98, 0, 0, 0, 99, 0, 101, 0, 107, 0, 126, 0, 108, 0, 0, 0, 99, 0, 103, 0, 55, 0, 126, 0, 57, 0, 0, 0, 99, 0, 104, 0, 98, 0, 126, 0, 100, 0, 0, 0, 99, 0, 104, 0, 102, 0, 126, 0, 104, 0, 0, 0, 99, 0, 104, 0, 106, 0, 126, 0, 114, 0, 0, 0, 99, 0, 104, 0, 119, 0, 126, 0, 122, 0, 0, 0, 99, 0, 105, 0, 97, 0, 126, 0, 101, 0, 0, 0, 99, 0, 105, 0, 109, 0, 126, 0, 110, 0, 0, 0, 99, 0, 106, 0, 104, 0, 126, 0, 105, 0, 0, 0, 99, 0, 106, 0, 109, 0, 126, 0, 112, 0, 0, 0, 99, 0, 107, 0, 110, 0, 126, 0, 111, 0, 0, 0, 99, 0, 107, 0, 113, 0, 126, 0, 118, 0, 0, 0, 99, 0, 107, 0, 120, 0, 126, 0, 122, 0, 0, 0, 99, 0, 108, 0, 104, 0, 126, 0, 109, 0, 0, 0, 99, 0, 108, 0, 116, 0, 126, 0, 117, 0, 0, 0, 99, 0, 109, 0, 108, 0, 126, 0, 109, 0, 0, 0, 99, 0, 109, 0, 114, 0, 126, 0, 116, 0, 0, 0, 99, 0, 110, 0, 97, 0, 126, 0, 99, 0, 0, 0, 99, 0, 110, 0, 103, 0, 126, 0, 105, 0, 0, 0, 99, 0, 110, 0, 107, 0, 126, 0, 108, 0, 0, 0, 99, 0, 110, 0, 115, 0, 126, 0, 117, 0, 0, 0, 99, 0, 110, 0, 119, 0, 126, 0, 120, 0, 0, 0, 99, 0, 111, 0, 97, 0, 126, 0, 104, 0, 0, 0, 99, 0, 111, 0, 106, 0, 126, 0, 113, 0, 0, 0, 99, 0, 111, 0, 114, 0, 110, 0, 117, 0, 0, 0, 99, 0, 111, 0, 116, 0, 126, 0, 120, 0, 0, 0, 99, 0, 112, 0, 97, 0, 126, 0, 99, 0, 0, 0, 99, 0, 112, 0, 101, 0, 126, 0, 103, 0, 0, 0, 99, 0, 112, 0, 110, 0, 126, 0, 112, 0, 0, 0, 99, 0, 112, 0, 120, 0, 126, 0, 121, 0, 0, 0, 99, 0, 114, 0, 97, 0, 126, 0, 100, 0, 0, 0, 99, 0, 114, 0, 102, 0, 126, 0, 116, 0, 0, 0, 99, 0, 114, 0, 103, 0, 126, 0, 104, 0, 0, 0, 99, 0, 114, 0, 118, 0, 126, 0, 122, 0, 0, 0, 99, 0, 115, 0, 97, 0, 126, 0, 111, 0, 0, 0, 99, 0, 115, 0, 113, 0, 126, 0, 119, 0, 0, 0, 99, 0, 115, 0, 121, 0, 126, 0, 122, 0, 0, 0, 99, 0, 116, 0, 99, 0, 126, 0, 101, 0, 0, 0, 99, 0, 116, 0, 103, 0, 126, 0, 104, 0, 0, 0, 99, 0, 116, 0, 108, 0, 126, 0, 112, 0, 0, 0, 99, 0, 116, 0, 115, 0, 126, 0, 117, 0, 0, 0, 99, 0, 117, 0, 97, 0, 126, 0, 99, 0, 0, 0, 99, 0, 117, 0, 103, 0, 126, 0, 108, 0, 0, 0, 99, 0, 117, 0, 111, 0, 126, 0, 121, 0, 0, 0, 99, 0, 119, 0, 97, 0, 126, 0, 98, 0, 0, 0, 99, 0, 121, 0, 97, 0, 126, 0, 98, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 49, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 53, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 54, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 55, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 57, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 97, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 98, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 99, 0, 0, 0, 99, 0, 122, 0, 51, 0, 49, 0, 49, 0, 0, 0, 99, 0, 122, 0, 51, 0, 49, 0, 51, 0, 0, 0, 99, 0, 122, 0, 51, 0, 49, 0, 52, 0, 0, 0, 99, 0, 122, 0, 51, 0, 49, 0, 53, 0, 0, 0, 99, 0, 122, 0, 51, 0, 49, 0, 55, 0, 0, 0, 99, 0, 122, 0, 51, 0, 50, 0, 49, 0, 0, 0, 99, 0, 122, 0, 51, 0, 50, 0, 50, 0, 0, 0, 99, 0, 122, 0, 51, 0, 50, 0, 51, 0, 0, 0, 99, 0, 122, 0, 51, 0, 50, 0, 53, 0, 0, 0, 99, 0, 122, 0, 51, 0, 50, 0, 54, 0, 0, 0, 99, 0, 122, 0, 51, 0, 50, 0, 55, 0, 0, 0, 99, 0, 122, 0, 52, 0, 49, 0, 49, 0, 0, 0, 99, 0, 122, 0, 52, 0, 49, 0, 50, 0, 0, 0, 99, 0, 122, 0, 52, 0, 49, 0, 51, 0, 0, 0, 99, 0, 122, 0, 52, 0, 50, 0, 49, 0, 0, 0, 99, 0, 122, 0, 52, 0, 50, 0, 51, 0, 0, 0, 99, 0, 122, 0, 52, 0, 50, 0, 52, 0, 0, 0, 99, 0, 122, 0, 52, 0, 50, 0, 53, 0, 0, 0, 99, 0, 122, 0, 52, 0, 50, 0, 55, 0, 0, 0, 99, 0, 122, 0, 53, 0, 49, 0, 49, 0, 0, 0, 99, 0, 122, 0, 53, 0, 49, 0, 51, 0, 0, 0, 99, 0, 122, 0, 53, 0, 49, 0, 52, 0, 0, 0, 99, 0, 122, 0, 53, 0, 50, 0, 49, 0, 0, 0, 99, 0, 122, 0, 53, 0, 50, 0, 50, 0, 0, 0, 99, 0, 122, 0, 53, 0, 50, 0, 51, 0, 0, 0, 99, 0, 122, 0, 53, 0, 50, 0, 53, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 49, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 50, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 51, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 53, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 49, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 52, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 53, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 55, 0, 0, 0, 99, 0, 122, 0, 55, 0, 49, 0, 49, 0, 0, 0, 99, 0, 122, 0, 55, 0, 49, 0, 50, 0, 0, 0, 99, 0, 122, 0, 55, 0, 49, 0, 51, 0, 0, 0, 99, 0, 122, 0, 55, 0, 49, 0, 52, 0, 0, 0, 99, 0, 122, 0, 55, 0, 49, 0, 53, 0, 0, 0, 99, 0, 122, 0, 55, 0, 50, 0, 49, 0, 0, 0, 99, 0, 122, 0, 55, 0, 50, 0, 50, 0, 0, 0, 99, 0, 122, 0, 55, 0, 50, 0, 51, 0, 0, 0, 99, 0, 122, 0, 56, 0, 48, 0, 49, 0, 0, 0, 99, 0, 122, 0, 56, 0, 48, 0, 50, 0, 0, 0, 99, 0, 122, 0, 56, 0, 48, 0, 51, 0, 0, 0, 99, 0, 122, 0, 56, 0, 48, 0, 53, 0, 0, 0, 99, 0, 122, 0, 56, 0, 48, 0, 54, 0, 0, 0, 99, 0, 122, 0, 110, 0, 126, 0, 111, 0, 0, 0, 100, 0, 97, 0, 99, 0, 126, 0, 101, 0, 0, 0, 100, 0, 97, 0, 103, 0, 126, 0, 109, 0, 0, 0, 100, 0, 97, 0, 110, 0, 103, 0, 105, 0, 0, 0, 100, 0, 97, 0, 113, 0, 126, 0, 115, 0, 0, 0, 100, 0, 97, 0, 117, 0, 126, 0, 122, 0, 0, 0, 100, 0, 98, 0, 100, 0, 126, 0, 103, 0, 0, 0, 100, 0, 98, 0, 105, 0, 126, 0, 106, 0, 0, 0, 100, 0, 98, 0, 108, 0, 126, 0, 114, 0, 0, 0, 100, 0, 98, 0, 116, 0, 126, 0, 119, 0, 0, 0, 100, 0, 100, 0, 100, 0, 126, 0, 101, 0, 0, 0, 100, 0, 100, 0, 105, 0, 126, 0, 106, 0, 0, 0, 100, 0, 100, 0, 110, 0, 126, 0, 111, 0, 0, 0, 100, 0, 100, 0, 114, 0, 126, 0, 115, 0, 0, 0, 100, 0, 101, 0, 99, 0, 126, 0, 105, 0, 0, 0, 100, 0, 101, 0, 107, 0, 126, 0, 110, 0, 0, 0, 100, 0, 101, 0, 112, 0, 126, 0, 115, 0, 0, 0, 100, 0, 103, 0, 97, 0, 126, 0, 101, 0, 0, 0, 100, 0, 103, 0, 103, 0, 126, 0, 105, 0, 0, 0, 100, 0, 103, 0, 107, 0, 126, 0, 108, 0, 0, 0, 100, 0, 103, 0, 114, 0, 126, 0, 117, 0, 0, 0, 100, 0, 103, 0, 119, 0, 126, 0, 120, 0, 0, 0, 100, 0, 104, 0, 108, 0, 126, 0, 111, 0, 0, 0, 100, 0, 104, 0, 114, 0, 126, 0, 115, 0, 0, 0, 100, 0, 104, 0, 117, 0, 126, 0, 120, 0, 0, 0, 100, 0, 105, 0, 97, 0, 126, 0, 100, 0, 0, 0, 100, 0, 105, 0, 102, 0, 126, 0, 106, 0, 0, 0, 100, 0, 105, 0, 108, 0, 126, 0, 112, 0, 0, 0, 100, 0, 105, 0, 114, 0, 126, 0, 115, 0, 0, 0, 100, 0, 105, 0, 119, 0, 126, 0, 122, 0, 0, 0, 100, 0, 106, 0, 97, 0, 126, 0, 102, 0, 0, 0, 100, 0, 106, 0, 105, 0, 126, 0, 107, 0, 0, 0, 100, 0, 106, 0, 109, 0, 126, 0, 111, 0, 0, 0, 100, 0, 107, 0, 114, 0, 126, 0, 115, 0, 0, 0, 100, 0, 108, 0, 109, 0, 126, 0, 110, 0, 0, 0, 100, 0, 109, 0, 97, 0, 126, 0, 101, 0, 0, 0, 100, 0, 109, 0, 107, 0, 126, 0, 111, 0, 0, 0, 100, 0, 109, 0, 114, 0, 126, 0, 115, 0, 0, 0, 100, 0, 109, 0, 117, 0, 126, 0, 121, 0, 0, 0, 100, 0, 110, 0, 100, 0, 126, 0, 101, 0, 0, 0, 100, 0, 110, 0, 105, 0, 126, 0, 107, 0, 0, 0, 100, 0, 110, 0, 110, 0, 126, 0, 111, 0, 0, 0, 100, 0, 110, 0, 116, 0, 126, 0, 119, 0, 0, 0, 100, 0, 111, 0, 97, 0, 126, 0, 99, 0, 0, 0, 100, 0, 111, 0, 101, 0, 126, 0, 102, 0, 0, 0, 100, 0, 111, 0, 104, 0, 126, 0, 105, 0, 0, 0, 100, 0, 111, 0, 107, 0, 126, 0, 108, 0, 0, 0, 100, 0, 111, 0, 110, 0, 126, 0, 116, 0, 0, 0, 100, 0, 111, 0, 118, 0, 126, 0, 122, 0, 0, 0, 100, 0, 114, 0, 97, 0, 126, 0, 101, 0, 0, 0, 100, 0, 114, 0, 110, 0, 126, 0, 111, 0, 0, 0, 100, 0, 114, 0, 113, 0, 126, 0, 117, 0, 0, 0, 100, 0, 115, 0, 104, 0, 126, 0, 105, 0, 0, 0, 100, 0, 116, 0, 97, 0, 126, 0, 98, 0, 0, 0, 100, 0, 116, 0, 104, 0, 126, 0, 105, 0, 0, 0, 100, 0, 116, 0, 109, 0, 126, 0, 112, 0, 0, 0, 100, 0, 116, 0, 114, 0, 126, 0, 117, 0, 0, 0, 100, 0, 117, 0, 97, 0, 126, 0, 99, 0, 0, 0, 100, 0, 117, 0, 101, 0, 126, 0, 105, 0, 0, 0, 100, 0, 117, 0, 107, 0, 126, 0, 115, 0, 0, 0, 100, 0, 117, 0, 117, 0, 126, 0, 122, 0, 0, 0, 100, 0, 119, 0, 114, 0, 126, 0, 115, 0, 0, 0, 100, 0, 119, 0, 121, 0, 126, 0, 122, 0, 0, 0, 100, 0, 121, 0, 97, 0, 126, 0, 98, 0, 0, 0, 100, 0, 121, 0, 109, 0, 126, 0, 111, 0, 0, 0, 101, 0, 99, 0, 97, 0, 126, 0, 105, 0, 0, 0, 101, 0, 99, 0, 108, 0, 126, 0, 112, 0, 0, 0, 101, 0, 99, 0, 116, 0, 126, 0, 117, 0, 0, 0, 101, 0, 99, 0, 119, 0, 126, 0, 122, 0, 0, 0, 101, 0, 103, 0, 120, 0, 126, 0, 121, 0, 0, 0, 101, 0, 107, 0, 108, 0, 126, 0, 109, 0, 0, 0, 101, 0, 107, 0, 111, 0, 126, 0, 112, 0, 0, 0, 101, 0, 108, 0, 104, 0, 126, 0, 105, 0, 0, 0, 101, 0, 109, 0, 97, 0, 126, 0, 98, 0, 0, 0, 101, 0, 109, 0, 109, 0, 126, 0, 110, 0, 0, 0, 101, 0, 109, 0, 119, 0, 126, 0, 121, 0, 0, 0, 101, 0, 110, 0, 95, 0, 66, 0, 90, 0, 0, 0, 101, 0, 110, 0, 95, 0, 67, 0, 72, 0, 0, 0, 101, 0, 110, 0, 95, 0, 72, 0, 75, 0, 0, 0, 101, 0, 110, 0, 95, 0, 77, 0, 79, 0, 0, 0, 101, 0, 110, 0, 97, 0, 126, 0, 100, 0, 0, 0, 101, 0, 110, 0, 108, 0, 126, 0, 111, 0, 0, 0, 101, 0, 110, 0, 113, 0, 126, 0, 114, 0, 0, 0, 101, 0, 110, 0, 117, 0, 126, 0, 120, 0, 0, 0, 101, 0, 114, 0, 103, 0, 126, 0, 105, 0, 0, 0, 101, 0, 114, 0, 114, 0, 126, 0, 116, 0, 0, 0, 101, 0, 115, 0, 95, 0, 66, 0, 82, 0, 0, 0, 101, 0, 115, 0, 95, 0, 85, 0, 83, 0, 0, 0, 101, 0, 115, 0, 103, 0, 126, 0, 105, 0, 0, 0, 101, 0, 115, 0, 108, 0, 126, 0, 111, 0, 0, 0, 101, 0, 115, 0, 120, 0, 126, 0, 121, 0, 0, 0, 101, 0, 116, 0, 98, 0, 126, 0, 99, 0, 0, 0, 101, 0, 116, 0, 110, 0, 126, 0, 111, 0, 0, 0, 101, 0, 116, 0, 114, 0, 126, 0, 117, 0, 0, 0, 102, 0, 97, 0, 97, 0, 126, 0, 98, 0, 0, 0, 102, 0, 97, 0, 102, 0, 126, 0, 110, 0, 0, 0, 102, 0, 97, 0, 120, 0, 126, 0, 122, 0, 0, 0, 102, 0, 105, 0, 116, 0, 126, 0, 117, 0, 0, 0, 102, 0, 108, 0, 104, 0, 126, 0, 105, 0, 0, 0, 102, 0, 111, 0, 109, 0, 126, 0, 110, 0, 0, 0, 102, 0, 111, 0, 114, 0, 126, 0, 115, 0, 0, 0, 102, 0, 114, 0, 97, 0, 126, 0, 98, 0, 0, 0, 102, 0, 114, 0, 99, 0, 126, 0, 100, 0, 0, 0, 102, 0, 114, 0, 100, 0, 126, 0, 103, 0, 0, 0, 102, 0, 114, 0, 104, 0, 126, 0, 109, 0, 0, 0, 102, 0, 114, 0, 111, 0, 126, 0, 112, 0, 0, 0, 102, 0, 114, 0, 111, 0, 126, 0, 116, 0, 0, 0, 102, 0, 114, 0, 113, 0, 126, 0, 114, 0, 0, 0, 102, 0, 114, 0, 115, 0, 126, 0, 116, 0, 0, 0, 102, 0, 114, 0, 117, 0, 126, 0, 118, 0, 0, 0, 102, 0, 117, 0, 100, 0, 126, 0, 102, 0, 0, 0, 102, 0, 117, 0, 104, 0, 126, 0, 106, 0, 0, 0, 102, 0, 117, 0, 109, 0, 126, 0, 110, 0, 0, 0, 102, 0, 117, 0, 113, 0, 126, 0, 114, 0, 0, 0, 102, 0, 117, 0, 116, 0, 126, 0, 118, 0, 0, 0, 103, 0, 97, 0, 49, 0, 126, 0, 57, 0, 0, 0, 103, 0, 97, 0, 97, 0, 126, 0, 117, 0, 0, 0, 103, 0, 97, 0, 119, 0, 126, 0, 121, 0, 0, 0, 103, 0, 98, 0, 97, 0, 98, 0, 99, 0, 0, 0, 103, 0, 98, 0, 97, 0, 98, 0, 100, 0, 0, 0, 103, 0, 98, 0, 97, 0, 98, 0, 101, 0, 0, 0, 103, 0, 98, 0, 97, 0, 110, 0, 116, 0, 0, 0, 103, 0, 98, 0, 97, 0, 114, 0, 100, 0, 0, 0, 103, 0, 98, 0, 97, 0, 126, 0, 98, 0, 0, 0, 103, 0, 98, 0, 98, 0, 100, 0, 103, 0, 0, 0, 103, 0, 98, 0, 98, 0, 108, 0, 97, 0, 0, 0, 103, 0, 98, 0, 98, 0, 108, 0, 121, 0, 0, 0, 103, 0, 98, 0, 98, 0, 110, 0, 98, 0, 0, 0, 103, 0, 98, 0, 98, 0, 114, 0, 99, 0, 0, 0, 103, 0, 98, 0, 98, 0, 114, 0, 100, 0, 0, 0, 103, 0, 98, 0, 99, 0, 103, 0, 118, 0, 0, 0, 103, 0, 98, 0, 99, 0, 107, 0, 102, 0, 0, 0, 103, 0, 98, 0, 99, 0, 107, 0, 116, 0, 0, 0, 103, 0, 98, 0, 99, 0, 108, 0, 114, 0, 0, 0, 103, 0, 98, 0, 99, 0, 115, 0, 114, 0, 0, 0, 103, 0, 98, 0, 100, 0, 110, 0, 99, 0, 0, 0, 103, 0, 98, 0, 100, 0, 110, 0, 100, 0, 0, 0, 103, 0, 98, 0, 100, 0, 111, 0, 119, 0, 0, 0, 103, 0, 98, 0, 100, 0, 126, 0, 110, 0, 0, 0, 103, 0, 98, 0, 101, 0, 110, 0, 103, 0, 0, 0, 103, 0, 98, 0, 104, 0, 97, 0, 108, 0, 0, 0, 103, 0, 98, 0, 104, 0, 97, 0, 109, 0, 0, 0, 103, 0, 98, 0, 108, 0, 109, 0, 118, 0, 0, 0, 103, 0, 98, 0, 108, 0, 114, 0, 110, 0, 0, 0, 103, 0, 98, 0, 109, 0, 102, 0, 116, 0, 0, 0, 103, 0, 98, 0, 109, 0, 121, 0, 108, 0, 0, 0, 103, 0, 98, 0, 110, 0, 100, 0, 110, 0, 0, 0, 103, 0, 98, 0, 110, 0, 116, 0, 97, 0, 0, 0, 103, 0, 98, 0, 110, 0, 121, 0, 109, 0, 0, 0, 103, 0, 98, 0, 111, 0, 109, 0, 104, 0, 0, 0, 103, 0, 98, 0, 112, 0, 126, 0, 115, 0, 0, 0, 103, 0, 98, 0, 115, 0, 108, 0, 102, 0, 0, 0, 103, 0, 98, 0, 115, 0, 108, 0, 103, 0, 0, 0, 103, 0, 98, 0, 115, 0, 111, 0, 108, 0, 0, 0, 103, 0, 98, 0, 115, 0, 111, 0, 109, 0, 0, 0, 103, 0, 98, 0, 115, 0, 116, 0, 98, 0, 0, 0, 103, 0, 98, 0, 115, 0, 116, 0, 103, 0, 0, 0, 103, 0, 98, 0, 115, 0, 116, 0, 104, 0, 0, 0, 103, 0, 98, 0, 115, 0, 116, 0, 116, 0, 0, 0, 103, 0, 98, 0, 117, 0, 126, 0, 122, 0, 0, 0, 103, 0, 99, 0, 99, 0, 126, 0, 102, 0, 0, 0, 103, 0, 100, 0, 97, 0, 126, 0, 111, 0, 0, 0, 103, 0, 100, 0, 113, 0, 126, 0, 117, 0, 0, 0, 103, 0, 101, 0, 97, 0, 126, 0, 100, 0, 0, 0, 103, 0, 101, 0, 103, 0, 126, 0, 109, 0, 0, 0, 103, 0, 101, 0, 118, 0, 126, 0, 122, 0, 0, 0, 103, 0, 103, 0, 97, 0, 126, 0, 98, 0, 0, 0, 103, 0, 103, 0, 100, 0, 126, 0, 101, 0, 0, 0, 103, 0, 103, 0, 107, 0, 126, 0, 108, 0, 0, 0, 103, 0, 103, 0, 110, 0, 126, 0, 111, 0, 0, 0, 103, 0, 103, 0, 116, 0, 126, 0, 117, 0, 0, 0, 103, 0, 104, 0, 107, 0, 126, 0, 108, 0, 0, 0, 103, 0, 104, 0, 110, 0, 126, 0, 111, 0, 0, 0, 103, 0, 104, 0, 114, 0, 126, 0, 116, 0, 0, 0, 103, 0, 105, 0, 97, 0, 126, 0, 101, 0, 0, 0, 103, 0, 105, 0, 103, 0, 126, 0, 104, 0, 0, 0, 103, 0, 105, 0, 108, 0, 126, 0, 110, 0, 0, 0, 103, 0, 105, 0, 112, 0, 126, 0, 117, 0, 0, 0, 103, 0, 105, 0, 119, 0, 126, 0, 122, 0, 0, 0, 103, 0, 106, 0, 109, 0, 126, 0, 110, 0, 0, 0, 103, 0, 107, 0, 110, 0, 126, 0, 112, 0, 0, 0, 103, 0, 108, 0, 99, 0, 126, 0, 100, 0, 0, 0, 103, 0, 108, 0, 104, 0, 126, 0, 108, 0, 0, 0, 103, 0, 109, 0, 97, 0, 126, 0, 98, 0, 0, 0, 103, 0, 109, 0, 100, 0, 126, 0, 101, 0, 0, 0, 103, 0, 109, 0, 103, 0, 126, 0, 104, 0, 0, 0, 103, 0, 109, 0, 117, 0, 126, 0, 122, 0, 0, 0, 103, 0, 110, 0, 97, 0, 126, 0, 101, 0, 0, 0, 103, 0, 110, 0, 103, 0, 126, 0, 110, 0, 0, 0, 103, 0, 110, 0, 113, 0, 126, 0, 114, 0, 0, 0, 103, 0, 110, 0, 116, 0, 126, 0, 117, 0, 0, 0, 103, 0, 111, 0, 97, 0, 126, 0, 117, 0, 0, 0, 103, 0, 111, 0, 119, 0, 126, 0, 122, 0, 0, 0, 103, 0, 114, 0, 97, 0, 126, 0, 100, 0, 0, 0, 103, 0, 114, 0, 97, 0, 126, 0, 109, 0, 0, 0, 103, 0, 114, 0, 103, 0, 126, 0, 107, 0, 0, 0, 103, 0, 114, 0, 113, 0, 126, 0, 122, 0, 0, 0, 103, 0, 115, 0, 108, 0, 126, 0, 112, 0, 0, 0, 103, 0, 117, 0, 97, 0, 126, 0, 102, 0, 0, 0, 103, 0, 117, 0, 104, 0, 126, 0, 105, 0, 0, 0, 103, 0, 117, 0, 107, 0, 126, 0, 117, 0, 0, 0, 103, 0, 117, 0, 119, 0, 126, 0, 120, 0, 0, 0, 103, 0, 118, 0, 101, 0, 126, 0, 102, 0, 0, 0, 103, 0, 118, 0, 108, 0, 126, 0, 112, 0, 0, 0, 103, 0, 118, 0, 114, 0, 126, 0, 115, 0, 0, 0, 103, 0, 119, 0, 97, 0, 126, 0, 103, 0, 0, 0, 103, 0, 119, 0, 105, 0, 126, 0, 106, 0, 0, 0, 103, 0, 119, 0, 109, 0, 126, 0, 110, 0, 0, 0, 103, 0, 119, 0, 116, 0, 126, 0, 117, 0, 0, 0, 103, 0, 119, 0, 119, 0, 126, 0, 120, 0, 0, 0, 103, 0, 121, 0, 100, 0, 126, 0, 103, 0, 0, 0, 103, 0, 121, 0, 108, 0, 126, 0, 111, 0, 0, 0, 104, 0, 97, 0, 97, 0, 126, 0, 115, 0, 0, 0, 104, 0, 97, 0, 118, 0, 126, 0, 122, 0, 0, 0, 104, 0, 98, 0, 97, 0, 126, 0, 98, 0, 0, 0, 104, 0, 98, 0, 110, 0, 126, 0, 111, 0, 0, 0, 104, 0, 101, 0, 103, 0, 126, 0, 105, 0, 0, 0, 104, 0, 105, 0, 97, 0, 126, 0, 98, 0, 0, 0, 104, 0, 105, 0, 102, 0, 126, 0, 108, 0, 0, 0, 104, 0, 105, 0, 119, 0, 126, 0, 120, 0, 0, 0, 104, 0, 108, 0, 97, 0, 126, 0, 98, 0, 0, 0, 104, 0, 108, 0, 100, 0, 126, 0, 101, 0, 0, 0, 104, 0, 108, 0, 116, 0, 126, 0, 117, 0, 0, 0, 104, 0, 109, 0, 97, 0, 126, 0, 110, 0, 0, 0, 104, 0, 109, 0, 112, 0, 126, 0, 122, 0, 0, 0, 104, 0, 110, 0, 100, 0, 126, 0, 101, 0, 0, 0, 104, 0, 110, 0, 104, 0, 126, 0, 106, 0, 0, 0, 104, 0, 110, 0, 110, 0, 126, 0, 111, 0, 0, 0, 104, 0, 111, 0, 97, 0, 126, 0, 101, 0, 0, 0, 104, 0, 111, 0, 104, 0, 126, 0, 109, 0, 0, 0, 104, 0, 111, 0, 111, 0, 126, 0, 112, 0, 0, 0, 104, 0, 111, 0, 114, 0, 126, 0, 116, 0, 0, 0, 104, 0, 111, 0, 118, 0, 126, 0, 119, 0, 0, 0, 104, 0, 111, 0, 121, 0, 126, 0, 122, 0, 0, 0, 104, 0, 114, 0, 111, 0, 126, 0, 112, 0, 0, 0, 104, 0, 114, 0, 116, 0, 126, 0, 117, 0, 0, 0, 104, 0, 114, 0, 119, 0, 126, 0, 120, 0, 0, 0, 104, 0, 117, 0, 98, 0, 126, 0, 109, 0, 0, 0, 104, 0, 117, 0, 111, 0, 126, 0, 122, 0, 0, 0, 104, 0, 121, 0, 119, 0, 126, 0, 120, 0, 0, 0, 105, 0, 98, 0, 97, 0, 126, 0, 98, 0, 0, 0, 105, 0, 98, 0, 100, 0, 126, 0, 101, 0, 0, 0, 105, 0, 98, 0, 103, 0, 126, 0, 104, 0, 0, 0, 105, 0, 98, 0, 108, 0, 126, 0, 110, 0, 0, 0, 105, 0, 100, 0, 97, 0, 126, 0, 101, 0, 0, 0, 105, 0, 100, 0, 114, 0, 126, 0, 117, 0, 0, 0, 105, 0, 102, 0, 97, 0, 126, 0, 98, 0, 0, 0, 105, 0, 102, 0, 101, 0, 126, 0, 102, 0, 0, 0, 105, 0, 103, 0, 108, 0, 126, 0, 111, 0, 0, 0, 105, 0, 106, 0, 110, 0, 126, 0, 111, 0, 0, 0, 105, 0, 107, 0, 107, 0, 126, 0, 108, 0, 0, 0, 105, 0, 107, 0, 111, 0, 126, 0, 112, 0, 0, 0, 105, 0, 107, 0, 114, 0, 126, 0, 116, 0, 0, 0, 105, 0, 107, 0, 118, 0, 126, 0, 120, 0, 0, 0, 105, 0, 108, 0, 97, 0, 126, 0, 98, 0, 0, 0, 105, 0, 108, 0, 111, 0, 126, 0, 112, 0, 0, 0, 105, 0, 108, 0, 117, 0, 126, 0, 118, 0, 0, 0, 105, 0, 109, 0, 110, 0, 126, 0, 111, 0, 0, 0, 105, 0, 109, 0, 114, 0, 126, 0, 115, 0, 0, 0, 105, 0, 110, 0, 98, 0, 126, 0, 99, 0, 0, 0, 105, 0, 110, 0, 103, 0, 126, 0, 104, 0, 0, 0, 105, 0, 110, 0, 108, 0, 126, 0, 112, 0, 0, 0, 105, 0, 110, 0, 115, 0, 126, 0, 116, 0, 0, 0, 105, 0, 114, 0, 104, 0, 126, 0, 105, 0, 0, 0, 105, 0, 114, 0, 110, 0, 126, 0, 111, 0, 0, 0, 105, 0, 114, 0, 120, 0, 126, 0, 121, 0, 0, 0, 105, 0, 115, 0, 49, 0, 126, 0, 56, 0, 0, 0, 105, 0, 115, 0, 99, 0, 126, 0, 101, 0, 0, 0, 105, 0, 115, 0, 103, 0, 126, 0, 105, 0, 0, 0, 105, 0, 115, 0, 109, 0, 126, 0, 111, 0, 0, 0, 105, 0, 115, 0, 116, 0, 126, 0, 117, 0, 0, 0, 105, 0, 116, 0, 98, 0, 126, 0, 101, 0, 0, 0, 105, 0, 116, 0, 107, 0, 126, 0, 109, 0, 0, 0, 105, 0, 116, 0, 114, 0, 126, 0, 116, 0, 0, 0, 105, 0, 116, 0, 118, 0, 126, 0, 122, 0, 0, 0, 106, 0, 97, 0, 97, 0, 126, 0, 102, 0, 0, 0, 106, 0, 97, 0, 106, 0, 126, 0, 111, 0, 0, 0, 106, 0, 97, 0, 115, 0, 126, 0, 117, 0, 0, 0, 106, 0, 97, 0, 117, 0, 101, 0, 114, 0, 0, 0, 106, 0, 97, 0, 120, 0, 126, 0, 122, 0, 0, 0, 106, 0, 98, 0, 105, 0, 126, 0, 107, 0, 0, 0, 106, 0, 98, 0, 110, 0, 126, 0, 111, 0, 0, 0, 106, 0, 98, 0, 116, 0, 126, 0, 117, 0, 0, 0, 106, 0, 99, 0, 115, 0, 126, 0, 116, 0, 0, 0, 106, 0, 101, 0, 104, 0, 126, 0, 105, 0, 0, 0, 106, 0, 101, 0, 107, 0, 126, 0, 108, 0, 0, 0, 106, 0, 101, 0, 116, 0, 126, 0, 117, 0, 0, 0, 106, 0, 105, 0, 97, 0, 126, 0, 101, 0, 0, 0, 106, 0, 105, 0, 103, 0, 126, 0, 105, 0, 0, 0, 106, 0, 105, 0, 108, 0, 126, 0, 109, 0, 0, 0, 106, 0, 105, 0, 116, 0, 126, 0, 118, 0, 0, 0, 106, 0, 107, 0, 111, 0, 126, 0, 112, 0, 0, 0, 106, 0, 109, 0, 97, 0, 126, 0, 100, 0, 0, 0, 106, 0, 109, 0, 114, 0, 126, 0, 115, 0, 0, 0, 106, 0, 109, 0, 119, 0, 126, 0, 120, 0, 0, 0, 106, 0, 110, 0, 105, 0, 126, 0, 106, 0, 0, 0, 106, 0, 111, 0, 114, 0, 126, 0, 115, 0, 0, 0, 106, 0, 114, 0, 97, 0, 126, 0, 98, 0, 0, 0, 106, 0, 114, 0, 116, 0, 126, 0, 117, 0, 0, 0, 106, 0, 117, 0, 97, 0, 126, 0, 100, 0, 0, 0, 106, 0, 117, 0, 104, 0, 126, 0, 105, 0, 0, 0, 106, 0, 117, 0, 107, 0, 126, 0, 112, 0, 0, 0, 106, 0, 117, 0, 114, 0, 126, 0, 117, 0, 0, 0, 107, 0, 97, 0, 97, 0, 126, 0, 107, 0, 0, 0, 107, 0, 97, 0, 111, 0, 126, 0, 114, 0, 0, 0, 107, 0, 97, 0, 118, 0, 126, 0, 121, 0, 0, 0, 107, 0, 98, 0, 97, 0, 126, 0, 101, 0, 0, 0, 107, 0, 98, 0, 103, 0, 126, 0, 122, 0, 0, 0, 107, 0, 99, 0, 97, 0, 126, 0, 122, 0, 0, 0, 107, 0, 100, 0, 99, 0, 126, 0, 114, 0, 0, 0, 107, 0, 100, 0, 116, 0, 126, 0, 117, 0, 0, 0, 107, 0, 100, 0, 119, 0, 126, 0, 122, 0, 0, 0, 107, 0, 101, 0, 97, 0, 126, 0, 122, 0, 0, 0, 107, 0, 102, 0, 97, 0, 126, 0, 122, 0, 0, 0, 107, 0, 103, 0, 97, 0, 126, 0, 98, 0, 0, 0, 107, 0, 103, 0, 98, 0, 126, 0, 99, 0, 0, 0, 107, 0, 103, 0, 99, 0, 126, 0, 100, 0, 0, 0, 107, 0, 103, 0, 101, 0, 126, 0, 103, 0, 0, 0, 107, 0, 103, 0, 105, 0, 126, 0, 121, 0, 0, 0, 107, 0, 103, 0, 110, 0, 126, 0, 111, 0, 0, 0, 107, 0, 104, 0, 51, 0, 126, 0, 57, 0, 0, 0, 107, 0, 104, 0, 97, 0, 126, 0, 106, 0, 0, 0, 107, 0, 104, 0, 110, 0, 126, 0, 122, 0, 0, 0, 107, 0, 105, 0, 97, 0, 126, 0, 106, 0, 0, 0, 107, 0, 105, 0, 108, 0, 126, 0, 109, 0, 0, 0, 107, 0, 105, 0, 111, 0, 126, 0, 113, 0, 0, 0, 107, 0, 105, 0, 115, 0, 126, 0, 122, 0, 0, 0, 107, 0, 106, 0, 97, 0, 126, 0, 118, 0, 0, 0, 107, 0, 106, 0, 120, 0, 126, 0, 122, 0, 0, 0, 107, 0, 107, 0, 97, 0, 126, 0, 122, 0, 0, 0, 107, 0, 107, 0, 99, 0, 111, 0, 114, 0, 0, 0, 107, 0, 108, 0, 97, 0, 126, 0, 122, 0, 0, 0, 107, 0, 109, 0, 97, 0, 126, 0, 113, 0, 0, 0, 107, 0, 109, 0, 115, 0, 126, 0, 122, 0, 0, 0, 107, 0, 110, 0, 97, 0, 126, 0, 98, 0, 0, 0, 107, 0, 110, 0, 100, 0, 126, 0, 102, 0, 0, 0, 107, 0, 110, 0, 105, 0, 126, 0, 109, 0, 0, 0, 107, 0, 110, 0, 111, 0, 126, 0, 122, 0, 0, 0, 107, 0, 111, 0, 99, 0, 126, 0, 105, 0, 0, 0, 107, 0, 111, 0, 107, 0, 126, 0, 108, 0, 0, 0, 107, 0, 111, 0, 111, 0, 126, 0, 113, 0, 0, 0, 107, 0, 111, 0, 115, 0, 126, 0, 119, 0, 0, 0, 107, 0, 111, 0, 121, 0, 126, 0, 122, 0, 0, 0, 107, 0, 112, 0, 97, 0, 126, 0, 111, 0, 0, 0, 107, 0, 112, 0, 113, 0, 126, 0, 117, 0, 0, 0, 107, 0, 112, 0, 119, 0, 126, 0, 122, 0, 0, 0, 107, 0, 113, 0, 97, 0, 126, 0, 122, 0, 0, 0, 107, 0, 114, 0, 97, 0, 126, 0, 102, 0, 0, 0, 107, 0, 114, 0, 104, 0, 126, 0, 108, 0, 0, 0, 107, 0, 114, 0, 110, 0, 126, 0, 112, 0, 0, 0, 107, 0, 114, 0, 114, 0, 126, 0, 122, 0, 0, 0, 107, 0, 115, 0, 97, 0, 126, 0, 122, 0, 0, 0, 107, 0, 115, 0, 99, 0, 111, 0, 114, 0, 0, 0, 107, 0, 116, 0, 97, 0, 126, 0, 113, 0, 0, 0, 107, 0, 116, 0, 115, 0, 126, 0, 122, 0, 0, 0, 107, 0, 117, 0, 98, 0, 126, 0, 113, 0, 0, 0, 107, 0, 117, 0, 115, 0, 126, 0, 122, 0, 0, 0, 107, 0, 118, 0, 97, 0, 126, 0, 114, 0, 0, 0, 107, 0, 118, 0, 116, 0, 126, 0, 122, 0, 0, 0, 107, 0, 119, 0, 97, 0, 126, 0, 112, 0, 0, 0, 107, 0, 119, 0, 114, 0, 126, 0, 122, 0, 0, 0, 107, 0, 120, 0, 97, 0, 126, 0, 100, 0, 0, 0, 107, 0, 120, 0, 104, 0, 126, 0, 122, 0, 0, 0, 107, 0, 121, 0, 97, 0, 126, 0, 122, 0, 0, 0, 107, 0, 122, 0, 97, 0, 126, 0, 103, 0, 0, 0, 107, 0, 122, 0, 107, 0, 126, 0, 115, 0, 0, 0, 107, 0, 122, 0, 117, 0, 126, 0, 122, 0, 0, 0, 108, 0, 97, 0, 97, 0, 126, 0, 110, 0, 0, 0, 108, 0, 97, 0, 112, 0, 126, 0, 115, 0, 0, 0, 108, 0, 97, 0, 119, 0, 126, 0, 122, 0, 0, 0, 108, 0, 98, 0, 98, 0, 126, 0, 99, 0, 0, 0, 108, 0, 98, 0, 101, 0, 126, 0, 103, 0, 0, 0, 108, 0, 98, 0, 105, 0, 126, 0, 106, 0, 0, 0, 108, 0, 98, 0, 108, 0, 126, 0, 111, 0, 0, 0, 108, 0, 98, 0, 113, 0, 126, 0, 122, 0, 0, 0, 108, 0, 99, 0, 99, 0, 126, 0, 102, 0, 0, 0, 108, 0, 99, 0, 108, 0, 126, 0, 109, 0, 0, 0, 108, 0, 99, 0, 112, 0, 126, 0, 113, 0, 0, 0, 108, 0, 100, 0, 97, 0, 126, 0, 98, 0, 0, 0, 108, 0, 100, 0, 103, 0, 126, 0, 113, 0, 0, 0, 108, 0, 101, 0, 97, 0, 126, 0, 102, 0, 0, 0, 108, 0, 101, 0, 104, 0, 126, 0, 122, 0, 0, 0, 108, 0, 103, 0, 97, 0, 126, 0, 98, 0, 0, 0, 108, 0, 103, 0, 103, 0, 126, 0, 105, 0, 0, 0, 108, 0, 103, 0, 107, 0, 126, 0, 110, 0, 0, 0, 108, 0, 103, 0, 113, 0, 126, 0, 114, 0, 0, 0, 108, 0, 103, 0, 116, 0, 126, 0, 117, 0, 0, 0, 108, 0, 104, 0, 104, 0, 126, 0, 105, 0, 0, 0, 108, 0, 104, 0, 108, 0, 126, 0, 110, 0, 0, 0, 108, 0, 104, 0, 115, 0, 126, 0, 117, 0, 0, 0, 108, 0, 105, 0, 97, 0, 126, 0, 104, 0, 0, 0, 108, 0, 105, 0, 106, 0, 126, 0, 108, 0, 0, 0, 108, 0, 105, 0, 111, 0, 126, 0, 115, 0, 0, 0, 108, 0, 105, 0, 112, 0, 97, 0, 119, 0, 0, 0, 108, 0, 105, 0, 117, 0, 126, 0, 122, 0, 0, 0, 108, 0, 106, 0, 119, 0, 126, 0, 120, 0, 0, 0, 108, 0, 107, 0, 97, 0, 126, 0, 101, 0, 0, 0, 108, 0, 107, 0, 104, 0, 126, 0, 106, 0, 0, 0, 108, 0, 107, 0, 108, 0, 126, 0, 111, 0, 0, 0, 108, 0, 107, 0, 114, 0, 126, 0, 117, 0, 0, 0, 108, 0, 108, 0, 97, 0, 126, 0, 110, 0, 0, 0, 108, 0, 108, 0, 112, 0, 126, 0, 113, 0, 0, 0, 108, 0, 109, 0, 97, 0, 126, 0, 108, 0, 0, 0, 108, 0, 109, 0, 110, 0, 126, 0, 114, 0, 0, 0, 108, 0, 109, 0, 117, 0, 126, 0, 122, 0, 0, 0, 108, 0, 110, 0, 97, 0, 126, 0, 98, 0, 0, 0, 108, 0, 110, 0, 103, 0, 126, 0, 106, 0, 0, 0, 108, 0, 110, 0, 108, 0, 126, 0, 111, 0, 0, 0, 108, 0, 111, 0, 97, 0, 126, 0, 99, 0, 0, 0, 108, 0, 111, 0, 101, 0, 126, 0, 122, 0, 0, 0, 108, 0, 112, 0, 110, 0, 126, 0, 111, 0, 0, 0, 108, 0, 114, 0, 107, 0, 126, 0, 111, 0, 0, 0, 108, 0, 115, 0, 97, 0, 126, 0, 104, 0, 0, 0, 108, 0, 115, 0, 100, 0, 126, 0, 101, 0, 0, 0, 108, 0, 115, 0, 104, 0, 126, 0, 105, 0, 0, 0, 108, 0, 115, 0, 106, 0, 126, 0, 107, 0, 0, 0, 108, 0, 115, 0, 108, 0, 126, 0, 112, 0, 0, 0, 108, 0, 115, 0, 114, 0, 126, 0, 116, 0, 0, 0, 108, 0, 116, 0, 103, 0, 126, 0, 105, 0, 0, 0, 108, 0, 116, 0, 110, 0, 126, 0, 111, 0, 0, 0, 108, 0, 117, 0, 99, 0, 126, 0, 102, 0, 0, 0, 108, 0, 117, 0, 105, 0, 126, 0, 119, 0, 0, 0, 108, 0, 117, 0, 121, 0, 126, 0, 122, 0, 0, 0, 108, 0, 118, 0, 48, 0, 48, 0, 49, 0, 0, 0, 108, 0, 118, 0, 48, 0, 48, 0, 50, 0, 0, 0, 108, 0, 118, 0, 48, 0, 48, 0, 52, 0, 0, 0, 108, 0, 118, 0, 48, 0, 48, 0, 54, 0, 0, 0, 108, 0, 118, 0, 48, 0, 48, 0, 55, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 48, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 50, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 51, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 53, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 54, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 55, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 51, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 55, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 56, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 49, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 50, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 51, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 53, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 54, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 55, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 48, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 53, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 56, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 50, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 51, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 53, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 54, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 55, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 49, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 51, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 53, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 55, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 56, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 48, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 49, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 50, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 51, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 53, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 55, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 56, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 49, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 51, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 52, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 53, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 55, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 56, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 57, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 48, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 49, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 50, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 51, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 53, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 54, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 55, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 57, 0, 0, 0, 108, 0, 119, 0, 103, 0, 126, 0, 104, 0, 0, 0, 108, 0, 119, 0, 108, 0, 126, 0, 109, 0, 0, 0, 108, 0, 119, 0, 115, 0, 126, 0, 117, 0, 0, 0, 109, 0, 97, 0, 97, 0, 126, 0, 98, 0, 0, 0, 109, 0, 97, 0, 98, 0, 101, 0, 114, 0, 0, 0, 109, 0, 97, 0, 98, 0, 101, 0, 115, 0, 0, 0, 109, 0, 97, 0, 100, 0, 126, 0, 103, 0, 0, 0, 109, 0, 97, 0, 103, 0, 117, 0, 101, 0, 0, 0, 109, 0, 97, 0, 103, 0, 117, 0, 102, 0, 0, 0, 109, 0, 97, 0, 105, 0, 126, 0, 107, 0, 0, 0, 109, 0, 97, 0, 107, 0, 104, 0, 110, 0, 0, 0, 109, 0, 97, 0, 107, 0, 104, 0, 111, 0, 0, 0, 109, 0, 97, 0, 109, 0, 109, 0, 100, 0, 0, 0, 109, 0, 97, 0, 109, 0, 109, 0, 110, 0, 0, 0, 109, 0, 97, 0, 109, 0, 126, 0, 110, 0, 0, 0, 109, 0, 97, 0, 112, 0, 126, 0, 113, 0, 0, 0, 109, 0, 97, 0, 115, 0, 105, 0, 108, 0, 0, 0, 109, 0, 97, 0, 115, 0, 121, 0, 98, 0, 0, 0, 109, 0, 97, 0, 115, 0, 126, 0, 120, 0, 0, 0, 109, 0, 98, 0, 97, 0, 126, 0, 102, 0, 0, 0, 109, 0, 98, 0, 104, 0, 126, 0, 122, 0, 0, 0, 109, 0, 99, 0, 97, 0, 126, 0, 122, 0, 0, 0, 109, 0, 100, 0, 97, 0, 126, 0, 110, 0, 0, 0, 109, 0, 100, 0, 112, 0, 126, 0, 122, 0, 0, 0, 109, 0, 101, 0, 97, 0, 126, 0, 102, 0, 0, 0, 109, 0, 101, 0, 104, 0, 126, 0, 119, 0, 0, 0, 109, 0, 101, 0, 121, 0, 126, 0, 122, 0, 0, 0, 109, 0, 102, 0, 97, 0, 126, 0, 122, 0, 0, 0, 109, 0, 103, 0, 97, 0, 126, 0, 119, 0, 0, 0, 109, 0, 103, 0, 116, 0, 126, 0, 117, 0, 0, 0, 109, 0, 103, 0, 121, 0, 126, 0, 122, 0, 0, 0, 109, 0, 104, 0, 97, 0, 108, 0, 107, 0, 0, 0, 109, 0, 104, 0, 97, 0, 108, 0, 108, 0, 0, 0, 109, 0, 104, 0, 97, 0, 126, 0, 103, 0, 0, 0, 109, 0, 104, 0, 105, 0, 126, 0, 113, 0, 0, 0, 109, 0, 104, 0, 115, 0, 126, 0, 117, 0, 0, 0, 109, 0, 104, 0, 119, 0, 126, 0, 122, 0, 0, 0, 109, 0, 105, 0, 97, 0, 126, 0, 114, 0, 0, 0, 109, 0, 105, 0, 116, 0, 126, 0, 117, 0, 0, 0, 109, 0, 105, 0, 119, 0, 126, 0, 122, 0, 0, 0, 109, 0, 106, 0, 98, 0, 126, 0, 101, 0, 0, 0, 109, 0, 106, 0, 103, 0, 126, 0, 122, 0, 0, 0, 109, 0, 107, 0, 97, 0, 126, 0, 99, 0, 0, 0, 109, 0, 107, 0, 101, 0, 126, 0, 122, 0, 0, 0, 109, 0, 108, 0, 50, 0, 126, 0, 57, 0, 0, 0, 109, 0, 108, 0, 97, 0, 126, 0, 99, 0, 0, 0, 109, 0, 108, 0, 101, 0, 126, 0, 102, 0, 0, 0, 109, 0, 108, 0, 104, 0, 126, 0, 115, 0, 0, 0, 109, 0, 108, 0, 117, 0, 126, 0, 120, 0, 0, 0, 109, 0, 109, 0, 97, 0, 126, 0, 114, 0, 0, 0, 109, 0, 109, 0, 116, 0, 126, 0, 122, 0, 0, 0, 109, 0, 110, 0, 48, 0, 52, 0, 54, 0, 0, 0, 109, 0, 110, 0, 48, 0, 52, 0, 55, 0, 0, 0, 109, 0, 110, 0, 48, 0, 54, 0, 51, 0, 0, 0, 109, 0, 110, 0, 48, 0, 54, 0, 52, 0, 0, 0, 109, 0, 110, 0, 48, 0, 54, 0, 53, 0, 0, 0, 109, 0, 110, 0, 97, 0, 126, 0, 106, 0, 0, 0, 109, 0, 110, 0, 108, 0, 126, 0, 115, 0, 0, 0, 109, 0, 110, 0, 117, 0, 126, 0, 122, 0, 0, 0, 109, 0, 111, 0, 99, 0, 126, 0, 101, 0, 0, 0, 109, 0, 111, 0, 103, 0, 126, 0, 107, 0, 0, 0, 109, 0, 111, 0, 111, 0, 126, 0, 122, 0, 0, 0, 109, 0, 112, 0, 97, 0, 126, 0, 101, 0, 0, 0, 109, 0, 112, 0, 103, 0, 126, 0, 122, 0, 0, 0, 109, 0, 113, 0, 97, 0, 126, 0, 99, 0, 0, 0, 109, 0, 113, 0, 101, 0, 126, 0, 122, 0, 0, 0, 109, 0, 114, 0, 97, 0, 126, 0, 104, 0, 0, 0, 109, 0, 114, 0, 106, 0, 126, 0, 122, 0, 0, 0, 109, 0, 114, 0, 110, 0, 107, 0, 99, 0, 0, 0, 109, 0, 115, 0, 98, 0, 126, 0, 115, 0, 0, 0, 109, 0, 115, 0, 117, 0, 126, 0, 122, 0, 0, 0, 109, 0, 116, 0, 97, 0, 126, 0, 121, 0, 0, 0, 109, 0, 117, 0, 97, 0, 126, 0, 101, 0, 0, 0, 109, 0, 117, 0, 103, 0, 126, 0, 107, 0, 0, 0, 109, 0, 117, 0, 109, 0, 126, 0, 111, 0, 0, 0, 109, 0, 117, 0, 113, 0, 126, 0, 118, 0, 0, 0, 109, 0, 117, 0, 120, 0, 126, 0, 122, 0, 0, 0, 109, 0, 118, 0, 97, 0, 126, 0, 98, 0, 0, 0, 109, 0, 118, 0, 100, 0, 126, 0, 105, 0, 0, 0, 109, 0, 118, 0, 107, 0, 126, 0, 122, 0, 0, 0, 109, 0, 119, 0, 97, 0, 126, 0, 99, 0, 0, 0, 109, 0, 119, 0, 101, 0, 126, 0, 105, 0, 0, 0, 109, 0, 119, 0, 107, 0, 126, 0, 119, 0, 0, 0, 109, 0, 119, 0, 120, 0, 126, 0, 121, 0, 0, 0, 109, 0, 120, 0, 97, 0, 126, 0, 122, 0, 0, 0, 109, 0, 120, 0, 100, 0, 105, 0, 102, 0, 0, 0, 109, 0, 121, 0, 98, 0, 126, 0, 99, 0, 0, 0, 109, 0, 121, 0, 101, 0, 126, 0, 104, 0, 0, 0, 109, 0, 121, 0, 106, 0, 126, 0, 112, 0, 0, 0, 109, 0, 121, 0, 114, 0, 126, 0, 115, 0, 0, 0, 109, 0, 121, 0, 117, 0, 126, 0, 122, 0, 0, 0, 109, 0, 122, 0, 97, 0, 126, 0, 98, 0, 0, 0, 109, 0, 122, 0, 97, 0, 126, 0, 101, 0, 0, 0, 109, 0, 122, 0, 103, 0, 126, 0, 122, 0, 0, 0, 109, 0, 122, 0, 112, 0, 126, 0, 113, 0, 0, 0, 109, 0, 122, 0, 115, 0, 126, 0, 116, 0, 0, 0, 110, 0, 97, 0, 97, 0, 126, 0, 99, 0, 0, 0, 110, 0, 97, 0, 101, 0, 126, 0, 116, 0, 0, 0, 110, 0, 97, 0, 119, 0, 126, 0, 122, 0, 0, 0, 110, 0, 98, 0, 97, 0, 126, 0, 101, 0, 0, 0, 110, 0, 98, 0, 103, 0, 126, 0, 107, 0, 0, 0, 110, 0, 98, 0, 109, 0, 126, 0, 119, 0, 0, 0, 110, 0, 99, 0, 97, 0, 126, 0, 111, 0, 0, 0, 110, 0, 99, 0, 113, 0, 126, 0, 117, 0, 0, 0, 110, 0, 100, 0, 97, 0, 126, 0, 100, 0, 0, 0, 110, 0, 100, 0, 102, 0, 126, 0, 110, 0, 0, 0, 110, 0, 100, 0, 112, 0, 126, 0, 122, 0, 0, 0, 110, 0, 101, 0, 49, 0, 126, 0, 56, 0, 0, 0, 110, 0, 101, 0, 97, 0, 126, 0, 107, 0, 0, 0, 110, 0, 101, 0, 100, 0, 105, 0, 115, 0, 0, 0, 110, 0, 101, 0, 109, 0, 126, 0, 111, 0, 0, 0, 110, 0, 101, 0, 113, 0, 126, 0, 122, 0, 0, 0, 110, 0, 103, 0, 97, 0, 126, 0, 122, 0, 0, 0, 110, 0, 104, 0, 97, 0, 126, 0, 105, 0, 0, 0, 110, 0, 104, 0, 109, 0, 126, 0, 114, 0, 0, 0, 110, 0, 104, 0, 116, 0, 126, 0, 122, 0, 0, 0, 110, 0, 105, 0, 97, 0, 126, 0, 111, 0, 0, 0, 110, 0, 105, 0, 113, 0, 126, 0, 122, 0, 0, 0, 110, 0, 106, 0, 97, 0, 126, 0, 98, 0, 0, 0, 110, 0, 106, 0, 104, 0, 126, 0, 106, 0, 0, 0, 110, 0, 106, 0, 105, 0, 118, 0, 97, 0, 0, 0, 110, 0, 106, 0, 108, 0, 126, 0, 111, 0, 0, 0, 110, 0, 106, 0, 114, 0, 126, 0, 117, 0, 0, 0, 110, 0, 106, 0, 120, 0, 126, 0, 122, 0, 0, 0, 110, 0, 107, 0, 97, 0, 126, 0, 107, 0, 0, 0, 110, 0, 107, 0, 109, 0, 126, 0, 120, 0, 0, 0, 110, 0, 108, 0, 98, 0, 113, 0, 49, 0, 0, 0, 110, 0, 108, 0, 98, 0, 113, 0, 50, 0, 0, 0, 110, 0, 108, 0, 98, 0, 113, 0, 51, 0, 0, 0, 110, 0, 108, 0, 105, 0, 126, 0, 109, 0, 0, 0, 110, 0, 108, 0, 117, 0, 126, 0, 122, 0, 0, 0, 110, 0, 109, 0, 97, 0, 126, 0, 122, 0, 0, 0, 110, 0, 110, 0, 97, 0, 126, 0, 110, 0, 0, 0, 110, 0, 110, 0, 112, 0, 126, 0, 114, 0, 0, 0, 110, 0, 110, 0, 116, 0, 126, 0, 119, 0, 0, 0, 110, 0, 110, 0, 121, 0, 126, 0, 122, 0, 0, 0, 110, 0, 111, 0, 99, 0, 126, 0, 110, 0, 0, 0, 110, 0, 111, 0, 112, 0, 126, 0, 113, 0, 0, 0, 110, 0, 111, 0, 115, 0, 126, 0, 119, 0, 0, 0, 110, 0, 111, 0, 121, 0, 126, 0, 122, 0, 0, 0, 110, 0, 112, 0, 49, 0, 126, 0, 53, 0, 0, 0, 110, 0, 112, 0, 97, 0, 126, 0, 98, 0, 0, 0, 110, 0, 112, 0, 103, 0, 126, 0, 104, 0, 0, 0, 110, 0, 112, 0, 110, 0, 126, 0, 111, 0, 0, 0, 110, 0, 112, 0, 120, 0, 126, 0, 121, 0, 0, 0, 110, 0, 113, 0, 107, 0, 126, 0, 111, 0, 0, 0, 110, 0, 114, 0, 97, 0, 126, 0, 99, 0, 0, 0, 110, 0, 114, 0, 101, 0, 126, 0, 103, 0, 0, 0, 110, 0, 114, 0, 107, 0, 126, 0, 110, 0, 0, 0, 110, 0, 114, 0, 116, 0, 126, 0, 117, 0, 0, 0, 110, 0, 115, 0, 99, 0, 126, 0, 105, 0, 0, 0, 110, 0, 115, 0, 107, 0, 126, 0, 122, 0, 0, 0, 110, 0, 116, 0, 100, 0, 126, 0, 101, 0, 0, 0, 110, 0, 116, 0, 105, 0, 126, 0, 107, 0, 0, 0, 110, 0, 116, 0, 111, 0, 126, 0, 112, 0, 0, 0, 110, 0, 116, 0, 119, 0, 126, 0, 122, 0, 0, 0, 110, 0, 117, 0, 97, 0, 126, 0, 122, 0, 0, 0, 110, 0, 117, 0, 108, 0, 105, 0, 107, 0, 0, 0, 110, 0, 119, 0, 97, 0, 126, 0, 99, 0, 0, 0, 110, 0, 119, 0, 120, 0, 126, 0, 121, 0, 0, 0, 110, 0, 120, 0, 100, 0, 126, 0, 101, 0, 0, 0, 110, 0, 120, 0, 107, 0, 126, 0, 111, 0, 0, 0, 110, 0, 120, 0, 113, 0, 126, 0, 114, 0, 0, 0, 110, 0, 121, 0, 98, 0, 126, 0, 121, 0, 0, 0, 110, 0, 122, 0, 97, 0, 126, 0, 98, 0, 0, 0, 110, 0, 122, 0, 121, 0, 126, 0, 122, 0, 0, 0, 111, 0, 98, 0, 107, 0, 126, 0, 109, 0, 0, 0, 111, 0, 98, 0, 116, 0, 126, 0, 117, 0, 0, 0, 111, 0, 100, 0, 116, 0, 126, 0, 117, 0, 0, 0, 111, 0, 103, 0, 98, 0, 126, 0, 99, 0, 0, 0, 111, 0, 104, 0, 116, 0, 126, 0, 117, 0, 0, 0, 111, 0, 106, 0, 98, 0, 126, 0, 99, 0, 0, 0, 111, 0, 106, 0, 118, 0, 126, 0, 119, 0, 0, 0, 111, 0, 107, 0, 97, 0, 126, 0, 98, 0, 0, 0, 111, 0, 107, 0, 100, 0, 126, 0, 101, 0, 0, 0, 111, 0, 107, 0, 103, 0, 126, 0, 111, 0, 0, 0, 111, 0, 107, 0, 114, 0, 126, 0, 115, 0, 0, 0, 111, 0, 107, 0, 117, 0, 126, 0, 118, 0, 0, 0, 111, 0, 108, 0, 100, 0, 126, 0, 101, 0, 0, 0, 111, 0, 108, 0, 116, 0, 126, 0, 117, 0, 0, 0, 111, 0, 109, 0, 97, 0, 126, 0, 99, 0, 0, 0, 111, 0, 109, 0, 107, 0, 126, 0, 108, 0, 0, 0, 111, 0, 109, 0, 110, 0, 126, 0, 114, 0, 0, 0, 111, 0, 109, 0, 116, 0, 126, 0, 120, 0, 0, 0, 111, 0, 110, 0, 97, 0, 126, 0, 98, 0, 0, 0, 111, 0, 110, 0, 105, 0, 126, 0, 107, 0, 0, 0, 111, 0, 110, 0, 110, 0, 126, 0, 112, 0, 0, 0, 111, 0, 110, 0, 114, 0, 126, 0, 117, 0, 0, 0, 111, 0, 110, 0, 119, 0, 126, 0, 120, 0, 0, 0, 111, 0, 111, 0, 114, 0, 126, 0, 115, 0, 0, 0, 111, 0, 114, 0, 103, 0, 126, 0, 104, 0, 0, 0, 111, 0, 114, 0, 110, 0, 126, 0, 111, 0, 0, 0, 111, 0, 114, 0, 114, 0, 126, 0, 120, 0, 0, 0, 111, 0, 115, 0, 111, 0, 126, 0, 112, 0, 0, 0, 111, 0, 115, 0, 116, 0, 126, 0, 117, 0, 0, 0, 111, 0, 116, 0, 97, 0, 126, 0, 98, 0, 0, 0, 111, 0, 116, 0, 100, 0, 126, 0, 101, 0, 0, 0, 111, 0, 116, 0, 107, 0, 126, 0, 111, 0, 0, 0, 111, 0, 116, 0, 113, 0, 126, 0, 117, 0, 0, 0, 111, 0, 116, 0, 119, 0, 126, 0, 122, 0, 0, 0, 111, 0, 117, 0, 97, 0, 126, 0, 98, 0, 0, 0, 112, 0, 97, 0, 50, 0, 126, 0, 57, 0, 0, 0, 112, 0, 97, 0, 97, 0, 126, 0, 105, 0, 0, 0, 112, 0, 97, 0, 107, 0, 126, 0, 109, 0, 0, 0, 112, 0, 97, 0, 111, 0, 126, 0, 122, 0, 0, 0, 112, 0, 98, 0, 98, 0, 126, 0, 99, 0, 0, 0, 112, 0, 98, 0, 101, 0, 126, 0, 105, 0, 0, 0, 112, 0, 98, 0, 108, 0, 126, 0, 112, 0, 0, 0, 112, 0, 98, 0, 114, 0, 126, 0, 116, 0, 0, 0, 112, 0, 99, 0, 97, 0, 126, 0, 110, 0, 0, 0, 112, 0, 100, 0, 110, 0, 126, 0, 111, 0, 0, 0, 112, 0, 100, 0, 116, 0, 126, 0, 117, 0, 0, 0, 112, 0, 101, 0, 97, 0, 126, 0, 98, 0, 0, 0, 112, 0, 101, 0, 100, 0, 126, 0, 109, 0, 0, 0, 112, 0, 101, 0, 108, 0, 97, 0, 108, 0, 0, 0, 112, 0, 101, 0, 108, 0, 97, 0, 109, 0, 0, 0, 112, 0, 101, 0, 111, 0, 126, 0, 113, 0, 0, 0, 112, 0, 101, 0, 120, 0, 126, 0, 122, 0, 0, 0, 112, 0, 103, 0, 107, 0, 126, 0, 108, 0, 0, 0, 112, 0, 103, 0, 109, 0, 112, 0, 108, 0, 0, 0, 112, 0, 103, 0, 109, 0, 112, 0, 109, 0, 0, 0, 112, 0, 104, 0, 98, 0, 117, 0, 107, 0, 0, 0, 112, 0, 104, 0, 98, 0, 117, 0, 108, 0, 0, 0, 112, 0, 104, 0, 99, 0, 97, 0, 109, 0, 0, 0, 112, 0, 104, 0, 99, 0, 97, 0, 110, 0, 0, 0, 112, 0, 104, 0, 99, 0, 97, 0, 115, 0, 0, 0, 112, 0, 104, 0, 99, 0, 97, 0, 116, 0, 0, 0, 112, 0, 104, 0, 103, 0, 126, 0, 105, 0, 0, 0, 112, 0, 104, 0, 107, 0, 126, 0, 111, 0, 0, 0, 112, 0, 104, 0, 112, 0, 97, 0, 109, 0, 0, 0, 112, 0, 104, 0, 112, 0, 97, 0, 110, 0, 0, 0, 112, 0, 104, 0, 113, 0, 126, 0, 114, 0, 0, 0, 112, 0, 104, 0, 116, 0, 126, 0, 119, 0, 0, 0, 112, 0, 105, 0, 97, 0, 126, 0, 106, 0, 0, 0, 112, 0, 105, 0, 108, 0, 126, 0, 112, 0, 0, 0, 112, 0, 105, 0, 114, 0, 126, 0, 122, 0, 0, 0, 112, 0, 107, 0, 97, 0, 126, 0, 99, 0, 0, 0, 112, 0, 107, 0, 103, 0, 126, 0, 104, 0, 0, 0, 112, 0, 107, 0, 110, 0, 126, 0, 112, 0, 0, 0, 112, 0, 107, 0, 114, 0, 126, 0, 117, 0, 0, 0, 112, 0, 108, 0, 97, 0, 126, 0, 104, 0, 0, 0, 112, 0, 108, 0, 106, 0, 126, 0, 108, 0, 0, 0, 112, 0, 108, 0, 110, 0, 126, 0, 115, 0, 0, 0, 112, 0, 108, 0, 117, 0, 126, 0, 119, 0, 0, 0, 112, 0, 108, 0, 121, 0, 126, 0, 122, 0, 0, 0, 112, 0, 109, 0, 97, 0, 126, 0, 98, 0, 0, 0, 112, 0, 109, 0, 100, 0, 126, 0, 102, 0, 0, 0, 112, 0, 109, 0, 104, 0, 126, 0, 111, 0, 0, 0, 112, 0, 109, 0, 113, 0, 126, 0, 116, 0, 0, 0, 112, 0, 109, 0, 119, 0, 126, 0, 122, 0, 0, 0, 112, 0, 110, 0, 99, 0, 126, 0, 101, 0, 0, 0, 112, 0, 110, 0, 103, 0, 126, 0, 122, 0, 0, 0, 112, 0, 111, 0, 101, 0, 126, 0, 105, 0, 0, 0, 112, 0, 111, 0, 109, 0, 126, 0, 113, 0, 0, 0, 112, 0, 111, 0, 115, 0, 126, 0, 116, 0, 0, 0, 112, 0, 111, 0, 118, 0, 126, 0, 122, 0, 0, 0, 112, 0, 112, 0, 107, 0, 126, 0, 113, 0, 0, 0, 112, 0, 112, 0, 115, 0, 126, 0, 117, 0, 0, 0, 112, 0, 114, 0, 99, 0, 126, 0, 105, 0, 0, 0, 112, 0, 114, 0, 107, 0, 126, 0, 114, 0, 0, 0, 112, 0, 114, 0, 116, 0, 126, 0, 117, 0, 0, 0, 112, 0, 114, 0, 119, 0, 126, 0, 120, 0, 0, 0, 112, 0, 115, 0, 99, 0, 126, 0, 101, 0, 0, 0, 112, 0, 115, 0, 103, 0, 126, 0, 105, 0, 0, 0, 112, 0, 115, 0, 106, 0, 101, 0, 109, 0, 0, 0, 112, 0, 115, 0, 108, 0, 126, 0, 117, 0, 0, 0, 112, 0, 116, 0, 104, 0, 126, 0, 105, 0, 0, 0, 112, 0, 116, 0, 110, 0, 126, 0, 114, 0, 0, 0, 112, 0, 116, 0, 116, 0, 126, 0, 119, 0, 0, 0, 112, 0, 117, 0, 97, 0, 126, 0, 103, 0, 0, 0, 112, 0, 117, 0, 105, 0, 126, 0, 106, 0, 0, 0, 112, 0, 117, 0, 111, 0, 126, 0, 114, 0, 0, 0, 112, 0, 117, 0, 116, 0, 101, 0, 114, 0, 0, 0, 112, 0, 117, 0, 116, 0, 126, 0, 117, 0, 0, 0, 112, 0, 117, 0, 119, 0, 126, 0, 121, 0, 0, 0, 112, 0, 119, 0, 50, 0, 50, 0, 55, 0, 0, 0, 112, 0, 119, 0, 50, 0, 50, 0, 56, 0, 0, 0, 112, 0, 119, 0, 97, 0, 126, 0, 98, 0, 0, 0, 112, 0, 119, 0, 109, 0, 126, 0, 111, 0, 0, 0, 112, 0, 121, 0, 50, 0, 126, 0, 57, 0, 0, 0, 112, 0, 121, 0, 109, 0, 126, 0, 110, 0, 0, 0, 112, 0, 121, 0, 120, 0, 126, 0, 121, 0, 0, 0, 113, 0, 97, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 98, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 99, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 100, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 101, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 102, 0, 97, 0, 126, 0, 121, 0, 0, 0, 113, 0, 103, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 104, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 105, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 106, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 107, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 108, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 109, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 110, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 111, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 112, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 113, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 114, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 115, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 116, 0, 97, 0, 126, 0, 122, 0, 0, 0, 113, 0, 117, 0, 97, 0, 126, 0, 100, 0, 0, 0, 113, 0, 117, 0, 102, 0, 126, 0, 105, 0, 0, 0, 113, 0, 117, 0, 107, 0, 126, 0, 110, 0, 0, 0, 113, 0, 117, 0, 112, 0, 126, 0, 115, 0, 0, 0, 113, 0, 117, 0, 118, 0, 126, 0, 121, 0, 0, 0, 113, 0, 118, 0, 104, 0, 126, 0, 106, 0, 0, 0, 113, 0, 118, 0, 108, 0, 126, 0, 112, 0, 0, 0, 113, 0, 118, 0, 121, 0, 126, 0, 122, 0, 0, 0, 113, 0, 119, 0, 115, 0, 126, 0, 116, 0, 0, 0, 113, 0, 120, 0, 110, 0, 126, 0, 117, 0, 0, 0, 114, 0, 97, 0, 97, 0, 126, 0, 100, 0, 0, 0, 114, 0, 97, 0, 102, 0, 126, 0, 122, 0, 0, 0, 114, 0, 98, 0, 107, 0, 126, 0, 108, 0, 0, 0, 114, 0, 101, 0, 97, 0, 126, 0, 98, 0, 0, 0, 114, 0, 101, 0, 105, 0, 126, 0, 106, 0, 0, 0, 114, 0, 101, 0, 108, 0, 126, 0, 110, 0, 0, 0, 114, 0, 101, 0, 114, 0, 126, 0, 116, 0, 0, 0, 114, 0, 103, 0, 114, 0, 126, 0, 115, 0, 0, 0, 114, 0, 105, 0, 103, 0, 105, 0, 107, 0, 0, 0, 114, 0, 105, 0, 108, 0, 126, 0, 110, 0, 0, 0, 114, 0, 105, 0, 116, 0, 126, 0, 117, 0, 0, 0, 114, 0, 107, 0, 97, 0, 126, 0, 98, 0, 0, 0, 114, 0, 107, 0, 104, 0, 126, 0, 105, 0, 0, 0, 114, 0, 109, 0, 97, 0, 126, 0, 105, 0, 0, 0, 114, 0, 109, 0, 107, 0, 126, 0, 113, 0, 0, 0, 114, 0, 109, 0, 115, 0, 126, 0, 120, 0, 0, 0, 114, 0, 111, 0, 97, 0, 126, 0, 103, 0, 0, 0, 114, 0, 111, 0, 108, 0, 126, 0, 109, 0, 0, 0, 114, 0, 111, 0, 111, 0, 126, 0, 112, 0, 0, 0, 114, 0, 111, 0, 122, 0, 97, 0, 106, 0, 0, 0, 114, 0, 115, 0, 108, 0, 126, 0, 109, 0, 0, 0, 114, 0, 117, 0, 98, 0, 126, 0, 99, 0, 0, 0, 114, 0, 117, 0, 101, 0, 126, 0, 105, 0, 0, 0, 114, 0, 117, 0, 109, 0, 103, 0, 114, 0, 0, 0, 114, 0, 117, 0, 111, 0, 126, 0, 113, 0, 0, 0, 114, 0, 117, 0, 116, 0, 126, 0, 117, 0, 0, 0, 114, 0, 117, 0, 121, 0, 126, 0, 122, 0, 0, 0, 115, 0, 97, 0, 97, 0, 126, 0, 102, 0, 0, 0, 115, 0, 97, 0, 104, 0, 126, 0, 109, 0, 0, 0, 115, 0, 97, 0, 113, 0, 126, 0, 122, 0, 0, 0, 115, 0, 98, 0, 97, 0, 126, 0, 122, 0, 0, 0, 115, 0, 99, 0, 101, 0, 126, 0, 105, 0, 0, 0, 115, 0, 99, 0, 107, 0, 126, 0, 108, 0, 0, 0, 115, 0, 99, 0, 110, 0, 126, 0, 113, 0, 0, 0, 115, 0, 99, 0, 115, 0, 126, 0, 120, 0, 0, 0, 115, 0, 100, 0, 97, 0, 126, 0, 99, 0, 0, 0, 115, 0, 100, 0, 101, 0, 126, 0, 104, 0, 0, 0, 115, 0, 100, 0, 106, 0, 126, 0, 112, 0, 0, 0, 115, 0, 100, 0, 114, 0, 126, 0, 118, 0, 0, 0, 115, 0, 101, 0, 97, 0, 126, 0, 119, 0, 0, 0, 115, 0, 101, 0, 99, 0, 126, 0, 105, 0, 0, 0, 115, 0, 101, 0, 109, 0, 126, 0, 111, 0, 0, 0, 115, 0, 101, 0, 115, 0, 126, 0, 117, 0, 0, 0, 115, 0, 101, 0, 119, 0, 126, 0, 122, 0, 0, 0, 115, 0, 101, 0, 121, 0, 126, 0, 122, 0, 0, 0, 115, 0, 103, 0, 97, 0, 126, 0, 101, 0, 0, 0, 115, 0, 103, 0, 103, 0, 126, 0, 107, 0, 0, 0, 115, 0, 103, 0, 109, 0, 126, 0, 110, 0, 0, 0, 115, 0, 103, 0, 114, 0, 126, 0, 117, 0, 0, 0, 115, 0, 103, 0, 119, 0, 126, 0, 122, 0, 0, 0, 115, 0, 104, 0, 97, 0, 126, 0, 101, 0, 0, 0, 115, 0, 104, 0, 103, 0, 126, 0, 122, 0, 0, 0, 115, 0, 105, 0, 48, 0, 48, 0, 49, 0, 0, 0, 115, 0, 105, 0, 48, 0, 48, 0, 50, 0, 0, 0, 115, 0, 105, 0, 48, 0, 48, 0, 51, 0, 0, 0, 115, 0, 105, 0, 48, 0, 48, 0, 52, 0, 0, 0, 115, 0, 105, 0, 48, 0, 48, 0, 53, 0, 0, 0, 115, 0, 105, 0, 48, 0, 48, 0, 54, 0, 0, 0, 115, 0, 105, 0, 48, 0, 48, 0, 55, 0, 0, 0, 115, 0, 105, 0, 48, 0, 48, 0, 56, 0, 0, 0, 115, 0, 105, 0, 48, 0, 48, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 48, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 49, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 50, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 51, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 52, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 53, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 54, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 55, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 56, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 48, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 49, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 50, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 51, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 52, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 53, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 54, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 55, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 56, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 48, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 49, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 50, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 51, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 52, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 53, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 54, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 55, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 56, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 48, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 49, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 50, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 51, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 52, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 53, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 54, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 55, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 56, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 48, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 49, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 50, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 51, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 52, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 53, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 54, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 55, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 56, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 48, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 49, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 50, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 51, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 52, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 53, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 54, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 55, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 56, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 48, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 49, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 50, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 51, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 52, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 53, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 54, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 55, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 56, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 48, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 49, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 50, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 51, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 52, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 53, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 54, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 55, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 56, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 48, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 49, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 50, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 51, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 52, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 53, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 54, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 55, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 56, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 48, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 49, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 50, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 51, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 53, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 54, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 55, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 48, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 49, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 51, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 53, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 55, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 49, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 50, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 51, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 53, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 54, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 55, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 48, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 49, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 51, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 53, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 55, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 49, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 54, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 55, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 48, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 51, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 55, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 48, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 49, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 51, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 53, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 55, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 49, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 50, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 51, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 54, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 55, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 49, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 50, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 51, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 53, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 54, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 55, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 48, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 51, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 53, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 55, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 56, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 57, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 48, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 49, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 50, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 51, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 52, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 53, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 54, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 55, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 56, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 57, 0, 0, 0, 115, 0, 105, 0, 50, 0, 49, 0, 48, 0, 0, 0, 115, 0, 105, 0, 50, 0, 49, 0, 49, 0, 0, 0, 115, 0, 105, 0, 50, 0, 49, 0, 50, 0, 0, 0, 115, 0, 105, 0, 50, 0, 49, 0, 51, 0, 0, 0, 115, 0, 105, 0, 97, 0, 126, 0, 98, 0, 0, 0, 115, 0, 105, 0, 100, 0, 126, 0, 109, 0, 0, 0, 115, 0, 105, 0, 111, 0, 126, 0, 122, 0, 0, 0, 115, 0, 106, 0, 97, 0, 126, 0, 98, 0, 0, 0, 115, 0, 106, 0, 100, 0, 126, 0, 101, 0, 0, 0, 115, 0, 106, 0, 107, 0, 126, 0, 112, 0, 0, 0, 115, 0, 106, 0, 114, 0, 126, 0, 117, 0, 0, 0, 115, 0, 107, 0, 97, 0, 126, 0, 106, 0, 0, 0, 115, 0, 107, 0, 109, 0, 126, 0, 122, 0, 0, 0, 115, 0, 108, 0, 99, 0, 126, 0, 106, 0, 0, 0, 115, 0, 108, 0, 108, 0, 126, 0, 110, 0, 0, 0, 115, 0, 108, 0, 112, 0, 126, 0, 117, 0, 0, 0, 115, 0, 108, 0, 119, 0, 126, 0, 122, 0, 0, 0, 115, 0, 109, 0, 97, 0, 126, 0, 100, 0, 0, 0, 115, 0, 109, 0, 102, 0, 126, 0, 110, 0, 0, 0, 115, 0, 109, 0, 112, 0, 126, 0, 122, 0, 0, 0, 115, 0, 110, 0, 98, 0, 126, 0, 99, 0, 0, 0, 115, 0, 110, 0, 101, 0, 126, 0, 103, 0, 0, 0, 115, 0, 110, 0, 105, 0, 126, 0, 115, 0, 0, 0, 115, 0, 110, 0, 117, 0, 126, 0, 122, 0, 0, 0, 115, 0, 111, 0, 97, 0, 126, 0, 101, 0, 0, 0, 115, 0, 111, 0, 103, 0, 126, 0, 108, 0, 0, 0, 115, 0, 111, 0, 108, 0, 98, 0, 97, 0, 0, 0, 115, 0, 111, 0, 110, 0, 126, 0, 115, 0, 0, 0, 115, 0, 111, 0, 117, 0, 126, 0, 122, 0, 0, 0, 115, 0, 112, 0, 98, 0, 126, 0, 101, 0, 0, 0, 115, 0, 112, 0, 107, 0, 126, 0, 118, 0, 0, 0, 115, 0, 113, 0, 106, 0, 126, 0, 107, 0, 0, 0, 115, 0, 113, 0, 109, 0, 126, 0, 111, 0, 0, 0, 115, 0, 113, 0, 113, 0, 126, 0, 117, 0, 0, 0, 115, 0, 114, 0, 97, 0, 126, 0, 98, 0, 0, 0, 115, 0, 114, 0, 101, 0, 126, 0, 105, 0, 0, 0, 115, 0, 114, 0, 107, 0, 126, 0, 111, 0, 0, 0, 115, 0, 114, 0, 113, 0, 126, 0, 122, 0, 0, 0, 115, 0, 115, 0, 97, 0, 126, 0, 118, 0, 0, 0, 115, 0, 115, 0, 120, 0, 126, 0, 122, 0, 0, 0, 115, 0, 116, 0, 97, 0, 126, 0, 98, 0, 0, 0, 115, 0, 116, 0, 100, 0, 126, 0, 119, 0, 0, 0, 115, 0, 117, 0, 97, 0, 126, 0, 99, 0, 0, 0, 115, 0, 117, 0, 105, 0, 126, 0, 107, 0, 0, 0, 115, 0, 117, 0, 108, 0, 126, 0, 109, 0, 0, 0, 115, 0, 117, 0, 113, 0, 126, 0, 116, 0, 0, 0, 115, 0, 117, 0, 118, 0, 126, 0, 122, 0, 0, 0, 115, 0, 118, 0, 97, 0, 126, 0, 99, 0, 0, 0, 115, 0, 119, 0, 102, 0, 126, 0, 103, 0, 0, 0, 115, 0, 119, 0, 105, 0, 126, 0, 121, 0, 0, 0, 115, 0, 120, 0, 98, 0, 126, 0, 99, 0, 0, 0, 115, 0, 120, 0, 107, 0, 126, 0, 111, 0, 0, 0, 115, 0, 120, 0, 114, 0, 126, 0, 115, 0, 0, 0, 115, 0, 121, 0, 97, 0, 126, 0, 100, 0, 0, 0, 115, 0, 121, 0, 107, 0, 126, 0, 111, 0, 0, 0, 115, 0, 121, 0, 114, 0, 126, 0, 115, 0, 0, 0, 115, 0, 121, 0, 119, 0, 126, 0, 121, 0, 0, 0, 115, 0, 122, 0, 97, 0, 126, 0, 101, 0, 0, 0, 115, 0, 122, 0, 118, 0, 126, 0, 119, 0, 0, 0, 116, 0, 97, 0, 97, 0, 126, 0, 103, 0, 0, 0, 116, 0, 97, 0, 105, 0, 126, 0, 108, 0, 0, 0, 116, 0, 97, 0, 110, 0, 126, 0, 115, 0, 0, 0, 116, 0, 97, 0, 117, 0, 126, 0, 122, 0, 0, 0, 116, 0, 98, 0, 97, 0, 126, 0, 122, 0, 0, 0, 116, 0, 99, 0, 97, 0, 126, 0, 105, 0, 0, 0, 116, 0, 99, 0, 107, 0, 126, 0, 113, 0, 0, 0, 116, 0, 99, 0, 115, 0, 126, 0, 117, 0, 0, 0, 116, 0, 99, 0, 119, 0, 126, 0, 122, 0, 0, 0, 116, 0, 100, 0, 97, 0, 126, 0, 111, 0, 0, 0, 116, 0, 100, 0, 113, 0, 126, 0, 116, 0, 0, 0, 116, 0, 100, 0, 120, 0, 126, 0, 121, 0, 0, 0, 116, 0, 101, 0, 97, 0, 126, 0, 105, 0, 0, 0, 116, 0, 101, 0, 109, 0, 126, 0, 122, 0, 0, 0, 116, 0, 102, 0, 110, 0, 126, 0, 111, 0, 0, 0, 116, 0, 103, 0, 97, 0, 126, 0, 102, 0, 0, 0, 116, 0, 103, 0, 104, 0, 126, 0, 106, 0, 0, 0, 116, 0, 103, 0, 110, 0, 126, 0, 122, 0, 0, 0, 116, 0, 104, 0, 100, 0, 126, 0, 102, 0, 0, 0, 116, 0, 104, 0, 104, 0, 126, 0, 105, 0, 0, 0, 116, 0, 104, 0, 107, 0, 126, 0, 110, 0, 0, 0, 116, 0, 104, 0, 112, 0, 126, 0, 119, 0, 0, 0, 116, 0, 104, 0, 121, 0, 126, 0, 122, 0, 0, 0, 116, 0, 105, 0, 100, 0, 126, 0, 101, 0, 0, 0, 116, 0, 105, 0, 102, 0, 126, 0, 113, 0, 0, 0, 116, 0, 105, 0, 115, 0, 126, 0, 122, 0, 0, 0, 116, 0, 106, 0, 105, 0, 126, 0, 106, 0, 0, 0, 116, 0, 106, 0, 108, 0, 126, 0, 112, 0, 0, 0, 116, 0, 107, 0, 97, 0, 126, 0, 98, 0, 0, 0, 116, 0, 107, 0, 100, 0, 126, 0, 103, 0, 0, 0, 116, 0, 107, 0, 108, 0, 126, 0, 110, 0, 0, 0, 116, 0, 107, 0, 112, 0, 126, 0, 120, 0, 0, 0, 116, 0, 108, 0, 97, 0, 126, 0, 100, 0, 0, 0, 116, 0, 108, 0, 102, 0, 126, 0, 118, 0, 0, 0, 116, 0, 108, 0, 120, 0, 126, 0, 121, 0, 0, 0, 116, 0, 109, 0, 97, 0, 126, 0, 98, 0, 0, 0, 116, 0, 109, 0, 97, 0, 126, 0, 111, 0, 0, 0, 116, 0, 109, 0, 108, 0, 126, 0, 109, 0, 0, 0, 116, 0, 109, 0, 113, 0, 126, 0, 119, 0, 0, 0, 116, 0, 109, 0, 121, 0, 126, 0, 122, 0, 0, 0, 116, 0, 110, 0, 97, 0, 126, 0, 100, 0, 0, 0, 116, 0, 110, 0, 101, 0, 126, 0, 102, 0, 0, 0, 116, 0, 110, 0, 103, 0, 126, 0, 105, 0, 0, 0, 116, 0, 110, 0, 107, 0, 126, 0, 122, 0, 0, 0, 116, 0, 111, 0, 98, 0, 126, 0, 100, 0, 0, 0, 116, 0, 111, 0, 102, 0, 126, 0, 106, 0, 0, 0, 116, 0, 111, 0, 108, 0, 126, 0, 109, 0, 0, 0, 116, 0, 111, 0, 111, 0, 126, 0, 115, 0, 0, 0, 116, 0, 111, 0, 117, 0, 126, 0, 122, 0, 0, 0, 116, 0, 112, 0, 101, 0, 126, 0, 103, 0, 0, 0, 116, 0, 112, 0, 105, 0, 126, 0, 114, 0, 0, 0, 116, 0, 112, 0, 116, 0, 126, 0, 122, 0, 0, 0, 116, 0, 113, 0, 108, 0, 126, 0, 114, 0, 0, 0, 116, 0, 113, 0, 116, 0, 126, 0, 117, 0, 0, 0, 116, 0, 114, 0, 97, 0, 126, 0, 122, 0, 0, 0, 116, 0, 115, 0, 97, 0, 126, 0, 101, 0, 0, 0, 116, 0, 115, 0, 103, 0, 126, 0, 109, 0, 0, 0, 116, 0, 115, 0, 112, 0, 126, 0, 122, 0, 0, 0, 116, 0, 116, 0, 97, 0, 126, 0, 112, 0, 0, 0, 116, 0, 116, 0, 101, 0, 116, 0, 111, 0, 0, 0, 116, 0, 116, 0, 114, 0, 99, 0, 109, 0, 0, 0, 116, 0, 116, 0, 114, 0, 126, 0, 119, 0, 0, 0, 116, 0, 116, 0, 119, 0, 116, 0, 111, 0, 0, 0, 116, 0, 116, 0, 121, 0, 126, 0, 122, 0, 0, 0, 116, 0, 117, 0, 97, 0, 126, 0, 106, 0, 0, 0, 116, 0, 117, 0, 108, 0, 126, 0, 113, 0, 0, 0, 116, 0, 117, 0, 115, 0, 126, 0, 122, 0, 0, 0, 116, 0, 118, 0, 100, 0, 126, 0, 101, 0, 0, 0, 116, 0, 118, 0, 107, 0, 126, 0, 111, 0, 0, 0, 116, 0, 118, 0, 115, 0, 126, 0, 117, 0, 0, 0, 116, 0, 118, 0, 119, 0, 126, 0, 121, 0, 0, 0, 116, 0, 119, 0, 97, 0, 126, 0, 104, 0, 0, 0, 116, 0, 119, 0, 107, 0, 104, 0, 113, 0, 0, 0, 116, 0, 119, 0, 108, 0, 126, 0, 114, 0, 0, 0, 116, 0, 119, 0, 116, 0, 110, 0, 113, 0, 0, 0, 116, 0, 119, 0, 116, 0, 112, 0, 113, 0, 0, 0, 116, 0, 119, 0, 116, 0, 120, 0, 113, 0, 0, 0, 116, 0, 119, 0, 116, 0, 126, 0, 117, 0, 0, 0, 116, 0, 119, 0, 119, 0, 126, 0, 121, 0, 0, 0, 116, 0, 120, 0, 97, 0, 126, 0, 99, 0, 0, 0, 116, 0, 120, 0, 103, 0, 126, 0, 106, 0, 0, 0, 116, 0, 120, 0, 109, 0, 126, 0, 111, 0, 0, 0, 116, 0, 120, 0, 113, 0, 126, 0, 117, 0, 0, 0, 116, 0, 120, 0, 120, 0, 126, 0, 121, 0, 0, 0, 116, 0, 121, 0, 104, 0, 126, 0, 106, 0, 0, 0, 116, 0, 121, 0, 114, 0, 126, 0, 118, 0, 0, 0, 116, 0, 122, 0, 108, 0, 126, 0, 111, 0, 0, 0, 117, 0, 97, 0, 109, 0, 126, 0, 110, 0, 0, 0, 117, 0, 99, 0, 99, 0, 111, 0, 114, 0, 0, 0, 117, 0, 100, 0, 105, 0, 126, 0, 106, 0, 0, 0, 117, 0, 100, 0, 108, 0, 126, 0, 109, 0, 0, 0, 117, 0, 103, 0, 49, 0, 48, 0, 49, 0, 0, 0, 117, 0, 103, 0, 49, 0, 48, 0, 50, 0, 0, 0, 117, 0, 103, 0, 49, 0, 48, 0, 51, 0, 0, 0, 117, 0, 103, 0, 49, 0, 48, 0, 52, 0, 0, 0, 117, 0, 103, 0, 49, 0, 48, 0, 53, 0, 0, 0, 117, 0, 103, 0, 49, 0, 48, 0, 54, 0, 0, 0, 117, 0, 103, 0, 49, 0, 48, 0, 55, 0, 0, 0, 117, 0, 103, 0, 49, 0, 48, 0, 56, 0, 0, 0, 117, 0, 103, 0, 49, 0, 48, 0, 57, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 48, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 49, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 50, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 51, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 52, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 53, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 54, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 55, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 56, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 57, 0, 0, 0, 117, 0, 103, 0, 49, 0, 50, 0, 48, 0, 0, 0, 117, 0, 103, 0, 49, 0, 50, 0, 49, 0, 0, 0, 117, 0, 103, 0, 49, 0, 50, 0, 50, 0, 0, 0, 117, 0, 103, 0, 49, 0, 50, 0, 51, 0, 0, 0, 117, 0, 103, 0, 49, 0, 50, 0, 52, 0, 0, 0, 117, 0, 103, 0, 49, 0, 50, 0, 53, 0, 0, 0, 117, 0, 103, 0, 49, 0, 50, 0, 54, 0, 0, 0, 117, 0, 103, 0, 50, 0, 48, 0, 49, 0, 0, 0, 117, 0, 103, 0, 50, 0, 48, 0, 50, 0, 0, 0, 117, 0, 103, 0, 50, 0, 48, 0, 51, 0, 0, 0, 117, 0, 103, 0, 50, 0, 48, 0, 52, 0, 0, 0, 117, 0, 103, 0, 50, 0, 48, 0, 53, 0, 0, 0, 117, 0, 103, 0, 50, 0, 48, 0, 54, 0, 0, 0, 117, 0, 103, 0, 50, 0, 48, 0, 55, 0, 0, 0, 117, 0, 103, 0, 50, 0, 48, 0, 56, 0, 0, 0, 117, 0, 103, 0, 50, 0, 48, 0, 57, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 48, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 49, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 50, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 51, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 52, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 53, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 54, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 55, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 56, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 57, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 48, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 49, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 50, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 51, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 52, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 53, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 54, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 55, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 56, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 57, 0, 0, 0, 117, 0, 103, 0, 50, 0, 51, 0, 53, 0, 0, 0, 117, 0, 103, 0, 50, 0, 51, 0, 54, 0, 0, 0, 117, 0, 103, 0, 51, 0, 48, 0, 49, 0, 0, 0, 117, 0, 103, 0, 51, 0, 48, 0, 50, 0, 0, 0, 117, 0, 103, 0, 51, 0, 48, 0, 51, 0, 0, 0, 117, 0, 103, 0, 51, 0, 48, 0, 53, 0, 0, 0, 117, 0, 103, 0, 51, 0, 48, 0, 54, 0, 0, 0, 117, 0, 103, 0, 51, 0, 48, 0, 55, 0, 0, 0, 117, 0, 103, 0, 51, 0, 48, 0, 57, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 48, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 49, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 50, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 51, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 52, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 53, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 54, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 55, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 56, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 57, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 49, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 50, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 51, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 52, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 53, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 54, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 55, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 57, 0, 0, 0, 117, 0, 103, 0, 51, 0, 51, 0, 48, 0, 0, 0, 117, 0, 103, 0, 51, 0, 51, 0, 49, 0, 0, 0, 117, 0, 103, 0, 51, 0, 51, 0, 51, 0, 0, 0, 117, 0, 103, 0, 52, 0, 48, 0, 49, 0, 0, 0, 117, 0, 103, 0, 52, 0, 48, 0, 50, 0, 0, 0, 117, 0, 103, 0, 52, 0, 48, 0, 51, 0, 0, 0, 117, 0, 103, 0, 52, 0, 48, 0, 53, 0, 0, 0, 117, 0, 103, 0, 52, 0, 48, 0, 54, 0, 0, 0, 117, 0, 103, 0, 52, 0, 48, 0, 55, 0, 0, 0, 117, 0, 103, 0, 52, 0, 48, 0, 57, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 49, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 50, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 51, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 53, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 54, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 48, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 49, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 50, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 51, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 52, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 53, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 54, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 55, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 57, 0, 0, 0, 117, 0, 103, 0, 52, 0, 51, 0, 49, 0, 0, 0, 117, 0, 103, 0, 52, 0, 51, 0, 50, 0, 0, 0, 117, 0, 103, 0, 97, 0, 126, 0, 98, 0, 0, 0, 117, 0, 103, 0, 110, 0, 126, 0, 111, 0, 0, 0, 117, 0, 107, 0, 103, 0, 126, 0, 104, 0, 0, 0, 117, 0, 107, 0, 112, 0, 126, 0, 113, 0, 0, 0, 117, 0, 108, 0, 97, 0, 126, 0, 99, 0, 0, 0, 117, 0, 108, 0, 101, 0, 126, 0, 102, 0, 0, 0, 117, 0, 108, 0, 107, 0, 126, 0, 110, 0, 0, 0, 117, 0, 109, 0, 97, 0, 126, 0, 100, 0, 0, 0, 117, 0, 109, 0, 109, 0, 126, 0, 112, 0, 0, 0, 117, 0, 109, 0, 114, 0, 126, 0, 115, 0, 0, 0, 117, 0, 110, 0, 109, 0, 126, 0, 110, 0, 0, 0, 117, 0, 114, 0, 97, 0, 126, 0, 99, 0, 0, 0, 117, 0, 114, 0, 101, 0, 126, 0, 112, 0, 0, 0, 117, 0, 114, 0, 116, 0, 126, 0, 122, 0, 0, 0, 117, 0, 115, 0, 104, 0, 126, 0, 105, 0, 0, 0, 117, 0, 117, 0, 109, 0, 126, 0, 110, 0, 0, 0, 118, 0, 97, 0, 101, 0, 126, 0, 106, 0, 0, 0, 118, 0, 97, 0, 108, 0, 126, 0, 112, 0, 0, 0, 118, 0, 97, 0, 114, 0, 126, 0, 115, 0, 0, 0, 118, 0, 97, 0, 117, 0, 126, 0, 118, 0, 0, 0, 118, 0, 101, 0, 97, 0, 126, 0, 112, 0, 0, 0, 118, 0, 101, 0, 99, 0, 126, 0, 100, 0, 0, 0, 118, 0, 101, 0, 108, 0, 126, 0, 109, 0, 0, 0, 118, 0, 101, 0, 111, 0, 126, 0, 112, 0, 0, 0, 118, 0, 101, 0, 114, 0, 126, 0, 122, 0, 0, 0, 118, 0, 105, 0, 99, 0, 126, 0, 100, 0, 0, 0, 118, 0, 105, 0, 102, 0, 126, 0, 103, 0, 0, 0, 118, 0, 105, 0, 115, 0, 126, 0, 116, 0, 0, 0, 118, 0, 107, 0, 105, 0, 126, 0, 109, 0, 0, 0, 118, 0, 107, 0, 111, 0, 126, 0, 112, 0, 0, 0, 118, 0, 107, 0, 116, 0, 126, 0, 117, 0, 0, 0, 118, 0, 109, 0, 97, 0, 126, 0, 109, 0, 0, 0, 118, 0, 109, 0, 112, 0, 126, 0, 115, 0, 0, 0, 118, 0, 109, 0, 117, 0, 126, 0, 122, 0, 0, 0, 118, 0, 114, 0, 115, 0, 126, 0, 116, 0, 0, 0, 118, 0, 117, 0, 109, 0, 126, 0, 110, 0, 0, 0, 119, 0, 97, 0, 97, 0, 126, 0, 122, 0, 0, 0, 119, 0, 98, 0, 97, 0, 126, 0, 98, 0, 0, 0, 119, 0, 98, 0, 101, 0, 126, 0, 102, 0, 0, 0, 119, 0, 98, 0, 104, 0, 126, 0, 109, 0, 0, 0, 119, 0, 98, 0, 112, 0, 126, 0, 116, 0, 0, 0, 119, 0, 98, 0, 118, 0, 126, 0, 119, 0, 0, 0, 119, 0, 100, 0, 106, 0, 126, 0, 107, 0, 0, 0, 119, 0, 101, 0, 99, 0, 126, 0, 100, 0, 0, 0, 119, 0, 101, 0, 103, 0, 126, 0, 105, 0, 0, 0, 119, 0, 101, 0, 109, 0, 126, 0, 112, 0, 0, 0, 119, 0, 101, 0, 114, 0, 126, 0, 117, 0, 0, 0, 119, 0, 103, 0, 97, 0, 126, 0, 98, 0, 0, 0, 119, 0, 105, 0, 98, 0, 126, 0, 99, 0, 0, 0, 119, 0, 105, 0, 101, 0, 126, 0, 110, 0, 0, 0, 119, 0, 105, 0, 117, 0, 126, 0, 118, 0, 0, 0, 119, 0, 107, 0, 97, 0, 126, 0, 98, 0, 0, 0, 119, 0, 108, 0, 107, 0, 126, 0, 109, 0, 0, 0, 119, 0, 108, 0, 114, 0, 126, 0, 115, 0, 0, 0, 119, 0, 108, 0, 117, 0, 126, 0, 121, 0, 0, 0, 119, 0, 109, 0, 97, 0, 126, 0, 101, 0, 0, 0, 119, 0, 109, 0, 104, 0, 126, 0, 105, 0, 0, 0, 119, 0, 109, 0, 109, 0, 126, 0, 111, 0, 0, 0, 119, 0, 109, 0, 115, 0, 126, 0, 116, 0, 0, 0, 119, 0, 109, 0, 119, 0, 126, 0, 120, 0, 0, 0, 119, 0, 110, 0, 98, 0, 126, 0, 101, 0, 0, 0, 119, 0, 110, 0, 109, 0, 126, 0, 112, 0, 0, 0, 119, 0, 111, 0, 97, 0, 126, 0, 103, 0, 0, 0, 119, 0, 111, 0, 109, 0, 126, 0, 111, 0, 0, 0, 119, 0, 111, 0, 114, 0, 126, 0, 115, 0, 0, 0, 119, 0, 114, 0, 97, 0, 126, 0, 98, 0, 0, 0, 119, 0, 114, 0, 103, 0, 126, 0, 105, 0, 0, 0, 119, 0, 114, 0, 107, 0, 126, 0, 112, 0, 0, 0, 119, 0, 114, 0, 114, 0, 126, 0, 115, 0, 0, 0, 119, 0, 114, 0, 117, 0, 126, 0, 122, 0, 0, 0, 119, 0, 115, 0, 114, 0, 126, 0, 115, 0, 0, 0, 119, 0, 115, 0, 117, 0, 126, 0, 118, 0, 0, 0, 119, 0, 116, 0, 104, 0, 126, 0, 105, 0, 0, 0, 119, 0, 117, 0, 97, 0, 126, 0, 98, 0, 0, 0, 119, 0, 117, 0, 108, 0, 126, 0, 110, 0, 0, 0, 119, 0, 117, 0, 116, 0, 126, 0, 118, 0, 0, 0, 119, 0, 117, 0, 120, 0, 126, 0, 121, 0, 0, 0, 119, 0, 119, 0, 97, 0, 126, 0, 98, 0, 0, 0, 119, 0, 121, 0, 97, 0, 126, 0, 98, 0, 0, 0, 120, 0, 97, 0, 97, 0, 126, 0, 101, 0, 0, 0, 120, 0, 97, 0, 105, 0, 126, 0, 119, 0, 0, 0, 120, 0, 98, 0, 98, 0, 126, 0, 101, 0, 0, 0, 120, 0, 98, 0, 105, 0, 126, 0, 106, 0, 0, 0, 120, 0, 98, 0, 109, 0, 126, 0, 112, 0, 0, 0, 120, 0, 99, 0, 98, 0, 126, 0, 99, 0, 0, 0, 120, 0, 99, 0, 103, 0, 126, 0, 104, 0, 0, 0, 120, 0, 99, 0, 108, 0, 126, 0, 111, 0, 0, 0, 120, 0, 99, 0, 116, 0, 126, 0, 119, 0, 0, 0, 120, 0, 101, 0, 108, 0, 126, 0, 109, 0, 0, 0, 120, 0, 101, 0, 114, 0, 126, 0, 117, 0, 0, 0, 120, 0, 103, 0, 97, 0, 126, 0, 98, 0, 0, 0, 120, 0, 103, 0, 102, 0, 126, 0, 103, 0, 0, 0, 120, 0, 103, 0, 108, 0, 126, 0, 110, 0, 0, 0, 120, 0, 104, 0, 99, 0, 126, 0, 101, 0, 0, 0, 120, 0, 104, 0, 116, 0, 126, 0, 118, 0, 0, 0, 120, 0, 105, 0, 114, 0, 126, 0, 115, 0, 0, 0, 120, 0, 107, 0, 97, 0, 126, 0, 103, 0, 0, 0, 120, 0, 107, 0, 105, 0, 126, 0, 108, 0, 0, 0, 120, 0, 107, 0, 110, 0, 126, 0, 122, 0, 0, 0, 120, 0, 108, 0, 97, 0, 126, 0, 101, 0, 0, 0, 120, 0, 108, 0, 110, 0, 126, 0, 112, 0, 0, 0, 120, 0, 109, 0, 97, 0, 126, 0, 104, 0, 0, 0, 120, 0, 109, 0, 106, 0, 126, 0, 122, 0, 0, 0, 120, 0, 110, 0, 97, 0, 126, 0, 98, 0, 0, 0, 120, 0, 110, 0, 103, 0, 126, 0, 105, 0, 0, 0, 120, 0, 110, 0, 110, 0, 126, 0, 111, 0, 0, 0, 120, 0, 110, 0, 114, 0, 126, 0, 117, 0, 0, 0, 120, 0, 110, 0, 121, 0, 126, 0, 122, 0, 0, 0, 120, 0, 111, 0, 99, 0, 126, 0, 100, 0, 0, 0, 120, 0, 111, 0, 109, 0, 126, 0, 112, 0, 0, 0, 120, 0, 112, 0, 105, 0, 126, 0, 107, 0, 0, 0, 120, 0, 112, 0, 109, 0, 126, 0, 117, 0, 0, 0, 120, 0, 114, 0, 97, 0, 126, 0, 98, 0, 0, 0, 120, 0, 114, 0, 100, 0, 126, 0, 101, 0, 0, 0, 120, 0, 114, 0, 109, 0, 126, 0, 110, 0, 0, 0, 120, 0, 114, 0, 113, 0, 126, 0, 114, 0, 0, 0, 120, 0, 114, 0, 116, 0, 126, 0, 117, 0, 0, 0, 120, 0, 115, 0, 97, 0, 126, 0, 101, 0, 0, 0, 120, 0, 115, 0, 104, 0, 126, 0, 105, 0, 0, 0, 120, 0, 115, 0, 109, 0, 126, 0, 115, 0, 0, 0, 120, 0, 115, 0, 117, 0, 126, 0, 118, 0, 0, 0, 120, 0, 116, 0, 97, 0, 126, 0, 101, 0, 0, 0, 120, 0, 116, 0, 103, 0, 126, 0, 106, 0, 0, 0, 120, 0, 116, 0, 108, 0, 126, 0, 119, 0, 0, 0, 120, 0, 116, 0, 121, 0, 126, 0, 122, 0, 0, 0, 120, 0, 117, 0, 97, 0, 126, 0, 98, 0, 0, 0, 120, 0, 117, 0, 108, 0, 126, 0, 112, 0, 0, 0, 120, 0, 117, 0, 116, 0, 126, 0, 117, 0, 0, 0, 120, 0, 118, 0, 110, 0, 126, 0, 111, 0, 0, 0, 120, 0, 119, 0, 99, 0, 126, 0, 101, 0, 0, 0, 120, 0, 119, 0, 106, 0, 126, 0, 108, 0, 0, 0, 120, 0, 121, 0, 97, 0, 126, 0, 98, 0, 0, 0, 120, 0, 121, 0, 106, 0, 126, 0, 108, 0, 0, 0, 121, 0, 97, 0, 97, 0, 126, 0, 122, 0, 0, 0, 121, 0, 98, 0, 97, 0, 126, 0, 98, 0, 0, 0, 121, 0, 98, 0, 104, 0, 126, 0, 111, 0, 0, 0, 121, 0, 98, 0, 120, 0, 126, 0, 121, 0, 0, 0, 121, 0, 101, 0, 105, 0, 126, 0, 106, 0, 0, 0, 121, 0, 101, 0, 114, 0, 126, 0, 118, 0, 0, 0, 121, 0, 103, 0, 108, 0, 126, 0, 109, 0, 0, 0, 121, 0, 103, 0, 114, 0, 126, 0, 115, 0, 0, 0, 121, 0, 105, 0, 102, 0, 126, 0, 110, 0, 0, 0, 121, 0, 105, 0, 112, 0, 126, 0, 118, 0, 0, 0, 121, 0, 107, 0, 107, 0, 126, 0, 111, 0, 0, 0, 121, 0, 107, 0, 116, 0, 126, 0, 117, 0, 0, 0, 121, 0, 108, 0, 97, 0, 126, 0, 98, 0, 0, 0, 121, 0, 108, 0, 108, 0, 126, 0, 111, 0, 0, 0, 121, 0, 109, 0, 98, 0, 126, 0, 101, 0, 0, 0, 121, 0, 109, 0, 103, 0, 126, 0, 105, 0, 0, 0, 121, 0, 109, 0, 107, 0, 126, 0, 115, 0, 0, 0, 121, 0, 110, 0, 100, 0, 126, 0, 101, 0, 0, 0, 121, 0, 110, 0, 107, 0, 126, 0, 108, 0, 0, 0, 121, 0, 110, 0, 110, 0, 126, 0, 111, 0, 0, 0, 121, 0, 111, 0, 107, 0, 126, 0, 110, 0, 0, 0, 121, 0, 111, 0, 120, 0, 126, 0, 121, 0, 0, 0, 121, 0, 112, 0, 97, 0, 126, 0, 98, 0, 0, 0, 121, 0, 112, 0, 103, 0, 126, 0, 104, 0, 0, 0, 121, 0, 112, 0, 109, 0, 126, 0, 112, 0, 0, 0, 121, 0, 114, 0, 97, 0, 126, 0, 98, 0, 0, 0, 121, 0, 114, 0, 107, 0, 126, 0, 111, 0, 0, 0, 121, 0, 115, 0, 99, 0, 126, 0, 100, 0, 0, 0, 121, 0, 115, 0, 110, 0, 126, 0, 112, 0, 0, 0, 121, 0, 115, 0, 114, 0, 126, 0, 115, 0, 0, 0, 121, 0, 117, 0, 97, 0, 126, 0, 103, 0, 0, 0, 121, 0, 117, 0, 105, 0, 126, 0, 110, 0, 0, 0, 121, 0, 117, 0, 112, 0, 126, 0, 114, 0, 0, 0, 121, 0, 117, 0, 119, 0, 126, 0, 122, 0, 0, 0, 121, 0, 119, 0, 113, 0, 126, 0, 114, 0, 0, 0, 121, 0, 119, 0, 116, 0, 126, 0, 117, 0, 0, 0, 121, 0, 120, 0, 108, 0, 126, 0, 109, 0, 0, 0, 122, 0, 97, 0, 97, 0, 126, 0, 104, 0, 0, 0, 122, 0, 97, 0, 106, 0, 126, 0, 109, 0, 0, 0, 122, 0, 97, 0, 111, 0, 126, 0, 122, 0, 0, 0, 122, 0, 101, 0, 103, 0, 126, 0, 104, 0, 0, 0, 122, 0, 103, 0, 97, 0, 126, 0, 98, 0, 0, 0, 122, 0, 103, 0, 109, 0, 126, 0, 110, 0, 0, 0, 122, 0, 104, 0, 119, 0, 126, 0, 120, 0, 0, 0, 122, 0, 105, 0, 97, 0, 126, 0, 98, 0, 0, 0, 122, 0, 105, 0, 107, 0, 126, 0, 110, 0, 0, 0, 122, 0, 107, 0, 97, 0, 126, 0, 98, 0, 0, 0, 122, 0, 107, 0, 103, 0, 126, 0, 104, 0, 0, 0, 122, 0, 107, 0, 110, 0, 126, 0, 112, 0, 0, 0, 122, 0, 107, 0, 116, 0, 126, 0, 118, 0, 0, 0, 122, 0, 108, 0, 109, 0, 126, 0, 110, 0, 0, 0, 122, 0, 109, 0, 97, 0, 126, 0, 122, 0, 0, 0, 122, 0, 110, 0, 100, 0, 126, 0, 101, 0, 0, 0, 122, 0, 111, 0, 113, 0, 126, 0, 115, 0, 0, 0, 122, 0, 112, 0, 97, 0, 126, 0, 122, 0, 0, 0, 122, 0, 114, 0, 110, 0, 126, 0, 112, 0, 0, 0, 122, 0, 115, 0, 107, 0, 126, 0, 108, 0, 0, 0, 122, 0, 116, 0, 108, 0, 126, 0, 110, 0, 0, 0, 122, 0, 116, 0, 112, 0, 126, 0, 113, 0, 0, 0, 122, 0, 116, 0, 115, 0, 126, 0, 117, 0, 0, 0, 122, 0, 116, 0, 120, 0, 126, 0, 121, 0, 0, 0, 122, 0, 117, 0, 109, 0, 126, 0, 110, 0, 0, 0, 116, 0, 114, 0, 56, 0, 48, 0, 126, 0, 49, 0, 0, 0, 99, 0, 104, 0, 122, 0, 103, 0, 126, 0, 104, 0, 0, 0, 112, 0, 116, 0, 95, 0, 42, 0, 95, 0, 42, 0, 0, 0, 101, 0, 110, 0, 95, 0, 42, 0, 95, 0, 42, 0, 0, 0, 101, 0, 115, 0, 95, 0, 42, 0, 95, 0, 42, 0, 0, 0, 103, 0, 114, 0, 105, 0, 116, 0, 97, 0, 108, 0, 0, 0, 97, 0, 111, 0, 49, 0, 57, 0, 57, 0, 48, 0, 0, 0, 97, 0, 114, 0, 95, 0, 42, 0, 95, 0, 42, 0, 0, 0, 105, 0, 100, 0, 98, 0, 97, 0, 126, 0, 98, 0, 0, 0, 105, 0, 116, 0, 55, 0, 55, 0, 126, 0, 56, 0, 0, 0, 109, 0, 100, 0, 115, 0, 110, 0, 126, 0, 111, 0, 0, 0, 110, 0, 103, 0, 107, 0, 100, 0, 126, 0, 101, 0, 0, 0, 114, 0, 117, 0, 107, 0, 107, 0, 126, 0, 108, 0, 0, 0, 99, 0, 111, 0, 112, 0, 116, 0, 105, 0, 99, 0, 0, 0, 110, 0, 100, 0, 121, 0, 117, 0, 107, 0, 97, 0, 0, 0, 112, 0, 105, 0, 110, 0, 121, 0, 105, 0, 110, 0, 0, 0, 66, 0, 114, 0, 97, 0, 104, 0, 126, 0, 105, 0, 0, 0, 72, 0, 97, 0, 110, 0, 115, 0, 126, 0, 116, 0, 0, 0, 76, 0, 105, 0, 110, 0, 97, 0, 126, 0, 98, 0, 0, 0, 81, 0, 97, 0, 97, 0, 97, 0, 126, 0, 102, 0, 0, 0, 81, 0, 97, 0, 97, 0, 106, 0, 126, 0, 112, 0, 0, 0, 81, 0, 97, 0, 97, 0, 113, 0, 126, 0, 122, 0, 0, 0, 81, 0, 97, 0, 98, 0, 97, 0, 126, 0, 120, 0, 0, 0, 97, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 100, 0, 48, 0, 50, 0, 126, 0, 56, 0, 0, 0, 97, 0, 100, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 102, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 103, 0, 48, 0, 51, 0, 126, 0, 56, 0, 0, 0, 97, 0, 103, 0, 49, 0, 48, 0, 126, 0, 49, 0, 0, 0, 97, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 105, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 108, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 97, 0, 108, 0, 49, 0, 48, 0, 126, 0, 50, 0, 0, 0, 97, 0, 108, 0, 107, 0, 97, 0, 126, 0, 99, 0, 0, 0, 97, 0, 108, 0, 112, 0, 113, 0, 126, 0, 114, 0, 0, 0, 97, 0, 108, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 109, 0, 95, 0, 42, 0, 95, 0, 42, 0, 0, 0, 97, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 113, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 114, 0, 97, 0, 110, 0, 101, 0, 115, 0, 0, 0, 97, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 115, 0, 97, 0, 110, 0, 116, 0, 101, 0, 0, 0, 97, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 117, 0, 118, 0, 101, 0, 114, 0, 110, 0, 0, 0, 97, 0, 117, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 119, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 120, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 97, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 98, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 98, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 98, 0, 98, 0, 49, 0, 48, 0, 126, 0, 49, 0, 0, 0, 98, 0, 98, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 100, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 98, 0, 100, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 98, 0, 100, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 98, 0, 100, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 98, 0, 100, 0, 52, 0, 48, 0, 126, 0, 57, 0, 0, 0, 98, 0, 100, 0, 53, 0, 48, 0, 126, 0, 57, 0, 0, 0, 98, 0, 100, 0, 54, 0, 48, 0, 126, 0, 52, 0, 0, 0, 98, 0, 100, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 102, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 98, 0, 102, 0, 49, 0, 48, 0, 126, 0, 51, 0, 0, 0, 98, 0, 102, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 103, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 98, 0, 103, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 98, 0, 103, 0, 50, 0, 48, 0, 126, 0, 56, 0, 0, 0, 98, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 104, 0, 49, 0, 51, 0, 126, 0, 53, 0, 0, 0, 98, 0, 104, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 105, 0, 98, 0, 108, 0, 126, 0, 109, 0, 0, 0, 98, 0, 105, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 106, 0, 97, 0, 107, 0, 126, 0, 108, 0, 0, 0, 98, 0, 106, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 108, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 113, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 114, 0, 97, 0, 108, 0, 126, 0, 109, 0, 0, 0, 98, 0, 114, 0, 109, 0, 115, 0, 126, 0, 116, 0, 0, 0, 98, 0, 114, 0, 112, 0, 97, 0, 126, 0, 98, 0, 0, 0, 98, 0, 114, 0, 114, 0, 110, 0, 126, 0, 111, 0, 0, 0, 98, 0, 114, 0, 114, 0, 114, 0, 126, 0, 115, 0, 0, 0, 98, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 115, 0, 110, 0, 111, 0, 126, 0, 112, 0, 0, 0, 98, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 116, 0, 49, 0, 49, 0, 126, 0, 53, 0, 0, 0, 98, 0, 116, 0, 50, 0, 49, 0, 126, 0, 52, 0, 0, 0, 98, 0, 116, 0, 51, 0, 49, 0, 126, 0, 52, 0, 0, 0, 98, 0, 116, 0, 52, 0, 49, 0, 126, 0, 53, 0, 0, 0, 98, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 118, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 119, 0, 115, 0, 111, 0, 126, 0, 112, 0, 0, 0, 98, 0, 119, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 121, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 98, 0, 122, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 97, 0, 110, 0, 115, 0, 126, 0, 117, 0, 0, 0, 99, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 100, 0, 104, 0, 107, 0, 126, 0, 108, 0, 0, 0, 99, 0, 100, 0, 109, 0, 110, 0, 126, 0, 111, 0, 0, 0, 99, 0, 100, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 102, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 103, 0, 49, 0, 49, 0, 126, 0, 54, 0, 0, 0, 99, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 104, 0, 115, 0, 103, 0, 126, 0, 104, 0, 0, 0, 99, 0, 104, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 105, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 99, 0, 105, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 99, 0, 105, 0, 115, 0, 97, 0, 117, 0, 112, 0, 0, 0, 99, 0, 105, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 108, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 110, 0, 49, 0, 49, 0, 126, 0, 53, 0, 0, 0, 99, 0, 110, 0, 50, 0, 49, 0, 126, 0, 51, 0, 0, 0, 99, 0, 110, 0, 51, 0, 49, 0, 126, 0, 55, 0, 0, 0, 99, 0, 110, 0, 52, 0, 49, 0, 126, 0, 54, 0, 0, 0, 99, 0, 110, 0, 53, 0, 48, 0, 126, 0, 52, 0, 0, 0, 99, 0, 110, 0, 54, 0, 49, 0, 126, 0, 53, 0, 0, 0, 99, 0, 110, 0, 57, 0, 49, 0, 126, 0, 50, 0, 0, 0, 99, 0, 110, 0, 104, 0, 97, 0, 126, 0, 98, 0, 0, 0, 99, 0, 110, 0, 104, 0, 107, 0, 126, 0, 108, 0, 0, 0, 99, 0, 110, 0, 115, 0, 99, 0, 126, 0, 100, 0, 0, 0, 99, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 112, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 114, 0, 101, 0, 105, 0, 115, 0, 115, 0, 0, 0, 99, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 117, 0, 48, 0, 51, 0, 126, 0, 57, 0, 0, 0, 99, 0, 117, 0, 49, 0, 48, 0, 126, 0, 54, 0, 0, 0, 99, 0, 117, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 118, 0, 115, 0, 108, 0, 126, 0, 109, 0, 0, 0, 99, 0, 118, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 119, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 120, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 121, 0, 48, 0, 49, 0, 126, 0, 54, 0, 0, 0, 99, 0, 121, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 100, 0, 97, 0, 106, 0, 110, 0, 107, 0, 111, 0, 0, 0, 100, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 100, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 100, 0, 106, 0, 97, 0, 114, 0, 126, 0, 115, 0, 0, 0, 100, 0, 106, 0, 100, 0, 105, 0, 126, 0, 106, 0, 0, 0, 100, 0, 106, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 100, 0, 107, 0, 56, 0, 49, 0, 126, 0, 53, 0, 0, 0, 100, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 100, 0, 109, 0, 48, 0, 50, 0, 126, 0, 57, 0, 0, 0, 100, 0, 109, 0, 49, 0, 48, 0, 126, 0, 49, 0, 0, 0, 100, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 100, 0, 111, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 100, 0, 111, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 100, 0, 111, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 100, 0, 111, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 100, 0, 111, 0, 52, 0, 48, 0, 126, 0, 50, 0, 0, 0, 100, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 100, 0, 122, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 100, 0, 122, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 100, 0, 122, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 100, 0, 122, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 100, 0, 122, 0, 52, 0, 48, 0, 126, 0, 56, 0, 0, 0, 100, 0, 122, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 101, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 101, 0, 99, 0, 115, 0, 100, 0, 126, 0, 101, 0, 0, 0, 101, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 101, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 101, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 101, 0, 104, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 101, 0, 110, 0, 95, 0, 49, 0, 53, 0, 48, 0, 0, 0, 101, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 101, 0, 115, 0, 95, 0, 52, 0, 49, 0, 57, 0, 0, 0, 101, 0, 115, 0, 97, 0, 114, 0, 126, 0, 115, 0, 0, 0, 101, 0, 115, 0, 99, 0, 97, 0, 126, 0, 99, 0, 0, 0, 101, 0, 115, 0, 99, 0, 108, 0, 126, 0, 109, 0, 0, 0, 101, 0, 115, 0, 99, 0, 114, 0, 126, 0, 117, 0, 0, 0, 101, 0, 115, 0, 109, 0, 99, 0, 126, 0, 100, 0, 0, 0, 101, 0, 115, 0, 116, 0, 101, 0, 126, 0, 102, 0, 0, 0, 101, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 101, 0, 116, 0, 115, 0, 110, 0, 126, 0, 111, 0, 0, 0, 101, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 102, 0, 105, 0, 48, 0, 50, 0, 126, 0, 57, 0, 0, 0, 102, 0, 105, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 102, 0, 105, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 102, 0, 106, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 102, 0, 106, 0, 49, 0, 48, 0, 126, 0, 52, 0, 0, 0, 102, 0, 106, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 102, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 102, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 102, 0, 111, 0, 110, 0, 105, 0, 112, 0, 97, 0, 0, 0, 102, 0, 111, 0, 110, 0, 117, 0, 112, 0, 97, 0, 0, 0, 102, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 102, 0, 114, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 102, 0, 114, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 102, 0, 114, 0, 50, 0, 49, 0, 126, 0, 57, 0, 0, 0, 102, 0, 114, 0, 50, 0, 97, 0, 126, 0, 98, 0, 0, 0, 102, 0, 114, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 102, 0, 114, 0, 52, 0, 48, 0, 126, 0, 57, 0, 0, 0, 102, 0, 114, 0, 53, 0, 48, 0, 126, 0, 57, 0, 0, 0, 102, 0, 114, 0, 54, 0, 48, 0, 126, 0, 57, 0, 0, 0, 102, 0, 114, 0, 55, 0, 48, 0, 126, 0, 57, 0, 0, 0, 102, 0, 114, 0, 56, 0, 48, 0, 126, 0, 57, 0, 0, 0, 102, 0, 114, 0, 57, 0, 48, 0, 126, 0, 53, 0, 0, 0, 102, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 97, 0, 115, 0, 99, 0, 111, 0, 110, 0, 0, 0, 103, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 98, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 100, 0, 48, 0, 49, 0, 126, 0, 54, 0, 0, 0, 103, 0, 100, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 101, 0, 115, 0, 106, 0, 126, 0, 107, 0, 0, 0, 103, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 102, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 104, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 105, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 108, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 110, 0, 98, 0, 101, 0, 126, 0, 102, 0, 0, 0, 103, 0, 110, 0, 107, 0, 97, 0, 126, 0, 98, 0, 0, 0, 103, 0, 110, 0, 107, 0, 100, 0, 126, 0, 101, 0, 0, 0, 103, 0, 110, 0, 107, 0, 110, 0, 126, 0, 111, 0, 0, 0, 103, 0, 110, 0, 109, 0, 99, 0, 126, 0, 100, 0, 0, 0, 103, 0, 110, 0, 109, 0, 108, 0, 126, 0, 109, 0, 0, 0, 103, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 112, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 113, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 114, 0, 48, 0, 51, 0, 126, 0, 55, 0, 0, 0, 103, 0, 114, 0, 49, 0, 49, 0, 126, 0, 55, 0, 0, 0, 103, 0, 114, 0, 50, 0, 49, 0, 126, 0, 52, 0, 0, 0, 103, 0, 114, 0, 51, 0, 49, 0, 126, 0, 52, 0, 0, 0, 103, 0, 114, 0, 52, 0, 49, 0, 126, 0, 52, 0, 0, 0, 103, 0, 114, 0, 53, 0, 49, 0, 126, 0, 57, 0, 0, 0, 103, 0, 114, 0, 54, 0, 49, 0, 126, 0, 52, 0, 0, 0, 103, 0, 114, 0, 55, 0, 49, 0, 126, 0, 51, 0, 0, 0, 103, 0, 114, 0, 56, 0, 49, 0, 126, 0, 53, 0, 0, 0, 103, 0, 114, 0, 57, 0, 49, 0, 126, 0, 52, 0, 0, 0, 103, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 117, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 119, 0, 98, 0, 108, 0, 126, 0, 109, 0, 0, 0, 103, 0, 119, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 103, 0, 121, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 104, 0, 101, 0, 98, 0, 114, 0, 101, 0, 119, 0, 0, 0, 104, 0, 101, 0, 112, 0, 108, 0, 111, 0, 99, 0, 0, 0, 104, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 104, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 104, 0, 110, 0, 99, 0, 108, 0, 126, 0, 109, 0, 0, 0, 104, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 104, 0, 114, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 104, 0, 114, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 104, 0, 114, 0, 50, 0, 48, 0, 126, 0, 49, 0, 0, 0, 104, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 104, 0, 116, 0, 110, 0, 100, 0, 126, 0, 101, 0, 0, 0, 104, 0, 116, 0, 115, 0, 100, 0, 126, 0, 101, 0, 0, 0, 104, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 104, 0, 117, 0, 115, 0, 110, 0, 126, 0, 111, 0, 0, 0, 104, 0, 117, 0, 115, 0, 115, 0, 126, 0, 116, 0, 0, 0, 104, 0, 117, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 100, 0, 106, 0, 97, 0, 126, 0, 98, 0, 0, 0, 105, 0, 100, 0, 107, 0, 97, 0, 126, 0, 98, 0, 0, 0, 105, 0, 100, 0, 107, 0, 114, 0, 126, 0, 117, 0, 0, 0, 105, 0, 100, 0, 110, 0, 116, 0, 126, 0, 117, 0, 0, 0, 105, 0, 100, 0, 112, 0, 97, 0, 126, 0, 98, 0, 0, 0, 105, 0, 100, 0, 115, 0, 97, 0, 126, 0, 98, 0, 0, 0, 105, 0, 100, 0, 115, 0, 108, 0, 126, 0, 110, 0, 0, 0, 105, 0, 100, 0, 115, 0, 114, 0, 126, 0, 117, 0, 0, 0, 105, 0, 100, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 101, 0, 99, 0, 110, 0, 126, 0, 111, 0, 0, 0, 105, 0, 101, 0, 109, 0, 110, 0, 126, 0, 111, 0, 0, 0, 105, 0, 101, 0, 119, 0, 119, 0, 126, 0, 120, 0, 0, 0, 105, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 108, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 110, 0, 97, 0, 114, 0, 126, 0, 115, 0, 0, 0, 105, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 0, 0, 105, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 113, 0, 98, 0, 97, 0, 126, 0, 98, 0, 0, 0, 105, 0, 113, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 114, 0, 48, 0, 49, 0, 126, 0, 56, 0, 0, 0, 105, 0, 114, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 105, 0, 114, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 105, 0, 114, 0, 51, 0, 48, 0, 126, 0, 50, 0, 0, 0, 105, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 105, 0, 116, 0, 97, 0, 110, 0, 126, 0, 114, 0, 0, 0, 105, 0, 116, 0, 98, 0, 110, 0, 126, 0, 111, 0, 0, 0, 105, 0, 116, 0, 98, 0, 114, 0, 126, 0, 116, 0, 0, 0, 105, 0, 116, 0, 99, 0, 97, 0, 126, 0, 98, 0, 0, 0, 105, 0, 116, 0, 99, 0, 104, 0, 126, 0, 105, 0, 0, 0, 105, 0, 116, 0, 99, 0, 110, 0, 126, 0, 111, 0, 0, 0, 105, 0, 116, 0, 99, 0, 114, 0, 126, 0, 116, 0, 0, 0, 105, 0, 116, 0, 108, 0, 116, 0, 126, 0, 117, 0, 0, 0, 105, 0, 116, 0, 109, 0, 98, 0, 126, 0, 99, 0, 0, 0, 105, 0, 116, 0, 109, 0, 110, 0, 126, 0, 111, 0, 0, 0, 105, 0, 116, 0, 109, 0, 115, 0, 126, 0, 116, 0, 0, 0, 105, 0, 116, 0, 112, 0, 99, 0, 126, 0, 101, 0, 0, 0, 105, 0, 116, 0, 112, 0, 110, 0, 126, 0, 111, 0, 0, 0, 105, 0, 116, 0, 112, 0, 116, 0, 126, 0, 118, 0, 0, 0, 105, 0, 116, 0, 114, 0, 109, 0, 126, 0, 111, 0, 0, 0, 105, 0, 116, 0, 115, 0, 111, 0, 126, 0, 112, 0, 0, 0, 105, 0, 116, 0, 115, 0, 114, 0, 126, 0, 115, 0, 0, 0, 105, 0, 116, 0, 116, 0, 110, 0, 126, 0, 112, 0, 0, 0, 105, 0, 116, 0, 116, 0, 114, 0, 126, 0, 115, 0, 0, 0, 105, 0, 116, 0, 118, 0, 97, 0, 126, 0, 99, 0, 0, 0, 105, 0, 116, 0, 118, 0, 114, 0, 126, 0, 116, 0, 0, 0, 105, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 106, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 106, 0, 109, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 106, 0, 109, 0, 49, 0, 48, 0, 126, 0, 52, 0, 0, 0, 106, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 106, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 106, 0, 112, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 106, 0, 112, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 106, 0, 112, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 106, 0, 112, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 106, 0, 112, 0, 52, 0, 48, 0, 126, 0, 55, 0, 0, 0, 106, 0, 112, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 101, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 107, 0, 101, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 107, 0, 101, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 107, 0, 101, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 107, 0, 101, 0, 52, 0, 48, 0, 126, 0, 55, 0, 0, 0, 107, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 104, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 107, 0, 104, 0, 50, 0, 48, 0, 126, 0, 53, 0, 0, 0, 107, 0, 104, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 105, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 110, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 107, 0, 110, 0, 49, 0, 48, 0, 126, 0, 51, 0, 0, 0, 107, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 112, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 107, 0, 112, 0, 49, 0, 51, 0, 126, 0, 52, 0, 0, 0, 107, 0, 112, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 114, 0, 50, 0, 54, 0, 126, 0, 57, 0, 0, 0, 107, 0, 114, 0, 51, 0, 48, 0, 126, 0, 49, 0, 0, 0, 107, 0, 114, 0, 52, 0, 49, 0, 126, 0, 57, 0, 0, 0, 107, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 119, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 121, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 97, 0, 98, 0, 107, 0, 126, 0, 108, 0, 0, 0, 108, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 98, 0, 98, 0, 104, 0, 126, 0, 105, 0, 0, 0, 108, 0, 98, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 99, 0, 48, 0, 49, 0, 126, 0, 51, 0, 0, 0, 108, 0, 99, 0, 48, 0, 53, 0, 126, 0, 56, 0, 0, 0, 108, 0, 99, 0, 49, 0, 48, 0, 126, 0, 50, 0, 0, 0, 108, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 105, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 108, 0, 105, 0, 49, 0, 48, 0, 126, 0, 49, 0, 0, 0, 108, 0, 105, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 107, 0, 49, 0, 49, 0, 126, 0, 51, 0, 0, 0, 108, 0, 107, 0, 50, 0, 49, 0, 126, 0, 51, 0, 0, 0, 108, 0, 107, 0, 51, 0, 49, 0, 126, 0, 51, 0, 0, 0, 108, 0, 107, 0, 52, 0, 49, 0, 126, 0, 53, 0, 0, 0, 108, 0, 107, 0, 53, 0, 49, 0, 126, 0, 51, 0, 0, 0, 108, 0, 107, 0, 54, 0, 49, 0, 126, 0, 50, 0, 0, 0, 108, 0, 107, 0, 55, 0, 49, 0, 126, 0, 50, 0, 0, 0, 108, 0, 107, 0, 56, 0, 49, 0, 126, 0, 50, 0, 0, 0, 108, 0, 107, 0, 57, 0, 49, 0, 126, 0, 50, 0, 0, 0, 108, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 116, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 108, 0, 116, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 116, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 116, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 116, 0, 52, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 116, 0, 53, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 117, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 118, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 108, 0, 121, 0, 109, 0, 105, 0, 126, 0, 106, 0, 0, 0, 108, 0, 121, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 97, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 109, 0, 97, 0, 49, 0, 48, 0, 126, 0, 50, 0, 0, 0, 109, 0, 97, 0, 49, 0, 51, 0, 126, 0, 54, 0, 0, 0, 109, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 99, 0, 115, 0, 111, 0, 126, 0, 112, 0, 0, 0, 109, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 100, 0, 98, 0, 114, 0, 126, 0, 115, 0, 0, 0, 109, 0, 100, 0, 99, 0, 108, 0, 126, 0, 109, 0, 0, 0, 109, 0, 100, 0, 99, 0, 114, 0, 126, 0, 117, 0, 0, 0, 109, 0, 100, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 101, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 109, 0, 101, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 109, 0, 101, 0, 50, 0, 48, 0, 126, 0, 51, 0, 0, 0, 109, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 102, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 104, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 107, 0, 48, 0, 50, 0, 126, 0, 56, 0, 0, 0, 109, 0, 107, 0, 49, 0, 48, 0, 126, 0, 52, 0, 0, 0, 109, 0, 107, 0, 49, 0, 56, 0, 126, 0, 57, 0, 0, 0, 109, 0, 107, 0, 50, 0, 48, 0, 126, 0, 55, 0, 0, 0, 109, 0, 107, 0, 50, 0, 56, 0, 126, 0, 57, 0, 0, 0, 109, 0, 107, 0, 51, 0, 50, 0, 126, 0, 55, 0, 0, 0, 109, 0, 107, 0, 51, 0, 56, 0, 126, 0, 57, 0, 0, 0, 109, 0, 107, 0, 52, 0, 48, 0, 126, 0, 57, 0, 0, 0, 109, 0, 107, 0, 53, 0, 48, 0, 126, 0, 54, 0, 0, 0, 109, 0, 107, 0, 53, 0, 56, 0, 126, 0, 57, 0, 0, 0, 109, 0, 107, 0, 54, 0, 48, 0, 126, 0, 55, 0, 0, 0, 109, 0, 107, 0, 55, 0, 48, 0, 126, 0, 54, 0, 0, 0, 109, 0, 107, 0, 56, 0, 48, 0, 126, 0, 51, 0, 0, 0, 109, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 108, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 109, 0, 48, 0, 49, 0, 126, 0, 55, 0, 0, 0, 109, 0, 109, 0, 49, 0, 49, 0, 126, 0, 56, 0, 0, 0, 109, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 112, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 113, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 114, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 109, 0, 114, 0, 49, 0, 48, 0, 126, 0, 53, 0, 0, 0, 109, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 116, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 109, 0, 116, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 109, 0, 116, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 109, 0, 116, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 109, 0, 116, 0, 52, 0, 48, 0, 126, 0, 57, 0, 0, 0, 109, 0, 116, 0, 53, 0, 48, 0, 126, 0, 57, 0, 0, 0, 109, 0, 116, 0, 54, 0, 48, 0, 126, 0, 56, 0, 0, 0, 109, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 117, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 118, 0, 48, 0, 48, 0, 126, 0, 53, 0, 0, 0, 109, 0, 118, 0, 48, 0, 55, 0, 126, 0, 56, 0, 0, 0, 109, 0, 118, 0, 49, 0, 50, 0, 126, 0, 52, 0, 0, 0, 109, 0, 118, 0, 50, 0, 51, 0, 126, 0, 57, 0, 0, 0, 109, 0, 118, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 119, 0, 107, 0, 114, 0, 126, 0, 115, 0, 0, 0, 109, 0, 119, 0, 109, 0, 103, 0, 126, 0, 104, 0, 0, 0, 109, 0, 119, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 120, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 121, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 109, 0, 121, 0, 49, 0, 48, 0, 126, 0, 54, 0, 0, 0, 109, 0, 121, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 97, 0, 111, 0, 115, 0, 126, 0, 116, 0, 0, 0, 110, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 102, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 103, 0, 107, 0, 110, 0, 126, 0, 111, 0, 0, 0, 110, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 105, 0, 99, 0, 97, 0, 114, 0, 100, 0, 0, 0, 110, 0, 105, 0, 109, 0, 115, 0, 126, 0, 116, 0, 0, 0, 110, 0, 105, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 108, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 111, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 110, 0, 111, 0, 49, 0, 48, 0, 126, 0, 50, 0, 0, 0, 110, 0, 111, 0, 49, 0, 52, 0, 126, 0, 53, 0, 0, 0, 110, 0, 111, 0, 49, 0, 54, 0, 126, 0, 55, 0, 0, 0, 110, 0, 111, 0, 49, 0, 56, 0, 126, 0, 57, 0, 0, 0, 110, 0, 111, 0, 50, 0, 48, 0, 126, 0, 51, 0, 0, 0, 110, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 112, 0, 112, 0, 49, 0, 126, 0, 55, 0, 0, 0, 110, 0, 112, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 114, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 110, 0, 114, 0, 49, 0, 48, 0, 126, 0, 52, 0, 0, 0, 110, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 117, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 111, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 97, 0, 109, 0, 97, 0, 107, 0, 97, 0, 0, 0, 112, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 102, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 104, 0, 48, 0, 48, 0, 126, 0, 51, 0, 0, 0, 112, 0, 104, 0, 48, 0, 53, 0, 126, 0, 57, 0, 0, 0, 112, 0, 104, 0, 49, 0, 48, 0, 126, 0, 53, 0, 0, 0, 112, 0, 104, 0, 52, 0, 48, 0, 126, 0, 49, 0, 0, 0, 112, 0, 104, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 108, 0, 115, 0, 107, 0, 126, 0, 108, 0, 0, 0, 112, 0, 108, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 116, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 112, 0, 116, 0, 49, 0, 48, 0, 126, 0, 56, 0, 0, 0, 112, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 119, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 112, 0, 121, 0, 49, 0, 48, 0, 126, 0, 54, 0, 0, 0, 112, 0, 121, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 113, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 114, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 114, 0, 111, 0, 99, 0, 115, 0, 126, 0, 116, 0, 0, 0, 114, 0, 111, 0, 116, 0, 108, 0, 126, 0, 109, 0, 0, 0, 114, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 114, 0, 115, 0, 48, 0, 48, 0, 126, 0, 57, 0, 0, 0, 114, 0, 115, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 114, 0, 115, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 114, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 114, 0, 117, 0, 107, 0, 98, 0, 126, 0, 99, 0, 0, 0, 114, 0, 117, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 114, 0, 119, 0, 48, 0, 49, 0, 126, 0, 53, 0, 0, 0, 114, 0, 119, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 97, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 115, 0, 97, 0, 49, 0, 48, 0, 126, 0, 50, 0, 0, 0, 115, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 98, 0, 109, 0, 107, 0, 126, 0, 108, 0, 0, 0, 115, 0, 98, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 99, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 115, 0, 99, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 99, 0, 50, 0, 48, 0, 126, 0, 53, 0, 0, 0, 115, 0, 99, 0, 111, 0, 117, 0, 115, 0, 101, 0, 0, 0, 115, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 100, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 101, 0, 97, 0, 98, 0, 126, 0, 99, 0, 0, 0, 115, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 103, 0, 48, 0, 49, 0, 126, 0, 53, 0, 0, 0, 115, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 104, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 105, 0, 109, 0, 112, 0, 108, 0, 101, 0, 0, 0, 115, 0, 105, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 106, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 108, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 109, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 115, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 110, 0, 107, 0, 100, 0, 126, 0, 101, 0, 0, 0, 115, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 118, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 120, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 121, 0, 104, 0, 108, 0, 126, 0, 109, 0, 0, 0, 115, 0, 121, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 97, 0, 114, 0, 97, 0, 115, 0, 107, 0, 0, 0, 116, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 100, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 102, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 104, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 104, 0, 50, 0, 48, 0, 126, 0, 55, 0, 0, 0, 116, 0, 104, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 104, 0, 52, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 104, 0, 53, 0, 48, 0, 126, 0, 56, 0, 0, 0, 116, 0, 104, 0, 54, 0, 48, 0, 126, 0, 55, 0, 0, 0, 116, 0, 104, 0, 55, 0, 48, 0, 126, 0, 55, 0, 0, 0, 116, 0, 104, 0, 56, 0, 48, 0, 126, 0, 54, 0, 0, 0, 116, 0, 104, 0, 57, 0, 48, 0, 126, 0, 54, 0, 0, 0, 116, 0, 104, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 106, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 108, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 110, 0, 49, 0, 49, 0, 126, 0, 52, 0, 0, 0, 116, 0, 110, 0, 50, 0, 49, 0, 126, 0, 51, 0, 0, 0, 116, 0, 110, 0, 51, 0, 49, 0, 126, 0, 52, 0, 0, 0, 116, 0, 110, 0, 52, 0, 49, 0, 126, 0, 51, 0, 0, 0, 116, 0, 110, 0, 53, 0, 49, 0, 126, 0, 51, 0, 0, 0, 116, 0, 110, 0, 55, 0, 49, 0, 126, 0, 51, 0, 0, 0, 116, 0, 110, 0, 56, 0, 49, 0, 126, 0, 51, 0, 0, 0, 116, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 111, 0, 48, 0, 49, 0, 126, 0, 53, 0, 0, 0, 116, 0, 111, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 114, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 116, 0, 114, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 114, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 114, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 114, 0, 52, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 114, 0, 53, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 114, 0, 54, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 114, 0, 55, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 114, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 118, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 119, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 116, 0, 122, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 116, 0, 122, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 122, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 116, 0, 122, 0, 51, 0, 48, 0, 126, 0, 49, 0, 0, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 117, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 117, 0, 99, 0, 114, 0, 99, 0, 111, 0, 114, 0, 0, 0, 117, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 117, 0, 108, 0, 115, 0, 116, 0, 101, 0, 114, 0, 0, 0, 117, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 117, 0, 110, 0, 105, 0, 102, 0, 111, 0, 110, 0, 0, 0, 117, 0, 115, 0, 97, 0, 107, 0, 126, 0, 108, 0, 0, 0, 117, 0, 115, 0, 109, 0, 100, 0, 126, 0, 101, 0, 0, 0, 117, 0, 115, 0, 109, 0, 110, 0, 126, 0, 111, 0, 0, 0, 117, 0, 115, 0, 109, 0, 115, 0, 126, 0, 116, 0, 0, 0, 117, 0, 115, 0, 110, 0, 99, 0, 126, 0, 101, 0, 0, 0, 117, 0, 115, 0, 115, 0, 99, 0, 126, 0, 100, 0, 0, 0, 117, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 117, 0, 121, 0, 114, 0, 110, 0, 126, 0, 111, 0, 0, 0, 117, 0, 121, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 117, 0, 122, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 118, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 118, 0, 99, 0, 48, 0, 49, 0, 126, 0, 54, 0, 0, 0, 118, 0, 99, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 118, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 118, 0, 103, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 118, 0, 105, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 118, 0, 110, 0, 48, 0, 49, 0, 126, 0, 55, 0, 0, 0, 118, 0, 110, 0, 49, 0, 51, 0, 126, 0, 52, 0, 0, 0, 118, 0, 110, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 118, 0, 110, 0, 51, 0, 48, 0, 126, 0, 55, 0, 0, 0, 118, 0, 110, 0, 52, 0, 48, 0, 126, 0, 49, 0, 0, 0, 118, 0, 110, 0, 52, 0, 51, 0, 126, 0, 55, 0, 0, 0, 118, 0, 110, 0, 53, 0, 48, 0, 126, 0, 57, 0, 0, 0, 118, 0, 110, 0, 54, 0, 54, 0, 126, 0, 57, 0, 0, 0, 118, 0, 110, 0, 55, 0, 48, 0, 126, 0, 51, 0, 0, 0, 118, 0, 110, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 118, 0, 117, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 119, 0, 102, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 119, 0, 115, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 120, 0, 107, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 121, 0, 101, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 121, 0, 116, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 122, 0, 97, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 122, 0, 109, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 122, 0, 109, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 122, 0, 119, 0, 109, 0, 118, 0, 126, 0, 119, 0, 0, 0, 122, 0, 119, 0, 122, 0, 122, 0, 122, 0, 122, 0, 0, 0, 117, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 117, 0, 122, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 0, 0, 117, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 110, 0, 101, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 0, 0, 101, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 105, 0, 115, 0, 108, 0, 97, 0, 109, 0, 105, 0, 99, 0, 0, 0, 98, 0, 110, 0, 95, 0, 66, 0, 101, 0, 110, 0, 103, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 0, 0, 114, 0, 117, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 0, 0, 116, 0, 105, 0, 95, 0, 69, 0, 116, 0, 104, 0, 105, 0, 0, 0, 99, 0, 104, 0, 105, 0, 110, 0, 101, 0, 115, 0, 101, 0, 0, 0, 121, 0, 105, 0, 95, 0, 72, 0, 101, 0, 98, 0, 114, 0, 0, 0, 109, 0, 121, 0, 95, 0, 77, 0, 121, 0, 109, 0, 114, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 0, 0, 116, 0, 101, 0, 95, 0, 84, 0, 101, 0, 108, 0, 117, 0, 0, 0, 101, 0, 115, 0, 95, 0, 42, 0, 95, 0, 69, 0, 83, 0, 0, 0, 107, 0, 97, 0, 95, 0, 71, 0, 101, 0, 111, 0, 114, 0, 0, 0, 108, 0, 111, 0, 95, 0, 76, 0, 97, 0, 111, 0, 111, 0, 0, 0, 115, 0, 105, 0, 95, 0, 83, 0, 105, 0, 110, 0, 104, 0, 0, 0, 107, 0, 110, 0, 95, 0, 75, 0, 110, 0, 100, 0, 97, 0, 0, 0, 111, 0, 114, 0, 95, 0, 79, 0, 114, 0, 121, 0, 97, 0, 0, 0, 116, 0, 97, 0, 95, 0, 84, 0, 97, 0, 109, 0, 108, 0, 0, 0, 98, 0, 109, 0, 95, 0, 78, 0, 107, 0, 111, 0, 111, 0, 0, 0, 101, 0, 110, 0, 95, 0, 83, 0, 104, 0, 97, 0, 119, 0, 0, 0, 104, 0, 121, 0, 95, 0, 65, 0, 114, 0, 109, 0, 110, 0, 0, 0, 107, 0, 109, 0, 95, 0, 75, 0, 104, 0, 109, 0, 114, 0, 0, 0, 109, 0, 108, 0, 95, 0, 77, 0, 108, 0, 121, 0, 109, 0, 0, 0, 112, 0, 97, 0, 95, 0, 71, 0, 117, 0, 114, 0, 117, 0, 0, 0, 101, 0, 110, 0, 95, 0, 42, 0, 95, 0, 71, 0, 66, 0, 0, 0, 115, 0, 114, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 109, 0, 110, 0, 95, 0, 77, 0, 111, 0, 110, 0, 103, 0, 0, 0, 101, 0, 110, 0, 95, 0, 68, 0, 115, 0, 114, 0, 116, 0, 0, 0, 102, 0, 102, 0, 95, 0, 65, 0, 100, 0, 108, 0, 109, 0, 0, 0, 115, 0, 100, 0, 95, 0, 75, 0, 104, 0, 111, 0, 106, 0, 0, 0, 115, 0, 100, 0, 95, 0, 83, 0, 105, 0, 110, 0, 100, 0, 0, 0, 97, 0, 122, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 112, 0, 101, 0, 114, 0, 115, 0, 105, 0, 97, 0, 110, 0, 0, 0, 112, 0, 115, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 112, 0, 116, 0, 95, 0, 42, 0, 95, 0, 66, 0, 82, 0, 0, 0, 112, 0, 116, 0, 95, 0, 42, 0, 95, 0, 85, 0, 83, 0, 0, 0, 115, 0, 100, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 115, 0, 114, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 0, 0, 116, 0, 107, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 117, 0, 114, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 99, 0, 117, 0, 95, 0, 71, 0, 108, 0, 97, 0, 103, 0, 0, 0, 97, 0, 98, 0, 108, 0, 49, 0, 57, 0, 52, 0, 51, 0, 0, 0, 98, 0, 97, 0, 117, 0, 100, 0, 100, 0, 104, 0, 97, 0, 0, 0, 109, 0, 101, 0, 116, 0, 101, 0, 108, 0, 107, 0, 111, 0, 0, 0, 97, 0, 108, 0, 97, 0, 108, 0, 99, 0, 57, 0, 55, 0, 0, 0, 97, 0, 102, 0, 98, 0, 97, 0, 108, 0, 126, 0, 109, 0, 0, 0, 97, 0, 107, 0, 117, 0, 97, 0, 112, 0, 101, 0, 109, 0, 0, 0, 97, 0, 111, 0, 99, 0, 110, 0, 110, 0, 126, 0, 111, 0, 0, 0, 97, 0, 114, 0, 101, 0, 118, 0, 101, 0, 108, 0, 97, 0, 0, 0, 97, 0, 114, 0, 101, 0, 118, 0, 109, 0, 100, 0, 97, 0, 0, 0, 97, 0, 122, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 97, 0, 122, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 0, 0, 97, 0, 122, 0, 115, 0, 97, 0, 107, 0, 126, 0, 108, 0, 0, 0, 97, 0, 122, 0, 122, 0, 97, 0, 113, 0, 126, 0, 114, 0, 0, 0, 98, 0, 97, 0, 108, 0, 97, 0, 110, 0, 107, 0, 97, 0, 0, 0, 98, 0, 102, 0, 98, 0, 97, 0, 108, 0, 126, 0, 110, 0, 0, 0, 98, 0, 102, 0, 107, 0, 111, 0, 115, 0, 126, 0, 116, 0, 0, 0, 98, 0, 111, 0, 104, 0, 111, 0, 114, 0, 105, 0, 99, 0, 0, 0, 98, 0, 115, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 0, 0, 99, 0, 122, 0, 49, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 99, 0, 122, 0, 49, 0, 48, 0, 97, 0, 126, 0, 102, 0, 0, 0, 99, 0, 122, 0, 49, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 99, 0, 122, 0, 49, 0, 50, 0, 48, 0, 126, 0, 50, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 99, 0, 122, 0, 50, 0, 48, 0, 97, 0, 126, 0, 99, 0, 0, 0, 99, 0, 122, 0, 51, 0, 49, 0, 49, 0, 126, 0, 55, 0, 0, 0, 99, 0, 122, 0, 51, 0, 50, 0, 49, 0, 126, 0, 55, 0, 0, 0, 99, 0, 122, 0, 52, 0, 49, 0, 49, 0, 126, 0, 51, 0, 0, 0, 99, 0, 122, 0, 52, 0, 50, 0, 49, 0, 126, 0, 55, 0, 0, 0, 99, 0, 122, 0, 53, 0, 49, 0, 49, 0, 126, 0, 52, 0, 0, 0, 99, 0, 122, 0, 53, 0, 50, 0, 49, 0, 126, 0, 53, 0, 0, 0, 99, 0, 122, 0, 53, 0, 51, 0, 49, 0, 126, 0, 52, 0, 0, 0, 99, 0, 122, 0, 54, 0, 49, 0, 49, 0, 126, 0, 53, 0, 0, 0, 99, 0, 122, 0, 54, 0, 50, 0, 49, 0, 126, 0, 55, 0, 0, 0, 99, 0, 122, 0, 54, 0, 51, 0, 49, 0, 126, 0, 53, 0, 0, 0, 99, 0, 122, 0, 54, 0, 52, 0, 49, 0, 126, 0, 55, 0, 0, 0, 99, 0, 122, 0, 55, 0, 49, 0, 49, 0, 126, 0, 53, 0, 0, 0, 99, 0, 122, 0, 55, 0, 50, 0, 49, 0, 126, 0, 52, 0, 0, 0, 99, 0, 122, 0, 56, 0, 48, 0, 49, 0, 126, 0, 54, 0, 0, 0, 101, 0, 109, 0, 111, 0, 100, 0, 101, 0, 110, 0, 103, 0, 0, 0, 101, 0, 110, 0, 95, 0, 42, 0, 95, 0, 85, 0, 83, 0, 0, 0, 102, 0, 102, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 102, 0, 111, 0, 110, 0, 110, 0, 97, 0, 112, 0, 97, 0, 0, 0, 103, 0, 98, 0, 97, 0, 98, 0, 99, 0, 126, 0, 101, 0, 0, 0, 103, 0, 98, 0, 98, 0, 100, 0, 102, 0, 126, 0, 103, 0, 0, 0, 103, 0, 98, 0, 98, 0, 114, 0, 99, 0, 126, 0, 100, 0, 0, 0, 103, 0, 98, 0, 100, 0, 110, 0, 99, 0, 126, 0, 100, 0, 0, 0, 103, 0, 98, 0, 101, 0, 110, 0, 102, 0, 126, 0, 103, 0, 0, 0, 103, 0, 98, 0, 104, 0, 97, 0, 108, 0, 126, 0, 109, 0, 0, 0, 103, 0, 98, 0, 115, 0, 108, 0, 102, 0, 126, 0, 103, 0, 0, 0, 103, 0, 98, 0, 115, 0, 111, 0, 108, 0, 126, 0, 109, 0, 0, 0, 103, 0, 98, 0, 115, 0, 116, 0, 103, 0, 126, 0, 104, 0, 0, 0, 103, 0, 98, 0, 115, 0, 116, 0, 115, 0, 126, 0, 116, 0, 0, 0, 103, 0, 114, 0, 99, 0, 108, 0, 97, 0, 115, 0, 115, 0, 0, 0, 103, 0, 114, 0, 109, 0, 105, 0, 115, 0, 116, 0, 114, 0, 0, 0, 104, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 104, 0, 101, 0, 112, 0, 98, 0, 117, 0, 114, 0, 110, 0, 0, 0, 105, 0, 116, 0, 105, 0, 104, 0, 97, 0, 115, 0, 97, 0, 0, 0, 105, 0, 117, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 107, 0, 107, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 107, 0, 117, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 107, 0, 121, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 107, 0, 121, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 108, 0, 97, 0, 117, 0, 107, 0, 105, 0, 107, 0, 97, 0, 0, 0, 108, 0, 101, 0, 109, 0, 111, 0, 115, 0, 105, 0, 110, 0, 0, 0, 108, 0, 118, 0, 48, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 52, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 53, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 54, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 55, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 56, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 118, 0, 48, 0, 57, 0, 48, 0, 126, 0, 57, 0, 0, 0, 108, 0, 118, 0, 49, 0, 48, 0, 48, 0, 126, 0, 57, 0, 0, 0, 109, 0, 97, 0, 98, 0, 101, 0, 114, 0, 126, 0, 115, 0, 0, 0, 109, 0, 97, 0, 103, 0, 117, 0, 101, 0, 126, 0, 102, 0, 0, 0, 109, 0, 97, 0, 107, 0, 104, 0, 110, 0, 126, 0, 111, 0, 0, 0, 109, 0, 97, 0, 115, 0, 105, 0, 107, 0, 126, 0, 108, 0, 0, 0, 109, 0, 104, 0, 97, 0, 108, 0, 107, 0, 126, 0, 108, 0, 0, 0, 109, 0, 108, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 109, 0, 110, 0, 48, 0, 52, 0, 54, 0, 126, 0, 55, 0, 0, 0, 109, 0, 110, 0, 48, 0, 54, 0, 51, 0, 126, 0, 53, 0, 0, 0, 109, 0, 111, 0, 110, 0, 111, 0, 116, 0, 111, 0, 110, 0, 0, 0, 109, 0, 115, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 110, 0, 108, 0, 98, 0, 113, 0, 49, 0, 126, 0, 51, 0, 0, 0, 112, 0, 97, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 112, 0, 101, 0, 108, 0, 97, 0, 108, 0, 126, 0, 109, 0, 0, 0, 112, 0, 103, 0, 109, 0, 112, 0, 108, 0, 126, 0, 109, 0, 0, 0, 112, 0, 104, 0, 98, 0, 117, 0, 107, 0, 126, 0, 108, 0, 0, 0, 112, 0, 104, 0, 99, 0, 97, 0, 109, 0, 126, 0, 110, 0, 0, 0, 112, 0, 104, 0, 99, 0, 97, 0, 115, 0, 126, 0, 116, 0, 0, 0, 112, 0, 104, 0, 112, 0, 97, 0, 109, 0, 126, 0, 110, 0, 0, 0, 112, 0, 111, 0, 108, 0, 121, 0, 116, 0, 111, 0, 110, 0, 0, 0, 112, 0, 114, 0, 111, 0, 118, 0, 101, 0, 110, 0, 99, 0, 0, 0, 112, 0, 115, 0, 106, 0, 101, 0, 109, 0, 126, 0, 110, 0, 0, 0, 112, 0, 119, 0, 50, 0, 50, 0, 54, 0, 126, 0, 56, 0, 0, 0, 115, 0, 100, 0, 95, 0, 68, 0, 101, 0, 118, 0, 97, 0, 0, 0, 115, 0, 105, 0, 48, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 52, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 53, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 54, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 55, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 56, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 48, 0, 57, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 48, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 51, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 48, 0, 126, 0, 52, 0, 0, 0, 115, 0, 105, 0, 49, 0, 52, 0, 54, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 53, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 54, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 55, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 56, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 49, 0, 57, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 50, 0, 48, 0, 48, 0, 126, 0, 57, 0, 0, 0, 115, 0, 105, 0, 50, 0, 49, 0, 48, 0, 126, 0, 51, 0, 0, 0, 115, 0, 111, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 115, 0, 117, 0, 114, 0, 115, 0, 105, 0, 108, 0, 118, 0, 0, 0, 115, 0, 117, 0, 116, 0, 115, 0, 105, 0, 108, 0, 118, 0, 0, 0, 115, 0, 119, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 116, 0, 103, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 117, 0, 103, 0, 49, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 49, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 49, 0, 50, 0, 48, 0, 126, 0, 54, 0, 0, 0, 117, 0, 103, 0, 50, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 50, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 50, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 50, 0, 51, 0, 48, 0, 126, 0, 54, 0, 0, 0, 117, 0, 103, 0, 51, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 51, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 51, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 51, 0, 51, 0, 48, 0, 126, 0, 52, 0, 0, 0, 117, 0, 103, 0, 52, 0, 48, 0, 49, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 52, 0, 49, 0, 48, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 52, 0, 50, 0, 48, 0, 126, 0, 57, 0, 0, 0, 117, 0, 103, 0, 52, 0, 51, 0, 48, 0, 126, 0, 50, 0, 0, 0, 117, 0, 103, 0, 95, 0, 67, 0, 121, 0, 114, 0, 108, 0, 0, 0, 118, 0, 97, 0, 105, 0, 100, 0, 105, 0, 107, 0, 97, 0, 0, 0, 119, 0, 111, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 100, 0, 121, 0, 111, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 105, 0, 106, 0, 101, 0, 107, 0, 97, 0, 118, 0, 115, 0, 107, 0, 0, 0, 98, 0, 111, 0, 114, 0, 110, 0, 104, 0, 111, 0, 108, 0, 109, 0, 0, 0, 120, 0, 115, 0, 105, 0, 115, 0, 116, 0, 101, 0, 109, 0, 111, 0, 0, 0, 107, 0, 111, 0, 99, 0, 105, 0, 101, 0, 119, 0, 105, 0, 101, 0, 0, 0, 49, 0, 54, 0, 48, 0, 54, 0, 110, 0, 105, 0, 99, 0, 116, 0, 0, 0, 49, 0, 54, 0, 57, 0, 52, 0, 97, 0, 99, 0, 97, 0, 100, 0, 0, 0, 49, 0, 57, 0, 53, 0, 57, 0, 97, 0, 99, 0, 97, 0, 100, 0, 0, 0, 98, 0, 97, 0, 107, 0, 117, 0, 49, 0, 57, 0, 50, 0, 54, 0, 0, 0, 98, 0, 97, 0, 115, 0, 105, 0, 99, 0, 101, 0, 110, 0, 103, 0, 0, 0, 98, 0, 105, 0, 115, 0, 99, 0, 97, 0, 121, 0, 97, 0, 110, 0, 0, 0, 98, 0, 108, 0, 116, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 98, 0, 117, 0, 100, 0, 100, 0, 104, 0, 105, 0, 115, 0, 116, 0, 0, 0, 98, 0, 121, 0, 110, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 99, 0, 111, 0, 108, 0, 98, 0, 49, 0, 57, 0, 52, 0, 53, 0, 0, 0, 100, 0, 106, 0, 101, 0, 95, 0, 65, 0, 114, 0, 97, 0, 98, 0, 0, 0, 101, 0, 110, 0, 95, 0, 42, 0, 95, 0, 48, 0, 48, 0, 49, 0, 0, 0, 101, 0, 115, 0, 95, 0, 42, 0, 95, 0, 52, 0, 49, 0, 57, 0, 0, 0, 101, 0, 116, 0, 104, 0, 105, 0, 111, 0, 112, 0, 105, 0, 99, 0, 0, 0, 102, 0, 111, 0, 110, 0, 107, 0, 105, 0, 114, 0, 115, 0, 104, 0, 0, 0, 102, 0, 111, 0, 110, 0, 120, 0, 115, 0, 97, 0, 109, 0, 112, 0, 0, 0, 104, 0, 111, 0, 103, 0, 110, 0, 111, 0, 114, 0, 115, 0, 107, 0, 0, 0, 104, 0, 115, 0, 105, 0, 115, 0, 116, 0, 101, 0, 109, 0, 111, 0, 0, 0, 105, 0, 118, 0, 97, 0, 110, 0, 99, 0, 104, 0, 111, 0, 118, 0, 0, 0, 106, 0, 97, 0, 112, 0, 97, 0, 110, 0, 101, 0, 115, 0, 101, 0, 0, 0, 106, 0, 121, 0, 117, 0, 116, 0, 112, 0, 105, 0, 110, 0, 103, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 97, 0, 100, 0, 111, 0, 99, 0, 0, 0, 108, 0, 117, 0, 110, 0, 97, 0, 49, 0, 57, 0, 49, 0, 56, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 116, 0, 111, 0, 110, 0, 0, 0, 110, 0, 101, 0, 119, 0, 102, 0, 111, 0, 117, 0, 110, 0, 100, 0, 0, 0, 111, 0, 120, 0, 101, 0, 110, 0, 100, 0, 105, 0, 99, 0, 116, 0, 0, 0, 112, 0, 101, 0, 116, 0, 114, 0, 49, 0, 55, 0, 48, 0, 56, 0, 0, 0, 115, 0, 99, 0, 111, 0, 116, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 115, 0, 104, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 115, 0, 112, 0, 97, 0, 110, 0, 103, 0, 108, 0, 105, 0, 115, 0, 0, 0, 115, 0, 117, 0, 114, 0, 109, 0, 105, 0, 114, 0, 97, 0, 110, 0, 0, 0, 118, 0, 97, 0, 105, 0, 95, 0, 76, 0, 97, 0, 116, 0, 110, 0, 0, 0, 118, 0, 97, 0, 108, 0, 101, 0, 110, 0, 99, 0, 105, 0, 97, 0, 0, 0, 118, 0, 97, 0, 108, 0, 108, 0, 97, 0, 100, 0, 101, 0, 114, 0, 0, 0, 118, 0, 105, 0, 118, 0, 97, 0, 114, 0, 97, 0, 117, 0, 112, 0, 0, 0, 119, 0, 97, 0, 100, 0, 101, 0, 103, 0, 105, 0, 108, 0, 101, 0, 0, 0, 121, 0, 117, 0, 101, 0, 95, 0, 72, 0, 97, 0, 110, 0, 115, 0, 0, 0, 103, 0, 114, 0, 101, 0, 103, 0, 111, 0, 114, 0, 105, 0, 97, 0, 110, 0, 0, 0, 102, 0, 111, 0, 111, 0, 116, 0, 32, 0, 105, 0, 110, 0, 99, 0, 104, 0, 0, 0, 108, 0, 117, 0, 110, 0, 105, 0, 115, 0, 111, 0, 108, 0, 97, 0, 114, 0, 0, 0, 97, 0, 114, 0, 101, 0, 97, 0, 45, 0, 97, 0, 99, 0, 114, 0, 101, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 42, 0, 0, 0, 108, 0, 105, 0, 103, 0, 104, 0, 116, 0, 45, 0, 108, 0, 117, 0, 120, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 103, 0, 114, 0, 97, 0, 109, 0, 0, 0, 112, 0, 97, 0, 104, 0, 97, 0, 119, 0, 104, 0, 50, 0, 126, 0, 52, 0, 0, 0, 119, 0, 101, 0, 101, 0, 107, 0, 79, 0, 102, 0, 89, 0, 101, 0, 97, 0, 114, 0, 0, 0, 119, 0, 101, 0, 101, 0, 107, 0, 79, 0, 102, 0, 68, 0, 97, 0, 116, 0, 101, 0, 0, 0, 101, 0, 110, 0, 95, 0, 42, 0, 95, 0, 36, 0, 101, 0, 110, 0, 85, 0, 83, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 72, 0, 75, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 77, 0, 79, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 112, 0, 111, 0, 117, 0, 110, 0, 100, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 115, 0, 116, 0, 111, 0, 110, 0, 101, 0, 0, 0, 97, 0, 114, 0, 101, 0, 97, 0, 45, 0, 100, 0, 117, 0, 110, 0, 97, 0, 109, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 99, 0, 97, 0, 114, 0, 97, 0, 116, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 111, 0, 117, 0, 110, 0, 99, 0, 101, 0, 0, 0, 112, 0, 111, 0, 119, 0, 101, 0, 114, 0, 45, 0, 119, 0, 97, 0, 116, 0, 116, 0, 0, 0, 115, 0, 112, 0, 101, 0, 101, 0, 100, 0, 45, 0, 107, 0, 110, 0, 111, 0, 116, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 99, 0, 117, 0, 112, 0, 0, 0, 119, 0, 101, 0, 101, 0, 107, 0, 79, 0, 102, 0, 77, 0, 111, 0, 110, 0, 116, 0, 104, 0, 0, 0, 101, 0, 110, 0, 95, 0, 42, 0, 95, 0, 36, 0, 33, 0, 101, 0, 110, 0, 85, 0, 83, 0, 0, 0, 112, 0, 111, 0, 117, 0, 110, 0, 100, 0, 32, 0, 111, 0, 117, 0, 110, 0, 99, 0, 101, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 102, 0, 111, 0, 111, 0, 116, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 121, 0, 97, 0, 114, 0, 100, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 98, 0, 105, 0, 116, 0, 0, 0, 103, 0, 114, 0, 97, 0, 112, 0, 104, 0, 105, 0, 99, 0, 115, 0, 45, 0, 101, 0, 109, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 105, 0, 110, 0, 99, 0, 104, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 109, 0, 105, 0, 108, 0, 101, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 100, 0, 97, 0, 108, 0, 116, 0, 111, 0, 110, 0, 0, 0, 115, 0, 116, 0, 111, 0, 110, 0, 101, 0, 32, 0, 112, 0, 111, 0, 117, 0, 110, 0, 100, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 112, 0, 105, 0, 110, 0, 116, 0, 0, 0, 105, 0, 115, 0, 108, 0, 97, 0, 109, 0, 105, 0, 99, 0, 45, 0, 116, 0, 98, 0, 108, 0, 97, 0, 0, 0, 97, 0, 114, 0, 101, 0, 97, 0, 45, 0, 104, 0, 101, 0, 99, 0, 116, 0, 97, 0, 114, 0, 101, 0, 0, 0, 97, 0, 110, 0, 103, 0, 108, 0, 101, 0, 45, 0, 100, 0, 101, 0, 103, 0, 114, 0, 101, 0, 101, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 113, 0, 117, 0, 97, 0, 114, 0, 116, 0, 0, 0, 97, 0, 110, 0, 103, 0, 108, 0, 101, 0, 45, 0, 114, 0, 97, 0, 100, 0, 105, 0, 97, 0, 110, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 98, 0, 121, 0, 116, 0, 101, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 100, 0, 97, 0, 121, 0, 0, 0, 101, 0, 108, 0, 101, 0, 99, 0, 116, 0, 114, 0, 105, 0, 99, 0, 45, 0, 111, 0, 104, 0, 109, 0, 0, 0, 101, 0, 110, 0, 101, 0, 114, 0, 103, 0, 121, 0, 45, 0, 106, 0, 111, 0, 117, 0, 108, 0, 101, 0, 0, 0, 102, 0, 111, 0, 114, 0, 99, 0, 101, 0, 45, 0, 110, 0, 101, 0, 119, 0, 116, 0, 111, 0, 110, 0, 0, 0, 105, 0, 115, 0, 108, 0, 97, 0, 109, 0, 105, 0, 99, 0, 45, 0, 114, 0, 103, 0, 115, 0, 97, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 112, 0, 111, 0, 105, 0, 110, 0, 116, 0, 0, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 117, 0, 114, 0, 101, 0, 45, 0, 98, 0, 97, 0, 114, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 108, 0, 105, 0, 116, 0, 101, 0, 114, 0, 0, 0, 105, 0, 115, 0, 108, 0, 97, 0, 109, 0, 105, 0, 99, 0, 45, 0, 99, 0, 105, 0, 118, 0, 105, 0, 108, 0, 0, 0, 107, 0, 105, 0, 108, 0, 111, 0, 103, 0, 114, 0, 97, 0, 109, 0, 32, 0, 103, 0, 114, 0, 97, 0, 109, 0, 0, 0, 97, 0, 114, 0, 95, 0, 42, 0, 95, 0, 36, 0, 109, 0, 97, 0, 103, 0, 104, 0, 114, 0, 101, 0, 98, 0, 0, 0, 109, 0, 105, 0, 110, 0, 117, 0, 116, 0, 101, 0, 32, 0, 115, 0, 101, 0, 99, 0, 111, 0, 110, 0, 100, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 103, 0, 114, 0, 97, 0, 109, 0, 0, 0, 99, 0, 111, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 114, 0, 45, 0, 109, 0, 111, 0, 108, 0, 101, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 104, 0, 111, 0, 117, 0, 114, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 119, 0, 101, 0, 101, 0, 107, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 121, 0, 101, 0, 97, 0, 114, 0, 0, 0, 101, 0, 108, 0, 101, 0, 99, 0, 116, 0, 114, 0, 105, 0, 99, 0, 45, 0, 118, 0, 111, 0, 108, 0, 116, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 102, 0, 97, 0, 116, 0, 104, 0, 111, 0, 109, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 112, 0, 97, 0, 114, 0, 115, 0, 101, 0, 99, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 98, 0, 97, 0, 114, 0, 114, 0, 101, 0, 108, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 98, 0, 117, 0, 115, 0, 104, 0, 101, 0, 108, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 103, 0, 97, 0, 108, 0, 108, 0, 111, 0, 110, 0, 0, 0, 119, 0, 101, 0, 101, 0, 107, 0, 79, 0, 102, 0, 73, 0, 110, 0, 116, 0, 101, 0, 114, 0, 118, 0, 97, 0, 108, 0, 0, 0, 97, 0, 114, 0, 95, 0, 42, 0, 95, 0, 36, 0, 33, 0, 109, 0, 97, 0, 103, 0, 104, 0, 114, 0, 101, 0, 98, 0, 0, 0, 101, 0, 115, 0, 95, 0, 42, 0, 95, 0, 36, 0, 97, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 115, 0, 0, 0, 112, 0, 116, 0, 95, 0, 42, 0, 95, 0, 36, 0, 97, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 115, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 36, 0, 99, 0, 110, 0, 115, 0, 97, 0, 114, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 109, 0, 105, 0, 99, 0, 114, 0, 111, 0, 103, 0, 114, 0, 97, 0, 109, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 102, 0, 117, 0, 114, 0, 108, 0, 111, 0, 110, 0, 103, 0, 0, 0, 99, 0, 111, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 114, 0, 45, 0, 107, 0, 97, 0, 114, 0, 97, 0, 116, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 109, 0, 111, 0, 110, 0, 116, 0, 104, 0, 0, 0, 101, 0, 110, 0, 101, 0, 114, 0, 103, 0, 121, 0, 45, 0, 99, 0, 97, 0, 108, 0, 111, 0, 114, 0, 105, 0, 101, 0, 0, 0, 103, 0, 114, 0, 97, 0, 112, 0, 104, 0, 105, 0, 99, 0, 115, 0, 45, 0, 112, 0, 105, 0, 120, 0, 101, 0, 108, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 103, 0, 114, 0, 97, 0, 109, 0, 0, 0, 112, 0, 111, 0, 119, 0, 101, 0, 114, 0, 45, 0, 103, 0, 105, 0, 103, 0, 97, 0, 119, 0, 97, 0, 116, 0, 116, 0, 0, 0, 112, 0, 111, 0, 119, 0, 101, 0, 114, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 119, 0, 97, 0, 116, 0, 116, 0, 0, 0, 112, 0, 111, 0, 119, 0, 101, 0, 114, 0, 45, 0, 109, 0, 101, 0, 103, 0, 97, 0, 119, 0, 97, 0, 116, 0, 116, 0, 0, 0, 101, 0, 115, 0, 95, 0, 42, 0, 95, 0, 36, 0, 33, 0, 97, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 115, 0, 0, 0, 112, 0, 116, 0, 95, 0, 42, 0, 95, 0, 36, 0, 33, 0, 97, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 115, 0, 0, 0, 122, 0, 104, 0, 95, 0, 72, 0, 97, 0, 110, 0, 116, 0, 95, 0, 36, 0, 33, 0, 99, 0, 110, 0, 115, 0, 97, 0, 114, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 109, 0, 105, 0, 110, 0, 117, 0, 116, 0, 101, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 103, 0, 105, 0, 103, 0, 97, 0, 98, 0, 105, 0, 116, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 98, 0, 105, 0, 116, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 109, 0, 101, 0, 103, 0, 97, 0, 98, 0, 105, 0, 116, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 116, 0, 101, 0, 114, 0, 97, 0, 98, 0, 105, 0, 116, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 100, 0, 101, 0, 99, 0, 97, 0, 100, 0, 101, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 115, 0, 101, 0, 99, 0, 111, 0, 110, 0, 100, 0, 0, 0, 101, 0, 108, 0, 101, 0, 99, 0, 116, 0, 114, 0, 105, 0, 99, 0, 45, 0, 97, 0, 109, 0, 112, 0, 101, 0, 114, 0, 101, 0, 0, 0, 101, 0, 110, 0, 101, 0, 114, 0, 103, 0, 121, 0, 45, 0, 116, 0, 104, 0, 101, 0, 114, 0, 109, 0, 45, 0, 117, 0, 115, 0, 0, 0, 102, 0, 114, 0, 101, 0, 113, 0, 117, 0, 101, 0, 110, 0, 99, 0, 121, 0, 45, 0, 104, 0, 101, 0, 114, 0, 116, 0, 122, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 101, 0, 97, 0, 114, 0, 116, 0, 104, 0, 45, 0, 109, 0, 97, 0, 115, 0, 115, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 109, 0, 101, 0, 116, 0, 114, 0, 105, 0, 99, 0, 45, 0, 116, 0, 111, 0, 110, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 111, 0, 117, 0, 110, 0, 99, 0, 101, 0, 45, 0, 116, 0, 114, 0, 111, 0, 121, 0, 0, 0, 109, 0, 97, 0, 115, 0, 115, 0, 45, 0, 115, 0, 111, 0, 108, 0, 97, 0, 114, 0, 45, 0, 109, 0, 97, 0, 115, 0, 115, 0, 0, 0, 112, 0, 111, 0, 119, 0, 101, 0, 114, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 119, 0, 97, 0, 116, 0, 116, 0, 0, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 117, 0, 114, 0, 101, 0, 45, 0, 112, 0, 97, 0, 115, 0, 99, 0, 97, 0, 108, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 116, 0, 101, 0, 97, 0, 115, 0, 112, 0, 111, 0, 111, 0, 110, 0, 0, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 32, 0, 99, 0, 101, 0, 110, 0, 116, 0, 105, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 105, 0, 115, 0, 108, 0, 97, 0, 109, 0, 105, 0, 99, 0, 45, 0, 117, 0, 109, 0, 97, 0, 108, 0, 113, 0, 117, 0, 114, 0, 97, 0, 0, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 117, 0, 114, 0, 101, 0, 45, 0, 105, 0, 110, 0, 99, 0, 104, 0, 45, 0, 104, 0, 103, 0, 0, 0, 97, 0, 110, 0, 103, 0, 108, 0, 101, 0, 45, 0, 97, 0, 114, 0, 99, 0, 45, 0, 109, 0, 105, 0, 110, 0, 117, 0, 116, 0, 101, 0, 0, 0, 97, 0, 110, 0, 103, 0, 108, 0, 101, 0, 45, 0, 97, 0, 114, 0, 99, 0, 45, 0, 115, 0, 101, 0, 99, 0, 111, 0, 110, 0, 100, 0, 0, 0, 97, 0, 110, 0, 103, 0, 108, 0, 101, 0, 45, 0, 114, 0, 101, 0, 118, 0, 111, 0, 108, 0, 117, 0, 116, 0, 105, 0, 111, 0, 110, 0, 0, 0, 97, 0, 114, 0, 101, 0, 97, 0, 45, 0, 115, 0, 113, 0, 117, 0, 97, 0, 114, 0, 101, 0, 45, 0, 102, 0, 111, 0, 111, 0, 116, 0, 0, 0, 97, 0, 114, 0, 101, 0, 97, 0, 45, 0, 115, 0, 113, 0, 117, 0, 97, 0, 114, 0, 101, 0, 45, 0, 105, 0, 110, 0, 99, 0, 104, 0, 0, 0, 97, 0, 114, 0, 101, 0, 97, 0, 45, 0, 115, 0, 113, 0, 117, 0, 97, 0, 114, 0, 101, 0, 45, 0, 109, 0, 105, 0, 108, 0, 101, 0, 0, 0, 97, 0, 114, 0, 101, 0, 97, 0, 45, 0, 115, 0, 113, 0, 117, 0, 97, 0, 114, 0, 101, 0, 45, 0, 121, 0, 97, 0, 114, 0, 100, 0, 0, 0, 99, 0, 111, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 114, 0, 45, 0, 112, 0, 101, 0, 114, 0, 99, 0, 101, 0, 110, 0, 116, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 103, 0, 105, 0, 103, 0, 97, 0, 98, 0, 121, 0, 116, 0, 101, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 98, 0, 121, 0, 116, 0, 101, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 109, 0, 101, 0, 103, 0, 97, 0, 98, 0, 121, 0, 116, 0, 101, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 112, 0, 101, 0, 116, 0, 97, 0, 98, 0, 121, 0, 116, 0, 101, 0, 0, 0, 100, 0, 105, 0, 103, 0, 105, 0, 116, 0, 97, 0, 108, 0, 45, 0, 116, 0, 101, 0, 114, 0, 97, 0, 98, 0, 121, 0, 116, 0, 101, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 99, 0, 101, 0, 110, 0, 116, 0, 117, 0, 114, 0, 121, 0, 0, 0, 101, 0, 110, 0, 101, 0, 114, 0, 103, 0, 121, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 106, 0, 111, 0, 117, 0, 108, 0, 101, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 100, 0, 101, 0, 99, 0, 105, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 110, 0, 97, 0, 110, 0, 111, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 112, 0, 105, 0, 99, 0, 111, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 112, 0, 111, 0, 119, 0, 101, 0, 114, 0, 45, 0, 104, 0, 111, 0, 114, 0, 115, 0, 101, 0, 112, 0, 111, 0, 119, 0, 101, 0, 114, 0, 0, 0, 112, 0, 114, 0, 111, 0, 112, 0, 111, 0, 114, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 107, 0, 97, 0, 114, 0, 97, 0, 116, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 97, 0, 99, 0, 114, 0, 101, 0, 45, 0, 102, 0, 111, 0, 111, 0, 116, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 100, 0, 101, 0, 99, 0, 105, 0, 108, 0, 105, 0, 116, 0, 101, 0, 114, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 109, 0, 101, 0, 103, 0, 97, 0, 108, 0, 105, 0, 116, 0, 101, 0, 114, 0, 0, 0, 100, 0, 101, 0, 95, 0, 102, 0, 97, 0, 99, 0, 116, 0, 111, 0, 95, 0, 111, 0, 102, 0, 102, 0, 105, 0, 99, 0, 105, 0, 97, 0, 108, 0, 0, 0, 111, 0, 102, 0, 102, 0, 105, 0, 99, 0, 105, 0, 97, 0, 108, 0, 95, 0, 114, 0, 101, 0, 103, 0, 105, 0, 111, 0, 110, 0, 97, 0, 108, 0, 0, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 117, 0, 114, 0, 101, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 98, 0, 97, 0, 114, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 109, 0, 105, 0, 99, 0, 114, 0, 111, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 77, 0, 65, 0, 43, 0, 68, 0, 90, 0, 43, 0, 84, 0, 78, 0, 43, 0, 76, 0, 89, 0, 43, 0, 77, 0, 82, 0, 43, 0, 69, 0, 72, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 104, 0, 101, 0, 99, 0, 116, 0, 111, 0, 108, 0, 105, 0, 116, 0, 101, 0, 114, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 116, 0, 97, 0, 98, 0, 108, 0, 101, 0, 115, 0, 112, 0, 111, 0, 111, 0, 110, 0, 0, 0, 97, 0, 114, 0, 101, 0, 97, 0, 45, 0, 115, 0, 113, 0, 117, 0, 97, 0, 114, 0, 101, 0, 45, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 99, 0, 111, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 114, 0, 45, 0, 112, 0, 101, 0, 114, 0, 109, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 102, 0, 111, 0, 114, 0, 99, 0, 101, 0, 45, 0, 112, 0, 111, 0, 117, 0, 110, 0, 100, 0, 45, 0, 102, 0, 111, 0, 114, 0, 99, 0, 101, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 99, 0, 101, 0, 110, 0, 116, 0, 105, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 108, 0, 105, 0, 103, 0, 104, 0, 116, 0, 45, 0, 121, 0, 101, 0, 97, 0, 114, 0, 0, 0, 116, 0, 111, 0, 114, 0, 113, 0, 117, 0, 101, 0, 45, 0, 112, 0, 111, 0, 117, 0, 110, 0, 100, 0, 45, 0, 102, 0, 111, 0, 111, 0, 116, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 99, 0, 101, 0, 110, 0, 116, 0, 105, 0, 108, 0, 105, 0, 116, 0, 101, 0, 114, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 99, 0, 117, 0, 98, 0, 105, 0, 99, 0, 45, 0, 102, 0, 111, 0, 111, 0, 116, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 99, 0, 117, 0, 98, 0, 105, 0, 99, 0, 45, 0, 105, 0, 110, 0, 99, 0, 104, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 99, 0, 117, 0, 98, 0, 105, 0, 99, 0, 45, 0, 109, 0, 105, 0, 108, 0, 101, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 99, 0, 117, 0, 98, 0, 105, 0, 99, 0, 45, 0, 121, 0, 97, 0, 114, 0, 100, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 99, 0, 117, 0, 112, 0, 45, 0, 109, 0, 101, 0, 116, 0, 114, 0, 105, 0, 99, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 108, 0, 105, 0, 116, 0, 101, 0, 114, 0, 0, 0, 101, 0, 110, 0, 101, 0, 114, 0, 103, 0, 121, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 99, 0, 97, 0, 108, 0, 111, 0, 114, 0, 105, 0, 101, 0, 0, 0, 101, 0, 110, 0, 101, 0, 114, 0, 103, 0, 121, 0, 45, 0, 102, 0, 111, 0, 111, 0, 100, 0, 99, 0, 97, 0, 108, 0, 111, 0, 114, 0, 105, 0, 101, 0, 0, 0, 116, 0, 101, 0, 109, 0, 112, 0, 101, 0, 114, 0, 97, 0, 116, 0, 117, 0, 114, 0, 101, 0, 45, 0, 107, 0, 101, 0, 108, 0, 118, 0, 105, 0, 110, 0, 0, 0, 99, 0, 111, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 114, 0, 45, 0, 112, 0, 101, 0, 114, 0, 109, 0, 121, 0, 114, 0, 105, 0, 97, 0, 100, 0, 0, 0, 103, 0, 114, 0, 97, 0, 112, 0, 104, 0, 105, 0, 99, 0, 115, 0, 45, 0, 109, 0, 101, 0, 103, 0, 97, 0, 112, 0, 105, 0, 120, 0, 101, 0, 108, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 99, 0, 117, 0, 98, 0, 105, 0, 99, 0, 45, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 102, 0, 108, 0, 117, 0, 105, 0, 100, 0, 45, 0, 111, 0, 117, 0, 110, 0, 99, 0, 101, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 112, 0, 105, 0, 110, 0, 116, 0, 45, 0, 109, 0, 101, 0, 116, 0, 114, 0, 105, 0, 99, 0, 0, 0, 115, 0, 112, 0, 101, 0, 101, 0, 100, 0, 45, 0, 109, 0, 105, 0, 108, 0, 101, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 104, 0, 111, 0, 117, 0, 114, 0, 0, 0, 116, 0, 101, 0, 109, 0, 112, 0, 101, 0, 114, 0, 97, 0, 116, 0, 117, 0, 114, 0, 101, 0, 45, 0, 99, 0, 101, 0, 108, 0, 115, 0, 105, 0, 117, 0, 115, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 100, 0, 97, 0, 121, 0, 45, 0, 112, 0, 101, 0, 114, 0, 115, 0, 111, 0, 110, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 110, 0, 97, 0, 110, 0, 111, 0, 115, 0, 101, 0, 99, 0, 111, 0, 110, 0, 100, 0, 0, 0, 101, 0, 110, 0, 101, 0, 114, 0, 103, 0, 121, 0, 45, 0, 101, 0, 108, 0, 101, 0, 99, 0, 116, 0, 114, 0, 111, 0, 110, 0, 118, 0, 111, 0, 108, 0, 116, 0, 0, 0, 102, 0, 114, 0, 101, 0, 113, 0, 117, 0, 101, 0, 110, 0, 99, 0, 121, 0, 45, 0, 103, 0, 105, 0, 103, 0, 97, 0, 104, 0, 101, 0, 114, 0, 116, 0, 122, 0, 0, 0, 102, 0, 114, 0, 101, 0, 113, 0, 117, 0, 101, 0, 110, 0, 99, 0, 121, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 104, 0, 101, 0, 114, 0, 116, 0, 122, 0, 0, 0, 102, 0, 114, 0, 101, 0, 113, 0, 117, 0, 101, 0, 110, 0, 99, 0, 121, 0, 45, 0, 109, 0, 101, 0, 103, 0, 97, 0, 104, 0, 101, 0, 114, 0, 116, 0, 122, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 115, 0, 111, 0, 108, 0, 97, 0, 114, 0, 45, 0, 114, 0, 97, 0, 100, 0, 105, 0, 117, 0, 115, 0, 0, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 117, 0, 114, 0, 101, 0, 45, 0, 97, 0, 116, 0, 109, 0, 111, 0, 115, 0, 112, 0, 104, 0, 101, 0, 114, 0, 101, 0, 0, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 117, 0, 114, 0, 101, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 112, 0, 97, 0, 115, 0, 99, 0, 97, 0, 108, 0, 0, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 117, 0, 114, 0, 101, 0, 45, 0, 109, 0, 101, 0, 103, 0, 97, 0, 112, 0, 97, 0, 115, 0, 99, 0, 97, 0, 108, 0, 0, 0, 116, 0, 101, 0, 109, 0, 112, 0, 101, 0, 114, 0, 97, 0, 116, 0, 117, 0, 114, 0, 101, 0, 45, 0, 103, 0, 101, 0, 110, 0, 101, 0, 114, 0, 105, 0, 99, 0, 0, 0, 116, 0, 111, 0, 114, 0, 113, 0, 117, 0, 101, 0, 45, 0, 110, 0, 101, 0, 119, 0, 116, 0, 111, 0, 110, 0, 45, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 110, 0, 97, 0, 117, 0, 116, 0, 105, 0, 99, 0, 97, 0, 108, 0, 45, 0, 109, 0, 105, 0, 108, 0, 101, 0, 0, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 117, 0, 114, 0, 101, 0, 45, 0, 104, 0, 101, 0, 99, 0, 116, 0, 111, 0, 112, 0, 97, 0, 115, 0, 99, 0, 97, 0, 108, 0, 0, 0, 97, 0, 99, 0, 99, 0, 101, 0, 108, 0, 101, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 103, 0, 45, 0, 102, 0, 111, 0, 114, 0, 99, 0, 101, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 109, 0, 105, 0, 99, 0, 114, 0, 111, 0, 115, 0, 101, 0, 99, 0, 111, 0, 110, 0, 100, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 115, 0, 101, 0, 99, 0, 111, 0, 110, 0, 100, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 119, 0, 101, 0, 101, 0, 107, 0, 45, 0, 112, 0, 101, 0, 114, 0, 115, 0, 111, 0, 110, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 121, 0, 101, 0, 97, 0, 114, 0, 45, 0, 112, 0, 101, 0, 114, 0, 115, 0, 111, 0, 110, 0, 0, 0, 101, 0, 108, 0, 101, 0, 99, 0, 116, 0, 114, 0, 105, 0, 99, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 97, 0, 109, 0, 112, 0, 101, 0, 114, 0, 101, 0, 0, 0, 101, 0, 110, 0, 101, 0, 114, 0, 103, 0, 121, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 119, 0, 97, 0, 116, 0, 116, 0, 45, 0, 104, 0, 111, 0, 117, 0, 114, 0, 0, 0, 103, 0, 114, 0, 97, 0, 112, 0, 104, 0, 105, 0, 99, 0, 115, 0, 45, 0, 100, 0, 111, 0, 116, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 105, 0, 110, 0, 99, 0, 104, 0, 0, 0, 97, 0, 114, 0, 101, 0, 97, 0, 45, 0, 115, 0, 113, 0, 117, 0, 97, 0, 114, 0, 101, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 100, 0, 117, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 109, 0, 111, 0, 110, 0, 116, 0, 104, 0, 45, 0, 112, 0, 101, 0, 114, 0, 115, 0, 111, 0, 110, 0, 0, 0, 115, 0, 112, 0, 101, 0, 101, 0, 100, 0, 45, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 115, 0, 101, 0, 99, 0, 111, 0, 110, 0, 100, 0, 0, 0, 116, 0, 101, 0, 109, 0, 112, 0, 101, 0, 114, 0, 97, 0, 116, 0, 117, 0, 114, 0, 101, 0, 45, 0, 102, 0, 97, 0, 104, 0, 114, 0, 101, 0, 110, 0, 104, 0, 101, 0, 105, 0, 116, 0, 0, 0, 97, 0, 114, 0, 101, 0, 97, 0, 45, 0, 115, 0, 113, 0, 117, 0, 97, 0, 114, 0, 101, 0, 45, 0, 99, 0, 101, 0, 110, 0, 116, 0, 105, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 108, 0, 105, 0, 103, 0, 104, 0, 116, 0, 45, 0, 115, 0, 111, 0, 108, 0, 97, 0, 114, 0, 45, 0, 108, 0, 117, 0, 109, 0, 105, 0, 110, 0, 111, 0, 115, 0, 105, 0, 116, 0, 121, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 99, 0, 117, 0, 98, 0, 105, 0, 99, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 103, 0, 97, 0, 108, 0, 108, 0, 111, 0, 110, 0, 45, 0, 105, 0, 109, 0, 112, 0, 101, 0, 114, 0, 105, 0, 97, 0, 108, 0, 0, 0, 65, 0, 83, 0, 43, 0, 71, 0, 85, 0, 43, 0, 77, 0, 72, 0, 43, 0, 77, 0, 80, 0, 43, 0, 80, 0, 82, 0, 43, 0, 85, 0, 77, 0, 43, 0, 85, 0, 83, 0, 43, 0, 86, 0, 73, 0, 0, 0, 103, 0, 114, 0, 97, 0, 112, 0, 104, 0, 105, 0, 99, 0, 115, 0, 45, 0, 112, 0, 105, 0, 120, 0, 101, 0, 108, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 105, 0, 110, 0, 99, 0, 104, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 99, 0, 117, 0, 98, 0, 105, 0, 99, 0, 45, 0, 99, 0, 101, 0, 110, 0, 116, 0, 105, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 115, 0, 112, 0, 101, 0, 101, 0, 100, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 104, 0, 111, 0, 117, 0, 114, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 109, 0, 105, 0, 108, 0, 101, 0, 45, 0, 115, 0, 99, 0, 97, 0, 110, 0, 100, 0, 105, 0, 110, 0, 97, 0, 118, 0, 105, 0, 97, 0, 110, 0, 0, 0, 108, 0, 101, 0, 110, 0, 103, 0, 116, 0, 104, 0, 45, 0, 97, 0, 115, 0, 116, 0, 114, 0, 111, 0, 110, 0, 111, 0, 109, 0, 105, 0, 99, 0, 97, 0, 108, 0, 45, 0, 117, 0, 110, 0, 105, 0, 116, 0, 0, 0, 121, 0, 101, 0, 97, 0, 114, 0, 45, 0, 112, 0, 101, 0, 114, 0, 115, 0, 111, 0, 110, 0, 32, 0, 109, 0, 111, 0, 110, 0, 116, 0, 104, 0, 45, 0, 112, 0, 101, 0, 114, 0, 115, 0, 111, 0, 110, 0, 0, 0, 99, 0, 111, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 114, 0, 45, 0, 112, 0, 97, 0, 114, 0, 116, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 111, 0, 110, 0, 0, 0, 103, 0, 114, 0, 97, 0, 112, 0, 104, 0, 105, 0, 99, 0, 115, 0, 45, 0, 100, 0, 111, 0, 116, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 99, 0, 101, 0, 110, 0, 116, 0, 105, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 99, 0, 111, 0, 110, 0, 115, 0, 117, 0, 109, 0, 112, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 109, 0, 105, 0, 108, 0, 101, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 103, 0, 97, 0, 108, 0, 108, 0, 111, 0, 110, 0, 0, 0, 101, 0, 110, 0, 101, 0, 114, 0, 103, 0, 121, 0, 45, 0, 98, 0, 114, 0, 105, 0, 116, 0, 105, 0, 115, 0, 104, 0, 45, 0, 116, 0, 104, 0, 101, 0, 114, 0, 109, 0, 97, 0, 108, 0, 45, 0, 117, 0, 110, 0, 105, 0, 116, 0, 0, 0, 118, 0, 111, 0, 108, 0, 117, 0, 109, 0, 101, 0, 45, 0, 102, 0, 108, 0, 117, 0, 105, 0, 100, 0, 45, 0, 111, 0, 117, 0, 110, 0, 99, 0, 101, 0, 45, 0, 105, 0, 109, 0, 112, 0, 101, 0, 114, 0, 105, 0, 97, 0, 108, 0, 0, 0, 99, 0, 111, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 114, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 109, 0, 111, 0, 108, 0, 101, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 108, 0, 105, 0, 116, 0, 101, 0, 114, 0, 0, 0, 103, 0, 114, 0, 97, 0, 112, 0, 104, 0, 105, 0, 99, 0, 115, 0, 45, 0, 112, 0, 105, 0, 120, 0, 101, 0, 108, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 99, 0, 101, 0, 110, 0, 116, 0, 105, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 117, 0, 114, 0, 101, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 45, 0, 111, 0, 102, 0, 45, 0, 109, 0, 101, 0, 114, 0, 99, 0, 117, 0, 114, 0, 121, 0, 0, 0, 112, 0, 114, 0, 101, 0, 115, 0, 115, 0, 117, 0, 114, 0, 101, 0, 45, 0, 112, 0, 111, 0, 117, 0, 110, 0, 100, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 115, 0, 113, 0, 117, 0, 97, 0, 114, 0, 101, 0, 45, 0, 105, 0, 110, 0, 99, 0, 104, 0, 0, 0, 99, 0, 111, 0, 110, 0, 115, 0, 117, 0, 109, 0, 112, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 108, 0, 105, 0, 116, 0, 101, 0, 114, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 107, 0, 105, 0, 108, 0, 111, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 0, 0, 99, 0, 111, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 114, 0, 45, 0, 109, 0, 105, 0, 108, 0, 108, 0, 105, 0, 103, 0, 114, 0, 97, 0, 109, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 100, 0, 101, 0, 99, 0, 105, 0, 108, 0, 105, 0, 116, 0, 101, 0, 114, 0, 0, 0, 99, 0, 111, 0, 110, 0, 115, 0, 117, 0, 109, 0, 112, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 108, 0, 105, 0, 116, 0, 101, 0, 114, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 49, 0, 48, 0, 48, 0, 107, 0, 105, 0, 108, 0, 111, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 115, 0, 0, 0, 99, 0, 111, 0, 110, 0, 115, 0, 117, 0, 109, 0, 112, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 109, 0, 105, 0, 108, 0, 101, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 103, 0, 97, 0, 108, 0, 108, 0, 111, 0, 110, 0, 45, 0, 105, 0, 109, 0, 112, 0, 101, 0, 114, 0, 105, 0, 97, 0, 108, 0, 0, 0, 97, 0, 99, 0, 99, 0, 101, 0, 108, 0, 101, 0, 114, 0, 97, 0, 116, 0, 105, 0, 111, 0, 110, 0, 45, 0, 109, 0, 101, 0, 116, 0, 101, 0, 114, 0, 45, 0, 112, 0, 101, 0, 114, 0, 45, 0, 115, 0, 101, 0, 99, 0, 111, 0, 110, 0, 100, 0, 45, 0, 115, 0, 113, 0, 117, 0, 97, 0, 114, 0, 101, 0, 100, 0, 0, 0, 1, 0, 23, 199, 5, 0, 23, 199, 77, 204, 221, 191, 73, 201, 134, 200, 5, 0, 213, 192, 23, 199, 221, 191, 73, 201, 134, 200, 5, 0, 23, 199, 77, 204, 221, 191, 73, 201, 134, 200, 2, 0, 23, 199, 5, 192, 2, 0, 23, 199, 5, 192, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 5, 0, 23, 199, 186, 174, 221, 191, 73, 201, 134, 200, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 2, 0, 23, 199, 63, 198, 2, 0, 23, 199, 5, 192, 5, 0, 23, 199, 32, 181, 221, 191, 73, 201, 134, 200, 2, 0, 23, 199, 7, 182, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 5, 0, 213, 192, 23, 199, 221, 191, 73, 201, 134, 200, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 2, 0, 23, 199, 117, 198, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 2, 0, 23, 199, 199, 139, 5, 0, 23, 199, 77, 204, 221, 191, 73, 201, 134, 200, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 2, 0, 23, 199, 5, 192, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 5, 0, 23, 199, 77, 204, 221, 191, 73, 201, 134, 200, 4, 0, 77, 204, 23, 199, 221, 191, 8, 201, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 2, 0, 23, 199, 5, 192, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 2, 0, 9, 198, 23, 199, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 3, 0, 23, 199, 175, 32, 5, 192, 4, 0, 23, 199, 221, 191, 73, 201, 134, 200, 3, 0, 84, 3, 123, 4, 239, 11, 3, 0, 176, 1, 203, 131, 19, 12, 3, 0, 67, 0, 75, 4, 43, 12, 3, 0, 108, 1, 103, 111, 255, 11, 3, 0, 46, 109, 215, 131, 67, 12, 3, 0, 52, 109, 223, 10, 7, 12, 3, 0, 88, 2, 179, 131, 11, 12, 3, 0, 52, 2, 13, 132, 51, 12, 3, 0, 120, 2, 135, 10, 23, 12, 3, 0, 60, 107, 209, 131, 3, 12, 3, 0, 139, 0, 109, 111, 63, 12, 3, 0, 92, 108, 227, 131, 47, 12, 3, 0, 97, 0, 197, 131, 59, 12, 3, 0, 124, 106, 251, 131, 79, 12, 3, 0, 58, 109, 245, 131, 75, 12, 3, 0, 100, 0, 123, 112, 243, 11, 3, 0, 142, 0, 15, 10, 27, 2, 3, 0, 212, 1, 221, 131, 83, 12, 3, 0, 156, 1, 61, 132, 139, 12, 3, 0, 64, 109, 19, 132, 167, 12, 3, 0, 160, 1, 115, 111, 123, 12, 3, 0, 130, 106, 37, 132, 23, 5, 3, 0, 43, 0, 79, 11, 119, 12, 3, 0, 22, 0, 121, 111, 143, 1, 3, 0, 103, 0, 7, 132, 131, 12, 3, 0, 46, 0, 207, 112, 99, 12, 3, 0, 70, 0, 77, 131, 111, 12, 3, 0, 145, 0, 215, 10, 143, 12, 3, 0, 218, 111, 43, 132, 155, 12, 3, 0, 44, 2, 109, 132, 63, 5, 3, 0, 156, 107, 55, 132, 51, 5, 3, 0, 106, 0, 139, 10, 115, 12, 3, 0, 210, 107, 79, 132, 163, 12, 3, 0, 70, 109, 1, 132, 135, 12, 3, 0, 109, 0, 49, 132, 23, 3, 3, 0, 148, 0, 207, 109, 187, 12, 3, 0, 151, 0, 73, 132, 191, 12, 3, 0, 14, 108, 67, 132, 195, 12, 3, 0, 216, 1, 133, 132, 147, 12, 3, 0, 98, 108, 85, 132, 151, 12, 3, 0, 178, 106, 219, 112, 211, 12, 3, 0, 76, 109, 199, 132, 215, 12, 3, 0, 84, 1, 5, 113, 243, 12, 3, 0, 228, 1, 145, 132, 207, 12, 3, 0, 60, 2, 211, 132, 247, 12, 3, 0, 160, 106, 55, 4, 27, 3, 3, 0, 25, 0, 83, 10, 227, 12, 3, 0, 224, 111, 217, 132, 251, 12, 3, 0, 104, 108, 243, 112, 219, 12, 3, 0, 90, 107, 139, 132, 235, 12, 3, 0, 48, 1, 237, 109, 223, 12, 3, 0, 148, 109, 249, 112, 255, 12, 3, 0, 110, 108, 11, 113, 11, 13, 3, 0, 108, 2, 119, 4, 59, 2, 3, 0, 154, 109, 23, 113, 23, 13, 3, 0, 200, 108, 225, 112, 7, 13, 3, 0, 112, 0, 95, 131, 27, 13, 3, 0, 230, 111, 193, 132, 143, 9, 3, 0, 184, 106, 223, 132, 135, 5, 3, 0, 16, 0, 81, 112, 35, 13, 3, 0, 148, 2, 55, 10, 43, 13, 3, 0, 58, 106, 51, 10, 51, 13, 3, 0, 28, 0, 103, 3, 55, 13, 3, 0, 190, 106, 87, 112, 59, 13, 3, 0, 20, 108, 133, 111, 67, 1, 3, 0, 116, 108, 33, 109, 63, 13, 3, 0, 10, 0, 185, 131, 67, 13, 3, 0, 162, 107, 139, 111, 71, 13, 3, 0, 92, 1, 241, 132, 95, 13, 3, 0, 49, 0, 91, 4, 83, 13, 3, 0, 112, 206, 19, 11, 91, 13, 3, 0, 232, 106, 29, 113, 87, 13, 3, 0, 50, 192, 119, 131, 167, 5, 3, 0, 7, 0, 235, 132, 75, 1, 3, 0, 144, 2, 19, 3, 163, 6, 3, 0, 196, 106, 11, 10, 103, 13, 3, 0, 238, 106, 99, 3, 107, 13, 3, 0, 242, 111, 95, 3, 147, 1, 3, 0, 244, 106, 171, 10, 123, 13, 3, 0, 80, 2, 247, 132, 119, 13, 3, 0, 118, 106, 23, 10, 111, 13, 3, 0, 157, 0, 19, 10, 127, 13, 3, 0, 68, 2, 39, 10, 131, 13, 3, 0, 88, 106, 95, 4, 135, 13, 3, 0, 82, 109, 3, 133, 207, 5, 3, 0, 52, 0, 43, 10, 143, 13, 3, 0, 48, 2, 27, 10, 199, 13, 3, 0, 248, 111, 63, 11, 147, 13, 3, 0, 202, 106, 59, 10, 151, 13, 3, 0, 88, 109, 111, 3, 159, 13, 3, 0, 160, 0, 253, 132, 195, 13, 3, 0, 26, 108, 107, 3, 171, 13, 3, 0, 116, 1, 93, 112, 163, 13, 3, 0, 124, 2, 83, 131, 167, 13, 3, 0, 66, 107, 127, 132, 179, 13, 3, 0, 144, 1, 67, 10, 191, 13, 3, 0, 140, 2, 7, 10, 179, 15, 3, 0, 122, 108, 35, 113, 31, 2, 3, 0, 115, 0, 89, 131, 35, 2, 3, 0, 206, 108, 203, 10, 175, 13, 3, 0, 32, 108, 41, 113, 203, 13, 3, 0, 144, 199, 119, 3, 207, 1, 3, 0, 116, 2, 9, 133, 215, 13, 3, 0, 160, 109, 115, 3, 219, 13, 3, 0, 96, 1, 17, 113, 227, 13, 3, 0, 73, 0, 47, 113, 231, 13, 3, 0, 31, 0, 123, 3, 235, 13, 3, 0, 4, 0, 135, 3, 239, 13, 3, 0, 250, 106, 75, 10, 11, 14, 3, 0, 96, 107, 147, 3, 27, 14, 3, 0, 94, 109, 71, 11, 255, 13, 3, 0, 46, 106, 131, 3, 3, 14, 3, 0, 254, 111, 91, 132, 7, 14, 3, 0, 34, 0, 143, 3, 19, 14, 3, 0, 52, 1, 139, 3, 15, 14, 3, 0, 118, 0, 127, 3, 23, 14, 3, 0, 104, 1, 79, 10, 111, 2, 3, 0, 100, 109, 67, 11, 39, 14, 3, 0, 102, 107, 151, 3, 35, 14, 3, 0, 76, 0, 163, 3, 43, 14, 3, 0, 121, 0, 155, 3, 47, 14, 3, 0, 94, 106, 53, 113, 51, 14, 3, 0, 208, 1, 77, 113, 55, 14, 3, 0, 124, 0, 213, 112, 115, 2, 3, 0, 38, 108, 63, 10, 59, 14, 3, 0, 152, 1, 255, 112, 3, 13, 3, 0, 106, 109, 219, 10, 63, 14, 3, 0, 168, 1, 59, 113, 51, 15, 3, 0, 244, 1, 65, 113, 67, 14, 3, 0, 79, 0, 71, 113, 79, 14, 3, 0, 4, 112, 231, 112, 31, 13, 3, 0, 37, 0, 159, 3, 211, 1, 3, 0, 28, 2, 83, 113, 83, 14, 3, 0, 236, 1, 99, 112, 43, 2, 3, 0, 112, 109, 227, 10, 91, 14, 3, 0, 184, 1, 87, 10, 95, 14, 3, 0, 148, 1, 237, 112, 99, 14, 3, 0, 118, 109, 95, 113, 87, 14, 3, 0, 44, 108, 105, 112, 219, 1, 3, 0, 124, 1, 171, 3, 107, 14, 3, 0, 166, 106, 91, 10, 123, 14, 3, 0, 132, 1, 15, 133, 223, 1, 3, 0, 82, 0, 167, 3, 215, 1, 3, 0, 68, 1, 211, 3, 135, 14, 3, 0, 127, 0, 111, 10, 139, 14, 3, 0, 84, 2, 207, 3, 143, 14, 3, 0, 130, 0, 115, 10, 127, 1, 3, 0, 4, 2, 231, 10, 67, 6, 3, 0, 108, 107, 95, 10, 151, 14, 3, 0, 85, 0, 175, 10, 163, 14, 3, 0, 0, 2, 87, 4, 31, 3, 3, 0, 40, 0, 99, 10, 183, 1, 3, 0, 248, 1, 207, 109, 175, 14, 3, 0, 56, 2, 203, 3, 179, 14, 3, 0, 226, 109, 175, 3, 75, 3, 3, 0, 136, 2, 163, 10, 187, 1, 3, 0, 163, 0, 107, 10, 199, 14, 3, 0, 64, 2, 191, 3, 187, 14, 3, 0, 10, 112, 119, 10, 231, 1, 3, 0, 208, 106, 103, 10, 123, 1, 3, 0, 0, 107, 195, 3, 203, 14, 3, 0, 133, 0, 187, 3, 155, 14, 3, 0, 114, 107, 179, 3, 211, 14, 3, 0, 128, 108, 199, 3, 159, 14, 3, 0, 6, 107, 183, 3, 215, 14, 3, 0, 136, 106, 123, 10, 183, 14, 3, 0, 214, 106, 215, 3, 51, 2, 3, 0, 136, 0, 147, 10, 227, 14, 3, 0, 128, 1, 151, 10, 231, 14, 3, 0, 16, 112, 159, 10, 235, 14, 3, 0, 100, 106, 231, 3, 239, 14, 3, 0, 134, 108, 227, 3, 99, 2, 3, 0, 112, 106, 131, 10, 247, 14, 3, 0, 55, 0, 235, 3, 255, 14, 3, 0, 188, 1, 117, 112, 87, 1, 3, 0, 120, 107, 127, 10, 39, 1, 3, 0, 12, 2, 143, 10, 3, 15, 3, 0, 50, 108, 155, 10, 243, 14, 3, 0, 126, 107, 223, 3, 7, 15, 3, 0, 240, 1, 111, 112, 55, 2, 3, 0, 140, 108, 187, 10, 119, 2, 3, 0, 146, 108, 251, 3, 19, 15, 3, 0, 204, 1, 31, 10, 71, 15, 3, 0, 132, 107, 243, 3, 39, 15, 3, 0, 64, 106, 255, 3, 27, 15, 3, 0, 76, 106, 239, 3, 11, 15, 3, 0, 88, 1, 195, 10, 43, 15, 3, 0, 128, 2, 235, 10, 127, 2, 3, 0, 22, 112, 191, 10, 47, 1, 3, 0, 166, 109, 207, 10, 47, 15, 3, 0, 166, 0, 47, 10, 63, 15, 3, 0, 172, 106, 199, 10, 55, 15, 3, 0, 138, 107, 179, 10, 31, 15, 3, 0, 168, 107, 247, 3, 59, 15, 3, 0, 88, 0, 129, 112, 75, 15, 3, 0, 187, 0, 127, 4, 79, 15, 3, 0, 190, 0, 131, 4, 83, 15, 3, 0, 169, 0, 135, 4, 87, 15, 3, 0, 193, 0, 139, 4, 91, 15, 3, 0, 156, 9, 143, 4, 95, 15, 3, 0, 196, 0, 147, 4, 99, 15, 3, 0, 199, 0, 151, 4, 103, 15, 3, 0, 202, 0, 195, 11, 107, 15, 3, 0, 172, 0, 19, 3, 163, 6, 3, 0, 205, 0, 155, 4, 111, 15, 3, 0, 208, 0, 159, 4, 115, 15, 3, 0, 211, 0, 163, 4, 119, 15, 3, 0, 214, 0, 167, 4, 123, 15, 3, 0, 217, 0, 171, 4, 127, 15, 3, 0, 252, 1, 211, 10, 131, 15, 3, 0, 72, 1, 113, 131, 143, 15, 3, 0, 64, 1, 251, 10, 199, 15, 3, 0, 40, 1, 135, 112, 151, 15, 3, 0, 12, 107, 141, 112, 155, 15, 3, 0, 72, 2, 15, 4, 159, 15, 3, 0, 124, 109, 97, 132, 235, 1, 3, 0, 144, 107, 19, 4, 231, 15, 3, 0, 18, 107, 15, 11, 167, 15, 3, 0, 24, 107, 51, 4, 223, 15, 3, 0, 30, 107, 27, 4, 123, 2, 3, 0, 28, 112, 7, 4, 183, 15, 3, 0, 142, 106, 3, 11, 195, 1, 3, 0, 40, 2, 27, 11, 187, 15, 3, 0, 80, 1, 255, 10, 219, 15, 3, 0, 36, 107, 23, 4, 195, 15, 3, 0, 76, 2, 243, 10, 63, 2, 3, 0, 44, 1, 247, 10, 175, 15, 3, 0, 220, 1, 35, 4, 239, 1, 3, 0, 232, 1, 23, 11, 227, 6, 3, 0, 56, 108, 43, 4, 207, 15, 3, 0, 232, 109, 11, 4, 131, 2, 3, 0, 20, 2, 59, 11, 211, 15, 3, 0, 152, 108, 145, 111, 131, 1, 3, 0, 130, 109, 107, 131, 151, 9, 3, 0, 58, 0, 59, 4, 235, 15, 3, 0, 62, 108, 47, 4, 227, 15, 3, 0, 34, 112, 51, 11, 239, 15, 3, 0, 180, 1, 163, 132, 243, 15, 3, 0, 104, 2, 35, 10, 103, 2, 3, 0, 172, 1, 35, 11, 247, 15, 3, 0, 76, 1, 63, 4, 251, 15, 3, 0, 91, 0, 31, 11, 255, 15, 3, 0, 136, 109, 39, 11, 7, 16, 3, 0, 212, 108, 3, 4, 11, 16, 3, 0, 32, 2, 47, 11, 151, 1, 3, 0, 140, 1, 79, 4, 79, 3, 3, 0, 68, 108, 67, 4, 19, 16, 3, 0, 132, 2, 3, 4, 135, 2, 3, 0, 19, 0, 43, 11, 35, 16, 3, 0, 74, 108, 71, 4, 31, 16, 3, 0, 80, 108, 55, 11, 39, 16, 3, 0, 175, 0, 187, 132, 43, 16, 3, 0, 148, 106, 99, 4, 47, 16, 3, 0, 120, 1, 125, 131, 243, 1, 3, 0, 82, 106, 83, 4, 67, 2, 3, 0, 36, 2, 167, 10, 55, 16, 3, 0, 133, 199, 103, 4, 71, 2, 3, 0, 172, 109, 107, 4, 59, 16, 3, 0, 94, 0, 111, 4, 71, 16, 3, 0, 224, 1, 71, 10, 75, 16, 3, 0, 86, 108, 239, 10, 79, 16, 3, 0, 158, 108, 83, 11, 87, 16, 3, 0, 142, 109, 103, 132, 91, 16, 3, 0, 207, 210, 75, 11, 51, 1, 3, 0, 196, 1, 31, 4, 95, 16, 3, 0, 42, 107, 219, 3, 103, 16, 3, 0, 136, 1, 87, 11, 107, 16, 3, 0, 150, 107, 115, 4, 75, 2, 3, 0, 220, 0, 175, 4, 83, 3, 3, 0, 223, 0, 179, 4, 71, 7, 3, 0, 226, 0, 183, 4, 115, 16, 3, 0, 229, 0, 187, 4, 119, 16, 3, 0, 232, 0, 191, 4, 91, 1, 3, 0, 235, 0, 195, 4, 123, 16, 3, 0, 238, 0, 199, 4, 131, 16, 3, 0, 241, 0, 203, 4, 135, 16, 3, 0, 244, 0, 207, 4, 139, 16, 3, 0, 247, 0, 199, 11, 143, 16, 3, 0, 61, 0, 203, 11, 147, 16, 3, 0, 250, 0, 211, 4, 151, 16, 3, 0, 152, 9, 215, 4, 247, 1, 3, 0, 68, 3, 219, 4, 155, 16, 3, 0, 253, 0, 207, 11, 159, 16, 3, 0, 148, 9, 211, 11, 163, 16, 3, 0, 0, 1, 215, 11, 155, 9, 3, 0, 144, 9, 140, 174, 167, 16, 3, 0, 3, 1, 219, 11, 171, 16, 3, 0, 6, 1, 223, 11, 175, 16, 3, 0, 9, 1, 227, 11, 179, 16, 3, 0, 72, 3, 1, 44, 183, 16, 3, 0, 12, 1, 231, 11, 187, 16, 3, 0, 96, 2, 35, 70, 95, 2, 3, 0, 15, 1, 223, 4, 191, 16, 3, 0, 18, 1, 235, 11, 195, 16, 3, 0, 8, 2, 11, 11, 83, 2, 3, 0, 64, 0, 95, 11, 199, 16, 3, 0, 192, 1, 205, 132, 191, 1, 3, 0, 16, 2, 119, 4, 207, 16, 3, 0, 70, 106, 39, 4, 107, 1, 3, 0, 48, 107, 99, 11, 215, 16, 3, 0, 178, 0, 5, 113, 35, 3, 3, 0, 54, 107, 7, 11, 231, 16, 3, 0, 88, 3, 227, 4, 87, 3, 2, 0, 231, 4, 75, 4, 2, 0, 235, 4, 167, 4, 2, 0, 239, 4, 163, 174, 2, 0, 243, 4, 57, 44, 2, 0, 247, 4, 101, 131, 2, 0, 251, 4, 175, 4, 2, 0, 63, 1, 226, 70, 2, 0, 255, 4, 231, 70, 2, 0, 3, 5, 123, 112, 2, 0, 7, 5, 147, 11, 2, 0, 11, 5, 191, 4, 2, 0, 15, 5, 62, 44, 2, 0, 19, 5, 162, 44, 2, 0, 27, 5, 183, 4, 2, 0, 31, 5, 241, 70, 2, 0, 35, 5, 207, 112, 2, 0, 39, 5, 167, 44, 2, 0, 43, 5, 102, 39, 2, 0, 47, 5, 15, 71, 2, 0, 55, 5, 211, 4, 2, 0, 59, 5, 219, 4, 2, 0, 67, 5, 236, 70, 2, 0, 23, 3, 67, 44, 2, 0, 71, 5, 72, 44, 2, 0, 75, 5, 179, 4, 2, 0, 79, 5, 246, 70, 2, 0, 83, 5, 219, 112, 2, 0, 87, 5, 187, 4, 2, 0, 27, 3, 155, 11, 2, 0, 91, 5, 55, 4, 2, 0, 95, 5, 159, 11, 2, 0, 231, 12, 140, 174, 2, 0, 99, 5, 243, 112, 2, 0, 103, 5, 237, 109, 2, 0, 107, 5, 237, 109, 2, 0, 111, 5, 249, 112, 2, 0, 115, 5, 163, 4, 2, 0, 119, 5, 11, 113, 2, 0, 123, 5, 119, 11, 2, 0, 127, 5, 23, 113, 2, 0, 131, 5, 225, 112, 2, 0, 139, 5, 81, 112, 2, 0, 143, 5, 103, 3, 2, 0, 147, 5, 87, 112, 2, 0, 151, 5, 33, 109, 2, 0, 155, 5, 35, 71, 2, 0, 159, 5, 91, 4, 2, 0, 163, 5, 29, 113, 2, 0, 171, 5, 229, 132, 2, 0, 175, 5, 195, 4, 2, 0, 179, 5, 99, 3, 2, 0, 167, 1, 95, 3, 2, 0, 183, 5, 95, 4, 2, 0, 187, 5, 207, 4, 2, 0, 191, 5, 127, 11, 2, 0, 195, 5, 111, 3, 2, 0, 199, 5, 107, 3, 2, 0, 203, 5, 93, 112, 2, 0, 211, 5, 35, 113, 2, 0, 215, 5, 41, 113, 2, 0, 219, 5, 119, 3, 2, 0, 223, 5, 115, 3, 2, 0, 227, 5, 17, 113, 2, 0, 171, 1, 47, 113, 2, 0, 231, 5, 123, 3, 2, 0, 235, 5, 135, 3, 2, 0, 239, 5, 147, 3, 2, 0, 243, 5, 131, 3, 2, 0, 247, 5, 143, 3, 2, 0, 251, 5, 139, 3, 2, 0, 79, 1, 127, 3, 2, 0, 255, 5, 151, 3, 2, 0, 3, 6, 163, 3, 2, 0, 7, 6, 155, 3, 2, 0, 11, 6, 53, 113, 2, 0, 15, 6, 77, 113, 2, 0, 95, 1, 213, 112, 2, 0, 3, 2, 255, 112, 2, 0, 19, 6, 59, 113, 2, 0, 23, 6, 65, 113, 2, 0, 27, 6, 71, 113, 2, 0, 31, 6, 231, 112, 2, 0, 35, 6, 159, 3, 2, 0, 39, 6, 83, 113, 2, 0, 43, 6, 99, 112, 2, 0, 47, 6, 237, 112, 2, 0, 51, 6, 95, 113, 2, 0, 55, 6, 105, 112, 2, 0, 59, 6, 171, 3, 2, 0, 7, 2, 167, 3, 2, 0, 63, 6, 211, 3, 2, 0, 71, 6, 207, 3, 2, 0, 75, 6, 159, 4, 2, 0, 31, 3, 87, 4, 2, 0, 79, 6, 207, 109, 2, 0, 11, 2, 203, 3, 2, 0, 83, 6, 175, 3, 2, 0, 71, 1, 191, 3, 2, 0, 87, 6, 111, 11, 2, 0, 91, 6, 195, 3, 2, 0, 95, 6, 187, 3, 2, 0, 99, 6, 179, 3, 2, 0, 67, 3, 199, 3, 2, 0, 71, 3, 199, 4, 2, 0, 103, 6, 183, 3, 2, 0, 107, 6, 33, 193, 2, 0, 175, 1, 215, 3, 2, 0, 115, 1, 231, 3, 2, 0, 111, 6, 227, 3, 2, 0, 115, 6, 235, 3, 2, 0, 119, 6, 117, 112, 2, 0, 123, 6, 223, 3, 2, 0, 127, 6, 111, 112, 2, 0, 131, 6, 183, 10, 2, 0, 135, 6, 251, 3, 2, 0, 139, 6, 243, 3, 2, 0, 143, 6, 255, 3, 2, 0, 147, 6, 239, 3, 2, 0, 151, 6, 215, 4, 2, 0, 155, 6, 247, 3, 2, 0, 159, 6, 129, 112, 2, 0, 167, 6, 151, 11, 2, 0, 171, 6, 143, 11, 2, 0, 175, 6, 135, 112, 2, 0, 135, 1, 141, 112, 2, 0, 179, 6, 15, 4, 2, 0, 183, 6, 67, 39, 2, 0, 187, 6, 19, 4, 2, 0, 191, 6, 135, 11, 2, 0, 195, 6, 51, 4, 2, 0, 199, 6, 27, 4, 2, 0, 203, 6, 7, 4, 2, 0, 207, 6, 23, 4, 2, 0, 211, 6, 35, 4, 2, 0, 215, 6, 155, 4, 2, 0, 219, 6, 43, 4, 2, 0, 223, 6, 11, 4, 2, 0, 139, 1, 115, 11, 2, 0, 231, 6, 59, 4, 2, 0, 235, 6, 47, 4, 2, 0, 239, 6, 63, 4, 2, 0, 243, 6, 171, 4, 2, 0, 247, 6, 123, 11, 2, 0, 251, 6, 79, 4, 2, 0, 255, 6, 67, 4, 2, 0, 3, 7, 163, 11, 2, 0, 179, 1, 71, 4, 2, 0, 7, 7, 30, 70, 2, 0, 11, 7, 99, 4, 2, 0, 15, 7, 203, 4, 2, 0, 19, 7, 83, 4, 2, 0, 23, 7, 103, 4, 2, 0, 43, 1, 223, 4, 2, 0, 27, 7, 139, 11, 2, 0, 15, 2, 107, 4, 2, 0, 31, 7, 103, 11, 2, 0, 35, 7, 111, 4, 2, 0, 39, 7, 131, 11, 2, 0, 43, 7, 107, 11, 2, 0, 47, 7, 31, 4, 2, 0, 51, 7, 219, 3, 2, 0, 55, 7, 115, 4, 2, 0, 59, 7, 167, 11, 2, 0, 63, 7, 135, 4, 2, 0, 67, 7, 127, 4, 2, 0, 155, 1, 183, 11, 2, 0, 71, 7, 187, 11, 2, 0, 111, 16, 191, 11, 2, 0, 159, 1, 123, 4, 2, 0, 83, 1, 171, 11, 2, 0, 75, 7, 131, 4, 2, 0, 79, 7, 175, 11, 2, 0, 83, 7, 147, 4, 2, 0, 203, 1, 179, 11, 2, 0, 87, 7, 139, 4, 2, 0, 19, 2, 1, 44, 2, 0, 91, 7, 143, 4, 2, 0, 119, 1, 151, 4, 2, 0, 95, 2, 227, 4, 2, 0, 95, 7, 91, 11, 2, 0, 35, 3, 39, 4, 2, 0, 99, 7, 19, 3, 138, 0, 247, 11, 251, 11, 15, 12, 27, 12, 31, 12, 35, 12, 39, 12, 67, 133, 55, 12, 71, 12, 87, 12, 91, 12, 95, 12, 103, 12, 107, 12, 23, 5, 73, 133, 127, 12, 51, 5, 159, 12, 55, 5, 79, 133, 171, 12, 63, 5, 175, 12, 179, 12, 183, 12, 199, 12, 75, 5, 27, 3, 95, 5, 85, 133, 103, 5, 239, 12, 115, 5, 15, 13, 19, 13, 135, 5, 39, 13, 47, 13, 107, 2, 75, 13, 79, 13, 91, 133, 167, 5, 99, 13, 115, 13, 139, 13, 155, 13, 183, 13, 187, 13, 207, 5, 207, 13, 211, 13, 223, 13, 243, 13, 247, 13, 251, 13, 39, 2, 31, 14, 71, 14, 75, 14, 59, 6, 103, 14, 111, 14, 115, 14, 119, 14, 127, 14, 131, 14, 67, 6, 227, 1, 147, 14, 47, 2, 167, 14, 171, 14, 71, 1, 191, 14, 195, 14, 207, 14, 147, 9, 71, 3, 219, 14, 223, 14, 99, 2, 251, 14, 15, 15, 23, 15, 35, 15, 67, 15, 135, 15, 139, 15, 147, 15, 163, 15, 171, 15, 103, 1, 191, 15, 203, 15, 223, 6, 227, 6, 215, 15, 3, 16, 15, 16, 23, 16, 27, 16, 51, 16, 139, 2, 43, 1, 63, 16, 27, 7, 67, 16, 31, 7, 83, 16, 39, 7, 99, 16, 63, 7, 67, 7, 103, 133, 75, 7, 143, 2, 79, 2, 127, 16, 83, 7, 87, 7, 251, 1, 19, 2, 91, 7, 119, 1, 147, 2, 203, 16, 109, 133, 211, 16, 87, 2, 255, 1, 219, 16, 223, 16, 227, 16, 235, 16, 239, 16, 155, 0, 231, 4, 235, 4, 239, 4, 243, 4, 247, 4, 251, 4, 63, 1, 255, 4, 3, 5, 7, 5, 11, 5, 15, 5, 19, 5, 27, 5, 31, 5, 35, 5, 39, 5, 43, 5, 47, 5, 59, 5, 67, 5, 23, 3, 71, 5, 203, 12, 79, 5, 83, 5, 87, 5, 91, 5, 99, 5, 107, 5, 111, 5, 119, 5, 123, 5, 127, 5, 131, 5, 139, 5, 143, 5, 147, 5, 151, 5, 155, 5, 159, 5, 163, 5, 171, 5, 175, 5, 179, 5, 167, 1, 183, 5, 187, 5, 191, 5, 195, 5, 199, 5, 203, 5, 211, 5, 215, 5, 219, 5, 223, 5, 227, 5, 171, 1, 231, 5, 235, 5, 239, 5, 243, 5, 247, 5, 251, 5, 79, 1, 255, 5, 3, 6, 7, 6, 11, 6, 15, 6, 95, 1, 3, 2, 19, 6, 23, 6, 27, 6, 31, 6, 35, 6, 39, 6, 43, 6, 47, 6, 51, 6, 55, 6, 7, 2, 63, 6, 71, 6, 75, 6, 31, 3, 79, 6, 11, 2, 83, 6, 87, 6, 91, 6, 95, 6, 99, 6, 67, 3, 103, 6, 107, 6, 175, 1, 115, 1, 111, 6, 115, 6, 119, 6, 123, 6, 127, 6, 131, 6, 135, 6, 139, 6, 143, 6, 147, 6, 151, 6, 155, 6, 159, 6, 167, 6, 171, 6, 175, 6, 135, 1, 179, 6, 183, 6, 187, 6, 191, 6, 195, 6, 199, 6, 203, 6, 207, 6, 211, 6, 215, 6, 219, 6, 139, 1, 231, 6, 235, 6, 239, 6, 243, 6, 247, 6, 251, 6, 255, 6, 3, 7, 179, 1, 7, 7, 11, 7, 15, 7, 19, 7, 23, 7, 15, 2, 35, 7, 43, 7, 47, 7, 51, 7, 55, 7, 59, 7, 83, 1, 79, 7, 203, 1, 95, 7, 35, 3, 99, 7, 237, 0, 247, 16, 39, 17, 67, 17, 71, 17, 99, 17, 115, 17, 80, 40, 203, 17, 243, 17, 39, 18, 43, 18, 55, 18, 53, 110, 71, 18, 75, 18, 103, 18, 220, 109, 107, 18, 119, 18, 123, 7, 131, 18, 139, 18, 179, 18, 187, 18, 199, 18, 203, 18, 211, 18, 219, 18, 231, 18, 239, 18, 255, 18, 99, 19, 107, 19, 107, 110, 147, 19, 113, 110, 151, 19, 155, 19, 183, 19, 203, 19, 215, 19, 251, 19, 255, 19, 8, 42, 41, 193, 19, 20, 27, 20, 31, 20, 39, 41, 39, 20, 44, 41, 87, 20, 95, 20, 131, 110, 115, 20, 131, 20, 167, 20, 38, 42, 75, 21, 95, 21, 189, 197, 130, 70, 183, 21, 51, 22, 127, 22, 135, 22, 139, 22, 238, 42, 171, 22, 171, 143, 179, 22, 199, 22, 3, 23, 95, 23, 103, 23, 71, 110, 139, 23, 187, 23, 199, 23, 231, 23, 43, 24, 119, 24, 127, 24, 227, 24, 243, 24, 247, 24, 251, 24, 204, 174, 252, 9, 143, 25, 120, 70, 191, 25, 152, 44, 22, 44, 95, 26, 99, 26, 29, 147, 107, 26, 143, 26, 147, 26, 151, 26, 155, 26, 59, 3, 163, 9, 167, 26, 175, 26, 97, 43, 179, 26, 183, 26, 195, 26, 199, 26, 207, 26, 215, 26, 219, 26, 223, 26, 172, 43, 231, 26, 251, 26, 15, 27, 55, 27, 80, 70, 179, 8, 84, 41, 235, 27, 3, 28, 7, 28, 11, 28, 23, 28, 59, 28, 67, 28, 71, 28, 216, 70, 79, 28, 95, 28, 103, 28, 111, 28, 115, 28, 249, 152, 123, 28, 127, 28, 131, 28, 135, 28, 251, 110, 135, 70, 159, 28, 219, 41, 89, 41, 243, 28, 247, 28, 94, 41, 152, 39, 255, 28, 184, 41, 71, 29, 219, 29, 251, 29, 39, 30, 75, 30, 147, 30, 159, 30, 167, 30, 191, 30, 235, 30, 255, 30, 3, 31, 7, 31, 11, 31, 19, 31, 27, 31, 39, 31, 59, 31, 63, 31, 67, 31, 95, 31, 103, 31, 227, 31, 144, 207, 135, 32, 196, 39, 143, 32, 132, 43, 83, 110, 182, 43, 83, 33, 87, 33, 147, 39, 103, 33, 197, 43, 119, 33, 24, 41, 201, 164, 155, 33, 167, 33, 171, 33, 223, 33, 243, 33, 247, 33, 251, 33, 173, 165, 19, 34, 232, 9, 27, 34, 39, 34, 25, 166, 47, 34, 59, 34, 63, 34, 236, 9, 171, 34, 183, 34, 199, 34, 251, 34, 207, 35, 191, 2, 227, 35, 99, 36, 111, 36, 191, 36, 207, 36, 227, 36, 39, 37, 71, 37, 75, 37, 155, 37, 175, 37, 187, 37, 199, 37, 239, 37, 207, 43, 247, 2, 35, 38, 219, 9, 75, 38, 99, 38, 17, 110, 145, 40, 31, 39, 15, 0, 175, 31, 199, 157, 205, 157, 211, 157, 217, 157, 223, 157, 229, 157, 235, 157, 241, 157, 247, 157, 253, 157, 3, 158, 9, 158, 15, 158, 21, 158, 26, 10, 180, 2, 115, 133, 121, 133, 127, 133, 133, 133, 255, 16, 218, 108, 139, 133, 145, 133, 151, 133, 157, 133, 163, 133, 169, 133, 175, 133, 181, 133, 187, 133, 35, 17, 193, 133, 199, 133, 205, 133, 70, 112, 211, 133, 217, 133, 223, 133, 47, 17, 51, 17, 229, 133, 31, 40, 235, 133, 253, 133, 3, 134, 59, 17, 9, 134, 15, 134, 63, 17, 21, 134, 27, 134, 33, 134, 39, 134, 45, 134, 51, 134, 57, 134, 75, 17, 63, 134, 79, 17, 83, 17, 87, 17, 91, 17, 103, 7, 95, 17, 103, 17, 107, 17, 252, 107, 69, 134, 75, 134, 221, 110, 81, 134, 87, 134, 46, 112, 93, 134, 99, 134, 105, 134, 212, 111, 111, 134, 117, 134, 129, 134, 119, 17, 123, 17, 135, 134, 135, 17, 141, 134, 139, 17, 147, 134, 143, 17, 147, 17, 151, 17, 155, 17, 52, 106, 159, 134, 165, 134, 177, 134, 64, 112, 189, 134, 195, 134, 201, 134, 213, 134, 219, 134, 225, 134, 231, 134, 237, 134, 28, 133, 251, 17, 255, 17, 108, 42, 243, 134, 249, 134, 255, 134, 5, 135, 7, 18, 11, 135, 17, 135, 11, 18, 113, 42, 15, 18, 23, 135, 19, 18, 200, 2, 29, 135, 35, 135, 41, 135, 47, 135, 53, 135, 59, 135, 204, 107, 65, 135, 51, 18, 71, 135, 59, 18, 63, 18, 12, 44, 95, 135, 23, 110, 101, 135, 113, 135, 119, 135, 125, 135, 131, 135, 137, 135, 143, 135, 155, 135, 0, 3, 161, 135, 197, 107, 167, 135, 173, 135, 209, 135, 204, 2, 215, 135, 221, 135, 227, 135, 233, 135, 239, 135, 176, 9, 245, 135, 251, 135, 7, 136, 13, 136, 19, 136, 25, 136, 127, 18, 31, 136, 37, 136, 43, 136, 49, 136, 5, 71, 55, 136, 143, 18, 74, 42, 61, 136, 67, 136, 10, 71, 73, 136, 85, 136, 109, 136, 115, 136, 171, 18, 121, 136, 127, 136, 133, 136, 16, 40, 139, 136, 145, 136, 167, 39, 151, 136, 157, 136, 175, 18, 163, 136, 169, 136, 175, 136, 181, 136, 187, 136, 193, 136, 199, 136, 205, 136, 211, 136, 217, 136, 223, 136, 207, 18, 229, 136, 215, 18, 235, 136, 241, 136, 223, 18, 247, 136, 244, 9, 253, 136, 3, 137, 9, 137, 235, 18, 15, 137, 21, 137, 27, 137, 243, 18, 247, 18, 33, 137, 39, 137, 251, 18, 45, 137, 51, 137, 3, 19, 57, 137, 63, 137, 69, 137, 75, 137, 192, 9, 81, 137, 11, 19, 87, 137, 15, 19, 19, 19, 23, 19, 27, 19, 31, 19, 35, 19, 39, 19, 148, 42, 55, 19, 59, 19, 63, 19, 248, 9, 99, 137, 105, 137, 111, 137, 239, 110, 117, 137, 123, 137, 67, 19, 71, 19, 129, 137, 75, 19, 79, 19, 83, 19, 87, 19, 91, 19, 153, 42, 135, 137, 95, 19, 141, 137, 103, 19, 151, 2, 139, 7, 39, 3, 111, 19, 115, 19, 147, 137, 153, 137, 159, 137, 158, 42, 119, 19, 123, 19, 165, 137, 127, 19, 171, 137, 131, 19, 135, 19, 163, 42, 168, 42, 139, 19, 173, 42, 143, 19, 177, 137, 178, 42, 183, 137, 189, 137, 195, 137, 201, 137, 71, 131, 207, 137, 213, 137, 51, 40, 219, 137, 225, 137, 237, 137, 159, 19, 243, 137, 249, 137, 163, 19, 255, 137, 167, 19, 171, 19, 5, 138, 179, 19, 196, 9, 11, 138, 17, 138, 29, 138, 97, 44, 35, 138, 41, 138, 47, 138, 195, 19, 53, 138, 59, 138, 65, 138, 71, 138, 199, 19, 87, 44, 77, 138, 83, 138, 89, 138, 127, 44, 207, 19, 211, 19, 95, 138, 219, 19, 223, 19, 227, 19, 21, 40, 101, 138, 231, 19, 171, 7, 235, 19, 181, 139, 243, 19, 164, 1, 247, 19, 187, 139, 193, 139, 55, 111, 205, 139, 211, 139, 152, 174, 217, 139, 223, 139, 229, 139, 235, 139, 119, 110, 3, 20, 180, 70, 7, 20, 241, 139, 11, 20, 247, 139, 253, 139, 3, 140, 82, 43, 127, 39, 9, 140, 15, 140, 21, 140, 125, 110, 15, 20, 27, 140, 33, 140, 39, 140, 143, 131, 45, 140, 51, 140, 87, 43, 23, 20, 183, 42, 57, 140, 63, 140, 69, 140, 75, 140, 81, 140, 87, 140, 93, 140, 35, 20, 99, 140, 105, 140, 111, 140, 117, 140, 43, 20, 47, 20, 13, 42, 162, 43, 51, 20, 123, 140, 55, 20, 59, 20, 63, 20, 129, 140, 135, 140, 67, 20, 141, 140, 147, 140, 153, 140, 71, 20, 159, 140, 75, 20, 165, 140, 171, 140, 92, 43, 177, 140, 79, 20, 183, 140, 189, 140, 195, 140, 201, 140, 207, 140, 213, 140, 18, 42, 219, 140, 83, 20, 188, 42, 91, 20, 225, 140, 231, 140, 149, 131, 145, 70, 99, 20, 237, 140, 150, 70, 166, 174, 103, 20, 243, 140, 107, 20, 249, 140, 111, 20, 255, 140, 5, 141, 175, 7, 11, 141, 17, 141, 23, 141, 29, 141, 172, 9, 123, 20, 127, 20, 35, 141, 135, 20, 139, 20, 41, 141, 47, 141, 143, 20, 147, 20, 151, 20, 53, 141, 155, 20, 159, 20, 168, 9, 163, 20, 171, 20, 175, 20, 179, 20, 183, 20, 187, 20, 191, 20, 195, 20, 19, 111, 49, 111, 199, 20, 191, 117, 179, 7, 224, 2, 15, 21, 19, 21, 23, 21, 27, 21, 31, 21, 35, 21, 39, 21, 83, 141, 227, 43, 43, 21, 90, 210, 47, 21, 232, 43, 154, 41, 51, 21, 33, 42, 55, 21, 59, 21, 89, 141, 95, 141, 63, 21, 43, 42, 212, 9, 67, 21, 101, 141, 48, 42, 53, 42, 71, 21, 41, 192, 79, 21, 107, 141, 83, 21, 87, 21, 91, 21, 113, 141, 99, 21, 103, 21, 107, 21, 119, 141, 32, 1, 149, 141, 155, 131, 115, 21, 155, 141, 161, 141, 167, 141, 184, 2, 119, 21, 123, 21, 237, 43, 173, 141, 230, 40, 70, 70, 179, 141, 137, 110, 107, 39, 9, 192, 197, 141, 203, 141, 127, 21, 173, 107, 242, 43, 209, 141, 186, 107, 215, 141, 114, 41, 221, 141, 227, 141, 131, 21, 135, 21, 188, 2, 235, 40, 139, 21, 143, 21, 147, 21, 151, 21, 155, 21, 159, 21, 203, 7, 163, 21, 167, 21, 147, 44, 233, 141, 171, 21, 239, 141, 175, 21, 179, 21, 187, 21, 245, 141, 191, 21, 195, 21, 161, 131, 13, 0, 199, 21, 207, 7, 203, 21, 168, 2, 207, 21, 211, 21, 43, 3, 215, 21, 219, 21, 251, 141, 223, 21, 42, 44, 227, 21, 231, 21, 235, 21, 1, 142, 239, 21, 143, 110, 243, 21, 247, 21, 49, 41, 251, 21, 255, 21, 131, 39, 3, 22, 224, 108, 7, 22, 11, 22, 7, 142, 13, 142, 15, 22, 19, 22, 23, 22, 82, 39, 25, 142, 31, 22, 35, 22, 49, 142, 54, 41, 43, 22, 47, 22, 231, 7, 73, 142, 79, 142, 85, 142, 91, 142, 97, 142, 75, 22, 79, 22, 83, 22, 87, 22, 156, 2, 1, 40, 109, 142, 115, 142, 151, 142, 241, 142, 57, 143, 111, 143, 117, 143, 143, 22, 147, 22, 243, 42, 151, 22, 208, 2, 123, 143, 129, 143, 155, 22, 135, 143, 141, 143, 159, 22, 95, 110, 147, 143, 163, 22, 167, 22, 153, 143, 159, 143, 175, 22, 165, 143, 177, 143, 183, 22, 5, 110, 187, 22, 191, 22, 195, 22, 183, 143, 189, 143, 195, 143, 201, 143, 207, 143, 213, 143, 219, 143, 225, 143, 248, 42, 203, 22, 231, 143, 207, 22, 211, 22, 215, 22, 173, 174, 237, 143, 223, 22, 164, 108, 243, 143, 249, 143, 227, 22, 231, 22, 235, 22, 239, 22, 243, 22, 255, 143, 5, 144, 11, 144, 153, 122, 255, 22, 17, 144, 154, 106, 23, 144, 29, 144, 35, 144, 41, 144, 6, 44, 7, 23, 47, 144, 53, 144, 15, 23, 19, 23, 23, 23, 27, 23, 47, 3, 31, 23, 35, 23, 39, 23, 59, 144, 71, 144, 31, 8, 31, 111, 77, 144, 83, 23, 87, 23, 83, 144, 91, 23, 51, 3, 2, 43, 99, 23, 228, 2, 89, 144, 95, 144, 101, 144, 107, 144, 111, 23, 40, 112, 115, 23, 119, 23, 113, 144, 123, 23, 119, 144, 125, 144, 127, 23, 131, 144, 137, 144, 143, 144, 131, 23, 149, 144, 155, 144, 135, 23, 143, 23, 161, 144, 147, 23, 167, 144, 151, 23, 155, 23, 91, 111, 159, 23, 163, 23, 182, 108, 173, 144, 179, 144, 185, 144, 191, 144, 175, 23, 179, 23, 183, 23, 191, 23, 195, 23, 102, 44, 203, 23, 197, 144, 207, 23, 211, 23, 215, 23, 219, 23, 223, 23, 227, 23, 112, 1, 203, 144, 37, 111, 209, 144, 235, 23, 239, 23, 243, 23, 215, 144, 43, 8, 247, 23, 251, 23, 255, 23, 3, 24, 40, 70, 221, 144, 227, 144, 233, 144, 239, 144, 245, 144, 7, 24, 251, 144, 1, 145, 7, 145, 155, 110, 11, 24, 32, 44, 13, 145, 19, 145, 25, 145, 31, 145, 37, 145, 43, 145, 15, 24, 19, 24, 162, 39, 23, 24, 27, 24, 31, 24, 35, 24, 39, 24, 49, 145, 55, 145, 61, 145, 47, 24, 51, 24, 105, 70, 55, 24, 47, 8, 59, 24, 8, 108, 59, 41, 63, 24, 67, 24, 71, 24, 75, 24, 184, 9, 67, 145, 73, 145, 79, 24, 83, 24, 87, 24, 91, 24, 95, 24, 99, 24, 103, 24, 107, 24, 176, 108, 111, 24, 79, 145, 21, 1, 232, 2, 115, 24, 48, 211, 123, 24, 85, 145, 91, 145, 97, 145, 103, 145, 245, 40, 131, 24, 135, 24, 7, 43, 139, 24, 143, 24, 147, 24, 137, 39, 109, 145, 155, 24, 115, 145, 199, 197, 121, 145, 127, 145, 159, 24, 163, 24, 67, 111, 155, 2, 2, 108, 167, 24, 171, 24, 175, 24, 133, 145, 179, 24, 183, 24, 187, 24, 191, 24, 195, 24, 199, 24, 200, 9, 203, 24, 207, 24, 211, 24, 215, 24, 219, 24, 139, 145, 223, 24, 22, 133, 250, 40, 145, 145, 151, 145, 157, 145, 163, 145, 235, 24, 169, 145, 64, 41, 73, 111, 239, 24, 55, 3, 175, 145, 79, 111, 181, 145, 255, 40, 29, 110, 255, 24, 187, 145, 193, 145, 4, 41, 199, 145, 245, 110, 205, 145, 3, 25, 211, 145, 217, 145, 7, 25, 160, 9, 11, 25, 15, 25, 161, 110, 19, 25, 23, 25, 27, 25, 31, 25, 45, 70, 35, 25, 223, 145, 157, 111, 229, 145, 39, 25, 43, 25, 235, 145, 172, 2, 85, 111, 247, 145, 253, 145, 79, 25, 3, 146, 83, 25, 9, 146, 192, 2, 15, 146, 87, 25, 21, 146, 91, 25, 27, 146, 33, 146, 214, 109, 95, 25, 69, 41, 99, 25, 103, 25, 107, 25, 111, 25, 115, 25, 119, 25, 123, 25, 127, 25, 131, 25, 135, 25, 139, 25, 147, 25, 74, 41, 117, 44, 39, 146, 151, 25, 45, 146, 9, 41, 51, 146, 63, 146, 159, 25, 69, 146, 75, 146, 163, 25, 81, 146, 167, 25, 87, 146, 171, 25, 175, 25, 179, 25, 183, 25, 187, 25, 93, 146, 99, 146, 121, 132, 195, 25, 105, 146, 167, 43, 199, 25, 203, 25, 207, 25, 211, 25, 215, 25, 111, 146, 117, 146, 123, 146, 159, 9, 219, 25, 129, 146, 223, 25, 227, 25, 231, 25, 235, 25, 239, 25, 135, 146, 243, 25, 247, 25, 251, 25, 255, 25, 141, 146, 3, 26, 7, 26, 11, 26, 147, 146, 153, 146, 15, 26, 19, 26, 210, 40, 159, 146, 23, 26, 27, 26, 119, 41, 31, 26, 35, 26, 165, 146, 11, 110, 39, 26, 43, 26, 47, 26, 51, 26, 171, 146, 55, 26, 177, 146, 59, 26, 183, 146, 189, 146, 195, 146, 201, 146, 63, 26, 67, 26, 152, 2, 71, 26, 75, 26, 79, 26, 83, 26, 87, 26, 91, 26, 100, 41, 207, 146, 163, 111, 213, 146, 219, 146, 225, 146, 231, 146, 237, 146, 169, 111, 243, 146, 249, 146, 255, 146, 5, 147, 11, 147, 236, 2, 17, 147, 35, 147, 41, 147, 59, 147, 167, 110, 65, 147, 216, 9, 71, 147, 77, 147, 83, 147, 89, 147, 24, 1, 95, 147, 101, 147, 4, 3, 107, 147, 151, 40, 119, 147, 41, 40, 125, 147, 131, 147, 36, 40, 137, 147, 143, 147, 149, 147, 155, 147, 52, 112, 159, 26, 161, 147, 167, 147, 173, 147, 179, 147, 185, 147, 191, 147, 197, 147, 203, 147, 209, 147, 157, 44, 215, 147, 221, 147, 227, 147, 233, 147, 20, 71, 239, 147, 251, 147, 1, 148, 192, 39, 7, 148, 13, 148, 107, 44, 19, 148, 25, 148, 156, 40, 31, 148, 37, 148, 43, 148, 211, 26, 49, 148, 240, 2, 55, 148, 61, 148, 102, 43, 67, 148, 73, 148, 6, 40, 79, 148, 85, 148, 227, 26, 91, 148, 48, 43, 97, 148, 103, 148, 109, 148, 115, 148, 121, 148, 127, 148, 239, 26, 133, 148, 139, 148, 243, 26, 145, 148, 247, 26, 151, 148, 157, 148, 163, 148, 255, 26, 3, 27, 244, 2, 169, 148, 175, 148, 181, 148, 187, 148, 193, 148, 7, 27, 58, 42, 199, 148, 205, 148, 11, 27, 211, 148, 211, 70, 217, 148, 223, 148, 229, 148, 241, 148, 23, 27, 27, 27, 31, 27, 35, 27, 39, 27, 247, 148, 253, 148, 3, 149, 9, 149, 15, 149, 51, 27, 21, 149, 27, 149, 59, 27, 195, 70, 63, 27, 33, 149, 39, 149, 45, 149, 170, 108, 51, 149, 179, 110, 57, 149, 63, 149, 67, 27, 71, 27, 107, 43, 75, 27, 252, 39, 69, 149, 75, 149, 79, 27, 83, 27, 81, 149, 227, 110, 87, 27, 91, 27, 101, 110, 95, 27, 171, 8, 87, 149, 99, 27, 103, 27, 107, 27, 111, 27, 175, 8, 99, 149, 105, 149, 117, 149, 123, 149, 135, 27, 139, 27, 206, 70, 143, 27, 129, 149, 135, 149, 183, 8, 155, 27, 142, 44, 83, 107, 141, 149, 147, 149, 153, 149, 178, 109, 171, 27, 177, 43, 175, 27, 179, 27, 183, 27, 187, 27, 7, 151, 13, 151, 195, 27, 19, 151, 199, 27, 203, 27, 207, 27, 211, 27, 120, 40, 215, 27, 219, 27, 223, 27, 25, 151, 43, 151, 61, 151, 91, 151, 97, 151, 115, 151, 231, 27, 121, 151, 127, 151, 133, 151, 139, 151, 145, 151, 151, 151, 157, 151, 163, 151, 169, 151, 97, 39, 175, 151, 187, 151, 25, 71, 205, 151, 211, 151, 217, 151, 223, 151, 182, 39, 229, 151, 235, 151, 241, 151, 247, 151, 253, 151, 188, 108, 3, 152, 9, 152, 164, 2, 21, 152, 27, 152, 33, 152, 39, 152, 63, 28, 45, 152, 51, 152, 232, 39, 87, 152, 93, 152, 99, 152, 75, 28, 105, 152, 111, 152, 83, 28, 117, 152, 123, 152, 129, 152, 135, 152, 141, 152, 26, 192, 147, 152, 153, 152, 77, 39, 165, 152, 171, 152, 248, 2, 177, 152, 183, 152, 189, 152, 195, 152, 201, 152, 207, 152, 213, 152, 219, 152, 225, 152, 231, 152, 237, 152, 243, 152, 119, 28, 255, 152, 197, 39, 11, 153, 17, 153, 23, 153, 29, 153, 35, 153, 47, 153, 53, 153, 141, 40, 71, 153, 77, 153, 83, 153, 142, 39, 89, 153, 95, 153, 101, 153, 163, 28, 107, 153, 113, 153, 167, 28, 171, 28, 204, 9, 119, 153, 125, 153, 131, 153, 227, 39, 143, 153, 155, 153, 161, 153, 164, 41, 207, 28, 211, 28, 169, 41, 215, 28, 166, 40, 167, 153, 173, 153, 110, 70, 179, 153, 185, 153, 191, 153, 197, 153, 203, 153, 227, 28, 209, 153, 221, 153, 227, 153, 233, 153, 239, 153, 245, 153, 231, 28, 177, 39, 174, 41, 235, 28, 253, 108, 112, 43, 13, 154, 179, 41, 239, 28, 19, 154, 25, 154, 72, 107, 31, 154, 37, 154, 43, 154, 49, 154, 251, 28, 55, 154, 61, 154, 67, 154, 73, 154, 85, 154, 91, 154, 23, 29, 97, 154, 27, 29, 31, 29, 103, 154, 35, 29, 109, 154, 39, 29, 43, 29, 217, 39, 115, 154, 121, 154, 47, 29, 127, 154, 51, 29, 55, 29, 133, 154, 59, 29, 63, 29, 77, 110, 139, 154, 145, 154, 151, 154, 63, 42, 157, 154, 67, 29, 163, 154, 68, 42, 14, 41, 169, 154, 175, 154, 17, 44, 75, 29, 79, 29, 83, 29, 187, 154, 87, 29, 91, 29, 95, 29, 191, 110, 99, 29, 103, 29, 193, 154, 107, 29, 199, 154, 111, 29, 115, 29, 205, 154, 211, 154, 119, 29, 123, 29, 196, 2, 217, 154, 223, 154, 131, 29, 135, 29, 139, 29, 143, 29, 255, 8, 159, 2, 229, 154, 151, 29, 155, 29, 52, 43, 159, 29, 163, 29, 235, 154, 167, 29, 117, 43, 241, 154, 8, 3, 171, 29, 175, 29, 179, 29, 183, 29, 187, 29, 191, 29, 247, 154, 195, 29, 199, 29, 203, 29, 253, 154, 122, 43, 207, 29, 75, 70, 35, 110, 3, 155, 211, 29, 215, 29, 164, 9, 9, 155, 223, 29, 115, 132, 15, 155, 21, 155, 27, 155, 33, 155, 39, 155, 45, 155, 231, 29, 235, 29, 51, 155, 239, 29, 180, 197, 243, 29, 247, 29, 57, 155, 200, 1, 63, 155, 255, 29, 3, 30, 69, 155, 75, 155, 81, 155, 87, 155, 176, 199, 128, 202, 93, 155, 99, 155, 105, 155, 111, 155, 7, 30, 11, 30, 147, 206, 117, 155, 15, 30, 19, 30, 23, 30, 200, 40, 205, 40, 27, 30, 180, 107, 215, 40, 3, 9, 181, 111, 123, 155, 129, 155, 135, 155, 43, 30, 220, 9, 127, 43, 47, 30, 51, 30, 141, 155, 147, 155, 55, 30, 153, 155, 159, 155, 59, 30, 165, 155, 171, 155, 177, 155, 183, 155, 63, 30, 67, 30, 71, 30, 79, 30, 83, 30, 87, 30, 91, 30, 95, 30, 99, 30, 103, 30, 107, 30, 105, 41, 195, 155, 201, 155, 207, 155, 213, 155, 219, 155, 225, 155, 231, 155, 151, 30, 155, 30, 237, 155, 163, 30, 171, 30, 175, 30, 179, 30, 183, 30, 243, 155, 249, 155, 255, 155, 5, 156, 23, 156, 195, 30, 29, 156, 199, 30, 203, 30, 207, 30, 125, 40, 211, 30, 215, 30, 219, 30, 35, 156, 223, 30, 227, 30, 231, 30, 239, 30, 243, 30, 247, 30, 89, 156, 95, 156, 113, 156, 119, 156, 180, 9, 125, 156, 131, 156, 137, 156, 251, 30, 143, 156, 149, 156, 155, 156, 161, 156, 221, 70, 167, 156, 173, 156, 179, 156, 185, 156, 191, 156, 197, 156, 203, 156, 209, 156, 215, 156, 221, 156, 17, 43, 227, 156, 233, 156, 15, 31, 239, 156, 23, 31, 245, 156, 251, 156, 1, 157, 31, 31, 35, 31, 7, 157, 13, 157, 43, 31, 47, 31, 51, 31, 55, 31, 22, 43, 19, 157, 25, 157, 31, 157, 37, 157, 71, 31, 172, 39, 27, 43, 43, 157, 49, 157, 61, 157, 79, 31, 83, 31, 92, 44, 87, 31, 67, 157, 73, 157, 79, 157, 91, 31, 85, 157, 91, 157, 99, 31, 97, 157, 109, 157, 115, 157, 133, 157, 107, 31, 111, 31, 139, 157, 115, 31, 119, 31, 123, 31, 159, 31, 151, 157, 163, 31, 167, 31, 157, 157, 171, 31, 72, 39, 27, 158, 33, 158, 39, 158, 45, 158, 51, 158, 19, 9, 23, 9, 27, 9, 57, 158, 63, 158, 31, 9, 35, 9, 69, 158, 110, 40, 39, 9, 11, 32, 43, 9, 15, 32, 75, 158, 47, 9, 51, 9, 23, 2, 55, 9, 81, 158, 59, 9, 47, 32, 51, 32, 87, 158, 93, 158, 59, 32, 99, 158, 63, 32, 67, 32, 197, 110, 105, 158, 169, 200, 71, 32, 111, 158, 117, 158, 123, 158, 75, 32, 79, 32, 83, 32, 118, 42, 87, 32, 129, 158, 91, 32, 95, 32, 99, 32, 103, 32, 107, 32, 141, 158, 111, 32, 147, 158, 115, 32, 119, 32, 123, 32, 153, 158, 159, 158, 127, 32, 55, 70, 131, 32, 46, 40, 165, 158, 171, 158, 177, 158, 139, 32, 213, 43, 147, 32, 151, 32, 155, 32, 159, 32, 163, 32, 167, 32, 171, 32, 55, 71, 183, 158, 189, 158, 195, 158, 179, 32, 183, 32, 187, 32, 191, 32, 195, 32, 123, 42, 100, 70, 199, 32, 203, 32, 207, 158, 207, 32, 211, 32, 215, 32, 219, 32, 223, 32, 87, 39, 213, 158, 219, 158, 227, 32, 231, 158, 237, 158, 243, 158, 252, 2, 137, 43, 231, 32, 235, 32, 239, 32, 243, 32, 247, 32, 251, 32, 255, 32, 3, 33, 7, 33, 11, 33, 207, 39, 249, 158, 255, 158, 15, 33, 5, 159, 11, 159, 222, 39, 125, 70, 17, 159, 23, 159, 29, 159, 35, 159, 26, 40, 41, 159, 47, 159, 53, 159, 59, 159, 23, 33, 167, 9, 27, 44, 65, 159, 95, 159, 63, 33, 67, 33, 71, 33, 75, 33, 79, 33, 224, 9, 101, 159, 107, 159, 113, 159, 91, 33, 119, 159, 125, 159, 131, 159, 137, 159, 66, 40, 217, 163, 223, 163, 229, 163, 235, 163, 241, 163, 152, 43, 247, 163, 253, 163, 95, 33, 188, 9, 3, 164, 9, 164, 228, 9, 192, 43, 15, 164, 21, 164, 27, 164, 33, 164, 146, 40, 39, 164, 45, 164, 51, 164, 255, 41, 57, 164, 63, 164, 69, 164, 75, 164, 56, 40, 81, 164, 87, 164, 99, 164, 105, 164, 111, 164, 107, 33, 111, 33, 117, 164, 115, 33, 226, 106, 123, 33, 127, 33, 123, 164, 129, 164, 135, 164, 187, 39, 141, 164, 147, 164, 153, 164, 159, 164, 137, 44, 165, 164, 171, 164, 110, 41, 177, 164, 183, 164, 203, 110, 212, 39, 189, 164, 139, 33, 143, 33, 195, 164, 207, 164, 213, 164, 161, 40, 219, 164, 187, 111, 147, 33, 151, 33, 222, 43, 159, 33, 11, 40, 163, 33, 225, 164, 231, 164, 237, 164, 175, 33, 179, 33, 243, 164, 249, 164, 183, 33, 187, 33, 255, 164, 191, 33, 5, 165, 11, 165, 17, 165, 23, 165, 199, 33, 203, 33, 207, 33, 211, 33, 215, 33, 29, 165, 219, 33, 202, 39, 35, 165, 41, 165, 47, 165, 53, 165, 59, 165, 65, 165, 71, 165, 77, 165, 83, 165, 89, 165, 95, 165, 171, 9, 101, 165, 12, 3, 107, 165, 235, 33, 113, 165, 78, 42, 119, 165, 83, 42, 239, 33, 16, 3, 125, 165, 131, 165, 137, 165, 122, 44, 143, 165, 149, 165, 155, 165, 161, 165, 167, 165, 194, 108, 255, 33, 189, 174, 179, 165, 185, 165, 189, 41, 3, 34, 191, 165, 197, 165, 7, 34, 194, 41, 11, 34, 71, 40, 203, 165, 209, 165, 215, 165, 221, 165, 23, 34, 227, 165, 233, 165, 239, 165, 251, 165, 7, 166, 13, 166, 94, 42, 19, 166, 31, 166, 37, 166, 76, 40, 43, 166, 49, 166, 55, 166, 61, 166, 67, 166, 88, 42, 170, 70, 73, 166, 79, 166, 85, 166, 51, 34, 91, 166, 97, 166, 55, 34, 176, 2, 103, 166, 52, 44, 109, 166, 115, 166, 121, 166, 61, 40, 127, 166, 145, 166, 157, 166, 163, 166, 169, 166, 175, 166, 175, 70, 181, 166, 187, 166, 193, 166, 199, 166, 205, 166, 217, 166, 241, 166, 247, 166, 253, 166, 71, 34, 3, 167, 9, 167, 15, 167, 21, 167, 240, 9, 79, 34, 83, 34, 27, 167, 87, 34, 91, 34, 95, 34, 33, 167, 99, 34, 103, 34, 199, 41, 107, 34, 111, 34, 39, 167, 115, 34, 45, 167, 119, 34, 57, 43, 175, 9, 193, 42, 198, 42, 62, 43, 123, 34, 67, 43, 224, 41, 127, 34, 57, 167, 63, 167, 229, 41, 128, 42, 167, 2, 212, 2, 205, 169, 63, 3, 211, 169, 234, 41, 131, 34, 135, 34, 171, 2, 139, 34, 143, 34, 58, 112, 196, 174, 217, 169, 180, 174, 223, 169, 147, 34, 65, 131, 151, 34, 155, 34, 229, 169, 235, 169, 159, 34, 241, 169, 133, 42, 163, 34, 247, 169, 167, 34, 239, 41, 253, 169, 3, 170, 175, 34, 179, 34, 60, 70, 244, 41, 9, 170, 216, 39, 187, 34, 191, 34, 195, 34, 179, 9, 203, 34, 216, 2, 15, 170, 21, 170, 203, 42, 27, 170, 72, 43, 33, 170, 249, 41, 211, 34, 140, 70, 215, 34, 77, 43, 56, 203, 219, 34, 223, 34, 175, 2, 227, 34, 39, 170, 231, 34, 235, 34, 3, 42, 239, 34, 243, 34, 240, 40, 247, 34, 192, 107, 135, 9, 179, 2, 45, 170, 51, 170, 57, 170, 63, 170, 255, 34, 3, 35, 7, 35, 208, 9, 75, 170, 81, 170, 87, 170, 3, 109, 95, 35, 99, 35, 37, 44, 99, 170, 105, 170, 83, 201, 182, 207, 111, 170, 103, 35, 107, 35, 117, 170, 123, 170, 129, 170, 119, 35, 123, 35, 135, 170, 141, 170, 147, 170, 127, 35, 131, 35, 135, 35, 132, 44, 193, 111, 139, 35, 143, 35, 139, 9, 153, 170, 147, 35, 115, 70, 160, 40, 151, 35, 159, 170, 155, 35, 159, 35, 0, 10, 165, 170, 171, 170, 177, 170, 183, 170, 189, 170, 195, 170, 204, 41, 163, 35, 167, 35, 171, 35, 201, 170, 209, 110, 175, 35, 179, 35, 207, 170, 213, 170, 219, 170, 225, 170, 183, 35, 187, 35, 231, 170, 191, 35, 195, 35, 199, 35, 203, 35, 211, 35, 252, 43, 215, 35, 219, 35, 183, 9, 237, 170, 243, 170, 223, 35, 249, 170, 231, 35, 98, 42, 235, 35, 255, 170, 239, 35, 150, 40, 185, 70, 191, 39, 155, 40, 243, 35, 247, 35, 251, 35, 255, 35, 47, 110, 208, 42, 5, 171, 29, 41, 11, 171, 17, 171, 23, 171, 29, 171, 35, 171, 41, 171, 47, 171, 53, 171, 3, 36, 213, 42, 218, 42, 59, 171, 223, 42, 34, 41, 195, 2, 220, 2, 65, 171, 209, 41, 215, 110, 71, 171, 77, 171, 7, 36, 11, 36, 15, 36, 83, 171, 19, 36, 89, 171, 95, 171, 101, 171, 107, 171, 228, 42, 23, 36, 27, 36, 187, 9, 113, 171, 119, 171, 31, 36, 125, 171, 35, 36, 39, 36, 43, 36, 131, 171, 47, 36, 51, 36, 137, 171, 55, 36, 143, 171, 149, 171, 155, 171, 63, 36, 67, 36, 71, 36, 75, 36, 79, 36, 161, 171, 91, 2, 83, 36, 87, 36, 91, 36, 167, 171, 95, 36, 173, 171, 199, 2, 179, 171, 203, 2, 185, 171, 191, 171, 103, 36, 107, 36, 115, 36, 197, 171, 191, 9, 203, 171, 209, 171, 195, 9, 215, 171, 119, 36, 123, 36, 127, 36, 131, 36, 135, 36, 139, 36, 143, 36, 147, 36, 151, 36, 155, 36, 221, 171, 159, 36, 227, 171, 163, 36, 233, 171, 207, 2, 239, 171, 167, 36, 245, 171, 171, 36, 175, 36, 179, 36, 24, 2, 183, 36, 251, 171, 187, 36, 1, 172, 195, 36, 199, 9, 199, 36, 203, 36, 7, 172, 211, 36, 215, 36, 219, 36, 223, 36, 13, 172, 19, 172, 25, 172, 31, 172, 231, 36, 235, 36, 37, 172, 239, 36, 243, 36, 247, 36, 43, 172, 49, 172, 55, 172, 203, 9, 61, 172, 7, 37, 67, 172, 73, 172, 79, 172, 85, 172, 11, 37, 15, 37, 19, 37, 91, 172, 23, 37, 27, 37, 31, 37, 35, 37, 43, 37, 97, 172, 103, 172, 47, 37, 51, 37, 55, 37, 109, 172, 115, 172, 59, 37, 63, 37, 121, 172, 127, 172, 133, 172, 67, 37, 139, 172, 145, 172, 151, 172, 157, 172, 79, 37, 163, 172, 169, 172, 175, 172, 181, 172, 187, 172, 83, 37, 211, 2, 87, 37, 193, 172, 215, 2, 199, 172, 207, 9, 91, 37, 205, 172, 95, 37, 99, 37, 211, 172, 103, 37, 217, 172, 219, 2, 107, 37, 111, 37, 115, 37, 119, 37, 123, 37, 127, 37, 131, 37, 135, 37, 223, 172, 229, 172, 139, 37, 143, 37, 59, 1, 147, 37, 151, 37, 235, 172, 241, 172, 159, 37, 247, 172, 253, 172, 163, 37, 202, 43, 167, 37, 171, 37, 163, 1, 126, 39, 179, 37, 183, 37, 191, 37, 195, 37, 223, 2, 3, 173, 211, 9, 9, 173, 203, 37, 207, 37, 211, 37, 15, 173, 215, 37, 21, 173, 227, 2, 219, 37, 157, 43, 223, 37, 200, 70, 227, 37, 92, 2, 231, 2, 27, 173, 33, 173, 235, 37, 243, 37, 247, 37, 235, 2, 215, 9, 39, 173, 251, 37, 45, 173, 239, 2, 51, 173, 255, 37, 243, 2, 3, 38, 57, 173, 7, 38, 11, 38, 15, 38, 63, 173, 69, 173, 75, 173, 19, 38, 23, 38, 27, 38, 81, 173, 31, 38, 87, 173, 93, 173, 138, 42, 39, 38, 43, 38, 221, 40, 47, 38, 51, 38, 55, 38, 99, 173, 59, 38, 105, 173, 111, 173, 117, 173, 63, 38, 123, 173, 67, 38, 129, 173, 71, 38, 135, 173, 79, 38, 251, 2, 83, 38, 141, 173, 223, 9, 227, 9, 147, 173, 153, 173, 87, 38, 201, 39, 231, 9, 91, 38, 235, 9, 239, 9, 159, 173, 165, 173, 171, 173, 214, 41, 177, 173, 103, 38, 107, 38, 143, 42, 111, 38, 115, 38, 119, 38, 183, 173, 189, 173, 123, 38, 127, 38, 131, 38, 195, 173, 135, 38, 139, 38, 143, 38, 147, 38, 151, 38, 155, 38, 159, 38, 4, 10, 201, 173, 207, 173, 213, 173, 163, 38, 255, 2, 167, 38, 171, 38, 175, 38, 243, 9, 247, 9, 179, 38, 183, 38, 219, 173, 31, 1, 225, 173, 187, 38, 231, 173, 191, 38, 60, 1, 27, 1, 195, 38, 111, 1, 199, 38, 237, 173, 243, 173, 249, 173, 65, 110, 251, 9, 203, 38, 255, 173, 207, 38, 5, 174, 3, 3, 11, 174, 156, 44, 17, 174, 211, 38, 215, 38, 219, 38, 23, 174, 227, 38, 231, 38, 235, 38, 29, 174, 140, 40, 35, 174, 165, 40, 239, 38, 243, 38, 7, 3, 247, 38, 199, 1, 251, 38, 41, 174, 47, 174, 255, 38, 3, 39, 7, 39, 53, 174, 11, 39, 206, 39, 59, 174, 186, 39, 211, 39, 11, 3, 15, 3, 65, 174, 71, 174, 77, 174, 83, 174, 160, 2, 15, 39, 19, 39, 89, 174, 23, 39, 255, 9, 35, 39, 39, 39, 43, 39, 47, 39, 3, 10, 51, 39, 6, 0, 163, 157, 169, 157, 175, 157, 181, 157, 187, 157, 193, 157, 3, 0, 19, 28, 185, 110, 27, 39, 12, 0, 120, 2, 148, 0, 108, 2, 148, 2, 157, 0, 12, 2, 172, 0, 20, 2, 132, 2, 8, 2, 16, 2, 178, 0, 24, 0, 37, 133, 183, 109, 153, 112, 189, 109, 43, 133, 49, 133, 201, 109, 221, 107, 183, 112, 55, 133, 189, 112, 103, 106, 195, 112, 201, 112, 31, 133, 91, 3, 61, 133, 37, 109, 75, 112, 89, 113, 144, 2, 184, 0, 169, 0, 80, 3, 2, 0, 222, 72, 227, 72, 120, 0, 68, 71, 52, 109, 73, 71, 60, 107, 78, 71, 83, 71, 212, 1, 93, 71, 98, 71, 103, 71, 108, 71, 113, 71, 118, 71, 178, 106, 128, 71, 133, 71, 138, 71, 110, 108, 143, 71, 58, 106, 236, 111, 58, 71, 20, 108, 116, 108, 10, 0, 154, 0, 162, 107, 92, 1, 60, 39, 163, 71, 173, 71, 244, 106, 80, 2, 118, 106, 178, 71, 183, 71, 188, 71, 193, 71, 206, 108, 32, 108, 144, 199, 208, 71, 96, 1, 213, 71, 238, 71, 243, 71, 248, 71, 100, 109, 102, 107, 253, 71, 94, 106, 7, 72, 12, 72, 168, 1, 244, 1, 79, 0, 17, 72, 37, 72, 184, 1, 148, 1, 42, 72, 82, 0, 68, 1, 57, 72, 62, 72, 214, 106, 136, 0, 77, 72, 134, 108, 112, 106, 82, 72, 120, 107, 50, 108, 126, 107, 240, 1, 140, 108, 102, 72, 107, 72, 112, 72, 138, 107, 168, 107, 88, 0, 252, 1, 72, 1, 64, 1, 40, 1, 12, 107, 157, 72, 162, 72, 167, 72, 152, 108, 172, 72, 112, 2, 182, 72, 187, 72, 192, 72, 19, 0, 74, 108, 197, 72, 148, 106, 120, 1, 82, 106, 36, 2, 133, 199, 207, 72, 224, 1, 86, 108, 158, 108, 142, 109, 207, 210, 196, 1, 42, 107, 136, 1, 150, 107, 61, 0, 64, 0, 192, 1, 70, 106, 48, 107, 54, 107, 4, 0, 84, 3, 122, 72, 127, 72, 132, 72, 2, 0, 242, 174, 249, 174, 157, 0, 184, 192, 170, 44, 88, 71, 200, 191, 175, 44, 120, 192, 180, 44, 185, 44, 190, 44, 123, 71, 195, 44, 232, 191, 39, 70, 199, 40, 207, 174, 235, 39, 200, 44, 205, 44, 90, 39, 210, 44, 169, 40, 215, 44, 148, 71, 204, 40, 220, 44, 192, 191, 208, 191, 153, 71, 158, 71, 176, 192, 225, 44, 230, 44, 235, 44, 168, 71, 0, 192, 56, 192, 24, 193, 198, 71, 203, 71, 240, 44, 245, 44, 115, 39, 250, 44, 144, 192, 218, 71, 223, 71, 255, 44, 4, 45, 214, 174, 228, 71, 16, 192, 44, 70, 233, 71, 9, 45, 49, 70, 54, 70, 59, 70, 240, 39, 2, 72, 14, 45, 19, 45, 174, 40, 179, 40, 22, 72, 128, 192, 192, 192, 27, 72, 80, 192, 24, 45, 32, 72, 29, 45, 64, 192, 184, 191, 34, 45, 39, 45, 221, 174, 44, 45, 49, 45, 54, 45, 59, 45, 67, 72, 64, 45, 69, 45, 64, 70, 72, 72, 74, 45, 79, 45, 69, 70, 136, 192, 84, 45, 168, 192, 89, 45, 94, 45, 204, 70, 24, 192, 87, 72, 99, 45, 104, 45, 92, 72, 104, 192, 97, 72, 109, 45, 114, 45, 119, 45, 88, 192, 124, 45, 129, 45, 134, 45, 117, 72, 139, 45, 184, 40, 144, 45, 149, 45, 154, 45, 159, 45, 164, 45, 209, 40, 152, 72, 215, 39, 189, 40, 79, 40, 169, 45, 5, 44, 112, 192, 174, 45, 179, 45, 200, 192, 72, 192, 177, 72, 74, 70, 214, 40, 219, 40, 184, 45, 250, 39, 155, 39, 189, 45, 194, 45, 199, 45, 204, 45, 96, 192, 202, 72, 209, 45, 40, 192, 130, 39, 214, 45, 219, 45, 110, 39, 120, 39, 224, 45, 229, 45, 224, 40, 234, 45, 212, 72, 239, 45, 194, 40, 232, 72, 244, 45, 3, 0, 228, 174, 142, 72, 235, 174, 8, 0, 137, 72, 237, 72, 242, 72, 247, 72, 252, 72, 1, 73, 6, 73, 11, 73, 199, 0, 156, 73, 161, 73, 166, 73, 171, 73, 176, 73, 181, 73, 186, 73, 191, 73, 196, 73, 201, 73, 206, 73, 211, 73, 77, 175, 216, 73, 221, 73, 226, 73, 231, 73, 236, 73, 241, 73, 246, 73, 79, 70, 251, 73, 0, 74, 5, 74, 84, 175, 10, 74, 84, 70, 89, 70, 15, 74, 20, 74, 25, 74, 30, 74, 35, 74, 196, 175, 40, 74, 128, 76, 62, 77, 77, 77, 82, 77, 97, 77, 111, 177, 118, 177, 160, 177, 167, 177, 174, 177, 181, 177, 188, 177, 195, 177, 147, 77, 202, 177, 173, 193, 181, 193, 189, 193, 197, 193, 21, 194, 29, 194, 26, 78, 31, 78, 98, 41, 155, 44, 209, 70, 214, 70, 36, 78, 103, 41, 219, 70, 41, 78, 108, 41, 46, 78, 51, 78, 56, 78, 32, 52, 217, 52, 19, 142, 222, 52, 27, 22, 227, 52, 31, 142, 232, 52, 237, 52, 37, 142, 242, 52, 247, 52, 39, 22, 252, 52, 43, 142, 1, 53, 6, 53, 55, 142, 11, 53, 61, 142, 16, 53, 67, 142, 21, 53, 26, 53, 139, 142, 145, 142, 129, 131, 163, 142, 169, 142, 175, 142, 193, 142, 199, 142, 205, 142, 211, 142, 217, 142, 141, 131, 235, 142, 147, 131, 159, 131, 9, 143, 15, 143, 165, 131, 21, 143, 27, 143, 33, 143, 39, 143, 45, 143, 51, 143, 87, 143, 31, 83, 91, 83, 169, 180, 176, 180, 183, 180, 190, 180, 197, 180, 204, 180, 211, 180, 218, 180, 225, 180, 232, 180, 96, 83, 35, 1, 85, 89, 159, 27, 163, 27, 167, 27, 230, 184, 79, 151, 85, 151, 109, 151, 97, 92, 137, 92, 167, 92, 172, 92, 107, 185, 227, 92, 121, 185, 91, 93, 146, 93, 186, 93, 191, 93, 216, 93, 159, 152, 18, 96, 23, 96, 150, 39, 28, 96, 33, 96, 38, 96, 43, 96, 5, 153, 87, 62, 92, 62, 142, 62, 1, 187, 147, 29, 255, 8, 117, 97, 189, 70, 207, 97, 212, 97, 46, 43, 217, 97, 194, 70, 222, 97, 227, 97, 232, 97, 237, 97, 242, 97, 247, 97, 162, 187, 252, 97, 1, 98, 6, 98, 44, 65, 133, 166, 139, 166, 151, 166, 211, 166, 223, 166, 229, 166, 235, 166, 1, 68, 36, 68, 86, 68, 131, 68, 151, 68, 166, 68, 86, 11, 7, 175, 249, 45, 254, 45, 3, 46, 8, 46, 229, 40, 13, 46, 234, 40, 61, 193, 253, 109, 99, 113, 105, 113, 111, 113, 117, 113, 123, 113, 3, 110, 129, 113, 135, 113, 141, 113, 9, 110, 147, 113, 153, 113, 159, 113, 165, 113, 171, 113, 177, 113, 183, 113, 189, 113, 195, 113, 201, 113, 207, 113, 213, 113, 219, 113, 225, 113, 231, 113, 237, 113, 243, 113, 249, 107, 249, 113, 255, 113, 5, 114, 35, 175, 42, 175, 63, 175, 70, 175, 18, 46, 226, 107, 23, 46, 28, 46, 33, 46, 38, 46, 43, 46, 48, 46, 53, 46, 58, 46, 63, 46, 11, 114, 17, 114, 23, 114, 29, 114, 35, 114, 77, 193, 41, 114, 47, 114, 53, 114, 59, 114, 65, 114, 81, 107, 71, 114, 77, 114, 83, 114, 255, 107, 15, 110, 153, 134, 171, 134, 183, 134, 207, 134, 89, 114, 95, 114, 68, 46, 101, 114, 73, 46, 107, 114, 113, 114, 239, 40, 119, 114, 125, 114, 131, 114, 137, 114, 143, 114, 149, 114, 155, 114, 10, 44, 161, 114, 21, 110, 167, 114, 173, 114, 179, 114, 185, 114, 191, 114, 197, 114, 203, 114, 209, 114, 255, 39, 215, 114, 221, 114, 227, 114, 233, 114, 239, 114, 27, 110, 245, 114, 251, 114, 1, 115, 7, 115, 4, 40, 13, 115, 19, 115, 25, 115, 31, 115, 180, 39, 139, 40, 37, 115, 15, 44, 78, 46, 33, 110, 43, 115, 49, 115, 55, 115, 61, 115, 67, 115, 73, 115, 79, 115, 39, 110, 83, 46, 85, 115, 91, 115, 97, 115, 117, 193, 103, 115, 109, 115, 115, 115, 121, 115, 127, 115, 144, 40, 133, 115, 139, 115, 185, 39, 145, 115, 151, 115, 157, 115, 163, 115, 88, 46, 169, 115, 175, 115, 181, 115, 187, 115, 193, 115, 93, 46, 199, 115, 209, 111, 125, 193, 205, 115, 211, 115, 217, 115, 210, 175, 217, 175, 231, 175, 238, 175, 245, 175, 252, 175, 3, 176, 10, 176, 17, 176, 149, 135, 223, 115, 229, 115, 235, 115, 241, 115, 247, 115, 253, 115, 3, 116, 9, 116, 15, 116, 5, 108, 45, 110, 21, 116, 27, 116, 38, 176, 45, 176, 141, 193, 51, 110, 33, 116, 39, 116, 45, 116, 51, 116, 57, 116, 63, 116, 69, 116, 75, 116, 81, 116, 87, 116, 93, 116, 99, 116, 105, 116, 111, 116, 149, 193, 117, 116, 123, 116, 129, 116, 135, 116, 141, 116, 147, 116, 153, 116, 159, 116, 165, 116, 171, 116, 177, 116, 183, 116, 57, 110, 189, 116, 195, 116, 201, 116, 207, 116, 213, 116, 219, 116, 225, 116, 231, 116, 237, 116, 63, 110, 243, 116, 249, 116, 59, 176, 66, 176, 73, 176, 87, 176, 98, 46, 103, 46, 101, 176, 244, 40, 108, 46, 113, 46, 118, 46, 249, 40, 123, 46, 128, 46, 254, 40, 133, 46, 138, 46, 3, 41, 143, 46, 148, 46, 153, 46, 158, 46, 115, 176, 8, 41, 163, 46, 168, 46, 173, 46, 178, 46, 183, 46, 188, 46, 193, 46, 198, 46, 203, 46, 208, 46, 213, 46, 218, 46, 13, 41, 79, 136, 15, 109, 229, 108, 91, 136, 103, 136, 223, 46, 228, 46, 233, 46, 238, 46, 164, 176, 243, 46, 248, 46, 253, 46, 2, 47, 7, 47, 12, 47, 17, 47, 22, 47, 171, 176, 178, 176, 27, 47, 32, 47, 37, 47, 42, 47, 185, 176, 192, 176, 47, 47, 52, 47, 57, 47, 62, 47, 67, 47, 72, 47, 77, 47, 82, 47, 87, 47, 92, 47, 97, 47, 102, 47, 107, 47, 18, 41, 112, 47, 117, 47, 214, 107, 122, 47, 127, 47, 132, 47, 137, 47, 142, 47, 147, 47, 152, 47, 157, 47, 206, 176, 162, 47, 23, 41, 167, 47, 172, 47, 177, 47, 182, 47, 187, 47, 9, 40, 192, 47, 220, 176, 227, 176, 234, 176, 241, 176, 197, 47, 202, 47, 207, 47, 212, 47, 80, 39, 217, 47, 222, 47, 20, 44, 227, 47, 149, 40, 154, 40, 28, 41, 232, 47, 33, 41, 25, 44, 6, 177, 237, 47, 14, 40, 242, 47, 30, 44, 160, 39, 247, 47, 252, 47, 35, 44, 1, 48, 19, 40, 255, 116, 6, 48, 220, 39, 5, 117, 11, 48, 16, 48, 21, 48, 26, 48, 31, 48, 34, 177, 36, 48, 41, 48, 46, 48, 51, 48, 56, 48, 61, 48, 66, 48, 71, 48, 55, 177, 76, 48, 38, 41, 81, 48, 86, 48, 91, 48, 43, 41, 96, 48, 101, 48, 106, 48, 111, 48, 116, 48, 62, 177, 121, 48, 126, 48, 131, 48, 136, 48, 141, 48, 146, 48, 151, 48, 156, 48, 161, 48, 166, 48, 11, 117, 171, 48, 176, 48, 181, 48, 186, 48, 191, 48, 196, 48, 201, 48, 206, 48, 48, 41, 211, 48, 216, 48, 53, 41, 221, 48, 226, 48, 83, 177, 127, 7, 131, 7, 93, 137, 17, 117, 231, 48, 236, 48, 241, 48, 246, 48, 251, 48, 0, 49, 5, 49, 10, 49, 15, 49, 20, 49, 25, 49, 30, 49, 35, 49, 40, 49, 45, 49, 97, 177, 50, 49, 55, 49, 60, 49, 58, 41, 65, 49, 70, 49, 75, 49, 102, 174, 80, 49, 85, 49, 90, 49, 95, 49, 100, 49, 105, 49, 63, 41, 110, 49, 115, 49, 120, 49, 68, 41, 125, 49, 130, 49, 73, 41, 135, 49, 140, 49, 145, 49, 150, 49, 155, 49, 160, 49, 165, 49, 170, 49, 175, 49, 180, 49, 185, 49, 190, 49, 195, 49, 78, 41, 200, 49, 83, 41, 205, 49, 210, 49, 215, 49, 220, 49, 225, 49, 230, 49, 235, 49, 240, 49, 245, 49, 250, 49, 255, 49, 4, 50, 9, 50, 40, 44, 14, 50, 19, 50, 24, 50, 29, 50, 209, 177, 34, 50, 39, 50, 216, 177, 44, 50, 49, 50, 54, 50, 59, 50, 88, 41, 64, 50, 69, 50, 93, 41, 74, 50, 223, 177, 79, 50, 84, 50, 89, 50, 94, 50, 99, 50, 104, 50, 109, 50, 84, 40, 114, 50, 23, 117, 29, 117, 35, 117, 41, 117, 227, 108, 47, 117, 53, 117, 59, 117, 65, 117, 71, 117, 77, 117, 83, 117, 89, 117, 95, 117, 119, 50, 101, 117, 69, 110, 107, 117, 113, 117, 119, 117, 125, 117, 131, 117, 75, 110, 137, 117, 143, 117, 149, 117, 155, 117, 81, 110, 161, 117, 167, 117, 173, 117, 179, 117, 185, 117, 143, 7, 147, 7, 23, 138, 151, 7, 155, 7, 124, 50, 129, 50, 9, 178, 16, 178, 45, 44, 163, 7, 134, 50, 139, 50, 144, 50, 149, 50, 154, 50, 159, 50, 164, 50, 169, 50, 174, 50, 179, 50, 184, 50, 189, 50, 194, 50, 167, 7, 24, 40, 199, 50, 30, 178, 204, 50, 209, 50, 214, 50, 219, 50, 50, 44, 58, 178, 224, 50, 229, 50, 205, 193, 213, 193, 234, 50, 221, 193, 239, 50, 229, 193, 244, 50, 237, 193, 249, 50, 245, 193, 55, 44, 253, 193, 60, 44, 5, 194, 254, 50, 13, 194, 3, 51, 37, 194, 65, 44, 45, 194, 8, 51, 53, 194, 70, 44, 61, 194, 75, 44, 69, 194, 13, 51, 18, 51, 23, 51, 28, 51, 33, 51, 38, 51, 43, 51, 48, 51, 53, 51, 58, 51, 63, 51, 68, 51, 89, 40, 73, 51, 94, 40, 113, 41, 100, 178, 107, 178, 118, 41, 78, 51, 121, 178, 135, 178, 142, 178, 156, 178, 163, 178, 170, 178, 177, 178, 184, 178, 198, 178, 205, 178, 212, 178, 219, 178, 226, 178, 59, 141, 65, 141, 191, 117, 247, 178, 71, 141, 77, 141, 83, 51, 88, 51, 93, 51, 98, 51, 103, 51, 108, 51, 113, 51, 118, 51, 123, 51, 128, 51, 133, 51, 138, 51, 143, 51, 148, 51, 153, 51, 197, 117, 203, 117, 209, 117, 158, 51, 163, 51, 215, 117, 183, 7, 168, 51, 173, 51, 221, 117, 178, 51, 183, 51, 188, 51, 193, 51, 198, 51, 227, 117, 203, 51, 208, 51, 123, 41, 233, 117, 213, 51, 239, 117, 245, 117, 251, 117, 1, 118, 7, 118, 13, 118, 218, 51, 223, 51, 228, 51, 233, 51, 238, 51, 243, 51, 130, 40, 248, 51, 253, 51, 2, 52, 47, 179, 7, 52, 187, 7, 12, 52, 17, 52, 22, 52, 191, 7, 54, 179, 27, 52, 61, 179, 37, 52, 68, 179, 128, 41, 42, 52, 133, 41, 47, 52, 52, 52, 57, 52, 135, 40, 138, 41, 62, 52, 81, 44, 90, 40, 67, 52, 72, 52, 77, 52, 241, 108, 82, 52, 75, 179, 87, 52, 92, 52, 97, 52, 102, 52, 105, 40, 107, 52, 195, 7, 112, 52, 117, 52, 143, 41, 122, 52, 173, 107, 127, 52, 132, 52, 137, 52, 142, 52, 147, 52, 95, 40, 82, 179, 152, 52, 199, 7, 157, 52, 162, 52, 167, 52, 100, 40, 172, 52, 177, 52, 29, 40, 182, 52, 187, 52, 192, 52, 197, 52, 202, 52, 207, 52, 96, 179, 212, 52, 110, 179, 117, 179, 131, 179, 138, 179, 211, 7, 215, 7, 219, 7, 223, 7, 227, 7, 19, 118, 25, 118, 31, 118, 37, 118, 187, 179, 194, 179, 201, 179, 208, 179, 215, 179, 222, 179, 229, 179, 236, 179, 243, 179, 250, 179, 1, 180, 87, 110, 43, 118, 49, 118, 55, 118, 61, 118, 93, 110, 67, 118, 73, 118, 79, 118, 99, 110, 85, 118, 91, 118, 97, 118, 103, 118, 109, 118, 115, 118, 103, 142, 109, 194, 121, 118, 127, 118, 133, 118, 139, 118, 145, 118, 151, 118, 157, 118, 117, 194, 163, 118, 169, 118, 175, 118, 181, 118, 187, 118, 193, 118, 199, 118, 205, 118, 211, 118, 217, 118, 223, 118, 229, 118, 235, 118, 125, 194, 241, 118, 247, 118, 253, 118, 3, 119, 9, 119, 15, 119, 21, 119, 27, 119, 33, 119, 39, 119, 105, 110, 45, 119, 51, 119, 57, 119, 111, 110, 63, 119, 69, 119, 75, 119, 81, 119, 87, 119, 93, 119, 117, 110, 99, 119, 105, 119, 123, 110, 111, 119, 133, 194, 117, 119, 123, 119, 129, 110, 129, 119, 135, 119, 141, 119, 147, 119, 153, 119, 159, 119, 165, 119, 171, 119, 141, 194, 135, 110, 177, 119, 171, 107, 183, 119, 189, 119, 195, 119, 141, 110, 201, 119, 207, 119, 213, 119, 219, 119, 225, 119, 147, 110, 233, 108, 149, 194, 231, 119, 237, 119, 243, 119, 249, 119, 255, 119, 5, 120, 153, 110, 11, 120, 17, 120, 23, 120, 29, 120, 35, 120, 159, 110, 41, 120, 47, 120, 53, 120, 59, 120, 165, 110, 65, 120, 71, 120, 77, 120, 83, 120, 89, 120, 95, 120, 171, 110, 101, 120, 107, 120, 113, 120, 119, 120, 125, 120, 177, 110, 131, 120, 137, 120, 143, 120, 149, 120, 155, 120, 161, 120, 167, 120, 173, 120, 179, 120, 185, 120, 191, 120, 183, 110, 197, 120, 203, 120, 209, 120, 215, 120, 221, 120, 227, 120, 233, 120, 239, 120, 245, 120, 251, 120, 1, 121, 7, 121, 13, 121, 19, 121, 25, 121, 189, 110, 31, 121, 37, 121, 43, 121, 49, 121, 55, 121, 61, 121, 67, 121, 73, 121, 79, 121, 85, 121, 91, 121, 97, 121, 103, 121, 109, 121, 115, 121, 195, 110, 121, 121, 127, 121, 133, 121, 139, 121, 145, 121, 151, 121, 157, 121, 163, 121, 169, 121, 175, 121, 181, 121, 213, 107, 187, 121, 193, 121, 199, 121, 205, 121, 157, 194, 211, 121, 217, 121, 165, 194, 223, 121, 229, 121, 235, 121, 173, 194, 241, 121, 181, 194, 201, 110, 247, 121, 253, 121, 3, 122, 9, 122, 15, 122, 21, 122, 27, 122, 33, 122, 39, 122, 45, 122, 51, 122, 161, 108, 57, 122, 63, 122, 207, 110, 69, 122, 75, 122, 81, 122, 87, 122, 93, 122, 167, 108, 99, 122, 175, 109, 105, 122, 111, 122, 213, 110, 117, 122, 123, 122, 129, 122, 135, 122, 141, 122, 147, 122, 177, 107, 183, 107, 36, 180, 31, 53, 36, 53, 41, 53, 46, 53, 51, 53, 153, 41, 56, 53, 61, 53, 66, 53, 50, 180, 99, 40, 71, 53, 76, 53, 81, 53, 86, 53, 91, 53, 96, 53, 101, 53, 106, 53, 111, 53, 116, 53, 121, 53, 126, 53, 131, 53, 136, 53, 141, 53, 158, 41, 239, 7, 153, 122, 243, 7, 247, 7, 251, 7, 106, 180, 146, 53, 255, 7, 151, 53, 3, 8, 156, 53, 161, 53, 166, 53, 171, 53, 7, 8, 163, 41, 176, 53, 168, 41, 181, 53, 186, 53, 11, 8, 113, 180, 120, 180, 127, 180, 191, 53, 15, 8, 173, 41, 196, 53, 178, 41, 19, 8, 134, 180, 141, 180, 23, 8, 201, 53, 183, 41, 206, 53, 211, 53, 216, 53, 221, 53, 226, 53, 231, 53, 236, 53, 27, 8, 241, 53, 47, 3, 34, 40, 246, 53, 251, 53, 0, 54, 65, 144, 5, 54, 10, 54, 15, 54, 20, 54, 25, 54, 30, 54, 35, 54, 40, 54, 188, 41, 193, 41, 45, 54, 50, 54, 55, 54, 60, 54, 65, 54, 70, 54, 75, 54, 80, 54, 85, 54, 90, 54, 95, 54, 198, 41, 100, 54, 11, 181, 165, 39, 203, 41, 105, 54, 35, 8, 235, 108, 208, 41, 110, 54, 39, 8, 115, 54, 120, 54, 85, 44, 125, 39, 125, 54, 105, 39, 130, 54, 135, 54, 90, 44, 140, 54, 213, 41, 145, 54, 150, 54, 60, 181, 218, 41, 155, 54, 160, 54, 165, 54, 170, 54, 175, 54, 180, 54, 185, 54, 190, 54, 195, 54, 200, 54, 205, 54, 210, 54, 215, 54, 74, 181, 81, 181, 88, 181, 232, 107, 220, 54, 225, 54, 102, 181, 230, 54, 235, 54, 240, 54, 109, 181, 245, 54, 250, 54, 255, 54, 4, 55, 9, 55, 14, 55, 95, 44, 223, 41, 228, 41, 19, 55, 24, 55, 29, 55, 34, 55, 233, 41, 39, 55, 100, 44, 44, 55, 49, 55, 54, 55, 39, 40, 105, 44, 238, 41, 243, 41, 59, 55, 64, 55, 69, 55, 170, 39, 74, 55, 79, 55, 84, 55, 248, 41, 123, 181, 130, 181, 89, 55, 94, 55, 99, 55, 104, 55, 2, 42, 109, 55, 114, 55, 119, 55, 124, 55, 151, 174, 129, 55, 134, 55, 7, 42, 151, 181, 139, 55, 144, 55, 149, 55, 12, 42, 158, 181, 154, 55, 165, 181, 159, 55, 164, 55, 169, 55, 174, 55, 179, 55, 172, 181, 179, 181, 17, 42, 184, 55, 186, 181, 189, 55, 193, 181, 200, 181, 194, 55, 51, 8, 22, 42, 214, 181, 27, 42, 55, 8, 199, 55, 59, 8, 32, 42, 37, 42, 42, 42, 63, 8, 204, 55, 209, 55, 47, 42, 52, 42, 214, 55, 67, 8, 219, 55, 221, 181, 224, 55, 229, 55, 104, 40, 234, 55, 71, 8, 239, 55, 244, 55, 228, 181, 75, 8, 57, 42, 249, 55, 55, 3, 254, 55, 79, 8, 3, 56, 8, 56, 0, 182, 13, 56, 255, 209, 18, 56, 23, 56, 28, 56, 33, 56, 38, 56, 43, 56, 48, 56, 53, 56, 58, 56, 63, 56, 68, 56, 73, 56, 78, 56, 83, 56, 88, 56, 93, 56, 98, 56, 103, 56, 108, 56, 113, 56, 118, 56, 123, 56, 128, 56, 133, 56, 62, 42, 138, 56, 67, 42, 143, 56, 148, 56, 153, 56, 158, 56, 163, 56, 28, 182, 168, 56, 173, 56, 178, 56, 183, 56, 188, 56, 193, 56, 198, 56, 203, 56, 208, 56, 213, 56, 218, 56, 223, 56, 228, 56, 109, 40, 42, 182, 49, 182, 56, 182, 63, 182, 241, 145, 233, 56, 238, 56, 243, 56, 248, 56, 253, 56, 2, 57, 7, 57, 12, 57, 17, 57, 22, 57, 27, 57, 32, 57, 37, 57, 42, 57, 47, 57, 52, 57, 158, 174, 57, 57, 62, 57, 67, 57, 132, 174, 84, 182, 72, 57, 77, 57, 82, 57, 87, 57, 92, 57, 97, 57, 91, 182, 98, 182, 102, 57, 105, 182, 107, 57, 112, 182, 112, 57, 119, 182, 126, 182, 117, 57, 122, 57, 127, 57, 132, 57, 137, 57, 77, 42, 142, 57, 82, 42, 147, 57, 129, 85, 152, 57, 157, 57, 162, 57, 167, 57, 172, 57, 177, 57, 182, 57, 187, 57, 133, 182, 140, 182, 192, 57, 197, 57, 147, 182, 154, 182, 202, 57, 207, 57, 212, 57, 164, 85, 217, 57, 169, 85, 87, 42, 161, 182, 222, 57, 227, 57, 168, 182, 232, 57, 175, 182, 237, 57, 242, 57, 247, 57, 252, 57, 1, 58, 6, 58, 182, 182, 11, 58, 16, 58, 189, 182, 196, 182, 21, 58, 26, 58, 31, 58, 203, 182, 210, 182, 36, 58, 3, 86, 217, 182, 41, 58, 46, 58, 224, 182, 51, 58, 245, 182, 252, 182, 56, 58, 61, 58, 66, 58, 71, 58, 76, 58, 81, 58, 86, 58, 91, 58, 96, 58, 101, 58, 106, 58, 111, 58, 17, 183, 24, 183, 31, 183, 38, 183, 45, 183, 59, 183, 66, 183, 73, 183, 80, 183, 87, 183, 23, 147, 116, 58, 121, 58, 83, 8, 47, 147, 87, 8, 91, 8, 95, 8, 108, 183, 99, 8, 115, 183, 53, 147, 103, 8, 247, 108, 107, 8, 111, 8, 115, 8, 119, 8, 143, 183, 150, 183, 126, 58, 123, 8, 59, 3, 164, 183, 131, 58, 171, 183, 136, 58, 185, 183, 192, 183, 199, 183, 141, 58, 146, 58, 151, 58, 156, 58, 97, 42, 190, 39, 161, 58, 159, 122, 165, 122, 219, 110, 171, 122, 177, 122, 183, 122, 189, 122, 195, 122, 201, 122, 207, 122, 213, 122, 219, 122, 225, 122, 173, 108, 231, 122, 189, 107, 237, 122, 179, 108, 102, 42, 234, 183, 114, 40, 166, 58, 171, 58, 176, 58, 181, 58, 186, 58, 191, 58, 196, 58, 201, 58, 107, 42, 206, 58, 211, 58, 112, 42, 216, 58, 221, 58, 226, 58, 231, 58, 236, 58, 248, 183, 241, 58, 246, 58, 251, 58, 6, 184, 13, 184, 20, 184, 34, 184, 41, 184, 131, 8, 55, 184, 135, 8, 62, 184, 139, 8, 69, 184, 143, 8, 76, 184, 147, 8, 83, 184, 151, 8, 90, 184, 155, 8, 97, 184, 159, 8, 104, 184, 163, 8, 111, 184, 0, 59, 5, 59, 10, 59, 15, 59, 20, 59, 117, 42, 25, 59, 30, 59, 35, 59, 40, 59, 195, 39, 45, 59, 50, 59, 122, 42, 55, 59, 93, 149, 111, 149, 139, 184, 146, 184, 153, 184, 160, 184, 167, 184, 174, 184, 60, 59, 65, 59, 70, 59, 75, 59, 80, 59, 85, 59, 90, 59, 95, 59, 100, 59, 105, 59, 110, 59, 115, 59, 120, 59, 125, 59, 130, 59, 127, 42, 135, 59, 132, 42, 140, 59, 145, 59, 44, 40, 150, 59, 155, 59, 29, 195, 37, 195, 45, 195, 53, 195, 61, 195, 69, 195, 77, 195, 85, 195, 93, 195, 101, 195, 109, 195, 243, 122, 37, 112, 249, 122, 255, 122, 5, 123, 225, 110, 11, 123, 17, 123, 231, 110, 23, 123, 160, 59, 165, 59, 170, 59, 175, 59, 115, 44, 180, 59, 185, 59, 190, 59, 195, 59, 200, 59, 202, 184, 205, 59, 175, 39, 137, 42, 210, 59, 215, 59, 220, 59, 142, 42, 225, 59, 230, 59, 235, 59, 216, 184, 223, 184, 29, 123, 35, 123, 41, 123, 47, 123, 195, 107, 117, 195, 53, 123, 59, 123, 65, 123, 71, 123, 77, 123, 83, 123, 237, 110, 89, 123, 95, 123, 101, 123, 239, 108, 107, 123, 113, 123, 119, 123, 125, 123, 125, 195, 131, 123, 137, 123, 143, 123, 149, 123, 243, 110, 155, 123, 161, 123, 167, 123, 173, 123, 179, 123, 133, 195, 185, 123, 191, 123, 225, 107, 197, 123, 203, 123, 209, 123, 215, 123, 221, 123, 227, 123, 249, 110, 233, 123, 239, 123, 245, 123, 251, 123, 1, 124, 7, 124, 13, 124, 19, 124, 25, 124, 255, 110, 5, 111, 31, 124, 37, 124, 141, 195, 43, 124, 49, 124, 55, 124, 61, 124, 67, 124, 73, 124, 201, 107, 79, 124, 85, 124, 91, 124, 97, 124, 103, 124, 109, 124, 115, 124, 240, 59, 245, 59, 250, 59, 147, 42, 152, 42, 157, 42, 162, 42, 167, 42, 172, 42, 177, 42, 255, 59, 4, 60, 9, 60, 244, 184, 14, 60, 19, 60, 24, 60, 29, 60, 34, 60, 2, 185, 39, 60, 9, 185, 16, 185, 44, 60, 49, 60, 54, 60, 59, 60, 64, 60, 69, 60, 74, 60, 79, 60, 182, 42, 84, 60, 89, 60, 94, 60, 99, 60, 104, 60, 109, 60, 187, 42, 114, 60, 119, 60, 165, 174, 124, 60, 129, 60, 134, 60, 139, 60, 144, 60, 30, 185, 37, 185, 44, 185, 191, 8, 195, 8, 199, 8, 203, 8, 181, 151, 149, 195, 11, 111, 121, 124, 17, 111, 127, 124, 133, 124, 139, 124, 245, 108, 145, 124, 207, 8, 151, 124, 157, 124, 163, 124, 169, 124, 175, 124, 181, 124, 187, 124, 185, 108, 193, 124, 199, 124, 211, 8, 205, 124, 191, 108, 211, 124, 217, 124, 79, 185, 86, 185, 149, 60, 93, 185, 100, 185, 154, 60, 114, 185, 128, 185, 135, 185, 142, 185, 149, 185, 159, 60, 156, 185, 164, 60, 163, 185, 169, 60, 215, 8, 174, 60, 15, 152, 223, 124, 184, 185, 191, 185, 229, 124, 235, 124, 241, 124, 247, 124, 253, 124, 165, 195, 3, 125, 9, 125, 15, 125, 21, 125, 27, 125, 33, 125, 39, 125, 173, 195, 45, 125, 51, 125, 57, 125, 63, 125, 219, 8, 233, 185, 240, 185, 5, 186, 12, 186, 19, 186, 26, 186, 33, 186, 40, 186, 47, 186, 179, 60, 192, 42, 197, 42, 184, 60, 189, 60, 194, 60, 199, 60, 204, 60, 209, 60, 214, 60, 219, 60, 224, 60, 229, 60, 234, 60, 202, 42, 239, 60, 244, 60, 68, 186, 75, 186, 82, 186, 249, 60, 254, 60, 89, 186, 69, 125, 3, 61, 8, 61, 223, 8, 13, 61, 18, 61, 23, 61, 28, 61, 33, 61, 103, 186, 207, 42, 38, 61, 43, 61, 110, 186, 48, 61, 53, 61, 58, 61, 227, 8, 140, 39, 225, 39, 212, 42, 217, 42, 63, 61, 222, 42, 68, 61, 85, 39, 231, 8, 227, 42, 120, 44, 73, 61, 75, 125, 81, 125, 87, 125, 93, 125, 99, 125, 105, 125, 111, 125, 23, 111, 117, 125, 123, 125, 29, 111, 129, 125, 35, 111, 135, 125, 141, 125, 147, 125, 153, 125, 159, 125, 251, 108, 165, 125, 171, 125, 177, 125, 183, 125, 189, 125, 195, 125, 201, 125, 207, 125, 43, 112, 213, 125, 1, 109, 219, 125, 225, 125, 131, 186, 138, 186, 41, 153, 235, 8, 239, 8, 243, 8, 231, 125, 247, 8, 59, 153, 65, 153, 78, 61, 83, 61, 88, 61, 93, 61, 98, 61, 103, 61, 108, 61, 113, 61, 118, 61, 123, 61, 128, 61, 159, 186, 133, 61, 137, 153, 138, 61, 143, 61, 148, 61, 232, 42, 153, 61, 158, 61, 237, 42, 163, 61, 242, 42, 168, 61, 173, 61, 178, 61, 183, 61, 188, 61, 193, 61, 198, 61, 203, 61, 247, 42, 172, 174, 194, 186, 208, 61, 213, 61, 218, 61, 223, 61, 228, 61, 233, 61, 238, 61, 243, 61, 248, 61, 253, 61, 252, 42, 2, 62, 1, 43, 7, 62, 12, 62, 17, 62, 22, 62, 27, 62, 6, 43, 32, 62, 37, 62, 42, 62, 47, 62, 52, 62, 57, 62, 62, 62, 67, 62, 215, 186, 72, 62, 77, 62, 82, 62, 197, 195, 97, 62, 102, 62, 107, 62, 112, 62, 117, 62, 122, 62, 127, 62, 132, 62, 137, 62, 147, 62, 152, 62, 119, 40, 236, 186, 243, 186, 250, 186, 8, 187, 15, 187, 79, 154, 157, 62, 162, 62, 167, 62, 124, 40, 172, 62, 177, 62, 11, 43, 182, 62, 187, 62, 192, 62, 16, 43, 29, 187, 21, 43, 26, 43, 31, 43, 43, 187, 50, 187, 237, 125, 41, 111, 243, 125, 249, 125, 255, 125, 5, 126, 11, 126, 17, 126, 23, 126, 29, 126, 35, 126, 41, 126, 47, 126, 53, 126, 151, 106, 49, 112, 59, 126, 197, 62, 202, 62, 207, 62, 212, 62, 217, 62, 222, 62, 227, 62, 232, 62, 237, 62, 36, 43, 242, 62, 7, 9, 247, 62, 189, 155, 252, 62, 1, 63, 6, 63, 71, 126, 77, 126, 83, 126, 89, 126, 95, 126, 101, 126, 107, 126, 113, 126, 119, 126, 125, 126, 131, 126, 137, 126, 213, 195, 143, 126, 149, 126, 155, 126, 161, 126, 167, 126, 173, 126, 211, 109, 179, 126, 185, 126, 191, 126, 7, 109, 197, 126, 203, 126, 209, 126, 47, 111, 215, 126, 221, 126, 227, 126, 233, 126, 239, 126, 245, 126, 251, 126, 221, 195, 1, 127, 7, 127, 13, 127, 19, 127, 25, 127, 31, 127, 37, 127, 43, 127, 49, 127, 55, 127, 120, 187, 127, 187, 134, 187, 141, 187, 61, 127, 67, 127, 73, 127, 79, 127, 85, 127, 91, 127, 97, 127, 103, 127, 109, 127, 115, 127, 121, 127, 127, 127, 13, 109, 133, 127, 139, 127, 229, 195, 145, 127, 237, 195, 151, 127, 245, 195, 157, 127, 19, 109, 163, 127, 53, 111, 169, 127, 175, 127, 181, 127, 187, 127, 193, 127, 59, 111, 65, 111, 71, 111, 199, 127, 77, 111, 83, 111, 205, 127, 211, 127, 217, 127, 223, 127, 229, 127, 235, 127, 241, 127, 247, 127, 253, 127, 3, 128, 9, 128, 15, 128, 21, 128, 27, 128, 33, 128, 39, 128, 45, 128, 51, 128, 57, 128, 63, 128, 253, 195, 69, 128, 75, 128, 81, 128, 87, 128, 93, 128, 99, 128, 105, 128, 111, 128, 25, 109, 117, 128, 123, 128, 55, 112, 129, 128, 135, 128, 231, 107, 141, 128, 147, 128, 153, 128, 61, 112, 159, 128, 165, 128, 11, 63, 41, 43, 16, 63, 21, 63, 26, 63, 31, 63, 36, 63, 41, 63, 46, 63, 51, 63, 56, 63, 61, 63, 66, 63, 71, 63, 76, 63, 81, 63, 86, 63, 91, 63, 96, 63, 101, 63, 106, 63, 111, 63, 116, 63, 121, 63, 171, 128, 177, 128, 89, 111, 183, 128, 21, 196, 189, 128, 195, 128, 201, 128, 207, 128, 213, 128, 217, 109, 219, 128, 225, 128, 231, 128, 237, 128, 204, 187, 211, 187, 126, 63, 131, 63, 95, 111, 101, 111, 107, 111, 113, 111, 119, 111, 125, 111, 131, 111, 31, 109, 137, 111, 143, 111, 243, 128, 29, 196, 249, 128, 255, 128, 15, 9, 232, 187, 136, 63, 145, 157, 5, 129, 141, 63, 146, 63, 151, 63, 156, 63, 161, 63, 166, 63, 171, 63, 176, 63, 181, 63, 186, 63, 51, 43, 63, 9, 191, 63, 196, 63, 201, 63, 206, 63, 211, 63, 216, 63, 221, 63, 226, 63, 231, 63, 4, 188, 125, 44, 236, 63, 241, 63, 246, 63, 251, 63, 0, 64, 5, 64, 10, 64, 15, 64, 20, 64, 25, 64, 30, 64, 35, 64, 40, 64, 45, 64, 50, 64, 55, 64, 60, 64, 65, 64, 70, 64, 75, 64, 11, 188, 80, 64, 85, 64, 90, 64, 95, 64, 25, 188, 32, 188, 39, 188, 100, 64, 130, 44, 105, 64, 110, 64, 11, 129, 17, 129, 149, 111, 23, 129, 56, 43, 29, 129, 35, 129, 61, 43, 115, 64, 41, 129, 47, 129, 120, 64, 66, 43, 125, 64, 155, 111, 53, 129, 161, 111, 53, 188, 167, 111, 59, 129, 65, 129, 173, 111, 71, 129, 77, 129, 83, 129, 179, 174, 89, 129, 130, 64, 95, 129, 135, 64, 101, 129, 107, 129, 113, 129, 119, 129, 125, 129, 140, 64, 145, 64, 131, 129, 137, 129, 143, 129, 149, 129, 155, 129, 161, 129, 167, 129, 173, 129, 179, 111, 179, 129, 185, 129, 191, 129, 197, 129, 203, 129, 209, 129, 215, 129, 71, 43, 241, 109, 221, 129, 237, 107, 150, 64, 247, 109, 227, 129, 233, 129, 185, 111, 76, 43, 239, 129, 245, 129, 251, 129, 1, 130, 155, 64, 7, 130, 160, 64, 13, 130, 19, 130, 25, 130, 31, 130, 191, 111, 37, 130, 243, 107, 43, 130, 215, 108, 67, 188, 81, 188, 88, 188, 165, 64, 170, 64, 175, 64, 180, 64, 185, 64, 190, 64, 102, 188, 195, 64, 200, 64, 205, 64, 116, 188, 123, 188, 130, 188, 210, 64, 215, 64, 220, 64, 225, 64, 81, 43, 230, 64, 235, 64, 86, 43, 240, 64, 245, 64, 250, 64, 255, 64, 4, 65, 9, 65, 91, 43, 14, 65, 19, 65, 24, 65, 158, 188, 29, 65, 71, 159, 67, 9, 77, 159, 83, 159, 89, 159, 172, 188, 34, 65, 39, 65, 45, 196, 53, 196, 61, 196, 69, 196, 77, 196, 85, 196, 93, 196, 101, 196, 109, 196, 117, 196, 125, 196, 133, 196, 141, 196, 149, 196, 157, 196, 165, 196, 173, 196, 181, 196, 189, 196, 197, 196, 205, 196, 213, 196, 221, 196, 49, 65, 54, 65, 59, 65, 64, 65, 96, 43, 69, 65, 74, 65, 101, 43, 27, 109, 71, 9, 106, 43, 75, 9, 79, 9, 228, 188, 79, 65, 84, 65, 89, 65, 94, 65, 242, 188, 99, 65, 111, 43, 104, 65, 109, 65, 114, 65, 119, 65, 124, 65, 129, 65, 134, 65, 139, 65, 144, 65, 116, 43, 149, 65, 154, 65, 121, 43, 159, 65, 164, 65, 169, 65, 174, 65, 179, 65, 126, 43, 184, 65, 189, 65, 194, 65, 199, 65, 204, 65, 209, 65, 214, 65, 219, 65, 224, 65, 229, 65, 234, 65, 239, 65, 244, 65, 131, 43, 136, 43, 249, 65, 254, 65, 141, 43, 3, 66, 146, 43, 151, 43, 8, 66, 13, 66, 18, 66, 23, 66, 87, 9, 147, 112, 28, 66, 33, 66, 38, 66, 43, 66, 48, 66, 53, 66, 58, 66, 63, 66, 68, 66, 73, 66, 135, 44, 159, 40, 78, 66, 83, 66, 88, 66, 93, 66, 98, 66, 156, 43, 103, 66, 42, 189, 135, 39, 108, 66, 70, 39, 113, 66, 118, 66, 123, 66, 128, 66, 133, 66, 140, 44, 138, 66, 145, 39, 143, 66, 148, 66, 153, 66, 158, 66, 163, 66, 168, 66, 173, 66, 178, 66, 161, 43, 183, 66, 188, 66, 193, 66, 198, 66, 203, 66, 208, 66, 213, 66, 218, 66, 223, 66, 228, 66, 233, 66, 238, 66, 243, 66, 248, 66, 91, 9, 95, 9, 99, 9, 103, 9, 107, 9, 105, 189, 112, 189, 119, 189, 126, 189, 133, 189, 140, 189, 147, 189, 154, 189, 161, 189, 111, 9, 253, 66, 166, 43, 2, 67, 7, 67, 12, 67, 17, 67, 22, 67, 171, 43, 27, 67, 32, 67, 37, 67, 42, 67, 47, 67, 52, 67, 57, 67, 62, 67, 67, 67, 176, 43, 245, 165, 119, 9, 1, 166, 111, 44, 203, 189, 210, 189, 217, 189, 224, 189, 231, 189, 72, 67, 238, 189, 245, 189, 3, 190, 17, 190, 24, 190, 31, 190, 38, 190, 45, 190, 52, 190, 59, 190, 66, 190, 95, 174, 49, 130, 55, 130, 61, 130, 67, 130, 73, 130, 79, 130, 85, 130, 91, 130, 97, 130, 221, 108, 103, 130, 109, 130, 115, 130, 121, 130, 127, 130, 133, 130, 139, 130, 145, 130, 151, 130, 157, 130, 163, 130, 169, 130, 175, 130, 181, 130, 187, 130, 193, 130, 223, 106, 199, 130, 205, 130, 211, 130, 217, 130, 223, 130, 229, 130, 235, 130, 241, 130, 197, 111, 247, 130, 253, 130, 3, 131, 9, 131, 69, 107, 15, 131, 21, 131, 27, 131, 33, 131, 101, 190, 108, 190, 115, 190, 122, 190, 77, 67, 82, 67, 87, 67, 92, 67, 97, 67, 102, 67, 107, 67, 112, 67, 117, 67, 122, 67, 127, 67, 132, 67, 137, 67, 142, 67, 147, 67, 152, 67, 157, 67, 162, 67, 167, 67, 172, 67, 177, 67, 182, 67, 187, 67, 192, 67, 197, 67, 202, 67, 207, 67, 13, 197, 21, 197, 29, 197, 37, 197, 45, 197, 53, 197, 61, 197, 69, 197, 77, 197, 85, 197, 93, 197, 101, 197, 109, 197, 117, 197, 125, 197, 123, 9, 63, 3, 127, 9, 131, 9, 212, 67, 217, 67, 222, 67, 227, 67, 232, 67, 237, 67, 242, 67, 247, 67, 252, 67, 178, 190, 238, 107, 6, 68, 181, 43, 11, 68, 16, 68, 186, 43, 21, 68, 26, 68, 31, 68, 41, 68, 46, 68, 51, 68, 56, 68, 61, 68, 66, 68, 71, 68, 191, 43, 76, 68, 185, 190, 81, 68, 192, 190, 199, 190, 206, 190, 196, 43, 91, 68, 96, 68, 101, 68, 106, 68, 111, 68, 116, 68, 121, 68, 126, 68, 136, 68, 213, 190, 141, 68, 146, 68, 156, 68, 161, 68, 171, 68, 176, 68, 181, 68, 186, 68, 191, 68, 244, 107, 196, 68, 201, 43, 49, 40, 201, 68, 206, 68, 211, 68, 216, 68, 206, 43, 221, 68, 226, 68, 227, 190, 231, 68, 236, 68, 241, 68, 54, 40, 200, 39, 59, 40, 246, 68, 251, 68, 145, 44, 150, 44, 164, 40, 0, 69, 5, 69, 10, 69, 205, 39, 64, 40, 210, 39, 69, 40, 74, 40, 15, 69, 255, 190, 69, 170, 93, 170, 34, 191, 20, 69, 41, 191, 25, 69, 48, 191, 55, 191, 30, 69, 62, 191, 69, 191, 35, 69, 76, 191, 40, 69, 45, 69, 83, 191, 90, 191, 50, 69, 55, 69, 60, 69, 65, 69, 70, 69, 39, 131, 45, 131, 51, 131, 203, 111, 67, 112, 57, 131, 75, 69, 80, 69, 216, 43, 85, 69, 90, 69, 95, 69, 100, 69, 105, 69, 110, 69, 115, 69, 120, 69, 125, 69, 130, 69, 221, 43, 135, 69, 140, 69, 145, 69, 150, 69, 155, 69, 160, 69, 165, 69, 170, 69, 226, 43, 175, 69, 231, 43, 180, 69, 185, 69, 190, 69, 195, 69, 236, 43, 129, 40, 200, 69, 134, 40, 205, 69, 241, 43, 210, 69, 246, 43, 215, 69, 220, 69, 225, 69, 230, 69, 235, 69, 240, 69, 245, 69, 250, 69, 153, 191, 255, 69, 4, 70, 251, 43, 9, 70, 14, 70, 19, 70, 24, 70, 230, 39, 75, 39, 167, 191, 0, 1, 0, 175, 14, 175, 21, 175, 28, 175, 49, 175, 56, 175, 91, 175, 105, 175, 112, 175, 119, 175, 133, 175, 147, 175, 154, 175, 168, 175, 175, 175, 182, 175, 189, 175, 203, 175, 224, 175, 24, 176, 31, 176, 52, 176, 80, 176, 94, 176, 108, 176, 122, 176, 129, 176, 136, 176, 143, 176, 150, 176, 157, 176, 199, 176, 213, 176, 248, 176, 255, 176, 13, 177, 20, 177, 27, 177, 41, 177, 48, 177, 69, 177, 76, 177, 90, 177, 104, 177, 132, 177, 139, 177, 146, 177, 153, 177, 230, 177, 237, 177, 244, 177, 2, 178, 23, 178, 37, 178, 44, 178, 51, 178, 65, 178, 72, 178, 86, 178, 93, 178, 114, 178, 128, 178, 149, 178, 191, 178, 233, 178, 240, 178, 254, 178, 5, 179, 12, 179, 19, 179, 33, 179, 89, 179, 103, 179, 124, 179, 145, 179, 152, 179, 159, 179, 180, 179, 8, 180, 22, 180, 29, 180, 43, 180, 57, 180, 64, 180, 71, 180, 78, 180, 85, 180, 92, 180, 99, 180, 148, 180, 155, 180, 162, 180, 239, 180, 246, 180, 253, 180, 4, 181, 18, 181, 25, 181, 46, 181, 53, 181, 67, 181, 95, 181, 116, 181, 137, 181, 144, 181, 207, 181, 235, 181, 242, 181, 249, 181, 14, 182, 21, 182, 35, 182, 70, 182, 77, 182, 231, 182, 238, 182, 3, 183, 10, 183, 52, 183, 94, 183, 101, 183, 122, 183, 129, 183, 136, 183, 157, 183, 178, 183, 206, 183, 213, 183, 220, 183, 227, 183, 241, 183, 255, 183, 27, 184, 48, 184, 118, 184, 125, 184, 132, 184, 181, 184, 188, 184, 195, 184, 209, 184, 237, 184, 251, 184, 23, 185, 51, 185, 58, 185, 65, 185, 72, 185, 170, 185, 177, 185, 198, 185, 205, 185, 212, 185, 219, 185, 226, 185, 247, 185, 254, 185, 54, 186, 61, 186, 96, 186, 117, 186, 124, 186, 145, 186, 152, 186, 166, 186, 173, 186, 180, 186, 187, 186, 201, 186, 222, 186, 229, 186, 22, 187, 36, 187, 57, 187, 64, 187, 71, 187, 78, 187, 92, 187, 99, 187, 106, 187, 113, 187, 148, 187, 155, 187, 169, 187, 176, 187, 183, 187, 190, 187, 197, 187, 218, 187, 225, 187, 239, 187, 246, 187, 253, 187, 18, 188, 46, 188, 60, 188, 74, 188, 95, 188, 109, 188, 144, 188, 151, 188, 165, 188, 179, 188, 186, 188, 200, 188, 207, 188, 214, 188, 221, 188, 235, 188, 249, 188, 0, 189, 7, 189, 14, 189, 21, 189, 28, 189, 35, 189, 49, 189, 56, 189, 70, 189, 77, 189, 84, 189, 91, 189, 98, 189, 168, 189, 175, 189, 182, 189, 189, 189, 196, 189, 252, 189, 10, 190, 73, 190, 80, 190, 87, 190, 94, 190, 129, 190, 136, 190, 150, 190, 164, 190, 220, 190, 234, 190, 241, 190, 248, 190, 6, 191, 13, 191, 20, 191, 27, 191, 97, 191, 104, 191, 111, 191, 118, 191, 125, 191, 132, 191, 139, 191, 146, 191, 160, 191, 174, 191, 172, 0, 91, 209, 243, 212, 111, 204, 128, 204, 160, 200, 186, 200, 145, 204, 53, 199, 147, 200, 94, 210, 162, 204, 179, 204, 4, 210, 151, 206, 196, 204, 213, 204, 180, 199, 132, 202, 137, 212, 240, 211, 102, 211, 230, 204, 169, 206, 186, 207, 143, 201, 170, 212, 105, 212, 156, 211, 206, 212, 50, 200, 199, 200, 60, 203, 247, 204, 76, 203, 8, 205, 92, 203, 25, 205, 42, 205, 108, 203, 59, 205, 76, 205, 124, 203, 212, 200, 65, 208, 157, 201, 112, 209, 133, 209, 44, 203, 147, 202, 26, 210, 85, 208, 140, 203, 171, 201, 154, 209, 185, 201, 175, 209, 156, 203, 196, 209, 225, 200, 199, 201, 162, 202, 148, 207, 238, 200, 129, 207, 93, 205, 217, 209, 105, 208, 172, 203, 184, 211, 187, 206, 251, 200, 125, 208, 188, 203, 145, 208, 165, 208, 128, 211, 238, 209, 62, 200, 205, 207, 177, 202, 13, 212, 210, 210, 52, 211, 205, 206, 110, 205, 213, 201, 26, 200, 117, 202, 74, 200, 127, 205, 223, 206, 21, 201, 61, 206, 86, 200, 27, 211, 79, 206, 144, 205, 49, 209, 227, 201, 161, 205, 34, 201, 38, 200, 185, 208, 73, 199, 117, 210, 191, 199, 83, 199, 129, 201, 220, 203, 102, 202, 192, 202, 202, 199, 236, 203, 158, 199, 169, 199, 153, 198, 98, 200, 204, 203, 252, 203, 207, 202, 178, 205, 222, 202, 237, 202, 12, 204, 213, 199, 205, 208, 70, 209, 94, 204, 47, 201, 43, 206, 43, 212, 74, 212, 28, 204, 225, 208, 245, 208, 2, 211, 224, 199, 48, 210, 25, 208, 45, 208, 71, 210, 9, 209, 167, 207, 241, 206, 29, 209, 212, 205, 255, 201, 3, 207, 234, 210, 21, 207, 39, 207, 140, 210, 224, 207, 57, 207, 75, 207, 235, 199, 93, 207, 229, 205, 243, 207, 212, 211, 13, 202, 163, 210, 115, 206, 60, 201, 246, 205, 111, 207, 122, 200, 6, 208, 173, 200, 133, 206, 44, 204, 241, 201, 3, 0, 85, 193, 93, 193, 39, 181, 96, 0, 202, 197, 211, 197, 29, 70, 220, 197, 0, 44, 34, 70, 29, 193, 69, 193, 53, 193, 63, 131, 137, 174, 126, 175, 140, 175, 161, 175, 229, 197, 133, 193, 107, 135, 238, 197, 37, 193, 247, 197, 1, 136, 157, 193, 97, 136, 175, 197, 125, 177, 27, 198, 231, 137, 251, 177, 79, 178, 168, 197, 77, 194, 166, 179, 72, 198, 101, 194, 173, 179, 81, 198, 15, 180, 189, 194, 130, 174, 197, 194, 213, 194, 90, 198, 99, 198, 166, 197, 221, 194, 108, 198, 57, 146, 126, 198, 113, 147, 193, 197, 245, 147, 13, 195, 21, 195, 135, 198, 235, 148, 144, 198, 45, 193, 181, 195, 193, 174, 149, 153, 162, 198, 208, 186, 215, 153, 181, 154, 113, 132, 171, 198, 93, 199, 85, 187, 180, 198, 200, 174, 5, 196, 13, 196, 103, 157, 135, 158, 201, 158, 225, 158, 189, 198, 137, 188, 193, 188, 93, 164, 25, 133, 207, 198, 216, 198, 237, 196, 245, 196, 63, 189, 51, 167, 143, 190, 157, 190, 171, 190, 141, 197, 234, 198, 243, 198, 252, 198, 5, 199, 184, 197, 1, 0, 72, 4, 184, 191, 2, 0, 28, 0, 7, 0, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 52, 0, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 202, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 4, 0, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 82, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 180, 44, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 140, 1, 1, 0, 72, 4, 184, 191, 2, 0, 214, 106, 70, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 155, 39, 2, 0, 179, 40, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 202, 106, 1, 0, 72, 4, 194, 40, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 61, 0, 1, 0, 72, 4, 192, 191, 2, 0, 72, 4, 90, 4, 0, 192, 7, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 184, 191, 26, 0, 67, 0, 103, 0, 28, 0, 10, 0, 49, 0, 112, 206, 232, 106, 96, 107, 34, 0, 76, 0, 152, 1, 79, 0, 236, 1, 82, 0, 68, 1, 64, 2, 240, 1, 166, 0, 88, 0, 72, 2, 18, 107, 220, 1, 58, 0, 180, 1, 140, 1, 64, 0, 2, 0, 52, 1, 56, 108, 3, 0, 175, 44, 104, 45, 134, 45, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 10, 0, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 68, 1, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 49, 0, 1, 0, 72, 4, 232, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 50, 192, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 2, 0, 72, 4, 90, 4, 184, 191, 156, 107, 3, 0, 200, 191, 192, 191, 184, 191, 3, 0, 34, 0, 52, 1, 40, 1, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 200, 191, 3, 0, 108, 1, 52, 1, 76, 106, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 185, 44, 4, 0, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 184, 191, 2, 0, 124, 106, 58, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 190, 44, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 195, 44, 4, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 25, 0, 2, 0, 72, 4, 90, 4, 192, 191, 216, 1, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 18, 107, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 48, 107, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 4, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 96, 192, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 120, 39, 1, 0, 72, 4, 208, 191, 2, 0, 72, 4, 90, 4, 192, 191, 22, 0, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 76, 106, 1, 0, 72, 4, 115, 39, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 90, 4, 64, 106, 1, 0, 72, 4, 208, 191, 3, 0, 46, 106, 0, 107, 188, 1, 2, 0, 72, 4, 90, 4, 184, 191, 42, 107, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 64, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 100, 106, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 4, 0, 1, 0, 90, 4, 44, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 44, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 209, 45, 2, 0, 184, 191, 104, 192, 1, 0, 90, 4, 40, 0, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 232, 191, 160, 1, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 120, 39, 1, 0, 90, 4, 48, 1, 1, 0, 72, 4, 232, 191, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 52, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 200, 191, 2, 0, 72, 4, 90, 4, 184, 191, 76, 106, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 2, 0, 192, 191, 184, 191, 1, 0, 90, 4, 44, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 192, 1, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 235, 39, 4, 0, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 90, 107, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 0, 192, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 176, 1, 1, 0, 90, 4, 50, 192, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 232, 191, 205, 44, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 64, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 82, 106, 2, 0, 72, 4, 90, 4, 184, 191, 115, 0, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 244, 106, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 90, 39, 1, 0, 72, 4, 215, 44, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 169, 40, 1, 0, 72, 4, 169, 40, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 200, 191, 2, 0, 34, 0, 52, 1, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 3, 0, 200, 191, 204, 40, 115, 39, 1, 0, 72, 4, 184, 191, 2, 0, 90, 39, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 90, 39, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 90, 39, 1, 0, 72, 4, 90, 39, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 90, 39, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 144, 107, 2, 0, 72, 4, 90, 4, 184, 191, 16, 0, 1, 0, 90, 4, 80, 1, 2, 0, 72, 4, 90, 4, 184, 191, 88, 1, 1, 0, 72, 4, 90, 39, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 88, 106, 2, 0, 72, 4, 90, 4, 184, 191, 190, 106, 1, 0, 90, 4, 58, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 46, 106, 6, 0, 124, 106, 130, 106, 160, 106, 58, 106, 184, 1, 166, 106, 12, 0, 210, 107, 16, 0, 190, 106, 118, 106, 88, 106, 31, 0, 37, 0, 112, 106, 88, 1, 142, 106, 80, 1, 133, 199, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 90, 39, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 128, 1, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 25, 0, 1, 0, 72, 4, 200, 191, 2, 0, 72, 4, 90, 4, 194, 45, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 219, 45, 133, 0, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 200, 191, 44, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 43, 0, 2, 0, 72, 4, 90, 4, 120, 39, 109, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 202, 106, 172, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 100, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 230, 44, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 174, 40, 2, 0, 184, 106, 144, 1, 89, 0, 46, 109, 52, 109, 97, 0, 58, 109, 64, 109, 46, 0, 218, 111, 70, 109, 14, 108, 98, 108, 178, 106, 76, 109, 224, 111, 90, 107, 230, 111, 236, 111, 20, 108, 232, 106, 238, 106, 242, 111, 244, 106, 88, 106, 82, 109, 248, 111, 202, 106, 88, 109, 26, 108, 115, 0, 32, 108, 144, 199, 250, 106, 94, 109, 46, 106, 254, 111, 100, 109, 102, 107, 94, 106, 38, 108, 106, 109, 4, 112, 112, 109, 118, 109, 44, 108, 108, 107, 85, 0, 136, 2, 10, 112, 208, 106, 0, 107, 114, 107, 214, 106, 16, 112, 100, 106, 120, 107, 50, 108, 126, 107, 132, 107, 64, 106, 76, 106, 22, 112, 166, 109, 138, 107, 12, 107, 124, 109, 144, 107, 18, 107, 30, 107, 28, 112, 36, 107, 56, 108, 130, 109, 62, 108, 34, 112, 136, 109, 68, 108, 74, 108, 80, 108, 148, 106, 82, 106, 36, 2, 133, 199, 86, 108, 142, 109, 207, 210, 42, 107, 150, 107, 70, 106, 48, 107, 54, 107, 2, 0, 176, 192, 112, 192, 49, 0, 76, 3, 67, 0, 92, 108, 124, 106, 156, 1, 160, 1, 130, 106, 22, 0, 210, 107, 160, 106, 104, 108, 184, 106, 16, 0, 58, 106, 190, 106, 10, 0, 92, 1, 49, 0, 50, 192, 7, 0, 196, 106, 118, 106, 144, 1, 96, 1, 31, 0, 96, 107, 34, 0, 104, 1, 76, 0, 37, 0, 236, 1, 148, 1, 124, 1, 166, 106, 132, 1, 68, 1, 128, 108, 6, 107, 112, 106, 88, 1, 172, 106, 72, 1, 24, 107, 142, 106, 80, 1, 112, 2, 76, 1, 19, 0, 64, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 23, 0, 92, 108, 156, 107, 104, 108, 148, 109, 110, 108, 154, 109, 116, 108, 154, 0, 162, 107, 50, 192, 66, 107, 122, 108, 160, 109, 100, 2, 128, 108, 134, 108, 140, 108, 146, 108, 166, 109, 168, 107, 152, 108, 172, 109, 158, 108, 9, 0, 176, 1, 98, 108, 58, 106, 118, 106, 88, 109, 64, 106, 172, 106, 72, 1, 133, 199, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 92, 1, 2, 0, 240, 39, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 50, 192, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 108, 1, 52, 1, 1, 0, 90, 4, 76, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 66, 107, 1, 0, 72, 4, 184, 191, 2, 0, 116, 1, 44, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 40, 0, 2, 0, 72, 4, 90, 4, 184, 191, 196, 106, 2, 0, 92, 1, 24, 107, 1, 0, 72, 4, 200, 191, 2, 0, 72, 4, 90, 4, 184, 191, 64, 106, 2, 0, 72, 4, 90, 4, 214, 45, 133, 199, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 238, 106, 2, 0, 72, 4, 90, 4, 184, 191, 80, 2, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 70, 0, 45, 0, 130, 106, 43, 0, 46, 0, 70, 0, 145, 0, 178, 106, 84, 1, 228, 1, 60, 2, 160, 106, 25, 0, 90, 107, 28, 0, 10, 0, 118, 106, 68, 2, 48, 2, 116, 1, 124, 2, 66, 107, 73, 0, 152, 1, 166, 106, 68, 1, 127, 0, 4, 2, 108, 107, 40, 0, 163, 0, 0, 107, 136, 0, 128, 1, 204, 1, 128, 2, 252, 1, 12, 107, 144, 107, 44, 1, 58, 0, 180, 1, 172, 1, 140, 1, 42, 107, 136, 1, 192, 1, 8, 0, 58, 106, 88, 106, 104, 1, 112, 106, 172, 106, 72, 1, 104, 2, 133, 199, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 136, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 128, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 100, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 18, 107, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 112, 106, 2, 0, 72, 4, 90, 4, 184, 191, 250, 106, 1, 0, 90, 4, 88, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 202, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 240, 191, 1, 0, 90, 4, 48, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 48, 2, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 88, 106, 1, 0, 72, 4, 0, 192, 2, 0, 72, 4, 90, 4, 184, 191, 38, 108, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 50, 192, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 52, 1, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 168, 107, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 2, 0, 208, 191, 40, 192, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 4, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 240, 44, 1, 0, 72, 4, 184, 191, 3, 0, 220, 44, 115, 39, 52, 72, 1, 0, 72, 4, 232, 191, 2, 0, 160, 106, 184, 1, 1, 0, 90, 4, 58, 106, 1, 0, 72, 4, 250, 44, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 94, 106, 2, 0, 72, 4, 90, 4, 184, 191, 94, 109, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 184, 191, 2, 0, 200, 191, 184, 191, 2, 0, 128, 1, 100, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 240, 191, 1, 0, 90, 4, 48, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 133, 199, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 108, 1, 2, 0, 72, 4, 90, 4, 16, 192, 96, 107, 2, 0, 72, 4, 90, 4, 208, 191, 46, 106, 2, 0, 238, 106, 70, 106, 2, 0, 208, 191, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 64, 106, 1, 0, 72, 4, 194, 40, 1, 0, 72, 4, 159, 45, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 9, 45, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 64, 192, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 4, 45, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 76, 106, 2, 0, 72, 4, 90, 4, 184, 191, 132, 107, 1, 0, 72, 4, 208, 191, 2, 0, 72, 4, 90, 4, 234, 45, 46, 106, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 2, 0, 156, 1, 96, 1, 3, 0, 124, 106, 64, 1, 142, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 240, 191, 1, 0, 90, 4, 48, 1, 2, 0, 72, 4, 90, 4, 184, 191, 73, 0, 2, 0, 72, 4, 90, 4, 184, 191, 31, 0, 3, 0, 124, 106, 72, 1, 64, 1, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 120, 192, 52, 2, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 100, 106, 2, 0, 72, 4, 90, 4, 184, 191, 4, 0, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 100, 106, 1, 0, 72, 4, 244, 45, 2, 0, 72, 4, 90, 4, 184, 191, 48, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 178, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 64, 106, 1, 0, 72, 4, 192, 191, 2, 0, 200, 191, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 118, 0, 4, 0, 160, 106, 104, 1, 76, 2, 224, 1, 5, 0, 58, 106, 118, 106, 96, 1, 208, 106, 133, 199, 2, 0, 90, 39, 184, 191, 1, 0, 90, 4, 178, 106, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 19, 45, 121, 0, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 102, 107, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 16, 192, 1, 0, 72, 4, 16, 192, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 14, 45, 4, 0, 2, 0, 72, 4, 90, 4, 56, 192, 52, 0, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 10, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 94, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 148, 106, 1, 0, 72, 4, 110, 39, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 200, 108, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 84, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 232, 191, 46, 106, 1, 0, 72, 4, 204, 45, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 24, 192, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 94, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 214, 106, 1, 0, 72, 4, 64, 192, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 2, 0, 200, 191, 192, 191, 1, 0, 90, 4, 48, 1, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 160, 0, 1, 0, 90, 4, 190, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 94, 106, 2, 0, 72, 4, 90, 4, 128, 192, 124, 0, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 60, 107, 1, 0, 72, 4, 80, 192, 1, 0, 90, 4, 46, 106, 1, 0, 90, 4, 44, 1, 2, 0, 168, 1, 244, 1, 2, 0, 48, 1, 133, 199, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 36, 107, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 2, 0, 200, 191, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 3, 0, 200, 191, 192, 191, 184, 191, 2, 0, 58, 0, 19, 0, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 2, 0, 72, 4, 90, 4, 139, 45, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 110, 39, 1, 0, 90, 4, 76, 1, 1, 0, 72, 4, 200, 191, 3, 0, 200, 191, 192, 191, 184, 191, 1, 0, 72, 4, 174, 40, 2, 0, 72, 4, 90, 4, 184, 191, 224, 1, 1, 0, 72, 4, 47, 72, 1, 0, 72, 4, 16, 192, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 76, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 82, 106, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 166, 106, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 110, 39, 1, 0, 72, 4, 34, 45, 1, 0, 72, 4, 192, 191, 2, 0, 72, 4, 90, 4, 170, 44, 40, 1, 2, 0, 192, 191, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 82, 106, 1, 0, 72, 4, 184, 191, 2, 0, 208, 191, 39, 45, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 44, 45, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 4, 0, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 40, 192, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 84, 1, 2, 0, 72, 4, 90, 4, 64, 192, 28, 2, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 52, 1, 2, 0, 72, 4, 90, 4, 184, 191, 124, 1, 1, 0, 90, 4, 88, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 84, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 84, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 94, 106, 1, 0, 72, 4, 232, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 94, 106, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 52, 1, 2, 0, 72, 4, 90, 4, 184, 191, 132, 1, 1, 0, 72, 4, 110, 39, 1, 0, 72, 4, 240, 191, 2, 0, 56, 192, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 4, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 208, 191, 2, 0, 46, 106, 188, 1, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 235, 39, 4, 0, 2, 0, 184, 191, 104, 192, 2, 0, 26, 108, 116, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 64, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 235, 39, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 74, 45, 36, 107, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 94, 106, 1, 0, 90, 4, 44, 1, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 76, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 0, 107, 1, 0, 90, 4, 44, 1, 2, 0, 72, 4, 90, 4, 184, 191, 108, 107, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 136, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 85, 0, 2, 0, 72, 4, 90, 4, 184, 191, 126, 107, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 4, 0, 2, 0, 72, 4, 90, 4, 192, 191, 0, 2, 1, 0, 72, 4, 136, 192, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 192, 191, 56, 2, 2, 0, 168, 192, 184, 40, 1, 0, 72, 4, 168, 192, 1, 0, 72, 4, 232, 191, 2, 0, 72, 4, 90, 4, 94, 45, 46, 106, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 24, 192, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 43, 0, 1, 0, 72, 4, 208, 191, 2, 0, 72, 4, 90, 4, 84, 45, 46, 106, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 89, 45, 2, 0, 200, 191, 184, 191, 3, 0, 44, 2, 6, 107, 30, 107, 2, 0, 76, 109, 4, 0, 2, 0, 72, 4, 90, 4, 184, 191, 208, 106, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 24, 192, 248, 1, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 82, 106, 1, 0, 72, 4, 64, 45, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 52, 1, 2, 0, 72, 4, 90, 4, 184, 191, 120, 107, 1, 0, 72, 4, 240, 191, 1, 0, 90, 4, 48, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 55, 0, 40, 2, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 54, 107, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 136, 106, 1, 0, 72, 4, 184, 191, 2, 0, 58, 106, 112, 106, 2, 0, 72, 4, 90, 4, 208, 191, 188, 1, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 214, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 136, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 50, 108, 1, 0, 72, 4, 184, 191, 7, 0, 100, 0, 130, 106, 106, 0, 112, 0, 112, 106, 232, 1, 130, 109, 1, 0, 90, 4, 58, 106, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 55, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 29, 45, 1, 0, 90, 4, 76, 1, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 215, 39, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 104, 192, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 70, 106, 1, 0, 72, 4, 90, 39, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 70, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 114, 107, 1, 0, 90, 4, 48, 107, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 148, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 82, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 118, 106, 1, 0, 72, 4, 90, 39, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 0, 192, 7, 0, 1, 0, 72, 4, 88, 192, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 52, 0, 1, 0, 72, 4, 124, 45, 1, 0, 72, 4, 184, 191, 2, 0, 240, 39, 184, 191, 1, 0, 72, 4, 119, 45, 2, 0, 200, 191, 144, 192, 2, 0, 46, 106, 76, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 64, 106, 2, 0, 144, 45, 149, 45, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 64, 106, 2, 0, 100, 0, 112, 0, 1, 0, 90, 4, 106, 0, 2, 0, 72, 4, 90, 4, 184, 191, 138, 107, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 100, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 239, 45, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 154, 45, 3, 0, 208, 191, 72, 192, 110, 39, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 88, 1, 1, 0, 90, 4, 120, 1, 1, 0, 72, 4, 184, 191, 3, 0, 192, 191, 115, 39, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 244, 106, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 200, 191, 108, 1, 1, 0, 90, 4, 76, 106, 10, 0, 60, 107, 210, 107, 200, 108, 66, 107, 206, 108, 226, 109, 136, 106, 172, 106, 232, 109, 212, 108, 1, 0, 72, 4, 184, 191, 3, 0, 156, 107, 162, 107, 146, 108, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 122, 108, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 252, 1, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 209, 40, 4, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 184, 191, 130, 39, 1, 0, 90, 4, 68, 1, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 232, 191, 2, 0, 160, 1, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 160, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 52, 1, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 46, 0, 1, 0, 72, 4, 184, 191, 2, 0, 84, 2, 72, 1, 2, 0, 72, 4, 90, 4, 192, 191, 64, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 5, 0, 216, 1, 208, 1, 37, 0, 40, 1, 120, 1, 9, 0, 22, 0, 58, 106, 92, 1, 124, 1, 132, 1, 88, 1, 40, 2, 91, 0, 94, 0, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 12, 107, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 179, 40, 5, 0, 208, 191, 245, 44, 174, 45, 179, 45, 72, 192, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 2, 0, 16, 192, 189, 40, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 4, 0, 1, 0, 72, 4, 184, 191, 4, 0, 232, 191, 208, 191, 114, 45, 88, 192, 1, 0, 90, 4, 44, 1, 1, 0, 72, 4, 169, 45, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 104, 1, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 88, 106, 1, 0, 72, 4, 184, 191, 2, 0, 200, 191, 208, 191, 2, 0, 192, 192, 200, 192, 2, 0, 46, 106, 76, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 52, 1, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 192, 191, 55, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 25, 0, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 136, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 228, 1, 2, 0, 184, 191, 109, 45, 1, 0, 72, 4, 184, 191, 3, 0, 200, 191, 184, 191, 130, 39, 1, 0, 90, 4, 68, 1, 1, 0, 72, 4, 24, 192, 1, 0, 90, 4, 248, 1, 2, 0, 72, 4, 90, 4, 72, 192, 148, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 7, 0, 2, 0, 72, 4, 90, 4, 184, 191, 80, 1, 2, 0, 16, 0, 64, 1, 1, 0, 72, 4, 200, 191, 1, 0, 90, 4, 76, 106, 2, 0, 72, 4, 90, 4, 184, 191, 142, 106, 1, 0, 90, 4, 124, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 97, 0, 150, 107, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 189, 40, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 54, 107, 1, 0, 90, 4, 44, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 40, 0, 2, 0, 72, 4, 90, 4, 184, 191, 220, 1, 2, 0, 200, 191, 129, 45, 2, 0, 28, 0, 7, 0, 1, 0, 72, 4, 110, 39, 1, 0, 90, 4, 76, 1, 2, 0, 88, 2, 61, 0, 2, 0, 0, 2, 64, 1, 2, 0, 192, 191, 184, 191, 4, 0, 156, 1, 130, 0, 64, 1, 61, 0, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 214, 40, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 232, 1, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 44, 1, 1, 0, 72, 4, 208, 191, 2, 0, 72, 4, 90, 4, 184, 191, 62, 108, 1, 0, 90, 4, 70, 106, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 44, 108, 1, 0, 90, 4, 70, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 45, 4, 0, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 148, 106, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 200, 191, 116, 1, 3, 0, 142, 0, 196, 106, 24, 107, 3, 0, 94, 106, 148, 106, 82, 106, 1, 0, 90, 4, 84, 1, 1, 0, 72, 4, 200, 191, 2, 0, 72, 4, 90, 4, 184, 191, 192, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 232, 191, 2, 0, 72, 4, 90, 4, 250, 39, 160, 1, 1, 0, 72, 4, 155, 39, 1, 0, 72, 4, 184, 191, 2, 0, 148, 1, 30, 107, 2, 0, 46, 106, 6, 107, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 120, 39, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 189, 45, 1, 0, 72, 4, 80, 192, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 199, 45, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 120, 39, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 40, 192, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 36, 107, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 82, 106, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 212, 108, 3, 0, 200, 191, 192, 191, 184, 191, 2, 0, 72, 4, 90, 4, 110, 39, 76, 1, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 208, 191, 2, 0, 72, 4, 90, 4, 0, 192, 232, 106, 1, 0, 90, 4, 7, 0, 1, 0, 72, 4, 0, 192, 1, 0, 90, 4, 232, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 100, 106, 3, 0, 200, 191, 192, 191, 184, 191, 2, 0, 108, 1, 52, 1, 2, 0, 72, 4, 90, 4, 184, 191, 136, 109, 2, 0, 192, 191, 184, 191, 1, 0, 72, 4, 208, 191, 1, 0, 72, 4, 184, 191, 3, 0, 200, 191, 192, 191, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 128, 1, 2, 0, 72, 4, 90, 4, 184, 191, 14, 108, 1, 0, 90, 4, 70, 106, 1, 0, 90, 4, 44, 1, 2, 0, 72, 4, 90, 4, 184, 191, 68, 108, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 132, 107, 2, 0, 184, 106, 19, 0, 2, 0, 72, 4, 90, 4, 200, 191, 58, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 155, 39, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 136, 106, 70, 106, 1, 0, 72, 4, 115, 39, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 64, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 120, 39, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 110, 39, 1, 0, 90, 4, 76, 1, 2, 0, 192, 191, 184, 191, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 114, 107, 2, 0, 72, 4, 90, 4, 184, 191, 80, 108, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 204, 1, 1, 0, 72, 4, 192, 191, 1, 0, 90, 4, 40, 1, 2, 0, 184, 191, 130, 39, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 192, 191, 2, 0, 72, 4, 90, 4, 184, 191, 40, 1, 2, 0, 200, 191, 192, 191, 2, 0, 72, 4, 90, 4, 184, 191, 48, 1, 1, 0, 72, 4, 229, 45, 2, 0, 72, 4, 90, 4, 192, 191, 120, 1, 1, 0, 90, 4, 64, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 60, 107, 5, 0, 139, 0, 151, 0, 181, 0, 140, 2, 116, 2, 2, 0, 232, 191, 208, 191, 1, 0, 90, 4, 46, 106, 2, 0, 232, 191, 208, 191, 2, 0, 72, 4, 90, 4, 200, 191, 76, 106, 1, 0, 90, 4, 46, 106, 3, 0, 200, 191, 192, 191, 184, 191, 1, 0, 90, 4, 108, 1, 2, 0, 184, 191, 224, 40, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 70, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 184, 191, 196, 1, 2, 0, 72, 4, 90, 4, 255, 44, 133, 199, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 130, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 58, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 136, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 0, 192, 1, 0, 90, 4, 7, 0, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 64, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 40, 192, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 136, 1, 2, 0, 72, 4, 90, 4, 200, 191, 152, 1, 2, 0, 72, 4, 90, 4, 184, 191, 44, 1, 1, 0, 72, 4, 200, 191, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 240, 191, 1, 0, 90, 4, 48, 1, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 210, 44, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 70, 106, 1, 0, 72, 4, 49, 45, 1, 0, 72, 4, 54, 45, 1, 0, 72, 4, 56, 192, 1, 0, 72, 4, 69, 45, 1, 0, 72, 4, 79, 45, 1, 0, 72, 4, 99, 45, 1, 0, 72, 4, 208, 191, 1, 0, 90, 4, 46, 106, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 82, 106, 1, 0, 72, 4, 164, 45, 1, 0, 72, 4, 79, 40, 1, 0, 72, 4, 208, 191, 2, 0, 240, 39, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 16, 192, 2, 0, 72, 4, 90, 4, 184, 191, 100, 106, 1, 0, 72, 4, 192, 191, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 240, 191, 32, 192, 2, 0, 48, 1, 144, 199, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 240, 191, 48, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 2, 0, 72, 4, 90, 4, 200, 191, 152, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 130, 39, 1, 0, 72, 4, 130, 39, 1, 0, 90, 4, 68, 1, 2, 0, 240, 191, 32, 192, 5, 0, 48, 1, 144, 199, 226, 109, 30, 107, 175, 0, 2, 0, 199, 40, 184, 40, 5, 0, 4, 0, 6, 107, 76, 1, 133, 199, 196, 1, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 70, 106, 1, 0, 72, 4, 184, 191, 1, 0, 72, 4, 184, 191, 1, 0, 90, 4, 19, 0, 4, 0, 152, 39, 142, 39, 47, 44, 37, 1, 4, 0, 162, 39, 167, 39, 102, 39, 37, 1, 4, 0, 147, 39, 167, 39, 102, 39, 37, 1, 4, 0, 147, 39, 162, 39, 102, 39, 37, 1, 4, 0, 147, 39, 187, 39, 102, 39, 37, 1, 4, 0, 83, 9, 180, 2, 102, 39, 37, 1, 4, 0, 51, 3, 127, 39, 102, 39, 100, 174, 4, 0, 48, 43, 127, 39, 102, 39, 100, 174, 4, 0, 164, 1, 152, 39, 247, 39, 37, 1, 4, 0, 164, 1, 142, 39, 247, 39, 37, 1, 4, 0, 218, 108, 87, 39, 67, 39, 100, 174, 4, 0, 115, 40, 32, 1, 67, 39, 100, 174, 4, 0, 31, 40, 177, 39, 67, 39, 100, 174, 4, 0, 252, 107, 32, 1, 67, 39, 100, 174, 4, 0, 98, 175, 149, 192, 67, 39, 100, 174, 4, 0, 200, 2, 107, 39, 67, 39, 100, 174, 4, 0, 204, 107, 87, 39, 67, 39, 100, 174, 4, 0, 205, 192, 245, 191, 67, 39, 100, 174, 4, 0, 0, 3, 87, 39, 67, 39, 100, 174, 4, 0, 197, 107, 32, 1, 67, 39, 100, 174, 4, 0, 111, 18, 112, 1, 67, 39, 100, 174, 4, 0, 74, 42, 32, 1, 67, 39, 100, 174, 4, 0, 229, 191, 213, 191, 67, 39, 100, 174, 4, 0, 16, 40, 82, 39, 67, 39, 100, 174, 4, 0, 21, 109, 43, 3, 67, 39, 100, 174, 4, 0, 135, 7, 32, 1, 67, 39, 100, 174, 4, 0, 39, 3, 52, 106, 67, 39, 100, 174, 4, 0, 51, 40, 82, 39, 67, 39, 100, 174, 4, 0, 159, 7, 82, 39, 67, 39, 100, 174, 4, 0, 21, 40, 32, 1, 67, 39, 100, 174, 4, 0, 224, 2, 32, 1, 67, 39, 100, 174, 4, 0, 184, 2, 32, 1, 67, 39, 100, 174, 4, 0, 186, 107, 168, 2, 67, 39, 100, 174, 4, 0, 188, 2, 107, 39, 67, 39, 100, 174, 4, 0, 224, 108, 164, 1, 67, 39, 100, 174, 4, 0, 156, 2, 177, 39, 67, 39, 100, 174, 4, 0, 1, 40, 32, 1, 67, 39, 100, 174, 4, 0, 235, 7, 32, 1, 67, 39, 100, 174, 4, 0, 208, 2, 32, 1, 67, 39, 100, 174, 4, 0, 164, 108, 107, 39, 67, 39, 100, 174, 4, 0, 154, 106, 107, 39, 67, 39, 100, 174, 4, 0, 228, 2, 112, 1, 67, 39, 100, 174, 4, 0, 182, 108, 32, 1, 67, 39, 100, 174, 4, 0, 113, 192, 32, 1, 67, 39, 100, 174, 4, 0, 8, 108, 82, 39, 67, 39, 100, 174, 4, 0, 176, 108, 87, 39, 67, 39, 100, 174, 4, 0, 117, 192, 245, 191, 67, 39, 100, 174, 4, 0, 232, 2, 32, 1, 67, 39, 100, 174, 4, 0, 2, 108, 32, 1, 67, 39, 100, 174, 4, 0, 172, 2, 32, 1, 67, 39, 100, 174, 4, 0, 152, 2, 137, 39, 67, 39, 100, 174, 4, 0, 53, 192, 213, 191, 67, 39, 100, 174, 4, 0, 100, 41, 32, 1, 67, 39, 100, 174, 4, 0, 236, 2, 82, 39, 67, 39, 100, 174, 4, 0, 4, 3, 87, 39, 67, 39, 100, 174, 4, 0, 41, 40, 32, 1, 67, 39, 100, 174, 4, 0, 125, 192, 213, 191, 67, 39, 100, 174, 4, 0, 36, 40, 32, 1, 67, 39, 100, 174, 4, 0, 77, 192, 213, 191, 67, 39, 100, 174, 4, 0, 127, 8, 32, 1, 67, 39, 100, 174, 4, 0, 192, 39, 176, 2, 67, 39, 100, 174, 4, 0, 240, 2, 87, 39, 67, 39, 100, 174, 4, 0, 6, 40, 192, 2, 67, 39, 100, 174, 4, 0, 244, 2, 32, 1, 67, 39, 100, 174, 4, 0, 170, 108, 82, 39, 67, 39, 100, 174, 4, 0, 252, 39, 32, 1, 67, 39, 100, 174, 4, 0, 61, 192, 213, 191, 67, 39, 100, 174, 4, 0, 167, 8, 32, 1, 67, 39, 100, 174, 4, 0, 83, 107, 82, 39, 67, 39, 100, 174, 4, 0, 187, 8, 32, 1, 67, 39, 100, 174, 4, 0, 97, 39, 82, 39, 67, 39, 100, 174, 4, 0, 182, 39, 32, 1, 67, 39, 100, 174, 4, 0, 188, 108, 204, 2, 67, 39, 100, 174, 4, 0, 164, 2, 32, 1, 67, 39, 100, 174, 4, 0, 133, 192, 213, 191, 67, 39, 100, 174, 4, 0, 232, 39, 87, 39, 67, 39, 100, 174, 4, 0, 26, 192, 112, 1, 67, 39, 100, 174, 4, 0, 77, 39, 137, 39, 67, 39, 100, 174, 4, 0, 248, 2, 32, 1, 67, 39, 100, 174, 4, 0, 197, 39, 32, 1, 67, 39, 100, 174, 4, 0, 21, 192, 213, 191, 67, 39, 100, 174, 4, 0, 227, 39, 32, 1, 67, 39, 100, 174, 4, 0, 205, 191, 213, 191, 67, 39, 100, 174, 4, 0, 72, 107, 142, 39, 67, 39, 37, 1, 4, 0, 72, 107, 152, 39, 67, 39, 100, 174, 4, 0, 251, 8, 32, 1, 67, 39, 100, 174, 4, 0, 196, 2, 32, 1, 67, 39, 100, 174, 4, 0, 85, 40, 32, 1, 67, 39, 100, 174, 4, 0, 8, 3, 82, 39, 67, 39, 100, 174, 4, 0, 200, 1, 32, 1, 67, 39, 100, 174, 4, 0, 180, 107, 32, 1, 67, 39, 100, 174, 4, 0, 85, 192, 213, 191, 67, 39, 100, 174, 4, 0, 105, 41, 32, 1, 67, 39, 100, 174, 4, 0, 141, 192, 213, 191, 67, 39, 100, 174, 4, 0, 11, 9, 32, 1, 67, 39, 100, 174, 4, 0, 172, 39, 32, 1, 67, 39, 100, 174, 4, 0, 221, 192, 213, 191, 67, 39, 100, 174, 4, 0, 72, 39, 107, 39, 67, 39, 100, 174, 4, 0, 46, 40, 127, 39, 67, 39, 100, 174, 4, 0, 213, 43, 32, 1, 67, 39, 100, 174, 4, 0, 252, 2, 82, 39, 67, 39, 100, 174, 4, 0, 207, 39, 112, 1, 67, 39, 100, 174, 4, 0, 26, 40, 32, 1, 67, 39, 100, 174, 4, 0, 245, 192, 213, 191, 67, 39, 100, 174, 4, 0, 66, 40, 32, 1, 67, 39, 100, 174, 4, 0, 69, 192, 213, 191, 67, 39, 100, 174, 4, 0, 255, 41, 32, 1, 67, 39, 100, 174, 4, 0, 56, 40, 32, 1, 67, 39, 100, 174, 4, 0, 226, 106, 32, 1, 67, 39, 100, 174, 4, 0, 110, 41, 32, 1, 67, 39, 100, 174, 4, 0, 212, 39, 137, 39, 67, 39, 100, 174, 4, 0, 11, 40, 32, 1, 67, 39, 100, 174, 4, 0, 202, 39, 32, 1, 67, 39, 100, 174, 4, 0, 93, 192, 213, 191, 67, 39, 100, 174, 4, 0, 12, 3, 32, 1, 67, 39, 100, 174, 4, 0, 37, 192, 213, 191, 67, 39, 100, 174, 4, 0, 16, 3, 87, 39, 67, 39, 100, 174, 4, 0, 194, 108, 32, 1, 67, 39, 100, 174, 4, 0, 253, 191, 213, 191, 67, 39, 100, 174, 4, 0, 71, 40, 87, 39, 67, 39, 100, 174, 4, 0, 5, 193, 245, 191, 67, 39, 100, 174, 4, 0, 115, 9, 32, 1, 67, 39, 100, 174, 4, 0, 94, 42, 32, 1, 67, 39, 100, 174, 4, 0, 76, 40, 32, 1, 67, 39, 100, 174, 4, 0, 61, 40, 87, 39, 67, 39, 100, 174, 4, 0, 9, 109, 32, 1, 67, 39, 100, 174, 4, 0, 212, 2, 60, 1, 67, 39, 100, 174, 4, 0, 216, 2, 32, 1, 67, 39, 100, 174, 4, 0, 13, 193, 213, 191, 67, 39, 100, 174, 4, 0, 192, 107, 87, 39, 67, 39, 100, 174, 4, 0, 181, 191, 245, 191, 67, 39, 100, 174, 4, 0, 220, 2, 82, 39, 67, 39, 100, 174, 4, 0, 24, 2, 32, 1, 67, 39, 100, 174, 4, 0, 92, 2, 32, 1, 67, 39, 100, 174, 4, 0, 13, 192, 213, 191, 67, 39, 100, 174, 4, 0, 221, 40, 32, 1, 67, 39, 100, 174, 4, 0, 160, 2, 32, 1, 67, 39, 100, 174, 4, 0, 157, 192, 253, 192, 150, 41, 37, 1, 4, 0, 237, 191, 29, 192, 251, 70, 100, 174, 4, 0, 29, 192, 237, 191, 30, 71, 100, 174, 4, 0, 136, 199, 147, 199, 58, 193, 37, 1, 4, 0, 229, 192, 237, 192, 102, 39, 37, 1, 4, 0, 237, 192, 109, 174, 247, 39, 37, 1, 4, 0, 229, 192, 109, 174, 247, 39, 37, 1, 4, 0, 109, 174, 109, 174, 102, 39, 37, 1, 4, 0, 85, 194, 116, 174, 0, 71, 37, 1, 4, 0, 149, 192, 116, 174, 102, 39, 37, 1, 4, 0, 45, 198, 116, 174, 102, 39, 37, 1, 4, 0, 116, 174, 116, 174, 150, 41, 37, 1, 4, 0, 45, 192, 123, 174, 247, 39, 37, 1, 4, 0, 54, 198, 123, 174, 102, 39, 37, 1, 4, 0, 123, 174, 123, 174, 150, 41, 37, 1, 4, 0, 71, 199, 71, 199, 50, 71, 37, 1, 4, 0, 112, 174, 112, 174, 167, 44, 37, 1, 4, 0, 75, 107, 75, 107, 102, 39, 37, 1, 4, 0, 77, 14, 86, 14, 92, 14, 97, 14, 177, 112, 97, 133, 186, 210, 97, 206, 6, 0, 32, 1, 85, 106, 107, 39, 40, 179, 207, 107, 169, 106, 4, 0, 152, 39, 142, 39, 100, 174, 37, 1, 4, 0, 162, 39, 167, 39, 107, 106, 37, 1, 4, 0, 147, 39, 167, 39, 107, 106, 37, 1, 4, 0, 147, 39, 162, 39, 107, 106, 37, 1, 4, 0, 147, 39, 187, 39, 107, 106, 37, 1, 4, 0, 83, 9, 180, 2, 107, 106, 37, 1, 4, 0, 51, 3, 127, 39, 107, 106, 100, 174, 4, 0, 48, 43, 127, 39, 107, 106, 100, 174, 4, 0, 164, 1, 152, 39, 163, 174, 37, 1, 4, 0, 164, 1, 142, 39, 163, 174, 37, 1, 4, 0, 218, 108, 87, 39, 57, 39, 100, 174, 4, 0, 115, 40, 32, 1, 57, 39, 100, 174, 4, 0, 31, 40, 177, 39, 57, 39, 100, 174, 4, 0, 252, 107, 32, 1, 57, 39, 100, 174, 4, 0, 200, 2, 107, 39, 57, 39, 100, 174, 4, 0, 204, 107, 87, 39, 57, 39, 100, 174, 4, 0, 0, 3, 87, 39, 57, 39, 100, 174, 4, 0, 197, 107, 32, 1, 57, 39, 100, 174, 4, 0, 220, 109, 112, 1, 57, 39, 100, 174, 4, 0, 74, 42, 32, 1, 57, 39, 100, 174, 4, 0, 16, 40, 82, 39, 57, 39, 100, 174, 4, 0, 21, 109, 43, 3, 57, 39, 100, 174, 4, 0, 135, 7, 32, 1, 57, 39, 100, 174, 4, 0, 39, 3, 52, 106, 57, 39, 100, 174, 4, 0, 51, 40, 82, 39, 57, 39, 100, 174, 4, 0, 159, 7, 82, 39, 57, 39, 100, 174, 4, 0, 21, 40, 32, 1, 57, 39, 100, 174, 4, 0, 224, 2, 32, 1, 57, 39, 100, 174, 4, 0, 184, 2, 32, 1, 57, 39, 100, 174, 4, 0, 186, 107, 168, 2, 57, 39, 100, 174, 4, 0, 188, 2, 107, 39, 57, 39, 100, 174, 4, 0, 224, 108, 164, 1, 57, 39, 100, 174, 4, 0, 156, 2, 177, 39, 57, 39, 100, 174, 4, 0, 1, 40, 32, 1, 57, 39, 100, 174, 4, 0, 235, 7, 32, 1, 57, 39, 100, 174, 4, 0, 208, 2, 32, 1, 57, 39, 100, 174, 4, 0, 164, 108, 107, 39, 57, 39, 100, 174, 4, 0, 154, 106, 107, 39, 57, 39, 100, 174, 4, 0, 228, 2, 112, 1, 57, 39, 100, 174, 4, 0, 182, 108, 32, 1, 57, 39, 100, 174, 4, 0, 113, 192, 32, 1, 57, 39, 100, 174, 4, 0, 8, 108, 82, 39, 57, 39, 100, 174, 4, 0, 176, 108, 87, 39, 57, 39, 100, 174, 4, 0, 232, 2, 32, 1, 57, 39, 100, 174, 4, 0, 2, 108, 32, 1, 57, 39, 100, 174, 4, 0, 172, 2, 32, 1, 57, 39, 100, 174, 4, 0, 152, 2, 137, 39, 57, 39, 100, 174, 4, 0, 100, 41, 32, 1, 57, 39, 100, 174, 4, 0, 236, 2, 82, 39, 57, 39, 100, 174, 4, 0, 4, 3, 87, 39, 57, 39, 100, 174, 4, 0, 41, 40, 32, 1, 57, 39, 100, 174, 4, 0, 36, 40, 32, 1, 57, 39, 100, 174, 4, 0, 127, 8, 32, 1, 57, 39, 100, 174, 4, 0, 192, 39, 176, 2, 57, 39, 100, 174, 4, 0, 240, 2, 87, 39, 57, 39, 100, 174, 4, 0, 6, 40, 192, 2, 57, 39, 100, 174, 4, 0, 244, 2, 32, 1, 57, 39, 100, 174, 4, 0, 170, 108, 82, 39, 57, 39, 100, 174, 4, 0, 252, 39, 32, 1, 57, 39, 100, 174, 4, 0, 167, 8, 32, 1, 57, 39, 100, 174, 4, 0, 83, 107, 82, 39, 57, 39, 100, 174, 4, 0, 187, 8, 32, 1, 57, 39, 100, 174, 4, 0, 97, 39, 82, 39, 57, 39, 100, 174, 4, 0, 182, 39, 32, 1, 57, 39, 100, 174, 4, 0, 188, 108, 204, 2, 57, 39, 100, 174, 4, 0, 164, 2, 32, 1, 57, 39, 100, 174, 4, 0, 232, 39, 87, 39, 57, 39, 100, 174, 4, 0, 26, 192, 112, 1, 57, 39, 100, 174, 4, 0, 77, 39, 137, 39, 57, 39, 100, 174, 4, 0, 248, 2, 32, 1, 57, 39, 100, 174, 4, 0, 197, 39, 32, 1, 57, 39, 100, 174, 4, 0, 227, 39, 32, 1, 57, 39, 100, 174, 4, 0, 72, 107, 142, 39, 57, 39, 37, 1, 4, 0, 72, 107, 152, 39, 57, 39, 100, 174, 4, 0, 251, 8, 32, 1, 57, 39, 100, 174, 4, 0, 196, 2, 32, 1, 57, 39, 100, 174, 4, 0, 85, 40, 32, 1, 57, 39, 100, 174, 4, 0, 8, 3, 82, 39, 57, 39, 100, 174, 4, 0, 200, 1, 32, 1, 57, 39, 100, 174, 4, 0, 180, 107, 32, 1, 57, 39, 100, 174, 4, 0, 105, 41, 32, 1, 57, 39, 100, 174, 4, 0, 11, 9, 32, 1, 57, 39, 100, 174, 4, 0, 172, 39, 32, 1, 57, 39, 100, 174, 4, 0, 72, 39, 107, 39, 57, 39, 100, 174, 4, 0, 46, 40, 127, 39, 57, 39, 100, 174, 4, 0, 213, 43, 32, 1, 57, 39, 100, 174, 4, 0, 252, 2, 82, 39, 57, 39, 100, 174, 4, 0, 207, 39, 112, 1, 57, 39, 100, 174, 4, 0, 26, 40, 32, 1, 57, 39, 100, 174, 4, 0, 66, 40, 32, 1, 57, 39, 100, 174, 4, 0, 255, 41, 32, 1, 57, 39, 100, 174, 4, 0, 56, 40, 32, 1, 57, 39, 100, 174, 4, 0, 226, 106, 32, 1, 57, 39, 100, 174, 4, 0, 110, 41, 32, 1, 57, 39, 100, 174, 4, 0, 212, 39, 137, 39, 57, 39, 100, 174, 4, 0, 11, 40, 32, 1, 57, 39, 100, 174, 4, 0, 202, 39, 32, 1, 57, 39, 100, 174, 4, 0, 12, 3, 32, 1, 57, 39, 100, 174, 4, 0, 16, 3, 87, 39, 57, 39, 100, 174, 4, 0, 194, 108, 32, 1, 57, 39, 100, 174, 4, 0, 71, 40, 87, 39, 57, 39, 100, 174, 4, 0, 115, 9, 32, 1, 57, 39, 100, 174, 4, 0, 94, 42, 32, 1, 57, 39, 100, 174, 4, 0, 76, 40, 32, 1, 57, 39, 100, 174, 4, 0, 61, 40, 87, 39, 57, 39, 100, 174, 4, 0, 9, 109, 32, 1, 57, 39, 100, 174, 4, 0, 212, 2, 60, 1, 57, 39, 100, 174, 4, 0, 216, 2, 32, 1, 57, 39, 100, 174, 4, 0, 192, 107, 87, 39, 57, 39, 100, 174, 4, 0, 220, 2, 82, 39, 57, 39, 100, 174, 4, 0, 24, 2, 32, 1, 57, 39, 100, 174, 4, 0, 92, 2, 32, 1, 57, 39, 100, 174, 4, 0, 221, 40, 32, 1, 57, 39, 100, 174, 4, 0, 160, 2, 32, 1, 57, 39, 100, 174, 4, 0, 251, 16, 52, 106, 57, 39, 100, 174, 4, 0, 3, 17, 52, 106, 57, 39, 100, 174, 4, 0, 7, 17, 52, 106, 57, 39, 100, 174, 4, 0, 11, 17, 52, 106, 57, 39, 100, 174, 4, 0, 15, 17, 52, 106, 57, 39, 100, 174, 4, 0, 19, 17, 52, 106, 57, 39, 100, 174, 4, 0, 23, 17, 52, 106, 57, 39, 100, 174, 4, 0, 27, 17, 52, 106, 57, 39, 100, 174, 4, 0, 31, 17, 52, 106, 57, 39, 100, 174, 4, 0, 43, 17, 52, 106, 57, 39, 100, 174, 4, 0, 247, 43, 52, 106, 57, 39, 100, 174, 4, 0, 55, 17, 52, 106, 57, 39, 100, 174, 4, 0, 103, 7, 52, 106, 57, 39, 100, 174, 4, 0, 127, 17, 52, 106, 57, 39, 100, 174, 4, 0, 131, 17, 52, 106, 57, 39, 100, 174, 4, 0, 107, 7, 52, 106, 57, 39, 100, 174, 4, 0, 111, 7, 52, 106, 57, 39, 100, 174, 4, 0, 115, 7, 52, 106, 57, 39, 100, 174, 4, 0, 119, 7, 52, 106, 57, 39, 100, 174, 4, 0, 247, 17, 52, 106, 57, 39, 100, 174, 4, 0, 3, 18, 52, 106, 57, 39, 100, 174, 4, 0, 23, 18, 52, 106, 57, 39, 100, 174, 4, 0, 27, 18, 52, 106, 57, 39, 100, 174, 4, 0, 31, 18, 52, 106, 57, 39, 100, 174, 4, 0, 35, 18, 52, 106, 57, 39, 100, 174, 4, 0, 67, 18, 52, 106, 57, 39, 100, 174, 4, 0, 125, 40, 52, 106, 57, 39, 100, 174, 4, 0, 139, 41, 52, 106, 57, 39, 100, 174, 4, 0, 135, 33, 52, 106, 57, 39, 100, 174, 4, 0, 47, 18, 204, 107, 57, 39, 100, 174, 4, 0, 139, 9, 186, 107, 57, 39, 100, 174, 4, 0, 207, 7, 13, 0, 57, 39, 100, 174, 4, 0, 231, 7, 13, 0, 57, 39, 100, 174, 4, 0, 55, 22, 13, 0, 57, 39, 100, 174, 4, 0, 59, 22, 13, 0, 57, 39, 100, 174, 4, 0, 63, 22, 13, 0, 57, 39, 100, 174, 4, 0, 67, 22, 13, 0, 57, 39, 100, 174, 4, 0, 71, 22, 13, 0, 57, 39, 100, 174, 4, 0, 217, 43, 13, 0, 57, 39, 100, 174, 4, 0, 6, 44, 154, 106, 57, 39, 100, 174, 4, 0, 61, 111, 154, 106, 57, 39, 100, 174, 4, 0, 107, 23, 154, 106, 57, 39, 100, 174, 4, 0, 219, 28, 154, 106, 57, 39, 100, 174, 4, 0, 231, 24, 214, 109, 57, 39, 100, 174, 4, 0, 111, 21, 99, 1, 57, 39, 100, 174, 4, 0, 203, 7, 99, 1, 57, 39, 100, 174, 4, 0, 223, 28, 99, 1, 57, 39, 100, 174, 4, 0, 227, 29, 99, 1, 57, 39, 100, 174, 4, 0, 12, 43, 99, 1, 57, 39, 100, 174, 4, 0, 134, 41, 99, 1, 57, 39, 100, 174, 4, 0, 231, 33, 99, 1, 57, 39, 100, 174, 4, 0, 67, 34, 99, 1, 57, 39, 100, 174, 4, 0, 11, 23, 163, 26, 57, 39, 100, 174, 4, 0, 219, 22, 171, 26, 57, 39, 100, 174, 4, 0, 151, 24, 55, 1, 57, 39, 100, 174, 4, 0, 43, 27, 55, 1, 57, 39, 100, 174, 4, 0, 49, 193, 55, 1, 57, 39, 100, 174, 4, 0, 47, 27, 55, 1, 57, 39, 100, 174, 4, 0, 171, 8, 55, 1, 57, 39, 100, 174, 4, 0, 79, 41, 55, 1, 57, 39, 100, 174, 4, 0, 159, 41, 55, 1, 57, 39, 100, 174, 4, 0, 151, 27, 55, 1, 57, 39, 100, 174, 4, 0, 183, 8, 55, 1, 57, 39, 100, 174, 4, 0, 191, 27, 55, 1, 57, 39, 100, 174, 4, 0, 253, 108, 55, 1, 57, 39, 100, 174, 4, 0, 127, 29, 55, 1, 57, 39, 100, 174, 4, 0, 55, 32, 55, 1, 57, 39, 100, 174, 4, 0, 147, 27, 178, 109, 57, 39, 100, 174, 4, 0, 115, 18, 97, 39, 57, 39, 100, 174, 4, 0, 123, 7, 97, 39, 57, 39, 100, 174, 4, 0, 135, 18, 97, 39, 57, 39, 100, 174, 4, 0, 227, 18, 97, 39, 57, 39, 100, 174, 4, 0, 190, 70, 97, 39, 57, 39, 100, 174, 4, 0, 99, 33, 97, 39, 57, 39, 100, 174, 4, 0, 227, 33, 97, 39, 57, 39, 100, 174, 4, 0, 15, 34, 97, 39, 57, 39, 100, 174, 4, 0, 75, 34, 97, 39, 57, 39, 100, 174, 4, 0, 255, 36, 97, 39, 57, 39, 100, 174, 4, 0, 3, 37, 97, 39, 57, 39, 100, 174, 4, 0, 107, 28, 232, 39, 57, 39, 100, 174, 4, 0, 123, 18, 77, 39, 57, 39, 100, 174, 4, 0, 183, 18, 77, 39, 57, 39, 100, 174, 4, 0, 191, 18, 77, 39, 57, 39, 100, 174, 4, 0, 195, 18, 77, 39, 57, 39, 100, 174, 4, 0, 25, 111, 77, 39, 57, 39, 100, 174, 4, 0, 119, 20, 77, 39, 57, 39, 100, 174, 4, 0, 43, 8, 77, 39, 57, 39, 100, 174, 4, 0, 137, 39, 77, 39, 57, 39, 100, 174, 4, 0, 95, 70, 77, 39, 57, 39, 100, 174, 4, 0, 155, 25, 77, 39, 57, 39, 100, 174, 4, 0, 187, 26, 77, 39, 57, 39, 100, 174, 4, 0, 191, 26, 77, 39, 57, 39, 100, 174, 4, 0, 203, 26, 77, 39, 57, 39, 100, 174, 4, 0, 173, 110, 77, 39, 57, 39, 100, 174, 4, 0, 235, 26, 77, 39, 57, 39, 100, 174, 4, 0, 19, 27, 77, 39, 57, 39, 100, 174, 4, 0, 227, 27, 77, 39, 57, 39, 100, 174, 4, 0, 239, 27, 77, 39, 57, 39, 100, 174, 4, 0, 243, 27, 77, 39, 57, 39, 100, 174, 4, 0, 247, 27, 77, 39, 57, 39, 100, 174, 4, 0, 15, 28, 77, 39, 57, 39, 100, 174, 4, 0, 87, 28, 77, 39, 57, 39, 100, 174, 4, 0, 91, 28, 77, 39, 57, 39, 100, 174, 4, 0, 99, 28, 77, 39, 57, 39, 100, 174, 4, 0, 31, 30, 77, 39, 57, 39, 100, 174, 4, 0, 35, 30, 77, 39, 57, 39, 100, 174, 4, 0, 187, 30, 77, 39, 57, 39, 100, 174, 4, 0, 32, 43, 77, 39, 57, 39, 100, 174, 4, 0, 43, 34, 77, 39, 57, 39, 100, 174, 4, 0, 207, 34, 77, 39, 57, 39, 100, 174, 4, 0, 111, 35, 77, 39, 57, 39, 100, 174, 4, 0, 115, 35, 77, 39, 57, 39, 100, 174, 4, 0, 251, 36, 77, 39, 57, 39, 100, 174, 4, 0, 223, 38, 77, 39, 57, 39, 100, 174, 4, 0, 181, 39, 77, 39, 57, 39, 100, 174, 4, 0, 175, 7, 227, 39, 57, 39, 100, 174, 4, 0, 131, 22, 200, 1, 57, 39, 100, 174, 4, 0, 167, 23, 200, 1, 57, 39, 100, 174, 4, 0, 3, 9, 200, 1, 57, 39, 100, 174, 4, 0, 144, 41, 180, 107, 57, 39, 100, 174, 4, 0, 143, 30, 172, 39, 57, 39, 100, 174, 4, 0, 75, 31, 172, 39, 57, 39, 100, 174, 4, 0, 179, 31, 72, 39, 57, 39, 100, 174, 4, 0, 183, 31, 72, 39, 57, 39, 100, 174, 4, 0, 187, 31, 72, 39, 57, 39, 100, 174, 4, 0, 191, 31, 72, 39, 57, 39, 100, 174, 4, 0, 195, 31, 72, 39, 57, 39, 100, 174, 4, 0, 199, 31, 72, 39, 57, 39, 100, 174, 4, 0, 203, 31, 72, 39, 57, 39, 100, 174, 4, 0, 207, 31, 72, 39, 57, 39, 100, 174, 4, 0, 211, 31, 72, 39, 57, 39, 100, 174, 4, 0, 41, 110, 72, 39, 57, 39, 100, 174, 4, 0, 215, 31, 72, 39, 57, 39, 100, 174, 4, 0, 219, 31, 72, 39, 57, 39, 100, 174, 4, 0, 223, 31, 72, 39, 57, 39, 100, 174, 4, 0, 19, 9, 72, 39, 57, 39, 100, 174, 4, 0, 23, 9, 72, 39, 57, 39, 100, 174, 4, 0, 27, 9, 72, 39, 57, 39, 100, 174, 4, 0, 231, 31, 72, 39, 57, 39, 100, 174, 4, 0, 235, 31, 72, 39, 57, 39, 100, 174, 4, 0, 239, 31, 72, 39, 57, 39, 100, 174, 4, 0, 243, 31, 72, 39, 57, 39, 100, 174, 4, 0, 247, 31, 72, 39, 57, 39, 100, 174, 4, 0, 251, 31, 72, 39, 57, 39, 100, 174, 4, 0, 255, 31, 72, 39, 57, 39, 100, 174, 4, 0, 3, 32, 72, 39, 57, 39, 100, 174, 4, 0, 31, 9, 72, 39, 57, 39, 100, 174, 4, 0, 35, 9, 72, 39, 57, 39, 100, 174, 4, 0, 7, 32, 72, 39, 57, 39, 100, 174, 4, 0, 110, 40, 72, 39, 57, 39, 100, 174, 4, 0, 39, 9, 72, 39, 57, 39, 100, 174, 4, 0, 43, 9, 72, 39, 57, 39, 100, 174, 4, 0, 19, 32, 72, 39, 57, 39, 100, 174, 4, 0, 47, 9, 72, 39, 57, 39, 100, 174, 4, 0, 51, 9, 72, 39, 57, 39, 100, 174, 4, 0, 23, 2, 72, 39, 57, 39, 100, 174, 4, 0, 55, 9, 72, 39, 57, 39, 100, 174, 4, 0, 23, 32, 72, 39, 57, 39, 100, 174, 4, 0, 27, 32, 72, 39, 57, 39, 100, 174, 4, 0, 31, 32, 72, 39, 57, 39, 100, 174, 4, 0, 35, 32, 72, 39, 57, 39, 100, 174, 4, 0, 39, 32, 72, 39, 57, 39, 100, 174, 4, 0, 43, 32, 72, 39, 57, 39, 100, 174, 4, 0, 59, 9, 72, 39, 57, 39, 100, 174, 4, 0, 187, 43, 222, 39, 57, 39, 100, 174, 4, 0, 19, 33, 222, 39, 57, 39, 100, 174, 4, 0, 131, 33, 222, 39, 57, 39, 100, 174, 4, 0, 243, 16, 226, 106, 57, 39, 100, 174, 4, 0, 103, 42, 226, 106, 57, 39, 100, 174, 4, 0, 111, 17, 226, 106, 57, 39, 100, 174, 4, 0, 225, 40, 195, 33, 57, 39, 100, 174, 4, 0, 135, 9, 192, 107, 57, 39, 100, 174, 4, 0, 231, 37, 92, 2, 57, 39, 100, 174, 4, 0, 7, 19, 60, 1, 57, 39, 100, 174, 4, 0, 139, 7, 60, 1, 57, 39, 100, 174, 4, 0, 175, 19, 60, 1, 57, 39, 100, 174, 4, 0, 171, 7, 60, 1, 57, 39, 100, 174, 4, 0, 239, 19, 60, 1, 57, 39, 100, 174, 4, 0, 233, 42, 60, 1, 57, 39, 100, 174, 4, 0, 171, 23, 60, 1, 57, 39, 100, 174, 4, 0, 47, 8, 60, 1, 57, 39, 100, 174, 4, 0, 120, 40, 60, 1, 57, 39, 100, 174, 4, 0, 50, 70, 60, 1, 57, 39, 100, 174, 4, 0, 199, 111, 60, 1, 57, 39, 100, 174, 4, 0, 59, 36, 60, 1, 57, 39, 100, 174, 4, 0, 95, 38, 60, 1, 57, 39, 100, 174, 4, 0, 71, 199, 71, 199, 77, 44, 37, 1, 4, 0, 205, 192, 245, 191, 57, 39, 100, 174, 4, 0, 229, 191, 213, 191, 57, 39, 100, 174, 4, 0, 117, 192, 245, 191, 57, 39, 100, 174, 4, 0, 53, 192, 213, 191, 57, 39, 100, 174, 4, 0, 125, 192, 213, 191, 57, 39, 100, 174, 4, 0, 77, 192, 213, 191, 57, 39, 100, 174, 4, 0, 61, 192, 213, 191, 57, 39, 100, 174, 4, 0, 133, 192, 213, 191, 57, 39, 100, 174, 4, 0, 21, 192, 213, 191, 57, 39, 100, 174, 4, 0, 205, 191, 213, 191, 57, 39, 100, 174, 4, 0, 85, 192, 213, 191, 57, 39, 100, 174, 4, 0, 141, 192, 213, 191, 57, 39, 100, 174, 4, 0, 221, 192, 213, 191, 57, 39, 100, 174, 4, 0, 245, 192, 213, 191, 57, 39, 100, 174, 4, 0, 69, 192, 213, 191, 57, 39, 100, 174, 4, 0, 93, 192, 213, 191, 57, 39, 100, 174, 4, 0, 37, 192, 213, 191, 57, 39, 100, 174, 4, 0, 253, 191, 213, 191, 57, 39, 100, 174, 4, 0, 5, 193, 245, 191, 57, 39, 100, 174, 4, 0, 13, 193, 213, 191, 57, 39, 100, 174, 4, 0, 181, 191, 245, 191, 57, 39, 100, 174, 4, 0, 13, 192, 213, 191, 57, 39, 100, 174, 4, 0, 157, 192, 253, 192, 41, 109, 37, 1, 4, 0, 237, 191, 29, 192, 40, 71, 100, 174, 4, 0, 29, 192, 237, 191, 45, 71, 100, 174, 4, 0, 112, 174, 112, 174, 162, 44, 37, 1, 4, 0, 101, 201, 101, 201, 107, 106, 37, 1, 4, 0, 42, 202, 42, 202, 107, 106, 37, 1, 4, 0, 144, 174, 144, 174, 41, 109, 37, 1, 4, 0, 125, 199, 125, 199, 107, 106, 37, 1, 4, 0, 2, 200, 149, 192, 35, 133, 37, 1, 4, 0, 2, 200, 2, 200, 107, 106, 37, 1, 4, 0, 116, 174, 116, 174, 41, 109, 37, 1, 4, 0, 57, 202, 57, 202, 107, 106, 37, 1, 4, 0, 252, 202, 252, 202, 107, 106, 37, 1, 4, 0, 123, 174, 123, 174, 41, 109, 37, 1, 4, 0, 72, 202, 72, 202, 107, 106, 37, 1, 4, 0, 12, 203, 12, 203, 107, 106, 37, 1, 4, 0, 109, 174, 109, 174, 41, 109, 37, 1, 4, 0, 87, 202, 87, 202, 107, 106, 37, 1, 4, 0, 28, 203, 28, 203, 107, 106, 37, 1, 4, 0, 63, 199, 63, 199, 41, 109, 37, 1, 4, 0, 75, 107, 75, 107, 107, 106, 37, 1, 85, 0, 26, 179, 43, 109, 49, 109, 55, 109, 61, 109, 215, 111, 67, 109, 11, 108, 125, 141, 175, 106, 73, 109, 221, 111, 87, 107, 227, 111, 181, 106, 233, 111, 17, 108, 229, 106, 235, 106, 239, 111, 241, 106, 85, 106, 79, 109, 245, 111, 199, 106, 85, 109, 23, 108, 29, 108, 137, 141, 247, 106, 93, 107, 91, 109, 43, 106, 251, 111, 97, 109, 99, 107, 91, 106, 35, 108, 103, 109, 1, 112, 109, 109, 115, 109, 41, 108, 105, 107, 143, 141, 7, 112, 205, 106, 253, 106, 111, 107, 3, 107, 211, 106, 13, 112, 97, 106, 117, 107, 47, 108, 123, 107, 129, 107, 61, 106, 73, 106, 19, 112, 135, 107, 9, 107, 121, 109, 141, 107, 15, 107, 27, 107, 25, 112, 33, 107, 53, 108, 127, 109, 59, 108, 31, 112, 133, 109, 65, 108, 71, 108, 77, 108, 145, 106, 79, 106, 83, 108, 139, 109, 39, 107, 147, 107, 67, 106, 45, 107, 51, 107, 9, 0, 121, 106, 127, 106, 131, 141, 55, 106, 187, 106, 193, 106, 109, 106, 21, 107, 139, 106, 22, 0, 89, 108, 153, 107, 185, 141, 95, 108, 101, 108, 145, 109, 107, 108, 151, 109, 113, 108, 159, 107, 119, 108, 157, 109, 125, 108, 131, 108, 137, 108, 143, 108, 163, 109, 165, 107, 149, 108, 191, 141, 169, 109, 155, 108, 11, 0, 57, 107, 157, 106, 197, 108, 115, 106, 63, 107, 203, 108, 163, 106, 223, 109, 133, 106, 229, 109, 209, 108, 39, 0, 101, 193, 109, 193, 0, 198, 101, 192, 165, 193, 18, 198, 21, 193, 36, 198, 157, 197, 173, 192, 109, 192, 181, 192, 93, 194, 205, 194, 229, 194, 237, 194, 245, 194, 253, 194, 5, 195, 157, 195, 165, 192, 189, 195, 205, 195, 37, 196, 189, 192, 197, 192, 198, 198, 229, 196, 157, 192, 253, 196, 5, 197, 133, 197, 197, 191, 189, 191, 225, 198, 149, 197, 158, 197, 14, 199, 29, 192, 7, 0, 21, 73, 26, 73, 31, 73, 36, 73, 41, 73, 46, 73, 51, 73, 7, 0, 249, 45, 254, 45, 3, 46, 8, 46, 229, 40, 13, 46, 234, 40, 34, 0, 241, 133, 247, 133, 253, 109, 99, 113, 105, 113, 111, 113, 117, 113, 123, 113, 3, 110, 129, 113, 135, 113, 141, 113, 9, 110, 147, 113, 153, 113, 159, 113, 165, 113, 171, 113, 177, 113, 183, 113, 189, 113, 195, 113, 201, 113, 207, 113, 213, 113, 219, 113, 225, 113, 231, 113, 237, 113, 243, 113, 249, 107, 249, 113, 255, 113, 5, 114, 8, 0, 56, 73, 61, 73, 66, 73, 71, 73, 76, 73, 81, 73, 86, 73, 91, 73, 12, 0, 96, 73, 101, 73, 106, 73, 111, 73, 116, 73, 121, 73, 126, 73, 131, 73, 136, 73, 141, 73, 146, 73, 151, 73, 11, 0, 18, 46, 226, 107, 23, 46, 28, 46, 33, 46, 38, 46, 43, 46, 48, 46, 53, 46, 58, 46, 63, 46, 18, 0, 11, 114, 17, 114, 23, 114, 29, 114, 35, 114, 123, 134, 69, 131, 41, 114, 47, 114, 53, 114, 59, 114, 65, 114, 81, 107, 71, 114, 77, 114, 83, 114, 255, 107, 15, 110, 24, 0, 89, 110, 80, 40, 65, 70, 46, 200, 156, 200, 159, 17, 163, 17, 167, 17, 171, 17, 151, 111, 175, 17, 131, 131, 13, 111, 179, 17, 107, 7, 183, 17, 111, 7, 182, 200, 187, 17, 191, 17, 195, 17, 199, 17, 115, 7, 119, 7, 9, 0, 207, 17, 211, 17, 215, 17, 219, 17, 223, 17, 227, 17, 231, 17, 235, 17, 239, 17, 8, 0, 89, 114, 95, 114, 68, 46, 101, 114, 73, 46, 107, 114, 113, 114, 239, 40, 70, 0, 119, 114, 125, 114, 131, 114, 137, 114, 143, 114, 149, 114, 155, 114, 10, 44, 21, 110, 167, 114, 173, 114, 179, 114, 185, 114, 191, 114, 203, 114, 209, 114, 255, 39, 215, 114, 221, 114, 227, 114, 233, 114, 239, 114, 27, 110, 245, 114, 251, 114, 7, 115, 4, 40, 13, 115, 19, 115, 25, 115, 31, 115, 180, 39, 139, 40, 37, 115, 78, 46, 33, 110, 49, 115, 55, 115, 61, 115, 67, 115, 73, 115, 79, 115, 39, 110, 83, 46, 85, 115, 77, 135, 83, 135, 109, 115, 115, 115, 121, 115, 127, 115, 144, 40, 133, 115, 139, 115, 185, 39, 145, 115, 151, 115, 157, 115, 163, 115, 88, 46, 169, 115, 175, 115, 181, 115, 187, 115, 193, 115, 93, 46, 199, 115, 209, 111, 89, 135, 49, 106, 3, 0, 205, 115, 211, 115, 217, 115, 11, 0, 45, 74, 50, 74, 55, 74, 60, 74, 65, 74, 70, 74, 75, 74, 80, 74, 85, 74, 90, 74, 95, 74, 8, 0, 79, 18, 83, 18, 87, 18, 91, 18, 95, 18, 137, 131, 255, 109, 99, 18, 3, 0, 223, 115, 241, 115, 9, 116, 13, 0, 164, 75, 169, 75, 174, 75, 179, 75, 184, 75, 189, 75, 194, 75, 199, 75, 204, 75, 209, 75, 214, 75, 219, 75, 224, 75, 28, 0, 229, 75, 234, 75, 239, 75, 244, 75, 249, 75, 254, 75, 3, 76, 8, 76, 13, 76, 18, 76, 23, 76, 28, 76, 33, 76, 38, 76, 43, 76, 48, 76, 53, 76, 58, 76, 63, 76, 68, 76, 73, 76, 78, 76, 83, 76, 88, 76, 93, 76, 98, 76, 103, 76, 108, 76, 4, 0, 113, 76, 118, 76, 123, 76, 98, 46, 18, 0, 103, 46, 133, 76, 138, 76, 244, 40, 108, 46, 113, 46, 118, 46, 249, 40, 123, 46, 128, 46, 254, 40, 133, 46, 138, 46, 3, 41, 143, 46, 148, 46, 153, 46, 158, 46, 12, 0, 94, 70, 143, 76, 8, 41, 163, 46, 168, 46, 173, 46, 178, 46, 183, 46, 188, 46, 193, 46, 198, 46, 203, 46, 4, 0, 208, 46, 213, 46, 218, 46, 13, 41, 9, 0, 147, 18, 151, 18, 15, 109, 229, 108, 155, 18, 159, 18, 43, 111, 163, 18, 167, 18, 3, 0, 223, 46, 228, 46, 233, 46, 27, 0, 238, 46, 148, 76, 153, 76, 243, 46, 248, 46, 253, 46, 2, 47, 7, 47, 12, 47, 17, 47, 22, 47, 158, 76, 163, 76, 168, 76, 173, 76, 27, 47, 32, 47, 37, 47, 42, 47, 178, 76, 99, 70, 183, 76, 188, 76, 47, 47, 52, 47, 57, 47, 62, 47, 32, 0, 67, 47, 72, 47, 77, 47, 82, 47, 87, 47, 92, 47, 97, 47, 102, 47, 107, 47, 18, 41, 112, 47, 117, 47, 214, 107, 122, 47, 127, 47, 132, 47, 137, 47, 142, 47, 147, 47, 152, 47, 157, 47, 193, 76, 198, 76, 162, 47, 23, 41, 167, 47, 172, 47, 177, 47, 182, 47, 187, 47, 9, 40, 192, 47, 20, 0, 203, 76, 208, 76, 213, 76, 218, 76, 223, 76, 228, 76, 233, 76, 238, 76, 243, 76, 248, 76, 253, 76, 2, 77, 7, 77, 12, 77, 17, 77, 22, 77, 27, 77, 32, 77, 197, 47, 202, 47, 16, 0, 207, 47, 212, 47, 80, 39, 217, 47, 222, 47, 20, 44, 227, 47, 149, 40, 154, 40, 28, 41, 232, 47, 33, 41, 25, 44, 37, 77, 42, 77, 237, 47, 7, 0, 14, 40, 242, 47, 30, 44, 160, 39, 247, 47, 252, 47, 35, 44, 6, 0, 1, 48, 19, 40, 255, 116, 6, 48, 220, 39, 5, 117, 13, 0, 11, 48, 16, 48, 21, 48, 26, 48, 31, 48, 47, 77, 52, 77, 57, 77, 36, 48, 41, 48, 46, 48, 51, 48, 56, 48, 26, 0, 61, 48, 66, 48, 71, 48, 67, 77, 72, 77, 76, 48, 38, 41, 81, 48, 86, 48, 91, 48, 43, 41, 96, 48, 101, 48, 106, 48, 111, 48, 116, 48, 87, 77, 92, 77, 121, 48, 126, 48, 131, 48, 136, 48, 141, 48, 146, 48, 151, 48, 156, 48, 17, 0, 161, 48, 166, 48, 11, 117, 171, 48, 176, 48, 181, 48, 186, 48, 191, 48, 196, 48, 201, 48, 206, 48, 48, 41, 211, 48, 216, 48, 53, 41, 221, 48, 226, 48, 12, 0, 127, 7, 131, 7, 43, 19, 47, 19, 51, 19, 102, 77, 107, 77, 112, 77, 117, 77, 122, 77, 127, 77, 17, 117, 26, 0, 231, 48, 236, 48, 241, 48, 246, 48, 251, 48, 0, 49, 5, 49, 10, 49, 15, 49, 20, 49, 25, 49, 30, 49, 35, 49, 40, 49, 45, 49, 132, 77, 104, 70, 50, 49, 55, 49, 60, 49, 58, 41, 65, 49, 70, 49, 75, 49, 137, 77, 142, 77, 14, 0, 80, 49, 85, 49, 90, 49, 95, 49, 100, 49, 105, 49, 63, 41, 110, 49, 115, 49, 120, 49, 68, 41, 125, 49, 130, 49, 73, 41, 16, 0, 135, 49, 140, 49, 145, 49, 150, 49, 155, 49, 160, 49, 165, 49, 170, 49, 175, 49, 180, 49, 185, 49, 190, 49, 195, 49, 78, 41, 200, 49, 83, 41, 10, 0, 205, 49, 210, 49, 215, 49, 220, 49, 225, 49, 230, 49, 235, 49, 240, 49, 245, 49, 250, 49, 34, 0, 255, 49, 4, 50, 9, 50, 40, 44, 14, 50, 19, 50, 24, 50, 29, 50, 152, 77, 157, 77, 34, 50, 39, 50, 109, 70, 162, 77, 44, 50, 49, 50, 54, 50, 59, 50, 88, 41, 64, 50, 69, 50, 93, 41, 74, 50, 167, 77, 172, 77, 79, 50, 84, 50, 89, 50, 94, 50, 99, 50, 104, 50, 109, 50, 84, 40, 114, 50, 33, 0, 23, 117, 29, 117, 35, 117, 41, 117, 227, 108, 47, 117, 53, 117, 59, 117, 65, 117, 71, 117, 77, 117, 83, 117, 89, 117, 95, 117, 119, 50, 101, 117, 69, 110, 107, 117, 113, 117, 119, 117, 125, 117, 131, 117, 75, 110, 137, 117, 143, 117, 149, 117, 155, 117, 81, 110, 161, 117, 167, 117, 173, 117, 179, 117, 185, 117, 7, 0, 143, 7, 147, 7, 187, 19, 191, 19, 151, 7, 155, 7, 124, 50, 16, 0, 129, 50, 177, 77, 182, 77, 187, 77, 192, 77, 197, 77, 202, 77, 207, 77, 212, 77, 217, 77, 222, 77, 227, 77, 232, 77, 237, 77, 242, 77, 45, 44, 2, 0, 163, 7, 167, 7, 6, 0, 252, 77, 1, 78, 6, 78, 11, 78, 16, 78, 21, 78, 14, 0, 224, 50, 229, 50, 234, 50, 239, 50, 244, 50, 249, 50, 55, 44, 60, 44, 254, 50, 3, 51, 65, 44, 8, 51, 70, 44, 75, 44, 16, 0, 13, 51, 18, 51, 23, 51, 28, 51, 33, 51, 38, 51, 43, 51, 48, 51, 53, 51, 58, 51, 63, 51, 68, 51, 89, 40, 73, 51, 94, 40, 113, 41, 6, 0, 119, 70, 61, 78, 66, 78, 71, 78, 118, 41, 78, 51, 5, 0, 76, 78, 81, 78, 86, 78, 91, 78, 96, 78, 10, 0, 101, 78, 106, 78, 111, 78, 116, 78, 121, 78, 126, 78, 131, 78, 136, 78, 141, 78, 146, 78, 10, 0, 55, 79, 60, 79, 65, 79, 70, 79, 75, 79, 80, 79, 85, 79, 90, 79, 95, 79, 100, 79, 48, 0, 105, 79, 110, 79, 115, 79, 120, 79, 125, 79, 130, 79, 135, 79, 140, 79, 145, 79, 150, 79, 155, 79, 160, 79, 165, 79, 170, 79, 175, 79, 180, 79, 185, 79, 190, 79, 195, 79, 200, 79, 205, 79, 210, 79, 215, 79, 220, 79, 225, 79, 230, 79, 235, 79, 240, 79, 245, 79, 250, 79, 255, 79, 224, 70, 4, 80, 9, 80, 14, 80, 229, 70, 19, 80, 24, 80, 29, 80, 34, 80, 39, 80, 44, 80, 49, 80, 234, 70, 54, 80, 59, 80, 64, 80, 239, 70, 24, 0, 203, 20, 207, 20, 211, 20, 215, 20, 23, 42, 219, 20, 223, 20, 227, 20, 231, 20, 235, 20, 239, 20, 155, 70, 160, 70, 243, 20, 247, 20, 251, 20, 69, 80, 74, 80, 255, 20, 3, 21, 28, 42, 7, 21, 179, 7, 11, 21, 15, 0, 83, 51, 88, 51, 93, 51, 98, 51, 103, 51, 108, 51, 113, 51, 118, 51, 123, 51, 128, 51, 133, 51, 138, 51, 143, 51, 148, 51, 153, 51, 27, 0, 197, 117, 203, 117, 209, 117, 158, 51, 163, 51, 215, 117, 183, 7, 168, 51, 173, 51, 221, 117, 178, 51, 183, 51, 188, 51, 193, 51, 198, 51, 227, 117, 203, 51, 208, 51, 123, 41, 233, 117, 213, 51, 239, 117, 245, 117, 251, 117, 1, 118, 7, 118, 13, 118, 6, 0, 218, 51, 223, 51, 228, 51, 233, 51, 238, 51, 243, 51, 19, 0, 2, 52, 79, 80, 84, 80, 124, 70, 27, 52, 99, 80, 104, 80, 32, 52, 119, 80, 128, 41, 42, 52, 62, 52, 129, 80, 134, 80, 87, 52, 102, 52, 143, 41, 122, 52, 162, 52, 11, 0, 177, 52, 29, 40, 182, 52, 187, 52, 192, 52, 197, 52, 202, 52, 207, 52, 149, 80, 154, 80, 212, 52, 19, 0, 217, 52, 159, 80, 164, 80, 169, 80, 174, 80, 179, 80, 184, 80, 189, 80, 194, 80, 199, 80, 204, 80, 209, 80, 214, 80, 219, 80, 224, 80, 229, 80, 234, 80, 239, 80, 244, 80, 5, 0, 211, 7, 215, 7, 219, 7, 223, 7, 227, 7, 4, 0, 19, 118, 25, 118, 31, 118, 37, 118, 26, 0, 87, 110, 43, 118, 222, 52, 49, 118, 55, 118, 227, 52, 61, 118, 93, 110, 232, 52, 67, 118, 73, 118, 79, 118, 99, 110, 85, 118, 242, 52, 247, 52, 91, 118, 252, 52, 97, 118, 103, 118, 109, 118, 115, 118, 1, 53, 6, 53, 16, 53, 21, 53, 9, 0, 91, 22, 95, 22, 99, 22, 103, 22, 107, 22, 111, 22, 115, 22, 119, 22, 123, 22, 7, 0, 141, 119, 247, 142, 213, 119, 233, 120, 169, 121, 51, 122, 99, 122, 7, 0, 1, 83, 6, 83, 11, 83, 16, 83, 21, 83, 26, 83, 31, 53, 12, 0, 36, 53, 41, 53, 46, 53, 51, 53, 153, 41, 56, 53, 61, 53, 66, 53, 80, 44, 129, 70, 99, 40, 71, 53, 10, 0, 76, 53, 81, 53, 86, 53, 91, 53, 96, 53, 101, 53, 106, 53, 111, 53, 116, 53, 121, 53, 5, 0, 126, 53, 131, 53, 136, 53, 141, 53, 158, 41, 6, 0, 239, 7, 247, 22, 251, 22, 124, 41, 243, 7, 247, 7, 8, 0, 251, 7, 255, 7, 3, 8, 7, 8, 11, 8, 15, 8, 19, 8, 23, 8, 2, 0, 27, 8, 47, 3, 14, 0, 0, 54, 43, 23, 47, 23, 51, 23, 55, 23, 149, 110, 59, 23, 63, 23, 67, 23, 253, 42, 71, 23, 75, 23, 79, 23, 31, 8, 22, 0, 5, 54, 10, 54, 15, 54, 20, 54, 25, 54, 30, 54, 35, 54, 40, 54, 188, 41, 193, 41, 45, 54, 50, 54, 55, 54, 60, 54, 65, 54, 70, 54, 75, 54, 80, 54, 85, 54, 90, 54, 95, 54, 198, 41, 4, 0, 165, 39, 35, 8, 235, 108, 39, 8, 10, 0, 120, 54, 85, 44, 125, 39, 125, 54, 105, 39, 130, 54, 135, 54, 90, 44, 140, 54, 213, 41, 18, 0, 145, 54, 150, 54, 106, 83, 111, 83, 218, 41, 155, 54, 160, 54, 165, 54, 170, 54, 175, 54, 180, 54, 185, 54, 190, 54, 195, 54, 200, 54, 205, 54, 210, 54, 215, 54, 21, 0, 116, 83, 121, 83, 126, 83, 131, 83, 136, 83, 141, 83, 146, 83, 151, 83, 156, 83, 161, 83, 166, 83, 171, 83, 176, 83, 181, 83, 186, 83, 191, 83, 196, 83, 201, 83, 206, 83, 211, 83, 216, 83, 10, 0, 232, 107, 220, 54, 225, 54, 221, 83, 226, 83, 230, 54, 235, 54, 240, 54, 231, 83, 236, 83, 43, 0, 245, 54, 250, 54, 255, 54, 4, 55, 9, 55, 14, 55, 95, 44, 223, 41, 228, 41, 19, 55, 24, 55, 29, 55, 34, 55, 233, 41, 39, 55, 100, 44, 44, 55, 49, 55, 54, 55, 39, 40, 105, 44, 238, 41, 243, 41, 59, 55, 64, 55, 69, 55, 170, 39, 74, 55, 79, 55, 84, 55, 248, 41, 253, 41, 241, 83, 139, 70, 246, 83, 89, 55, 94, 55, 99, 55, 104, 55, 2, 42, 109, 55, 114, 55, 119, 55, 7, 0, 12, 42, 15, 84, 169, 55, 50, 84, 17, 42, 149, 70, 70, 84, 4, 0, 51, 8, 63, 8, 67, 8, 71, 8, 6, 0, 75, 8, 57, 42, 249, 55, 55, 3, 254, 55, 79, 8, 36, 0, 3, 56, 8, 56, 120, 84, 125, 84, 13, 56, 255, 209, 18, 56, 23, 56, 28, 56, 33, 56, 38, 56, 43, 56, 48, 56, 53, 56, 58, 56, 63, 56, 68, 56, 73, 56, 78, 56, 83, 56, 88, 56, 93, 56, 98, 56, 103, 56, 108, 56, 113, 56, 118, 56, 123, 56, 128, 56, 133, 56, 62, 42, 138, 56, 67, 42, 143, 56, 148, 56, 153, 56, 18, 0, 158, 56, 163, 56, 130, 84, 135, 84, 168, 56, 173, 56, 178, 56, 183, 56, 188, 56, 193, 56, 198, 56, 203, 56, 208, 56, 213, 56, 218, 56, 223, 56, 228, 56, 109, 40, 31, 0, 140, 84, 145, 84, 150, 84, 155, 84, 160, 84, 165, 84, 170, 84, 175, 84, 180, 84, 185, 84, 190, 84, 195, 84, 200, 84, 205, 84, 210, 84, 215, 84, 220, 84, 225, 84, 230, 84, 235, 84, 240, 84, 245, 84, 250, 84, 255, 84, 4, 85, 9, 85, 14, 85, 19, 85, 24, 85, 29, 85, 34, 85, 8, 0, 47, 25, 51, 25, 55, 25, 59, 25, 63, 25, 67, 25, 71, 25, 75, 25, 20, 0, 233, 56, 238, 56, 243, 56, 248, 56, 253, 56, 2, 57, 7, 57, 12, 57, 17, 57, 22, 57, 27, 57, 32, 57, 37, 57, 42, 57, 47, 57, 52, 57, 39, 85, 44, 85, 57, 57, 62, 57, 14, 0, 28, 86, 33, 86, 38, 86, 43, 86, 48, 86, 53, 86, 58, 86, 63, 86, 68, 86, 73, 86, 78, 86, 83, 86, 88, 86, 93, 86, 12, 0, 56, 58, 61, 58, 66, 58, 71, 58, 76, 58, 81, 58, 86, 58, 91, 58, 96, 58, 101, 58, 106, 58, 111, 58, 47, 0, 98, 86, 103, 86, 108, 86, 113, 86, 118, 86, 123, 86, 128, 86, 133, 86, 138, 86, 143, 86, 148, 86, 153, 86, 158, 86, 163, 86, 168, 86, 173, 86, 178, 86, 183, 86, 188, 86, 193, 86, 198, 86, 203, 86, 208, 86, 213, 86, 218, 86, 223, 86, 228, 86, 233, 86, 238, 86, 243, 86, 248, 86, 253, 86, 2, 87, 7, 87, 12, 87, 17, 87, 22, 87, 27, 87, 32, 87, 37, 87, 42, 87, 47, 87, 52, 87, 57, 87, 62, 87, 67, 87, 72, 87, 47, 0, 77, 87, 82, 87, 87, 87, 92, 87, 97, 87, 102, 87, 107, 87, 112, 87, 117, 87, 122, 87, 127, 87, 132, 87, 137, 87, 142, 87, 147, 87, 152, 87, 157, 87, 162, 87, 167, 87, 172, 87, 177, 87, 182, 87, 187, 87, 192, 87, 197, 87, 202, 87, 207, 87, 212, 87, 217, 87, 222, 87, 227, 87, 232, 87, 237, 87, 242, 87, 247, 87, 252, 87, 1, 88, 6, 88, 11, 88, 16, 88, 21, 88, 26, 88, 31, 88, 36, 88, 41, 88, 46, 88, 51, 88, 9, 0, 42, 43, 103, 26, 116, 58, 121, 58, 83, 8, 175, 111, 111, 26, 87, 8, 91, 8, 25, 0, 95, 8, 99, 8, 115, 26, 119, 26, 123, 26, 127, 26, 131, 26, 135, 26, 139, 26, 56, 88, 61, 88, 66, 88, 71, 88, 76, 88, 81, 88, 86, 88, 91, 88, 96, 88, 101, 88, 106, 88, 111, 88, 116, 88, 121, 88, 126, 88, 131, 88, 3, 0, 103, 8, 247, 108, 107, 8, 3, 0, 111, 8, 115, 8, 119, 8, 2, 0, 123, 8, 59, 3, 12, 0, 201, 88, 206, 88, 211, 88, 216, 88, 221, 88, 226, 88, 231, 88, 236, 88, 241, 88, 131, 58, 246, 88, 251, 88, 17, 0, 136, 58, 0, 89, 5, 89, 10, 89, 15, 89, 20, 89, 25, 89, 30, 89, 35, 89, 40, 89, 45, 89, 50, 89, 55, 89, 60, 89, 65, 89, 70, 89, 141, 58, 6, 0, 146, 58, 151, 58, 156, 58, 97, 42, 190, 39, 161, 58, 18, 0, 159, 122, 165, 122, 219, 110, 171, 122, 177, 122, 183, 122, 189, 122, 195, 122, 201, 122, 207, 122, 213, 122, 219, 122, 225, 122, 173, 108, 231, 122, 189, 107, 237, 122, 179, 108, 18, 0, 102, 42, 75, 89, 80, 89, 114, 40, 166, 58, 171, 58, 176, 58, 181, 58, 186, 58, 191, 58, 196, 58, 201, 58, 107, 42, 206, 58, 211, 58, 112, 42, 216, 58, 221, 58, 8, 0, 226, 58, 231, 58, 236, 58, 90, 89, 95, 89, 241, 58, 246, 58, 251, 58, 10, 0, 100, 89, 105, 89, 110, 89, 115, 89, 120, 89, 125, 89, 130, 89, 135, 89, 140, 89, 145, 89, 11, 0, 150, 89, 155, 89, 160, 89, 165, 89, 170, 89, 175, 89, 180, 89, 185, 89, 190, 89, 195, 89, 200, 89, 9, 0, 131, 8, 135, 8, 139, 8, 143, 8, 147, 8, 151, 8, 155, 8, 159, 8, 163, 8, 15, 0, 0, 59, 5, 59, 10, 59, 15, 59, 20, 59, 117, 42, 25, 59, 30, 59, 35, 59, 40, 59, 195, 39, 45, 59, 50, 59, 122, 42, 55, 59, 10, 0, 175, 8, 167, 131, 115, 27, 119, 27, 123, 27, 127, 27, 179, 8, 85, 70, 131, 27, 90, 70, 70, 0, 74, 90, 79, 90, 84, 90, 89, 90, 94, 90, 99, 90, 104, 90, 109, 90, 114, 90, 119, 90, 124, 90, 129, 90, 134, 90, 139, 90, 144, 90, 149, 90, 154, 90, 159, 90, 164, 90, 169, 90, 174, 90, 179, 90, 184, 90, 189, 90, 194, 90, 199, 90, 204, 90, 209, 90, 214, 90, 219, 90, 224, 90, 229, 90, 234, 90, 239, 90, 244, 90, 249, 90, 254, 90, 3, 91, 8, 91, 13, 91, 18, 91, 23, 91, 28, 91, 33, 91, 38, 91, 43, 91, 48, 91, 53, 91, 58, 91, 63, 91, 68, 91, 73, 91, 78, 91, 83, 91, 88, 91, 93, 91, 98, 91, 103, 91, 108, 91, 60, 59, 65, 59, 70, 59, 75, 59, 80, 59, 85, 59, 90, 59, 95, 59, 100, 59, 105, 59, 110, 59, 12, 0, 115, 59, 120, 59, 125, 59, 130, 59, 127, 42, 135, 59, 132, 42, 140, 59, 145, 59, 44, 40, 150, 59, 155, 59, 119, 0, 159, 149, 165, 149, 171, 131, 171, 149, 181, 109, 177, 149, 183, 149, 177, 131, 151, 112, 189, 149, 187, 109, 183, 131, 193, 109, 157, 112, 189, 131, 195, 131, 163, 112, 169, 112, 175, 112, 201, 131, 199, 109, 195, 149, 201, 149, 207, 131, 207, 149, 213, 149, 219, 149, 213, 131, 219, 107, 181, 112, 219, 131, 225, 131, 225, 149, 231, 131, 237, 131, 243, 131, 231, 149, 237, 149, 187, 112, 249, 131, 243, 149, 249, 149, 255, 149, 255, 131, 5, 150, 11, 150, 17, 150, 5, 132, 23, 150, 29, 150, 11, 132, 17, 132, 23, 132, 29, 132, 35, 150, 35, 132, 193, 112, 41, 150, 47, 150, 41, 132, 199, 112, 53, 150, 59, 150, 47, 132, 65, 150, 71, 150, 77, 150, 53, 132, 83, 150, 59, 132, 89, 150, 65, 132, 95, 150, 71, 132, 101, 150, 77, 132, 107, 150, 113, 150, 119, 150, 125, 150, 131, 150, 137, 150, 143, 150, 83, 132, 149, 150, 89, 132, 155, 150, 161, 150, 167, 150, 95, 132, 173, 150, 101, 132, 179, 150, 185, 150, 191, 150, 107, 132, 197, 150, 203, 150, 209, 150, 215, 150, 221, 150, 227, 150, 233, 150, 205, 109, 239, 150, 245, 150, 251, 150, 205, 112, 1, 151, 243, 122, 37, 112, 249, 122, 255, 122, 5, 123, 225, 110, 11, 123, 17, 123, 231, 110, 23, 123, 22, 0, 160, 59, 165, 59, 170, 59, 175, 59, 115, 44, 180, 59, 185, 59, 190, 59, 195, 59, 200, 59, 113, 91, 118, 91, 205, 59, 175, 39, 137, 42, 210, 59, 215, 59, 220, 59, 142, 42, 225, 59, 230, 59, 235, 59, 12, 0, 123, 91, 128, 91, 133, 91, 138, 91, 143, 91, 148, 91, 153, 91, 158, 91, 163, 91, 168, 91, 173, 91, 178, 91, 17, 0, 240, 59, 245, 59, 250, 59, 147, 42, 152, 42, 157, 42, 162, 42, 167, 42, 172, 42, 177, 42, 255, 59, 4, 60, 9, 60, 183, 91, 188, 91, 14, 60, 19, 60, 37, 0, 24, 60, 29, 60, 34, 60, 193, 91, 198, 91, 39, 60, 203, 91, 208, 91, 179, 70, 213, 91, 218, 91, 223, 91, 44, 60, 49, 60, 54, 60, 59, 60, 64, 60, 69, 60, 74, 60, 79, 60, 182, 42, 84, 60, 89, 60, 94, 60, 99, 60, 104, 60, 109, 60, 187, 42, 114, 60, 119, 60, 228, 91, 233, 91, 124, 60, 129, 60, 134, 60, 139, 60, 144, 60, 23, 0, 238, 91, 243, 91, 248, 91, 253, 91, 2, 92, 7, 92, 12, 92, 17, 92, 22, 92, 27, 92, 32, 92, 37, 92, 42, 92, 47, 92, 52, 92, 57, 92, 62, 92, 67, 92, 72, 92, 77, 92, 82, 92, 87, 92, 92, 92, 6, 0, 191, 8, 195, 8, 199, 8, 203, 8, 251, 27, 255, 27, 2, 0, 207, 8, 211, 8, 71, 0, 102, 92, 107, 92, 112, 92, 117, 92, 122, 92, 127, 92, 132, 92, 142, 92, 147, 92, 152, 92, 157, 92, 162, 92, 149, 60, 177, 92, 182, 92, 187, 92, 192, 92, 197, 92, 202, 92, 207, 92, 212, 92, 217, 92, 222, 92, 154, 60, 232, 92, 237, 92, 242, 92, 247, 92, 252, 92, 1, 93, 6, 93, 11, 93, 16, 93, 21, 93, 26, 93, 31, 93, 36, 93, 41, 93, 46, 93, 51, 93, 56, 93, 61, 93, 66, 93, 71, 93, 76, 93, 81, 93, 86, 93, 96, 93, 101, 93, 106, 93, 111, 93, 116, 93, 121, 93, 126, 93, 131, 93, 136, 93, 141, 93, 159, 60, 151, 93, 156, 93, 161, 93, 166, 93, 171, 93, 176, 93, 181, 93, 164, 60, 196, 93, 201, 93, 206, 93, 211, 93, 169, 60, 11, 0, 215, 8, 27, 28, 31, 28, 35, 28, 39, 28, 43, 28, 47, 28, 51, 28, 55, 28, 174, 60, 223, 124, 15, 0, 221, 93, 226, 93, 231, 93, 236, 93, 241, 93, 246, 93, 251, 93, 0, 94, 5, 94, 10, 94, 15, 94, 20, 94, 25, 94, 30, 94, 35, 94, 22, 0, 219, 8, 229, 124, 235, 124, 241, 124, 247, 124, 253, 124, 57, 152, 63, 152, 3, 125, 9, 125, 15, 125, 21, 125, 27, 125, 33, 125, 39, 125, 69, 152, 75, 152, 81, 152, 45, 125, 51, 125, 57, 125, 63, 125, 15, 0, 40, 94, 45, 94, 50, 94, 55, 94, 60, 94, 65, 94, 70, 94, 75, 94, 80, 94, 85, 94, 90, 94, 95, 94, 100, 94, 105, 94, 110, 94, 68, 0, 115, 94, 120, 94, 125, 94, 130, 94, 135, 94, 140, 94, 145, 94, 150, 94, 155, 94, 160, 94, 165, 94, 170, 94, 175, 94, 180, 94, 185, 94, 190, 94, 195, 94, 200, 94, 205, 94, 210, 94, 215, 94, 220, 94, 225, 94, 230, 94, 235, 94, 240, 94, 245, 94, 250, 94, 255, 94, 4, 95, 9, 95, 14, 95, 19, 95, 24, 95, 29, 95, 34, 95, 39, 95, 44, 95, 49, 95, 54, 95, 59, 95, 64, 95, 69, 95, 74, 95, 79, 95, 84, 95, 89, 95, 94, 95, 99, 95, 160, 44, 104, 95, 109, 95, 114, 95, 119, 95, 124, 95, 129, 95, 134, 95, 139, 95, 144, 95, 165, 44, 149, 95, 154, 95, 159, 95, 164, 95, 169, 95, 174, 95, 179, 95, 13, 71, 17, 0, 179, 60, 192, 42, 197, 42, 184, 60, 189, 60, 194, 60, 199, 60, 204, 60, 209, 60, 214, 60, 219, 60, 224, 60, 229, 60, 234, 60, 202, 42, 239, 60, 244, 60, 21, 0, 184, 95, 189, 95, 194, 95, 199, 95, 204, 95, 209, 95, 214, 95, 219, 95, 224, 95, 229, 95, 234, 95, 249, 60, 254, 60, 239, 95, 244, 95, 249, 95, 254, 95, 3, 96, 8, 96, 13, 96, 69, 125, 3, 0, 223, 8, 227, 8, 231, 8, 32, 0, 75, 125, 81, 125, 87, 125, 93, 125, 99, 125, 105, 125, 111, 125, 23, 111, 117, 125, 123, 125, 29, 111, 129, 125, 35, 111, 135, 125, 141, 125, 147, 125, 153, 125, 159, 125, 251, 108, 165, 125, 171, 125, 177, 125, 183, 125, 189, 125, 195, 125, 201, 125, 207, 125, 43, 112, 213, 125, 1, 109, 219, 125, 225, 125, 16, 0, 48, 96, 53, 96, 58, 96, 63, 96, 68, 96, 73, 96, 78, 96, 83, 96, 88, 96, 93, 96, 98, 96, 103, 96, 108, 96, 113, 96, 118, 96, 123, 96, 11, 0, 37, 43, 139, 28, 235, 8, 239, 8, 243, 8, 231, 125, 247, 8, 143, 28, 147, 28, 151, 28, 155, 28, 14, 0, 78, 61, 83, 61, 88, 61, 93, 61, 98, 61, 103, 61, 108, 61, 113, 61, 118, 61, 123, 61, 128, 61, 128, 96, 133, 96, 133, 61, 8, 0, 175, 28, 179, 28, 183, 28, 187, 28, 191, 28, 195, 28, 199, 28, 203, 28, 37, 0, 138, 61, 143, 61, 148, 61, 232, 42, 153, 61, 158, 61, 237, 42, 163, 61, 242, 42, 168, 61, 173, 61, 178, 61, 183, 61, 188, 61, 193, 61, 198, 61, 203, 61, 247, 42, 138, 96, 143, 96, 148, 96, 153, 96, 208, 61, 213, 61, 218, 61, 223, 61, 228, 61, 233, 61, 238, 61, 243, 61, 248, 61, 253, 61, 252, 42, 2, 62, 1, 43, 7, 62, 12, 62, 17, 0, 17, 62, 22, 62, 27, 62, 6, 43, 32, 62, 37, 62, 42, 62, 47, 62, 52, 62, 57, 62, 62, 62, 67, 62, 158, 96, 163, 96, 72, 62, 77, 62, 82, 62, 18, 0, 87, 62, 251, 153, 1, 154, 7, 154, 92, 62, 97, 62, 102, 62, 107, 62, 112, 62, 117, 62, 122, 62, 127, 62, 132, 62, 137, 62, 142, 62, 147, 62, 152, 62, 119, 40, 20, 0, 168, 96, 173, 96, 178, 96, 183, 96, 188, 96, 193, 96, 198, 96, 203, 96, 208, 96, 213, 96, 218, 96, 223, 96, 228, 96, 233, 96, 238, 96, 243, 96, 248, 96, 253, 96, 2, 97, 7, 97, 12, 0, 3, 29, 7, 29, 11, 29, 15, 29, 19, 29, 12, 97, 17, 97, 22, 97, 27, 97, 32, 97, 37, 97, 42, 97, 14, 0, 47, 97, 52, 97, 57, 97, 62, 97, 67, 97, 72, 97, 77, 97, 82, 97, 87, 97, 92, 97, 97, 97, 102, 97, 107, 97, 112, 97, 17, 0, 237, 125, 41, 111, 243, 125, 249, 125, 255, 125, 5, 126, 11, 126, 17, 126, 23, 126, 29, 126, 35, 126, 41, 126, 47, 126, 53, 126, 151, 106, 49, 112, 59, 126, 11, 0, 197, 62, 202, 62, 207, 62, 212, 62, 217, 62, 222, 62, 227, 62, 232, 62, 237, 62, 36, 43, 242, 62, 13, 0, 7, 9, 111, 30, 115, 30, 119, 30, 123, 30, 127, 30, 131, 30, 135, 30, 139, 30, 247, 62, 252, 62, 1, 63, 6, 63, 26, 0, 71, 126, 77, 126, 83, 126, 89, 126, 95, 126, 101, 126, 107, 126, 113, 126, 119, 126, 125, 126, 131, 126, 137, 126, 11, 156, 17, 156, 143, 126, 149, 126, 155, 126, 161, 126, 167, 126, 173, 126, 211, 109, 179, 126, 185, 126, 191, 126, 7, 109, 197, 126, 22, 0, 203, 126, 209, 126, 47, 111, 215, 126, 221, 126, 227, 126, 233, 126, 239, 126, 245, 126, 251, 126, 41, 156, 47, 156, 1, 127, 7, 127, 13, 127, 19, 127, 25, 127, 31, 127, 37, 127, 43, 127, 49, 127, 55, 127, 17, 0, 122, 97, 127, 97, 132, 97, 137, 97, 142, 97, 147, 97, 152, 97, 157, 97, 162, 97, 167, 97, 172, 97, 177, 97, 182, 97, 187, 97, 192, 97, 197, 97, 202, 97, 8, 0, 11, 63, 41, 43, 16, 63, 21, 63, 26, 63, 31, 63, 36, 63, 41, 63, 16, 0, 46, 63, 51, 63, 56, 63, 61, 63, 66, 63, 71, 63, 76, 63, 81, 63, 86, 63, 91, 63, 96, 63, 101, 63, 106, 63, 111, 63, 116, 63, 121, 63, 16, 0, 171, 128, 177, 128, 89, 111, 183, 128, 55, 157, 119, 132, 189, 128, 195, 128, 201, 128, 207, 128, 213, 128, 217, 109, 219, 128, 225, 128, 231, 128, 237, 128, 20, 0, 11, 98, 16, 98, 21, 98, 26, 98, 31, 98, 36, 98, 41, 98, 46, 98, 51, 98, 56, 98, 61, 98, 66, 98, 71, 98, 76, 98, 81, 98, 86, 98, 91, 98, 96, 98, 126, 63, 131, 63, 16, 0, 95, 111, 101, 111, 107, 111, 113, 111, 119, 111, 125, 111, 131, 111, 31, 109, 137, 111, 143, 111, 243, 128, 125, 132, 121, 157, 127, 157, 249, 128, 255, 128, 18, 0, 15, 9, 127, 31, 131, 31, 135, 31, 139, 31, 143, 31, 147, 31, 151, 31, 155, 31, 101, 98, 106, 98, 111, 98, 116, 98, 121, 98, 126, 98, 131, 98, 136, 63, 5, 129, 8, 0, 141, 63, 146, 63, 151, 63, 156, 63, 161, 63, 166, 63, 171, 63, 176, 63, 42, 0, 181, 63, 186, 63, 51, 43, 63, 9, 191, 63, 196, 63, 201, 63, 206, 63, 211, 63, 216, 63, 221, 63, 226, 63, 231, 63, 136, 98, 141, 98, 125, 44, 236, 63, 241, 63, 246, 63, 251, 63, 0, 64, 5, 64, 10, 64, 15, 64, 20, 64, 25, 64, 30, 64, 35, 64, 40, 64, 45, 64, 50, 64, 55, 64, 60, 64, 65, 64, 70, 64, 75, 64, 146, 98, 151, 98, 80, 64, 85, 64, 90, 64, 95, 64, 20, 0, 156, 98, 196, 98, 201, 98, 206, 98, 211, 98, 216, 98, 221, 98, 226, 98, 231, 98, 236, 98, 241, 98, 246, 98, 251, 98, 0, 99, 5, 99, 10, 99, 15, 99, 20, 99, 100, 64, 130, 44, 83, 0, 105, 64, 110, 64, 11, 129, 17, 129, 149, 111, 23, 129, 56, 43, 29, 129, 35, 129, 61, 43, 115, 64, 41, 129, 47, 129, 120, 64, 66, 43, 125, 64, 155, 111, 53, 129, 161, 111, 50, 99, 55, 99, 167, 111, 59, 129, 65, 129, 173, 111, 71, 129, 77, 129, 83, 129, 60, 99, 65, 99, 89, 129, 130, 64, 95, 129, 135, 64, 101, 129, 107, 129, 113, 129, 119, 129, 125, 129, 140, 64, 145, 64, 131, 129, 137, 129, 143, 129, 149, 129, 155, 129, 161, 129, 167, 129, 173, 129, 179, 111, 179, 129, 185, 129, 191, 129, 197, 129, 203, 129, 209, 129, 215, 129, 71, 43, 241, 109, 221, 129, 237, 107, 150, 64, 247, 109, 227, 129, 233, 129, 185, 111, 76, 43, 239, 129, 245, 129, 251, 129, 1, 130, 155, 64, 7, 130, 160, 64, 13, 130, 19, 130, 25, 130, 31, 130, 191, 111, 37, 130, 243, 107, 43, 130, 215, 108, 5, 0, 70, 99, 75, 99, 80, 99, 85, 99, 90, 99, 13, 0, 95, 99, 100, 99, 105, 99, 110, 99, 115, 99, 120, 99, 125, 99, 130, 99, 135, 99, 140, 99, 145, 99, 150, 99, 165, 64, 10, 0, 170, 64, 175, 64, 180, 64, 185, 64, 190, 64, 155, 99, 160, 99, 195, 64, 200, 64, 205, 64, 25, 0, 165, 99, 170, 99, 175, 99, 180, 99, 185, 99, 190, 99, 195, 99, 200, 99, 205, 99, 210, 99, 215, 99, 220, 99, 225, 99, 230, 99, 235, 99, 240, 99, 245, 99, 250, 99, 255, 99, 4, 100, 9, 100, 14, 100, 19, 100, 24, 100, 29, 100, 18, 0, 210, 64, 215, 64, 220, 64, 225, 64, 81, 43, 230, 64, 235, 64, 86, 43, 240, 64, 245, 64, 250, 64, 255, 64, 4, 65, 9, 65, 91, 43, 14, 65, 19, 65, 24, 65, 21, 0, 34, 100, 39, 100, 29, 65, 142, 43, 27, 33, 205, 111, 1, 111, 19, 41, 31, 33, 35, 33, 67, 9, 39, 33, 59, 110, 43, 33, 47, 33, 7, 111, 51, 33, 147, 43, 129, 41, 55, 33, 59, 33, 5, 0, 44, 100, 49, 100, 54, 100, 59, 100, 64, 100, 3, 0, 34, 65, 39, 65, 44, 65, 212, 0, 143, 159, 149, 159, 155, 159, 161, 159, 167, 159, 173, 159, 179, 159, 185, 159, 191, 159, 197, 159, 203, 159, 209, 159, 215, 159, 221, 159, 227, 159, 233, 159, 239, 159, 245, 159, 251, 159, 1, 160, 7, 160, 13, 160, 19, 160, 25, 160, 31, 160, 37, 160, 43, 160, 49, 160, 55, 160, 61, 160, 67, 160, 73, 160, 79, 160, 85, 160, 91, 160, 97, 160, 103, 160, 109, 160, 115, 160, 121, 160, 127, 160, 133, 160, 139, 160, 145, 160, 151, 160, 157, 160, 163, 160, 169, 160, 175, 160, 181, 160, 187, 160, 193, 160, 199, 160, 205, 160, 211, 160, 217, 160, 223, 160, 229, 160, 235, 160, 241, 160, 247, 160, 253, 160, 3, 161, 9, 161, 15, 161, 21, 161, 27, 161, 33, 161, 39, 161, 45, 161, 51, 161, 57, 161, 63, 161, 69, 161, 75, 161, 81, 161, 87, 161, 93, 161, 99, 161, 105, 161, 111, 161, 117, 161, 123, 161, 129, 161, 135, 161, 141, 161, 147, 161, 153, 161, 159, 161, 165, 161, 171, 161, 177, 161, 183, 161, 189, 161, 195, 161, 201, 161, 207, 161, 213, 161, 219, 161, 225, 161, 231, 161, 237, 161, 243, 161, 249, 161, 255, 161, 5, 162, 11, 162, 17, 162, 23, 162, 29, 162, 35, 162, 131, 132, 41, 162, 47, 162, 53, 162, 211, 112, 59, 162, 65, 162, 71, 162, 137, 132, 77, 162, 83, 162, 89, 162, 217, 112, 95, 162, 101, 162, 107, 162, 113, 162, 119, 162, 125, 162, 131, 162, 223, 112, 137, 162, 143, 162, 149, 162, 229, 112, 155, 162, 161, 162, 167, 162, 143, 132, 173, 162, 149, 132, 155, 132, 235, 112, 179, 162, 185, 162, 161, 132, 191, 162, 197, 162, 167, 132, 241, 112, 203, 162, 173, 132, 179, 132, 235, 109, 209, 162, 185, 132, 215, 162, 221, 162, 227, 162, 191, 132, 233, 162, 239, 162, 245, 162, 197, 132, 251, 162, 1, 163, 7, 163, 247, 112, 13, 163, 19, 163, 25, 163, 253, 112, 203, 132, 31, 163, 37, 163, 209, 132, 43, 163, 3, 113, 49, 163, 55, 163, 61, 163, 215, 132, 67, 163, 73, 163, 79, 163, 9, 113, 85, 163, 91, 163, 15, 113, 21, 113, 97, 163, 103, 163, 109, 163, 221, 132, 115, 163, 121, 163, 127, 163, 133, 163, 139, 163, 145, 163, 151, 163, 157, 163, 163, 163, 169, 163, 175, 163, 181, 163, 187, 163, 193, 163, 199, 163, 205, 163, 211, 163, 8, 0, 49, 65, 54, 65, 59, 65, 64, 65, 96, 43, 69, 65, 74, 65, 101, 43, 5, 0, 27, 109, 71, 9, 106, 43, 75, 9, 79, 9, 9, 0, 69, 100, 74, 100, 79, 100, 84, 100, 89, 100, 94, 100, 99, 100, 104, 100, 109, 100, 14, 0, 79, 65, 84, 65, 89, 65, 94, 65, 114, 100, 119, 100, 99, 65, 111, 43, 104, 65, 109, 65, 114, 65, 119, 65, 124, 65, 129, 65, 18, 0, 134, 65, 139, 65, 144, 65, 116, 43, 149, 65, 154, 65, 121, 43, 159, 65, 164, 65, 169, 65, 174, 65, 179, 65, 126, 43, 184, 65, 189, 65, 194, 65, 199, 65, 204, 65, 10, 0, 209, 65, 214, 65, 219, 65, 224, 65, 229, 65, 234, 65, 239, 65, 244, 65, 131, 43, 136, 43, 10, 0, 249, 65, 254, 65, 141, 43, 3, 66, 146, 43, 151, 43, 8, 66, 13, 66, 18, 66, 23, 66, 2, 0, 87, 9, 147, 112, 14, 0, 28, 66, 33, 66, 38, 66, 43, 66, 48, 66, 53, 66, 58, 66, 63, 66, 68, 66, 73, 66, 135, 44, 159, 40, 78, 66, 83, 66, 14, 0, 88, 66, 93, 66, 98, 66, 156, 43, 103, 66, 199, 70, 124, 100, 135, 39, 108, 66, 70, 39, 113, 66, 118, 66, 123, 66, 128, 66, 4, 0, 133, 66, 140, 44, 138, 66, 145, 39, 23, 0, 143, 66, 148, 66, 153, 66, 158, 66, 163, 66, 168, 66, 173, 66, 178, 66, 161, 43, 183, 66, 188, 66, 193, 66, 198, 66, 203, 66, 208, 66, 213, 66, 218, 66, 223, 66, 228, 66, 233, 66, 238, 66, 243, 66, 248, 66, 5, 0, 91, 9, 95, 9, 99, 9, 103, 9, 107, 9, 78, 0, 129, 100, 134, 100, 139, 100, 144, 100, 149, 100, 154, 100, 159, 100, 164, 100, 169, 100, 174, 100, 179, 100, 184, 100, 189, 100, 194, 100, 199, 100, 204, 100, 209, 100, 214, 100, 219, 100, 224, 100, 229, 100, 234, 100, 239, 100, 244, 100, 249, 100, 254, 100, 3, 101, 8, 101, 13, 101, 18, 101, 23, 101, 28, 101, 33, 101, 38, 101, 43, 101, 48, 101, 53, 101, 58, 101, 63, 101, 68, 101, 73, 101, 78, 101, 83, 101, 88, 101, 93, 101, 98, 101, 103, 101, 108, 101, 113, 101, 118, 101, 123, 101, 128, 101, 133, 101, 138, 101, 143, 101, 148, 101, 153, 101, 158, 101, 163, 101, 168, 101, 173, 101, 178, 101, 183, 101, 188, 101, 193, 101, 198, 101, 203, 101, 208, 101, 213, 101, 218, 101, 223, 101, 228, 101, 233, 101, 238, 101, 243, 101, 248, 101, 100, 39, 111, 9, 5, 0, 253, 66, 166, 43, 2, 67, 7, 67, 12, 67, 13, 0, 17, 67, 22, 67, 171, 43, 27, 67, 32, 67, 37, 67, 42, 67, 47, 67, 52, 67, 57, 67, 62, 67, 67, 67, 176, 43, 6, 0, 31, 34, 165, 70, 119, 9, 163, 2, 35, 34, 111, 44, 24, 0, 253, 101, 2, 102, 7, 102, 12, 102, 17, 102, 22, 102, 27, 102, 32, 102, 37, 102, 42, 102, 47, 102, 52, 102, 57, 102, 62, 102, 67, 102, 72, 102, 77, 102, 72, 67, 82, 102, 87, 102, 92, 102, 97, 102, 102, 102, 107, 102, 5, 0, 112, 102, 117, 102, 122, 102, 127, 102, 132, 102, 81, 0, 137, 102, 142, 102, 147, 102, 152, 102, 157, 102, 162, 102, 167, 102, 172, 102, 177, 102, 182, 102, 187, 102, 192, 102, 197, 102, 202, 102, 207, 102, 212, 102, 217, 102, 222, 102, 227, 102, 232, 102, 237, 102, 242, 102, 247, 102, 252, 102, 1, 103, 6, 103, 11, 103, 16, 103, 21, 103, 26, 103, 31, 103, 36, 103, 41, 103, 46, 103, 51, 103, 56, 103, 61, 103, 66, 103, 71, 103, 76, 103, 81, 103, 86, 103, 91, 103, 96, 103, 101, 103, 106, 103, 111, 103, 116, 103, 121, 103, 126, 103, 131, 103, 136, 103, 141, 103, 146, 103, 151, 103, 156, 103, 161, 103, 166, 103, 171, 103, 176, 103, 181, 103, 186, 103, 191, 103, 196, 103, 201, 103, 206, 103, 211, 103, 216, 103, 221, 103, 226, 103, 231, 103, 236, 103, 241, 103, 246, 103, 251, 103, 0, 104, 5, 104, 10, 104, 15, 104, 20, 104, 28, 71, 15, 0, 49, 130, 55, 130, 61, 130, 67, 130, 73, 130, 79, 130, 85, 130, 91, 130, 97, 130, 221, 108, 103, 130, 109, 130, 115, 130, 121, 130, 127, 130, 8, 0, 133, 130, 139, 130, 145, 130, 151, 130, 157, 130, 163, 130, 169, 130, 175, 130, 22, 0, 181, 130, 187, 130, 193, 130, 223, 106, 199, 130, 205, 130, 211, 130, 217, 130, 223, 130, 229, 130, 235, 130, 241, 130, 197, 111, 247, 130, 253, 130, 3, 131, 9, 131, 69, 107, 15, 131, 21, 131, 27, 131, 33, 131, 31, 0, 25, 104, 30, 104, 35, 104, 40, 104, 45, 104, 50, 104, 55, 104, 60, 104, 65, 104, 55, 39, 70, 104, 33, 71, 75, 104, 80, 104, 38, 71, 85, 104, 90, 104, 95, 104, 43, 71, 48, 71, 100, 104, 105, 104, 110, 104, 115, 104, 120, 104, 125, 104, 130, 104, 135, 104, 140, 104, 145, 104, 150, 104, 27, 0, 77, 67, 82, 67, 87, 67, 92, 67, 97, 67, 102, 67, 107, 67, 112, 67, 117, 67, 122, 67, 127, 67, 132, 67, 137, 67, 142, 67, 147, 67, 152, 67, 157, 67, 162, 67, 167, 67, 172, 67, 177, 67, 182, 67, 187, 67, 192, 67, 197, 67, 202, 67, 207, 67, 4, 0, 123, 9, 63, 3, 127, 9, 131, 9, 9, 0, 212, 67, 217, 67, 222, 67, 227, 67, 232, 67, 237, 67, 242, 67, 247, 67, 252, 67, 57, 0, 242, 109, 155, 104, 238, 107, 1, 68, 6, 68, 181, 43, 11, 68, 16, 68, 186, 43, 21, 68, 26, 68, 31, 68, 36, 68, 41, 68, 46, 68, 51, 68, 56, 68, 61, 68, 66, 68, 71, 68, 191, 43, 76, 68, 160, 104, 165, 104, 81, 68, 170, 104, 248, 109, 86, 68, 175, 104, 180, 104, 185, 104, 190, 104, 195, 104, 196, 43, 91, 68, 96, 68, 101, 68, 106, 68, 111, 68, 116, 68, 121, 68, 126, 68, 131, 68, 136, 68, 200, 104, 205, 104, 141, 68, 146, 68, 151, 68, 156, 68, 161, 68, 166, 68, 171, 68, 176, 68, 181, 68, 186, 68, 191, 68, 19, 0, 244, 107, 196, 68, 201, 43, 49, 40, 201, 68, 206, 68, 211, 68, 216, 68, 206, 43, 221, 68, 226, 68, 211, 43, 53, 71, 231, 68, 236, 68, 241, 68, 54, 40, 200, 39, 59, 40, 14, 0, 246, 68, 251, 68, 145, 44, 150, 44, 164, 40, 0, 69, 5, 69, 10, 69, 205, 39, 64, 40, 210, 39, 69, 40, 74, 40, 15, 69, 6, 0, 210, 104, 215, 104, 220, 104, 225, 104, 230, 104, 235, 104, 25, 0, 11, 35, 15, 35, 19, 35, 23, 35, 27, 35, 31, 35, 35, 35, 39, 35, 43, 35, 47, 35, 51, 35, 55, 35, 59, 35, 233, 110, 183, 2, 63, 35, 3, 109, 67, 35, 71, 35, 187, 2, 75, 35, 79, 35, 83, 35, 87, 35, 91, 35, 63, 0, 240, 104, 245, 104, 250, 104, 255, 104, 4, 105, 9, 105, 14, 105, 20, 69, 19, 105, 24, 105, 25, 69, 29, 105, 34, 105, 39, 105, 44, 105, 49, 105, 54, 105, 59, 105, 64, 105, 69, 105, 74, 105, 79, 105, 84, 105, 89, 105, 94, 105, 99, 105, 104, 105, 109, 105, 114, 105, 30, 69, 119, 105, 124, 105, 129, 105, 134, 105, 139, 105, 144, 105, 149, 105, 35, 69, 154, 105, 159, 105, 164, 105, 169, 105, 174, 105, 179, 105, 184, 105, 189, 105, 194, 105, 199, 105, 40, 69, 45, 69, 204, 105, 209, 105, 214, 105, 219, 105, 224, 105, 229, 105, 234, 105, 239, 105, 50, 69, 55, 69, 60, 69, 65, 69, 70, 69, 6, 0, 39, 131, 45, 131, 51, 131, 203, 111, 67, 112, 57, 131, 3, 0, 75, 69, 80, 69, 216, 43, 11, 0, 85, 69, 90, 69, 95, 69, 100, 69, 105, 69, 110, 69, 115, 69, 120, 69, 125, 69, 130, 69, 221, 43, 22, 0, 135, 69, 140, 69, 145, 69, 150, 69, 155, 69, 160, 69, 165, 69, 170, 69, 226, 43, 175, 69, 231, 43, 180, 69, 185, 69, 190, 69, 195, 69, 236, 43, 129, 40, 200, 69, 134, 40, 205, 69, 241, 43, 210, 69, 9, 0, 246, 43, 215, 69, 220, 69, 225, 69, 230, 69, 235, 69, 240, 69, 245, 69, 250, 69, 10, 0, 244, 105, 249, 105, 254, 105, 3, 106, 8, 106, 13, 106, 18, 106, 23, 106, 28, 106, 255, 69, 10, 0, 4, 70, 251, 43, 9, 70, 14, 70, 19, 70, 24, 70, 230, 39, 75, 39, 33, 106, 38, 106, 8, 0, 161, 114, 197, 114, 1, 115, 15, 44, 43, 115, 91, 115, 97, 115, 103, 115, 6, 0, 105, 74, 125, 74, 130, 74, 220, 74, 89, 75, 94, 75, 11, 0, 100, 74, 115, 74, 135, 74, 140, 74, 145, 74, 150, 74, 175, 74, 240, 74, 250, 74, 74, 75, 119, 75, 13, 0, 160, 74, 170, 74, 180, 74, 185, 74, 225, 74, 4, 75, 14, 75, 19, 75, 39, 75, 49, 75, 104, 75, 149, 75, 154, 75, 10, 0, 120, 74, 155, 74, 205, 74, 210, 74, 230, 74, 245, 74, 24, 75, 34, 75, 54, 75, 129, 75, 8, 0, 110, 74, 215, 74, 59, 75, 64, 75, 79, 75, 84, 75, 109, 75, 134, 75, 8, 0, 165, 74, 190, 74, 235, 74, 255, 74, 69, 75, 99, 75, 114, 75, 159, 75, 4, 0, 195, 74, 29, 75, 139, 75, 144, 75, 4, 0, 200, 74, 9, 75, 44, 75, 124, 75, 5, 0, 229, 115, 235, 115, 247, 115, 253, 115, 3, 116, 5, 0, 15, 116, 5, 108, 45, 110, 21, 116, 27, 116, 6, 0, 179, 135, 191, 135, 197, 135, 135, 116, 153, 116, 213, 116, 2, 0, 51, 116, 123, 116, 3, 0, 39, 116, 111, 116, 203, 135, 3, 0, 185, 135, 141, 116, 195, 116, 4, 0, 45, 116, 189, 116, 201, 116, 63, 110, 3, 0, 51, 110, 147, 116, 249, 116, 5, 0, 63, 116, 69, 116, 99, 116, 105, 116, 219, 116, 3, 0, 75, 116, 93, 116, 225, 116, 4, 0, 129, 116, 177, 116, 237, 116, 243, 116, 3, 0, 57, 116, 117, 116, 165, 116, 4, 0, 171, 116, 57, 110, 207, 116, 231, 116, 4, 0, 33, 116, 81, 116, 159, 116, 183, 116, 8, 0, 139, 50, 169, 50, 174, 50, 184, 50, 189, 50, 114, 70, 214, 50, 50, 44, 14, 0, 134, 50, 144, 50, 149, 50, 154, 50, 159, 50, 164, 50, 179, 50, 194, 50, 24, 40, 199, 50, 247, 77, 204, 50, 209, 50, 219, 50, 12, 0, 137, 138, 143, 138, 149, 138, 107, 138, 73, 112, 79, 112, 75, 131, 113, 138, 119, 138, 125, 138, 85, 112, 131, 138, 7, 0, 155, 138, 81, 131, 161, 138, 167, 138, 173, 138, 87, 131, 179, 138, 7, 0, 185, 138, 191, 138, 197, 138, 91, 112, 203, 138, 209, 138, 215, 138, 3, 0, 221, 138, 227, 138, 233, 138, 7, 0, 239, 138, 97, 112, 245, 138, 251, 138, 1, 139, 103, 112, 7, 139, 4, 0, 13, 139, 109, 112, 19, 139, 25, 139, 5, 0, 31, 139, 37, 139, 43, 139, 115, 112, 49, 139, 4, 0, 93, 131, 99, 131, 121, 112, 105, 131, 5, 0, 55, 139, 61, 139, 67, 139, 127, 112, 73, 139, 7, 0, 79, 139, 111, 131, 133, 112, 85, 139, 91, 139, 139, 112, 97, 139, 5, 0, 103, 139, 109, 139, 115, 139, 121, 139, 127, 139, 4, 0, 133, 139, 139, 139, 145, 139, 117, 131, 6, 0, 151, 139, 157, 139, 163, 139, 123, 131, 169, 139, 175, 139, 4, 0, 176, 78, 216, 78, 241, 78, 246, 78, 4, 0, 171, 78, 221, 78, 20, 79, 25, 79, 3, 0, 191, 78, 236, 78, 15, 79, 3, 0, 211, 78, 10, 79, 30, 79, 2, 0, 181, 78, 0, 79, 4, 0, 161, 78, 166, 78, 196, 78, 226, 78, 3, 0, 5, 79, 35, 79, 40, 79, 2, 0, 151, 78, 50, 79, 4, 0, 156, 78, 231, 78, 251, 78, 45, 79, 3, 0, 186, 78, 201, 78, 206, 78, 8, 0, 253, 51, 89, 80, 37, 52, 52, 52, 135, 40, 81, 44, 82, 52, 132, 52, 3, 0, 138, 41, 139, 80, 100, 40, 9, 0, 7, 52, 22, 52, 67, 52, 195, 7, 127, 52, 137, 52, 142, 52, 157, 52, 172, 52, 5, 0, 248, 51, 109, 80, 124, 80, 57, 52, 152, 52, 2, 0, 133, 41, 144, 80, 4, 0, 187, 7, 47, 52, 90, 40, 95, 40, 2, 0, 12, 52, 94, 80, 4, 0, 191, 7, 77, 52, 107, 52, 117, 52, 3, 0, 17, 52, 147, 52, 167, 52, 3, 0, 130, 40, 114, 80, 199, 7, 5, 0, 33, 81, 38, 81, 48, 81, 53, 81, 58, 81, 3, 0, 8, 81, 13, 81, 18, 81, 3, 0, 254, 80, 3, 81, 23, 81, 3, 0, 249, 80, 28, 81, 43, 81, 12, 0, 63, 81, 73, 81, 93, 81, 133, 81, 183, 81, 253, 81, 17, 82, 22, 82, 122, 82, 152, 82, 172, 82, 177, 82, 8, 0, 158, 81, 178, 81, 2, 82, 97, 82, 157, 82, 162, 82, 242, 82, 65, 39, 4, 0, 163, 81, 198, 81, 238, 81, 87, 82, 2, 0, 203, 81, 208, 81, 6, 0, 148, 81, 193, 81, 243, 81, 248, 81, 12, 82, 32, 82, 10, 0, 98, 81, 108, 81, 62, 82, 67, 82, 77, 82, 82, 82, 92, 82, 142, 82, 147, 82, 237, 82, 5, 0, 68, 81, 102, 82, 107, 82, 117, 82, 207, 82, 8, 0, 182, 82, 192, 82, 197, 82, 247, 82, 245, 39, 252, 82, 254, 70, 148, 41, 12, 0, 138, 81, 143, 81, 153, 81, 168, 81, 173, 81, 228, 81, 7, 82, 42, 82, 127, 82, 202, 82, 227, 82, 232, 82, 5, 0, 128, 81, 188, 81, 57, 82, 112, 82, 187, 82, 13, 0, 103, 81, 113, 81, 118, 81, 213, 81, 218, 81, 223, 81, 233, 81, 37, 82, 47, 82, 132, 82, 137, 82, 212, 82, 217, 82, 6, 0, 78, 81, 83, 81, 88, 81, 123, 81, 222, 82, 244, 70, 5, 0, 27, 82, 52, 82, 72, 82, 167, 82, 249, 70, 152, 0, 151, 118, 157, 118, 135, 131, 157, 142, 163, 118, 169, 118, 193, 118, 199, 118, 205, 118, 211, 118, 217, 118, 223, 118, 229, 118, 235, 118, 181, 142, 187, 142, 241, 118, 247, 118, 253, 118, 3, 119, 15, 119, 33, 119, 39, 119, 105, 110, 51, 119, 57, 119, 63, 119, 69, 119, 81, 119, 93, 119, 117, 110, 105, 119, 123, 110, 223, 142, 117, 119, 129, 110, 129, 119, 135, 119, 153, 131, 177, 119, 171, 107, 183, 119, 207, 119, 225, 119, 147, 110, 253, 142, 3, 143, 231, 119, 237, 119, 243, 119, 249, 119, 5, 120, 153, 110, 11, 120, 17, 120, 23, 120, 29, 120, 35, 120, 159, 110, 41, 120, 53, 120, 59, 120, 165, 110, 65, 120, 71, 120, 77, 120, 83, 120, 95, 120, 171, 110, 101, 120, 107, 120, 113, 120, 119, 120, 125, 120, 177, 110, 131, 120, 137, 120, 143, 120, 149, 120, 161, 120, 179, 120, 203, 120, 209, 120, 215, 120, 221, 120, 227, 120, 245, 120, 1, 121, 7, 121, 19, 121, 25, 121, 189, 110, 37, 121, 43, 121, 55, 121, 73, 121, 79, 121, 85, 121, 97, 121, 103, 121, 109, 121, 195, 110, 121, 121, 133, 121, 139, 121, 145, 121, 151, 121, 175, 121, 181, 121, 213, 107, 187, 121, 193, 121, 199, 121, 205, 121, 63, 143, 69, 143, 217, 121, 75, 143, 81, 143, 223, 121, 229, 121, 235, 121, 99, 143, 241, 121, 145, 112, 105, 143, 201, 110, 253, 121, 3, 122, 9, 122, 15, 122, 21, 122, 27, 122, 39, 122, 45, 122, 57, 122, 63, 122, 69, 122, 75, 122, 81, 122, 87, 122, 93, 122, 175, 109, 105, 122, 111, 122, 213, 110, 117, 122, 123, 122, 129, 122, 141, 122, 147, 122, 177, 107, 11, 0, 121, 142, 133, 118, 139, 118, 175, 118, 21, 119, 123, 119, 201, 119, 89, 120, 155, 120, 183, 110, 251, 120, 32, 0, 127, 142, 133, 142, 121, 118, 145, 118, 45, 119, 111, 119, 229, 142, 147, 119, 153, 119, 159, 119, 165, 119, 171, 119, 135, 110, 189, 119, 195, 119, 219, 119, 255, 119, 47, 120, 167, 120, 185, 120, 197, 120, 239, 120, 49, 121, 67, 121, 127, 121, 157, 121, 163, 121, 211, 121, 93, 143, 207, 110, 167, 108, 183, 107, 22, 0, 127, 118, 181, 118, 187, 118, 9, 119, 27, 119, 111, 110, 75, 119, 87, 119, 99, 119, 141, 110, 233, 108, 173, 120, 191, 120, 13, 121, 31, 121, 61, 121, 91, 121, 115, 121, 247, 121, 33, 122, 161, 108, 135, 122, 5, 0, 134, 70, 146, 53, 168, 41, 181, 53, 61, 83, 5, 0, 151, 53, 171, 53, 176, 53, 51, 83, 211, 53, 4, 0, 156, 53, 161, 53, 163, 41, 191, 53, 5, 0, 41, 83, 56, 83, 66, 83, 76, 83, 206, 53, 5, 0, 46, 83, 173, 41, 196, 53, 81, 83, 216, 53, 3, 0, 166, 53, 86, 83, 183, 41, 6, 0, 36, 83, 186, 53, 178, 41, 71, 83, 201, 53, 221, 53, 4, 0, 241, 53, 34, 40, 246, 53, 251, 53, 3, 0, 226, 53, 231, 53, 236, 53, 2, 0, 100, 54, 105, 54, 3, 0, 3, 71, 203, 41, 208, 41, 3, 0, 101, 83, 110, 54, 115, 54, 6, 0, 134, 55, 10, 84, 139, 55, 144, 55, 149, 55, 194, 55, 5, 0, 20, 84, 154, 55, 30, 84, 35, 84, 40, 84, 2, 0, 164, 55, 174, 55, 3, 0, 251, 83, 179, 55, 45, 84, 2, 0, 55, 84, 60, 84, 6, 0, 7, 42, 65, 84, 189, 55, 75, 84, 80, 84, 90, 84, 10, 0, 124, 55, 0, 84, 129, 55, 5, 84, 25, 84, 159, 55, 184, 55, 144, 70, 85, 84, 95, 84, 5, 0, 59, 8, 52, 42, 105, 84, 229, 55, 104, 40, 12, 0, 27, 42, 55, 8, 32, 42, 37, 42, 204, 55, 209, 55, 214, 55, 219, 55, 224, 55, 244, 55, 110, 84, 115, 84, 6, 0, 22, 42, 159, 70, 42, 42, 47, 42, 234, 55, 239, 55, 3, 0, 154, 70, 199, 55, 100, 84, 8, 0, 132, 174, 72, 57, 92, 57, 104, 85, 207, 57, 244, 85, 8, 86, 13, 86, 12, 0, 87, 57, 79, 85, 109, 85, 114, 85, 172, 57, 187, 57, 164, 70, 197, 57, 149, 85, 199, 85, 224, 85, 174, 70, 2, 0, 102, 57, 92, 42, 7, 0, 97, 57, 174, 85, 214, 85, 36, 58, 41, 58, 46, 58, 18, 86, 4, 0, 147, 57, 157, 57, 229, 85, 21, 58, 9, 0, 8, 71, 127, 57, 132, 57, 154, 85, 169, 70, 232, 57, 242, 57, 252, 57, 209, 85, 10, 0, 69, 85, 77, 42, 152, 57, 182, 57, 139, 85, 110, 44, 227, 57, 184, 85, 189, 85, 16, 58, 2, 0, 222, 57, 254, 85, 5, 0, 49, 85, 59, 85, 142, 57, 144, 85, 194, 85, 5, 0, 82, 42, 134, 85, 6, 58, 204, 85, 23, 86, 4, 0, 64, 85, 99, 85, 179, 85, 31, 58, 2, 0, 94, 85, 162, 57, 5, 0, 77, 57, 72, 42, 107, 57, 202, 57, 11, 58, 6, 0, 82, 57, 74, 85, 84, 85, 137, 57, 177, 57, 26, 58, 2, 0, 159, 85, 237, 57, 5, 0, 119, 85, 117, 57, 167, 57, 247, 57, 51, 58, 9, 0, 67, 57, 112, 57, 124, 85, 122, 57, 192, 57, 87, 42, 1, 58, 234, 85, 249, 85, 5, 0, 89, 85, 212, 57, 217, 57, 219, 85, 239, 85, 9, 0, 136, 88, 141, 88, 146, 88, 161, 88, 171, 88, 176, 88, 186, 88, 196, 88, 126, 58, 5, 0, 151, 88, 156, 88, 166, 88, 181, 88, 191, 88, 3, 0, 24, 90, 44, 90, 54, 90, 3, 0, 205, 89, 235, 89, 19, 90, 3, 0, 210, 89, 245, 89, 250, 89, 5, 0, 215, 89, 220, 89, 255, 89, 34, 90, 69, 90, 3, 0, 225, 89, 240, 89, 4, 90, 2, 0, 14, 90, 59, 90, 2, 0, 230, 89, 49, 90, 2, 0, 39, 90, 64, 90, 2, 0, 9, 90, 29, 90, 8, 0, 77, 123, 107, 123, 143, 123, 191, 123, 197, 123, 1, 124, 79, 124, 97, 124, 8, 0, 31, 151, 89, 123, 119, 123, 55, 151, 161, 123, 249, 110, 251, 123, 55, 124, 9, 0, 59, 123, 113, 123, 131, 123, 149, 123, 209, 123, 227, 123, 255, 110, 61, 124, 201, 107, 8, 0, 167, 123, 179, 123, 233, 123, 7, 124, 25, 124, 19, 133, 103, 151, 43, 124, 5, 0, 47, 123, 195, 107, 125, 123, 67, 151, 73, 151, 9, 0, 37, 151, 65, 123, 71, 123, 237, 110, 155, 123, 203, 123, 221, 123, 5, 111, 31, 124, 8, 0, 83, 123, 101, 123, 137, 123, 173, 123, 225, 107, 13, 124, 19, 124, 109, 124, 5, 0, 95, 123, 215, 123, 239, 123, 85, 124, 115, 124, 5, 0, 29, 123, 243, 110, 67, 124, 73, 124, 91, 124, 4, 0, 41, 123, 49, 151, 37, 124, 103, 124, 4, 0, 53, 123, 239, 108, 185, 123, 49, 124, 2, 0, 35, 123, 245, 123, 14, 0, 199, 151, 17, 111, 127, 124, 133, 124, 139, 124, 245, 108, 145, 124, 151, 124, 157, 124, 185, 108, 193, 124, 199, 124, 205, 124, 211, 124, 10, 0, 193, 151, 11, 111, 121, 124, 163, 124, 169, 124, 175, 124, 181, 124, 187, 124, 191, 108, 217, 124, 9, 0, 28, 61, 33, 61, 18, 71, 207, 42, 43, 61, 212, 42, 217, 42, 222, 42, 227, 42, 6, 0, 23, 61, 184, 70, 38, 61, 58, 61, 140, 39, 85, 39, 13, 0, 3, 61, 8, 61, 13, 61, 18, 61, 95, 39, 23, 71, 48, 61, 53, 61, 225, 39, 63, 61, 68, 61, 120, 44, 73, 61, 3, 0, 157, 62, 172, 62, 16, 43, 3, 0, 162, 62, 177, 62, 21, 43, 3, 0, 167, 62, 124, 40, 182, 62, 3, 0, 11, 43, 192, 62, 26, 43, 2, 0, 187, 62, 31, 43, 4, 0, 199, 127, 77, 111, 235, 127, 107, 156, 5, 0, 139, 127, 145, 127, 83, 111, 63, 128, 81, 128, 7, 0, 103, 127, 109, 127, 59, 156, 57, 128, 101, 156, 231, 107, 159, 128, 6, 0, 85, 127, 71, 156, 77, 156, 83, 156, 253, 127, 123, 128, 6, 0, 79, 127, 91, 127, 151, 127, 59, 111, 71, 111, 39, 128, 4, 0, 13, 109, 19, 109, 45, 128, 111, 128, 6, 0, 127, 127, 193, 127, 229, 127, 51, 128, 25, 109, 147, 128, 4, 0, 115, 127, 153, 128, 61, 112, 165, 128, 4, 0, 53, 156, 65, 156, 21, 128, 27, 128, 7, 0, 163, 127, 53, 111, 169, 127, 175, 127, 187, 127, 99, 128, 105, 128, 3, 0, 217, 127, 33, 128, 55, 112, 5, 0, 67, 127, 73, 127, 181, 127, 129, 128, 135, 128, 4, 0, 223, 127, 247, 127, 117, 128, 141, 128, 6, 0, 61, 127, 97, 127, 121, 127, 65, 111, 205, 127, 15, 128, 5, 0, 133, 127, 157, 127, 211, 127, 75, 128, 87, 128, 5, 0, 241, 127, 3, 128, 9, 128, 69, 128, 93, 128, 5, 0, 25, 99, 30, 99, 35, 99, 40, 99, 45, 99, 7, 0, 161, 98, 166, 98, 171, 98, 176, 98, 181, 98, 186, 98, 191, 98, 26, 0, 69, 167, 75, 167, 81, 167, 87, 167, 93, 167, 99, 167, 105, 167, 111, 167, 117, 167, 123, 167, 129, 167, 135, 167, 141, 167, 147, 167, 153, 167, 159, 167, 165, 167, 171, 167, 177, 167, 183, 167, 189, 167, 195, 167, 201, 167, 207, 167, 213, 167, 219, 167, 36, 0, 225, 167, 231, 167, 237, 167, 243, 167, 249, 167, 255, 167, 5, 168, 11, 168, 17, 168, 23, 168, 29, 168, 35, 168, 41, 168, 47, 168, 53, 168, 59, 168, 65, 168, 71, 168, 77, 168, 83, 168, 89, 168, 95, 168, 101, 168, 107, 168, 113, 168, 119, 168, 125, 168, 131, 168, 137, 168, 227, 132, 233, 132, 27, 113, 239, 132, 245, 132, 143, 168, 149, 168, 34, 0, 155, 168, 161, 168, 167, 168, 251, 132, 173, 168, 179, 168, 185, 168, 1, 133, 191, 168, 197, 168, 203, 168, 209, 168, 215, 168, 221, 168, 227, 168, 233, 168, 239, 168, 245, 168, 251, 168, 33, 113, 1, 169, 7, 169, 13, 169, 19, 169, 25, 169, 31, 169, 37, 169, 39, 113, 43, 169, 49, 169, 55, 169, 45, 113, 61, 169, 7, 133, 32, 0, 67, 169, 73, 169, 79, 169, 51, 113, 85, 169, 91, 169, 97, 169, 57, 113, 103, 169, 63, 113, 109, 169, 115, 169, 121, 169, 69, 113, 127, 169, 133, 169, 75, 113, 81, 113, 87, 113, 139, 169, 145, 169, 151, 169, 157, 169, 163, 169, 169, 169, 175, 169, 181, 169, 13, 133, 187, 169, 93, 113, 193, 169, 199, 169, 5, 0, 177, 112, 97, 111, 127, 111, 151, 132, 153, 112, 5, 0, 191, 131, 189, 109, 165, 112, 159, 112, 171, 112, 16, 0, 92, 108, 156, 107, 210, 107, 151, 0, 104, 108, 148, 109, 162, 107, 242, 111, 48, 2, 140, 2, 32, 108, 146, 108, 168, 107, 232, 1, 172, 109, 158, 108, 5, 0, 25, 132, 31, 132, 195, 112, 201, 112, 169, 0, 17, 0, 43, 0, 70, 0, 25, 0, 200, 108, 202, 106, 26, 108, 116, 1, 206, 108, 118, 109, 40, 0, 64, 2, 128, 1, 100, 106, 28, 112, 36, 107, 44, 1, 172, 1, 8, 0, 98, 108, 110, 108, 122, 108, 160, 109, 128, 108, 134, 108, 140, 108, 152, 108, 22, 0, 46, 0, 28, 0, 232, 106, 7, 0, 254, 111, 94, 106, 152, 1, 108, 107, 0, 107, 114, 107, 136, 106, 252, 1, 12, 107, 144, 107, 220, 1, 56, 108, 104, 2, 148, 106, 82, 106, 192, 1, 48, 107, 54, 107, 9, 0, 10, 0, 49, 0, 112, 206, 82, 0, 68, 1, 18, 107, 140, 1, 154, 0, 100, 2, 9, 0, 60, 107, 84, 1, 228, 1, 60, 2, 90, 107, 68, 2, 66, 107, 232, 109, 180, 1, 5, 0, 14, 108, 44, 108, 214, 106, 62, 108, 70, 106, 4, 0, 201, 109, 195, 109, 221, 107, 183, 109, 5, 0, 218, 111, 178, 106, 160, 0, 128, 2, 133, 199, 28, 0, 46, 109, 52, 109, 100, 0, 64, 109, 145, 0, 106, 0, 70, 109, 154, 109, 112, 0, 20, 108, 116, 108, 82, 109, 124, 2, 73, 0, 102, 107, 106, 109, 4, 112, 112, 109, 4, 2, 163, 0, 10, 112, 166, 109, 130, 109, 34, 112, 74, 108, 86, 108, 142, 109, 207, 210, 8, 0, 48, 1, 144, 199, 121, 0, 168, 1, 244, 1, 56, 2, 226, 109, 175, 0, 9, 0, 108, 1, 160, 1, 109, 0, 46, 106, 52, 1, 148, 1, 133, 0, 188, 1, 76, 106, 11, 0, 44, 2, 4, 0, 124, 0, 28, 2, 248, 1, 6, 107, 64, 106, 30, 107, 76, 1, 212, 108, 196, 1, 17, 0, 176, 1, 88, 2, 156, 1, 50, 192, 88, 109, 144, 1, 96, 1, 104, 1, 130, 0, 0, 2, 208, 106, 64, 1, 172, 106, 142, 106, 76, 2, 224, 1, 61, 0, 6, 0, 58, 109, 76, 109, 230, 111, 116, 2, 16, 112, 126, 107, 5, 0, 238, 106, 136, 0, 132, 107, 124, 109, 42, 107, 8, 0, 244, 106, 115, 0, 38, 108, 85, 0, 136, 2, 120, 107, 138, 107, 36, 2, 10, 0, 97, 0, 224, 111, 50, 108, 204, 1, 22, 112, 136, 109, 68, 108, 80, 108, 136, 1, 150, 107, 5, 0, 91, 3, 157, 132, 183, 112, 233, 131, 239, 131, 5, 0, 32, 2, 91, 0, 208, 1, 37, 0, 94, 0, 18, 0, 67, 0, 52, 2, 212, 1, 103, 0, 184, 106, 52, 0, 96, 107, 34, 0, 76, 0, 79, 0, 236, 1, 240, 1, 166, 0, 88, 0, 72, 2, 58, 0, 19, 0, 64, 0, 4, 0, 175, 132, 181, 132, 169, 132, 189, 112, 10, 0, 22, 0, 216, 1, 16, 0, 31, 0, 84, 2, 88, 1, 72, 1, 40, 1, 80, 1, 120, 1, 16, 0, 248, 111, 94, 109, 100, 109, 142, 0, 190, 106, 92, 1, 196, 106, 80, 2, 88, 106, 250, 106, 118, 0, 124, 1, 132, 1, 55, 0, 24, 107, 40, 2, 9, 0, 124, 106, 130, 106, 160, 106, 58, 106, 118, 106, 184, 1, 166, 106, 127, 0, 112, 106, 5, 0, 139, 0, 76, 3, 181, 0, 236, 111, 112, 2, 3, 0, 144, 2, 184, 0, 80, 3, 2, 0, 173, 131, 89, 113, 2, 0, 148, 0, 132, 2, 2, 0, 108, 2, 16, 2, 2, 0, 12, 2, 8, 2, 2, 0, 148, 2, 157, 0, 3, 0, 201, 109, 195, 109, 221, 107, 4, 0, 189, 109, 165, 112, 159, 112, 171, 112, 3, 0, 195, 109, 221, 107, 183, 109, 28, 0, 124, 106, 130, 106, 184, 106, 16, 0, 58, 106, 190, 106, 92, 1, 50, 192, 196, 106, 118, 106, 88, 106, 144, 1, 96, 1, 31, 0, 250, 106, 104, 1, 124, 1, 166, 106, 132, 1, 208, 106, 112, 106, 88, 1, 172, 106, 24, 107, 142, 106, 80, 1, 22, 0, 72, 1, 19, 0, 124, 106, 130, 106, 184, 106, 58, 106, 92, 1, 50, 192, 196, 106, 118, 106, 144, 1, 250, 106, 104, 1, 124, 1, 166, 106, 132, 1, 208, 106, 112, 106, 172, 106, 142, 106, 80, 1, 193, 0, 176, 1, 67, 0, 108, 1, 46, 109, 88, 2, 52, 2, 60, 107, 92, 108, 124, 106, 58, 109, 212, 1, 156, 1, 64, 109, 160, 1, 130, 106, 43, 0, 22, 0, 103, 0, 46, 0, 70, 0, 44, 2, 156, 107, 210, 107, 70, 109, 109, 0, 14, 108, 216, 1, 98, 108, 178, 106, 84, 1, 228, 1, 60, 2, 160, 106, 25, 0, 104, 108, 90, 107, 48, 1, 148, 109, 110, 108, 154, 109, 200, 108, 184, 106, 16, 0, 58, 106, 28, 0, 190, 106, 20, 108, 116, 108, 10, 0, 162, 107, 92, 1, 49, 0, 232, 106, 50, 192, 7, 0, 196, 106, 238, 106, 244, 106, 118, 106, 68, 2, 88, 106, 82, 109, 52, 0, 202, 106, 26, 108, 116, 1, 66, 107, 144, 1, 122, 108, 206, 108, 32, 108, 160, 109, 96, 1, 73, 0, 31, 0, 4, 0, 250, 106, 96, 107, 46, 106, 34, 0, 52, 1, 118, 0, 104, 1, 102, 107, 76, 0, 121, 0, 94, 106, 208, 1, 124, 0, 38, 108, 152, 1, 106, 109, 168, 1, 244, 1, 79, 0, 37, 0, 28, 2, 236, 1, 112, 109, 184, 1, 148, 1, 118, 109, 44, 108, 124, 1, 166, 106, 132, 1, 82, 0, 68, 1, 127, 0, 84, 2, 130, 0, 108, 107, 85, 0, 0, 2, 40, 0, 248, 1, 56, 2, 64, 2, 208, 106, 0, 107, 133, 0, 128, 108, 114, 107, 6, 107, 136, 106, 214, 106, 128, 1, 100, 106, 134, 108, 112, 106, 55, 0, 120, 107, 188, 1, 126, 107, 240, 1, 140, 108, 146, 108, 132, 107, 64, 106, 76, 106, 88, 1, 172, 106, 138, 107, 168, 107, 88, 0, 72, 1, 64, 1, 40, 1, 12, 107, 72, 2, 124, 109, 144, 107, 18, 107, 24, 107, 30, 107, 142, 106, 80, 1, 36, 107, 76, 2, 44, 1, 220, 1, 232, 1, 56, 108, 232, 109, 152, 108, 58, 0, 62, 108, 180, 1, 172, 1, 76, 1, 91, 0, 212, 108, 32, 2, 140, 1, 68, 108, 19, 0, 74, 108, 80, 108, 148, 106, 120, 1, 82, 106, 133, 199, 172, 109, 94, 0, 86, 108, 158, 108, 196, 1, 42, 107, 150, 107, 64, 0, 70, 106, 48, 107, 54, 107, 2, 0, 63, 39, 107, 174, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 4, 0, 63, 39, 28, 1, 1, 0, 107, 174, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 3, 0, 107, 174, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 2, 0, 107, 174, 63, 39, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 3, 0, 63, 39, 1, 0, 107, 174, 3, 0, 63, 39, 1, 0, 107, 174, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 3, 0, 107, 174, 1, 0, 63, 39, 2, 0, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 3, 0, 63, 39, 1, 0, 107, 174, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 2, 0, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 28, 1, 1, 0, 107, 174, 63, 39, 4, 0, 63, 39, 1, 0, 107, 174, 28, 1, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 2, 0, 107, 174, 63, 39, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 2, 0, 1, 0, 63, 39, 3, 0, 63, 39, 107, 174, 1, 0, 2, 0, 63, 39, 1, 0, 3, 0, 63, 39, 1, 0, 107, 174, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 3, 0, 63, 39, 107, 174, 1, 0, 4, 0, 1, 0, 28, 1, 107, 174, 63, 39, 4, 0, 107, 174, 63, 39, 1, 0, 28, 1, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 63, 39, 28, 1, 1, 0, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 2, 0, 107, 174, 63, 39, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 107, 174, 63, 39, 1, 0, 28, 1, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 63, 39, 1, 0, 107, 174, 28, 1, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 2, 0, 107, 174, 63, 39, 4, 0, 63, 39, 1, 0, 107, 174, 28, 1, 4, 0, 1, 0, 28, 1, 107, 174, 63, 39, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 3, 0, 63, 39, 1, 0, 107, 174, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 2, 0, 107, 174, 63, 39, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 1, 0, 107, 174, 28, 1, 4, 0, 107, 174, 63, 39, 28, 1, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 1, 0, 28, 1, 107, 174, 63, 39, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 2, 0, 107, 174, 63, 39, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 2, 0, 1, 0, 63, 39, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 3, 0, 63, 39, 107, 174, 61, 71, 4, 0, 1, 0, 28, 1, 63, 39, 107, 174, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 1, 0, 107, 174, 63, 39, 28, 1, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 107, 174, 63, 39, 1, 0, 28, 1, 4, 0, 107, 174, 63, 39, 1, 0, 28, 1, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 28, 1, 1, 0, 107, 174, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 3, 0, 63, 39, 1, 0, 107, 174, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 2, 0, 107, 174, 63, 39, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 3, 0, 63, 39, 107, 174, 1, 0, 4, 0, 63, 39, 1, 0, 28, 1, 107, 174, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 2, 0, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 3, 0, 63, 39, 1, 0, 107, 174, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 4, 0, 1, 0, 28, 1, 63, 39, 107, 174, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 1, 0, 28, 1, 107, 174, 63, 39, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 28, 1, 1, 0, 107, 174, 63, 39, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 63, 39, 1, 0, 28, 1, 2, 0, 63, 39, 1, 0, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 2, 0, 63, 39, 1, 0, 3, 0, 63, 39, 107, 174, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 4, 0, 107, 174, 63, 39, 1, 0, 28, 1, 4, 0, 63, 39, 1, 0, 107, 174, 28, 1, 3, 0, 63, 39, 107, 174, 1, 0, 2, 0, 107, 174, 63, 39, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 3, 0, 107, 174, 1, 0, 63, 39, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 63, 39, 1, 0, 28, 1, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 2, 0, 63, 39, 1, 0, 2, 0, 107, 174, 63, 39, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 2, 0, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 3, 0, 63, 39, 1, 0, 107, 174, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 3, 0, 63, 39, 107, 174, 1, 0, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 2, 0, 63, 39, 1, 0, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 2, 0, 240, 22, 248, 22, 63, 39, 63, 39, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 3, 0, 63, 39, 107, 174, 1, 0, 3, 0, 63, 39, 107, 174, 1, 0, 2, 0, 107, 174, 63, 39, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 3, 0, 107, 174, 63, 39, 1, 0, 3, 0, 63, 39, 107, 174, 1, 0, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 1, 0, 28, 1, 107, 174, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 2, 0, 107, 174, 63, 39, 2, 0, 63, 39, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 1, 0, 28, 1, 107, 174, 63, 39, 4, 0, 1, 0, 28, 1, 63, 39, 107, 174, 3, 0, 63, 39, 1, 0, 107, 174, 4, 0, 1, 0, 28, 1, 63, 39, 107, 174, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 3, 0, 63, 39, 1, 0, 107, 174, 3, 0, 63, 39, 107, 174, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 107, 174, 63, 39, 1, 0, 28, 1, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 2, 0, 107, 174, 63, 39, 2, 0, 63, 39, 1, 0, 2, 0, 107, 174, 63, 39, 3, 0, 63, 39, 1, 0, 107, 174, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 2, 0, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 28, 1, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 107, 174, 1, 0, 28, 1, 63, 39, 3, 0, 63, 39, 107, 174, 1, 0, 4, 0, 107, 174, 28, 1, 63, 39, 1, 0, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 4, 0, 63, 39, 107, 174, 1, 0, 28, 1, 3, 0, 63, 39, 107, 174, 1, 0, 3, 0, 63, 39, 107, 174, 1, 0, 4, 0, 1, 0, 28, 1, 107, 174, 63, 39, 3, 0, 1, 0, 107, 174, 63, 39, 3, 0, 63, 39, 107, 174, 1, 0, 3, 0, 63, 39, 107, 174, 1, 0, 3, 0, 1, 0, 107, 174, 63, 39, 3, 0, 1, 0, 107, 174, 63, 39, 4, 0, 1, 0, 28, 1, 107, 174, 63, 39, 4, 0, 1, 0, 28, 1, 107, 174, 63, 39, 4, 0, 1, 0, 107, 174, 28, 1, 63, 39, 3, 0, 1, 0, 107, 174, 63, 39, 4, 0, 63, 39, 1, 0, 28, 1, 107, 174, 27, 0, 154, 23, 173, 23, 192, 23, 211, 23, 234, 23, 3, 24, 24, 24, 44, 24, 64, 24, 76, 24, 96, 24, 110, 24, 130, 24, 146, 24, 158, 24, 176, 24, 192, 24, 207, 24, 223, 24, 245, 24, 1, 25, 19, 25, 37, 25, 55, 25, 66, 25, 85, 25, 105, 25, 152, 200, 152, 200, 152, 200, 146, 212, 182, 212, 115, 201, 77, 211, 115, 201, 155, 207, 136, 207, 145, 211, 145, 211, 86, 206, 127, 212, 73, 206, 145, 211, 73, 206, 127, 212, 73, 206, 134, 201, 112, 202, 79, 209, 8, 211, 8, 211, 57, 208, 57, 208, 127, 206, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 96, 24, 60, 204, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 96, 24, 60, 204, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 4, 0, 234, 23, 125, 25, 130, 24, 19, 25, 117, 212, 60, 204, 145, 211, 52, 206, 1, 0, 85, 25, 83, 210, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 85, 25, 83, 210, 4, 0, 211, 23, 148, 25, 125, 25, 182, 25, 249, 211, 218, 212, 33, 199, 255, 209, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 2, 0, 211, 23, 125, 25, 249, 211, 60, 204, 1, 0, 211, 23, 249, 211, 2, 0, 125, 25, 207, 24, 60, 204, 73, 206, 2, 0, 211, 23, 125, 25, 249, 211, 60, 204, 1, 0, 211, 23, 249, 211, 2, 0, 96, 24, 1, 25, 60, 204, 87, 201, 1, 0, 211, 23, 249, 211, 3, 0, 96, 24, 1, 25, 19, 25, 60, 204, 87, 201, 52, 206, 2, 0, 96, 24, 1, 25, 60, 204, 87, 201, 2, 0, 211, 23, 55, 25, 249, 211, 54, 210, 1, 0, 211, 23, 249, 211, 1, 0, 96, 24, 60, 204, 16, 0, 154, 23, 173, 23, 192, 23, 211, 23, 148, 25, 125, 25, 182, 25, 146, 24, 158, 24, 192, 24, 207, 24, 223, 24, 245, 24, 1, 25, 19, 25, 37, 25, 58, 199, 58, 199, 58, 199, 249, 211, 218, 212, 33, 199, 255, 209, 65, 209, 45, 200, 33, 199, 65, 209, 33, 200, 110, 200, 14, 200, 52, 206, 31, 208, 1, 0, 211, 23, 249, 211, 3, 0, 211, 23, 96, 24, 1, 25, 249, 211, 60, 204, 14, 200, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 96, 24, 60, 204, 1, 0, 211, 23, 249, 211, 2, 0, 96, 24, 19, 25, 60, 204, 52, 206, 1, 0, 211, 23, 249, 211, 3, 0, 125, 25, 182, 25, 19, 25, 33, 199, 255, 209, 103, 204, 1, 0, 211, 23, 249, 211, 2, 0, 234, 23, 96, 24, 117, 212, 60, 204, 2, 0, 96, 24, 1, 25, 60, 204, 87, 201, 1, 0, 234, 23, 117, 212, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 2, 0, 234, 23, 55, 25, 117, 212, 54, 210, 1, 0, 211, 23, 249, 211, 1, 0, 85, 25, 83, 210, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 4, 0, 234, 23, 125, 25, 192, 24, 19, 25, 117, 212, 60, 204, 60, 204, 52, 212, 3, 0, 211, 23, 234, 23, 96, 24, 249, 211, 117, 212, 60, 204, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 4, 0, 211, 23, 148, 25, 125, 25, 207, 24, 249, 211, 117, 212, 60, 204, 73, 206, 3, 0, 211, 23, 125, 25, 55, 25, 249, 211, 60, 204, 54, 210, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 2, 0, 125, 25, 55, 25, 60, 204, 54, 210, 1, 0, 85, 25, 83, 210, 1, 0, 125, 25, 60, 204, 1, 0, 85, 25, 83, 210, 1, 0, 211, 23, 249, 211, 4, 0, 211, 23, 125, 25, 19, 25, 55, 25, 249, 211, 60, 204, 52, 212, 54, 210, 2, 0, 96, 24, 1, 25, 60, 204, 87, 201, 4, 0, 211, 23, 96, 24, 211, 25, 55, 25, 249, 211, 60, 204, 34, 211, 54, 210, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 2, 0, 234, 23, 19, 25, 117, 212, 52, 206, 1, 0, 211, 23, 249, 211, 2, 0, 234, 23, 96, 24, 117, 212, 60, 204, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 22, 0, 154, 23, 173, 23, 192, 23, 234, 23, 96, 24, 125, 25, 110, 24, 130, 24, 146, 24, 158, 24, 176, 24, 192, 24, 207, 24, 223, 24, 245, 24, 1, 25, 19, 25, 37, 25, 55, 25, 66, 25, 85, 25, 105, 25, 58, 199, 58, 199, 58, 199, 168, 211, 255, 209, 33, 199, 255, 209, 255, 209, 65, 209, 33, 200, 255, 209, 33, 199, 65, 209, 33, 200, 163, 199, 14, 200, 103, 204, 31, 208, 31, 208, 83, 210, 83, 210, 177, 211, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 2, 0, 211, 23, 96, 24, 249, 211, 60, 204, 1, 0, 211, 23, 249, 211, 1, 0, 211, 23, 249, 211, 3, 0, 114, 199, 27, 202, 246, 199, 2, 0, 103, 199, 246, 199, 2, 0, 103, 199, 246, 199, 2, 0, 27, 202, 246, 199, 3, 0, 114, 199, 246, 199, 27, 202, 2, 0, 114, 199, 246, 199, 2, 0, 103, 199, 246, 199, 2, 0, 103, 199, 246, 199, 2, 0, 103, 199, 246, 199, 2, 0, 103, 199, 246, 199, 3, 0, 114, 199, 246, 199, 27, 202, 2, 0, 103, 199, 246, 199, 2, 0, 114, 199, 246, 199, 2, 0, 103, 199, 246, 199, 2, 0, 246, 199, 114, 199, 2, 0, 246, 199, 27, 202, 3, 0, 103, 199, 114, 199, 246, 199, 2, 0, 103, 199, 246, 199, 2, 0, 246, 199, 27, 202, 3, 0, 103, 199, 246, 199, 27, 202, 2, 0, 246, 199, 27, 202, 3, 0, 114, 199, 246, 199, 27, 202, 2, 0, 114, 199, 246, 199, 2, 0, 114, 199, 246, 199, 2, 0, 246, 199, 103, 199, 3, 0, 246, 199, 114, 199, 103, 199, 2, 0, 103, 199, 246, 199, 4, 0, 246, 199, 27, 202, 114, 199, 103, 199, 2, 0, 103, 199, 246, 199, 2, 0, 246, 199, 27, 202, 2, 0, 103, 199, 246, 199, 3, 0, 246, 199, 114, 199, 103, 199, 3, 0, 103, 199, 246, 199, 27, 202, 2, 0, 246, 199, 103, 199, 4, 0, 246, 199, 103, 199, 27, 202, 114, 199, 3, 0, 114, 199, 103, 199, 246, 199, 3, 0, 103, 199, 246, 199, 27, 202, 3, 0, 246, 199, 114, 199, 103, 199, 2, 0, 246, 199, 27, 202, 3, 0, 114, 199, 246, 199, 27, 202, 2, 0, 27, 202, 246, 199, 2, 0, 27, 202, 246, 199, 2, 0, 246, 199, 27, 202, 2, 0, 103, 199, 246, 199, 4, 0, 246, 199, 27, 202, 114, 199, 103, 199, 2, 0, 246, 199, 27, 202, 3, 0, 103, 199, 246, 199, 27, 202, 3, 0, 246, 199, 27, 202, 103, 199, 2, 0, 103, 199, 246, 199, 3, 0, 246, 199, 27, 202, 103, 199, 2, 0, 103, 199, 246, 199, 4, 0, 246, 199, 103, 199, 27, 202, 114, 199, 2, 0, 246, 199, 27, 202, 3, 0, 114, 199, 246, 199, 27, 202, 2, 0, 246, 199, 27, 202, 3, 0, 103, 199, 246, 199, 27, 202, 3, 0, 114, 199, 246, 199, 27, 202, 3, 0, 103, 199, 114, 199, 246, 199, 2, 0, 103, 199, 27, 202, 3, 0, 0, 0, 226, 253, 255, 255, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 173, 158, 0, 224, 1, 0, 70, 2, 177, 158, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 179, 158, 0, 32, 47, 199, 0, 96, 3, 0, 0, 0, 180, 245, 255, 255, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 185, 158, 0, 224, 1, 0, 70, 2, 189, 158, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 191, 158, 0, 32, 43, 199, 0, 96, 3, 0, 0, 0, 28, 1, 0, 0, 8, 0, 0, 0, 28, 0, 0, 0, 1, 0, 87, 0, 197, 158, 0, 224, 3, 0, 0, 0, 28, 1, 0, 0, 8, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 203, 158, 0, 224, 2, 0, 70, 2, 13, 20, 170, 170, 201, 158, 0, 32, 207, 158, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 209, 158, 0, 32, 65, 126, 0, 96, 3, 0, 0, 0, 228, 246, 255, 255, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 217, 158, 0, 224, 1, 0, 70, 2, 221, 158, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 223, 158, 0, 32, 43, 199, 0, 96, 3, 0, 0, 0, 8, 0, 0, 0, 8, 0, 0, 0, 28, 0, 0, 0, 1, 0, 87, 0, 229, 158, 0, 224, 3, 0, 0, 0, 8, 0, 0, 0, 8, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 235, 158, 0, 224, 2, 0, 70, 2, 13, 20, 170, 170, 233, 158, 0, 32, 239, 158, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 241, 158, 0, 32, 65, 126, 0, 96, 3, 0, 0, 0, 140, 234, 255, 255, 8, 0, 0, 0, 29, 0, 0, 0, 1, 0, 87, 0, 249, 158, 0, 224, 1, 0, 70, 2, 253, 158, 0, 32, 1, 0, 54, 0, 255, 158, 0, 32, 3, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 0, 31, 0, 0, 0, 1, 0, 87, 0, 3, 159, 0, 224, 3, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 9, 159, 0, 224, 2, 0, 70, 2, 13, 20, 170, 170, 7, 159, 0, 32, 13, 159, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 15, 159, 0, 32, 47, 199, 0, 96, 3, 0, 0, 0, 80, 241, 255, 255, 10, 0, 0, 0, 7, 0, 0, 0, 1, 0, 59, 0, 23, 159, 0, 224, 1, 0, 70, 2, 27, 159, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 29, 159, 0, 32, 43, 199, 0, 96, 3, 0, 0, 0, 79, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 35, 159, 0, 224, 1, 0, 70, 2, 39, 159, 0, 32, 1, 0, 54, 0, 41, 159, 0, 32, 3, 0, 0, 0, 110, 2, 0, 0, 7, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 45, 159, 0, 224, 1, 0, 70, 2, 49, 159, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 51, 159, 0, 32, 217, 106, 0, 96, 3, 0, 0, 0, 110, 2, 0, 0, 7, 0, 0, 0, 16, 0, 0, 0, 1, 0, 59, 0, 57, 159, 0, 224, 1, 0, 70, 2, 61, 159, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 63, 159, 0, 32, 217, 106, 0, 96, 3, 0, 0, 0, 110, 2, 0, 0, 7, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 69, 159, 0, 224, 1, 0, 70, 2, 73, 159, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 75, 159, 0, 32, 217, 106, 0, 96, 3, 0, 0, 0, 110, 2, 0, 0, 7, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 81, 159, 0, 224, 1, 0, 70, 2, 85, 159, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 87, 159, 0, 32, 217, 106, 0, 96, 3, 0, 0, 0, 110, 2, 0, 0, 7, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 93, 159, 0, 224, 1, 0, 70, 2, 97, 159, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 99, 159, 0, 32, 217, 106, 0, 96, 3, 0, 0, 0, 133, 2, 0, 0, 6, 0, 0, 0, 19, 0, 0, 0, 1, 0, 59, 0, 105, 159, 0, 224, 3, 0, 0, 0, 138, 2, 0, 0, 2, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 111, 159, 0, 224, 3, 0, 0, 0, 217, 2, 0, 0, 8, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 117, 159, 0, 224, 3, 0, 0, 0, 145, 4, 0, 0, 4, 0, 0, 0, 8, 0, 0, 0, 1, 0, 59, 0, 123, 159, 0, 224, 3, 0, 0, 0, 147, 4, 0, 0, 4, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 129, 159, 0, 224, 3, 0, 0, 0, 151, 4, 0, 0, 7, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 135, 159, 0, 224, 3, 0, 0, 0, 153, 4, 0, 0, 8, 0, 0, 0, 4, 0, 0, 0, 1, 0, 59, 0, 141, 159, 0, 224, 3, 0, 0, 0, 157, 4, 0, 0, 7, 0, 0, 0, 14, 0, 0, 0, 1, 0, 59, 0, 147, 159, 0, 224, 3, 0, 0, 0, 158, 4, 0, 0, 5, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 153, 159, 0, 224, 3, 0, 0, 0, 160, 4, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 1, 0, 59, 0, 159, 159, 0, 224, 3, 0, 0, 0, 161, 4, 0, 0, 8, 0, 0, 0, 14, 0, 0, 0, 1, 0, 59, 0, 165, 159, 0, 224, 3, 0, 0, 0, 166, 4, 0, 0, 4, 0, 0, 0, 11, 0, 0, 0, 1, 0, 59, 0, 171, 159, 0, 224, 3, 0, 0, 0, 175, 4, 0, 0, 4, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 177, 159, 0, 224, 3, 0, 0, 0, 237, 2, 0, 0, 4, 0, 0, 0, 14, 0, 0, 0, 1, 0, 59, 0, 183, 159, 0, 224, 3, 0, 0, 0, 177, 4, 0, 0, 2, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 189, 159, 0, 224, 3, 0, 0, 0, 180, 4, 0, 0, 2, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 195, 159, 0, 224, 3, 0, 0, 0, 182, 4, 0, 0, 4, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 201, 159, 0, 224, 3, 0, 0, 0, 183, 4, 0, 0, 10, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 207, 159, 0, 224, 3, 0, 0, 0, 187, 4, 0, 0, 3, 0, 0, 0, 9, 0, 0, 0, 1, 0, 59, 0, 213, 159, 0, 224, 3, 0, 0, 0, 189, 4, 0, 0, 12, 0, 0, 0, 6, 0, 0, 0, 1, 0, 59, 0, 219, 159, 0, 224, 3, 0, 0, 0, 195, 4, 0, 0, 4, 0, 0, 0, 12, 0, 0, 0, 1, 0, 59, 0, 225, 159, 0, 224, 3, 0, 0, 0, 198, 4, 0, 0, 4, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 231, 159, 0, 224, 3, 0, 0, 0, 200, 4, 0, 0, 11, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 237, 159, 0, 224, 3, 0, 0, 0, 201, 4, 0, 0, 4, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 243, 159, 0, 224, 3, 0, 0, 0, 237, 2, 0, 0, 7, 0, 0, 0, 2, 0, 0, 0, 1, 0, 59, 0, 249, 159, 0, 224, 3, 0, 0, 0, 203, 4, 0, 0, 12, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 255, 159, 0, 224, 3, 0, 0, 0, 205, 4, 0, 0, 3, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 5, 160, 0, 224, 3, 0, 0, 0, 208, 4, 0, 0, 4, 0, 0, 0, 2, 0, 0, 0, 1, 0, 59, 0, 11, 160, 0, 224, 3, 0, 0, 0, 209, 4, 0, 0, 4, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 17, 160, 0, 224, 3, 0, 0, 0, 210, 4, 0, 0, 11, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 23, 160, 0, 224, 3, 0, 0, 0, 211, 4, 0, 0, 9, 0, 0, 0, 19, 0, 0, 0, 1, 0, 59, 0, 29, 160, 0, 224, 3, 0, 0, 0, 214, 4, 0, 0, 11, 0, 0, 0, 23, 0, 0, 0, 1, 0, 59, 0, 35, 160, 0, 224, 3, 0, 0, 0, 215, 4, 0, 0, 2, 0, 0, 0, 7, 0, 0, 0, 1, 0, 59, 0, 41, 160, 0, 224, 3, 0, 0, 0, 216, 4, 0, 0, 7, 0, 0, 0, 16, 0, 0, 0, 1, 0, 59, 0, 47, 160, 0, 224, 3, 0, 0, 0, 219, 4, 0, 0, 2, 0, 0, 0, 26, 0, 0, 0, 1, 0, 59, 0, 53, 160, 0, 224, 3, 0, 0, 0, 245, 2, 0, 0, 8, 0, 0, 0, 18, 0, 0, 0, 1, 0, 59, 0, 59, 160, 0, 224, 3, 0, 0, 0, 223, 4, 0, 0, 2, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 65, 160, 0, 224, 3, 0, 0, 0, 225, 4, 0, 0, 3, 0, 0, 0, 18, 0, 0, 0, 1, 0, 59, 0, 71, 160, 0, 224, 3, 0, 0, 0, 232, 4, 0, 0, 10, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 77, 160, 0, 224, 3, 0, 0, 0, 233, 4, 0, 0, 3, 0, 0, 0, 14, 0, 0, 0, 1, 0, 59, 0, 83, 160, 0, 224, 3, 0, 0, 0, 235, 4, 0, 0, 3, 0, 0, 0, 26, 0, 0, 0, 1, 0, 59, 0, 89, 160, 0, 224, 3, 0, 0, 0, 236, 4, 0, 0, 4, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 95, 160, 0, 224, 3, 0, 0, 0, 237, 4, 0, 0, 2, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 101, 160, 0, 224, 3, 0, 0, 0, 240, 4, 0, 0, 2, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 107, 160, 0, 224, 3, 0, 0, 0, 251, 4, 0, 0, 4, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 113, 160, 0, 224, 3, 0, 0, 0, 254, 4, 0, 0, 2, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 119, 160, 0, 224, 3, 0, 0, 0, 253, 2, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 1, 0, 59, 0, 125, 160, 0, 224, 3, 0, 0, 0, 8, 5, 0, 0, 4, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 131, 160, 0, 224, 3, 0, 0, 0, 13, 5, 0, 0, 8, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 137, 160, 0, 224, 3, 0, 0, 0, 19, 5, 0, 0, 4, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 143, 160, 0, 224, 3, 0, 0, 0, 22, 5, 0, 0, 11, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 149, 160, 0, 224, 3, 0, 0, 0, 23, 5, 0, 0, 8, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 155, 160, 0, 224, 3, 0, 0, 0, 26, 5, 0, 0, 12, 0, 0, 0, 14, 0, 0, 0, 1, 0, 59, 0, 161, 160, 0, 224, 3, 0, 0, 0, 28, 5, 0, 0, 10, 0, 0, 0, 9, 0, 0, 0, 1, 0, 59, 0, 167, 160, 0, 224, 3, 0, 0, 0, 31, 5, 0, 0, 4, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 173, 160, 0, 224, 3, 0, 0, 0, 32, 5, 0, 0, 3, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 179, 160, 0, 224, 3, 0, 0, 0, 37, 5, 0, 0, 2, 0, 0, 0, 3, 0, 0, 0, 1, 0, 59, 0, 185, 160, 0, 224, 3, 0, 0, 0, 255, 2, 0, 0, 8, 0, 0, 0, 16, 0, 0, 0, 1, 0, 59, 0, 191, 160, 0, 224, 3, 0, 0, 0, 39, 5, 0, 0, 4, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 197, 160, 0, 224, 3, 0, 0, 0, 41, 5, 0, 0, 2, 0, 0, 0, 23, 0, 0, 0, 1, 0, 59, 0, 203, 160, 0, 224, 3, 0, 0, 0, 44, 5, 0, 0, 12, 0, 0, 0, 9, 0, 0, 0, 1, 0, 59, 0, 209, 160, 0, 224, 3, 0, 0, 0, 46, 5, 0, 0, 4, 0, 0, 0, 26, 0, 0, 0, 1, 0, 59, 0, 215, 160, 0, 224, 3, 0, 0, 0, 49, 5, 0, 0, 8, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 221, 160, 0, 224, 3, 0, 0, 0, 51, 5, 0, 0, 8, 0, 0, 0, 9, 0, 0, 0, 1, 0, 59, 0, 227, 160, 0, 224, 3, 0, 0, 0, 54, 5, 0, 0, 1, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 233, 160, 0, 224, 3, 0, 0, 0, 56, 5, 0, 0, 2, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 239, 160, 0, 224, 3, 0, 0, 0, 60, 5, 0, 0, 4, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 245, 160, 0, 224, 3, 0, 0, 0, 66, 5, 0, 0, 12, 0, 0, 0, 8, 0, 0, 0, 1, 0, 59, 0, 251, 160, 0, 224, 3, 0, 0, 0, 2, 3, 0, 0, 10, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 1, 161, 0, 224, 3, 0, 0, 0, 90, 5, 0, 0, 7, 0, 0, 0, 24, 0, 0, 0, 1, 0, 59, 0, 7, 161, 0, 224, 3, 0, 0, 0, 92, 5, 0, 0, 4, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 13, 161, 0, 224, 3, 0, 0, 0, 95, 5, 0, 0, 5, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 19, 161, 0, 224, 3, 0, 0, 0, 99, 5, 0, 0, 3, 0, 0, 0, 22, 0, 0, 0, 1, 0, 59, 0, 25, 161, 0, 224, 3, 0, 0, 0, 101, 5, 0, 0, 2, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 31, 161, 0, 224, 3, 0, 0, 0, 104, 5, 0, 0, 4, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 37, 161, 0, 224, 3, 0, 0, 0, 107, 5, 0, 0, 8, 0, 0, 0, 22, 0, 0, 0, 1, 0, 59, 0, 43, 161, 0, 224, 3, 0, 0, 0, 107, 5, 0, 0, 8, 0, 0, 0, 23, 0, 0, 0, 1, 0, 59, 0, 49, 161, 0, 224, 3, 0, 0, 0, 109, 5, 0, 0, 2, 0, 0, 0, 9, 0, 0, 0, 1, 0, 59, 0, 55, 161, 0, 224, 3, 0, 0, 0, 110, 5, 0, 0, 3, 0, 0, 0, 26, 0, 0, 0, 1, 0, 59, 0, 61, 161, 0, 224, 3, 0, 0, 0, 13, 3, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 67, 161, 0, 224, 3, 0, 0, 0, 114, 5, 0, 0, 7, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 73, 161, 0, 224, 3, 0, 0, 0, 148, 5, 0, 0, 4, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 79, 161, 0, 224, 3, 0, 0, 0, 149, 5, 0, 0, 9, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 85, 161, 0, 224, 3, 0, 0, 0, 161, 5, 0, 0, 2, 0, 0, 0, 17, 0, 0, 0, 1, 0, 59, 0, 91, 161, 0, 224, 3, 0, 0, 0, 164, 5, 0, 0, 2, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 97, 161, 0, 224, 3, 0, 0, 0, 169, 5, 0, 0, 7, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 103, 161, 0, 224, 3, 0, 0, 0, 172, 5, 0, 0, 7, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 109, 161, 0, 224, 3, 0, 0, 0, 175, 5, 0, 0, 7, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 115, 161, 0, 224, 3, 0, 0, 0, 177, 5, 0, 0, 9, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 121, 161, 0, 224, 3, 0, 0, 0, 180, 5, 0, 0, 12, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 127, 161, 0, 224, 3, 0, 0, 0, 14, 3, 0, 0, 8, 0, 0, 0, 19, 0, 0, 0, 1, 0, 59, 0, 133, 161, 0, 224, 3, 0, 0, 0, 186, 5, 0, 0, 2, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 139, 161, 0, 224, 3, 0, 0, 0, 187, 5, 0, 0, 3, 0, 0, 0, 3, 0, 0, 0, 1, 0, 59, 0, 145, 161, 0, 224, 3, 0, 0, 0, 189, 5, 0, 0, 4, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 151, 161, 0, 224, 3, 0, 0, 0, 207, 5, 0, 0, 7, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 157, 161, 0, 224, 3, 0, 0, 0, 209, 5, 0, 0, 8, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 163, 161, 0, 224, 3, 0, 0, 0, 212, 5, 0, 0, 7, 0, 0, 0, 19, 0, 0, 0, 1, 0, 59, 0, 169, 161, 0, 224, 3, 0, 0, 0, 221, 5, 0, 0, 2, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 175, 161, 0, 224, 3, 0, 0, 0, 224, 5, 0, 0, 2, 0, 0, 0, 30, 0, 0, 0, 1, 0, 59, 0, 181, 161, 0, 224, 3, 0, 0, 0, 241, 5, 0, 0, 8, 0, 0, 0, 23, 0, 0, 0, 1, 0, 59, 0, 187, 161, 0, 224, 3, 0, 0, 0, 248, 5, 0, 0, 8, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 193, 161, 0, 224, 3, 0, 0, 0, 38, 3, 0, 0, 5, 0, 0, 0, 18, 0, 0, 0, 1, 0, 59, 0, 199, 161, 0, 224, 3, 0, 0, 0, 252, 5, 0, 0, 7, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 205, 161, 0, 224, 3, 0, 0, 0, 19, 6, 0, 0, 10, 0, 0, 0, 23, 0, 0, 0, 1, 0, 59, 0, 211, 161, 0, 224, 3, 0, 0, 0, 22, 6, 0, 0, 2, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 217, 161, 0, 224, 3, 0, 0, 0, 34, 6, 0, 0, 4, 0, 0, 0, 23, 0, 0, 0, 1, 0, 59, 0, 223, 161, 0, 224, 3, 0, 0, 0, 37, 6, 0, 0, 7, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 229, 161, 0, 224, 3, 0, 0, 0, 56, 6, 0, 0, 12, 0, 0, 0, 8, 0, 0, 0, 1, 0, 59, 0, 235, 161, 0, 224, 3, 0, 0, 0, 60, 6, 0, 0, 10, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 241, 161, 0, 224, 3, 0, 0, 0, 79, 6, 0, 0, 7, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 247, 161, 0, 224, 3, 0, 0, 0, 88, 6, 0, 0, 2, 0, 0, 0, 30, 0, 0, 0, 1, 0, 59, 0, 253, 161, 0, 224, 3, 0, 0, 0, 108, 6, 0, 0, 12, 0, 0, 0, 16, 0, 0, 0, 1, 0, 59, 0, 3, 162, 0, 224, 3, 0, 0, 0, 160, 2, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 9, 162, 0, 224, 3, 0, 0, 0, 42, 3, 0, 0, 9, 0, 0, 0, 19, 0, 0, 0, 1, 0, 59, 0, 15, 162, 0, 224, 3, 0, 0, 0, 112, 6, 0, 0, 2, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 21, 162, 0, 224, 3, 0, 0, 0, 116, 6, 0, 0, 9, 0, 0, 0, 18, 0, 0, 0, 1, 0, 59, 0, 27, 162, 0, 224, 3, 0, 0, 0, 119, 6, 0, 0, 4, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 33, 162, 0, 224, 3, 0, 0, 0, 122, 6, 0, 0, 7, 0, 0, 0, 23, 0, 0, 0, 1, 0, 59, 0, 39, 162, 0, 224, 3, 0, 0, 0, 125, 6, 0, 0, 4, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 45, 162, 0, 224, 3, 0, 0, 0, 137, 6, 0, 0, 9, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 51, 162, 0, 224, 3, 0, 0, 0, 145, 6, 0, 0, 9, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 57, 162, 0, 224, 3, 0, 0, 0, 148, 6, 0, 0, 2, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 63, 162, 0, 224, 3, 0, 0, 0, 152, 6, 0, 0, 9, 0, 0, 0, 30, 0, 0, 0, 1, 0, 59, 0, 69, 162, 0, 224, 3, 0, 0, 0, 168, 6, 0, 0, 3, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 75, 162, 0, 224, 3, 0, 0, 0, 56, 3, 0, 0, 1, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 81, 162, 0, 224, 3, 0, 0, 0, 175, 6, 0, 0, 4, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 87, 162, 0, 224, 3, 0, 0, 0, 180, 6, 0, 0, 6, 0, 0, 0, 22, 0, 0, 0, 1, 0, 59, 0, 93, 162, 0, 224, 3, 0, 0, 0, 200, 6, 0, 0, 4, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 99, 162, 0, 224, 3, 0, 0, 0, 205, 6, 0, 0, 2, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 105, 162, 0, 224, 3, 0, 0, 0, 208, 6, 0, 0, 2, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 111, 162, 0, 224, 3, 0, 0, 0, 212, 6, 0, 0, 7, 0, 0, 0, 12, 0, 0, 0, 1, 0, 59, 0, 117, 162, 0, 224, 3, 0, 0, 0, 215, 6, 0, 0, 10, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 123, 162, 0, 224, 3, 0, 0, 0, 228, 6, 0, 0, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 59, 0, 129, 162, 0, 224, 3, 0, 0, 0, 236, 6, 0, 0, 11, 0, 0, 0, 16, 0, 0, 0, 1, 0, 59, 0, 135, 162, 0, 224, 3, 0, 0, 0, 245, 6, 0, 0, 4, 0, 0, 0, 2, 0, 0, 0, 1, 0, 59, 0, 141, 162, 0, 224, 3, 0, 0, 0, 66, 3, 0, 0, 1, 0, 0, 0, 3, 0, 0, 0, 1, 0, 59, 0, 147, 162, 0, 224, 3, 0, 0, 0, 253, 6, 0, 0, 1, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 153, 162, 0, 224, 3, 0, 0, 0, 9, 7, 0, 0, 2, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 159, 162, 0, 224, 3, 0, 0, 0, 12, 7, 0, 0, 2, 0, 0, 0, 11, 0, 0, 0, 1, 0, 59, 0, 165, 162, 0, 224, 3, 0, 0, 0, 26, 7, 0, 0, 4, 0, 0, 0, 22, 0, 0, 0, 1, 0, 59, 0, 171, 162, 0, 224, 3, 0, 0, 0, 38, 7, 0, 0, 12, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 177, 162, 0, 224, 3, 0, 0, 0, 52, 7, 0, 0, 12, 0, 0, 0, 2, 0, 0, 0, 1, 0, 59, 0, 183, 162, 0, 224, 3, 0, 0, 0, 56, 7, 0, 0, 2, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 189, 162, 0, 224, 3, 0, 0, 0, 62, 7, 0, 0, 11, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 195, 162, 0, 224, 3, 0, 0, 0, 68, 7, 0, 0, 3, 0, 0, 0, 18, 0, 0, 0, 1, 0, 59, 0, 201, 162, 0, 224, 3, 0, 0, 0, 69, 7, 0, 0, 2, 0, 0, 0, 19, 0, 0, 0, 1, 0, 59, 0, 207, 162, 0, 224, 3, 0, 0, 0, 80, 3, 0, 0, 6, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 213, 162, 0, 224, 3, 0, 0, 0, 72, 7, 0, 0, 2, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 219, 162, 0, 224, 3, 0, 0, 0, 73, 7, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 1, 0, 59, 0, 225, 162, 0, 224, 3, 0, 0, 0, 76, 7, 0, 0, 9, 0, 0, 0, 8, 0, 0, 0, 1, 0, 59, 0, 231, 162, 0, 224, 3, 0, 0, 0, 120, 7, 0, 0, 7, 0, 0, 0, 30, 0, 0, 0, 1, 0, 59, 0, 237, 162, 0, 224, 3, 0, 0, 0, 134, 7, 0, 0, 12, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 243, 162, 0, 224, 3, 0, 0, 0, 197, 7, 0, 0, 1, 0, 0, 0, 8, 0, 0, 0, 1, 0, 59, 0, 249, 162, 0, 224, 3, 0, 0, 0, 227, 7, 0, 0, 5, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 255, 162, 0, 224, 3, 0, 0, 0, 83, 3, 0, 0, 4, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 5, 163, 0, 224, 3, 0, 0, 0, 86, 3, 0, 0, 11, 0, 0, 0, 30, 0, 0, 0, 1, 0, 59, 0, 11, 163, 0, 224, 3, 0, 0, 0, 89, 3, 0, 0, 2, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 17, 163, 0, 224, 3, 0, 0, 0, 91, 3, 0, 0, 4, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 23, 163, 0, 224, 3, 0, 0, 0, 109, 3, 0, 0, 4, 0, 0, 0, 16, 0, 0, 0, 1, 0, 59, 0, 29, 163, 0, 224, 3, 0, 0, 0, 117, 3, 0, 0, 2, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 35, 163, 0, 224, 3, 0, 0, 0, 174, 2, 0, 0, 7, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 41, 163, 0, 224, 3, 0, 0, 0, 121, 3, 0, 0, 4, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 47, 163, 0, 224, 3, 0, 0, 0, 130, 3, 0, 0, 4, 0, 0, 0, 26, 0, 0, 0, 1, 0, 59, 0, 53, 163, 0, 224, 3, 0, 0, 0, 133, 3, 0, 0, 7, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 59, 163, 0, 224, 3, 0, 0, 0, 155, 3, 0, 0, 4, 0, 0, 0, 11, 0, 0, 0, 1, 0, 59, 0, 65, 163, 0, 224, 3, 0, 0, 0, 163, 3, 0, 0, 4, 0, 0, 0, 26, 0, 0, 0, 1, 0, 59, 0, 71, 163, 0, 224, 3, 0, 0, 0, 170, 3, 0, 0, 5, 0, 0, 0, 22, 0, 0, 0, 1, 0, 59, 0, 77, 163, 0, 224, 3, 0, 0, 0, 179, 3, 0, 0, 4, 0, 0, 0, 22, 0, 0, 0, 1, 0, 59, 0, 83, 163, 0, 224, 3, 0, 0, 0, 189, 3, 0, 0, 10, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 89, 163, 0, 224, 3, 0, 0, 0, 193, 3, 0, 0, 2, 0, 0, 0, 16, 0, 0, 0, 1, 0, 59, 0, 95, 163, 0, 224, 3, 0, 0, 0, 196, 3, 0, 0, 7, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 101, 163, 0, 224, 3, 0, 0, 0, 189, 2, 0, 0, 3, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 107, 163, 0, 224, 3, 0, 0, 0, 200, 3, 0, 0, 8, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 113, 163, 0, 224, 3, 0, 0, 0, 202, 3, 0, 0, 3, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 119, 163, 0, 224, 3, 0, 0, 0, 205, 3, 0, 0, 12, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 125, 163, 0, 224, 3, 0, 0, 0, 208, 3, 0, 0, 7, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 131, 163, 0, 224, 3, 0, 0, 0, 210, 3, 0, 0, 11, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 137, 163, 0, 224, 3, 0, 0, 0, 215, 3, 0, 0, 4, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 143, 163, 0, 224, 3, 0, 0, 0, 217, 3, 0, 0, 4, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 149, 163, 0, 224, 3, 0, 0, 0, 219, 3, 0, 0, 4, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 155, 163, 0, 224, 3, 0, 0, 0, 221, 3, 0, 0, 8, 0, 0, 0, 8, 0, 0, 0, 1, 0, 59, 0, 161, 163, 0, 224, 3, 0, 0, 0, 222, 3, 0, 0, 11, 0, 0, 0, 7, 0, 0, 0, 1, 0, 59, 0, 167, 163, 0, 224, 3, 0, 0, 0, 192, 2, 0, 0, 5, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 173, 163, 0, 224, 3, 0, 0, 0, 227, 3, 0, 0, 2, 0, 0, 0, 22, 0, 0, 0, 1, 0, 59, 0, 179, 163, 0, 224, 3, 0, 0, 0, 231, 3, 0, 0, 1, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 185, 163, 0, 224, 3, 0, 0, 0, 236, 3, 0, 0, 7, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 191, 163, 0, 224, 3, 0, 0, 0, 244, 3, 0, 0, 12, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 197, 163, 0, 224, 3, 0, 0, 0, 249, 3, 0, 0, 4, 0, 0, 0, 23, 0, 0, 0, 1, 0, 59, 0, 203, 163, 0, 224, 3, 0, 0, 0, 253, 3, 0, 0, 2, 0, 0, 0, 2, 0, 0, 0, 1, 0, 59, 0, 209, 163, 0, 224, 3, 0, 0, 0, 0, 4, 0, 0, 7, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 215, 163, 0, 224, 3, 0, 0, 0, 4, 4, 0, 0, 7, 0, 0, 0, 25, 0, 0, 0, 1, 0, 59, 0, 221, 163, 0, 224, 3, 0, 0, 0, 13, 4, 0, 0, 4, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 227, 163, 0, 224, 3, 0, 0, 0, 16, 4, 0, 0, 11, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 233, 163, 0, 224, 3, 0, 0, 0, 196, 2, 0, 0, 1, 0, 0, 0, 11, 0, 0, 0, 1, 0, 59, 0, 239, 163, 0, 224, 3, 0, 0, 0, 20, 4, 0, 0, 11, 0, 0, 0, 24, 0, 0, 0, 1, 0, 59, 0, 245, 163, 0, 224, 3, 0, 0, 0, 22, 4, 0, 0, 4, 0, 0, 0, 14, 0, 0, 0, 1, 0, 59, 0, 251, 163, 0, 224, 3, 0, 0, 0, 29, 4, 0, 0, 1, 0, 0, 0, 11, 0, 0, 0, 1, 0, 59, 0, 1, 164, 0, 224, 3, 0, 0, 0, 34, 4, 0, 0, 8, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 7, 164, 0, 224, 3, 0, 0, 0, 41, 4, 0, 0, 8, 0, 0, 0, 2, 0, 0, 0, 1, 0, 59, 0, 13, 164, 0, 224, 3, 0, 0, 0, 45, 4, 0, 0, 4, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 19, 164, 0, 224, 3, 0, 0, 0, 50, 4, 0, 0, 8, 0, 0, 0, 23, 0, 0, 0, 1, 0, 59, 0, 25, 164, 0, 224, 3, 0, 0, 0, 53, 4, 0, 0, 11, 0, 0, 0, 17, 0, 0, 0, 1, 0, 59, 0, 31, 164, 0, 224, 3, 0, 0, 0, 57, 4, 0, 0, 2, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 37, 164, 0, 224, 3, 0, 0, 0, 60, 4, 0, 0, 2, 0, 0, 0, 7, 0, 0, 0, 1, 0, 59, 0, 43, 164, 0, 224, 3, 0, 0, 0, 203, 2, 0, 0, 9, 0, 0, 0, 2, 0, 0, 0, 1, 0, 59, 0, 49, 164, 0, 224, 3, 0, 0, 0, 63, 4, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 1, 0, 59, 0, 55, 164, 0, 224, 3, 0, 0, 0, 70, 4, 0, 0, 12, 0, 0, 0, 15, 0, 0, 0, 1, 0, 59, 0, 61, 164, 0, 224, 3, 0, 0, 0, 72, 4, 0, 0, 12, 0, 0, 0, 17, 0, 0, 0, 1, 0, 59, 0, 67, 164, 0, 224, 3, 0, 0, 0, 73, 4, 0, 0, 11, 0, 0, 0, 21, 0, 0, 0, 1, 0, 59, 0, 73, 164, 0, 224, 3, 0, 0, 0, 75, 4, 0, 0, 8, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 79, 164, 0, 224, 3, 0, 0, 0, 80, 4, 0, 0, 2, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 85, 164, 0, 224, 3, 0, 0, 0, 82, 4, 0, 0, 4, 0, 0, 0, 9, 0, 0, 0, 1, 0, 59, 0, 91, 164, 0, 224, 3, 0, 0, 0, 84, 4, 0, 0, 8, 0, 0, 0, 3, 0, 0, 0, 1, 0, 59, 0, 97, 164, 0, 224, 3, 0, 0, 0, 86, 4, 0, 0, 7, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 103, 164, 0, 224, 3, 0, 0, 0, 89, 4, 0, 0, 7, 0, 0, 0, 13, 0, 0, 0, 1, 0, 59, 0, 109, 164, 0, 224, 3, 0, 0, 0, 205, 2, 0, 0, 11, 0, 0, 0, 17, 0, 0, 0, 1, 0, 59, 0, 115, 164, 0, 224, 3, 0, 0, 0, 94, 4, 0, 0, 4, 0, 0, 0, 3, 0, 0, 0, 1, 0, 59, 0, 121, 164, 0, 224, 3, 0, 0, 0, 96, 4, 0, 0, 4, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 127, 164, 0, 224, 3, 0, 0, 0, 100, 4, 0, 0, 4, 0, 0, 0, 3, 0, 0, 0, 1, 0, 59, 0, 133, 164, 0, 224, 3, 0, 0, 0, 102, 4, 0, 0, 1, 0, 0, 0, 22, 0, 0, 0, 1, 0, 59, 0, 139, 164, 0, 224, 3, 0, 0, 0, 107, 4, 0, 0, 1, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 145, 164, 0, 224, 3, 0, 0, 0, 108, 4, 0, 0, 8, 0, 0, 0, 11, 0, 0, 0, 1, 0, 59, 0, 151, 164, 0, 224, 3, 0, 0, 0, 111, 4, 0, 0, 4, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 157, 164, 0, 224, 3, 0, 0, 0, 117, 4, 0, 0, 7, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 163, 164, 0, 224, 3, 0, 0, 0, 118, 4, 0, 0, 4, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 169, 164, 0, 224, 3, 0, 0, 0, 120, 4, 0, 0, 2, 0, 0, 0, 23, 0, 0, 0, 1, 0, 59, 0, 175, 164, 0, 224, 3, 0, 0, 0, 212, 2, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 1, 0, 59, 0, 181, 164, 0, 224, 3, 0, 0, 0, 121, 4, 0, 0, 7, 0, 0, 0, 22, 0, 0, 0, 1, 0, 59, 0, 187, 164, 0, 224, 3, 0, 0, 0, 127, 4, 0, 0, 1, 0, 0, 0, 26, 0, 0, 0, 1, 0, 59, 0, 193, 164, 0, 224, 3, 0, 0, 0, 130, 4, 0, 0, 10, 0, 0, 0, 28, 0, 0, 0, 1, 0, 59, 0, 199, 164, 0, 224, 3, 0, 0, 0, 132, 4, 0, 0, 4, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 205, 164, 0, 224, 3, 0, 0, 0, 135, 4, 0, 0, 4, 0, 0, 0, 20, 0, 0, 0, 1, 0, 59, 0, 211, 164, 0, 224, 3, 0, 0, 0, 136, 4, 0, 0, 1, 0, 0, 0, 10, 0, 0, 0, 1, 0, 59, 0, 217, 164, 0, 224, 3, 0, 0, 0, 137, 4, 0, 0, 9, 0, 0, 0, 4, 0, 0, 0, 1, 0, 59, 0, 223, 164, 0, 224, 3, 0, 0, 0, 139, 4, 0, 0, 3, 0, 0, 0, 29, 0, 0, 0, 1, 0, 59, 0, 229, 164, 0, 224, 3, 0, 0, 0, 141, 4, 0, 0, 6, 0, 0, 0, 5, 0, 0, 0, 1, 0, 59, 0, 235, 164, 0, 224, 3, 0, 0, 0, 142, 4, 0, 0, 8, 0, 0, 0, 27, 0, 0, 0, 1, 0, 59, 0, 241, 164, 0, 224, 237, 0, 70, 2, 13, 20, 73, 20, 224, 0, 228, 0, 232, 0, 236, 0, 240, 0, 244, 0, 248, 0, 252, 0, 0, 1, 4, 1, 78, 20, 8, 1, 12, 1, 16, 1, 20, 1, 24, 1, 28, 1, 32, 1, 36, 1, 40, 1, 44, 1, 83, 20, 48, 1, 52, 1, 56, 1, 60, 1, 64, 1, 68, 1, 72, 1, 76, 1, 80, 1, 84, 1, 88, 20, 88, 1, 92, 1, 96, 1, 100, 1, 104, 1, 108, 1, 112, 1, 116, 1, 120, 1, 124, 1, 93, 20, 128, 1, 132, 1, 136, 1, 140, 1, 144, 1, 148, 1, 152, 1, 156, 1, 160, 1, 164, 1, 98, 20, 45, 15, 168, 1, 172, 1, 176, 1, 180, 1, 184, 1, 188, 1, 192, 1, 196, 1, 200, 1, 173, 2, 204, 1, 208, 1, 212, 1, 216, 1, 220, 1, 224, 1, 228, 1, 232, 1, 236, 1, 240, 1, 177, 2, 244, 1, 248, 1, 252, 1, 0, 2, 4, 2, 8, 2, 12, 2, 16, 2, 20, 2, 24, 2, 181, 2, 28, 2, 32, 2, 36, 2, 40, 2, 44, 2, 48, 2, 52, 2, 56, 2, 60, 2, 64, 2, 53, 15, 68, 2, 72, 2, 76, 2, 80, 2, 84, 2, 88, 2, 92, 2, 96, 2, 100, 2, 104, 2, 17, 20, 127, 17, 108, 2, 112, 2, 116, 2, 120, 2, 124, 2, 128, 2, 132, 2, 136, 2, 140, 2, 144, 2, 50, 19, 148, 2, 152, 2, 156, 2, 160, 2, 164, 2, 168, 2, 172, 2, 176, 2, 180, 2, 184, 2, 197, 2, 188, 2, 192, 2, 196, 2, 200, 2, 204, 2, 208, 2, 212, 2, 216, 2, 220, 2, 224, 2, 55, 19, 228, 2, 232, 2, 236, 2, 240, 2, 244, 2, 248, 2, 252, 2, 205, 2, 60, 19, 213, 2, 217, 2, 221, 2, 225, 2, 21, 20, 229, 2, 132, 17, 137, 17, 192, 17, 70, 19, 202, 17, 207, 17, 212, 17, 217, 17, 222, 17, 25, 20, 227, 17, 142, 17, 147, 17, 141, 1, 145, 1, 80, 19, 153, 1, 157, 1, 161, 1, 165, 1, 29, 20, 46, 15, 152, 17, 157, 17, 162, 17, 181, 1, 90, 19, 189, 1, 95, 19, 197, 1, 201, 1, 173, 19, 205, 1, 209, 1, 100, 19, 167, 17, 172, 17, 105, 19, 229, 1, 110, 19, 237, 1, 241, 1, 177, 19, 245, 1, 177, 17, 182, 17, 1, 2, 5, 2, 120, 19, 13, 2, 125, 19, 130, 19, 25, 2, 181, 19, 187, 17, 33, 2, 135, 19, 41, 2, 45, 2, 49, 2, 53, 2, 57, 2, 140, 19, 65, 2, 185, 19, 69, 2, 73, 2, 77, 2, 81, 2, 85, 2, 89, 2, 93, 2, 97, 2, 101, 2, 105, 2, 109, 159, 0, 32, 115, 159, 0, 32, 121, 159, 0, 32, 127, 159, 0, 32, 133, 159, 0, 32, 139, 159, 0, 32, 145, 159, 0, 32, 151, 159, 0, 32, 157, 159, 0, 32, 163, 159, 0, 32, 169, 159, 0, 32, 175, 159, 0, 32, 181, 159, 0, 32, 187, 159, 0, 32, 193, 159, 0, 32, 199, 159, 0, 32, 205, 159, 0, 32, 211, 159, 0, 32, 217, 159, 0, 32, 223, 159, 0, 32, 229, 159, 0, 32, 235, 159, 0, 32, 241, 159, 0, 32, 247, 159, 0, 32, 253, 159, 0, 32, 3, 160, 0, 32, 9, 160, 0, 32, 15, 160, 0, 32, 21, 160, 0, 32, 27, 160, 0, 32, 33, 160, 0, 32, 39, 160, 0, 32, 45, 160, 0, 32, 51, 160, 0, 32, 57, 160, 0, 32, 63, 160, 0, 32, 69, 160, 0, 32, 75, 160, 0, 32, 81, 160, 0, 32, 87, 160, 0, 32, 93, 160, 0, 32, 99, 160, 0, 32, 105, 160, 0, 32, 111, 160, 0, 32, 117, 160, 0, 32, 123, 160, 0, 32, 129, 160, 0, 32, 135, 160, 0, 32, 141, 160, 0, 32, 147, 160, 0, 32, 153, 160, 0, 32, 159, 160, 0, 32, 165, 160, 0, 32, 171, 160, 0, 32, 177, 160, 0, 32, 183, 160, 0, 32, 189, 160, 0, 32, 195, 160, 0, 32, 201, 160, 0, 32, 207, 160, 0, 32, 213, 160, 0, 32, 219, 160, 0, 32, 225, 160, 0, 32, 231, 160, 0, 32, 237, 160, 0, 32, 243, 160, 0, 32, 249, 160, 0, 32, 255, 160, 0, 32, 5, 161, 0, 32, 11, 161, 0, 32, 17, 161, 0, 32, 23, 161, 0, 32, 29, 161, 0, 32, 35, 161, 0, 32, 41, 161, 0, 32, 47, 161, 0, 32, 53, 161, 0, 32, 59, 161, 0, 32, 65, 161, 0, 32, 71, 161, 0, 32, 77, 161, 0, 32, 83, 161, 0, 32, 89, 161, 0, 32, 95, 161, 0, 32, 101, 161, 0, 32, 107, 161, 0, 32, 113, 161, 0, 32, 119, 161, 0, 32, 125, 161, 0, 32, 131, 161, 0, 32, 137, 161, 0, 32, 143, 161, 0, 32, 149, 161, 0, 32, 155, 161, 0, 32, 161, 161, 0, 32, 167, 161, 0, 32, 173, 161, 0, 32, 179, 161, 0, 32, 185, 161, 0, 32, 191, 161, 0, 32, 197, 161, 0, 32, 203, 161, 0, 32, 209, 161, 0, 32, 215, 161, 0, 32, 221, 161, 0, 32, 227, 161, 0, 32, 233, 161, 0, 32, 239, 161, 0, 32, 245, 161, 0, 32, 251, 161, 0, 32, 1, 162, 0, 32, 7, 162, 0, 32, 13, 162, 0, 32, 19, 162, 0, 32, 25, 162, 0, 32, 31, 162, 0, 32, 37, 162, 0, 32, 43, 162, 0, 32, 49, 162, 0, 32, 55, 162, 0, 32, 61, 162, 0, 32, 67, 162, 0, 32, 73, 162, 0, 32, 79, 162, 0, 32, 85, 162, 0, 32, 91, 162, 0, 32, 97, 162, 0, 32, 103, 162, 0, 32, 109, 162, 0, 32, 115, 162, 0, 32, 121, 162, 0, 32, 127, 162, 0, 32, 133, 162, 0, 32, 139, 162, 0, 32, 145, 162, 0, 32, 151, 162, 0, 32, 157, 162, 0, 32, 163, 162, 0, 32, 169, 162, 0, 32, 175, 162, 0, 32, 181, 162, 0, 32, 187, 162, 0, 32, 193, 162, 0, 32, 199, 162, 0, 32, 205, 162, 0, 32, 211, 162, 0, 32, 217, 162, 0, 32, 223, 162, 0, 32, 229, 162, 0, 32, 235, 162, 0, 32, 241, 162, 0, 32, 247, 162, 0, 32, 253, 162, 0, 32, 3, 163, 0, 32, 9, 163, 0, 32, 15, 163, 0, 32, 21, 163, 0, 32, 27, 163, 0, 32, 33, 163, 0, 32, 39, 163, 0, 32, 45, 163, 0, 32, 51, 163, 0, 32, 57, 163, 0, 32, 63, 163, 0, 32, 69, 163, 0, 32, 75, 163, 0, 32, 81, 163, 0, 32, 87, 163, 0, 32, 93, 163, 0, 32, 99, 163, 0, 32, 105, 163, 0, 32, 111, 163, 0, 32, 117, 163, 0, 32, 123, 163, 0, 32, 129, 163, 0, 32, 135, 163, 0, 32, 141, 163, 0, 32, 147, 163, 0, 32, 153, 163, 0, 32, 159, 163, 0, 32, 165, 163, 0, 32, 171, 163, 0, 32, 177, 163, 0, 32, 183, 163, 0, 32, 189, 163, 0, 32, 195, 163, 0, 32, 201, 163, 0, 32, 207, 163, 0, 32, 213, 163, 0, 32, 219, 163, 0, 32, 225, 163, 0, 32, 231, 163, 0, 32, 237, 163, 0, 32, 243, 163, 0, 32, 249, 163, 0, 32, 255, 163, 0, 32, 5, 164, 0, 32, 11, 164, 0, 32, 17, 164, 0, 32, 23, 164, 0, 32, 29, 164, 0, 32, 35, 164, 0, 32, 41, 164, 0, 32, 47, 164, 0, 32, 53, 164, 0, 32, 59, 164, 0, 32, 65, 164, 0, 32, 71, 164, 0, 32, 77, 164, 0, 32, 83, 164, 0, 32, 89, 164, 0, 32, 95, 164, 0, 32, 101, 164, 0, 32, 107, 164, 0, 32, 113, 164, 0, 32, 119, 164, 0, 32, 125, 164, 0, 32, 131, 164, 0, 32, 137, 164, 0, 32, 143, 164, 0, 32, 149, 164, 0, 32, 155, 164, 0, 32, 161, 164, 0, 32, 167, 164, 0, 32, 173, 164, 0, 32, 179, 164, 0, 32, 185, 164, 0, 32, 191, 164, 0, 32, 197, 164, 0, 32, 203, 164, 0, 32, 209, 164, 0, 32, 215, 164, 0, 32, 221, 164, 0, 32, 227, 164, 0, 32, 233, 164, 0, 32, 239, 164, 0, 32, 245, 164, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 247, 164, 0, 32, 47, 199, 0, 96, 3, 0, 0, 0, 110, 2, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 95, 166, 0, 224, 1, 0, 70, 2, 99, 166, 0, 32, 2, 0, 54, 0, 65, 0, 170, 170, 101, 166, 0, 32, 47, 199, 0, 96, 3, 0, 0, 0, 119, 7, 0, 0, 12, 0, 0, 0, 31, 0, 0, 0, 1, 0, 87, 0, 107, 166, 0, 224, 3, 0, 0, 0, 120, 7, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 59, 0, 113, 166, 0, 224, 2, 0, 70, 2, 13, 20, 170, 170, 111, 166, 0, 32, 117, 166, 0, 32, 1, 0, 54, 0, 119, 166, 0, 32, 17, 0, 45, 0, 72, 0, 80, 0, 91, 0, 97, 0, 106, 0, 126, 0, 136, 0, 143, 0, 150, 0, 158, 0, 172, 0, 185, 0, 198, 0, 215, 0, 0, 3, 8, 3, 181, 158, 0, 32, 193, 158, 0, 32, 213, 158, 0, 32, 225, 158, 0, 32, 245, 158, 0, 32, 1, 159, 0, 32, 19, 159, 0, 32, 31, 159, 0, 32, 43, 159, 0, 32, 53, 159, 0, 32, 65, 159, 0, 32, 77, 159, 0, 32, 89, 159, 0, 32, 101, 159, 0, 32, 91, 166, 0, 32, 103, 166, 0, 32, 123, 166, 0, 32, 39, 0, 11, 23, 35, 3, 38, 3, 41, 3, 44, 3, 47, 3, 50, 3, 53, 3, 56, 3, 59, 3, 62, 3, 65, 3, 75, 15, 68, 3, 72, 23, 71, 3, 74, 3, 77, 3, 80, 3, 83, 3, 86, 3, 89, 3, 92, 3, 95, 3, 98, 3, 101, 3, 104, 3, 107, 3, 110, 3, 113, 3, 116, 3, 119, 3, 122, 3, 125, 3, 128, 3, 131, 3, 134, 3, 28, 26, 137, 3, 25, 213, 0, 144, 27, 213, 0, 144, 33, 213, 0, 144, 39, 213, 0, 144, 45, 213, 0, 144, 48, 213, 0, 144, 51, 213, 0, 144, 56, 213, 0, 144, 61, 213, 0, 144, 67, 213, 0, 144, 72, 213, 0, 144, 77, 213, 0, 144, 80, 213, 0, 144, 83, 213, 0, 144, 89, 213, 0, 144, 92, 213, 0, 144, 97, 213, 0, 144, 103, 213, 0, 144, 108, 213, 0, 144, 111, 213, 0, 144, 116, 213, 0, 144, 119, 213, 0, 144, 125, 213, 0, 144, 130, 213, 0, 144, 135, 213, 0, 144, 140, 213, 0, 144, 143, 213, 0, 144, 148, 213, 0, 144, 153, 213, 0, 144, 158, 213, 0, 144, 164, 213, 0, 144, 169, 213, 0, 144, 174, 213, 0, 144, 177, 213, 0, 144, 182, 213, 0, 144, 187, 213, 0, 144, 190, 213, 0, 144, 195, 213, 0, 144, 199, 213, 0, 144, 47, 1, 0, 0, 204, 213, 0, 144, 208, 213, 0, 144, 212, 213, 0, 144, 216, 213, 0, 144, 220, 213, 0, 144, 224, 213, 0, 144, 228, 213, 0, 144, 232, 213, 0, 144, 236, 213, 0, 144, 240, 213, 0, 144, 244, 213, 0, 144, 248, 213, 0, 144, 252, 213, 0, 144, 0, 214, 0, 144, 4, 214, 0, 144, 8, 214, 0, 144, 12, 214, 0, 144, 16, 214, 0, 144, 20, 214, 0, 144, 24, 214, 0, 144, 28, 214, 0, 144, 32, 214, 0, 144, 36, 214, 0, 144, 40, 214, 0, 144, 44, 214, 0, 144, 48, 214, 0, 144, 52, 214, 0, 144, 56, 214, 0, 144, 60, 214, 0, 144, 64, 214, 0, 144, 68, 214, 0, 144, 72, 214, 0, 144, 76, 214, 0, 144, 80, 214, 0, 144, 84, 214, 0, 144, 88, 214, 0, 144, 92, 214, 0, 144, 96, 214, 0, 144, 100, 214, 0, 144, 104, 214, 0, 144, 108, 214, 0, 144, 112, 214, 0, 144, 116, 214, 0, 144, 120, 214, 0, 144, 124, 214, 0, 144, 128, 214, 0, 144, 132, 214, 0, 144, 136, 214, 0, 144, 140, 214, 0, 144, 144, 214, 0, 144, 148, 214, 0, 144, 152, 214, 0, 144, 156, 214, 0, 144, 160, 214, 0, 144, 164, 214, 0, 144, 168, 214, 0, 144, 172, 214, 0, 144, 176, 214, 0, 144, 180, 214, 0, 144, 184, 214, 0, 144, 188, 214, 0, 144, 192, 214, 0, 144, 196, 214, 0, 144, 200, 214, 0, 144, 204, 214, 0, 144, 208, 214, 0, 144, 212, 214, 0, 144, 216, 214, 0, 144, 220, 214, 0, 144, 224, 214, 0, 144, 228, 214, 0, 144, 232, 214, 0, 144, 236, 214, 0, 144, 240, 214, 0, 144, 244, 214, 0, 144, 248, 214, 0, 144, 252, 214, 0, 144, 0, 215, 0, 144, 4, 215, 0, 144, 8, 215, 0, 144, 12, 215, 0, 144, 16, 215, 0, 144, 20, 215, 0, 144, 24, 215, 0, 144, 28, 215, 0, 144, 32, 215, 0, 144, 36, 215, 0, 144, 40, 215, 0, 144, 44, 215, 0, 144, 48, 215, 0, 144, 52, 215, 0, 144, 56, 215, 0, 144, 60, 215, 0, 144, 64, 215, 0, 144, 68, 215, 0, 144, 72, 215, 0, 144, 76, 215, 0, 144, 80, 215, 0, 144, 84, 215, 0, 144, 88, 215, 0, 144, 92, 215, 0, 144, 96, 215, 0, 144, 100, 215, 0, 144, 104, 215, 0, 144, 108, 215, 0, 144, 112, 215, 0, 144, 116, 215, 0, 144, 120, 215, 0, 144, 124, 215, 0, 144, 128, 215, 0, 144, 132, 215, 0, 144, 136, 215, 0, 144, 140, 215, 0, 144, 144, 215, 0, 144, 148, 215, 0, 144, 152, 215, 0, 144, 156, 215, 0, 144, 160, 215, 0, 144, 164, 215, 0, 144, 168, 215, 0, 144, 172, 215, 0, 144, 176, 215, 0, 144, 180, 215, 0, 144, 184, 215, 0, 144, 188, 215, 0, 144, 192, 215, 0, 144, 196, 215, 0, 144, 200, 215, 0, 144, 204, 215, 0, 144, 208, 215, 0, 144, 212, 215, 0, 144, 216, 215, 0, 144, 220, 215, 0, 144, 224, 215, 0, 144, 228, 215, 0, 144, 232, 215, 0, 144, 236, 215, 0, 144, 240, 215, 0, 144, 244, 215, 0, 144, 248, 215, 0, 144, 252, 215, 0, 144, 0, 216, 0, 144, 4, 216, 0, 144, 8, 216, 0, 144, 12, 216, 0, 144, 16, 216, 0, 144, 20, 216, 0, 144, 24, 216, 0, 144, 28, 216, 0, 144, 32, 216, 0, 144, 36, 216, 0, 144, 40, 216, 0, 144, 44, 216, 0, 144, 48, 216, 0, 144, 52, 216, 0, 144, 56, 216, 0, 144, 60, 216, 0, 144, 64, 216, 0, 144, 68, 216, 0, 144, 72, 216, 0, 144, 76, 216, 0, 144, 80, 216, 0, 144, 84, 216, 0, 144, 88, 216, 0, 144, 92, 216, 0, 144, 96, 216, 0, 144, 100, 216, 0, 144, 104, 216, 0, 144, 108, 216, 0, 144, 112, 216, 0, 144, 116, 216, 0, 144, 120, 216, 0, 144, 124, 216, 0, 144, 128, 216, 0, 144, 132, 216, 0, 144, 136, 216, 0, 144, 140, 216, 0, 144, 144, 216, 0, 144, 148, 216, 0, 144, 152, 216, 0, 144, 156, 216, 0, 144, 160, 216, 0, 144, 164, 216, 0, 144, 168, 216, 0, 144, 172, 216, 0, 144, 176, 216, 0, 144, 180, 216, 0, 144, 184, 216, 0, 144, 188, 216, 0, 144, 192, 216, 0, 144, 196, 216, 0, 144, 200, 216, 0, 144, 204, 216, 0, 144, 208, 216, 0, 144, 212, 216, 0, 144, 216, 216, 0, 144, 220, 216, 0, 144, 224, 216, 0, 144, 228, 216, 0, 144, 232, 216, 0, 144, 236, 216, 0, 144, 240, 216, 0, 144, 244, 216, 0, 144, 248, 216, 0, 144, 252, 216, 0, 144, 0, 217, 0, 144, 4, 217, 0, 144, 8, 217, 0, 144, 12, 217, 0, 144, 16, 217, 0, 144, 20, 217, 0, 144, 24, 217, 0, 144, 28, 217, 0, 144, 32, 217, 0, 144, 36, 217, 0, 144, 40, 217, 0, 144, 44, 217, 0, 144, 48, 217, 0, 144, 52, 217, 0, 144, 56, 217, 0, 144, 60, 217, 0, 144, 64, 217, 0, 144, 68, 217, 0, 144, 72, 217, 0, 144, 76, 217, 0, 144, 80, 217, 0, 144, 84, 217, 0, 144, 88, 217, 0, 144, 92, 217, 0, 144, 96, 217, 0, 144, 100, 217, 0, 144, 104, 217, 0, 144, 108, 217, 0, 144, 112, 217, 0, 144, 116, 217, 0, 144, 120, 217, 0, 144, 124, 217, 0, 144, 128, 217, 0, 144, 132, 217, 0, 144, 136, 217, 0, 144, 140, 217, 0, 144, 144, 217, 0, 144, 148, 217, 0, 144, 152, 217, 0, 144, 156, 217, 0, 144, 160, 217, 0, 144, 164, 217, 0, 144, 168, 217, 0, 144, 172, 217, 0, 144, 176, 217, 0, 144, 180, 217, 0, 144, 184, 217, 0, 144, 188, 217, 0, 144, 192, 217, 0, 144, 196, 217, 0, 144, 200, 217, 0, 144, 204, 217, 0, 144, 208, 217, 0, 144, 212, 217, 0, 144, 216, 217, 0, 144, 220, 217, 0, 144, 224, 217, 0, 144, 228, 217, 0, 144, 232, 217, 0, 144, 236, 217, 0, 144, 240, 217, 0, 144, 244, 217, 0, 144, 248, 217, 0, 144, 252, 217, 0, 144, 0, 218, 0, 144, 4, 218, 0, 144, 8, 218, 0, 144, 12, 218, 0, 144, 16, 218, 0, 144, 20, 218, 0, 144, 24, 218, 0, 144, 28, 218, 0, 144, 32, 218, 0, 144, 36, 218, 0, 144, 40, 218, 0, 144, 44, 218, 0, 144, 48, 218, 0, 144, 52, 218, 0, 144, 56, 218, 0, 144, 60, 218, 0, 144, 64, 218, 0, 144, 68, 218, 0, 144, 72, 218, 0, 144, 76, 218, 0, 144, 80, 218, 0, 144, 84, 218, 0, 144, 88, 218, 0, 144, 92, 218, 0, 144, 96, 218, 0, 144, 100, 218, 0, 144, 104, 218, 0, 144, 108, 218, 0, 144, 112, 218, 0, 144, 116, 218, 0, 144, 120, 218, 0, 144, 124, 218, 0, 144, 128, 218, 0, 144, 132, 218, 0, 144, 182, 0, 0, 0, 136, 218, 0, 144, 139, 218, 0, 144, 142, 218, 0, 144, 145, 218, 0, 144, 148, 218, 0, 144, 151, 218, 0, 144, 154, 218, 0, 144, 157, 218, 0, 144, 160, 218, 0, 144, 163, 218, 0, 144, 166, 218, 0, 144, 169, 218, 0, 144, 172, 218, 0, 144, 175, 218, 0, 144, 178, 218, 0, 144, 181, 218, 0, 144, 184, 218, 0, 144, 187, 218, 0, 144, 190, 218, 0, 144, 193, 218, 0, 144, 196, 218, 0, 144, 199, 218, 0, 144, 202, 218, 0, 144, 205, 218, 0, 144, 208, 218, 0, 144, 211, 218, 0, 144, 214, 218, 0, 144, 217, 218, 0, 144, 220, 218, 0, 144, 223, 218, 0, 144, 226, 218, 0, 144, 229, 218, 0, 144, 232, 218, 0, 144, 235, 218, 0, 144, 238, 218, 0, 144, 241, 218, 0, 144, 244, 218, 0, 144, 247, 218, 0, 144, 250, 218, 0, 144, 253, 218, 0, 144, 0, 219, 0, 144, 3, 219, 0, 144, 6, 219, 0, 144, 9, 219, 0, 144, 12, 219, 0, 144, 15, 219, 0, 144, 18, 219, 0, 144, 21, 219, 0, 144, 24, 219, 0, 144, 27, 219, 0, 144, 30, 219, 0, 144, 33, 219, 0, 144, 36, 219, 0, 144, 39, 219, 0, 144, 42, 219, 0, 144, 45, 219, 0, 144, 48, 219, 0, 144, 51, 219, 0, 144, 54, 219, 0, 144, 57, 219, 0, 144, 60, 219, 0, 144, 63, 219, 0, 144, 66, 219, 0, 144, 69, 219, 0, 144, 72, 219, 0, 144, 75, 219, 0, 144, 78, 219, 0, 144, 81, 219, 0, 144, 84, 219, 0, 144, 87, 219, 0, 144, 90, 219, 0, 144, 93, 219, 0, 144, 96, 219, 0, 144, 99, 219, 0, 144, 102, 219, 0, 144, 105, 219, 0, 144, 108, 219, 0, 144, 111, 219, 0, 144, 114, 219, 0, 144, 117, 219, 0, 144, 120, 219, 0, 144, 123, 219, 0, 144, 126, 219, 0, 144, 129, 219, 0, 144, 132, 219, 0, 144, 135, 219, 0, 144, 138, 219, 0, 144, 141, 219, 0, 144, 144, 219, 0, 144, 147, 219, 0, 144, 150, 219, 0, 144, 153, 219, 0, 144, 156, 219, 0, 144, 159, 219, 0, 144, 162, 219, 0, 144, 165, 219, 0, 144, 168, 219, 0, 144, 171, 219, 0, 144, 174, 219, 0, 144, 177, 219, 0, 144, 180, 219, 0, 144, 183, 219, 0, 144, 186, 219, 0, 144, 189, 219, 0, 144, 192, 219, 0, 144, 195, 219, 0, 144, 198, 219, 0, 144, 201, 219, 0, 144, 204, 219, 0, 144, 207, 219, 0, 144, 210, 219, 0, 144, 213, 219, 0, 144, 216, 219, 0, 144, 219, 219, 0, 144, 222, 219, 0, 144, 225, 219, 0, 144, 228, 219, 0, 144, 231, 219, 0, 144, 234, 219, 0, 144, 237, 219, 0, 144, 240, 219, 0, 144, 243, 219, 0, 144, 246, 219, 0, 144, 249, 219, 0, 144, 252, 219, 0, 144, 255, 219, 0, 144, 2, 220, 0, 144, 5, 220, 0, 144, 8, 220, 0, 144, 11, 220, 0, 144, 14, 220, 0, 144, 17, 220, 0, 144, 20, 220, 0, 144, 23, 220, 0, 144, 26, 220, 0, 144, 29, 220, 0, 144, 32, 220, 0, 144, 35, 220, 0, 144, 38, 220, 0, 144, 41, 220, 0, 144, 44, 220, 0, 144, 47, 220, 0, 144, 50, 220, 0, 144, 53, 220, 0, 144, 56, 220, 0, 144, 59, 220, 0, 144, 62, 220, 0, 144, 65, 220, 0, 144, 68, 220, 0, 144, 71, 220, 0, 144, 74, 220, 0, 144, 77, 220, 0, 144, 80, 220, 0, 144, 83, 220, 0, 144, 86, 220, 0, 144, 89, 220, 0, 144, 92, 220, 0, 144, 95, 220, 0, 144, 98, 220, 0, 144, 101, 220, 0, 144, 104, 220, 0, 144, 107, 220, 0, 144, 110, 220, 0, 144, 113, 220, 0, 144, 116, 220, 0, 144, 119, 220, 0, 144, 122, 220, 0, 144, 125, 220, 0, 144, 128, 220, 0, 144, 131, 220, 0, 144, 134, 220, 0, 144, 137, 220, 0, 144, 140, 220, 0, 144, 143, 220, 0, 144, 146, 220, 0, 144, 149, 220, 0, 144, 152, 220, 0, 144, 155, 220, 0, 144, 158, 220, 0, 144, 161, 220, 0, 144, 164, 220, 0, 144, 167, 220, 0, 144, 3, 0, 206, 3, 217, 3, 225, 3, 170, 220, 0, 144, 53, 221, 0, 144, 95, 2, 0, 96, 6, 0, 206, 3, 242, 3, 217, 3, 254, 3, 7, 4, 225, 3, 170, 170, 209, 221, 0, 144, 191, 222, 0, 144, 207, 222, 0, 144, 234, 232, 0, 144, 241, 232, 0, 144, 165, 199, 0, 96, 7, 0, 206, 3, 15, 4, 242, 3, 217, 3, 254, 3, 7, 4, 225, 3, 245, 232, 0, 144, 2, 233, 0, 144, 27, 233, 0, 144, 30, 233, 0, 144, 151, 233, 0, 144, 217, 72, 0, 96, 88, 3, 0, 96, 6, 0, 206, 3, 242, 3, 217, 3, 254, 3, 7, 4, 225, 3, 170, 170, 147, 72, 0, 96, 156, 233, 0, 144, 159, 233, 0, 144, 61, 234, 0, 144, 65, 234, 0, 144, 16, 73, 0, 96, 3, 0, 206, 3, 217, 3, 225, 3, 74, 234, 0, 144, 18, 235, 0, 144, 105, 246, 0, 144, 2, 0, 206, 3, 217, 3, 170, 170, 195, 205, 0, 96, 106, 247, 0, 144, 2, 0, 206, 3, 217, 3, 170, 170, 23, 248, 0, 144, 27, 248, 0, 144, 7, 0, 197, 3, 233, 3, 20, 4, 27, 4, 34, 4, 46, 4, 253, 25, 185, 168, 0, 32, 190, 168, 0, 32, 200, 168, 0, 32, 211, 168, 0, 32, 221, 168, 0, 32, 226, 168, 0, 32, 230, 168, 0, 32, 1, 0, 90, 4, 127, 248, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 124, 248, 0, 80, 245, 168, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 130, 248, 0, 80, 133, 248, 0, 80, 1, 0, 64, 4, 136, 248, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 139, 248, 0, 80, 142, 248, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 145, 248, 0, 80, 148, 248, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 151, 248, 0, 80, 154, 248, 0, 80, 1, 0, 64, 4, 157, 248, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 160, 248, 0, 80, 163, 248, 0, 80, 1, 0, 80, 4, 166, 248, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 169, 248, 0, 80, 172, 248, 0, 80, 1, 0, 90, 4, 178, 248, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 175, 248, 0, 80, 25, 169, 0, 32, 1, 0, 64, 4, 181, 248, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 184, 248, 0, 80, 187, 248, 0, 80, 1, 0, 72, 4, 190, 248, 0, 144, 1, 0, 80, 4, 37, 169, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 193, 248, 0, 80, 196, 248, 0, 80, 1, 0, 80, 4, 199, 248, 0, 80, 1, 0, 64, 4, 202, 248, 0, 80, 1, 0, 64, 4, 205, 248, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 208, 248, 0, 80, 211, 248, 0, 80, 1, 0, 64, 4, 214, 248, 0, 80, 1, 0, 64, 4, 217, 248, 0, 80, 1, 0, 64, 4, 222, 248, 0, 80, 1, 0, 64, 4, 225, 248, 0, 80, 1, 0, 80, 4, 228, 248, 0, 80, 1, 0, 64, 4, 231, 248, 0, 80, 1, 0, 64, 4, 234, 248, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 200, 191, 0, 96, 237, 248, 0, 144, 2, 0, 72, 4, 90, 4, 170, 170, 155, 39, 0, 96, 8, 249, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 69, 169, 0, 32, 73, 169, 0, 32, 1, 0, 72, 4, 11, 249, 0, 144, 1, 0, 80, 4, 81, 169, 0, 32, 1, 0, 64, 4, 15, 249, 0, 80, 1, 0, 64, 4, 18, 249, 0, 80, 1, 0, 64, 4, 21, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 24, 249, 0, 80, 27, 249, 0, 80, 1, 0, 64, 4, 30, 249, 0, 80, 1, 0, 80, 4, 33, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 36, 249, 0, 80, 39, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 42, 249, 0, 80, 45, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 48, 249, 0, 80, 51, 249, 0, 80, 1, 0, 64, 4, 54, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 57, 249, 0, 80, 60, 249, 0, 80, 1, 0, 64, 4, 63, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 66, 249, 0, 80, 69, 249, 0, 80, 1, 0, 80, 4, 72, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 75, 249, 0, 80, 78, 249, 0, 80, 1, 0, 64, 4, 81, 249, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 86, 249, 0, 144, 212, 1, 0, 96, 1, 0, 90, 4, 90, 249, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 131, 169, 0, 32, 135, 169, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 94, 249, 0, 80, 97, 249, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 103, 249, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 100, 249, 0, 80, 145, 169, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 107, 249, 0, 80, 110, 249, 0, 80, 1, 0, 64, 4, 115, 249, 0, 80, 1, 0, 90, 4, 121, 249, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 118, 249, 0, 80, 159, 169, 0, 32, 1, 0, 64, 4, 124, 249, 0, 80, 1, 0, 64, 4, 127, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 130, 249, 0, 80, 133, 249, 0, 80, 1, 0, 64, 4, 138, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 141, 249, 0, 80, 144, 249, 0, 80, 1, 0, 64, 4, 147, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 152, 249, 0, 80, 155, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 158, 249, 0, 80, 161, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 164, 249, 0, 80, 167, 249, 0, 80, 1, 0, 64, 4, 170, 249, 0, 80, 1, 0, 64, 4, 173, 249, 0, 80, 1, 0, 64, 4, 176, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 179, 249, 0, 80, 182, 249, 0, 80, 1, 0, 64, 4, 185, 249, 0, 80, 1, 0, 64, 4, 188, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 193, 249, 0, 80, 196, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 199, 249, 0, 80, 202, 249, 0, 80, 1, 0, 64, 4, 205, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 208, 249, 0, 80, 211, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 214, 249, 0, 80, 217, 249, 0, 80, 1, 0, 80, 4, 220, 249, 0, 80, 1, 0, 90, 4, 226, 249, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 223, 249, 0, 80, 227, 169, 0, 32, 1, 0, 64, 4, 230, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 235, 249, 0, 80, 238, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 241, 249, 0, 80, 244, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 247, 249, 0, 80, 250, 249, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 253, 249, 0, 80, 0, 250, 0, 80, 1, 0, 80, 4, 3, 250, 0, 80, 1, 0, 64, 4, 6, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 9, 250, 0, 80, 12, 250, 0, 80, 1, 0, 64, 4, 15, 250, 0, 80, 1, 0, 64, 4, 18, 250, 0, 80, 1, 0, 72, 4, 21, 250, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 7, 170, 0, 32, 24, 250, 0, 80, 1, 0, 64, 4, 27, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 30, 250, 0, 80, 35, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 38, 250, 0, 80, 41, 250, 0, 80, 1, 0, 64, 4, 44, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 47, 250, 0, 80, 50, 250, 0, 80, 1, 0, 64, 4, 53, 250, 0, 80, 1, 0, 64, 4, 56, 250, 0, 80, 1, 0, 64, 4, 59, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 62, 250, 0, 80, 65, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 70, 250, 0, 80, 73, 250, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 76, 250, 0, 144, 156, 1, 0, 96, 1, 0, 64, 4, 43, 170, 0, 32, 1, 0, 80, 4, 79, 250, 0, 80, 1, 0, 64, 4, 82, 250, 0, 80, 1, 0, 64, 4, 85, 250, 0, 80, 1, 0, 64, 4, 88, 250, 0, 80, 1, 0, 64, 4, 91, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 94, 250, 0, 80, 97, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 100, 250, 0, 80, 103, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 108, 250, 0, 80, 111, 250, 0, 80, 1, 0, 64, 4, 114, 250, 0, 80, 1, 0, 64, 4, 117, 250, 0, 80, 1, 0, 64, 4, 120, 250, 0, 80, 1, 0, 64, 4, 123, 250, 0, 80, 1, 0, 64, 4, 126, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 129, 250, 0, 80, 134, 250, 0, 80, 1, 0, 64, 4, 137, 250, 0, 80, 1, 0, 64, 4, 140, 250, 0, 80, 1, 0, 64, 4, 143, 250, 0, 80, 1, 0, 64, 4, 146, 250, 0, 80, 1, 0, 72, 4, 149, 250, 0, 144, 1, 0, 64, 4, 93, 170, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 152, 250, 0, 80, 155, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 158, 250, 0, 80, 161, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 164, 250, 0, 80, 167, 250, 0, 80, 1, 0, 64, 4, 170, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 175, 250, 0, 80, 178, 250, 0, 80, 1, 0, 64, 4, 181, 250, 0, 80, 1, 0, 80, 4, 184, 250, 0, 80, 1, 0, 64, 4, 187, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 190, 250, 0, 80, 193, 250, 0, 80, 1, 0, 64, 4, 196, 250, 0, 80, 1, 0, 64, 4, 199, 250, 0, 80, 1, 0, 64, 4, 202, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 205, 250, 0, 80, 208, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 211, 250, 0, 80, 214, 250, 0, 80, 1, 0, 64, 4, 217, 250, 0, 80, 1, 0, 90, 4, 223, 250, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 220, 250, 0, 80, 141, 170, 0, 32, 1, 0, 64, 4, 226, 250, 0, 80, 1, 0, 64, 4, 229, 250, 0, 80, 1, 0, 72, 4, 232, 250, 0, 144, 1, 0, 80, 4, 151, 170, 0, 32, 1, 0, 64, 4, 236, 250, 0, 80, 1, 0, 72, 4, 239, 250, 0, 144, 1, 0, 64, 4, 157, 170, 0, 32, 1, 0, 64, 4, 242, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 245, 250, 0, 80, 248, 250, 0, 80, 1, 0, 64, 4, 251, 250, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 254, 250, 0, 80, 1, 251, 0, 80, 1, 0, 64, 4, 4, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 7, 251, 0, 80, 10, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 13, 251, 0, 80, 18, 251, 0, 80, 1, 0, 80, 4, 21, 251, 0, 80, 1, 0, 64, 4, 26, 251, 0, 80, 1, 0, 64, 4, 29, 251, 0, 80, 1, 0, 80, 4, 32, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 35, 251, 0, 80, 38, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 41, 251, 0, 80, 44, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 47, 251, 0, 80, 52, 251, 0, 80, 1, 0, 64, 4, 55, 251, 0, 80, 1, 0, 64, 4, 58, 251, 0, 80, 1, 0, 64, 4, 61, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 64, 251, 0, 80, 67, 251, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 70, 251, 0, 144, 2, 0, 72, 4, 90, 4, 170, 170, 215, 39, 0, 96, 77, 251, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 213, 170, 0, 32, 217, 170, 0, 32, 1, 0, 64, 4, 90, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 93, 251, 0, 80, 96, 251, 0, 80, 1, 0, 64, 4, 99, 251, 0, 80, 1, 0, 64, 4, 102, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 105, 251, 0, 80, 108, 251, 0, 80, 1, 0, 64, 4, 111, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 114, 251, 0, 80, 117, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 120, 251, 0, 80, 123, 251, 0, 80, 1, 0, 64, 4, 128, 251, 0, 80, 1, 0, 64, 4, 131, 251, 0, 80, 1, 0, 64, 4, 134, 251, 0, 80, 1, 0, 64, 4, 137, 251, 0, 80, 1, 0, 64, 4, 140, 251, 0, 80, 1, 0, 80, 4, 143, 251, 0, 80, 1, 0, 64, 4, 146, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 151, 251, 0, 80, 154, 251, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 159, 251, 0, 80, 162, 251, 0, 80, 1, 0, 64, 4, 165, 251, 0, 80, 1, 0, 64, 4, 170, 251, 0, 80, 1, 0, 90, 4, 176, 251, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 173, 251, 0, 80, 19, 171, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 179, 251, 0, 80, 182, 251, 0, 80, 1, 0, 64, 4, 185, 251, 0, 80, 1, 0, 80, 4, 188, 251, 0, 80, 1, 0, 64, 4, 191, 251, 0, 80, 1, 0, 64, 4, 194, 251, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 115, 39, 0, 96, 197, 251, 0, 144, 1, 0, 64, 4, 37, 171, 0, 32, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 200, 251, 0, 144, 2, 0, 72, 4, 90, 4, 170, 170, 34, 252, 0, 144, 37, 252, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 43, 171, 0, 32, 47, 171, 0, 32, 1, 0, 80, 4, 87, 252, 0, 80, 1, 0, 64, 4, 90, 252, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 93, 252, 0, 144, 1, 0, 90, 4, 117, 252, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 59, 171, 0, 32, 63, 171, 0, 32, 1, 0, 64, 4, 127, 252, 0, 80, 1, 0, 64, 4, 130, 252, 0, 80, 1, 0, 72, 4, 135, 252, 0, 144, 1, 0, 80, 4, 73, 171, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 138, 252, 0, 80, 141, 252, 0, 80, 1, 0, 64, 4, 144, 252, 0, 80, 1, 0, 64, 4, 147, 252, 0, 80, 1, 0, 64, 4, 150, 252, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 200, 191, 0, 96, 153, 252, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 87, 171, 0, 32, 156, 252, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 159, 252, 0, 80, 162, 252, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 192, 0, 96, 168, 252, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 165, 252, 0, 80, 99, 171, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 171, 252, 0, 80, 174, 252, 0, 80, 1, 0, 90, 4, 182, 252, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 177, 252, 0, 80, 111, 171, 0, 32, 1, 0, 64, 4, 185, 252, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 188, 252, 0, 80, 193, 252, 0, 80, 1, 0, 64, 4, 198, 252, 0, 80, 1, 0, 64, 4, 201, 252, 0, 80, 1, 0, 64, 4, 206, 252, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 211, 252, 0, 80, 214, 252, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 217, 252, 0, 144, 2, 0, 72, 4, 90, 4, 170, 170, 225, 44, 0, 96, 7, 253, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 133, 171, 0, 32, 137, 171, 0, 32, 1, 0, 64, 4, 16, 253, 0, 80, 1, 0, 80, 4, 19, 253, 0, 80, 1, 0, 80, 4, 22, 253, 0, 80, 1, 0, 64, 4, 25, 253, 0, 80, 1, 0, 64, 4, 28, 253, 0, 80, 1, 0, 64, 4, 31, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 34, 253, 0, 80, 37, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 40, 253, 0, 80, 43, 253, 0, 80, 1, 0, 64, 4, 46, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 49, 253, 0, 80, 52, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 55, 253, 0, 80, 58, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 61, 253, 0, 80, 64, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 67, 253, 0, 80, 72, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 75, 253, 0, 80, 78, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 81, 253, 0, 80, 84, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 87, 253, 0, 80, 90, 253, 0, 80, 1, 0, 64, 4, 93, 253, 0, 80, 1, 0, 64, 4, 96, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 99, 253, 0, 80, 102, 253, 0, 80, 1, 0, 64, 4, 105, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 108, 253, 0, 80, 111, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 114, 253, 0, 80, 117, 253, 0, 80, 1, 0, 80, 4, 120, 253, 0, 80, 1, 0, 64, 4, 123, 253, 0, 80, 1, 0, 64, 4, 128, 253, 0, 80, 1, 0, 64, 4, 131, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 134, 253, 0, 80, 137, 253, 0, 80, 1, 0, 64, 4, 140, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 143, 253, 0, 80, 146, 253, 0, 80, 1, 0, 80, 4, 149, 253, 0, 80, 1, 0, 64, 4, 152, 253, 0, 80, 1, 0, 80, 4, 157, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 160, 253, 0, 80, 163, 253, 0, 80, 1, 0, 72, 4, 166, 253, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 241, 171, 0, 32, 169, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 172, 253, 0, 80, 175, 253, 0, 80, 1, 0, 64, 4, 178, 253, 0, 80, 1, 0, 80, 4, 181, 253, 0, 80, 1, 0, 64, 4, 184, 253, 0, 80, 1, 0, 72, 4, 187, 253, 0, 144, 1, 0, 80, 4, 1, 172, 0, 32, 1, 0, 64, 4, 191, 253, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 194, 253, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 7, 172, 0, 32, 197, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 200, 253, 0, 80, 203, 253, 0, 80, 1, 0, 64, 4, 206, 253, 0, 80, 1, 0, 64, 4, 209, 253, 0, 80, 1, 0, 64, 4, 212, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 215, 253, 0, 80, 218, 253, 0, 80, 1, 0, 64, 4, 221, 253, 0, 80, 1, 0, 64, 4, 226, 253, 0, 80, 1, 0, 64, 4, 229, 253, 0, 80, 1, 0, 72, 4, 232, 253, 0, 144, 1, 0, 90, 4, 235, 253, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 35, 172, 0, 32, 37, 172, 0, 32, 1, 0, 64, 4, 238, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 241, 253, 0, 80, 244, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 247, 253, 0, 80, 250, 253, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 253, 253, 0, 80, 0, 254, 0, 80, 1, 0, 64, 4, 3, 254, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 59, 45, 0, 96, 13, 254, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 8, 254, 0, 80, 59, 172, 0, 32, 2, 0, 72, 4, 90, 4, 170, 170, 16, 254, 0, 144, 238, 106, 0, 96, 1, 0, 64, 4, 67, 172, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 19, 254, 0, 80, 22, 254, 0, 80, 1, 0, 80, 4, 25, 254, 0, 80, 1, 0, 64, 4, 28, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 31, 254, 0, 80, 34, 254, 0, 80, 1, 0, 64, 4, 37, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 40, 254, 0, 80, 43, 254, 0, 80, 1, 0, 64, 4, 46, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 49, 254, 0, 80, 52, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 55, 254, 0, 80, 58, 254, 0, 80, 1, 0, 64, 4, 61, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 66, 254, 0, 80, 69, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 74, 254, 0, 80, 77, 254, 0, 80, 1, 0, 64, 4, 80, 254, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 83, 254, 0, 144, 1, 0, 90, 4, 86, 254, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 113, 172, 0, 32, 117, 172, 0, 32, 1, 0, 64, 4, 90, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 93, 254, 0, 80, 96, 254, 0, 80, 1, 0, 64, 4, 99, 254, 0, 80, 1, 0, 90, 4, 109, 254, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 104, 254, 0, 80, 131, 172, 0, 32, 1, 0, 64, 4, 113, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 116, 254, 0, 80, 121, 254, 0, 80, 1, 0, 64, 4, 124, 254, 0, 80, 1, 0, 80, 4, 127, 254, 0, 80, 1, 0, 64, 4, 130, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 133, 254, 0, 80, 136, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 139, 254, 0, 80, 144, 254, 0, 80, 1, 0, 64, 4, 147, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 150, 254, 0, 80, 153, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 156, 254, 0, 80, 159, 254, 0, 80, 1, 0, 64, 4, 164, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 167, 254, 0, 80, 170, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 173, 254, 0, 80, 176, 254, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 182, 254, 0, 144, 40, 1, 0, 96, 2, 0, 64, 4, 80, 4, 170, 170, 179, 254, 0, 80, 177, 172, 0, 32, 1, 0, 64, 4, 185, 254, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 190, 254, 0, 144, 1, 0, 90, 4, 195, 254, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 187, 172, 0, 32, 191, 172, 0, 32, 1, 0, 72, 4, 201, 254, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 197, 172, 0, 32, 204, 254, 0, 80, 1, 0, 64, 4, 207, 254, 0, 80, 1, 0, 64, 4, 210, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 215, 254, 0, 80, 218, 254, 0, 80, 1, 0, 64, 4, 221, 254, 0, 80, 1, 0, 64, 4, 224, 254, 0, 80, 1, 0, 64, 4, 227, 254, 0, 80, 1, 0, 64, 4, 230, 254, 0, 80, 1, 0, 64, 4, 233, 254, 0, 80, 1, 0, 80, 4, 236, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 239, 254, 0, 80, 242, 254, 0, 80, 1, 0, 64, 4, 247, 254, 0, 80, 1, 0, 64, 4, 252, 254, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 255, 254, 0, 80, 2, 255, 0, 80, 1, 0, 64, 4, 5, 255, 0, 80, 1, 0, 64, 4, 8, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 11, 255, 0, 80, 14, 255, 0, 80, 1, 0, 64, 4, 17, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 20, 255, 0, 80, 23, 255, 0, 80, 1, 0, 64, 4, 26, 255, 0, 80, 1, 0, 64, 4, 29, 255, 0, 80, 1, 0, 64, 4, 32, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 35, 255, 0, 80, 38, 255, 0, 80, 1, 0, 64, 4, 41, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 44, 255, 0, 80, 47, 255, 0, 80, 1, 0, 64, 4, 50, 255, 0, 80, 1, 0, 64, 4, 53, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 56, 255, 0, 80, 59, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 62, 255, 0, 80, 65, 255, 0, 80, 1, 0, 64, 4, 68, 255, 0, 80, 1, 0, 64, 4, 71, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 74, 255, 0, 80, 77, 255, 0, 80, 1, 0, 64, 4, 82, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 85, 255, 0, 80, 88, 255, 0, 80, 1, 0, 64, 4, 91, 255, 0, 80, 1, 0, 64, 4, 94, 255, 0, 80, 1, 0, 64, 4, 97, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 100, 255, 0, 80, 103, 255, 0, 80, 1, 0, 64, 4, 106, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 109, 255, 0, 80, 112, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 115, 255, 0, 80, 118, 255, 0, 80, 1, 0, 64, 4, 121, 255, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 124, 255, 0, 144, 37, 0, 0, 96, 2, 0, 64, 4, 80, 4, 170, 170, 57, 173, 0, 32, 127, 255, 0, 80, 1, 0, 64, 4, 130, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 133, 255, 0, 80, 138, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 141, 255, 0, 80, 144, 255, 0, 80, 1, 0, 64, 4, 147, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 152, 255, 0, 80, 155, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 158, 255, 0, 80, 161, 255, 0, 80, 1, 0, 80, 4, 164, 255, 0, 80, 2, 0, 72, 4, 90, 4, 170, 170, 24, 45, 0, 96, 167, 255, 0, 144, 1, 0, 90, 4, 170, 255, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 87, 173, 0, 32, 91, 173, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 173, 255, 0, 80, 176, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 179, 255, 0, 80, 182, 255, 0, 80, 1, 0, 64, 4, 185, 255, 0, 80, 1, 0, 64, 4, 188, 255, 0, 80, 1, 0, 64, 4, 191, 255, 0, 80, 1, 0, 64, 4, 194, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 197, 255, 0, 80, 200, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 203, 255, 0, 80, 206, 255, 0, 80, 1, 0, 64, 4, 209, 255, 0, 80, 1, 0, 64, 4, 212, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 215, 255, 0, 80, 218, 255, 0, 80, 1, 0, 72, 4, 221, 255, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 129, 173, 0, 32, 224, 255, 0, 80, 1, 0, 64, 4, 227, 255, 0, 80, 1, 0, 64, 4, 230, 255, 0, 80, 1, 0, 64, 4, 233, 255, 0, 80, 1, 0, 72, 4, 236, 255, 0, 144, 1, 0, 90, 4, 240, 255, 0, 144, 2, 0, 64, 4, 80, 4, 170, 170, 141, 173, 0, 32, 143, 173, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 243, 255, 0, 80, 246, 255, 0, 80, 1, 0, 64, 4, 249, 255, 0, 80, 2, 0, 64, 4, 80, 4, 170, 170, 252, 255, 0, 80, 255, 255, 0, 80, 1, 0, 64, 4, 4, 0, 1, 80, 1, 0, 64, 4, 7, 0, 1, 80, 1, 0, 64, 4, 10, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 13, 0, 1, 80, 16, 0, 1, 80, 1, 0, 64, 4, 19, 0, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 22, 0, 1, 144, 208, 1, 0, 96, 1, 0, 64, 4, 171, 173, 0, 32, 1, 0, 64, 4, 26, 0, 1, 80, 1, 0, 80, 4, 29, 0, 1, 80, 1, 0, 80, 4, 34, 0, 1, 80, 1, 0, 64, 4, 37, 0, 1, 80, 1, 0, 64, 4, 40, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 43, 0, 1, 80, 46, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 49, 0, 1, 80, 52, 0, 1, 80, 1, 0, 64, 4, 55, 0, 1, 80, 1, 0, 64, 4, 58, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 63, 0, 1, 80, 66, 0, 1, 80, 1, 0, 64, 4, 69, 0, 1, 80, 1, 0, 64, 4, 72, 0, 1, 80, 1, 0, 64, 4, 75, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 78, 0, 1, 80, 81, 0, 1, 80, 1, 0, 72, 4, 86, 0, 1, 144, 1, 0, 80, 4, 213, 173, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 89, 0, 1, 80, 92, 0, 1, 80, 1, 0, 64, 4, 95, 0, 1, 80, 1, 0, 72, 4, 98, 0, 1, 144, 1, 0, 64, 4, 223, 173, 0, 32, 1, 0, 64, 4, 101, 0, 1, 80, 1, 0, 64, 4, 104, 0, 1, 80, 1, 0, 80, 4, 107, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 110, 0, 1, 80, 113, 0, 1, 80, 1, 0, 64, 4, 116, 0, 1, 80, 1, 0, 64, 4, 119, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 122, 0, 1, 80, 125, 0, 1, 80, 1, 0, 64, 4, 128, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 131, 0, 1, 80, 134, 0, 1, 80, 1, 0, 64, 4, 137, 0, 1, 80, 1, 0, 64, 4, 142, 0, 1, 80, 1, 0, 64, 4, 145, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 148, 0, 1, 80, 151, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 154, 0, 1, 80, 159, 0, 1, 80, 1, 0, 64, 4, 162, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 165, 0, 1, 80, 168, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 171, 0, 1, 80, 174, 0, 1, 80, 1, 0, 80, 4, 177, 0, 1, 80, 1, 0, 64, 4, 180, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 183, 0, 1, 80, 186, 0, 1, 80, 1, 0, 64, 4, 189, 0, 1, 80, 1, 0, 80, 4, 192, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 195, 0, 1, 80, 198, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 201, 0, 1, 80, 204, 0, 1, 80, 1, 0, 64, 4, 207, 0, 1, 80, 1, 0, 64, 4, 212, 0, 1, 80, 1, 0, 80, 4, 215, 0, 1, 80, 1, 0, 72, 4, 218, 0, 1, 144, 1, 0, 64, 4, 45, 174, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 221, 0, 1, 80, 224, 0, 1, 80, 1, 0, 64, 4, 227, 0, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 230, 0, 1, 80, 233, 0, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 224, 45, 0, 96, 239, 0, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 236, 0, 1, 80, 59, 174, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 242, 0, 1, 80, 245, 0, 1, 80, 1, 0, 72, 4, 250, 0, 1, 144, 1, 0, 90, 4, 253, 0, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 71, 174, 0, 32, 73, 174, 0, 32, 1, 0, 64, 4, 0, 1, 1, 80, 1, 0, 64, 4, 3, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 6, 1, 1, 80, 9, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 12, 1, 1, 80, 15, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 18, 1, 1, 80, 21, 1, 1, 80, 1, 0, 64, 4, 24, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 27, 1, 1, 80, 30, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 35, 1, 1, 80, 38, 1, 1, 80, 1, 0, 80, 4, 41, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 44, 1, 1, 80, 47, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 50, 1, 1, 80, 53, 1, 1, 80, 1, 0, 80, 4, 56, 1, 1, 80, 1, 0, 64, 4, 59, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 64, 1, 1, 80, 67, 1, 1, 80, 1, 0, 64, 4, 70, 1, 1, 80, 1, 0, 64, 4, 73, 1, 1, 80, 1, 0, 64, 4, 76, 1, 1, 80, 1, 0, 64, 4, 79, 1, 1, 80, 1, 0, 64, 4, 84, 1, 1, 80, 1, 0, 64, 4, 89, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 92, 1, 1, 80, 95, 1, 1, 80, 1, 0, 64, 4, 98, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 103, 1, 1, 80, 106, 1, 1, 80, 1, 0, 64, 4, 109, 1, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 117, 1, 1, 144, 48, 1, 0, 96, 2, 0, 64, 4, 80, 4, 170, 170, 112, 1, 1, 80, 147, 174, 0, 32, 1, 0, 80, 4, 120, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 123, 1, 1, 80, 126, 1, 1, 80, 1, 0, 64, 4, 131, 1, 1, 80, 1, 0, 64, 4, 134, 1, 1, 80, 1, 0, 64, 4, 137, 1, 1, 80, 1, 0, 64, 4, 140, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 143, 1, 1, 80, 146, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 149, 1, 1, 80, 152, 1, 1, 80, 1, 0, 64, 4, 157, 1, 1, 80, 1, 0, 64, 4, 160, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 163, 1, 1, 80, 166, 1, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 169, 1, 1, 144, 172, 1, 1, 144, 1, 0, 90, 4, 176, 1, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 185, 174, 0, 32, 189, 174, 0, 32, 1, 0, 64, 4, 179, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 184, 1, 1, 80, 187, 1, 1, 80, 1, 0, 64, 4, 190, 1, 1, 80, 1, 0, 64, 4, 193, 1, 1, 80, 1, 0, 64, 4, 196, 1, 1, 80, 1, 0, 64, 4, 199, 1, 1, 80, 1, 0, 64, 4, 202, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 205, 1, 1, 80, 208, 1, 1, 80, 1, 0, 64, 4, 211, 1, 1, 80, 1, 0, 64, 4, 214, 1, 1, 80, 1, 0, 64, 4, 217, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 222, 1, 1, 80, 225, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 228, 1, 1, 80, 231, 1, 1, 80, 1, 0, 80, 4, 234, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 237, 1, 1, 80, 240, 1, 1, 80, 1, 0, 64, 4, 243, 1, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 248, 1, 1, 80, 251, 1, 1, 80, 1, 0, 64, 4, 254, 1, 1, 80, 1, 0, 64, 4, 1, 2, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 4, 2, 1, 144, 1, 0, 64, 4, 245, 174, 0, 32, 1, 0, 64, 4, 7, 2, 1, 80, 1, 0, 64, 4, 10, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 15, 2, 1, 80, 18, 2, 1, 80, 1, 0, 90, 4, 24, 2, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 21, 2, 1, 80, 3, 175, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 27, 2, 1, 80, 32, 2, 1, 80, 1, 0, 64, 4, 35, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 38, 2, 1, 80, 41, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 44, 2, 1, 80, 47, 2, 1, 80, 1, 0, 64, 4, 50, 2, 1, 80, 1, 0, 64, 4, 53, 2, 1, 80, 1, 0, 64, 4, 56, 2, 1, 80, 1, 0, 64, 4, 59, 2, 1, 80, 1, 0, 64, 4, 62, 2, 1, 80, 1, 0, 64, 4, 67, 2, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 70, 2, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 35, 175, 0, 32, 78, 2, 1, 80, 1, 0, 64, 4, 81, 2, 1, 80, 1, 0, 64, 4, 84, 2, 1, 80, 1, 0, 64, 4, 89, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 92, 2, 1, 80, 95, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 98, 2, 1, 80, 101, 2, 1, 80, 1, 0, 64, 4, 104, 2, 1, 80, 1, 0, 80, 4, 107, 2, 1, 80, 1, 0, 80, 4, 110, 2, 1, 80, 1, 0, 64, 4, 113, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 116, 2, 1, 80, 119, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 122, 2, 1, 80, 125, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 128, 2, 1, 80, 131, 2, 1, 80, 1, 0, 64, 4, 134, 2, 1, 80, 1, 0, 64, 4, 137, 2, 1, 80, 1, 0, 64, 4, 140, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 143, 2, 1, 80, 148, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 151, 2, 1, 80, 154, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 157, 2, 1, 80, 160, 2, 1, 80, 1, 0, 64, 4, 163, 2, 1, 80, 1, 0, 64, 4, 166, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 169, 2, 1, 80, 172, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 175, 2, 1, 80, 178, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 181, 2, 1, 80, 184, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 189, 2, 1, 80, 192, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 195, 2, 1, 80, 198, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 201, 2, 1, 80, 204, 2, 1, 80, 1, 0, 72, 4, 207, 2, 1, 144, 1, 0, 80, 4, 123, 175, 0, 32, 1, 0, 80, 4, 210, 2, 1, 80, 1, 0, 72, 4, 213, 2, 1, 144, 1, 0, 90, 4, 216, 2, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 129, 175, 0, 32, 131, 175, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 219, 2, 1, 80, 222, 2, 1, 80, 1, 0, 72, 4, 225, 2, 1, 144, 1, 0, 80, 4, 141, 175, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 228, 2, 1, 80, 231, 2, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 234, 2, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 149, 175, 0, 32, 237, 2, 1, 80, 1, 0, 64, 4, 240, 2, 1, 80, 1, 0, 64, 4, 245, 2, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 248, 2, 1, 80, 251, 2, 1, 80, 1, 0, 64, 4, 254, 2, 1, 80, 1, 0, 64, 4, 1, 3, 1, 80, 1, 0, 80, 4, 4, 3, 1, 80, 1, 0, 64, 4, 7, 3, 1, 80, 1, 0, 80, 4, 10, 3, 1, 80, 1, 0, 72, 4, 13, 3, 1, 144, 1, 0, 80, 4, 175, 175, 0, 32, 1, 0, 64, 4, 17, 3, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 20, 3, 1, 80, 25, 3, 1, 80, 1, 0, 64, 4, 28, 3, 1, 80, 1, 0, 72, 4, 31, 3, 1, 144, 1, 0, 64, 4, 187, 175, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 35, 3, 1, 80, 38, 3, 1, 80, 1, 0, 64, 4, 41, 3, 1, 80, 1, 0, 80, 4, 44, 3, 1, 80, 1, 0, 80, 4, 47, 3, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 50, 3, 1, 80, 55, 3, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 58, 3, 1, 144, 1, 0, 64, 4, 205, 175, 0, 32, 1, 0, 64, 4, 69, 3, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 72, 3, 1, 144, 1, 0, 64, 4, 213, 175, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 76, 3, 1, 80, 79, 3, 1, 80, 1, 0, 64, 4, 82, 3, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 85, 3, 1, 80, 88, 3, 1, 80, 1, 0, 64, 4, 91, 3, 1, 80, 1, 0, 64, 4, 94, 3, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 97, 3, 1, 80, 100, 3, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 103, 3, 1, 80, 106, 3, 1, 80, 1, 0, 64, 4, 111, 3, 1, 80, 1, 0, 64, 4, 114, 3, 1, 80, 1, 0, 72, 4, 117, 3, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 245, 175, 0, 32, 120, 3, 1, 80, 1, 0, 64, 4, 123, 3, 1, 80, 1, 0, 90, 4, 129, 3, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 126, 3, 1, 80, 253, 175, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 132, 3, 1, 80, 135, 3, 1, 80, 1, 0, 64, 4, 138, 3, 1, 80, 1, 0, 64, 4, 141, 3, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 144, 3, 1, 80, 147, 3, 1, 80, 1, 0, 64, 4, 150, 3, 1, 80, 1, 0, 64, 4, 153, 3, 1, 80, 1, 0, 64, 4, 158, 3, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 161, 3, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 21, 176, 0, 32, 164, 3, 1, 80, 1, 0, 64, 4, 169, 3, 1, 80, 1, 0, 64, 4, 172, 3, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 175, 3, 1, 80, 178, 3, 1, 80, 1, 0, 64, 4, 181, 3, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 192, 191, 0, 96, 184, 3, 1, 144, 1, 0, 90, 4, 190, 3, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 39, 176, 0, 32, 43, 176, 0, 32, 1, 0, 64, 4, 200, 3, 1, 80, 1, 0, 64, 4, 203, 3, 1, 80, 1, 0, 64, 4, 206, 3, 1, 80, 1, 0, 64, 4, 209, 3, 1, 80, 1, 0, 64, 4, 214, 3, 1, 80, 1, 0, 64, 4, 217, 3, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 220, 3, 1, 144, 46, 106, 0, 96, 1, 0, 80, 4, 61, 176, 0, 32, 1, 0, 64, 4, 226, 3, 1, 80, 1, 0, 64, 4, 229, 3, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 232, 3, 1, 80, 235, 3, 1, 80, 1, 0, 72, 4, 238, 3, 1, 144, 1, 0, 80, 4, 75, 176, 0, 32, 1, 0, 64, 4, 241, 3, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 244, 3, 1, 80, 247, 3, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 253, 3, 1, 144, 46, 106, 0, 96, 2, 0, 64, 4, 80, 4, 170, 170, 250, 3, 1, 80, 85, 176, 0, 32, 1, 0, 80, 4, 2, 4, 1, 80, 1, 0, 64, 4, 5, 4, 1, 80, 1, 0, 64, 4, 8, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 11, 4, 1, 80, 14, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 17, 4, 1, 80, 20, 4, 1, 80, 1, 0, 64, 4, 23, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 26, 4, 1, 80, 29, 4, 1, 80, 1, 0, 64, 4, 32, 4, 1, 80, 1, 0, 72, 4, 35, 4, 1, 144, 2, 0, 72, 4, 90, 4, 170, 170, 38, 4, 1, 144, 41, 4, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 115, 176, 0, 32, 117, 176, 0, 32, 1, 0, 64, 4, 44, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 47, 4, 1, 80, 50, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 53, 4, 1, 80, 56, 4, 1, 80, 1, 0, 64, 4, 61, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 64, 4, 1, 80, 67, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 70, 4, 1, 80, 73, 4, 1, 80, 1, 0, 64, 4, 76, 4, 1, 80, 1, 0, 80, 4, 79, 4, 1, 80, 1, 0, 64, 4, 82, 4, 1, 80, 1, 0, 64, 4, 85, 4, 1, 80, 1, 0, 72, 4, 90, 4, 1, 144, 1, 0, 80, 4, 153, 176, 0, 32, 1, 0, 64, 4, 93, 4, 1, 80, 1, 0, 72, 4, 96, 4, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 159, 176, 0, 32, 100, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 103, 4, 1, 80, 106, 4, 1, 80, 1, 0, 64, 4, 109, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 114, 4, 1, 80, 117, 4, 1, 80, 1, 0, 90, 4, 125, 4, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 120, 4, 1, 80, 175, 176, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 128, 4, 1, 80, 131, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 134, 4, 1, 80, 139, 4, 1, 80, 1, 0, 64, 4, 142, 4, 1, 80, 1, 0, 64, 4, 145, 4, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 148, 4, 1, 144, 1, 0, 64, 4, 193, 176, 0, 32, 1, 0, 64, 4, 151, 4, 1, 80, 1, 0, 64, 4, 154, 4, 1, 80, 1, 0, 64, 4, 157, 4, 1, 80, 1, 0, 80, 4, 160, 4, 1, 80, 1, 0, 64, 4, 163, 4, 1, 80, 1, 0, 64, 4, 166, 4, 1, 80, 1, 0, 80, 4, 171, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 174, 4, 1, 80, 177, 4, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 185, 4, 1, 144, 188, 4, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 180, 4, 1, 80, 217, 176, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 191, 4, 1, 80, 194, 4, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 197, 4, 1, 144, 2, 0, 72, 4, 90, 4, 170, 170, 235, 44, 0, 96, 200, 4, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 229, 176, 0, 32, 233, 176, 0, 32, 2, 0, 72, 4, 90, 4, 170, 170, 203, 4, 1, 144, 206, 4, 1, 144, 1, 0, 64, 4, 241, 176, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 211, 4, 1, 80, 214, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 217, 4, 1, 80, 220, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 223, 4, 1, 80, 226, 4, 1, 80, 1, 0, 64, 4, 229, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 232, 4, 1, 80, 237, 4, 1, 80, 1, 0, 64, 4, 240, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 243, 4, 1, 80, 248, 4, 1, 80, 1, 0, 64, 4, 251, 4, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 254, 4, 1, 80, 1, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 6, 5, 1, 80, 9, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 12, 5, 1, 80, 15, 5, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 20, 5, 1, 144, 1, 0, 64, 4, 29, 177, 0, 32, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 24, 5, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 35, 177, 0, 32, 28, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 31, 5, 1, 80, 34, 5, 1, 80, 1, 0, 64, 4, 39, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 42, 5, 1, 80, 45, 5, 1, 80, 1, 0, 64, 4, 48, 5, 1, 80, 1, 0, 64, 4, 51, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 54, 5, 1, 80, 57, 5, 1, 80, 1, 0, 80, 4, 62, 5, 1, 80, 1, 0, 64, 4, 65, 5, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 96, 192, 0, 96, 68, 5, 1, 144, 1, 0, 90, 4, 71, 5, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 65, 177, 0, 32, 69, 177, 0, 32, 1, 0, 64, 4, 74, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 77, 5, 1, 80, 80, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 83, 5, 1, 80, 86, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 89, 5, 1, 80, 92, 5, 1, 80, 1, 0, 64, 4, 95, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 98, 5, 1, 80, 101, 5, 1, 80, 1, 0, 64, 4, 104, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 107, 5, 1, 80, 110, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 113, 5, 1, 80, 116, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 119, 5, 1, 80, 122, 5, 1, 80, 1, 0, 64, 4, 125, 5, 1, 80, 1, 0, 64, 4, 128, 5, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 133, 5, 1, 144, 91, 0, 0, 96, 1, 0, 64, 4, 113, 177, 0, 32, 1, 0, 64, 4, 137, 5, 1, 80, 1, 0, 64, 4, 142, 5, 1, 80, 1, 0, 64, 4, 145, 5, 1, 80, 1, 0, 64, 4, 148, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 151, 5, 1, 80, 156, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 159, 5, 1, 80, 162, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 165, 5, 1, 80, 168, 5, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 171, 5, 1, 144, 32, 2, 0, 96, 1, 0, 90, 4, 175, 5, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 139, 177, 0, 32, 143, 177, 0, 32, 1, 0, 64, 4, 178, 5, 1, 80, 1, 0, 72, 4, 183, 5, 1, 144, 1, 0, 64, 4, 151, 177, 0, 32, 1, 0, 64, 4, 186, 5, 1, 80, 1, 0, 64, 4, 189, 5, 1, 80, 1, 0, 72, 4, 192, 5, 1, 144, 1, 0, 64, 4, 159, 177, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 196, 5, 1, 80, 199, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 202, 5, 1, 80, 207, 5, 1, 80, 1, 0, 80, 4, 210, 5, 1, 80, 1, 0, 64, 4, 213, 5, 1, 80, 1, 0, 64, 4, 218, 5, 1, 80, 1, 0, 64, 4, 221, 5, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 184, 191, 0, 96, 226, 5, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 179, 177, 0, 32, 229, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 234, 5, 1, 80, 237, 5, 1, 80, 1, 0, 64, 4, 240, 5, 1, 80, 1, 0, 90, 4, 246, 5, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 243, 5, 1, 80, 193, 177, 0, 32, 1, 0, 64, 4, 249, 5, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 252, 5, 1, 80, 255, 5, 1, 80, 1, 0, 64, 4, 2, 6, 1, 80, 1, 0, 64, 4, 5, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 8, 6, 1, 80, 11, 6, 1, 80, 1, 0, 64, 4, 14, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 17, 6, 1, 80, 20, 6, 1, 80, 1, 0, 72, 4, 23, 6, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 219, 177, 0, 32, 26, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 29, 6, 1, 80, 32, 6, 1, 80, 1, 0, 64, 4, 35, 6, 1, 80, 1, 0, 64, 4, 40, 6, 1, 80, 1, 0, 64, 4, 43, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 48, 6, 1, 80, 51, 6, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 54, 6, 1, 144, 68, 1, 0, 96, 1, 0, 64, 4, 239, 177, 0, 32, 1, 0, 64, 4, 57, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 60, 6, 1, 80, 63, 6, 1, 80, 1, 0, 72, 4, 68, 6, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 251, 177, 0, 32, 71, 6, 1, 80, 1, 0, 80, 4, 76, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 79, 6, 1, 80, 84, 6, 1, 80, 1, 0, 64, 4, 87, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 90, 6, 1, 80, 93, 6, 1, 80, 1, 0, 90, 4, 96, 6, 1, 144, 1, 0, 80, 4, 13, 178, 0, 32, 1, 0, 72, 4, 102, 6, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 17, 178, 0, 32, 105, 6, 1, 80, 1, 0, 72, 4, 108, 6, 1, 144, 1, 0, 64, 4, 23, 178, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 111, 6, 1, 80, 116, 6, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 119, 6, 1, 144, 94, 0, 0, 96, 2, 0, 64, 4, 80, 4, 170, 170, 31, 178, 0, 32, 123, 6, 1, 80, 1, 0, 72, 4, 126, 6, 1, 144, 1, 0, 64, 4, 39, 178, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 129, 6, 1, 80, 132, 6, 1, 80, 1, 0, 64, 4, 135, 6, 1, 80, 1, 0, 64, 4, 138, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 141, 6, 1, 80, 146, 6, 1, 80, 1, 0, 64, 4, 151, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 154, 6, 1, 80, 157, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 160, 6, 1, 80, 163, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 166, 6, 1, 80, 169, 6, 1, 80, 1, 0, 80, 4, 172, 6, 1, 80, 1, 0, 80, 4, 175, 6, 1, 80, 1, 0, 64, 4, 178, 6, 1, 80, 1, 0, 64, 4, 181, 6, 1, 80, 1, 0, 64, 4, 184, 6, 1, 80, 1, 0, 64, 4, 187, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 190, 6, 1, 80, 193, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 196, 6, 1, 80, 199, 6, 1, 80, 1, 0, 64, 4, 202, 6, 1, 80, 1, 0, 64, 4, 205, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 208, 6, 1, 80, 211, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 214, 6, 1, 80, 217, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 220, 6, 1, 80, 223, 6, 1, 80, 1, 0, 64, 4, 226, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 231, 6, 1, 80, 236, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 239, 6, 1, 80, 242, 6, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 245, 6, 1, 80, 248, 6, 1, 80, 1, 0, 64, 4, 251, 6, 1, 80, 1, 0, 64, 4, 254, 6, 1, 80, 1, 0, 80, 4, 1, 7, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 4, 7, 1, 80, 7, 7, 1, 80, 1, 0, 80, 4, 10, 7, 1, 80, 1, 0, 80, 4, 13, 7, 1, 80, 1, 0, 64, 4, 16, 7, 1, 80, 1, 0, 80, 4, 19, 7, 1, 80, 1, 0, 80, 4, 22, 7, 1, 80, 1, 0, 80, 4, 25, 7, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 28, 7, 1, 80, 31, 7, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 34, 7, 1, 80, 37, 7, 1, 80, 1, 0, 80, 4, 40, 7, 1, 80, 1, 0, 80, 4, 43, 7, 1, 80, 1, 0, 64, 4, 46, 7, 1, 80, 1, 0, 72, 4, 49, 7, 1, 144, 1, 0, 80, 4, 155, 178, 0, 32, 1, 0, 64, 4, 52, 7, 1, 80, 1, 0, 64, 4, 55, 7, 1, 80, 1, 0, 64, 4, 58, 7, 1, 80, 1, 0, 64, 4, 61, 7, 1, 80, 1, 0, 64, 4, 64, 7, 1, 80, 1, 0, 64, 4, 67, 7, 1, 80, 1, 0, 64, 4, 72, 7, 1, 80, 1, 0, 64, 4, 75, 7, 1, 80, 1, 0, 64, 4, 78, 7, 1, 80, 1, 0, 72, 4, 81, 7, 1, 144, 1, 0, 90, 4, 84, 7, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 177, 178, 0, 32, 179, 178, 0, 32, 2, 0, 64, 4, 80, 4, 170, 170, 87, 7, 1, 80, 90, 7, 1, 80, 1, 0, 64, 4, 95, 7, 1, 80, 1, 0, 64, 4, 98, 7, 1, 80, 1, 0, 64, 4, 101, 7, 1, 80, 1, 0, 64, 4, 106, 7, 1, 80, 1, 0, 80, 4, 109, 7, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 112, 7, 1, 80, 115, 7, 1, 80, 2, 0, 72, 4, 90, 4, 170, 170, 118, 7, 1, 144, 121, 7, 1, 144, 2, 0, 72, 4, 90, 4, 170, 170, 127, 7, 1, 144, 130, 7, 1, 144, 2, 0, 64, 4, 80, 4, 170, 170, 203, 178, 0, 32, 207, 178, 0, 32, 1, 0, 64, 4, 136, 7, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 139, 7, 1, 80, 142, 7, 1, 80, 1, 0, 64, 4, 145, 7, 1, 80, 2, 0, 64, 4, 80, 4, 170, 170, 148, 7, 1, 80, 151, 7, 1, 80, 9, 3, 46, 8, 125, 12, 102, 4, 106, 4, 110, 4, 114, 4, 118, 4, 122, 4, 108, 13, 126, 4, 175, 11, 130, 4, 134, 4, 138, 4, 178, 9, 142, 4, 146, 4, 150, 4, 154, 4, 158, 4, 183, 11, 162, 4, 242, 17, 166, 4, 170, 4, 174, 4, 112, 13, 178, 4, 182, 4, 186, 4, 190, 4, 194, 4, 198, 4, 202, 4, 68, 4, 206, 4, 116, 13, 210, 4, 214, 4, 218, 4, 215, 13, 222, 4, 226, 4, 148, 18, 199, 11, 243, 7, 230, 4, 234, 4, 238, 4, 242, 4, 246, 4, 250, 4, 254, 4, 2, 5, 6, 5, 36, 9, 10, 5, 14, 5, 18, 5, 22, 5, 26, 5, 30, 5, 34, 5, 38, 5, 42, 5, 45, 5, 49, 5, 53, 5, 57, 5, 61, 5, 65, 5, 69, 5, 73, 5, 76, 5, 80, 5, 84, 5, 88, 5, 92, 5, 96, 5, 100, 5, 194, 0, 104, 5, 44, 7, 108, 5, 112, 5, 115, 5, 118, 5, 122, 5, 126, 5, 128, 13, 130, 5, 134, 5, 138, 5, 142, 5, 145, 5, 149, 5, 153, 5, 157, 5, 161, 5, 165, 5, 169, 5, 173, 5, 177, 5, 181, 5, 185, 5, 189, 5, 193, 5, 74, 8, 197, 5, 201, 5, 205, 5, 209, 5, 213, 5, 19, 26, 217, 5, 221, 5, 108, 10, 225, 5, 229, 5, 233, 5, 237, 5, 241, 5, 245, 5, 249, 5, 253, 5, 1, 6, 5, 6, 9, 6, 13, 6, 17, 6, 211, 11, 21, 6, 25, 6, 152, 13, 29, 6, 33, 6, 37, 6, 41, 6, 45, 6, 49, 6, 215, 11, 53, 6, 57, 6, 61, 6, 65, 6, 68, 6, 137, 12, 9, 17, 71, 6, 75, 6, 79, 6, 83, 6, 30, 13, 87, 6, 91, 6, 95, 6, 99, 6, 103, 6, 107, 6, 111, 6, 115, 6, 119, 6, 123, 6, 127, 6, 131, 6, 135, 6, 139, 6, 143, 6, 146, 6, 150, 6, 154, 6, 157, 6, 227, 11, 161, 6, 165, 6, 169, 6, 173, 6, 177, 6, 52, 25, 251, 13, 181, 6, 156, 12, 247, 11, 185, 6, 164, 12, 189, 6, 45, 19, 193, 6, 197, 6, 201, 6, 222, 9, 205, 6, 209, 6, 212, 6, 162, 6, 216, 6, 220, 6, 224, 6, 228, 6, 37, 21, 231, 6, 102, 8, 235, 6, 239, 6, 243, 6, 247, 6, 251, 6, 255, 6, 3, 7, 7, 7, 11, 7, 15, 7, 19, 7, 106, 8, 38, 13, 23, 7, 27, 7, 31, 7, 35, 7, 39, 7, 43, 7, 47, 7, 51, 7, 55, 7, 58, 7, 62, 7, 66, 7, 70, 7, 124, 10, 74, 7, 78, 7, 82, 7, 130, 20, 86, 7, 90, 7, 94, 7, 98, 7, 102, 7, 106, 7, 110, 7, 114, 7, 118, 7, 122, 7, 126, 7, 129, 7, 133, 7, 137, 7, 141, 7, 145, 7, 148, 7, 152, 7, 121, 8, 156, 7, 160, 7, 164, 7, 168, 7, 128, 10, 255, 11, 172, 7, 176, 7, 180, 7, 184, 7, 188, 7, 192, 7, 196, 7, 200, 7, 204, 7, 208, 7, 238, 5, 212, 7, 216, 7, 220, 7, 179, 12, 224, 7, 228, 7, 252, 19, 232, 7, 235, 7, 250, 5, 239, 7, 99, 11, 242, 7, 246, 7, 7, 12, 250, 7, 186, 12, 135, 4, 77, 5, 254, 7, 2, 8, 6, 8, 68, 9, 48, 4, 144, 10, 10, 8, 2, 6, 14, 8, 18, 8, 22, 8, 26, 8, 30, 8, 34, 8, 38, 8, 42, 8, 174, 6, 45, 8, 49, 8, 53, 8, 57, 8, 61, 8, 65, 8, 69, 8, 73, 8, 77, 8, 81, 8, 85, 8, 89, 8, 93, 8, 97, 8, 101, 8, 105, 8, 109, 8, 112, 8, 116, 8, 120, 8, 124, 8, 128, 8, 132, 8, 136, 8, 140, 8, 80, 9, 144, 8, 157, 8, 148, 8, 152, 8, 143, 4, 156, 8, 194, 12, 160, 8, 113, 20, 164, 8, 168, 8, 171, 8, 29, 11, 175, 8, 179, 8, 183, 8, 187, 8, 191, 8, 198, 12, 195, 8, 199, 8, 203, 8, 207, 8, 211, 8, 215, 8, 218, 8, 222, 8, 226, 8, 101, 5, 230, 8, 234, 8, 238, 8, 241, 8, 245, 8, 249, 8, 252, 8, 0, 9, 178, 6, 4, 9, 195, 0, 8, 9, 12, 9, 16, 9, 20, 9, 24, 9, 28, 9, 32, 9, 35, 9, 39, 9, 43, 9, 47, 9, 51, 9, 55, 9, 43, 17, 42, 13, 59, 9, 63, 9, 67, 9, 71, 9, 75, 9, 79, 9, 83, 9, 87, 9, 91, 9, 161, 8, 3, 8, 95, 9, 99, 9, 103, 9, 170, 23, 107, 9, 111, 9, 114, 9, 118, 9, 122, 9, 126, 9, 130, 9, 134, 9, 138, 9, 142, 9, 146, 9, 149, 9, 153, 9, 157, 9, 161, 9, 165, 9, 169, 9, 173, 9, 177, 9, 181, 9, 185, 9, 189, 9, 193, 9, 197, 9, 201, 9, 205, 9, 209, 9, 213, 9, 217, 9, 221, 9, 225, 9, 229, 9, 152, 10, 233, 9, 237, 9, 241, 9, 245, 9, 214, 12, 3, 14, 154, 0, 249, 9, 253, 9, 27, 8, 0, 10, 171, 13, 4, 10, 8, 10, 12, 10, 16, 10, 20, 10, 24, 10, 28, 10, 175, 13, 32, 10, 36, 10, 40, 10, 39, 12, 123, 11, 44, 10, 48, 10, 52, 10, 56, 10, 60, 10, 64, 10, 68, 10, 72, 10, 76, 10, 80, 10, 83, 10, 87, 10, 91, 10, 95, 10, 179, 13, 99, 10, 103, 10, 153, 18, 206, 18, 107, 10, 50, 13, 111, 10, 115, 10, 197, 7, 119, 10, 131, 11, 123, 10, 127, 10, 131, 10, 135, 10, 139, 10, 143, 10, 147, 10, 222, 18, 151, 10, 148, 19, 155, 10, 159, 10, 163, 10, 167, 10, 171, 10, 175, 10, 179, 10, 183, 13, 183, 10, 187, 10, 191, 10, 195, 10, 198, 10, 202, 10, 205, 10, 209, 10, 213, 10, 217, 10, 9, 3, 217, 7, 143, 11, 160, 18, 29, 10, 221, 10, 225, 10, 229, 10, 233, 10, 236, 10, 240, 10, 244, 10, 248, 10, 252, 10, 0, 11, 4, 11, 8, 11, 12, 11, 16, 11, 20, 11, 24, 11, 226, 12, 28, 11, 32, 11, 35, 11, 39, 11, 43, 11, 47, 11, 51, 11, 55, 11, 26, 6, 31, 4, 59, 11, 63, 11, 66, 11, 70, 11, 74, 11, 78, 11, 82, 11, 86, 11, 90, 11, 94, 11, 98, 11, 102, 11, 106, 11, 110, 11, 240, 6, 114, 11, 118, 11, 122, 11, 126, 11, 62, 12, 130, 11, 100, 13, 134, 11, 138, 11, 142, 11, 146, 11, 230, 12, 150, 11, 154, 11, 158, 11, 203, 4, 162, 11, 166, 11, 195, 13, 170, 11, 174, 11, 178, 11, 182, 11, 186, 11, 251, 17, 190, 11, 194, 11, 198, 11, 202, 11, 226, 10, 206, 11, 15, 18, 210, 11, 214, 11, 238, 12, 218, 11, 222, 11, 251, 3, 226, 11, 230, 11, 234, 11, 238, 11, 242, 11, 246, 11, 242, 12, 30, 18, 250, 11, 254, 11, 2, 12, 246, 12, 6, 12, 184, 10, 10, 12, 24, 19, 14, 12, 18, 12, 29, 19, 22, 12, 26, 12, 30, 12, 34, 12, 38, 12, 229, 7, 42, 12, 46, 12, 188, 10, 50, 12, 54, 12, 199, 13, 57, 12, 61, 12, 65, 12, 69, 12, 150, 5, 73, 12, 51, 0, 77, 12, 186, 6, 81, 12, 85, 12, 89, 12, 123, 7, 92, 12, 96, 12, 100, 12, 104, 12, 108, 12, 112, 12, 116, 12, 120, 12, 154, 14, 124, 12, 128, 12, 132, 12, 136, 12, 140, 12, 144, 12, 148, 12, 152, 12, 68, 0, 155, 12, 159, 12, 163, 12, 108, 9, 167, 12, 170, 12, 174, 12, 178, 12, 182, 12, 185, 12, 189, 12, 230, 10, 193, 12, 197, 12, 201, 12, 205, 12, 209, 12, 213, 12, 210, 22, 217, 12, 154, 5, 221, 12, 225, 12, 45, 10, 229, 12, 233, 12, 2, 13, 237, 12, 241, 12, 245, 12, 249, 12, 6, 13, 253, 12, 1, 13, 5, 13, 9, 13, 13, 13, 17, 13, 220, 24, 21, 13, 25, 13, 29, 13, 33, 13, 155, 11, 37, 13, 82, 12, 41, 13, 45, 13, 49, 13, 53, 13, 57, 13, 138, 7, 143, 9, 61, 13, 65, 13, 68, 13, 72, 13, 76, 13, 79, 13, 83, 13, 87, 13, 91, 13, 118, 20, 95, 13, 99, 13, 103, 13, 227, 4, 107, 13, 48, 17, 111, 13, 115, 13, 119, 13, 123, 13, 127, 13, 201, 18, 131, 13, 198, 6, 135, 13, 139, 13, 143, 13, 147, 13, 151, 13, 155, 13, 158, 13, 162, 13, 166, 13, 170, 13, 174, 13, 178, 13, 182, 13, 186, 13, 190, 13, 194, 13, 198, 13, 202, 13, 206, 13, 210, 13, 214, 13, 218, 13, 109, 12, 214, 10, 222, 13, 245, 21, 226, 13, 230, 13, 14, 14, 234, 13, 238, 13, 242, 13, 246, 13, 250, 13, 254, 13, 154, 9, 2, 14, 6, 14, 9, 14, 13, 14, 247, 168, 0, 32, 251, 168, 0, 32, 255, 168, 0, 32, 1, 169, 0, 32, 5, 169, 0, 32, 9, 169, 0, 32, 13, 169, 0, 32, 15, 169, 0, 32, 19, 169, 0, 32, 21, 169, 0, 32, 27, 169, 0, 32, 31, 169, 0, 32, 33, 169, 0, 32, 39, 169, 0, 32, 41, 169, 0, 32, 45, 169, 0, 32, 47, 169, 0, 32, 49, 169, 0, 32, 51, 169, 0, 32, 55, 169, 0, 32, 57, 169, 0, 32, 59, 169, 0, 32, 61, 169, 0, 32, 63, 169, 0, 32, 65, 169, 0, 32, 67, 169, 0, 32, 77, 169, 0, 32, 83, 169, 0, 32, 85, 169, 0, 32, 87, 169, 0, 32, 89, 169, 0, 32, 91, 169, 0, 32, 95, 169, 0, 32, 97, 169, 0, 32, 99, 169, 0, 32, 103, 169, 0, 32, 107, 169, 0, 32, 111, 169, 0, 32, 113, 169, 0, 32, 117, 169, 0, 32, 119, 169, 0, 32, 123, 169, 0, 32, 125, 169, 0, 32, 129, 169, 0, 32, 137, 169, 0, 32, 141, 169, 0, 32, 149, 169, 0, 32, 153, 169, 0, 32, 157, 169, 0, 32, 161, 169, 0, 32, 165, 169, 0, 32, 167, 169, 0, 32, 169, 169, 0, 32, 173, 169, 0, 32, 175, 169, 0, 32, 179, 169, 0, 32, 181, 169, 0, 32, 185, 169, 0, 32, 189, 169, 0, 32, 193, 169, 0, 32, 195, 169, 0, 32, 197, 169, 0, 32, 199, 169, 0, 32, 203, 169, 0, 32, 205, 169, 0, 32, 207, 169, 0, 32, 211, 169, 0, 32, 215, 169, 0, 32, 217, 169, 0, 32, 221, 169, 0, 32, 225, 169, 0, 32, 229, 169, 0, 32, 233, 169, 0, 32, 235, 169, 0, 32, 239, 169, 0, 32, 243, 169, 0, 32, 247, 169, 0, 32, 251, 169, 0, 32, 253, 169, 0, 32, 255, 169, 0, 32, 3, 170, 0, 32, 5, 170, 0, 32, 9, 170, 0, 32, 13, 170, 0, 32, 15, 170, 0, 32, 19, 170, 0, 32, 23, 170, 0, 32, 25, 170, 0, 32, 29, 170, 0, 32, 31, 170, 0, 32, 33, 170, 0, 32, 35, 170, 0, 32, 39, 170, 0, 32, 47, 170, 0, 32, 49, 170, 0, 32, 51, 170, 0, 32, 53, 170, 0, 32, 55, 170, 0, 32, 57, 170, 0, 32, 59, 170, 0, 32, 63, 170, 0, 32, 67, 170, 0, 32, 71, 170, 0, 32, 73, 170, 0, 32, 75, 170, 0, 32, 77, 170, 0, 32, 79, 170, 0, 32, 81, 170, 0, 32, 85, 170, 0, 32, 87, 170, 0, 32, 89, 170, 0, 32, 91, 170, 0, 32, 95, 170, 0, 32, 97, 170, 0, 32, 101, 170, 0, 32, 105, 170, 0, 32, 109, 170, 0, 32, 111, 170, 0, 32, 115, 170, 0, 32, 117, 170, 0, 32, 119, 170, 0, 32, 121, 170, 0, 32, 125, 170, 0, 32, 127, 170, 0, 32, 129, 170, 0, 32, 131, 170, 0, 32, 135, 170, 0, 32, 139, 170, 0, 32, 143, 170, 0, 32, 147, 170, 0, 32, 149, 170, 0, 32, 153, 170, 0, 32, 155, 170, 0, 32, 159, 170, 0, 32, 161, 170, 0, 32, 163, 170, 0, 32, 167, 170, 0, 32, 169, 170, 0, 32, 173, 170, 0, 32, 175, 170, 0, 32, 179, 170, 0, 32, 183, 170, 0, 32, 185, 170, 0, 32, 187, 170, 0, 32, 189, 170, 0, 32, 191, 170, 0, 32, 195, 170, 0, 32, 199, 170, 0, 32, 203, 170, 0, 32, 205, 170, 0, 32, 207, 170, 0, 32, 209, 170, 0, 32, 221, 170, 0, 32, 225, 170, 0, 32, 227, 170, 0, 32, 231, 170, 0, 32, 233, 170, 0, 32, 235, 170, 0, 32, 239, 170, 0, 32, 241, 170, 0, 32, 245, 170, 0, 32, 249, 170, 0, 32, 251, 170, 0, 32, 253, 170, 0, 32, 255, 170, 0, 32, 1, 171, 0, 32, 3, 171, 0, 32, 5, 171, 0, 32, 7, 171, 0, 32, 11, 171, 0, 32, 15, 171, 0, 32, 17, 171, 0, 32, 21, 171, 0, 32, 25, 171, 0, 32, 29, 171, 0, 32, 31, 171, 0, 32, 33, 171, 0, 32, 35, 171, 0, 32, 41, 171, 0, 32, 51, 171, 0, 32, 55, 171, 0, 32, 57, 171, 0, 32, 65, 171, 0, 32, 69, 171, 0, 32, 71, 171, 0, 32, 75, 171, 0, 32, 77, 171, 0, 32, 81, 171, 0, 32, 83, 171, 0, 32, 85, 171, 0, 32, 91, 171, 0, 32, 95, 171, 0, 32, 103, 171, 0, 32, 107, 171, 0, 32, 113, 171, 0, 32, 117, 171, 0, 32, 119, 171, 0, 32, 123, 171, 0, 32, 125, 171, 0, 32, 127, 171, 0, 32, 129, 171, 0, 32, 141, 171, 0, 32, 145, 171, 0, 32, 147, 171, 0, 32, 149, 171, 0, 32, 151, 171, 0, 32, 153, 171, 0, 32, 155, 171, 0, 32, 157, 171, 0, 32, 161, 171, 0, 32, 165, 171, 0, 32, 167, 171, 0, 32, 171, 171, 0, 32, 175, 171, 0, 32, 179, 171, 0, 32, 183, 171, 0, 32, 187, 171, 0, 32, 191, 171, 0, 32, 195, 171, 0, 32, 197, 171, 0, 32, 199, 171, 0, 32, 203, 171, 0, 32, 205, 171, 0, 32, 209, 171, 0, 32, 213, 171, 0, 32, 215, 171, 0, 32, 217, 171, 0, 32, 219, 171, 0, 32, 221, 171, 0, 32, 225, 171, 0, 32, 227, 171, 0, 32, 231, 171, 0, 32, 233, 171, 0, 32, 235, 171, 0, 32, 237, 171, 0, 32, 243, 171, 0, 32, 247, 171, 0, 32, 251, 171, 0, 32, 253, 171, 0, 32, 255, 171, 0, 32, 3, 172, 0, 32, 5, 172, 0, 32, 11, 172, 0, 32, 15, 172, 0, 32, 19, 172, 0, 32, 21, 172, 0, 32, 23, 172, 0, 32, 25, 172, 0, 32, 29, 172, 0, 32, 31, 172, 0, 32, 33, 172, 0, 32, 39, 172, 0, 32, 43, 172, 0, 32, 45, 172, 0, 32, 49, 172, 0, 32, 53, 172, 0, 32, 57, 172, 0, 32, 63, 172, 0, 32, 71, 172, 0, 32, 73, 172, 0, 32, 77, 172, 0, 32, 79, 172, 0, 32, 81, 172, 0, 32, 85, 172, 0, 32, 87, 172, 0, 32, 91, 172, 0, 32, 93, 172, 0, 32, 97, 172, 0, 32, 101, 172, 0, 32, 103, 172, 0, 32, 107, 172, 0, 32, 111, 172, 0, 32, 119, 172, 0, 32, 123, 172, 0, 32, 125, 172, 0, 32, 129, 172, 0, 32, 133, 172, 0, 32, 137, 172, 0, 32, 139, 172, 0, 32, 143, 172, 0, 32, 145, 172, 0, 32, 147, 172, 0, 32, 149, 172, 0, 32, 153, 172, 0, 32, 157, 172, 0, 32, 159, 172, 0, 32, 163, 172, 0, 32, 167, 172, 0, 32, 169, 172, 0, 32, 173, 172, 0, 32, 181, 172, 0, 32, 185, 172, 0, 32, 193, 172, 0, 32, 199, 172, 0, 32, 203, 172, 0, 32, 205, 172, 0, 32, 207, 172, 0, 32, 211, 172, 0, 32, 213, 172, 0, 32, 215, 172, 0, 32, 217, 172, 0, 32, 219, 172, 0, 32, 221, 172, 0, 32, 223, 172, 0, 32, 227, 172, 0, 32, 229, 172, 0, 32, 231, 172, 0, 32, 235, 172, 0, 32, 237, 172, 0, 32, 239, 172, 0, 32, 243, 172, 0, 32, 245, 172, 0, 32, 249, 172, 0, 32, 251, 172, 0, 32, 253, 172, 0, 32, 255, 172, 0, 32, 3, 173, 0, 32, 5, 173, 0, 32, 9, 173, 0, 32, 11, 173, 0, 32, 13, 173, 0, 32, 17, 173, 0, 32, 21, 173, 0, 32, 23, 173, 0, 32, 25, 173, 0, 32, 29, 173, 0, 32, 31, 173, 0, 32, 35, 173, 0, 32, 37, 173, 0, 32, 39, 173, 0, 32, 41, 173, 0, 32, 45, 173, 0, 32, 47, 173, 0, 32, 51, 173, 0, 32, 55, 173, 0, 32, 61, 173, 0, 32, 65, 173, 0, 32, 67, 173, 0, 32, 71, 173, 0, 32, 75, 173, 0, 32, 77, 173, 0, 32, 81, 173, 0, 32, 85, 173, 0, 32, 93, 173, 0, 32, 97, 173, 0, 32, 101, 173, 0, 32, 105, 173, 0, 32, 107, 173, 0, 32, 109, 173, 0, 32, 111, 173, 0, 32, 113, 173, 0, 32, 117, 173, 0, 32, 121, 173, 0, 32, 123, 173, 0, 32, 125, 173, 0, 32, 131, 173, 0, 32, 135, 173, 0, 32, 137, 173, 0, 32, 139, 173, 0, 32, 145, 173, 0, 32, 149, 173, 0, 32, 153, 173, 0, 32, 155, 173, 0, 32, 159, 173, 0, 32, 161, 173, 0, 32, 163, 173, 0, 32, 165, 173, 0, 32, 169, 173, 0, 32, 175, 173, 0, 32, 177, 173, 0, 32, 179, 173, 0, 32, 181, 173, 0, 32, 183, 173, 0, 32, 185, 173, 0, 32, 187, 173, 0, 32, 191, 173, 0, 32, 195, 173, 0, 32, 197, 173, 0, 32, 199, 173, 0, 32, 203, 173, 0, 32, 205, 173, 0, 32, 207, 173, 0, 32, 209, 173, 0, 32, 215, 173, 0, 32, 217, 173, 0, 32, 221, 173, 0, 32, 225, 173, 0, 32, 227, 173, 0, 32, 229, 173, 0, 32, 231, 173, 0, 32, 233, 173, 0, 32, 237, 173, 0, 32, 239, 173, 0, 32, 241, 173, 0, 32, 245, 173, 0, 32, 247, 173, 0, 32, 251, 173, 0, 32, 253, 173, 0, 32, 255, 173, 0, 32, 1, 174, 0, 32, 5, 174, 0, 32, 9, 174, 0, 32, 11, 174, 0, 32, 15, 174, 0, 32, 19, 174, 0, 32, 21, 174, 0, 32, 23, 174, 0, 32, 27, 174, 0, 32, 29, 174, 0, 32, 31, 174, 0, 32, 35, 174, 0, 32, 39, 174, 0, 32, 41, 174, 0, 32, 43, 174, 0, 32, 47, 174, 0, 32, 49, 174, 0, 32, 53, 174, 0, 32, 55, 174, 0, 32, 63, 174, 0, 32, 67, 174, 0, 32, 75, 174, 0, 32, 79, 174, 0, 32, 81, 174, 0, 32, 83, 174, 0, 32, 87, 174, 0, 32, 91, 174, 0, 32, 95, 174, 0, 32, 97, 174, 0, 32, 101, 174, 0, 32, 105, 174, 0, 32, 107, 174, 0, 32, 111, 174, 0, 32, 115, 174, 0, 32, 117, 174, 0, 32, 119, 174, 0, 32, 123, 174, 0, 32, 125, 174, 0, 32, 127, 174, 0, 32, 129, 174, 0, 32, 131, 174, 0, 32, 133, 174, 0, 32, 135, 174, 0, 32, 139, 174, 0, 32, 141, 174, 0, 32, 145, 174, 0, 32, 151, 174, 0, 32, 155, 174, 0, 32, 157, 174, 0, 32, 161, 174, 0, 32, 163, 174, 0, 32, 165, 174, 0, 32, 167, 174, 0, 32, 169, 174, 0, 32, 173, 174, 0, 32, 177, 174, 0, 32, 179, 174, 0, 32, 181, 174, 0, 32, 191, 174, 0, 32, 195, 174, 0, 32, 197, 174, 0, 32, 201, 174, 0, 32, 203, 174, 0, 32, 205, 174, 0, 32, 207, 174, 0, 32, 209, 174, 0, 32, 211, 174, 0, 32, 215, 174, 0, 32, 217, 174, 0, 32, 219, 174, 0, 32, 221, 174, 0, 32, 225, 174, 0, 32, 229, 174, 0, 32, 231, 174, 0, 32, 235, 174, 0, 32, 237, 174, 0, 32, 241, 174, 0, 32, 243, 174, 0, 32, 249, 174, 0, 32, 251, 174, 0, 32, 253, 174, 0, 32, 255, 174, 0, 32, 5, 175, 0, 32, 9, 175, 0, 32, 13, 175, 0, 32, 15, 175, 0, 32, 19, 175, 0, 32, 23, 175, 0, 32, 25, 175, 0, 32, 27, 175, 0, 32, 29, 175, 0, 32, 31, 175, 0, 32, 33, 175, 0, 32, 39, 175, 0, 32, 43, 175, 0, 32, 45, 175, 0, 32, 47, 175, 0, 32, 49, 175, 0, 32, 53, 175, 0, 32, 57, 175, 0, 32, 59, 175, 0, 32, 61, 175, 0, 32, 63, 175, 0, 32, 65, 175, 0, 32, 69, 175, 0, 32, 73, 175, 0, 32, 77, 175, 0, 32, 79, 175, 0, 32, 81, 175, 0, 32, 83, 175, 0, 32, 87, 175, 0, 32, 91, 175, 0, 32, 95, 175, 0, 32, 97, 175, 0, 32, 99, 175, 0, 32, 103, 175, 0, 32, 107, 175, 0, 32, 111, 175, 0, 32, 115, 175, 0, 32, 119, 175, 0, 32, 125, 175, 0, 32, 127, 175, 0, 32, 133, 175, 0, 32, 137, 175, 0, 32, 143, 175, 0, 32, 145, 175, 0, 32, 153, 175, 0, 32, 157, 175, 0, 32, 159, 175, 0, 32, 161, 175, 0, 32, 165, 175, 0, 32, 167, 175, 0, 32, 169, 175, 0, 32, 171, 175, 0, 32, 173, 175, 0, 32, 177, 175, 0, 32, 179, 175, 0, 32, 181, 175, 0, 32, 185, 175, 0, 32, 189, 175, 0, 32, 191, 175, 0, 32, 195, 175, 0, 32, 197, 175, 0, 32, 199, 175, 0, 32, 201, 175, 0, 32, 209, 175, 0, 32, 211, 175, 0, 32, 217, 175, 0, 32, 219, 175, 0, 32, 223, 175, 0, 32, 225, 175, 0, 32, 229, 175, 0, 32, 231, 175, 0, 32, 233, 175, 0, 32, 237, 175, 0, 32, 241, 175, 0, 32, 243, 175, 0, 32, 247, 175, 0, 32, 251, 175, 0, 32, 255, 175, 0, 32, 3, 176, 0, 32, 7, 176, 0, 32, 9, 176, 0, 32, 11, 176, 0, 32, 15, 176, 0, 32, 17, 176, 0, 32, 19, 176, 0, 32, 25, 176, 0, 32, 29, 176, 0, 32, 31, 176, 0, 32, 33, 176, 0, 32, 37, 176, 0, 32, 45, 176, 0, 32, 49, 176, 0, 32, 51, 176, 0, 32, 53, 176, 0, 32, 55, 176, 0, 32, 57, 176, 0, 32, 59, 176, 0, 32, 65, 176, 0, 32, 67, 176, 0, 32, 69, 176, 0, 32, 71, 176, 0, 32, 77, 176, 0, 32, 79, 176, 0, 32, 81, 176, 0, 32, 89, 176, 0, 32, 93, 176, 0, 32, 95, 176, 0, 32, 97, 176, 0, 32, 99, 176, 0, 32, 103, 176, 0, 32, 107, 176, 0, 32, 109, 176, 0, 32, 113, 176, 0, 32, 121, 176, 0, 32, 125, 176, 0, 32, 127, 176, 0, 32, 131, 176, 0, 32, 135, 176, 0, 32, 137, 176, 0, 32, 141, 176, 0, 32, 145, 176, 0, 32, 147, 176, 0, 32, 149, 176, 0, 32, 151, 176, 0, 32, 155, 176, 0, 32, 157, 176, 0, 32, 161, 176, 0, 32, 165, 176, 0, 32, 169, 176, 0, 32, 171, 176, 0, 32, 177, 176, 0, 32, 181, 176, 0, 32, 185, 176, 0, 32, 189, 176, 0, 32, 191, 176, 0, 32, 197, 176, 0, 32, 199, 176, 0, 32, 201, 176, 0, 32, 203, 176, 0, 32, 205, 176, 0, 32, 207, 176, 0, 32, 209, 176, 0, 32, 211, 176, 0, 32, 213, 176, 0, 32, 221, 176, 0, 32, 225, 176, 0, 32, 237, 176, 0, 32, 245, 176, 0, 32, 247, 176, 0, 32, 251, 176, 0, 32, 255, 176, 0, 32, 3, 177, 0, 32, 5, 177, 0, 32, 9, 177, 0, 32, 11, 177, 0, 32, 15, 177, 0, 32, 17, 177, 0, 32, 21, 177, 0, 32, 25, 177, 0, 32, 33, 177, 0, 32, 39, 177, 0, 32, 43, 177, 0, 32, 47, 177, 0, 32, 49, 177, 0, 32, 53, 177, 0, 32, 55, 177, 0, 32, 57, 177, 0, 32, 61, 177, 0, 32, 63, 177, 0, 32, 71, 177, 0, 32, 75, 177, 0, 32, 77, 177, 0, 32, 81, 177, 0, 32, 85, 177, 0, 32, 89, 177, 0, 32, 91, 177, 0, 32, 95, 177, 0, 32, 97, 177, 0, 32, 101, 177, 0, 32, 105, 177, 0, 32, 109, 177, 0, 32, 111, 177, 0, 32, 117, 177, 0, 32, 119, 177, 0, 32, 121, 177, 0, 32, 123, 177, 0, 32, 125, 177, 0, 32, 127, 177, 0, 32, 131, 177, 0, 32, 135, 177, 0, 32, 145, 177, 0, 32, 149, 177, 0, 32, 153, 177, 0, 32, 155, 177, 0, 32, 157, 177, 0, 32, 161, 177, 0, 32, 163, 177, 0, 32, 167, 177, 0, 32, 171, 177, 0, 32, 173, 177, 0, 32, 175, 177, 0, 32, 177, 177, 0, 32, 183, 177, 0, 32, 187, 177, 0, 32, 191, 177, 0, 32, 195, 177, 0, 32, 199, 177, 0, 32, 201, 177, 0, 32, 205, 177, 0, 32, 207, 177, 0, 32, 209, 177, 0, 32, 213, 177, 0, 32, 215, 177, 0, 32, 221, 177, 0, 32, 225, 177, 0, 32, 229, 177, 0, 32, 231, 177, 0, 32, 233, 177, 0, 32, 235, 177, 0, 32, 243, 177, 0, 32, 245, 177, 0, 32, 247, 177, 0, 32, 253, 177, 0, 32, 1, 178, 0, 32, 3, 178, 0, 32, 7, 178, 0, 32, 9, 178, 0, 32, 15, 178, 0, 32, 19, 178, 0, 32, 25, 178, 0, 32, 27, 178, 0, 32, 35, 178, 0, 32, 41, 178, 0, 32, 43, 178, 0, 32, 47, 178, 0, 32, 49, 178, 0, 32, 51, 178, 0, 32, 55, 178, 0, 32, 57, 178, 0, 32, 61, 178, 0, 32, 65, 178, 0, 32, 69, 178, 0, 32, 71, 178, 0, 32, 73, 178, 0, 32, 75, 178, 0, 32, 77, 178, 0, 32, 79, 178, 0, 32, 81, 178, 0, 32, 85, 178, 0, 32, 89, 178, 0, 32, 91, 178, 0, 32, 93, 178, 0, 32, 97, 178, 0, 32, 101, 178, 0, 32, 105, 178, 0, 32, 107, 178, 0, 32, 111, 178, 0, 32, 115, 178, 0, 32, 119, 178, 0, 32, 121, 178, 0, 32, 123, 178, 0, 32, 125, 178, 0, 32, 129, 178, 0, 32, 131, 178, 0, 32, 133, 178, 0, 32, 135, 178, 0, 32, 137, 178, 0, 32, 139, 178, 0, 32, 141, 178, 0, 32, 145, 178, 0, 32, 149, 178, 0, 32, 151, 178, 0, 32, 153, 178, 0, 32, 157, 178, 0, 32, 159, 178, 0, 32, 161, 178, 0, 32, 163, 178, 0, 32, 165, 178, 0, 32, 167, 178, 0, 32, 169, 178, 0, 32, 171, 178, 0, 32, 173, 178, 0, 32, 175, 178, 0, 32, 181, 178, 0, 32, 185, 178, 0, 32, 189, 178, 0, 32, 191, 178, 0, 32, 193, 178, 0, 32, 195, 178, 0, 32, 197, 178, 0, 32, 199, 178, 0, 32, 211, 178, 0, 32, 215, 178, 0, 32, 217, 178, 0, 32, 221, 178, 0, 32, 223, 178, 0, 32, 155, 0, 0, 0, 154, 7, 1, 144, 159, 7, 1, 144, 164, 7, 1, 144, 169, 7, 1, 144, 174, 7, 1, 144, 179, 7, 1, 144, 184, 7, 1, 144, 189, 7, 1, 144, 194, 7, 1, 144, 199, 7, 1, 144, 204, 7, 1, 144, 209, 7, 1, 144, 214, 7, 1, 144, 219, 7, 1, 144, 224, 7, 1, 144, 229, 7, 1, 144, 234, 7, 1, 144, 239, 7, 1, 144, 244, 7, 1, 144, 249, 7, 1, 144, 254, 7, 1, 144, 3, 8, 1, 144, 8, 8, 1, 144, 13, 8, 1, 144, 18, 8, 1, 144, 23, 8, 1, 144, 28, 8, 1, 144, 33, 8, 1, 144, 38, 8, 1, 144, 43, 8, 1, 144, 48, 8, 1, 144, 53, 8, 1, 144, 58, 8, 1, 144, 63, 8, 1, 144, 68, 8, 1, 144, 73, 8, 1, 144, 78, 8, 1, 144, 83, 8, 1, 144, 88, 8, 1, 144, 93, 8, 1, 144, 98, 8, 1, 144, 103, 8, 1, 144, 108, 8, 1, 144, 113, 8, 1, 144, 118, 8, 1, 144, 123, 8, 1, 144, 128, 8, 1, 144, 133, 8, 1, 144, 138, 8, 1, 144, 143, 8, 1, 144, 148, 8, 1, 144, 153, 8, 1, 144, 158, 8, 1, 144, 163, 8, 1, 144, 168, 8, 1, 144, 173, 8, 1, 144, 178, 8, 1, 144, 183, 8, 1, 144, 188, 8, 1, 144, 193, 8, 1, 144, 198, 8, 1, 144, 203, 8, 1, 144, 208, 8, 1, 144, 213, 8, 1, 144, 218, 8, 1, 144, 223, 8, 1, 144, 228, 8, 1, 144, 233, 8, 1, 144, 238, 8, 1, 144, 243, 8, 1, 144, 248, 8, 1, 144, 253, 8, 1, 144, 2, 9, 1, 144, 7, 9, 1, 144, 12, 9, 1, 144, 17, 9, 1, 144, 22, 9, 1, 144, 27, 9, 1, 144, 32, 9, 1, 144, 37, 9, 1, 144, 42, 9, 1, 144, 47, 9, 1, 144, 52, 9, 1, 144, 57, 9, 1, 144, 62, 9, 1, 144, 67, 9, 1, 144, 72, 9, 1, 144, 77, 9, 1, 144, 82, 9, 1, 144, 87, 9, 1, 144, 92, 9, 1, 144, 97, 9, 1, 144, 102, 9, 1, 144, 107, 9, 1, 144, 112, 9, 1, 144, 117, 9, 1, 144, 122, 9, 1, 144, 127, 9, 1, 144, 132, 9, 1, 144, 137, 9, 1, 144, 142, 9, 1, 144, 147, 9, 1, 144, 152, 9, 1, 144, 157, 9, 1, 144, 162, 9, 1, 144, 167, 9, 1, 144, 172, 9, 1, 144, 177, 9, 1, 144, 182, 9, 1, 144, 187, 9, 1, 144, 192, 9, 1, 144, 197, 9, 1, 144, 202, 9, 1, 144, 207, 9, 1, 144, 212, 9, 1, 144, 217, 9, 1, 144, 222, 9, 1, 144, 227, 9, 1, 144, 232, 9, 1, 144, 237, 9, 1, 144, 242, 9, 1, 144, 247, 9, 1, 144, 252, 9, 1, 144, 1, 10, 1, 144, 6, 10, 1, 144, 11, 10, 1, 144, 16, 10, 1, 144, 21, 10, 1, 144, 26, 10, 1, 144, 31, 10, 1, 144, 36, 10, 1, 144, 41, 10, 1, 144, 46, 10, 1, 144, 51, 10, 1, 144, 56, 10, 1, 144, 61, 10, 1, 144, 66, 10, 1, 144, 71, 10, 1, 144, 76, 10, 1, 144, 81, 10, 1, 144, 86, 10, 1, 144, 91, 10, 1, 144, 96, 10, 1, 144, 101, 10, 1, 144, 106, 10, 1, 144, 111, 10, 1, 144, 116, 10, 1, 144, 121, 10, 1, 144, 126, 10, 1, 144, 131, 10, 1, 144, 136, 10, 1, 144, 141, 10, 1, 144, 146, 10, 1, 144, 151, 10, 1, 144, 156, 10, 1, 144, 1, 0, 34, 14, 113, 183, 0, 128, 2, 0, 63, 14, 105, 14, 170, 170, 161, 10, 1, 80, 170, 10, 1, 144, 1, 0, 34, 14, 15, 184, 0, 32, 88, 1, 0, 0, 177, 10, 1, 144, 182, 10, 1, 144, 187, 10, 1, 144, 192, 10, 1, 144, 197, 10, 1, 144, 202, 10, 1, 144, 207, 10, 1, 144, 212, 10, 1, 144, 217, 10, 1, 144, 222, 10, 1, 144, 227, 10, 1, 144, 232, 10, 1, 144, 237, 10, 1, 144, 242, 10, 1, 144, 247, 10, 1, 144, 252, 10, 1, 144, 1, 11, 1, 144, 6, 11, 1, 144, 11, 11, 1, 144, 16, 11, 1, 144, 21, 11, 1, 144, 26, 11, 1, 144, 31, 11, 1, 144, 36, 11, 1, 144, 41, 11, 1, 144, 46, 11, 1, 144, 51, 11, 1, 144, 56, 11, 1, 144, 61, 11, 1, 144, 66, 11, 1, 144, 71, 11, 1, 144, 76, 11, 1, 144, 81, 11, 1, 144, 86, 11, 1, 144, 91, 11, 1, 144, 96, 11, 1, 144, 101, 11, 1, 144, 106, 11, 1, 144, 111, 11, 1, 144, 116, 11, 1, 144, 121, 11, 1, 144, 126, 11, 1, 144, 131, 11, 1, 144, 136, 11, 1, 144, 141, 11, 1, 144, 146, 11, 1, 144, 151, 11, 1, 144, 156, 11, 1, 144, 161, 11, 1, 144, 166, 11, 1, 144, 171, 11, 1, 144, 176, 11, 1, 144, 181, 11, 1, 144, 186, 11, 1, 144, 191, 11, 1, 144, 196, 11, 1, 144, 201, 11, 1, 144, 206, 11, 1, 144, 211, 11, 1, 144, 216, 11, 1, 144, 221, 11, 1, 144, 226, 11, 1, 144, 231, 11, 1, 144, 236, 11, 1, 144, 241, 11, 1, 144, 246, 11, 1, 144, 251, 11, 1, 144, 0, 12, 1, 144, 5, 12, 1, 144, 10, 12, 1, 144, 15, 12, 1, 144, 20, 12, 1, 144, 25, 12, 1, 144, 30, 12, 1, 144, 35, 12, 1, 144, 40, 12, 1, 144, 45, 12, 1, 144, 50, 12, 1, 144, 55, 12, 1, 144, 60, 12, 1, 144, 65, 12, 1, 144, 70, 12, 1, 144, 75, 12, 1, 144, 80, 12, 1, 144, 85, 12, 1, 144, 90, 12, 1, 144, 95, 12, 1, 144, 100, 12, 1, 144, 105, 12, 1, 144, 110, 12, 1, 144, 115, 12, 1, 144, 120, 12, 1, 144, 125, 12, 1, 144, 130, 12, 1, 144, 135, 12, 1, 144, 140, 12, 1, 144, 145, 12, 1, 144, 150, 12, 1, 144, 155, 12, 1, 144, 160, 12, 1, 144, 165, 12, 1, 144, 170, 12, 1, 144, 175, 12, 1, 144, 180, 12, 1, 144, 185, 12, 1, 144, 190, 12, 1, 144, 195, 12, 1, 144, 200, 12, 1, 144, 205, 12, 1, 144, 210, 12, 1, 144, 215, 12, 1, 144, 220, 12, 1, 144, 225, 12, 1, 144, 230, 12, 1, 144, 235, 12, 1, 144, 240, 12, 1, 144, 245, 12, 1, 144, 250, 12, 1, 144, 255, 12, 1, 144, 4, 13, 1, 144, 9, 13, 1, 144, 14, 13, 1, 144, 19, 13, 1, 144, 24, 13, 1, 144, 29, 13, 1, 144, 34, 13, 1, 144, 39, 13, 1, 144, 44, 13, 1, 144, 49, 13, 1, 144, 54, 13, 1, 144, 59, 13, 1, 144, 64, 13, 1, 144, 69, 13, 1, 144, 74, 13, 1, 144, 79, 13, 1, 144, 84, 13, 1, 144, 89, 13, 1, 144, 94, 13, 1, 144, 99, 13, 1, 144, 104, 13, 1, 144, 109, 13, 1, 144, 114, 13, 1, 144, 119, 13, 1, 144, 124, 13, 1, 144, 129, 13, 1, 144, 134, 13, 1, 144, 139, 13, 1, 144, 144, 13, 1, 144, 149, 13, 1, 144, 154, 13, 1, 144, 159, 13, 1, 144, 164, 13, 1, 144, 169, 13, 1, 144, 174, 13, 1, 144, 179, 13, 1, 144, 184, 13, 1, 144, 189, 13, 1, 144, 194, 13, 1, 144, 199, 13, 1, 144, 204, 13, 1, 144, 209, 13, 1, 144, 214, 13, 1, 144, 219, 13, 1, 144, 224, 13, 1, 144, 229, 13, 1, 144, 234, 13, 1, 144, 239, 13, 1, 144, 244, 13, 1, 144, 249, 13, 1, 144, 254, 13, 1, 144, 3, 14, 1, 144, 8, 14, 1, 144, 13, 14, 1, 144, 18, 14, 1, 144, 23, 14, 1, 144, 28, 14, 1, 144, 33, 14, 1, 144, 38, 14, 1, 144, 43, 14, 1, 144, 48, 14, 1, 144, 53, 14, 1, 144, 58, 14, 1, 144, 63, 14, 1, 144, 68, 14, 1, 144, 73, 14, 1, 144, 78, 14, 1, 144, 83, 14, 1, 144, 88, 14, 1, 144, 93, 14, 1, 144, 98, 14, 1, 144, 103, 14, 1, 144, 108, 14, 1, 144, 113, 14, 1, 144, 118, 14, 1, 144, 123, 14, 1, 144, 128, 14, 1, 144, 133, 14, 1, 144, 138, 14, 1, 144, 143, 14, 1, 144, 148, 14, 1, 144, 153, 14, 1, 144, 158, 14, 1, 144, 163, 14, 1, 144, 168, 14, 1, 144, 173, 14, 1, 144, 178, 14, 1, 144, 183, 14, 1, 144, 188, 14, 1, 144, 193, 14, 1, 144, 198, 14, 1, 144, 203, 14, 1, 144, 208, 14, 1, 144, 213, 14, 1, 144, 218, 14, 1, 144, 223, 14, 1, 144, 228, 14, 1, 144, 233, 14, 1, 144, 238, 14, 1, 144, 243, 14, 1, 144, 248, 14, 1, 144, 253, 14, 1, 144, 2, 15, 1, 144, 7, 15, 1, 144, 12, 15, 1, 144, 17, 15, 1, 144, 22, 15, 1, 144, 27, 15, 1, 144, 32, 15, 1, 144, 37, 15, 1, 144, 42, 15, 1, 144, 47, 15, 1, 144, 52, 15, 1, 144, 57, 15, 1, 144, 62, 15, 1, 144, 67, 15, 1, 144, 72, 15, 1, 144, 77, 15, 1, 144, 82, 15, 1, 144, 87, 15, 1, 144, 92, 15, 1, 144, 97, 15, 1, 144, 102, 15, 1, 144, 107, 15, 1, 144, 112, 15, 1, 144, 117, 15, 1, 144, 122, 15, 1, 144, 127, 15, 1, 144, 132, 15, 1, 144, 137, 15, 1, 144, 142, 15, 1, 144, 147, 15, 1, 144, 152, 15, 1, 144, 157, 15, 1, 144, 162, 15, 1, 144, 167, 15, 1, 144, 172, 15, 1, 144, 177, 15, 1, 144, 182, 15, 1, 144, 187, 15, 1, 144, 192, 15, 1, 144, 197, 15, 1, 144, 202, 15, 1, 144, 207, 15, 1, 144, 212, 15, 1, 144, 217, 15, 1, 144, 222, 15, 1, 144, 227, 15, 1, 144, 232, 15, 1, 144, 237, 15, 1, 144, 242, 15, 1, 144, 247, 15, 1, 144, 252, 15, 1, 144, 1, 16, 1, 144, 6, 16, 1, 144, 11, 16, 1, 144, 16, 16, 1, 144, 21, 16, 1, 144, 26, 16, 1, 144, 31, 16, 1, 144, 36, 16, 1, 144, 41, 16, 1, 144, 46, 16, 1, 144, 51, 16, 1, 144, 56, 16, 1, 144, 61, 16, 1, 144, 66, 16, 1, 144, 71, 16, 1, 144, 76, 16, 1, 144, 81, 16, 1, 144, 86, 16, 1, 144, 91, 16, 1, 144, 96, 16, 1, 144, 101, 16, 1, 144, 106, 16, 1, 144, 111, 16, 1, 144, 116, 16, 1, 144, 121, 16, 1, 144, 126, 16, 1, 144, 131, 16, 1, 144, 136, 16, 1, 144, 141, 16, 1, 144, 146, 16, 1, 144, 151, 16, 1, 144, 156, 16, 1, 144, 161, 16, 1, 144, 166, 16, 1, 144, 171, 16, 1, 144, 176, 16, 1, 144, 181, 16, 1, 144, 186, 16, 1, 144, 191, 16, 1, 144, 196, 16, 1, 144, 201, 16, 1, 144, 206, 16, 1, 144, 211, 16, 1, 144, 216, 16, 1, 144, 221, 16, 1, 144, 226, 16, 1, 144, 231, 16, 1, 144, 236, 16, 1, 144, 241, 16, 1, 144, 246, 16, 1, 144, 251, 16, 1, 144, 0, 17, 1, 144, 5, 17, 1, 144, 10, 17, 1, 144, 15, 17, 1, 144, 20, 17, 1, 144, 25, 17, 1, 144, 30, 17, 1, 144, 35, 17, 1, 144, 40, 17, 1, 144, 45, 17, 1, 144, 50, 17, 1, 144, 55, 17, 1, 144, 60, 17, 1, 144, 65, 17, 1, 144, 70, 17, 1, 144, 75, 17, 1, 144, 80, 17, 1, 144, 85, 17, 1, 144, 90, 17, 1, 144, 95, 17, 1, 144, 100, 17, 1, 144, 1, 0, 34, 14, 21, 184, 0, 128, 2, 0, 0, 0, 41, 1, 0, 0, 210, 0, 0, 0, 2, 0, 157, 14, 175, 14, 170, 170, 0, 0, 0, 112, 112, 185, 0, 224, 1, 0, 214, 14, 1, 0, 0, 112, 1, 0, 188, 14, 119, 185, 0, 32, 1, 0, 214, 14, 1, 0, 0, 112, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 2, 0, 188, 14, 175, 14, 170, 170, 123, 185, 0, 32, 125, 185, 0, 224, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 132, 185, 0, 224, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 137, 185, 0, 224, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 142, 185, 0, 224, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 147, 185, 0, 224, 1, 0, 157, 14, 2, 0, 0, 112, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 154, 185, 0, 224, 1, 0, 214, 14, 1, 0, 0, 112, 1, 0, 188, 14, 159, 185, 0, 32, 1, 0, 214, 14, 0, 0, 0, 112, 2, 0, 157, 14, 188, 14, 170, 170, 1, 0, 0, 112, 163, 185, 0, 32, 1, 0, 214, 14, 0, 0, 0, 112, 2, 0, 157, 14, 188, 14, 170, 170, 1, 0, 0, 112, 169, 185, 0, 32, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 175, 185, 0, 224, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 180, 185, 0, 224, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 185, 185, 0, 224, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 190, 185, 0, 224, 1, 0, 214, 14, 1, 0, 0, 112, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 2, 0, 188, 14, 175, 14, 170, 170, 195, 185, 0, 32, 197, 185, 0, 224, 1, 0, 214, 14, 1, 0, 0, 112, 1, 0, 188, 14, 204, 185, 0, 32, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 208, 185, 0, 224, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 2, 0, 157, 14, 175, 14, 170, 170, 1, 0, 0, 112, 213, 185, 0, 224, 2, 0, 0, 0, 23, 1, 0, 0, 216, 0, 0, 0, 1, 0, 175, 14, 220, 185, 0, 224, 21, 0, 11, 23, 185, 14, 226, 14, 60, 23, 229, 14, 232, 14, 235, 14, 238, 14, 241, 14, 244, 14, 247, 14, 250, 14, 253, 14, 0, 15, 3, 15, 6, 15, 9, 15, 12, 15, 15, 15, 94, 14, 18, 15, 115, 185, 0, 32, 121, 185, 0, 32, 128, 185, 0, 32, 135, 185, 0, 32, 140, 185, 0, 32, 145, 185, 0, 32, 150, 185, 0, 32, 152, 185, 0, 32, 157, 185, 0, 32, 161, 185, 0, 32, 165, 185, 0, 32, 171, 185, 0, 32, 178, 185, 0, 32, 183, 185, 0, 32, 188, 185, 0, 32, 193, 185, 0, 32, 200, 185, 0, 32, 206, 185, 0, 32, 211, 185, 0, 32, 216, 185, 0, 32, 223, 185, 0, 32, 6, 0, 35, 15, 42, 15, 49, 15, 56, 15, 62, 15, 67, 15, 170, 170, 105, 17, 1, 144, 191, 17, 1, 144, 201, 17, 1, 144, 224, 17, 1, 144, 236, 17, 1, 144, 147, 199, 0, 96, 131, 1, 101, 15, 35, 3, 38, 3, 104, 15, 107, 15, 110, 15, 113, 15, 116, 15, 119, 15, 122, 15, 125, 15, 128, 15, 131, 15, 134, 15, 137, 15, 140, 15, 143, 15, 41, 3, 146, 15, 149, 15, 152, 15, 24, 23, 155, 15, 30, 23, 185, 14, 158, 15, 161, 15, 164, 15, 226, 14, 60, 23, 167, 15, 170, 15, 173, 15, 84, 23, 176, 15, 229, 14, 179, 15, 44, 3, 232, 14, 235, 14, 182, 15, 185, 15, 188, 15, 191, 15, 194, 15, 50, 3, 197, 15, 200, 15, 203, 15, 53, 3, 36, 23, 206, 15, 56, 3, 62, 3, 42, 23, 65, 3, 209, 15, 212, 15, 215, 15, 218, 15, 221, 15, 238, 14, 224, 15, 227, 15, 230, 15, 233, 15, 236, 15, 239, 15, 48, 23, 242, 15, 241, 14, 245, 15, 248, 15, 251, 15, 254, 15, 1, 16, 4, 16, 7, 16, 10, 16, 68, 3, 72, 23, 71, 3, 74, 3, 13, 16, 90, 23, 16, 16, 77, 3, 80, 3, 19, 16, 22, 16, 25, 16, 28, 16, 83, 3, 31, 16, 34, 16, 86, 3, 89, 3, 37, 16, 40, 16, 92, 3, 43, 16, 46, 16, 49, 16, 247, 14, 52, 16, 55, 16, 58, 16, 61, 16, 95, 3, 98, 3, 64, 16, 67, 16, 70, 16, 73, 16, 76, 16, 79, 16, 82, 16, 250, 14, 85, 16, 104, 3, 88, 16, 91, 16, 94, 16, 97, 16, 253, 14, 100, 16, 103, 16, 106, 16, 109, 16, 112, 16, 0, 15, 115, 16, 118, 16, 121, 16, 124, 16, 127, 16, 107, 3, 3, 15, 54, 23, 130, 16, 6, 15, 133, 16, 136, 16, 110, 3, 59, 15, 12, 15, 139, 16, 113, 3, 142, 16, 145, 16, 148, 16, 151, 16, 116, 3, 154, 16, 157, 16, 119, 3, 160, 16, 122, 3, 163, 16, 166, 16, 169, 16, 172, 16, 175, 16, 178, 16, 181, 16, 184, 16, 187, 16, 190, 16, 15, 15, 125, 3, 193, 16, 128, 3, 196, 16, 131, 3, 199, 16, 202, 16, 205, 16, 134, 3, 208, 16, 211, 16, 214, 16, 217, 16, 28, 26, 220, 16, 223, 16, 226, 16, 229, 16, 94, 14, 232, 16, 235, 16, 238, 16, 18, 15, 241, 16, 244, 16, 247, 16, 250, 16, 137, 3, 132, 23, 253, 16, 0, 17, 3, 17, 8, 17, 12, 17, 16, 17, 20, 17, 24, 17, 28, 17, 32, 17, 36, 17, 40, 17, 46, 17, 52, 17, 57, 17, 62, 17, 67, 17, 72, 17, 77, 17, 82, 17, 87, 17, 92, 17, 97, 17, 102, 17, 107, 17, 112, 17, 117, 17, 121, 17, 125, 17, 130, 17, 135, 17, 140, 17, 145, 17, 150, 17, 155, 17, 160, 17, 165, 17, 170, 17, 175, 17, 180, 17, 185, 17, 190, 17, 195, 17, 200, 17, 205, 17, 210, 17, 215, 17, 220, 17, 225, 17, 230, 17, 235, 17, 240, 17, 245, 17, 250, 17, 255, 17, 4, 18, 9, 18, 14, 18, 19, 18, 24, 18, 29, 18, 34, 18, 39, 18, 44, 18, 49, 18, 54, 18, 59, 18, 64, 18, 69, 18, 73, 18, 77, 18, 81, 18, 85, 18, 91, 18, 97, 18, 103, 18, 109, 18, 115, 18, 121, 18, 127, 18, 133, 18, 139, 18, 145, 18, 151, 18, 157, 18, 163, 18, 169, 18, 175, 18, 181, 18, 187, 18, 193, 18, 199, 18, 205, 18, 209, 18, 213, 18, 217, 18, 221, 18, 225, 18, 229, 18, 233, 18, 237, 18, 241, 18, 245, 18, 249, 18, 253, 18, 2, 19, 7, 19, 12, 19, 17, 19, 22, 19, 27, 19, 32, 19, 36, 19, 40, 19, 44, 19, 48, 19, 53, 19, 58, 19, 63, 19, 68, 19, 73, 19, 78, 19, 83, 19, 88, 19, 93, 19, 98, 19, 103, 19, 108, 19, 113, 19, 118, 19, 123, 19, 128, 19, 133, 19, 138, 19, 143, 19, 147, 19, 151, 19, 155, 19, 159, 19, 163, 19, 167, 19, 171, 19, 175, 19, 179, 19, 183, 19, 187, 19, 192, 19, 197, 19, 202, 19, 207, 19, 212, 19, 217, 19, 222, 19, 227, 19, 232, 19, 237, 19, 242, 19, 247, 19, 251, 19, 255, 19, 3, 20, 7, 20, 11, 20, 15, 20, 19, 20, 23, 20, 27, 20, 31, 20, 36, 20, 41, 20, 46, 20, 51, 20, 56, 20, 61, 20, 66, 20, 71, 20, 76, 20, 81, 20, 86, 20, 91, 20, 96, 20, 101, 20, 106, 20, 111, 20, 116, 20, 121, 20, 125, 20, 129, 20, 133, 20, 20, 18, 1, 144, 28, 18, 1, 144, 36, 18, 1, 144, 71, 18, 1, 144, 80, 18, 1, 144, 93, 18, 1, 144, 105, 18, 1, 144, 124, 18, 1, 144, 149, 18, 1, 144, 159, 18, 1, 144, 168, 18, 1, 144, 239, 18, 1, 144, 243, 18, 1, 144, 255, 18, 1, 144, 8, 19, 1, 144, 12, 19, 1, 144, 26, 19, 1, 144, 55, 19, 1, 144, 60, 19, 1, 144, 79, 19, 1, 144, 92, 19, 1, 144, 97, 19, 1, 144, 107, 19, 1, 144, 111, 19, 1, 144, 139, 19, 1, 144, 172, 19, 1, 144, 193, 19, 1, 144, 210, 19, 1, 144, 218, 19, 1, 144, 225, 19, 1, 144, 239, 19, 1, 144, 10, 20, 1, 144, 28, 20, 1, 144, 41, 20, 1, 144, 68, 20, 1, 144, 83, 20, 1, 144, 100, 20, 1, 144, 111, 20, 1, 144, 146, 20, 1, 144, 180, 20, 1, 144, 188, 20, 1, 144, 205, 20, 1, 144, 208, 20, 1, 144, 215, 20, 1, 144, 230, 20, 1, 144, 247, 20, 1, 144, 254, 20, 1, 144, 4, 21, 1, 144, 15, 21, 1, 144, 26, 21, 1, 144, 75, 21, 1, 144, 100, 21, 1, 144, 116, 21, 1, 144, 144, 21, 1, 144, 151, 21, 1, 144, 171, 21, 1, 144, 183, 21, 1, 144, 203, 21, 1, 144, 209, 21, 1, 144, 214, 21, 1, 144, 241, 21, 1, 144, 251, 21, 1, 144, 3, 22, 1, 144, 11, 22, 1, 144, 24, 22, 1, 144, 35, 22, 1, 144, 41, 22, 1, 144, 48, 22, 1, 144, 57, 22, 1, 144, 60, 22, 1, 144, 75, 22, 1, 144, 98, 22, 1, 144, 103, 22, 1, 144, 114, 22, 1, 144, 133, 22, 1, 144, 155, 22, 1, 144, 166, 22, 1, 144, 210, 22, 1, 144, 218, 22, 1, 144, 223, 22, 1, 144, 230, 22, 1, 144, 11, 23, 1, 144, 30, 23, 1, 144, 62, 23, 1, 144, 71, 23, 1, 144, 92, 23, 1, 144, 107, 23, 1, 144, 120, 23, 1, 144, 168, 23, 1, 144, 216, 23, 1, 144, 226, 23, 1, 144, 252, 23, 1, 144, 0, 24, 1, 144, 4, 24, 1, 144, 7, 24, 1, 144, 20, 24, 1, 144, 38, 24, 1, 144, 45, 24, 1, 144, 64, 24, 1, 144, 83, 24, 1, 144, 92, 24, 1, 144, 103, 24, 1, 144, 115, 24, 1, 144, 125, 24, 1, 144, 141, 24, 1, 144, 152, 24, 1, 144, 223, 24, 1, 144, 236, 24, 1, 144, 100, 25, 1, 144, 123, 25, 1, 144, 136, 25, 1, 144, 154, 25, 1, 144, 192, 25, 1, 144, 216, 25, 1, 144, 223, 25, 1, 144, 226, 25, 1, 144, 42, 26, 1, 144, 54, 26, 1, 144, 70, 26, 1, 144, 93, 26, 1, 144, 109, 26, 1, 144, 178, 26, 1, 144, 196, 26, 1, 144, 218, 26, 1, 144, 222, 26, 1, 144, 255, 26, 1, 144, 16, 27, 1, 144, 28, 27, 1, 144, 43, 27, 1, 144, 52, 27, 1, 144, 90, 27, 1, 144, 108, 27, 1, 144, 127, 27, 1, 144, 148, 27, 1, 144, 161, 27, 1, 144, 176, 27, 1, 144, 194, 27, 1, 144, 206, 27, 1, 144, 220, 27, 1, 144, 247, 27, 1, 144, 14, 28, 1, 144, 32, 28, 1, 144, 41, 28, 1, 144, 58, 28, 1, 144, 75, 28, 1, 144, 96, 28, 1, 144, 113, 28, 1, 144, 132, 28, 1, 144, 141, 28, 1, 144, 184, 28, 1, 144, 205, 28, 1, 144, 33, 29, 1, 144, 39, 29, 1, 144, 53, 29, 1, 144, 64, 29, 1, 144, 90, 29, 1, 144, 109, 29, 1, 144, 131, 29, 1, 144, 137, 29, 1, 144, 141, 29, 1, 144, 98, 30, 1, 144, 107, 30, 1, 144, 113, 30, 1, 144, 123, 30, 1, 144, 138, 30, 1, 144, 157, 30, 1, 144, 168, 30, 1, 144, 179, 30, 1, 144, 182, 30, 1, 144, 197, 30, 1, 144, 212, 30, 1, 144, 217, 30, 1, 144, 241, 30, 1, 144, 247, 30, 1, 144, 70, 31, 1, 144, 76, 31, 1, 144, 90, 31, 1, 144, 97, 31, 1, 144, 122, 31, 1, 144, 128, 31, 1, 144, 210, 31, 1, 144, 226, 31, 1, 144, 235, 31, 1, 144, 2, 32, 1, 144, 34, 32, 1, 144, 62, 32, 1, 144, 67, 32, 1, 144, 77, 32, 1, 144, 135, 32, 1, 144, 155, 32, 1, 144, 170, 32, 1, 144, 177, 32, 1, 144, 203, 32, 1, 144, 11, 33, 1, 144, 18, 33, 1, 144, 22, 33, 1, 144, 34, 33, 1, 144, 57, 33, 1, 144, 67, 33, 1, 144, 78, 33, 1, 144, 89, 33, 1, 144, 98, 33, 1, 144, 105, 33, 1, 144, 117, 33, 1, 144, 131, 33, 1, 144, 142, 33, 1, 144, 151, 33, 1, 144, 160, 33, 1, 144, 165, 33, 1, 144, 170, 33, 1, 144, 176, 33, 1, 144, 182, 33, 1, 144, 189, 33, 1, 144, 87, 116, 0, 96, 192, 33, 1, 144, 196, 33, 1, 144, 200, 33, 1, 144, 205, 33, 1, 144, 209, 33, 1, 144, 215, 33, 1, 144, 219, 33, 1, 144, 224, 33, 1, 144, 228, 33, 1, 144, 233, 33, 1, 144, 238, 33, 1, 144, 247, 33, 1, 144, 6, 34, 1, 144, 19, 34, 1, 144, 27, 34, 1, 144, 35, 34, 1, 144, 39, 34, 1, 144, 47, 34, 1, 144, 52, 34, 1, 144, 58, 34, 1, 144, 63, 34, 1, 144, 69, 34, 1, 144, 77, 34, 1, 144, 83, 34, 1, 144, 88, 34, 1, 144, 95, 34, 1, 144, 100, 34, 1, 144, 105, 34, 1, 144, 109, 34, 1, 144, 113, 34, 1, 144, 116, 34, 1, 144, 121, 34, 1, 144, 125, 34, 1, 144, 128, 34, 1, 144, 133, 34, 1, 144, 137, 34, 1, 144, 146, 34, 1, 144, 105, 40, 0, 96, 173, 107, 0, 96, 150, 34, 1, 144, 160, 34, 1, 144, 166, 34, 1, 144, 169, 34, 1, 144, 174, 34, 1, 144, 177, 34, 1, 144, 112, 52, 0, 96, 92, 52, 0, 96, 241, 108, 0, 96, 97, 52, 0, 96, 182, 34, 1, 144, 72, 52, 0, 96, 186, 34, 1, 144, 190, 34, 1, 144, 196, 34, 1, 144, 200, 34, 1, 144, 204, 34, 1, 144, 208, 34, 1, 144, 221, 34, 1, 144, 230, 34, 1, 144, 235, 34, 1, 144, 238, 34, 1, 144, 245, 34, 1, 144, 237, 52, 0, 96, 0, 35, 1, 144, 6, 35, 1, 144, 11, 53, 0, 96, 26, 53, 0, 96, 15, 35, 1, 144, 28, 35, 1, 144, 34, 35, 1, 144, 48, 35, 1, 144, 55, 35, 1, 144, 61, 35, 1, 144, 214, 35, 1, 144, 226, 35, 1, 144, 3, 36, 1, 144, 26, 36, 1, 144, 32, 36, 1, 144, 38, 36, 1, 144, 43, 36, 1, 144, 49, 36, 1, 144, 55, 36, 1, 144, 59, 36, 1, 144, 66, 36, 1, 144, 71, 36, 1, 144, 75, 36, 1, 144, 78, 36, 1, 144, 82, 36, 1, 144, 86, 36, 1, 144, 93, 36, 1, 144, 99, 36, 1, 144, 102, 36, 1, 144, 106, 36, 1, 144, 109, 36, 1, 144, 116, 36, 1, 144, 127, 36, 1, 144, 133, 36, 1, 144, 146, 36, 1, 144, 153, 36, 1, 144, 157, 36, 1, 144, 54, 85, 0, 96, 166, 36, 1, 144, 179, 36, 1, 144, 182, 36, 1, 144, 190, 36, 1, 144, 195, 36, 1, 144, 205, 36, 1, 144, 216, 36, 1, 144, 219, 36, 1, 144, 225, 36, 1, 144, 231, 36, 1, 144, 236, 36, 1, 144, 239, 36, 1, 144, 245, 36, 1, 144, 252, 36, 1, 144, 255, 36, 1, 144, 5, 37, 1, 144, 15, 37, 1, 144, 21, 37, 1, 144, 31, 37, 1, 144, 37, 37, 1, 144, 41, 37, 1, 144, 45, 37, 1, 144, 49, 37, 1, 144, 55, 37, 1, 144, 59, 37, 1, 144, 62, 37, 1, 144, 65, 37, 1, 144, 68, 37, 1, 144, 71, 37, 1, 144, 80, 37, 1, 144, 89, 37, 1, 144, 99, 37, 1, 144, 108, 37, 1, 144, 114, 37, 1, 144, 124, 37, 1, 144, 133, 37, 1, 144, 139, 37, 1, 144, 145, 37, 1, 144, 150, 37, 1, 144, 155, 37, 1, 144, 158, 37, 1, 144, 173, 37, 1, 144, 184, 37, 1, 144, 194, 37, 1, 144, 201, 37, 1, 144, 215, 37, 1, 144, 219, 37, 1, 144, 223, 37, 1, 144, 227, 37, 1, 144, 231, 37, 1, 144, 234, 37, 1, 144, 239, 37, 1, 144, 245, 37, 1, 144, 253, 37, 1, 144, 4, 38, 1, 144, 11, 38, 1, 144, 16, 38, 1, 144, 23, 38, 1, 144, 28, 38, 1, 144, 33, 38, 1, 144, 41, 38, 1, 144, 45, 38, 1, 144, 51, 38, 1, 144, 56, 38, 1, 144, 63, 38, 1, 144, 69, 38, 1, 144, 75, 38, 1, 144, 81, 38, 1, 144, 89, 38, 1, 144, 116, 38, 1, 144, 153, 38, 1, 144, 188, 38, 1, 144, 3, 0, 11, 23, 158, 20, 194, 20, 22, 40, 1, 144, 75, 112, 0, 96, 26, 40, 1, 144, 8, 0, 11, 23, 186, 20, 202, 20, 214, 20, 218, 20, 148, 1, 168, 1, 184, 1, 170, 170, 172, 0, 0, 96, 178, 0, 0, 96, 120, 2, 0, 96, 29, 40, 1, 144, 32, 40, 1, 144, 35, 40, 1, 144, 20, 2, 0, 96, 38, 40, 1, 144, 6, 0, 13, 21, 116, 2, 52, 15, 17, 21, 20, 21, 23, 21, 170, 170, 41, 40, 1, 144, 45, 40, 1, 144, 50, 40, 1, 144, 54, 40, 1, 144, 83, 40, 1, 144, 103, 40, 1, 144, 32, 0, 11, 23, 158, 20, 162, 20, 166, 20, 170, 20, 174, 20, 178, 20, 182, 20, 186, 20, 190, 20, 194, 20, 198, 20, 202, 20, 206, 20, 210, 20, 214, 20, 218, 20, 222, 20, 226, 20, 230, 20, 234, 20, 136, 1, 140, 1, 148, 1, 45, 15, 168, 1, 180, 1, 184, 1, 238, 20, 241, 20, 206, 3, 4, 21, 170, 170, 221, 38, 1, 144, 227, 38, 1, 144, 233, 38, 1, 144, 250, 38, 1, 144, 0, 39, 1, 144, 18, 39, 1, 144, 27, 39, 1, 144, 50, 39, 1, 144, 60, 39, 1, 144, 70, 39, 1, 144, 76, 39, 1, 144, 81, 39, 1, 144, 87, 39, 1, 144, 116, 39, 1, 144, 125, 39, 1, 144, 135, 39, 1, 144, 147, 39, 1, 144, 165, 39, 1, 144, 172, 39, 1, 144, 178, 39, 1, 144, 187, 39, 1, 144, 198, 39, 1, 144, 204, 39, 1, 144, 210, 39, 1, 144, 229, 39, 1, 144, 234, 39, 1, 144, 245, 39, 1, 144, 6, 40, 1, 144, 16, 40, 1, 144, 80, 188, 0, 32, 85, 188, 0, 32, 98, 188, 0, 32, 1, 0, 43, 21, 112, 201, 250, 114, 3, 0, 0, 0, 220, 49, 108, 3, 112, 201, 250, 2, 224, 204, 39, 3, 2, 0, 251, 13, 60, 21, 170, 170, 157, 188, 0, 32, 159, 188, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 118, 243, 114, 1, 0, 43, 21, 240, 61, 242, 114, 1, 0, 43, 21, 48, 165, 230, 114, 3, 0, 0, 0, 92, 88, 137, 3, 32, 119, 252, 2, 120, 13, 69, 3, 4, 0, 74, 8, 247, 11, 102, 8, 60, 21, 170, 170, 167, 188, 0, 32, 171, 188, 0, 32, 173, 188, 0, 32, 175, 188, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 149, 247, 114, 1, 0, 43, 21, 112, 238, 223, 114, 1, 0, 43, 21, 96, 79, 243, 114, 1, 0, 43, 21, 48, 82, 223, 114, 1, 0, 43, 21, 96, 26, 231, 114, 1, 0, 43, 21, 208, 216, 224, 114, 3, 0, 0, 0, 0, 104, 173, 3, 224, 105, 249, 2, 148, 75, 101, 3, 7, 0, 112, 13, 230, 4, 251, 13, 222, 9, 27, 8, 26, 6, 60, 21, 186, 188, 0, 32, 190, 188, 0, 32, 192, 188, 0, 32, 194, 188, 0, 32, 196, 188, 0, 32, 198, 188, 0, 32, 200, 188, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 240, 98, 215, 114, 2, 0, 43, 21, 86, 21, 170, 170, 176, 198, 214, 114, 32, 13, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 79, 243, 114, 1, 0, 43, 21, 176, 108, 229, 114, 1, 0, 43, 21, 208, 87, 183, 114, 1, 0, 43, 21, 192, 64, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 61, 242, 114, 3, 0, 0, 0, 228, 31, 158, 3, 232, 247, 239, 2, 32, 21, 107, 3, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 16, 4, 223, 114, 1, 0, 43, 21, 160, 196, 187, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 240, 151, 227, 114, 11, 0, 230, 4, 49, 5, 222, 9, 168, 7, 100, 21, 47, 11, 26, 6, 60, 21, 230, 10, 155, 11, 108, 21, 215, 188, 0, 32, 219, 188, 0, 32, 223, 188, 0, 32, 227, 188, 0, 32, 229, 188, 0, 32, 231, 188, 0, 32, 233, 188, 0, 32, 237, 188, 0, 224, 241, 188, 0, 32, 245, 188, 0, 32, 247, 188, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 205, 248, 114, 1, 0, 43, 21, 16, 4, 223, 114, 3, 0, 0, 0, 120, 237, 135, 3, 112, 201, 250, 2, 228, 154, 70, 3, 3, 0, 251, 13, 31, 4, 60, 21, 12, 189, 0, 32, 16, 189, 0, 32, 18, 189, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 3, 0, 0, 0, 168, 175, 119, 3, 48, 45, 250, 2, 12, 162, 58, 3, 2, 0, 251, 13, 60, 21, 170, 170, 27, 189, 0, 32, 31, 189, 0, 224, 1, 0, 43, 21, 48, 82, 223, 114, 1, 0, 43, 21, 176, 85, 212, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 164, 5, 153, 3, 128, 115, 250, 2, 58, 40, 91, 3, 4, 0, 52, 25, 253, 9, 54, 12, 60, 21, 170, 170, 39, 189, 0, 32, 41, 189, 0, 32, 43, 189, 0, 32, 47, 189, 0, 224, 1, 0, 43, 21, 0, 0, 0, 112, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 1, 0, 43, 21, 16, 117, 225, 114, 3, 0, 0, 0, 8, 218, 151, 3, 224, 224, 250, 2, 206, 32, 91, 3, 4, 0, 199, 11, 250, 5, 101, 5, 60, 21, 170, 170, 58, 189, 0, 32, 60, 189, 0, 32, 64, 189, 0, 32, 66, 189, 0, 224, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 208, 126, 239, 114, 1, 0, 43, 21, 240, 98, 215, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 231, 245, 114, 3, 0, 0, 0, 128, 189, 165, 3, 64, 108, 246, 2, 7, 98, 106, 3, 1, 0, 43, 21, 16, 27, 240, 114, 5, 0, 164, 8, 161, 8, 31, 4, 60, 21, 45, 13, 77, 189, 0, 32, 81, 189, 0, 32, 83, 189, 0, 32, 87, 189, 0, 224, 91, 189, 0, 32, 3, 0, 0, 0, 0, 118, 104, 3, 112, 201, 250, 2, 224, 8, 30, 3, 1, 0, 43, 21, 32, 119, 252, 114, 2, 0, 60, 21, 49, 13, 170, 170, 101, 189, 0, 224, 105, 189, 0, 32, 1, 0, 43, 21, 160, 249, 199, 114, 1, 0, 43, 21, 96, 26, 231, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 1, 0, 43, 21, 112, 19, 197, 114, 3, 0, 0, 0, 52, 219, 176, 3, 120, 158, 250, 2, 30, 146, 108, 3, 5, 0, 137, 12, 251, 13, 247, 11, 130, 20, 60, 21, 111, 189, 0, 32, 113, 189, 0, 32, 115, 189, 0, 32, 119, 189, 0, 32, 121, 189, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 80, 123, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 201, 250, 114, 3, 0, 0, 0, 208, 12, 127, 3, 80, 123, 250, 2, 228, 186, 63, 3, 3, 0, 251, 13, 29, 19, 60, 21, 133, 189, 0, 32, 137, 189, 0, 32, 141, 189, 0, 224, 1, 0, 43, 21, 48, 45, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 123, 250, 114, 1, 0, 43, 21, 208, 209, 246, 114, 1, 0, 43, 21, 240, 91, 237, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 192, 64, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 33, 209, 114, 1, 0, 43, 21, 160, 39, 234, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 16, 207, 210, 114, 3, 0, 0, 0, 232, 131, 169, 3, 96, 162, 250, 2, 233, 232, 99, 3, 9, 0, 242, 4, 30, 13, 251, 13, 102, 8, 179, 12, 232, 7, 48, 4, 24, 19, 60, 21, 150, 189, 0, 32, 152, 189, 0, 32, 156, 189, 0, 32, 158, 189, 0, 32, 160, 189, 0, 32, 164, 189, 0, 32, 168, 189, 0, 32, 170, 189, 0, 32, 174, 189, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 64, 84, 250, 114, 1, 0, 43, 21, 48, 82, 223, 114, 3, 0, 0, 0, 0, 243, 179, 3, 112, 201, 250, 2, 13, 85, 105, 3, 1, 0, 43, 21, 48, 149, 191, 114, 1, 0, 43, 21, 80, 160, 223, 114, 5, 0, 251, 13, 48, 4, 60, 21, 119, 13, 140, 21, 192, 189, 0, 32, 196, 189, 0, 32, 198, 189, 0, 224, 202, 189, 0, 32, 204, 189, 0, 32, 1, 0, 43, 21, 160, 99, 224, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 123, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 253, 244, 114, 3, 0, 0, 0, 248, 156, 138, 3, 128, 115, 250, 2, 32, 3, 73, 3, 4, 0, 251, 13, 222, 18, 248, 10, 60, 21, 170, 170, 214, 189, 0, 32, 216, 189, 0, 32, 220, 189, 0, 32, 224, 189, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 201, 250, 114, 3, 0, 0, 0, 136, 50, 131, 3, 32, 119, 252, 2, 240, 248, 59, 3, 2, 0, 89, 12, 60, 21, 170, 170, 235, 189, 0, 32, 239, 189, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 66, 249, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 135, 235, 114, 1, 0, 43, 21, 64, 211, 208, 114, 3, 0, 0, 0, 232, 105, 165, 3, 176, 232, 250, 2, 177, 72, 103, 3, 1, 0, 43, 21, 192, 154, 207, 114, 1, 0, 43, 21, 32, 96, 235, 114, 1, 0, 43, 21, 32, 133, 208, 114, 7, 0, 199, 11, 154, 21, 101, 5, 60, 21, 197, 12, 209, 12, 5, 13, 247, 189, 0, 32, 251, 189, 0, 32, 255, 189, 0, 32, 1, 190, 0, 224, 5, 190, 0, 32, 7, 190, 0, 32, 9, 190, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 201, 250, 114, 3, 0, 71, 21, 43, 21, 86, 21, 16, 206, 0, 96, 112, 201, 250, 114, 32, 13, 228, 114, 1, 0, 43, 21, 16, 140, 242, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 131, 237, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 52, 237, 114, 3, 0, 71, 21, 43, 21, 86, 21, 16, 206, 0, 96, 224, 52, 237, 114, 32, 13, 228, 114, 3, 0, 0, 0, 44, 94, 154, 3, 96, 162, 250, 2, 221, 93, 92, 3, 7, 0, 142, 5, 162, 21, 251, 13, 179, 12, 199, 13, 170, 21, 60, 21, 22, 190, 0, 32, 26, 190, 0, 32, 31, 190, 0, 32, 33, 190, 0, 32, 37, 190, 0, 32, 41, 190, 0, 32, 46, 190, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 8, 59, 140, 3, 200, 228, 250, 2, 203, 180, 75, 3, 2, 0, 251, 13, 60, 21, 170, 170, 61, 190, 0, 32, 65, 190, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 1, 0, 43, 21, 32, 133, 208, 114, 1, 0, 43, 21, 96, 109, 238, 114, 1, 0, 43, 21, 16, 11, 201, 114, 1, 0, 43, 21, 48, 149, 191, 114, 1, 0, 43, 21, 160, 166, 192, 114, 1, 0, 43, 21, 16, 94, 208, 114, 2, 0, 119, 21, 43, 21, 170, 170, 128, 187, 238, 114, 16, 87, 230, 114, 2, 0, 119, 21, 43, 21, 170, 170, 112, 5, 241, 114, 32, 13, 228, 114, 3, 0, 0, 0, 188, 81, 173, 3, 40, 124, 244, 2, 93, 113, 119, 3, 10, 0, 112, 5, 213, 5, 251, 13, 118, 7, 8, 10, 40, 10, 80, 10, 110, 11, 112, 12, 60, 21, 170, 170, 73, 190, 0, 32, 77, 190, 0, 32, 79, 190, 0, 32, 81, 190, 0, 32, 83, 190, 0, 32, 85, 190, 0, 32, 87, 190, 0, 32, 89, 190, 0, 32, 93, 190, 0, 32, 97, 190, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 9, 239, 114, 1, 0, 43, 21, 240, 174, 244, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 122, 241, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 18, 244, 114, 3, 0, 0, 0, 112, 220, 170, 3, 112, 201, 250, 2, 60, 132, 103, 3, 1, 0, 43, 21, 224, 52, 237, 114, 2, 0, 43, 21, 86, 21, 170, 170, 160, 69, 229, 114, 32, 13, 228, 114, 7, 0, 30, 13, 251, 13, 102, 8, 222, 18, 60, 21, 83, 13, 227, 4, 117, 190, 0, 32, 121, 190, 0, 32, 123, 190, 0, 32, 127, 190, 0, 32, 131, 190, 0, 224, 135, 190, 0, 32, 137, 190, 0, 32, 1, 0, 43, 21, 64, 144, 240, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 0, 0, 0, 112, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 9, 239, 114, 1, 0, 43, 21, 192, 200, 241, 114, 3, 0, 0, 0, 100, 255, 152, 3, 88, 15, 240, 2, 115, 195, 104, 3, 6, 0, 150, 6, 209, 6, 178, 21, 102, 8, 28, 10, 60, 21, 170, 170, 152, 190, 0, 32, 154, 190, 0, 32, 156, 190, 0, 32, 158, 190, 0, 32, 162, 190, 0, 32, 164, 190, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 1, 0, 43, 21, 0, 161, 232, 114, 1, 0, 43, 21, 208, 126, 239, 114, 1, 0, 43, 21, 176, 48, 239, 114, 3, 0, 0, 0, 220, 32, 165, 3, 0, 178, 250, 2, 214, 66, 97, 3, 1, 0, 43, 21, 240, 91, 237, 114, 6, 0, 42, 5, 30, 13, 251, 13, 230, 12, 60, 21, 45, 10, 170, 170, 178, 190, 0, 32, 182, 190, 0, 32, 184, 190, 0, 32, 186, 190, 0, 32, 188, 190, 0, 224, 192, 190, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 244, 248, 114, 1, 0, 43, 21, 16, 117, 225, 114, 3, 0, 0, 0, 20, 99, 158, 3, 144, 29, 250, 2, 138, 177, 88, 3, 3, 0, 112, 13, 27, 8, 60, 21, 204, 190, 0, 32, 208, 190, 0, 32, 210, 190, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 253, 197, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 174, 244, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 75, 245, 114, 1, 0, 43, 21, 224, 36, 198, 114, 3, 0, 0, 0, 124, 124, 144, 3, 64, 239, 245, 2, 237, 142, 103, 3, 5, 0, 251, 13, 102, 8, 62, 12, 123, 7, 60, 21, 219, 190, 0, 32, 223, 190, 0, 32, 227, 190, 0, 32, 231, 190, 0, 32, 233, 190, 0, 224, 1, 0, 43, 21, 208, 126, 239, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 5, 241, 114, 3, 0, 0, 0, 204, 102, 151, 3, 128, 38, 242, 2, 61, 123, 103, 3, 1, 0, 43, 21, 48, 165, 230, 114, 4, 0, 231, 6, 102, 8, 60, 21, 214, 10, 170, 170, 245, 190, 0, 32, 247, 190, 0, 32, 251, 190, 0, 224, 255, 190, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 84, 250, 114, 3, 0, 0, 0, 152, 230, 120, 3, 112, 201, 250, 2, 32, 164, 51, 3, 2, 0, 102, 8, 60, 21, 170, 170, 8, 191, 0, 32, 12, 191, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 184, 249, 114, 3, 0, 0, 0, 28, 158, 141, 3, 96, 162, 250, 2, 208, 213, 66, 3, 2, 0, 251, 13, 60, 21, 170, 170, 20, 191, 0, 32, 24, 191, 0, 224, 1, 0, 43, 21, 32, 43, 223, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 223, 249, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 13, 228, 114, 3, 0, 0, 0, 12, 178, 152, 3, 208, 60, 250, 2, 197, 27, 78, 3, 1, 0, 43, 21, 240, 91, 237, 114, 5, 0, 251, 13, 39, 12, 192, 21, 60, 21, 140, 21, 32, 191, 0, 32, 34, 191, 0, 32, 38, 191, 0, 32, 42, 191, 0, 224, 46, 191, 0, 32, 1, 0, 43, 21, 224, 43, 176, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 187, 238, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 253, 244, 114, 1, 0, 43, 21, 144, 7, 212, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 144, 240, 114, 3, 0, 0, 0, 80, 77, 160, 3, 192, 152, 249, 2, 231, 121, 103, 3, 6, 0, 186, 4, 148, 18, 247, 11, 130, 20, 63, 11, 60, 21, 170, 170, 56, 191, 0, 32, 58, 191, 0, 32, 62, 191, 0, 32, 66, 191, 0, 32, 68, 191, 0, 32, 72, 191, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 161, 232, 114, 1, 0, 43, 21, 80, 10, 248, 114, 3, 0, 0, 0, 16, 93, 123, 3, 64, 84, 250, 2, 192, 6, 59, 3, 3, 0, 222, 18, 248, 10, 60, 21, 86, 191, 0, 32, 90, 191, 0, 32, 92, 191, 0, 224, 1, 0, 43, 21, 0, 251, 217, 114, 1, 0, 43, 21, 0, 161, 232, 114, 1, 0, 43, 21, 192, 101, 195, 114, 1, 0, 43, 21, 128, 118, 187, 114, 1, 0, 43, 21, 128, 111, 209, 114, 1, 0, 43, 21, 16, 94, 208, 114, 1, 0, 43, 21, 32, 110, 191, 114, 1, 0, 43, 21, 208, 27, 193, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 144, 249, 114, 3, 0, 0, 0, 64, 0, 183, 3, 128, 121, 249, 2, 79, 50, 120, 3, 1, 0, 43, 21, 64, 248, 181, 114, 1, 0, 43, 21, 128, 148, 182, 114, 12, 0, 30, 13, 251, 13, 247, 11, 129, 7, 48, 4, 2, 6, 116, 8, 29, 11, 31, 4, 60, 21, 147, 13, 245, 21, 170, 170, 101, 191, 0, 32, 103, 191, 0, 32, 105, 191, 0, 32, 107, 191, 0, 32, 109, 191, 0, 32, 111, 191, 0, 32, 113, 191, 0, 32, 115, 191, 0, 32, 117, 191, 0, 32, 121, 191, 0, 224, 125, 191, 0, 32, 127, 191, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 24, 94, 149, 3, 160, 68, 250, 2, 26, 79, 76, 3, 2, 0, 251, 13, 60, 21, 170, 170, 148, 191, 0, 32, 152, 191, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 218, 242, 114, 1, 0, 43, 21, 240, 91, 237, 114, 1, 0, 43, 21, 112, 95, 226, 114, 1, 0, 43, 21, 80, 70, 238, 114, 3, 0, 0, 0, 52, 67, 143, 3, 192, 188, 243, 2, 125, 237, 82, 3, 1, 0, 43, 21, 48, 248, 237, 114, 6, 0, 154, 6, 251, 13, 47, 9, 197, 7, 60, 21, 249, 12, 170, 170, 160, 191, 0, 32, 164, 191, 0, 32, 166, 191, 0, 32, 168, 191, 0, 32, 170, 191, 0, 224, 174, 191, 0, 32, 3, 0, 0, 0, 148, 212, 62, 3, 112, 201, 250, 2, 32, 119, 252, 2, 1, 0, 43, 21, 32, 119, 252, 114, 2, 0, 60, 21, 49, 13, 170, 170, 186, 191, 0, 224, 190, 191, 0, 32, 1, 0, 43, 21, 112, 72, 209, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 10, 248, 114, 3, 0, 0, 0, 212, 122, 153, 3, 120, 170, 248, 2, 142, 236, 89, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 36, 245, 114, 4, 0, 175, 11, 251, 13, 60, 21, 210, 22, 170, 170, 196, 191, 0, 32, 198, 191, 0, 32, 202, 191, 0, 224, 206, 191, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 131, 237, 114, 3, 0, 0, 0, 8, 134, 165, 3, 224, 224, 250, 2, 178, 7, 101, 3, 3, 0, 36, 9, 230, 12, 60, 21, 217, 191, 0, 32, 221, 191, 0, 32, 225, 191, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 1, 0, 43, 21, 0, 244, 239, 114, 3, 0, 0, 0, 200, 45, 137, 3, 40, 106, 247, 2, 254, 30, 77, 3, 3, 0, 251, 13, 247, 11, 60, 21, 234, 191, 0, 32, 238, 191, 0, 32, 240, 191, 0, 224, 1, 0, 43, 21, 128, 88, 192, 114, 1, 0, 43, 21, 144, 217, 177, 114, 1, 0, 43, 21, 112, 215, 206, 114, 1, 0, 43, 21, 192, 244, 192, 114, 1, 0, 43, 21, 112, 215, 206, 114, 1, 0, 43, 21, 112, 49, 192, 114, 1, 0, 43, 21, 80, 227, 191, 114, 1, 0, 43, 21, 96, 10, 192, 114, 1, 0, 43, 21, 48, 82, 223, 114, 1, 0, 43, 21, 64, 105, 184, 114, 1, 0, 43, 21, 224, 239, 185, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 205, 248, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 9, 239, 114, 1, 0, 43, 21, 64, 22, 177, 114, 3, 0, 119, 21, 71, 21, 43, 21, 32, 66, 240, 114, 25, 206, 0, 96, 48, 149, 191, 114, 1, 0, 43, 21, 64, 121, 223, 114, 3, 0, 119, 21, 71, 21, 43, 21, 32, 66, 240, 114, 25, 206, 0, 96, 32, 80, 196, 114, 3, 0, 119, 21, 71, 21, 43, 21, 32, 66, 240, 114, 25, 206, 0, 96, 32, 80, 196, 114, 1, 0, 43, 21, 144, 240, 194, 114, 1, 0, 43, 21, 96, 153, 189, 114, 1, 0, 43, 21, 80, 174, 179, 114, 1, 0, 43, 21, 64, 211, 208, 114, 1, 0, 43, 21, 112, 252, 179, 114, 3, 0, 0, 0, 120, 192, 180, 3, 112, 201, 250, 2, 225, 57, 107, 3, 1, 0, 43, 21, 80, 197, 196, 114, 25, 0, 218, 4, 241, 5, 152, 13, 33, 6, 37, 6, 41, 6, 45, 6, 57, 6, 30, 13, 91, 6, 95, 6, 251, 13, 102, 8, 152, 7, 254, 7, 48, 4, 144, 10, 203, 21, 20, 10, 24, 10, 183, 10, 12, 11, 214, 11, 60, 21, 109, 12, 249, 191, 0, 32, 251, 191, 0, 32, 253, 191, 0, 32, 255, 191, 0, 32, 1, 192, 0, 32, 3, 192, 0, 32, 5, 192, 0, 32, 7, 192, 0, 32, 9, 192, 0, 32, 11, 192, 0, 32, 13, 192, 0, 32, 15, 192, 0, 32, 19, 192, 0, 32, 23, 192, 0, 32, 25, 192, 0, 32, 30, 192, 0, 32, 32, 192, 0, 32, 37, 192, 0, 32, 42, 192, 0, 32, 44, 192, 0, 32, 46, 192, 0, 32, 48, 192, 0, 32, 50, 192, 0, 32, 52, 192, 0, 224, 56, 192, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 80, 70, 238, 114, 1, 0, 43, 21, 112, 88, 248, 114, 3, 0, 0, 0, 36, 150, 106, 3, 112, 201, 250, 2, 32, 141, 34, 3, 3, 0, 251, 13, 192, 21, 60, 21, 96, 192, 0, 32, 100, 192, 0, 32, 102, 192, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 56, 226, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 240, 181, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 240, 38, 225, 114, 1, 0, 43, 21, 144, 120, 214, 114, 1, 0, 43, 21, 112, 238, 223, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 18, 235, 114, 1, 0, 43, 21, 32, 246, 210, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 79, 243, 114, 3, 0, 0, 0, 176, 254, 157, 3, 160, 223, 245, 2, 138, 195, 114, 3, 9, 0, 102, 8, 109, 8, 161, 8, 95, 9, 111, 9, 114, 9, 203, 4, 123, 7, 60, 21, 111, 192, 0, 32, 115, 192, 0, 32, 119, 192, 0, 32, 123, 192, 0, 32, 125, 192, 0, 32, 127, 192, 0, 32, 131, 192, 0, 32, 133, 192, 0, 32, 137, 192, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 40, 243, 114, 1, 0, 43, 21, 64, 211, 208, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 40, 243, 114, 3, 0, 0, 0, 248, 112, 137, 3, 48, 81, 244, 2, 42, 66, 95, 3, 4, 0, 102, 8, 161, 8, 242, 12, 60, 21, 170, 170, 155, 192, 0, 32, 159, 192, 0, 32, 161, 192, 0, 32, 165, 192, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 127, 248, 114, 1, 0, 43, 21, 128, 21, 224, 114, 3, 0, 0, 0, 12, 3, 152, 3, 176, 119, 248, 2, 90, 55, 94, 3, 3, 0, 102, 8, 161, 8, 60, 21, 176, 192, 0, 32, 180, 192, 0, 32, 182, 192, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 209, 246, 114, 1, 0, 43, 21, 16, 253, 244, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 226, 238, 114, 3, 0, 71, 21, 43, 21, 86, 21, 7, 206, 0, 96, 80, 153, 245, 114, 32, 13, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 251, 226, 114, 2, 0, 43, 21, 86, 21, 170, 170, 144, 173, 226, 114, 32, 13, 228, 114, 1, 0, 43, 21, 32, 156, 225, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 224, 202, 212, 114, 1, 0, 43, 21, 144, 150, 209, 114, 3, 0, 0, 0, 156, 196, 170, 3, 112, 201, 250, 2, 97, 37, 99, 3, 1, 0, 43, 21, 128, 254, 206, 114, 11, 0, 30, 13, 251, 13, 102, 8, 122, 7, 48, 4, 91, 9, 31, 4, 240, 6, 118, 11, 60, 21, 107, 13, 191, 192, 0, 32, 195, 192, 0, 32, 197, 192, 0, 32, 201, 192, 0, 32, 206, 192, 0, 32, 210, 192, 0, 32, 214, 192, 0, 32, 216, 192, 0, 32, 220, 192, 0, 32, 222, 192, 0, 224, 226, 192, 0, 32, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 240, 91, 237, 114, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 208, 193, 207, 114, 2, 0, 119, 21, 43, 21, 170, 170, 160, 242, 221, 114, 128, 134, 226, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 40, 243, 114, 1, 0, 43, 21, 48, 172, 208, 114, 2, 0, 43, 21, 86, 21, 170, 170, 176, 251, 226, 114, 32, 13, 228, 114, 3, 0, 0, 0, 80, 90, 162, 3, 232, 92, 244, 2, 14, 194, 105, 3, 7, 0, 6, 5, 126, 5, 111, 6, 102, 8, 97, 8, 230, 11, 60, 21, 245, 192, 0, 32, 249, 192, 0, 32, 253, 192, 0, 32, 1, 193, 0, 32, 5, 193, 0, 32, 7, 193, 0, 32, 11, 193, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 3, 0, 0, 0, 252, 149, 121, 3, 48, 45, 250, 2, 184, 129, 54, 3, 2, 0, 251, 13, 60, 21, 170, 170, 26, 193, 0, 32, 30, 193, 0, 224, 1, 0, 43, 21, 240, 181, 222, 114, 1, 0, 43, 21, 240, 234, 234, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 3, 0, 0, 0, 68, 175, 169, 3, 208, 185, 250, 2, 117, 124, 104, 3, 4, 0, 182, 4, 251, 13, 247, 11, 60, 21, 170, 170, 38, 193, 0, 32, 40, 193, 0, 32, 42, 193, 0, 32, 46, 193, 0, 224, 2, 0, 119, 21, 43, 21, 170, 170, 128, 187, 238, 114, 160, 76, 207, 114, 1, 0, 43, 21, 48, 29, 211, 114, 2, 0, 119, 21, 43, 21, 170, 170, 208, 126, 239, 114, 192, 64, 222, 114, 1, 0, 43, 21, 192, 64, 222, 114, 2, 0, 119, 21, 43, 21, 170, 170, 208, 126, 239, 114, 224, 142, 222, 114, 2, 0, 119, 21, 43, 21, 170, 170, 32, 66, 240, 114, 80, 220, 213, 114, 2, 0, 43, 21, 86, 21, 170, 170, 208, 103, 222, 114, 32, 13, 228, 114, 2, 0, 119, 21, 43, 21, 170, 170, 32, 66, 240, 114, 32, 103, 213, 114, 1, 0, 43, 21, 224, 112, 227, 114, 2, 0, 119, 21, 43, 21, 170, 170, 48, 248, 237, 114, 176, 25, 222, 114, 2, 0, 119, 21, 43, 21, 170, 170, 208, 126, 239, 114, 192, 124, 212, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 122, 241, 114, 2, 0, 119, 21, 43, 21, 170, 170, 48, 248, 237, 114, 240, 38, 225, 114, 1, 0, 43, 21, 64, 234, 225, 114, 1, 0, 43, 21, 0, 0, 0, 112, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 14, 246, 114, 1, 0, 43, 21, 160, 76, 207, 114, 2, 0, 119, 21, 43, 21, 170, 170, 32, 66, 240, 114, 224, 89, 210, 114, 1, 0, 43, 21, 32, 103, 213, 114, 1, 0, 43, 21, 192, 11, 210, 114, 1, 0, 43, 21, 96, 116, 216, 114, 2, 0, 43, 21, 86, 21, 170, 170, 176, 85, 212, 114, 32, 13, 228, 114, 1, 0, 43, 21, 160, 242, 221, 114, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 176, 62, 195, 114, 2, 0, 119, 21, 43, 21, 170, 170, 0, 161, 232, 114, 224, 142, 222, 114, 3, 0, 0, 0, 168, 48, 161, 3, 104, 143, 246, 2, 218, 169, 105, 3, 1, 0, 43, 21, 224, 96, 188, 114, 2, 0, 119, 21, 43, 21, 170, 170, 64, 121, 223, 114, 0, 221, 222, 114, 28, 0, 130, 4, 112, 13, 246, 4, 250, 4, 2, 5, 26, 5, 96, 5, 149, 5, 173, 5, 185, 5, 135, 6, 251, 13, 197, 6, 209, 6, 178, 21, 102, 8, 217, 21, 18, 8, 156, 8, 222, 8, 165, 9, 237, 9, 48, 10, 151, 10, 155, 10, 60, 21, 214, 13, 218, 13, 170, 170, 57, 193, 0, 32, 61, 193, 0, 32, 63, 193, 0, 32, 67, 193, 0, 32, 69, 193, 0, 32, 73, 193, 0, 32, 77, 193, 0, 32, 81, 193, 0, 32, 85, 193, 0, 32, 87, 193, 0, 32, 91, 193, 0, 32, 95, 193, 0, 32, 99, 193, 0, 32, 103, 193, 0, 32, 105, 193, 0, 32, 107, 193, 0, 32, 111, 193, 0, 32, 113, 193, 0, 32, 117, 193, 0, 32, 119, 193, 0, 32, 121, 193, 0, 32, 123, 193, 0, 32, 127, 193, 0, 32, 129, 193, 0, 32, 133, 193, 0, 32, 137, 193, 0, 224, 141, 193, 0, 32, 143, 193, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 55, 208, 114, 1, 0, 43, 21, 16, 131, 181, 114, 1, 0, 43, 21, 16, 4, 223, 114, 1, 0, 43, 21, 112, 238, 223, 114, 1, 0, 43, 21, 208, 216, 224, 114, 2, 0, 119, 21, 43, 21, 170, 170, 0, 214, 244, 114, 128, 81, 214, 114, 1, 0, 43, 21, 176, 205, 192, 114, 1, 0, 43, 21, 208, 223, 202, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 176, 115, 207, 114, 1, 0, 43, 21, 160, 23, 195, 114, 1, 0, 43, 21, 224, 126, 183, 114, 1, 0, 43, 21, 80, 197, 196, 114, 1, 0, 43, 21, 0, 0, 0, 112, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 33, 209, 114, 1, 0, 43, 21, 48, 82, 223, 114, 1, 0, 43, 21, 0, 145, 193, 114, 1, 0, 43, 21, 224, 43, 176, 114, 1, 0, 43, 21, 176, 205, 192, 114, 3, 0, 0, 0, 100, 216, 196, 3, 24, 49, 250, 2, 165, 97, 134, 3, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 48, 142, 213, 114, 1, 0, 43, 21, 128, 125, 165, 114, 1, 0, 43, 21, 32, 4, 167, 114, 1, 0, 43, 21, 192, 147, 229, 114, 1, 0, 43, 21, 64, 91, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 176, 228, 209, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 105, 249, 114, 26, 0, 115, 5, 251, 13, 31, 7, 160, 7, 228, 7, 135, 4, 124, 8, 100, 21, 29, 11, 225, 21, 43, 9, 67, 9, 153, 9, 233, 21, 99, 10, 198, 10, 230, 12, 140, 12, 60, 21, 155, 11, 241, 21, 76, 13, 139, 13, 249, 21, 14, 14, 154, 9, 170, 170, 190, 193, 0, 32, 194, 193, 0, 32, 196, 193, 0, 32, 198, 193, 0, 32, 200, 193, 0, 32, 202, 193, 0, 32, 206, 193, 0, 32, 208, 193, 0, 32, 210, 193, 0, 32, 214, 193, 0, 32, 216, 193, 0, 32, 218, 193, 0, 32, 220, 193, 0, 32, 222, 193, 0, 32, 226, 193, 0, 32, 228, 193, 0, 32, 230, 193, 0, 32, 232, 193, 0, 32, 234, 193, 0, 224, 238, 193, 0, 32, 242, 193, 0, 32, 244, 193, 0, 32, 246, 193, 0, 32, 248, 193, 0, 32, 250, 193, 0, 32, 254, 193, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 223, 249, 114, 1, 0, 43, 21, 112, 72, 209, 114, 3, 0, 0, 0, 240, 164, 173, 3, 128, 246, 249, 2, 218, 25, 109, 3, 3, 0, 247, 11, 133, 7, 60, 21, 42, 194, 0, 32, 46, 194, 0, 32, 48, 194, 0, 224, 3, 0, 0, 0, 192, 113, 62, 3, 112, 201, 250, 2, 32, 119, 252, 2, 1, 0, 43, 21, 32, 119, 252, 114, 2, 0, 60, 21, 49, 13, 170, 170, 57, 194, 0, 224, 61, 194, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 3, 0, 0, 0, 232, 85, 160, 3, 248, 95, 250, 2, 26, 26, 94, 3, 2, 0, 247, 11, 60, 21, 170, 170, 67, 194, 0, 32, 71, 194, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 104, 224, 164, 3, 176, 232, 250, 2, 124, 114, 103, 3, 2, 0, 247, 11, 60, 21, 170, 170, 79, 194, 0, 32, 83, 194, 0, 224, 1, 0, 43, 21, 240, 144, 249, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 71, 247, 114, 3, 0, 0, 0, 36, 8, 138, 3, 168, 162, 248, 2, 245, 231, 79, 3, 3, 0, 93, 8, 31, 4, 60, 21, 91, 194, 0, 32, 93, 194, 0, 32, 97, 194, 0, 224, 1, 0, 43, 21, 96, 56, 226, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 161, 232, 114, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 80, 10, 248, 114, 3, 0, 0, 0, 160, 10, 137, 3, 64, 84, 250, 2, 161, 134, 73, 3, 4, 0, 247, 11, 222, 18, 248, 10, 60, 21, 170, 170, 106, 194, 0, 32, 108, 194, 0, 32, 112, 194, 0, 32, 116, 194, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 75, 245, 114, 3, 0, 0, 0, 80, 204, 118, 3, 112, 201, 250, 2, 148, 20, 44, 3, 2, 0, 251, 13, 60, 21, 170, 170, 127, 194, 0, 32, 131, 194, 0, 224, 1, 0, 43, 21, 112, 215, 206, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 1, 0, 43, 21, 208, 209, 246, 114, 1, 0, 43, 21, 96, 26, 231, 114, 1, 0, 43, 21, 32, 133, 208, 114, 3, 0, 0, 0, 104, 96, 152, 3, 184, 189, 250, 2, 61, 97, 88, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 48, 239, 114, 7, 0, 112, 13, 52, 25, 251, 13, 102, 8, 250, 5, 60, 21, 45, 10, 139, 194, 0, 32, 141, 194, 0, 32, 145, 194, 0, 32, 147, 194, 0, 32, 149, 194, 0, 32, 151, 194, 0, 224, 155, 194, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 1, 0, 43, 21, 48, 248, 237, 114, 1, 0, 43, 21, 240, 204, 239, 114, 1, 0, 43, 21, 208, 163, 212, 114, 1, 0, 43, 21, 64, 31, 238, 114, 3, 0, 0, 0, 156, 133, 168, 3, 112, 201, 250, 2, 175, 97, 103, 3, 6, 0, 215, 11, 30, 13, 251, 13, 32, 11, 184, 10, 60, 21, 170, 170, 170, 194, 0, 32, 174, 194, 0, 32, 176, 194, 0, 32, 178, 194, 0, 32, 180, 194, 0, 32, 182, 194, 0, 224, 2, 0, 43, 21, 86, 21, 170, 170, 80, 70, 238, 114, 32, 13, 228, 114, 1, 0, 43, 21, 64, 121, 223, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 144, 249, 114, 2, 0, 43, 21, 86, 21, 170, 170, 176, 235, 187, 114, 32, 13, 228, 114, 1, 0, 43, 21, 32, 246, 210, 114, 1, 0, 43, 21, 64, 114, 245, 114, 1, 0, 43, 21, 192, 147, 229, 114, 1, 0, 43, 21, 96, 109, 238, 114, 1, 0, 43, 21, 64, 188, 191, 114, 1, 0, 43, 21, 208, 117, 178, 114, 2, 0, 43, 21, 86, 21, 170, 170, 112, 238, 223, 114, 32, 13, 228, 114, 1, 0, 43, 21, 176, 55, 217, 114, 2, 0, 43, 21, 86, 21, 170, 170, 128, 88, 192, 114, 32, 13, 228, 114, 1, 0, 43, 21, 96, 26, 231, 114, 1, 0, 43, 21, 160, 189, 209, 114, 1, 0, 43, 21, 224, 59, 215, 114, 2, 0, 43, 21, 86, 21, 170, 170, 0, 131, 237, 114, 32, 13, 228, 114, 1, 0, 43, 21, 160, 39, 234, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 144, 150, 209, 114, 1, 0, 43, 21, 192, 147, 229, 114, 1, 0, 43, 21, 0, 122, 176, 114, 2, 0, 43, 21, 86, 21, 170, 170, 160, 242, 221, 114, 32, 13, 228, 114, 3, 0, 0, 0, 188, 115, 184, 3, 112, 201, 250, 2, 33, 7, 114, 3, 1, 0, 43, 21, 144, 60, 224, 114, 1, 0, 43, 21, 192, 147, 229, 114, 26, 0, 242, 4, 9, 17, 30, 13, 119, 6, 52, 25, 251, 13, 247, 11, 102, 8, 251, 6, 255, 6, 122, 7, 179, 12, 224, 7, 48, 4, 226, 8, 101, 5, 115, 10, 222, 18, 20, 11, 32, 11, 230, 12, 77, 12, 96, 12, 60, 21, 45, 10, 87, 13, 170, 170, 196, 194, 0, 32, 200, 194, 0, 32, 202, 194, 0, 32, 206, 194, 0, 32, 210, 194, 0, 32, 212, 194, 0, 32, 214, 194, 0, 32, 216, 194, 0, 32, 218, 194, 0, 32, 220, 194, 0, 32, 222, 194, 0, 32, 226, 194, 0, 32, 228, 194, 0, 32, 232, 194, 0, 32, 234, 194, 0, 32, 236, 194, 0, 32, 238, 194, 0, 32, 242, 194, 0, 32, 244, 194, 0, 32, 246, 194, 0, 32, 248, 194, 0, 32, 250, 194, 0, 32, 252, 194, 0, 32, 0, 195, 0, 224, 4, 195, 0, 32, 6, 195, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 112, 201, 250, 114, 3, 0, 0, 0, 152, 45, 105, 3, 112, 201, 250, 2, 32, 22, 33, 3, 2, 0, 251, 13, 60, 21, 170, 170, 48, 195, 0, 32, 52, 195, 0, 224, 1, 0, 43, 21, 224, 22, 242, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 143, 231, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 160, 223, 114, 1, 0, 43, 21, 208, 239, 241, 114, 3, 0, 0, 0, 160, 210, 137, 3, 152, 10, 246, 2, 241, 184, 84, 3, 5, 0, 46, 8, 112, 13, 102, 8, 188, 10, 60, 21, 60, 195, 0, 32, 62, 195, 0, 32, 66, 195, 0, 32, 70, 195, 0, 32, 72, 195, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 223, 249, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 48, 218, 242, 114, 1, 0, 43, 21, 160, 205, 248, 114, 1, 0, 43, 21, 32, 246, 210, 114, 1, 0, 43, 21, 0, 0, 0, 112, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 128, 254, 206, 114, 1, 0, 43, 21, 16, 170, 237, 114, 3, 0, 0, 0, 120, 45, 167, 3, 112, 201, 250, 2, 86, 91, 95, 3, 8, 0, 9, 17, 30, 13, 251, 13, 37, 21, 38, 8, 194, 12, 89, 12, 60, 21, 170, 170, 84, 195, 0, 32, 88, 195, 0, 32, 92, 195, 0, 32, 94, 195, 0, 32, 96, 195, 0, 32, 98, 195, 0, 32, 102, 195, 0, 32, 104, 195, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 6, 250, 114, 3, 0, 0, 0, 24, 245, 128, 3, 32, 6, 250, 2, 46, 69, 67, 3, 2, 0, 251, 13, 60, 21, 170, 170, 121, 195, 0, 32, 125, 195, 0, 224, 1, 0, 43, 21, 96, 206, 201, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 149, 247, 114, 3, 0, 0, 0, 8, 109, 165, 3, 200, 109, 249, 2, 140, 82, 103, 3, 3, 0, 251, 13, 247, 11, 60, 21, 133, 195, 0, 32, 135, 195, 0, 32, 139, 195, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 248, 246, 114, 1, 0, 43, 21, 112, 88, 248, 114, 1, 0, 43, 21, 96, 26, 231, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 187, 238, 114, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 224, 82, 232, 114, 3, 0, 0, 0, 48, 102, 172, 3, 48, 194, 246, 2, 127, 27, 108, 3, 6, 0, 112, 13, 194, 4, 251, 13, 102, 8, 49, 8, 60, 21, 170, 170, 148, 195, 0, 32, 152, 195, 0, 32, 154, 195, 0, 32, 156, 195, 0, 32, 160, 195, 0, 32, 164, 195, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 3, 0, 0, 0, 120, 132, 140, 3, 168, 150, 250, 2, 176, 133, 73, 3, 2, 0, 247, 11, 60, 21, 170, 170, 178, 195, 0, 32, 182, 195, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 84, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 70, 238, 114, 1, 0, 43, 21, 144, 30, 229, 114, 3, 0, 0, 0, 40, 187, 165, 3, 96, 168, 249, 2, 185, 68, 104, 3, 4, 0, 247, 11, 63, 11, 70, 11, 60, 21, 170, 170, 190, 195, 0, 32, 194, 195, 0, 32, 198, 195, 0, 32, 200, 195, 0, 224, 1, 0, 43, 21, 96, 79, 243, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 131, 246, 114, 1, 0, 43, 21, 144, 226, 238, 114, 1, 0, 43, 21, 192, 57, 244, 114, 3, 0, 0, 0, 244, 225, 153, 3, 176, 232, 250, 2, 13, 100, 88, 3, 1, 0, 43, 21, 144, 30, 229, 114, 6, 0, 251, 13, 164, 12, 162, 6, 230, 12, 60, 21, 99, 13, 170, 170, 211, 195, 0, 32, 213, 195, 0, 32, 217, 195, 0, 32, 219, 195, 0, 32, 221, 195, 0, 224, 225, 195, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 6, 250, 114, 1, 0, 43, 21, 64, 114, 245, 114, 1, 0, 43, 21, 80, 54, 199, 114, 1, 0, 43, 21, 112, 5, 241, 114, 3, 0, 0, 0, 208, 225, 179, 3, 248, 244, 246, 2, 149, 235, 116, 3, 5, 0, 112, 13, 206, 4, 52, 25, 251, 13, 60, 21, 237, 195, 0, 32, 241, 195, 0, 32, 243, 195, 0, 32, 245, 195, 0, 32, 247, 195, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 132, 215, 130, 3, 96, 79, 243, 2, 223, 175, 80, 3, 2, 0, 112, 13, 60, 21, 170, 170, 3, 196, 0, 32, 7, 196, 0, 224, 1, 0, 43, 21, 64, 234, 225, 114, 3, 0, 71, 21, 43, 21, 86, 21, 16, 206, 0, 96, 16, 230, 227, 114, 32, 13, 228, 114, 1, 0, 43, 21, 208, 103, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 174, 244, 114, 1, 0, 43, 21, 64, 234, 225, 114, 3, 0, 0, 0, 104, 157, 146, 3, 168, 49, 246, 2, 73, 154, 95, 3, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 0, 214, 244, 114, 1, 0, 43, 21, 96, 109, 238, 114, 8, 0, 46, 8, 112, 13, 181, 5, 251, 13, 73, 12, 60, 21, 182, 12, 185, 12, 170, 170, 15, 196, 0, 32, 17, 196, 0, 32, 22, 196, 0, 32, 24, 196, 0, 32, 28, 196, 0, 32, 30, 196, 0, 224, 34, 196, 0, 32, 38, 196, 0, 32, 1, 0, 43, 21, 192, 214, 197, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 208, 103, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 80, 70, 238, 114, 1, 0, 43, 21, 192, 87, 239, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 201, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 64, 121, 223, 114, 1, 0, 43, 21, 208, 163, 212, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 26, 231, 114, 3, 0, 0, 0, 8, 194, 180, 3, 168, 150, 250, 2, 63, 71, 109, 3, 9, 0, 242, 17, 214, 4, 74, 8, 251, 13, 247, 11, 45, 19, 201, 6, 124, 10, 60, 21, 53, 196, 0, 32, 55, 196, 0, 32, 59, 196, 0, 32, 63, 196, 0, 32, 65, 196, 0, 32, 69, 196, 0, 32, 73, 196, 0, 32, 75, 196, 0, 32, 79, 196, 0, 224, 1, 0, 43, 21, 224, 142, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 183, 240, 114, 1, 0, 43, 21, 240, 61, 242, 114, 1, 0, 43, 21, 64, 144, 240, 114, 1, 0, 43, 21, 48, 195, 225, 114, 1, 0, 43, 21, 192, 147, 229, 114, 3, 0, 0, 0, 216, 216, 165, 3, 176, 161, 241, 2, 226, 169, 118, 3, 1, 0, 43, 21, 192, 147, 229, 114, 1, 0, 43, 21, 32, 43, 223, 114, 9, 0, 46, 8, 183, 11, 251, 13, 143, 11, 6, 12, 188, 10, 60, 21, 182, 12, 48, 17, 97, 196, 0, 32, 99, 196, 0, 32, 103, 196, 0, 32, 105, 196, 0, 32, 107, 196, 0, 32, 109, 196, 0, 32, 111, 196, 0, 224, 115, 196, 0, 32, 117, 196, 0, 32, 1, 0, 43, 21, 96, 109, 238, 114, 1, 0, 43, 21, 160, 92, 246, 114, 1, 0, 43, 21, 112, 215, 206, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 6, 250, 114, 1, 0, 43, 21, 32, 163, 203, 114, 1, 0, 43, 21, 208, 133, 217, 114, 1, 0, 43, 21, 192, 101, 195, 114, 1, 0, 43, 21, 48, 149, 191, 114, 1, 0, 43, 21, 48, 149, 191, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 101, 242, 114, 3, 0, 0, 0, 228, 133, 166, 3, 32, 119, 252, 2, 8, 241, 94, 3, 11, 0, 30, 13, 251, 13, 164, 12, 162, 6, 114, 11, 230, 12, 251, 3, 30, 12, 38, 12, 89, 12, 60, 21, 133, 196, 0, 32, 135, 196, 0, 32, 137, 196, 0, 32, 139, 196, 0, 32, 143, 196, 0, 32, 145, 196, 0, 32, 147, 196, 0, 32, 149, 196, 0, 32, 151, 196, 0, 32, 153, 196, 0, 32, 157, 196, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 6, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 161, 241, 114, 1, 0, 43, 21, 0, 101, 242, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 239, 241, 114, 1, 0, 43, 21, 112, 72, 209, 114, 3, 0, 0, 0, 116, 111, 145, 3, 104, 250, 249, 2, 4, 94, 85, 3, 6, 0, 251, 13, 228, 6, 255, 11, 172, 7, 146, 11, 60, 21, 170, 170, 178, 196, 0, 32, 182, 196, 0, 32, 186, 196, 0, 32, 188, 196, 0, 32, 192, 196, 0, 32, 194, 196, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 192, 27, 249, 114, 3, 0, 0, 0, 192, 40, 120, 3, 112, 201, 250, 2, 120, 152, 45, 3, 2, 0, 251, 13, 60, 21, 170, 170, 208, 196, 0, 32, 212, 196, 0, 224, 1, 0, 43, 21, 32, 66, 240, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 96, 244, 114, 1, 0, 43, 21, 208, 43, 232, 114, 1, 0, 43, 21, 176, 48, 239, 114, 3, 0, 0, 0, 224, 81, 122, 3, 208, 66, 249, 2, 155, 208, 72, 3, 1, 0, 43, 21, 208, 216, 224, 114, 1, 0, 43, 21, 0, 48, 230, 114, 7, 0, 225, 5, 251, 13, 183, 8, 43, 11, 60, 21, 41, 13, 210, 13, 220, 196, 0, 32, 222, 196, 0, 32, 226, 196, 0, 32, 228, 196, 0, 32, 230, 196, 0, 224, 234, 196, 0, 32, 236, 196, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 3, 0, 0, 0, 100, 82, 135, 3, 112, 201, 250, 2, 100, 194, 63, 3, 2, 0, 37, 21, 60, 21, 170, 170, 249, 196, 0, 32, 253, 196, 0, 224, 2, 0, 119, 21, 43, 21, 170, 170, 224, 255, 224, 114, 240, 211, 217, 114, 1, 0, 43, 21, 208, 193, 207, 114, 2, 0, 43, 21, 86, 21, 170, 170, 64, 211, 208, 114, 32, 13, 228, 114, 1, 0, 43, 21, 32, 13, 228, 114, 1, 0, 43, 21, 176, 161, 241, 114, 1, 0, 43, 21, 16, 170, 237, 114, 1, 0, 43, 21, 144, 37, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 201, 250, 114, 1, 0, 43, 21, 96, 63, 204, 114, 2, 0, 43, 21, 86, 21, 170, 170, 112, 12, 219, 114, 32, 13, 228, 114, 1, 0, 43, 21, 64, 112, 162, 114, 1, 0, 43, 21, 16, 4, 223, 114, 1, 0, 43, 21, 144, 37, 207, 114, 2, 0, 43, 21, 86, 21, 170, 170, 224, 255, 224, 114, 32, 13, 228, 114, 1, 0, 43, 21, 176, 25, 222, 114, 1, 0, 43, 21, 208, 103, 222, 114, 3, 0, 0, 0, 32, 103, 182, 3, 112, 201, 250, 2, 172, 7, 112, 3, 17, 0, 128, 13, 74, 8, 211, 11, 30, 13, 251, 13, 247, 11, 45, 19, 102, 8, 247, 6, 122, 7, 99, 11, 48, 4, 222, 18, 9, 3, 0, 11, 31, 4, 60, 21, 5, 197, 0, 32, 9, 197, 0, 32, 11, 197, 0, 32, 15, 197, 0, 32, 17, 197, 0, 32, 19, 197, 0, 32, 21, 197, 0, 32, 23, 197, 0, 32, 27, 197, 0, 32, 29, 197, 0, 32, 33, 197, 0, 32, 35, 197, 0, 32, 37, 197, 0, 32, 39, 197, 0, 32, 43, 197, 0, 32, 45, 197, 0, 32, 47, 197, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 75, 245, 114, 1, 0, 43, 21, 160, 39, 234, 114, 3, 0, 0, 0, 8, 31, 153, 3, 208, 66, 249, 2, 192, 185, 89, 3, 3, 0, 102, 8, 59, 11, 60, 21, 77, 197, 0, 32, 81, 197, 0, 32, 83, 197, 0, 224, 1, 0, 43, 21, 240, 98, 215, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 144, 233, 216, 114, 1, 0, 43, 21, 192, 147, 229, 114, 1, 0, 43, 21, 208, 50, 210, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 201, 250, 114, 1, 0, 43, 21, 112, 148, 238, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 32, 223, 193, 114, 3, 0, 71, 21, 43, 21, 86, 21, 25, 206, 0, 96, 176, 2, 205, 114, 32, 13, 228, 114, 1, 0, 43, 21, 208, 50, 210, 114, 1, 0, 43, 21, 240, 15, 208, 114, 1, 0, 43, 21, 48, 96, 179, 114, 1, 0, 43, 21, 176, 62, 195, 114, 1, 0, 43, 21, 176, 55, 217, 114, 2, 0, 43, 21, 86, 21, 170, 170, 176, 138, 224, 114, 32, 13, 228, 114, 1, 0, 43, 21, 240, 241, 212, 114, 3, 0, 0, 0, 20, 130, 182, 3, 112, 201, 250, 2, 108, 175, 111, 3, 1, 0, 43, 21, 144, 97, 197, 114, 1, 0, 43, 21, 32, 103, 213, 114, 18, 0, 112, 5, 137, 12, 30, 13, 52, 25, 251, 13, 102, 8, 38, 13, 55, 7, 48, 4, 215, 8, 249, 8, 27, 8, 233, 10, 210, 11, 112, 12, 60, 21, 109, 12, 140, 21, 170, 170, 92, 197, 0, 32, 94, 197, 0, 32, 98, 197, 0, 32, 100, 197, 0, 32, 102, 197, 0, 32, 106, 197, 0, 32, 108, 197, 0, 32, 112, 197, 0, 32, 117, 197, 0, 32, 119, 197, 0, 32, 121, 197, 0, 32, 123, 197, 0, 32, 125, 197, 0, 32, 127, 197, 0, 32, 131, 197, 0, 32, 133, 197, 0, 224, 137, 197, 0, 32, 139, 197, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 84, 250, 114, 3, 0, 0, 0, 8, 195, 134, 3, 64, 84, 250, 2, 15, 242, 72, 3, 2, 0, 251, 13, 60, 21, 170, 170, 169, 197, 0, 32, 173, 197, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 199, 223, 114, 1, 0, 43, 21, 96, 26, 231, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 205, 248, 114, 1, 0, 43, 21, 80, 190, 218, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 199, 223, 114, 1, 0, 43, 21, 160, 39, 234, 114, 3, 0, 0, 0, 164, 155, 153, 3, 200, 228, 250, 2, 65, 2, 94, 3, 1, 0, 43, 21, 240, 91, 237, 114, 8, 0, 125, 12, 250, 5, 174, 6, 101, 5, 29, 10, 230, 12, 60, 21, 166, 13, 170, 170, 181, 197, 0, 32, 185, 197, 0, 32, 187, 197, 0, 32, 191, 197, 0, 32, 193, 197, 0, 32, 197, 197, 0, 32, 199, 197, 0, 224, 203, 197, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 110, 247, 114, 1, 0, 43, 21, 224, 165, 239, 114, 3, 0, 0, 0, 156, 162, 134, 3, 112, 88, 248, 2, 21, 71, 74, 3, 1, 0, 43, 21, 144, 60, 224, 114, 4, 0, 102, 8, 51, 7, 60, 21, 140, 21, 170, 170, 218, 197, 0, 32, 222, 197, 0, 32, 224, 197, 0, 224, 228, 197, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 68, 142, 137, 3, 32, 119, 252, 2, 218, 38, 66, 3, 2, 0, 251, 13, 60, 21, 170, 170, 237, 197, 0, 32, 241, 197, 0, 224, 1, 0, 43, 21, 32, 13, 228, 114, 1, 0, 43, 21, 224, 255, 224, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 176, 161, 241, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 240, 91, 237, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 226, 238, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 0, 0, 0, 112, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 192, 177, 224, 114, 1, 0, 43, 21, 48, 195, 225, 114, 1, 0, 43, 21, 32, 73, 218, 114, 1, 0, 43, 21, 160, 242, 221, 114, 1, 0, 43, 21, 96, 10, 192, 114, 3, 0, 0, 0, 176, 212, 164, 3, 56, 151, 246, 2, 1, 10, 106, 3, 13, 0, 106, 4, 118, 4, 178, 9, 227, 11, 251, 13, 209, 6, 178, 21, 23, 7, 137, 7, 121, 8, 217, 10, 174, 11, 60, 21, 249, 197, 0, 32, 251, 197, 0, 32, 253, 197, 0, 32, 1, 198, 0, 32, 5, 198, 0, 32, 9, 198, 0, 32, 11, 198, 0, 32, 13, 198, 0, 32, 17, 198, 0, 32, 19, 198, 0, 32, 21, 198, 0, 32, 23, 198, 0, 32, 25, 198, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 227, 247, 114, 1, 0, 43, 21, 96, 79, 243, 114, 3, 0, 0, 0, 48, 99, 135, 3, 64, 227, 247, 2, 82, 120, 60, 3, 3, 0, 251, 13, 247, 11, 60, 21, 49, 198, 0, 32, 53, 198, 0, 32, 55, 198, 0, 224, 1, 0, 43, 21, 32, 209, 237, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 127, 248, 114, 3, 0, 0, 0, 84, 243, 135, 3, 32, 119, 252, 2, 14, 199, 64, 3, 3, 0, 9, 17, 194, 12, 60, 21, 64, 198, 0, 32, 66, 198, 0, 32, 70, 198, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 192, 200, 241, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 16, 27, 240, 114, 3, 0, 0, 0, 16, 191, 140, 3, 88, 122, 243, 2, 121, 175, 89, 3, 5, 0, 251, 13, 209, 6, 178, 21, 181, 9, 60, 21, 79, 198, 0, 32, 83, 198, 0, 32, 85, 198, 0, 32, 87, 198, 0, 32, 89, 198, 0, 224, 1, 0, 43, 21, 224, 165, 239, 114, 1, 0, 43, 21, 0, 0, 0, 112, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 27, 240, 114, 1, 0, 43, 21, 96, 56, 226, 114, 1, 0, 43, 21, 176, 48, 239, 114, 1, 0, 43, 21, 32, 13, 228, 114, 1, 0, 43, 21, 240, 91, 237, 114, 3, 0, 0, 0, 148, 203, 151, 3, 208, 239, 241, 2, 90, 143, 103, 3, 8, 0, 209, 6, 178, 21, 102, 8, 187, 8, 29, 22, 179, 10, 85, 12, 60, 21, 170, 170, 101, 198, 0, 32, 103, 198, 0, 32, 105, 198, 0, 32, 109, 198, 0, 32, 111, 198, 0, 32, 113, 198, 0, 32, 115, 198, 0, 32, 117, 198, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 105, 249, 114, 3, 0, 0, 0, 4, 161, 137, 3, 224, 105, 249, 2, 104, 36, 78, 3, 2, 0, 102, 8, 60, 21, 170, 170, 134, 198, 0, 32, 138, 198, 0, 224, 1, 0, 43, 21, 240, 121, 232, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 244, 248, 114, 1, 0, 43, 21, 112, 118, 243, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 217, 233, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 233, 160, 114, 3, 0, 0, 0, 64, 86, 152, 3, 240, 13, 250, 2, 209, 102, 83, 3, 6, 0, 177, 5, 247, 11, 205, 6, 102, 8, 31, 4, 60, 21, 170, 170, 146, 198, 0, 32, 148, 198, 0, 32, 152, 198, 0, 32, 154, 198, 0, 32, 158, 198, 0, 32, 162, 198, 0, 224, 1, 0, 43, 21, 112, 72, 209, 114, 1, 0, 43, 21, 32, 13, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 201, 250, 114, 1, 0, 43, 21, 112, 118, 243, 114, 1, 0, 43, 21, 160, 39, 234, 114, 1, 0, 43, 21, 0, 221, 222, 114, 1, 0, 43, 21, 80, 17, 226, 114, 1, 0, 43, 21, 80, 24, 204, 114, 3, 0, 0, 0, 100, 90, 167, 3, 8, 135, 250, 2, 159, 100, 103, 3, 1, 0, 43, 21, 192, 64, 222, 114, 1, 0, 43, 21, 112, 139, 177, 114, 11, 0, 42, 5, 30, 13, 52, 25, 251, 13, 102, 8, 253, 9, 39, 11, 54, 12, 60, 21, 45, 10, 237, 12, 176, 198, 0, 32, 178, 198, 0, 32, 180, 198, 0, 32, 184, 198, 0, 32, 186, 198, 0, 32, 188, 198, 0, 32, 190, 198, 0, 32, 192, 198, 0, 32, 194, 198, 0, 224, 198, 198, 0, 32, 200, 198, 0, 32, 3, 0, 0, 0, 132, 241, 84, 3, 112, 201, 250, 2, 0, 64, 13, 3, 2, 0, 119, 21, 43, 21, 170, 170, 32, 119, 252, 114, 32, 119, 252, 114, 2, 0, 60, 21, 49, 13, 170, 170, 219, 198, 0, 224, 223, 198, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 223, 249, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 26, 231, 114, 3, 0, 0, 0, 180, 228, 164, 3, 24, 67, 247, 2, 245, 71, 104, 3, 3, 0, 247, 11, 66, 11, 60, 21, 231, 198, 0, 32, 235, 198, 0, 32, 239, 198, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 9, 239, 114, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 240, 144, 249, 114, 3, 0, 0, 0, 164, 27, 141, 3, 112, 201, 250, 2, 28, 203, 73, 3, 3, 0, 108, 10, 251, 13, 60, 21, 248, 198, 0, 32, 252, 198, 0, 32, 0, 199, 0, 224, 1, 0, 43, 21, 160, 233, 160, 114, 1, 0, 43, 21, 160, 233, 160, 114, 1, 0, 43, 21, 160, 99, 224, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 108, 27, 137, 3, 104, 30, 244, 2, 29, 74, 89, 3, 5, 0, 209, 6, 178, 21, 171, 8, 31, 4, 60, 21, 9, 199, 0, 32, 11, 199, 0, 32, 13, 199, 0, 32, 15, 199, 0, 32, 19, 199, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 20, 226, 141, 3, 48, 176, 249, 2, 13, 137, 82, 3, 2, 0, 251, 13, 60, 21, 170, 170, 31, 199, 0, 32, 35, 199, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 118, 243, 114, 3, 0, 0, 0, 28, 190, 169, 3, 152, 242, 249, 2, 182, 127, 97, 3, 1, 0, 43, 21, 224, 105, 249, 114, 1, 0, 43, 21, 32, 13, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 5, 0, 251, 13, 60, 21, 230, 13, 154, 9, 140, 21, 43, 199, 0, 32, 47, 199, 0, 224, 51, 199, 0, 32, 53, 199, 0, 32, 55, 199, 0, 32, 3, 0, 0, 0, 116, 22, 64, 3, 112, 201, 250, 2, 32, 119, 252, 2, 2, 0, 119, 21, 43, 21, 170, 170, 32, 119, 252, 114, 32, 119, 252, 114, 2, 0, 60, 21, 49, 13, 170, 170, 67, 199, 0, 224, 71, 199, 0, 32, 1, 0, 43, 21, 128, 224, 211, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 149, 247, 114, 3, 0, 0, 0, 152, 151, 154, 3, 120, 170, 248, 2, 5, 129, 100, 3, 3, 0, 251, 13, 247, 11, 60, 21, 79, 199, 0, 32, 81, 199, 0, 32, 85, 199, 0, 224, 1, 0, 43, 21, 80, 40, 243, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 201, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 221, 222, 114, 3, 0, 0, 0, 20, 88, 164, 3, 136, 197, 250, 2, 40, 2, 93, 3, 4, 0, 251, 13, 179, 12, 48, 4, 60, 21, 170, 170, 94, 199, 0, 32, 96, 199, 0, 32, 100, 199, 0, 32, 104, 199, 0, 224, 3, 0, 119, 21, 71, 21, 43, 21, 32, 119, 252, 114, 16, 206, 0, 96, 240, 151, 227, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 10, 248, 114, 3, 0, 0, 0, 20, 147, 150, 3, 152, 28, 243, 2, 172, 101, 103, 3, 3, 0, 102, 8, 252, 19, 60, 21, 115, 199, 0, 32, 120, 199, 0, 32, 124, 199, 0, 224, 1, 0, 43, 21, 96, 109, 238, 114, 1, 0, 43, 21, 128, 187, 238, 114, 1, 0, 43, 21, 224, 255, 224, 114, 1, 0, 43, 21, 192, 11, 210, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 1, 0, 43, 21, 240, 68, 220, 114, 1, 0, 43, 21, 128, 254, 206, 114, 1, 0, 43, 21, 176, 175, 197, 114, 3, 0, 0, 0, 128, 52, 167, 3, 112, 201, 250, 2, 42, 124, 101, 3, 9, 0, 30, 13, 251, 13, 102, 8, 179, 12, 232, 7, 100, 13, 184, 10, 24, 19, 60, 21, 133, 199, 0, 32, 135, 199, 0, 32, 137, 199, 0, 32, 139, 199, 0, 32, 141, 199, 0, 32, 145, 199, 0, 32, 147, 199, 0, 32, 149, 199, 0, 32, 151, 199, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 3, 0, 0, 0, 176, 17, 159, 3, 168, 150, 250, 2, 195, 177, 89, 3, 2, 0, 247, 11, 60, 21, 170, 170, 169, 199, 0, 32, 173, 199, 0, 224, 1, 0, 43, 21, 224, 142, 222, 114, 1, 0, 43, 21, 112, 72, 209, 114, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 32, 43, 223, 114, 1, 0, 43, 21, 128, 51, 219, 114, 1, 0, 43, 21, 80, 160, 223, 114, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 240, 181, 222, 114, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 0, 221, 222, 114, 1, 0, 43, 21, 128, 254, 206, 114, 1, 0, 43, 21, 80, 107, 211, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 114, 245, 114, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 96, 222, 240, 114, 1, 0, 43, 21, 192, 11, 210, 114, 1, 0, 43, 21, 160, 76, 207, 114, 1, 0, 43, 21, 144, 37, 207, 114, 1, 0, 43, 21, 16, 177, 215, 114, 2, 0, 119, 21, 43, 21, 170, 170, 192, 200, 241, 114, 240, 8, 230, 114, 1, 0, 43, 21, 80, 77, 216, 114, 1, 0, 43, 21, 64, 241, 203, 114, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 224, 255, 224, 114, 1, 0, 43, 21, 224, 112, 227, 114, 1, 0, 43, 21, 0, 145, 193, 114, 1, 0, 43, 21, 16, 207, 210, 114, 1, 0, 43, 21, 160, 46, 212, 114, 1, 0, 43, 21, 112, 215, 206, 114, 1, 0, 43, 21, 208, 246, 219, 114, 1, 0, 43, 21, 224, 36, 198, 114, 1, 0, 43, 21, 0, 131, 237, 114, 1, 0, 43, 21, 64, 241, 203, 114, 3, 0, 0, 0, 8, 1, 183, 3, 64, 215, 249, 2, 3, 5, 121, 3, 1, 0, 43, 21, 128, 51, 219, 114, 30, 0, 110, 4, 174, 4, 234, 4, 254, 4, 18, 5, 88, 5, 169, 5, 35, 7, 98, 7, 7, 12, 42, 8, 112, 8, 241, 8, 39, 9, 75, 9, 161, 9, 177, 9, 201, 9, 249, 9, 192, 21, 72, 10, 139, 10, 90, 11, 134, 11, 251, 17, 18, 12, 186, 6, 104, 12, 60, 21, 140, 21, 170, 170, 181, 199, 0, 32, 183, 199, 0, 32, 185, 199, 0, 32, 189, 199, 0, 32, 191, 199, 0, 32, 193, 199, 0, 32, 197, 199, 0, 32, 201, 199, 0, 32, 203, 199, 0, 32, 205, 199, 0, 32, 209, 199, 0, 32, 213, 199, 0, 32, 215, 199, 0, 32, 217, 199, 0, 32, 219, 199, 0, 32, 221, 199, 0, 32, 225, 199, 0, 32, 227, 199, 0, 32, 229, 199, 0, 32, 233, 199, 0, 32, 235, 199, 0, 32, 237, 199, 0, 32, 239, 199, 0, 32, 241, 199, 0, 32, 243, 199, 0, 32, 245, 199, 0, 32, 247, 199, 0, 32, 249, 199, 0, 32, 251, 199, 0, 224, 255, 199, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 1, 0, 43, 21, 80, 70, 238, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 9, 239, 114, 3, 0, 0, 0, 84, 45, 168, 3, 112, 201, 250, 2, 181, 57, 94, 3, 4, 0, 251, 13, 102, 8, 38, 13, 60, 21, 170, 170, 47, 200, 0, 32, 51, 200, 0, 32, 53, 200, 0, 32, 57, 200, 0, 224, 1, 0, 43, 21, 112, 42, 214, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 187, 238, 114, 1, 0, 43, 21, 144, 166, 248, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 1, 0, 43, 21, 160, 242, 221, 114, 1, 0, 43, 21, 208, 103, 222, 114, 1, 0, 43, 21, 112, 102, 204, 114, 1, 0, 43, 21, 240, 181, 222, 114, 1, 0, 43, 21, 80, 17, 226, 114, 1, 0, 43, 21, 240, 91, 237, 114, 3, 0, 0, 0, 236, 159, 167, 3, 56, 127, 250, 2, 250, 88, 99, 3, 1, 0, 43, 21, 128, 254, 206, 114, 1, 0, 43, 21, 224, 255, 224, 114, 13, 0, 183, 11, 112, 13, 251, 13, 128, 10, 232, 7, 12, 9, 27, 8, 32, 11, 100, 13, 230, 12, 60, 21, 182, 12, 109, 12, 68, 200, 0, 32, 70, 200, 0, 32, 74, 200, 0, 32, 76, 200, 0, 32, 80, 200, 0, 32, 82, 200, 0, 32, 84, 200, 0, 32, 86, 200, 0, 32, 88, 200, 0, 32, 90, 200, 0, 32, 92, 200, 0, 224, 96, 200, 0, 32, 98, 200, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 82, 223, 114, 3, 0, 0, 0, 224, 161, 142, 3, 112, 201, 250, 2, 246, 157, 69, 3, 3, 0, 251, 13, 145, 7, 60, 21, 120, 200, 0, 32, 124, 200, 0, 32, 128, 200, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 208, 103, 222, 114, 2, 0, 43, 21, 86, 21, 170, 170, 48, 82, 223, 114, 32, 13, 228, 114, 1, 0, 43, 21, 240, 105, 193, 114, 1, 0, 43, 21, 32, 27, 184, 114, 1, 0, 43, 21, 208, 140, 195, 114, 2, 0, 119, 21, 43, 21, 170, 170, 176, 18, 244, 114, 192, 64, 222, 114, 1, 0, 43, 21, 128, 254, 206, 114, 1, 0, 43, 21, 64, 241, 203, 114, 2, 0, 119, 21, 43, 21, 170, 170, 32, 66, 240, 114, 112, 238, 223, 114, 2, 0, 119, 21, 43, 21, 170, 170, 0, 214, 244, 114, 64, 181, 213, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 16, 200, 232, 114, 1, 0, 43, 21, 48, 149, 191, 114, 1, 0, 43, 21, 128, 5, 185, 114, 1, 0, 43, 21, 208, 230, 180, 114, 1, 0, 43, 21, 160, 76, 207, 114, 1, 0, 43, 21, 224, 156, 178, 114, 1, 0, 43, 21, 64, 45, 194, 114, 1, 0, 43, 21, 240, 68, 220, 114, 1, 0, 43, 21, 0, 55, 208, 114, 1, 0, 43, 21, 128, 125, 165, 114, 1, 0, 43, 21, 64, 112, 162, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 148, 238, 114, 1, 0, 43, 21, 112, 72, 209, 114, 1, 0, 43, 21, 192, 11, 210, 114, 1, 0, 43, 21, 64, 211, 208, 114, 1, 0, 43, 21, 208, 253, 197, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 208, 73, 227, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 239, 241, 114, 1, 0, 43, 21, 176, 25, 222, 114, 1, 0, 43, 21, 176, 2, 205, 114, 1, 0, 43, 21, 160, 106, 202, 114, 1, 0, 43, 21, 48, 89, 201, 114, 1, 0, 43, 21, 32, 133, 208, 114, 3, 0, 119, 21, 71, 21, 43, 21, 16, 27, 240, 114, 25, 206, 0, 96, 128, 28, 202, 114, 1, 0, 43, 21, 176, 115, 207, 114, 1, 0, 43, 21, 96, 17, 170, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 80, 17, 226, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 16, 207, 210, 114, 1, 0, 43, 21, 240, 15, 208, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 80, 107, 211, 114, 1, 0, 43, 21, 208, 117, 178, 114, 1, 0, 43, 21, 112, 252, 179, 114, 1, 0, 43, 21, 224, 156, 178, 114, 1, 0, 43, 21, 112, 72, 209, 114, 2, 0, 119, 21, 43, 21, 170, 170, 32, 66, 240, 114, 192, 64, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 192, 64, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 78, 225, 114, 1, 0, 43, 21, 112, 215, 206, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 26, 231, 114, 1, 0, 43, 21, 144, 180, 204, 114, 1, 0, 43, 21, 192, 64, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 64, 181, 213, 114, 1, 0, 43, 21, 240, 105, 193, 114, 1, 0, 43, 21, 144, 37, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 78, 225, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 192, 177, 224, 114, 1, 0, 43, 21, 96, 176, 206, 114, 1, 0, 43, 21, 80, 227, 191, 114, 2, 0, 119, 21, 43, 21, 170, 170, 128, 187, 238, 114, 112, 185, 211, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 122, 176, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 48, 142, 213, 114, 1, 0, 43, 21, 80, 84, 194, 114, 1, 0, 43, 21, 224, 232, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 33, 209, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 32, 223, 193, 114, 1, 0, 43, 21, 176, 62, 195, 114, 1, 0, 43, 21, 128, 111, 209, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 176, 108, 229, 114, 1, 0, 43, 21, 176, 115, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 128, 104, 231, 114, 3, 0, 0, 0, 72, 128, 192, 3, 96, 67, 245, 2, 83, 63, 134, 3, 1, 0, 43, 21, 96, 63, 204, 114, 1, 0, 43, 21, 128, 58, 197, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 32, 13, 228, 114, 1, 0, 43, 21, 224, 232, 207, 114, 1, 0, 43, 21, 176, 115, 207, 114, 2, 0, 119, 21, 43, 21, 170, 170, 208, 126, 239, 114, 160, 46, 212, 114, 1, 0, 43, 21, 192, 154, 207, 114, 78, 0, 116, 13, 226, 4, 30, 5, 34, 5, 38, 5, 45, 5, 57, 5, 61, 5, 69, 5, 84, 5, 112, 5, 115, 5, 118, 5, 130, 5, 138, 5, 157, 5, 213, 5, 83, 6, 115, 6, 143, 6, 154, 6, 251, 13, 43, 7, 90, 7, 94, 7, 118, 7, 126, 7, 255, 11, 196, 7, 212, 7, 216, 7, 101, 8, 105, 8, 120, 8, 128, 8, 136, 8, 168, 8, 179, 8, 211, 8, 215, 8, 20, 9, 47, 9, 59, 9, 87, 9, 169, 9, 173, 9, 27, 8, 8, 10, 175, 13, 44, 10, 68, 10, 197, 7, 147, 10, 163, 10, 160, 18, 233, 10, 74, 11, 98, 11, 110, 11, 195, 13, 190, 11, 198, 11, 206, 11, 238, 12, 56, 22, 69, 12, 100, 12, 154, 14, 136, 12, 152, 12, 60, 21, 53, 13, 57, 13, 138, 7, 123, 13, 127, 13, 135, 13, 182, 13, 170, 170, 137, 200, 0, 32, 141, 200, 0, 32, 145, 200, 0, 32, 147, 200, 0, 32, 149, 200, 0, 32, 151, 200, 0, 32, 155, 200, 0, 32, 157, 200, 0, 32, 159, 200, 0, 32, 163, 200, 0, 32, 167, 200, 0, 32, 171, 200, 0, 32, 173, 200, 0, 32, 175, 200, 0, 32, 177, 200, 0, 32, 179, 200, 0, 32, 181, 200, 0, 32, 183, 200, 0, 32, 185, 200, 0, 32, 187, 200, 0, 32, 189, 200, 0, 32, 191, 200, 0, 32, 195, 200, 0, 32, 197, 200, 0, 32, 199, 200, 0, 32, 201, 200, 0, 32, 203, 200, 0, 32, 207, 200, 0, 32, 211, 200, 0, 32, 213, 200, 0, 32, 215, 200, 0, 32, 217, 200, 0, 32, 219, 200, 0, 32, 221, 200, 0, 32, 226, 200, 0, 32, 228, 200, 0, 32, 230, 200, 0, 32, 234, 200, 0, 32, 238, 200, 0, 32, 240, 200, 0, 32, 244, 200, 0, 32, 246, 200, 0, 32, 248, 200, 0, 32, 250, 200, 0, 32, 252, 200, 0, 32, 0, 201, 0, 32, 4, 201, 0, 32, 8, 201, 0, 32, 10, 201, 0, 32, 14, 201, 0, 32, 16, 201, 0, 32, 18, 201, 0, 32, 22, 201, 0, 32, 24, 201, 0, 32, 26, 201, 0, 32, 30, 201, 0, 32, 34, 201, 0, 32, 36, 201, 0, 32, 38, 201, 0, 32, 42, 201, 0, 32, 46, 201, 0, 32, 50, 201, 0, 32, 52, 201, 0, 32, 54, 201, 0, 32, 58, 201, 0, 32, 62, 201, 0, 32, 64, 201, 0, 32, 66, 201, 0, 32, 70, 201, 0, 32, 72, 201, 0, 32, 76, 201, 0, 224, 80, 201, 0, 32, 82, 201, 0, 32, 84, 201, 0, 32, 88, 201, 0, 32, 90, 201, 0, 32, 92, 201, 0, 32, 96, 201, 0, 32, 3, 0, 119, 21, 71, 21, 43, 21, 32, 119, 252, 114, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 144, 104, 119, 3, 112, 201, 250, 2, 112, 14, 46, 3, 2, 0, 251, 13, 60, 21, 170, 170, 216, 201, 0, 32, 221, 201, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 14, 246, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 32, 43, 223, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 128, 187, 238, 114, 1, 0, 43, 21, 112, 5, 241, 114, 1, 0, 43, 21, 48, 112, 218, 114, 1, 0, 43, 21, 144, 120, 214, 114, 1, 0, 43, 21, 224, 202, 212, 114, 3, 0, 0, 0, 148, 177, 172, 3, 168, 168, 247, 2, 86, 226, 107, 3, 8, 0, 112, 13, 67, 22, 13, 6, 251, 13, 222, 9, 103, 9, 116, 12, 60, 21, 170, 170, 229, 201, 0, 32, 233, 201, 0, 32, 237, 201, 0, 32, 241, 201, 0, 32, 243, 201, 0, 32, 245, 201, 0, 32, 247, 201, 0, 32, 249, 201, 0, 224, 1, 0, 43, 21, 64, 121, 223, 114, 1, 0, 43, 21, 192, 87, 239, 114, 1, 0, 43, 21, 64, 121, 223, 114, 2, 0, 43, 21, 86, 21, 170, 170, 64, 181, 213, 114, 32, 13, 228, 114, 1, 0, 43, 21, 224, 142, 222, 114, 1, 0, 43, 21, 112, 95, 226, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 31, 247, 114, 1, 0, 43, 21, 64, 75, 189, 114, 1, 0, 43, 21, 224, 112, 227, 114, 1, 0, 43, 21, 64, 211, 208, 114, 1, 0, 43, 21, 240, 188, 200, 114, 1, 0, 43, 21, 128, 35, 180, 114, 1, 0, 43, 21, 128, 194, 216, 114, 2, 0, 119, 21, 43, 21, 170, 170, 224, 52, 237, 114, 80, 160, 223, 114, 1, 0, 43, 21, 192, 64, 222, 114, 1, 0, 43, 21, 32, 13, 228, 114, 1, 0, 43, 21, 224, 202, 212, 114, 1, 0, 43, 21, 192, 154, 207, 114, 2, 0, 119, 21, 43, 21, 170, 170, 160, 242, 221, 114, 48, 82, 223, 114, 1, 0, 43, 21, 80, 17, 226, 114, 3, 0, 0, 0, 32, 140, 180, 3, 144, 166, 248, 2, 103, 107, 114, 3, 1, 0, 43, 21, 192, 177, 224, 114, 22, 0, 112, 13, 67, 22, 230, 4, 49, 5, 122, 5, 13, 6, 222, 9, 47, 7, 78, 7, 250, 5, 174, 6, 100, 21, 79, 9, 103, 9, 142, 9, 95, 10, 47, 11, 26, 6, 122, 11, 222, 11, 60, 21, 230, 10, 170, 170, 10, 202, 0, 32, 12, 202, 0, 32, 14, 202, 0, 32, 16, 202, 0, 32, 20, 202, 0, 32, 22, 202, 0, 32, 24, 202, 0, 32, 28, 202, 0, 32, 30, 202, 0, 32, 32, 202, 0, 32, 34, 202, 0, 32, 36, 202, 0, 32, 38, 202, 0, 32, 40, 202, 0, 32, 44, 202, 0, 32, 46, 202, 0, 32, 48, 202, 0, 32, 50, 202, 0, 32, 52, 202, 0, 32, 56, 202, 0, 32, 58, 202, 0, 224, 62, 202, 0, 32, 1, 0, 43, 21, 208, 20, 215, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 40, 77, 150, 3, 112, 201, 250, 2, 158, 121, 76, 3, 3, 0, 9, 17, 68, 9, 60, 21, 98, 202, 0, 32, 100, 202, 0, 32, 104, 202, 0, 224, 1, 0, 43, 21, 176, 62, 195, 114, 1, 0, 43, 21, 0, 221, 222, 114, 1, 0, 43, 21, 160, 136, 197, 114, 1, 0, 43, 21, 176, 62, 195, 114, 1, 0, 43, 21, 96, 222, 240, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 240, 8, 230, 114, 2, 0, 43, 21, 86, 21, 170, 170, 64, 15, 199, 114, 32, 13, 228, 114, 1, 0, 43, 21, 192, 48, 183, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 1, 0, 43, 21, 32, 73, 218, 114, 1, 0, 43, 21, 96, 123, 194, 114, 2, 0, 43, 21, 86, 21, 170, 170, 208, 246, 219, 114, 32, 13, 228, 114, 1, 0, 43, 21, 112, 192, 189, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 16, 4, 223, 114, 2, 0, 43, 21, 86, 21, 170, 170, 224, 172, 217, 114, 32, 13, 228, 114, 1, 0, 43, 21, 208, 193, 207, 114, 1, 0, 43, 21, 208, 193, 207, 114, 3, 0, 0, 0, 148, 148, 181, 3, 112, 201, 250, 2, 195, 63, 111, 3, 1, 0, 43, 21, 208, 103, 222, 114, 20, 0, 74, 8, 30, 13, 165, 6, 52, 25, 251, 13, 102, 8, 11, 7, 179, 12, 48, 4, 63, 9, 91, 9, 103, 10, 35, 11, 94, 11, 226, 10, 15, 18, 218, 11, 24, 19, 60, 21, 68, 13, 170, 170, 113, 202, 0, 32, 115, 202, 0, 32, 117, 202, 0, 32, 119, 202, 0, 32, 121, 202, 0, 32, 123, 202, 0, 32, 127, 202, 0, 32, 131, 202, 0, 32, 133, 202, 0, 32, 137, 202, 0, 32, 139, 202, 0, 32, 141, 202, 0, 32, 145, 202, 0, 32, 147, 202, 0, 32, 149, 202, 0, 32, 151, 202, 0, 32, 155, 202, 0, 32, 157, 202, 0, 32, 159, 202, 0, 224, 163, 202, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 3, 0, 0, 0, 36, 196, 140, 3, 112, 201, 250, 2, 52, 44, 71, 3, 2, 0, 251, 13, 60, 21, 170, 170, 196, 202, 0, 32, 200, 202, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 1, 0, 43, 21, 48, 45, 250, 114, 3, 0, 0, 0, 248, 128, 151, 3, 176, 244, 248, 2, 121, 200, 90, 3, 3, 0, 251, 13, 14, 8, 60, 21, 208, 202, 0, 32, 212, 202, 0, 32, 214, 202, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 1, 0, 43, 21, 16, 140, 242, 114, 3, 0, 0, 0, 144, 27, 161, 3, 88, 80, 250, 2, 200, 88, 103, 3, 3, 0, 112, 13, 251, 13, 60, 21, 223, 202, 0, 32, 227, 202, 0, 32, 229, 202, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 1, 0, 43, 21, 0, 25, 213, 114, 2, 0, 43, 21, 86, 21, 170, 170, 144, 233, 216, 114, 32, 13, 228, 114, 3, 0, 0, 0, 172, 89, 186, 3, 112, 201, 250, 2, 88, 239, 118, 3, 4, 0, 2, 6, 29, 11, 166, 11, 60, 21, 170, 170, 238, 202, 0, 32, 242, 202, 0, 32, 244, 202, 0, 32, 248, 202, 0, 224, 1, 0, 43, 21, 96, 236, 196, 114, 2, 0, 43, 21, 86, 21, 170, 170, 224, 172, 217, 114, 32, 13, 228, 114, 2, 0, 119, 21, 43, 21, 170, 170, 160, 242, 221, 114, 240, 181, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 148, 238, 114, 1, 0, 43, 21, 32, 110, 191, 114, 1, 0, 43, 21, 16, 230, 227, 114, 1, 0, 43, 21, 192, 4, 232, 114, 1, 0, 43, 21, 80, 70, 238, 114, 1, 0, 43, 21, 192, 4, 232, 114, 1, 0, 43, 21, 32, 96, 235, 114, 1, 0, 43, 21, 240, 91, 237, 114, 2, 0, 119, 21, 43, 21, 170, 170, 96, 79, 243, 114, 0, 221, 222, 114, 1, 0, 43, 21, 128, 134, 226, 114, 1, 0, 43, 21, 176, 85, 212, 114, 1, 0, 43, 21, 208, 27, 193, 114, 2, 0, 119, 21, 43, 21, 170, 170, 160, 242, 221, 114, 16, 177, 215, 114, 2, 0, 119, 21, 43, 21, 170, 170, 160, 242, 221, 114, 160, 46, 212, 114, 1, 0, 43, 21, 208, 103, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 192, 245, 114, 1, 0, 43, 21, 96, 116, 216, 114, 3, 0, 0, 0, 180, 72, 165, 3, 80, 4, 249, 2, 199, 34, 109, 3, 21, 0, 112, 13, 79, 6, 157, 6, 251, 13, 126, 7, 141, 7, 61, 8, 80, 9, 160, 8, 126, 9, 138, 9, 185, 9, 213, 9, 143, 11, 233, 10, 28, 11, 186, 11, 188, 10, 123, 7, 155, 12, 60, 21, 3, 203, 0, 32, 5, 203, 0, 32, 9, 203, 0, 32, 13, 203, 0, 32, 17, 203, 0, 32, 19, 203, 0, 32, 21, 203, 0, 32, 23, 203, 0, 32, 25, 203, 0, 32, 27, 203, 0, 32, 29, 203, 0, 32, 31, 203, 0, 32, 35, 203, 0, 32, 37, 203, 0, 32, 39, 203, 0, 32, 41, 203, 0, 32, 45, 203, 0, 32, 49, 203, 0, 32, 51, 203, 0, 32, 55, 203, 0, 32, 57, 203, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 1, 243, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 44, 241, 114, 3, 0, 0, 0, 76, 15, 151, 3, 64, 209, 250, 2, 226, 106, 95, 3, 3, 0, 178, 6, 230, 12, 60, 21, 93, 203, 0, 32, 97, 203, 0, 32, 101, 203, 0, 224, 1, 0, 43, 21, 0, 221, 222, 114, 1, 0, 43, 21, 112, 215, 206, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 66, 249, 114, 3, 0, 0, 0, 52, 83, 157, 3, 248, 244, 246, 2, 207, 66, 104, 3, 4, 0, 1, 6, 89, 8, 113, 20, 60, 21, 170, 170, 110, 203, 0, 32, 112, 203, 0, 32, 114, 203, 0, 32, 118, 203, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 214, 244, 114, 3, 0, 0, 0, 56, 121, 120, 3, 224, 105, 249, 2, 247, 230, 72, 3, 3, 0, 251, 13, 62, 7, 60, 21, 129, 203, 0, 32, 133, 203, 0, 32, 137, 203, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 192, 245, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 192, 57, 244, 114, 3, 0, 0, 0, 252, 71, 134, 3, 120, 51, 247, 2, 108, 195, 83, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 222, 240, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 83, 241, 114, 5, 0, 112, 13, 102, 8, 60, 21, 131, 13, 242, 13, 146, 203, 0, 32, 150, 203, 0, 32, 154, 203, 0, 224, 158, 203, 0, 32, 162, 203, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 3, 0, 0, 0, 56, 162, 134, 3, 144, 154, 250, 2, 124, 19, 64, 3, 2, 0, 251, 13, 60, 21, 170, 170, 174, 203, 0, 32, 178, 203, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 192, 27, 249, 114, 3, 0, 0, 0, 128, 161, 153, 3, 32, 119, 252, 2, 179, 159, 105, 3, 2, 0, 29, 11, 60, 21, 170, 170, 186, 203, 0, 32, 190, 203, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 108, 38, 181, 3, 120, 158, 250, 2, 197, 152, 109, 3, 2, 0, 29, 11, 60, 21, 170, 170, 198, 203, 0, 32, 202, 203, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 228, 52, 167, 3, 56, 2, 250, 2, 63, 241, 90, 3, 2, 0, 112, 13, 60, 21, 170, 170, 210, 203, 0, 32, 214, 203, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 3, 0, 0, 0, 12, 24, 136, 3, 136, 197, 250, 2, 34, 18, 65, 3, 2, 0, 251, 13, 60, 21, 170, 170, 222, 203, 0, 32, 226, 203, 0, 224, 1, 0, 43, 21, 0, 48, 230, 114, 1, 0, 43, 21, 48, 248, 237, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 114, 245, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 192, 170, 246, 114, 3, 0, 0, 0, 200, 22, 170, 3, 200, 228, 250, 2, 117, 156, 104, 3, 1, 0, 43, 21, 64, 121, 223, 114, 6, 0, 30, 13, 251, 13, 143, 4, 230, 12, 60, 21, 78, 22, 170, 170, 234, 203, 0, 32, 236, 203, 0, 32, 238, 203, 0, 32, 242, 203, 0, 32, 246, 203, 0, 224, 250, 203, 0, 32, 1, 0, 43, 21, 192, 207, 219, 114, 1, 0, 43, 21, 160, 69, 229, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 53, 246, 114, 3, 0, 0, 0, 88, 14, 155, 3, 24, 198, 246, 2, 217, 135, 97, 3, 4, 0, 89, 8, 148, 8, 3, 8, 60, 21, 170, 170, 6, 204, 0, 32, 8, 204, 0, 32, 10, 204, 0, 32, 14, 204, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 205, 248, 114, 1, 0, 43, 21, 192, 200, 241, 114, 1, 0, 43, 21, 16, 207, 210, 114, 1, 0, 43, 21, 64, 91, 228, 114, 1, 0, 43, 21, 16, 4, 223, 114, 3, 0, 0, 0, 68, 254, 160, 3, 64, 90, 249, 2, 216, 204, 95, 3, 6, 0, 112, 13, 251, 13, 102, 8, 250, 5, 86, 22, 60, 21, 170, 170, 25, 204, 0, 32, 29, 204, 0, 32, 31, 204, 0, 32, 33, 204, 0, 32, 35, 204, 0, 32, 37, 204, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 105, 249, 114, 3, 0, 0, 0, 184, 37, 136, 3, 200, 109, 249, 2, 70, 194, 73, 3, 2, 0, 251, 13, 60, 21, 170, 170, 51, 204, 0, 32, 55, 204, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 71, 21, 43, 21, 86, 21, 7, 206, 0, 96, 144, 166, 248, 114, 32, 13, 228, 114, 3, 0, 0, 0, 72, 221, 139, 3, 32, 119, 252, 2, 62, 219, 61, 3, 1, 0, 43, 21, 32, 156, 225, 114, 4, 0, 30, 13, 122, 7, 60, 21, 107, 13, 170, 170, 63, 204, 0, 32, 67, 204, 0, 32, 72, 204, 0, 224, 76, 204, 0, 32, 1, 0, 43, 21, 224, 52, 237, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 14, 246, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 248, 237, 114, 3, 0, 0, 0, 152, 254, 166, 3, 192, 152, 249, 2, 38, 50, 105, 3, 4, 0, 251, 13, 246, 12, 154, 14, 60, 21, 170, 170, 85, 204, 0, 32, 87, 204, 0, 32, 91, 204, 0, 32, 95, 204, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 88, 248, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 32, 209, 237, 114, 1, 0, 43, 21, 192, 124, 212, 114, 3, 0, 0, 0, 64, 152, 141, 3, 64, 245, 244, 2, 209, 212, 93, 3, 1, 0, 43, 21, 160, 99, 224, 114, 1, 0, 43, 21, 0, 0, 0, 112, 8, 0, 251, 13, 209, 6, 178, 21, 187, 8, 209, 9, 60, 21, 61, 13, 94, 22, 170, 170, 106, 204, 0, 32, 110, 204, 0, 32, 112, 204, 0, 32, 114, 204, 0, 32, 116, 204, 0, 32, 118, 204, 0, 224, 122, 204, 0, 32, 124, 204, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 204, 239, 114, 1, 0, 43, 21, 128, 21, 224, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 3, 0, 0, 0, 192, 108, 142, 3, 64, 90, 249, 2, 150, 124, 89, 3, 1, 0, 43, 21, 240, 68, 220, 114, 1, 0, 43, 21, 32, 209, 237, 114, 6, 0, 251, 13, 150, 5, 51, 0, 60, 21, 155, 13, 6, 14, 170, 170, 139, 204, 0, 32, 143, 204, 0, 32, 145, 204, 0, 32, 149, 204, 0, 224, 153, 204, 0, 32, 155, 204, 0, 32, 1, 0, 43, 21, 32, 209, 237, 114, 1, 0, 43, 21, 160, 122, 241, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 205, 248, 114, 1, 0, 43, 21, 64, 227, 247, 114, 1, 0, 43, 21, 0, 0, 0, 112, 3, 0, 0, 0, 12, 124, 161, 3, 200, 228, 250, 2, 32, 193, 90, 3, 6, 0, 30, 13, 251, 13, 170, 23, 230, 12, 250, 11, 60, 21, 170, 170, 167, 204, 0, 32, 169, 204, 0, 32, 171, 204, 0, 32, 175, 204, 0, 32, 177, 204, 0, 32, 179, 204, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 75, 245, 114, 1, 0, 43, 21, 192, 57, 244, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 244, 248, 114, 3, 0, 71, 21, 43, 21, 86, 21, 16, 206, 0, 96, 112, 231, 245, 114, 32, 13, 228, 114, 1, 0, 43, 21, 64, 31, 238, 114, 3, 0, 0, 0, 44, 1, 157, 3, 32, 119, 252, 2, 4, 122, 80, 3, 6, 0, 30, 13, 251, 13, 102, 8, 32, 9, 31, 4, 60, 21, 170, 170, 193, 204, 0, 32, 197, 204, 0, 32, 199, 204, 0, 32, 203, 204, 0, 32, 208, 204, 0, 32, 210, 204, 0, 224, 1, 0, 43, 21, 32, 179, 242, 114, 1, 0, 43, 21, 144, 0, 234, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 253, 244, 114, 1, 0, 43, 21, 160, 122, 241, 114, 3, 0, 0, 0, 40, 197, 155, 3, 176, 232, 250, 2, 100, 109, 89, 3, 5, 0, 251, 13, 107, 9, 146, 9, 230, 12, 60, 21, 224, 204, 0, 32, 226, 204, 0, 32, 228, 204, 0, 32, 232, 204, 0, 32, 234, 204, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 248, 246, 114, 3, 0, 0, 0, 180, 251, 156, 3, 88, 86, 249, 2, 123, 204, 96, 3, 2, 0, 112, 13, 60, 21, 170, 170, 246, 204, 0, 32, 250, 204, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 36, 245, 114, 1, 0, 43, 21, 176, 244, 248, 114, 1, 0, 43, 21, 32, 209, 237, 114, 1, 0, 43, 21, 144, 210, 199, 114, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 128, 187, 238, 114, 2, 0, 43, 21, 86, 21, 170, 170, 16, 230, 227, 114, 32, 13, 228, 114, 2, 0, 43, 21, 86, 21, 170, 170, 16, 230, 227, 114, 32, 13, 228, 114, 1, 0, 43, 21, 112, 178, 233, 114, 1, 0, 43, 21, 112, 178, 233, 114, 3, 0, 0, 0, 168, 87, 167, 3, 88, 235, 245, 2, 165, 252, 106, 3, 3, 0, 119, 21, 71, 21, 43, 21, 208, 126, 239, 114, 16, 206, 0, 96, 32, 96, 235, 114, 1, 0, 43, 21, 160, 9, 239, 114, 12, 0, 112, 13, 68, 4, 251, 13, 247, 11, 102, 8, 102, 11, 103, 22, 254, 11, 112, 22, 60, 21, 25, 13, 254, 13, 170, 170, 2, 205, 0, 32, 6, 205, 0, 32, 8, 205, 0, 32, 10, 205, 0, 32, 12, 205, 0, 32, 16, 205, 0, 32, 20, 205, 0, 32, 24, 205, 0, 32, 26, 205, 0, 32, 28, 205, 0, 224, 32, 205, 0, 32, 37, 205, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 201, 250, 114, 3, 0, 0, 0, 160, 249, 143, 3, 112, 201, 250, 2, 198, 169, 60, 3, 2, 0, 102, 8, 60, 21, 170, 170, 58, 205, 0, 32, 62, 205, 0, 224, 1, 0, 43, 21, 224, 195, 234, 114, 1, 0, 43, 21, 16, 117, 225, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 75, 245, 114, 1, 0, 43, 21, 224, 255, 224, 114, 3, 0, 0, 0, 144, 37, 151, 3, 112, 201, 250, 2, 220, 188, 91, 3, 1, 0, 43, 21, 32, 209, 237, 114, 6, 0, 42, 5, 27, 7, 100, 13, 230, 12, 60, 21, 82, 12, 170, 170, 70, 205, 0, 32, 72, 205, 0, 32, 74, 205, 0, 32, 78, 205, 0, 32, 80, 205, 0, 224, 84, 205, 0, 32, 1, 0, 43, 21, 240, 121, 232, 114, 1, 0, 43, 21, 32, 13, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 208, 55, 149, 3, 232, 181, 250, 2, 41, 155, 80, 3, 4, 0, 54, 12, 199, 13, 170, 21, 60, 21, 170, 170, 96, 205, 0, 32, 98, 205, 0, 32, 100, 205, 0, 32, 104, 205, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 220, 147, 125, 3, 112, 201, 250, 2, 152, 230, 60, 3, 2, 0, 102, 8, 60, 21, 170, 170, 115, 205, 0, 32, 119, 205, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 109, 238, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 53, 246, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 105, 249, 114, 3, 0, 0, 0, 164, 155, 153, 3, 200, 133, 245, 2, 132, 171, 105, 3, 4, 0, 251, 13, 102, 8, 152, 10, 60, 21, 170, 170, 127, 205, 0, 32, 131, 205, 0, 32, 135, 205, 0, 32, 139, 205, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 223, 249, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 209, 246, 114, 3, 0, 0, 0, 32, 0, 120, 3, 104, 250, 249, 2, 232, 133, 67, 3, 3, 0, 251, 13, 214, 12, 60, 21, 150, 205, 0, 32, 154, 205, 0, 32, 158, 205, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 231, 245, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 208, 126, 239, 114, 3, 0, 0, 0, 236, 49, 152, 3, 80, 123, 250, 2, 168, 171, 89, 3, 1, 0, 43, 21, 48, 195, 225, 114, 4, 0, 253, 9, 54, 12, 60, 21, 45, 10, 170, 170, 167, 205, 0, 32, 171, 205, 0, 32, 175, 205, 0, 224, 179, 205, 0, 32, 1, 0, 43, 21, 96, 229, 218, 114, 1, 0, 43, 21, 32, 179, 242, 114, 1, 0, 43, 21, 64, 121, 223, 114, 1, 0, 43, 21, 32, 73, 218, 114, 1, 0, 43, 21, 16, 34, 218, 114, 1, 0, 43, 21, 176, 25, 222, 114, 1, 0, 43, 21, 208, 43, 232, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 179, 242, 114, 1, 0, 43, 21, 160, 242, 221, 114, 1, 0, 43, 21, 16, 4, 223, 114, 1, 0, 43, 21, 32, 13, 228, 114, 1, 0, 43, 21, 32, 156, 225, 114, 1, 0, 43, 21, 176, 108, 229, 114, 3, 0, 0, 0, 40, 209, 153, 3, 240, 198, 240, 2, 43, 144, 104, 3, 1, 0, 43, 21, 80, 160, 223, 114, 15, 0, 112, 13, 44, 7, 124, 22, 108, 5, 189, 5, 123, 6, 212, 6, 102, 8, 65, 8, 132, 8, 60, 10, 246, 11, 46, 12, 60, 21, 213, 12, 188, 205, 0, 32, 190, 205, 0, 32, 192, 205, 0, 32, 194, 205, 0, 32, 196, 205, 0, 32, 198, 205, 0, 32, 200, 205, 0, 32, 202, 205, 0, 32, 206, 205, 0, 32, 208, 205, 0, 32, 210, 205, 0, 32, 212, 205, 0, 32, 214, 205, 0, 32, 216, 205, 0, 224, 220, 205, 0, 32, 1, 0, 43, 21, 16, 4, 223, 114, 1, 0, 43, 21, 0, 62, 186, 114, 1, 0, 43, 21, 240, 181, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 114, 245, 114, 1, 0, 43, 21, 0, 48, 230, 114, 3, 0, 0, 0, 136, 209, 167, 3, 88, 211, 249, 2, 1, 61, 110, 3, 6, 0, 53, 8, 136, 8, 16, 10, 80, 10, 2, 12, 60, 21, 170, 170, 245, 205, 0, 32, 247, 205, 0, 32, 249, 205, 0, 32, 251, 205, 0, 32, 255, 205, 0, 32, 1, 206, 0, 224, 1, 0, 43, 21, 128, 104, 231, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 223, 249, 114, 1, 0, 43, 21, 144, 37, 207, 114, 3, 0, 0, 0, 244, 150, 153, 3, 240, 138, 250, 2, 71, 58, 91, 3, 1, 0, 43, 21, 128, 201, 194, 114, 1, 0, 43, 21, 224, 142, 222, 114, 6, 0, 100, 21, 171, 13, 230, 12, 60, 21, 78, 22, 154, 9, 170, 170, 15, 206, 0, 32, 17, 206, 0, 32, 21, 206, 0, 32, 23, 206, 0, 224, 27, 206, 0, 32, 29, 206, 0, 32, 1, 0, 43, 21, 112, 238, 223, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 13, 228, 114, 3, 0, 0, 0, 120, 124, 158, 3, 160, 68, 250, 2, 68, 124, 80, 3, 1, 0, 43, 21, 32, 13, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 5, 0, 251, 13, 31, 4, 60, 21, 154, 9, 140, 21, 41, 206, 0, 32, 43, 206, 0, 32, 47, 206, 0, 224, 51, 206, 0, 32, 53, 206, 0, 32, 1, 0, 43, 21, 96, 109, 238, 114, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 80, 123, 250, 114, 3, 0, 0, 0, 232, 41, 134, 3, 80, 123, 250, 2, 132, 232, 63, 3, 3, 0, 108, 10, 251, 13, 60, 21, 65, 206, 0, 32, 67, 206, 0, 32, 71, 206, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 3, 0, 0, 0, 52, 154, 141, 3, 96, 162, 250, 2, 99, 227, 77, 3, 2, 0, 102, 8, 60, 21, 170, 170, 80, 206, 0, 32, 84, 206, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 166, 248, 114, 1, 0, 43, 21, 144, 30, 229, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 80, 70, 238, 114, 3, 0, 0, 0, 0, 42, 150, 3, 80, 159, 244, 2, 43, 90, 92, 3, 1, 0, 43, 21, 96, 33, 209, 114, 6, 0, 112, 13, 209, 6, 178, 21, 102, 8, 60, 21, 198, 6, 170, 170, 92, 206, 0, 32, 96, 206, 0, 32, 98, 206, 0, 32, 100, 206, 0, 32, 102, 206, 0, 224, 106, 206, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 192, 245, 114, 3, 0, 0, 0, 104, 144, 119, 3, 80, 123, 250, 2, 108, 211, 48, 3, 2, 0, 251, 13, 60, 21, 170, 170, 118, 206, 0, 32, 122, 206, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 192, 27, 249, 114, 1, 0, 43, 21, 240, 91, 237, 114, 1, 0, 43, 21, 192, 57, 244, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 88, 119, 150, 3, 160, 199, 249, 2, 211, 21, 78, 3, 5, 0, 251, 13, 102, 8, 48, 4, 123, 11, 60, 21, 130, 206, 0, 32, 134, 206, 0, 32, 136, 206, 0, 32, 138, 206, 0, 32, 142, 206, 0, 224, 1, 0, 43, 21, 240, 204, 239, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 192, 170, 246, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 255, 224, 114, 1, 0, 43, 21, 224, 105, 249, 114, 1, 0, 43, 21, 144, 60, 224, 114, 3, 0, 0, 0, 76, 215, 151, 3, 0, 59, 249, 2, 236, 146, 88, 3, 1, 0, 43, 21, 64, 91, 228, 114, 7, 0, 69, 5, 251, 13, 102, 8, 225, 9, 154, 14, 60, 21, 138, 7, 154, 206, 0, 32, 156, 206, 0, 32, 160, 206, 0, 32, 164, 206, 0, 32, 166, 206, 0, 32, 168, 206, 0, 224, 172, 206, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 6, 250, 114, 3, 0, 0, 0, 116, 204, 142, 3, 0, 178, 250, 2, 217, 56, 77, 3, 2, 0, 143, 6, 60, 21, 170, 170, 185, 206, 0, 32, 189, 206, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 75, 245, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 75, 245, 114, 3, 0, 0, 0, 200, 242, 150, 3, 32, 24, 247, 2, 90, 199, 104, 3, 1, 0, 43, 21, 224, 29, 220, 114, 1, 0, 43, 21, 64, 61, 233, 114, 1, 0, 43, 21, 208, 50, 210, 114, 6, 0, 251, 13, 202, 10, 60, 21, 221, 12, 9, 13, 6, 14, 170, 170, 197, 206, 0, 32, 201, 206, 0, 32, 205, 206, 0, 224, 209, 206, 0, 32, 211, 206, 0, 32, 213, 206, 0, 32, 1, 0, 43, 21, 16, 170, 237, 114, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 112, 88, 248, 114, 1, 0, 43, 21, 224, 89, 210, 114, 1, 0, 43, 21, 240, 15, 208, 114, 1, 0, 43, 21, 48, 29, 211, 114, 1, 0, 43, 21, 48, 29, 211, 114, 1, 0, 43, 21, 96, 17, 170, 114, 3, 0, 0, 0, 156, 205, 181, 3, 152, 242, 249, 2, 135, 238, 118, 3, 1, 0, 43, 21, 240, 98, 215, 114, 9, 0, 251, 13, 247, 11, 189, 9, 107, 10, 127, 10, 131, 10, 238, 11, 60, 21, 226, 13, 225, 206, 0, 32, 227, 206, 0, 32, 231, 206, 0, 32, 233, 206, 0, 32, 235, 206, 0, 32, 237, 206, 0, 32, 239, 206, 0, 32, 241, 206, 0, 224, 245, 206, 0, 32, 1, 0, 43, 21, 128, 88, 192, 114, 1, 0, 43, 21, 112, 245, 201, 114, 1, 0, 43, 21, 64, 181, 213, 114, 1, 0, 43, 21, 144, 226, 238, 114, 1, 0, 43, 21, 144, 60, 224, 114, 1, 0, 43, 21, 192, 64, 222, 114, 1, 0, 43, 21, 176, 115, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 31, 247, 114, 1, 0, 43, 21, 160, 212, 226, 114, 3, 0, 0, 0, 244, 6, 177, 3, 248, 226, 249, 2, 209, 154, 106, 3, 1, 0, 43, 21, 80, 70, 238, 114, 1, 0, 43, 21, 192, 64, 222, 114, 12, 0, 88, 5, 169, 5, 127, 6, 251, 13, 242, 7, 42, 8, 27, 8, 39, 12, 154, 14, 60, 21, 140, 21, 2, 14, 170, 170, 5, 207, 0, 32, 7, 207, 0, 32, 9, 207, 0, 32, 11, 207, 0, 32, 13, 207, 0, 32, 15, 207, 0, 32, 17, 207, 0, 32, 19, 207, 0, 32, 23, 207, 0, 32, 25, 207, 0, 224, 29, 207, 0, 32, 31, 207, 0, 32, 1, 0, 43, 21, 208, 73, 227, 114, 1, 0, 43, 21, 48, 135, 235, 114, 1, 0, 43, 21, 64, 204, 230, 114, 1, 0, 43, 21, 160, 99, 224, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 204, 239, 114, 1, 0, 43, 21, 32, 156, 225, 114, 1, 0, 43, 21, 224, 112, 227, 114, 1, 0, 43, 21, 160, 23, 195, 114, 3, 0, 0, 0, 212, 47, 153, 3, 168, 61, 244, 2, 18, 232, 105, 3, 1, 0, 43, 21, 240, 121, 232, 114, 1, 0, 43, 21, 16, 170, 237, 114, 1, 0, 43, 21, 128, 21, 224, 114, 1, 0, 43, 21, 96, 183, 184, 114, 13, 0, 233, 9, 111, 10, 123, 10, 202, 10, 31, 4, 130, 11, 234, 11, 123, 7, 60, 21, 2, 13, 91, 13, 206, 13, 6, 14, 52, 207, 0, 32, 54, 207, 0, 32, 56, 207, 0, 32, 58, 207, 0, 32, 60, 207, 0, 32, 64, 207, 0, 32, 66, 207, 0, 32, 68, 207, 0, 32, 70, 207, 0, 224, 74, 207, 0, 32, 76, 207, 0, 32, 78, 207, 0, 32, 80, 207, 0, 32, 1, 0, 43, 21, 240, 31, 247, 114, 1, 0, 43, 21, 96, 229, 218, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 26, 231, 114, 1, 0, 43, 21, 176, 78, 234, 114, 1, 0, 43, 21, 112, 5, 241, 114, 1, 0, 43, 21, 240, 91, 237, 114, 1, 0, 43, 21, 144, 226, 238, 114, 3, 0, 0, 0, 16, 150, 151, 3, 0, 59, 249, 2, 138, 91, 90, 3, 1, 0, 43, 21, 64, 181, 213, 114, 9, 0, 175, 11, 30, 13, 251, 13, 239, 7, 157, 8, 153, 18, 131, 11, 60, 21, 210, 22, 102, 207, 0, 32, 104, 207, 0, 32, 106, 207, 0, 32, 110, 207, 0, 32, 112, 207, 0, 32, 114, 207, 0, 32, 116, 207, 0, 32, 118, 207, 0, 224, 122, 207, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 84, 250, 114, 3, 0, 0, 0, 252, 56, 149, 3, 16, 92, 250, 2, 66, 140, 75, 3, 2, 0, 102, 8, 60, 21, 170, 170, 138, 207, 0, 32, 142, 207, 0, 224, 1, 0, 43, 21, 16, 94, 208, 114, 1, 0, 43, 21, 80, 70, 238, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 0, 0, 0, 112, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 27, 240, 114, 1, 0, 43, 21, 96, 26, 231, 114, 1, 0, 43, 21, 208, 239, 241, 114, 3, 0, 0, 0, 232, 220, 150, 3, 88, 15, 240, 2, 70, 200, 104, 3, 1, 0, 43, 21, 192, 147, 229, 114, 1, 0, 43, 21, 0, 2, 196, 114, 10, 0, 112, 13, 103, 6, 209, 6, 178, 21, 102, 8, 7, 7, 121, 8, 60, 21, 213, 12, 17, 13, 170, 170, 150, 207, 0, 32, 152, 207, 0, 32, 154, 207, 0, 32, 156, 207, 0, 32, 158, 207, 0, 32, 162, 207, 0, 32, 164, 207, 0, 32, 166, 207, 0, 224, 170, 207, 0, 32, 172, 207, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 84, 250, 114, 3, 0, 0, 0, 28, 239, 115, 3, 112, 201, 250, 2, 16, 98, 43, 3, 2, 0, 251, 13, 60, 21, 170, 170, 190, 207, 0, 32, 194, 207, 0, 224, 1, 0, 43, 21, 176, 235, 187, 114, 1, 0, 43, 21, 240, 188, 200, 114, 1, 0, 43, 21, 48, 255, 215, 114, 1, 0, 43, 21, 208, 27, 193, 114, 1, 0, 43, 21, 224, 142, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 196, 243, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 0, 0, 0, 112, 2, 0, 119, 21, 43, 21, 170, 170, 128, 187, 238, 114, 48, 165, 230, 114, 1, 0, 43, 21, 16, 170, 237, 114, 1, 0, 43, 21, 160, 242, 221, 114, 1, 0, 43, 21, 224, 142, 222, 114, 1, 0, 43, 21, 16, 170, 237, 114, 1, 0, 43, 21, 16, 94, 208, 114, 1, 0, 43, 21, 80, 24, 204, 114, 1, 0, 43, 21, 144, 226, 238, 114, 3, 0, 0, 0, 100, 193, 179, 3, 200, 8, 245, 2, 61, 29, 120, 3, 2, 0, 119, 21, 43, 21, 170, 170, 208, 126, 239, 114, 0, 221, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 170, 237, 114, 19, 0, 162, 4, 112, 13, 80, 5, 209, 5, 161, 6, 251, 13, 209, 6, 178, 21, 15, 7, 121, 8, 217, 21, 246, 7, 186, 12, 57, 8, 77, 8, 4, 11, 60, 21, 189, 12, 214, 10, 202, 207, 0, 32, 204, 207, 0, 32, 206, 207, 0, 32, 208, 207, 0, 32, 210, 207, 0, 32, 212, 207, 0, 32, 216, 207, 0, 32, 218, 207, 0, 32, 220, 207, 0, 32, 224, 207, 0, 32, 226, 207, 0, 32, 228, 207, 0, 32, 230, 207, 0, 32, 232, 207, 0, 32, 234, 207, 0, 32, 236, 207, 0, 32, 238, 207, 0, 224, 242, 207, 0, 32, 246, 207, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 149, 247, 114, 3, 0, 0, 0, 224, 20, 153, 3, 32, 149, 247, 2, 9, 199, 95, 3, 2, 0, 247, 11, 60, 21, 170, 170, 23, 208, 0, 32, 27, 208, 0, 224, 2, 0, 43, 21, 86, 21, 170, 170, 176, 131, 246, 114, 32, 13, 228, 114, 1, 0, 43, 21, 224, 105, 249, 114, 1, 0, 43, 21, 16, 27, 240, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 176, 251, 226, 114, 2, 0, 43, 21, 86, 21, 170, 170, 64, 234, 225, 114, 32, 13, 228, 114, 1, 0, 43, 21, 32, 43, 223, 114, 2, 0, 43, 21, 86, 21, 170, 170, 112, 208, 228, 114, 32, 13, 228, 114, 2, 0, 43, 21, 86, 21, 170, 170, 240, 91, 237, 114, 32, 13, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 1, 0, 43, 21, 192, 64, 222, 114, 3, 0, 0, 0, 48, 228, 176, 3, 112, 201, 250, 2, 56, 94, 104, 3, 1, 0, 43, 21, 192, 64, 222, 114, 1, 0, 43, 21, 224, 142, 222, 114, 13, 0, 30, 13, 251, 13, 102, 8, 106, 8, 102, 7, 7, 12, 42, 13, 115, 10, 222, 18, 103, 22, 60, 21, 45, 10, 246, 13, 35, 208, 0, 32, 39, 208, 0, 32, 41, 208, 0, 32, 43, 208, 0, 32, 47, 208, 0, 32, 51, 208, 0, 32, 53, 208, 0, 32, 57, 208, 0, 32, 61, 208, 0, 32, 65, 208, 0, 32, 67, 208, 0, 224, 71, 208, 0, 32, 73, 208, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 126, 239, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 144, 150, 209, 114, 3, 0, 0, 0, 80, 154, 168, 3, 32, 119, 252, 2, 131, 176, 94, 3, 4, 0, 206, 18, 148, 19, 251, 3, 60, 21, 170, 170, 95, 208, 0, 32, 99, 208, 0, 32, 103, 208, 0, 32, 107, 208, 0, 224, 1, 0, 43, 21, 96, 199, 223, 114, 1, 0, 43, 21, 240, 181, 222, 114, 1, 0, 43, 21, 32, 103, 213, 114, 1, 0, 43, 21, 64, 204, 230, 114, 1, 0, 43, 21, 96, 63, 204, 114, 1, 0, 43, 21, 64, 211, 208, 114, 1, 0, 43, 21, 144, 60, 224, 114, 1, 0, 43, 21, 224, 255, 224, 114, 1, 0, 43, 21, 144, 150, 209, 114, 1, 0, 43, 21, 96, 146, 211, 114, 1, 0, 43, 21, 0, 78, 225, 114, 1, 0, 43, 21, 48, 36, 189, 114, 1, 0, 43, 21, 176, 25, 222, 114, 1, 0, 43, 21, 240, 91, 237, 114, 1, 0, 43, 21, 176, 25, 222, 114, 1, 0, 43, 21, 240, 211, 217, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 101, 242, 114, 1, 0, 43, 21, 16, 117, 225, 114, 2, 0, 119, 21, 43, 21, 170, 170, 112, 231, 245, 114, 128, 224, 211, 114, 1, 0, 43, 21, 112, 185, 211, 114, 1, 0, 43, 21, 208, 103, 222, 114, 1, 0, 43, 21, 128, 254, 206, 114, 3, 0, 0, 0, 92, 156, 159, 3, 112, 112, 244, 2, 24, 73, 106, 3, 1, 0, 43, 21, 64, 121, 223, 114, 1, 0, 43, 21, 160, 242, 221, 114, 1, 0, 43, 21, 192, 64, 222, 114, 1, 0, 43, 21, 64, 211, 208, 114, 1, 0, 43, 21, 176, 205, 192, 114, 1, 0, 43, 21, 0, 25, 213, 114, 29, 0, 226, 4, 238, 4, 38, 5, 69, 5, 112, 5, 115, 5, 131, 6, 251, 13, 148, 7, 255, 11, 26, 8, 47, 9, 59, 9, 173, 9, 241, 9, 32, 10, 197, 7, 119, 10, 106, 11, 128, 12, 144, 12, 148, 12, 60, 21, 170, 12, 174, 12, 178, 12, 201, 12, 147, 22, 198, 13, 118, 208, 0, 32, 120, 208, 0, 32, 122, 208, 0, 32, 124, 208, 0, 32, 126, 208, 0, 32, 128, 208, 0, 32, 130, 208, 0, 32, 132, 208, 0, 32, 134, 208, 0, 32, 136, 208, 0, 32, 138, 208, 0, 32, 140, 208, 0, 32, 142, 208, 0, 32, 144, 208, 0, 32, 146, 208, 0, 32, 148, 208, 0, 32, 150, 208, 0, 32, 154, 208, 0, 32, 156, 208, 0, 32, 160, 208, 0, 32, 162, 208, 0, 32, 164, 208, 0, 32, 166, 208, 0, 224, 170, 208, 0, 32, 172, 208, 0, 32, 174, 208, 0, 32, 176, 208, 0, 32, 178, 208, 0, 32, 180, 208, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 123, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 131, 246, 114, 3, 0, 0, 0, 128, 115, 119, 3, 112, 201, 250, 2, 48, 129, 55, 3, 3, 0, 251, 13, 179, 13, 60, 21, 226, 208, 0, 32, 230, 208, 0, 32, 234, 208, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 92, 246, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 92, 246, 114, 3, 0, 0, 0, 68, 71, 103, 3, 48, 45, 250, 2, 72, 47, 43, 3, 3, 0, 251, 13, 143, 10, 60, 21, 243, 208, 0, 32, 247, 208, 0, 32, 251, 208, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 96, 162, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 192, 177, 224, 114, 3, 0, 0, 0, 136, 171, 165, 3, 112, 201, 250, 2, 163, 109, 93, 3, 3, 0, 251, 13, 3, 14, 60, 21, 4, 209, 0, 32, 8, 209, 0, 32, 12, 209, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 10, 248, 114, 1, 0, 43, 21, 16, 230, 227, 114, 1, 0, 43, 21, 160, 129, 219, 114, 3, 0, 0, 0, 212, 175, 165, 3, 200, 240, 248, 2, 228, 210, 91, 3, 4, 0, 112, 13, 230, 4, 222, 9, 60, 21, 170, 170, 21, 209, 0, 32, 25, 209, 0, 32, 27, 209, 0, 32, 29, 209, 0, 224, 1, 0, 43, 21, 32, 209, 237, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 53, 246, 114, 3, 0, 0, 0, 228, 95, 164, 3, 8, 10, 250, 2, 160, 74, 92, 3, 1, 0, 43, 21, 192, 154, 207, 114, 4, 0, 251, 13, 247, 11, 60, 21, 140, 21, 170, 170, 40, 209, 0, 32, 42, 209, 0, 32, 46, 209, 0, 224, 50, 209, 0, 32, 1, 0, 43, 21, 0, 221, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 209, 246, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 248, 237, 114, 3, 0, 0, 0, 28, 90, 169, 3, 64, 90, 249, 2, 32, 136, 106, 3, 4, 0, 148, 18, 247, 11, 63, 11, 60, 21, 170, 170, 59, 209, 0, 32, 61, 209, 0, 32, 65, 209, 0, 32, 69, 209, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 253, 244, 114, 3, 0, 0, 0, 72, 165, 140, 3, 96, 162, 250, 2, 5, 170, 75, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 105, 240, 114, 1, 0, 43, 21, 224, 82, 232, 114, 4, 0, 102, 8, 60, 21, 220, 24, 140, 21, 170, 170, 80, 209, 0, 32, 84, 209, 0, 224, 88, 209, 0, 32, 92, 209, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 79, 243, 114, 3, 0, 71, 21, 43, 21, 86, 21, 16, 206, 0, 96, 80, 160, 223, 114, 32, 13, 228, 114, 3, 0, 0, 0, 76, 34, 152, 3, 192, 51, 245, 2, 13, 55, 97, 3, 3, 0, 119, 21, 71, 21, 43, 21, 16, 140, 242, 114, 16, 206, 0, 96, 176, 131, 246, 114, 4, 0, 251, 13, 238, 5, 60, 21, 225, 12, 170, 170, 101, 209, 0, 32, 105, 209, 0, 32, 110, 209, 0, 224, 114, 209, 0, 32, 1, 0, 43, 21, 112, 238, 223, 114, 1, 0, 43, 21, 224, 255, 224, 114, 1, 0, 43, 21, 0, 62, 186, 114, 1, 0, 43, 21, 128, 111, 209, 114, 3, 0, 119, 21, 71, 21, 43, 21, 16, 170, 237, 114, 25, 206, 0, 96, 192, 87, 239, 114, 1, 0, 43, 21, 224, 59, 215, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 114, 245, 114, 1, 0, 43, 21, 48, 105, 240, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 214, 244, 114, 3, 0, 119, 21, 71, 21, 43, 21, 0, 161, 232, 114, 25, 206, 0, 96, 64, 61, 233, 114, 1, 0, 43, 21, 128, 88, 192, 114, 3, 0, 119, 21, 71, 21, 43, 21, 224, 52, 237, 114, 25, 206, 0, 96, 0, 18, 235, 114, 1, 0, 43, 21, 48, 29, 211, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 192, 64, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 224, 142, 222, 114, 1, 0, 43, 21, 112, 238, 223, 114, 2, 0, 119, 21, 43, 21, 170, 170, 128, 44, 241, 114, 32, 253, 188, 114, 3, 0, 0, 0, 208, 43, 176, 3, 208, 60, 250, 2, 37, 160, 118, 3, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 176, 25, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 208, 216, 224, 114, 1, 0, 43, 21, 80, 77, 216, 114, 21, 0, 65, 5, 76, 5, 100, 5, 153, 5, 217, 5, 25, 6, 251, 13, 247, 11, 220, 6, 176, 7, 204, 7, 2, 8, 203, 8, 197, 9, 236, 10, 244, 10, 132, 12, 60, 21, 241, 12, 111, 13, 140, 21, 126, 209, 0, 32, 128, 209, 0, 32, 130, 209, 0, 32, 132, 209, 0, 32, 134, 209, 0, 32, 139, 209, 0, 32, 141, 209, 0, 32, 145, 209, 0, 32, 147, 209, 0, 32, 151, 209, 0, 32, 156, 209, 0, 32, 158, 209, 0, 32, 163, 209, 0, 32, 165, 209, 0, 32, 169, 209, 0, 32, 173, 209, 0, 32, 175, 209, 0, 32, 179, 209, 0, 224, 183, 209, 0, 32, 187, 209, 0, 32, 191, 209, 0, 32, 1, 0, 43, 21, 80, 17, 226, 114, 1, 0, 43, 21, 224, 232, 207, 114, 2, 0, 43, 21, 86, 21, 170, 170, 80, 220, 213, 114, 32, 13, 228, 114, 1, 0, 43, 21, 208, 103, 222, 114, 1, 0, 43, 21, 176, 205, 192, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 79, 243, 114, 1, 0, 43, 21, 224, 59, 215, 114, 1, 0, 43, 21, 112, 215, 206, 114, 1, 0, 43, 21, 0, 55, 208, 114, 1, 0, 43, 21, 80, 107, 211, 114, 1, 0, 43, 21, 192, 64, 222, 114, 1, 0, 43, 21, 176, 115, 207, 114, 1, 0, 43, 21, 208, 110, 200, 114, 1, 0, 43, 21, 192, 154, 207, 114, 1, 0, 43, 21, 112, 215, 206, 114, 1, 0, 43, 21, 192, 200, 241, 114, 1, 0, 43, 21, 160, 76, 207, 114, 1, 0, 43, 21, 160, 92, 246, 114, 1, 0, 43, 21, 48, 248, 237, 114, 1, 0, 43, 21, 0, 131, 237, 114, 2, 0, 119, 21, 43, 21, 170, 170, 160, 242, 221, 114, 176, 78, 234, 114, 3, 0, 0, 0, 244, 169, 179, 3, 200, 14, 244, 2, 119, 46, 120, 3, 1, 0, 43, 21, 208, 50, 210, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 24, 0, 230, 4, 34, 5, 49, 5, 134, 5, 157, 5, 251, 13, 222, 9, 66, 7, 70, 7, 192, 7, 208, 7, 140, 8, 215, 8, 245, 8, 0, 9, 20, 9, 56, 10, 162, 22, 26, 6, 238, 12, 10, 12, 60, 21, 170, 22, 138, 7, 170, 170, 225, 209, 0, 32, 227, 209, 0, 32, 229, 209, 0, 32, 233, 209, 0, 32, 235, 209, 0, 32, 237, 209, 0, 32, 241, 209, 0, 32, 243, 209, 0, 32, 245, 209, 0, 32, 247, 209, 0, 32, 249, 209, 0, 32, 251, 209, 0, 32, 253, 209, 0, 32, 255, 209, 0, 32, 1, 210, 0, 32, 3, 210, 0, 32, 5, 210, 0, 32, 7, 210, 0, 32, 9, 210, 0, 32, 11, 210, 0, 32, 13, 210, 0, 32, 17, 210, 0, 224, 21, 210, 0, 32, 23, 210, 0, 32, 1, 0, 43, 21, 96, 3, 214, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 144, 37, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 112, 148, 238, 114, 1, 0, 43, 21, 80, 183, 240, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 208, 27, 193, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 84, 250, 114, 1, 0, 43, 21, 96, 109, 238, 114, 1, 0, 43, 21, 112, 162, 194, 114, 1, 0, 43, 21, 208, 103, 222, 114, 3, 0, 0, 0, 88, 195, 179, 3, 200, 228, 250, 2, 15, 157, 107, 3, 1, 0, 43, 21, 48, 29, 211, 114, 11, 0, 36, 9, 53, 6, 30, 13, 251, 13, 170, 23, 32, 11, 230, 12, 14, 12, 120, 12, 60, 21, 82, 12, 64, 210, 0, 32, 66, 210, 0, 32, 70, 210, 0, 32, 74, 210, 0, 32, 76, 210, 0, 32, 80, 210, 0, 32, 84, 210, 0, 32, 86, 210, 0, 32, 88, 210, 0, 32, 90, 210, 0, 224, 94, 210, 0, 32, 1, 0, 43, 21, 32, 156, 225, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 223, 249, 114, 3, 0, 0, 0, 52, 255, 120, 3, 112, 201, 250, 2, 92, 16, 49, 3, 3, 0, 251, 13, 102, 8, 60, 21, 113, 210, 0, 32, 115, 210, 0, 32, 119, 210, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 166, 248, 114, 3, 0, 0, 0, 168, 49, 90, 3, 112, 201, 250, 2, 32, 112, 18, 3, 2, 0, 251, 13, 60, 21, 170, 170, 128, 210, 0, 32, 132, 210, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 80, 40, 243, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 244, 248, 114, 3, 0, 0, 0, 16, 197, 164, 3, 152, 117, 249, 2, 247, 132, 91, 3, 3, 0, 251, 13, 247, 11, 60, 21, 140, 210, 0, 32, 144, 210, 0, 32, 148, 210, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 232, 195, 150, 3, 232, 56, 250, 2, 161, 144, 93, 3, 2, 0, 112, 13, 60, 21, 170, 170, 157, 210, 0, 32, 161, 210, 0, 224, 1, 0, 43, 21, 240, 204, 239, 114, 1, 0, 43, 21, 224, 52, 237, 114, 1, 0, 43, 21, 48, 248, 237, 114, 1, 0, 43, 21, 160, 76, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 84, 250, 114, 3, 0, 0, 0, 52, 148, 167, 3, 208, 60, 250, 2, 195, 84, 103, 3, 6, 0, 251, 13, 247, 11, 102, 8, 124, 10, 31, 4, 60, 21, 170, 170, 169, 210, 0, 32, 171, 210, 0, 32, 173, 210, 0, 32, 175, 210, 0, 32, 177, 210, 0, 32, 181, 210, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 178, 233, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 248, 246, 114, 3, 0, 0, 0, 192, 9, 121, 3, 0, 184, 249, 2, 248, 65, 59, 3, 3, 0, 251, 13, 252, 10, 60, 21, 195, 210, 0, 32, 199, 210, 0, 32, 203, 210, 0, 224, 1, 0, 43, 21, 208, 216, 224, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 78, 225, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 227, 247, 114, 3, 0, 0, 0, 12, 24, 161, 3, 56, 2, 250, 2, 112, 54, 97, 3, 4, 0, 30, 13, 247, 11, 130, 20, 60, 21, 170, 170, 212, 210, 0, 32, 214, 210, 0, 32, 218, 210, 0, 32, 222, 210, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 66, 249, 114, 1, 0, 43, 21, 240, 91, 237, 114, 1, 0, 43, 21, 128, 111, 209, 114, 3, 0, 0, 0, 108, 247, 167, 3, 248, 95, 250, 2, 69, 25, 90, 3, 4, 0, 112, 13, 222, 9, 27, 8, 60, 21, 170, 170, 233, 210, 0, 32, 237, 210, 0, 32, 239, 210, 0, 32, 241, 210, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 66, 249, 114, 1, 0, 43, 21, 176, 131, 246, 114, 1, 0, 43, 21, 48, 248, 237, 114, 3, 0, 0, 0, 60, 148, 139, 3, 192, 27, 249, 2, 64, 64, 83, 3, 4, 0, 102, 8, 86, 11, 154, 14, 60, 21, 170, 170, 252, 210, 0, 32, 0, 211, 0, 32, 2, 211, 0, 32, 4, 211, 0, 224, 1, 0, 43, 21, 112, 162, 194, 114, 1, 0, 43, 21, 16, 94, 208, 114, 1, 0, 43, 21, 176, 205, 192, 114, 1, 0, 43, 21, 48, 105, 240, 114, 1, 0, 43, 21, 224, 52, 237, 114, 1, 0, 43, 21, 80, 70, 238, 114, 1, 0, 43, 21, 32, 126, 230, 114, 1, 0, 43, 21, 80, 227, 191, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 105, 249, 114, 1, 0, 43, 21, 128, 254, 206, 114, 3, 0, 0, 0, 136, 41, 170, 3, 168, 150, 250, 2, 107, 6, 105, 3, 1, 0, 43, 21, 144, 37, 207, 114, 12, 0, 42, 5, 30, 13, 52, 25, 251, 13, 247, 11, 102, 8, 232, 7, 32, 11, 100, 13, 170, 21, 60, 21, 45, 10, 170, 170, 15, 211, 0, 32, 17, 211, 0, 32, 19, 211, 0, 32, 21, 211, 0, 32, 23, 211, 0, 32, 25, 211, 0, 32, 27, 211, 0, 32, 29, 211, 0, 32, 31, 211, 0, 32, 35, 211, 0, 32, 37, 211, 0, 224, 41, 211, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 144, 90, 219, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 191, 227, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 80, 160, 223, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 16, 34, 218, 114, 1, 0, 43, 21, 112, 148, 238, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 201, 250, 114, 3, 0, 71, 21, 43, 21, 86, 21, 16, 206, 0, 96, 112, 201, 250, 114, 32, 13, 228, 114, 3, 0, 0, 0, 36, 102, 164, 3, 96, 162, 250, 2, 227, 74, 97, 3, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 0, 0, 112, 9, 0, 179, 12, 232, 7, 100, 13, 184, 10, 54, 12, 199, 13, 170, 21, 60, 21, 82, 12, 62, 211, 0, 32, 66, 211, 0, 32, 70, 211, 0, 32, 74, 211, 0, 32, 78, 211, 0, 32, 80, 211, 0, 32, 84, 211, 0, 32, 89, 211, 0, 224, 93, 211, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 85, 203, 114, 1, 0, 43, 21, 96, 10, 192, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 48, 29, 211, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 80, 24, 204, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 208, 103, 222, 114, 1, 0, 43, 21, 32, 133, 208, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 224, 59, 215, 114, 1, 0, 43, 21, 16, 207, 210, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 208, 103, 222, 114, 1, 0, 43, 21, 96, 33, 209, 114, 1, 0, 43, 21, 64, 188, 191, 114, 2, 0, 119, 21, 43, 21, 170, 170, 96, 79, 243, 114, 0, 251, 217, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 192, 154, 207, 114, 1, 0, 43, 21, 160, 233, 160, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 176, 228, 209, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 80, 197, 196, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 208, 193, 207, 114, 1, 0, 43, 21, 160, 106, 202, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 55, 208, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 224, 232, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 206, 201, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 224, 232, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 16, 94, 208, 114, 1, 0, 43, 21, 48, 239, 176, 114, 1, 0, 43, 21, 208, 27, 193, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 176, 228, 209, 114, 1, 0, 43, 21, 192, 11, 210, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 6, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 192, 11, 210, 114, 1, 0, 43, 21, 112, 252, 179, 114, 3, 0, 0, 0, 64, 44, 184, 3, 200, 228, 250, 2, 171, 45, 119, 3, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 224, 142, 222, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 144, 37, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 32, 246, 210, 114, 1, 0, 43, 21, 208, 117, 178, 114, 1, 0, 43, 21, 0, 0, 0, 112, 37, 0, 122, 4, 158, 4, 215, 13, 154, 21, 243, 7, 161, 5, 19, 26, 229, 5, 65, 6, 68, 6, 75, 6, 162, 6, 250, 5, 6, 8, 10, 8, 69, 8, 175, 8, 195, 8, 207, 8, 230, 8, 238, 8, 35, 9, 51, 9, 193, 9, 171, 13, 36, 10, 83, 10, 29, 10, 230, 12, 178, 11, 170, 21, 60, 21, 6, 13, 21, 13, 33, 13, 72, 13, 95, 13, 111, 211, 0, 32, 115, 211, 0, 32, 117, 211, 0, 32, 121, 211, 0, 32, 125, 211, 0, 32, 129, 211, 0, 32, 131, 211, 0, 32, 135, 211, 0, 32, 137, 211, 0, 32, 139, 211, 0, 32, 141, 211, 0, 32, 143, 211, 0, 32, 145, 211, 0, 32, 149, 211, 0, 32, 153, 211, 0, 32, 155, 211, 0, 32, 159, 211, 0, 32, 163, 211, 0, 32, 167, 211, 0, 32, 169, 211, 0, 32, 173, 211, 0, 32, 177, 211, 0, 32, 181, 211, 0, 32, 185, 211, 0, 32, 189, 211, 0, 32, 191, 211, 0, 32, 193, 211, 0, 32, 197, 211, 0, 32, 199, 211, 0, 32, 203, 211, 0, 32, 207, 211, 0, 32, 209, 211, 0, 224, 213, 211, 0, 32, 217, 211, 0, 32, 221, 211, 0, 32, 225, 211, 0, 32, 227, 211, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 248, 237, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 205, 192, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 110, 247, 114, 3, 0, 0, 0, 16, 75, 151, 3, 152, 135, 246, 2, 82, 156, 103, 3, 4, 0, 251, 13, 102, 8, 203, 4, 60, 21, 170, 170, 29, 212, 0, 32, 33, 212, 0, 32, 37, 212, 0, 32, 41, 212, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 2, 0, 43, 21, 86, 21, 170, 170, 224, 255, 224, 114, 32, 13, 228, 114, 3, 0, 0, 0, 220, 192, 180, 3, 128, 252, 248, 2, 223, 204, 106, 3, 3, 0, 112, 13, 198, 4, 60, 21, 52, 212, 0, 32, 56, 212, 0, 32, 60, 212, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 1, 0, 43, 21, 240, 181, 222, 114, 3, 0, 0, 0, 72, 76, 134, 3, 104, 131, 248, 2, 89, 78, 81, 3, 3, 0, 251, 13, 154, 11, 60, 21, 69, 212, 0, 32, 73, 212, 0, 32, 75, 212, 0, 224, 1, 0, 43, 21, 96, 162, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 122, 241, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 214, 244, 114, 3, 0, 0, 0, 248, 118, 136, 3, 48, 176, 249, 2, 26, 106, 70, 3, 4, 0, 49, 6, 251, 13, 102, 8, 60, 21, 170, 170, 84, 212, 0, 32, 86, 212, 0, 32, 90, 212, 0, 32, 94, 212, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 253, 244, 114, 1, 0, 43, 21, 96, 169, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 253, 244, 114, 1, 0, 43, 21, 240, 211, 217, 114, 1, 0, 43, 21, 176, 138, 224, 114, 1, 0, 43, 21, 32, 43, 223, 114, 1, 0, 43, 21, 240, 68, 220, 114, 3, 0, 0, 0, 56, 126, 165, 3, 216, 166, 246, 2, 168, 84, 108, 3, 1, 0, 43, 21, 240, 241, 212, 114, 9, 0, 112, 13, 10, 5, 251, 13, 216, 6, 19, 7, 217, 21, 0, 10, 60, 21, 234, 13, 105, 212, 0, 32, 109, 212, 0, 32, 111, 212, 0, 32, 115, 212, 0, 32, 117, 212, 0, 32, 119, 212, 0, 32, 121, 212, 0, 32, 123, 212, 0, 224, 127, 212, 0, 32, 1, 0, 43, 21, 160, 205, 248, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 96, 199, 223, 114, 1, 0, 43, 21, 48, 142, 213, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 112, 102, 204, 114, 1, 0, 43, 21, 208, 50, 210, 114, 2, 0, 119, 21, 43, 21, 170, 170, 240, 31, 247, 114, 32, 57, 179, 114, 1, 0, 43, 21, 112, 49, 192, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 3, 0, 0, 0, 176, 176, 170, 3, 112, 201, 250, 2, 122, 72, 103, 3, 1, 0, 43, 21, 96, 123, 194, 114, 11, 0, 251, 13, 162, 6, 224, 6, 99, 11, 126, 11, 251, 3, 22, 12, 26, 12, 89, 12, 60, 21, 109, 12, 143, 212, 0, 32, 145, 212, 0, 32, 149, 212, 0, 32, 151, 212, 0, 32, 153, 212, 0, 32, 155, 212, 0, 32, 157, 212, 0, 32, 161, 212, 0, 32, 163, 212, 0, 32, 167, 212, 0, 224, 171, 212, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 223, 249, 114, 1, 0, 43, 21, 208, 193, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 209, 237, 114, 1, 0, 43, 21, 192, 154, 207, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 160, 223, 114, 3, 0, 0, 0, 36, 216, 170, 3, 88, 80, 250, 2, 47, 164, 95, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 110, 247, 114, 7, 0, 251, 13, 27, 8, 39, 12, 233, 10, 154, 14, 60, 21, 154, 9, 190, 212, 0, 32, 194, 212, 0, 32, 196, 212, 0, 32, 200, 212, 0, 32, 202, 212, 0, 32, 206, 212, 0, 224, 210, 212, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 53, 246, 114, 3, 0, 0, 0, 24, 127, 106, 3, 80, 123, 250, 2, 36, 174, 52, 3, 2, 0, 251, 13, 60, 21, 170, 170, 225, 212, 0, 32, 229, 212, 0, 224, 1, 0, 43, 21, 224, 22, 242, 114, 1, 0, 43, 21, 240, 174, 244, 114, 1, 0, 43, 21, 16, 253, 244, 114, 1, 0, 43, 21, 128, 224, 211, 114, 1, 0, 43, 21, 240, 15, 208, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 244, 248, 114, 3, 0, 0, 0, 108, 101, 158, 3, 200, 228, 250, 2, 37, 179, 89, 3, 7, 0, 30, 13, 251, 13, 179, 12, 232, 7, 48, 4, 24, 19, 60, 21, 237, 212, 0, 32, 239, 212, 0, 32, 241, 212, 0, 32, 243, 212, 0, 32, 245, 212, 0, 32, 247, 212, 0, 32, 251, 212, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 135, 244, 114, 1, 0, 43, 21, 32, 179, 242, 114, 3, 0, 0, 0, 132, 206, 119, 3, 32, 119, 252, 2, 60, 168, 44, 3, 3, 0, 206, 18, 230, 12, 60, 21, 10, 213, 0, 32, 14, 213, 0, 32, 16, 213, 0, 224, 1, 0, 43, 21, 48, 218, 242, 114, 1, 0, 43, 21, 160, 9, 239, 114, 1, 0, 43, 21, 224, 165, 239, 114, 1, 0, 43, 21, 240, 91, 237, 114, 1, 0, 43, 21, 144, 90, 219, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 105, 249, 114, 3, 0, 0, 0, 52, 135, 165, 3, 224, 224, 250, 2, 248, 204, 94, 3, 1, 0, 43, 21, 48, 82, 223, 114, 8, 0, 215, 11, 30, 13, 251, 13, 232, 7, 32, 11, 184, 10, 60, 21, 82, 12, 170, 170, 25, 213, 0, 32, 27, 213, 0, 32, 29, 213, 0, 32, 31, 213, 0, 32, 33, 213, 0, 32, 35, 213, 0, 32, 39, 213, 0, 224, 43, 213, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 5, 241, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 48, 45, 250, 114, 1, 0, 43, 21, 240, 204, 239, 114, 2, 0, 119, 21, 43, 21, 170, 170, 192, 147, 229, 114, 224, 165, 239, 114, 3, 0, 0, 0, 44, 74, 149, 3, 168, 73, 242, 2, 181, 31, 96, 3, 7, 0, 251, 13, 209, 6, 178, 21, 199, 8, 209, 9, 68, 0, 60, 21, 58, 213, 0, 32, 62, 213, 0, 32, 64, 213, 0, 32, 66, 213, 0, 32, 68, 213, 0, 32, 70, 213, 0, 32, 74, 213, 0, 224, 1, 0, 43, 21, 80, 190, 218, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 66, 249, 114, 3, 0, 0, 0, 0, 107, 135, 3, 64, 84, 250, 2, 254, 32, 61, 3, 3, 0, 156, 12, 48, 4, 60, 21, 89, 213, 0, 32, 91, 213, 0, 32, 95, 213, 0, 224, 3, 0, 119, 21, 71, 21, 43, 21, 32, 119, 252, 114, 25, 206, 0, 96, 144, 120, 214, 114, 3, 0, 119, 21, 71, 21, 43, 21, 224, 52, 237, 114, 25, 206, 0, 96, 144, 180, 204, 114, 3, 0, 119, 21, 71, 21, 43, 21, 224, 52, 237, 114, 25, 206, 0, 96, 160, 99, 224, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 144, 226, 238, 114, 1, 0, 43, 21, 0, 0, 0, 112, 3, 0, 119, 21, 71, 21, 43, 21, 32, 119, 252, 114, 16, 206, 0, 96, 176, 161, 241, 114, 3, 0, 119, 21, 71, 21, 43, 21, 224, 52, 237, 114, 25, 206, 0, 96, 16, 94, 208, 114, 3, 0, 119, 21, 71, 21, 43, 21, 224, 52, 237, 114, 25, 206, 0, 96, 128, 58, 197, 114, 3, 0, 119, 21, 71, 21, 43, 21, 224, 52, 237, 114, 25, 206, 0, 96, 144, 233, 216, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 240, 181, 222, 114, 3, 0, 119, 21, 71, 21, 43, 21, 224, 52, 237, 114, 25, 206, 0, 96, 64, 211, 208, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 240, 91, 237, 114, 3, 0, 0, 0, 160, 227, 155, 3, 168, 67, 243, 2, 1, 11, 104, 3, 3, 0, 119, 21, 71, 21, 43, 21, 224, 52, 237, 114, 25, 206, 0, 96, 240, 105, 193, 114, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 160, 92, 246, 114, 15, 0, 92, 5, 145, 5, 146, 6, 209, 6, 178, 21, 102, 8, 171, 8, 217, 9, 229, 9, 194, 11, 42, 12, 65, 12, 60, 21, 217, 12, 198, 6, 104, 213, 0, 32, 109, 213, 0, 32, 114, 213, 0, 32, 119, 213, 0, 32, 123, 213, 0, 32, 125, 213, 0, 32, 130, 213, 0, 32, 135, 213, 0, 32, 140, 213, 0, 32, 145, 213, 0, 32, 149, 213, 0, 32, 154, 213, 0, 32, 158, 213, 0, 224, 162, 213, 0, 32, 167, 213, 0, 32, 3, 0, 119, 21, 71, 21, 43, 21, 112, 201, 250, 114, 16, 206, 0, 96, 96, 222, 240, 114, 1, 0, 43, 21, 96, 146, 211, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 149, 247, 114, 1, 0, 43, 21, 64, 121, 223, 114, 3, 0, 0, 0, 112, 165, 150, 3, 208, 114, 241, 2, 14, 120, 103, 3, 5, 0, 112, 13, 143, 11, 188, 10, 123, 7, 60, 21, 194, 213, 0, 32, 199, 213, 0, 32, 201, 213, 0, 32, 205, 213, 0, 32, 207, 213, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 105, 249, 114, 2, 0, 119, 21, 43, 21, 170, 170, 240, 31, 247, 114, 128, 14, 246, 114, 3, 0, 0, 0, 128, 134, 145, 3, 120, 33, 250, 2, 103, 91, 80, 3, 1, 0, 43, 21, 192, 64, 222, 114, 4, 0, 222, 18, 61, 12, 60, 21, 140, 21, 170, 170, 219, 213, 0, 32, 223, 213, 0, 32, 227, 213, 0, 224, 231, 213, 0, 32, 1, 0, 43, 21, 240, 204, 239, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 204, 239, 114, 1, 0, 43, 21, 128, 247, 228, 114, 3, 0, 0, 0, 164, 148, 150, 3, 240, 204, 239, 2, 222, 78, 103, 3, 4, 0, 112, 13, 251, 13, 191, 10, 60, 21, 170, 170, 240, 213, 0, 32, 242, 213, 0, 32, 246, 213, 0, 32, 248, 213, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 166, 248, 114, 3, 0, 0, 0, 48, 122, 127, 3, 24, 73, 246, 2, 102, 90, 74, 3, 2, 0, 31, 4, 60, 21, 170, 170, 3, 214, 0, 32, 7, 214, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 66, 249, 114, 3, 0, 0, 0, 212, 85, 155, 3, 8, 147, 248, 2, 231, 238, 95, 3, 2, 0, 247, 11, 60, 21, 170, 170, 15, 214, 0, 32, 19, 214, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 128, 14, 246, 114, 1, 0, 43, 21, 240, 91, 237, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 17, 226, 114, 3, 0, 0, 0, 104, 151, 122, 3, 112, 201, 250, 2, 146, 124, 62, 3, 1, 0, 43, 21, 144, 143, 231, 114, 5, 0, 251, 13, 247, 11, 222, 18, 60, 21, 79, 13, 27, 214, 0, 32, 31, 214, 0, 32, 33, 214, 0, 32, 37, 214, 0, 224, 41, 214, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 227, 247, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 176, 108, 229, 114, 1, 0, 43, 21, 32, 43, 223, 114, 1, 0, 43, 21, 0, 161, 232, 114, 2, 0, 43, 21, 86, 21, 170, 170, 192, 184, 202, 114, 32, 13, 228, 114, 3, 0, 0, 0, 16, 51, 155, 3, 104, 131, 248, 2, 47, 184, 104, 3, 6, 0, 112, 13, 102, 8, 250, 5, 101, 5, 116, 12, 60, 21, 170, 170, 51, 214, 0, 32, 55, 214, 0, 32, 59, 214, 0, 32, 61, 214, 0, 32, 63, 214, 0, 32, 67, 214, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 227, 247, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 135, 244, 114, 3, 0, 0, 0, 32, 76, 149, 3, 240, 19, 249, 2, 176, 38, 88, 3, 1, 0, 43, 21, 16, 4, 223, 114, 1, 0, 43, 21, 64, 204, 230, 114, 5, 0, 251, 13, 150, 5, 60, 21, 2, 13, 6, 14, 81, 214, 0, 32, 85, 214, 0, 32, 89, 214, 0, 224, 93, 214, 0, 32, 95, 214, 0, 32, 1, 0, 43, 21, 112, 201, 250, 114, 3, 0, 0, 0, 172, 162, 103, 3, 112, 201, 250, 2, 56, 167, 29, 3, 2, 0, 251, 13, 60, 21, 170, 170, 105, 214, 0, 32, 107, 214, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 3, 0, 0, 0, 64, 167, 126, 3, 96, 162, 250, 2, 50, 43, 64, 3, 2, 0, 251, 13, 60, 21, 170, 170, 115, 214, 0, 32, 119, 214, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 70, 238, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 165, 239, 114, 3, 0, 0, 0, 248, 228, 151, 3, 16, 21, 241, 2, 187, 42, 104, 3, 3, 0, 112, 13, 102, 8, 60, 21, 127, 214, 0, 32, 131, 214, 0, 32, 135, 214, 0, 224, 2, 0, 119, 21, 43, 21, 170, 170, 32, 119, 252, 114, 32, 119, 252, 114, 3, 0, 0, 0, 92, 140, 95, 3, 112, 201, 250, 2, 224, 151, 27, 3, 2, 0, 102, 8, 60, 21, 170, 170, 144, 214, 0, 32, 148, 214, 0, 224, 1, 0, 43, 21, 80, 70, 238, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 253, 244, 114, 2, 0, 119, 21, 43, 21, 170, 170, 48, 248, 237, 114, 208, 103, 222, 114, 3, 0, 0, 0, 164, 129, 149, 3, 160, 229, 244, 2, 237, 247, 98, 3, 4, 0, 227, 11, 102, 8, 250, 7, 60, 21, 170, 170, 156, 214, 0, 32, 158, 214, 0, 32, 162, 214, 0, 32, 166, 214, 0, 224, 1, 0, 43, 21, 240, 204, 239, 114, 2, 0, 119, 21, 43, 21, 170, 170, 96, 79, 243, 114, 192, 124, 212, 114, 1, 0, 43, 21, 16, 4, 223, 114, 2, 0, 119, 21, 43, 21, 170, 170, 208, 126, 239, 114, 32, 110, 191, 114, 1, 0, 43, 21, 32, 110, 191, 114, 1, 0, 43, 21, 32, 13, 228, 114, 1, 0, 43, 21, 208, 193, 207, 114, 2, 0, 43, 21, 86, 21, 170, 170, 0, 18, 235, 114, 32, 13, 228, 114, 1, 0, 43, 21, 128, 141, 204, 114, 2, 0, 43, 21, 86, 21, 170, 170, 0, 161, 232, 114, 32, 13, 228, 114, 3, 0, 0, 0, 80, 238, 179, 3, 152, 242, 249, 2, 143, 56, 112, 3, 3, 0, 119, 21, 71, 21, 43, 21, 16, 223, 249, 114, 16, 206, 0, 96, 64, 227, 247, 114, 2, 0, 43, 21, 86, 21, 170, 170, 192, 87, 239, 114, 32, 13, 228, 114, 1, 0, 43, 21, 32, 43, 223, 114, 14, 0, 251, 13, 89, 8, 252, 8, 43, 9, 149, 9, 221, 9, 16, 10, 159, 10, 2, 12, 50, 12, 60, 21, 167, 12, 1, 13, 140, 21, 170, 170, 177, 214, 0, 32, 179, 214, 0, 32, 183, 214, 0, 32, 185, 214, 0, 32, 189, 214, 0, 32, 191, 214, 0, 32, 193, 214, 0, 32, 195, 214, 0, 32, 199, 214, 0, 32, 201, 214, 0, 32, 205, 214, 0, 224, 209, 214, 0, 32, 214, 214, 0, 32, 218, 214, 0, 32, 1, 0, 43, 21, 64, 188, 191, 114, 1, 0, 43, 21, 160, 16, 217, 114, 1, 0, 43, 21, 0, 131, 237, 114, 3, 0, 0, 0, 140, 221, 151, 3, 200, 228, 250, 2, 72, 159, 99, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 5, 0, 112, 13, 222, 9, 230, 12, 60, 21, 108, 9, 242, 214, 0, 32, 244, 214, 0, 32, 246, 214, 0, 32, 248, 214, 0, 224, 252, 214, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 240, 199, 88, 3, 32, 6, 250, 2, 52, 173, 42, 3, 3, 0, 71, 21, 43, 21, 86, 21, 16, 206, 0, 96, 32, 119, 252, 114, 32, 13, 228, 114, 3, 0, 251, 13, 60, 21, 193, 12, 8, 215, 0, 32, 12, 215, 0, 224, 16, 215, 0, 32, 3, 0, 119, 21, 71, 21, 43, 21, 32, 119, 252, 114, 16, 206, 0, 96, 240, 174, 244, 114, 3, 0, 0, 0, 136, 153, 143, 3, 152, 147, 244, 2, 97, 130, 88, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 174, 244, 114, 3, 0, 31, 4, 60, 21, 163, 12, 26, 215, 0, 32, 31, 215, 0, 224, 35, 215, 0, 32, 1, 0, 43, 21, 64, 68, 211, 114, 1, 0, 43, 21, 0, 131, 237, 114, 3, 0, 0, 0, 84, 94, 164, 3, 224, 224, 250, 2, 173, 191, 94, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 92, 246, 114, 1, 0, 43, 21, 160, 39, 234, 114, 5, 0, 101, 5, 230, 12, 60, 21, 230, 10, 143, 9, 44, 215, 0, 32, 46, 215, 0, 32, 48, 215, 0, 224, 52, 215, 0, 32, 56, 215, 0, 32, 1, 0, 43, 21, 224, 105, 249, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 105, 249, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 248, 246, 114, 3, 0, 0, 0, 36, 227, 164, 3, 24, 192, 247, 2, 26, 130, 103, 3, 4, 0, 126, 4, 112, 13, 102, 8, 60, 21, 170, 170, 66, 215, 0, 32, 68, 215, 0, 32, 72, 215, 0, 32, 76, 215, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 244, 239, 114, 3, 0, 0, 0, 24, 7, 126, 3, 112, 201, 250, 2, 94, 219, 72, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 3, 0, 251, 13, 60, 21, 154, 5, 87, 215, 0, 32, 91, 215, 0, 224, 95, 215, 0, 32, 1, 0, 43, 21, 80, 31, 182, 114, 1, 0, 43, 21, 48, 29, 211, 114, 1, 0, 43, 21, 64, 181, 213, 114, 1, 0, 43, 21, 96, 116, 216, 114, 1, 0, 43, 21, 208, 20, 215, 114, 1, 0, 43, 21, 96, 146, 211, 114, 1, 0, 43, 21, 160, 46, 212, 114, 1, 0, 43, 21, 80, 31, 182, 114, 1, 0, 43, 21, 80, 70, 238, 114, 1, 0, 43, 21, 0, 115, 198, 114, 1, 0, 43, 21, 0, 115, 198, 114, 1, 0, 43, 21, 144, 233, 216, 114, 1, 0, 43, 21, 240, 15, 208, 114, 1, 0, 43, 21, 96, 17, 170, 114, 1, 0, 43, 21, 112, 208, 228, 114, 1, 0, 43, 21, 32, 200, 176, 114, 1, 0, 43, 21, 64, 45, 194, 114, 1, 0, 43, 21, 208, 27, 193, 114, 2, 0, 43, 21, 86, 21, 170, 170, 64, 45, 194, 114, 32, 13, 228, 114, 3, 0, 0, 0, 104, 97, 181, 3, 8, 10, 250, 2, 92, 38, 114, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 16, 223, 249, 114, 1, 0, 43, 21, 144, 74, 180, 114, 1, 0, 43, 21, 192, 78, 178, 114, 1, 0, 43, 21, 224, 142, 222, 114, 24, 0, 125, 12, 122, 4, 112, 13, 199, 11, 67, 22, 42, 5, 53, 5, 52, 25, 251, 13, 250, 5, 174, 6, 69, 8, 144, 8, 143, 4, 101, 5, 205, 22, 157, 9, 54, 12, 170, 21, 60, 21, 45, 10, 229, 12, 143, 9, 13, 14, 170, 170, 104, 215, 0, 32, 106, 215, 0, 32, 108, 215, 0, 32, 110, 215, 0, 32, 112, 215, 0, 32, 114, 215, 0, 32, 116, 215, 0, 32, 118, 215, 0, 32, 120, 215, 0, 32, 122, 215, 0, 32, 124, 215, 0, 32, 126, 215, 0, 32, 128, 215, 0, 32, 130, 215, 0, 32, 132, 215, 0, 32, 134, 215, 0, 32, 136, 215, 0, 32, 138, 215, 0, 32, 140, 215, 0, 32, 144, 215, 0, 224, 148, 215, 0, 32, 152, 215, 0, 32, 154, 215, 0, 32, 156, 215, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 192, 27, 249, 114, 1, 0, 43, 21, 224, 89, 210, 114, 3, 0, 0, 0, 212, 16, 154, 3, 160, 193, 250, 2, 209, 88, 88, 3, 3, 0, 251, 13, 247, 11, 60, 21, 195, 215, 0, 32, 199, 215, 0, 32, 201, 215, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 18, 235, 114, 3, 0, 0, 0, 224, 40, 108, 3, 48, 45, 250, 2, 238, 172, 57, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 66, 249, 114, 3, 0, 251, 13, 60, 21, 13, 13, 210, 215, 0, 32, 214, 215, 0, 224, 218, 215, 0, 32, 3, 0, 0, 0, 244, 219, 179, 3, 40, 88, 250, 2, 4, 88, 105, 3, 1, 0, 43, 21, 192, 244, 192, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 45, 250, 114, 3, 0, 60, 21, 233, 12, 140, 21, 227, 215, 0, 224, 231, 215, 0, 32, 233, 215, 0, 32, 1, 0, 43, 21, 192, 64, 222, 114, 1, 0, 43, 21, 16, 4, 223, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 53, 246, 114, 1, 0, 43, 21, 96, 116, 216, 114, 1, 0, 43, 21, 144, 60, 224, 114, 1, 0, 43, 21, 16, 4, 223, 114, 1, 0, 43, 21, 48, 112, 218, 114, 2, 0, 119, 21, 43, 21, 170, 170, 96, 79, 243, 114, 240, 181, 222, 114, 1, 0, 43, 21, 224, 142, 222, 114, 1, 0, 43, 21, 16, 117, 225, 114, 1, 0, 43, 21, 96, 116, 216, 114, 1, 0, 43, 21, 32, 133, 208, 114, 2, 0, 119, 21, 43, 21, 170, 170, 160, 242, 221, 114, 0, 55, 208, 114, 1, 0, 43, 21, 112, 178, 233, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 105, 249, 114, 3, 0, 0, 0, 4, 68, 165, 3, 176, 6, 246, 2, 81, 54, 110, 3, 1, 0, 43, 21, 96, 116, 216, 114, 17, 0, 210, 4, 22, 5, 251, 13, 22, 8, 85, 8, 218, 8, 16, 9, 185, 9, 245, 9, 205, 10, 138, 11, 162, 11, 202, 11, 81, 12, 123, 7, 60, 21, 103, 13, 242, 215, 0, 32, 244, 215, 0, 32, 246, 215, 0, 32, 250, 215, 0, 32, 252, 215, 0, 32, 254, 215, 0, 32, 0, 216, 0, 32, 2, 216, 0, 32, 6, 216, 0, 32, 8, 216, 0, 32, 10, 216, 0, 32, 12, 216, 0, 32, 14, 216, 0, 32, 18, 216, 0, 32, 20, 216, 0, 32, 24, 216, 0, 224, 28, 216, 0, 32, 1, 0, 43, 21, 240, 211, 217, 114, 1, 0, 43, 21, 208, 163, 212, 114, 1, 0, 43, 21, 64, 181, 213, 114, 1, 0, 43, 21, 128, 88, 192, 114, 1, 0, 43, 21, 16, 207, 210, 114, 1, 0, 43, 21, 128, 21, 224, 114, 1, 0, 43, 21, 0, 25, 213, 114, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 32, 179, 242, 114, 1, 0, 43, 21, 192, 64, 222, 114, 3, 0, 0, 0, 0, 109, 168, 3, 200, 228, 250, 2, 35, 117, 108, 3, 1, 0, 43, 21, 96, 146, 211, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 153, 245, 114, 1, 0, 43, 21, 208, 103, 222, 114, 13, 0, 36, 9, 42, 5, 29, 6, 52, 25, 232, 7, 32, 11, 100, 13, 230, 12, 150, 11, 60, 21, 45, 10, 82, 12, 109, 12, 56, 216, 0, 32, 58, 216, 0, 32, 60, 216, 0, 32, 62, 216, 0, 32, 64, 216, 0, 32, 66, 216, 0, 32, 68, 216, 0, 32, 70, 216, 0, 32, 74, 216, 0, 32, 76, 216, 0, 224, 80, 216, 0, 32, 82, 216, 0, 32, 86, 216, 0, 32, 1, 0, 43, 21, 80, 17, 226, 114, 1, 0, 43, 21, 96, 169, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 95, 226, 114, 1, 0, 43, 21, 160, 226, 182, 114, 1, 0, 43, 21, 96, 56, 226, 114, 1, 0, 43, 21, 16, 170, 237, 114, 1, 0, 43, 21, 208, 216, 224, 114, 1, 0, 43, 21, 240, 8, 230, 114, 1, 0, 43, 21, 80, 160, 223, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 240, 31, 247, 114, 1, 0, 43, 21, 112, 95, 226, 114, 3, 0, 0, 0, 252, 34, 161, 3, 160, 217, 246, 2, 25, 252, 107, 3, 1, 0, 43, 21, 48, 82, 223, 114, 1, 0, 43, 21, 80, 130, 228, 114, 14, 0, 114, 4, 221, 5, 251, 13, 255, 11, 24, 9, 43, 17, 87, 10, 209, 10, 203, 4, 123, 7, 155, 12, 60, 21, 253, 12, 186, 13, 170, 170, 108, 216, 0, 32, 110, 216, 0, 32, 112, 216, 0, 32, 116, 216, 0, 32, 118, 216, 0, 32, 120, 216, 0, 32, 122, 216, 0, 32, 124, 216, 0, 32, 126, 216, 0, 32, 128, 216, 0, 32, 132, 216, 0, 32, 134, 216, 0, 224, 138, 216, 0, 32, 140, 216, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 28, 155, 104, 3, 112, 201, 250, 2, 96, 71, 30, 3, 2, 0, 251, 13, 60, 21, 170, 170, 164, 216, 0, 32, 168, 216, 0, 224, 1, 0, 43, 21, 80, 174, 179, 114, 2, 0, 43, 21, 86, 21, 170, 170, 16, 101, 186, 114, 32, 13, 228, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 240, 165, 183, 114, 1, 0, 43, 21, 176, 85, 212, 114, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 64, 84, 250, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 0, 18, 235, 114, 1, 0, 43, 21, 48, 66, 184, 114, 1, 0, 43, 21, 96, 146, 211, 114, 1, 0, 43, 21, 64, 181, 213, 114, 1, 0, 43, 21, 128, 118, 187, 114, 2, 0, 71, 21, 43, 21, 170, 170, 25, 206, 0, 96, 208, 57, 188, 114, 2, 0, 43, 21, 86, 21, 170, 170, 192, 78, 178, 114, 32, 13, 228, 114, 1, 0, 43, 21, 224, 89, 210, 114, 1, 0, 43, 21, 160, 189, 209, 114, 1, 0, 43, 21, 208, 117, 178, 114, 1, 0, 43, 21, 0, 122, 176, 114, 1, 0, 43, 21, 160, 136, 197, 114, 1, 0, 43, 21, 0, 0, 0, 112, 1, 0, 43, 21, 240, 218, 195, 114, 1, 0, 43, 21, 64, 211, 208, 114, 3, 0, 0, 0, 20, 71, 196, 3, 112, 201, 250, 2, 168, 8, 122, 3, 1, 0, 43, 21, 224, 89, 210, 114, 1, 0, 43, 21, 144, 97, 197, 114, 1, 0, 43, 21, 16, 177, 215, 114, 25, 0, 237, 5, 245, 5, 253, 5, 71, 6, 30, 13, 251, 13, 247, 11, 185, 6, 220, 6, 102, 8, 235, 6, 164, 7, 77, 5, 48, 4, 29, 11, 83, 9, 52, 10, 195, 10, 221, 10, 8, 11, 230, 12, 60, 21, 76, 13, 109, 12, 140, 21, 176, 216, 0, 32, 178, 216, 0, 32, 182, 216, 0, 32, 184, 216, 0, 32, 186, 216, 0, 32, 188, 216, 0, 32, 192, 216, 0, 32, 196, 216, 0, 32, 198, 216, 0, 32, 200, 216, 0, 32, 202, 216, 0, 32, 204, 216, 0, 32, 208, 216, 0, 32, 212, 216, 0, 32, 214, 216, 0, 32, 216, 216, 0, 32, 218, 216, 0, 32, 220, 216, 0, 32, 222, 216, 0, 32, 224, 216, 0, 32, 226, 216, 0, 32, 228, 216, 0, 224, 232, 216, 0, 32, 234, 216, 0, 32, 236, 216, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 192, 27, 249, 114, 3, 0, 0, 0, 32, 116, 159, 3, 72, 166, 250, 2, 34, 162, 91, 3, 2, 0, 247, 11, 60, 21, 170, 170, 20, 217, 0, 32, 24, 217, 0, 224, 1, 0, 43, 21, 0, 221, 222, 114, 1, 0, 43, 21, 32, 209, 237, 114, 3, 0, 0, 0, 88, 48, 166, 3, 16, 217, 250, 2, 13, 85, 106, 3, 1, 0, 43, 21, 128, 194, 216, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 144, 166, 248, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 48, 248, 237, 114, 6, 0, 45, 8, 230, 12, 60, 21, 45, 10, 143, 9, 241, 21, 170, 170, 32, 217, 0, 32, 34, 217, 0, 32, 36, 217, 0, 224, 40, 217, 0, 32, 42, 217, 0, 32, 46, 217, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 96, 49, 248, 114, 1, 0, 43, 21, 96, 49, 248, 114, 3, 0, 0, 0, 40, 110, 107, 3, 32, 119, 252, 2, 224, 61, 42, 3, 3, 0, 48, 4, 195, 0, 60, 21, 60, 217, 0, 32, 64, 217, 0, 32, 66, 217, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 84, 250, 114, 3, 0, 0, 0, 228, 50, 134, 3, 64, 84, 250, 2, 148, 201, 72, 3, 2, 0, 251, 13, 60, 21, 170, 170, 75, 217, 0, 32, 79, 217, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 49, 248, 114, 3, 0, 0, 0, 224, 155, 168, 3, 184, 64, 250, 2, 28, 150, 106, 3, 2, 0, 247, 11, 60, 21, 170, 170, 87, 217, 0, 32, 91, 217, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 162, 250, 114, 3, 0, 0, 0, 160, 163, 124, 3, 144, 154, 250, 2, 4, 112, 61, 3, 2, 0, 251, 13, 60, 21, 170, 170, 99, 217, 0, 32, 103, 217, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 7, 206, 0, 96, 240, 31, 247, 114, 3, 0, 0, 0, 64, 45, 138, 3, 112, 201, 250, 2, 161, 221, 72, 3, 2, 0, 251, 13, 60, 21, 170, 170, 111, 217, 0, 32, 115, 217, 0, 224, 2, 0, 119, 21, 43, 21, 170, 170, 0, 214, 244, 114, 16, 124, 203, 114, 3, 0, 0, 0, 60, 175, 172, 3, 176, 238, 249, 2, 227, 142, 116, 3, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 160, 205, 248, 114, 1, 0, 43, 21, 176, 25, 222, 114, 4, 0, 5, 6, 60, 21, 76, 13, 140, 21, 170, 170, 123, 217, 0, 32, 127, 217, 0, 224, 131, 217, 0, 32, 135, 217, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 105, 249, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 112, 88, 248, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 79, 243, 114, 3, 0, 0, 0, 32, 202, 128, 3, 64, 96, 248, 2, 229, 160, 75, 3, 4, 0, 73, 5, 251, 13, 102, 8, 60, 21, 170, 170, 144, 217, 0, 32, 148, 217, 0, 32, 152, 217, 0, 32, 156, 217, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 1, 243, 114, 1, 0, 43, 21, 48, 105, 240, 114, 3, 0, 0, 0, 0, 232, 110, 3, 96, 79, 243, 2, 62, 97, 58, 3, 1, 0, 43, 21, 0, 214, 244, 114, 4, 0, 102, 8, 3, 7, 60, 21, 201, 18, 170, 170, 167, 217, 0, 32, 171, 217, 0, 32, 173, 217, 0, 224, 177, 217, 0, 32, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 96, 199, 223, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 119, 252, 114, 3, 0, 0, 0, 228, 0, 134, 3, 160, 193, 250, 2, 36, 78, 74, 3, 3, 0, 251, 13, 29, 19, 60, 21, 186, 217, 0, 32, 190, 217, 0, 32, 194, 217, 0, 224, 1, 0, 43, 21, 224, 248, 246, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 184, 249, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 13, 228, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 32, 13, 228, 114, 3, 0, 0, 0, 160, 132, 150, 3, 24, 180, 249, 2, 39, 103, 89, 3, 5, 0, 154, 4, 54, 12, 199, 13, 170, 21, 60, 21, 203, 217, 0, 32, 205, 217, 0, 32, 209, 217, 0, 32, 213, 217, 0, 32, 217, 217, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 248, 246, 114, 1, 0, 43, 21, 160, 39, 234, 114, 3, 0, 0, 0, 44, 195, 158, 3, 8, 165, 245, 2, 16, 32, 106, 3, 3, 0, 112, 13, 251, 13, 60, 21, 229, 217, 0, 32, 233, 217, 0, 32, 235, 217, 0, 224, 1, 0, 43, 21, 176, 48, 239, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 208, 96, 244, 114, 1, 0, 43, 21, 224, 142, 222, 114, 1, 0, 43, 21, 192, 27, 249, 114, 3, 0, 0, 0, 128, 143, 131, 3, 0, 184, 249, 2, 144, 49, 74, 3, 5, 0, 165, 5, 102, 8, 123, 7, 92, 12, 60, 21, 244, 217, 0, 32, 246, 217, 0, 32, 250, 217, 0, 32, 252, 217, 0, 32, 254, 217, 0, 224, 3, 0, 119, 21, 71, 21, 43, 21, 112, 201, 250, 114, 25, 206, 0, 96, 16, 170, 237, 114, 3, 0, 119, 21, 71, 21, 43, 21, 112, 201, 250, 114, 16, 206, 0, 96, 48, 105, 240, 114, 2, 0, 119, 21, 43, 21, 170, 170, 144, 53, 246, 114, 64, 121, 223, 114, 3, 0, 119, 21, 71, 21, 43, 21, 96, 79, 243, 114, 25, 206, 0, 96, 0, 221, 222, 114, 3, 0, 119, 21, 71, 21, 43, 21, 96, 79, 243, 114, 25, 206, 0, 96, 224, 195, 234, 114, 3, 0, 119, 21, 71, 21, 43, 21, 96, 79, 243, 114, 25, 206, 0, 96, 176, 138, 224, 114, 3, 0, 119, 21, 71, 21, 43, 21, 96, 79, 243, 114, 25, 206, 0, 96, 240, 121, 232, 114, 2, 0, 119, 21, 43, 21, 170, 170, 96, 79, 243, 114, 96, 100, 177, 114, 3, 0, 0, 0, 32, 126, 174, 3, 16, 223, 249, 2, 138, 51, 110, 3, 3, 0, 119, 21, 71, 21, 43, 21, 96, 79, 243, 114, 25, 206, 0, 96, 32, 239, 232, 114, 3, 0, 119, 21, 71, 21, 43, 21, 96, 79, 243, 114, 25, 206, 0, 96, 192, 34, 227, 114, 3, 0, 119, 21, 71, 21, 43, 21, 96, 79, 243, 114, 25, 206, 0, 96, 112, 238, 223, 114, 3, 0, 119, 21, 71, 21, 43, 21, 96, 79, 243, 114, 25, 206, 0, 96, 96, 109, 238, 114, 3, 0, 119, 21, 71, 21, 43, 21, 96, 79, 243, 114, 25, 206, 0, 96, 192, 87, 239, 114, 14, 0, 175, 11, 251, 13, 255, 11, 183, 13, 187, 10, 150, 5, 51, 0, 123, 7, 60, 21, 210, 22, 2, 13, 65, 13, 155, 13, 6, 14, 170, 170, 10, 218, 0, 32, 15, 218, 0, 32, 20, 218, 0, 32, 24, 218, 0, 32, 29, 218, 0, 32, 34, 218, 0, 32, 39, 218, 0, 32, 44, 218, 0, 32, 48, 218, 0, 224, 52, 218, 0, 32, 57, 218, 0, 32, 62, 218, 0, 32, 67, 218, 0, 32, 72, 218, 0, 32, 1, 0, 43, 21, 48, 105, 240, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 64, 31, 238, 114, 1, 0, 43, 21, 192, 147, 229, 114, 1, 0, 43, 21, 48, 248, 237, 114, 3, 0, 0, 0, 148, 11, 158, 3, 176, 12, 245, 2, 163, 66, 104, 3, 5, 0, 14, 5, 251, 13, 99, 9, 202, 10, 60, 21, 99, 218, 0, 32, 101, 218, 0, 32, 105, 218, 0, 32, 107, 218, 0, 32, 109, 218, 0, 224, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 224, 22, 242, 114, 1, 0, 43, 21, 80, 130, 228, 114, 1, 0, 43, 21, 16, 87, 230, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 0, 131, 237, 114, 1, 0, 43, 21, 208, 186, 229, 114, 1, 0, 43, 21, 48, 82, 223, 114, 2, 0, 71, 21, 43, 21, 170, 170, 16, 206, 0, 96, 80, 10, 248, 114, 3, 0, 0, 0, 172, 193, 152, 3, 224, 111, 248, 2, 80, 228, 103, 3, 1, 0, 43, 21, 32, 133, 208, 114, 1, 0, 43, 21, 192, 237, 214, 114, 10, 0, 251, 13, 81, 8, 76, 10, 50, 13, 111, 10, 202, 10, 229, 7, 60, 21, 210, 22, 65, 13, 170, 170, 121, 218, 0, 32, 125, 218, 0, 32, 127, 218, 0, 32, 129, 218, 0, 32, 133, 218, 0, 32, 135, 218, 0, 32, 137, 218, 0, 32, 141, 218, 0, 224, 145, 218, 0, 32, 147, 218, 0, 32, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 60, 21, 165, 218, 0, 224, 1, 1, 40, 21, 101, 15, 35, 3, 38, 3, 104, 15, 116, 21, 107, 15, 110, 15, 113, 15, 134, 21, 116, 15, 137, 21, 119, 15, 122, 15, 148, 21, 151, 21, 125, 15, 128, 15, 131, 15, 134, 15, 137, 15, 140, 15, 143, 15, 41, 3, 146, 15, 149, 15, 186, 21, 189, 21, 152, 15, 24, 23, 155, 15, 30, 23, 185, 14, 158, 15, 200, 21, 161, 15, 164, 15, 226, 14, 60, 23, 211, 21, 167, 15, 170, 15, 173, 15, 84, 23, 176, 15, 214, 21, 229, 14, 179, 15, 44, 3, 232, 14, 2, 22, 235, 14, 182, 15, 185, 15, 5, 22, 47, 3, 188, 15, 191, 15, 194, 15, 8, 22, 50, 3, 197, 15, 200, 15, 203, 15, 53, 3, 11, 22, 36, 23, 206, 15, 56, 3, 59, 3, 62, 3, 42, 23, 65, 3, 209, 15, 212, 15, 14, 22, 215, 15, 17, 22, 218, 15, 221, 15, 238, 14, 224, 15, 227, 15, 20, 22, 23, 22, 230, 15, 26, 22, 233, 15, 236, 15, 239, 15, 38, 22, 48, 23, 242, 15, 41, 22, 241, 14, 44, 22, 245, 15, 248, 15, 75, 15, 47, 22, 251, 15, 254, 15, 1, 16, 4, 16, 50, 22, 7, 16, 10, 16, 68, 3, 53, 22, 72, 23, 64, 22, 71, 3, 74, 3, 13, 16, 90, 23, 75, 22, 16, 16, 77, 3, 80, 3, 19, 16, 22, 16, 25, 16, 28, 16, 83, 3, 31, 16, 34, 16, 86, 3, 89, 3, 244, 14, 37, 16, 40, 16, 92, 3, 43, 16, 46, 16, 49, 16, 247, 14, 52, 16, 55, 16, 58, 16, 61, 16, 95, 3, 98, 3, 64, 16, 67, 16, 70, 16, 121, 22, 73, 16, 76, 16, 79, 16, 82, 16, 250, 14, 85, 16, 101, 3, 132, 22, 135, 22, 104, 3, 138, 22, 88, 16, 91, 16, 94, 16, 97, 16, 253, 14, 100, 16, 103, 16, 106, 16, 141, 22, 109, 16, 144, 22, 112, 16, 0, 15, 115, 16, 118, 16, 121, 16, 124, 16, 156, 22, 127, 16, 107, 3, 3, 15, 54, 23, 159, 22, 130, 16, 6, 15, 133, 16, 136, 16, 178, 22, 181, 22, 9, 15, 110, 3, 59, 15, 12, 15, 139, 16, 113, 3, 184, 22, 142, 16, 145, 16, 148, 16, 151, 16, 116, 3, 154, 16, 157, 16, 119, 3, 160, 16, 122, 3, 163, 16, 166, 16, 187, 22, 169, 16, 172, 16, 175, 16, 178, 16, 181, 16, 184, 16, 187, 16, 190, 16, 15, 15, 190, 22, 125, 3, 193, 16, 193, 22, 196, 22, 128, 3, 199, 22, 196, 16, 131, 3, 199, 16, 202, 22, 202, 16, 205, 16, 134, 3, 208, 16, 211, 16, 214, 16, 217, 16, 28, 26, 220, 16, 223, 16, 226, 16, 229, 16, 94, 14, 232, 16, 235, 16, 213, 22, 238, 16, 18, 15, 216, 22, 219, 22, 241, 16, 244, 16, 247, 16, 250, 16, 222, 22, 137, 3, 225, 22, 132, 23, 253, 16, 0, 17, 228, 22, 163, 188, 0, 32, 179, 188, 0, 32, 204, 188, 0, 32, 251, 188, 0, 32, 22, 189, 0, 32, 35, 189, 0, 32, 51, 189, 0, 32, 70, 189, 0, 32, 93, 189, 0, 32, 107, 189, 0, 32, 125, 189, 0, 32, 145, 189, 0, 32, 178, 189, 0, 32, 206, 189, 0, 32, 228, 189, 0, 32, 243, 189, 0, 32, 11, 190, 0, 32, 50, 190, 0, 32, 69, 190, 0, 32, 101, 190, 0, 32, 141, 190, 0, 32, 168, 190, 0, 32, 194, 190, 0, 32, 214, 190, 0, 32, 237, 190, 0, 32, 1, 191, 0, 32, 16, 191, 0, 32, 28, 191, 0, 32, 48, 191, 0, 32, 76, 191, 0, 32, 96, 191, 0, 32, 129, 191, 0, 32, 156, 191, 0, 32, 176, 191, 0, 32, 192, 191, 0, 32, 210, 191, 0, 32, 229, 191, 0, 32, 244, 191, 0, 32, 58, 192, 0, 32, 106, 192, 0, 32, 141, 192, 0, 32, 169, 192, 0, 32, 186, 192, 0, 32, 228, 192, 0, 32, 15, 193, 0, 32, 34, 193, 0, 32, 50, 193, 0, 32, 147, 193, 0, 32, 2, 194, 0, 32, 52, 194, 0, 32, 63, 194, 0, 32, 75, 194, 0, 32, 87, 194, 0, 32, 101, 194, 0, 32, 120, 194, 0, 32, 135, 194, 0, 32, 159, 194, 0, 32, 186, 194, 0, 32, 8, 195, 0, 32, 56, 195, 0, 32, 76, 195, 0, 32, 108, 195, 0, 32, 129, 195, 0, 32, 143, 195, 0, 32, 168, 195, 0, 32, 186, 195, 0, 32, 204, 195, 0, 32, 227, 195, 0, 32, 251, 195, 0, 32, 11, 196, 0, 32, 40, 196, 0, 32, 83, 196, 0, 32, 119, 196, 0, 32, 161, 196, 0, 32, 198, 196, 0, 32, 216, 196, 0, 32, 238, 196, 0, 32, 1, 197, 0, 32, 51, 197, 0, 32, 87, 197, 0, 32, 141, 197, 0, 32, 177, 197, 0, 32, 205, 197, 0, 32, 230, 197, 0, 32, 245, 197, 0, 32, 29, 198, 0, 32, 59, 198, 0, 32, 74, 198, 0, 32, 93, 198, 0, 32, 121, 198, 0, 32, 142, 198, 0, 32, 166, 198, 0, 32, 202, 198, 0, 32, 227, 198, 0, 32, 243, 198, 0, 32, 4, 199, 0, 32, 23, 199, 0, 32, 39, 199, 0, 32, 59, 199, 0, 32, 75, 199, 0, 32, 89, 199, 0, 32, 108, 199, 0, 32, 128, 199, 0, 32, 155, 199, 0, 32, 177, 199, 0, 32, 1, 200, 0, 32, 61, 200, 0, 32, 100, 200, 0, 32, 132, 200, 0, 32, 98, 201, 0, 32, 225, 201, 0, 32, 253, 201, 0, 32, 64, 202, 0, 32, 108, 202, 0, 32, 165, 202, 0, 32, 204, 202, 0, 32, 218, 202, 0, 32, 233, 202, 0, 32, 252, 202, 0, 32, 61, 203, 0, 32, 105, 203, 0, 32, 122, 203, 0, 32, 141, 203, 0, 32, 166, 203, 0, 32, 182, 203, 0, 32, 194, 203, 0, 32, 206, 203, 0, 32, 218, 203, 0, 32, 230, 203, 0, 32, 252, 203, 0, 32, 18, 204, 0, 32, 41, 204, 0, 32, 59, 204, 0, 32, 78, 204, 0, 32, 99, 204, 0, 32, 126, 204, 0, 32, 157, 204, 0, 32, 183, 204, 0, 32, 214, 204, 0, 32, 238, 204, 0, 32, 254, 204, 0, 32, 39, 205, 0, 32, 66, 205, 0, 32, 86, 205, 0, 32, 108, 205, 0, 32, 123, 205, 0, 32, 143, 205, 0, 32, 162, 205, 0, 32, 181, 205, 0, 32, 222, 205, 0, 32, 5, 206, 0, 32, 31, 206, 0, 32, 57, 206, 0, 32, 75, 206, 0, 32, 88, 206, 0, 32, 108, 206, 0, 32, 126, 206, 0, 32, 146, 206, 0, 32, 174, 206, 0, 32, 193, 206, 0, 32, 215, 206, 0, 32, 247, 206, 0, 32, 33, 207, 0, 32, 82, 207, 0, 32, 124, 207, 0, 32, 146, 207, 0, 32, 174, 207, 0, 32, 198, 207, 0, 32, 250, 207, 0, 32, 31, 208, 0, 32, 75, 208, 0, 32, 111, 208, 0, 32, 182, 208, 0, 32, 238, 208, 0, 32, 255, 208, 0, 32, 16, 209, 0, 32, 33, 209, 0, 32, 52, 209, 0, 32, 73, 209, 0, 32, 94, 209, 0, 32, 119, 209, 0, 32, 193, 209, 0, 32, 27, 210, 0, 32, 96, 210, 0, 32, 123, 210, 0, 32, 136, 210, 0, 32, 152, 210, 0, 32, 165, 210, 0, 32, 185, 210, 0, 32, 207, 210, 0, 32, 226, 210, 0, 32, 245, 210, 0, 32, 8, 211, 0, 32, 43, 211, 0, 32, 97, 211, 0, 32, 229, 211, 0, 32, 45, 212, 0, 32, 64, 212, 0, 32, 79, 212, 0, 32, 98, 212, 0, 32, 129, 212, 0, 32, 173, 212, 0, 32, 214, 212, 0, 32, 233, 212, 0, 32, 255, 212, 0, 32, 20, 213, 0, 32, 45, 213, 0, 32, 78, 213, 0, 32, 99, 213, 0, 32, 171, 213, 0, 32, 211, 213, 0, 32, 233, 213, 0, 32, 252, 213, 0, 32, 11, 214, 0, 32, 23, 214, 0, 32, 43, 214, 0, 32, 71, 214, 0, 32, 97, 214, 0, 32, 111, 214, 0, 32, 123, 214, 0, 32, 139, 214, 0, 32, 152, 214, 0, 32, 170, 214, 0, 32, 220, 214, 0, 32, 0, 215, 0, 32, 21, 215, 0, 32, 39, 215, 0, 32, 58, 215, 0, 32, 80, 215, 0, 32, 99, 215, 0, 32, 158, 215, 0, 32, 205, 215, 0, 32, 222, 215, 0, 32, 237, 215, 0, 32, 30, 216, 0, 32, 88, 216, 0, 32, 142, 216, 0, 32, 172, 216, 0, 32, 238, 216, 0, 32, 28, 217, 0, 32, 50, 217, 0, 32, 70, 217, 0, 32, 83, 217, 0, 32, 95, 217, 0, 32, 107, 217, 0, 32, 119, 217, 0, 32, 137, 217, 0, 32, 160, 217, 0, 32, 179, 217, 0, 32, 198, 217, 0, 32, 221, 217, 0, 32, 239, 217, 0, 32, 2, 218, 0, 32, 77, 218, 0, 32, 113, 218, 0, 32, 149, 218, 0, 32, 169, 218, 0, 32, 2, 0, 240, 22, 248, 22, 170, 170, 41, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 44, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 49, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 52, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 57, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 62, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 67, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 72, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 76, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 79, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 82, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 87, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 90, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 93, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 98, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 106, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 110, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 114, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 119, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 123, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 126, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 129, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 133, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 138, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 141, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 144, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 149, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 154, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 164, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 167, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 172, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 175, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 180, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 185, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 190, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 195, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 198, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 202, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 205, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 209, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 212, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 217, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 222, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 226, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 231, 41, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 241, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 246, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 251, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 254, 41, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 3, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 13, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 16, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 21, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 29, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 34, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 39, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 44, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 49, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 54, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 57, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 62, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 67, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 70, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 75, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 85, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 90, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 95, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 100, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 103, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 106, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 111, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 116, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 120, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 123, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 128, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 131, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 136, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 141, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 144, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 147, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 152, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 157, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 162, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 167, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 170, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 175, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 180, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 185, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 188, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 198, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 203, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 206, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 211, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 214, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 219, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 224, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 232, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 235, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 240, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 245, 42, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 250, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 254, 42, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 3, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 8, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 13, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 18, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 23, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 28, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 33, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 38, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 43, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 48, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 51, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 56, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 61, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 66, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 70, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 75, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 80, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 83, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 88, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 92, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 97, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 102, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 107, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 110, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 113, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 117, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 120, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 125, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 135, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 140, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 145, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 150, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 155, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 158, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 163, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 168, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 173, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 178, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 183, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 186, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 191, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 199, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 204, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 209, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 214, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 217, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 221, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 226, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 231, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 236, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 241, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 246, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 251, 43, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 255, 43, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 2, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 7, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 11, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 14, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 19, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 24, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 29, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 32, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 35, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 40, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 45, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 48, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 51, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 60, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 65, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 70, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 74, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 84, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 89, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 94, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 107, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 112, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 116, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 120, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 123, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 126, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 131, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 134, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 139, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 144, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 149, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 154, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 159, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 164, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 168, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 172, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 175, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 180, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 185, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 188, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 191, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 196, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 201, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 206, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 210, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 215, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 220, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 225, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 230, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 234, 44, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 238, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 243, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 248, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 253, 44, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 2, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 5, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 8, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 11, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 15, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 20, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 23, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 28, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 33, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 38, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 43, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 47, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 52, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 57, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 62, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 67, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 72, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 77, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 82, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 86, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 90, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 95, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 99, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 103, 45, 1, 144, 63, 39, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 107, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 111, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 115, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 120, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 125, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 130, 45, 1, 144, 107, 174, 0, 96, 2, 0, 240, 22, 248, 22, 170, 170, 134, 45, 1, 144, 63, 39, 0, 96, 255, 0, 11, 23, 40, 21, 101, 15, 35, 3, 38, 3, 104, 15, 116, 21, 107, 15, 110, 15, 113, 15, 116, 15, 137, 21, 119, 15, 122, 15, 148, 21, 151, 21, 125, 15, 128, 15, 131, 15, 134, 15, 137, 15, 140, 15, 143, 15, 41, 3, 149, 15, 186, 21, 189, 21, 152, 15, 24, 23, 155, 15, 30, 23, 185, 14, 158, 15, 161, 15, 226, 14, 60, 23, 211, 21, 167, 15, 170, 15, 173, 15, 84, 23, 176, 15, 214, 21, 229, 14, 179, 15, 44, 3, 232, 14, 2, 22, 235, 14, 182, 15, 185, 15, 47, 3, 188, 15, 191, 15, 194, 15, 8, 22, 50, 3, 197, 15, 200, 15, 203, 15, 53, 3, 11, 22, 36, 23, 206, 15, 56, 3, 59, 3, 62, 3, 42, 23, 65, 3, 209, 15, 212, 15, 14, 22, 215, 15, 218, 15, 221, 15, 238, 14, 224, 15, 227, 15, 20, 22, 23, 22, 230, 15, 26, 22, 236, 15, 239, 15, 38, 22, 48, 23, 242, 15, 241, 14, 44, 22, 245, 15, 248, 15, 75, 15, 251, 15, 254, 15, 50, 22, 7, 16, 10, 16, 68, 3, 53, 22, 72, 23, 64, 22, 71, 3, 74, 3, 13, 16, 90, 23, 75, 22, 16, 16, 77, 3, 80, 3, 19, 16, 22, 16, 25, 16, 28, 16, 83, 3, 31, 16, 34, 16, 86, 3, 89, 3, 244, 14, 37, 16, 40, 16, 92, 3, 43, 16, 46, 16, 49, 16, 247, 14, 52, 16, 55, 16, 58, 16, 61, 16, 95, 3, 98, 3, 64, 16, 67, 16, 70, 16, 121, 22, 76, 16, 79, 16, 82, 16, 250, 14, 85, 16, 101, 3, 132, 22, 135, 22, 104, 3, 138, 22, 97, 16, 253, 14, 100, 16, 103, 16, 106, 16, 141, 22, 109, 16, 144, 22, 112, 16, 0, 15, 115, 16, 121, 16, 124, 16, 156, 22, 127, 16, 107, 3, 3, 15, 54, 23, 159, 22, 130, 16, 6, 15, 133, 16, 178, 22, 181, 22, 9, 15, 110, 3, 59, 15, 12, 15, 139, 16, 113, 3, 184, 22, 142, 16, 145, 16, 148, 16, 116, 3, 154, 16, 157, 16, 119, 3, 160, 16, 122, 3, 163, 16, 166, 16, 187, 22, 169, 16, 172, 16, 175, 16, 178, 16, 181, 16, 184, 16, 187, 16, 190, 16, 15, 15, 190, 22, 125, 3, 193, 16, 193, 22, 196, 22, 128, 3, 199, 22, 196, 16, 202, 16, 134, 3, 208, 16, 211, 16, 214, 16, 28, 26, 220, 16, 223, 16, 226, 16, 229, 16, 94, 14, 232, 16, 235, 16, 213, 22, 238, 16, 18, 15, 216, 22, 219, 22, 244, 16, 247, 16, 250, 16, 222, 22, 137, 3, 225, 22, 132, 23, 253, 16, 2, 23, 8, 23, 15, 23, 35, 15, 21, 23, 27, 23, 33, 23, 39, 23, 45, 23, 51, 23, 57, 23, 63, 23, 69, 23, 75, 23, 81, 23, 87, 23, 93, 23, 99, 23, 105, 23, 111, 23, 117, 23, 123, 23, 129, 23, 45, 220, 0, 32, 49, 220, 0, 32, 53, 220, 0, 32, 57, 220, 0, 32, 61, 220, 0, 32, 65, 220, 0, 32, 69, 220, 0, 32, 73, 220, 0, 32, 77, 220, 0, 32, 81, 220, 0, 32, 85, 220, 0, 32, 89, 220, 0, 32, 93, 220, 0, 32, 97, 220, 0, 32, 101, 220, 0, 32, 101, 41, 1, 80, 105, 220, 0, 32, 109, 220, 0, 32, 113, 220, 0, 32, 117, 220, 0, 32, 121, 220, 0, 32, 125, 220, 0, 32, 129, 220, 0, 32, 133, 220, 0, 32, 137, 220, 0, 32, 141, 220, 0, 32, 145, 220, 0, 32, 149, 220, 0, 32, 153, 220, 0, 32, 159, 41, 1, 80, 157, 220, 0, 32, 161, 220, 0, 32, 165, 220, 0, 32, 169, 220, 0, 32, 173, 220, 0, 32, 177, 220, 0, 32, 181, 220, 0, 32, 185, 220, 0, 32, 189, 220, 0, 32, 193, 220, 0, 32, 197, 220, 0, 32, 201, 220, 0, 32, 205, 220, 0, 32, 209, 220, 0, 32, 213, 220, 0, 32, 217, 220, 0, 32, 221, 220, 0, 32, 236, 41, 1, 80, 225, 220, 0, 32, 229, 220, 0, 32, 233, 220, 0, 32, 237, 220, 0, 32, 241, 220, 0, 32, 8, 42, 1, 80, 245, 220, 0, 32, 249, 220, 0, 32, 253, 220, 0, 32, 24, 42, 1, 80, 1, 221, 0, 32, 5, 221, 0, 32, 9, 221, 0, 32, 13, 221, 0, 32, 17, 221, 0, 32, 21, 221, 0, 32, 25, 221, 0, 32, 29, 221, 0, 32, 33, 221, 0, 32, 37, 221, 0, 32, 41, 221, 0, 32, 80, 42, 1, 80, 45, 221, 0, 32, 49, 221, 0, 32, 53, 221, 0, 32, 57, 221, 0, 32, 61, 221, 0, 32, 65, 221, 0, 32, 69, 221, 0, 32, 73, 221, 0, 32, 77, 221, 0, 32, 81, 221, 0, 32, 85, 221, 0, 32, 89, 221, 0, 32, 93, 221, 0, 32, 97, 221, 0, 32, 101, 221, 0, 32, 105, 221, 0, 32, 109, 221, 0, 32, 113, 221, 0, 32, 117, 221, 0, 32, 121, 221, 0, 32, 125, 221, 0, 32, 129, 221, 0, 32, 133, 221, 0, 32, 137, 221, 0, 32, 141, 221, 0, 32, 193, 42, 1, 80, 145, 221, 0, 32, 149, 221, 0, 32, 153, 221, 0, 32, 157, 221, 0, 32, 161, 221, 0, 32, 165, 221, 0, 32, 169, 221, 0, 32, 227, 42, 1, 80, 173, 221, 0, 32, 177, 221, 0, 32, 181, 221, 0, 32, 185, 221, 0, 32, 189, 221, 0, 32, 193, 221, 0, 32, 197, 221, 0, 32, 201, 221, 0, 32, 205, 221, 0, 32, 209, 221, 0, 32, 213, 221, 0, 32, 217, 221, 0, 32, 221, 221, 0, 32, 225, 221, 0, 32, 229, 221, 0, 32, 233, 221, 0, 32, 237, 221, 0, 32, 241, 221, 0, 32, 245, 221, 0, 32, 249, 221, 0, 32, 253, 221, 0, 32, 1, 222, 0, 32, 5, 222, 0, 32, 9, 222, 0, 32, 13, 222, 0, 32, 17, 222, 0, 32, 21, 222, 0, 32, 25, 222, 0, 32, 29, 222, 0, 32, 33, 222, 0, 32, 37, 222, 0, 32, 41, 222, 0, 32, 45, 222, 0, 32, 49, 222, 0, 32, 130, 43, 1, 80, 53, 222, 0, 32, 57, 222, 0, 32, 61, 222, 0, 32, 65, 222, 0, 32, 69, 222, 0, 32, 73, 222, 0, 32, 77, 222, 0, 32, 81, 222, 0, 32, 85, 222, 0, 32, 89, 222, 0, 32, 93, 222, 0, 32, 97, 222, 0, 32, 101, 222, 0, 32, 194, 43, 1, 80, 105, 222, 0, 32, 109, 222, 0, 32, 113, 222, 0, 32, 117, 222, 0, 32, 121, 222, 0, 32, 125, 222, 0, 32, 129, 222, 0, 32, 133, 222, 0, 32, 137, 222, 0, 32, 141, 222, 0, 32, 145, 222, 0, 32, 149, 222, 0, 32, 153, 222, 0, 32, 157, 222, 0, 32, 161, 222, 0, 32, 165, 222, 0, 32, 169, 222, 0, 32, 173, 222, 0, 32, 177, 222, 0, 32, 181, 222, 0, 32, 185, 222, 0, 32, 189, 222, 0, 32, 193, 222, 0, 32, 197, 222, 0, 32, 201, 222, 0, 32, 205, 222, 0, 32, 55, 44, 1, 80, 209, 222, 0, 32, 213, 222, 0, 32, 217, 222, 0, 32, 221, 222, 0, 32, 79, 44, 1, 80, 225, 222, 0, 32, 229, 222, 0, 32, 233, 222, 0, 32, 97, 44, 1, 80, 102, 44, 1, 80, 237, 222, 0, 32, 241, 222, 0, 32, 245, 222, 0, 32, 249, 222, 0, 32, 253, 222, 0, 32, 1, 223, 0, 32, 5, 223, 0, 32, 9, 223, 0, 32, 13, 223, 0, 32, 17, 223, 0, 32, 21, 223, 0, 32, 25, 223, 0, 32, 29, 223, 0, 32, 33, 223, 0, 32, 37, 223, 0, 32, 41, 223, 0, 32, 45, 223, 0, 32, 49, 223, 0, 32, 53, 223, 0, 32, 57, 223, 0, 32, 61, 223, 0, 32, 65, 223, 0, 32, 69, 223, 0, 32, 73, 223, 0, 32, 77, 223, 0, 32, 81, 223, 0, 32, 85, 223, 0, 32, 89, 223, 0, 32, 93, 223, 0, 32, 97, 223, 0, 32, 101, 223, 0, 32, 105, 223, 0, 32, 109, 223, 0, 32, 113, 223, 0, 32, 117, 223, 0, 32, 121, 223, 0, 32, 125, 223, 0, 32, 129, 223, 0, 32, 133, 223, 0, 32, 137, 223, 0, 32, 141, 223, 0, 32, 145, 223, 0, 32, 149, 223, 0, 32, 153, 223, 0, 32, 157, 223, 0, 32, 161, 223, 0, 32, 165, 223, 0, 32, 169, 223, 0, 32, 173, 223, 0, 32, 177, 223, 0, 32, 181, 223, 0, 32, 185, 223, 0, 32, 189, 223, 0, 32, 193, 223, 0, 32, 197, 223, 0, 32, 201, 223, 0, 32, 205, 223, 0, 32, 209, 223, 0, 32, 213, 223, 0, 32, 217, 223, 0, 32, 221, 223, 0, 32, 225, 223, 0, 32, 229, 223, 0, 32, 233, 223, 0, 32, 237, 223, 0, 32, 96, 0, 11, 23, 104, 15, 116, 21, 113, 15, 119, 15, 122, 15, 128, 15, 137, 15, 143, 15, 41, 3, 189, 21, 152, 15, 30, 23, 185, 14, 161, 15, 164, 15, 226, 14, 60, 23, 84, 23, 179, 15, 44, 3, 191, 15, 194, 15, 197, 15, 200, 15, 53, 3, 206, 15, 56, 3, 42, 23, 209, 15, 212, 15, 218, 15, 238, 14, 224, 15, 75, 15, 254, 15, 4, 16, 7, 16, 10, 16, 68, 3, 53, 22, 72, 23, 13, 16, 90, 23, 77, 3, 80, 3, 19, 16, 31, 16, 86, 3, 89, 3, 244, 14, 37, 16, 43, 16, 46, 16, 55, 16, 58, 16, 61, 16, 70, 16, 73, 16, 79, 16, 101, 3, 138, 22, 88, 16, 91, 16, 253, 14, 100, 16, 103, 16, 106, 16, 115, 16, 118, 16, 127, 16, 107, 3, 130, 16, 136, 16, 9, 15, 59, 15, 12, 15, 145, 16, 148, 16, 116, 3, 160, 16, 122, 3, 166, 16, 169, 16, 196, 22, 131, 3, 208, 16, 211, 16, 223, 16, 226, 16, 94, 14, 238, 16, 216, 22, 241, 16, 244, 16, 132, 23, 170, 170, 139, 45, 1, 80, 194, 45, 1, 80, 197, 45, 1, 80, 200, 45, 1, 80, 203, 45, 1, 80, 206, 45, 1, 80, 209, 45, 1, 80, 212, 45, 1, 80, 215, 45, 1, 80, 218, 45, 1, 80, 221, 45, 1, 80, 224, 45, 1, 80, 227, 45, 1, 80, 236, 45, 1, 80, 239, 45, 1, 80, 242, 45, 1, 80, 245, 45, 1, 80, 248, 45, 1, 80, 1, 46, 1, 80, 4, 46, 1, 80, 7, 46, 1, 80, 12, 46, 1, 80, 15, 46, 1, 80, 20, 46, 1, 80, 25, 46, 1, 80, 28, 46, 1, 80, 33, 46, 1, 80, 36, 46, 1, 80, 43, 46, 1, 80, 48, 46, 1, 80, 53, 46, 1, 80, 56, 46, 1, 80, 59, 46, 1, 80, 92, 46, 1, 80, 95, 46, 1, 80, 102, 46, 1, 80, 105, 46, 1, 80, 108, 46, 1, 80, 111, 46, 1, 80, 114, 46, 1, 80, 119, 46, 1, 80, 122, 46, 1, 80, 129, 46, 1, 80, 132, 46, 1, 80, 137, 46, 1, 80, 142, 46, 1, 80, 145, 46, 1, 80, 148, 46, 1, 80, 151, 46, 1, 80, 156, 46, 1, 80, 159, 46, 1, 80, 162, 46, 1, 80, 165, 46, 1, 80, 168, 46, 1, 80, 171, 46, 1, 80, 174, 46, 1, 80, 177, 46, 1, 80, 180, 46, 1, 80, 183, 46, 1, 80, 186, 46, 1, 80, 189, 46, 1, 80, 192, 46, 1, 80, 195, 46, 1, 80, 198, 46, 1, 80, 201, 46, 1, 80, 210, 46, 1, 80, 217, 46, 1, 80, 220, 46, 1, 80, 223, 46, 1, 80, 232, 46, 1, 80, 239, 46, 1, 80, 242, 46, 1, 80, 245, 46, 1, 80, 248, 46, 1, 80, 253, 46, 1, 80, 0, 47, 1, 80, 3, 47, 1, 80, 6, 47, 1, 80, 9, 47, 1, 80, 18, 47, 1, 80, 23, 47, 1, 80, 32, 47, 1, 80, 35, 47, 1, 80, 38, 47, 1, 80, 41, 47, 1, 80, 44, 47, 1, 80, 49, 47, 1, 80, 52, 47, 1, 80, 57, 47, 1, 80, 60, 47, 1, 80, 63, 47, 1, 80, 108, 47, 1, 80, 111, 47, 1, 80, 114, 47, 1, 80, 119, 47, 1, 80, 122, 47, 1, 80, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 6, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 7, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 2, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 6, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 92, 38, 5, 156, 0, 11, 23, 101, 15, 35, 3, 38, 3, 104, 15, 116, 21, 107, 15, 110, 15, 241, 25, 116, 15, 137, 21, 119, 15, 122, 15, 151, 21, 125, 15, 128, 15, 134, 15, 137, 15, 143, 15, 41, 3, 189, 21, 152, 15, 30, 23, 185, 14, 158, 15, 161, 15, 164, 15, 226, 14, 60, 23, 84, 23, 229, 14, 179, 15, 44, 3, 232, 14, 235, 14, 188, 15, 191, 15, 194, 15, 50, 3, 197, 15, 200, 15, 203, 15, 53, 3, 36, 23, 206, 15, 56, 3, 42, 23, 65, 3, 209, 15, 212, 15, 17, 22, 218, 15, 238, 14, 227, 15, 20, 22, 23, 22, 26, 22, 38, 22, 242, 15, 241, 14, 44, 22, 75, 15, 251, 15, 254, 15, 4, 16, 7, 16, 10, 16, 68, 3, 53, 22, 72, 23, 71, 3, 74, 3, 13, 16, 90, 23, 75, 22, 16, 16, 77, 3, 80, 3, 19, 16, 22, 16, 25, 16, 86, 3, 89, 3, 37, 16, 40, 16, 92, 3, 46, 16, 49, 16, 55, 16, 58, 16, 61, 16, 95, 3, 64, 16, 67, 16, 70, 16, 76, 16, 79, 16, 250, 14, 85, 16, 101, 3, 135, 22, 88, 16, 94, 16, 253, 14, 100, 16, 103, 16, 0, 15, 115, 16, 118, 16, 121, 16, 127, 16, 107, 3, 3, 15, 54, 23, 6, 15, 133, 16, 136, 16, 9, 15, 59, 15, 139, 16, 113, 3, 184, 22, 142, 16, 145, 16, 148, 16, 116, 3, 119, 3, 160, 16, 122, 3, 166, 16, 187, 22, 169, 16, 175, 16, 15, 15, 125, 3, 131, 3, 199, 16, 205, 16, 211, 16, 214, 16, 28, 26, 223, 16, 226, 16, 229, 16, 94, 14, 232, 16, 235, 16, 213, 22, 18, 15, 219, 22, 241, 16, 250, 16, 222, 22, 137, 3, 132, 23, 0, 17, 170, 170, 1, 226, 0, 224, 8, 226, 0, 224, 15, 226, 0, 224, 22, 226, 0, 224, 29, 226, 0, 224, 36, 226, 0, 224, 43, 226, 0, 224, 50, 226, 0, 224, 57, 226, 0, 224, 64, 226, 0, 224, 71, 226, 0, 224, 78, 226, 0, 224, 85, 226, 0, 224, 92, 226, 0, 224, 99, 226, 0, 224, 106, 226, 0, 224, 113, 226, 0, 224, 120, 226, 0, 224, 127, 226, 0, 224, 134, 226, 0, 224, 141, 226, 0, 224, 148, 226, 0, 224, 155, 226, 0, 224, 162, 226, 0, 224, 169, 226, 0, 224, 176, 226, 0, 224, 183, 226, 0, 224, 190, 226, 0, 224, 197, 226, 0, 224, 204, 226, 0, 224, 211, 226, 0, 224, 218, 226, 0, 224, 225, 226, 0, 224, 232, 226, 0, 224, 239, 226, 0, 224, 246, 226, 0, 224, 253, 226, 0, 224, 4, 227, 0, 224, 11, 227, 0, 224, 18, 227, 0, 224, 25, 227, 0, 224, 32, 227, 0, 224, 39, 227, 0, 224, 46, 227, 0, 224, 53, 227, 0, 224, 60, 227, 0, 224, 67, 227, 0, 224, 74, 227, 0, 224, 81, 227, 0, 224, 88, 227, 0, 224, 95, 227, 0, 224, 102, 227, 0, 224, 109, 227, 0, 224, 116, 227, 0, 224, 123, 227, 0, 224, 130, 227, 0, 224, 137, 227, 0, 224, 144, 227, 0, 224, 151, 227, 0, 224, 158, 227, 0, 224, 165, 227, 0, 224, 172, 227, 0, 224, 179, 227, 0, 224, 186, 227, 0, 224, 193, 227, 0, 224, 200, 227, 0, 224, 207, 227, 0, 224, 214, 227, 0, 224, 221, 227, 0, 224, 228, 227, 0, 224, 235, 227, 0, 224, 242, 227, 0, 224, 249, 227, 0, 224, 0, 228, 0, 224, 7, 228, 0, 224, 14, 228, 0, 224, 21, 228, 0, 224, 28, 228, 0, 224, 35, 228, 0, 224, 42, 228, 0, 224, 49, 228, 0, 224, 56, 228, 0, 224, 63, 228, 0, 224, 70, 228, 0, 224, 77, 228, 0, 224, 84, 228, 0, 224, 91, 228, 0, 224, 98, 228, 0, 224, 105, 228, 0, 224, 112, 228, 0, 224, 119, 228, 0, 224, 126, 228, 0, 224, 133, 228, 0, 224, 140, 228, 0, 224, 147, 228, 0, 224, 154, 228, 0, 224, 161, 228, 0, 224, 168, 228, 0, 224, 175, 228, 0, 224, 182, 228, 0, 224, 189, 228, 0, 224, 196, 228, 0, 224, 203, 228, 0, 224, 210, 228, 0, 224, 217, 228, 0, 224, 224, 228, 0, 224, 231, 228, 0, 224, 238, 228, 0, 224, 245, 228, 0, 224, 252, 228, 0, 224, 3, 229, 0, 224, 10, 229, 0, 224, 17, 229, 0, 224, 24, 229, 0, 224, 31, 229, 0, 224, 38, 229, 0, 224, 45, 229, 0, 224, 52, 229, 0, 224, 59, 229, 0, 224, 66, 229, 0, 224, 73, 229, 0, 224, 80, 229, 0, 224, 87, 229, 0, 224, 94, 229, 0, 224, 101, 229, 0, 224, 108, 229, 0, 224, 115, 229, 0, 224, 122, 229, 0, 224, 129, 229, 0, 224, 136, 229, 0, 224, 143, 229, 0, 224, 150, 229, 0, 224, 157, 229, 0, 224, 164, 229, 0, 224, 171, 229, 0, 224, 178, 229, 0, 224, 185, 229, 0, 224, 192, 229, 0, 224, 199, 229, 0, 224, 206, 229, 0, 224, 213, 229, 0, 224, 220, 229, 0, 224, 227, 229, 0, 224, 234, 229, 0, 224, 241, 229, 0, 224, 248, 229, 0, 224, 255, 229, 0, 224, 6, 230, 0, 224, 13, 230, 0, 224, 20, 230, 0, 224, 27, 230, 0, 224, 34, 230, 0, 224, 41, 230, 0, 224, 48, 230, 0, 224, 55, 230, 0, 224, 62, 230, 0, 224, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 238, 14, 48, 231, 0, 224, 73, 0, 175, 11, 183, 11, 112, 13, 199, 11, 36, 9, 42, 5, 112, 5, 142, 5, 74, 8, 215, 11, 137, 12, 9, 17, 30, 13, 52, 25, 251, 13, 247, 11, 164, 12, 45, 19, 222, 9, 162, 6, 220, 6, 102, 8, 124, 10, 126, 7, 255, 11, 179, 12, 232, 7, 250, 5, 7, 12, 68, 9, 48, 4, 22, 26, 2, 6, 174, 6, 143, 4, 113, 20, 168, 8, 29, 11, 178, 6, 3, 8, 170, 23, 146, 9, 253, 9, 27, 8, 171, 13, 175, 13, 39, 12, 222, 18, 209, 7, 233, 10, 32, 11, 31, 4, 100, 13, 230, 12, 246, 12, 184, 10, 24, 19, 54, 12, 199, 13, 89, 12, 123, 7, 154, 14, 152, 12, 167, 12, 45, 10, 82, 12, 49, 13, 138, 7, 143, 9, 76, 13, 154, 9, 25, 26, 6, 14, 125, 47, 1, 144, 129, 47, 1, 144, 246, 199, 0, 96, 132, 47, 1, 144, 135, 47, 1, 144, 138, 47, 1, 144, 142, 47, 1, 144, 145, 47, 1, 144, 114, 199, 0, 96, 148, 47, 1, 144, 151, 47, 1, 144, 154, 47, 1, 144, 157, 47, 1, 144, 161, 47, 1, 144, 164, 47, 1, 144, 114, 199, 0, 96, 167, 47, 1, 144, 170, 47, 1, 144, 173, 47, 1, 144, 176, 47, 1, 144, 246, 199, 0, 96, 114, 199, 0, 96, 114, 199, 0, 96, 246, 199, 0, 96, 180, 47, 1, 144, 183, 47, 1, 144, 246, 199, 0, 96, 246, 199, 0, 96, 246, 199, 0, 96, 186, 47, 1, 144, 190, 47, 1, 144, 193, 47, 1, 144, 197, 47, 1, 144, 200, 47, 1, 144, 246, 199, 0, 96, 203, 47, 1, 144, 206, 47, 1, 144, 246, 199, 0, 96, 210, 47, 1, 144, 213, 47, 1, 144, 218, 47, 1, 144, 221, 47, 1, 144, 224, 47, 1, 144, 227, 47, 1, 144, 231, 47, 1, 144, 235, 47, 1, 144, 238, 47, 1, 144, 243, 47, 1, 144, 247, 47, 1, 144, 251, 47, 1, 144, 255, 47, 1, 144, 2, 48, 1, 144, 6, 48, 1, 144, 9, 48, 1, 144, 12, 48, 1, 144, 15, 48, 1, 144, 27, 202, 0, 96, 18, 48, 1, 144, 23, 48, 1, 144, 26, 48, 1, 144, 30, 48, 1, 144, 34, 48, 1, 144, 37, 48, 1, 144, 41, 48, 1, 144, 44, 48, 1, 144, 49, 48, 1, 144, 103, 199, 0, 96, 52, 48, 1, 144, 56, 48, 1, 144, 59, 48, 1, 144, 63, 48, 1, 144, 67, 48, 1, 144, 71, 48, 1, 144, 20, 0, 32, 0, 12, 3, 140, 3, 152, 3, 165, 3, 186, 3, 51, 4, 17, 14, 42, 14, 121, 14, 141, 14, 21, 15, 78, 15, 137, 20, 26, 21, 231, 22, 135, 23, 232, 25, 244, 25, 5, 26, 170, 170, 125, 166, 0, 32, 151, 166, 0, 32, 63, 71, 0, 96, 210, 166, 0, 128, 2, 168, 0, 128, 234, 168, 0, 32, 227, 178, 0, 32, 13, 184, 0, 32, 19, 184, 0, 32, 110, 185, 0, 32, 225, 185, 0, 32, 1, 186, 0, 32, 11, 186, 0, 32, 108, 188, 0, 32, 171, 218, 0, 32, 241, 223, 0, 32, 112, 225, 0, 32, 69, 230, 0, 32, 50, 231, 0, 32, 52, 231, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 136, 19, 0, 32, 7, 0, 0, 0, 220, 8, 0, 0, 141, 19, 0, 0, 141, 19, 0, 0, 205, 1, 0, 0, 1, 0, 0, 0, 130, 19, 0, 0, 98, 99, 112, 84, 121, 112, 101, 65, 108, 105, 97, 115, 0, 116, 122, 0, 97, 113, 97, 109, 115, 0, 99, 110, 99, 107, 103, 0, 99, 110, 104, 114, 98, 0, 99, 110, 107, 104, 103, 0, 117, 115, 110, 97, 118, 97, 106, 111, 0, 116, 121, 112, 101, 65, 108, 105, 97, 115, 0, 116, 105, 109, 101, 122, 111, 110, 101, 0, 65, 102, 114, 105, 99, 97, 58, 65, 115, 109, 97, 114, 97, 0, 65, 102, 114, 105, 99, 97, 58, 84, 105, 109, 98, 117, 107, 116, 117, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 66, 117, 101, 110, 111, 115, 95, 65, 105, 114, 101, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 67, 97, 116, 97, 109, 97, 114, 99, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 67, 111, 109, 111, 100, 82, 105, 118, 97, 100, 97, 118, 105, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 67, 111, 114, 100, 111, 98, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 74, 117, 106, 117, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 77, 101, 110, 100, 111, 122, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 116, 105, 107, 111, 107, 97, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 116, 107, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 69, 110, 115, 101, 110, 97, 100, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 70, 111, 114, 116, 95, 87, 97, 121, 110, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 73, 110, 100, 105, 97, 110, 97, 112, 111, 108, 105, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 75, 101, 110, 116, 117, 99, 107, 121, 58, 76, 111, 117, 105, 115, 118, 105, 108, 108, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 75, 110, 111, 120, 95, 73, 78, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 111, 114, 116, 111, 95, 65, 99, 114, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 111, 115, 97, 114, 105, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 104, 105, 112, 114, 111, 99, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 86, 105, 114, 103, 105, 110, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 83, 111, 117, 116, 104, 95, 80, 111, 108, 101, 0, 65, 115, 105, 97, 58, 65, 115, 104, 107, 104, 97, 98, 97, 100, 0, 65, 115, 105, 97, 58, 67, 104, 111, 110, 103, 113, 105, 110, 103, 0, 65, 115, 105, 97, 58, 67, 104, 117, 110, 103, 107, 105, 110, 103, 0, 65, 115, 105, 97, 58, 68, 97, 99, 99, 97, 0, 65, 115, 105, 97, 58, 72, 97, 114, 98, 105, 110, 0, 65, 115, 105, 97, 58, 72, 111, 95, 67, 104, 105, 95, 77, 105, 110, 104, 0, 65, 115, 105, 97, 58, 73, 115, 116, 97, 110, 98, 117, 108, 0, 65, 115, 105, 97, 58, 75, 97, 115, 104, 103, 97, 114, 0, 65, 115, 105, 97, 58, 75, 97, 116, 104, 109, 97, 110, 100, 117, 0, 65, 115, 105, 97, 58, 75, 111, 108, 107, 97, 116, 97, 0, 65, 115, 105, 97, 58, 77, 97, 99, 97, 111, 0, 65, 115, 105, 97, 58, 84, 101, 108, 95, 65, 118, 105, 118, 0, 65, 115, 105, 97, 58, 84, 104, 105, 109, 98, 117, 0, 65, 115, 105, 97, 58, 85, 106, 117, 110, 103, 95, 80, 97, 110, 100, 97, 110, 103, 0, 65, 115, 105, 97, 58, 85, 108, 97, 110, 95, 66, 97, 116, 111, 114, 0, 65, 115, 105, 97, 58, 89, 97, 110, 103, 111, 110, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 70, 97, 114, 111, 101, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 74, 97, 110, 95, 77, 97, 121, 101, 110, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 65, 67, 84, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 67, 97, 110, 98, 101, 114, 114, 97, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 76, 72, 73, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 78, 83, 87, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 78, 111, 114, 116, 104, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 81, 117, 101, 101, 110, 115, 108, 97, 110, 100, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 83, 111, 117, 116, 104, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 84, 97, 115, 109, 97, 110, 105, 97, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 86, 105, 99, 116, 111, 114, 105, 97, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 87, 101, 115, 116, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 89, 97, 110, 99, 111, 119, 105, 110, 110, 97, 0, 66, 114, 97, 122, 105, 108, 58, 65, 99, 114, 101, 0, 66, 114, 97, 122, 105, 108, 58, 68, 101, 78, 111, 114, 111, 110, 104, 97, 0, 66, 114, 97, 122, 105, 108, 58, 69, 97, 115, 116, 0, 66, 114, 97, 122, 105, 108, 58, 87, 101, 115, 116, 0, 67, 97, 110, 97, 100, 97, 58, 65, 116, 108, 97, 110, 116, 105, 99, 0, 67, 97, 110, 97, 100, 97, 58, 67, 101, 110, 116, 114, 97, 108, 0, 67, 97, 110, 97, 100, 97, 58, 69, 97, 115, 116, 45, 83, 97, 115, 107, 97, 116, 99, 104, 101, 119, 97, 110, 0, 67, 97, 110, 97, 100, 97, 58, 69, 97, 115, 116, 101, 114, 110, 0, 67, 97, 110, 97, 100, 97, 58, 77, 111, 117, 110, 116, 97, 105, 110, 0, 67, 97, 110, 97, 100, 97, 58, 78, 101, 119, 102, 111, 117, 110, 100, 108, 97, 110, 100, 0, 67, 97, 110, 97, 100, 97, 58, 80, 97, 99, 105, 102, 105, 99, 0, 67, 97, 110, 97, 100, 97, 58, 83, 97, 115, 107, 97, 116, 99, 104, 101, 119, 97, 110, 0, 67, 97, 110, 97, 100, 97, 58, 89, 117, 107, 111, 110, 0, 67, 104, 105, 108, 101, 58, 67, 111, 110, 116, 105, 110, 101, 110, 116, 97, 108, 0, 67, 104, 105, 108, 101, 58, 69, 97, 115, 116, 101, 114, 73, 115, 108, 97, 110, 100, 0, 69, 116, 99, 58, 71, 77, 84, 43, 48, 0, 69, 116, 99, 58, 71, 77, 84, 45, 48, 0, 69, 116, 99, 58, 71, 77, 84, 48, 0, 69, 116, 99, 58, 71, 114, 101, 101, 110, 119, 105, 99, 104, 0, 69, 116, 99, 58, 85, 67, 84, 0, 69, 116, 99, 58, 85, 110, 105, 118, 101, 114, 115, 97, 108, 0, 69, 116, 99, 58, 90, 117, 108, 117, 0, 69, 117, 114, 111, 112, 101, 58, 66, 101, 108, 102, 97, 115, 116, 0, 69, 117, 114, 111, 112, 101, 58, 78, 105, 99, 111, 115, 105, 97, 0, 69, 117, 114, 111, 112, 101, 58, 84, 105, 114, 97, 115, 112, 111, 108, 0, 77, 101, 120, 105, 99, 111, 58, 66, 97, 106, 97, 78, 111, 114, 116, 101, 0, 77, 101, 120, 105, 99, 111, 58, 66, 97, 106, 97, 83, 117, 114, 0, 77, 101, 120, 105, 99, 111, 58, 71, 101, 110, 101, 114, 97, 108, 0, 80, 97, 99, 105, 102, 105, 99, 58, 67, 104, 117, 117, 107, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 111, 104, 110, 112, 101, 105, 0, 80, 97, 99, 105, 102, 105, 99, 58, 83, 97, 109, 111, 97, 0, 80, 97, 99, 105, 102, 105, 99, 58, 89, 97, 112, 0, 85, 83, 58, 65, 108, 97, 115, 107, 97, 0, 85, 83, 58, 65, 108, 101, 117, 116, 105, 97, 110, 0, 85, 83, 58, 65, 114, 105, 122, 111, 110, 97, 0, 85, 83, 58, 67, 101, 110, 116, 114, 97, 108, 0, 85, 83, 58, 69, 97, 115, 116, 45, 73, 110, 100, 105, 97, 110, 97, 0, 85, 83, 58, 69, 97, 115, 116, 101, 114, 110, 0, 85, 83, 58, 72, 97, 119, 97, 105, 105, 0, 85, 83, 58, 73, 110, 100, 105, 97, 110, 97, 45, 83, 116, 97, 114, 107, 101, 0, 85, 83, 58, 77, 105, 99, 104, 105, 103, 97, 110, 0, 85, 83, 58, 77, 111, 117, 110, 116, 97, 105, 110, 0, 85, 83, 58, 80, 97, 99, 105, 102, 105, 99, 0, 85, 83, 58, 80, 97, 99, 105, 102, 105, 99, 45, 78, 101, 119, 0, 85, 83, 58, 83, 97, 109, 111, 97, 0, 67, 117, 98, 97, 0, 69, 83, 84, 0, 69, 103, 121, 112, 116, 0, 71, 66, 0, 71, 66, 45, 69, 105, 114, 101, 0, 72, 83, 84, 0, 72, 111, 110, 103, 107, 111, 110, 103, 0, 73, 99, 101, 108, 97, 110, 100, 0, 73, 114, 97, 110, 0, 73, 115, 114, 97, 101, 108, 0, 74, 97, 112, 97, 110, 0, 76, 105, 98, 121, 97, 0, 77, 83, 84, 0, 78, 90, 0, 78, 90, 45, 67, 72, 65, 84, 0, 78, 97, 118, 97, 106, 111, 0, 80, 82, 67, 0, 80, 111, 108, 97, 110, 100, 0, 80, 111, 114, 116, 117, 103, 97, 108, 0, 82, 79, 67, 0, 82, 79, 75, 0, 84, 117, 114, 107, 101, 121, 0, 87, 45, 83, 85, 0, 116, 121, 112, 101, 77, 97, 112, 0, 65, 102, 114, 105, 99, 97, 58, 65, 98, 105, 100, 106, 97, 110, 0, 65, 102, 114, 105, 99, 97, 58, 65, 99, 99, 114, 97, 0, 65, 102, 114, 105, 99, 97, 58, 65, 100, 100, 105, 115, 95, 65, 98, 97, 98, 97, 0, 65, 102, 114, 105, 99, 97, 58, 65, 108, 103, 105, 101, 114, 115, 0, 65, 102, 114, 105, 99, 97, 58, 65, 115, 109, 101, 114, 97, 0, 65, 102, 114, 105, 99, 97, 58, 66, 97, 109, 97, 107, 111, 0, 65, 102, 114, 105, 99, 97, 58, 66, 97, 110, 103, 117, 105, 0, 65, 102, 114, 105, 99, 97, 58, 66, 97, 110, 106, 117, 108, 0, 65, 102, 114, 105, 99, 97, 58, 66, 105, 115, 115, 97, 117, 0, 65, 102, 114, 105, 99, 97, 58, 66, 108, 97, 110, 116, 121, 114, 101, 0, 65, 102, 114, 105, 99, 97, 58, 66, 114, 97, 122, 122, 97, 118, 105, 108, 108, 101, 0, 65, 102, 114, 105, 99, 97, 58, 66, 117, 106, 117, 109, 98, 117, 114, 97, 0, 65, 102, 114, 105, 99, 97, 58, 67, 97, 105, 114, 111, 0, 65, 102, 114, 105, 99, 97, 58, 67, 97, 115, 97, 98, 108, 97, 110, 99, 97, 0, 65, 102, 114, 105, 99, 97, 58, 67, 101, 117, 116, 97, 0, 65, 102, 114, 105, 99, 97, 58, 67, 111, 110, 97, 107, 114, 121, 0, 65, 102, 114, 105, 99, 97, 58, 68, 97, 107, 97, 114, 0, 65, 102, 114, 105, 99, 97, 58, 68, 97, 114, 95, 101, 115, 95, 83, 97, 108, 97, 97, 109, 0, 65, 102, 114, 105, 99, 97, 58, 68, 106, 105, 98, 111, 117, 116, 105, 0, 65, 102, 114, 105, 99, 97, 58, 68, 111, 117, 97, 108, 97, 0, 65, 102, 114, 105, 99, 97, 58, 69, 108, 95, 65, 97, 105, 117, 110, 0, 65, 102, 114, 105, 99, 97, 58, 70, 114, 101, 101, 116, 111, 119, 110, 0, 65, 102, 114, 105, 99, 97, 58, 71, 97, 98, 111, 114, 111, 110, 101, 0, 65, 102, 114, 105, 99, 97, 58, 72, 97, 114, 97, 114, 101, 0, 65, 102, 114, 105, 99, 97, 58, 74, 111, 104, 97, 110, 110, 101, 115, 98, 117, 114, 103, 0, 65, 102, 114, 105, 99, 97, 58, 74, 117, 98, 97, 0, 65, 102, 114, 105, 99, 97, 58, 75, 97, 109, 112, 97, 108, 97, 0, 65, 102, 114, 105, 99, 97, 58, 75, 104, 97, 114, 116, 111, 117, 109, 0, 65, 102, 114, 105, 99, 97, 58, 75, 105, 103, 97, 108, 105, 0, 65, 102, 114, 105, 99, 97, 58, 75, 105, 110, 115, 104, 97, 115, 97, 0, 65, 102, 114, 105, 99, 97, 58, 76, 97, 103, 111, 115, 0, 65, 102, 114, 105, 99, 97, 58, 76, 105, 98, 114, 101, 118, 105, 108, 108, 101, 0, 65, 102, 114, 105, 99, 97, 58, 76, 111, 109, 101, 0, 65, 102, 114, 105, 99, 97, 58, 76, 117, 97, 110, 100, 97, 0, 65, 102, 114, 105, 99, 97, 58, 76, 117, 98, 117, 109, 98, 97, 115, 104, 105, 0, 65, 102, 114, 105, 99, 97, 58, 76, 117, 115, 97, 107, 97, 0, 65, 102, 114, 105, 99, 97, 58, 77, 97, 108, 97, 98, 111, 0, 65, 102, 114, 105, 99, 97, 58, 77, 97, 112, 117, 116, 111, 0, 65, 102, 114, 105, 99, 97, 58, 77, 97, 115, 101, 114, 117, 0, 65, 102, 114, 105, 99, 97, 58, 77, 98, 97, 98, 97, 110, 101, 0, 65, 102, 114, 105, 99, 97, 58, 77, 111, 103, 97, 100, 105, 115, 104, 117, 0, 65, 102, 114, 105, 99, 97, 58, 77, 111, 110, 114, 111, 118, 105, 97, 0, 65, 102, 114, 105, 99, 97, 58, 78, 97, 105, 114, 111, 98, 105, 0, 65, 102, 114, 105, 99, 97, 58, 78, 100, 106, 97, 109, 101, 110, 97, 0, 65, 102, 114, 105, 99, 97, 58, 78, 105, 97, 109, 101, 121, 0, 65, 102, 114, 105, 99, 97, 58, 78, 111, 117, 97, 107, 99, 104, 111, 116, 116, 0, 65, 102, 114, 105, 99, 97, 58, 79, 117, 97, 103, 97, 100, 111, 117, 103, 111, 117, 0, 65, 102, 114, 105, 99, 97, 58, 80, 111, 114, 116, 111, 45, 78, 111, 118, 111, 0, 65, 102, 114, 105, 99, 97, 58, 83, 97, 111, 95, 84, 111, 109, 101, 0, 65, 102, 114, 105, 99, 97, 58, 84, 114, 105, 112, 111, 108, 105, 0, 65, 102, 114, 105, 99, 97, 58, 84, 117, 110, 105, 115, 0, 65, 102, 114, 105, 99, 97, 58, 87, 105, 110, 100, 104, 111, 101, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 100, 97, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 110, 99, 104, 111, 114, 97, 103, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 110, 103, 117, 105, 108, 108, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 110, 116, 105, 103, 117, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 97, 103, 117, 97, 105, 110, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 76, 97, 95, 82, 105, 111, 106, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 82, 105, 111, 95, 71, 97, 108, 108, 101, 103, 111, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 83, 97, 108, 116, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 83, 97, 110, 95, 74, 117, 97, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 83, 97, 110, 95, 76, 117, 105, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 84, 117, 99, 117, 109, 97, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 103, 101, 110, 116, 105, 110, 97, 58, 85, 115, 104, 117, 97, 105, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 114, 117, 98, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 65, 115, 117, 110, 99, 105, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 97, 104, 105, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 97, 104, 105, 97, 95, 66, 97, 110, 100, 101, 114, 97, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 97, 114, 98, 97, 100, 111, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 101, 108, 101, 109, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 101, 108, 105, 122, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 108, 97, 110, 99, 45, 83, 97, 98, 108, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 111, 97, 95, 86, 105, 115, 116, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 111, 103, 111, 116, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 111, 105, 115, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 66, 117, 101, 110, 111, 115, 95, 65, 105, 114, 101, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 109, 98, 114, 105, 100, 103, 101, 95, 66, 97, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 109, 112, 111, 95, 71, 114, 97, 110, 100, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 110, 99, 117, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 114, 97, 99, 97, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 116, 97, 109, 97, 114, 99, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 121, 101, 110, 110, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 97, 121, 109, 97, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 104, 105, 99, 97, 103, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 104, 105, 104, 117, 97, 104, 117, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 111, 114, 97, 108, 95, 72, 97, 114, 98, 111, 117, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 111, 114, 100, 111, 98, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 111, 115, 116, 97, 95, 82, 105, 99, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 114, 101, 115, 116, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 117, 105, 97, 98, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 67, 117, 114, 97, 99, 97, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 97, 110, 109, 97, 114, 107, 115, 104, 97, 118, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 97, 119, 115, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 97, 119, 115, 111, 110, 95, 67, 114, 101, 101, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 101, 110, 118, 101, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 101, 116, 114, 111, 105, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 68, 111, 109, 105, 110, 105, 99, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 69, 100, 109, 111, 110, 116, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 69, 105, 114, 117, 110, 101, 112, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 69, 108, 95, 83, 97, 108, 118, 97, 100, 111, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 70, 111, 114, 116, 95, 78, 101, 108, 115, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 70, 111, 114, 116, 97, 108, 101, 122, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 108, 97, 99, 101, 95, 66, 97, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 111, 100, 116, 104, 97, 98, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 111, 111, 115, 101, 95, 66, 97, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 114, 97, 110, 100, 95, 84, 117, 114, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 114, 101, 110, 97, 100, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 117, 97, 100, 101, 108, 111, 117, 112, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 117, 97, 116, 101, 109, 97, 108, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 117, 97, 121, 97, 113, 117, 105, 108, 0, 65, 109, 101, 114, 105, 99, 97, 58, 71, 117, 121, 97, 110, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 72, 97, 108, 105, 102, 97, 120, 0, 65, 109, 101, 114, 105, 99, 97, 58, 72, 97, 118, 97, 110, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 72, 101, 114, 109, 111, 115, 105, 108, 108, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 75, 110, 111, 120, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 77, 97, 114, 101, 110, 103, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 80, 101, 116, 101, 114, 115, 98, 117, 114, 103, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 84, 101, 108, 108, 95, 67, 105, 116, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 86, 101, 118, 97, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 86, 105, 110, 99, 101, 110, 110, 101, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 58, 87, 105, 110, 97, 109, 97, 99, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 100, 105, 97, 110, 97, 112, 111, 108, 105, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 110, 117, 118, 105, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 73, 113, 97, 108, 117, 105, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 74, 97, 109, 97, 105, 99, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 74, 117, 106, 117, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 74, 117, 110, 101, 97, 117, 0, 65, 109, 101, 114, 105, 99, 97, 58, 75, 101, 110, 116, 117, 99, 107, 121, 58, 77, 111, 110, 116, 105, 99, 101, 108, 108, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 75, 114, 97, 108, 101, 110, 100, 105, 106, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 76, 97, 95, 80, 97, 122, 0, 65, 109, 101, 114, 105, 99, 97, 58, 76, 105, 109, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 76, 111, 115, 95, 65, 110, 103, 101, 108, 101, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 76, 111, 117, 105, 115, 118, 105, 108, 108, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 76, 111, 119, 101, 114, 95, 80, 114, 105, 110, 99, 101, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 99, 101, 105, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 110, 97, 103, 117, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 110, 97, 117, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 114, 105, 103, 111, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 114, 116, 105, 110, 105, 113, 117, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 116, 97, 109, 111, 114, 111, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 97, 122, 97, 116, 108, 97, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 101, 110, 100, 111, 122, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 101, 110, 111, 109, 105, 110, 101, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 101, 114, 105, 100, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 101, 116, 108, 97, 107, 97, 116, 108, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 101, 120, 105, 99, 111, 95, 67, 105, 116, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 105, 113, 117, 101, 108, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 111, 110, 99, 116, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 111, 110, 116, 101, 114, 114, 101, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 111, 110, 116, 101, 118, 105, 100, 101, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 111, 110, 116, 114, 101, 97, 108, 0, 65, 109, 101, 114, 105, 99, 97, 58, 77, 111, 110, 116, 115, 101, 114, 114, 97, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 97, 115, 115, 97, 117, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 101, 119, 95, 89, 111, 114, 107, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 105, 112, 105, 103, 111, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 111, 109, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 111, 114, 111, 110, 104, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 111, 114, 116, 104, 95, 68, 97, 107, 111, 116, 97, 58, 66, 101, 117, 108, 97, 104, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 111, 114, 116, 104, 95, 68, 97, 107, 111, 116, 97, 58, 67, 101, 110, 116, 101, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 78, 111, 114, 116, 104, 95, 68, 97, 107, 111, 116, 97, 58, 78, 101, 119, 95, 83, 97, 108, 101, 109, 0, 65, 109, 101, 114, 105, 99, 97, 58, 79, 106, 105, 110, 97, 103, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 97, 110, 97, 109, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 97, 110, 103, 110, 105, 114, 116, 117, 110, 103, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 97, 114, 97, 109, 97, 114, 105, 98, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 104, 111, 101, 110, 105, 120, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 111, 114, 116, 45, 97, 117, 45, 80, 114, 105, 110, 99, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 111, 114, 116, 95, 111, 102, 95, 83, 112, 97, 105, 110, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 111, 114, 116, 111, 95, 86, 101, 108, 104, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 117, 101, 114, 116, 111, 95, 82, 105, 99, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 80, 117, 110, 116, 97, 95, 65, 114, 101, 110, 97, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 97, 105, 110, 121, 95, 82, 105, 118, 101, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 97, 110, 107, 105, 110, 95, 73, 110, 108, 101, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 101, 99, 105, 102, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 101, 103, 105, 110, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 101, 115, 111, 108, 117, 116, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 82, 105, 111, 95, 66, 114, 97, 110, 99, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 97, 110, 116, 97, 95, 73, 115, 97, 98, 101, 108, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 97, 110, 116, 97, 114, 101, 109, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 97, 110, 116, 105, 97, 103, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 97, 110, 116, 111, 95, 68, 111, 109, 105, 110, 103, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 97, 111, 95, 80, 97, 117, 108, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 99, 111, 114, 101, 115, 98, 121, 115, 117, 110, 100, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 105, 116, 107, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 66, 97, 114, 116, 104, 101, 108, 101, 109, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 74, 111, 104, 110, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 75, 105, 116, 116, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 76, 117, 99, 105, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 84, 104, 111, 109, 97, 115, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 116, 95, 86, 105, 110, 99, 101, 110, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 83, 119, 105, 102, 116, 95, 67, 117, 114, 114, 101, 110, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 101, 103, 117, 99, 105, 103, 97, 108, 112, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 104, 117, 108, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 104, 117, 110, 100, 101, 114, 95, 66, 97, 121, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 105, 106, 117, 97, 110, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 111, 114, 111, 110, 116, 111, 0, 65, 109, 101, 114, 105, 99, 97, 58, 84, 111, 114, 116, 111, 108, 97, 0, 65, 109, 101, 114, 105, 99, 97, 58, 86, 97, 110, 99, 111, 117, 118, 101, 114, 0, 65, 109, 101, 114, 105, 99, 97, 58, 87, 104, 105, 116, 101, 104, 111, 114, 115, 101, 0, 65, 109, 101, 114, 105, 99, 97, 58, 87, 105, 110, 110, 105, 112, 101, 103, 0, 65, 109, 101, 114, 105, 99, 97, 58, 89, 97, 107, 117, 116, 97, 116, 0, 65, 109, 101, 114, 105, 99, 97, 58, 89, 101, 108, 108, 111, 119, 107, 110, 105, 102, 101, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 67, 97, 115, 101, 121, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 68, 97, 118, 105, 115, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 68, 117, 109, 111, 110, 116, 68, 85, 114, 118, 105, 108, 108, 101, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 77, 97, 99, 113, 117, 97, 114, 105, 101, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 77, 97, 119, 115, 111, 110, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 77, 99, 77, 117, 114, 100, 111, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 80, 97, 108, 109, 101, 114, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 82, 111, 116, 104, 101, 114, 97, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 83, 121, 111, 119, 97, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 84, 114, 111, 108, 108, 0, 65, 110, 116, 97, 114, 99, 116, 105, 99, 97, 58, 86, 111, 115, 116, 111, 107, 0, 65, 114, 99, 116, 105, 99, 58, 76, 111, 110, 103, 121, 101, 97, 114, 98, 121, 101, 110, 0, 65, 115, 105, 97, 58, 65, 100, 101, 110, 0, 65, 115, 105, 97, 58, 65, 108, 109, 97, 116, 121, 0, 65, 115, 105, 97, 58, 65, 109, 109, 97, 110, 0, 65, 115, 105, 97, 58, 65, 110, 97, 100, 121, 114, 0, 65, 115, 105, 97, 58, 65, 113, 116, 97, 117, 0, 65, 115, 105, 97, 58, 65, 113, 116, 111, 98, 101, 0, 65, 115, 105, 97, 58, 65, 115, 104, 103, 97, 98, 97, 116, 0, 65, 115, 105, 97, 58, 65, 116, 121, 114, 97, 117, 0, 65, 115, 105, 97, 58, 66, 97, 103, 104, 100, 97, 100, 0, 65, 115, 105, 97, 58, 66, 97, 104, 114, 97, 105, 110, 0, 65, 115, 105, 97, 58, 66, 97, 107, 117, 0, 65, 115, 105, 97, 58, 66, 97, 110, 103, 107, 111, 107, 0, 65, 115, 105, 97, 58, 66, 97, 114, 110, 97, 117, 108, 0, 65, 115, 105, 97, 58, 66, 101, 105, 114, 117, 116, 0, 65, 115, 105, 97, 58, 66, 105, 115, 104, 107, 101, 107, 0, 65, 115, 105, 97, 58, 66, 114, 117, 110, 101, 105, 0, 65, 115, 105, 97, 58, 67, 97, 108, 99, 117, 116, 116, 97, 0, 65, 115, 105, 97, 58, 67, 104, 105, 116, 97, 0, 65, 115, 105, 97, 58, 67, 104, 111, 105, 98, 97, 108, 115, 97, 110, 0, 65, 115, 105, 97, 58, 67, 111, 108, 111, 109, 98, 111, 0, 65, 115, 105, 97, 58, 68, 97, 109, 97, 115, 99, 117, 115, 0, 65, 115, 105, 97, 58, 68, 104, 97, 107, 97, 0, 65, 115, 105, 97, 58, 68, 105, 108, 105, 0, 65, 115, 105, 97, 58, 68, 117, 98, 97, 105, 0, 65, 115, 105, 97, 58, 68, 117, 115, 104, 97, 110, 98, 101, 0, 65, 115, 105, 97, 58, 70, 97, 109, 97, 103, 117, 115, 116, 97, 0, 65, 115, 105, 97, 58, 71, 97, 122, 97, 0, 65, 115, 105, 97, 58, 72, 101, 98, 114, 111, 110, 0, 65, 115, 105, 97, 58, 72, 111, 110, 103, 95, 75, 111, 110, 103, 0, 65, 115, 105, 97, 58, 72, 111, 118, 100, 0, 65, 115, 105, 97, 58, 73, 114, 107, 117, 116, 115, 107, 0, 65, 115, 105, 97, 58, 74, 97, 107, 97, 114, 116, 97, 0, 65, 115, 105, 97, 58, 74, 97, 121, 97, 112, 117, 114, 97, 0, 65, 115, 105, 97, 58, 74, 101, 114, 117, 115, 97, 108, 101, 109, 0, 65, 115, 105, 97, 58, 75, 97, 98, 117, 108, 0, 65, 115, 105, 97, 58, 75, 97, 109, 99, 104, 97, 116, 107, 97, 0, 65, 115, 105, 97, 58, 75, 97, 114, 97, 99, 104, 105, 0, 65, 115, 105, 97, 58, 75, 97, 116, 109, 97, 110, 100, 117, 0, 65, 115, 105, 97, 58, 75, 104, 97, 110, 100, 121, 103, 97, 0, 65, 115, 105, 97, 58, 75, 114, 97, 115, 110, 111, 121, 97, 114, 115, 107, 0, 65, 115, 105, 97, 58, 75, 117, 97, 108, 97, 95, 76, 117, 109, 112, 117, 114, 0, 65, 115, 105, 97, 58, 75, 117, 99, 104, 105, 110, 103, 0, 65, 115, 105, 97, 58, 75, 117, 119, 97, 105, 116, 0, 65, 115, 105, 97, 58, 77, 97, 99, 97, 117, 0, 65, 115, 105, 97, 58, 77, 97, 103, 97, 100, 97, 110, 0, 65, 115, 105, 97, 58, 77, 97, 107, 97, 115, 115, 97, 114, 0, 65, 115, 105, 97, 58, 77, 97, 110, 105, 108, 97, 0, 65, 115, 105, 97, 58, 77, 117, 115, 99, 97, 116, 0, 65, 115, 105, 97, 58, 78, 105, 99, 111, 115, 105, 97, 0, 65, 115, 105, 97, 58, 78, 111, 118, 111, 107, 117, 122, 110, 101, 116, 115, 107, 0, 65, 115, 105, 97, 58, 78, 111, 118, 111, 115, 105, 98, 105, 114, 115, 107, 0, 65, 115, 105, 97, 58, 79, 109, 115, 107, 0, 65, 115, 105, 97, 58, 79, 114, 97, 108, 0, 65, 115, 105, 97, 58, 80, 104, 110, 111, 109, 95, 80, 101, 110, 104, 0, 65, 115, 105, 97, 58, 80, 111, 110, 116, 105, 97, 110, 97, 107, 0, 65, 115, 105, 97, 58, 80, 121, 111, 110, 103, 121, 97, 110, 103, 0, 65, 115, 105, 97, 58, 81, 97, 116, 97, 114, 0, 65, 115, 105, 97, 58, 81, 111, 115, 116, 97, 110, 97, 121, 0, 65, 115, 105, 97, 58, 81, 121, 122, 121, 108, 111, 114, 100, 97, 0, 65, 115, 105, 97, 58, 82, 97, 110, 103, 111, 111, 110, 0, 65, 115, 105, 97, 58, 82, 105, 121, 97, 100, 104, 0, 65, 115, 105, 97, 58, 83, 97, 105, 103, 111, 110, 0, 65, 115, 105, 97, 58, 83, 97, 107, 104, 97, 108, 105, 110, 0, 65, 115, 105, 97, 58, 83, 97, 109, 97, 114, 107, 97, 110, 100, 0, 65, 115, 105, 97, 58, 83, 101, 111, 117, 108, 0, 65, 115, 105, 97, 58, 83, 104, 97, 110, 103, 104, 97, 105, 0, 65, 115, 105, 97, 58, 83, 105, 110, 103, 97, 112, 111, 114, 101, 0, 65, 115, 105, 97, 58, 83, 114, 101, 100, 110, 101, 107, 111, 108, 121, 109, 115, 107, 0, 65, 115, 105, 97, 58, 84, 97, 105, 112, 101, 105, 0, 65, 115, 105, 97, 58, 84, 97, 115, 104, 107, 101, 110, 116, 0, 65, 115, 105, 97, 58, 84, 98, 105, 108, 105, 115, 105, 0, 65, 115, 105, 97, 58, 84, 101, 104, 114, 97, 110, 0, 65, 115, 105, 97, 58, 84, 104, 105, 109, 112, 104, 117, 0, 65, 115, 105, 97, 58, 84, 111, 107, 121, 111, 0, 65, 115, 105, 97, 58, 84, 111, 109, 115, 107, 0, 65, 115, 105, 97, 58, 85, 108, 97, 97, 110, 98, 97, 97, 116, 97, 114, 0, 65, 115, 105, 97, 58, 85, 114, 117, 109, 113, 105, 0, 65, 115, 105, 97, 58, 85, 115, 116, 45, 78, 101, 114, 97, 0, 65, 115, 105, 97, 58, 86, 105, 101, 110, 116, 105, 97, 110, 101, 0, 65, 115, 105, 97, 58, 86, 108, 97, 100, 105, 118, 111, 115, 116, 111, 107, 0, 65, 115, 105, 97, 58, 89, 97, 107, 117, 116, 115, 107, 0, 65, 115, 105, 97, 58, 89, 101, 107, 97, 116, 101, 114, 105, 110, 98, 117, 114, 103, 0, 65, 115, 105, 97, 58, 89, 101, 114, 101, 118, 97, 110, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 65, 122, 111, 114, 101, 115, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 66, 101, 114, 109, 117, 100, 97, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 67, 97, 110, 97, 114, 121, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 67, 97, 112, 101, 95, 86, 101, 114, 100, 101, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 70, 97, 101, 114, 111, 101, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 77, 97, 100, 101, 105, 114, 97, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 82, 101, 121, 107, 106, 97, 118, 105, 107, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 83, 111, 117, 116, 104, 95, 71, 101, 111, 114, 103, 105, 97, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 83, 116, 95, 72, 101, 108, 101, 110, 97, 0, 65, 116, 108, 97, 110, 116, 105, 99, 58, 83, 116, 97, 110, 108, 101, 121, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 65, 100, 101, 108, 97, 105, 100, 101, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 66, 114, 105, 115, 98, 97, 110, 101, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 66, 114, 111, 107, 101, 110, 95, 72, 105, 108, 108, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 67, 117, 114, 114, 105, 101, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 68, 97, 114, 119, 105, 110, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 69, 117, 99, 108, 97, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 72, 111, 98, 97, 114, 116, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 76, 105, 110, 100, 101, 109, 97, 110, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 76, 111, 114, 100, 95, 72, 111, 119, 101, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 77, 101, 108, 98, 111, 117, 114, 110, 101, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 80, 101, 114, 116, 104, 0, 65, 117, 115, 116, 114, 97, 108, 105, 97, 58, 83, 121, 100, 110, 101, 121, 0, 69, 116, 99, 58, 71, 77, 84, 0, 69, 116, 99, 58, 71, 77, 84, 43, 49, 0, 69, 116, 99, 58, 71, 77, 84, 43, 49, 48, 0, 69, 116, 99, 58, 71, 77, 84, 43, 49, 49, 0, 69, 116, 99, 58, 71, 77, 84, 43, 49, 50, 0, 69, 116, 99, 58, 71, 77, 84, 43, 50, 0, 69, 116, 99, 58, 71, 77, 84, 43, 51, 0, 69, 116, 99, 58, 71, 77, 84, 43, 52, 0, 69, 116, 99, 58, 71, 77, 84, 43, 53, 0, 69, 116, 99, 58, 71, 77, 84, 43, 54, 0, 69, 116, 99, 58, 71, 77, 84, 43, 55, 0, 69, 116, 99, 58, 71, 77, 84, 43, 56, 0, 69, 116, 99, 58, 71, 77, 84, 43, 57, 0, 69, 116, 99, 58, 71, 77, 84, 45, 49, 0, 69, 116, 99, 58, 71, 77, 84, 45, 49, 48, 0, 69, 116, 99, 58, 71, 77, 84, 45, 49, 49, 0, 69, 116, 99, 58, 71, 77, 84, 45, 49, 50, 0, 69, 116, 99, 58, 71, 77, 84, 45, 49, 51, 0, 69, 116, 99, 58, 71, 77, 84, 45, 49, 52, 0, 69, 116, 99, 58, 71, 77, 84, 45, 50, 0, 69, 116, 99, 58, 71, 77, 84, 45, 51, 0, 69, 116, 99, 58, 71, 77, 84, 45, 52, 0, 69, 116, 99, 58, 71, 77, 84, 45, 53, 0, 69, 116, 99, 58, 71, 77, 84, 45, 54, 0, 69, 116, 99, 58, 71, 77, 84, 45, 55, 0, 69, 116, 99, 58, 71, 77, 84, 45, 56, 0, 69, 116, 99, 58, 71, 77, 84, 45, 57, 0, 69, 116, 99, 58, 85, 84, 67, 0, 69, 116, 99, 58, 85, 110, 107, 110, 111, 119, 110, 0, 69, 117, 114, 111, 112, 101, 58, 65, 109, 115, 116, 101, 114, 100, 97, 109, 0, 69, 117, 114, 111, 112, 101, 58, 65, 110, 100, 111, 114, 114, 97, 0, 69, 117, 114, 111, 112, 101, 58, 65, 115, 116, 114, 97, 107, 104, 97, 110, 0, 69, 117, 114, 111, 112, 101, 58, 65, 116, 104, 101, 110, 115, 0, 69, 117, 114, 111, 112, 101, 58, 66, 101, 108, 103, 114, 97, 100, 101, 0, 69, 117, 114, 111, 112, 101, 58, 66, 101, 114, 108, 105, 110, 0, 69, 117, 114, 111, 112, 101, 58, 66, 114, 97, 116, 105, 115, 108, 97, 118, 97, 0, 69, 117, 114, 111, 112, 101, 58, 66, 114, 117, 115, 115, 101, 108, 115, 0, 69, 117, 114, 111, 112, 101, 58, 66, 117, 99, 104, 97, 114, 101, 115, 116, 0, 69, 117, 114, 111, 112, 101, 58, 66, 117, 100, 97, 112, 101, 115, 116, 0, 69, 117, 114, 111, 112, 101, 58, 66, 117, 115, 105, 110, 103, 101, 110, 0, 69, 117, 114, 111, 112, 101, 58, 67, 104, 105, 115, 105, 110, 97, 117, 0, 69, 117, 114, 111, 112, 101, 58, 67, 111, 112, 101, 110, 104, 97, 103, 101, 110, 0, 69, 117, 114, 111, 112, 101, 58, 68, 117, 98, 108, 105, 110, 0, 69, 117, 114, 111, 112, 101, 58, 71, 105, 98, 114, 97, 108, 116, 97, 114, 0, 69, 117, 114, 111, 112, 101, 58, 71, 117, 101, 114, 110, 115, 101, 121, 0, 69, 117, 114, 111, 112, 101, 58, 72, 101, 108, 115, 105, 110, 107, 105, 0, 69, 117, 114, 111, 112, 101, 58, 73, 115, 108, 101, 95, 111, 102, 95, 77, 97, 110, 0, 69, 117, 114, 111, 112, 101, 58, 73, 115, 116, 97, 110, 98, 117, 108, 0, 69, 117, 114, 111, 112, 101, 58, 74, 101, 114, 115, 101, 121, 0, 69, 117, 114, 111, 112, 101, 58, 75, 97, 108, 105, 110, 105, 110, 103, 114, 97, 100, 0, 69, 117, 114, 111, 112, 101, 58, 75, 105, 101, 118, 0, 69, 117, 114, 111, 112, 101, 58, 75, 105, 114, 111, 118, 0, 69, 117, 114, 111, 112, 101, 58, 76, 105, 115, 98, 111, 110, 0, 69, 117, 114, 111, 112, 101, 58, 76, 106, 117, 98, 108, 106, 97, 110, 97, 0, 69, 117, 114, 111, 112, 101, 58, 76, 111, 110, 100, 111, 110, 0, 69, 117, 114, 111, 112, 101, 58, 76, 117, 120, 101, 109, 98, 111, 117, 114, 103, 0, 69, 117, 114, 111, 112, 101, 58, 77, 97, 100, 114, 105, 100, 0, 69, 117, 114, 111, 112, 101, 58, 77, 97, 108, 116, 97, 0, 69, 117, 114, 111, 112, 101, 58, 77, 97, 114, 105, 101, 104, 97, 109, 110, 0, 69, 117, 114, 111, 112, 101, 58, 77, 105, 110, 115, 107, 0, 69, 117, 114, 111, 112, 101, 58, 77, 111, 110, 97, 99, 111, 0, 69, 117, 114, 111, 112, 101, 58, 77, 111, 115, 99, 111, 119, 0, 69, 117, 114, 111, 112, 101, 58, 79, 115, 108, 111, 0, 69, 117, 114, 111, 112, 101, 58, 80, 97, 114, 105, 115, 0, 69, 117, 114, 111, 112, 101, 58, 80, 111, 100, 103, 111, 114, 105, 99, 97, 0, 69, 117, 114, 111, 112, 101, 58, 80, 114, 97, 103, 117, 101, 0, 69, 117, 114, 111, 112, 101, 58, 82, 105, 103, 97, 0, 69, 117, 114, 111, 112, 101, 58, 82, 111, 109, 101, 0, 69, 117, 114, 111, 112, 101, 58, 83, 97, 109, 97, 114, 97, 0, 69, 117, 114, 111, 112, 101, 58, 83, 97, 110, 95, 77, 97, 114, 105, 110, 111, 0, 69, 117, 114, 111, 112, 101, 58, 83, 97, 114, 97, 106, 101, 118, 111, 0, 69, 117, 114, 111, 112, 101, 58, 83, 97, 114, 97, 116, 111, 118, 0, 69, 117, 114, 111, 112, 101, 58, 83, 105, 109, 102, 101, 114, 111, 112, 111, 108, 0, 69, 117, 114, 111, 112, 101, 58, 83, 107, 111, 112, 106, 101, 0, 69, 117, 114, 111, 112, 101, 58, 83, 111, 102, 105, 97, 0, 69, 117, 114, 111, 112, 101, 58, 83, 116, 111, 99, 107, 104, 111, 108, 109, 0, 69, 117, 114, 111, 112, 101, 58, 84, 97, 108, 108, 105, 110, 110, 0, 69, 117, 114, 111, 112, 101, 58, 84, 105, 114, 97, 110, 101, 0, 69, 117, 114, 111, 112, 101, 58, 85, 108, 121, 97, 110, 111, 118, 115, 107, 0, 69, 117, 114, 111, 112, 101, 58, 85, 122, 104, 103, 111, 114, 111, 100, 0, 69, 117, 114, 111, 112, 101, 58, 86, 97, 100, 117, 122, 0, 69, 117, 114, 111, 112, 101, 58, 86, 97, 116, 105, 99, 97, 110, 0, 69, 117, 114, 111, 112, 101, 58, 86, 105, 101, 110, 110, 97, 0, 69, 117, 114, 111, 112, 101, 58, 86, 105, 108, 110, 105, 117, 115, 0, 69, 117, 114, 111, 112, 101, 58, 86, 111, 108, 103, 111, 103, 114, 97, 100, 0, 69, 117, 114, 111, 112, 101, 58, 87, 97, 114, 115, 97, 119, 0, 69, 117, 114, 111, 112, 101, 58, 90, 97, 103, 114, 101, 98, 0, 69, 117, 114, 111, 112, 101, 58, 90, 97, 112, 111, 114, 111, 122, 104, 121, 101, 0, 69, 117, 114, 111, 112, 101, 58, 90, 117, 114, 105, 99, 104, 0, 73, 110, 100, 105, 97, 110, 58, 65, 110, 116, 97, 110, 97, 110, 97, 114, 105, 118, 111, 0, 73, 110, 100, 105, 97, 110, 58, 67, 104, 97, 103, 111, 115, 0, 73, 110, 100, 105, 97, 110, 58, 67, 104, 114, 105, 115, 116, 109, 97, 115, 0, 73, 110, 100, 105, 97, 110, 58, 67, 111, 99, 111, 115, 0, 73, 110, 100, 105, 97, 110, 58, 67, 111, 109, 111, 114, 111, 0, 73, 110, 100, 105, 97, 110, 58, 75, 101, 114, 103, 117, 101, 108, 101, 110, 0, 73, 110, 100, 105, 97, 110, 58, 77, 97, 104, 101, 0, 73, 110, 100, 105, 97, 110, 58, 77, 97, 108, 100, 105, 118, 101, 115, 0, 73, 110, 100, 105, 97, 110, 58, 77, 97, 117, 114, 105, 116, 105, 117, 115, 0, 73, 110, 100, 105, 97, 110, 58, 77, 97, 121, 111, 116, 116, 101, 0, 73, 110, 100, 105, 97, 110, 58, 82, 101, 117, 110, 105, 111, 110, 0, 80, 97, 99, 105, 102, 105, 99, 58, 65, 112, 105, 97, 0, 80, 97, 99, 105, 102, 105, 99, 58, 65, 117, 99, 107, 108, 97, 110, 100, 0, 80, 97, 99, 105, 102, 105, 99, 58, 66, 111, 117, 103, 97, 105, 110, 118, 105, 108, 108, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 67, 104, 97, 116, 104, 97, 109, 0, 80, 97, 99, 105, 102, 105, 99, 58, 69, 97, 115, 116, 101, 114, 0, 80, 97, 99, 105, 102, 105, 99, 58, 69, 102, 97, 116, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 69, 110, 100, 101, 114, 98, 117, 114, 121, 0, 80, 97, 99, 105, 102, 105, 99, 58, 70, 97, 107, 97, 111, 102, 111, 0, 80, 97, 99, 105, 102, 105, 99, 58, 70, 105, 106, 105, 0, 80, 97, 99, 105, 102, 105, 99, 58, 70, 117, 110, 97, 102, 117, 116, 105, 0, 80, 97, 99, 105, 102, 105, 99, 58, 71, 97, 108, 97, 112, 97, 103, 111, 115, 0, 80, 97, 99, 105, 102, 105, 99, 58, 71, 97, 109, 98, 105, 101, 114, 0, 80, 97, 99, 105, 102, 105, 99, 58, 71, 117, 97, 100, 97, 108, 99, 97, 110, 97, 108, 0, 80, 97, 99, 105, 102, 105, 99, 58, 71, 117, 97, 109, 0, 80, 97, 99, 105, 102, 105, 99, 58, 72, 111, 110, 111, 108, 117, 108, 117, 0, 80, 97, 99, 105, 102, 105, 99, 58, 74, 111, 104, 110, 115, 116, 111, 110, 0, 80, 97, 99, 105, 102, 105, 99, 58, 75, 105, 114, 105, 116, 105, 109, 97, 116, 105, 0, 80, 97, 99, 105, 102, 105, 99, 58, 75, 111, 115, 114, 97, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 75, 119, 97, 106, 97, 108, 101, 105, 110, 0, 80, 97, 99, 105, 102, 105, 99, 58, 77, 97, 106, 117, 114, 111, 0, 80, 97, 99, 105, 102, 105, 99, 58, 77, 97, 114, 113, 117, 101, 115, 97, 115, 0, 80, 97, 99, 105, 102, 105, 99, 58, 77, 105, 100, 119, 97, 121, 0, 80, 97, 99, 105, 102, 105, 99, 58, 78, 97, 117, 114, 117, 0, 80, 97, 99, 105, 102, 105, 99, 58, 78, 105, 117, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 78, 111, 114, 102, 111, 108, 107, 0, 80, 97, 99, 105, 102, 105, 99, 58, 78, 111, 117, 109, 101, 97, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 97, 103, 111, 95, 80, 97, 103, 111, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 97, 108, 97, 117, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 105, 116, 99, 97, 105, 114, 110, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 111, 110, 97, 112, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 80, 111, 114, 116, 95, 77, 111, 114, 101, 115, 98, 121, 0, 80, 97, 99, 105, 102, 105, 99, 58, 82, 97, 114, 111, 116, 111, 110, 103, 97, 0, 80, 97, 99, 105, 102, 105, 99, 58, 83, 97, 105, 112, 97, 110, 0, 80, 97, 99, 105, 102, 105, 99, 58, 84, 97, 104, 105, 116, 105, 0, 80, 97, 99, 105, 102, 105, 99, 58, 84, 97, 114, 97, 119, 97, 0, 80, 97, 99, 105, 102, 105, 99, 58, 84, 111, 110, 103, 97, 116, 97, 112, 117, 0, 80, 97, 99, 105, 102, 105, 99, 58, 84, 114, 117, 107, 0, 80, 97, 99, 105, 102, 105, 99, 58, 87, 97, 107, 101, 0, 80, 97, 99, 105, 102, 105, 99, 58, 87, 97, 108, 108, 105, 115, 0, 67, 83, 84, 54, 67, 68, 84, 0, 69, 83, 84, 53, 69, 68, 84, 0, 77, 83, 84, 55, 77, 68, 84, 0, 80, 83, 84, 56, 80, 68, 84, 0, 170, 0, 0, 103, 0, 109, 0, 116, 0, 0, 0, 117, 0, 110, 0, 107, 0, 0, 0, 117, 0, 116, 0, 99, 0, 0, 0, 103, 0, 97, 0, 122, 0, 97, 0, 0, 0, 99, 0, 110, 0, 115, 0, 104, 0, 97, 0, 0, 0, 99, 0, 110, 0, 117, 0, 114, 0, 99, 0, 0, 0, 110, 0, 122, 0, 97, 0, 107, 0, 108, 0, 0, 0, 117, 0, 115, 0, 100, 0, 101, 0, 110, 0, 0, 0, 97, 0, 100, 0, 97, 0, 108, 0, 118, 0, 0, 0, 97, 0, 101, 0, 100, 0, 120, 0, 98, 0, 0, 0, 97, 0, 102, 0, 107, 0, 98, 0, 108, 0, 0, 0, 97, 0, 103, 0, 97, 0, 110, 0, 117, 0, 0, 0, 97, 0, 105, 0, 97, 0, 120, 0, 97, 0, 0, 0, 97, 0, 108, 0, 116, 0, 105, 0, 97, 0, 0, 0, 97, 0, 109, 0, 101, 0, 118, 0, 110, 0, 0, 0, 97, 0, 110, 0, 99, 0, 117, 0, 114, 0, 0, 0, 97, 0, 111, 0, 108, 0, 97, 0, 100, 0, 0, 0, 97, 0, 113, 0, 99, 0, 97, 0, 115, 0, 0, 0, 97, 0, 113, 0, 100, 0, 97, 0, 118, 0, 0, 0, 97, 0, 113, 0, 100, 0, 100, 0, 117, 0, 0, 0, 97, 0, 113, 0, 109, 0, 97, 0, 119, 0, 0, 0, 97, 0, 113, 0, 109, 0, 99, 0, 109, 0, 0, 0, 97, 0, 113, 0, 112, 0, 108, 0, 109, 0, 0, 0, 97, 0, 113, 0, 114, 0, 111, 0, 116, 0, 0, 0, 97, 0, 113, 0, 115, 0, 121, 0, 119, 0, 0, 0, 97, 0, 113, 0, 116, 0, 114, 0, 108, 0, 0, 0, 97, 0, 113, 0, 118, 0, 111, 0, 115, 0, 0, 0, 97, 0, 114, 0, 98, 0, 117, 0, 101, 0, 0, 0, 97, 0, 114, 0, 99, 0, 111, 0, 114, 0, 0, 0, 97, 0, 114, 0, 99, 0, 116, 0, 99, 0, 0, 0, 97, 0, 114, 0, 105, 0, 114, 0, 106, 0, 0, 0, 97, 0, 114, 0, 106, 0, 117, 0, 106, 0, 0, 0, 97, 0, 114, 0, 108, 0, 117, 0, 113, 0, 0, 0, 97, 0, 114, 0, 109, 0, 100, 0, 122, 0, 0, 0, 97, 0, 114, 0, 114, 0, 103, 0, 108, 0, 0, 0, 97, 0, 114, 0, 115, 0, 108, 0, 97, 0, 0, 0, 97, 0, 114, 0, 116, 0, 117, 0, 99, 0, 0, 0, 97, 0, 114, 0, 117, 0, 97, 0, 113, 0, 0, 0, 97, 0, 114, 0, 117, 0, 115, 0, 104, 0, 0, 0, 97, 0, 115, 0, 112, 0, 112, 0, 103, 0, 0, 0, 97, 0, 116, 0, 118, 0, 105, 0, 101, 0, 0, 0, 97, 0, 117, 0, 97, 0, 100, 0, 108, 0, 0, 0, 97, 0, 117, 0, 98, 0, 104, 0, 113, 0, 0, 0, 97, 0, 117, 0, 98, 0, 110, 0, 101, 0, 0, 0, 97, 0, 117, 0, 100, 0, 114, 0, 119, 0, 0, 0, 97, 0, 117, 0, 101, 0, 117, 0, 99, 0, 0, 0, 97, 0, 117, 0, 104, 0, 98, 0, 97, 0, 0, 0, 97, 0, 117, 0, 107, 0, 110, 0, 115, 0, 0, 0, 97, 0, 117, 0, 108, 0, 100, 0, 99, 0, 0, 0, 97, 0, 117, 0, 108, 0, 100, 0, 104, 0, 0, 0, 97, 0, 117, 0, 109, 0, 101, 0, 108, 0, 0, 0, 97, 0, 117, 0, 109, 0, 113, 0, 105, 0, 0, 0, 97, 0, 117, 0, 112, 0, 101, 0, 114, 0, 0, 0, 97, 0, 117, 0, 115, 0, 121, 0, 100, 0, 0, 0, 97, 0, 119, 0, 97, 0, 117, 0, 97, 0, 0, 0, 97, 0, 122, 0, 98, 0, 97, 0, 107, 0, 0, 0, 98, 0, 97, 0, 115, 0, 106, 0, 106, 0, 0, 0, 98, 0, 98, 0, 98, 0, 103, 0, 105, 0, 0, 0, 98, 0, 100, 0, 100, 0, 97, 0, 99, 0, 0, 0, 98, 0, 101, 0, 98, 0, 114, 0, 117, 0, 0, 0, 98, 0, 102, 0, 111, 0, 117, 0, 97, 0, 0, 0, 98, 0, 103, 0, 115, 0, 111, 0, 102, 0, 0, 0, 98, 0, 104, 0, 98, 0, 97, 0, 104, 0, 0, 0, 98, 0, 105, 0, 98, 0, 106, 0, 109, 0, 0, 0, 98, 0, 106, 0, 112, 0, 116, 0, 110, 0, 0, 0, 98, 0, 109, 0, 98, 0, 100, 0, 97, 0, 0, 0, 98, 0, 110, 0, 98, 0, 119, 0, 110, 0, 0, 0, 98, 0, 111, 0, 108, 0, 112, 0, 98, 0, 0, 0, 98, 0, 113, 0, 107, 0, 114, 0, 97, 0, 0, 0, 98, 0, 114, 0, 97, 0, 117, 0, 120, 0, 0, 0, 98, 0, 114, 0, 98, 0, 101, 0, 108, 0, 0, 0, 98, 0, 114, 0, 98, 0, 118, 0, 98, 0, 0, 0, 98, 0, 114, 0, 99, 0, 103, 0, 98, 0, 0, 0, 98, 0, 114, 0, 99, 0, 103, 0, 114, 0, 0, 0, 98, 0, 114, 0, 101, 0, 114, 0, 110, 0, 0, 0, 98, 0, 114, 0, 102, 0, 101, 0, 110, 0, 0, 0, 98, 0, 114, 0, 102, 0, 111, 0, 114, 0, 0, 0, 98, 0, 114, 0, 109, 0, 97, 0, 111, 0, 0, 0, 98, 0, 114, 0, 109, 0, 99, 0, 122, 0, 0, 0, 98, 0, 114, 0, 112, 0, 118, 0, 104, 0, 0, 0, 98, 0, 114, 0, 114, 0, 98, 0, 114, 0, 0, 0, 98, 0, 114, 0, 114, 0, 101, 0, 99, 0, 0, 0, 98, 0, 114, 0, 115, 0, 97, 0, 111, 0, 0, 0, 98, 0, 114, 0, 115, 0, 115, 0, 97, 0, 0, 0, 98, 0, 114, 0, 115, 0, 116, 0, 109, 0, 0, 0, 98, 0, 115, 0, 110, 0, 97, 0, 115, 0, 0, 0, 98, 0, 116, 0, 116, 0, 104, 0, 105, 0, 0, 0, 98, 0, 119, 0, 103, 0, 98, 0, 101, 0, 0, 0, 98, 0, 121, 0, 109, 0, 115, 0, 113, 0, 0, 0, 98, 0, 122, 0, 98, 0, 122, 0, 101, 0, 0, 0, 99, 0, 97, 0, 99, 0, 102, 0, 113, 0, 0, 0, 99, 0, 97, 0, 101, 0, 100, 0, 109, 0, 0, 0, 99, 0, 97, 0, 102, 0, 102, 0, 115, 0, 0, 0, 99, 0, 97, 0, 102, 0, 110, 0, 101, 0, 0, 0, 99, 0, 97, 0, 103, 0, 108, 0, 98, 0, 0, 0, 99, 0, 97, 0, 103, 0, 111, 0, 111, 0, 0, 0, 99, 0, 97, 0, 104, 0, 97, 0, 108, 0, 0, 0, 99, 0, 97, 0, 105, 0, 113, 0, 108, 0, 0, 0, 99, 0, 97, 0, 109, 0, 111, 0, 110, 0, 0, 0, 99, 0, 97, 0, 109, 0, 116, 0, 114, 0, 0, 0, 99, 0, 97, 0, 110, 0, 112, 0, 103, 0, 0, 0, 99, 0, 97, 0, 112, 0, 110, 0, 116, 0, 0, 0, 99, 0, 97, 0, 114, 0, 101, 0, 98, 0, 0, 0, 99, 0, 97, 0, 114, 0, 101, 0, 103, 0, 0, 0, 99, 0, 97, 0, 115, 0, 106, 0, 102, 0, 0, 0, 99, 0, 97, 0, 116, 0, 104, 0, 117, 0, 0, 0, 99, 0, 97, 0, 116, 0, 111, 0, 114, 0, 0, 0, 99, 0, 97, 0, 118, 0, 97, 0, 110, 0, 0, 0, 99, 0, 97, 0, 119, 0, 110, 0, 112, 0, 0, 0, 99, 0, 97, 0, 121, 0, 98, 0, 120, 0, 0, 0, 99, 0, 97, 0, 121, 0, 99, 0, 98, 0, 0, 0, 99, 0, 97, 0, 121, 0, 100, 0, 97, 0, 0, 0, 99, 0, 97, 0, 121, 0, 100, 0, 113, 0, 0, 0, 99, 0, 97, 0, 121, 0, 101, 0, 107, 0, 0, 0, 99, 0, 97, 0, 121, 0, 101, 0, 118, 0, 0, 0, 99, 0, 97, 0, 121, 0, 120, 0, 121, 0, 0, 0, 99, 0, 97, 0, 121, 0, 121, 0, 110, 0, 0, 0, 99, 0, 97, 0, 121, 0, 122, 0, 102, 0, 0, 0, 99, 0, 97, 0, 121, 0, 122, 0, 115, 0, 0, 0, 99, 0, 99, 0, 99, 0, 99, 0, 107, 0, 0, 0, 99, 0, 100, 0, 102, 0, 98, 0, 109, 0, 0, 0, 99, 0, 100, 0, 102, 0, 105, 0, 104, 0, 0, 0, 99, 0, 102, 0, 98, 0, 103, 0, 102, 0, 0, 0, 99, 0, 103, 0, 98, 0, 122, 0, 118, 0, 0, 0, 99, 0, 104, 0, 122, 0, 114, 0, 104, 0, 0, 0, 99, 0, 105, 0, 97, 0, 98, 0, 106, 0, 0, 0, 99, 0, 107, 0, 114, 0, 97, 0, 114, 0, 0, 0, 99, 0, 108, 0, 105, 0, 112, 0, 99, 0, 0, 0, 99, 0, 108, 0, 112, 0, 117, 0, 113, 0, 0, 0, 99, 0, 108, 0, 115, 0, 99, 0, 108, 0, 0, 0, 99, 0, 109, 0, 100, 0, 108, 0, 97, 0, 0, 0, 99, 0, 111, 0, 98, 0, 111, 0, 103, 0, 0, 0, 99, 0, 114, 0, 115, 0, 106, 0, 111, 0, 0, 0, 99, 0, 117, 0, 104, 0, 97, 0, 118, 0, 0, 0, 99, 0, 118, 0, 114, 0, 97, 0, 105, 0, 0, 0, 99, 0, 120, 0, 120, 0, 99, 0, 104, 0, 0, 0, 99, 0, 121, 0, 102, 0, 109, 0, 103, 0, 0, 0, 99, 0, 121, 0, 110, 0, 105, 0, 99, 0, 0, 0, 99, 0, 122, 0, 112, 0, 114, 0, 103, 0, 0, 0, 100, 0, 101, 0, 98, 0, 101, 0, 114, 0, 0, 0, 100, 0, 106, 0, 106, 0, 105, 0, 98, 0, 0, 0, 100, 0, 107, 0, 99, 0, 112, 0, 104, 0, 0, 0, 100, 0, 109, 0, 100, 0, 111, 0, 109, 0, 0, 0, 100, 0, 111, 0, 115, 0, 100, 0, 113, 0, 0, 0, 100, 0, 122, 0, 97, 0, 108, 0, 103, 0, 0, 0, 101, 0, 99, 0, 103, 0, 112, 0, 115, 0, 0, 0, 101, 0, 99, 0, 103, 0, 121, 0, 101, 0, 0, 0, 101, 0, 101, 0, 116, 0, 108, 0, 108, 0, 0, 0, 101, 0, 103, 0, 99, 0, 97, 0, 105, 0, 0, 0, 101, 0, 104, 0, 101, 0, 97, 0, 105, 0, 0, 0, 101, 0, 114, 0, 97, 0, 115, 0, 109, 0, 0, 0, 101, 0, 115, 0, 99, 0, 101, 0, 117, 0, 0, 0, 101, 0, 115, 0, 108, 0, 112, 0, 97, 0, 0, 0, 101, 0, 115, 0, 109, 0, 97, 0, 100, 0, 0, 0, 101, 0, 116, 0, 97, 0, 100, 0, 100, 0, 0, 0, 102, 0, 105, 0, 104, 0, 101, 0, 108, 0, 0, 0, 102, 0, 105, 0, 109, 0, 104, 0, 113, 0, 0, 0, 102, 0, 106, 0, 115, 0, 117, 0, 118, 0, 0, 0, 102, 0, 107, 0, 112, 0, 115, 0, 121, 0, 0, 0, 102, 0, 109, 0, 107, 0, 115, 0, 97, 0, 0, 0, 102, 0, 109, 0, 112, 0, 110, 0, 105, 0, 0, 0, 102, 0, 109, 0, 116, 0, 107, 0, 107, 0, 0, 0, 102, 0, 111, 0, 116, 0, 104, 0, 111, 0, 0, 0, 102, 0, 114, 0, 112, 0, 97, 0, 114, 0, 0, 0, 103, 0, 97, 0, 108, 0, 98, 0, 118, 0, 0, 0, 103, 0, 98, 0, 108, 0, 111, 0, 110, 0, 0, 0, 103, 0, 100, 0, 103, 0, 110, 0, 100, 0, 0, 0, 103, 0, 101, 0, 116, 0, 98, 0, 115, 0, 0, 0, 103, 0, 102, 0, 99, 0, 97, 0, 121, 0, 0, 0, 103, 0, 103, 0, 103, 0, 99, 0, 105, 0, 0, 0, 103, 0, 104, 0, 97, 0, 99, 0, 99, 0, 0, 0, 103, 0, 105, 0, 103, 0, 105, 0, 98, 0, 0, 0, 103, 0, 108, 0, 103, 0, 111, 0, 104, 0, 0, 0, 103, 0, 108, 0, 111, 0, 98, 0, 121, 0, 0, 0, 103, 0, 108, 0, 116, 0, 104, 0, 117, 0, 0, 0, 103, 0, 109, 0, 98, 0, 106, 0, 108, 0, 0, 0, 103, 0, 110, 0, 99, 0, 107, 0, 121, 0, 0, 0, 103, 0, 112, 0, 98, 0, 98, 0, 114, 0, 0, 0, 103, 0, 112, 0, 109, 0, 115, 0, 98, 0, 0, 0, 103, 0, 112, 0, 115, 0, 98, 0, 104, 0, 0, 0, 103, 0, 113, 0, 115, 0, 115, 0, 103, 0, 0, 0, 103, 0, 114, 0, 97, 0, 116, 0, 104, 0, 0, 0, 103, 0, 115, 0, 103, 0, 114, 0, 118, 0, 0, 0, 103, 0, 116, 0, 103, 0, 117, 0, 97, 0, 0, 0, 103, 0, 117, 0, 103, 0, 117, 0, 109, 0, 0, 0, 103, 0, 119, 0, 111, 0, 120, 0, 98, 0, 0, 0, 103, 0, 121, 0, 103, 0, 101, 0, 111, 0, 0, 0, 104, 0, 107, 0, 104, 0, 107, 0, 103, 0, 0, 0, 104, 0, 110, 0, 116, 0, 103, 0, 117, 0, 0, 0, 104, 0, 114, 0, 122, 0, 97, 0, 103, 0, 0, 0, 104, 0, 116, 0, 112, 0, 97, 0, 112, 0, 0, 0, 104, 0, 117, 0, 98, 0, 117, 0, 100, 0, 0, 0, 105, 0, 100, 0, 100, 0, 106, 0, 106, 0, 0, 0, 105, 0, 100, 0, 106, 0, 107, 0, 116, 0, 0, 0, 105, 0, 100, 0, 109, 0, 97, 0, 107, 0, 0, 0, 105, 0, 100, 0, 112, 0, 110, 0, 107, 0, 0, 0, 105, 0, 101, 0, 100, 0, 117, 0, 98, 0, 0, 0, 105, 0, 109, 0, 100, 0, 103, 0, 115, 0, 0, 0, 105, 0, 110, 0, 99, 0, 99, 0, 117, 0, 0, 0, 105, 0, 111, 0, 100, 0, 103, 0, 97, 0, 0, 0, 105, 0, 113, 0, 98, 0, 103, 0, 119, 0, 0, 0, 105, 0, 114, 0, 116, 0, 104, 0, 114, 0, 0, 0, 105, 0, 115, 0, 114, 0, 101, 0, 121, 0, 0, 0, 105, 0, 116, 0, 114, 0, 111, 0, 109, 0, 0, 0, 106, 0, 101, 0, 115, 0, 116, 0, 104, 0, 0, 0, 106, 0, 109, 0, 107, 0, 105, 0, 110, 0, 0, 0, 106, 0, 111, 0, 97, 0, 109, 0, 109, 0, 0, 0, 106, 0, 112, 0, 116, 0, 121, 0, 111, 0, 0, 0, 107, 0, 101, 0, 110, 0, 98, 0, 111, 0, 0, 0, 107, 0, 103, 0, 102, 0, 114, 0, 117, 0, 0, 0, 107, 0, 104, 0, 112, 0, 110, 0, 104, 0, 0, 0, 107, 0, 105, 0, 99, 0, 120, 0, 105, 0, 0, 0, 107, 0, 105, 0, 112, 0, 104, 0, 111, 0, 0, 0, 107, 0, 105, 0, 116, 0, 114, 0, 119, 0, 0, 0, 107, 0, 109, 0, 121, 0, 118, 0, 97, 0, 0, 0, 107, 0, 110, 0, 98, 0, 97, 0, 115, 0, 0, 0, 107, 0, 112, 0, 102, 0, 110, 0, 106, 0, 0, 0, 107, 0, 114, 0, 115, 0, 101, 0, 108, 0, 0, 0, 107, 0, 119, 0, 107, 0, 119, 0, 105, 0, 0, 0, 107, 0, 121, 0, 103, 0, 101, 0, 99, 0, 0, 0, 107, 0, 122, 0, 97, 0, 97, 0, 117, 0, 0, 0, 107, 0, 122, 0, 97, 0, 107, 0, 120, 0, 0, 0, 107, 0, 122, 0, 97, 0, 108, 0, 97, 0, 0, 0, 107, 0, 122, 0, 103, 0, 117, 0, 119, 0, 0, 0, 107, 0, 122, 0, 107, 0, 115, 0, 110, 0, 0, 0, 107, 0, 122, 0, 107, 0, 122, 0, 111, 0, 0, 0, 107, 0, 122, 0, 117, 0, 114, 0, 97, 0, 0, 0, 108, 0, 97, 0, 118, 0, 116, 0, 101, 0, 0, 0, 108, 0, 98, 0, 98, 0, 101, 0, 121, 0, 0, 0, 108, 0, 99, 0, 99, 0, 97, 0, 115, 0, 0, 0, 108, 0, 105, 0, 118, 0, 100, 0, 122, 0, 0, 0, 108, 0, 107, 0, 99, 0, 109, 0, 98, 0, 0, 0, 108, 0, 114, 0, 109, 0, 108, 0, 119, 0, 0, 0, 108, 0, 115, 0, 109, 0, 115, 0, 117, 0, 0, 0, 108, 0, 116, 0, 118, 0, 110, 0, 111, 0, 0, 0, 108, 0, 117, 0, 108, 0, 117, 0, 120, 0, 0, 0, 108, 0, 118, 0, 114, 0, 105, 0, 120, 0, 0, 0, 108, 0, 121, 0, 116, 0, 105, 0, 112, 0, 0, 0, 109, 0, 97, 0, 99, 0, 97, 0, 115, 0, 0, 0, 109, 0, 99, 0, 109, 0, 111, 0, 110, 0, 0, 0, 109, 0, 100, 0, 107, 0, 105, 0, 118, 0, 0, 0, 109, 0, 101, 0, 116, 0, 103, 0, 100, 0, 0, 0, 109, 0, 103, 0, 116, 0, 110, 0, 114, 0, 0, 0, 109, 0, 104, 0, 107, 0, 119, 0, 97, 0, 0, 0, 109, 0, 104, 0, 109, 0, 97, 0, 106, 0, 0, 0, 109, 0, 107, 0, 115, 0, 107, 0, 112, 0, 0, 0, 109, 0, 108, 0, 98, 0, 107, 0, 111, 0, 0, 0, 109, 0, 109, 0, 114, 0, 103, 0, 110, 0, 0, 0, 109, 0, 110, 0, 99, 0, 111, 0, 113, 0, 0, 0, 109, 0, 110, 0, 104, 0, 118, 0, 100, 0, 0, 0, 109, 0, 110, 0, 117, 0, 108, 0, 110, 0, 0, 0, 109, 0, 111, 0, 109, 0, 102, 0, 109, 0, 0, 0, 109, 0, 112, 0, 115, 0, 112, 0, 110, 0, 0, 0, 109, 0, 113, 0, 102, 0, 100, 0, 102, 0, 0, 0, 109, 0, 114, 0, 110, 0, 107, 0, 99, 0, 0, 0, 109, 0, 115, 0, 109, 0, 110, 0, 105, 0, 0, 0, 109, 0, 116, 0, 109, 0, 108, 0, 97, 0, 0, 0, 109, 0, 117, 0, 112, 0, 108, 0, 117, 0, 0, 0, 109, 0, 118, 0, 109, 0, 108, 0, 101, 0, 0, 0, 109, 0, 119, 0, 98, 0, 108, 0, 122, 0, 0, 0, 109, 0, 120, 0, 99, 0, 104, 0, 105, 0, 0, 0, 109, 0, 120, 0, 99, 0, 117, 0, 110, 0, 0, 0, 109, 0, 120, 0, 104, 0, 109, 0, 111, 0, 0, 0, 109, 0, 120, 0, 109, 0, 97, 0, 109, 0, 0, 0, 109, 0, 120, 0, 109, 0, 101, 0, 120, 0, 0, 0, 109, 0, 120, 0, 109, 0, 105, 0, 100, 0, 0, 0, 109, 0, 120, 0, 109, 0, 116, 0, 121, 0, 0, 0, 109, 0, 120, 0, 109, 0, 122, 0, 116, 0, 0, 0, 109, 0, 120, 0, 111, 0, 106, 0, 105, 0, 0, 0, 109, 0, 120, 0, 112, 0, 118, 0, 114, 0, 0, 0, 109, 0, 120, 0, 116, 0, 105, 0, 106, 0, 0, 0, 109, 0, 121, 0, 107, 0, 99, 0, 104, 0, 0, 0, 109, 0, 121, 0, 107, 0, 117, 0, 108, 0, 0, 0, 109, 0, 122, 0, 109, 0, 112, 0, 109, 0, 0, 0, 110, 0, 97, 0, 119, 0, 100, 0, 104, 0, 0, 0, 110, 0, 99, 0, 110, 0, 111, 0, 117, 0, 0, 0, 110, 0, 101, 0, 110, 0, 105, 0, 109, 0, 0, 0, 110, 0, 102, 0, 110, 0, 108, 0, 107, 0, 0, 0, 110, 0, 103, 0, 108, 0, 111, 0, 115, 0, 0, 0, 110, 0, 105, 0, 109, 0, 103, 0, 97, 0, 0, 0, 110, 0, 108, 0, 97, 0, 109, 0, 115, 0, 0, 0, 110, 0, 111, 0, 111, 0, 115, 0, 108, 0, 0, 0, 110, 0, 112, 0, 107, 0, 116, 0, 109, 0, 0, 0, 110, 0, 114, 0, 105, 0, 110, 0, 117, 0, 0, 0, 110, 0, 117, 0, 105, 0, 117, 0, 101, 0, 0, 0, 110, 0, 122, 0, 99, 0, 104, 0, 116, 0, 0, 0, 111, 0, 109, 0, 109, 0, 99, 0, 116, 0, 0, 0, 112, 0, 97, 0, 112, 0, 116, 0, 121, 0, 0, 0, 112, 0, 101, 0, 108, 0, 105, 0, 109, 0, 0, 0, 112, 0, 102, 0, 103, 0, 109, 0, 114, 0, 0, 0, 112, 0, 102, 0, 110, 0, 104, 0, 118, 0, 0, 0, 112, 0, 102, 0, 112, 0, 112, 0, 116, 0, 0, 0, 112, 0, 103, 0, 112, 0, 111, 0, 109, 0, 0, 0, 112, 0, 103, 0, 114, 0, 97, 0, 119, 0, 0, 0, 112, 0, 104, 0, 109, 0, 110, 0, 108, 0, 0, 0, 112, 0, 107, 0, 107, 0, 104, 0, 105, 0, 0, 0, 112, 0, 108, 0, 119, 0, 97, 0, 119, 0, 0, 0, 112, 0, 109, 0, 109, 0, 113, 0, 99, 0, 0, 0, 112, 0, 110, 0, 112, 0, 99, 0, 110, 0, 0, 0, 112, 0, 114, 0, 115, 0, 106, 0, 117, 0, 0, 0, 112, 0, 116, 0, 102, 0, 110, 0, 99, 0, 0, 0, 112, 0, 116, 0, 108, 0, 105, 0, 115, 0, 0, 0, 112, 0, 116, 0, 112, 0, 100, 0, 108, 0, 0, 0, 112, 0, 119, 0, 114, 0, 111, 0, 114, 0, 0, 0, 112, 0, 121, 0, 97, 0, 115, 0, 117, 0, 0, 0, 113, 0, 97, 0, 100, 0, 111, 0, 104, 0, 0, 0, 114, 0, 101, 0, 114, 0, 101, 0, 117, 0, 0, 0, 114, 0, 111, 0, 98, 0, 117, 0, 104, 0, 0, 0, 114, 0, 115, 0, 98, 0, 101, 0, 103, 0, 0, 0, 114, 0, 117, 0, 97, 0, 115, 0, 102, 0, 0, 0, 114, 0, 117, 0, 98, 0, 97, 0, 120, 0, 0, 0, 114, 0, 117, 0, 100, 0, 121, 0, 114, 0, 0, 0, 114, 0, 117, 0, 103, 0, 100, 0, 120, 0, 0, 0, 114, 0, 117, 0, 105, 0, 107, 0, 116, 0, 0, 0, 114, 0, 117, 0, 107, 0, 103, 0, 100, 0, 0, 0, 114, 0, 117, 0, 107, 0, 114, 0, 97, 0, 0, 0, 114, 0, 117, 0, 107, 0, 117, 0, 102, 0, 0, 0, 114, 0, 117, 0, 107, 0, 118, 0, 120, 0, 0, 0, 114, 0, 117, 0, 109, 0, 111, 0, 119, 0, 0, 0, 114, 0, 117, 0, 110, 0, 111, 0, 122, 0, 0, 0, 114, 0, 117, 0, 111, 0, 109, 0, 115, 0, 0, 0, 114, 0, 117, 0, 111, 0, 118, 0, 98, 0, 0, 0, 114, 0, 117, 0, 112, 0, 107, 0, 99, 0, 0, 0, 114, 0, 117, 0, 114, 0, 116, 0, 119, 0, 0, 0, 114, 0, 117, 0, 116, 0, 111, 0, 102, 0, 0, 0, 114, 0, 117, 0, 117, 0, 108, 0, 121, 0, 0, 0, 114, 0, 117, 0, 117, 0, 117, 0, 115, 0, 0, 0, 114, 0, 117, 0, 118, 0, 111, 0, 103, 0, 0, 0, 114, 0, 117, 0, 118, 0, 118, 0, 111, 0, 0, 0, 114, 0, 117, 0, 121, 0, 101, 0, 107, 0, 0, 0, 114, 0, 117, 0, 121, 0, 107, 0, 115, 0, 0, 0, 114, 0, 119, 0, 107, 0, 103, 0, 108, 0, 0, 0, 115, 0, 97, 0, 114, 0, 117, 0, 104, 0, 0, 0, 115, 0, 98, 0, 104, 0, 105, 0, 114, 0, 0, 0, 115, 0, 99, 0, 109, 0, 97, 0, 119, 0, 0, 0, 115, 0, 100, 0, 107, 0, 114, 0, 116, 0, 0, 0, 115, 0, 101, 0, 115, 0, 116, 0, 111, 0, 0, 0, 115, 0, 103, 0, 115, 0, 105, 0, 110, 0, 0, 0, 115, 0, 104, 0, 115, 0, 104, 0, 110, 0, 0, 0, 115, 0, 105, 0, 108, 0, 106, 0, 117, 0, 0, 0, 115, 0, 106, 0, 108, 0, 121, 0, 114, 0, 0, 0, 115, 0, 107, 0, 98, 0, 116, 0, 115, 0, 0, 0, 115, 0, 108, 0, 102, 0, 110, 0, 97, 0, 0, 0, 115, 0, 109, 0, 115, 0, 97, 0, 105, 0, 0, 0, 115, 0, 110, 0, 100, 0, 107, 0, 114, 0, 0, 0, 115, 0, 111, 0, 109, 0, 103, 0, 113, 0, 0, 0, 115, 0, 114, 0, 112, 0, 98, 0, 109, 0, 0, 0, 115, 0, 115, 0, 106, 0, 117, 0, 98, 0, 0, 0, 115, 0, 116, 0, 116, 0, 109, 0, 115, 0, 0, 0, 115, 0, 118, 0, 115, 0, 97, 0, 108, 0, 0, 0, 115, 0, 120, 0, 112, 0, 104, 0, 105, 0, 0, 0, 115, 0, 121, 0, 100, 0, 97, 0, 109, 0, 0, 0, 115, 0, 122, 0, 113, 0, 109, 0, 110, 0, 0, 0, 116, 0, 99, 0, 103, 0, 100, 0, 116, 0, 0, 0, 116, 0, 100, 0, 110, 0, 100, 0, 106, 0, 0, 0, 116, 0, 102, 0, 112, 0, 102, 0, 114, 0, 0, 0, 116, 0, 103, 0, 108, 0, 102, 0, 119, 0, 0, 0, 116, 0, 104, 0, 98, 0, 107, 0, 107, 0, 0, 0, 116, 0, 106, 0, 100, 0, 121, 0, 117, 0, 0, 0, 116, 0, 107, 0, 102, 0, 107, 0, 111, 0, 0, 0, 116, 0, 108, 0, 100, 0, 105, 0, 108, 0, 0, 0, 116, 0, 109, 0, 97, 0, 115, 0, 98, 0, 0, 0, 116, 0, 110, 0, 116, 0, 117, 0, 110, 0, 0, 0, 116, 0, 111, 0, 116, 0, 98, 0, 117, 0, 0, 0, 116, 0, 114, 0, 105, 0, 115, 0, 116, 0, 0, 0, 116, 0, 116, 0, 112, 0, 111, 0, 115, 0, 0, 0, 116, 0, 118, 0, 102, 0, 117, 0, 110, 0, 0, 0, 116, 0, 119, 0, 116, 0, 112, 0, 101, 0, 0, 0, 116, 0, 122, 0, 100, 0, 97, 0, 114, 0, 0, 0, 117, 0, 97, 0, 105, 0, 101, 0, 118, 0, 0, 0, 117, 0, 97, 0, 111, 0, 122, 0, 104, 0, 0, 0, 117, 0, 97, 0, 115, 0, 105, 0, 112, 0, 0, 0, 117, 0, 97, 0, 117, 0, 122, 0, 104, 0, 0, 0, 117, 0, 103, 0, 107, 0, 108, 0, 97, 0, 0, 0, 117, 0, 109, 0, 97, 0, 119, 0, 107, 0, 0, 0, 117, 0, 109, 0, 106, 0, 111, 0, 110, 0, 0, 0, 117, 0, 109, 0, 109, 0, 100, 0, 121, 0, 0, 0, 117, 0, 115, 0, 97, 0, 100, 0, 107, 0, 0, 0, 117, 0, 115, 0, 97, 0, 101, 0, 103, 0, 0, 0, 117, 0, 115, 0, 97, 0, 110, 0, 99, 0, 0, 0, 117, 0, 115, 0, 98, 0, 111, 0, 105, 0, 0, 0, 117, 0, 115, 0, 99, 0, 104, 0, 105, 0, 0, 0, 117, 0, 115, 0, 100, 0, 101, 0, 116, 0, 0, 0, 117, 0, 115, 0, 104, 0, 110, 0, 108, 0, 0, 0, 117, 0, 115, 0, 105, 0, 110, 0, 100, 0, 0, 0, 117, 0, 115, 0, 106, 0, 110, 0, 117, 0, 0, 0, 117, 0, 115, 0, 107, 0, 110, 0, 120, 0, 0, 0, 117, 0, 115, 0, 108, 0, 97, 0, 120, 0, 0, 0, 117, 0, 115, 0, 108, 0, 117, 0, 105, 0, 0, 0, 117, 0, 115, 0, 109, 0, 110, 0, 109, 0, 0, 0, 117, 0, 115, 0, 109, 0, 111, 0, 99, 0, 0, 0, 117, 0, 115, 0, 109, 0, 116, 0, 109, 0, 0, 0, 117, 0, 115, 0, 110, 0, 121, 0, 99, 0, 0, 0, 117, 0, 115, 0, 111, 0, 101, 0, 97, 0, 0, 0, 117, 0, 115, 0, 111, 0, 109, 0, 101, 0, 0, 0, 117, 0, 115, 0, 112, 0, 104, 0, 120, 0, 0, 0, 117, 0, 115, 0, 115, 0, 105, 0, 116, 0, 0, 0, 117, 0, 115, 0, 116, 0, 101, 0, 108, 0, 0, 0, 117, 0, 115, 0, 119, 0, 108, 0, 122, 0, 0, 0, 117, 0, 115, 0, 119, 0, 115, 0, 113, 0, 0, 0, 117, 0, 115, 0, 120, 0, 117, 0, 108, 0, 0, 0, 117, 0, 115, 0, 121, 0, 97, 0, 107, 0, 0, 0, 117, 0, 121, 0, 109, 0, 118, 0, 100, 0, 0, 0, 117, 0, 122, 0, 115, 0, 107, 0, 100, 0, 0, 0, 117, 0, 122, 0, 116, 0, 97, 0, 115, 0, 0, 0, 118, 0, 97, 0, 118, 0, 97, 0, 116, 0, 0, 0, 118, 0, 99, 0, 115, 0, 118, 0, 100, 0, 0, 0, 118, 0, 101, 0, 99, 0, 99, 0, 115, 0, 0, 0, 118, 0, 103, 0, 116, 0, 111, 0, 118, 0, 0, 0, 118, 0, 105, 0, 115, 0, 116, 0, 116, 0, 0, 0, 118, 0, 110, 0, 115, 0, 103, 0, 110, 0, 0, 0, 118, 0, 117, 0, 118, 0, 108, 0, 105, 0, 0, 0, 119, 0, 102, 0, 109, 0, 97, 0, 117, 0, 0, 0, 119, 0, 115, 0, 97, 0, 112, 0, 119, 0, 0, 0, 121, 0, 101, 0, 97, 0, 100, 0, 101, 0, 0, 0, 121, 0, 116, 0, 109, 0, 97, 0, 109, 0, 0, 0, 122, 0, 97, 0, 106, 0, 110, 0, 98, 0, 0, 0, 122, 0, 109, 0, 108, 0, 117, 0, 110, 0, 0, 0, 122, 0, 119, 0, 104, 0, 114, 0, 101, 0, 0, 0, 104, 0, 101, 0, 98, 0, 114, 0, 111, 0, 110, 0, 0, 0, 109, 0, 120, 0, 115, 0, 116, 0, 105, 0, 115, 0, 0, 0, 114, 0, 117, 0, 115, 0, 114, 0, 101, 0, 100, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 48, 0, 49, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 48, 0, 50, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 48, 0, 51, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 48, 0, 52, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 48, 0, 53, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 48, 0, 54, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 48, 0, 55, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 48, 0, 56, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 48, 0, 57, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 49, 0, 48, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 49, 0, 49, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 49, 0, 50, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 49, 0, 51, 0, 0, 0, 117, 0, 116, 0, 99, 0, 101, 0, 49, 0, 52, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 48, 0, 49, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 48, 0, 50, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 48, 0, 51, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 48, 0, 52, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 48, 0, 53, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 48, 0, 54, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 48, 0, 55, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 48, 0, 56, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 48, 0, 57, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 49, 0, 48, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 49, 0, 49, 0, 0, 0, 117, 0, 116, 0, 99, 0, 119, 0, 49, 0, 50, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 85, 0, 84, 0, 67, 0, 0, 0, 99, 0, 115, 0, 116, 0, 54, 0, 99, 0, 100, 0, 116, 0, 0, 0, 100, 0, 101, 0, 98, 0, 115, 0, 110, 0, 103, 0, 110, 0, 0, 0, 101, 0, 115, 0, 116, 0, 53, 0, 101, 0, 100, 0, 116, 0, 0, 0, 106, 0, 101, 0, 114, 0, 117, 0, 115, 0, 108, 0, 109, 0, 0, 0, 109, 0, 115, 0, 116, 0, 55, 0, 109, 0, 100, 0, 116, 0, 0, 0, 112, 0, 115, 0, 116, 0, 56, 0, 112, 0, 100, 0, 116, 0, 0, 0, 114, 0, 117, 0, 99, 0, 104, 0, 105, 0, 116, 0, 97, 0, 0, 0, 114, 0, 117, 0, 107, 0, 104, 0, 110, 0, 100, 0, 103, 0, 0, 0, 114, 0, 117, 0, 117, 0, 110, 0, 101, 0, 114, 0, 97, 0, 0, 0, 117, 0, 115, 0, 105, 0, 110, 0, 118, 0, 101, 0, 118, 0, 0, 0, 117, 0, 115, 0, 110, 0, 100, 0, 99, 0, 110, 0, 116, 0, 0, 0, 117, 0, 115, 0, 110, 0, 100, 0, 110, 0, 115, 0, 108, 0, 0, 0, 103, 0, 108, 0, 100, 0, 107, 0, 115, 0, 104, 0, 118, 0, 110, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 53, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 55, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 104, 0, 97, 0, 107, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 99, 0, 97, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 101, 0, 111, 0, 117, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 111, 0, 107, 0, 121, 0, 111, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 49, 0, 48, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 97, 0, 105, 0, 103, 0, 111, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 97, 0, 105, 0, 112, 0, 101, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 101, 0, 104, 0, 114, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 114, 0, 117, 0, 109, 0, 113, 0, 105, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 100, 0, 97, 0, 107, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 114, 0, 117, 0, 107, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 105, 0, 114, 0, 111, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 78, 0, 105, 0, 99, 0, 111, 0, 115, 0, 105, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 82, 0, 97, 0, 110, 0, 103, 0, 111, 0, 111, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 104, 0, 105, 0, 109, 0, 112, 0, 104, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 104, 0, 97, 0, 110, 0, 103, 0, 104, 0, 97, 0, 105, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 111, 0, 110, 0, 100, 0, 111, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 115, 0, 109, 0, 101, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 109, 0, 97, 0, 107, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 117, 0, 106, 0, 117, 0, 121, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 115, 0, 104, 0, 103, 0, 97, 0, 98, 0, 97, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 97, 0, 108, 0, 99, 0, 117, 0, 116, 0, 116, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 116, 0, 109, 0, 97, 0, 110, 0, 100, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 107, 0, 97, 0, 115, 0, 115, 0, 97, 0, 114, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 68, 0, 117, 0, 98, 0, 108, 0, 105, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 105, 0, 115, 0, 98, 0, 111, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 111, 0, 115, 0, 99, 0, 111, 0, 119, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 87, 0, 97, 0, 114, 0, 115, 0, 97, 0, 119, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 101, 0, 110, 0, 118, 0, 101, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 101, 0, 103, 0, 105, 0, 110, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 74, 0, 101, 0, 114, 0, 117, 0, 115, 0, 97, 0, 108, 0, 101, 0, 109, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 114, 0, 105, 0, 112, 0, 111, 0, 108, 0, 105, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 118, 0, 97, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 110, 0, 97, 0, 117, 0, 115, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 110, 0, 103, 0, 95, 0, 75, 0, 111, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 105, 0, 110, 0, 103, 0, 97, 0, 112, 0, 111, 0, 114, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 111, 0, 110, 0, 97, 0, 112, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 114, 0, 100, 0, 111, 0, 98, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 105, 0, 106, 0, 117, 0, 97, 0, 110, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 73, 0, 115, 0, 116, 0, 97, 0, 110, 0, 98, 0, 117, 0, 108, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 104, 0, 105, 0, 99, 0, 97, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 101, 0, 116, 0, 114, 0, 111, 0, 105, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 108, 0, 105, 0, 102, 0, 97, 0, 120, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 97, 0, 109, 0, 97, 0, 105, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 110, 0, 100, 0, 111, 0, 122, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 111, 0, 110, 0, 104, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 104, 0, 111, 0, 101, 0, 110, 0, 105, 0, 120, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 111, 0, 114, 0, 111, 0, 110, 0, 116, 0, 111, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 70, 0, 97, 0, 101, 0, 114, 0, 111, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 80, 0, 101, 0, 114, 0, 116, 0, 104, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 67, 0, 104, 0, 105, 0, 115, 0, 105, 0, 110, 0, 97, 0, 117, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 67, 0, 104, 0, 97, 0, 116, 0, 104, 0, 97, 0, 109, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 83, 0, 121, 0, 100, 0, 110, 0, 101, 0, 121, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 65, 0, 117, 0, 99, 0, 107, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 100, 0, 109, 0, 111, 0, 110, 0, 116, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 122, 0, 97, 0, 116, 0, 108, 0, 97, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 101, 0, 119, 0, 95, 0, 89, 0, 111, 0, 114, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 105, 0, 97, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 74, 0, 111, 0, 104, 0, 110, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 87, 0, 105, 0, 110, 0, 110, 0, 105, 0, 112, 0, 101, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 108, 0, 97, 0, 97, 0, 110, 0, 98, 0, 97, 0, 97, 0, 116, 0, 97, 0, 114, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 68, 0, 97, 0, 114, 0, 119, 0, 105, 0, 110, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 98, 0, 97, 0, 114, 0, 116, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 72, 0, 111, 0, 110, 0, 111, 0, 108, 0, 117, 0, 108, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 116, 0, 97, 0, 109, 0, 97, 0, 114, 0, 99, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 97, 0, 103, 0, 111, 0, 95, 0, 80, 0, 97, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 99, 0, 104, 0, 111, 0, 114, 0, 97, 0, 103, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 111, 0, 95, 0, 80, 0, 97, 0, 117, 0, 108, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 84, 0, 104, 0, 111, 0, 109, 0, 97, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 86, 0, 97, 0, 110, 0, 99, 0, 111, 0, 117, 0, 118, 0, 101, 0, 114, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 75, 0, 119, 0, 97, 0, 106, 0, 97, 0, 108, 0, 101, 0, 105, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 105, 0, 111, 0, 95, 0, 66, 0, 114, 0, 97, 0, 110, 0, 99, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 117, 0, 105, 0, 115, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 87, 0, 104, 0, 105, 0, 116, 0, 101, 0, 104, 0, 111, 0, 114, 0, 115, 0, 101, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 82, 0, 101, 0, 121, 0, 107, 0, 106, 0, 97, 0, 118, 0, 105, 0, 107, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 65, 0, 100, 0, 101, 0, 108, 0, 97, 0, 105, 0, 100, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 66, 0, 114, 0, 105, 0, 115, 0, 98, 0, 97, 0, 110, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 115, 0, 95, 0, 65, 0, 110, 0, 103, 0, 101, 0, 108, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 95, 0, 67, 0, 105, 0, 116, 0, 121, 0, 0, 0, 65, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 47, 0, 76, 0, 111, 0, 110, 0, 103, 0, 121, 0, 101, 0, 97, 0, 114, 0, 98, 0, 121, 0, 101, 0, 110, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 76, 0, 111, 0, 114, 0, 100, 0, 95, 0, 72, 0, 111, 0, 119, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 77, 0, 101, 0, 108, 0, 98, 0, 111, 0, 117, 0, 114, 0, 110, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 112, 0, 111, 0, 108, 0, 105, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 75, 0, 110, 0, 111, 0, 120, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 117, 0, 101, 0, 110, 0, 111, 0, 115, 0, 95, 0, 65, 0, 105, 0, 114, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 114, 0, 97, 0, 108, 0, 95, 0, 72, 0, 97, 0, 114, 0, 98, 0, 111, 0, 117, 0, 114, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 66, 0, 114, 0, 111, 0, 107, 0, 101, 0, 110, 0, 95, 0, 72, 0, 105, 0, 108, 0, 108, 0, 0, 0, 5, 0, 48, 0, 54, 0, 60, 0, 66, 0, 72, 0, 30, 0, 18, 0, 18, 0, 24, 0, 36, 0, 130, 0, 100, 0, 114, 0, 130, 0, 161, 0, 189, 0, 222, 0, 248, 0, 16, 1, 42, 1, 59, 1, 72, 1, 89, 1, 108, 1, 137, 1, 165, 1, 181, 1, 200, 1, 216, 1, 233, 1, 248, 1, 14, 2, 29, 2, 44, 2, 59, 2, 70, 2, 82, 2, 99, 2, 113, 2, 126, 2, 141, 2, 154, 2, 165, 2, 179, 2, 191, 2, 210, 2, 226, 2, 238, 2, 253, 2, 16, 3, 30, 3, 49, 3, 63, 3, 77, 3, 93, 3, 114, 3, 130, 3, 149, 3, 168, 3, 183, 3, 204, 3, 216, 3, 233, 3, 245, 3, 1, 4, 17, 4, 32, 4, 57, 4, 72, 4, 88, 4, 108, 4, 123, 4, 143, 4, 156, 4, 174, 4, 62, 6, 67, 6, 71, 6, 83, 6, 193, 4, 203, 4, 213, 4, 222, 4, 236, 4, 244, 4, 2, 5, 11, 5, 26, 5, 41, 5, 77, 6, 80, 6, 124, 27, 197, 4, 207, 4, 217, 4, 226, 4, 88, 6, 92, 6, 101, 6, 109, 6, 114, 6, 68, 15, 121, 6, 12, 34, 127, 6, 133, 6, 57, 5, 74, 5, 89, 5, 137, 6, 140, 6, 148, 6, 155, 6, 104, 5, 118, 5, 134, 5, 148, 5, 159, 6, 166, 6, 175, 6, 179, 6, 254, 24, 183, 6, 240, 4, 160, 5, 170, 5, 182, 5, 193, 5, 204, 5, 220, 5, 231, 5, 241, 5, 3, 6, 15, 6, 27, 6, 38, 6, 53, 6, 144, 28, 248, 4, 190, 6, 6, 5, 245, 11, 3, 12, 95, 16, 225, 14, 225, 14, 37, 13, 17, 12, 149, 13, 116, 16, 139, 11, 53, 13, 53, 16, 53, 16, 114, 15, 74, 16, 95, 15, 37, 13, 143, 12, 41, 15, 38, 14, 31, 12, 217, 11, 217, 11, 36, 11, 217, 11, 91, 11, 69, 13, 127, 11, 59, 12, 45, 12, 47, 11, 173, 12, 204, 11, 73, 12, 157, 14, 191, 11, 213, 13, 249, 15, 21, 14, 21, 14, 13, 16, 21, 14, 174, 14, 190, 15, 171, 15, 191, 14, 33, 16, 229, 13, 138, 16, 95, 15, 165, 13, 23, 15, 218, 12, 117, 13, 140, 14, 158, 12, 197, 13, 55, 14, 123, 14, 59, 15, 158, 12, 133, 15, 106, 14, 7, 13, 203, 12, 16, 11, 165, 11, 87, 12, 151, 10, 151, 10, 151, 10, 151, 10, 159, 10, 159, 10, 159, 10, 231, 11, 178, 11, 245, 13, 231, 11, 231, 11, 151, 10, 151, 10, 151, 10, 151, 10, 151, 10, 80, 11, 233, 12, 152, 15, 115, 11, 173, 12, 133, 13, 69, 11, 77, 15, 188, 12, 26, 11, 53, 13, 72, 14, 229, 15, 38, 14, 5, 14, 143, 12, 217, 11, 152, 11, 22, 13, 243, 14, 152, 11, 129, 12, 101, 12, 103, 11, 58, 11, 248, 12, 69, 13, 159, 10, 5, 15, 139, 11, 181, 13, 85, 13, 53, 16, 89, 14, 208, 14, 74, 16, 101, 13, 143, 12, 209, 15, 209, 15, 243, 14, 159, 10, 159, 10, 115, 12, 159, 10, 205, 1, 203, 6, 218, 6, 231, 6, 250, 6, 9, 7, 23, 7, 37, 7, 51, 7, 65, 7, 79, 7, 95, 7, 114, 7, 131, 7, 144, 7, 162, 7, 175, 7, 190, 7, 203, 7, 224, 7, 240, 7, 254, 7, 14, 8, 30, 8, 46, 8, 60, 8, 80, 8, 92, 8, 107, 8, 123, 8, 137, 8, 153, 8, 166, 8, 184, 8, 196, 8, 210, 8, 228, 8, 242, 8, 0, 9, 14, 9, 28, 9, 43, 9, 60, 9, 76, 9, 91, 9, 107, 9, 121, 9, 139, 9, 158, 9, 176, 9, 192, 9, 207, 9, 220, 9, 236, 9, 249, 9, 11, 10, 28, 10, 44, 10, 62, 10, 89, 10, 120, 10, 144, 10, 171, 10, 198, 10, 224, 10, 250, 10, 8, 11, 25, 11, 39, 11, 62, 11, 79, 11, 93, 11, 108, 11, 129, 11, 147, 11, 162, 11, 176, 11, 197, 11, 219, 11, 240, 11, 255, 11, 15, 12, 33, 12, 49, 12, 64, 12, 80, 12, 98, 12, 120, 12, 136, 12, 155, 12, 171, 12, 186, 12, 202, 12, 223, 12, 238, 12, 3, 13, 18, 13, 34, 13, 51, 13, 68, 13, 85, 13, 105, 13, 125, 13, 143, 13, 161, 13, 177, 13, 195, 13, 214, 13, 230, 13, 249, 13, 11, 14, 29, 14, 44, 14, 60, 14, 75, 14, 94, 14, 115, 14, 139, 14, 166, 14, 192, 14, 214, 14, 240, 14, 8, 15, 29, 15, 44, 15, 60, 15, 76, 15, 90, 15, 105, 15, 133, 15, 152, 15, 167, 15, 180, 15, 200, 15, 219, 15, 241, 15, 0, 16, 16, 16, 31, 16, 47, 16, 66, 16, 84, 16, 101, 16, 117, 16, 135, 16, 150, 16, 169, 16, 189, 16, 206, 16, 222, 16, 240, 16, 3, 17, 20, 17, 39, 17, 54, 17, 71, 17, 87, 17, 100, 17, 116, 17, 144, 17, 172, 17, 203, 17, 219, 17, 234, 17, 254, 17, 17, 18, 33, 18, 56, 18, 78, 18, 98, 18, 118, 18, 139, 18, 159, 18, 180, 18, 195, 18, 210, 18, 227, 18, 246, 18, 11, 19, 28, 19, 45, 19, 67, 19, 85, 19, 106, 19, 120, 19, 142, 19, 159, 19, 176, 19, 193, 19, 211, 19, 230, 19, 252, 19, 16, 20, 30, 20, 50, 20, 66, 20, 82, 20, 98, 20, 116, 20, 135, 20, 152, 20, 168, 20, 188, 20, 205, 20, 222, 20, 248, 20, 13, 21, 31, 21, 50, 21, 68, 21, 87, 21, 104, 21, 121, 21, 139, 21, 159, 21, 169, 21, 181, 21, 192, 21, 204, 21, 215, 21, 227, 21, 241, 21, 253, 21, 10, 22, 23, 22, 33, 22, 46, 22, 59, 22, 71, 22, 84, 22, 96, 22, 110, 22, 121, 22, 137, 22, 150, 22, 164, 22, 175, 22, 185, 22, 196, 22, 210, 22, 225, 22, 235, 22, 247, 22, 6, 23, 16, 23, 29, 23, 42, 23, 56, 23, 71, 23, 82, 23, 97, 23, 110, 23, 124, 23, 138, 23, 155, 23, 173, 23, 186, 23, 198, 23, 209, 23, 222, 23, 236, 23, 248, 23, 4, 24, 17, 24, 35, 24, 52, 24, 62, 24, 72, 24, 88, 24, 103, 24, 118, 24, 129, 24, 143, 24, 158, 24, 171, 24, 183, 24, 195, 24, 209, 24, 224, 24, 235, 24, 249, 24, 8, 25, 27, 25, 39, 25, 53, 25, 66, 25, 78, 25, 91, 25, 102, 25, 113, 25, 130, 25, 142, 25, 156, 25, 171, 25, 188, 25, 201, 25, 220, 25, 233, 25, 249, 25, 10, 26, 26, 26, 46, 26, 62, 26, 79, 26, 98, 26, 121, 26, 140, 26, 157, 26, 176, 26, 195, 26, 217, 26, 234, 26, 251, 26, 11, 27, 28, 27, 47, 27, 67, 27, 87, 27, 103, 27, 79, 35, 87, 35, 120, 27, 128, 27, 138, 27, 149, 27, 160, 27, 171, 27, 181, 27, 191, 27, 201, 27, 211, 27, 221, 27, 231, 27, 241, 27, 251, 27, 5, 28, 16, 28, 27, 28, 38, 28, 49, 28, 60, 28, 70, 28, 80, 28, 90, 28, 100, 28, 110, 28, 120, 28, 130, 28, 140, 28, 148, 28, 160, 28, 177, 28, 192, 28, 209, 28, 223, 28, 239, 28, 253, 28, 15, 29, 31, 29, 48, 29, 64, 29, 80, 29, 96, 29, 114, 29, 128, 29, 145, 29, 161, 29, 177, 29, 196, 29, 212, 29, 226, 29, 245, 29, 1, 30, 14, 30, 28, 30, 45, 30, 59, 30, 77, 30, 91, 30, 104, 30, 121, 30, 134, 30, 148, 30, 162, 30, 174, 30, 187, 30, 204, 30, 218, 30, 230, 30, 242, 30, 0, 31, 18, 31, 34, 31, 49, 31, 67, 31, 81, 31, 94, 31, 111, 31, 126, 31, 140, 31, 157, 31, 173, 31, 186, 31, 201, 31, 215, 31, 230, 31, 247, 31, 5, 32, 19, 32, 37, 32, 51, 32, 71, 32, 85, 32, 102, 32, 115, 32, 129, 32, 146, 32, 158, 32, 174, 32, 191, 32, 206, 32, 95, 35, 103, 35, 221, 32, 234, 32, 251, 32, 16, 33, 32, 33, 47, 33, 61, 33, 79, 33, 95, 33, 108, 33, 125, 33, 143, 33, 159, 33, 179, 33, 192, 33, 209, 33, 226, 33, 245, 33, 4, 34, 22, 34, 37, 34, 55, 34, 70, 34, 84, 34, 97, 34, 113, 34, 128, 34, 146, 34, 160, 34, 177, 34, 192, 34, 213, 34, 231, 34, 246, 34, 5, 35, 20, 35, 38, 35, 51, 35, 64, 35, 232, 2, 246, 3, 150, 3, 90, 3, 126, 3, 190, 5, 214, 2, 20, 4, 80, 4, 12, 6, 220, 2, 116, 1, 114, 3, 142, 5, 132, 3, 26, 4, 10, 8, 154, 8, 66, 3, 6, 3, 120, 3, 254, 7, 4, 2, 198, 9, 186, 9, 28, 8, 184, 8, 212, 7, 188, 7, 208, 2, 126, 6, 210, 3, 82, 8, 90, 0, 202, 2, 192, 9, 50, 4, 96, 6, 112, 5, 58, 8, 16, 8, 106, 5, 218, 4, 70, 8, 114, 6, 238, 5, 98, 1, 122, 1, 34, 8, 136, 5, 118, 8, 102, 6, 208, 8, 220, 8, 66, 0, 60, 0, 152, 1, 174, 0, 198, 0, 204, 0, 216, 0, 186, 0, 210, 0, 222, 0, 62, 1, 26, 7, 236, 1, 72, 6, 80, 1, 158, 1, 16, 2, 136, 2, 164, 1, 12, 3, 226, 8, 156, 0, 142, 2, 176, 1, 24, 6, 132, 9, 168, 0, 234, 3, 28, 5, 232, 8, 18, 6, 190, 2, 162, 0, 18, 3, 22, 2, 170, 1, 84, 0, 7, 11, 148, 2, 154, 2, 36, 0, 238, 8, 78, 3, 28, 2, 182, 1, 40, 8, 40, 2, 194, 1, 46, 2, 2, 4, 52, 2, 64, 8, 222, 3, 32, 4, 68, 4, 102, 3, 86, 4, 58, 2, 24, 3, 30, 6, 6, 9, 214, 8, 84, 9, 72, 9, 239, 10, 48, 9, 78, 9, 250, 8, 166, 2, 64, 2, 200, 4, 180, 0, 0, 9, 30, 9, 146, 1, 140, 1, 186, 6, 12, 9, 18, 9, 46, 8, 206, 1, 132, 6, 200, 1, 38, 4, 232, 5, 36, 6, 60, 6, 192, 0, 24, 9, 48, 6, 36, 9, 42, 6, 240, 6, 70, 2, 54, 6, 102, 9, 76, 2, 244, 5, 248, 1, 42, 9, 82, 2, 54, 9, 188, 1, 90, 9, 247, 10, 255, 10, 66, 6, 180, 6, 88, 2, 22, 8, 60, 9, 110, 4, 136, 8, 212, 1, 252, 6, 250, 2, 34, 2, 160, 2, 224, 1, 100, 2, 94, 2, 218, 1, 211, 9, 242, 1, 0, 3, 84, 3, 230, 1, 8, 4, 66, 9, 44, 4, 106, 2, 4, 5, 88, 5, 144, 9, 126, 9, 178, 2, 98, 4, 14, 4, 112, 2, 78, 6, 118, 2, 138, 9, 124, 2, 172, 2, 130, 2, 96, 9, 184, 2, 96, 0, 102, 0, 108, 0, 44, 1, 114, 0, 120, 0, 126, 0, 132, 0, 138, 0, 144, 0, 150, 0, 242, 7, 174, 9, 46, 5, 206, 4, 68, 7, 34, 5, 40, 5, 112, 8, 52, 5, 170, 4, 110, 1, 68, 1, 88, 8, 62, 7, 82, 5, 224, 4, 134, 1, 158, 4, 215, 10, 202, 5, 100, 5, 52, 8, 86, 1, 106, 8, 48, 0, 94, 8, 42, 3, 13, 0, 204, 9, 92, 4, 208, 5, 80, 7, 128, 4, 122, 4, 191, 10, 54, 0, 134, 7, 228, 6, 150, 6, 223, 10, 92, 7, 90, 6, 84, 6, 22, 5, 220, 5, 74, 7, 134, 4, 222, 6, 174, 6, 48, 3, 116, 7, 128, 7, 122, 7, 70, 5, 230, 4, 140, 4, 10, 5, 32, 7, 58, 5, 64, 5, 196, 5, 194, 7, 150, 9, 158, 7, 108, 9, 16, 5, 18, 0, 224, 7, 218, 9, 148, 8, 114, 9, 228, 3, 176, 4, 254, 1, 212, 4, 146, 7, 214, 5, 24, 0, 231, 10, 76, 5, 170, 7, 182, 7, 176, 7, 78, 0, 14, 7, 128, 1, 138, 3, 30, 3, 198, 3, 2, 7, 182, 4, 62, 4, 230, 7, 174, 3, 240, 0, 252, 0, 246, 0, 20, 1, 2, 1, 8, 1, 14, 1, 26, 1, 32, 1, 38, 1, 50, 1, 56, 1, 167, 10, 183, 10, 1, 0, 67, 10, 130, 10, 137, 10, 144, 10, 74, 10, 81, 10, 88, 10, 95, 10, 102, 10, 109, 10, 116, 10, 123, 10, 225, 9, 32, 10, 39, 10, 46, 10, 53, 10, 60, 10, 232, 9, 239, 9, 246, 9, 253, 9, 4, 10, 11, 10, 18, 10, 25, 10, 9, 0, 5, 0, 138, 6, 42, 0, 56, 7, 56, 4, 50, 7, 60, 3, 248, 7, 92, 1, 44, 7, 116, 4, 175, 10, 154, 5, 72, 3, 146, 4, 252, 3, 240, 3, 156, 3, 152, 4, 130, 8, 194, 4, 86, 7, 160, 8, 104, 7, 8, 7, 236, 7, 216, 3, 124, 5, 144, 3, 250, 5, 162, 3, 10, 2, 148, 5, 110, 7, 144, 6, 204, 3, 160, 5, 54, 3, 130, 5, 188, 4, 98, 7, 4, 8, 74, 1, 140, 7, 172, 8, 184, 5, 104, 1, 218, 7, 108, 3, 72, 0, 152, 7, 178, 8, 94, 5, 120, 9, 234, 0, 118, 5, 164, 7, 234, 6, 104, 4, 166, 8, 226, 2, 166, 5, 164, 4, 36, 3, 196, 2, 254, 4, 76, 8, 206, 7, 6, 6, 0, 6, 180, 9, 38, 7, 199, 10, 207, 10, 168, 9, 30, 0, 216, 6, 168, 6, 244, 2, 156, 9, 242, 4, 100, 8, 168, 3, 142, 8, 96, 3, 192, 6, 200, 7, 74, 4, 244, 8, 196, 8, 236, 4, 180, 3, 172, 5, 178, 5, 198, 6, 202, 8, 156, 6, 162, 6, 120, 6, 108, 6, 228, 0, 20, 7, 246, 6, 186, 3, 210, 6, 238, 2, 226, 5, 204, 6, 248, 4, 124, 8, 192, 3, 190, 8, 162, 9, 170, 170, 1, 0, 45, 0, 160, 16, 0, 80, 1, 0, 91, 0, 171, 16, 0, 80, 1, 0, 91, 0, 176, 17, 0, 80, 3, 0, 32, 0, 81, 0, 195, 6, 130, 19, 0, 32, 132, 19, 0, 32, 134, 19, 0, 32, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 245, 21, 0, 32, 7, 0, 0, 0, 183, 3, 0, 0, 247, 21, 0, 0, 247, 21, 0, 0, 137, 0, 0, 0, 1, 0, 0, 0, 39, 21, 0, 0, 109, 97, 112, 84, 105, 109, 101, 122, 111, 110, 101, 115, 0, 65, 85, 83, 32, 67, 101, 110, 116, 114, 97, 108, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 48, 48, 49, 0, 65, 85, 0, 65, 85, 83, 32, 69, 97, 115, 116, 101, 114, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 102, 103, 104, 97, 110, 105, 115, 116, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 70, 0, 65, 108, 97, 115, 107, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 85, 83, 0, 65, 108, 101, 117, 116, 105, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 108, 116, 97, 105, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 82, 85, 0, 65, 114, 97, 98, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 72, 0, 75, 87, 0, 81, 65, 0, 83, 65, 0, 89, 69, 0, 65, 114, 97, 98, 105, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 69, 0, 79, 77, 0, 90, 90, 0, 65, 114, 97, 98, 105, 99, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 73, 81, 0, 65, 114, 103, 101, 110, 116, 105, 110, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 82, 0, 65, 115, 116, 114, 97, 107, 104, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 116, 108, 97, 110, 116, 105, 99, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 77, 0, 67, 65, 0, 71, 76, 0, 65, 117, 115, 32, 67, 101, 110, 116, 114, 97, 108, 32, 87, 46, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 122, 101, 114, 98, 97, 105, 106, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 90, 0, 65, 122, 111, 114, 101, 115, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 80, 84, 0, 66, 97, 104, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 82, 0, 66, 97, 110, 103, 108, 97, 100, 101, 115, 104, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 68, 0, 66, 84, 0, 66, 101, 108, 97, 114, 117, 115, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 89, 0, 66, 111, 117, 103, 97, 105, 110, 118, 105, 108, 108, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 80, 71, 0, 67, 97, 110, 97, 100, 97, 32, 67, 101, 110, 116, 114, 97, 108, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 97, 112, 101, 32, 86, 101, 114, 100, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 86, 0, 67, 97, 117, 99, 97, 115, 117, 115, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 77, 0, 67, 101, 110, 46, 32, 65, 117, 115, 116, 114, 97, 108, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 101, 110, 116, 114, 97, 108, 32, 65, 109, 101, 114, 105, 99, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 90, 0, 67, 82, 0, 69, 67, 0, 71, 84, 0, 72, 78, 0, 78, 73, 0, 83, 86, 0, 65, 81, 0, 67, 78, 0, 73, 79, 0, 75, 71, 0, 75, 90, 0, 67, 101, 110, 116, 114, 97, 108, 32, 66, 114, 97, 122, 105, 108, 105, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 101, 110, 116, 114, 97, 108, 32, 69, 117, 114, 111, 112, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 76, 0, 67, 90, 0, 72, 85, 0, 77, 69, 0, 82, 83, 0, 83, 73, 0, 83, 75, 0, 67, 101, 110, 116, 114, 97, 108, 32, 69, 117, 114, 111, 112, 101, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 65, 0, 72, 82, 0, 77, 75, 0, 80, 76, 0, 67, 101, 110, 116, 114, 97, 108, 32, 80, 97, 99, 105, 102, 105, 99, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 70, 77, 0, 78, 67, 0, 83, 66, 0, 86, 85, 0, 67, 101, 110, 116, 114, 97, 108, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 32, 40, 77, 101, 120, 105, 99, 111, 41, 0, 77, 88, 0, 67, 104, 97, 116, 104, 97, 109, 32, 73, 115, 108, 97, 110, 100, 115, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 78, 90, 0, 67, 104, 105, 110, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 72, 75, 0, 77, 79, 0, 67, 117, 98, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 85, 0, 68, 97, 116, 101, 108, 105, 110, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 69, 46, 32, 65, 102, 114, 105, 99, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 68, 74, 0, 69, 82, 0, 69, 84, 0, 75, 69, 0, 75, 77, 0, 77, 71, 0, 83, 79, 0, 83, 83, 0, 84, 90, 0, 85, 71, 0, 89, 84, 0, 69, 46, 32, 65, 117, 115, 116, 114, 97, 108, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 69, 46, 32, 69, 117, 114, 111, 112, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 77, 68, 0, 69, 46, 32, 83, 111, 117, 116, 104, 32, 65, 109, 101, 114, 105, 99, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 69, 97, 115, 116, 101, 114, 32, 73, 115, 108, 97, 110, 100, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 76, 0, 69, 97, 115, 116, 101, 114, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 32, 40, 77, 101, 120, 105, 99, 111, 41, 0, 66, 83, 0, 69, 103, 121, 112, 116, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 69, 71, 0, 69, 107, 97, 116, 101, 114, 105, 110, 98, 117, 114, 103, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 70, 76, 69, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 88, 0, 66, 71, 0, 69, 69, 0, 70, 73, 0, 76, 84, 0, 76, 86, 0, 85, 65, 0, 70, 105, 106, 105, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 70, 74, 0, 71, 77, 84, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 69, 83, 0, 70, 79, 0, 71, 66, 0, 71, 71, 0, 73, 69, 0, 73, 77, 0, 74, 69, 0, 71, 84, 66, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 89, 0, 71, 82, 0, 82, 79, 0, 71, 101, 111, 114, 103, 105, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 71, 69, 0, 71, 114, 101, 101, 110, 108, 97, 110, 100, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 71, 114, 101, 101, 110, 119, 105, 99, 104, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 70, 0, 67, 73, 0, 71, 72, 0, 71, 77, 0, 71, 78, 0, 71, 87, 0, 73, 83, 0, 76, 82, 0, 77, 76, 0, 77, 82, 0, 83, 72, 0, 83, 76, 0, 83, 78, 0, 84, 71, 0, 72, 97, 105, 116, 105, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 72, 84, 0, 72, 97, 119, 97, 105, 105, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 75, 0, 80, 70, 0, 85, 77, 0, 73, 110, 100, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 73, 78, 0, 73, 114, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 73, 82, 0, 73, 115, 114, 97, 101, 108, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 73, 76, 0, 74, 111, 114, 100, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 74, 79, 0, 75, 97, 108, 105, 110, 105, 110, 103, 114, 97, 100, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 75, 82, 0, 76, 105, 98, 121, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 76, 89, 0, 76, 105, 110, 101, 32, 73, 115, 108, 97, 110, 100, 115, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 75, 73, 0, 76, 111, 114, 100, 32, 72, 111, 119, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 77, 97, 103, 97, 100, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 77, 97, 103, 97, 108, 108, 97, 110, 101, 115, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 77, 97, 114, 113, 117, 101, 115, 97, 115, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 77, 97, 117, 114, 105, 116, 105, 117, 115, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 77, 85, 0, 82, 69, 0, 83, 67, 0, 77, 105, 100, 100, 108, 101, 32, 69, 97, 115, 116, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 76, 66, 0, 77, 111, 110, 116, 101, 118, 105, 100, 101, 111, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 85, 89, 0, 77, 111, 114, 111, 99, 99, 111, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 69, 72, 0, 77, 65, 0, 77, 111, 117, 110, 116, 97, 105, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 32, 40, 77, 101, 120, 105, 99, 111, 41, 0, 77, 121, 97, 110, 109, 97, 114, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 67, 0, 77, 77, 0, 78, 46, 32, 67, 101, 110, 116, 114, 97, 108, 32, 65, 115, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 78, 97, 109, 105, 98, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 78, 65, 0, 78, 101, 112, 97, 108, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 78, 80, 0, 78, 101, 119, 32, 90, 101, 97, 108, 97, 110, 100, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 78, 101, 119, 102, 111, 117, 110, 100, 108, 97, 110, 100, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 78, 111, 114, 102, 111, 108, 107, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 78, 70, 0, 78, 111, 114, 116, 104, 32, 65, 115, 105, 97, 32, 69, 97, 115, 116, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 78, 111, 114, 116, 104, 32, 65, 115, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 78, 111, 114, 116, 104, 32, 75, 111, 114, 101, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 75, 80, 0, 79, 109, 115, 107, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 80, 97, 99, 105, 102, 105, 99, 32, 83, 65, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 80, 97, 99, 105, 102, 105, 99, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 32, 40, 77, 101, 120, 105, 99, 111, 41, 0, 80, 97, 107, 105, 115, 116, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 80, 75, 0, 80, 97, 114, 97, 103, 117, 97, 121, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 80, 89, 0, 81, 121, 122, 121, 108, 111, 114, 100, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 82, 111, 109, 97, 110, 99, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 69, 0, 68, 75, 0, 70, 82, 0, 82, 117, 115, 115, 105, 97, 32, 84, 105, 109, 101, 32, 90, 111, 110, 101, 32, 49, 48, 0, 82, 117, 115, 115, 105, 97, 32, 84, 105, 109, 101, 32, 90, 111, 110, 101, 32, 49, 49, 0, 82, 117, 115, 115, 105, 97, 32, 84, 105, 109, 101, 32, 90, 111, 110, 101, 32, 51, 0, 82, 117, 115, 115, 105, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 83, 65, 32, 69, 97, 115, 116, 101, 114, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 70, 75, 0, 71, 70, 0, 83, 82, 0, 83, 65, 32, 80, 97, 99, 105, 102, 105, 99, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 79, 0, 74, 77, 0, 75, 89, 0, 80, 65, 0, 80, 69, 0, 83, 65, 32, 87, 101, 115, 116, 101, 114, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 71, 0, 65, 73, 0, 65, 87, 0, 66, 66, 0, 66, 76, 0, 66, 79, 0, 66, 81, 0, 67, 87, 0, 68, 77, 0, 68, 79, 0, 71, 68, 0, 71, 80, 0, 71, 89, 0, 75, 78, 0, 76, 67, 0, 77, 70, 0, 77, 81, 0, 77, 83, 0, 80, 82, 0, 83, 88, 0, 84, 84, 0, 86, 67, 0, 86, 71, 0, 86, 73, 0, 83, 69, 32, 65, 115, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 67, 88, 0, 73, 68, 0, 75, 72, 0, 76, 65, 0, 84, 72, 0, 86, 78, 0, 83, 97, 105, 110, 116, 32, 80, 105, 101, 114, 114, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 80, 77, 0, 83, 97, 107, 104, 97, 108, 105, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 83, 97, 109, 111, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 87, 83, 0, 83, 97, 111, 32, 84, 111, 109, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 83, 84, 0, 83, 97, 114, 97, 116, 111, 118, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 83, 105, 110, 103, 97, 112, 111, 114, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 78, 0, 77, 89, 0, 80, 72, 0, 83, 71, 0, 83, 111, 117, 116, 104, 32, 65, 102, 114, 105, 99, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 66, 73, 0, 66, 87, 0, 67, 68, 0, 76, 83, 0, 77, 87, 0, 77, 90, 0, 82, 87, 0, 83, 90, 0, 90, 65, 0, 90, 77, 0, 90, 87, 0, 83, 114, 105, 32, 76, 97, 110, 107, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 76, 75, 0, 83, 117, 100, 97, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 83, 68, 0, 83, 121, 114, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 83, 89, 0, 84, 97, 105, 112, 101, 105, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 84, 87, 0, 84, 97, 115, 109, 97, 110, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 84, 111, 99, 97, 110, 116, 105, 110, 115, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 84, 111, 107, 121, 111, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 74, 80, 0, 80, 87, 0, 84, 76, 0, 84, 111, 109, 115, 107, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 84, 111, 110, 103, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 84, 79, 0, 84, 114, 97, 110, 115, 98, 97, 105, 107, 97, 108, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 84, 117, 114, 107, 101, 121, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 84, 82, 0, 84, 117, 114, 107, 115, 32, 65, 110, 100, 32, 67, 97, 105, 99, 111, 115, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 85, 83, 32, 77, 111, 117, 110, 116, 97, 105, 110, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 85, 84, 67, 0, 85, 84, 67, 43, 49, 50, 0, 77, 72, 0, 78, 82, 0, 84, 86, 0, 87, 70, 0, 85, 84, 67, 43, 49, 51, 0, 84, 75, 0, 85, 84, 67, 45, 48, 50, 0, 71, 83, 0, 85, 84, 67, 45, 48, 56, 0, 80, 78, 0, 85, 84, 67, 45, 48, 57, 0, 85, 84, 67, 45, 49, 49, 0, 65, 83, 0, 78, 85, 0, 85, 108, 97, 97, 110, 98, 97, 97, 116, 97, 114, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 77, 78, 0, 86, 101, 110, 101, 122, 117, 101, 108, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 86, 69, 0, 86, 108, 97, 100, 105, 118, 111, 115, 116, 111, 107, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 86, 111, 108, 103, 111, 103, 114, 97, 100, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 87, 46, 32, 65, 117, 115, 116, 114, 97, 108, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 87, 46, 32, 67, 101, 110, 116, 114, 97, 108, 32, 65, 102, 114, 105, 99, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 79, 0, 66, 74, 0, 67, 70, 0, 67, 71, 0, 67, 77, 0, 68, 90, 0, 71, 65, 0, 71, 81, 0, 78, 69, 0, 78, 71, 0, 84, 68, 0, 84, 78, 0, 87, 46, 32, 69, 117, 114, 111, 112, 101, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 65, 68, 0, 65, 84, 0, 67, 72, 0, 68, 69, 0, 71, 73, 0, 73, 84, 0, 76, 73, 0, 76, 85, 0, 77, 67, 0, 77, 84, 0, 78, 76, 0, 78, 79, 0, 83, 69, 0, 83, 74, 0, 83, 77, 0, 86, 65, 0, 87, 46, 32, 77, 111, 110, 103, 111, 108, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 87, 101, 115, 116, 32, 65, 115, 105, 97, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 77, 86, 0, 84, 70, 0, 84, 74, 0, 84, 77, 0, 85, 90, 0, 87, 101, 115, 116, 32, 66, 97, 110, 107, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 80, 83, 0, 87, 101, 115, 116, 32, 80, 97, 99, 105, 102, 105, 99, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 71, 85, 0, 77, 80, 0, 89, 97, 107, 117, 116, 115, 107, 32, 83, 116, 97, 110, 100, 97, 114, 100, 32, 84, 105, 109, 101, 0, 170, 170, 0, 0, 67, 0, 83, 0, 84, 0, 54, 0, 67, 0, 68, 0, 84, 0, 0, 0, 69, 0, 83, 0, 84, 0, 53, 0, 69, 0, 68, 0, 84, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 0, 0, 77, 0, 83, 0, 84, 0, 55, 0, 77, 0, 68, 0, 84, 0, 0, 0, 80, 0, 83, 0, 84, 0, 56, 0, 80, 0, 68, 0, 84, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 107, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 118, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 79, 0, 109, 0, 115, 0, 107, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 50, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 56, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 57, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 100, 0, 101, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 105, 0, 108, 0, 105, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 49, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 51, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 52, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 53, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 54, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 55, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 50, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 51, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 52, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 53, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 54, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 55, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 56, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 57, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 109, 0, 109, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 104, 0, 105, 0, 116, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 104, 0, 97, 0, 107, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 117, 0, 98, 0, 97, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 98, 0, 117, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 101, 0, 111, 0, 117, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 111, 0, 107, 0, 121, 0, 111, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 111, 0, 109, 0, 115, 0, 107, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 49, 0, 49, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 49, 0, 50, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 50, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 51, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 99, 0, 97, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 81, 0, 97, 0, 116, 0, 97, 0, 114, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 49, 0, 48, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 48, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 49, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 52, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 101, 0, 105, 0, 114, 0, 117, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 82, 0, 105, 0, 121, 0, 97, 0, 100, 0, 104, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 97, 0, 105, 0, 112, 0, 101, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 101, 0, 104, 0, 114, 0, 97, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 117, 0, 98, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 109, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 108, 0, 109, 0, 97, 0, 116, 0, 121, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 114, 0, 117, 0, 110, 0, 101, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 101, 0, 98, 0, 114, 0, 111, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 117, 0, 119, 0, 97, 0, 105, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 110, 0, 105, 0, 108, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 117, 0, 115, 0, 99, 0, 97, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 97, 0, 105, 0, 103, 0, 111, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 114, 0, 117, 0, 109, 0, 113, 0, 105, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 75, 0, 105, 0, 101, 0, 118, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 79, 0, 115, 0, 108, 0, 111, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 82, 0, 105, 0, 103, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 82, 0, 111, 0, 109, 0, 101, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 104, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 105, 0, 114, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 97, 0, 103, 0, 111, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 100, 0, 97, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 103, 0, 104, 0, 100, 0, 97, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 110, 0, 103, 0, 107, 0, 111, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 114, 0, 110, 0, 97, 0, 117, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 111, 0, 108, 0, 111, 0, 109, 0, 98, 0, 111, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 73, 0, 114, 0, 107, 0, 117, 0, 116, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 114, 0, 97, 0, 99, 0, 104, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 103, 0, 97, 0, 100, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 82, 0, 97, 0, 110, 0, 103, 0, 111, 0, 111, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 98, 0, 105, 0, 108, 0, 105, 0, 115, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 101, 0, 114, 0, 101, 0, 118, 0, 97, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 105, 0, 110, 0, 115, 0, 107, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 80, 0, 97, 0, 114, 0, 105, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 65, 0, 112, 0, 105, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 70, 0, 105, 0, 106, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 99, 0, 99, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 107, 0, 97, 0, 114, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 117, 0, 110, 0, 105, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 105, 0, 109, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 104, 0, 114, 0, 97, 0, 105, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 105, 0, 115, 0, 104, 0, 107, 0, 101, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 104, 0, 105, 0, 109, 0, 112, 0, 104, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 97, 0, 107, 0, 117, 0, 116, 0, 115, 0, 107, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 97, 0, 108, 0, 116, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 111, 0, 102, 0, 105, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 97, 0, 100, 0, 117, 0, 122, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 111, 0, 99, 0, 111, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 117, 0, 97, 0, 109, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 105, 0, 117, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 114, 0, 117, 0, 107, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 87, 0, 97, 0, 107, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 104, 0, 105, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 97, 0, 108, 0, 99, 0, 117, 0, 116, 0, 116, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 97, 0, 109, 0, 97, 0, 115, 0, 99, 0, 117, 0, 115, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 116, 0, 109, 0, 97, 0, 110, 0, 100, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 97, 0, 107, 0, 104, 0, 97, 0, 108, 0, 105, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 104, 0, 97, 0, 110, 0, 103, 0, 104, 0, 97, 0, 105, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 111, 0, 110, 0, 100, 0, 111, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 109, 0, 97, 0, 114, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 87, 0, 97, 0, 114, 0, 115, 0, 97, 0, 119, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 115, 0, 109, 0, 101, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 109, 0, 97, 0, 107, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 110, 0, 103, 0, 117, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 110, 0, 106, 0, 117, 0, 108, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 105, 0, 115, 0, 115, 0, 97, 0, 117, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 111, 0, 117, 0, 97, 0, 108, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 114, 0, 97, 0, 114, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 105, 0, 103, 0, 97, 0, 108, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 117, 0, 97, 0, 110, 0, 100, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 117, 0, 115, 0, 97, 0, 107, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 108, 0, 97, 0, 98, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 112, 0, 117, 0, 116, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 115, 0, 101, 0, 114, 0, 117, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 105, 0, 97, 0, 109, 0, 101, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 117, 0, 98, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 104, 0, 117, 0, 108, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 115, 0, 104, 0, 103, 0, 97, 0, 98, 0, 97, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 117, 0, 115, 0, 104, 0, 97, 0, 110, 0, 98, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 74, 0, 97, 0, 121, 0, 97, 0, 112, 0, 117, 0, 114, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 107, 0, 97, 0, 115, 0, 115, 0, 97, 0, 114, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 97, 0, 115, 0, 104, 0, 107, 0, 101, 0, 110, 0, 116, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 116, 0, 104, 0, 101, 0, 110, 0, 115, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 101, 0, 114, 0, 108, 0, 105, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 68, 0, 117, 0, 98, 0, 108, 0, 105, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 74, 0, 101, 0, 114, 0, 115, 0, 101, 0, 121, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 111, 0, 110, 0, 97, 0, 99, 0, 111, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 111, 0, 115, 0, 99, 0, 111, 0, 119, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 80, 0, 114, 0, 97, 0, 103, 0, 117, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 107, 0, 111, 0, 112, 0, 106, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 84, 0, 105, 0, 114, 0, 97, 0, 110, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 105, 0, 101, 0, 110, 0, 110, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 90, 0, 97, 0, 103, 0, 114, 0, 101, 0, 98, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 90, 0, 117, 0, 114, 0, 105, 0, 99, 0, 104, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 104, 0, 97, 0, 103, 0, 111, 0, 115, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 111, 0, 109, 0, 111, 0, 114, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 69, 0, 102, 0, 97, 0, 116, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 97, 0, 117, 0, 114, 0, 117, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 97, 0, 108, 0, 97, 0, 117, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 97, 0, 105, 0, 114, 0, 111, 0, 98, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 114, 0, 105, 0, 112, 0, 111, 0, 108, 0, 105, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 111, 0, 103, 0, 111, 0, 116, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 110, 0, 99, 0, 117, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 118, 0, 97, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 97, 0, 95, 0, 80, 0, 97, 0, 122, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 74, 0, 101, 0, 114, 0, 117, 0, 115, 0, 97, 0, 108, 0, 101, 0, 109, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 80, 0, 121, 0, 111, 0, 110, 0, 103, 0, 121, 0, 97, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 81, 0, 121, 0, 122, 0, 121, 0, 108, 0, 111, 0, 114, 0, 100, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 105, 0, 110, 0, 103, 0, 97, 0, 112, 0, 111, 0, 114, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 114, 0, 97, 0, 116, 0, 111, 0, 118, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 98, 0, 105, 0, 100, 0, 106, 0, 97, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 108, 0, 103, 0, 105, 0, 101, 0, 114, 0, 115, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 110, 0, 97, 0, 107, 0, 114, 0, 121, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 97, 0, 109, 0, 112, 0, 97, 0, 108, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 98, 0, 97, 0, 98, 0, 97, 0, 110, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 101, 0, 108, 0, 105, 0, 122, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 121, 0, 109, 0, 97, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 117, 0, 105, 0, 97, 0, 98, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 101, 0, 110, 0, 118, 0, 101, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 121, 0, 97, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 97, 0, 115, 0, 115, 0, 97, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 97, 0, 110, 0, 97, 0, 109, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 101, 0, 103, 0, 105, 0, 110, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 110, 0, 103, 0, 95, 0, 75, 0, 111, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 109, 0, 99, 0, 104, 0, 97, 0, 116, 0, 107, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 86, 0, 105, 0, 101, 0, 110, 0, 116, 0, 105, 0, 97, 0, 110, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 110, 0, 100, 0, 111, 0, 114, 0, 114, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 84, 0, 97, 0, 108, 0, 108, 0, 105, 0, 110, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 97, 0, 116, 0, 105, 0, 99, 0, 97, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 105, 0, 108, 0, 110, 0, 105, 0, 117, 0, 115, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 121, 0, 111, 0, 116, 0, 116, 0, 101, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 82, 0, 101, 0, 117, 0, 110, 0, 105, 0, 111, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 77, 0, 105, 0, 100, 0, 119, 0, 97, 0, 121, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 111, 0, 117, 0, 109, 0, 101, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 83, 0, 97, 0, 105, 0, 112, 0, 97, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 97, 0, 104, 0, 105, 0, 116, 0, 105, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 97, 0, 114, 0, 97, 0, 119, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 87, 0, 97, 0, 108, 0, 108, 0, 105, 0, 115, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 104, 0, 97, 0, 114, 0, 116, 0, 111, 0, 117, 0, 109, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 111, 0, 95, 0, 84, 0, 111, 0, 109, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 87, 0, 105, 0, 110, 0, 100, 0, 104, 0, 111, 0, 101, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 114, 0, 97, 0, 99, 0, 97, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 121, 0, 101, 0, 110, 0, 110, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 111, 0, 100, 0, 116, 0, 104, 0, 97, 0, 98, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 104, 0, 111, 0, 101, 0, 110, 0, 105, 0, 120, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 65, 0, 122, 0, 111, 0, 114, 0, 101, 0, 115, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 69, 0, 117, 0, 99, 0, 108, 0, 97, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 80, 0, 101, 0, 114, 0, 116, 0, 104, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 117, 0, 100, 0, 97, 0, 112, 0, 101, 0, 115, 0, 116, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 67, 0, 104, 0, 105, 0, 115, 0, 105, 0, 110, 0, 97, 0, 117, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 73, 0, 115, 0, 116, 0, 97, 0, 110, 0, 98, 0, 117, 0, 108, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 67, 0, 104, 0, 97, 0, 116, 0, 104, 0, 97, 0, 109, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 111, 0, 114, 0, 102, 0, 111, 0, 108, 0, 107, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 108, 0, 97, 0, 110, 0, 116, 0, 121, 0, 114, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 106, 0, 105, 0, 98, 0, 111, 0, 117, 0, 116, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 108, 0, 95, 0, 65, 0, 97, 0, 105, 0, 117, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 70, 0, 114, 0, 101, 0, 101, 0, 116, 0, 111, 0, 119, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 97, 0, 98, 0, 111, 0, 114, 0, 111, 0, 110, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 105, 0, 110, 0, 115, 0, 104, 0, 97, 0, 115, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 114, 0, 111, 0, 118, 0, 105, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 100, 0, 106, 0, 97, 0, 109, 0, 101, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 116, 0, 105, 0, 103, 0, 117, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 104, 0, 105, 0, 99, 0, 97, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 117, 0, 114, 0, 97, 0, 99, 0, 97, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 114, 0, 101, 0, 110, 0, 97, 0, 100, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 108, 0, 105, 0, 102, 0, 97, 0, 120, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 97, 0, 109, 0, 97, 0, 105, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 110, 0, 97, 0, 103, 0, 117, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 114, 0, 105, 0, 103, 0, 111, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 111, 0, 110, 0, 104, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 79, 0, 106, 0, 105, 0, 110, 0, 97, 0, 103, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 105, 0, 106, 0, 117, 0, 97, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 111, 0, 114, 0, 116, 0, 111, 0, 108, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 80, 0, 104, 0, 110, 0, 111, 0, 109, 0, 95, 0, 80, 0, 101, 0, 110, 0, 104, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 67, 0, 97, 0, 110, 0, 97, 0, 114, 0, 121, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 70, 0, 97, 0, 101, 0, 114, 0, 111, 0, 101, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 32, 0, 69, 0, 116, 0, 99, 0, 47, 0, 85, 0, 84, 0, 67, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 101, 0, 108, 0, 103, 0, 114, 0, 97, 0, 100, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 114, 0, 117, 0, 115, 0, 115, 0, 101, 0, 108, 0, 115, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 71, 0, 117, 0, 101, 0, 114, 0, 110, 0, 115, 0, 101, 0, 121, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 72, 0, 101, 0, 108, 0, 115, 0, 105, 0, 110, 0, 107, 0, 105, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 114, 0, 97, 0, 106, 0, 101, 0, 118, 0, 111, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 108, 0, 100, 0, 105, 0, 118, 0, 101, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 70, 0, 97, 0, 107, 0, 97, 0, 111, 0, 102, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 97, 0, 109, 0, 98, 0, 105, 0, 101, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 115, 0, 117, 0, 110, 0, 99, 0, 105, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 105, 0, 113, 0, 117, 0, 101, 0, 108, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 105, 0, 97, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 74, 0, 111, 0, 104, 0, 110, 0, 115, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 78, 0, 111, 0, 118, 0, 111, 0, 115, 0, 105, 0, 98, 0, 105, 0, 114, 0, 115, 0, 107, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 68, 0, 97, 0, 114, 0, 119, 0, 105, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 117, 0, 99, 0, 104, 0, 97, 0, 114, 0, 101, 0, 115, 0, 116, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 111, 0, 108, 0, 103, 0, 111, 0, 103, 0, 114, 0, 97, 0, 100, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 117, 0, 114, 0, 105, 0, 116, 0, 105, 0, 117, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 65, 0, 117, 0, 99, 0, 107, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 72, 0, 111, 0, 110, 0, 111, 0, 108, 0, 117, 0, 108, 0, 117, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 117, 0, 106, 0, 117, 0, 109, 0, 98, 0, 117, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 103, 0, 97, 0, 100, 0, 105, 0, 115, 0, 104, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 103, 0, 117, 0, 105, 0, 108, 0, 108, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 114, 0, 98, 0, 97, 0, 100, 0, 111, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 111, 0, 109, 0, 105, 0, 110, 0, 105, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 101, 0, 119, 0, 95, 0, 89, 0, 111, 0, 114, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 75, 0, 105, 0, 116, 0, 116, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 76, 0, 117, 0, 99, 0, 105, 0, 97, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 115, 0, 101, 0, 121, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 118, 0, 105, 0, 115, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 121, 0, 111, 0, 119, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 114, 0, 97, 0, 115, 0, 110, 0, 111, 0, 121, 0, 97, 0, 114, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 108, 0, 97, 0, 97, 0, 110, 0, 98, 0, 97, 0, 97, 0, 116, 0, 97, 0, 114, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 86, 0, 108, 0, 97, 0, 100, 0, 105, 0, 118, 0, 111, 0, 115, 0, 116, 0, 111, 0, 107, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 66, 0, 101, 0, 114, 0, 109, 0, 117, 0, 100, 0, 97, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 83, 0, 116, 0, 97, 0, 110, 0, 108, 0, 101, 0, 121, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 98, 0, 97, 0, 114, 0, 116, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 83, 0, 121, 0, 100, 0, 110, 0, 101, 0, 121, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 109, 0, 115, 0, 116, 0, 101, 0, 114, 0, 100, 0, 97, 0, 109, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 115, 0, 116, 0, 114, 0, 97, 0, 107, 0, 104, 0, 97, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 71, 0, 105, 0, 98, 0, 114, 0, 97, 0, 108, 0, 116, 0, 97, 0, 114, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 106, 0, 117, 0, 98, 0, 108, 0, 106, 0, 97, 0, 110, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 97, 0, 114, 0, 105, 0, 101, 0, 104, 0, 97, 0, 109, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 80, 0, 111, 0, 100, 0, 103, 0, 111, 0, 114, 0, 105, 0, 99, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 116, 0, 111, 0, 99, 0, 107, 0, 104, 0, 111, 0, 108, 0, 109, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 104, 0, 114, 0, 105, 0, 115, 0, 116, 0, 109, 0, 97, 0, 115, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 75, 0, 101, 0, 114, 0, 103, 0, 117, 0, 101, 0, 108, 0, 101, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 70, 0, 117, 0, 110, 0, 97, 0, 102, 0, 117, 0, 116, 0, 105, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 74, 0, 111, 0, 104, 0, 110, 0, 115, 0, 116, 0, 111, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 105, 0, 116, 0, 99, 0, 97, 0, 105, 0, 114, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 115, 0, 97, 0, 98, 0, 108, 0, 97, 0, 110, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 97, 0, 103, 0, 117, 0, 97, 0, 105, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 97, 0, 116, 0, 101, 0, 109, 0, 97, 0, 108, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 111, 0, 95, 0, 80, 0, 97, 0, 117, 0, 108, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 77, 0, 97, 0, 114, 0, 113, 0, 117, 0, 101, 0, 115, 0, 97, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 111, 0, 110, 0, 103, 0, 97, 0, 116, 0, 97, 0, 112, 0, 117, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 105, 0, 98, 0, 114, 0, 101, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 117, 0, 98, 0, 117, 0, 109, 0, 98, 0, 97, 0, 115, 0, 104, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 117, 0, 97, 0, 107, 0, 99, 0, 104, 0, 111, 0, 116, 0, 116, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 111, 0, 45, 0, 78, 0, 111, 0, 118, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 99, 0, 104, 0, 111, 0, 114, 0, 97, 0, 103, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 104, 0, 105, 0, 104, 0, 117, 0, 97, 0, 104, 0, 117, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 97, 0, 121, 0, 97, 0, 113, 0, 117, 0, 105, 0, 108, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 116, 0, 97, 0, 109, 0, 111, 0, 114, 0, 111, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 84, 0, 104, 0, 111, 0, 109, 0, 97, 0, 115, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 119, 0, 115, 0, 111, 0, 110, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 86, 0, 111, 0, 115, 0, 116, 0, 111, 0, 107, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 114, 0, 97, 0, 116, 0, 105, 0, 115, 0, 108, 0, 97, 0, 118, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 67, 0, 111, 0, 112, 0, 101, 0, 110, 0, 104, 0, 97, 0, 103, 0, 101, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 117, 0, 120, 0, 101, 0, 109, 0, 98, 0, 111, 0, 117, 0, 114, 0, 103, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 110, 0, 95, 0, 77, 0, 97, 0, 114, 0, 105, 0, 110, 0, 111, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 105, 0, 109, 0, 102, 0, 101, 0, 114, 0, 111, 0, 112, 0, 111, 0, 108, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 69, 0, 110, 0, 100, 0, 101, 0, 114, 0, 98, 0, 117, 0, 114, 0, 121, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 97, 0, 108, 0, 97, 0, 112, 0, 97, 0, 103, 0, 111, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 97, 0, 103, 0, 111, 0, 95, 0, 80, 0, 97, 0, 103, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 82, 0, 97, 0, 114, 0, 111, 0, 116, 0, 111, 0, 110, 0, 103, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 114, 0, 97, 0, 110, 0, 100, 0, 95, 0, 84, 0, 117, 0, 114, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 101, 0, 118, 0, 105, 0, 100, 0, 101, 0, 111, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 114, 0, 101, 0, 100, 0, 110, 0, 101, 0, 107, 0, 111, 0, 108, 0, 121, 0, 109, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 101, 0, 107, 0, 97, 0, 116, 0, 101, 0, 114, 0, 105, 0, 110, 0, 98, 0, 117, 0, 114, 0, 103, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 82, 0, 101, 0, 121, 0, 107, 0, 106, 0, 97, 0, 118, 0, 105, 0, 107, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 75, 0, 97, 0, 108, 0, 105, 0, 110, 0, 105, 0, 110, 0, 103, 0, 114, 0, 97, 0, 100, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 75, 0, 105, 0, 114, 0, 105, 0, 116, 0, 105, 0, 109, 0, 97, 0, 116, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 100, 0, 100, 0, 105, 0, 115, 0, 95, 0, 65, 0, 98, 0, 97, 0, 98, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 114, 0, 97, 0, 122, 0, 122, 0, 97, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 79, 0, 117, 0, 97, 0, 103, 0, 97, 0, 100, 0, 111, 0, 117, 0, 103, 0, 111, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 115, 0, 116, 0, 97, 0, 95, 0, 82, 0, 105, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 97, 0, 100, 0, 101, 0, 108, 0, 111, 0, 117, 0, 112, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 101, 0, 114, 0, 109, 0, 111, 0, 115, 0, 105, 0, 108, 0, 108, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 114, 0, 97, 0, 108, 0, 101, 0, 110, 0, 100, 0, 105, 0, 106, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 114, 0, 116, 0, 105, 0, 110, 0, 105, 0, 113, 0, 117, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 115, 0, 101, 0, 114, 0, 114, 0, 97, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 97, 0, 114, 0, 97, 0, 109, 0, 97, 0, 114, 0, 105, 0, 98, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 86, 0, 105, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 99, 0, 77, 0, 117, 0, 114, 0, 100, 0, 111, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 83, 0, 116, 0, 95, 0, 72, 0, 101, 0, 108, 0, 101, 0, 110, 0, 97, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 65, 0, 100, 0, 101, 0, 108, 0, 97, 0, 105, 0, 100, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 66, 0, 114, 0, 105, 0, 115, 0, 98, 0, 97, 0, 110, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 73, 0, 115, 0, 108, 0, 101, 0, 95, 0, 111, 0, 102, 0, 95, 0, 77, 0, 97, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 111, 0, 104, 0, 97, 0, 110, 0, 110, 0, 101, 0, 115, 0, 98, 0, 117, 0, 114, 0, 103, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 115, 0, 95, 0, 65, 0, 110, 0, 103, 0, 101, 0, 108, 0, 101, 0, 115, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 67, 0, 97, 0, 112, 0, 101, 0, 95, 0, 86, 0, 101, 0, 114, 0, 100, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 76, 0, 111, 0, 114, 0, 100, 0, 95, 0, 72, 0, 111, 0, 119, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 117, 0, 97, 0, 100, 0, 97, 0, 108, 0, 99, 0, 97, 0, 110, 0, 97, 0, 108, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 108, 0, 95, 0, 83, 0, 97, 0, 108, 0, 118, 0, 97, 0, 100, 0, 111, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 95, 0, 67, 0, 105, 0, 116, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 117, 0, 101, 0, 114, 0, 116, 0, 111, 0, 95, 0, 82, 0, 105, 0, 99, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 101, 0, 103, 0, 117, 0, 99, 0, 105, 0, 103, 0, 97, 0, 108, 0, 112, 0, 97, 0, 0, 0, 65, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 47, 0, 76, 0, 111, 0, 110, 0, 103, 0, 121, 0, 101, 0, 97, 0, 114, 0, 98, 0, 121, 0, 101, 0, 110, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 65, 0, 110, 0, 116, 0, 97, 0, 110, 0, 97, 0, 110, 0, 97, 0, 114, 0, 105, 0, 118, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 117, 0, 110, 0, 116, 0, 97, 0, 95, 0, 65, 0, 114, 0, 101, 0, 110, 0, 97, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 66, 0, 111, 0, 117, 0, 103, 0, 97, 0, 105, 0, 110, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 95, 0, 77, 0, 111, 0, 114, 0, 101, 0, 115, 0, 98, 0, 121, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 114, 0, 95, 0, 101, 0, 115, 0, 95, 0, 83, 0, 97, 0, 108, 0, 97, 0, 97, 0, 109, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 108, 0, 97, 0, 110, 0, 99, 0, 45, 0, 83, 0, 97, 0, 98, 0, 108, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 117, 0, 101, 0, 110, 0, 111, 0, 115, 0, 95, 0, 65, 0, 105, 0, 114, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 110, 0, 109, 0, 97, 0, 114, 0, 107, 0, 115, 0, 104, 0, 97, 0, 118, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 112, 0, 111, 0, 108, 0, 105, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 99, 0, 111, 0, 114, 0, 101, 0, 115, 0, 98, 0, 121, 0, 115, 0, 117, 0, 110, 0, 100, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 99, 0, 113, 0, 117, 0, 97, 0, 114, 0, 105, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 114, 0, 97, 0, 108, 0, 95, 0, 72, 0, 97, 0, 114, 0, 98, 0, 111, 0, 117, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 119, 0, 101, 0, 114, 0, 95, 0, 80, 0, 114, 0, 105, 0, 110, 0, 99, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 95, 0, 111, 0, 102, 0, 95, 0, 83, 0, 112, 0, 97, 0, 105, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 111, 0, 95, 0, 68, 0, 111, 0, 109, 0, 105, 0, 110, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 66, 0, 97, 0, 114, 0, 116, 0, 104, 0, 101, 0, 108, 0, 101, 0, 109, 0, 121, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 101, 0, 98, 0, 114, 0, 111, 0, 110, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 71, 0, 97, 0, 122, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 45, 0, 97, 0, 117, 0, 45, 0, 80, 0, 114, 0, 105, 0, 110, 0, 99, 0, 101, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 83, 0, 111, 0, 117, 0, 116, 0, 104, 0, 95, 0, 71, 0, 101, 0, 111, 0, 114, 0, 103, 0, 105, 0, 97, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 117, 0, 109, 0, 111, 0, 110, 0, 116, 0, 68, 0, 85, 0, 114, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 108, 0, 109, 0, 97, 0, 116, 0, 121, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 81, 0, 111, 0, 115, 0, 116, 0, 97, 0, 110, 0, 97, 0, 121, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 109, 0, 99, 0, 104, 0, 97, 0, 116, 0, 107, 0, 97, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 110, 0, 97, 0, 100, 0, 121, 0, 114, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 97, 0, 107, 0, 117, 0, 116, 0, 115, 0, 107, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 104, 0, 97, 0, 110, 0, 100, 0, 121, 0, 103, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 97, 0, 100, 0, 114, 0, 105, 0, 100, 0, 32, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 101, 0, 117, 0, 116, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 111, 0, 115, 0, 99, 0, 111, 0, 119, 0, 32, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 75, 0, 105, 0, 114, 0, 111, 0, 118, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 74, 0, 97, 0, 107, 0, 97, 0, 114, 0, 116, 0, 97, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 80, 0, 111, 0, 110, 0, 116, 0, 105, 0, 97, 0, 110, 0, 97, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 78, 0, 105, 0, 99, 0, 111, 0, 115, 0, 105, 0, 97, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 70, 0, 97, 0, 109, 0, 97, 0, 103, 0, 117, 0, 115, 0, 116, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 101, 0, 110, 0, 118, 0, 101, 0, 114, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 111, 0, 105, 0, 115, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 97, 0, 115, 0, 104, 0, 107, 0, 101, 0, 110, 0, 116, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 97, 0, 109, 0, 97, 0, 114, 0, 107, 0, 97, 0, 110, 0, 100, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 101, 0, 114, 0, 108, 0, 105, 0, 110, 0, 32, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 117, 0, 115, 0, 105, 0, 110, 0, 103, 0, 101, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 111, 0, 110, 0, 97, 0, 112, 0, 101, 0, 32, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 75, 0, 111, 0, 115, 0, 114, 0, 97, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 117, 0, 97, 0, 108, 0, 97, 0, 95, 0, 76, 0, 117, 0, 109, 0, 112, 0, 117, 0, 114, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 117, 0, 99, 0, 104, 0, 105, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 86, 0, 108, 0, 97, 0, 100, 0, 105, 0, 118, 0, 111, 0, 115, 0, 116, 0, 111, 0, 107, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 115, 0, 116, 0, 45, 0, 78, 0, 101, 0, 114, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 105, 0, 115, 0, 98, 0, 111, 0, 110, 0, 32, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 77, 0, 97, 0, 100, 0, 101, 0, 105, 0, 114, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 108, 0, 97, 0, 97, 0, 110, 0, 98, 0, 97, 0, 97, 0, 116, 0, 97, 0, 114, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 104, 0, 111, 0, 105, 0, 98, 0, 97, 0, 108, 0, 115, 0, 97, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 77, 0, 97, 0, 106, 0, 117, 0, 114, 0, 111, 0, 32, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 75, 0, 119, 0, 97, 0, 106, 0, 97, 0, 108, 0, 101, 0, 105, 0, 110, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 98, 0, 97, 0, 114, 0, 116, 0, 32, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 67, 0, 117, 0, 114, 0, 114, 0, 105, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 115, 0, 116, 0, 114, 0, 97, 0, 107, 0, 104, 0, 97, 0, 110, 0, 32, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 85, 0, 108, 0, 121, 0, 97, 0, 110, 0, 111, 0, 118, 0, 115, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 104, 0, 105, 0, 104, 0, 117, 0, 97, 0, 104, 0, 117, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 122, 0, 97, 0, 116, 0, 108, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 114, 0, 97, 0, 115, 0, 110, 0, 111, 0, 121, 0, 97, 0, 114, 0, 115, 0, 107, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 78, 0, 111, 0, 118, 0, 111, 0, 107, 0, 117, 0, 122, 0, 110, 0, 101, 0, 116, 0, 115, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 117, 0, 105, 0, 97, 0, 98, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 109, 0, 112, 0, 111, 0, 95, 0, 71, 0, 114, 0, 97, 0, 110, 0, 100, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 105, 0, 111, 0, 95, 0, 66, 0, 114, 0, 97, 0, 110, 0, 99, 0, 111, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 105, 0, 114, 0, 117, 0, 110, 0, 101, 0, 112, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 101, 0, 103, 0, 105, 0, 110, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 119, 0, 105, 0, 102, 0, 116, 0, 95, 0, 67, 0, 117, 0, 114, 0, 114, 0, 101, 0, 110, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 105, 0, 106, 0, 117, 0, 97, 0, 110, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 97, 0, 95, 0, 73, 0, 115, 0, 97, 0, 98, 0, 101, 0, 108, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 111, 0, 116, 0, 104, 0, 101, 0, 114, 0, 97, 0, 32, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 97, 0, 108, 0, 109, 0, 101, 0, 114, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 83, 0, 121, 0, 100, 0, 110, 0, 101, 0, 121, 0, 32, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 77, 0, 101, 0, 108, 0, 98, 0, 111, 0, 117, 0, 114, 0, 110, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 66, 0, 114, 0, 105, 0, 115, 0, 98, 0, 97, 0, 110, 0, 101, 0, 32, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 76, 0, 105, 0, 110, 0, 100, 0, 101, 0, 109, 0, 97, 0, 110, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 65, 0, 100, 0, 101, 0, 108, 0, 97, 0, 105, 0, 100, 0, 101, 0, 32, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 66, 0, 114, 0, 111, 0, 107, 0, 101, 0, 110, 0, 95, 0, 72, 0, 105, 0, 108, 0, 108, 0, 0, 0, 44, 220, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 79, 0, 114, 0, 97, 0, 108, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 113, 0, 116, 0, 97, 0, 117, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 113, 0, 116, 0, 111, 0, 98, 0, 101, 0, 32, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 116, 0, 121, 0, 114, 0, 97, 0, 117, 0, 0, 0, 45, 220, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 75, 0, 105, 0, 101, 0, 118, 0, 32, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 85, 0, 122, 0, 104, 0, 103, 0, 111, 0, 114, 0, 111, 0, 100, 0, 32, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 90, 0, 97, 0, 112, 0, 111, 0, 114, 0, 111, 0, 122, 0, 104, 0, 121, 0, 101, 0, 0, 0, 51, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 86, 0, 97, 0, 110, 0, 99, 0, 111, 0, 117, 0, 118, 0, 101, 0, 114, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 119, 0, 115, 0, 111, 0, 110, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 87, 0, 104, 0, 105, 0, 116, 0, 101, 0, 104, 0, 111, 0, 114, 0, 115, 0, 101, 0, 0, 0, 52, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 110, 0, 97, 0, 117, 0, 115, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 111, 0, 97, 0, 95, 0, 86, 0, 105, 0, 115, 0, 116, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 111, 0, 95, 0, 86, 0, 101, 0, 108, 0, 104, 0, 111, 0, 0, 0, 56, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 119, 0, 115, 0, 111, 0, 110, 0, 95, 0, 67, 0, 114, 0, 101, 0, 101, 0, 107, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 114, 0, 101, 0, 115, 0, 116, 0, 111, 0, 110, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 70, 0, 111, 0, 114, 0, 116, 0, 95, 0, 78, 0, 101, 0, 108, 0, 115, 0, 111, 0, 110, 0, 0, 0, 66, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 112, 0, 111, 0, 108, 0, 105, 0, 115, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 77, 0, 97, 0, 114, 0, 101, 0, 110, 0, 103, 0, 111, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 86, 0, 101, 0, 118, 0, 97, 0, 121, 0, 0, 0, 67, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 108, 0, 105, 0, 102, 0, 97, 0, 120, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 108, 0, 97, 0, 99, 0, 101, 0, 95, 0, 66, 0, 97, 0, 121, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 111, 0, 111, 0, 115, 0, 101, 0, 95, 0, 66, 0, 97, 0, 121, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 99, 0, 116, 0, 111, 0, 110, 0, 0, 0, 73, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 100, 0, 109, 0, 111, 0, 110, 0, 116, 0, 111, 0, 110, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 109, 0, 98, 0, 114, 0, 105, 0, 100, 0, 103, 0, 101, 0, 95, 0, 66, 0, 97, 0, 121, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 117, 0, 118, 0, 105, 0, 107, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 89, 0, 101, 0, 108, 0, 108, 0, 111, 0, 119, 0, 107, 0, 110, 0, 105, 0, 102, 0, 101, 0, 0, 0, 74, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 87, 0, 105, 0, 110, 0, 110, 0, 105, 0, 112, 0, 101, 0, 103, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 97, 0, 105, 0, 110, 0, 121, 0, 95, 0, 82, 0, 105, 0, 118, 0, 101, 0, 114, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 97, 0, 110, 0, 107, 0, 105, 0, 110, 0, 95, 0, 73, 0, 110, 0, 108, 0, 101, 0, 116, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 101, 0, 115, 0, 111, 0, 108, 0, 117, 0, 116, 0, 101, 0, 0, 0, 75, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 95, 0, 67, 0, 105, 0, 116, 0, 121, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 104, 0, 105, 0, 97, 0, 95, 0, 66, 0, 97, 0, 110, 0, 100, 0, 101, 0, 114, 0, 97, 0, 115, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 114, 0, 105, 0, 100, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 101, 0, 114, 0, 114, 0, 101, 0, 121, 0, 0, 0, 78, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 70, 0, 111, 0, 114, 0, 116, 0, 97, 0, 108, 0, 101, 0, 122, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 101, 0, 108, 0, 101, 0, 109, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 99, 0, 101, 0, 105, 0, 111, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 101, 0, 99, 0, 105, 0, 102, 0, 101, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 97, 0, 114, 0, 101, 0, 109, 0, 0, 0, 94, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 99, 0, 104, 0, 111, 0, 114, 0, 97, 0, 103, 0, 101, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 117, 0, 110, 0, 101, 0, 97, 0, 117, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 116, 0, 108, 0, 97, 0, 107, 0, 97, 0, 116, 0, 108, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 109, 0, 101, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 105, 0, 116, 0, 107, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 89, 0, 97, 0, 107, 0, 117, 0, 116, 0, 97, 0, 116, 0, 0, 0, 104, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 111, 0, 114, 0, 111, 0, 110, 0, 116, 0, 111, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 113, 0, 97, 0, 108, 0, 117, 0, 105, 0, 116, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 114, 0, 101, 0, 97, 0, 108, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 105, 0, 112, 0, 105, 0, 103, 0, 111, 0, 110, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 97, 0, 110, 0, 103, 0, 110, 0, 105, 0, 114, 0, 116, 0, 117, 0, 110, 0, 103, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 104, 0, 117, 0, 110, 0, 100, 0, 101, 0, 114, 0, 95, 0, 66, 0, 97, 0, 121, 0, 0, 0, 156, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 101, 0, 119, 0, 95, 0, 89, 0, 111, 0, 114, 0, 107, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 101, 0, 116, 0, 114, 0, 111, 0, 105, 0, 116, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 80, 0, 101, 0, 116, 0, 101, 0, 114, 0, 115, 0, 98, 0, 117, 0, 114, 0, 103, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 86, 0, 105, 0, 110, 0, 99, 0, 101, 0, 110, 0, 110, 0, 101, 0, 115, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 87, 0, 105, 0, 110, 0, 97, 0, 109, 0, 97, 0, 99, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 101, 0, 110, 0, 116, 0, 117, 0, 99, 0, 107, 0, 121, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 105, 0, 99, 0, 101, 0, 108, 0, 108, 0, 111, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 117, 0, 105, 0, 115, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 167, 220, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 104, 0, 105, 0, 99, 0, 97, 0, 103, 0, 111, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 75, 0, 110, 0, 111, 0, 120, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 84, 0, 101, 0, 108, 0, 108, 0, 95, 0, 67, 0, 105, 0, 116, 0, 121, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 110, 0, 111, 0, 109, 0, 105, 0, 110, 0, 101, 0, 101, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 116, 0, 104, 0, 95, 0, 68, 0, 97, 0, 107, 0, 111, 0, 116, 0, 97, 0, 47, 0, 66, 0, 101, 0, 117, 0, 108, 0, 97, 0, 104, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 116, 0, 104, 0, 95, 0, 68, 0, 97, 0, 107, 0, 111, 0, 116, 0, 97, 0, 47, 0, 67, 0, 101, 0, 110, 0, 116, 0, 101, 0, 114, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 116, 0, 104, 0, 95, 0, 68, 0, 97, 0, 107, 0, 111, 0, 116, 0, 97, 0, 47, 0, 78, 0, 101, 0, 119, 0, 95, 0, 83, 0, 97, 0, 108, 0, 101, 0, 109, 0, 0, 0, 16, 221, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 117, 0, 101, 0, 110, 0, 111, 0, 115, 0, 95, 0, 65, 0, 105, 0, 114, 0, 101, 0, 115, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 76, 0, 97, 0, 95, 0, 82, 0, 105, 0, 111, 0, 106, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 82, 0, 105, 0, 111, 0, 95, 0, 71, 0, 97, 0, 108, 0, 108, 0, 101, 0, 103, 0, 111, 0, 115, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 83, 0, 97, 0, 108, 0, 116, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 95, 0, 74, 0, 117, 0, 97, 0, 110, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 95, 0, 76, 0, 117, 0, 105, 0, 115, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 84, 0, 117, 0, 99, 0, 117, 0, 109, 0, 97, 0, 110, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 85, 0, 115, 0, 104, 0, 117, 0, 97, 0, 105, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 116, 0, 97, 0, 109, 0, 97, 0, 114, 0, 99, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 114, 0, 100, 0, 111, 0, 98, 0, 97, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 117, 0, 106, 0, 117, 0, 121, 0, 32, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 110, 0, 100, 0, 111, 0, 122, 0, 97, 0, 0, 0, 2, 0, 71, 0, 75, 0, 149, 12, 149, 12, 2, 0, 71, 0, 75, 0, 28, 14, 16, 24, 2, 0, 71, 0, 130, 0, 59, 1, 59, 1, 2, 0, 71, 0, 155, 0, 173, 15, 67, 27, 2, 0, 71, 0, 155, 0, 211, 2, 211, 2, 2, 0, 71, 0, 201, 0, 250, 2, 250, 2, 6, 0, 71, 0, 223, 0, 226, 0, 229, 0, 232, 0, 235, 0, 225, 1, 202, 3, 65, 2, 158, 1, 225, 1, 101, 0, 4, 0, 71, 0, 4, 1, 7, 1, 10, 1, 48, 1, 48, 1, 89, 2, 211, 0, 2, 0, 71, 0, 34, 1, 224, 2, 224, 2, 2, 0, 71, 0, 61, 1, 199, 19, 84, 29, 2, 0, 71, 0, 201, 0, 62, 14, 241, 22, 4, 0, 71, 0, 111, 1, 114, 1, 117, 1, 0, 11, 233, 13, 202, 25, 182, 5, 2, 0, 71, 0, 75, 0, 208, 9, 208, 9, 2, 0, 71, 0, 174, 1, 41, 0, 41, 0, 3, 0, 71, 0, 117, 1, 198, 1, 192, 9, 6, 20, 192, 9, 2, 0, 71, 0, 221, 1, 102, 4, 102, 4, 3, 0, 71, 0, 249, 1, 252, 1, 37, 1, 37, 1, 228, 3, 2, 0, 71, 0, 21, 2, 98, 3, 98, 3, 2, 0, 71, 0, 51, 2, 115, 19, 115, 19, 2, 0, 71, 0, 114, 1, 96, 8, 161, 23, 3, 0, 71, 0, 108, 2, 10, 1, 170, 18, 170, 18, 121, 0, 2, 0, 71, 0, 134, 2, 85, 3, 85, 3, 2, 0, 71, 0, 75, 0, 73, 18, 91, 24, 9, 0, 71, 0, 196, 2, 199, 2, 202, 2, 205, 2, 208, 2, 211, 2, 214, 2, 10, 1, 29, 15, 247, 7, 139, 17, 151, 16, 29, 15, 34, 19, 32, 11, 230, 18, 161, 0, 7, 0, 71, 0, 217, 2, 220, 2, 223, 2, 226, 2, 229, 2, 10, 1, 29, 2, 25, 16, 113, 2, 178, 6, 215, 3, 252, 20, 231, 0, 2, 0, 71, 0, 221, 1, 21, 8, 89, 23, 8, 0, 71, 0, 37, 3, 40, 3, 43, 3, 46, 3, 49, 3, 52, 3, 55, 3, 240, 9, 122, 6, 94, 6, 240, 9, 130, 14, 192, 11, 96, 14, 43, 16, 5, 0, 71, 0, 89, 3, 92, 3, 95, 3, 98, 3, 214, 4, 0, 12, 150, 6, 108, 6, 214, 4, 7, 0, 71, 0, 75, 0, 131, 3, 134, 3, 137, 3, 140, 3, 10, 1, 210, 18, 27, 20, 18, 22, 5, 9, 210, 18, 206, 6, 191, 1, 5, 0, 71, 0, 114, 1, 174, 3, 155, 0, 10, 1, 208, 10, 90, 26, 227, 15, 171, 28, 1, 0, 2, 0, 71, 0, 174, 3, 250, 18, 166, 26, 2, 0, 71, 0, 207, 3, 32, 10, 32, 10, 4, 0, 71, 0, 220, 2, 230, 3, 233, 3, 172, 4, 172, 4, 111, 8, 147, 1, 2, 0, 71, 0, 255, 3, 52, 7, 52, 7, 2, 0, 71, 0, 10, 1, 114, 1, 114, 1, 14, 0, 71, 0, 217, 2, 49, 4, 52, 4, 55, 4, 58, 4, 61, 4, 64, 4, 67, 4, 70, 4, 73, 4, 76, 4, 79, 4, 10, 1, 248, 6, 165, 13, 80, 10, 228, 4, 82, 17, 248, 6, 192, 6, 74, 19, 12, 13, 5, 2, 157, 19, 217, 7, 216, 8, 201, 0, 2, 0, 71, 0, 75, 0, 92, 18, 53, 24, 2, 0, 71, 0, 133, 4, 0, 10, 0, 10, 2, 0, 71, 0, 221, 1, 47, 15, 47, 15, 2, 0, 71, 0, 195, 4, 157, 7, 157, 7, 5, 0, 71, 0, 229, 4, 114, 1, 155, 0, 10, 1, 80, 13, 66, 8, 163, 27, 13, 28, 9, 0, 2, 0, 71, 0, 174, 3, 37, 7, 37, 7, 2, 0, 71, 0, 252, 4, 185, 2, 185, 2, 2, 0, 71, 0, 201, 0, 6, 17, 6, 17, 8, 0, 71, 0, 44, 5, 47, 5, 50, 5, 53, 5, 56, 5, 59, 5, 62, 5, 125, 2, 113, 14, 11, 4, 171, 8, 240, 11, 201, 8, 149, 2, 178, 24, 2, 0, 71, 0, 84, 5, 137, 3, 137, 3, 9, 0, 71, 0, 105, 5, 108, 5, 111, 5, 114, 5, 117, 5, 120, 5, 123, 5, 198, 1, 186, 4, 144, 11, 160, 11, 186, 4, 224, 11, 38, 6, 111, 18, 52, 6, 110, 22, 4, 0, 71, 0, 144, 5, 147, 5, 150, 5, 166, 12, 158, 21, 10, 6, 166, 12, 2, 0, 71, 0, 176, 5, 72, 3, 72, 3, 2, 0, 71, 0, 117, 1, 160, 9, 160, 9, 15, 0, 71, 0, 227, 5, 230, 5, 233, 5, 236, 5, 239, 5, 242, 5, 245, 5, 248, 5, 251, 5, 254, 5, 1, 6, 4, 6, 7, 6, 10, 6, 25, 17, 120, 17, 172, 7, 150, 3, 14, 5, 202, 7, 28, 5, 25, 17, 160, 10, 242, 4, 137, 15, 54, 18, 112, 10, 163, 3, 17, 2, 2, 0, 71, 0, 33, 6, 180, 20, 180, 20, 6, 0, 71, 0, 59, 6, 62, 6, 65, 6, 155, 0, 10, 1, 234, 12, 187, 16, 35, 9, 215, 14, 234, 12, 169, 1, 2, 0, 71, 0, 88, 6, 116, 4, 116, 4, 2, 0, 71, 0, 110, 6, 249, 1, 249, 1, 2, 0, 71, 0, 134, 6, 82, 7, 82, 7, 2, 0, 71, 0, 158, 6, 15, 1, 15, 1, 2, 0, 71, 0, 201, 0, 44, 17, 44, 17, 2, 0, 71, 0, 187, 6, 70, 1, 70, 1, 2, 0, 71, 0, 210, 6, 7, 7, 7, 7, 3, 0, 71, 0, 240, 6, 10, 1, 63, 17, 63, 17, 202, 1, 2, 0, 71, 0, 75, 0, 190, 18, 190, 18, 2, 0, 71, 0, 201, 0, 46, 3, 46, 3, 2, 0, 71, 0, 195, 4, 94, 19, 94, 19, 2, 0, 71, 0, 62, 6, 65, 15, 65, 15, 4, 0, 71, 0, 106, 7, 109, 7, 112, 7, 200, 12, 200, 12, 231, 8, 173, 2, 2, 0, 71, 0, 141, 7, 213, 1, 213, 1, 2, 0, 71, 0, 169, 7, 224, 16, 224, 16, 3, 0, 71, 0, 194, 7, 197, 7, 249, 14, 96, 10, 249, 14, 5, 0, 71, 0, 114, 1, 174, 3, 155, 0, 10, 1, 36, 8, 15, 26, 80, 11, 186, 21, 25, 0, 2, 0, 71, 0, 174, 3, 191, 15, 19, 23, 3, 0, 71, 0, 254, 7, 1, 8, 59, 3, 37, 4, 59, 3, 2, 0, 71, 0, 201, 0, 132, 12, 132, 12, 2, 0, 71, 0, 56, 8, 112, 9, 112, 9, 2, 0, 71, 0, 79, 8, 144, 4, 144, 4, 3, 0, 71, 0, 217, 2, 207, 3, 217, 12, 35, 18, 217, 12, 2, 0, 71, 0, 114, 1, 115, 12, 115, 12, 2, 0, 71, 0, 157, 8, 48, 10, 48, 10, 2, 0, 71, 0, 201, 0, 20, 3, 20, 3, 2, 0, 71, 0, 201, 0, 182, 13, 54, 23, 2, 0, 71, 0, 241, 8, 97, 7, 97, 7, 2, 0, 71, 0, 201, 0, 61, 0, 61, 0, 2, 0, 71, 0, 195, 4, 98, 12, 98, 12, 4, 0, 71, 0, 114, 1, 155, 0, 10, 1, 150, 18, 225, 24, 150, 18, 33, 0, 2, 0, 71, 0, 174, 3, 96, 11, 198, 23, 2, 0, 71, 0, 86, 9, 33, 3, 33, 3, 2, 0, 71, 0, 112, 9, 64, 12, 64, 12, 2, 0, 71, 0, 229, 2, 112, 7, 112, 7, 5, 0, 71, 0, 161, 9, 164, 9, 105, 5, 167, 9, 111, 3, 208, 11, 61, 16, 76, 21, 111, 3, 2, 0, 71, 0, 201, 0, 243, 16, 243, 16, 2, 0, 71, 0, 201, 0, 126, 8, 22, 21, 2, 0, 71, 0, 201, 0, 200, 4, 200, 4, 3, 0, 71, 0, 201, 0, 62, 5, 80, 6, 103, 21, 115, 16, 7, 0, 71, 0, 217, 2, 221, 1, 20, 10, 23, 10, 26, 10, 10, 1, 144, 9, 235, 23, 243, 26, 250, 13, 144, 9, 253, 17, 131, 0, 10, 0, 71, 0, 221, 1, 114, 1, 54, 10, 202, 2, 57, 10, 60, 10, 63, 10, 66, 10, 10, 1, 22, 7, 125, 23, 48, 20, 22, 7, 209, 15, 16, 11, 6, 8, 81, 8, 189, 3, 151, 0, 28, 0, 71, 0, 94, 10, 97, 10, 100, 10, 103, 10, 106, 10, 109, 10, 112, 10, 221, 1, 114, 1, 115, 10, 118, 10, 121, 10, 124, 10, 127, 10, 130, 10, 133, 10, 136, 10, 139, 10, 142, 10, 145, 10, 148, 10, 151, 10, 154, 10, 157, 10, 160, 10, 163, 10, 10, 1, 67, 7, 192, 10, 29, 13, 168, 5, 46, 13, 136, 20, 67, 7, 196, 17, 22, 25, 178, 19, 224, 10, 63, 13, 114, 20, 240, 10, 158, 17, 51, 8, 97, 13, 114, 13, 48, 11, 215, 17, 234, 17, 14, 19, 70, 20, 92, 20, 16, 18, 112, 11, 245, 15, 141, 0, 9, 0, 71, 0, 217, 2, 188, 10, 191, 10, 194, 10, 197, 10, 200, 10, 203, 10, 10, 1, 237, 2, 148, 13, 164, 14, 130, 21, 128, 11, 141, 8, 237, 2, 101, 2, 241, 0, 2, 0, 71, 0, 233, 10, 81, 12, 81, 12, 2, 0, 71, 0, 201, 0, 158, 4, 158, 4, 2, 0, 71, 0, 23, 11, 124, 3, 124, 3, 2, 0, 71, 0, 49, 11, 96, 9, 96, 9, 2, 0, 71, 0, 201, 0, 142, 7, 142, 7, 8, 0, 71, 0, 217, 2, 98, 11, 191, 10, 101, 11, 104, 11, 107, 11, 10, 1, 127, 7, 131, 13, 41, 2, 238, 5, 48, 22, 77, 2, 127, 7, 251, 0, 13, 0, 71, 0, 137, 11, 140, 11, 143, 11, 146, 11, 149, 11, 152, 11, 155, 11, 158, 11, 161, 11, 164, 11, 167, 11, 10, 1, 130, 18, 251, 12, 128, 10, 119, 15, 140, 5, 64, 10, 126, 5, 70, 5, 232, 7, 130, 18, 98, 5, 56, 5, 191, 0, 2, 0, 71, 0, 194, 11, 7, 3, 7, 3, 2, 0, 71, 0, 217, 11, 80, 9, 80, 9, 2, 0, 71, 0, 240, 11, 130, 4, 130, 4, 2, 0, 71, 0, 8, 12, 237, 1, 237, 1, 2, 0, 71, 0, 75, 0, 11, 14, 207, 22, 2, 0, 71, 0, 221, 1, 11, 15, 11, 15, 6, 0, 71, 0, 191, 10, 78, 12, 81, 12, 84, 12, 10, 1, 81, 1, 224, 5, 81, 1, 234, 6, 111, 0, 5, 1, 2, 0, 71, 0, 201, 0, 92, 1, 92, 1, 2, 0, 71, 0, 127, 12, 83, 15, 83, 15, 2, 0, 71, 0, 201, 0, 26, 1, 26, 1, 2, 0, 71, 0, 177, 12, 16, 10, 16, 10, 2, 0, 71, 0, 238, 12, 205, 16, 205, 16, 2, 0, 71, 0, 155, 0, 241, 19, 134, 25, 5, 0, 71, 0, 114, 1, 174, 3, 155, 0, 10, 1, 176, 9, 76, 25, 177, 17, 176, 9, 171, 0, 3, 0, 71, 0, 117, 1, 10, 1, 17, 0, 220, 19, 176, 11, 8, 0, 71, 0, 240, 6, 248, 12, 251, 12, 254, 12, 65, 6, 1, 13, 10, 1, 125, 1, 50, 9, 174, 22, 220, 6, 198, 14, 89, 4, 65, 9, 125, 1, 4, 0, 71, 0, 240, 6, 11, 13, 10, 1, 136, 1, 133, 16, 32, 12, 136, 1, 4, 0, 71, 0, 221, 1, 21, 13, 10, 1, 71, 0, 64, 11, 203, 20, 71, 0, 3, 0, 71, 0, 31, 13, 10, 1, 81, 0, 232, 14, 81, 0, 3, 0, 71, 0, 62, 6, 10, 1, 91, 0, 48, 12, 91, 0, 5, 0, 71, 0, 48, 13, 51, 13, 65, 6, 10, 1, 103, 1, 169, 16, 63, 4, 246, 8, 103, 1, 2, 0, 71, 0, 80, 13, 199, 13, 141, 22, 2, 0, 71, 0, 107, 13, 128, 9, 128, 9, 2, 0, 71, 0, 201, 0, 216, 13, 79, 22, 2, 0, 71, 0, 201, 0, 183, 12, 183, 12, 2, 0, 71, 0, 75, 0, 224, 9, 224, 9, 15, 0, 71, 0, 219, 13, 222, 13, 143, 11, 225, 13, 228, 13, 231, 13, 234, 13, 237, 13, 240, 13, 243, 13, 246, 13, 249, 13, 252, 13, 10, 1, 198, 2, 84, 5, 155, 15, 144, 10, 0, 5, 101, 17, 42, 5, 187, 7, 101, 15, 112, 5, 154, 5, 198, 2, 176, 10, 176, 3, 181, 0, 17, 0, 71, 0, 23, 14, 26, 14, 29, 14, 32, 14, 35, 14, 38, 14, 41, 14, 44, 14, 47, 14, 50, 14, 53, 14, 56, 14, 59, 14, 62, 14, 65, 14, 68, 14, 24, 6, 156, 8, 136, 6, 164, 6, 244, 21, 79, 14, 161, 2, 24, 4, 79, 16, 66, 6, 254, 3, 45, 14, 137, 2, 147, 14, 54, 19, 97, 16, 186, 8, 2, 0, 71, 0, 80, 13, 51, 0, 51, 0, 9, 0, 71, 0, 217, 2, 229, 2, 121, 14, 124, 14, 127, 14, 130, 14, 133, 14, 10, 1, 252, 5, 7, 16, 132, 24, 16, 12, 181, 14, 210, 5, 196, 5, 215, 21, 221, 0, 2, 0, 71, 0, 160, 14, 53, 2, 158, 20, 7, 0, 71, 0, 217, 2, 131, 3, 190, 14, 193, 14, 51, 2, 10, 1, 136, 19, 226, 20, 76, 4, 50, 4, 20, 9, 136, 19, 180, 1, 2, 0, 71, 0, 201, 0, 241, 3, 49, 21, 170, 170, 137, 0, 45, 0, 78, 0, 104, 0, 133, 0, 158, 0, 181, 0, 204, 0, 238, 0, 13, 1, 37, 1, 64, 1, 88, 1, 120, 1, 149, 1, 177, 1, 201, 1, 224, 1, 255, 1, 24, 2, 54, 2, 83, 2, 111, 2, 137, 2, 166, 2, 7, 8, 232, 2, 8, 3, 58, 3, 101, 3, 61, 2, 143, 3, 177, 3, 210, 3, 236, 3, 2, 4, 25, 4, 82, 4, 109, 4, 136, 4, 167, 4, 82, 0, 198, 4, 232, 4, 255, 4, 26, 5, 65, 5, 87, 5, 126, 5, 153, 5, 179, 5, 203, 5, 13, 6, 36, 6, 68, 6, 91, 6, 113, 6, 137, 6, 161, 6, 221, 8, 190, 6, 213, 6, 243, 6, 11, 7, 33, 7, 58, 7, 82, 7, 115, 7, 144, 7, 172, 7, 214, 12, 200, 7, 232, 7, 4, 8, 34, 8, 59, 8, 82, 8, 108, 8, 135, 8, 160, 8, 190, 8, 215, 8, 244, 8, 7, 9, 168, 14, 32, 9, 63, 9, 89, 9, 115, 9, 139, 9, 170, 9, 190, 9, 210, 9, 229, 9, 251, 9, 29, 10, 69, 10, 166, 10, 206, 10, 236, 10, 3, 11, 26, 11, 52, 11, 74, 11, 110, 11, 170, 11, 197, 11, 220, 11, 243, 11, 11, 12, 34, 12, 58, 12, 87, 12, 107, 12, 130, 12, 156, 12, 180, 12, 79, 0, 211, 12, 237, 12, 241, 12, 4, 13, 14, 13, 24, 13, 34, 13, 41, 13, 54, 13, 83, 13, 110, 13, 136, 13, 160, 13, 187, 13, 255, 13, 71, 14, 97, 14, 136, 14, 163, 14, 196, 14, 102, 30, 0, 80, 107, 30, 0, 80, 112, 30, 0, 80, 117, 30, 0, 80, 122, 30, 0, 80, 127, 30, 0, 80, 132, 30, 0, 80, 145, 30, 0, 80, 154, 30, 0, 80, 159, 30, 0, 80, 164, 30, 0, 80, 169, 30, 0, 80, 178, 30, 0, 80, 183, 30, 0, 80, 188, 30, 0, 80, 195, 30, 0, 80, 200, 30, 0, 80, 207, 30, 0, 80, 212, 30, 0, 80, 217, 30, 0, 80, 222, 30, 0, 80, 229, 30, 0, 80, 234, 30, 0, 80, 239, 30, 0, 80, 2, 31, 0, 80, 17, 31, 0, 80, 22, 31, 0, 80, 39, 31, 0, 80, 50, 31, 0, 80, 65, 31, 0, 80, 76, 31, 0, 80, 81, 31, 0, 80, 86, 31, 0, 80, 95, 31, 0, 80, 100, 31, 0, 80, 105, 31, 0, 80, 134, 31, 0, 80, 139, 31, 0, 80, 144, 31, 0, 80, 149, 31, 0, 80, 154, 31, 0, 80, 165, 31, 0, 80, 170, 31, 0, 80, 175, 31, 0, 80, 180, 31, 0, 80, 197, 31, 0, 80, 202, 31, 0, 80, 221, 31, 0, 80, 230, 31, 0, 80, 235, 31, 0, 80, 240, 31, 0, 80, 15, 32, 0, 80, 20, 32, 0, 80, 33, 32, 0, 80, 38, 32, 0, 80, 43, 32, 0, 80, 48, 32, 0, 80, 53, 32, 0, 80, 58, 32, 0, 80, 63, 32, 0, 80, 68, 32, 0, 80, 75, 32, 0, 80, 80, 32, 0, 80, 85, 32, 0, 80, 90, 32, 0, 80, 95, 32, 0, 80, 104, 32, 0, 80, 109, 32, 0, 80, 114, 32, 0, 80, 121, 32, 0, 80, 132, 32, 0, 80, 137, 32, 0, 80, 144, 32, 0, 80, 149, 32, 0, 80, 154, 32, 0, 80, 159, 32, 0, 80, 166, 32, 0, 80, 171, 32, 0, 80, 176, 32, 0, 80, 181, 32, 0, 80, 186, 32, 0, 80, 191, 32, 0, 80, 196, 32, 0, 80, 201, 32, 0, 80, 210, 32, 0, 80, 215, 32, 0, 80, 220, 32, 0, 80, 225, 32, 0, 80, 230, 32, 0, 80, 241, 32, 0, 80, 246, 32, 0, 80, 251, 32, 0, 80, 0, 33, 0, 80, 7, 33, 0, 80, 22, 33, 0, 80, 43, 33, 0, 80, 100, 33, 0, 80, 119, 33, 0, 80, 124, 33, 0, 80, 129, 33, 0, 80, 134, 33, 0, 80, 139, 33, 0, 80, 144, 33, 0, 80, 161, 33, 0, 80, 188, 33, 0, 80, 193, 33, 0, 80, 198, 33, 0, 80, 203, 33, 0, 80, 208, 33, 0, 80, 213, 33, 0, 80, 218, 33, 0, 80, 231, 33, 0, 80, 236, 33, 0, 80, 241, 33, 0, 80, 246, 33, 0, 80, 251, 33, 0, 80, 0, 34, 0, 80, 5, 34, 0, 80, 16, 34, 0, 80, 23, 34, 0, 80, 40, 34, 0, 80, 49, 34, 0, 80, 58, 34, 0, 80, 65, 34, 0, 80, 72, 34, 0, 80, 83, 34, 0, 80, 88, 34, 0, 80, 93, 34, 0, 80, 98, 34, 0, 80, 103, 34, 0, 80, 108, 34, 0, 80, 139, 34, 0, 80, 174, 34, 0, 80, 179, 34, 0, 80, 198, 34, 0, 80, 203, 34, 0, 80, 218, 34, 0, 80, 1, 0, 32, 0, 39, 21, 0, 32, 170, 170, 170, 170, 32, 0, 218, 39, 20, 0, 0, 0, 0, 0, 2, 0, 82, 101, 115, 66, 2, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 176, 151, 0, 32, 7, 0, 0, 0, 78, 0, 0, 0, 184, 151, 0, 0, 184, 151, 0, 0, 29, 0, 0, 0, 1, 0, 0, 0, 113, 22, 0, 0, 84, 90, 86, 101, 114, 115, 105, 111, 110, 0, 90, 111, 110, 101, 115, 0, 116, 114, 97, 110, 115, 0, 116, 121, 112, 101, 79, 102, 102, 115, 101, 116, 115, 0, 116, 121, 112, 101, 77, 97, 112, 0, 108, 105, 110, 107, 115, 0, 116, 114, 97, 110, 115, 80, 114, 101, 51, 50, 0, 116, 114, 97, 110, 115, 80, 111, 115, 116, 51, 50, 0, 102, 105, 110, 97, 108, 82, 117, 108, 101, 0, 102, 105, 110, 97, 108, 82, 97, 119, 0, 102, 105, 110, 97, 108, 89, 101, 97, 114, 0, 78, 97, 109, 101, 115, 0, 82, 117, 108, 101, 115, 0, 65, 78, 0, 65, 83, 0, 65, 84, 0, 65, 86, 0, 67, 45, 69, 117, 114, 0, 67, 97, 110, 97, 100, 97, 0, 67, 104, 97, 116, 104, 97, 109, 0, 67, 104, 105, 108, 101, 0, 67, 117, 98, 97, 0, 69, 85, 0, 69, 85, 65, 115, 105, 97, 0, 70, 105, 106, 105, 0, 72, 97, 105, 116, 105, 0, 73, 114, 97, 110, 0, 74, 111, 114, 100, 97, 110, 0, 76, 72, 0, 76, 101, 98, 97, 110, 111, 110, 0, 77, 101, 120, 105, 99, 111, 0, 77, 111, 108, 100, 111, 118, 97, 0, 78, 90, 0, 80, 97, 108, 101, 115, 116, 105, 110, 101, 0, 80, 97, 114, 97, 0, 83, 121, 114, 105, 97, 0, 83, 121, 115, 116, 101, 109, 86, 0, 84, 104, 117, 108, 101, 0, 84, 114, 111, 108, 108, 0, 85, 83, 0, 87, 83, 0, 90, 105, 111, 110, 0, 82, 101, 103, 105, 111, 110, 115, 0, 170, 170, 0, 0, 85, 0, 83, 0, 0, 0, 69, 0, 85, 0, 0, 0, 67, 0, 65, 0, 0, 0, 82, 0, 85, 0, 0, 0, 65, 0, 85, 0, 0, 0, 66, 0, 82, 0, 0, 0, 65, 0, 82, 0, 0, 0, 77, 0, 88, 0, 0, 0, 65, 0, 81, 0, 0, 0, 67, 0, 78, 0, 0, 0, 78, 0, 90, 0, 0, 0, 75, 0, 90, 0, 0, 0, 70, 0, 77, 0, 0, 0, 65, 0, 83, 0, 0, 0, 67, 0, 76, 0, 0, 0, 71, 0, 66, 0, 0, 0, 73, 0, 68, 0, 0, 0, 71, 0, 76, 0, 0, 0, 77, 0, 78, 0, 0, 0, 80, 0, 84, 0, 0, 0, 85, 0, 65, 0, 0, 0, 66, 0, 68, 0, 0, 0, 67, 0, 89, 0, 0, 0, 69, 0, 71, 0, 0, 0, 69, 0, 83, 0, 0, 0, 73, 0, 76, 0, 0, 0, 74, 0, 80, 0, 0, 0, 75, 0, 73, 0, 0, 0, 77, 0, 72, 0, 0, 0, 80, 0, 70, 0, 0, 0, 84, 0, 82, 0, 0, 0, 85, 0, 77, 0, 0, 0, 86, 0, 78, 0, 0, 0, 87, 0, 83, 0, 0, 0, 65, 0, 77, 0, 0, 0, 65, 0, 78, 0, 0, 0, 66, 0, 84, 0, 0, 0, 67, 0, 68, 0, 0, 0, 67, 0, 73, 0, 0, 0, 67, 0, 85, 0, 0, 0, 68, 0, 69, 0, 0, 0, 69, 0, 67, 0, 0, 0, 69, 0, 82, 0, 0, 0, 70, 0, 79, 0, 0, 0, 70, 0, 82, 0, 0, 0, 72, 0, 75, 0, 0, 0, 73, 0, 69, 0, 0, 0, 73, 0, 82, 0, 0, 0, 73, 0, 83, 0, 0, 0, 74, 0, 77, 0, 0, 0, 75, 0, 69, 0, 0, 0, 75, 0, 82, 0, 0, 0, 76, 0, 89, 0, 0, 0, 77, 0, 68, 0, 0, 0, 77, 0, 77, 0, 0, 0, 77, 0, 79, 0, 0, 0, 77, 0, 89, 0, 0, 0, 77, 0, 90, 0, 0, 0, 78, 0, 79, 0, 0, 0, 78, 0, 80, 0, 0, 0, 80, 0, 71, 0, 0, 0, 80, 0, 75, 0, 0, 0, 80, 0, 76, 0, 0, 0, 80, 0, 82, 0, 0, 0, 80, 0, 83, 0, 0, 0, 83, 0, 66, 0, 0, 0, 83, 0, 71, 0, 0, 0, 84, 0, 77, 0, 0, 0, 84, 0, 87, 0, 0, 0, 85, 0, 90, 0, 0, 0, 65, 0, 68, 0, 0, 0, 65, 0, 69, 0, 0, 0, 65, 0, 70, 0, 0, 0, 65, 0, 71, 0, 0, 0, 65, 0, 73, 0, 0, 0, 65, 0, 76, 0, 0, 0, 65, 0, 79, 0, 0, 0, 65, 0, 86, 0, 0, 0, 65, 0, 87, 0, 0, 0, 65, 0, 88, 0, 0, 0, 65, 0, 90, 0, 0, 0, 66, 0, 65, 0, 0, 0, 66, 0, 66, 0, 0, 0, 66, 0, 69, 0, 0, 0, 66, 0, 70, 0, 0, 0, 66, 0, 71, 0, 0, 0, 66, 0, 72, 0, 0, 0, 66, 0, 73, 0, 0, 0, 66, 0, 74, 0, 0, 0, 66, 0, 76, 0, 0, 0, 66, 0, 77, 0, 0, 0, 66, 0, 78, 0, 0, 0, 66, 0, 79, 0, 0, 0, 66, 0, 81, 0, 0, 0, 66, 0, 83, 0, 0, 0, 66, 0, 87, 0, 0, 0, 66, 0, 89, 0, 0, 0, 66, 0, 90, 0, 0, 0, 67, 0, 67, 0, 0, 0, 67, 0, 70, 0, 0, 0, 67, 0, 71, 0, 0, 0, 67, 0, 72, 0, 0, 0, 67, 0, 75, 0, 0, 0, 67, 0, 77, 0, 0, 0, 67, 0, 79, 0, 0, 0, 67, 0, 82, 0, 0, 0, 67, 0, 86, 0, 0, 0, 67, 0, 87, 0, 0, 0, 67, 0, 88, 0, 0, 0, 67, 0, 90, 0, 0, 0, 68, 0, 74, 0, 0, 0, 68, 0, 75, 0, 0, 0, 68, 0, 77, 0, 0, 0, 68, 0, 79, 0, 0, 0, 68, 0, 90, 0, 0, 0, 69, 0, 69, 0, 0, 0, 69, 0, 72, 0, 0, 0, 70, 0, 73, 0, 0, 0, 70, 0, 74, 0, 0, 0, 70, 0, 75, 0, 0, 0, 71, 0, 65, 0, 0, 0, 71, 0, 68, 0, 0, 0, 71, 0, 69, 0, 0, 0, 71, 0, 70, 0, 0, 0, 71, 0, 71, 0, 0, 0, 71, 0, 72, 0, 0, 0, 71, 0, 73, 0, 0, 0, 71, 0, 77, 0, 0, 0, 71, 0, 78, 0, 0, 0, 71, 0, 80, 0, 0, 0, 71, 0, 81, 0, 0, 0, 71, 0, 82, 0, 0, 0, 71, 0, 83, 0, 0, 0, 71, 0, 85, 0, 0, 0, 71, 0, 87, 0, 0, 0, 71, 0, 89, 0, 0, 0, 72, 0, 78, 0, 0, 0, 72, 0, 82, 0, 0, 0, 72, 0, 84, 0, 0, 0, 72, 0, 85, 0, 0, 0, 73, 0, 77, 0, 0, 0, 73, 0, 79, 0, 0, 0, 73, 0, 81, 0, 0, 0, 74, 0, 69, 0, 0, 0, 74, 0, 79, 0, 0, 0, 75, 0, 71, 0, 0, 0, 75, 0, 72, 0, 0, 0, 75, 0, 77, 0, 0, 0, 75, 0, 78, 0, 0, 0, 75, 0, 80, 0, 0, 0, 75, 0, 87, 0, 0, 0, 75, 0, 89, 0, 0, 0, 76, 0, 65, 0, 0, 0, 76, 0, 66, 0, 0, 0, 76, 0, 67, 0, 0, 0, 76, 0, 72, 0, 0, 0, 76, 0, 73, 0, 0, 0, 76, 0, 75, 0, 0, 0, 76, 0, 82, 0, 0, 0, 76, 0, 83, 0, 0, 0, 76, 0, 85, 0, 0, 0, 76, 0, 86, 0, 0, 0, 77, 0, 65, 0, 0, 0, 77, 0, 67, 0, 0, 0, 77, 0, 69, 0, 0, 0, 77, 0, 70, 0, 0, 0, 77, 0, 71, 0, 0, 0, 77, 0, 75, 0, 0, 0, 77, 0, 76, 0, 0, 0, 77, 0, 80, 0, 0, 0, 77, 0, 81, 0, 0, 0, 77, 0, 82, 0, 0, 0, 77, 0, 83, 0, 0, 0, 77, 0, 85, 0, 0, 0, 77, 0, 86, 0, 0, 0, 77, 0, 87, 0, 0, 0, 78, 0, 65, 0, 0, 0, 78, 0, 67, 0, 0, 0, 78, 0, 69, 0, 0, 0, 78, 0, 70, 0, 0, 0, 78, 0, 71, 0, 0, 0, 78, 0, 73, 0, 0, 0, 78, 0, 76, 0, 0, 0, 78, 0, 82, 0, 0, 0, 78, 0, 85, 0, 0, 0, 79, 0, 77, 0, 0, 0, 80, 0, 65, 0, 0, 0, 80, 0, 69, 0, 0, 0, 80, 0, 72, 0, 0, 0, 80, 0, 77, 0, 0, 0, 80, 0, 78, 0, 0, 0, 80, 0, 87, 0, 0, 0, 80, 0, 89, 0, 0, 0, 81, 0, 65, 0, 0, 0, 82, 0, 69, 0, 0, 0, 82, 0, 79, 0, 0, 0, 82, 0, 83, 0, 0, 0, 82, 0, 87, 0, 0, 0, 83, 0, 65, 0, 0, 0, 83, 0, 67, 0, 0, 0, 83, 0, 68, 0, 0, 0, 83, 0, 69, 0, 0, 0, 83, 0, 72, 0, 0, 0, 83, 0, 73, 0, 0, 0, 83, 0, 74, 0, 0, 0, 83, 0, 75, 0, 0, 0, 83, 0, 76, 0, 0, 0, 83, 0, 77, 0, 0, 0, 83, 0, 78, 0, 0, 0, 83, 0, 79, 0, 0, 0, 83, 0, 82, 0, 0, 0, 83, 0, 83, 0, 0, 0, 83, 0, 86, 0, 0, 0, 83, 0, 88, 0, 0, 0, 83, 0, 89, 0, 0, 0, 83, 0, 90, 0, 0, 0, 84, 0, 68, 0, 0, 0, 84, 0, 70, 0, 0, 0, 84, 0, 71, 0, 0, 0, 84, 0, 72, 0, 0, 0, 84, 0, 74, 0, 0, 0, 84, 0, 75, 0, 0, 0, 84, 0, 76, 0, 0, 0, 84, 0, 78, 0, 0, 0, 84, 0, 79, 0, 0, 0, 84, 0, 86, 0, 0, 0, 84, 0, 90, 0, 0, 0, 85, 0, 71, 0, 0, 0, 85, 0, 89, 0, 0, 0, 86, 0, 65, 0, 0, 0, 86, 0, 67, 0, 0, 0, 86, 0, 69, 0, 0, 0, 86, 0, 71, 0, 0, 0, 86, 0, 73, 0, 0, 0, 86, 0, 85, 0, 0, 0, 87, 0, 70, 0, 0, 0, 89, 0, 69, 0, 0, 0, 89, 0, 84, 0, 0, 0, 90, 0, 65, 0, 0, 0, 90, 0, 77, 0, 0, 0, 90, 0, 87, 0, 0, 0, 48, 0, 48, 0, 49, 0, 0, 0, 67, 0, 84, 0, 84, 0, 0, 0, 65, 0, 71, 0, 84, 0, 0, 0, 77, 0, 73, 0, 84, 0, 0, 0, 80, 0, 76, 0, 84, 0, 0, 0, 86, 0, 83, 0, 84, 0, 0, 0, 87, 0, 69, 0, 84, 0, 0, 0, 65, 0, 69, 0, 84, 0, 0, 0, 65, 0, 82, 0, 84, 0, 0, 0, 65, 0, 83, 0, 84, 0, 0, 0, 66, 0, 69, 0, 84, 0, 0, 0, 66, 0, 83, 0, 84, 0, 0, 0, 67, 0, 65, 0, 84, 0, 0, 0, 67, 0, 69, 0, 84, 0, 0, 0, 67, 0, 78, 0, 84, 0, 0, 0, 67, 0, 83, 0, 84, 0, 0, 0, 69, 0, 65, 0, 84, 0, 0, 0, 69, 0, 67, 0, 84, 0, 0, 0, 69, 0, 69, 0, 84, 0, 0, 0, 69, 0, 83, 0, 84, 0, 0, 0, 72, 0, 83, 0, 84, 0, 0, 0, 73, 0, 69, 0, 84, 0, 0, 0, 73, 0, 83, 0, 84, 0, 0, 0, 74, 0, 83, 0, 84, 0, 0, 0, 77, 0, 69, 0, 84, 0, 0, 0, 77, 0, 83, 0, 84, 0, 0, 0, 78, 0, 69, 0, 84, 0, 0, 0, 78, 0, 83, 0, 84, 0, 0, 0, 80, 0, 78, 0, 84, 0, 0, 0, 80, 0, 82, 0, 67, 0, 0, 0, 80, 0, 82, 0, 84, 0, 0, 0, 80, 0, 83, 0, 84, 0, 0, 0, 82, 0, 79, 0, 67, 0, 0, 0, 82, 0, 79, 0, 75, 0, 0, 0, 83, 0, 83, 0, 84, 0, 0, 0, 67, 0, 117, 0, 98, 0, 97, 0, 0, 0, 73, 0, 114, 0, 97, 0, 110, 0, 0, 0, 80, 0, 97, 0, 114, 0, 97, 0, 0, 0, 87, 0, 45, 0, 83, 0, 85, 0, 0, 0, 90, 0, 105, 0, 111, 0, 110, 0, 0, 0, 67, 0, 45, 0, 69, 0, 117, 0, 114, 0, 0, 0, 67, 0, 104, 0, 105, 0, 108, 0, 101, 0, 0, 0, 50, 0, 48, 0, 49, 0, 57, 0, 99, 0, 0, 0, 69, 0, 103, 0, 121, 0, 112, 0, 116, 0, 0, 0, 72, 0, 97, 0, 105, 0, 116, 0, 105, 0, 0, 0, 74, 0, 97, 0, 112, 0, 97, 0, 110, 0, 0, 0, 76, 0, 105, 0, 98, 0, 121, 0, 97, 0, 0, 0, 83, 0, 121, 0, 114, 0, 105, 0, 97, 0, 0, 0, 67, 0, 97, 0, 110, 0, 97, 0, 100, 0, 97, 0, 0, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 0, 0, 69, 0, 85, 0, 65, 0, 115, 0, 105, 0, 97, 0, 0, 0, 73, 0, 115, 0, 114, 0, 97, 0, 101, 0, 108, 0, 0, 0, 74, 0, 111, 0, 114, 0, 100, 0, 97, 0, 110, 0, 0, 0, 78, 0, 97, 0, 118, 0, 97, 0, 106, 0, 111, 0, 0, 0, 80, 0, 111, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 84, 0, 117, 0, 114, 0, 107, 0, 101, 0, 121, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 0, 0, 78, 0, 90, 0, 45, 0, 67, 0, 72, 0, 65, 0, 84, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 85, 0, 84, 0, 67, 0, 0, 0, 71, 0, 66, 0, 45, 0, 69, 0, 105, 0, 114, 0, 101, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 85, 0, 67, 0, 84, 0, 0, 0, 70, 0, 97, 0, 99, 0, 116, 0, 111, 0, 114, 0, 121, 0, 0, 0, 73, 0, 99, 0, 101, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 76, 0, 101, 0, 98, 0, 97, 0, 110, 0, 111, 0, 110, 0, 0, 0, 77, 0, 111, 0, 108, 0, 100, 0, 111, 0, 118, 0, 97, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 48, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 90, 0, 117, 0, 108, 0, 117, 0, 0, 0, 72, 0, 111, 0, 110, 0, 103, 0, 107, 0, 111, 0, 110, 0, 103, 0, 0, 0, 80, 0, 111, 0, 114, 0, 116, 0, 117, 0, 103, 0, 97, 0, 108, 0, 0, 0, 85, 0, 83, 0, 47, 0, 83, 0, 97, 0, 109, 0, 111, 0, 97, 0, 0, 0, 80, 0, 97, 0, 108, 0, 101, 0, 115, 0, 116, 0, 105, 0, 110, 0, 101, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 48, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 48, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 100, 0, 101, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 107, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 105, 0, 108, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 71, 0, 97, 0, 122, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 118, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 79, 0, 109, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 79, 0, 114, 0, 97, 0, 108, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 49, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 50, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 51, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 52, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 53, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 54, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 55, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 56, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 57, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 50, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 51, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 52, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 53, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 54, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 55, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 56, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 57, 0, 0, 0, 85, 0, 83, 0, 47, 0, 65, 0, 108, 0, 97, 0, 115, 0, 107, 0, 97, 0, 0, 0, 85, 0, 83, 0, 47, 0, 72, 0, 97, 0, 119, 0, 97, 0, 105, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 109, 0, 109, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 113, 0, 116, 0, 97, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 104, 0, 105, 0, 116, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 97, 0, 99, 0, 99, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 104, 0, 97, 0, 107, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 117, 0, 98, 0, 97, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 98, 0, 117, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 99, 0, 97, 0, 111, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 99, 0, 97, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 81, 0, 97, 0, 116, 0, 97, 0, 114, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 101, 0, 111, 0, 117, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 111, 0, 107, 0, 121, 0, 111, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 111, 0, 109, 0, 115, 0, 107, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 49, 0, 48, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 49, 0, 49, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 43, 0, 49, 0, 50, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 48, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 49, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 50, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 51, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 77, 0, 84, 0, 45, 0, 49, 0, 52, 0, 0, 0, 85, 0, 83, 0, 47, 0, 65, 0, 114, 0, 105, 0, 122, 0, 111, 0, 110, 0, 97, 0, 0, 0, 85, 0, 83, 0, 47, 0, 67, 0, 101, 0, 110, 0, 116, 0, 114, 0, 97, 0, 108, 0, 0, 0, 85, 0, 83, 0, 47, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 85, 0, 83, 0, 47, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 117, 0, 98, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 109, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 108, 0, 109, 0, 97, 0, 116, 0, 121, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 110, 0, 97, 0, 100, 0, 121, 0, 114, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 113, 0, 116, 0, 111, 0, 98, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 116, 0, 121, 0, 114, 0, 97, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 101, 0, 105, 0, 114, 0, 117, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 114, 0, 117, 0, 110, 0, 101, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 97, 0, 114, 0, 98, 0, 105, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 101, 0, 98, 0, 114, 0, 111, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 117, 0, 119, 0, 97, 0, 105, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 110, 0, 105, 0, 108, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 117, 0, 115, 0, 99, 0, 97, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 82, 0, 105, 0, 121, 0, 97, 0, 100, 0, 104, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 97, 0, 105, 0, 103, 0, 111, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 97, 0, 105, 0, 112, 0, 101, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 101, 0, 104, 0, 114, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 104, 0, 105, 0, 109, 0, 98, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 114, 0, 117, 0, 109, 0, 113, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 97, 0, 110, 0, 103, 0, 111, 0, 110, 0, 0, 0, 66, 0, 114, 0, 97, 0, 122, 0, 105, 0, 108, 0, 47, 0, 65, 0, 99, 0, 114, 0, 101, 0, 0, 0, 66, 0, 114, 0, 97, 0, 122, 0, 105, 0, 108, 0, 47, 0, 69, 0, 97, 0, 115, 0, 116, 0, 0, 0, 66, 0, 114, 0, 97, 0, 122, 0, 105, 0, 108, 0, 47, 0, 87, 0, 101, 0, 115, 0, 116, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 85, 0, 110, 0, 107, 0, 110, 0, 111, 0, 119, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 75, 0, 105, 0, 101, 0, 118, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 79, 0, 115, 0, 108, 0, 111, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 82, 0, 105, 0, 103, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 82, 0, 111, 0, 109, 0, 101, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 104, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 89, 0, 97, 0, 112, 0, 0, 0, 85, 0, 83, 0, 47, 0, 65, 0, 108, 0, 101, 0, 117, 0, 116, 0, 105, 0, 97, 0, 110, 0, 0, 0, 85, 0, 83, 0, 47, 0, 77, 0, 105, 0, 99, 0, 104, 0, 105, 0, 103, 0, 97, 0, 110, 0, 0, 0, 85, 0, 83, 0, 47, 0, 77, 0, 111, 0, 117, 0, 110, 0, 116, 0, 97, 0, 105, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 70, 0, 105, 0, 106, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 99, 0, 99, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 105, 0, 114, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 101, 0, 117, 0, 116, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 107, 0, 97, 0, 114, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 97, 0, 103, 0, 111, 0, 115, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 117, 0, 110, 0, 105, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 100, 0, 97, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 116, 0, 107, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 105, 0, 109, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 109, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 103, 0, 104, 0, 100, 0, 97, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 104, 0, 114, 0, 97, 0, 105, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 110, 0, 103, 0, 107, 0, 111, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 97, 0, 114, 0, 110, 0, 97, 0, 117, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 66, 0, 105, 0, 115, 0, 104, 0, 107, 0, 101, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 111, 0, 108, 0, 111, 0, 109, 0, 98, 0, 111, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 73, 0, 114, 0, 107, 0, 117, 0, 116, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 74, 0, 97, 0, 107, 0, 97, 0, 114, 0, 116, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 114, 0, 97, 0, 99, 0, 104, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 115, 0, 104, 0, 103, 0, 97, 0, 114, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 111, 0, 108, 0, 107, 0, 97, 0, 116, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 117, 0, 99, 0, 104, 0, 105, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 103, 0, 97, 0, 100, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 78, 0, 105, 0, 99, 0, 111, 0, 115, 0, 105, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 82, 0, 97, 0, 110, 0, 103, 0, 111, 0, 111, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 98, 0, 105, 0, 108, 0, 105, 0, 115, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 104, 0, 105, 0, 109, 0, 112, 0, 104, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 97, 0, 107, 0, 117, 0, 116, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 101, 0, 114, 0, 101, 0, 118, 0, 97, 0, 110, 0, 0, 0, 67, 0, 97, 0, 110, 0, 97, 0, 100, 0, 97, 0, 47, 0, 89, 0, 117, 0, 107, 0, 111, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 75, 0, 105, 0, 114, 0, 111, 0, 118, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 97, 0, 108, 0, 116, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 105, 0, 110, 0, 115, 0, 107, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 80, 0, 97, 0, 114, 0, 105, 0, 115, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 111, 0, 102, 0, 105, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 97, 0, 100, 0, 117, 0, 122, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 111, 0, 99, 0, 111, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 65, 0, 112, 0, 105, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 117, 0, 97, 0, 109, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 105, 0, 117, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 114, 0, 117, 0, 107, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 87, 0, 97, 0, 107, 0, 101, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 65, 0, 83, 0, 84, 0, 52, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 67, 0, 83, 0, 84, 0, 54, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 69, 0, 83, 0, 84, 0, 53, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 77, 0, 83, 0, 84, 0, 55, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 80, 0, 83, 0, 84, 0, 56, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 89, 0, 83, 0, 84, 0, 57, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 71, 0, 114, 0, 101, 0, 101, 0, 110, 0, 119, 0, 105, 0, 99, 0, 104, 0, 0, 0, 69, 0, 116, 0, 99, 0, 47, 0, 85, 0, 110, 0, 105, 0, 118, 0, 101, 0, 114, 0, 115, 0, 97, 0, 108, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 104, 0, 117, 0, 108, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 65, 0, 67, 0, 84, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 115, 0, 109, 0, 97, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 115, 0, 109, 0, 101, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 109, 0, 97, 0, 107, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 110, 0, 103, 0, 117, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 110, 0, 106, 0, 117, 0, 108, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 105, 0, 115, 0, 115, 0, 97, 0, 117, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 111, 0, 117, 0, 97, 0, 108, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 114, 0, 97, 0, 114, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 105, 0, 103, 0, 97, 0, 108, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 117, 0, 97, 0, 110, 0, 100, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 117, 0, 115, 0, 97, 0, 107, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 108, 0, 97, 0, 98, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 112, 0, 117, 0, 116, 0, 111, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 115, 0, 101, 0, 114, 0, 117, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 105, 0, 97, 0, 109, 0, 101, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 117, 0, 98, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 104, 0, 105, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 101, 0, 108, 0, 101, 0, 109, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 111, 0, 105, 0, 115, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 117, 0, 106, 0, 117, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 105, 0, 116, 0, 107, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 115, 0, 104, 0, 103, 0, 97, 0, 98, 0, 97, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 97, 0, 108, 0, 99, 0, 117, 0, 116, 0, 116, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 97, 0, 109, 0, 97, 0, 115, 0, 99, 0, 117, 0, 115, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 68, 0, 117, 0, 115, 0, 104, 0, 97, 0, 110, 0, 98, 0, 101, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 73, 0, 115, 0, 116, 0, 97, 0, 110, 0, 98, 0, 117, 0, 108, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 74, 0, 97, 0, 121, 0, 97, 0, 112, 0, 117, 0, 114, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 116, 0, 109, 0, 97, 0, 110, 0, 100, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 104, 0, 97, 0, 110, 0, 100, 0, 121, 0, 103, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 77, 0, 97, 0, 107, 0, 97, 0, 115, 0, 115, 0, 97, 0, 114, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 81, 0, 111, 0, 115, 0, 116, 0, 97, 0, 110, 0, 97, 0, 121, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 97, 0, 107, 0, 104, 0, 97, 0, 108, 0, 105, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 104, 0, 97, 0, 110, 0, 103, 0, 104, 0, 97, 0, 105, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 97, 0, 115, 0, 104, 0, 107, 0, 101, 0, 110, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 84, 0, 101, 0, 108, 0, 95, 0, 65, 0, 118, 0, 105, 0, 118, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 115, 0, 116, 0, 45, 0, 78, 0, 101, 0, 114, 0, 97, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 76, 0, 72, 0, 73, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 78, 0, 83, 0, 87, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 116, 0, 104, 0, 101, 0, 110, 0, 115, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 101, 0, 114, 0, 108, 0, 105, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 68, 0, 117, 0, 98, 0, 108, 0, 105, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 74, 0, 101, 0, 114, 0, 115, 0, 101, 0, 121, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 105, 0, 115, 0, 98, 0, 111, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 111, 0, 110, 0, 100, 0, 111, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 97, 0, 100, 0, 114, 0, 105, 0, 100, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 111, 0, 110, 0, 97, 0, 99, 0, 111, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 111, 0, 115, 0, 99, 0, 111, 0, 119, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 80, 0, 114, 0, 97, 0, 103, 0, 117, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 109, 0, 97, 0, 114, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 107, 0, 111, 0, 112, 0, 106, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 84, 0, 105, 0, 114, 0, 97, 0, 110, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 105, 0, 101, 0, 110, 0, 110, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 87, 0, 97, 0, 114, 0, 115, 0, 97, 0, 119, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 90, 0, 97, 0, 103, 0, 114, 0, 101, 0, 98, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 90, 0, 117, 0, 114, 0, 105, 0, 99, 0, 104, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 104, 0, 97, 0, 103, 0, 111, 0, 115, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 111, 0, 109, 0, 111, 0, 114, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 67, 0, 104, 0, 117, 0, 117, 0, 107, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 69, 0, 102, 0, 97, 0, 116, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 97, 0, 117, 0, 114, 0, 117, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 97, 0, 108, 0, 97, 0, 117, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 83, 0, 97, 0, 109, 0, 111, 0, 97, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 72, 0, 83, 0, 84, 0, 49, 0, 48, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 105, 0, 110, 0, 103, 0, 97, 0, 112, 0, 111, 0, 114, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 98, 0, 105, 0, 100, 0, 106, 0, 97, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 108, 0, 103, 0, 105, 0, 101, 0, 114, 0, 115, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 110, 0, 97, 0, 107, 0, 114, 0, 121, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 97, 0, 109, 0, 112, 0, 97, 0, 108, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 98, 0, 97, 0, 98, 0, 97, 0, 110, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 97, 0, 105, 0, 114, 0, 111, 0, 98, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 114, 0, 105, 0, 112, 0, 111, 0, 108, 0, 105, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 101, 0, 108, 0, 105, 0, 122, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 111, 0, 103, 0, 111, 0, 116, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 110, 0, 99, 0, 117, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 121, 0, 109, 0, 97, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 117, 0, 105, 0, 97, 0, 98, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 119, 0, 115, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 101, 0, 110, 0, 118, 0, 101, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 121, 0, 97, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 118, 0, 97, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 117, 0, 118, 0, 105, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 117, 0, 110, 0, 101, 0, 97, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 97, 0, 95, 0, 80, 0, 97, 0, 122, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 99, 0, 101, 0, 105, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 110, 0, 97, 0, 117, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 114, 0, 105, 0, 100, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 97, 0, 115, 0, 115, 0, 97, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 97, 0, 110, 0, 97, 0, 109, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 101, 0, 99, 0, 105, 0, 102, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 101, 0, 103, 0, 105, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 86, 0, 105, 0, 114, 0, 103, 0, 105, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 65, 0, 115, 0, 104, 0, 107, 0, 104, 0, 97, 0, 98, 0, 97, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 104, 0, 111, 0, 110, 0, 103, 0, 113, 0, 105, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 104, 0, 117, 0, 110, 0, 103, 0, 107, 0, 105, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 70, 0, 97, 0, 109, 0, 97, 0, 103, 0, 117, 0, 115, 0, 116, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 110, 0, 103, 0, 95, 0, 75, 0, 111, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 74, 0, 101, 0, 114, 0, 117, 0, 115, 0, 97, 0, 108, 0, 101, 0, 109, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 109, 0, 99, 0, 104, 0, 97, 0, 116, 0, 107, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 97, 0, 116, 0, 104, 0, 109, 0, 97, 0, 110, 0, 100, 0, 117, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 80, 0, 111, 0, 110, 0, 116, 0, 105, 0, 97, 0, 110, 0, 97, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 80, 0, 121, 0, 111, 0, 110, 0, 103, 0, 121, 0, 97, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 81, 0, 121, 0, 122, 0, 121, 0, 108, 0, 111, 0, 114, 0, 100, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 97, 0, 109, 0, 97, 0, 114, 0, 107, 0, 97, 0, 110, 0, 100, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 86, 0, 105, 0, 101, 0, 110, 0, 116, 0, 105, 0, 97, 0, 110, 0, 101, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 70, 0, 97, 0, 114, 0, 111, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 87, 0, 101, 0, 115, 0, 116, 0, 0, 0, 67, 0, 97, 0, 110, 0, 97, 0, 100, 0, 97, 0, 47, 0, 67, 0, 101, 0, 110, 0, 116, 0, 114, 0, 97, 0, 108, 0, 0, 0, 67, 0, 97, 0, 110, 0, 97, 0, 100, 0, 97, 0, 47, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 110, 0, 0, 0, 67, 0, 97, 0, 110, 0, 97, 0, 100, 0, 97, 0, 47, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 110, 0, 100, 0, 111, 0, 114, 0, 114, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 101, 0, 108, 0, 102, 0, 97, 0, 115, 0, 116, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 78, 0, 105, 0, 99, 0, 111, 0, 115, 0, 105, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 114, 0, 97, 0, 116, 0, 111, 0, 118, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 84, 0, 97, 0, 108, 0, 108, 0, 105, 0, 110, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 97, 0, 116, 0, 105, 0, 99, 0, 97, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 105, 0, 108, 0, 110, 0, 105, 0, 117, 0, 115, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 121, 0, 111, 0, 116, 0, 116, 0, 101, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 82, 0, 101, 0, 117, 0, 110, 0, 105, 0, 111, 0, 110, 0, 0, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 47, 0, 66, 0, 97, 0, 106, 0, 97, 0, 83, 0, 117, 0, 114, 0, 0, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 47, 0, 71, 0, 101, 0, 110, 0, 101, 0, 114, 0, 97, 0, 108, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 75, 0, 111, 0, 115, 0, 114, 0, 97, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 77, 0, 97, 0, 106, 0, 117, 0, 114, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 77, 0, 105, 0, 100, 0, 119, 0, 97, 0, 121, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 111, 0, 117, 0, 109, 0, 101, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 111, 0, 110, 0, 97, 0, 112, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 83, 0, 97, 0, 105, 0, 112, 0, 97, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 97, 0, 104, 0, 105, 0, 116, 0, 105, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 97, 0, 114, 0, 97, 0, 119, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 87, 0, 97, 0, 108, 0, 108, 0, 105, 0, 115, 0, 0, 0, 85, 0, 83, 0, 47, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 45, 0, 78, 0, 101, 0, 119, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 110, 0, 111, 0, 120, 0, 95, 0, 73, 0, 78, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 97, 0, 109, 0, 97, 0, 105, 0, 99, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 67, 0, 104, 0, 97, 0, 116, 0, 104, 0, 97, 0, 109, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 67, 0, 83, 0, 84, 0, 54, 0, 67, 0, 68, 0, 84, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 69, 0, 83, 0, 84, 0, 53, 0, 69, 0, 68, 0, 84, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 77, 0, 83, 0, 84, 0, 55, 0, 77, 0, 68, 0, 84, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 80, 0, 83, 0, 84, 0, 56, 0, 80, 0, 68, 0, 84, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 108, 0, 97, 0, 110, 0, 116, 0, 121, 0, 114, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 106, 0, 105, 0, 98, 0, 111, 0, 117, 0, 116, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 108, 0, 95, 0, 65, 0, 97, 0, 105, 0, 117, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 70, 0, 114, 0, 101, 0, 101, 0, 116, 0, 111, 0, 119, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 97, 0, 98, 0, 111, 0, 114, 0, 111, 0, 110, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 104, 0, 97, 0, 114, 0, 116, 0, 111, 0, 117, 0, 109, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 105, 0, 110, 0, 115, 0, 104, 0, 97, 0, 115, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 114, 0, 111, 0, 118, 0, 105, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 100, 0, 106, 0, 97, 0, 109, 0, 101, 0, 110, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 111, 0, 95, 0, 84, 0, 111, 0, 109, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 105, 0, 109, 0, 98, 0, 117, 0, 107, 0, 116, 0, 117, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 87, 0, 105, 0, 110, 0, 100, 0, 104, 0, 111, 0, 101, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 116, 0, 105, 0, 103, 0, 117, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 114, 0, 97, 0, 99, 0, 97, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 121, 0, 101, 0, 110, 0, 110, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 104, 0, 105, 0, 99, 0, 97, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 114, 0, 100, 0, 111, 0, 98, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 114, 0, 101, 0, 115, 0, 116, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 117, 0, 114, 0, 97, 0, 99, 0, 97, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 101, 0, 116, 0, 114, 0, 111, 0, 105, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 111, 0, 100, 0, 116, 0, 104, 0, 97, 0, 98, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 114, 0, 101, 0, 110, 0, 97, 0, 100, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 97, 0, 108, 0, 105, 0, 102, 0, 97, 0, 120, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 113, 0, 97, 0, 108, 0, 117, 0, 105, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 110, 0, 97, 0, 103, 0, 117, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 114, 0, 105, 0, 103, 0, 111, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 110, 0, 100, 0, 111, 0, 122, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 99, 0, 116, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 105, 0, 112, 0, 105, 0, 103, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 111, 0, 110, 0, 104, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 79, 0, 106, 0, 105, 0, 110, 0, 97, 0, 103, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 104, 0, 111, 0, 101, 0, 110, 0, 105, 0, 120, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 111, 0, 115, 0, 97, 0, 114, 0, 105, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 105, 0, 106, 0, 117, 0, 97, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 111, 0, 114, 0, 111, 0, 110, 0, 116, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 111, 0, 114, 0, 116, 0, 111, 0, 108, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 89, 0, 97, 0, 107, 0, 117, 0, 116, 0, 97, 0, 116, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 67, 0, 104, 0, 111, 0, 105, 0, 98, 0, 97, 0, 108, 0, 115, 0, 97, 0, 110, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 80, 0, 104, 0, 110, 0, 111, 0, 109, 0, 95, 0, 80, 0, 101, 0, 110, 0, 104, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 108, 0, 97, 0, 110, 0, 95, 0, 66, 0, 97, 0, 116, 0, 111, 0, 114, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 65, 0, 122, 0, 111, 0, 114, 0, 101, 0, 115, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 67, 0, 97, 0, 110, 0, 97, 0, 114, 0, 121, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 70, 0, 97, 0, 101, 0, 114, 0, 111, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 69, 0, 117, 0, 99, 0, 108, 0, 97, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 116, 0, 104, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 80, 0, 101, 0, 114, 0, 116, 0, 104, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 83, 0, 111, 0, 117, 0, 116, 0, 104, 0, 0, 0, 67, 0, 97, 0, 110, 0, 97, 0, 100, 0, 97, 0, 47, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 0, 0, 67, 0, 97, 0, 110, 0, 97, 0, 100, 0, 97, 0, 47, 0, 77, 0, 111, 0, 117, 0, 110, 0, 116, 0, 97, 0, 105, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 101, 0, 108, 0, 103, 0, 114, 0, 97, 0, 100, 0, 101, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 114, 0, 117, 0, 115, 0, 115, 0, 101, 0, 108, 0, 115, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 117, 0, 100, 0, 97, 0, 112, 0, 101, 0, 115, 0, 116, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 117, 0, 115, 0, 105, 0, 110, 0, 103, 0, 101, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 67, 0, 104, 0, 105, 0, 115, 0, 105, 0, 110, 0, 97, 0, 117, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 71, 0, 117, 0, 101, 0, 114, 0, 110, 0, 115, 0, 101, 0, 121, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 72, 0, 101, 0, 108, 0, 115, 0, 105, 0, 110, 0, 107, 0, 105, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 73, 0, 115, 0, 116, 0, 97, 0, 110, 0, 98, 0, 117, 0, 108, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 114, 0, 97, 0, 106, 0, 101, 0, 118, 0, 111, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 84, 0, 105, 0, 114, 0, 97, 0, 115, 0, 112, 0, 111, 0, 108, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 85, 0, 122, 0, 104, 0, 103, 0, 111, 0, 114, 0, 111, 0, 100, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 108, 0, 100, 0, 105, 0, 118, 0, 101, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 70, 0, 97, 0, 107, 0, 97, 0, 111, 0, 102, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 97, 0, 109, 0, 98, 0, 105, 0, 101, 0, 114, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 78, 0, 111, 0, 114, 0, 102, 0, 111, 0, 108, 0, 107, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 111, 0, 104, 0, 110, 0, 112, 0, 101, 0, 105, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 65, 0, 83, 0, 84, 0, 52, 0, 65, 0, 68, 0, 84, 0, 0, 0, 83, 0, 121, 0, 115, 0, 116, 0, 101, 0, 109, 0, 86, 0, 47, 0, 89, 0, 83, 0, 84, 0, 57, 0, 89, 0, 68, 0, 84, 0, 0, 0, 85, 0, 83, 0, 47, 0, 69, 0, 97, 0, 115, 0, 116, 0, 45, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 114, 0, 111, 0, 108, 0, 108, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 117, 0, 106, 0, 117, 0, 109, 0, 98, 0, 117, 0, 114, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 103, 0, 97, 0, 100, 0, 105, 0, 115, 0, 104, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 103, 0, 117, 0, 105, 0, 108, 0, 108, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 115, 0, 117, 0, 110, 0, 99, 0, 105, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 116, 0, 105, 0, 107, 0, 111, 0, 107, 0, 97, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 114, 0, 98, 0, 97, 0, 100, 0, 111, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 111, 0, 109, 0, 105, 0, 110, 0, 105, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 100, 0, 109, 0, 111, 0, 110, 0, 116, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 105, 0, 114, 0, 117, 0, 110, 0, 101, 0, 112, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 110, 0, 115, 0, 101, 0, 110, 0, 97, 0, 100, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 122, 0, 97, 0, 116, 0, 108, 0, 97, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 105, 0, 113, 0, 117, 0, 101, 0, 108, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 114, 0, 101, 0, 97, 0, 108, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 101, 0, 119, 0, 95, 0, 89, 0, 111, 0, 114, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 101, 0, 115, 0, 111, 0, 108, 0, 117, 0, 116, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 97, 0, 114, 0, 101, 0, 109, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 105, 0, 97, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 104, 0, 105, 0, 112, 0, 114, 0, 111, 0, 99, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 74, 0, 111, 0, 104, 0, 110, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 75, 0, 105, 0, 116, 0, 116, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 76, 0, 117, 0, 99, 0, 105, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 87, 0, 105, 0, 110, 0, 110, 0, 105, 0, 112, 0, 101, 0, 103, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 115, 0, 101, 0, 121, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 118, 0, 105, 0, 115, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 121, 0, 111, 0, 119, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 95, 0, 67, 0, 104, 0, 105, 0, 95, 0, 77, 0, 105, 0, 110, 0, 104, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 114, 0, 97, 0, 115, 0, 110, 0, 111, 0, 121, 0, 97, 0, 114, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 78, 0, 111, 0, 118, 0, 111, 0, 115, 0, 105, 0, 98, 0, 105, 0, 114, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 108, 0, 97, 0, 97, 0, 110, 0, 98, 0, 97, 0, 97, 0, 116, 0, 97, 0, 114, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 86, 0, 108, 0, 97, 0, 100, 0, 105, 0, 118, 0, 111, 0, 115, 0, 116, 0, 111, 0, 107, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 66, 0, 101, 0, 114, 0, 109, 0, 117, 0, 100, 0, 97, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 77, 0, 97, 0, 100, 0, 101, 0, 105, 0, 114, 0, 97, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 83, 0, 116, 0, 97, 0, 110, 0, 108, 0, 101, 0, 121, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 67, 0, 117, 0, 114, 0, 114, 0, 105, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 68, 0, 97, 0, 114, 0, 119, 0, 105, 0, 110, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 72, 0, 111, 0, 98, 0, 97, 0, 114, 0, 116, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 83, 0, 121, 0, 100, 0, 110, 0, 101, 0, 121, 0, 0, 0, 66, 0, 114, 0, 97, 0, 122, 0, 105, 0, 108, 0, 47, 0, 68, 0, 101, 0, 78, 0, 111, 0, 114, 0, 111, 0, 110, 0, 104, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 109, 0, 115, 0, 116, 0, 101, 0, 114, 0, 100, 0, 97, 0, 109, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 65, 0, 115, 0, 116, 0, 114, 0, 97, 0, 107, 0, 104, 0, 97, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 117, 0, 99, 0, 104, 0, 97, 0, 114, 0, 101, 0, 115, 0, 116, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 71, 0, 105, 0, 98, 0, 114, 0, 97, 0, 108, 0, 116, 0, 97, 0, 114, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 106, 0, 117, 0, 98, 0, 108, 0, 106, 0, 97, 0, 110, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 77, 0, 97, 0, 114, 0, 105, 0, 101, 0, 104, 0, 97, 0, 109, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 80, 0, 111, 0, 100, 0, 103, 0, 111, 0, 114, 0, 105, 0, 99, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 116, 0, 111, 0, 99, 0, 107, 0, 104, 0, 111, 0, 108, 0, 109, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 85, 0, 108, 0, 121, 0, 97, 0, 110, 0, 111, 0, 118, 0, 115, 0, 107, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 86, 0, 111, 0, 108, 0, 103, 0, 111, 0, 103, 0, 114, 0, 97, 0, 100, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 67, 0, 104, 0, 114, 0, 105, 0, 115, 0, 116, 0, 109, 0, 97, 0, 115, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 75, 0, 101, 0, 114, 0, 103, 0, 117, 0, 101, 0, 108, 0, 101, 0, 110, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 77, 0, 97, 0, 117, 0, 114, 0, 105, 0, 116, 0, 105, 0, 117, 0, 115, 0, 0, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 47, 0, 66, 0, 97, 0, 106, 0, 97, 0, 78, 0, 111, 0, 114, 0, 116, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 65, 0, 117, 0, 99, 0, 107, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 70, 0, 117, 0, 110, 0, 97, 0, 102, 0, 117, 0, 116, 0, 105, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 72, 0, 111, 0, 110, 0, 111, 0, 108, 0, 117, 0, 108, 0, 117, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 74, 0, 111, 0, 104, 0, 110, 0, 115, 0, 116, 0, 111, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 105, 0, 116, 0, 99, 0, 97, 0, 105, 0, 114, 0, 110, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 75, 0, 119, 0, 97, 0, 106, 0, 97, 0, 108, 0, 101, 0, 105, 0, 110, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 115, 0, 97, 0, 98, 0, 108, 0, 97, 0, 110, 0, 99, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 105, 0, 98, 0, 114, 0, 101, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 117, 0, 98, 0, 117, 0, 109, 0, 98, 0, 97, 0, 115, 0, 104, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 117, 0, 97, 0, 107, 0, 99, 0, 104, 0, 111, 0, 116, 0, 116, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 111, 0, 45, 0, 78, 0, 111, 0, 118, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 110, 0, 99, 0, 104, 0, 111, 0, 114, 0, 97, 0, 103, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 97, 0, 103, 0, 117, 0, 97, 0, 105, 0, 110, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 111, 0, 97, 0, 95, 0, 86, 0, 105, 0, 115, 0, 116, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 116, 0, 97, 0, 109, 0, 97, 0, 114, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 104, 0, 105, 0, 104, 0, 117, 0, 97, 0, 104, 0, 117, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 70, 0, 111, 0, 114, 0, 116, 0, 97, 0, 108, 0, 101, 0, 122, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 108, 0, 97, 0, 99, 0, 101, 0, 95, 0, 66, 0, 97, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 111, 0, 111, 0, 115, 0, 101, 0, 95, 0, 66, 0, 97, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 97, 0, 116, 0, 101, 0, 109, 0, 97, 0, 108, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 97, 0, 121, 0, 97, 0, 113, 0, 117, 0, 105, 0, 108, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 116, 0, 97, 0, 109, 0, 111, 0, 114, 0, 111, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 110, 0, 111, 0, 109, 0, 105, 0, 110, 0, 101, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 101, 0, 114, 0, 114, 0, 101, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 111, 0, 95, 0, 80, 0, 97, 0, 117, 0, 108, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 84, 0, 104, 0, 111, 0, 109, 0, 97, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 86, 0, 97, 0, 110, 0, 99, 0, 111, 0, 117, 0, 118, 0, 101, 0, 114, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 119, 0, 115, 0, 111, 0, 110, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 97, 0, 108, 0, 109, 0, 101, 0, 114, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 86, 0, 111, 0, 115, 0, 116, 0, 111, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 75, 0, 117, 0, 97, 0, 108, 0, 97, 0, 95, 0, 76, 0, 117, 0, 109, 0, 112, 0, 117, 0, 114, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 78, 0, 111, 0, 118, 0, 111, 0, 107, 0, 117, 0, 122, 0, 110, 0, 101, 0, 116, 0, 115, 0, 107, 0, 0, 0, 67, 0, 104, 0, 105, 0, 108, 0, 101, 0, 47, 0, 67, 0, 111, 0, 110, 0, 116, 0, 105, 0, 110, 0, 101, 0, 110, 0, 116, 0, 97, 0, 108, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 66, 0, 114, 0, 97, 0, 116, 0, 105, 0, 115, 0, 108, 0, 97, 0, 118, 0, 97, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 67, 0, 111, 0, 112, 0, 101, 0, 110, 0, 104, 0, 97, 0, 103, 0, 101, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 76, 0, 117, 0, 120, 0, 101, 0, 109, 0, 98, 0, 111, 0, 117, 0, 114, 0, 103, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 97, 0, 110, 0, 95, 0, 77, 0, 97, 0, 114, 0, 105, 0, 110, 0, 111, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 83, 0, 105, 0, 109, 0, 102, 0, 101, 0, 114, 0, 111, 0, 112, 0, 111, 0, 108, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 90, 0, 97, 0, 112, 0, 111, 0, 114, 0, 111, 0, 122, 0, 104, 0, 121, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 69, 0, 110, 0, 100, 0, 101, 0, 114, 0, 98, 0, 117, 0, 114, 0, 121, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 97, 0, 108, 0, 97, 0, 112, 0, 97, 0, 103, 0, 111, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 77, 0, 97, 0, 114, 0, 113, 0, 117, 0, 101, 0, 115, 0, 97, 0, 115, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 97, 0, 103, 0, 111, 0, 95, 0, 80, 0, 97, 0, 103, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 82, 0, 97, 0, 114, 0, 111, 0, 116, 0, 111, 0, 110, 0, 103, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 84, 0, 111, 0, 110, 0, 103, 0, 97, 0, 116, 0, 97, 0, 112, 0, 117, 0, 0, 0, 85, 0, 83, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 45, 0, 83, 0, 116, 0, 97, 0, 114, 0, 107, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 100, 0, 100, 0, 105, 0, 115, 0, 95, 0, 65, 0, 98, 0, 97, 0, 98, 0, 97, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 114, 0, 97, 0, 122, 0, 122, 0, 97, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 79, 0, 117, 0, 97, 0, 103, 0, 97, 0, 100, 0, 111, 0, 117, 0, 103, 0, 111, 0, 117, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 115, 0, 116, 0, 97, 0, 95, 0, 82, 0, 105, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 70, 0, 111, 0, 114, 0, 116, 0, 95, 0, 87, 0, 97, 0, 121, 0, 110, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 114, 0, 97, 0, 110, 0, 100, 0, 95, 0, 84, 0, 117, 0, 114, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 71, 0, 117, 0, 97, 0, 100, 0, 101, 0, 108, 0, 111, 0, 117, 0, 112, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 72, 0, 101, 0, 114, 0, 109, 0, 111, 0, 115, 0, 105, 0, 108, 0, 108, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 114, 0, 97, 0, 108, 0, 101, 0, 110, 0, 100, 0, 105, 0, 106, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 117, 0, 105, 0, 115, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 114, 0, 116, 0, 105, 0, 110, 0, 105, 0, 113, 0, 117, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 116, 0, 108, 0, 97, 0, 107, 0, 97, 0, 116, 0, 108, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 101, 0, 118, 0, 105, 0, 100, 0, 101, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 115, 0, 101, 0, 114, 0, 114, 0, 97, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 97, 0, 114, 0, 97, 0, 109, 0, 97, 0, 114, 0, 105, 0, 98, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 111, 0, 95, 0, 65, 0, 99, 0, 114, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 105, 0, 111, 0, 95, 0, 66, 0, 114, 0, 97, 0, 110, 0, 99, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 86, 0, 105, 0, 110, 0, 99, 0, 101, 0, 110, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 87, 0, 104, 0, 105, 0, 116, 0, 101, 0, 104, 0, 111, 0, 114, 0, 115, 0, 101, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 99, 0, 77, 0, 117, 0, 114, 0, 100, 0, 111, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 111, 0, 116, 0, 104, 0, 101, 0, 114, 0, 97, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 83, 0, 114, 0, 101, 0, 100, 0, 110, 0, 101, 0, 107, 0, 111, 0, 108, 0, 121, 0, 109, 0, 115, 0, 107, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 85, 0, 106, 0, 117, 0, 110, 0, 103, 0, 95, 0, 80, 0, 97, 0, 110, 0, 100, 0, 97, 0, 110, 0, 103, 0, 0, 0, 65, 0, 115, 0, 105, 0, 97, 0, 47, 0, 89, 0, 101, 0, 107, 0, 97, 0, 116, 0, 101, 0, 114, 0, 105, 0, 110, 0, 98, 0, 117, 0, 114, 0, 103, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 74, 0, 97, 0, 110, 0, 95, 0, 77, 0, 97, 0, 121, 0, 101, 0, 110, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 82, 0, 101, 0, 121, 0, 107, 0, 106, 0, 97, 0, 118, 0, 105, 0, 107, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 83, 0, 116, 0, 95, 0, 72, 0, 101, 0, 108, 0, 101, 0, 110, 0, 97, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 65, 0, 100, 0, 101, 0, 108, 0, 97, 0, 105, 0, 100, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 66, 0, 114, 0, 105, 0, 115, 0, 98, 0, 97, 0, 110, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 67, 0, 97, 0, 110, 0, 98, 0, 101, 0, 114, 0, 114, 0, 97, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 76, 0, 105, 0, 110, 0, 100, 0, 101, 0, 109, 0, 97, 0, 110, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 84, 0, 97, 0, 115, 0, 109, 0, 97, 0, 110, 0, 105, 0, 97, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 86, 0, 105, 0, 99, 0, 116, 0, 111, 0, 114, 0, 105, 0, 97, 0, 0, 0, 67, 0, 104, 0, 105, 0, 108, 0, 101, 0, 47, 0, 69, 0, 97, 0, 115, 0, 116, 0, 101, 0, 114, 0, 73, 0, 115, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 73, 0, 115, 0, 108, 0, 101, 0, 95, 0, 111, 0, 102, 0, 95, 0, 77, 0, 97, 0, 110, 0, 0, 0, 69, 0, 117, 0, 114, 0, 111, 0, 112, 0, 101, 0, 47, 0, 75, 0, 97, 0, 108, 0, 105, 0, 110, 0, 105, 0, 110, 0, 103, 0, 114, 0, 97, 0, 100, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 75, 0, 105, 0, 114, 0, 105, 0, 116, 0, 105, 0, 109, 0, 97, 0, 116, 0, 105, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 74, 0, 111, 0, 104, 0, 97, 0, 110, 0, 110, 0, 101, 0, 115, 0, 98, 0, 117, 0, 114, 0, 103, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 69, 0, 108, 0, 95, 0, 83, 0, 97, 0, 108, 0, 118, 0, 97, 0, 100, 0, 111, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 70, 0, 111, 0, 114, 0, 116, 0, 95, 0, 78, 0, 101, 0, 108, 0, 115, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 115, 0, 95, 0, 65, 0, 110, 0, 103, 0, 101, 0, 108, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 101, 0, 120, 0, 105, 0, 99, 0, 111, 0, 95, 0, 67, 0, 105, 0, 116, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 97, 0, 110, 0, 103, 0, 110, 0, 105, 0, 114, 0, 116, 0, 117, 0, 110, 0, 103, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 111, 0, 95, 0, 86, 0, 101, 0, 108, 0, 104, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 117, 0, 101, 0, 114, 0, 116, 0, 111, 0, 95, 0, 82, 0, 105, 0, 99, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 97, 0, 105, 0, 110, 0, 121, 0, 95, 0, 82, 0, 105, 0, 118, 0, 101, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 101, 0, 103, 0, 117, 0, 99, 0, 105, 0, 103, 0, 97, 0, 108, 0, 112, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 84, 0, 104, 0, 117, 0, 110, 0, 100, 0, 101, 0, 114, 0, 95, 0, 66, 0, 97, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 89, 0, 101, 0, 108, 0, 108, 0, 111, 0, 119, 0, 107, 0, 110, 0, 105, 0, 102, 0, 101, 0, 0, 0, 65, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 47, 0, 76, 0, 111, 0, 110, 0, 103, 0, 121, 0, 101, 0, 97, 0, 114, 0, 98, 0, 121, 0, 101, 0, 110, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 67, 0, 97, 0, 112, 0, 101, 0, 95, 0, 86, 0, 101, 0, 114, 0, 100, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 76, 0, 111, 0, 114, 0, 100, 0, 95, 0, 72, 0, 111, 0, 119, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 77, 0, 101, 0, 108, 0, 98, 0, 111, 0, 117, 0, 114, 0, 110, 0, 101, 0, 0, 0, 67, 0, 97, 0, 110, 0, 97, 0, 100, 0, 97, 0, 47, 0, 78, 0, 101, 0, 119, 0, 102, 0, 111, 0, 117, 0, 110, 0, 100, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 67, 0, 97, 0, 110, 0, 97, 0, 100, 0, 97, 0, 47, 0, 83, 0, 97, 0, 115, 0, 107, 0, 97, 0, 116, 0, 99, 0, 104, 0, 101, 0, 119, 0, 97, 0, 110, 0, 0, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 47, 0, 65, 0, 110, 0, 116, 0, 97, 0, 110, 0, 97, 0, 110, 0, 97, 0, 114, 0, 105, 0, 118, 0, 111, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 71, 0, 117, 0, 97, 0, 100, 0, 97, 0, 108, 0, 99, 0, 97, 0, 110, 0, 97, 0, 108, 0, 0, 0, 65, 0, 102, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 114, 0, 95, 0, 101, 0, 115, 0, 95, 0, 83, 0, 97, 0, 108, 0, 97, 0, 97, 0, 109, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 108, 0, 97, 0, 110, 0, 99, 0, 45, 0, 83, 0, 97, 0, 98, 0, 108, 0, 111, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 117, 0, 101, 0, 110, 0, 111, 0, 115, 0, 95, 0, 65, 0, 105, 0, 114, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 109, 0, 112, 0, 111, 0, 95, 0, 71, 0, 114, 0, 97, 0, 110, 0, 100, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 110, 0, 109, 0, 97, 0, 114, 0, 107, 0, 115, 0, 104, 0, 97, 0, 118, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 97, 0, 119, 0, 115, 0, 111, 0, 110, 0, 95, 0, 67, 0, 114, 0, 101, 0, 101, 0, 107, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 75, 0, 110, 0, 111, 0, 120, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 112, 0, 111, 0, 108, 0, 105, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 117, 0, 110, 0, 116, 0, 97, 0, 95, 0, 65, 0, 114, 0, 101, 0, 110, 0, 97, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 82, 0, 97, 0, 110, 0, 107, 0, 105, 0, 110, 0, 95, 0, 73, 0, 110, 0, 108, 0, 101, 0, 116, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 97, 0, 95, 0, 73, 0, 115, 0, 97, 0, 98, 0, 101, 0, 108, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 99, 0, 111, 0, 114, 0, 101, 0, 115, 0, 98, 0, 121, 0, 115, 0, 117, 0, 110, 0, 100, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 77, 0, 97, 0, 99, 0, 113, 0, 117, 0, 97, 0, 114, 0, 105, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 81, 0, 117, 0, 101, 0, 101, 0, 110, 0, 115, 0, 108, 0, 97, 0, 110, 0, 100, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 89, 0, 97, 0, 110, 0, 99, 0, 111, 0, 119, 0, 105, 0, 110, 0, 110, 0, 97, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 66, 0, 111, 0, 117, 0, 103, 0, 97, 0, 105, 0, 110, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 80, 0, 97, 0, 99, 0, 105, 0, 102, 0, 105, 0, 99, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 95, 0, 77, 0, 111, 0, 114, 0, 101, 0, 115, 0, 98, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 97, 0, 109, 0, 98, 0, 114, 0, 105, 0, 100, 0, 103, 0, 101, 0, 95, 0, 66, 0, 97, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 67, 0, 111, 0, 114, 0, 97, 0, 108, 0, 95, 0, 72, 0, 97, 0, 114, 0, 98, 0, 111, 0, 117, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 86, 0, 101, 0, 118, 0, 97, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 76, 0, 111, 0, 119, 0, 101, 0, 114, 0, 95, 0, 80, 0, 114, 0, 105, 0, 110, 0, 99, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 95, 0, 111, 0, 102, 0, 95, 0, 83, 0, 112, 0, 97, 0, 105, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 116, 0, 111, 0, 95, 0, 68, 0, 111, 0, 109, 0, 105, 0, 110, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 116, 0, 95, 0, 66, 0, 97, 0, 114, 0, 116, 0, 104, 0, 101, 0, 108, 0, 101, 0, 109, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 119, 0, 105, 0, 102, 0, 116, 0, 95, 0, 67, 0, 117, 0, 114, 0, 114, 0, 101, 0, 110, 0, 116, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 83, 0, 111, 0, 117, 0, 116, 0, 104, 0, 95, 0, 80, 0, 111, 0, 108, 0, 101, 0, 0, 0, 65, 0, 117, 0, 115, 0, 116, 0, 114, 0, 97, 0, 108, 0, 105, 0, 97, 0, 47, 0, 66, 0, 114, 0, 111, 0, 107, 0, 101, 0, 110, 0, 95, 0, 72, 0, 105, 0, 108, 0, 108, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 66, 0, 97, 0, 104, 0, 105, 0, 97, 0, 95, 0, 66, 0, 97, 0, 110, 0, 100, 0, 101, 0, 114, 0, 97, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 80, 0, 111, 0, 114, 0, 116, 0, 45, 0, 97, 0, 117, 0, 45, 0, 80, 0, 114, 0, 105, 0, 110, 0, 99, 0, 101, 0, 0, 0, 65, 0, 116, 0, 108, 0, 97, 0, 110, 0, 116, 0, 105, 0, 99, 0, 47, 0, 83, 0, 111, 0, 117, 0, 116, 0, 104, 0, 95, 0, 71, 0, 101, 0, 111, 0, 114, 0, 103, 0, 105, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 74, 0, 117, 0, 106, 0, 117, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 83, 0, 97, 0, 108, 0, 116, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 77, 0, 97, 0, 114, 0, 101, 0, 110, 0, 103, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 87, 0, 105, 0, 110, 0, 97, 0, 109, 0, 97, 0, 99, 0, 0, 0, 67, 0, 97, 0, 110, 0, 97, 0, 100, 0, 97, 0, 47, 0, 69, 0, 97, 0, 115, 0, 116, 0, 45, 0, 83, 0, 97, 0, 115, 0, 107, 0, 97, 0, 116, 0, 99, 0, 104, 0, 101, 0, 119, 0, 97, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 67, 0, 111, 0, 114, 0, 100, 0, 111, 0, 98, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 77, 0, 101, 0, 110, 0, 100, 0, 111, 0, 122, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 84, 0, 117, 0, 99, 0, 117, 0, 109, 0, 97, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 85, 0, 115, 0, 104, 0, 117, 0, 97, 0, 105, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 84, 0, 101, 0, 108, 0, 108, 0, 95, 0, 67, 0, 105, 0, 116, 0, 121, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 86, 0, 105, 0, 110, 0, 99, 0, 101, 0, 110, 0, 110, 0, 101, 0, 115, 0, 0, 0, 65, 0, 110, 0, 116, 0, 97, 0, 114, 0, 99, 0, 116, 0, 105, 0, 99, 0, 97, 0, 47, 0, 68, 0, 117, 0, 109, 0, 111, 0, 110, 0, 116, 0, 68, 0, 85, 0, 114, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 76, 0, 97, 0, 95, 0, 82, 0, 105, 0, 111, 0, 106, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 95, 0, 74, 0, 117, 0, 97, 0, 110, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 83, 0, 97, 0, 110, 0, 95, 0, 76, 0, 117, 0, 105, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 80, 0, 101, 0, 116, 0, 101, 0, 114, 0, 115, 0, 98, 0, 117, 0, 114, 0, 103, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 67, 0, 97, 0, 116, 0, 97, 0, 109, 0, 97, 0, 114, 0, 99, 0, 97, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 101, 0, 110, 0, 116, 0, 117, 0, 99, 0, 107, 0, 121, 0, 47, 0, 76, 0, 111, 0, 117, 0, 105, 0, 115, 0, 118, 0, 105, 0, 108, 0, 108, 0, 101, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 75, 0, 101, 0, 110, 0, 116, 0, 117, 0, 99, 0, 107, 0, 121, 0, 47, 0, 77, 0, 111, 0, 110, 0, 116, 0, 105, 0, 99, 0, 101, 0, 108, 0, 108, 0, 111, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 116, 0, 104, 0, 95, 0, 68, 0, 97, 0, 107, 0, 111, 0, 116, 0, 97, 0, 47, 0, 66, 0, 101, 0, 117, 0, 108, 0, 97, 0, 104, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 116, 0, 104, 0, 95, 0, 68, 0, 97, 0, 107, 0, 111, 0, 116, 0, 97, 0, 47, 0, 67, 0, 101, 0, 110, 0, 116, 0, 101, 0, 114, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 47, 0, 73, 0, 110, 0, 100, 0, 105, 0, 97, 0, 110, 0, 97, 0, 112, 0, 111, 0, 108, 0, 105, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 66, 0, 117, 0, 101, 0, 110, 0, 111, 0, 115, 0, 95, 0, 65, 0, 105, 0, 114, 0, 101, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 82, 0, 105, 0, 111, 0, 95, 0, 71, 0, 97, 0, 108, 0, 108, 0, 101, 0, 103, 0, 111, 0, 115, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 78, 0, 111, 0, 114, 0, 116, 0, 104, 0, 95, 0, 68, 0, 97, 0, 107, 0, 111, 0, 116, 0, 97, 0, 47, 0, 78, 0, 101, 0, 119, 0, 95, 0, 83, 0, 97, 0, 108, 0, 101, 0, 109, 0, 0, 0, 65, 0, 109, 0, 101, 0, 114, 0, 105, 0, 99, 0, 97, 0, 47, 0, 65, 0, 114, 0, 103, 0, 101, 0, 110, 0, 116, 0, 105, 0, 110, 0, 97, 0, 47, 0, 67, 0, 111, 0, 109, 0, 111, 0, 100, 0, 82, 0, 105, 0, 118, 0, 97, 0, 100, 0, 97, 0, 118, 0, 105, 0, 97, 0, 0, 0, 122, 2, 218, 10, 240, 2, 220, 2, 244, 2, 248, 2, 95, 14, 54, 8, 184, 29, 110, 14, 222, 10, 236, 10, 250, 10, 8, 11, 22, 11, 36, 11, 188, 18, 203, 29, 13, 23, 67, 8, 232, 26, 80, 8, 125, 14, 93, 8, 7, 34, 204, 18, 50, 11, 220, 18, 236, 18, 252, 18, 64, 11, 119, 32, 157, 6, 140, 14, 12, 19, 78, 11, 28, 19, 106, 8, 250, 26, 169, 6, 92, 11, 12, 27, 106, 11, 120, 11, 134, 11, 148, 11, 155, 14, 30, 23, 44, 19, 170, 14, 60, 19, 162, 11, 30, 27, 222, 29, 48, 27, 76, 19, 92, 19, 185, 14, 119, 8, 108, 19, 132, 8, 66, 27, 47, 23, 124, 19, 84, 27, 209, 38, 40, 38, 46, 39, 6, 37, 141, 36, 188, 37, 32, 37, 240, 38, 165, 36, 215, 37, 242, 37, 58, 37, 84, 37, 176, 11, 64, 23, 81, 23, 145, 8, 190, 11, 72, 36, 98, 23, 204, 11, 200, 14, 28, 34, 102, 27, 215, 14, 218, 11, 49, 34, 108, 35, 70, 34, 230, 14, 140, 19, 120, 27, 156, 19, 245, 14, 172, 19, 138, 27, 130, 35, 188, 19, 241, 29, 204, 19, 4, 15, 220, 19, 91, 34, 19, 15, 112, 34, 34, 15, 236, 19, 115, 23, 132, 23, 149, 23, 139, 32, 166, 23, 159, 32, 4, 30, 156, 27, 174, 27, 252, 19, 192, 27, 23, 30, 12, 20, 42, 30, 210, 27, 228, 27, 49, 15, 28, 20, 64, 15, 61, 30, 180, 38, 133, 34, 189, 36, 13, 38, 110, 37, 152, 35, 136, 37, 213, 36, 154, 34, 79, 15, 44, 20, 92, 18, 232, 11, 94, 15, 68, 38, 96, 38, 76, 18, 80, 30, 109, 15, 158, 8, 179, 32, 99, 30, 174, 35, 124, 15, 60, 20, 139, 15, 76, 20, 118, 30, 246, 27, 183, 23, 92, 20, 8, 28, 154, 15, 137, 30, 199, 32, 200, 23, 108, 20, 26, 28, 156, 30, 217, 23, 175, 30, 169, 15, 234, 23, 124, 20, 171, 8, 140, 20, 124, 38, 152, 38, 15, 39, 156, 20, 184, 15, 219, 32, 194, 30, 172, 20, 95, 36, 196, 35, 213, 30, 239, 32, 3, 33, 175, 34, 23, 33, 196, 34, 199, 15, 214, 15, 251, 23, 232, 30, 188, 20, 217, 34, 12, 24, 29, 24, 218, 35, 44, 28, 238, 34, 46, 24, 246, 11, 240, 35, 63, 24, 80, 24, 97, 24, 62, 28, 251, 30, 6, 36, 43, 33, 194, 10, 63, 33, 204, 20, 220, 20, 236, 20, 80, 28, 229, 15, 14, 31, 114, 24, 252, 20, 83, 33, 131, 24, 148, 24, 162, 37, 3, 35, 98, 28, 33, 31, 116, 28, 52, 31, 28, 36, 165, 24, 252, 22, 134, 28, 103, 33, 124, 4, 181, 6, 138, 5, 193, 6, 149, 5, 205, 6, 4, 12, 244, 15, 217, 6, 184, 8, 197, 8, 134, 4, 210, 8, 223, 8, 229, 6, 236, 8, 241, 6, 18, 12, 160, 5, 12, 21, 3, 16, 18, 16, 249, 8, 171, 5, 32, 12, 182, 5, 144, 4, 193, 5, 46, 12, 33, 16, 154, 4, 253, 6, 9, 7, 182, 24, 48, 16, 164, 4, 6, 9, 60, 12, 19, 9, 74, 12, 63, 16, 204, 5, 78, 16, 32, 9, 45, 9, 93, 16, 88, 12, 102, 12, 58, 9, 199, 24, 152, 28, 71, 9, 21, 7, 215, 5, 226, 5, 84, 9, 116, 12, 33, 7, 45, 7, 97, 9, 170, 28, 216, 24, 174, 4, 184, 4, 28, 21, 108, 16, 123, 16, 237, 5, 130, 12, 138, 16, 110, 9, 57, 7, 69, 7, 144, 12, 153, 16, 248, 5, 158, 12, 80, 14, 71, 31, 81, 7, 172, 12, 123, 9, 93, 7, 186, 12, 105, 7, 136, 9, 3, 6, 14, 6, 90, 31, 233, 24, 44, 21, 117, 7, 200, 12, 168, 16, 250, 24, 149, 9, 129, 7, 109, 31, 162, 9, 60, 21, 11, 25, 76, 21, 123, 33, 92, 21, 183, 16, 128, 31, 28, 25, 147, 31, 118, 36, 166, 31, 45, 25, 208, 10, 185, 31, 204, 31, 50, 36, 223, 31, 62, 25, 79, 25, 108, 21, 96, 25, 214, 12, 242, 31, 143, 33, 163, 33, 228, 12, 124, 21, 140, 21, 24, 35, 156, 21, 113, 25, 5, 32, 24, 32, 198, 16, 45, 35, 252, 2, 0, 3, 141, 7, 130, 25, 153, 7, 165, 7, 4, 3, 8, 3, 12, 3, 16, 3, 132, 18, 216, 2, 172, 21, 213, 16, 237, 36, 228, 16, 188, 21, 183, 33, 243, 16, 203, 33, 175, 9, 188, 28, 43, 32, 96, 3, 20, 3, 24, 3, 28, 3, 32, 3, 148, 18, 139, 3, 4, 4, 241, 3, 104, 4, 194, 4, 25, 6, 36, 6, 47, 6, 204, 4, 214, 4, 224, 4, 234, 4, 244, 4, 254, 4, 8, 5, 18, 5, 114, 4, 28, 5, 58, 6, 69, 6, 80, 6, 91, 6, 102, 6, 38, 5, 48, 5, 58, 5, 68, 5, 78, 5, 88, 5, 98, 5, 108, 5, 49, 4, 166, 10, 9, 4, 249, 3, 180, 10, 177, 7, 58, 4, 147, 25, 2, 17, 164, 25, 242, 12, 17, 17, 204, 21, 0, 13, 206, 28, 220, 21, 181, 25, 236, 21, 252, 21, 12, 22, 224, 28, 14, 13, 198, 25, 28, 22, 44, 22, 62, 32, 60, 22, 28, 13, 81, 32, 189, 7, 188, 9, 42, 13, 215, 25, 56, 13, 242, 28, 70, 13, 201, 9, 232, 25, 214, 9, 84, 13, 98, 13, 32, 17, 201, 7, 227, 9, 249, 25, 112, 13, 213, 7, 225, 7, 126, 13, 4, 29, 76, 22, 47, 17, 22, 29, 140, 13, 240, 9, 10, 26, 62, 17, 154, 13, 92, 22, 27, 26, 108, 22, 253, 9, 77, 17, 168, 13, 92, 17, 44, 26, 182, 13, 196, 13, 40, 29, 210, 13, 17, 4, 46, 0, 1, 4, 245, 3, 108, 4, 118, 4, 53, 4, 170, 10, 36, 3, 67, 4, 40, 3, 44, 3, 25, 4, 223, 33, 224, 13, 61, 26, 10, 10, 238, 13, 78, 26, 237, 7, 124, 22, 95, 26, 107, 17, 122, 17, 101, 3, 190, 3, 48, 3, 100, 18, 151, 3, 222, 26, 157, 3, 52, 3, 224, 2, 56, 3, 164, 18, 112, 26, 137, 17, 152, 17, 60, 3, 64, 3, 31, 0, 233, 3, 204, 3, 228, 2, 68, 3, 72, 3, 76, 3, 80, 3, 180, 18, 23, 10, 129, 26, 66, 35, 108, 18, 252, 13, 167, 17, 10, 14, 58, 29, 140, 22, 41, 8, 146, 26, 76, 29, 156, 22, 243, 33, 36, 10, 163, 26, 180, 26, 100, 32, 182, 17, 214, 26, 197, 17, 94, 29, 212, 17, 24, 14, 49, 10, 172, 22, 227, 17, 112, 29, 38, 14, 197, 26, 188, 22, 242, 17, 87, 35, 130, 29, 1, 18, 52, 14, 16, 18, 31, 18, 148, 29, 62, 10, 75, 10, 46, 18, 249, 7, 211, 3, 76, 4, 84, 3, 88, 3, 92, 3, 85, 14, 88, 10, 204, 22, 101, 10, 124, 18, 114, 10, 140, 18, 66, 14, 127, 10, 156, 18, 140, 10, 172, 18, 153, 10, 220, 22, 218, 3, 13, 4, 118, 5, 5, 8, 113, 6, 124, 6, 236, 22, 135, 6, 128, 5, 166, 29, 17, 8, 29, 8, 146, 6, 61, 18, 85, 4, 253, 3, 184, 10, 232, 2, 111, 3, 236, 2, 62, 4, 122, 2, 13, 0, 13, 0, 19, 0, 70, 0, 1, 0, 115, 0, 120, 1, 237, 2, 87, 1, 127, 0, 127, 0, 249, 1, 42, 1, 126, 1, 147, 1, 14, 2, 45, 1, 6, 1, 70, 0, 231, 1, 73, 0, 129, 1, 113, 2, 167, 2, 75, 1, 54, 1, 93, 1, 107, 2, 30, 1, 209, 2, 203, 2, 122, 2, 170, 2, 89, 2, 80, 2, 112, 0, 29, 2, 105, 1, 143, 2, 229, 0, 112, 0, 206, 2, 135, 1, 172, 0, 222, 1, 134, 2, 116, 2, 219, 1, 151, 0, 137, 2, 23, 2, 2, 2, 253, 0, 9, 1, 233, 2, 115, 0, 157, 0, 158, 2, 17, 2, 1, 0, 1, 0, 223, 0, 220, 0, 16, 0, 19, 0, 19, 0, 19, 0, 19, 0, 19, 0, 19, 0, 19, 0, 19, 0, 19, 0, 19, 0, 19, 0, 19, 0, 19, 0, 235, 0, 65, 2, 7, 0, 1, 0, 16, 0, 22, 0, 247, 0, 16, 0, 36, 1, 7, 0, 16, 0, 57, 1, 1, 0, 19, 0, 7, 0, 16, 0, 22, 0, 182, 2, 19, 0, 114, 1, 198, 1, 1, 0, 22, 0, 7, 0, 19, 0, 60, 1, 7, 0, 16, 0, 66, 1, 52, 0, 7, 0, 7, 0, 1, 0, 1, 0, 81, 1, 7, 0, 16, 0, 125, 2, 22, 0, 7, 0, 1, 0, 16, 0, 7, 0, 52, 0, 7, 0, 254, 3, 108, 1, 132, 1, 221, 2, 124, 0, 150, 1, 7, 0, 118, 0, 22, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 7, 0, 7, 0, 148, 0, 19, 0, 1, 0, 1, 0, 1, 0, 1, 0, 24, 1, 21, 1, 50, 2, 1, 0, 1, 0, 128, 2, 16, 0, 32, 2, 16, 0, 240, 1, 255, 1, 22, 0, 22, 0, 19, 0, 1, 0, 22, 0, 1, 0, 22, 0, 56, 2, 7, 0, 22, 0, 173, 2, 7, 0, 5, 2, 27, 1, 1, 0, 7, 0, 1, 0, 16, 0, 1, 0, 1, 0, 1, 0, 22, 0, 47, 2, 7, 0, 119, 2, 1, 0, 159, 1, 217, 2, 16, 0, 16, 0, 190, 0, 43, 0, 7, 0, 7, 0, 16, 0, 7, 0, 7, 0, 16, 0, 19, 0, 22, 0, 16, 0, 43, 0, 84, 1, 16, 0, 52, 0, 1, 0, 1, 0, 12, 1, 7, 0, 189, 1, 207, 1, 188, 2, 179, 2, 7, 0, 153, 1, 52, 0, 7, 0, 22, 0, 7, 0, 185, 2, 7, 0, 217, 2, 7, 0, 7, 0, 1, 0, 7, 0, 25, 0, 25, 0, 25, 0, 13, 0, 25, 0, 25, 0, 25, 0, 25, 0, 31, 0, 25, 0, 25, 0, 25, 0, 101, 2, 197, 2, 34, 0, 177, 1, 10, 0, 34, 0, 34, 0, 202, 0, 202, 0, 34, 0, 171, 1, 3, 1, 241, 0, 146, 2, 10, 0, 204, 1, 180, 1, 18, 1, 89, 18, 10, 0, 55, 0, 28, 0, 28, 0, 216, 1, 64, 0, 131, 2, 64, 0, 155, 2, 214, 0, 149, 2, 67, 0, 193, 0, 28, 0, 193, 0, 97, 0, 136, 0, 55, 0, 10, 0, 91, 0, 49, 0, 49, 0, 76, 0, 217, 0, 10, 0, 184, 0, 28, 0, 178, 0, 178, 0, 10, 0, 89, 18, 10, 0, 169, 0, 169, 0, 195, 1, 166, 0, 166, 0, 10, 0, 49, 0, 53, 2, 44, 2, 67, 0, 10, 0, 10, 0, 10, 0, 34, 0, 183, 1, 49, 0, 192, 1, 68, 2, 34, 0, 34, 0, 163, 0, 83, 2, 97, 0, 10, 0, 208, 0, 154, 0, 28, 0, 199, 0, 10, 0, 205, 0, 208, 0, 111, 1, 142, 0, 76, 0, 109, 0, 109, 0, 79, 0, 10, 0, 49, 0, 55, 0, 55, 0, 28, 0, 10, 0, 201, 1, 10, 0, 10, 0, 163, 0, 10, 0, 103, 0, 58, 0, 15, 1, 73, 0, 63, 1, 130, 0, 130, 0, 175, 0, 58, 0, 145, 0, 141, 1, 95, 2, 102, 1, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 13, 0, 16, 0, 64, 0, 16, 0, 16, 0, 16, 0, 16, 0, 172, 0, 212, 2, 7, 0, 1, 0, 212, 2, 28, 0, 7, 0, 7, 0, 7, 0, 7, 0, 7, 0, 7, 0, 7, 0, 7, 0, 7, 0, 43, 0, 43, 0, 118, 0, 151, 0, 133, 0, 212, 2, 212, 2, 212, 2, 70, 0, 139, 0, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 35, 2, 211, 0, 10, 0, 138, 1, 46, 0, 77, 2, 121, 0, 104, 2, 250, 0, 74, 2, 162, 1, 121, 0, 160, 0, 78, 1, 139, 0, 123, 1, 117, 1, 96, 1, 165, 1, 91, 0, 174, 1, 10, 0, 61, 0, 10, 0, 58, 0, 98, 2, 46, 0, 225, 1, 73, 0, 246, 3, 238, 0, 33, 1, 234, 1, 10, 0, 67, 0, 175, 0, 133, 0, 237, 1, 72, 1, 228, 1, 225, 2, 10, 0, 110, 2, 244, 0, 10, 0, 61, 0, 246, 1, 0, 1, 92, 2, 90, 1, 226, 0, 160, 0, 10, 0, 61, 0, 213, 1, 176, 2, 238, 3, 229, 2, 10, 0, 187, 0, 156, 1, 61, 0, 48, 1, 212, 2, 46, 0, 46, 0, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 136, 0, 1, 0, 89, 18, 145, 0, 243, 1, 168, 1, 69, 1, 39, 1, 186, 1, 140, 2, 86, 2, 11, 2, 8, 2, 200, 2, 71, 2, 142, 0, 76, 0, 79, 0, 148, 0, 79, 0, 85, 0, 157, 0, 212, 2, 100, 0, 212, 2, 212, 2, 22, 0, 22, 0, 22, 0, 103, 0, 31, 0, 31, 0, 31, 0, 1, 0, 184, 0, 1, 0, 28, 0, 190, 0, 1, 0, 212, 2, 100, 0, 31, 0, 181, 0, 31, 0, 37, 0, 43, 0, 191, 2, 82, 0, 152, 2, 99, 1, 164, 2, 124, 0, 88, 0, 196, 0, 144, 1, 1, 0, 94, 0, 82, 0, 37, 0, 85, 0, 85, 0, 88, 0, 94, 0, 38, 2, 41, 2, 26, 2, 20, 2, 40, 0, 62, 2, 59, 2, 37, 0, 37, 0, 181, 0, 51, 1, 252, 1, 40, 0, 88, 0, 82, 0, 161, 2, 37, 0, 94, 0, 194, 2, 37, 0, 187, 0, 58, 0, 205, 0, 154, 0, 196, 0, 199, 0, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 212, 2, 91, 0, 212, 2, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 40, 0, 212, 2, 212, 2, 97, 0, 10, 0, 212, 2, 212, 2, 170, 170, 11, 0, 0, 0, 9, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 9, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 9, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 9, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 8, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 10, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 8, 0, 0, 0, 226, 255, 255, 255, 255, 255, 255, 255, 172, 38, 0, 0, 1, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 172, 38, 0, 0, 1, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 8, 0, 0, 0, 2, 0, 0, 0, 255, 255, 255, 255, 64, 56, 0, 0, 2, 0, 0, 0, 3, 0, 0, 0, 2, 0, 0, 0, 255, 255, 255, 255, 48, 42, 0, 0, 2, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 8, 0, 0, 0, 255, 255, 255, 255, 0, 0, 0, 0, 1, 0, 0, 0, 10, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 0, 0, 0, 0, 1, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 16, 14, 0, 0, 2, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 16, 14, 0, 0, 2, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 16, 14, 0, 0, 2, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 16, 14, 0, 0, 2, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 10, 0, 0, 0, 8, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 0, 0, 0, 255, 255, 255, 255, 48, 42, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 8, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 10, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 20, 0, 0, 0, 0, 0, 0, 0, 128, 81, 1, 0, 0, 0, 0, 0, 8, 0, 0, 0, 20, 0, 0, 0, 0, 0, 0, 0, 128, 81, 1, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 225, 255, 255, 255, 251, 255, 255, 255, 128, 81, 1, 0, 0, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 250, 255, 255, 255, 0, 0, 0, 0, 1, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 9, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 8, 7, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 48, 42, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 8, 0, 0, 0, 226, 255, 255, 255, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 1, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 225, 255, 255, 255, 250, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 249, 255, 255, 255, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 9, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 22, 0, 0, 0, 255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 225, 255, 255, 255, 250, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 250, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 3, 0, 0, 0, 226, 255, 255, 255, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 8, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 10, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 16, 14, 0, 0, 2, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 16, 14, 0, 0, 2, 0, 0, 0, 32, 28, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 8, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 10, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 8, 0, 0, 0, 226, 255, 255, 255, 255, 255, 255, 255, 48, 42, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255, 64, 56, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 11, 0, 0, 0, 2, 0, 0, 0, 23, 0, 0, 0, 250, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 225, 255, 255, 255, 255, 255, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 29, 0, 144, 0, 147, 0, 150, 0, 153, 0, 156, 0, 162, 0, 169, 0, 177, 0, 183, 0, 188, 0, 191, 0, 198, 0, 203, 0, 209, 0, 214, 0, 221, 0, 224, 0, 232, 0, 239, 0, 247, 0, 250, 0, 4, 1, 9, 1, 15, 1, 23, 1, 29, 1, 35, 1, 38, 1, 41, 1, 113, 22, 0, 224, 125, 22, 0, 224, 137, 22, 0, 224, 149, 22, 0, 224, 161, 22, 0, 224, 173, 22, 0, 224, 185, 22, 0, 224, 197, 22, 0, 224, 209, 22, 0, 224, 221, 22, 0, 224, 233, 22, 0, 224, 245, 22, 0, 224, 1, 23, 0, 224, 13, 23, 0, 224, 25, 23, 0, 224, 37, 23, 0, 224, 49, 23, 0, 224, 61, 23, 0, 224, 73, 23, 0, 224, 85, 23, 0, 224, 97, 23, 0, 224, 109, 23, 0, 224, 121, 23, 0, 224, 133, 23, 0, 224, 145, 23, 0, 224, 157, 23, 0, 224, 169, 23, 0, 224, 181, 23, 0, 224, 193, 23, 0, 224, 11, 0, 0, 0, 5, 0, 0, 0, 11, 0, 0, 0, 13, 0, 0, 0, 21, 0, 0, 0, 22, 0, 0, 0, 27, 0, 0, 0, 38, 0, 0, 0, 51, 0, 0, 0, 52, 0, 0, 0, 55, 0, 0, 0, 91, 1, 0, 0, 1, 0, 0, 0, 72, 146, 230, 146, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 56, 252, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 249, 23, 0, 224, 5, 24, 0, 224, 7, 24, 0, 16, 9, 24, 0, 224, 47, 0, 0, 0, 180, 102, 48, 158, 128, 123, 52, 163, 80, 252, 211, 163, 0, 175, 21, 165, 208, 47, 181, 165, 128, 226, 246, 166, 80, 99, 150, 167, 0, 22, 216, 168, 208, 150, 119, 169, 0, 155, 186, 170, 208, 27, 90, 171, 128, 206, 155, 172, 80, 79, 59, 173, 0, 2, 125, 174, 208, 130, 28, 175, 128, 53, 94, 176, 80, 182, 253, 176, 128, 186, 64, 178, 80, 59, 224, 178, 0, 238, 33, 180, 208, 110, 193, 180, 128, 33, 3, 182, 80, 162, 162, 182, 0, 85, 228, 183, 208, 213, 131, 184, 0, 218, 198, 185, 208, 90, 102, 186, 128, 13, 168, 187, 80, 142, 71, 188, 0, 65, 137, 189, 208, 193, 40, 190, 128, 116, 106, 191, 80, 245, 9, 192, 128, 249, 76, 193, 80, 122, 236, 193, 0, 45, 46, 195, 208, 173, 205, 195, 128, 96, 15, 197, 80, 225, 174, 197, 0, 148, 240, 198, 208, 20, 144, 199, 0, 25, 211, 200, 208, 153, 114, 201, 128, 76, 180, 202, 80, 205, 83, 203, 0, 128, 149, 204, 208, 0, 53, 205, 170, 170, 170, 170, 170, 170, 170, 170, 47, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 204, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 176, 4, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 21, 24, 0, 224, 71, 24, 0, 16, 84, 24, 0, 224, 33, 0, 0, 0, 79, 80, 96, 145, 240, 120, 71, 155, 112, 44, 215, 155, 112, 145, 188, 156, 240, 72, 192, 157, 112, 254, 137, 158, 240, 42, 160, 159, 240, 165, 96, 160, 240, 12, 128, 161, 240, 18, 46, 162, 240, 76, 122, 163, 240, 129, 53, 164, 112, 6, 184, 164, 112, 6, 255, 198, 128, 186, 88, 199, 160, 9, 218, 199, 16, 52, 146, 207, 0, 0, 138, 208, 16, 22, 114, 209, 112, 36, 78, 210, 112, 7, 75, 212, 0, 211, 206, 229, 240, 176, 92, 243, 240, 193, 120, 2, 240, 200, 67, 3, 0, 215, 207, 13, 240, 68, 173, 14, 0, 90, 120, 15, 16, 89, 104, 16, 112, 67, 118, 18, 128, 66, 102, 19, 16, 124, 95, 20, 0, 95, 79, 21, 2, 0, 0, 0, 255, 255, 255, 255, 224, 73, 200, 107, 170, 170, 170, 170, 170, 170, 170, 170, 34, 0, 0, 0, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 4, 5, 4, 5, 4, 1, 4, 1, 2, 1, 2, 4, 5, 4, 1, 2, 1, 4, 170, 170, 12, 0, 0, 0, 220, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 49, 2, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 96, 24, 0, 224, 130, 24, 0, 224, 135, 24, 0, 16, 145, 24, 0, 224, 2, 0, 0, 0, 144, 156, 230, 146, 16, 97, 103, 9, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 100, 241, 255, 255, 0, 0, 0, 0, 240, 241, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 165, 24, 0, 224, 171, 24, 0, 16, 173, 24, 0, 224, 3, 0, 0, 0, 3, 0, 0, 0, 18, 0, 0, 0, 145, 1, 0, 0, 126, 0, 0, 0, 224, 180, 147, 200, 208, 123, 250, 200, 224, 239, 252, 201, 208, 232, 199, 202, 96, 174, 203, 203, 208, 41, 223, 204, 224, 225, 172, 205, 208, 244, 198, 206, 224, 102, 143, 207, 208, 121, 169, 208, 224, 96, 132, 209, 80, 173, 138, 210, 96, 99, 54, 232, 80, 45, 244, 232, 96, 185, 11, 234, 208, 96, 213, 234, 240, 250, 236, 235, 0, 109, 181, 236, 240, 127, 207, 237, 0, 242, 151, 238, 112, 179, 176, 239, 128, 37, 121, 240, 240, 230, 145, 241, 0, 89, 90, 242, 112, 26, 115, 243, 128, 140, 59, 244, 112, 159, 85, 245, 128, 17, 30, 246, 240, 210, 54, 247, 0, 69, 255, 247, 112, 6, 24, 249, 0, 202, 225, 249, 240, 57, 249, 250, 128, 253, 194, 251, 240, 190, 219, 252, 128, 130, 165, 253, 112, 242, 188, 254, 0, 182, 134, 255, 240, 37, 158, 0, 128, 233, 103, 1, 112, 89, 127, 2, 0, 29, 73, 3, 112, 222, 97, 4, 0, 162, 43, 5, 240, 17, 67, 6, 128, 213, 12, 7, 112, 69, 36, 8, 0, 9, 238, 8, 240, 120, 5, 10, 128, 60, 207, 10, 240, 253, 231, 11, 128, 193, 177, 12, 112, 49, 201, 13, 0, 245, 146, 14, 240, 100, 170, 15, 128, 40, 116, 16, 112, 152, 139, 17, 0, 92, 85, 18, 112, 29, 110, 19, 0, 225, 55, 20, 240, 80, 79, 21, 128, 20, 25, 22, 240, 147, 160, 23, 0, 72, 250, 23, 240, 163, 112, 25, 128, 123, 219, 25, 240, 60, 244, 26, 128, 0, 190, 27, 112, 112, 213, 28, 0, 52, 159, 29, 240, 163, 182, 30, 128, 103, 128, 31, 112, 215, 151, 32, 0, 155, 97, 33, 112, 92, 122, 34, 0, 32, 68, 35, 112, 39, 98, 36, 128, 83, 37, 37, 112, 195, 60, 38, 0, 135, 6, 39, 240, 246, 29, 40, 128, 186, 231, 40, 240, 123, 0, 42, 128, 63, 202, 42, 112, 175, 225, 43, 0, 115, 171, 44, 240, 226, 194, 45, 128, 166, 140, 46, 224, 19, 160, 47, 208, 12, 107, 48, 224, 245, 127, 49, 208, 238, 74, 50, 224, 215, 95, 51, 208, 208, 42, 52, 224, 185, 63, 53, 208, 178, 10, 54, 96, 214, 40, 55, 80, 207, 243, 55, 96, 184, 8, 57, 80, 177, 211, 57, 96, 154, 232, 58, 80, 147, 179, 59, 96, 124, 200, 60, 80, 117, 147, 61, 96, 94, 168, 62, 80, 87, 115, 63, 224, 122, 145, 64, 208, 115, 92, 65, 224, 92, 113, 66, 208, 85, 60, 67, 224, 62, 81, 68, 80, 253, 18, 69, 224, 32, 49, 70, 80, 106, 224, 70, 224, 2, 17, 72, 208, 17, 183, 72, 224, 228, 240, 73, 80, 185, 141, 74, 96, 1, 218, 75, 208, 189, 97, 76, 224, 88, 137, 76, 80, 250, 164, 76, 224, 56, 117, 83, 208, 137, 172, 83, 96, 188, 218, 83, 80, 130, 36, 84, 2, 0, 0, 0, 255, 255, 255, 255, 171, 77, 189, 125, 127, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 85, 29, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 185, 24, 0, 224, 189, 24, 0, 224, 60, 25, 0, 224, 63, 25, 0, 16, 96, 25, 0, 224, 94, 0, 0, 0, 156, 249, 81, 150, 128, 20, 255, 198, 112, 172, 88, 199, 128, 237, 217, 199, 240, 50, 161, 210, 0, 164, 53, 219, 240, 39, 238, 219, 64, 114, 37, 251, 112, 239, 194, 251, 128, 132, 107, 8, 240, 109, 198, 8, 0, 12, 232, 11, 240, 71, 97, 12, 128, 63, 201, 13, 112, 242, 142, 14, 128, 81, 211, 15, 112, 163, 39, 16, 0, 166, 183, 26, 240, 111, 24, 30, 128, 230, 65, 72, 112, 34, 187, 72, 0, 26, 35, 74, 112, 213, 141, 74, 128, 192, 220, 75, 112, 229, 93, 76, 128, 184, 151, 77, 240, 140, 52, 78, 160, 160, 156, 79, 160, 187, 8, 80, 32, 154, 49, 80, 160, 167, 103, 80, 160, 130, 124, 81, 160, 203, 216, 81, 160, 158, 5, 82, 160, 115, 108, 82, 160, 122, 55, 83, 160, 33, 174, 83, 32, 70, 220, 83, 160, 85, 76, 84, 160, 92, 23, 85, 32, 224, 124, 85, 160, 4, 171, 85, 160, 55, 44, 86, 160, 62, 247, 86, 160, 135, 83, 87, 32, 172, 129, 87, 32, 84, 21, 88, 160, 32, 215, 88, 160, 244, 32, 89, 160, 83, 88, 89, 32, 54, 245, 89, 160, 2, 183, 90, 32, 156, 247, 90, 160, 192, 37, 91, 160, 67, 206, 92, 32, 104, 252, 92, 160, 176, 155, 94, 32, 213, 201, 94, 32, 88, 114, 96, 160, 124, 160, 96, 32, 197, 63, 98, 32, 36, 119, 98, 160, 108, 22, 100, 32, 145, 68, 100, 32, 20, 237, 101, 160, 56, 27, 102, 32, 129, 186, 103, 32, 224, 241, 103, 160, 40, 145, 105, 32, 77, 191, 105, 32, 208, 103, 107, 160, 244, 149, 107, 32, 61, 53, 109, 160, 97, 99, 109, 160, 228, 11, 111, 32, 9, 58, 111, 160, 81, 217, 112, 160, 176, 16, 113, 32, 249, 175, 114, 160, 29, 222, 114, 160, 160, 134, 116, 32, 197, 180, 116, 160, 13, 84, 118, 160, 108, 139, 118, 32, 181, 42, 120, 160, 217, 88, 120, 32, 34, 248, 121, 32, 129, 47, 122, 160, 201, 206, 123, 32, 238, 252, 123, 32, 113, 165, 125, 160, 149, 211, 125, 32, 222, 114, 127, 32, 61, 170, 127, 204, 0, 0, 0, 0, 0, 0, 0, 160, 133, 73, 129, 0, 0, 0, 0, 32, 170, 119, 129, 0, 0, 0, 0, 32, 45, 32, 131, 0, 0, 0, 0, 160, 81, 78, 131, 0, 0, 0, 0, 32, 154, 237, 132, 0, 0, 0, 0, 32, 249, 36, 133, 0, 0, 0, 0, 160, 65, 196, 134, 0, 0, 0, 0, 32, 102, 242, 134, 0, 0, 0, 0, 160, 174, 145, 136, 0, 0, 0, 0, 160, 13, 201, 136, 0, 0, 0, 0, 32, 86, 104, 138, 0, 0, 0, 0, 160, 122, 150, 138, 0, 0, 0, 0, 160, 253, 62, 140, 0, 0, 0, 0, 32, 34, 109, 140, 0, 0, 0, 0, 160, 106, 12, 142, 0, 0, 0, 0, 160, 201, 67, 142, 0, 0, 0, 0, 32, 18, 227, 143, 0, 0, 0, 0, 160, 54, 17, 144, 0, 0, 0, 0, 160, 185, 185, 145, 0, 0, 0, 0, 32, 222, 231, 145, 0, 0, 0, 0, 160, 38, 135, 147, 0, 0, 0, 0, 160, 133, 190, 147, 0, 0, 0, 0, 32, 206, 93, 149, 0, 0, 0, 0, 160, 242, 139, 149, 0, 0, 0, 0, 32, 59, 43, 151, 0, 0, 0, 0, 32, 154, 98, 151, 0, 0, 0, 0, 160, 226, 1, 153, 0, 0, 0, 0, 32, 7, 48, 153, 0, 0, 0, 0, 32, 138, 216, 154, 0, 0, 0, 0, 160, 174, 6, 155, 0, 0, 0, 0, 32, 247, 165, 156, 0, 0, 0, 0, 32, 86, 221, 156, 0, 0, 0, 0, 160, 158, 124, 158, 0, 0, 0, 0, 32, 195, 170, 158, 0, 0, 0, 0, 32, 70, 83, 160, 0, 0, 0, 0, 160, 106, 129, 160, 0, 0, 0, 0, 32, 179, 32, 162, 0, 0, 0, 0, 32, 18, 88, 162, 0, 0, 0, 0, 160, 90, 247, 163, 0, 0, 0, 0, 32, 127, 37, 164, 0, 0, 0, 0, 160, 199, 196, 165, 0, 0, 0, 0, 160, 38, 252, 165, 0, 0, 0, 0, 32, 111, 155, 167, 0, 0, 0, 0, 160, 147, 201, 167, 0, 0, 0, 0, 160, 22, 114, 169, 0, 0, 0, 0, 32, 59, 160, 169, 0, 0, 0, 0, 160, 131, 63, 171, 0, 0, 0, 0, 160, 226, 118, 171, 0, 0, 0, 0, 32, 43, 22, 173, 0, 0, 0, 0, 160, 79, 68, 173, 0, 0, 0, 0, 160, 210, 236, 174, 0, 0, 0, 0, 32, 247, 26, 175, 0, 0, 0, 0, 160, 63, 186, 176, 0, 0, 0, 0, 32, 100, 232, 176, 0, 0, 0, 0, 32, 231, 144, 178, 0, 0, 0, 0, 160, 11, 191, 178, 0, 0, 0, 0, 32, 84, 94, 180, 0, 0, 0, 0, 32, 179, 149, 180, 0, 0, 0, 0, 160, 251, 52, 182, 0, 0, 0, 0, 32, 32, 99, 182, 0, 0, 0, 0, 32, 163, 11, 184, 0, 0, 0, 0, 160, 199, 57, 184, 0, 0, 0, 0, 32, 16, 217, 185, 0, 0, 0, 0, 32, 111, 16, 186, 0, 0, 0, 0, 160, 183, 175, 187, 0, 0, 0, 0, 32, 220, 221, 187, 0, 0, 0, 0, 32, 95, 134, 189, 0, 0, 0, 0, 160, 131, 180, 189, 0, 0, 0, 0, 32, 204, 83, 191, 0, 0, 0, 0, 160, 240, 129, 191, 0, 0, 0, 0, 160, 115, 42, 193, 0, 0, 0, 0, 32, 152, 88, 193, 0, 0, 0, 0, 160, 224, 247, 194, 0, 0, 0, 0, 160, 63, 47, 195, 0, 0, 0, 0, 32, 136, 206, 196, 0, 0, 0, 0, 160, 172, 252, 196, 0, 0, 0, 0, 160, 47, 165, 198, 0, 0, 0, 0, 32, 84, 211, 198, 0, 0, 0, 0, 160, 156, 114, 200, 0, 0, 0, 0, 160, 251, 169, 200, 0, 0, 0, 0, 32, 68, 73, 202, 0, 0, 0, 0, 160, 104, 119, 202, 0, 0, 0, 0, 160, 235, 31, 204, 0, 0, 0, 0, 32, 16, 78, 204, 0, 0, 0, 0, 160, 88, 237, 205, 0, 0, 0, 0, 32, 125, 27, 206, 0, 0, 0, 0, 32, 0, 196, 207, 0, 0, 0, 0, 160, 36, 242, 207, 0, 0, 0, 0, 32, 109, 145, 209, 0, 0, 0, 0, 32, 204, 200, 209, 0, 0, 0, 0, 160, 20, 104, 211, 0, 0, 0, 0, 32, 57, 150, 211, 0, 0, 0, 0, 32, 188, 62, 213, 0, 0, 0, 0, 160, 224, 108, 213, 0, 0, 0, 0, 32, 41, 12, 215, 0, 0, 0, 0, 32, 136, 67, 215, 0, 0, 0, 0, 160, 208, 226, 216, 0, 0, 0, 0, 32, 245, 16, 217, 0, 0, 0, 0, 32, 120, 185, 218, 0, 0, 0, 0, 160, 156, 231, 218, 0, 0, 0, 0, 32, 229, 134, 220, 0, 0, 0, 0, 160, 9, 181, 220, 196, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 8, 0, 0, 0, 228, 248, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 91, 0, 66, 0, 54, 0, 170, 170, 111, 25, 0, 224, 206, 25, 0, 224, 155, 26, 0, 16, 205, 26, 0, 224, 43, 0, 0, 0, 112, 117, 214, 158, 96, 110, 161, 159, 112, 239, 5, 170, 0, 110, 231, 170, 240, 167, 201, 173, 0, 50, 167, 174, 112, 79, 160, 175, 0, 20, 135, 176, 0, 122, 137, 177, 128, 48, 112, 178, 64, 114, 37, 251, 112, 239, 194, 251, 128, 132, 107, 8, 240, 109, 198, 8, 0, 12, 232, 11, 240, 71, 97, 12, 128, 63, 201, 13, 112, 242, 142, 14, 128, 81, 211, 15, 112, 163, 39, 16, 0, 166, 183, 26, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 0, 181, 54, 126, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 44, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 10, 0, 0, 0, 4, 251, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 221, 26, 0, 224, 9, 27, 0, 224, 15, 27, 0, 16, 27, 27, 0, 224, 83, 0, 0, 0, 224, 240, 72, 188, 144, 176, 209, 11, 0, 12, 232, 11, 240, 71, 97, 12, 128, 63, 201, 13, 112, 242, 142, 14, 128, 81, 211, 15, 112, 163, 39, 16, 128, 230, 65, 72, 112, 34, 187, 72, 0, 26, 35, 74, 112, 213, 141, 74, 128, 192, 220, 75, 112, 229, 93, 76, 128, 184, 151, 77, 240, 140, 52, 78, 160, 160, 156, 79, 160, 187, 8, 80, 32, 154, 49, 80, 160, 167, 103, 80, 160, 130, 124, 81, 160, 203, 216, 81, 160, 158, 5, 82, 160, 115, 108, 82, 160, 122, 55, 83, 160, 33, 174, 83, 32, 70, 220, 83, 160, 85, 76, 84, 160, 92, 23, 85, 32, 224, 124, 85, 160, 4, 171, 85, 160, 55, 44, 86, 160, 62, 247, 86, 160, 135, 83, 87, 32, 172, 129, 87, 32, 84, 21, 88, 160, 32, 215, 88, 160, 244, 32, 89, 160, 83, 88, 89, 32, 54, 245, 89, 160, 2, 183, 90, 32, 156, 247, 90, 160, 192, 37, 91, 160, 67, 206, 92, 32, 104, 252, 92, 160, 176, 155, 94, 32, 213, 201, 94, 32, 88, 114, 96, 160, 124, 160, 96, 32, 197, 63, 98, 32, 36, 119, 98, 160, 108, 22, 100, 32, 145, 68, 100, 32, 20, 237, 101, 160, 56, 27, 102, 32, 129, 186, 103, 32, 224, 241, 103, 160, 40, 145, 105, 32, 77, 191, 105, 32, 208, 103, 107, 160, 244, 149, 107, 32, 61, 53, 109, 160, 97, 99, 109, 160, 228, 11, 111, 32, 9, 58, 111, 160, 81, 217, 112, 160, 176, 16, 113, 32, 249, 175, 114, 160, 29, 222, 114, 160, 160, 134, 116, 32, 197, 180, 116, 160, 13, 84, 118, 160, 108, 139, 118, 32, 181, 42, 120, 160, 217, 88, 120, 32, 34, 248, 121, 32, 129, 47, 122, 160, 201, 206, 123, 32, 238, 252, 123, 32, 113, 165, 125, 160, 149, 211, 125, 32, 222, 114, 127, 32, 61, 170, 127, 204, 0, 0, 0, 0, 0, 0, 0, 160, 133, 73, 129, 0, 0, 0, 0, 32, 170, 119, 129, 0, 0, 0, 0, 32, 45, 32, 131, 0, 0, 0, 0, 160, 81, 78, 131, 0, 0, 0, 0, 32, 154, 237, 132, 0, 0, 0, 0, 32, 249, 36, 133, 0, 0, 0, 0, 160, 65, 196, 134, 0, 0, 0, 0, 32, 102, 242, 134, 0, 0, 0, 0, 160, 174, 145, 136, 0, 0, 0, 0, 160, 13, 201, 136, 0, 0, 0, 0, 32, 86, 104, 138, 0, 0, 0, 0, 160, 122, 150, 138, 0, 0, 0, 0, 160, 253, 62, 140, 0, 0, 0, 0, 32, 34, 109, 140, 0, 0, 0, 0, 160, 106, 12, 142, 0, 0, 0, 0, 160, 201, 67, 142, 0, 0, 0, 0, 32, 18, 227, 143, 0, 0, 0, 0, 160, 54, 17, 144, 0, 0, 0, 0, 160, 185, 185, 145, 0, 0, 0, 0, 32, 222, 231, 145, 0, 0, 0, 0, 160, 38, 135, 147, 0, 0, 0, 0, 160, 133, 190, 147, 0, 0, 0, 0, 32, 206, 93, 149, 0, 0, 0, 0, 160, 242, 139, 149, 0, 0, 0, 0, 32, 59, 43, 151, 0, 0, 0, 0, 32, 154, 98, 151, 0, 0, 0, 0, 160, 226, 1, 153, 0, 0, 0, 0, 32, 7, 48, 153, 0, 0, 0, 0, 32, 138, 216, 154, 0, 0, 0, 0, 160, 174, 6, 155, 0, 0, 0, 0, 32, 247, 165, 156, 0, 0, 0, 0, 32, 86, 221, 156, 0, 0, 0, 0, 160, 158, 124, 158, 0, 0, 0, 0, 32, 195, 170, 158, 0, 0, 0, 0, 32, 70, 83, 160, 0, 0, 0, 0, 160, 106, 129, 160, 0, 0, 0, 0, 32, 179, 32, 162, 0, 0, 0, 0, 32, 18, 88, 162, 0, 0, 0, 0, 160, 90, 247, 163, 0, 0, 0, 0, 32, 127, 37, 164, 0, 0, 0, 0, 160, 199, 196, 165, 0, 0, 0, 0, 160, 38, 252, 165, 0, 0, 0, 0, 32, 111, 155, 167, 0, 0, 0, 0, 160, 147, 201, 167, 0, 0, 0, 0, 160, 22, 114, 169, 0, 0, 0, 0, 32, 59, 160, 169, 0, 0, 0, 0, 160, 131, 63, 171, 0, 0, 0, 0, 160, 226, 118, 171, 0, 0, 0, 0, 32, 43, 22, 173, 0, 0, 0, 0, 160, 79, 68, 173, 0, 0, 0, 0, 160, 210, 236, 174, 0, 0, 0, 0, 32, 247, 26, 175, 0, 0, 0, 0, 160, 63, 186, 176, 0, 0, 0, 0, 32, 100, 232, 176, 0, 0, 0, 0, 32, 231, 144, 178, 0, 0, 0, 0, 160, 11, 191, 178, 0, 0, 0, 0, 32, 84, 94, 180, 0, 0, 0, 0, 32, 179, 149, 180, 0, 0, 0, 0, 160, 251, 52, 182, 0, 0, 0, 0, 32, 32, 99, 182, 0, 0, 0, 0, 32, 163, 11, 184, 0, 0, 0, 0, 160, 199, 57, 184, 0, 0, 0, 0, 32, 16, 217, 185, 0, 0, 0, 0, 32, 111, 16, 186, 0, 0, 0, 0, 160, 183, 175, 187, 0, 0, 0, 0, 32, 220, 221, 187, 0, 0, 0, 0, 32, 95, 134, 189, 0, 0, 0, 0, 160, 131, 180, 189, 0, 0, 0, 0, 32, 204, 83, 191, 0, 0, 0, 0, 160, 240, 129, 191, 0, 0, 0, 0, 160, 115, 42, 193, 0, 0, 0, 0, 32, 152, 88, 193, 0, 0, 0, 0, 160, 224, 247, 194, 0, 0, 0, 0, 160, 63, 47, 195, 0, 0, 0, 0, 32, 136, 206, 196, 0, 0, 0, 0, 160, 172, 252, 196, 0, 0, 0, 0, 160, 47, 165, 198, 0, 0, 0, 0, 32, 84, 211, 198, 0, 0, 0, 0, 160, 156, 114, 200, 0, 0, 0, 0, 160, 251, 169, 200, 0, 0, 0, 0, 32, 68, 73, 202, 0, 0, 0, 0, 160, 104, 119, 202, 0, 0, 0, 0, 160, 235, 31, 204, 0, 0, 0, 0, 32, 16, 78, 204, 0, 0, 0, 0, 160, 88, 237, 205, 0, 0, 0, 0, 32, 125, 27, 206, 0, 0, 0, 0, 32, 0, 196, 207, 0, 0, 0, 0, 160, 36, 242, 207, 0, 0, 0, 0, 32, 109, 145, 209, 0, 0, 0, 0, 32, 204, 200, 209, 0, 0, 0, 0, 160, 20, 104, 211, 0, 0, 0, 0, 32, 57, 150, 211, 0, 0, 0, 0, 32, 188, 62, 213, 0, 0, 0, 0, 160, 224, 108, 213, 0, 0, 0, 0, 32, 41, 12, 215, 0, 0, 0, 0, 32, 136, 67, 215, 0, 0, 0, 0, 160, 208, 226, 216, 0, 0, 0, 0, 32, 245, 16, 217, 0, 0, 0, 0, 32, 120, 185, 218, 0, 0, 0, 0, 160, 156, 231, 218, 0, 0, 0, 0, 32, 229, 134, 220, 0, 0, 0, 0, 160, 9, 181, 220, 170, 170, 170, 170, 185, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 170, 170, 170, 8, 0, 0, 0, 160, 243, 255, 255, 0, 0, 0, 0, 240, 241, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 4, 0, 48, 0, 91, 0, 66, 0, 54, 0, 170, 170, 49, 27, 0, 224, 133, 27, 0, 224, 83, 28, 0, 16, 131, 28, 0, 224, 3, 0, 0, 0, 30, 0, 0, 0, 44, 0, 0, 0, 45, 0, 0, 0, 5, 0, 0, 0, 104, 207, 70, 130, 128, 140, 174, 204, 112, 111, 158, 205, 128, 110, 142, 206, 112, 81, 126, 207, 2, 0, 0, 0, 255, 255, 255, 255, 64, 65, 123, 109, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 6, 0, 0, 0, 1, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 64, 26, 0, 0, 0, 0, 0, 0, 24, 21, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 147, 28, 0, 224, 151, 28, 0, 224, 157, 28, 0, 224, 163, 28, 0, 16, 166, 28, 0, 224, 34, 0, 0, 0, 220, 218, 163, 182, 224, 23, 158, 0, 80, 52, 122, 1, 224, 249, 125, 2, 208, 103, 91, 3, 224, 126, 96, 4, 208, 236, 61, 5, 224, 96, 64, 6, 80, 32, 31, 7, 224, 66, 32, 8, 208, 83, 0, 9, 224, 36, 0, 10, 80, 135, 225, 10, 224, 6, 224, 11, 80, 12, 196, 12, 224, 232, 191, 13, 208, 63, 165, 14, 96, 5, 169, 15, 80, 115, 134, 16, 96, 231, 136, 17, 208, 166, 103, 18, 96, 201, 104, 19, 208, 43, 74, 20, 96, 171, 72, 21, 80, 95, 43, 22, 96, 141, 40, 23, 208, 146, 12, 24, 96, 111, 8, 25, 80, 198, 237, 25, 224, 139, 241, 26, 80, 75, 208, 27, 224, 109, 209, 28, 208, 126, 177, 29, 32, 69, 128, 56, 170, 170, 170, 170, 34, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 170, 170, 8, 0, 0, 0, 164, 29, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 183, 28, 0, 224, 219, 28, 0, 16, 229, 28, 0, 224, 35, 0, 0, 0, 0, 218, 163, 182, 224, 23, 158, 0, 80, 52, 122, 1, 224, 249, 125, 2, 208, 103, 91, 3, 224, 126, 96, 4, 208, 236, 61, 5, 224, 96, 64, 6, 80, 32, 31, 7, 224, 66, 32, 8, 208, 83, 0, 9, 224, 36, 0, 10, 80, 135, 225, 10, 224, 6, 224, 11, 80, 12, 196, 12, 224, 232, 191, 13, 208, 63, 165, 14, 96, 5, 169, 15, 80, 115, 134, 16, 96, 231, 136, 17, 208, 166, 103, 18, 96, 201, 104, 19, 208, 43, 74, 20, 96, 171, 72, 21, 80, 95, 43, 22, 96, 141, 40, 23, 208, 146, 12, 24, 96, 111, 8, 25, 80, 198, 237, 25, 224, 139, 241, 26, 80, 75, 208, 27, 224, 109, 209, 28, 208, 126, 177, 29, 32, 69, 128, 56, 80, 228, 248, 89, 35, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 170, 8, 0, 0, 0, 128, 30, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 243, 28, 0, 224, 23, 29, 0, 16, 33, 29, 0, 224, 10, 0, 0, 0, 12, 0, 0, 0, 16, 0, 0, 0, 25, 0, 0, 0, 35, 0, 0, 0, 36, 0, 0, 0, 37, 0, 0, 0, 39, 0, 0, 0, 42, 0, 0, 0, 50, 0, 0, 0, 53, 0, 0, 0, 1, 0, 0, 0, 80, 243, 81, 161, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 48, 3, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 47, 29, 0, 224, 58, 29, 0, 224, 63, 29, 0, 16, 65, 29, 0, 224, 9, 0, 0, 0, 15, 0, 0, 0, 17, 0, 0, 0, 28, 0, 0, 0, 29, 0, 0, 0, 34, 0, 0, 0, 40, 0, 0, 0, 41, 0, 0, 0, 43, 0, 0, 0, 122, 1, 0, 0, 1, 0, 0, 0, 244, 197, 70, 130, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 140, 30, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 77, 29, 0, 224, 87, 29, 0, 224, 91, 29, 0, 16, 93, 29, 0, 224, 2, 0, 0, 0, 156, 108, 95, 160, 110, 90, 202, 3, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 228, 245, 255, 255, 0, 0, 0, 0, 146, 245, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 105, 29, 0, 224, 111, 29, 0, 16, 113, 29, 0, 224, 12, 0, 0, 0, 7, 0, 0, 0, 9, 0, 0, 0, 10, 0, 0, 0, 23, 0, 0, 0, 24, 0, 0, 0, 32, 0, 0, 0, 46, 0, 0, 0, 48, 0, 0, 0, 140, 1, 0, 0, 3, 2, 0, 0, 7, 2, 0, 0, 12, 2, 0, 0, 4, 0, 0, 0, 252, 218, 238, 177, 208, 154, 194, 180, 216, 71, 145, 199, 212, 225, 47, 237, 4, 0, 0, 0, 3, 1, 2, 3, 8, 0, 0, 0, 132, 34, 0, 0, 0, 0, 0, 0, 40, 35, 0, 0, 0, 0, 0, 0, 172, 38, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 125, 29, 0, 224, 138, 29, 0, 224, 143, 29, 0, 16, 145, 29, 0, 224, 3, 0, 0, 0, 100, 128, 230, 146, 112, 113, 102, 18, 96, 222, 38, 19, 170, 170, 170, 170, 170, 170, 170, 170, 3, 0, 0, 0, 1, 2, 1, 170, 6, 0, 0, 0, 28, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 161, 29, 0, 224, 167, 29, 0, 16, 169, 29, 0, 224, 3, 0, 0, 0, 128, 142, 230, 146, 16, 136, 73, 90, 144, 187, 42, 92, 2, 0, 0, 0, 255, 255, 255, 255, 48, 253, 60, 94, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 4, 0, 0, 0, 1, 2, 3, 2, 8, 0, 0, 0, 80, 6, 0, 0, 0, 0, 0, 0, 99, 247, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 181, 29, 0, 224, 185, 29, 0, 224, 191, 29, 0, 16, 193, 29, 0, 224, 2, 0, 0, 0, 56, 0, 0, 0, 20, 2, 0, 0, 32, 0, 0, 0, 36, 193, 242, 161, 16, 177, 187, 221, 96, 173, 35, 222, 16, 210, 120, 225, 224, 101, 231, 225, 112, 63, 47, 229, 224, 204, 169, 229, 240, 198, 78, 235, 96, 66, 146, 22, 112, 247, 8, 23, 224, 43, 250, 23, 240, 42, 234, 24, 96, 95, 219, 25, 240, 175, 204, 26, 96, 228, 189, 27, 240, 122, 180, 28, 224, 23, 159, 29, 112, 11, 147, 30, 96, 238, 130, 31, 112, 74, 112, 32, 224, 126, 97, 33, 112, 207, 82, 34, 224, 3, 68, 35, 240, 2, 52, 36, 96, 55, 37, 37, 240, 183, 64, 38, 96, 241, 78, 50, 112, 54, 68, 51, 224, 106, 53, 52, 0, 153, 157, 80, 128, 217, 84, 81, 128, 180, 105, 82, 170, 170, 170, 170, 170, 170, 170, 170, 32, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 2, 3, 1, 2, 3, 8, 0, 0, 0, 92, 12, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 209, 29, 0, 224, 212, 29, 0, 224, 247, 29, 0, 16, 0, 30, 0, 224, 33, 0, 0, 0, 79, 80, 96, 145, 224, 136, 58, 198, 96, 158, 88, 199, 224, 34, 219, 199, 224, 84, 226, 202, 240, 105, 173, 203, 16, 75, 231, 204, 144, 23, 169, 205, 0, 22, 194, 205, 16, 176, 204, 205, 0, 53, 162, 206, 16, 52, 146, 207, 224, 227, 137, 208, 16, 22, 114, 209, 96, 22, 78, 210, 240, 223, 199, 13, 112, 172, 137, 14, 240, 100, 170, 15, 112, 26, 116, 16, 240, 58, 163, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 112, 195, 60, 38, 112, 39, 5, 39, 240, 13, 116, 66, 0, 128, 60, 67, 144, 231, 37, 68, 16, 253, 67, 69, 144, 201, 5, 70, 16, 223, 35, 71, 16, 230, 238, 71, 16, 193, 3, 73, 2, 0, 0, 0, 255, 255, 255, 255, 244, 19, 70, 89, 170, 170, 170, 170, 170, 170, 170, 170, 34, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 140, 9, 0, 0, 0, 0, 0, 0, 49, 2, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 16, 30, 0, 224, 50, 30, 0, 224, 55, 30, 0, 16, 65, 30, 0, 224, 52, 0, 0, 0, 104, 207, 70, 130, 128, 140, 174, 204, 112, 111, 158, 205, 96, 199, 140, 45, 16, 28, 105, 46, 0, 233, 125, 47, 16, 254, 72, 48, 128, 5, 103, 49, 16, 224, 40, 50, 128, 231, 70, 51, 144, 252, 17, 52, 128, 201, 38, 53, 144, 222, 241, 53, 128, 171, 6, 55, 144, 192, 209, 55, 128, 141, 230, 56, 144, 162, 177, 57, 128, 111, 198, 58, 144, 132, 145, 59, 0, 140, 175, 60, 144, 102, 113, 61, 0, 110, 143, 62, 16, 131, 90, 63, 0, 80, 111, 64, 16, 101, 58, 65, 0, 50, 79, 66, 16, 71, 26, 67, 0, 20, 47, 68, 16, 41, 250, 68, 0, 246, 14, 70, 16, 11, 218, 70, 128, 18, 248, 71, 144, 39, 195, 72, 128, 244, 215, 73, 144, 9, 163, 74, 128, 214, 183, 75, 144, 235, 130, 76, 128, 184, 151, 77, 144, 205, 98, 78, 128, 154, 119, 79, 144, 175, 66, 80, 0, 183, 96, 81, 144, 145, 34, 82, 0, 153, 64, 83, 16, 174, 11, 84, 0, 123, 32, 85, 16, 144, 235, 85, 0, 93, 0, 87, 16, 114, 203, 87, 0, 63, 224, 88, 16, 84, 171, 89, 96, 102, 238, 89, 2, 0, 0, 0, 255, 255, 255, 255, 120, 75, 123, 109, 170, 170, 170, 170, 170, 170, 170, 170, 53, 0, 0, 0, 3, 4, 5, 4, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 170, 170, 170, 12, 0, 0, 0, 8, 16, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 24, 21, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 81, 30, 0, 224, 134, 30, 0, 224, 139, 30, 0, 16, 154, 30, 0, 224, 3, 0, 0, 0, 59, 0, 0, 0, 80, 0, 0, 0, 104, 2, 0, 0, 80, 0, 0, 0, 208, 68, 137, 203, 64, 80, 97, 210, 80, 113, 184, 254, 64, 84, 168, 255, 80, 83, 152, 0, 64, 54, 136, 1, 80, 53, 120, 2, 192, 82, 113, 3, 208, 81, 97, 4, 192, 52, 81, 5, 208, 51, 65, 6, 192, 22, 49, 7, 208, 109, 141, 7, 192, 248, 16, 9, 80, 233, 173, 9, 192, 218, 240, 10, 208, 217, 224, 11, 64, 247, 217, 12, 208, 187, 192, 13, 64, 217, 185, 14, 80, 216, 169, 15, 64, 187, 153, 16, 80, 186, 137, 17, 64, 157, 121, 18, 80, 156, 105, 19, 64, 127, 89, 20, 80, 126, 73, 21, 64, 97, 57, 22, 80, 96, 41, 23, 192, 125, 34, 24, 80, 66, 9, 25, 192, 95, 2, 26, 192, 80, 242, 26, 176, 51, 226, 27, 192, 50, 210, 28, 176, 21, 194, 29, 192, 20, 178, 30, 176, 247, 161, 31, 64, 71, 118, 32, 176, 217, 129, 33, 64, 41, 86, 34, 48, 246, 106, 35, 64, 11, 54, 36, 48, 216, 74, 37, 64, 237, 21, 38, 48, 186, 42, 39, 192, 9, 255, 39, 48, 156, 10, 41, 192, 235, 222, 41, 48, 126, 234, 42, 192, 205, 190, 43, 176, 154, 211, 44, 192, 175, 158, 45, 176, 124, 179, 46, 192, 145, 126, 47, 176, 94, 147, 48, 64, 174, 103, 49, 176, 64, 115, 50, 64, 144, 71, 51, 176, 34, 83, 52, 64, 114, 39, 53, 176, 4, 51, 54, 64, 84, 7, 55, 48, 33, 28, 56, 64, 54, 231, 56, 48, 3, 252, 57, 64, 24, 199, 58, 48, 229, 219, 59, 192, 52, 176, 60, 48, 199, 187, 61, 192, 22, 144, 62, 48, 169, 155, 63, 192, 248, 111, 64, 176, 197, 132, 65, 192, 218, 79, 66, 176, 167, 100, 67, 192, 188, 47, 68, 176, 137, 68, 69, 64, 239, 243, 69, 48, 166, 45, 71, 4, 0, 0, 0, 255, 255, 255, 255, 209, 253, 194, 63, 255, 255, 255, 255, 94, 90, 135, 125, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 82, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 170, 170, 12, 0, 0, 0, 226, 171, 0, 0, 0, 0, 0, 0, 98, 90, 255, 255, 0, 0, 0, 0, 80, 101, 255, 255, 0, 0, 0, 0, 80, 101, 255, 255, 16, 14, 0, 0, 96, 115, 255, 255, 0, 0, 0, 0, 96, 115, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 96, 115, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 174, 30, 0, 224, 178, 30, 0, 224, 3, 31, 0, 224, 11, 31, 0, 16, 33, 31, 0, 224, 3, 0, 0, 0, 4, 0, 0, 0, 60, 0, 0, 0, 103, 2, 0, 0, 80, 0, 0, 0, 192, 54, 137, 203, 48, 66, 97, 210, 64, 99, 184, 254, 48, 70, 168, 255, 64, 69, 152, 0, 48, 40, 136, 1, 64, 39, 120, 2, 176, 68, 113, 3, 192, 67, 97, 4, 176, 38, 81, 5, 192, 37, 65, 6, 176, 8, 49, 7, 192, 95, 141, 7, 176, 234, 16, 9, 64, 219, 173, 9, 176, 204, 240, 10, 192, 203, 224, 11, 48, 233, 217, 12, 192, 173, 192, 13, 48, 203, 185, 14, 64, 202, 169, 15, 48, 173, 153, 16, 64, 172, 137, 17, 48, 143, 121, 18, 64, 142, 105, 19, 48, 113, 89, 20, 64, 112, 73, 21, 48, 83, 57, 22, 64, 82, 41, 23, 176, 111, 34, 24, 64, 52, 9, 25, 176, 81, 2, 26, 176, 66, 242, 26, 160, 37, 226, 27, 176, 36, 210, 28, 160, 7, 194, 29, 176, 6, 178, 30, 160, 233, 161, 31, 48, 57, 118, 32, 160, 203, 129, 33, 48, 27, 86, 34, 32, 232, 106, 35, 48, 253, 53, 36, 32, 202, 74, 37, 48, 223, 21, 38, 32, 172, 42, 39, 176, 251, 254, 39, 32, 142, 10, 41, 176, 221, 222, 41, 32, 112, 234, 42, 176, 191, 190, 43, 160, 140, 211, 44, 176, 161, 158, 45, 160, 110, 179, 46, 176, 131, 126, 47, 160, 80, 147, 48, 48, 160, 103, 49, 160, 50, 115, 50, 48, 130, 71, 51, 160, 20, 83, 52, 48, 100, 39, 53, 160, 246, 50, 54, 48, 70, 7, 55, 32, 19, 28, 56, 48, 40, 231, 56, 32, 245, 251, 57, 48, 10, 199, 58, 32, 215, 219, 59, 176, 38, 176, 60, 32, 185, 187, 61, 176, 8, 144, 62, 32, 155, 155, 63, 176, 234, 111, 64, 160, 183, 132, 65, 176, 204, 79, 66, 160, 153, 100, 67, 176, 174, 47, 68, 160, 123, 68, 69, 48, 225, 243, 69, 32, 152, 45, 71, 4, 0, 0, 0, 255, 255, 255, 255, 209, 253, 194, 63, 255, 255, 255, 255, 72, 65, 135, 125, 170, 170, 170, 170, 170, 170, 170, 170, 82, 0, 0, 0, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 170, 170, 12, 0, 0, 0, 248, 196, 0, 0, 0, 0, 0, 0, 96, 115, 255, 255, 0, 0, 0, 0, 96, 115, 255, 255, 16, 14, 0, 0, 120, 115, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 112, 129, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 59, 31, 0, 224, 63, 31, 0, 224, 144, 31, 0, 224, 151, 31, 0, 16, 173, 31, 0, 224, 51, 0, 0, 0, 48, 116, 170, 150, 224, 73, 15, 184, 160, 64, 253, 184, 48, 52, 241, 185, 32, 116, 222, 186, 48, 174, 56, 218, 48, 250, 235, 218, 176, 225, 25, 220, 32, 89, 185, 220, 48, 21, 251, 221, 32, 222, 155, 222, 48, 154, 221, 223, 32, 51, 84, 224, 176, 255, 151, 244, 32, 94, 5, 245, 48, 100, 192, 246, 160, 30, 14, 247, 48, 44, 81, 248, 32, 197, 199, 248, 176, 210, 10, 250, 160, 248, 168, 250, 48, 6, 236, 251, 160, 125, 139, 252, 48, 142, 201, 29, 160, 215, 120, 30, 176, 53, 160, 31, 160, 207, 51, 32, 48, 105, 129, 33, 160, 200, 11, 34, 176, 16, 88, 35, 32, 112, 226, 35, 176, 242, 55, 37, 32, 199, 212, 37, 48, 121, 128, 48, 160, 77, 29, 49, 176, 32, 87, 50, 32, 106, 6, 51, 48, 84, 56, 52, 32, 193, 248, 52, 48, 31, 32, 54, 160, 104, 207, 54, 176, 198, 246, 55, 32, 133, 184, 56, 48, 227, 223, 57, 160, 44, 143, 58, 176, 255, 200, 59, 160, 14, 111, 60, 48, 145, 196, 61, 160, 240, 78, 62, 48, 101, 131, 80, 160, 57, 32, 81, 51, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 208, 210, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 199, 31, 0, 224, 251, 31, 0, 16, 9, 32, 0, 224, 3, 0, 0, 0, 2, 0, 0, 0, 64, 0, 0, 0, 90, 0, 0, 0, 60, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 48, 15, 33, 39, 160, 88, 208, 39, 48, 241, 0, 41, 160, 58, 176, 41, 48, 211, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 176, 9, 119, 71, 32, 127, 220, 71, 176, 162, 250, 72, 32, 97, 188, 73, 2, 0, 0, 0, 255, 255, 255, 255, 76, 168, 156, 114, 170, 170, 170, 170, 170, 170, 170, 170, 61, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 4, 5, 4, 5, 4, 170, 170, 170, 12, 0, 0, 0, 52, 201, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 21, 32, 0, 224, 25, 32, 0, 224, 86, 32, 0, 224, 91, 32, 0, 16, 108, 32, 0, 224, 3, 0, 0, 0, 65, 0, 0, 0, 66, 0, 0, 0, 95, 0, 0, 0, 60, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 48, 15, 33, 39, 160, 88, 208, 39, 64, 255, 0, 41, 160, 58, 176, 41, 48, 211, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 48, 241, 187, 64, 192, 11, 213, 64, 176, 9, 119, 71, 32, 127, 220, 71, 2, 0, 0, 0, 255, 255, 255, 255, 44, 175, 156, 114, 170, 170, 170, 170, 170, 170, 170, 170, 61, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 2, 5, 4, 5, 4, 3, 4, 2, 4, 5, 4, 170, 170, 170, 12, 0, 0, 0, 84, 194, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 129, 32, 0, 224, 133, 32, 0, 224, 194, 32, 0, 224, 199, 32, 0, 16, 216, 32, 0, 224, 3, 0, 0, 0, 67, 0, 0, 0, 101, 0, 0, 0, 197, 0, 0, 0, 60, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 48, 15, 33, 39, 160, 88, 208, 39, 64, 255, 0, 41, 160, 58, 176, 41, 48, 211, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 176, 9, 119, 71, 32, 127, 220, 71, 176, 162, 250, 72, 32, 97, 188, 73, 170, 170, 170, 170, 60, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 1, 4, 3, 4, 3, 2, 3, 4, 3, 4, 3, 10, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 237, 32, 0, 224, 241, 32, 0, 224, 47, 33, 0, 16, 63, 33, 0, 224, 2, 0, 0, 0, 68, 0, 0, 0, 143, 0, 0, 0, 58, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 192, 87, 42, 39, 176, 219, 226, 39, 64, 138, 238, 40, 160, 58, 176, 41, 48, 211, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 176, 9, 119, 71, 32, 127, 220, 71, 2, 0, 0, 0, 255, 255, 255, 255, 184, 174, 156, 114, 170, 170, 170, 170, 59, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 2, 3, 2, 5, 4, 5, 4, 3, 4, 5, 4, 170, 12, 0, 0, 0, 200, 194, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 81, 33, 0, 224, 84, 33, 0, 224, 143, 33, 0, 224, 147, 33, 0, 16, 163, 33, 0, 224, 61, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 48, 15, 33, 39, 160, 181, 205, 39, 64, 38, 38, 40, 48, 241, 0, 41, 160, 58, 176, 41, 48, 211, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 48, 241, 187, 64, 192, 11, 213, 64, 176, 9, 119, 71, 32, 127, 220, 71, 2, 0, 0, 0, 255, 255, 255, 255, 44, 176, 156, 114, 170, 170, 170, 170, 170, 170, 170, 170, 62, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 2, 4, 5, 4, 5, 4, 3, 4, 2, 4, 5, 4, 170, 170, 12, 0, 0, 0, 84, 193, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 184, 33, 0, 224, 246, 33, 0, 224, 251, 33, 0, 16, 12, 34, 0, 224, 2, 0, 0, 0, 70, 0, 0, 0, 161, 0, 0, 0, 60, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 64, 52, 25, 39, 176, 195, 205, 39, 192, 103, 250, 40, 176, 72, 176, 41, 64, 225, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 176, 19, 176, 64, 192, 62, 86, 65, 176, 9, 119, 71, 32, 127, 220, 71, 2, 0, 0, 0, 255, 255, 255, 255, 4, 178, 156, 114, 61, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 2, 3, 2, 3, 2, 5, 4, 3, 4, 2, 4, 5, 4, 170, 170, 170, 12, 0, 0, 0, 124, 191, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 32, 34, 0, 224, 35, 34, 0, 224, 96, 34, 0, 224, 99, 34, 0, 16, 116, 34, 0, 224, 60, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 48, 15, 33, 39, 160, 88, 208, 39, 48, 241, 0, 41, 160, 58, 176, 41, 48, 211, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 48, 241, 187, 64, 192, 11, 213, 64, 176, 9, 119, 71, 32, 127, 220, 71, 2, 0, 0, 0, 255, 255, 255, 255, 100, 178, 156, 114, 170, 170, 170, 170, 170, 170, 170, 170, 61, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 4, 2, 4, 5, 4, 170, 170, 170, 12, 0, 0, 0, 28, 191, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 137, 34, 0, 224, 198, 34, 0, 224, 203, 34, 0, 16, 220, 34, 0, 224, 58, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 48, 15, 33, 39, 160, 88, 208, 39, 64, 255, 0, 41, 160, 58, 176, 41, 48, 211, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 176, 9, 119, 71, 32, 127, 220, 71, 2, 0, 0, 0, 255, 255, 255, 255, 212, 174, 156, 114, 170, 170, 170, 170, 59, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 2, 5, 4, 5, 4, 3, 4, 5, 4, 170, 12, 0, 0, 0, 172, 194, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 240, 34, 0, 224, 43, 35, 0, 224, 47, 35, 0, 16, 63, 35, 0, 224, 61, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 48, 15, 33, 39, 160, 181, 205, 39, 64, 38, 38, 40, 48, 241, 0, 41, 160, 58, 176, 41, 48, 211, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 176, 159, 186, 64, 64, 48, 3, 65, 176, 9, 119, 71, 32, 127, 220, 71, 2, 0, 0, 0, 255, 255, 255, 255, 188, 177, 156, 114, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 62, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 2, 4, 5, 4, 5, 4, 3, 4, 2, 4, 5, 4, 170, 170, 12, 0, 0, 0, 196, 191, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 83, 35, 0, 224, 145, 35, 0, 224, 151, 35, 0, 16, 168, 35, 0, 224, 61, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 165, 253, 37, 64, 52, 25, 39, 176, 195, 205, 39, 192, 27, 71, 40, 176, 198, 246, 55, 176, 42, 191, 56, 176, 159, 186, 64, 64, 48, 3, 65, 176, 9, 119, 71, 160, 252, 147, 71, 176, 82, 211, 71, 64, 118, 241, 72, 176, 52, 179, 73, 64, 88, 209, 74, 2, 0, 0, 0, 255, 255, 255, 255, 180, 175, 156, 114, 170, 170, 170, 170, 170, 170, 170, 170, 62, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 2, 3, 2, 4, 3, 4, 2, 4, 5, 3, 2, 3, 2, 4, 170, 170, 12, 0, 0, 0, 204, 193, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 188, 35, 0, 224, 250, 35, 0, 224, 255, 35, 0, 16, 16, 36, 0, 224, 62, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 48, 15, 33, 39, 160, 88, 208, 39, 64, 255, 0, 41, 160, 58, 176, 41, 48, 211, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 48, 241, 187, 64, 64, 209, 203, 64, 176, 9, 119, 71, 32, 127, 220, 71, 176, 162, 250, 72, 32, 97, 188, 73, 2, 0, 0, 0, 255, 255, 255, 255, 164, 174, 156, 114, 170, 170, 170, 170, 63, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 2, 5, 4, 5, 4, 3, 4, 2, 4, 5, 4, 5, 4, 170, 12, 0, 0, 0, 220, 194, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 36, 36, 0, 224, 99, 36, 0, 224, 103, 36, 0, 16, 120, 36, 0, 224, 60, 0, 0, 0, 48, 143, 146, 162, 64, 82, 123, 182, 176, 201, 26, 183, 64, 143, 30, 184, 48, 112, 212, 184, 192, 125, 23, 186, 176, 163, 181, 186, 64, 177, 248, 187, 48, 215, 150, 188, 192, 228, 217, 189, 176, 10, 120, 190, 64, 24, 187, 191, 176, 143, 90, 192, 64, 157, 157, 193, 48, 195, 59, 194, 192, 208, 126, 195, 176, 246, 28, 196, 64, 4, 96, 197, 48, 42, 254, 197, 192, 55, 65, 199, 48, 175, 224, 199, 64, 148, 129, 200, 176, 161, 77, 202, 192, 134, 238, 202, 48, 255, 77, 206, 192, 237, 176, 206, 176, 53, 41, 211, 192, 100, 67, 212, 48, 8, 61, 244, 192, 246, 159, 244, 48, 108, 5, 245, 64, 16, 50, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 181, 148, 35, 160, 148, 16, 36, 176, 242, 55, 37, 160, 118, 240, 37, 48, 15, 33, 39, 160, 88, 208, 39, 48, 241, 0, 41, 160, 58, 176, 41, 48, 211, 224, 42, 32, 87, 153, 43, 176, 198, 246, 55, 176, 42, 191, 56, 48, 78, 185, 64, 192, 11, 213, 64, 176, 9, 119, 71, 32, 127, 220, 71, 2, 0, 0, 0, 255, 255, 255, 255, 136, 177, 156, 114, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 61, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 4, 2, 4, 5, 4, 170, 170, 170, 12, 0, 0, 0, 248, 191, 255, 255, 0, 0, 0, 0, 208, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 140, 36, 0, 224, 201, 36, 0, 224, 207, 36, 0, 16, 224, 36, 0, 224, 80, 0, 0, 0, 144, 245, 23, 184, 64, 218, 43, 5, 176, 240, 252, 7, 192, 116, 207, 10, 176, 202, 151, 11, 192, 249, 177, 12, 48, 254, 120, 13, 64, 45, 147, 14, 176, 49, 90, 15, 192, 96, 116, 16, 176, 67, 100, 17, 64, 148, 85, 18, 176, 200, 70, 19, 64, 25, 56, 20, 48, 252, 39, 21, 192, 76, 25, 22, 176, 47, 9, 23, 64, 128, 250, 23, 48, 99, 234, 24, 192, 179, 219, 25, 48, 232, 204, 26, 192, 56, 190, 27, 176, 27, 174, 28, 64, 108, 159, 29, 48, 79, 143, 30, 192, 159, 128, 31, 176, 130, 112, 32, 64, 211, 97, 33, 176, 7, 83, 34, 64, 88, 68, 35, 48, 59, 52, 36, 64, 59, 65, 37, 176, 110, 21, 38, 64, 191, 6, 39, 48, 162, 246, 39, 64, 138, 238, 40, 176, 72, 176, 41, 192, 189, 207, 42, 48, 9, 185, 43, 64, 171, 171, 44, 176, 12, 112, 45, 192, 222, 140, 46, 176, 238, 79, 47, 64, 18, 110, 48, 48, 104, 54, 49, 192, 46, 87, 50, 176, 178, 15, 51, 192, 16, 55, 52, 48, 207, 248, 52, 192, 242, 22, 54, 176, 235, 225, 54, 192, 212, 246, 55, 176, 205, 193, 56, 192, 182, 214, 57, 176, 175, 161, 58, 64, 211, 191, 59, 48, 182, 175, 60, 192, 144, 113, 61, 48, 152, 143, 62, 64, 173, 90, 63, 48, 122, 111, 64, 64, 238, 113, 65, 176, 172, 51, 66, 64, 208, 81, 67, 176, 142, 19, 68, 64, 178, 49, 69, 176, 112, 243, 69, 192, 206, 26, 71, 176, 82, 211, 71, 192, 176, 250, 72, 176, 52, 179, 73, 192, 146, 218, 74, 48, 59, 193, 75, 192, 255, 167, 76, 48, 29, 161, 77, 192, 225, 135, 78, 48, 255, 128, 79, 64, 254, 112, 80, 48, 108, 78, 81, 64, 224, 80, 82, 170, 170, 170, 170, 170, 170, 170, 170, 80, 0, 0, 0, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 8, 0, 0, 0, 240, 201, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 192, 199, 255, 127, 106, 3, 0, 96, 222, 7, 0, 112, 244, 36, 0, 224, 71, 37, 0, 16, 92, 37, 0, 224, 2, 0, 0, 0, 79, 0, 0, 0, 100, 0, 0, 0, 4, 0, 0, 0, 128, 161, 184, 158, 112, 249, 186, 159, 96, 87, 248, 200, 240, 9, 97, 210, 2, 0, 0, 0, 255, 255, 255, 255, 100, 132, 238, 114, 170, 170, 170, 170, 5, 0, 0, 0, 1, 2, 1, 2, 3, 170, 170, 170, 8, 0, 0, 0, 28, 170, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 111, 37, 0, 224, 114, 37, 0, 224, 119, 37, 0, 224, 123, 37, 0, 16, 126, 37, 0, 224, 61, 0, 0, 0, 28, 107, 170, 150, 224, 73, 15, 184, 160, 64, 253, 184, 48, 52, 241, 185, 32, 116, 222, 186, 48, 174, 56, 218, 48, 250, 235, 218, 176, 225, 25, 220, 32, 89, 185, 220, 48, 21, 251, 221, 32, 222, 155, 222, 48, 154, 221, 223, 32, 51, 84, 224, 176, 255, 151, 244, 32, 94, 5, 245, 48, 100, 192, 246, 160, 30, 14, 247, 48, 44, 81, 248, 32, 197, 199, 248, 176, 210, 10, 250, 160, 248, 168, 250, 48, 6, 236, 251, 160, 125, 139, 252, 48, 142, 201, 29, 160, 215, 120, 30, 176, 53, 160, 31, 160, 207, 51, 32, 48, 105, 129, 33, 160, 200, 11, 34, 176, 16, 88, 35, 32, 112, 226, 35, 176, 242, 55, 37, 32, 199, 212, 37, 48, 15, 33, 39, 160, 227, 189, 39, 48, 241, 0, 41, 32, 139, 148, 41, 176, 13, 234, 42, 160, 50, 107, 43, 48, 181, 192, 44, 32, 196, 102, 45, 48, 151, 160, 46, 32, 166, 70, 47, 48, 121, 128, 48, 160, 77, 29, 49, 176, 32, 87, 50, 32, 106, 6, 51, 48, 84, 56, 52, 32, 193, 248, 52, 48, 31, 32, 54, 160, 104, 207, 54, 176, 198, 246, 55, 32, 133, 184, 56, 48, 227, 223, 57, 160, 44, 143, 58, 176, 255, 200, 59, 160, 14, 111, 60, 48, 145, 196, 61, 160, 240, 78, 62, 176, 72, 154, 78, 32, 146, 73, 79, 170, 170, 170, 170, 170, 170, 170, 170, 61, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 228, 219, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 143, 37, 0, 224, 207, 37, 0, 16, 224, 37, 0, 224, 41, 0, 0, 0, 112, 232, 182, 165, 224, 110, 242, 175, 96, 86, 102, 182, 96, 210, 67, 183, 96, 54, 12, 184, 240, 134, 253, 184, 96, 113, 234, 203, 240, 180, 145, 216, 128, 112, 0, 0, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 144, 18, 245, 58, 0, 209, 182, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 144, 206, 111, 64, 128, 155, 132, 65, 144, 176, 79, 66, 128, 125, 100, 67, 144, 146, 47, 68, 128, 95, 68, 69, 144, 116, 15, 70, 128, 65, 36, 71, 16, 145, 248, 71, 128, 35, 4, 73, 16, 115, 216, 73, 128, 5, 228, 74, 16, 85, 184, 75, 240, 19, 205, 76, 0, 41, 152, 77, 240, 245, 172, 78, 170, 170, 170, 170, 41, 0, 0, 0, 2, 4, 2, 4, 2, 4, 2, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 5, 4, 5, 4, 170, 170, 170, 12, 0, 0, 0, 84, 157, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 160, 171, 255, 127, 176, 3, 0, 96, 220, 7, 0, 112, 236, 37, 0, 224, 23, 38, 0, 16, 35, 38, 0, 224, 9, 0, 0, 0, 229, 99, 133, 184, 224, 242, 0, 14, 208, 140, 148, 14, 224, 0, 151, 15, 208, 110, 116, 16, 224, 226, 118, 17, 208, 80, 84, 18, 96, 255, 95, 19, 80, 62, 48, 20, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 9, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 27, 200, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 58, 38, 0, 224, 71, 38, 0, 16, 75, 38, 0, 224, 29, 0, 0, 0, 116, 116, 170, 150, 224, 73, 15, 184, 160, 64, 253, 184, 48, 52, 241, 185, 32, 116, 222, 186, 48, 174, 56, 218, 48, 250, 235, 218, 176, 225, 25, 220, 32, 89, 185, 220, 48, 21, 251, 221, 32, 222, 155, 222, 48, 154, 221, 223, 32, 51, 84, 224, 176, 255, 151, 244, 32, 94, 5, 245, 48, 100, 192, 246, 160, 30, 14, 247, 48, 44, 81, 248, 32, 197, 199, 248, 176, 210, 10, 250, 160, 248, 168, 250, 48, 6, 236, 251, 160, 125, 139, 252, 48, 142, 201, 29, 160, 215, 120, 30, 176, 53, 160, 31, 160, 207, 51, 32, 48, 105, 129, 33, 160, 200, 11, 34, 170, 170, 170, 170, 170, 170, 170, 170, 29, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 140, 210, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 87, 38, 0, 224, 119, 38, 0, 16, 128, 38, 0, 224, 55, 0, 0, 0, 176, 217, 94, 147, 224, 59, 159, 159, 216, 81, 69, 160, 224, 29, 127, 161, 88, 110, 46, 162, 224, 255, 94, 163, 88, 80, 14, 164, 224, 225, 62, 165, 88, 50, 238, 165, 96, 254, 39, 167, 88, 20, 206, 167, 96, 224, 7, 169, 88, 246, 173, 169, 96, 194, 231, 170, 216, 18, 151, 171, 96, 164, 199, 172, 216, 244, 118, 173, 96, 134, 167, 174, 216, 214, 86, 175, 96, 104, 135, 176, 216, 184, 54, 177, 224, 132, 112, 178, 216, 154, 22, 179, 224, 102, 80, 180, 216, 124, 246, 180, 224, 72, 48, 182, 88, 153, 223, 182, 224, 42, 16, 184, 88, 123, 191, 184, 224, 12, 240, 185, 88, 93, 159, 186, 96, 41, 217, 187, 88, 63, 127, 188, 96, 11, 185, 189, 88, 33, 95, 190, 96, 237, 152, 191, 88, 3, 63, 192, 96, 207, 120, 193, 216, 31, 40, 194, 96, 177, 88, 195, 216, 1, 8, 196, 96, 147, 56, 197, 216, 227, 231, 197, 224, 175, 33, 199, 216, 197, 199, 199, 224, 145, 1, 201, 216, 167, 167, 201, 224, 115, 225, 202, 88, 196, 144, 203, 224, 85, 193, 204, 88, 166, 112, 205, 96, 219, 98, 7, 80, 208, 185, 7, 96, 113, 97, 24, 80, 55, 171, 24, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 55, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 3, 1, 170, 8, 0, 0, 0, 80, 173, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 8, 7, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 140, 38, 0, 224, 199, 38, 0, 16, 214, 38, 0, 224, 4, 0, 0, 0, 96, 133, 184, 158, 80, 221, 186, 159, 96, 226, 136, 203, 208, 237, 96, 210, 2, 0, 0, 0, 255, 255, 255, 255, 12, 57, 61, 94, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 5, 0, 0, 0, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 116, 202, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 228, 38, 0, 224, 233, 38, 0, 224, 239, 38, 0, 16, 242, 38, 0, 224, 33, 0, 0, 0, 224, 127, 170, 150, 240, 87, 15, 184, 176, 78, 253, 184, 64, 66, 241, 185, 48, 130, 222, 186, 64, 188, 56, 218, 64, 8, 236, 218, 192, 239, 25, 220, 48, 103, 185, 220, 64, 35, 251, 221, 48, 236, 155, 222, 64, 168, 221, 223, 48, 65, 84, 224, 192, 13, 152, 244, 48, 108, 5, 245, 64, 114, 192, 246, 176, 44, 14, 247, 64, 58, 81, 248, 48, 211, 199, 248, 192, 224, 10, 250, 176, 6, 169, 250, 64, 20, 236, 251, 176, 139, 139, 252, 64, 156, 201, 29, 176, 229, 120, 30, 192, 67, 160, 31, 176, 221, 51, 32, 64, 119, 129, 33, 176, 214, 11, 34, 192, 212, 246, 55, 48, 147, 184, 56, 64, 241, 223, 57, 176, 29, 233, 57, 170, 170, 170, 170, 33, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 32, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 0, 39, 0, 224, 35, 39, 0, 16, 45, 39, 0, 224, 3, 0, 0, 0, 112, 85, 88, 152, 80, 115, 3, 42, 64, 93, 190, 43, 170, 170, 170, 170, 170, 170, 170, 170, 3, 0, 0, 0, 1, 2, 1, 170, 6, 0, 0, 0, 144, 186, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 57, 39, 0, 224, 63, 39, 0, 16, 65, 39, 0, 224, 89, 0, 0, 0, 160, 72, 166, 158, 144, 21, 187, 159, 160, 42, 134, 160, 144, 247, 154, 161, 32, 76, 70, 168, 144, 12, 137, 203, 0, 24, 97, 210, 16, 117, 248, 250, 0, 88, 232, 251, 16, 87, 216, 252, 0, 58, 200, 253, 16, 57, 184, 254, 0, 28, 168, 255, 16, 27, 152, 0, 0, 254, 135, 1, 16, 253, 119, 2, 128, 26, 113, 3, 144, 25, 97, 4, 128, 252, 80, 5, 144, 251, 64, 6, 128, 222, 48, 7, 144, 31, 178, 7, 128, 192, 16, 9, 16, 177, 173, 9, 128, 162, 240, 10, 144, 161, 224, 11, 0, 191, 217, 12, 144, 131, 192, 13, 0, 161, 185, 14, 16, 160, 169, 15, 0, 131, 153, 16, 16, 130, 137, 17, 0, 101, 121, 18, 16, 100, 105, 19, 0, 71, 89, 20, 16, 70, 73, 21, 0, 41, 57, 22, 16, 40, 41, 23, 128, 69, 34, 24, 16, 10, 9, 25, 128, 39, 2, 26, 144, 38, 242, 26, 128, 9, 226, 27, 144, 8, 210, 28, 128, 235, 193, 29, 144, 234, 177, 30, 128, 205, 161, 31, 16, 29, 118, 32, 128, 175, 129, 33, 16, 255, 85, 34, 0, 204, 106, 35, 16, 225, 53, 36, 0, 174, 74, 37, 16, 195, 21, 38, 0, 144, 42, 39, 144, 223, 254, 39, 0, 114, 10, 41, 144, 193, 222, 41, 0, 84, 234, 42, 144, 163, 190, 43, 128, 112, 211, 44, 144, 133, 158, 45, 128, 82, 179, 46, 144, 103, 126, 47, 128, 52, 147, 48, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 16, 238, 198, 58, 0, 187, 219, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 144, 206, 111, 64, 128, 155, 132, 65, 144, 176, 79, 66, 128, 125, 100, 67, 144, 146, 47, 68, 128, 95, 68, 69, 16, 197, 243, 69, 0, 124, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 192, 26, 4, 94, 170, 170, 170, 170, 90, 0, 0, 0, 1, 2, 1, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 10, 0, 0, 0, 15, 147, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 144, 157, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 77, 39, 0, 224, 167, 39, 0, 224, 171, 39, 0, 16, 195, 39, 0, 224, 62, 0, 0, 0, 128, 205, 242, 161, 144, 12, 137, 203, 0, 24, 97, 210, 112, 90, 47, 247, 240, 133, 40, 248, 16, 100, 105, 19, 0, 71, 89, 20, 16, 70, 73, 21, 0, 41, 57, 22, 16, 40, 41, 23, 128, 69, 34, 24, 16, 10, 9, 25, 128, 39, 2, 26, 144, 38, 242, 26, 128, 9, 226, 27, 144, 8, 210, 28, 128, 235, 193, 29, 144, 234, 177, 30, 128, 205, 161, 31, 16, 29, 118, 32, 128, 175, 129, 33, 16, 255, 85, 34, 0, 204, 106, 35, 16, 225, 53, 36, 0, 174, 74, 37, 16, 195, 21, 38, 0, 144, 42, 39, 144, 223, 254, 39, 0, 114, 10, 41, 144, 193, 222, 41, 0, 84, 234, 42, 144, 163, 190, 43, 128, 112, 211, 44, 144, 133, 158, 45, 128, 82, 179, 46, 144, 103, 126, 47, 128, 52, 147, 48, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 80, 233, 4, 58, 16, 238, 198, 58, 0, 187, 219, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 144, 206, 111, 64, 128, 155, 132, 65, 144, 176, 79, 66, 128, 125, 100, 67, 144, 146, 47, 68, 128, 95, 68, 69, 16, 197, 243, 69, 0, 124, 45, 71, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 62, 0, 0, 0, 1, 2, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 5, 6, 4, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 144, 157, 255, 255, 32, 28, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 144, 157, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 217, 39, 0, 224, 27, 40, 0, 16, 44, 40, 0, 224, 91, 0, 0, 0, 52, 122, 170, 150, 240, 87, 15, 184, 176, 78, 253, 184, 64, 66, 241, 185, 48, 130, 222, 186, 64, 188, 56, 218, 64, 8, 236, 218, 192, 239, 25, 220, 48, 103, 185, 220, 64, 35, 251, 221, 48, 236, 155, 222, 64, 168, 221, 223, 48, 65, 84, 224, 192, 13, 152, 244, 48, 108, 5, 245, 64, 114, 192, 246, 176, 44, 14, 247, 64, 58, 81, 248, 48, 211, 199, 248, 192, 224, 10, 250, 176, 6, 169, 250, 64, 20, 236, 251, 176, 139, 139, 252, 64, 156, 201, 29, 176, 229, 120, 30, 192, 67, 160, 31, 176, 221, 51, 32, 64, 119, 129, 33, 176, 214, 11, 34, 192, 30, 88, 35, 48, 126, 226, 35, 192, 0, 56, 37, 48, 213, 212, 37, 64, 29, 33, 39, 176, 241, 189, 39, 64, 255, 0, 41, 48, 153, 148, 41, 192, 27, 234, 42, 176, 64, 107, 43, 64, 195, 192, 44, 48, 210, 102, 45, 64, 165, 160, 46, 48, 180, 70, 47, 64, 135, 128, 48, 176, 91, 29, 49, 192, 46, 87, 50, 48, 120, 6, 51, 64, 98, 56, 52, 48, 207, 248, 52, 64, 45, 32, 54, 176, 118, 207, 54, 192, 212, 246, 55, 48, 147, 184, 56, 64, 241, 223, 57, 176, 58, 143, 58, 192, 13, 201, 59, 176, 28, 111, 60, 64, 159, 196, 61, 176, 254, 78, 62, 64, 12, 146, 63, 176, 224, 46, 64, 64, 6, 135, 65, 48, 253, 23, 66, 64, 208, 81, 67, 48, 223, 247, 67, 192, 97, 77, 69, 176, 251, 224, 69, 64, 148, 17, 71, 48, 163, 183, 71, 192, 176, 250, 72, 48, 133, 151, 73, 192, 146, 218, 74, 176, 161, 128, 75, 192, 116, 186, 76, 176, 131, 96, 77, 192, 86, 154, 78, 48, 160, 73, 79, 64, 115, 131, 80, 176, 71, 32, 81, 64, 85, 99, 82, 176, 41, 0, 83, 64, 55, 67, 84, 48, 70, 233, 84, 64, 25, 35, 86, 48, 40, 201, 86, 64, 251, 2, 88, 48, 10, 169, 88, 64, 221, 226, 89, 48, 236, 136, 90, 192, 110, 222, 91, 48, 206, 104, 92, 170, 170, 170, 170, 170, 170, 170, 170, 91, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 204, 204, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 69, 40, 0, 224, 163, 40, 0, 16, 187, 40, 0, 224, 42, 0, 0, 0, 96, 218, 182, 165, 96, 213, 134, 22, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 0, 196, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 128, 4, 245, 58, 240, 194, 182, 59, 128, 252, 175, 60, 240, 142, 187, 61, 128, 222, 143, 62, 240, 112, 155, 63, 128, 192, 111, 64, 112, 141, 132, 65, 128, 162, 79, 66, 112, 111, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 128, 102, 15, 70, 112, 51, 36, 71, 0, 131, 248, 71, 112, 21, 4, 73, 0, 101, 216, 73, 112, 247, 227, 74, 0, 71, 184, 75, 240, 19, 205, 76, 0, 41, 152, 77, 240, 245, 172, 78, 0, 11, 120, 79, 240, 215, 140, 80, 128, 39, 97, 81, 240, 185, 108, 82, 128, 9, 65, 83, 240, 155, 76, 84, 0, 221, 205, 84, 170, 170, 170, 170, 42, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 170, 170, 10, 0, 0, 0, 168, 174, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 199, 40, 0, 224, 243, 40, 0, 16, 255, 40, 0, 224, 4, 0, 0, 0, 60, 44, 30, 147, 72, 236, 152, 246, 112, 146, 91, 71, 112, 169, 37, 87, 2, 0, 0, 0, 255, 255, 255, 255, 64, 26, 135, 105, 5, 0, 0, 0, 2, 1, 3, 1, 3, 170, 170, 170, 8, 0, 0, 0, 64, 193, 255, 255, 0, 0, 0, 0, 184, 192, 255, 255, 0, 0, 0, 0, 68, 193, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 15, 41, 0, 224, 20, 41, 0, 224, 23, 41, 0, 16, 26, 41, 0, 224, 2, 0, 0, 0, 144, 43, 244, 145, 192, 53, 195, 251, 170, 170, 170, 170, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 240, 206, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 42, 41, 0, 224, 47, 41, 0, 16, 49, 41, 0, 224, 3, 0, 0, 0, 98, 0, 0, 0, 125, 1, 0, 0, 106, 2, 0, 0, 174, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 0, 116, 203, 162, 240, 247, 131, 163, 128, 210, 69, 164, 240, 217, 99, 165, 0, 217, 83, 166, 112, 151, 21, 167, 0, 187, 51, 168, 240, 179, 254, 168, 0, 157, 19, 170, 240, 149, 222, 170, 0, 127, 243, 171, 240, 119, 190, 172, 0, 97, 211, 173, 240, 89, 158, 174, 0, 67, 179, 175, 240, 59, 126, 176, 128, 95, 156, 177, 112, 88, 103, 178, 128, 65, 124, 179, 112, 58, 71, 180, 128, 35, 92, 181, 112, 28, 39, 182, 128, 5, 60, 183, 112, 254, 6, 184, 128, 231, 27, 185, 112, 224, 230, 185, 0, 4, 5, 187, 112, 194, 198, 187, 0, 230, 228, 188, 240, 222, 175, 189, 0, 200, 196, 190, 240, 192, 143, 191, 0, 214, 90, 192, 112, 60, 176, 193, 0, 140, 132, 194, 240, 132, 79, 195, 0, 110, 100, 196, 240, 102, 47, 197, 128, 138, 77, 198, 240, 72, 15, 199, 128, 108, 45, 200, 112, 101, 248, 200, 128, 78, 13, 202, 112, 71, 216, 202, 128, 254, 136, 203, 240, 9, 97, 210, 0, 243, 117, 211, 240, 235, 64, 212, 0, 213, 85, 213, 240, 205, 32, 214, 0, 183, 53, 215, 240, 175, 0, 216, 0, 153, 21, 217, 240, 145, 224, 217, 128, 181, 254, 218, 240, 115, 192, 219, 128, 151, 222, 220, 112, 144, 169, 221, 128, 121, 190, 222, 112, 114, 137, 223, 128, 91, 158, 224, 112, 84, 105, 225, 128, 61, 126, 226, 112, 54, 73, 227, 128, 31, 94, 228, 240, 60, 87, 229, 0, 60, 71, 230, 240, 30, 55, 231, 0, 30, 39, 232, 240, 0, 23, 233, 0, 0, 7, 234, 240, 226, 246, 234, 0, 226, 230, 235, 240, 196, 214, 236, 0, 196, 198, 237, 112, 225, 191, 238, 128, 224, 175, 239, 112, 195, 159, 240, 128, 194, 143, 241, 112, 165, 127, 242, 128, 164, 111, 243, 112, 135, 95, 244, 128, 134, 79, 245, 112, 105, 63, 246, 128, 104, 47, 247, 240, 133, 40, 248, 128, 74, 15, 249, 240, 103, 8, 250, 0, 103, 248, 250, 240, 73, 232, 251, 0, 73, 216, 252, 240, 43, 200, 253, 0, 43, 184, 254, 240, 13, 168, 255, 0, 13, 152, 0, 240, 239, 135, 1, 0, 239, 119, 2, 112, 12, 113, 3, 128, 11, 97, 4, 112, 238, 80, 5, 128, 237, 64, 6, 112, 208, 48, 7, 128, 39, 141, 7, 112, 178, 16, 9, 0, 163, 173, 9, 112, 148, 240, 10, 128, 147, 224, 11, 240, 176, 217, 12, 128, 117, 192, 13, 240, 146, 185, 14, 0, 146, 169, 15, 240, 116, 153, 16, 0, 116, 137, 17, 240, 86, 121, 18, 0, 86, 105, 19, 240, 56, 89, 20, 0, 56, 73, 21, 240, 26, 57, 22, 0, 26, 41, 23, 112, 55, 34, 24, 0, 252, 8, 25, 112, 25, 2, 26, 128, 24, 242, 26, 112, 251, 225, 27, 128, 250, 209, 28, 112, 221, 193, 29, 128, 220, 177, 30, 112, 191, 161, 31, 0, 15, 118, 32, 112, 161, 129, 33, 0, 241, 85, 34, 240, 189, 106, 35, 0, 211, 53, 36, 240, 159, 74, 37, 0, 181, 21, 38, 240, 129, 42, 39, 128, 209, 254, 39, 240, 99, 10, 41, 128, 179, 222, 41, 240, 69, 234, 42, 128, 149, 190, 43, 112, 98, 211, 44, 128, 119, 158, 45, 112, 68, 179, 46, 128, 89, 126, 47, 112, 38, 147, 48, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 0, 224, 198, 58, 240, 172, 219, 59, 128, 252, 175, 60, 240, 142, 187, 61, 128, 222, 143, 62, 240, 112, 155, 63, 128, 192, 111, 64, 112, 141, 132, 65, 128, 162, 79, 66, 112, 111, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 175, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 8, 0, 0, 0, 212, 173, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 160, 171, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 61, 41, 0, 224, 65, 41, 0, 224, 240, 41, 0, 224, 243, 41, 0, 16, 32, 42, 0, 224, 20, 0, 0, 0, 112, 232, 182, 165, 224, 110, 242, 175, 96, 86, 102, 182, 96, 210, 67, 183, 96, 54, 12, 184, 240, 134, 253, 184, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 144, 18, 245, 58, 0, 209, 182, 59, 144, 10, 176, 60, 0, 157, 187, 61, 20, 0, 0, 0, 1, 3, 1, 3, 1, 3, 4, 3, 4, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 10, 0, 0, 0, 140, 156, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 144, 157, 255, 127, 176, 3, 0, 96, 211, 7, 0, 112, 54, 42, 0, 224, 75, 42, 0, 16, 81, 42, 0, 224, 9, 0, 0, 0, 77, 22, 232, 163, 96, 73, 54, 17, 80, 110, 183, 17, 96, 43, 22, 19, 80, 80, 151, 19, 96, 224, 151, 39, 208, 182, 110, 40, 96, 194, 119, 41, 208, 217, 194, 41, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 9, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 51, 177, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 102, 42, 0, 224, 115, 42, 0, 16, 119, 42, 0, 224, 2, 0, 0, 0, 112, 75, 214, 155, 0, 59, 249, 158, 2, 0, 0, 0, 255, 255, 255, 255, 188, 112, 61, 94, 170, 170, 170, 170, 170, 170, 170, 170, 3, 0, 0, 0, 2, 1, 2, 170, 6, 0, 0, 0, 196, 146, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 131, 42, 0, 224, 134, 42, 0, 224, 139, 42, 0, 16, 141, 42, 0, 224, 89, 0, 0, 0, 148, 123, 170, 150, 240, 87, 15, 184, 176, 78, 253, 184, 64, 66, 241, 185, 48, 130, 222, 186, 64, 188, 56, 218, 64, 8, 236, 218, 192, 239, 25, 220, 48, 103, 185, 220, 64, 35, 251, 221, 48, 236, 155, 222, 64, 168, 221, 223, 48, 65, 84, 224, 192, 13, 152, 244, 48, 108, 5, 245, 64, 114, 192, 246, 176, 44, 14, 247, 64, 58, 81, 248, 48, 211, 199, 248, 192, 224, 10, 250, 176, 6, 169, 250, 64, 20, 236, 251, 176, 139, 139, 252, 64, 156, 201, 29, 176, 229, 120, 30, 192, 67, 160, 31, 176, 221, 51, 32, 64, 119, 129, 33, 176, 214, 11, 34, 192, 30, 88, 35, 48, 126, 226, 35, 192, 0, 56, 37, 48, 213, 212, 37, 64, 29, 33, 39, 176, 241, 189, 39, 64, 255, 0, 41, 48, 153, 148, 41, 192, 27, 234, 42, 176, 64, 107, 43, 64, 195, 192, 44, 48, 210, 102, 45, 64, 165, 160, 46, 48, 180, 70, 47, 64, 135, 128, 48, 176, 91, 29, 49, 192, 46, 87, 50, 48, 120, 6, 51, 64, 98, 56, 52, 48, 207, 248, 52, 64, 45, 32, 54, 176, 118, 207, 54, 192, 212, 246, 55, 48, 147, 184, 56, 64, 241, 223, 57, 176, 58, 143, 58, 192, 13, 201, 59, 176, 28, 111, 60, 64, 159, 196, 61, 176, 254, 78, 62, 64, 6, 135, 65, 48, 253, 23, 66, 64, 208, 81, 67, 48, 223, 247, 67, 192, 97, 77, 69, 176, 251, 224, 69, 64, 148, 17, 71, 48, 163, 183, 71, 192, 176, 250, 72, 48, 133, 151, 73, 192, 146, 218, 74, 176, 161, 128, 75, 192, 116, 186, 76, 176, 131, 96, 77, 192, 86, 154, 78, 48, 160, 73, 79, 64, 115, 131, 80, 176, 71, 32, 81, 64, 85, 99, 82, 176, 41, 0, 83, 64, 55, 67, 84, 48, 70, 233, 84, 64, 25, 35, 86, 48, 40, 201, 86, 64, 251, 2, 88, 48, 10, 169, 88, 64, 221, 226, 89, 48, 236, 136, 90, 192, 110, 222, 91, 48, 206, 104, 92, 170, 170, 170, 170, 170, 170, 170, 170, 89, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 108, 203, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 155, 42, 0, 224, 247, 42, 0, 16, 15, 43, 0, 224, 4, 0, 0, 0, 77, 0, 0, 0, 105, 0, 0, 0, 148, 0, 0, 0, 153, 0, 0, 0, 2, 0, 0, 0, 35, 46, 30, 147, 72, 236, 152, 246, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 93, 191, 255, 255, 0, 0, 0, 0, 184, 192, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 27, 43, 0, 224, 32, 43, 0, 224, 35, 43, 0, 16, 37, 43, 0, 224, 34, 0, 0, 0, 0, 73, 128, 155, 80, 124, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 48, 78, 231, 48, 170, 170, 170, 170, 34, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 170, 170, 8, 0, 0, 0, 128, 238, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 51, 43, 0, 224, 87, 43, 0, 16, 97, 43, 0, 224, 65, 0, 0, 0, 176, 203, 184, 158, 160, 35, 187, 159, 176, 12, 208, 160, 128, 210, 162, 161, 176, 40, 137, 203, 32, 52, 97, 210, 144, 118, 47, 247, 16, 162, 40, 248, 144, 236, 48, 7, 32, 114, 105, 19, 16, 85, 89, 20, 32, 84, 73, 21, 16, 55, 57, 22, 32, 54, 41, 23, 144, 83, 34, 24, 32, 24, 9, 25, 144, 53, 2, 26, 160, 52, 242, 26, 144, 23, 226, 27, 160, 22, 210, 28, 144, 249, 193, 29, 160, 248, 177, 30, 144, 219, 161, 31, 32, 43, 118, 32, 144, 189, 129, 33, 32, 13, 86, 34, 16, 218, 106, 35, 32, 239, 53, 36, 16, 188, 74, 37, 32, 209, 21, 38, 16, 158, 42, 39, 160, 237, 254, 39, 16, 128, 10, 41, 160, 207, 222, 41, 16, 98, 234, 42, 160, 177, 190, 43, 144, 126, 211, 44, 160, 147, 158, 45, 144, 96, 179, 46, 160, 117, 126, 47, 144, 66, 147, 48, 32, 146, 103, 49, 144, 36, 115, 50, 32, 116, 71, 51, 144, 6, 83, 52, 32, 86, 39, 53, 144, 232, 50, 54, 32, 56, 7, 55, 16, 5, 28, 56, 32, 26, 231, 56, 16, 231, 251, 57, 32, 252, 198, 58, 16, 201, 219, 59, 160, 24, 176, 60, 16, 171, 187, 61, 160, 250, 143, 62, 16, 141, 155, 63, 160, 220, 111, 64, 144, 169, 132, 65, 160, 190, 79, 66, 144, 139, 100, 67, 160, 160, 47, 68, 144, 109, 68, 69, 32, 211, 243, 69, 16, 138, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 180, 142, 134, 125, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 66, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 3, 1, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 170, 170, 12, 0, 0, 0, 76, 125, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 16, 14, 0, 0, 112, 129, 255, 255, 32, 28, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 128, 143, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 111, 43, 0, 224, 177, 43, 0, 224, 183, 43, 0, 16, 201, 43, 0, 224, 56, 0, 0, 0, 160, 189, 184, 158, 144, 21, 187, 159, 160, 26, 137, 203, 16, 38, 97, 210, 32, 241, 85, 213, 16, 234, 32, 214, 32, 211, 53, 215, 16, 204, 0, 216, 32, 181, 21, 217, 16, 174, 224, 217, 160, 209, 254, 218, 16, 144, 192, 219, 160, 179, 222, 220, 144, 172, 169, 221, 160, 149, 190, 222, 144, 142, 137, 223, 160, 119, 158, 224, 144, 112, 105, 225, 160, 89, 126, 226, 144, 82, 73, 227, 160, 59, 94, 228, 144, 52, 41, 229, 32, 88, 71, 230, 16, 81, 18, 231, 32, 58, 39, 232, 16, 51, 242, 232, 32, 28, 7, 234, 16, 21, 210, 234, 32, 254, 230, 235, 16, 247, 177, 236, 32, 224, 198, 237, 16, 217, 145, 238, 160, 252, 175, 239, 16, 187, 113, 240, 160, 222, 143, 241, 144, 193, 127, 242, 160, 192, 111, 243, 144, 163, 95, 244, 160, 162, 79, 245, 144, 133, 63, 246, 160, 132, 47, 247, 16, 162, 40, 248, 160, 102, 15, 249, 16, 132, 8, 250, 32, 131, 248, 250, 16, 102, 232, 251, 32, 101, 216, 252, 16, 72, 200, 253, 32, 71, 184, 254, 16, 42, 168, 255, 32, 41, 152, 0, 16, 12, 136, 1, 32, 11, 120, 2, 144, 40, 113, 3, 160, 39, 97, 4, 144, 240, 1, 5, 2, 0, 0, 0, 255, 255, 255, 255, 56, 116, 61, 94, 170, 170, 170, 170, 170, 170, 170, 170, 57, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 170, 170, 170, 8, 0, 0, 0, 72, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 225, 43, 0, 224, 26, 44, 0, 224, 31, 44, 0, 16, 47, 44, 0, 224, 4, 0, 0, 0, 109, 0, 0, 0, 204, 0, 0, 0, 32, 2, 0, 0, 112, 2, 0, 0, 96, 0, 0, 0, 144, 58, 166, 158, 128, 7, 187, 159, 144, 28, 134, 160, 128, 233, 154, 161, 144, 254, 101, 162, 0, 6, 132, 163, 144, 224, 69, 164, 128, 166, 143, 164, 144, 12, 137, 203, 0, 24, 97, 210, 144, 118, 47, 247, 0, 148, 40, 248, 144, 88, 15, 249, 0, 118, 8, 250, 16, 117, 248, 250, 0, 88, 232, 251, 16, 87, 216, 252, 0, 58, 200, 253, 16, 57, 184, 254, 0, 28, 168, 255, 16, 27, 152, 0, 0, 254, 135, 1, 16, 253, 119, 2, 128, 26, 113, 3, 144, 25, 97, 4, 128, 252, 80, 5, 144, 251, 64, 6, 128, 222, 48, 7, 144, 53, 141, 7, 128, 192, 16, 9, 16, 177, 173, 9, 128, 162, 240, 10, 144, 161, 224, 11, 0, 191, 217, 12, 144, 131, 192, 13, 0, 161, 185, 14, 16, 160, 169, 15, 0, 131, 153, 16, 16, 130, 137, 17, 0, 101, 121, 18, 16, 100, 105, 19, 0, 71, 89, 20, 16, 70, 73, 21, 0, 41, 57, 22, 16, 40, 41, 23, 128, 69, 34, 24, 16, 10, 9, 25, 128, 39, 2, 26, 144, 38, 242, 26, 128, 9, 226, 27, 144, 8, 210, 28, 128, 235, 193, 29, 144, 234, 177, 30, 128, 205, 161, 31, 16, 29, 118, 32, 128, 175, 129, 33, 16, 255, 85, 34, 0, 204, 106, 35, 16, 225, 53, 36, 0, 174, 74, 37, 16, 195, 21, 38, 0, 144, 42, 39, 144, 223, 254, 39, 0, 114, 10, 41, 144, 193, 222, 41, 0, 84, 234, 42, 144, 163, 190, 43, 128, 112, 211, 44, 144, 133, 158, 45, 128, 82, 179, 46, 144, 103, 126, 47, 128, 52, 147, 48, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 16, 238, 198, 58, 0, 187, 219, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 144, 206, 111, 64, 128, 155, 132, 65, 144, 176, 79, 66, 128, 125, 100, 67, 144, 146, 47, 68, 128, 95, 68, 69, 16, 197, 243, 69, 0, 124, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 176, 12, 4, 94, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 97, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 148, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 144, 157, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 63, 44, 0, 224, 68, 44, 0, 224, 165, 44, 0, 224, 171, 44, 0, 16, 197, 44, 0, 224, 2, 0, 0, 0, 110, 0, 0, 0, 111, 2, 0, 0, 80, 0, 0, 0, 91, 34, 189, 133, 0, 148, 60, 153, 112, 240, 136, 203, 224, 251, 96, 210, 240, 168, 53, 215, 224, 161, 0, 216, 140, 144, 51, 251, 224, 59, 232, 251, 240, 58, 216, 252, 224, 29, 200, 253, 112, 223, 64, 6, 96, 194, 48, 7, 112, 25, 141, 7, 96, 164, 16, 9, 112, 163, 0, 10, 96, 134, 240, 10, 112, 133, 224, 11, 224, 162, 217, 12, 112, 103, 192, 13, 224, 132, 185, 14, 240, 131, 169, 15, 224, 102, 153, 16, 240, 101, 137, 17, 224, 72, 121, 18, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 112, 10, 242, 26, 96, 237, 225, 27, 112, 236, 209, 28, 96, 207, 193, 29, 112, 206, 177, 30, 96, 177, 161, 31, 240, 0, 118, 32, 96, 147, 129, 33, 240, 226, 85, 34, 224, 175, 106, 35, 240, 196, 53, 36, 224, 145, 74, 37, 240, 166, 21, 38, 224, 115, 42, 39, 112, 195, 254, 39, 224, 85, 10, 41, 112, 165, 222, 41, 224, 55, 234, 42, 112, 135, 190, 43, 96, 84, 211, 44, 112, 105, 158, 45, 96, 54, 179, 46, 112, 75, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 240, 239, 230, 56, 224, 188, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 170, 170, 170, 170, 170, 170, 170, 170, 80, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 8, 0, 0, 0, 37, 178, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 176, 185, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 217, 44, 0, 224, 220, 44, 0, 224, 47, 45, 0, 16, 68, 45, 0, 224, 2, 0, 0, 0, 112, 0, 0, 0, 132, 1, 0, 0, 89, 0, 0, 0, 224, 206, 222, 136, 144, 175, 184, 158, 128, 7, 187, 159, 144, 145, 152, 160, 128, 133, 210, 160, 144, 232, 138, 162, 0, 6, 132, 163, 144, 202, 106, 164, 128, 195, 53, 165, 16, 231, 83, 166, 128, 165, 21, 167, 16, 201, 51, 168, 0, 194, 254, 168, 144, 12, 137, 203, 0, 24, 97, 210, 16, 227, 85, 213, 0, 220, 32, 214, 144, 25, 97, 4, 128, 252, 80, 5, 144, 251, 64, 6, 128, 222, 48, 7, 144, 221, 32, 8, 128, 192, 16, 9, 144, 191, 0, 10, 128, 162, 240, 10, 144, 161, 224, 11, 0, 191, 217, 12, 144, 131, 192, 13, 0, 161, 185, 14, 16, 160, 169, 15, 0, 131, 153, 16, 16, 130, 137, 17, 0, 101, 121, 18, 16, 100, 105, 19, 0, 71, 89, 20, 16, 70, 73, 21, 0, 41, 57, 22, 16, 40, 41, 23, 128, 69, 34, 24, 16, 10, 9, 25, 128, 39, 2, 26, 144, 38, 242, 26, 128, 9, 226, 27, 144, 8, 210, 28, 128, 235, 193, 29, 144, 234, 177, 30, 128, 205, 161, 31, 16, 29, 118, 32, 128, 175, 129, 33, 16, 255, 85, 34, 0, 204, 106, 35, 16, 225, 53, 36, 0, 174, 74, 37, 16, 195, 21, 38, 0, 144, 42, 39, 144, 223, 254, 39, 0, 114, 10, 41, 144, 193, 222, 41, 0, 84, 234, 42, 144, 163, 190, 43, 128, 112, 211, 44, 144, 133, 158, 45, 128, 82, 179, 46, 144, 103, 126, 47, 128, 52, 147, 48, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 16, 238, 198, 58, 0, 187, 219, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 144, 206, 111, 64, 128, 155, 132, 65, 144, 176, 79, 66, 128, 125, 100, 67, 144, 146, 47, 68, 128, 95, 68, 69, 16, 197, 243, 69, 0, 124, 45, 71, 170, 170, 170, 170, 170, 170, 170, 170, 89, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 160, 149, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 144, 157, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 88, 45, 0, 224, 91, 45, 0, 224, 183, 45, 0, 16, 207, 45, 0, 224, 33, 0, 0, 0, 128, 136, 170, 150, 0, 102, 15, 184, 192, 92, 253, 184, 80, 80, 241, 185, 64, 144, 222, 186, 80, 202, 56, 218, 80, 22, 236, 218, 208, 253, 25, 220, 64, 117, 185, 220, 80, 49, 251, 221, 64, 250, 155, 222, 80, 182, 221, 223, 64, 79, 84, 224, 208, 27, 152, 244, 64, 122, 5, 245, 80, 128, 192, 246, 192, 58, 14, 247, 80, 72, 81, 248, 64, 225, 199, 248, 208, 238, 10, 250, 192, 20, 169, 250, 80, 34, 236, 251, 192, 153, 139, 252, 80, 170, 201, 29, 192, 243, 120, 30, 208, 81, 160, 31, 192, 235, 51, 32, 80, 133, 129, 33, 192, 228, 11, 34, 80, 209, 192, 44, 64, 224, 102, 45, 80, 127, 96, 72, 192, 4, 127, 82, 33, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 170, 170, 170, 8, 0, 0, 0, 128, 190, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 225, 45, 0, 224, 3, 46, 0, 16, 13, 46, 0, 224, 5, 0, 0, 0, 32, 166, 213, 163, 224, 220, 154, 32, 80, 155, 92, 33, 224, 190, 122, 34, 80, 125, 60, 35, 170, 170, 170, 170, 170, 170, 170, 170, 5, 0, 0, 0, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 96, 172, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 27, 46, 0, 224, 35, 46, 0, 16, 38, 46, 0, 224, 141, 0, 0, 0, 160, 189, 184, 158, 144, 21, 187, 159, 160, 26, 137, 203, 16, 38, 97, 210, 32, 241, 85, 213, 16, 234, 32, 214, 32, 211, 53, 215, 16, 204, 0, 216, 32, 181, 21, 217, 16, 174, 224, 217, 160, 209, 254, 218, 16, 144, 192, 219, 160, 179, 222, 220, 144, 172, 169, 221, 160, 149, 190, 222, 144, 142, 137, 223, 160, 119, 158, 224, 144, 112, 105, 225, 160, 89, 126, 226, 144, 82, 73, 227, 160, 59, 94, 228, 144, 52, 41, 229, 32, 88, 71, 230, 16, 81, 18, 231, 32, 58, 39, 232, 16, 51, 242, 232, 32, 28, 7, 234, 16, 21, 210, 234, 32, 254, 230, 235, 16, 247, 177, 236, 32, 224, 198, 237, 16, 217, 145, 238, 160, 252, 175, 239, 16, 187, 113, 240, 160, 222, 143, 241, 144, 193, 127, 242, 160, 192, 111, 243, 144, 163, 95, 244, 160, 162, 79, 245, 144, 133, 63, 246, 160, 132, 47, 247, 16, 162, 40, 248, 160, 102, 15, 249, 16, 132, 8, 250, 32, 131, 248, 250, 16, 102, 232, 251, 32, 101, 216, 252, 16, 72, 200, 253, 32, 71, 184, 254, 16, 42, 168, 255, 32, 41, 152, 0, 16, 12, 136, 1, 32, 11, 120, 2, 144, 40, 113, 3, 160, 39, 97, 4, 144, 10, 81, 5, 160, 9, 65, 6, 144, 236, 48, 7, 160, 235, 32, 8, 144, 206, 16, 9, 160, 205, 0, 10, 144, 176, 240, 10, 160, 175, 224, 11, 16, 205, 217, 12, 160, 145, 192, 13, 16, 175, 185, 14, 32, 174, 169, 15, 16, 145, 153, 16, 32, 144, 137, 17, 16, 115, 121, 18, 32, 114, 105, 19, 16, 85, 89, 20, 32, 84, 73, 21, 16, 55, 57, 22, 32, 54, 41, 23, 144, 83, 34, 24, 32, 24, 9, 25, 144, 53, 2, 26, 160, 52, 242, 26, 144, 23, 226, 27, 160, 22, 210, 28, 144, 249, 193, 29, 160, 248, 177, 30, 144, 219, 161, 31, 32, 43, 118, 32, 144, 189, 129, 33, 32, 13, 86, 34, 16, 218, 106, 35, 32, 239, 53, 36, 16, 188, 74, 37, 32, 209, 21, 38, 16, 158, 42, 39, 160, 237, 254, 39, 16, 128, 10, 41, 160, 207, 222, 41, 16, 98, 234, 42, 160, 177, 190, 43, 144, 126, 211, 44, 160, 147, 158, 45, 144, 96, 179, 46, 160, 117, 126, 47, 144, 66, 147, 48, 32, 146, 103, 49, 144, 36, 115, 50, 32, 116, 71, 51, 144, 6, 83, 52, 32, 86, 39, 53, 144, 232, 50, 54, 32, 56, 7, 55, 16, 5, 28, 56, 32, 26, 231, 56, 16, 231, 251, 57, 32, 252, 198, 58, 16, 201, 219, 59, 160, 24, 176, 60, 16, 171, 187, 61, 160, 250, 143, 62, 16, 141, 155, 63, 160, 220, 111, 64, 144, 169, 132, 65, 160, 190, 79, 66, 144, 139, 100, 67, 160, 160, 47, 68, 144, 109, 68, 69, 32, 211, 243, 69, 16, 138, 45, 71, 32, 181, 211, 71, 16, 108, 13, 73, 32, 151, 179, 73, 16, 78, 237, 74, 160, 179, 156, 75, 144, 106, 214, 76, 160, 149, 124, 77, 144, 76, 182, 78, 160, 119, 92, 79, 144, 46, 150, 80, 160, 89, 60, 81, 144, 16, 118, 82, 160, 59, 28, 83, 144, 242, 85, 84, 160, 29, 252, 84, 2, 0, 0, 0, 255, 255, 255, 255, 135, 118, 61, 94, 142, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 170, 170, 8, 0, 0, 0, 249, 140, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 50, 46, 0, 224, 192, 46, 0, 224, 195, 46, 0, 16, 232, 46, 0, 224, 39, 0, 0, 0, 24, 107, 170, 150, 224, 73, 15, 184, 160, 64, 253, 184, 48, 52, 241, 185, 32, 116, 222, 186, 48, 174, 56, 218, 48, 250, 235, 218, 176, 225, 25, 220, 32, 89, 185, 220, 48, 21, 251, 221, 32, 222, 155, 222, 48, 154, 221, 223, 32, 51, 84, 224, 176, 255, 151, 244, 32, 94, 5, 245, 48, 100, 192, 246, 160, 30, 14, 247, 48, 44, 81, 248, 32, 197, 199, 248, 176, 210, 10, 250, 160, 248, 168, 250, 48, 6, 236, 251, 160, 125, 139, 252, 48, 142, 201, 29, 160, 215, 120, 30, 176, 53, 160, 31, 160, 207, 51, 32, 48, 105, 129, 33, 160, 200, 11, 34, 176, 16, 88, 35, 32, 112, 226, 35, 176, 242, 55, 37, 32, 199, 212, 37, 176, 198, 246, 55, 32, 133, 184, 56, 48, 227, 223, 57, 32, 74, 242, 57, 176, 255, 200, 59, 160, 14, 111, 60, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 39, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 232, 219, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 248, 46, 0, 224, 35, 47, 0, 16, 46, 47, 0, 224, 79, 0, 0, 0, 52, 168, 241, 128, 96, 133, 184, 158, 80, 221, 186, 159, 96, 226, 136, 203, 208, 237, 96, 210, 96, 63, 158, 224, 80, 56, 105, 225, 96, 239, 96, 4, 80, 210, 80, 5, 96, 209, 64, 6, 80, 180, 48, 7, 96, 179, 32, 8, 80, 150, 16, 9, 96, 149, 0, 10, 80, 120, 240, 10, 96, 119, 224, 11, 208, 148, 217, 12, 96, 89, 192, 13, 208, 118, 185, 14, 224, 117, 169, 15, 208, 88, 153, 16, 224, 87, 137, 17, 208, 58, 121, 18, 224, 57, 105, 19, 208, 28, 89, 20, 224, 27, 73, 21, 208, 254, 56, 22, 224, 253, 40, 23, 80, 27, 34, 24, 224, 223, 8, 25, 80, 253, 1, 26, 96, 252, 241, 26, 80, 223, 225, 27, 96, 222, 209, 28, 80, 193, 193, 29, 96, 192, 177, 30, 80, 163, 161, 31, 224, 242, 117, 32, 80, 133, 129, 33, 224, 212, 85, 34, 208, 161, 106, 35, 224, 182, 53, 36, 208, 131, 74, 37, 224, 152, 21, 38, 208, 101, 42, 39, 96, 181, 254, 39, 208, 71, 10, 41, 96, 151, 222, 41, 208, 41, 234, 42, 96, 121, 190, 43, 80, 70, 211, 44, 96, 91, 158, 45, 80, 40, 179, 46, 96, 61, 126, 47, 80, 10, 147, 48, 224, 89, 103, 49, 80, 236, 114, 50, 224, 59, 71, 51, 80, 206, 82, 52, 224, 29, 39, 53, 80, 176, 50, 54, 224, 255, 6, 55, 208, 204, 27, 56, 224, 225, 230, 56, 208, 174, 251, 57, 224, 195, 198, 58, 208, 144, 219, 59, 96, 224, 175, 60, 208, 114, 187, 61, 96, 194, 143, 62, 208, 84, 155, 63, 96, 164, 111, 64, 80, 113, 132, 65, 96, 134, 79, 66, 80, 83, 100, 67, 96, 104, 47, 68, 80, 53, 68, 69, 224, 154, 243, 69, 208, 81, 45, 71, 170, 170, 170, 170, 79, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 204, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 192, 199, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 58, 47, 0, 224, 139, 47, 0, 16, 160, 47, 0, 224, 35, 0, 0, 0, 0, 104, 128, 155, 80, 124, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 35, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 128, 207, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 208, 213, 255, 127, 4, 0, 0, 96, 205, 7, 0, 112, 177, 47, 0, 224, 215, 47, 0, 16, 225, 47, 0, 224, 152, 0, 0, 0, 140, 126, 184, 158, 124, 214, 186, 159, 108, 77, 158, 190, 56, 49, 184, 192, 168, 239, 121, 193, 56, 19, 152, 194, 168, 209, 89, 195, 56, 245, 119, 196, 168, 179, 57, 197, 184, 17, 97, 198, 168, 149, 25, 199, 184, 243, 64, 200, 40, 178, 2, 201, 184, 213, 32, 202, 40, 148, 226, 202, 184, 183, 0, 204, 200, 230, 96, 210, 216, 68, 136, 211, 72, 3, 74, 212, 216, 38, 104, 213, 72, 229, 41, 214, 216, 8, 72, 215, 72, 199, 9, 216, 216, 234, 39, 217, 72, 169, 233, 217, 88, 7, 17, 219, 200, 197, 210, 219, 88, 116, 222, 220, 72, 109, 169, 221, 88, 86, 190, 222, 72, 79, 137, 223, 88, 56, 158, 224, 72, 49, 105, 225, 88, 26, 126, 226, 72, 19, 73, 227, 88, 252, 93, 228, 72, 245, 40, 229, 216, 24, 71, 230, 200, 17, 18, 231, 216, 250, 38, 232, 200, 243, 241, 232, 216, 220, 6, 234, 200, 213, 209, 234, 216, 190, 230, 235, 200, 183, 177, 236, 216, 160, 198, 237, 72, 190, 191, 238, 88, 189, 175, 239, 72, 160, 159, 240, 88, 159, 143, 241, 72, 130, 127, 242, 88, 129, 111, 243, 72, 100, 95, 244, 88, 99, 79, 245, 72, 70, 63, 246, 88, 69, 47, 247, 200, 98, 40, 248, 88, 107, 218, 248, 96, 46, 15, 249, 208, 75, 8, 250, 224, 74, 248, 250, 208, 45, 232, 251, 224, 44, 216, 252, 208, 15, 200, 253, 224, 14, 184, 254, 208, 241, 167, 255, 224, 240, 151, 0, 208, 211, 135, 1, 224, 210, 119, 2, 80, 240, 112, 3, 96, 239, 96, 4, 80, 210, 80, 5, 96, 209, 64, 6, 80, 180, 48, 7, 96, 179, 32, 8, 80, 150, 16, 9, 96, 149, 0, 10, 80, 120, 240, 10, 96, 119, 224, 11, 208, 148, 217, 12, 96, 89, 192, 13, 208, 118, 185, 14, 224, 117, 169, 15, 208, 88, 153, 16, 224, 87, 137, 17, 208, 58, 121, 18, 224, 57, 105, 19, 208, 28, 89, 20, 224, 27, 73, 21, 208, 254, 56, 22, 224, 253, 40, 23, 80, 27, 34, 24, 224, 223, 8, 25, 80, 253, 1, 26, 96, 252, 241, 26, 80, 223, 225, 27, 96, 222, 209, 28, 80, 193, 193, 29, 96, 192, 177, 30, 80, 163, 161, 31, 252, 214, 117, 32, 108, 105, 129, 33, 252, 184, 85, 34, 220, 119, 106, 35, 252, 154, 53, 36, 236, 103, 74, 37, 252, 124, 21, 38, 236, 73, 42, 39, 124, 153, 254, 39, 236, 43, 10, 41, 124, 123, 222, 41, 236, 13, 234, 42, 124, 93, 190, 43, 108, 42, 211, 44, 124, 63, 158, 45, 108, 12, 179, 46, 124, 33, 126, 47, 108, 238, 146, 48, 252, 61, 103, 49, 108, 208, 114, 50, 252, 31, 71, 51, 108, 178, 82, 52, 252, 1, 39, 53, 108, 148, 50, 54, 252, 227, 6, 55, 236, 176, 27, 56, 252, 197, 230, 56, 236, 146, 251, 57, 252, 167, 198, 58, 236, 116, 219, 59, 124, 196, 175, 60, 236, 86, 187, 61, 124, 166, 143, 62, 236, 56, 155, 63, 124, 136, 111, 64, 108, 85, 132, 65, 124, 106, 79, 66, 108, 55, 100, 67, 124, 76, 47, 68, 108, 25, 68, 69, 252, 126, 243, 69, 236, 53, 45, 71, 252, 96, 211, 71, 236, 23, 13, 73, 252, 66, 179, 73, 236, 249, 236, 74, 124, 95, 156, 75, 108, 22, 214, 76, 124, 65, 124, 77, 80, 20, 182, 78, 96, 63, 92, 79, 80, 246, 149, 80, 2, 0, 0, 0, 255, 255, 255, 255, 36, 60, 61, 94, 170, 170, 170, 170, 153, 0, 0, 0, 4, 5, 4, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 16, 0, 0, 0, 92, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 192, 199, 255, 255, 32, 28, 0, 0, 148, 206, 255, 255, 0, 0, 0, 0, 148, 206, 255, 255, 16, 14, 0, 0, 200, 206, 255, 255, 0, 0, 0, 0, 200, 206, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 192, 199, 255, 127, 169, 3, 0, 96, 221, 7, 0, 112, 242, 47, 0, 224, 139, 48, 0, 224, 143, 48, 0, 16, 183, 48, 0, 224, 77, 0, 0, 0, 254, 180, 15, 147, 240, 101, 137, 17, 224, 72, 121, 18, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 112, 10, 242, 26, 96, 237, 225, 27, 112, 236, 209, 28, 96, 207, 193, 29, 112, 206, 177, 30, 96, 177, 161, 31, 240, 0, 118, 32, 96, 147, 129, 33, 240, 226, 85, 34, 224, 175, 106, 35, 240, 196, 53, 36, 224, 145, 74, 37, 240, 166, 21, 38, 224, 115, 42, 39, 112, 195, 254, 39, 224, 85, 10, 41, 112, 165, 222, 41, 224, 55, 234, 42, 112, 135, 190, 43, 96, 84, 211, 44, 112, 105, 158, 45, 96, 54, 179, 46, 112, 75, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 240, 239, 230, 56, 224, 188, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 240, 138, 211, 71, 224, 65, 13, 73, 240, 108, 179, 73, 224, 35, 237, 74, 112, 137, 156, 75, 96, 64, 214, 76, 112, 107, 124, 77, 96, 34, 182, 78, 112, 77, 92, 79, 96, 4, 150, 80, 112, 47, 60, 81, 96, 230, 117, 82, 112, 17, 28, 83, 96, 200, 85, 84, 112, 243, 251, 84, 96, 170, 53, 86, 240, 211, 164, 90, 224, 138, 222, 91, 2, 0, 0, 0, 255, 255, 255, 255, 48, 30, 135, 105, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 78, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 4, 3, 2, 170, 170, 10, 0, 0, 0, 80, 189, 255, 255, 0, 0, 0, 0, 2, 184, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 1, 0, 0, 96, 227, 7, 0, 112, 211, 48, 0, 224, 33, 49, 0, 224, 39, 49, 0, 16, 60, 49, 0, 224, 9, 0, 0, 0, 220, 234, 157, 159, 96, 172, 85, 7, 208, 150, 205, 7, 96, 120, 44, 25, 80, 228, 207, 25, 224, 238, 234, 39, 208, 92, 200, 40, 96, 82, 84, 68, 80, 75, 31, 69, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 9, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 36, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 82, 49, 0, 224, 95, 49, 0, 16, 99, 49, 0, 224, 3, 0, 0, 0, 24, 66, 164, 182, 208, 252, 22, 43, 64, 230, 113, 43, 2, 0, 0, 0, 255, 255, 255, 255, 88, 38, 135, 105, 170, 170, 170, 170, 4, 0, 0, 0, 1, 2, 3, 2, 8, 0, 0, 0, 40, 181, 255, 255, 0, 0, 0, 0, 104, 182, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 111, 49, 0, 224, 115, 49, 0, 224, 119, 49, 0, 16, 121, 49, 0, 224, 3, 0, 0, 0, 136, 121, 217, 152, 60, 180, 125, 10, 48, 251, 127, 39, 170, 170, 170, 170, 170, 170, 170, 170, 3, 0, 0, 0, 2, 3, 1, 170, 8, 0, 0, 0, 120, 201, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 68, 203, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 137, 49, 0, 224, 143, 49, 0, 16, 145, 49, 0, 224, 2, 0, 0, 0, 128, 0, 0, 0, 128, 1, 0, 0, 167, 0, 0, 0, 160, 171, 241, 128, 192, 222, 228, 154, 48, 19, 214, 155, 96, 133, 184, 158, 80, 221, 186, 159, 64, 23, 157, 162, 48, 177, 48, 163, 64, 86, 122, 164, 48, 31, 27, 165, 192, 160, 83, 166, 176, 82, 252, 166, 64, 189, 60, 168, 176, 52, 220, 168, 64, 159, 28, 170, 48, 58, 205, 170, 64, 129, 252, 171, 48, 145, 191, 172, 64, 216, 238, 173, 48, 254, 140, 174, 64, 69, 188, 175, 48, 85, 127, 176, 64, 156, 174, 177, 176, 112, 75, 178, 64, 126, 142, 179, 48, 187, 36, 180, 64, 96, 110, 181, 176, 192, 21, 182, 64, 66, 78, 183, 176, 23, 8, 184, 192, 233, 36, 185, 176, 249, 231, 185, 192, 203, 4, 187, 48, 22, 209, 187, 64, 93, 0, 189, 176, 49, 157, 189, 64, 180, 242, 190, 48, 218, 144, 191, 192, 231, 211, 192, 48, 71, 94, 193, 64, 142, 141, 194, 48, 158, 80, 195, 64, 112, 109, 196, 48, 128, 48, 197, 64, 60, 114, 198, 48, 98, 16, 199, 192, 110, 54, 200, 176, 126, 249, 200, 192, 80, 22, 202, 176, 96, 217, 202, 96, 226, 136, 203, 208, 237, 96, 210, 224, 214, 117, 211, 208, 207, 64, 212, 224, 184, 85, 213, 208, 177, 32, 214, 224, 154, 53, 215, 208, 147, 0, 216, 224, 124, 21, 217, 208, 117, 224, 217, 96, 123, 222, 220, 80, 116, 169, 221, 96, 93, 190, 222, 80, 86, 137, 223, 96, 63, 158, 224, 80, 56, 105, 225, 96, 33, 126, 226, 80, 26, 73, 227, 224, 31, 71, 230, 208, 24, 18, 231, 224, 1, 39, 232, 208, 250, 241, 232, 224, 227, 6, 234, 208, 220, 209, 234, 224, 197, 230, 235, 208, 190, 177, 236, 96, 166, 143, 241, 80, 137, 127, 242, 96, 136, 111, 243, 80, 107, 95, 244, 96, 106, 79, 245, 80, 77, 63, 246, 96, 76, 47, 247, 208, 105, 40, 248, 96, 46, 15, 249, 208, 75, 8, 250, 224, 74, 248, 250, 208, 45, 232, 251, 224, 44, 216, 252, 208, 15, 200, 253, 224, 14, 184, 254, 208, 241, 167, 255, 224, 240, 151, 0, 208, 211, 135, 1, 224, 210, 119, 2, 80, 240, 112, 3, 96, 239, 96, 4, 80, 210, 80, 5, 96, 209, 64, 6, 80, 180, 48, 7, 96, 179, 32, 8, 80, 150, 16, 9, 96, 149, 0, 10, 80, 120, 240, 10, 96, 119, 224, 11, 208, 148, 217, 12, 96, 89, 192, 13, 208, 118, 185, 14, 224, 117, 169, 15, 208, 88, 153, 16, 224, 87, 137, 17, 208, 58, 121, 18, 224, 57, 105, 19, 208, 28, 89, 20, 224, 27, 73, 21, 208, 254, 56, 22, 224, 253, 40, 23, 80, 27, 34, 24, 224, 223, 8, 25, 80, 253, 1, 26, 96, 252, 241, 26, 80, 223, 225, 27, 96, 222, 209, 28, 80, 193, 193, 29, 96, 192, 177, 30, 80, 163, 161, 31, 224, 242, 117, 32, 80, 133, 129, 33, 224, 212, 85, 34, 208, 161, 106, 35, 224, 182, 53, 36, 208, 131, 74, 37, 224, 152, 21, 38, 208, 101, 42, 39, 96, 181, 254, 39, 208, 71, 10, 41, 96, 151, 222, 41, 208, 41, 234, 42, 96, 121, 190, 43, 80, 70, 211, 44, 96, 91, 158, 45, 80, 40, 179, 46, 96, 61, 126, 47, 80, 10, 147, 48, 224, 89, 103, 49, 80, 236, 114, 50, 224, 59, 71, 51, 80, 206, 82, 52, 224, 29, 39, 53, 80, 176, 50, 54, 224, 255, 6, 55, 208, 204, 27, 56, 224, 225, 230, 56, 208, 174, 251, 57, 224, 195, 198, 58, 208, 144, 219, 59, 96, 224, 175, 60, 208, 114, 187, 61, 96, 194, 143, 62, 208, 84, 155, 63, 96, 164, 111, 64, 80, 113, 132, 65, 96, 134, 79, 66, 80, 83, 100, 67, 96, 104, 47, 68, 80, 53, 68, 69, 224, 154, 243, 69, 208, 81, 45, 71, 170, 170, 170, 170, 167, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 96, 196, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 192, 199, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 159, 49, 0, 224, 162, 49, 0, 224, 75, 50, 0, 16, 118, 50, 0, 224, 2, 0, 0, 0, 129, 0, 0, 0, 139, 1, 0, 0, 107, 0, 0, 0, 128, 194, 98, 172, 80, 148, 211, 177, 64, 93, 116, 178, 208, 102, 91, 200, 64, 81, 211, 200, 208, 72, 59, 202, 192, 109, 188, 202, 80, 101, 36, 204, 192, 79, 156, 204, 80, 11, 196, 209, 192, 245, 59, 210, 80, 237, 163, 211, 192, 215, 27, 212, 208, 5, 96, 247, 64, 125, 255, 247, 208, 68, 61, 249, 192, 83, 227, 249, 208, 59, 219, 250, 64, 134, 167, 251, 208, 169, 197, 252, 64, 104, 135, 253, 208, 0, 184, 254, 192, 227, 167, 255, 208, 226, 151, 0, 192, 197, 135, 1, 208, 196, 119, 2, 64, 226, 112, 3, 80, 225, 96, 4, 192, 20, 53, 5, 80, 195, 64, 6, 64, 72, 22, 7, 80, 165, 32, 8, 192, 123, 247, 8, 80, 135, 0, 10, 64, 106, 240, 10, 80, 105, 224, 11, 192, 134, 217, 12, 80, 75, 192, 13, 192, 104, 185, 14, 80, 162, 178, 15, 64, 155, 125, 16, 208, 234, 81, 17, 192, 183, 102, 18, 208, 204, 49, 19, 192, 153, 70, 20, 208, 130, 91, 21, 192, 123, 38, 22, 208, 100, 59, 23, 192, 93, 6, 24, 208, 70, 27, 25, 192, 63, 230, 25, 208, 40, 251, 26, 64, 92, 207, 27, 208, 10, 219, 28, 64, 62, 175, 29, 80, 83, 122, 30, 64, 32, 143, 31, 80, 53, 90, 32, 64, 2, 111, 33, 208, 81, 67, 34, 64, 228, 78, 35, 208, 51, 35, 36, 64, 198, 46, 37, 208, 138, 21, 38, 192, 226, 23, 39, 80, 167, 254, 39, 208, 210, 247, 40, 80, 137, 222, 41, 208, 180, 215, 42, 80, 107, 190, 43, 208, 150, 183, 44, 80, 77, 158, 45, 208, 120, 151, 46, 80, 47, 126, 47, 208, 90, 119, 48, 208, 75, 103, 49, 208, 60, 87, 50, 208, 45, 71, 51, 80, 89, 64, 52, 80, 213, 29, 53, 80, 176, 50, 54, 80, 183, 253, 54, 208, 204, 27, 56, 208, 211, 230, 56, 208, 174, 251, 57, 208, 181, 198, 58, 208, 144, 219, 59, 80, 210, 175, 60, 208, 114, 187, 61, 80, 180, 143, 62, 208, 84, 155, 63, 208, 91, 102, 64, 80, 53, 68, 69, 208, 140, 243, 69, 80, 23, 36, 71, 80, 169, 220, 71, 80, 249, 3, 73, 208, 80, 179, 73, 80, 219, 227, 74, 80, 109, 156, 75, 208, 247, 204, 76, 208, 137, 133, 77, 208, 78, 191, 78, 208, 224, 119, 79, 80, 246, 149, 80, 80, 19, 60, 81, 80, 216, 117, 82, 2, 0, 0, 0, 255, 255, 255, 255, 184, 40, 135, 105, 170, 170, 170, 170, 108, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 8, 0, 0, 0, 200, 178, 255, 255, 0, 0, 0, 0, 192, 178, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 176, 185, 255, 127, 96, 3, 0, 96, 222, 7, 0, 112, 136, 50, 0, 224, 139, 50, 0, 224, 247, 50, 0, 224, 251, 50, 0, 16, 23, 51, 0, 224, 15, 0, 0, 0, 112, 232, 182, 165, 224, 110, 242, 175, 96, 86, 102, 182, 96, 210, 67, 183, 96, 54, 12, 184, 240, 134, 253, 184, 96, 113, 234, 203, 240, 180, 145, 216, 128, 112, 0, 0, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 170, 170, 170, 170, 170, 170, 170, 170, 15, 0, 0, 0, 2, 4, 2, 4, 2, 4, 2, 1, 2, 3, 2, 3, 2, 3, 2, 170, 10, 0, 0, 0, 248, 151, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 45, 51, 0, 224, 63, 51, 0, 16, 68, 51, 0, 224, 5, 0, 0, 0, 117, 0, 0, 0, 131, 0, 0, 0, 139, 0, 0, 0, 0, 2, 0, 0, 107, 2, 0, 0, 37, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 34, 87, 202, 112, 71, 216, 202, 128, 254, 136, 203, 240, 9, 97, 210, 0, 243, 117, 211, 240, 235, 64, 212, 0, 213, 85, 213, 240, 205, 32, 214, 0, 183, 53, 215, 240, 175, 0, 216, 0, 153, 21, 217, 240, 145, 224, 217, 128, 181, 254, 218, 240, 115, 192, 219, 128, 151, 222, 220, 112, 144, 169, 221, 128, 121, 190, 222, 112, 114, 137, 223, 128, 91, 158, 224, 112, 84, 105, 225, 128, 61, 126, 226, 112, 54, 73, 227, 128, 31, 94, 228, 240, 22, 242, 232, 0, 0, 7, 234, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 38, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 10, 0, 0, 0, 58, 175, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 176, 185, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 84, 51, 0, 224, 90, 51, 0, 224, 128, 51, 0, 224, 131, 51, 0, 16, 142, 51, 0, 224, 3, 0, 0, 0, 132, 0, 0, 0, 147, 0, 0, 0, 110, 2, 0, 0, 92, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 254, 136, 203, 240, 9, 97, 210, 0, 213, 85, 213, 240, 205, 32, 214, 0, 183, 53, 215, 240, 175, 0, 216, 0, 153, 21, 217, 240, 145, 224, 217, 128, 181, 254, 218, 240, 115, 192, 219, 128, 151, 222, 220, 112, 144, 169, 221, 128, 121, 190, 222, 112, 114, 137, 223, 128, 91, 158, 224, 112, 84, 105, 225, 128, 61, 126, 226, 112, 54, 73, 227, 128, 31, 94, 228, 240, 60, 87, 229, 0, 60, 71, 230, 240, 30, 55, 231, 0, 30, 39, 232, 240, 22, 242, 232, 0, 0, 7, 234, 240, 248, 209, 234, 0, 226, 230, 235, 240, 196, 214, 236, 0, 196, 198, 237, 112, 225, 191, 238, 128, 224, 175, 239, 112, 195, 159, 240, 128, 194, 143, 241, 112, 135, 95, 244, 0, 103, 248, 250, 240, 73, 232, 251, 0, 73, 216, 252, 240, 43, 200, 253, 0, 43, 184, 254, 240, 13, 168, 255, 0, 13, 152, 0, 240, 239, 135, 1, 0, 239, 119, 2, 112, 12, 113, 3, 128, 11, 97, 4, 112, 238, 80, 5, 128, 237, 64, 6, 112, 208, 48, 7, 128, 39, 141, 7, 112, 178, 16, 9, 0, 163, 173, 9, 112, 148, 240, 10, 128, 147, 224, 11, 240, 176, 217, 12, 128, 117, 192, 13, 240, 146, 185, 14, 0, 146, 169, 15, 240, 116, 153, 16, 0, 116, 137, 17, 240, 86, 121, 18, 0, 86, 105, 19, 240, 56, 89, 20, 0, 56, 73, 21, 240, 26, 57, 22, 0, 26, 41, 23, 112, 55, 34, 24, 0, 252, 8, 25, 112, 25, 2, 26, 128, 24, 242, 26, 112, 251, 225, 27, 128, 250, 209, 28, 112, 221, 193, 29, 128, 220, 177, 30, 112, 191, 161, 31, 0, 15, 118, 32, 112, 161, 129, 33, 0, 241, 85, 34, 240, 189, 106, 35, 0, 211, 53, 36, 240, 159, 74, 37, 0, 181, 21, 38, 240, 129, 42, 39, 128, 209, 254, 39, 240, 99, 10, 41, 112, 118, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 170, 170, 170, 170, 93, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 1, 2, 1, 170, 170, 170, 8, 0, 0, 0, 202, 174, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 160, 171, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 166, 51, 0, 224, 170, 51, 0, 224, 7, 52, 0, 224, 11, 52, 0, 16, 36, 52, 0, 224, 41, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 254, 136, 203, 240, 9, 97, 210, 128, 151, 222, 220, 112, 144, 169, 221, 128, 61, 126, 226, 112, 54, 73, 227, 128, 31, 94, 228, 112, 24, 41, 229, 0, 60, 71, 230, 240, 52, 18, 231, 0, 30, 39, 232, 240, 22, 242, 232, 0, 0, 7, 234, 240, 248, 209, 234, 0, 226, 230, 235, 240, 218, 177, 236, 0, 196, 198, 237, 240, 188, 145, 238, 128, 224, 175, 239, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 240, 224, 119, 2, 96, 254, 112, 3, 112, 253, 96, 4, 96, 224, 80, 5, 112, 223, 64, 6, 96, 194, 48, 7, 112, 25, 141, 7, 112, 178, 16, 9, 240, 148, 173, 9, 96, 134, 240, 10, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 42, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 3, 4, 3, 4, 3, 4, 3, 170, 170, 10, 0, 0, 0, 13, 175, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 58, 52, 0, 224, 100, 52, 0, 224, 103, 52, 0, 16, 115, 52, 0, 224, 56, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 254, 136, 203, 240, 9, 97, 210, 224, 61, 103, 228, 112, 24, 41, 229, 0, 60, 71, 230, 240, 52, 18, 231, 0, 30, 39, 232, 240, 22, 242, 232, 0, 0, 7, 234, 240, 248, 209, 234, 0, 226, 230, 235, 240, 218, 177, 236, 0, 196, 198, 237, 240, 188, 145, 238, 128, 224, 175, 239, 112, 195, 159, 240, 128, 194, 143, 241, 112, 165, 127, 242, 128, 164, 111, 243, 112, 135, 95, 244, 128, 134, 79, 245, 112, 105, 63, 246, 128, 104, 47, 247, 240, 103, 8, 250, 0, 103, 248, 250, 240, 73, 232, 251, 0, 73, 216, 252, 240, 43, 200, 253, 0, 43, 184, 254, 240, 13, 168, 255, 0, 13, 152, 0, 240, 239, 135, 1, 0, 239, 119, 2, 112, 12, 113, 3, 128, 11, 97, 4, 112, 238, 80, 5, 128, 237, 64, 6, 112, 208, 48, 7, 128, 39, 141, 7, 112, 178, 16, 9, 0, 163, 173, 9, 112, 148, 240, 10, 128, 147, 224, 11, 240, 176, 217, 12, 128, 117, 192, 13, 240, 146, 185, 14, 112, 118, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 240, 138, 211, 71, 224, 65, 13, 73, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 170, 170, 170, 170, 170, 170, 170, 170, 57, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 1, 2, 3, 4, 3, 170, 170, 170, 10, 0, 0, 0, 45, 174, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 1, 0, 0, 96, 217, 7, 0, 112, 137, 52, 0, 224, 194, 52, 0, 224, 199, 52, 0, 16, 215, 52, 0, 224, 36, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 254, 136, 203, 240, 9, 97, 210, 224, 61, 103, 228, 112, 24, 41, 229, 0, 60, 71, 230, 240, 52, 18, 231, 0, 30, 39, 232, 240, 22, 242, 232, 0, 0, 7, 234, 240, 248, 209, 234, 0, 226, 230, 235, 240, 218, 177, 236, 0, 196, 198, 237, 240, 188, 145, 238, 128, 224, 175, 239, 112, 195, 159, 240, 128, 194, 143, 241, 112, 165, 127, 242, 128, 164, 111, 243, 112, 135, 95, 244, 128, 134, 79, 245, 240, 73, 232, 251, 0, 73, 216, 252, 240, 43, 200, 253, 0, 43, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 112, 118, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 170, 170, 170, 170, 170, 170, 170, 170, 37, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 2, 1, 4, 3, 4, 3, 2, 1, 2, 1, 170, 170, 170, 10, 0, 0, 0, 169, 174, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 160, 171, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 237, 52, 0, 224, 18, 53, 0, 224, 23, 53, 0, 16, 34, 53, 0, 224, 19, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 254, 136, 203, 240, 9, 97, 210, 128, 61, 126, 226, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 240, 224, 119, 2, 96, 254, 112, 3, 112, 253, 96, 4, 96, 224, 80, 5, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 20, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 10, 0, 0, 0, 64, 176, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 56, 53, 0, 224, 76, 53, 0, 224, 79, 53, 0, 16, 85, 53, 0, 224, 41, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 254, 136, 203, 240, 9, 97, 210, 0, 243, 117, 211, 240, 235, 64, 212, 128, 91, 158, 224, 112, 84, 105, 225, 128, 61, 126, 226, 112, 54, 73, 227, 224, 61, 103, 228, 112, 24, 41, 229, 0, 60, 71, 230, 240, 52, 18, 231, 0, 30, 39, 232, 240, 22, 242, 232, 0, 0, 7, 234, 240, 248, 209, 234, 0, 226, 230, 235, 240, 218, 177, 236, 0, 196, 198, 237, 112, 225, 191, 238, 128, 224, 175, 239, 240, 158, 113, 240, 128, 194, 143, 241, 112, 165, 127, 242, 128, 164, 111, 243, 112, 135, 95, 244, 128, 134, 79, 245, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 112, 118, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 240, 138, 211, 71, 224, 65, 13, 73, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 42, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 4, 3, 4, 3, 2, 1, 2, 3, 4, 3, 170, 170, 10, 0, 0, 0, 241, 173, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 1, 0, 0, 96, 217, 7, 0, 112, 107, 53, 0, 224, 149, 53, 0, 224, 155, 53, 0, 16, 167, 53, 0, 224, 47, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 254, 136, 203, 240, 9, 97, 210, 0, 243, 117, 211, 240, 235, 64, 212, 0, 213, 85, 213, 240, 205, 32, 214, 0, 183, 53, 215, 240, 175, 0, 216, 0, 153, 21, 217, 240, 145, 224, 217, 128, 181, 254, 218, 240, 115, 192, 219, 128, 151, 222, 220, 112, 144, 169, 221, 128, 121, 190, 222, 112, 114, 137, 223, 128, 91, 158, 224, 112, 84, 105, 225, 128, 61, 126, 226, 112, 54, 73, 227, 128, 31, 94, 228, 240, 60, 87, 229, 0, 60, 71, 230, 240, 30, 55, 231, 0, 30, 39, 232, 240, 22, 242, 232, 0, 0, 7, 234, 240, 248, 209, 234, 0, 226, 230, 235, 240, 218, 177, 236, 0, 196, 198, 237, 240, 188, 145, 238, 128, 224, 175, 239, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 112, 118, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 224, 95, 45, 71, 240, 138, 211, 71, 224, 65, 13, 73, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 48, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 4, 3, 4, 3, 2, 1, 4, 3, 4, 3, 10, 0, 0, 0, 207, 174, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 1, 0, 0, 96, 217, 7, 0, 112, 189, 53, 0, 224, 237, 53, 0, 224, 243, 53, 0, 16, 0, 54, 0, 224, 60, 0, 0, 0, 128, 78, 6, 224, 128, 104, 47, 247, 0, 148, 40, 248, 32, 144, 137, 17, 16, 100, 105, 19, 0, 71, 89, 20, 16, 70, 73, 21, 0, 41, 57, 22, 16, 40, 41, 23, 128, 69, 34, 24, 16, 10, 9, 25, 128, 39, 2, 26, 144, 38, 242, 26, 128, 9, 226, 27, 144, 8, 210, 28, 128, 235, 193, 29, 144, 234, 177, 30, 128, 205, 161, 31, 16, 29, 118, 32, 128, 175, 129, 33, 16, 255, 85, 34, 0, 204, 106, 35, 16, 225, 53, 36, 0, 174, 74, 37, 16, 195, 21, 38, 0, 144, 42, 39, 144, 223, 254, 39, 0, 114, 10, 41, 144, 193, 222, 41, 0, 84, 234, 42, 144, 163, 190, 43, 128, 112, 211, 44, 144, 133, 158, 45, 128, 82, 179, 46, 144, 103, 126, 47, 128, 52, 147, 48, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 16, 238, 198, 58, 0, 187, 219, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 144, 206, 111, 64, 128, 155, 132, 65, 144, 176, 79, 66, 128, 125, 100, 67, 144, 146, 47, 68, 128, 95, 68, 69, 16, 197, 243, 69, 0, 124, 45, 71, 60, 0, 0, 0, 1, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 32, 28, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 144, 157, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 22, 54, 0, 224, 83, 54, 0, 16, 99, 54, 0, 224, 60, 0, 0, 0, 128, 161, 108, 204, 224, 251, 96, 210, 80, 62, 47, 247, 208, 105, 40, 248, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 112, 10, 242, 26, 96, 237, 225, 27, 112, 236, 209, 28, 96, 207, 193, 29, 112, 206, 177, 30, 96, 177, 161, 31, 240, 0, 118, 32, 96, 147, 129, 33, 240, 226, 85, 34, 224, 175, 106, 35, 240, 196, 53, 36, 224, 145, 74, 37, 240, 166, 21, 38, 224, 115, 42, 39, 112, 195, 254, 39, 224, 85, 10, 41, 112, 165, 222, 41, 224, 55, 234, 42, 112, 135, 190, 43, 96, 84, 211, 44, 112, 105, 158, 45, 96, 54, 179, 46, 112, 75, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 170, 170, 170, 170, 170, 170, 170, 170, 60, 0, 0, 0, 4, 3, 5, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 32, 28, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 176, 185, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 120, 54, 0, 224, 183, 54, 0, 16, 199, 54, 0, 224, 2, 0, 0, 0, 142, 0, 0, 0, 17, 2, 0, 0, 21, 0, 0, 0, 254, 180, 15, 147, 112, 25, 141, 7, 96, 164, 16, 9, 240, 148, 173, 9, 96, 134, 240, 10, 112, 133, 224, 11, 224, 162, 217, 12, 112, 103, 192, 13, 224, 132, 185, 14, 240, 131, 169, 15, 224, 102, 153, 16, 240, 101, 137, 17, 224, 72, 121, 18, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 21, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 2, 184, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 222, 54, 0, 224, 225, 54, 0, 224, 247, 54, 0, 16, 254, 54, 0, 224, 80, 0, 0, 0, 160, 26, 137, 203, 16, 38, 97, 210, 32, 71, 184, 254, 16, 42, 168, 255, 32, 41, 152, 0, 16, 12, 136, 1, 32, 11, 120, 2, 144, 40, 113, 3, 160, 39, 97, 4, 144, 10, 81, 5, 160, 9, 65, 6, 144, 236, 48, 7, 160, 67, 141, 7, 144, 206, 16, 9, 32, 191, 173, 9, 144, 176, 240, 10, 160, 175, 224, 11, 16, 205, 217, 12, 160, 145, 192, 13, 16, 175, 185, 14, 32, 174, 169, 15, 16, 145, 153, 16, 32, 144, 137, 17, 16, 115, 121, 18, 32, 114, 105, 19, 32, 99, 89, 20, 32, 84, 73, 21, 16, 55, 57, 22, 32, 54, 41, 23, 144, 83, 34, 24, 32, 24, 9, 25, 144, 53, 2, 26, 176, 66, 242, 26, 160, 37, 226, 27, 176, 36, 210, 28, 160, 7, 194, 29, 176, 6, 178, 30, 160, 233, 161, 31, 48, 57, 118, 32, 160, 203, 129, 33, 48, 27, 86, 34, 32, 232, 106, 35, 48, 253, 53, 36, 32, 202, 74, 37, 48, 223, 21, 38, 32, 172, 42, 39, 176, 251, 254, 39, 32, 142, 10, 41, 176, 221, 222, 41, 32, 112, 234, 42, 176, 191, 190, 43, 160, 140, 211, 44, 176, 161, 158, 45, 160, 110, 179, 46, 176, 131, 126, 47, 160, 80, 147, 48, 48, 160, 103, 49, 160, 50, 115, 50, 48, 130, 71, 51, 160, 20, 83, 52, 48, 100, 39, 53, 160, 246, 50, 54, 48, 70, 7, 55, 32, 19, 28, 56, 48, 40, 231, 56, 32, 245, 251, 57, 48, 10, 199, 58, 32, 215, 219, 59, 176, 38, 176, 60, 32, 185, 187, 61, 176, 8, 144, 62, 32, 155, 155, 63, 176, 234, 111, 64, 160, 183, 132, 65, 176, 204, 79, 66, 160, 153, 100, 67, 176, 174, 47, 68, 160, 123, 68, 69, 48, 225, 243, 69, 32, 152, 45, 71, 4, 0, 0, 0, 255, 255, 255, 255, 209, 253, 194, 63, 255, 255, 255, 255, 197, 50, 135, 125, 170, 170, 170, 170, 82, 0, 0, 0, 3, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 2, 4, 5, 4, 5, 4, 5, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 12, 0, 0, 0, 123, 211, 0, 0, 0, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 16, 14, 0, 0, 251, 129, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 112, 129, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 12, 55, 0, 224, 93, 55, 0, 224, 99, 55, 0, 16, 121, 55, 0, 224, 2, 0, 0, 0, 145, 0, 0, 0, 152, 0, 0, 0, 116, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 0, 247, 115, 164, 112, 17, 22, 165, 128, 78, 13, 202, 112, 71, 216, 202, 128, 254, 136, 203, 240, 9, 97, 210, 28, 215, 117, 211, 112, 9, 164, 211, 128, 181, 254, 218, 240, 115, 192, 219, 128, 151, 222, 220, 112, 144, 169, 221, 128, 121, 190, 222, 112, 114, 137, 223, 128, 91, 158, 224, 112, 84, 105, 225, 128, 61, 126, 226, 112, 54, 73, 227, 128, 31, 94, 228, 112, 24, 41, 229, 0, 60, 71, 230, 240, 30, 55, 231, 0, 30, 39, 232, 240, 0, 23, 233, 0, 0, 7, 234, 240, 226, 246, 234, 0, 226, 230, 235, 240, 196, 214, 236, 0, 196, 198, 237, 112, 225, 191, 238, 128, 224, 175, 239, 112, 144, 30, 240, 240, 58, 216, 252, 224, 29, 200, 253, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 240, 224, 119, 2, 96, 254, 112, 3, 112, 253, 96, 4, 96, 224, 80, 5, 112, 223, 64, 6, 96, 194, 48, 7, 112, 25, 141, 7, 112, 178, 16, 9, 240, 148, 173, 9, 96, 134, 240, 10, 112, 133, 224, 11, 224, 162, 217, 12, 112, 103, 192, 13, 224, 132, 185, 14, 240, 131, 169, 15, 224, 102, 153, 16, 240, 101, 137, 17, 224, 72, 121, 18, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 112, 10, 242, 26, 96, 237, 225, 27, 112, 236, 209, 28, 96, 207, 193, 29, 112, 206, 177, 30, 96, 177, 161, 31, 240, 0, 118, 32, 96, 147, 129, 33, 240, 226, 85, 34, 224, 175, 106, 35, 240, 196, 53, 36, 224, 145, 74, 37, 240, 166, 21, 38, 224, 115, 42, 39, 112, 195, 254, 39, 224, 85, 10, 41, 112, 165, 222, 41, 224, 55, 234, 42, 112, 135, 190, 43, 96, 84, 211, 44, 112, 105, 158, 45, 96, 54, 179, 46, 112, 75, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 240, 239, 230, 56, 224, 188, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 117, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 170, 10, 0, 0, 0, 154, 175, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 176, 185, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 145, 55, 0, 224, 148, 55, 0, 224, 9, 56, 0, 224, 15, 56, 0, 16, 46, 56, 0, 224, 86, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 254, 136, 203, 240, 9, 97, 210, 0, 73, 216, 252, 240, 43, 200, 253, 0, 43, 184, 254, 240, 13, 168, 255, 0, 13, 152, 0, 240, 239, 135, 1, 0, 239, 119, 2, 112, 12, 113, 3, 128, 11, 97, 4, 112, 238, 80, 5, 128, 237, 64, 6, 112, 208, 48, 7, 128, 39, 141, 7, 112, 178, 16, 9, 0, 163, 173, 9, 112, 148, 240, 10, 128, 147, 224, 11, 240, 176, 217, 12, 128, 117, 192, 13, 240, 146, 185, 14, 0, 146, 169, 15, 240, 116, 153, 16, 0, 116, 137, 17, 240, 86, 121, 18, 0, 86, 105, 19, 240, 56, 89, 20, 0, 56, 73, 21, 240, 26, 57, 22, 0, 26, 41, 23, 112, 55, 34, 24, 0, 252, 8, 25, 112, 25, 2, 26, 128, 24, 242, 26, 112, 251, 225, 27, 128, 250, 209, 28, 112, 221, 193, 29, 128, 220, 177, 30, 112, 191, 161, 31, 0, 15, 118, 32, 112, 161, 129, 33, 0, 241, 85, 34, 240, 189, 106, 35, 0, 211, 53, 36, 240, 159, 74, 37, 0, 181, 21, 38, 240, 129, 42, 39, 128, 209, 254, 39, 240, 99, 10, 41, 128, 179, 222, 41, 240, 69, 234, 42, 128, 149, 190, 43, 112, 98, 211, 44, 128, 119, 158, 45, 112, 68, 179, 46, 128, 89, 126, 47, 112, 38, 147, 48, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 160, 254, 3, 94, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 87, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 10, 0, 0, 0, 116, 176, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 70, 56, 0, 224, 157, 56, 0, 224, 163, 56, 0, 16, 186, 56, 0, 224, 2, 0, 0, 0, 228, 150, 30, 184, 212, 213, 238, 184, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 28, 192, 255, 255, 0, 0, 0, 0, 28, 192, 255, 255, 16, 14, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 208, 56, 0, 224, 211, 56, 0, 16, 213, 56, 0, 224, 15, 0, 0, 0, 212, 64, 116, 140, 80, 74, 207, 195, 64, 227, 69, 196, 208, 74, 47, 197, 192, 45, 31, 198, 208, 44, 15, 199, 192, 15, 255, 199, 80, 196, 24, 30, 64, 93, 143, 30, 208, 247, 249, 31, 192, 144, 112, 32, 208, 227, 158, 37, 192, 124, 21, 38, 80, 3, 37, 45, 64, 156, 155, 45, 2, 0, 0, 0, 255, 255, 255, 255, 188, 35, 135, 105, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 16, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 8, 0, 0, 0, 196, 183, 255, 255, 0, 0, 0, 0, 172, 183, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 225, 56, 0, 224, 241, 56, 0, 224, 247, 56, 0, 16, 252, 56, 0, 224, 4, 0, 0, 0, 151, 0, 0, 0, 37, 2, 0, 0, 113, 2, 0, 0, 114, 2, 0, 0, 124, 0, 0, 0, 160, 72, 166, 158, 144, 21, 187, 159, 160, 42, 134, 160, 144, 247, 154, 161, 160, 26, 137, 203, 16, 38, 97, 210, 92, 116, 254, 214, 144, 173, 128, 216, 144, 195, 254, 218, 16, 144, 192, 219, 144, 165, 222, 220, 144, 172, 169, 221, 144, 135, 190, 222, 144, 142, 137, 223, 144, 105, 158, 224, 144, 112, 105, 225, 144, 75, 126, 226, 144, 82, 73, 227, 144, 45, 94, 228, 144, 52, 41, 229, 16, 74, 71, 230, 16, 81, 18, 231, 16, 44, 39, 232, 16, 51, 242, 232, 16, 14, 7, 234, 16, 21, 210, 234, 16, 240, 230, 235, 16, 247, 177, 236, 16, 210, 198, 237, 16, 217, 145, 238, 144, 238, 175, 239, 16, 187, 113, 240, 144, 208, 143, 241, 144, 193, 127, 242, 144, 178, 111, 243, 144, 163, 95, 244, 144, 148, 79, 245, 144, 133, 63, 246, 144, 118, 47, 247, 16, 162, 40, 248, 144, 88, 15, 249, 16, 132, 8, 250, 32, 131, 248, 250, 16, 102, 232, 251, 32, 101, 216, 252, 16, 72, 200, 253, 32, 71, 184, 254, 16, 42, 168, 255, 32, 41, 152, 0, 16, 12, 136, 1, 32, 11, 120, 2, 144, 40, 113, 3, 160, 39, 97, 4, 144, 10, 81, 5, 160, 9, 65, 6, 144, 236, 48, 7, 160, 67, 141, 7, 144, 206, 16, 9, 32, 191, 173, 9, 144, 176, 240, 10, 160, 175, 224, 11, 16, 205, 217, 12, 160, 145, 192, 13, 16, 175, 185, 14, 32, 174, 169, 15, 16, 145, 153, 16, 32, 144, 137, 17, 16, 115, 121, 18, 32, 114, 105, 19, 16, 85, 89, 20, 32, 84, 73, 21, 16, 55, 57, 22, 32, 54, 41, 23, 144, 83, 34, 24, 32, 24, 9, 25, 144, 53, 2, 26, 160, 52, 242, 26, 144, 23, 226, 27, 160, 22, 210, 28, 144, 249, 193, 29, 160, 248, 177, 30, 144, 219, 161, 31, 32, 43, 118, 32, 144, 189, 129, 33, 32, 13, 86, 34, 16, 218, 106, 35, 32, 239, 53, 36, 16, 188, 74, 37, 32, 209, 21, 38, 16, 158, 42, 39, 160, 237, 254, 39, 16, 128, 10, 41, 160, 207, 222, 41, 16, 98, 234, 42, 160, 177, 190, 43, 144, 126, 211, 44, 160, 147, 158, 45, 144, 96, 179, 46, 160, 117, 126, 47, 144, 66, 147, 48, 32, 146, 103, 49, 144, 36, 115, 50, 32, 116, 71, 51, 144, 6, 83, 52, 32, 86, 39, 53, 144, 232, 50, 54, 32, 56, 7, 55, 16, 5, 28, 56, 32, 26, 231, 56, 16, 231, 251, 57, 32, 252, 198, 58, 16, 201, 219, 59, 160, 24, 176, 60, 16, 171, 187, 61, 160, 250, 143, 62, 16, 141, 155, 63, 160, 220, 111, 64, 144, 169, 132, 65, 160, 190, 79, 66, 144, 139, 100, 67, 160, 160, 47, 68, 144, 109, 68, 69, 32, 211, 243, 69, 16, 138, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 192, 26, 4, 94, 170, 170, 170, 170, 170, 170, 170, 170, 125, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 38, 145, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 128, 143, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 12, 57, 0, 224, 17, 57, 0, 224, 142, 57, 0, 224, 147, 57, 0, 16, 180, 57, 0, 224, 41, 0, 0, 0, 124, 104, 170, 150, 224, 73, 15, 184, 160, 64, 253, 184, 48, 52, 241, 185, 32, 116, 222, 186, 48, 174, 56, 218, 48, 250, 235, 218, 176, 225, 25, 220, 32, 89, 185, 220, 48, 21, 251, 221, 32, 222, 155, 222, 48, 154, 221, 223, 32, 51, 84, 224, 176, 255, 151, 244, 32, 94, 5, 245, 48, 100, 192, 246, 160, 30, 14, 247, 48, 44, 81, 248, 32, 197, 199, 248, 176, 210, 10, 250, 160, 248, 168, 250, 48, 6, 236, 251, 160, 125, 139, 252, 48, 142, 201, 29, 160, 215, 120, 30, 176, 53, 160, 31, 160, 207, 51, 32, 48, 105, 129, 33, 160, 200, 11, 34, 176, 16, 88, 35, 32, 112, 226, 35, 176, 242, 55, 37, 32, 199, 212, 37, 48, 121, 128, 48, 160, 77, 29, 49, 176, 198, 246, 55, 32, 133, 184, 56, 48, 227, 223, 57, 32, 74, 242, 57, 176, 255, 200, 59, 160, 14, 111, 60, 170, 170, 170, 170, 41, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 132, 222, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 200, 57, 0, 224, 243, 57, 0, 16, 255, 57, 0, 224, 15, 0, 0, 0, 232, 72, 45, 189, 96, 116, 67, 6, 80, 62, 164, 9, 224, 248, 81, 17, 80, 111, 212, 17, 224, 218, 49, 19, 80, 81, 180, 19, 32, 145, 97, 41, 80, 75, 193, 42, 224, 221, 67, 43, 80, 239, 201, 50, 224, 192, 88, 66, 80, 105, 63, 67, 128, 110, 84, 68, 96, 89, 31, 69, 2, 0, 0, 0, 255, 255, 255, 255, 100, 44, 135, 105, 170, 170, 170, 170, 16, 0, 0, 0, 3, 1, 4, 1, 2, 1, 2, 1, 4, 1, 4, 1, 2, 1, 2, 1, 10, 0, 0, 0, 28, 175, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 24, 175, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 11, 58, 0, 224, 27, 58, 0, 224, 31, 58, 0, 16, 36, 58, 0, 224, 2, 0, 0, 0, 156, 0, 0, 0, 121, 1, 0, 0, 31, 0, 0, 0, 68, 127, 170, 150, 240, 87, 15, 184, 176, 78, 253, 184, 64, 66, 241, 185, 48, 130, 222, 186, 64, 188, 56, 218, 64, 8, 236, 218, 192, 239, 25, 220, 48, 103, 185, 220, 64, 35, 251, 221, 48, 236, 155, 222, 64, 168, 221, 223, 48, 65, 84, 224, 192, 13, 152, 244, 48, 108, 5, 245, 64, 114, 192, 246, 176, 44, 14, 247, 64, 58, 81, 248, 48, 211, 199, 248, 192, 224, 10, 250, 176, 6, 169, 250, 64, 20, 236, 251, 176, 139, 139, 252, 64, 156, 201, 29, 176, 229, 120, 30, 192, 67, 160, 31, 176, 221, 51, 32, 64, 119, 129, 33, 176, 214, 11, 34, 64, 195, 192, 44, 48, 210, 102, 45, 170, 170, 170, 170, 170, 170, 170, 170, 31, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 188, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 54, 58, 0, 224, 57, 58, 0, 224, 91, 58, 0, 16, 100, 58, 0, 224, 3, 0, 0, 0, 68, 200, 163, 145, 64, 110, 77, 19, 176, 22, 52, 20, 170, 170, 170, 170, 3, 0, 0, 0, 1, 2, 1, 170, 6, 0, 0, 0, 188, 198, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 114, 58, 0, 224, 119, 58, 0, 16, 121, 58, 0, 224, 33, 0, 0, 0, 96, 218, 182, 165, 0, 241, 85, 34, 240, 189, 106, 35, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 128, 4, 245, 58, 240, 194, 182, 59, 128, 252, 175, 60, 240, 142, 187, 61, 128, 222, 143, 62, 240, 112, 155, 63, 128, 192, 111, 64, 112, 141, 132, 65, 128, 162, 79, 66, 112, 111, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 128, 102, 15, 70, 112, 51, 36, 71, 0, 131, 248, 71, 112, 21, 4, 73, 0, 101, 216, 73, 112, 247, 227, 74, 128, 151, 156, 75, 112, 78, 214, 76, 33, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 64, 162, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 160, 171, 255, 127, 1, 0, 0, 96, 219, 7, 0, 112, 133, 58, 0, 224, 167, 58, 0, 16, 177, 58, 0, 224, 2, 0, 0, 0, 160, 0, 0, 0, 26, 2, 0, 0, 23, 0, 0, 0, 112, 232, 182, 165, 224, 110, 242, 175, 96, 86, 102, 182, 96, 210, 67, 183, 96, 54, 12, 184, 240, 134, 253, 184, 96, 113, 234, 203, 240, 180, 145, 216, 128, 112, 0, 0, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 144, 18, 245, 58, 0, 209, 182, 59, 144, 10, 176, 60, 0, 157, 187, 61, 170, 170, 170, 170, 170, 170, 170, 170, 23, 0, 0, 0, 2, 4, 2, 4, 2, 4, 2, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 10, 0, 0, 0, 60, 156, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 144, 157, 255, 127, 176, 3, 0, 96, 211, 7, 0, 112, 194, 58, 0, 224, 197, 58, 0, 224, 223, 58, 0, 16, 230, 58, 0, 224, 81, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 254, 136, 203, 240, 9, 97, 210, 0, 243, 117, 211, 240, 235, 64, 212, 128, 74, 15, 249, 240, 103, 8, 250, 0, 43, 184, 254, 112, 223, 64, 6, 112, 208, 48, 7, 128, 39, 141, 7, 112, 178, 16, 9, 0, 163, 173, 9, 112, 148, 240, 10, 128, 147, 224, 11, 240, 176, 217, 12, 128, 117, 192, 13, 240, 146, 185, 14, 0, 146, 169, 15, 240, 116, 153, 16, 0, 116, 137, 17, 240, 86, 121, 18, 0, 86, 105, 19, 240, 56, 89, 20, 0, 56, 73, 21, 240, 26, 57, 22, 0, 26, 41, 23, 112, 55, 34, 24, 0, 252, 8, 25, 112, 25, 2, 26, 128, 24, 242, 26, 112, 251, 225, 27, 128, 250, 209, 28, 112, 221, 193, 29, 128, 220, 177, 30, 112, 191, 161, 31, 0, 15, 118, 32, 112, 161, 129, 33, 0, 241, 85, 34, 240, 189, 106, 35, 0, 211, 53, 36, 240, 159, 74, 37, 0, 181, 21, 38, 240, 129, 42, 39, 128, 209, 254, 39, 240, 99, 10, 41, 128, 179, 222, 41, 240, 69, 234, 42, 128, 149, 190, 43, 112, 98, 211, 44, 128, 119, 158, 45, 112, 68, 179, 46, 128, 89, 126, 47, 112, 38, 147, 48, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 0, 224, 198, 58, 240, 172, 219, 59, 128, 252, 175, 60, 240, 142, 187, 61, 128, 222, 143, 62, 240, 112, 155, 63, 128, 192, 111, 64, 112, 141, 132, 65, 128, 162, 79, 66, 112, 111, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 99, 73, 119, 97, 170, 170, 170, 170, 170, 170, 170, 170, 82, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 8, 0, 0, 0, 221, 173, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 160, 171, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 252, 58, 0, 224, 78, 59, 0, 224, 83, 59, 0, 16, 105, 59, 0, 224, 17, 0, 0, 0, 96, 218, 182, 165, 96, 213, 134, 22, 80, 75, 76, 24, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 128, 4, 245, 58, 240, 194, 182, 59, 128, 252, 175, 60, 240, 142, 187, 61, 17, 0, 0, 0, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 8, 0, 0, 0, 252, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 160, 171, 255, 127, 176, 3, 0, 96, 211, 7, 0, 112, 125, 59, 0, 224, 143, 59, 0, 16, 149, 59, 0, 224, 42, 0, 0, 0, 160, 26, 137, 203, 16, 38, 97, 210, 32, 71, 184, 254, 16, 42, 168, 255, 32, 41, 152, 0, 16, 12, 136, 1, 32, 11, 120, 2, 144, 40, 113, 3, 160, 39, 97, 4, 144, 10, 81, 5, 160, 9, 65, 6, 144, 236, 48, 7, 160, 67, 141, 7, 144, 206, 16, 9, 32, 191, 173, 9, 144, 176, 240, 10, 160, 175, 224, 11, 16, 205, 217, 12, 160, 145, 192, 13, 16, 175, 185, 14, 32, 174, 169, 15, 16, 145, 153, 16, 32, 144, 137, 17, 16, 115, 121, 18, 32, 114, 105, 19, 16, 85, 89, 20, 32, 84, 73, 21, 16, 55, 57, 22, 32, 54, 41, 23, 144, 83, 34, 24, 32, 24, 9, 25, 144, 53, 2, 26, 160, 226, 53, 86, 48, 72, 229, 86, 32, 255, 30, 88, 48, 42, 197, 88, 32, 225, 254, 89, 48, 12, 165, 90, 32, 195, 222, 91, 160, 70, 68, 92, 48, 238, 132, 92, 32, 165, 190, 93, 4, 0, 0, 0, 255, 255, 255, 255, 209, 253, 194, 63, 255, 255, 255, 255, 26, 48, 135, 125, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 44, 0, 0, 0, 3, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 1, 2, 1, 2, 1, 2, 4, 1, 2, 1, 12, 0, 0, 0, 38, 214, 0, 0, 0, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 16, 14, 0, 0, 166, 132, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 112, 129, 255, 127, 1, 0, 0, 96, 228, 7, 0, 112, 168, 59, 0, 224, 211, 59, 0, 224, 219, 59, 0, 16, 231, 59, 0, 224, 2, 0, 0, 0, 165, 0, 0, 0, 27, 2, 0, 0, 28, 0, 0, 0, 112, 232, 182, 165, 224, 110, 242, 175, 96, 86, 102, 182, 96, 210, 67, 183, 96, 54, 12, 184, 240, 134, 253, 184, 96, 176, 222, 197, 80, 52, 151, 198, 224, 241, 85, 201, 80, 221, 234, 201, 224, 198, 2, 207, 80, 86, 183, 207, 224, 21, 153, 218, 208, 131, 118, 219, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 128, 4, 245, 58, 240, 194, 182, 59, 128, 252, 175, 60, 240, 142, 187, 61, 28, 0, 0, 0, 1, 2, 1, 2, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 8, 0, 0, 0, 12, 163, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 160, 171, 255, 127, 176, 3, 0, 96, 211, 7, 0, 112, 255, 59, 0, 224, 2, 60, 0, 224, 31, 60, 0, 16, 39, 60, 0, 224, 44, 0, 0, 0, 168, 56, 182, 145, 192, 99, 110, 19, 208, 228, 117, 32, 64, 119, 129, 33, 208, 198, 85, 34, 192, 147, 106, 35, 208, 168, 53, 36, 192, 117, 74, 37, 208, 138, 21, 38, 192, 87, 42, 39, 80, 167, 254, 39, 192, 57, 10, 41, 80, 137, 222, 41, 192, 27, 234, 42, 80, 107, 190, 43, 64, 56, 211, 44, 80, 77, 158, 45, 64, 26, 179, 46, 80, 47, 126, 47, 64, 252, 146, 48, 208, 75, 103, 49, 64, 222, 114, 50, 208, 45, 71, 51, 64, 192, 82, 52, 208, 15, 39, 53, 64, 162, 50, 54, 208, 241, 6, 55, 192, 190, 27, 56, 208, 211, 230, 56, 192, 160, 251, 57, 208, 181, 198, 58, 192, 130, 219, 59, 80, 210, 175, 60, 192, 100, 187, 61, 80, 180, 143, 62, 192, 70, 155, 63, 80, 150, 111, 64, 64, 99, 132, 65, 80, 120, 79, 66, 64, 69, 100, 67, 80, 90, 47, 68, 64, 39, 68, 69, 208, 140, 243, 69, 192, 67, 45, 71, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 44, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 8, 0, 0, 0, 88, 203, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 208, 213, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 59, 60, 0, 224, 107, 60, 0, 16, 119, 60, 0, 224, 145, 0, 0, 0, 80, 182, 241, 128, 96, 133, 184, 158, 80, 221, 186, 159, 208, 56, 60, 187, 64, 35, 180, 187, 208, 26, 28, 189, 64, 5, 148, 189, 208, 252, 251, 190, 64, 231, 115, 191, 208, 222, 219, 192, 64, 201, 83, 193, 208, 192, 187, 194, 64, 171, 51, 195, 208, 162, 155, 196, 64, 141, 19, 197, 208, 248, 112, 198, 64, 205, 13, 199, 208, 241, 72, 200, 64, 175, 237, 200, 208, 94, 22, 202, 192, 203, 214, 202, 96, 226, 136, 203, 208, 237, 96, 210, 224, 214, 117, 211, 208, 207, 64, 212, 224, 184, 85, 213, 208, 177, 32, 214, 224, 154, 53, 215, 208, 147, 0, 216, 224, 124, 21, 217, 208, 117, 224, 217, 96, 153, 254, 218, 208, 87, 192, 219, 96, 123, 222, 220, 80, 116, 169, 221, 96, 93, 190, 222, 80, 86, 137, 223, 96, 63, 158, 224, 80, 56, 105, 225, 96, 33, 126, 226, 80, 26, 73, 227, 96, 3, 94, 228, 80, 252, 40, 229, 224, 31, 71, 230, 208, 24, 18, 231, 224, 1, 39, 232, 208, 228, 22, 233, 224, 227, 6, 234, 208, 198, 246, 234, 224, 197, 230, 235, 208, 168, 214, 236, 224, 167, 198, 237, 80, 197, 191, 238, 96, 196, 175, 239, 80, 167, 159, 240, 96, 166, 143, 241, 80, 137, 127, 242, 96, 136, 111, 243, 80, 107, 95, 244, 96, 106, 79, 245, 80, 77, 63, 246, 96, 76, 47, 247, 208, 105, 40, 248, 96, 46, 15, 249, 208, 75, 8, 250, 224, 74, 248, 250, 208, 45, 232, 251, 224, 44, 216, 252, 208, 15, 200, 253, 224, 14, 184, 254, 208, 241, 167, 255, 224, 240, 151, 0, 208, 211, 135, 1, 224, 210, 119, 2, 80, 240, 112, 3, 96, 239, 96, 4, 80, 210, 80, 5, 96, 179, 32, 8, 80, 150, 16, 9, 96, 149, 0, 10, 80, 120, 240, 10, 96, 119, 224, 11, 208, 148, 217, 12, 96, 89, 192, 13, 208, 118, 185, 14, 224, 117, 169, 15, 208, 88, 153, 16, 224, 87, 137, 17, 208, 58, 121, 18, 224, 57, 105, 19, 208, 28, 89, 20, 224, 27, 73, 21, 208, 254, 56, 22, 224, 253, 40, 23, 80, 27, 34, 24, 224, 223, 8, 25, 80, 253, 1, 26, 96, 252, 241, 26, 80, 223, 225, 27, 96, 222, 209, 28, 80, 193, 193, 29, 96, 192, 177, 30, 80, 163, 161, 31, 224, 242, 117, 32, 80, 133, 129, 33, 224, 212, 85, 34, 208, 161, 106, 35, 224, 182, 53, 36, 208, 131, 74, 37, 224, 152, 21, 38, 208, 101, 42, 39, 96, 181, 254, 39, 208, 71, 10, 41, 96, 151, 222, 41, 208, 41, 234, 42, 124, 93, 190, 43, 108, 42, 211, 44, 124, 63, 158, 45, 108, 12, 179, 46, 124, 33, 126, 47, 108, 238, 146, 48, 252, 61, 103, 49, 108, 208, 114, 50, 252, 31, 71, 51, 108, 178, 82, 52, 252, 1, 39, 53, 108, 148, 50, 54, 252, 227, 6, 55, 236, 176, 27, 56, 252, 197, 230, 56, 236, 146, 251, 57, 252, 167, 198, 58, 236, 116, 219, 59, 124, 196, 175, 60, 236, 86, 187, 61, 124, 166, 143, 62, 236, 56, 155, 63, 124, 136, 111, 64, 108, 85, 132, 65, 124, 106, 79, 66, 108, 55, 100, 67, 124, 76, 47, 68, 108, 25, 68, 69, 224, 154, 243, 69, 208, 81, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 188, 237, 30, 94, 146, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 68, 195, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 192, 199, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 138, 60, 0, 224, 28, 61, 0, 224, 31, 61, 0, 16, 69, 61, 0, 224, 17, 0, 0, 0, 96, 218, 182, 165, 0, 241, 85, 34, 240, 189, 106, 35, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 128, 4, 245, 58, 240, 194, 182, 59, 128, 252, 175, 60, 240, 142, 187, 61, 17, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 244, 161, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 160, 171, 255, 127, 176, 3, 0, 96, 211, 7, 0, 112, 89, 61, 0, 224, 107, 61, 0, 16, 113, 61, 0, 224, 85, 0, 0, 0, 179, 135, 146, 162, 64, 219, 255, 168, 176, 15, 241, 169, 56, 89, 226, 170, 48, 67, 210, 171, 184, 140, 195, 172, 176, 118, 179, 173, 184, 181, 244, 187, 176, 181, 191, 188, 184, 151, 212, 189, 176, 151, 159, 190, 184, 121, 180, 191, 176, 121, 127, 192, 184, 91, 148, 193, 176, 91, 95, 194, 56, 120, 125, 195, 176, 61, 63, 196, 56, 90, 93, 197, 176, 31, 31, 198, 56, 82, 24, 199, 48, 60, 8, 200, 56, 30, 29, 201, 48, 30, 232, 201, 56, 159, 139, 202, 48, 198, 30, 205, 40, 102, 149, 205, 176, 133, 11, 236, 40, 53, 242, 236, 176, 74, 69, 237, 32, 214, 133, 237, 176, 114, 19, 247, 32, 27, 250, 247, 48, 62, 254, 252, 40, 17, 246, 253, 48, 117, 150, 0, 32, 82, 216, 0, 176, 138, 87, 4, 160, 58, 198, 4, 176, 27, 150, 7, 152, 218, 223, 7, 40, 159, 198, 8, 48, 78, 90, 9, 32, 115, 219, 9, 48, 18, 26, 13, 160, 135, 127, 13, 48, 127, 231, 14, 160, 105, 95, 15, 48, 214, 217, 16, 160, 75, 63, 17, 176, 45, 137, 17, 160, 162, 49, 19, 48, 84, 195, 33, 32, 120, 39, 34, 176, 228, 161, 35, 160, 148, 16, 36, 176, 103, 74, 37, 32, 60, 231, 37, 48, 15, 33, 39, 160, 88, 208, 39, 176, 43, 10, 41, 160, 58, 176, 41, 48, 211, 224, 42, 160, 28, 144, 43, 48, 246, 76, 65, 192, 47, 70, 66, 208, 163, 72, 67, 192, 156, 19, 68, 80, 75, 31, 69, 192, 126, 243, 69, 208, 103, 8, 71, 192, 96, 211, 71, 208, 73, 232, 72, 192, 66, 179, 73, 208, 43, 200, 74, 64, 95, 156, 75, 208, 13, 168, 76, 64, 65, 124, 77, 208, 239, 135, 78, 64, 35, 92, 79, 80, 12, 113, 80, 64, 5, 60, 81, 80, 238, 80, 82, 64, 231, 27, 83, 80, 208, 48, 84, 64, 201, 251, 84, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 85, 0, 0, 0, 1, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 5, 4, 5, 4, 6, 4, 6, 4, 5, 4, 6, 4, 6, 4, 7, 5, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 6, 4, 170, 170, 170, 16, 0, 0, 0, 77, 203, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 200, 206, 255, 255, 0, 0, 0, 0, 200, 206, 255, 255, 8, 7, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 8, 7, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 24, 21, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 130, 61, 0, 224, 219, 61, 0, 16, 242, 61, 0, 224, 89, 0, 0, 0, 138, 66, 55, 147, 112, 120, 79, 245, 96, 91, 63, 246, 112, 90, 47, 247, 224, 119, 40, 248, 112, 60, 15, 249, 224, 89, 8, 250, 240, 88, 248, 250, 224, 59, 232, 251, 240, 58, 216, 252, 224, 29, 200, 253, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 240, 224, 119, 2, 96, 254, 112, 3, 112, 253, 96, 4, 96, 224, 80, 5, 112, 223, 64, 6, 96, 194, 48, 7, 112, 193, 32, 8, 96, 164, 16, 9, 112, 163, 0, 10, 96, 134, 240, 10, 112, 133, 224, 11, 224, 162, 217, 12, 112, 103, 192, 13, 224, 132, 185, 14, 240, 131, 169, 15, 224, 102, 153, 16, 240, 101, 137, 17, 224, 72, 121, 18, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 112, 10, 242, 26, 96, 237, 225, 27, 112, 236, 209, 28, 96, 207, 193, 29, 112, 206, 177, 30, 96, 177, 161, 31, 240, 0, 118, 32, 96, 147, 129, 33, 240, 226, 85, 34, 224, 175, 106, 35, 240, 196, 53, 36, 224, 145, 74, 37, 240, 166, 21, 38, 224, 115, 42, 39, 112, 195, 254, 39, 224, 85, 10, 41, 112, 165, 222, 41, 224, 55, 234, 42, 112, 135, 190, 43, 96, 84, 211, 44, 112, 105, 158, 45, 96, 54, 179, 46, 112, 75, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 240, 239, 230, 56, 224, 188, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 170, 170, 170, 170, 89, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 118, 183, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 176, 185, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 8, 62, 0, 224, 99, 62, 0, 16, 123, 62, 0, 224, 2, 0, 0, 0, 173, 0, 0, 0, 108, 2, 0, 0, 174, 0, 0, 0, 112, 30, 166, 158, 96, 235, 186, 159, 112, 0, 134, 160, 96, 205, 154, 161, 112, 226, 101, 162, 224, 233, 131, 163, 112, 174, 106, 164, 96, 167, 53, 165, 240, 202, 83, 166, 96, 137, 21, 167, 240, 172, 51, 168, 224, 165, 254, 168, 240, 142, 19, 170, 224, 135, 222, 170, 240, 112, 243, 171, 224, 105, 190, 172, 240, 82, 211, 173, 224, 75, 158, 174, 240, 52, 179, 175, 224, 45, 126, 176, 112, 81, 156, 177, 96, 74, 103, 178, 112, 51, 124, 179, 96, 44, 71, 180, 112, 21, 92, 181, 96, 14, 39, 182, 112, 247, 59, 183, 96, 240, 6, 184, 112, 217, 27, 185, 96, 210, 230, 185, 240, 245, 4, 187, 96, 180, 198, 187, 240, 215, 228, 188, 224, 208, 175, 189, 240, 185, 196, 190, 224, 178, 143, 191, 240, 155, 164, 192, 224, 148, 111, 193, 240, 125, 132, 194, 224, 118, 79, 195, 240, 95, 100, 196, 224, 88, 47, 197, 112, 124, 77, 198, 224, 58, 15, 199, 112, 94, 45, 200, 96, 87, 248, 200, 112, 64, 13, 202, 96, 57, 216, 202, 112, 240, 136, 203, 224, 251, 96, 210, 240, 228, 117, 211, 224, 221, 64, 212, 240, 198, 85, 213, 224, 191, 32, 214, 240, 168, 53, 215, 224, 161, 0, 216, 240, 138, 21, 217, 224, 131, 224, 217, 112, 167, 254, 218, 224, 101, 192, 219, 112, 137, 222, 220, 96, 130, 169, 221, 112, 107, 190, 222, 96, 100, 137, 223, 112, 77, 158, 224, 96, 70, 105, 225, 112, 47, 126, 226, 96, 40, 73, 227, 112, 17, 94, 228, 224, 46, 87, 229, 240, 45, 71, 230, 224, 16, 55, 231, 240, 15, 39, 232, 224, 242, 22, 233, 240, 241, 6, 234, 224, 212, 246, 234, 240, 211, 230, 235, 224, 182, 214, 236, 240, 181, 198, 237, 96, 211, 191, 238, 112, 210, 175, 239, 96, 181, 159, 240, 112, 180, 143, 241, 96, 151, 127, 242, 112, 150, 111, 243, 96, 121, 95, 244, 112, 120, 79, 245, 96, 91, 63, 246, 112, 90, 47, 247, 224, 119, 40, 248, 112, 60, 15, 249, 224, 89, 8, 250, 240, 88, 248, 250, 224, 59, 232, 251, 240, 58, 216, 252, 224, 29, 200, 253, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 240, 224, 119, 2, 96, 254, 112, 3, 112, 253, 96, 4, 96, 224, 80, 5, 112, 223, 64, 6, 96, 194, 48, 7, 112, 25, 141, 7, 96, 164, 16, 9, 240, 148, 173, 9, 96, 134, 240, 10, 112, 133, 224, 11, 224, 162, 217, 12, 112, 103, 192, 13, 224, 132, 185, 14, 240, 131, 169, 15, 224, 102, 153, 16, 240, 101, 137, 17, 224, 72, 121, 18, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 112, 10, 242, 26, 96, 237, 225, 27, 112, 236, 209, 28, 96, 207, 193, 29, 112, 206, 177, 30, 96, 177, 161, 31, 240, 0, 118, 32, 96, 147, 129, 33, 240, 226, 85, 34, 224, 175, 106, 35, 240, 196, 53, 36, 224, 145, 74, 37, 240, 166, 21, 38, 224, 115, 42, 39, 112, 195, 254, 39, 224, 85, 10, 41, 112, 165, 222, 41, 224, 55, 234, 42, 112, 135, 190, 43, 96, 84, 211, 44, 112, 105, 158, 45, 96, 54, 179, 46, 112, 75, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 240, 239, 230, 56, 224, 188, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 144, 240, 3, 94, 170, 170, 170, 170, 170, 170, 170, 170, 175, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 158, 186, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 176, 185, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 140, 62, 0, 224, 143, 62, 0, 224, 62, 63, 0, 224, 67, 63, 0, 16, 112, 63, 0, 224, 72, 0, 0, 0, 112, 147, 184, 158, 96, 235, 186, 159, 80, 73, 248, 200, 224, 251, 96, 210, 112, 193, 32, 8, 96, 164, 16, 9, 112, 163, 0, 10, 96, 134, 240, 10, 112, 133, 224, 11, 224, 162, 217, 12, 112, 103, 192, 13, 224, 132, 185, 14, 240, 131, 169, 15, 224, 102, 153, 16, 240, 101, 137, 17, 224, 72, 121, 18, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 112, 10, 242, 26, 96, 237, 225, 27, 112, 236, 209, 28, 96, 207, 193, 29, 112, 206, 177, 30, 96, 177, 161, 31, 240, 0, 118, 32, 96, 147, 129, 33, 240, 226, 85, 34, 224, 175, 106, 35, 240, 196, 53, 36, 224, 145, 74, 37, 240, 166, 21, 38, 224, 115, 42, 39, 112, 195, 254, 39, 224, 85, 10, 41, 112, 165, 222, 41, 224, 55, 234, 42, 112, 135, 190, 43, 96, 84, 211, 44, 112, 105, 158, 45, 96, 54, 179, 46, 112, 75, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 240, 239, 230, 56, 224, 188, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 64, 129, 238, 114, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 73, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 64, 173, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 132, 63, 0, 224, 205, 63, 0, 224, 211, 63, 0, 16, 231, 63, 0, 224, 80, 0, 0, 0, 208, 68, 137, 203, 64, 80, 97, 210, 80, 113, 184, 254, 64, 84, 168, 255, 80, 83, 152, 0, 64, 54, 136, 1, 80, 53, 120, 2, 192, 82, 113, 3, 208, 81, 97, 4, 192, 52, 81, 5, 208, 51, 65, 6, 192, 22, 49, 7, 208, 109, 141, 7, 192, 248, 16, 9, 80, 233, 173, 9, 192, 218, 240, 10, 208, 217, 224, 11, 64, 247, 217, 12, 208, 187, 192, 13, 64, 217, 185, 14, 80, 216, 169, 15, 64, 187, 153, 16, 80, 186, 137, 17, 64, 157, 121, 18, 80, 156, 105, 19, 64, 127, 89, 20, 80, 126, 73, 21, 64, 97, 57, 22, 80, 96, 41, 23, 192, 125, 34, 24, 80, 66, 9, 25, 192, 95, 2, 26, 176, 66, 242, 26, 160, 37, 226, 27, 176, 36, 210, 28, 160, 7, 194, 29, 176, 6, 178, 30, 160, 233, 161, 31, 48, 57, 118, 32, 160, 203, 129, 33, 48, 27, 86, 34, 32, 232, 106, 35, 48, 253, 53, 36, 32, 202, 74, 37, 48, 223, 21, 38, 32, 172, 42, 39, 176, 251, 254, 39, 32, 142, 10, 41, 176, 221, 222, 41, 32, 112, 234, 42, 176, 191, 190, 43, 160, 140, 211, 44, 176, 161, 158, 45, 160, 110, 179, 46, 176, 131, 126, 47, 160, 80, 147, 48, 48, 160, 103, 49, 160, 50, 115, 50, 48, 130, 71, 51, 160, 20, 83, 52, 48, 100, 39, 53, 160, 246, 50, 54, 48, 70, 7, 55, 32, 19, 28, 56, 48, 40, 231, 56, 32, 245, 251, 57, 48, 10, 199, 58, 32, 215, 219, 59, 176, 38, 176, 60, 32, 185, 187, 61, 176, 8, 144, 62, 32, 155, 155, 63, 176, 234, 111, 64, 160, 183, 132, 65, 176, 204, 79, 66, 160, 153, 100, 67, 176, 174, 47, 68, 160, 123, 68, 69, 48, 225, 243, 69, 32, 152, 45, 71, 4, 0, 0, 0, 255, 255, 255, 255, 209, 253, 194, 63, 255, 255, 255, 255, 210, 79, 135, 125, 82, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 170, 170, 12, 0, 0, 0, 110, 182, 0, 0, 0, 0, 0, 0, 238, 100, 255, 255, 0, 0, 0, 0, 80, 101, 255, 255, 0, 0, 0, 0, 80, 101, 255, 255, 16, 14, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 112, 129, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 249, 63, 0, 224, 74, 64, 0, 224, 79, 64, 0, 16, 101, 64, 0, 224, 2, 0, 0, 0, 176, 0, 0, 0, 119, 1, 0, 0, 39, 0, 0, 0, 100, 101, 170, 150, 208, 59, 15, 184, 144, 50, 253, 184, 32, 38, 241, 185, 16, 102, 222, 186, 32, 160, 56, 218, 32, 236, 235, 218, 160, 211, 25, 220, 16, 75, 185, 220, 32, 7, 251, 221, 16, 208, 155, 222, 32, 140, 221, 223, 16, 37, 84, 224, 160, 241, 151, 244, 16, 80, 5, 245, 32, 86, 192, 246, 144, 16, 14, 247, 32, 30, 81, 248, 16, 183, 199, 248, 160, 196, 10, 250, 144, 234, 168, 250, 32, 248, 235, 251, 144, 111, 139, 252, 32, 128, 201, 29, 144, 201, 120, 30, 160, 39, 160, 31, 144, 193, 51, 32, 32, 91, 129, 33, 144, 186, 11, 34, 160, 2, 88, 35, 16, 98, 226, 35, 160, 228, 55, 37, 16, 185, 212, 37, 160, 184, 246, 55, 16, 119, 184, 56, 32, 213, 223, 57, 144, 1, 233, 57, 160, 241, 200, 59, 144, 0, 111, 60, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 39, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 156, 225, 255, 255, 0, 0, 0, 0, 224, 227, 255, 255, 0, 0, 0, 0, 224, 227, 255, 255, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 125, 64, 0, 224, 128, 64, 0, 224, 171, 64, 0, 16, 182, 64, 0, 224, 96, 0, 0, 0, 144, 58, 166, 158, 128, 7, 187, 159, 144, 28, 134, 160, 128, 233, 154, 161, 144, 12, 137, 203, 0, 24, 97, 210, 16, 117, 248, 250, 0, 88, 232, 251, 16, 87, 216, 252, 0, 58, 200, 253, 16, 57, 184, 254, 0, 28, 168, 255, 16, 27, 152, 0, 0, 254, 135, 1, 16, 253, 119, 2, 128, 26, 113, 3, 144, 25, 97, 4, 128, 252, 80, 5, 144, 251, 64, 6, 128, 222, 48, 7, 144, 53, 141, 7, 128, 192, 16, 9, 16, 177, 173, 9, 128, 162, 240, 10, 144, 161, 224, 11, 0, 191, 217, 12, 144, 131, 192, 13, 0, 161, 185, 14, 16, 160, 169, 15, 0, 131, 153, 16, 16, 130, 137, 17, 0, 101, 121, 18, 16, 100, 105, 19, 0, 71, 89, 20, 16, 70, 73, 21, 0, 41, 57, 22, 16, 40, 41, 23, 128, 69, 34, 24, 16, 10, 9, 25, 128, 39, 2, 26, 144, 38, 242, 26, 128, 9, 226, 27, 144, 8, 210, 28, 128, 235, 193, 29, 144, 234, 177, 30, 128, 205, 161, 31, 16, 29, 118, 32, 128, 175, 129, 33, 16, 255, 85, 34, 0, 204, 106, 35, 16, 225, 53, 36, 0, 174, 74, 37, 16, 195, 21, 38, 0, 144, 42, 39, 144, 223, 254, 39, 0, 114, 10, 41, 144, 193, 222, 41, 0, 84, 234, 42, 144, 163, 190, 43, 128, 112, 211, 44, 144, 133, 158, 45, 128, 82, 179, 46, 144, 103, 126, 47, 128, 52, 147, 48, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 16, 238, 198, 58, 0, 187, 219, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 144, 206, 111, 64, 128, 155, 132, 65, 144, 176, 79, 66, 128, 125, 100, 67, 144, 146, 47, 68, 128, 95, 68, 69, 16, 197, 243, 69, 0, 124, 45, 71, 16, 167, 211, 71, 0, 94, 13, 73, 16, 137, 179, 73, 0, 64, 237, 74, 144, 165, 156, 75, 128, 92, 214, 76, 128, 121, 124, 77, 112, 48, 182, 78, 2, 0, 0, 0, 255, 255, 255, 255, 176, 12, 4, 94, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 97, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 4, 3, 170, 170, 170, 10, 0, 0, 0, 149, 160, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 160, 171, 255, 127, 1, 0, 0, 96, 220, 7, 0, 112, 196, 64, 0, 224, 37, 65, 0, 224, 43, 65, 0, 16, 69, 65, 0, 224, 88, 0, 0, 0, 144, 58, 166, 158, 128, 7, 187, 159, 144, 28, 134, 160, 128, 233, 154, 161, 144, 12, 137, 203, 0, 24, 97, 210, 16, 117, 248, 250, 0, 88, 232, 251, 16, 87, 216, 252, 0, 58, 200, 253, 16, 57, 184, 254, 0, 28, 168, 255, 16, 27, 152, 0, 0, 254, 135, 1, 16, 253, 119, 2, 128, 26, 113, 3, 144, 25, 97, 4, 128, 252, 80, 5, 144, 251, 64, 6, 128, 222, 48, 7, 144, 53, 141, 7, 128, 192, 16, 9, 16, 177, 173, 9, 128, 162, 240, 10, 144, 161, 224, 11, 0, 191, 217, 12, 144, 131, 192, 13, 0, 161, 185, 14, 16, 160, 169, 15, 0, 131, 153, 16, 16, 130, 137, 17, 0, 101, 121, 18, 16, 100, 105, 19, 0, 71, 89, 20, 16, 70, 73, 21, 0, 41, 57, 22, 16, 40, 41, 23, 128, 69, 34, 24, 16, 10, 9, 25, 128, 39, 2, 26, 144, 38, 242, 26, 128, 9, 226, 27, 144, 8, 210, 28, 128, 235, 193, 29, 144, 234, 177, 30, 128, 205, 161, 31, 16, 29, 118, 32, 128, 175, 129, 33, 16, 255, 85, 34, 0, 204, 106, 35, 16, 225, 53, 36, 0, 174, 74, 37, 16, 195, 21, 38, 0, 144, 42, 39, 144, 223, 254, 39, 0, 114, 10, 41, 144, 193, 222, 41, 0, 84, 234, 42, 128, 149, 190, 43, 112, 98, 211, 44, 128, 119, 158, 45, 112, 68, 179, 46, 128, 89, 126, 47, 112, 38, 147, 48, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 0, 224, 198, 58, 240, 172, 219, 59, 128, 252, 175, 60, 240, 142, 187, 61, 128, 222, 143, 62, 240, 112, 155, 63, 128, 192, 111, 64, 112, 141, 132, 65, 128, 162, 79, 66, 112, 111, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 176, 12, 4, 94, 89, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 170, 10, 0, 0, 0, 8, 161, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 160, 171, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 91, 65, 0, 224, 180, 65, 0, 224, 183, 65, 0, 16, 207, 65, 0, 224, 88, 0, 0, 0, 144, 58, 166, 158, 128, 7, 187, 159, 144, 28, 134, 160, 128, 233, 154, 161, 144, 12, 137, 203, 0, 24, 97, 210, 16, 117, 248, 250, 0, 88, 232, 251, 16, 87, 216, 252, 0, 58, 200, 253, 16, 57, 184, 254, 0, 28, 168, 255, 16, 27, 152, 0, 0, 254, 135, 1, 16, 253, 119, 2, 128, 26, 113, 3, 144, 25, 97, 4, 128, 252, 80, 5, 144, 251, 64, 6, 128, 222, 48, 7, 144, 53, 141, 7, 128, 192, 16, 9, 16, 177, 173, 9, 128, 162, 240, 10, 144, 161, 224, 11, 0, 191, 217, 12, 144, 131, 192, 13, 0, 161, 185, 14, 16, 160, 169, 15, 0, 131, 153, 16, 16, 130, 137, 17, 0, 101, 121, 18, 16, 100, 105, 19, 0, 71, 89, 20, 16, 70, 73, 21, 0, 41, 57, 22, 16, 40, 41, 23, 128, 69, 34, 24, 16, 10, 9, 25, 128, 39, 2, 26, 144, 38, 242, 26, 128, 9, 226, 27, 144, 8, 210, 28, 128, 235, 193, 29, 144, 234, 177, 30, 128, 205, 161, 31, 16, 29, 118, 32, 128, 175, 129, 33, 16, 255, 85, 34, 0, 204, 106, 35, 16, 225, 53, 36, 0, 174, 74, 37, 16, 195, 21, 38, 0, 144, 42, 39, 144, 223, 254, 39, 0, 114, 10, 41, 144, 193, 222, 41, 0, 84, 234, 42, 144, 163, 190, 43, 128, 112, 211, 44, 144, 133, 158, 45, 128, 82, 179, 46, 144, 103, 126, 47, 128, 52, 147, 48, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 16, 238, 198, 58, 0, 187, 219, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 128, 192, 111, 64, 112, 141, 132, 65, 128, 162, 79, 66, 112, 111, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 176, 12, 4, 94, 170, 170, 170, 170, 170, 170, 170, 170, 89, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 170, 10, 0, 0, 0, 237, 160, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 160, 171, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 229, 65, 0, 224, 62, 66, 0, 224, 67, 66, 0, 16, 91, 66, 0, 224, 36, 0, 0, 0, 112, 232, 182, 165, 224, 110, 242, 175, 96, 86, 102, 182, 96, 210, 67, 183, 96, 54, 12, 184, 240, 134, 253, 184, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 144, 18, 245, 58, 0, 209, 182, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 144, 206, 111, 64, 128, 155, 132, 65, 144, 176, 79, 66, 128, 125, 100, 67, 144, 146, 47, 68, 128, 95, 68, 69, 144, 116, 15, 70, 128, 65, 36, 71, 16, 145, 248, 71, 128, 35, 4, 73, 16, 115, 216, 73, 128, 5, 228, 74, 144, 165, 156, 75, 128, 92, 214, 76, 170, 170, 170, 170, 36, 0, 0, 0, 1, 3, 1, 3, 1, 3, 4, 3, 4, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 10, 0, 0, 0, 28, 158, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 144, 157, 255, 127, 1, 0, 0, 96, 219, 7, 0, 112, 113, 66, 0, 224, 151, 66, 0, 16, 161, 66, 0, 224, 2, 0, 0, 0, 97, 0, 0, 0, 181, 0, 0, 0, 1, 0, 0, 0, 232, 97, 244, 139, 2, 0, 0, 0, 255, 255, 255, 255, 16, 38, 135, 105, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 112, 181, 255, 255, 0, 0, 0, 0, 24, 181, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 182, 66, 0, 224, 185, 66, 0, 224, 187, 66, 0, 224, 191, 66, 0, 16, 193, 66, 0, 224, 61, 0, 0, 0, 128, 82, 213, 163, 96, 226, 136, 203, 208, 237, 96, 210, 64, 48, 47, 247, 192, 91, 40, 248, 224, 57, 105, 19, 208, 28, 89, 20, 224, 27, 73, 21, 208, 254, 56, 22, 224, 253, 40, 23, 80, 27, 34, 24, 224, 223, 8, 25, 80, 253, 1, 26, 96, 252, 241, 26, 80, 223, 225, 27, 96, 222, 209, 28, 80, 193, 193, 29, 96, 192, 177, 30, 80, 163, 161, 31, 224, 242, 117, 32, 80, 133, 129, 33, 224, 212, 85, 34, 208, 161, 106, 35, 224, 182, 53, 36, 208, 131, 74, 37, 224, 152, 21, 38, 208, 101, 42, 39, 96, 181, 254, 39, 208, 71, 10, 41, 96, 151, 222, 41, 208, 41, 234, 42, 96, 121, 190, 43, 80, 70, 211, 44, 96, 91, 158, 45, 80, 40, 179, 46, 96, 61, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 170, 170, 170, 170, 61, 0, 0, 0, 5, 6, 5, 7, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 4, 3, 4, 3, 4, 3, 4, 3, 4, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 170, 16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 192, 199, 255, 255, 32, 28, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 176, 185, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 208, 66, 0, 224, 15, 67, 0, 16, 32, 67, 0, 224, 4, 0, 0, 0, 184, 142, 5, 145, 196, 75, 42, 190, 180, 44, 98, 210, 184, 49, 190, 27, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 4, 0, 0, 0, 1, 2, 3, 4, 10, 0, 0, 0, 72, 204, 255, 255, 0, 0, 0, 0, 60, 204, 255, 255, 0, 0, 0, 0, 76, 204, 255, 255, 0, 0, 0, 0, 200, 206, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 59, 67, 0, 224, 67, 67, 0, 16, 69, 67, 0, 224, 3, 0, 0, 0, 184, 0, 0, 0, 34, 2, 0, 0, 105, 2, 0, 0, 10, 0, 0, 0, 144, 58, 166, 158, 128, 7, 187, 159, 144, 28, 134, 160, 128, 233, 154, 161, 144, 12, 137, 203, 28, 223, 23, 207, 172, 229, 143, 207, 28, 26, 129, 208, 16, 117, 248, 250, 0, 88, 232, 251, 2, 0, 0, 0, 255, 255, 255, 255, 176, 12, 4, 94, 11, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 238, 150, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 85, 67, 0, 224, 89, 67, 0, 224, 100, 67, 0, 224, 103, 67, 0, 16, 107, 67, 0, 224, 45, 0, 0, 0, 252, 113, 110, 156, 208, 70, 27, 25, 64, 239, 1, 26, 80, 238, 241, 26, 64, 209, 225, 27, 80, 208, 209, 28, 64, 179, 193, 29, 80, 178, 177, 30, 64, 149, 161, 31, 80, 148, 145, 32, 64, 119, 129, 33, 224, 212, 85, 34, 224, 175, 106, 35, 224, 182, 53, 36, 224, 145, 74, 37, 224, 152, 21, 38, 224, 115, 42, 39, 96, 181, 254, 39, 224, 85, 10, 41, 96, 151, 222, 41, 224, 55, 234, 42, 96, 121, 190, 43, 96, 84, 211, 44, 96, 91, 158, 45, 96, 54, 179, 46, 96, 61, 126, 47, 96, 24, 147, 48, 224, 89, 103, 49, 96, 250, 114, 50, 224, 59, 71, 51, 96, 220, 82, 52, 80, 120, 79, 66, 64, 69, 100, 67, 80, 90, 47, 68, 64, 39, 68, 69, 112, 77, 92, 79, 96, 4, 150, 80, 112, 47, 60, 81, 96, 230, 117, 82, 112, 17, 28, 83, 96, 200, 85, 84, 112, 243, 251, 84, 96, 170, 53, 86, 240, 241, 196, 88, 224, 168, 254, 89, 2, 0, 0, 0, 255, 255, 255, 255, 80, 31, 135, 105, 46, 0, 0, 0, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 8, 0, 0, 0, 48, 188, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 68, 188, 255, 255, 0, 0, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 145, 3, 0, 96, 226, 7, 0, 112, 122, 67, 0, 224, 168, 67, 0, 224, 171, 67, 0, 16, 184, 67, 0, 224, 15, 0, 0, 0, 61, 0, 0, 0, 62, 0, 0, 0, 111, 0, 0, 0, 123, 0, 0, 0, 124, 0, 0, 0, 157, 0, 0, 0, 171, 0, 0, 0, 186, 0, 0, 0, 206, 0, 0, 0, 208, 0, 0, 0, 209, 0, 0, 0, 210, 0, 0, 0, 211, 0, 0, 0, 218, 0, 0, 0, 220, 0, 0, 0, 1, 0, 0, 0, 172, 51, 55, 147, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 84, 198, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 204, 67, 0, 224, 220, 67, 0, 224, 223, 67, 0, 16, 225, 67, 0, 224, 29, 0, 0, 0, 232, 130, 170, 150, 240, 87, 15, 184, 176, 78, 253, 184, 64, 66, 241, 185, 48, 130, 222, 186, 64, 188, 56, 218, 64, 8, 236, 218, 192, 239, 25, 220, 48, 103, 185, 220, 64, 35, 251, 221, 48, 236, 155, 222, 64, 168, 221, 223, 48, 65, 84, 224, 192, 13, 152, 244, 48, 108, 5, 245, 64, 114, 192, 246, 176, 44, 14, 247, 64, 58, 81, 248, 48, 211, 199, 248, 192, 224, 10, 250, 176, 6, 169, 250, 64, 20, 236, 251, 176, 139, 139, 252, 64, 156, 201, 29, 176, 229, 120, 30, 192, 67, 160, 31, 176, 221, 51, 32, 64, 119, 129, 33, 176, 214, 11, 34, 29, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 24, 196, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 237, 67, 0, 224, 11, 68, 0, 16, 20, 68, 0, 224, 2, 0, 0, 0, 189, 0, 0, 0, 36, 2, 0, 0, 2, 0, 0, 0, 192, 50, 246, 203, 208, 237, 96, 210, 2, 0, 0, 0, 255, 255, 255, 255, 185, 149, 230, 122, 170, 170, 170, 170, 170, 170, 170, 170, 3, 0, 0, 0, 1, 2, 1, 170, 6, 0, 0, 0, 7, 194, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 32, 68, 0, 224, 35, 68, 0, 224, 38, 68, 0, 224, 43, 68, 0, 16, 45, 68, 0, 224, 115, 0, 0, 0, 70, 71, 48, 143, 80, 229, 92, 155, 198, 226, 124, 159, 192, 113, 0, 161, 198, 119, 94, 176, 64, 61, 119, 177, 208, 0, 65, 178, 192, 112, 88, 179, 80, 52, 34, 180, 64, 164, 57, 181, 208, 103, 3, 182, 192, 215, 26, 183, 80, 155, 228, 183, 192, 92, 253, 184, 80, 32, 199, 185, 64, 110, 28, 204, 208, 231, 108, 204, 192, 85, 51, 213, 64, 146, 118, 213, 64, 60, 209, 253, 176, 250, 146, 254, 192, 205, 204, 255, 176, 220, 114, 0, 192, 80, 117, 1, 176, 73, 64, 2, 192, 50, 85, 3, 176, 43, 32, 4, 64, 79, 62, 5, 176, 13, 0, 6, 64, 188, 11, 7, 176, 239, 223, 7, 64, 19, 254, 8, 176, 209, 191, 9, 64, 245, 221, 10, 48, 238, 168, 11, 64, 215, 189, 12, 48, 208, 136, 13, 64, 185, 157, 14, 48, 178, 104, 15, 192, 213, 134, 16, 48, 148, 72, 17, 192, 183, 102, 18, 48, 118, 40, 19, 192, 153, 70, 20, 176, 146, 17, 21, 192, 123, 38, 22, 176, 116, 241, 22, 192, 93, 6, 24, 176, 86, 209, 24, 192, 63, 230, 25, 176, 56, 177, 26, 64, 92, 207, 27, 176, 26, 145, 28, 64, 62, 175, 29, 176, 252, 112, 30, 64, 32, 143, 31, 48, 3, 127, 32, 64, 2, 111, 33, 48, 251, 57, 34, 64, 228, 78, 35, 48, 221, 25, 36, 192, 0, 56, 37, 48, 191, 249, 37, 192, 248, 242, 38, 48, 161, 217, 39, 192, 196, 247, 40, 176, 189, 194, 41, 192, 166, 215, 42, 176, 159, 162, 43, 192, 136, 183, 44, 176, 129, 130, 45, 192, 106, 151, 46, 176, 99, 98, 47, 64, 135, 128, 48, 176, 69, 66, 49, 64, 105, 96, 50, 48, 215, 61, 51, 64, 75, 64, 52, 48, 68, 11, 53, 64, 184, 13, 54, 176, 213, 6, 55, 64, 15, 0, 56, 48, 8, 203, 56, 192, 43, 233, 57, 48, 234, 170, 58, 192, 13, 201, 59, 48, 204, 138, 60, 192, 239, 168, 61, 48, 174, 106, 62, 192, 209, 136, 63, 176, 202, 83, 64, 192, 179, 104, 65, 176, 172, 51, 66, 192, 149, 72, 67, 176, 142, 19, 68, 64, 178, 49, 69, 176, 112, 243, 69, 64, 148, 17, 71, 48, 2, 239, 71, 64, 118, 241, 72, 48, 111, 188, 73, 64, 88, 209, 74, 176, 0, 184, 75, 64, 58, 177, 76, 48, 7, 198, 77, 192, 130, 80, 78, 176, 174, 156, 79, 192, 217, 66, 80, 176, 144, 124, 81, 64, 246, 43, 82, 176, 114, 92, 83, 64, 216, 11, 84, 48, 230, 55, 87, 192, 236, 175, 87, 176, 134, 67, 88, 2, 0, 0, 0, 255, 255, 255, 255, 252, 29, 135, 105, 116, 0, 0, 0, 3, 1, 3, 4, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 4, 1, 4, 1, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 6, 14, 0, 0, 0, 132, 189, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 186, 189, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 60, 68, 0, 224, 176, 68, 0, 224, 179, 68, 0, 16, 209, 68, 0, 224, 72, 0, 0, 0, 128, 161, 184, 158, 112, 249, 186, 159, 96, 87, 248, 200, 240, 9, 97, 210, 128, 207, 32, 8, 112, 178, 16, 9, 128, 177, 0, 10, 112, 148, 240, 10, 128, 147, 224, 11, 240, 176, 217, 12, 128, 117, 192, 13, 240, 146, 185, 14, 0, 146, 169, 15, 240, 116, 153, 16, 0, 116, 137, 17, 240, 86, 121, 18, 0, 86, 105, 19, 240, 56, 89, 20, 0, 56, 73, 21, 240, 26, 57, 22, 0, 26, 41, 23, 112, 55, 34, 24, 0, 252, 8, 25, 112, 25, 2, 26, 128, 24, 242, 26, 112, 251, 225, 27, 128, 250, 209, 28, 112, 221, 193, 29, 128, 220, 177, 30, 112, 191, 161, 31, 0, 15, 118, 32, 112, 161, 129, 33, 0, 241, 85, 34, 240, 189, 106, 35, 0, 211, 53, 36, 240, 159, 74, 37, 0, 181, 21, 38, 240, 129, 42, 39, 128, 209, 254, 39, 240, 99, 10, 41, 128, 179, 222, 41, 240, 69, 234, 42, 128, 149, 190, 43, 112, 98, 211, 44, 128, 119, 158, 45, 112, 68, 179, 46, 128, 89, 126, 47, 112, 38, 147, 48, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 0, 224, 198, 58, 240, 172, 219, 59, 128, 252, 175, 60, 240, 142, 187, 61, 128, 222, 143, 62, 240, 112, 155, 63, 128, 192, 111, 64, 112, 141, 132, 65, 128, 162, 79, 66, 112, 111, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 40, 135, 238, 114, 73, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 88, 167, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 160, 171, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 231, 68, 0, 224, 48, 69, 0, 224, 51, 69, 0, 16, 71, 69, 0, 224, 59, 0, 0, 0, 0, 110, 140, 231, 96, 76, 47, 247, 224, 119, 40, 248, 0, 86, 105, 19, 240, 56, 89, 20, 0, 56, 73, 21, 240, 26, 57, 22, 0, 26, 41, 23, 112, 55, 34, 24, 0, 252, 8, 25, 112, 25, 2, 26, 128, 24, 242, 26, 112, 251, 225, 27, 128, 250, 209, 28, 112, 221, 193, 29, 128, 220, 177, 30, 112, 191, 161, 31, 0, 15, 118, 32, 112, 161, 129, 33, 0, 241, 85, 34, 240, 189, 106, 35, 0, 211, 53, 36, 240, 159, 74, 37, 0, 181, 21, 38, 240, 129, 42, 39, 128, 209, 254, 39, 240, 99, 10, 41, 128, 179, 222, 41, 240, 69, 234, 42, 128, 149, 190, 43, 112, 98, 211, 44, 128, 119, 158, 45, 112, 68, 179, 46, 128, 89, 126, 47, 112, 38, 147, 48, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 0, 224, 198, 58, 240, 172, 219, 59, 128, 252, 175, 60, 240, 142, 187, 61, 128, 222, 143, 62, 240, 112, 155, 63, 128, 192, 111, 64, 112, 141, 132, 65, 128, 162, 79, 66, 112, 111, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 170, 170, 170, 170, 170, 170, 170, 170, 59, 0, 0, 0, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 32, 28, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 160, 171, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 89, 69, 0, 224, 151, 69, 0, 16, 167, 69, 0, 224, 39, 0, 0, 0, 184, 103, 170, 150, 224, 73, 15, 184, 160, 64, 253, 184, 48, 52, 241, 185, 32, 116, 222, 186, 48, 174, 56, 218, 48, 250, 235, 218, 176, 225, 25, 220, 32, 89, 185, 220, 48, 21, 251, 221, 32, 222, 155, 222, 48, 154, 221, 223, 32, 51, 84, 224, 176, 255, 151, 244, 32, 94, 5, 245, 48, 100, 192, 246, 160, 30, 14, 247, 48, 44, 81, 248, 32, 197, 199, 248, 176, 210, 10, 250, 160, 248, 168, 250, 48, 6, 236, 251, 160, 125, 139, 252, 48, 142, 201, 29, 160, 215, 120, 30, 176, 53, 160, 31, 160, 207, 51, 32, 48, 105, 129, 33, 160, 200, 11, 34, 176, 16, 88, 35, 32, 112, 226, 35, 176, 242, 55, 37, 32, 199, 212, 37, 176, 198, 246, 55, 32, 133, 184, 56, 48, 227, 223, 57, 160, 15, 233, 57, 176, 255, 200, 59, 160, 14, 111, 60, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 39, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 72, 223, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 188, 69, 0, 224, 231, 69, 0, 16, 242, 69, 0, 224, 3, 0, 0, 0, 194, 0, 0, 0, 130, 1, 0, 0, 135, 1, 0, 0, 52, 0, 0, 0, 28, 147, 253, 134, 144, 175, 184, 158, 128, 7, 187, 159, 240, 79, 101, 181, 224, 72, 48, 182, 240, 49, 69, 183, 224, 42, 16, 184, 240, 19, 37, 185, 224, 12, 240, 185, 112, 48, 14, 187, 224, 238, 207, 187, 112, 18, 238, 188, 96, 11, 185, 189, 240, 8, 114, 194, 224, 235, 97, 195, 240, 234, 81, 196, 96, 147, 56, 197, 240, 204, 49, 198, 224, 175, 33, 199, 112, 233, 26, 200, 96, 204, 10, 201, 112, 203, 250, 201, 96, 174, 234, 202, 144, 12, 137, 203, 0, 24, 97, 210, 16, 140, 99, 211, 0, 111, 83, 212, 16, 227, 85, 213, 0, 220, 32, 214, 16, 197, 53, 215, 0, 190, 0, 216, 16, 167, 21, 217, 0, 160, 224, 217, 144, 195, 254, 218, 0, 130, 192, 219, 144, 165, 222, 220, 128, 158, 169, 221, 144, 135, 190, 222, 128, 128, 137, 223, 144, 105, 158, 224, 128, 98, 105, 225, 144, 75, 126, 226, 128, 68, 73, 227, 144, 45, 94, 228, 128, 38, 41, 229, 16, 74, 71, 230, 0, 67, 18, 231, 16, 44, 39, 232, 0, 37, 242, 232, 16, 240, 230, 235, 0, 211, 214, 236, 16, 210, 198, 237, 52, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 8, 0, 0, 0, 228, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 254, 69, 0, 224, 2, 70, 0, 224, 55, 70, 0, 16, 69, 70, 0, 224, 59, 0, 0, 0, 128, 129, 251, 213, 96, 76, 47, 247, 224, 119, 40, 248, 0, 86, 105, 19, 240, 56, 89, 20, 0, 56, 73, 21, 240, 26, 57, 22, 0, 26, 41, 23, 112, 55, 34, 24, 0, 252, 8, 25, 112, 25, 2, 26, 128, 24, 242, 26, 112, 251, 225, 27, 128, 250, 209, 28, 112, 221, 193, 29, 128, 220, 177, 30, 112, 191, 161, 31, 0, 15, 118, 32, 112, 161, 129, 33, 0, 241, 85, 34, 240, 189, 106, 35, 0, 211, 53, 36, 240, 159, 74, 37, 0, 181, 21, 38, 240, 129, 42, 39, 128, 209, 254, 39, 240, 99, 10, 41, 128, 179, 222, 41, 240, 69, 234, 42, 128, 149, 190, 43, 112, 98, 211, 44, 128, 119, 158, 45, 112, 68, 179, 46, 128, 89, 126, 47, 112, 38, 147, 48, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 0, 224, 198, 58, 240, 172, 219, 59, 128, 252, 175, 60, 240, 142, 187, 61, 128, 222, 143, 62, 240, 112, 155, 63, 128, 192, 111, 64, 112, 141, 132, 65, 128, 162, 79, 66, 112, 111, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 170, 170, 170, 170, 170, 170, 170, 170, 59, 0, 0, 0, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 2, 1, 170, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 32, 28, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 160, 171, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 85, 70, 0, 224, 147, 70, 0, 16, 163, 70, 0, 224, 3, 0, 0, 0, 187, 0, 0, 0, 196, 0, 0, 0, 118, 1, 0, 0, 31, 0, 0, 0, 144, 134, 170, 150, 0, 102, 15, 184, 192, 92, 253, 184, 80, 80, 241, 185, 64, 144, 222, 186, 80, 202, 56, 218, 80, 22, 236, 218, 208, 253, 25, 220, 64, 117, 185, 220, 80, 49, 251, 221, 64, 250, 155, 222, 80, 182, 221, 223, 64, 79, 84, 224, 208, 27, 152, 244, 64, 122, 5, 245, 80, 128, 192, 246, 192, 58, 14, 247, 80, 72, 81, 248, 64, 225, 199, 248, 208, 238, 10, 250, 192, 20, 169, 250, 80, 34, 236, 251, 192, 153, 139, 252, 80, 170, 201, 29, 192, 243, 120, 30, 208, 81, 160, 31, 192, 235, 51, 32, 80, 133, 129, 33, 192, 228, 11, 34, 80, 127, 96, 72, 192, 4, 127, 82, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 31, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 170, 8, 0, 0, 0, 112, 192, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 184, 70, 0, 224, 188, 70, 0, 224, 223, 70, 0, 16, 232, 70, 0, 224, 30, 0, 0, 0, 72, 122, 170, 150, 240, 87, 15, 184, 176, 78, 253, 184, 64, 66, 241, 185, 48, 130, 222, 186, 64, 188, 56, 218, 64, 8, 236, 218, 192, 239, 25, 220, 48, 103, 185, 220, 64, 35, 251, 221, 48, 236, 155, 222, 64, 168, 221, 223, 48, 65, 84, 224, 192, 13, 152, 244, 48, 108, 5, 245, 64, 114, 192, 246, 176, 44, 14, 247, 64, 58, 81, 248, 48, 211, 199, 248, 192, 224, 10, 250, 176, 6, 169, 250, 64, 20, 236, 251, 176, 139, 139, 252, 64, 156, 201, 29, 176, 229, 120, 30, 192, 67, 160, 31, 176, 221, 51, 32, 64, 119, 129, 33, 176, 214, 11, 34, 64, 113, 96, 72, 30, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 170, 170, 8, 0, 0, 0, 184, 204, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 248, 70, 0, 224, 23, 71, 0, 16, 32, 71, 0, 224, 2, 0, 0, 0, 200, 0, 0, 0, 137, 1, 0, 0, 122, 0, 0, 0, 70, 71, 48, 143, 80, 229, 92, 155, 198, 226, 124, 159, 192, 113, 0, 161, 198, 119, 94, 176, 64, 61, 119, 177, 208, 0, 65, 178, 192, 112, 88, 179, 80, 52, 34, 180, 64, 164, 57, 181, 208, 103, 3, 182, 192, 215, 26, 183, 80, 155, 228, 183, 192, 92, 253, 184, 80, 32, 199, 185, 64, 110, 28, 204, 208, 231, 108, 204, 192, 143, 220, 211, 176, 201, 27, 212, 192, 85, 51, 213, 64, 146, 118, 213, 64, 60, 209, 253, 176, 250, 146, 254, 192, 205, 204, 255, 176, 220, 114, 0, 192, 80, 117, 1, 176, 73, 64, 2, 192, 50, 85, 3, 176, 43, 32, 4, 64, 79, 62, 5, 176, 13, 0, 6, 64, 188, 11, 7, 176, 239, 223, 7, 64, 19, 254, 8, 176, 209, 191, 9, 64, 245, 221, 10, 48, 238, 168, 11, 64, 215, 189, 12, 48, 208, 136, 13, 64, 185, 157, 14, 48, 178, 104, 15, 192, 213, 134, 16, 48, 148, 72, 17, 192, 183, 102, 18, 48, 118, 40, 19, 192, 153, 70, 20, 176, 146, 17, 21, 192, 123, 38, 22, 176, 116, 241, 22, 192, 93, 6, 24, 176, 86, 209, 24, 192, 63, 230, 25, 176, 56, 177, 26, 64, 92, 207, 27, 176, 26, 145, 28, 64, 62, 175, 29, 176, 252, 112, 30, 64, 32, 143, 31, 48, 3, 127, 32, 64, 2, 111, 33, 48, 251, 57, 34, 64, 228, 78, 35, 48, 221, 25, 36, 192, 0, 56, 37, 48, 191, 249, 37, 192, 248, 242, 38, 48, 161, 217, 39, 192, 196, 247, 40, 176, 189, 194, 41, 192, 166, 215, 42, 176, 159, 162, 43, 192, 136, 183, 44, 176, 129, 130, 45, 192, 106, 151, 46, 176, 99, 98, 47, 64, 135, 128, 48, 176, 69, 66, 49, 64, 105, 96, 50, 48, 215, 61, 51, 64, 75, 64, 52, 48, 68, 11, 53, 64, 184, 13, 54, 176, 213, 6, 55, 64, 15, 0, 56, 48, 8, 203, 56, 192, 43, 233, 57, 48, 234, 170, 58, 192, 13, 201, 59, 48, 204, 138, 60, 192, 239, 168, 61, 48, 174, 106, 62, 192, 209, 136, 63, 176, 202, 83, 64, 192, 179, 104, 65, 176, 172, 51, 66, 192, 149, 72, 67, 176, 142, 19, 68, 64, 178, 49, 69, 176, 112, 243, 69, 64, 148, 17, 71, 48, 2, 239, 71, 64, 118, 241, 72, 48, 111, 188, 73, 64, 88, 209, 74, 176, 0, 184, 75, 64, 58, 177, 76, 48, 7, 198, 77, 192, 130, 80, 78, 176, 174, 156, 79, 192, 217, 66, 80, 176, 144, 124, 81, 64, 246, 43, 82, 176, 114, 92, 83, 64, 216, 11, 84, 48, 230, 55, 87, 192, 236, 175, 87, 48, 200, 23, 89, 192, 206, 143, 89, 48, 170, 247, 90, 192, 176, 111, 91, 176, 103, 169, 92, 192, 124, 116, 93, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 122, 0, 0, 0, 1, 0, 3, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 3, 4, 3, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 170, 170, 10, 0, 0, 0, 186, 189, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 192, 199, 255, 127, 127, 3, 0, 96, 228, 7, 0, 112, 46, 71, 0, 224, 49, 71, 0, 224, 175, 71, 0, 16, 207, 71, 0, 224, 16, 0, 0, 0, 96, 66, 223, 186, 208, 75, 8, 250, 64, 195, 167, 250, 208, 241, 167, 255, 200, 123, 67, 0, 208, 211, 135, 1, 72, 127, 250, 1, 80, 240, 112, 3, 72, 4, 221, 3, 80, 210, 80, 5, 72, 137, 191, 5, 80, 180, 48, 7, 200, 188, 160, 7, 80, 150, 16, 9, 224, 188, 251, 57, 96, 225, 41, 58, 2, 0, 0, 0, 255, 255, 255, 255, 8, 29, 135, 105, 170, 170, 170, 170, 170, 170, 170, 170, 17, 0, 0, 0, 4, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 5, 1, 5, 170, 170, 170, 12, 0, 0, 0, 120, 190, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 8, 7, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 96, 190, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 229, 71, 0, 224, 246, 71, 0, 224, 251, 71, 0, 16, 1, 72, 0, 224, 3, 0, 0, 0, 202, 0, 0, 0, 116, 1, 0, 0, 120, 1, 0, 0, 91, 0, 0, 0, 180, 114, 170, 150, 224, 73, 15, 184, 160, 64, 253, 184, 48, 52, 241, 185, 32, 116, 222, 186, 48, 174, 56, 218, 48, 250, 235, 218, 176, 225, 25, 220, 32, 89, 185, 220, 48, 21, 251, 221, 32, 222, 155, 222, 48, 154, 221, 223, 32, 51, 84, 224, 48, 9, 90, 244, 32, 94, 5, 245, 48, 100, 192, 246, 160, 30, 14, 247, 48, 44, 81, 248, 32, 197, 199, 248, 176, 210, 10, 250, 160, 248, 168, 250, 48, 6, 236, 251, 160, 125, 139, 252, 48, 142, 201, 29, 160, 215, 120, 30, 176, 53, 160, 31, 160, 207, 51, 32, 48, 105, 129, 33, 160, 200, 11, 34, 176, 16, 88, 35, 32, 112, 226, 35, 176, 242, 55, 37, 32, 199, 212, 37, 48, 15, 33, 39, 160, 227, 189, 39, 48, 241, 0, 41, 32, 139, 148, 41, 176, 13, 234, 42, 160, 50, 107, 43, 48, 181, 192, 44, 32, 196, 102, 45, 48, 151, 160, 46, 32, 166, 70, 47, 48, 121, 128, 48, 160, 77, 29, 49, 176, 32, 87, 50, 32, 106, 6, 51, 48, 84, 56, 52, 32, 193, 248, 52, 48, 31, 32, 54, 160, 104, 207, 54, 176, 198, 246, 55, 32, 133, 184, 56, 48, 227, 223, 57, 160, 44, 143, 58, 176, 255, 200, 59, 160, 14, 111, 60, 48, 145, 196, 61, 160, 240, 78, 62, 48, 254, 145, 63, 160, 210, 46, 64, 48, 248, 134, 65, 32, 239, 23, 66, 48, 194, 81, 67, 32, 209, 247, 67, 176, 83, 77, 69, 160, 237, 224, 69, 48, 134, 17, 71, 32, 149, 183, 71, 176, 162, 250, 72, 32, 119, 151, 73, 176, 132, 218, 74, 160, 147, 128, 75, 176, 102, 186, 76, 160, 117, 96, 77, 176, 72, 154, 78, 32, 146, 73, 79, 48, 101, 131, 80, 160, 57, 32, 81, 48, 71, 99, 82, 160, 27, 0, 83, 48, 41, 67, 84, 32, 56, 233, 84, 48, 11, 35, 86, 32, 26, 201, 86, 48, 237, 2, 88, 32, 252, 168, 88, 48, 207, 226, 89, 32, 222, 136, 90, 176, 96, 222, 91, 32, 192, 104, 92, 170, 170, 170, 170, 170, 170, 170, 170, 91, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 76, 212, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 21, 72, 0, 224, 25, 72, 0, 224, 119, 72, 0, 16, 143, 72, 0, 224, 35, 0, 0, 0, 24, 76, 128, 155, 64, 110, 77, 19, 192, 36, 52, 20, 160, 249, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 35, 0, 0, 0, 1, 2, 1, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 10, 0, 0, 0, 104, 235, 255, 255, 0, 0, 0, 0, 224, 227, 255, 255, 0, 0, 0, 0, 224, 227, 255, 255, 16, 14, 0, 0, 240, 241, 255, 255, 0, 0, 0, 0, 240, 241, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 240, 241, 255, 127, 4, 0, 0, 96, 205, 7, 0, 112, 157, 72, 0, 224, 195, 72, 0, 16, 205, 72, 0, 224, 80, 0, 0, 0, 160, 26, 137, 203, 16, 38, 97, 210, 32, 71, 184, 254, 16, 42, 168, 255, 32, 41, 152, 0, 16, 12, 136, 1, 32, 11, 120, 2, 144, 40, 113, 3, 160, 39, 97, 4, 144, 10, 81, 5, 160, 9, 65, 6, 144, 236, 48, 7, 160, 67, 141, 7, 144, 206, 16, 9, 32, 191, 173, 9, 144, 176, 240, 10, 160, 175, 224, 11, 16, 205, 217, 12, 160, 145, 192, 13, 16, 175, 185, 14, 32, 174, 169, 15, 16, 145, 153, 16, 32, 144, 137, 17, 16, 115, 121, 18, 32, 114, 105, 19, 16, 85, 89, 20, 32, 84, 73, 21, 16, 55, 57, 22, 32, 54, 41, 23, 144, 83, 34, 24, 32, 24, 9, 25, 144, 53, 2, 26, 176, 66, 242, 26, 160, 37, 226, 27, 176, 36, 210, 28, 160, 7, 194, 29, 176, 6, 178, 30, 160, 233, 161, 31, 48, 57, 118, 32, 160, 203, 129, 33, 48, 27, 86, 34, 32, 232, 106, 35, 48, 253, 53, 36, 32, 202, 74, 37, 48, 223, 21, 38, 32, 172, 42, 39, 176, 251, 254, 39, 32, 142, 10, 41, 176, 221, 222, 41, 32, 112, 234, 42, 176, 191, 190, 43, 160, 140, 211, 44, 176, 161, 158, 45, 160, 110, 179, 46, 176, 131, 126, 47, 160, 80, 147, 48, 48, 160, 103, 49, 160, 50, 115, 50, 48, 130, 71, 51, 160, 20, 83, 52, 48, 100, 39, 53, 160, 246, 50, 54, 48, 70, 7, 55, 32, 19, 28, 56, 48, 40, 231, 56, 32, 245, 251, 57, 48, 10, 199, 58, 32, 215, 219, 59, 176, 38, 176, 60, 32, 185, 187, 61, 176, 8, 144, 62, 32, 155, 155, 63, 176, 234, 111, 64, 160, 183, 132, 65, 176, 204, 79, 66, 160, 153, 100, 67, 176, 174, 47, 68, 160, 123, 68, 69, 48, 225, 243, 69, 32, 152, 45, 71, 4, 0, 0, 0, 255, 255, 255, 255, 209, 253, 194, 63, 255, 255, 255, 255, 153, 51, 135, 125, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 82, 0, 0, 0, 1, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 12, 0, 0, 0, 167, 210, 0, 0, 0, 0, 0, 0, 39, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 16, 14, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 112, 129, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 226, 72, 0, 224, 51, 73, 0, 224, 59, 73, 0, 16, 81, 73, 0, 224, 3, 0, 0, 0, 207, 0, 0, 0, 124, 1, 0, 0, 133, 1, 0, 0, 187, 0, 0, 0, 12, 98, 207, 156, 252, 230, 164, 157, 140, 126, 184, 158, 124, 214, 186, 159, 220, 136, 182, 160, 76, 255, 56, 161, 92, 25, 149, 162, 76, 252, 132, 163, 92, 251, 116, 164, 76, 222, 100, 165, 220, 23, 94, 166, 76, 192, 68, 167, 220, 249, 61, 168, 76, 162, 36, 169, 220, 219, 29, 170, 76, 132, 4, 171, 220, 189, 253, 171, 76, 102, 228, 172, 220, 159, 221, 173, 204, 130, 205, 174, 220, 129, 189, 175, 204, 100, 173, 176, 92, 158, 166, 177, 204, 70, 141, 178, 92, 128, 134, 179, 204, 40, 109, 180, 92, 98, 102, 181, 204, 10, 77, 182, 92, 68, 70, 183, 204, 236, 44, 184, 92, 38, 38, 185, 76, 9, 22, 186, 220, 66, 15, 187, 76, 235, 245, 187, 220, 36, 239, 188, 76, 205, 213, 189, 108, 77, 158, 190, 168, 6, 207, 190, 24, 175, 181, 191, 56, 49, 184, 192, 168, 239, 121, 193, 56, 19, 152, 194, 168, 209, 89, 195, 56, 245, 119, 196, 168, 179, 57, 197, 184, 17, 97, 198, 168, 149, 25, 199, 184, 243, 64, 200, 40, 178, 2, 201, 184, 213, 32, 202, 40, 148, 226, 202, 184, 183, 0, 204, 200, 230, 96, 210, 216, 68, 136, 211, 72, 3, 74, 212, 216, 38, 104, 213, 72, 229, 41, 214, 216, 8, 72, 215, 72, 199, 9, 216, 216, 234, 39, 217, 72, 169, 233, 217, 88, 7, 17, 219, 200, 197, 210, 219, 88, 116, 222, 220, 72, 109, 169, 221, 88, 86, 190, 222, 72, 79, 137, 223, 88, 56, 158, 224, 72, 49, 105, 225, 88, 26, 126, 226, 72, 19, 73, 227, 88, 252, 93, 228, 72, 245, 40, 229, 216, 24, 71, 230, 200, 17, 18, 231, 216, 250, 38, 232, 200, 243, 241, 232, 216, 220, 6, 234, 200, 213, 209, 234, 216, 190, 230, 235, 200, 183, 177, 236, 216, 160, 198, 237, 72, 190, 191, 238, 88, 189, 175, 239, 72, 160, 159, 240, 88, 159, 143, 241, 72, 130, 127, 242, 88, 129, 111, 243, 72, 100, 95, 244, 88, 99, 79, 245, 72, 70, 63, 246, 88, 69, 47, 247, 200, 98, 40, 248, 88, 39, 15, 249, 200, 68, 8, 250, 216, 67, 248, 250, 200, 38, 232, 251, 216, 37, 216, 252, 200, 8, 200, 253, 216, 7, 184, 254, 200, 234, 167, 255, 216, 233, 151, 0, 200, 204, 135, 1, 216, 203, 119, 2, 72, 233, 112, 3, 88, 232, 96, 4, 72, 203, 80, 5, 88, 202, 64, 6, 72, 173, 48, 7, 88, 172, 32, 8, 72, 143, 16, 9, 88, 142, 0, 10, 72, 113, 240, 10, 88, 112, 224, 11, 200, 141, 217, 12, 88, 82, 192, 13, 200, 111, 185, 14, 216, 110, 169, 15, 200, 81, 153, 16, 216, 80, 137, 17, 200, 51, 121, 18, 216, 50, 105, 19, 200, 21, 89, 20, 216, 20, 73, 21, 200, 247, 56, 22, 216, 246, 40, 23, 72, 20, 34, 24, 216, 216, 8, 25, 72, 246, 1, 26, 88, 245, 241, 26, 72, 216, 225, 27, 88, 215, 209, 28, 72, 186, 193, 29, 88, 185, 177, 30, 72, 156, 161, 31, 244, 207, 117, 32, 100, 98, 129, 33, 244, 177, 85, 34, 212, 112, 106, 35, 244, 147, 53, 36, 228, 96, 74, 37, 244, 117, 21, 38, 228, 66, 42, 39, 116, 146, 254, 39, 228, 36, 10, 41, 116, 116, 222, 41, 228, 6, 234, 42, 116, 86, 190, 43, 100, 35, 211, 44, 116, 56, 158, 45, 100, 5, 179, 46, 116, 26, 126, 47, 100, 231, 146, 48, 244, 54, 103, 49, 100, 201, 114, 50, 244, 24, 71, 51, 100, 171, 82, 52, 244, 250, 38, 53, 100, 141, 50, 54, 244, 220, 6, 55, 228, 169, 27, 56, 244, 190, 230, 56, 228, 139, 251, 57, 244, 160, 198, 58, 228, 109, 219, 59, 116, 189, 175, 60, 228, 79, 187, 61, 116, 159, 143, 62, 228, 49, 155, 63, 116, 129, 111, 64, 100, 78, 132, 65, 116, 99, 79, 66, 100, 48, 100, 67, 116, 69, 47, 68, 100, 18, 68, 69, 244, 119, 243, 69, 228, 46, 45, 71, 244, 89, 211, 71, 228, 16, 13, 73, 244, 59, 179, 73, 228, 242, 236, 74, 116, 88, 156, 75, 100, 15, 214, 76, 116, 58, 124, 77, 72, 13, 182, 78, 88, 56, 92, 79, 72, 239, 149, 80, 170, 170, 170, 170, 170, 170, 170, 170, 187, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 10, 0, 0, 0, 148, 206, 255, 255, 0, 0, 0, 0, 148, 206, 255, 255, 16, 14, 0, 0, 200, 206, 255, 255, 0, 0, 0, 0, 200, 206, 255, 255, 16, 14, 0, 0, 200, 206, 255, 255, 32, 28, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 200, 206, 255, 127, 169, 3, 0, 96, 221, 7, 0, 112, 105, 73, 0, 224, 109, 73, 0, 224, 43, 74, 0, 16, 91, 74, 0, 224, 22, 0, 0, 0, 24, 150, 253, 134, 144, 175, 184, 158, 128, 7, 187, 159, 144, 12, 137, 203, 0, 24, 97, 210, 16, 1, 118, 211, 0, 111, 83, 212, 16, 227, 85, 213, 0, 220, 32, 214, 16, 197, 53, 215, 0, 190, 0, 216, 16, 167, 21, 217, 0, 160, 224, 217, 16, 44, 39, 232, 0, 15, 23, 233, 16, 240, 230, 235, 0, 211, 214, 236, 16, 210, 198, 237, 0, 203, 145, 238, 144, 238, 175, 239, 0, 173, 113, 240, 144, 25, 97, 4, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 22, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 170, 170, 8, 0, 0, 0, 232, 154, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 113, 74, 0, 224, 139, 74, 0, 16, 146, 74, 0, 224, 7, 0, 0, 0, 68, 75, 76, 164, 224, 220, 154, 32, 80, 155, 92, 33, 224, 190, 122, 34, 80, 125, 60, 35, 224, 140, 93, 68, 208, 200, 214, 68, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 7, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 60, 174, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 160, 74, 0, 224, 171, 74, 0, 16, 174, 74, 0, 224, 35, 0, 0, 0, 252, 119, 128, 155, 224, 122, 245, 39, 208, 93, 229, 40, 224, 92, 213, 41, 208, 63, 197, 42, 96, 121, 190, 43, 80, 70, 211, 44, 96, 91, 158, 45, 80, 40, 179, 46, 96, 61, 126, 47, 80, 10, 147, 48, 224, 89, 103, 49, 80, 236, 114, 50, 224, 59, 71, 51, 80, 206, 82, 52, 224, 29, 39, 53, 80, 176, 50, 54, 224, 255, 6, 55, 208, 204, 27, 56, 224, 225, 230, 56, 208, 174, 251, 57, 224, 195, 198, 58, 208, 144, 219, 59, 96, 224, 175, 60, 208, 114, 187, 61, 96, 194, 143, 62, 208, 84, 155, 63, 96, 164, 111, 64, 80, 113, 132, 65, 96, 134, 79, 66, 80, 83, 100, 67, 96, 104, 47, 68, 80, 53, 68, 69, 224, 154, 243, 69, 208, 81, 45, 71, 170, 170, 170, 170, 35, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 132, 191, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 192, 199, 255, 127, 202, 10, 0, 96, 216, 7, 0, 112, 186, 74, 0, 224, 223, 74, 0, 16, 233, 74, 0, 224, 77, 0, 0, 0, 224, 123, 36, 143, 112, 240, 136, 203, 224, 251, 96, 210, 240, 254, 151, 0, 224, 225, 135, 1, 240, 224, 119, 2, 96, 254, 112, 3, 112, 253, 96, 4, 96, 224, 80, 5, 112, 193, 32, 8, 96, 164, 16, 9, 112, 163, 0, 10, 96, 134, 240, 10, 112, 133, 224, 11, 224, 162, 217, 12, 112, 103, 192, 13, 224, 132, 185, 14, 240, 131, 169, 15, 224, 102, 153, 16, 240, 101, 137, 17, 224, 72, 121, 18, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 112, 10, 242, 26, 96, 237, 225, 27, 112, 236, 209, 28, 96, 207, 193, 29, 112, 206, 177, 30, 96, 177, 161, 31, 240, 0, 118, 32, 96, 147, 129, 33, 240, 226, 85, 34, 224, 175, 106, 35, 240, 196, 53, 36, 224, 145, 74, 37, 240, 166, 21, 38, 224, 115, 42, 39, 112, 195, 254, 39, 224, 85, 10, 41, 112, 165, 222, 41, 224, 55, 234, 42, 112, 135, 190, 43, 96, 84, 211, 44, 112, 105, 158, 45, 96, 54, 179, 46, 112, 75, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 240, 239, 230, 56, 224, 188, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 44, 130, 238, 114, 78, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 84, 172, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 250, 74, 0, 224, 72, 75, 0, 224, 75, 75, 0, 16, 96, 75, 0, 224, 4, 0, 0, 0, 115, 0, 0, 0, 198, 0, 0, 0, 216, 0, 0, 0, 25, 2, 0, 0, 94, 0, 0, 0, 128, 246, 182, 165, 112, 79, 121, 169, 240, 124, 242, 175, 112, 100, 102, 182, 0, 16, 27, 183, 240, 242, 10, 184, 128, 141, 234, 203, 112, 186, 153, 210, 0, 89, 27, 215, 240, 180, 145, 216, 144, 75, 126, 226, 144, 82, 73, 227, 144, 45, 94, 228, 144, 52, 41, 229, 16, 74, 71, 230, 16, 81, 18, 231, 16, 44, 39, 232, 16, 51, 242, 232, 16, 14, 7, 234, 16, 21, 210, 234, 16, 240, 230, 235, 16, 247, 177, 236, 16, 210, 198, 237, 16, 217, 145, 238, 160, 175, 224, 11, 16, 205, 217, 12, 160, 145, 192, 13, 16, 175, 185, 14, 32, 174, 169, 15, 16, 145, 153, 16, 32, 144, 137, 17, 16, 115, 121, 18, 32, 114, 105, 19, 16, 85, 89, 20, 32, 84, 73, 21, 16, 55, 57, 22, 32, 54, 41, 23, 144, 83, 34, 24, 32, 24, 9, 25, 144, 53, 2, 26, 160, 52, 242, 26, 144, 23, 226, 27, 160, 22, 210, 28, 144, 249, 193, 29, 160, 248, 177, 30, 144, 219, 161, 31, 32, 43, 118, 32, 144, 189, 129, 33, 32, 13, 86, 34, 16, 218, 106, 35, 32, 239, 53, 36, 16, 188, 74, 37, 32, 209, 21, 38, 16, 158, 42, 39, 160, 237, 254, 39, 16, 128, 10, 41, 160, 207, 222, 41, 16, 98, 234, 42, 160, 177, 190, 43, 144, 126, 211, 44, 160, 147, 158, 45, 144, 96, 179, 46, 160, 117, 126, 47, 144, 66, 147, 48, 32, 146, 103, 49, 144, 36, 115, 50, 32, 116, 71, 51, 144, 6, 83, 52, 32, 86, 39, 53, 144, 232, 50, 54, 32, 56, 7, 55, 16, 5, 28, 56, 32, 26, 231, 56, 16, 231, 251, 57, 32, 252, 198, 58, 16, 201, 219, 59, 160, 24, 176, 60, 16, 171, 187, 61, 160, 250, 143, 62, 16, 141, 155, 63, 160, 220, 111, 64, 144, 169, 132, 65, 160, 190, 79, 66, 144, 139, 100, 67, 160, 160, 47, 68, 144, 109, 68, 69, 160, 130, 15, 70, 144, 79, 36, 71, 32, 159, 248, 71, 144, 49, 4, 73, 32, 129, 216, 73, 144, 19, 228, 74, 160, 179, 156, 75, 144, 106, 214, 76, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 94, 0, 0, 0, 3, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 8, 0, 0, 0, 76, 146, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 128, 143, 255, 127, 1, 0, 0, 96, 219, 7, 0, 112, 116, 75, 0, 224, 121, 75, 0, 224, 219, 75, 0, 16, 244, 75, 0, 224, 3, 0, 0, 0, 170, 0, 0, 0, 217, 0, 0, 0, 131, 1, 0, 0, 170, 0, 0, 0, 112, 147, 184, 158, 96, 235, 186, 159, 200, 46, 135, 160, 64, 177, 154, 161, 240, 6, 148, 162, 64, 169, 85, 163, 240, 93, 134, 164, 96, 120, 40, 165, 240, 63, 102, 166, 224, 78, 12, 167, 240, 33, 70, 168, 224, 48, 236, 168, 112, 201, 28, 170, 96, 77, 213, 170, 112, 171, 252, 171, 96, 47, 181, 172, 112, 141, 220, 173, 96, 17, 149, 174, 112, 111, 188, 175, 224, 45, 126, 176, 112, 81, 156, 177, 96, 74, 103, 178, 112, 51, 124, 179, 96, 44, 71, 180, 112, 21, 92, 181, 96, 14, 39, 182, 112, 247, 59, 183, 96, 240, 6, 184, 240, 19, 37, 185, 96, 210, 230, 185, 240, 245, 4, 187, 224, 238, 207, 187, 240, 215, 228, 188, 224, 208, 175, 189, 240, 185, 196, 190, 224, 178, 143, 191, 240, 155, 164, 192, 224, 148, 111, 193, 240, 125, 132, 194, 224, 118, 79, 195, 240, 95, 100, 196, 224, 88, 47, 197, 112, 124, 77, 198, 224, 58, 15, 199, 112, 94, 45, 200, 224, 251, 96, 210, 240, 228, 117, 211, 224, 221, 64, 212, 208, 170, 85, 213, 192, 163, 32, 214, 208, 140, 53, 215, 192, 133, 0, 216, 208, 110, 21, 217, 64, 118, 51, 218, 112, 167, 254, 218, 96, 116, 19, 220, 112, 137, 222, 220, 96, 130, 169, 221, 112, 107, 190, 222, 96, 100, 137, 223, 112, 77, 158, 224, 96, 70, 105, 225, 112, 47, 126, 226, 96, 40, 73, 227, 112, 17, 94, 228, 96, 10, 41, 229, 240, 45, 71, 230, 224, 38, 18, 231, 240, 15, 39, 232, 224, 242, 22, 233, 240, 241, 6, 234, 224, 212, 246, 234, 240, 211, 230, 235, 224, 182, 214, 236, 240, 181, 198, 237, 96, 211, 191, 238, 112, 210, 175, 239, 96, 181, 159, 240, 112, 180, 143, 241, 96, 151, 127, 242, 112, 150, 111, 243, 96, 121, 95, 244, 112, 120, 79, 245, 96, 91, 63, 246, 112, 90, 47, 247, 224, 119, 40, 248, 112, 60, 15, 249, 224, 89, 8, 250, 240, 88, 248, 250, 224, 59, 232, 251, 240, 58, 216, 252, 224, 29, 200, 253, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 240, 224, 119, 2, 96, 254, 112, 3, 112, 253, 96, 4, 96, 224, 80, 5, 112, 223, 64, 6, 96, 194, 48, 7, 112, 193, 32, 8, 96, 164, 16, 9, 112, 163, 0, 10, 96, 134, 240, 10, 112, 133, 224, 11, 224, 162, 217, 12, 112, 103, 192, 13, 224, 132, 185, 14, 240, 131, 169, 15, 224, 102, 153, 16, 240, 101, 137, 17, 224, 72, 121, 18, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 112, 10, 242, 26, 96, 237, 225, 27, 112, 236, 209, 28, 96, 207, 193, 29, 112, 206, 177, 30, 96, 177, 161, 31, 240, 0, 118, 32, 96, 147, 129, 33, 240, 226, 85, 34, 224, 175, 106, 35, 240, 196, 53, 36, 224, 145, 74, 37, 240, 166, 21, 38, 224, 115, 42, 39, 112, 195, 254, 39, 224, 85, 10, 41, 112, 165, 222, 41, 224, 55, 234, 42, 112, 135, 190, 43, 96, 84, 211, 44, 112, 105, 158, 45, 96, 54, 179, 46, 112, 75, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 240, 239, 230, 56, 224, 188, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 236, 120, 238, 114, 170, 170, 170, 170, 171, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 148, 181, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 176, 185, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 8, 76, 0, 224, 12, 76, 0, 224, 183, 76, 0, 224, 187, 76, 0, 16, 231, 76, 0, 224, 2, 0, 0, 0, 219, 0, 0, 0, 134, 1, 0, 0, 128, 0, 0, 0, 160, 189, 184, 158, 144, 21, 187, 159, 160, 26, 137, 203, 16, 38, 97, 210, 32, 15, 118, 211, 16, 8, 65, 212, 32, 241, 85, 213, 16, 234, 32, 214, 32, 211, 53, 215, 16, 204, 0, 216, 32, 181, 21, 217, 16, 174, 224, 217, 160, 209, 254, 218, 16, 144, 192, 219, 160, 179, 222, 220, 144, 172, 169, 221, 160, 149, 190, 222, 144, 142, 137, 223, 160, 119, 158, 224, 144, 112, 105, 225, 160, 89, 126, 226, 144, 82, 73, 227, 160, 59, 94, 228, 144, 52, 41, 229, 32, 88, 71, 230, 16, 81, 18, 231, 32, 58, 39, 232, 16, 51, 242, 232, 32, 28, 7, 234, 16, 21, 210, 234, 32, 254, 230, 235, 16, 247, 177, 236, 32, 224, 198, 237, 16, 217, 145, 238, 160, 252, 175, 239, 16, 187, 113, 240, 160, 222, 143, 241, 144, 193, 127, 242, 160, 192, 111, 243, 144, 163, 95, 244, 160, 162, 79, 245, 144, 133, 63, 246, 160, 132, 47, 247, 16, 162, 40, 248, 160, 102, 15, 249, 16, 132, 8, 250, 32, 131, 248, 250, 16, 102, 232, 251, 32, 101, 216, 252, 16, 72, 200, 253, 32, 71, 184, 254, 16, 42, 168, 255, 32, 41, 152, 0, 16, 12, 136, 1, 32, 11, 120, 2, 144, 40, 113, 3, 160, 39, 97, 4, 144, 10, 81, 5, 160, 9, 65, 6, 144, 236, 48, 7, 160, 235, 32, 8, 144, 206, 16, 9, 160, 205, 0, 10, 144, 176, 240, 10, 160, 175, 224, 11, 16, 205, 217, 12, 160, 145, 192, 13, 16, 175, 185, 14, 32, 174, 169, 15, 16, 145, 153, 16, 32, 144, 137, 17, 16, 115, 121, 18, 32, 114, 105, 19, 16, 85, 89, 20, 32, 84, 73, 21, 16, 55, 57, 22, 32, 54, 41, 23, 144, 83, 34, 24, 32, 24, 9, 25, 144, 53, 2, 26, 160, 52, 242, 26, 144, 23, 226, 27, 160, 22, 210, 28, 144, 249, 193, 29, 160, 248, 177, 30, 144, 219, 161, 31, 32, 43, 118, 32, 144, 189, 129, 33, 32, 13, 86, 34, 16, 218, 106, 35, 32, 239, 53, 36, 16, 188, 74, 37, 32, 209, 21, 38, 16, 158, 42, 39, 160, 237, 254, 39, 16, 128, 10, 41, 160, 207, 222, 41, 16, 98, 234, 42, 160, 177, 190, 43, 144, 126, 211, 44, 160, 147, 158, 45, 144, 96, 179, 46, 160, 117, 126, 47, 144, 66, 147, 48, 32, 146, 103, 49, 144, 36, 115, 50, 32, 116, 71, 51, 144, 6, 83, 52, 32, 86, 39, 53, 144, 232, 50, 54, 32, 56, 7, 55, 16, 5, 28, 56, 32, 26, 231, 56, 16, 231, 251, 57, 32, 252, 198, 58, 16, 201, 219, 59, 160, 24, 176, 60, 16, 171, 187, 61, 160, 250, 143, 62, 16, 141, 155, 63, 160, 220, 111, 64, 144, 169, 132, 65, 160, 190, 79, 66, 144, 139, 100, 67, 160, 160, 47, 68, 144, 109, 68, 69, 32, 211, 243, 69, 16, 138, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 236, 118, 61, 94, 170, 170, 170, 170, 129, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 148, 140, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 128, 143, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 251, 76, 0, 224, 254, 76, 0, 224, 127, 77, 0, 224, 131, 77, 0, 16, 165, 77, 0, 224, 2, 0, 0, 0, 221, 0, 0, 0, 136, 1, 0, 0, 65, 0, 0, 0, 176, 203, 184, 158, 160, 35, 187, 159, 176, 12, 208, 160, 128, 210, 162, 161, 176, 40, 137, 203, 32, 52, 97, 210, 144, 118, 47, 247, 16, 162, 40, 248, 16, 95, 29, 251, 32, 114, 105, 19, 16, 85, 89, 20, 32, 84, 73, 21, 16, 55, 57, 22, 32, 54, 41, 23, 144, 83, 34, 24, 32, 24, 9, 25, 144, 53, 2, 26, 160, 52, 242, 26, 144, 23, 226, 27, 160, 22, 210, 28, 144, 249, 193, 29, 160, 248, 177, 30, 144, 219, 161, 31, 32, 43, 118, 32, 144, 189, 129, 33, 32, 13, 86, 34, 16, 218, 106, 35, 32, 239, 53, 36, 16, 188, 74, 37, 32, 209, 21, 38, 16, 158, 42, 39, 160, 237, 254, 39, 16, 128, 10, 41, 160, 207, 222, 41, 16, 98, 234, 42, 160, 177, 190, 43, 144, 126, 211, 44, 160, 147, 158, 45, 144, 96, 179, 46, 160, 117, 126, 47, 144, 66, 147, 48, 32, 146, 103, 49, 144, 36, 115, 50, 32, 116, 71, 51, 144, 6, 83, 52, 32, 86, 39, 53, 144, 232, 50, 54, 32, 56, 7, 55, 16, 5, 28, 56, 32, 26, 231, 56, 16, 231, 251, 57, 32, 252, 198, 58, 16, 201, 219, 59, 160, 24, 176, 60, 16, 171, 187, 61, 160, 250, 143, 62, 16, 141, 155, 63, 160, 220, 111, 64, 144, 169, 132, 65, 160, 190, 79, 66, 144, 139, 100, 67, 160, 160, 47, 68, 144, 109, 68, 69, 32, 211, 243, 69, 16, 138, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 156, 138, 134, 125, 170, 170, 170, 170, 170, 170, 170, 170, 66, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 3, 1, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 170, 170, 12, 0, 0, 0, 100, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 16, 14, 0, 0, 112, 129, 255, 255, 32, 28, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 128, 143, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 185, 77, 0, 224, 188, 77, 0, 224, 254, 77, 0, 224, 3, 78, 0, 16, 21, 78, 0, 224, 2, 0, 0, 0, 222, 0, 0, 0, 129, 1, 0, 0, 124, 0, 0, 0, 224, 251, 1, 155, 80, 186, 195, 155, 128, 161, 184, 158, 112, 249, 186, 159, 128, 59, 160, 194, 240, 132, 79, 195, 128, 254, 136, 203, 240, 9, 97, 210, 0, 104, 136, 211, 240, 96, 83, 212, 0, 213, 85, 213, 240, 205, 32, 214, 0, 183, 53, 215, 240, 175, 0, 216, 0, 153, 21, 217, 240, 145, 224, 217, 0, 7, 0, 219, 240, 92, 200, 219, 128, 151, 222, 220, 112, 144, 169, 221, 128, 121, 190, 222, 112, 114, 137, 223, 128, 91, 158, 224, 112, 84, 105, 225, 128, 61, 126, 226, 112, 54, 73, 227, 128, 31, 94, 228, 112, 24, 41, 229, 0, 60, 71, 230, 240, 52, 18, 231, 0, 30, 39, 232, 240, 22, 242, 232, 0, 0, 7, 234, 240, 248, 209, 234, 0, 226, 230, 235, 240, 196, 214, 236, 0, 196, 198, 237, 240, 188, 145, 238, 128, 164, 111, 243, 240, 98, 49, 244, 128, 74, 15, 249, 0, 118, 8, 250, 0, 103, 248, 250, 0, 88, 232, 251, 0, 73, 216, 252, 0, 58, 200, 253, 0, 43, 184, 254, 0, 28, 168, 255, 0, 13, 152, 0, 0, 254, 135, 1, 0, 239, 119, 2, 128, 26, 113, 3, 128, 11, 97, 4, 128, 252, 80, 5, 128, 237, 64, 6, 128, 222, 48, 7, 128, 207, 32, 8, 128, 192, 16, 9, 128, 177, 0, 10, 128, 162, 240, 10, 128, 147, 224, 11, 0, 191, 217, 12, 128, 117, 192, 13, 0, 161, 185, 14, 0, 146, 169, 15, 0, 131, 153, 16, 0, 116, 137, 17, 0, 101, 121, 18, 0, 86, 105, 19, 0, 71, 89, 20, 0, 56, 73, 21, 0, 41, 57, 22, 0, 26, 41, 23, 128, 69, 34, 24, 0, 252, 8, 25, 128, 39, 2, 26, 128, 24, 242, 26, 128, 9, 226, 27, 128, 250, 209, 28, 128, 235, 193, 29, 128, 220, 177, 30, 128, 205, 161, 31, 0, 15, 118, 32, 128, 175, 129, 33, 0, 241, 85, 34, 0, 204, 106, 35, 0, 211, 53, 36, 0, 174, 74, 37, 0, 181, 21, 38, 0, 144, 42, 39, 128, 209, 254, 39, 0, 114, 10, 41, 128, 179, 222, 41, 0, 84, 234, 42, 128, 149, 190, 43, 128, 112, 211, 44, 128, 119, 158, 45, 128, 82, 179, 46, 128, 89, 126, 47, 128, 52, 147, 48, 0, 118, 103, 49, 128, 22, 115, 50, 0, 88, 71, 51, 128, 248, 82, 52, 0, 58, 39, 53, 128, 218, 50, 54, 0, 28, 7, 55, 0, 247, 27, 56, 0, 254, 230, 56, 0, 217, 251, 57, 0, 224, 198, 58, 0, 187, 219, 59, 128, 252, 175, 60, 0, 157, 187, 61, 128, 222, 143, 62, 0, 127, 155, 63, 128, 192, 111, 64, 128, 155, 132, 65, 128, 162, 79, 66, 128, 125, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 2, 0, 0, 0, 255, 255, 255, 255, 148, 176, 228, 100, 170, 170, 170, 170, 125, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 236, 164, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 160, 171, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 47, 78, 0, 224, 50, 78, 0, 224, 175, 78, 0, 224, 179, 78, 0, 16, 212, 78, 0, 224, 80, 0, 0, 0, 176, 40, 137, 203, 32, 52, 97, 210, 48, 85, 184, 254, 32, 56, 168, 255, 48, 55, 152, 0, 32, 26, 136, 1, 48, 25, 120, 2, 160, 54, 113, 3, 176, 53, 97, 4, 160, 24, 81, 5, 176, 23, 65, 6, 160, 250, 48, 7, 176, 81, 141, 7, 160, 220, 16, 9, 48, 205, 173, 9, 160, 190, 240, 10, 176, 189, 224, 11, 32, 219, 217, 12, 176, 159, 192, 13, 32, 189, 185, 14, 48, 188, 169, 15, 32, 159, 153, 16, 48, 158, 137, 17, 32, 129, 121, 18, 48, 128, 105, 19, 32, 99, 89, 20, 48, 98, 73, 21, 32, 69, 57, 22, 48, 68, 41, 23, 160, 97, 34, 24, 48, 38, 9, 25, 160, 67, 2, 26, 176, 66, 242, 26, 160, 37, 226, 27, 176, 36, 210, 28, 160, 7, 194, 29, 176, 6, 178, 30, 160, 233, 161, 31, 48, 57, 118, 32, 160, 203, 129, 33, 48, 27, 86, 34, 32, 232, 106, 35, 48, 253, 53, 36, 32, 202, 74, 37, 48, 223, 21, 38, 32, 172, 42, 39, 176, 251, 254, 39, 32, 142, 10, 41, 176, 221, 222, 41, 32, 112, 234, 42, 176, 191, 190, 43, 160, 140, 211, 44, 176, 161, 158, 45, 160, 110, 179, 46, 176, 131, 126, 47, 160, 80, 147, 48, 48, 160, 103, 49, 160, 50, 115, 50, 48, 130, 71, 51, 160, 20, 83, 52, 48, 100, 39, 53, 160, 246, 50, 54, 48, 70, 7, 55, 32, 19, 28, 56, 48, 40, 231, 56, 32, 245, 251, 57, 48, 10, 199, 58, 32, 215, 219, 59, 176, 38, 176, 60, 32, 185, 187, 61, 176, 8, 144, 62, 32, 155, 155, 63, 176, 234, 111, 64, 160, 183, 132, 65, 176, 204, 79, 66, 160, 153, 100, 67, 176, 174, 47, 68, 160, 123, 68, 69, 48, 225, 243, 69, 32, 152, 45, 71, 4, 0, 0, 0, 255, 255, 255, 255, 209, 253, 194, 63, 255, 255, 255, 255, 191, 55, 135, 125, 170, 170, 170, 170, 82, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 129, 206, 0, 0, 0, 0, 0, 0, 1, 125, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 112, 129, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 232, 78, 0, 224, 57, 79, 0, 224, 63, 79, 0, 16, 85, 79, 0, 224, 61, 0, 0, 0, 0, 24, 42, 190, 144, 12, 137, 203, 0, 24, 97, 210, 112, 90, 47, 247, 240, 133, 40, 248, 16, 100, 105, 19, 0, 71, 89, 20, 16, 70, 73, 21, 0, 41, 57, 22, 16, 40, 41, 23, 128, 69, 34, 24, 16, 10, 9, 25, 128, 39, 2, 26, 144, 38, 242, 26, 128, 9, 226, 27, 144, 8, 210, 28, 128, 235, 193, 29, 144, 234, 177, 30, 128, 205, 161, 31, 16, 29, 118, 32, 128, 175, 129, 33, 16, 255, 85, 34, 0, 204, 106, 35, 16, 225, 53, 36, 0, 174, 74, 37, 16, 195, 21, 38, 0, 144, 42, 39, 144, 223, 254, 39, 0, 114, 10, 41, 144, 193, 222, 41, 0, 84, 234, 42, 144, 163, 190, 43, 128, 112, 211, 44, 144, 133, 158, 45, 128, 82, 179, 46, 144, 103, 126, 47, 128, 52, 147, 48, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 16, 238, 198, 58, 0, 187, 219, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 144, 206, 111, 64, 128, 155, 132, 65, 144, 176, 79, 66, 128, 125, 100, 67, 144, 146, 47, 68, 128, 95, 68, 69, 16, 197, 243, 69, 0, 124, 45, 71, 61, 0, 0, 0, 1, 2, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 144, 157, 255, 255, 32, 28, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 144, 157, 255, 127, 169, 3, 0, 96, 216, 7, 0, 112, 105, 79, 0, 224, 167, 79, 0, 16, 184, 79, 0, 224, 7, 0, 0, 0, 128, 204, 30, 254, 32, 6, 218, 74, 240, 202, 143, 75, 32, 156, 169, 78, 144, 205, 67, 79, 128, 59, 10, 88, 16, 15, 164, 90, 7, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 203, 79, 0, 224, 211, 79, 0, 16, 214, 79, 0, 224, 7, 0, 0, 0, 0, 64, 156, 231, 16, 223, 71, 246, 0, 171, 71, 254, 48, 20, 218, 74, 64, 250, 151, 75, 48, 170, 169, 78, 192, 247, 67, 79, 170, 170, 170, 170, 7, 0, 0, 0, 2, 0, 2, 1, 2, 1, 2, 170, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 226, 79, 0, 224, 235, 79, 0, 16, 238, 79, 0, 224, 3, 0, 0, 0, 128, 118, 188, 212, 96, 96, 52, 222, 128, 2, 60, 231, 170, 170, 170, 170, 3, 0, 0, 0, 1, 0, 1, 170, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 250, 79, 0, 224, 255, 79, 0, 16, 1, 80, 0, 224, 90, 0, 0, 0, 128, 120, 213, 155, 240, 32, 188, 156, 96, 180, 135, 160, 0, 104, 12, 215, 0, 141, 194, 251, 0, 126, 178, 252, 0, 89, 199, 253, 128, 176, 118, 254, 0, 59, 167, 255, 128, 146, 86, 0, 0, 29, 135, 1, 0, 175, 63, 2, 128, 57, 112, 3, 0, 28, 13, 4, 128, 27, 80, 5, 128, 56, 246, 5, 128, 253, 47, 7, 128, 26, 214, 7, 128, 223, 15, 9, 128, 252, 181, 9, 128, 193, 239, 10, 0, 25, 159, 11, 0, 222, 216, 12, 0, 251, 126, 13, 0, 192, 184, 14, 0, 221, 94, 15, 0, 162, 152, 16, 0, 191, 62, 17, 0, 132, 120, 18, 0, 161, 30, 19, 0, 102, 88, 20, 0, 131, 254, 20, 0, 72, 56, 22, 0, 79, 3, 23, 128, 100, 33, 24, 0, 49, 227, 24, 128, 70, 1, 26, 128, 99, 167, 26, 128, 40, 225, 27, 128, 69, 135, 28, 128, 10, 193, 29, 128, 39, 103, 30, 0, 178, 151, 31, 128, 126, 89, 32, 128, 206, 128, 33, 0, 155, 66, 34, 0, 235, 105, 35, 0, 125, 34, 36, 0, 205, 73, 37, 0, 95, 2, 38, 0, 175, 41, 39, 0, 182, 244, 39, 128, 225, 237, 40, 0, 152, 212, 41, 128, 195, 205, 42, 0, 122, 180, 43, 128, 165, 173, 44, 0, 92, 148, 45, 128, 135, 141, 46, 0, 62, 116, 47, 128, 105, 109, 48, 128, 90, 93, 49, 0, 134, 86, 50, 128, 60, 61, 51, 0, 104, 54, 52, 128, 30, 29, 53, 0, 74, 22, 54, 128, 0, 253, 54, 0, 44, 246, 55, 128, 226, 220, 56, 128, 233, 167, 57, 128, 196, 188, 58, 128, 42, 191, 59, 0, 225, 165, 60, 128, 12, 159, 61, 0, 195, 133, 62, 128, 238, 126, 63, 0, 165, 101, 64, 128, 208, 94, 65, 0, 135, 69, 66, 128, 178, 62, 67, 128, 163, 46, 68, 128, 148, 30, 69, 0, 75, 5, 70, 0, 177, 7, 71, 0, 162, 247, 71, 0, 147, 231, 72, 0, 132, 215, 73, 0, 117, 199, 74, 0, 102, 183, 75, 2, 0, 0, 0, 255, 255, 255, 255, 0, 22, 5, 124, 170, 170, 170, 170, 170, 170, 170, 170, 91, 0, 0, 0, 1, 2, 1, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 170, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 11, 80, 0, 224, 102, 80, 0, 224, 107, 80, 0, 16, 131, 80, 0, 224, 2, 0, 0, 0, 128, 50, 32, 226, 64, 34, 218, 74, 170, 170, 170, 170, 2, 0, 0, 0, 2, 1, 170, 170, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 147, 80, 0, 224, 151, 80, 0, 16, 153, 80, 0, 224, 82, 0, 0, 0, 0, 173, 152, 246, 176, 159, 230, 246, 192, 67, 19, 248, 48, 211, 199, 248, 64, 119, 244, 249, 176, 54, 211, 250, 192, 53, 195, 251, 48, 83, 188, 252, 64, 82, 172, 253, 48, 53, 156, 254, 64, 52, 140, 255, 176, 74, 163, 7, 160, 111, 36, 8, 176, 188, 48, 23, 192, 93, 6, 24, 176, 86, 209, 24, 192, 63, 230, 25, 176, 56, 177, 26, 64, 92, 207, 27, 176, 26, 145, 28, 64, 62, 175, 29, 176, 252, 112, 30, 64, 32, 143, 31, 48, 3, 127, 32, 64, 2, 111, 33, 48, 251, 57, 34, 64, 228, 78, 35, 48, 221, 25, 36, 192, 0, 56, 37, 48, 191, 249, 37, 192, 248, 242, 38, 48, 161, 217, 39, 192, 196, 247, 40, 176, 189, 194, 41, 192, 166, 215, 42, 176, 159, 162, 43, 192, 136, 183, 44, 176, 129, 130, 45, 192, 106, 151, 46, 176, 99, 98, 47, 64, 135, 128, 48, 176, 69, 66, 49, 64, 105, 96, 50, 48, 215, 61, 51, 64, 75, 64, 52, 48, 68, 11, 53, 64, 184, 13, 54, 176, 213, 6, 55, 64, 15, 0, 56, 48, 8, 203, 56, 192, 43, 233, 57, 48, 234, 170, 58, 192, 13, 201, 59, 48, 204, 138, 60, 192, 239, 168, 61, 48, 174, 106, 62, 192, 209, 136, 63, 176, 202, 83, 64, 192, 179, 104, 65, 176, 172, 51, 66, 192, 149, 72, 67, 176, 142, 19, 68, 64, 178, 49, 69, 176, 112, 243, 69, 64, 148, 17, 71, 48, 2, 239, 71, 64, 118, 241, 72, 48, 111, 188, 73, 64, 88, 209, 74, 176, 0, 184, 75, 64, 58, 177, 76, 48, 7, 198, 77, 192, 130, 80, 78, 176, 174, 156, 79, 192, 217, 66, 80, 176, 144, 124, 81, 64, 246, 43, 82, 176, 114, 92, 83, 64, 216, 11, 84, 48, 230, 55, 87, 192, 236, 175, 87, 176, 134, 67, 88, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 82, 0, 0, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 4, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 170, 170, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 165, 80, 0, 224, 251, 80, 0, 16, 17, 81, 0, 224, 1, 0, 0, 0, 0, 45, 2, 13, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 33, 81, 0, 224, 35, 81, 0, 16, 37, 81, 0, 224, 1, 0, 0, 0, 0, 88, 177, 231, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 47, 81, 0, 224, 51, 81, 0, 16, 53, 81, 0, 224, 2, 0, 0, 0, 144, 5, 70, 66, 16, 27, 100, 67, 170, 170, 170, 170, 2, 0, 0, 0, 1, 0, 170, 170, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 0, 0, 0, 112, 7, 23, 0, 96, 214, 7, 0, 112, 63, 81, 0, 224, 67, 81, 0, 16, 69, 81, 0, 224, 1, 0, 0, 0, 128, 137, 88, 233, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 84, 81, 0, 224, 87, 81, 0, 16, 89, 81, 0, 224, 51, 0, 0, 0, 220, 123, 25, 170, 48, 239, 163, 181, 160, 125, 39, 21, 16, 178, 24, 22, 32, 177, 8, 23, 144, 229, 249, 23, 160, 228, 233, 24, 16, 25, 219, 25, 160, 105, 204, 26, 192, 118, 188, 27, 192, 103, 172, 28, 192, 88, 156, 29, 192, 73, 140, 30, 192, 58, 124, 31, 192, 43, 108, 32, 192, 28, 92, 33, 192, 13, 76, 34, 192, 254, 59, 35, 192, 239, 43, 36, 192, 224, 27, 37, 192, 209, 11, 38, 64, 253, 4, 39, 64, 238, 244, 39, 80, 237, 228, 40, 80, 149, 120, 41, 64, 208, 212, 41, 64, 193, 196, 42, 64, 178, 180, 43, 64, 163, 164, 44, 64, 148, 148, 45, 64, 133, 132, 46, 64, 118, 116, 47, 64, 103, 100, 48, 192, 146, 93, 49, 192, 109, 114, 50, 192, 116, 61, 51, 192, 79, 82, 52, 192, 86, 29, 53, 192, 49, 50, 54, 192, 56, 253, 54, 64, 78, 27, 56, 192, 26, 221, 56, 64, 48, 251, 57, 192, 252, 188, 58, 64, 18, 219, 59, 64, 25, 166, 60, 64, 244, 186, 61, 64, 251, 133, 62, 64, 214, 154, 63, 64, 221, 101, 64, 192, 242, 131, 65, 51, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 10, 0, 0, 0, 36, 72, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 99, 81, 0, 224, 151, 81, 0, 16, 165, 81, 0, 224, 71, 0, 0, 0, 208, 214, 163, 182, 224, 121, 114, 6, 80, 171, 12, 7, 96, 55, 36, 8, 208, 222, 237, 8, 224, 106, 5, 10, 80, 18, 207, 10, 224, 239, 231, 11, 208, 117, 218, 12, 96, 35, 201, 13, 208, 202, 146, 14, 96, 5, 169, 15, 208, 172, 114, 16, 96, 213, 173, 28, 208, 9, 159, 29, 96, 253, 146, 30, 80, 224, 130, 31, 96, 223, 114, 32, 80, 194, 98, 33, 96, 193, 82, 34, 208, 222, 75, 35, 96, 188, 100, 36, 208, 192, 43, 37, 96, 111, 55, 38, 208, 162, 11, 39, 224, 115, 11, 40, 80, 74, 226, 40, 96, 190, 228, 41, 208, 102, 203, 42, 224, 101, 187, 43, 208, 72, 171, 44, 224, 71, 155, 45, 208, 181, 120, 46, 96, 100, 132, 47, 224, 165, 88, 48, 96, 70, 100, 49, 96, 194, 65, 50, 96, 40, 68, 51, 96, 164, 33, 52, 96, 10, 36, 53, 96, 134, 1, 54, 96, 147, 122, 55, 224, 162, 234, 55, 224, 124, 226, 56, 96, 191, 211, 57, 224, 94, 194, 58, 96, 161, 179, 59, 96, 146, 163, 60, 96, 131, 147, 61, 96, 116, 131, 62, 96, 79, 152, 63, 96, 86, 99, 64, 224, 246, 110, 65, 224, 114, 76, 66, 224, 99, 60, 67, 224, 84, 44, 68, 224, 47, 65, 69, 224, 54, 12, 70, 224, 17, 33, 71, 224, 24, 236, 71, 96, 46, 10, 73, 224, 250, 203, 73, 96, 16, 234, 74, 224, 220, 171, 75, 96, 242, 201, 76, 96, 249, 148, 77, 96, 212, 169, 78, 96, 219, 116, 79, 80, 94, 179, 82, 96, 159, 52, 83, 224, 180, 82, 84, 170, 170, 170, 170, 170, 170, 170, 170, 71, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 176, 33, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 197, 3, 0, 96, 223, 7, 0, 112, 181, 81, 0, 224, 255, 81, 0, 16, 18, 82, 0, 224, 64, 0, 0, 0, 156, 29, 25, 170, 192, 140, 163, 181, 48, 27, 39, 21, 160, 79, 24, 22, 176, 78, 8, 23, 48, 145, 249, 23, 64, 144, 233, 24, 176, 196, 218, 25, 64, 21, 204, 26, 96, 34, 188, 27, 96, 19, 172, 28, 96, 4, 156, 29, 96, 245, 139, 30, 96, 230, 123, 31, 96, 215, 107, 32, 96, 200, 91, 33, 96, 185, 75, 34, 96, 170, 59, 35, 96, 155, 43, 36, 96, 140, 27, 37, 96, 125, 11, 38, 224, 168, 4, 39, 224, 153, 244, 39, 240, 152, 228, 40, 240, 64, 120, 41, 224, 123, 212, 41, 224, 108, 196, 42, 224, 93, 180, 43, 224, 78, 164, 44, 224, 63, 148, 45, 224, 48, 132, 46, 224, 33, 116, 47, 224, 18, 100, 48, 96, 62, 93, 49, 96, 25, 114, 50, 96, 32, 61, 51, 96, 251, 81, 52, 96, 2, 29, 53, 96, 221, 49, 54, 96, 228, 252, 54, 224, 249, 26, 56, 96, 198, 220, 56, 224, 219, 250, 57, 96, 168, 188, 58, 224, 189, 218, 59, 224, 196, 165, 60, 224, 159, 186, 61, 224, 166, 133, 62, 224, 129, 154, 63, 224, 136, 101, 64, 96, 158, 131, 65, 224, 106, 69, 66, 96, 128, 99, 67, 224, 76, 37, 68, 96, 98, 67, 69, 224, 46, 5, 70, 96, 68, 35, 71, 96, 75, 238, 71, 96, 38, 3, 73, 96, 45, 206, 73, 96, 8, 227, 74, 96, 15, 174, 75, 240, 50, 204, 76, 112, 255, 141, 77, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 64, 0, 0, 0, 3, 5, 6, 5, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 14, 0, 0, 0, 100, 166, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 16, 14, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 16, 14, 0, 0, 208, 182, 0, 0, 0, 0, 0, 0, 208, 182, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 35, 82, 0, 224, 103, 82, 0, 16, 120, 82, 0, 224, 50, 0, 0, 0, 224, 148, 25, 170, 64, 253, 163, 181, 48, 206, 24, 22, 32, 177, 8, 23, 160, 243, 249, 23, 176, 242, 233, 24, 32, 39, 219, 25, 176, 119, 204, 26, 208, 132, 188, 27, 208, 117, 172, 28, 208, 102, 156, 29, 208, 87, 140, 30, 208, 72, 124, 31, 208, 57, 108, 32, 208, 42, 92, 33, 208, 27, 76, 34, 208, 12, 60, 35, 208, 253, 43, 36, 208, 238, 27, 37, 208, 223, 11, 38, 80, 11, 5, 39, 80, 252, 244, 39, 96, 251, 228, 40, 96, 163, 120, 41, 80, 222, 212, 41, 80, 207, 196, 42, 80, 192, 180, 43, 80, 177, 164, 44, 80, 162, 148, 45, 80, 147, 132, 46, 96, 146, 116, 47, 96, 131, 100, 48, 224, 174, 93, 49, 224, 137, 114, 50, 224, 144, 61, 51, 224, 107, 82, 52, 224, 114, 29, 53, 224, 77, 50, 54, 224, 84, 253, 54, 96, 106, 27, 56, 224, 54, 221, 56, 96, 76, 251, 57, 224, 24, 189, 58, 96, 46, 219, 59, 96, 53, 166, 60, 96, 16, 187, 61, 96, 23, 134, 62, 96, 242, 154, 63, 96, 249, 101, 64, 224, 14, 132, 65, 50, 0, 0, 0, 1, 3, 5, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 170, 170, 12, 0, 0, 0, 32, 47, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 140, 82, 0, 224, 191, 82, 0, 16, 205, 82, 0, 224, 51, 0, 0, 0, 104, 142, 25, 170, 64, 253, 163, 181, 176, 139, 39, 21, 32, 192, 24, 22, 32, 177, 8, 23, 160, 243, 249, 23, 176, 242, 233, 24, 32, 39, 219, 25, 176, 119, 204, 26, 208, 132, 188, 27, 208, 117, 172, 28, 208, 102, 156, 29, 208, 87, 140, 30, 208, 72, 124, 31, 208, 57, 108, 32, 208, 42, 92, 33, 208, 27, 76, 34, 208, 12, 60, 35, 208, 253, 43, 36, 208, 238, 27, 37, 208, 223, 11, 38, 80, 11, 5, 39, 80, 252, 244, 39, 96, 251, 228, 40, 96, 163, 120, 41, 80, 222, 212, 41, 80, 207, 196, 42, 80, 192, 180, 43, 80, 177, 164, 44, 80, 162, 148, 45, 80, 147, 132, 46, 80, 132, 116, 47, 80, 117, 100, 48, 208, 160, 93, 49, 208, 123, 114, 50, 208, 130, 61, 51, 208, 93, 82, 52, 208, 100, 29, 53, 208, 63, 50, 54, 208, 70, 253, 54, 80, 92, 27, 56, 208, 40, 221, 56, 80, 62, 251, 57, 208, 10, 189, 58, 80, 32, 219, 59, 80, 39, 166, 60, 80, 2, 187, 61, 80, 9, 134, 62, 80, 228, 154, 63, 80, 235, 101, 64, 208, 0, 132, 65, 51, 0, 0, 0, 1, 3, 4, 5, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 12, 0, 0, 0, 152, 53, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 223, 82, 0, 224, 19, 83, 0, 16, 33, 83, 0, 224, 2, 0, 0, 0, 244, 0, 0, 0, 245, 0, 0, 0, 25, 0, 0, 0, 68, 141, 25, 170, 64, 253, 163, 181, 176, 139, 39, 21, 32, 192, 24, 22, 48, 191, 8, 23, 160, 243, 249, 23, 176, 242, 233, 24, 32, 39, 219, 25, 176, 119, 204, 26, 208, 132, 188, 27, 208, 117, 172, 28, 208, 102, 156, 29, 208, 87, 140, 30, 208, 72, 124, 31, 208, 57, 108, 32, 208, 42, 92, 33, 208, 27, 76, 34, 208, 12, 60, 35, 208, 253, 43, 36, 208, 238, 27, 37, 208, 223, 11, 38, 80, 11, 5, 39, 80, 252, 244, 39, 96, 251, 228, 40, 96, 163, 120, 41, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 25, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 170, 170, 170, 10, 0, 0, 0, 188, 54, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 51, 83, 0, 224, 54, 83, 0, 224, 83, 83, 0, 16, 91, 83, 0, 224, 50, 0, 0, 0, 80, 147, 25, 170, 80, 11, 164, 181, 48, 206, 24, 22, 32, 177, 8, 23, 160, 243, 249, 23, 176, 242, 233, 24, 32, 39, 219, 25, 176, 119, 204, 26, 208, 132, 188, 27, 208, 117, 172, 28, 208, 102, 156, 29, 208, 87, 140, 30, 208, 72, 124, 31, 208, 57, 108, 32, 208, 42, 92, 33, 208, 27, 76, 34, 208, 12, 60, 35, 208, 253, 43, 36, 208, 238, 27, 37, 208, 223, 11, 38, 80, 11, 5, 39, 80, 252, 244, 39, 96, 251, 228, 40, 96, 163, 120, 41, 80, 222, 212, 41, 80, 207, 196, 42, 80, 192, 180, 43, 80, 177, 164, 44, 80, 162, 148, 45, 80, 147, 132, 46, 80, 132, 116, 47, 80, 117, 100, 48, 208, 160, 93, 49, 208, 123, 114, 50, 208, 130, 61, 51, 208, 93, 82, 52, 208, 100, 29, 53, 208, 63, 50, 54, 208, 70, 253, 54, 96, 106, 27, 56, 224, 54, 221, 56, 96, 76, 251, 57, 224, 24, 189, 58, 96, 46, 219, 59, 96, 53, 166, 60, 96, 16, 187, 61, 96, 23, 134, 62, 96, 242, 154, 63, 96, 249, 101, 64, 224, 14, 132, 65, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 50, 0, 0, 0, 1, 4, 6, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 4, 170, 170, 14, 0, 0, 0, 176, 48, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 109, 83, 0, 224, 163, 83, 0, 16, 177, 83, 0, 224, 53, 0, 0, 0, 224, 60, 48, 158, 80, 104, 48, 23, 192, 15, 250, 23, 80, 189, 232, 24, 64, 67, 219, 25, 208, 147, 204, 26, 64, 200, 189, 27, 80, 199, 173, 28, 224, 116, 156, 29, 224, 101, 140, 30, 224, 86, 124, 31, 224, 71, 108, 32, 224, 56, 92, 33, 224, 41, 76, 34, 224, 26, 60, 35, 224, 11, 44, 36, 224, 252, 27, 37, 224, 237, 11, 38, 96, 25, 5, 39, 0, 120, 246, 39, 128, 186, 231, 40, 0, 253, 216, 41, 128, 63, 202, 42, 128, 48, 186, 43, 0, 115, 171, 44, 0, 100, 155, 45, 128, 166, 140, 46, 128, 151, 124, 47, 0, 218, 109, 48, 128, 28, 95, 49, 0, 95, 80, 50, 0, 80, 64, 51, 128, 146, 49, 52, 128, 131, 33, 53, 0, 198, 18, 54, 0, 183, 2, 55, 128, 249, 243, 55, 0, 60, 229, 56, 128, 126, 214, 57, 128, 111, 198, 58, 0, 178, 183, 59, 0, 163, 167, 60, 128, 229, 152, 61, 128, 214, 136, 62, 0, 25, 122, 63, 128, 91, 107, 64, 0, 158, 92, 65, 0, 143, 76, 66, 128, 209, 61, 67, 128, 194, 45, 68, 0, 5, 31, 69, 0, 246, 14, 70, 128, 56, 0, 71, 2, 0, 0, 0, 255, 255, 255, 255, 220, 177, 134, 105, 170, 170, 170, 170, 54, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 164, 41, 0, 0, 0, 0, 0, 0, 160, 41, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 197, 83, 0, 224, 251, 83, 0, 224, 255, 83, 0, 16, 14, 84, 0, 224, 66, 0, 0, 0, 68, 149, 25, 170, 80, 12, 218, 231, 192, 153, 39, 21, 48, 206, 24, 22, 64, 205, 8, 23, 176, 1, 250, 23, 192, 0, 234, 24, 48, 53, 219, 25, 192, 133, 204, 26, 224, 146, 188, 27, 224, 131, 172, 28, 224, 116, 156, 29, 224, 101, 140, 30, 224, 86, 124, 31, 224, 71, 108, 32, 224, 56, 92, 33, 224, 41, 76, 34, 224, 26, 60, 35, 224, 11, 44, 36, 224, 252, 27, 37, 224, 237, 11, 38, 96, 25, 5, 39, 96, 10, 245, 39, 112, 9, 229, 40, 112, 250, 212, 41, 112, 235, 196, 42, 16, 217, 93, 49, 16, 180, 114, 50, 0, 173, 61, 51, 0, 136, 82, 52, 0, 143, 29, 53, 0, 106, 50, 54, 0, 113, 253, 54, 128, 134, 27, 56, 0, 83, 221, 56, 128, 104, 251, 57, 0, 53, 189, 58, 128, 74, 219, 59, 128, 81, 166, 60, 128, 44, 187, 61, 128, 51, 134, 62, 128, 14, 155, 63, 128, 21, 102, 64, 0, 43, 132, 65, 128, 247, 69, 66, 0, 13, 100, 67, 128, 217, 37, 68, 0, 239, 67, 69, 128, 187, 5, 70, 0, 209, 35, 71, 0, 216, 238, 71, 0, 179, 3, 73, 0, 186, 206, 73, 0, 149, 227, 74, 0, 156, 174, 75, 128, 177, 204, 76, 0, 126, 142, 77, 128, 147, 172, 78, 0, 96, 110, 79, 128, 117, 140, 80, 128, 124, 87, 81, 128, 87, 108, 82, 128, 94, 55, 83, 128, 57, 76, 84, 128, 64, 23, 85, 128, 27, 44, 86, 170, 170, 170, 170, 170, 170, 170, 170, 66, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 10, 0, 0, 0, 188, 46, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 30, 84, 0, 224, 99, 84, 0, 16, 117, 84, 0, 224, 3, 0, 0, 0, 250, 0, 0, 0, 46, 1, 0, 0, 75, 1, 0, 0, 1, 0, 0, 0, 196, 103, 106, 162, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 60, 94, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 133, 84, 0, 224, 137, 84, 0, 224, 139, 84, 0, 16, 141, 84, 0, 224, 67, 0, 0, 0, 252, 125, 213, 161, 32, 225, 163, 181, 144, 111, 39, 21, 0, 164, 24, 22, 16, 163, 8, 23, 128, 215, 249, 23, 144, 214, 233, 24, 0, 11, 219, 25, 144, 91, 204, 26, 176, 104, 188, 27, 176, 89, 172, 28, 176, 74, 156, 29, 176, 59, 140, 30, 176, 44, 124, 31, 176, 29, 108, 32, 176, 14, 92, 33, 176, 255, 75, 34, 176, 240, 59, 35, 176, 225, 43, 36, 176, 210, 27, 37, 176, 195, 11, 38, 48, 239, 4, 39, 48, 224, 244, 39, 64, 223, 228, 40, 64, 135, 120, 41, 48, 194, 212, 41, 48, 179, 196, 42, 48, 164, 180, 43, 48, 149, 164, 44, 48, 134, 148, 45, 48, 119, 132, 46, 48, 104, 116, 47, 128, 76, 199, 47, 64, 103, 100, 48, 192, 146, 93, 49, 192, 109, 114, 50, 192, 116, 61, 51, 192, 79, 82, 52, 192, 86, 29, 53, 192, 49, 50, 54, 192, 56, 253, 54, 64, 78, 27, 56, 192, 26, 221, 56, 64, 48, 251, 57, 192, 252, 188, 58, 64, 18, 219, 59, 64, 25, 166, 60, 64, 244, 186, 61, 64, 251, 133, 62, 64, 214, 154, 63, 64, 221, 101, 64, 192, 242, 131, 65, 64, 191, 69, 66, 192, 212, 99, 67, 64, 161, 37, 68, 192, 182, 67, 69, 64, 131, 5, 70, 192, 152, 35, 71, 192, 159, 238, 71, 192, 122, 3, 73, 192, 129, 206, 73, 192, 92, 227, 74, 192, 99, 174, 75, 64, 121, 204, 76, 192, 69, 142, 77, 48, 243, 75, 84, 64, 234, 246, 86, 170, 170, 170, 170, 170, 170, 170, 170, 67, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 3, 170, 10, 0, 0, 0, 132, 78, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 153, 84, 0, 224, 223, 84, 0, 16, 241, 84, 0, 224, 64, 0, 0, 0, 224, 99, 101, 162, 80, 130, 123, 163, 96, 128, 78, 164, 208, 180, 63, 165, 224, 39, 37, 166, 208, 127, 39, 167, 224, 243, 41, 168, 80, 178, 235, 168, 224, 133, 42, 232, 80, 45, 244, 232, 96, 185, 11, 234, 208, 96, 213, 234, 224, 236, 236, 235, 80, 148, 182, 236, 224, 113, 207, 237, 80, 25, 153, 238, 96, 165, 176, 239, 208, 76, 122, 240, 96, 94, 166, 4, 208, 119, 43, 5, 224, 3, 67, 6, 80, 171, 12, 7, 96, 55, 36, 8, 208, 222, 237, 8, 224, 106, 5, 10, 80, 18, 207, 10, 224, 239, 231, 11, 80, 151, 177, 12, 96, 35, 201, 13, 208, 202, 146, 14, 96, 5, 169, 15, 208, 172, 114, 16, 224, 46, 244, 26, 208, 156, 209, 27, 96, 98, 213, 28, 80, 208, 178, 29, 224, 149, 182, 30, 208, 3, 148, 31, 96, 201, 151, 32, 80, 55, 117, 33, 224, 44, 163, 34, 80, 188, 87, 35, 96, 95, 103, 36, 208, 239, 56, 37, 96, 181, 60, 38, 80, 35, 26, 39, 224, 232, 29, 40, 208, 86, 251, 40, 224, 109, 0, 42, 208, 9, 206, 42, 96, 206, 180, 43, 80, 177, 164, 44, 96, 176, 148, 45, 80, 147, 132, 46, 96, 146, 116, 47, 80, 117, 100, 48, 224, 174, 93, 49, 208, 145, 77, 50, 224, 144, 61, 51, 208, 115, 45, 52, 224, 114, 29, 53, 208, 85, 13, 54, 224, 84, 253, 54, 80, 92, 27, 56, 2, 0, 0, 0, 255, 255, 255, 255, 184, 194, 182, 86, 170, 170, 170, 170, 170, 170, 170, 170, 65, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 72, 33, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 32, 28, 0, 112, 33, 4, 0, 96, 208, 7, 0, 112, 1, 85, 0, 224, 66, 85, 0, 224, 71, 85, 0, 16, 89, 85, 0, 224, 52, 0, 0, 0, 16, 126, 25, 170, 48, 239, 163, 181, 160, 125, 39, 21, 16, 178, 24, 22, 32, 177, 8, 23, 144, 229, 249, 23, 160, 228, 233, 24, 16, 25, 219, 25, 160, 105, 204, 26, 192, 118, 188, 27, 192, 103, 172, 28, 192, 88, 156, 29, 192, 73, 140, 30, 192, 58, 124, 31, 192, 43, 108, 32, 192, 28, 92, 33, 192, 13, 76, 34, 192, 254, 59, 35, 192, 239, 43, 36, 192, 224, 27, 37, 192, 209, 11, 38, 64, 253, 4, 39, 64, 238, 244, 39, 192, 163, 190, 40, 48, 55, 231, 41, 32, 165, 196, 42, 48, 25, 199, 43, 32, 135, 164, 44, 48, 251, 166, 45, 32, 105, 132, 46, 48, 221, 134, 47, 32, 75, 100, 48, 48, 191, 102, 49, 160, 103, 77, 50, 216, 137, 61, 51, 200, 86, 82, 52, 216, 107, 29, 53, 200, 56, 50, 54, 216, 77, 253, 54, 72, 85, 27, 56, 216, 47, 221, 56, 72, 55, 251, 57, 216, 17, 189, 58, 72, 25, 219, 59, 88, 46, 166, 60, 72, 251, 186, 61, 88, 16, 134, 62, 72, 221, 154, 63, 88, 242, 101, 64, 200, 249, 131, 65, 88, 212, 69, 66, 32, 146, 251, 66, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 52, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 10, 0, 0, 0, 240, 69, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 107, 85, 0, 224, 163, 85, 0, 16, 177, 85, 0, 224, 2, 0, 0, 0, 68, 2, 138, 173, 136, 71, 103, 186, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 188, 107, 0, 0, 0, 0, 0, 0, 120, 105, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 193, 85, 0, 224, 199, 85, 0, 16, 201, 85, 0, 224, 66, 0, 0, 0, 160, 249, 219, 161, 0, 197, 163, 181, 112, 83, 39, 21, 224, 135, 24, 22, 240, 134, 8, 23, 96, 187, 249, 23, 112, 186, 233, 24, 224, 238, 218, 25, 112, 63, 204, 26, 144, 76, 188, 27, 144, 61, 172, 28, 144, 46, 156, 29, 144, 31, 140, 30, 144, 16, 124, 31, 144, 1, 108, 32, 144, 242, 91, 33, 144, 227, 75, 34, 144, 212, 59, 35, 144, 197, 43, 36, 144, 182, 27, 37, 144, 167, 11, 38, 16, 211, 4, 39, 16, 196, 244, 39, 32, 195, 228, 40, 32, 107, 120, 41, 16, 166, 212, 41, 16, 151, 196, 42, 16, 136, 180, 43, 16, 121, 164, 44, 16, 106, 148, 45, 16, 91, 132, 46, 16, 76, 116, 47, 16, 61, 100, 48, 144, 104, 93, 49, 144, 67, 114, 50, 144, 74, 61, 51, 144, 37, 82, 52, 144, 44, 29, 53, 144, 7, 50, 54, 144, 14, 253, 54, 16, 36, 27, 56, 144, 240, 220, 56, 16, 6, 251, 57, 144, 210, 188, 58, 16, 232, 218, 59, 16, 239, 165, 60, 16, 202, 186, 61, 16, 209, 133, 62, 16, 172, 154, 63, 16, 179, 101, 64, 144, 200, 131, 65, 16, 149, 69, 66, 144, 170, 99, 67, 16, 119, 37, 68, 144, 140, 67, 69, 16, 89, 5, 70, 144, 110, 35, 71, 144, 117, 238, 71, 144, 80, 3, 73, 144, 87, 206, 73, 144, 50, 227, 74, 144, 57, 174, 75, 16, 79, 204, 76, 144, 27, 142, 77, 0, 201, 75, 84, 32, 206, 246, 86, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 66, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 1, 3, 170, 170, 12, 0, 0, 0, 96, 106, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 16, 14, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 213, 85, 0, 224, 27, 86, 0, 16, 45, 86, 0, 224, 51, 0, 0, 0, 40, 231, 211, 134, 144, 220, 11, 15, 128, 200, 233, 24, 224, 238, 218, 25, 112, 63, 204, 26, 96, 34, 188, 27, 112, 33, 172, 28, 96, 4, 156, 29, 112, 3, 140, 30, 96, 230, 123, 31, 112, 229, 107, 32, 96, 200, 91, 33, 112, 199, 75, 34, 96, 170, 59, 35, 112, 169, 43, 36, 96, 140, 27, 37, 112, 139, 11, 38, 224, 168, 4, 39, 240, 167, 244, 39, 224, 138, 228, 40, 240, 137, 212, 41, 224, 108, 196, 42, 240, 107, 180, 43, 224, 78, 164, 44, 240, 77, 148, 45, 224, 48, 132, 46, 240, 47, 116, 47, 224, 18, 100, 48, 112, 76, 93, 49, 96, 47, 77, 50, 112, 46, 61, 51, 96, 17, 45, 52, 112, 16, 29, 53, 96, 243, 12, 54, 144, 165, 233, 58, 128, 158, 180, 59, 144, 157, 164, 60, 128, 128, 148, 61, 144, 127, 132, 62, 128, 98, 116, 63, 144, 97, 100, 64, 128, 68, 84, 65, 144, 67, 68, 66, 128, 38, 52, 67, 144, 37, 36, 68, 0, 67, 29, 69, 240, 170, 239, 71, 160, 154, 21, 85, 112, 97, 5, 86, 160, 124, 245, 86, 112, 67, 229, 87, 51, 0, 0, 0, 1, 2, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 2, 3, 2, 3, 2, 170, 12, 0, 0, 0, 88, 107, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 63, 86, 0, 224, 115, 86, 0, 16, 129, 86, 0, 224, 7, 0, 0, 0, 28, 189, 157, 135, 40, 28, 90, 203, 160, 43, 149, 204, 56, 128, 117, 210, 40, 0, 166, 49, 32, 0, 113, 50, 40, 234, 63, 68, 2, 0, 0, 0, 255, 255, 255, 255, 36, 153, 182, 86, 170, 170, 170, 170, 8, 0, 0, 0, 1, 2, 3, 4, 2, 6, 5, 2, 14, 0, 0, 0, 220, 74, 0, 0, 0, 0, 0, 0, 228, 74, 0, 0, 0, 0, 0, 0, 88, 77, 0, 0, 0, 0, 0, 0, 88, 77, 0, 0, 8, 7, 0, 0, 88, 77, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 104, 91, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 147, 86, 0, 224, 155, 86, 0, 224, 159, 86, 0, 16, 162, 86, 0, 224, 101, 0, 0, 0, 120, 171, 242, 161, 128, 47, 129, 162, 112, 157, 94, 163, 128, 17, 97, 164, 112, 127, 62, 165, 128, 243, 64, 166, 112, 97, 30, 167, 128, 213, 32, 168, 240, 125, 7, 169, 0, 82, 143, 241, 112, 156, 91, 242, 128, 40, 115, 243, 112, 126, 59, 244, 128, 173, 85, 245, 240, 84, 31, 246, 0, 225, 54, 247, 240, 54, 255, 247, 0, 218, 14, 249, 240, 187, 225, 249, 0, 72, 249, 250, 112, 239, 194, 251, 0, 205, 219, 252, 112, 116, 165, 253, 128, 0, 189, 254, 240, 167, 134, 255, 0, 52, 158, 0, 112, 219, 103, 1, 128, 103, 127, 2, 240, 14, 73, 3, 128, 236, 97, 4, 240, 147, 43, 5, 0, 32, 67, 6, 112, 199, 12, 7, 128, 83, 36, 8, 240, 250, 237, 8, 0, 135, 5, 10, 112, 46, 207, 10, 0, 12, 232, 11, 112, 179, 177, 12, 128, 63, 201, 13, 240, 89, 107, 14, 0, 115, 170, 15, 112, 141, 76, 16, 0, 197, 244, 24, 112, 109, 219, 25, 0, 74, 215, 26, 112, 242, 189, 27, 0, 35, 85, 30, 112, 229, 138, 31, 0, 122, 71, 32, 240, 25, 137, 33, 0, 116, 60, 34, 240, 158, 107, 35, 128, 191, 50, 36, 112, 69, 37, 37, 128, 68, 21, 38, 112, 39, 5, 39, 224, 91, 246, 39, 80, 144, 231, 40, 96, 27, 226, 41, 80, 21, 202, 42, 96, 43, 178, 43, 208, 95, 163, 44, 224, 71, 155, 45, 80, 124, 140, 46, 96, 123, 124, 47, 208, 175, 109, 48, 96, 0, 95, 49, 208, 52, 80, 50, 96, 226, 62, 51, 80, 104, 49, 52, 96, 196, 30, 53, 208, 155, 18, 54, 224, 154, 2, 55, 80, 207, 243, 55, 224, 31, 229, 56, 80, 84, 214, 57, 96, 83, 198, 58, 208, 135, 183, 59, 224, 134, 167, 60, 80, 187, 152, 61, 96, 186, 136, 62, 208, 238, 121, 63, 96, 63, 107, 64, 208, 115, 92, 65, 224, 114, 76, 66, 80, 167, 61, 67, 96, 166, 45, 68, 80, 253, 18, 69, 224, 54, 12, 70, 80, 62, 42, 71, 96, 83, 245, 71, 208, 113, 11, 73, 224, 250, 203, 73, 80, 2, 234, 74, 96, 23, 181, 75, 80, 228, 201, 76, 96, 249, 148, 77, 80, 198, 169, 78, 96, 219, 116, 79, 80, 168, 137, 80, 170, 170, 170, 170, 101, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 8, 34, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 163, 3, 0, 96, 221, 7, 0, 112, 184, 86, 0, 224, 31, 87, 0, 16, 58, 87, 0, 224, 3, 0, 0, 0, 5, 1, 0, 0, 7, 1, 0, 0, 117, 1, 0, 0, 6, 0, 0, 0, 176, 134, 219, 202, 24, 113, 5, 204, 168, 50, 149, 204, 152, 210, 168, 221, 16, 196, 59, 74, 144, 216, 60, 75, 2, 0, 0, 0, 255, 255, 255, 255, 188, 134, 134, 105, 170, 170, 170, 170, 170, 170, 170, 170, 7, 0, 0, 0, 2, 5, 1, 5, 3, 4, 3, 170, 12, 0, 0, 0, 196, 84, 0, 0, 0, 0, 0, 0, 88, 77, 0, 0, 0, 0, 0, 0, 208, 82, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 104, 91, 0, 0, 0, 0, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 75, 87, 0, 224, 79, 87, 0, 224, 86, 87, 0, 224, 91, 87, 0, 16, 94, 87, 0, 224, 4, 0, 0, 0, 196, 24, 230, 146, 240, 50, 153, 203, 112, 48, 234, 11, 0, 153, 195, 57, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 4, 0, 0, 0, 1, 2, 1, 2, 6, 0, 0, 0, 188, 117, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 115, 87, 0, 224, 123, 87, 0, 16, 125, 87, 0, 224, 2, 0, 0, 0, 9, 1, 0, 0, 40, 1, 0, 0, 1, 0, 0, 0, 168, 153, 242, 161, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 216, 51, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 137, 87, 0, 224, 140, 87, 0, 224, 143, 87, 0, 16, 145, 87, 0, 224, 24, 0, 0, 0, 128, 131, 25, 170, 48, 239, 163, 181, 160, 125, 39, 21, 16, 178, 24, 22, 32, 177, 8, 23, 144, 229, 249, 23, 160, 228, 233, 24, 16, 25, 219, 25, 160, 105, 204, 26, 192, 118, 188, 27, 192, 103, 172, 28, 192, 88, 156, 29, 192, 73, 140, 30, 192, 58, 124, 31, 192, 43, 108, 32, 192, 28, 92, 33, 192, 13, 76, 34, 192, 254, 59, 35, 192, 239, 43, 36, 192, 224, 27, 37, 192, 209, 11, 38, 64, 253, 4, 39, 64, 238, 244, 39, 80, 143, 202, 40, 170, 170, 170, 170, 24, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 10, 0, 0, 0, 128, 64, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 157, 87, 0, 224, 183, 87, 0, 16, 190, 87, 0, 224, 88, 0, 0, 0, 44, 30, 119, 165, 224, 175, 237, 9, 208, 146, 221, 10, 224, 100, 250, 11, 80, 198, 190, 12, 96, 57, 164, 13, 208, 225, 138, 14, 96, 27, 132, 15, 208, 79, 117, 16, 96, 253, 99, 17, 80, 224, 83, 18, 224, 25, 77, 19, 80, 194, 51, 20, 96, 193, 35, 21, 80, 164, 19, 22, 96, 163, 3, 23, 80, 134, 243, 23, 96, 133, 227, 24, 80, 104, 211, 25, 96, 103, 195, 26, 208, 132, 188, 27, 224, 131, 172, 28, 208, 102, 156, 29, 224, 101, 140, 30, 208, 72, 124, 31, 224, 71, 108, 32, 208, 42, 92, 33, 224, 41, 76, 34, 208, 12, 60, 35, 224, 11, 44, 36, 208, 238, 27, 37, 224, 237, 11, 38, 80, 11, 5, 39, 96, 10, 245, 39, 80, 237, 228, 40, 96, 236, 212, 41, 80, 207, 196, 42, 96, 206, 180, 43, 80, 177, 164, 44, 96, 176, 148, 45, 80, 147, 132, 46, 96, 146, 116, 47, 80, 117, 100, 48, 224, 174, 93, 49, 208, 145, 77, 50, 224, 144, 61, 51, 208, 115, 45, 52, 224, 114, 29, 53, 16, 120, 50, 54, 16, 127, 253, 54, 144, 148, 27, 56, 16, 97, 221, 56, 144, 118, 251, 57, 16, 67, 189, 58, 144, 88, 219, 59, 144, 95, 166, 60, 144, 58, 187, 61, 144, 65, 134, 62, 144, 28, 155, 63, 144, 35, 102, 64, 16, 57, 132, 65, 144, 5, 70, 66, 16, 27, 100, 67, 144, 231, 37, 68, 16, 253, 67, 69, 144, 201, 5, 70, 16, 223, 35, 71, 16, 230, 238, 71, 16, 193, 3, 73, 16, 200, 206, 73, 16, 163, 227, 74, 16, 170, 174, 75, 144, 191, 204, 76, 16, 140, 142, 77, 144, 161, 172, 78, 16, 110, 110, 79, 144, 131, 140, 80, 144, 138, 87, 81, 144, 101, 108, 82, 144, 108, 55, 83, 144, 71, 76, 84, 144, 78, 23, 85, 144, 41, 44, 86, 144, 48, 247, 86, 208, 127, 208, 87, 16, 40, 245, 89, 144, 244, 182, 90, 16, 10, 213, 91, 88, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 1, 2, 1, 8, 0, 0, 0, 212, 31, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 183, 3, 0, 96, 227, 7, 0, 112, 206, 87, 0, 224, 39, 88, 0, 16, 62, 88, 0, 224, 110, 0, 0, 0, 224, 178, 89, 200, 80, 193, 229, 204, 0, 254, 172, 205, 208, 244, 198, 206, 224, 102, 143, 207, 208, 121, 169, 208, 224, 96, 132, 209, 112, 201, 138, 210, 128, 176, 101, 211, 208, 224, 107, 212, 96, 99, 54, 232, 80, 45, 244, 232, 96, 185, 11, 234, 208, 96, 213, 234, 240, 250, 236, 235, 0, 109, 181, 236, 240, 127, 207, 237, 0, 242, 151, 238, 112, 179, 176, 239, 128, 37, 121, 240, 240, 230, 145, 241, 0, 89, 90, 242, 112, 26, 115, 243, 128, 140, 59, 244, 112, 159, 85, 245, 128, 17, 30, 246, 240, 210, 54, 247, 0, 69, 255, 247, 112, 6, 24, 249, 0, 202, 225, 249, 240, 57, 249, 250, 80, 66, 39, 251, 224, 139, 124, 8, 208, 176, 253, 8, 96, 234, 246, 9, 208, 51, 166, 10, 224, 170, 232, 19, 224, 9, 32, 20, 224, 116, 249, 26, 224, 28, 141, 27, 224, 248, 190, 28, 208, 241, 137, 29, 96, 255, 204, 30, 80, 153, 96, 31, 96, 177, 130, 32, 208, 181, 73, 33, 224, 158, 94, 34, 80, 93, 32, 35, 96, 48, 90, 36, 80, 63, 0, 37, 224, 237, 11, 38, 208, 230, 214, 38, 224, 207, 235, 39, 80, 3, 192, 40, 96, 236, 212, 41, 208, 31, 169, 42, 224, 101, 187, 43, 208, 1, 137, 44, 224, 71, 155, 45, 80, 169, 95, 46, 224, 41, 123, 47, 208, 197, 72, 48, 96, 70, 100, 49, 96, 194, 65, 50, 96, 40, 68, 51, 96, 164, 33, 52, 96, 10, 36, 53, 96, 134, 1, 54, 96, 97, 22, 55, 80, 68, 6, 56, 224, 125, 255, 56, 208, 96, 239, 57, 224, 95, 223, 58, 208, 66, 207, 59, 224, 65, 191, 60, 208, 36, 175, 61, 224, 35, 159, 62, 208, 6, 143, 63, 224, 5, 127, 64, 224, 129, 92, 65, 224, 231, 94, 66, 240, 183, 65, 67, 96, 166, 45, 68, 80, 253, 18, 69, 224, 217, 14, 70, 112, 111, 232, 70, 224, 24, 236, 71, 208, 17, 183, 72, 224, 250, 203, 73, 96, 60, 160, 74, 156, 46, 173, 75, 208, 189, 97, 76, 156, 249, 148, 77, 80, 194, 53, 78, 96, 219, 116, 79, 224, 145, 91, 80, 96, 189, 84, 81, 80, 160, 68, 82, 96, 159, 52, 83, 80, 108, 73, 84, 224, 210, 21, 85, 80, 78, 41, 86, 240, 194, 245, 86, 96, 202, 19, 88, 240, 164, 213, 88, 96, 172, 243, 89, 240, 134, 181, 90, 96, 142, 211, 91, 224, 67, 157, 92, 96, 112, 179, 93, 2, 0, 0, 0, 255, 255, 255, 255, 176, 74, 189, 125, 111, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 80, 32, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 32, 28, 0, 112, 94, 4, 0, 96, 228, 7, 0, 112, 81, 88, 0, 224, 192, 88, 0, 224, 195, 88, 0, 16, 224, 88, 0, 224, 112, 0, 0, 0, 224, 178, 89, 200, 80, 193, 229, 204, 0, 254, 172, 205, 208, 244, 198, 206, 224, 102, 143, 207, 208, 121, 169, 208, 224, 96, 132, 209, 112, 201, 138, 210, 128, 176, 101, 211, 208, 224, 107, 212, 96, 99, 54, 232, 80, 45, 244, 232, 96, 185, 11, 234, 208, 96, 213, 234, 240, 250, 236, 235, 0, 109, 181, 236, 240, 127, 207, 237, 0, 242, 151, 238, 112, 179, 176, 239, 128, 37, 121, 240, 240, 230, 145, 241, 0, 89, 90, 242, 112, 26, 115, 243, 128, 140, 59, 244, 112, 159, 85, 245, 128, 17, 30, 246, 240, 210, 54, 247, 0, 69, 255, 247, 112, 6, 24, 249, 0, 202, 225, 249, 240, 57, 249, 250, 80, 66, 39, 251, 224, 139, 124, 8, 208, 176, 253, 8, 96, 234, 246, 9, 208, 51, 166, 10, 224, 170, 232, 19, 224, 9, 32, 20, 224, 116, 249, 26, 224, 28, 141, 27, 224, 248, 190, 28, 208, 241, 137, 29, 96, 255, 204, 30, 80, 153, 96, 31, 96, 177, 130, 32, 208, 181, 73, 33, 224, 158, 94, 34, 80, 93, 32, 35, 96, 48, 90, 36, 80, 63, 0, 37, 224, 237, 11, 38, 208, 230, 214, 38, 224, 207, 235, 39, 80, 3, 192, 40, 96, 236, 212, 41, 208, 31, 169, 42, 224, 101, 187, 43, 208, 1, 137, 44, 224, 71, 155, 45, 80, 169, 95, 46, 224, 41, 123, 47, 208, 197, 72, 48, 96, 70, 100, 49, 96, 194, 65, 50, 96, 40, 68, 51, 96, 164, 33, 52, 96, 10, 36, 53, 96, 134, 1, 54, 96, 97, 22, 55, 80, 68, 6, 56, 224, 125, 255, 56, 208, 96, 239, 57, 224, 95, 223, 58, 208, 66, 207, 59, 224, 65, 191, 60, 208, 36, 175, 61, 224, 35, 159, 62, 208, 6, 143, 63, 224, 5, 127, 64, 224, 129, 92, 65, 224, 231, 94, 66, 240, 183, 65, 67, 96, 166, 45, 68, 80, 253, 18, 69, 224, 217, 14, 70, 112, 111, 232, 70, 224, 24, 236, 71, 80, 6, 187, 72, 224, 250, 203, 73, 96, 60, 160, 74, 224, 220, 171, 75, 208, 189, 97, 76, 156, 249, 148, 77, 80, 194, 53, 78, 224, 11, 92, 78, 80, 220, 132, 78, 96, 219, 116, 79, 224, 145, 91, 80, 96, 189, 84, 81, 80, 160, 68, 82, 96, 159, 52, 83, 80, 108, 73, 84, 224, 210, 21, 85, 80, 78, 41, 86, 240, 194, 245, 86, 96, 202, 19, 88, 240, 164, 213, 88, 96, 172, 243, 89, 240, 134, 181, 90, 96, 142, 211, 91, 224, 67, 157, 92, 96, 112, 179, 93, 2, 0, 0, 0, 255, 255, 255, 255, 25, 74, 189, 125, 170, 170, 170, 170, 113, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 231, 32, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 32, 28, 0, 112, 94, 4, 0, 96, 228, 7, 0, 112, 242, 88, 0, 224, 99, 89, 0, 224, 103, 89, 0, 16, 133, 89, 0, 224, 3, 0, 0, 0, 15, 1, 0, 0, 54, 1, 0, 0, 118, 2, 0, 0, 9, 0, 0, 0, 128, 67, 140, 136, 10, 43, 163, 145, 128, 230, 53, 205, 112, 206, 89, 209, 240, 62, 59, 210, 16, 187, 50, 213, 128, 228, 182, 228, 0, 152, 47, 237, 0, 199, 61, 10, 170, 170, 170, 170, 170, 170, 170, 170, 9, 0, 0, 0, 2, 1, 3, 4, 1, 3, 1, 3, 1, 170, 170, 170, 10, 0, 0, 0, 0, 100, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 246, 99, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 151, 89, 0, 224, 155, 89, 0, 224, 167, 89, 0, 16, 171, 89, 0, 224, 2, 0, 0, 0, 16, 1, 0, 0, 255, 1, 0, 0, 69, 0, 0, 0, 144, 99, 105, 133, 48, 49, 77, 202, 48, 147, 219, 202, 120, 113, 75, 203, 144, 222, 160, 210, 128, 215, 107, 211, 184, 88, 147, 212, 56, 176, 66, 213, 184, 58, 115, 214, 184, 65, 62, 215, 184, 50, 46, 216, 184, 57, 249, 216, 184, 20, 14, 218, 184, 27, 217, 218, 184, 246, 237, 219, 184, 253, 184, 220, 184, 216, 205, 221, 56, 26, 162, 222, 56, 245, 182, 223, 56, 252, 129, 224, 40, 201, 150, 225, 56, 105, 79, 226, 40, 171, 118, 227, 56, 75, 47, 228, 168, 199, 95, 229, 56, 45, 15, 230, 168, 169, 63, 231, 184, 73, 248, 231, 168, 139, 31, 233, 184, 43, 216, 233, 168, 109, 255, 234, 184, 13, 184, 235, 168, 79, 223, 236, 184, 239, 151, 237, 40, 108, 200, 238, 184, 209, 119, 239, 40, 78, 168, 240, 184, 179, 87, 241, 40, 48, 136, 242, 56, 208, 64, 243, 40, 18, 104, 244, 56, 178, 32, 245, 40, 244, 71, 246, 56, 126, 37, 247, 40, 97, 21, 248, 56, 96, 5, 249, 40, 67, 245, 249, 56, 66, 229, 250, 168, 95, 222, 251, 184, 94, 206, 252, 168, 65, 190, 253, 184, 64, 174, 254, 168, 35, 158, 255, 184, 34, 142, 0, 168, 5, 126, 1, 184, 4, 110, 2, 168, 231, 93, 3, 184, 230, 77, 4, 40, 4, 71, 5, 56, 3, 55, 6, 40, 230, 38, 7, 56, 61, 131, 7, 40, 200, 6, 9, 56, 199, 246, 9, 40, 170, 230, 10, 56, 169, 214, 11, 40, 140, 198, 12, 56, 57, 155, 17, 168, 108, 111, 18, 170, 170, 170, 170, 69, 0, 0, 0, 1, 3, 2, 4, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 170, 170, 170, 10, 0, 0, 0, 10, 107, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 8, 7, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 189, 89, 0, 224, 192, 89, 0, 224, 7, 90, 0, 16, 26, 90, 0, 224, 50, 0, 0, 0, 148, 252, 211, 134, 160, 234, 11, 15, 144, 214, 233, 24, 0, 11, 219, 25, 144, 91, 204, 26, 128, 62, 188, 27, 144, 61, 172, 28, 128, 32, 156, 29, 144, 31, 140, 30, 128, 2, 124, 31, 144, 1, 108, 32, 128, 228, 91, 33, 144, 227, 75, 34, 128, 198, 59, 35, 144, 197, 43, 36, 128, 168, 27, 37, 144, 167, 11, 38, 0, 197, 4, 39, 16, 196, 244, 39, 0, 167, 228, 40, 16, 166, 212, 41, 0, 137, 196, 42, 16, 136, 180, 43, 0, 107, 164, 44, 16, 106, 148, 45, 0, 77, 132, 46, 16, 76, 116, 47, 0, 47, 100, 48, 144, 104, 93, 49, 128, 75, 77, 50, 144, 74, 61, 51, 128, 45, 45, 52, 144, 44, 29, 53, 128, 15, 13, 54, 176, 193, 233, 58, 160, 186, 180, 59, 176, 185, 164, 60, 160, 156, 148, 61, 176, 155, 132, 62, 160, 126, 116, 63, 176, 125, 100, 64, 160, 96, 84, 65, 176, 95, 68, 66, 160, 66, 52, 67, 176, 65, 36, 68, 32, 95, 29, 69, 176, 168, 21, 85, 128, 111, 5, 86, 176, 138, 245, 86, 128, 81, 229, 87, 50, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 236, 85, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 44, 90, 0, 224, 95, 90, 0, 16, 109, 90, 0, 224, 65, 0, 0, 0, 191, 15, 18, 162, 16, 211, 163, 181, 128, 97, 39, 21, 240, 149, 24, 22, 0, 149, 8, 23, 112, 201, 249, 23, 128, 200, 233, 24, 240, 252, 218, 25, 128, 77, 204, 26, 160, 90, 188, 27, 160, 75, 172, 28, 160, 60, 156, 29, 160, 45, 140, 30, 160, 30, 124, 31, 160, 15, 108, 32, 160, 0, 92, 33, 160, 241, 75, 34, 160, 226, 59, 35, 160, 211, 43, 36, 160, 196, 27, 37, 160, 181, 11, 38, 32, 225, 4, 39, 32, 210, 244, 39, 48, 209, 228, 40, 48, 121, 120, 41, 32, 180, 212, 41, 32, 165, 196, 42, 32, 150, 180, 43, 32, 135, 164, 44, 32, 120, 148, 45, 32, 105, 132, 46, 32, 90, 116, 47, 32, 75, 100, 48, 160, 118, 93, 49, 160, 81, 114, 50, 160, 88, 61, 51, 160, 51, 82, 52, 160, 58, 29, 53, 160, 21, 50, 54, 160, 28, 253, 54, 32, 50, 27, 56, 160, 254, 220, 56, 32, 20, 251, 57, 160, 224, 188, 58, 32, 246, 218, 59, 32, 253, 165, 60, 32, 216, 186, 61, 32, 223, 133, 62, 32, 186, 154, 63, 32, 193, 101, 64, 160, 214, 131, 65, 32, 163, 69, 66, 160, 184, 99, 67, 32, 133, 37, 68, 160, 154, 67, 69, 32, 103, 5, 70, 160, 124, 35, 71, 160, 131, 238, 71, 160, 94, 3, 73, 160, 101, 206, 73, 160, 64, 227, 74, 160, 71, 174, 75, 32, 93, 204, 76, 160, 41, 142, 77, 16, 215, 75, 84, 170, 170, 170, 170, 170, 170, 170, 170, 65, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 3, 170, 170, 170, 12, 0, 0, 0, 193, 97, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 16, 14, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 123, 90, 0, 224, 191, 90, 0, 16, 209, 90, 0, 224, 7, 0, 0, 0, 224, 133, 120, 169, 96, 222, 22, 186, 136, 131, 191, 203, 112, 238, 86, 210, 8, 198, 60, 215, 0, 38, 255, 218, 136, 190, 181, 244, 7, 0, 0, 0, 2, 3, 5, 3, 4, 3, 1, 170, 12, 0, 0, 0, 32, 100, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 32, 103, 0, 0, 0, 0, 0, 0, 120, 105, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 227, 90, 0, 224, 235, 90, 0, 16, 238, 90, 0, 224, 3, 0, 0, 0, 152, 193, 22, 186, 240, 185, 88, 208, 104, 162, 181, 244, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 3, 0, 0, 0, 1, 2, 1, 170, 6, 0, 0, 0, 232, 131, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 152, 133, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 0, 91, 0, 224, 7, 91, 0, 16, 9, 91, 0, 224, 3, 0, 0, 0, 22, 1, 0, 0, 65, 1, 0, 0, 15, 2, 0, 0, 98, 0, 0, 0, 136, 69, 48, 158, 224, 178, 89, 200, 80, 193, 229, 204, 0, 254, 172, 205, 208, 244, 198, 206, 224, 102, 143, 207, 208, 121, 169, 208, 224, 96, 132, 209, 112, 201, 138, 210, 128, 176, 101, 211, 208, 224, 107, 212, 96, 20, 90, 215, 192, 31, 223, 215, 112, 181, 47, 216, 224, 70, 30, 217, 240, 232, 16, 218, 224, 179, 235, 218, 0, 52, 180, 219, 224, 32, 185, 220, 0, 141, 224, 221, 128, 206, 180, 222, 128, 191, 164, 223, 0, 118, 139, 224, 0, 125, 86, 225, 96, 74, 190, 226, 208, 52, 54, 227, 0, 247, 156, 228, 208, 22, 22, 229, 224, 211, 116, 230, 128, 210, 17, 231, 0, 255, 39, 232, 208, 79, 232, 232, 224, 139, 124, 8, 208, 176, 253, 8, 96, 234, 246, 9, 208, 51, 166, 10, 224, 170, 232, 19, 224, 9, 32, 20, 224, 116, 249, 26, 224, 28, 141, 27, 224, 248, 190, 28, 208, 241, 137, 29, 96, 255, 204, 30, 80, 153, 96, 31, 96, 177, 130, 32, 208, 181, 73, 33, 224, 158, 94, 34, 80, 93, 32, 35, 96, 48, 90, 36, 80, 63, 0, 37, 224, 237, 11, 38, 208, 230, 214, 38, 224, 207, 235, 39, 80, 3, 192, 40, 96, 236, 212, 41, 208, 31, 169, 42, 224, 101, 187, 43, 208, 1, 137, 44, 224, 71, 155, 45, 80, 169, 95, 46, 224, 41, 123, 47, 208, 197, 72, 48, 224, 150, 72, 49, 80, 110, 60, 50, 96, 179, 49, 51, 208, 254, 26, 52, 96, 149, 17, 53, 80, 166, 241, 53, 128, 8, 4, 55, 112, 1, 207, 55, 128, 95, 246, 56, 224, 249, 220, 57, 112, 237, 208, 58, 96, 91, 174, 59, 112, 160, 163, 60, 96, 178, 160, 61, 112, 130, 131, 62, 224, 159, 124, 63, 112, 54, 115, 64, 96, 164, 80, 65, 0, 143, 76, 66, 112, 79, 72, 67, 0, 113, 44, 68, 240, 246, 30, 69, 0, 83, 12, 70, 240, 99, 236, 70, 0, 53, 236, 71, 112, 245, 231, 72, 0, 23, 204, 73, 240, 156, 190, 74, 0, 249, 171, 75, 240, 9, 140, 76, 128, 21, 149, 77, 112, 155, 135, 78, 128, 247, 116, 79, 240, 66, 94, 80, 128, 217, 84, 81, 112, 73, 108, 82, 2, 0, 0, 0, 255, 255, 255, 255, 250, 194, 182, 86, 99, 0, 0, 0, 4, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 10, 0, 0, 0, 6, 33, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 32, 28, 0, 0, 248, 32, 0, 0, 0, 0, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 116, 3, 0, 96, 222, 7, 0, 112, 21, 91, 0, 224, 25, 91, 0, 224, 124, 91, 0, 224, 127, 91, 0, 16, 153, 91, 0, 224, 1, 0, 0, 0, 64, 215, 249, 208, 2, 0, 0, 0, 255, 255, 255, 255, 160, 154, 134, 105, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 224, 64, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 72, 63, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 177, 91, 0, 224, 179, 91, 0, 224, 183, 91, 0, 16, 185, 91, 0, 224, 64, 0, 0, 0, 196, 150, 82, 167, 208, 154, 163, 181, 64, 41, 39, 21, 176, 93, 24, 22, 192, 92, 8, 23, 48, 145, 249, 23, 64, 144, 233, 24, 176, 196, 218, 25, 64, 21, 204, 26, 96, 34, 188, 27, 96, 19, 172, 28, 96, 4, 156, 29, 96, 245, 139, 30, 96, 230, 123, 31, 96, 215, 107, 32, 96, 200, 91, 33, 96, 185, 75, 34, 96, 170, 59, 35, 96, 155, 43, 36, 96, 140, 27, 37, 96, 125, 11, 38, 224, 168, 4, 39, 224, 153, 244, 39, 240, 152, 228, 40, 240, 64, 120, 41, 224, 123, 212, 41, 224, 108, 196, 42, 224, 93, 180, 43, 224, 78, 164, 44, 224, 63, 148, 45, 224, 48, 132, 46, 224, 33, 116, 47, 224, 18, 100, 48, 96, 62, 93, 49, 96, 25, 114, 50, 96, 32, 61, 51, 96, 251, 81, 52, 96, 2, 29, 53, 96, 221, 49, 54, 96, 228, 252, 54, 224, 249, 26, 56, 96, 198, 220, 56, 224, 219, 250, 57, 96, 168, 188, 58, 224, 189, 218, 59, 224, 196, 165, 60, 224, 159, 186, 61, 224, 166, 133, 62, 224, 129, 154, 63, 224, 136, 101, 64, 96, 158, 131, 65, 224, 106, 69, 66, 96, 128, 99, 67, 224, 76, 37, 68, 96, 98, 67, 69, 224, 46, 5, 70, 96, 68, 35, 71, 96, 75, 238, 71, 96, 38, 3, 73, 96, 45, 206, 73, 96, 8, 227, 74, 96, 15, 174, 75, 240, 50, 204, 76, 112, 255, 141, 77, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 64, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 10, 0, 0, 0, 188, 148, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 16, 14, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 199, 91, 0, 224, 11, 92, 0, 16, 28, 92, 0, 224, 2, 0, 0, 0, 25, 1, 0, 0, 33, 2, 0, 0, 10, 0, 0, 0, 164, 252, 126, 137, 168, 50, 149, 204, 152, 18, 116, 210, 168, 224, 168, 221, 176, 69, 175, 60, 160, 40, 159, 61, 48, 160, 65, 72, 160, 71, 11, 73, 48, 221, 228, 73, 32, 123, 236, 74, 170, 170, 170, 170, 10, 0, 0, 0, 3, 4, 3, 1, 2, 1, 2, 1, 2, 1, 170, 170, 10, 0, 0, 0, 220, 62, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 88, 77, 0, 0, 0, 0, 0, 0, 88, 77, 0, 0, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 44, 92, 0, 224, 47, 92, 0, 224, 59, 92, 0, 16, 63, 92, 0, 224, 2, 0, 0, 0, 27, 1, 0, 0, 28, 1, 0, 0, 2, 0, 0, 0, 132, 125, 242, 161, 168, 48, 24, 30, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 252, 79, 0, 0, 0, 0, 0, 0, 88, 77, 0, 0, 0, 0, 0, 0, 220, 80, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 81, 92, 0, 224, 84, 92, 0, 224, 87, 92, 0, 16, 89, 92, 0, 224, 67, 0, 0, 0, 235, 228, 219, 161, 0, 197, 163, 181, 112, 83, 39, 21, 224, 135, 24, 22, 240, 134, 8, 23, 96, 187, 249, 23, 112, 186, 233, 24, 224, 238, 218, 25, 112, 63, 204, 26, 144, 76, 188, 27, 144, 61, 172, 28, 144, 46, 156, 29, 144, 31, 140, 30, 144, 16, 124, 31, 144, 1, 108, 32, 144, 242, 91, 33, 144, 227, 75, 34, 144, 212, 59, 35, 144, 197, 43, 36, 144, 182, 27, 37, 144, 167, 11, 38, 16, 211, 4, 39, 16, 196, 244, 39, 32, 195, 228, 40, 32, 107, 120, 41, 16, 166, 212, 41, 16, 151, 196, 42, 16, 136, 180, 43, 16, 121, 164, 44, 16, 106, 148, 45, 16, 91, 132, 46, 16, 76, 116, 47, 16, 61, 100, 48, 144, 104, 93, 49, 144, 67, 114, 50, 144, 74, 61, 51, 144, 37, 82, 52, 144, 44, 29, 53, 144, 7, 50, 54, 144, 14, 253, 54, 16, 36, 27, 56, 144, 240, 220, 56, 16, 6, 251, 57, 144, 210, 188, 58, 16, 232, 218, 59, 16, 239, 165, 60, 16, 202, 186, 61, 16, 209, 133, 62, 16, 172, 154, 63, 112, 228, 242, 63, 0, 165, 101, 64, 128, 186, 131, 65, 0, 135, 69, 66, 128, 156, 99, 67, 0, 105, 37, 68, 128, 126, 67, 69, 0, 75, 5, 70, 128, 96, 35, 71, 128, 103, 238, 71, 128, 66, 3, 73, 128, 73, 206, 73, 128, 36, 227, 74, 128, 43, 174, 75, 0, 65, 204, 76, 128, 13, 142, 77, 80, 2, 110, 78, 0, 201, 75, 84, 67, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 7, 5, 3, 170, 16, 0, 0, 0, 21, 127, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 16, 14, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 103, 92, 0, 224, 171, 92, 0, 16, 189, 92, 0, 224, 3, 0, 0, 0, 255, 0, 0, 0, 30, 1, 0, 0, 1, 2, 0, 0, 5, 0, 0, 0, 186, 188, 157, 135, 40, 140, 219, 202, 24, 113, 5, 204, 168, 50, 149, 204, 152, 18, 116, 210, 4, 0, 0, 0, 255, 255, 255, 255, 40, 24, 186, 38, 255, 255, 255, 255, 48, 235, 231, 67, 170, 170, 170, 170, 7, 0, 0, 0, 4, 1, 2, 3, 2, 3, 2, 170, 10, 0, 0, 0, 216, 82, 0, 0, 0, 0, 0, 0, 70, 75, 0, 0, 0, 0, 0, 0, 88, 77, 0, 0, 0, 0, 0, 0, 88, 77, 0, 0, 16, 14, 0, 0, 208, 82, 0, 0, 0, 0, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 211, 92, 0, 224, 215, 92, 0, 224, 221, 92, 0, 224, 227, 92, 0, 16, 230, 92, 0, 224, 65, 0, 0, 0, 242, 13, 249, 161, 32, 225, 163, 181, 144, 111, 39, 21, 0, 164, 24, 22, 16, 163, 8, 23, 128, 215, 249, 23, 144, 214, 233, 24, 0, 11, 219, 25, 144, 91, 204, 26, 176, 104, 188, 27, 176, 89, 172, 28, 176, 74, 156, 29, 176, 59, 140, 30, 176, 44, 124, 31, 176, 29, 108, 32, 176, 14, 92, 33, 176, 255, 75, 34, 176, 240, 59, 35, 176, 225, 43, 36, 176, 210, 27, 37, 176, 195, 11, 38, 48, 239, 4, 39, 48, 224, 244, 39, 64, 223, 228, 40, 64, 135, 120, 41, 48, 194, 212, 41, 48, 179, 196, 42, 48, 164, 180, 43, 48, 149, 164, 44, 48, 134, 148, 45, 48, 119, 132, 46, 48, 104, 116, 47, 48, 89, 100, 48, 176, 132, 93, 49, 176, 95, 114, 50, 176, 102, 61, 51, 176, 65, 82, 52, 176, 72, 29, 53, 176, 35, 50, 54, 176, 42, 253, 54, 48, 64, 27, 56, 176, 12, 221, 56, 48, 34, 251, 57, 176, 238, 188, 58, 48, 4, 219, 59, 48, 11, 166, 60, 48, 230, 186, 61, 48, 237, 133, 62, 48, 200, 154, 63, 48, 207, 101, 64, 176, 228, 131, 65, 48, 177, 69, 66, 176, 198, 99, 67, 48, 147, 37, 68, 176, 168, 67, 69, 48, 117, 5, 70, 176, 138, 35, 71, 176, 145, 238, 71, 176, 108, 3, 73, 176, 115, 206, 73, 176, 78, 227, 74, 176, 85, 174, 75, 48, 107, 204, 76, 176, 55, 142, 77, 32, 229, 75, 84, 65, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 3, 170, 170, 170, 12, 0, 0, 0, 14, 87, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 16, 14, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 249, 92, 0, 224, 59, 93, 0, 16, 77, 93, 0, 224, 7, 0, 0, 0, 163, 133, 131, 134, 144, 78, 103, 186, 96, 228, 10, 192, 96, 229, 179, 202, 8, 95, 145, 203, 240, 109, 72, 210, 8, 245, 145, 22, 2, 0, 0, 0, 255, 255, 255, 255, 170, 85, 54, 126, 170, 170, 170, 170, 8, 0, 0, 0, 1, 2, 3, 4, 5, 7, 5, 6, 16, 0, 0, 0, 86, 95, 0, 0, 0, 0, 0, 0, 93, 97, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 176, 4, 0, 0, 32, 103, 0, 0, 0, 0, 0, 0, 120, 105, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 95, 93, 0, 224, 103, 93, 0, 224, 107, 93, 0, 16, 110, 93, 0, 224, 18, 0, 0, 0, 144, 6, 138, 173, 136, 71, 103, 186, 128, 39, 123, 191, 80, 27, 243, 191, 128, 172, 93, 193, 80, 160, 213, 193, 0, 224, 62, 195, 208, 211, 182, 195, 128, 19, 32, 197, 80, 7, 152, 197, 0, 71, 1, 199, 208, 58, 121, 199, 0, 204, 227, 200, 208, 191, 91, 201, 128, 255, 196, 202, 80, 243, 60, 203, 0, 88, 145, 203, 240, 109, 72, 210, 170, 170, 170, 170, 170, 170, 170, 170, 18, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 2, 170, 170, 10, 0, 0, 0, 112, 103, 0, 0, 0, 0, 0, 0, 120, 105, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 176, 4, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 134, 93, 0, 224, 155, 93, 0, 16, 161, 93, 0, 224, 2, 0, 0, 0, 35, 1, 0, 0, 36, 1, 0, 0, 71, 0, 0, 0, 142, 91, 105, 133, 240, 117, 71, 203, 224, 202, 242, 203, 80, 186, 251, 204, 96, 254, 211, 205, 208, 165, 157, 206, 112, 122, 97, 210, 112, 248, 120, 211, 240, 173, 66, 212, 112, 171, 75, 213, 240, 76, 116, 214, 240, 83, 63, 215, 240, 68, 47, 216, 112, 250, 248, 216, 112, 213, 13, 218, 112, 220, 216, 218, 112, 183, 237, 219, 112, 190, 184, 220, 240, 234, 206, 221, 240, 218, 161, 222, 240, 181, 182, 223, 240, 188, 129, 224, 240, 151, 150, 225, 240, 41, 79, 226, 240, 121, 118, 227, 240, 11, 47, 228, 112, 150, 95, 229, 240, 237, 14, 230, 168, 169, 63, 231, 184, 73, 248, 231, 168, 139, 31, 233, 184, 43, 216, 233, 168, 109, 255, 234, 184, 13, 184, 235, 168, 79, 223, 236, 184, 239, 151, 237, 40, 108, 200, 238, 184, 209, 119, 239, 40, 78, 168, 240, 184, 179, 87, 241, 40, 48, 136, 242, 56, 208, 64, 243, 40, 18, 104, 244, 56, 178, 32, 245, 40, 244, 71, 246, 56, 126, 37, 247, 24, 83, 21, 248, 56, 96, 5, 249, 24, 53, 245, 249, 56, 66, 229, 250, 168, 95, 222, 251, 184, 94, 206, 252, 168, 65, 190, 253, 184, 64, 174, 254, 168, 35, 158, 255, 184, 34, 142, 0, 168, 5, 126, 1, 184, 4, 110, 2, 168, 231, 93, 3, 184, 230, 77, 4, 40, 4, 71, 5, 56, 3, 55, 6, 40, 230, 38, 7, 56, 61, 131, 7, 40, 200, 6, 9, 56, 199, 246, 9, 40, 170, 230, 10, 56, 169, 214, 11, 40, 140, 198, 12, 56, 57, 155, 17, 168, 108, 111, 18, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 71, 0, 0, 0, 1, 3, 4, 3, 4, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 10, 0, 0, 0, 114, 106, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 177, 93, 0, 224, 180, 93, 0, 224, 255, 93, 0, 16, 18, 94, 0, 224, 66, 0, 0, 0, 160, 54, 25, 170, 224, 168, 163, 181, 80, 55, 39, 21, 192, 107, 24, 22, 208, 106, 8, 23, 64, 159, 249, 23, 80, 158, 233, 24, 192, 210, 218, 25, 80, 35, 204, 26, 112, 48, 188, 27, 112, 33, 172, 28, 112, 18, 156, 29, 112, 3, 140, 30, 112, 244, 123, 31, 112, 229, 107, 32, 112, 214, 91, 33, 112, 199, 75, 34, 112, 184, 59, 35, 112, 169, 43, 36, 112, 154, 27, 37, 112, 139, 11, 38, 240, 182, 4, 39, 240, 167, 244, 39, 0, 167, 228, 40, 0, 79, 120, 41, 240, 137, 212, 41, 240, 122, 196, 42, 240, 107, 180, 43, 240, 92, 164, 44, 240, 77, 148, 45, 240, 62, 132, 46, 240, 47, 116, 47, 240, 32, 100, 48, 112, 76, 93, 49, 112, 39, 114, 50, 112, 46, 61, 51, 112, 9, 82, 52, 112, 16, 29, 53, 112, 235, 49, 54, 112, 242, 252, 54, 240, 7, 27, 56, 112, 212, 220, 56, 240, 233, 250, 57, 112, 182, 188, 58, 240, 203, 218, 59, 240, 210, 165, 60, 240, 173, 186, 61, 240, 180, 133, 62, 240, 143, 154, 63, 240, 150, 101, 64, 112, 172, 131, 65, 240, 120, 69, 66, 112, 142, 99, 67, 240, 90, 37, 68, 112, 112, 67, 69, 240, 60, 5, 70, 112, 82, 35, 71, 112, 89, 238, 71, 112, 52, 3, 73, 112, 59, 206, 73, 112, 22, 227, 74, 112, 29, 174, 75, 240, 50, 204, 76, 112, 255, 141, 77, 224, 172, 75, 84, 0, 156, 27, 87, 66, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 1, 3, 170, 170, 12, 0, 0, 0, 96, 141, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 16, 14, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 36, 94, 0, 224, 103, 94, 0, 16, 121, 94, 0, 224, 2, 0, 0, 0, 38, 1, 0, 0, 70, 1, 0, 0, 3, 0, 0, 0, 144, 213, 22, 186, 128, 29, 136, 203, 112, 238, 86, 210, 170, 170, 170, 170, 3, 0, 0, 0, 1, 2, 1, 170, 6, 0, 0, 0, 240, 111, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 139, 94, 0, 224, 142, 94, 0, 224, 147, 94, 0, 16, 149, 94, 0, 224, 8, 0, 0, 0, 128, 244, 156, 193, 112, 48, 22, 194, 0, 231, 242, 203, 112, 37, 169, 208, 0, 57, 108, 226, 240, 162, 213, 226, 128, 70, 117, 15, 240, 122, 102, 16, 4, 0, 0, 0, 255, 255, 255, 255, 16, 220, 225, 20, 255, 255, 255, 255, 144, 63, 31, 123, 170, 170, 170, 170, 170, 170, 170, 170, 10, 0, 0, 0, 3, 1, 2, 1, 4, 1, 2, 1, 2, 1, 170, 170, 10, 0, 0, 0, 240, 31, 255, 255, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 112, 113, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 163, 94, 0, 224, 172, 94, 0, 224, 179, 94, 0, 16, 183, 94, 0, 224, 2, 0, 0, 0, 41, 1, 0, 0, 217, 1, 0, 0, 51, 0, 0, 0, 184, 30, 119, 165, 224, 175, 237, 9, 208, 146, 221, 10, 224, 100, 250, 11, 80, 198, 190, 12, 96, 57, 164, 13, 208, 225, 138, 14, 96, 27, 132, 15, 208, 79, 117, 16, 96, 253, 99, 17, 80, 224, 83, 18, 224, 25, 77, 19, 80, 194, 51, 20, 96, 193, 35, 21, 80, 164, 19, 22, 96, 163, 3, 23, 80, 134, 243, 23, 96, 133, 227, 24, 80, 104, 211, 25, 96, 103, 195, 26, 208, 132, 188, 27, 224, 131, 172, 28, 208, 102, 156, 29, 224, 101, 140, 30, 208, 72, 124, 31, 224, 71, 108, 32, 208, 42, 92, 33, 224, 41, 76, 34, 208, 12, 60, 35, 224, 11, 44, 36, 208, 238, 27, 37, 224, 237, 11, 38, 80, 11, 5, 39, 96, 10, 245, 39, 80, 237, 228, 40, 96, 236, 212, 41, 80, 207, 196, 42, 96, 206, 180, 43, 80, 177, 164, 44, 96, 176, 148, 45, 80, 147, 132, 46, 96, 146, 116, 47, 80, 117, 100, 48, 224, 174, 93, 49, 208, 145, 77, 50, 224, 144, 61, 51, 208, 115, 45, 52, 224, 114, 29, 53, 16, 120, 50, 54, 16, 127, 253, 54, 144, 148, 27, 56, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 51, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 72, 31, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 32, 28, 0, 112, 183, 3, 0, 96, 208, 7, 0, 112, 201, 94, 0, 224, 204, 94, 0, 224, 3, 95, 0, 16, 17, 95, 0, 224, 64, 0, 0, 0, 192, 32, 24, 170, 32, 225, 163, 181, 144, 111, 39, 21, 0, 164, 24, 22, 16, 163, 8, 23, 128, 215, 249, 23, 144, 214, 233, 24, 0, 11, 219, 25, 144, 91, 204, 26, 176, 104, 188, 27, 176, 89, 172, 28, 176, 74, 156, 29, 176, 59, 140, 30, 176, 44, 124, 31, 176, 29, 108, 32, 176, 14, 92, 33, 176, 255, 75, 34, 176, 240, 59, 35, 176, 225, 43, 36, 176, 210, 27, 37, 176, 195, 11, 38, 48, 239, 4, 39, 48, 224, 244, 39, 64, 223, 228, 40, 64, 135, 120, 41, 48, 194, 212, 41, 48, 179, 196, 42, 48, 164, 180, 43, 48, 149, 164, 44, 48, 134, 148, 45, 48, 119, 132, 46, 48, 104, 116, 47, 48, 89, 100, 48, 176, 132, 93, 49, 176, 95, 114, 50, 176, 102, 61, 51, 176, 65, 82, 52, 176, 72, 29, 53, 176, 35, 50, 54, 176, 42, 253, 54, 48, 64, 27, 56, 176, 12, 221, 56, 48, 34, 251, 57, 176, 238, 188, 58, 48, 4, 219, 59, 48, 11, 166, 60, 48, 230, 186, 61, 48, 237, 133, 62, 48, 200, 154, 63, 48, 207, 101, 64, 176, 228, 131, 65, 48, 177, 69, 66, 176, 198, 99, 67, 48, 147, 37, 68, 176, 168, 67, 69, 48, 117, 5, 70, 176, 138, 35, 71, 176, 145, 238, 71, 176, 108, 3, 73, 176, 115, 206, 73, 176, 78, 227, 74, 176, 85, 174, 75, 64, 121, 204, 76, 192, 69, 142, 77, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 64, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 10, 0, 0, 0, 192, 81, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 35, 95, 0, 224, 103, 95, 0, 16, 120, 95, 0, 224, 67, 0, 0, 0, 36, 25, 219, 161, 32, 225, 163, 181, 144, 111, 39, 21, 0, 164, 24, 22, 16, 163, 8, 23, 128, 215, 249, 23, 144, 214, 233, 24, 0, 11, 219, 25, 144, 91, 204, 26, 176, 104, 188, 27, 176, 89, 172, 28, 176, 74, 156, 29, 176, 59, 140, 30, 176, 44, 124, 31, 176, 29, 108, 32, 176, 14, 92, 33, 176, 255, 75, 34, 176, 240, 59, 35, 176, 225, 43, 36, 176, 210, 27, 37, 176, 195, 11, 38, 48, 239, 4, 39, 48, 224, 244, 39, 64, 223, 228, 40, 64, 135, 120, 41, 48, 194, 212, 41, 48, 179, 196, 42, 48, 164, 180, 43, 0, 78, 254, 43, 64, 163, 164, 44, 64, 148, 148, 45, 64, 133, 132, 46, 64, 118, 116, 47, 64, 103, 100, 48, 192, 146, 93, 49, 192, 109, 114, 50, 192, 116, 61, 51, 192, 79, 82, 52, 192, 86, 29, 53, 192, 49, 50, 54, 192, 56, 253, 54, 64, 78, 27, 56, 192, 26, 221, 56, 64, 48, 251, 57, 192, 252, 188, 58, 64, 18, 219, 59, 64, 25, 166, 60, 64, 244, 186, 61, 64, 251, 133, 62, 64, 214, 154, 63, 64, 221, 101, 64, 192, 242, 131, 65, 64, 191, 69, 66, 192, 212, 99, 67, 64, 161, 37, 68, 192, 182, 67, 69, 64, 131, 5, 70, 192, 152, 35, 71, 192, 159, 238, 71, 192, 122, 3, 73, 192, 129, 206, 73, 192, 92, 227, 74, 192, 99, 174, 75, 64, 121, 204, 76, 192, 69, 142, 77, 48, 243, 75, 84, 192, 204, 147, 87, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 67, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 3, 170, 10, 0, 0, 0, 188, 77, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 136, 95, 0, 224, 207, 95, 0, 16, 225, 95, 0, 224, 65, 0, 0, 0, 182, 64, 179, 161, 48, 239, 163, 181, 160, 125, 39, 21, 16, 178, 24, 22, 32, 177, 8, 23, 144, 229, 249, 23, 160, 228, 233, 24, 16, 25, 219, 25, 160, 105, 204, 26, 192, 118, 188, 27, 192, 103, 172, 28, 192, 88, 156, 29, 192, 73, 140, 30, 192, 58, 124, 31, 192, 43, 108, 32, 192, 28, 92, 33, 192, 13, 76, 34, 192, 254, 59, 35, 192, 239, 43, 36, 192, 224, 27, 37, 192, 209, 11, 38, 64, 253, 4, 39, 64, 238, 244, 39, 80, 237, 228, 40, 80, 149, 120, 41, 64, 208, 212, 41, 64, 193, 196, 42, 64, 178, 180, 43, 64, 163, 164, 44, 64, 148, 148, 45, 64, 133, 132, 46, 64, 118, 116, 47, 64, 103, 100, 48, 192, 146, 93, 49, 192, 109, 114, 50, 192, 116, 61, 51, 192, 79, 82, 52, 192, 86, 29, 53, 192, 49, 50, 54, 192, 56, 253, 54, 64, 78, 27, 56, 192, 26, 221, 56, 64, 48, 251, 57, 192, 252, 188, 58, 64, 18, 219, 59, 64, 25, 166, 60, 64, 244, 186, 61, 64, 251, 133, 62, 64, 214, 154, 63, 64, 221, 101, 64, 192, 242, 131, 65, 64, 191, 69, 66, 192, 212, 99, 67, 64, 161, 37, 68, 192, 182, 67, 69, 64, 131, 5, 70, 192, 152, 35, 71, 192, 159, 238, 71, 192, 122, 3, 73, 192, 129, 206, 73, 192, 92, 227, 74, 192, 99, 174, 75, 64, 121, 204, 76, 192, 69, 142, 77, 48, 243, 75, 84, 65, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 3, 170, 170, 170, 12, 0, 0, 0, 202, 68, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 241, 95, 0, 224, 51, 96, 0, 16, 69, 96, 0, 224, 51, 0, 0, 0, 220, 147, 25, 170, 80, 11, 164, 181, 176, 139, 39, 21, 32, 192, 24, 22, 32, 177, 8, 23, 160, 243, 249, 23, 176, 242, 233, 24, 32, 39, 219, 25, 176, 119, 204, 26, 208, 132, 188, 27, 208, 117, 172, 28, 208, 102, 156, 29, 208, 87, 140, 30, 208, 72, 124, 31, 208, 57, 108, 32, 208, 42, 92, 33, 208, 27, 76, 34, 208, 12, 60, 35, 208, 253, 43, 36, 224, 252, 27, 37, 224, 237, 11, 38, 96, 25, 5, 39, 96, 10, 245, 39, 96, 251, 228, 40, 96, 163, 120, 41, 80, 222, 212, 41, 96, 221, 196, 42, 96, 206, 180, 43, 96, 191, 164, 44, 96, 176, 148, 45, 96, 161, 132, 46, 96, 146, 116, 47, 96, 131, 100, 48, 224, 174, 93, 49, 224, 137, 114, 50, 224, 144, 61, 51, 224, 107, 82, 52, 224, 114, 29, 53, 224, 77, 50, 54, 224, 84, 253, 54, 96, 106, 27, 56, 224, 54, 221, 56, 96, 76, 251, 57, 224, 24, 189, 58, 96, 46, 219, 59, 96, 53, 166, 60, 96, 16, 187, 61, 96, 23, 134, 62, 96, 242, 154, 63, 96, 249, 101, 64, 224, 14, 132, 65, 51, 0, 0, 0, 1, 4, 5, 6, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 2, 3, 2, 3, 2, 4, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 4, 170, 14, 0, 0, 0, 36, 48, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 87, 96, 0, 224, 139, 96, 0, 16, 153, 96, 0, 224, 7, 0, 0, 0, 0, 223, 22, 186, 8, 164, 121, 203, 112, 238, 86, 210, 8, 198, 60, 215, 0, 38, 255, 218, 136, 190, 181, 244, 128, 116, 218, 33, 170, 170, 170, 170, 170, 170, 170, 170, 7, 0, 0, 0, 2, 4, 2, 3, 2, 3, 1, 170, 10, 0, 0, 0, 128, 102, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 120, 105, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 173, 96, 0, 224, 183, 96, 0, 16, 186, 96, 0, 224, 4, 0, 0, 0, 156, 241, 215, 139, 248, 22, 230, 146, 112, 2, 206, 85, 112, 117, 236, 90, 4, 0, 0, 0, 1, 2, 1, 2, 6, 0, 0, 0, 228, 117, 0, 0, 0, 0, 0, 0, 136, 119, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 202, 96, 0, 224, 207, 96, 0, 16, 209, 96, 0, 224, 2, 0, 0, 0, 248, 0, 0, 0, 49, 1, 0, 0, 2, 0, 0, 0, 48, 157, 242, 161, 192, 146, 138, 4, 2, 0, 0, 0, 2, 1, 170, 170, 6, 0, 0, 0, 80, 48, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 221, 96, 0, 224, 224, 96, 0, 224, 227, 96, 0, 16, 229, 96, 0, 224, 51, 0, 0, 0, 92, 136, 25, 170, 64, 253, 163, 181, 176, 139, 39, 21, 32, 192, 24, 22, 32, 177, 8, 23, 160, 243, 249, 23, 176, 242, 233, 24, 32, 39, 219, 25, 176, 119, 204, 26, 208, 132, 188, 27, 208, 117, 172, 28, 208, 102, 156, 29, 208, 87, 140, 30, 208, 72, 124, 31, 208, 57, 108, 32, 208, 42, 92, 33, 208, 27, 76, 34, 208, 12, 60, 35, 208, 253, 43, 36, 208, 238, 27, 37, 208, 223, 11, 38, 80, 11, 5, 39, 80, 252, 244, 39, 96, 251, 228, 40, 96, 163, 120, 41, 80, 222, 212, 41, 80, 207, 196, 42, 80, 192, 180, 43, 80, 177, 164, 44, 80, 162, 148, 45, 80, 147, 132, 46, 80, 132, 116, 47, 80, 117, 100, 48, 208, 160, 93, 49, 208, 123, 114, 50, 208, 130, 61, 51, 208, 93, 82, 52, 208, 100, 29, 53, 208, 63, 50, 54, 208, 70, 253, 54, 80, 92, 27, 56, 208, 40, 221, 56, 80, 62, 251, 57, 208, 10, 189, 58, 80, 32, 219, 59, 80, 39, 166, 60, 80, 2, 187, 61, 80, 9, 134, 62, 80, 228, 154, 63, 80, 235, 101, 64, 208, 0, 132, 65, 51, 0, 0, 0, 1, 3, 4, 5, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 5, 170, 12, 0, 0, 0, 164, 59, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 243, 96, 0, 224, 39, 97, 0, 16, 53, 97, 0, 224, 52, 0, 0, 0, 160, 134, 25, 170, 64, 253, 163, 181, 176, 139, 39, 21, 32, 192, 24, 22, 32, 177, 8, 23, 160, 243, 249, 23, 176, 242, 233, 24, 32, 39, 219, 25, 176, 119, 204, 26, 208, 132, 188, 27, 208, 117, 172, 28, 208, 102, 156, 29, 208, 87, 140, 30, 208, 72, 124, 31, 208, 57, 108, 32, 208, 42, 92, 33, 208, 27, 76, 34, 208, 12, 60, 35, 208, 253, 43, 36, 208, 238, 27, 37, 208, 223, 11, 38, 80, 11, 5, 39, 80, 252, 244, 39, 96, 251, 228, 40, 80, 149, 120, 41, 64, 208, 212, 41, 80, 207, 196, 42, 80, 192, 180, 43, 80, 177, 164, 44, 80, 162, 148, 45, 80, 147, 132, 46, 80, 132, 116, 47, 80, 117, 100, 48, 208, 160, 93, 49, 208, 123, 114, 50, 208, 130, 61, 51, 208, 93, 82, 52, 208, 100, 29, 53, 208, 63, 50, 54, 208, 70, 253, 54, 80, 92, 27, 56, 208, 40, 221, 56, 80, 62, 251, 57, 208, 10, 189, 58, 80, 32, 219, 59, 80, 39, 166, 60, 80, 2, 187, 61, 80, 9, 134, 62, 80, 228, 154, 63, 80, 235, 101, 64, 208, 0, 132, 65, 160, 216, 27, 92, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 52, 0, 0, 0, 1, 3, 4, 5, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 3, 5, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 5, 3, 12, 0, 0, 0, 96, 61, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 71, 97, 0, 224, 127, 97, 0, 16, 141, 97, 0, 224, 3, 0, 0, 0, 238, 0, 0, 0, 34, 1, 0, 0, 53, 1, 0, 0, 1, 0, 0, 0, 180, 54, 27, 213, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 204, 43, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 159, 97, 0, 224, 163, 97, 0, 224, 167, 97, 0, 16, 169, 97, 0, 224, 66, 0, 0, 0, 184, 205, 240, 134, 240, 178, 48, 210, 80, 55, 39, 21, 192, 107, 24, 22, 208, 106, 8, 23, 64, 159, 249, 23, 80, 158, 233, 24, 192, 210, 218, 25, 80, 35, 204, 26, 112, 48, 188, 27, 112, 33, 172, 28, 112, 18, 156, 29, 112, 3, 140, 30, 112, 244, 123, 31, 112, 229, 107, 32, 112, 214, 91, 33, 112, 199, 75, 34, 112, 184, 59, 35, 112, 169, 43, 36, 112, 154, 27, 37, 112, 139, 11, 38, 240, 182, 4, 39, 240, 167, 244, 39, 0, 167, 228, 40, 0, 79, 120, 41, 240, 137, 212, 41, 240, 122, 196, 42, 240, 107, 180, 43, 240, 92, 164, 44, 240, 77, 148, 45, 240, 62, 132, 46, 240, 47, 116, 47, 240, 32, 100, 48, 112, 76, 93, 49, 112, 39, 114, 50, 112, 46, 61, 51, 128, 23, 82, 52, 128, 30, 29, 53, 128, 249, 49, 54, 128, 0, 253, 54, 0, 22, 27, 56, 128, 226, 220, 56, 0, 248, 250, 57, 128, 196, 188, 58, 0, 218, 218, 59, 0, 225, 165, 60, 0, 188, 186, 61, 0, 195, 133, 62, 0, 158, 154, 63, 0, 165, 101, 64, 128, 186, 131, 65, 0, 135, 69, 66, 128, 156, 99, 67, 0, 105, 37, 68, 128, 126, 67, 69, 0, 75, 5, 70, 128, 96, 35, 71, 128, 103, 238, 71, 128, 66, 3, 73, 128, 73, 206, 73, 128, 36, 227, 74, 128, 43, 174, 75, 0, 65, 204, 76, 128, 13, 142, 77, 240, 186, 75, 84, 0, 178, 246, 86, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 66, 0, 0, 0, 1, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 4, 2, 4, 170, 170, 12, 0, 0, 0, 200, 133, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 181, 97, 0, 224, 251, 97, 0, 16, 13, 98, 0, 224, 24, 0, 0, 0, 55, 133, 25, 170, 64, 253, 163, 181, 176, 139, 39, 21, 32, 192, 24, 22, 32, 177, 8, 23, 160, 243, 249, 23, 176, 242, 233, 24, 32, 39, 219, 25, 176, 119, 204, 26, 208, 132, 188, 27, 208, 117, 172, 28, 208, 102, 156, 29, 208, 87, 140, 30, 208, 72, 124, 31, 208, 57, 108, 32, 208, 42, 92, 33, 208, 27, 76, 34, 208, 12, 60, 35, 208, 253, 43, 36, 208, 238, 27, 37, 208, 223, 11, 38, 80, 11, 5, 39, 80, 252, 244, 39, 80, 237, 228, 40, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 24, 0, 0, 0, 1, 2, 3, 4, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 10, 0, 0, 0, 201, 62, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 31, 98, 0, 224, 59, 98, 0, 16, 66, 98, 0, 224, 2, 0, 0, 0, 57, 1, 0, 0, 85, 2, 0, 0, 28, 0, 0, 0, 120, 240, 215, 139, 248, 22, 230, 146, 112, 143, 101, 215, 96, 157, 238, 215, 112, 250, 248, 216, 224, 45, 205, 217, 240, 138, 215, 218, 224, 15, 173, 219, 240, 226, 230, 220, 224, 241, 140, 221, 240, 41, 79, 226, 248, 183, 107, 228, 104, 24, 19, 229, 120, 3, 98, 230, 232, 76, 17, 231, 120, 112, 47, 232, 104, 244, 231, 232, 120, 82, 15, 234, 104, 214, 199, 234, 120, 52, 239, 235, 104, 184, 167, 236, 120, 22, 207, 237, 104, 154, 135, 238, 120, 113, 53, 240, 144, 96, 163, 32, 144, 103, 110, 33, 144, 66, 131, 34, 144, 73, 78, 35, 170, 170, 170, 170, 28, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 4, 3, 4, 3, 10, 0, 0, 0, 8, 119, 0, 0, 0, 0, 0, 0, 136, 119, 0, 0, 0, 0, 0, 0, 136, 119, 0, 0, 16, 14, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 82, 98, 0, 224, 85, 98, 0, 224, 115, 98, 0, 16, 123, 98, 0, 224, 6, 0, 0, 0, 2, 1, 0, 0, 3, 1, 0, 0, 13, 1, 0, 0, 58, 1, 0, 0, 127, 1, 0, 0, 35, 2, 0, 0, 26, 0, 0, 0, 128, 94, 89, 200, 112, 249, 9, 201, 0, 189, 211, 201, 240, 138, 5, 203, 0, 64, 124, 203, 240, 62, 59, 210, 128, 123, 139, 211, 240, 173, 66, 212, 0, 34, 69, 213, 240, 191, 76, 214, 0, 191, 60, 215, 112, 102, 6, 216, 128, 242, 29, 217, 240, 124, 65, 217, 32, 82, 186, 30, 144, 155, 105, 31, 160, 132, 126, 32, 144, 125, 73, 33, 32, 161, 103, 34, 144, 95, 41, 35, 32, 131, 71, 36, 16, 124, 18, 37, 32, 101, 39, 38, 16, 94, 242, 38, 32, 71, 7, 40, 16, 64, 210, 40, 2, 0, 0, 0, 255, 255, 255, 255, 41, 67, 54, 126, 170, 170, 170, 170, 27, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 215, 113, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 141, 98, 0, 224, 148, 98, 0, 224, 175, 98, 0, 224, 179, 98, 0, 16, 187, 98, 0, 224, 2, 0, 0, 0, 59, 1, 0, 0, 87, 2, 0, 0, 7, 0, 0, 0, 163, 133, 131, 134, 144, 78, 103, 186, 96, 228, 10, 192, 96, 229, 179, 202, 8, 95, 145, 203, 240, 109, 72, 210, 8, 245, 145, 22, 170, 170, 170, 170, 170, 170, 170, 170, 7, 0, 0, 0, 1, 2, 3, 4, 6, 4, 5, 170, 14, 0, 0, 0, 93, 97, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 176, 4, 0, 0, 32, 103, 0, 0, 0, 0, 0, 0, 120, 105, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 202, 98, 0, 224, 205, 98, 0, 224, 215, 98, 0, 16, 218, 98, 0, 224, 65, 0, 0, 0, 228, 51, 25, 170, 224, 168, 163, 181, 80, 55, 39, 21, 192, 107, 24, 22, 208, 106, 8, 23, 64, 159, 249, 23, 80, 158, 233, 24, 192, 210, 218, 25, 80, 35, 204, 26, 112, 48, 188, 27, 112, 33, 172, 28, 112, 18, 156, 29, 112, 3, 140, 30, 112, 244, 123, 31, 112, 229, 107, 32, 112, 214, 91, 33, 112, 199, 75, 34, 112, 184, 59, 35, 112, 169, 43, 36, 112, 154, 27, 37, 112, 139, 11, 38, 240, 182, 4, 39, 240, 167, 244, 39, 0, 167, 228, 40, 0, 79, 120, 41, 240, 137, 212, 41, 240, 122, 196, 42, 240, 107, 180, 43, 240, 92, 164, 44, 240, 77, 148, 45, 240, 62, 132, 46, 240, 47, 116, 47, 240, 32, 100, 48, 112, 76, 93, 49, 112, 39, 114, 50, 112, 46, 61, 51, 112, 9, 82, 52, 112, 16, 29, 53, 112, 235, 49, 54, 112, 242, 252, 54, 240, 7, 27, 56, 112, 212, 220, 56, 240, 233, 250, 57, 112, 182, 188, 58, 240, 203, 218, 59, 240, 210, 165, 60, 240, 173, 186, 61, 240, 180, 133, 62, 240, 143, 154, 63, 240, 150, 101, 64, 112, 172, 131, 65, 240, 120, 69, 66, 112, 142, 99, 67, 240, 90, 37, 68, 112, 112, 67, 69, 240, 60, 5, 70, 112, 82, 35, 71, 112, 89, 238, 71, 112, 52, 3, 73, 112, 59, 206, 73, 112, 22, 227, 74, 112, 29, 174, 75, 240, 50, 204, 76, 112, 255, 141, 77, 224, 172, 75, 84, 170, 170, 170, 170, 65, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 3, 170, 170, 170, 12, 0, 0, 0, 28, 144, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 16, 14, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 240, 98, 0, 224, 51, 99, 0, 16, 69, 99, 0, 224, 2, 0, 0, 0, 61, 1, 0, 0, 84, 2, 0, 0, 40, 0, 0, 0, 128, 73, 85, 195, 128, 89, 84, 210, 128, 123, 139, 211, 240, 173, 66, 212, 0, 34, 69, 213, 240, 191, 76, 214, 0, 191, 60, 215, 112, 102, 6, 216, 128, 242, 29, 217, 240, 153, 231, 217, 0, 38, 255, 218, 112, 205, 200, 219, 128, 89, 224, 220, 240, 0, 170, 221, 0, 115, 114, 222, 112, 100, 181, 223, 0, 133, 124, 224, 240, 151, 150, 225, 128, 184, 93, 226, 112, 203, 119, 227, 0, 236, 62, 228, 112, 32, 48, 229, 0, 113, 33, 230, 112, 165, 18, 231, 128, 164, 2, 232, 240, 216, 243, 232, 0, 216, 227, 233, 112, 12, 213, 234, 128, 11, 197, 235, 240, 63, 182, 236, 0, 252, 247, 237, 240, 196, 152, 238, 128, 47, 217, 239, 112, 248, 121, 240, 0, 86, 252, 7, 112, 138, 237, 8, 128, 137, 221, 9, 240, 189, 206, 10, 128, 161, 219, 17, 112, 221, 84, 18, 2, 0, 0, 0, 255, 255, 255, 255, 24, 240, 206, 116, 170, 170, 170, 170, 41, 0, 0, 0, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 8, 0, 0, 0, 232, 113, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 87, 99, 0, 224, 90, 99, 0, 224, 131, 99, 0, 224, 135, 99, 0, 16, 147, 99, 0, 224, 24, 0, 0, 0, 9, 131, 25, 170, 48, 239, 163, 181, 160, 125, 39, 21, 16, 178, 24, 22, 32, 177, 8, 23, 144, 229, 249, 23, 160, 228, 233, 24, 16, 25, 219, 25, 160, 105, 204, 26, 192, 118, 188, 27, 192, 103, 172, 28, 192, 88, 156, 29, 192, 73, 140, 30, 192, 58, 124, 31, 192, 43, 108, 32, 192, 28, 92, 33, 192, 13, 76, 34, 192, 254, 59, 35, 192, 239, 43, 36, 192, 224, 27, 37, 192, 209, 11, 38, 64, 253, 4, 39, 64, 238, 244, 39, 80, 237, 228, 40, 170, 170, 170, 170, 170, 170, 170, 170, 24, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 10, 0, 0, 0, 247, 64, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 164, 99, 0, 224, 191, 99, 0, 16, 198, 99, 0, 224, 50, 0, 0, 0, 1, 154, 25, 170, 80, 12, 218, 231, 192, 153, 39, 21, 48, 206, 24, 22, 64, 205, 8, 23, 176, 1, 250, 23, 192, 0, 234, 24, 48, 53, 219, 25, 192, 133, 204, 26, 224, 146, 188, 27, 224, 131, 172, 28, 224, 116, 156, 29, 224, 101, 140, 30, 224, 86, 124, 31, 224, 71, 108, 32, 224, 56, 92, 33, 224, 41, 76, 34, 224, 26, 60, 35, 224, 11, 44, 36, 224, 252, 27, 37, 224, 237, 11, 38, 96, 25, 5, 39, 96, 10, 245, 39, 112, 9, 229, 40, 80, 222, 212, 41, 64, 193, 196, 42, 80, 192, 180, 43, 64, 163, 164, 44, 80, 162, 148, 45, 64, 133, 132, 46, 64, 118, 116, 47, 48, 89, 100, 48, 192, 146, 93, 49, 176, 65, 82, 52, 192, 86, 29, 53, 176, 35, 50, 54, 192, 56, 253, 54, 48, 64, 27, 56, 192, 26, 221, 56, 48, 34, 251, 57, 192, 252, 188, 58, 48, 4, 219, 59, 64, 25, 166, 60, 48, 230, 186, 61, 64, 251, 133, 62, 48, 200, 154, 63, 64, 221, 101, 64, 176, 199, 221, 64, 240, 28, 132, 65, 112, 233, 69, 66, 170, 170, 170, 170, 170, 170, 170, 170, 50, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 2, 1, 2, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 2, 1, 3, 170, 170, 10, 0, 0, 0, 255, 41, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 214, 99, 0, 224, 11, 100, 0, 16, 25, 100, 0, 224, 2, 0, 0, 0, 64, 1, 0, 0, 14, 2, 0, 0, 99, 0, 0, 0, 200, 18, 219, 210, 72, 162, 187, 14, 64, 45, 116, 15, 48, 64, 142, 16, 64, 58, 237, 16, 200, 103, 85, 17, 184, 74, 69, 18, 200, 236, 55, 19, 184, 21, 45, 20, 200, 118, 32, 40, 184, 157, 219, 40, 200, 156, 203, 41, 184, 34, 190, 42, 72, 208, 172, 43, 56, 86, 159, 44, 200, 3, 142, 45, 184, 137, 128, 46, 72, 55, 111, 47, 56, 189, 97, 48, 200, 106, 80, 49, 184, 240, 66, 50, 200, 239, 50, 51, 184, 117, 37, 52, 72, 35, 20, 53, 56, 169, 6, 54, 200, 86, 245, 54, 184, 220, 231, 55, 72, 138, 214, 56, 56, 16, 201, 57, 72, 15, 185, 58, 56, 149, 171, 59, 200, 66, 154, 60, 184, 200, 140, 61, 72, 118, 123, 62, 56, 252, 109, 63, 200, 169, 92, 64, 184, 47, 79, 65, 200, 46, 63, 66, 184, 180, 49, 67, 72, 201, 226, 71, 56, 79, 213, 72, 72, 78, 197, 73, 56, 212, 183, 74, 200, 129, 166, 75, 184, 7, 153, 76, 72, 181, 135, 77, 56, 59, 122, 78, 200, 232, 104, 79, 184, 110, 91, 80, 200, 109, 75, 81, 184, 243, 61, 82, 72, 161, 44, 83, 56, 39, 31, 84, 200, 212, 13, 85, 184, 90, 0, 86, 72, 8, 239, 86, 56, 142, 225, 87, 72, 141, 209, 88, 56, 19, 196, 89, 200, 192, 178, 90, 184, 70, 165, 91, 72, 244, 147, 92, 56, 122, 134, 93, 200, 39, 117, 94, 184, 173, 103, 95, 200, 172, 87, 96, 184, 50, 74, 97, 72, 224, 56, 98, 56, 102, 43, 99, 200, 19, 26, 100, 184, 153, 12, 101, 72, 71, 251, 101, 56, 205, 237, 102, 72, 204, 221, 103, 56, 82, 208, 104, 200, 255, 190, 105, 184, 133, 177, 106, 72, 51, 160, 107, 56, 185, 146, 108, 200, 102, 129, 109, 184, 236, 115, 110, 72, 154, 98, 111, 56, 32, 85, 112, 72, 31, 69, 113, 56, 165, 55, 114, 200, 82, 38, 115, 184, 216, 24, 116, 72, 134, 7, 117, 56, 12, 250, 117, 200, 185, 232, 118, 184, 63, 219, 119, 200, 62, 203, 120, 184, 196, 189, 121, 72, 114, 172, 122, 56, 248, 158, 123, 200, 165, 141, 124, 184, 43, 128, 125, 72, 217, 110, 126, 56, 95, 97, 127, 204, 0, 0, 0, 0, 0, 0, 0, 72, 94, 81, 128, 0, 0, 0, 0, 56, 228, 67, 129, 0, 0, 0, 0, 200, 145, 50, 130, 0, 0, 0, 0, 184, 23, 37, 131, 0, 0, 0, 0, 72, 197, 19, 132, 0, 0, 0, 0, 56, 75, 6, 133, 0, 0, 0, 0, 200, 248, 244, 133, 0, 0, 0, 0, 184, 126, 231, 134, 0, 0, 0, 0, 200, 125, 215, 135, 0, 0, 0, 0, 184, 3, 202, 136, 0, 0, 0, 0, 72, 177, 184, 137, 0, 0, 0, 0, 56, 55, 171, 138, 0, 0, 0, 0, 200, 228, 153, 139, 0, 0, 0, 0, 184, 106, 140, 140, 0, 0, 0, 0, 72, 24, 123, 141, 0, 0, 0, 0, 56, 158, 109, 142, 0, 0, 0, 0, 72, 157, 93, 143, 0, 0, 0, 0, 56, 35, 80, 144, 0, 0, 0, 0, 200, 208, 62, 145, 0, 0, 0, 0, 184, 86, 49, 146, 0, 0, 0, 0, 72, 4, 32, 147, 0, 0, 0, 0, 56, 138, 18, 148, 0, 0, 0, 0, 200, 55, 1, 149, 0, 0, 0, 0, 184, 189, 243, 149, 0, 0, 0, 0, 200, 188, 227, 150, 0, 0, 0, 0, 184, 66, 214, 151, 0, 0, 0, 0, 72, 240, 196, 152, 0, 0, 0, 0, 56, 118, 183, 153, 0, 0, 0, 0, 200, 35, 166, 154, 0, 0, 0, 0, 184, 169, 152, 155, 0, 0, 0, 0, 72, 87, 135, 156, 0, 0, 0, 0, 56, 221, 121, 157, 0, 0, 0, 0, 72, 220, 105, 158, 0, 0, 0, 0, 56, 98, 92, 159, 0, 0, 0, 0, 200, 15, 75, 160, 0, 0, 0, 0, 184, 149, 61, 161, 0, 0, 0, 0, 72, 67, 44, 162, 0, 0, 0, 0, 56, 201, 30, 163, 0, 0, 0, 0, 200, 118, 13, 164, 0, 0, 0, 0, 184, 252, 255, 164, 0, 0, 0, 0, 200, 251, 239, 165, 0, 0, 0, 0, 184, 129, 226, 166, 0, 0, 0, 0, 72, 47, 209, 167, 0, 0, 0, 0, 56, 181, 195, 168, 0, 0, 0, 0, 200, 98, 178, 169, 0, 0, 0, 0, 184, 232, 164, 170, 0, 0, 0, 0, 72, 150, 147, 171, 0, 0, 0, 0, 56, 28, 134, 172, 0, 0, 0, 0, 200, 201, 116, 173, 0, 0, 0, 0, 184, 79, 103, 174, 0, 0, 0, 0, 200, 78, 87, 175, 0, 0, 0, 0, 184, 212, 73, 176, 0, 0, 0, 0, 72, 130, 56, 177, 0, 0, 0, 0, 56, 8, 43, 178, 0, 0, 0, 0, 200, 181, 25, 179, 0, 0, 0, 0, 184, 59, 12, 180, 0, 0, 0, 0, 72, 233, 250, 180, 0, 0, 0, 0, 56, 111, 237, 181, 0, 0, 0, 0, 72, 110, 221, 182, 0, 0, 0, 0, 56, 244, 207, 183, 0, 0, 0, 0, 200, 161, 190, 184, 0, 0, 0, 0, 184, 39, 177, 185, 0, 0, 0, 0, 72, 213, 159, 186, 0, 0, 0, 0, 56, 91, 146, 187, 0, 0, 0, 0, 200, 8, 129, 188, 0, 0, 0, 0, 184, 142, 115, 189, 0, 0, 0, 0, 200, 141, 99, 190, 0, 0, 0, 0, 184, 19, 86, 191, 0, 0, 0, 0, 72, 193, 68, 192, 0, 0, 0, 0, 56, 71, 55, 193, 0, 0, 0, 0, 200, 244, 37, 194, 0, 0, 0, 0, 184, 122, 24, 195, 0, 0, 0, 0, 72, 40, 7, 196, 0, 0, 0, 0, 56, 174, 249, 196, 0, 0, 0, 0, 72, 173, 233, 197, 0, 0, 0, 0, 56, 51, 220, 198, 0, 0, 0, 0, 200, 224, 202, 199, 0, 0, 0, 0, 184, 102, 189, 200, 0, 0, 0, 0, 72, 20, 172, 201, 0, 0, 0, 0, 56, 154, 158, 202, 0, 0, 0, 0, 200, 71, 141, 203, 0, 0, 0, 0, 184, 205, 127, 204, 0, 0, 0, 0, 200, 204, 111, 205, 0, 0, 0, 0, 184, 82, 98, 206, 0, 0, 0, 0, 72, 0, 81, 207, 0, 0, 0, 0, 56, 134, 67, 208, 0, 0, 0, 0, 200, 51, 50, 209, 0, 0, 0, 0, 184, 185, 36, 210, 0, 0, 0, 0, 72, 103, 19, 211, 0, 0, 0, 0, 56, 237, 5, 212, 0, 0, 0, 0, 72, 236, 245, 212, 0, 0, 0, 0, 56, 114, 232, 213, 0, 0, 0, 0, 200, 31, 215, 214, 0, 0, 0, 0, 184, 165, 201, 215, 0, 0, 0, 0, 72, 83, 184, 216, 0, 0, 0, 0, 56, 217, 170, 217, 0, 0, 0, 0, 200, 134, 153, 218, 0, 0, 0, 0, 184, 12, 140, 219, 0, 0, 0, 0, 200, 11, 124, 220, 0, 0, 0, 0, 184, 145, 110, 221, 0, 0, 0, 0, 72, 63, 93, 222, 0, 0, 0, 0, 56, 197, 79, 223, 170, 170, 170, 170, 170, 170, 170, 170, 201, 0, 0, 0, 1, 3, 4, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 10, 0, 0, 0, 56, 48, 0, 0, 0, 0, 0, 0, 56, 49, 0, 0, 0, 0, 0, 0, 56, 49, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 91, 0, 66, 0, 54, 0, 170, 170, 56, 49, 0, 112, 101, 3, 0, 96, 41, 8, 0, 112, 41, 100, 0, 224, 44, 100, 0, 224, 144, 100, 0, 224, 95, 101, 0, 16, 147, 101, 0, 224, 2, 0, 0, 0, 66, 1, 0, 0, 67, 1, 0, 0, 2, 0, 0, 0, 116, 21, 230, 213, 168, 77, 97, 33, 170, 170, 170, 170, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 12, 84, 0, 0, 0, 0, 0, 0, 88, 77, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 171, 101, 0, 224, 174, 101, 0, 224, 179, 101, 0, 16, 181, 101, 0, 224, 3, 0, 0, 0, 68, 1, 0, 0, 16, 2, 0, 0, 18, 2, 0, 0, 8, 0, 0, 0, 112, 2, 62, 215, 240, 89, 237, 215, 112, 250, 248, 216, 240, 59, 205, 217, 240, 0, 7, 219, 240, 29, 173, 219, 240, 226, 230, 220, 240, 255, 140, 221, 2, 0, 0, 0, 255, 255, 255, 255, 112, 164, 194, 101, 9, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 3, 131, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 195, 101, 0, 224, 199, 101, 0, 224, 208, 101, 0, 224, 211, 101, 0, 16, 215, 101, 0, 224, 67, 0, 0, 0, 217, 78, 229, 161, 32, 225, 163, 181, 144, 111, 39, 21, 0, 164, 24, 22, 16, 163, 8, 23, 128, 215, 249, 23, 144, 214, 233, 24, 0, 11, 219, 25, 144, 91, 204, 26, 176, 104, 188, 27, 176, 89, 172, 28, 176, 74, 156, 29, 176, 59, 140, 30, 176, 44, 124, 31, 176, 29, 108, 32, 176, 14, 92, 33, 176, 255, 75, 34, 176, 240, 59, 35, 176, 225, 43, 36, 176, 210, 27, 37, 176, 195, 11, 38, 48, 239, 4, 39, 48, 224, 244, 39, 64, 223, 228, 40, 64, 135, 120, 41, 48, 194, 212, 41, 48, 179, 196, 42, 48, 164, 180, 43, 48, 149, 164, 44, 48, 134, 148, 45, 48, 119, 132, 46, 48, 104, 116, 47, 48, 89, 100, 48, 176, 132, 93, 49, 176, 95, 114, 50, 176, 102, 61, 51, 176, 65, 82, 52, 176, 72, 29, 53, 176, 35, 50, 54, 176, 42, 253, 54, 48, 64, 27, 56, 176, 12, 221, 56, 48, 34, 251, 57, 176, 238, 188, 58, 48, 4, 219, 59, 48, 11, 166, 60, 176, 233, 206, 60, 64, 244, 186, 61, 64, 251, 133, 62, 64, 214, 154, 63, 64, 221, 101, 64, 192, 242, 131, 65, 64, 191, 69, 66, 192, 212, 99, 67, 64, 161, 37, 68, 192, 182, 67, 69, 64, 131, 5, 70, 192, 152, 35, 71, 192, 159, 238, 71, 192, 122, 3, 73, 192, 129, 206, 73, 192, 92, 227, 74, 192, 99, 174, 75, 64, 121, 204, 76, 192, 69, 142, 77, 48, 243, 75, 84, 192, 248, 73, 87, 170, 170, 170, 170, 67, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 3, 170, 10, 0, 0, 0, 167, 79, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 16, 14, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 230, 101, 0, 224, 43, 102, 0, 16, 61, 102, 0, 224, 2, 0, 0, 0, 71, 1, 0, 0, 72, 1, 0, 0, 50, 0, 0, 0, 76, 238, 211, 134, 144, 220, 11, 15, 128, 200, 233, 24, 240, 252, 218, 25, 128, 77, 204, 26, 112, 48, 188, 27, 128, 47, 172, 28, 112, 18, 156, 29, 128, 17, 140, 30, 112, 244, 123, 31, 128, 243, 107, 32, 112, 214, 91, 33, 128, 213, 75, 34, 112, 184, 59, 35, 128, 183, 43, 36, 112, 154, 27, 37, 128, 153, 11, 38, 240, 182, 4, 39, 0, 182, 244, 39, 240, 152, 228, 40, 0, 152, 212, 41, 240, 122, 196, 42, 0, 122, 180, 43, 240, 92, 164, 44, 0, 92, 148, 45, 240, 62, 132, 46, 0, 62, 116, 47, 240, 32, 100, 48, 128, 90, 93, 49, 112, 61, 77, 50, 128, 60, 61, 51, 112, 31, 45, 52, 128, 30, 29, 53, 112, 1, 13, 54, 160, 179, 233, 58, 144, 172, 180, 59, 160, 171, 164, 60, 144, 142, 148, 61, 160, 141, 132, 62, 144, 112, 116, 63, 160, 111, 100, 64, 144, 82, 84, 65, 160, 81, 68, 66, 144, 52, 52, 67, 160, 51, 36, 68, 16, 81, 29, 69, 160, 154, 21, 85, 112, 97, 5, 86, 160, 124, 245, 86, 112, 67, 229, 87, 50, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 52, 100, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 77, 102, 0, 224, 80, 102, 0, 224, 131, 102, 0, 16, 145, 102, 0, 224, 2, 0, 0, 0, 26, 1, 0, 0, 73, 1, 0, 0, 1, 0, 0, 0, 100, 186, 254, 176, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 28, 82, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 161, 102, 0, 224, 164, 102, 0, 224, 167, 102, 0, 16, 169, 102, 0, 224, 66, 0, 0, 0, 186, 221, 219, 161, 0, 197, 163, 181, 112, 83, 39, 21, 192, 107, 24, 22, 208, 106, 8, 23, 64, 159, 249, 23, 80, 158, 233, 24, 192, 210, 218, 25, 80, 35, 204, 26, 112, 48, 188, 27, 112, 33, 172, 28, 112, 18, 156, 29, 112, 3, 140, 30, 112, 244, 123, 31, 112, 229, 107, 32, 112, 214, 91, 33, 112, 199, 75, 34, 112, 184, 59, 35, 112, 169, 43, 36, 112, 154, 27, 37, 112, 139, 11, 38, 240, 182, 4, 39, 240, 167, 244, 39, 0, 167, 228, 40, 0, 79, 120, 41, 240, 137, 212, 41, 240, 122, 196, 42, 240, 107, 180, 43, 240, 92, 164, 44, 240, 77, 148, 45, 240, 62, 132, 46, 240, 47, 116, 47, 240, 32, 100, 48, 112, 76, 93, 49, 112, 39, 114, 50, 112, 46, 61, 51, 112, 9, 82, 52, 112, 16, 29, 53, 112, 235, 49, 54, 112, 242, 252, 54, 240, 7, 27, 56, 112, 212, 220, 56, 240, 233, 250, 57, 112, 182, 188, 58, 240, 203, 218, 59, 240, 210, 165, 60, 240, 173, 186, 61, 240, 180, 133, 62, 240, 143, 154, 63, 240, 150, 101, 64, 112, 172, 131, 65, 240, 120, 69, 66, 112, 142, 99, 67, 240, 90, 37, 68, 112, 112, 67, 69, 240, 60, 5, 70, 112, 82, 35, 71, 112, 89, 238, 71, 112, 52, 3, 73, 112, 59, 206, 73, 112, 22, 227, 74, 112, 29, 174, 75, 240, 50, 204, 76, 112, 255, 141, 77, 64, 244, 109, 78, 240, 186, 75, 84, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 66, 0, 0, 0, 1, 2, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 4, 3, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 7, 5, 3, 170, 170, 16, 0, 0, 0, 70, 134, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 16, 14, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 181, 102, 0, 224, 251, 102, 0, 16, 13, 103, 0, 224, 65, 0, 0, 0, 93, 71, 89, 167, 240, 182, 163, 181, 96, 69, 39, 21, 208, 121, 24, 22, 224, 120, 8, 23, 80, 173, 249, 23, 96, 172, 233, 24, 208, 224, 218, 25, 96, 49, 204, 26, 128, 62, 188, 27, 128, 47, 172, 28, 128, 32, 156, 29, 128, 17, 140, 30, 128, 2, 124, 31, 128, 243, 107, 32, 128, 228, 91, 33, 128, 213, 75, 34, 128, 198, 59, 35, 128, 183, 43, 36, 128, 168, 27, 37, 128, 153, 11, 38, 0, 197, 4, 39, 0, 182, 244, 39, 16, 181, 228, 40, 16, 93, 120, 41, 0, 152, 212, 41, 0, 137, 196, 42, 0, 122, 180, 43, 0, 107, 164, 44, 0, 92, 148, 45, 0, 77, 132, 46, 0, 62, 116, 47, 0, 47, 100, 48, 128, 90, 93, 49, 128, 53, 114, 50, 128, 60, 61, 51, 128, 23, 82, 52, 128, 30, 29, 53, 128, 249, 49, 54, 128, 0, 253, 54, 0, 22, 27, 56, 128, 226, 220, 56, 0, 248, 250, 57, 128, 196, 188, 58, 0, 218, 218, 59, 0, 225, 165, 60, 0, 188, 186, 61, 0, 195, 133, 62, 0, 158, 154, 63, 0, 165, 101, 64, 128, 186, 131, 65, 0, 135, 69, 66, 128, 156, 99, 67, 0, 105, 37, 68, 128, 126, 67, 69, 0, 75, 5, 70, 128, 96, 35, 71, 128, 103, 238, 71, 128, 66, 3, 73, 128, 73, 206, 73, 128, 36, 227, 74, 128, 43, 174, 75, 0, 65, 204, 76, 128, 13, 142, 77, 240, 186, 75, 84, 170, 170, 170, 170, 170, 170, 170, 170, 65, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 3, 170, 170, 170, 12, 0, 0, 0, 163, 123, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 16, 14, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 35, 103, 0, 224, 103, 103, 0, 16, 121, 103, 0, 224, 65, 0, 0, 0, 94, 234, 219, 161, 0, 197, 163, 181, 112, 83, 39, 21, 224, 135, 24, 22, 240, 134, 8, 23, 96, 187, 249, 23, 112, 186, 233, 24, 224, 238, 218, 25, 112, 63, 204, 26, 144, 76, 188, 27, 144, 61, 172, 28, 144, 46, 156, 29, 144, 31, 140, 30, 144, 16, 124, 31, 144, 1, 108, 32, 144, 242, 91, 33, 144, 227, 75, 34, 144, 212, 59, 35, 144, 197, 43, 36, 144, 182, 27, 37, 144, 167, 11, 38, 16, 211, 4, 39, 16, 196, 244, 39, 32, 195, 228, 40, 32, 107, 120, 41, 16, 166, 212, 41, 16, 151, 196, 42, 16, 136, 180, 43, 16, 121, 164, 44, 16, 106, 148, 45, 16, 91, 132, 46, 16, 76, 116, 47, 16, 61, 100, 48, 144, 104, 93, 49, 144, 67, 114, 50, 144, 74, 61, 51, 144, 37, 82, 52, 144, 44, 29, 53, 144, 7, 50, 54, 144, 14, 253, 54, 16, 36, 27, 56, 144, 240, 220, 56, 16, 6, 251, 57, 144, 210, 188, 58, 16, 232, 218, 59, 16, 239, 165, 60, 16, 202, 186, 61, 16, 209, 133, 62, 16, 172, 154, 63, 16, 179, 101, 64, 144, 200, 131, 65, 16, 149, 69, 66, 144, 170, 99, 67, 16, 119, 37, 68, 144, 140, 67, 69, 16, 89, 5, 70, 144, 110, 35, 71, 144, 117, 238, 71, 144, 80, 3, 73, 144, 87, 206, 73, 144, 50, 227, 74, 144, 57, 174, 75, 16, 79, 204, 76, 144, 27, 142, 77, 0, 201, 75, 84, 170, 170, 170, 170, 170, 170, 170, 170, 65, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 3, 170, 170, 170, 12, 0, 0, 0, 162, 121, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 16, 14, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 139, 103, 0, 224, 207, 103, 0, 16, 225, 103, 0, 224, 2, 0, 0, 0, 52, 1, 0, 0, 78, 1, 0, 0, 3, 0, 0, 0, 81, 115, 242, 161, 24, 252, 242, 203, 240, 103, 154, 209, 170, 170, 170, 170, 3, 0, 0, 0, 1, 2, 1, 170, 6, 0, 0, 0, 47, 90, 0, 0, 0, 0, 0, 0, 104, 91, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 243, 103, 0, 224, 246, 103, 0, 224, 251, 103, 0, 16, 253, 103, 0, 224, 66, 0, 0, 0, 39, 9, 95, 155, 255, 177, 18, 161, 64, 253, 163, 181, 176, 139, 39, 21, 32, 192, 24, 22, 48, 191, 8, 23, 160, 243, 249, 23, 176, 242, 233, 24, 32, 39, 219, 25, 176, 119, 204, 26, 208, 132, 188, 27, 208, 117, 172, 28, 208, 102, 156, 29, 208, 87, 140, 30, 208, 72, 124, 31, 208, 57, 108, 32, 208, 42, 92, 33, 208, 27, 76, 34, 208, 12, 60, 35, 208, 253, 43, 36, 208, 238, 27, 37, 208, 223, 11, 38, 80, 11, 5, 39, 80, 252, 244, 39, 96, 251, 228, 40, 96, 163, 120, 41, 80, 222, 212, 41, 80, 207, 196, 42, 80, 192, 180, 43, 80, 177, 164, 44, 80, 162, 148, 45, 80, 147, 132, 46, 80, 132, 116, 47, 80, 117, 100, 48, 208, 160, 93, 49, 208, 123, 114, 50, 208, 130, 61, 51, 208, 93, 82, 52, 208, 100, 29, 53, 208, 63, 50, 54, 208, 70, 253, 54, 80, 92, 27, 56, 208, 40, 221, 56, 80, 62, 251, 57, 208, 10, 189, 58, 80, 32, 219, 59, 80, 39, 166, 60, 80, 2, 187, 61, 80, 9, 134, 62, 80, 228, 154, 63, 80, 235, 101, 64, 208, 0, 132, 65, 80, 205, 69, 66, 208, 226, 99, 67, 80, 175, 37, 68, 208, 196, 67, 69, 80, 145, 5, 70, 208, 166, 35, 71, 208, 173, 238, 71, 208, 136, 3, 73, 208, 143, 206, 73, 208, 106, 227, 74, 208, 113, 174, 75, 80, 135, 204, 76, 208, 83, 142, 77, 64, 1, 76, 84, 170, 170, 170, 170, 66, 0, 0, 0, 1, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 6, 4, 170, 170, 14, 0, 0, 0, 217, 56, 0, 0, 0, 0, 0, 0, 193, 52, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 16, 14, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 11, 104, 0, 224, 79, 104, 0, 16, 97, 104, 0, 224, 2, 0, 0, 0, 80, 1, 0, 0, 28, 2, 0, 0, 62, 0, 0, 0, 72, 154, 25, 170, 80, 12, 218, 231, 192, 153, 39, 21, 48, 206, 24, 22, 64, 205, 8, 23, 176, 1, 250, 23, 192, 0, 234, 24, 48, 53, 219, 25, 192, 133, 204, 26, 224, 146, 188, 27, 224, 131, 172, 28, 224, 116, 156, 29, 224, 101, 140, 30, 224, 86, 124, 31, 224, 71, 108, 32, 224, 56, 92, 33, 224, 41, 76, 34, 224, 26, 60, 35, 224, 11, 44, 36, 224, 252, 27, 37, 224, 237, 11, 38, 96, 25, 5, 39, 96, 10, 245, 39, 112, 9, 229, 40, 112, 250, 212, 41, 112, 235, 196, 42, 112, 220, 180, 43, 112, 205, 164, 44, 112, 190, 148, 45, 112, 175, 132, 46, 112, 160, 116, 47, 112, 145, 100, 48, 224, 144, 61, 51, 224, 107, 82, 52, 224, 114, 29, 53, 224, 77, 50, 54, 224, 84, 253, 54, 96, 106, 27, 56, 224, 54, 221, 56, 96, 76, 251, 57, 224, 24, 189, 58, 96, 46, 219, 59, 96, 53, 166, 60, 96, 16, 187, 61, 96, 23, 134, 62, 96, 242, 154, 63, 96, 249, 101, 64, 224, 14, 132, 65, 96, 219, 69, 66, 224, 240, 99, 67, 96, 189, 37, 68, 224, 210, 67, 69, 96, 159, 5, 70, 224, 180, 35, 71, 224, 187, 238, 71, 224, 150, 3, 73, 224, 157, 206, 73, 224, 120, 227, 74, 224, 127, 174, 75, 96, 149, 204, 76, 224, 97, 142, 77, 96, 119, 172, 78, 62, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 10, 0, 0, 0, 184, 41, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 117, 104, 0, 224, 120, 104, 0, 224, 183, 104, 0, 16, 200, 104, 0, 224, 138, 0, 0, 0, 160, 170, 230, 146, 144, 137, 75, 155, 160, 227, 254, 155, 144, 9, 157, 156, 144, 159, 201, 157, 144, 142, 127, 158, 16, 211, 170, 159, 144, 112, 95, 160, 144, 6, 140, 161, 144, 245, 65, 162, 144, 139, 110, 163, 16, 41, 35, 164, 16, 191, 79, 165, 144, 11, 6, 170, 16, 171, 244, 170, 16, 196, 201, 173, 16, 64, 167, 174, 144, 107, 160, 175, 16, 34, 135, 176, 16, 136, 137, 177, 144, 62, 112, 178, 144, 164, 114, 179, 144, 32, 80, 180, 144, 104, 50, 183, 144, 228, 15, 184, 144, 213, 255, 184, 144, 198, 239, 185, 16, 212, 200, 188, 16, 197, 184, 189, 144, 123, 159, 190, 16, 167, 152, 191, 16, 13, 155, 192, 16, 137, 120, 193, 16, 122, 104, 194, 16, 107, 88, 195, 144, 33, 63, 196, 16, 77, 56, 197, 16, 179, 58, 198, 144, 200, 88, 199, 144, 251, 217, 199, 144, 75, 1, 201, 144, 60, 241, 201, 16, 127, 226, 202, 16, 111, 181, 203, 0, 192, 236, 203, 0, 104, 128, 204, 16, 191, 220, 204, 16, 81, 149, 205, 128, 103, 195, 205, 0, 191, 114, 206, 144, 219, 197, 206, 16, 51, 117, 207, 0, 132, 172, 207, 0, 161, 82, 208, 144, 189, 165, 208, 16, 21, 85, 209, 0, 102, 140, 209, 0, 131, 50, 210, 144, 159, 133, 210, 16, 225, 89, 211, 16, 210, 73, 212, 64, 237, 57, 213, 64, 222, 41, 214, 64, 207, 25, 215, 64, 192, 9, 216, 64, 177, 249, 216, 64, 162, 233, 217, 64, 117, 185, 220, 192, 160, 178, 221, 192, 145, 162, 222, 192, 130, 146, 223, 192, 115, 130, 224, 192, 100, 114, 225, 192, 85, 98, 226, 192, 70, 82, 227, 192, 55, 66, 228, 192, 40, 50, 229, 192, 25, 34, 230, 64, 69, 27, 231, 64, 54, 11, 232, 64, 39, 251, 232, 64, 24, 235, 233, 64, 9, 219, 234, 64, 250, 202, 235, 64, 235, 186, 236, 64, 220, 170, 237, 64, 205, 154, 238, 64, 190, 138, 239, 64, 175, 122, 240, 64, 160, 106, 241, 192, 203, 99, 242, 192, 188, 83, 243, 192, 173, 67, 244, 192, 158, 51, 245, 192, 143, 35, 246, 192, 128, 19, 247, 192, 113, 3, 248, 192, 98, 243, 248, 16, 41, 155, 13, 16, 26, 139, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 160, 38, 84, 18, 144, 9, 68, 19, 160, 8, 52, 20, 160, 249, 35, 21, 160, 234, 19, 22, 160, 219, 3, 23, 160, 204, 243, 23, 176, 203, 227, 24, 160, 174, 211, 25, 160, 159, 195, 26, 32, 203, 188, 27, 32, 188, 172, 28, 32, 173, 156, 29, 32, 158, 140, 30, 32, 143, 124, 31, 32, 128, 108, 32, 32, 113, 92, 33, 32, 98, 76, 34, 32, 83, 60, 35, 32, 68, 44, 36, 32, 53, 28, 37, 32, 38, 12, 38, 160, 81, 5, 39, 160, 66, 245, 39, 160, 51, 229, 40, 160, 36, 213, 41, 160, 21, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 144, 27, 61, 94, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 139, 0, 0, 0, 4, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 1, 2, 3, 2, 1, 2, 3, 2, 1, 2, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 7, 6, 5, 6, 5, 6, 5, 6, 5, 170, 16, 0, 0, 0, 240, 231, 255, 255, 0, 0, 0, 0, 224, 227, 255, 255, 0, 0, 0, 0, 224, 227, 255, 255, 16, 14, 0, 0, 224, 227, 255, 255, 32, 28, 0, 0, 40, 229, 255, 255, 0, 0, 0, 0, 240, 241, 255, 255, 0, 0, 0, 0, 240, 241, 255, 255, 16, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 240, 241, 255, 127, 4, 0, 0, 96, 205, 7, 0, 112, 218, 104, 0, 224, 101, 105, 0, 224, 107, 105, 0, 16, 143, 105, 0, 224, 69, 0, 0, 0, 230, 29, 195, 180, 96, 179, 32, 8, 80, 150, 16, 9, 96, 149, 0, 10, 80, 120, 240, 10, 96, 119, 224, 11, 208, 148, 217, 12, 96, 89, 192, 13, 208, 118, 185, 14, 224, 117, 169, 15, 208, 88, 153, 16, 224, 87, 137, 17, 208, 58, 121, 18, 224, 57, 105, 19, 208, 28, 89, 20, 224, 27, 73, 21, 208, 254, 56, 22, 224, 253, 40, 23, 80, 27, 34, 24, 224, 223, 8, 25, 80, 253, 1, 26, 96, 252, 241, 26, 80, 223, 225, 27, 96, 222, 209, 28, 80, 193, 193, 29, 96, 192, 177, 30, 80, 163, 161, 31, 224, 242, 117, 32, 80, 133, 129, 33, 224, 212, 85, 34, 208, 161, 106, 35, 224, 182, 53, 36, 208, 131, 74, 37, 224, 152, 21, 38, 208, 101, 42, 39, 96, 181, 254, 39, 208, 71, 10, 41, 96, 151, 222, 41, 208, 41, 234, 42, 96, 121, 190, 43, 80, 70, 211, 44, 96, 91, 158, 45, 80, 40, 179, 46, 96, 61, 126, 47, 80, 10, 147, 48, 224, 89, 103, 49, 80, 236, 114, 50, 224, 59, 71, 51, 80, 206, 82, 52, 224, 29, 39, 53, 80, 176, 50, 54, 224, 255, 6, 55, 208, 204, 27, 56, 224, 225, 230, 56, 208, 174, 251, 57, 224, 195, 198, 58, 208, 144, 219, 59, 96, 224, 175, 60, 208, 114, 187, 61, 96, 194, 143, 62, 208, 84, 155, 63, 96, 164, 111, 64, 80, 113, 132, 65, 96, 134, 79, 66, 80, 83, 100, 67, 96, 104, 47, 68, 80, 53, 68, 69, 224, 154, 243, 69, 208, 81, 45, 71, 170, 170, 170, 170, 170, 170, 170, 170, 69, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 58, 195, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 192, 199, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 171, 105, 0, 224, 243, 105, 0, 16, 6, 106, 0, 224, 36, 0, 0, 0, 240, 92, 4, 166, 32, 247, 65, 212, 0, 54, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 36, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 8, 0, 0, 0, 144, 241, 255, 255, 0, 0, 0, 0, 240, 241, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 0, 0, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 23, 106, 0, 224, 63, 106, 0, 16, 73, 106, 0, 224, 4, 0, 0, 0, 160, 170, 230, 146, 32, 156, 149, 204, 16, 124, 116, 210, 64, 247, 23, 11, 170, 170, 170, 170, 170, 170, 170, 170, 4, 0, 0, 0, 1, 2, 1, 3, 8, 0, 0, 0, 244, 233, 255, 255, 0, 0, 0, 0, 224, 227, 255, 255, 0, 0, 0, 0, 224, 227, 255, 255, 16, 14, 0, 0, 240, 241, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 92, 106, 0, 224, 99, 106, 0, 16, 101, 106, 0, 224, 2, 0, 0, 0, 85, 1, 0, 0, 86, 1, 0, 0, 33, 0, 0, 0, 88, 164, 109, 139, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 33, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 168, 249, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 0, 0, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 115, 106, 0, 224, 118, 106, 0, 224, 155, 106, 0, 16, 165, 106, 0, 224, 138, 0, 0, 0, 144, 156, 230, 146, 128, 123, 75, 155, 144, 213, 254, 155, 128, 251, 156, 156, 128, 145, 201, 157, 128, 128, 127, 158, 0, 197, 170, 159, 128, 98, 95, 160, 128, 248, 139, 161, 128, 231, 65, 162, 128, 125, 110, 163, 0, 27, 35, 164, 0, 177, 79, 165, 128, 253, 5, 170, 0, 157, 244, 170, 0, 182, 201, 173, 0, 50, 167, 174, 128, 93, 160, 175, 0, 20, 135, 176, 0, 122, 137, 177, 128, 48, 112, 178, 128, 150, 114, 179, 128, 18, 80, 180, 128, 90, 50, 183, 128, 214, 15, 184, 128, 199, 255, 184, 128, 184, 239, 185, 0, 198, 200, 188, 0, 183, 184, 189, 128, 109, 159, 190, 0, 153, 152, 191, 0, 255, 154, 192, 0, 123, 120, 193, 0, 108, 104, 194, 0, 93, 88, 195, 128, 19, 63, 196, 0, 63, 56, 197, 0, 165, 58, 198, 128, 186, 88, 199, 128, 237, 217, 199, 128, 61, 1, 201, 128, 46, 241, 201, 0, 113, 226, 202, 0, 97, 181, 203, 240, 177, 236, 203, 240, 89, 128, 204, 0, 177, 220, 204, 0, 67, 149, 205, 112, 89, 195, 205, 240, 176, 114, 206, 128, 205, 197, 206, 0, 37, 117, 207, 240, 117, 172, 207, 240, 146, 82, 208, 128, 175, 165, 208, 0, 7, 85, 209, 240, 87, 140, 209, 240, 116, 50, 210, 128, 145, 133, 210, 0, 211, 89, 211, 0, 196, 73, 212, 48, 223, 57, 213, 48, 208, 41, 214, 48, 193, 25, 215, 48, 178, 9, 216, 48, 163, 249, 216, 48, 148, 233, 217, 48, 103, 185, 220, 176, 146, 178, 221, 176, 131, 162, 222, 176, 116, 146, 223, 176, 101, 130, 224, 176, 86, 114, 225, 176, 71, 98, 226, 176, 56, 82, 227, 176, 41, 66, 228, 176, 26, 50, 229, 176, 11, 34, 230, 48, 55, 27, 231, 48, 40, 11, 232, 48, 25, 251, 232, 48, 10, 235, 233, 48, 251, 218, 234, 48, 236, 202, 235, 48, 221, 186, 236, 48, 206, 170, 237, 48, 191, 154, 238, 48, 176, 138, 239, 48, 161, 122, 240, 48, 146, 106, 241, 176, 189, 99, 242, 176, 174, 83, 243, 176, 159, 67, 244, 176, 144, 51, 245, 176, 129, 35, 246, 176, 114, 19, 247, 176, 99, 3, 248, 176, 84, 243, 248, 0, 27, 155, 13, 0, 12, 139, 14, 128, 55, 132, 15, 128, 40, 116, 16, 128, 25, 100, 17, 144, 24, 84, 18, 128, 251, 67, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 160, 189, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 138, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 1, 2, 3, 2, 1, 2, 3, 2, 1, 2, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 170, 170, 12, 0, 0, 0, 40, 240, 255, 255, 0, 0, 0, 0, 240, 241, 255, 255, 0, 0, 0, 0, 240, 241, 255, 255, 16, 14, 0, 0, 240, 241, 255, 255, 32, 28, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 0, 0, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 183, 106, 0, 224, 67, 107, 0, 16, 103, 107, 0, 224, 2, 0, 0, 0, 89, 1, 0, 0, 2, 2, 0, 0, 68, 0, 0, 0, 160, 131, 96, 139, 0, 30, 145, 156, 144, 136, 209, 157, 128, 81, 114, 158, 16, 3, 213, 159, 0, 133, 83, 160, 144, 54, 182, 161, 128, 39, 60, 164, 16, 116, 185, 164, 0, 26, 77, 198, 32, 39, 61, 199, 176, 23, 218, 199, 160, 67, 38, 201, 32, 38, 195, 201, 160, 37, 6, 203, 160, 66, 172, 203, 32, 205, 220, 204, 160, 36, 140, 205, 32, 175, 188, 206, 160, 6, 108, 207, 32, 145, 156, 208, 160, 232, 75, 209, 160, 173, 133, 210, 160, 202, 43, 211, 160, 143, 101, 212, 32, 209, 57, 213, 160, 113, 69, 214, 32, 179, 25, 215, 160, 83, 37, 216, 32, 149, 249, 216, 32, 112, 14, 218, 32, 119, 217, 218, 160, 23, 229, 219, 32, 89, 185, 220, 32, 52, 206, 221, 160, 117, 162, 222, 32, 22, 174, 223, 160, 87, 130, 224, 32, 248, 141, 225, 160, 57, 98, 226, 32, 218, 109, 227, 160, 27, 66, 228, 32, 188, 77, 229, 160, 253, 33, 230, 160, 216, 54, 231, 32, 26, 11, 232, 160, 186, 22, 233, 32, 252, 234, 233, 160, 156, 246, 234, 32, 222, 202, 235, 160, 126, 214, 236, 32, 192, 170, 237, 160, 96, 182, 238, 32, 162, 138, 239, 160, 66, 150, 240, 32, 132, 106, 241, 32, 95, 127, 242, 160, 160, 83, 243, 32, 65, 95, 244, 160, 130, 51, 245, 32, 35, 63, 246, 160, 100, 19, 247, 32, 5, 31, 248, 160, 70, 243, 248, 32, 231, 254, 249, 160, 40, 211, 250, 160, 3, 232, 251, 32, 69, 188, 252, 170, 170, 170, 170, 68, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 8, 0, 0, 0, 96, 235, 255, 255, 0, 0, 0, 0, 240, 241, 255, 255, 0, 0, 0, 0, 240, 241, 255, 255, 16, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 126, 107, 0, 224, 129, 107, 0, 224, 199, 107, 0, 16, 217, 107, 0, 224, 2, 0, 0, 0, 255, 255, 255, 255, 192, 253, 134, 105, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 192, 221, 255, 255, 0, 0, 0, 0, 224, 227, 255, 255, 0, 0, 0, 0, 3, 0, 80, 0, 66, 0, 54, 0, 233, 107, 0, 224, 239, 107, 0, 16, 241, 107, 0, 224, 69, 0, 0, 0, 60, 95, 68, 147, 192, 90, 79, 195, 48, 3, 54, 196, 192, 60, 47, 197, 48, 229, 21, 198, 64, 89, 24, 199, 176, 1, 255, 199, 64, 59, 248, 200, 176, 227, 222, 201, 64, 29, 216, 202, 176, 197, 190, 203, 64, 255, 183, 204, 48, 129, 54, 205, 64, 254, 17, 25, 176, 188, 211, 25, 32, 196, 241, 26, 48, 100, 170, 27, 32, 166, 209, 28, 48, 70, 138, 29, 176, 91, 168, 30, 64, 54, 106, 31, 176, 61, 136, 32, 64, 24, 74, 33, 176, 31, 104, 34, 64, 250, 41, 35, 176, 1, 72, 36, 64, 220, 9, 37, 48, 30, 49, 38, 64, 190, 233, 38, 48, 0, 17, 40, 192, 218, 210, 40, 48, 226, 240, 41, 192, 188, 178, 42, 48, 196, 208, 43, 192, 158, 146, 44, 48, 166, 176, 45, 192, 128, 114, 46, 48, 136, 144, 47, 192, 98, 82, 48, 176, 164, 121, 49, 64, 127, 59, 50, 176, 134, 89, 51, 64, 97, 27, 52, 176, 104, 57, 53, 64, 67, 251, 53, 176, 74, 25, 55, 64, 37, 219, 55, 176, 44, 249, 56, 64, 7, 187, 57, 208, 42, 217, 58, 224, 202, 145, 59, 80, 71, 194, 60, 224, 172, 113, 61, 80, 41, 162, 62, 96, 201, 90, 63, 80, 11, 130, 64, 96, 171, 58, 65, 80, 237, 97, 66, 96, 141, 26, 67, 80, 207, 65, 68, 96, 111, 250, 68, 80, 177, 33, 70, 96, 81, 218, 70, 208, 205, 10, 72, 224, 109, 195, 72, 208, 175, 234, 73, 224, 79, 163, 74, 208, 145, 202, 75, 224, 49, 131, 76, 170, 170, 170, 170, 170, 170, 170, 170, 69, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 4, 3, 4, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 170, 170, 170, 10, 0, 0, 0, 196, 201, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 208, 213, 255, 255, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 251, 107, 0, 224, 67, 108, 0, 16, 86, 108, 0, 224, 2, 0, 0, 0, 94, 1, 0, 0, 110, 1, 0, 0, 83, 0, 0, 0, 164, 173, 78, 156, 248, 39, 188, 156, 8, 186, 84, 203, 120, 94, 199, 203, 136, 93, 183, 204, 120, 64, 167, 205, 8, 122, 160, 206, 120, 34, 135, 207, 136, 64, 112, 3, 8, 35, 13, 4, 136, 34, 80, 5, 136, 63, 246, 5, 136, 4, 48, 7, 136, 33, 214, 7, 136, 230, 15, 9, 136, 3, 182, 9, 136, 200, 239, 10, 8, 32, 159, 11, 8, 229, 216, 12, 8, 2, 127, 13, 8, 199, 184, 14, 8, 228, 94, 15, 8, 169, 152, 16, 8, 198, 62, 17, 8, 139, 120, 18, 8, 168, 30, 19, 8, 109, 88, 20, 8, 138, 254, 20, 8, 79, 56, 22, 136, 166, 231, 22, 136, 107, 33, 24, 136, 136, 199, 24, 136, 77, 1, 26, 136, 106, 167, 26, 136, 47, 225, 27, 136, 76, 135, 28, 136, 17, 193, 29, 136, 163, 121, 30, 8, 185, 151, 31, 136, 133, 89, 32, 136, 213, 128, 33, 8, 162, 66, 34, 8, 242, 105, 35, 8, 132, 34, 36, 8, 212, 73, 37, 8, 102, 2, 38, 8, 182, 41, 39, 8, 211, 207, 39, 8, 152, 9, 41, 136, 100, 203, 41, 8, 122, 233, 42, 136, 209, 152, 43, 136, 150, 210, 44, 136, 40, 139, 45, 136, 120, 178, 46, 8, 69, 116, 47, 136, 90, 146, 48, 136, 97, 93, 49, 136, 60, 114, 50, 136, 67, 61, 51, 136, 30, 82, 52, 136, 37, 29, 53, 136, 0, 50, 54, 136, 7, 253, 54, 8, 29, 27, 56, 136, 233, 220, 56, 8, 255, 250, 57, 136, 203, 188, 58, 8, 225, 218, 59, 8, 232, 165, 60, 8, 195, 186, 61, 8, 202, 133, 62, 8, 165, 154, 63, 8, 172, 101, 64, 136, 193, 131, 65, 8, 142, 69, 66, 136, 163, 99, 67, 136, 170, 46, 68, 136, 133, 67, 69, 8, 82, 5, 70, 136, 103, 35, 71, 8, 169, 247, 71, 8, 154, 231, 72, 4, 0, 0, 0, 255, 255, 255, 255, 20, 139, 22, 115, 255, 255, 255, 255, 112, 3, 18, 123, 170, 170, 170, 170, 85, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 170, 170, 170, 8, 0, 0, 0, 236, 129, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 152, 133, 0, 0, 0, 0, 0, 0, 152, 133, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 152, 133, 0, 112, 40, 0, 0, 96, 217, 7, 0, 112, 102, 108, 0, 224, 105, 108, 0, 224, 189, 108, 0, 224, 195, 108, 0, 16, 218, 108, 0, 224, 2, 0, 0, 0, 95, 1, 0, 0, 109, 1, 0, 0, 16, 0, 0, 0, 156, 166, 78, 156, 240, 32, 188, 156, 0, 179, 84, 203, 112, 87, 199, 203, 128, 86, 183, 204, 112, 57, 167, 205, 0, 115, 160, 206, 112, 27, 135, 207, 128, 57, 112, 3, 0, 28, 13, 4, 0, 205, 73, 37, 0, 234, 239, 37, 0, 175, 41, 39, 0, 204, 207, 39, 0, 145, 9, 41, 0, 174, 175, 41, 2, 0, 0, 0, 255, 255, 255, 255, 8, 159, 237, 114, 17, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 120, 143, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 240, 108, 0, 224, 243, 108, 0, 224, 4, 109, 0, 224, 7, 109, 0, 16, 13, 109, 0, 224, 2, 0, 0, 0, 96, 1, 0, 0, 115, 1, 0, 0, 83, 0, 0, 0, 164, 173, 78, 156, 248, 39, 188, 156, 8, 186, 84, 203, 120, 94, 199, 203, 136, 93, 183, 204, 120, 64, 167, 205, 8, 122, 160, 206, 120, 34, 135, 207, 136, 64, 112, 3, 8, 35, 13, 4, 136, 34, 80, 5, 136, 63, 246, 5, 136, 4, 48, 7, 136, 33, 214, 7, 136, 230, 15, 9, 136, 3, 182, 9, 136, 200, 239, 10, 8, 32, 159, 11, 8, 229, 216, 12, 8, 2, 127, 13, 8, 199, 184, 14, 8, 228, 94, 15, 8, 169, 152, 16, 8, 198, 62, 17, 8, 139, 120, 18, 8, 168, 30, 19, 8, 109, 88, 20, 8, 138, 254, 20, 8, 79, 56, 22, 136, 144, 12, 23, 136, 107, 33, 24, 136, 136, 199, 24, 136, 77, 1, 26, 136, 106, 167, 26, 136, 47, 225, 27, 136, 76, 135, 28, 136, 17, 193, 29, 136, 163, 121, 30, 8, 185, 151, 31, 136, 133, 89, 32, 136, 213, 128, 33, 8, 162, 66, 34, 8, 242, 105, 35, 8, 132, 34, 36, 8, 212, 73, 37, 8, 241, 239, 37, 8, 182, 41, 39, 8, 211, 207, 39, 8, 152, 9, 41, 8, 181, 175, 41, 8, 122, 233, 42, 136, 209, 152, 43, 136, 150, 210, 44, 136, 179, 120, 45, 136, 120, 178, 46, 136, 149, 88, 47, 136, 90, 146, 48, 136, 97, 93, 49, 136, 60, 114, 50, 136, 67, 61, 51, 136, 30, 82, 52, 136, 37, 29, 53, 136, 0, 50, 54, 136, 7, 253, 54, 8, 29, 27, 56, 136, 233, 220, 56, 8, 255, 250, 57, 136, 203, 188, 58, 8, 225, 218, 59, 8, 232, 165, 60, 8, 195, 186, 61, 8, 202, 133, 62, 8, 165, 154, 63, 8, 172, 101, 64, 136, 193, 131, 65, 8, 142, 69, 66, 136, 163, 99, 67, 136, 170, 46, 68, 136, 133, 67, 69, 8, 82, 5, 70, 136, 103, 35, 71, 8, 169, 247, 71, 8, 154, 231, 72, 6, 0, 0, 0, 255, 255, 255, 255, 100, 136, 22, 115, 255, 255, 255, 255, 224, 165, 4, 118, 255, 255, 255, 255, 112, 3, 18, 123, 170, 170, 170, 170, 86, 0, 0, 0, 4, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 170, 170, 10, 0, 0, 0, 156, 132, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 152, 133, 0, 0, 0, 0, 0, 0, 152, 133, 0, 0, 16, 14, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 152, 133, 0, 112, 40, 0, 0, 96, 217, 7, 0, 112, 28, 109, 0, 224, 31, 109, 0, 224, 115, 109, 0, 224, 123, 109, 0, 16, 146, 109, 0, 224, 83, 0, 0, 0, 128, 120, 213, 155, 240, 32, 188, 156, 0, 179, 84, 203, 112, 87, 199, 203, 128, 86, 183, 204, 112, 57, 167, 205, 0, 115, 160, 206, 112, 27, 135, 207, 128, 57, 112, 3, 0, 28, 13, 4, 128, 27, 80, 5, 128, 56, 246, 5, 128, 253, 47, 7, 128, 26, 214, 7, 128, 223, 15, 9, 128, 252, 181, 9, 128, 193, 239, 10, 0, 25, 159, 11, 0, 222, 216, 12, 0, 251, 126, 13, 0, 192, 184, 14, 0, 221, 94, 15, 0, 162, 152, 16, 0, 191, 62, 17, 0, 132, 120, 18, 0, 161, 30, 19, 0, 102, 88, 20, 0, 131, 254, 20, 0, 72, 56, 22, 0, 79, 3, 23, 128, 100, 33, 24, 0, 49, 227, 24, 128, 70, 1, 26, 128, 99, 167, 26, 128, 40, 225, 27, 128, 69, 135, 28, 128, 10, 193, 29, 128, 39, 103, 30, 0, 178, 151, 31, 128, 126, 89, 32, 128, 206, 128, 33, 0, 155, 66, 34, 0, 235, 105, 35, 0, 125, 34, 36, 0, 205, 73, 37, 0, 95, 2, 38, 0, 175, 41, 39, 0, 182, 244, 39, 128, 225, 237, 40, 0, 152, 212, 41, 128, 195, 205, 42, 0, 122, 180, 43, 128, 165, 173, 44, 0, 92, 148, 45, 128, 135, 141, 46, 0, 62, 116, 47, 128, 105, 109, 48, 128, 90, 93, 49, 0, 134, 86, 50, 128, 60, 61, 51, 0, 104, 54, 52, 128, 30, 29, 53, 0, 74, 22, 54, 128, 0, 253, 54, 0, 44, 246, 55, 128, 226, 220, 56, 128, 233, 167, 57, 128, 196, 188, 58, 128, 42, 191, 59, 0, 225, 165, 60, 128, 12, 159, 61, 0, 195, 133, 62, 128, 238, 126, 63, 0, 165, 101, 64, 128, 208, 94, 65, 0, 135, 69, 66, 128, 178, 62, 67, 128, 163, 46, 68, 128, 148, 30, 69, 0, 75, 5, 70, 0, 177, 7, 71, 0, 162, 247, 71, 0, 147, 231, 72, 2, 0, 0, 0, 255, 255, 255, 255, 32, 4, 46, 116, 170, 170, 170, 170, 170, 170, 170, 170, 84, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 6, 0, 0, 0, 224, 134, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 160, 140, 0, 112, 238, 3, 0, 96, 217, 7, 0, 112, 170, 109, 0, 224, 254, 109, 0, 224, 3, 110, 0, 16, 25, 110, 0, 224, 3, 0, 0, 0, 0, 0, 0, 0, 99, 1, 0, 0, 107, 1, 0, 0, 8, 0, 0, 0, 164, 173, 78, 156, 248, 39, 188, 156, 8, 186, 84, 203, 120, 94, 199, 203, 136, 93, 183, 204, 120, 64, 167, 205, 8, 122, 160, 206, 120, 34, 135, 207, 4, 0, 0, 0, 255, 255, 255, 255, 88, 146, 22, 115, 255, 255, 255, 255, 112, 3, 18, 123, 170, 170, 170, 170, 170, 170, 170, 170, 10, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 168, 122, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 152, 133, 0, 0, 0, 0, 0, 0, 152, 133, 0, 0, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 43, 110, 0, 224, 47, 110, 0, 224, 56, 110, 0, 224, 63, 110, 0, 16, 67, 110, 0, 224, 18, 0, 0, 0, 48, 184, 78, 156, 132, 50, 188, 156, 148, 196, 84, 203, 4, 105, 199, 203, 20, 104, 183, 204, 4, 75, 167, 205, 20, 241, 15, 9, 20, 14, 182, 9, 20, 88, 1, 26, 20, 117, 167, 26, 20, 82, 37, 41, 148, 191, 175, 41, 148, 180, 113, 69, 148, 92, 5, 70, 20, 114, 35, 71, 20, 121, 238, 71, 20, 84, 3, 73, 20, 91, 206, 73, 2, 0, 0, 0, 255, 255, 255, 255, 176, 10, 166, 116, 170, 170, 170, 170, 19, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 208, 120, 0, 0, 0, 0, 0, 0, 12, 123, 0, 0, 0, 0, 0, 0, 12, 123, 0, 0, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 84, 110, 0, 224, 103, 110, 0, 224, 107, 110, 0, 16, 113, 110, 0, 224, 2, 0, 0, 0, 101, 1, 0, 0, 112, 1, 0, 0, 91, 0, 0, 0, 128, 120, 213, 155, 240, 32, 188, 156, 0, 179, 84, 203, 112, 87, 199, 203, 128, 86, 183, 204, 112, 57, 167, 205, 0, 115, 160, 206, 112, 27, 135, 207, 0, 141, 194, 251, 0, 126, 178, 252, 0, 89, 199, 253, 128, 176, 118, 254, 0, 59, 167, 255, 128, 146, 86, 0, 0, 29, 135, 1, 0, 175, 63, 2, 128, 57, 112, 3, 0, 28, 13, 4, 128, 27, 80, 5, 128, 56, 246, 5, 128, 253, 47, 7, 128, 26, 214, 7, 128, 223, 15, 9, 128, 252, 181, 9, 128, 193, 239, 10, 0, 25, 159, 11, 0, 222, 216, 12, 0, 251, 126, 13, 0, 192, 184, 14, 0, 221, 94, 15, 0, 162, 152, 16, 0, 191, 62, 17, 0, 132, 120, 18, 0, 161, 30, 19, 0, 102, 88, 20, 0, 131, 254, 20, 0, 72, 56, 22, 0, 79, 3, 23, 128, 100, 33, 24, 0, 49, 227, 24, 128, 70, 1, 26, 128, 99, 167, 26, 128, 40, 225, 27, 128, 69, 135, 28, 128, 10, 193, 29, 128, 39, 103, 30, 0, 178, 151, 31, 128, 126, 89, 32, 128, 206, 128, 33, 0, 155, 66, 34, 0, 235, 105, 35, 0, 125, 34, 36, 0, 205, 73, 37, 0, 95, 2, 38, 0, 175, 41, 39, 0, 182, 244, 39, 128, 225, 237, 40, 0, 152, 212, 41, 128, 195, 205, 42, 0, 122, 180, 43, 128, 165, 173, 44, 0, 92, 148, 45, 128, 135, 141, 46, 0, 62, 116, 47, 128, 105, 109, 48, 128, 90, 93, 49, 0, 134, 86, 50, 128, 60, 61, 51, 0, 104, 54, 52, 128, 30, 29, 53, 0, 74, 22, 54, 128, 0, 253, 54, 0, 44, 246, 55, 128, 226, 220, 56, 128, 233, 167, 57, 128, 196, 188, 58, 128, 42, 191, 59, 0, 225, 165, 60, 128, 12, 159, 61, 0, 195, 133, 62, 128, 238, 126, 63, 0, 165, 101, 64, 128, 208, 94, 65, 0, 135, 69, 66, 128, 178, 62, 67, 128, 163, 46, 68, 128, 148, 30, 69, 0, 75, 5, 70, 0, 177, 7, 71, 0, 162, 247, 71, 0, 147, 231, 72, 2, 0, 0, 0, 255, 255, 255, 255, 228, 0, 46, 116, 170, 170, 170, 170, 170, 170, 170, 170, 92, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 6, 0, 0, 0, 28, 138, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 160, 140, 0, 112, 238, 3, 0, 96, 217, 7, 0, 112, 127, 110, 0, 224, 130, 110, 0, 224, 222, 110, 0, 224, 227, 110, 0, 16, 251, 110, 0, 224, 20, 0, 0, 0, 156, 166, 78, 156, 240, 32, 188, 156, 0, 179, 84, 203, 112, 87, 199, 203, 128, 86, 183, 204, 112, 57, 167, 205, 0, 115, 160, 206, 112, 27, 135, 207, 128, 57, 112, 3, 0, 28, 13, 4, 0, 205, 73, 37, 0, 234, 239, 37, 0, 175, 41, 39, 0, 204, 207, 39, 0, 145, 9, 41, 0, 174, 175, 41, 0, 115, 233, 42, 128, 202, 152, 43, 128, 143, 210, 44, 128, 172, 120, 45, 2, 0, 0, 0, 255, 255, 255, 255, 212, 162, 237, 114, 21, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 172, 139, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 15, 111, 0, 224, 36, 111, 0, 224, 39, 111, 0, 16, 46, 111, 0, 224, 2, 0, 0, 0, 102, 1, 0, 0, 104, 1, 0, 0, 56, 0, 0, 0, 224, 102, 254, 20, 248, 64, 56, 22, 104, 138, 231, 22, 120, 93, 33, 24, 104, 108, 199, 24, 120, 63, 1, 26, 104, 78, 167, 26, 120, 33, 225, 27, 104, 48, 135, 28, 120, 3, 193, 29, 112, 142, 121, 30, 248, 170, 151, 31, 112, 112, 89, 32, 120, 199, 128, 33, 240, 140, 66, 34, 248, 227, 105, 35, 240, 110, 34, 36, 248, 197, 73, 37, 240, 219, 239, 37, 248, 167, 41, 39, 240, 189, 207, 39, 248, 137, 9, 41, 240, 159, 175, 41, 248, 107, 233, 42, 112, 188, 152, 43, 120, 136, 210, 44, 112, 158, 120, 45, 120, 106, 178, 46, 112, 128, 88, 47, 120, 76, 146, 48, 112, 76, 93, 49, 120, 46, 114, 50, 112, 46, 61, 51, 120, 16, 82, 52, 112, 16, 29, 53, 120, 242, 49, 54, 112, 242, 252, 54, 248, 14, 27, 56, 112, 212, 220, 56, 120, 226, 167, 57, 112, 182, 188, 58, 248, 210, 218, 59, 240, 210, 165, 60, 248, 180, 186, 61, 240, 180, 133, 62, 248, 150, 154, 63, 240, 150, 101, 64, 120, 179, 131, 65, 240, 120, 69, 66, 120, 149, 99, 67, 112, 149, 46, 68, 120, 119, 67, 69, 240, 60, 5, 70, 120, 89, 35, 71, 240, 147, 247, 71, 248, 139, 231, 72, 2, 0, 0, 0, 255, 255, 255, 255, 220, 119, 22, 115, 57, 0, 0, 0, 1, 2, 4, 2, 4, 2, 4, 2, 4, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 170, 170, 170, 10, 0, 0, 0, 36, 149, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 168, 147, 0, 0, 0, 0, 0, 0, 168, 147, 0, 0, 8, 7, 0, 0, 168, 147, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 168, 147, 0, 112, 210, 1, 0, 96, 217, 7, 0, 112, 60, 111, 0, 224, 63, 111, 0, 224, 120, 111, 0, 224, 123, 111, 0, 16, 139, 111, 0, 224, 2, 0, 0, 0, 105, 1, 0, 0, 113, 1, 0, 0, 83, 0, 0, 0, 156, 166, 78, 156, 240, 32, 188, 156, 0, 179, 84, 203, 112, 87, 199, 203, 128, 86, 183, 204, 112, 57, 167, 205, 0, 115, 160, 206, 112, 27, 135, 207, 128, 57, 112, 3, 0, 28, 13, 4, 128, 27, 80, 5, 128, 56, 246, 5, 128, 253, 47, 7, 128, 26, 214, 7, 128, 223, 15, 9, 128, 252, 181, 9, 128, 193, 239, 10, 0, 25, 159, 11, 0, 222, 216, 12, 0, 251, 126, 13, 0, 192, 184, 14, 0, 221, 94, 15, 0, 162, 152, 16, 0, 191, 62, 17, 0, 132, 120, 18, 0, 161, 30, 19, 0, 102, 88, 20, 0, 131, 254, 20, 0, 72, 56, 22, 128, 159, 231, 22, 128, 100, 33, 24, 128, 129, 199, 24, 128, 70, 1, 26, 128, 99, 167, 26, 128, 40, 225, 27, 128, 69, 135, 28, 128, 10, 193, 29, 128, 156, 121, 30, 0, 178, 151, 31, 128, 126, 89, 32, 0, 148, 119, 33, 0, 155, 66, 34, 0, 235, 105, 35, 0, 125, 34, 36, 0, 205, 73, 37, 0, 95, 2, 38, 0, 175, 41, 39, 0, 204, 207, 39, 0, 145, 9, 41, 0, 174, 175, 41, 0, 115, 233, 42, 128, 202, 152, 43, 128, 143, 210, 44, 128, 172, 120, 45, 128, 113, 178, 46, 0, 62, 116, 47, 128, 83, 146, 48, 128, 90, 93, 49, 128, 53, 114, 50, 128, 60, 61, 51, 128, 23, 82, 52, 128, 30, 29, 53, 128, 249, 49, 54, 128, 0, 253, 54, 0, 22, 27, 56, 128, 226, 220, 56, 128, 233, 167, 57, 128, 196, 188, 58, 0, 218, 218, 59, 0, 225, 165, 60, 0, 188, 186, 61, 0, 195, 133, 62, 0, 158, 154, 63, 0, 165, 101, 64, 128, 186, 131, 65, 0, 135, 69, 66, 128, 156, 99, 67, 128, 163, 46, 68, 128, 126, 67, 69, 0, 75, 5, 70, 128, 96, 35, 71, 0, 162, 247, 71, 0, 147, 231, 72, 2, 0, 0, 0, 255, 255, 255, 255, 24, 133, 22, 115, 170, 170, 170, 170, 170, 170, 170, 170, 84, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 6, 0, 0, 0, 232, 135, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 160, 140, 0, 112, 232, 0, 0, 96, 217, 7, 0, 112, 163, 111, 0, 224, 166, 111, 0, 224, 250, 111, 0, 224, 255, 111, 0, 16, 21, 112, 0, 224, 2, 0, 0, 0, 108, 1, 0, 0, 114, 1, 0, 0, 18, 0, 0, 0, 188, 194, 78, 156, 16, 61, 188, 156, 32, 207, 84, 203, 144, 115, 199, 203, 160, 114, 183, 204, 144, 85, 167, 205, 160, 251, 15, 9, 160, 24, 182, 9, 160, 98, 1, 26, 160, 127, 167, 26, 160, 92, 37, 41, 32, 202, 175, 41, 32, 191, 113, 69, 32, 103, 5, 70, 160, 124, 35, 71, 160, 131, 238, 71, 160, 94, 3, 73, 160, 101, 206, 73, 2, 0, 0, 0, 255, 255, 255, 255, 228, 22, 166, 116, 170, 170, 170, 170, 19, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 156, 108, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 128, 112, 0, 0, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 41, 112, 0, 224, 44, 112, 0, 224, 63, 112, 0, 224, 67, 112, 0, 16, 73, 112, 0, 224, 5, 0, 0, 0, 1, 0, 0, 0, 93, 1, 0, 0, 97, 1, 0, 0, 106, 1, 0, 0, 111, 1, 0, 0, 83, 0, 0, 0, 156, 166, 78, 156, 240, 32, 188, 156, 0, 179, 84, 203, 112, 87, 199, 203, 128, 86, 183, 204, 112, 57, 167, 205, 0, 115, 160, 206, 112, 27, 135, 207, 128, 57, 112, 3, 0, 28, 13, 4, 128, 27, 80, 5, 128, 56, 246, 5, 128, 253, 47, 7, 128, 26, 214, 7, 128, 223, 15, 9, 128, 252, 181, 9, 128, 193, 239, 10, 0, 25, 159, 11, 0, 222, 216, 12, 0, 251, 126, 13, 0, 192, 184, 14, 0, 221, 94, 15, 0, 162, 152, 16, 0, 191, 62, 17, 0, 132, 120, 18, 0, 161, 30, 19, 0, 102, 88, 20, 0, 131, 254, 20, 0, 72, 56, 22, 128, 137, 12, 23, 128, 100, 33, 24, 128, 129, 199, 24, 128, 70, 1, 26, 128, 99, 167, 26, 128, 40, 225, 27, 128, 69, 135, 28, 128, 10, 193, 29, 128, 156, 121, 30, 0, 178, 151, 31, 128, 126, 89, 32, 128, 206, 128, 33, 0, 155, 66, 34, 0, 235, 105, 35, 0, 125, 34, 36, 0, 205, 73, 37, 0, 234, 239, 37, 0, 175, 41, 39, 0, 204, 207, 39, 0, 145, 9, 41, 0, 174, 175, 41, 0, 115, 233, 42, 128, 202, 152, 43, 128, 143, 210, 44, 128, 172, 120, 45, 128, 113, 178, 46, 128, 142, 88, 47, 128, 83, 146, 48, 128, 90, 93, 49, 128, 53, 114, 50, 128, 60, 61, 51, 128, 23, 82, 52, 128, 30, 29, 53, 128, 249, 49, 54, 128, 0, 253, 54, 0, 22, 27, 56, 128, 226, 220, 56, 128, 233, 167, 57, 128, 196, 188, 58, 0, 218, 218, 59, 0, 225, 165, 60, 0, 188, 186, 61, 0, 195, 133, 62, 0, 158, 154, 63, 0, 165, 101, 64, 128, 186, 131, 65, 0, 135, 69, 66, 128, 156, 99, 67, 128, 163, 46, 68, 128, 126, 67, 69, 0, 75, 5, 70, 128, 96, 35, 71, 0, 162, 247, 71, 0, 147, 231, 72, 2, 0, 0, 0, 255, 255, 255, 255, 60, 127, 22, 115, 170, 170, 170, 170, 170, 170, 170, 170, 84, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 6, 0, 0, 0, 196, 141, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 160, 140, 0, 112, 106, 0, 0, 96, 217, 7, 0, 112, 88, 112, 0, 224, 94, 112, 0, 224, 178, 112, 0, 224, 183, 112, 0, 16, 205, 112, 0, 224, 54, 0, 0, 0, 96, 23, 12, 155, 240, 218, 213, 155, 144, 174, 217, 156, 144, 181, 164, 157, 144, 144, 185, 158, 144, 151, 132, 159, 144, 113, 9, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 16, 22, 114, 209, 144, 64, 78, 210, 144, 99, 164, 13, 16, 26, 139, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 54, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 170, 170, 4, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 16, 14, 0, 112, 121, 3, 0, 96, 205, 7, 0, 112, 225, 112, 0, 224, 27, 113, 0, 16, 42, 113, 0, 224, 88, 0, 0, 0, 128, 44, 166, 158, 112, 249, 186, 159, 128, 14, 134, 160, 112, 219, 154, 161, 128, 254, 136, 203, 240, 9, 97, 210, 0, 103, 248, 250, 240, 73, 232, 251, 0, 73, 216, 252, 240, 43, 200, 253, 0, 43, 184, 254, 240, 13, 168, 255, 0, 13, 152, 0, 240, 239, 135, 1, 0, 239, 119, 2, 112, 12, 113, 3, 128, 11, 97, 4, 112, 238, 80, 5, 128, 237, 64, 6, 112, 208, 48, 7, 128, 39, 141, 7, 112, 178, 16, 9, 0, 163, 173, 9, 112, 148, 240, 10, 128, 147, 224, 11, 240, 176, 217, 12, 128, 117, 192, 13, 240, 146, 185, 14, 0, 146, 169, 15, 240, 116, 153, 16, 0, 116, 137, 17, 240, 86, 121, 18, 0, 86, 105, 19, 240, 56, 89, 20, 0, 56, 73, 21, 240, 26, 57, 22, 0, 26, 41, 23, 112, 55, 34, 24, 0, 252, 8, 25, 112, 25, 2, 26, 128, 24, 242, 26, 112, 251, 225, 27, 128, 250, 209, 28, 112, 221, 193, 29, 128, 220, 177, 30, 112, 191, 161, 31, 0, 15, 118, 32, 112, 161, 129, 33, 0, 241, 85, 34, 240, 189, 106, 35, 0, 211, 53, 36, 240, 159, 74, 37, 0, 181, 21, 38, 240, 129, 42, 39, 128, 209, 254, 39, 240, 99, 10, 41, 128, 179, 222, 41, 240, 69, 234, 42, 128, 149, 190, 43, 112, 98, 211, 44, 128, 119, 158, 45, 112, 68, 179, 46, 128, 89, 126, 47, 112, 38, 147, 48, 0, 118, 103, 49, 112, 8, 115, 50, 0, 88, 71, 51, 112, 234, 82, 52, 0, 58, 39, 53, 112, 204, 50, 54, 0, 28, 7, 55, 240, 232, 27, 56, 0, 254, 230, 56, 240, 202, 251, 57, 0, 224, 198, 58, 240, 172, 219, 59, 128, 252, 175, 60, 240, 142, 187, 61, 128, 222, 143, 62, 240, 112, 155, 63, 128, 192, 111, 64, 112, 141, 132, 65, 128, 162, 79, 66, 112, 111, 100, 67, 128, 132, 47, 68, 112, 81, 68, 69, 0, 183, 243, 69, 240, 109, 45, 71, 170, 170, 170, 170, 88, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 4, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 160, 171, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 57, 113, 0, 224, 147, 113, 0, 16, 170, 113, 0, 224, 40, 0, 0, 0, 144, 99, 164, 13, 16, 26, 139, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 40, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 4, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 185, 113, 0, 224, 227, 113, 0, 16, 238, 113, 0, 224, 2, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 253, 113, 0, 224, 88, 0, 0, 0, 112, 30, 166, 158, 96, 235, 186, 159, 112, 0, 134, 160, 96, 205, 154, 161, 112, 240, 136, 203, 224, 251, 96, 210, 240, 88, 248, 250, 224, 59, 232, 251, 240, 58, 216, 252, 224, 29, 200, 253, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 240, 224, 119, 2, 96, 254, 112, 3, 112, 253, 96, 4, 96, 224, 80, 5, 112, 223, 64, 6, 96, 194, 48, 7, 112, 25, 141, 7, 96, 164, 16, 9, 240, 148, 173, 9, 96, 134, 240, 10, 112, 133, 224, 11, 224, 162, 217, 12, 112, 103, 192, 13, 224, 132, 185, 14, 240, 131, 169, 15, 224, 102, 153, 16, 240, 101, 137, 17, 224, 72, 121, 18, 240, 71, 105, 19, 224, 42, 89, 20, 240, 41, 73, 21, 224, 12, 57, 22, 240, 11, 41, 23, 96, 41, 34, 24, 240, 237, 8, 25, 96, 11, 2, 26, 112, 10, 242, 26, 96, 237, 225, 27, 112, 236, 209, 28, 96, 207, 193, 29, 112, 206, 177, 30, 96, 177, 161, 31, 240, 0, 118, 32, 96, 147, 129, 33, 240, 226, 85, 34, 224, 175, 106, 35, 240, 196, 53, 36, 224, 145, 74, 37, 240, 166, 21, 38, 224, 115, 42, 39, 112, 195, 254, 39, 224, 85, 10, 41, 112, 165, 222, 41, 224, 55, 234, 42, 112, 135, 190, 43, 96, 84, 211, 44, 112, 105, 158, 45, 96, 54, 179, 46, 112, 75, 126, 47, 96, 24, 147, 48, 240, 103, 103, 49, 96, 250, 114, 50, 240, 73, 71, 51, 96, 220, 82, 52, 240, 43, 39, 53, 96, 190, 50, 54, 240, 13, 7, 55, 224, 218, 27, 56, 240, 239, 230, 56, 224, 188, 251, 57, 240, 209, 198, 58, 224, 158, 219, 59, 112, 238, 175, 60, 224, 128, 187, 61, 112, 208, 143, 62, 224, 98, 155, 63, 112, 178, 111, 64, 96, 127, 132, 65, 112, 148, 79, 66, 96, 97, 100, 67, 112, 118, 47, 68, 96, 67, 68, 69, 240, 168, 243, 69, 224, 95, 45, 71, 88, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 4, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 176, 185, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 2, 114, 0, 224, 91, 114, 0, 16, 114, 114, 0, 224, 10, 0, 0, 0, 147, 1, 0, 0, 148, 1, 0, 0, 161, 1, 0, 0, 176, 1, 0, 0, 177, 1, 0, 0, 249, 1, 0, 0, 250, 1, 0, 0, 251, 1, 0, 0, 252, 1, 0, 0, 253, 1, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 74, 0, 54, 0, 170, 170, 129, 114, 0, 224, 140, 114, 0, 224, 2, 0, 0, 0, 240, 241, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 147, 114, 0, 224, 2, 0, 0, 0, 96, 115, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 152, 114, 0, 224, 2, 0, 0, 0, 80, 101, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 157, 114, 0, 224, 2, 0, 0, 0, 64, 87, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 162, 114, 0, 224, 2, 0, 0, 0, 224, 227, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 167, 114, 0, 224, 2, 0, 0, 0, 208, 213, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 172, 114, 0, 224, 2, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 177, 114, 0, 224, 2, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 182, 114, 0, 224, 2, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 187, 114, 0, 224, 2, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 192, 114, 0, 224, 2, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 197, 114, 0, 224, 2, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 202, 114, 0, 224, 2, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 207, 114, 0, 224, 2, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 212, 114, 0, 224, 2, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 217, 114, 0, 224, 2, 0, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 222, 114, 0, 224, 2, 0, 0, 0, 208, 182, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 227, 114, 0, 224, 2, 0, 0, 0, 224, 196, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 232, 114, 0, 224, 2, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 237, 114, 0, 224, 2, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 242, 114, 0, 224, 2, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 247, 114, 0, 224, 2, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 252, 114, 0, 224, 2, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 1, 115, 0, 224, 2, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 6, 115, 0, 224, 2, 0, 0, 0, 128, 112, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 11, 115, 0, 224, 2, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 16, 115, 0, 224, 8, 0, 0, 0, 178, 1, 0, 0, 179, 1, 0, 0, 180, 1, 0, 0, 182, 1, 0, 0, 102, 2, 0, 0, 116, 2, 0, 0, 117, 2, 0, 0, 121, 2, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 74, 0, 54, 0, 170, 170, 21, 115, 0, 224, 30, 115, 0, 224, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 37, 115, 0, 224, 97, 0, 0, 0, 236, 46, 12, 155, 92, 214, 213, 155, 12, 184, 217, 156, 12, 191, 164, 157, 12, 37, 167, 158, 12, 22, 151, 159, 140, 65, 144, 160, 12, 248, 118, 161, 140, 35, 112, 162, 12, 218, 86, 163, 140, 5, 80, 164, 12, 188, 54, 165, 140, 91, 37, 166, 140, 193, 39, 167, 140, 227, 94, 168, 140, 163, 7, 169, 12, 90, 238, 169, 140, 133, 231, 170, 12, 226, 39, 172, 140, 103, 199, 172, 12, 102, 237, 173, 140, 73, 167, 174, 140, 153, 206, 175, 140, 43, 135, 176, 140, 30, 177, 177, 12, 72, 112, 178, 12, 82, 146, 179, 12, 42, 80, 180, 140, 133, 115, 181, 12, 12, 48, 182, 12, 185, 84, 183, 12, 238, 15, 184, 140, 120, 64, 185, 12, 208, 239, 185, 140, 113, 24, 187, 140, 236, 216, 187, 12, 165, 249, 188, 140, 206, 184, 189, 140, 216, 218, 190, 140, 176, 152, 191, 140, 93, 189, 192, 140, 146, 120, 193, 140, 203, 167, 194, 92, 93, 220, 194, 112, 116, 88, 195, 112, 196, 127, 196, 112, 86, 56, 197, 240, 247, 96, 198, 240, 114, 33, 199, 80, 178, 68, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 16, 22, 114, 209, 144, 64, 78, 210, 144, 99, 164, 13, 16, 26, 139, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 97, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 3, 2, 3, 2, 3, 2, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 170, 170, 170, 12, 0, 0, 0, 148, 4, 0, 0, 0, 0, 0, 0, 148, 4, 0, 0, 16, 14, 0, 0, 176, 4, 0, 0, 0, 0, 0, 0, 176, 4, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 42, 115, 0, 224, 143, 115, 0, 16, 169, 115, 0, 224, 25, 0, 0, 0, 0, 219, 65, 212, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 148, 179, 54, 126, 170, 170, 170, 170, 170, 170, 170, 170, 26, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 108, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 192, 115, 0, 224, 218, 115, 0, 224, 223, 115, 0, 16, 231, 115, 0, 224, 64, 0, 0, 0, 116, 69, 24, 170, 80, 11, 164, 181, 192, 153, 39, 21, 48, 206, 24, 22, 64, 205, 8, 23, 176, 1, 250, 23, 192, 0, 234, 24, 48, 53, 219, 25, 192, 133, 204, 26, 224, 146, 188, 27, 224, 131, 172, 28, 224, 116, 156, 29, 224, 101, 140, 30, 224, 86, 124, 31, 224, 71, 108, 32, 224, 56, 92, 33, 224, 41, 76, 34, 224, 26, 60, 35, 224, 11, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 112, 39, 5, 39, 112, 24, 245, 39, 96, 236, 212, 41, 112, 235, 196, 42, 112, 220, 180, 43, 112, 205, 164, 44, 112, 190, 148, 45, 112, 175, 132, 46, 112, 160, 116, 47, 112, 145, 100, 48, 240, 188, 93, 49, 240, 151, 114, 50, 240, 158, 61, 51, 240, 121, 82, 52, 240, 128, 29, 53, 240, 91, 50, 54, 240, 98, 253, 54, 112, 120, 27, 56, 240, 68, 221, 56, 112, 90, 251, 57, 240, 38, 189, 58, 112, 60, 219, 59, 112, 67, 166, 60, 112, 30, 187, 61, 112, 37, 134, 62, 112, 0, 155, 63, 112, 7, 102, 64, 240, 28, 132, 65, 112, 233, 69, 66, 240, 254, 99, 67, 112, 203, 37, 68, 240, 224, 67, 69, 112, 173, 5, 70, 240, 194, 35, 71, 240, 201, 238, 71, 240, 164, 3, 73, 240, 171, 206, 73, 240, 134, 227, 74, 240, 141, 174, 75, 112, 163, 204, 76, 240, 111, 142, 77, 96, 29, 76, 84, 112, 20, 247, 86, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 64, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 2, 1, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 3, 10, 0, 0, 0, 12, 45, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 251, 115, 0, 224, 63, 116, 0, 16, 80, 116, 0, 224, 55, 0, 0, 0, 128, 33, 128, 155, 224, 233, 124, 185, 208, 175, 198, 185, 224, 99, 242, 201, 80, 168, 16, 202, 16, 75, 231, 204, 240, 76, 170, 205, 224, 24, 162, 206, 112, 105, 147, 207, 96, 158, 19, 223, 80, 10, 183, 223, 96, 94, 236, 9, 96, 244, 24, 11, 0, 174, 205, 11, 0, 159, 189, 12, 128, 85, 164, 13, 128, 93, 140, 14, 128, 55, 132, 15, 16, 252, 106, 16, 240, 123, 100, 17, 240, 170, 82, 18, 96, 130, 70, 19, 80, 194, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 55, 0, 0, 0, 3, 4, 3, 4, 2, 1, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 10, 0, 0, 0, 60, 22, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 96, 116, 0, 224, 155, 116, 0, 16, 170, 116, 0, 224, 6, 0, 0, 0, 188, 1, 0, 0, 208, 1, 0, 0, 220, 1, 0, 0, 226, 1, 0, 0, 229, 1, 0, 0, 243, 1, 0, 0, 36, 0, 0, 0, 224, 53, 2, 202, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 16, 140, 161, 209, 144, 64, 78, 210, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 72, 240, 60, 94, 170, 170, 170, 170, 37, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 56, 19, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 191, 116, 0, 224, 198, 116, 0, 224, 235, 116, 0, 224, 239, 116, 0, 16, 250, 116, 0, 224, 60, 0, 0, 0, 96, 23, 12, 155, 240, 218, 213, 155, 144, 174, 217, 156, 144, 181, 164, 157, 144, 144, 185, 158, 144, 151, 132, 159, 144, 113, 9, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 16, 22, 114, 209, 0, 150, 182, 209, 128, 190, 88, 210, 16, 79, 161, 210, 144, 27, 99, 211, 144, 35, 75, 212, 32, 209, 57, 213, 144, 231, 103, 213, 0, 115, 168, 213, 16, 180, 41, 214, 16, 26, 44, 215, 16, 150, 9, 216, 144, 193, 2, 217, 16, 120, 233, 217, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 248, 97, 162, 111, 170, 170, 170, 170, 61, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 1, 2, 1, 2, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 8, 0, 0, 0, 136, 12, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 14, 117, 0, 224, 75, 117, 0, 224, 79, 117, 0, 16, 96, 117, 0, 224, 101, 0, 0, 0, 128, 73, 68, 152, 112, 37, 12, 155, 240, 218, 213, 155, 144, 174, 217, 156, 144, 181, 164, 157, 144, 144, 185, 158, 144, 151, 132, 159, 48, 248, 206, 159, 240, 165, 96, 160, 112, 187, 126, 161, 240, 18, 46, 162, 240, 76, 122, 163, 240, 129, 53, 164, 112, 35, 94, 165, 240, 53, 37, 166, 240, 155, 39, 167, 240, 1, 42, 168, 240, 125, 7, 169, 112, 52, 238, 169, 240, 95, 231, 170, 240, 80, 215, 171, 240, 65, 199, 172, 240, 167, 201, 173, 240, 35, 167, 174, 112, 79, 160, 175, 240, 5, 135, 176, 240, 107, 137, 177, 160, 76, 112, 178, 160, 178, 114, 179, 160, 46, 80, 180, 32, 90, 73, 181, 160, 16, 48, 182, 160, 118, 50, 183, 160, 242, 15, 184, 160, 227, 255, 184, 160, 212, 239, 185, 32, 139, 214, 186, 32, 241, 216, 187, 32, 226, 200, 188, 32, 211, 184, 189, 160, 137, 159, 190, 32, 181, 152, 191, 32, 27, 155, 192, 32, 151, 120, 193, 32, 136, 104, 194, 32, 121, 88, 195, 160, 47, 63, 196, 32, 91, 56, 197, 32, 193, 58, 198, 160, 214, 88, 199, 160, 9, 218, 199, 32, 25, 74, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 144, 94, 110, 208, 16, 22, 114, 209, 144, 64, 78, 210, 16, 64, 145, 211, 144, 35, 75, 212, 144, 99, 164, 13, 16, 26, 139, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 0, 200, 232, 109, 170, 170, 170, 170, 170, 170, 170, 170, 102, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 10, 0, 0, 0, 26, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 116, 117, 0, 224, 218, 117, 0, 224, 223, 117, 0, 16, 250, 117, 0, 224, 55, 0, 0, 0, 8, 210, 176, 183, 96, 243, 62, 185, 96, 156, 239, 185, 96, 141, 223, 186, 96, 126, 207, 187, 224, 169, 200, 188, 224, 154, 184, 189, 224, 139, 168, 190, 224, 124, 152, 191, 224, 109, 136, 192, 224, 94, 120, 193, 224, 79, 104, 194, 224, 64, 88, 195, 224, 49, 72, 196, 224, 34, 56, 197, 224, 19, 40, 198, 224, 4, 24, 199, 96, 209, 173, 17, 80, 224, 83, 18, 208, 11, 77, 19, 96, 208, 51, 20, 128, 221, 35, 21, 128, 206, 19, 22, 128, 191, 3, 23, 128, 176, 243, 23, 128, 161, 227, 24, 128, 146, 211, 25, 128, 131, 195, 26, 0, 175, 188, 27, 0, 160, 172, 28, 0, 145, 156, 29, 0, 130, 140, 30, 0, 115, 124, 31, 0, 100, 108, 32, 0, 85, 92, 33, 0, 70, 76, 34, 0, 55, 60, 35, 0, 40, 44, 36, 0, 25, 28, 37, 0, 10, 12, 38, 128, 53, 5, 39, 96, 10, 245, 39, 96, 251, 228, 40, 96, 236, 212, 41, 96, 221, 196, 42, 96, 206, 180, 43, 96, 191, 164, 44, 96, 176, 148, 45, 80, 147, 132, 46, 96, 146, 116, 47, 80, 117, 100, 48, 224, 174, 93, 49, 208, 123, 114, 50, 16, 187, 61, 51, 16, 150, 82, 52, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 55, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 120, 24, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 4, 0, 0, 96, 206, 7, 0, 112, 16, 118, 0, 224, 75, 118, 0, 16, 90, 118, 0, 224, 68, 0, 0, 0, 96, 23, 12, 155, 240, 218, 213, 155, 144, 174, 217, 156, 144, 181, 164, 157, 160, 41, 167, 158, 144, 151, 132, 159, 32, 210, 154, 160, 144, 194, 192, 161, 112, 195, 243, 201, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 224, 120, 153, 209, 96, 187, 138, 210, 144, 166, 80, 211, 16, 210, 73, 212, 16, 195, 57, 213, 16, 180, 41, 214, 16, 165, 25, 215, 16, 150, 9, 216, 144, 193, 2, 217, 16, 120, 233, 217, 144, 47, 237, 218, 16, 91, 230, 219, 240, 168, 162, 226, 96, 242, 81, 227, 112, 220, 131, 228, 224, 37, 51, 229, 240, 225, 116, 230, 96, 182, 17, 231, 0, 210, 84, 232, 144, 194, 241, 232, 0, 54, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 28, 179, 238, 106, 69, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 228, 17, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 107, 118, 0, 224, 176, 118, 0, 224, 179, 118, 0, 16, 198, 118, 0, 224, 2, 0, 0, 0, 195, 1, 0, 0, 234, 1, 0, 0, 60, 0, 0, 0, 12, 159, 107, 158, 8, 210, 176, 183, 96, 243, 62, 185, 96, 156, 239, 185, 96, 141, 223, 186, 96, 126, 207, 187, 224, 169, 200, 188, 224, 154, 184, 189, 224, 139, 168, 190, 224, 124, 152, 191, 224, 109, 136, 192, 224, 94, 120, 193, 224, 79, 104, 194, 224, 64, 88, 195, 224, 49, 72, 196, 224, 34, 56, 197, 224, 19, 40, 198, 224, 4, 24, 199, 96, 147, 188, 200, 80, 125, 119, 202, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 96, 144, 78, 208, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 224, 76, 67, 38, 128, 53, 5, 39, 128, 38, 245, 39, 128, 23, 229, 40, 96, 236, 212, 41, 80, 207, 196, 42, 96, 206, 180, 43, 80, 177, 164, 44, 96, 176, 148, 45, 80, 147, 132, 46, 96, 146, 116, 47, 80, 117, 100, 48, 224, 174, 93, 49, 208, 123, 114, 50, 0, 173, 61, 51, 0, 136, 82, 52, 2, 0, 0, 0, 255, 255, 255, 255, 248, 200, 182, 86, 61, 0, 0, 0, 4, 3, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 2, 1, 2, 1, 2, 7, 8, 7, 8, 7, 8, 7, 8, 7, 8, 7, 8, 7, 8, 7, 8, 7, 8, 7, 8, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 170, 170, 170, 18, 0, 0, 0, 8, 27, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 120, 24, 0, 0, 0, 0, 0, 0, 244, 26, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 41, 4, 0, 96, 206, 7, 0, 112, 216, 118, 0, 224, 219, 118, 0, 224, 24, 119, 0, 224, 27, 119, 0, 16, 44, 119, 0, 224, 50, 0, 0, 0, 96, 140, 30, 155, 208, 190, 213, 155, 112, 87, 67, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 16, 22, 114, 209, 144, 16, 36, 210, 16, 133, 121, 211, 144, 173, 27, 212, 16, 173, 94, 213, 16, 224, 223, 213, 144, 201, 71, 215, 16, 194, 191, 215, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 52, 239, 12, 113, 170, 170, 170, 170, 51, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 204, 11, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 76, 119, 0, 224, 127, 119, 0, 224, 131, 119, 0, 16, 145, 119, 0, 224, 2, 0, 0, 0, 146, 1, 0, 0, 197, 1, 0, 0, 145, 0, 0, 0, 145, 179, 38, 155, 17, 11, 214, 155, 160, 48, 207, 156, 160, 195, 164, 157, 160, 157, 156, 158, 160, 26, 151, 159, 32, 186, 133, 160, 160, 252, 118, 161, 32, 156, 101, 162, 160, 200, 123, 163, 160, 184, 78, 164, 32, 251, 63, 165, 32, 96, 37, 166, 32, 198, 39, 167, 32, 44, 42, 168, 160, 248, 235, 168, 160, 211, 0, 170, 32, 21, 213, 170, 32, 240, 233, 171, 32, 108, 199, 172, 32, 210, 201, 173, 32, 78, 167, 174, 160, 121, 160, 175, 32, 48, 135, 176, 160, 208, 146, 177, 160, 76, 112, 178, 160, 178, 114, 179, 160, 46, 80, 180, 32, 90, 73, 181, 160, 16, 48, 182, 160, 118, 50, 183, 160, 242, 15, 184, 160, 88, 18, 185, 160, 212, 239, 185, 32, 0, 233, 186, 32, 241, 216, 187, 32, 87, 219, 188, 32, 211, 184, 189, 160, 254, 177, 190, 32, 181, 152, 191, 32, 27, 155, 192, 32, 151, 120, 193, 32, 253, 122, 194, 32, 121, 88, 195, 160, 164, 81, 196, 32, 91, 56, 197, 32, 193, 58, 198, 160, 214, 88, 199, 160, 9, 218, 199, 32, 224, 73, 212, 160, 33, 30, 213, 32, 172, 78, 214, 32, 40, 44, 215, 32, 142, 46, 216, 32, 149, 249, 216, 32, 112, 14, 218, 32, 236, 235, 218, 160, 23, 229, 219, 32, 206, 203, 220, 160, 249, 196, 221, 160, 234, 180, 222, 32, 22, 174, 223, 160, 204, 148, 224, 160, 72, 114, 225, 32, 116, 107, 226, 160, 42, 82, 227, 160, 144, 84, 228, 160, 12, 50, 229, 32, 173, 61, 230, 32, 41, 27, 231, 160, 84, 20, 232, 32, 11, 251, 232, 32, 113, 253, 233, 32, 237, 218, 234, 32, 83, 221, 235, 32, 207, 186, 236, 160, 250, 179, 237, 32, 177, 154, 238, 160, 103, 129, 239, 32, 125, 159, 240, 160, 73, 97, 241, 32, 95, 127, 242, 32, 102, 74, 243, 32, 65, 95, 244, 160, 13, 33, 245, 32, 35, 63, 246, 160, 239, 0, 247, 32, 5, 31, 248, 160, 209, 224, 248, 32, 231, 254, 249, 160, 179, 192, 250, 160, 3, 232, 251, 160, 171, 123, 252, 112, 187, 199, 253, 32, 198, 112, 3, 32, 88, 41, 4, 32, 168, 80, 5, 32, 58, 9, 6, 32, 138, 48, 7, 32, 28, 233, 7, 32, 108, 16, 9, 32, 254, 200, 9, 32, 78, 240, 10, 160, 26, 178, 11, 32, 48, 208, 12, 160, 252, 145, 13, 32, 18, 176, 14, 160, 222, 113, 15, 160, 46, 153, 16, 160, 192, 81, 17, 160, 16, 121, 18, 160, 162, 49, 19, 160, 242, 88, 20, 144, 235, 35, 21, 144, 198, 56, 22, 144, 205, 3, 23, 144, 168, 24, 24, 144, 175, 227, 24, 144, 138, 248, 25, 144, 145, 195, 26, 16, 167, 225, 27, 16, 174, 172, 28, 16, 137, 193, 29, 16, 144, 140, 30, 16, 107, 161, 31, 16, 114, 108, 32, 16, 77, 129, 33, 16, 84, 76, 34, 16, 47, 97, 35, 16, 54, 44, 36, 144, 75, 74, 37, 16, 24, 12, 38, 144, 45, 42, 39, 144, 52, 245, 39, 144, 15, 10, 41, 144, 22, 213, 41, 144, 241, 233, 42, 144, 248, 180, 43, 144, 211, 201, 44, 144, 218, 148, 45, 144, 181, 169, 46, 144, 188, 116, 47, 144, 151, 137, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 220, 10, 209, 87, 146, 0, 0, 0, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 5, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 12, 0, 0, 0, 36, 250, 255, 255, 0, 0, 0, 0, 15, 250, 255, 255, 0, 0, 0, 0, 15, 250, 255, 255, 16, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 0, 0, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 163, 119, 0, 224, 166, 119, 0, 224, 56, 120, 0, 224, 59, 120, 0, 16, 97, 120, 0, 224, 115, 0, 0, 0, 160, 173, 38, 155, 32, 5, 214, 155, 160, 48, 207, 156, 160, 195, 164, 157, 160, 157, 156, 158, 160, 26, 151, 159, 32, 186, 133, 160, 160, 252, 118, 161, 32, 156, 101, 162, 160, 200, 123, 163, 160, 184, 78, 164, 32, 251, 63, 165, 32, 96, 37, 166, 32, 198, 39, 167, 32, 44, 42, 168, 160, 248, 235, 168, 160, 211, 0, 170, 32, 21, 213, 170, 32, 240, 233, 171, 32, 108, 199, 172, 32, 210, 201, 173, 32, 78, 167, 174, 160, 121, 160, 175, 32, 48, 135, 176, 160, 208, 146, 177, 160, 76, 112, 178, 160, 178, 114, 179, 160, 46, 80, 180, 32, 90, 73, 181, 160, 16, 48, 182, 160, 118, 50, 183, 160, 242, 15, 184, 160, 88, 18, 185, 160, 212, 239, 185, 32, 0, 233, 186, 32, 241, 216, 187, 32, 87, 219, 188, 32, 211, 184, 189, 160, 254, 177, 190, 32, 181, 152, 191, 32, 27, 155, 192, 32, 151, 120, 193, 32, 253, 122, 194, 32, 121, 88, 195, 160, 164, 81, 196, 32, 91, 56, 197, 32, 193, 58, 198, 160, 214, 88, 199, 160, 9, 218, 199, 144, 38, 22, 202, 144, 89, 151, 202, 144, 30, 209, 203, 144, 59, 119, 204, 144, 0, 177, 205, 16, 88, 96, 206, 144, 226, 144, 207, 144, 94, 110, 208, 16, 22, 114, 209, 16, 50, 251, 209, 32, 254, 105, 210, 160, 41, 99, 211, 32, 224, 73, 212, 160, 33, 30, 213, 144, 253, 66, 213, 16, 224, 223, 213, 32, 172, 78, 214, 160, 3, 254, 214, 32, 142, 46, 216, 32, 149, 249, 216, 32, 112, 14, 218, 32, 236, 235, 218, 160, 23, 229, 219, 32, 206, 203, 220, 160, 249, 196, 221, 160, 234, 180, 222, 32, 22, 174, 223, 160, 204, 148, 224, 160, 72, 114, 225, 32, 116, 107, 226, 160, 42, 82, 227, 160, 144, 84, 228, 160, 12, 50, 229, 32, 173, 61, 230, 32, 41, 27, 231, 160, 84, 20, 232, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 4, 10, 209, 87, 170, 170, 170, 170, 116, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 1, 2, 1, 2, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 12, 0, 0, 0, 252, 250, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 123, 120, 0, 224, 239, 120, 0, 224, 243, 120, 0, 16, 17, 121, 0, 224, 2, 0, 0, 0, 200, 1, 0, 0, 213, 1, 0, 0, 35, 0, 0, 0, 27, 111, 115, 164, 96, 81, 206, 203, 96, 229, 192, 204, 128, 221, 35, 21, 128, 206, 19, 22, 128, 191, 3, 23, 128, 176, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 35, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 101, 23, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 32, 28, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 41, 121, 0, 224, 44, 121, 0, 224, 83, 121, 0, 16, 93, 121, 0, 224, 3, 0, 0, 0, 19, 1, 0, 0, 202, 1, 0, 0, 101, 2, 0, 0, 114, 0, 0, 0, 152, 245, 139, 144, 96, 23, 12, 155, 208, 190, 213, 155, 224, 99, 101, 162, 80, 130, 123, 163, 96, 128, 78, 164, 208, 180, 63, 165, 224, 39, 37, 166, 208, 127, 39, 167, 96, 40, 40, 170, 208, 253, 225, 170, 224, 137, 249, 171, 80, 49, 195, 172, 224, 63, 129, 200, 80, 19, 1, 201, 96, 245, 74, 201, 80, 128, 206, 202, 96, 174, 203, 203, 80, 9, 107, 210, 96, 57, 162, 211, 80, 2, 67, 212, 224, 13, 76, 213, 208, 123, 41, 214, 224, 239, 43, 215, 208, 93, 9, 216, 96, 151, 2, 217, 208, 63, 233, 217, 224, 179, 235, 218, 80, 92, 210, 219, 96, 208, 212, 220, 80, 62, 178, 221, 96, 185, 244, 241, 80, 239, 98, 244, 96, 6, 104, 245, 208, 56, 31, 246, 112, 147, 110, 6, 112, 154, 57, 7, 0, 117, 251, 7, 112, 124, 25, 9, 0, 203, 208, 9, 112, 94, 249, 10, 128, 254, 177, 11, 112, 64, 217, 12, 128, 85, 164, 13, 112, 173, 166, 14, 128, 55, 132, 15, 80, 17, 248, 15, 112, 176, 137, 25, 224, 176, 220, 25, 240, 208, 230, 27, 240, 239, 198, 28, 112, 49, 155, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 112, 39, 5, 39, 112, 24, 245, 39, 112, 9, 229, 40, 112, 250, 212, 41, 112, 235, 196, 42, 112, 220, 180, 43, 112, 205, 164, 44, 240, 131, 139, 45, 112, 175, 132, 46, 112, 160, 116, 47, 112, 145, 100, 48, 240, 188, 93, 49, 240, 151, 114, 50, 240, 158, 61, 51, 240, 121, 82, 52, 240, 128, 29, 53, 240, 91, 50, 54, 240, 98, 253, 54, 112, 120, 27, 56, 240, 68, 221, 56, 112, 90, 251, 57, 240, 38, 189, 58, 112, 60, 219, 59, 112, 67, 166, 60, 112, 30, 187, 61, 112, 37, 134, 62, 112, 0, 155, 63, 112, 7, 102, 64, 240, 28, 132, 65, 112, 233, 69, 66, 240, 254, 99, 67, 112, 203, 37, 68, 240, 224, 67, 69, 144, 201, 5, 70, 16, 223, 35, 71, 16, 230, 238, 71, 16, 193, 3, 73, 16, 200, 206, 73, 16, 163, 227, 74, 16, 170, 174, 75, 144, 191, 204, 76, 144, 221, 143, 77, 144, 161, 172, 78, 16, 110, 110, 79, 144, 131, 140, 80, 144, 138, 87, 81, 144, 101, 108, 82, 16, 190, 56, 83, 144, 71, 76, 84, 144, 78, 23, 85, 144, 158, 62, 86, 144, 48, 247, 86, 80, 46, 207, 87, 2, 0, 0, 0, 255, 255, 255, 255, 216, 200, 182, 86, 170, 170, 170, 170, 170, 170, 170, 170, 115, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 4, 5, 4, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 4, 170, 12, 0, 0, 0, 40, 27, 0, 0, 0, 0, 0, 0, 104, 27, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 111, 121, 0, 224, 115, 121, 0, 224, 230, 121, 0, 224, 235, 121, 0, 16, 9, 122, 0, 224, 79, 0, 0, 0, 96, 23, 12, 155, 240, 218, 213, 155, 144, 174, 217, 156, 144, 181, 164, 157, 144, 144, 185, 158, 144, 151, 132, 159, 144, 113, 9, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 16, 22, 114, 209, 224, 119, 124, 209, 96, 132, 149, 209, 80, 173, 138, 210, 224, 182, 89, 211, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 0, 25, 28, 37, 0, 10, 12, 38, 128, 53, 5, 39, 128, 38, 245, 39, 128, 23, 229, 40, 128, 8, 213, 41, 128, 249, 196, 42, 128, 234, 180, 43, 128, 219, 164, 44, 128, 204, 148, 45, 128, 189, 132, 46, 128, 174, 116, 47, 128, 159, 100, 48, 0, 203, 93, 49, 0, 166, 114, 50, 0, 173, 61, 51, 0, 136, 82, 52, 0, 143, 29, 53, 0, 106, 50, 54, 0, 113, 253, 54, 128, 134, 27, 56, 0, 83, 221, 56, 128, 104, 251, 57, 0, 53, 189, 58, 128, 74, 219, 59, 128, 81, 166, 60, 128, 44, 187, 61, 128, 51, 134, 62, 128, 14, 155, 63, 128, 21, 102, 64, 0, 43, 132, 65, 128, 247, 69, 66, 0, 13, 100, 67, 128, 217, 37, 68, 0, 239, 67, 69, 128, 187, 5, 70, 0, 209, 35, 71, 0, 216, 238, 71, 0, 179, 3, 73, 0, 186, 206, 73, 0, 149, 227, 74, 0, 156, 174, 75, 128, 177, 204, 76, 0, 126, 142, 77, 112, 43, 76, 84, 2, 0, 0, 0, 255, 255, 255, 255, 72, 91, 162, 111, 170, 170, 170, 170, 170, 170, 170, 170, 80, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 4, 3, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 3, 14, 0, 0, 0, 56, 19, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 30, 122, 0, 224, 110, 122, 0, 224, 115, 122, 0, 16, 136, 122, 0, 224, 38, 0, 0, 0, 100, 167, 25, 170, 96, 25, 164, 181, 208, 46, 205, 202, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 112, 168, 205, 206, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 224, 32, 141, 38, 128, 23, 229, 40, 96, 236, 212, 41, 80, 207, 196, 42, 96, 206, 180, 43, 80, 177, 164, 44, 96, 176, 148, 45, 80, 147, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 38, 0, 0, 0, 3, 5, 2, 1, 2, 1, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 14, 0, 0, 0, 156, 28, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 158, 122, 0, 224, 199, 122, 0, 16, 210, 122, 0, 224, 63, 0, 0, 0, 128, 57, 0, 161, 80, 11, 164, 181, 192, 153, 39, 21, 48, 206, 24, 22, 64, 205, 8, 23, 176, 1, 250, 23, 192, 0, 234, 24, 48, 53, 219, 25, 192, 133, 204, 26, 224, 146, 188, 27, 224, 131, 172, 28, 224, 116, 156, 29, 224, 101, 140, 30, 224, 86, 124, 31, 224, 71, 108, 32, 224, 56, 92, 33, 224, 41, 76, 34, 224, 26, 60, 35, 224, 11, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 112, 39, 5, 39, 112, 24, 245, 39, 96, 236, 212, 41, 112, 235, 196, 42, 112, 220, 180, 43, 112, 205, 164, 44, 112, 190, 148, 45, 112, 175, 132, 46, 112, 160, 116, 47, 112, 145, 100, 48, 240, 188, 93, 49, 240, 151, 114, 50, 240, 158, 61, 51, 240, 121, 82, 52, 240, 128, 29, 53, 240, 91, 50, 54, 240, 98, 253, 54, 112, 120, 27, 56, 240, 68, 221, 56, 112, 90, 251, 57, 240, 38, 189, 58, 112, 60, 219, 59, 112, 67, 166, 60, 112, 30, 187, 61, 112, 37, 134, 62, 112, 0, 155, 63, 112, 7, 102, 64, 240, 28, 132, 65, 112, 233, 69, 66, 240, 254, 99, 67, 112, 203, 37, 68, 240, 224, 67, 69, 112, 173, 5, 70, 240, 194, 35, 71, 240, 201, 238, 71, 240, 164, 3, 73, 240, 171, 206, 73, 240, 134, 227, 74, 240, 141, 174, 75, 112, 163, 204, 76, 240, 111, 142, 77, 96, 29, 76, 84, 63, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 2, 1, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 170, 10, 0, 0, 0, 152, 46, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 235, 122, 0, 224, 43, 123, 0, 16, 60, 123, 0, 224, 2, 0, 0, 0, 207, 1, 0, 0, 83, 2, 0, 0, 139, 0, 0, 0, 128, 142, 230, 146, 112, 109, 75, 155, 128, 199, 254, 155, 112, 237, 156, 156, 112, 131, 201, 157, 112, 114, 127, 158, 240, 182, 170, 159, 112, 84, 95, 160, 112, 234, 139, 161, 112, 217, 65, 162, 112, 111, 110, 163, 240, 12, 35, 164, 240, 162, 79, 165, 112, 239, 5, 170, 240, 142, 244, 170, 240, 167, 201, 173, 240, 35, 167, 174, 112, 79, 160, 175, 240, 5, 135, 176, 240, 107, 137, 177, 112, 34, 112, 178, 112, 136, 114, 179, 112, 4, 80, 180, 112, 76, 50, 183, 112, 200, 15, 184, 112, 185, 255, 184, 112, 170, 239, 185, 240, 183, 200, 188, 240, 168, 184, 189, 112, 95, 159, 190, 240, 138, 152, 191, 240, 240, 154, 192, 240, 108, 120, 193, 240, 93, 104, 194, 240, 78, 88, 195, 112, 5, 63, 196, 240, 48, 56, 197, 240, 150, 58, 198, 112, 172, 88, 199, 112, 223, 217, 199, 112, 47, 1, 201, 112, 32, 241, 201, 240, 98, 226, 202, 240, 82, 181, 203, 224, 163, 236, 203, 224, 75, 128, 204, 240, 162, 220, 204, 240, 52, 149, 205, 96, 75, 195, 205, 224, 162, 114, 206, 112, 191, 197, 206, 240, 22, 117, 207, 224, 103, 172, 207, 224, 132, 82, 208, 112, 161, 165, 208, 240, 248, 84, 209, 224, 73, 140, 209, 224, 102, 50, 210, 112, 131, 133, 210, 240, 196, 89, 211, 240, 181, 73, 212, 32, 209, 57, 213, 32, 194, 41, 214, 32, 179, 25, 215, 32, 164, 9, 216, 32, 149, 249, 216, 32, 134, 233, 217, 32, 89, 185, 220, 160, 132, 178, 221, 160, 117, 162, 222, 160, 102, 146, 223, 160, 87, 130, 224, 160, 72, 114, 225, 160, 57, 98, 226, 160, 42, 82, 227, 160, 27, 66, 228, 160, 12, 50, 229, 160, 253, 33, 230, 32, 41, 27, 231, 32, 26, 11, 232, 32, 11, 251, 232, 32, 252, 234, 233, 32, 237, 218, 234, 32, 222, 202, 235, 32, 207, 186, 236, 32, 192, 170, 237, 32, 177, 154, 238, 32, 162, 138, 239, 32, 147, 122, 240, 32, 132, 106, 241, 160, 175, 99, 242, 160, 160, 83, 243, 160, 145, 67, 244, 160, 130, 51, 245, 160, 115, 35, 246, 160, 100, 19, 247, 160, 85, 3, 248, 160, 70, 243, 248, 0, 42, 171, 12, 0, 27, 155, 13, 0, 12, 139, 14, 128, 55, 132, 15, 128, 40, 116, 16, 128, 25, 100, 17, 144, 24, 84, 18, 128, 251, 67, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 160, 189, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 139, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 1, 2, 3, 2, 1, 2, 3, 2, 1, 2, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 4, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 5, 4, 5, 4, 5, 4, 2, 1, 170, 12, 0, 0, 0, 99, 247, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 0, 0, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 76, 123, 0, 224, 79, 123, 0, 224, 219, 123, 0, 16, 255, 123, 0, 224, 7, 0, 0, 0, 187, 1, 0, 0, 199, 1, 0, 0, 201, 1, 0, 0, 203, 1, 0, 0, 209, 1, 0, 0, 247, 1, 0, 0, 248, 1, 0, 0, 159, 0, 0, 0, 160, 173, 38, 155, 32, 5, 214, 155, 160, 48, 207, 156, 160, 195, 164, 157, 160, 157, 156, 158, 160, 26, 151, 159, 32, 186, 133, 160, 160, 252, 118, 161, 32, 156, 101, 162, 160, 200, 123, 163, 160, 184, 78, 164, 32, 251, 63, 165, 32, 96, 37, 166, 32, 198, 39, 167, 32, 44, 42, 168, 160, 248, 235, 168, 160, 211, 0, 170, 32, 21, 213, 170, 32, 240, 233, 171, 32, 108, 199, 172, 32, 210, 201, 173, 32, 78, 167, 174, 160, 121, 160, 175, 32, 48, 135, 176, 160, 208, 146, 177, 160, 76, 112, 178, 160, 178, 114, 179, 160, 46, 80, 180, 32, 90, 73, 181, 160, 16, 48, 182, 160, 118, 50, 183, 160, 242, 15, 184, 160, 88, 18, 185, 160, 212, 239, 185, 32, 0, 233, 186, 32, 241, 216, 187, 32, 87, 219, 188, 32, 211, 184, 189, 160, 254, 177, 190, 32, 181, 152, 191, 32, 27, 155, 192, 32, 151, 120, 193, 32, 253, 122, 194, 32, 121, 88, 195, 160, 164, 81, 196, 32, 91, 56, 197, 32, 193, 58, 198, 160, 214, 88, 199, 160, 9, 218, 199, 144, 38, 22, 202, 144, 89, 151, 202, 144, 30, 209, 203, 144, 59, 119, 204, 144, 0, 177, 205, 16, 88, 96, 206, 144, 226, 144, 207, 144, 94, 110, 208, 16, 22, 114, 209, 16, 50, 251, 209, 32, 254, 105, 210, 160, 41, 99, 211, 32, 224, 73, 212, 160, 33, 30, 213, 144, 253, 66, 213, 16, 224, 223, 213, 32, 172, 78, 214, 160, 3, 254, 214, 32, 142, 46, 216, 32, 149, 249, 216, 32, 112, 14, 218, 32, 236, 235, 218, 160, 23, 229, 219, 32, 206, 203, 220, 160, 249, 196, 221, 160, 234, 180, 222, 32, 22, 174, 223, 160, 204, 148, 224, 160, 72, 114, 225, 32, 116, 107, 226, 160, 42, 82, 227, 160, 144, 84, 228, 160, 12, 50, 229, 32, 173, 61, 230, 32, 41, 27, 231, 160, 84, 20, 232, 32, 11, 251, 232, 32, 113, 253, 233, 32, 237, 218, 234, 32, 83, 221, 235, 32, 207, 186, 236, 160, 250, 179, 237, 32, 177, 154, 238, 160, 103, 129, 239, 32, 125, 159, 240, 160, 73, 97, 241, 32, 95, 127, 242, 32, 102, 74, 243, 32, 65, 95, 244, 160, 13, 33, 245, 32, 35, 63, 246, 160, 239, 0, 247, 32, 5, 31, 248, 160, 209, 224, 248, 32, 231, 254, 249, 160, 179, 192, 250, 160, 3, 232, 251, 160, 171, 123, 252, 112, 187, 199, 253, 32, 198, 112, 3, 32, 88, 41, 4, 32, 168, 80, 5, 32, 58, 9, 6, 32, 138, 48, 7, 32, 28, 233, 7, 32, 108, 16, 9, 32, 254, 200, 9, 32, 78, 240, 10, 160, 26, 178, 11, 32, 48, 208, 12, 160, 252, 145, 13, 32, 18, 176, 14, 160, 222, 113, 15, 160, 46, 153, 16, 160, 192, 81, 17, 160, 16, 121, 18, 160, 162, 49, 19, 160, 242, 88, 20, 144, 235, 35, 21, 144, 198, 56, 22, 144, 205, 3, 23, 144, 168, 24, 24, 144, 175, 227, 24, 144, 138, 248, 25, 144, 145, 195, 26, 16, 167, 225, 27, 16, 174, 172, 28, 16, 137, 193, 29, 16, 144, 140, 30, 16, 107, 161, 31, 16, 114, 108, 32, 16, 77, 129, 33, 16, 84, 76, 34, 16, 47, 97, 35, 16, 54, 44, 36, 144, 75, 74, 37, 16, 24, 12, 38, 144, 45, 42, 39, 144, 52, 245, 39, 144, 15, 10, 41, 144, 22, 213, 41, 144, 241, 233, 42, 144, 248, 180, 43, 144, 211, 201, 44, 144, 218, 148, 45, 144, 181, 169, 46, 144, 188, 116, 47, 144, 151, 137, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 203, 9, 93, 26, 170, 170, 170, 170, 160, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 1, 2, 1, 2, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 10, 0, 0, 0, 181, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 0, 0, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 23, 124, 0, 224, 31, 124, 0, 224, 191, 124, 0, 224, 195, 124, 0, 16, 236, 124, 0, 224, 101, 0, 0, 0, 188, 173, 162, 132, 96, 140, 30, 155, 240, 218, 213, 155, 224, 167, 234, 156, 112, 153, 164, 157, 144, 144, 185, 158, 144, 151, 132, 159, 112, 196, 224, 159, 240, 165, 96, 160, 160, 229, 126, 161, 240, 18, 46, 162, 16, 105, 122, 163, 240, 129, 53, 164, 144, 63, 94, 165, 240, 53, 37, 166, 0, 170, 39, 167, 240, 1, 42, 168, 16, 154, 7, 169, 112, 52, 238, 169, 0, 110, 231, 170, 112, 162, 216, 171, 0, 80, 199, 172, 240, 167, 201, 173, 0, 50, 167, 174, 112, 79, 160, 175, 0, 20, 135, 176, 240, 107, 137, 177, 128, 48, 112, 178, 112, 136, 114, 179, 160, 46, 80, 180, 32, 90, 73, 181, 160, 16, 48, 182, 160, 118, 50, 183, 160, 242, 15, 184, 160, 227, 255, 184, 160, 212, 239, 185, 32, 139, 214, 186, 32, 241, 216, 187, 32, 226, 200, 188, 32, 211, 184, 189, 160, 137, 159, 190, 32, 181, 152, 191, 32, 27, 155, 192, 32, 151, 120, 193, 32, 136, 104, 194, 32, 121, 88, 195, 160, 47, 63, 196, 32, 91, 56, 197, 32, 193, 58, 198, 160, 214, 88, 199, 160, 9, 218, 199, 32, 48, 66, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 176, 111, 208, 16, 22, 114, 209, 144, 64, 78, 210, 16, 64, 145, 211, 144, 35, 75, 212, 144, 99, 164, 13, 16, 26, 139, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 101, 0, 0, 0, 3, 4, 3, 4, 3, 4, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 170, 10, 0, 0, 0, 196, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 4, 125, 0, 224, 107, 125, 0, 16, 134, 125, 0, 224, 79, 0, 0, 0, 240, 197, 186, 158, 0, 57, 160, 159, 240, 27, 144, 160, 128, 108, 129, 161, 112, 239, 5, 170, 0, 110, 231, 170, 240, 167, 201, 173, 0, 50, 167, 174, 112, 79, 160, 175, 0, 20, 135, 176, 0, 122, 137, 177, 128, 48, 112, 178, 112, 136, 114, 179, 128, 18, 80, 180, 240, 236, 201, 194, 0, 93, 88, 195, 240, 63, 72, 196, 224, 27, 109, 196, 96, 116, 57, 197, 128, 91, 33, 199, 240, 142, 245, 199, 96, 222, 245, 203, 240, 113, 149, 204, 96, 75, 195, 205, 112, 213, 160, 206, 96, 45, 163, 207, 112, 183, 128, 208, 96, 15, 131, 209, 112, 153, 96, 210, 96, 241, 98, 211, 112, 123, 64, 212, 224, 70, 30, 217, 240, 91, 233, 217, 224, 205, 13, 8, 112, 146, 244, 8, 224, 175, 237, 9, 112, 116, 212, 10, 224, 28, 187, 11, 240, 27, 171, 12, 96, 57, 164, 13, 240, 253, 138, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 0, 181, 54, 126, 170, 170, 170, 170, 80, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 1, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 12, 0, 0, 0, 140, 252, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 155, 125, 0, 224, 235, 125, 0, 224, 239, 125, 0, 16, 4, 126, 0, 224, 86, 0, 0, 0, 112, 248, 56, 155, 224, 204, 213, 155, 240, 203, 197, 156, 96, 0, 183, 157, 112, 254, 137, 158, 224, 28, 160, 159, 240, 165, 96, 160, 96, 173, 126, 161, 112, 55, 92, 162, 96, 26, 76, 163, 240, 53, 108, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 144, 226, 144, 207, 144, 94, 110, 208, 16, 22, 114, 209, 240, 210, 76, 210, 144, 49, 62, 211, 16, 210, 73, 212, 112, 247, 29, 213, 240, 151, 41, 214, 144, 128, 235, 214, 16, 150, 9, 216, 240, 181, 51, 249, 224, 196, 217, 249, 112, 210, 28, 251, 240, 180, 185, 251, 112, 180, 252, 252, 240, 150, 153, 253, 240, 208, 229, 254, 112, 179, 130, 255, 240, 178, 197, 0, 112, 149, 98, 1, 112, 90, 156, 2, 112, 119, 66, 3, 240, 118, 133, 4, 240, 147, 43, 5, 112, 51, 26, 6, 112, 36, 10, 7, 112, 22, 23, 8, 112, 52, 218, 8, 144, 20, 247, 9, 128, 13, 194, 10, 144, 246, 214, 11, 128, 239, 161, 12, 144, 216, 182, 13, 128, 209, 129, 14, 144, 186, 150, 15, 128, 179, 97, 16, 144, 156, 118, 17, 128, 149, 65, 18, 16, 91, 69, 19, 0, 178, 42, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 100, 211, 189, 112, 170, 170, 170, 170, 87, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 156, 13, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 28, 126, 0, 224, 115, 126, 0, 224, 119, 126, 0, 16, 142, 126, 0, 224, 67, 0, 0, 0, 56, 170, 25, 170, 96, 25, 164, 181, 208, 112, 94, 202, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 96, 2, 10, 208, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 112, 24, 245, 39, 128, 23, 229, 40, 128, 8, 213, 41, 128, 249, 196, 42, 128, 234, 180, 43, 128, 219, 164, 44, 128, 204, 148, 45, 128, 189, 132, 46, 128, 174, 116, 47, 128, 159, 100, 48, 0, 203, 93, 49, 0, 166, 114, 50, 0, 173, 61, 51, 0, 136, 82, 52, 0, 143, 29, 53, 0, 106, 50, 54, 0, 113, 253, 54, 128, 134, 27, 56, 0, 83, 221, 56, 128, 104, 251, 57, 0, 53, 189, 58, 128, 74, 219, 59, 128, 81, 166, 60, 128, 44, 187, 61, 128, 51, 134, 62, 128, 14, 155, 63, 128, 21, 102, 64, 0, 43, 132, 65, 128, 247, 69, 66, 0, 13, 100, 67, 128, 217, 37, 68, 0, 239, 67, 69, 128, 187, 5, 70, 0, 209, 35, 71, 0, 216, 238, 71, 0, 179, 3, 73, 0, 186, 206, 73, 0, 149, 227, 74, 0, 156, 174, 75, 128, 177, 204, 76, 0, 126, 142, 77, 2, 0, 0, 0, 255, 255, 255, 255, 40, 202, 182, 86, 68, 0, 0, 0, 3, 4, 6, 2, 1, 2, 1, 2, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 6, 16, 0, 0, 0, 216, 25, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 200, 25, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 160, 126, 0, 224, 228, 126, 0, 224, 231, 126, 0, 16, 249, 126, 0, 224, 102, 0, 0, 0, 79, 80, 96, 145, 240, 120, 71, 155, 112, 44, 215, 155, 112, 145, 188, 156, 240, 72, 192, 157, 112, 254, 137, 158, 240, 42, 160, 159, 240, 165, 96, 160, 240, 12, 128, 161, 240, 18, 46, 162, 240, 76, 122, 163, 240, 129, 53, 164, 112, 35, 94, 165, 240, 53, 37, 166, 240, 155, 39, 167, 112, 38, 88, 168, 240, 125, 7, 169, 112, 52, 238, 169, 240, 95, 231, 170, 240, 80, 215, 171, 240, 65, 199, 172, 240, 167, 201, 173, 240, 35, 167, 174, 112, 79, 160, 175, 240, 5, 135, 176, 240, 107, 137, 177, 112, 34, 112, 178, 112, 136, 114, 179, 112, 4, 80, 180, 240, 47, 73, 181, 112, 230, 47, 182, 112, 76, 50, 183, 112, 200, 15, 184, 112, 185, 255, 184, 112, 170, 239, 185, 240, 96, 214, 186, 240, 198, 216, 187, 240, 183, 200, 188, 240, 168, 184, 189, 112, 95, 159, 190, 240, 138, 152, 191, 240, 240, 154, 192, 240, 108, 120, 193, 240, 93, 104, 194, 240, 78, 88, 195, 112, 5, 63, 196, 240, 48, 56, 197, 240, 150, 58, 198, 112, 172, 88, 199, 160, 9, 218, 199, 240, 91, 23, 202, 224, 84, 226, 202, 240, 105, 173, 203, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 240, 241, 137, 208, 16, 22, 114, 209, 144, 64, 78, 210, 0, 57, 187, 11, 240, 27, 171, 12, 144, 99, 164, 13, 16, 26, 139, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 148, 69, 200, 107, 103, 0, 0, 0, 4, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 170, 14, 0, 0, 0, 236, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 49, 2, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 17, 127, 0, 224, 120, 127, 0, 224, 123, 127, 0, 16, 150, 127, 0, 224, 2, 0, 0, 0, 216, 1, 0, 0, 119, 2, 0, 0, 77, 0, 0, 0, 199, 30, 95, 155, 121, 242, 62, 157, 249, 238, 42, 158, 105, 57, 247, 158, 249, 87, 132, 159, 233, 108, 216, 160, 128, 57, 0, 161, 64, 166, 60, 161, 192, 109, 16, 164, 176, 50, 61, 164, 176, 104, 21, 165, 192, 3, 61, 165, 80, 69, 30, 167, 96, 25, 164, 181, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 112, 39, 5, 39, 112, 24, 245, 39, 128, 23, 229, 40, 128, 191, 120, 41, 112, 250, 212, 41, 112, 235, 196, 42, 112, 220, 180, 43, 112, 205, 164, 44, 112, 190, 148, 45, 112, 175, 132, 46, 112, 160, 116, 47, 112, 145, 100, 48, 240, 188, 93, 49, 240, 151, 114, 50, 240, 158, 61, 51, 240, 121, 82, 52, 240, 128, 29, 53, 240, 91, 50, 54, 240, 98, 253, 54, 112, 120, 27, 56, 240, 68, 221, 56, 112, 90, 251, 57, 240, 38, 189, 58, 112, 60, 219, 59, 112, 67, 166, 60, 112, 30, 187, 61, 112, 37, 134, 62, 112, 0, 155, 63, 112, 7, 102, 64, 240, 28, 132, 65, 112, 233, 69, 66, 240, 254, 99, 67, 112, 203, 37, 68, 240, 224, 67, 69, 112, 173, 5, 70, 240, 194, 35, 71, 240, 201, 238, 71, 240, 164, 3, 73, 240, 171, 206, 73, 240, 134, 227, 74, 240, 141, 174, 75, 112, 163, 204, 76, 240, 111, 142, 77, 96, 29, 76, 84, 170, 170, 170, 170, 170, 170, 170, 170, 77, 0, 0, 0, 3, 4, 3, 5, 4, 5, 7, 6, 7, 8, 7, 6, 1, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 2, 1, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 9, 6, 170, 170, 170, 20, 0, 0, 0, 57, 35, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 119, 35, 0, 0, 0, 0, 0, 0, 119, 35, 0, 0, 16, 14, 0, 0, 119, 35, 0, 0, 32, 28, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 32, 28, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 176, 127, 0, 224, 179, 127, 0, 224, 3, 128, 0, 16, 24, 128, 0, 224, 3, 0, 0, 0, 237, 0, 0, 0, 87, 1, 0, 0, 218, 1, 0, 0, 58, 0, 0, 0, 0, 227, 39, 155, 96, 123, 212, 155, 96, 77, 183, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 16, 22, 114, 209, 16, 7, 98, 210, 144, 32, 175, 235, 16, 76, 168, 236, 16, 61, 152, 237, 16, 46, 136, 238, 16, 31, 120, 239, 16, 16, 104, 240, 16, 1, 88, 241, 16, 242, 71, 242, 16, 227, 55, 243, 16, 212, 39, 244, 16, 197, 23, 245, 144, 240, 16, 246, 16, 6, 47, 247, 144, 210, 240, 247, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 108, 36, 238, 114, 170, 170, 170, 170, 59, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 20, 10, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 52, 128, 0, 224, 56, 128, 0, 224, 115, 128, 0, 224, 119, 128, 0, 16, 135, 128, 0, 224, 2, 0, 0, 0, 141, 1, 0, 0, 219, 1, 0, 0, 101, 0, 0, 0, 139, 80, 96, 145, 240, 120, 71, 155, 112, 44, 215, 155, 112, 145, 188, 156, 240, 72, 192, 157, 112, 254, 137, 158, 240, 42, 160, 159, 240, 165, 96, 160, 240, 12, 128, 161, 240, 18, 46, 162, 240, 76, 122, 163, 240, 129, 53, 164, 112, 35, 94, 165, 240, 53, 37, 166, 240, 155, 39, 167, 112, 38, 88, 168, 240, 125, 7, 169, 112, 52, 238, 169, 240, 95, 231, 170, 240, 80, 215, 171, 240, 65, 199, 172, 240, 167, 201, 173, 240, 35, 167, 174, 112, 79, 160, 175, 240, 5, 135, 176, 240, 107, 137, 177, 112, 34, 112, 178, 112, 136, 114, 179, 112, 4, 80, 180, 240, 47, 73, 181, 112, 230, 47, 182, 112, 76, 50, 183, 112, 200, 15, 184, 112, 185, 255, 184, 112, 170, 239, 185, 240, 96, 214, 186, 240, 198, 216, 187, 240, 183, 200, 188, 240, 168, 184, 189, 112, 95, 159, 190, 240, 138, 152, 191, 240, 240, 154, 192, 240, 108, 120, 193, 240, 93, 104, 194, 240, 78, 88, 195, 112, 5, 63, 196, 240, 48, 56, 197, 240, 150, 58, 198, 112, 172, 88, 199, 160, 9, 218, 199, 224, 39, 108, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 224, 225, 79, 208, 240, 241, 137, 208, 16, 22, 114, 209, 144, 64, 78, 210, 0, 57, 187, 11, 240, 27, 171, 12, 144, 99, 164, 13, 16, 26, 139, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 101, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 5, 4, 5, 4, 5, 3, 2, 3, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 170, 170, 170, 12, 0, 0, 0, 49, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 155, 128, 0, 224, 158, 128, 0, 224, 7, 129, 0, 16, 34, 129, 0, 224, 2, 0, 0, 0, 190, 1, 0, 0, 221, 1, 0, 0, 60, 0, 0, 0, 96, 23, 12, 155, 240, 218, 213, 155, 144, 174, 217, 156, 144, 181, 164, 157, 144, 144, 185, 158, 144, 151, 132, 159, 144, 113, 9, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 16, 22, 114, 209, 16, 7, 98, 210, 144, 28, 128, 211, 16, 210, 73, 212, 32, 180, 147, 212, 32, 114, 2, 213, 16, 56, 76, 213, 16, 180, 41, 214, 16, 26, 44, 215, 16, 150, 9, 216, 16, 112, 1, 217, 16, 120, 233, 217, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 248, 234, 207, 108, 170, 170, 170, 170, 170, 170, 170, 170, 61, 0, 0, 0, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 170, 8, 0, 0, 0, 136, 13, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 58, 129, 0, 224, 61, 129, 0, 224, 122, 129, 0, 224, 127, 129, 0, 16, 144, 129, 0, 224, 53, 0, 0, 0, 254, 135, 185, 158, 254, 142, 132, 159, 126, 70, 136, 160, 254, 130, 203, 160, 222, 241, 231, 173, 96, 100, 175, 200, 80, 101, 98, 202, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 112, 137, 144, 208, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 0, 25, 28, 37, 0, 10, 12, 38, 128, 53, 5, 39, 128, 38, 245, 39, 128, 23, 229, 40, 128, 8, 213, 41, 128, 249, 196, 42, 128, 234, 180, 43, 128, 219, 164, 44, 128, 204, 148, 45, 128, 189, 132, 46, 128, 174, 116, 47, 128, 159, 100, 48, 0, 203, 93, 49, 0, 188, 77, 50, 16, 187, 61, 51, 16, 150, 82, 52, 16, 157, 29, 53, 16, 120, 50, 54, 16, 127, 253, 54, 144, 148, 27, 56, 16, 67, 189, 58, 144, 88, 219, 59, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 53, 0, 0, 0, 3, 0, 3, 0, 4, 6, 2, 1, 2, 1, 2, 1, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 170, 170, 170, 16, 0, 0, 0, 162, 22, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 162, 22, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 4, 0, 0, 96, 210, 7, 0, 112, 166, 129, 0, 224, 223, 129, 0, 16, 238, 129, 0, 224, 3, 0, 0, 0, 223, 1, 0, 0, 225, 1, 0, 0, 238, 1, 0, 0, 86, 0, 0, 0, 112, 248, 56, 155, 224, 204, 213, 155, 240, 203, 197, 156, 96, 0, 183, 157, 112, 254, 137, 158, 224, 28, 160, 159, 240, 165, 96, 160, 96, 173, 126, 161, 112, 55, 92, 162, 96, 26, 76, 163, 240, 53, 108, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 144, 94, 110, 208, 16, 22, 114, 209, 240, 210, 76, 210, 144, 49, 62, 211, 16, 210, 73, 212, 112, 247, 29, 213, 240, 151, 41, 214, 144, 128, 235, 214, 16, 150, 9, 216, 240, 181, 51, 249, 224, 196, 217, 249, 112, 210, 28, 251, 240, 180, 185, 251, 112, 180, 252, 252, 240, 150, 153, 253, 240, 208, 229, 254, 112, 179, 130, 255, 240, 178, 197, 0, 112, 149, 98, 1, 112, 90, 156, 2, 112, 119, 66, 3, 240, 118, 133, 4, 240, 147, 43, 5, 112, 147, 110, 6, 240, 117, 11, 7, 240, 58, 69, 8, 240, 87, 235, 8, 112, 87, 46, 10, 240, 57, 203, 10, 112, 57, 14, 12, 240, 27, 171, 12, 240, 224, 228, 13, 240, 253, 138, 14, 112, 253, 205, 15, 112, 26, 116, 16, 112, 223, 173, 17, 112, 252, 83, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 112, 129, 188, 112, 87, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 180, 11, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 9, 130, 0, 224, 13, 130, 0, 224, 100, 130, 0, 224, 103, 130, 0, 16, 126, 130, 0, 224, 64, 0, 0, 0, 128, 57, 0, 161, 80, 11, 164, 181, 192, 153, 39, 21, 48, 206, 24, 22, 64, 205, 8, 23, 176, 1, 250, 23, 192, 0, 234, 24, 48, 53, 219, 25, 192, 133, 204, 26, 224, 146, 188, 27, 224, 131, 172, 28, 224, 116, 156, 29, 224, 101, 140, 30, 224, 86, 124, 31, 224, 71, 108, 32, 224, 56, 92, 33, 224, 41, 76, 34, 224, 26, 60, 35, 224, 11, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 112, 39, 5, 39, 112, 24, 245, 39, 128, 23, 229, 40, 0, 199, 0, 41, 96, 236, 212, 41, 96, 221, 196, 42, 96, 206, 180, 43, 96, 191, 164, 44, 96, 176, 148, 45, 96, 161, 132, 46, 96, 146, 116, 47, 96, 131, 100, 48, 224, 174, 93, 49, 224, 137, 114, 50, 224, 144, 61, 51, 224, 107, 82, 52, 224, 114, 29, 53, 224, 77, 50, 54, 224, 84, 253, 54, 96, 106, 27, 56, 224, 54, 221, 56, 96, 76, 251, 57, 224, 24, 189, 58, 96, 46, 219, 59, 96, 53, 166, 60, 96, 16, 187, 61, 96, 23, 134, 62, 96, 242, 154, 63, 96, 249, 101, 64, 224, 14, 132, 65, 96, 219, 69, 66, 224, 240, 99, 67, 96, 189, 37, 68, 224, 210, 67, 69, 96, 159, 5, 70, 224, 180, 35, 71, 224, 187, 238, 71, 224, 150, 3, 73, 224, 157, 206, 73, 224, 120, 227, 74, 224, 127, 174, 75, 112, 163, 204, 76, 240, 111, 142, 77, 64, 0, 0, 0, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 2, 3, 2, 1, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 3, 2, 4, 12, 0, 0, 0, 244, 46, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 146, 130, 0, 224, 211, 130, 0, 16, 228, 130, 0, 224, 64, 0, 0, 0, 128, 57, 0, 161, 80, 11, 164, 181, 192, 153, 39, 21, 48, 206, 24, 22, 64, 205, 8, 23, 176, 1, 250, 23, 192, 0, 234, 24, 48, 53, 219, 25, 192, 133, 204, 26, 224, 146, 188, 27, 224, 131, 172, 28, 224, 116, 156, 29, 224, 101, 140, 30, 224, 86, 124, 31, 224, 71, 108, 32, 224, 56, 92, 33, 224, 41, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 112, 39, 5, 39, 112, 24, 245, 39, 96, 236, 212, 41, 112, 235, 196, 42, 112, 220, 180, 43, 112, 205, 164, 44, 112, 190, 148, 45, 112, 175, 132, 46, 112, 160, 116, 47, 112, 145, 100, 48, 240, 188, 93, 49, 240, 151, 114, 50, 240, 158, 61, 51, 240, 121, 82, 52, 240, 128, 29, 53, 240, 91, 50, 54, 240, 98, 253, 54, 112, 120, 27, 56, 240, 68, 221, 56, 112, 90, 251, 57, 240, 38, 189, 58, 112, 60, 219, 59, 112, 67, 166, 60, 112, 30, 187, 61, 112, 37, 134, 62, 112, 0, 155, 63, 112, 7, 102, 64, 240, 28, 132, 65, 112, 233, 69, 66, 240, 254, 99, 67, 112, 203, 37, 68, 240, 224, 67, 69, 112, 173, 5, 70, 240, 194, 35, 71, 240, 201, 238, 71, 240, 164, 3, 73, 240, 171, 206, 73, 240, 134, 227, 74, 240, 141, 174, 75, 112, 163, 204, 76, 240, 111, 142, 77, 96, 29, 76, 84, 112, 78, 67, 88, 64, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 2, 1, 2, 1, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 3, 10, 0, 0, 0, 50, 43, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 246, 130, 0, 224, 55, 131, 0, 16, 72, 131, 0, 224, 74, 0, 0, 0, 32, 164, 25, 170, 96, 25, 164, 181, 208, 141, 4, 203, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 224, 56, 159, 207, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 240, 46, 141, 38, 96, 236, 212, 41, 80, 207, 196, 42, 96, 206, 180, 43, 80, 177, 164, 44, 96, 176, 148, 45, 208, 198, 194, 45, 64, 133, 132, 46, 80, 132, 116, 47, 64, 103, 100, 48, 208, 160, 93, 49, 0, 166, 114, 50, 16, 187, 61, 51, 16, 150, 82, 52, 16, 157, 29, 53, 16, 120, 50, 54, 16, 127, 253, 54, 144, 148, 27, 56, 16, 97, 221, 56, 144, 118, 251, 57, 16, 67, 189, 58, 144, 88, 219, 59, 144, 95, 166, 60, 144, 58, 187, 61, 144, 65, 134, 62, 144, 28, 155, 63, 144, 35, 102, 64, 16, 57, 132, 65, 144, 5, 70, 66, 16, 27, 100, 67, 144, 231, 37, 68, 16, 253, 67, 69, 144, 201, 5, 70, 16, 223, 35, 71, 16, 230, 238, 71, 16, 193, 3, 73, 16, 200, 206, 73, 16, 163, 227, 74, 16, 170, 174, 75, 144, 191, 204, 76, 16, 140, 142, 77, 144, 161, 172, 78, 16, 110, 110, 79, 144, 131, 140, 80, 144, 138, 87, 81, 144, 101, 108, 82, 128, 94, 55, 83, 96, 29, 76, 84, 2, 0, 0, 0, 255, 255, 255, 255, 8, 196, 182, 86, 170, 170, 170, 170, 75, 0, 0, 0, 5, 3, 6, 2, 1, 2, 1, 2, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 3, 4, 3, 4, 3, 4, 7, 6, 7, 6, 7, 6, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 8, 6, 170, 18, 0, 0, 0, 248, 31, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 224, 31, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 88, 131, 0, 224, 163, 131, 0, 224, 167, 131, 0, 16, 187, 131, 0, 224, 44, 0, 0, 0, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 32, 36, 114, 209, 80, 239, 99, 17, 224, 63, 85, 18, 208, 11, 77, 19, 224, 33, 53, 20, 208, 237, 44, 21, 112, 192, 19, 22, 208, 207, 12, 23, 128, 176, 243, 23, 128, 161, 227, 24, 128, 146, 211, 25, 128, 131, 195, 26, 0, 175, 188, 27, 0, 160, 172, 28, 0, 145, 156, 29, 0, 130, 140, 30, 0, 115, 124, 31, 0, 100, 108, 32, 0, 85, 92, 33, 0, 70, 76, 34, 0, 55, 60, 35, 0, 40, 44, 36, 0, 25, 28, 37, 0, 10, 12, 38, 128, 53, 5, 39, 96, 10, 245, 39, 80, 237, 228, 40, 96, 236, 212, 41, 80, 207, 196, 42, 96, 206, 180, 43, 80, 177, 164, 44, 96, 176, 148, 45, 80, 147, 132, 46, 96, 146, 116, 47, 80, 117, 100, 48, 224, 174, 93, 49, 208, 123, 114, 50, 16, 187, 61, 51, 16, 150, 82, 52, 4, 0, 0, 0, 255, 255, 255, 255, 36, 206, 182, 86, 255, 255, 255, 255, 24, 227, 195, 114, 46, 0, 0, 0, 3, 4, 1, 2, 1, 2, 1, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 170, 170, 12, 0, 0, 0, 220, 21, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 104, 27, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 32, 28, 0, 112, 4, 0, 0, 96, 206, 7, 0, 112, 213, 131, 0, 224, 2, 132, 0, 224, 7, 132, 0, 16, 20, 132, 0, 224, 36, 0, 0, 0, 96, 140, 30, 155, 240, 218, 213, 155, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 4, 0, 0, 0, 255, 255, 255, 255, 148, 159, 213, 84, 255, 255, 255, 255, 98, 115, 85, 124, 170, 170, 170, 170, 38, 0, 0, 0, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 8, 0, 0, 0, 236, 16, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 30, 14, 0, 0, 0, 0, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 44, 132, 0, 224, 81, 132, 0, 224, 87, 132, 0, 16, 98, 132, 0, 224, 52, 0, 0, 0, 204, 45, 89, 158, 144, 144, 185, 158, 144, 151, 132, 159, 112, 43, 0, 161, 76, 111, 115, 164, 224, 181, 176, 200, 80, 151, 198, 202, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 224, 203, 116, 208, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 0, 25, 28, 37, 0, 10, 12, 38, 128, 53, 5, 39, 128, 38, 245, 39, 128, 23, 229, 40, 128, 8, 213, 41, 128, 249, 196, 42, 128, 234, 180, 43, 128, 219, 164, 44, 128, 204, 148, 45, 128, 189, 132, 46, 128, 174, 116, 47, 128, 159, 100, 48, 0, 203, 93, 49, 0, 166, 114, 50, 0, 173, 61, 51, 0, 136, 82, 52, 0, 143, 29, 53, 16, 120, 50, 54, 16, 127, 253, 54, 144, 148, 27, 56, 144, 95, 166, 60, 144, 58, 187, 61, 52, 0, 0, 0, 1, 2, 1, 0, 3, 5, 2, 1, 2, 1, 2, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 14, 0, 0, 0, 52, 23, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 32, 28, 0, 112, 4, 0, 0, 96, 211, 7, 0, 112, 118, 132, 0, 224, 171, 132, 0, 16, 185, 132, 0, 224, 51, 0, 0, 0, 104, 52, 170, 150, 112, 135, 109, 200, 16, 75, 231, 204, 144, 23, 169, 205, 144, 233, 184, 205, 240, 57, 40, 8, 96, 62, 239, 8, 240, 120, 5, 10, 224, 113, 208, 10, 112, 79, 233, 11, 96, 72, 180, 12, 240, 107, 210, 13, 96, 42, 148, 14, 112, 252, 176, 15, 96, 12, 116, 16, 112, 222, 144, 17, 96, 238, 83, 18, 112, 192, 112, 19, 96, 185, 59, 20, 112, 185, 72, 21, 96, 178, 19, 22, 240, 213, 49, 23, 224, 206, 252, 23, 112, 148, 0, 25, 96, 95, 219, 25, 240, 175, 204, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 51, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 152, 18, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 210, 132, 0, 224, 7, 133, 0, 16, 21, 133, 0, 224, 66, 0, 0, 0, 128, 57, 0, 161, 80, 11, 164, 181, 192, 153, 39, 21, 48, 206, 24, 22, 64, 205, 8, 23, 176, 1, 250, 23, 192, 0, 234, 24, 48, 53, 219, 25, 192, 133, 204, 26, 224, 146, 188, 27, 224, 131, 172, 28, 224, 116, 156, 29, 224, 101, 140, 30, 224, 86, 124, 31, 224, 71, 108, 32, 224, 56, 92, 33, 224, 41, 76, 34, 224, 26, 60, 35, 224, 11, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 112, 39, 5, 39, 112, 24, 245, 39, 128, 23, 229, 40, 128, 191, 120, 41, 112, 250, 212, 41, 112, 235, 196, 42, 112, 220, 180, 43, 112, 205, 164, 44, 112, 190, 148, 45, 112, 175, 132, 46, 112, 160, 116, 47, 112, 145, 100, 48, 240, 188, 93, 49, 240, 151, 114, 50, 240, 158, 61, 51, 240, 121, 82, 52, 240, 128, 29, 53, 240, 91, 50, 54, 240, 98, 253, 54, 112, 120, 27, 56, 240, 68, 221, 56, 112, 90, 251, 57, 240, 38, 189, 58, 112, 60, 219, 59, 112, 67, 166, 60, 112, 30, 187, 61, 112, 37, 134, 62, 112, 0, 155, 63, 112, 7, 102, 64, 240, 28, 132, 65, 112, 233, 69, 66, 240, 254, 99, 67, 112, 203, 37, 68, 240, 224, 67, 69, 112, 173, 5, 70, 240, 194, 35, 71, 240, 201, 238, 71, 240, 164, 3, 73, 240, 171, 206, 73, 240, 134, 227, 74, 240, 141, 174, 75, 112, 163, 204, 76, 240, 111, 142, 77, 96, 29, 76, 84, 112, 20, 247, 86, 170, 170, 170, 170, 170, 170, 170, 170, 66, 0, 0, 0, 3, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 4, 3, 4, 3, 2, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 5, 3, 5, 170, 170, 14, 0, 0, 0, 96, 45, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 38, 133, 0, 224, 107, 133, 0, 16, 125, 133, 0, 224, 37, 0, 0, 0, 144, 113, 9, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 224, 158, 161, 208, 240, 253, 229, 209, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 240, 46, 141, 38, 160, 66, 245, 39, 96, 236, 212, 41, 80, 207, 196, 42, 96, 206, 180, 43, 80, 177, 164, 44, 96, 176, 148, 45, 80, 147, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 24, 176, 238, 106, 170, 170, 170, 170, 38, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 170, 170, 14, 0, 0, 0, 232, 20, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 32, 28, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 145, 133, 0, 224, 183, 133, 0, 224, 187, 133, 0, 16, 198, 133, 0, 224, 56, 0, 0, 0, 96, 23, 12, 155, 240, 218, 213, 155, 144, 174, 217, 156, 144, 181, 164, 157, 144, 144, 185, 158, 144, 151, 132, 159, 16, 26, 112, 162, 144, 91, 68, 163, 144, 113, 9, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 16, 22, 114, 209, 16, 69, 127, 209, 144, 27, 99, 211, 144, 35, 75, 212, 16, 195, 57, 213, 16, 180, 41, 214, 16, 26, 44, 215, 16, 150, 9, 216, 240, 39, 77, 19, 96, 208, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 47, 95, 162, 111, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 57, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 81, 15, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 224, 133, 0, 224, 25, 134, 0, 224, 31, 134, 0, 16, 47, 134, 0, 224, 51, 0, 0, 0, 80, 31, 79, 156, 152, 74, 133, 161, 240, 48, 241, 162, 96, 120, 102, 163, 112, 207, 172, 200, 208, 42, 89, 202, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 224, 61, 48, 208, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 0, 25, 28, 37, 0, 10, 12, 38, 128, 53, 5, 39, 128, 38, 245, 39, 128, 23, 229, 40, 128, 8, 213, 41, 128, 249, 196, 42, 128, 234, 180, 43, 128, 219, 164, 44, 128, 204, 148, 45, 128, 189, 132, 46, 128, 174, 116, 47, 128, 159, 100, 48, 0, 203, 93, 49, 0, 166, 114, 50, 0, 173, 61, 51, 0, 136, 82, 52, 16, 157, 29, 53, 16, 120, 50, 54, 16, 127, 253, 54, 144, 148, 27, 56, 144, 65, 134, 62, 144, 28, 155, 63, 2, 0, 0, 0, 255, 255, 255, 255, 68, 204, 182, 86, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 52, 0, 0, 0, 3, 4, 1, 5, 1, 7, 2, 1, 2, 1, 2, 7, 8, 7, 8, 7, 8, 7, 8, 7, 8, 7, 8, 7, 8, 7, 8, 7, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 2, 1, 2, 5, 6, 5, 18, 0, 0, 0, 188, 23, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 176, 19, 0, 0, 0, 0, 0, 0, 104, 22, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 32, 28, 0, 112, 4, 0, 0, 96, 212, 7, 0, 112, 65, 134, 0, 224, 117, 134, 0, 224, 123, 134, 0, 16, 137, 134, 0, 224, 64, 0, 0, 0, 220, 70, 245, 161, 80, 11, 164, 181, 192, 153, 39, 21, 48, 206, 24, 22, 64, 205, 8, 23, 176, 1, 250, 23, 192, 0, 234, 24, 48, 53, 219, 25, 192, 133, 204, 26, 224, 146, 188, 27, 224, 131, 172, 28, 224, 116, 156, 29, 224, 101, 140, 30, 224, 86, 124, 31, 224, 71, 108, 32, 224, 56, 92, 33, 224, 41, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 112, 39, 5, 39, 112, 24, 245, 39, 96, 236, 212, 41, 112, 235, 196, 42, 112, 220, 180, 43, 112, 205, 164, 44, 112, 190, 148, 45, 112, 175, 132, 46, 112, 160, 116, 47, 112, 145, 100, 48, 240, 188, 93, 49, 240, 151, 114, 50, 240, 158, 61, 51, 240, 121, 82, 52, 240, 128, 29, 53, 240, 91, 50, 54, 240, 98, 253, 54, 112, 120, 27, 56, 240, 68, 221, 56, 112, 90, 251, 57, 240, 38, 189, 58, 112, 60, 219, 59, 112, 67, 166, 60, 112, 30, 187, 61, 112, 37, 134, 62, 112, 0, 155, 63, 112, 7, 102, 64, 240, 28, 132, 65, 112, 233, 69, 66, 240, 254, 99, 67, 112, 203, 37, 68, 240, 224, 67, 69, 112, 173, 5, 70, 240, 194, 35, 71, 240, 201, 238, 71, 240, 164, 3, 73, 240, 171, 206, 73, 240, 134, 227, 74, 240, 141, 174, 75, 112, 163, 204, 76, 240, 111, 142, 77, 96, 29, 76, 84, 240, 237, 212, 91, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 64, 0, 0, 0, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 2, 1, 2, 1, 2, 1, 3, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 1, 3, 10, 0, 0, 0, 164, 41, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 167, 134, 0, 224, 235, 134, 0, 16, 252, 134, 0, 224, 2, 0, 0, 0, 242, 1, 0, 0, 82, 2, 0, 0, 82, 0, 0, 0, 208, 42, 168, 153, 96, 23, 12, 155, 240, 218, 213, 155, 144, 174, 217, 156, 144, 181, 164, 157, 144, 144, 185, 158, 144, 151, 132, 159, 0, 182, 154, 160, 0, 189, 101, 161, 96, 124, 125, 166, 16, 222, 118, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 0, 186, 132, 208, 112, 146, 149, 209, 96, 187, 138, 210, 112, 255, 98, 211, 144, 35, 75, 212, 16, 173, 94, 213, 16, 180, 41, 214, 16, 26, 44, 215, 16, 150, 9, 216, 144, 193, 2, 217, 16, 120, 233, 217, 0, 210, 84, 232, 128, 180, 241, 232, 128, 165, 225, 233, 128, 150, 209, 234, 0, 150, 20, 236, 0, 179, 186, 236, 0, 164, 170, 237, 0, 149, 154, 238, 0, 90, 212, 239, 0, 119, 122, 240, 0, 60, 180, 241, 0, 89, 90, 242, 0, 30, 148, 243, 0, 59, 58, 244, 128, 58, 125, 245, 0, 29, 26, 246, 128, 85, 164, 13, 0, 12, 139, 14, 128, 55, 132, 15, 128, 40, 116, 16, 128, 25, 100, 17, 128, 10, 84, 18, 0, 54, 77, 19, 128, 236, 51, 20, 128, 221, 35, 21, 128, 206, 19, 22, 128, 191, 3, 23, 128, 176, 243, 23, 128, 161, 227, 24, 128, 146, 211, 25, 128, 131, 195, 26, 0, 175, 188, 27, 0, 160, 172, 28, 0, 145, 156, 29, 0, 130, 140, 30, 0, 115, 124, 31, 0, 100, 108, 32, 0, 85, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 82, 0, 0, 0, 1, 2, 1, 2, 1, 2, 3, 4, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 10, 0, 0, 0, 176, 19, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 12, 135, 0, 224, 15, 135, 0, 224, 99, 135, 0, 16, 121, 135, 0, 224, 39, 0, 0, 0, 48, 163, 25, 170, 96, 25, 164, 181, 208, 231, 170, 202, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 112, 214, 189, 206, 208, 167, 39, 21, 64, 220, 24, 22, 80, 219, 8, 23, 192, 15, 250, 23, 208, 14, 234, 24, 64, 67, 219, 25, 208, 147, 204, 26, 240, 160, 188, 27, 240, 145, 172, 28, 240, 130, 156, 29, 240, 115, 140, 30, 240, 100, 124, 31, 240, 85, 108, 32, 240, 70, 92, 33, 240, 55, 76, 34, 240, 40, 60, 35, 240, 25, 44, 36, 240, 10, 28, 37, 240, 251, 11, 38, 112, 39, 5, 39, 112, 24, 245, 39, 80, 237, 228, 40, 96, 236, 212, 41, 80, 207, 196, 42, 96, 206, 180, 43, 80, 177, 164, 44, 96, 176, 148, 45, 80, 147, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 2, 0, 0, 0, 255, 255, 255, 255, 8, 195, 182, 86, 170, 170, 170, 170, 40, 0, 0, 0, 5, 3, 6, 2, 1, 2, 1, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 7, 6, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 16, 0, 0, 0, 248, 32, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 32, 28, 0, 0, 0, 0, 0, 0, 32, 28, 0, 0, 16, 14, 0, 0, 208, 32, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 0, 0, 0, 0, 48, 42, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 32, 28, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 143, 135, 0, 224, 183, 135, 0, 224, 187, 135, 0, 16, 198, 135, 0, 224, 3, 0, 0, 0, 194, 1, 0, 0, 237, 1, 0, 0, 245, 1, 0, 0, 36, 0, 0, 0, 0, 106, 23, 202, 0, 113, 226, 202, 0, 76, 247, 203, 0, 83, 194, 204, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 4, 0, 0, 0, 255, 255, 255, 255, 128, 234, 240, 36, 255, 255, 255, 255, 134, 6, 212, 113, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 38, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 0, 8, 0, 0, 0, 0, 0, 0, 250, 6, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 16, 14, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 226, 135, 0, 224, 230, 135, 0, 224, 11, 136, 0, 224, 19, 136, 0, 16, 30, 136, 0, 224, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 54, 0, 52, 136, 0, 224, 2, 0, 0, 0, 96, 115, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 57, 136, 0, 224, 2, 0, 0, 0, 156, 247, 126, 137, 176, 221, 230, 48, 170, 170, 170, 170, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 228, 67, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 96, 84, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 62, 136, 0, 224, 67, 136, 0, 16, 69, 136, 0, 224, 2, 0, 0, 0, 255, 255, 255, 255, 228, 169, 22, 115, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 28, 99, 0, 0, 0, 0, 0, 0, 112, 98, 0, 0, 0, 0, 0, 0, 3, 0, 80, 0, 66, 0, 54, 0, 81, 136, 0, 224, 87, 136, 0, 16, 89, 136, 0, 224, 2, 0, 0, 0, 255, 255, 255, 255, 164, 38, 85, 124, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 220, 90, 0, 0, 0, 0, 0, 0, 104, 91, 0, 0, 0, 0, 0, 0, 3, 0, 80, 0, 66, 0, 54, 0, 99, 136, 0, 224, 103, 136, 0, 16, 105, 136, 0, 224, 1, 0, 0, 0, 128, 98, 97, 218, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 115, 136, 0, 224, 119, 136, 0, 16, 121, 136, 0, 224, 1, 0, 0, 0, 132, 230, 100, 136, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 252, 51, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 131, 136, 0, 224, 135, 136, 0, 16, 137, 136, 0, 224, 1, 0, 0, 0, 152, 195, 47, 237, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 232, 68, 0, 0, 0, 0, 0, 0, 80, 70, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 147, 136, 0, 224, 151, 136, 0, 16, 153, 136, 0, 224, 5, 0, 0, 0, 152, 5, 127, 137, 64, 237, 5, 24, 48, 114, 219, 24, 224, 150, 3, 73, 208, 143, 206, 73, 170, 170, 170, 170, 170, 170, 170, 170, 5, 0, 0, 0, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 232, 53, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 163, 136, 0, 224, 171, 136, 0, 16, 174, 136, 0, 224, 1, 0, 0, 0, 128, 57, 204, 145, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 0, 52, 0, 0, 0, 0, 0, 0, 64, 56, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 186, 136, 0, 224, 191, 136, 0, 16, 193, 136, 0, 224, 54, 0, 0, 0, 96, 23, 12, 155, 240, 218, 213, 155, 144, 174, 217, 156, 144, 181, 164, 157, 144, 144, 185, 158, 144, 151, 132, 159, 144, 113, 9, 200, 16, 75, 231, 204, 144, 23, 169, 205, 16, 67, 162, 206, 16, 52, 146, 207, 16, 37, 130, 208, 16, 22, 114, 209, 144, 64, 78, 210, 144, 99, 164, 13, 16, 26, 139, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 54, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 170, 170, 4, 0, 0, 0, 16, 14, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 16, 14, 0, 112, 121, 3, 0, 96, 205, 7, 0, 112, 203, 136, 0, 224, 3, 137, 0, 16, 18, 137, 0, 224, 2, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 33, 137, 0, 224, 88, 0, 0, 0, 144, 58, 166, 158, 128, 7, 187, 159, 144, 28, 134, 160, 128, 233, 154, 161, 144, 12, 137, 203, 0, 24, 97, 210, 16, 117, 248, 250, 0, 88, 232, 251, 16, 87, 216, 252, 0, 58, 200, 253, 16, 57, 184, 254, 0, 28, 168, 255, 16, 27, 152, 0, 0, 254, 135, 1, 16, 253, 119, 2, 128, 26, 113, 3, 144, 25, 97, 4, 128, 252, 80, 5, 144, 251, 64, 6, 128, 222, 48, 7, 144, 53, 141, 7, 128, 192, 16, 9, 16, 177, 173, 9, 128, 162, 240, 10, 144, 161, 224, 11, 0, 191, 217, 12, 144, 131, 192, 13, 0, 161, 185, 14, 16, 160, 169, 15, 0, 131, 153, 16, 16, 130, 137, 17, 0, 101, 121, 18, 16, 100, 105, 19, 0, 71, 89, 20, 16, 70, 73, 21, 0, 41, 57, 22, 16, 40, 41, 23, 128, 69, 34, 24, 16, 10, 9, 25, 128, 39, 2, 26, 144, 38, 242, 26, 128, 9, 226, 27, 144, 8, 210, 28, 128, 235, 193, 29, 144, 234, 177, 30, 128, 205, 161, 31, 16, 29, 118, 32, 128, 175, 129, 33, 16, 255, 85, 34, 0, 204, 106, 35, 16, 225, 53, 36, 0, 174, 74, 37, 16, 195, 21, 38, 0, 144, 42, 39, 144, 223, 254, 39, 0, 114, 10, 41, 144, 193, 222, 41, 0, 84, 234, 42, 144, 163, 190, 43, 128, 112, 211, 44, 144, 133, 158, 45, 128, 82, 179, 46, 144, 103, 126, 47, 128, 52, 147, 48, 16, 132, 103, 49, 128, 22, 115, 50, 16, 102, 71, 51, 128, 248, 82, 52, 16, 72, 39, 53, 128, 218, 50, 54, 16, 42, 7, 55, 0, 247, 27, 56, 16, 12, 231, 56, 0, 217, 251, 57, 16, 238, 198, 58, 0, 187, 219, 59, 144, 10, 176, 60, 0, 157, 187, 61, 144, 236, 143, 62, 0, 127, 155, 63, 144, 206, 111, 64, 128, 155, 132, 65, 144, 176, 79, 66, 128, 125, 100, 67, 144, 146, 47, 68, 128, 95, 68, 69, 16, 197, 243, 69, 0, 124, 45, 71, 88, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 4, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 144, 157, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 38, 137, 0, 224, 127, 137, 0, 16, 150, 137, 0, 224, 88, 0, 0, 0, 160, 72, 166, 158, 144, 21, 187, 159, 160, 42, 134, 160, 144, 247, 154, 161, 160, 26, 137, 203, 16, 38, 97, 210, 32, 131, 248, 250, 16, 102, 232, 251, 32, 101, 216, 252, 16, 72, 200, 253, 32, 71, 184, 254, 16, 42, 168, 255, 32, 41, 152, 0, 16, 12, 136, 1, 32, 11, 120, 2, 144, 40, 113, 3, 160, 39, 97, 4, 144, 10, 81, 5, 160, 9, 65, 6, 144, 236, 48, 7, 160, 67, 141, 7, 144, 206, 16, 9, 32, 191, 173, 9, 144, 176, 240, 10, 160, 175, 224, 11, 16, 205, 217, 12, 160, 145, 192, 13, 16, 175, 185, 14, 32, 174, 169, 15, 16, 145, 153, 16, 32, 144, 137, 17, 16, 115, 121, 18, 32, 114, 105, 19, 16, 85, 89, 20, 32, 84, 73, 21, 16, 55, 57, 22, 32, 54, 41, 23, 144, 83, 34, 24, 32, 24, 9, 25, 144, 53, 2, 26, 160, 52, 242, 26, 144, 23, 226, 27, 160, 22, 210, 28, 144, 249, 193, 29, 160, 248, 177, 30, 144, 219, 161, 31, 32, 43, 118, 32, 144, 189, 129, 33, 32, 13, 86, 34, 16, 218, 106, 35, 32, 239, 53, 36, 16, 188, 74, 37, 32, 209, 21, 38, 16, 158, 42, 39, 160, 237, 254, 39, 16, 128, 10, 41, 160, 207, 222, 41, 16, 98, 234, 42, 160, 177, 190, 43, 144, 126, 211, 44, 160, 147, 158, 45, 144, 96, 179, 46, 160, 117, 126, 47, 144, 66, 147, 48, 32, 146, 103, 49, 144, 36, 115, 50, 32, 116, 71, 51, 144, 6, 83, 52, 32, 86, 39, 53, 144, 232, 50, 54, 32, 56, 7, 55, 16, 5, 28, 56, 32, 26, 231, 56, 16, 231, 251, 57, 32, 252, 198, 58, 16, 201, 219, 59, 160, 24, 176, 60, 16, 171, 187, 61, 160, 250, 143, 62, 16, 141, 155, 63, 160, 220, 111, 64, 144, 169, 132, 65, 160, 190, 79, 66, 144, 139, 100, 67, 160, 160, 47, 68, 144, 109, 68, 69, 32, 211, 243, 69, 16, 138, 45, 71, 170, 170, 170, 170, 88, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 4, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 128, 143, 255, 127, 1, 0, 0, 96, 216, 7, 0, 112, 165, 137, 0, 224, 255, 137, 0, 16, 22, 138, 0, 224, 2, 0, 0, 0, 22, 2, 0, 0, 39, 2, 0, 0, 8, 0, 0, 0, 0, 252, 5, 145, 56, 4, 98, 218, 176, 39, 159, 76, 224, 43, 151, 77, 96, 226, 125, 78, 160, 139, 253, 78, 224, 13, 119, 79, 224, 254, 102, 80, 2, 0, 0, 0, 255, 255, 255, 255, 0, 201, 61, 110, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 9, 0, 0, 0, 2, 1, 3, 4, 3, 4, 6, 5, 6, 170, 170, 170, 14, 0, 0, 0, 128, 176, 0, 0, 0, 0, 0, 0, 72, 94, 255, 255, 0, 0, 0, 0, 0, 95, 255, 255, 0, 0, 0, 0, 80, 101, 255, 255, 0, 0, 0, 0, 80, 101, 255, 255, 16, 14, 0, 0, 208, 182, 0, 0, 0, 0, 0, 0, 208, 182, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 208, 182, 0, 112, 100, 0, 0, 96, 221, 7, 0, 112, 37, 138, 0, 224, 40, 138, 0, 224, 49, 138, 0, 224, 55, 138, 0, 16, 59, 138, 0, 224, 5, 0, 0, 0, 230, 0, 0, 0, 233, 0, 0, 0, 29, 2, 0, 0, 30, 2, 0, 0, 40, 2, 0, 0, 97, 0, 0, 0, 232, 178, 180, 176, 88, 135, 81, 177, 104, 229, 120, 178, 96, 229, 67, 179, 104, 199, 88, 180, 96, 199, 35, 181, 104, 169, 56, 182, 96, 169, 3, 183, 104, 139, 24, 184, 224, 197, 236, 184, 104, 109, 248, 185, 224, 167, 204, 186, 104, 79, 216, 187, 224, 232, 227, 188, 232, 246, 174, 189, 224, 202, 195, 190, 232, 216, 142, 191, 224, 172, 163, 192, 232, 186, 110, 193, 224, 142, 131, 194, 232, 156, 78, 195, 224, 112, 99, 196, 232, 126, 46, 197, 96, 141, 76, 198, 232, 96, 14, 199, 96, 111, 44, 200, 104, 125, 247, 200, 64, 154, 218, 210, 224, 253, 24, 9, 224, 165, 172, 9, 96, 165, 239, 10, 224, 252, 158, 11, 224, 193, 216, 12, 224, 222, 126, 13, 224, 163, 184, 14, 224, 192, 94, 15, 224, 133, 152, 16, 224, 162, 62, 17, 224, 103, 120, 18, 224, 132, 30, 19, 224, 73, 88, 20, 224, 102, 254, 20, 224, 43, 56, 22, 96, 131, 231, 22, 96, 72, 33, 24, 96, 101, 199, 24, 96, 42, 1, 26, 96, 71, 167, 26, 96, 12, 225, 27, 96, 41, 135, 28, 96, 238, 192, 29, 96, 11, 103, 30, 96, 208, 160, 31, 96, 237, 70, 32, 96, 178, 128, 33, 224, 9, 48, 34, 224, 206, 105, 35, 224, 235, 15, 36, 96, 1, 46, 37, 224, 66, 2, 38, 96, 227, 13, 39, 224, 36, 226, 39, 96, 197, 237, 40, 224, 6, 194, 41, 96, 167, 205, 42, 96, 35, 171, 43, 96, 137, 173, 44, 96, 5, 139, 45, 96, 107, 141, 46, 96, 231, 106, 47, 96, 77, 109, 48, 96, 201, 74, 49, 224, 105, 86, 50, 96, 171, 42, 51, 224, 75, 54, 52, 96, 141, 10, 53, 224, 45, 22, 54, 224, 169, 243, 54, 224, 15, 246, 55, 224, 139, 211, 56, 224, 241, 213, 57, 224, 109, 179, 58, 96, 14, 191, 59, 224, 79, 147, 60, 96, 240, 158, 61, 224, 49, 115, 62, 96, 210, 126, 63, 96, 78, 92, 64, 96, 180, 94, 65, 96, 48, 60, 66, 96, 150, 62, 67, 96, 18, 28, 68, 96, 120, 30, 69, 96, 244, 251, 69, 96, 90, 254, 70, 224, 133, 247, 71, 96, 60, 222, 72, 2, 0, 0, 0, 255, 255, 255, 255, 168, 76, 183, 65, 170, 170, 170, 170, 98, 0, 0, 0, 1, 3, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 4, 5, 170, 170, 12, 0, 0, 0, 216, 163, 0, 0, 0, 0, 0, 0, 184, 161, 0, 0, 0, 0, 0, 0, 184, 161, 0, 0, 8, 7, 0, 0, 184, 161, 0, 0, 16, 14, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 192, 168, 0, 112, 31, 0, 0, 96, 217, 7, 0, 112, 87, 138, 0, 224, 93, 138, 0, 224, 191, 138, 0, 224, 195, 138, 0, 16, 221, 138, 0, 224, 3, 0, 0, 0, 96, 54, 67, 204, 240, 108, 43, 210, 128, 215, 158, 84, 4, 0, 0, 0, 255, 255, 255, 255, 40, 82, 182, 86, 255, 255, 255, 255, 144, 164, 237, 114, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 5, 0, 0, 0, 2, 3, 1, 3, 4, 170, 170, 170, 10, 0, 0, 0, 216, 145, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 240, 137, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 247, 138, 0, 224, 251, 138, 0, 224, 3, 139, 0, 16, 6, 139, 0, 224, 2, 0, 0, 0, 31, 2, 0, 0, 42, 2, 0, 0, 70, 0, 0, 0, 188, 150, 218, 210, 224, 253, 24, 9, 224, 165, 172, 9, 96, 165, 239, 10, 224, 252, 158, 11, 224, 193, 216, 12, 224, 222, 126, 13, 224, 163, 184, 14, 224, 192, 94, 15, 224, 133, 152, 16, 224, 162, 62, 17, 224, 103, 120, 18, 224, 132, 30, 19, 224, 73, 88, 20, 224, 102, 254, 20, 224, 43, 56, 22, 96, 131, 231, 22, 96, 72, 33, 24, 96, 101, 199, 24, 96, 42, 1, 26, 96, 71, 167, 26, 96, 12, 225, 27, 96, 41, 135, 28, 96, 238, 192, 29, 96, 11, 103, 30, 96, 208, 160, 31, 96, 237, 70, 32, 96, 178, 128, 33, 224, 9, 48, 34, 224, 206, 105, 35, 224, 235, 15, 36, 96, 1, 46, 37, 224, 66, 2, 38, 96, 227, 13, 39, 224, 36, 226, 39, 96, 197, 237, 40, 224, 6, 194, 41, 96, 167, 205, 42, 96, 35, 171, 43, 96, 137, 173, 44, 96, 5, 139, 45, 96, 107, 141, 46, 96, 231, 106, 47, 96, 77, 109, 48, 96, 201, 74, 49, 224, 105, 86, 50, 96, 171, 42, 51, 224, 75, 54, 52, 96, 141, 10, 53, 224, 45, 22, 54, 224, 169, 243, 54, 224, 15, 246, 55, 224, 139, 211, 56, 224, 241, 213, 57, 224, 109, 179, 58, 96, 14, 191, 59, 224, 79, 147, 60, 96, 240, 158, 61, 224, 49, 115, 62, 96, 210, 126, 63, 96, 78, 92, 64, 96, 180, 94, 65, 96, 48, 60, 66, 96, 150, 62, 67, 96, 18, 28, 68, 96, 120, 30, 69, 96, 244, 251, 69, 96, 90, 254, 70, 224, 133, 247, 71, 96, 60, 222, 72, 2, 0, 0, 0, 255, 255, 255, 255, 132, 68, 183, 65, 170, 170, 170, 170, 170, 170, 170, 170, 71, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 170, 8, 0, 0, 0, 252, 171, 0, 0, 0, 0, 0, 0, 68, 172, 0, 0, 0, 0, 0, 0, 76, 179, 0, 0, 0, 0, 0, 0, 76, 179, 0, 0, 16, 14, 0, 0, 8, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 76, 179, 0, 112, 116, 18, 0, 96, 217, 7, 0, 112, 24, 139, 0, 224, 27, 139, 0, 224, 98, 139, 0, 224, 103, 139, 0, 16, 122, 139, 0, 224, 3, 0, 0, 0, 43, 2, 0, 0, 78, 2, 0, 0, 81, 2, 0, 0, 4, 0, 0, 0, 224, 163, 17, 152, 240, 249, 57, 160, 96, 10, 234, 201, 240, 14, 17, 210, 4, 0, 0, 0, 255, 255, 255, 255, 52, 191, 225, 20, 255, 255, 255, 255, 180, 38, 54, 126, 170, 170, 170, 170, 6, 0, 0, 0, 3, 2, 1, 2, 1, 2, 170, 170, 8, 0, 0, 0, 204, 60, 255, 255, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 76, 142, 0, 0, 0, 0, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 144, 139, 0, 224, 148, 139, 0, 224, 153, 139, 0, 224, 159, 139, 0, 16, 162, 139, 0, 224, 2, 0, 0, 0, 138, 1, 0, 0, 44, 2, 0, 0, 102, 0, 0, 0, 136, 64, 199, 185, 64, 60, 209, 253, 176, 250, 146, 254, 192, 205, 204, 255, 176, 220, 114, 0, 192, 80, 117, 1, 176, 73, 64, 2, 192, 50, 85, 3, 176, 43, 32, 4, 64, 79, 62, 5, 176, 13, 0, 6, 64, 188, 11, 7, 176, 239, 223, 7, 64, 19, 254, 8, 176, 209, 191, 9, 64, 245, 221, 10, 48, 238, 168, 11, 64, 215, 189, 12, 48, 208, 136, 13, 64, 185, 157, 14, 48, 178, 104, 15, 192, 213, 134, 16, 48, 148, 72, 17, 192, 183, 102, 18, 48, 118, 40, 19, 192, 153, 70, 20, 176, 146, 17, 21, 192, 123, 38, 22, 176, 116, 241, 22, 192, 93, 6, 24, 176, 86, 209, 24, 192, 63, 230, 25, 176, 56, 177, 26, 64, 92, 207, 27, 176, 26, 145, 28, 64, 62, 175, 29, 176, 252, 112, 30, 64, 32, 143, 31, 48, 3, 127, 32, 64, 2, 111, 33, 48, 251, 57, 34, 64, 228, 78, 35, 48, 221, 25, 36, 192, 0, 56, 37, 48, 191, 249, 37, 192, 248, 242, 38, 48, 161, 217, 39, 192, 196, 247, 40, 176, 189, 194, 41, 192, 166, 215, 42, 176, 159, 162, 43, 192, 136, 183, 44, 176, 129, 130, 45, 192, 106, 151, 46, 176, 99, 98, 47, 64, 135, 128, 48, 176, 69, 66, 49, 64, 105, 96, 50, 48, 215, 61, 51, 64, 75, 64, 52, 48, 68, 11, 53, 64, 184, 13, 54, 176, 213, 6, 55, 64, 15, 0, 56, 48, 8, 203, 56, 192, 43, 233, 57, 48, 234, 170, 58, 192, 13, 201, 59, 48, 204, 138, 60, 192, 239, 168, 61, 48, 174, 106, 62, 192, 209, 136, 63, 176, 202, 83, 64, 192, 179, 104, 65, 176, 172, 51, 66, 192, 149, 72, 67, 176, 142, 19, 68, 64, 178, 49, 69, 176, 112, 243, 69, 64, 148, 17, 71, 48, 2, 239, 71, 64, 118, 241, 72, 48, 111, 188, 73, 64, 88, 209, 74, 176, 0, 184, 75, 64, 58, 177, 76, 48, 7, 198, 77, 192, 130, 80, 78, 176, 174, 156, 79, 192, 217, 66, 80, 176, 144, 124, 81, 64, 246, 43, 82, 176, 114, 92, 83, 64, 216, 11, 84, 48, 230, 55, 87, 192, 236, 175, 87, 48, 200, 23, 89, 192, 206, 143, 89, 48, 170, 247, 90, 192, 176, 111, 91, 176, 103, 169, 92, 192, 124, 116, 93, 170, 170, 170, 170, 170, 170, 170, 170, 102, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 170, 170, 10, 0, 0, 0, 120, 153, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 74, 0, 48, 0, 66, 0, 54, 0, 160, 171, 255, 127, 127, 3, 0, 96, 228, 7, 0, 112, 179, 139, 0, 224, 182, 139, 0, 224, 31, 140, 0, 16, 58, 140, 0, 224, 21, 0, 0, 0, 180, 194, 245, 146, 208, 247, 210, 25, 192, 218, 194, 26, 208, 102, 218, 27, 192, 188, 162, 28, 80, 246, 155, 29, 192, 158, 130, 30, 80, 216, 123, 31, 64, 187, 107, 32, 80, 186, 91, 33, 64, 157, 75, 34, 80, 156, 59, 35, 64, 127, 43, 36, 80, 126, 27, 37, 64, 97, 11, 38, 80, 96, 251, 38, 64, 67, 235, 39, 208, 124, 228, 40, 64, 81, 129, 41, 208, 72, 233, 42, 64, 51, 97, 43, 170, 170, 170, 170, 21, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 170, 170, 170, 6, 0, 0, 0, 204, 157, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 80, 140, 0, 224, 103, 140, 0, 16, 110, 140, 0, 224, 2, 0, 0, 0, 192, 4, 86, 18, 176, 57, 5, 47, 2, 0, 0, 0, 255, 255, 255, 255, 100, 85, 55, 126, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 3, 0, 0, 0, 1, 2, 3, 170, 8, 0, 0, 0, 156, 95, 255, 255, 0, 0, 0, 0, 64, 87, 255, 255, 0, 0, 0, 0, 80, 101, 255, 255, 0, 0, 0, 0, 208, 182, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 122, 140, 0, 224, 125, 140, 0, 224, 131, 140, 0, 16, 133, 140, 0, 224, 1, 0, 0, 0, 176, 153, 253, 78, 2, 0, 0, 0, 255, 255, 255, 255, 136, 85, 55, 126, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 120, 95, 255, 255, 0, 0, 0, 0, 80, 101, 255, 255, 0, 0, 0, 0, 208, 182, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 149, 140, 0, 224, 151, 140, 0, 224, 155, 140, 0, 16, 157, 140, 0, 224, 26, 0, 0, 0, 192, 177, 19, 154, 224, 23, 59, 54, 96, 250, 215, 54, 96, 52, 36, 56, 96, 220, 183, 56, 224, 44, 17, 75, 96, 15, 174, 75, 96, 234, 194, 76, 224, 65, 114, 77, 96, 204, 162, 78, 224, 196, 26, 79, 96, 174, 130, 80, 224, 166, 250, 80, 224, 202, 107, 82, 208, 122, 218, 82, 96, 231, 84, 84, 224, 106, 186, 84, 96, 201, 52, 86, 224, 76, 154, 86, 224, 229, 29, 88, 224, 46, 122, 88, 224, 199, 253, 89, 224, 16, 90, 90, 224, 169, 221, 91, 224, 242, 57, 92, 96, 198, 198, 93, 170, 170, 170, 170, 26, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 170, 170, 6, 0, 0, 0, 192, 167, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 192, 168, 0, 112, 49, 8, 0, 96, 228, 7, 0, 112, 171, 140, 0, 224, 199, 140, 0, 16, 207, 140, 0, 224, 2, 0, 0, 0, 255, 255, 255, 255, 252, 12, 54, 126, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 4, 168, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 3, 0, 80, 0, 66, 0, 54, 0, 224, 140, 0, 224, 227, 140, 0, 16, 229, 140, 0, 224, 4, 0, 0, 0, 128, 76, 164, 182, 80, 196, 24, 30, 224, 10, 23, 43, 80, 244, 113, 43, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 4, 0, 0, 0, 3, 1, 2, 1, 8, 0, 0, 0, 0, 172, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 239, 140, 0, 224, 247, 140, 0, 16, 249, 140, 0, 224, 1, 0, 0, 0, 4, 72, 80, 148, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 124, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 7, 141, 0, 224, 11, 141, 0, 16, 13, 141, 0, 224, 2, 0, 0, 0, 52, 2, 0, 0, 86, 2, 0, 0, 1, 0, 0, 0, 140, 51, 79, 148, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 244, 149, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 4, 0, 74, 0, 48, 0, 66, 0, 54, 0, 170, 170, 23, 141, 0, 224, 26, 141, 0, 224, 31, 141, 0, 16, 33, 141, 0, 224, 2, 0, 0, 0, 53, 2, 0, 0, 73, 2, 0, 0, 18, 0, 0, 0, 224, 149, 55, 203, 240, 137, 46, 208, 0, 190, 55, 236, 240, 248, 54, 239, 0, 0, 155, 251, 140, 39, 63, 254, 0, 30, 1, 255, 240, 88, 93, 255, 0, 44, 151, 0, 112, 117, 70, 1, 0, 14, 119, 2, 112, 87, 38, 3, 0, 151, 112, 7, 240, 209, 204, 7, 0, 145, 8, 12, 44, 135, 124, 12, 128, 148, 191, 13, 112, 163, 101, 14, 4, 0, 0, 0, 255, 255, 255, 255, 204, 197, 225, 20, 255, 255, 255, 255, 76, 45, 54, 126, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 20, 0, 0, 0, 2, 3, 1, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 10, 0, 0, 0, 52, 54, 255, 255, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 180, 135, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 16, 14, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 45, 141, 0, 224, 48, 141, 0, 224, 67, 141, 0, 224, 75, 141, 0, 16, 81, 141, 0, 224, 3, 0, 0, 0, 54, 2, 0, 0, 55, 2, 0, 0, 109, 2, 0, 0, 5, 0, 0, 0, 72, 67, 5, 187, 88, 113, 33, 187, 200, 61, 137, 203, 56, 73, 97, 210, 72, 115, 141, 213, 2, 0, 0, 0, 255, 255, 255, 255, 190, 112, 224, 116, 170, 170, 170, 170, 170, 170, 170, 170, 6, 0, 0, 0, 1, 2, 1, 2, 1, 3, 170, 170, 8, 0, 0, 0, 2, 108, 255, 255, 0, 0, 0, 0, 88, 108, 255, 255, 0, 0, 0, 0, 88, 108, 255, 255, 16, 14, 0, 0, 96, 115, 255, 255, 0, 0, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 100, 141, 0, 224, 104, 141, 0, 224, 110, 141, 0, 224, 115, 141, 0, 16, 118, 141, 0, 224, 2, 0, 0, 0, 0, 242, 85, 18, 160, 43, 5, 47, 2, 0, 0, 0, 255, 255, 255, 255, 128, 72, 55, 126, 170, 170, 170, 170, 170, 170, 170, 170, 3, 0, 0, 0, 1, 2, 3, 170, 8, 0, 0, 0, 128, 108, 255, 255, 0, 0, 0, 0, 0, 106, 255, 255, 0, 0, 0, 0, 96, 115, 255, 255, 0, 0, 0, 0, 224, 196, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 135, 141, 0, 224, 138, 141, 0, 224, 143, 141, 0, 16, 145, 141, 0, 224, 7, 0, 0, 0, 208, 149, 17, 152, 240, 249, 57, 160, 208, 53, 237, 193, 96, 10, 234, 201, 240, 14, 17, 210, 80, 27, 134, 255, 64, 103, 139, 54, 4, 0, 0, 0, 255, 255, 255, 255, 180, 180, 225, 20, 255, 255, 255, 255, 52, 28, 54, 126, 170, 170, 170, 170, 9, 0, 0, 0, 3, 4, 1, 4, 2, 1, 4, 5, 4, 170, 170, 170, 12, 0, 0, 0, 76, 71, 255, 255, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 204, 152, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 161, 141, 0, 224, 169, 141, 0, 224, 175, 141, 0, 16, 179, 141, 0, 224, 2, 0, 0, 0, 19, 2, 0, 0, 58, 2, 0, 0, 5, 0, 0, 0, 208, 53, 237, 193, 96, 10, 234, 201, 240, 129, 70, 207, 80, 27, 134, 255, 64, 14, 118, 44, 2, 0, 0, 0, 255, 255, 255, 255, 32, 24, 54, 126, 6, 0, 0, 0, 4, 3, 2, 4, 1, 5, 170, 170, 12, 0, 0, 0, 224, 156, 0, 0, 0, 0, 0, 0, 64, 87, 255, 255, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 199, 141, 0, 224, 202, 141, 0, 224, 208, 141, 0, 224, 211, 141, 0, 16, 214, 141, 0, 224, 6, 0, 0, 0, 208, 149, 17, 152, 240, 249, 57, 160, 208, 53, 237, 193, 96, 10, 234, 201, 112, 71, 61, 207, 80, 27, 134, 255, 2, 0, 0, 0, 255, 255, 255, 255, 128, 20, 54, 126, 170, 170, 170, 170, 170, 170, 170, 170, 7, 0, 0, 0, 3, 1, 3, 2, 1, 3, 4, 170, 10, 0, 0, 0, 128, 160, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 235, 141, 0, 224, 242, 141, 0, 224, 247, 141, 0, 16, 250, 141, 0, 224, 1, 0, 0, 0, 72, 76, 80, 148, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 56, 125, 255, 255, 0, 0, 0, 0, 104, 122, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 12, 142, 0, 224, 15, 142, 0, 16, 17, 142, 0, 224, 4, 0, 0, 0, 4, 43, 231, 163, 200, 233, 144, 204, 240, 39, 67, 210, 232, 168, 33, 17, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 4, 0, 0, 0, 2, 1, 2, 3, 8, 0, 0, 0, 124, 156, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 184, 161, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 27, 142, 0, 224, 35, 142, 0, 16, 37, 142, 0, 224, 2, 0, 0, 0, 96, 53, 67, 220, 56, 202, 116, 16, 2, 0, 0, 0, 255, 255, 255, 255, 76, 84, 55, 126, 170, 170, 170, 170, 170, 170, 170, 170, 3, 0, 0, 0, 2, 1, 3, 170, 8, 0, 0, 0, 180, 96, 255, 255, 0, 0, 0, 0, 72, 94, 255, 255, 0, 0, 0, 0, 160, 96, 255, 255, 0, 0, 0, 0, 80, 101, 255, 255, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 51, 142, 0, 224, 54, 142, 0, 224, 59, 142, 0, 16, 61, 142, 0, 224, 7, 0, 0, 0, 128, 248, 65, 220, 104, 202, 15, 9, 104, 231, 181, 9, 104, 230, 15, 86, 240, 175, 152, 93, 240, 160, 136, 94, 240, 145, 120, 95, 2, 0, 0, 0, 255, 255, 255, 255, 136, 23, 54, 126, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 8, 0, 0, 0, 3, 4, 5, 4, 1, 2, 1, 2, 12, 0, 0, 0, 120, 157, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 16, 14, 0, 0, 128, 157, 0, 0, 0, 0, 0, 0, 184, 161, 0, 0, 0, 0, 0, 0, 184, 161, 0, 0, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 154, 0, 112, 106, 0, 0, 96, 229, 7, 0, 112, 77, 142, 0, 224, 85, 142, 0, 224, 91, 142, 0, 16, 94, 142, 0, 224, 7, 0, 0, 0, 116, 196, 245, 146, 80, 186, 230, 14, 192, 187, 86, 15, 80, 156, 198, 16, 64, 239, 55, 17, 240, 75, 160, 50, 112, 68, 24, 51, 170, 170, 170, 170, 7, 0, 0, 0, 1, 2, 1, 2, 1, 2, 1, 170, 6, 0, 0, 0, 12, 156, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 16, 14, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 118, 142, 0, 224, 127, 142, 0, 16, 130, 142, 0, 224, 4, 0, 0, 0, 61, 2, 0, 0, 66, 2, 0, 0, 74, 2, 0, 0, 115, 2, 0, 0, 1, 0, 0, 0, 8, 251, 5, 145, 2, 0, 0, 0, 255, 255, 255, 255, 8, 200, 61, 110, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 120, 177, 0, 0, 0, 0, 0, 0, 248, 95, 255, 255, 0, 0, 0, 0, 80, 101, 255, 255, 0, 0, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 142, 142, 0, 224, 147, 142, 0, 224, 149, 142, 0, 224, 155, 142, 0, 16, 157, 142, 0, 224, 4, 0, 0, 0, 255, 255, 255, 255, 108, 207, 225, 20, 255, 255, 255, 255, 236, 54, 54, 126, 170, 170, 170, 170, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 148, 44, 255, 255, 0, 0, 0, 0, 20, 126, 0, 0, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 3, 0, 80, 0, 66, 0, 54, 0, 172, 142, 0, 224, 179, 142, 0, 16, 181, 142, 0, 224, 1, 0, 0, 0, 8, 66, 68, 53, 2, 0, 0, 0, 255, 255, 255, 255, 244, 46, 55, 126, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 12, 134, 255, 255, 0, 0, 0, 0, 120, 136, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 193, 142, 0, 224, 195, 142, 0, 224, 199, 142, 0, 16, 201, 142, 0, 224, 2, 0, 0, 0, 69, 2, 0, 0, 70, 2, 0, 0, 5, 0, 0, 0, 208, 149, 17, 152, 240, 249, 57, 160, 208, 53, 237, 193, 96, 10, 234, 201, 240, 14, 17, 210, 4, 0, 0, 0, 255, 255, 255, 255, 44, 185, 225, 20, 255, 255, 255, 255, 172, 32, 54, 126, 170, 170, 170, 170, 170, 170, 170, 170, 7, 0, 0, 0, 3, 4, 1, 4, 2, 1, 4, 170, 10, 0, 0, 0, 212, 66, 255, 255, 0, 0, 0, 0, 144, 126, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 84, 148, 0, 0, 0, 0, 0, 0, 176, 154, 0, 0, 0, 0, 0, 0, 5, 0, 74, 0, 48, 0, 80, 0, 66, 0, 54, 0, 215, 142, 0, 224, 218, 142, 0, 224, 224, 142, 0, 224, 231, 142, 0, 16, 234, 142, 0, 224, 4, 0, 0, 0, 255, 255, 255, 255, 8, 90, 182, 86, 255, 255, 255, 255, 144, 164, 237, 114, 170, 170, 170, 170, 2, 0, 0, 0, 1, 2, 170, 170, 6, 0, 0, 0, 248, 137, 0, 0, 0, 0, 0, 0, 240, 137, 0, 0, 0, 0, 0, 0, 160, 140, 0, 0, 0, 0, 0, 0, 3, 0, 80, 0, 66, 0, 54, 0, 253, 142, 0, 224, 3, 143, 0, 16, 5, 143, 0, 224, 26, 0, 0, 0, 40, 27, 172, 16, 24, 181, 63, 17, 32, 129, 121, 18, 24, 151, 31, 19, 32, 99, 89, 20, 24, 121, 255, 20, 32, 69, 57, 22, 152, 149, 232, 22, 160, 97, 34, 24, 152, 119, 200, 24, 160, 67, 2, 26, 152, 89, 168, 26, 160, 37, 226, 27, 152, 59, 136, 28, 160, 7, 194, 29, 152, 29, 104, 30, 160, 233, 161, 31, 152, 255, 71, 32, 160, 203, 129, 33, 24, 28, 49, 34, 32, 232, 106, 35, 24, 254, 16, 36, 32, 202, 74, 37, 24, 224, 240, 37, 32, 172, 42, 39, 24, 194, 208, 39, 2, 0, 0, 0, 255, 255, 255, 255, 200, 74, 55, 126, 27, 0, 0, 0, 1, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 8, 0, 0, 0, 56, 106, 255, 255, 0, 0, 0, 0, 88, 108, 255, 255, 0, 0, 0, 0, 96, 115, 255, 255, 0, 0, 0, 0, 96, 115, 255, 255, 8, 7, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 17, 143, 0, 224, 44, 143, 0, 224, 47, 143, 0, 16, 55, 143, 0, 224, 1, 0, 0, 0, 184, 85, 80, 148, 170, 170, 170, 170, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 200, 115, 255, 255, 0, 0, 0, 0, 96, 115, 255, 255, 0, 0, 0, 0, 3, 0, 48, 0, 66, 0, 54, 0, 71, 143, 0, 224, 75, 143, 0, 16, 77, 143, 0, 224, 2, 0, 0, 0, 255, 255, 255, 255, 204, 18, 54, 126, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 52, 162, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 3, 0, 80, 0, 66, 0, 54, 0, 87, 143, 0, 224, 91, 143, 0, 16, 93, 143, 0, 224, 9, 0, 0, 0, 144, 66, 115, 201, 208, 71, 251, 55, 208, 125, 211, 56, 80, 8, 4, 58, 64, 184, 114, 58, 80, 234, 227, 59, 64, 154, 82, 60, 208, 215, 29, 88, 208, 32, 122, 88, 2, 0, 0, 0, 255, 255, 255, 255, 184, 7, 54, 126, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 10, 0, 0, 0, 1, 2, 3, 2, 3, 2, 3, 2, 3, 2, 170, 170, 8, 0, 0, 0, 72, 173, 0, 0, 0, 0, 0, 0, 112, 173, 0, 0, 0, 0, 0, 0, 208, 182, 0, 0, 0, 0, 0, 0, 208, 182, 0, 0, 16, 14, 0, 0, 4, 0, 48, 0, 80, 0, 66, 0, 54, 0, 170, 170, 103, 143, 0, 224, 113, 143, 0, 224, 119, 143, 0, 16, 123, 143, 0, 224, 2, 0, 0, 0, 255, 255, 255, 255, 204, 24, 54, 126, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 52, 156, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 3, 0, 80, 0, 66, 0, 54, 0, 139, 143, 0, 224, 143, 143, 0, 16, 145, 143, 0, 224, 2, 0, 0, 0, 255, 255, 255, 255, 168, 8, 54, 126, 170, 170, 170, 170, 1, 0, 0, 0, 1, 170, 170, 170, 4, 0, 0, 0, 88, 172, 0, 0, 0, 0, 0, 0, 192, 168, 0, 0, 0, 0, 0, 0, 3, 0, 80, 0, 66, 0, 54, 0, 155, 143, 0, 224, 159, 143, 0, 16, 161, 143, 0, 224, 2, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 171, 143, 0, 224, 150, 0, 0, 0, 224, 42, 177, 128, 208, 13, 161, 129, 224, 12, 145, 130, 208, 239, 128, 131, 224, 238, 112, 132, 80, 12, 106, 133, 96, 11, 90, 134, 80, 238, 73, 135, 96, 237, 57, 136, 80, 208, 41, 137, 96, 207, 25, 138, 80, 178, 9, 139, 96, 177, 249, 139, 80, 148, 233, 140, 96, 147, 217, 141, 208, 176, 210, 142, 96, 117, 185, 143, 208, 146, 178, 144, 224, 145, 162, 145, 208, 116, 146, 146, 224, 115, 130, 147, 208, 86, 114, 148, 224, 85, 98, 149, 208, 56, 82, 150, 224, 55, 66, 151, 208, 26, 50, 152, 224, 25, 34, 153, 80, 55, 27, 154, 96, 54, 11, 155, 80, 25, 251, 155, 96, 24, 235, 156, 80, 251, 218, 157, 96, 250, 202, 158, 80, 221, 186, 159, 96, 220, 170, 160, 80, 191, 154, 161, 96, 190, 138, 162, 208, 219, 131, 163, 96, 160, 106, 164, 208, 189, 99, 165, 224, 188, 83, 166, 208, 159, 67, 167, 224, 158, 51, 168, 208, 129, 35, 169, 224, 128, 19, 170, 208, 99, 3, 171, 224, 98, 243, 171, 208, 69, 227, 172, 224, 68, 211, 173, 80, 98, 204, 174, 224, 38, 179, 175, 80, 68, 172, 176, 96, 67, 156, 177, 80, 38, 140, 178, 96, 37, 124, 179, 80, 8, 108, 180, 96, 7, 92, 181, 80, 234, 75, 182, 96, 233, 59, 183, 80, 204, 43, 184, 96, 203, 27, 185, 208, 232, 20, 186, 224, 231, 4, 187, 208, 202, 244, 187, 224, 201, 228, 188, 208, 172, 212, 189, 224, 171, 196, 190, 208, 142, 180, 191, 224, 141, 164, 192, 208, 112, 148, 193, 224, 111, 132, 194, 80, 141, 125, 195, 224, 81, 100, 196, 80, 111, 93, 197, 96, 110, 77, 198, 80, 81, 61, 199, 96, 80, 45, 200, 80, 51, 29, 201, 96, 50, 13, 202, 80, 21, 253, 202, 96, 20, 237, 203, 80, 247, 220, 204, 96, 246, 204, 205, 208, 19, 198, 206, 224, 18, 182, 207, 208, 245, 165, 208, 224, 244, 149, 209, 208, 215, 133, 210, 224, 214, 117, 211, 208, 185, 101, 212, 224, 184, 85, 213, 208, 155, 69, 214, 224, 154, 53, 215, 80, 184, 46, 216, 224, 124, 21, 217, 80, 154, 14, 218, 96, 153, 254, 218, 80, 124, 238, 219, 96, 123, 222, 220, 80, 94, 206, 221, 96, 93, 190, 222, 80, 64, 174, 223, 96, 63, 158, 224, 80, 34, 142, 225, 96, 33, 126, 226, 208, 62, 119, 227, 96, 3, 94, 228, 208, 32, 87, 229, 224, 31, 71, 230, 208, 2, 55, 231, 224, 1, 39, 232, 208, 228, 22, 233, 224, 227, 6, 234, 208, 198, 246, 234, 224, 197, 230, 235, 208, 168, 214, 236, 224, 167, 198, 237, 80, 197, 191, 238, 96, 196, 175, 239, 80, 167, 159, 240, 96, 166, 143, 241, 80, 137, 127, 242, 96, 136, 111, 243, 80, 107, 95, 244, 96, 106, 79, 245, 80, 77, 63, 246, 96, 76, 47, 247, 208, 105, 40, 248, 96, 46, 15, 249, 208, 75, 8, 250, 224, 74, 248, 250, 208, 45, 232, 251, 224, 44, 216, 252, 208, 15, 200, 253, 224, 14, 184, 254, 208, 241, 167, 255, 224, 240, 151, 0, 208, 211, 135, 1, 224, 210, 119, 2, 80, 240, 112, 3, 96, 239, 96, 4, 80, 210, 80, 5, 96, 209, 64, 6, 80, 180, 48, 7, 96, 11, 141, 7, 80, 128, 53, 9, 224, 134, 173, 9, 80, 120, 240, 10, 96, 119, 224, 11, 208, 148, 217, 12, 8, 0, 0, 0, 255, 255, 255, 255, 224, 102, 241, 124, 255, 255, 255, 255, 208, 73, 225, 125, 255, 255, 255, 255, 224, 72, 209, 126, 255, 255, 255, 255, 208, 43, 193, 127, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 154, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 170, 170, 4, 0, 0, 0, 192, 199, 255, 255, 0, 0, 0, 0, 192, 199, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 192, 199, 255, 127, 225, 3, 0, 96, 185, 7, 0, 112, 176, 143, 0, 224, 71, 144, 0, 224, 83, 144, 0, 16, 123, 144, 0, 224, 2, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 139, 144, 0, 224, 150, 0, 0, 0, 0, 71, 177, 128, 240, 41, 161, 129, 0, 41, 145, 130, 240, 11, 129, 131, 0, 11, 113, 132, 112, 40, 106, 133, 128, 39, 90, 134, 112, 10, 74, 135, 128, 9, 58, 136, 112, 236, 41, 137, 128, 235, 25, 138, 112, 206, 9, 139, 128, 205, 249, 139, 112, 176, 233, 140, 128, 175, 217, 141, 240, 204, 210, 142, 128, 145, 185, 143, 240, 174, 178, 144, 0, 174, 162, 145, 240, 144, 146, 146, 0, 144, 130, 147, 240, 114, 114, 148, 0, 114, 98, 149, 240, 84, 82, 150, 0, 84, 66, 151, 240, 54, 50, 152, 0, 54, 34, 153, 112, 83, 27, 154, 128, 82, 11, 155, 112, 53, 251, 155, 128, 52, 235, 156, 112, 23, 219, 157, 128, 22, 203, 158, 112, 249, 186, 159, 128, 248, 170, 160, 112, 219, 154, 161, 128, 218, 138, 162, 240, 247, 131, 163, 128, 188, 106, 164, 240, 217, 99, 165, 0, 217, 83, 166, 240, 187, 67, 167, 0, 187, 51, 168, 240, 157, 35, 169, 0, 157, 19, 170, 240, 127, 3, 171, 0, 127, 243, 171, 240, 97, 227, 172, 0, 97, 211, 173, 112, 126, 204, 174, 0, 67, 179, 175, 112, 96, 172, 176, 128, 95, 156, 177, 112, 66, 140, 178, 128, 65, 124, 179, 112, 36, 108, 180, 128, 35, 92, 181, 112, 6, 76, 182, 128, 5, 60, 183, 112, 232, 43, 184, 128, 231, 27, 185, 240, 4, 21, 186, 0, 4, 5, 187, 240, 230, 244, 187, 0, 230, 228, 188, 240, 200, 212, 189, 0, 200, 196, 190, 240, 170, 180, 191, 0, 170, 164, 192, 240, 140, 148, 193, 0, 140, 132, 194, 112, 169, 125, 195, 0, 110, 100, 196, 112, 139, 93, 197, 128, 138, 77, 198, 112, 109, 61, 199, 128, 108, 45, 200, 112, 79, 29, 201, 128, 78, 13, 202, 112, 49, 253, 202, 128, 48, 237, 203, 112, 19, 221, 204, 128, 18, 205, 205, 240, 47, 198, 206, 0, 47, 182, 207, 240, 17, 166, 208, 0, 17, 150, 209, 240, 243, 133, 210, 0, 243, 117, 211, 240, 213, 101, 212, 0, 213, 85, 213, 240, 183, 69, 214, 0, 183, 53, 215, 112, 212, 46, 216, 0, 153, 21, 217, 112, 182, 14, 218, 128, 181, 254, 218, 112, 152, 238, 219, 128, 151, 222, 220, 112, 122, 206, 221, 128, 121, 190, 222, 112, 92, 174, 223, 128, 91, 158, 224, 112, 62, 142, 225, 128, 61, 126, 226, 240, 90, 119, 227, 128, 31, 94, 228, 240, 60, 87, 229, 0, 60, 71, 230, 240, 30, 55, 231, 0, 30, 39, 232, 240, 0, 23, 233, 0, 0, 7, 234, 240, 226, 246, 234, 0, 226, 230, 235, 240, 196, 214, 236, 0, 196, 198, 237, 112, 225, 191, 238, 128, 224, 175, 239, 112, 195, 159, 240, 128, 194, 143, 241, 112, 165, 127, 242, 128, 164, 111, 243, 112, 135, 95, 244, 128, 134, 79, 245, 112, 105, 63, 246, 128, 104, 47, 247, 240, 133, 40, 248, 128, 74, 15, 249, 240, 103, 8, 250, 0, 103, 248, 250, 240, 73, 232, 251, 0, 73, 216, 252, 240, 43, 200, 253, 0, 43, 184, 254, 240, 13, 168, 255, 0, 13, 152, 0, 240, 239, 135, 1, 0, 239, 119, 2, 112, 12, 113, 3, 128, 11, 97, 4, 112, 238, 80, 5, 128, 237, 64, 6, 112, 208, 48, 7, 128, 39, 141, 7, 112, 156, 53, 9, 0, 163, 173, 9, 112, 148, 240, 10, 128, 147, 224, 11, 240, 176, 217, 12, 8, 0, 0, 0, 255, 255, 255, 255, 0, 131, 241, 124, 255, 255, 255, 255, 240, 101, 225, 125, 255, 255, 255, 255, 0, 101, 209, 126, 255, 255, 255, 255, 240, 71, 193, 127, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 154, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 170, 170, 4, 0, 0, 0, 160, 171, 255, 255, 0, 0, 0, 0, 160, 171, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 160, 171, 255, 127, 225, 3, 0, 96, 185, 7, 0, 112, 144, 144, 0, 224, 39, 145, 0, 224, 51, 145, 0, 16, 91, 145, 0, 224, 2, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 107, 145, 0, 224, 150, 0, 0, 0, 240, 56, 177, 128, 224, 27, 161, 129, 240, 26, 145, 130, 224, 253, 128, 131, 240, 252, 112, 132, 96, 26, 106, 133, 112, 25, 90, 134, 96, 252, 73, 135, 112, 251, 57, 136, 96, 222, 41, 137, 112, 221, 25, 138, 96, 192, 9, 139, 112, 191, 249, 139, 96, 162, 233, 140, 112, 161, 217, 141, 224, 190, 210, 142, 112, 131, 185, 143, 224, 160, 178, 144, 240, 159, 162, 145, 224, 130, 146, 146, 240, 129, 130, 147, 224, 100, 114, 148, 240, 99, 98, 149, 224, 70, 82, 150, 240, 69, 66, 151, 224, 40, 50, 152, 240, 39, 34, 153, 96, 69, 27, 154, 112, 68, 11, 155, 96, 39, 251, 155, 112, 38, 235, 156, 96, 9, 219, 157, 112, 8, 203, 158, 96, 235, 186, 159, 112, 234, 170, 160, 96, 205, 154, 161, 112, 204, 138, 162, 224, 233, 131, 163, 112, 174, 106, 164, 224, 203, 99, 165, 240, 202, 83, 166, 224, 173, 67, 167, 240, 172, 51, 168, 224, 143, 35, 169, 240, 142, 19, 170, 224, 113, 3, 171, 240, 112, 243, 171, 224, 83, 227, 172, 240, 82, 211, 173, 96, 112, 204, 174, 240, 52, 179, 175, 96, 82, 172, 176, 112, 81, 156, 177, 96, 52, 140, 178, 112, 51, 124, 179, 96, 22, 108, 180, 112, 21, 92, 181, 96, 248, 75, 182, 112, 247, 59, 183, 96, 218, 43, 184, 112, 217, 27, 185, 224, 246, 20, 186, 240, 245, 4, 187, 224, 216, 244, 187, 240, 215, 228, 188, 224, 186, 212, 189, 240, 185, 196, 190, 224, 156, 180, 191, 240, 155, 164, 192, 224, 126, 148, 193, 240, 125, 132, 194, 96, 155, 125, 195, 240, 95, 100, 196, 96, 125, 93, 197, 112, 124, 77, 198, 96, 95, 61, 199, 112, 94, 45, 200, 96, 65, 29, 201, 112, 64, 13, 202, 96, 35, 253, 202, 112, 34, 237, 203, 96, 5, 221, 204, 112, 4, 205, 205, 224, 33, 198, 206, 240, 32, 182, 207, 224, 3, 166, 208, 240, 2, 150, 209, 224, 229, 133, 210, 240, 228, 117, 211, 224, 199, 101, 212, 240, 198, 85, 213, 224, 169, 69, 214, 240, 168, 53, 215, 96, 198, 46, 216, 240, 138, 21, 217, 96, 168, 14, 218, 112, 167, 254, 218, 96, 138, 238, 219, 112, 137, 222, 220, 96, 108, 206, 221, 112, 107, 190, 222, 96, 78, 174, 223, 112, 77, 158, 224, 96, 48, 142, 225, 112, 47, 126, 226, 224, 76, 119, 227, 112, 17, 94, 228, 224, 46, 87, 229, 240, 45, 71, 230, 224, 16, 55, 231, 240, 15, 39, 232, 224, 242, 22, 233, 240, 241, 6, 234, 224, 212, 246, 234, 240, 211, 230, 235, 224, 182, 214, 236, 240, 181, 198, 237, 96, 211, 191, 238, 112, 210, 175, 239, 96, 181, 159, 240, 112, 180, 143, 241, 96, 151, 127, 242, 112, 150, 111, 243, 96, 121, 95, 244, 112, 120, 79, 245, 96, 91, 63, 246, 112, 90, 47, 247, 224, 119, 40, 248, 112, 60, 15, 249, 224, 89, 8, 250, 240, 88, 248, 250, 224, 59, 232, 251, 240, 58, 216, 252, 224, 29, 200, 253, 240, 28, 184, 254, 224, 255, 167, 255, 240, 254, 151, 0, 224, 225, 135, 1, 240, 224, 119, 2, 96, 254, 112, 3, 112, 253, 96, 4, 96, 224, 80, 5, 112, 223, 64, 6, 96, 194, 48, 7, 112, 25, 141, 7, 96, 142, 53, 9, 240, 148, 173, 9, 96, 134, 240, 10, 112, 133, 224, 11, 224, 162, 217, 12, 8, 0, 0, 0, 255, 255, 255, 255, 240, 116, 241, 124, 255, 255, 255, 255, 224, 87, 225, 125, 255, 255, 255, 255, 240, 86, 209, 126, 255, 255, 255, 255, 224, 57, 193, 127, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 154, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 170, 170, 4, 0, 0, 0, 176, 185, 255, 255, 0, 0, 0, 0, 176, 185, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 176, 185, 255, 127, 225, 3, 0, 96, 185, 7, 0, 112, 112, 145, 0, 224, 7, 146, 0, 224, 19, 146, 0, 16, 59, 146, 0, 224, 2, 0, 0, 0, 96, 115, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 75, 146, 0, 224, 2, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 80, 146, 0, 224, 150, 0, 0, 0, 16, 85, 177, 128, 0, 56, 161, 129, 16, 55, 145, 130, 0, 26, 129, 131, 16, 25, 113, 132, 128, 54, 106, 133, 144, 53, 90, 134, 128, 24, 74, 135, 144, 23, 58, 136, 128, 250, 41, 137, 144, 249, 25, 138, 128, 220, 9, 139, 144, 219, 249, 139, 128, 190, 233, 140, 144, 189, 217, 141, 0, 219, 210, 142, 144, 159, 185, 143, 0, 189, 178, 144, 16, 188, 162, 145, 0, 159, 146, 146, 16, 158, 130, 147, 0, 129, 114, 148, 16, 128, 98, 149, 0, 99, 82, 150, 16, 98, 66, 151, 0, 69, 50, 152, 16, 68, 34, 153, 128, 97, 27, 154, 144, 96, 11, 155, 128, 67, 251, 155, 144, 66, 235, 156, 128, 37, 219, 157, 144, 36, 203, 158, 128, 7, 187, 159, 144, 6, 171, 160, 128, 233, 154, 161, 144, 232, 138, 162, 0, 6, 132, 163, 144, 202, 106, 164, 0, 232, 99, 165, 16, 231, 83, 166, 0, 202, 67, 167, 16, 201, 51, 168, 0, 172, 35, 169, 16, 171, 19, 170, 0, 142, 3, 171, 16, 141, 243, 171, 0, 112, 227, 172, 16, 111, 211, 173, 128, 140, 204, 174, 16, 81, 179, 175, 128, 110, 172, 176, 144, 109, 156, 177, 128, 80, 140, 178, 144, 79, 124, 179, 128, 50, 108, 180, 144, 49, 92, 181, 128, 20, 76, 182, 144, 19, 60, 183, 128, 246, 43, 184, 144, 245, 27, 185, 0, 19, 21, 186, 16, 18, 5, 187, 0, 245, 244, 187, 16, 244, 228, 188, 0, 215, 212, 189, 16, 214, 196, 190, 0, 185, 180, 191, 16, 184, 164, 192, 0, 155, 148, 193, 16, 154, 132, 194, 128, 183, 125, 195, 16, 124, 100, 196, 128, 153, 93, 197, 144, 152, 77, 198, 128, 123, 61, 199, 144, 122, 45, 200, 128, 93, 29, 201, 144, 92, 13, 202, 128, 63, 253, 202, 144, 62, 237, 203, 128, 33, 221, 204, 144, 32, 205, 205, 0, 62, 198, 206, 16, 61, 182, 207, 0, 32, 166, 208, 16, 31, 150, 209, 0, 2, 134, 210, 16, 1, 118, 211, 0, 228, 101, 212, 16, 227, 85, 213, 0, 198, 69, 214, 16, 197, 53, 215, 128, 226, 46, 216, 16, 167, 21, 217, 128, 196, 14, 218, 144, 195, 254, 218, 128, 166, 238, 219, 144, 165, 222, 220, 128, 136, 206, 221, 144, 135, 190, 222, 128, 106, 174, 223, 144, 105, 158, 224, 128, 76, 142, 225, 144, 75, 126, 226, 0, 105, 119, 227, 144, 45, 94, 228, 0, 75, 87, 229, 16, 74, 71, 230, 0, 45, 55, 231, 16, 44, 39, 232, 0, 15, 23, 233, 16, 14, 7, 234, 0, 241, 246, 234, 16, 240, 230, 235, 0, 211, 214, 236, 16, 210, 198, 237, 128, 239, 191, 238, 144, 238, 175, 239, 128, 209, 159, 240, 144, 208, 143, 241, 128, 179, 127, 242, 144, 178, 111, 243, 128, 149, 95, 244, 144, 148, 79, 245, 128, 119, 63, 246, 144, 118, 47, 247, 0, 148, 40, 248, 144, 88, 15, 249, 0, 118, 8, 250, 16, 117, 248, 250, 0, 88, 232, 251, 16, 87, 216, 252, 0, 58, 200, 253, 16, 57, 184, 254, 0, 28, 168, 255, 16, 27, 152, 0, 0, 254, 135, 1, 16, 253, 119, 2, 128, 26, 113, 3, 144, 25, 97, 4, 128, 252, 80, 5, 144, 251, 64, 6, 128, 222, 48, 7, 144, 53, 141, 7, 128, 170, 53, 9, 16, 177, 173, 9, 128, 162, 240, 10, 144, 161, 224, 11, 0, 191, 217, 12, 8, 0, 0, 0, 255, 255, 255, 255, 16, 145, 241, 124, 255, 255, 255, 255, 0, 116, 225, 125, 255, 255, 255, 255, 16, 115, 209, 126, 255, 255, 255, 255, 0, 86, 193, 127, 170, 170, 170, 170, 170, 170, 170, 170, 154, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 170, 170, 4, 0, 0, 0, 144, 157, 255, 255, 0, 0, 0, 0, 144, 157, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 144, 157, 255, 127, 225, 3, 0, 96, 185, 7, 0, 112, 85, 146, 0, 224, 236, 146, 0, 224, 247, 146, 0, 16, 31, 147, 0, 224, 2, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 47, 147, 0, 224, 150, 0, 0, 0, 32, 99, 177, 128, 16, 70, 161, 129, 32, 69, 145, 130, 16, 40, 129, 131, 32, 39, 113, 132, 144, 68, 106, 133, 160, 67, 90, 134, 144, 38, 74, 135, 160, 37, 58, 136, 144, 8, 42, 137, 160, 7, 26, 138, 144, 234, 9, 139, 160, 233, 249, 139, 144, 204, 233, 140, 160, 203, 217, 141, 16, 233, 210, 142, 160, 173, 185, 143, 16, 203, 178, 144, 32, 202, 162, 145, 16, 173, 146, 146, 32, 172, 130, 147, 16, 143, 114, 148, 32, 142, 98, 149, 16, 113, 82, 150, 32, 112, 66, 151, 16, 83, 50, 152, 32, 82, 34, 153, 144, 111, 27, 154, 160, 110, 11, 155, 144, 81, 251, 155, 160, 80, 235, 156, 144, 51, 219, 157, 160, 50, 203, 158, 144, 21, 187, 159, 160, 20, 171, 160, 144, 247, 154, 161, 160, 246, 138, 162, 16, 20, 132, 163, 160, 216, 106, 164, 16, 246, 99, 165, 32, 245, 83, 166, 16, 216, 67, 167, 32, 215, 51, 168, 16, 186, 35, 169, 32, 185, 19, 170, 16, 156, 3, 171, 32, 155, 243, 171, 16, 126, 227, 172, 32, 125, 211, 173, 144, 154, 204, 174, 32, 95, 179, 175, 144, 124, 172, 176, 160, 123, 156, 177, 144, 94, 140, 178, 160, 93, 124, 179, 144, 64, 108, 180, 160, 63, 92, 181, 144, 34, 76, 182, 160, 33, 60, 183, 144, 4, 44, 184, 160, 3, 28, 185, 16, 33, 21, 186, 32, 32, 5, 187, 16, 3, 245, 187, 32, 2, 229, 188, 16, 229, 212, 189, 32, 228, 196, 190, 16, 199, 180, 191, 32, 198, 164, 192, 16, 169, 148, 193, 32, 168, 132, 194, 144, 197, 125, 195, 32, 138, 100, 196, 144, 167, 93, 197, 160, 166, 77, 198, 144, 137, 61, 199, 160, 136, 45, 200, 144, 107, 29, 201, 160, 106, 13, 202, 144, 77, 253, 202, 160, 76, 237, 203, 144, 47, 221, 204, 160, 46, 205, 205, 16, 76, 198, 206, 32, 75, 182, 207, 16, 46, 166, 208, 32, 45, 150, 209, 16, 16, 134, 210, 32, 15, 118, 211, 16, 242, 101, 212, 32, 241, 85, 213, 16, 212, 69, 214, 32, 211, 53, 215, 144, 240, 46, 216, 32, 181, 21, 217, 144, 210, 14, 218, 160, 209, 254, 218, 144, 180, 238, 219, 160, 179, 222, 220, 144, 150, 206, 221, 160, 149, 190, 222, 144, 120, 174, 223, 160, 119, 158, 224, 144, 90, 142, 225, 160, 89, 126, 226, 16, 119, 119, 227, 160, 59, 94, 228, 16, 89, 87, 229, 32, 88, 71, 230, 16, 59, 55, 231, 32, 58, 39, 232, 16, 29, 23, 233, 32, 28, 7, 234, 16, 255, 246, 234, 32, 254, 230, 235, 16, 225, 214, 236, 32, 224, 198, 237, 144, 253, 191, 238, 160, 252, 175, 239, 144, 223, 159, 240, 160, 222, 143, 241, 144, 193, 127, 242, 160, 192, 111, 243, 144, 163, 95, 244, 160, 162, 79, 245, 144, 133, 63, 246, 160, 132, 47, 247, 16, 162, 40, 248, 160, 102, 15, 249, 16, 132, 8, 250, 32, 131, 248, 250, 16, 102, 232, 251, 32, 101, 216, 252, 16, 72, 200, 253, 32, 71, 184, 254, 16, 42, 168, 255, 32, 41, 152, 0, 16, 12, 136, 1, 32, 11, 120, 2, 144, 40, 113, 3, 160, 39, 97, 4, 144, 10, 81, 5, 160, 9, 65, 6, 144, 236, 48, 7, 160, 67, 141, 7, 144, 184, 53, 9, 32, 191, 173, 9, 144, 176, 240, 10, 160, 175, 224, 11, 16, 205, 217, 12, 8, 0, 0, 0, 255, 255, 255, 255, 32, 159, 241, 124, 255, 255, 255, 255, 16, 130, 225, 125, 255, 255, 255, 255, 32, 129, 209, 126, 255, 255, 255, 255, 16, 100, 193, 127, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 154, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 170, 170, 4, 0, 0, 0, 128, 143, 255, 255, 0, 0, 0, 0, 128, 143, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 128, 143, 255, 127, 225, 3, 0, 96, 185, 7, 0, 112, 52, 147, 0, 224, 203, 147, 0, 224, 215, 147, 0, 16, 255, 147, 0, 224, 2, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 1, 0, 54, 0, 15, 148, 0, 224, 150, 0, 0, 0, 48, 113, 177, 128, 32, 84, 161, 129, 48, 83, 145, 130, 32, 54, 129, 131, 48, 53, 113, 132, 160, 82, 106, 133, 176, 81, 90, 134, 160, 52, 74, 135, 176, 51, 58, 136, 160, 22, 42, 137, 176, 21, 26, 138, 160, 248, 9, 139, 176, 247, 249, 139, 160, 218, 233, 140, 176, 217, 217, 141, 32, 247, 210, 142, 176, 187, 185, 143, 32, 217, 178, 144, 48, 216, 162, 145, 32, 187, 146, 146, 48, 186, 130, 147, 32, 157, 114, 148, 48, 156, 98, 149, 32, 127, 82, 150, 48, 126, 66, 151, 32, 97, 50, 152, 48, 96, 34, 153, 160, 125, 27, 154, 176, 124, 11, 155, 160, 95, 251, 155, 176, 94, 235, 156, 160, 65, 219, 157, 176, 64, 203, 158, 160, 35, 187, 159, 176, 34, 171, 160, 160, 5, 155, 161, 176, 4, 139, 162, 32, 34, 132, 163, 176, 230, 106, 164, 32, 4, 100, 165, 48, 3, 84, 166, 32, 230, 67, 167, 48, 229, 51, 168, 32, 200, 35, 169, 48, 199, 19, 170, 32, 170, 3, 171, 48, 169, 243, 171, 32, 140, 227, 172, 48, 139, 211, 173, 160, 168, 204, 174, 48, 109, 179, 175, 160, 138, 172, 176, 176, 137, 156, 177, 160, 108, 140, 178, 176, 107, 124, 179, 160, 78, 108, 180, 176, 77, 92, 181, 160, 48, 76, 182, 176, 47, 60, 183, 160, 18, 44, 184, 176, 17, 28, 185, 32, 47, 21, 186, 48, 46, 5, 187, 32, 17, 245, 187, 48, 16, 229, 188, 32, 243, 212, 189, 48, 242, 196, 190, 32, 213, 180, 191, 48, 212, 164, 192, 32, 183, 148, 193, 48, 182, 132, 194, 160, 211, 125, 195, 48, 152, 100, 196, 160, 181, 93, 197, 176, 180, 77, 198, 160, 151, 61, 199, 176, 150, 45, 200, 160, 121, 29, 201, 176, 120, 13, 202, 160, 91, 253, 202, 176, 90, 237, 203, 160, 61, 221, 204, 176, 60, 205, 205, 32, 90, 198, 206, 48, 89, 182, 207, 32, 60, 166, 208, 48, 59, 150, 209, 32, 30, 134, 210, 48, 29, 118, 211, 32, 0, 102, 212, 48, 255, 85, 213, 32, 226, 69, 214, 48, 225, 53, 215, 160, 254, 46, 216, 48, 195, 21, 217, 160, 224, 14, 218, 176, 223, 254, 218, 160, 194, 238, 219, 176, 193, 222, 220, 160, 164, 206, 221, 176, 163, 190, 222, 160, 134, 174, 223, 176, 133, 158, 224, 160, 104, 142, 225, 176, 103, 126, 226, 32, 133, 119, 227, 176, 73, 94, 228, 32, 103, 87, 229, 48, 102, 71, 230, 32, 73, 55, 231, 48, 72, 39, 232, 32, 43, 23, 233, 48, 42, 7, 234, 32, 13, 247, 234, 48, 12, 231, 235, 32, 239, 214, 236, 48, 238, 198, 237, 160, 11, 192, 238, 176, 10, 176, 239, 160, 237, 159, 240, 176, 236, 143, 241, 160, 207, 127, 242, 176, 206, 111, 243, 160, 177, 95, 244, 176, 176, 79, 245, 160, 147, 63, 246, 176, 146, 47, 247, 32, 176, 40, 248, 176, 116, 15, 249, 32, 146, 8, 250, 48, 145, 248, 250, 32, 116, 232, 251, 48, 115, 216, 252, 32, 86, 200, 253, 48, 85, 184, 254, 32, 56, 168, 255, 48, 55, 152, 0, 32, 26, 136, 1, 48, 25, 120, 2, 160, 54, 113, 3, 176, 53, 97, 4, 160, 24, 81, 5, 176, 23, 65, 6, 160, 250, 48, 7, 176, 81, 141, 7, 160, 198, 53, 9, 48, 205, 173, 9, 160, 190, 240, 10, 176, 189, 224, 11, 32, 219, 217, 12, 8, 0, 0, 0, 255, 255, 255, 255, 48, 173, 241, 124, 255, 255, 255, 255, 32, 144, 225, 125, 255, 255, 255, 255, 48, 143, 209, 126, 255, 255, 255, 255, 32, 114, 193, 127, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 154, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 170, 170, 4, 0, 0, 0, 112, 129, 255, 255, 0, 0, 0, 0, 112, 129, 255, 255, 16, 14, 0, 0, 7, 0, 113, 0, 103, 0, 122, 0, 48, 0, 80, 0, 66, 0, 54, 0, 112, 129, 255, 127, 225, 3, 0, 96, 185, 7, 0, 112, 20, 148, 0, 224, 171, 148, 0, 224, 183, 148, 0, 16, 223, 148, 0, 224, 40, 0, 0, 0, 144, 99, 164, 13, 16, 26, 139, 14, 144, 69, 132, 15, 144, 54, 116, 16, 144, 39, 100, 17, 144, 24, 84, 18, 16, 68, 77, 19, 144, 250, 51, 20, 144, 235, 35, 21, 144, 220, 19, 22, 144, 205, 3, 23, 144, 190, 243, 23, 144, 175, 227, 24, 144, 160, 211, 25, 144, 145, 195, 26, 16, 189, 188, 27, 16, 174, 172, 28, 16, 159, 156, 29, 16, 144, 140, 30, 16, 129, 124, 31, 16, 114, 108, 32, 16, 99, 92, 33, 16, 84, 76, 34, 16, 69, 60, 35, 16, 54, 44, 36, 16, 39, 28, 37, 16, 24, 12, 38, 144, 67, 5, 39, 144, 52, 245, 39, 144, 37, 229, 40, 144, 22, 213, 41, 144, 7, 197, 42, 144, 248, 180, 43, 144, 233, 164, 44, 144, 218, 148, 45, 144, 203, 132, 46, 144, 188, 116, 47, 144, 173, 100, 48, 16, 217, 93, 49, 16, 180, 114, 50, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 170, 40, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 14, 0, 0, 6, 0, 113, 0, 103, 0, 122, 0, 48, 0, 66, 0, 54, 0, 170, 170, 0, 0, 0, 112, 4, 0, 0, 96, 205, 7, 0, 112, 239, 148, 0, 224, 27, 149, 0, 16, 38, 149, 0, 224, 122, 2, 0, 0, 99, 1, 0, 112, 111, 1, 0, 112, 64, 0, 0, 112, 18, 0, 0, 112, 60, 0, 0, 112, 14, 24, 0, 32, 91, 24, 0, 32, 48, 0, 0, 112, 158, 24, 0, 32, 48, 0, 0, 112, 48, 0, 0, 112, 5, 0, 0, 112, 36, 0, 0, 112, 5, 0, 0, 112, 180, 24, 0, 32, 43, 0, 0, 112, 36, 0, 0, 112, 43, 0, 0, 112, 103, 25, 0, 32, 214, 26, 0, 32, 38, 27, 0, 32, 5, 0, 0, 112, 5, 0, 0, 112, 48, 0, 0, 112, 48, 0, 0, 112, 36, 0, 0, 112, 140, 28, 0, 32, 5, 0, 0, 112, 43, 0, 0, 112, 43, 0, 0, 112, 175, 28, 0, 32, 238, 28, 0, 32, 48, 0, 0, 112, 42, 29, 0, 32, 43, 0, 0, 112, 36, 0, 0, 112, 70, 29, 0, 32, 36, 0, 0, 112, 5, 0, 0, 112, 36, 0, 0, 112, 43, 0, 0, 112, 43, 0, 0, 112, 36, 0, 0, 112, 98, 29, 0, 32, 30, 0, 0, 112, 30, 0, 0, 112, 48, 0, 0, 112, 120, 29, 0, 32, 154, 29, 0, 32, 176, 29, 0, 32, 36, 0, 0, 112, 5, 0, 0, 112, 5, 0, 0, 112, 36, 0, 0, 112, 202, 29, 0, 32, 5, 0, 0, 112, 9, 30, 0, 32, 74, 30, 0, 32, 167, 30, 0, 32, 46, 31, 0, 32, 186, 31, 0, 32, 186, 0, 0, 112, 186, 0, 0, 112, 16, 32, 0, 32, 121, 32, 0, 32, 229, 32, 0, 32, 65, 0, 0, 112, 74, 33, 0, 32, 176, 33, 0, 32, 25, 34, 0, 32, 129, 34, 0, 32, 233, 34, 0, 32, 76, 35, 0, 32, 181, 35, 0, 32, 29, 36, 0, 32, 133, 36, 0, 32, 237, 36, 0, 32, 105, 0, 0, 112, 101, 37, 0, 32, 135, 37, 0, 32, 59, 0, 0, 112, 231, 37, 0, 32, 48, 38, 0, 32, 82, 38, 0, 32, 135, 38, 0, 32, 223, 38, 0, 32, 249, 38, 0, 32, 52, 39, 0, 32, 72, 39, 0, 32, 206, 39, 0, 32, 64, 0, 0, 112, 59, 40, 0, 32, 194, 40, 0, 32, 10, 41, 0, 32, 35, 41, 0, 32, 65, 0, 0, 112, 56, 41, 0, 32, 181, 0, 0, 112, 41, 42, 0, 32, 92, 42, 0, 32, 79, 0, 0, 112, 67, 0, 0, 112, 126, 42, 0, 32, 148, 42, 0, 32, 22, 43, 0, 32, 44, 43, 0, 32, 106, 43, 0, 32, 214, 43, 0, 32, 56, 44, 0, 32, 204, 44, 0, 32, 77, 45, 0, 32, 186, 0, 0, 112, 214, 45, 0, 32, 22, 46, 0, 32, 45, 46, 0, 32, 216, 0, 0, 112, 241, 46, 0, 32, 131, 0, 0, 112, 53, 47, 0, 32, 167, 47, 0, 32, 232, 47, 0, 32, 200, 48, 0, 32, 71, 49, 0, 32, 186, 0, 0, 112, 186, 0, 0, 112, 106, 49, 0, 32, 130, 49, 0, 32, 154, 49, 0, 32, 125, 50, 0, 32, 32, 51, 0, 32, 79, 51, 0, 32, 153, 51, 0, 32, 45, 52, 0, 32, 126, 52, 0, 32, 226, 52, 0, 32, 45, 53, 0, 32, 96, 53, 0, 32, 178, 53, 0, 32, 11, 54, 0, 32, 131, 0, 0, 112, 110, 54, 0, 32, 212, 54, 0, 32, 5, 55, 0, 32, 68, 0, 0, 112, 134, 55, 0, 32, 57, 56, 0, 32, 197, 56, 0, 32, 132, 0, 0, 112, 105, 0, 0, 112, 220, 56, 0, 32, 5, 57, 0, 32, 187, 57, 0, 32, 145, 0, 0, 112, 105, 0, 0, 112, 6, 58, 0, 32, 47, 58, 0, 32, 107, 58, 0, 32, 186, 0, 0, 112, 128, 58, 0, 32, 184, 58, 0, 32, 241, 58, 0, 32, 70, 0, 0, 112, 114, 59, 0, 32, 158, 59, 0, 32, 244, 59, 0, 32, 48, 60, 0, 32, 128, 60, 0, 32, 78, 61, 0, 32, 120, 61, 0, 32, 3, 62, 0, 32, 217, 0, 0, 112, 186, 0, 0, 112, 130, 62, 0, 32, 119, 63, 0, 32, 238, 63, 0, 32, 114, 64, 0, 32, 189, 64, 0, 32, 80, 65, 0, 32, 218, 65, 0, 32, 102, 66, 0, 32, 172, 66, 0, 32, 200, 66, 0, 32, 49, 67, 0, 32, 80, 67, 0, 32, 114, 67, 0, 32, 193, 67, 0, 32, 230, 67, 0, 32, 196, 0, 0, 112, 27, 68, 0, 32, 52, 68, 0, 32, 224, 68, 0, 32, 78, 69, 0, 32, 178, 69, 0, 32, 249, 69, 0, 32, 78, 70, 0, 32, 174, 70, 0, 32, 241, 70, 0, 32, 67, 0, 0, 112, 216, 0, 0, 112, 41, 71, 0, 32, 218, 71, 0, 32, 14, 72, 0, 32, 150, 72, 0, 32, 216, 72, 0, 32, 109, 0, 0, 112, 94, 73, 0, 32, 186, 0, 0, 112, 102, 74, 0, 32, 186, 0, 0, 112, 186, 0, 0, 112, 186, 0, 0, 112, 186, 0, 0, 112, 155, 74, 0, 32, 181, 74, 0, 32, 240, 74, 0, 32, 105, 75, 0, 32, 253, 75, 0, 32, 238, 76, 0, 32, 186, 0, 0, 112, 172, 77, 0, 32, 186, 0, 0, 112, 34, 78, 0, 32, 219, 78, 0, 32, 94, 79, 0, 32, 193, 79, 0, 32, 221, 79, 0, 32, 245, 79, 0, 32, 6, 80, 0, 32, 140, 80, 0, 32, 160, 80, 0, 32, 40, 2, 0, 112, 28, 81, 0, 32, 42, 81, 0, 32, 40, 2, 0, 112, 58, 81, 0, 32, 74, 81, 0, 32, 94, 81, 0, 32, 218, 1, 0, 112, 53, 1, 0, 112, 176, 81, 0, 32, 25, 82, 0, 32, 135, 82, 0, 32, 218, 82, 0, 32, 46, 83, 0, 32, 102, 83, 0, 32, 244, 0, 0, 112, 192, 83, 0, 32, 23, 84, 0, 32, 49, 1, 0, 112, 128, 84, 0, 32, 146, 84, 0, 32, 252, 84, 0, 32, 96, 85, 0, 32, 188, 85, 0, 32, 208, 85, 0, 32, 30, 1, 0, 112, 58, 86, 0, 32, 142, 86, 0, 32, 58, 1, 0, 112, 58, 1, 0, 112, 177, 86, 0, 32, 7, 1, 0, 112, 65, 87, 0, 32, 107, 87, 0, 32, 132, 87, 0, 32, 150, 87, 0, 32, 201, 87, 0, 32, 71, 88, 0, 32, 231, 88, 0, 32, 58, 1, 0, 112, 140, 89, 0, 32, 182, 89, 0, 32, 37, 90, 0, 32, 118, 90, 0, 32, 222, 90, 0, 32, 202, 1, 0, 112, 251, 90, 0, 32, 16, 91, 0, 32, 164, 91, 0, 32, 192, 91, 0, 32, 39, 92, 0, 32, 74, 92, 0, 32, 73, 1, 0, 112, 96, 92, 0, 32, 27, 1, 0, 112, 206, 92, 0, 32, 241, 92, 0, 32, 90, 93, 0, 32, 127, 93, 0, 32, 172, 93, 0, 32, 53, 1, 0, 112, 36, 1, 0, 112, 29, 94, 0, 32, 134, 94, 0, 32, 156, 94, 0, 32, 194, 94, 0, 32, 9, 1, 0, 112, 24, 95, 0, 32, 131, 95, 0, 32, 236, 95, 0, 32, 82, 96, 0, 32, 168, 96, 0, 32, 250, 0, 0, 112, 197, 96, 0, 32, 216, 96, 0, 32, 236, 96, 0, 32, 66, 97, 0, 32, 154, 97, 0, 32, 78, 1, 0, 112, 174, 97, 0, 32, 15, 1, 0, 112, 26, 98, 0, 32, 77, 98, 0, 32, 134, 98, 0, 32, 194, 98, 0, 32, 233, 98, 0, 32, 82, 99, 0, 32, 156, 99, 0, 32, 209, 99, 0, 32, 36, 100, 0, 32, 158, 101, 0, 32, 22, 1, 0, 112, 67, 1, 0, 112, 188, 101, 0, 32, 222, 101, 0, 32, 72, 102, 0, 32, 38, 1, 0, 112, 154, 102, 0, 32, 71, 1, 0, 112, 174, 102, 0, 32, 30, 103, 0, 32, 250, 0, 0, 112, 134, 103, 0, 32, 238, 103, 0, 32, 4, 104, 0, 32, 112, 104, 0, 32, 211, 104, 0, 32, 160, 105, 0, 32, 13, 106, 0, 32, 82, 106, 0, 32, 110, 106, 0, 32, 86, 1, 0, 112, 172, 106, 0, 32, 218, 1, 0, 112, 116, 107, 0, 32, 226, 107, 0, 32, 246, 107, 0, 32, 5, 0, 0, 112, 97, 108, 0, 32, 111, 1, 0, 112, 227, 108, 0, 32, 20, 109, 0, 32, 157, 109, 0, 32, 111, 1, 0, 112, 32, 110, 0, 32, 76, 110, 0, 32, 120, 110, 0, 32, 2, 111, 0, 32, 104, 1, 0, 112, 53, 111, 0, 32, 150, 111, 0, 32, 28, 112, 0, 32, 111, 1, 0, 112, 99, 1, 0, 112, 80, 112, 0, 32, 95, 1, 0, 112, 94, 1, 0, 112, 212, 112, 0, 32, 101, 1, 0, 112, 105, 1, 0, 112, 108, 1, 0, 112, 96, 1, 0, 112, 202, 0, 0, 112, 7, 1, 0, 112, 196, 0, 0, 112, 176, 0, 0, 112, 202, 0, 0, 112, 156, 0, 0, 112, 43, 0, 0, 112, 47, 113, 0, 32, 207, 0, 0, 112, 98, 0, 0, 112, 175, 113, 0, 32, 58, 1, 0, 112, 128, 0, 0, 112, 222, 0, 0, 112, 194, 0, 0, 112, 217, 0, 0, 112, 112, 0, 0, 112, 207, 0, 0, 112, 219, 0, 0, 112, 194, 0, 0, 112, 221, 0, 0, 112, 200, 0, 0, 112, 44, 2, 0, 112, 129, 0, 0, 112, 48, 0, 0, 112, 219, 1, 0, 112, 243, 113, 0, 32, 0, 114, 0, 32, 119, 114, 0, 32, 18, 0, 0, 112, 197, 1, 0, 112, 143, 114, 0, 32, 147, 1, 0, 112, 150, 114, 0, 32, 155, 114, 0, 32, 160, 114, 0, 32, 165, 114, 0, 32, 170, 114, 0, 32, 175, 114, 0, 32, 180, 114, 0, 32, 185, 114, 0, 32, 190, 114, 0, 32, 195, 114, 0, 32, 200, 114, 0, 32, 205, 114, 0, 32, 147, 1, 0, 112, 210, 114, 0, 32, 215, 114, 0, 32, 220, 114, 0, 32, 225, 114, 0, 32, 230, 114, 0, 32, 235, 114, 0, 32, 240, 114, 0, 32, 245, 114, 0, 32, 250, 114, 0, 32, 255, 114, 0, 32, 4, 115, 0, 32, 9, 115, 0, 32, 14, 115, 0, 32, 19, 115, 0, 32, 147, 1, 0, 112, 147, 1, 0, 112, 179, 1, 0, 112, 33, 115, 0, 32, 179, 1, 0, 112, 40, 115, 0, 32, 179, 1, 0, 112, 182, 115, 0, 32, 240, 115, 0, 32, 91, 116, 0, 32, 181, 116, 0, 32, 209, 1, 0, 112, 1, 117, 0, 32, 105, 117, 0, 32, 221, 1, 0, 112, 5, 118, 0, 32, 97, 118, 0, 32, 205, 118, 0, 32, 245, 1, 0, 112, 63, 119, 0, 32, 152, 119, 0, 32, 110, 120, 0, 32, 30, 121, 0, 32, 209, 1, 0, 112, 100, 121, 0, 32, 209, 1, 0, 112, 22, 122, 0, 32, 209, 1, 0, 112, 151, 122, 0, 32, 225, 122, 0, 32, 71, 123, 0, 32, 12, 124, 0, 32, 188, 1, 0, 112, 247, 124, 0, 32, 145, 125, 0, 32, 17, 126, 0, 32, 149, 126, 0, 32, 200, 1, 0, 112, 10, 127, 0, 32, 165, 127, 0, 32, 45, 128, 0, 32, 41, 1, 0, 112, 142, 128, 0, 32, 47, 129, 0, 32, 188, 1, 0, 112, 153, 129, 0, 32, 255, 129, 0, 32, 133, 130, 0, 32, 241, 130, 0, 32, 223, 1, 0, 112, 188, 1, 0, 112, 83, 131, 0, 32, 206, 131, 0, 32, 188, 1, 0, 112, 33, 132, 0, 32, 107, 132, 0, 32, 200, 132, 0, 32, 28, 133, 0, 32, 195, 1, 0, 112, 140, 133, 0, 32, 213, 133, 0, 32, 245, 1, 0, 112, 223, 1, 0, 112, 54, 134, 0, 32, 156, 134, 0, 32, 7, 135, 0, 32, 132, 135, 0, 32, 188, 1, 0, 112, 215, 135, 0, 32, 39, 136, 0, 32, 55, 136, 0, 32, 209, 1, 0, 112, 209, 1, 0, 112, 147, 1, 0, 112, 147, 1, 0, 112, 147, 1, 0, 112, 147, 1, 0, 112, 147, 1, 0, 112, 60, 136, 0, 32, 16, 1, 0, 112, 131, 0, 0, 112, 30, 1, 0, 112, 89, 1, 0, 112, 48, 0, 0, 112, 76, 136, 0, 32, 94, 136, 0, 32, 110, 136, 0, 32, 48, 0, 0, 112, 126, 136, 0, 32, 142, 136, 0, 32, 158, 136, 0, 32, 181, 136, 0, 32, 48, 0, 0, 112, 198, 136, 0, 32, 64, 1, 0, 112, 22, 1, 0, 112, 68, 1, 0, 112, 142, 0, 0, 112, 68, 1, 0, 112, 58, 2, 0, 112, 56, 0, 0, 112, 23, 137, 0, 32, 39, 2, 0, 112, 36, 137, 0, 32, 155, 137, 0, 32, 216, 0, 0, 112, 160, 0, 0, 112, 165, 0, 0, 112, 80, 1, 0, 112, 40, 2, 0, 112, 40, 2, 0, 112, 42, 2, 0, 112, 109, 0, 0, 112, 25, 1, 0, 112, 184, 0, 0, 112, 58, 1, 0, 112, 189, 0, 0, 112, 151, 0, 0, 112, 27, 138, 0, 32, 74, 138, 0, 32, 234, 138, 0, 32, 17, 139, 0, 32, 131, 139, 0, 32, 171, 139, 0, 32, 69, 140, 0, 32, 117, 140, 0, 32, 142, 140, 0, 32, 164, 140, 0, 32, 214, 140, 0, 32, 234, 140, 0, 32, 2, 141, 0, 32, 18, 141, 0, 32, 38, 141, 0, 32, 92, 141, 0, 32, 127, 141, 0, 32, 54, 2, 0, 112, 154, 141, 0, 32, 192, 141, 0, 32, 227, 141, 0, 32, 5, 142, 0, 32, 22, 142, 0, 32, 66, 2, 0, 112, 46, 142, 0, 32, 70, 142, 0, 32, 107, 142, 0, 32, 137, 142, 0, 32, 164, 142, 0, 32, 188, 142, 0, 32, 208, 142, 0, 32, 245, 142, 0, 32, 69, 2, 0, 112, 12, 143, 0, 32, 64, 143, 0, 32, 53, 2, 0, 112, 66, 2, 0, 112, 82, 143, 0, 32, 98, 143, 0, 32, 132, 143, 0, 32, 43, 2, 0, 112, 150, 143, 0, 32, 166, 143, 0, 32, 43, 2, 0, 112, 242, 1, 0, 112, 207, 1, 0, 112, 61, 1, 0, 112, 57, 1, 0, 112, 52, 2, 0, 112, 59, 1, 0, 112, 174, 143, 0, 32, 128, 144, 0, 32, 142, 144, 0, 32, 96, 145, 0, 32, 110, 145, 0, 32, 64, 146, 0, 32, 78, 146, 0, 32, 83, 146, 0, 32, 36, 147, 0, 32, 50, 147, 0, 32, 4, 148, 0, 32, 18, 148, 0, 32, 228, 148, 0, 32, 202, 1, 0, 112, 179, 1, 0, 112, 60, 0, 0, 112, 59, 0, 0, 112, 184, 0, 0, 112, 98, 0, 0, 112, 131, 0, 0, 112, 173, 0, 0, 112, 54, 2, 0, 112, 132, 0, 0, 112, 110, 0, 0, 112, 109, 0, 0, 112, 151, 0, 0, 112, 151, 0, 0, 112, 66, 2, 0, 112, 179, 1, 0, 112, 179, 1, 0, 112, 15, 1, 0, 112, 216, 1, 0, 112, 43, 149, 0, 32, 179, 1, 0, 112, 5, 0, 132, 0, 46, 1, 138, 0, 32, 0, 42, 0, 79, 39, 0, 144, 202, 41, 0, 144, 205, 23, 0, 32, 133, 3, 0, 96, 53, 149, 0, 128
};
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
 **********************************************************************
 *   Copyright (C) 1996-2016, International Business Machines
 *   Corporation and others.  All Rights Reserved.
 **********************************************************************
 *
 * Provides functionality for mapping between
 * LCID and Posix IDs or ICU locale to codepage
 *
 * Note: All classes and code in this file are
 *       intended for internal use only.
 *
 * Methods of interest:
 *   unsigned long convertToLCID(const char*);
 *   const char* convertToPosix(unsigned long);
 *
 * Kathleen Wilson, 4/30/96
 *
 *  Date        Name        Description
 *  3/11/97     aliu        Fixed off-by-one bug in assignment operator. Added
 *                          setId() method and safety check against 
 *                          MAX_ID_LENGTH.
 * 04/23/99     stephen     Added C wrapper for convertToPosix.
 * 09/18/00     george      Removed the memory leaks.
 * 08/23/01     george      Convert to C
 */






#if U_PLATFORM_HAS_WIN32_API && UCONFIG_USE_WINDOWS_LCID_MAPPING_API
#include <windows.h>
#include <winnls.h> // LCIDToLocaleName and LocaleNameToLCID
#endif

/*
 * Note:
 * The mapping from Win32 locale ID numbers to POSIX locale strings should
 * be the faster one.
 *
 * Windows LCIDs are defined at https://msdn.microsoft.com/en-us/library/cc233965.aspx
 * [MS-LCID] Windows Language Code Identifier (LCID) Reference
 */

/*
////////////////////////////////////////////////
//
// Internal Classes for LCID <--> POSIX Mapping
//
/////////////////////////////////////////////////
*/

typedef struct ILcidPosixElement
{
    const uint32_t hostID;
    const char * const posixID;
} ILcidPosixElement;

typedef struct ILcidPosixMap
{
    const uint32_t numRegions;
    const struct ILcidPosixElement* const regionMaps;
} ILcidPosixMap;


/*
/////////////////////////////////////////////////
//
// Easy macros to make the LCID <--> POSIX Mapping
//
/////////////////////////////////////////////////
*/

/**
 * The standard one language/one country mapping for LCID.
 * The first element must be the language, and the following
 * elements are the language with the country.
 * @param hostID LCID in host format such as 0x044d
 * @param languageID posix ID of just the language such as 'de'
 * @param posixID posix ID of the language_TERRITORY such as 'de_CH'
 */
#define ILCID_POSIX_ELEMENT_ARRAY(hostID, languageID, posixID) \
static const ILcidPosixElement locmap_ ## languageID [] = { \
    {LANGUAGE_LCID(hostID), #languageID},     /* parent locale */ \
    {hostID, #posixID}, \
};

/**
 * Define a subtable by ID
 * @param id the POSIX ID, either a language or language_TERRITORY
 */
#define ILCID_POSIX_SUBTABLE(id) \
static const ILcidPosixElement locmap_ ## id [] =


/**
 * Create the map for the posixID. This macro supposes that the language string
 * name is the same as the global variable name, and that the first element
 * in the ILcidPosixElement is just the language.
 * @param _posixID the full POSIX ID for this entry.
 */
#define ILCID_POSIX_MAP(_posixID) \
    {UPRV_LENGTHOF(locmap_ ## _posixID), locmap_ ## _posixID}

/*
////////////////////////////////////////////
//
// Create the table of LCID to POSIX Mapping
// None of it should be dynamically created.
//
// Keep static locale variables inside the function so that
// it can be created properly during static init.
//
// Note: This table should be updated periodically. Check the [MS-LCID] Windows Language Code Identifier 
//       (LCID) Reference defined at https://msdn.microsoft.com/en-us/library/cc233965.aspx
//
//       Microsoft is moving away from LCID in favor of locale name as of Vista.  This table needs to be
//       maintained for support of older Windows version.
//       Update: Windows 7 (091130)
//
// Note: Microsoft assign a different LCID if a locale has a sorting variant. POSIX IDs below may contain
//       @collation=XXX, but no other keywords are allowed (at least for now). When uprv_convertToLCID() is
//       called from uloc_getLCID(), keywords other than collation are already removed. If we really need
//       to support other keywords in this mapping data, we must update the implementation.
////////////////////////////////////////////
*/

// TODO: For Windows ideally this table would be a list of exceptions rather than a complete list as 
// LocaleNameToLCID and LCIDToLocaleName provide 90% of these.

ILCID_POSIX_ELEMENT_ARRAY(0x0436, af, af_ZA)

ILCID_POSIX_SUBTABLE(ar) {
    {0x01,   "ar"},
    {0x3801, "ar_AE"},
    {0x3c01, "ar_BH"},
    {0x1401, "ar_DZ"},
    {0x0c01, "ar_EG"},
    {0x0801, "ar_IQ"},
    {0x2c01, "ar_JO"},
    {0x3401, "ar_KW"},
    {0x3001, "ar_LB"},
    {0x1001, "ar_LY"},
    {0x1801, "ar_MA"},
    {0x1801, "ar_MO"},
    {0x2001, "ar_OM"},
    {0x4001, "ar_QA"},
    {0x0401, "ar_SA"},
    {0x2801, "ar_SY"},
    {0x1c01, "ar_TN"},
    {0x2401, "ar_YE"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x044d, as, as_IN)
ILCID_POSIX_ELEMENT_ARRAY(0x045e, am, am_ET)
ILCID_POSIX_ELEMENT_ARRAY(0x047a, arn,arn_CL)

ILCID_POSIX_SUBTABLE(az) {
    {0x2c,   "az"},
    {0x082c, "az_Cyrl_AZ"},  /* Cyrillic based */
    {0x742c, "az_Cyrl"},  /* Cyrillic based */
    {0x042c, "az_Latn_AZ"}, /* Latin based */
    {0x782c, "az_Latn"}, /* Latin based */
    {0x042c, "az_AZ"} /* Latin based */
};

ILCID_POSIX_ELEMENT_ARRAY(0x046d, ba, ba_RU)
ILCID_POSIX_ELEMENT_ARRAY(0x0423, be, be_BY)

/*ILCID_POSIX_SUBTABLE(ber) {
    {0x5f,   "ber"},
    {0x045f, "ber_Arab_DZ"},
    {0x045f, "ber_Arab"},
    {0x085f, "ber_Latn_DZ"},
    {0x085f, "ber_Latn"}
};*/

ILCID_POSIX_ELEMENT_ARRAY(0x0402, bg, bg_BG)

ILCID_POSIX_SUBTABLE(bin) {
    {0x66, "bin"},
    {0x0466, "bin_NG"}
};

ILCID_POSIX_SUBTABLE(bn) {
    {0x45,   "bn"},
    {0x0845, "bn_BD"},
    {0x0445, "bn_IN"}
};

ILCID_POSIX_SUBTABLE(bo) {
    {0x51,   "bo"},
    {0x0851, "bo_BT"},
    {0x0451, "bo_CN"},
    {0x0c51, "dz_BT"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x047e, br, br_FR)

ILCID_POSIX_SUBTABLE(ca) {
    {0x03,   "ca"},
    {0x0403, "ca_ES"},
    {0x0803, "ca_ES_VALENCIA"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0483, co, co_FR)

ILCID_POSIX_SUBTABLE(chr) {
    {0x05c,  "chr"},
    {0x7c5c, "chr_Cher"},
    {0x045c, "chr_Cher_US"},
    {0x045c, "chr_US"}
};

// ICU has chosen different names for these.
ILCID_POSIX_SUBTABLE(ckb) {
    {0x92,   "ckb"},
    {0x7c92, "ckb_Arab"},
    {0x0492, "ckb_Arab_IQ"}
};

/* Declared as cs_CZ to get around compiler errors on z/OS, which defines cs as a function */
ILCID_POSIX_ELEMENT_ARRAY(0x0405, cs, cs_CZ)

ILCID_POSIX_ELEMENT_ARRAY(0x0452, cy, cy_GB)
ILCID_POSIX_ELEMENT_ARRAY(0x0406, da, da_DK)

// Windows doesn't know POSIX or BCP47 Unicode phonebook sort names
ILCID_POSIX_SUBTABLE(de) {
    {0x07,   "de"},
    {0x0c07, "de_AT"},
    {0x0807, "de_CH"},
    {0x0407, "de_DE"},
    {0x1407, "de_LI"},
    {0x1007, "de_LU"},
    {0x10407,"de_DE@collation=phonebook"},  /*This is really de_DE_PHONEBOOK on Windows*/
    {0x10407,"de@collation=phonebook"}  /*This is really de_DE_PHONEBOOK on Windows*/
};

ILCID_POSIX_ELEMENT_ARRAY(0x0465, dv, dv_MV)
ILCID_POSIX_ELEMENT_ARRAY(0x0408, el, el_GR)

// Windows uses an empty string for 'invariant'
ILCID_POSIX_SUBTABLE(en) {
    {0x09,   "en"},
    {0x0c09, "en_AU"},
    {0x2809, "en_BZ"},
    {0x1009, "en_CA"},
    {0x0809, "en_GB"},
    {0x3c09, "en_HK"},
    {0x3809, "en_ID"},
    {0x1809, "en_IE"},
    {0x4009, "en_IN"},
    {0x2009, "en_JM"},
    {0x4409, "en_MY"},
    {0x1409, "en_NZ"},
    {0x3409, "en_PH"},
    {0x4809, "en_SG"},
    {0x2C09, "en_TT"},
    {0x0409, "en_US"},
    {0x007f, "en_US_POSIX"}, /* duplicate for round-tripping */
    {0x2409, "en_029"},
    {0x1c09, "en_ZA"},
    {0x3009, "en_ZW"},
    {0x2409, "en_VI"},  /* Virgin Islands AKA Caribbean Islands (en_CB). On Windows8+ This is 0x1000 or dynamically assigned */
    {0x0409, "en_AS"},  /* Alias for en_US. Leave last.  On Windows8+ This is 0x1000 or dynamically assigned */
    {0x0409, "en_GU"},  /* Alias for en_US. Leave last.  On Windows8+ This is 0x1000 or dynamically assigned */
    {0x0409, "en_MH"},  /* Alias for en_US. Leave last.  On Windows8+ This is 0x1000 or dynamically assigned */
    {0x0409, "en_MP"},  /* Alias for en_US. Leave last.  On Windows8+ This is 0x1000 or dynamically assigned */
    {0x0409, "en_UM"}   /* Alias for en_US. Leave last.  On Windows8+ This is 0x1000 or dynamically assigned */
};

ILCID_POSIX_SUBTABLE(en_US_POSIX) {
    {0x007f, "en_US_POSIX"} /* duplicate for roundtripping */
};

// Windows doesn't know POSIX or BCP47 Unicode traditional sort names
ILCID_POSIX_SUBTABLE(es) {
    {0x0a,   "es"},
    {0x2c0a, "es_AR"},
    {0x400a, "es_BO"},
    {0x340a, "es_CL"},
    {0x240a, "es_CO"},
    {0x140a, "es_CR"},
    {0x5c0a, "es_CU"},
    {0x1c0a, "es_DO"},
    {0x300a, "es_EC"},
    {0x0c0a, "es_ES"},      /*Modern sort.*/
    {0x100a, "es_GT"},
    {0x480a, "es_HN"},
    {0x080a, "es_MX"},
    {0x4c0a, "es_NI"},
    {0x180a, "es_PA"},
    {0x280a, "es_PE"},
    {0x500a, "es_PR"},
    {0x3c0a, "es_PY"},
    {0x440a, "es_SV"},
    {0x540a, "es_US"},
    {0x380a, "es_UY"},
    {0x200a, "es_VE"},
    {0x580a, "es_419"},
    {0x040a, "es_ES@collation=traditional"},
    {0x040a, "es@collation=traditional"}        // Windows will treat this as es-ES@collation=traditional
};

ILCID_POSIX_ELEMENT_ARRAY(0x0425, et, et_EE)
ILCID_POSIX_ELEMENT_ARRAY(0x042d, eu, eu_ES)

/* ISO-639 doesn't distinguish between Persian and Dari.*/
ILCID_POSIX_SUBTABLE(fa) {
    {0x29,   "fa"},
    {0x0429, "fa_IR"},  /* Persian/Farsi (Iran) */
    {0x048c, "fa_AF"}   /* Persian/Dari (Afghanistan) */
};


/* duplicate for roundtripping */
ILCID_POSIX_SUBTABLE(fa_AF) {
    {0x8c,   "fa_AF"},  /* Persian/Dari (Afghanistan) */
    {0x048c, "fa_AF"}   /* Persian/Dari (Afghanistan) */
};

ILCID_POSIX_SUBTABLE(ff) {
    {0x67,   "ff"},
    {0x7c67, "ff_Latn"},
    {0x0867, "ff_Latn_SN"},
    {0x0467, "ff_NG"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x040b, fi, fi_FI)
ILCID_POSIX_ELEMENT_ARRAY(0x0464, fil,fil_PH)
ILCID_POSIX_ELEMENT_ARRAY(0x0438, fo, fo_FO)

ILCID_POSIX_SUBTABLE(fr) {
    {0x0c,   "fr"},
    {0x080c, "fr_BE"},
    {0x0c0c, "fr_CA"},
    {0x240c, "fr_CD"},
    {0x240c, "fr_CG"},
    {0x100c, "fr_CH"},
    {0x300c, "fr_CI"},
    {0x2c0c, "fr_CM"},
    {0x040c, "fr_FR"},
    {0x3c0c, "fr_HT"},
    {0x140c, "fr_LU"},
    {0x380c, "fr_MA"},
    {0x180c, "fr_MC"},
    {0x340c, "fr_ML"},
    {0x200c, "fr_RE"},
    {0x280c, "fr_SN"},
    {0xe40c, "fr_015"},
    {0x1c0c, "fr_029"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0467, fuv, fuv_NG)

ILCID_POSIX_ELEMENT_ARRAY(0x0462, fy, fy_NL)

ILCID_POSIX_SUBTABLE(ga) { /* Gaelic (Ireland) */
    {0x3c,   "ga"},
    {0x083c, "ga_IE"},
    {0x043c, "gd_GB"}
};

ILCID_POSIX_SUBTABLE(gd) { /* Gaelic (Scotland) */
    {0x91,   "gd"},
    {0x0491, "gd_GB"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0456, gl, gl_ES)
ILCID_POSIX_ELEMENT_ARRAY(0x0447, gu, gu_IN)
ILCID_POSIX_ELEMENT_ARRAY(0x0474, gn, gn_PY)
ILCID_POSIX_ELEMENT_ARRAY(0x0484, gsw,gsw_FR)

ILCID_POSIX_SUBTABLE(ha) {
    {0x68,   "ha"},
    {0x7c68, "ha_Latn"},
    {0x0468, "ha_Latn_NG"},
};

ILCID_POSIX_ELEMENT_ARRAY(0x0475, haw,haw_US)
ILCID_POSIX_ELEMENT_ARRAY(0x040d, he, he_IL)
ILCID_POSIX_ELEMENT_ARRAY(0x0439, hi, hi_IN)

/* This LCID is really four different locales.*/
ILCID_POSIX_SUBTABLE(hr) {
    {0x1a,   "hr"},
    {0x141a, "bs_Latn_BA"},  /* Bosnian, Bosnia and Herzegovina */
    {0x681a, "bs_Latn"},  /* Bosnian, Bosnia and Herzegovina */
    {0x141a, "bs_BA"},  /* Bosnian, Bosnia and Herzegovina */
    {0x781a, "bs"},     /* Bosnian */
    {0x201a, "bs_Cyrl_BA"},  /* Bosnian, Bosnia and Herzegovina */
    {0x641a, "bs_Cyrl"},  /* Bosnian, Bosnia and Herzegovina */
    {0x101a, "hr_BA"},  /* Croatian in Bosnia */
    {0x041a, "hr_HR"},  /* Croatian*/
    {0x2c1a, "sr_Latn_ME"},
    {0x241a, "sr_Latn_RS"},
    {0x181a, "sr_Latn_BA"}, /* Serbo-Croatian in Bosnia */
    {0x081a, "sr_Latn_CS"}, /* Serbo-Croatian*/
    {0x701a, "sr_Latn"},    /* It's 0x1a or 0x081a, pick one to make the test program happy. */
    {0x1c1a, "sr_Cyrl_BA"}, /* Serbo-Croatian in Bosnia */
    {0x0c1a, "sr_Cyrl_CS"}, /* Serbian*/
    {0x301a, "sr_Cyrl_ME"},
    {0x281a, "sr_Cyrl_RS"},
    {0x6c1a, "sr_Cyrl"},    /* It's 0x1a or 0x0c1a, pick one to make the test program happy. */
    {0x7c1a, "sr"}          /* In CLDR sr is sr_Cyrl. */
};

ILCID_POSIX_SUBTABLE(hsb) {
    {0x2E,   "hsb"},
    {0x042E, "hsb_DE"},
    {0x082E, "dsb_DE"},
    {0x7C2E, "dsb"},
};

ILCID_POSIX_ELEMENT_ARRAY(0x040e, hu, hu_HU)
ILCID_POSIX_ELEMENT_ARRAY(0x042b, hy, hy_AM)

ILCID_POSIX_SUBTABLE(ibb) {
    {0x69, "ibb"},
    {0x0469, "ibb_NG"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0421, id, id_ID)
ILCID_POSIX_ELEMENT_ARRAY(0x0470, ig, ig_NG)
ILCID_POSIX_ELEMENT_ARRAY(0x0478, ii, ii_CN)
ILCID_POSIX_ELEMENT_ARRAY(0x040f, is, is_IS)

ILCID_POSIX_SUBTABLE(it) {
    {0x10,   "it"},
    {0x0810, "it_CH"},
    {0x0410, "it_IT"}
};

ILCID_POSIX_SUBTABLE(iu) {
    {0x5d,   "iu"},
    {0x045d, "iu_Cans_CA"},
    {0x785d, "iu_Cans"},
    {0x085d, "iu_Latn_CA"},
    {0x7c5d, "iu_Latn"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x040d, iw, iw_IL)    /*Left in for compatibility*/
ILCID_POSIX_ELEMENT_ARRAY(0x0411, ja, ja_JP)
ILCID_POSIX_ELEMENT_ARRAY(0x0437, ka, ka_GE)
ILCID_POSIX_ELEMENT_ARRAY(0x043f, kk, kk_KZ)
ILCID_POSIX_ELEMENT_ARRAY(0x046f, kl, kl_GL)
ILCID_POSIX_ELEMENT_ARRAY(0x0453, km, km_KH)
ILCID_POSIX_ELEMENT_ARRAY(0x044b, kn, kn_IN)

ILCID_POSIX_SUBTABLE(ko) {
    {0x12,   "ko"},
    {0x0812, "ko_KP"},
    {0x0412, "ko_KR"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0457, kok, kok_IN)
ILCID_POSIX_ELEMENT_ARRAY(0x0471, kr,  kr_NG)

ILCID_POSIX_SUBTABLE(ks) {         /* We could add PK and CN too */
    {0x60,   "ks"},
    {0x0460, "ks_Arab_IN"},
    {0x0860, "ks_Deva_IN"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0440, ky, ky_KG)   /* Kyrgyz is spoken in Kyrgyzstan */

ILCID_POSIX_SUBTABLE(la) {
    {0x76,   "la"},
    {0x0476, "la_001"},
    {0x0476, "la_IT"}       /*Left in for compatibility*/
};

ILCID_POSIX_ELEMENT_ARRAY(0x046e, lb, lb_LU)
ILCID_POSIX_ELEMENT_ARRAY(0x0454, lo, lo_LA)
ILCID_POSIX_ELEMENT_ARRAY(0x0427, lt, lt_LT)
ILCID_POSIX_ELEMENT_ARRAY(0x0426, lv, lv_LV)
ILCID_POSIX_ELEMENT_ARRAY(0x0481, mi, mi_NZ)
ILCID_POSIX_ELEMENT_ARRAY(0x042f, mk, mk_MK)
ILCID_POSIX_ELEMENT_ARRAY(0x044c, ml, ml_IN)

ILCID_POSIX_SUBTABLE(mn) {
    {0x50,   "mn"},
    {0x0450, "mn_MN"},
    {0x7c50, "mn_Mong"},
    {0x0850, "mn_Mong_CN"},
    {0x0850, "mn_CN"},
    {0x7850, "mn_Cyrl"},
    {0x0c50, "mn_Mong_MN"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0458, mni,mni_IN)
ILCID_POSIX_ELEMENT_ARRAY(0x047c, moh,moh_CA)
ILCID_POSIX_ELEMENT_ARRAY(0x044e, mr, mr_IN)

ILCID_POSIX_SUBTABLE(ms) {
    {0x3e,   "ms"},
    {0x083e, "ms_BN"},   /* Brunei Darussalam*/
    {0x043e, "ms_MY"}    /* Malaysia*/
};

ILCID_POSIX_ELEMENT_ARRAY(0x043a, mt, mt_MT)
ILCID_POSIX_ELEMENT_ARRAY(0x0455, my, my_MM)

ILCID_POSIX_SUBTABLE(ne) {
    {0x61,   "ne"},
    {0x0861, "ne_IN"},   /* India*/
    {0x0461, "ne_NP"}    /* Nepal*/
};

ILCID_POSIX_SUBTABLE(nl) {
    {0x13,   "nl"},
    {0x0813, "nl_BE"},
    {0x0413, "nl_NL"}
};

/* The "no" locale split into nb and nn.  By default in ICU, "no" is nb.*/
// TODO: Not all of these are needed on Windows, but I don't know how ICU treats preferred ones here.
ILCID_POSIX_SUBTABLE(no) {
    {0x14,   "no"},     /* really nb_NO - actually Windows differentiates between neutral (no region) and specific (with region) */ 
    {0x7c14, "nb"},     /* really nb */
    {0x0414, "nb_NO"},  /* really nb_NO. Keep first in the 414 list. */
    {0x0414, "no_NO"},  /* really nb_NO */
    {0x0814, "nn_NO"},  /* really nn_NO. Keep first in the 814 list.  */
    {0x7814, "nn"},     /* It's 0x14 or 0x814, pick one to make the test program happy. */
    {0x0814, "no_NO_NY"}/* really nn_NO */
};

ILCID_POSIX_ELEMENT_ARRAY(0x046c, nso,nso_ZA)   /* TODO: Verify the ISO-639 code */
ILCID_POSIX_ELEMENT_ARRAY(0x0482, oc, oc_FR)

ILCID_POSIX_SUBTABLE(om) { /* TODO: Verify the country */
    {0x72,   "om"},
    {0x0472, "om_ET"},
    {0x0472, "gaz_ET"}
};

/* Declared as or_IN to get around compiler errors*/
ILCID_POSIX_SUBTABLE(or_IN) {
    {0x48,   "or"},
    {0x0448, "or_IN"},
};

ILCID_POSIX_SUBTABLE(pa) {
    {0x46,   "pa"},
    {0x0446, "pa_IN"},
    {0x0846, "pa_Arab_PK"},
    {0x0846, "pa_PK"}
};

ILCID_POSIX_SUBTABLE(pap) {
    {0x79, "pap"},
    {0x0479, "pap_029"},
    {0x0479, "pap_AN"}     /*Left in for compatibility*/
};

ILCID_POSIX_ELEMENT_ARRAY(0x0415, pl, pl_PL)
ILCID_POSIX_ELEMENT_ARRAY(0x0463, ps, ps_AF)

ILCID_POSIX_SUBTABLE(pt) {
    {0x16,   "pt"},
    {0x0416, "pt_BR"},
    {0x0816, "pt_PT"}
};

ILCID_POSIX_SUBTABLE(qu) {
    {0x6b,   "qu"},
    {0x046b, "qu_BO"},
    {0x086b, "qu_EC"},
    {0x0C6b, "qu_PE"},
    {0x046b, "quz_BO"},
    {0x086b, "quz_EC"},
    {0x0C6b, "quz_PE"}
};

ILCID_POSIX_SUBTABLE(quc) {
    {0x93,   "quc"},
    {0x0493, "quc_CO"},
    /*
        "quc_Latn_GT" is an exceptional case. Language ID of "quc"
        is 0x93, but LCID of "quc_Latn_GT" is 0x486, which should be
        under the group of "qut". "qut" is a retired ISO 639-3 language
        code for West Central Quiche, and merged to "quc".
        It looks Windows previously reserved "qut" for K'iche', but,
        decided to use "quc" when adding a locale for K'iche' (Guatemala).

        This data structure used here assumes language ID bits in
        LCID is unique for alphabetic language code. But this is not true
        for "quc_Latn_GT". If we don't have the data below, LCID look up
        by alphabetic locale ID (POSIX) will fail. The same entry is found
        under "qut" below, which is required for reverse look up.
    */
    {0x0486, "quc_Latn_GT"}
};

ILCID_POSIX_SUBTABLE(qut) {
    {0x86,   "qut"},
    {0x0486, "qut_GT"},
    /*
        See the note in "quc" above.
    */
    {0x0486, "quc_Latn_GT"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0417, rm, rm_CH)

ILCID_POSIX_SUBTABLE(ro) {
    {0x18,   "ro"},
    {0x0418, "ro_RO"},
    {0x0818, "ro_MD"}
};

// TODO: This is almost certainly 'wrong'.  0 in Windows is a synonym for LOCALE_USER_DEFAULT.
// More likely this is a similar concept to the Windows 0x7f Invariant locale ""
// (Except that it's not invariant in ICU)
ILCID_POSIX_SUBTABLE(root) {
    {0x00,   "root"}
};

ILCID_POSIX_SUBTABLE(ru) {
    {0x19,   "ru"},
    {0x0419, "ru_RU"},
    {0x0819, "ru_MD"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0487, rw, rw_RW)
ILCID_POSIX_ELEMENT_ARRAY(0x044f, sa, sa_IN)
ILCID_POSIX_ELEMENT_ARRAY(0x0485, sah,sah_RU)

ILCID_POSIX_SUBTABLE(sd) {
    {0x59,   "sd"},
    {0x0459, "sd_Deva_IN"},
    {0x0459, "sd_IN"},
    {0x0859, "sd_Arab_PK"},
    {0x0859, "sd_PK"},
    {0x7c59, "sd_Arab"}
};

ILCID_POSIX_SUBTABLE(se) {
    {0x3b,   "se"},
    {0x0c3b, "se_FI"},
    {0x043b, "se_NO"},
    {0x083b, "se_SE"},
    {0x783b, "sma"},
    {0x183b, "sma_NO"},
    {0x1c3b, "sma_SE"},
    {0x7c3b, "smj"},
    {0x703b, "smn"},
    {0x743b, "sms"},
    {0x103b, "smj_NO"},
    {0x143b, "smj_SE"},
    {0x243b, "smn_FI"},
    {0x203b, "sms_FI"},
};

ILCID_POSIX_ELEMENT_ARRAY(0x045b, si, si_LK)
ILCID_POSIX_ELEMENT_ARRAY(0x041b, sk, sk_SK)
ILCID_POSIX_ELEMENT_ARRAY(0x0424, sl, sl_SI)

ILCID_POSIX_SUBTABLE(so) {
    {0x77,   "so"},
    {0x0477, "so_SO"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x041c, sq, sq_AL)
ILCID_POSIX_ELEMENT_ARRAY(0x0430, st, st_ZA)

ILCID_POSIX_SUBTABLE(sv) {
    {0x1d,   "sv"},
    {0x081d, "sv_FI"},
    {0x041d, "sv_SE"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0441, sw, sw_KE)
ILCID_POSIX_ELEMENT_ARRAY(0x045A, syr, syr_SY)

ILCID_POSIX_SUBTABLE(ta) {
    {0x49,   "ta"},
    {0x0449, "ta_IN"},
    {0x0849, "ta_LK"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x044a, te, te_IN)

/* Cyrillic based by default */
ILCID_POSIX_SUBTABLE(tg) {
    {0x28,   "tg"},
    {0x7c28, "tg_Cyrl"},
    {0x0428, "tg_Cyrl_TJ"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x041e, th, th_TH)

ILCID_POSIX_SUBTABLE(ti) {
    {0x73,   "ti"},
    {0x0873, "ti_ER"},
    {0x0473, "ti_ET"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0442, tk, tk_TM)

ILCID_POSIX_SUBTABLE(tn) {
    {0x32,   "tn"},
    {0x0832, "tn_BW"},
    {0x0432, "tn_ZA"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x041f, tr, tr_TR)
ILCID_POSIX_ELEMENT_ARRAY(0x0431, ts, ts_ZA)
ILCID_POSIX_ELEMENT_ARRAY(0x0444, tt, tt_RU)

ILCID_POSIX_SUBTABLE(tzm) {
    {0x5f,   "tzm"},
    {0x7c5f, "tzm_Latn"},
    {0x085f, "tzm_Latn_DZ"},
    {0x105f, "tzm_Tfng_MA"},
    {0x045f, "tzm_Arab_MA"},
    {0x045f, "tmz"}
};

ILCID_POSIX_SUBTABLE(ug) {
    {0x80,   "ug"},
    {0x0480, "ug_CN"},
    {0x0480, "ug_Arab_CN"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x0422, uk, uk_UA)

ILCID_POSIX_SUBTABLE(ur) {
    {0x20,   "ur"},
    {0x0820, "ur_IN"},
    {0x0420, "ur_PK"}
};

ILCID_POSIX_SUBTABLE(uz) {
    {0x43,   "uz"},
    {0x0843, "uz_Cyrl_UZ"},  /* Cyrillic based */
    {0x7843, "uz_Cyrl"},  /* Cyrillic based */
    {0x0843, "uz_UZ"},  /* Cyrillic based */
    {0x0443, "uz_Latn_UZ"}, /* Latin based */
    {0x7c43, "uz_Latn"} /* Latin based */
};

ILCID_POSIX_SUBTABLE(ve) { /* TODO: Verify the country */
    {0x33,   "ve"},
    {0x0433, "ve_ZA"},
    {0x0433, "ven_ZA"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x042a, vi, vi_VN)
ILCID_POSIX_ELEMENT_ARRAY(0x0488, wo, wo_SN)
ILCID_POSIX_ELEMENT_ARRAY(0x0434, xh, xh_ZA)

ILCID_POSIX_SUBTABLE(yi) {
    {0x003d, "yi"},
    {0x043d, "yi_001"}
};

ILCID_POSIX_ELEMENT_ARRAY(0x046a, yo, yo_NG)

// Windows & ICU tend to different names for some of these
// TODO: Windows probably does not need all of these entries, but I don't know how the precedence works.
ILCID_POSIX_SUBTABLE(zh) {
    {0x0004, "zh_Hans"},
    {0x7804, "zh"},
    {0x0804, "zh_CN"},
    {0x0804, "zh_Hans_CN"},
    {0x0c04, "zh_Hant_HK"},
    {0x0c04, "zh_HK"},
    {0x1404, "zh_Hant_MO"},
    {0x1404, "zh_MO"},
    {0x1004, "zh_Hans_SG"},
    {0x1004, "zh_SG"},
    {0x0404, "zh_Hant_TW"},
    {0x7c04, "zh_Hant"},
    {0x0404, "zh_TW"},
    {0x30404,"zh_Hant_TW"},     /* Bopomofo order */
    {0x30404,"zh_TW"},          /* Bopomofo order */
    {0x20004,"zh@collation=stroke"},
    {0x20404,"zh_Hant@collation=stroke"},
    {0x20404,"zh_Hant_TW@collation=stroke"},
    {0x20404,"zh_TW@collation=stroke"},
    {0x20804,"zh_Hans@collation=stroke"},
    {0x20804,"zh_Hans_CN@collation=stroke"},
    {0x20804,"zh_CN@collation=stroke"}
    // TODO: Alternate collations for other LCIDs are missing, eg: 0x50804
};

ILCID_POSIX_ELEMENT_ARRAY(0x0435, zu, zu_ZA)

/* This must be static and grouped by LCID. */
static const ILcidPosixMap gPosixIDmap[] = {
    ILCID_POSIX_MAP(af),    /*  af  Afrikaans                 0x36 */
    ILCID_POSIX_MAP(am),    /*  am  Amharic                   0x5e */
    ILCID_POSIX_MAP(ar),    /*  ar  Arabic                    0x01 */
    ILCID_POSIX_MAP(arn),   /*  arn Araucanian/Mapudungun     0x7a */
    ILCID_POSIX_MAP(as),    /*  as  Assamese                  0x4d */
    ILCID_POSIX_MAP(az),    /*  az  Azerbaijani               0x2c */
    ILCID_POSIX_MAP(ba),    /*  ba  Bashkir                   0x6d */
    ILCID_POSIX_MAP(be),    /*  be  Belarusian                0x23 */
/*    ILCID_POSIX_MAP(ber),     ber Berber/Tamazight          0x5f */
    ILCID_POSIX_MAP(bg),    /*  bg  Bulgarian                 0x02 */
    ILCID_POSIX_MAP(bin),   /*  bin Edo                       0x66 */
    ILCID_POSIX_MAP(bn),    /*  bn  Bengali; Bangla           0x45 */
    ILCID_POSIX_MAP(bo),    /*  bo  Tibetan                   0x51 */
    ILCID_POSIX_MAP(br),    /*  br  Breton                    0x7e */
    ILCID_POSIX_MAP(ca),    /*  ca  Catalan                   0x03 */
    ILCID_POSIX_MAP(chr),   /*  chr Cherokee                  0x5c */
    ILCID_POSIX_MAP(ckb),   /*  ckb Sorani (Central Kurdish)  0x92 */
    ILCID_POSIX_MAP(co),    /*  co  Corsican                  0x83 */
    ILCID_POSIX_MAP(cs),    /*  cs  Czech                     0x05 */
    ILCID_POSIX_MAP(cy),    /*  cy  Welsh                     0x52 */
    ILCID_POSIX_MAP(da),    /*  da  Danish                    0x06 */
    ILCID_POSIX_MAP(de),    /*  de  German                    0x07 */
    ILCID_POSIX_MAP(dv),    /*  dv  Divehi                    0x65 */
    ILCID_POSIX_MAP(el),    /*  el  Greek                     0x08 */
    ILCID_POSIX_MAP(en),    /*  en  English                   0x09 */
    ILCID_POSIX_MAP(en_US_POSIX), /*    invariant             0x7f */
    ILCID_POSIX_MAP(es),    /*  es  Spanish                   0x0a */
    ILCID_POSIX_MAP(et),    /*  et  Estonian                  0x25 */
    ILCID_POSIX_MAP(eu),    /*  eu  Basque                    0x2d */
    ILCID_POSIX_MAP(fa),    /*  fa  Persian/Farsi             0x29 */
    ILCID_POSIX_MAP(fa_AF), /*  fa  Persian/Dari              0x8c */
    ILCID_POSIX_MAP(ff),    /*  ff  Fula                      0x67 */
    ILCID_POSIX_MAP(fi),    /*  fi  Finnish                   0x0b */
    ILCID_POSIX_MAP(fil),   /*  fil Filipino                  0x64 */
    ILCID_POSIX_MAP(fo),    /*  fo  Faroese                   0x38 */
    ILCID_POSIX_MAP(fr),    /*  fr  French                    0x0c */
    ILCID_POSIX_MAP(fuv),   /*  fuv Fulfulde - Nigeria        0x67 */
    ILCID_POSIX_MAP(fy),    /*  fy  Frisian                   0x62 */
    ILCID_POSIX_MAP(ga),    /*  *   Gaelic (Ireland,Scotland) 0x3c */
    ILCID_POSIX_MAP(gd),    /*  gd  Gaelic (United Kingdom)   0x91 */
    ILCID_POSIX_MAP(gl),    /*  gl  Galician                  0x56 */
    ILCID_POSIX_MAP(gn),    /*  gn  Guarani                   0x74 */
    ILCID_POSIX_MAP(gsw),   /*  gsw Alemanic/Alsatian/Swiss German 0x84 */
    ILCID_POSIX_MAP(gu),    /*  gu  Gujarati                  0x47 */
    ILCID_POSIX_MAP(ha),    /*  ha  Hausa                     0x68 */
    ILCID_POSIX_MAP(haw),   /*  haw Hawaiian                  0x75 */
    ILCID_POSIX_MAP(he),    /*  he  Hebrew (formerly iw)      0x0d */
    ILCID_POSIX_MAP(hi),    /*  hi  Hindi                     0x39 */
    ILCID_POSIX_MAP(hr),    /*  *   Croatian and others       0x1a */
    ILCID_POSIX_MAP(hsb),   /*  hsb Upper Sorbian             0x2e */
    ILCID_POSIX_MAP(hu),    /*  hu  Hungarian                 0x0e */
    ILCID_POSIX_MAP(hy),    /*  hy  Armenian                  0x2b */
    ILCID_POSIX_MAP(ibb),   /*  ibb Ibibio - Nigeria          0x69 */
    ILCID_POSIX_MAP(id),    /*  id  Indonesian (formerly in)  0x21 */
    ILCID_POSIX_MAP(ig),    /*  ig  Igbo                      0x70 */
    ILCID_POSIX_MAP(ii),    /*  ii  Sichuan Yi                0x78 */
    ILCID_POSIX_MAP(is),    /*  is  Icelandic                 0x0f */
    ILCID_POSIX_MAP(it),    /*  it  Italian                   0x10 */
    ILCID_POSIX_MAP(iu),    /*  iu  Inuktitut                 0x5d */
    ILCID_POSIX_MAP(iw),    /*  iw  Hebrew                    0x0d */
    ILCID_POSIX_MAP(ja),    /*  ja  Japanese                  0x11 */
    ILCID_POSIX_MAP(ka),    /*  ka  Georgian                  0x37 */
    ILCID_POSIX_MAP(kk),    /*  kk  Kazakh                    0x3f */
    ILCID_POSIX_MAP(kl),    /*  kl  Kalaallisut               0x6f */
    ILCID_POSIX_MAP(km),    /*  km  Khmer                     0x53 */
    ILCID_POSIX_MAP(kn),    /*  kn  Kannada                   0x4b */
    ILCID_POSIX_MAP(ko),    /*  ko  Korean                    0x12 */
    ILCID_POSIX_MAP(kok),   /*  kok Konkani                   0x57 */
    ILCID_POSIX_MAP(kr),    /*  kr  Kanuri                    0x71 */
    ILCID_POSIX_MAP(ks),    /*  ks  Kashmiri                  0x60 */
    ILCID_POSIX_MAP(ky),    /*  ky  Kyrgyz                    0x40 */
    ILCID_POSIX_MAP(lb),    /*  lb  Luxembourgish             0x6e */
    ILCID_POSIX_MAP(la),    /*  la  Latin                     0x76 */
    ILCID_POSIX_MAP(lo),    /*  lo  Lao                       0x54 */
    ILCID_POSIX_MAP(lt),    /*  lt  Lithuanian                0x27 */
    ILCID_POSIX_MAP(lv),    /*  lv  Latvian, Lettish          0x26 */
    ILCID_POSIX_MAP(mi),    /*  mi  Maori                     0x81 */
    ILCID_POSIX_MAP(mk),    /*  mk  Macedonian                0x2f */
    ILCID_POSIX_MAP(ml),    /*  ml  Malayalam                 0x4c */
    ILCID_POSIX_MAP(mn),    /*  mn  Mongolian                 0x50 */
    ILCID_POSIX_MAP(mni),   /*  mni Manipuri                  0x58 */
    ILCID_POSIX_MAP(moh),   /*  moh Mohawk                    0x7c */
    ILCID_POSIX_MAP(mr),    /*  mr  Marathi                   0x4e */
    ILCID_POSIX_MAP(ms),    /*  ms  Malay                     0x3e */
    ILCID_POSIX_MAP(mt),    /*  mt  Maltese                   0x3a */
    ILCID_POSIX_MAP(my),    /*  my  Burmese                   0x55 */
/*    ILCID_POSIX_MAP(nb),    //  no  Norwegian                 0x14 */
    ILCID_POSIX_MAP(ne),    /*  ne  Nepali                    0x61 */
    ILCID_POSIX_MAP(nl),    /*  nl  Dutch                     0x13 */
/*    ILCID_POSIX_MAP(nn),    //  no  Norwegian                 0x14 */
    ILCID_POSIX_MAP(no),    /*  *   Norwegian                 0x14 */
    ILCID_POSIX_MAP(nso),   /*  nso Sotho, Northern (Sepedi dialect) 0x6c */
    ILCID_POSIX_MAP(oc),    /*  oc  Occitan                   0x82 */
    ILCID_POSIX_MAP(om),    /*  om  Oromo                     0x72 */
    ILCID_POSIX_MAP(or_IN), /*  or  Oriya                     0x48 */
    ILCID_POSIX_MAP(pa),    /*  pa  Punjabi                   0x46 */
    ILCID_POSIX_MAP(pap),   /*  pap Papiamentu                0x79 */
    ILCID_POSIX_MAP(pl),    /*  pl  Polish                    0x15 */
    ILCID_POSIX_MAP(ps),    /*  ps  Pashto                    0x63 */
    ILCID_POSIX_MAP(pt),    /*  pt  Portuguese                0x16 */
    ILCID_POSIX_MAP(qu),    /*  qu  Quechua                   0x6B */
    ILCID_POSIX_MAP(quc),   /*  quc K'iche                    0x93 */
    ILCID_POSIX_MAP(qut),   /*  qut K'iche                    0x86 */
    ILCID_POSIX_MAP(rm),    /*  rm  Raeto-Romance/Romansh     0x17 */
    ILCID_POSIX_MAP(ro),    /*  ro  Romanian                  0x18 */
    ILCID_POSIX_MAP(root),  /*  root                          0x00 */
    ILCID_POSIX_MAP(ru),    /*  ru  Russian                   0x19 */
    ILCID_POSIX_MAP(rw),    /*  rw  Kinyarwanda               0x87 */
    ILCID_POSIX_MAP(sa),    /*  sa  Sanskrit                  0x4f */
    ILCID_POSIX_MAP(sah),   /*  sah Yakut                     0x85 */
    ILCID_POSIX_MAP(sd),    /*  sd  Sindhi                    0x59 */
    ILCID_POSIX_MAP(se),    /*  se  Sami                      0x3b */
/*    ILCID_POSIX_MAP(sh),    //  sh  Serbo-Croatian            0x1a */
    ILCID_POSIX_MAP(si),    /*  si  Sinhalese                 0x5b */
    ILCID_POSIX_MAP(sk),    /*  sk  Slovak                    0x1b */
    ILCID_POSIX_MAP(sl),    /*  sl  Slovenian                 0x24 */
    ILCID_POSIX_MAP(so),    /*  so  Somali                    0x77 */
    ILCID_POSIX_MAP(sq),    /*  sq  Albanian                  0x1c */
/*    ILCID_POSIX_MAP(sr),    //  sr  Serbian                   0x1a */
    ILCID_POSIX_MAP(st),    /*  st  Sutu                      0x30 */
    ILCID_POSIX_MAP(sv),    /*  sv  Swedish                   0x1d */
    ILCID_POSIX_MAP(sw),    /*  sw  Swahili                   0x41 */
    ILCID_POSIX_MAP(syr),   /*  syr Syriac                    0x5A */
    ILCID_POSIX_MAP(ta),    /*  ta  Tamil                     0x49 */
    ILCID_POSIX_MAP(te),    /*  te  Telugu                    0x4a */
    ILCID_POSIX_MAP(tg),    /*  tg  Tajik                     0x28 */
    ILCID_POSIX_MAP(th),    /*  th  Thai                      0x1e */
    ILCID_POSIX_MAP(ti),    /*  ti  Tigrigna                  0x73 */
    ILCID_POSIX_MAP(tk),    /*  tk  Turkmen                   0x42 */
    ILCID_POSIX_MAP(tn),    /*  tn  Tswana                    0x32 */
    ILCID_POSIX_MAP(tr),    /*  tr  Turkish                   0x1f */
    ILCID_POSIX_MAP(ts),    /*  ts  Tsonga                    0x31 */
    ILCID_POSIX_MAP(tt),    /*  tt  Tatar                     0x44 */
    ILCID_POSIX_MAP(tzm),   /*  tzm Tamazight                 0x5f */
    ILCID_POSIX_MAP(ug),    /*  ug  Uighur                    0x80 */
    ILCID_POSIX_MAP(uk),    /*  uk  Ukrainian                 0x22 */
    ILCID_POSIX_MAP(ur),    /*  ur  Urdu                      0x20 */
    ILCID_POSIX_MAP(uz),    /*  uz  Uzbek                     0x43 */
    ILCID_POSIX_MAP(ve),    /*  ve  Venda                     0x33 */
    ILCID_POSIX_MAP(vi),    /*  vi  Vietnamese                0x2a */
    ILCID_POSIX_MAP(wo),    /*  wo  Wolof                     0x88 */
    ILCID_POSIX_MAP(xh),    /*  xh  Xhosa                     0x34 */
    ILCID_POSIX_MAP(yi),    /*  yi  Yiddish                   0x3d */
    ILCID_POSIX_MAP(yo),    /*  yo  Yoruba                    0x6a */
    ILCID_POSIX_MAP(zh),    /*  zh  Chinese                   0x04 */
    ILCID_POSIX_MAP(zu),    /*  zu  Zulu                      0x35 */
};

static const uint32_t gLocaleCount = UPRV_LENGTHOF(gPosixIDmap);

/**
 * Do not call this function. It is called by hostID.
 * The function is not private because this struct must stay as a C struct,
 * and this is an internal class.
 */
static int32_t
idCmp(const char* id1, const char* id2)
{
    int32_t diffIdx = 0;
    while (*id1 == *id2 && *id1 != 0) {
        diffIdx++;
        id1++;
        id2++;
    }
    return diffIdx;
}

/**
 * Searches for a Windows LCID
 *
 * @param posixID the Posix style locale id.
 * @param status gets set to U_ILLEGAL_ARGUMENT_ERROR when the Posix ID has
 *               no equivalent Windows LCID.
 * @return the LCID
 */
static uint32_t
getHostID(const ILcidPosixMap *this_0, const char* posixID, UErrorCode* status)
{
    int32_t bestIdx = 0;
    int32_t bestIdxDiff = 0;
    int32_t posixIDlen = (int32_t)uprv_strlen(posixID);
    uint32_t idx;

    for (idx = 0; idx < this_0->numRegions; idx++ ) {
        int32_t sameChars = idCmp(posixID, this_0->regionMaps[idx].posixID);
        if (sameChars > bestIdxDiff && this_0->regionMaps[idx].posixID[sameChars] == 0) {
            if (posixIDlen == sameChars) {
                /* Exact match */
                return this_0->regionMaps[idx].hostID;
            }
            bestIdxDiff = sameChars;
            bestIdx = idx;
        }
    }
    /* We asked for something unusual, like en_ZZ, and we try to return the number for the same language. */
    /* We also have to make sure that sid and si and similar string subsets don't match. */
    if ((posixID[bestIdxDiff] == '_' || posixID[bestIdxDiff] == '@')
        && this_0->regionMaps[bestIdx].posixID[bestIdxDiff] == 0)
    {
        *status = U_USING_FALLBACK_WARNING;
        return this_0->regionMaps[bestIdx].hostID;
    }

    /*no match found */
    *status = U_ILLEGAL_ARGUMENT_ERROR;
    return this_0->regionMaps->hostID;
}

static const char*
getPosixID(const ILcidPosixMap *this_0, uint32_t hostID)
{
    uint32_t i;
    for (i = 0; i < this_0->numRegions; i++)
    {
        if (this_0->regionMaps[i].hostID == hostID)
        {
            return this_0->regionMaps[i].posixID;
        }
    }

    /* If you get here, then no matching region was found,
       so return the language id with the wild card region. */
    return this_0->regionMaps[0].posixID;
}

/*
//////////////////////////////////////
//
// LCID --> POSIX
//
/////////////////////////////////////
*/
#if U_PLATFORM_HAS_WIN32_API && UCONFIG_USE_WINDOWS_LCID_MAPPING_API
/*
 * Various language tags needs to be changed:
 * quz -> qu
 * prs -> fa
 */
#define FIX_LANGUAGE_ID_TAG(buffer, len) \
    if (len >= 3) { \
        if (buffer[0] == 'q' && buffer[1] == 'u' && buffer[2] == 'z') {\
            buffer[2] = 0; \
            uprv_strcat(buffer, buffer+3); \
        } else if (buffer[0] == 'p' && buffer[1] == 'r' && buffer[2] == 's') {\
            buffer[0] = 'f'; buffer[1] = 'a'; buffer[2] = 0; \
            uprv_strcat(buffer, buffer+3); \
        } \
    }

#endif

U_CAPI int32_t
uprv_convertToPosix(uint32_t hostid, char *posixID, int32_t posixIDCapacity, UErrorCode* status)
{
    uint16_t langID;
    uint32_t localeIndex;
    UBool bLookup = TRUE;
    const char *pPosixID = NULL;

#if U_PLATFORM_HAS_WIN32_API && UCONFIG_USE_WINDOWS_LCID_MAPPING_API
    static_assert(ULOC_FULLNAME_CAPACITY > LOCALE_NAME_MAX_LENGTH, "Windows locale names have smaller length than ICU locale names.");

    char locName[LOCALE_NAME_MAX_LENGTH] = {};

    // Note: Windows primary lang ID 0x92 in LCID is used for Central Kurdish and
    // GetLocaleInfo() maps such LCID to "ku". However, CLDR uses "ku" for
    // Northern Kurdish and "ckb" for Central Kurdish. For this reason, we cannot
    // use the Windows API to resolve locale ID for this specific case.
    if ((hostid & 0x3FF) != 0x92) {
        int32_t tmpLen = 0;
        char16_t windowsLocaleName[LOCALE_NAME_MAX_LENGTH] = {};

        // Note: LOCALE_ALLOW_NEUTRAL_NAMES was enabled in Windows7+, prior versions did not handle neutral (no-region) locale names.
        tmpLen = LCIDToLocaleName(hostid, (PWSTR)windowsLocaleName, UPRV_LENGTHOF(windowsLocaleName), LOCALE_ALLOW_NEUTRAL_NAMES);
        if (tmpLen > 1) {
            int32_t i = 0;
            // Only need to look up in table if have _, eg for de-de_phoneb type alternate sort.
            bLookup = FALSE;
            for (i = 0; i < UPRV_LENGTHOF(locName); i++)
            {
                locName[i] = (char)(windowsLocaleName[i]);

                // Windows locale name may contain sorting variant, such as "es-ES_tradnl".
                // In such cases, we need special mapping data found in the hardcoded table
                // in this source file.
                if (windowsLocaleName[i] == L'_')
                {
                    // Keep the base locale, without variant
                    // TODO: Should these be mapped from _phoneb to @collation=phonebook, etc.?
                    locName[i] = '\0';
                    tmpLen = i;
                    bLookup = TRUE;
                    break;
                }
                else if (windowsLocaleName[i] == L'-')
                {
                    // Windows names use -, ICU uses _
                    locName[i] = '_';
                }
                else if (windowsLocaleName[i] == L'\0')
                {
                    // No point in doing more work than necessary
                    break;
                }
            }
            // TODO: Need to understand this better, why isn't it an alias?
            FIX_LANGUAGE_ID_TAG(locName, tmpLen);
            pPosixID = locName;
        }
    }
#endif

    if (bLookup) {
        const char *pCandidate = NULL;
        langID = LANGUAGE_LCID(hostid);

        for (localeIndex = 0; localeIndex < gLocaleCount; localeIndex++) {
            if (langID == gPosixIDmap[localeIndex].regionMaps->hostID) {
                pCandidate = getPosixID(&gPosixIDmap[localeIndex], hostid);
                break;
            }
        }

        /* On Windows, when locale name has a variant, we still look up the hardcoded table.
           If a match in the hardcoded table is longer than the Windows locale name without
           variant, we use the one as the result */
        if (pCandidate && (pPosixID == NULL || uprv_strlen(pCandidate) > uprv_strlen(pPosixID))) {
            pPosixID = pCandidate;
        }
    }

    if (pPosixID) {
        int32_t resLen = static_cast<int32_t>(uprv_strlen(pPosixID));
        int32_t copyLen = resLen <= posixIDCapacity ? resLen : posixIDCapacity;
        uprv_memcpy(posixID, pPosixID, copyLen);
        if (resLen < posixIDCapacity) {
            posixID[resLen] = 0;
            if (*status == U_STRING_NOT_TERMINATED_WARNING) {
                *status = U_ZERO_ERROR;
            }
        } else if (resLen == posixIDCapacity) {
            *status = U_STRING_NOT_TERMINATED_WARNING;
        } else {
            *status = U_BUFFER_OVERFLOW_ERROR;
        }
        return resLen;
    }

    /* no match found */
    *status = U_ILLEGAL_ARGUMENT_ERROR;
    return -1;
}

/*
//////////////////////////////////////
//
// POSIX --> LCID
// This should only be called from uloc_getLCID.
// The locale ID must be in canonical form.
//
/////////////////////////////////////
*/
U_CAPI uint32_t
uprv_convertToLCIDPlatform(const char* localeID, UErrorCode* status)
{
    if (U_FAILURE(*status)) {
        return 0;
    }

    // The purpose of this function is to leverage the Windows platform name->lcid
    // conversion functionality when available.
#if U_PLATFORM_HAS_WIN32_API && UCONFIG_USE_WINDOWS_LCID_MAPPING_API
    int32_t len;
    char collVal[ULOC_KEYWORDS_CAPACITY] = {};
    char baseName[ULOC_FULLNAME_CAPACITY] = {};
    const char * mylocaleID = localeID;

    // Check any for keywords.
    if (uprv_strchr(localeID, '@'))
    {
        len = uloc_getKeywordValue(localeID, "collation", collVal, UPRV_LENGTHOF(collVal) - 1, status);
        if (U_SUCCESS(*status) && len > 0)
        {
            // If it contains the keyword collation, return 0 so that the LCID lookup table will be used.
            return 0;
        }
        else
        {
            // If the locale ID contains keywords other than collation, just use the base name.
            len = uloc_getBaseName(localeID, baseName, UPRV_LENGTHOF(baseName) - 1, status);

            if (U_SUCCESS(*status) && len > 0)
            {
                baseName[len] = 0;
                mylocaleID = baseName;
            }
        }
    }

    char asciiBCP47Tag[LOCALE_NAME_MAX_LENGTH] = {};
    // this will change it from de_DE@collation=phonebook to de-DE-u-co-phonebk form
    (void)uloc_toLanguageTag(mylocaleID, asciiBCP47Tag, UPRV_LENGTHOF(asciiBCP47Tag), FALSE, status);

    if (U_SUCCESS(*status))
    {
        // Need it to be UTF-16, not 8-bit
        wchar_t bcp47Tag[LOCALE_NAME_MAX_LENGTH] = {};
        int32_t i;
        for (i = 0; i < UPRV_LENGTHOF(bcp47Tag); i++)
        {
            if (asciiBCP47Tag[i] == '\0')
            {
                break;
            }
            else
            {
                // Copy the character
                bcp47Tag[i] = static_cast<wchar_t>(asciiBCP47Tag[i]);
            }
        }

        if (i < (UPRV_LENGTHOF(bcp47Tag) - 1))
        {
            // Ensure it's null terminated
            bcp47Tag[i] = L'\0';
            LCID lcid = LocaleNameToLCID(bcp47Tag, LOCALE_ALLOW_NEUTRAL_NAMES);
            if (lcid > 0)
            {
                // Found LCID from windows, return that one, unless its completely ambiguous
                // LOCALE_USER_DEFAULT and transients are OK because they will round trip
                // for this process.
                if (lcid != LOCALE_CUSTOM_UNSPECIFIED)
                {
                    return lcid;
                }
            }
        }
    }
#else
    (void) localeID; // Suppress unused variable warning.
#endif

    // Nothing found, or not implemented.
    return 0;
}

U_CAPI uint32_t
uprv_convertToLCID(const char *langID, const char* posixID, UErrorCode* status)
{
    // This function does the table lookup when native platform name->lcid conversion isn't available,
    // or for locales that don't follow patterns the platform expects.
    uint32_t   low    = 0;
    uint32_t   high   = gLocaleCount;
    uint32_t   mid;
    uint32_t   oldmid = 0;
    int32_t    compVal;

    uint32_t   value         = 0;
    uint32_t   fallbackValue = (uint32_t)-1;
    UErrorCode myStatus;
    uint32_t   idx;

    /* Check for incomplete id. */
    if (!langID || !posixID || uprv_strlen(langID) < 2 || uprv_strlen(posixID) < 2) {
        return 0;
    }

    /*Binary search for the map entry for normal cases */

    while (high > low)  /*binary search*/{

        mid = (high+low) >> 1; /*Finds median*/

        if (mid == oldmid) 
            break;

        compVal = uprv_strcmp(langID, gPosixIDmap[mid].regionMaps->posixID);
        if (compVal < 0){
            high = mid;
        }
        else if (compVal > 0){
            low = mid;
        }
        else /*we found it*/{
            return getHostID(&gPosixIDmap[mid], posixID, status);
        }
        oldmid = mid;
    }

    /*
     * Sometimes we can't do a binary search on posixID because some LCIDs
     * go to different locales.  We hit one of those special cases.
     */
    for (idx = 0; idx < gLocaleCount; idx++ ) {
        myStatus = U_ZERO_ERROR;
        value = getHostID(&gPosixIDmap[idx], posixID, &myStatus);
        if (myStatus == U_ZERO_ERROR) {
            return value;
        }
        else if (myStatus == U_USING_FALLBACK_WARNING) {
            fallbackValue = value;
        }
    }

    if (fallbackValue != (uint32_t)-1) {
        *status = U_USING_FALLBACK_WARNING;
        return fallbackValue;
    }

    /* no match found */
    *status = U_ILLEGAL_ARGUMENT_ERROR;
    return 0;   /* return international (root) */
}
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1997-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************
*
*  FILE NAME : putil.c (previously putil.cpp and ptypes.cpp)
*
*   Date        Name        Description
*   04/14/97    aliu        Creation.
*   04/24/97    aliu        Added getDefaultDataDirectory() and
*                            getDefaultLocaleID().
*   04/28/97    aliu        Rewritten to assume Unix and apply general methods
*                            for assumed case.  Non-UNIX platforms must be
*                            special-cased.  Rewrote numeric methods dealing
*                            with NaN and Infinity to be platform independent
*                             over all IEEE 754 platforms.
*   05/13/97    aliu        Restored sign of timezone
*                            (semantics are hours West of GMT)
*   06/16/98    erm         Added IEEE_754 stuff, cleaned up isInfinite, isNan,
*                             nextDouble..
*   07/22/98    stephen     Added remainder, max, min, trunc
*   08/13/98    stephen     Added isNegativeInfinity, isPositiveInfinity
*   08/24/98    stephen     Added longBitsFromDouble
*   09/08/98    stephen     Minor changes for Mac Port
*   03/02/99    stephen     Removed openFile().  Added AS400 support.
*                            Fixed EBCDIC tables
*   04/15/99    stephen     Converted to C.
*   06/28/99    stephen     Removed mutex locking in u_isBigEndian().
*   08/04/99    jeffrey R.  Added OS/2 changes
*   11/15/99    helena      Integrated S/390 IEEE support.
*   04/26/01    Barry N.    OS/400 support for uprv_getDefaultLocaleID
*   08/15/01    Steven H.   OS/400 support for uprv_getDefaultCodepage
*   01/03/08    Steven L.   Fake Time Support
******************************************************************************
*/

// Defines _XOPEN_SOURCE for access to POSIX functions.
// Must be before any other #includes.
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
*   Copyright (C) 2011-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
*******************************************************************************
*   file name:  uposixdefs.h
*   encoding:   UTF-8
*   tab size:   8 (not used)
*   indentation:4
*
*   created on: 2011jul25
*   created by: Markus W. Scherer
*
*   Common definitions for implementation files working with POSIX functions.
*   *Important*: #include this file before any other header files!
*/


/*
 * Define _XOPEN_SOURCE for access to POSIX functions.
 *
 * We cannot use U_PLATFORM from platform.h/utypes.h because
 * "The Open Group Base Specifications"
 * chapter "2.2 The Compilation Environment" says:
 * "In the compilation of an application that #defines a feature test macro
 * specified by IEEE Std 1003.1-2001,
 * no header defined by IEEE Std 1003.1-2001 shall be included prior to
 * the definition of the feature test macro."
 */
#ifdef _XOPEN_SOURCE
    /* Use the predefined value. */
#else
    /*
     * Version 6.0:
     * The Open Group Base Specifications Issue 6 (IEEE Std 1003.1, 2004 Edition)
     * also known as
     * SUSv3 = Open Group Single UNIX Specification, Version 3 (UNIX03)
     *
     * Note: This definition used to be in C source code (e.g., putil.c)
     * and define _XOPEN_SOURCE to different values depending on __STDC_VERSION__.
     * In C++ source code (e.g., putil.cpp), __STDC_VERSION__ is not defined at all.
     */
#   define _XOPEN_SOURCE 600
#endif

/*
 * Make sure things like readlink and such functions work.
 * Poorly upgraded Solaris machines can't have this defined.
 * Cleanly installed Solaris can use this #define.
 *
 * z/OS needs this definition for timeval and to get usleep.
 */
#if !defined(_XOPEN_SOURCE_EXTENDED) && defined(__TOS_MVS__)
#   define _XOPEN_SOURCE_EXTENDED 1
#endif

/**
 * Solaris says:
 *   "...it is invalid to compile an XPG6 or a POSIX.1-2001 application with anything other
 *   than a c99 or later compiler."
 * Apparently C++11 is not "or later". Work around this.
 */
#if defined(__cplusplus) && (defined(sun) || defined(__sun)) && !defined (_STDC_C99)
#   define _STDC_C99
#endif



// First, the platform type. Need this for U_PLATFORM.


#if U_PLATFORM == U_PF_MINGW && defined __STRICT_ANSI__
/* tzset isn't defined in strict ANSI on MinGW. */
#undef __STRICT_ANSI__
#endif

/*
 * Cygwin with GCC requires inclusion of time.h after the above disabling strict asci mode statement.
 */
#include <time.h>

#if !U_PLATFORM_USES_ONLY_WIN32_API
#include <sys/time.h>
#endif

/* include the rest of the ICU headers */











/* Include standard headers. */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <locale.h>
#include <float.h>

#ifndef U_COMMON_IMPLEMENTATION
#error U_COMMON_IMPLEMENTATION not set - must be set for all ICU source files in common/ - see http://userguide.icu-project.org/howtouseicu
#endif


/* include system headers */
#if U_PLATFORM_USES_ONLY_WIN32_API
    /*
     * TODO: U_PLATFORM_USES_ONLY_WIN32_API includes MinGW.
     * Should Cygwin be included as well (U_PLATFORM_HAS_WIN32_API)
     * to use native APIs as much as possible?
     */
#ifndef WIN32_LEAN_AND_MEAN
#   define WIN32_LEAN_AND_MEAN
#endif
#   define VC_EXTRALEAN
#   define NOUSER
#   define NOSERVICE
#   define NOIME
#   define NOMCX
#   include <windows.h>

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 2005-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File WINTZ.H
*
********************************************************************************
*/




#if U_PLATFORM_USES_ONLY_WIN32_API

/**
 * \file
 * \brief C API: Utilities for dealing w/ Windows time zones.
 */

U_CDECL_BEGIN
/* Forward declarations for Windows types... */
typedef struct _TIME_ZONE_INFORMATION TIME_ZONE_INFORMATION;
U_CDECL_END

U_INTERNAL const char* U_EXPORT2
uprv_detectWindowsTimeZone();

#endif /* U_PLATFORM_USES_ONLY_WIN32_API  */


#elif U_PLATFORM == U_PF_OS400
#   include <float.h>
#   include <qusec.h>       /* error code structure */
#   include <qusrjobi.h>
#   include <qliept.h>      /* EPT_CALL macro  - this include must be after all other "QSYSINCs" */
#   include <mih/testptr.h> /* For uprv_maximumPtr */
#elif U_PLATFORM == U_PF_OS390
   /* Needed for UCNV_SWAP_LFNL_OPTION_STRING */
#elif U_PLATFORM_IS_DARWIN_BASED || U_PLATFORM_IS_LINUX_BASED || U_PLATFORM == U_PF_BSD || U_PLATFORM == U_PF_SOLARIS
#   include <limits.h>
#   include <unistd.h>
#   if U_PLATFORM == U_PF_SOLARIS
#       ifndef _XPG4_2
#           define _XPG4_2
#       endif
#   endif
#elif U_PLATFORM == U_PF_QNX
#   include <sys/neutrino.h>
#endif

/*
 * Only include langinfo.h if we have a way to get the codeset. If we later
 * depend on more feature, we can test on U_HAVE_NL_LANGINFO.
 *
 */

#if U_HAVE_NL_LANGINFO_CODESET
#include <langinfo.h>
#endif

/**
 * Simple things (presence of functions, etc) should just go in configure.in and be added to
 * icucfg.h via autoheader.
 */
#if U_PLATFORM_IMPLEMENTS_POSIX
#   if U_PLATFORM == U_PF_OS400
#    define HAVE_DLFCN_H 0
#    define HAVE_DLOPEN 0
#   else
#   ifndef HAVE_DLFCN_H
#    define HAVE_DLFCN_H 1
#   endif
#   ifndef HAVE_DLOPEN
#    define HAVE_DLOPEN 1
#   endif
#   endif
#   ifndef HAVE_GETTIMEOFDAY
#    define HAVE_GETTIMEOFDAY 1
#   endif
#else
#   define HAVE_DLFCN_H 0
#   define HAVE_DLOPEN 0
#   define HAVE_GETTIMEOFDAY 0
#endif

U_NAMESPACE_USE

/* Define the extension for data files, again... */
#define DATA_TYPE "dat"

/* Leave this copyright notice here! */
// static const char copyright[] = U_COPYRIGHT_STRING;

/* floating point implementations ------------------------------------------- */

/* We return QNAN rather than SNAN*/
#define SIGN 0x80000000U

/* Make it easy to define certain types of constants */
typedef union {
    int64_t i64; /* This must be defined first in order to allow the initialization to work. This is a C89 feature. */
    double d64;
} BitPatternConversion;
static const BitPatternConversion gNan = { (int64_t) INT64_C(0x7FF8000000000000) };
static const BitPatternConversion gInf = { (int64_t) INT64_C(0x7FF0000000000000) };

/*---------------------------------------------------------------------------
  Platform utilities
  Our general strategy is to assume we're on a POSIX platform.  Platforms which
  are non-POSIX must declare themselves so.  The default POSIX implementation
  will sometimes work for non-POSIX platforms as well (e.g., the NaN-related
  functions).
  ---------------------------------------------------------------------------*/

#if U_PLATFORM_USES_ONLY_WIN32_API || U_PLATFORM == U_PF_OS400
#   undef U_POSIX_LOCALE
#else
#   define U_POSIX_LOCALE    1
#endif

/*
    WARNING! u_topNBytesOfDouble and u_bottomNBytesOfDouble
    can't be properly optimized by the gcc compiler sometimes (i.e. gcc 3.2).
*/
#if !IEEE_754
static char*
u_topNBytesOfDouble(double* d, int n)
{
#if U_IS_BIG_ENDIAN
    return (char*)d;
#else
    return (char*)(d + 1) - n;
#endif
}

static char*
u_bottomNBytesOfDouble(double* d, int n)
{
#if U_IS_BIG_ENDIAN
    return (char*)(d + 1) - n;
#else
    return (char*)d;
#endif
}
#endif   /* !IEEE_754 */

#if IEEE_754
static UBool
u_signBit(double d) {
    uint8_t hiByte;
#if U_IS_BIG_ENDIAN
    hiByte = *(uint8_t *)&d;
#else
    hiByte = *(((uint8_t *)&d) + sizeof(double) - 1);
#endif
    return (hiByte & 0x80) != 0;
}
#endif



#if defined (U_DEBUG_FAKETIME)
/* Override the clock to test things without having to move the system clock.
 * Assumes POSIX gettimeofday() will function
 */
UDate fakeClock_t0 = 0; /** Time to start the clock from **/
UDate fakeClock_dt = 0; /** Offset (fake time - real time) **/
UBool fakeClock_set = FALSE; /** True if fake clock has spun up **/

static UDate getUTCtime_real() {
    struct timeval posixTime;
    gettimeofday(&posixTime, NULL);
    return (UDate)(((int64_t)posixTime.tv_sec * U_MILLIS_PER_SECOND) + (posixTime.tv_usec/1000));
}

static UDate getUTCtime_fake() {
    static UMutex fakeClockMutex;
    umtx_lock(&fakeClockMutex);
    if(!fakeClock_set) {
        UDate real = getUTCtime_real();
        const char *fake_start = getenv("U_FAKETIME_START");
        if((fake_start!=NULL) && (fake_start[0]!=0)) {
            sscanf(fake_start,"%lf",&fakeClock_t0);
            fakeClock_dt = fakeClock_t0 - real;
            fprintf(stderr,"U_DEBUG_FAKETIME was set at compile time, so the ICU clock will start at a preset value\n"
                    "env variable U_FAKETIME_START=%.0f (%s) for an offset of %.0f ms from the current time %.0f\n",
                    fakeClock_t0, fake_start, fakeClock_dt, real);
        } else {
          fakeClock_dt = 0;
            fprintf(stderr,"U_DEBUG_FAKETIME was set at compile time, but U_FAKETIME_START was not set.\n"
                    "Set U_FAKETIME_START to the number of milliseconds since 1/1/1970 to set the ICU clock.\n");
        }
        fakeClock_set = TRUE;
    }
    umtx_unlock(&fakeClockMutex);

    return getUTCtime_real() + fakeClock_dt;
}
#endif

#if U_PLATFORM_USES_ONLY_WIN32_API
typedef union {
    int64_t int64;
    FILETIME fileTime;
} FileTimeConversion;   /* This is like a ULARGE_INTEGER */

/* Number of 100 nanoseconds from 1/1/1601 to 1/1/1970 */
#define EPOCH_BIAS  INT64_C(116444736000000000)
#define HECTONANOSECOND_PER_MILLISECOND   10000

#endif

/*---------------------------------------------------------------------------
  Universal Implementations
  These are designed to work on all platforms.  Try these, and if they
  don't work on your platform, then special case your platform with new
  implementations.
---------------------------------------------------------------------------*/

U_CAPI UDate U_EXPORT2
uprv_getUTCtime()
{
#if defined(U_DEBUG_FAKETIME)
    return getUTCtime_fake(); /* Hook for overriding the clock */
#else
    return uprv_getRawUTCtime();
#endif
}

/* Return UTC (GMT) time measured in milliseconds since 0:00 on 1/1/70.*/
U_CAPI UDate U_EXPORT2
uprv_getRawUTCtime()
{
#if U_PLATFORM_USES_ONLY_WIN32_API

    FileTimeConversion winTime;
    GetSystemTimeAsFileTime(&winTime.fileTime);
    return (UDate)((winTime.int64 - EPOCH_BIAS) / HECTONANOSECOND_PER_MILLISECOND);
#else

#if HAVE_GETTIMEOFDAY
    struct timeval posixTime;
    gettimeofday(&posixTime, NULL);
    return (UDate)(((int64_t)posixTime.tv_sec * U_MILLIS_PER_SECOND) + (posixTime.tv_usec/1000));
#else
    time_t epochtime;
    time(&epochtime);
    return (UDate)epochtime * U_MILLIS_PER_SECOND;
#endif

#endif
}

/*-----------------------------------------------------------------------------
  IEEE 754
  These methods detect and return NaN and infinity values for doubles
  conforming to IEEE 754.  Platforms which support this standard include X86,
  Mac 680x0, Mac PowerPC, AIX RS/6000, and most others.
  If this doesn't work on your platform, you have non-IEEE floating-point, and
  will need to code your own versions.  A naive implementation is to return 0.0
  for getNaN and getInfinity, and false for isNaN and isInfinite.
  ---------------------------------------------------------------------------*/

U_CAPI UBool U_EXPORT2
uprv_isNaN(double number)
{
#if IEEE_754
    BitPatternConversion convertedNumber;
    convertedNumber.d64 = number;
    /* Infinity is 0x7FF0000000000000U. Anything greater than that is a NaN */
    return (UBool)((convertedNumber.i64 & U_INT64_MAX) > gInf.i64);

#elif U_PLATFORM == U_PF_OS390
    uint32_t highBits = *(uint32_t*)u_topNBytesOfDouble(&number,
                        sizeof(uint32_t));
    uint32_t lowBits  = *(uint32_t*)u_bottomNBytesOfDouble(&number,
                        sizeof(uint32_t));

    return ((highBits & 0x7F080000L) == 0x7F080000L) &&
      (lowBits == 0x00000000L);

#else
    /* If your platform doesn't support IEEE 754 but *does* have an NaN value,*/
    /* you'll need to replace this default implementation with what's correct*/
    /* for your platform.*/
    return number != number;
#endif
}

U_CAPI UBool U_EXPORT2
uprv_isInfinite(double number)
{
#if IEEE_754
    BitPatternConversion convertedNumber;
    convertedNumber.d64 = number;
    /* Infinity is exactly 0x7FF0000000000000U. */
    return (UBool)((convertedNumber.i64 & U_INT64_MAX) == gInf.i64);
#elif U_PLATFORM == U_PF_OS390
    uint32_t highBits = *(uint32_t*)u_topNBytesOfDouble(&number,
                        sizeof(uint32_t));
    uint32_t lowBits  = *(uint32_t*)u_bottomNBytesOfDouble(&number,
                        sizeof(uint32_t));

    return ((highBits  & ~SIGN) == 0x70FF0000L) && (lowBits == 0x00000000L);

#else
    /* If your platform doesn't support IEEE 754 but *does* have an infinity*/
    /* value, you'll need to replace this default implementation with what's*/
    /* correct for your platform.*/
    return number == (2.0 * number);
#endif
}

U_CAPI UBool U_EXPORT2
uprv_isPositiveInfinity(double number)
{
#if IEEE_754 || U_PLATFORM == U_PF_OS390
    return (UBool)(number > 0 && uprv_isInfinite(number));
#else
    return uprv_isInfinite(number);
#endif
}

U_CAPI UBool U_EXPORT2
uprv_isNegativeInfinity(double number)
{
#if IEEE_754 || U_PLATFORM == U_PF_OS390
    return (UBool)(number < 0 && uprv_isInfinite(number));

#else
    uint32_t highBits = *(uint32_t*)u_topNBytesOfDouble(&number,
                        sizeof(uint32_t));
    return((highBits & SIGN) && uprv_isInfinite(number));

#endif
}

U_CAPI double U_EXPORT2
uprv_getNaN()
{
#if IEEE_754 || U_PLATFORM == U_PF_OS390
    return gNan.d64;
#else
    /* If your platform doesn't support IEEE 754 but *does* have an NaN value,*/
    /* you'll need to replace this default implementation with what's correct*/
    /* for your platform.*/
    return 0.0;
#endif
}

U_CAPI double U_EXPORT2
uprv_getInfinity()
{
#if IEEE_754 || U_PLATFORM == U_PF_OS390
    return gInf.d64;
#else
    /* If your platform doesn't support IEEE 754 but *does* have an infinity*/
    /* value, you'll need to replace this default implementation with what's*/
    /* correct for your platform.*/
    return 0.0;
#endif
}

U_CAPI double U_EXPORT2
uprv_floor(double x)
{
    return floor(x);
}

U_CAPI double U_EXPORT2
uprv_ceil(double x)
{
    return ceil(x);
}

U_CAPI double U_EXPORT2
uprv_round(double x)
{
    return uprv_floor(x + 0.5);
}

U_CAPI double U_EXPORT2
uprv_fabs(double x)
{
    return fabs(x);
}

U_CAPI double U_EXPORT2
uprv_modf(double x, double* y)
{
    return modf(x, y);
}

U_CAPI double U_EXPORT2
uprv_fmod(double x, double y)
{
    return fmod(x, y);
}

U_CAPI double U_EXPORT2
uprv_pow(double x, double y)
{
    /* This is declared as "double pow(double x, double y)" */
    return pow(x, y);
}

U_CAPI double U_EXPORT2
uprv_pow10(int32_t x)
{
    return pow(10.0, (double)x);
}

U_CAPI double U_EXPORT2
uprv_fmax(double x, double y)
{
#if IEEE_754
    /* first handle NaN*/
    if(uprv_isNaN(x) || uprv_isNaN(y))
        return uprv_getNaN();

    /* check for -0 and 0*/
    if(x == 0.0 && y == 0.0 && u_signBit(x))
        return y;

#endif

    /* this should work for all flt point w/o NaN and Inf special cases */
    return (x > y ? x : y);
}

U_CAPI double U_EXPORT2
uprv_fmin(double x, double y)
{
#if IEEE_754
    /* first handle NaN*/
    if(uprv_isNaN(x) || uprv_isNaN(y))
        return uprv_getNaN();

    /* check for -0 and 0*/
    if(x == 0.0 && y == 0.0 && u_signBit(y))
        return y;

#endif

    /* this should work for all flt point w/o NaN and Inf special cases */
    return (x > y ? y : x);
}

U_CAPI UBool U_EXPORT2
uprv_add32_overflow(int32_t a, int32_t b, int32_t* res) {
    // NOTE: Some compilers (GCC, Clang) have primitives available, like __builtin_add_overflow.
    // This function could be optimized by calling one of those primitives.
    auto a64 = static_cast<int64_t>(a);
    auto b64 = static_cast<int64_t>(b);
    int64_t res64 = a64 + b64;
    *res = static_cast<int32_t>(res64);
    return res64 != *res;
}

U_CAPI UBool U_EXPORT2
uprv_mul32_overflow(int32_t a, int32_t b, int32_t* res) {
    // NOTE: Some compilers (GCC, Clang) have primitives available, like __builtin_mul_overflow.
    // This function could be optimized by calling one of those primitives.
    auto a64 = static_cast<int64_t>(a);
    auto b64 = static_cast<int64_t>(b);
    int64_t res64 = a64 * b64;
    *res = static_cast<int32_t>(res64);
    return res64 != *res;
}

/**
 * Truncates the given double.
 * trunc(3.3) = 3.0, trunc (-3.3) = -3.0
 * This is different than calling floor() or ceil():
 * floor(3.3) = 3, floor(-3.3) = -4
 * ceil(3.3) = 4, ceil(-3.3) = -3
 */
U_CAPI double U_EXPORT2
uprv_trunc(double d)
{
#if IEEE_754
    /* handle error cases*/
    if(uprv_isNaN(d))
        return uprv_getNaN();
    if(uprv_isInfinite(d))
        return uprv_getInfinity();

    if(u_signBit(d))    /* Signbit() picks up -0.0;  d<0 does not. */
        return ceil(d);
    else
        return floor(d);

#else
    return d >= 0 ? floor(d) : ceil(d);

#endif
}

/**
 * Return the largest positive number that can be represented by an integer
 * type of arbitrary bit length.
 */
U_CAPI double U_EXPORT2
uprv_maxMantissa(void)
{
    return pow(2.0, DBL_MANT_DIG + 1.0) - 1.0;
}

U_CAPI double U_EXPORT2
uprv_log(double d)
{
    return log(d);
}

U_CAPI void * U_EXPORT2
uprv_maximumPtr(void * base)
{
#if U_PLATFORM == U_PF_OS400
    /*
     * With the provided function we should never be out of range of a given segment
     * (a traditional/typical segment that is).  Our segments have 5 bytes for the
     * id and 3 bytes for the offset.  The key is that the casting takes care of
     * only retrieving the offset portion minus x1000.  Hence, the smallest offset
     * seen in a program is x001000 and when casted to an int would be 0.
     * That's why we can only add 0xffefff.  Otherwise, we would exceed the segment.
     *
     * Currently, 16MB is the current addressing limitation on i5/OS if the activation is
     * non-TERASPACE.  If it is TERASPACE it is 2GB - 4k(header information).
     * This function determines the activation based on the pointer that is passed in and
     * calculates the appropriate maximum available size for
     * each pointer type (TERASPACE and non-TERASPACE)
     *
     * Unlike other operating systems, the pointer model isn't determined at
     * compile time on i5/OS.
     */
    if ((base != NULL) && (_TESTPTR(base, _C_TERASPACE_CHECK))) {
        /* if it is a TERASPACE pointer the max is 2GB - 4k */
        return ((void *)(((char *)base)-((uint32_t)(base))+((uint32_t)0x7fffefff)));
    }
    /* otherwise 16MB since NULL ptr is not checkable or the ptr is not TERASPACE */
    return ((void *)(((char *)base)-((uint32_t)(base))+((uint32_t)0xffefff)));

#else
    return U_MAX_PTR(base);
#endif
}

/*---------------------------------------------------------------------------
  Platform-specific Implementations
  Try these, and if they don't work on your platform, then special case your
  platform with new implementations.
  ---------------------------------------------------------------------------*/

/* Generic time zone layer -------------------------------------------------- */

/* Time zone utilities */
U_CAPI void U_EXPORT2
uprv_tzset()
{
#if defined(U_TZSET)
    U_TZSET();
#else
    /* no initialization*/
#endif
}

U_CAPI int32_t U_EXPORT2
uprv_timezone()
{
#ifdef U_TIMEZONE
    return U_TIMEZONE;
#else
    time_t t, t1, t2;
    struct tm tmrec;
    int32_t tdiff = 0;

    time(&t);
    uprv_memcpy( &tmrec, localtime(&t), sizeof(tmrec) );
#if U_PLATFORM != U_PF_IPHONE
    UBool dst_checked = (tmrec.tm_isdst != 0); /* daylight savings time is checked*/
#endif
    t1 = mktime(&tmrec);                 /* local time in seconds*/
    uprv_memcpy( &tmrec, gmtime(&t), sizeof(tmrec) );
    t2 = mktime(&tmrec);                 /* GMT (or UTC) in seconds*/
    tdiff = t2 - t1;

#if U_PLATFORM != U_PF_IPHONE
    /* imitate NT behaviour, which returns same timezone offset to GMT for
       winter and summer.
       This does not work on all platforms. For instance, on glibc on Linux
       and on Mac OS 10.5, tdiff calculated above remains the same
       regardless of whether DST is in effect or not. iOS is another
       platform where this does not work. Linux + glibc and Mac OS 10.5
       have U_TIMEZONE defined so that this code is not reached.
    */
    if (dst_checked)
        tdiff += 3600;
#endif
    return tdiff;
#endif
}

/* Note that U_TZNAME does *not* have to be tzname, but if it is,
   some platforms need to have it declared here. */

#if defined(U_TZNAME) && (U_PLATFORM == U_PF_IRIX || U_PLATFORM_IS_DARWIN_BASED)
/* RS6000 and others reject char **tzname.  */
extern U_IMPORT char *U_TZNAME[];
#endif

#if !UCONFIG_NO_FILE_IO && ((U_PLATFORM_IS_DARWIN_BASED && (U_PLATFORM != U_PF_IPHONE || defined(U_TIMEZONE))) || U_PLATFORM_IS_LINUX_BASED || U_PLATFORM == U_PF_BSD || U_PLATFORM == U_PF_SOLARIS)
/* These platforms are likely to use Olson timezone IDs. */
/* common targets of the symbolic link at TZDEFAULT are:
 * "/usr/share/zoneinfo/<olsonID>" default, older Linux distros, macOS to 10.12
 * "../usr/share/zoneinfo/<olsonID>" newer Linux distros: Red Hat Enterprise Linux 7, Ubuntu 16, SuSe Linux 12
 * "/usr/share/lib/zoneinfo/<olsonID>" Solaris
 * "../usr/share/lib/zoneinfo/<olsonID>" Solaris
 * "/var/db/timezone/zoneinfo/<olsonID>" macOS 10.13
 * To avoid checking lots of paths, just check that the target path
 * before the <olsonID> ends with "/zoneinfo/", and the <olsonID> is valid.
 */

#define CHECK_LOCALTIME_LINK 1
#if U_PLATFORM_IS_DARWIN_BASED
#include <tzfile.h>
#define TZZONEINFO      (TZDIR "/")
#elif U_PLATFORM == U_PF_SOLARIS
#define TZDEFAULT       "/etc/localtime"
#define TZZONEINFO      "/usr/share/lib/zoneinfo/"
#define TZ_ENV_CHECK    "localtime"
#else
#define TZDEFAULT       "/etc/localtime"
#define TZZONEINFO      "/usr/share/zoneinfo/"
#endif
#define TZZONEINFOTAIL  "/zoneinfo/"
#if U_HAVE_DIRENT_H
#define TZFILE_SKIP     "posixrules" /* tz file to skip when searching. */
/* Some Linux distributions have 'localtime' in /usr/share/zoneinfo
   symlinked to /etc/localtime, which makes searchForTZFile return
   'localtime' when it's the first match. */
#define TZFILE_putil_SKIP2    "localtime"
#define SEARCH_TZFILE
#include <dirent.h>  /* Needed to search through system timezone files */
#endif
static char gTimeZoneBuffer[PATH_MAX];
static char *gTimeZoneBufferPtr = NULL;
#endif

#if !U_PLATFORM_USES_ONLY_WIN32_API
#define isNonDigit(ch) (ch < '0' || '9' < ch)
static UBool isValidOlsonID(const char *id) {
    int32_t idx = 0;

    /* Determine if this is something like Iceland (Olson ID)
    or AST4ADT (non-Olson ID) */
    while (id[idx] && isNonDigit(id[idx]) && id[idx] != ',') {
        idx++;
    }

    /* If we went through the whole string, then it might be okay.
    The timezone is sometimes set to "CST-7CDT", "CST6CDT5,J129,J131/19:30",
    "GRNLNDST3GRNLNDDT" or similar, so we cannot use it.
    The rest of the time it could be an Olson ID. George */
    return (UBool)(id[idx] == 0
        || uprv_strcmp(id, "PST8PDT") == 0
        || uprv_strcmp(id, "MST7MDT") == 0
        || uprv_strcmp(id, "CST6CDT") == 0
        || uprv_strcmp(id, "EST5EDT") == 0);
}

/* On some Unix-like OS, 'posix' subdirectory in
   /usr/share/zoneinfo replicates the top-level contents. 'right'
   subdirectory has the same set of files, but individual files
   are different from those in the top-level directory or 'posix'
   because 'right' has files for TAI (Int'l Atomic Time) while 'posix'
   has files for UTC.
   When the first match for /etc/localtime is in either of them
   (usually in posix because 'right' has different file contents),
   or TZ environment variable points to one of them, createTimeZone
   fails because, say, 'posix/America/New_York' is not an Olson
   timezone id ('America/New_York' is). So, we have to skip
   'posix/' and 'right/' at the beginning. */
static void skipZoneIDPrefix(const char** id) {
    if (uprv_strncmp(*id, "posix/", 6) == 0
        || uprv_strncmp(*id, "right/", 6) == 0)
    {
        *id += 6;
    }
}
#endif

#if defined(U_TZNAME) && !U_PLATFORM_USES_ONLY_WIN32_API

#define CONVERT_HOURS_TO_SECONDS(offset) (int32_t)(offset*3600)
typedef struct OffsetZoneMapping {
    int32_t offsetSeconds;
    int32_t daylightType; /* 0=U_DAYLIGHT_NONE, 1=daylight in June-U_DAYLIGHT_JUNE, 2=daylight in December=U_DAYLIGHT_DECEMBER*/
    const char *stdID;
    const char *dstID;
    const char *olsonID;
} OffsetZoneMapping;

enum { U_DAYLIGHT_NONE=0,U_DAYLIGHT_JUNE=1,U_DAYLIGHT_DECEMBER=2 };

/*
This list tries to disambiguate a set of abbreviated timezone IDs and offsets
and maps it to an Olson ID.
Before adding anything to this list, take a look at
icu/source/tools/tzcode/tz.alias
Sometimes no daylight savings (0) is important to define due to aliases.
This list can be tested with icu/source/test/compat/tzone.pl
More values could be added to daylightType to increase precision.
*/
static const struct OffsetZoneMapping OFFSET_ZONE_MAPPINGS[] = {
    {-45900, 2, "CHAST", "CHADT", "Pacific/Chatham"},
    {-43200, 1, "PETT", "PETST", "Asia/Kamchatka"},
    {-43200, 2, "NZST", "NZDT", "Pacific/Auckland"},
    {-43200, 1, "ANAT", "ANAST", "Asia/Anadyr"},
    {-39600, 1, "MAGT", "MAGST", "Asia/Magadan"},
    {-37800, 2, "LHST", "LHST", "Australia/Lord_Howe"},
    {-36000, 2, "EST", "EST", "Australia/Sydney"},
    {-36000, 1, "SAKT", "SAKST", "Asia/Sakhalin"},
    {-36000, 1, "VLAT", "VLAST", "Asia/Vladivostok"},
    {-34200, 2, "CST", "CST", "Australia/South"},
    {-32400, 1, "YAKT", "YAKST", "Asia/Yakutsk"},
    {-32400, 1, "CHOT", "CHOST", "Asia/Choibalsan"},
    {-31500, 2, "CWST", "CWST", "Australia/Eucla"},
    {-28800, 1, "IRKT", "IRKST", "Asia/Irkutsk"},
    {-28800, 1, "ULAT", "ULAST", "Asia/Ulaanbaatar"},
    {-28800, 2, "WST", "WST", "Australia/West"},
    {-25200, 1, "HOVT", "HOVST", "Asia/Hovd"},
    {-25200, 1, "KRAT", "KRAST", "Asia/Krasnoyarsk"},
    {-21600, 1, "NOVT", "NOVST", "Asia/Novosibirsk"},
    {-21600, 1, "OMST", "OMSST", "Asia/Omsk"},
    {-18000, 1, "YEKT", "YEKST", "Asia/Yekaterinburg"},
    {-14400, 1, "SAMT", "SAMST", "Europe/Samara"},
    {-14400, 1, "AMT", "AMST", "Asia/Yerevan"},
    {-14400, 1, "AZT", "AZST", "Asia/Baku"},
    {-10800, 1, "AST", "ADT", "Asia/Baghdad"},
    {-10800, 1, "MSK", "MSD", "Europe/Moscow"},
    {-10800, 1, "VOLT", "VOLST", "Europe/Volgograd"},
    {-7200, 0, "EET", "CEST", "Africa/Tripoli"},
    {-7200, 1, "EET", "EEST", "Europe/Athens"}, /* Conflicts with Africa/Cairo */
    {-7200, 1, "IST", "IDT", "Asia/Jerusalem"},
    {-3600, 0, "CET", "WEST", "Africa/Algiers"},
    {-3600, 2, "WAT", "WAST", "Africa/Windhoek"},
    {0, 1, "GMT", "IST", "Europe/Dublin"},
    {0, 1, "GMT", "BST", "Europe/London"},
    {0, 0, "WET", "WEST", "Africa/Casablanca"},
    {0, 0, "WET", "WET", "Africa/El_Aaiun"},
    {3600, 1, "AZOT", "AZOST", "Atlantic/Azores"},
    {3600, 1, "EGT", "EGST", "America/Scoresbysund"},
    {10800, 1, "PMST", "PMDT", "America/Miquelon"},
    {10800, 2, "UYT", "UYST", "America/Montevideo"},
    {10800, 1, "WGT", "WGST", "America/Godthab"},
    {10800, 2, "BRT", "BRST", "Brazil/East"},
    {12600, 1, "NST", "NDT", "America/St_Johns"},
    {14400, 1, "AST", "ADT", "Canada/Atlantic"},
    {14400, 2, "AMT", "AMST", "America/Cuiaba"},
    {14400, 2, "CLT", "CLST", "Chile/Continental"},
    {14400, 2, "FKT", "FKST", "Atlantic/Stanley"},
    {14400, 2, "PYT", "PYST", "America/Asuncion"},
    {18000, 1, "CST", "CDT", "America/Havana"},
    {18000, 1, "EST", "EDT", "US/Eastern"}, /* Conflicts with America/Grand_Turk */
    {21600, 2, "EAST", "EASST", "Chile/EasterIsland"},
    {21600, 0, "CST", "MDT", "Canada/Saskatchewan"},
    {21600, 0, "CST", "CDT", "America/Guatemala"},
    {21600, 1, "CST", "CDT", "US/Central"}, /* Conflicts with Mexico/General */
    {25200, 1, "MST", "MDT", "US/Mountain"}, /* Conflicts with Mexico/BajaSur */
    {28800, 0, "PST", "PST", "Pacific/Pitcairn"},
    {28800, 1, "PST", "PDT", "US/Pacific"}, /* Conflicts with Mexico/BajaNorte */
    {32400, 1, "AKST", "AKDT", "US/Alaska"},
    {36000, 1, "HAST", "HADT", "US/Aleutian"}
};

/*#define DEBUG_TZNAME*/

static const char* remapShortTimeZone(const char *stdID, const char *dstID, int32_t daylightType, int32_t offset)
{
    int32_t idx;
#ifdef DEBUG_TZNAME
    fprintf(stderr, "TZ=%s std=%s dst=%s daylight=%d offset=%d\n", getenv("TZ"), stdID, dstID, daylightType, offset);
#endif
    for (idx = 0; idx < UPRV_LENGTHOF(OFFSET_ZONE_MAPPINGS); idx++)
    {
        if (offset == OFFSET_ZONE_MAPPINGS[idx].offsetSeconds
            && daylightType == OFFSET_ZONE_MAPPINGS[idx].daylightType
            && strcmp(OFFSET_ZONE_MAPPINGS[idx].stdID, stdID) == 0
            && strcmp(OFFSET_ZONE_MAPPINGS[idx].dstID, dstID) == 0)
        {
            return OFFSET_ZONE_MAPPINGS[idx].olsonID;
        }
    }
    return NULL;
}
#endif

#ifdef SEARCH_TZFILE
#define MAX_READ_SIZE 512

typedef struct DefaultTZInfo {
    char* defaultTZBuffer;
    int64_t defaultTZFileSize;
    FILE* defaultTZFilePtr;
    UBool defaultTZstatus;
    int32_t defaultTZPosition;
} DefaultTZInfo;

/*
 * This method compares the two files given to see if they are a match.
 * It is currently use to compare two TZ files.
 */
static UBool compareBinaryFiles(const char* defaultTZFileName, const char* TZFileName, DefaultTZInfo* tzInfo) {
    FILE* file;
    int64_t sizeFile;
    int64_t sizeFileLeft;
    int32_t sizeFileRead;
    int32_t sizeFileToRead;
    char bufferFile[MAX_READ_SIZE];
    UBool result = TRUE;

    if (tzInfo->defaultTZFilePtr == NULL) {
        tzInfo->defaultTZFilePtr = fopen(defaultTZFileName, "r");
    }
    file = fopen(TZFileName, "r");

    tzInfo->defaultTZPosition = 0; /* reset position to begin search */

    if (file != NULL && tzInfo->defaultTZFilePtr != NULL) {
        /* First check that the file size are equal. */
        if (tzInfo->defaultTZFileSize == 0) {
            fseek(tzInfo->defaultTZFilePtr, 0, SEEK_END);
            tzInfo->defaultTZFileSize = ftell(tzInfo->defaultTZFilePtr);
        }
        fseek(file, 0, SEEK_END);
        sizeFile = ftell(file);
        sizeFileLeft = sizeFile;

        if (sizeFile != tzInfo->defaultTZFileSize) {
            result = FALSE;
        } else {
            /* Store the data from the files in seperate buffers and
             * compare each byte to determine equality.
             */
            if (tzInfo->defaultTZBuffer == NULL) {
                rewind(tzInfo->defaultTZFilePtr);
                tzInfo->defaultTZBuffer = (char*)uprv_malloc(sizeof(char) * tzInfo->defaultTZFileSize);
                sizeFileRead = fread(tzInfo->defaultTZBuffer, 1, tzInfo->defaultTZFileSize, tzInfo->defaultTZFilePtr);
            }
            rewind(file);
            while(sizeFileLeft > 0) {
                uprv_memset(bufferFile, 0, MAX_READ_SIZE);
                sizeFileToRead = sizeFileLeft < MAX_READ_SIZE ? sizeFileLeft : MAX_READ_SIZE;

                sizeFileRead = fread(bufferFile, 1, sizeFileToRead, file);
                if (memcmp(tzInfo->defaultTZBuffer + tzInfo->defaultTZPosition, bufferFile, sizeFileRead) != 0) {
                    result = FALSE;
                    break;
                }
                sizeFileLeft -= sizeFileRead;
                tzInfo->defaultTZPosition += sizeFileRead;
            }
        }
    } else {
        result = FALSE;
    }

    if (file != NULL) {
        fclose(file);
    }

    return result;
}


/* dirent also lists two entries: "." and ".." that we can safely ignore. */
#define putil_SKIP1 "."
#define putil_SKIP2 ".."
static UBool U_CALLCONV putil_cleanup(void);
static CharString *gSearchTZFileResult = NULL;

/*
 * This method recursively traverses the directory given for a matching TZ file and returns the first match.
 * This function is not thread safe - it uses a global, gSearchTZFileResult, to hold its results.
 */
static char* searchForTZFile(const char* path, DefaultTZInfo* tzInfo) {
    DIR* dirp = NULL;
    struct dirent* dirEntry = NULL;
    char* result = NULL;
    UErrorCode status = U_ZERO_ERROR;

    /* Save the current path */
    CharString curpath(path, -1, status);
    if (U_FAILURE(status)) {
        goto cleanupAndReturn;
    }

    dirp = opendir(path);
    if (dirp == NULL) {
        goto cleanupAndReturn;
    }

    if (gSearchTZFileResult == NULL) {
        gSearchTZFileResult = new CharString;
        if (gSearchTZFileResult == NULL) {
            goto cleanupAndReturn;
        }
        ucln_common_registerCleanup(UCLN_COMMON_PUTIL, putil_cleanup);
    }

    /* Check each entry in the directory. */
    while((dirEntry = readdir(dirp)) != NULL) {
        const char* dirName = dirEntry->d_name;
        if (uprv_strcmp(dirName, putil_SKIP1) != 0 && uprv_strcmp(dirName, putil_SKIP2) != 0
            && uprv_strcmp(TZFILE_SKIP, dirName) != 0 && uprv_strcmp(TZFILE_putil_SKIP2, dirName) != 0) {
            /* Create a newpath with the new entry to test each entry in the directory. */
            CharString newpath(curpath, status);
            newpath.append(dirName, -1, status);
            if (U_FAILURE(status)) {
                break;
            }

            DIR* subDirp = NULL;
            if ((subDirp = opendir(newpath.data())) != NULL) {
                /* If this new path is a directory, make a recursive call with the newpath. */
                closedir(subDirp);
                newpath.append('/', status);
                if (U_FAILURE(status)) {
                    break;
                }
                result = searchForTZFile(newpath.data(), tzInfo);
                /*
                 Have to get out here. Otherwise, we'd keep looking
                 and return the first match in the top-level directory
                 if there's a match in the top-level. If not, this function
                 would return NULL and set gTimeZoneBufferPtr to NULL in initDefault().
                 It worked without this in most cases because we have a fallback of calling
                 localtime_r to figure out the default timezone.
                */
                if (result != NULL)
                    break;
            } else {
                if(compareBinaryFiles(TZDEFAULT, newpath.data(), tzInfo)) {
                    int32_t amountToSkip = sizeof(TZZONEINFO) - 1;
                    if (amountToSkip > newpath.length()) {
                        amountToSkip = newpath.length();
                    }
                    const char* zoneid = newpath.data() + amountToSkip;
                    skipZoneIDPrefix(&zoneid);
                    gSearchTZFileResult->clear();
                    gSearchTZFileResult->append(zoneid, -1, status);
                    if (U_FAILURE(status)) {
                        break;
                    }
                    result = gSearchTZFileResult->data();
                    /* Get out after the first one found. */
                    break;
                }
            }
        }
    }

  cleanupAndReturn:
    if (dirp) {
        closedir(dirp);
    }
    return result;
}
#endif

U_CAPI void U_EXPORT2
uprv_tzname_clear_cache()
{
#if defined(CHECK_LOCALTIME_LINK) && !defined(DEBUG_SKIP_LOCALTIME_LINK)
    gTimeZoneBufferPtr = NULL;
#endif
}

U_CAPI const char* U_EXPORT2
uprv_tzname(int n)
{
    (void)n; // Avoid unreferenced parameter warning.
    const char *tzid = NULL;
#if U_PLATFORM_USES_ONLY_WIN32_API
    tzid = uprv_detectWindowsTimeZone();

    if (tzid != NULL) {
        return tzid;
    }

#ifndef U_TZNAME
    // The return value is free'd in timezone.cpp on Windows because
    // the other code path returns a pointer to a heap location.
    // If we don't have a name already, then tzname wouldn't be any
    // better, so just fall back.
    return uprv_strdup("");
#endif // !U_TZNAME

#else

/*#if U_PLATFORM_IS_DARWIN_BASED
    int ret;

    tzid = getenv("TZFILE");
    if (tzid != NULL) {
        return tzid;
    }
#endif*/

/* This code can be temporarily disabled to test tzname resolution later on. */
#ifndef DEBUG_TZNAME
    tzid = getenv("TZ");
    if (tzid != NULL && isValidOlsonID(tzid)
#if U_PLATFORM == U_PF_SOLARIS
    /* When TZ equals localtime on Solaris, check the /etc/localtime file. */
        && uprv_strcmp(tzid, TZ_ENV_CHECK) != 0
#endif
    ) {
        /* The colon forces tzset() to treat the remainder as zoneinfo path */
        if (tzid[0] == ':') {
            tzid++;
        }
        /* This might be a good Olson ID. */
        skipZoneIDPrefix(&tzid);
        return tzid;
    }
    /* else U_TZNAME will give a better result. */
#endif

#if defined(CHECK_LOCALTIME_LINK) && !defined(DEBUG_SKIP_LOCALTIME_LINK)
    /* Caller must handle threading issues */
    if (gTimeZoneBufferPtr == NULL) {
        /*
        This is a trick to look at the name of the link to get the Olson ID
        because the tzfile contents is underspecified.
        This isn't guaranteed to work because it may not be a symlink.
        */
        int32_t ret = (int32_t)readlink(TZDEFAULT, gTimeZoneBuffer, sizeof(gTimeZoneBuffer)-1);
        if (0 < ret) {
            int32_t tzZoneInfoTailLen = uprv_strlen(TZZONEINFOTAIL);
            gTimeZoneBuffer[ret] = 0;
            char *  tzZoneInfoTailPtr = uprv_strstr(gTimeZoneBuffer, TZZONEINFOTAIL);

            if (tzZoneInfoTailPtr != NULL
                && isValidOlsonID(tzZoneInfoTailPtr + tzZoneInfoTailLen))
            {
                return (gTimeZoneBufferPtr = tzZoneInfoTailPtr + tzZoneInfoTailLen);
            }
        } else {
#if defined(SEARCH_TZFILE)
            DefaultTZInfo* tzInfo = (DefaultTZInfo*)uprv_malloc(sizeof(DefaultTZInfo));
            if (tzInfo != NULL) {
                tzInfo->defaultTZBuffer = NULL;
                tzInfo->defaultTZFileSize = 0;
                tzInfo->defaultTZFilePtr = NULL;
                tzInfo->defaultTZstatus = FALSE;
                tzInfo->defaultTZPosition = 0;

                gTimeZoneBufferPtr = searchForTZFile(TZZONEINFO, tzInfo);

                /* Free previously allocated memory */
                if (tzInfo->defaultTZBuffer != NULL) {
                    uprv_free(tzInfo->defaultTZBuffer);
                }
                if (tzInfo->defaultTZFilePtr != NULL) {
                    fclose(tzInfo->defaultTZFilePtr);
                }
                uprv_free(tzInfo);
            }

            if (gTimeZoneBufferPtr != NULL && isValidOlsonID(gTimeZoneBufferPtr)) {
                return gTimeZoneBufferPtr;
            }
#endif
        }
    }
    else {
        return gTimeZoneBufferPtr;
    }
#endif
#endif

#ifdef U_TZNAME
#if U_PLATFORM_USES_ONLY_WIN32_API
    /* The return value is free'd in timezone.cpp on Windows because
     * the other code path returns a pointer to a heap location. */
    return uprv_strdup(U_TZNAME[n]);
#else
    /*
    U_TZNAME is usually a non-unique abbreviation, which isn't normally usable.
    So we remap the abbreviation to an olson ID.

    Since Windows exposes a little more timezone information,
    we normally don't use this code on Windows because
    uprv_detectWindowsTimeZone should have already given the correct answer.
    */
    {
        struct tm juneSol, decemberSol;
        int daylightType;
        static const time_t juneSolstice=1182478260; /*2007-06-21 18:11 UT*/
        static const time_t decemberSolstice=1198332540; /*2007-12-22 06:09 UT*/

        /* This probing will tell us when daylight savings occurs.  */
        localtime_r(&juneSolstice, &juneSol);
        localtime_r(&decemberSolstice, &decemberSol);
        if(decemberSol.tm_isdst > 0) {
          daylightType = U_DAYLIGHT_DECEMBER;
        } else if(juneSol.tm_isdst > 0) {
          daylightType = U_DAYLIGHT_JUNE;
        } else {
          daylightType = U_DAYLIGHT_NONE;
        }
        tzid = remapShortTimeZone(U_TZNAME[0], U_TZNAME[1], daylightType, uprv_timezone());
        if (tzid != NULL) {
            return tzid;
        }
    }
    return U_TZNAME[n];
#endif
#else
    return "";
#endif
}

/* Get and set the ICU data directory --------------------------------------- */

static icu::UInitOnce gDataDirInitOnce = U_INITONCE_INITIALIZER;
static char *gDataDirectory = NULL;

UInitOnce gTimeZoneFilesInitOnce = U_INITONCE_INITIALIZER;
static CharString *gTimeZoneFilesDirectory = NULL;

#if U_POSIX_LOCALE || U_PLATFORM_USES_ONLY_WIN32_API
 static const char *gCorrectedPOSIXLocale = NULL; /* Sometimes heap allocated */
 static bool gCorrectedPOSIXLocaleHeapAllocated = false;
#endif

static UBool U_CALLCONV putil_cleanup(void)
{
    if (gDataDirectory && *gDataDirectory) {
        uprv_free(gDataDirectory);
    }
    gDataDirectory = NULL;
    gDataDirInitOnce.reset();

    delete gTimeZoneFilesDirectory;
    gTimeZoneFilesDirectory = NULL;
    gTimeZoneFilesInitOnce.reset();

#ifdef SEARCH_TZFILE
    delete gSearchTZFileResult;
    gSearchTZFileResult = NULL;
#endif

#if U_POSIX_LOCALE || U_PLATFORM_USES_ONLY_WIN32_API
    if (gCorrectedPOSIXLocale && gCorrectedPOSIXLocaleHeapAllocated) {
        uprv_free(const_cast<char *>(gCorrectedPOSIXLocale));
        gCorrectedPOSIXLocale = NULL;
        gCorrectedPOSIXLocaleHeapAllocated = false;
    }
#endif
    return TRUE;
}

/*
 * Set the data directory.
 *    Make a copy of the passed string, and set the global data dir to point to it.
 */
U_CAPI void U_EXPORT2
u_setDataDirectory(const char *directory) {
    char *newDataDir;
    int32_t length;

    if(directory==NULL || *directory==0) {
        /* A small optimization to prevent the malloc and copy when the
        shared library is used, and this is a way to make sure that NULL
        is never returned.
        */
        newDataDir = (char *)"";
    }
    else {
        length=(int32_t)uprv_strlen(directory);
        newDataDir = (char *)uprv_malloc(length + 2);
        /* Exit out if newDataDir could not be created. */
        if (newDataDir == NULL) {
            return;
        }
        uprv_strcpy(newDataDir, directory);

#if (U_FILE_SEP_CHAR != U_FILE_ALT_SEP_CHAR)
        {
            char *p;
            while((p = uprv_strchr(newDataDir, U_FILE_ALT_SEP_CHAR)) != NULL) {
                *p = U_FILE_SEP_CHAR;
            }
        }
#endif
    }

    if (gDataDirectory && *gDataDirectory) {
        uprv_free(gDataDirectory);
    }
    gDataDirectory = newDataDir;
    ucln_common_registerCleanup(UCLN_COMMON_PUTIL, putil_cleanup);
}

U_CAPI UBool U_EXPORT2
uprv_pathIsAbsolute(const char *path)
{
  if(!path || !*path) {
    return FALSE;
  }

  if(*path == U_FILE_SEP_CHAR) {
    return TRUE;
  }

#if (U_FILE_SEP_CHAR != U_FILE_ALT_SEP_CHAR)
  if(*path == U_FILE_ALT_SEP_CHAR) {
    return TRUE;
  }
#endif

#if U_PLATFORM_USES_ONLY_WIN32_API
  if( (((path[0] >= 'A') && (path[0] <= 'Z')) ||
       ((path[0] >= 'a') && (path[0] <= 'z'))) &&
      path[1] == ':' ) {
    return TRUE;
  }
#endif

  return FALSE;
}

/* Backup setting of ICU_DATA_DIR_PREFIX_ENV_VAR
   (needed for some Darwin ICU build environments) */
#if U_PLATFORM_IS_DARWIN_BASED && TARGET_OS_SIMULATOR
# if !defined(ICU_DATA_DIR_PREFIX_ENV_VAR)
#  define ICU_DATA_DIR_PREFIX_ENV_VAR "IPHONE_SIMULATOR_ROOT"
# endif
#endif

#if defined(ICU_DATA_DIR_WINDOWS)
// Helper function to get the ICU Data Directory under the Windows directory location.
static BOOL U_CALLCONV getIcuDataDirectoryUnderWindowsDirectory(char* directoryBuffer, UINT bufferLength)
{
    wchar_t windowsPath[MAX_PATH];
    char windowsPathUtf8[MAX_PATH];

    UINT length = GetSystemWindowsDirectoryW(windowsPath, UPRV_LENGTHOF(windowsPath));
    if ((length > 0) && (length < (UPRV_LENGTHOF(windowsPath) - 1))) {
        // Convert UTF-16 to a UTF-8 string.
        UErrorCode status = U_ZERO_ERROR;
        int32_t windowsPathUtf8Len = 0;
        u_strToUTF8(windowsPathUtf8, static_cast<int32_t>(UPRV_LENGTHOF(windowsPathUtf8)),
            &windowsPathUtf8Len, reinterpret_cast<const UChar*>(windowsPath), -1, &status);

        if (U_SUCCESS(status) && (status != U_STRING_NOT_TERMINATED_WARNING) &&
            (windowsPathUtf8Len < (UPRV_LENGTHOF(windowsPathUtf8) - 1))) {
            // Ensure it always has a separator, so we can append the ICU data path.
            if (windowsPathUtf8[windowsPathUtf8Len - 1] != U_FILE_SEP_CHAR) {
                windowsPathUtf8[windowsPathUtf8Len++] = U_FILE_SEP_CHAR;
                windowsPathUtf8[windowsPathUtf8Len] = '\0';
            }
            // Check if the concatenated string will fit.
            if ((windowsPathUtf8Len + UPRV_LENGTHOF(ICU_DATA_DIR_WINDOWS)) < bufferLength) {
                uprv_strcpy(directoryBuffer, windowsPathUtf8);
                uprv_strcat(directoryBuffer, ICU_DATA_DIR_WINDOWS);
                return TRUE;
            }
        }
    }

    return FALSE;
}
#endif

static void U_CALLCONV dataDirectoryInitFn() {
    /* If we already have the directory, then return immediately. Will happen if user called
     * u_setDataDirectory().
     */
    if (gDataDirectory) {
        return;
    }

    const char *path = NULL;
#if defined(ICU_DATA_DIR_PREFIX_ENV_VAR)
    char datadir_path_buffer[PATH_MAX];
#endif

    /*
    When ICU_NO_USER_DATA_OVERRIDE is defined, users aren't allowed to
    override ICU's data with the ICU_DATA environment variable. This prevents
    problems where multiple custom copies of ICU's specific version of data
    are installed on a system. Either the application must define the data
    directory with u_setDataDirectory, define ICU_DATA_DIR when compiling
    ICU, set the data with udata_setCommonData or trust that all of the
    required data is contained in ICU's data library that contains
    the entry point defined by U_ICUDATA_ENTRY_POINT.

    There may also be some platforms where environment variables
    are not allowed.
    */
#   if !defined(ICU_NO_USER_DATA_OVERRIDE) && !UCONFIG_NO_FILE_IO
    /* First try to get the environment variable */
#     if U_PLATFORM_HAS_WINUWP_API == 0  // Windows UWP does not support getenv
        path=getenv("ICU_DATA");
#     endif
#   endif

    /* ICU_DATA_DIR may be set as a compile option.
     * U_ICU_DATA_DEFAULT_DIR is provided and is set by ICU at compile time
     * and is used only when data is built in archive mode eliminating the need
     * for ICU_DATA_DIR to be set. U_ICU_DATA_DEFAULT_DIR is set to the installation
     * directory of the data dat file. Users should use ICU_DATA_DIR if they want to
     * set their own path.
     */
#if defined(ICU_DATA_DIR) || defined(U_ICU_DATA_DEFAULT_DIR)
    if(path==NULL || *path==0) {
# if defined(ICU_DATA_DIR_PREFIX_ENV_VAR)
        const char *prefix = getenv(ICU_DATA_DIR_PREFIX_ENV_VAR);
# endif
# ifdef ICU_DATA_DIR
        path=ICU_DATA_DIR;
# else
        path=U_ICU_DATA_DEFAULT_DIR;
# endif
# if defined(ICU_DATA_DIR_PREFIX_ENV_VAR)
        if (prefix != NULL) {
            snprintf(datadir_path_buffer, PATH_MAX, "%s%s", prefix, path);
            path=datadir_path_buffer;
        }
# endif
    }
#endif

#if defined(ICU_DATA_DIR_WINDOWS)
    char datadir_path_buffer[MAX_PATH];
    if (getIcuDataDirectoryUnderWindowsDirectory(datadir_path_buffer, UPRV_LENGTHOF(datadir_path_buffer))) {
        path = datadir_path_buffer;
    }
#endif

    if(path==NULL) {
        /* It looks really bad, set it to something. */
        path = "";
    }

    u_setDataDirectory(path);
    return;
}

U_CAPI const char * U_EXPORT2
u_getDataDirectory(void) {
    umtx_initOnce(gDataDirInitOnce, &dataDirectoryInitFn);
    return gDataDirectory;
}

static void setTimeZoneFilesDir(const char *path, UErrorCode &status) {
    if (U_FAILURE(status)) {
        return;
    }
    gTimeZoneFilesDirectory->clear();
    gTimeZoneFilesDirectory->append(path, status);
#if (U_FILE_SEP_CHAR != U_FILE_ALT_SEP_CHAR)
    char *p = gTimeZoneFilesDirectory->data();
    while ((p = uprv_strchr(p, U_FILE_ALT_SEP_CHAR)) != NULL) {
        *p = U_FILE_SEP_CHAR;
    }
#endif
}

#define PUTIL_TO_STRING(x) PUTIL_TO_STRING_2(x)
#define PUTIL_TO_STRING_2(x) #x

static void U_CALLCONV TimeZoneDataDirInitFn(UErrorCode &status) {
    U_ASSERT(gTimeZoneFilesDirectory == NULL);
    ucln_common_registerCleanup(UCLN_COMMON_PUTIL, putil_cleanup);
    gTimeZoneFilesDirectory = new CharString();
    if (gTimeZoneFilesDirectory == NULL) {
        status = U_MEMORY_ALLOCATION_ERROR;
        return;
    }

    const char *dir = "";

#if U_PLATFORM_HAS_WINUWP_API == 1
// The UWP version does not support the environment variable setting.

# if defined(ICU_DATA_DIR_WINDOWS)
    // When using the Windows system data, we can possibly pick up time zone data from the Windows directory.
    char datadir_path_buffer[MAX_PATH];
    if (getIcuDataDirectoryUnderWindowsDirectory(datadir_path_buffer, UPRV_LENGTHOF(datadir_path_buffer))) {
        dir = datadir_path_buffer;
    }
# endif

#else
    dir = getenv("ICU_TIMEZONE_FILES_DIR");
#endif // U_PLATFORM_HAS_WINUWP_API

#if defined(U_TIMEZONE_FILES_DIR)
    if (dir == NULL) {
        // Build time configuration setting.
        dir = PUTIL_TO_STRING(U_TIMEZONE_FILES_DIR);
    }
#endif

    if (dir == NULL) {
        dir = "";
    }

    setTimeZoneFilesDir(dir, status);
}


U_CAPI const char * U_EXPORT2
u_getTimeZoneFilesDirectory(UErrorCode *status) {
    umtx_initOnce(gTimeZoneFilesInitOnce, &TimeZoneDataDirInitFn, *status);
    return U_SUCCESS(*status) ? gTimeZoneFilesDirectory->data() : "";
}

U_CAPI void U_EXPORT2
u_setTimeZoneFilesDirectory(const char *path, UErrorCode *status) {
    umtx_initOnce(gTimeZoneFilesInitOnce, &TimeZoneDataDirInitFn, *status);
    setTimeZoneFilesDir(path, *status);

    // Note: this function does some extra churn, first setting based on the
    //       environment, then immediately replacing with the value passed in.
    //       The logic is simpler that way, and performance shouldn't be an issue.
}


#if U_POSIX_LOCALE
/* A helper function used by uprv_getPOSIXIDForDefaultLocale and
 * uprv_getPOSIXIDForDefaultCodepage. Returns the posix locale id for
 * LC_CTYPE and LC_MESSAGES. It doesn't support other locale categories.
 */
static const char *uprv_getPOSIXIDForCategory(int category)
{
    const char* posixID = NULL;
    if (category == LC_MESSAGES || category == LC_CTYPE) {
        /*
        * On Solaris two different calls to setlocale can result in
        * different values. Only get this value once.
        *
        * We must check this first because an application can set this.
        *
        * LC_ALL can't be used because it's platform dependent. The LANG
        * environment variable seems to affect LC_CTYPE variable by default.
        * Here is what setlocale(LC_ALL, NULL) can return.
        * HPUX can return 'C C C C C C C'
        * Solaris can return /en_US/C/C/C/C/C on the second try.
        * Linux can return LC_CTYPE=C;LC_NUMERIC=C;...
        *
        * The default codepage detection also needs to use LC_CTYPE.
        *
        * Do not call setlocale(LC_*, "")! Using an empty string instead
        * of NULL, will modify the libc behavior.
        */
        posixID = setlocale(category, NULL);
        if ((posixID == 0)
            || (uprv_strcmp("C", posixID) == 0)
            || (uprv_strcmp("POSIX", posixID) == 0))
        {
            /* Maybe we got some garbage.  Try something more reasonable */
            posixID = getenv("LC_ALL");
            /* Solaris speaks POSIX -  See IEEE Std 1003.1-2008
             * This is needed to properly handle empty env. variables
             */
#if U_PLATFORM == U_PF_SOLARIS
            if ((posixID == 0) || (posixID[0] == '\0')) {
                posixID = getenv(category == LC_MESSAGES ? "LC_MESSAGES" : "LC_CTYPE");
                if ((posixID == 0) || (posixID[0] == '\0')) {
#else
            if (posixID == 0) {
                posixID = getenv(category == LC_MESSAGES ? "LC_MESSAGES" : "LC_CTYPE");
                if (posixID == 0) {
#endif
                    posixID = getenv("LANG");
                }
            }
        }
    }
    if ((posixID==0)
        || (uprv_strcmp("C", posixID) == 0)
        || (uprv_strcmp("POSIX", posixID) == 0))
    {
        /* Nothing worked.  Give it a nice POSIX default value. */
        posixID = "en_US_POSIX";
        // Note: this test will not catch 'C.UTF-8',
        // that will be handled in uprv_getDefaultLocaleID().
        // Leave this mapping here for the uprv_getPOSIXIDForDefaultCodepage()
        // caller which expects to see "en_US_POSIX" in many branches.
    }
    return posixID;
}

/* Return just the POSIX id for the default locale, whatever happens to be in
 * it. It gets the value from LC_MESSAGES and indirectly from LC_ALL and LANG.
 */
static const char *uprv_getPOSIXIDForDefaultLocale(void)
{
    static const char* posixID = NULL;
    if (posixID == 0) {
        posixID = uprv_getPOSIXIDForCategory(LC_MESSAGES);
    }
    return posixID;
}

#if !U_CHARSET_IS_UTF8
/* Return just the POSIX id for the default codepage, whatever happens to be in
 * it. It gets the value from LC_CTYPE and indirectly from LC_ALL and LANG.
 */
static const char *uprv_getPOSIXIDForDefaultCodepage(void)
{
    static const char* posixID = NULL;
    if (posixID == 0) {
        posixID = uprv_getPOSIXIDForCategory(LC_CTYPE);
    }
    return posixID;
}
#endif
#endif

/* NOTE: The caller should handle thread safety */
U_CAPI const char* U_EXPORT2
uprv_getDefaultLocaleID()
{
#if U_POSIX_LOCALE
/*
  Note that:  (a '!' means the ID is improper somehow)
     LC_ALL  ---->     default_loc          codepage
--------------------------------------------------------
     ab.CD             ab                   CD
     ab@CD             ab__CD               -
     ab@CD.EF          ab__CD               EF

     ab_CD.EF@GH       ab_CD_GH             EF

Some 'improper' ways to do the same as above:
  !  ab_CD@GH.EF       ab_CD_GH             EF
  !  ab_CD.EF@GH.IJ    ab_CD_GH             EF
  !  ab_CD@ZZ.EF@GH.IJ ab_CD_GH             EF

     _CD@GH            _CD_GH               -
     _CD.EF@GH         _CD_GH               EF

The variant cannot have dots in it.
The 'rightmost' variant (@xxx) wins.
The leftmost codepage (.xxx) wins.
*/
    const char* posixID = uprv_getPOSIXIDForDefaultLocale();

    /* Format: (no spaces)
    ll [ _CC ] [ . MM ] [ @ VV]

      l = lang, C = ctry, M = charmap, V = variant
    */

    if (gCorrectedPOSIXLocale != nullptr) {
        return gCorrectedPOSIXLocale;
    }

    // Copy the ID into owned memory.
    // Over-allocate in case we replace "C" with "en_US_POSIX" (+10), + null termination
    char *correctedPOSIXLocale = static_cast<char *>(uprv_malloc(uprv_strlen(posixID) + 10 + 1));
    if (correctedPOSIXLocale == nullptr) {
        return nullptr;
    }
    uprv_strcpy(correctedPOSIXLocale, posixID);

    char *limit;
    if ((limit = uprv_strchr(correctedPOSIXLocale, '.')) != nullptr) {
        *limit = 0;
    }
    if ((limit = uprv_strchr(correctedPOSIXLocale, '@')) != nullptr) {
        *limit = 0;
    }

    if ((uprv_strcmp("C", correctedPOSIXLocale) == 0) // no @ variant
        || (uprv_strcmp("POSIX", correctedPOSIXLocale) == 0)) {
      // Raw input was C.* or POSIX.*, Give it a nice POSIX default value.
      // (The "C"/"POSIX" case is handled in uprv_getPOSIXIDForCategory())
      uprv_strcpy(correctedPOSIXLocale, "en_US_POSIX");
    }

    /* Note that we scan the *uncorrected* ID. */
    const char *p;
    if ((p = uprv_strrchr(posixID, '@')) != nullptr) {
        p++;

        /* Take care of any special cases here.. */
        if (!uprv_strcmp(p, "nynorsk")) {
            p = "NY";
            /* Don't worry about no__NY. In practice, it won't appear. */
        }

        if (uprv_strchr(correctedPOSIXLocale,'_') == nullptr) {
            uprv_strcat(correctedPOSIXLocale, "__"); /* aa@b -> aa__b (note this can make the new locale 1 char longer) */
        }
        else {
            uprv_strcat(correctedPOSIXLocale, "_"); /* aa_CC@b -> aa_CC_b */
        }

        const char *q;
        if ((q = uprv_strchr(p, '.')) != nullptr) {
            /* How big will the resulting string be? */
            int32_t len = (int32_t)(uprv_strlen(correctedPOSIXLocale) + (q-p));
            uprv_strncat(correctedPOSIXLocale, p, q-p); // do not include charset
            correctedPOSIXLocale[len] = 0;
        }
        else {
            /* Anything following the @ sign */
            uprv_strcat(correctedPOSIXLocale, p);
        }

        /* Should there be a map from 'no@nynorsk' -> no_NO_NY here?
         * How about 'russian' -> 'ru'?
         * Many of the other locales using ISO codes will be handled by the
         * canonicalization functions in uloc_getDefault.
         */
    }

    if (gCorrectedPOSIXLocale == nullptr) {
        gCorrectedPOSIXLocale = correctedPOSIXLocale;
        gCorrectedPOSIXLocaleHeapAllocated = true;
        ucln_common_registerCleanup(UCLN_COMMON_PUTIL, putil_cleanup);
        correctedPOSIXLocale = nullptr;
    }
    posixID = gCorrectedPOSIXLocale;

    if (correctedPOSIXLocale != nullptr) {  /* Was already set - clean up. */
        uprv_free(correctedPOSIXLocale);
    }

    return posixID;

#elif U_PLATFORM_USES_ONLY_WIN32_API
#define POSIX_LOCALE_CAPACITY 64
    UErrorCode status = U_ZERO_ERROR;
    char *correctedPOSIXLocale = nullptr;

    // If we have already figured this out just use the cached value
    if (gCorrectedPOSIXLocale != nullptr) {
        return gCorrectedPOSIXLocale;
    }

    // No cached value, need to determine the current value
    static WCHAR windowsLocale[LOCALE_NAME_MAX_LENGTH] = {};
    int length = GetLocaleInfoEx(LOCALE_NAME_USER_DEFAULT, LOCALE_SNAME, windowsLocale, LOCALE_NAME_MAX_LENGTH);

    // Now we should have a Windows locale name that needs converted to the POSIX style.
    if (length > 0) // If length is 0, then the GetLocaleInfoEx failed.
    {
        // First we need to go from UTF-16 to char (and also convert from _ to - while we're at it.)
        char modifiedWindowsLocale[LOCALE_NAME_MAX_LENGTH] = {};

        int32_t i;
        for (i = 0; i < UPRV_LENGTHOF(modifiedWindowsLocale); i++)
        {
            if (windowsLocale[i] == '_')
            {
                modifiedWindowsLocale[i] = '-';
            }
            else
            {
                modifiedWindowsLocale[i] = static_cast<char>(windowsLocale[i]);
            }

            if (modifiedWindowsLocale[i] == '\0')
            {
                break;
            }
        }

        if (i >= UPRV_LENGTHOF(modifiedWindowsLocale))
        {
            // Ran out of room, can't really happen, maybe we'll be lucky about a matching
            // locale when tags are dropped
            modifiedWindowsLocale[UPRV_LENGTHOF(modifiedWindowsLocale) - 1] = '\0';
        }

        // Now normalize the resulting name
        correctedPOSIXLocale = static_cast<char *>(uprv_malloc(POSIX_LOCALE_CAPACITY + 1));
        /* TODO: Should we just exit on memory allocation failure? */
        if (correctedPOSIXLocale)
        {
            int32_t posixLen = uloc_canonicalize(modifiedWindowsLocale, correctedPOSIXLocale, POSIX_LOCALE_CAPACITY, &status);
            if (U_SUCCESS(status))
            {
                *(correctedPOSIXLocale + posixLen) = 0;
                gCorrectedPOSIXLocale = correctedPOSIXLocale;
                gCorrectedPOSIXLocaleHeapAllocated = true;
                ucln_common_registerCleanup(UCLN_COMMON_PUTIL, putil_cleanup);
            }
            else
            {
                uprv_free(correctedPOSIXLocale);
            }
        }
    }

    // If unable to find a locale we can agree upon, use en-US by default
    if (gCorrectedPOSIXLocale == nullptr) {
        gCorrectedPOSIXLocale = "en_US";
    }
    return gCorrectedPOSIXLocale;

#elif U_PLATFORM == U_PF_OS400
    /* locales are process scoped and are by definition thread safe */
    static char correctedLocale[64];
    const  char *localeID = getenv("LC_ALL");
           char *p;

    if (localeID == NULL)
        localeID = getenv("LANG");
    if (localeID == NULL)
        localeID = setlocale(LC_ALL, NULL);
    /* Make sure we have something... */
    if (localeID == NULL)
        return "en_US_POSIX";

    /* Extract the locale name from the path. */
    if((p = uprv_strrchr(localeID, '/')) != NULL)
    {
        /* Increment p to start of locale name. */
        p++;
        localeID = p;
    }

    /* Copy to work location. */
    uprv_strcpy(correctedLocale, localeID);

    /* Strip off the '.locale' extension. */
    if((p = uprv_strchr(correctedLocale, '.')) != NULL) {
        *p = 0;
    }

    /* Upper case the locale name. */
    T_CString_toUpperCase(correctedLocale);

    /* See if we are using the POSIX locale.  Any of the
    * following are equivalent and use the same QLGPGCMA
    * (POSIX) locale.
    * QLGPGCMA2 means UCS2
    * QLGPGCMA_4 means UTF-32
    * QLGPGCMA_8 means UTF-8
    */
    if ((uprv_strcmp("C", correctedLocale) == 0) ||
        (uprv_strcmp("POSIX", correctedLocale) == 0) ||
        (uprv_strncmp("QLGPGCMA", correctedLocale, 8) == 0))
    {
        uprv_strcpy(correctedLocale, "en_US_POSIX");
    }
    else
    {
        int16_t LocaleLen;

        /* Lower case the lang portion. */
        for(p = correctedLocale; *p != 0 && *p != '_'; p++)
        {
            *p = uprv_tolower(*p);
        }

        /* Adjust for Euro.  After '_E' add 'URO'. */
        LocaleLen = uprv_strlen(correctedLocale);
        if (correctedLocale[LocaleLen - 2] == '_' &&
            correctedLocale[LocaleLen - 1] == 'E')
        {
            uprv_strcat(correctedLocale, "URO");
        }

        /* If using Lotus-based locale then convert to
         * equivalent non Lotus.
         */
        else if (correctedLocale[LocaleLen - 2] == '_' &&
            correctedLocale[LocaleLen - 1] == 'L')
        {
            correctedLocale[LocaleLen - 2] = 0;
        }

        /* There are separate simplified and traditional
         * locales called zh_HK_S and zh_HK_T.
         */
        else if (uprv_strncmp(correctedLocale, "zh_HK", 5) == 0)
        {
            uprv_strcpy(correctedLocale, "zh_HK");
        }

        /* A special zh_CN_GBK locale...
        */
        else if (uprv_strcmp(correctedLocale, "zh_CN_GBK") == 0)
        {
            uprv_strcpy(correctedLocale, "zh_CN");
        }

    }

    return correctedLocale;
#endif

}

#if !U_CHARSET_IS_UTF8
#if U_POSIX_LOCALE
/*
Due to various platform differences, one platform may specify a charset,
when they really mean a different charset. Remap the names so that they are
compatible with ICU. Only conflicting/ambiguous aliases should be resolved
here. Before adding anything to this function, please consider adding unique
names to the ICU alias table in the data directory.
*/
static const char*
remapPlatformDependentCodepage(const char *locale, const char *name) {
    if (locale != NULL && *locale == 0) {
        /* Make sure that an empty locale is handled the same way. */
        locale = NULL;
    }
    if (name == NULL) {
        return NULL;
    }
#if U_PLATFORM == U_PF_AIX
    if (uprv_strcmp(name, "IBM-943") == 0) {
        /* Use the ASCII compatible ibm-943 */
        name = "Shift-JIS";
    }
    else if (uprv_strcmp(name, "IBM-1252") == 0) {
        /* Use the windows-1252 that contains the Euro */
        name = "IBM-5348";
    }
#elif U_PLATFORM == U_PF_SOLARIS
    if (locale != NULL && uprv_strcmp(name, "EUC") == 0) {
        /* Solaris underspecifies the "EUC" name. */
        if (uprv_strcmp(locale, "zh_CN") == 0) {
            name = "EUC-CN";
        }
        else if (uprv_strcmp(locale, "zh_TW") == 0) {
            name = "EUC-TW";
        }
        else if (uprv_strcmp(locale, "ko_KR") == 0) {
            name = "EUC-KR";
        }
    }
    else if (uprv_strcmp(name, "eucJP") == 0) {
        /*
        ibm-954 is the best match.
        ibm-33722 is the default for eucJP (similar to Windows).
        */
        name = "eucjis";
    }
    else if (uprv_strcmp(name, "646") == 0) {
        /*
         * The default codepage given by Solaris is 646 but the C library routines treat it as if it was
         * ISO-8859-1 instead of US-ASCII(646).
         */
        name = "ISO-8859-1";
    }
#elif U_PLATFORM_IS_DARWIN_BASED
    if (locale == NULL && *name == 0) {
        /*
        No locale was specified, and an empty name was passed in.
        This usually indicates that nl_langinfo didn't return valid information.
        Mac OS X uses UTF-8 by default (especially the locale data and console).
        */
        name = "UTF-8";
    }
    else if (uprv_strcmp(name, "CP949") == 0) {
        /* Remap CP949 to a similar codepage to avoid issues with backslash and won symbol. */
        name = "EUC-KR";
    }
    else if (locale != NULL && uprv_strcmp(locale, "en_US_POSIX") != 0 && uprv_strcmp(name, "US-ASCII") == 0) {
        /*
         * For non C/POSIX locale, default the code page to UTF-8 instead of US-ASCII.
         */
        name = "UTF-8";
    }
#elif U_PLATFORM == U_PF_BSD
    if (uprv_strcmp(name, "CP949") == 0) {
        /* Remap CP949 to a similar codepage to avoid issues with backslash and won symbol. */
        name = "EUC-KR";
    }
#elif U_PLATFORM == U_PF_HPUX
    if (locale != NULL && uprv_strcmp(locale, "zh_HK") == 0 && uprv_strcmp(name, "big5") == 0) {
        /* HP decided to extend big5 as hkbig5 even though it's not compatible :-( */
        /* zh_TW.big5 is not the same charset as zh_HK.big5! */
        name = "hkbig5";
    }
    else if (uprv_strcmp(name, "eucJP") == 0) {
        /*
        ibm-1350 is the best match, but unavailable.
        ibm-954 is mostly a superset of ibm-1350.
        ibm-33722 is the default for eucJP (similar to Windows).
        */
        name = "eucjis";
    }
#elif U_PLATFORM == U_PF_LINUX
    if (locale != NULL && uprv_strcmp(name, "euc") == 0) {
        /* Linux underspecifies the "EUC" name. */
        if (uprv_strcmp(locale, "korean") == 0) {
            name = "EUC-KR";
        }
        else if (uprv_strcmp(locale, "japanese") == 0) {
            /* See comment below about eucJP */
            name = "eucjis";
        }
    }
    else if (uprv_strcmp(name, "eucjp") == 0) {
        /*
        ibm-1350 is the best match, but unavailable.
        ibm-954 is mostly a superset of ibm-1350.
        ibm-33722 is the default for eucJP (similar to Windows).
        */
        name = "eucjis";
    }
    else if (locale != NULL && uprv_strcmp(locale, "en_US_POSIX") != 0 &&
            (uprv_strcmp(name, "ANSI_X3.4-1968") == 0 || uprv_strcmp(name, "US-ASCII") == 0)) {
        /*
         * For non C/POSIX locale, default the code page to UTF-8 instead of US-ASCII.
         */
        name = "UTF-8";
    }
    /*
     * Linux returns ANSI_X3.4-1968 for C/POSIX, but the call site takes care of
     * it by falling back to 'US-ASCII' when NULL is returned from this
     * function. So, we don't have to worry about it here.
     */
#endif
    /* return NULL when "" is passed in */
    if (*name == 0) {
        name = NULL;
    }
    return name;
}

static const char*
getCodepageFromPOSIXID(const char *localeName, char * buffer, int32_t buffCapacity)
{
    char localeBuf[100];
    const char *name = NULL;
    char *variant = NULL;

    if (localeName != NULL && (name = (uprv_strchr(localeName, '.'))) != NULL) {
        size_t localeCapacity = uprv_min(sizeof(localeBuf), (name-localeName)+1);
        uprv_strncpy(localeBuf, localeName, localeCapacity);
        localeBuf[localeCapacity-1] = 0; /* ensure NULL termination */
        name = uprv_strncpy(buffer, name+1, buffCapacity);
        buffer[buffCapacity-1] = 0; /* ensure NULL termination */
        if ((variant = const_cast<char *>(uprv_strchr(name, '@'))) != NULL) {
            *variant = 0;
        }
        name = remapPlatformDependentCodepage(localeBuf, name);
    }
    return name;
}
#endif

static const char*
int_getDefaultCodepage()
{
#if U_PLATFORM == U_PF_OS400
    uint32_t ccsid = 37; /* Default to ibm-37 */
    static char codepage[64];
    Qwc_JOBI0400_t jobinfo;
    Qus_EC_t error = { sizeof(Qus_EC_t) }; /* SPI error code */

    EPT_CALL(QUSRJOBI)(&jobinfo, sizeof(jobinfo), "JOBI0400",
        "*                         ", "                ", &error);

    if (error.Bytes_Available == 0) {
        if (jobinfo.Coded_Char_Set_ID != 0xFFFF) {
            ccsid = (uint32_t)jobinfo.Coded_Char_Set_ID;
        }
        else if (jobinfo.Default_Coded_Char_Set_Id != 0xFFFF) {
            ccsid = (uint32_t)jobinfo.Default_Coded_Char_Set_Id;
        }
        /* else use the default */
    }
    sprintf(codepage,"ibm-%d", ccsid);
    return codepage;

#elif U_PLATFORM == U_PF_OS390
    static char codepage[64];

    strncpy(codepage, nl_langinfo(CODESET),63-strlen(UCNV_SWAP_LFNL_OPTION_STRING));
    strcat(codepage,UCNV_SWAP_LFNL_OPTION_STRING);
    codepage[63] = 0; /* NULL terminate */

    return codepage;

#elif U_PLATFORM_USES_ONLY_WIN32_API
    static char codepage[64];
    DWORD codepageNumber = 0;

#if U_PLATFORM_HAS_WINUWP_API == 1
    // UWP doesn't have a direct API to get the default ACP as Microsoft would rather
    // have folks use Unicode than a "system" code page, however this is the same
    // codepage as the system default locale codepage.  (FWIW, the system locale is
    // ONLY used for codepage, it should never be used for anything else)
    GetLocaleInfoEx(LOCALE_NAME_SYSTEM_DEFAULT, LOCALE_IDEFAULTANSICODEPAGE | LOCALE_RETURN_NUMBER,
        (LPWSTR)&codepageNumber, sizeof(codepageNumber) / sizeof(WCHAR));
#else
    // Win32 apps can call GetACP
    codepageNumber = GetACP();
#endif
    // Special case for UTF-8
    if (codepageNumber == 65001)
    {
        return "UTF-8";
    }
    // Windows codepages can look like windows-1252, so format the found number
    // the numbers are eclectic, however all valid system code pages, besides UTF-8
    // are between 3 and 19999
    if (codepageNumber > 0 && codepageNumber < 20000)
    {
        sprintf(codepage, "windows-%ld", codepageNumber);
        return codepage;
    }
    // If the codepage number call failed then return UTF-8
    return "UTF-8";

#elif U_POSIX_LOCALE
    static char codesetName[100];
    const char *localeName = NULL;
    const char *name = NULL;

    localeName = uprv_getPOSIXIDForDefaultCodepage();
    uprv_memset(codesetName, 0, sizeof(codesetName));
    /* On Solaris nl_langinfo returns C locale values unless setlocale
     * was called earlier.
     */
#if (U_HAVE_NL_LANGINFO_CODESET && U_PLATFORM != U_PF_SOLARIS)
    /* When available, check nl_langinfo first because it usually gives more
       useful names. It depends on LC_CTYPE.
       nl_langinfo may use the same buffer as setlocale. */
    {
        const char *codeset = nl_langinfo(U_NL_LANGINFO_CODESET);
#if U_PLATFORM_IS_DARWIN_BASED || U_PLATFORM_IS_LINUX_BASED
        /*
         * On Linux and MacOSX, ensure that default codepage for non C/POSIX locale is UTF-8
         * instead of ASCII.
         */
        if (uprv_strcmp(localeName, "en_US_POSIX") != 0) {
            codeset = remapPlatformDependentCodepage(localeName, codeset);
        } else
#endif
        {
            codeset = remapPlatformDependentCodepage(NULL, codeset);
        }

        if (codeset != NULL) {
            uprv_strncpy(codesetName, codeset, sizeof(codesetName));
            codesetName[sizeof(codesetName)-1] = 0;
            return codesetName;
        }
    }
#endif

    /* Use setlocale in a nice way, and then check some environment variables.
       Maybe the application used setlocale already.
    */
    uprv_memset(codesetName, 0, sizeof(codesetName));
    name = getCodepageFromPOSIXID(localeName, codesetName, sizeof(codesetName));
    if (name) {
        /* if we can find the codeset name from setlocale, return that. */
        return name;
    }

    if (*codesetName == 0)
    {
        /* Everything failed. Return US ASCII (ISO 646). */
        (void)uprv_strcpy(codesetName, "US-ASCII");
    }
    return codesetName;
#else
    return "US-ASCII";
#endif
}


U_CAPI const char*  U_EXPORT2
uprv_getDefaultCodepage()
{
    static char const  *name = NULL;
    umtx_lock(NULL);
    if (name == NULL) {
        name = int_getDefaultCodepage();
    }
    umtx_unlock(NULL);
    return name;
}
#endif  /* !U_CHARSET_IS_UTF8 */


/* end of platform-specific implementation -------------- */

/* version handling --------------------------------------------------------- */

U_CAPI void U_EXPORT2
u_versionFromString(UVersionInfo versionArray, const char *versionString) {
    char *end;
    uint16_t part=0;

    if(versionArray==NULL) {
        return;
    }

    if(versionString!=NULL) {
        for(;;) {
            versionArray[part]=(uint8_t)uprv_strtoul(versionString, &end, 10);
            if(end==versionString || ++part==U_MAX_VERSION_LENGTH || *end!=U_VERSION_DELIMITER) {
                break;
            }
            versionString=end+1;
        }
    }

    while(part<U_MAX_VERSION_LENGTH) {
        versionArray[part++]=0;
    }
}

U_CAPI void U_EXPORT2
u_versionFromUString(UVersionInfo versionArray, const UChar *versionString) {
    if(versionArray!=NULL && versionString!=NULL) {
        char versionChars[U_MAX_VERSION_STRING_LENGTH+1];
        int32_t len = u_strlen(versionString);
        if(len>U_MAX_VERSION_STRING_LENGTH) {
            len = U_MAX_VERSION_STRING_LENGTH;
        }
        u_UCharsToChars(versionString, versionChars, len);
        versionChars[len]=0;
        u_versionFromString(versionArray, versionChars);
    }
}

U_CAPI void U_EXPORT2
u_versionToString(const UVersionInfo versionArray, char *versionString) {
    uint16_t count, part;
    uint8_t field;

    if(versionString==NULL) {
        return;
    }

    if(versionArray==NULL) {
        versionString[0]=0;
        return;
    }

    /* count how many fields need to be written */
    for(count=4; count>0 && versionArray[count-1]==0; --count) {
    }

    if(count <= 1) {
        count = 2;
    }

    /* write the first part */
    /* write the decimal field value */
    field=versionArray[0];
    if(field>=100) {
        *versionString++=(char)('0'+field/100);
        field%=100;
    }
    if(field>=10) {
        *versionString++=(char)('0'+field/10);
        field%=10;
    }
    *versionString++=(char)('0'+field);

    /* write the following parts */
    for(part=1; part<count; ++part) {
        /* write a dot first */
        *versionString++=U_VERSION_DELIMITER;

        /* write the decimal field value */
        field=versionArray[part];
        if(field>=100) {
            *versionString++=(char)('0'+field/100);
            field%=100;
        }
        if(field>=10) {
            *versionString++=(char)('0'+field/10);
            field%=10;
        }
        *versionString++=(char)('0'+field);
    }

    /* NUL-terminate */
    *versionString=0;
}

U_CAPI void U_EXPORT2
u_getVersion(UVersionInfo versionArray) {
    // (void)copyright;   // Suppress unused variable warning from clang.
    u_versionFromString(versionArray, U_ICU_VERSION);
}

/*
 * Hey, Emacs, please set the following:
 *
 * Local Variables:
 * indent-tabs-mode: nil
 * End:
 *
 */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
******************************************************************************
*
*   Copyright (C) 1999-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
*
******************************************************************************/


/*----------------------------------------------------------------------------
 *
 *       Memory mapped file wrappers for use by the ICU Data Implementation
 *       All of the platform-specific implementation for mapping data files
 *         is here.  The rest of the ICU Data implementation uses only the
 *         wrapper functions.
 *
 *----------------------------------------------------------------------------*/
/* Defines _XOPEN_SOURCE for access to POSIX functions.
 * Must be before any other #includes. */







/* memory-mapping base definitions ------------------------------------------ */

#if MAP_IMPLEMENTATION==MAP_WIN32
#ifndef WIN32_LEAN_AND_MEAN
#   define WIN32_LEAN_AND_MEAN
#endif
#   define VC_EXTRALEAN
#   define NOUSER
#   define NOSERVICE
#   define NOIME
#   define NOMCX

#   if U_PLATFORM_HAS_WINUWP_API == 1
        // Some previous versions of the Windows 10 SDK don't expose various APIs for UWP applications
        // to use, even though UWP apps are allowed to call and use them.  Temporarily change the
        // WINAPI family partition below to Desktop, so that function declarations are visible for UWP.
#       include <winapifamily.h>
#       if !(WINAPI_FAMILY_PARTITION(WINAPI_PARTITION_DESKTOP | WINAPI_PARTITION_SYSTEM))
#           pragma push_macro("WINAPI_PARTITION_DESKTOP")
#           undef WINAPI_PARTITION_DESKTOP
#           define WINAPI_PARTITION_DESKTOP 1
#           define CHANGED_WINAPI_PARTITION_DESKTOP_VALUE
#       endif
#   endif

#   include <windows.h>

#   if U_PLATFORM_HAS_WINUWP_API == 1 && defined(CHANGED_WINAPI_PARTITION_DESKTOP_VALUE)
#       pragma pop_macro("WINAPI_PARTITION_DESKTOP")
#   endif



typedef HANDLE MemoryMap;

#   define IS_MAP(map) ((map)!=nullptr)

#elif MAP_IMPLEMENTATION==MAP_POSIX || MAP_IMPLEMENTATION==MAP_390DLL
    typedef size_t MemoryMap;

#   define IS_MAP(map) ((map)!=0)

#   include <unistd.h>
#   include <sys/mman.h>
#   include <sys/stat.h>
#   include <fcntl.h>

#   ifndef MAP_FAILED
#       define MAP_FAILED ((void*)-1)
#   endif

#   if MAP_IMPLEMENTATION==MAP_390DLL
        /*   No memory mapping for 390 batch mode.  Fake it using dll loading.  */
#       include <dll.h>



#       define LIB_PREFIX "lib"
#       define LIB_SUFFIX ".dll"
        /* This is inconvenient until we figure out what to do with U_ICUDATA_NAME in utypes.h */
#       define U_ICUDATA_ENTRY_NAME "icudt" U_ICU_VERSION_SHORT U_LIB_SUFFIX_C_NAME_STRING "_dat"
#   endif
#elif MAP_IMPLEMENTATION==MAP_STDIO
#   include <stdio.h>


    typedef void *MemoryMap;

#   define IS_MAP(map) ((map)!=nullptr)
#endif

/*----------------------------------------------------------------------------*
 *                                                                            *
 *   Memory Mapped File support.  Platform dependent implementation of        *
 *                           functions used by the rest of the implementation.*
 *                                                                            *
 *----------------------------------------------------------------------------*/
#if MAP_IMPLEMENTATION==MAP_NONE
    U_CFUNC UBool
    uprv_mapFile(UDataMemory *pData, const char *path, UErrorCode *status) {
        if (U_FAILURE(*status)) {
            return FALSE;
        }
        UDataMemory_init(pData); /* Clear the output struct. */
        return FALSE;            /* no file access */
    }

    U_CFUNC void uprv_unmapFile(UDataMemory *pData) {
        /* nothing to do */
    }
#elif MAP_IMPLEMENTATION==MAP_WIN32
    U_CFUNC UBool
    uprv_mapFile(
         UDataMemory *pData,    /* Fill in with info on the result doing the mapping. */
                                /*   Output only; any original contents are cleared.  */
         const char *path,      /* File path to be opened/mapped.                     */
         UErrorCode *status     /* Error status, used to report out-of-memory errors. */
         )
    {
        if (U_FAILURE(*status)) {
            return FALSE;
        }

        HANDLE map = nullptr;
        HANDLE file = INVALID_HANDLE_VALUE;

        UDataMemory_init(pData); /* Clear the output struct.        */

        /* open the input file */
#if U_PLATFORM_HAS_WINUWP_API == 0
        // Note: In the non-UWP code-path (ie: Win32), the value of the path variable might have come from 
        // the CRT 'getenv' function, and would be therefore be encoded in the default ANSI code page.
        // This means that we can't call the *W version of API below, whereas in the UWP code-path
        // there is no 'getenv' call, and thus the string will be only UTF-8/Invariant characters.
        file=CreateFileA(path, GENERIC_READ, FILE_SHARE_READ, nullptr,
            OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL|FILE_FLAG_RANDOM_ACCESS, nullptr);
#else
        // Convert from UTF-8 string to UTF-16 string.
        wchar_t utf16Path[MAX_PATH];
        int32_t pathUtf16Len = 0;
        u_strFromUTF8(reinterpret_cast<UChar*>(utf16Path), static_cast<int32_t>(UPRV_LENGTHOF(utf16Path)), &pathUtf16Len, path, -1, status);

        if (U_FAILURE(*status)) {
            return FALSE;
        }
        if (*status == U_STRING_NOT_TERMINATED_WARNING) {
            // Report back an error instead of a warning.
            *status = U_BUFFER_OVERFLOW_ERROR;
            return FALSE;
        }

        file = CreateFileW(utf16Path, GENERIC_READ, FILE_SHARE_READ, nullptr,
            OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS, nullptr);
#endif
        if (file == INVALID_HANDLE_VALUE) {
            // If we failed to open the file due to an out-of-memory error, then we want
            // to report that error back to the caller.
            if (HRESULT_FROM_WIN32(GetLastError()) == E_OUTOFMEMORY) {
                *status = U_MEMORY_ALLOCATION_ERROR;
            }
            return FALSE;
        }

        // Note: We use NULL/nullptr for lpAttributes parameter below.
        // This means our handle cannot be inherited and we will get the default security descriptor.
        /* create an unnamed Windows file-mapping object for the specified file */
        map = CreateFileMappingW(file, nullptr, PAGE_READONLY, 0, 0, nullptr);

        CloseHandle(file);
        if (map == nullptr) {
            // If we failed to create the mapping due to an out-of-memory error, then 
            // we want to report that error back to the caller.
            if (HRESULT_FROM_WIN32(GetLastError()) == E_OUTOFMEMORY) {
                *status = U_MEMORY_ALLOCATION_ERROR;
            }
            return FALSE;
        }

        /* map a view of the file into our address space */
        pData->pHeader = reinterpret_cast<const DataHeader *>(MapViewOfFile(map, FILE_MAP_READ, 0, 0, 0));
        if (pData->pHeader == nullptr) {
            CloseHandle(map);
            return FALSE;
        }
        pData->map = map;
        return TRUE;
    }

    U_CFUNC void
    uprv_unmapFile(UDataMemory *pData) {
        if (pData != nullptr && pData->map != nullptr) {
            UnmapViewOfFile(pData->pHeader);
            CloseHandle(pData->map);
            pData->pHeader = nullptr;
            pData->map = nullptr;
        }
    }



#elif MAP_IMPLEMENTATION==MAP_POSIX
    U_CFUNC UBool
    uprv_mapFile(UDataMemory *pData, const char *path, UErrorCode *status) {
        int fd;
        int length;
        struct stat mystat;
        void *data;

        if (U_FAILURE(*status)) {
            return FALSE;
        }

        UDataMemory_init(pData); /* Clear the output struct.        */

        /* determine the length of the file */
        if(stat(path, &mystat)!=0 || mystat.st_size<=0) {
            return FALSE;
        }
        length=mystat.st_size;

        /* open the file */
        fd=open(path, O_RDONLY);
        if(fd==-1) {
            return FALSE;
        }

        /* get a view of the mapping */
#if U_PLATFORM != U_PF_HPUX
        data=mmap(0, length, PROT_READ, MAP_SHARED,  fd, 0);
#else
        data=mmap(0, length, PROT_READ, MAP_PRIVATE, fd, 0);
#endif
        close(fd); /* no longer needed */
        if(data==MAP_FAILED) {
            // Possibly check the errno value for ENOMEM, and report U_MEMORY_ALLOCATION_ERROR?
            return FALSE;
        }

        pData->map = (char *)data + length;
        pData->pHeader=(const DataHeader *)data;
        pData->mapAddr = data;
#if U_PLATFORM == U_PF_IPHONE
        posix_madvise(data, length, POSIX_MADV_RANDOM);
#endif
        return TRUE;
    }

    U_CFUNC void
    uprv_unmapFile(UDataMemory *pData) {
        if(pData!=nullptr && pData->map!=nullptr) {
            size_t dataLen = (char *)pData->map - (char *)pData->mapAddr;
            if(munmap(pData->mapAddr, dataLen)==-1) {
            }
            pData->pHeader=nullptr;
            pData->map=0;
            pData->mapAddr=nullptr;
        }
    }



#elif MAP_IMPLEMENTATION==MAP_STDIO
    /* copy of the filestrm.c/T_FileStream_size() implementation */
    static int32_t
    umap_fsize(FILE *f) {
        int32_t savedPos = ftell(f);
        int32_t size = 0;

        /*Changes by Bertrand A. D. doesn't affect the current position
        goes to the end of the file before ftell*/
        fseek(f, 0, SEEK_END);
        size = (int32_t)ftell(f);
        fseek(f, savedPos, SEEK_SET);
        return size;
    }

    U_CFUNC UBool
    uprv_mapFile(UDataMemory *pData, const char *path, UErrorCode *status) {
        FILE *file;
        int32_t fileLength;
        void *p;

        if (U_FAILURE(*status)) {
            return FALSE;
        }

        UDataMemory_init(pData); /* Clear the output struct.        */
        /* open the input file */
        file=fopen(path, "rb");
        if(file==nullptr) {
            return FALSE;
        }

        /* get the file length */
        fileLength=umap_fsize(file);
        if(ferror(file) || fileLength<=20) {
            fclose(file);
            return FALSE;
        }

        /* allocate the memory to hold the file data */
        p=uprv_malloc(fileLength);
        if(p==nullptr) {
            fclose(file);
            *status = U_MEMORY_ALLOCATION_ERROR;
            return FALSE;
        }

        /* read the file */
        if(fileLength!=fread(p, 1, fileLength, file)) {
            uprv_free(p);
            fclose(file);
            return FALSE;
        }

        fclose(file);
        pData->map=p;
        pData->pHeader=(const DataHeader *)p;
        pData->mapAddr=p;
        return TRUE;
    }

    U_CFUNC void
    uprv_unmapFile(UDataMemory *pData) {
        if(pData!=nullptr && pData->map!=nullptr) {
            uprv_free(pData->map);
            pData->map     = nullptr;
            pData->mapAddr = nullptr;
            pData->pHeader = nullptr;
        }
    }


#elif MAP_IMPLEMENTATION==MAP_390DLL
    /*  390 specific Library Loading.
     *  This is the only platform left that dynamically loads an ICU Data Library.
     *  All other platforms use .data files when dynamic loading is required, but
     *  this turn out to be awkward to support in 390 batch mode.
     *
     *  The idea here is to hide the fact that 390 is using dll loading from the
     *   rest of ICU, and make it look like there is file loading happening.
     *
     */

    static char *strcpy_returnEnd(char *dest, const char *src)
    {
        while((*dest=*src)!=0) {
            ++dest;
            ++src;
        }
        return dest;
    }
    
    /*------------------------------------------------------------------------------
     *                                                                              
     *  computeDirPath   given a user-supplied path of an item to be opened,             
     *                         compute and return 
     *                            - the full directory path to be used 
     *                              when opening the file.
     *                            - Pointer to null at end of above returned path    
     *
     *                       Parameters:
     *                          path:        input path.  Buffer is not altered.
     *                          pathBuffer:  Output buffer.  Any contents are overwritten.
     *
     *                       Returns:
     *                          Pointer to null termination in returned pathBuffer.
     *
     *                    TODO:  This works the way ICU historically has, but the
     *                           whole data fallback search path is so complicated that
     *                           probably almost no one will ever really understand it,
     *                           the potential for confusion is large.  (It's not just 
     *                           this one function, but the whole scheme.)
     *                            
     *------------------------------------------------------------------------------*/
    static char *uprv_computeDirPath(const char *path, char *pathBuffer)
    {
        char   *finalSlash;       /* Ptr to last dir separator in input path, or null if none. */
        int32_t pathLen;          /* Length of the returned directory path                     */
        
        finalSlash = 0;
        if (path != 0) {
            finalSlash = uprv_strrchr(path, U_FILE_SEP_CHAR);
        }
        
        *pathBuffer = 0;
        if (finalSlash == 0) {
        /* No user-supplied path.  
            * Copy the ICU_DATA path to the path buffer and return that*/
            const char *icuDataDir;
            icuDataDir=u_getDataDirectory();
            if(icuDataDir!=nullptr && *icuDataDir!=0) {
                return strcpy_returnEnd(pathBuffer, icuDataDir);
            } else {
                /* there is no icuDataDir either.  Just return the empty pathBuffer. */
                return pathBuffer;
            }
        } 
        
        /* User supplied path did contain a directory portion.
        * Copy it to the output path buffer */
        pathLen = (int32_t)(finalSlash - path + 1);
        uprv_memcpy(pathBuffer, path, pathLen);
        *(pathBuffer+pathLen) = 0;
        return pathBuffer+pathLen;
    }
    

#   define DATA_TYPE "dat"

    U_CFUNC UBool uprv_mapFile(UDataMemory *pData, const char *path, UErrorCode *status) {
        const char *inBasename;
        char *basename;
        char pathBuffer[1024];
        const DataHeader *pHeader;
        dllhandle *handle;
        void *val=0;

        if (U_FAILURE(*status)) {
            return FALSE;
        }

        inBasename=uprv_strrchr(path, U_FILE_SEP_CHAR);
        if(inBasename==nullptr) {
            inBasename = path;
        } else {
            inBasename++;
        }
        basename=uprv_computeDirPath(path, pathBuffer);
        if(uprv_strcmp(inBasename, U_ICUDATA_NAME".dat") != 0) {
            /* must mmap file... for build */
            int fd;
            int length;
            struct stat mystat;
            void *data;
            UDataMemory_init(pData); /* Clear the output struct. */

            /* determine the length of the file */
            if(stat(path, &mystat)!=0 || mystat.st_size<=0) {
                return FALSE;
            }
            length=mystat.st_size;

            /* open the file */
            fd=open(path, O_RDONLY);
            if(fd==-1) {
                return FALSE;
            }

            /* get a view of the mapping */
            data=mmap(0, length, PROT_READ, MAP_PRIVATE, fd, 0);
            close(fd); /* no longer needed */
            if(data==MAP_FAILED) {
                // Possibly check the errorno value for ENOMEM, and report U_MEMORY_ALLOCATION_ERROR?
                return FALSE;
            }
            pData->map = (char *)data + length;
            pData->pHeader=(const DataHeader *)data;
            pData->mapAddr = data;
            return TRUE;
        }

#       ifdef OS390BATCH
            /* ### hack: we still need to get u_getDataDirectory() fixed
            for OS/390 (batch mode - always return "//"? )
            and this here straightened out with LIB_PREFIX and LIB_SUFFIX (both empty?!)
            This is probably due to the strange file system on OS/390.  It's more like
            a database with short entry names than a typical file system. */
            /* U_ICUDATA_NAME should always have the correct name */
            /* BUT FOR BATCH MODE IT IS AN EXCEPTION BECAUSE */
            /* THE FIRST THREE LETTERS ARE PREASSIGNED TO THE */
            /* PROJECT!!!!! */
            uprv_strcpy(pathBuffer, "IXMI" U_ICU_VERSION_SHORT "DA");
#       else
            /* set up the library name */
            uprv_strcpy(basename, LIB_PREFIX U_LIBICUDATA_NAME U_ICU_VERSION_SHORT LIB_SUFFIX);
#       endif

#       ifdef UDATA_DEBUG
             fprintf(stderr, "dllload: %s ", pathBuffer);
#       endif

        handle=dllload(pathBuffer);

#       ifdef UDATA_DEBUG
               fprintf(stderr, " -> %08X\n", handle );
#       endif

        if(handle != nullptr) {
               /* we have a data DLL - what kind of lookup do we need here? */
               /* try to find the Table of Contents */
               UDataMemory_init(pData); /* Clear the output struct.        */
               val=dllqueryvar((dllhandle*)handle, U_ICUDATA_ENTRY_NAME);
               if(val == 0) {
                    /* failed... so keep looking */
                    return FALSE;
               }
#              ifdef UDATA_DEBUG
                    fprintf(stderr, "dllqueryvar(%08X, %s) -> %08X\n", handle, U_ICUDATA_ENTRY_NAME, val);
#              endif

               pData->pHeader=(const DataHeader *)val;
               return TRUE;
         } else {
               return FALSE; /* no handle */
         }
    }

    U_CFUNC void uprv_unmapFile(UDataMemory *pData) {
        if(pData!=nullptr && pData->map!=nullptr) {
            uprv_free(pData->map);
            pData->map     = nullptr;
            pData->mapAddr = nullptr;
            pData->pHeader = nullptr;
        }   
    }

#else
#   error MAP_IMPLEMENTATION is set incorrectly
#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 2005-2015, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File WINTZ.CPP
*
********************************************************************************
*/



#if U_PLATFORM_USES_ONLY_WIN32_API









#ifndef WIN32_LEAN_AND_MEAN
#   define WIN32_LEAN_AND_MEAN
#endif
#   define VC_EXTRALEAN
#   define NOUSER
#   define NOSERVICE
#   define NOIME
#   define NOMCX
#include <windows.h>

U_NAMESPACE_BEGIN

// The max size of TimeZoneKeyName is 128, defined in DYNAMIC_TIME_ZONE_INFORMATION
#define MAX_TIMEZONE_ID_LENGTH 128

/**
* Main Windows time zone detection function.
* Returns the Windows time zone converted to an ICU time zone as a heap-allocated buffer, or nullptr upon failure.
* Note: We use the Win32 API GetDynamicTimeZoneInformation to get the current time zone info.
* This API returns a non-localized time zone name, which we can then map to an ICU time zone name.
*/
U_INTERNAL const char* U_EXPORT2
uprv_detectWindowsTimeZone()
{
    UErrorCode status = U_ZERO_ERROR;
    char* icuid = nullptr;
    char dynamicTZKeyName[MAX_TIMEZONE_ID_LENGTH];
    char tmpid[MAX_TIMEZONE_ID_LENGTH];
    int32_t len;
    int id = GEOID_NOT_AVAILABLE;
    int errorCode;
    wchar_t ISOcodeW[3] = {}; /* 2 letter ISO code in UTF-16 */
    char ISOcode[3] = {}; /* 2 letter ISO code in UTF-8 */

    DYNAMIC_TIME_ZONE_INFORMATION dynamicTZI;
    uprv_memset(&dynamicTZI, 0, sizeof(dynamicTZI));
    uprv_memset(dynamicTZKeyName, 0, sizeof(dynamicTZKeyName));
    uprv_memset(tmpid, 0, sizeof(tmpid));

    /* Obtain TIME_ZONE_INFORMATION from the API and get the non-localized time zone name. */
    if (TIME_ZONE_ID_INVALID == GetDynamicTimeZoneInformation(&dynamicTZI)) {
        return nullptr;
    }

    id = GetUserGeoID(GEOCLASS_NATION);
    errorCode = GetGeoInfoW(id, GEO_ISO2, ISOcodeW, 3, 0);

    // convert from wchar_t* (UTF-16 on Windows) to char* (UTF-8).
    u_strToUTF8(ISOcode, UPRV_LENGTHOF(ISOcode), nullptr,
        reinterpret_cast<const UChar*>(ISOcodeW), UPRV_LENGTHOF(ISOcodeW), &status);

    LocalUResourceBundlePointer bundle(ures_openDirect(nullptr, "windowsZones", &status));
    ures_getByKey(bundle.getAlias(), "mapTimezones", bundle.getAlias(), &status);

    // convert from wchar_t* (UTF-16 on Windows) to char* (UTF-8).
    u_strToUTF8(dynamicTZKeyName, UPRV_LENGTHOF(dynamicTZKeyName), nullptr,
        reinterpret_cast<const UChar*>(dynamicTZI.TimeZoneKeyName), -1, &status);

    if (U_FAILURE(status)) {
        return nullptr;
    }

    if (dynamicTZI.TimeZoneKeyName[0] != 0) {
        StackUResourceBundle winTZ;
        ures_getByKey(bundle.getAlias(), dynamicTZKeyName, winTZ.getAlias(), &status);

        if (U_SUCCESS(status)) {
            const UChar* icuTZ = nullptr;
            if (errorCode != 0) {
                icuTZ = ures_getStringByKey(winTZ.getAlias(), ISOcode, &len, &status);
            }
            if (errorCode == 0 || icuTZ == nullptr) {
                /* fallback to default "001" and reset status */
                status = U_ZERO_ERROR;
                icuTZ = ures_getStringByKey(winTZ.getAlias(), "001", &len, &status);
            }

            if (U_SUCCESS(status)) {
                int index = 0;

                while (!(*icuTZ == '\0' || *icuTZ == ' ')) {
                    // time zone IDs only contain ASCII invariant characters.
                    tmpid[index++] = (char)(*icuTZ++);
                }
                tmpid[index] = '\0';
            }
        }
    }

    // Copy the timezone ID to icuid to be returned.
    if (tmpid[0] != 0) {
        icuid = uprv_strdup(tmpid);
    }

    return icuid;
}

U_NAMESPACE_END
#endif /* U_PLATFORM_USES_ONLY_WIN32_API  */
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
*******************************************************************************
* Copyright (C) 2003-2009,2012,2016 International Business Machines Corporation and
* others. All Rights Reserved.
*******************************************************************************
*
* File JAPANCAL.CPP
*
* Modification History:
*  05/16/2003    srl     copied from buddhcal.cpp
*
*/



#if !UCONFIG_NO_FORMATTING
#if U_PLATFORM_HAS_WINUWP_API == 0
#include <stdlib.h> // getenv() is not available in UWP env
#else
#ifndef WIN32_LEAN_AND_MEAN
#   define WIN32_LEAN_AND_MEAN
#endif
#   define VC_EXTRALEAN
#   define NOUSER
#   define NOSERVICE
#   define NOIME
#   define NOMCX
#include <windows.h>
#endif









static icu::EraRules * gJapaneseEraRules = nullptr;
static icu::UInitOnce gJapaneseEraRulesInitOnce = U_INITONCE_INITIALIZER;
static int32_t gCurrentEra = 0;

U_CDECL_BEGIN
static UBool japanese_calendar_cleanup(void) {
    if (gJapaneseEraRules) {
        delete gJapaneseEraRules;
        gJapaneseEraRules = nullptr;
    }
    gCurrentEra = 0;
    gJapaneseEraRulesInitOnce.reset();
    return TRUE;
}
U_CDECL_END

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(JapaneseCalendar)

static const int32_t japancal_kGregorianEpoch = 1970;    // used as the default value of EXTENDED_YEAR
static const char* TENTATIVE_ERA_VAR_NAME = "ICU_ENABLE_TENTATIVE_ERA";


// Export the following for use by test code.
UBool JapaneseCalendar::enableTentativeEra() {
    // Although start date of next Japanese era is planned ahead, a name of
    // new era might not be available. This implementation allows tester to
    // check a new era without era names by settings below (in priority order).
    // By default, such tentative era is disabled.

    // 1. Environment variable ICU_ENABLE_TENTATIVE_ERA=true or false

    UBool includeTentativeEra = FALSE;

#if U_PLATFORM_HAS_WINUWP_API == 1
    // UWP doesn't allow access to getenv(), but we can call GetEnvironmentVariableW to do the same thing.
    UChar varName[26] = {};
    u_charsToUChars(TENTATIVE_ERA_VAR_NAME, varName, static_cast<int32_t>(uprv_strlen(TENTATIVE_ERA_VAR_NAME)));
    WCHAR varValue[5] = {};
    DWORD ret = GetEnvironmentVariableW(reinterpret_cast<WCHAR*>(varName), varValue, UPRV_LENGTHOF(varValue));
    if ((ret == 4) && (_wcsicmp(varValue, L"true") == 0)) {
        includeTentativeEra = TRUE;
    }
#else
    char *envVarVal = getenv(TENTATIVE_ERA_VAR_NAME);
    if (envVarVal != NULL && uprv_stricmp(envVarVal, "true") == 0) {
        includeTentativeEra = TRUE;
    }
#endif
    return includeTentativeEra;
}


// Initialize global Japanese era data
static void U_CALLCONV initializeEras(UErrorCode &status) {
    gJapaneseEraRules = EraRules::createInstance("japanese", JapaneseCalendar::enableTentativeEra(), status);
    if (U_FAILURE(status)) {
        return;
    }
    gCurrentEra = gJapaneseEraRules->getCurrentEraIndex();
}

static void init(UErrorCode &status) {
    umtx_initOnce(gJapaneseEraRulesInitOnce, &initializeEras, status);
    ucln_i18n_registerCleanup(UCLN_I18N_JAPANESE_CALENDAR, japanese_calendar_cleanup);
}

/* Some platforms don't like to export constants, like old Palm OS and some z/OS configurations. */
uint32_t JapaneseCalendar::getCurrentEra() {
    return gCurrentEra;
}

JapaneseCalendar::JapaneseCalendar(const Locale& aLocale, UErrorCode& success)
:   GregorianCalendar(aLocale, success)
{
    init(success);
    setTimeInMillis(getNow(), success); // Call this again now that the vtable is set up properly.
}

JapaneseCalendar::~JapaneseCalendar()
{
}

JapaneseCalendar::JapaneseCalendar(const JapaneseCalendar& source)
: GregorianCalendar(source)
{
    UErrorCode status = U_ZERO_ERROR;
    init(status);
    U_ASSERT(U_SUCCESS(status));
}

JapaneseCalendar& JapaneseCalendar::operator= ( const JapaneseCalendar& right)
{
    GregorianCalendar::operator=(right);
    return *this;
}

JapaneseCalendar* JapaneseCalendar::clone() const
{
    return new JapaneseCalendar(*this);
}

const char *JapaneseCalendar::getType() const
{
    return "japanese";
}

int32_t JapaneseCalendar::getDefaultMonthInYear(int32_t eyear)
{
    int32_t era = internalGetEra();
    // TODO do we assume we can trust 'era'?  What if it is denormalized?

    int32_t month = 0;

    // Find out if we are at the edge of an era
    int32_t eraStart[3] = { 0,0,0 };
    UErrorCode status = U_ZERO_ERROR;
    gJapaneseEraRules->getStartDate(era, eraStart, status);
    U_ASSERT(U_SUCCESS(status));
    if(eyear == eraStart[0]) {
        // Yes, we're in the first year of this era.
        return eraStart[1]  // month
                -1;         // return 0-based month
    }

    return month;
}

int32_t JapaneseCalendar::getDefaultDayInMonth(int32_t eyear, int32_t month)
{
    int32_t era = internalGetEra();
    int32_t day = 1;

    int32_t eraStart[3] = { 0,0,0 };
    UErrorCode status = U_ZERO_ERROR;
    gJapaneseEraRules->getStartDate(era, eraStart, status);
    U_ASSERT(U_SUCCESS(status));
    if(eyear == eraStart[0]) {
        if(month == eraStart[1] - 1) {
            return eraStart[2];
        }
    }

    return day;
}


int32_t JapaneseCalendar::internalGetEra() const
{
    return internalGet(UCAL_ERA, gCurrentEra);
}

int32_t JapaneseCalendar::handleGetExtendedYear()
{
    // EXTENDED_YEAR in JapaneseCalendar is a Gregorian year
    // The default value of EXTENDED_YEAR is 1970 (Showa 45)
    int32_t year;

    if (newerField(UCAL_EXTENDED_YEAR, UCAL_YEAR) == UCAL_EXTENDED_YEAR &&
        newerField(UCAL_EXTENDED_YEAR, UCAL_ERA) == UCAL_EXTENDED_YEAR) {
        year = internalGet(UCAL_EXTENDED_YEAR, japancal_kGregorianEpoch);
    } else {
        UErrorCode status = U_ZERO_ERROR;
        int32_t eraStartYear = gJapaneseEraRules->getStartYear(internalGet(UCAL_ERA, gCurrentEra), status);
        U_ASSERT(U_SUCCESS(status));

        // extended year is a gregorian year, where 1 = 1AD,  0 = 1BC, -1 = 2BC, etc
        year = internalGet(UCAL_YEAR, 1)    // pin to minimum of year 1 (first year)
            + eraStartYear                  // add gregorian starting year
            - 1;                            // Subtract one because year starts at 1
    }
    return year;
}


void JapaneseCalendar::handleComputeFields(int32_t julianDay, UErrorCode& status)
{
    //Calendar::timeToFields(theTime, quick, status);
    GregorianCalendar::handleComputeFields(julianDay, status);
    int32_t year = internalGet(UCAL_EXTENDED_YEAR); // Gregorian year
    int32_t eraIdx = gJapaneseEraRules->getEraIndex(year, internalGet(UCAL_MONTH) + 1, internalGet(UCAL_DAY_OF_MONTH), status);

    internalSet(UCAL_ERA, eraIdx);
    internalSet(UCAL_YEAR, year - gJapaneseEraRules->getStartYear(eraIdx, status) + 1);
}

/*
Disable pivoting
*/
UBool JapaneseCalendar::haveDefaultCentury() const
{
    return FALSE;
}

UDate JapaneseCalendar::defaultCenturyStart() const
{
    return 0;// WRONG
}

int32_t JapaneseCalendar::defaultCenturyStartYear() const
{
    return 0;
}

int32_t JapaneseCalendar::handleGetLimit(UCalendarDateFields field, ELimitType limitType) const
{
    switch(field) {
    case UCAL_ERA:
        if (limitType == UCAL_LIMIT_MINIMUM || limitType == UCAL_LIMIT_GREATEST_MINIMUM) {
            return 0;
        }
        return gJapaneseEraRules->getNumberOfEras() - 1; // max known era, not gCurrentEra
    case UCAL_YEAR:
        {
            switch (limitType) {
            case UCAL_LIMIT_MINIMUM:
            case UCAL_LIMIT_GREATEST_MINIMUM:
                return 1;
            case UCAL_LIMIT_LEAST_MAXIMUM:
                return 1;
            case  UCAL_LIMIT_COUNT: //added to avoid warning
            case UCAL_LIMIT_MAXIMUM:
            {
                UErrorCode status = U_ZERO_ERROR;
                int32_t eraStartYear = gJapaneseEraRules->getStartYear(gCurrentEra, status);
                U_ASSERT(U_SUCCESS(status));
                return GregorianCalendar::handleGetLimit(UCAL_YEAR, UCAL_LIMIT_MAXIMUM) - eraStartYear;
            }
            default:
                return 1;    // Error condition, invalid limitType
            }
        }
    default:
        return GregorianCalendar::handleGetLimit(field,limitType);
    }
}

int32_t JapaneseCalendar::getActualMaximum(UCalendarDateFields field, UErrorCode& status) const {
    if (field == UCAL_YEAR) {
        int32_t era = get(UCAL_ERA, status);
        if (U_FAILURE(status)) {
            return 0; // error case... any value
        }
        if (era == gJapaneseEraRules->getNumberOfEras() - 1) { // max known era, not gCurrentEra
            // TODO: Investigate what value should be used here - revisit after 4.0.
            return handleGetLimit(UCAL_YEAR, UCAL_LIMIT_MAXIMUM);
        } else {
            int32_t nextEraStart[3] = { 0,0,0 };
            gJapaneseEraRules->getStartDate(era + 1, nextEraStart, status);
            int32_t nextEraYear = nextEraStart[0];
            int32_t nextEraMonth = nextEraStart[1]; // 1-base
            int32_t nextEraDate = nextEraStart[2];

            int32_t eraStartYear = gJapaneseEraRules->getStartYear(era, status);
            int32_t maxYear = nextEraYear - eraStartYear + 1;   // 1-base
            if (nextEraMonth == 1 && nextEraDate == 1) {
                // Subtract 1, because the next era starts at Jan 1
                maxYear--;
            }
            return maxYear;
        }
    }
    return GregorianCalendar::getActualMaximum(field, status);
}

U_NAMESPACE_END

#endif
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 2005-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File WINDTFMT.CPP
*
********************************************************************************
*/



#if U_PLATFORM_USES_ONLY_WIN32_API

#if !UCONFIG_NO_FORMATTING

















// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 2008-2011, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File WINTZIMPL.H
*
********************************************************************************
*/




#if U_PLATFORM_USES_ONLY_WIN32_API
/**
 * \file
 * \brief C API: Utilities for dealing w/ Windows time zones.
 */
U_CDECL_BEGIN
/* Forward declarations for Windows types... */
typedef struct _TIME_ZONE_INFORMATION TIME_ZONE_INFORMATION;
U_CDECL_END

/*
 * This method was moved over from common/wintz.h to allow for access to i18n functions
 * needed to get the Windows time zone information without using static tables.
 */
U_CAPI UBool U_EXPORT2
uprv_getWindowsTimeZoneInfo(TIME_ZONE_INFORMATION *zoneInfo, const UChar *icuid, int32_t length);


#endif /* U_PLATFORM_USES_ONLY_WIN32_API */



#ifndef WIN32_LEAN_AND_MEAN
#   define WIN32_LEAN_AND_MEAN
#endif
#   define VC_EXTRALEAN
#   define NOUSER
#   define NOSERVICE
#   define NOIME
#   define NOMCX
#include <windows.h>

U_NAMESPACE_BEGIN

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(Win32DateFormat)

#define NEW_ARRAY(type,count) (type *) uprv_malloc((count) * sizeof(type))
#define DELETE_ARRAY(array) uprv_free((void *) (array))

#define windtfmt_STACK_BUFFER_SIZE 64

UnicodeString* Win32DateFormat::getTimeDateFormat(const Calendar *cal, const Locale *locale, UErrorCode &status) const
{
    UnicodeString *result = NULL;
    const char *type = cal->getType();
    const char *base = locale->getBaseName();
    UResourceBundle *topBundle = ures_open((char *) 0, base, &status);
    UResourceBundle *calBundle = ures_getByKey(topBundle, "calendar", NULL, &status);
    UResourceBundle *typBundle = ures_getByKeyWithFallback(calBundle, type, NULL, &status);
    UResourceBundle *patBundle = ures_getByKeyWithFallback(typBundle, "DateTimePatterns", NULL, &status);

    if (status == U_MISSING_RESOURCE_ERROR) {
        status = U_ZERO_ERROR;
        typBundle = ures_getByKeyWithFallback(calBundle, "gregorian", typBundle, &status);
        patBundle = ures_getByKeyWithFallback(typBundle, "DateTimePatterns", patBundle, &status);
    }

    if (U_FAILURE(status)) {
        static const UChar defaultPattern[] = {0x007B, 0x0031, 0x007D, 0x0020, 0x007B, 0x0030, 0x007D, 0x0000}; // "{1} {0}"
        return new UnicodeString(defaultPattern, UPRV_LENGTHOF(defaultPattern));
    }

    int32_t resStrLen = 0;
    int32_t glueIndex = DateFormat::kDateTime;
    int32_t patSize = ures_getSize(patBundle);
    if (patSize >= (DateFormat::kDateTimeOffset + DateFormat::kShort + 1)) {
        // Get proper date time format
        glueIndex = (int32_t)(DateFormat::kDateTimeOffset + (fDateStyle - DateFormat::kDateOffset));
    }
    const UChar *resStr = ures_getStringByIndex(patBundle, glueIndex, &resStrLen, &status);

    result = new UnicodeString(TRUE, resStr, resStrLen);

    ures_close(patBundle);
    ures_close(typBundle);
    ures_close(calBundle);
    ures_close(topBundle);

    return result;
}

// TODO: This is copied in both winnmfmt.cpp and windtfmt.cpp, but really should
// be factored out into a common helper for both.
static UErrorCode windtfmt_GetEquivalentWindowsLocaleName(const Locale& locale, UnicodeString** buffer)
{
    UErrorCode status = U_ZERO_ERROR;
    char asciiBCP47Tag[LOCALE_NAME_MAX_LENGTH] = {};

    // Convert from names like "en_CA" and "de_DE@collation=phonebook" to "en-CA" and "de-DE-u-co-phonebk".
    (void)uloc_toLanguageTag(locale.getName(), asciiBCP47Tag, UPRV_LENGTHOF(asciiBCP47Tag), FALSE, &status);

    if (U_SUCCESS(status))
    {
        // Need it to be UTF-16, not 8-bit
        // TODO: This seems like a good thing for a helper
        wchar_t bcp47Tag[LOCALE_NAME_MAX_LENGTH] = {};
        int32_t i;
        for (i = 0; i < UPRV_LENGTHOF(bcp47Tag); i++)
        {
            if (asciiBCP47Tag[i] == '\0')
            {
                break;
            }
            else
            {
                // normally just copy the character
                bcp47Tag[i] = static_cast<wchar_t>(asciiBCP47Tag[i]);
            }
        }

        // Ensure it's null terminated
        if (i < (UPRV_LENGTHOF(bcp47Tag) - 1))
        {
            bcp47Tag[i] = L'\0';
        }
        else
        {
            // Ran out of room.
            bcp47Tag[UPRV_LENGTHOF(bcp47Tag) - 1] = L'\0';
        }


        wchar_t windowsLocaleName[LOCALE_NAME_MAX_LENGTH] = {};

        // Note: On Windows versions below 10, there is no support for locale name aliases.
        // This means that it will fail for locales where ICU has a completely different
        // name (like ku vs ckb), and it will also not work for alternate sort locale
        // names like "de-DE-u-co-phonebk".

        // TODO: We could add some sort of exception table for cases like ku vs ckb.

        int length = ResolveLocaleName(bcp47Tag, windowsLocaleName, UPRV_LENGTHOF(windowsLocaleName));

        if (length > 0)
        {
            *buffer = new UnicodeString(windowsLocaleName);
        }
        else
        {
            status = U_UNSUPPORTED_ERROR;
        }
    }
    return status;
}

// TODO: Range-check timeStyle, dateStyle
Win32DateFormat::Win32DateFormat(DateFormat::EStyle timeStyle, DateFormat::EStyle dateStyle, const Locale &locale, UErrorCode &status)
  : DateFormat(), fDateTimeMsg(NULL), fTimeStyle(timeStyle), fDateStyle(dateStyle), fLocale(locale), fZoneID(), fWindowsLocaleName(nullptr)
{
    if (U_SUCCESS(status)) {
        windtfmt_GetEquivalentWindowsLocaleName(locale, &fWindowsLocaleName);
        // Note: In the previous code, it would look up the LCID for the locale, and if
        // the locale was not recognized then it would get an LCID of 0, which is a
        // synonym for LOCALE_USER_DEFAULT on Windows.
        // If the above method fails, then fWindowsLocaleName will remain as nullptr, and
        // then we will pass nullptr to API GetLocaleInfoEx, which is the same as passing
        // LOCALE_USER_DEFAULT.

        fTZI = NEW_ARRAY(TIME_ZONE_INFORMATION, 1);
        uprv_memset(fTZI, 0, sizeof(TIME_ZONE_INFORMATION));
        adoptCalendar(Calendar::createInstance(locale, status));
    }
}

Win32DateFormat::Win32DateFormat(const Win32DateFormat &other)
  : DateFormat(other)
{
    *this = other;
}

Win32DateFormat::~Win32DateFormat()
{
//    delete fCalendar;
    uprv_free(fTZI);
    delete fDateTimeMsg;
    delete fWindowsLocaleName;
}

Win32DateFormat &Win32DateFormat::operator=(const Win32DateFormat &other)
{
    // The following handles fCalendar
    DateFormat::operator=(other);

//    delete fCalendar;

    this->fDateTimeMsg = other.fDateTimeMsg == NULL ? NULL : new UnicodeString(*other.fDateTimeMsg);
    this->fTimeStyle   = other.fTimeStyle;
    this->fDateStyle   = other.fDateStyle;
    this->fLocale      = other.fLocale;
//    this->fCalendar    = other.fCalendar->clone();
    this->fZoneID      = other.fZoneID;

    this->fTZI = NEW_ARRAY(TIME_ZONE_INFORMATION, 1);
    *this->fTZI = *other.fTZI;

    this->fWindowsLocaleName = other.fWindowsLocaleName == NULL ? NULL : new UnicodeString(*other.fWindowsLocaleName);

    return *this;
}

Win32DateFormat *Win32DateFormat::clone() const
{
    return new Win32DateFormat(*this);
}

// TODO: Is just ignoring pos the right thing?
UnicodeString &Win32DateFormat::format(Calendar &cal, UnicodeString &appendTo, FieldPosition & /* pos */) const
{
    FILETIME ft;
    SYSTEMTIME st_gmt;
    SYSTEMTIME st_local;
    TIME_ZONE_INFORMATION tzi = *fTZI;
    UErrorCode status = U_ZERO_ERROR;
    const TimeZone &tz = cal.getTimeZone();
    int64_t uct, uft;

    setTimeZoneInfo(&tzi, tz);

    uct = utmscale_fromInt64((int64_t) cal.getTime(status), UDTS_ICU4C_TIME, &status);
    uft = utmscale_toInt64(uct, UDTS_WINDOWS_FILE_TIME, &status);

    ft.dwLowDateTime =  (DWORD) (uft & 0xFFFFFFFF);
    ft.dwHighDateTime = (DWORD) ((uft >> 32) & 0xFFFFFFFF);

    FileTimeToSystemTime(&ft, &st_gmt);
    SystemTimeToTzSpecificLocalTime(&tzi, &st_gmt, &st_local);


    if (fDateStyle != DateFormat::kNone && fTimeStyle != DateFormat::kNone) {
        UnicodeString date;
        UnicodeString time;
        UnicodeString *pattern = fDateTimeMsg;

        formatDate(&st_local, date);
        formatTime(&st_local, time);

        if (strcmp(fCalendar->getType(), cal.getType()) != 0) {
            pattern = getTimeDateFormat(&cal, &fLocale, status);
        }

        SimpleFormatter(*pattern, 2, 2, status).format(time, date, appendTo, status);
    } else if (fDateStyle != DateFormat::kNone) {
        formatDate(&st_local, appendTo);
    } else if (fTimeStyle != DateFormat::kNone) {
        formatTime(&st_local, appendTo);
    }

    return appendTo;
}

void Win32DateFormat::parse(const UnicodeString& /* text */, Calendar& /* cal */, ParsePosition& pos) const
{
    pos.setErrorIndex(pos.getIndex());
}

void Win32DateFormat::adoptCalendar(Calendar *newCalendar)
{
    if (fCalendar == NULL || strcmp(fCalendar->getType(), newCalendar->getType()) != 0) {
        UErrorCode status = U_ZERO_ERROR;

        if (fDateStyle != DateFormat::kNone && fTimeStyle != DateFormat::kNone) {
            delete fDateTimeMsg;
            fDateTimeMsg = getTimeDateFormat(newCalendar, &fLocale, status);
        }
    }

    delete fCalendar;
    fCalendar = newCalendar;

    fZoneID = setTimeZoneInfo(fTZI, fCalendar->getTimeZone());
}

void Win32DateFormat::setCalendar(const Calendar &newCalendar)
{
    adoptCalendar(newCalendar.clone());
}

void Win32DateFormat::adoptTimeZone(TimeZone *zoneToAdopt)
{
    fZoneID = setTimeZoneInfo(fTZI, *zoneToAdopt);
    fCalendar->adoptTimeZone(zoneToAdopt);
}

void Win32DateFormat::setTimeZone(const TimeZone& zone)
{
    fZoneID = setTimeZoneInfo(fTZI, zone);
    fCalendar->setTimeZone(zone);
}

static const DWORD dfFlags[] = {DATE_LONGDATE, DATE_LONGDATE, DATE_SHORTDATE, DATE_SHORTDATE};

void Win32DateFormat::formatDate(const SYSTEMTIME *st, UnicodeString &appendTo) const
{
    int result=0;
    wchar_t stackBuffer[windtfmt_STACK_BUFFER_SIZE];
    wchar_t *buffer = stackBuffer;
    const wchar_t *localeName = nullptr;

    if (fWindowsLocaleName != nullptr)
    {
        localeName = reinterpret_cast<const wchar_t*>(toOldUCharPtr(fWindowsLocaleName->getTerminatedBuffer()));
    }

    result = GetDateFormatEx(localeName, dfFlags[fDateStyle - kDateOffset], st, NULL, buffer, windtfmt_STACK_BUFFER_SIZE, NULL);

    if (result == 0) {
        if (GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
            int newLength = GetDateFormatEx(localeName, dfFlags[fDateStyle - kDateOffset], st, NULL, NULL, 0, NULL);

            buffer = NEW_ARRAY(wchar_t, newLength);

            GetDateFormatEx(localeName, dfFlags[fDateStyle - kDateOffset], st, NULL, buffer, newLength, NULL);
        }
    }

    appendTo.append((const UChar *)buffer, (int32_t) wcslen(buffer));

    if (buffer != stackBuffer) {
        DELETE_ARRAY(buffer);
    }
}

static const DWORD tfFlags[] = {0, 0, 0, TIME_NOSECONDS};

void Win32DateFormat::formatTime(const SYSTEMTIME *st, UnicodeString &appendTo) const
{
    int result;
    wchar_t stackBuffer[windtfmt_STACK_BUFFER_SIZE];
    wchar_t *buffer = stackBuffer;
    const wchar_t *localeName = nullptr;

    if (fWindowsLocaleName != nullptr)
    {
        localeName = reinterpret_cast<const wchar_t*>(toOldUCharPtr(fWindowsLocaleName->getTerminatedBuffer()));
    }

    result = GetTimeFormatEx(localeName, tfFlags[fTimeStyle], st, NULL, buffer, windtfmt_STACK_BUFFER_SIZE);

    if (result == 0) {
        if (GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
            int newLength = GetTimeFormatEx(localeName, tfFlags[fTimeStyle], st, NULL, NULL, 0);

            buffer = NEW_ARRAY(wchar_t, newLength);

            GetTimeFormatEx(localeName, tfFlags[fTimeStyle], st, NULL, buffer, newLength);
        }
    }

    appendTo.append((const UChar *)buffer, (int32_t) wcslen(buffer));

    if (buffer != stackBuffer) {
        DELETE_ARRAY(buffer);
    }
}

UnicodeString Win32DateFormat::setTimeZoneInfo(TIME_ZONE_INFORMATION *tzi, const TimeZone &zone) const
{
    UnicodeString zoneID;

    zone.getID(zoneID);

    if (zoneID.compare(fZoneID) != 0) {
        UnicodeString icuid;

        zone.getID(icuid);
        if (! uprv_getWindowsTimeZoneInfo(tzi, icuid.getBuffer(), icuid.length())) {
            UBool found = FALSE;
            int32_t ec = TimeZone::countEquivalentIDs(icuid);

            for (int z = 0; z < ec; z += 1) {
                UnicodeString equiv = TimeZone::getEquivalentID(icuid, z);

                found = uprv_getWindowsTimeZoneInfo(tzi, equiv.getBuffer(), equiv.length());
                if (found) {
                    break;
                }
            }

            if (! found) {
                GetTimeZoneInformation(tzi);
            }
        }
    }

    return zoneID;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif // U_PLATFORM_USES_ONLY_WIN32_API

// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 2005-2016, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File WINNMFMT.CPP
*
********************************************************************************
*/



#if U_PLATFORM_USES_ONLY_WIN32_API

#if !UCONFIG_NO_FORMATTING












#ifndef WIN32_LEAN_AND_MEAN
#   define WIN32_LEAN_AND_MEAN
#endif
#   define VC_EXTRALEAN
#   define NOUSER
#   define NOSERVICE
#   define NOIME
#   define NOMCX
#include <windows.h>
#include <stdio.h>

U_NAMESPACE_BEGIN

union FormatInfo
{
    NUMBERFMTW   number;
    CURRENCYFMTW currency;
};

UOBJECT_DEFINE_RTTI_IMPLEMENTATION(Win32NumberFormat)

#define NEW_ARRAY(type,count) (type *) uprv_malloc((count) * sizeof(type))
#define DELETE_ARRAY(array) uprv_free((void *) (array))

#define winnmfmt_STACK_BUFFER_SIZE 32

/*
 * Turns a string of the form "3;2;0" into the grouping UINT
 * needed for NUMBERFMT and CURRENCYFMT. If the string does not
 * end in ";0" then the return value should be multiplied by 10.
 * (e.g. "3" => 30, "3;2" => 320)
 */
static UINT getGrouping(const wchar_t *grouping)
{
    UINT g = 0;
    const wchar_t *s;

    for (s = grouping; *s != L'\0'; s += 1) {
        if (*s > L'0' && *s < L'9') {
            g = g * 10 + (*s - L'0');
        } else if (*s != L';') {
            break;
        }
    }

    if (*s != L'0') {
        g *= 10;
    }

    return g;
}

static void getNumberFormat(NUMBERFMTW *fmt, const wchar_t *windowsLocaleName)
{
    wchar_t buf[10];

    GetLocaleInfoEx(windowsLocaleName, LOCALE_RETURN_NUMBER|LOCALE_IDIGITS, (LPWSTR) &fmt->NumDigits, sizeof(UINT));
    GetLocaleInfoEx(windowsLocaleName, LOCALE_RETURN_NUMBER|LOCALE_ILZERO,  (LPWSTR) &fmt->LeadingZero, sizeof(UINT));

    GetLocaleInfoEx(windowsLocaleName, LOCALE_SGROUPING, (LPWSTR)buf, 10);
    fmt->Grouping = getGrouping(buf);

    fmt->lpDecimalSep = NEW_ARRAY(wchar_t, 6);
    GetLocaleInfoEx(windowsLocaleName, LOCALE_SDECIMAL,  fmt->lpDecimalSep,  6);

    fmt->lpThousandSep = NEW_ARRAY(wchar_t, 6);
    GetLocaleInfoEx(windowsLocaleName, LOCALE_STHOUSAND, fmt->lpThousandSep, 6);

    GetLocaleInfoEx(windowsLocaleName, LOCALE_RETURN_NUMBER|LOCALE_INEGNUMBER, (LPWSTR) &fmt->NegativeOrder, sizeof(UINT));
}

static void freeNumberFormat(NUMBERFMTW *fmt)
{
    if (fmt != NULL) {
        DELETE_ARRAY(fmt->lpThousandSep);
        DELETE_ARRAY(fmt->lpDecimalSep);
    }
}

static void getCurrencyFormat(CURRENCYFMTW *fmt, const wchar_t *windowsLocaleName)
{
    wchar_t buf[10];

    GetLocaleInfoEx(windowsLocaleName, LOCALE_RETURN_NUMBER|LOCALE_ICURRDIGITS, (LPWSTR) &fmt->NumDigits, sizeof(UINT));
    GetLocaleInfoEx(windowsLocaleName, LOCALE_RETURN_NUMBER|LOCALE_ILZERO, (LPWSTR) &fmt->LeadingZero, sizeof(UINT));

    GetLocaleInfoEx(windowsLocaleName, LOCALE_SMONGROUPING, (LPWSTR)buf, sizeof(buf));
    fmt->Grouping = getGrouping(buf);

    fmt->lpDecimalSep = NEW_ARRAY(wchar_t, 6);
    GetLocaleInfoEx(windowsLocaleName, LOCALE_SMONDECIMALSEP,  fmt->lpDecimalSep,  6);

    fmt->lpThousandSep = NEW_ARRAY(wchar_t, 6);
    GetLocaleInfoEx(windowsLocaleName, LOCALE_SMONTHOUSANDSEP, fmt->lpThousandSep, 6);

    GetLocaleInfoEx(windowsLocaleName, LOCALE_RETURN_NUMBER|LOCALE_INEGCURR,  (LPWSTR) &fmt->NegativeOrder, sizeof(UINT));
    GetLocaleInfoEx(windowsLocaleName, LOCALE_RETURN_NUMBER|LOCALE_ICURRENCY, (LPWSTR) &fmt->PositiveOrder, sizeof(UINT));

    fmt->lpCurrencySymbol = NEW_ARRAY(wchar_t, 8);
    GetLocaleInfoEx(windowsLocaleName, LOCALE_SCURRENCY, (LPWSTR) fmt->lpCurrencySymbol, 8);
}

static void freeCurrencyFormat(CURRENCYFMTW *fmt)
{
    if (fmt != NULL) {
        DELETE_ARRAY(fmt->lpCurrencySymbol);
        DELETE_ARRAY(fmt->lpThousandSep);
        DELETE_ARRAY(fmt->lpDecimalSep);
    }
}

// TODO: This is copied in both winnmfmt.cpp and windtfmt.cpp, but really should
// be factored out into a common helper for both.
static UErrorCode winnmfmt_GetEquivalentWindowsLocaleName(const Locale& locale, UnicodeString** buffer)
{
    UErrorCode status = U_ZERO_ERROR;
    char asciiBCP47Tag[LOCALE_NAME_MAX_LENGTH] = {};

    // Convert from names like "en_CA" and "de_DE@collation=phonebook" to "en-CA" and "de-DE-u-co-phonebk".
    (void) uloc_toLanguageTag(locale.getName(), asciiBCP47Tag, UPRV_LENGTHOF(asciiBCP47Tag), FALSE, &status);

    if (U_SUCCESS(status))
    {
        // Need it to be UTF-16, not 8-bit
        // TODO: This seems like a good thing for a helper
        wchar_t bcp47Tag[LOCALE_NAME_MAX_LENGTH] = {};
        int32_t i;
        for (i = 0; i < UPRV_LENGTHOF(bcp47Tag); i++)
        {
            if (asciiBCP47Tag[i] == '\0')
            {
                break;
            }
            else
            {
                // normally just copy the character
                bcp47Tag[i] = static_cast<wchar_t>(asciiBCP47Tag[i]);
            }
        }

        // Ensure it's null terminated
        if (i < (UPRV_LENGTHOF(bcp47Tag) - 1))
        {
            bcp47Tag[i] = L'\0';
        }
        else
        {
            // Ran out of room.
            bcp47Tag[UPRV_LENGTHOF(bcp47Tag) - 1] = L'\0';
        }


        wchar_t windowsLocaleName[LOCALE_NAME_MAX_LENGTH] = {};

        // Note: On Windows versions below 10, there is no support for locale name aliases.
        // This means that it will fail for locales where ICU has a completely different
        // name (like ku vs ckb), and it will also not work for alternate sort locale
        // names like "de-DE-u-co-phonebk".

        // TODO: We could add some sort of exception table for cases like ku vs ckb.

        int length = ResolveLocaleName(bcp47Tag, windowsLocaleName, UPRV_LENGTHOF(windowsLocaleName));

        if (length > 0)
        {
            *buffer = new UnicodeString(windowsLocaleName);
        }
        else
        {
            status = U_UNSUPPORTED_ERROR;
        }
    }
    return status;
}

Win32NumberFormat::Win32NumberFormat(const Locale &locale, UBool currency, UErrorCode &status)
  : NumberFormat(), fCurrency(currency), fFormatInfo(NULL), fFractionDigitsSet(FALSE), fWindowsLocaleName(nullptr)
{
    if (!U_FAILURE(status)) {
        fLCID = locale.getLCID();

        winnmfmt_GetEquivalentWindowsLocaleName(locale, &fWindowsLocaleName);
        // Note: In the previous code, it would look up the LCID for the locale, and if
        // the locale was not recognized then it would get an LCID of 0, which is a
        // synonym for LOCALE_USER_DEFAULT on Windows.
        // If the above method fails, then fWindowsLocaleName will remain as nullptr, and
        // then we will pass nullptr to API GetLocaleInfoEx, which is the same as passing
        // LOCALE_USER_DEFAULT.

        // Resolve actual locale to be used later
        UErrorCode tmpsts = U_ZERO_ERROR;
        char tmpLocID[ULOC_FULLNAME_CAPACITY];
        int32_t len = uloc_getLocaleForLCID(fLCID, tmpLocID, UPRV_LENGTHOF(tmpLocID) - 1, &tmpsts);
        if (U_SUCCESS(tmpsts)) {
            tmpLocID[len] = 0;
            fLocale = Locale((const char*)tmpLocID);
        }

        const wchar_t *localeName = nullptr;

        if (fWindowsLocaleName != nullptr)
        {
            localeName = reinterpret_cast<const wchar_t*>(toOldUCharPtr(fWindowsLocaleName->getTerminatedBuffer()));
        }

        fFormatInfo = (FormatInfo*)uprv_malloc(sizeof(FormatInfo));

        if (fCurrency) {
            getCurrencyFormat(&fFormatInfo->currency, localeName);
        } else {
            getNumberFormat(&fFormatInfo->number, localeName);
        }
    }
}

Win32NumberFormat::Win32NumberFormat(const Win32NumberFormat &other)
  : NumberFormat(other), fFormatInfo((FormatInfo*)uprv_malloc(sizeof(FormatInfo)))
{
    if (fFormatInfo != NULL) {
        uprv_memset(fFormatInfo, 0, sizeof(*fFormatInfo));
    }
    *this = other;
}

Win32NumberFormat::~Win32NumberFormat()
{
    if (fFormatInfo != NULL) {
        if (fCurrency) {
            freeCurrencyFormat(&fFormatInfo->currency);
        } else {
            freeNumberFormat(&fFormatInfo->number);
        }

        uprv_free(fFormatInfo);
    }
    delete fWindowsLocaleName;
}

Win32NumberFormat &Win32NumberFormat::operator=(const Win32NumberFormat &other)
{
    NumberFormat::operator=(other);

    this->fCurrency          = other.fCurrency;
    this->fLocale            = other.fLocale;
    this->fLCID              = other.fLCID;
    this->fFractionDigitsSet = other.fFractionDigitsSet;
    this->fWindowsLocaleName = other.fWindowsLocaleName == NULL ? NULL : new UnicodeString(*other.fWindowsLocaleName);

    const wchar_t *localeName = nullptr;

    if (fWindowsLocaleName != nullptr)
    {
        localeName = reinterpret_cast<const wchar_t*>(toOldUCharPtr(fWindowsLocaleName->getTerminatedBuffer()));
    }

    if (fCurrency) {
        freeCurrencyFormat(&fFormatInfo->currency);
        getCurrencyFormat(&fFormatInfo->currency, localeName);
    } else {
        freeNumberFormat(&fFormatInfo->number);
        getNumberFormat(&fFormatInfo->number, localeName);
    }

    return *this;
}

Win32NumberFormat *Win32NumberFormat::clone() const
{
    return new Win32NumberFormat(*this);
}

UnicodeString& Win32NumberFormat::format(double number, UnicodeString& appendTo, FieldPosition& /* pos */) const
{
    return format(getMaximumFractionDigits(), appendTo, L"%.16f", number);
}

UnicodeString& Win32NumberFormat::format(int32_t number, UnicodeString& appendTo, FieldPosition& /* pos */) const
{
    return format(getMinimumFractionDigits(), appendTo, L"%I32d", number);
}

UnicodeString& Win32NumberFormat::format(int64_t number, UnicodeString& appendTo, FieldPosition& /* pos */) const
{
    return format(getMinimumFractionDigits(), appendTo, L"%I64d", number);
}

void Win32NumberFormat::parse(const UnicodeString& text, Formattable& result, ParsePosition& parsePosition) const
{
    UErrorCode status = U_ZERO_ERROR;
    NumberFormat *nf = fCurrency? NumberFormat::createCurrencyInstance(fLocale, status) : NumberFormat::createInstance(fLocale, status);

    nf->parse(text, result, parsePosition);
    delete nf;
}
void Win32NumberFormat::setMaximumFractionDigits(int32_t newValue)
{
    fFractionDigitsSet = TRUE;
    NumberFormat::setMaximumFractionDigits(newValue);
}

void Win32NumberFormat::setMinimumFractionDigits(int32_t newValue)
{
    fFractionDigitsSet = TRUE;
    NumberFormat::setMinimumFractionDigits(newValue);
}

UnicodeString &Win32NumberFormat::format(int32_t numDigits, UnicodeString &appendTo, const wchar_t *fmt, ...) const
{
    wchar_t nStackBuffer[winnmfmt_STACK_BUFFER_SIZE];
    wchar_t *nBuffer = nStackBuffer;
    va_list args;
    int result;

    nBuffer[0] = 0x0000;

    /* Due to the arguments causing a result to be <= 23 characters (+2 for NULL and minus),
    we don't need to reallocate the buffer. */
    va_start(args, fmt);
    result = _vsnwprintf(nBuffer, winnmfmt_STACK_BUFFER_SIZE, fmt, args);
    va_end(args);

    /* Just to make sure of the above statement, we add this assert */
    U_ASSERT(result >=0);
    // The following code is not used because _vscwprintf isn't available on MinGW at the moment.
    /*if (result < 0) {
        int newLength;

        va_start(args, fmt);
        newLength = _vscwprintf(fmt, args);
        va_end(args);

        nBuffer = NEW_ARRAY(UChar, newLength + 1);

        va_start(args, fmt);
        result = _vsnwprintf(nBuffer, newLength + 1, fmt, args);
        va_end(args);
    }*/

    // vswprintf is sensitive to the locale set by setlocale. For some locales
    // it doesn't use "." as the decimal separator, which is what GetNumberFormatW
    // and GetCurrencyFormatW both expect to see.
    //
    // To fix this, we scan over the string and replace the first non-digits, except
    // for a leading "-", with a "."
    //
    // Note: (nBuffer[0] == L'-') will evaluate to 1 if there is a leading '-' in the
    // number, and 0 otherwise.
    for (wchar_t *p = &nBuffer[nBuffer[0] == L'-']; *p != L'\0'; p += 1) {
        if (*p < L'0' || *p > L'9') {
            *p = L'.';
            break;
        }
    }

    wchar_t stackBuffer[winnmfmt_STACK_BUFFER_SIZE];
    wchar_t *buffer = stackBuffer;
    FormatInfo formatInfo;

    formatInfo = *fFormatInfo;
    buffer[0] = 0x0000;

    const wchar_t *localeName = nullptr;

    if (fWindowsLocaleName != nullptr)
    {
        localeName = reinterpret_cast<const wchar_t*>(toOldUCharPtr(fWindowsLocaleName->getTerminatedBuffer()));
    }

    if (fCurrency) {
        if (fFractionDigitsSet) {
            formatInfo.currency.NumDigits = (UINT) numDigits;
        }

        if (!isGroupingUsed()) {
            formatInfo.currency.Grouping = 0;
        }

        result = GetCurrencyFormatEx(localeName, 0, nBuffer, &formatInfo.currency, buffer, winnmfmt_STACK_BUFFER_SIZE);

        if (result == 0) {
            DWORD lastError = GetLastError();

            if (lastError == ERROR_INSUFFICIENT_BUFFER) {
                int newLength = GetCurrencyFormatEx(localeName, 0, nBuffer, &formatInfo.currency, NULL, 0);

                buffer = NEW_ARRAY(wchar_t, newLength);
                buffer[0] = 0x0000;
                GetCurrencyFormatEx(localeName, 0, nBuffer,  &formatInfo.currency, buffer, newLength);
            }
        }
    } else {
        if (fFractionDigitsSet) {
            formatInfo.number.NumDigits = (UINT) numDigits;
        }

        if (!isGroupingUsed()) {
            formatInfo.number.Grouping = 0;
        }

        result = GetNumberFormatEx(localeName, 0, nBuffer, &formatInfo.number, buffer, winnmfmt_STACK_BUFFER_SIZE);

        if (result == 0) {
            if (GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
                int newLength = GetNumberFormatEx(localeName, 0, nBuffer, &formatInfo.number, NULL, 0);

                buffer = NEW_ARRAY(wchar_t, newLength);
                buffer[0] = 0x0000;
                GetNumberFormatEx(localeName, 0, nBuffer, &formatInfo.number, buffer, newLength);
            }
        }
    }

    appendTo.append((UChar *)buffer, (int32_t) wcslen(buffer));

    if (buffer != stackBuffer) {
        DELETE_ARRAY(buffer);
    }

    /*if (nBuffer != nStackBuffer) {
        DELETE_ARRAY(nBuffer);
    }*/

    return appendTo;
}

U_NAMESPACE_END

#endif /* #if !UCONFIG_NO_FORMATTING */

#endif // U_PLATFORM_USES_ONLY_WIN32_API
// © 2016 and later: Unicode, Inc. and others.
// License & terms of use: http://www.unicode.org/copyright.html
/*
********************************************************************************
*   Copyright (C) 2009-2013, International Business Machines
*   Corporation and others.  All Rights Reserved.
********************************************************************************
*
* File WINTZIMPL.CPP
*
********************************************************************************
*/



#if U_PLATFORM_USES_ONLY_WIN32_API && !UCONFIG_NO_FORMATTING 










#ifndef WIN32_LEAN_AND_MEAN
#   define WIN32_LEAN_AND_MEAN
#endif
#   define VC_EXTRALEAN
#   define NOUSER
#   define NOSERVICE
#   define NOIME
#   define NOMCX

#include <windows.h>

U_NAMESPACE_USE

static UBool getSystemTimeInformation(TimeZone *tz, SYSTEMTIME &daylightDate, SYSTEMTIME &standardDate, int32_t &bias, int32_t &daylightBias, int32_t &standardBias) {
    UErrorCode status = U_ZERO_ERROR;
    UBool result = TRUE;
    BasicTimeZone *btz = (BasicTimeZone*)tz; // we should check type
    InitialTimeZoneRule *initial = NULL;
    AnnualTimeZoneRule *std = NULL, *dst = NULL;

    btz->getSimpleRulesNear(uprv_getUTCtime(), initial, std, dst, status);
    if (U_SUCCESS(status)) {
        if (std == NULL || dst == NULL) {
            bias = -1 * (initial->getRawOffset()/60000);
            standardBias = 0;
            daylightBias = 0;
            // Do not use DST.  Set 0 to all stadardDate/daylightDate fields
            standardDate.wYear = standardDate.wMonth  = standardDate.wDayOfWeek = standardDate.wDay = 
            standardDate.wHour = standardDate.wMinute = standardDate.wSecond    = standardDate.wMilliseconds = 0;
            daylightDate.wYear = daylightDate.wMonth  = daylightDate.wDayOfWeek = daylightDate.wDay =
            daylightDate.wHour = daylightDate.wMinute = daylightDate.wSecond    = daylightDate.wMilliseconds = 0;
        } else {
            U_ASSERT(std->getRule()->getDateRuleType() == DateTimeRule::DOW);
            U_ASSERT(dst->getRule()->getDateRuleType() == DateTimeRule::DOW);

            bias = -1 * (std->getRawOffset()/60000);
            standardBias = 0;
            daylightBias = -1 * (dst->getDSTSavings()/60000);
            // Always use DOW type rule
            int32_t hour, min, sec, mil;
            standardDate.wYear = 0;
            standardDate.wMonth = static_cast<WORD>(std->getRule()->getRuleMonth()) + 1;
            standardDate.wDay = static_cast<WORD>(std->getRule()->getRuleWeekInMonth());
            if (standardDate.wDay < 0) {
                standardDate.wDay = 5;
            }
            standardDate.wDayOfWeek = static_cast<WORD>(std->getRule()->getRuleDayOfWeek()) - 1;

            mil = std->getRule()->getRuleMillisInDay();
            hour = mil/3600000;
            mil %= 3600000;
            min = mil/60000;
            mil %= 60000;
            sec = mil/1000;
            mil %= 1000;

            standardDate.wHour = static_cast<WORD>(hour);
            standardDate.wMinute = static_cast<WORD>(min);
            standardDate.wSecond = static_cast<WORD>(sec);
            standardDate.wMilliseconds = static_cast<WORD>(mil);

            daylightDate.wYear = 0;
            daylightDate.wMonth = static_cast<WORD>(dst->getRule()->getRuleMonth()) + 1;
            daylightDate.wDay = static_cast<WORD>(dst->getRule()->getRuleWeekInMonth());
            if (daylightDate.wDay < 0) {
                daylightDate.wDay = 5;
            }
            daylightDate.wDayOfWeek = static_cast<WORD>(dst->getRule()->getRuleDayOfWeek()) - 1;

            mil = dst->getRule()->getRuleMillisInDay();
            hour = mil/3600000;
            mil %= 3600000;
            min = mil/60000;
            mil %= 60000;
            sec = mil/1000;
            mil %= 1000;

            daylightDate.wHour = static_cast<WORD>(hour);
            daylightDate.wMinute = static_cast<WORD>(min);
            daylightDate.wSecond = static_cast<WORD>(sec);
            daylightDate.wMilliseconds = static_cast<WORD>(mil);
        }
    } else {
        result = FALSE;
    }

    delete initial;
    delete std;
    delete dst;

    return result;
}

static UBool getWindowsTimeZoneInfo(TIME_ZONE_INFORMATION *zoneInfo, const UChar *icuid, int32_t length) {
    UBool result = FALSE;
    UnicodeString id = UnicodeString(icuid, length);
    TimeZone *tz = TimeZone::createTimeZone(id);
    
    if (tz != NULL) {
        int32_t bias;
        int32_t daylightBias;
        int32_t standardBias;
        SYSTEMTIME daylightDate;
        SYSTEMTIME standardDate;

        if (getSystemTimeInformation(tz, daylightDate, standardDate, bias, daylightBias, standardBias)) {
            uprv_memset(zoneInfo, 0, sizeof(TIME_ZONE_INFORMATION)); // We do not set standard/daylight names, so nullify first.
            zoneInfo->Bias          = bias;
            zoneInfo->DaylightBias  = daylightBias;
            zoneInfo->StandardBias  = standardBias;
            zoneInfo->DaylightDate  = daylightDate;
            zoneInfo->StandardDate  = standardDate;

            result = TRUE;
        }
    }

    return result;
}

/*
 * Given the timezone icuid, fill in zoneInfo by calling auxillary functions that creates a timezone and extract the 
 * information to put into zoneInfo. This includes bias and standard time date and daylight saving date.
 */
U_CAPI UBool U_EXPORT2
uprv_getWindowsTimeZoneInfo(TIME_ZONE_INFORMATION *zoneInfo, const UChar *icuid, int32_t length)
{
    if (getWindowsTimeZoneInfo(zoneInfo, icuid, length)) {
        return TRUE;
    } else {
        return FALSE;
    }
}

#endif
